/* xoreos-tools - Tools to help with xoreos development
 *
 * xoreos-tools is the legal property of its developers, whose names
 * can be found in the AUTHORS file distributed with this source
 * distribution.
 *
 * xoreos-tools is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * xoreos-tools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with xoreos-tools. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file
 *  Resolve a hash found in hashed Dragon Age: Origins and Dragon Age 2
 *  archives back to the filename.
 */

#include "src/common/util.h"
#include "src/common/binsearch.h"

#include "src/files_dragonage.h"

typedef Common::BinSearchValue<uint64, const char *> DragonAgeFileHash;

/** All currently known Dragon Age: Origins and Dragon Age 2 file names,
 *  together with their FNV64 hashes.
 *
 *  Based on fnvdump-20110527.filelist, courtesy of tazpn, Thought Process,
 *  4Aces, a friendly donor, and Mephales. fnvdump-20110527.filelist can
 *  be found in fnvdump-20110527.zip on the BioWare Social Network page
 *  for pyGFF (<http://social.bioware.com/project/1936/>).
 *
 *  Note: For the binary search to work, this list needs to stay sorted by
 *  hash value!
 */
static const DragonAgeFileHash kDragonAgeFilesHashes[] = {
	{0x000023C542D3A2D7ULL, "145396997.ogg"},
	{0x000058C598998F7AULL, "vxca_caverns1_mines_ikf_0_0.vlm"},
	{0x0000816D7858B6D5ULL, "6032202_m.fxe"},
	{0x00008AD55410A2C2ULL, "253679411.ogg"},
	{0x0000FB05E0D9A812ULL, "vxlt_0_lowtown_h25_0_0.vlm"},
	{0x000154378DC66970ULL, "6153973_m.fxe"},
	{0x000223914E32B0F7ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.mmh"},
	{0x00026CA3C4FC1740ULL, "w_encounter2_camp_27_0.phy"},
	{0x000293524CE74D78ULL, "211629641.ogg"},
	{0x00029AB70FAA0D5FULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chestcast.gad"},
	{0x0002EE258B36F956ULL, "6094605_m.fxe"},
	{0x0002F52B87C0D0B1ULL, "6129133_m.fxe"},
	{0x00032C77163301CEULL, "mag110_vincento6216162.cl"},
	{0x0003546C439AB2B5ULL, "vxlt_foundry_n_xm_0_0.vlm"},
	{0x0003683F1DE7D362ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45se.phy"},
	{0x0004C93D6A7C586AULL, "art\\vfx\\cinematic\\vfx_d_aoe_shockwave_p.impa.gad"},
	{0x00055A53B019426FULL, "6040938_f.fxe"},
	{0x0005923A728B292BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv1.gad"},
	{0x0005C286F484043EULL, "follower_banter6148479.cl"},
	{0x0006C6D7418AF881ULL, "grey_wardens.stg"},
	{0x00076CA1071F84B6ULL, "mer341_earl6054222.cl"},
	{0x0007701A829773ECULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wi_galowint.phy"},
	{0x0007D4A0FE6A9980ULL, "991581810.ogg"},
	{0x0007E97BA5508DB9ULL, "hlac_w_sundermount_-131074_0y.dds"},
	{0x00080FCC98BF7103ULL, "vxlt_0_lowtown_a2_n_do0_0_0.vlm"},
	{0x00087195F8744707ULL, "gen00fl_anders6111103.cl"},
	{0x00088BA3C500B7F8ULL, "w_encounter2_camp_27_0.rml"},
	{0x0008943BD1D0C3EFULL, "llac_lt_0_lowtown_a2_-131071_0.dds"},
	{0x00098FB70FB0194CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chestcast.evt"},
	{0x000A20B161C3C7A9ULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.mmh"},
	{0x000A46DD2C4F6ECBULL, "art\\vfx\\cinematic\\vfx_ship_wake_p.dura.evt"},
	{0x000A588D5292AE5DULL, "166862660.ogg"},
	{0x000A891F7A385692ULL, "one100_solivitus6060372.cl"},
	{0x000AAA9B2660F775ULL, "gen_im_arm_glv_med_act2_01.uti"},
	{0x000ACB2763CA1E3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_lp.ani"},
	{0x000B0D8BF2AED7EFULL, "qun110_petrice6018526.cl"},
	{0x000B4C8DE698AA43ULL, "364196154.ogg"},
	{0x000B8F3A728F8F90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv1.evt"},
	{0x000BAC3D6A824501ULL, "art\\vfx\\cinematic\\vfx_d_aoe_shockwave_p.impa.evt"},
	{0x000BBC77F664E751ULL, "747383522.ogg"},
	{0x000C004545999560ULL, "da2_prc_seb_en-us.cre"},
	{0x000CBB4520CC7EFDULL, "28876310.ogg"},
	{0x000D393A73708095ULL, "vxht_0_hightown_a2_b9i_0_0.vlm"},
	{0x000D85464CAE0335ULL, "6103310_f.fxe"},
	{0x000DCF332026338FULL, "static_static_p31.psh.fxo.dx11"},
	{0x000E3889E8BE20A7ULL, "6174227_m.fxe"},
	{0x000E6F9A4581E5B7ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgte.mmh"},
	{0x000E739A4581ECC4ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgte.mao"},
	{0x000E859A45820B29ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgte.msh"},
	{0x000EC6B5DD954981ULL, "vxw_bonepitmining_a2_e4_0_0.vlm"},
	{0x000EDD0496EFA897ULL, "6126758_m.fxe"},
	{0x000F3F46A4BF6171ULL, "t3_arm_slr.tnt"},
	{0x000FB978BCBAA7E3ULL, "art\\characters\\creatures\\ragedemon\\animation\\die_lp_v1.ani"},
	{0x000FC4CD04078490ULL, "gen00fl_aveline6122738.cl"},
	{0x000FCA4BC549B6F5ULL, "vxlt_0_lowtown_a2_ehy_0_0.vlm"},
	{0x001080714944FDCEULL, "6053103_m.fxe"},
	{0x0010BE39535B7A6EULL, "6125185_m.fxe"},
	{0x0010C6D0061874BAULL, "vxlt_0_lowtown_a2_gnv_0_0.vlm"},
	{0x00112E739BBA0E9EULL, "hlac_lt_0_lowtown_a2_n_c_65535_3y.dds"},
	{0x001152F33057DCB6ULL, "vxlt_gallowstemplar_5yb_0_0.vlm"},
	{0x00115FDD2C55B78CULL, "art\\vfx\\cinematic\\vfx_ship_wake_p.dura.gad"},
	{0x001225B0EF71BE05ULL, "6126352_m.fxe"},
	{0x00127E26D28D1E4AULL, "6135698_m.fxe"},
	{0x0012D1EFAA79F394ULL, "cth_nb5.tnt"},
	{0x00130334E047807AULL, "vxlt_0_lowtown_a2_n_e1j_0_0.vlm"},
	{0x0013157FE714D0CDULL, "585520048.ogg"},
	{0x0013A918EBCC12E9ULL, "gen00fl_aveline6053600.cl"},
	{0x0013CF421D07FF84ULL, "lt_0_lowtown_docks_a3_184_0.mmh"},
	{0x0013D3421D0806D7ULL, "lt_0_lowtown_docks_a3_184_0.mao"},
	{0x0013E5421D082526ULL, "lt_0_lowtown_docks_a3_184_0.msh"},
	{0x0014063BCDD43F1CULL, "headgrab_flip.dds"},
	{0x001431B76B6EF550ULL, "vxlt_undercitydungeon_188_0_0.vlm"},
	{0x001492FCC3E5B650ULL, "105956307.ogg"},
	{0x0015059462359891ULL, "62300253.ogg"},
	{0x00151E52E1AA7EB5ULL, "vxlt_gallowstemplar_f_5wt_0_0.vlm"},
	{0x00151E9A8EF1E6BEULL, "llac_ht_playershouse_n_855_18.dds"},
	{0x00152DB9F56C1C12ULL, "vxca_cavern1_haunted_ei8_0_0.vlm"},
	{0x001677B70FBA7AA5ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chestcast.ani"},
	{0x00169440A3E8D8A2ULL, "mag330_walter6045963.cl"},
	{0x0016A0EE214B301CULL, "vxdr_0_deeproads_cvk_0_0.vlm"},
	{0x0016A3BB48E30FF6ULL, "vxlt_gallowstemplar_f_47z_0_0.vlm"},
	{0x001708A8FDED2A42ULL, "vxht_mansion2b_61_0_0.vlm"},
	{0x00172027EC26E3CCULL, "mrl150_merrill6088663.cl"},
	{0x001735E21A4F64E6ULL, "6107303_m.fxe"},
	{0x00173F0E5AA2D86DULL, "gen00fl_sebastian6178347.cl"},
	{0x0017523170273A9CULL, "6080771_f.fxe"},
	{0x00175EDD2C5A218AULL, "art\\vfx\\cinematic\\vfx_ship_wake_p.dura.ani"},
	{0x0017C5E3A27DAA6FULL, "amb_scene_sleeping_snore.bnk"},
	{0x0017E32763D4D10FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_lp.evt"},
	{0x0017F0839A77765DULL, "vxca_caverns1_mines_ei3_0_0.vlm"},
	{0x0018115DC791A492ULL, "6082691_m.fxe"},
	{0x001853AB659092B9ULL, "rightarrow.dds"},
	{0x00187AA89ADB7CC6ULL, "96555776.ogg"},
	{0x0018CD6178758E85ULL, "6095519_m.fxe"},
	{0x0018D001972FADF2ULL, "ash_01.dds"},
	{0x001927FFEEF5AB38ULL, "dre000_sibling_has_blight6036454.cl"},
	{0x0019827F0005258CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_basic_1.mao"},
	{0x0019A43D6A8E7550ULL, "art\\vfx\\cinematic\\vfx_d_aoe_shockwave_p.impa.ani"},
	{0x0019C134E21397E3ULL, "llac_ht_mansion2b_818_16.dds"},
	{0x0019DDA1A7A7F169ULL, "411858169.ogg"},
	{0x001A43F139F053F6ULL, "6107164_f.fxe"},
	{0x001A66DBC144C270ULL, "vxht_0_hightown_n_aut_0_0.vlm"},
	{0x001A77B8A9B3C8E7ULL, "6142789_m.fxe"},
	{0x001B250033CCA938ULL, "6156350_m.fxe"},
	{0x001B33289076A5E3ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postc.msh"},
	{0x001B41289076BD3DULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postc.mmh"},
	{0x001B45289076C44EULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postc.mao"},
	{0x001B62B0E0069490ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_ceilingcap.mao"},
	{0x001B66B0E0069BA3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_ceilingcap.mmh"},
	{0x001B6F7DE45B7487ULL, "476576503.ogg"},
	{0x001B74B0E006B3FDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_ceilingcap.msh"},
	{0x001C41F2756B8BCCULL, "vxht_keep_3g3_0_0.vlm"},
	{0x001C4E08C9E0A178ULL, "6141732_m.fxe"},
	{0x001CC21555DC8275ULL, "qun110_petrice6205991.cl"},
	{0x001DBCD0664F6AB4ULL, "vxdr_primevaltomb_6sk_0_0.vlm"},
	{0x001DFB72FA149F85ULL, "vxdr_primevalentrance_2qj_0_2.vlm"},
	{0x001E027EDAED2642ULL, "mag101_alain6052992.cl"},
	{0x001E28CB7B061CC1ULL, "gen00fl_aveline6139566.cl"},
	{0x001EB81553E2D67CULL, "vxht_mansion2b_1pt_0_0.vlm"},
	{0x001EE42763DAF1B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_lp.gad"},
	{0x001F11EF8F4DFC32ULL, "643291818.ogg"},
	{0x001F96FA49E53804ULL, "vxlt_gallowstemplar_5c2_0_0.vlm"},
	{0x001FA1A972DFBC37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor.gad"},
	{0x00202E1FDE123E07ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_act_c.anb"},
	{0x002030400135DD0CULL, "zz_dae_debug6029804.cl"},
	{0x0020A6F624463F27ULL, "vxlt_0_lowtown_docks_a3_n_aqa_0_0.vlm"},
	{0x0020CBD49A7979E8ULL, "350740614.ogg"},
	{0x0021603C5781AEEAULL, "6107075_m.fxe"},
	{0x0021B162A89704AFULL, "class_abilities_rogue.gda"},
	{0x0021F0847D52F7D9ULL, "203975825.ogg"},
	{0x002216C8AC1D30BFULL, "gen00fl_anders6125214.cl"},
	{0x00222348434A4CF2ULL, "37734295.ogg"},
	{0x00225105228C69A2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06b.phy"},
	{0x002277A3B1CFC16BULL, "270529028.ogg"},
	{0x0022A12276ABFD59ULL, "425294863.ogg"},
	{0x00232AB58903105AULL, "460069760.ogg"},
	{0x0023A18C5FB66AE7ULL, "hlac_lt_0_lowtown_n_0_4y.dds"},
	{0x0023D078BCCB77CDULL, "art\\characters\\creatures\\ragedemon\\animation\\die_lp_v1.gad"},
	{0x0023DF1A5429E0FFULL, "w_bonepit_123_0.phy"},
	{0x0023F5A8A2C3D4A5ULL, "35985850.ogg"},
	{0x0024C53F25972C87ULL, "gen00fl_merrill6105618.cl"},
	{0x0024D01A7D5E883CULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.anb"},
	{0x0024DB7DFE428D5BULL, "follower_banter6161217.cl"},
	{0x00254E5A625AC565ULL, "12127855.ogg"},
	{0x002571A327C86AB3ULL, "231971006.ogg"},
	{0x00257CA41C82AE87ULL, "6163595_m.fxe"},
	{0x0025B6A972E44BC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor.evt"},
	{0x0025C426446AD38DULL, "785649323.ogg"},
	{0x00260E22A2D86169ULL, "539311964.ogg"},
	{0x002618C156235BFBULL, "mer240_jansen6050408.cl"},
	{0x0026348732202654ULL, "dae000_player_generic_find6179052.cl"},
	{0x00264D270259CCFCULL, "vxlt_0_lowtown_a2_n_gnc_0_0.vlm"},
	{0x00264EE2AB43CBFDULL, "mrl000_marethari6129284.cl"},
	{0x0026833AB987BD12ULL, "6085791_m.fxe"},
	{0x002731E864A1D4A5ULL, "vxht_keep_64n_0_0.vlm"},
	{0x002769D4B2DEF4CBULL, "ser270_emeric6149349.cl"},
	{0x002783141D4D1A70ULL, "ring_stolen_shadows.dds"},
	{0x002789A3C51B64D5ULL, "w_encounter2_camp_27_0.msh"},
	{0x00278BA3C51B684BULL, "w_encounter2_camp_27_0.mmh"},
	{0x002797A3C51B7CE8ULL, "w_encounter2_camp_27_0.mao"},
	{0x0027A98543B16D73ULL, "vxht_0_hightown_n_c68_0_0.vlm"},
	{0x0027EF45CEE302B9ULL, "vxlt_0_lowtown_n_fh5_0_0.vlm"},
	{0x00281D5D5A0F7AC4ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtb.msh"},
	{0x00282F5D5A0F992DULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtb.mao"},
	{0x0028335D5A0FA066ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtb.mmh"},
	{0x0028433F1E06B016ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45se.mao"},
	{0x00284F3F1E06C4B5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45se.mmh"},
	{0x0028513F1E06C82BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45se.msh"},
	{0x0028D69D941BA8DFULL, "mag291_ntbanter6056606.cl"},
	{0x0028D9B65CC05D32ULL, "llac_dr_dungeon_3216_0.dds"},
	{0x0028DCBE06380CD6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_lgta_0s.dds"},
	{0x0028FA07511EFB6AULL, "6040688_m.fxe"},
	{0x002955EC23DC2AABULL, "385617670.ogg"},
	{0x0029F034AA427EADULL, "608514247.ogg"},
	{0x002A43F92345D39BULL, "vxht_keep_n_6ji_0_0.vlm"},
	{0x002A5AACADE5DCE9ULL, "zz_dae_debug6047641.cl"},
	{0x002A5DDA8B9F46BAULL, "gen_im_trs_krk_junk2_03.uti"},
	{0x002A6F380AD2BC1EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_die_lp_v1.ani"},
	{0x002AE278BB663F5FULL, "6063821_f.fxe"},
	{0x002B50699D0E1EC5ULL, "vxca_caverns1_mines_ikm_0_0.vlm"},
	{0x002C053BD3F9B5B6ULL, "hlac_lt_undercityserialkiller_-1_40z.dds"},
	{0x002C2EAD58458334ULL, "vxdr_primevaltomb_6vh_0_0.vlm"},
	{0x002C3AE74A8B164FULL, "w_bonepit_134_0.phy"},
	{0x002D6AC4E0AC271BULL, "908341707.ogg"},
	{0x002E4DB07FCF0C41ULL, "6019098_m.fxe"},
	{0x002E6A53B49F3802ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_enter.gad"},
	{0x002EC8A206FAE2F4ULL, "hlac_lt_0_lowtown_a2_n_c_131071_7z.dds"},
	{0x002EF7B1AC6FC59AULL, "841232134.ogg"},
	{0x002FD53E9EC6E030ULL, "hlac_ht_chantry_2534_41x.dds"},
	{0x00300C2DA908BAF8ULL, "6071146_f.fxe"},
	{0x00307D97DD7D1651ULL, "vxlt_darktown_4t6_0_1.vlm"},
	{0x00309F019743EC70ULL, "ash_01.mao"},
	{0x0030E89E7D164272ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls107a_0n.dds"},
	{0x0031192CB30D2B31ULL, "vxlt_gallowstemplar_f_3s6_0_0.vlm"},
	{0x00312EE300415E85ULL, "6109697_f.fxe"},
	{0x0031684FE73956BBULL, "art\\characters\\creatures\\strider\\animation\\groundstompenter.ani"},
	{0x0031E31D496F9C10ULL, "vxht_0_hightown_caq_0_0.vlm"},
	{0x0031EE6CC31C7D33ULL, "6108888_m.fxe"},
	{0x0031FFA4634EEC27ULL, "hlac_lt_0_lowtown_a2_-131071_0x.dds"},
	{0x00322A5D5A17E531ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtb.phy"},
	{0x003246D9905DAB42ULL, "hlac_w_encounter3_-131074_0z.dds"},
	{0x00331DB161E6C7DEULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.anb"},
	{0x00332F6BED2345DFULL, "465890089.ogg"},
	{0x0033558038859080ULL, "6083597_m.fxe"},
	{0x0033781207307763ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rockblend_drelief.dds"},
	{0x0033CEA972F0B1BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor.ani"},
	{0x0033D5EA417CAB38ULL, "6087320_m.fxe"},
	{0x003444511391D53CULL, "gen00fl_varric6097185.cl"},
	{0x003448CB51C1FD6FULL, "hlac_w_encounter1_-131071_0z.dds"},
	{0x0034CAA15281391CULL, "hlac_lt_0_lowtown_docks_65537_0z.dds"},
	{0x0034D43D7E63A75DULL, "vxca_caverns1_mines_ec1_0_0.vlm"},
	{0x0034E01A54385504ULL, "w_bonepit_123_0.mmh"},
	{0x0034E41A54385C57ULL, "w_bonepit_123_0.mao"},
	{0x0034F61A54387AA6ULL, "w_bonepit_123_0.msh"},
	{0x0035653DADF62DBAULL, "6039688_m.fxe"},
	{0x0035776F9BEB2863ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphangb.mao"},
	{0x00357B6F9BEB2EB0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphangb.mmh"},
	{0x0035896F9BEB46CAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphangb.msh"},
	{0x00358D53B4A592F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_enter.evt"},
	{0x0035A3C55697ED25ULL, "vxht_keep_61m_0_0.vlm"},
	{0x0035DEF54A285DDFULL, "gen00fl_merrill6121256.cl"},
	{0x0035ECE2F84845A3ULL, "6065910_m.fxe"},
	{0x003655017F0E091BULL, "vxlt_0_lowtown_a2_eor_0_0.vlm"},
	{0x0036B057F20896D1ULL, "73253986.ogg"},
	{0x00371275793CA396ULL, "dae000_gen_turn_in_16178395.cl"},
	{0x00378C58017C5323ULL, "gen00fl_fenris6113868.cl"},
	{0x0037995F215677C2ULL, "vxht_0_hightown_a3_bb6_0_0.vlm"},
	{0x00382B0522927834ULL, "follower_banter6171554.cl"},
	{0x0038BC9B587FFDF0ULL, "gen_im_wep_rog_arc_act0_03.uti"},
	{0x0038BF621EA3E6EFULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs103a_0n.dds"},
	{0x0038DFBD861C552FULL, "dae021_chantry_initiate_mob6145199.cl"},
	{0x0039617FDC3433ADULL, "dr_primevaltomb_658.rml"},
	{0x0039804B1A65E5F7ULL, "6110275_f.fxe"},
	{0x0039C3C329123AAFULL, "vxlt_hangedman_1qv_0_0.vlm"},
	{0x003A987FF637D907ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_sit_loop.gad"},
	{0x003AA0B73B59B883ULL, "qun131cr_taskmaster.utc"},
	{0x003ADDF09E8D22CAULL, "vxlt_hangedman_n_5c_0_0.vlm"},
	{0x003B9CE40977B8FFULL, "vxlt_0_lowtown_fge_0_0.vlm"},
	{0x003BAEF9559038E2ULL, "dre000_vampire6034317.cl"},
	{0x003BDE97682E1B95ULL, "vxdr_primevaltomb_3hn_0_0.vlm"},
	{0x003C53CACDA7F309ULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.mmh"},
	{0x003CA53D9A6655E3ULL, "gen00fl_isabela_init6110275.cl"},
	{0x003CEC1932C40F11ULL, "vxht_brothel_n_5b7_0_0.vlm"},
	{0x003D113E9C145BD9ULL, "vxdr_0_deeproads_ck9_0_0.vlm"},
	{0x003D44EADA8830D4ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.anb"},
	{0x003DDDA762E5E295ULL, "follower_barks6153820.cl"},
	{0x003DFAD8E4BC0FC5ULL, "vxlt_0_lowtown_a2_n_e1a_0_0.vlm"},
	{0x003DFFDAAB99CF7FULL, "318769168.ogg"},
	{0x003E704FE743EEB2ULL, "art\\characters\\creatures\\strider\\animation\\groundstompenter.evt"},
	{0x003E88380AE38F98ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_die_lp_v1.gad"},
	{0x003E9ED2FD062390ULL, "vxht_0_hightown_a3_aqv_0_0.vlm"},
	{0x003EB0C2FD1B4609ULL, "dae170ar_gallows.lst"},
	{0x003EDB8AB8AAA0E1ULL, "vxca_caverns1_mines_im7_0_0.vlm"},
	{0x003F69A68C3B720DULL, "follower_banter6173010.cl"},
	{0x003FA0BFF2377705ULL, "6102797_f.fxe"},
	{0x003FB9BAD9C39DA5ULL, "heallight3.dds"},
	{0x003FCD67074CFE0AULL, "34683419.ogg"},
	{0x0040A1C054A8C586ULL, "vxca_caverns1_mines_k2n_0_1.vlm"},
	{0x004146CCAF787990ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef06.mmh"},
	{0x004154CCAF7891AAULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef06.msh"},
	{0x0041B0A1BD365856ULL, "341869908.ogg"},
	{0x0041F684F72D799CULL, "vxdr_0_deeproads_cen_0_0.vlm"},
	{0x0042D339F761B03EULL, "vxw_bonepitmining_a2_1lg_0_0.vlm"},
	{0x004352EC3AB3D01BULL, "775212395.ogg"},
	{0x00439553B4B1DDE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_enter.ani"},
	{0x0043D265B7DBC76DULL, "6153756_m.fxe"},
	{0x004416E276753B28ULL, "6064165_m.fxe"},
	{0x0044AFBE0634098EULL, "558687102.ogg"},
	{0x0044C829C0BB970AULL, "llac_lt_0_lowtown_docks_a3_131071_0.dds"},
	{0x0044CD4AC08410CAULL, "hlac_ht_keep_n_4519_10z.dds"},
	{0x0044E5F4972DCA79ULL, "vxlt_undercityencounter_wj_0_0.vlm"},
	{0x0044EA1592AE002BULL, "355069094.ogg"},
	{0x004588C2FD212140ULL, "dae170ar_gallows.ncs"},
	{0x00458F4FE74A4245ULL, "art\\characters\\creatures\\strider\\animation\\groundstompenter.gad"},
	{0x0045C39A7447B334ULL, "6150183_m.fxe"},
	{0x00460ED72C76E873ULL, "hlac_lt_0_lowtown_docks_n_-3_6z.dds"},
	{0x0046EFB42AA77BA6ULL, "487737712.ogg"},
	{0x00472F872A501F58ULL, "hlac_lt_0_lowtown_a3_n_c_65535_7x.dds"},
	{0x004794CC37697B5EULL, "6105610_f.fxe"},
	{0x0047C3A7D825B658ULL, "664692917.ogg"},
	{0x0048380522AD0EF5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06b.mmh"},
	{0x00483A0522AD126BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06b.msh"},
	{0x00483B95DCA2FE64ULL, "follower_banter6163562.cl"},
	{0x0048471A82CEC4B7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wi_galowint.mmh"},
	{0x00484B1A82CECBC4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wi_galowint.mao"},
	{0x00485D1A82CEEA29ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wi_galowint.msh"},
	{0x0048644079FF4CC8ULL, "vxca_caverns1_mines_im2_0_0.vlm"},
	{0x00492B1FDE353E08ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_act_c.mmh"},
	{0x004946BF3873F4A1ULL, "hlac_ca_cavern1_haunted_24426_7z.dds"},
	{0x00499694061BFD81ULL, "6099209_m.fxe"},
	{0x0049D139505C1271ULL, "6084751_m.fxe"},
	{0x004A1AE58D64EA5BULL, "vxlt_gallowstemplar_f_5yx_0_0.vlm"},
	{0x004A23EADA928297ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.mmh"},
	{0x004A789F2C9725B4ULL, "mrl150tr_undead_pre_camp.ncs"},
	{0x004AEC911FC78184ULL, "vxlt_gallowstemplar_5r1_0_0.vlm"},
	{0x004B01F98A0A5648ULL, "6175160_m.fxe"},
	{0x004BADA854448F04ULL, "zz_followers_banter6158799.cl"},
	{0x004BE3D9DB10F2A7ULL, "350863035.ogg"},
	{0x004C56DB5143BAF7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3a.msh"},
	{0x004C68DB5143D95AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3a.mao"},
	{0x004C6CDB5143E069ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3a.mmh"},
	{0x004C776E4670B73FULL, "817862058.ogg"},
	{0x004C899D00441B00ULL, "vxca_caverns1_icy_0_0.vlm"},
	{0x004D4D67AAA0B961ULL, "vxca_caverns1_mines_ij6_0_0.vlm"},
	{0x004D4FC6E09A0A41ULL, "vxdr_0_deeproads_f5f_0_0.vlm"},
	{0x004DB7C787E55088ULL, "6119706_m.fxe"},
	{0x004DE4ACE0B35C66ULL, "6202275_m.fxe"},
	{0x004E455F63411FA9ULL, "lt_0_lowtown_docks_a3_n_32_0_lc.mmh"},
	{0x004ECED115B0F2DDULL, "365834235.ogg"},
	{0x004F58265D1712BEULL, "6134994_m.fxe"},
	{0x004F7511D48BD696ULL, "vxlt_hangedman_n_2dl_0_0.vlm"},
	{0x004F857FF64A154DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_sit_loop.ani"},
	{0x004FA9C2FD29ADAEULL, "dae170ar_gallows.are"},
	{0x0050025975EF2870ULL, "41536900.ogg"},
	{0x0050481970F5AAC9ULL, "vxdr_primevaltomb_85r_0_0.vlm"},
	{0x00508237CE26A505ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_homa_0n.dds"},
	{0x0050A1B3B878AC27ULL, "188256296.ogg"},
	{0x0050B43620E6BAAFULL, "zz_one000_barber6179396.cl"},
	{0x0050F9AFE614EB1EULL, "mer300pt_mine_massacre.ncs"},
	{0x0051231C30AD8D15ULL, "694489519.ogg"},
	{0x00512F2AB1A7F42EULL, "dre000_dougal6131164.cl"},
	{0x0051BCB8E7985AACULL, "hlac_lt_gallowstemplar_f_2391_6x.dds"},
	{0x005201702AAA53F3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medp_0t.dds"},
	{0x00524442A9D1DFEDULL, "6156456_m.fxe"},
	{0x0052C99C87B2E2B1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masg_0t.dds"},
	{0x005357D18923E2F9ULL, "vxlt_undercityencounter_ri_0_0.vlm"},
	{0x00542628D92EEBD7ULL, "zz_isa_debug6175468.cl"},
	{0x005499A8EA5ECDA1ULL, "and111_refugee_barks6111541.cl"},
	{0x0054CC1D2D8BDC23ULL, "vxlt_gallowscourtyard_n_651_0_0.vlm"},
	{0x0054DC0EE9550CAFULL, "hlac_lt_gallowscourtyard_-65536_1z.dds"},
	{0x005555E60260158FULL, "6149909_f.fxe"},
	{0x005581F36B3F9B7FULL, "6056840_m.fxe"},
	{0x005596342A4887CDULL, "1008660080.ogg"},
	{0x00560BC55111BB60ULL, "6170277_m.fxe"},
	{0x00566E8F96A06971ULL, "vxlt_gallowstemplar_n_64z_0_0.vlm"},
	{0x0056C4A916F4BD09ULL, "vxht_bartrand_n_5t9_0_0.vlm"},
	{0x0056D61D6BF56548ULL, "vxca_caverns1_mines_ij3_0_0.vlm"},
	{0x0056F4193B5650F7ULL, "gencr_coterie_archer_f.utc"},
	{0x005769F7233A1142ULL, "vxht_0_hightown_a3_aux_0_0.vlm"},
	{0x00583459F857D40AULL, "6171038_m.fxe"},
	{0x0058716FBF7E1AB2ULL, "vxw_woundedcoast_19q_0_0.vlm"},
	{0x00594BEED57AEA48ULL, "46517586.ogg"},
	{0x005978FFC1A1A1E4ULL, "static_static_p38.psh.fxo.dx11"},
	{0x0059CFD316B04011ULL, "vxht_brothel_n_5h5_0_0.vlm"},
	{0x005A21A3B8516287ULL, "gen00fl_isabela6097012.cl"},
	{0x005A67CCCEBE38F3ULL, "vxw_woundedcoast_ow_0_0.vlm"},
	{0x005A8E034B4316E8ULL, "603570096.ogg"},
	{0x005AB7BA3B1F7BF2ULL, "vxht_0_hightown_akc_0_0.vlm"},
	{0x005B141B42784C7DULL, "pro000_merchant6051845.cl"},
	{0x005B394BB87ADD1CULL, "vxht_chantry_2zi_0_0.vlm"},
	{0x005B6E022D75118CULL, "vxca_caverns1_mines_hm3_0_0.vlm"},
	{0x005B7307D7598323ULL, "hlac_lt_warehouse_1_5z.dds"},
	{0x005C19A21A763AA2ULL, "hlac_ht_keep_3732_9y.dds"},
	{0x005D6E0B5EC4ABC5ULL, "hlac_lt_0_lowtown_a3_n_c_1_3z.dds"},
	{0x005D9C3279C1804EULL, "6146919_m.fxe"},
	{0x005EDCD3384AEE78ULL, "1063123879.ogg"},
	{0x0060164A2721259DULL, "gen00fl_fenris6107188.cl"},
	{0x00605338DE9198FDULL, "6173941_m.fxe"},
	{0x0060D4F2876347E5ULL, "gen00fl_isabela_init6075936.cl"},
	{0x00614B3D6E2DFCF6ULL, "vxht_mansion2_3dh_0_0.vlm"},
	{0x006158732238BB65ULL, "hightown_male_elfservnt_mob6170079.cl"},
	{0x0061F8C2D0F0EF60ULL, "hla_ht_0_hightown_n_-65537_0x.dds"},
	{0x00620407007FD66AULL, "vxlt_hangedman_258_0_0.vlm"},
	{0x0062579F2CAB7FE1ULL, "mrl150tr_undead_pre_camp.utt"},
	{0x006290455D8BF87DULL, "zz_approval_debug6003969.cl"},
	{0x0062C08CDAB9905BULL, "llac_lt_gallowsprison_n_-1_4.dds"},
	{0x0062DD44DD7A5E0BULL, "face_face_p11.psh.fxo.dx11"},
	{0x006312C1C7216939ULL, "mag210_arianni6021205.cl"},
	{0x00635454F0AF57AAULL, "220513863.ogg"},
	{0x0063B001D0FD0D81ULL, "follower_banter6148737.cl"},
	{0x0063C62BAC68501DULL, "6032044_m.fxe"},
	{0x0063D1C13964E8B1ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.anb"},
	{0x0063E63ADDE8368EULL, "vxht_keep_5hf_0_0.vlm"},
	{0x0063EBBDED1F9111ULL, "184995067.ogg"},
	{0x00649659C3F1E295ULL, "vxht_brothel_n_4fj_0_0.vlm"},
	{0x0064A78D3DD1D8F9ULL, "mag360_thrask6044272.cl"},
	{0x0064A81F9168A9A3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_cartrock.mao"},
	{0x0064AC1F9168B0F0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_cartrock.mmh"},
	{0x0064BA1F9168C80AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_cartrock.msh"},
	{0x0064C5E24F57D7D9ULL, "vxht_chantry_2y9_0_0.vlm"},
	{0x0064ECD6B213E9D0ULL, "6039692_f.fxe"},
	{0x006550CACDCAF33EULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.anb"},
	{0x0065B877D8AE4DCAULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_hta.mmh"},
	{0x0065DBD4153029C2ULL, "vxht_0_hightown_a3_ary_0_0.vlm"},
	{0x0065FA741FA74228ULL, "gen00fl_isabela6135692.cl"},
	{0x0065FDD170D80137ULL, "vxdr_primevaltomb_75s_0_0.vlm"},
	{0x00660EC079CDEA22ULL, "61611133.ogg"},
	{0x00669CD6DAAB6D97ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medh_0n.dds"},
	{0x00669F569A60D0ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_eyejitter.gad"},
	{0x0066B4A9E5252F87ULL, "241811738.ogg"},
	{0x0067327514E925E0ULL, "vxw_bonepitmining_a2_2ae_0_0.vlm"},
	{0x006792AA493FE3D1ULL, "mer320_hubert6053728.cl"},
	{0x0067A15A76D5D4ECULL, "vxlt_darktown_680_0_0.vlm"},
	{0x0067D04B03B3B284ULL, "vxlt_gallowstemplar_5x7_0_0.vlm"},
	{0x00685E6789F14761ULL, "hlac_w_blightlands_0_8z.dds"},
	{0x00688ED3E3C91180ULL, "one350_variel6088060.cl"},
	{0x0068D9A9C0B45173ULL, "vxw_woundedcoast_dv_0_0.vlm"},
	{0x0068E87C2D072130ULL, "hlac_lt_0_lowtown_docks_a3_-2_1y.dds"},
	{0x00695D58EECEC0BBULL, "783507773.ogg"},
	{0x00696D56E4304C00ULL, "vxca_caverns1_iig_0_0.vlm"},
	{0x0069802899AF7021ULL, "570192197.ogg"},
	{0x0069B5807D6AE7EBULL, "mag291_demon_marethari6166177.cl"},
	{0x0069B597CEA283A7ULL, "follower_banter6163646.cl"},
	{0x0069EC828CA002DDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scroll_d.dds"},
	{0x006A450A690DA4D8ULL, "gen00fl_aveline6132937.cl"},
	{0x006A697CFF2B246FULL, "rdr111dg_martin6015087.cl"},
	{0x006B89BEE40BA807ULL, "vxht_0_hightown_a83_0_0.vlm"},
	{0x006CE432D69E6421ULL, "6151471_m.fxe"},
	{0x006D6E982566601FULL, "vxlt_darktown_5nn_0_0.vlm"},
	{0x006D9E569A66EDAAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_eyejitter.ani"},
	{0x006DC90A686C03E9ULL, "vxht_0_hightown_a3_ajv_0_0.vlm"},
	{0x006E083FB639CD72ULL, "730212427.ogg"},
	{0x006E0DE428101A3DULL, "hlac_lt_0_lowtown_docks_n_65534_1x.dds"},
	{0x006E7029E9E71D13ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_post.phy"},
	{0x006E76AB30B32109ULL, "474878209.ogg"},
	{0x006E7ADC8CBF22DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\2hand_stop.evt"},
	{0x006EEB9EEADE402FULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90r.gad"},
	{0x006EEDFB37CD7CFBULL, "vxht_0_hightown_a3_50n_0_0.vlm"},
	{0x006EF4F94110D12EULL, "800210944.ogg"},
	{0x006F389E05E5209EULL, "zz_vault_debug6203262.cl"},
	{0x007020B763BBAB95ULL, "isa111_isabela6066580.cl"},
	{0x007075E3F275EEEAULL, "vxlt_hangedman_229_0_0.vlm"},
	{0x007092C2BE4035DAULL, "977107726.ogg"},
	{0x0070F2C1396FAB2AULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.mmh"},
	{0x007170C6E9DD6E04ULL, "zz_pride_demon2_v3.cut"},
	{0x00718FC612EA3021ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtc_0n.dds"},
	{0x007255E70EF40E90ULL, "hlac_ca_cavern1_haunted_24426_14y.dds"},
	{0x007327294FED60B8ULL, "6108578_f.fxe"},
	{0x007337BF414DF059ULL, "vxht_chantry_2r8_0_0.vlm"},
	{0x007383067C9300A1ULL, "static_motionvectors_p0.psh.fxo.dx11.assembly"},
	{0x00738B36F0EA9CAFULL, "vxlt_hangedman_1mr_0_0.vlm"},
	{0x0073D5A18DCE4CD6ULL, "6106195_m.fxe"},
	{0x00740502BBA02E5DULL, "770134267.ogg"},
	{0x0074356370E3BFDAULL, "zz_followers_banter6158991.cl"},
	{0x007441EFF481E0BBULL, "6111210_f.fxe"},
	{0x00746CDB7A8B704BULL, "6180817_m.fxe"},
	{0x0074C0CAE6508451ULL, "vxlt_0_lowtown_a3_n_c_gof_0_0.vlm"},
	{0x0074D120E5706286ULL, "llac_lt_0_lowtown_docks_0_4.dds"},
	{0x0074EBBA22DC4403ULL, "706267296.ogg"},
	{0x007573DC8CC5351CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\2hand_stop.gad"},
	{0x00758304FE7B9AA6ULL, "6096323_f.fxe"},
	{0x007595192E71CA4FULL, "gen_im_arm_glv_med_act1_02.uti"},
	{0x0075DE0D1707DB67ULL, "6122923_m.fxe"},
	{0x0076009EEAE482DCULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90r.evt"},
	{0x0076957364FF91A7ULL, "449179346.ogg"},
	{0x0076B1AB8538B911ULL, "art\\levels\\races\\design\\prp_moneypoucha.mao"},
	{0x0076DC3B52D19BFEULL, "gen_im_trs_qun_lgt_cht_r2.uti"},
	{0x0076E13AA25DF37FULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean_3.mao"},
	{0x0076F45F3D383B03ULL, "ico_boots_light_4.dds"},
	{0x00774B86B2AA69F3ULL, "vxw_woundedcoast_ay_0_0.vlm"},
	{0x007786D744097058ULL, "gen00fl_dog6171156.cl"},
	{0x00778B12F8021B48ULL, "scaleform_sf_main_p0.psh.fxo.dx11.assembly"},
	{0x007790F516F8A7C1ULL, "28613663.ogg"},
	{0x0077DCFF367E5004ULL, "6156215_m.fxe"},
	{0x0077DF33D6266480ULL, "vxca_caverns1_ild_0_0.vlm"},
	{0x0077E1D4BA39867AULL, "971430361.ogg"},
	{0x0077F9CBF056B84DULL, "lt_0_lowtown_docks_n_228_0.mmh"},
	{0x0077FDCBF056BE9EULL, "lt_0_lowtown_docks_n_228_0.mao"},
	{0x00780BCBF056D6D3ULL, "lt_0_lowtown_docks_n_228_0.msh"},
	{0x00784C9275FE3720ULL, "6116739_f.fxe"},
	{0x0078690223316BF1ULL, "aip_dalishwarrior2h_base.gda"},
	{0x00786C8F5C547E64ULL, "923631154.ogg"},
	{0x0078AB0C160C8E0EULL, "20275664.ogg"},
	{0x00795B334E0224A6ULL, "hlac_ca_caverns2_5802_39y.dds"},
	{0x007996A044D7BFEDULL, "gen00fl_sebastian6195856.cl"},
	{0x007A54B5533D3769ULL, "931355191.ogg"},
	{0x007A76F7E98E8D63ULL, "287836797.ogg"},
	{0x007AAAFCDC67FE54ULL, "art\\vfx\\combat\\vfx_gib_corm_1.mao"},
	{0x007ADD7E31C57477ULL, "6145212_m.fxe"},
	{0x007AE89ECE8EFE00ULL, "dae000ip_crowd_mf5_drunk1.utp"},
	{0x007B582DF9ED6908ULL, "hlac_lt_0_lowtown_65533_0y.dds"},
	{0x007B6DEB2FC99E24ULL, "zz_followers_banter6159666.cl"},
	{0x007BD7B47D7C3714ULL, "hlac_w_bonepitmining_a2_262149_0z.dds"},
	{0x007C0FE30C4B96F8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad_6.pwk"},
	{0x007C308AC0C24B53ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.impa.evt"},
	{0x007C72DC8CCB521AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\2hand_stop.ani"},
	{0x007CAF8C2DF50C6FULL, "llac_lt_0_lowtown_a2_-2_1.dds"},
	{0x007CF595A213DA26ULL, "564717246.ogg"},
	{0x007D311766C701F6ULL, "fx_2003_diff.dds"},
	{0x007D58186F12B8F1ULL, "1003059896.ogg"},
	{0x007D678F80951E2FULL, "6201276_f.fxe"},
	{0x007DBFC2C99F27C9ULL, "vxw_bonepitdestroyed_a3_1pz_0_0.vlm"},
	{0x007DFCAC32E95D76ULL, "945667779.ogg"},
	{0x007E048F5739EB76ULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180r.evt"},
	{0x007FB02AD24CE1BEULL, "hlac_lt_undercityserialkiller_-1_17z.dds"},
	{0x007FD95535039EC1ULL, "llac_lt_0_lowtown_a2_n_c_-262143_0.dds"},
	{0x008027AB4C54C7CFULL, "vxlt_0_lowtown_a3_n_c_fiw_0_0.vlm"},
	{0x00807F9597EBFD7BULL, "zz_prmpt_debug6198272.cl"},
	{0x008109755BA917F5ULL, "6143267_m.fxe"},
	{0x00813F04F58A5D63ULL, "follower_banter6148036.cl"},
	{0x0081DCA3E840C21AULL, "dae000ip_craft_potions3.utp"},
	{0x0081EFDF6A10064FULL, "430871833.ogg"},
	{0x00824963898C07B2ULL, "hlac_ht_brothel_n_3370_0z.dds"},
	{0x00824DB3CFADAE3DULL, "vxca_caverns1_mines_cjp_0_0.vlm"},
	{0x00824DCCAFB01CABULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef06.phy"},
	{0x0082C2852638E5D5ULL, "hlac_lt_0_lowtown_n_-196608_2x.dds"},
	{0x0082F679FB4421C8ULL, "bloodsmall5.dds"},
	{0x0083298AC0C85D94ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.impa.gad"},
	{0x00836A84DE83E573ULL, "vxlt_backalley_g_1pc_0_1.vlm"},
	{0x008383B7FA146A80ULL, "392929790.ogg"},
	{0x0083FB8F573E4701ULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180r.gad"},
	{0x0084089EEAF0CDF5ULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90r.ani"},
	{0x008450C07A35FDFAULL, "vxlt_0_lowtown_docks_a3_n_cx9_0_0.vlm"},
	{0x0084F63824C73C29ULL, "vxht_0_hightown_8rr_0_0.vlm"},
	{0x00851CF2C9CE94D7ULL, "lt_0_lowtown_docks_n_172_0.mmh"},
	{0x008520F2C9CE9BE4ULL, "lt_0_lowtown_docks_n_172_0.mao"},
	{0x0085258A67D1D237ULL, "gen_im_arm_cht_noba_t2.uti"},
	{0x008526F70450F8ACULL, "vxlt_0_lowtown_docks_a3_n_8rn_0_0.vlm"},
	{0x008532F2C9CEBA49ULL, "lt_0_lowtown_docks_n_172_0.msh"},
	{0x00857E115EDD209DULL, "6126525_m.fxe"},
	{0x0085966D5F06B379ULL, "6108809_f.fxe"},
	{0x0085C878542797FBULL, "w_bonepitdestroyed_a3_221_0.msh"},
	{0x0085DA785427B646ULL, "w_bonepitdestroyed_a3_221_0.mao"},
	{0x0085E6785427CAE5ULL, "w_bonepitdestroyed_a3_221_0.mmh"},
	{0x00860B61FEB6A68EULL, "vxht_0_hightown_n_ccr_0_0.vlm"},
	{0x008704B29759CA4AULL, "6068545_m.fxe"},
	{0x008758B04EAE2632ULL, "445424801.ogg"},
	{0x0087756683D62C13ULL, "hlac_lt_0_lowtown_docks_n_131073_1x.dds"},
	{0x0087EB6935FAC94BULL, "fx_impact02.dds"},
	{0x0087F1AB2ADE4081ULL, "zz_vrc_debug6200241.cl"},
	{0x0088B9E9963871B3ULL, "p_leather_stretcher_01n.dds"},
	{0x0089B4DD7B0CE205ULL, "6107289_m.fxe"},
	{0x008A388AC0CE9672ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.impa.ani"},
	{0x008A52520952911FULL, "vxlt_darktown_5th_0_0.vlm"},
	{0x008AA535D28171B6ULL, "vxht_0_hightown_b5f_0_0.vlm"},
	{0x008ADEF3B7F31885ULL, "40117880.ogg"},
	{0x008AF1451C28D760ULL, "nat320_delilah6065926.cl"},
	{0x008AFC8F574467AFULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180r.ani"},
	{0x008B79B8B8EA7396ULL, "hlac_w_bonepitdestroyed_a3_327686_0x.dds"},
	{0x008BCD2F746C1077ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_3z.dds"},
	{0x008C06B8A5683B53ULL, "6044195_m.fxe"},
	{0x008C167280DE2AADULL, "734590368.ogg"},
	{0x008C8C3436EC955EULL, "zz_one_debug6078374.cl"},
	{0x008CD8C55BDCC4E3ULL, "936640766.ogg"},
	{0x008DAE37B235F02EULL, "6195814_m.fxe"},
	{0x008DD575178A481BULL, "6033690_m.fxe"},
	{0x008E1CC60FF3AEA7ULL, "hlac_lt_0_lowtown_n_-65537_1y.dds"},
	{0x008F3314EFD7647DULL, "vxht_0_hightown_a3_9aw_0_0.vlm"},
	{0x008F3EC50935429AULL, "vxlt_gallowstemplar_f_3n2_0_0.vlm"},
	{0x008F668B23757B40ULL, "vxdr_primevalentrance_1qb_0_0.vlm"},
	{0x009077FBF2F6D343ULL, "vxlt_gallowstemplar_f_2j6_0_0.vlm"},
	{0x0090FDB7D8BB7E37ULL, "933121158.ogg"},
	{0x009144284C3D7864ULL, "art\\levels\\races\\natural\\highlands\\foliage\\_textures\\plc_flowerwild_01_0_sa.dds"},
	{0x00917787EE4A9621ULL, "zz_mag_debug6024441.cl"},
	{0x0091A73F5F1E0459ULL, "139669065.ogg"},
	{0x0091BF191D832111ULL, "52572992.ogg"},
	{0x0093034733B9AF05ULL, "pro000_captain_wright6004379.cl"},
	{0x00932345FF525584ULL, "804509941.ogg"},
	{0x00933E48313AEF98ULL, "6054492_m.fxe"},
	{0x0093600E645B4808ULL, "vxlt_hangedman_ny_0_0.vlm"},
	{0x0093753245C60492ULL, "zz_one_debug6099937.cl"},
	{0x00944E907988FFF8ULL, "lgt000ip_tombs_seal.utp"},
	{0x009484CF10DD101CULL, "ht_brothel_n_2995.anb"},
	{0x009576A331EA1715ULL, "vxdr_primevaltomb_6y0_0_0.vlm"},
	{0x0095A1144932F469ULL, "6142542_f.fxe"},
	{0x0095A342CB8F88CCULL, "6004201_f.fxe"},
	{0x0095F3E06D1E3DD3ULL, "vxlt_0_lowtown_a3_n_c_e1x_0_0.vlm"},
	{0x009638539567815AULL, "zz_mag_debug6051361.cl"},
	{0x009638893AFCE56FULL, "llac_lt_0_lowtown_docks_131072_1.dds"},
	{0x009641308DAB7B0BULL, "hlac_lt_0_lowtown_a2_-2_2y.dds"},
	{0x0096A72447FC63DAULL, "vxlt_undercitydungeon_1il_0_1.vlm"},
	{0x00970A97DFB98FAFULL, "vxw_encounter2_camp_9h_0_0.vlm"},
	{0x00974C6C6B38A073ULL, "vxw_bonepit_1wq_0_0.vlm"},
	{0x0097D60D8640A026ULL, "879358767.ogg"},
	{0x00980394879376F3ULL, "gen00fl_sebastian6179764.cl"},
	{0x00989BBA11F08109ULL, "vxht_keep_4nf_0_0.vlm"},
	{0x0099022DA1CC18BDULL, "vxca_caverns2_97o_0_0.vlm"},
	{0x00990C63C983D7A0ULL, "vxlt_gallowsprison_n_6nq_0_0.vlm"},
	{0x00995C50CABF5B11ULL, "zz_ser_debug6078557.cl"},
	{0x0099BEA9052357D5ULL, "510828892.ogg"},
	{0x009AC7ADF63E9DCFULL, "gen00fl_anders6114456.cl"},
	{0x009ACCA16703180CULL, "and111_lirene6084381.cl"},
	{0x009AEE93A23A420EULL, "vxlt_gallowstemplar_453_0_0.vlm"},
	{0x009B6D9F036786C4ULL, "6112119_f.fxe"},
	{0x009B7DE632D19657ULL, "gen00fl_isabela6098215.cl"},
	{0x009C229007CC7748ULL, "vxca_caverns1_mines_j2t_0_0.vlm"},
	{0x009CBB3849384CA8ULL, "vxlt_0_lowtown_eoq_0_0.vlm"},
	{0x009D032D2157BE0DULL, "art\\characters\\creatures\\dog\\textures\\c_eye_dogamo_0n.dds"},
	{0x009DF1091831FE74ULL, "ca_caverns1_24426_lc.mmh"},
	{0x009E0DE154F2E3F7ULL, "634622183.ogg"},
	{0x009F02A5AC47D0CEULL, "6093687_f.fxe"},
	{0x009F2B704DC1F73EULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.mmh"},
	{0x009F6D9B87B49B54ULL, "vxlt_0_lowtown_a3_n_c_eqk_0_0.vlm"},
	{0x009F886042913287ULL, "hm_desp_refugee.mor"},
	{0x009F9DBF04E19265ULL, "vxlt_undercityserialkiller_yx_0_0.vlm"},
	{0x009FB479EC8BF356ULL, "vxlt_0_lowtown_a2_gpd_0_0.vlm"},
	{0x009FFA353C2BBFDDULL, "181858611.ogg"},
	{0x00A0236E80AB1A6BULL, "187655715.ogg"},
	{0x00A0B3AEA86BDE6EULL, "364075432.ogg"},
	{0x00A0B4B734FC9EB7ULL, "vxlt_0_lowtown_n_gc3_0_0.vlm"},
	{0x00A13C9CE3DACF78ULL, "919390249.ogg"},
	{0x00A176EEA3EE30AAULL, "335667790.ogg"},
	{0x00A18F0B3C06388BULL, "6142536_m.fxe"},
	{0x00A27EBCFD1CFA0CULL, "vxht_0_hightown_a3_c8s_0_0.vlm"},
	{0x00A28114776FB76DULL, "face2_ds1.psh.fxo"},
	{0x00A2A11D888CB373ULL, "vxw_woundedcoast_pz_0_0.vlm"},
	{0x00A2E5DAF8F45440ULL, "6150965_m.fxe"},
	{0x00A3A32D974FB699ULL, "hlac_dr_dungeon_3215_2z.dds"},
	{0x00A3DE0344918E20ULL, "vxht_mansion_n_4zp_0_0.vlm"},
	{0x00A3FD4A94B84CFFULL, "gen00fl_aveline6121895.cl"},
	{0x00A403B8992BD8A5ULL, "hlac_lt_gallowstemplar_n_2394_7y.dds"},
	{0x00A42A06AA0F0496ULL, "1002264399.ogg"},
	{0x00A4393FA2CD410AULL, "framebuffereffects_passthrough_p0.psh.fxo"},
	{0x00A475B61D5B739CULL, "6128019_f.fxe"},
	{0x00A4A248E1C9DEA8ULL, "vxca_cave_1cu_0_0.vlm"},
	{0x00A57204D9206010ULL, "vxht_0_hightown_c9y_0_0.vlm"},
	{0x00A57C74D1AFA82FULL, "vxw_encounter2_camp_4k_0_0.vlm"},
	{0x00A5B31F91A0530BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_cartrock.phy"},
	{0x00A612B4409566CEULL, "lgt261tr_girl_flee.utt"},
	{0x00A659E533E8D4DCULL, "500646603.ogg"},
	{0x00A668F9B865E199ULL, "zz_tbe100_barber6196160.cl"},
	{0x00A6712BC984BE20ULL, "gen00fl_isabela6015842.cl"},
	{0x00A6881676F0EAA5ULL, "ht_mansion_n_3369.rml"},
	{0x00A688BC2BDA9C57ULL, "vxw_encounter2_camp_po_0_0.vlm"},
	{0x00A6DC4A844293AEULL, "363069456.ogg"},
	{0x00A6F5B735259CE2ULL, "6169750_m.fxe"},
	{0x00A727FDA94A999AULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_6x.dds"},
	{0x00A783CBC31345CEULL, "vxht_0_hightown_avh_0_0.vlm"},
	{0x00A7931CB1B4AC39ULL, "hlac_w_bonepitmining_a2_131072_0y.dds"},
	{0x00A7D0C9ABBAD271ULL, "vxlt_0_lowtown_n_i4f_0_0.vlm"},
	{0x00A8A8A1D57A0F18ULL, "900522910.ogg"},
	{0x00A8CCC73C4CB446ULL, "1056624136.ogg"},
	{0x00A8EF2AEB09AF0AULL, "6027863_m.fxe"},
	{0x00A961411374A758ULL, "llac_lt_0_lowtown_docks_a3_n_65533_2.dds"},
	{0x00A9979C36AB80A3ULL, "clf260_harley6088770.cl"},
	{0x00A9AFD9358C5B3FULL, "gen00fl_aveline6115474.cl"},
	{0x00A9C0932CC4168BULL, "vxlt_0_lowtown_n_e1k_0_0.vlm"},
	{0x00A9DF38AD8023EAULL, "llac_ca_caverns1_24424_14.dds"},
	{0x00AA78FABAC09C74ULL, "lgt000cr_merch_turn_in_4.utc"},
	{0x00AAA1E768368AC8ULL, "hf_mother_h1s2.mor"},
	{0x00AAA76218FF99C7ULL, "one000_ilen6037077.cl"},
	{0x00ACEB2859EFFD6CULL, "llac_lt_gallowstemplar_7198_5.dds"},
	{0x00AD0A704DCDFCFDULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.anb"},
	{0x00AD13B8979208DFULL, "6036522_f.fxe"},
	{0x00AD2BE36499C489ULL, "hlac_ht_mansion2_818_47z.dds"},
	{0x00AD669B45C9F231ULL, "366095080.ogg"},
	{0x00AD693D43CF15E9ULL, "6149364_m.fxe"},
	{0x00AD7F3636B4EC5DULL, "vxw_encounter2_camp_rg_0_0.vlm"},
	{0x00AD813E955BE0A3ULL, "gen_im_wep_rog_dua_act1_01.uti"},
	{0x00AE2656DE820BD6ULL, "vxlt_0_lowtown_a2_gwe_0_0.vlm"},
	{0x00AE3A0891DD3150ULL, "al_mage_t06.gda"},
	{0x00AF081B9A229492ULL, "896936067.ogg"},
	{0x00AF48647EAA8D45ULL, "vxlt_0_lowtown_docks_a3_n_cx2_0_0.vlm"},
	{0x00AF963ADA4ECE68ULL, "zz_vis_debug6072534.cl"},
	{0x00AFA5CE6190D4C2ULL, "285934508.ogg"},
	{0x00AFB3B1F71C70D3ULL, "471630697.ogg"},
	{0x00AFC168B9629AE7ULL, "766805326.ogg"},
	{0x00AFEB202D999FD8ULL, "qcr000_fake_siege.cnv"},
	{0x00B015297D731688ULL, "lgt000im_ultra_swatch.uti"},
	{0x00B0346DF0C43D8AULL, "ser120_ghyslain6104804.cl"},
	{0x00B03FAAEC17A740ULL, "zz_hawke_cu_v2.cut"},
	{0x00B0A02881567BDFULL, "55122978.ogg"},
	{0x00B0FB5211166477ULL, "6076015_m.fxe"},
	{0x00B1B5125A3C5B04ULL, "267162787.ogg"},
	{0x00B1F80C3DF81F29ULL, "framebuffereffects_radialblur_p1.vsh.fxo.dx11"},
	{0x00B201986350CBB8ULL, "zz_dae_debug6029413.cl"},
	{0x00B22E1D060B260DULL, "mcr310_argument6074338.cl"},
	{0x00B248CAB0DA2EE4ULL, "6062794_m.fxe"},
	{0x00B2CDC3430BEBE0ULL, "t3_wdg_w23.tnt"},
	{0x00B3760026B2F9FBULL, "6078707_m.fxe"},
	{0x00B399B92598B656ULL, "sf_solid.psh.fxo"},
	{0x00B3CD202D9D55E0ULL, "qcr000_fake_siege.bnk"},
	{0x00B44C71FD328F2DULL, "6165563_f.fxe"},
	{0x00B467B41C7EB8B3ULL, "hlac_lt_0_lowtown_a2_131071_3x.dds"},
	{0x00B4A3181F276F89ULL, "dae000_flemeth6000962.cl"},
	{0x00B4FA991DD0B4D7ULL, "vxw_encounter2_camp_kn_0_0.vlm"},
	{0x00B5A5E5B1622E15ULL, "449708548.ogg"},
	{0x00B619711E5F921EULL, "hlac_ca_caverns1_24427_2x.dds"},
	{0x00B633BE4209E103ULL, "21972141.ogg"},
	{0x00B6367E2E9736AEULL, "fx_smokeghost_fb.mao"},
	{0x00B6CFDB2C83FB72ULL, "hlac_lt_0_lowtown_docks_65534_3x.dds"},
	{0x00B6E6B375FD3801ULL, "6128400_m.fxe"},
	{0x00B706FA2A76F508ULL, "hf_qmq_minion6.mor"},
	{0x00B707B3E4FD87D0ULL, "hlac_lt_0_lowtown_n_131069_0z.dds"},
	{0x00B721411598624EULL, "38737293.ogg"},
	{0x00B72812AB6B6373ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_laundry_n.dds"},
	{0x00B761EF3DCD0B7DULL, "seb221st_flora_drunk.stg"},
	{0x00B7DC4061EE2F60ULL, "978777270.ogg"},
	{0x00B7E4BC9601F82DULL, "llac_dr_0_deeproads_5849_7.dds"},
	{0x00B7F75E445F11A1ULL, "vxw_bonepit_zs_0_0.vlm"},
	{0x00B8FE0C0EEB828AULL, "hlac_ht_0_hightown_a3_-3_0z.dds"},
	{0x00BAB87D21C0C708ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_lashesm1.mao"},
	{0x00BB1FA008F6998DULL, "661855604.ogg"},
	{0x00BB31EAE608F5E2ULL, "gen00fl_aveline_post_patrol.stg"},
	{0x00BB5941F233C494ULL, "6093776_m.fxe"},
	{0x00BB7495421E3AE5ULL, "fxc_bm_drop.mao"},
	{0x00BBA9C31B9A8529ULL, "449793971.ogg"},
	{0x00BBB32D5B1ED2D8ULL, "ser221_gascard6149261.cl"},
	{0x00BBCD80252C5F98ULL, "dae000_oldmangossip6175769.cl"},
	{0x00BBE35A514FCA38ULL, "dae000ip_crowd_lm5_talk5.utp"},
	{0x00BBF74C0D9738B0ULL, "vxlt_0_lowtown_a3_n_c_fep_0_0.vlm"},
	{0x00BC014FCC61F54DULL, "d15_link_smo2.dds"},
	{0x00BC51F0A7FE5639ULL, "6054725_m.fxe"},
	{0x00BC624D0B994B50ULL, "gen_im_trs_rdr_hvy_glv_r0.uti"},
	{0x00BC81F2F918E8F6ULL, "cod_hst_kirk_arch_1a_9.utp"},
	{0x00BCB2946E8F6E83ULL, "vxca_caverns1_mines_e4z_0_0.vlm"},
	{0x00BCB86810BBA9AFULL, "vxca_caverns1_j6j_0_0.vlm"},
	{0x00BD0999AB3F22B3ULL, "hlac_w_encounter2_3_0x.dds"},
	{0x00BD25B3927CE7E3ULL, "hlac_lt_0_lowtown_a3_n_c_1_0y.dds"},
	{0x00BDC26E4E0A8F36ULL, "587554613.ogg"},
	{0x00BE05B440A9E2F7ULL, "lgt261tr_girl_flee.ncs"},
	{0x00BE090CB40E9F4EULL, "fx_dotblur_alpha.mao"},
	{0x00BE9AB8D2C22B21ULL, "hlac_ht_keep_4519_1x.dds"},
	{0x00BF2F5660C98A68ULL, "vxht_brothel_n_5ki_0_0.vlm"},
	{0x00BF6276E1092B0CULL, "vxht_0_hightown_a3_c6q_0_0.vlm"},
	{0x00BF7D2F181687B6ULL, "vxdr_primevaltomb_6iw_0_0.vlm"},
	{0x00BF9D000E6924D7ULL, "and100_anders6111659.cl"},
	{0x00C017EE093F64D5ULL, "hlac_lt_0_lowtown_docks_a3_n_131073_0x.dds"},
	{0x00C031700838BA93ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_001a_0n.dds"},
	{0x00C03512697714C9ULL, "vxca_caverns2prc_90c_0_0.vlm"},
	{0x00C049454793E96EULL, "vxlt_0_lowtown_n_ers_0_0.vlm"},
	{0x00C059FB6FF9951DULL, "gen_aoe_player_detect.ncs"},
	{0x00C0707960B46D0EULL, "348519769.ogg"},
	{0x00C076AB535F4C97ULL, "434639282.ogg"},
	{0x00C0BAA7A2D04764ULL, "6069694_m.fxe"},
	{0x00C0D9DC6A3BB2E8ULL, "lt_0_lowtown_docks_a3_n_185_0.msh"},
	{0x00C0EBDC6A3BD179ULL, "lt_0_lowtown_docks_a3_n_185_0.mao"},
	{0x00C0F294A1CE68A1ULL, "framebuffereffects_composedof_p0.psh.fxo.dx11"},
	{0x00C0F7DC6A3BE512ULL, "lt_0_lowtown_docks_a3_n_185_0.mmh"},
	{0x00C103A8979C3992ULL, "art\\vfx\\materials\\vfx_lightning_add.mao"},
	{0x00C23B44E1B0A897ULL, "6134731_m.fxe"},
	{0x00C29D2511928CAAULL, "6161652_f.fxe"},
	{0x00C36C760FC6CD57ULL, "vxw_encounter2_camp_fi_0_0.vlm"},
	{0x00C389C0C0DE5A64ULL, "vxca_cave_vk_0_0.vlm"},
	{0x00C39E136EF88CAAULL, "hlac_ht_mansion_n_3369_1x.dds"},
	{0x00C46432ECE8D28AULL, "gen00fl_varric6201278.cl"},
	{0x00C53F7F3DE3BE5EULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.impa.evt"},
	{0x00C54A98F1B5FD3EULL, "mer220_hubert6059919.cl"},
	{0x00C58324EB25590AULL, "893824633.ogg"},
	{0x00C67141A4C12B36ULL, "llac_lt_undercityserialkiller_-1_44.dds"},
	{0x00C6BE31D202EDEAULL, "seb321_leliana6180969.cl"},
	{0x00C6C2A94FCCCF39ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containerb_n.dds"},
	{0x00C71967B0A9059DULL, "mcr310_sibling.bnk"},
	{0x00C827D2420536DFULL, "cod_bks_demons_gift.utp"},
	{0x00C84F871496A1F6ULL, "vxw_encounter3_fg_0_0.vlm"},
	{0x00C85783205CB16DULL, "6030495_m.fxe"},
	{0x00C86C327DC50341ULL, "vrc221st_varric_outsidemans.stg"},
	{0x00C88DF27F341000ULL, "6170840_m.fxe"},
	{0x00C8A99BC31AB527ULL, "vxht_mansion2b_1xk_0_0.vlm"},
	{0x00C90B3E44DC4936ULL, "23351594.ogg"},
	{0x00C91C09E1651B73ULL, "6056615_m.fxe"},
	{0x00C9287788D8E11BULL, "vxht_keep_n_63b_0_0.vlm"},
	{0x00C967C7AC031A79ULL, "gen00fl_isabela_init6078166.cl"},
	{0x00C9B34F13A1930DULL, "dae000_bartrand6064039.cl"},
	{0x00CA31ED107CD797ULL, "lt_warehouse_lc.mmh"},
	{0x00CA400E8CB6064FULL, "one000_ilen6059253.cl"},
	{0x00CA569E67A4FB3BULL, "163572584.ogg"},
	{0x00CA5A4297ED4CFEULL, "32561661.ogg"},
	{0x00CA7E490E789194ULL, "141880707.ogg"},
	{0x00CA803739234357ULL, "llac_lt_foundry_n_320_2.dds"},
	{0x00CA8DFCA93F5224ULL, "w_encounter3_126_0.mmh"},
	{0x00CA91FCA93F5977ULL, "w_encounter3_126_0.mao"},
	{0x00CAA3FCA93F77C6ULL, "w_encounter3_126_0.msh"},
	{0x00CAA6D4807D58FDULL, "zz_followers_banter6159045.cl"},
	{0x00CAC3CD99F46EB9ULL, "6051825_m.fxe"},
	{0x00CAFB67B0ACBBA5ULL, "mcr310_sibling.cnv"},
	{0x00CB0151FD359DB3ULL, "vxca_caverns2_9xp_0_0.vlm"},
	{0x00CB40DF8C0FAD1EULL, "6062183_m.fxe"},
	{0x00CC567F3DEA0389ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.impa.gad"},
	{0x00CCA39804DD065FULL, "llac_lt_0_lowtown_docks_n_262143_0.dds"},
	{0x00CCCAE3FAD8BAF8ULL, "vxca_cavern1_haunted_jlj_0_0.vlm"},
	{0x00CCDBB71BDC9997ULL, "vxdr_0_deeproads_cp6_0_0.vlm"},
	{0x00CD5F6E86484F75ULL, "mcr310_argument6075096.cl"},
	{0x00CD63C76C834CD4ULL, "21050686.ogg"},
	{0x00CDCE29EE9D1135ULL, "254702025.ogg"},
	{0x00CDE9D147695B0DULL, "650341972.ogg"},
	{0x00CE19FFC29A6098ULL, "vxlt_undercityserialkiller_14e_0_0.vlm"},
	{0x00CEA4F3B75BA414ULL, "33780184.ogg"},
	{0x00CEB8274D7AC8CFULL, "783738038.ogg"},
	{0x00CEBB22398A01EEULL, "vxlt_0_lowtown_n_eot_0_0.vlm"},
	{0x00CECB840FE8A3BFULL, "vxlt_gallowstemplar_f_46y_0_1.vlm"},
	{0x00CF299E2374B4B3ULL, "art\\vfx\\combat\\vfx_r_arc_front_close_c.impa.evt"},
	{0x00CF6940292F968DULL, "dae100_viscount6118516.cl"},
	{0x00CFB3AC2C4F2CC8ULL, "hlac_lt_undercityserialkiller_-1_2x.dds"},
	{0x00CFC240D3368D63ULL, "llac_w_woundedcoast_65539_0.dds"},
	{0x00CFD56994056DF5ULL, "vxht_mansion_n_6jc_0_0.vlm"},
	{0x00D04C7748044931ULL, "vxlt_undercityencounter_wr_0_0.vlm"},
	{0x00D05641145A790EULL, "6060282_m.fxe"},
	{0x00D06929EA39DB88ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_post.mmh"},
	{0x00D07529EA39F02BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_post.mao"},
	{0x00D075415B522ABBULL, "dre000_sandal6028367.cl"},
	{0x00D08729EA3A0EB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_post.msh"},
	{0x00D0B7A2C1EB6423ULL, "vxht_bartrand_n_64h_0_0.vlm"},
	{0x00D0BCBE4757D070ULL, "hlac_w_bonepit_393217_0y.dds"},
	{0x00D0F5D2D2BCD4A7ULL, "739967534.ogg"},
	{0x00D131641982AACBULL, "253416240.ogg"},
	{0x00D155653B0E77C4ULL, "128504251.ogg"},
	{0x00D23400F5EAFD55ULL, "pro000_wesley6007722.cl"},
	{0x00D2646D7EEC12C2ULL, "451492700.ogg"},
	{0x00D3277F3DEFD2D7ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.impa.ani"},
	{0x00D3334A3EB0DAC8ULL, "hlac_ca_caverns2prc_5802_4x.dds"},
	{0x00D34F7F001D129AULL, "85216000.ogg"},
	{0x00D35BF19EB32229ULL, "943794326.ogg"},
	{0x00D4E5EA685FC594ULL, "51108397.ogg"},
	{0x00D56A73F4923077ULL, "6139457_m.fxe"},
	{0x00D583CF1114A5ACULL, "ht_brothel_n_2995.rml"},
	{0x00D5E17991ADE8E0ULL, "305153607.ogg"},
	{0x00D6229E237AC7F4ULL, "art\\vfx\\combat\\vfx_r_arc_front_close_c.impa.gad"},
	{0x00D632454E3D3A87ULL, "6105623_m.fxe"},
	{0x00D6C164068CBA76ULL, "vxw_encounter3_ch_0_0.vlm"},
	{0x00D71B78B510CDA7ULL, "vxht_mansion2b_1sj_0_0.vlm"},
	{0x00D79B18E2BAD953ULL, "vxlt_0_lowtown_a2_bxp_0_0.vlm"},
	{0x00D7AB60116AF14FULL, "hlac_w_encounter2_camp_3_0z.dds"},
	{0x00D7D40E2FB71EDDULL, "genip_cin_dagger.utp"},
	{0x00D7E2C58ECD0CECULL, "hlac_lt_0_lowtown_docks_0_0x.dds"},
	{0x00D818C2275420A3ULL, "6182028_m.fxe"},
	{0x00D867A0797CD544ULL, "vxlt_0_lowtown_docks_n_bte_0_0.vlm"},
	{0x00D89A6E305B1E4AULL, "6065393_m.fxe"},
	{0x00D8A2B98AD457E8ULL, "gen00fl_anders6097437.cl"},
	{0x00D8AE80BF681676ULL, "lowtown_fem_lowlife_mob6169188.cl"},
	{0x00D8D4CD0C9735DDULL, "vxw_encounter2_camp_cd_0_0.vlm"},
	{0x00D97BEDB4645556ULL, "vxlt_0_lowtown_a2_gdh_0_0.vlm"},
	{0x00D98716771C42CDULL, "ht_mansion_n_3369.anb"},
	{0x00DA339FE5B239CBULL, "gen00fl_varric6105795.cl"},
	{0x00DA51BCA4A7B0E6ULL, "zz_rockwraith2.cut"},
	{0x00DB0790DC82DCF3ULL, "827305660.ogg"},
	{0x00DB3126E5F0B849ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumtile_e.dds"},
	{0x00DB4D940DD2B217ULL, "vxdr_0_deeproads_cm1_0_0.vlm"},
	{0x00DB670DF6645011ULL, "hlac_w_sundermount_196613_0x.dds"},
	{0x00DB7DD77257D2C8ULL, "79440042.ogg"},
	{0x00DB8C1C01B3EE6DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_basic.mmh"},
	{0x00DB901C01B3F4BEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_basic.mao"},
	{0x00DB9E1C01B40CF3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_basic.msh"},
	{0x00DBE5E5CE5A6375ULL, "617597084.ogg"},
	{0x00DC5B4F3D34C77EULL, "6113074_m.fxe"},
	{0x00DCE190B31FB391ULL, "zz_vault_debug6203379.cl"},
	{0x00DCFC3E73E11000ULL, "vxca_caverns1_i1o_0_0.vlm"},
	{0x00DD2CFF2B801A6EULL, "vxlt_0_lowtown_n_ehu_0_0.vlm"},
	{0x00DD319E2380FFD2ULL, "art\\vfx\\combat\\vfx_r_arc_front_close_c.impa.ani"},
	{0x00DD34EF61877BE4ULL, "ave200_donnic6112154.cl"},
	{0x00DD823114D8C796ULL, "isa300ip_trail_marker.ncs"},
	{0x00DDB3009C5B406EULL, "combat_deathblow_1.gda"},
	{0x00DE474685FB8675ULL, "vxht_mansion_n_6mb_0_0.vlm"},
	{0x00DEBE5439FA61B1ULL, "vxlt_undercityencounter_rq_0_0.vlm"},
	{0x00DF14871F9D8769ULL, "hlac_ht_0_hightown_n_-196608_2z.dds"},
	{0x00DF2898BC150827ULL, "6068526_m.fxe"},
	{0x00DF861CAFEB9038ULL, "dae170ip_at_templarhall.utp"},
	{0x00DF95D94556B30DULL, "character_character_p53.psh.fxo.dx11"},
	{0x00DF9A8C19B5A054ULL, "vxlt_0_lowtown_docks_b06_0_0.vlm"},
	{0x00DFA897D37485D4ULL, "vxdr_primevaltomb_507_0_2.vlm"},
	{0x00DFD43AF9AC1E6DULL, "hlac_w_bonepitmining_a2_327683_0x.dds"},
	{0x00DFE058FC99EF7DULL, "physics1.psh.fxo"},
	{0x00E02592B5BA0734ULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.mmh"},
	{0x00E02CC5BE8E97F2ULL, "6156464_m.fxe"},
	{0x00E0B05A238ADD02ULL, "vxht_mansion_n_509_0_0.vlm"},
	{0x00E11D22A642D338ULL, "framebuffereffects_gaussblur.csh.hlsl_p0.csh.fxo.dx11"},
	{0x00E14B3B442F1E60ULL, "64533346.ogg"},
	{0x00E14B3F8AEBA7CEULL, "vxht_0_hightown_ajl_0_0.vlm"},
	{0x00E157EDB60F0618ULL, "dae030cr_lowlife_male.utc"},
	{0x00E188BDE77366F3ULL, "454174884.ogg"},
	{0x00E1BE90F6F4DBC1ULL, "6119695_m.fxe"},
	{0x00E206EE8DF6CCC4ULL, "6110262_m.fxe"},
	{0x00E26C870C25A216ULL, "83078354.ogg"},
	{0x00E27ADBC276A9C7ULL, "lgt000im_ultra_seal.uti"},
	{0x00E30CCB7A75D8EAULL, "240504924.ogg"},
	{0x00E34CF51A415B21ULL, "vxw_bonepit_kp_0_0.vlm"},
	{0x00E3A9E99EA05E54ULL, "115635243.ogg"},
	{0x00E4A28FEDD9E374ULL, "22317869.ogg"},
	{0x00E4C027F5BB3878ULL, "hlac_ca_caverns2prc_5802_49y.dds"},
	{0x00E58D55A706E627ULL, "vxht_mansion2b_1ve_0_0.vlm"},
	{0x00E592963278EC56ULL, "153084807.ogg"},
	{0x00E5973FF86293A2ULL, "vxht_mansion2_1rh_0_0.vlm"},
	{0x00E5982D6B1D0EF1ULL, "dae200_bran6094815.cl"},
	{0x00E5C3667ED32959ULL, "zz_followers_banter6159096.cl"},
	{0x00E5D6DB91FE27B3ULL, "hlac_ca_caverns2prc_5802_38z.dds"},
	{0x00E5D76FC1F49ED3ULL, "vxlt_warehouse_n_137_0_0.vlm"},
	{0x00E60AC1C11C3F73ULL, "dae000ip_crowd_lm5_drunk2.utp"},
	{0x00E6856E3891B424ULL, "vxlt_darktown_5m0_0_0.vlm"},
	{0x00E6904E4112D690ULL, "zz_followers_banter6158742.cl"},
	{0x00E6D38ADD632BEEULL, "6139823_m.fxe"},
	{0x00E739D02828E3EAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernflow.mao"},
	{0x00E73DD02828EA39ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernflow.mmh"},
	{0x00E747D02828FB27ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernflow.msh"},
	{0x00E74AC9B354FA36ULL, "hlac_w_encounter3_0_6y.dds"},
	{0x00E791161A881B6BULL, "6110258_m.fxe"},
	{0x00E7E50BE121CEB3ULL, "vxca_caverns2_9fr_0_0.vlm"},
	{0x00E7F7487F271FACULL, "6120034_m.fxe"},
	{0x00E80C453A760628ULL, "vxht_mansion_n_6p8_0_0.vlm"},
	{0x00E823682FC02AABULL, "835630958.ogg"},
	{0x00E8BBA6DF3B2110ULL, "330837361.ogg"},
	{0x00E934F2BFB74195ULL, "vxdr_primevaltomb_34r_0_0.vlm"},
	{0x00E975738026DF83ULL, "vxlt_hangedman_25u_0_0.vlm"},
	{0x00E9AFF746D146B0ULL, "hlac_lt_0_lowtown_a2_n_c_65536_5x.dds"},
	{0x00E9BB764CD553A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent_override.evt"},
	{0x00E9BF70FFC8CA97ULL, "vxdr_0_deeproads_cj0_0_0.vlm"},
	{0x00EAA07AAAF5473EULL, "follower_banter6161096.cl"},
	{0x00EAE29276794280ULL, "528518661.ogg"},
	{0x00EB5E5BA7306D9BULL, "w_sundermount_127_0.mmh"},
	{0x00EB6A5BA7308238ULL, "w_sundermount_127_0.mao"},
	{0x00EB7C5BA730A085ULL, "w_sundermount_127_0.msh"},
	{0x00EBCE878211075EULL, "6087281_m.fxe"},
	{0x00EC1FEBD0DA5451ULL, "character_character_p71.psh.fxo.dx11.assembly"},
	{0x00EC4D70BCAEC2EFULL, "194588415.ogg"},
	{0x00EC851C01C2552AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_basic.phy"},
	{0x00ECE2EF461E6F0EULL, "vxht_brothel_n_5gk_0_0.vlm"},
	{0x00ECF1F1C4CDD1F1ULL, "610810005.ogg"},
	{0x00ED2E33385F6A9DULL, "hlac_w_sundermount_65532_0y.dds"},
	{0x00ED89EE8BA54C11ULL, "mag170_hugh6080708.cl"},
	{0x00ED98733C5FD856ULL, "mag170_thrask6054044.cl"},
	{0x00EED12AD7A35BB6ULL, "vxlt_undercityserialkiller_wu_0_0.vlm"},
	{0x00EF3579B7A1F2F5ULL, "6129740_m.fxe"},
	{0x00EF582583493D52ULL, "vxlt_warehouse_n_132_0_0.vlm"},
	{0x00EFD2764CD9E6F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent_override.gad"},
	{0x00F02E5A38FD6238ULL, "zz_approval_debug6003553.cl"},
	{0x00F0459079D74E99ULL, "lgt000ip_tombs_seal.ncs"},
	{0x00F0944D9033C450ULL, "6209891_f.fxe"},
	{0x00F12A872D359F3FULL, "601022978.ogg"},
	{0x00F14B6A44FEC78AULL, "437129781.ogg"},
	{0x00F15F62F23269D5ULL, "6119321_m.fxe"},
	{0x00F20DF24F7FDA61ULL, "vxlt_warehouse_hz_0_0.vlm"},
	{0x00F21663825B4BD5ULL, "3032894.ogg"},
	{0x00F3E1A61B7BF02DULL, "hlac_lt_0_lowtown_docks_a3_n_65537_0y.dds"},
	{0x00F4091CEA58AC22ULL, "vxht_mansion2_1ui_0_0.vlm"},
	{0x00F4CD274272A9EDULL, "437099647.ogg"},
	{0x00F4D0F625577F45ULL, "6135823_f.fxe"},
	{0x00F4E9DF86451327ULL, "hlac_lt_0_lowtown_a2_n_65534_3y.dds"},
	{0x00F51A707DD755A3ULL, "t1_skn_h03.tnt"},
	{0x00F54EE5F116E65BULL, "vxlt_gallowstemplar_n_5qx_0_0.vlm"},
	{0x00F56CAAA0734009ULL, "llac_ht_0_hightown_a2_65536_1.dds"},
	{0x00F5800AB1232476ULL, "vxht_keep_6hu_0_0.vlm"},
	{0x00F582A016094124ULL, "lt_0_lowtown_docks_130_0.msh"},
	{0x00F594A016095F8DULL, "lt_0_lowtown_docks_130_0.mao"},
	{0x00F598A0160966C6ULL, "lt_0_lowtown_docks_130_0.mmh"},
	{0x00F5D142694C4E52ULL, "vxht_0_hightown_a2_9b4_0_0.vlm"},
	{0x00F5D1E2AB5F3B9EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_lp.evt"},
	{0x00F62CDC4358B701ULL, "6116492_m.fxe"},
	{0x00F6D3764CE0069FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent_override.ani"},
	{0x00F70C7177EA564CULL, "vxca_caverns2prc_9jh_0_0.vlm"},
	{0x00F715A88D0BFEFCULL, "llac_w_sundermount_65541_0.dds"},
	{0x00F74D06B90B5D63ULL, "vxlt_gallowscourtyard_4v5_0_0.vlm"},
	{0x00F77140E4EE105BULL, "vxlt_gallowstemplar_f_55t_0_0.vlm"},
	{0x00F7E750721CF803ULL, "vxlt_hangedman_22t_0_0.vlm"},
	{0x00F81DB5F105D31EULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.cess.ani"},
	{0x00F8791B26920958ULL, "6164180_m.fxe"},
	{0x00F8DDAFCC3E99F2ULL, "6087634_m.fxe"},
	{0x00F8F744712E9165ULL, "qun110_mugger6055638.cl"},
	{0x00F9395681DD7708ULL, "vxht_mansion2_2e8_0_0.vlm"},
	{0x00F94496704AB55FULL, "vxlt_gallowsprison_n_6bf_0_0.vlm"},
	{0x00F9ACD3D9225525ULL, "6135067_f.fxe"},
	{0x00FA19C70D34DCB5ULL, "vxlt_gallowstemplar_n_5cp_0_0.vlm"},
	{0x00FA23F0FA236DDCULL, "gen00fl_fenris6110842.cl"},
	{0x00FA39DE65375AAFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p.gad"},
	{0x00FAC08757B20BCBULL, "hlac_lt_gallowstemplar_2388_0x.dds"},
	{0x00FAE12D8E5EE448ULL, "143043216.ogg"},
	{0x00FB54CC3814878EULL, "vxht_brothel_n_5jl_0_0.vlm"},
	{0x00FBC2EB665D0988ULL, "ser210_witness6168145.cl"},
	{0x00FBE8E2AB63CEC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_lp.gad"},
	{0x00FBEAF54FD320AFULL, "llac_ht_chantry_n_2534_47.dds"},
	{0x00FC6060D7A7A23CULL, "llac_lt_warehouse_n_1_3.dds"},
	{0x00FCA7F64780A6ACULL, "20731078.ogg"},
	{0x00FD6CF4376FF93EULL, "vxdr_0_deeproads_bcc_0_0.vlm"},
	{0x00FDFF5F81AC5561ULL, "blank_exterior_mmi.gff"},
	{0x00FE1F4B777C72EEULL, "273181833.ogg"},
	{0x00FE3A103E9D5DF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up.gad"},
	{0x00FEA24ADAE10CC1ULL, "6113095_m.fxe"},
	{0x00FEDBF5B28FA8A5ULL, "vxca_caverns1_igt_0_0.vlm"},
	{0x010042B62CC68459ULL, "6064029_m.fxe"},
	{0x01007BB52DB0DB27ULL, "126194848.ogg"},
	{0x010089F49F70EE60ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright_1.mao"},
	{0x0100CF188AE10B2AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_noba_0d.dds"},
	{0x01010C3DE73E7201ULL, "787507502.ogg"},
	{0x010116FFC09E035BULL, "vxht_bartrand_n_3yr_0_0.vlm"},
	{0x01014EDE653D9D5CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p.evt"},
	{0x01016AEF341B616DULL, "crt_cycle_4.dds"},
	{0x0101BB1DBB102CF9ULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_ex.ani"},
	{0x0101C35BA9947B3DULL, "llac_w_encounter1_0_2.dds"},
	{0x01022D02493B455CULL, "hlac_ht_keep_n_4519_17y.dds"},
	{0x0102B9E2AB699D17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_lp.ani"},
	{0x0102D17DA41F91B3ULL, "hlac_dr_dungeon_3217_3y.dds"},
	{0x0102FD8DA26289EFULL, "var_merchant.gda"},
	{0x01037042541B145CULL, "art\\vfx\\combat\\vfx_glb_pois_act_c.impa.ani"},
	{0x01038D3114F9A9A3ULL, "isa300ip_trail_marker.utp"},
	{0x0103C16F6908319EULL, "vxlt_0_lowtown_n_f0f_0_0.vlm"},
	{0x0103DC5535BF10D5ULL, "tools_sound.phy"},
	{0x01044965DB0B970CULL, "840089393.ogg"},
	{0x010457763B36C761ULL, "306192549.ogg"},
	{0x01047016AA64250BULL, "6183957_m.fxe"},
	{0x01048BE67F8AB903ULL, "6208052_m.fxe"},
	{0x0104CE59710B142DULL, "art\\characters\\playercharacter\\humanfemale\\cth_comc.mmh"},
	{0x0104D259710B1B7EULL, "art\\characters\\playercharacter\\humanfemale\\cth_comc.mao"},
	{0x0104D541EF9D93E2ULL, "380246162.ogg"},
	{0x0104E059710B32B3ULL, "art\\characters\\playercharacter\\humanfemale\\cth_comc.msh"},
	{0x01054F103EA39F84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up.evt"},
	{0x01059A30C0243A96ULL, "lowtown_male_alienage1_sta6162776.cl"},
	{0x010635B5F11239EFULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.cess.evt"},
	{0x010640279D0C7646ULL, "925420644.ogg"},
	{0x01068A1CF3E5FB68ULL, "129277583.ogg"},
	{0x0106B470903BDF02ULL, "6127257_m.fxe"},
	{0x0107600D85D70BA9ULL, "267281836.ogg"},
	{0x01077A09B1D5C1DEULL, "mag110cr_vincento.utc"},
	{0x0107814E4746C9B1ULL, "665893479.ogg"},
	{0x01079F0CAD20B885ULL, "hlac_lt_gallowscourtyard_-2_3y.dds"},
	{0x0108229E552FB27CULL, "gen00fl_aveline6128549.cl"},
	{0x0108961DBB160C3BULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_ex.gad"},
	{0x0108A5646CA6D8D4ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_player_victim.gad"},
	{0x0108F8F7A7DC3601ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_pillar.phy"},
	{0x0108F9AC415E162FULL, "6101952_f.fxe"},
	{0x01092892B5DD11EBULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.anb"},
	{0x010942727B270A1FULL, "6106917_f.fxe"},
	{0x01094ED2F608377BULL, "hlac_ht_chantry_2534_51z.dds"},
	{0x010975EE29505833ULL, "388531366.ogg"},
	{0x01097D42541F959EULL, "art\\vfx\\combat\\vfx_glb_pois_act_c.impa.gad"},
	{0x01099AD1672E27CCULL, "mag291ip_chest.utp"},
	{0x0109C6A92A0AA00EULL, "vxht_brothel_n_5mm_0_0.vlm"},
	{0x0109C929D5952D93ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_hammer_r.dds"},
	{0x0109EE462A7BED28ULL, "137590041.ogg"},
	{0x010A0621396A0DA7ULL, "vxlt_0_lowtown_a2_n_gbx_0_0.vlm"},
	{0x010A51105F62AA77ULL, "786950035.ogg"},
	{0x010A8A885C1AFE71ULL, "blood_conj_01_l3.mao"},
	{0x010AD1369C12D9E6ULL, "6136003_f.fxe"},
	{0x010AD870E73A54B5ULL, "ico_boots_light_2.dds"},
	{0x010B0B8055703875ULL, "vxlt_undercitydungeon_189_0_0.vlm"},
	{0x010B15704C7B2438ULL, "124445040.ogg"},
	{0x010B4C474B9A6D01ULL, "llac_ht_0_hightown_n_-1_1.dds"},
	{0x010B6A9895001DF3ULL, "985165113.ogg"},
	{0x010BB61C9A7DE02EULL, "art\\vfx\\combat\\vfx_w_2nd_atk_lng_c.mmh"},
	{0x010BFAEFF9BD075CULL, "473932488.ogg"},
	{0x010C04B79577EFCBULL, "vxlt_0_lowtown_a2_n_byx_0_0.vlm"},
	{0x010C25569EEDD472ULL, "hlac_ht_0_hightown_a2_131073_0x.dds"},
	{0x010C26AC0048FC46ULL, "6114466_m.fxe"},
	{0x010C2D4C4F3C08BCULL, "hlac_lt_0_lowtown_docks_-196610_2z.dds"},
	{0x010C31C1AC38156CULL, "vxdr_0_deeproads_fht_0_0.vlm"},
	{0x010C36B5F116A698ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.cess.gad"},
	{0x010CF3B9076AB940ULL, "6005304_m.fxe"},
	{0x010CFB6EA9C4CECCULL, "38878772.ogg"},
	{0x010D4DD2A485C125ULL, "vxca_caverns1_ijs_0_0.vlm"},
	{0x010D94A85FFFBF90ULL, "physics2.vsh.fxo"},
	{0x010DBE7B21BB4595ULL, "1071404251.ogg"},
	{0x010E431CD61E9AD5ULL, "6113521_m.fxe"},
	{0x010E5D71676F30A7ULL, "6152722_m.fxe"},
	{0x010ECD928D981722ULL, "qun160_vashothlead6053926.cl"},
	{0x010EF1AC336C0B61ULL, "vxlt_warehouse_vx_0_0.vlm"},
	{0x010F1C93C996487AULL, "323927471.ogg"},
	{0x010F56DE6549E875ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p.ani"},
	{0x010FB31DBB1C5C20ULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_ex.evt"},
	{0x010FB4646CAD10B2ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_player_victim.ani"},
	{0x010FEA3EAE113B74ULL, "vxlt_undercitydungeon_1ke_0_0.vlm"},
	{0x010FF7E027BD4539ULL, "llac_ca_cavern1_haunted_24425_11.dds"},
	{0x01100EB95967626CULL, "art\\characters\\playercharacter\\humanmale\\glv_wic.phy"},
	{0x01105842542575BDULL, "art\\vfx\\combat\\vfx_glb_pois_act_c.impa.evt"},
	{0x0110620B13F7727FULL, "998393354.ogg"},
	{0x0110ECD6CE44DD22ULL, "vxht_mansion2_1ck_0_0.vlm"},
	{0x0110FD4DE5F0AC63ULL, "hlac_ca_caverns1_mines_24424_25z.dds"},
	{0x0111225B246837D3ULL, "zz_followers_banter6158646.cl"},
	{0x011177CEC9E16E7BULL, "zz_qcr_debug6052119.cl"},
	{0x0112329FD503175BULL, "vxlt_gallowstemplar_n_5oz_0_0.vlm"},
	{0x011267103EAE537DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up.ani"},
	{0x0112B3C1E0837664ULL, "6169741_m.fxe"},
	{0x0112D7605290F997ULL, "271606913.ogg"},
	{0x0112F681CDA56E93ULL, "6055613_f.fxe"},
	{0x0113663AE69682F0ULL, "vxca_caverns1_d5z_0_0.vlm"},
	{0x0113984428897F07ULL, "254143813.ogg"},
	{0x0114FE5ECB115437ULL, "6121966_m.fxe"},
	{0x0115C75971197BEAULL, "art\\characters\\playercharacter\\humanfemale\\cth_comc.phy"},
	{0x01160BF628F61B90ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww12_galloalcovlo.mao"},
	{0x01160FF628F622A3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww12_galloalcovlo.mmh"},
	{0x01161DF628F63AFDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww12_galloalcovlo.msh"},
	{0x0116820FA6B89848ULL, "6164809_m.fxe"},
	{0x0116E2E0D13866F0ULL, "w_bonepit_56_0.mao"},
	{0x0116E6E0D1386D03ULL, "w_bonepit_56_0.mmh"},
	{0x0116F4E0D138855DULL, "w_bonepit_56_0.msh"},
	{0x0117763EEBBFC653ULL, "6061235_m.fxe"},
	{0x01179A5B873D1AABULL, "llac_lt_0_lowtown_a2_n_c_131071_0.dds"},
	{0x0117BC81CA913247ULL, "llac_lt_gallowstemplar_n_2388_5.dds"},
	{0x0117DC367036B8A1ULL, "vxdr_primevaltomb_5ao_0_0.vlm"},
	{0x0117E8FFFA8C0735ULL, "vxdr_0_deeproads_d36_0_0.vlm"},
	{0x0118CB8CC9ECB419ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf.ani"},
	{0x0118E114D2E36056ULL, "vxht_0_hightown_5eq_0_0.vlm"},
	{0x01190729F386AD1EULL, "6075288_m.fxe"},
	{0x011938249CB28DD0ULL, "hlac_lt_0_lowtown_a2_-262143_0z.dds"},
	{0x011943138E08F266ULL, "6131103_f.fxe"},
	{0x01197959A32D6F34ULL, "hlac_ht_keep_3732_0x.dds"},
	{0x0119D51C9A8A522DULL, "art\\vfx\\combat\\vfx_w_2nd_atk_lng_c.anb"},
	{0x011A444CECB86AD2ULL, "follower_banter6151284.cl"},
	{0x011A462C7F6D69D7ULL, "dae000_flemeth6000681.cl"},
	{0x011A7EDFB28E3E53ULL, "6119099_m.fxe"},
	{0x011B8D1893046BEBULL, "hlac_lt_warehouse_n_65536_2x.dds"},
	{0x011BBFAF967BD9A5ULL, "vxca_caverns1_imr_0_0.vlm"},
	{0x011BFE9FBB55094CULL, "33440783.ogg"},
	{0x011C058829A5A8C3ULL, "6020423_m.fxe"},
	{0x011C7F72C36848BDULL, "6023954_m.fxe"},
	{0x011C8A2DFED965E1ULL, "6045107_m.fxe"},
	{0x011CE19B1CDF3991ULL, "t1_skn_e04.tnt"},
	{0x011D2041B143EA87ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.dura.gad"},
	{0x011D2FEB24028A28ULL, "vrc221_varric6048576.cl"},
	{0x011D8EEA70D464B1ULL, "6054031_m.fxe"},
	{0x011E51B0273AC2BDULL, "gen_im_trs_chn_shd_wdn_r2.uti"},
	{0x011E5C1BA00753F4ULL, "vxlt_undercitydungeon_1nj_0_0.vlm"},
	{0x011E649C3CFBB5C1ULL, "sib111_mother6098960.cl"},
	{0x011ED719DABDAC4DULL, "vxlt_foundry_n_zo_0_0.vlm"},
	{0x011F219404818A80ULL, "6094769_railing1.cut"},
	{0x011FA68CC9F2945BULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf.gad"},
	{0x0120252EBA1CE904ULL, "614766196.ogg"},
	{0x0120B98FCAED64AAULL, "765117665.ogg"},
	{0x0120C9C726429C29ULL, "vxlt_gallowstemplar_5cc_0_0.vlm"},
	{0x01211F3CF1245794ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_landscapes_d.dds"},
	{0x01212FDE27846117ULL, "6096260_m.fxe"},
	{0x012146B77500027AULL, "6173047_m.fxe"},
	{0x0121A329FC9E6433ULL, "vxht_0_hightown_aoz_0_0.vlm"},
	{0x012285FDA3F08232ULL, "p_cyc.mao"},
	{0x01229B6CCC772B03ULL, "mag291_demon_marethari6027832.cl"},
	{0x0122E6E877C12A75ULL, "vxdr_primevaltomb_vy_0_0.vlm"},
	{0x012324591D3E721BULL, "vxlt_undercitydungeon_143_0_0.vlm"},
	{0x0123ABEF6D81B168ULL, "lt_0_lowtown_docks_a3_102_0.msh"},
	{0x0123BDEF6D81CFF9ULL, "lt_0_lowtown_docks_a3_102_0.mao"},
	{0x0123BEC874C14776ULL, "art\\characters\\creatures\\strider\\animation\\attackbackleft.evt"},
	{0x0123C9EF6D81E392ULL, "lt_0_lowtown_docks_a3_102_0.mmh"},
	{0x0123D6BEAE12153CULL, "dae000_mother6121048.cl"},
	{0x01240B75F5623F25ULL, "56645124.ogg"},
	{0x01241541B149F694ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.dura.evt"},
	{0x0125A0BEB7E78CF5ULL, "6071690_m.fxe"},
	{0x0125B39C968B9393ULL, "vxlt_hangedman_n_1t6_0_0.vlm"},
	{0x0126353A6B765499ULL, "108928135.ogg"},
	{0x01264F912C85A688ULL, "zz_ave_debug6056781.cl"},
	{0x0126648C7D6255EAULL, "fen110_anso6094958.cl"},
	{0x0126BD1F9545A718ULL, "hlac_w_woundedcoast_131072_0y.dds"},
	{0x0126C38CC9F8E440ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf.evt"},
	{0x0126D2B58EA776E6ULL, "vxlt_0_lowtown_n_feu_0_0.vlm"},
	{0x01270BEB92F36C78ULL, "vxlt_backalley_g_1li_0_0.vlm"},
	{0x012710F6290497C8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww12_galloalcovlo.phy"},
	{0x01291217D2904F26ULL, "screentexrect.psh.fxo"},
	{0x0129180F4937AB66ULL, "hlac_lt_0_lowtown_n_65537_5y.dds"},
	{0x012984B887928882ULL, "6060450_f.fxe"},
	{0x0129B5C874C5A301ULL, "art\\characters\\creatures\\strider\\animation\\attackbackleft.gad"},
	{0x012A77651B9BC143ULL, "6027523_m.fxe"},
	{0x012A97B5606AF907ULL, "vxht_0_hightown_a2_amt_0_0.vlm"},
	{0x012BA3F48A00FCD2ULL, "w_encounter1_13_0_ll.mmh"},
	{0x012C11884576479EULL, "689008926.ogg"},
	{0x012C144C62775FECULL, "814973217.ogg"},
	{0x012C5D4318A85157ULL, "t3_wdg_w06.tnt"},
	{0x012C7C3A979D63C1ULL, "6076314_m.fxe"},
	{0x012CD86B16F48E02ULL, "44776808.ogg"},
	{0x012D488C85F17E10ULL, "6135985_f.fxe"},
	{0x012D846A8C48DA96ULL, "743929453.ogg"},
	{0x012DC2A9EC61BBB1ULL, "257990337.ogg"},
	{0x012DD090B2310E22ULL, "vxht_mansion2_1iu_0_0.vlm"},
	{0x012DEC623210291DULL, "162851701.ogg"},
	{0x012E07FB96DC6503ULL, "6135627_f.fxe"},
	{0x012EE10AF231C3A4ULL, "vxlt_0_lowtown_docks_a3_n_d3i_0_0.vlm"},
	{0x012F3ECC86606343ULL, "vxw_bonepitmining_a2_6g_0_0.vlm"},
	{0x012F72DEE00CC0A7ULL, "129556009.ogg"},
	{0x012F8582BE3E9466ULL, "931499584.ogg"},
	{0x012FA4122C223AD0ULL, "vxw_sundermount_25l_0_0.vlm"},
	{0x0130774418722DADULL, "vxht_0_hightown_a3_c86_0_0.vlm"},
	{0x0130A10078CD784DULL, "vxlt_gallowstemplar_f_5qf_0_0.vlm"},
	{0x0130B6C874CBC3AFULL, "art\\characters\\creatures\\strider\\animation\\attackbackleft.ani"},
	{0x0130F3E2C8BB8A83ULL, "mrl150tr_mtn_undead_2.ncs"},
	{0x0131E982EFD59C66ULL, "fx_magma.mao"},
	{0x0131F67E87B42EC0ULL, "hlac_ht_mansion2_818_66z.dds"},
	{0x01320D41B15626CDULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.dura.ani"},
	{0x01323EBC024DC36FULL, "347433975.ogg"},
	{0x0132582CF6BB92C9ULL, "22864467.ogg"},
	{0x01327314F7F5EF06ULL, "696922482.ogg"},
	{0x01328371A4AEE3BEULL, "6174774_m.fxe"},
	{0x0132BE3DF07A4E68ULL, "vxht_brothel_n_53a_0_0.vlm"},
	{0x01348E20EB10648DULL, "llac_w_sundermount_-262142_0.dds"},
	{0x0134D85ED2B14B53ULL, "vxlt_0_lowtown_a3_n_c_eje_0_0.vlm"},
	{0x01354BBEDDD6F27FULL, "741108543.ogg"},
	{0x0135803867154B27ULL, "6086852_m.fxe"},
	{0x0135AF1F6174F4E6ULL, "hlac_lt_0_lowtown_a3_n_c_-3_3x.dds"},
	{0x0135E741485C6BAFULL, "nodiffuse.dds"},
	{0x01367970C32F3492ULL, "art\\vfx\\cinematic\\vfx_orsino_swirly_c.impa.gad"},
	{0x01369AB1FF1A209CULL, "vxlt_gallowstemplar_n_687_0_0.vlm"},
	{0x0136B69C77F925B5ULL, "6127027_m.fxe"},
	{0x01372BF5C5D66FECULL, "vxlt_0_lowtown_a2_n_fi7_0_0.vlm"},
	{0x0137BA517C0C76CAULL, "6053929_m.fxe"},
	{0x01381139B6670AC7ULL, "309872747.ogg"},
	{0x0138D596DF2D0C1BULL, "hlac_lt_blackemporium_1055_2y.dds"},
	{0x0138DA73035AEC00ULL, "6208279_m.fxe"},
	{0x01393FBC4F6C8563ULL, "llac_dr_0_deeproads_5842_15.dds"},
	{0x01395E5BA7728F28ULL, "w_sundermount_127_0.rml"},
	{0x013971D880A5993BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v4.evt"},
	{0x013975D2BD66C94FULL, "vxlt_0_lowtown_h2d_0_0.vlm"},
	{0x0139782E93A012BFULL, "mrl210_merrill6117417.cl"},
	{0x0139BDA5A60F96FEULL, "6160645_m.fxe"},
	{0x0139E6708359127FULL, "hlac_w_encounter1_0_0x.dds"},
	{0x013A202C551B2C67ULL, "6037391_m.fxe"},
	{0x013A90A42BB14AD6ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_candle.msh"},
	{0x013A9AA42BB15BD4ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_candle.mmh"},
	{0x013A9EA42BB162E7ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_candle.mao"},
	{0x013AD249563419ECULL, "hlac_w_woundedcoast_-131067_0z.dds"},
	{0x013B386961E62802ULL, "vxw_encounter2_camp_q1_0_0.vlm"},
	{0x013B5596AFED28B1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop.ani"},
	{0x013BC067DF63294CULL, "521157243.ogg"},
	{0x013C1EB8D1D9EAA6ULL, "hlac_w_sundermount_-3_0y.dds"},
	{0x013C3004121031DCULL, "gen00fl_varric6076210.cl"},
	{0x013C30FE5F05C672ULL, "vxca_caverns2prc_7o7_0_0.vlm"},
	{0x013C6AF7691B91CCULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoreflower.phy"},
	{0x013CC0CE9BF11A2DULL, "6167193_m.fxe"},
	{0x013CC50C83C22C59ULL, "hlac_dr_primevalentrance_1726_9x.dds"},
	{0x013CCAA252760A50ULL, "493272999.ogg"},
	{0x013CE099C0D23EB4ULL, "343299574.ogg"},
	{0x013D308FC5883493ULL, "6129333_m.fxe"},
	{0x013D48AA19FCECFDULL, "6098206_m.fxe"},
	{0x013D7C70C33558C9ULL, "art\\vfx\\cinematic\\vfx_orsino_swirly_c.impa.evt"},
	{0x013D9E754324070CULL, "and100_anders6095972.cl"},
	{0x013DF958A812EBC5ULL, "hla_ht_0_hightown_a3_65534_0y.dds"},
	{0x013EFA0404FEDE3FULL, "6089732_m.fxe"},
	{0x013F6AD880A9F97CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v4.gad"},
	{0x013FAD1E5F891E84ULL, "866068617.ogg"},
	{0x01400E7395361879ULL, "866889873.ogg"},
	{0x014029F3C80AD1D5ULL, "mag320_orsino6071237.cl"},
	{0x014031F0B4020620ULL, "mag321_comtesse_de_launcet6064905.cl"},
	{0x01404CB08ABBB3E8ULL, "6173409_m.fxe"},
	{0x0140E4A8649D527BULL, "hlac_lt_0_lowtown_a2_65533_3y.dds"},
	{0x0140E79DE9519F92ULL, "one100dg_lia6064065.cl"},
	{0x0141010CE37BB9B6ULL, "502891069.ogg"},
	{0x0141204AD0571C2EULL, "6081816_m.fxe"},
	{0x01413793AA04C8A3ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkr.ani"},
	{0x01426096AFF35A53ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop.gad"},
	{0x0142724ED0396EC6ULL, "head_flipv2.dds"},
	{0x014287471D0CBAC8ULL, "6122043_m.fxe"},
	{0x0142D330067B50CAULL, "145825638.ogg"},
	{0x0143063D7B021C12ULL, "vxlt_0_lowtown_a2_n_c_hq1_0_0.vlm"},
	{0x0143BD625CA6D3DCULL, "vxca_caverns1_i13_0_0.vlm"},
	{0x0143E2C1298D1582ULL, "6139658_m.fxe"},
	{0x0143F13B387B56FEULL, "172994011.ogg"},
	{0x01442A86476C550DULL, "1005578092.ogg"},
	{0x014482A012A6C1CDULL, "250315779.ogg"},
	{0x01450C8EF110391CULL, "vxlt_gallowstemplar_n_654_0_0.vlm"},
	{0x0145450AEF991EECULL, "gen00fl_dog6167696.cl"},
	{0x014638A8407667C9ULL, "vxw_bonepitmining_a2_1wl_0_1.vlm"},
	{0x014669D880B0167AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v4.ani"},
	{0x0146A1FB7CEC8CFBULL, "zz_one_debug6074012.cl"},
	{0x0146C1998CB491EBULL, "vxlt_undercitydungeon_17j_0_0.vlm"},
	{0x0146C8F5923069A1ULL, "6137202_m.fxe"},
	{0x014759FA6CDB2ED2ULL, "368571035.ogg"},
	{0x0147D629BDC0829AULL, "vxht_0_hightown_7lx_0_0.vlm"},
	{0x014802E14DBEED2CULL, "6036503_f.fxe"},
	{0x014840C89DBC5748ULL, "hlac_w_woundedcoast_5_0x.dds"},
	{0x0148428B34FC7477ULL, "prm000im_evra_belt.uti"},
	{0x01485D96AFF7C0B8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop.evt"},
	{0x0148F34CD24BC17DULL, "gen_im_npc_arm_hum_boo_cotc.uti"},
	{0x014997AA81929C6EULL, "vxw_bonepitmining_a2_n6_0_0.vlm"},
	{0x01499CB31D4534EDULL, "vxw_bonepitmining_a2_gv_0_0.vlm"},
	{0x0149AA4653DC4082ULL, "vxw_encounter2_camp_l2_0_0.vlm"},
	{0x0149BC8C2BDF8395ULL, "zz_one000_crafter6179395.cl"},
	{0x014A7AA3EC6090EFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmred_0.pwk"},
	{0x014AC63026A835E2ULL, "6155292_m.fxe"},
	{0x014AC892977F21FEULL, "830992032.ogg"},
	{0x014AEF5535FB33F0ULL, "tools_sound.msh"},
	{0x014AFD5535FB4B0AULL, "tools_sound.mmh"},
	{0x014B34262CF68E7FULL, "isa311_isabela6092510.cl"},
	{0x014B5D256F89C314ULL, "414734373.ogg"},
	{0x014B6470C3416D38ULL, "art\\vfx\\cinematic\\vfx_orsino_swirly_c.impa.ani"},
	{0x014B6589E1604F1DULL, "vxlt_warehouse_11x_0_0.vlm"},
	{0x014B6DEA47F778AFULL, "6140942_f.fxe"},
	{0x014B92EC3E3A9520ULL, "vxca_caverns2_92x_0_1.vlm"},
	{0x014C22866A4C9443ULL, "vxw_bonepitmining_a2_8i_0_0.vlm"},
	{0x014C4F86A526E65FULL, "vxht_mansion_n_4ab_0_0.vlm"},
	{0x014CC8940247A9A3ULL, "mag370_meredith6046874.cl"},
	{0x014D2410E1EBB334ULL, "6151236_m.fxe"},
	{0x014D2A3785788455ULL, "vxca_cavern1_haunted_e4l_0_0.vlm"},
	{0x014D6AE54C825AA7ULL, "zz_vault_debug6177766.cl"},
	{0x014E078F33247033ULL, "llac_ca_caverns2_5802_0.dds"},
	{0x014E337A0A532782ULL, "hlac_ht_bartrand_n_3374_0x.dds"},
	{0x014EB38A9CF0ACAAULL, "zz_otl_debug6060958.cl"},
	{0x014F4F93AA112E9AULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkr.evt"},
	{0x014FE464E8C779C8ULL, "llac_ht_0_hightown_a2_-131074_0.dds"},
	{0x014FF778E015D973ULL, "vxw_woundedcoast_4f_0_0.vlm"},
	{0x0150E5B9599EB337ULL, "art\\characters\\playercharacter\\humanmale\\glv_wic.mmh"},
	{0x0150E9B9599EBA44ULL, "art\\characters\\playercharacter\\humanmale\\glv_wic.mao"},
	{0x0150FBB9599ED8A9ULL, "art\\characters\\playercharacter\\humanmale\\glv_wic.msh"},
	{0x0150FD3C10B690A1ULL, "382643223.ogg"},
	{0x0152502A0B46B08DULL, "vxlt_0_lowtown_a2_n_c_ejf_0_0.vlm"},
	{0x01527981B94CC797ULL, "300814449.ogg"},
	{0x015292F002AC074DULL, "one220_yevhen6115738.cl"},
	{0x0152A666450ED107ULL, "6122817_m.fxe"},
	{0x0152A77A4A8E2442ULL, "vxht_brothel_n_5b8_0_0.vlm"},
	{0x015303F62E2578D0ULL, "mag360_samson6044901.cl"},
	{0x01537B215C0B0B2CULL, "gen_im_npc_arm_hum_glv_cota.uti"},
	{0x01539528E0ACCA95ULL, "335745138.ogg"},
	{0x0153E35FA48885EDULL, "lt_0_lowtown_docks_a3_n_252_0.mmh"},
	{0x0153E75FA4888C3EULL, "lt_0_lowtown_docks_a3_n_252_0.mao"},
	{0x0153F55FA488A473ULL, "lt_0_lowtown_docks_a3_n_252_0.msh"},
	{0x0153FF685A8EA83BULL, "vxw_bonepit_1wh_0_1.vlm"},
	{0x0154416ADAEC5989ULL, "hlac_lt_darktown_-1_84y.dds"},
	{0x0154A6CC49B1A7F4ULL, "llac_lt_backalley_g_65536_1.dds"},
	{0x0154A6FFFD2A70BFULL, "pro000_refugee_angry6052171.cl"},
	{0x0154E82043F10FC9ULL, "vxca_cavern1_haunted_j29_0_0.vlm"},
	{0x015525854AAA9523ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_p.anb"},
	{0x0155637D0F5B1E05ULL, "vxlt_backalley_g_1lp_0_0.vlm"},
	{0x01558E93278468BAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_noba_0t.dds"},
	{0x0155E851C3E4D5D2ULL, "6161619_m.fxe"},
	{0x015646F346309F1EULL, "359616565.ogg"},
	{0x01564E93AA174B8DULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkr.gad"},
	{0x01565EAB438EE962ULL, "361616212.ogg"},
	{0x0156ACEFFC96E0DEULL, "follower_banter6170645.cl"},
	{0x0156C0E31A3D43E1ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower2_varric_.gad"},
	{0x01574C0FBAFE4C34ULL, "308458129.ogg"},
	{0x0157DEF065A8C45FULL, "6124741_m.fxe"},
	{0x01580E0BE72558EEULL, "667286197.ogg"},
	{0x01580E900F3B4D6DULL, "vxw_bonepitmining_a2_bu_0_0.vlm"},
	{0x01581C2345D25902ULL, "vxw_encounter2_camp_g3_0_0.vlm"},
	{0x015828D6E6F975BAULL, "vxlt_gallowstemplar_n_6sd_0_0.vlm"},
	{0x0158602D2F2D1E87ULL, "dae000_fenarel6147434.cl"},
	{0x01588EC1508640B9ULL, "vxca_caverns2prc_9fe_0_0.vlm"},
	{0x0158975ABFE2A788ULL, "341775883.ogg"},
	{0x01596A2F065A4E59ULL, "6175998_m.fxe"},
	{0x0159BBC07253907BULL, "6139765_m.fxe"},
	{0x0159C07577EB1727ULL, "zz_mrl_debug6184211.cl"},
	{0x0159C259FA524382ULL, "qun120cr_javaris6044137.cl"},
	{0x0159D766D356679DULL, "vxlt_warehouse_14w_0_0.vlm"},
	{0x015AC163971CFEDFULL, "vxht_mansion_n_4dm_0_0.vlm"},
	{0x015AC46B618D763EULL, "dae000_fem_priest_mob.cnv"},
	{0x015AD8866A0CB8BEULL, "dae000_meredith6092581.cl"},
	{0x015B2F802EF67489ULL, "vxdr_dungeon_1lv_0_0.vlm"},
	{0x015B3CCE77F21AD2ULL, "vxlt_0_lowtown_a3_n_c_epb_0_0.vlm"},
	{0x015B4000AED2E468ULL, "6060508_f.fxe"},
	{0x015B80FF928D3651ULL, "6077843_m.fxe"},
	{0x015B9B790D8EF9FBULL, "6015998_f.fxe"},
	{0x015BDDBDE8EB0F3FULL, "6087532_m.fxe"},
	{0x015C16CF19E60A0EULL, "6166557_m.fxe"},
	{0x015C67A87BD06C26ULL, "art\\characters\\creatures\\spider\\animation\\kd_exit.evt"},
	{0x015C8246C27F39D9ULL, "6112154_f.fxe"},
	{0x015C8362BAC9B730ULL, "6089825_m.fxe"},
	{0x015C905A268648A3ULL, "655447868.ogg"},
	{0x015CC1E31A41B08FULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower2_varric_.ani"},
	{0x015CC58A36B8EE0FULL, "303257207.ogg"},
	{0x015D2A64172ADB61ULL, "728034344.ogg"},
	{0x015E49AC778A6946ULL, "lgt000cr_kirk_turn_in_6.utc"},
	{0x015E5725A2A51C93ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_bnt_dblw_lp.ani"},
	{0x015E6955D20BF1F3ULL, "vxw_woundedcoast_1i_0_0.vlm"},
	{0x015E913F0398E003ULL, "gen00fl_fenris6126554.cl"},
	{0x015EA66B61912C4EULL, "dae000_fem_priest_mob.bnk"},
	{0x015F1AC8C7FB4F41ULL, "vxht_bartrand_n_642_0_0.vlm"},
	{0x015F567CEE7DDF99ULL, "ht_0_hightown_a2_16_0.anb"},
	{0x015F6D1FFF80DCD5ULL, "vxlt_hangedman_n_258_0_0.vlm"},
	{0x0160673FAAA43040ULL, "vxlt_backalley_n_1eo_0_0.vlm"},
	{0x016083DF8E752897ULL, "156249868.ogg"},
	{0x0160AF47C7CCF9C6ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potc_0.pwk"},
	{0x0160C206FD3CC90DULL, "vxlt_0_lowtown_a2_n_c_eoe_0_0.vlm"},
	{0x0160D35F5ACBCE78ULL, "vxlt_backalley_g_1xe_0_0.vlm"},
	{0x016119573C843CC2ULL, "vxht_brothel_n_5e9_0_0.vlm"},
	{0x01611DFD3C29933BULL, "292046921.ogg"},
	{0x0162350A86D3E685ULL, "231316768.ogg"},
	{0x01627EA87BD4FF71ULL, "art\\characters\\creatures\\spider\\animation\\kd_exit.gad"},
	{0x0162A98FF8EDD571ULL, "vxdr_primevaltomb_535_0_0.vlm"},
	{0x0162BA1C952C2741ULL, "vfx.mat.dx11"},
	{0x0162E700DEE7830DULL, "6075770_f.fxe"},
	{0x016302854AB6974CULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_p.mmh"},
	{0x0163654A0994EE9BULL, "vxca_caverns1_ckh_0_0.vlm"},
	{0x01637CF5EFAA18ADULL, "llac_lt_0_lowtown_a2_65533_2.dds"},
	{0x0163BC8EB0254CF1ULL, "6132888_f.fxe"},
	{0x0163BEAEB7598F34ULL, "6103451_m.fxe"},
	{0x0163EBC0AF8EAF3EULL, "vxlt_blackemporium_vs_0_0.vlm"},
	{0x0164093038FC4A80ULL, "or_db_sh.mao"},
	{0x01641BFF90FB9B9AULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffc.msh"},
	{0x016429FF90FBB3D3ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffc.mao"},
	{0x01642DFF90FBBAE0ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffc.mmh"},
	{0x0164748835D1EBCFULL, "dae000_flemeth6013252.cl"},
	{0x0164946BBE846A10ULL, "802140597.ogg"},
	{0x0164BA7A42F0E866ULL, "dae000_pc_home_ambients6208169.cl"},
	{0x0164F78A22ECDA7CULL, "hlac_lt_backalley_n_65537_1z.dds"},
	{0x0165096F67A1D36DULL, "vxdr_0_deeproads_cui_0_2.vlm"},
	{0x016526CBE831CB66ULL, "ave121_brennan6105896.cl"},
	{0x01653C2D20505A14ULL, "t3_tra.tnt"},
	{0x01653C6C0ED4E917ULL, "root_tree.mao"},
	{0x01670065185A3B7CULL, "45067022.ogg"},
	{0x016770C27E8B2BF5ULL, "hlac_w_bonepit_327680_0y.dds"},
	{0x0167B3E7116CDA51ULL, "dae000_arishok6067932.cl"},
	{0x01681FE80856A25AULL, "dae000ip_crowd_lm3_talk5.utp"},
	{0x0168BC5FA3D6E214ULL, "6090037_f.fxe"},
	{0x0168C815AB70C561ULL, "vxlt_blackemporium_15x_0_0.vlm"},
	{0x01690317E01116A9ULL, "vxlt_gallowstemplar_5xf_0_0.vlm"},
	{0x01697FA87BDB1F1FULL, "art\\characters\\creatures\\spider\\animation\\kd_exit.ani"},
	{0x01698A186187E7FAULL, "vxlt_house_dr_0_0.vlm"},
	{0x0169F0A12783856AULL, "vxlt_0_lowtown_a2_n_ep4_0_1.vlm"},
	{0x016A767E35B1F4C1ULL, "6165964_m.fxe"},
	{0x016AE2C9204E10FCULL, "6185501_m.fxe"},
	{0x016B01DB296C5B79ULL, "386194802.ogg"},
	{0x016B1503D25218B0ULL, "vxca_caverns2_7k0_0_0.vlm"},
	{0x016B261A3C025D52ULL, "al_mage_t04.gda"},
	{0x016BB729C376BC8FULL, "llac_lt_0_lowtown_docks_a3_n_65532_0.dds"},
	{0x016BFDF7A830BC3DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_pillar.mao"},
	{0x016C01F7A830C376ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_pillar.mmh"},
	{0x016C0BF7A830D474ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_pillar.msh"},
	{0x016C63E2FBDF7288ULL, "gen00fl_varric6075642.cl"},
	{0x016CC10D8B063C8AULL, "vxlt_gallowstemplar_47y_0_0.vlm"},
	{0x016CE6D0BF6A5452ULL, "llac_lt_gallowstemplar_n_2382_2.dds"},
	{0x016CEE1F505D15A5ULL, "hightown_male_noble2_mob6170931.cl"},
	{0x016D40007ACA0E2DULL, "6071715_m.fxe"},
	{0x016E9ECE2EC33ED0ULL, "hlac_lt_0_lowtown_docks_a3_-131074_5y.dds"},
	{0x016EB1C67BA66617ULL, "hlac_lt_0_lowtown_-65537_0y.dds"},
	{0x016EEECECD055145ULL, "398046902.ogg"},
	{0x016F50089BF7F9ABULL, "6124723_m.fxe"},
	{0x016F53337D647BE3ULL, "el_lr_middle_dusk.mmh"},
	{0x016FB68A0D24FF2FULL, "vxw_encounter2_camp_ny_0_0.vlm"},
	{0x01701625ACA080B8ULL, "mag100_cullen6082314.cl"},
	{0x0170A76A544CB240ULL, "442011668.ogg"},
	{0x0170C059C9910CE8ULL, "vxw_sundermount_1zn_0_0.vlm"},
	{0x0170DA050B73880AULL, "525232403.ogg"},
	{0x01710866529B8C73ULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.dura.ani"},
	{0x017123C29D63D195ULL, "hlac_ht_brothel_n_3373_2z.dds"},
	{0x0171C99841C58323ULL, "hlac_lt_0_lowtown_-196609_4z.dds"},
	{0x0171D726FB8B071BULL, "vxca_caverns1_cnk_0_0.vlm"},
	{0x017255CCC9D97F0FULL, "z_dust8_white.dds"},
	{0x01727B35EEEC28AAULL, "vxht_bartrand_n_1pn_0_0.vlm"},
	{0x01727CDE65EB9EC0ULL, "423126312.ogg"},
	{0x01728B95A81D1730ULL, "w_blightlands_60_0.mao"},
	{0x01728F95A81D1E43ULL, "w_blightlands_60_0.mmh"},
	{0x01729D95A81D359DULL, "w_blightlands_60_0.msh"},
	{0x01732E25A2B7333DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_bnt_dblw_lp.gad"},
	{0x017335CB52ABB632ULL, "438398380.ogg"},
	{0x01735808884E5626ULL, "545792343.ogg"},
	{0x0173987F1EA20DCBULL, "vxlt_0_lowtown_a2_ghu_0_0.vlm"},
	{0x0173BF968DDE5021ULL, "vxw_bonepit_9n_0_0.vlm"},
	{0x01742EAF0D34630FULL, "ss_rogue1.cut"},
	{0x017532D860541CD9ULL, "6064589_f.fxe"},
	{0x01756E0A6B72CB2DULL, "dae000_elegant6166890.cl"},
	{0x0175997AB8787D5BULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_021a_0s.dds"},
	{0x0175DEE1BA4DEBF1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masb_0s.dds"},
	{0x0176539F5CC6F271ULL, "vxw_bonepit_p1_0_0.vlm"},
	{0x01766FD5FDD5856DULL, "6097405_m.fxe"},
	{0x0176ABDD0A2C3055ULL, "985918524.ogg"},
	{0x0176E4335BA19667ULL, "vxlt_0_lowtown_a2_n_c_e3c_0_0.vlm"},
	{0x017733D74266EED0ULL, "35235318.ogg"},
	{0x01779AB25B86DC19ULL, "zz_dre_debug6025538.cl"},
	{0x017933259222BDD3ULL, "vxlt_0_lowtown_docks_n_b61_0_1.vlm"},
	{0x01798F29B4D7143AULL, "215721774.ogg"},
	{0x0179D043F0359CE7ULL, "vxca_caverns2_7o7_0_0.vlm"},
	{0x017A0646B11CC6CFULL, "284604047.ogg"},
	{0x017A12D2FD58FCB9ULL, "gen_im_trs_apo_lgt_boo_r2.uti"},
	{0x017A5A813B384639ULL, "719895107.ogg"},
	{0x017A93E5B44E5600ULL, "vxlt_0_lowtown_euw_0_0.vlm"},
	{0x017B18C9027FC4C1ULL, "9312095.ogg"},
	{0x017B589D4D918382ULL, "325686806.ogg"},
	{0x017BA3E19313D603ULL, "6208243_m.fxe"},
	{0x017BD9E2089CE89EULL, "mcr371_idunna6087279.cl"},
	{0x017C0A4F0C248002ULL, "vxlt_0_lowtown_a3_n_c_ghh_0_0.vlm"},
	{0x017C13CC7F9D9E5BULL, "llac_dr_slavers_3214_9.dds"},
	{0x017C288650533DEBULL, "1039700506.ogg"},
	{0x017C41F769512F97ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoreflower.mmh"},
	{0x017C45F6A607EE48ULL, "855403700.ogg"},
	{0x017C45F7695136A4ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoreflower.mao"},
	{0x017C57F769515509ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoreflower.msh"},
	{0x017D11A2CA5C4EF6ULL, "art\\characters\\creatures\\stonegolem\\animation\\blast.evt"},
	{0x017D4AF98E906140ULL, "vxlt_backalley_n_1km_0_0.vlm"},
	{0x017D82F550C6D0C2ULL, "seb201_lady_harimann6177402.cl"},
	{0x017E006652A608AAULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.dura.evt"},
	{0x017E2866FF1B17AFULL, "vxw_encounter2_camp_ix_0_0.vlm"},
	{0x017E5AAF9ABE6C78ULL, "art\\vfx\\combat\\vfx_g_spear_mod_c.anb"},
	{0x017E8E4AF41FF507ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.impa.gad"},
	{0x017EAEA98C28A492ULL, "plt_cod_crt_pride_demon.plo"},
	{0x017F14914D7B4EFDULL, "440700053.ogg"},
	{0x017F2F0EA206A0EFULL, "lt_0_lowtown_a3_n_c_36_0.anb"},
	{0x017F480178D83C6DULL, "atl_startmenu2_dxt5_dat.xml"},
	{0x017FC8EF6DD0408DULL, "lt_0_lowtown_docks_a3_102_0.phy"},
	{0x017FEFED8AFD005BULL, "6088151_m.fxe"},
	{0x01808259E885290EULL, "6087164_m.fxe"},
	{0x0180F3200C37CA5BULL, "lt_0_lowtown_docks_n_178_0.msh"},
	{0x018105200C37E8A6ULL, "lt_0_lowtown_docks_n_178_0.mao"},
	{0x018111200C37FD45ULL, "lt_0_lowtown_docks_n_178_0.mmh"},
	{0x01816FE7BF146648ULL, "plt_mer110pt_back_to_work.nss"},
	{0x01818B099D1ED7DCULL, "620875888.ogg"},
	{0x01826EA6D697D443ULL, "1054891784.ogg"},
	{0x018308A2CA60AA81ULL, "art\\characters\\creatures\\stonegolem\\animation\\blast.gad"},
	{0x0183855EF7599D4BULL, "middle_dusk_rig.msh"},
	{0x0183B4BAF01C2C34ULL, "880501347.ogg"},
	{0x0183EA1596E4F77FULL, "art\\levels\\races\\design\\prp_largetreasur.msh"},
	{0x0183F01596E50121ULL, "art\\levels\\races\\design\\prp_largetreasur.mmh"},
	{0x0183FC1596E515C2ULL, "art\\levels\\races\\design\\prp_largetreasur.mao"},
	{0x018429A73E2B4E18ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potplant_dp.dds"},
	{0x018470AEED6D4E50ULL, "6196685_f.fxe"},
	{0x01848FAE33C343BBULL, "831679755.ogg"},
	{0x0184DF6652ABEF1DULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.dura.gad"},
	{0x0184E4D4DA477040ULL, "765546665.ogg"},
	{0x01851B2FE0FDE728ULL, "fxa_unholy_bt.dds"},
	{0x0185665F6A476BFAULL, "w_bonepitdestroyed_a3_231_0.mmh"},
	{0x01856A5F6A477231ULL, "w_bonepitdestroyed_a3_231_0.mao"},
	{0x0185785F6A478A40ULL, "w_bonepitdestroyed_a3_231_0.msh"},
	{0x0185834AF4260114ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.impa.evt"},
	{0x0185A16D86075AC8ULL, "hlac_ht_mansion2_818_31z.dds"},
	{0x0185C453DC166D4FULL, "6079050_m.fxe"},
	{0x0185DB24FE08A686ULL, "mcr371_orsino6083353.cl"},
	{0x0185F8E2C903D31AULL, "mrl150tr_mtn_undead_2.utt"},
	{0x018613C558306793ULL, "hlac_lt_hangedman_n_-1_3x.dds"},
	{0x018660D21644436DULL, "gen00fl_merrill6106363.cl"},
	{0x01869084ACF9E8AFULL, "6119679_m.fxe"},
	{0x01872B6B22C2BB5FULL, "547978045.ogg"},
	{0x018731FF8F1AF19CULL, "1004859216.ogg"},
	{0x018785175B328BD5ULL, "490171856.ogg"},
	{0x0187BD53DB7F2ACEULL, "6071037_m.fxe"},
	{0x018859EE0232BCA6ULL, "t3_0000052.tnt"},
	{0x018878E7BF1A945BULL, "plt_mer110pt_back_to_work.plo"},
	{0x0188A892F33A0ED5ULL, "mer240_jansen6050334.cl"},
	{0x0188F2F77EE99675ULL, "and110_refugee6111601.cl"},
	{0x01894804B2E81C4AULL, "vxlt_gallowsprison_n_5qu_0_0.vlm"},
	{0x0189586A448A27E2ULL, "dae000ip_crowd_ff1_talk11.utp"},
	{0x018A09A2CA66CB2FULL, "art\\characters\\creatures\\stonegolem\\animation\\blast.ani"},
	{0x018A80629B3B5802ULL, "zz_dog_debug6127920.cl"},
	{0x018A8F95A831B060ULL, "w_blightlands_60_0.rml"},
	{0x018AF8FD30DCD4B5ULL, "6048824_f.fxe"},
	{0x018B13A756CE6AFFULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.cess.gad"},
	{0x018B4481BA91F138ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_arm_cas_0s.dds"},
	{0x018BE5292F5DEE01ULL, "hlac_w_encounter2_camp_65534_0z.dds"},
	{0x018C3FDF2351AB6CULL, "771313688.ogg"},
	{0x018CB59396951FD3ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_011a_0n.dds"},
	{0x018D06BF9F3E6A54ULL, "54311468.ogg"},
	{0x018DE10B6CAD12E8ULL, "lgt311tr_shw_mer_mettin.utt"},
	{0x018DF14772550F90ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_1y.dds"},
	{0x018DF259353D6875ULL, "zz_facefx_test.cnv"},
	{0x018E0EBB17BB1F48ULL, "llac_lt_0_lowtown_a2_n_c_0_0.dds"},
	{0x018E1CF2FE4D7D5EULL, "cs_mcr370cs_epilogue.bnk"},
	{0x018E758F3F000025ULL, "vxw_encounter3_60_0_0.vlm"},
	{0x018E80219F50A151ULL, "art\\characters\\heraldry\\hld_circle_of_mages_1.dds"},
	{0x018ED943AF635BB5ULL, "zz_rogueslate_white2.cut"},
	{0x018F464FECE88CE2ULL, "ht_keep_9215.rml"},
	{0x018F87043EE92DB7ULL, "348179205.ogg"},
	{0x018FF10F9505ED5FULL, "llac_w_encounter2_camp_65538_0.dds"},
	{0x019077CA90682E11ULL, "follower_banter6136471.cl"},
	{0x01907C39028E3ECBULL, "vxlt_0_lowtown_a2_gbw_0_0.vlm"},
	{0x0190E4D99A8CA4BFULL, "6101490_m.fxe"},
	{0x019108A756D2C3ECULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.cess.evt"},
	{0x01914CA1814DEF13ULL, "mag370_meredith6204104.cl"},
	{0x01916D26B016DA86ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf.ani"},
	{0x0191731A65F8D95FULL, "zz_idle_test.cut"},
	{0x019199A98E6CA7BAULL, "gen00fl_fenris6126342.cl"},
	{0x019213AC1BE2032CULL, "llac_lt_gallowscourtyard_-131072_1.dds"},
	{0x019253A1E3D9FA9DULL, "one140cr_veld.utc"},
	{0x01926F5A946FF6D8ULL, "vxca_caverns2_907_0_0.vlm"},
	{0x0192955A3EE20117ULL, "follower_banter6165608.cl"},
	{0x0192D19496F024F1ULL, "6173398_m.fxe"},
	{0x0192FC7023F088E7ULL, "273076312.ogg"},
	{0x0193049835CBF643ULL, "follower_banter6161240.cl"},
	{0x01934AF62108D663ULL, "6053592_f.fxe"},
	{0x01937B4AF432314DULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.impa.ani"},
	{0x0193F4D314D66E77ULL, "mcr370_cut_player_survives6171832.cl"},
	{0x0193F89D18E50A3BULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyo.phy"},
	{0x01942D3980108A35ULL, "vxht_0_hightown_a3_c91_0_0.vlm"},
	{0x019445B0A4D4154CULL, "144333932.ogg"},
	{0x019470CC8178BC3DULL, "6037106_m.fxe"},
	{0x019481BC65FC7745ULL, "vxlt_house_dy_0_0.vlm"},
	{0x0194D1752BCF5DA6ULL, "vxht_mansion2_3ek_0_0.vlm"},
	{0x0194DE59354364D0ULL, "zz_facefx_test.are"},
	{0x0195B2D41D9F080AULL, "plt_rdr300pt_boss.plo"},
	{0x01963AB1116A60DAULL, "140731671.ogg"},
	{0x019699A42BFFB8CFULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_candle.phy"},
	{0x0196C3CA7D392D66ULL, "mag170_thrask6016958.cl"},
	{0x0196DFF829C75918ULL, "vxht_bartrand_n_755_0_0.vlm"},
	{0x0196FC5FAD39A3F5ULL, "vfxchunkydd.vsh.fxo"},
	{0x019703A6B55E6B9EULL, "vxht_keep_n_3ba_0_0.vlm"},
	{0x01979F4477FD832CULL, "vxlt_gallowscourtyard_n_3bb_0_0.vlm"},
	{0x0197A4D13258201FULL, "vxht_mansion2_3g0_0_0.vlm"},
	{0x019804111482207AULL, "6102555_m.fxe"},
	{0x01982EABA4695FD3ULL, "102664365.ogg"},
	{0x0198A2D2D7F30393ULL, "6064719_m.fxe"},
	{0x019929CDB213B456ULL, "6125407_f.fxe"},
	{0x0199577CEEAF5999ULL, "ht_0_hightown_a2_16_0.rml"},
	{0x019AA1859AF0647EULL, "132105732.ogg"},
	{0x019AD648FA0486E9ULL, "vxlt_0_lowtown_n_e28_0_0.vlm"},
	{0x019B8DBD88EF1AE5ULL, "rdr111dg_martin6014998.cl"},
	{0x019CC0C96AB4BA80ULL, "zz_qun_debug6045400.cl"},
	{0x019DC0741822B625ULL, "vxca_caverns1_i8y_0_0.vlm"},
	{0x019DC8A0FCC65169ULL, "bur_ho_clu_glo.mao"},
	{0x019E61DE5A262129ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef04.phy"},
	{0x019E922B6209AAE7ULL, "355926805.ogg"},
	{0x019EB180B0CCD63FULL, "hlac_lt_0_lowtown_a2_-2_6y.dds"},
	{0x019EED59354BD2C3ULL, "zz_facefx_test.lst"},
	{0x019EF0A756DED845ULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.cess.ani"},
	{0x019F51BDA8630C57ULL, "one220_yevhen6102813.cl"},
	{0x019FB2B41AE1C4C3ULL, "6174876_m.fxe"},
	{0x01A037FA82DB54F5ULL, "vxlt_0_lowtown_a3_n_c_ho9_0_0.vlm"},
	{0x01A0487ADB18238CULL, "17075002.ogg"},
	{0x01A09CB39B1BA582ULL, "6060641_f.fxe"},
	{0x01A0BD1A250E2917ULL, "hlac_w_bonepitmining_a2_3_0z.dds"},
	{0x01A0E13786660F58ULL, "vxca_caverns2_954_0_0.vlm"},
	{0x01A10B0A0157E607ULL, "evl_dot.mao"},
	{0x01A1397A47C932B3ULL, "hm_mag_controlled_3.mor"},
	{0x01A1F632D57A8860ULL, "934948865.ogg"},
	{0x01A2B923587C865DULL, "303419649.ogg"},
	{0x01A2DC7FA8DC4E7AULL, "6157194_f.fxe"},
	{0x01A2E2A9736ED4BDULL, "6030006_m.fxe"},
	{0x01A310BC28D7F3AFULL, "6053897_f.fxe"},
	{0x01A3DA2B86D60AD4ULL, "art\\vfx\\combat\\vfx_r_var_activation2_c.mmh"},
	{0x01A3E4243450B303ULL, "6134726_m.fxe"},
	{0x01A468EED5468B54ULL, "vxw_woundedcoast_y2_0_0.vlm"},
	{0x01A4825D52C18091ULL, "6143713_m.fxe"},
	{0x01A511BC6091CA9EULL, "one300_nuncio6073802.cl"},
	{0x01A517C8B55BA252ULL, "hlac_w_sundermount_393215_0z.dds"},
	{0x01A536B9E03AB19FULL, "6080189.cut"},
	{0x01A55626B0275C80ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf.gad"},
	{0x01A55FF21B0D94BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_8.pwk"},
	{0x01A5841188D61BB0ULL, "vxht_chantry_n_2zf_0_0.vlm"},
	{0x01A58C44448942E3ULL, "6048343_f.fxe"},
	{0x01A5DA8D16E56C25ULL, "ser270_templar6164087.cl"},
	{0x01A656DE86A7B2C1ULL, "6175712_m.fxe"},
	{0x01A659AF9ADFBC13ULL, "art\\vfx\\combat\\vfx_g_spear_mod_c.mmh"},
	{0x01A663EB16740FF4ULL, "hlac_lt_0_lowtown_a2_-65535_0z.dds"},
	{0x01A7042BA5CB5D5EULL, "221075995.ogg"},
	{0x01A75FE5E206C982ULL, "vxlt_0_lowtown_a3_n_c_gyk_0_0.vlm"},
	{0x01A78AFC5AD85BD5ULL, "rdr111dg_harbormaster6016142.cl"},
	{0x01A79BAAA409CCD6ULL, "6122507_f.fxe"},
	{0x01A7D263A8F67B39ULL, "gen00fl_fenris6099801.cl"},
	{0x01A8A14D793BEB0FULL, "6052923_m.fxe"},
	{0x01A90D8B8B23C6D5ULL, "vxca_caverns2prc_a7f_0_0.vlm"},
	{0x01A925F36F12BD7DULL, "vxw_woundedcoast_19j_0_0.vlm"},
	{0x01A929AB56EB361FULL, "6143122_m.fxe"},
	{0x01A985D9B6FDE0A9ULL, "hlac_lt_0_lowtown_a3_n_c_1_10x.dds"},
	{0x01A9F1F04FAF6943ULL, "vxlt_0_lowtown_n_ep4_0_0.vlm"},
	{0x01A9FCF93755D4A8ULL, "vxlt_0_lowtown_a3_n_c_hta_0_0.vlm"},
	{0x01AA9AE04A5B1341ULL, "gen00fl_aveline6137717.cl"},
	{0x01AAA2317B744198ULL, "369490588.ogg"},
	{0x01AAAE7FD1D3FE69ULL, "mag131_taraneh6094536.cl"},
	{0x01AAF9E5A5FE42C2ULL, "6079288_m.fxe"},
	{0x01AB0C0EA22B9F93ULL, "lt_0_lowtown_a3_n_c_36_0.rml"},
	{0x01AB581D068BA7F3ULL, "vxw_bonepitmining_a2_1ow_0_0.vlm"},
	{0x01AB6DD17CFEA83FULL, "mag360_alain6044656.cl"},
	{0x01AB762B7868B338ULL, "6126397_m.fxe"},
	{0x01AB88CB984EF531ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_right_1.mao"},
	{0x01AB9EEA1B8431FEULL, "46457837.ogg"},
	{0x01AC8D17D4103989ULL, "6032853_m.fxe"},
	{0x01AC9A284789126BULL, "ht_0_hightown_a3_35_0_ll.mmh"},
	{0x01AC9C7A71BA9D00ULL, "plt_cod_hst_qunari.plo"},
	{0x01ACAE762DCECE1BULL, "vxw_encounter2_nm_0_0.vlm"},
	{0x01AD92A20C2F69E5ULL, "940437419.ogg"},
	{0x01AD9EB44185B843ULL, "art\\characters\\creatures\\ragedemon\\animation\\shuffle.ani"},
	{0x01ADB039AF136844ULL, "vxca_caverns1_mines_cna_0_0.vlm"},
	{0x01ADFE76ED90F803ULL, "ser270_templar6165680.cl"},
	{0x01AE1BE0F94AC596ULL, "261595125.ogg"},
	{0x01AE3E989995F9CAULL, "vxlt_hangedman_1ht_0_0.vlm"},
	{0x01AE4A8CD7B9D972ULL, "361233576.ogg"},
	{0x01AE96A50A611DACULL, "325402529.ogg"},
	{0x01AEE6DBE00245CEULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.dura.ani"},
	{0x01AF41C6939919E8ULL, "lgt110_orz_harrowmont6172607.cl"},
	{0x01AFDF096FA80778ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_crumble.phy"},
	{0x01B01B13249F5471ULL, "vxw_bonepit_l5_0_0.vlm"},
	{0x01B03098FDB02C44ULL, "follower_banter6153976.cl"},
	{0x01B0987CE43D64EBULL, "vxlt_0_lowtown_a3_n_c_hqi_0_0.vlm"},
	{0x01B0A71419D2608FULL, "zz_followers_banter6157856.cl"},
	{0x01B0A9962331E38EULL, "mcr371_hostile_anders6082118.cl"},
	{0x01B0CA0CDF4526C4ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_8z.dds"},
	{0x01B0F115970BF09EULL, "art\\levels\\races\\design\\prp_largetreasur.phy"},
	{0x01B12D2517B76DF2ULL, "896880102.ogg"},
	{0x01B1C2CFEC005159ULL, "vxlt_shop_t1_0_0.vlm"},
	{0x01B1E13170A2CC7EULL, "docks_scene_dockworkers6176219.cl"},
	{0x01B1F4023E623143ULL, "6149649_m.fxe"},
	{0x01B201A679B4DE43ULL, "vxca_caverns2_981_0_1.vlm"},
	{0x01B20A50D3B75653ULL, "ser121_jethann6106564.cl"},
	{0x01B20DE9CFCDE06CULL, "hlac_ht_mansion2_818_68x.dds"},
	{0x01B2A98542665558ULL, "vxlt_0_lowtown_docks_n_ba7_0_1.vlm"},
	{0x01B2B9408AFF51D3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsma.phy"},
	{0x01B2DACBC73CA3D4ULL, "vxw_woundedcoast_v3_0_0.vlm"},
	{0x01B2E69D15E3FD4BULL, "dae000ip_crowd_ff8_talk8.utp"},
	{0x01B3E3B77FDC0B13ULL, "hlac_ht_mansion2_818_9x.dds"},
	{0x01B3E649E3630DBAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtb_0n.dds"},
	{0x01B43E8D247D09CBULL, "636924627.ogg"},
	{0x01B48B39A0D27F21ULL, "vxlt_0_lowtown_glo_0_0.vlm"},
	{0x01B4BEE78B37161BULL, "6087200_m.fxe"},
	{0x01B507B73613CB4EULL, "vxdr_primevaltomb_7kh_0_0.vlm"},
	{0x01B607C6AFC85230ULL, "6082214_f.fxe"},
	{0x01B6433111B8C2A9ULL, "6208479_f.fxe"},
	{0x01B6580FE9A98BC6ULL, "hlac_lt_hangedman_-1_28y.dds"},
	{0x01B66B12DCD6F70EULL, "vxht_0_hightown_a2_c8p_0_0.vlm"},
	{0x01B76B454BAE2A69ULL, "hlac_ht_0_hightown_65536_0y.dds"},
	{0x01B7903418681E44ULL, "vxca_caverns1_mines_gk3_0_1.vlm"},
	{0x01B79CB3B1CD4C56ULL, "hlac_lt_0_lowtown_a2_n_c_131071_7x.dds"},
	{0x01B80969F42EA9CCULL, "192986629.ogg"},
	{0x01B811C9BFC3F455ULL, "blood_imp_01.mao"},
	{0x01B82C2D891BE625ULL, "vxlt_0_lowtown_a2_ep1_0_0.vlm"},
	{0x01B83FA5CE58C36DULL, "201017059.ogg"},
	{0x01B84FD60C2A2C27ULL, "isa000ip_bottle_container.utp"},
	{0x01B8686C82ECD32AULL, "842637075.ogg"},
	{0x01B89FAD2B7A76F1ULL, "vxca_caverns2prc_9qx_0_0.vlm"},
	{0x01B967A96654C8A5ULL, "22087330.ogg"},
	{0x01B97FCAA2CC6769ULL, "hlac_ht_0_hightown_a3_131071_2z.dds"},
	{0x01BADAB18F75317CULL, "159883868.ogg"},
	{0x01BAE15A534D56A2ULL, "vxht_0_hightown_n_8pd_0_0.vlm"},
	{0x01BB096C228531DEULL, "vxlt_gallowscourtyard_n_79y_0_0.vlm"},
	{0x01BB20531FC4E69BULL, "vxw_encounter2_il_0_0.vlm"},
	{0x01BB339AA2F2E459ULL, "vxht_mansion2_24z_0_0.vlm"},
	{0x01BB5154530A4EB8ULL, "6172583_m.fxe"},
	{0x01BB600E52FB4BE0ULL, "6133346_m.fxe"},
	{0x01BC48E5157C6D6BULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm_eyesm1.mao"},
	{0x01BCDEDBE00E751FULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.dura.evt"},
	{0x01BCE8593564C7A4ULL, "zz_facefx_test.utc"},
	{0x01BCFBD4CDAE2F7FULL, "one100dg_kelder6060540.cl"},
	{0x01BCFE196FBA2BAFULL, "zz_ave_debug6064626.cl"},
	{0x01BE7A2A12516178ULL, "ht_0_hightown_n_52_0.rml"},
	{0x01BF0ED5167E5281ULL, "hlac_ca_caverns2prc_5802_23x.dds"},
	{0x01BFD4A99432E3E1ULL, "vxca_caverns2_62n_0_1.vlm"},
	{0x01BFDF9D190A1ADAULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyo.msh"},
	{0x01BFE09060C8E7C0ULL, "ave310_donnic6134769.cl"},
	{0x01BFED9D190A3213ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyo.mao"},
	{0x01BFF19D190A3920ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyo.mmh"},
	{0x01C016B39FBF2FD0ULL, "vxw_sundermount_2kr_0_0.vlm"},
	{0x01C02CB2484CF8CFULL, "430105954.ogg"},
	{0x01C0A0D34B6ED2EDULL, "91346626.ogg"},
	{0x01C0E74DFC94DEC2ULL, "11061155.ogg"},
	{0x01C1425D1780AA70ULL, "vxlt_0_lowtown_docks_cvx_0_1.vlm"},
	{0x01C14CA8B932BC54ULL, "vxw_woundedcoast_s0_0_0.vlm"},
	{0x01C16D8ABD44A686ULL, "gen00fl_fenris6113885.cl"},
	{0x01C1B5B44196882DULL, "art\\characters\\creatures\\ragedemon\\animation\\shuffle.gad"},
	{0x01C2FD1692C897A1ULL, "vxlt_0_lowtown_gkn_0_0.vlm"},
	{0x01C37A1B295B7C61ULL, "w_encounter2_39_0.phy"},
	{0x01C3BFDBE0145F08ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.dura.gad"},
	{0x01C3C3E73E4E78AFULL, "6043724_f.fxe"},
	{0x01C3EC59356AED26ULL, "zz_facefx_test.stg"},
	{0x01C4E1DB41202740ULL, "549434560.ogg"},
	{0x01C555E2A9D326D1ULL, "follower_banter6170572.cl"},
	{0x01C558C8EF5DFAECULL, "6206813_m.fxe"},
	{0x01C64A56F9D54FFBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured.evt"},
	{0x01C659AA7E582F5FULL, "hlac_w_sundermount_327677_0z.dds"},
	{0x01C67E7110C30F98ULL, "572360438.ogg"},
	{0x01C6939DD04745C6ULL, "mrl000_marethari6142116.cl"},
	{0x01C6BA492B497A47ULL, "vxw_bonepitmining_a2_8u_0_0.vlm"},
	{0x01C6EF06FA04B371ULL, "llac_ht_chantry_n_2534_49.dds"},
	{0x01C7992C8FC3BECFULL, "art\\vfx\\combat\\vfx_m_pri_petrify_conj_c.mmh"},
	{0x01C7D2C46646C9B2ULL, "6053693_m.fxe"},
	{0x01C7DBC8CBE834DDULL, "llac_lt_0_lowtown_a2_n_c_-65535_2.dds"},
	{0x01C7F0B1B5E98F69ULL, "gen00fl_merrill6115792.cl"},
	{0x01C819CE1E7B9F75ULL, "628918599.ogg"},
	{0x01C85EA0B7A3025FULL, "vxlt_gallowscourtyard_748_0_0.vlm"},
	{0x01C8A36AE9182C95ULL, "w_blightlands_217_0.msh"},
	{0x01C8A56AE918300BULL, "w_blightlands_217_0.mmh"},
	{0x01C8B16AE91844A8ULL, "w_blightlands_217_0.mao"},
	{0x01C8B17499F65627ULL, "zz_approval_debug6003703.cl"},
	{0x01C8E1127CFBA9E3ULL, "vxw_encounter1_e1_0_0.vlm"},
	{0x01C97114320C831BULL, "6077012_m.fxe"},
	{0x01C98E4694F49D57ULL, "dae210_gate_guard6085083.cl"},
	{0x01C9DDBD15885ACBULL, "gen00fl_varric6078675.cl"},
	{0x01C9F3F659A786BEULL, "6111933_m.fxe"},
	{0x01CA4ADE5A4B354EULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef04.mmh"},
	{0x01CA54C0A42AD190ULL, "6127620_m.fxe"},
	{0x01CA64DE5A4B613CULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef04.msh"},
	{0x01CA831B2961AA7DULL, "w_encounter2_39_0.rml"},
	{0x01CAE661F60E3B53ULL, "dae000_arishok6068531.cl"},
	{0x01CAFDE3A8FA9856ULL, "zz_qun_debug6085864.cl"},
	{0x01CBA4739D33B153ULL, "739936143.ogg"},
	{0x01CC39D7F4A7A40AULL, "mcr310_argument6087315.cl"},
	{0x01CC817A0695615AULL, "crt_pro_cycle.mao"},
	{0x01CCCD2E12B8132CULL, "zz_qcr_debug6074958.cl"},
	{0x01CCDD2B86F9148BULL, "art\\vfx\\combat\\vfx_r_var_activation2_c.anb"},
	{0x01CCF199227E8F72ULL, "490959097.ogg"},
	{0x01CD4356F9DB623CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured.gad"},
	{0x01CD518E886244EDULL, "hlac_lt_gallowscourtyard_-1_1z.dds"},
	{0x01CD56C6D43D64A1ULL, "938928808.ogg"},
	{0x01CDD10B214B7572ULL, "hla_ht_0_hightown_a2_131073_1z.dds"},
	{0x01CDD823E3C3F753ULL, "one110_gustav6089601.cl"},
	{0x01CDEEF2FE831364ULL, "cs_mcr370cs_epilogue.wbd"},
	{0x01CE132CC2410420ULL, "6133830_m.fxe"},
	{0x01CE1B16CD6F9B8AULL, "follower_banter6171538.cl"},
	{0x01CE33A4E68D8EB2ULL, "6156006_m.fxe"},
	{0x01CE86C0BB5D8A85ULL, "217899971.ogg"},
	{0x01CE889091B54850ULL, "vxw_sundermount_2ls_0_0.vlm"},
	{0x01CEA1E4FC17764FULL, "pro000cs_act1_crim_mage.cut"},
	{0x01CECDD4ED81CE62ULL, "t3_wdg_w21.tnt"},
	{0x01CECFF85CAEBF76ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_act_c.impa.evt"},
	{0x01CF595D0A64DA25ULL, "vxht_0_hightown_n_b27_0_0.vlm"},
	{0x01CF852069068A69ULL, "519941235.ogg"},
	{0x01CFADB71D023A3EULL, "636760602.ogg"},
	{0x01CFEF0298B5F033ULL, "vxw_sundermount_313_0_1.vlm"},
	{0x01D0075E01072065ULL, "mcr371_orsino6083327.cl"},
	{0x01D04F68F67EBF5DULL, "vxlt_0_lowtown_docks_c02_0_0.vlm"},
	{0x01D0AF78AE85B426ULL, "gen00fl_isabela6099417.cl"},
	{0x01D1BDF98864ECCAULL, "vxw_encounter2_camp_4b_0_0.vlm"},
	{0x01D20CBFB5603844ULL, "482685270.ogg"},
	{0x01D2E874EDE79F5AULL, "178305163.ogg"},
	{0x01D2EDBC2287E3C2ULL, "vxht_0_hightown_n_c0g_0_0.vlm"},
	{0x01D3400790900693ULL, "sound_item_materials.gda"},
	{0x01D37ABC68B63262ULL, "ttx_pillar_spire_1.mmh"},
	{0x01D37CBC68B63518ULL, "ttx_pillar_spire_1.msh"},
	{0x01D3B113F3CDBE12ULL, "129696791.ogg"},
	{0x01D40C8057281B9EULL, "dre000_cut_rock_wraith.cnv"},
	{0x01D44256F9E17F3AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured.ani"},
	{0x01D46C967FC72789ULL, "character_rendertodepth_p3.psh.fxo.dx11"},
	{0x01D4C6F85CB31B01ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_act_c.impa.gad"},
	{0x01D5264D695AC836ULL, "rog_criticalstrike.dds"},
	{0x01D5C33E81C14148ULL, "hlac_ca_caverns2prc_5802_55x.dds"},
	{0x01D6DC8B8DCBA639ULL, "129076433.ogg"},
	{0x01D6EE80572A1EAEULL, "dre000_cut_rock_wraith.bnk"},
	{0x01D7462625D1320CULL, "def_cobble_n.dds"},
	{0x01D752EF6EF1C263ULL, "vxw_encounter1_j0_0_0.vlm"},
	{0x01D77BF86F3AC2BCULL, "w_blightlands_125_0.mao"},
	{0x01D787F86F3AD75FULL, "w_blightlands_125_0.mmh"},
	{0x01D78DF86F3AE101ULL, "w_blightlands_125_0.msh"},
	{0x01D7B00BE9E4F970ULL, "lt_0_lowtown_docks_233_0.msh"},
	{0x01D7BE0BE9E5108AULL, "lt_0_lowtown_docks_233_0.mmh"},
	{0x01D7C20BE9E517C1ULL, "lt_0_lowtown_docks_233_0.mao"},
	{0x01D7C216FB332441ULL, "79106815.ogg"},
	{0x01D7C83798B8C8DDULL, "mag210_arianni6024530.cl"},
	{0x01D8C9C8CF3B59A4ULL, "6178484_m.fxe"},
	{0x01D8DAF53F82E084ULL, "35009841.ogg"},
	{0x01D92BC16EE16261ULL, "vxlt_warehouse_8j_0_0.vlm"},
	{0x01D94F4C4D935BFCULL, "one300_nuncio6075046.cl"},
	{0x01D95A680822BF16ULL, "vxlt_gallowstemplar_6so_0_0.vlm"},
	{0x01DAC3A455D58B94ULL, "mag210_arianni6057320.cl"},
	{0x01DACB140B2F7369ULL, "6137628_m.fxe"},
	{0x01DBB1E152299DA8ULL, "6035905_m.fxe"},
	{0x01DBC7F85CB93BAFULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_act_c.impa.ani"},
	{0x01DBD25B5520470DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha11a.mmh"},
	{0x01DBDE068D45A0AFULL, "vxlt_0_lowtown_a3_n_c_gz3_0_0.vlm"},
	{0x01DBE45B55206593ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha11a.msh"},
	{0x01DC601DE9BC67D9ULL, "609686389.ogg"},
	{0x01DC6F715BBB37E7ULL, "6207707_m.fxe"},
	{0x01DC750F55CC9CDCULL, "vxlt_hangedman_n_1jw_0_0.vlm"},
	{0x01DCFA6D83AB60D0ULL, "vxw_sundermount_2qp_0_0.vlm"},
	{0x01DDBBFB5C703E80ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrel.phy"},
	{0x01DDEFDC19FFC3DEULL, "gen_im_arm_glv_hvy_act2_03.uti"},
	{0x01DE4F31208A4155ULL, "vxlt_0_lowtown_n_eny_0_1.vlm"},
	{0x01DE7872A8E8A30AULL, "plt_cod_mgc_possession.plo"},
	{0x01DE9176A15D54FAULL, "art\\vfx\\models\\vfx_cin_rock_02.msh"},
	{0x01DEA376A15D7340ULL, "art\\vfx\\models\\vfx_cin_rock_02.mmh"},
	{0x01DEF6A3E20082D6ULL, "6138805_f.fxe"},
	{0x01DFB448161AFB47ULL, "476205535.ogg"},
	{0x01DFD610F7259C3BULL, "189029176.ogg"},
	{0x01DFE447AFDF4DA0ULL, "mir_rip_sliver.mao"},
	{0x01DFFFE903795315ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_filecabinet.phy"},
	{0x01E04910B8315ADBULL, "mag100_karras6034708.cl"},
	{0x01E05D4E0BF614CEULL, "vxdr_primevaltomb_7zk_0_0.vlm"},
	{0x01E06548E7FB361BULL, "349712331.ogg"},
	{0x01E114B157B50FF3ULL, "vxht_playershouse_n_1j7_0_0.vlm"},
	{0x01E141D7600B4D00ULL, "735759448.ogg"},
	{0x01E15D673A098BB9ULL, "oip_dust.mao"},
	{0x01E18BCC339D10E3ULL, "188420035.ogg"},
	{0x01E1D6315D82B7B3ULL, "one100ip_st_ilen.utp"},
	{0x01E26224CEAC4D75ULL, "dae000_merchant_access_hf6162977.cl"},
	{0x01E2B47730360E48ULL, "vxlt_warehouse_8o_0_0.vlm"},
	{0x01E33D0B7B247AD4ULL, "6104941_f.fxe"},
	{0x01E37FB89980A04BULL, "sib131_slaver_lord6098067.cl"},
	{0x01E38D02F0D6FAFBULL, "w_encounter3_112_0.msh"},
	{0x01E39F02F0D71946ULL, "w_encounter3_112_0.mao"},
	{0x01E3AB02F0D72DE5ULL, "w_encounter3_112_0.mmh"},
	{0x01E3B7B504618AB8ULL, "vxdr_primevaltomb_81z_0_1.vlm"},
	{0x01E402D8704E693FULL, "art\\vfx\\creatures\\fxc_spider_dppr_c.impa.gad"},
	{0x01E51222BF3FC4BEULL, "125306649.ogg"},
	{0x01E52DDA3819F7DCULL, "270296176.ogg"},
	{0x01E52E924A85706CULL, "art\\vfx\\materials\\vfx_particle_static.mmh"},
	{0x01E534FC3871891AULL, "vxlt_0_lowtown_f0e_0_0.vlm"},
	{0x01E55154849E8463ULL, "qun200_party_comment6090454.cl"},
	{0x01E575D1DBBE2CCDULL, "vxht_0_hightown_a3_ca1_0_0.vlm"},
	{0x01E6566484968BA5ULL, "534968233.ogg"},
	{0x01E73BA5C1317224ULL, "6173584_m.fxe"},
	{0x01E784F789981856ULL, "lowtown_male_beggar_sta6151021.cl"},
	{0x01E7BBFB98AA6CACULL, "fex_spt_pil.msh"},
	{0x01E7C1FB98AA769EULL, "fex_spt_pil.mmh"},
	{0x01E7DF58323CF685ULL, "mrl150_dalish_guard6170402.cl"},
	{0x01E7FB6DB8759C32ULL, "78376993.ogg"},
	{0x01E8B80CDBC6AEE5ULL, "gen00fl_isabela6100004.cl"},
	{0x01E8D00B6CF9A115ULL, "lgt311tr_shw_mer_mettin.ncs"},
	{0x01E914B6C1B3580BULL, "character_character_p80.psh.fxo"},
	{0x01E9E71C1AE6C7B9ULL, "520390239.ogg"},
	{0x01EA4B97D35A13D0ULL, "mer140_bonepit_comments6062992.cl"},
	{0x01EAB659BF527201ULL, "643471853.ogg"},
	{0x01EAF7D87054752CULL, "art\\vfx\\creatures\\fxc_spider_dppr_c.impa.evt"},
	{0x01EB6C4A75A17950ULL, "vxw_sundermount_2zq_0_0.vlm"},
	{0x01EB9136731F74FDULL, "mag160_feynriel6020093.cl"},
	{0x01EBCA11840CBFC7ULL, "isa111_isabela6075421.cl"},
	{0x01EBD45D0DBE3604ULL, "zz_one_debug6025820.cl"},
	{0x01EBDD9EE6451D35ULL, "44718964.ogg"},
	{0x01EC3ABAAC66A5FDULL, "734087125.ogg"},
	{0x01EC3D16EE510B25ULL, "vxht_0_hightown_n_b85_0_0.vlm"},
	{0x01EC5730C87FDBC0ULL, "6073781_m.fxe"},
	{0x01EC5843569BDEF1ULL, "651545614.ogg"},
	{0x01ECA81E1ED83EDAULL, "187830120.ogg"},
	{0x01ECC10E128059D5ULL, "vxlt_0_lowtown_n_ekf_0_1.vlm"},
	{0x01ECCB5B552EAECAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha11a.phy"},
	{0x01ECDB5D4EE54B5FULL, "art\\effects\\particlelightparams.dat"},
	{0x01ED2C7D7EE92C8CULL, "zz_dae_debug6049998.cl"},
	{0x01EDA595D6A93AE1ULL, "1061589519.ogg"},
	{0x01EE6296C21D7E8CULL, "vxlt_hangedman_28i_0_0.vlm"},
	{0x01EECBAC64CCE5F0ULL, "hlac_w_woundedcoast_65541_0x.dds"},
	{0x01EEF63635BE2E52ULL, "vxca_cavern1_haunted_j4x_0_0.vlm"},
	{0x01EF00119EA81C4EULL, "zz_dae_debug6029918.cl"},
	{0x01EF17276D553AB5ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.mmh"},
	{0x01EF2ED9E7616DC2ULL, "dae000ip_crowd_ff2_talk3.utp"},
	{0x01EFDD6AF963F3B8ULL, "vxht_0_hightown_a3_c8g_0_0.vlm"},
	{0x01EFE5633FCD7789ULL, "vxca_caverns1_mines_ehl_0_0.vlm"},
	{0x01F09A2C8FE6C50CULL, "art\\vfx\\combat\\vfx_m_pri_petrify_conj_c.anb"},
	{0x01F0E68F9B836937ULL, "dae020cr_hm_gossip3.utc"},
	{0x01F12FAB647FA073ULL, "6049471_m.fxe"},
	{0x01F1AA934DBDA559ULL, "dae000_meredith6092607.cl"},
	{0x01F1D3EA6074BEDDULL, "40364646.ogg"},
	{0x01F29122E595AFB5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_basic.phy"},
	{0x01F2969E8D4E9FEAULL, "and121_karl6093893.cl"},
	{0x01F2FD6140E35CA9ULL, "hlac_ca_caverns2prc_5802_61z.dds"},
	{0x01F3916DC05A2273ULL, "vxw_bonepitmining_a2_1pz_0_0.vlm"},
	{0x01F3A6D92A67A19AULL, "vxlt_0_lowtown_f5f_0_0.vlm"},
	{0x01F3BA7EBC85E2BEULL, "hlac_ht_0_hightown_n_-131074_2y.dds"},
	{0x01F3F22B99185311ULL, "vxw_woundedcoast_39_0_0.vlm"},
	{0x01F43030A6246FDFULL, "hlac_w_bonepitmining_a2_262146_0x.dds"},
	{0x01F4B055E8C886F9ULL, "6161300_m.fxe"},
	{0x01F4BCF1DF097E12ULL, "hlac_ht_chantry_2534_38z.dds"},
	{0x01F4CD35980B40E2ULL, "qcr220cr_nobleman3.utc"},
	{0x01F55214D7F33377ULL, "fx_glow03.dds"},
	{0x01F5F4D677D9FFAEULL, "gen00fl_aveline6004159.cl"},
	{0x01F6102B6ECFE1A0ULL, "6160962_m.fxe"},
	{0x01F6A81CA6DCCC69ULL, "zz_followers_banter6157847.cl"},
	{0x01F6AB37B131AAD9ULL, "250282369.ogg"},
	{0x01F6ACF173B2673BULL, "gen00fl_aveline6132833.cl"},
	{0x01F6B5372D24D43EULL, "vxlt_undercityserialkiller_1t3_0_0.vlm"},
	{0x01F6C1BCD93C2AB4ULL, "6039003_f.fxe"},
	{0x01F6E36B29668D97ULL, "art\\vfx\\models\\vfx_debris_04.mmh"},
	{0x01F6E76B296694A4ULL, "art\\vfx\\models\\vfx_debris_04.mao"},
	{0x01F6F96B2966B309ULL, "art\\vfx\\models\\vfx_debris_04.msh"},
	{0x01F72542E05B91A3ULL, "139941280.ogg"},
	{0x01F78693B3A9708BULL, "character_character_p71.psh.fxo"},
	{0x01F7C5D32F72B1FFULL, "one100dg_kelder6055449.cl"},
	{0x01F81C0385390678ULL, "and230_mage_girl6116260.cl"},
	{0x01F8616067C2AF74ULL, "147824282.ogg"},
	{0x01F8AF0EACB6737AULL, "art\\characters\\playercharacter\\humanfemale\\glv_mede.mmh"},
	{0x01F8B30EACB679B1ULL, "art\\characters\\playercharacter\\humanfemale\\glv_mede.mao"},
	{0x01F8C10EACB691C0ULL, "art\\characters\\playercharacter\\humanfemale\\glv_mede.msh"},
	{0x01F8D1CD9E0BDF6BULL, "vxlt_0_lowtown_a3_n_c_hjl_0_0.vlm"},
	{0x01F8DFD870608985ULL, "art\\vfx\\creatures\\fxc_spider_dppr_c.impa.ani"},
	{0x01F971E86013E254ULL, "art\\vfx\\combat\\vfx_glb_potion_drink_c.impa.ani"},
	{0x01FA3BEDFDCF98DBULL, "vxdr_0_deeproads_du2_0_0.vlm"},
	{0x01FA636E6D4B320CULL, "637429153.ogg"},
	{0x01FAF921B518FF7DULL, "mag121_idunna6084231.cl"},
	{0x01FB0131B2BB4039ULL, "lgt160_ntb_elf_conflict6163368.cl"},
	{0x01FB0E02909236D4ULL, "712523648.ogg"},
	{0x01FB41FF74394BB4ULL, "224939151.ogg"},
	{0x01FB7F99A77317E0ULL, "framebuffereffects_diffusiondof_p11.psh.fxo.dx11"},
	{0x01FBA14D7327376BULL, "vxw_encounter2_r9_0_0.vlm"},
	{0x01FC11EA3CE5C6EAULL, "vxlt_0_lowtown_docks_cp1_0_0.vlm"},
	{0x01FC3A760D82A851ULL, "6088226_f.fxe"},
	{0x01FCA0C6AA8C8D94ULL, "gen00fl_aveline6105463.cl"},
	{0x01FCA83E6932A203ULL, "6104758_m.fxe"},
	{0x01FCC1DBD012CB9FULL, "vxht_mansion2_389_0_0.vlm"},
	{0x01FCD473B413970CULL, "vxlt_hangedman_25n_0_0.vlm"},
	{0x01FCE50B0A7D73BBULL, "290658571.ogg"},
	{0x01FCF5B8E831E27AULL, "follower_banter6155712.cl"},
	{0x01FD14276D617342ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.anb"},
	{0x01FD681327B446D2ULL, "vxca_cavern1_haunted_j3y_0_0.vlm"},
	{0x01FD7ABBF41CF9DFULL, "mer320_hubert6053744.cl"},
	{0x01FDE7C475DE8C12ULL, "one350cr_variel.utc"},
	{0x01FE3C28B06B567AULL, "6098902_m.fxe"},
	{0x01FE91BC325444BAULL, "vxca_caverns2prc_9sq_0_0.vlm"},
	{0x01FEA91AD8C0C51FULL, "vxw_encounter3_g1_0_0.vlm"},
	{0x01FEC6C17ABE5160ULL, "vxlt_blackemporium_z9_0_0.vlm"},
	{0x01FECED23654916FULL, "930583148.ogg"},
	{0x01FEEB552E079626ULL, "gen_im_trs_gen_bracelet_03.uti"},
	{0x01FF14B801284C21ULL, "497417416.ogg"},
	{0x01FF198ED1C34459ULL, "vxlt_gallowscourtyard_773_0_0.vlm"},
	{0x01FFB2154FADC7F1ULL, "mag311_keran6062058.cl"},
	{0x01FFDDC6E3D1C40AULL, "gen00fl_bethany6121314.cl"},
	{0x0200403161D83F59ULL, "gen00fl_aveline6116622.cl"},
	{0x020059DD3721E11EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_exit.ani"},
	{0x02005A475A5C820DULL, "6101027_m.fxe"},
	{0x02006EE86019FBB6ULL, "art\\vfx\\combat\\vfx_glb_potion_drink_c.impa.gad"},
	{0x0200D8FDE548F171ULL, "vxca_caverns2prc_9hs_0_0.vlm"},
	{0x0201AD918EEB91A1ULL, "vxht_chantry_n_36j_0_0.vlm"},
	{0x02020B77AE6B181CULL, "ave100_donnic6058927.cl"},
	{0x02023FC99B3F1D36ULL, "vxlt_0_lowtown_a2_n_c_i18_0_0.vlm"},
	{0x0202D8FB5C8F8C15ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrel.msh"},
	{0x0202DAFB5C8F8F8BULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrel.mmh"},
	{0x0202E6FB5C8FA428ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrel.mao"},
	{0x0202F81252463789ULL, "prty_bethany_armor_1.dds"},
	{0x02032232DABE9F79ULL, "6166200_m.fxe"},
	{0x0203470170345992ULL, "989617390.ogg"},
	{0x02034C4ACC3BB18DULL, "llac_w_sundermount_-65538_0.dds"},
	{0x02035FB4FA25328DULL, "vxlt_0_lowtown_a2_ekj_0_2.vlm"},
	{0x0203A2729CCDE70EULL, "lt_0_lowtown_docks_a3_39_0.anb"},
	{0x0203BF19448E1B01ULL, "vxlt_0_lowtown_euz_0_0.vlm"},
	{0x02044B9CD07F75D3ULL, "677141380.ogg"},
	{0x02045C64946646C2ULL, "mag121_viveka6157199.cl"},
	{0x0204820860C5FA20ULL, "6165662_m.fxe"},
	{0x0204B41B9F1A189BULL, "6153541_m.fxe"},
	{0x0205D2E089622C02ULL, "331129245.ogg"},
	{0x0205F870A59F890BULL, "character_character_p22.psh.fxo"},
	{0x02061EAF4BD2A109ULL, "art\\vfx\\creatures\\fxc_gol_sla_c.mmh"},
	{0x020629E261C2FBA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_1h_sw.ani"},
	{0x0206381474F1ABA5ULL, "6188720_f.fxe"},
	{0x020689E8601E9515ULL, "art\\vfx\\combat\\vfx_glb_potion_drink_c.impa.evt"},
	{0x0206D4B70F5C8D19ULL, "6036479_m.fxe"},
	{0x020725685F9CEB36ULL, "fen110_anso6095012.cl"},
	{0x02072A5B35612331ULL, "ht_0_hightown_a3_21_0_lc.mmh"},
	{0x02072A97B261BDCDULL, "vxlt_gallowstemplar_n_5jc_0_0.vlm"},
	{0x02076B57D021141BULL, "6166462_m.fxe"},
	{0x0207865662F84123ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_crew_0t.dds"},
	{0x0207AEC938AC8924ULL, "dre000_dougal6129743.cl"},
	{0x020803AB2EE647ACULL, "hlac_dr_primevaltomb_11985_0x.dds"},
	{0x02081FC2A3DDADB4ULL, "hlac_w_bonepitdestroyed_a3_393217_0y.dds"},
	{0x020824017CBE5AC9ULL, "horizontalblur5.vsh.fxo"},
	{0x0208774ADBFA0E7AULL, "gen_im_trs_krk_junk1_02.uti"},
	{0x020898C8B70A4CC1ULL, "zz_followers_banter6158460.cl"},
	{0x0208E4D1AA552E55ULL, "dae000_arishok6072839.cl"},
	{0x02093FEC64169731ULL, "645263111.ogg"},
	{0x020965B502B09D7FULL, "vxlt_0_lowtown_a3_n_c_gjs_0_0.vlm"},
	{0x020A25D419B9ADC0ULL, "follower_banter6163720.cl"},
	{0x020A3A5541ED7731ULL, "atl_startmenu2_dxt1_dat.xml"},
	{0x020A479BC42E1B26ULL, "vxw_encounter3_bk_0_0.vlm"},
	{0x020A6AF8362991B7ULL, "6172531_m.fxe"},
	{0x020A9ACAA3B22F5BULL, "static_staticreliefmap_p0.gsh.fxo.dx11"},
	{0x020ACC096FF4920DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_crumble.msh"},
	{0x020ADA096FF4AA40ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_crumble.mao"},
	{0x020ADE096FF4B093ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_crumble.mmh"},
	{0x020B3E46717F51D3ULL, "vxlt_darktown_6j1_0_0.vlm"},
	{0x020B428209C90D32ULL, "vxlt_0_lowtown_docks_n_bba_0_0.vlm"},
	{0x020B4650A609AF8CULL, "vxlt_hangedman_22o_0_0.vlm"},
	{0x020BCFD8BA2CBBE6ULL, "6004095_m.fxe"},
	{0x020C0AB6C549229FULL, "6180963_m.fxe"},
	{0x020CC124DD5024B8ULL, "vxht_0_hightown_a3_c6e_0_0.vlm"},
	{0x020D44EFDF4009A8ULL, "6173870_m.fxe"},
	{0x020D6D06B3CE06D8ULL, "one120_athenril6088920.cl"},
	{0x020DA7D3C64A6445ULL, "307771646.ogg"},
	{0x020E0346DDC07662ULL, "vxdr_primevaltomb_4im_0_0.vlm"},
	{0x020E71DD372E47EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_exit.evt"},
	{0x020EB68AC54E7B52ULL, "6092258_m.fxe"},
	{0x020ED5DB041E2A84ULL, "6095499_m.fxe"},
	{0x020FCC7679D26F17ULL, "6133629_f.fxe"},
	{0x02101AC2ECB96A55ULL, "vxw_woundedcoast_1fk_0_0.vlm"},
	{0x02101F6E80E1AA21ULL, "vxht_chantry_n_33i_0_0.vlm"},
	{0x02101FB4EA80F1FAULL, "hlac_dr_0_deeproads_5845_14y.dds"},
	{0x02104681F05A4770ULL, "isa210ip_at_foundry.ncs"},
	{0x0210559E5F195828ULL, "6072324_m.fxe"},
	{0x021088A84B7C9566ULL, "447407321.ogg"},
	{0x0210C1617F6FC45FULL, "w_woundedcoast_87_0.rml"},
	{0x0211421C0717D53EULL, "vxlt_blackemporium_2o_0_0.vlm"},
	{0x02121A055A16A6C7ULL, "6153985_m.fxe"},
	{0x02121EDCF5CDAE1BULL, "vxw_blightlands_7i_0_0.vlm"},
	{0x0212ECF18D5EB487ULL, "313346623.ogg"},
	{0x02135BA33750A54AULL, "vxlt_hangedman_15s_0_0.vlm"},
	{0x02136EAD83BFD78DULL, "vxlt_undercityserialkiller_144_0_0.vlm"},
	{0x0213A4920FEE4C7AULL, "846670694.ogg"},
	{0x0213B1004BB83B7EULL, "937310838.ogg"},
	{0x021441E261CF619AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_1h_sw.evt"},
	{0x021446153DBEC8E3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushb.msh"},
	{0x021454153DBEE03DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushb.mmh"},
	{0x021472DD3732B498ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_exit.gad"},
	{0x02150C740131EF0DULL, "6155655_m.fxe"},
	{0x0215375A2F9C1B35ULL, "zz_vault_debug.cnv"},
	{0x0215B2408B53C348ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsma.mmh"},
	{0x0215D0408B53F672ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsma.msh"},
	{0x0216D239AD763A26ULL, "vxht_mansion2_30t_0_0.vlm"},
	{0x0216F4B2E2C3E057ULL, "65601744.ogg"},
	{0x02171BD218907065ULL, "enemytrap_ret.mao"},
	{0x0217FFD838128397ULL, "lgt110_gxa_coa_spy6117450.cl"},
	{0x02180BFE242D6575ULL, "vxw_sundermount_20n_0_0.vlm"},
	{0x021813B04C2198FCULL, "gen00fl_anders6125435.cl"},
	{0x0218506CFE2A1775ULL, "455117835.ogg"},
	{0x0218561D7ACD280FULL, "vxlt_0_lowtown_a3_n_c_gx9_0_0.vlm"},
	{0x021884921EDB3423ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\scaredmage_trim.ani"},
	{0x021884F2EE723085ULL, "hlac_lt_0_lowtown_a2_65537_1y.dds"},
	{0x02189D8FA95FE952ULL, "gen_im_arm_cht_hvy_act1_04.uti"},
	{0x0219FBFDEF4973C3ULL, "329007565.ogg"},
	{0x021A2418127276CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_2.gad"},
	{0x021A4BCC7365B4BAULL, "hlac_lt_gallowstemplar_2392_0z.dds"},
	{0x021B01E10E703C72ULL, "gen00fl_anders6196732.cl"},
	{0x021B3AFA15AFC109ULL, "vxca_caverns1_mines_egq_0_0.vlm"},
	{0x021B40E261D57E8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_1h_sw.gad"},
	{0x021B5BBF4DAB4A9CULL, "vxlt_0_lowtown_a2_n_hcs_0_0.vlm"},
	{0x021BAFD318C689C6ULL, "static1_relief.vsh.fxo"},
	{0x021BB65DB9705C27ULL, "dae000_epic_merchant_3.bnk"},
	{0x021BC7B5F05315D0ULL, "21736382.ogg"},
	{0x021BD3E98B462C04ULL, "llac_ht_0_hightown_n_-65535_1.dds"},
	{0x021D1016CC2ABF54ULL, "953924992.ogg"},
	{0x021D847D96D1EB35ULL, "ser100_party_comments6120593.cl"},
	{0x021DCE0EACD5C385ULL, "art\\characters\\playercharacter\\humanfemale\\glv_mede.phy"},
	{0x021DD3E336EF5688ULL, "425418143.ogg"},
	{0x021E46BB61253996ULL, "711405394.ogg"},
	{0x021E8070DD77A9E6ULL, "character_character_p142.psh.fxo"},
	{0x021E8C9FDEAF82D5ULL, "vxw_woundedcoast_1aj_0_0.vlm"},
	{0x021F3A5A2FA47492ULL, "zz_vault_debug.ncs"},
	{0x021F83D797C59D54ULL, "vxca_caverns1_eb2_0_0.vlm"},
	{0x021FA63E3632CCF8ULL, "vxht_mansion2b_2f9_0_0.vlm"},
	{0x021FD85DB9747E9FULL, "dae000_epic_merchant_3.cnv"},
	{0x022041A66A74531FULL, "6143373_m.fxe"},
	{0x0220539132B85BD3ULL, "6113025_f.fxe"},
	{0x022090B9E7C3C69BULL, "vxw_blightlands_8h_0_0.vlm"},
	{0x0220B44EF3679CC2ULL, "genpt_personality.ncs"},
	{0x0220F2EC952EA49BULL, "mag100_cullen6093555.cl"},
	{0x0220FD04B3A6F406ULL, "crt_sf_dust.mao"},
	{0x02210662F24DDBADULL, "llac_dr_primevaltomb_659_10.dds"},
	{0x022139181278B97CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_2.evt"},
	{0x022146CA31057E97ULL, "vxht_mansion_n_4hq_0_0.vlm"},
	{0x02216A07F6B39A5DULL, "mag101_grace6034239.cl"},
	{0x02220C0D9DAF1A86ULL, "fxc_puke_dis.mao"},
	{0x02221DF3D5BB82B4ULL, "vxca_caverns2prc_3nv_0_0.vlm"},
	{0x0223165BADE5E317ULL, "art\\characters\\weapons\\longbows\\w_lbw_dslb.mmh"},
	{0x02231A02D0DD0E6AULL, "313801561.ogg"},
	{0x02231A5BADE5EA24ULL, "art\\characters\\weapons\\longbows\\w_lbw_dslb.mao"},
	{0x02232C5BADE60889ULL, "art\\characters\\weapons\\longbows\\w_lbw_dslb.msh"},
	{0x0223C24584FF7D78ULL, "dae000cr_grand_cleric6147717.cl"},
	{0x0223D18750FAAD40ULL, "356469182.ogg"},
	{0x0223E79AB7338C50ULL, "montage_subs_demohack6217123.cl"},
	{0x02244FA19FF218E0ULL, "509054391.ogg"},
	{0x0224A16B49DAF3D1ULL, "zz_one_debug6099943.cl"},
	{0x0224CE15BC8A146BULL, "6110980_m.fxe"},
	{0x0224CE7897B803C3ULL, "vxw_bonepitmining_a2_mv_0_0.vlm"},
	{0x0224F6C7000B0012ULL, "6208282_m.fxe"},
	{0x0224FFA0317DD02BULL, "6143867_m.fxe"},
	{0x022544169F6C52A6ULL, "vxht_mansion2_37u_0_0.vlm"},
	{0x0225E1FAB4724019ULL, "296174770.ogg"},
	{0x0225F3883D00DB5AULL, "156872740.ogg"},
	{0x022626E07E913D2BULL, "ave200_donnic6110647.cl"},
	{0x022650F4FF41D747ULL, "6110395_m.fxe"},
	{0x0226525556C858BDULL, "static_static_p6.vsh.fxo.dx11"},
	{0x02267DDB16237DF5ULL, "vxw_sundermount_25m_0_0.vlm"},
	{0x02267F1B29B0019DULL, "w_encounter2_39_0.mao"},
	{0x0226831B29B008D6ULL, "w_encounter2_39_0.mmh"},
	{0x02268B87C40DB3BEULL, "626923295.ogg"},
	{0x02268D1B29B019D4ULL, "w_encounter2_39_0.msh"},
	{0x02269C921EE79A1AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\scaredmage_trim.evt"},
	{0x022712E903B57630ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_filecabinet.msh"},
	{0x022720E903B58E4AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_filecabinet.mmh"},
	{0x022724E903B59481ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_filecabinet.mao"},
	{0x0227782A7DA59625ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_add_scl_2.mao"},
	{0x02288362DB2D91C1ULL, "sib131cr_slave_a.utc"},
	{0x0228EB315DBEDE66ULL, "one100ip_st_ilen.ncs"},
	{0x02297069505CAD3AULL, "hlac_lt_0_lowtown_docks_a3_0_5z.dds"},
	{0x0229789D9621920FULL, "fighting60c.mao"},
	{0x02297A90CA00B246ULL, "lgt250_way_dalish_conflict6147155.cl"},
	{0x02297ED8A6821B5BULL, "vxlt_0_lowtown_a2_fff_0_0.vlm"},
	{0x0229C0E22BA8E719ULL, "dae221ip_at_brothel.utp"},
	{0x0229D93B61934B4CULL, "ave151cr_ambush_bandit_f.utc"},
	{0x022A0BF9C052CB0EULL, "255286280.ogg"},
	{0x022A2060EB9A2FD7ULL, "696054786.ogg"},
	{0x022A2FF029B681AFULL, "55116379.ogg"},
	{0x022A41FFFAEFF8DBULL, "narrative_varric6172470.cl"},
	{0x022AEF604BCE4D9FULL, "708385550.ogg"},
	{0x022AF737E6885349ULL, "6117188_f.fxe"},
	{0x022B2065301EC8BEULL, "gen00fl_dog6117796.cl"},
	{0x022B7CDE538D2966ULL, "hlac_lt_undercitydungeon_-1_2y.dds"},
	{0x022B8402F114A092ULL, "w_encounter3_112_0.phy"},
	{0x022BBA10BC3EC603ULL, "w_sundermount_252_0.mao"},
	{0x022BBE10BC3ECD50ULL, "w_sundermount_252_0.mmh"},
	{0x022BCC10BC3EE56AULL, "w_sundermount_252_0.msh"},
	{0x022BD6E29EC0331BULL, "6087011_m.fxe"},
	{0x022BEA2EDABD429BULL, "vxht_brothel_n_6kq_0_0.vlm"},
	{0x022C1A8228640065ULL, "t1_skn_h05.tnt"},
	{0x022C20A9B074545AULL, "44305587.ogg"},
	{0x022C810B09C24F09ULL, "6148129_m.fxe"},
	{0x022CABAF58A3D784ULL, "6027047_m.fxe"},
	{0x022CFE7CD0A59B55ULL, "vxw_woundedcoast_1li_0_0.vlm"},
	{0x022D0FAECE2B6AC9ULL, "vxw_bonepitmining_a2_2a0_0_0.vlm"},
	{0x022D102C9C5C6E4FULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_p.gad"},
	{0x022D4042C717F6D6ULL, "gen_im_arm_hlm_lgt_act2_01.uti"},
	{0x022D9B921EEDB70DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\scaredmage_trim.gad"},
	{0x022DD0EA30F6FB30ULL, "follower_banter6141683.cl"},
	{0x022E0533028631FBULL, "382468831.ogg"},
	{0x022E2549C1DE90BDULL, "lt_undercityencounter_mmap.dds"},
	{0x022E2B14C6109F2CULL, "413602488.ogg"},
	{0x022E3A9BD18729B2ULL, "hlac_ht_0_hightown_a2_131070_0y.dds"},
	{0x022E411812835095ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_2.ani"},
	{0x022E47AFABA6C426ULL, "vxht_0_hightown_n_bxb_0_0.vlm"},
	{0x022E6F1014FE819CULL, "1030398000.ogg"},
	{0x022F1BAF4BF5A13EULL, "art\\vfx\\creatures\\fxc_gol_sla_c.anb"},
	{0x022F6D2268E51C7CULL, "seb300im_resolution_helm.uti"},
	{0x022F7888F7967737ULL, "art\\characters\\playercharacter\\qunarimale\\eye_com.mao"},
	{0x022FA9CB0228115AULL, "273915825.ogg"},
	{0x022FBB681154635CULL, "genip_node_c2.utp"},
	{0x02304233C216AD2FULL, "6128346_f.fxe"},
	{0x0230AA96B8FC5A00ULL, "412768.ogg"},
	{0x0230F3537AB9D84CULL, "hlac_dr_dungeon_3207_5y.dds"},
	{0x0231226A18485B20ULL, "vxlt_0_lowtown_a2_bxw_0_0.vlm"},
	{0x02312D96A8C36BBCULL, "llac_ht_keep_n_3732_14.dds"},
	{0x02316B69A8B0F7A0ULL, "6155319_m.fxe"},
	{0x02316BAC1577D587ULL, "9311751.ogg"},
	{0x023176DD16F23B17ULL, "one200cr_solivitus.ncs"},
	{0x0231C92927ADFF4FULL, "664892581.ogg"},
	{0x0233405589AE1C43ULL, "vxw_bonepitmining_a2_hy_0_0.vlm"},
	{0x023347B99DC0C2CFULL, "vxlt_0_lowtown_a2_n_c_hje_0_0.vlm"},
	{0x023370D4612154A7ULL, "834426679.ogg"},
	{0x02338D738DCA32DEULL, "art\\characters\\weapons\\arrows\\w_arw_001.dura.evt"},
	{0x0234252C9C62B0FCULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_p.evt"},
	{0x02346CF2A2973BEDULL, "444449558.ogg"},
	{0x023478FAFC86D99DULL, "150179660.ogg"},
	{0x0234B1625ACE0E8EULL, "w_blightlands_41_0.msh"},
	{0x0234BF625ACE26DFULL, "w_blightlands_41_0.mao"},
	{0x0234CB625ACE3B7CULL, "w_blightlands_41_0.mmh"},
	{0x0234D2C11CC35A27ULL, "636418678.ogg"},
	{0x0234F5579BA063EAULL, "w_bonepit_221_0.mao"},
	{0x0234F9579BA06A39ULL, "w_bonepit_221_0.mmh"},
	{0x023503579BA07B27ULL, "w_bonepit_221_0.msh"},
	{0x02353E3470ACF388ULL, "hlac_lt_0_lowtown_docks_a3_n_1_0y.dds"},
	{0x02358850CE6D8385ULL, "zz_ave_debug6064130.cl"},
	{0x02369818DFAB56B2ULL, "vxlt_0_lowtown_docks_n_bql_0_0.vlm"},
	{0x0236A5729CF94672ULL, "lt_0_lowtown_docks_a3_39_0.rml"},
	{0x0236C8E828DB4925ULL, "919519717.ogg"},
	{0x0236D91C72665AACULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.cess.ani"},
	{0x023711D3D038C902ULL, "6042255_m.fxe"},
	{0x02376D565624CBCDULL, "730904185.ogg"},
	{0x023775C84B66D460ULL, "ip_restrict_class_war_rog.ncs"},
	{0x02379B459957CADBULL, "521084878.ogg"},
	{0x0237B09A45BA0AC4ULL, "hlac_lt_0_lowtown_a3_n_c_65534_1y.dds"},
	{0x02380E91693C8BD1ULL, "6207576_m.fxe"},
	{0x02383A3BB10EEE1BULL, "6121024_f.fxe"},
	{0x0238592FFA2CA6BAULL, "vxca_caverns2prc_9om_0_0.vlm"},
	{0x02386052D3342779ULL, "hlac_ht_keep_3732_18z.dds"},
	{0x023892F15A957DEFULL, "gen_im_trs_gen_junk2_02.uti"},
	{0x023901ACC75552B3ULL, "t1_skn_e06.tnt"},
	{0x02393FD23C6B59CBULL, "llac_ht_0_hightown_n_131073_0.dds"},
	{0x023990F86F8DB0A4ULL, "w_blightlands_125_0.phy"},
	{0x0239EBF74E7D864BULL, "zz_lgt_3_debug6140435.cl"},
	{0x023A2797B625495CULL, "6122432_m.fxe"},
	{0x023A5C0BCCB35B1BULL, "vxht_brothel_n_6tp_0_0.vlm"},
	{0x023AA422E5D385D0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_basic.msh"},
	{0x023AA4738DD07809ULL, "art\\characters\\weapons\\arrows\\w_arw_001.dura.gad"},
	{0x023AB222E5D39DEAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_basic.mmh"},
	{0x023AB622E5D3A421ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_basic.mao"},
	{0x023ACEDF28B5683EULL, "narrative_banter6149658.cl"},
	{0x023AEEEB4CA80219ULL, "location_states.gda"},
	{0x023AF13AD977130BULL, "hlac_w_encounter2_camp_65534_0x.dds"},
	{0x023AF2E7FBB86789ULL, "6145229_m.fxe"},
	{0x023B12C99AE3C0A6ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_015a_0d.dds"},
	{0x023B829FE687C7D8ULL, "55378402.ogg"},
	{0x023BC63A69B22AA3ULL, "103297414.ogg"},
	{0x023C51D991B1ABBFULL, "6107082_f.fxe"},
	{0x023C67917BB1CE54ULL, "vxca_caverns1_eh0_0_0.vlm"},
	{0x023C6CAA45CCD789ULL, "gen00fl_isabela6096765.cl"},
	{0x023C6D36CD3DEFDFULL, "552445884.ogg"},
	{0x023CB98C9D9CDCA6ULL, "vxht_0_hightown_n_bcc_0_0.vlm"},
	{0x023CFB00BA82A834ULL, "hlac_w_blightlands_393214_0y.dds"},
	{0x023CFEEFEBB90F24ULL, "209325487.ogg"},
	{0x023D13B3B437A7ACULL, "mage_chainlightning.dds"},
	{0x023D2DFFEEE79C2CULL, "mrl310_party_comments6143069.cl"},
	{0x023D588AC92047B0ULL, "gen00fl_anders6142414.cl"},
	{0x023DE5916805FF12ULL, "vxlt_hangedman_n_2k1_0_0.vlm"},
	{0x023DE61C726C8FEEULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.cess.gad"},
	{0x023DFCAE9062AA8BULL, "vxlt_0_lowtown_h8v_0_0.vlm"},
	{0x023F0C06C4E2F102ULL, "darktown_fem_fereldan_sta6156752.cl"},
	{0x023F5EC0F78220A0ULL, "vxlt_warehouse_n_130_0_0.vlm"},
	{0x023F77C784D65EC2ULL, "isa211_isabela6071610.cl"},
	{0x02405262AC4CEC25ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_maso_0n.dds"},
	{0x0240F0AAB0B0B3E5ULL, "56464989.ogg"},
	{0x024104650A71895CULL, "store_lgt_epic2_1.utm"},
	{0x02412D2C9C6D4815ULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_p.ani"},
	{0x024175738DD64757ULL, "art\\characters\\weapons\\arrows\\w_arw_001.dura.ani"},
	{0x024198B2C6E9EDE4ULL, "hlac_ht_chantry_n_2534_28x.dds"},
	{0x0241F5666B754204ULL, "vxlt_backalley_n_1ki_0_0.vlm"},
	{0x02423D1EEF273011ULL, "680076734.ogg"},
	{0x02427BECCF24B08CULL, "gen00fl_fenris6105592.cl"},
	{0x02429F267389183EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction1.evt"},
	{0x024332C194FC60EAULL, "hlac_lt_gallowstemplar_2383_4x.dds"},
	{0x02436136C592C0F5ULL, "lgt360_mag_orlanna6156292.cl"},
	{0x0243FD9C8908DC89ULL, "6099414_m.fxe"},
	{0x02444DABB743CFCAULL, "lt_0_lowtown_a2_n_35_0_lc.mmh"},
	{0x024472CEB9466F31ULL, "6180415_m.fxe"},
	{0x0244E11C7272A5CDULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.cess.evt"},
	{0x0244E5756B805EADULL, "94908601.ogg"},
	{0x0245510C17FF0D8CULL, "vxlt_0_lowtown_docks_b50_0_1.vlm"},
	{0x024559EE033AFB3AULL, "349964445.ogg"},
	{0x0245781855D23A11ULL, "vxlt_gamlenshouse_n_az_0_0.vlm"},
	{0x0245929737115EB7ULL, "rdr111dg_martin6015058.cl"},
	{0x024689B9825FDA2AULL, "746612217.ogg"},
	{0x0246C2A424BDCAE3ULL, "6142217_m.fxe"},
	{0x0246C305BCBFB1EAULL, "940825965.ogg"},
	{0x024732FC9C7EE002ULL, "938400262.ogg"},
	{0x02479D5450111B2AULL, "6206561_m.fxe"},
	{0x0247F1F83EC63AC4ULL, "one200ip_ink_chest.utp"},
	{0x02481FB5E8D4762AULL, "277450829.ogg"},
	{0x02483ACF2FAA7EB4ULL, "zz_followers_banter6158940.cl"},
	{0x02486C7E93A43003ULL, "vxht_0_hightown_n_aig_0_0.vlm"},
	{0x0248B63F031AB599ULL, "6059304_m.fxe"},
	{0x0248BC3898D429C1ULL, "104177900.ogg"},
	{0x0248C6A5A56E2ADFULL, "vxdr_primevalentrance_2zj_0_1.vlm"},
	{0x0248DCF7A5C296A9ULL, "6182080_m.fxe"},
	{0x0249781C911DC7B6ULL, "140498356.ogg"},
	{0x0249A98708109D64ULL, "vxht_0_hightown_bc2_0_0.vlm"},
	{0x0249B626738F5E69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction1.gad"},
	{0x0249BF8654013CAFULL, "6128006_m.fxe"},
	{0x0249E236B491CC55ULL, "vxw_woundedcoast_1rw_0_0.vlm"},
	{0x024AE9DE505D4296ULL, "vxlt_0_lowtown_docks_n_arz_0_0.vlm"},
	{0x024B9863716321D5ULL, "mrl150_merrill6085803.cl"},
	{0x024CD78290B08730ULL, "vxca_cave_1o8_0_1.vlm"},
	{0x024D0EE3502FF0E9ULL, "6025978_m.fxe"},
	{0x024D1D7C666FD1BFULL, "6207853_m.fxe"},
	{0x024D358AB812FF52ULL, "6127316_m.fxe"},
	{0x024DB6DC9E96F290ULL, "20033468.ogg"},
	{0x024DBE802F71D5EAULL, "hlac_w_bonepitdestroyed_a3_393219_0y.dds"},
	{0x024E4416ABD5ED61ULL, "hlac_ht_0_hightown_a3_65536_0x.dds"},
	{0x024EAEBA0D5C9BAEULL, "6057349_m.fxe"},
	{0x024EB2A789E42631ULL, "858665221.ogg"},
	{0x024FB13E6C1A6365ULL, "hlac_ht_0_hightown_n_0_0y.dds"},
	{0x024FFF6108FDD31BULL, "art\\characters\\creatures\\desiredemon\\animation\\wkf.ani"},
	{0x025035D740BE56BEULL, "t3_dog_d06.tnt"},
	{0x02503D9CAE99ED09ULL, "vxht_0_hightown_a2_bau_0_0.vlm"},
	{0x025067435D6B5A84ULL, "vxlt_backalley_n_1vn_0_0.vlm"},
	{0x0250872673952CB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction1.ani"},
	{0x0250CCBAF0E97B23ULL, "gen00fl_varric6133853.cl"},
	{0x02514AF676212CC7ULL, "and111_lirene6084537.cl"},
	{0x0251BD6F6AE47170ULL, "6077312_m.fxe"},
	{0x0251E30F1EB89069ULL, "6137479_m.fxe"},
	{0x0251EE81EE5D82E4ULL, "lt_0_lowtown_docks_a3_20_0.anb"},
	{0x025274477DC543F2ULL, "218628858.ogg"},
	{0x0252C6FB0E76FEF7ULL, "6119532_m.fxe"},
	{0x0252F56283C46734ULL, "41634055.ogg"},
	{0x025307B14F6EEC2EULL, "6019174_m.fxe"},
	{0x02533357F6C6122EULL, "gen00fl_anders6113058.cl"},
	{0x025398AB2A385DE8ULL, "hlac_ht_playershouse_n_855_13y.dds"},
	{0x0253CB625AE8EBCFULL, "w_blightlands_41_0.rml"},
	{0x02540F97CB33E37AULL, "vxht_0_hightown_a82_0_0.vlm"},
	{0x0254560122173FD4ULL, "590983552.ogg"},
	{0x02545A400D40248CULL, "ss_tower.dds"},
	{0x0254D59018073006ULL, "gen_im_arm_glv_lgt_act3_02.uti"},
	{0x0255084CD5EB0D28ULL, "hlac_lt_hangedman_-1_21x.dds"},
	{0x0255491A13189741ULL, "vxlt_0_lowtown_docks_n_9k6_0_0.vlm"},
	{0x025554F761267F1CULL, "follower_banter6169708.cl"},
	{0x02556D8BF96A79F6ULL, "fen110_fenris6100237.cl"},
	{0x0255815440BF1254ULL, "hlac_lt_0_lowtown_a2_0_9z.dds"},
	{0x0255F47C27181A7DULL, "vxht_mansion2b_336_0_0.vlm"},
	{0x02560863D92237D8ULL, "hlac_w_encounter3_-262143_0z.dds"},
	{0x02562763DEB4C348ULL, "dae000_flemeth6013301.cl"},
	{0x025722FC402D01D6ULL, "724129579.ogg"},
	{0x02573FC5B09F8C1BULL, "vxht_brothel_n_6zr_0_0.vlm"},
	{0x0257954D98C78506ULL, "llac_dr_slavers_3216_6.dds"},
	{0x025815FD724FAA29ULL, "6044658_m.fxe"},
	{0x02581B63FA06B5E4ULL, "vxht_0_hightown_bl3_0_0.vlm"},
	{0x025845E5339C224FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_showcase_n.dds"},
	{0x02584E54CED5015BULL, "vxht_keep_73h_0_0.vlm"},
	{0x02592AAB72D76496ULL, "6037407_f.fxe"},
	{0x02599D4681890DA6ULL, "vxht_0_hightown_n_bie_0_0.vlm"},
	{0x025A18190885F1D9ULL, "vxht_mansion2_2og_0_0.vlm"},
	{0x025A279D1F061B93ULL, "hlac_lt_gallowstemplar_n_2384_5x.dds"},
	{0x025AA5E1FE9DF51BULL, "llac_ca_cavern1_haunted_24423_11.dds"},
	{0x025AF4B7BF523D8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_22base.phy"},
	{0x025B91C9286C25ABULL, "185781131.ogg"},
	{0x025BE91477DF6EC4ULL, "gen00fl_bethany6102892.cl"},
	{0x025C046A5662A4DDULL, "mcr310_cut_chantry_blows_up6087560.cl"},
	{0x025C5324CF28199FULL, "hlac_lt_0_lowtown_n_-131072_1y.dds"},
	{0x025CFECD9D653868ULL, "91242934.ogg"},
	{0x025D076109086B12ULL, "art\\characters\\creatures\\desiredemon\\animation\\wkf.evt"},
	{0x025D2B4299919CD0ULL, "6113606_f.fxe"},
	{0x025D41F9DC63EF03ULL, "hlac_lt_gallowstemplar_7198_0z.dds"},
	{0x025E3D81F09C5979ULL, "isa210ip_at_foundry.utp"},
	{0x025FAFE92D5EDBEBULL, "vxw_sundermount_2zo_0_1.vlm"},
	{0x025FB5617FB3844FULL, "w_woundedcoast_87_0.mao"},
	{0x025FB5F984E94411ULL, "hlac_lt_0_lowtown_docks_n_-65538_8x.dds"},
	{0x025FC1617FB398ECULL, "w_woundedcoast_87_0.mmh"},
	{0x025FC7617FB3A2DEULL, "w_woundedcoast_87_0.msh"},
	{0x025FD9D4E79286D4ULL, "art\\vfx\\combat\\vfx_gib_insect_bld_p.mmh"},
	{0x0260448CC75C9BC4ULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\opening.gad"},
	{0x02604DBE320AC89DULL, "6095018_m.fxe"},
	{0x02604F05337F1D9AULL, "mag360cr_cullen.utc"},
	{0x026100C5382F63E3ULL, "gst_hands_clasped.gda"},
	{0x026138D8006D1777ULL, "6114232_m.fxe"},
	{0x026159BEC9BAFDF8ULL, "bl_ar_cl.mao"},
	{0x02616FB3778F53E5ULL, "dae011_corff_sta6149647.cl"},
	{0x0261E9CFDFF19FAFULL, "hlac_w_encounter1_65536_0y.dds"},
	{0x0262AC0EB24D5D56ULL, "hlac_lt_0_lowtown_131070_0y.dds"},
	{0x0262AFB9ADCEA02BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_right.evt"},
	{0x0262F9ED9736A7BBULL, "hlac_lt_0_lowtown_a2_65536_1x.dds"},
	{0x026373D4B6512D51ULL, "vxlt_darktown_2r7_0_0.vlm"},
	{0x0263BAF7050EAFC1ULL, "vxlt_0_lowtown_docks_n_9f5_0_0.vlm"},
	{0x02642661090EBEA5ULL, "art\\characters\\creatures\\desiredemon\\animation\\wkf.gad"},
	{0x026446F9178BE069ULL, "690869382.ogg"},
	{0x02644FAABE544F43ULL, "pro000_colonel_ewald6040584.cl"},
	{0x02646B7FEC6D8EB9ULL, "fen310_fenris6118680.cl"},
	{0x0264731FA137195BULL, "hlac_w_sundermount_196613_0z.dds"},
	{0x02650B17D67936B8ULL, "vxht_keep_4ar_0_0.vlm"},
	{0x026510577E95B338ULL, "6136129_m.fxe"},
	{0x02656087314899C3ULL, "6191297_m.fxe"},
	{0x026579ECD184E77DULL, "vxlt_gallowstemplar_6q6_0_0.vlm"},
	{0x026599F8E706E699ULL, "6053104_m.fxe"},
	{0x0265A917E3455C68ULL, "6119419_f.fxe"},
	{0x0266538CC76120A2ULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\opening.ani"},
	{0x02666951C57B0D9FULL, "73321348.ogg"},
	{0x0266EFAE9FB18471ULL, "vxlt_0_lowtown_a3_n_c_ewh_0_0.vlm"},
	{0x02672D0D21D03746ULL, "658129838.ogg"},
	{0x0267C20032AD6CB6ULL, "104109959.ogg"},
	{0x0267D8AE3DCD4B0AULL, "105475712.ogg"},
	{0x026852AFB03EC354ULL, "ave260cr_beast.utc"},
	{0x0268A1525AE77B9BULL, "llac_lt_0_lowtown_docks_-3_1.dds"},
	{0x0268F211832C54DCULL, "dae000ip_wall_lean_left_d.utp"},
	{0x026958F5446DC387ULL, "follower_banter6172697.cl"},
	{0x0269BB4CED8ABAC8ULL, "vxht_keep_n_67y_0_0.vlm"},
	{0x0269C8B9ADD4E9ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_right.gad"},
	{0x026A14938E848677ULL, "vxlt_gallowscourtyard_63p_0_0.vlm"},
	{0x026A7099EEE93633ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur2_flemeth_fake.ani"},
	{0x026B738DA1A427A6ULL, "1041345516.ogg"},
	{0x026B80A850653455ULL, "venge_revival.dds"},
	{0x026B8F9CFD21E396ULL, "6103628_m.fxe"},
	{0x026B9A899F038477ULL, "mag210_arianni6182207.cl"},
	{0x026BF794A0D124F1ULL, "6068986_m.fxe"},
	{0x026C4D611EAA5BBAULL, "dre000_cut2d_ragstoriches6025103.cl"},
	{0x026C60B52DBCA36BULL, "vxlt_0_lowtown_a3_n_c_h2d_0_0.vlm"},
	{0x026C6BEA2C1E48F6ULL, "428323297.ogg"},
	{0x026CA745D32C53E7ULL, "6161618_m.fxe"},
	{0x026CACF2CA4B95ECULL, "6068001_f.fxe"},
	{0x026D0F2D659924CFULL, "vxlt_0_lowtown_a2_n_c_hva_0_0.vlm"},
	{0x026DB27050D3DD96ULL, "qun231_varnell6078097.cl"},
	{0x026DBEF36DEC23BEULL, "284466219.ogg"},
	{0x026DC69A2616B85CULL, "6067663_m.fxe"},
	{0x026DED693E661AE6ULL, "qun120cr_javaris6042809.cl"},
	{0x026E1AE6C097778BULL, "hlac_ca_caverns2prc_5802_23z.dds"},
	{0x026E89833C8C5DA3ULL, "vxht_0_hightown_a3_avh_0_0.vlm"},
	{0x026EA4741E2C2761ULL, "vxht_keep_n_62s_0_0.vlm"},
	{0x026ECE146E6ED207ULL, "fen311ar_hanged_man.ncs"},
	{0x02703FAA7AEBEEA3ULL, "vxht_0_hightown_ccp_0_0.vlm"},
	{0x027055C7CC34D629ULL, "652132592.ogg"},
	{0x0270C7B9ADDB06EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_right.ani"},
	{0x02714FA435EF58B4ULL, "6090247_m.fxe"},
	{0x0271E5B1A84745D1ULL, "vxlt_darktown_2u6_0_0.vlm"},
	{0x0271F2015B543116ULL, "vxlt_warehouse_v3_0_0.vlm"},
	{0x0272277EC0A3F129ULL, "vxdr_primevalentrance_1k5_0_0.vlm"},
	{0x027294DE91BC749CULL, "gen00fl_aveline6113990.cl"},
	{0x0272AB38769065F1ULL, "215467420.ogg"},
	{0x0273ABA495DBC0E2ULL, "6170639_m.fxe"},
	{0x0273C9A25DFE7C9FULL, "zz_vault_debug6203355.cl"},
	{0x0273DEC27C1EFB64ULL, "vxht_mansion_n_4ym_0_0.vlm"},
	{0x0273E49D2E460A14ULL, "hlac_lt_undercityencounter_-1_48z.dds"},
	{0x0274A076E86E0E2DULL, "6070882_m.fxe"},
	{0x0274AC6522D9DAC9ULL, "hlac_lt_undercitydungeon_-1_22x.dds"},
	{0x0274F11C6558C1ECULL, "lgt260cr_way_mage2.utc"},
	{0x0274F7511E0E6FCAULL, "6053011_m.fxe"},
	{0x027500347AE34E19ULL, "vxht_0_hightown_aib_0_0.vlm"},
	{0x0275181F42E682C3ULL, "324904485.ogg"},
	{0x02751D2C8B447C1EULL, "6028394_m.fxe"},
	{0x0275A626AF0C6EACULL, "hlac_lt_gallowstemplar_f_2383_8y.dds"},
	{0x0275AE146E74BAB6ULL, "fen311ar_hanged_man.lst"},
	{0x0275B9BC1B56FCB2ULL, "23518858.ogg"},
	{0x0275E59D9E3BA78FULL, "dae000_mother6121009.cl"},
	{0x02761B4308C67276ULL, "6135816_m.fxe"},
	{0x02764D0CE8B7A968ULL, "6120562_m.fxe"},
	{0x02768F474D28930AULL, "vxca_caverns2_78s_0_0.vlm"},
	{0x0276AB0A4CD0C18FULL, "68053612.ogg"},
	{0x0276ABA3C9107094ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\runf.evt"},
	{0x0276C541E8CD341BULL, "7123747.ogg"},
	{0x0276CBEA9AED5F4BULL, "vxlt_0_lowtown_docks_b5c_0_1.vlm"},
	{0x02782D29DF80D348ULL, "vxht_keep_n_62v_0_0.vlm"},
	{0x027912A2B1C1B1D9ULL, "6000980_m.fxe"},
	{0x02793056411C1423ULL, "mag130_danzig6016445.cl"},
	{0x027947B0745091D8ULL, "follower_banter6160978.cl"},
	{0x027A1FA4DF5EDB8DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxb.mmh"},
	{0x027A23A4DF5EE2DEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxb.mao"},
	{0x027A31A4DF5EFA13ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxb.msh"},
	{0x027ABABC73AC794FULL, "333123730.ogg"},
	{0x027B02A41EAC4A5DULL, "isa311_castillon6094901.cl"},
	{0x027B55109D08D0D7ULL, "vxw_sundermount_35r_0_0.vlm"},
	{0x027B8BA965055653ULL, "gen00fl_aveline6128412.cl"},
	{0x027BEA96B5DF7979ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robc_t.dds"},
	{0x027C8B1AE273381DULL, "355035268.ogg"},
	{0x027CFB602E827623ULL, "vxht_0_hightown_a3_ask_0_0.vlm"},
	{0x027D4655F65C3796ULL, "381827300.ogg"},
	{0x027D691B35BADDA3ULL, "lgt310_mag_melson6152119.cl"},
	{0x027D805C551D8E21ULL, "486914434.ogg"},
	{0x027D8B8AFA58576CULL, "hm_gen_templar_captain.mor"},
	{0x027E022F3EA45490ULL, "vxht_keep_n_73t_0_0.vlm"},
	{0x027E2C08B41071D2ULL, "llac_lt_0_lowtown_-1_5.dds"},
	{0x027E7D4037CD79B4ULL, "19407311.ogg"},
	{0x027EC06463B60E96ULL, "vxlt_0_lowtown_a2_n_eqf_0_0.vlm"},
	{0x027F073BCFA872C5ULL, "vxht_0_hightown_a89_0_0.vlm"},
	{0x027F4799EEFB4CDDULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur2_flemeth_fake.gad"},
	{0x027F525AAC233FE4ULL, "follower_banter6155901.cl"},
	{0x027FBD146E7D28A1ULL, "fen311ar_hanged_man.are"},
	{0x028052829CE9405EULL, "6128054_m.fxe"},
	{0x0280578E9A3D5E51ULL, "vxlt_darktown_2x5_0_0.vlm"},
	{0x0280A9DF5A67A19FULL, "vxw_encounter3_88_0_0.vlm"},
	{0x028127F2AE0BCA12ULL, "hlac_ca_cavern1_haunted_24421_0y.dds"},
	{0x02814BBEA0167552ULL, "6081849_m.fxe"},
	{0x0281570C79FD40BDULL, "393101068.ogg"},
	{0x02815C49515122C0ULL, "561160858.ogg"},
	{0x0281F84E911F847EULL, "one100dg_nabil6060508.cl"},
	{0x028246213483F73DULL, "6064681_m.fxe"},
	{0x028295C7C1C7D903ULL, "art\\characters\\weapons\\maces\\w_mce_mc105.mao"},
	{0x0282B1026295D5DDULL, "6103616_m.fxe"},
	{0x028318F739A8DDE7ULL, "gen00fl_anders6119190.cl"},
	{0x0283C24824A9EE93ULL, "vxdr_primevaltomb_50w_0_1.vlm"},
	{0x028403D91FDC1D77ULL, "lgt150_nat_were_supporter6110995.cl"},
	{0x02840BF9B2CC62C0ULL, "dae000ip_crowd_lm3_talk7.utp"},
	{0x028431D01D2A2FAEULL, "vxht_mansion2b_2ng_0_0.vlm"},
	{0x028488F452C7F3D9ULL, "171003944.ogg"},
	{0x0284EDEA00104D45ULL, "vrc221_varric6044049.cl"},
	{0x02851894803B1593ULL, "1001536621.ogg"},
	{0x028526D2629C0BFAULL, "vxlt_gallowstemplar_3gr_0_0.vlm"},
	{0x0285D0A0C2D7E1F8ULL, "6110657_m.fxe"},
	{0x02865593A6C1818EULL, "hlac_lt_0_lowtown_docks_-131075_7y.dds"},
	{0x02865E2FBC81FE89ULL, "llac_lt_0_lowtown_-65536_8.dds"},
	{0x02868F6403B0B67BULL, "hlac_lt_0_lowtown_a2_65535_0z.dds"},
	{0x02871D0C5F59DC6DULL, "hlac_lt_0_lowtown_a2_n_65535_5z.dds"},
	{0x02875097E3AD0251ULL, "6139396_m.fxe"},
	{0x0287EA650AADDB8BULL, "6117956_m.fxe"},
	{0x02881446918A4F93ULL, "vxca_cavern1_haunted_e64_0_0.vlm"},
	{0x0288DCD4E7B5908BULL, "art\\vfx\\combat\\vfx_gib_insect_bld_p.anb"},
	{0x0288FB2460E4DF09ULL, "75636167.ogg"},
	{0x0289081F20D67496ULL, "art\\characters\\creatures\\desiredemon\\c_desiredemon.msh"},
	{0x0289121F20D68594ULL, "art\\characters\\creatures\\desiredemon\\c_desiredemon.mmh"},
	{0x0289161F20D68CA7ULL, "art\\characters\\creatures\\desiredemon\\c_desiredemon.mao"},
	{0x0289C69925970997ULL, "vxlt_shop_f5_0_0.vlm"},
	{0x028B18A4DF6D434AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxb.phy"},
	{0x028B2817018DF415ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_1z.dds"},
	{0x028BFCE74A38EF16ULL, "237798594.ogg"},
	{0x028C2838DD7C3E01ULL, "6110523_f.fxe"},
	{0x028C486045D4F7A6ULL, "vxw_encounter3_7t_0_0.vlm"},
	{0x028CBF19C823A177ULL, "vxlt_0_lowtown_e2e_0_0.vlm"},
	{0x028D178474EF90CCULL, "6162436_m.fxe"},
	{0x028E4844E3B6A422ULL, "vxw_bonepitdestroyed_a3_260_0_0.vlm"},
	{0x028E9573104B5D44ULL, "vxht_0_hightown_apt_0_0.vlm"},
	{0x028ED2D9828BC92FULL, "20564381.ogg"},
	{0x028F9888DE575A14ULL, "328645088.ogg"},
	{0x02902EA9B3EAFC6EULL, "gen00fl_anders6121183.cl"},
	{0x02908FDB0897A939ULL, "hlac_lt_0_lowtown_n_131074_0y.dds"},
	{0x0290A5CF4D72AA83ULL, "vxht_0_hightown_n_avj_0_0.vlm"},
	{0x0290CBEDE72AB4EDULL, "vxlt_0_lowtown_a3_n_c_fhz_0_0.vlm"},
	{0x0290F45E66BED147ULL, "vxw_bonepitmining_a2_fk_0_0.vlm"},
	{0x02912BC480E8FD8BULL, "6131209_f.fxe"},
	{0x02915FD197BB9900ULL, "994519018.ogg"},
	{0x0291E2D7C1DF17E4ULL, "vxht_0_hightown_bx7_0_0.vlm"},
	{0x0291EF32F4D033D6ULL, "w_sundermount_3_0.msh"},
	{0x0291F932F4D044D4ULL, "w_sundermount_3_0.mmh"},
	{0x0291FD32F4D04BE7ULL, "w_sundermount_3_0.mao"},
	{0x029205075923471FULL, "6108509_m.fxe"},
	{0x02920D81EE939B44ULL, "lt_0_lowtown_docks_a3_20_0.rml"},
	{0x02922F76521EB8D8ULL, "vxlt_0_lowtown_f0w_0_0.vlm"},
	{0x02926986AAF3DF9DULL, "vxlt_gallowscourtyard_n_745_0_0.vlm"},
	{0x0292B2DAAE6B3493ULL, "919440501.ogg"},
	{0x0293B012A5B6FFFCULL, "vxlt_undercityserialkiller_1jb_0_1.vlm"},
	{0x0293C4971276B8E9ULL, "6118304_m.fxe"},
	{0x02943AE0A555C5DEULL, "579058972.ogg"},
	{0x02944A8CB229A022ULL, "vxlt_0_lowtown_a2_f52_0_0.vlm"},
	{0x029520FE17A4A2FCULL, "follower_banter6165366.cl"},
	{0x02957930E687D009ULL, "vxlt_0_lowtown_a2_hfz_0_0.vlm"},
	{0x0295918808DCFC26ULL, "pro000_merchant6051637.cl"},
	{0x0295E2393A16998CULL, "121507228.ogg"},
	{0x0296284E4444407DULL, "hlac_ht_keep_n_4519_0x.dds"},
	{0x02965E477E7B2CCFULL, "gen00fl_anders6119389.cl"},
	{0x02968A54F1891329ULL, "genip_trap_leghold.utp"},
	{0x0296B11A05029688ULL, "vxht_0_hightown_n_c41_0_0.vlm"},
	{0x029752EE71353978ULL, "ip_restrict_weapon_style.ncs"},
	{0x02980933716989DFULL, "6014873_m.fxe"},
	{0x02989F66BB7418CFULL, "lava_alpha.dds"},
	{0x0298CBCF93F42E1CULL, "vxca_caverns2prc_9u5_0_0.vlm"},
	{0x029903F1D2900E63ULL, "llac_ca_cavern1_haunted_24425_13.dds"},
	{0x0299152960DE80B0ULL, "109036135.ogg"},
	{0x029927306B611236ULL, "6044239_m.fxe"},
	{0x0299520594A2BD6CULL, "hlac_lt_0_lowtown_a2_n_c_-65537_5z.dds"},
	{0x029990E37071D58BULL, "vxht_chantry_n_2it_0_0.vlm"},
	{0x029A382F236533E6ULL, "6094726_m.fxe"},
	{0x029A572F0B92A813ULL, "vxlt_hangedman_n_27h_0_0.vlm"},
	{0x029A5D54C3643DB9ULL, "t3_wdg_w08.tnt"},
	{0x029A69A9F38BBC1FULL, "llac_dr_0_deeproads_5843_3.dds"},
	{0x029B413A054FDA12ULL, "505111542.ogg"},
	{0x029B6279246AC3F4ULL, "vxw_woundedcoast_122_0_0.vlm"},
	{0x029B9AD560F3802CULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb102a_0n.dds"},
	{0x029BCBB7BF898F57ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_22base.mmh"},
	{0x029BCFB7BF899664ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_22base.mao"},
	{0x029BDEBE23F09D14ULL, "6103114_m.fxe"},
	{0x029BE1B7BF89B4C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_22base.msh"},
	{0x029BF80EC3DA43CAULL, "vxw_encounter2_camp_np_0_0.vlm"},
	{0x029C5D4F8ACE9CD1ULL, "zz_ave_debug6174631.cl"},
	{0x029C5F88D3F51109ULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunariamulet_d.dds"},
	{0x029C726C3D0CAD9AULL, "genlock_emissary_def.utc"},
	{0x029CA635A90F0266ULL, "6105820_m.fxe"},
	{0x029CA8D18BACAEE2ULL, "6016158_f.fxe"},
	{0x029D10A5AE9D4244ULL, "vxca_caverns2_9fq_0_0.vlm"},
	{0x029D7E1890832C68ULL, "6078685_m.fxe"},
	{0x029D86201520919FULL, "hlac_ht_playershouse_n_855_26z.dds"},
	{0x029E3B5FC7A805A9ULL, "6206258_m.fxe"},
	{0x029ED15290881224ULL, "vxlt_warehouse_vq_0_0.vlm"},
	{0x029F1B86E8FC204FULL, "vxca_caverns2prc_a9v_0_0.vlm"},
	{0x029F7D576D1CC508ULL, "657507611.ogg"},
	{0x029F7E7288F2AAE4ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_3y.dds"},
	{0x029FDB70D2DF31D4ULL, "vxlt_backalley_g_1jw_0_0.vlm"},
	{0x02A00EA7E7A29926ULL, "follower_banter6160965.cl"},
	{0x02A06817953304B1ULL, "mag110_vincento6167359.cl"},
	{0x02A0D60ACC55E316ULL, "107044949.ogg"},
	{0x02A1158A011660AEULL, "vxht_mansion2b_2dy_0_0.vlm"},
	{0x02A1AC99751A82C5ULL, "vxlt_gallowstemplar_3jz_0_0.vlm"},
	{0x02A2327BDF3EA03DULL, "6116225_m.fxe"},
	{0x02A24D6B4DFFD896ULL, "hlac_ht_keep_3732_0z.dds"},
	{0x02A2917E696AA31BULL, "vxw_blightlands_ec_0_0.vlm"},
	{0x02A29A795D9C1DE6ULL, "art\\vfx\\combat\\vfx_f_and_fury_c.impa.evt"},
	{0x02A2CDB5D7B84DDCULL, "6125705_m.fxe"},
	{0x02A3162BC37D2441ULL, "dae000_mother6004193.cl"},
	{0x02A356A751197BA4ULL, "w_bonepitmining_a2_33_0.msh"},
	{0x02A35E2A8025D3F5ULL, "dae000_junar6147868.cl"},
	{0x02A368A751199A0DULL, "w_bonepitmining_a2_33_0.mao"},
	{0x02A36BA77502FEE6ULL, "6173779_m.fxe"},
	{0x02A36CA75119A146ULL, "w_bonepitmining_a2_33_0.mmh"},
	{0x02A38A746D2D08F5ULL, "6129029_m.fxe"},
	{0x02A3EB0DD87DE889ULL, "vxlt_0_lowtown_a2_hcy_0_0.vlm"},
	{0x02A4B96086580F00ULL, "vxlt_0_lowtown_a3_n_c_hpe_0_0.vlm"},
	{0x02A5456912F3BA69ULL, "rog_combo.dds"},
	{0x02A605DDFC4BEB59ULL, "hlac_lt_0_lowtown_a2_n_65536_3y.dds"},
	{0x02A67C4789B7AAA7ULL, "vxdr_0_deeproads_9ci_0_0.vlm"},
	{0x02A693299BCAE066ULL, "726154384.ogg"},
	{0x02A6ED64987AAF95ULL, "ser221_quentin6128963.cl"},
	{0x02A802C06267EE0BULL, "vxht_chantry_n_2dw_0_0.vlm"},
	{0x02A82ECBB8DCA85DULL, "vxca_caverns1_gje_0_0.vlm"},
	{0x02A8ECA3026133B1ULL, "6092276_f.fxe"},
	{0x02A8F7A8CAB05D6BULL, "vxw_encounter2_65_0_0.vlm"},
	{0x02A91A866AAE0460ULL, "851993401.ogg"},
	{0x02A9381821D4B551ULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_ambient.anb"},
	{0x02A99CF73CDAD747ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_hammer_d.dds"},
	{0x02A9B1795DA26331ULL, "art\\vfx\\combat\\vfx_f_and_fury_c.impa.gad"},
	{0x02AA509B15E6B594ULL, "6104014_m.fxe"},
	{0x02AA5C10D22EC4C9ULL, "208473144.ogg"},
	{0x02AA607DA483879FULL, "vxlt_0_lowtown_a2_er9_0_0.vlm"},
	{0x02AA69EBB5D05C4AULL, "vxw_encounter2_camp_iq_0_0.vlm"},
	{0x02AB22B5BC4EE34BULL, "201096058.ogg"},
	{0x02AB460EC832627BULL, "gen00fl_anders6196825.cl"},
	{0x02AB485921239CC2ULL, "mcr370_assignments6081743.cl"},
	{0x02AB926B64C44145ULL, "6129048_f.fxe"},
	{0x02AC1A761BE06725ULL, "6128807_f.fxe"},
	{0x02ACEE1D27DB45A1ULL, "gen00fl_aveline6197229.cl"},
	{0x02AD2C2979BF7CBCULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\w_var\\bianca_n.dds"},
	{0x02AD5AB299F451ECULL, "412519080.ogg"},
	{0x02AD96EBB42CF207ULL, "mag360_alain6044627.cl"},
	{0x02ADABA98E5340F4ULL, "483920033.ogg"},
	{0x02ADD8184AAB0247ULL, "vxw_bonepitmining_a2_he_0_0.vlm"},
	{0x02AE21AE22BCD664ULL, "8107244.ogg"},
	{0x02AE4D4DC4D54A54ULL, "vxlt_backalley_g_1wt_0_0.vlm"},
	{0x02AEE8C13D0F781FULL, "6106709_m.fxe"},
	{0x02AF5CF4A4286809ULL, "1007223374.ogg"},
	{0x02AF68C24650CF63ULL, "6149124_m.fxe"},
	{0x02B01EC7AEDDCEA5ULL, "atl_background_dxt5_00.dds"},
	{0x02B08D4D45F4083FULL, "6072288_m.fxe"},
	{0x02B0A850F662E9E9ULL, "6112104_m.fxe"},
	{0x02B0B2795DA883DFULL, "art\\vfx\\combat\\vfx_f_and_fury_c.impa.ani"},
	{0x02B1035B5B60BB9BULL, "vxw_blightlands_nb_0_0.vlm"},
	{0x02B13BC190343EDDULL, "vxw_blightlands_121_0_0.vlm"},
	{0x02B141F28DB271A7ULL, "mag170_recruit016094519.cl"},
	{0x02B1478052FA1823ULL, "and100_anders6095865.cl"},
	{0x02B14FF5FB25CC5CULL, "vxca_cavern1_haunted_hm3_0_0.vlm"},
	{0x02B15F866F9EFD36ULL, "442642754.ogg"},
	{0x02B1C2E7E3B5876DULL, "hlac_lt_0_lowtown_n_-131072_8x.dds"},
	{0x02B1ED34CAE92769ULL, "6146784.cut"},
	{0x02B1F35C610AB1C8ULL, "6166282_f.fxe"},
	{0x02B21EE3999357BFULL, "vxht_mansion2_28p_0_0.vlm"},
	{0x02B28C3469ECC7F7ULL, "follower_banter6151272.cl"},
	{0x02B2AF732AB13CC0ULL, "6115782_m.fxe"},
	{0x02B34CFAF091BBF3ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgtb_s.dds"},
	{0x02B39894AA042B7CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urva_0s.dds"},
	{0x02B4280E603BADD3ULL, "hlac_lt_gallowstemplar_2394_3x.dds"},
	{0x02B45C821AB9CA17ULL, "hlac_lt_0_lowtown_-65535_7z.dds"},
	{0x02B46CA75128140DULL, "w_bonepitmining_a2_33_0.rml"},
	{0x02B4EE247BADC327ULL, "vxdr_0_deeproads_9dh_0_0.vlm"},
	{0x02B514BE0BC46318ULL, "6136305_m.fxe"},
	{0x02B5207ABAC23960ULL, "6147198_m.fxe"},
	{0x02B5D6BE406DE493ULL, "6152313_m.fxe"},
	{0x02B5E04E3AB6E654ULL, "nat320_delilah6065917.cl"},
	{0x02B5FEFE8FF0DDA6ULL, "vxlt_0_lowtown_a2_ews_0_0.vlm"},
	{0x02B6085E46C0EAA4ULL, "762393266.ogg"},
	{0x02B614AEA820FBD7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03b.phy"},
	{0x02B651CAF5C50032ULL, "6089947_f.fxe"},
	{0x02B6621EB8A102FEULL, "86297039.ogg"},
	{0x02B6D0E7ADECFE5DULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firexxxl.mmh"},
	{0x02B6F07C89C072F5ULL, "vxw_sundermount_2ks_0_0.vlm"},
	{0x02B714B4EF0770ADULL, "6033692_m.fxe"},
	{0x02B76985BCA675EBULL, "vxw_encounter2_14_0_0.vlm"},
	{0x02B846330856F4F4ULL, "vxw_woundedcoast_180_0_0.vlm"},
	{0x02B8A6554ABD1C86ULL, "472143021.ogg"},
	{0x02B8A6A7D9AF8962ULL, "6139070_m.fxe"},
	{0x02B8BD69756A0C4DULL, "6140864_m.fxe"},
	{0x02B8D25A9679A01FULL, "vxlt_0_lowtown_a2_eu8_0_0.vlm"},
	{0x02B8D77F9DE18DC9ULL, "w_blightlands_3_0.rml"},
	{0x02B9D3C983A06B52ULL, "qun220cr_javaris6071778.cl"},
	{0x02B9E1253ABBC1D3ULL, "264513211.ogg"},
	{0x02B9F8E0384BBCCCULL, "6166611_f.fxe"},
	{0x02BA0CBC6632CD84ULL, "hlac_w_sundermount_196604_0y.dds"},
	{0x02BA7FA59BA51C84ULL, "zz_delete_me.cut"},
	{0x02BCEA6D563A19AAULL, "598101857.ogg"},
	{0x02BD5A51A477F7DEULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.impa.ani"},
	{0x02BD6EE6CBC01AC8ULL, "ts_slavers_prog_junk1.gda"},
	{0x02BD83A0A2FA3F28ULL, "355951251.ogg"},
	{0x02BE0D2E349CED62ULL, "vxht_chantry_n_20f_0_0.vlm"},
	{0x02BE1C600D73F9F1ULL, "hlac_lt_0_lowtown_196606_0x.dds"},
	{0x02BE444546437E07ULL, "335400291.ogg"},
	{0x02BE7E34FF8F9A60ULL, "zz_lgt_2_debug6211863.cl"},
	{0x02BEEE462A746DFAULL, "vxlt_gallowstemplar_3sn_0_0.vlm"},
	{0x02BF2DD73FAA79D5ULL, "vxlt_darktown_5q1_0_0.vlm"},
	{0x02BF5889B30C8288ULL, "dae000_arishok6067764.cl"},
	{0x02BF75384D56D41BULL, "vxw_blightlands_se_0_0.vlm"},
	{0x02BF9B49A87FC5D0ULL, "hlac_lt_0_lowtown_a2_n_-65537_1x.dds"},
	{0x02BF9CABB8ABBDDBULL, "vxw_encounter1_fd_0_0.vlm"},
	{0x02C0254AA047CF83ULL, "311608505.ogg"},
	{0x02C049AD7F364732ULL, "6110407_f.fxe"},
	{0x02C0F742544BCC94ULL, "555452862.ogg"},
	{0x02C121501CA75540ULL, "6110882_m.fxe"},
	{0x02C1BEF522F9A6F3ULL, "qun210cr_guard6087765.cl"},
	{0x02C1E77CD398204DULL, "art\\vfx\\models\\vfx_debris_02.mmh"},
	{0x02C1EB7CD398269EULL, "art\\vfx\\models\\vfx_debris_02.mao"},
	{0x02C1F97CD3983ED3ULL, "art\\vfx\\models\\vfx_debris_02.msh"},
	{0x02C224D8FEC11B27ULL, "vxlt_0_lowtown_n_feh_0_0.vlm"},
	{0x02C2530E3795C55AULL, "6169163_m.fxe"},
	{0x02C27228D8011F84ULL, "llac_lt_0_lowtown_docks_n_65533_3.dds"},
	{0x02C2783B4CC922E4ULL, "hlac_lt_0_lowtown_docks_a3_-131073_3z.dds"},
	{0x02C28C07BE7BB920ULL, "llac_lt_gallowstemplar_n_2391_4.dds"},
	{0x02C2CB7F9DE9CDA9ULL, "w_blightlands_3_0.mao"},
	{0x02C2D77F9DE9E242ULL, "w_blightlands_3_0.mmh"},
	{0x02C2D97F9DE9E5F8ULL, "w_blightlands_3_0.msh"},
	{0x02C3397DD5291221ULL, "6033128_m.fxe"},
	{0x02C3A0B89CE4DE41ULL, "fen210_dying_magister6108609.cl"},
	{0x02C3DBBCD07F7A94ULL, "6176006_m.fxe"},
	{0x02C44D6AA39E1869ULL, "1034795020.ogg"},
	{0x02C48D7D6155CA57ULL, "character_character_p127.psh.fxo.dx11.assembly"},
	{0x02C4C43D0E46E0ADULL, "lgt360_mag_orlanna6156289.cl"},
	{0x02C4C62F3D179B62ULL, "zz_store_dae110cr_store5.utm"},
	{0x02C51BF59D64BCE0ULL, "311737374.ogg"},
	{0x02C562597BB68B75ULL, "vxw_sundermount_2lr_0_0.vlm"},
	{0x02C596208828D545ULL, "hm_raider_2.mor"},
	{0x02C5C2880AC7C6B0ULL, "vxlt_0_lowtown_a2_n_c_hxl_0_0.vlm"},
	{0x02C63C032A18E496ULL, "591695867.ogg"},
	{0x02C68BEE51D1C443ULL, "6133741_m.fxe"},
	{0x02C6D1E1253DCEA5ULL, "qun231_varnell6099592.cl"},
	{0x02C7329FDAA2C9AAULL, "ave200_donnic6111052.cl"},
	{0x02C758157BAA17AAULL, "isa211st_walleyedsam.stg"},
	{0x02C762064FAABBBDULL, "vxdr_primevaltomb_7sw_0_0.vlm"},
	{0x02C7A68F2CF68BB1ULL, "286692659.ogg"},
	{0x02C7BE023347C5CDULL, "6054079_m.fxe"},
	{0x02C7D0CC3133AEC4ULL, "6170237_m.fxe"},
	{0x02C8598D0568C1C7ULL, "169017157.ogg"},
	{0x02C88F32A7CC012DULL, "6144936_m.fxe"},
	{0x02C91CD914F62503ULL, "plt_cod_mgc_chantry_founded.plo"},
	{0x02C9CBC35E3BAE9FULL, "6087122_f.fxe"},
	{0x02C9F2EDB73ABAE5ULL, "hlac_w_bonepit_0_3x.dds"},
	{0x02CA09546AD37403ULL, "w_woundedcoast_61_0.rml"},
	{0x02CA7251A482AAAFULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.impa.evt"},
	{0x02CB1D3F0CE1163CULL, "6201514_f.fxe"},
	{0x02CB2779894D9787ULL, "mag101_grace6053090.cl"},
	{0x02CB5D21F379ED3BULL, "dae000_junar6147853.cl"},
	{0x02CB60B7721F45BCULL, "plt_cod_cha_dao_hero.plo"},
	{0x02CC15F61E8D74BFULL, "229017019.ogg"},
	{0x02CC54AD4FEB766AULL, "968832196.ogg"},
	{0x02CD4BB7C293FECCULL, "6177841_m.fxe"},
	{0x02CDE2F4045461A7ULL, "follower_banter6168027.cl"},
	{0x02CDE7153F4CEC9BULL, "vxw_blightlands_td_0_0.vlm"},
	{0x02CE0C14FEC453DBULL, "w_sundermount_220_0.mmh"},
	{0x02CE0E88AAA1D65BULL, "vxw_encounter1_kg_0_0.vlm"},
	{0x02CE1814FEC46878ULL, "w_sundermount_220_0.mao"},
	{0x02CE2A14FEC486C5ULL, "w_sundermount_220_0.msh"},
	{0x02CE4DE8554EF6E6ULL, "hlac_lt_0_lowtown_n_65536_10x.dds"},
	{0x02CF0955518EAC9BULL, "vxw_sundermount_2gy_0_0.vlm"},
	{0x02CF3FE94F7CE3E3ULL, "gen00fl_anders6125441.cl"},
	{0x02CF56C82A2B1B8BULL, "rob_pl4.tnt"},
	{0x02CF5F617BCE2CA9ULL, "vxlt_warehouse_n_12z_0_0.vlm"},
	{0x02CFA838CD829A69ULL, "hlac_ht_keep_n_3732_12y.dds"},
	{0x02CFF80CBCCFDA47ULL, "vxw_encounter2_q1_0_1.vlm"},
	{0x02D05B7D93A94FDFULL, "zz_personality_debug6004631.cl"},
	{0x02D05C5D2F37A61EULL, "one280_merin6116761.cl"},
	{0x02D0961EAE5B0FC3ULL, "726448729.ogg"},
	{0x02D0ADFA163553DDULL, "77655423.ogg"},
	{0x02D0BB14DB37A956ULL, "6073875_m.fxe"},
	{0x02D17351A488CB58ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.impa.gad"},
	{0x02D2DB0E2902C694ULL, "abi_m_haste.bnk"},
	{0x02D38E7F6C8BDA12ULL, "vxlt_0_lowtown_docks_a3_n_d0v_0_0.vlm"},
	{0x02D39E26AFE0AF81ULL, "414171283.ogg"},
	{0x02D3BA05A29FCD81ULL, "vxca_cavern1_haunted_hmb_0_0.vlm"},
	{0x02D3C1E017D80F8AULL, "gen00fl_anders6119221.cl"},
	{0x02D4068E5C93A810ULL, "6175098_m.fxe"},
	{0x02D43464FCBDDF30ULL, "vxlt_0_lowtown_a2_n_c_hsm_0_0.vlm"},
	{0x02D4B3FB8DEDA2A1ULL, "6144284_m.fxe"},
	{0x02D526BB986A6050ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube.msh"},
	{0x02D534BB986A786AULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube.mmh"},
	{0x02D538BB986A7EA1ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube.mao"},
	{0x02D56B1989E18FC6ULL, "745940270.ogg"},
	{0x02D57DBA372C1645ULL, "620642287.ogg"},
	{0x02D61405551FB652ULL, "vxca_cavern1_haunted_jdh_0_0.vlm"},
	{0x02D6B2E9293FE213ULL, "gen00fl_merrill6106435.cl"},
	{0x02D6CEC3CA19D7C4ULL, "vxht_0_hightown_aoy_0_0.vlm"},
	{0x02D843C8EB36CB9DULL, "hlac_ht_chantry_n_2534_34x.dds"},
	{0x02D8C7AC1C2D1C77ULL, "6150180_m.fxe"},
	{0x02D8F1DFCDBEB7A0ULL, "6207968_m.fxe"},
	{0x02D91E3AE93C2AC6ULL, "vxca_caverns1_mines_drd_0_0.vlm"},
	{0x02DAC481D39B78FDULL, "zz_mrl_debug6185298.cl"},
	{0x02DAF11F57A0F102ULL, "zz_start_dre6080515.cl"},
	{0x02DBCAB9933BBFF6ULL, "vxca_caverns2prc_90a_0_1.vlm"},
	{0x02DC060BEBF2B13BULL, "mer210_seldrek6033962.cl"},
	{0x02DC1EC9D7478CCDULL, "zz_dae_debug6029111.cl"},
	{0x02DC34B32A3FAED1ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc102a_0s.dds"},
	{0x02DCC72D99D76EA2ULL, "6077949_m.fxe"},
	{0x02DCF7BFBF67B393ULL, "gen00fl_anders6125931.cl"},
	{0x02DD41CC9C9F91FBULL, "dae000_arishok6067800.cl"},
	{0x02DD7ADDF4841156ULL, "llac_lt_backalley_g_65536_3.dds"},
	{0x02DD7B324384C51BULL, "vxw_sundermount_2hx_0_0.vlm"},
	{0x02DD83C973DA8490ULL, "499737896.ogg"},
	{0x02DDB92B81CFA9CCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\walkf_injured.ani"},
	{0x02DDD3BCECFCC213ULL, "6185201_m.fxe"},
	{0x02DE80D44E307100ULL, "vxlt_0_lowtown_a3_n_c_hla_0_0.vlm"},
	{0x02DE989E53930909ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_exit.ani"},
	{0x02DF36C81B81F65AULL, "6163363_m.fxe"},
	{0x02DF6E3463E69F45ULL, "hlac_w_woundedcoast_-262143_0y.dds"},
	{0x02E0CA51FBAA3823ULL, "vxht_keep_5j5_0_0.vlm"},
	{0x02E1230644666A13ULL, "vxht_bartrand_n_4sg_0_0.vlm"},
	{0x02E133C7C01AD66FULL, "mag371_elsa6041246.cl"},
	{0x02E1AADF7717BB75ULL, "6167236_m.fxe"},
	{0x02E1C110A3A14BDDULL, "6035184_m.fxe"},
	{0x02E1CF49353BBB28ULL, "291896924.ogg"},
	{0x02E20A797B00059AULL, "655279093.ogg"},
	{0x02E246135FA2BC75ULL, "vxw_sundermount_2zp_0_0.vlm"},
	{0x02E2F478F682B43EULL, "dae200cr_viscount6087392.cl"},
	{0x02E32AEA06F6CEFAULL, "w_encounter3_120_0.mmh"},
	{0x02E32EEA06F6D531ULL, "w_encounter3_120_0.mao"},
	{0x02E33CEA06F6ED40ULL, "w_encounter3_120_0.msh"},
	{0x02E3D11C5BDA5D21ULL, "genip_trap_firecone.utp"},
	{0x02E47F88980FB5D7ULL, "vxlt_darktown_2qd_0_0.vlm"},
	{0x02E4989754E99218ULL, "pro000_athenril6048726.cl"},
	{0x02E4C62B81D5DE0EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\walkf_injured.gad"},
	{0x02E5AEE9FB8CFAACULL, "6097840_m.fxe"},
	{0x02E5B39E5399558BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_exit.gad"},
	{0x02E5D4C7F1954B2CULL, "vxlt_0_lowtown_docks_n_b0q_0_0.vlm"},
	{0x02E621F490122EA8ULL, "hlac_ca_caverns1_mines_24425_8y.dds"},
	{0x02E63008E704FB29ULL, "llac_w_encounter3_-65538_0.dds"},
	{0x02E66AEFE92109A8ULL, "6068486_m.fxe"},
	{0x02E6BC0BEDC7FC7DULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_icon.phy"},
	{0x02E6CC23C3DD381DULL, "19595233.ogg"},
	{0x02E6D3AB5DBAB85BULL, "mcr370_cullen6087462.cl"},
	{0x02E6F932F51884D7ULL, "w_sundermount_3_0.rml"},
	{0x02E739890E2334F7ULL, "6157080_m.fxe"},
	{0x02E74671EDB80317ULL, "538180206.ogg"},
	{0x02E76909A8B1074AULL, "912865520.ogg"},
	{0x02E78E24F5176A6AULL, "gen_im_trs_drk_amu_r1.uti"},
	{0x02E791406EAEB2CCULL, "284234708.ogg"},
	{0x02E7D6A250EE8847ULL, "6179497_m.fxe"},
	{0x02E7E7C6509906A3ULL, "pendant_morningfrost.dds"},
	{0x02E8B6759F3EAD5DULL, "vxlt_0_lowtown_docks_a3_d1i_0_0.vlm"},
	{0x02E8C4B4E05215F1ULL, "hlac_dr_0_deeproads_5839_0z.dds"},
	{0x02E8FAE6D3052EB4ULL, "dae000_gen_turn_in_26198699.cl"},
	{0x02E95581991DB24DULL, "qun220cr_javaris6071520.cl"},
	{0x02E957C2D80EE845ULL, "323500196.ogg"},
	{0x02E971C4CBAA6EBBULL, "zz_one_debug6113373.cl"},
	{0x02E9CAE7A34799EBULL, "gen00fl_aveline6114020.cl"},
	{0x02E9E5EA2EE8FD45ULL, "vxlt_gallowstemplar_3se_0_0.vlm"},
	{0x02EA4C4E963D7FC2ULL, "693270863.ogg"},
	{0x02EA8807530DDB1BULL, "119102636.ogg"},
	{0x02EAB952F3BD25CFULL, "lgt200_eau_tome6171791.cl"},
	{0x02EB2952992C40F9ULL, "zz_followers_banter6158679.cl"},
	{0x02EBAD96831A6475ULL, "hlac_ca_caverns2prc_5802_65z.dds"},
	{0x02EBE972259DE9C7ULL, "hlac_lt_0_lowtown_65537_5y.dds"},
	{0x02EC5C2AA1E427C5ULL, "lgt110_gxa_vgk_silver_order6117024.cl"},
	{0x02ECA86698F6C10BULL, "vxlt_0_lowtown_a2_n_c_er8_0_0.vlm"},
	{0x02ECB09E539F6FF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_exit.evt"},
	{0x02ED76B8B4F4F8BFULL, "514394588.ogg"},
	{0x02EE4217C30797D6ULL, "vxlt_gallowscourtyard_7f3_0_0.vlm"},
	{0x02EEDBD2E05982A6ULL, "6179333_m.fxe"},
	{0x02EFBEABF95F69DEULL, "gen00fl_varric6080086.cl"},
	{0x02EFE19662989EFDULL, "6175529_m.fxe"},
	{0x02F0C7FB82337724ULL, "and321tr_act3_opening.utt"},
	{0x02F1ACC47104D346ULL, "hlac_ht_chantry_n_2534_28z.dds"},
	{0x02F1B632BC04AA79ULL, "hlac_w_bonepit_196614_0z.dds"},
	{0x02F1BAD6CD2254BEULL, "422967507.ogg"},
	{0x02F221542D56D548ULL, "qcr220_grey_wardens6135912.cl"},
	{0x02F28B36020E4708ULL, "828167187.ogg"},
	{0x02F3385F579B46E8ULL, "6122854_m.fxe"},
	{0x02F3446E91885C72ULL, "llac_w_bonepit_131076_0.dds"},
	{0x02F35DFF22829048ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_4z.dds"},
	{0x02F3695C85A0F12AULL, "vxht_0_hightown_a3_b8l_0_0.vlm"},
	{0x02F3AEDA95ADF0FCULL, "vxht_keep_n_36w_0_0.vlm"},
	{0x02F408D6D73B4199ULL, "207475233.ogg"},
	{0x02F4A1A215A57E9FULL, "611107694.ogg"},
	{0x02F4ED2FDE9EF855ULL, "638217596.ogg"},
	{0x02F5107B11685DEEULL, "dae000ip_crowd_ff7_talk13.utp"},
	{0x02F53E944F1CF485ULL, "hf_lilley.mor"},
	{0x02F608943A6335B3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medh_0d.dds"},
	{0x02F61C06F7F98678ULL, "6111190_f.fxe"},
	{0x02F66E609D724B62ULL, "6118589_m.fxe"},
	{0x02F67E631CF5FF01ULL, "art\\vfx\\combat\\material__43.mao"},
	{0x02F6C8E7B055EF43ULL, "art\\vfx\\textures\\vfx_twirl01.dds"},
	{0x02F7BD0BEDD67082ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_icon.mmh"},
	{0x02F7BF0BEDD67438ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_icon.msh"},
	{0x02F8A7BC088CCC24ULL, "vxlt_blackemporium_15q_0_0.vlm"},
	{0x02F9088CDF27858FULL, "gen00fl_varric6118735.cl"},
	{0x02F99C4C0BFD0718ULL, "vxht_keep_61t_0_0.vlm"},
	{0x02FA5668ECBA77E4ULL, "703619480.ogg"},
	{0x02FA5EEC2770F61BULL, "vxw_sundermount_2vz_0_0.vlm"},
	{0x02FA7E655FF8CBD3ULL, "w_bonepitmining_a2_198_0.phy"},
	{0x02FA917C48485D7BULL, "dae000_quintus6154940.cl"},
	{0x02FAFB082202E820ULL, "lowtown_male_urchin_sta6164213.cl"},
	{0x02FB05ACC0E257ECULL, "zz_trans.ncc"},
	{0x02FB47AE51D0FD21ULL, "6164462_m.fxe"},
	{0x02FCECA9429C48C1ULL, "vxht_0_hightown_a3_av2_0_0.vlm"},
	{0x02FCF7546AFEAF1AULL, "w_woundedcoast_61_0.msh"},
	{0x02FD05546AFEC753ULL, "w_woundedcoast_61_0.mao"},
	{0x02FD09546AFECE60ULL, "w_woundedcoast_61_0.mmh"},
	{0x02FD2B01ECE5A6CAULL, "dae000_hate_follower_react6103217.cl"},
	{0x02FD8C54754D1CF6ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_esm_nrmi_0n.dds"},
	{0x02FDF5AEA85E7CCCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03b.mmh"},
	{0x02FDFBAEA85E86BEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03b.msh"},
	{0x02FE384F49BF5826ULL, "vxlt_0_lowtown_a2_enx_0_0.vlm"},
	{0x02FEB0D23DBB2D14ULL, "hlac_lt_gallowstemplar_f_2391_5y.dds"},
	{0x02FECD427A70D93BULL, "dae000ip_crowd_mf6_talk14.utp"},
	{0x02FF4C460013343FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_nud_0s.dds"},
	{0x02FF70A3DE753FB7ULL, "1025070583.ogg"},
	{0x02FF7939266DE08BULL, "6153664_m.fxe"},
	{0x02FFAB210D5FA98CULL, "642661986.ogg"},
	{0x02FFEDA8751E276DULL, "vxlt_hangedman_n_1xl_0_0.vlm"},
	{0x03004868E7B1CED3ULL, "769156700.ogg"},
	{0x0300B24BEE88E3A3ULL, "vxht_0_hightown_c5z_0_0.vlm"},
	{0x0300DE3DE112AA54ULL, "vxw_bonepitmining_a2_27k_0_0.vlm"},
	{0x03017F33D00557FAULL, "vfx0_cs_addv.vsh.fxo"},
	{0x0301824AA96F4FDBULL, "lt_0_lowtown_a3_n_c_20_0_lc.mmh"},
	{0x030197FA354D9EBBULL, "6066348_m.fxe"},
	{0x0301D2D98EA5E84BULL, "vxw_woundedcoast_hk_0_1.vlm"},
	{0x0301DB39779709AAULL, "vxht_0_hightown_a3_b7m_0_0.vlm"},
	{0x0302E0096BBDDC4AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_basic.mao"},
	{0x0302E4096BBDE299ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_basic.mmh"},
	{0x0302EE096BBDF387ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_basic.msh"},
	{0x0303377CF222F66BULL, "6177882_m.fxe"},
	{0x0303A7B3173C40FFULL, "6169171_m.fxe"},
	{0x0303CB1BFF0A12D4ULL, "vxw_bonepitmining_a2_1v3_0_0.vlm"},
	{0x030454E5AABD7BBEULL, "vxlt_backalley_g_1og_0_1.vlm"},
	{0x030461A35AE9B9D7ULL, "gen00fl_fenris6128057.cl"},
	{0x03047D52B352F746ULL, "ht_0_hightown_mmap.dds"},
	{0x0304BA5C34DAB947ULL, "6177697_m.fxe"},
	{0x0304DE64991FB543ULL, "ts_dwarven_prog_junk5.gda"},
	{0x0305301E7B5CB673ULL, "6094486_m.fxe"},
	{0x0305350BAE3F6407ULL, "6087042_m.fxe"},
	{0x03055213766524BBULL, "vxlt_0_lowtown_h95_0_0.vlm"},
	{0x0305AA6EFE110AD5ULL, "vxw_woundedcoast_11z_0_0.vlm"},
	{0x0306ABAF9CE27215ULL, "gen_im_trs_slv_hvy_hlm_r2.uti"},
	{0x0306FDE8DEC83250ULL, "vxlt_gallowstemplar_f_5ao_0_0.vlm"},
	{0x030782584DEF0E97ULL, "vxdr_primevaltomb_56h_0_0.vlm"},
	{0x030897F6522FC6B7ULL, "6094679_m.fxe"},
	{0x0308E899BDD04146ULL, "cr_spe_01.mao"},
	{0x0309680416C611ECULL, "vxht_mansion2_2mt_0_0.vlm"},
	{0x030987F88BEF7627ULL, "vxht_keep_3yf_0_0.vlm"},
	{0x0309A9CED5BFA168ULL, "llac_ht_mansion2b_818_55.dds"},
	{0x0309AD8CB0E26842ULL, "vxw_woundedcoast_36_0_0.vlm"},
	{0x0309B2EB65DC5221ULL, "hlac_lt_0_lowtown_a2_-262144_0z.dds"},
	{0x0309FEF5ED2770B8ULL, "691921873.ogg"},
	{0x030A708DFC14A671ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_talk_1.ani"},
	{0x030A747007754EE2ULL, "ser121_jethann6104565.cl"},
	{0x030B22C0B81340A1ULL, "ss_rogue3.cut"},
	{0x030B7A846F1DD2DCULL, "vxlt_0_lowtown_a3_n_c_hp9_0_0.vlm"},
	{0x030BACB6DADA7DB5ULL, "2106169.ogg"},
	{0x030BCB67F02A41B0ULL, "vxlt_warehouse_137_0_0.vlm"},
	{0x030BDC74D2B2367CULL, "ico_1h_axe_5.dds"},
	{0x030C037DC784B862ULL, "follower_barks6153854.cl"},
	{0x030C116CF7603BFFULL, "tbe200im_ponderer_helm.uti"},
	{0x030C14CB1C17AD67ULL, "vxlt_warehouse_109_0_0.vlm"},
	{0x030C36C986E731EAULL, "vxw_bonepit_268_0_0.vlm"},
	{0x030D3BD6CBB7C200ULL, "165630047.ogg"},
	{0x030D5DD4DE152B7CULL, "775278411.ogg"},
	{0x030E993D0311FE39ULL, "436089737.ogg"},
	{0x030F4DCAF236BFF7ULL, "vxca_caverns2_8d0_0_1.vlm"},
	{0x030F64F855DCE8F9ULL, "vxlt_darktown_4p3_0_0.vlm"},
	{0x030FD09D381F8C4EULL, "6102549_f.fxe"},
	{0x031017891A6548E5ULL, "6164435_m.fxe"},
	{0x03104161F063C3BDULL, "hlac_ht_mansion_n_3370_0z.dds"},
	{0x03105CAE3F4B1B25ULL, "180853079.ogg"},
	{0x0310F8B7B2F3F32DULL, "prty_carver_armor_1.dds"},
	{0x03117B8DFC1AD713ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_talk_1.gad"},
	{0x0311915F1F8301DEULL, "vxlt_warehouse_eb_0_0.vlm"},
	{0x0311AB3AFF67FFF0ULL, "961115753.ogg"},
	{0x0311CE9F238E7E75ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_floortile_s.dds"},
	{0x0311D0AB6D73C144ULL, "qun111_ketojan6041043.cl"},
	{0x0311D55E70474C02ULL, "art\\vfx\\combat\\vfx_glb_ele_n_additive_c.impa.gad"},
	{0x03123CF8F1002B54ULL, "vxw_bonepitmining_a2_1s0_0_0.vlm"},
	{0x0312ADB44EB60540ULL, "vxlt_gallowscourtyard_n_6yr_0_0.vlm"},
	{0x0312B61833B2C5E3ULL, "114051186.ogg"},
	{0x0313B9DA0A7B7C9CULL, "vxlt_0_lowtown_docks_bfx_0_0.vlm"},
	{0x0313F50E293A8A46ULL, "abi_m_haste.wbd"},
	{0x031461FB81778369ULL, "vxlt_0_lowtown_docks_a3_cyb_0_0.vlm"},
	{0x03149A9B3C87DC4FULL, "vxw_woundedcoast_1gz_0_0.vlm"},
	{0x03158EE6137F88A7ULL, "zz_ser_debug6091987.cl"},
	{0x0315AFE839262CA1ULL, "cd_dae000_bartrand.bnk"},
	{0x0315C865588FF381ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v2.ani"},
	{0x0315DEFB8252BAD1ULL, "and321tr_act3_opening.ncs"},
	{0x0316200AE570C452ULL, "543553393.ogg"},
	{0x03162D14FF024110ULL, "w_sundermount_220_0.phy"},
	{0x0316B0CE8652A5BBULL, "vxlt_0_lowtown_eql_0_0.vlm"},
	{0x0316E7755FDA7DBCULL, "hlac_lt_undercityserialkiller_-1_5y.dds"},
	{0x03170F9C548A54C0ULL, "670087323.ogg"},
	{0x03173A95A737BEE6ULL, "hlac_lt_0_lowtown_a2_n_65537_7y.dds"},
	{0x031755ADA5D13121ULL, "hlac_lt_gallowsprison_n_-1_12y.dds"},
	{0x03175CCDD1E98B3BULL, "6093606_f.fxe"},
	{0x0317A12E537644E8ULL, "crack2.dds"},
	{0x0317D24CDA644093ULL, "hlac_w_bonepitdestroyed_a3_65536_0z.dds"},
	{0x0317D9E108BC2A6CULL, "vxht_mansion2_2vu_0_0.vlm"},
	{0x03188D81D8CFBD8BULL, "6147975_m.fxe"},
	{0x0318F85E704DA6F9ULL, "art\\vfx\\combat\\vfx_glb_ele_n_additive_c.impa.evt"},
	{0x03192116CC505F2FULL, "gen00fl_sebastian6175001.cl"},
	{0x03192EB009D8BBFFULL, "llac_lt_0_lowtown_a3_n_c_-3_0.dds"},
	{0x0319344A378FBA8FULL, "hlac_w_woundedcoast_-196607_0x.dds"},
	{0x03197F0F9B773796ULL, "vxlt_house_eq_0_0.vlm"},
	{0x0319A27D215D62D9ULL, "vxht_keep_5zw_0_0.vlm"},
	{0x0319DADD698ED066ULL, "vxca_caverns1_j2a_0_2.vlm"},
	{0x0319E4DABCB3E455ULL, "20361247.ogg"},
	{0x031A3D44E2205A30ULL, "vxlt_warehouse_164_0_0.vlm"},
	{0x031A480DC188A4DEULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_act_c.mmh"},
	{0x031A6713723C2BBCULL, "hlac_ca_caverns2prc_5802_3y.dds"},
	{0x031AA8A678DD4A6AULL, "vxw_bonepit_219_0_0.vlm"},
	{0x031AB88B307076EFULL, "hlac_ht_keep_n_4519_9y.dds"},
	{0x031ABA0515CE1C00ULL, "gen00fl_fenris6126616.cl"},
	{0x031BD36558947223ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v2.gad"},
	{0x031CB410B94BE083ULL, "6190441_m.fxe"},
	{0x031CD162590A586DULL, "vxlt_hangedman_n_1rj_0_0.vlm"},
	{0x031D0A6EB2212141ULL, "vxca_cavern1_haunted_gkj_0_1.vlm"},
	{0x031DC1F7C4FEDB54ULL, "vxw_bonepitmining_a2_29i_0_0.vlm"},
	{0x031E0798349BC033ULL, "vxlt_warehouse_n_x5_0_0.vlm"},
	{0x031E36EF4695D52EULL, "mag311st_keran.stg"},
	{0x031E489FBDBE967FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v1.gad"},
	{0x031E71098096F07FULL, "mrl310_merrill6143819.cl"},
	{0x031EC4AA3747B2D9ULL, "713192954.ogg"},
	{0x031EEA4CB98D2BD2ULL, "vxht_brothel_n_3qw_0_0.vlm"},
	{0x031F035E96C09EF7ULL, "482505535.ogg"},
	{0x031F3853B6151093ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_enter.evt"},
	{0x031F72D83B76A305ULL, "zz_approval_debug6003118.cl"},
	{0x031F9D1D827FAE5DULL, "fen210_orana6133071.cl"},
	{0x03208B6CFB2871FFULL, "6163371_m.fxe"},
	{0x0320F3494EE415A1ULL, "hlac_w_bonepitdestroyed_a3_196612_0x.dds"},
	{0x03210C4E2552BA92ULL, "6143732_m.fxe"},
	{0x03213178681B92C5ULL, "vxht_0_hightown_a2_awp_0_0.vlm"},
	{0x03213760F980B743ULL, "6118438_m.fxe"},
	{0x03213E7C72FF1934ULL, "vxht_bartrand_n_6kj_0_0.vlm"},
	{0x03215AFB8C669358ULL, "454960271.ogg"},
	{0x0321B9278A9E5BF6ULL, "art\\vfx\\materials\\chunky\\vfx_c_spi_hand_vc.mao"},
	{0x0321E22183B4D04BULL, "6163421_m.fxe"},
	{0x0321E6C6C860654CULL, "717178414.ogg"},
	{0x0322352DFAFB37F0ULL, "vxht_chantry_25x_0_0.vlm"},
	{0x0322782E63B7B50AULL, "vxca_cavern1_haunted_fmw_0_0.vlm"},
	{0x03229E2FB5BA1CA7ULL, "266281053.ogg"},
	{0x0322B065589A5508ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v2.evt"},
	{0x0322BD6246656345ULL, "ser270st_templar.stg"},
	{0x0322BFECB844F23BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_roll_back.ani"},
	{0x0323164E82BD3D2EULL, "vxht_mansion2b_29r_0_0.vlm"},
	{0x03236D243D461125ULL, "254026138.ogg"},
	{0x03237876CCF29373ULL, "hlac_lt_undercitydungeon_-1_22z.dds"},
	{0x03237E38D6BA0271ULL, "1053829834.ogg"},
	{0x03239622B6920EEDULL, "ts_circleofmagi_prog_junk1.gda"},
	{0x0323A465993BDC5CULL, "vxlt_undercitydungeon_1c5_0_0.vlm"},
	{0x0323C35E527AF2EBULL, "6077229_m.fxe"},
	{0x03243D9FBDC2EF6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v1.evt"},
	{0x03253153B61970D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_enter.gad"},
	{0x0326211BF2FFFEFCULL, "vxlt_warehouse_n_13l_0_0.vlm"},
	{0x0326809713A77958ULL, "6084372_m.fxe"},
	{0x03269C002613FB0BULL, "6087554_m.fxe"},
	{0x0326D4572BE5C953ULL, "vxlt_backalley_g_1dn_0_0.vlm"},
	{0x0327005E7059F1E8ULL, "art\\vfx\\combat\\vfx_glb_ele_n_additive_c.impa.ani"},
	{0x0327270DC192F69DULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_act_c.anb"},
	{0x03279B5C249ECA7BULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_wood.dds"},
	{0x0327F999BDEB0EE8ULL, "cr_spe_01.dds"},
	{0x0328F3B19432771CULL, "vxht_keep_n_4nd_0_0.vlm"},
	{0x03292601330BAD62ULL, "gen_im_arm_boo_hvy_champion.uti"},
	{0x03292E55A93032E3ULL, "gen00fl_anders6121207.cl"},
	{0x032976A5AF247D44ULL, "131858664.ogg"},
	{0x0329B954FE983938ULL, "hlac_lt_0_lowtown_a2_n_c_65534_1z.dds"},
	{0x0329D201C83627A2ULL, "placeable_core.ncs"},
	{0x0329EF7436B47AA6ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_002a_s.dds"},
	{0x032A199E6DF6B897ULL, "zz_ave_debug6174721.cl"},
	{0x032A2493533D6E40ULL, "vxdr_0_deeproads_f5k_0_0.vlm"},
	{0x032A88DAC7E5D611ULL, "vxlt_undercityserialkiller_135_0_0.vlm"},
	{0x032B201996D6F33EULL, "6118145_m.fxe"},
	{0x032B30408D4B62E1ULL, "915557874.ogg"},
	{0x032B41EBE7F1DB6BULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.cess.ani"},
	{0x032BD971DD730E1EULL, "305643762.ogg"},
	{0x032C4053B61FA8B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_enter.ani"},
	{0x032C48B239C919F9ULL, "vxlt_darktown_4z1_0_0.vlm"},
	{0x032C9A38D9030FFCULL, "vxht_0_hightown_a2_9bj_0_0.vlm"},
	{0x032CAB13B7DFBCC4ULL, "6078335_m.fxe"},
	{0x032DB0A0A577C334ULL, "hlac_w_woundedcoast_0_11y.dds"},
	{0x032DE4941697D401ULL, "vxlt_0_lowtown_a3_n_c_hph_0_0.vlm"},
	{0x032DE4A9AC93387BULL, "crt_sw_bee_gl.mao"},
	{0x032E7519036F32DEULL, "vxlt_warehouse_sd_0_0.vlm"},
	{0x032ECDC3D4AEA212ULL, "mrl210_merrill6102153.cl"},
	{0x032F02722187B09CULL, "vxw_bonepitdestroyed_a3_1zm_0_0.vlm"},
	{0x032F0481B6C042DBULL, "vxlt_hangedman_1xq_0_0.vlm"},
	{0x032F34D3B0CE2E8CULL, "lt_0_lowtown_docks_a3_212_0.msh"},
	{0x032F37FCC1467956ULL, "hlac_lt_0_lowtown_a2_-65535_0x.dds"},
	{0x032F3AD3B0CE397EULL, "lt_0_lowtown_docks_a3_212_0.mmh"},
	{0x032F46D3B0CE4D15ULL, "lt_0_lowtown_docks_a3_212_0.mao"},
	{0x032F5E434C1AB49BULL, "zz_one_debug6077408.cl"},
	{0x032F9F0FF03682FAULL, "6171535_m.fxe"},
	{0x032FC7ECB84F8A32ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_roll_back.evt"},
	{0x032FF0329A650F2BULL, "llac_ca_caverns1_24427_5.dds"},
	{0x0330550CED3A67D2ULL, "vxlt_backalley_g_1dk_0_0.vlm"},
	{0x0330A23BBD605816ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_abilitycharge.ani"},
	{0x0330FD08B003911AULL, "vxlt_0_lowtown_a3_n_c_gjz_0_0.vlm"},
	{0x033149567426ED40ULL, "6165114_m.fxe"},
	{0x0331841A59B5D584ULL, "168460306.ogg"},
	{0x03318573C8A8D0B9ULL, "254109219.ogg"},
	{0x0331DA0C3D546757ULL, "608945896.ogg"},
	{0x0332259FBDCF03C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v1.ani"},
	{0x0332353B44710B6BULL, "6078329_m.fxe"},
	{0x033248236C51AA31ULL, "6153378_m.fxe"},
	{0x0332C3F28022C81AULL, "6068268_m.fxe"},
	{0x033346614B14E009ULL, "415140884.ogg"},
	{0x0333E75BC68C11E7ULL, "6097159_m.fxe"},
	{0x033453404FC9BA54ULL, "dae000cr_grand_cleric6182004.cl"},
	{0x033492F8E4F6177CULL, "vxlt_warehouse_n_18m_0_0.vlm"},
	{0x0334F4D36F9BFAEFULL, "285305121.ogg"},
	{0x0336421CCD410BACULL, "dre000st_varric_lyrium.stg"},
	{0x0336A89E8C83F825ULL, "vxca_cavern1_haunted_huf_0_0.vlm"},
	{0x0336E6ECB855DDC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_roll_back.gad"},
	{0x03375FD155B780F6ULL, "34100833.ogg"},
	{0x0337FF6F719FBDBFULL, "731050005.ogg"},
	{0x033829EBE7FC3CE2ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.cess.evt"},
	{0x033832BEFC56CBE4ULL, "326606094.ogg"},
	{0x0338CB385958275FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_diagonal.msh"},
	{0x0338D13859583101ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_diagonal.mmh"},
	{0x0338DD38595845A2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_diagonal.mao"},
	{0x03395966E74051ADULL, "6068501_f.fxe"},
	{0x033A591FDE00FDB1ULL, "6053603_m.fxe"},
	{0x033A8B01FCF23D56ULL, "vxht_keep_n_4h8_0_0.vlm"},
	{0x033ABA8F2BBF3279ULL, "vxlt_darktown_4a0_0_0.vlm"},
	{0x033BEE460B1E4A86ULL, "vxlt_hangedman_283_0_0.vlm"},
	{0x033BEFF44E96CFCFULL, "505882856.ogg"},
	{0x033CE5B4EFFB1553ULL, "gen00fl_aveline6122137.cl"},
	{0x033CE6F5F5654B5EULL, "vxlt_warehouse_te_0_0.vlm"},
	{0x033CF3419AA3CF4EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobc_0n.dds"},
	{0x033DBA8B3E04E919ULL, "ts_apostates_03.gda"},
	{0x033E223656EB4A34ULL, "vxht_bartrand_n_6qh_0_0.vlm"},
	{0x033E4FAE023D0693ULL, "vxlt_darktown_56r_0_0.vlm"},
	{0x033E82C528562082ULL, "846999526.ogg"},
	{0x033EAA3BBD6CA307ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_abilitycharge.evt"},
	{0x033ED6009F58A9BAULL, "1000753099.ogg"},
	{0x033EF01DD134B384ULL, "pro000_captain_wright6088112.cl"},
	{0x033EF139F91B52D9ULL, "arc_mit_st.mao"},
	{0x033EFCD1A9F711A0ULL, "vxca_cavern1_haunted_j4z_0_0.vlm"},
	{0x033F28EBE80259F5ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.cess.gad"},
	{0x033F373A4169DE38ULL, "dae300_bran6143087.cl"},
	{0x033F7A17B1DF9C94ULL, "hlac_lt_0_lowtown_docks_65536_0y.dds"},
	{0x033FB1F0CC0D382CULL, "vxlt_gallowstemplar_n_3k7_0_0.vlm"},
	{0x0340024E56B5ABE0ULL, "gen00fl_fenris6127985.cl"},
	{0x034023C8171555D0ULL, "combat_st_r1_a5.gda"},
	{0x034099C28A4776D6ULL, "vxw_bonepitdestroyed_a3_1x1_0_0.vlm"},
	{0x03409A1F49747387ULL, "vxlt_warehouse_n_rk_0_0.vlm"},
	{0x0341D6EE8F2D2615ULL, "framebuffereffects_diffusiondof_p5.psh.fxo.dx11"},
	{0x034275C580860B50ULL, "vxlt_backalley_n_1n0_0_2.vlm"},
	{0x0342AE2A2A2C3BD3ULL, "hlac_w_blightlands_327678_0y.dds"},
	{0x0343283347D2A629ULL, "vxlt_backalley_n_1k8_0_0.vlm"},
	{0x03434AF658CF43FFULL, "802579644.ogg"},
	{0x0343D5E4D62FB214ULL, "191764040.ogg"},
	{0x0344111AF65EA291ULL, "686657894.ogg"},
	{0x0344A1458C1A6CF4ULL, "6095868_m.fxe"},
	{0x0344A49DF8E575A2ULL, "vxca_caverns1_mines_edo_0_0.vlm"},
	{0x0344F84C90F732B8ULL, "6749365.ogg"},
	{0x0345444076488FFBULL, "vxca_cavern1_haunted_im4_0_1.vlm"},
	{0x0345CB3BBD72FA50ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_abilitycharge.gad"},
	{0x0345FB83CAA2CD27ULL, "dre200st_dougal_ambush.stg"},
	{0x0346407262B2BC8BULL, "vxlt_0_lowtown_a2_n_c_hov_0_0.vlm"},
	{0x03465E60D0AB18A4ULL, "vxlt_house_e3_0_0.vlm"},
	{0x034660FDC026F160ULL, "hlac_ht_keep_n_4519_18x.dds"},
	{0x034681FE87AA7517ULL, "vxw_bonepit_ni_0_0.vlm"},
	{0x03473C3F954F48A6ULL, "hlac_lt_0_lowtown_a3_n_c_-2_0z.dds"},
	{0x0347FDAC44753ED3ULL, "6143250_m.fxe"},
	{0x0348336A3DDEC326ULL, "6216955_m.fxe"},
	{0x0348875211A2DFF7ULL, "vxlt_0_lowtown_eqx_0_0.vlm"},
	{0x03492C6C1DB54AF9ULL, "vxlt_darktown_4l7_0_0.vlm"},
	{0x034949C4A7D16B62ULL, "6165255_m.fxe"},
	{0x034977353D8C7880ULL, "6015864_m.fxe"},
	{0x03497E5529DA48A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanmale_ambient_cheerwinning_2_darren.gad"},
	{0x034997CBC7A31B5BULL, "sst_ss_livingarmor_1.gda"},
	{0x0349BA98E450C731ULL, "327217579.ogg"},
	{0x034A0563CA114AC1ULL, "6070504_f.fxe"},
	{0x034A0C9A88C194ECULL, "89077315.ogg"},
	{0x034A3321B05A90B8ULL, "vxlt_hangedman_25z_0_0.vlm"},
	{0x034A41447EF4E956ULL, "vxlt_0_lowtown_a3_n_c_enx_0_0.vlm"},
	{0x034A6022FD146306ULL, "vxlt_hangedman_254_0_0.vlm"},
	{0x034B04D1F8CB3593ULL, "vxw_woundedcoast_1d7_0_0.vlm"},
	{0x034B2079ACDDBFFEULL, "ser120_ghyslain6105623.cl"},
	{0x034B3892778B5AF8ULL, "187907354.ogg"},
	{0x034B58D2E75B63DEULL, "vxlt_warehouse_yf_0_0.vlm"},
	{0x034BEEEC1FE9598BULL, "806652237.ogg"},
	{0x034C44FE46FD1134ULL, "6051396_m.fxe"},
	{0x034CC19C7BA636C2ULL, "vxlt_gamlenshouse_n_nt_0_0.vlm"},
	{0x034CD9216DA5708CULL, "6174215_m.fxe"},
	{0x034CFE993A143526ULL, "6067269_m.fxe"},
	{0x034D44AF9721551FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelg_0.pwk"},
	{0x034E4F5D00491F54ULL, "618982592.ogg"},
	{0x034E778FCCAB27EFULL, "gen00fl_aveline6002587.cl"},
	{0x034E87B1802DDFB3ULL, "6084919_m.fxe"},
	{0x034EBB99562F2CBFULL, "t3_psp_b02.tnt"},
	{0x034F0BFC3B6A8C07ULL, "vxlt_warehouse_n_yj_0_0.vlm"},
	{0x034F3249F7EA261DULL, "414573605.ogg"},
	{0x034F7892644E22DEULL, "6113496_m.fxe"},
	{0x035022EEB6D27FB9ULL, "vxlt_0_lowtown_a2_n_c_csg_0_0.vlm"},
	{0x0350F0DFB9FFCCC3ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb2a.mmh"},
	{0x0350FEDFB9FFE41DULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb2a.msh"},
	{0x03513BB49A766ADCULL, "hlac_w_encounter2_196611_0z.dds"},
	{0x0352044C4B99A7C1ULL, "hlac_w_sundermount_0_4x.dds"},
	{0x03522A72C3221461ULL, "vxw_encounter2_mz_0_0.vlm"},
	{0x03526999A2385A8CULL, "gen00fl_merrill6133144.cl"},
	{0x03526FE980F06D7BULL, "gencr_templar_meredith.utc"},
	{0x0352DA465310522DULL, "311475772.ogg"},
	{0x0353B9B09C9DE307ULL, "vxlt_darktown_4ub_0_0.vlm"},
	{0x03547025ED48EA16ULL, "6049434_m.fxe"},
	{0x035528CD38A9AF39ULL, "142073347.ogg"},
	{0x0355851E9D0DEC37ULL, "vxw_bonepitdestroyed_a3_p7_0_0.vlm"},
	{0x0355C13C0A21E72EULL, "vxlt_gallowstemplar_f_2bg_0_0.vlm"},
	{0x0355D39CD1A1D1ABULL, "hlac_lt_0_lowtown_a2_65538_4y.dds"},
	{0x0355FC4ECFCD10DBULL, "578376902.ogg"},
	{0x035601CDA5A30E02ULL, "6099699_m.fxe"},
	{0x03574EC3C1AD2C35ULL, "vxw_encounter2_camp_6s_0_0.vlm"},
	{0x0357E8B74F0CBE10ULL, "vxlt_0_lowtown_a2_n_c_ehy_0_0.vlm"},
	{0x0357FB8237A80385ULL, "6055142_m.fxe"},
	{0x0357FD6E3B1223BEULL, "vxht_keep_674_0_0.vlm"},
	{0x035808A565E562E2ULL, "vxlt_0_lowtown_h6q_0_0.vlm"},
	{0x035871A34AF6F39FULL, "vxlt_gallowstemplar_39e_0_0.vlm"},
	{0x0358C84BD143F1DBULL, "585070178.ogg"},
	{0x0358D1FFEF0A7B86ULL, "vxlt_hangedman_225_0_0.vlm"},
	{0x0359364B5F36C146ULL, "vxht_mansion2b_6u_0_0.vlm"},
	{0x035976AEEAC14E13ULL, "vxw_woundedcoast_1o6_0_0.vlm"},
	{0x035A0B9326490838ULL, "140349955.ogg"},
	{0x035A2E1512A34E21ULL, "461742218.ogg"},
	{0x035A6CD37E264391ULL, "590404810.ogg"},
	{0x035A96B49A8F924AULL, "6083825_m.fxe"},
	{0x035AF58383DCCB62ULL, "vxlt_0_lowtown_ei9_0_0.vlm"},
	{0x035B6F0E56886284ULL, "653549345.ogg"},
	{0x035B9EE5D4EDFF16ULL, "zz_dre_debug6054625.cl"},
	{0x035BF705B95DAFFDULL, "vxlt_0_lowtown_docks_n_bbz_0_0.vlm"},
	{0x035C74B022F6A4B5ULL, "250330734.ogg"},
	{0x035CE4DE0E29D75BULL, "589118319.ogg"},
	{0x035CF1AC9A2FFAE2ULL, "vxht_chantry_n_2yi_0_0.vlm"},
	{0x035D7765604D3D48ULL, "w_bonepitmining_a2_198_0.mmh"},
	{0x035D7D7C6E33A7D6ULL, "vxw_bonepitdestroyed_a3_1f3_0_0.vlm"},
	{0x035D8365604D51EBULL, "w_bonepitmining_a2_198_0.mao"},
	{0x035D90F77D7D7FCCULL, "vxlt_backalley_g_1vs_0_0.vlm"},
	{0x035D9565604D7072ULL, "w_bonepitmining_a2_198_0.msh"},
	{0x035DA95529EB3A88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanmale_ambient_cheerwinning_2_darren.ani"},
	{0x035DBDE83963FA53ULL, "cd_dae000_bartrand.wbd"},
	{0x035DFFCFC3D85F09ULL, "389201594.ogg"},
	{0x035E0328F82B9A8EULL, "qun231_party_comment6077137.cl"},
	{0x035E10ACD79762BBULL, "gen00fl_isabela6097009.cl"},
	{0x035E2F555020138DULL, "21294410.ogg"},
	{0x035EAD9FFE9E210EULL, "vxlt_0_lowtown_a2_n_c_f4x_0_0.vlm"},
	{0x035F3F3D24BED290ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_baserocks_1.mao"},
	{0x035F57EDD80EE336ULL, "fx_m3c6.mao"},
	{0x036044B572DD1E2EULL, "llac_lt_0_lowtown_a3_n_c_0_7.dds"},
	{0x0360BB9B0FC40AC1ULL, "vxht_keep_5gh_0_0.vlm"},
	{0x03612193CF96F56BULL, "vxlt_0_lowtown_a2_n_cmq_0_0.vlm"},
	{0x03619017311D2A5CULL, "follower_banter6160677.cl"},
	{0x0361D9D34CD8EE65ULL, "6047439_m.fxe"},
	{0x0361F1DFBA0E41E8ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb2a.phy"},
	{0x0361F3C045683B88ULL, "vxht_0_hightown_a2_cf8_0_0.vlm"},
	{0x03622BB0D393DB1EULL, "6113904_m.fxe"},
	{0x03626E08691455EAULL, "gen_im_wep_rog_dua_start.uti"},
	{0x036294FB14ACF004ULL, "vxca_caverns2prc_3p8_0_0.vlm"},
	{0x0362A34322DA0C58ULL, "qun221_petrice6083823.cl"},
	{0x0362BEDB452AD2B3ULL, "623643582.ogg"},
	{0x0362CECA72101FE6ULL, "hlac_lt_0_lowtown_docks_-65538_4x.dds"},
	{0x0362FC8AB226E0B5ULL, "gen00fl_anders6113037.cl"},
	{0x0363ED5F2FE9F133ULL, "gen00fl_dog6115072.cl"},
	{0x03644F7AAC5F617EULL, "505504122.ogg"},
	{0x036479429AEC5628ULL, "llac_lt_backalley_g_1_7.dds"},
	{0x0364C597A70CD3A0ULL, "vxlt_darktown_6j6_0_0.vlm"},
	{0x0364E5096C10AE86ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_basic.phy"},
	{0x0365C0A0B3A344B5ULL, "vxw_encounter2_camp_1r_0_0.vlm"},
	{0x03666F4B2D083C3EULL, "vxht_keep_605_0_0.vlm"},
	{0x036686EC2EF6E601ULL, "355940205.ogg"},
	{0x0367658A61E1F2A0ULL, "vxca_cavern1_haunted_gf1_0_0.vlm"},
	{0x03678E48FB713A3CULL, "ht_0_hightown_a2_21_0_lc.mmh"},
	{0x0367AC07DE703601ULL, "vxlt_0_lowtown_a3_n_c_hll_0_0.vlm"},
	{0x0367E88BDCB76693ULL, "vxw_woundedcoast_1j9_0_0.vlm"},
	{0x0368963A0936CA57ULL, "gen00fl_carver6108934.cl"},
	{0x0368C9E5E960129CULL, "vxw_bonepitdestroyed_a3_1na_0_0.vlm"},
	{0x0368FBB84BB36E0AULL, "6083710_f.fxe"},
	{0x036909B1D3C035B6ULL, "247101991.ogg"},
	{0x03690B67CBFD6924ULL, "vxlt_warehouse_8c_0_0.vlm"},
	{0x036916776B4AA0A2ULL, "6000741_m.fxe"},
	{0x036932E2C9EC2D0AULL, "vxlt_0_lowtown_a2_n_c_e1h_0_0.vlm"},
	{0x0369924438405536ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_015a_0t.dds"},
	{0x036A304A909BC593ULL, "gen00fl_anders6116119.cl"},
	{0x036A4CDEEAE82CD6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffidle.ani"},
	{0x036AF58BA36C905FULL, "27831013.ogg"},
	{0x036C018A49AAF34BULL, "lt_gallowstemplar_f_2394.rml"},
	{0x036CB504386FDE18ULL, "vxca_caverns1_mines_j42_0_0.vlm"},
	{0x036CEA7E3D66AE46ULL, "6075496_f.fxe"},
	{0x036D91F0F191728AULL, "gen00fl_fenris6112042.cl"},
	{0x036DA22BE8592401ULL, "6151116_m.fxe"},
	{0x036DDCC8B1810F71ULL, "dae000_gen_turn_in_26178485.cl"},
	{0x036E09107FD28502ULL, "mag160_feynriel6020423.cl"},
	{0x036E163CDF24D05AULL, "309465768.ogg"},
	{0x036E47D2E0A7B1A1ULL, "hlac_lt_0_lowtown_a2_n_131072_3x.dds"},
	{0x036EA4A0EDEA36DBULL, "follower_banter6163655.cl"},
	{0x036EBF898AB34F3CULL, "693183747.ogg"},
	{0x036F5825486A683BULL, "vxht_chantry_n_2c2_0_0.vlm"},
	{0x036FD46D5DF7D10BULL, "132643051.ogg"},
	{0x036FFED0DE598802ULL, "mcr371_hostile_fenris6082255.cl"},
	{0x03701BF6BBBAE67BULL, "hlac_ht_keep_4519_10z.dds"},
	{0x037041677DA2228CULL, "6071890_f.fxe"},
	{0x03705453B9D7E27EULL, "845927087.ogg"},
	{0x0371F672BB4E7E91ULL, "vxdr_0_deeproads_ewg_0_0.vlm"},
	{0x03728C4CF6E40846ULL, "hlac_lt_0_lowtown_docks_a3_-131073_3x.dds"},
	{0x0372E21B7D044978ULL, "hlac_w_blightlands_0_11y.dds"},
	{0x0372E587898F3F02ULL, "6093499_m.fxe"},
	{0x037307A85113FFCEULL, "212942974.ogg"},
	{0x037322A85E20286AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_cap.mao"},
	{0x037326A85E202EB9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_cap.mmh"},
	{0x037330A85E203FA7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_cap.msh"},
	{0x0373AD1A06A943CCULL, "100788062.ogg"},
	{0x0373BA87F3DF0577ULL, "gen00fl_sebastian6180841.cl"},
	{0x0373D199799DF6DEULL, "521205737.ogg"},
	{0x0373F57473176324ULL, "6166238_m.fxe"},
	{0x03755DEA508EC3A9ULL, "800259845.ogg"},
	{0x0375D716E7F0CB79ULL, "152554641.ogg"},
	{0x03761DE4D0664E81ULL, "vxlt_0_lowtown_a3_n_c_hic_0_0.vlm"},
	{0x03764F630AD7550AULL, "vxlt_0_lowtown_h83_0_0.vlm"},
	{0x03768B36E3B8D2BFULL, "hlac_ca_caverns2prc_5802_27z.dds"},
	{0x0376E54007E6BEACULL, "vxlt_0_lowtown_fgj_0_0.vlm"},
	{0x037754DEEAF2C4C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffidle.evt"},
	{0x037783D6113BE61EULL, "large_d.dds"},
	{0x0377B0E9BACA1B49ULL, "6164457_m.fxe"},
	{0x0377E0D20F482DD8ULL, "vxca_caverns2prc_7wu_0_0.vlm"},
	{0x03788686E405BD00ULL, "1032962997.ogg"},
	{0x0378D3644A350B47ULL, "mag170_hugh6080801.cl"},
	{0x03792F35CE1B4182ULL, "vxlt_darktown_2r8_0_0.vlm"},
	{0x0379931E9F77BDB5ULL, "gen00fl_merrill6151099.cl"},
	{0x0379DEDC5B7F298EULL, "543078483.ogg"},
	{0x0379FA767A3A0F69ULL, "w_bonepitdestroyed_a3_198_0.phy"},
	{0x037A74B16169B0CCULL, "vxlt_backalley_g_1xq_0_0.vlm"},
	{0x037AB6AAF319B9F2ULL, "ht_brothel_n_3370.anb"},
	{0x037AE2741CA86135ULL, "vxlt_foundry_n_142_0_0.vlm"},
	{0x037AE7D307F69F89ULL, "gen00fl_varric6087600.cl"},
	{0x037B24D12EBE4AA5ULL, "mag210_arianni6021223.cl"},
	{0x037B50583A01A549ULL, "isa321_velasco6094489.cl"},
	{0x037BC554C7E4FE5BULL, "6180097_m.fxe"},
	{0x037C256A723F5AD1ULL, "96182818.ogg"},
	{0x037C27CC05168EFAULL, "86324526.ogg"},
	{0x037CFC93F2AF4635ULL, "vxca_caverns1_j5q_0_0.vlm"},
	{0x037D1801882D3E41ULL, "mag370_meredith6092172.cl"},
	{0x037D4E59A605E556ULL, "6078460_f.fxe"},
	{0x037D68EDD827FDB8ULL, "fx_m3c6.dds"},
	{0x037D85A9641DBF01ULL, "gen00fl_sebastian6195849.cl"},
	{0x037DA77754B7BEA5ULL, "vxlt_undercitydungeon_1jc_0_0.vlm"},
	{0x037E6F033D890141ULL, "784904153.ogg"},
	{0x037E75DEEAF91B10ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffidle.gad"},
	{0x037EF5CB79C5FE17ULL, "vrc321ip_spooky6_obj_6.utp"},
	{0x037FD39971E4F0A5ULL, "6080834_m.fxe"},
	{0x03802555A296DDEBULL, "6094891_m.fxe"},
	{0x0380684FAD449711ULL, "vxdr_0_deeproads_epf_0_0.vlm"},
	{0x0380EBB167F49D4EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1c.msh"},
	{0x0380F9B167F4B49FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1c.mao"},
	{0x038105B167F4C93CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1c.mmh"},
	{0x03815695388B17ECULL, "dae000_mother6028523.cl"},
	{0x03825CEF969E5649ULL, "ico_gloves_heavy_4.dds"},
	{0x03825E433F4005E3ULL, "6106899_m.fxe"},
	{0x038287B4058A220AULL, "dae000_madam_lusine6151391.cl"},
	{0x038386C80EE345CAULL, "305820507.ogg"},
	{0x0383D6DEE4EE2C54ULL, "hlac_lt_0_lowtown_docks_0_3y.dds"},
	{0x0384338C88BBA5D8ULL, "t3_boy_b02.tnt"},
	{0x038438585BA931FCULL, "hlac_dr_slavers_3217_8y.dds"},
	{0x038439E3C2AAC42BULL, "combat_arishock_combo.gda"},
	{0x0384DBEEE12FC14CULL, "6105439_m.fxe"},
	{0x0385571CF9DCD72CULL, "vxlt_0_lowtown_fhk_0_0.vlm"},
	{0x0385970B01E09668ULL, "dre000_vampire6034337.cl"},
	{0x0386497ED977EAE9ULL, "160319597.ogg"},
	{0x0386A9D4C56E0717ULL, "hlac_lt_0_lowtown_0_4x.dds"},
	{0x0386B7F66F897C03ULL, "205614137.ogg"},
	{0x03874C9C8F3FF97DULL, "vxlt_0_lowtown_docks_n_bqw_0_0.vlm"},
	{0x03877155931A6F62ULL, "403055393.ogg"},
	{0x0387A112C0115A02ULL, "vxlt_darktown_2u9_0_0.vlm"},
	{0x038856B320127A62ULL, "426966247.ogg"},
	{0x03887B44BA860D40ULL, "24943529.ogg"},
	{0x0388E6B33ECBC904ULL, "vxca_cavern1_haunted_icq_0_1.vlm"},
	{0x03893C234B71A47DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_laundry_sa.dds"},
	{0x0389647BE07A165EULL, "vxw_encounter3_8u_0_0.vlm"},
	{0x0389831CCF95EB17ULL, "vxdr_primevaltomb_5cc_0_0.vlm"},
	{0x0389CA80D95C61F4ULL, "125220886.ogg"},
	{0x038A38B7A9C0B4BBULL, "6155802_m.fxe"},
	{0x038B090B71FAB5EDULL, "zz_lgt_2_debug6206587.cl"},
	{0x038B367FF78977CCULL, "478867279.ogg"},
	{0x038B38CF90972ED0ULL, "6033565_m.fxe"},
	{0x038B3A51B7A5F0A9ULL, "605741718.ogg"},
	{0x038C3572DFE28CA3ULL, "vxdr_primevaltomb_1ks_0_0.vlm"},
	{0x038C58E8368B8642ULL, "vxht_playershouse_n_1ld_0_0.vlm"},
	{0x038CBB74945C3D36ULL, "6094307_f.fxe"},
	{0x038CF6262E8661EFULL, "ser221_party_comments6209854.cl"},
	{0x038CFC28D6765150ULL, "213385057.ogg"},
	{0x038D942CABB00BBEULL, "6102919_m.fxe"},
	{0x038E21B60FEF18A9ULL, "652846022.ogg"},
	{0x038E4FC763F0E50FULL, "341866816.ogg"},
	{0x038E6BFA2FB87056ULL, "134042949.ogg"},
	{0x038E99D12969CD54ULL, "w_encounter1_8_0.phy"},
	{0x038F3B4B84E1F1D3ULL, "dae00st_gamlens_house.stg"},
	{0x038F94E32FE63DAAULL, "109549006.ogg"},
	{0x038FB0F4D371EF66ULL, "ht_0_hightown_a3_42_0_lc.mmh"},
	{0x038FE85E11FF2068ULL, "vxlt_undercityserialkiller_10i_0_0.vlm"},
	{0x0390259F07ED5CABULL, "6077420_f.fxe"},
	{0x03903542073422A8ULL, "gen00fl_merrill6108674.cl"},
	{0x03907F72E25E2633ULL, "51354205.ogg"},
	{0x0390D8D2CBA85FB7ULL, "zze_roguevshighdragon_track.cut"},
	{0x0390F4E642E984E4ULL, "vxdr_primevaltomb_6x6_0_0.vlm"},
	{0x03917FC99FA62D75ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_dwn_c.mmh"},
	{0x0392049EA8680FC3ULL, "6031307_f.fxe"},
	{0x03925B850A02BFF3ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.cess.evt"},
	{0x03927A45C32E7A54ULL, "284694658.ogg"},
	{0x0392FB4CE4769BDBULL, "vxlt_foundry_n_108_0_0.vlm"},
	{0x03933124060DBC14ULL, "6044228_m.fxe"},
	{0x039332FF6318F3FBULL, "vxlt_0_lowtown_n_cmm_0_0.vlm"},
	{0x0393573F36E56397ULL, "one280_iwan6096866.cl"},
	{0x03937AF17AAC606BULL, "zz_vrc_debug6064294.cl"},
	{0x0393A836068E3B03ULL, "vxw_bonepit_26u_0_0.vlm"},
	{0x0393B7190B5A7C3DULL, "zz_mer_debug6017035.cl"},
	{0x039468AA4143DAE9ULL, "vfx_lightnig_dot.dds"},
	{0x03954D7DACBC95F5ULL, "rog_slowcloud.dds"},
	{0x039590D1296FDCFCULL, "w_encounter1_8_0.rml"},
	{0x0395BA9E8C22591CULL, "vxlt_0_lowtown_docks_b5q_0_0.vlm"},
	{0x03969C5CA6BBA982ULL, "dae000_worthy6166568.cl"},
	{0x0397115745C25388ULL, "vxca_cavern1_haunted_flv_0_0.vlm"},
	{0x03973A88F76EBDD3ULL, "vxw_encounter2_camp_qu_0_1.vlm"},
	{0x039759FB96A49617ULL, "6126985_m.fxe"},
	{0x0397D658D2702EDEULL, "vxw_encounter3_5v_0_0.vlm"},
	{0x03984F5C0D09D443ULL, "6118669_m.fxe"},
	{0x039855BB74CEAE6DULL, "vrc321_varric6084444.cl"},
	{0x0398D4E642A698C2ULL, "zz_followers_banter6158292.cl"},
	{0x0399522E935BE1AEULL, "hlac_w_encounter1_-131072_0z.dds"},
	{0x039954850A08D234ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.cess.gad"},
	{0x039976D129739951ULL, "w_encounter1_8_0.msh"},
	{0x039984D12973B16CULL, "w_encounter1_8_0.mao"},
	{0x039990D12973C50FULL, "w_encounter1_8_0.mmh"},
	{0x0399974FC66B2A6EULL, "lt_0_lowtown_docks_29_0_ll.mmh"},
	{0x0399D26BA9DEA809ULL, "vxlt_0_lowtown_docks_c6l_0_0.vlm"},
	{0x039A035A35F4E7BDULL, "follower_banter6163204.cl"},
	{0x039A346F988EEF33ULL, "hlac_lt_0_lowtown_a2_n_-131073_3z.dds"},
	{0x039B03AEE19620BCULL, "art\\cinematics\\additional animation\\cs_femnobleiii_cinematicanim.ani"},
	{0x039B61BD0D43E992ULL, "vxht_mansion2b_322_0_0.vlm"},
	{0x039B83BDC8A33CBEULL, "gen00fl_aveline6139149.cl"},
	{0x039BBF69072D43B3ULL, "379748269.ogg"},
	{0x039C475C9D31D481ULL, "1063577718.ogg"},
	{0x039C9085C92F035AULL, "hlac_lt_0_lowtown_65533_1x.dds"},
	{0x039CD0D539384C41ULL, "rdr111dg_martin6054583.cl"},
	{0x039D4C099130C811ULL, "vxdr_0_deeproads_efd_0_0.vlm"},
	{0x039D8573A8A5FD83ULL, "llac_lt_darktown_-1_45.dds"},
	{0x039DE6912EF297BFULL, "6108308_f.fxe"},
	{0x039E15B8B7FDCC03ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.impa.evt"},
	{0x039E7CC99FB0B202ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_dwn_c.anb"},
	{0x039E7DBBB53A1F17ULL, "llac_w_encounter3_-131071_0.dds"},
	{0x039F1AACAF0DECFCULL, "100311459.ogg"},
	{0x039F63850A0D5712ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.cess.ani"},
	{0x039F8ED9A845E795ULL, "6098822_m.fxe"},
	{0x03A051CE9D50395AULL, "vxdr_0_deeproads_fbp_0_0.vlm"},
	{0x03A0581ABB155EFDULL, "6121173_m.fxe"},
	{0x03A092AE5447E641ULL, "tint_override.gda"},
	{0x03A098A3A08B143BULL, "vxca_cave_18c_0_0.vlm"},
	{0x03A0FBEC1718A0DDULL, "hlac_ht_bartrand_n_3365_1z.dds"},
	{0x03A134D51B1707A6ULL, "vxlt_0_lowtown_byr_0_0.vlm"},
	{0x03A210AEE19C55FEULL, "art\\cinematics\\additional animation\\cs_femnobleiii_cinematicanim.gad"},
	{0x03A3810A327CEC09ULL, "zz_mer_debug6062750.cl"},
	{0x03A3BE2213DF78B0ULL, "vxlt_hangedman_n_2xn_0_0.vlm"},
	{0x03A3C0E219700977ULL, "3174314.ogg"},
	{0x03A404CF815F620CULL, "vxht_0_hightown_n_avq_0_0.vlm"},
	{0x03A434A4C7415928ULL, "dre100cr_pickpocket.utc"},
	{0x03A46BE5E4734951ULL, "w_bonepitmining_a2_142_0.mmh"},
	{0x03A477E5E4735DF2ULL, "w_bonepitmining_a2_142_0.mao"},
	{0x03A485E5E473750FULL, "w_bonepitmining_a2_142_0.msh"},
	{0x03A4A079D2692737ULL, "smoke_water_add.mao"},
	{0x03A4E21CCF3E3AA1ULL, "vxht_bartrand_n_6r8_0_0.vlm"},
	{0x03A4EEB8B803A884ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.impa.gad"},
	{0x03A58F19F6A6464CULL, "6111386_m.fxe"},
	{0x03A5E3767A5F228EULL, "w_bonepitdestroyed_a3_198_0.mmh"},
	{0x03A5EF767A5F3725ULL, "w_bonepitdestroyed_a3_198_0.mao"},
	{0x03A5FD767A5F4F7CULL, "w_bonepitdestroyed_a3_198_0.msh"},
	{0x03A78BD3C4BA0A64ULL, "vxw_bonepit_zz_0_0.vlm"},
	{0x03A7A4C2123B4199ULL, "786890948.ogg"},
	{0x03A83350CCEF2559ULL, "gen00fl_isabela6135659.cl"},
	{0x03A83C10BCFDCC73ULL, "gen00fl_sebastian6189006.cl"},
	{0x03A86E1DC1683838ULL, "lgt200ip_eau_tome02.utp"},
	{0x03A8856F6493E6FFULL, "dae100_viscount6101974.cl"},
	{0x03A890CBAA5E1BB0ULL, "146076095.ogg"},
	{0x03A9C8D6851215E6ULL, "gen_im_npc_arm_hum_cht_cota.uti"},
	{0x03AA55A5B692746FULL, "hlac_lt_0_lowtown_n_0_3x.dds"},
	{0x03AB252A3B51D7D0ULL, "6141549_m.fxe"},
	{0x03AB280F74041E2FULL, "1005817762.ogg"},
	{0x03AB8B1717C24C70ULL, "460566625.ogg"},
	{0x03ABD9201B5B295CULL, "hlac_ht_0_hightown_131070_0y.dds"},
	{0x03ABFDB8B809E162ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.impa.ani"},
	{0x03AC173E9397A18CULL, "1068858671.ogg"},
	{0x03AC8A625F1A207AULL, "6177843_m.fxe"},
	{0x03ACE0349A804578ULL, "vxw_bonepitmining_a2_8f_0_0.vlm"},
	{0x03AD006C4A2971AAULL, "289791914.ogg"},
	{0x03AD57DF752E177CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic_3.mao"},
	{0x03AD64E4C523C283ULL, "109768444.ogg"},
	{0x03ADB9AAF345182EULL, "ht_brothel_n_3370.rml"},
	{0x03AE065B10866593ULL, "ht_0_hightown_a3_13_0.rml"},
	{0x03AE0F3B76F038FCULL, "535324543.ogg"},
	{0x03AE7BE980F950E8ULL, "zz_qun_debug6085732.cl"},
	{0x03AF16B8444E4647ULL, "6083994_m.fxe"},
	{0x03AF5D6E9E60D50CULL, "73393359.ogg"},
	{0x03B051B62DFA5C54ULL, "zz_ser_debug6091928.cl"},
	{0x03B0AFC2E38C7528ULL, "and111_lirene6044961.cl"},
	{0x03B0EC059EFCE810ULL, "ser210_witness6168174.cl"},
	{0x03B104309D912571ULL, "zz_approval_debug6003525.cl"},
	{0x03B13729A981565CULL, "llac_lt_0_lowtown_a2_n_2_0.dds"},
	{0x03B1519951F3C445ULL, "244521764.ogg"},
	{0x03B2069797310756ULL, "vxdr_primevaltomb_8v4_0_0.vlm"},
	{0x03B20E77C705C3C9ULL, "and100_anders6095878.cl"},
	{0x03B25B2096C6F64BULL, "one300_nuncio6084353.cl"},
	{0x03B27DB47EE2BAF6ULL, "gen_im_trs_bst_drag_gland_03.uti"},
	{0x03B2993E8F84B60AULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath1.gad"},
	{0x03B2A6F7518C3B9AULL, "462758899.ogg"},
	{0x03B2CCC762B3E08FULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_lar_p.impa.gad"},
	{0x03B303A8E48419BCULL, "6125208_m.fxe"},
	{0x03B31B1A2A95FC5EULL, "vxlt_0_lowtown_a2_ert_0_0.vlm"},
	{0x03B32CBB0FFB8B53ULL, "vxw_bonepitmining_a2_26g_0_0.vlm"},
	{0x03B33FE60727A7A6ULL, "6015087_f.fxe"},
	{0x03B34391FF5C3638ULL, "8763680.ogg"},
	{0x03B38FB26BE74A33ULL, "6016670_f.fxe"},
	{0x03B43E2E688BAED6ULL, "art\\characters\\creatures\\mubari\\animation\\eat_twitch.ani"},
	{0x03B458F6961DC645ULL, "gen_im_trs_gen_one_axe_r1.uti"},
	{0x03B4CACE24F7A385ULL, "37098996.ogg"},
	{0x03B57A34D1074759ULL, "6183797_m.fxe"},
	{0x03B58957F5A136B8ULL, "hlac_ht_chantry_2534_42y.dds"},
	{0x03B664875AD73C2DULL, "6117491_m.fxe"},
	{0x03B67B74EEA6DCC3ULL, "hlac_lt_0_lowtown_a2_n_c_-3_3x.dds"},
	{0x03B6D09B3A3999B5ULL, "follower_banter6169862.cl"},
	{0x03B73684F7876254ULL, "562759684.ogg"},
	{0x03B8867F6E9C6B85ULL, "6000802_f.fxe"},
	{0x03B8E1C762B8703CULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_lar_p.impa.evt"},
	{0x03B92124A63F3822ULL, "w_woundedcoast_73_0.rml"},
	{0x03B92F3DBC265776ULL, "758612263.ogg"},
	{0x03B9350F36F7E138ULL, "6048762_f.fxe"},
	{0x03B9CD050A6A55F8ULL, "6155838_m.fxe"},
	{0x03BA7355D93502F9ULL, "266599275.ogg"},
	{0x03BA75F0C9AFC220ULL, "gen000fl_merrill_house.stg"},
	{0x03BAD111CF2CE3BDULL, "6103353_m.fxe"},
	{0x03BADF442C304E39ULL, "lt_0_lowtown_a2_52_0.anb"},
	{0x03BAF9C75EB73015ULL, "gen00fl_fenris6110812.cl"},
	{0x03BC60F9A8C9CB17ULL, "6137033_m.fxe"},
	{0x03BCAD70D15029D2ULL, "vxw_bonepitmining_a2_26b_0_0.vlm"},
	{0x03BD3588813C6A5AULL, "vxdr_0_deeproads_fhr_0_0.vlm"},
	{0x03BD803D0B6D37F7ULL, "506510954.ogg"},
	{0x03BD84CB4C99E4ECULL, "vxht_0_hightown_c68_0_0.vlm"},
	{0x03BE540C32FBC605ULL, "vxw_bonepitmining_a2_6q_0_0.vlm"},
	{0x03BF105F040828B9ULL, "6143970_f.fxe"},
	{0x03C0040FA63F2B40ULL, "vxlt_gallowscourtyard_n_65f_0_0.vlm"},
	{0x03C0A1DBF7CBA9B0ULL, "vxlt_hangedman_n_2bl_0_0.vlm"},
	{0x03C0CFAD73BF6241ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.cess.ani"},
	{0x03C1131D7116E10AULL, "dae150_harshal6148854.cl"},
	{0x03C1462E689646C7ULL, "art\\characters\\creatures\\mubari\\animation\\eat_twitch.evt"},
	{0x03C16C647E06EFA3ULL, "hlac_ht_keep_3732_18x.dds"},
	{0x03C18CF71C8C14DEULL, "vxlt_0_lowtown_a2_euu_0_0.vlm"},
	{0x03C26CE5B010DCDFULL, "one200_gold_followers6120151.cl"},
	{0x03C2895040F8DDA1ULL, "6117107_m.fxe"},
	{0x03C2A8E10D6216C3ULL, "vxlt_0_lowtown_a3_n_c_f4x_0_0.vlm"},
	{0x03C2BBF14B34F0E1ULL, "6033322_m.fxe"},
	{0x03C2D47C67F4C298ULL, "vxca_caverns1_igm_0_0.vlm"},
	{0x03C332EC541E3496ULL, "vxht_keep_n_677_0_0.vlm"},
	{0x03C3DEE53715B513ULL, "6068067_m.fxe"},
	{0x03C3E625799ECD4CULL, "hlac_w_woundedcoast_262147_0y.dds"},
	{0x03C4400A20079B8EULL, "6061335_m.fxe"},
	{0x03C4A3FB34009649ULL, "vxlt_blackemporium_zl_0_0.vlm"},
	{0x03C4D5F1AFD0D93DULL, "489489928.ogg"},
	{0x03C587316176BD01ULL, "392717886.ogg"},
	{0x03C5EC76024AF164ULL, "gen00fl_anders6096618.cl"},
	{0x03C647B634BEE469ULL, "124506230.ogg"},
	{0x03C6743E8F9520F0ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath1.ani"},
	{0x03C69D149F8C68FCULL, "182625619.ogg"},
	{0x03C6E9C762C4BB55ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_lar_p.impa.ani"},
	{0x03C7DAAD73C593E3ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.cess.gad"},
	{0x03C84CA3E94741D8ULL, "608203063.ogg"},
	{0x03C85981382CE917ULL, "vxw_bonepitdestroyed_a3_1r2_0_0.vlm"},
	{0x03C8672E689C9D10ULL, "art\\characters\\creatures\\mubari\\animation\\eat_twitch.gad"},
	{0x03C8AB18ED109F3EULL, "and100_anders6095985.cl"},
	{0x03C8CAB3A89719DFULL, "11968675.ogg"},
	{0x03C93F3E2CAEE05BULL, "330019779.ogg"},
	{0x03C97FE11000FF17ULL, "gen00fl_aveline6122069.cl"},
	{0x03C9F920B79257B5ULL, "bl_tri_f.mao"},
	{0x03C9FE151B66226DULL, "mag291_torpor6025981.cl"},
	{0x03CA0B24A64D8677ULL, "w_woundedcoast_73_0.msh"},
	{0x03CA1D24A64DA4DAULL, "w_woundedcoast_73_0.mao"},
	{0x03CA2124A64DABE9ULL, "w_woundedcoast_73_0.mmh"},
	{0x03CA546E80209001ULL, "hlac_w_blightlands_65537_0x.dds"},
	{0x03CA71638B8593D9ULL, "23693978.ogg"},
	{0x03CB3D2850E2A575ULL, "dae000_merchant_hm_mob6174221.cl"},
	{0x03CB99467DBB4F57ULL, "6082812_m.fxe"},
	{0x03CD3BEDDF9D8F1AULL, "fx_c_grad_dis.mao"},
	{0x03CDDF98CF6F13CBULL, "vxlt_0_lowtown_docks_bc8_0_0.vlm"},
	{0x03CE1B30B5B3E8FCULL, "6116800_m.fxe"},
	{0x03CE57BF4DAD4F34ULL, "hlac_ht_0_hightown_a2_65532_0z.dds"},
	{0x03CEB7AD73CB76C8ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.cess.evt"},
	{0x03CEE81D70775331ULL, "907264014.ogg"},
	{0x03CF3AD88F4EB869ULL, "hlac_ca_cavern1_haunted_24426_2y.dds"},
	{0x03CF49B9AB60C730ULL, "cod_mgc_mana.utp"},
	{0x03CFFED40E822D5EULL, "vxlt_0_lowtown_a2_exv_0_0.vlm"},
	{0x03D01074F3E7BC53ULL, "vxw_bonepitmining_a2_28i_0_0.vlm"},
	{0x03D04EA432B9876CULL, "340521108.ogg"},
	{0x03D0587AA7603BF9ULL, "mag100_thrask6032048.cl"},
	{0x03D07B40E2B40643ULL, "w_blightlands_30_0.mao"},
	{0x03D07F40E2B40C90ULL, "w_blightlands_30_0.mmh"},
	{0x03D08D40E2B424AAULL, "w_blightlands_30_0.msh"},
	{0x03D128AE9BBCE524ULL, "lgt110_orz_harrowmont6108996.cl"},
	{0x03D1465959EADB18ULL, "vxca_caverns1_ijj_0_0.vlm"},
	{0x03D18C4CAB0CD77BULL, "1941608.ogg"},
	{0x03D199A1781A92BEULL, "vxht_mansion2b_2r2_0_1.vlm"},
	{0x03D2821AF2E72BF5ULL, "vxlt_gallowstemplar_f_5pa_0_0.vlm"},
	{0x03D29F4F3A0A161AULL, "lt_gallowscourtyard_30_0.rml"},
	{0x03D2E16A9A9C53E4ULL, "vxw_bonepit_ky_0_0.vlm"},
	{0x03D3232CCD34EA49ULL, "6029031_m.fxe"},
	{0x03D34B7577875244ULL, "6166014_m.fxe"},
	{0x03D3D6D24302B171ULL, "gen00fl_bethany6098055.cl"},
	{0x03D48BCEC2C71D3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanl.gad"},
	{0x03D49238F05A00C2ULL, "vxht_playershouse_n_1wa_0_0.vlm"},
	{0x03D4E4F67E8DCBCCULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_6x.dds"},
	{0x03D527A85E72FAA6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_cap.phy"},
	{0x03D5478B556BBFD4ULL, "531333280.ogg"},
	{0x03D5B7F6DD701617ULL, "578295991.ogg"},
	{0x03D5CE0DB8704B24ULL, "938927686.ogg"},
	{0x03D5D0293DB82CE6ULL, "and111_lirene6090210.cl"},
	{0x03D669778517A972ULL, "vxlt_0_lowtown_docks_a3_n_brz_0_0.vlm"},
	{0x03D758212B9AE604ULL, "vxlt_gallowscourtyard_n_6yv_0_0.vlm"},
	{0x03D75A8E57D1744FULL, "mag211_marethari6025148.cl"},
	{0x03D7C866EFE53957ULL, "6040577_m.fxe"},
	{0x03D8197304033E36ULL, "one300_zevran6089708.cl"},
	{0x03D88E9D89BD8C18ULL, "art\\levels\\races\\natural\\highlands\\foliage\\_textures\\plc_flowerwild_01_0_n.dds"},
	{0x03D8D7727976C55BULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he8a.mmh"},
	{0x03D8F5727976F845ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he8a.msh"},
	{0x03D92C4D5E183B41ULL, "dae000_flemeth6000823.cl"},
	{0x03D957BC870F53A5ULL, "445419301.ogg"},
	{0x03D967AE581FAE50ULL, "6002976_m.fxe"},
	{0x03D987D0CD88FEEFULL, "vxca_caverns1_mines_j1w_0_1.vlm"},
	{0x03D9912AB53C5AD2ULL, "vxw_bonepitmining_a2_28l_0_0.vlm"},
	{0x03D9CA925F17EEBFULL, "vxht_0_hightown_c34_0_0.vlm"},
	{0x03D9CD66A8B31E9AULL, "hlac_lt_0_lowtown_a2_n_c_65534_1x.dds"},
	{0x03D9DC52F1A94432ULL, "one100dg_kelder6056894.cl"},
	{0x03D9FEE8D7AA1CE9ULL, "109555343.ogg"},
	{0x03DA7267CC9DD01FULL, "vxlt_gallowstemplar_3nl_0_0.vlm"},
	{0x03DACC7201E358D4ULL, "6217114_m.fxe"},
	{0x03DB5D5EA5481C26ULL, "132187802.ogg"},
	{0x03DB65B1CB1BA396ULL, "6132794_m.fxe"},
	{0x03DB8ECEC2CD4151ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanl.evt"},
	{0x03DBC28B679CD356ULL, "171625570.ogg"},
	{0x03DBE93559E1E548ULL, "gen00fl_aveline6137585.cl"},
	{0x03DC1471FB91A2A1ULL, "6147779_m.fxe"},
	{0x03DC24CB4DF38FECULL, "344809040.ogg"},
	{0x03DC71EC8C3F91E6ULL, "isa300ip_warehouse_exit.utp"},
	{0x03DCC948F307D81FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_generalwood_n.dds"},
	{0x03DD24D95B7B841DULL, "vxlt_0_lowtown_a2_e16_0_0.vlm"},
	{0x03DD598C50691C57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\combatrun.gad"},
	{0x03DDDC5E80BD768AULL, "hlac_lt_hangedman_-1_21z.dds"},
	{0x03DE540E8483C632ULL, "vxht_0_hightown_apr_0_0.vlm"},
	{0x03DE9B7C9FAED58DULL, "107154232.ogg"},
	{0x03DF87EE859D3B53ULL, "54224187.ogg"},
	{0x03DFB8364BE0F398ULL, "vxca_caverns1_imk_0_0.vlm"},
	{0x03DFC75E4EDA4CD5ULL, "hlac_lt_0_lowtown_a2_n_-131071_2y.dds"},
	{0x03DFE06F731A1742ULL, "527720795.ogg"},
	{0x03E091334FC91A3CULL, "ht_0_hightown_n_34_0.anb"},
	{0x03E09C21EAC0D5E5ULL, "6201480_m.fxe"},
	{0x03E0D9C62F190CA4ULL, "gen00fl_anders6103909.cl"},
	{0x03E165A6DCF7BE76ULL, "vxw_sundermount_28a_0_0.vlm"},
	{0x03E19028C09F7ECAULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_c.anb"},
	{0x03E1B91A068ED380ULL, "125629908.ogg"},
	{0x03E1C6E57CDF1FF1ULL, "vxca_cavern1_haunted_jgb_0_0.vlm"},
	{0x03E233CEB6D32BCBULL, "6161603_m.fxe"},
	{0x03E28570F1B72AB4ULL, "one200ip_flower.ncs"},
	{0x03E3AC9512A42575ULL, "dae000_mother6121014.cl"},
	{0x03E43C1AFFFB2C9EULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_dwn_c.impa.ani"},
	{0x03E46E8C506F5EE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\combatrun.evt"},
	{0x03E516AECC56AB4FULL, "hlac_w_woundedcoast_-196604_0y.dds"},
	{0x03E5A3895B93838EULL, "zz_dae_debug6047608.cl"},
	{0x03E5A3D9D86F8946ULL, "6031751_f.fxe"},
	{0x03E5E6B4360056AEULL, "6069439_f.fxe"},
	{0x03E69C624BB3A6B6ULL, "6162773_m.fxe"},
	{0x03E6EE96A380264BULL, "ave200_donnic6110513.cl"},
	{0x03E713BEC4AE1FC9ULL, "6072563_m.fxe"},
	{0x03E7526161917727ULL, "759605591.ogg"},
	{0x03E834C408F2DF60ULL, "6027708_m.fxe"},
	{0x03E83C6F510E073FULL, "vxht_0_hightown_c67_0_0.vlm"},
	{0x03E8A6CEC2D7F420ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanl.ani"},
	{0x03E8C20B2FBC0D5BULL, "hlac_lt_0_lowtown_docks_n_-65538_8z.dds"},
	{0x03E8CE621C561680ULL, "ht_0_hightown_33_0_ll.mmh"},
	{0x03E90B55E4BBBB75ULL, "6166583_m.fxe"},
	{0x03E91DFF99B4F8D9ULL, "vxht_0_hightown_n_c5o_0_0.vlm"},
	{0x03E976F17A5211D1ULL, "stone_s_rock.mao"},
	{0x03E99D85A0A271F4ULL, "542484696.ogg"},
	{0x03E9C405E2E61B2DULL, "vxht_0_hightown_n_ai9_0_0.vlm"},
	{0x03E9CFF764842FBFULL, "682375734.ogg"},
	{0x03E9D3399DBA0FADULL, "vxdr_primevalentrance_19s_0_0.vlm"},
	{0x03EA38AB4EE513AFULL, "336041127.ogg"},
	{0x03EB7891AE747034ULL, "vxht_bartrand_n_65t_0_0.vlm"},
	{0x03EBEF8EAAC27AF9ULL, "6037599_m.fxe"},
	{0x03EBF482F0454CDFULL, "llac_lt_0_lowtown_a3_n_c_196607_1.dds"},
	{0x03EC00C1A614010FULL, "6174797_m.fxe"},
	{0x03EC7613B6C10510ULL, "998582898.ogg"},
	{0x03EC94F5A0B3B3D3ULL, "6071681_m.fxe"},
	{0x03ECCF4122D5AB32ULL, "vxca_caverns2_9fw_0_0.vlm"},
	{0x03ECE28DF26E5E5EULL, "vxlt_0_lowtown_a2_efh_0_0.vlm"},
	{0x03ED2F71ACA8DC2AULL, "6098726_m.fxe"},
	{0x03ED35AED938001DULL, "vxw_encounter3_e1_0_0.vlm"},
	{0x03EDD766ADC0BA24ULL, "68250502.ogg"},
	{0x03EE8F83CA3B4752ULL, "follower_banter6161067.cl"},
	{0x03EECF0030224F04ULL, "vxdr_0_deeproads_f5o_0_0.vlm"},
	{0x03EEEB026C12DBD5ULL, "llac_w_encounter2_1_0.dds"},
	{0x03EF10684A702E35ULL, "270682253.ogg"},
	{0x03EF9D783252AAAAULL, "6109028_m.fxe"},
	{0x03EFF99209E2DFC9ULL, "vxlt_blackemporium_uq_0_0.vlm"},
	{0x03F0123D895215A4ULL, "vxht_keep_n_67u_0_0.vlm"},
	{0x03F071D10CC8D3EDULL, "vxlt_undercitydungeon_130_0_0.vlm"},
	{0x03F137A940FA66BBULL, "hlac_lt_gallowscourtyard_-131072_4z.dds"},
	{0x03F1ED35BA275DF4ULL, "hlac_ht_0_hightown_a2_65535_1x.dds"},
	{0x03F2541B0007936FULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_dwn_c.impa.evt"},
	{0x03F2868C507BC4DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\combatrun.ani"},
	{0x03F2B0540989ABADULL, "6077010_m.fxe"},
	{0x03F30644BBEDC81CULL, "isa311_isabela6092498.cl"},
	{0x03F30EB19032A377ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_0x.dds"},
	{0x03F35467C616265CULL, "vxw_woundedcoast_1kq_0_0.vlm"},
	{0x03F38A2803BB0BD8ULL, "312160025.ogg"},
	{0x03F3BB5438292C52ULL, "follower_barks6153344.cl"},
	{0x03F449876A49ACBEULL, "263827842.ogg"},
	{0x03F44E5156E2FD7AULL, "lowtown_fem_prostitute_sta6152367.cl"},
	{0x03F4B09E2CB120D9ULL, "215284487.ogg"},
	{0x03F4CA5C2BF85D8FULL, "lt_0_lowtown_docks_31_0.anb"},
	{0x03F4DE42F720EC8FULL, "zz_isa_debug6133930.cl"},
	{0x03F4E0442C61C839ULL, "lt_0_lowtown_a2_52_0.rml"},
	{0x03F50E3F3DA9BF36ULL, "6167473_m.fxe"},
	{0x03F51CB7208B1894ULL, "vxw_encounter1_c4_0_0.vlm"},
	{0x03F5774A0690EA1DULL, "lt_0_lowtown_a3_n_c_44_0_lc.mmh"},
	{0x03F641889FE318ABULL, "matt_warrior_2.utc"},
	{0x03F6B7CD64F5CC65ULL, "vxht_0_hightown_n_7g5_0_0.vlm"},
	{0x03F77EB3175A3698ULL, "vxlt_warehouse_q8_0_0.vlm"},
	{0x03F784D4ECE37097ULL, "gen00fl_fenris6127117.cl"},
	{0x03F78FDC8BAB1159ULL, "vxht_0_hightown_n_c2n_0_0.vlm"},
	{0x03F82DBC2471FE72ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_atlas_jimmy_0s.dds"},
	{0x03F8551B000C0018ULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_dwn_c.impa.gad"},
	{0x03F8C0B9CCDCB938ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_flsa_0d.dds"},
	{0x03F8F700DC0BF690ULL, "520611234.ogg"},
	{0x03F909AAD653292AULL, "vxht_keep_55q_0_0.vlm"},
	{0x03F9529CAE64936EULL, "dae000_mother6099190.cl"},
	{0x03FA1F3C69F16AB3ULL, "49115056.ogg"},
	{0x03FA24D5DF24D875ULL, "lgt110_arl_bann_teagan6115626.cl"},
	{0x03FA3ADDD204223FULL, "vxlt_0_lowtown_docks_aln_0_0.vlm"},
	{0x03FA6470F1CB84E5ULL, "one200ip_flower.utp"},
	{0x03FB64A19CF42475ULL, "llac_w_encounter3_-131068_0.dds"},
	{0x03FC62DD152B7A5AULL, "hlac_lt_0_lowtown_a2_n_c_65535_2z.dds"},
	{0x03FC8B1E74A59A88ULL, "403448814.ogg"},
	{0x03FCBC8EA9775D00ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerb_n.dds"},
	{0x03FCC64EEC9E9836ULL, "vxht_bartrand_n_6qz_0_0.vlm"},
	{0x03FD49424F90D23AULL, "6119274_m.fxe"},
	{0x03FD844D865117F7ULL, "character_character_p178.psh.fxo.dx11"},
	{0x03FDB53F7A9259DEULL, "vxht_keep_n_6go_0_0.vlm"},
	{0x03FDD7B1C8C97575ULL, "vxlt_gallowstemplar_f_5an_0_0.vlm"},
	{0x03FE841A7B482E24ULL, "vxht_keep_n_62z_0_0.vlm"},
	{0x03FEFD58745622B5ULL, "vxca_caverns1_jfx_0_0.vlm"},
	{0x03FF7B3B8D550223ULL, "124543312.ogg"},
	{0x03FFEC23B95346DEULL, "llac_dr_0_deeproads_17039_2.dds"},
	{0x04001BD964F0E93FULL, "vxlt_undercitydungeon_1ie_0_1.vlm"},
	{0x0400A49E4B03E788ULL, "440696593.ogg"},
	{0x0400E3A3E2984BEBULL, "art\\vfx\\combat\\vfx_f_mer_emove_act_c.impa.ani"},
	{0x0401143B35884834ULL, "vxht_mansion2b_2k2_0_0.vlm"},
	{0x0401E97557F2F84DULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_com_0s.dds"},
	{0x0401EDD5CCDE8F76ULL, "w_encounter2_60_0.msh"},
	{0x0401F7D5CCDEA074ULL, "w_encounter2_60_0.mmh"},
	{0x0401FBD5CCDEA687ULL, "w_encounter2_60_0.mao"},
	{0x0402320B58FF2D46ULL, "823997369.ogg"},
	{0x04023911F137C663ULL, "gen00fl_sebastian6179799.cl"},
	{0x040256CE43D4427BULL, "199721422.ogg"},
	{0x040266DF879A1FC1ULL, "ser221ip_door_west.utp"},
	{0x04028C3A7B921441ULL, "nat380_temmerin6090888.cl"},
	{0x0402B24E38248D38ULL, "231802145.ogg"},
	{0x0402EE8582ABE5C2ULL, "6096809_m.fxe"},
	{0x04038E9412813114ULL, "vxw_encounter1_h5_0_0.vlm"},
	{0x04048AD20BE267E4ULL, "33835233.ogg"},
	{0x04048D94542E7D20ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\textures\\windowglowcombined_s.dds"},
	{0x0404C869ECFC5677ULL, "109017843.ogg"},
	{0x0404E674D383EE65ULL, "vxdr_0_deeproads_fiv_0_0.vlm"},
	{0x04050B0A24BB94A4ULL, "llac_dr_dungeon_3207_8.dds"},
	{0x0405349D513510F7ULL, "6104552_m.fxe"},
	{0x040596C6434FA9E9ULL, "55898743.ogg"},
	{0x0405A430A5837EE8ULL, "1009756150.ogg"},
	{0x0405F09009504F18ULL, "vxlt_warehouse_z9_0_0.vlm"},
	{0x0405FE26A956FEE9ULL, "lt_backalley_g_0_0.phy"},
	{0x0406B99E0C924CDBULL, "6110715_m.fxe"},
	{0x0406DCF586DBB1ECULL, "hand2.cur"},
	{0x04072D6652FFA503ULL, "209274311.ogg"},
	{0x040770ABF091B164ULL, "vxdr_primevaltomb_5af_0_0.vlm"},
	{0x040885AFD7BF7647ULL, "59911015.ogg"},
	{0x0408ACBAC3FA3ABFULL, "vxlt_0_lowtown_docks_aqi_0_0.vlm"},
	{0x0408D8C76E3CCC96ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_dbl_c.anb"},
	{0x04090AF42A645B6AULL, "gen_im_trs_dwv_med_glv_r0.uti"},
	{0x04094B9092876887ULL, "dae000_pc_home_ambients6208632.cl"},
	{0x0409C647D65A8F5EULL, "vxlt_0_lowtown_a2_elj_0_0.vlm"},
	{0x040A090C7A411AD7ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.cess.gad"},
	{0x040A7F40E2E58553ULL, "w_blightlands_30_0.rml"},
	{0x040A919A3CCC8461ULL, "hlac_lt_0_lowtown_docks_n_131071_1x.dds"},
	{0x040A9328C0C288BDULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_c.mmh"},
	{0x040AA9D96C98FCA5ULL, "art\\characters\\weapons\\maces\\w_mce_mc107.mao"},
	{0x040C26A3CC95D1E4ULL, "llac_lt_0_lowtown_docks_-4_1.dds"},
	{0x040CDFBD5589FBC0ULL, "zz_mer_debug6017858.cl"},
	{0x040CE88D24F615C2ULL, "llac_lt_0_lowtown_n_-65534_0.dds"},
	{0x040D0726A95D2CE5ULL, "lt_backalley_g_0_0.rml"},
	{0x040DCBA3E2A2AD62ULL, "art\\vfx\\combat\\vfx_f_mer_emove_act_c.impa.evt"},
	{0x040DF2131549EC79ULL, "llac_lt_gallowstemplar_f_2388_3.dds"},
	{0x040E0434C6C5664DULL, "vxw_bonepitdestroyed_a3_278_0_0.vlm"},
	{0x040E52FE27570C75ULL, "936784495.ogg"},
	{0x040E63ABBE44B18EULL, "73206073.ogg"},
	{0x040E66ED2FFA477BULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanr.evt"},
	{0x040E70BF6F343017ULL, "vxdr_primevaltomb_8o2_0_0.vlm"},
	{0x040E77DB1A8FAAD5ULL, "vxlt_0_lowtown_docks_a3_n_ctn_0_0.vlm"},
	{0x040E8C6419E6ECF1ULL, "hlac_lt_0_lowtown_-3_6z.dds"},
	{0x040EB6278F427C39ULL, "494588162.ogg"},
	{0x040FF0E861229613ULL, "gen00fl_sebastian6179889.cl"},
	{0x041030EB4CF00B72ULL, "vxlt_0_lowtown_docks_a3_n_bff_0_0.vlm"},
	{0x04103821AA02575CULL, "vxw_woundedcoast_1qw_0_0.vlm"},
	{0x041040E79C135DDDULL, "combat_lmm_r1_a3.gda"},
	{0x0410A229103B3C5BULL, "vxht_brothel_n_75b_0_0.vlm"},
	{0x0410B216049403ABULL, "gen00fl_isabela6096796.cl"},
	{0x0410B49C30B48E40ULL, "273684510.ogg"},
	{0x0410C6690D0D973FULL, "w_encounter1_32_0.phy"},
	{0x0410E1440FCBC670ULL, "vxlt_0_lowtown_a2_gvd_0_0.vlm"},
	{0x04111E0C7A475D64ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.cess.evt"},
	{0x04112FCF1CEA9050ULL, "vxht_keep_7aa_0_0.vlm"},
	{0x04115A4F66A68B39ULL, "852184223.ogg"},
	{0x04115D0762E74002ULL, "84585901.ogg"},
	{0x04119A4BCC2D6DCAULL, "mer210_coterie_contact6036771.cl"},
	{0x0411EA99B3E84BD9ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_electr.gad"},
	{0x0411F69F7D66856DULL, "cod_hst_orlais.utp"},
	{0x041204411E4A5A4AULL, "377156371.ogg"},
	{0x041264226BFC9975ULL, "681291877.ogg"},
	{0x041273A2E7DA5DD5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_medj_0s.dds"},
	{0x0412B76ED1D83426ULL, "6174536_m.fxe"},
	{0x0412DC64D869F209ULL, "6119509_m.fxe"},
	{0x0413974C1F814CABULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv3.ani"},
	{0x0413D8E7F62AE6A7ULL, "6167560_m.fxe"},
	{0x04144860C0BDDBAEULL, "55759269.ogg"},
	{0x0414CAA3E2A8CA75ULL, "art\\vfx\\combat\\vfx_f_mer_emove_act_c.impa.gad"},
	{0x0414CFC10ADDA68DULL, "895667437.ogg"},
	{0x04153B1EDC2F4E59ULL, "48373123.ogg"},
	{0x041553FE6C8CCD87ULL, "plt_gen00pt_generic_actions.plo"},
	{0x04155FED300059BCULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanr.gad"},
	{0x04162E1919445F59ULL, "vxht_brothel_n_59r_0_0.vlm"},
	{0x0416ABD88D2082FBULL, "100849290.ogg"},
	{0x0416AFE2C5F28861ULL, "6048445_m.fxe"},
	{0x0416B6A3C228009CULL, "6095845_f.fxe"},
	{0x0416DBC76E490F01ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_dbl_c.mmh"},
	{0x0418062935E5E9ABULL, "6071898_m.fxe"},
	{0x0418119A031BC7D1ULL, "252762106.ogg"},
	{0x041816E77BFD2697ULL, "gen00fl_anders6125690.cl"},
	{0x041883C52065BC45ULL, "6033717_m.fxe"},
	{0x0418E03CF81411C2ULL, "plt_qun100pt_mage_escort.plo"},
	{0x0418F457F9FAFC8BULL, "705361302.ogg"},
	{0x0418FB99B3EE8727ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_electr.ani"},
	{0x0419373F5CA6B5A5ULL, "one220_dera6099729.cl"},
	{0x0419761C3F6CBB8AULL, "vxlt_0_lowtown_a2_n_c_h24_0_0.vlm"},
	{0x0419EF1ABD362B22ULL, "zz_dae_debug6029921.cl"},
	{0x041A4BF5473469E9ULL, "mag110_arianni6016826.cl"},
	{0x041A5BB84A800696ULL, "hlac_ht_mansion_n_3373_3y.dds"},
	{0x041B81B38FC3E398ULL, "static_static_p27.psh.fxo.dx11"},
	{0x041B866A1FBB9218ULL, "37455165.ogg"},
	{0x041B9F845C34E0CEULL, "6078474_m.fxe"},
	{0x041BDD3D9DB5B5A0ULL, "tools_charpost.phy"},
	{0x041BE6B85DEED52FULL, "art\\characters\\creatures\\dragon\\textures\\cn_eye_drga_0d.dds"},
	{0x041C5EED300676BAULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanr.ani"},
	{0x041C6487FE976B61ULL, "vxw_encounter2_7t_0_0.vlm"},
	{0x041C900B410CFAAEULL, "vxht_0_hightown_cb6_0_0.vlm"},
	{0x041CF707B54C923CULL, "hlac_ca_caverns2prc_5802_52y.dds"},
	{0x041D23ED5912C581ULL, "6084235_f.fxe"},
	{0x041D48ED94FA0248ULL, "345223630.ogg"},
	{0x041D4F6878BE5663ULL, "lava0.vsh.fxo"},
	{0x041DB65A5FAF4D1DULL, "atl_demosplash_dxt5_00.dds"},
	{0x041DF7F519747934ULL, "vxht_mansion2b_2a0_0_0.vlm"},
	{0x041E0E633188D842ULL, "6032000_f.fxe"},
	{0x041E113BDFB88E4DULL, "follower_banter6161291.cl"},
	{0x041E6151EEA041D7ULL, "6061795_f.fxe"},
	{0x041EAC21016F5563ULL, "6118377_m.fxe"},
	{0x041F360C7A53C35DULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.cess.ani"},
	{0x0420275A05AA6E9AULL, "303610733.ogg"},
	{0x04202821F3B4BB1FULL, "6144941_m.fxe"},
	{0x04207F2DA02BACBCULL, "484631275.ogg"},
	{0x042080871E1041A2ULL, "6129596_m.fxe"},
	{0x0420A75C2C1D5D33ULL, "lt_0_lowtown_docks_31_0.rml"},
	{0x0420C18D7FDAE21CULL, "vxw_woundedcoast_hy_0_0.vlm"},
	{0x0420C89413B0AAADULL, "ave121_jeven6105631.cl"},
	{0x0420F87279B4B290ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he8a.phy"},
	{0x04217F4C1F8D6122ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv3.evt"},
	{0x042190335000AFCCULL, "ht_0_hightown_n_34_0.rml"},
	{0x0421B3C16841600CULL, "w_woundedcoast_254_0.msh"},
	{0x0421B9C168416AFEULL, "w_woundedcoast_254_0.mmh"},
	{0x0421C5C168417E95ULL, "w_woundedcoast_254_0.mao"},
	{0x0421C7690D1C0C44ULL, "w_encounter1_32_0.mmh"},
	{0x0421CB690D1C1297ULL, "w_encounter1_32_0.mao"},
	{0x0421D917DB0A4D94ULL, "6198675_m.fxe"},
	{0x0421DD690D1C31E6ULL, "w_encounter1_32_0.msh"},
	{0x04222C8E94578C5EULL, "6041904_m.fxe"},
	{0x0422314DB1CE6BECULL, "953888856.ogg"},
	{0x0422788697E52552ULL, "face0_s2.psh.fxo"},
	{0x042311B2E122DD41ULL, "162459204.ogg"},
	{0x04235098F4AAD0E3ULL, "w_bonepit_20_0.mao"},
	{0x04235498F4AAD730ULL, "w_bonepit_20_0.mmh"},
	{0x04236003F1EA8456ULL, "vxlt_0_lowtown_df1_0_0.vlm"},
	{0x04236298F4AAEF4AULL, "w_bonepit_20_0.msh"},
	{0x0423D87FEBE51429ULL, "hlac_ht_0_hightown_131072_0z.dds"},
	{0x0423EE04667DA8D9ULL, "vxht_0_hightown_a2_c7z_0_0.vlm"},
	{0x0423F688C6C2DA06ULL, "dae000_worthy6166616.cl"},
	{0x0424036916E5FC7CULL, "1028937705.ogg"},
	{0x042459FC80D12567ULL, "6054640_m.fxe"},
	{0x0424806A61302628ULL, "vxlt_gallowstemplar_f_5r2_0_0.vlm"},
	{0x04249817EAEDCDB5ULL, "mrl000_marethari6129385.cl"},
	{0x0424AC5B6D402B0EULL, "493262192.ogg"},
	{0x0424FBE9F94DE98FULL, "6034325_m.fxe"},
	{0x04250F32B9D17B6BULL, "6144431_m.fxe"},
	{0x04254A792EFA57F1ULL, "6129447_m.fxe"},
	{0x0425880E2E1A75DBULL, "vxca_caverns1_jgu_0_0.vlm"},
	{0x0425ED3DBFEC1748ULL, "vxw_encounter2_7q_0_0.vlm"},
	{0x04262AE99E006336ULL, "and231im_alrik_papers.uti"},
	{0x0426AF3F26E94FA2ULL, "689504468.ogg"},
	{0x0426C82A4C1D2E67ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden.mao"},
	{0x0426D2B88AF85DA9ULL, "230360171.ogg"},
	{0x0426E5388AAE6824ULL, "vrc321ip_spooky5_obj_6.utp"},
	{0x04277ADB7E411A53ULL, "62975446.ogg"},
	{0x042803739B2D2BB8ULL, "fen210st_tevniter_ambusher.stg"},
	{0x042826EB5D805820ULL, "48062705.ogg"},
	{0x042850F48FF06CE4ULL, "729684778.ogg"},
	{0x04286992177D06B2ULL, "aip_circlemage_base.gda"},
	{0x04286FEF8AC68AE8ULL, "6114416_f.fxe"},
	{0x04287E4C1F937E35ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv3.gad"},
	{0x042896EC1E7B0BDCULL, "dae000_adriano6163672.cl"},
	{0x042926A48DF7CA4AULL, "vxw_bonepitdestroyed_a3_73_0_0.vlm"},
	{0x04292BF2DDB7493BULL, "and111st_donation_box.stg"},
	{0x04294D753F503FD1ULL, "mag291cr_crowd_elf_1.utc"},
	{0x042A04A16A1AB94DULL, "hlac_ht_keep_3732_11y.dds"},
	{0x042A3B9B635A45F7ULL, "6111468_m.fxe"},
	{0x042A73338BE1F9CDULL, "hlac_lt_0_lowtown_docks_-4_0x.dds"},
	{0x042A79912CA167BBULL, "vxlt_0_lowtown_a2_n_c_hd7_0_0.vlm"},
	{0x042ACF100F806ACDULL, "hlac_dr_primevaltomb_663_2x.dds"},
	{0x042AD664F08D83E1ULL, "vxw_encounter2_2k_0_0.vlm"},
	{0x042B1189F81A9816ULL, "qun120_arishok6052909.cl"},
	{0x042B5B94FE7BDBD5ULL, "vxlt_0_lowtown_docks_a3_n_czl_0_0.vlm"},
	{0x042B5D5B13D31A69ULL, "204847640.ogg"},
	{0x042B5F72282F2076ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocklarge.phy"},
	{0x042C56924621B2DFULL, "78104986.ogg"},
	{0x042C600D23925282ULL, "art\\characters\\heraldry\\hld_city_guard.dds"},
	{0x042D1A0214F9714EULL, "6105558_m.fxe"},
	{0x042D1DFDF3656DE3ULL, "6117277_m.fxe"},
	{0x042D693D5C40FDB1ULL, "6019105_f.fxe"},
	{0x042DBB7698504D36ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pickaxe.msh"},
	{0x042DC57698505E34ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pickaxe.mmh"},
	{0x042DC97698506547ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pickaxe.mao"},
	{0x042E8860371BD474ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_crn_rdy.gad"},
	{0x042EB496F2E9AC78ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_105a_sa.dds"},
	{0x042EC30EF41E611CULL, "follower_banter6147410.cl"},
	{0x042F1E33D1A79782ULL, "dre000_dougal6131155.cl"},
	{0x042F3D90C9AEAD4FULL, "hlac_lt_blackemporium_1055_6y.dds"},
	{0x04307798F4B59D4FULL, "w_bonepit_20_0.anb"},
	{0x0430BED5F9D58846ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_lp.ani"},
	{0x0430DE287B252EFAULL, "6102660_f.fxe"},
	{0x0430E4F0BF0082F6ULL, "fen110_fenris6097047.cl"},
	{0x043150991C029792ULL, "llac_ca_cavern1_haunted_24424_4.dds"},
	{0x0431698552A81735ULL, "llac_lt_darktown_-1_43.dds"},
	{0x04317269014947FDULL, "6174614_m.fxe"},
	{0x0431E726A97C120EULL, "lt_backalley_g_0_0.mmh"},
	{0x0431F326A97C26A5ULL, "lt_backalley_g_0_0.mao"},
	{0x0431FD569CB495ADULL, "vxht_0_hightown_n_av4_0_0.vlm"},
	{0x04320126A97C3EFCULL, "lt_backalley_g_0_0.msh"},
	{0x0432230200C155F8ULL, "6070212_m.fxe"},
	{0x04325FE15873C159ULL, "vxht_0_hightown_a2_c2y_0_0.vlm"},
	{0x0432959DA9F90F0FULL, "205098535.ogg"},
	{0x0432A5D23A8B828DULL, "gen00fl_merrill6128918.cl"},
	{0x0432D0E24D564B9EULL, "30185661.ogg"},
	{0x0433EFD768F70419ULL, "gen00fl_isabela6102455.cl"},
	{0x04345F1AB1E22FC8ULL, "vxw_encounter2_2n_0_0.vlm"},
	{0x04349A286A7470D7ULL, "215434413.ogg"},
	{0x0434A9A167B92CACULL, "vxht_chantry_n_147_0_0.vlm"},
	{0x0434DA3E31C85178ULL, "6027511_m.fxe"},
	{0x0434F2A4FD934FECULL, "fx_dd_streak.mao"},
	{0x04350F0923109A9BULL, "llac_ht_mansion2_818_1.dds"},
	{0x04351CF2C10EFC13ULL, "hlac_dr_0_deeproads_17038_2x.dds"},
	{0x0435487228374D17ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocklarge.msh"},
	{0x04355A7228376C7AULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocklarge.mao"},
	{0x04355E7228377289ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocklarge.mmh"},
	{0x0435727981F5A31FULL, "91770588.ogg"},
	{0x0435976037220C52ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_crn_rdy.ani"},
	{0x0435A54FC86F7B31ULL, "116119439.ogg"},
	{0x0435E0944C0BE6B8ULL, "lt_0_lowtown_36_0_lc.mmh"},
	{0x0435E0D8F780F6DBULL, "w_bonepitdestroyed_a3_126_0.msh"},
	{0x0435F2D8F7811526ULL, "w_bonepitdestroyed_a3_126_0.mao"},
	{0x0435FED8F78129C5ULL, "w_bonepitdestroyed_a3_126_0.mmh"},
	{0x04360E964630BDA9ULL, "art\\vfx\\combat\\vfx_gib_ogre_bld_p.mmh"},
	{0x043625873327FA6AULL, "mrl251ip_loot_d.utp"},
	{0x043742D4CE6861DDULL, "6137245_f.fxe"},
	{0x0437D57BF14DC9E2ULL, "zz_and_debug6176464.cl"},
	{0x043833A182EDBA6DULL, "follower_banter6169879.cl"},
	{0x04392FBE3F80E053ULL, "gen_im_trs_chn_rng_r0.uti"},
	{0x043A1CEC39F19C54ULL, "6116552_m.fxe"},
	{0x043A25D33261D1D1ULL, "t3_wdg_dn1.tnt"},
	{0x043AF3A3C54E7B8AULL, "depth_speedtree.vsh.fxo"},
	{0x043AFEEE0318C69EULL, "art\\characters\\creatures\\shriek\\animation\\blendtree\\shriek_movement.anb"},
	{0x043B7B6BB2C35366ULL, "hlac_lt_darktown_-1_40z.dds"},
	{0x043BAD950FD3DD62ULL, "vxca_caverns2prc_afz_0_0.vlm"},
	{0x043C07ED14F9D171ULL, "vxlt_backalley_g_1eg_0_0.vlm"},
	{0x043C16BF6289C81FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared2a.phy"},
	{0x043C3066C5C70D27ULL, "521470572.ogg"},
	{0x043C6538688D59ADULL, "6156643_m.fxe"},
	{0x043C760155BC002EULL, "29562530.ogg"},
	{0x043CF9CE483913B9ULL, "vxw_bonepitmining_a2_21p_0_0.vlm"},
	{0x043DB6F0AD6631C5ULL, "525392193.ogg"},
	{0x043DE161CBA56EFDULL, "6033677_f.fxe"},
	{0x043E9A49EFAF47B5ULL, "vxdr_slavers_148_0_0.vlm"},
	{0x043EA6D5F9E19CB7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_lp.evt"},
	{0x043EE8296FBB1CF0ULL, "112412601.ogg"},
	{0x043F9CBB78FAE18EULL, "vxlt_undercityserialkiller_1id_0_0.vlm"},
	{0x043FC648E3FDA8C2ULL, "6126561_f.fxe"},
	{0x04400626A988840DULL, "lt_backalley_g_0_0.anb"},
	{0x044067BAE4ED6672ULL, "vxht_bartrand_n_5wg_0_0.vlm"},
	{0x044094DEF2B76E78ULL, "6075112_m.fxe"},
	{0x0440D1BE4A69D9D9ULL, "vxht_0_hightown_a2_c9x_0_0.vlm"},
	{0x0440ED03F5EAE5C7ULL, "zz_dae_debug6142957.cl"},
	{0x0440FA3D9DD50335ULL, "tools_charpost.msh"},
	{0x0440FC3D9DD506ABULL, "tools_charpost.mmh"},
	{0x0441846A069568CEULL, "gen00fl_aveline6137486.cl"},
	{0x0441A1C116C9E383ULL, "6121895_f.fxe"},
	{0x0441ECC6D084C4BFULL, "art\\characters\\playercharacter\\elfmale\\cth_comc.phy"},
	{0x0441FFAD4E4BA08EULL, "367230948.ogg"},
	{0x0442A92F6037ED4FULL, "491895825.ogg"},
	{0x0442D24FD9337958ULL, "6208672_f.fxe"},
	{0x04430683FAC85B73ULL, "6016174_m.fxe"},
	{0x04432453CE1C7CC8ULL, "vxw_blightlands_ax_0_0.vlm"},
	{0x04432E994EE7B453ULL, "co_starline.mao"},
	{0x044368AF544943E2ULL, "vxlt_0_lowtown_a2_n_c_hkw_0_0.vlm"},
	{0x044380A298AD92F2ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.imp.gad"},
	{0x0443C831BE42C986ULL, "129993528.ogg"},
	{0x044417F791DEF48DULL, "6127276_m.fxe"},
	{0x044433FB34766377ULL, "815842234.ogg"},
	{0x044462D4D1110CD0ULL, "623890158.ogg"},
	{0x04447FE41E27E7D9ULL, "6081341_m.fxe"},
	{0x0444D540E99B9F18ULL, "vxca_caverns1_i2b_0_0.vlm"},
	{0x0444D9DE4D206A6DULL, "vxlt_ruinedhouse_n_5b_0_0.vlm"},
	{0x0444DB24A3087DF3ULL, "6105638_m.fxe"},
	{0x0444DE7A95B6882DULL, "776518696.ogg"},
	{0x044531B8B1ACD3AAULL, "dae021_sister_alema_mob.cnv"},
	{0x04456A988C30B26DULL, "vxca_cavern1_haunted_e8p_0_0.vlm"},
	{0x044597D5815DFC55ULL, "788055769.ogg"},
	{0x0445A7D5F9E7BD40ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_lp.gad"},
	{0x0445C89323AB20B3ULL, "6103683_m.fxe"},
	{0x0445CCA3A043EB9AULL, "6102335_m.fxe"},
	{0x04467B4E1A4A41D6ULL, "w_encounter3_211_0.mao"},
	{0x044683C079A4DECDULL, "219821552.ogg"},
	{0x0446874E1A4A5675ULL, "w_encounter3_211_0.mmh"},
	{0x0446894E1A4A59EBULL, "w_encounter3_211_0.msh"},
	{0x0446C32AA24C6969ULL, "hlac_w_encounter3_-262144_0z.dds"},
	{0x044785F6C50EF95EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rainbow.ani"},
	{0x04479CE19544C4EBULL, "lla_ht_0_hightown_a3_-131073_0.dds"},
	{0x0447AF7EE9C5E4C3ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_2z.dds"},
	{0x0447BA1ED479B4E1ULL, "vxw_encounter2_8i_0_0.vlm"},
	{0x0447BA8FAC3B4FD3ULL, "gen00fl_fenris6103326.cl"},
	{0x0447E2B2972CEBAFULL, "dae000_gamlen6149406.cl"},
	{0x0448138EB6E6D65BULL, "vxw_encounter2_camp_ov_0_0.vlm"},
	{0x04484423C4320640ULL, "hlac_ht_chantry_n_2534_4x.dds"},
	{0x0448574D98D8B659ULL, "fen110_anso6095026.cl"},
	{0x04485B69DD412C42ULL, "w_blightlands_31_0.rml"},
	{0x0448A1056BBE04C7ULL, "ser191_emeric6150905.cl"},
	{0x0448FC342A5B760CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1_tool.msh"},
	{0x044902342A5B80FEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1_tool.mmh"},
	{0x04490E342A5B9495ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1_tool.mao"},
	{0x044913B8B1B089D2ULL, "dae021_sister_alema_mob.bnk"},
	{0x0449905C395601F3ULL, "vxca_caverns1_mines_ghl_0_0.vlm"},
	{0x0449BF0B075F129BULL, "364898384.ogg"},
	{0x0449F320B9528E12ULL, "gen_im_trs_rdr_two_2hs_r1.uti"},
	{0x044A4A7206659F40ULL, "6144322_m.fxe"},
	{0x044A7D8C5628ECE5ULL, "875637687.ogg"},
	{0x044AE7E0B0098CC4ULL, "331920600.ogg"},
	{0x044B5F5C4358DC5DULL, "266945119.ogg"},
	{0x044B65773E62DA3EULL, "994343206.ogg"},
	{0x044BAAC54FE2DDD2ULL, "combat_2h_mv_default.gda"},
	{0x044BEB6373C1F1EAULL, "glows.mao"},
	{0x044C2CC34D39468AULL, "dae000_pc_home_ambients6208693.cl"},
	{0x044CFFC922122830ULL, "hlac_lt_0_lowtown_docks_a3_-3_3x.dds"},
	{0x044D17BF62983D24ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared2a.mmh"},
	{0x044D1BBF62984477ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared2a.mao"},
	{0x044D2DBF629862C6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared2a.msh"},
	{0x044D79FB4A8477B5ULL, "vxht_0_hightown_n_an4_0_0.vlm"},
	{0x044E0E986AF0FA0EULL, "vxlt_undercityserialkiller_1ne_0_0.vlm"},
	{0x044E8D82A85866B5ULL, "projdecal_projdecal_p0.psh.fxo"},
	{0x044EA06FBB5B63CFULL, "6104923_m.fxe"},
	{0x044FD60137126FA8ULL, "vxlt_gallowstemplar_f_5c1_0_0.vlm"},
	{0x0450938AD2B4EBEEULL, "13484574.ogg"},
	{0x0450B55B2118A6CDULL, "845095204.ogg"},
	{0x0451138F78BE5C18ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookofinquisition_n.dds"},
	{0x045131456FDB310FULL, "zz_qcr_debug6058852.cl"},
	{0x045199AA1C9ED5B8ULL, "6136701_m.fxe"},
	{0x0451CB979E71CA09ULL, "920692722.ogg"},
	{0x0451CE380ACC615EULL, "532887636.ogg"},
	{0x0451E590329B8307ULL, "llac_ht_0_hightown_a3_-131073_1.dds"},
	{0x0452387826C760A1ULL, "vxht_bartrand_n_664_0_0.vlm"},
	{0x04528F6378E116C9ULL, "915052311.ogg"},
	{0x04529F1462685457ULL, "ts_raiders_prog_junk2.gda"},
	{0x0452EDC6D09339C4ULL, "art\\characters\\playercharacter\\elfmale\\cth_comc.mmh"},
	{0x0452F1C6D0934017ULL, "art\\characters\\playercharacter\\elfmale\\cth_comc.mao"},
	{0x045303C6D0935F66ULL, "art\\characters\\playercharacter\\elfmale\\cth_comc.msh"},
	{0x0453621C99921397ULL, "26864461.ogg"},
	{0x045375DD7CAFF2B4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexa.phy"},
	{0x0453787BC421732FULL, "vxw_encounter1_ip_0_1.vlm"},
	{0x045399874D27FA44ULL, "6046838_f.fxe"},
	{0x045401623B0BA6F1ULL, "qcr221cr_qunari_honor_guard.utc"},
	{0x04549DF6C519AC2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rainbow.evt"},
	{0x0454C2565D42FB57ULL, "w_encounter2_camp_mmap.dds"},
	{0x0454DF701F3B4D5CULL, "194085175.ogg"},
	{0x04551622763D5BADULL, "6087013_m.fxe"},
	{0x045564BFC8AD044EULL, "709392334.ogg"},
	{0x0455A2F8DE1662A2ULL, "vxdr_primevaltomb_3ir_0_0.vlm"},
	{0x0455FA5F55FCF2DCULL, "sw_sh.mao"},
	{0x045632933C296A48ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_highthistle.mmh"},
	{0x04563E933C297EEBULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_highthistle.mao"},
	{0x045650933C299D72ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_highthistle.msh"},
	{0x04567F57F0B401E1ULL, "vxw_blightlands_os_0_0.vlm"},
	{0x0456856BA8DCEEDBULL, "vxw_encounter2_camp_ji_0_0.vlm"},
	{0x045696D74CF15DE2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pulpit_sa.dds"},
	{0x0456AE825B5B56ADULL, "hightown_fem_cityguard_mob6168528.cl"},
	{0x0456B65859DFC864ULL, "zz_followers_banter6159561.cl"},
	{0x0456C90FCA3A888FULL, "544282309.ogg"},
	{0x0456F7D5CD26E077ULL, "w_encounter2_60_0.rml"},
	{0x0457B5A439E6C1A8ULL, "qun160_vashothlead6053946.cl"},
	{0x0457E97ECD8E7D7BULL, "vxlt_0_lowtown_docks_b6j_0_1.vlm"},
	{0x045802392B4C1A73ULL, "vxca_caverns1_mines_geo_0_0.vlm"},
	{0x045818EF9B2B0616ULL, "6103400_m.fxe"},
	{0x04584CE9891CFA52ULL, "1003989707.ogg"},
	{0x04586BA298BFCA98ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.imp.ani"},
	{0x0458B401DAD248D3ULL, "gen00fl_aveline6114011.cl"},
	{0x0458FF8E4AA0ED94ULL, "vxlt_hangedman_21j_0_0.vlm"},
	{0x045940355B54DB21ULL, "270346430.ogg"},
	{0x04594569DD4F7997ULL, "w_blightlands_31_0.msh"},
	{0x04595769DD4F98FAULL, "w_blightlands_31_0.mao"},
	{0x04595B69DD4F9F09ULL, "w_blightlands_31_0.mmh"},
	{0x0459B9B483FE104FULL, "6173811_m.fxe"},
	{0x0459E8B57AA41198ULL, "hlac_w_woundedcoast_327685_0z.dds"},
	{0x045A25A10A1515B9ULL, "vxlt_0_lowtown_a2_n_fe2_0_0.vlm"},
	{0x045A5805EFA73074ULL, "hlac_lt_0_lowtown_docks_n_-196609_1x.dds"},
	{0x045A7F33834A72E9ULL, "568985712.ogg"},
	{0x045AF368CF0AFFAEULL, "103727455.ogg"},
	{0x045B1ECFDE2FF3BBULL, "vxlt_hangedman_n_209_0_0.vlm"},
	{0x045B27FFEBDD4D3AULL, "6088910_m.fxe"},
	{0x045B9EF6C51FCCD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rainbow.gad"},
	{0x045BE39A0E185F93ULL, "6188673_m.fxe"},
	{0x045C934D27F49A41ULL, "one220_yevhen6099030.cl"},
	{0x045CF1A890540FD0ULL, "365862714.ogg"},
	{0x045CF6D5CD2B4A6FULL, "w_encounter2_60_0.phy"},
	{0x045D137EA0DCAEFEULL, "6078796_f.fxe"},
	{0x045D52DD7CB80BB1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexa.msh"},
	{0x045D5498F4DC4FF3ULL, "w_bonepit_20_0.rml"},
	{0x045D60DD7CB823CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexa.mao"},
	{0x045D6CDD7CB8386FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexa.mmh"},
	{0x045DA51B3C103DD8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bush.mao"},
	{0x045E523A430C08CBULL, "vxlt_0_lowtown_docks_b52_0_0.vlm"},
	{0x045E8E230AFC5188ULL, "vxht_playershouse_n_1o3_0_0.vlm"},
	{0x045E9BB113B9BC70ULL, "6101826_m.fxe"},
	{0x045EB7C9536CCE1AULL, "811027582.ogg"},
	{0x045EC8FF7B267AD1ULL, "hlac_lt_0_lowtown_docks_196607_2x.dds"},
	{0x045F0B964653BDDEULL, "art\\vfx\\combat\\vfx_gib_ogre_bld_p.anb"},
	{0x045F20DE85C5B18BULL, "hlac_ca_caverns1_mines_24427_3y.dds"},
	{0x045F25B94846B54CULL, "gen00fl_anders6125273.cl"},
	{0x045F88449A035450ULL, "vxlt_gallowstemplar_n_55s_0_0.vlm"},
	{0x045FB7AB30925EFDULL, "6162455_m.fxe"},
	{0x045FC53713F45759ULL, "mag320_nobleman_3.bnk"},
	{0x0460080DB208ADC8ULL, "vxw_blightlands_ov_0_0.vlm"},
	{0x04600B870E94EE38ULL, "6131601_m.fxe"},
	{0x0460A30C9AE79051ULL, "59123258.ogg"},
	{0x0460CBA0931EF756ULL, "vxw_bonepit_1q6_0_0.vlm"},
	{0x0460F4F67E9C8588ULL, "761207995.ogg"},
	{0x0460FDE2422B610CULL, "dre000_bethany6037317.cl"},
	{0x046107231E7DF43DULL, "6018667_f.fxe"},
	{0x04611F58989DD82CULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluvian_d.dds"},
	{0x04611FDA96094357ULL, "hlac_ht_chantry_n_2534_34z.dds"},
	{0x04612F8A8919AFC0ULL, "6096906_m.fxe"},
	{0x0462690C9778384BULL, "6127519_m.fxe"},
	{0x04628D4FEFC2C5CDULL, "vxdr_primevaltomb_917_0_0.vlm"},
	{0x0462F2639AA6D2D2ULL, "gen_im_trs_gem_ruby_04.uti"},
	{0x04631750C835A224ULL, "37164881.ogg"},
	{0x046395A2B54E5324ULL, "820369356.ogg"},
	{0x0463B55B32869027ULL, "t3_wdg_tb3.tnt"},
	{0x0463E73713F879B9ULL, "mag320_nobleman_3.cnv"},
	{0x0463FBF7B36424EBULL, "gen00fl_anders6125579.cl"},
	{0x04645CE3CD67FCAAULL, "hlac_lt_0_lowtown_a3_n_c_-65537_5y.dds"},
	{0x0464643735CAC369ULL, "243639855.ogg"},
	{0x0464F7489AD3075BULL, "vxw_encounter2_camp_eh_0_0.vlm"},
	{0x0465A15812D4F6CFULL, "vxlt_0_lowtown_docks_a3_n_c3u_0_0.vlm"},
	{0x0465B7C8E2C91209ULL, "6164035_m.fxe"},
	{0x04668BC097F64726ULL, "837076207.ogg"},
	{0x0466F5A2EB3C19C8ULL, "ht_playershouse_n.arl"},
	{0x046729F5BEAFA475ULL, "follower_banter6171579.cl"},
	{0x0467F5E4DB09E5CDULL, "one100dg_kelder6055425.cl"},
	{0x04686923B194F89FULL, "vxht_0_hightown_a2_ava_0_0.vlm"},
	{0x04686D0D2436E6C8ULL, "578012366.ogg"},
	{0x0468BA915647BF6DULL, "vxlt_0_lowtown_a3_n_c_hr2_0_0.vlm"},
	{0x0468CFD20FAAFFFFULL, "zz_ave_debug6175058.cl"},
	{0x046947275BF22397ULL, "6174964_m.fxe"},
	{0x046A046373DB19FCULL, "glows.dds"},
	{0x046A6D915483307DULL, "6034046_m.fxe"},
	{0x046AB1E0EA9EF38BULL, "206651454.ogg"},
	{0x046AB4E7ECF8FFBEULL, "6119511_m.fxe"},
	{0x046AD60854BB2727ULL, "632421234.ogg"},
	{0x046B160595C25A5CULL, "mag360_samson6044923.cl"},
	{0x046B47220D40FC61ULL, "6209945_m.fxe"},
	{0x046CBC1DA4A470B3ULL, "vxht_mansion2_1v1_0_0.vlm"},
	{0x046CC15ED63269CDULL, "hlac_lt_gallowscourtyard_-131073_1x.dds"},
	{0x046E1E816DA5E184ULL, "77426167.ogg"},
	{0x046E291B002CB497ULL, "mag311_keran6061822.cl"},
	{0x046E79EAA3FEC648ULL, "vxw_blightlands_pw_0_0.vlm"},
	{0x046EAF63C146F17CULL, "hlac_lt_0_lowtown_a3_n_c_-65536_5x.dds"},
	{0x046EBFED5C97D7E6ULL, "hlac_ht_0_hightown_0_2y.dds"},
	{0x046EDDC45726CE96ULL, "nat380_nathaniel6066618.cl"},
	{0x046EE33D6463BD96ULL, "vxlt_gallowstemplar_3ru_0_0.vlm"},
	{0x046EE6AFACE4A610ULL, "vxdr_0_deeproads_esf_0_0.vlm"},
	{0x046EF0E82DC60F20ULL, "vxht_brothel_n_3qu_0_0.vlm"},
	{0x046F3B3066DA7D2EULL, "vxlt_0_lowtown_n_exe_0_0.vlm"},
	{0x046F3D7D85150FD6ULL, "vxw_bonepit_1z7_0_0.vlm"},
	{0x046F4B6FCE808441ULL, "dre000_varric6027753.cl"},
	{0x046F64109DAB9D38ULL, "hlac_lt_0_lowtown_a2_n_c_65536_6y.dds"},
	{0x046F855039A9E214ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath3.gad"},
	{0x046FFCB667E2E161ULL, "qun120_arishok_guard6060604.cl"},
	{0x04701D459ED5E6F8ULL, "6163242_m.fxe"},
	{0x0470C02F6212FBEDULL, "vxca_cavern1_haunted_e7m_0_0.vlm"},
	{0x0470E2541C55135FULL, "6177830_f.fxe"},
	{0x0471510E993F64D9ULL, "6108774_m.fxe"},
	{0x04715234E26FF3C1ULL, "6101073_f.fxe"},
	{0x047159708D9EFC07ULL, "6181757_m.fxe"},
	{0x0471C69A05A61810ULL, "891693116.ogg"},
	{0x0471D8485094B376ULL, "vxw_sundermount_2nk_0_0.vlm"},
	{0x0471DFBFF4D37061ULL, "gen00fl_varric6105421.cl"},
	{0x0472365588D07766ULL, "6026976_m.fxe"},
	{0x0472D17158D025CDULL, "vxw_encounter2_camp_8e_0_0.vlm"},
	{0x04731CB3F372DF39ULL, "6099211_f.fxe"},
	{0x0473210D28765FF8ULL, "643539622.ogg"},
	{0x047342C8A3A4146BULL, "6066707_f.fxe"},
	{0x047351FE2C30680DULL, "6028163_f.fxe"},
	{0x04736311D4A032E1ULL, "vxw_blightlands_uq_0_0.vlm"},
	{0x0473C358AFCE4E25ULL, "det_smo.mao"},
	{0x0475090E978C2DD8ULL, "274053304.ogg"},
	{0x047566ABA40488E0ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hflatgrass_da.dds"},
	{0x0475B7C301502C5FULL, "6088663_m.fxe"},
	{0x0475D2CEF93E8E91ULL, "975791807.ogg"},
	{0x04760534C10E3B6DULL, "hlac_dr_slavers_3207_4y.dds"},
	{0x04764302182C367DULL, "132621602.ogg"},
	{0x0476945039B01AF2ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath3.ani"},
	{0x0476FDC319B008D8ULL, "one220_yevhen6135788.cl"},
	{0x047760A7804C93CFULL, "263196295.ogg"},
	{0x0477802647290DE2ULL, "plt_one000pt_talked_to.nss"},
	{0x0478F4003B6AC1F7ULL, "6178371_m.fxe"},
	{0x047902DAE944A083ULL, "6126992_m.fxe"},
	{0x0479077993F822DDULL, "ts_beasts_prog_liver.gda"},
	{0x04793AF9E4FA6BA2ULL, "hlac_w_encounter3_-262139_0y.dds"},
	{0x047960802A39B50BULL, "hlac_w_blightlands_65537_0z.dds"},
	{0x047967827E4C70FDULL, "docks_male_sailor_mob6176670.cl"},
	{0x0479B71306B7A3D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf.ani"},
	{0x047A710363F240EFULL, "6112339_m.fxe"},
	{0x047AC24466D772A3ULL, "hlac_w_bonepit_196614_0x.dds"},
	{0x047B19821F698E5AULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunarirelic_n.dds"},
	{0x047B2DFA969A8933ULL, "vxht_mansion2_1y0_0_0.vlm"},
	{0x047BABF8D890404FULL, "hlac_lt_0_lowtown_docks_-65537_3y.dds"},
	{0x047BB4D27162C850ULL, "hlac_lt_undercityencounter_-1_44z.dds"},
	{0x047C254BCC15EB0CULL, "gen_im_trs_rdr_med_boo_r0.uti"},
	{0x047D30231C21A5E2ULL, "vxlt_0_lowtown_a2_n_c_hws_0_0.vlm"},
	{0x047DAD0D58D095AEULL, "vxlt_0_lowtown_n_euf_0_0.vlm"},
	{0x047DB61B569927F5ULL, "vxlt_gallowstemplar_n_5xa_0_0.vlm"},
	{0x047DDFBFF481A674ULL, "hlac_ht_chantry_2534_27x.dds"},
	{0x047E667D91716503ULL, "w_encounter2_camp_60_0.mao"},
	{0x047E6A7D91716C50ULL, "w_encounter2_camp_60_0.mmh"},
	{0x047E787D9171846AULL, "w_encounter2_camp_60_0.msh"},
	{0x047EC08275A93292ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medd_0s.dds"},
	{0x047ED691ADC9844AULL, "665082386.ogg"},
	{0x047ED93CF86AB889ULL, "plt_qun100pt_mage_escort.nss"},
	{0x047F09BC5790DE76ULL, "6083318_f.fxe"},
	{0x047F102B26D8EE83ULL, "530781805.ogg"},
	{0x047F5ED5A4714ADDULL, "follower_banter6164490.cl"},
	{0x0480386049760591ULL, "t1_mue_or3.tnt"},
	{0x04804A25428ACBF6ULL, "vxw_sundermount_2sl_0_0.vlm"},
	{0x04809C3D2AE01F32ULL, "222395906.ogg"},
	{0x0480AF4645086021ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc105a_0d.dds"},
	{0x04813CD4BDEBC588ULL, "6120110_f.fxe"},
	{0x0481434E4AC63E4DULL, "vxw_encounter2_camp_3d_0_0.vlm"},
	{0x048146DDF9219DF2ULL, "dae200_bran6094766.cl"},
	{0x048160CFEFD7461BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galofloor.phy"},
	{0x04819243C447AD73ULL, "6138739_m.fxe"},
	{0x0481EE238D3840B8ULL, "llac_lt_gallowstemplar_f_2395_0.dds"},
	{0x04825869DD729F3EULL, "w_blightlands_31_0.anb"},
	{0x04827FCAF7CDE862ULL, "vxlt_hangedman_n_26b_0_0.vlm"},
	{0x0482F75E2BA32412ULL, "vxw_sundermount_2zr_0_1.vlm"},
	{0x0483518C1FD7D147ULL, "qun211cr_elf_warrior6072200.cl"},
	{0x048381330A38B59DULL, "766552943.ogg"},
	{0x0483BC12B37ED3B1ULL, "6110490_m.fxe"},
	{0x0483C8D691B512F0ULL, "6089577_m.fxe"},
	{0x0484299FF34644DFULL, "6083763_m.fxe"},
	{0x0484AE7C83240C04ULL, "vxlt_gallowscourtyard_n_65b_0_0.vlm"},
	{0x0484C1F861AF3B5AULL, "6207423_m.fxe"},
	{0x04858C26C7EBB4F8ULL, "vxw_bonepitmining_a2_my_0_0.vlm"},
	{0x04859E4B3A33F06DULL, "vxlt_0_lowtown_a3_n_c_hd0_0_0.vlm"},
	{0x0485BAB145AD4208ULL, "759675933.ogg"},
	{0x048618C3158EAD7EULL, "lowtown_male_alienage1_sta6161629.cl"},
	{0x0486BF1306C23BC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf.evt"},
	{0x04875BA299583F27ULL, "79879180.ogg"},
	{0x04876E020FB968B8ULL, "t3_arm_mer.tnt"},
	{0x0487B1D3D31893C2ULL, "vxdr_0_deeproads_ewh_0_0.vlm"},
	{0x048835C7540BE74AULL, "vxca_caverns1_gkq_0_0.vlm"},
	{0x0488A10B9EBBE0F2ULL, "vxht_bartrand_n_5nl_0_0.vlm"},
	{0x0488E39F82185C24ULL, "hlac_lt_0_lowtown_a2_-131072_4z.dds"},
	{0x0488FFE6238D36B1ULL, "mcr371_hostile_aveline6082178.cl"},
	{0x04894977D2D0002CULL, "llac_dr_primevalentrance_1725_3.dds"},
	{0x0489BCF49C72DEB9ULL, "gencr_gate_guardian_statue.utc"},
	{0x0489C31D0566E488ULL, "plt_gen00pt_money.plo"},
	{0x0489D9342A92D199ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1_tool.phy"},
	{0x0489FA496C788A91ULL, "6149760_m.fxe"},
	{0x048A64E2C4854F83ULL, "dae300_petrice6097867.cl"},
	{0x048AC8EA53641FDFULL, "6146811_m.fxe"},
	{0x048AFE4AB22583B5ULL, "6053337_m.fxe"},
	{0x048B4D1B307DD503ULL, "hightown_market_guard_sta6169601.cl"},
	{0x048C14C4E80AD1A7ULL, "vxlt_backalley_n_1y8_0_0.vlm"},
	{0x048C74A935428233ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.dura.evt"},
	{0x048C96CAC4776ABCULL, "mrl250pt_mirrorimg.ncs"},
	{0x048DE01306C89210ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf.gad"},
	{0x048EBC023480E476ULL, "vxw_sundermount_2tm_0_0.vlm"},
	{0x048EE3911B011E86ULL, "511744329.ogg"},
	{0x048FE58CEB8AFE03ULL, "alphamagicarc.mao"},
	{0x0490D5FD68A030DEULL, "gen00fl_sebastian6180495.cl"},
	{0x0490EA46B1A5D4ACULL, "304442948.ogg"},
	{0x04911001C088B897ULL, "vxca_cavern1_haunted_j5r_0_0.vlm"},
	{0x049182533D3B193BULL, "llac_lt_0_lowtown_a3_n_c_131072_4.dds"},
	{0x04923CAD361E2581ULL, "6142519_f.fxe"},
	{0x04927DE080920ACAULL, "llac_ht_mansion2b_818_57.dds"},
	{0x0492E243C44F86BAULL, "addtalent.ncs"},
	{0x04936DA935489574ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.dura.gad"},
	{0x049377AF0194DBBEULL, "vxlt_0_lowtown_e8b_0_0.vlm"},
	{0x049389A62D6353E9ULL, "qun160_vashothlead6052564.cl"},
	{0x0493A1C9330B038DULL, "vxlt_gallowscourtyard_n_7av_0_0.vlm"},
	{0x0494404CBEE47B5FULL, "vxw_bonepitmining_a2_2af_0_0.vlm"},
	{0x0495694958DB3D64ULL, "ach_kill_combo.dds"},
	{0x04958C02954E57CCULL, "64504191.ogg"},
	{0x0495A4DB479A1EFDULL, "6206867_m.fxe"},
	{0x0495C41C14F87263ULL, "6118504_m.fxe"},
	{0x049623B0C50EAC42ULL, "vxdr_0_deeproads_epi_0_0.vlm"},
	{0x0496246A570B4AE3ULL, "dae300_cullen6142067.cl"},
	{0x0496BE3FFDC73529ULL, "6082160_m.fxe"},
	{0x0496F881A1B6B3F9ULL, "50520841.ogg"},
	{0x049744CE0A572588ULL, "gen00fl_fenris6106706.cl"},
	{0x0497EDCA55BCF013ULL, "6143416_m.fxe"},
	{0x04980DF71F3A1344ULL, "character_character_p30.psh.fxo.dx11.assembly"},
	{0x04986C265E6EA311ULL, "6142660_m.fxe"},
	{0x04986DEAA7F518ECULL, "vxht_0_hightown_n_c7l_0_0.vlm"},
	{0x04987F4189D23FCDULL, "vxca_caverns1_j7c_0_0.vlm"},
	{0x049898F2EB91A676ULL, "332122720.ogg"},
	{0x0498EBC4909266AFULL, "rog_critsneak.dds"},
	{0x0499060459056395ULL, "422984921.ogg"},
	{0x049987A46269BFCFULL, "128332134.ogg"},
	{0x04998C2FFB59D1B2ULL, "hlac_lt_0_lowtown_docks_n_-4_7y.dds"},
	{0x049A50B13F017308ULL, "vxlt_warehouse_1dv_0_0.vlm"},
	{0x049A7CA9354ECD52ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.dura.ani"},
	{0x049AF4C26D2EE3E6ULL, "438809916.ogg"},
	{0x049B13AFF215FC40ULL, "hla_ht_0_hightown_a3_-65536_0y.dds"},
	{0x049B53472468C523ULL, "mag170_thrask6042848.cl"},
	{0x049BA901CEE8AF52ULL, "one100dg_vanard6017496.cl"},
	{0x049BC28E99979EA4ULL, "vxlt_gallowstemplar_3r7_0_0.vlm"},
	{0x049C119330F6B04EULL, "gen00fl_anders6114486.cl"},
	{0x049C2048B50BA91FULL, "hlac_lt_0_lowtown_a2_n_2_0y.dds"},
	{0x049C3C1609765E1CULL, "hlac_lt_hangedman_-1_26y.dds"},
	{0x049C58EFA30CBB34ULL, "635789606.ogg"},
	{0x049C706E83634454ULL, "6056829_m.fxe"},
	{0x049C8A2E31DC09EAULL, "vxlt_0_lowtown_a2_n_c_gwh_0_0.vlm"},
	{0x049CF1B76B7E81B3ULL, "art\\characters\\weapons\\mauls\\w_mal_ml106.mao"},
	{0x049D0D6BDEE00D75ULL, "zz_vault_debug6203285.cl"},
	{0x049D2DDF2676FCF6ULL, "vxw_sundermount_2yn_0_0.vlm"},
	{0x049D5147FA30154BULL, "341609295.ogg"},
	{0x049DC081AB733C87ULL, "vxht_mansion2_1wn_0_0.vlm"},
	{0x049DE12714668D4DULL, "gen00fl_isabela6100075.cl"},
	{0x049DE8AE51051CBCULL, "seb201_lady_harimann6197709.cl"},
	{0x049F3C0273D9445CULL, "dae021_male_worshipper_sta6154417.cl"},
	{0x049F7054C40117B1ULL, "debug.gfx"},
	{0x049F9AAF692F678AULL, "vxlt_0_lowtown_docks_a3_9nn_0_0.vlm"},
	{0x04A046D1F8ACAD03ULL, "w_woundedcoast_106_0.mao"},
	{0x04A04AD1F8ACB450ULL, "w_woundedcoast_106_0.mmh"},
	{0x04A058D1F8ACCC6AULL, "w_woundedcoast_106_0.msh"},
	{0x04A059D512255012ULL, "hm_feynriel_b.mor"},
	{0x04A12921DEE2D9CCULL, "mag211_marethari6025314.cl"},
	{0x04A14DDA97D0FDDAULL, "hlac_ca_caverns1_24427_1y.dds"},
	{0x04A1E98BF38AF43EULL, "vxlt_0_lowtown_e3c_0_0.vlm"},
	{0x04A217B8ECACD475ULL, "seb221_flora_harimann6177694.cl"},
	{0x04A24603C28A3298ULL, "ser221_gascard6165443.cl"},
	{0x04A3492BE7C0C46AULL, "art\\vfx\\combat\\vfx_r_var_kback_act_c.anb"},
	{0x04A435F906EE8AE3ULL, "6117404_m.fxe"},
	{0x04A4B26F8E01DF4FULL, "dae000_templar_recruit_1.bnk"},
	{0x04A4CE961CBCB2C1ULL, "vxht_0_hightown_a2_c4g_0_0.vlm"},
	{0x04A4D5DAA537B723ULL, "6144743_m.fxe"},
	{0x04A4E19BFCDCCE7AULL, "gen00fl_dog6141140.cl"},
	{0x04A5B4F5C97923BAULL, "llac_ca_cavern1_haunted_24421_9.dds"},
	{0x04A5B99C13173E14ULL, "llac_dr_primevaltomb_9621_2.dds"},
	{0x04A5F33375302EC0ULL, "hlac_ca_cavern1_haunted_24424_23y.dds"},
	{0x04A61E531B3CFF24ULL, "6054391_m.fxe"},
	{0x04A6B8EAF9142196ULL, "vxw_encounter2_b0_0_0.vlm"},
	{0x04A6D38D6632C36EULL, "vxlt_warehouse_13y_0_0.vlm"},
	{0x04A6D6F1D1DBB9D5ULL, "979369949.ogg"},
	{0x04A707363A7AF28CULL, "vxw_bonepit_26n_0_0.vlm"},
	{0x04A71F52C0025F26ULL, "6069951_m.fxe"},
	{0x04A76D2647521165ULL, "plt_one000pt_talked_to.plo"},
	{0x04A7A02A1301417BULL, "6068441_m.fxe"},
	{0x04A7D46F8E044EB7ULL, "dae000_templar_recruit_1.cnv"},
	{0x04A84FAEC656966AULL, "6056830_m.fxe"},
	{0x04A87FAB3138BBE7ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_0x.dds"},
	{0x04A8AF4FA5C5091AULL, "vxw_bonepitdestroyed_a3_9i_0_0.vlm"},
	{0x04A932559A91DBCBULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.cess.evt"},
	{0x04A94F0E62174736ULL, "llac_lt_0_lowtown_a2_2_1.dds"},
	{0x04A9FA9ADFDA7A87ULL, "6060705_f.fxe"},
	{0x04AA1CAA4427BE36ULL, "vxht_bartrand_n_65f_0_0.vlm"},
	{0x04AA40FED85C468AULL, "mcr310_argument6127606.cl"},
	{0x04AA83CE1CABE712ULL, "vxht_0_hightown_n_9xe_0_0.vlm"},
	{0x04AAFC0B23D2226AULL, "vxlt_0_lowtown_a2_n_c_gpi_0_0.vlm"},
	{0x04AB6CB8BB02E007ULL, "art\\characters\\weapons\\dagger\\w_dgr_105a.mao"},
	{0x04AB714F1323950BULL, "6054491_m.fxe"},
	{0x04AC71EC38F87438ULL, "gen_im_trs_krk_junk2_01.uti"},
	{0x04AC82C40A409EE9ULL, "vxdr_0_deeproads_eue_0_0.vlm"},
	{0x04AD0F4A8CFACB70ULL, "vxlt_gallowstemplar_f_6sv_0_0.vlm"},
	{0x04AD3237DDDA97DEULL, "mer140_bonepit_comments6077813.cl"},
	{0x04AD4F92F1B5E8BAULL, "vxlt_backalley_g_1vu_0_0.vlm"},
	{0x04AD55F719FBCDA4ULL, "ef_gen_act2.mor"},
	{0x04ADA1CC95AC66BBULL, "vxht_0_hightown_a3_9zt_0_0.vlm"},
	{0x04ADECDF7E2A8EA3ULL, "vxht_keep_n_70j_0_0.vlm"},
	{0x04AE176A37A19E8BULL, "6172395_m.fxe"},
	{0x04AE47CFEFFE09BAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galofloor.msh"},
	{0x04AE51A91F507C84ULL, "6136662_m.fxe"},
	{0x04AE55CFEFFE21F3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galofloor.mao"},
	{0x04AE59CFEFFE2800ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galofloor.mmh"},
	{0x04AE71AA7D574F42ULL, "365475068.ogg"},
	{0x04AEDCC26BFD87F6ULL, "tbe100_xenon6189674.cl"},
	{0x04AF8BE36F5A8CCAULL, "ave110_ambush_26105930.cl"},
	{0x04AFBAC0E81F74EEULL, "vxca_caverns2prc_a13_0_0.vlm"},
	{0x04AFC28E0D4D122EULL, "art\\vfx\\creatures\\fxc_ashwraith_c.dura.evt"},
	{0x04AFE07E94F11365ULL, "6043365_m.fxe"},
	{0x04AFFF94C2EB6E32ULL, "genip_node_d1.utp"},
	{0x04B04B559A98248CULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.cess.gad"},
	{0x04B0B84ED232DEBDULL, "art\\vfx\\combat\\cn_gut.mao"},
	{0x04B0F0B30CAA8911ULL, "llac_dr_primevaltomb_665_0.dds"},
	{0x04B1537ED7905580ULL, "233086860.ogg"},
	{0x04B1577FD717D30CULL, "vxca_caverns2_2qw_0_0.vlm"},
	{0x04B16305D1900806ULL, "vxdr_primevaltomb_2vo_0_1.vlm"},
	{0x04B1F3575E62DE66ULL, "6030603_m.fxe"},
	{0x04B209B1F995550DULL, "vxca_caverns1_mines_j4c_0_0.vlm"},
	{0x04B2150B6EC46275ULL, "511082057.ogg"},
	{0x04B23A5D926A88F9ULL, "vxht_0_hightown_a3_c5a_0_0.vlm"},
	{0x04B26E3C349FD885ULL, "hlac_lt_0_lowtown_a3_n_c_-1_6z.dds"},
	{0x04B2A7D5F8E4A363ULL, "6112704_m.fxe"},
	{0x04B340730EB2CB41ULL, "vxht_0_hightown_a2_c3f_0_0.vlm"},
	{0x04B354876223DADAULL, "413002776.ogg"},
	{0x04B398DB4CE6A4DEULL, "dae010_alienage_elf_f2.utc"},
	{0x04B3D7D90E0C7D91ULL, "dae000cr_grand_cleric6144096.cl"},
	{0x04B46F66CB45D6B5ULL, "hlac_ht_brothel_n_2995_3z.dds"},
	{0x04B4CE8B1D510361ULL, "hlac_lt_0_lowtown_a2_n_-131072_4z.dds"},
	{0x04B4EBC028B433D7ULL, "914868443.ogg"},
	{0x04B4F7C485091A14ULL, "hlac_dr_primevaltomb_11985_3y.dds"},
	{0x04B5223EB437B537ULL, "gen00fl_merrill6108437.cl"},
	{0x04B52AC7EB0A3A16ULL, "vxw_encounter2_m1_0_0.vlm"},
	{0x04B579132C710B0CULL, "vxw_bonepit_21o_0_0.vlm"},
	{0x04B5FD8BF9573525ULL, "vxw_sundermount_2m5_0_0.vlm"},
	{0x04B64A559A9C8E8AULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.cess.ani"},
	{0x04B67A80B96420D5ULL, "vxht_mansion2b_3ed_0_0.vlm"},
	{0x04B6917ABEEB890DULL, "343108042.ogg"},
	{0x04B6B98E0D5321D9ULL, "art\\vfx\\creatures\\fxc_ashwraith_c.dura.gad"},
	{0x04B6E22CB679B230ULL, "232855076.ogg"},
	{0x04B7200066B32090ULL, "vxdr_0_deeproads_ehc_0_0.vlm"},
	{0x04B73A876FEAEA12ULL, "666294689.ogg"},
	{0x04B73E0C3C129A96ULL, "174706719.ogg"},
	{0x04B755756F98CCC2ULL, "6107458_f.fxe"},
	{0x04B76A7D91A13113ULL, "w_encounter2_camp_60_0.rml"},
	{0x04B7C3EE50E846A5ULL, "and100_anders6173569.cl"},
	{0x04B7D80916ECADE2ULL, "820403571.ogg"},
	{0x04B80AD7B2331960ULL, "547249313.ogg"},
	{0x04B83860F34323EEULL, "81260081.ogg"},
	{0x04B8E2D4AEE8356BULL, "hlac_lt_0_lowtown_n_-131072_5y.dds"},
	{0x04B8EF713CDE4954ULL, "art\\levels\\races\\design\\prp_bodybaga_1.mao"},
	{0x04B9561F1136F2A6ULL, "art\\characters\\creatures\\spider\\textures\\c_corspidra_0d.dds"},
	{0x04B9BA34616F8A17ULL, "ts_circleofmagi_prog_junk3.gda"},
	{0x04B9E5012CE34DB7ULL, "127357187.ogg"},
	{0x04B9E95D87134D2BULL, "hlac_lt_shop_-1_0y.dds"},
	{0x04BA0144AF117774ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_dead_loop.gad"},
	{0x04BA37B2303A08F0ULL, "6057141_m.fxe"},
	{0x04BA43B925875B56ULL, "795014385.ogg"},
	{0x04BA59B3CFAB2A72ULL, "hlac_dr_primevaltomb_661_5x.dds"},
	{0x04BAAA70567CD21CULL, "w_blightlands_51_0.mao"},
	{0x04BAB670567CE6BFULL, "w_blightlands_51_0.mmh"},
	{0x04BABC70567CF161ULL, "w_blightlands_51_0.msh"},
	{0x04BACA7C8B4CCA90ULL, "6102580_m.fxe"},
	{0x04BADFC4D241BA7EULL, "vxdr_primevalentrance_16m_0_0.vlm"},
	{0x04BBF09DF7DAB6D5ULL, "gen00fl_anders6214666.cl"},
	{0x04BC3AB576B0B746ULL, "llac_lt_0_lowtown_docks_-4_3.dds"},
	{0x04BC619E048C59D8ULL, "mer240_jansen6050441.cl"},
	{0x04BCB015FD5E218CULL, "331563872.ogg"},
	{0x04BCD2B71B25BD46ULL, "vxlt_0_lowtown_docks_b4h_0_0.vlm"},
	{0x04BCD69CD2153688ULL, "ch2armtnt_dy.psh.fxo"},
	{0x04BD26E064B0B598ULL, "262200545.ogg"},
	{0x04BDA6BAC6CB6E27ULL, "llac_lt_hangedman_n_-1_27.dds"},
	{0x04BDCA8E0D595D27ULL, "art\\vfx\\creatures\\fxc_ashwraith_c.dura.ani"},
	{0x04BE717D91A75C6BULL, "w_encounter2_camp_60_0.phy"},
	{0x04BEBBEAA0753EAFULL, "vxca_cave_tz_0_0.vlm"},
	{0x04BEFBAF13625B21ULL, "llac_lt_0_lowtown_65533_0.dds"},
	{0x04BEFC8A5A6D4C91ULL, "dae000cr_grand_cleric6187949.cl"},
	{0x04BFE500703BC3DAULL, "736467513.ogg"},
	{0x04C080E4B4E1A2A7ULL, "lgt110_arl_bann_teagan6115602.cl"},
	{0x04C081571304F78EULL, "588873373.ogg"},
	{0x04C0FB2C83B9649FULL, "1067708547.ogg"},
	{0x04C11044AF17AF52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_dead_loop.ani"},
	{0x04C1FC9358DA7527ULL, "312653586.ogg"},
	{0x04C2B9C63A89808BULL, "6027952_m.fxe"},
	{0x04C2BE62531BB66FULL, "vxht_0_hightown_c8i_0_0.vlm"},
	{0x04C2F7F511975933ULL, "hlac_ht_mansion2_818_47x.dds"},
	{0x04C30220117C5552ULL, "gen00fl_varric6118967.cl"},
	{0x04C3258EAC693B3FULL, "vxht_0_hightown_n_c7c_0_0.vlm"},
	{0x04C39CA4DD005296ULL, "vxw_encounter2_h2_0_0.vlm"},
	{0x04C3BCD978D032CCULL, "character_character_p5.psh.fxo.dx11"},
	{0x04C40A66191CBBE2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_basic.phy"},
	{0x04C4C15510211B44ULL, "vxlt_warehouse_n_u1_0_0.vlm"},
	{0x04C541C560F394B1ULL, "112775813.ogg"},
	{0x04C5B7C685BCFAE9ULL, "docks_scene_dockworkers6176504.cl"},
	{0x04C63C794D13C3CEULL, "one300_zevran6089735.cl"},
	{0x04C668C58082FAF9ULL, "vxdr_0_deeproads_dqd_0_0.vlm"},
	{0x04C6755689B1EF66ULL, "vxw_sundermount_36w_0_1.vlm"},
	{0x04C6CDA13F17EE1DULL, "hlac_dr_primevalentrance_1726_3z.dds"},
	{0x04C6FB4034B9924EULL, "vxht_bartrand_n_5vh_0_0.vlm"},
	{0x04C705ABE121DDF8ULL, "plt_prm_cod_parth_staff.plo"},
	{0x04C86050F90F9371ULL, "vxw_bonepitmining_a2_21h_0_0.vlm"},
	{0x04C87A24CEA95FA4ULL, "mrl000_marethari6129203.cl"},
	{0x04C8BAEC31B36D43ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha14a.mmh"},
	{0x04C8BD713A6DEC4AULL, "vxlt_0_lowtown_docks_a3_d0u_0_0.vlm"},
	{0x04C8C8EC31B3849DULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha14a.msh"},
	{0x04C8E7081759A8C6ULL, "lgt160_ntb_elf_conflict6163394.cl"},
	{0x04C91354BCD88A30ULL, "vxht_0_hightown_a3_7g2_0_0.vlm"},
	{0x04C93BA2FDA154E6ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.cess.ani"},
	{0x04CA0B74BF11ED50ULL, "vxlt_0_lowtown_docks_n_b67_0_0.vlm"},
	{0x04CA334CD5A219BAULL, "vxlt_backalley_g_1xw_0_0.vlm"},
	{0x04CA900088298461ULL, "hlac_dr_0_deeproads_5842_8x.dds"},
	{0x04CAB82A14B37C3AULL, "6167532_m.fxe"},
	{0x04CB357D72B0A5AEULL, "dae000_gamlen6131099.cl"},
	{0x04CB4B41D6150A9FULL, "vxht_brothel_n_6g9_0_0.vlm"},
	{0x04CB63A55B3A7A43ULL, "6092299_m.fxe"},
	{0x04CBD0AF9B6C5FFCULL, "844188577.ogg"},
	{0x04CC4C2BE7E3CE5DULL, "art\\vfx\\combat\\vfx_r_var_kback_act_c.mmh"},
	{0x04CCA38367BE3E6DULL, "vxca_caverns2prc_a0w_0_0.vlm"},
	{0x04CCD3DB6EB0ACD9ULL, "574609747.ogg"},
	{0x04CD1C06AA79C4F3ULL, "hlac_ca_cavern1_haunted_24424_13y.dds"},
	{0x04CD37DD2C4B94FEULL, "vxht_0_hightown_a2_b8y_0_0.vlm"},
	{0x04CD8239DF962B34ULL, "vxht_0_hightown_a3_a6k_0_0.vlm"},
	{0x04CDEAF3744013A1ULL, "vxdr_primevaltomb_3lw_0_1.vlm"},
	{0x04CE099A4884FCEFULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax104a_0s.dds"},
	{0x04CE8BF9EE93053EULL, "vxca_caverns1_i99_0_0.vlm"},
	{0x04CF475E9C548F36ULL, "art\\vfx\\combat\\vfx_r_arc_front_close_c.anb"},
	{0x04CF690678EE394BULL, "darktown_male_sickly_sta.bnk"},
	{0x04CF7F3417F6A710ULL, "456273016.ogg"},
	{0x04CFB46AC4E22228ULL, "vxlt_gallowstemplar_n_5z2_0_0.vlm"},
	{0x04D0541898BC9568ULL, "135791899.ogg"},
	{0x04D0A4A54AB32C3FULL, "96253635.ogg"},
	{0x04D11C8D5C857B4AULL, "6125199_f.fxe"},
	{0x04D1923397965D6CULL, "6125338_m.fxe"},
	{0x04D1A2193696B6B8ULL, "vxdr_primevalentrance_2yi_0_0.vlm"},
	{0x04D1D1B2A4428C69ULL, "art\\characters\\playercharacter\\qunarimale\\mt_bas_qm_facem1.mao"},
	{0x04D1DC147CFB53B9ULL, "292693206.ogg"},
	{0x04D20E81CEF66B16ULL, "vxw_encounter2_s3_0_0.vlm"},
	{0x04D245143D38F8F1ULL, "ser270_emeric6113502.cl"},
	{0x04D26B0678F0722BULL, "darktown_male_sickly_sta.cnv"},
	{0x04D2B5D24333A10CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right.ani"},
	{0x04D302FF375A51D8ULL, "6144847_m.fxe"},
	{0x04D33332021733C4ULL, "vxlt_warehouse_n_x6_0_0.vlm"},
	{0x04D3C727659D13EAULL, "511724464.ogg"},
	{0x04D3F8BEA8DF5C0FULL, "537220783.ogg"},
	{0x04D41ACBF54E6CECULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potb.msh"},
	{0x04D420CBF54E76DEULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potb.mmh"},
	{0x04D42CCBF54E8B75ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potb.mao"},
	{0x04D45388F9D9898EULL, "hlac_ht_bartrand_n_3373_0y.dds"},
	{0x04D48B1615E235FDULL, "6112121_f.fxe"},
	{0x04D4B4A1D6851795ULL, "plt_cod_ltr_junk_starkprinc.plo"},
	{0x04D52DA3C30F544CULL, "6040955_f.fxe"},
	{0x04D53A76FD97FC0CULL, "w_blightlands_87_0.msh"},
	{0x04D54076FD9806FEULL, "w_blightlands_87_0.mmh"},
	{0x04D54C76FD981A95ULL, "w_blightlands_87_0.mao"},
	{0x04D56EC0E1ABBDB1ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc105a_0t.dds"},
	{0x04D5A31CA9E39771ULL, "294386887.ogg"},
	{0x04D5AC5D0AE7084DULL, "vxdr_primevalentrance_272_0_0.vlm"},
	{0x04D5F83DDDDCCDC5ULL, "945810808.ogg"},
	{0x04D65CB784D82F50ULL, "prty_seb.dds"},
	{0x04D6A668546AE58BULL, "vxlt_darktown_5oa_0_0.vlm"},
	{0x04D6E9C2C18260A6ULL, "vxht_brothel_n_6bs_0_0.vlm"},
	{0x04D723A2FDAD6957ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.cess.evt"},
	{0x04D787F5734B9E87ULL, "vxht_mansion2_1kj_0_0.vlm"},
	{0x04D7ADC25C3EEAF8ULL, "vxdr_primevaltomb_7ua_0_1.vlm"},
	{0x04D7B5BC9DC73138ULL, "ht_brothel_n_3372.rml"},
	{0x04D859D884017B39ULL, "vxlt_0_lowtown_docks_a3_n_d39_0_1.vlm"},
	{0x04D8B8EBF27A65ABULL, "404980372.ogg"},
	{0x04D8F1CAC7FBF82AULL, "hlac_lt_0_lowtown_n_-65538_0z.dds"},
	{0x04D91422467EBFF4ULL, "narrative_varric6124292.cl"},
	{0x04D9BBEC31C1E268ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha14a.phy"},
	{0x04D9C2D24339D64EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right.gad"},
	{0x04D9E4351AF8F0F1ULL, "6175990_m.fxe"},
	{0x04D9ED06DD7199C6ULL, "gen00fl_carver6108887.cl"},
	{0x04DA4630DB65513CULL, "one200pt_fools_gold.ncs"},
	{0x04DB06688F27D9FFULL, "372507193.ogg"},
	{0x04DB8DD76779FD95ULL, "6135921_m.fxe"},
	{0x04DC1AA7E5115755ULL, "506307671.ogg"},
	{0x04DC4A5E9C5F1EA1ULL, "art\\vfx\\combat\\vfx_r_arc_front_close_c.mmh"},
	{0x04DD2343F291F779ULL, "327899676.ogg"},
	{0x04DD7FE8111DBC86ULL, "lt_0_lowtown_a2_n_c_19_0_lc.mmh"},
	{0x04DDBA2C1D1B9B8CULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urvb_0d.dds"},
	{0x04DDCF59CEE708F6ULL, "lgt260_eau_tmp_angelique6143810.cl"},
	{0x04DE24A2FDB389E0ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.cess.gad"},
	{0x04DE5C16E009A2D5ULL, "dae000ip_crowd_mf4_talk18.utp"},
	{0x04DEDB40F2B91CFFULL, "1028048000.ogg"},
	{0x04DF0546383918DAULL, "mag291_ntbanter.cnv"},
	{0x04DF2C6FC42450A1ULL, "6177281_f.fxe"},
	{0x04DF883213665F23ULL, "6069412_f.fxe"},
	{0x04DF8E6A4E7B93CAULL, "6122099_f.fxe"},
	{0x04DF993A9C1F1172ULL, "vxlt_0_lowtown_a2_n_gv5_0_0.vlm"},
	{0x04DF9EE307E78524ULL, "s_ballista_06_l2.mao"},
	{0x04DFAE03299A9F0AULL, "97719090.ogg"},
	{0x04DFF356AD5BB1BCULL, "vxca_cave_2dw_0_0.vlm"},
	{0x04E00410898C75ECULL, "6122238_m.fxe"},
	{0x04E00A0B04ED8239ULL, "6079298_m.fxe"},
	{0x04E01FC971008DA2ULL, "vxlt_0_lowtown_docks_c1n_0_0.vlm"},
	{0x04E051D1F8E2A46BULL, "w_woundedcoast_106_0.phy"},
	{0x04E0BDD2433FEC2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right.evt"},
	{0x04E16278EECF4F6DULL, "6174666_f.fxe"},
	{0x04E20A19203E1B24ULL, "vxw_encounter2_ms_0_0.vlm"},
	{0x04E22746383B8822ULL, "mag291_ntbanter.bnk"},
	{0x04E22E1492443CEEULL, "dre000_vampire6034414.cl"},
	{0x04E24DA928F13893ULL, "vxca_caverns1_mines_ikk_0_0.vlm"},
	{0x04E28A40121B28EEULL, "hlac_ht_mansion2_818_62x.dds"},
	{0x04E2E626F140E2DDULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_magb_0s.dds"},
	{0x04E3B770569FEDFCULL, "w_blightlands_51_0.anb"},
	{0x04E51171B0648317ULL, "hlac_dr_primevaltomb_664_0y.dds"},
	{0x04E54033C11B0F5DULL, "vxht_chantry_2v8_0_0.vlm"},
	{0x04E56685FF98CE48ULL, "vxlt_darktown_468_0_0.vlm"},
	{0x04E5835094412A05ULL, "hlac_lt_0_lowtown_a3_n_c_2_1y.dds"},
	{0x04E59BCB97BC2C30ULL, "ser270_emeric6149353.cl"},
	{0x04E5F9D26541B707ULL, "vxht_mansion2_1ne_0_0.vlm"},
	{0x04E634E8C77C2C16ULL, "genip_trap_vert_fire_jet.utp"},
	{0x04E80F172C3980F1ULL, "330757309.ogg"},
	{0x04E8843A7197CC8EULL, "6068880_m.fxe"},
	{0x04E88CFC0763C42DULL, "follower_banter6171544.cl"},
	{0x04E8DF1C47BCACDEULL, "6120951_m.fxe"},
	{0x04E8E566193B9896ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_basic.mao"},
	{0x04E8F166193BAD35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_basic.mmh"},
	{0x04E8F366193BB0ABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_basic.msh"},
	{0x04E914549FECF0C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel.gad"},
	{0x04E9435207723503ULL, "29839064.ogg"},
	{0x04E95CA557434452ULL, "gen_im_npc_wep_arc_lbwb.uti"},
	{0x04E99EC095361849ULL, "lt_0_lowtown_docks_a3_n_1_0_lc.mmh"},
	{0x04E9CCD1FD9E1503ULL, "379643227.ogg"},
	{0x04EB80D62DC1FC8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e.gad"},
	{0x04EBA1F6CF300610ULL, "follower_banter6171929.cl"},
	{0x04EBA81683E4AE1BULL, "3295196.ogg"},
	{0x04EC1F87AD4477E8ULL, "lt_0_lowtown_docks_109_0.msh"},
	{0x04EC3187AD449679ULL, "lt_0_lowtown_docks_109_0.mao"},
	{0x04EC3D87AD44AA12ULL, "lt_0_lowtown_docks_109_0.mmh"},
	{0x04EC5360370118D9ULL, "857843093.ogg"},
	{0x04EC54B22E6EC59EULL, "hlac_w_sundermount_-196608_0z.dds"},
	{0x04EC7A3E5A5F67E5ULL, "mcr370_assignments6081719.cl"},
	{0x04ECD46D990750A2ULL, "vxdr_slavers_1ja_0_0.vlm"},
	{0x04ED2BF683869DF1ULL, "t3_hey_am3.tnt"},
	{0x04ED33F10E6DACE6ULL, "857451040.ogg"},
	{0x04ED4D5445BEBF8AULL, "llac_lt_backalley_g_1_5.dds"},
	{0x04ED79599F94C17CULL, "94285862.ogg"},
	{0x04EDCAE83D8B193AULL, "hlac_lt_0_lowtown_a3_n_c_131072_5z.dds"},
	{0x04EE3D40A3D53FECULL, "635891422.ogg"},
	{0x04EEE6822F5B7407ULL, "618963231.ogg"},
	{0x04EFDAB32DC1B5FCULL, "gen_im_trs_gen_hvy_boo_r1.uti"},
	{0x04F00172D6B22E15ULL, "944417318.ogg"},
	{0x04F009549FF2FCD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel.evt"},
	{0x04F01543C6B4DB4EULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.impa.ani"},
	{0x04F01F6402256B7CULL, "llac_ht_chantry_n_2534_13.dds"},
	{0x04F025D5268B5463ULL, "w_bonepitmining_a2_81_0.msh"},
	{0x04F026DAEB1F3352ULL, "hlac_lt_gallowstemplar_2383_3y.dds"},
	{0x04F033D5268B6BBDULL, "w_bonepitmining_a2_81_0.mmh"},
	{0x04F036AB6DB95D65ULL, "hlac_ht_0_hightown_a3_131073_0z.dds"},
	{0x04F037D5268B72CEULL, "w_bonepitmining_a2_81_0.mao"},
	{0x04F07BF6123433A4ULL, "vxw_encounter2_hp_0_0.vlm"},
	{0x04F0E1E5217855CAULL, "framebuffereffects_verticalblurshadows_p0.vsh.fxo"},
	{0x04F142070F468A64ULL, "vxht_chantry_n_36c_0_0.vlm"},
	{0x04F17FD62DC6668AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e.ani"},
	{0x04F1C1205F631719ULL, "rdr111dg_martin6055074.cl"},
	{0x04F1CEBF60312D53ULL, "hlac_w_bonepitdestroyed_a3_196610_0x.dds"},
	{0x04F250912713C035ULL, "seb222ar_ruin.are"},
	{0x04F2A1E27FA2515DULL, "983898147.ogg"},
	{0x04F30176B994BB82ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkr.gad"},
	{0x04F3323D936CB894ULL, "867033515.ogg"},
	{0x04F3CD7CA56E91A6ULL, "vxht_brothel_n_6hu_0_0.vlm"},
	{0x04F3D59FBF98DB38ULL, "889324293.ogg"},
	{0x04F409F1BBE34F90ULL, "622366784.ogg"},
	{0x04F43811949485B2ULL, "6001198_f.fxe"},
	{0x04F4BABE1A924780ULL, "hlac_lt_0_lowtown_docks_a3_-131074_4z.dds"},
	{0x04F4E784A51C777EULL, "6095857_m.fxe"},
	{0x04F59D9B10E4804CULL, "qcr220_meredith6054750.cl"},
	{0x04F625372C2946FBULL, "6146943_m.fxe"},
	{0x04F6DA4ADA7DA505ULL, "vxht_chantry_n_2th_0_1.vlm"},
	{0x04F6DBFD41B5329DULL, "face_rendertodepth_p1.vsh.fxo.dx11"},
	{0x04F72775E1698056ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qm_overlay.anb"},
	{0x04F76D5351D5D8A0ULL, "llac_lt_undercityserialkiller_-1_42.dds"},
	{0x04F79731DE6168EAULL, "6077344_f.fxe"},
	{0x04F7E40CB033FB14ULL, "vxlt_hangedman_2xq_0_0.vlm"},
	{0x04F8598C121EE128ULL, "lgt160_wk_avern_deaddrop.cnv"},
	{0x04F87006E999BB87ULL, "ser110ip_to_foundry.utp"},
	{0x04F967A4E491E003ULL, "6041946_f.fxe"},
	{0x04F977E1C23BC949ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_5z.dds"},
	{0x04F9AD6850B26810ULL, "vxlt_darktown_587_0_0.vlm"},
	{0x04FA15CFF4E0AD6BULL, "hlac_ht_playershouse_n_855_22z.dds"},
	{0x04FA1C21A2A111D1ULL, "6110619_f.fxe"},
	{0x04FA20DD280207ABULL, "6092178_m.fxe"},
	{0x04FA3FB1E7424E88ULL, "6095914_m.fxe"},
	{0x04FAA7F9641D74EFULL, "hlac_lt_gallowscourtyard_-131072_0z.dds"},
	{0x04FAF21A8FB155C5ULL, "follower_banter6161192.cl"},
	{0x04FB0A019AC46BA8ULL, "vxlt_gallowstemplar_n_5k1_0_0.vlm"},
	{0x04FB79DCA18F7768ULL, "dae000_meredith6085275.cl"},
	{0x04FB7B8C12215090ULL, "lgt160_wk_avern_deaddrop.bnk"},
	{0x04FB88C6138A2255ULL, "hlac_lt_0_lowtown_a2_n_c_-65538_0y.dds"},
	{0x04FB99A2A4C2342FULL, "6055001_m.fxe"},
	{0x04FC003F443D576DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightl.mmh"},
	{0x04FC043F443D5DBEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightl.mao"},
	{0x04FC123F443D75F3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightl.msh"},
	{0x04FC7191271C4C9BULL, "seb222ar_ruin.ncs"},
	{0x04FC8B506160B7B4ULL, "2068354.ogg"},
	{0x04FCF576195DB56DULL, "gen_im_acc_blt_act3_03.uti"},
	{0x04FD01549FFD790DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel.ani"},
	{0x04FD54ECA7A58FA5ULL, "follower_banter6155667.cl"},
	{0x04FD69790B5FE795ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_crumble_1.mao"},
	{0x04FD7EFB3B17C5E8ULL, "225279094.ogg"},
	{0x04FDB9540E473A84ULL, "6169819_m.fxe"},
	{0x04FDF5937931EB3FULL, "vxht_0_hightown_a2_c1v_0_0.vlm"},
	{0x04FE0D43C6C10A9FULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.impa.evt"},
	{0x04FF29514B92FACCULL, "hlac_lt_0_lowtown_docks_a3_n_131070_0x.dds"},
	{0x04FF956C55764F7AULL, "vxlt_undercitydungeon_1c2_0_1.vlm"},
	{0x04FFBDD9B2951138ULL, "6052087_m.fxe"},
	{0x0500438903DA1388ULL, "6086389_m.fxe"},
	{0x050065F7A6BE95ECULL, "1004779893.ogg"},
	{0x05009232F8DA3971ULL, "lgt260cr_way_merc_3.utc"},
	{0x0500CFA5ED938BB0ULL, "hlac_ht_store_187_3x.dds"},
	{0x05015F7DCFCE2868ULL, "t3_wpn_wht.tnt"},
	{0x050230C5DEAF28BEULL, "vxca_caverns1_mines_j20_0_2.vlm"},
	{0x05023F599764AA26ULL, "vxht_brothel_n_6uv_0_0.vlm"},
	{0x0502899EA954D0ABULL, "6018576_m.fxe"},
	{0x0502AB688C1ABD68ULL, "hlac_lt_0_lowtown_65534_2z.dds"},
	{0x0502EA5FA59376C2ULL, "868071912.ogg"},
	{0x050330317F1B2AF0ULL, "lt_0_lowtown_docks_a3_11_0.rml"},
	{0x050330D0BBBF7A95ULL, "llac_lt_0_lowtown_a3_n_c_65533_2.dds"},
	{0x0503519127223542ULL, "seb222ar_ruin.lst"},
	{0x05036E27330BEC52ULL, "zz_ave_debug6064144.cl"},
	{0x050396BC9DEC37F0ULL, "ht_brothel_n_3372.anb"},
	{0x0503BF6C7C2DE99EULL, "664560915.ogg"},
	{0x050479AB0BD721E3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2pil.msh"},
	{0x050487AB0BD7393DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2pil.mmh"},
	{0x05048BAB0BD7404EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2pil.mao"},
	{0x0504EE43C6C6F488ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.impa.gad"},
	{0x0504F8815E6178FAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_trip.gad"},
	{0x0505DF8C7AE1B71EULL, "6072986_m.fxe"},
	{0x050613A967E024EEULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsma.phy"},
	{0x050622712961134DULL, "gen00fl_carver6103887.cl"},
	{0x050628A1F19224EBULL, "6073785_f.fxe"},
	{0x05067877EF7CBAFAULL, "201366813.ogg"},
	{0x05067A01A8A306A6ULL, "vxw_sundermount_26c_0_0.vlm"},
	{0x0506810780D7EB50ULL, "hlac_lt_gallowstemplar_f_2391_9y.dds"},
	{0x0506974AD45160B4ULL, "sky_lt_0_lowtown_docks_a3.mao"},
	{0x0506B8A0136E4621ULL, "vxht_mansion2b_27o_0_0.vlm"},
	{0x0506E111A76A1BAAULL, "375618774.ogg"},
	{0x0506FF046F58317AULL, "6148089_m.fxe"},
	{0x05079FC988BA0CA9ULL, "36482741.ogg"},
	{0x05082C76B9A76168ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkr.ani"},
	{0x05087DC15EF85680ULL, "zz_mcr_debug6078868.cl"},
	{0x0508DF7E75358AC4ULL, "follower_banter6164071.cl"},
	{0x0509AA0230F6C608ULL, "867661756.ogg"},
	{0x050A0B7F96B84CAFULL, "6052101_m.fxe"},
	{0x050A0EB614FB7E0BULL, "vxca_caverns2_90a_0_1.vlm"},
	{0x050A7DC5E64D1363ULL, "190988183.ogg"},
	{0x050AB6E1CFBD226AULL, "43561290.ogg"},
	{0x050AFB815E65E911ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_trip.evt"},
	{0x050B38BC4D70D4B3ULL, "llac_ht_0_hightown_a2_65536_3.dds"},
	{0x050C67679EB55468ULL, "439717786.ogg"},
	{0x050CB32056BB992DULL, "llac_ht_keep_4519_1.dds"},
	{0x050CF93F444BBE2AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightl.phy"},
	{0x050D5156EFF8EA29ULL, "gen00fl_sebastian6179326.cl"},
	{0x050D6120043BA67BULL, "w_encounter2_mmi.gff"},
	{0x050D8647A987FD1DULL, "lgt320_mis_resp_swordhunter6150271.cl"},
	{0x050E81B79506800BULL, "gen_im_arm_cht_hvy_act1_01.uti"},
	{0x050E862F8EEB1B93ULL, "llac_lt_0_lowtown_docks_a3_n_0_3.dds"},
	{0x050F1EC253022B44ULL, "dae000_bartrand6206913.cl"},
	{0x050F3490EE880CCEULL, "vxht_bartrand_n_5mm_0_0.vlm"},
	{0x050FC48D769D51A1ULL, "gen00fl_bethany6101490.cl"},
	{0x050FCC2E01ED0761ULL, "vxdr_0_deeproads_gg1_0_0.vlm"},
	{0x0510342093E15155ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_diagonal.phy"},
	{0x05106496E3C8690FULL, "188606680.ogg"},
	{0x051094ECE5CDC5D5ULL, "vxlt_house_f3_0_0.vlm"},
	{0x0510ABE4C1136643ULL, "vxlt_0_lowtown_a2_n_c_hja_0_0.vlm"},
	{0x05111318A313A47BULL, "lt_0_lowtown_docks_a3_158_0.mmh"},
	{0x05111F18A313B818ULL, "lt_0_lowtown_docks_a3_158_0.mao"},
	{0x05113118A313D765ULL, "lt_0_lowtown_docks_a3_158_0.msh"},
	{0x0512726E95C01EC3ULL, "6093560_m.fxe"},
	{0x05127900B22656AAULL, "vxht_brothel_n_5bp_0_0.vlm"},
	{0x0512A7A1D6BA6372ULL, "plt_cod_ltr_junk_starkprinc.nss"},
	{0x0512E08148F4E9DBULL, "follower_banter6173113.cl"},
	{0x0512EBAEDE1D1419ULL, "hlac_ca_cavern1_haunted_24424_12x.dds"},
	{0x0513B0F1B01E2523ULL, "vrc321_varric6096939.cl"},
	{0x0513F7CBF5841579ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potb.phy"},
	{0x0514012E5B59C71FULL, "gen00fl_anders6113084.cl"},
	{0x05140D375E241940ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_0z.dds"},
	{0x05147ACA6ED99ED3ULL, "dae221ar_player_home.ncs"},
	{0x0514CCA761A02869ULL, "gen00fl_merrill6145142.cl"},
	{0x05151E26E81D0024ULL, "vxca_caverns2_657_0_0.vlm"},
	{0x051520E666F2250AULL, "6177331_m.fxe"},
	{0x051544FDB387D761ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_lowtownjimmy_t.dds"},
	{0x0515A67E93B793D1ULL, "6084712_f.fxe"},
	{0x05161776FDCF5799ULL, "w_blightlands_87_0.phy"},
	{0x05163C0104A26DAEULL, "darktown_fem_fereldan_sta6152806.cl"},
	{0x05169DD3C775CA9EULL, "hlac_lt_0_lowtown_-262143_0z.dds"},
	{0x051745B9F06E3BC1ULL, "153616720.ogg"},
	{0x051804244F294A5EULL, "6131162_m.fxe"},
	{0x0518A54A893F611BULL, "mag130_danzig6018810.cl"},
	{0x0518E067F74C14F8ULL, "sky_lt_hangedman.mao"},
	{0x0518E8F8CBD55240ULL, "vxht_mansion2_27j_0_0.vlm"},
	{0x051913815E724FE0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_trip.ani"},
	{0x05195266E56BB72BULL, "vxht_0_hightown_b8l_0_0.vlm"},
	{0x05195BFB92F786D1ULL, "hm_guard_2.mor"},
	{0x0519956DFD49898CULL, "llac_lt_undercityencounter_-1_31.dds"},
	{0x0519EDC3CA4529C8ULL, "dae000_sandal6147765.cl"},
	{0x051ABA07DBF649A9ULL, "6099321_f.fxe"},
	{0x051B085899395386ULL, "6118390_m.fxe"},
	{0x051B1ACFC244924AULL, "6041687_m.fxe"},
	{0x051B43E4639AB8CAULL, "lt_0_lowtown_docks_19_0_lc.mmh"},
	{0x051BAC557760E3C2ULL, "681964974.ogg"},
	{0x051BFD46FA5BE9C2ULL, "dre000_hireling26038687.cl"},
	{0x051C43029FCAA4E1ULL, "vxht_mansion2_1pl_0_1.vlm"},
	{0x051C8F63D87CF60FULL, "mag170_thrask6054021.cl"},
	{0x051DA52181A0B1F9ULL, "w_bonepitdestroyed_a3_201_0.msh"},
	{0x051DAF2181A0C2E7ULL, "w_bonepitdestroyed_a3_201_0.mmh"},
	{0x051DB32181A0C934ULL, "w_bonepitdestroyed_a3_201_0.mao"},
	{0x051DDC274CAFB4A0ULL, "mag101_aombanter6052745.cl"},
	{0x051E4C1B3CA51F61ULL, "llac_w_encounter2_327681_0.dds"},
	{0x051FB3D25AD4AA76ULL, "nat380_temmerin6120948.cl"},
	{0x05201E28A865299CULL, "107776671.ogg"},
	{0x05202B0EFFFE1603ULL, "mcr370_samson6087532.cl"},
	{0x0520B881D3F6F2DBULL, "w_encounter2_camp_52_0.mmh"},
	{0x0520BF8C20961D98ULL, "312186403.ogg"},
	{0x0520C481D3F70778ULL, "w_encounter2_camp_52_0.mao"},
	{0x0520D681D3F725C5ULL, "w_encounter2_camp_52_0.msh"},
	{0x0522423A339E73EDULL, "hlac_lt_0_lowtown_docks_a3_n_262142_0x.dds"},
	{0x05224C9D1D2BD38DULL, "vxht_mansion2_31x_0_0.vlm"},
	{0x05226C462465CC1EULL, "104078808.ogg"},
	{0x0522A038EAC5AE79ULL, "zz_mer_debug6189048.cl"},
	{0x0522A1D403D468C3ULL, "6076554_m.fxe"},
	{0x0522ABCA64EE4980ULL, "6117814_m.fxe"},
	{0x0523B67056D5CF8CULL, "w_blightlands_51_0.rml"},
	{0x0523CCBF1CD2B898ULL, "notifications.gfx"},
	{0x05244D1651744CBCULL, "gen00fl_varric6093624.cl"},
	{0x052456EE82B78C4EULL, "6177853_f.fxe"},
	{0x0524C3CFBD282CEAULL, "6037624_m.fxe"},
	{0x0525EA114090F714ULL, "lowtown_male_urchin_sta6164208.cl"},
	{0x0526AE08F1D3A7F5ULL, "lt_warehouse_3_0.phy"},
	{0x05275AD5BDCB6AC0ULL, "vxht_mansion2_20k_0_0.vlm"},
	{0x0527790C1B46E516ULL, "1003392060.ogg"},
	{0x0528728B3AF0D5DFULL, "6105448_f.fxe"},
	{0x0528C2055F2EF4C5ULL, "vxlt_warehouse_n_ed_0_0.vlm"},
	{0x052997C629753E20ULL, "vxw_bonepitmining_a2_28n_0_0.vlm"},
	{0x0529A6A1DDC1D0B1ULL, "6106270-mrlact1_pt3_fire.cut"},
	{0x052A6177EEE00AF9ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_hed_cb.mao"},
	{0x052A7D27436C02BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_twohanded.evt"},
	{0x052B04AE70D3064DULL, "6083703_m.fxe"},
	{0x052B74B790BC5245ULL, "677076202.ogg"},
	{0x052B9AE9F5BE9E2AULL, "6129725_m.fxe"},
	{0x052BC5BDF7FAA2A8ULL, "vxlt_0_lowtown_docks_n_bdy_0_0.vlm"},
	{0x052C37F4C0321FC4ULL, "mag100_cullen6096070.cl"},
	{0x052C59B11E391175ULL, "isa211_isabela6197506.cl"},
	{0x052CCDC0B307E95EULL, "84352073.ogg"},
	{0x052CEB013E43FF06ULL, "vxlt_blackemporium_o1_0_1.vlm"},
	{0x052D57DCA4E2D666ULL, "hlac_lt_0_lowtown_docks_262142_0y.dds"},
	{0x052DAA7CAAE3F541ULL, "vxca_cavern1_haunted_ckg_0_0.vlm"},
	{0x052DB379F3F74066ULL, "vxlt_0_lowtown_a2_n_fim_0_0.vlm"},
	{0x052DC190BE302824ULL, "lt_0_lowtown_docks_a3_230_0.msh"},
	{0x052DCF08F1D9FE21ULL, "lt_warehouse_3_0.rml"},
	{0x052DD390BE30468DULL, "lt_0_lowtown_docks_a3_230_0.mao"},
	{0x052DD790BE304DC6ULL, "lt_0_lowtown_docks_a3_230_0.mmh"},
	{0x052E1ACF6D69AEEBULL, "dae000_mother6004120.cl"},
	{0x052E5AFB1CBD7253ULL, "plt_lgt000pt_baz_golemite.plo"},
	{0x052E698F6A856826ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison.evt"},
	{0x052EF178A6AD6000ULL, "vxht_mansion2b_346_0_0.vlm"},
	{0x052F0892341868FDULL, "vxht_0_hightown_api_0_0.vlm"},
	{0x052F2F03514AEF31ULL, "vxht_0_hightown_a3_c0z_0_0.vlm"},
	{0x052F31317F406758ULL, "lt_0_lowtown_docks_a3_11_0.anb"},
	{0x052F5CBA961287AAULL, "vxht_brothel_n_5hr_0_0.vlm"},
	{0x052FB92D739A7509ULL, "952357242.ogg"},
	{0x052FC47455F4BA7FULL, "vxlt_0_lowtown_docks_n_coz_0_0.vlm"},
	{0x05307627437062FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_twohanded.gad"},
	{0x05319279F0930C4BULL, "vxw_woundedcoast_q1_0_0.vlm"},
	{0x053228953F07EAE6ULL, "6072339_m.fxe"},
	{0x053229365E22EB2CULL, "w_bonepit_97_0.msh"},
	{0x05322F365E22F51EULL, "w_bonepit_97_0.mmh"},
	{0x05323B365E2309B5ULL, "w_bonepit_97_0.mao"},
	{0x05328479E6C321DDULL, "6137066_m.fxe"},
	{0x053349478AA30CD1ULL, "vxht_mansion_n_4y4_0_0.vlm"},
	{0x05337E797411E9A1ULL, "hlac_lt_gallowscourtyard_65534_2z.dds"},
	{0x0533B0BAC953E1EDULL, "vxw_woundedcoast_1cl_0_0.vlm"},
	{0x0533B2F31F8F8B6DULL, "gencr_high_dragon.utc"},
	{0x0533BBEE27CD8E9CULL, "6179007_m.fxe"},
	{0x0533D382C842E35AULL, "6102717_m.fxe"},
	{0x05342F8D9B9F7AC9ULL, "250953656.ogg"},
	{0x05346C1CE5DE4076ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mh_m.mov_rf_leanr.evt"},
	{0x0534808F6A89FB71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison.gad"},
	{0x0534A23FB83931F1ULL, "vxht_mansion2b_2o7_0_0.vlm"},
	{0x0534C2828CAD4F5DULL, "vfx_vfx_p2.vsh.fxo"},
	{0x0535A374B72F33BFULL, "592159521.ogg"},
	{0x0535AAB5214C0BCEULL, "w_encounter3_249_0.mmh"},
	{0x0535ADF8FCD6AF22ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medb_0t.dds"},
	{0x0535B6B5214C2065ULL, "w_encounter3_249_0.mao"},
	{0x0535C4B5214C37BCULL, "w_encounter3_249_0.msh"},
	{0x0535D339BDDF2FA9ULL, "6051263_m.fxe"},
	{0x0535F443968DFADDULL, "hlac_ca_cavern1_haunted_24424_18z.dds"},
	{0x0535F5557E54F7A2ULL, "904977543.ogg"},
	{0x0536182B09389167ULL, "fen110_fenris6096425.cl"},
	{0x053621A9E917CC3BULL, "character_character_p51.psh.fxo.dx11.assembly"},
	{0x0536223B9FF1E190ULL, "gen00fl_anders6119119.cl"},
	{0x053650D49CC68A0CULL, "gen00fl_merrill6127224.cl"},
	{0x0536CA2FB93ABF15ULL, "vxht_0_hightown_a3_a71_0_0.vlm"},
	{0x0536CA4BD33B0A0CULL, "6081788_m.fxe"},
	{0x0537752743767FFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_twohanded.ani"},
	{0x053791D52292C77AULL, "vxlt_warehouse_n_yk_0_0.vlm"},
	{0x0537BD9B05869E90ULL, "738421465.ogg"},
	{0x053805818BB3BA55ULL, "hlac_lt_blackemporium_1055_12x.dds"},
	{0x05386EC97B90A0E7ULL, "vxht_mansion2_3ev_0_0.vlm"},
	{0x05391390605C2BDFULL, "hlac_lt_darktown_-1_37y.dds"},
	{0x05397545A24FD42EULL, "vxlt_0_lowtown_n_e2s_0_0.vlm"},
	{0x0539B12A7977850BULL, "seb201_lady_harimann.bnk"},
	{0x0539C2BD21C08E62ULL, "hla_ht_0_hightown_-65537_0y.dds"},
	{0x053A0502005FB8CAULL, "295059736.ogg"},
	{0x053A379AE9F0BB28ULL, "vxlt_0_lowtown_docks_n_bav_0_0.vlm"},
	{0x053A6D8D718CAAF3ULL, "vxlt_gallowscourtyard_6yz_0_0.vlm"},
	{0x053A769B217458AAULL, "vxlt_0_lowtown_docks_b4d_0_0.vlm"},
	{0x053A85119A840098ULL, "6134826_m.fxe"},
	{0x053AA0A1E7A49E15ULL, "hlac_lt_gallowstemplar_f_2388_3x.dds"},
	{0x053AB62F772F8F5DULL, "222187236.ogg"},
	{0x053B6F9D6DFAF881ULL, "vxlt_0_lowtown_docks_a3_n_c42_0_0.vlm"},
	{0x053B818F6A901B1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison.ani"},
	{0x053C241180BEC1B4ULL, "static_billboard_p0.psh.fxo.dx11.assembly"},
	{0x053C3F8983C61E7AULL, "vxlt_darktown_4uc_0_0.vlm"},
	{0x053CB5A157AA6186ULL, "art\\effects\\cloudshadows\\w_encounter2\\cloud1.dds"},
	{0x053CD360D87FAD8DULL, "vxlt_warehouse_qi_0_0.vlm"},
	{0x053CFD418BEAB09AULL, "gen_im_trs_dal_blt_r0.uti"},
	{0x053D7584077C4FCAULL, "1432136.ogg"},
	{0x053D83C4D26A4DFDULL, "vxca_caverns2_9fl_0_0.vlm"},
	{0x053D9BE026178E75ULL, "mag291cr_caress.utc"},
	{0x053DA0E0434107B1ULL, "vxht_0_hightown_a3_c5y_0_0.vlm"},
	{0x053DAB469F34F852ULL, "6102076_m.fxe"},
	{0x053DB32A797B71EBULL, "seb201_lady_harimann.cnv"},
	{0x053DDCD74A9A5E99ULL, "6127889_m.fxe"},
	{0x053E6C3725E9F3C7ULL, "follower_banter6166458.cl"},
	{0x053EBDE847FB6852ULL, "narrative_banter6149661.cl"},
	{0x053EE8496B9D25A8ULL, "hlac_lt_0_lowtown_a2_n_c_65533_2z.dds"},
	{0x053F5DDA28A726FDULL, "t1_mub_bk2.tnt"},
	{0x053F6BEB9EACAF79ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.dura.ani"},
	{0x053F9F0B1D816BA5ULL, "561664353.ogg"},
	{0x05405CCAFC452C21ULL, "316572553.ogg"},
	{0x0540F8E87295D83EULL, "vxht_mansion_n_3yh_0_0.vlm"},
	{0x054113CEDB2132E7ULL, "264075058.ogg"},
	{0x0541802354001436ULL, "qun211_braggart6075704.cl"},
	{0x054191940F339DE1ULL, "fx_wavering_dis.mao"},
	{0x0541C51A4372113FULL, "105756324.ogg"},
	{0x05421B633139F94FULL, "dae300_bran6143503.cl"},
	{0x05422297BB49FA6DULL, "vxw_woundedcoast_1nk_0_0.vlm"},
	{0x0542253B0873578AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_surrounding_male.gad"},
	{0x05429487F767F8C0ULL, "artcharactersplayercharacterhumanmaleanimationblendtreehumanoid_damage_reactions.anb"},
	{0x0542C7BD46BED4B2ULL, "vxlt_0_lowtown_a3_n_c_gik_0_0.vlm"},
	{0x054325F71B256AAEULL, "pro000_varric6170652.cl"},
	{0x05435C5798E059C9ULL, "lt_0_lowtown_docks_probe_bw7.dds"},
	{0x054379EF77A24DA4ULL, "18332247.ogg"},
	{0x0544019165092E33ULL, "vxlt_gallowscourtyard_3t8_0_0.vlm"},
	{0x054478B54A7BAD68ULL, "vxca_caverns1_mines_gkn_0_0.vlm"},
	{0x0544E3FEB4192957ULL, "gen_im_acc_amu_act3_08.uti"},
	{0x0545025AC3ABEA35ULL, "hlac_ca_cavern1_haunted_24427_2z.dds"},
	{0x0545466D6F9F8375ULL, "vxht_bartrand_n_33z_0_0.vlm"},
	{0x0545A5BF431B25C5ULL, "vxlt_warehouse_n_ob_0_0.vlm"},
	{0x0545BA010DF5E4B2ULL, "6112962_f.fxe"},
	{0x0546370D25119506ULL, "36593040.ogg"},
	{0x054646EB9EB28EBBULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.dura.gad"},
	{0x05466CFECD7571B3ULL, "1058832196.ogg"},
	{0x054709F4FC5F8785ULL, "hlac_lt_0_lowtown_docks_n_-65538_7y.dds"},
	{0x054712A96817B9B1ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsma.mmh"},
	{0x054724A9C98CC492ULL, "6108067_m.fxe"},
	{0x05472CA96817E66FULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsma.msh"},
	{0x05477A5CD6F14C2CULL, "crt_pro_02.mao"},
	{0x05478CC10049EEF8ULL, "6078991_m.fxe"},
	{0x05486030D976BFF1ULL, "6091883_m.fxe"},
	{0x0549083B08794521ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_surrounding_male.evt"},
	{0x054921602331816CULL, "6106401_m.fxe"},
	{0x054932D199E31A1CULL, "vxca_caverns1_e7u_0_0.vlm"},
	{0x05493A22B8C60BFAULL, "prm000im_seeker_shield.uti"},
	{0x05495FE2E55A1BA5ULL, "6055164_m.fxe"},
	{0x0549691710576C13ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkright.ani"},
	{0x0549982181C5D61CULL, "w_bonepitdestroyed_a3_201_0.phy"},
	{0x0549DA9C906A299EULL, "382166365.ogg"},
	{0x054A499CDED1F9CDULL, "vxca_cavern1_haunted_k2p_0_0.vlm"},
	{0x054A725CE57979BEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_wood_n.dds"},
	{0x054A735888EBC843ULL, "vxlt_0_lowtown_a2_n_c_hve_0_0.vlm"},
	{0x054AB0F7AAB1E60DULL, "997244204.ogg"},
	{0x054AFD1EC780D426ULL, "mag291_vincento6027669.cl"},
	{0x054B453DCA75C60DULL, "vxlt_warehouse_zh_0_0.vlm"},
	{0x054B52A1F3C11B64ULL, "hlac_ht_mansion2b_818_46z.dds"},
	{0x054C05C91597B415ULL, "art\\characters\\weapons\\mauls\\w_mal_ml104.mao"},
	{0x054C886171D62300ULL, "gen00fl_aveline6139198.cl"},
	{0x054CF8CB3FF6E274ULL, "hlac_lt_warehouse_65536_3y.dds"},
	{0x054D6302946C8BF0ULL, "vxdr_0_deeproads_cuc_0_1.vlm"},
	{0x054D63EB9EB8DEA0ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.dura.evt"},
	{0x054D900B00545B9EULL, "prp_str_dbone_d.dds"},
	{0x054E1F01320CDD9EULL, "177801756.ogg"},
	{0x054E5740219C4E19ULL, "lgt260_eau_tmp_dockworker6143602.cl"},
	{0x054F01BF19DC4089ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he4a.phy"},
	{0x054F40E84A4ECA23ULL, "zz_dre_debug6056985.cl"},
	{0x054F897BA5D86AFCULL, "vxlt_darktown_60x_0_0.vlm"},
	{0x054FA401C9CCE7B8ULL, "hlac_lt_0_lowtown_-196608_2x.dds"},
	{0x054FFC0592E3FE29ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\fx_deeproads.mao"},
	{0x055014E5F43547B5ULL, "hf_warden.mor"},
	{0x05509474AD4012EDULL, "vxw_woundedcoast_1ij_0_0.vlm"},
	{0x0550A3851FBF9175ULL, "ht_keep_n.arl"},
	{0x0550DFA252B1C628ULL, "6067779_m.fxe"},
	{0x05514D5CD0632951ULL, "vxht_0_hightown_bxn_0_0.vlm"},
	{0x055151A357A6E737ULL, "dae000cr_commoner_d3.utc"},
	{0x055185F99C2562F1ULL, "vxht_mansion2b_2e5_0_0.vlm"},
	{0x0551C79807D22404ULL, "279932747.ogg"},
	{0x05521198DD9DBAD0ULL, "mrl251ip_loot_f.utp"},
	{0x0552FC1942D09CC4ULL, "577147096.ogg"},
	{0x05533A788CBBAA9AULL, "669509920.ogg"},
	{0x05538DC1A97FAFFBULL, "hlac_lt_0_lowtown_a3_n_c_0_0z.dds"},
	{0x0553AD5636016023ULL, "gen_im_trs_gen_one_mac_r0.uti"},
	{0x0553D095B80A18EDULL, "vxlt_0_lowtown_a3_n_c_eki_0_1.vlm"},
	{0x055412788BCE3B07ULL, "584854807.ogg"},
	{0x0554179C35113E45ULL, "vxlt_warehouse_n_ra_0_0.vlm"},
	{0x055447B92E163B48ULL, "vxlt_warehouse_n_b8_0_0.vlm"},
	{0x0554D311324079C7ULL, "gen_im_cns_cft_pot_05.uti"},
	{0x055502A8B41135B0ULL, "one300_zevran6089113.cl"},
	{0x055539FF95672CFBULL, "and111_lirene6084506.cl"},
	{0x0555BB0D0C5AEBB9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_exit.ani"},
	{0x0555FB83454BEA9CULL, "vxlt_backalley_g_1u6_0_0.vlm"},
	{0x0556003B0883C270ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_surrounding_male.ani"},
	{0x055645E4DCD7EAF3ULL, "t3_wdg_dn3.tnt"},
	{0x0556620D84049775ULL, "vxlt_gallowstemplar_n_55r_0_0.vlm"},
	{0x05567426712CD664ULL, "vxlt_0_lowtown_a3_n_c_ep5_0_1.vlm"},
	{0x05572BD0F87FB896ULL, "hlac_ht_0_hightown_a2_65532_0x.dds"},
	{0x0557317B021A5298ULL, "vxht_keep_n_635_0_0.vlm"},
	{0x055733026A189AA7ULL, "t3_nbl_n03.tnt"},
	{0x05573EBE7E5610E2ULL, "6068004_m.fxe"},
	{0x05574720941D7470ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_diagonal.msh"},
	{0x05575520941D8B8AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_diagonal.mmh"},
	{0x055758669A2DCAAAULL, "hlac_dr_slavers_3214_2x.dds"},
	{0x05575920941D92C1ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_diagonal.mao"},
	{0x0557611710639C4AULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkright.evt"},
	{0x05578EF4BF0B1CC1ULL, "llac_lt_gallowscourtyard_131071_2.dds"},
	{0x0557C32BF5DA7D2AULL, "w_bonepitmining_a2_21_0_ll.mmh"},
	{0x0557F2C6771E5C97ULL, "336964561.ogg"},
	{0x0558CB8E73956AEEULL, "character_character_p98.psh.fxo.dx11.assembly"},
	{0x0558E47DEDA88C90ULL, "zz_mcr_debug6071544.cl"},
	{0x055941EAA8E4A37AULL, "hlac_w_bonepitdestroyed_a3_65538_0y.dds"},
	{0x0559EC0CB7D7A54FULL, "zz_approval_debug6133725.cl"},
	{0x0559FBDEAFE070D1ULL, "vxlt_gallowstemplar_5qa_0_0.vlm"},
	{0x055A577E6106E958ULL, "mrl150_dalish_guard6068739.cl"},
	{0x055A64921B5B95DDULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_add.mao"},
	{0x055AA40051FD9FB5ULL, "6167650_m.fxe"},
	{0x055AB22A4BB81440ULL, "185489731.ogg"},
	{0x055C19D602E50B6CULL, "vxca_cavern1_haunted_cny_0_0.vlm"},
	{0x055C451F9F750EABULL, "487173508.ogg"},
	{0x055C72B5DCED6F63ULL, "6057090_m.fxe"},
	{0x055C79D3455FB542ULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_f_c.mmh"},
	{0x055C8FE8BDE5769CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage2_.ani"},
	{0x055C960D0C60CBFBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_exit.gad"},
	{0x055D235911AAC54EULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.cess.evt"},
	{0x055D3154D2129A83ULL, "6149784_m.fxe"},
	{0x055D3256AE928331ULL, "qun110_petrice6042299.cl"},
	{0x055DCC4F136A855CULL, "hf_ave1_banditf.mor"},
	{0x055E4017106982BDULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkright.gad"},
	{0x055E41FB1CE66BF0ULL, "plt_lgt000pt_baz_golemite.nss"},
	{0x055E5489C9A7BE30ULL, "ave121_brennan6105814.cl"},
	{0x055F1C8654EE854CULL, "636098286.ogg"},
	{0x055FB4A61FBD0ED1ULL, "hlac_w_encounter2_-131071_0y.dds"},
	{0x05600E8579A3819EULL, "325396927.ogg"},
	{0x05605017DB9E7951ULL, "620009819.ogg"},
	{0x0560F814F6B6C8AFULL, "135094375.ogg"},
	{0x0560FDD42301056BULL, "2381692.ogg"},
	{0x056120BD771CB6E1ULL, "6140244_m.fxe"},
	{0x056137D57E9D0BCAULL, "hlac_lt_gallowstemplar_f_2382_3y.dds"},
	{0x05618DB5EA4D22C2ULL, "34285563.ogg"},
	{0x0561A1C00810A608ULL, "gen00fl_anders6142537.cl"},
	{0x05622BD13CB08DB5ULL, "6088863_m.fxe"},
	{0x0562B30D0C6568E0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_exit.evt"},
	{0x0562D85925EA6D15ULL, "gen_im_trs_dwv_med_cht_r1.uti"},
	{0x05633C02A12F235EULL, "vxw_encounter2_ea_0_0.vlm"},
	{0x05639AFA694D867FULL, "vxlt_0_lowtown_a2_n_gpn_0_0.vlm"},
	{0x05639CE8BDEBABDEULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage2_.gad"},
	{0x05641A5911B0D4F9ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.cess.gad"},
	{0x056432051C120CD9ULL, "513439048.ogg"},
	{0x05646D603742031CULL, "vxlt_backalley_g_1r7_0_0.vlm"},
	{0x0564BD5FF2CA7746ULL, "vxw_woundedcoast_cb_0_0.vlm"},
	{0x0564CB31C5719E31ULL, "qun131_party_comment6117262.cl"},
	{0x056509EC7FFDA377ULL, "mrl150_merrill6078328.cl"},
	{0x05659275B35E38EEULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfood_0.pwk"},
	{0x056631AC3FD34935ULL, "155133054.ogg"},
	{0x0566552066BF3511ULL, "6067187_m.fxe"},
	{0x05666A715B77A768ULL, "563503190.ogg"},
	{0x056691F53EA53606ULL, "6196688_m.fxe"},
	{0x0566C647A8BE1469ULL, "6202218_m.fxe"},
	{0x056703017E369307ULL, "qcr000_arishok6086005.cl"},
	{0x056733BA9F4BCD5EULL, "6173793_m.fxe"},
	{0x05678AECD24583F2ULL, "vxlt_0_lowtown_docks_n_c06_0_0.vlm"},
	{0x056792D3AE0BB0D6ULL, "vxlt_backalley_g_1wf_0_0.vlm"},
	{0x05689AE8D481CED2ULL, "606976967.ogg"},
	{0x0568D981D434E010ULL, "w_encounter2_camp_52_0.phy"},
	{0x056981D16F7FD9DAULL, "702117156.ogg"},
	{0x05698DC9C7597950ULL, "vxlt_0_lowtown_fj1_0_0.vlm"},
	{0x05699D9A172F7062ULL, "vxlt_hangedman_n_2fr_0_0.vlm"},
	{0x056A049040B982FEULL, "289088724.ogg"},
	{0x056A2B5911B55D47ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.cess.ani"},
	{0x056A77BB055B3EDAULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_c.impa.gad"},
	{0x056AFCD07EACC212ULL, "mag210ip_store_arthuris.utp"},
	{0x056B0478502CE7F1ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_lowtownjimmy_d.dds"},
	{0x056B23CEFBFEAC6AULL, "hm_one_mekel.mor"},
	{0x056B5C006B895CC1ULL, "6174489_m.fxe"},
	{0x056B89ADF7A41B6DULL, "hlac_lt_0_lowtown_n_65534_6z.dds"},
	{0x056B8F2FA41963B5ULL, "hlac_w_bonepitdestroyed_a3_262144_0x.dds"},
	{0x056BBB1E47981578ULL, "hlac_w_sundermount_-327681_0y.dds"},
	{0x056C123037557571ULL, "qun211_braggart6075680.cl"},
	{0x056C66B8452938B4ULL, "vxht_0_hightown_a3_amv_0_0.vlm"},
	{0x056C783579D92374ULL, "llac_w_bonepit_262149_0.dds"},
	{0x056CEC3931CC073FULL, "gen00fl_anders6108509.cl"},
	{0x056D00275E394102ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb107a_0s.dds"},
	{0x056D175DCA76FF4CULL, "166447101.ogg"},
	{0x056DC108F20FCA10ULL, "lt_warehouse_3_0.msh"},
	{0x056DCAB696C6024AULL, "6217088_m.fxe"},
	{0x056DCF08F20FE22AULL, "lt_warehouse_3_0.mmh"},
	{0x056DD308F20FE961ULL, "lt_warehouse_3_0.mao"},
	{0x056E7AE64BD2D11BULL, "vxlt_gallowstemplar_n_51x_0_0.vlm"},
	{0x056E97B0B78C9F0BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_loop.ani"},
	{0x056F0885DA4F521AULL, "6129685_f.fxe"},
	{0x056F0C6A483D70C3ULL, "139298434.ogg"},
	{0x056F6621B744BE92ULL, "gen00fl_aveline6120954.cl"},
	{0x056F701A9AB5B7F0ULL, "6084695_m.fxe"},
	{0x056FB0B7542FE9FEULL, "vxlt_0_lowtown_a3_n_c_cmo_0_0.vlm"},
	{0x056FB46E7B17E59BULL, "hlac_ht_0_hightown_a3_65536_1y.dds"},
	{0x056FB881D43AC768ULL, "w_encounter2_camp_52_0.rml"},
	{0x057084635EE1A54BULL, "follower_banter6165617.cl"},
	{0x0570B44F9BF649EDULL, "vxlt_0_lowtown_a3_n_c_eqg_0_1.vlm"},
	{0x05716622B4E68FD7ULL, "gen00fl_anders6202353.cl"},
	{0x05717ABB056162F1ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_c.impa.evt"},
	{0x0571B1DDB7A8511CULL, "vxw_bonepitmining_a2_1hy_0_0.vlm"},
	{0x05720CD75B439EFFULL, "vxlt_0_lowtown_a2_n_gmi_0_0.vlm"},
	{0x05722BB9DE463EE1ULL, "qun111_ketojan6040955.cl"},
	{0x0572B19DB84B512FULL, "6055276_m.fxe"},
	{0x0572D92C533A42DFULL, "6153884_m.fxe"},
	{0x0572FCA36C6CE288ULL, "vxlt_warehouse_11i_0_0.vlm"},
	{0x0573FFB24F7A503DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masc_0s.dds"},
	{0x0574577B6452EB88ULL, "681305092.ogg"},
	{0x05747AA4F6C91263ULL, "778003310.ogg"},
	{0x0574F5FEC7DB72AEULL, "dae321ar_chantry.ncs"},
	{0x0575154EAF15AB14ULL, "follower_banter6156203.cl"},
	{0x05752827A7E95EFFULL, "6102705_m.fxe"},
	{0x0575567457F8E0DAULL, "ts_coterie_arm_props.gda"},
	{0x057635EDD87C5E12ULL, "vxlt_gallowstemplar_37j_0_0.vlm"},
	{0x0576FDD1D883B0C0ULL, "860479082.ogg"},
	{0x0577DBE1D121BDBBULL, "vxht_0_hightown_a3_90b_0_0.vlm"},
	{0x05780F77092588E2ULL, "vxlt_hangedman_n_2ms_0_0.vlm"},
	{0x05783098FA0A6531ULL, "781775349.ogg"},
	{0x05783300D907E29CULL, "steam_01.mao"},
	{0x0578D617548CE18DULL, "hla_ht_0_hightown_a2_-65537_0z.dds"},
	{0x0578F5A59573FFCAULL, "vxlt_0_lowtown_docks_a3_n_aon_0_0.vlm"},
	{0x057940A53EC212DDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_rob_magb.mao"},
	{0x05794291C9B386DAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rb_basic.mmh"},
	{0x05794691C9B38D11ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rb_basic.mao"},
	{0x05795491C9B3A520ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rb_basic.msh"},
	{0x05797E18E4E39D11ULL, "148407676.ogg"},
	{0x0579C9AA87C48DABULL, "6217966_m.fxe"},
	{0x0579FB665A34616EULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.impa.evt"},
	{0x057A070151574445ULL, "gen00fl_fenris6126761.cl"},
	{0x057AC14764F9C756ULL, "hlac_ht_0_hightown_a2_65535_1z.dds"},
	{0x057AD895371F5134ULL, "vxht_0_hightown_a3_ajw_0_0.vlm"},
	{0x057AEABF1A0154AEULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he4a.mmh"},
	{0x057B04BF1A01809CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he4a.msh"},
	{0x057B146A15558868ULL, "llac_dr_primevalentrance_1726_11.dds"},
	{0x057B60029394552BULL, "6095485_m.fxe"},
	{0x057BA0001BABA076ULL, "vxlt_foundry_n_vt_0_0.vlm"},
	{0x057BEDFEC7E18387ULL, "dae321ar_chantry.lst"},
	{0x057C16A25511D115ULL, "vxht_0_hightown_a3_byt_0_0.vlm"},
	{0x057C2D01CA6FFFCAULL, "composedof.psh.fxo"},
	{0x057C7FB0B798B382ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_loop.evt"},
	{0x057C7FB20C4EC3C3ULL, "vxca_caverns2prc_a9r_0_0.vlm"},
	{0x057CCBB7DF930859ULL, "6132713_m.fxe"},
	{0x057CDAD41139B57DULL, "hlac_dr_primevaltomb_658_2z.dds"},
	{0x057D15334464F084ULL, "6133140_m.fxe"},
	{0x057D9136D1718924ULL, "t1_mue_bl1.tnt"},
	{0x057E4A99E2C94595ULL, "168434962.ogg"},
	{0x057E92BB056C15C0ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_c.impa.ani"},
	{0x057E95F0A52F2552ULL, "qcr000_arishok6060183.cl"},
	{0x057EBBBCCE4E8291ULL, "follower_banter6161013.cl"},
	{0x057EFE4F543385D8ULL, "gen00fl_dog6165777.cl"},
	{0x057F303A5B8CB5D4ULL, "one220_yevhen6200154.cl"},
	{0x057FBED894A43EE4ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingb_d.dds"},
	{0x057FF2665A38BD19ULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.impa.gad"},
	{0x05801FBC851B545EULL, "vxw_encounter2_kc_0_0.vlm"},
	{0x05806E6017A0B201ULL, "dae000_arishok6067730.cl"},
	{0x0580A2A746E8EBA1ULL, "102207188.ogg"},
	{0x05816E805E62FB08ULL, "vxlt_warehouse_14f_0_0.vlm"},
	{0x0581A119D6B6A846ULL, "vxw_woundedcoast_ud_0_0.vlm"},
	{0x05821FDEA367FA3EULL, "6049654_m.fxe"},
	{0x05824A24E65AFD30ULL, "dae000_arishok6072890.cl"},
	{0x0582D32E6C41382AULL, "mag110_arianni6015761.cl"},
	{0x05834F49E3E480C0ULL, "758310862.ogg"},
	{0x05837EB0B79ED095ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_loop.gad"},
	{0x058394ECDCA42931ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.anb"},
	{0x05839A0499DF777FULL, "6107805_m.fxe"},
	{0x05839C38EF5D4111ULL, "6166208_m.fxe"},
	{0x05839F03F46A4586ULL, "hightown_male_noble_gossip6171431.cl"},
	{0x05843EAA7BAB2E9CULL, "6054874_m.fxe"},
	{0x058441CF813DABDDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07b.phy"},
	{0x0584489D108E5CFDULL, "llac_lt_0_lowtown_a3_n_c_-2_3.dds"},
	{0x05844D44B9E88596ULL, "mag200_feynriel6023889.cl"},
	{0x0584768D91F7E1D6ULL, "vxlt_backalley_g_1yh_0_0.vlm"},
	{0x0585054EEFFE747DULL, "dae000cr_ineria.utc"},
	{0x058578D34582B8E9ULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_f_c.anb"},
	{0x0585B34798409E45ULL, "one300_murder_followers6089902.cl"},
	{0x05860F23C1532B8FULL, "l_env_wal_lowtown_int_tavern_bed.wbd"},
	{0x058612901CF52C64ULL, "hf_zmother_h3s2.mor"},
	{0x0586C7BA323BE370ULL, "777137567.ogg"},
	{0x058703665A3EF967ULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.impa.ani"},
	{0x058770F7BF2E827AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobd_0s.dds"},
	{0x05887AC62A49E326ULL, "vxw_sundermount_2cl_0_0.vlm"},
	{0x05891EC40FFCEEC0ULL, "6044821_f.fxe"},
	{0x058A39A1815B902EULL, "6139917_m.fxe"},
	{0x058A5AF382C8222FULL, "one140cr_charade6166014.cl"},
	{0x058AE5ECA69A43CBULL, "6140286_m.fxe"},
	{0x058B5345DFA95BABULL, "ht_0_hightown_a3_16_0_lc.mmh"},
	{0x058B7DA9BA3C0289ULL, "gen00fl_aveline6104957.cl"},
	{0x058C265275497936ULL, "vxht_0_hightown_a3_a6y_0_0.vlm"},
	{0x058CEEFEC7EFF8BCULL, "dae321ar_chantry.are"},
	{0x058D07EBD97DF876ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_1z.dds"},
	{0x058D37C9BF609A35ULL, "407091354.ogg"},
	{0x058D60C55B430EBCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_booster_d.dds"},
	{0x058E04335A10A16FULL, "637125805.ogg"},
	{0x058E0F2CF5EE9D48ULL, "vxlt_warehouse_n_v4_0_0.vlm"},
	{0x058E5867F852D69DULL, "ts_books_01.gda"},
	{0x058E8FA73BB7C28FULL, "510220805.ogg"},
	{0x058EA9643DF80DAAULL, "follower_banter6155802.cl"},
	{0x058F04DA5A886040ULL, "cod_hst_amell.utp"},
	{0x058F06D0BE6F6633ULL, "626516320.ogg"},
	{0x058F55D8A7F0C824ULL, "01_dust.mao"},
	{0x0590B5ECDCAEEBAAULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.mmh"},
	{0x0590C7328C5B1AF2ULL, "static_static_p3.psh.fxo.dx11.assembly"},
	{0x0590D94B8ECD771AULL, "art\\characters\\playercharacter\\humanmale\\cth_visa.phy"},
	{0x05912C1EBD2B9BD5ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_p_p.anb"},
	{0x0591B50905BC8884ULL, "6163808_m.fxe"},
	{0x0591C5AF6C2AAA89ULL, "30422172.ogg"},
	{0x05929A0801BBB7B9ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobc.msh"},
	{0x0592A40801BBC8A7ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobc.mmh"},
	{0x0592A80801BBCFF4ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobc.mao"},
	{0x059435C912FC06C1ULL, "vxlt_0_lowtown_a2_n_c_e84_0_0.vlm"},
	{0x05944295C41D2887ULL, "dae110ip_at_hh_house.utp"},
	{0x059542CF814C20E2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07b.mmh"},
	{0x059544CF814C2398ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07b.msh"},
	{0x059578820C900A5AULL, "mag110_vincento6016322.cl"},
	{0x0595B086E9DB44C4ULL, "221793484.ogg"},
	{0x0595FE266DC1B164ULL, "142796207.ogg"},
	{0x059693125B647492ULL, "461123575.ogg"},
	{0x0596ECA31C3FFBA6ULL, "vxw_sundermount_2dm_0_0.vlm"},
	{0x0596FE24C01CB4A3ULL, "llac_lt_gallowstemplar_f_2388_1.dds"},
	{0x059722E1969F1E3DULL, "gen00fl_fenris6110823.cl"},
	{0x05974D6400EEB683ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\energyburst.ani"},
	{0x05976C7BE4F5CE0AULL, "hlac_ht_playershouse_n_855_6y.dds"},
	{0x059793E47DFB3A79ULL, "vxca_caverns1_mines_e8l_0_0.vlm"},
	{0x0597AE21E74AA269ULL, "gen00fl_merrill6129881.cl"},
	{0x0597C2C3B58A910DULL, "vxw_encounter3_6x_0_0.vlm"},
	{0x0598742187292003ULL, "vxht_chantry_26v_0_0.vlm"},
	{0x0598BBECA64A7DECULL, "ave331_post_fight6138805.cl"},
	{0x0598C1562640DEFFULL, "95581331.ogg"},
	{0x0598FA5C38FE0215ULL, "vxht_0_hightown_a3_bsr_0_0.vlm"},
	{0x059910EBFBA1BA2DULL, "504325556.ogg"},
	{0x05992E4377A78AF3ULL, "vxht_keep_6y3_0_0.vlm"},
	{0x0599987D8C8CACE0ULL, "288114523.ogg"},
	{0x0599BFFF22C8BE1FULL, "vxw_sundermount_2j2_0_0.vlm"},
	{0x059A2D396DEDC9D3ULL, "6162123_m.fxe"},
	{0x059B0263A9C54066ULL, "llac_w_encounter2_262149_0.dds"},
	{0x059B03FB7C6DC7FBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cooker.msh"},
	{0x059B117EFF5DC657ULL, "vxlt_gallowsprison_n_6ga_0_0.vlm"},
	{0x059B15FB7C6DE646ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cooker.mao"},
	{0x059B21FB7C6DFAE5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cooker.mmh"},
	{0x059C5871F3EC1EB3ULL, "t1_mue_or1.tnt"},
	{0x059C9BF334AFF696ULL, "vxw_blightlands_a6_0_0.vlm"},
	{0x059CF084ABD7CB58ULL, "and100_anders6095954.cl"},
	{0x059D2578DBD67A2EULL, "6068857_m.fxe"},
	{0x059D656AC57AC3C2ULL, "fen210_orana6109765.cl"},
	{0x059DA3A0AE3F63A4ULL, "6065250_m.fxe"},
	{0x059DBFD8AEC47115ULL, "649703626.ogg"},
	{0x059DC51125BD4A92ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal_1.mao"},
	{0x059DEDA6A46A473CULL, "6000798_m.fxe"},
	{0x059E36C3417C89D3ULL, "vxw_bonepitmining_a2_5n_0_0.vlm"},
	{0x059EB2EAC2C4461DULL, "346517612.ogg"},
	{0x059F0DEE669D6C25ULL, "ser270_emeric6113223.cl"},
	{0x059F39B719684276ULL, "llac_w_bonepitmining_a2_327684_0.dds"},
	{0x059F6191C9D48AE5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rb_basic.phy"},
	{0x05A08DC8145B07A2ULL, "vxw_sundermount_317_0_0.vlm"},
	{0x05A0D224176A84A1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_quill_d.dds"},
	{0x05A0D611B953740EULL, "mcr370_negotiation6080170.cl"},
	{0x05A0FB3234D9E3ABULL, "qcr000_arishok6060108.cl"},
	{0x05A13EA1C895999EULL, "summon.ncs"},
	{0x05A177F0533521E8ULL, "hlac_ca_caverns1_mines_24424_19z.dds"},
	{0x05A24A28D3F2E9A0ULL, "gen00fl_anders6111083.cl"},
	{0x05A26584C03E52A4ULL, "6086932_m.fxe"},
	{0x05A29039D63903BAULL, "6078795_m.fxe"},
	{0x05A2DFDFE237C5D4ULL, "art\\levels\\races\\design\\animation\\gasbarrel.inactivetoactive.gad"},
	{0x05A41216699738F3ULL, "hlac_lt_gallowstemplar_f_2395_0y.dds"},
	{0x05A452082FD7F62CULL, "lt_0_lowtown_n_probe_gp0.dds"},
	{0x05A4656400F96A7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\energyburst.evt"},
	{0x05A47DFBE0C03E57ULL, "6016695_m.fxe"},
	{0x05A4BEC57A8774D9ULL, "vxht_keep_63b_0_1.vlm"},
	{0x05A55E800E361426ULL, "vxw_sundermount_2in_0_0.vlm"},
	{0x05A5953316062351ULL, "74674018.ogg"},
	{0x05A6916B31F09C6CULL, "vxlt_0_lowtown_docks_8rn_0_0.vlm"},
	{0x05A6AD3DFD075FD7ULL, "6174810_m.fxe"},
	{0x05A71C536311D4FAULL, "6126698_m.fxe"},
	{0x05A7A020699DA373ULL, "vxht_keep_6b2_0_0.vlm"},
	{0x05A7B77902D12852ULL, "vxw_bonepitmining_a2_5k_0_0.vlm"},
	{0x05A813C46D6218B7ULL, "6137464_m.fxe"},
	{0x05A821F69B9D43A7ULL, "ico_maul_5.dds"},
	{0x05A82F3F062EF89BULL, "zz_qun_debug6028219.cl"},
	{0x05A82FB4C599F171ULL, "6088337_m.fxe"},
	{0x05A843F36C5481F3ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_5x.dds"},
	{0x05A89F165FE3E253ULL, "6167223_m.fxe"},
	{0x05A90A28D8D8EFDCULL, "406088451.ogg"},
	{0x05A91F7E7A1FA40FULL, "6105989_f.fxe"},
	{0x05A9C82C69056018ULL, "hlac_lt_undercityserialkiller_-1_10y.dds"},
	{0x05A9EEDFE23DFDB2ULL, "art\\levels\\races\\design\\animation\\gasbarrel.inactivetoactive.ani"},
	{0x05AA0BE6B5370932ULL, "562094675.ogg"},
	{0x05AA0DF8196EAF3EULL, "hlac_w_bonepitdestroyed_a3_196608_0x.dds"},
	{0x05AA9834DF443B99ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\stealth.ani"},
	{0x05AB0DD026A60F16ULL, "vxw_blightlands_j7_0_0.vlm"},
	{0x05AB4662E4F0C2DBULL, "art\\vfx\\combat\\vfx_m_spi_crh_aoe_p.impa.ani"},
	{0x05AB646400FF876DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\energyburst.gad"},
	{0x05ABCAA9F40181D2ULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_1x.dds"},
	{0x05ABDA0F000AD45CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_action.ani"},
	{0x05AC2808891C2F4FULL, "6206875_f.fxe"},
	{0x05AC3203D23A2969ULL, "6076216_f.fxe"},
	{0x05ACE668626BB4B4ULL, "6039307_m.fxe"},
	{0x05ACF3D071355CEBULL, "797050997.ogg"},
	{0x05AD27F3EEC122A5ULL, "151698944.ogg"},
	{0x05ADBD2201B8E702ULL, "gen_im_trs_crc_lgt_glv_r2.uti"},
	{0x05ADE4F013EF5E67ULL, "vxw_encounter3_bv_0_0.vlm"},
	{0x05AE4115BE26F1EAULL, "667201325.ogg"},
	{0x05AE501FB37E1329ULL, "vxlt_0_lowtown_docks_a3_8pf_0_0.vlm"},
	{0x05AF20AA63AC93A0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglebody_n.dds"},
	{0x05AFCDCDE06B2732ULL, "6087432_m.fxe"},
	{0x05AFE22D9D4095FBULL, "578924389.ogg"},
	{0x05B02FDAB4BF9619ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_scra_0n.dds"},
	{0x05B0625440BF2197ULL, "532048302.ogg"},
	{0x05B0CDCAB711186CULL, "llac_ht_mansion_n_3382_0.dds"},
	{0x05B12D5C967FE398ULL, "vxlt_undercitydungeon_14d_0_0.vlm"},
	{0x05B17334DF4A1BDBULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\stealth.gad"},
	{0x05B185F4D372F23BULL, "mag360_alain6044616.cl"},
	{0x05B1D485F5D5CD95ULL, "425126052.ogg"},
	{0x05B1D79FA6A5F3A2ULL, "6083146_m.fxe"},
	{0x05B1E70F000F559EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_action.gad"},
	{0x05B247A41F3C9C39ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock2.ani"},
	{0x05B2A74F35559BDFULL, "6032198_m.fxe"},
	{0x05B2D7A89D80C243ULL, "pro000_cut2d_voyage6005291.cl"},
	{0x05B2ECD06D68FA3FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmla.msh"},
	{0x05B2F2D06D6904E1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmla.mmh"},
	{0x05B2FED06D691882ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmla.mao"},
	{0x05B34BC466F7579CULL, "266367170.ogg"},
	{0x05B393F1350AE539ULL, "plt_cod_bks_thrask_notes.nss"},
	{0x05B3D05D002C2CA6ULL, "vxw_sundermount_2ro_0_0.vlm"},
	{0x05B3D625CA63B8A2ULL, "276213801.ogg"},
	{0x05B4238F852C9865ULL, "712017955.ogg"},
	{0x05B43121A4975B61ULL, "mag291_torpor6026935.cl"},
	{0x05B4DED6B89732D0ULL, "vxht_mansion_n_5mh_0_0.vlm"},
	{0x05B4E3BB6DE56266ULL, "198921679.ogg"},
	{0x05B4F838330E748CULL, "501098935.ogg"},
	{0x05B6A191B79009F1ULL, "6083237_m.fxe"},
	{0x05B6AABD8CE67E10ULL, "vrc221_varric6175232.cl"},
	{0x05B7587408635C48ULL, "art\\vfx\\combat\\vfx_glb_ele_f_deathblow_c.anb"},
	{0x05B76605DFAAF591ULL, "vxht_mansion2_3cz_0_0.vlm"},
	{0x05B7A928A2FDD87FULL, "hlac_lt_0_lowtown_a2_n_c_-2_1y.dds"},
	{0x05B7C5EB4D4AF2C3ULL, "dae200_bran6094897.cl"},
	{0x05B7F538E349F757ULL, "vxlt_gallowsprison_n_6mc_0_0.vlm"},
	{0x05B89034DF506BC0ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\stealth.evt"},
	{0x05B8C20F001535BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_action.evt"},
	{0x05B922A41F427C7BULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock2.gad"},
	{0x05B92D1EBD4CEF16ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_p_p.mmh"},
	{0x05B92D79154E3FB1ULL, "22607826.ogg"},
	{0x05B92DD369271A91ULL, "vxlt_gallowscourtyard_n_7ev_0_0.vlm"},
	{0x05B94E62E4FD0DD2ULL, "art\\vfx\\combat\\vfx_m_spi_crh_aoe_p.impa.evt"},
	{0x05B97FAD189C2796ULL, "vxw_blightlands_o8_0_0.vlm"},
	{0x05B9B646C5DD6757ULL, "689130196.ogg"},
	{0x05BA139F53F6E571ULL, "gen00fl_isabela6101546.cl"},
	{0x05BA339CB95F7A88ULL, "qcr221_cut_isabelas_return6042635.cl"},
	{0x05BA7AF13510D912ULL, "plt_cod_bks_thrask_notes.plo"},
	{0x05BAB60B4DA98AA4ULL, "vxw_encounter2_2b_0_0.vlm"},
	{0x05BAB930B4768141ULL, "333270957.ogg"},
	{0x05BAFA59093292CFULL, "vxht_mansion2b_3bs_0_0.vlm"},
	{0x05BB18B37B756982ULL, "mer220_hubert6049752.cl"},
	{0x05BBFF19A4ADA2B1ULL, "6108020_m.fxe"},
	{0x05BC3010BD954C7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_loop.ani"},
	{0x05BC68AC969879ADULL, "6108030_f.fxe"},
	{0x05BCBB749922C39CULL, "hlac_w_encounter3_-262146_0y.dds"},
	{0x05BCD8778C287DFBULL, "spy100pt_among_shadows.ncs"},
	{0x05BD0C8A3187CDB4ULL, "hlac_lt_0_lowtown_a3_n_c_65534_0z.dds"},
	{0x05BD17A7581FB67BULL, "llac_lt_gallowstemplar_f_2391_3.dds"},
	{0x05BDB4C0046AEBE8ULL, "vxw_bonepitmining_a2_e1_0_0.vlm"},
	{0x05BE3CA4DD15E192ULL, "6108236_m.fxe"},
	{0x05BE8D0801E0DCDCULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobc.phy"},
	{0x05BEA038CBD23C89ULL, "hlac_lt_gallowstemplar_2388_5y.dds"},
	{0x05BEA61C1C65E7C5ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_loop_c.anb"},
	{0x05BF2955C7255C7FULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc101a_0n.dds"},
	{0x05BF40687732D5DBULL, "vxw_bonepit_247_0_0.vlm"},
	{0x05BF4D0BF1752E8FULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.cess.gad"},
	{0x05BF5789F7A17870ULL, "31463962.ogg"},
	{0x05BFE74875D71C41ULL, "204157844.ogg"},
	{0x05C06D62E5036165ULL, "art\\vfx\\combat\\vfx_m_spi_crh_aoe_p.impa.gad"},
	{0x05C071534D283B14ULL, "art\\vfx\\combat\\vfx_r_var_modal_act_c.mmh"},
	{0x05C0D9B17A10B018ULL, "vxca_cavern1_haunted_hic_0_0.vlm"},
	{0x05C0E53DA9133075ULL, "vxlt_0_lowtown_docks_n_b66_0_0.vlm"},
	{0x05C10461F628C105ULL, "hlac_ht_mansion_n_2995_0x.dds"},
	{0x05C14B7F46608485ULL, "vxlt_0_lowtown_a3_n_c_ffx_0_0.vlm"},
	{0x05C1927B973AE374ULL, "llac_lt_0_lowtown_docks_a3_n_196607_1.dds"},
	{0x05C1D51983D04BCFULL, "6199327_m.fxe"},
	{0x05C1E9C4D9ADA932ULL, "t3_wdg_w14.tnt"},
	{0x05C23D7EB5BD33F5ULL, "6188569_m.fxe"},
	{0x05C24F5B8438BC43ULL, "515519104.ogg"},
	{0x05C250B959F64CDDULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stand_sa.dds"},
	{0x05C2FA7522ED04ABULL, "isa211_isabela6175166.cl"},
	{0x05C2FCC929B3EA64ULL, "6068015_m.fxe"},
	{0x05C30FC947988D06ULL, "vxdr_primevaltomb_35b_0_0.vlm"},
	{0x05C34A1E7390A435ULL, "6162741_m.fxe"},
	{0x05C38EF569E9107FULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvianbroke_0.pwk"},
	{0x05C3BB4F84829346ULL, "6034279_f.fxe"},
	{0x05C4000D46FE05FAULL, "6128498_m.fxe"},
	{0x05C404479733736FULL, "vxht_keep_n_60v_0_1.vlm"},
	{0x05C47A3CC5F43D38ULL, "329491671.ogg"},
	{0x05C483DA4D89D473ULL, "vxht_keep_6h4_0_0.vlm"},
	{0x05C53763D2062A62ULL, "plt_lgt000pt_ultra_1_05.nss"},
	{0x05C53D3E7EC4DF16ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopright.mao"},
	{0x05C5493E7EC4F3B5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopright.mmh"},
	{0x05C54B3E7EC4F72BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopright.msh"},
	{0x05C5620BF179BE3CULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.cess.evt"},
	{0x05C596CBC506C701ULL, "6134975_m.fxe"},
	{0x05C5D7E2D1A10E11ULL, "vxht_mansion2_3fy_0_0.vlm"},
	{0x05C5E4F9615086ADULL, "zz_lgt_debug6109146.cl"},
	{0x05C62923C1893C65ULL, "l_env_wal_lowtown_int_tavern_bed.bnk"},
	{0x05C63BF477967388ULL, "llac_w_sundermount_-65531_0.dds"},
	{0x05C68186A11E27CCULL, "vxw_bonepitmining_a2_8r_0_0.vlm"},
	{0x05C7463E2FC8E882ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml104a_0s.dds"},
	{0x05C79FB05B1D3311ULL, "vxlt_gallowscourtyard_n_7hu_0_0.vlm"},
	{0x05C7A1B54360A4AEULL, "llac_lt_undercityserialkiller_-1_35.dds"},
	{0x05C7B7A31F42DC9CULL, "6177016_m.fxe"},
	{0x05C7D60CE9900601ULL, "mag311_keran6041512.cl"},
	{0x05C7EDC4AAA58B9FULL, "133994738.ogg"},
	{0x05C7EE6343C7A154ULL, "gen00fl_varric6118763.cl"},
	{0x05C7F18A0A924016ULL, "vxw_blightlands_p9_0_0.vlm"},
	{0x05C810DEFB276EB4ULL, "6164481_m.fxe"},
	{0x05C8896B25D957FFULL, "zz_qun_debug6028303.cl"},
	{0x05C8DBFC91CFBD55ULL, "lt_0_lowtown_docks_a3_133_0.phy"},
	{0x05C8DDA7DDC07F5AULL, "vxht_mansion2_2vs_0_0.vlm"},
	{0x05C97B4469E3D7A4ULL, "vxw_blightlands_at_0_0.vlm"},
	{0x05CA3810BDA19772ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_loop.evt"},
	{0x05CA480018FF5709ULL, "zz_dae_debug6049966.cl"},
	{0x05CB142C672EFF37ULL, "hlac_lt_0_lowtown_n_65535_0y.dds"},
	{0x05CBB1A077DB6173ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01a.phy"},
	{0x05CC47C8B3A77DF3ULL, "qun111_ketojan6040975.cl"},
	{0x05CC5B23EB70FFABULL, "6071151_m.fxe"},
	{0x05CD02BA6962B75BULL, "hlac_ht_keep_4519_0y.dds"},
	{0x05CD0CD3A13CF66CULL, "vxlt_0_lowtown_n_bz2_0_0.vlm"},
	{0x05CD4E082395F18BULL, "vxca_caverns1_mines_ihz_0_0.vlm"},
	{0x05CD5D8FE01A5FDDULL, "vxlt_gallowsprison_n_6hz_0_0.vlm"},
	{0x05CD8F56135D60FDULL, "6041415_m.fxe"},
	{0x05CD9281385B44EEULL, "6059772_m.fxe"},
	{0x05CDA445D091E997ULL, "100065249.ogg"},
	{0x05CDCE33E7E242D9ULL, "420848115.ogg"},
	{0x05CE7202B51989CBULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyo_0s.dds"},
	{0x05CF00413B5BBA10ULL, "vxlt_gallowscourtyard_n_79g_0_0.vlm"},
	{0x05CF1B4F4126BCA4ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagleheadsml.mmh"},
	{0x05CF1F4F4126C3F7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagleheadsml.mao"},
	{0x05CF314F4126E246ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagleheadsml.msh"},
	{0x05CF9298B6DC79C1ULL, "abi_w_taunt.wbd"},
	{0x05CFE26800E1C7CDULL, "one100dg_lia6061030.cl"},
	{0x05D00F8A9A4BA58AULL, "593523212.ogg"},
	{0x05D01B2CFA920590ULL, "76747625.ogg"},
	{0x05D0452DD8423AFCULL, "185396028.ogg"},
	{0x05D05710BDA63705ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_loop.gad"},
	{0x05D08382C927B04BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hat_medc_0t.dds"},
	{0x05D0B416E4185DA6ULL, "vxw_sundermount_2xq_0_0.vlm"},
	{0x05D0C18CACAB0BCFULL, "vxlt_gallowstemplar_f_37l_0_0.vlm"},
	{0x05D0CEC7698CAF2CULL, "819899703.ogg"},
	{0x05D0EEC767FE5478ULL, "6133733_m.fxe"},
	{0x05D0F61375028D2EULL, "hlac_ht_brothel_n_3367_1y.dds"},
	{0x05D1114683D5880DULL, "one300_zevran6075746.cl"},
	{0x05D1A41C7342A55CULL, "hlac_w_woundedcoast_327680_0y.dds"},
	{0x05D1B38221FE05F0ULL, "6120411_f.fxe"},
	{0x05D1BA7F5EB4FBE2ULL, "fen200ip_slaver_chest_3.utp"},
	{0x05D27C69137C301DULL, "vxlt_undercityencounter_up_0_0.vlm"},
	{0x05D36A0BF1860955ULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.cess.ani"},
	{0x05D37CF83BDEC6C7ULL, "vxlt_0_lowtown_a2_n_c_hqx_0_0.vlm"},
	{0x05D394242AC81F4BULL, "vxw_bonepit_100_0_0.vlm"},
	{0x05D3ABCFD0E4FECDULL, "6078735-offeriscommendable.cut"},
	{0x05D405DE5C277F5EULL, "vxht_brothel_n_6gt_0_0.vlm"},
	{0x05D490E80D186E8CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd03a.msh"},
	{0x05D496E80D18797EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd03a.mmh"},
	{0x05D55E695B0B4278ULL, "llac_lt_0_lowtown_docks_n_65534_4.dds"},
	{0x05D5D7EF529E278AULL, "ser270_templar6112551.cl"},
	{0x05D5E37D168026F1ULL, "6072691_m.fxe"},
	{0x05D629B7A31AAD14ULL, "vxlt_0_lowtown_docks_cvm_0_0.vlm"},
	{0x05D66163B31C55AEULL, "llac_lt_0_lowtown_docks_a3_-65538_5.dds"},
	{0x05D697C64074F30EULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urva_0s.dds"},
	{0x05D69D6447F782C8ULL, "6149981_m.fxe"},
	{0x05D6DB2F43F5FB80ULL, "483766983.ogg"},
	{0x05D72204D5CB314AULL, "58779280.ogg"},
	{0x05D7DE17169ED4E1ULL, "6113756_m.fxe"},
	{0x05D8ACE7A891F1FFULL, "254870243.ogg"},
	{0x05D9558F4E5082BDULL, "vxht_bartrand_n_5nw_0_0.vlm"},
	{0x05D9D495694F17FEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv1.evt"},
	{0x05D9D49E5334D122ULL, "6050274_m.fxe"},
	{0x05DA49808168F951ULL, "1053426831.ogg"},
	{0x05DA9CF761AB5363ULL, "1006894153.ogg"},
	{0x05DAB1CF02ABF4FAULL, "hlac_dr_0_deeproads_5838_2x.dds"},
	{0x05DB9E450FA37D05ULL, "6208043_m.fxe"},
	{0x05DCBC09F1812C15ULL, "hlac_lt_0_lowtown_docks_131070_2y.dds"},
	{0x05DD3B3B553E07F5ULL, "164432362.ogg"},
	{0x05DD41F71301DEC4ULL, "696384518.ogg"},
	{0x05DD9365A8BA3304ULL, "vxht_mansion2_27n_0_0.vlm"},
	{0x05DDBE99612A3A5EULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm_eyesm1.mao"},
	{0x05DEF3D06D8E415EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmla.phy"},
	{0x05DF077A53E88B7AULL, "785837411.ogg"},
	{0x05DF672FE4EAC444ULL, "lt_warehouse_n_0_0.msh"},
	{0x05DF792FE4EAE2ADULL, "lt_warehouse_n_0_0.mao"},
	{0x05DF7D2FE4EAE9E6ULL, "lt_warehouse_n_0_0.mmh"},
	{0x05DF81C1307926D3ULL, "lgt000ip_black_vial.utp"},
	{0x05DF8819965F539AULL, "284042965.ogg"},
	{0x05DF9C3B505AE9E9ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med2_c.impa.ani"},
	{0x05DFC8DDD430CFBFULL, "vxlt_0_lowtown_docks_a3_cto_0_0.vlm"},
	{0x05E00DD2F68D8CE1ULL, "zz_dae_debug6029128.cl"},
	{0x05E031A7C625F266ULL, "274237908.ogg"},
	{0x05E0577408865FE3ULL, "art\\vfx\\combat\\vfx_glb_ele_f_deathblow_c.mmh"},
	{0x05E0A2009E01DC74ULL, "hlac_lt_0_lowtown_a3_n_c_65533_1x.dds"},
	{0x05E0D313A8C036CEULL, "161653378.ogg"},
	{0x05E0E8388E5D612BULL, "isa211_isabela6202284.cl"},
	{0x05E0EB9569555D29ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv1.gad"},
	{0x05E10BCFCEE08B27ULL, "54791665.ogg"},
	{0x05E198CBA67CC999ULL, "vxw_bonepitdestroyed_a3_c0_0_0.vlm"},
	{0x05E222EBCB147EA6ULL, "hlac_lt_gallowscourtyard_0_0x.dds"},
	{0x05E24014D2E56319ULL, "vxlt_0_lowtown_docks_a3_n_bqt_0_0.vlm"},
	{0x05E26143F91912AAULL, "vxw_woundedcoast_cn_0_0.vlm"},
	{0x05E2A6BCF13096A4ULL, "hlac_ht_playershouse_n_855_36x.dds"},
	{0x05E2FAFB7CAB6D92ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cooker.phy"},
	{0x05E32C179A7999D6ULL, "hlac_lt_0_lowtown_docks_n_-196609_1z.dds"},
	{0x05E3522D83C95771ULL, "gen_im_trs_bst_spd_silk_02.uti"},
	{0x05E371BCC4282687ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium1_c.impa.gad"},
	{0x05E39BD3AE8C1AF9ULL, "hlac_lt_0_lowtown_a2_n_c_0_9x.dds"},
	{0x05E3D6241DA9CFFEULL, "545625298.ogg"},
	{0x05E444A0BD6C45A5ULL, "dae000ip_crowd_em2_talk11.utp"},
	{0x05E456E4DDC72262ULL, "pro000_captain_wright6004402.cl"},
	{0x05E49CDD9731B060ULL, "lt_0_lowtown_docks_a3_n_38_0_ll.mmh"},
	{0x05E58C567731779DULL, "6090810_f.fxe"},
	{0x05E61F7F304233B6ULL, "6060462_m.fxe"},
	{0x05E6A10C4D1E2EDDULL, "vxht_mansion_n_4wz_0_0.vlm"},
	{0x05E6B73B5061366BULL, "art\\vfx\\combat\\vfx_r_dwp_front_med2_c.impa.gad"},
	{0x05E6EF58FE2E1EFBULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha2a.phy"},
	{0x05E722ED8EE71DB0ULL, "602527660.ogg"},
	{0x05E75F60EF39A035ULL, "dae000_dalish_apprentices6147556.cl"},
	{0x05E7A71C1C88EE06ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_loop_c.mmh"},
	{0x05E7ABC30B309C09ULL, "6166998_m.fxe"},
	{0x05E7B75CAF157431ULL, "w_bonepitmining_a2_227_0.mmh"},
	{0x05E7BC95695B2C77ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv1.ani"},
	{0x05E7C35CAF1588D2ULL, "w_bonepitmining_a2_227_0.mao"},
	{0x05E7CB65618FA7C2ULL, "vxlt_gallowscourtyard_n_75i_0_0.vlm"},
	{0x05E7D15CAF15A0EFULL, "w_bonepitmining_a2_227_0.msh"},
	{0x05E7D599AE938736ULL, "art\\characters\\creatures\\spider\\textures\\c_corspidra_0t.dds"},
	{0x05E814DD39B95134ULL, "follower_banter6163759.cl"},
	{0x05E9742FE4F32EB1ULL, "lt_warehouse_n_0_0.phy"},
	{0x05E974534D4B45CBULL, "art\\vfx\\combat\\vfx_r_var_modal_act_c.anb"},
	{0x05E978B3FB13072BULL, "vxlt_warehouse_n_14z_0_0.vlm"},
	{0x05E98421235B7AD4ULL, "631790187.ogg"},
	{0x05E9E1BE9D3A8819ULL, "hlac_lt_gallowstemplar_f_2395_1x.dds"},
	{0x05EA66BCC42E3294ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium1_c.impa.evt"},
	{0x05EA70338542E39AULL, "6129761_m.fxe"},
	{0x05EC0507D8511B6DULL, "61869258.ogg"},
	{0x05EC05429AB04B84ULL, "vxht_mansion2_20o_0_0.vlm"},
	{0x05EC485F77941029ULL, "6171091_m.fxe"},
	{0x05ECD24B8F1BCA43ULL, "art\\characters\\playercharacter\\humanmale\\cth_visa.msh"},
	{0x05ECE04B8F1BE19DULL, "art\\characters\\playercharacter\\humanmale\\cth_visa.mmh"},
	{0x05ECE44B8F1BE8AEULL, "art\\characters\\playercharacter\\humanmale\\cth_visa.mao"},
	{0x05ED6DBB0FB4794FULL, "gen00fl_sebastian6195865.cl"},
	{0x05EDB43B50674FD0ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med2_c.impa.evt"},
	{0x05EECB48C83E3C86ULL, "585188236.ogg"},
	{0x05EFAC5A6574F4F9ULL, "vxlt_hangedman_20h_0_0.vlm"},
	{0x05F00A58A1685A21ULL, "vxht_0_hightown_a3_at4_0_0.vlm"},
	{0x05F00AA89872E219ULL, "vxw_bonepitdestroyed_a3_n7_0_0.vlm"},
	{0x05F012E5C7FFD56FULL, "vxdr_primevaltomb_4h3_0_0.vlm"},
	{0x05F07D2FE4F95CADULL, "lt_warehouse_n_0_0.rml"},
	{0x05F0A3F3915FA7D3ULL, "hlac_lt_0_lowtown_a2_-65536_6z.dds"},
	{0x05F11AEC905825E1ULL, "hlac_w_encounter2_camp_327680_0y.dds"},
	{0x05F17FBD0BAE1CABULL, "390172430.ogg"},
	{0x05F19378107C44A2ULL, "rdr200st_martin.utm"},
	{0x05F202AA24D75324ULL, "gencr_street_thug_2w_f.utc"},
	{0x05F216A79E7F4D2FULL, "follower_banter6161262.cl"},
	{0x05F22DCA47B1A021ULL, "194461907.ogg"},
	{0x05F26D86F59001A9ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sib_flemeth_real.ani"},
	{0x05F27F2D2AAB7064ULL, "1048112508.ogg"},
	{0x05F298500195F21EULL, "364509069.ogg"},
	{0x05F2FB8FA273FC93ULL, "345424339.ogg"},
	{0x05F30AE8EC0E88F3ULL, "120277321.ogg"},
	{0x05F4314AB87DF35BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainsack.mmh"},
	{0x05F43D4AB87E07F8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainsack.mao"},
	{0x05F443E30F142CBFULL, "lt_0_lowtown_docks_195_0.msh"},
	{0x05F449E30F143761ULL, "lt_0_lowtown_docks_195_0.mmh"},
	{0x05F44F4AB87E2645ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainsack.msh"},
	{0x05F455E30F144B02ULL, "lt_0_lowtown_docks_195_0.mao"},
	{0x05F4639B73E058B0ULL, "default_additive.mao"},
	{0x05F465E436C1EC38ULL, "vxw_bonepit_26z_0_0.vlm"},
	{0x05F492E5837BBE86ULL, "vxw_bonepit_264_0_0.vlm"},
	{0x05F4B498B6FBCFC3ULL, "abi_w_taunt.bnk"},
	{0x05F4B6D6CF66A04AULL, "vxlt_0_lowtown_docks_n_cs7_0_0.vlm"},
	{0x05F52463D22F2DE5ULL, "plt_lgt000pt_ultra_1_05.plo"},
	{0x05F532CC7B1DE7B0ULL, "727715760.ogg"},
	{0x05F578B8D883934CULL, "vxlt_hangedman_43_0_0.vlm"},
	{0x05F5E89D36134F4DULL, "835544679.ogg"},
	{0x05F6056BE1DAFA68ULL, "778446039.ogg"},
	{0x05F60A574B99F5A3ULL, "lowtown_male_lowlife1_sta6165031.cl"},
	{0x05F6C4831CCEDCCDULL, "hlac_lt_0_lowtown_a2_n_c_131073_2z.dds"},
	{0x05F6F8CEC64896F9ULL, "6152632_m.fxe"},
	{0x05F71774CD4BA9C7ULL, "vxca_caverns2prc_a9n_0_0.vlm"},
	{0x05F73B1431B60707ULL, "151354426.ogg"},
	{0x05F85EBCC43A62CDULL, "art\\vfx\\combat\\vfx_w_wns_front_medium1_c.impa.ani"},
	{0x05F9374E8A986435ULL, "6206913_m.fxe"},
	{0x05F94AE709B1E83BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall_tool.mmh"},
	{0x05F956E709B1FCD8ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall_tool.mao"},
	{0x05F968E709B21B25ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall_tool.msh"},
	{0x05F98886F5964E2BULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sib_flemeth_real.gad"},
	{0x05FA5961FA19A350ULL, "vxlt_hangedman_1ow_0_0.vlm"},
	{0x05FB66B39DDC00C6ULL, "hlac_ht_mansion2b_818_46x.dds"},
	{0x05FB838395B6EAAFULL, "801058010.ogg"},
	{0x05FBA3661C420EB9ULL, "mrl000_marethari6205603.cl"},
	{0x05FC02169ADBE49CULL, "hlac_lt_gallowstemplar_f_2383_9z.dds"},
	{0x05FC15370542AB25ULL, "325374599.ogg"},
	{0x05FC1A0CB9FE7011ULL, "841941646.ogg"},
	{0x05FCAA84433D1D30ULL, "vxca_caverns1_mines_icu_0_1.vlm"},
	{0x05FCC443FC0F6E5AULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf01.mmh"},
	{0x05FCD643FC0F8CA0ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf01.msh"},
	{0x05FDB6EB43D7150AULL, "18587229.ogg"},
	{0x05FE7C858A68FA99ULL, "vxw_bonepitdestroyed_a3_i6_0_0.vlm"},
	{0x05FF44FDDD0543AAULL, "vxw_woundedcoast_uh_0_0.vlm"},
	{0x05FFB81373E7CD1AULL, "hlac_lt_0_lowtown_-196608_2z.dds"},
	{0x05FFC36A2B48A710ULL, "994024761.ogg"},
	{0x06002D64FAC92F60ULL, "vrc321ip_spooky2_obj_1.utp"},
	{0x0600539CB2615264ULL, "6082676_f.fxe"},
	{0x0600A74CA41EB324ULL, "595003969.ogg"},
	{0x0601CCCEF3E0F829ULL, "vxht_bartrand_n_64j_0_0.vlm"},
	{0x060270EE54EF00A3ULL, "hairalpha_rendertodepth_p0.vsh.fxo.dx11"},
	{0x060304C27571D706ULL, "vxw_bonepit_215_0_0.vlm"},
	{0x060342B831BC39A4ULL, "vxw_blightlands_ux_0_0.vlm"},
	{0x0603F141F6BDCE1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_enter.ani"},
	{0x0604324FF1DC2D9EULL, "vxht_mansion2_2k0_0_0.vlm"},
	{0x060447A3B97615C4ULL, "bnn.80.dds"},
	{0x06049470747BE1CFULL, "486514339.ogg"},
	{0x0604C3FA60F6ADC7ULL, "mag360_samson6062122.cl"},
	{0x0604D4D37E78AFB4ULL, "58194906.ogg"},
	{0x0604EB98B8CB3060ULL, "t3_har_wht.tnt"},
	{0x060559E7C16FF923ULL, "1054308189.ogg"},
	{0x06058951BF41C247ULL, "vxca_caverns2prc_a4i_0_0.vlm"},
	{0x06058F9E8BF19ED6ULL, "w_encounter2_camp_47_0.rml"},
	{0x0605A521B6877E22ULL, "hla_ht_0_hightown_-65534_0x.dds"},
	{0x060619AFE5F7097DULL, "character_character_p31.psh.fxo.dx11"},
	{0x060644AB7702351AULL, "vxlt_0_lowtown_a2_n_db6_0_0.vlm"},
	{0x0606695C51A22764ULL, "gen00fl_varric6087481.cl"},
	{0x0606AA9283C9F48FULL, "plt_mrl200pt_n_siblingleft.plo"},
	{0x0606B3D19F540FD6ULL, "hlac_ht_chantry_2534_27z.dds"},
	{0x0606DF81E7F9396AULL, "hlac_ca_caverns2_5802_30z.dds"},
	{0x060705EDCC655615ULL, "6051461_m.fxe"},
	{0x0607502AE786A11FULL, "hlac_ht_keep_n_3732_0x.dds"},
	{0x06079FBC5B4B0339ULL, "hlac_dr_0_deeproads_5849_4y.dds"},
	{0x06085A159B0049BCULL, "lowtown_male_lowlife1_sta6164969.cl"},
	{0x060860CCBBDFE656ULL, "190034454.ogg"},
	{0x06088F684EDB3026ULL, "hlac_lt_0_lowtown_a2_131072_1z.dds"},
	{0x0608CB3EEC0FBBD0ULL, "vxlt_hangedman_1jt_0_0.vlm"},
	{0x0608E101C7FB007AULL, "lt_0_lowtown_docks_217_0.msh"},
	{0x0608EF01C7FB17B3ULL, "lt_0_lowtown_docks_217_0.mao"},
	{0x0608F301C7FB1EC0ULL, "lt_0_lowtown_docks_217_0.mmh"},
	{0x0609B2B2B974F18DULL, "vxlt_0_lowtown_a2_eot_0_0.vlm"},
	{0x0609EA8278A7DA81ULL, "hlac_lt_0_lowtown_docks_a3_131071_3x.dds"},
	{0x0609FC8D3B51E73CULL, "351599264.ogg"},
	{0x0609FDD55704D067ULL, "zz_ruins.cut"},
	{0x060A6E6F02B4EBF3ULL, "vxlt_0_lowtown_e0s_0_0.vlm"},
	{0x060AA471EF1F4FA1ULL, "dae000_mother6121700.cl"},
	{0x060AC24688D6C1EEULL, "llac_w_bonepit_0_0.dds"},
	{0x060ACE07CE1772AEULL, "804999991.ogg"},
	{0x060AFE41F6C4035EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_enter.gad"},
	{0x060B009BC81C87B6ULL, "zz_vrc_debug6200222.cl"},
	{0x060B13FB2F4B2947ULL, "vxlt_warehouse_156_0_0.vlm"},
	{0x060B8D4F90DFC8BEULL, "qun100ip_at_safehouse.utp"},
	{0x060BBE01BE6E0633ULL, "vxdr_dungeon_1m3_0_0.vlm"},
	{0x060C889E8BF7B1FAULL, "w_encounter2_camp_47_0.phy"},
	{0x060CA9623140CFA8ULL, "6162704_m.fxe"},
	{0x060CD58AC502492BULL, "vxht_0_hightown_a2_9wf_0_0.vlm"},
	{0x060D43E9906B4B01ULL, "6039947_f.fxe"},
	{0x060DB6DACEFB5C2AULL, "vxw_woundedcoast_ri_0_0.vlm"},
	{0x060E1EF310E36389ULL, "llac_lt_0_lowtown_131070_2.dds"},
	{0x060EFBC7EBB739C4ULL, "gen00fl_sebastian6174025.cl"},
	{0x060FD3D49AD0808DULL, "hlac_lt_0_lowtown_n_65535_1x.dds"},
	{0x060FEEFC920BE070ULL, "lt_0_lowtown_docks_a3_133_0.msh"},
	{0x060FFCFC920BF78AULL, "lt_0_lowtown_docks_a3_133_0.mmh"},
	{0x061000FC920BFEC1ULL, "lt_0_lowtown_docks_a3_133_0.mao"},
	{0x06108E9430080365ULL, "vxht_mansion2_2iu_0_0.vlm"},
	{0x061092927A219362ULL, "zz_dae_debug6047631.cl"},
	{0x0611D941F6C9E37DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_enter.evt"},
	{0x0612D658FE532E9AULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha2a.msh"},
	{0x0612E38D175C2AA8ULL, "vxlt_0_lowtown_docks_n_b4i_0_0.vlm"},
	{0x0612E858FE534DE0ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha2a.mmh"},
	{0x061349760CF36CF6ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_strt.evt"},
	{0x0613E4F99A6BAA3CULL, "sse_ss_lowclass_thugs_m_1.gda"},
	{0x0613FB2EB137DAC7ULL, "vxca_caverns2prc_a3h_0_0.vlm"},
	{0x06140BE54110C986ULL, "one350_terath6087147.cl"},
	{0x06144B57B2D21D70ULL, "vxca_cavern1_haunted_igx_0_0.vlm"},
	{0x06154BE486216986ULL, "follower_barks6153757.cl"},
	{0x06156DE80D4FCA19ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd03a.phy"},
	{0x0615AAAF1230A7BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_look_right.ani"},
	{0x0615B7C1DD0FFA44ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunarifort_sa.dds"},
	{0x0615F3D5EE88F62DULL, "lgt000cr_gifre.utc"},
	{0x0616C83FD930C760ULL, "579091014.ogg"},
	{0x06177E6898C883C3ULL, "mcr371_merebeth_st.stg"},
	{0x0617883A66270A45ULL, "782128283.ogg"},
	{0x06179CB2FD8CD145ULL, "vxlt_0_lowtown_docks_a3_n_b10_0_0.vlm"},
	{0x06189F98A631970CULL, "502211960.ogg"},
	{0x0618A8EC9A5FCABEULL, "art\\vfx\\combat\\vfx_r_dwp_front_long_c.impa.evt"},
	{0x0618F9EAFBE31BF8ULL, "6208179_f.fxe"},
	{0x06190F6F4F05581EULL, "mag121_eau_banter6107383.cl"},
	{0x061940760CF7C881ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_strt.gad"},
	{0x061980AF713EDB59ULL, "558954320.ogg"},
	{0x061A13DBE9C2EC9CULL, "6072316_f.fxe"},
	{0x061B1B3F2336E828ULL, "6167404_m.fxe"},
	{0x061B65660CDA8B63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\persuade_die.evt"},
	{0x061B66D125898803ULL, "849369558.ogg"},
	{0x061BBBC87408DB94ULL, "con_ico_charm.dds"},
	{0x061C1154E3E4B49FULL, "369161940.ogg"},
	{0x061C3B19222A2C73ULL, "6116384_m.fxe"},
	{0x061C9163AEE6D7EDULL, "llac_dr_dungeon_3214_14.dds"},
	{0x061C9C1495D37004ULL, "follower_banter6170092.cl"},
	{0x061C9E2FE3AF7B73ULL, "126333085.ogg"},
	{0x061CC04F7DC253C3ULL, "6102801_f.fxe"},
	{0x061CDFF26F155D94ULL, "vrc221_bartrand6065316.cl"},
	{0x061CE7F3BAA990EDULL, "isa311_castillon6092077.cl"},
	{0x061DD8C2E8E89650ULL, "vxw_encounter2_camp_7a_0_0.vlm"},
	{0x061EB4CCEBA64B59ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgta.mmh"},
	{0x061EBECCEBA65C47ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgta.msh"},
	{0x061FBFEC9A6610E9ULL, "art\\vfx\\combat\\vfx_r_dwp_front_long_c.impa.gad"},
	{0x061FC0C66056136EULL, "dae000_keep_noblewoman_sta6162045.cl"},
	{0x061FC5C6DDE033F4ULL, "gen00fl_anders6215204.cl"},
	{0x06202EE7076BEB1CULL, "882027049.ogg"},
	{0x062041760CFDE92FULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_strt.ani"},
	{0x06208951EC1F33A2ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_180_c.impa.gad"},
	{0x0620AE902E142600ULL, "hlac_ht_bartrand_n_3369_2z.dds"},
	{0x062105E5C45F4047ULL, "6170228-pridedrain-h.cut"},
	{0x06211609D5C85E9EULL, "vxht_mansion2_2q2_0_0.vlm"},
	{0x06216923AC1BC0B6ULL, "alphastreaks01.mao"},
	{0x0621C18720FF58F9ULL, "gen00fl_fenris6103319.cl"},
	{0x0621E7B7721FE444ULL, "vxht_mansion2_1v2_0_0.vlm"},
	{0x06223E660CE067E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\persuade_die.gad"},
	{0x06224D634734F624ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkrockend.msh"},
	{0x06225F634735148DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkrockend.mao"},
	{0x0622636347351BC6ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkrockend.mmh"},
	{0x0622B2AF123B3FB2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_look_right.evt"},
	{0x0622E343FC307265ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf01.phy"},
	{0x0622F1CD764F9CB2ULL, "vxlt_warehouse_n_x0_0_0.vlm"},
	{0x0622FA672ED1B931ULL, "vxca_caverns1_mines_e8t_0_0.vlm"},
	{0x0623668F3D95C7F8ULL, "vxlt_0_lowtown_a3_n_c_f4w_0_0.vlm"},
	{0x0623F628EF1F22BCULL, "vxw_bonepitdestroyed_a3_ci_0_0.vlm"},
	{0x06240BCCCF14D9F7ULL, "dae000_gamlen6149477.cl"},
	{0x06247E1467956503ULL, "18845896.ogg"},
	{0x0624ACDD9656AAB2ULL, "601233711.ogg"},
	{0x0624BAE0260C8C79ULL, "6143674_m.fxe"},
	{0x062503B0A0A7C6C7ULL, "352874749.ogg"},
	{0x0625272397D1463EULL, "ht_0_hightown_a2_34_0_lc.mmh"},
	{0x062540F86534613DULL, "553133580.ogg"},
	{0x0625E0E51725C14CULL, "656821536.ogg"},
	{0x062601D69EA44D96ULL, "prm000im_ivo_amulet.uti"},
	{0x062690EC9A6BDF37ULL, "art\\vfx\\combat\\vfx_r_dwp_front_long_c.impa.ani"},
	{0x06270CA40869D6DBULL, "vxlt_0_lowtown_a3_n_c_i0s_0_0.vlm"},
	{0x0627742BB92CA6D5ULL, "hlac_lt_0_lowtown_-1_1y.dds"},
	{0x06279F81D782F3B2ULL, "vxlt_darktown_4t8_0_0.vlm"},
	{0x0627AC51EC258D99ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_180_c.impa.evt"},
	{0x06280B481387BC8DULL, "282196207.ogg"},
	{0x06284D6CC5D63AA2ULL, "hlac_lt_0_lowtown_n_65536_7x.dds"},
	{0x062859F591BA3A71ULL, "102215895.ogg"},
	{0x06286EF292037A43ULL, "vxlt_hangedman_n_2zc_0_0.vlm"},
	{0x0628E8BDF379B20CULL, "vxlt_0_lowtown_a2_n_c_gz9_0_0.vlm"},
	{0x0628F08B57ED77EBULL, "6122780.cut"},
	{0x0628F228FEA7D547ULL, "hla_ht_0_hightown_a2_-65537_0x.dds"},
	{0x06293102D55AD0BDULL, "6109557_m.fxe"},
	{0x06294D660CE69FC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\persuade_die.ani"},
	{0x06295F0FC05BC5F0ULL, "fx_headcloud.mao"},
	{0x0629C84E3035B052ULL, "and111_lirene6084369.cl"},
	{0x0629D1AF12419345ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_look_right.gad"},
	{0x062A0C1C82DA5AA2ULL, "vxw_bonepit_2aa_0_0.vlm"},
	{0x062B12C46699DCDAULL, "26083346.ogg"},
	{0x062B5DA5DFF29D4EULL, "hlac_lt_0_lowtown_a2_n_-196609_0x.dds"},
	{0x062B751A0BDF3364ULL, "mag121_viveka6083366.cl"},
	{0x062BAB2006C554A5ULL, "gen00fl_aveline6114844.cl"},
	{0x062C31A6E6E1E9D5ULL, "vxlt_gallowstemplar_3e3_0_0.vlm"},
	{0x062C4A9FDADEAED0ULL, "vxw_encounter2_camp_2n_0_0.vlm"},
	{0x062CDC2C7809B311ULL, "vxw_encounter3_dz_0_0.vlm"},
	{0x062D897BD85EFD67ULL, "vxht_0_hightown_n_9p4_0_0.vlm"},
	{0x062DAAA0782E1FE8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01a.mmh"},
	{0x062DC8A0782E5212ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01a.msh"},
	{0x062DFFEF9CBAA8F0ULL, "dae011_corff_sta6142476.cl"},
	{0x062E83ED0CEC3B8BULL, "plt_vis160pt_traps.plo"},
	{0x062FAC9AF0092742ULL, "dae000_oldmangossip6175756.cl"},
	{0x062FE5B495963AE7ULL, "vxw_encounter3_7i_0_0.vlm"},
	{0x062FFC25A684E237ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem.phy"},
	{0x0630F74D2A4D4712ULL, "6016485_m.fxe"},
	{0x0631226116C5B582ULL, "gen00fl_anders6113061.cl"},
	{0x06315104DA6F8803ULL, "6169885_m.fxe"},
	{0x063161D12C42A27DULL, "ser221_gascard6122830.cl"},
	{0x0631A3E744DE632DULL, "one300_nuncio6075328.cl"},
	{0x0631A8EBE21D2CB9ULL, "6185496_m.fxe"},
	{0x0631AA630256823DULL, "6129480_f.fxe"},
	{0x0631E11941D34C07ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\cindes\\custom\\meleeattackv1_frozen.gad"},
	{0x06326805E1153B3CULL, "vxw_bonepitdestroyed_a3_nn_0_0.vlm"},
	{0x06329FE302FA839FULL, "6092354_m.fxe"},
	{0x063304F6CAC88F8AULL, "zz_approval_debug6003090.cl"},
	{0x06334BF365E9844FULL, "972006821.ogg"},
	{0x06337BB16D816551ULL, "vxdr_primevaltomb_6ni_0_0.vlm"},
	{0x0633CE3EDA1CD0C8ULL, "dae300_bran6143450.cl"},
	{0x0634134978E62C19ULL, "character_character_p95.psh.fxo"},
	{0x063421D151C8DBF7ULL, "lt_0_lowtown_docks_2_0.anb"},
	{0x0634938EF20906C8ULL, "dae000_templar_knight6165163.cl"},
	{0x063495ADEB2F23D3ULL, "6179909_m.fxe"},
	{0x06349A871C75BFD1ULL, "hlac_lt_0_lowtown_131071_1z.dds"},
	{0x0634B451EC302588ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_180_c.impa.ani"},
	{0x0634CF2C05B80234ULL, "269185751.ogg"},
	{0x06360DF9B7ECEFAEULL, "art\\vfx\\combat\\vfx_c_mer_thrust_c.mmh"},
	{0x063683815CEEE673ULL, "889917315.ogg"},
	{0x0637112CCE6E9047ULL, "652720674.ogg"},
	{0x06377C53ADA92B4BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_bdy_lgtb.mao"},
	{0x0637A2DFC750E93DULL, "6102457_m.fxe"},
	{0x0637E6940F07D8CBULL, "78783938.ogg"},
	{0x0637E7F9D1F4E798ULL, "vxlt_0_lowtown_a3_n_c_hfg_0_0.vlm"},
	{0x063818E7371D2C56ULL, "759644513.ogg"},
	{0x06389F2781768824ULL, "hlac_lt_undercitydungeon_-1_41z.dds"},
	{0x0638E829D4836A4DULL, "nat380_party_banter6066171.cl"},
	{0x06391C92D3A9D759ULL, "gen00fl_aveline6053327.cl"},
	{0x06399DAC15D1A050ULL, "6172594_m.fxe"},
	{0x0639AF10DE396811ULL, "vxlt_0_lowtown_a2_e8e_0_0.vlm"},
	{0x0639ECEB41AF2FD2ULL, "941997773.ogg"},
	{0x0639FCC4B77D525BULL, "llac_dr_primevaltomb_665_2.dds"},
	{0x063A1B0964651896ULL, "262197999.ogg"},
	{0x063AC04211449276ULL, "vxlt_gallowstemplar_f_51i_0_0.vlm"},
	{0x063AC8E32743926DULL, "351487108.ogg"},
	{0x063AD6C96EB69DB0ULL, "472996533.ogg"},
	{0x063B2EB80AADDD88ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheela.msh"},
	{0x063B40B80AADFC19ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheela.mao"},
	{0x063B457095B02918ULL, "hlac_w_woundedcoast_-65533_0z.dds"},
	{0x063B4CB80AAE10B2ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheela.mmh"},
	{0x063B4E0969FFCB91ULL, "vxw_encounter3_ay_0_0.vlm"},
	{0x063B817C9AE01A3DULL, "mag291st_ambient1.stg"},
	{0x063B8528EA500E80ULL, "6207986_m.fxe"},
	{0x063BB327124AF402ULL, "vxht_0_hightown_bif_0_0.vlm"},
	{0x063C2711512C6911ULL, "vxca_caverns1_in3_0_0.vlm"},
	{0x063C524AB8BBE090ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainsack.phy"},
	{0x063D3708CE3FCA58ULL, "509078959.ogg"},
	{0x063DA78767414EA0ULL, "147531661.ogg"},
	{0x063E58AF0931E74EULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig_d.dds"},
	{0x063EBE7F2FAD5C5AULL, "dae000_bartrand6206887.cl"},
	{0x063F49AFFD14A187ULL, "6137500_f.fxe"},
	{0x063F8307F2482BCFULL, "ts_beasts_spdr_prog_fang.gda"},
	{0x06400EA0B4469C4AULL, "6089470_m.fxe"},
	{0x06401327D41AF0ABULL, "45107947.ogg"},
	{0x06403D2B0B542EAFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottle.mao"},
	{0x0640492B0B54434CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottle.mmh"},
	{0x06404F2B0B544D3EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottle.msh"},
	{0x0640B0E1FC76AD1CULL, "w_encounter3_73_0.mao"},
	{0x0640BCE1FC76C1BFULL, "w_encounter3_73_0.mmh"},
	{0x0640C2E1FC76CC61ULL, "w_encounter3_73_0.msh"},
	{0x064191B45567747DULL, "gen00fl_varric6105537.cl"},
	{0x06419FFE0A1CD898ULL, "vxlt_undercitydungeon_1jz_0_0.vlm"},
	{0x0641E6BA71A21F92ULL, "45738317.ogg"},
	{0x064218550FB76FCCULL, "321974101.ogg"},
	{0x064238E27553D4C5ULL, "1066855077.ogg"},
	{0x06424CB0BD51D169ULL, "t3_arm_pl1.tnt"},
	{0x064285266ADC4499ULL, "character_character_p42.psh.fxo"},
	{0x06434E5F36036FD2ULL, "vxht_0_hightown_c13_0_0.vlm"},
	{0x06437E2E16C06A86ULL, "223278053.ogg"},
	{0x0643C7018CD5C347ULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_ex.gad"},
	{0x0643F9D70B5FD9A2ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg101a_0s.dds"},
	{0x06440EA4221BECC4ULL, "vxlt_0_lowtown_docks_a3_n_bss_0_0.vlm"},
	{0x06441FEE194217A8ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ci_basic_1.mao"},
	{0x06442CF9B7F961ADULL, "art\\vfx\\combat\\vfx_c_mer_thrust_c.anb"},
	{0x06443FE219478C5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_lying_injured_dead.anb"},
	{0x064486B4A4B26D5DULL, "632722863.ogg"},
	{0x0644EBC1B2F96061ULL, "vxht_0_hightown_a3_cco_0_0.vlm"},
	{0x06454F62FF5A1C58ULL, "hlac_lt_0_lowtown_a2_n_-65537_2y.dds"},
	{0x06459C668128EDBDULL, "vxht_mansion2b_1v7_0_0.vlm"},
	{0x0645AAD470ECEDE2ULL, "dae150ip_rockslide.utp"},
	{0x0646223C06E10CD0ULL, "459901378.ogg"},
	{0x064630725B5A98E7ULL, "6136179_m.fxe"},
	{0x06468997677B0CA4ULL, "hlac_lt_gallowstemplar_n_2383_10x.dds"},
	{0x06468D6F0125DCCCULL, "153796425.ogg"},
	{0x0646CE1941E5884DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\cindes\\custom\\meleeattackv1_frozen.ani"},
	{0x0646E50525CEE776ULL, "6099589_m.fxe"},
	{0x0647967C7586D58EULL, "dae221st_dougal_pc_home.stg"},
	{0x06481A75F581CD1EULL, "6151036_m.fxe"},
	{0x06483F8666A4E9D3ULL, "vxht_0_hightown_c49_0_0.vlm"},
	{0x06489F1DF0A7BC6CULL, "llac_w_encounter2_camp_262145_0.dds"},
	{0x0649A37C43CC9029ULL, "hlac_w_woundedcoast_327682_0z.dds"},
	{0x0649A4B1FD2848B6ULL, "vxht_0_hightown_cbn_0_0.vlm"},
	{0x0649EAECE40AA2AAULL, "2488493.ogg"},
	{0x064A00DF18E9F967ULL, "hlac_lt_gallowstemplar_n_2391_8y.dds"},
	{0x064ADC72ED7EC9F5ULL, "192745520.ogg"},
	{0x064B61AA0EB013A0ULL, "339075018.ogg"},
	{0x064BA7063BC20DC5ULL, "154452707.ogg"},
	{0x064BDB2F122BFF46ULL, "vxw_woundedcoast_3r_0_0.vlm"},
	{0x064C453BCA38CB3DULL, "6095090_m.fxe"},
	{0x064CBF12D0029200ULL, "dae000ip_crowd_mf5_talk9.utp"},
	{0x064D1EDFCFC96736ULL, "269483132.ogg"},
	{0x064DA87460174A9CULL, "6162304_m.fxe"},
	{0x064E1ED4552C709BULL, "437272125.ogg"},
	{0x064E23B76853C4B2ULL, "dae000_gamlen6048489.cl"},
	{0x064E34029D2A659EULL, "dae000_madam_lusine6151889.cl"},
	{0x064E400818B98250ULL, "6027357_m.fxe"},
	{0x064E4E3B73B62BB5ULL, "vxlt_0_lowtown_docks_a3_d2x_0_0.vlm"},
	{0x064E7477744E0C48ULL, "vxht_0_hightown_a3_ccj_0_0.vlm"},
	{0x064EC3BD0840C799ULL, "ico_locked_sustained.dds"},
	{0x064ED1790AC93157ULL, "gen00fl_merrill6111408.cl"},
	{0x064F18683D7FA376ULL, "497257866.ogg"},
	{0x064FA9B2E9EA65D7ULL, "hlac_dr_primevalentrance_1726_3x.dds"},
	{0x064FB3AC2431D553ULL, "6072117_f.fxe"},
	{0x065021A6EDFA0555ULL, "vxlt_gallowstemplar_n_3sb_0_0.vlm"},
	{0x06505F6B516D9651ULL, "vxdr_primevaltomb_6xo_0_0.vlm"},
	{0x065081C63A1C6FA8ULL, "fx_cloud.dds"},
	{0x06509CC67DCFE1BEULL, "694374863.ogg"},
	{0x0650DC405F825C11ULL, "vxht_0_hightown_n_bzc_0_0.vlm"},
	{0x0650F7035CD25D19ULL, "character_character_p33.psh.fxo"},
	{0x0651BE20B9FD7985ULL, "vxlt_0_lowtown_a3_n_c_f4n_0_0.vlm"},
	{0x0651C992840A1574ULL, "plt_mrl200pt_n_siblingleft.nss"},
	{0x065238EA908944F8ULL, "t3_wdg_g01.tnt"},
	{0x06533010DFF029C0ULL, "hlac_lt_0_lowtown_n_65537_4x.dds"},
	{0x06534B03205930E3ULL, "6172632_m.fxe"},
	{0x06535B9DECC4ACA0ULL, "471460666.ogg"},
	{0x0653F5542FECD575ULL, "361963953.ogg"},
	{0x06542C0A8E1E3814ULL, "hlac_ht_0_hightown_a2_65536_4z.dds"},
	{0x0654B54640B7E1A3ULL, "gen00fl_aveline6121925.cl"},
	{0x0654C68FDF49F083ULL, "hlac_ht_brothel_n_3381_1x.dds"},
	{0x0654F20C95C5493CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_loop.ani"},
	{0x0655D9A522E1F58DULL, "llac_ht_0_hightown_-131072_1.dds"},
	{0x0655FE439528E1D5ULL, "llac_lt_0_lowtown_a3_n_c_131070_4.dds"},
	{0x065736580E9553C1ULL, "179219130.ogg"},
	{0x06574D793CB0904FULL, "zz_ave_debug6056757.cl"},
	{0x06575BC80347AE1AULL, "vxlt_0_lowtown_docks_n_cos_0_0.vlm"},
	{0x0657873DBCC43355ULL, "vxlt_gallowstemplar_3t0_0_0.vlm"},
	{0x0657B4018CE64B8DULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_ex.ani"},
	{0x0657C5CE606008B5ULL, "6209890_f.fxe"},
	{0x0658BC188460B7EBULL, "ice_shard_body.dds"},
	{0x06591918A232B3D4ULL, "vxlt_gallowscourtyard_6z0_0_0.vlm"},
	{0x06591F0ADCCAF85EULL, "vxw_blightlands_bg_0_0.vlm"},
	{0x065978867713EE03ULL, "6111103_m.fxe"},
	{0x06599AD045216501ULL, "71737.ogg"},
	{0x06599C52DFB220E7ULL, "vxlt_0_lowtown_a2_ewn_0_0.vlm"},
	{0x0659AC5DA7F0DAB6ULL, "hlac_lt_0_lowtown_a2_n_-65536_4z.dds"},
	{0x0659C79E8CDE4022ULL, "llac_lt_0_lowtown_a2_n_65536_3.dds"},
	{0x0659CE7B26450B22ULL, "dae010cr_joker3.utc"},
	{0x0659F8F936179894ULL, "vxw_encounter2_p9_0_1.vlm"},
	{0x065A21401DD93D1EULL, "vxlt_hangedman_5n_0_0.vlm"},
	{0x065B77C809EA295AULL, "follower_banter6163586.cl"},
	{0x065B81A3526AD303ULL, "gen00fl_sebastian6177894.cl"},
	{0x065BFF0C95CB7E7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_loop.gad"},
	{0x065C2B7FD948C556ULL, "gr_02_glow.mao"},
	{0x065C4DAC2014CBA6ULL, "6071837_m.fxe"},
	{0x065C53DDE0EB1232ULL, "6133142_m.fxe"},
	{0x065D40171C57E3C1ULL, "dae011_gossips6149781.cl"},
	{0x065E488F97F43432ULL, "vxw_woundedcoast_1be_0_1.vlm"},
	{0x065E9E93CE473A5EULL, "64067283.ogg"},
	{0x065F659A24A30C81ULL, "gen_im_acc_rng_act3_06.uti"},
	{0x0660C25D0528F683ULL, "dae000_meredith6080996.cl"},
	{0x06611ED151EFC40BULL, "lt_0_lowtown_docks_2_0.rml"},
	{0x066155DFC963B5D5ULL, "216294867.ogg"},
	{0x066187D3B6B2421FULL, "vxlt_warehouse_n_aj_0_0.vlm"},
	{0x0661CB6FF138F76FULL, "6201543_m.fxe"},
	{0x0662DA0C95D15D9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_loop.evt"},
	{0x06631C7E2DF110FEULL, "vxca_caverns1_k2t_0_2.vlm"},
	{0x066333997E2052C1ULL, "307187591.ogg"},
	{0x06633B5418A5EE12ULL, "633983138.ogg"},
	{0x066423C0012E802EULL, "493957694.ogg"},
	{0x06645AA637DF1496ULL, "vxdr_0_deeproads_gby_0_0.vlm"},
	{0x066473C735E156F0ULL, "6014337_m.fxe"},
	{0x06648D077E87E7A0ULL, "rdr111dg_assistant6055108.cl"},
	{0x0664A3A0AAFFAE4CULL, "hlac_dr_0_deeproads_5845_16y.dds"},
	{0x0665174E148338A8ULL, "6049821_m.fxe"},
	{0x06651B6927675CF8ULL, "gen00fl_sebastian6180444.cl"},
	{0x06653DBBFF5B857CULL, "988915853.ogg"},
	{0x06654ED74BCA3DDEULL, "34813456.ogg"},
	{0x066581AD38443CADULL, "qun110_petrice6036418.cl"},
	{0x06659CA7CD869BDEULL, "hlac_lt_0_lowtown_docks_n_0_2y.dds"},
	{0x0665D1B4FA9EA5FEULL, "vxlt_hangedman_n_b_0_0.vlm"},
	{0x06665C477196D165ULL, "303949269.ogg"},
	{0x0666C0711077F039ULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgta.msh"},
	{0x0666C832DB628F40ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_roba_0s.dds"},
	{0x0666CA7110780127ULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgta.mmh"},
	{0x0666CE7110780874ULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgta.mao"},
	{0x0667053E010B38D7ULL, "vxdr_primevaltomb_5ud_0_1.vlm"},
	{0x06673857C35743C5ULL, "981012385.ogg"},
	{0x0667819E8C445123ULL, "w_encounter2_camp_47_0.msh"},
	{0x06678F9E8C44697DULL, "w_encounter2_camp_47_0.mmh"},
	{0x0667939E8C446F8EULL, "w_encounter2_camp_47_0.mao"},
	{0x0667EF37A19CC2A8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small_2.mao"},
	{0x06683EDC49299F17ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_2_sylvesta.gad"},
	{0x06699485A0AE25C2ULL, "358276984.ogg"},
	{0x0669FC5C12B24AC1ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge11.ani"},
	{0x066A5846B795A4FBULL, "vxlt_0_lowtown_docks_a3_bgc_0_0.vlm"},
	{0x066B01E00A83C86CULL, "vxdr_primevaltomb_2um_0_0.vlm"},
	{0x066B2C85DB33E52BULL, "72667407.ogg"},
	{0x066BF00343ED4264ULL, "hlac_w_encounter2_196607_0y.dds"},
	{0x066C64E5FC7B0FDBULL, "503900244.ogg"},
	{0x066C968920D765C7ULL, "vxw_woundedcoast_1gb_0_0.vlm"},
	{0x066CDC59759CADC3ULL, "hlac_w_sundermount_196612_0y.dds"},
	{0x066DA93DE6A12C6FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffdeselectedexit.gad"},
	{0x066DD2BDB166C9DCULL, "6086987_m.fxe"},
	{0x066E324979CE0AFCULL, "lt_backalley_n_0_0.anb"},
	{0x066E57B71A1BD059ULL, "6021223_f.fxe"},
	{0x066F6BC35A675556ULL, "mag100_cullen6095132.cl"},
	{0x066F88A3CD41DBD4ULL, "6034541_f.fxe"},
	{0x066FB73FC7AA8502ULL, "vxlt_gallowstemplar_5qn_0_0.vlm"},
	{0x067055AC40104B97ULL, "6117677_m.fxe"},
	{0x0670925F13FA99EEULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_act_c.impa.evt"},
	{0x0671075C12B87C63ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge11.gad"},
	{0x0671105222260F2EULL, "seb100im_vael_amulet.uti"},
	{0x0671240F1872FA4BULL, "vxw_bonepitmining_a2_25f_0_0.vlm"},
	{0x06719D8B7C78404FULL, "hlac_lt_0_lowtown_65537_2x.dds"},
	{0x0671EF324273A874ULL, "573112652.ogg"},
	{0x06720B5E604B096CULL, "1066940770.ogg"},
	{0x0672655AE545CA76ULL, "vxlt_gallowstemplar_n_5aa_0_0.vlm"},
	{0x06734DAE89526D27ULL, "vxca_cavern1_haunted_k2p_0_2.vlm"},
	{0x0674349339CE7E0CULL, "490157242.ogg"},
	{0x06745895431FC168ULL, "gen_im_trs_gen_med_glv_r1.uti"},
	{0x067461029B4C259BULL, "399227454.ogg"},
	{0x0674BE3DE6A76E1CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffdeselectedexit.evt"},
	{0x0674D8F12D7A7EDBULL, "vxlt_0_lowtown_docks_a3_d3u_0_0.vlm"},
	{0x0674FC2474726AE3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottlecin_d.dds"},
	{0x0675C2407711BEB3ULL, "vxlt_gallowstemplar_53a_0_0.vlm"},
	{0x0675EBF48C1A63CFULL, "753436621.ogg"},
	{0x06762303F47A463EULL, "284356213.ogg"},
	{0x067671A29D221814ULL, "mcr371ip_vfx_target.ncs"},
	{0x0676895F13FEF599ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_act_c.impa.gad"},
	{0x0676A83A508E95BCULL, "6085137_m.fxe"},
	{0x0676D125A6C09B8FULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem.mao"},
	{0x0676DD25A6C0B02CULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem.mmh"},
	{0x0676E325A6C0BA1EULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem.msh"},
	{0x0677132CC78532F8ULL, "lgt000_chanter6147199.cl"},
	{0x067817FEBEE1568EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25_0.pwk"},
	{0x06782562A2EE36A3ULL, "art\\vfx\\creatures\\fxc_drg_brth_c_c.impa.ani"},
	{0x06782654B991384CULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_sand.phy"},
	{0x0678980704C5C881ULL, "6117979_f.fxe"},
	{0x06789CB5CF416389ULL, "vxdr_0_deeproads_cqk_0_0.vlm"},
	{0x06793797C4D74332ULL, "6139742_m.fxe"},
	{0x067ADF33F1AE602EULL, "zz_followers_banter6157613.cl"},
	{0x067AF682C4697705ULL, "6134737_m.fxe"},
	{0x067B12DD164B73B1ULL, "vxlt_hangedman_20p_0_0.vlm"},
	{0x067B69AC1E5014F9ULL, "w_encounter3_136_0.msh"},
	{0x067B73AC1E5025E7ULL, "w_encounter3_136_0.mmh"},
	{0x067B77AC1E502C34ULL, "w_encounter3_136_0.mao"},
	{0x067C072AD92D2871ULL, "art\\vfx\\combat\\vfx_w_wns_front_close4_c.anb"},
	{0x067C6BDC493A949DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_2_sylvesta.ani"},
	{0x067CAA4139A745C5ULL, "llac_lt_0_lowtown_docks_a3_n_0_5.dds"},
	{0x067CD06BAED4ABDCULL, "vxht_mansion2b_3dz_0_0.vlm"},
	{0x067CEE27272BBFB8ULL, "one280_emrys6104279.cl"},
	{0x067D0073C9E86A1BULL, "vxw_encounter1_ei_0_0.vlm"},
	{0x067D517D04807244ULL, "49728967.ogg"},
	{0x067D5C822E2094FEULL, "438640351.ogg"},
	{0x067D9A5F140531E7ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_act_c.impa.ani"},
	{0x067DDA9C71740ED4ULL, "6134743_m.fxe"},
	{0x067E07DC8A8CDA0CULL, "dae000_bartrand6206962.cl"},
	{0x067E87C602D588EBULL, "character_character_p57.psh.fxo"},
	{0x067EC7802D164611ULL, "6117680_m.fxe"},
	{0x067EC78932066417ULL, "6112577_m.fxe"},
	{0x067EF3865858E497ULL, "dae000_mother6156397.cl"},
	{0x067F28FD3170DF7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_right.gad"},
	{0x067F47F0EDB57310ULL, "95120636.ogg"},
	{0x067F7BC96A2774BEULL, "540402038.ogg"},
	{0x0680344E2BA498D6ULL, "and100_anders6095884.cl"},
	{0x068098A7BA8BAC27ULL, "6126894_f.fxe"},
	{0x0680C947288E3A7CULL, "llac_lt_0_lowtown_131071_2.dds"},
	{0x0680D4200847B11BULL, "follower_banter6143370.cl"},
	{0x0680E0A3F85936B4ULL, "6118024_m.fxe"},
	{0x0681A61D7A5979DBULL, "explosion_v2.cut"},
	{0x0681AEB74C4FC355ULL, "llac_lt_gallowstemplar_n_2383_8.dds"},
	{0x0681B5CCEBFACA46ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgta.phy"},
	{0x0681C63DE6B20635ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffdeselectedexit.ani"},
	{0x0681C693492586F2ULL, "t3_wdg_t00.tnt"},
	{0x06827B271BC8E11AULL, "vrc221_bartrand6065273.cl"},
	{0x06827B8E8F1B657BULL, "79372845.ogg"},
	{0x068282CA28607B4BULL, "vxw_bonepitmining_a2_1q1_0_0.vlm"},
	{0x0682D4BEA32FF4E4ULL, "6180832_m.fxe"},
	{0x06834ACE1F70975BULL, "vxlt_0_lowtown_docks_a3_d4t_0_0.vlm"},
	{0x06838D4792C9127FULL, "6131186_f.fxe"},
	{0x0683EE09C1C6BB46ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.cess.ani"},
	{0x0684A2DAB7598268ULL, "vxca_caverns1_icq_0_0.vlm"},
	{0x0684C5CE60E768FDULL, "vxdr_primevaltomb_7rd_0_0.vlm"},
	{0x0685175562FB4DADULL, "fex_ruin_spt_a.mmh"},
	{0x06851DFD3175386CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_right.evt"},
	{0x0685295562FB6C33ULL, "fex_ruin_spt_a.msh"},
	{0x06855AE425BF1D95ULL, "vxlt_0_lowtown_docks_bgb_0_0.vlm"},
	{0x0686098902E33678ULL, "vxlt_0_lowtown_h7l_0_0.vlm"},
	{0x068623C25ED4294EULL, "630762602.ogg"},
	{0x06863D62A2FA9C9AULL, "art\\vfx\\creatures\\fxc_drg_brth_c_c.impa.evt"},
	{0x0686465A929B77FAULL, "vxht_mansion2_1wo_0_0.vlm"},
	{0x068681F1D1F530E1ULL, "754460595.ogg"},
	{0x0686ABD44C0EFCA2ULL, "zz_and_debug6178574.cl"},
	{0x0686E7BC560B6E63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfront.evt"},
	{0x06874129E621993AULL, "341269542.ogg"},
	{0x0687A167AB4C4ABAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackr.mmh"},
	{0x0687A567AB4C51F1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackr.mao"},
	{0x0687B367AB4C6900ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackr.msh"},
	{0x06881739975C73A6ULL, "hlac_lt_0_lowtown_a2_131069_0y.dds"},
	{0x06882883C20A023EULL, "mag311_nyssa6041969.cl"},
	{0x06897628676DB0AEULL, "character_character_p34.psh.fxo.dx11"},
	{0x06898535CF146A5BULL, "6071839_f.fxe"},
	{0x0689D9F2FB7B8314ULL, "z_cl_offset.dds"},
	{0x068A062DD32045A3ULL, "1038293607.ogg"},
	{0x068A282AD9399DEAULL, "art\\vfx\\combat\\vfx_w_wns_front_close4_c.mmh"},
	{0x068A60C55C697278ULL, "dae000_oldmangossip6174982.cl"},
	{0x068A60F58D6A35D6ULL, "637483964.ogg"},
	{0x068A979504D79F5AULL, "gen_im_wep_rog_arc_act3_04.uti"},
	{0x068AEDCA83D35E58ULL, "6048435_m.fxe"},
	{0x068B0360A33B1379ULL, "follower_banter6163420.cl"},
	{0x068B324E16A07A1AULL, "vxlt_0_lowtown_a2_n_gpg_0_0.vlm"},
	{0x068B5EFC25E7FBD8ULL, "pro000ip_corpse_2.utp"},
	{0x068B7250BBDE829BULL, "vxw_encounter1_jh_0_0.vlm"},
	{0x068B7F7A36ED26CAULL, "hlac_lt_0_lowtown_65534_2x.dds"},
	{0x068B839B0F0C0E65ULL, "vxlt_0_lowtown_docks_bv5_0_0.vlm"},
	{0x068BCA60257F91F8ULL, "vxca_cavern1_haunted_eib_0_0.vlm"},
	{0x068C29C038623D46ULL, "ht_brothel_n_3364_lc.mmh"},
	{0x068CC1CD1CFADCE2ULL, "ave200_aveline_bar6116201.cl"},
	{0x068CD8458800C1A2ULL, "zz_mag_debug6054483.cl"},
	{0x068CD84F96140684ULL, "113148028.ogg"},
	{0x068D3C62A300B98DULL, "art\\vfx\\creatures\\fxc_drg_brth_c_c.impa.gad"},
	{0x068D9E721D1E2A9AULL, "gen_im_trs_gen_amu_r0.uti"},
	{0x068DC0BC56114AE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfront.gad"},
	{0x068DE51E14BC3259ULL, "vxlt_blackemporium_t6_0_0.vlm"},
	{0x068DFFB51C335D0FULL, "6059782_m.fxe"},
	{0x068E12A0A9EB6B96ULL, "6062782_f.fxe"},
	{0x068E50A29D367245ULL, "mcr371ip_vfx_target.utp"},
	{0x068EB2191F9BEEDBULL, "gen00fl_isabela6075315.cl"},
	{0x068EEF3EDCEFCDC6ULL, "vxlt_gamlenshouse_n_n0_0_0.vlm"},
	{0x06901609EC155DA6ULL, "art\\vfx\\combat\\vfx_glb_ele_p_additive_c.impa.ani"},
	{0x069061D67A7F2A0DULL, "6164196_m.fxe"},
	{0x0690B2115864BF9EULL, "gen00fl_varric6099408.cl"},
	{0x0690BB9BDC937ECCULL, "vxw_bonepitmining_a2_fl_0_0.vlm"},
	{0x06916FDB7F5FDAC9ULL, "mag200_feynriel6023974.cl"},
	{0x0691D609C1D2CFB7ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.cess.evt"},
	{0x0691DC321CCF4901ULL, "vxlt_gallowstemplar_685_0_0.vlm"},
	{0x0692B371109D155CULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgta.phy"},
	{0x069305FD31814CC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_right.ani"},
	{0x069309907C9FC6A4ULL, "fx_darkbolt_add.mao"},
	{0x069363C6A78A82E7ULL, "vxlt_0_lowtown_a2_ekj_0_0.vlm"},
	{0x06936B37C6A9BB09ULL, "453903928.ogg"},
	{0x06939DD4712EF8BFULL, "dae150ip_rockslide.ncs"},
	{0x0693B432910A03C5ULL, "6076541_m.fxe"},
	{0x0694812FED096670ULL, "149231442.ogg"},
	{0x069482FA0B9777B3ULL, "6099590_f.fxe"},
	{0x06949AED0D431658ULL, "plt_vis160pt_traps.nss"},
	{0x0694A36D09986BBBULL, "vxca_caverns1_mines_ii9_0_0.vlm"},
	{0x0694AF3A5085E8F6ULL, "vxlt_0_lowtown_docks_n_ba8_0_0.vlm"},
	{0x0694CFBC561782C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfront.ani"},
	{0x069589BC56718349ULL, "6040575_m.fxe"},
	{0x0695ADDD8D77EE7BULL, "vxlt_0_lowtown_docks_a3_bvd_0_0.vlm"},
	{0x0695F5AEE2480D84ULL, "6153788_m.fxe"},
	{0x0695FB439E3625CFULL, "434541046.ogg"},
	{0x0696049613020EDCULL, "llac_lt_0_lowtown_a2_65536_2.dds"},
	{0x069639B7FC85DB71ULL, "313122261.ogg"},
	{0x0696D90442A35F57ULL, "486383232.ogg"},
	{0x06970BAE485B2564ULL, "6143402_m.fxe"},
	{0x06974BB80AFC6DADULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheela.phy"},
	{0x0698CBE414A71C74ULL, "6119135_m.fxe"},
	{0x0698D3B63966A202ULL, "gen00fl_sebastian6173743.cl"},
	{0x0698D709C1D8F040ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.cess.gad"},
	{0x0698F3AA07DC0795ULL, "face_face_p0.vsh.fxo.dx11.assembly"},
	{0x06991947EFE6F138ULL, "413877842.ogg"},
	{0x069954C0B60B66B4ULL, "hlac_lt_foundry_n_320_8y.dds"},
	{0x06995FA775C976D8ULL, "6043535_m.fxe"},
	{0x0699FEEAB50CCED5ULL, "vxht_0_hightown_asx_0_0.vlm"},
	{0x069A126775C735ECULL, "vxht_playershouse_n_1tb_0_0.vlm"},
	{0x069AAEB9E7D268B6ULL, "388813504.ogg"},
	{0x069AEF3278D0FBC3ULL, "444030646.ogg"},
	{0x069C4E81C9B94829ULL, "zz_ave_debug6056764.cl"},
	{0x069C70701F8FFB46ULL, "484605898.ogg"},
	{0x069C8628C33FC98DULL, "vxht_keep_n_63d_0_0.vlm"},
	{0x069C961E5D60886CULL, "182822612.ogg"},
	{0x069CA715B4814B5AULL, "302364844.ogg"},
	{0x069CA72C277234ACULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf05.phy"},
	{0x069CE7DC40935F16ULL, "6166903_f.fxe"},
	{0x069CFE09EC1FBF17ULL, "art\\vfx\\combat\\vfx_glb_ele_p_additive_c.impa.evt"},
	{0x069D49C6FCABF4E9ULL, "hlac_lt_gallowsprison_n_-1_11x.dds"},
	{0x069D6DF0732C549DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_d.dds"},
	{0x069E012D686E57E5ULL, "295456976.ogg"},
	{0x069E6BE381ACC531ULL, "mag291_torpor6196575.cl"},
	{0x069F2D78CE89974CULL, "vxw_bonepitmining_a2_mm_0_0.vlm"},
	{0x069FC6A4C3AC79A6ULL, "6156095_m.fxe"},
	{0x06A0009E988ECF74ULL, "872997646.ogg"},
	{0x06A0312CB1F867BEULL, "dae000_pc_home_ambients.cnv"},
	{0x06A04ABBB5927ECEULL, "vxht_0_hightown_bgd_0_0.vlm"},
	{0x06A051598284DF89ULL, "vxht_chantry_2ck_0_0.vlm"},
	{0x06A0744A56FC2B90ULL, "dre000_bodahn6031354.cl"},
	{0x06A0D94431790E7CULL, "lt_0_lowtown_docks_a3_28_0.rml"},
	{0x06A12E2FC255C613ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkb.evt"},
	{0x06A1441BEEE71491ULL, "vxca_caverns1_i34_0_0.vlm"},
	{0x06A186949B45B368ULL, "vxca_caverns1_iio_0_0.vlm"},
	{0x06A1A7CDAF76C471ULL, "1027632389.ogg"},
	{0x06A265CE65BE7F6BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupa_sa.dds"},
	{0x06A2783B90E211F7ULL, "358956274.ogg"},
	{0x06A27AA61708CEB7ULL, "373591425.ogg"},
	{0x06A2D1DDF4652AD7ULL, "mag330st_walter_1.stg"},
	{0x06A3044F390557F3ULL, "dae200_saemus6091705.cl"},
	{0x06A309143C39B75DULL, "hlac_ht_0_hightown_a2_131071_1z.dds"},
	{0x06A3FF09EC25DFA0ULL, "art\\vfx\\combat\\vfx_glb_ele_p_additive_c.impa.gad"},
	{0x06A40C70574580D8ULL, "vxca_caverns1_gkb_0_1.vlm"},
	{0x06A4132CB1FC1DCEULL, "dae000_pc_home_ambients.bnk"},
	{0x06A42B3CD4F2D96EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thinsmall.msh"},
	{0x06A4393CD4F2F0BFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thinsmall.mao"},
	{0x06A4453CD4F3055CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thinsmall.mmh"},
	{0x06A4E44E968DE465ULL, "622301975.ogg"},
	{0x06A4E5C172DB5A15ULL, "lgt260cr_eau_tmp_conrad.utc"},
	{0x06A536C5F7892B19ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\vst_hightown_n.dds"},
	{0x06A5E89369C4E47FULL, "6152119_m.fxe"},
	{0x06A5F26B349EDB84ULL, "hlac_lt_0_lowtown_a2_n_c_-131071_0x.dds"},
	{0x06A631FC016650EBULL, "hlac_dr_primevalentrance_1726_11y.dds"},
	{0x06A636A598AAB53AULL, "6125595_f.fxe"},
	{0x06A6D71369FE5EFDULL, "gen00fl_varric6033690.cl"},
	{0x06A70E37ABEACDEBULL, "474962373.ogg"},
	{0x06A7226046830C09ULL, "vxht_0_hightown_n_7fv_0_0.vlm"},
	{0x06A7272FC25A2654ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkb.gad"},
	{0x06A75B37301E7DDDULL, "6152197_m.fxe"},
	{0x06A7AC3368C72990ULL, "vxlt_gallowscourtyard_n_7fv_0_0.vlm"},
	{0x06A7E5D3C8B7D434ULL, "vxlt_0_lowtown_n_e29_0_0.vlm"},
	{0x06A812BF5B32C452ULL, "387893583.ogg"},
	{0x06A81607FA8CAB1AULL, "vxlt_0_lowtown_a2_n_gja_0_0.vlm"},
	{0x06A83C4D4A3B8B8FULL, "gen00fl_isabela6079298.cl"},
	{0x06A8A8E264D9B93EULL, "420859301.ogg"},
	{0x06A8B1FBBE9D661EULL, "6178459_m.fxe"},
	{0x06A8C2170C9E9801ULL, "6174795_m.fxe"},
	{0x06A8C870976846B6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4_0.pwk"},
	{0x06A8DDA2A4E81434ULL, "6147716_m.fxe"},
	{0x06A8DED24F143070ULL, "zz_isa_debug6175285.cl"},
	{0x06A91233756292FFULL, "nat320_delilah6065949.cl"},
	{0x06A947B198F757DCULL, "gen_im_arm_glv_lgt_act1_02.uti"},
	{0x06A950C89044B542ULL, "6075734_m.fxe"},
	{0x06AA328CA6B9398BULL, "vxlt_0_lowtown_a2_n_ett_0_0.vlm"},
	{0x06AAA8C6A16AC204ULL, "hlac_w_sundermount_-196607_0y.dds"},
	{0x06AB2376917FEEAAULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.cess.gad"},
	{0x06ABB89FA29875F5ULL, "vxht_mansion_n_5mi_0_0.vlm"},
	{0x06ABD321BB15D78CULL, "vxht_chantry_26m_0_0.vlm"},
	{0x06AC1A6763A9AC9EULL, "vxht_0_hightown_n_ca8_0_0.vlm"},
	{0x06ACAF388ACCA10DULL, "vxlt_0_lowtown_bxz_0_0.vlm"},
	{0x06ACC067AB6B9BC5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackr.phy"},
	{0x06ACE13CB03BC3F9ULL, "vxdr_primevalentrance_24s_0_0.vlm"},
	{0x06ACFDAE5E8E7BDDULL, "mer341_jansen6082993.cl"},
	{0x06AD0901AFBBBCACULL, "624818676.ogg"},
	{0x06AD146EF8B87274ULL, "6074387_f.fxe"},
	{0x06AD1B8F1D2B92AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ent.gad"},
	{0x06AD46D449DFD011ULL, "22523524.ogg"},
	{0x06AD9F55C07FAFCCULL, "vxw_bonepitmining_a2_hb_0_0.vlm"},
	{0x06AE31497A03EC8CULL, "lt_backalley_n_0_0.rml"},
	{0x06AE362FC2605E32ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkb.ani"},
	{0x06AE70ECE8D9FD7EULL, "gen00fl_merrill6119616.cl"},
	{0x06AE907C6C5EACABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_blessing.anb"},
	{0x06AEB194382F56DCULL, "llac_lt_warehouse_0_9.dds"},
	{0x06AEDCF0F9B08F91ULL, "vxw_encounter3_9q_0_0.vlm"},
	{0x06AF32805C499AA9ULL, "767256262.ogg"},
	{0x06AFB88509306F2AULL, "vxw_bonepit_11j_0_0.vlm"},
	{0x06AFB9C53B6596C1ULL, "l_isa121ar_chantry.bnk"},
	{0x06AFDB8E71C7A0FDULL, "plt_cod_hst_antiva.plo"},
	{0x06AFEB303CD990CAULL, "45836127.ogg"},
	{0x06B020B5A69DD163ULL, "971961642.ogg"},
	{0x06B028CC58438F27ULL, "382137704.ogg"},
	{0x06B0D4F6879DADE7ULL, "mag100_cullen6095081.cl"},
	{0x06B0EC789AD14B5EULL, "vxw_sundermount_2op_0_0.vlm"},
	{0x06B13DFE97100745ULL, "vxht_mansion2_1wd_0_0.vlm"},
	{0x06B13E760A9C8A75ULL, "mcr310_argument6182303.cl"},
	{0x06B1C9328703BFBDULL, "572169896.ogg"},
	{0x06B206769185DC41ULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.cess.evt"},
	{0x06B223FABAFD085EULL, "gen00fl_aveline6131097.cl"},
	{0x06B2AC2B99473852ULL, "mag320_nobleman_16060261.cl"},
	{0x06B33C575C5A21CFULL, "ch1armtnt_ds2.psh.fxo"},
	{0x06B3BA69F88E573EULL, "vxht_keep_n_8aw_0_0.vlm"},
	{0x06B4308F1D31D55CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ent.evt"},
	{0x06B4611A7F4AE805ULL, "vxlt_0_lowtown_h7u_0_0.vlm"},
	{0x06B4B5C60729948EULL, "6082695_m.fxe"},
	{0x06B569C1644A4A9FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.open.gad"},
	{0x06B57F6EB66F52B7ULL, "qun120_arishok6051312.cl"},
	{0x06B6303BC4ECBAF9ULL, "vxlt_gallowstemplar_2p1_0_0.vlm"},
	{0x06B646C98EAF07DDULL, "6026988_m.fxe"},
	{0x06B6582E88453FBFULL, "vxht_playershouse_n_1qb_0_0.vlm"},
	{0x06B6B7DA799DD497ULL, "llac_lt_0_lowtown_a2_n_c_-65535_0.dds"},
	{0x06B6E6889CDECEF8ULL, "620696377.ogg"},
	{0x06B70D78CAFE9245ULL, "6126514_m.fxe"},
	{0x06B72E317CB27779ULL, "1009157208.ogg"},
	{0x06B73E0334EE3DB6ULL, "vxca_caverns1_mines_i1j_0_2.vlm"},
	{0x06B7FD54B9C6D617ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_sand.mmh"},
	{0x06B81354B9C6FB89ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_sand.msh"},
	{0x06B83F5F446F6F87ULL, "fen110_anso6095006.cl"},
	{0x06B8F69422C0FF8BULL, "hlac_lt_0_lowtown_docks_a3_131071_3z.dds"},
	{0x06B925531352414CULL, "hlac_w_bonepit_3_0y.dds"},
	{0x06B952A8848F0096ULL, "vxlt_darktown_517_0_1.vlm"},
	{0x06B9D777089AADFBULL, "hlac_w_bonepitmining_a2_0_5x.dds"},
	{0x06B9E9BA5D3C166CULL, "138554042.ogg"},
	{0x06BA2C051D4B5E50ULL, "317738526.ogg"},
	{0x06BA54B1010D9CCAULL, "6030441_m.fxe"},
	{0x06BAE7B2FEA707D6ULL, "688482529.ogg"},
	{0x06BB06786B56C430ULL, "6139469_m.fxe"},
	{0x06BB2D9FC5FE5E96ULL, "isa120st_hayder.stg"},
	{0x06BB52EE0A11983FULL, "6167568_m.fxe"},
	{0x06BB58997076584FULL, "gen00fl_merrill6129928.cl"},
	{0x06BB864BEAAE8AF4ULL, "6079087_f.fxe"},
	{0x06BBD1B48FDFB0DEULL, "mag121_idunna6084191.cl"},
	{0x06BC046823366062ULL, "vxca_caverns1_mines_iob_0_0.vlm"},
	{0x06BC6057F6E11B2CULL, "vxw_bonepitdestroyed_a3_9o_0_0.vlm"},
	{0x06BC99F04062BC3AULL, "w_bonepitdestroyed_a3_2_0_lc.mmh"},
	{0x06BCABD3BFCA8AC4ULL, "hlac_ca_caverns2prc_5802_15x.dds"},
	{0x06BCC3CCBDD1C9A5ULL, "hlac_lt_0_lowtown_a2_n_c_-3_2y.dds"},
	{0x06BD0ABF7B66804BULL, "art\\vfx\\combat\\vfx_glb_fire_imp_maj_c.mmh"},
	{0x06BD273898968D94ULL, "pro000_cut_flemeth_rescue6001230.cl"},
	{0x06BD27C19D221143ULL, "dae020ip_merch_tailor.utp"},
	{0x06BD351366711089ULL, "vxht_chantry_2yi_0_0.vlm"},
	{0x06BD4ECDEBA6A811ULL, "vxw_encounter3_2p_0_0.vlm"},
	{0x06BDEE44D582789EULL, "vxht_0_hightown_a2_50n_0_0.vlm"},
	{0x06BDF8302C2E353FULL, "6110287_m.fxe"},
	{0x06BE679AE7009F55ULL, "vxlt_0_lowtown_n_ded_0_0.vlm"},
	{0x06BE6BE20706C0C2ULL, "6084231_m.fxe"},
	{0x06BEFE7691905890ULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.cess.ani"},
	{0x06BF3BA753780411ULL, "gen00fl_dog6117713.cl"},
	{0x06BFC17F5B66E013ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_loop.evt"},
	{0x06BFEBF6459A49E8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_alcovetop.msh"},
	{0x06BFEFE644EB7447ULL, "hlac_lt_0_lowtown_n_65535_1z.dds"},
	{0x06BFFDF6459A6879ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_alcovetop.mao"},
	{0x06C007A8415E4583ULL, "dae210_gate_guard6209980.cl"},
	{0x06C009F6459A7C12ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_alcovetop.mmh"},
	{0x06C00DCE5A73D9FAULL, "vxht_mansion2_1kk_0_0.vlm"},
	{0x06C0C8D14C3E1EFCULL, "307723460.ogg"},
	{0x06C1272C2DC99AFFULL, "603666314.ogg"},
	{0x06C135C1231A7C73ULL, "6127432_m.fxe"},
	{0x06C1AE934BAFB07BULL, "319725558.ogg"},
	{0x06C1E221059152AAULL, "131177778.ogg"},
	{0x06C1FE0400B06FE6ULL, "rdr100pt_keepers.ncs"},
	{0x06C2388F1D3E2075ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ent.ani"},
	{0x06C239679814A81BULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masa_0n.dds"},
	{0x06C31694A5A43D71ULL, "zz_followers_banter6159489.cl"},
	{0x06C34B31F65E407FULL, "vxlt_gallowstemplar_5cu_0_0.vlm"},
	{0x06C36E22D7955F0DULL, "6081208_m.fxe"},
	{0x06C39B62D0555210ULL, "6042406_m.fxe"},
	{0x06C3B2A6B4B40B51ULL, "499585223.ogg"},
	{0x06C3F062F650E0D5ULL, "vxca_cavern1_haunted_dpi_0_0.vlm"},
	{0x06C41F8D1C8DFEC6ULL, "vxw_bonepitmining_a2_1xc_0_0.vlm"},
	{0x06C4B2C5834F133AULL, "656171415.ogg"},
	{0x06C4DCF4A311139AULL, "646180225.ogg"},
	{0x06C548C007554D85ULL, "6103645_m.fxe"},
	{0x06C56E8AEFDE1B46ULL, "138118116.ogg"},
	{0x06C5BA7F5B6B4054ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_loop.gad"},
	{0x06C636832F3E9FE8ULL, "916578916.ogg"},
	{0x06C66B2495072715ULL, "197047975.ogg"},
	{0x06C6E8A1A5A02DC5ULL, "6172700_m.fxe"},
	{0x06C7BC5B166F8F0AULL, "hlac_lt_0_lowtown_a2_n_c_65533_2x.dds"},
	{0x06C81ADB47B4A68CULL, "493228196.ogg"},
	{0x06C837378D60A793ULL, "vxlt_gallowstemplar_38r_0_0.vlm"},
	{0x06C8AFDD0821E6AFULL, "495003840.ogg"},
	{0x06C8EB5D2636C269ULL, "522751623.ogg"},
	{0x06C8EF09071803CBULL, "art\\characters\\creatures\\mubari\\animation\\p_wkf.ani"},
	{0x06C9059D6410E4A7ULL, "vxlt_0_lowtown_a2_n_fip_0_0.vlm"},
	{0x06C9243CD511E927ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thinsmall.phy"},
	{0x06C9ED13226247C8ULL, "hlac_ca_caverns1_mines_24421_7x.dds"},
	{0x06CA070F880A558BULL, "6071128_m.fxe"},
	{0x06CA46C1645C6BE5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.open.ani"},
	{0x06CA8A0078043E14ULL, "mas.cut"},
	{0x06CABAF9CF1A6E80ULL, "vxlt_gallowstemplar_f_3n8_0_0.vlm"},
	{0x06CAD45D1B7744AEULL, "6086170_m.fxe"},
	{0x06CAD72CB374DF4AULL, "ser221_gascard6133877.cl"},
	{0x06CAEBBF7B7288B0ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_maj_c.anb"},
	{0x06CB0F2A0CC38704ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackright.ani"},
	{0x06CB3461369828F3ULL, "246794454.ogg"},
	{0x06CB7A5C7843B080ULL, "65821246.ogg"},
	{0x06CBC3279C0B4C3FULL, "hlac_lt_0_lowtown_a2_n_c_1_2x.dds"},
	{0x06CBEDE26A8A3E08ULL, "vxlt_0_lowtown_docks_a3_n_d4p_0_0.vlm"},
	{0x06CC9A220C569F52ULL, "and100_anders6095996.cl"},
	{0x06CCACBF6B21F0F2ULL, "vxht_0_hightown_a2_4n1_0_0.vlm"},
	{0x06CCC97F5B717832ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_loop.ani"},
	{0x06CCD3D447821DA5ULL, "gen00fl_aveline6112705.cl"},
	{0x06CCDA44319E4ACCULL, "lt_0_lowtown_docks_a3_28_0.anb"},
	{0x06CD13BD2DB69D01ULL, "1004621212.ogg"},
	{0x06CD2166F7750AC2ULL, "vxht_mansion2_3cu_0_0.vlm"},
	{0x06CD40777161573CULL, "hlac_ht_chantry_n_2534_20x.dds"},
	{0x06CD720310411DC3ULL, "68297658.ogg"},
	{0x06CD76D6AD1EDEBBULL, "hlac_ca_caverns2prc_5802_22y.dds"},
	{0x06CDE32D2B5462C9ULL, "gen00fl_anders6111318.cl"},
	{0x06CDF6AF10411B6CULL, "6173364_m.fxe"},
	{0x06CE7FAB4C69F27AULL, "vxht_mansion2_1nd_0_0.vlm"},
	{0x06CEB02185F977F5ULL, "ser221_gamlen6131640.cl"},
	{0x06CEE69F219FA037ULL, "ttx_roof_int_d_1.mmh"},
	{0x06CEE93480F12FC2ULL, "vxlt_gallowscourtyard_n_7ey_0_0.vlm"},
	{0x06CEFC9F219FC5A9ULL, "ttx_roof_int_d_1.msh"},
	{0x06CF07B84081C90CULL, "480460697.ogg"},
	{0x06CF61FB6952E5A6ULL, "vxlt_gallowstemplar_f_5ya_0_0.vlm"},
	{0x06CF6F75F923609EULL, "vxlt_gallowstemplar_46z_0_0.vlm"},
	{0x06CFA2019DD67671ULL, "400254932.ogg"},
	{0x06CFCF86CF383FA4ULL, "vxw_encounter2_rs_0_1.vlm"},
	{0x06CFDE1CC948E0ADULL, "861527240.ogg"},
	{0x06D08CAAE1DF71A5ULL, "ave200_party_comment6116370.cl"},
	{0x06D092A796BBC5EBULL, "ico_dagger_2.dds"},
	{0x06D0B6541455603CULL, "6116163_m.fxe"},
	{0x06D0FA9B6D4F9AE3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_pointing.ani"},
	{0x06D17848D613D507ULL, "vxca_caverns1_mines_ehr_0_0.vlm"},
	{0x06D178B9AC6573FAULL, "492494020.ogg"},
	{0x06D1BD0EE85458FFULL, "vxlt_gallowstemplar_5lt_0_0.vlm"},
	{0x06D1DA5DF4E88DB0ULL, "6150334_m.fxe"},
	{0x06D1F278B4AE0859ULL, "6120783_m.fxe"},
	{0x06D20C2A0CC9A166ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackright.gad"},
	{0x06D2916A0E841746ULL, "vxw_bonepitmining_a2_1ud_0_0.vlm"},
	{0x06D3119AFDF7B621ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_weightshift_gesture_left_hand.ani"},
	{0x06D31AD8B2574BFDULL, "one121_trap6112683.cl"},
	{0x06D3BA9CF94B6605ULL, "6104745_m.fxe"},
	{0x06D41485E531022BULL, "qun220cr_javaris6071953.cl"},
	{0x06D44F71002EB575ULL, "vxlt_hangedman_1eh_0_0.vlm"},
	{0x06D4A34779D02D7FULL, "360481153.ogg"},
	{0x06D4A8843E2A42F3ULL, "vxlt_0_lowtown_eja_0_0.vlm"},
	{0x06D4CF9206249817ULL, "276434610.ogg"},
	{0x06D4D5780A743E53ULL, "hlac_lt_0_lowtown_n_131073_1y.dds"},
	{0x06D537B05935AF12ULL, "hlac_ht_0_hightown_a3_-131073_1y.dds"},
	{0x06D56CFBCB0DCD41ULL, "llac_lt_0_lowtown_docks_a3_n_65534_4.dds"},
	{0x06D5730A8C84118FULL, "hlac_ht_keep_4519_4y.dds"},
	{0x06D59B82F8AB2FEDULL, "406212550.ogg"},
	{0x06D6C88650B00F3BULL, "7732721.ogg"},
	{0x06D6D70907241842ULL, "art\\characters\\creatures\\mubari\\animation\\p_wkf.evt"},
	{0x06D70A7BCD395D00ULL, "vxlt_gallowstemplar_n_395_0_0.vlm"},
	{0x06D73D30CFC119B1ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_moths.mmh"},
	{0x06D7A59E08E32644ULL, "617337283.ogg"},
	{0x06D7A84D75DDE1B5ULL, "snow_02_l2.mao"},
	{0x06D7BBCFDFA8B417ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_railend.mmh"},
	{0x06D7BFCFDFA8BB24ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_railend.mao"},
	{0x06D7D1CFDFA8D989ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_railend.msh"},
	{0x06D7F8286FD7CEE3ULL, "6116303_f.fxe"},
	{0x06D7FEF580F39C0FULL, "vxht_0_hightown_by7_0_0.vlm"},
	{0x06D8072A0CCE0445ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackright.evt"},
	{0x06D80F956A362CEDULL, "seb221_ruxton_harimann6177203.cl"},
	{0x06D8173DA2D49AFDULL, "gen00fl_anders6201156.cl"},
	{0x06D82B7AE452EAC0ULL, "6176288_m.fxe"},
	{0x06D840E1E0E34738ULL, "vxlt_0_lowtown_docks_a3_aos_0_0.vlm"},
	{0x06D87543D883DCD9ULL, "vxlt_undercityserialkiller_1ms_0_0.vlm"},
	{0x06D8938F43CC7591ULL, "vxlt_0_lowtown_a2_eqr_0_0.vlm"},
	{0x06D90D47ACF89D74ULL, "vxlt_0_lowtown_n_gpe_0_0.vlm"},
	{0x06D914EF6A8BFE96ULL, "vxlt_warehouse_n_sy_0_0.vlm"},
	{0x06D9648C7AA91423ULL, "vxlt_undercityserialkiller_1a5_0_0.vlm"},
	{0x06D966444A6F87ADULL, "481016883.ogg"},
	{0x06D98FCA7897F848ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brrckbrd_sa.dds"},
	{0x06D9E58AEACFB1FDULL, "mer140ip_at_caverns.utp"},
	{0x06D9E74687B6536FULL, "44933787.ogg"},
	{0x06D9EB1C0A3D2F85ULL, "6096266_m.fxe"},
	{0x06DA1C9AFDFDE7C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_weightshift_gesture_left_hand.gad"},
	{0x06DB2195B0370BD0ULL, "w_sundermount_228_0.mao"},
	{0x06DB2595B03712E3ULL, "w_sundermount_228_0.mmh"},
	{0x06DB3395B0372A3DULL, "w_sundermount_228_0.msh"},
	{0x06DBB9A1DFA3210BULL, "6131206_m.fxe"},
	{0x06DBE2E7F3C58855ULL, "one100dg_lia6049832.cl"},
	{0x06DCC5275B9E48B6ULL, "vxdr_dungeon_1lt_0_1.vlm"},
	{0x06DCD60907288255ULL, "art\\characters\\creatures\\mubari\\animation\\p_wkf.gad"},
	{0x06DCF25BD9DC5A14ULL, "llac_lt_0_lowtown_docks_-131075_4.dds"},
	{0x06DD5B1172E74842ULL, "vxlt_gallowscourtyard_n_7hz_0_0.vlm"},
	{0x06DD7E2C27A98677ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf05.mmh"},
	{0x06DD942C27A9ABE9ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf05.msh"},
	{0x06DE129B6D5A4DDAULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_pointing.evt"},
	{0x06DEE0A1A94DC315ULL, "fx_fl_smoke_add.mao"},
	{0x06DF24E37A0BDCC0ULL, "mag100_cullen6093547.cl"},
	{0x06DF35365CD5E088ULL, "6180443_m.fxe"},
	{0x06DF4E5872219A3CULL, "6082266_m.fxe"},
	{0x06DFF09FA5113FA7ULL, "6167484_f.fxe"},
	{0x06E02EEBDA4A717FULL, "vxlt_gallowstemplar_5iw_0_0.vlm"},
	{0x06E03A0B75D2EC53ULL, "fen210_fenris6108754.cl"},
	{0x06E0F99AFE03CAA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_weightshift_gesture_left_hand.evt"},
	{0x06E18FFF4B5212B4ULL, "vxht_keep_n_7aj_0_0.vlm"},
	{0x06E1BD3F5FD88081ULL, "6034327_m.fxe"},
	{0x06E23521A3C55992ULL, "vxlt_0_lowtown_a2_n_c_dow_0_0.vlm"},
	{0x06E250AC7AE95D0BULL, "gen00fl_fenris6128066.cl"},
	{0x06E2AC41DA3B7C68ULL, "vxlt_gallowscourtyard_62t_0_0.vlm"},
	{0x06E2F7544FCE717AULL, "glo_vfx_traps.bnk"},
	{0x06E300A72F484E74ULL, "12571673.ogg"},
	{0x06E31AC4A781C230ULL, "vxlt_gallowstemplar_n_38v_0_0.vlm"},
	{0x06E3E8645A254DB3ULL, "vxht_0_hightown_a2_bx0_0_0.vlm"},
	{0x06E3FBBD40E71F08ULL, "35915696.ogg"},
	{0x06E4941A3DA785D8ULL, "vxlt_0_lowtown_docks_a3_d0f_0_1.vlm"},
	{0x06E4A090802F911DULL, "hlac_lt_0_lowtown_docks_-2_4z.dds"},
	{0x06E5119B6D606ACDULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_pointing.gad"},
	{0x06E522FF27769D7BULL, "mag311_huon6042574.cl"},
	{0x06E52C1A67F5131BULL, "lt_0_lowtown_docks_126_0.mmh"},
	{0x06E5381A67F527B8ULL, "lt_0_lowtown_docks_126_0.mao"},
	{0x06E54A1A67F54605ULL, "lt_0_lowtown_docks_126_0.msh"},
	{0x06E57AC046A01A58ULL, "6092230_m.fxe"},
	{0x06E5985E82B67BF2ULL, "840231134.ogg"},
	{0x06E5AE25714138FCULL, "6143284_m.fxe"},
	{0x06E5B874C5633FABULL, "richard_test_st.stg"},
	{0x06E6224298BB6AE6ULL, "llac_lt_0_lowtown_n_1_7.dds"},
	{0x06E6458ECE491DC4ULL, "274953840.ogg"},
	{0x06E6DF88F7959AF9ULL, "root_clu.mao"},
	{0x06E70867D7B9D25DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven_0.pwk"},
	{0x06E722F701C97E76ULL, "vxht_0_hightown_n_5jc_0_0.vlm"},
	{0x06E743D9EEE94A39ULL, "6100042_m.fxe"},
	{0x06E7555F79BD410FULL, "513628503.ogg"},
	{0x06E77A0349A3CE95ULL, "187533396.ogg"},
	{0x06E786CC5C821716ULL, "vxlt_warehouse_n_vz_0_0.vlm"},
	{0x06E812D8245DED50ULL, "vxw_encounter2_camp_qs_0_0.vlm"},
	{0x06E8212A86F9189CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medc_0n.dds"},
	{0x06E85CBFC7434CDDULL, "37159721.ogg"},
	{0x06E892AB47C7D157ULL, "mag160_feynriel6016586.cl"},
	{0x06E8DD38188E04A9ULL, "71683659.ogg"},
	{0x06E91086BBD3AEA4ULL, "vxlt_warehouse_n_i5_0_0.vlm"},
	{0x06E96814BDACEB9DULL, "vxw_sundermount_219_0_0.vlm"},
	{0x06E980280B9ED268ULL, "mag370_meredith6064390.cl"},
	{0x06E9CFBD850E72C2ULL, "6045368_m.fxe"},
	{0x06EA3B810963A1CFULL, "vxlt_0_lowtown_n_d9b_0_0.vlm"},
	{0x06EAFCAD4B10600BULL, "777317973.ogg"},
	{0x06EB1AE4C9EA2B15ULL, "llac_w_encounter2_camp_196612_0.dds"},
	{0x06EB4FFC40A28AC2ULL, "gen00fl_varric6105745.cl"},
	{0x06EB6CC5AA6F82CFULL, "460220665.ogg"},
	{0x06EBEF20FDDCE744ULL, "hlac_w_sundermount_196603_0x.dds"},
	{0x06EC20E3B61BE543ULL, "432429324.ogg"},
	{0x06EC661E54569289ULL, "6025361_m.fxe"},
	{0x06ED49D8B94DF698ULL, "387769264.ogg"},
	{0x06ED5A5E3DFFD47AULL, "6119469_m.fxe"},
	{0x06EDF6BA1FC5C34DULL, "dae300_cullen6140647.cl"},
	{0x06EE35E8EE2FDD70ULL, "t3_dog_d08.tnt"},
	{0x06EE65539F3D6D8AULL, "llac_ht_keep_3733_1.dds"},
	{0x06EF0AFD09AFD6DBULL, "6155631_m.fxe"},
	{0x06EF1DD8C5972033ULL, "vxca_caverns2_8pa_0_0.vlm"},
	{0x06F0811EA8D3F070ULL, "vxht_brothel_n_5k1_0_0.vlm"},
	{0x06F08983F4C7F354ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_leaning_over_table.anb"},
	{0x06F09E1AB8929D00ULL, "w_woundedcoast_215_0.phy"},
	{0x06F12B826BA1D2A2ULL, "one300_zevran6089726.cl"},
	{0x06F1332AE41AE675ULL, "vxlt_hangedman_1ov_0_0.vlm"},
	{0x06F15287DF923A0AULL, "6112988_f.fxe"},
	{0x06F15B324432200CULL, "6083812_m.fxe"},
	{0x06F18B056635ED20ULL, "808370756.ogg"},
	{0x06F1C693FF2867B4ULL, "6107550_m.fxe"},
	{0x06F1F2572E9D15E9ULL, "vxht_mansion2_3eh_0_0.vlm"},
	{0x06F236B883A34000ULL, "mag100_thrask6034812.cl"},
	{0x06F256904DF1E395ULL, "hlac_lt_0_lowtown_-2_3z.dds"},
	{0x06F29A2E1351378BULL, "dre000_varric6034636.cl"},
	{0x06F3656EDFA97909ULL, "766333203.ogg"},
	{0x06F3EC9D389632D8ULL, "6097130_m.fxe"},
	{0x06F488A597DFBCAEULL, "579286869.ogg"},
	{0x06F493274B0D8699ULL, "llac_lt_0_lowtown_65536_5.dds"},
	{0x06F4C073DF123094ULL, "vxw_bonepit_27e_0_0.vlm"},
	{0x06F4C51AA8B3EEE5ULL, "llac_lt_0_lowtown_docks_n_-4_5.dds"},
	{0x06F5370B58404ADAULL, "6180508_m.fxe"},
	{0x06F59C46A4BC39CEULL, "hlac_ht_playershouse_n_855_2y.dds"},
	{0x06F5F10190E4CE74ULL, "vxlt_0_lowtown_n_gjk_0_0.vlm"},
	{0x06F5F970905FDFD2ULL, "6149187_m.fxe"},
	{0x06F633ED7AC0B997ULL, "fex_ruin_arch_b.msh"},
	{0x06F649ED7AC0DF09ULL, "fex_ruin_arch_b.mmh"},
	{0x06F67C874B9B3EB9ULL, "6067855_m.fxe"},
	{0x06F6820A57A55A8DULL, "vxlt_undercitydungeon_145_0_0.vlm"},
	{0x06F684B5165405D0ULL, "vxw_encounter2_camp_lp_0_0.vlm"},
	{0x06F6C7C53BA1B173ULL, "l_isa121ar_chantry.wbd"},
	{0x06F6D43D6C433A7CULL, "6031977_f.fxe"},
	{0x06F6F5E9615711CEULL, "vxht_0_hightown_byz_0_0.vlm"},
	{0x06F7305AD2CE3670ULL, "one100dg_vanard6048252.cl"},
	{0x06F73D09F0EFC683ULL, "6082598_f.fxe"},
	{0x06F78937CE0C151DULL, "mcr310cr_templar_escort.utc"},
	{0x06F7ABBE977D61B4ULL, "vxlt_0_lowtown_docks_a0m_0_0.vlm"},
	{0x06F7BE14770A0BA0ULL, "vxw_bonepitmining_a2_5i_0_0.vlm"},
	{0x06F81DEE335ABDB0ULL, "zz_lgt_3_debug6140427.cl"},
	{0x06F88149F57DF3B0ULL, "351726277.ogg"},
	{0x06F887E76918C709ULL, "70590875.ogg"},
	{0x06F89CC6147BCAADULL, "zz_followers_banter6158958.cl"},
	{0x06F94ECF5A2A75FDULL, "6128072_f.fxe"},
	{0x06F95C68316A5FBAULL, "vxw_bonepitmining_a2_6z_0_0.vlm"},
	{0x06F96010542DC8D3ULL, "hlac_lt_gallowstemplar_n_7198_7x.dds"},
	{0x06F9A1476CC832CFULL, "6147416_m.fxe"},
	{0x06F9A845AD0E6500ULL, "sky_sky2_p1.psh.fxo.dx11.assembly"},
	{0x06FA07E8B20634B0ULL, "zz_pro_debug6007863.cl"},
	{0x06FA37644D8C2363ULL, "gen00fl_isabela6101056.cl"},
	{0x06FA879EE3B17136ULL, "vxlt_0_lowtown_docks_aor_0_0.vlm"},
	{0x06FAB7923F352F26ULL, "vxlt_gallowstemplar_f_5jd_0_0.vlm"},
	{0x06FB02366D6D105DULL, "gen00fl_sebastian6180429.cl"},
	{0x06FB7C403599C80AULL, "6100311_m.fxe"},
	{0x06FBB5225ECDA472ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyn_0d.dds"},
	{0x06FBD5C3DE8EE1D3ULL, "6083626_m.fxe"},
	{0x06FBDD921AF986ADULL, "w_encounter2_61_0.mmh"},
	{0x06FBE1921AF98DFEULL, "w_encounter2_61_0.mao"},
	{0x06FBEF921AF9A533ULL, "w_encounter2_61_0.msh"},
	{0x06FBF3233502A2B6ULL, "seb221_brett_harimann6177249.cl"},
	{0x06FCCDDFABF61E87ULL, "vxca_caverns1_mines_ego_0_0.vlm"},
	{0x06FCE4E50401E9F7ULL, "vxht_mansion2b_389_0_0.vlm"},
	{0x06FD7189E53FA64BULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.mmh"},
	{0x06FD9C017DA1116FULL, "vxlt_0_lowtown_n_byp_0_0.vlm"},
	{0x06FDA10B14CA2E14ULL, "6174242_m.fxe"},
	{0x06FE48D49D09E91BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01flwrplnt.msh"},
	{0x06FE4ACD74B862D2ULL, "rand_underground.gda"},
	{0x06FE5AD49D0A0866ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01flwrplnt.mao"},
	{0x06FE66D49D0A1C05ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01flwrplnt.mmh"},
	{0x06FE90937A84EB61ULL, "vxht_mansion2_1a2_0_0.vlm"},
	{0x06FEF6EC42F9442EULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.mmh"},
	{0x06FF42FD6684B11BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hat_medc_0d.dds"},
	{0x06FFEE33B3654387ULL, "zz_ser_debug6078155.cl"},
	{0x06FFF229CA7BC6D7ULL, "vxlt_0_lowtown_a3_n_c_e3b_0_0.vlm"},
	{0x070060D82D872721ULL, "40988930.ogg"},
	{0x0700CC1E3E117EB3ULL, "vxht_0_hightown_a2_bf2_0_0.vlm"},
	{0x0700ED03054A0668ULL, "fex_ruin_arch_lrg_e.msh"},
	{0x07010B03054A3892ULL, "fex_ruin_arch_lrg_e.mmh"},
	{0x07012274FEC56507ULL, "vxlt_0_lowtown_docks_9fh_0_0.vlm"},
	{0x0701228A1C5E08C0ULL, "6110011_m.fxe"},
	{0x0701A62D3957506FULL, "vxlt_0_lowtown_a2_n_c_e1a_0_0.vlm"},
	{0x0701CD25773DBAF7ULL, "vxht_store_bt_0_0.vlm"},
	{0x07023AD45F4C4F84ULL, "vxht_keep_5ga_0_0.vlm"},
	{0x0702C6E8109CFA1EULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack2_c.impa.ani"},
	{0x07030DA0737CC0ABULL, "vxca_cavern1_haunted_gjz_0_0.vlm"},
	{0x070320CA36AD9E21ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast.ani"},
	{0x07033250D1084914ULL, "vxw_bonepit_22j_0_0.vlm"},
	{0x0703E46430000C0EULL, "vxlt_warehouse_w8_0_0.vlm"},
	{0x0703E87BC027F1CAULL, "llac_dr_primevalentrance_1726_13.dds"},
	{0x070456421CF4DDD8ULL, "vxlt_0_lowtown_docks_n_coa_0_0.vlm"},
	{0x0704AB407DB6ADB3ULL, "6082096_m.fxe"},
	{0x0704C8313ACC6391ULL, "6156452_m.fxe"},
	{0x0704F692084A1E50ULL, "vxw_encounter2_camp_gq_0_0.vlm"},
	{0x070512FB8C6E22E8ULL, "lt_0_lowtown_a2_n_c_52_0_ll.mmh"},
	{0x070585BCBE860B53ULL, "6065248_m.fxe"},
	{0x07058782DE3B4B47ULL, "fen110_fenris6100249.cl"},
	{0x0705DD921B01DB26ULL, "w_encounter2_61_0.rml"},
	{0x0705EA518304AB2DULL, "6046887_m.fxe"},
	{0x07064453E5561D0FULL, "hlac_w_bonepitdestroyed_a3_458754_0x.dds"},
	{0x070662AAC490513EULL, "gen00fl_sebastian6174926.cl"},
	{0x0707C4077C6ADCD5ULL, "one100ip_st_solivitus.utp"},
	{0x070811544FEDB918ULL, "glo_vfx_traps.wbd"},
	{0x070819493BD99748ULL, "vxht_mansion2_1a7_0_0.vlm"},
	{0x0708274561BCE89FULL, "llac_lt_foundry_n_320_11.dds"},
	{0x070829F985F6158BULL, "12089447.ogg"},
	{0x070839C55B870D97ULL, "hlac_lt_0_lowtown_a2_n_c_131072_4y.dds"},
	{0x07087F225C0FA46BULL, "vxw_bonepitdestroyed_a3_232_0_0.vlm"},
	{0x0708CCB85E14FA7FULL, "vxht_playershouse_n_1su_0_1.vlm"},
	{0x07090A02C89F3E5AULL, "369209083.ogg"},
	{0x0709FF07E399BDD8ULL, "153220355.ogg"},
	{0x070A2BCA36B3CFC3ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast.gad"},
	{0x070A3D37E389FC52ULL, "hlac_ca_cavern1_haunted_24424_0z.dds"},
	{0x070A47A0D084FA53ULL, "6086926_m.fxe"},
	{0x070ABCCF825FB589ULL, "20331479.ogg"},
	{0x070B28991FA2AFCFULL, "vxca_caverns2_6z9_0_0.vlm"},
	{0x070B5289E54BAEB0ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.anb"},
	{0x070BBBA933D9E625ULL, "82473490.ogg"},
	{0x070BC0D5725D214CULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyn_0d.dds"},
	{0x070BD19804C8B53CULL, "1002504282.ogg"},
	{0x070C0192A9835133ULL, "vxca_caverns2_8zc_0_0.vlm"},
	{0x070C55FCA9BBA910ULL, "662837619.ogg"},
	{0x070CD6921B07EE6AULL, "w_encounter2_61_0.phy"},
	{0x070CF8B3B853A73CULL, "ave310_donnic6134721.cl"},
	{0x070D02706C7B03E1ULL, "vxht_mansion2_1d1_0_0.vlm"},
	{0x070D15EC4305B62DULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.anb"},
	{0x070DEC928ACB2E2DULL, "6169366_m.fxe"},
	{0x070E769067AB08FDULL, "214762027.ogg"},
	{0x070EA2371155DA07ULL, "89334860.ogg"},
	{0x070EDF8AAA9DBA2BULL, "vxlt_darktown_6id_0_0.vlm"},
	{0x070F0EF06E6B2A2EULL, "6162052_m.fxe"},
	{0x070F1F1BCF78A52FULL, "vxca_caverns2_805_0_0.vlm"},
	{0x070F21A5991B9A39ULL, "vxw_bonepitmining_a2_do_0_0.vlm"},
	{0x070F29B154B76298ULL, "215561819.ogg"},
	{0x070F44D8E3DE8259ULL, "follower_banter6171070.cl"},
	{0x070FC873F479549CULL, "vxht_chantry_30v_0_0.vlm"},
	{0x070FE259384A9A94ULL, "6149893_m.fxe"},
	{0x0710ACB151426804ULL, "vxht_keep_5jf_0_0.vlm"},
	{0x0710D8D38DC13258ULL, "lt_0_lowtown_docks_a3_n_3_0_ll.mmh"},
	{0x0710DEE810A960EFULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack2_c.impa.evt"},
	{0x071108CA36B9B2A8ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast.evt"},
	{0x071187B43D71E3F3ULL, "vxht_mansion_n_6p7_0_0.vlm"},
	{0x071219E05286B19AULL, "w_sundermount_160_0.msh"},
	{0x071227E05286C9D3ULL, "w_sundermount_160_0.mao"},
	{0x07122BE05286D0E0ULL, "w_sundermount_160_0.mmh"},
	{0x07123CB7A05C3ED9ULL, "vxlt_undercityserialkiller_1qw_0_0.vlm"},
	{0x0712564121F6248EULL, "vxlt_warehouse_x9_0_0.vlm"},
	{0x0712A4D00B51A0FDULL, "vxdr_primevaltomb_91e_0_2.vlm"},
	{0x0713DC34F10F5EA3ULL, "477487556.ogg"},
	{0x0714631E84AD6CFBULL, "fx_c_woodgrain.mao"},
	{0x0714661D91B5F824ULL, "vxlt_warehouse_n_v8_0_0.vlm"},
	{0x07149357D7A9777BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_exit.evt"},
	{0x0715157C2AF67768ULL, "vxca_caverns1_i1g_0_0.vlm"},
	{0x071566720D9496ECULL, "tow_smoke.mao"},
	{0x0715BB1AB8B1EA95ULL, "w_woundedcoast_215_0.msh"},
	{0x0715BD1AB8B1EE0BULL, "w_woundedcoast_215_0.mmh"},
	{0x0715C91AB8B202A8ULL, "w_woundedcoast_215_0.mao"},
	{0x0715D61FA8EE5710ULL, "mag210_arianni6024568.cl"},
	{0x07161A6140CE569CULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die_loop.ani"},
	{0x07164022155690BAULL, "vxw_bonepitmining_a2_8t_0_0.vlm"},
	{0x0716708AD7F8F1A7ULL, "hlac_lt_0_lowtown_docks_-131074_4z.dds"},
	{0x07168B262DCFAFC8ULL, "vxht_mansion2_1d4_0_0.vlm"},
	{0x0716DFE810ADCD98ULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack2_c.impa.gad"},
	{0x0716E7D6B93D5C63ULL, "follower_banter6172716.cl"},
	{0x0717E6EB23876168ULL, "qun111_ketojan6019046.cl"},
	{0x07186727274D67C8ULL, "145548216.ogg"},
	{0x0718A4E7A8FCFB78ULL, "440931876.ogg"},
	{0x0718C8B8E9506B3DULL, "6051641_m.fxe"},
	{0x071926CA8441D50AULL, "6062806_m.fxe"},
	{0x07192C007C1BA687ULL, "vxlt_0_lowtown_n_ci5_0_0.vlm"},
	{0x071931D33E7932E2ULL, "6188622_f.fxe"},
	{0x07194BE00DCF71E8ULL, "6180883_m.fxe"},
	{0x071976EAA2E3547FULL, "852584759.ogg"},
	{0x0719C7DAEDA9CEEAULL, "vxlt_shop_j0_0_0.vlm"},
	{0x0719ED625EBB63DDULL, "al_varric_s01.gda"},
	{0x071B5A8C77A9C210ULL, "vxw_encounter3_gb_0_0.vlm"},
	{0x071B8C57D7AF89BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_exit.gad"},
	{0x071B9E3B25436CA2ULL, "l_ser211ar_foundry.bnk"},
	{0x071BB6835ECDFC59ULL, "follower_barks6173417.cl"},
	{0x071BC34CF5A7D29EULL, "6135896_m.fxe"},
	{0x071C08F645E8D90DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_alcovetop.phy"},
	{0x071C5B9EA59A5CB1ULL, "zz_mrl_debug6184194.cl"},
	{0x071CCA4F90251023ULL, "6110786_m.fxe"},
	{0x071D17973300DD21ULL, "vxlt_0_lowtown_n_e0n_0_0.vlm"},
	{0x071D23801873FE40ULL, "mage_firemastery.dds"},
	{0x071D276140D48BDEULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die_loop.gad"},
	{0x071D58431DE2279CULL, "648819365.ogg"},
	{0x071D80CD606142AEULL, "6167352_m.fxe"},
	{0x071D93828B11B2B9ULL, "vxw_bonepitmining_a2_cn_0_0.vlm"},
	{0x071DA10A57536A6DULL, "6083053_m.fxe"},
	{0x071DAFD821FDAFB3ULL, "vxht_0_hightown_a2_bl4_0_0.vlm"},
	{0x071DD19164F5499CULL, "56354090.ogg"},
	{0x071DDC79ED2E4091ULL, "6140983_m.fxe"},
	{0x071E54362A40B314ULL, "6142993_m.fxe"},
	{0x071EC0D8971AC8F7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channelwater.mmh"},
	{0x071EC4D8971ACF04ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channelwater.mao"},
	{0x071ED6D8971AEE69ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channelwater.msh"},
	{0x071EDD5F8653B5D0ULL, "19347031.ogg"},
	{0x071FABA001E97F71ULL, "vxlt_0_lowtown_n_ey1_0_0.vlm"},
	{0x071FDD1FC11A3FBAULL, "gen00fl_fenris6108085.cl"},
	{0x071FF586EB099D7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_drink.evt"},
	{0x07205BB0290B7DE1ULL, "qun120_arishok6051466.cl"},
	{0x0720CCDFFD9AF011ULL, "vxlt_0_lowtown_a2_ehu_0_0.vlm"},
	{0x07213068B53A1AADULL, "316495243.ogg"},
	{0x07213D9347A82173ULL, "411377213.ogg"},
	{0x07228B57D7B5A6BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_exit.ani"},
	{0x0722AFEFBB47D14AULL, "lt_0_lowtown_docks_n_237_0.mao"},
	{0x0722B3EFBB47D799ULL, "lt_0_lowtown_docks_n_237_0.mmh"},
	{0x0722BDEFBB47E887ULL, "lt_0_lowtown_docks_n_237_0.msh"},
	{0x0722EEC1585491C9ULL, "zz_soundset.ncs"},
	{0x0722F0F749EECB2DULL, "603291185.ogg"},
	{0x07237A2C8A297303ULL, "6135707_m.fxe"},
	{0x0723D7F28B4A84C8ULL, "vxlt_gallowscourtyard_n_46q_0_0.vlm"},
	{0x0724026140DA6BFDULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die_loop.evt"},
	{0x072511DC1EDD89AFULL, "lt_gallowscourtyard_48_0.rml"},
	{0x07258E3DA77FB253ULL, "water_water_p0.psh.fxo.dx11"},
	{0x07259D894CECF87AULL, "775797314.ogg"},
	{0x0725A708218B70E8ULL, "ht_0_hightown_n_41_0.rml"},
	{0x0727087CC27E6D89ULL, "6174966_m.fxe"},
	{0x07270C86EB0FE2A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_drink.gad"},
	{0x07279B0EB1F6B126ULL, "ico_attack.dds"},
	{0x07279B8016613755ULL, "6048385_f.fxe"},
	{0x072805D54EE98515ULL, "zz_stand_neutral_test.cut"},
	{0x07291A7BBF7AE60BULL, "274920584.ogg"},
	{0x072992495B0F77C2ULL, "genip_cin_crown.utp"},
	{0x0729BC379173B035ULL, "667288383.ogg"},
	{0x0729E70645134C43ULL, "722414801.ogg"},
	{0x072A4C01FE078B4AULL, "hlac_ca_caverns1_mines_24424_19x.dds"},
	{0x072A524724158DD3ULL, "pro000_aveline_intro6109378.cl"},
	{0x072A6C6B15035977ULL, "ca_caverns2prc_5802_lc.mmh"},
	{0x072A8F0A8D5B6603ULL, "dae000cr_grand_cleric6148160.cl"},
	{0x072AD04C6EB75F2DULL, "6163166_m.fxe"},
	{0x072B407F3D622AB3ULL, "6109634_m.fxe"},
	{0x072B5C2E66D156D5ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_cast_aoe_c.mmh"},
	{0x072B6B938FCAF706ULL, "one111ip_ggt_crate_2.utp"},
	{0x072BE054989F67DDULL, "6031816_m.fxe"},
	{0x072BFD64D84E723EULL, "686881272.ogg"},
	{0x072C01F20C0E7C9BULL, "hlac_lt_0_lowtown_docks_n_131071_0y.dds"},
	{0x072C055F7D07CB39ULL, "vxw_bonepitmining_a2_nm_0_0.vlm"},
	{0x072C904D5C3C0AEBULL, "vxlt_0_lowtown_f1o_0_0.vlm"},
	{0x072D1F342CE3D733ULL, "324947462.ogg"},
	{0x072D6D9DEBFE128DULL, "6110612_m.fxe"},
	{0x072D8BC7562A6BC5ULL, "825657466.ogg"},
	{0x072DDD86EB15B1F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_drink.ani"},
	{0x072DECA54C9FCB64ULL, "601123951.ogg"},
	{0x072E0453AE903E07ULL, "gen00fl_fenris6113950.cl"},
	{0x072E0AFBFB9127EEULL, "hlac_w_encounter2_camp_327679_0z.dds"},
	{0x072E57177938D44AULL, "877899518.ogg"},
	{0x072F9847159074A6ULL, "6113632_m.fxe"},
	{0x07304D7FBE7FD7CCULL, "ave310_brennan6136872.cl"},
	{0x07309FA9507D8B15ULL, "ss_qunari_arishok.bnk"},
	{0x0730BBFA1DC86458ULL, "dae000ip_crowd_ff6_talk16.utp"},
	{0x0730C0CE472E5F93ULL, "6016952_m.fxe"},
	{0x0730F425D2659045ULL, "hlac_dr_0_deeproads_5849_3x.dds"},
	{0x07317C1DB7E0325CULL, "vxlt_gallowstemplar_f_3nd_0_0.vlm"},
	{0x07318A78C090FD33ULL, "vxlt_gallowstemplar_5pl_0_0.vlm"},
	{0x0731B344079E94DAULL, "vxca_cave_23a_0_1.vlm"},
	{0x0731C01B0B1ECC03ULL, "vxlt_gallowstemplar_f_6qk_0_0.vlm"},
	{0x07320CC02FC65EA5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_basic_1.mao"},
	{0x07335EC1C5061989ULL, "359329134.ogg"},
	{0x0733B011EFAACB42ULL, "6125616_f.fxe"},
	{0x0733CA5A9B579C9EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v5.ani"},
	{0x0733D6BF4E7B6296ULL, "vis121_saemus6101339.cl"},
	{0x07340F70834F8EEDULL, "llac_lt_0_lowtown_docks_a3_n_-3_1.dds"},
	{0x07346DE191B8EAE2ULL, "812832990.ogg"},
	{0x07347E71616C22E3ULL, "lgt120cr_bhelen_supp_war.utc"},
	{0x0734DB6E7701A1F0ULL, "hf_amb_keep_servant.mor"},
	{0x0735BF83A5B71610ULL, "416701605.ogg"},
	{0x0735E1E59FC8580AULL, "605264988.ogg"},
	{0x0735FE24D663BBACULL, "175455843.ogg"},
	{0x0736570752E7F990ULL, "vxht_mansion2_32h_0_0.vlm"},
	{0x0736C2FB43FBD70DULL, "vxca_caverns1_ckn_0_0.vlm"},
	{0x0736DE92D449DE80ULL, "vxlt_0_lowtown_n_d46_0_0.vlm"},
	{0x07375013BEACE496ULL, "231099.ogg"},
	{0x07376BBCAF6C2197ULL, "36056791.ogg"},
	{0x073785709B903232ULL, "187668956.ogg"},
	{0x0737E2CD129A4835ULL, "genip_mug.utp"},
	{0x0738592E66DBDC62ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_cast_aoe_c.anb"},
	{0x0738B36640CC0440ULL, "follower_banter6162107.cl"},
	{0x073A5364A7438E91ULL, "841588050.ogg"},
	{0x073A5AE299C667F1ULL, "402073764.ogg"},
	{0x073A605C03EBF375ULL, "vxca_caverns1_j4n_0_0.vlm"},
	{0x073A6665F4FDD415ULL, "6177909_m.fxe"},
	{0x073A84DD4A4A0064ULL, "vxht_mansion2b_2vu_0_0.vlm"},
	{0x073AA76A268935B4ULL, "398275833.ogg"},
	{0x073AB8A0C685603EULL, "663420472.ogg"},
	{0x073AD4A09C850278ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_setprops.phy"},
	{0x073AE1D060891B59ULL, "588140410.ogg"},
	{0x073B0697747D6E7AULL, "240747354.ogg"},
	{0x073BA2F7A4999A64ULL, "6071435_m.fxe"},
	{0x073C22AAA233B981ULL, "follower_banter6169840.cl"},
	{0x073C48EE87040CA9ULL, "sky_sky2_p0.psh.fxo.dx11.assembly"},
	{0x073C7B5247EB958AULL, "vxca_caverns1_mines_j5c_0_0.vlm"},
	{0x073D1A47FBC67368ULL, "nat380_sibling6207554.cl"},
	{0x073D1AD01C8890E0ULL, "gen_im_trs_dal_hvy_glv_r1.uti"},
	{0x073D97E8A6842C4AULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.impa.gad"},
	{0x073DA6784380E4BBULL, "461785518.ogg"},
	{0x073DCB9FB5FF20DCULL, "vxlt_gallowstemplar_n_39i_0_0.vlm"},
	{0x073DFF46A68D7C2CULL, "ter_cracked01_s.dds"},
	{0x073E22D96EA58B75ULL, "627804293.ogg"},
	{0x073E4AABCDCCB36AULL, "299208642.ogg"},
	{0x073F0C5ED4A22D18ULL, "vxw_blightlands_n5_0_0.vlm"},
	{0x073F120BD6F6A49FULL, "318458356.ogg"},
	{0x073F204D537CDE99ULL, "313824710.ogg"},
	{0x073F5ABB9A98F322ULL, "zz_qun_debug6049745.cl"},
	{0x073F62D1601DE5BBULL, "vxdr_primevalentrance_392_0_1.vlm"},
	{0x073FC74718E1C0DFULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkfront.gad"},
	{0x07402E2CF2B7CFCBULL, "312233193.ogg"},
	{0x07404C8171D3B6E9ULL, "6096532_m.fxe"},
	{0x074056E647C9EF5DULL, "ave200_donnic6112144.cl"},
	{0x0740840DD9EBC583ULL, "6060540_m.fxe"},
	{0x0741883B256416C0ULL, "l_ser211ar_foundry.wbd"},
	{0x07419ECF1901095DULL, "391320634.ogg"},
	{0x0741A411CFA7141CULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.dura.ani"},
	{0x0741C37122DC9A2AULL, "vxw_bonepitmining_a2_1xo_0_0.vlm"},
	{0x0741E25A9B64036FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v5.evt"},
	{0x0742177F366D457FULL, "zz_dae_debug6143033.cl"},
	{0x07422937D211A43FULL, "59713491.ogg"},
	{0x07422D32D055EFCFULL, "fenris_2.cut"},
	{0x07423258597027A5ULL, "132602425.ogg"},
	{0x074269783E6D5B7FULL, "hlac_ht_0_hightown_n_131074_0y.dds"},
	{0x07426BEE1C45BB10ULL, "vxlt_warehouse_sz_0_0.vlm"},
	{0x0742978D8FD3C842ULL, "vxw_encounter3_du_0_0.vlm"},
	{0x0742C70263D871C6ULL, "46807652.ogg"},
	{0x07437095770065C2ULL, "vxca_caverns1_ik3_0_0.vlm"},
	{0x0743AAD9FB76D179ULL, "6136238_m.fxe"},
	{0x0743D7F74E0008FAULL, "hlac_lt_0_lowtown_docks_a3_131068_0x.dds"},
	{0x07443166F782A34CULL, "dae000_pc_home_ambients6208646.cl"},
	{0x07447AE8A68A19E1ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.impa.evt"},
	{0x0744D2296AD84AE7ULL, "vxlt_0_lowtown_byo_0_0.vlm"},
	{0x0744E0E011D412DCULL, "460914563.ogg"},
	{0x0744F27BD79A22E4ULL, "1004138122.ogg"},
	{0x07451EB6D2D319D0ULL, "833616524.ogg"},
	{0x07452C851EB3D821ULL, "42800917.ogg"},
	{0x074581BABF46F1CBULL, "419519299.ogg"},
	{0x0745E09BDC5A3716ULL, "hlac_lt_0_lowtown_a3_n_c_65534_0x.dds"},
	{0x0745F963B960FB35ULL, "vxht_0_hightown_a3_8pu_0_0.vlm"},
	{0x074607B275FF61B3ULL, "vxlt_0_lowtown_a2_hg7_0_0.vlm"},
	{0x07462DEC22F30BC3ULL, "gen00fl_varric6076228.cl"},
	{0x07469EA52243C258ULL, "vxlt_0_lowtown_a2_n_gmr_0_0.vlm"},
	{0x0746BC4718E7CCCCULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkfront.evt"},
	{0x07472E13FD00A7B0ULL, "zz_fol_debug6066136.cl"},
	{0x07478EF7C60BB5D3ULL, "hlac_lt_warehouse_1_4y.dds"},
	{0x0747E35A9B687018ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v5.gad"},
	{0x0747E6D2B9AC6375ULL, "mag170cr_thrask.utc"},
	{0x0747E6E4E0D44B57ULL, "vxdr_primevaltomb_5bl_0_0.vlm"},
	{0x074828499DE7269AULL, "6106960_m.fxe"},
	{0x07482925226E82DFULL, "mer320_hubert6083763.cl"},
	{0x074848E77FCD4CF9ULL, "vxht_keep_n_55i_0_0.vlm"},
	{0x0748B111CFAD495EULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.dura.gad"},
	{0x0748CED930C82BC3ULL, "6013313_m.fxe"},
	{0x0748E2843FB2ABB9ULL, "hlac_lt_gallowscourtyard_-65537_4x.dds"},
	{0x0749432FDE5CDE7CULL, "gen_im_wep_war_two_act1_04.uti"},
	{0x074971B17CF96DDAULL, "vxht_0_hightown_n_c7j_0_0.vlm"},
	{0x07499AC4685464C2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleftb.phy"},
	{0x074A668D420D4CD6ULL, "llac_lt_0_lowtown_docks_a3_n_196607_3.dds"},
	{0x074C3D7CA7F5395CULL, "vxlt_gallowstemplar_n_34n_0_0.vlm"},
	{0x074D0F831EE1A29DULL, "vxw_encounter2_9j_0_0.vlm"},
	{0x074D41FBB844B6D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_wary.evt"},
	{0x074D5AD5499D56D8ULL, "dae000_bartrand6031162.cl"},
	{0x074D666134B2E90EULL, "gen00fl_merrill6151102.cl"},
	{0x074D7E3BC6984598ULL, "vxw_blightlands_s2_0_0.vlm"},
	{0x074DAD47F2754703ULL, "qun120_arishok_guard6054843.cl"},
	{0x074E3B9D20719C98ULL, "670987113.ogg"},
	{0x074E96609316EB6BULL, "vxdr_primevaltomb_902_0_0.vlm"},
	{0x074EF18CDB16536DULL, "vxlt_0_lowtown_docks_a3_n_cym_0_0.vlm"},
	{0x074EF2397FECFBDDULL, "321024058.ogg"},
	{0x074F011C99A09115ULL, "746790619.ogg"},
	{0x074F49EA4EDFE75FULL, "183784624.ogg"},
	{0x074F634D06690554ULL, "6083086_f.fxe"},
	{0x074F6A71F6037D42ULL, "vxlt_0_lowtown_a2_e8j_0_0.vlm"},
	{0x074F83603D56586AULL, "11256007.ogg"},
	{0x074F8C11CFB3297DULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.dura.evt"},
	{0x074FB23510EB18ECULL, "hlac_lt_warehouse_65537_3z.dds"},
	{0x075018390E987FBCULL, "vrc221st_varric_postmans.stg"},
	{0x0750354E14D2B2AAULL, "vxw_bonepitmining_a2_1uh_0_0.vlm"},
	{0x07506266BFF63304ULL, "vxlt_blackemporium_rt_0_1.vlm"},
	{0x0750A2F2A51922B0ULL, "gen00fl_anders6119524.cl"},
	{0x0750E8649B4B63E1ULL, "vxlt_gallowscourtyard_4uz_0_0.vlm"},
	{0x0751096A81C9E0C2ULL, "vxw_encounter3_av_0_0.vlm"},
	{0x07513C88BEF452CDULL, "vxlt_undercityencounter_10t_0_0.vlm"},
	{0x075182B2EB45D5D4ULL, "vxlt_0_lowtown_haf_0_0.vlm"},
	{0x0751B7127AF385C3ULL, "13880782.ogg"},
	{0x0751FE8B92B05603ULL, "6117664_m.fxe"},
	{0x075272E8A6964930ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.impa.ani"},
	{0x07527756B8A351C1ULL, "dre000_sibling_has_blight6037032.cl"},
	{0x0752CD6882200E3EULL, "vxw_woundedcoast_1bx_0_0.vlm"},
	{0x0752D0075CD96110ULL, "and230tr_act2_opening.utt"},
	{0x0753A44718F22E25ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkfront.ani"},
	{0x0753D7DB62427FCEULL, "hlac_w_bonepitdestroyed_a3_393218_0z.dds"},
	{0x0753DA29CD3A13FCULL, "6129727_m.fxe"},
	{0x0753E62D5F5A3381ULL, "vxlt_gallowstemplar_f_3n5_0_0.vlm"},
	{0x07543AFBB84AC914ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_wary.gad"},
	{0x075510821439DAD8ULL, "vxlt_0_lowtown_a2_n_gjs_0_0.vlm"},
	{0x075594DFEA45A7C0ULL, "hlac_lt_undercitydungeon_-1_3x.dds"},
	{0x0755BBAF173188D3ULL, "201408721.ogg"},
	{0x07562082EE59CF64ULL, "mcr310_argument6182285.cl"},
	{0x075656A689B45F31ULL, "mcr310_argument6075135.cl"},
	{0x07576BBBEC719A41ULL, "vxht_0_hightown_c3k_0_1.vlm"},
	{0x075835C9E70522C4ULL, "383844884.ogg"},
	{0x07583FD270F1EB5DULL, "vxlt_gallowscourtyard_7c3_0_0.vlm"},
	{0x0758B3ED606640BAULL, "zz_pro_cutscenes6005832.cl"},
	{0x0758D68B864D698BULL, "gen_im_trs_gem_ruby_01.uti"},
	{0x07593975BDD23101ULL, "hurlock_lsword.utc"},
	{0x075955D58EC2E6C1ULL, "vxlt_undercityserialkiller_1lh_0_0.vlm"},
	{0x07596989B91B4CF4ULL, "vxca_caverns2prc_7oa_0_0.vlm"},
	{0x075995D824C7388FULL, "dae200_cullen6140309.cl"},
	{0x0759A18BD1F8CB44ULL, "481034042.ogg"},
	{0x0759DE80440353E0ULL, "vxlt_hangedman_n_27o_0_0.vlm"},
	{0x0759E7698C1953A0ULL, "gen00fl_dog6165706.cl"},
	{0x075A0A8E23546994ULL, "6135656_m.fxe"},
	{0x075A29AEE1F4855FULL, "hlac_ht_mansion2b_818_56x.dds"},
	{0x075A3C08131A1100ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_007a_0s.dds"},
	{0x075A757C0C619152ULL, "6105644_m.fxe"},
	{0x075B1A7C5AE5641FULL, "259813274.ogg"},
	{0x075B491D3A149DC9ULL, "vxht_chantry_n_216_0_0.vlm"},
	{0x075B49FBB85101F2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_wary.ani"},
	{0x075BC716F77E5036ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_103a_n.dds"},
	{0x075BF018B88E5E18ULL, "vxw_blightlands_t3_0_0.vlm"},
	{0x075CC2705714CB4CULL, "mrl250_pol6101318.cl"},
	{0x075CC5077CB31E24ULL, "one100ip_st_solivitus.ncs"},
	{0x075CC5A950A3069FULL, "ss_qunari_arishok.wbd"},
	{0x075CD33A1A2AC12BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_reaver.evt"},
	{0x075CD5F2AC0DAF92ULL, "316391160.ogg"},
	{0x075CD6966D10BD37ULL, "mag291_demon_marethari6027865.cl"},
	{0x075CF708A69B8463ULL, "zz_dae_debug6029311.cl"},
	{0x075D7CF7852AF7D0ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_act_c.mmh"},
	{0x075EA72B06C8CB2AULL, "vxw_bonepitmining_a2_1ri_0_0.vlm"},
	{0x075EAC3DF50A19B9ULL, "147438291.ogg"},
	{0x075ECFF029CDA1C5ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastwood_sa.dds"},
	{0x075F2BAFD4184008ULL, "6135670_m.fxe"},
	{0x075F4FA80031EC10ULL, "vxlt_warehouse_yx_0_0.vlm"},
	{0x076004AE731DFDCEULL, "hlac_lt_darktown_-1_24x.dds"},
	{0x076035AF5D792201ULL, "vxlt_gallowstemplar_n_398_0_0.vlm"},
	{0x076064C4D1EBAE5CULL, "mag320pt_otl_emile.ncs"},
	{0x0760706884A66E83ULL, "6112564_m.fxe"},
	{0x076070B3A8B749AEULL, "128912203.ogg"},
	{0x0760A323567FB6A2ULL, "ser270_emeric6111400.cl"},
	{0x07613F45741626BEULL, "vxw_woundedcoast_1my_0_0.vlm"},
	{0x0761968875139812ULL, "vxlt_0_lowtown_docks_n_b64_0_1.vlm"},
	{0x0761A0F5620411ADULL, "380873923.ogg"},
	{0x0761C8ACC0D88A3AULL, "hlac_ht_playershouse_n_855_12x.dds"},
	{0x0761DD504ADFD2BCULL, "23656240.ogg"},
	{0x0762832C902880AEULL, "vxca_cavern1_haunted_j2c_0_1.vlm"},
	{0x07629E240CDF7D6DULL, "llac_lt_darktown_-1_32.dds"},
	{0x0762E24ADE526F0FULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_04skeleton.mao"},
	{0x0762EE4ADE5283ACULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_04skeleton.mmh"},
	{0x0762F44ADE528D9EULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_04skeleton.msh"},
	{0x076316149AC59B3BULL, "vxlt_0_lowtown_a2_n_ep0_0_0.vlm"},
	{0x07636D171E3C8D85ULL, "vxdr_primevaltomb_8q0_0_2.vlm"},
	{0x07637E1967DFAF89ULL, "6025112_m.fxe"},
	{0x0763A0EA6111255FULL, "430702140.ogg"},
	{0x0763EC3A1A310AECULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_reaver.gad"},
	{0x076410063884B470ULL, "vxht_brothel_n_539_0_0.vlm"},
	{0x076443E37403CD34ULL, "gen00pt_skills.ncs"},
	{0x0765B19AF6CC08E5ULL, "darktown_male_sickly_sta6175324.cl"},
	{0x0765BB77A5B50F9DULL, "hlac_lt_0_lowtown_a2_n_-131071_1x.dds"},
	{0x0765DA74202C49E4ULL, "vxht_mansion2b_2ix_0_0.vlm"},
	{0x0765DEF5D0425D63ULL, "vxlt_0_lowtown_docks_n_c4o_0_0.vlm"},
	{0x07667B009A180031ULL, "vxht_0_hightown_n_a4n_0_0.vlm"},
	{0x07667CE2A6A459B8ULL, "6205706_m.fxe"},
	{0x076697A1774C7142ULL, "vxht_0_hightown_n_bzl_0_0.vlm"},
	{0x0766B0DB4132D3A8ULL, "vxw_encounter2_camp_93_0_0.vlm"},
	{0x0766B1AF62E803DDULL, "vxlt_gallowscourtyard_7d2_0_0.vlm"},
	{0x0766BDD632DE1EEDULL, "hlac_ca_caverns1_mines_24426_9z.dds"},
	{0x0767687DC6FBD369ULL, "vxw_encounter3_bh_0_0.vlm"},
	{0x0767C7B280B8FF41ULL, "vxlt_undercityserialkiller_1ao_0_0.vlm"},
	{0x0767E26C4BA1783BULL, "6144350_m.fxe"},
	{0x0767F60ABA6BBEF4ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.anb"},
	{0x0768105DF897A8CBULL, "hlac_w_sundermount_0_4z.dds"},
	{0x0768505D35F96C60ULL, "vxlt_hangedman_n_22l_0_0.vlm"},
	{0x076890C4AF5AAF7FULL, "blank_exterior.arl"},
	{0x07689988BAA3287EULL, "ser270_emeric6201574.cl"},
	{0x076912F83637B60CULL, "dae000_gamlen6124239.cl"},
	{0x0769761248D445D6ULL, "hlac_lt_0_lowtown_a3_n_c_65533_1z.dds"},
	{0x0769A9E5600DE8D3ULL, "640346969.ogg"},
	{0x076A56E25A51A36FULL, "vxht_chantry_n_2i8_0_0.vlm"},
	{0x076A6BBA155AF4BDULL, "vxht_0_hightown_bfp_0_0.vlm"},
	{0x076A750FC1B6C13DULL, "vxca_caverns2_8it_0_0.vlm"},
	{0x076A7F95EAD86A32ULL, "358427009.ogg"},
	{0x076A8210A696C5E7ULL, "ser270_templar6127404.cl"},
	{0x076A9FF78535BDEFULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_act_c.anb"},
	{0x076AEB3A1A3727EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_reaver.ani"},
	{0x076B6EBC15AD346FULL, "306628972.ogg"},
	{0x076C6994BF611C91ULL, "6037591_m.fxe"},
	{0x076CA7E5595EE323ULL, "hlac_lt_0_lowtown_a2_n_c_0_9z.dds"},
	{0x076D534DF2883568ULL, "6074493_f.fxe"},
	{0x076D955CE05E8FEAULL, "vxht_0_hightown_a3_bxl_0_0.vlm"},
	{0x076E0F8E3EA51B51ULL, "6059007_m.fxe"},
	{0x076E75C468734276ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleftb.mao"},
	{0x076E81C468735615ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleftb.mmh"},
	{0x076E83C46873598BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleftb.msh"},
	{0x076F18A3BAB3389DULL, "297301126.ogg"},
	{0x076F6A61DADA6896ULL, "6054551_m.fxe"},
	{0x076F79D0C3D4A8E1ULL, "dae000_bartrand6207005.cl"},
	{0x076FB122660C3F3EULL, "vxw_woundedcoast_1hz_0_0.vlm"},
	{0x0770B75F41E22962ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_galowint.msh"},
	{0x0770B95F41E22C18ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_galowint.mmh"},
	{0x0770C55F41E240BBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_galowint.mao"},
	{0x07715BD512DD1E98ULL, "fen110_fenris6117929.cl"},
	{0x07716992F1471778ULL, "vxlt_0_lowtown_a2_n_c_hjn_0_0.vlm"},
	{0x077187F18CBBB3BBULL, "vxlt_0_lowtown_a2_n_eu3_0_0.vlm"},
	{0x07719A4A763F9E0DULL, "vxw_encounter2_cl_0_0.vlm"},
	{0x0771A652E6584D32ULL, "vxht_mansion2_1v4_0_0.vlm"},
	{0x0771FD0272B03829ULL, "dae000cr_fereldan_f.utc"},
	{0x0772841924A1940CULL, "594841100.ogg"},
	{0x077364C744B37950ULL, "follower_banter6151350.cl"},
	{0x077368423E86E6C9ULL, "558430963.ogg"},
	{0x0773C164EE2BC005ULL, "hlac_lt_0_lowtown_a2_131073_1z.dds"},
	{0x0773C794B8D3E4E5ULL, "332714298.ogg"},
	{0x0773CB87D6A31A3AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct2window.phy"},
	{0x0773F31BB043F269ULL, "vxht_mansion2_30w_0_0.vlm"},
	{0x07744C5112226264ULL, "vxht_mansion2b_2by_0_0.vlm"},
	{0x077457EE70ADADDAULL, "6205991_m.fxe"},
	{0x07747E80C31B4633ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_chainhead.phy"},
	{0x077484A9BB96FD86ULL, "gen_im_trs_crc_lgt_boo_r1.uti"},
	{0x0774A3A13C9E6BDEULL, "lgt311_shw_mer_mettin6156468.cl"},
	{0x0774C57C8EE7E33FULL, "mag110_vincento6017674.cl"},
	{0x0775097E694289C2ULL, "vxht_0_hightown_n_bem_0_0.vlm"},
	{0x077522B83328EC28ULL, "vxw_encounter2_camp_40_0_0.vlm"},
	{0x0775238C54DE1C5DULL, "vxlt_gallowscourtyard_7i1_0_0.vlm"},
	{0x077554C0A6F4CC60ULL, "vxht_bartrand_n_4de_0_0.vlm"},
	{0x0775D50ABA77C3B7ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.mmh"},
	{0x0775E2073C663A22ULL, "661098789.ogg"},
	{0x0776398F72AF17C1ULL, "vxlt_undercityserialkiller_1fn_0_0.vlm"},
	{0x07765CB29522CC71ULL, "vxdr_0_deeproads_cun_0_1.vlm"},
	{0x077669D78B9E76C3ULL, "6123233_f.fxe"},
	{0x0776746D3C7A7E06ULL, "vxlt_0_lowtown_a2_n_c_gzc_0_0.vlm"},
	{0x07770C167CCE2D87ULL, "vxlt_0_lowtown_a2_hfd_0_0.vlm"},
	{0x07775DD6F4EDDB8FULL, "6033477_f.fxe"},
	{0x0777A68350397220ULL, "6071295_m.fxe"},
	{0x0777AA0EA0802536ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfr.ani"},
	{0x07785507FC891BFAULL, "6128397_m.fxe"},
	{0x07785F8672CD7897ULL, "vxht_mansion2b_25c_0_0.vlm"},
	{0x07788F833403BB9BULL, "6160943_m.fxe"},
	{0x077896662D0B691FULL, "996266580.ogg"},
	{0x0778AFAB9B4F0F35ULL, "777001793.ogg"},
	{0x0778BF075CFA13CDULL, "and230tr_act2_opening.ncs"},
	{0x07795C809A9974BEULL, "ico_ring_3.dds"},
	{0x07795D0AACAE9617ULL, "6000829_m.fxe"},
	{0x07797A293E774F07ULL, "6143210_m.fxe"},
	{0x0779DA77EF771225ULL, "dae000_madam_lusine6151628.cl"},
	{0x077A1FC2696342DAULL, "w_sundermount_55_0.anb"},
	{0x077A3A07C3E99A42ULL, "llac_lt_gallowstemplar_f_2384_0.dds"},
	{0x077A5C2114391F06ULL, "cage_05.mao"},
	{0x077A95FBCD04E1A6ULL, "vxlt_gallowstemplar_n_5qa_0_0.vlm"},
	{0x077AC05F41EA8CA3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_galowint.phy"},
	{0x077ADB71B1573511ULL, "6030491_m.fxe"},
	{0x077AEF91C2992A15ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1table.msh"},
	{0x077AF191C2992D8BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1table.mmh"},
	{0x077AFD91C2994228ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1table.mao"},
	{0x077B36D8062DE1BBULL, "6059988_m.fxe"},
	{0x077BC52AE47E4DE8ULL, "6079793_f.fxe"},
	{0x077C544DFB296A47ULL, "art\\characters\\creatures\\desiredemon\\animation\\inst.gad"},
	{0x077D3CA7931A5E8DULL, "vxlt_0_lowtown_a3_n_c_hf6_0_0.vlm"},
	{0x077D42F253D1B630ULL, "6208213_m.fxe"},
	{0x077D8CA662F5811BULL, "706450533.ogg"},
	{0x077DC76747AADFA3ULL, "and330st_base_justice_post1.stg"},
	{0x077DEE6B71BC4579ULL, "vxht_0_hightown_a2_b82_0_0.vlm"},
	{0x077DF90DECF1E67FULL, "6042287_m.fxe"},
	{0x077E0FAA6D2E79AAULL, "hf_mother_h3s2.mor"},
	{0x077E1CFF819DFE61ULL, "lav_crust02d.dds"},
	{0x077E7EF7BFDE3165ULL, "307693223.ogg"},
	{0x077F39B987FE9B26ULL, "6115880_f.fxe"},
	{0x077F86F03A409239ULL, "715014997.ogg"},
	{0x077F9A76DF5EC51CULL, "llac_lt_gallowscourtyard_n_-2_0.dds"},
	{0x07800C276835B68DULL, "vxw_encounter2_nk_0_0.vlm"},
	{0x0780182FD84E65B2ULL, "vxht_mansion2_1y5_0_0.vlm"},
	{0x0780784ABEB7E3FBULL, "vxdr_0_deeproads_f0o_0_0.vlm"},
	{0x078093C3EF4814FEULL, "6095743_m.fxe"},
	{0x0781163B2E7857C0ULL, "one280_merin6120920.cl"},
	{0x0781C15427CB8319ULL, "dae000ip_crowd_mf6_talk16.utp"},
	{0x0781C187185F5F67ULL, "1033526843.ogg"},
	{0x0783494DFB2F7654ULL, "art\\characters\\creatures\\desiredemon\\animation\\inst.evt"},
	{0x0784C65991151C1EULL, "vxht_0_hightown_n_c47_0_0.vlm"},
	{0x0784E366F5019BBFULL, "zz_christest.cut"},
	{0x07851BCA95AC4BF4ULL, "240241348.ogg"},
	{0x07854637103FF08AULL, "922015221.ogg"},
	{0x07857DF36EC44607ULL, "vxlt_0_lowtown_a2_hcg_0_0.vlm"},
	{0x07858BA2D07A6D1DULL, "1008466646.ogg"},
	{0x0785B20EA08C7027ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfr.evt"},
	{0x0785F3FF1209A95DULL, "llac_w_encounter2_327685_0.dds"},
	{0x0786C15F7A38865EULL, "472225132.ogg"},
	{0x07875FA50D322E44ULL, "t1_skn_004.tnt"},
	{0x0787635987BC2DBCULL, "llac_lt_0_lowtown_docks_n_-65540_0.dds"},
	{0x07880E761638CDFFULL, "6082585_bespecific.cut"},
	{0x0788D787F6352B0BULL, "dae000_arishok6078457.cl"},
	{0x07899495FFEFD8CCULL, "87459977.ogg"},
	{0x0789ABFFEEE17AE8ULL, "gen00fl_fenris6105866.cl"},
	{0x0789C344BF8DA072ULL, "6115797_m.fxe"},
	{0x0789F536AD3CEEEAULL, "plt_cod_hst_sundermount.plo"},
	{0x078A07CC34D4AC92ULL, "lowtown_male_alienage1_sta6161632.cl"},
	{0x078A770FB4638FE2ULL, "vxlt_0_lowtown_a2_n_erk_0_0.vlm"},
	{0x078AA866AD0F2EB5ULL, "zz_followers_banter6157678.cl"},
	{0x078BBFEDEFA75402ULL, "hlac_lt_backalley_n_0_12z.dds"},
	{0x078BD30EA0911470ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfr.gad"},
	{0x078D4E242DF26129ULL, "6103339_m.fxe"},
	{0x078DAC1C93D3A945ULL, "vxht_keep_4nr_0_0.vlm"},
	{0x078DB352B3B8D6F7ULL, "70402972.ogg"},
	{0x078DC20B4D4AD418ULL, "lgt220pt_shp_kill_qun.ncs"},
	{0x078E59A2580355C6ULL, "vxw_bonepitmining_a2_16q_0_0.vlm"},
	{0x078E7E045A2BCF0DULL, "vxw_encounter2_ij_0_0.vlm"},
	{0x078ED83A99975779ULL, "lyriumghost_magicresist.dds"},
	{0x078F673AE4A5C61EULL, "6115072_m.fxe"},
	{0x078FCB3A2A4F3DE7ULL, "war_critimmunity.dds"},
	{0x0790153F81562F40ULL, "vxht_0_hightown_a2_c3k_0_0.vlm"},
	{0x07903CCF058F214AULL, "hlac_ca_caverns1_24422_4y.dds"},
	{0x0790414DFB39F28DULL, "art\\characters\\creatures\\desiredemon\\animation\\inst.ani"},
	{0x079055EA22CEF89FULL, "art\\characters\\weapons\\maces\\w_mce_mc105a.msh"},
	{0x07905BEA22CF0341ULL, "art\\characters\\weapons\\maces\\w_mce_mc105a.mmh"},
	{0x0790A97901D37906ULL, "fx_dazzed_halo.mao"},
	{0x079199C41271A5CCULL, "qun221_petrice6077987.cl"},
	{0x0791A074768F662FULL, "one280_iwan6096927.cl"},
	{0x0791ED384D2EBAC2ULL, "vxht_0_hightown_n_bko_0_0.vlm"},
	{0x0792120B354F4775ULL, "6016297_m.fxe"},
	{0x079262979B1E8315ULL, "vxw_woundedcoast_oa_0_0.vlm"},
	{0x0792704B3357A15EULL, "mcr370_templar16086968.cl"},
	{0x079288EF872B6026ULL, "6147641_f.fxe"},
	{0x07929FA0A478E1E9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windownk.phy"},
	{0x07934159153D9DB9ULL, "dae000_tomwise6167323.cl"},
	{0x079354FAAA3C5320ULL, "vxht_0_hightown_c11_0_0.vlm"},
	{0x079358272066AF06ULL, "vxlt_0_lowtown_a2_n_c_gpe_0_0.vlm"},
	{0x0793EFD060BA5E87ULL, "vxlt_0_lowtown_a2_hhf_0_0.vlm"},
	{0x07940EE53DDC2D11ULL, "826508734.ogg"},
	{0x079477AE5B1BB930ULL, "l_mag311ar_warehouse.bnk"},
	{0x07959BB2E95D1DCFULL, "6137508_f.fxe"},
	{0x0795C1A09CD18D0DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_setprops.msh"},
	{0x0795CFA09CD1A540ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_setprops.mao"},
	{0x0795D3A09CD1AB93ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_setprops.mmh"},
	{0x079623CB956077ABULL, "6149704_m.fxe"},
	{0x079624161B63CF5CULL, "vxca_caverns2_95p_0_0.vlm"},
	{0x0796BF8E7CA1C6DFULL, "llac_lt_0_lowtown_docks_-65537_4.dds"},
	{0x0796CB27F69E2A01ULL, "vxlt_0_lowtown_docks_n_b0p_0_0.vlm"},
	{0x0796D2DBA5B734B4ULL, "6051399_f.fxe"},
	{0x0796DCB632C0BFF9ULL, "vxht_0_hightown_n_aua_0_0.vlm"},
	{0x0797BEDDA20FBC0FULL, "vxht_brothel_n_4cc_0_0.vlm"},
	{0x0797E6F5C251AF02ULL, "17813400.ogg"},
	{0x079817387B9A9CB5ULL, "vxht_mansion2b_1af_0_0.vlm"},
	{0x079819512C7794DDULL, "12972369.ogg"},
	{0x07983521B183B8F2ULL, "6110897_m.fxe"},
	{0x07986D21145239A8ULL, "cage_05.dds"},
	{0x0798813AAD1FC871ULL, "hlac_lt_0_lowtown_n_-131071_0y.dds"},
	{0x07988CE964F3F33AULL, "isa111_isabela6068965.cl"},
	{0x0798B20C71F589A5ULL, "6150379_m.fxe"},
	{0x0798E8ECA659A862ULL, "vxlt_0_lowtown_a2_n_ewt_0_0.vlm"},
	{0x0798FF6E04ADA452ULL, "bld_drian_tu3.dds"},
	{0x07990A99AAA6059BULL, "6149058_m.fxe"},
	{0x079910C8282AC02FULL, "296548502.ogg"},
	{0x079941D3A4AE21B0ULL, "285098305.ogg"},
	{0x07998607731E5F3AULL, "17537944.ogg"},
	{0x0799BE966081D64FULL, "6121207_m.fxe"},
	{0x0799E0EDB86F2D66ULL, "follower_banter6136364.cl"},
	{0x079A0B4C2B56BDFBULL, "mrl310_party_comments6142849.cl"},
	{0x079BA109D01A41C0ULL, "746979851.ogg"},
	{0x079BEF6A50F3A48AULL, "vxlt_warehouse_aj_0_0.vlm"},
	{0x079C25A0463BB1DDULL, "vxw_bonepitdestroyed_a3_1p6_0_0.vlm"},
	{0x079C4903423F93C8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicechair.msh"},
	{0x079C5B03423FB259ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicechair.mao"},
	{0x079C6703423FC6F2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicechair.mmh"},
	{0x079D596854B51CEDULL, "455142024.ogg"},
	{0x079D5B50F9509E90ULL, "vxw_encounter3_8k_0_0.vlm"},
	{0x079F7C6BB2C03D9CULL, "follower_banter6163535.cl"},
	{0x079F9EAF6E81F1A9ULL, "vxht_0_hightown_a3_avj_0_0.vlm"},
	{0x079FBF9345CA6760ULL, "6115755_m.fxe"},
	{0x079FD839D1194E00ULL, "6147859_m.fxe"},
	{0x079FFD8CDBF605D6ULL, "126757341.ogg"},
	{0x07A04268A503D490ULL, "6108937_m.fxe"},
	{0x07A08AAB9EEA5DEFULL, "vxlt_0_lowtown_a2_n_c_exj_0_0.vlm"},
	{0x07A0B95DB570FE25ULL, "210799245.ogg"},
	{0x07A0E329BACC1713ULL, "vxlt_gallowstemplar_3sc_0_0.vlm"},
	{0x07A154D6ACE182A9ULL, "vxht_0_hightown_ccr_0_0.vlm"},
	{0x07A15C440092CBC1ULL, "586777993.ogg"},
	{0x07A191472A64814CULL, "6070219_f.fxe"},
	{0x07A20620F3790224ULL, "hlac_w_sundermount_-327685_0y.dds"},
	{0x07A208588500619CULL, "vxw_bonepitmining_a2_73_0_0.vlm"},
	{0x07A222C26984998DULL, "w_sundermount_55_0.mmh"},
	{0x07A226C26984A0DEULL, "w_sundermount_55_0.mao"},
	{0x07A234C26984B813ULL, "w_sundermount_55_0.msh"},
	{0x07A28BEEA0BEC17BULL, "gen00fl_isabela6096820.cl"},
	{0x07A2ADC4AF8C58CDULL, "blank_exterior.rml"},
	{0x07A2B4510CB6AAF1ULL, "mag321_comtesse_de_launcet6043255.cl"},
	{0x07A2E7E6890C48DDULL, "mcr370_meredith6085911.cl"},
	{0x07A3111ED1954788ULL, "vxca_caverns2prc_8mm_0_1.vlm"},
	{0x07A355930C07C9E3ULL, "803028275.ogg"},
	{0x07A369EC8950D391ULL, "6092510_m.fxe"},
	{0x07A392C520F29D64ULL, "w_woundedcoast_157_0.msh"},
	{0x07A3A4C520F2BBCDULL, "w_woundedcoast_157_0.mao"},
	{0x07A3A8C520F2C206ULL, "w_woundedcoast_157_0.mmh"},
	{0x07A3AA9D50257441ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_2x.dds"},
	{0x07A3CA44FE7BE2D7ULL, "vxw_bonepit_8o_0_0.vlm"},
	{0x07A4E3A533F18612ULL, "hlac_dr_dungeon_3207_0x.dds"},
	{0x07A5953ACC3BF506ULL, "175081927.ogg"},
	{0x07A59A7288C49CE6ULL, "vxht_0_hightown_c3a_0_0.vlm"},
	{0x07A59EE13FB0A434ULL, "lt_0_lowtown_n_19_0_lc.mmh"},
	{0x07A5D15842FAB5FEULL, "llac_lt_0_lowtown_docks_a3_n_-65538_4.dds"},
	{0x07A689156D90B535ULL, "vxht_mansion2b_1de_0_0.vlm"},
	{0x07A6E9FAB2ACB9BCULL, "6055434_f.fxe"},
	{0x07A6EAB8ACF5168EULL, "ef_bar_flirt.mor"},
	{0x07A7298FBF9B4CB8ULL, "552606664.ogg"},
	{0x07A747C161E24B38ULL, "leave_chunky_l3.mao"},
	{0x07A77C769C9C1E1BULL, "6142158_m.fxe"},
	{0x07A81B42AB6EB904ULL, "vxht_brothel_n_5dx_0_0.vlm"},
	{0x07A83082117C2D9BULL, "vxlt_hangedman_18q_0_0.vlm"},
	{0x07A8B53FA8A63412ULL, "llac_ht_brothel_n_3374_0.dds"},
	{0x07A8B99A35C7A937ULL, "gencr_desire_demon.utc"},
	{0x07A8E86D0D32D5ABULL, "hm_nobleman_1.mor"},
	{0x07A8FA053BE83271ULL, "vxw_bonepitmining_a2_cv_0_0.vlm"},
	{0x07A90598852A126EULL, "hlac_ca_cavern1_haunted_24426_11x.dds"},
	{0x07A9CF15D1B94718ULL, "llac_dr_primevaltomb_661_1.dds"},
	{0x07A9EAA7C65BF617ULL, "seb321_leliana6197874.cl"},
	{0x07AA2CB96BD5D3AAULL, "storage_n.dds"},
	{0x07AA2FEA6F6ED560ULL, "254787041.ogg"},
	{0x07AA977D3831CA5DULL, "vxw_bonepitdestroyed_a3_1u5_0_0.vlm"},
	{0x07AAE76B5E549F8CULL, "6044875_m.fxe"},
	{0x07AB1827883E8449ULL, "vxlt_hangedman_n_1td_0_0.vlm"},
	{0x07AB3AE75B7662A8ULL, "820697048.ogg"},
	{0x07AB751D546DBA3FULL, "673121504.ogg"},
	{0x07AB804A787A212CULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.dura.ani"},
	{0x07ABCD2DEB46B710ULL, "vxw_encounter3_5h_0_0.vlm"},
	{0x07AC2B9435B75243ULL, "dre000_sibling_has_blight6036482.cl"},
	{0x07AC953FD096B54CULL, "vxlt_0_lowtown_a3_n_c_ffu_0_0.vlm"},
	{0x07ACAD9F01DB07A0ULL, "6153937_m.fxe"},
	{0x07ACDEC489082DA4ULL, "770637477.ogg"},
	{0x07AD01E573F4A55AULL, "zz_followers_banter6158559.cl"},
	{0x07AD22496EA5F100ULL, "vxlt_0_lowtown_fhg_0_0.vlm"},
	{0x07AD22C2698EA106ULL, "w_sundermount_55_0.rml"},
	{0x07AD48852204ECD7ULL, "vxlt_0_lowtown_a3_n_c_eov_0_0.vlm"},
	{0x07AD4F573B3AE94AULL, "mag360_samson6042088.cl"},
	{0x07AD8AA0A8F86BC3ULL, "llac_dr_0_deeproads_5845_18.dds"},
	{0x07ADA45F82D77C0EULL, "6088420_m.fxe"},
	{0x07ADAF274569187AULL, "134177803.ogg"},
	{0x07ADCB9E55305C07ULL, "w_woundedcoast_15_0_lc.mmh"},
	{0x07ADD2CB84CBAB95ULL, "vxw_sundermount_1yz_0_0.vlm"},
	{0x07ADFA3E9E367B43ULL, "llac_lt_backalley_g_0_1.dds"},
	{0x07AE317037C07FE0ULL, "6110855_m.fxe"},
	{0x07AE4A16C30F6680ULL, "6140959_m.fxe"},
	{0x07AE7B2A964F0C47ULL, "6092400_m.fxe"},
	{0x07AE8EBCD1972F19ULL, "877931651.ogg"},
	{0x07AEAE741F948A36ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_2x.dds"},
	{0x07AEBE5833C03698ULL, "llac_w_bonepit_0_6.dds"},
	{0x07AEC6C807168794ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs104a_0n.dds"},
	{0x07AED024A32023CBULL, "882362404.ogg"},
	{0x07AF05D36D231654ULL, "character_character_p179.psh.fxo.dx11"},
	{0x07AF2A896FFB3848ULL, "vxlt_0_lowtown_docks_b54_0_1.vlm"},
	{0x07AF46517F0AB415ULL, "vxw_woundedcoast_ao_0_0.vlm"},
	{0x07AF4BCEDA73DAC3ULL, "dae211ar_tavern.are"},
	{0x07AFAFA9248D71F7ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm5.gad"},
	{0x07AFD1F6796946EAULL, "601840597.ogg"},
	{0x07B0168BD72BAAE1ULL, "or_db_clo.mao"},
	{0x07B0BBB3338AC02FULL, "vxw_sundermount_31g_0_1.vlm"},
	{0x07B0E18E92DB5245ULL, "22221941.ogg"},
	{0x07B133B8D7ECBAAFULL, "6104945_m.fxe"},
	{0x07B140B50D23AE43ULL, "vxht_mansion2b_32g_0_0.vlm"},
	{0x07B1AC9D747E03D8ULL, "vxlt_0_lowtown_docks_n_c3u_0_0.vlm"},
	{0x07B22C695CDFF021ULL, "6060906_m.fxe"},
	{0x07B26DB8598451CEULL, "6036515_m.fxe"},
	{0x07B28D4A7880566EULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.dura.gad"},
	{0x07B28DCA3F2B1789ULL, "736528541.ogg"},
	{0x07B2A1EA40C10EB0ULL, "1013623299.ogg"},
	{0x07B300D78E4EBCDCULL, "hlac_w_woundedcoast_-65536_0y.dds"},
	{0x07B31BC26993014AULL, "w_sundermount_55_0.phy"},
	{0x07B34A87F4440C10ULL, "lgt360_mag_orlanna6155391.cl"},
	{0x07B383E04D34C2B6ULL, "589877431.ogg"},
	{0x07B38FBF501341E1ULL, "6077303_f.fxe"},
	{0x07B3E6544945001CULL, "6106971_m.fxe"},
	{0x07B3FCB9051C8214ULL, "combat_smcc_range.gda"},
	{0x07B40C4F7ABAB566ULL, "vxht_0_hightown_c6b_0_0.vlm"},
	{0x07B45363D2204596ULL, "381032126.ogg"},
	{0x07B4F12BC1D5B558ULL, "95636241.ogg"},
	{0x07B53FDC196E4CEFULL, "994898071.ogg"},
	{0x07B54C5C72061B92ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.mmh"},
	{0x07B57C6E1817527CULL, "vxht_0_hightown_a2_5gy_0_0.vlm"},
	{0x07B585200DD5E221ULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.anb"},
	{0x07B5B694A1A313F9ULL, "6053342_m.fxe"},
	{0x07B5DB4D7CE8B2B7ULL, "vxlt_warehouse_n_199_0_0.vlm"},
	{0x07B5E7C58EA2D9CAULL, "174419138.ogg"},
	{0x07B6281F32FDF995ULL, "hlac_ht_0_hightown_a3_196607_0z.dds"},
	{0x07B639C6379DF431ULL, "zz_followers_banter6158349.cl"},
	{0x07B6BE288456C5E9ULL, "vxht_0_hightown_n_bif_0_0.vlm"},
	{0x07B6E3E20376BFBFULL, "308729787.ogg"},
	{0x07B76BE22DDE4AF1ULL, "vxw_bonepitmining_a2_nu_0_0.vlm"},
	{0x07B7B43422F45761ULL, "vxlt_0_lowtown_a2_fiy_0_0.vlm"},
	{0x07B7CC6F2BBAAA9BULL, "mcr310_argument6075803.cl"},
	{0x07B849EBC9ACD37BULL, "6002799_m.fxe"},
	{0x07B9884A78866C4DULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.dura.evt"},
	{0x07B98A047A349CC9ULL, "vxlt_hangedman_n_1qk_0_0.vlm"},
	{0x07B9C50993EBB0B4ULL, "llac_ca_caverns1_mines_24424_8.dds"},
	{0x07BA2B1B2AE03BB5ULL, "6170557_m.fxe"},
	{0x07BA59AE5B3C55FAULL, "l_mag311ar_warehouse.wbd"},
	{0x07BAACB93A3E8A6BULL, "163963612.ogg"},
	{0x07BB2EBB5F4185B0ULL, "vxht_chantry_n_1yh_0_0.vlm"},
	{0x07BC163C74CD948EULL, "6083520_m.fxe"},
	{0x07BC414D704117D5ULL, "174587843.ogg"},
	{0x07BC50BC3009F982ULL, "521976231.ogg"},
	{0x07BC5CEA22F43FBEULL, "art\\characters\\weapons\\maces\\w_mce_mc105a.phy"},
	{0x07BC6C6F060CC992ULL, "1024719107.ogg"},
	{0x07BC6FEB1E917977ULL, "232933825.ogg"},
	{0x07BC7D7630B42CD3ULL, "plt_cod_itm_blackfeather.plo"},
	{0x07BC875C762D633CULL, "fxc_spider_web.dds"},
	{0x07BCAAC75DDC7DEFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabsedge.mao"},
	{0x07BCB6C75DDC918CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabsedge.mmh"},
	{0x07BCBCC75DDC9C7EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabsedge.msh"},
	{0x07BD13171E7E16C4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_chantrybasic.msh"},
	{0x07BD25171E7E352DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_chantrybasic.mao"},
	{0x07BD27B6165FCEEEULL, "vxht_mansion2b_3g5_0_0.vlm"},
	{0x07BD29171E7E3C66ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_chantrybasic.mmh"},
	{0x07BD6E6582D68EEFULL, "vxlt_0_lowtown_a2_n_c_ert_0_0.vlm"},
	{0x07BDC12AFF2F6F52ULL, "181919069.ogg"},
	{0x07BE88A0A49DF50EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windownk.mmh"},
	{0x07BE94A0A49E09A5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windownk.mao"},
	{0x07BEA2A0A49E21FCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windownk.msh"},
	{0x07BEEC1268EC929CULL, "vxw_bonepitmining_a2_91_0_0.vlm"},
	{0x07BF1E690FC0B451ULL, "vxht_chantry_n_339_0_0.vlm"},
	{0x07BF64DFF6FD123CULL, "904428238.ogg"},
	{0x07BFA1B7883D1FAFULL, "vxw_bonepit_bu_0_0.vlm"},
	{0x07BFD7F04E17E561ULL, "845524399.ogg"},
	{0x07C0E99B8BE979C4ULL, "vxdr_dungeon_1m0_0_0.vlm"},
	{0x07C23264A8D08F15ULL, "mcr370_meredith6084398.cl"},
	{0x07C2EB931DD1E0C9ULL, "6090039_m.fxe"},
	{0x07C366200DE1EA9AULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.mmh"},
	{0x07C369545B364919ULL, "vxdr_primevalentrance_1k4_0_1.vlm"},
	{0x07C3C0238BAC6FCCULL, "344964187.ogg"},
	{0x07C3DCA9249E677DULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm5.ani"},
	{0x07C3E0E55A4DBD65ULL, "vxw_woundedcoast_o1_0_0.vlm"},
	{0x07C40F74A30CA4B5ULL, "hlac_lt_0_lowtown_a3_n_c_131071_2y.dds"},
	{0x07C4F9EA32C68A71ULL, "6081349_m.fxe"},
	{0x07C5119A7AF699D4ULL, "vxlt_0_lowtown_h9n_0_0.vlm"},
	{0x07C545B8F27C340EULL, "6216072_m.fxe"},
	{0x07C56FEDD42955DBULL, "lgt110_arl_bann_teagan6115619.cl"},
	{0x07C5A17748BE5696ULL, "lt_0_lowtown_docks_a3_11_0_ll.mmh"},
	{0x07C60A4FC5ECF7F4ULL, "6172543_m.fxe"},
	{0x07C711EF436C8490ULL, "vxlt_0_lowtown_a2_erj_0_0.vlm"},
	{0x07C720171E868131ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_chantrybasic.phy"},
	{0x07C7FE7898EE0254ULL, "vxlt_0_lowtown_ev6_0_0.vlm"},
	{0x07C86BE01FAC33D4ULL, "character_character_p252.psh.fxo.dx11.assembly"},
	{0x07C90DF2FEC9FA75ULL, "llac_lt_0_lowtown_docks_a3_n_-65540_2.dds"},
	{0x07C9F571D16F3F08ULL, "llac_w_sundermount_5_0.dds"},
	{0x07C9FC437CF16314ULL, "vxlt_darktown_2v2_0_0.vlm"},
	{0x07CA2C3F05F11DD7ULL, "vxlt_0_lowtown_a3_n_c_eup_0_0.vlm"},
	{0x07CA7ED251DFF8EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv3.gad"},
	{0x07CAC273B683F907ULL, "6127991_m.fxe"},
	{0x07CB6B6DCC8894C3ULL, "6120846_m.fxe"},
	{0x07CB75E45FBFF9C4ULL, "6155823_m.fxe"},
	{0x07CB9E9BA4087FEDULL, "vxht_mansion2b_3cv_0_0.vlm"},
	{0x07CC3AC30C45AB72ULL, "follower_banter6155631.cl"},
	{0x07CC69ABCF5D4D37ULL, "gen_im_acc_rng_act3_08.uti"},
	{0x07CCC59C904E29D9ULL, "hlac_lt_0_lowtown_a2_1_7z.dds"},
	{0x07CD02A5BD559525ULL, "6208025_m.fxe"},
	{0x07CD2E8BF6D20F48ULL, "qun110_petrice6018527.cl"},
	{0x07CD51A9978277C7ULL, "6115819_m.fxe"},
	{0x07CD84DB701F39A9ULL, "6048135_m.fxe"},
	{0x07CDFF55769A5655ULL, "llac_ht_chantry_2534_46.dds"},
	{0x07CE13947A33382FULL, "vxw_bonepit_mt_0_0.vlm"},
	{0x07CF47373798EA95ULL, "vxw_bonepit_1wg_0_0.vlm"},
	{0x07CFB6632E6AF9C7ULL, "729158139.ogg"},
	{0x07CFC487D6F16D63ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct2window.msh"},
	{0x07CFD287D6F184BDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct2window.mmh"},
	{0x07CFD687D6F18BCEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct2window.mao"},
	{0x07CFF08EE58EB8D5ULL, "105379879.ogg"},
	{0x07CFF9369A54D896ULL, "vxlt_gallowscourtyard_n_4r3_0_0.vlm"},
	{0x07D02A95812DF46AULL, "692712829.ogg"},
	{0x07D08387DE683419ULL, "qun221_viscount.bnk"},
	{0x07D11890EFD9ED38ULL, "356147462.ogg"},
	{0x07D11B9884579072ULL, "70752520.ogg"},
	{0x07D137EB96D77FCDULL, "624611782.ogg"},
	{0x07D141F628313D7BULL, "887281049.ogg"},
	{0x07D155AFFBBA550FULL, "6080278_m.fxe"},
	{0x07D161D251E5E581ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv3.evt"},
	{0x07D1CC2C0F3756F7ULL, "gen00fl_merrill6162019.cl"},
	{0x07D1DEAC4372FEB5ULL, "vxht_mansion2b_1ub_0_0.vlm"},
	{0x07D226B87A1D0DCFULL, "hlac_w_bonepitdestroyed_a3_458757_0y.dds"},
	{0x07D252C24C43D5E5ULL, "vxw_woundedcoast_d0_0_0.vlm"},
	{0x07D2AFD31EA8630FULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.dura.gad"},
	{0x07D2B3E278EC7EB8ULL, "hlac_lt_0_lowtown_docks_a3_-3_0y.dds"},
	{0x07D2BAD7D9F0672DULL, "mag101_aombanter6052830.cl"},
	{0x07D311662E5CD5B6ULL, "6064992_m.fxe"},
	{0x07D326FA453B4DA8ULL, "docks_scene_dockworkers6176473.cl"},
	{0x07D3652E6154B23DULL, "w_bonepitmining_a2_233_0.mao"},
	{0x07D3692E6154B976ULL, "w_bonepitmining_a2_233_0.mmh"},
	{0x07D3732E6154CA74ULL, "w_bonepitmining_a2_233_0.msh"},
	{0x07D3A587DE6AA479ULL, "qun221_viscount.cnv"},
	{0x07D3AF6A30A3CBB1ULL, "vxht_keep_n_55q_0_0.vlm"},
	{0x07D3BAF0E4FB3162ULL, "hlac_lt_backalley_g_65537_7y.dds"},
	{0x07D47026BC7CCD8BULL, "6108079_m.fxe"},
	{0x07D4A65078E14E53ULL, "hlac_w_encounter3_-327681_0z.dds"},
	{0x07D4C955427BD818ULL, "6064859_m.fxe"},
	{0x07D519A017DC7450ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12.mmh"},
	{0x07D522B9D6078F68ULL, "6109206_m.fxe"},
	{0x07D561ACE1C282B2ULL, "115839401.ogg"},
	{0x07D5902D6D099B0EULL, "gen_im_trs_slv_junk2_02.uti"},
	{0x07D5C8E72988AC19ULL, "rdr111dg_martin6014947.cl"},
	{0x07D6E726D041ECEFULL, "hlac_ca_caverns2prc_5802_26y.dds"},
	{0x07D7668AFC694D34ULL, "84425335.ogg"},
	{0x07D76CAE641B4229ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_6.dds"},
	{0x07D77780C36FB7A8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_chainhead.mmh"},
	{0x07D78380C36FCC4BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_chainhead.mao"},
	{0x07D79580C36FEAD2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_chainhead.msh"},
	{0x07D7FAEFE020517BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ferelden_jimmy_da.dds"},
	{0x07D8213AB9E453AAULL, "335318809.ogg"},
	{0x07D833F6FDFEFFB0ULL, "910051035.ogg"},
	{0x07D85CD647C8F9CAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzardb_d.dds"},
	{0x07D86E206EE77B94ULL, "vxlt_darktown_2y3_0_0.vlm"},
	{0x07D89E1BF7E73657ULL, "vxlt_0_lowtown_a3_n_c_exs_0_0.vlm"},
	{0x07D8A26F67EAAB2FULL, "lt_0_lowtown_a2_n_46_0_ll.mmh"},
	{0x07D8A57D66AE40D8ULL, "isa120cr_letter_thug.utc"},
	{0x07D8C4D31EACF2BCULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.dura.evt"},
	{0x07D8D74AAD1D01C7ULL, "vxlt_hangedman_21d_0_1.vlm"},
	{0x07D8D890218061C3ULL, "6021338_m.fxe"},
	{0x07D8FD66EB8E1EF1ULL, "744479932.ogg"},
	{0x07D9889835872B05ULL, "vxlt_0_lowtown_a2_n_c_hvs_0_0.vlm"},
	{0x07DA1EB7C7114AEBULL, "vxlt_0_lowtown_n_gbx_0_0.vlm"},
	{0x07DA52F1584554D2ULL, "959699377.ogg"},
	{0x07DA760481DC04DBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_weightshift_medium.ani"},
	{0x07DA9BE854ECFD95ULL, "vxw_woundedcoast_pl_0_0.vlm"},
	{0x07DB413690946A88ULL, "hlac_lt_0_lowtown_1_8x.dds"},
	{0x07DB54CEDA9924A8ULL, "dae211ar_tavern.lst"},
	{0x07DB8E2C40A908E2ULL, "vxlt_0_lowtown_docks_n_be0_0_0.vlm"},
	{0x07DBBF84405DD60CULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanl.ani"},
	{0x07DBD9C2654AD3F8ULL, "lgt260cr_eau_asst_harbor.utc"},
	{0x07DC3072E359355AULL, "328889444.ogg"},
	{0x07DC48803540D00FULL, "vxlt_0_lowtown_a2_h1u_0_0.vlm"},
	{0x07DC7BB97EDA0D6CULL, "hlac_ht_mansion2_818_6z.dds"},
	{0x07DC85716C2950AFULL, "vxw_bonepit_hw_0_0.vlm"},
	{0x07DD462C7AAA20CDULL, "6108239_m.fxe"},
	{0x07DDB914298F0315ULL, "vxw_bonepit_1xf_0_0.vlm"},
	{0x07DE138E807BFF10ULL, "dae000_porfiria6155264.cl"},
	{0x07DE4B5C72291F39ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.anb"},
	{0x07DE9D338C5B0833ULL, "6094811_m.fxe"},
	{0x07DEECA10E5F3075ULL, "vxw_encounter2_camp_qr_0_0.vlm"},
	{0x07DF59D251F215D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv3.ani"},
	{0x07DF82C7FCED5DBBULL, "vxlt_0_lowtown_docks_n_bvh_0_0.vlm"},
	{0x07DFA19A8B544504ULL, "340634248.ogg"},
	{0x07DFC299A3EE0525ULL, "vxht_keep_703_0_0.vlm"},
	{0x07E05894B00A22DBULL, "shi_star.mao"},
	{0x07E07A3750CBE7BAULL, "vxw_bonepitmining_a2_fj_0_0.vlm"},
	{0x07E09B7CBE6DF8BDULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi04.phy"},
	{0x07E0D7FDCC56A711ULL, "vxht_mansion_n_6jo_0_0.vlm"},
	{0x07E0E28263338CE2ULL, "vxlt_gallowstemplar_f_34x_0_0.vlm"},
	{0x07E13174F29AC538ULL, "gen00fl_fenris6126982.cl"},
	{0x07E1C96D73E0B1F0ULL, "641646339.ogg"},
	{0x07E23CDE7D42F8DDULL, "vxdr_primevaltomb_8m6_0_0.vlm"},
	{0x07E254CEDA9F43F9ULL, "dae211ar_tavern.ncs"},
	{0x07E264263137FA0DULL, "vxht_brothel_n_6ey_0_0.vlm"},
	{0x07E26D9E5D431D15ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg105a_0t.dds"},
	{0x07E2CC8440640B4EULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanl.gad"},
	{0x07E32562B70EF7C3ULL, "llac_ca_caverns2_5802_22.dds"},
	{0x07E33208635BECAAULL, "vxlt_undercityserialkiller_19p_0_0.vlm"},
	{0x07E3E1591FA46B7CULL, "vxlt_gallowstemplar_4sk_0_0.vlm"},
	{0x07E3F5A92758B590ULL, "vxlt_0_lowtown_a2_exh_0_0.vlm"},
	{0x07E41DC0BDEE8015ULL, "dr_primevalentrance_mmap.dds"},
	{0x07E45EF1000A2C5DULL, "vxw_bonepitdestroyed_a3_1i1_0_0.vlm"},
	{0x07E4D2FAD846C2D3ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\choke.evt"},
	{0x07E57C8E3034968DULL, "vxw_blightlands_7o_0_0.vlm"},
	{0x07E5E4C4ECE2B977ULL, "gen_im_npc_wep_two_qunb.uti"},
	{0x07E603B4B25B4E77ULL, "vxlt_0_lowtown_docks_n_bgc_0_0.vlm"},
	{0x07E693FD5E894D6EULL, "art\\vfx\\cinematic\\vfx_scaredmage_c.mmh"},
	{0x07E69EB749A6C522ULL, "vxht_chantry_2zg_0_0.vlm"},
	{0x07E6A094E5F4C513ULL, "vxht_0_hightown_a2_ary_0_0.vlm"},
	{0x07E6CCD31EB93DD5ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.dura.ani"},
	{0x07E779EF15BACC63ULL, "226769632.ogg"},
	{0x07E7A62D9A702A07ULL, "6129791_m.fxe"},
	{0x07E7CF12E8AA3761ULL, "6041110_m.fxe"},
	{0x07E87E0481E84FD2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_weightshift_medium.evt"},
	{0x07E881991774C3BEULL, "923356097.ogg"},
	{0x07E9694248A2AFE9ULL, "vxw_encounter3_7w_0_0.vlm"},
	{0x07E9C784406A212DULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_leanl.evt"},
	{0x07EA0009329F2162ULL, "vxlt_0_lowtown_docks_n_bb1_0_0.vlm"},
	{0x07EA03CCEE93549FULL, "vxht_mansion2_1sj_0_0.vlm"},
	{0x07EAE92A4DA7B9BCULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r90_p.gad"},
	{0x07EB4922AFFF9559ULL, "gen00fl_isabela6075234.cl"},
	{0x07EBCBFAD84CD514ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\choke.gad"},
	{0x07EC0058576B7C9BULL, "hlac_dr_0_deeproads_5842_9y.dds"},
	{0x07EC3D4E290513E7ULL, "6160819_m.fxe"},
	{0x07EC7C0526A9C441ULL, "follower_banter6171557.cl"},
	{0x07ECE3CCE9C58EB8ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appc_0n.dds"},
	{0x07ED30F39F45CB24ULL, "6052999_m.fxe"},
	{0x07ED5E7E005548F5ULL, "vxw_encounter2_camp_lq_0_0.vlm"},
	{0x07ED9C31351CC759ULL, "dae200_bran6094781.cl"},
	{0x07EDD51AE276F81EULL, "dae200_bran6097002.cl"},
	{0x07EDE70A36CF65E9ULL, "122676454.ogg"},
	{0x07EDF4A4EEE3763BULL, "vxlt_0_lowtown_docks_n_bsk_0_0.vlm"},
	{0x07EDF5E57743203CULL, "llac_w_encounter2_-131068_0.dds"},
	{0x07EE1BD205F8CE23ULL, "37104844.ogg"},
	{0x07EE4A514B28C2B8ULL, "qcr220_orsino6061367.cl"},
	{0x07EE9403A8EEC7C2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww22_galocell.msh"},
	{0x07EE9603A8EECB78ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww22_galocell.mmh"},
	{0x07EEA203A8EEDF1BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww22_galocell.mao"},
	{0x07EEE32FDFF0981BULL, "mrl000_marethari6129354.cl"},
	{0x07EF49DABE4CBF91ULL, "vxht_mansion_n_6mn_0_0.vlm"},
	{0x07EF993F519C7F61ULL, "6173699_m.fxe"},
	{0x07EF9D0481EEA365ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_weightshift_medium.gad"},
	{0x07F088A30F120B35ULL, "dae000_hate_enemies6167621.cl"},
	{0x07F0F99C0EB0A979ULL, "6177695_m.fxe"},
	{0x07F13BDDC521EECBULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm_facem1.mao"},
	{0x07F171431856CC88ULL, "vxlt_gallowstemplar_n_481_0_0.vlm"},
	{0x07F17DCDECD97B13ULL, "6161632_m.fxe"},
	{0x07F19C7CBE7C6DC2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi04.mmh"},
	{0x07F19E7CBE7C7178ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi04.msh"},
	{0x07F1E82A4DADD6BAULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r90_p.ani"},
	{0x07F1EFDFC6A1D2ABULL, "6096331_f.fxe"},
	{0x07F242E2E00B378DULL, "hlac_dr_0_deeproads_5840_1y.dds"},
	{0x07F29B7A2CD89D44ULL, "mrl150_merrill6150737.cl"},
	{0x07F2DAFAD8530DF2ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\choke.ani"},
	{0x07F3519A6FB0F42CULL, "vxw_sundermount_372_0_1.vlm"},
	{0x07F3B2FD5E940C6DULL, "art\\vfx\\cinematic\\vfx_scaredmage_c.anb"},
	{0x07F3EE6B222AAF0DULL, "vxw_blightlands_8n_0_0.vlm"},
	{0x07F4127539BA4520ULL, "hlac_lt_0_lowtown_docks_a3_196607_0z.dds"},
	{0x07F43EEF7A477A97ULL, "6128316_m.fxe"},
	{0x07F46205132A4314ULL, "hlac_ht_0_hightown_n_65533_2x.dds"},
	{0x07F51271D7EADD93ULL, "vxht_0_hightown_a2_amx_0_0.vlm"},
	{0x07F557332BE1BC52ULL, "vxht_chantry_n_27s_0_0.vlm"},
	{0x07F59DF137159CF4ULL, "6149281_f.fxe"},
	{0x07F61D882DE9EFD4ULL, "808050658.ogg"},
	{0x07F62282D3D6B0A9ULL, "842861886.ogg"},
	{0x07F65A924FDE629CULL, "6141581_m.fxe"},
	{0x07F76603428C70EDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicechair.phy"},
	{0x07F87A8C63AFA595ULL, "6059973_m.fxe"},
	{0x07F886528F9CC767ULL, "vxht_brothel_n_61w_0_0.vlm"},
	{0x07F89D03A8F72B03ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww22_galocell.phy"},
	{0x07F8B0C40AAF68B7ULL, "hlac_ht_0_hightown_a3_-65537_3y.dds"},
	{0x07F8B339F01A4173ULL, "6054734_m.fxe"},
	{0x07F8D7E54E99006EULL, "hlac_ht_chantry_n_2534_0z.dds"},
	{0x07F931CFB75CD138ULL, "vxht_chantry_26y_0_0.vlm"},
	{0x07F94206D0BE3B3FULL, "vxht_bartrand_n_734_0_0.vlm"},
	{0x07F996AFAA410E2BULL, "782189194.ogg"},
	{0x07FA61BD00B15FD6ULL, "vxw_encounter2_camp_4o_0_1.vlm"},
	{0x07FA983386411101ULL, "6148118_m.fxe"},
	{0x07FAFE5BC197CE09ULL, "gen00fl_isabela6101069.cl"},
	{0x07FB4B5E64552F67ULL, "vxw_encounter2_gb_0_0.vlm"},
	{0x07FB7580021D5823ULL, "736649680.ogg"},
	{0x07FBD05AF24B6175ULL, "vxw_encounter2_camp_gp_0_0.vlm"},
	{0x07FC50E970B17472ULL, "vxlt_warehouse_n_154_0_0.vlm"},
	{0x07FC56529079789FULL, "llac_lt_undercitydungeon_-1_28.dds"},
	{0x07FD11233804281EULL, "456708600.ogg"},
	{0x07FD2CD6487981E6ULL, "llac_lt_0_lowtown_a2_n_-65535_7.dds"},
	{0x07FD44A3AF16EFE1ULL, "6156550_f.fxe"},
	{0x07FDACB30F21084EULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.dura.ani"},
	{0x07FDCB82F62A12B2ULL, "vxht_bartrand_n_5pr_0_0.vlm"},
	{0x07FDD6FF3AADB260ULL, "vxlt_darktown_56u_0_0.vlm"},
	{0x07FDE067C455C221ULL, "677318038.ogg"},
	{0x07FE1C42BF8D8367ULL, "mrl000_marethari6142331.cl"},
	{0x07FE1EA22ABB9579ULL, "follower_banter6026240.cl"},
	{0x07FE20D334896900ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firecoal_e.dds"},
	{0x07FE8933454F2A6AULL, "gen_im_trs_cot_lgt_hlm_r2.uti"},
	{0x07FEA233BF5409B9ULL, "qun221_petrice6083802.cl"},
	{0x07FEEF1EC22F9A6FULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.impa.gad"},
	{0x07FF6C1156118EF7ULL, "pro000_opening6207400.cl"},
	{0x07FFDFF253161E65ULL, "l_env_wal_lt_qcr_mcr_cityfear_end_game.wbd"},
	{0x08000445E685187AULL, "618493915.ogg"},
	{0x080010307F82E7FBULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls107a_0s.dds"},
	{0x080037E1A45CD917ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_black.phy"},
	{0x08009069BDFE12A7ULL, "175557346.ogg"},
	{0x0800B3114F57D5A4ULL, "6154107_m.fxe"},
	{0x0800D466F214FC7FULL, "6102899_m.fxe"},
	{0x0800D4A31417BBEEULL, "fen310_fenris6118195.cl"},
	{0x08012F388329261BULL, "vxw_sundermount_30h_0_1.vlm"},
	{0x08016DD60E066C33ULL, "ss_lowclass_thugs_m_1.bnk"},
	{0x0802470AB86E95DDULL, "6102817_m.fxe"},
	{0x0802716FE03A7B75ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_act_c.mmh"},
	{0x0802BE99893E3AC5ULL, "576591666.ogg"},
	{0x0802F463EB5BA6F6ULL, "vxlt_hangedman_n_sm_0_0.vlm"},
	{0x080305C497A0F344ULL, "dae000_gamlen6102525.cl"},
	{0x08032A1A400EC688ULL, "gen00fl_isabela6085616.cl"},
	{0x08033A37A172C990ULL, "vxht_keep_n_6lp_0_0.vlm"},
	{0x08034C7FF04EEA12ULL, "hlac_dr_slavers_3214_5y.dds"},
	{0x0803A6E9FC3A8CC6ULL, "vxdr_0_deeproads_ec0_0_0.vlm"},
	{0x080465D4F055EA6DULL, "201246150.ogg"},
	{0x0805189B895521A4ULL, "10482083.ogg"},
	{0x0806041EC235DC1CULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.impa.evt"},
	{0x08060F1C698AF2EBULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicerchair_0.pwk"},
	{0x0806539E3614E7FAULL, "t3_boy_b00.tnt"},
	{0x080677077DDFCD69ULL, "571674441.ogg"},
	{0x0806A286B61183FAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_basic.msh"},
	{0x0806B086B6119B33ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_basic.mao"},
	{0x0806B486B611A240ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_basic.mmh"},
	{0x08070A8D5C4B0FB9ULL, "246469178.ogg"},
	{0x080711DCC281A008ULL, "art\\vfx\\materials\\vfx_smoke_puffy.mao"},
	{0x08073368A616BDEEULL, "37199919.ogg"},
	{0x080781E508C637F9ULL, "234216133.ogg"},
	{0x0807A3ACA952E9B8ULL, "vxht_chantry_23v_0_0.vlm"},
	{0x0807FEB7FECD899EULL, "461209484.ogg"},
	{0x08089AC7F455077FULL, "hm_rdr_dockworker2.mor"},
	{0x0808A93BD708DF28ULL, "ht_0_hightown_a2_33_0_ll.mmh"},
	{0x08090EAAFF714C95ULL, "vxw_bonepit_1kc_0_0.vlm"},
	{0x0809958B1554B7BBULL, "6214888_m.fxe"},
	{0x0809F6F7BD5E3699ULL, "follower_banter6172839.cl"},
	{0x080A3CB45DD81CB9ULL, "mag291_vincento6023831.cl"},
	{0x080A5F13F019824FULL, "flip_cover_pressed.cur"},
	{0x080A7F491D875940ULL, "vxca_cavern1_haunted_ckj_0_0.vlm"},
	{0x080A8C23313BE160ULL, "vxw_woundedcoast_1d0_0_0.vlm"},
	{0x080AB1FB34B9F696ULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.impa.ani"},
	{0x080AB8020C7BA420ULL, "gen_im_trs_dal_med_glv_r2.uti"},
	{0x080BA4B30F2D379FULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.dura.evt"},
	{0x080C0B6CF5C34F6DULL, "zz_followers_banter6157667.cl"},
	{0x080C4EC01729F66CULL, "6085256_m.fxe"},
	{0x080C6272F1CE8073ULL, "392896239.ogg"},
	{0x080CA69E10A48420ULL, "mag101_grace6034078.cl"},
	{0x080CC0656500FB37ULL, "vxlt_backalley_n_1ny_0_0.vlm"},
	{0x080D04C10873914DULL, "vxlt_undercityencounter_1qg_0_0.vlm"},
	{0x080D9C379BCB5911ULL, "vxlt_undercitydungeon_156_0_0.vlm"},
	{0x080DB9BD071A438DULL, "vxht_brothel_n_6tv_0_0.vlm"},
	{0x080E20AFF0E5720CULL, "pro000cr_soldier2.utc"},
	{0x080E2B4047B30D90ULL, "zz_prmpt_debug6200941.cl"},
	{0x080E5DD70C67243DULL, "337179379.ogg"},
	{0x080ECD25165E8821ULL, "6074419_m.fxe"},
	{0x080EE281F9886BABULL, "lla_ht_0_hightown_a2_65538_0.dds"},
	{0x080EF867853145FBULL, "w_woundedcoast_4_0.mmh"},
	{0x080F046785315998ULL, "w_woundedcoast_4_0.mao"},
	{0x080F1667853178E5ULL, "w_woundedcoast_4_0.msh"},
	{0x080F3761E8BE3D65ULL, "vxw_bonepit_1x6_0_0.vlm"},
	{0x080F4B3FFD3AFF10ULL, "vxlt_0_lowtown_a2_eiw_0_0.vlm"},
	{0x080F4E62E09388C8ULL, "958720939.ogg"},
	{0x080F5C93A394FA53ULL, "hlac_lt_backalley_n_65536_1z.dds"},
	{0x080F6E6FE0450002ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_act_c.anb"},
	{0x080FC5CBEF20087AULL, "w_sundermount_159_0.msh"},
	{0x080FC5CEA7F188AFULL, "allan_test.arl"},
	{0x080FD3CBEF201FB3ULL, "w_sundermount_159_0.mao"},
	{0x080FD7CBEF2026C0ULL, "w_sundermount_159_0.mmh"},
	{0x08105487EC171298ULL, "art\\characters\\creatures\\shade\\animation\\blendtree\\shade_movement.anb"},
	{0x0810B8E7AA64AE5DULL, "6107717_m.fxe"},
	{0x0811A0EE1A6D6889ULL, "hm_leonato.mor"},
	{0x0811AC149368E210ULL, "vxht_keep_n_6gq_0_0.vlm"},
	{0x081216427419BFB6ULL, "hlac_ht_bartrand_n_3368_0y.dds"},
	{0x08122F9483DAC72AULL, "qun110_petrice6018536.cl"},
	{0x081285B30F332188ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.dura.gad"},
	{0x08130BC257616FD4ULL, "vxca_cavern1_haunted_edy_0_0.vlm"},
	{0x08130C1EC2407435ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.impa.ani"},
	{0x08131FE092BE95E9ULL, "vxlt_0_lowtown_a2_ewp_0_0.vlm"},
	{0x08134CEC6339CC31ULL, "combat_dw_mv_default.gda"},
	{0x08136A61A537B2A4ULL, "6105490_f.fxe"},
	{0x0813D15BF1A12631ULL, "vxht_0_hightown_n_apz_0_0.vlm"},
	{0x0813E933BA199B1BULL, "vxw_encounter2_camp_kz_0_0.vlm"},
	{0x0814B457BD5F12DAULL, "vxht_0_hightown_a2_ccq_0_0.vlm"},
	{0x0814CD4FDF6B2D51ULL, "qun211cr_braggart_ally_f4.utc"},
	{0x081520A01812646BULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12.phy"},
	{0x081646FB5B6EEEB1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beama.msh"},
	{0x081654FB5B6F06CCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beama.mao"},
	{0x081660FB5B6F1B6FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beama.mmh"},
	{0x08167791E36A905CULL, "vxlt_gallowstemplar_n_3rx_0_0.vlm"},
	{0x08167A76E1B90DF9ULL, "hlac_lt_0_lowtown_n_-262144_1z.dds"},
	{0x08170F1EB31FD75CULL, "hlac_lt_darktown_-1_48x.dds"},
	{0x08178087F1676515ULL, "vxw_bonepit_1lb_0_0.vlm"},
	{0x08178FA516C3A4A0ULL, "art\\characters\\weapons\\promo\\shields\\textures\\w_shd_011a_0s.dds"},
	{0x0817AECE2D490709ULL, "follower_banter6172729.cl"},
	{0x0817BEAE3825D5CBULL, "fx_ghosta.mao"},
	{0x0817C2F434C6FA67ULL, "394715659.ogg"},
	{0x0817D9D06C5FCDBAULL, "llac_ht_mansion2b_818_44.dds"},
	{0x0817E558645DBD54ULL, "hlac_lt_0_lowtown_-262144_3x.dds"},
	{0x0817E69BC3DDBBCCULL, "712761899.ogg"},
	{0x0818B9FB34C64187ULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.impa.evt"},
	{0x0818D11327001CDAULL, "vxca_cave_2cq_0_1.vlm"},
	{0x0818E16B4FC8786CULL, "hlac_ht_0_hightown_131071_2x.dds"},
	{0x0818F1260F7D71C0ULL, "vxca_cavern1_haunted_clk_0_0.vlm"},
	{0x081940E73359AA27ULL, "vxht_mansion2_288_0_0.vlm"},
	{0x08195D8681D04B3BULL, "art\\characters\\creatures\\shriek\\animation\\combat\\teleport.ani"},
	{0x0819957B59657988ULL, "default_player.utc"},
	{0x0819B16542C91E5EULL, "hlac_ht_0_hightown_a2_-65536_1x.dds"},
	{0x081AE57A4DAF959AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_bridgebusted.msh"},
	{0x081AF37A4DAFADD3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_bridgebusted.mao"},
	{0x081AF77A4DAFB4E0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_bridgebusted.mmh"},
	{0x081B45B8D6B46ACEULL, "art\\vfx\\combat\\vfx_r_arc_sgrenade_act_c.impa.ani"},
	{0x081B9A325D5C08F7ULL, "hlac_w_encounter1_-131070_0y.dds"},
	{0x081C0452451B5ADDULL, "vxdr_primevaltomb_8y2_0_0.vlm"},
	{0x081C167F59EE0092ULL, "w_encounter2_30_0_ll.mmh"},
	{0x081C3E17831D2407ULL, "vxdr_primevaltomb_3ac_0_0.vlm"},
	{0x081C64040D69D3D4ULL, "6095464_m.fxe"},
	{0x081C94ED978C884FULL, "356182457.ogg"},
	{0x081E560A6D2483A5ULL, "gen00fl_aveline6132934.cl"},
	{0x081ED6734D522E9EULL, "547384812.ogg"},
	{0x081F8D2AB1FF5310ULL, "226344326.ogg"},
	{0x081FDAFB34CC98D0ULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.impa.gad"},
	{0x0820652CD9CAA556ULL, "dae150_harshal6148872.cl"},
	{0x0820A4414F02B14DULL, "213942103.ogg"},
	{0x0820B130FE56B500ULL, "vxlt_0_lowtown_f0o_0_0.vlm"},
	{0x0820D1AF3F8968F6ULL, "17154077.ogg"},
	{0x08211B6A9EDAA80AULL, "vxlt_blackemporium_13e_0_0.vlm"},
	{0x082127854089F5C7ULL, "gen00fl_fenris6110890.cl"},
	{0x0821339B117BF5DDULL, "418222130.ogg"},
	{0x082135975379B2A8ULL, "zz_qcr_debug6056622.cl"},
	{0x08215EF34F3D238EULL, "dae120cr_hm_guard_2.utc"},
	{0x08217186502AD0B4ULL, "vxlt_0_lowtown_docks_a3_cyc_0_0.vlm"},
	{0x08224338E3973EB1ULL, "vxht_0_hightown_n_auy_0_0.vlm"},
	{0x08225FD9C0CBAFC2ULL, "vxht_0_hightown_n_b9q_0_0.vlm"},
	{0x0822C44B667F8B81ULL, "6097558_f.fxe"},
	{0x0823016298BAA865ULL, "ter_stoney2_d.dds"},
	{0x08234A963A789D69ULL, "hlac_ht_0_hightown_a3_-131072_1y.dds"},
	{0x0824763BABE3C616ULL, "6127477_m.fxe"},
	{0x0824FF28AFA3ABA3ULL, "gen00fl_anders6136005.cl"},
	{0x082519921CE4E508ULL, "vxht_keep_6a4_0_0.vlm"},
	{0x0825437F0F9BD9EBULL, "255456765.ogg"},
	{0x0825847069A41BBAULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he03.phy"},
	{0x08258DC1DC9E03CDULL, "6149742_m.fxe"},
	{0x082617B54E063100ULL, "follower_banter6149184.cl"},
	{0x0826489BFD0514EBULL, "873052622.ogg"},
	{0x0826658681DAE332ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\teleport.evt"},
	{0x082671651FDCD222ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_exit.gad"},
	{0x08268D3CC350AC32ULL, "mrl000_marethari6129278.cl"},
	{0x082725A23F8FFCDCULL, "6107446_m.fxe"},
	{0x08287AFA57D5514DULL, "vxlt_0_lowtown_docks_a3_n_app_0_0.vlm"},
	{0x08292D8D78316145ULL, "vxca_caverns1_mines_e7u_0_0.vlm"},
	{0x08293DB8D6C09A1FULL, "art\\vfx\\combat\\vfx_r_arc_sgrenade_act_c.impa.evt"},
	{0x082A5A9E042F9D1DULL, "6015941_f.fxe"},
	{0x082A9D76EB06748DULL, "vxht_brothel_n_6zt_0_0.vlm"},
	{0x082B00375ED8A7A9ULL, "hlac_w_woundedcoast_-65534_0z.dds"},
	{0x082B795AAF8B7DBEULL, "vxw_woundedcoast_17w_0_0.vlm"},
	{0x082B80AE34D364E1ULL, "vxdr_primevalentrance_2q5_0_1.vlm"},
	{0x082B8FA6E705DEBFULL, "character_character_p273.psh.fxo.dx11.assembly"},
	{0x082B967C0434A616ULL, "one200_gold_followers6101131.cl"},
	{0x082BA25BD303A7D9ULL, "hlac_lt_undercityencounter_-1_19y.dds"},
	{0x082BB2D4D16D3D29ULL, "6140319_f.fxe"},
	{0x082BB30F2DCA486CULL, "vxht_mansion2_af_0_0.vlm"},
	{0x082D7977A316CD9AULL, "follower_banter6136720.cl"},
	{0x082D7E426319B2CFULL, "20967362.ogg"},
	{0x082D848681E136C5ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\teleport.gad"},
	{0x082D94651FE32C19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_exit.evt"},
	{0x082DF40C4F9E5442ULL, "vxdr_primevalentrance_37p_0_0.vlm"},
	{0x082E11584E2E6049ULL, "177772745.ogg"},
	{0x082E5CE1B65128F4ULL, "hlac_w_encounter3_-196603_0y.dds"},
	{0x082E6EDAA6312318ULL, "6135595_m.fxe"},
	{0x082EBE11DA29150FULL, "art\\vfx\\combat\\vfx_w_wns_front_close2_c.impa.gad"},
	{0x082F17AC4CFBDD54ULL, "vxht_chantry_n_2ci_0_0.vlm"},
	{0x082F230DF04CCD80ULL, "vxlt_0_lowtown_f5l_0_0.vlm"},
	{0x08301EB8D6C68408ULL, "art\\vfx\\combat\\vfx_r_arc_sgrenade_act_c.impa.gad"},
	{0x0830B6225455CCBDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_highlandplant01_sa.dds"},
	{0x08315829BED140E4ULL, "mcr370_meredith6083626.cl"},
	{0x0831C778207E77A9ULL, "vxht_0_hightown_c5x_0_0.vlm"},
	{0x0831CAF39A154CC0ULL, "mrl310_party_comments6142901.cl"},
	{0x0832E10C9D5C55F7ULL, "6109750_f.fxe"},
	{0x08331A387DFF097DULL, "vxdr_primevalentrance_2st_0_2.vlm"},
	{0x08338B6F0EDAFD88ULL, "vxht_keep_6j5_0_0.vlm"},
	{0x0833E3297713BF94ULL, "lgt270ip_eau_docks_scroll.utp"},
	{0x08341D8CBAAE37A4ULL, "6111323_f.fxe"},
	{0x08344F9AD2C39C6FULL, "6104021_f.fxe"},
	{0x08349E7B597C509CULL, "default_player.mor"},
	{0x0834CFA88946EC3BULL, "6208676_m.fxe"},
	{0x0834D311DA2DA4BCULL, "art\\vfx\\combat\\vfx_w_wns_front_close2_c.impa.evt"},
	{0x0835C567BDED1B54ULL, "6098931_m.fxe"},
	{0x0835E96344E02925ULL, "44354168.ogg"},
	{0x08360B640DCAC701ULL, "636136473.ogg"},
	{0x083663D0C411604BULL, "zz_otl_debug6060672.cl"},
	{0x0836884115650542ULL, "vxlt_0_lowtown_a2_ehz_0_0.vlm"},
	{0x0836E08C837CB8C6ULL, "vxht_chantry_33i_0_0.vlm"},
	{0x08372F911FF6C5C3ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_loop.ani"},
	{0x0837D11740379E64ULL, "6135620_f.fxe"},
	{0x08388D576C3EFD34ULL, "33522670.ogg"},
	{0x08391D01D2F04617ULL, "221582263.ogg"},
	{0x08396D267F14E939ULL, "vxlt_0_lowtown_a3_n_c_fee_0_0.vlm"},
	{0x0839A5A15FAD2243ULL, "w_bonepitdestroyed_a3_33_0.mao"},
	{0x0839A9A15FAD2890ULL, "w_bonepitdestroyed_a3_33_0.mmh"},
	{0x0839B7A15FAD40AAULL, "w_bonepitdestroyed_a3_33_0.msh"},
	{0x0839D729D7447233ULL, "1030072561.ogg"},
	{0x0839EB37A181963EULL, "vxw_woundedcoast_12h_0_0.vlm"},
	{0x083A8A346248EF8BULL, "lightning_f.dds"},
	{0x083A957B121ABB36ULL, "6084430_f.fxe"},
	{0x083A9C651FEDC408ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_exit.ani"},
	{0x083AB90C16423FBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_exit.ani"},
	{0x083AC2A4742BC629ULL, "hlac_lt_0_lowtown_a2_n_-131072_1y.dds"},
	{0x083B5022A3BA0A6BULL, "6175649_f.fxe"},
	{0x083C92DD60182A8CULL, "6067124_m.fxe"},
	{0x083CF961299A5B38ULL, "war_autokillweak.dds"},
	{0x083D706C3E50EA0DULL, "hlac_ht_0_hightown_a3_-65537_2x.dds"},
	{0x083D9B6821B92B6BULL, "6020412_m.fxe"},
	{0x083E259B3C186C0DULL, "mag210_arianni6057338.cl"},
	{0x083F8F4538DE6287ULL, "hlac_lt_0_lowtown_docks_-4_0z.dds"},
	{0x083FB9F2534BC10FULL, "l_env_wal_lt_qcr_mcr_cityfear_end_game.bnk"},
	{0x083FEA7A0A511DF7ULL, "6039304_m.fxe"},
	{0x083FEB21BC7CD387ULL, "hlac_dr_primevaltomb_663_2z.dds"},
	{0x083FEC9D5B5E1638ULL, "668047676.ogg"},
	{0x08405268486BC4FCULL, "926960578.ogg"},
	{0x0840794C4468F539ULL, "hlac_lt_undercityserialkiller_-1_31y.dds"},
	{0x084090E6E3D59767ULL, "172255434.ogg"},
	{0x0841266375D9EE4DULL, "6015692_m.fxe"},
	{0x084152973DF03D40ULL, "hlac_lt_0_lowtown_a2_n_65537_6x.dds"},
	{0x0841634D5FC7CBF0ULL, "w_bonepitmining_a2_218_0.mao"},
	{0x0841674D5FC7D203ULL, "w_bonepitmining_a2_218_0.mmh"},
	{0x0841754D5FC7EA5DULL, "w_bonepitmining_a2_218_0.msh"},
	{0x0841A4A800E06D6AULL, "6098928_m.fxe"},
	{0x0841C60C164874FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_exit.gad"},
	{0x0841E108F5390ABCULL, "6129305_m.fxe"},
	{0x0842718D97803D99ULL, "vxht_0_hightown_bai_0_0.vlm"},
	{0x0842DB11DA39EFD5ULL, "art\\vfx\\combat\\vfx_w_wns_front_close2_c.impa.ani"},
	{0x0843384682256F42ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_7y.dds"},
	{0x0843DA4089747027ULL, "vxlt_hangedman_sm_0_0.vlm"},
	{0x08441B89D81CEBAEULL, "294877139.ogg"},
	{0x08449DE21175BDADULL, "llac_dr_primevaltomb_658_0.dds"},
	{0x0845479120032BBAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_loop.evt"},
	{0x084552697572D146ULL, "vxht_chantry_36j_0_0.vlm"},
	{0x0846435B3FB6FB2FULL, "mcr310_sibling6076414.cl"},
	{0x08464DCF6C44F784ULL, "vxht_0_hightown_a2_c4n_0_0.vlm"},
	{0x08470CE1A498936FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_black.mao"},
	{0x084718E1A498A70CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_black.mmh"},
	{0x08471EE1A498B1FEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_black.msh"},
	{0x08478D499E3801D1ULL, "hlac_ca_caverns1_24421_7z.dds"},
	{0x0847ACF7DF77B6E9ULL, "6095887_m.fxe"},
	{0x08488BD60E42A191ULL, "ss_lowclass_thugs_m_1.wbd"},
	{0x084895D433D11686ULL, "6178428_m.fxe"},
	{0x0848A10C164E541DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_exit.evt"},
	{0x0848FED5B0969F65ULL, "vxw_bonepit_1l2_0_0.vlm"},
	{0x08495608E2A8486BULL, "6070211_m.fxe"},
	{0x0849629E71815FB3ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmloop.ani"},
	{0x0849BCB185808576ULL, "vxht_mansion_n_5bc_0_0.vlm"},
	{0x0849BD97AB247293ULL, "hlac_lt_gallowscourtyard_-65535_2z.dds"},
	{0x0849C00C2417B543ULL, "6173688_m.fxe"},
	{0x0849C27F6867A919ULL, "ico_gloves_medium_4.dds"},
	{0x0849C3066E2EAEBAULL, "character_character_p205.psh.fxo.dx11"},
	{0x0849F05A54F62CA4ULL, "ser221_gamlen6135081.cl"},
	{0x084A99484AC988D6ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_act_c.anb"},
	{0x084B0E2E4D2A382AULL, "556695877.ogg"},
	{0x084B4691200795ADULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_loop.gad"},
	{0x084B4DF9A7080856ULL, "one100dg_nabil6063822.cl"},
	{0x084B5519E8F5A2D9ULL, "llac_ht_0_hightown_131074_0.dds"},
	{0x084BDF66EA257BAFULL, "chargen_preload.gda"},
	{0x084C3281015852D9ULL, "zz_vrc_debug6200216.cl"},
	{0x084C3C81F794842AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_chantryfloor_d.dds"},
	{0x084C61A5B2B8B2D7ULL, "llac_w_bonepit_458758_0.dds"},
	{0x084C7064DCDDBC80ULL, "6135804_f.fxe"},
	{0x084E0234B9E74018ULL, "130874223.ogg"},
	{0x084F677030FF9461ULL, "524477835.ogg"},
	{0x084FE22D737C423CULL, "6113963_f.fxe"},
	{0x08501C23D07BEC84ULL, "1018586397.ogg"},
	{0x08502FFAC622E6E2ULL, "833990235.ogg"},
	{0x08504BB9E7E665DFULL, "405924032.ogg"},
	{0x0850D4778100E0A9ULL, "w_woundedcoast_81_0.rml"},
	{0x0851795538F4F947ULL, "428736614.ogg"},
	{0x0852897DBD97AEDCULL, "6146931_m.fxe"},
	{0x0852ABFB0A12A898ULL, "vxht_bartrand_n_4a3_0_1.vlm"},
	{0x0852BC74900E30F8ULL, "vxw_woundedcoast_1gq_0_0.vlm"},
	{0x0852E0D50E81A09FULL, "6075421_m.fxe"},
	{0x0852E6B734EA97D9ULL, "95733110.ogg"},
	{0x085388944459C49EULL, "hlac_lt_0_lowtown_docks_a3_65533_1x.dds"},
	{0x0853C4234B5AAE86ULL, "codex.gda"},
	{0x08547B78D677ED1CULL, "233853545.ogg"},
	{0x0854BFAC5E3B1004ULL, "vxht_0_hightown_a2_c3o_0_0.vlm"},
	{0x0854C7F299015CE3ULL, "8410723.ogg"},
	{0x0854DAAAE06CB915ULL, "901461952.ogg"},
	{0x085571F46EC4CFCAULL, "vxca_caverns2_78q_0_2.vlm"},
	{0x085663BBB21E15A5ULL, "mag100_cullen6093746.cl"},
	{0x0856834BDA43767DULL, "zz_approval_debug6003608.cl"},
	{0x085683F571D3C590ULL, "vxht_0_hightown_cbi_0_1.vlm"},
	{0x08569E30E50FA21AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stackshelf_d.dds"},
	{0x0856CEF1856DC73EULL, "vxw_woundedcoast_18j_0_0.vlm"},
	{0x0856E90C5D7ECB42ULL, "6166896_m.fxe"},
	{0x0857362A988ABDF4ULL, "388643105.ogg"},
	{0x08575A9E718D8FEAULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmloop.evt"},
	{0x085781C484482949ULL, "6117773_f.fxe"},
	{0x085781CFBF85B83FULL, "489106010.ogg"},
	{0x085788BFE8DC7D09ULL, "vxlt_gallowscourtyard_77c_0_0.vlm"},
	{0x08578F217B6CBF0CULL, "vxht_0_hightown_a3_2xj_0_0.vlm"},
	{0x08579C484AD41841ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_act_c.mmh"},
	{0x0857A954EB620782ULL, "6077130_m.fxe"},
	{0x0857BC70CEE7203EULL, "6129032_f.fxe"},
	{0x0857E23DBAC8C4C3ULL, "zz_followers_banter6157809.cl"},
	{0x08589AFF4113CDB6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelwht_0.pwk"},
	{0x0858F1B1D35735A4ULL, "one300_zevran6089108.cl"},
	{0x0858FBFF8A1DF994ULL, "gen00fl_sebastian6178362.cl"},
	{0x08590CB967408C59ULL, "gen00fl_merrill6130050.cl"},
	{0x085974004B1F5510ULL, "6118091_m.fxe"},
	{0x0859C32978C314BDULL, "966806491.ogg"},
	{0x0859D2297734724DULL, "lgt270ip_eau_docks_scroll.ncs"},
	{0x085A414F765ABA52ULL, "gen00fl_anders6125154.cl"},
	{0x085AC87781092089ULL, "w_woundedcoast_81_0.mao"},
	{0x085AD47781093522ULL, "w_woundedcoast_81_0.mmh"},
	{0x085AD677810938D8ULL, "w_woundedcoast_81_0.msh"},
	{0x085B5D68400E80F0ULL, "vxlt_0_lowtown_a2_n_c_cml_0_0.vlm"},
	{0x085B8C8E0934A5B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyejitter.gad"},
	{0x085BB34E949556C7ULL, "57643385.ogg"},
	{0x085BC035FC4D0357ULL, "w_blightlands_140_0.phy"},
	{0x085BF2A27FD2ED0AULL, "llac_ca_caverns1_24426_9.dds"},
	{0x085C033344643676ULL, "344996939.ogg"},
	{0x085C8A6CFE9B74E2ULL, "w_blightlands_9_0.msh"},
	{0x085C8C6CFE9B7798ULL, "w_blightlands_9_0.mmh"},
	{0x085C986CFE9B8C3BULL, "w_blightlands_9_0.mao"},
	{0x085CF86785736688ULL, "w_woundedcoast_4_0.rml"},
	{0x085D83D9AF9302F0ULL, "331597471.ogg"},
	{0x085E1C9C18410ABAULL, "ts_beasts_02.gda"},
	{0x085E399E7193765DULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmloop.gad"},
	{0x085E4EAE5E2F1799ULL, "hlac_lt_0_lowtown_docks_n_196607_1z.dds"},
	{0x085F5E9D27C83B19ULL, "vxca_caverns2_8xk_0_0.vlm"},
	{0x085F8C6CFE9DADABULL, "w_blightlands_9_0.rml"},
	{0x085FB15CA5B7592EULL, "hlac_lt_gallowscourtyard_-131071_0z.dds"},
	{0x0860582AA51C8954ULL, "vxca_cave_188_0_0.vlm"},
	{0x0860D99CF04E3049ULL, "mrl150_dalish_guard6068673.cl"},
	{0x0860FB5AAF8DC75CULL, "6143631_m.fxe"},
	{0x0861435F0A00F5C9ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyo_0n.dds"},
	{0x086190E8A26D1116ULL, "vxht_mansion2_1ay_0_0.vlm"},
	{0x0861A0E9209DE5FFULL, "625500654.ogg"},
	{0x086207F09D42FD2CULL, "zz_t_trailer_isabela.cut"},
	{0x08628D8E093AC65FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyejitter.ani"},
	{0x0862EB0A7D6C32F8ULL, "vxlt_darktown_590_0_0.vlm"},
	{0x086300B552A5FA87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v4.gad"},
	{0x0863660DC783CE02ULL, "19094807.ogg"},
	{0x08643755776F6FAEULL, "vxlt_warehouse_10l_0_0.vlm"},
	{0x08646ADDE80AC3B3ULL, "static2.psh.fxo"},
	{0x08648F69E20D476FULL, "vxlt_hangedman_n_216_0_0.vlm"},
	{0x0864B2265B479D0CULL, "llac_ht_mansion2_818_44.dds"},
	{0x0864D9EF67B6202AULL, "32253957.ogg"},
	{0x08655C96460378C7ULL, "qcr221_cut_diplomatic_sol6201297.cl"},
	{0x0865C5D6A4297168ULL, "6179003_m.fxe"},
	{0x0865DF36A7F68820ULL, "6121091_m.fxe"},
	{0x0865FA9CDAD29589ULL, "vxlt_gallowscourtyard_78b_0_0.vlm"},
	{0x08661500A4C1025AULL, "vxlt_0_lowtown_docks_a3_ctf_0_0.vlm"},
	{0x0866A06B696CB676ULL, "vxht_mansion_n_5he_0_0.vlm"},
	{0x08670A1736C117D8ULL, "6121925_m.fxe"},
	{0x0867478A01E3757AULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightsand_d.dds"},
	{0x0867F3D87D8F7DE8ULL, "art\\characters\\playercharacter\\elfmale\\cth_coma.msh"},
	{0x086805D87D8F9C79ULL, "art\\characters\\playercharacter\\elfmale\\cth_coma.mao"},
	{0x086811D87D8FB012ULL, "art\\characters\\playercharacter\\elfmale\\cth_coma.mmh"},
	{0x08681C6AFBA1EB1CULL, "gen00fl_sebastian6179875.cl"},
	{0x08685883D3E48ADBULL, "zz_vault_debug6203254.cl"},
	{0x0868800AE7E463E7ULL, "hlac_ca_caverns1_24424_15x.dds"},
	{0x0868C39C9B027CA1ULL, "528381606.ogg"},
	{0x086913941C25A2EEULL, "6070351_m.fxe"},
	{0x08696F27AAE9EE93ULL, "6195820_m.fxe"},
	{0x0869EB1E0BDEA52FULL, "mag311cr_nyssa.utc"},
	{0x0869F5B552AC0694ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v4.evt"},
	{0x086A40C720A9BCD8ULL, "llac_lt_0_lowtown_a3_n_c_0_1.dds"},
	{0x086A6C929CD0A86BULL, "6169041_m.fxe"},
	{0x086A9FF43BEA114BULL, "vxw_bonepitmining_a2_7p_0_0.vlm"},
	{0x086ADB7FB9291E63ULL, "zz_mag_debug6051238.cl"},
	{0x086B0CD05463C02CULL, "vxlt_0_lowtown_docks_n_cou_0_0.vlm"},
	{0x086B12DC0B2884AEULL, "711015632.ogg"},
	{0x086B2F3CA1A7A94DULL, "hlac_ca_caverns2prc_5802_37y.dds"},
	{0x086BC9C87EF7D8EDULL, "hlac_dr_primevaltomb_657_1z.dds"},
	{0x086BD0C01EF0AF9DULL, "dre000_dougal6129779.cl"},
	{0x086BF54E2DBF4282ULL, "llac_ht_chantry_n_2534_8.dds"},
	{0x086C80DC10F3947AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galo2floor.mmh"},
	{0x086C84DC10F39AB1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galo2floor.mao"},
	{0x086C92DC10F3B2C0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galo2floor.msh"},
	{0x086CD0DF4132F106ULL, "6088907_m.fxe"},
	{0x086D58CF568CAEF7ULL, "vxlt_gallowstemplar_f_3e2_0_0.vlm"},
	{0x086D7A249E0099E1ULL, "vxw_bonepitdestroyed_a3_29v_0_0.vlm"},
	{0x086D7CA22B0208D7ULL, "hlac_lt_0_lowtown_docks_-2_4x.dds"},
	{0x086D9DB6A82A53CBULL, "861525396.ogg"},
	{0x086E6708D004F6D7ULL, "512246595.ogg"},
	{0x086EFD932D58B209ULL, "zz_otl_debug6060964.cl"},
	{0x086F4374A1540BEAULL, "ser200_aveline6201545.cl"},
	{0x086FA3654A52F53DULL, "hlac_lt_0_lowtown_a3_n_c_-65538_0x.dds"},
	{0x087019BB3CB5EDF2ULL, "vxlt_backalley_n_1hv_0_0.vlm"},
	{0x087039754A3B170CULL, "art\\characters\\playercharacter\\humanmale\\boo_medb.phy"},
	{0x08704884EF134F4AULL, "vxht_keep_3zn_0_0.vlm"},
	{0x08710A45A2EAE12AULL, "1066186631.ogg"},
	{0x08710A95859F6AC0ULL, "vxlt_0_lowtown_a2_n_c_e89_0_0.vlm"},
	{0x08711AD983BE1D2EULL, "6144096_m.fxe"},
	{0x0871A823180736C1ULL, "zz_dae_debug6131012.cl"},
	{0x0871E0CC4BE9D0FFULL, "vxlt_warehouse_n_12l_0_0.vlm"},
	{0x0872A78CEDA0A6B0ULL, "llac_w_sundermount_0_5.dds"},
	{0x0872AE41DB1CF4DBULL, "465017862.ogg"},
	{0x0872FBBAE5E4A368ULL, "vxlt_undercitydungeon_10h_0_0.vlm"},
	{0x08737E29C1325B09ULL, "6031841_m.fxe"},
	{0x0873A9A15FDEA153ULL, "w_bonepitdestroyed_a3_33_0.rml"},
	{0x0873B72147E9065CULL, "948611953.ogg"},
	{0x0873C1BCC12ED884ULL, "llac_lt_gallowstemplar_f_2383_5.dds"},
	{0x08746A23DB831C87ULL, "fxa_gas_smoke_add.mao"},
	{0x08747437506EC5A3ULL, "vxht_keep_648_0_0.vlm"},
	{0x08748026B9B26552ULL, "6180690_m.fxe"},
	{0x087485E5C30E2CE9ULL, "927798562.ogg"},
	{0x08754BC0056AEA63ULL, "582270531.ogg"},
	{0x0875CF43A5068605ULL, "391285232.ogg"},
	{0x0875EF2FA3D18B8DULL, "vxw_blightlands_ee_0_0.vlm"},
	{0x087634B43F2134D4ULL, "6331090.ogg"},
	{0x087669B464621B75ULL, "vxlt_0_lowtown_a3_n_c_gv5_0_0.vlm"},
	{0x0876F4D1E3C3ECBFULL, "458174934.ogg"},
	{0x087702DA5F5545C8ULL, "vxw_bonepitdestroyed_a3_29s_0_0.vlm"},
	{0x087760F501138267ULL, "wm_ico_sundermount_roll.dds"},
	{0x087785710E1BBB6EULL, "6075051_m.fxe"},
	{0x0877EDB552B836CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v4.ani"},
	{0x08782D1B7AF47448ULL, "dae000cr_grand_cleric6124703.cl"},
	{0x08783BA876172ADBULL, "art\\levels\\races\\proxy\\props\\_textures\\fence_n.dds"},
	{0x0878A2013ACA764BULL, "980530323.ogg"},
	{0x0878B4AE2CA872FCULL, "399496367.ogg"},
	{0x0878C322EBDD8855ULL, "6053272_m.fxe"},
	{0x087956D16C7666F6ULL, "346128455.ogg"},
	{0x0879A912DA3A0288ULL, "905791155.ogg"},
	{0x0879A99CD58E41BDULL, "w_woundedcoast_92_0.mao"},
	{0x0879ACB71508C8D7ULL, "6492903.ogg"},
	{0x0879AD9CD58E48F6ULL, "w_woundedcoast_92_0.mmh"},
	{0x0879B79CD58E59F4ULL, "w_woundedcoast_92_0.msh"},
	{0x087A1755A5BA317CULL, "vxht_0_hightown_c1m_0_0.vlm"},
	{0x087AA27BCD244295ULL, "mag170_thrask6054078.cl"},
	{0x087AAD18700BB7E4ULL, "1061516262.ogg"},
	{0x087AC4A2A95DDE15ULL, "mag291_vincento6027354.cl"},
	{0x087B0136E721D941ULL, "cry_star.mao"},
	{0x087B77D21BDD47AAULL, "default_alpha.mao"},
	{0x087B951B5FDC5DB7ULL, "545884007.ogg"},
	{0x087BE0F8EEF14B2BULL, "7119463.ogg"},
	{0x087BEC018FF6B261ULL, "vxw_bonepitdestroyed_a3_24u_0_0.vlm"},
	{0x087C240E801B3566ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_imp_p.anb"},
	{0x087CAFD477483DDBULL, "hlac_lt_0_lowtown_a2_n_131072_2y.dds"},
	{0x087D0325213284C0ULL, "6075202_m.fxe"},
	{0x087D1CECDB121D8FULL, "6076478_f.fxe"},
	{0x087E03867E48B053ULL, "lt_gallowstemplar_n_8165_ll.mmh"},
	{0x087E109A76414B5AULL, "vxlt_0_lowtown_n_bz4_0_0.vlm"},
	{0x087E416B33EFD078ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_001a_0s.dds"},
	{0x087F51079CD51D9EULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srdb_0d.dds"},
	{0x087F53C58CAAB398ULL, "hlac_lt_warehouse_n_65537_1y.dds"},
	{0x087FB5551BA1F769ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyn_0n.dds"},
	{0x087FD7C3BC548EE0ULL, "vxlt_darktown_5cz_0_0.vlm"},
	{0x088016DD35F9A3D7ULL, "gen_im_trs_chn_junk2_03.uti"},
	{0x08804BF54A935E51ULL, "978512188.ogg"},
	{0x08805AF3F2D91A42ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackl.msh"},
	{0x08805CF3F2D91DF8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackl.mmh"},
	{0x088068F3F2D9319BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackl.mao"},
	{0x088114060C75E285ULL, "vxw_woundedcoast_1gh_0_0.vlm"},
	{0x08817D7069F26EE3ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he03.msh"},
	{0x0881860BF925395FULL, "1027488832.ogg"},
	{0x0881872B7F452901ULL, "one280_emrys6104302.cl"},
	{0x08818B7069F2863DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he03.mmh"},
	{0x0881B76302EBE433ULL, "amb_scene_dishdrop.bnk"},
	{0x0881C2B08253F26BULL, "llac_ht_keep_3743_1.dds"},
	{0x0881DD70832ED287ULL, "hlac_lt_gallowscourtyard_-131073_1z.dds"},
	{0x08820B012392BCCEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04a.phy"},
	{0x0882681559F532FFULL, "396242984.ogg"},
	{0x08827F2BB4FFDF27ULL, "6036464_m.fxe"},
	{0x0882A7A07356FEF5ULL, "348704072.ogg"},
	{0x0882A9908815A268ULL, "6177203_m.fxe"},
	{0x0882E3296E53C803ULL, "tbe300im_maker_dagger.uti"},
	{0x08830A46E4E82D37ULL, "hlac_ca_cave_2194_0x.dds"},
	{0x0883AB6DC30B4343ULL, "110307944.ogg"},
	{0x0883B92C863F78D2ULL, "6042048_m.fxe"},
	{0x0883C11E5A9C5037ULL, "hlac_ca_caverns1_24423_3z.dds"},
	{0x0883D377812C38C9ULL, "w_woundedcoast_81_0.anb"},
	{0x0883E763992800E6ULL, "darktown_male_lowlife_mob6174406.cl"},
	{0x08844B31045E3761ULL, "217970546.ogg"},
	{0x0884610C95C7A40DULL, "vxw_blightlands_nd_0_0.vlm"},
	{0x08853FEBA2510A02ULL, "hlac_w_bonepit_458753_0x.dds"},
	{0x088574B7514B5E48ULL, "vxw_bonepitdestroyed_a3_24p_0_0.vlm"},
	{0x08858593BF0D697CULL, "mag110_cut_abomination6037101.cl"},
	{0x0885E4F7E54A4491ULL, "vxlt_0_lowtown_a3_n_c_hx0_0_0.vlm"},
	{0x088683BF446DE238ULL, "hlac_ht_store_187_0y.dds"},
	{0x08872462C4414976ULL, "vxlt_gallowsprison_n_5qi_0_0.vlm"},
	{0x088735A4A3CA06F9ULL, "zz_fol_debug6092334.cl"},
	{0x08874CE014EAA976ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_tileweb.mao"},
	{0x088758E014EABD15ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_tileweb.mmh"},
	{0x08875AE014EAC08BULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_tileweb.msh"},
	{0x08879F8C4C1D406DULL, "gen_im_trs_krk_hgh_junk1_01.uti"},
	{0x0887C0C13DB90399ULL, "99099205.ogg"},
	{0x0887C43BF2028EFCULL, "329931146.ogg"},
	{0x088809571609D40AULL, "860298652.ogg"},
	{0x0888386CEA88B8C0ULL, "311103735.ogg"},
	{0x088873ECA98BE09BULL, "art\\characters\\playercharacter\\humanmale\\boo_maso.mmh"},
	{0x08887FECA98BF538ULL, "art\\characters\\playercharacter\\humanmale\\boo_maso.mao"},
	{0x08888881A74DC6F4ULL, "6067700_m.fxe"},
	{0x0888893297B049FCULL, "vxht_0_hightown_c4b_0_0.vlm"},
	{0x08888F6A8C294AA2ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_maj_c.impa.gad"},
	{0x088891ECA98C1385ULL, "art\\characters\\playercharacter\\humanmale\\boo_maso.msh"},
	{0x0888B6726957210BULL, "6110221_m.fxe"},
	{0x0888C54023CB71FCULL, "m_mer351ar_cave.wbd"},
	{0x0888FFEEC6E04BEEULL, "6106775_m.fxe"},
	{0x0889270E8025C411ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_imp_p.mmh"},
	{0x0889300A4434418AULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.cess.gad"},
	{0x0889679119B14426ULL, "one140st_charade.stg"},
	{0x0889757AD3366BBFULL, "follower_barks6173406.cl"},
	{0x088988C04CA34E1AULL, "6059995_f.fxe"},
	{0x088994D93D066B69ULL, "dae300_cullen6135889.cl"},
	{0x0889B499251F2206ULL, "6086707_m.fxe"},
	{0x088A1E5443A4DF90ULL, "llac_lt_0_lowtown_n_1_1.dds"},
	{0x088A2A1DD64E02EEULL, "6169761_m.fxe"},
	{0x088A63F3F2E17D83ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackl.phy"},
	{0x088A8ACC3584D86BULL, "532867125.ogg"},
	{0x088B3E638F9AB2B0ULL, "6164056_m.fxe"},
	{0x088B6D3FF9E637F9ULL, "vxw_bonepit_28i_0_0.vlm"},
	{0x088BE56845989741ULL, "vxca_caverns1_mines_gjx_0_0.vlm"},
	{0x088C115C023658E7ULL, "vxw_blightlands_9c_0_0.vlm"},
	{0x088C4EBB857E73E1ULL, "vxdr_primevaltomb_3oz_0_1.vlm"},
	{0x088C74E3C177A489ULL, "ht_0_hightownexplosion_n_35_0_lc.mmh"},
	{0x088C7C4E426FDA5EULL, "vxlt_gallowscourtyard_n_4wd_0_0.vlm"},
	{0x088C8502744E6312ULL, "art\\vfx\\combat\\vfx_c_profane_s_crust_c.mmh"},
	{0x088C94C70F805C6EULL, "vxlt_gallowstemplar_n_379_0_0.vlm"},
	{0x088C9CF76CA6F400ULL, "hlac_lt_gallowstemplar_f_2391_8z.dds"},
	{0x088CB18BEF297694ULL, "6166875_f.fxe"},
	{0x088CC4E178074868ULL, "72743126.ogg"},
	{0x088CFCAB8E2E3361ULL, "w_encounter1_56_0.phy"},
	{0x088D2E29F1EE43C2ULL, "6163903_m.fxe"},
	{0x088D4926934E3BC3ULL, "vxca_caverns1_mines_i2d_0_0.vlm"},
	{0x088D6A813D43217CULL, "rand_highway.gda"},
	{0x088DE45A139DED34ULL, "6167952_m.fxe"},
	{0x088EB7220357866EULL, "vxlt_gallowstemplar_f_3s5_0_0.vlm"},
	{0x088EF56ED615E966ULL, "vxht_0_hightown_n_c76_0_0.vlm"},
	{0x088F1FEBCC0867E0ULL, "131396371.ogg"},
	{0x088F8084E0309F50ULL, "llac_lt_0_lowtown_docks_-65540_1.dds"},
	{0x088FB26A8C2FA499ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_maj_c.impa.evt"},
	{0x088FD50948715FD9ULL, "clf260_harley6089568.cl"},
	{0x088FD7FDC4CE4097ULL, "hlac_ht_bartrand_n_3365_1x.dds"},
	{0x0890130A443A2F21ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.cess.evt"},
	{0x08901E50A2FE8B41ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v3_c.anb"},
	{0x089038BF000A0937ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloora.msh"},
	{0x08904ABF000A279AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloora.mao"},
	{0x08904EBF000A2EA9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloora.mmh"},
	{0x089172A6D8E2540FULL, "83012575.ogg"},
	{0x08919FDC1112E485ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galo2floor.phy"},
	{0x0891AB15B61FA76DULL, "83334929.ogg"},
	{0x0891DD67154F4867ULL, "121215777.ogg"},
	{0x08926AFD3A61CB2BULL, "6111183_m.fxe"},
	{0x0892D2E987BDBC8DULL, "vxw_blightlands_sc_0_0.vlm"},
	{0x089340C7D218AF3CULL, "vxht_0_hightown_bj1_0_0.vlm"},
	{0x0893463862018CECULL, "52362697.ogg"},
	{0x0893DAF2DFF6B871ULL, "107773372.ogg"},
	{0x0894234B42C6859BULL, "art\\vfx\\models\\vfx_forearm_opened.msh"},
	{0x0894354B42C6A4E6ULL, "art\\vfx\\models\\vfx_forearm_opened.mao"},
	{0x0894355394992EFEULL, "gen00fl_anders6119122.cl"},
	{0x0894414B42C6B885ULL, "art\\vfx\\models\\vfx_forearm_opened.mmh"},
	{0x089510F64F220EA6ULL, "fen121im_magisters_key.uti"},
	{0x08955425BE3E3DFEULL, "llac_ca_caverns2prc_5802_11.dds"},
	{0x0895E431D912915CULL, "vxlt_0_lowtown_f50_0_0.vlm"},
	{0x08968CF1D311A701ULL, "316360927.ogg"},
	{0x0896B15F95A93E92ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.gad"},
	{0x0896E0BAF88C08B9ULL, "dae000cr_grand_cleric6153122.cl"},
	{0x089708DA7DD4ED07ULL, "vxdr_primevalentrance_1a9_0_0.vlm"},
	{0x0897164AB0CF0CCFULL, "zz_fen_debug6134335.cl"},
	{0x08972CE794F756E5ULL, "436143794.ogg"},
	{0x089739775B9C0F27ULL, "449884873.ogg"},
	{0x08975AB1022A48D8ULL, "w_bonepitdestroyed_a3_24_0_lc.mmh"},
	{0x089761703B2D6C96ULL, "gen00fl_anders6096483.cl"},
	{0x0897DCFFBD5E63D8ULL, "vxlt_gallowstemplar_5cn_0_0.vlm"},
	{0x0897FB9AF47ED25CULL, "631934082.ogg"},
	{0x08985E842C853643ULL, "vxht_mansion2b_3bw_0_0.vlm"},
	{0x08985EF55735B469ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_019a_0d.dds"},
	{0x0898D1C455F235BDULL, "vxlt_undercityserialkiller_xs_0_0.vlm"},
	{0x089913FE27A0C144ULL, "vxht_0_hightown_a2_bx3_0_0.vlm"},
	{0x0899764E01540D55ULL, "6043981_m.fxe"},
	{0x089A3C882C12BA6AULL, "gen_im_trs_gem_tigerseye_03.uti"},
	{0x089A82B7A5F98F44ULL, "ave200im_donnic_gift.uti"},
	{0x089AB58279FAB481ULL, "6126709_m.fxe"},
	{0x089ACE985BCCE636ULL, "hlac_w_sundermount_131077_0z.dds"},
	{0x089B25B9AF0B4793ULL, "art\\vfx\\textures\\vfx_fire_tile.dds"},
	{0x089B4B00466C9C3BULL, "vxlt_0_lowtown_docks_n_b7g_0_0.vlm"},
	{0x089B55F81E243679ULL, "6083996_m.fxe"},
	{0x089BA4205B6260DCULL, "hlac_ca_caverns2_5802_70x.dds"},
	{0x089C0F86FA6BF770ULL, "dae000_bodhan6146853.cl"},
	{0x089C10337C9248B1ULL, "6085477_m.fxe"},
	{0x089CBA6A8C3A3C88ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_maj_c.impa.ani"},
	{0x089D0B0A4444AC70ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.cess.ani"},
	{0x089D7FBA12E5DEA2ULL, "xenonreload.psh.fxo"},
	{0x089DA6C84CC831F2ULL, "782658592.ogg"},
	{0x089DB45F95AF62C9ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.evt"},
	{0x089DD51A46F1D329ULL, "6038587_m.fxe"},
	{0x089DF7B12F14B786ULL, "hlac_lt_0_lowtown_a2_-131072_4x.dds"},
	{0x089DFF50A30A93BAULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v3_c.mmh"},
	{0x089E3D0817271EB6ULL, "gen00fl_sebastian6178359.cl"},
	{0x089E9D00E0BB9878ULL, "fx_energy01.mao"},
	{0x089FAADB098118CCULL, "vxlt_0_lowtown_h7x_0_0.vlm"},
	{0x089FC80080BDE67AULL, "6104449_m.fxe"},
	{0x08A00290B7449EF2ULL, "zz_vault_debug6203378.cl"},
	{0x08A0177A9D8A1812ULL, "6142310_f.fxe"},
	{0x08A144C679B3D50DULL, "vxw_blightlands_tb_0_0.vlm"},
	{0x08A14A6E57110CDDULL, "sse_ss_revenant_1.gda"},
	{0x08A278657F577A71ULL, "ser221_party_comments6122854.cl"},
	{0x08A27872F2301D99ULL, "lowtown_fem_alien_elf_2_mob6166449.cl"},
	{0x08A28B10817910EEULL, "6170143_m.fxe"},
	{0x08A2C8B1C9367591ULL, "vxlt_0_lowtown_a3_n_c_hb6_0_0.vlm"},
	{0x08A2D1A2E1B67470ULL, "957406270.ogg"},
	{0x08A2D26782F9E0E9ULL, "hlac_ht_0_hightown_a3_-131069_0z.dds"},
	{0x08A2EA0AAE1A8161ULL, "vxlt_0_lowtown_csz_0_0.vlm"},
	{0x08A30C3FCBCD58E7ULL, "443505163.ogg"},
	{0x08A39535FC8A6FAFULL, "w_blightlands_140_0.mao"},
	{0x08A3A135FC8A844CULL, "w_blightlands_140_0.mmh"},
	{0x08A3A735FC8A8E3EULL, "w_blightlands_140_0.msh"},
	{0x08A3A7ADDC4A0AA8ULL, "6156765_m.fxe"},
	{0x08A42299BB3E151EULL, "807062034.ogg"},
	{0x08A44972931293C4ULL, "vxca_caverns2_8pb_0_0.vlm"},
	{0x08A4F7048FEF2D8AULL, "vxht_keep_n_6ai_0_0.vlm"},
	{0x08A5DCADE725ED67ULL, "vxht_brothel_n_6uk_0_0.vlm"},
	{0x08A609E23E87A72CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_manhole.phy"},
	{0x08A61C457E95BC11ULL, "6124050_f.fxe"},
	{0x08A6B3F5A092B6C0ULL, "mag100_thrask6031631.cl"},
	{0x08A70DD7BA3A33EEULL, "6163561_m.fxe"},
	{0x08A72434484D3B37ULL, "425370666.ogg"},
	{0x08A75D85758C35E3ULL, "vxht_mansion_n_4c0_0_0.vlm"},
	{0x08A79EB5B40C5C53ULL, "vxht_chantry_n_2ve_0_0.vlm"},
	{0x08A8179C21EA1390ULL, "hlac_dr_0_deeproads_5845_19x.dds"},
	{0x08A830E752FBCDFDULL, "vxca_caverns2prc_62o_0_1.vlm"},
	{0x08A904A47F409087ULL, "gen00fl_aveline6053347.cl"},
	{0x08A92634A500BB86ULL, "gen00fl_dog6146963.cl"},
	{0x08A962C5600C9B60ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_mushroom_e.dds"},
	{0x08A98A9757BE5AE3ULL, "920824244.ogg"},
	{0x08A99EE98378DE96ULL, "twitch_core_settings.gda"},
	{0x08A9B388A6F69F27ULL, "vxht_mansion2_2v2_0_0.vlm"},
	{0x08A9D03B01DFF91BULL, "319427210.ogg"},
	{0x08A9DC648E967A01ULL, "vxlt_0_lowtown_f0b_0_0.vlm"},
	{0x08AA201D296DE018ULL, "hlac_lt_0_lowtown_docks_a3_262142_0z.dds"},
	{0x08AB2D5A9236D860ULL, "vxlt_darktown_5ry_0_0.vlm"},
	{0x08AB766834C69C2EULL, "409749874.ogg"},
	{0x08AB9C5F95BB7738ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.ani"},
	{0x08AB9FE482F64FEDULL, "vxlt_gallowstemplar_f_3ru_0_0.vlm"},
	{0x08ABB77F51DC8CC1ULL, "115300159.ogg"},
	{0x08ABFD07C266CBFFULL, "6180127_m.fxe"},
	{0x08AC2F78AB408165ULL, "6101391_m.fxe"},
	{0x08AC5671861D0990ULL, "follower_banter6155691.cl"},
	{0x08AD55F63AFD6339ULL, "vxca_cavern1_haunted_fn1_0_0.vlm"},
	{0x08ADC5ECF7088665ULL, "vxlt_0_lowtown_docks_a3_cuz_0_0.vlm"},
	{0x08AE25B7CFFDAF43ULL, "vxlt_gallowstemplar_f_37h_0_0.vlm"},
	{0x08AE8BDE3E5F4642ULL, "6039943_m.fxe"},
	{0x08AEDD02165020B6ULL, "cd_mer141_jansen.wbd"},
	{0x08AF2130BD2FA08EULL, "vxlt_undercitydungeon_17k_0_0.vlm"},
	{0x08AF8931055FD4ABULL, "plt_and000pt_end_stage.plo"},
	{0x08AF92A84E5010F9ULL, "120251091.ogg"},
	{0x08AFBF3D6BD1E575ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax102a_0s.dds"},
	{0x08B02481B604E03CULL, "vxht_0_hightown_bp7_0_0.vlm"},
	{0x08B0251787F69F39ULL, "vxca_caverns2_62f_0_1.vlm"},
	{0x08B0568F37572622ULL, "lt_gallowstemplar_f.arl"},
	{0x08B05BBDFDFD1293ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargegreen.phy"},
	{0x08B0A9F29B54C77CULL, "357781841.ogg"},
	{0x08B0AAEE8FCA3772ULL, "mag170_hugh6080709.cl"},
	{0x08B0B83859A728EBULL, "869624794.ogg"},
	{0x08B0F43E64F8312EULL, "247145366.ogg"},
	{0x08B110754A7268D7ULL, "art\\characters\\playercharacter\\humanmale\\boo_medb.mmh"},
	{0x08B114754A726FE4ULL, "art\\characters\\playercharacter\\humanmale\\boo_medb.mao"},
	{0x08B126754A728E49ULL, "art\\characters\\playercharacter\\humanmale\\boo_medb.msh"},
	{0x08B16D420CD4ABA9ULL, "qun221_petrice6093256.cl"},
	{0x08B185CA5A3A4842ULL, "w_bonepitdestroyed_a3_32_0.rml"},
	{0x08B20C9277D08E1CULL, "147841104.ogg"},
	{0x08B25B63D4E938E2ULL, "568637517.ogg"},
	{0x08B2874A1F9F2A89ULL, "vxht_mansion2_2o7_0_0.vlm"},
	{0x08B2CE37C79A0557ULL, "519455370.ogg"},
	{0x08B371F443A95045ULL, "6172568_m.fxe"},
	{0x08B3C4BB1E913AE4ULL, "417230139.ogg"},
	{0x08B5222D2F09D027ULL, "hlac_ht_bartrand_n_3381_1z.dds"},
	{0x08B561FE945BDB98ULL, "6147190_m.fxe"},
	{0x08B58402747166B9ULL, "art\\vfx\\combat\\vfx_c_profane_s_crust_c.anb"},
	{0x08B5C34023F25C76ULL, "m_mer351ar_cave.bnk"},
	{0x08B5E60B59B01B3BULL, "zz_dae_debug6029904.cl"},
	{0x08B5F7B80B8CF244ULL, "vxht_0_hightown_a2_bf1_0_0.vlm"},
	{0x08B60D1BEB0DBA2FULL, "character_character_p224.psh.fxo.dx11"},
	{0x08B6E4B1E4CED667ULL, "dae000_hate_enemies6167601.cl"},
	{0x08B79CCDFF02BAC5ULL, "6208421_m.fxe"},
	{0x08B80AD2AF2DE9F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_buildup.ani"},
	{0x08B84E41808C9281ULL, "vxlt_0_lowtown_f5a_0_0.vlm"},
	{0x08B881890AEE1109ULL, "ave110tr_enter_dpatrol.ncs"},
	{0x08B8C639AE487450ULL, "vxca_caverns2prc_7tu_0_1.vlm"},
	{0x08B8D28A5CEADD7BULL, "6144516_m.fxe"},
	{0x08B8DFFD6D6E53EAULL, "165556938.ogg"},
	{0x08B951300A3DC2DEULL, "gen00fl_sebastian6188657.cl"},
	{0x08B96B385B25CE34ULL, "tbe000im_unknown_gloves.uti"},
	{0x08B99F37842CF0E0ULL, "vxlt_darktown_5wf_0_0.vlm"},
	{0x08B9FE21BD3C107AULL, "vxca_caverns1_mines_ehs_0_0.vlm"},
	{0x08BA299C92CDD5A7ULL, "dae000_maren6148431.cl"},
	{0x08BA8AA935462198ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4.phy"},
	{0x08BB27F5D67E072AULL, "gen00fl_bethany6097824.cl"},
	{0x08BB68F9B4A8995DULL, "819019979.ogg"},
	{0x08BBDBE3ED5F52F0ULL, "vxlt_backalley_g_1kd_0_0.vlm"},
	{0x08BC2FB0771369A9ULL, "667240170.ogg"},
	{0x08BD62100D1F6723ULL, "6020544_m.fxe"},
	{0x08BD77C8B0F8C598ULL, "gen00fl_anders6125215.cl"},
	{0x08BDA461E0176816ULL, "w_blightlands_17_0.rml"},
	{0x08BDAE00E0D665FAULL, "fx_energy01.dds"},
	{0x08BE3E2F611BA09FULL, "6027053_m.fxe"},
	{0x08BE695267949A0DULL, "6084360_m.fxe"},
	{0x08BFD338DEBD4CA5ULL, "lgt310cr_mag_melson_guard1.utc"},
	{0x08C019D16924F2D5ULL, "6180954_m.fxe"},
	{0x08C02D13761DACC4ULL, "594721002.ogg"},
	{0x08C04F23B0368EE9ULL, "6118999_m.fxe"},
	{0x08C05DB50677FA37ULL, "vxlt_0_lowtown_docks_a3_bqq_0_0.vlm"},
	{0x08C07D7A271451A0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd.phy"},
	{0x08C09339EBBB5C7FULL, "vxht_0_hightown_a2_9zm_0_0.vlm"},
	{0x08C0DAB3720C5888ULL, "fex_ruin_circ_b.mmh"},
	{0x08C0F8B3720C8BB2ULL, "fex_ruin_circ_b.msh"},
	{0x08C1043568F61D18ULL, "620373766.ogg"},
	{0x08C1DBF588159BDCULL, "832353037.ogg"},
	{0x08C20A0123C89E91ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04a.mmh"},
	{0x08C20BC304B3AECDULL, "vxht_mansion2b_1fw_0_0.vlm"},
	{0x08C21A746A58F241ULL, "lt_0_lowtown_docks_a3_n_9_0_lc.mmh"},
	{0x08C2240123C8CB4FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04a.msh"},
	{0x08C239C15B003694ULL, "mer240_jansen6050409.cl"},
	{0x08C26FCA5A489597ULL, "w_bonepitdestroyed_a3_32_0.msh"},
	{0x08C26FE2B020A69EULL, "mrl000_marethari6129285.cl"},
	{0x08C27BC0DFEDD16BULL, "dae000ip_crowd_mf1_talk11.utp"},
	{0x08C281CA5A48B4FAULL, "w_bonepitdestroyed_a3_32_0.mao"},
	{0x08C285CA5A48BB09ULL, "w_bonepitdestroyed_a3_32_0.mmh"},
	{0x08C2B9DE949CB42EULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocksmall.phy"},
	{0x08C2F64D7BC16CAAULL, "vxht_chantry_36f_0_0.vlm"},
	{0x08C30EA706F03AE8ULL, "mcr370_desten6087319.cl"},
	{0x08C38B9288D8ED2DULL, "qcr221cs_keep_intro.cut"},
	{0x08C3E9007966D3DFULL, "vxlt_0_lowtown_a2_csh_0_0.vlm"},
	{0x08C410D87DDE0D0DULL, "art\\characters\\playercharacter\\elfmale\\cth_coma.phy"},
	{0x08C424885CDA2158ULL, "6024529_m.fxe"},
	{0x08C46994FD830AC4ULL, "vxht_0_hightown_a2_bk6_0_0.vlm"},
	{0x08C46A92FAA82E28ULL, "cs_pro000cs_avaline_intro.bnk"},
	{0x08C4826F97F88D53ULL, "vxht_chantry_n_2lg_0_0.vlm"},
	{0x08C4A5C5686A9EADULL, "vxlt_gallowstemplar_f_6t6_0_0.vlm"},
	{0x08C502D2AF38662AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_buildup.evt"},
	{0x08C55E4EB694942AULL, "hlac_lt_0_lowtown_docks_n_-65539_6z.dds"},
	{0x08C59A4408D14554ULL, "vxca_caverns2prc_aeu_0_0.vlm"},
	{0x08C5CEEC2265B28DULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_screen_n.dds"},
	{0x08C659CE4A4F4C19ULL, "6127676_m.fxe"},
	{0x08C6E22300E38C5BULL, "6101227_m.fxe"},
	{0x08C6FEF729BEA01AULL, "6067867_m.fxe"},
	{0x08C744674EE0F5FBULL, "6141616_m.fxe"},
	{0x08C744A84CA2F4BBULL, "hlac_lt_gallowstemplar_n_2391_3x.dds"},
	{0x08C7B8B60BC1137EULL, "ts_templars_acc_props.gda"},
	{0x08C7D71F7C53121EULL, "vxlt_0_lowtown_n_bxz_0_0.vlm"},
	{0x08C8450514AA3820ULL, "6111092_m.fxe"},
	{0x08C89EC795A2886AULL, "ave121_brennan6105926.cl"},
	{0x08C8D56303281991ULL, "amb_scene_dishdrop.wbd"},
	{0x08C93AA8A243A831ULL, "mag101cr_alain.utc"},
	{0x08C93F833A4BDA25ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic_2.mao"},
	{0x08C96A890AFC5DECULL, "ave110tr_enter_dpatrol.utt"},
	{0x08C9A6ED7FE58C8FULL, "629556143.ogg"},
	{0x08C9D258AA203FC4ULL, "vxlt_gallowstemplar_5ji_0_0.vlm"},
	{0x08CA2D5BE348CF7AULL, "1019429390.ogg"},
	{0x08CA4DC0DF556B70ULL, "vxlt_backalley_g_1pe_0_0.vlm"},
	{0x08CAD14AAF9216E1ULL, "hlac_lt_0_lowtown_docks_196606_0y.dds"},
	{0x08CB3638A6C4D3B0ULL, "6018544_m.fxe"},
	{0x08CB39BEB1E3A184ULL, "494174111.ogg"},
	{0x08CB4F7CB7EF47A5ULL, "6096248_m.fxe"},
	{0x08CB965173FA1183ULL, "vxca_caverns2prc_9le_0_0.vlm"},
	{0x08CBAEB15DFF066AULL, "6178536_m.fxe"},
	{0x08CBC09A49E575F2ULL, "nat380cr_temmerin.utc"},
	{0x08CBE1D2AF3E4C9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_buildup.gad"},
	{0x08CC07BF85B0B486ULL, "299148302.ogg"},
	{0x08CC204A40804CDCULL, "6154198_m.fxe"},
	{0x08CC6015E4E57CF9ULL, "al_rogue_talents.gda"},
	{0x08CD121159D0CAD3ULL, "hlac_lt_0_lowtown_docks_a3_-65540_2y.dds"},
	{0x08CD2183D1BCD677ULL, "vxlt_0_lowtown_docks_n_b7s_0_0.vlm"},
	{0x08CD38B1780F4264ULL, "887352302.ogg"},
	{0x08CD555E4D46ABD4ULL, "vxlt_0_lowtown_a2_n_c_hnn_0_0.vlm"},
	{0x08CDDF9236B29C7BULL, "152058437.ogg"},
	{0x08CE14593902D616ULL, "art\\characters\\creatures\\mubari\\animation\\com_cavort.evt"},
	{0x08CE8A7E3DB23225ULL, "vxlt_0_lowtown_a3_n_c_ep0_0_1.vlm"},
	{0x08CEC66B759F2D2BULL, "hlac_w_bonepitmining_a2_5_0x.dds"},
	{0x08CF29B5FA6C3A04ULL, "vxca_cavern1_haunted_ckn_0_0.vlm"},
	{0x08CF94ECA9C81BD0ULL, "art\\characters\\playercharacter\\humanmale\\boo_maso.phy"},
	{0x08D02669DA78D9B4ULL, "43750690.ogg"},
	{0x08D0DCC8C364B6EAULL, "vxht_brothel_n_5cc_0_0.vlm"},
	{0x08D14D5B0B0CA5B3ULL, "6171394_m.fxe"},
	{0x08D1EE95EA8574E3ULL, "hlac_lt_gallowscourtyard_-65537_4z.dds"},
	{0x08D210181C86A14BULL, "family_base.gda"},
	{0x08D24F3C9E9E43F7ULL, "482631263.ogg"},
	{0x08D2AD5806577BDCULL, "gen00fl_fenris6113869.cl"},
	{0x08D2F64548E0D117ULL, "vxlt_0_lowtown_n_eqo_0_0.vlm"},
	{0x08D2FE891E5751ABULL, "6080578_f.fxe"},
	{0x08D30389B0A242A5ULL, "vxlt_0_lowtown_f4z_0_0.vlm"},
	{0x08D3F9C825D194C0ULL, "1045838744.ogg"},
	{0x08D450C10E502355ULL, "6143344_m.fxe"},
	{0x08D4D0823979C1EEULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srdb_0t.dds"},
	{0x08D4D9CA278AC882ULL, "vxht_chantry_n_336_0_0.vlm"},
	{0x08D4E0E3A0D11935ULL, "gen00fl_sebastian6187752.cl"},
	{0x08D5077B25AE268EULL, "follower_barks6153564.cl"},
	{0x08D50B593908E5A1ULL, "art\\characters\\creatures\\mubari\\animation\\com_cavort.gad"},
	{0x08D539BF666062E3ULL, "mcr371_scared_mage_m6086909.cl"},
	{0x08D5CC11AFAD9F66ULL, "6090362_m.fxe"},
	{0x08D5E7CECCE78503ULL, "6128367_m.fxe"},
	{0x08D5F58FC0A1E741ULL, "w_woundedcoast_78_0.rml"},
	{0x08D62E404B808961ULL, "mag131_keran6086265.cl"},
	{0x08D632D3C0BA6773ULL, "lla_ht_0_hightown_1_0.dds"},
	{0x08D64037795A799FULL, "zz_dae_debug6029816.cl"},
	{0x08D64C95E17380B1ULL, "follower_banter6163561.cl"},
	{0x08D793F636B39FDDULL, "6214947_m.fxe"},
	{0x08D7B625A3F68100ULL, "mbv.cut"},
	{0x08D7BA8D41FAEF1CULL, "hlac_lt_0_lowtown_a3_n_c_65536_12y.dds"},
	{0x08D7CD9A9D4B9D81ULL, "6140136_m.fxe"},
	{0x08D83007CBDE1277ULL, "vxca_caverns1_mines_cka_0_0.vlm"},
	{0x08D857FF0764F837ULL, "580427038.ogg"},
	{0x08DA325B808CA097ULL, "820695290.ogg"},
	{0x08DA597E0A84F2EAULL, "509173713.ogg"},
	{0x08DA6F64448F79DDULL, "hlac_lt_0_lowtown_docks_a3_-196610_1y.dds"},
	{0x08DAA289D1745E46ULL, "hm_sib_tobrius.mor"},
	{0x08DAB45FE64A68EBULL, "6068908_m.fxe"},
	{0x08DB030216759C24ULL, "cd_mer141_jansen.bnk"},
	{0x08DBC73B3F3CC454ULL, "vxlt_0_lowtown_a2_n_c_hio_0_0.vlm"},
	{0x08DBDAC990D3A0BDULL, "plt_lgt210pt_qun_blkpwdr.plo"},
	{0x08DC0C59390F064FULL, "art\\characters\\creatures\\mubari\\animation\\com_cavort.ani"},
	{0x08DC5BBCFDEFB1D8ULL, "6080646_m.fxe"},
	{0x08DC67D7F1FAD8EEULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha16a.phy"},
	{0x08DC811D1A5F12A1ULL, "ser270_cut_aftermath6162278.cl"},
	{0x08DCEFBA5570036BULL, "hlac_lt_0_lowtown_docks_196608_1z.dds"},
	{0x08DCFA84A0B73743ULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf.ani"},
	{0x08DD956AAAC2FCA6ULL, "vxht_keep_73q_0_0.vlm"},
	{0x08DD963E189FE766ULL, "6126809_m.fxe"},
	{0x08DD9FB815286BF9ULL, "gen_im_trs_dal_hvy_cht_r2.uti"},
	{0x08DDA5A5543F141DULL, "vxlt_gallowstemplar_f_5k8_0_0.vlm"},
	{0x08DDCF3ED3C2BE84ULL, "vxht_keep_n_709_0_0.vlm"},
	{0x08DDE99A64065148ULL, "472745481.ogg"},
	{0x08DE6993A64EE285ULL, "384732723.ogg"},
	{0x08DEC1E9EAA2C6ADULL, "gen00fl_aveline6121992.cl"},
	{0x08DEE47DDDAAF8A8ULL, "gen00fl_fenris6113906.cl"},
	{0x08DFADE598969467ULL, "6106873_m.fxe"},
	{0x08DFBF37FD02BE33ULL, "6174094_m.fxe"},
	{0x08DFCB9B661D6C52ULL, "character_character_p50.psh.fxo.dx11.assembly"},
	{0x08E009553EF5F6D0ULL, "95817530.ogg"},
	{0x08E0159EE5DE1EB6ULL, "vxlt_0_lowtown_docks_a3_cyq_0_0.vlm"},
	{0x08E03775D091177FULL, "gen00fl_isabela6090321.cl"},
	{0x08E0C9CEAD0E22ACULL, "art\\characters\\playercharacter\\humanfemale\\arm_nuda.msh"},
	{0x08E0CFCEAD0E2C9EULL, "art\\characters\\playercharacter\\humanfemale\\arm_nuda.mmh"},
	{0x08E0DBCEAD0E4135ULL, "art\\characters\\playercharacter\\humanfemale\\arm_nuda.mao"},
	{0x08E11284F1C7C5C4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_temp_rock_n.dds"},
	{0x08E1DDC990D811EAULL, "plt_lgt210pt_qun_blkpwdr.nss"},
	{0x08E26D08A4663622ULL, "one300_zevran6071479.cl"},
	{0x08E29DE070D5FD64ULL, "vxlt_gallowstemplar_n_2mx_0_0.vlm"},
	{0x08E2F75DF1157FAFULL, "hlac_lt_0_lowtown_a2_65535_4z.dds"},
	{0x08E3092FA63E9920ULL, "art\\vfx\\materials\\vfx_flare01a_add.mao"},
	{0x08E333F44EFCB9F4ULL, "6088839_m.fxe"},
	{0x08E352E7D2EBF215ULL, "vxw_bonepitmining_a2_1oa_0_0.vlm"},
	{0x08E37A61BEE1AE8CULL, "179992175.ogg"},
	{0x08E458F16E8C4D3BULL, "art\\characters\\weapons\\longswords\\w_lsw_ls103a.phy"},
	{0x08E45A5D6B0C9A5CULL, "995673915.ogg"},
	{0x08E481C345E8F87DULL, "vxlt_0_lowtown_docks_a3_n_bsn_0_0.vlm"},
	{0x08E4F2461097A36DULL, "vxca_caverns2_96l_0_0.vlm"},
	{0x08E4F5C5C1B09FC5ULL, "vxca_caverns1_mines_ehx_0_0.vlm"},
	{0x08E507916B69FA48ULL, "849999264.ogg"},
	{0x08E53D38122A0F90ULL, "hlac_ht_0_hightown_131074_0y.dds"},
	{0x08E59A7A27339F35ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd.msh"},
	{0x08E59C7A2733A2ABULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd.mmh"},
	{0x08E5A87A2733B748ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd.mao"},
	{0x08E608C09A437C23ULL, "vxlt_gallowscourtyard_n_7dk_0_0.vlm"},
	{0x08E67FE9952E1D59ULL, "dae021_bureaucrat_1.utc"},
	{0x08E68BB00EE229C2ULL, "hlac_lt_gallowstemplar_f_2383_11y.dds"},
	{0x08E6B6128E0C70C4ULL, "vxlt_gallowstemplar_5po_0_0.vlm"},
	{0x08E6E0E23EBEF8F7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_manhole.mmh"},
	{0x08E6E4E23EBEFF04ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_manhole.mao"},
	{0x08E6F6E23EBF1E69ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_manhole.msh"},
	{0x08E7317AC3419C70ULL, "vxlt_backalley_g_1zc_0_0.vlm"},
	{0x08E79DC3103202D9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2bwin.msh"},
	{0x08E7A0BE29932393ULL, "gen00fl_varric6088830.cl"},
	{0x08E7A7C3103213C7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2bwin.mmh"},
	{0x08E7ABC310321A14ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2bwin.mao"},
	{0x08E7DD56C8FED284ULL, "6068464_m.fxe"},
	{0x08E82C6E169B0534ULL, "6085928_m.fxe"},
	{0x08E884A87370F222ULL, "vxht_0_hightown_a3_b4p_0_0.vlm"},
	{0x08E8B16DEE358294ULL, "l_rdr370ar_finale.bnk"},
	{0x08E99B84180EE28CULL, "6076860_f.fxe"},
	{0x08E9AA1F04E6F861ULL, "ss_circlemages_f_1.bnk"},
	{0x08E9D83F7CCDB45EULL, "18872200.ogg"},
	{0x08EA333A8836E423ULL, "6169193_m.fxe"},
	{0x08EA5C92FAC8E5D2ULL, "cs_pro000cs_avaline_intro.wbd"},
	{0x08EB0EE03363D3BFULL, "718107386.ogg"},
	{0x08EB1284A0C39D3AULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf.evt"},
	{0x08EBD960BD0862FDULL, "mage_forceresistance.dds"},
	{0x08EC441951C23B8AULL, "jagged_rock.mao"},
	{0x08ECCD7E8C2C6ECCULL, "dae100_gamlen_letter6146996.cl"},
	{0x08ED0E8698CEBDD5ULL, "vxlt_warehouse_n_z9_0_0.vlm"},
	{0x08ED8C7DE86BDDC7ULL, "hlac_ht_0_hightown_a3_-65537_2z.dds"},
	{0x08EDD6F4B041C6FFULL, "hf_cgn_006.mop"},
	{0x08EE101BBBD88E04ULL, "hlac_lt_0_lowtown_docks_n_-4_5y.dds"},
	{0x08EE84ADC485AF0BULL, "vxht_0_hightown_a2_akj_0_0.vlm"},
	{0x08EE978950878757ULL, "hlac_lt_0_lowtown_a2_n_-131071_1z.dds"},
	{0x08EEB5E17A94117AULL, "sum_conj_03.mao"},
	{0x08EF3C575E6E24B8ULL, "6173085_m.fxe"},
	{0x08EF71F37BA1D91EULL, "gen_im_trs_krk_g_dagger.uti"},
	{0x08EFBAA8EF39F642ULL, "and111_refugee_barks6111540.cl"},
	{0x08EFC44FEBC982CDULL, "lt_0_lowtown_docks_n_112_0.mmh"},
	{0x08EFC80161DFE5A3ULL, "face_face_p19.psh.fxo.dx11"},
	{0x08EFC84FEBC9891EULL, "lt_0_lowtown_docks_n_112_0.mao"},
	{0x08EFD64FEBC9A153ULL, "lt_0_lowtown_docks_n_112_0.msh"},
	{0x08F001AB8E82B89DULL, "w_encounter1_56_0.mao"},
	{0x08F005AB8E82BFD6ULL, "w_encounter1_56_0.mmh"},
	{0x08F00FAB8E82D0D4ULL, "w_encounter1_56_0.msh"},
	{0x08F04A7940120BE2ULL, "6160959_m.fxe"},
	{0x08F0733AEFCEB895ULL, "follower_banter6169475.cl"},
	{0x08F079CDDC543563ULL, "6074372_m.fxe"},
	{0x08F08581B4637BC2ULL, "759626661.ogg"},
	{0x08F09A4C3FFDAD0EULL, "33704797.ogg"},
	{0x08F0C6FED60678DCULL, "685754883.ogg"},
	{0x08F11184A0C8072DULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf.gad"},
	{0x08F181A87D3E68CAULL, "294247523.ogg"},
	{0x08F1A5D140F2D274ULL, "6083939_m.fxe"},
	{0x08F2F3A037DF10FDULL, "vxlt_0_lowtown_docks_a3_n_bvm_0_0.vlm"},
	{0x08F3CCE06EA9132CULL, "ener_lit.mao"},
	{0x08F3DBC366DAB670ULL, "vxht_chantry_n_353_0_1.vlm"},
	{0x08F3FA6795E0BF4AULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_lf.gad"},
	{0x08F4EBE2EBF282EBULL, "hlac_lt_gallowscourtyard_-65536_3x.dds"},
	{0x08F4F22958F89AEFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_channelwater_n.dds"},
	{0x08F5E9B764FABE87ULL, "38979494.ogg"},
	{0x08F5E9FEAD9CC1ECULL, "242997766.ogg"},
	{0x08F6A87BD823A34EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley_2.mao"},
	{0x08F807D59E87B0BCULL, "37313870.ogg"},
	{0x08F80F99D8491D06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_action.evt"},
	{0x08F811E6437A9F0FULL, "and340ip_brimstone.utp"},
	{0x08F81464DAF616D3ULL, "and231cr_lyrium_smuggler.utc"},
	{0x08F8AAA62DC1BDD1ULL, "qun221_petrice_temple.stg"},
	{0x08F8F24CF2E1B36EULL, "595816543.ogg"},
	{0x08F93E0C0D8A0A3AULL, "llac_lt_0_lowtown_a2_n_-1_2.dds"},
	{0x08F97E53755DF891ULL, "vxht_brothel_n_6dr_0_0.vlm"},
	{0x08F985C21204A547ULL, "12002886.ogg"},
	{0x08FA3F4353D5F023ULL, "290116499.ogg"},
	{0x08FADD6795E6ACE1ULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_lf.evt"},
	{0x08FB127FE8EF8838ULL, "follower_banter6155769.cl"},
	{0x08FB334C437AD544ULL, "vxlt_0_lowtown_a2_hg4_0_0.vlm"},
	{0x08FB3B9E8BE6CC7FULL, "hlac_lt_gallowscourtyard_n_-1_0x.dds"},
	{0x08FBE4D72924C2FCULL, "hlac_ca_cavern1_haunted_24422_4z.dds"},
	{0x08FBF9CC9A0F7C1EULL, "453353414.ogg"},
	{0x08FC31F0FAD06DD3ULL, "hf_ave1_raider_assassin.mor"},
	{0x08FC749944EFB73EULL, "702908570.ogg"},
	{0x08FD9AD1E2463BD7ULL, "pow_in_clo.mao"},
	{0x08FE55B71DE4BFE9ULL, "vxlt_0_lowtown_byq_0_0.vlm"},
	{0x08FEBC5632082462ULL, "6165659_m.fxe"},
	{0x08FF23155590F382ULL, "6195876_m.fxe"},
	{0x08FF25A09F523589ULL, "6201224_m.fxe"},
	{0x08FF2699D84F6351ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_action.gad"},
	{0x08FF36EC69394A08ULL, "228146478.ogg"},
	{0x08FF92A29E15B289ULL, "6135774_m.fxe"},
	{0x08FFAB7B4AB27CE8ULL, "366330823.ogg"},
	{0x08FFD101D5D9E922ULL, "follower_banter6148736.cl"},
	{0x08FFFC4CBB81C489ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.cess.ani"},
	{0x090017AE32E8EAF4ULL, "6086639_m.fxe"},
	{0x090023AE9F31206EULL, "vxlt_gallowstemplar_n_3o1_0_0.vlm"},
	{0x0900C74E22A9541CULL, "vxlt_gallowscourtyard_6xr_0_0.vlm"},
	{0x0900F53F14B031D9ULL, "6064722_m.fxe"},
	{0x09010C4B27621E51ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs107a_0d.dds"},
	{0x09016C708CC238D9ULL, "6087228_m.fxe"},
	{0x0902B8DE94D296F1ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocksmall.mmh"},
	{0x0902BF1030852DA9ULL, "ser221_gamlen6162352.cl"},
	{0x0902C4DE94D2AA92ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocksmall.mao"},
	{0x0902D2DE94D2C2AFULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrocksmall.msh"},
	{0x0902D75D567EB0C1ULL, "vxht_keep_64b_0_0.vlm"},
	{0x0902D84FFFCE0D9EULL, "131265196.ogg"},
	{0x090305A49349490CULL, "hlac_lt_0_lowtown_docks_-131075_0z.dds"},
	{0x09031E7AA4867CE3ULL, "6168042_m.fxe"},
	{0x0903277524E03988ULL, "6123350_m.fxe"},
	{0x09037195AF232B29ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_p.impa.ani"},
	{0x09046D75DB1DB18CULL, "mer320_hubert6090238.cl"},
	{0x09048F344068900CULL, "6217099_m.fxe"},
	{0x0904ABD2D0AC6882ULL, "55650100.ogg"},
	{0x0905053BD616D4C2ULL, "vxht_mansion_n_6ma_0_0.vlm"},
	{0x09052799D853D0FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_action.ani"},
	{0x09054072D2211B65ULL, "l_qun101ar_talvashoth_cave.wbd"},
	{0x0907174CBB88110BULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.cess.gad"},
	{0x090733D8EABEAF84ULL, "922868106.ogg"},
	{0x0907F03067541111ULL, "vxht_brothel_n_6aq_0_0.vlm"},
	{0x0907F4B65F43DB42ULL, "gen_im_wep_mag_sta_act0_02.uti"},
	{0x090822556B6B0934ULL, "gen00fl_isabela6075299.cl"},
	{0x09089C555617253FULL, "6078354_m.fxe"},
	{0x0908C5C4EA308FB1ULL, "fonts_pseudo.gfx"},
	{0x0908D56795F2DC30ULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_lf.ani"},
	{0x0908E08B67F3C4B3ULL, "347571273.ogg"},
	{0x09091EE6438928BAULL, "and340ip_brimstone.ncs"},
	{0x09097334E79DCC00ULL, "hlac_ca_caverns2_5802_38x.dds"},
	{0x0909DBB579B2B3DFULL, "da2terrain_terrain_tools_p1.psh.fxo"},
	{0x090A1B393C27EDF5ULL, "6170319_m.fxe"},
	{0x090A8C95AF2977ABULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_p.impa.gad"},
	{0x090B0494A2454FEDULL, "6180095_m.fxe"},
	{0x090BCC6AFB7FFA56ULL, "gen00fl_isabela6196859.cl"},
	{0x090C1E64EC31A994ULL, "714964153.ogg"},
	{0x090C3D191D8AEC79ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_htwood_sa.dds"},
	{0x090C783066127B94ULL, "6128276_m.fxe"},
	{0x090CB26149DE4171ULL, "102010252.ogg"},
	{0x090CDE37553FAD92ULL, "211618669.ogg"},
	{0x090CFA95B58B2DEAULL, "art\\characters\\creatures\\dragon\\animation\\p_wkf.gad"},
	{0x090D58B5AFFC349FULL, "560690973.ogg"},
	{0x090E144CBB8E2B70ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.cess.evt"},
	{0x090E3EBC81E83C11ULL, "mrl351cr_faketamlen.utc"},
	{0x090E4FFCD9D06A9CULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d04_0t.dds"},
	{0x090EF1E3A1CDC6ACULL, "43165186.ogg"},
	{0x090F14B8F51E8434ULL, "vxlt_gamlenshouse_n_2v_0_0.vlm"},
	{0x090F85729E111483ULL, "6175152_m.fxe"},
	{0x090FA1933AE96E41ULL, "264958685.ogg"},
	{0x091038DAD0C27FA6ULL, "vxht_0_hightown_n_56w_0_0.vlm"},
	{0x09103FF16EB15DDAULL, "art\\characters\\weapons\\longswords\\w_lsw_ls103a.msh"},
	{0x09104B5C9792E945ULL, "vxca_caverns1_mines_ege_0_0.vlm"},
	{0x091051F16EB17C20ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls103a.mmh"},
	{0x09107FF8C737BAE9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_1_sylvesta.ani"},
	{0x09108995AF2DDE10ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_p.impa.evt"},
	{0x091120DC55ED98BCULL, "hlac_ca_caverns1_mines_24421_0y.dds"},
	{0x0911493A4874C941ULL, "vxht_keep_61a_0_0.vlm"},
	{0x0911876775619ABCULL, "161326788.ogg"},
	{0x0911CE1C87EBEF8AULL, "6153342_m.fxe"},
	{0x091254BDFE4FD108ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargegreen.mmh"},
	{0x091260BDFE4FE5ABULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargegreen.mao"},
	{0x091272BDFE500432ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargegreen.msh"},
	{0x091294076A3A60E2ULL, "gen00fl_varric6076094.cl"},
	{0x0912B4BAEC579FACULL, "zz_dre_debug6025524.cl"},
	{0x0912F646ED821C47ULL, "vxht_mansion2b_3bk_0_0.vlm"},
	{0x0913144F063A2EB0ULL, "318382847.ogg"},
	{0x0913674C743D88FEULL, "71061690.ogg"},
	{0x091381D9B81DF811ULL, "dae100_alistair_drunk6135156.cl"},
	{0x0913CF55BDB3EB2BULL, "dre000_dougal6142217.cl"},
	{0x0913DD95B5911A81ULL, "art\\characters\\creatures\\dragon\\animation\\p_wkf.evt"},
	{0x0913E4B5A03E9FF7ULL, "6118027_m.fxe"},
	{0x09147011C55021F7ULL, "zz_dae_debug6049946.cl"},
	{0x091490C31058DAFCULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2bwin.phy"},
	{0x0914D135921B2C65ULL, "vxw_bonepitmining_a2_1o1_0_0.vlm"},
	{0x09152E2A3F1341D1ULL, "hlac_dr_0_deeproads_5843_1x.dds"},
	{0x091565CBD10D7B82ULL, "hlac_lt_0_lowtown_docks_a3_196608_1z.dds"},
	{0x09158EAF07152654ULL, "vxlt_0_lowtown_a2_n_c_hus_0_0.vlm"},
	{0x09159CA643BC1DF5ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastgravel_n.dds"},
	{0x09159EBA607DA9AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180.gad"},
	{0x0915A03105B6AF78ULL, "plt_and000pt_end_stage.nss"},
	{0x0915E5D5E91B2DF2ULL, "vxca_cavern1_haunted_gk4_0_0.vlm"},
	{0x0916085350C7A0F0ULL, "462820969.ogg"},
	{0x091624E1248DA001ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_3y.dds"},
	{0x0916620D594A2991ULL, "vxht_brothel_n_6jp_0_0.vlm"},
	{0x091689A935947EB3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4.mmh"},
	{0x09169A120545D868ULL, "vxlt_0_lowtown_a3_n_c_hli_0_0.vlm"},
	{0x0916A26030A19705ULL, "seb_module.dds"},
	{0x0916B785A8F3BD93ULL, "mer320_hubert6089916.cl"},
	{0x0916E78FC0D96630ULL, "w_woundedcoast_78_0.msh"},
	{0x0916F58FC0D97E4AULL, "w_woundedcoast_78_0.mmh"},
	{0x0916F98FC0D98481ULL, "w_woundedcoast_78_0.mao"},
	{0x09179AF8C73E076BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_1_sylvesta.gad"},
	{0x0918A950422363D5ULL, "vfx_chunkyparticle_p2.vsh.fxo.dx11.assembly"},
	{0x0918F81414B8738EULL, "vxlt_0_lowtown_a3_n_c_hqh_0_0.vlm"},
	{0x0919755BAB640D0BULL, "gen00fl_aveline6140807.cl"},
	{0x091980620E9724DFULL, "692639367.ogg"},
	{0x09198719195A4ED0ULL, "6135075_f.fxe"},
	{0x0919D03F542F7C8AULL, "803181148.ogg"},
	{0x091A2F43FDBAB068ULL, "984050000.ogg"},
	{0x091A633AF78EE314ULL, "63339503.ogg"},
	{0x091A860B059BCD2DULL, "165680980.ogg"},
	{0x091B814006639EB9ULL, "zz_dae_debug6029807.cl"},
	{0x091C81BA60839741ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180.evt"},
	{0x091D66D7F2326DB1ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha16a.mmh"},
	{0x091D80D7F2329A6FULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha16a.msh"},
	{0x091DD9491335088EULL, "mag170_thrask6054013.cl"},
	{0x091DFB20A12B2D2AULL, "6144082_f.fxe"},
	{0x091F05DFC61BCE1FULL, "hlac_lt_gamlenshouse_n_-1_3x.dds"},
	{0x091F33335018ADABULL, "hlac_w_sundermount_-65539_0y.dds"},
	{0x091FBCAB6A8CECE0ULL, "ser191_emeric6104714.cl"},
	{0x091FF81CB459DDF4ULL, "gen00fl_anders6128380.cl"},
	{0x09203FF979E2080AULL, "6156042_m.fxe"},
	{0x0920965050E39758ULL, "6160764_m.fxe"},
	{0x09209661E06BCE63ULL, "w_blightlands_17_0.msh"},
	{0x0920A461E06BE5BDULL, "w_blightlands_17_0.mmh"},
	{0x0920A861E06BECCEULL, "w_blightlands_17_0.mao"},
	{0x0920FB0DC39C3C43ULL, "hlac_lt_0_lowtown_docks_a3_65534_3z.dds"},
	{0x0921A6CEAD457E39ULL, "art\\characters\\playercharacter\\humanfemale\\arm_nuda.phy"},
	{0x0921AA8AD9E1C2E3ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_4webdetail.mao"},
	{0x0921AE8AD9E1C930ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_4webdetail.mmh"},
	{0x0921BC8AD9E1E14AULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_4webdetail.msh"},
	{0x0921D595B59D4AD0ULL, "art\\characters\\creatures\\dragon\\animation\\p_wkf.ani"},
	{0x0921DCBE56A67AA2ULL, "6129847_f.fxe"},
	{0x0922F18CA7E9AB39ULL, "gen00fl_bethany6101199.cl"},
	{0x09235798B3956E42ULL, "vxlt_undercitydungeon_159_0_0.vlm"},
	{0x0923AE6CFA1ED6E5ULL, "6133787_m.fxe"},
	{0x0923BE551260B5D2ULL, "gen00fl_aveline6106950.cl"},
	{0x0923C8CF3FC9CA13ULL, "325931453.ogg"},
	{0x09240A7269EAEB64ULL, "dae000cr_grand_cleric6187933.cl"},
	{0x09242578B7FA9FC2ULL, "317868142.ogg"},
	{0x0924358DEA88285FULL, "hlac_lt_0_lowtown_a2_n_1_0x.dds"},
	{0x09245E78556B5EE8ULL, "lt_0_lowtown_docks_a3_209_0.msh"},
	{0x09246F5A32AA3CAFULL, "475785724.ogg"},
	{0x09247078556B7D79ULL, "lt_0_lowtown_docks_a3_209_0.mao"},
	{0x09247C78556B9112ULL, "lt_0_lowtown_docks_a3_209_0.mmh"},
	{0x0925C2496640E865ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hanginga_n.dds"},
	{0x0925C62023AF5749ULL, "gen00fl_anders6136018.cl"},
	{0x0925F3C5CEE21E6EULL, "594645931.ogg"},
	{0x09264AB000F83BFFULL, "600476901.ogg"},
	{0x0926E03C22253DE1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep_3.mao"},
	{0x0926F7C911960A0AULL, "vxht_keep_n_64b_0_0.vlm"},
	{0x0927113EBEF251EFULL, "301848913.ogg"},
	{0x09271273FFB00AD2ULL, "6077040_m.fxe"},
	{0x0928C64E931EC094ULL, "hlac_lt_darktown_-1_53x.dds"},
	{0x0928CDADF149A2E2ULL, "179264182.ogg"},
	{0x09290E925C7E9E91ULL, "vxw_sundermount_20b_0_0.vlm"},
	{0x092979BA608E1390ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180.ani"},
	{0x09299DB5C9D5F11AULL, "543998019.ogg"},
	{0x0929CB6DEE6D4646ULL, "l_rdr370ar_finale.wbd"},
	{0x092A5204097BC7BBULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.dura.evt"},
	{0x092AD1426CE72E73ULL, "hlac_lt_0_lowtown_65538_2y.dds"},
	{0x092B542C2CA8987FULL, "326166667.ogg"},
	{0x092B65115441C8A8ULL, "33490386.ogg"},
	{0x092B8C388CBA6C95ULL, "vxw_bonepitmining_a2_1pl_0_0.vlm"},
	{0x092C181A7A0B2667ULL, "vxw_sundermount_2cq_0_0.vlm"},
	{0x092DE4AE9AB73EF1ULL, "character_character_p257.psh.fxo.dx11.assembly"},
	{0x092E06A2DBA0C200ULL, "mcr371_post_harvester6077595.cl"},
	{0x092E1701A69C9589ULL, "6117675_m.fxe"},
	{0x092E2D96EF3BB149ULL, "vxlt_hangedman_n_229_0_0.vlm"},
	{0x092E30203BEB0B51ULL, "589952063.ogg"},
	{0x092E765A4823B229ULL, "162721642.ogg"},
	{0x092EA162D5A6911CULL, "vxlt_darktown_5ko_0_1.vlm"},
	{0x092F6C7ECBDA20C7ULL, "332940664.ogg"},
	{0x092FDEB239CC94E4ULL, "6016946_m.fxe"},
	{0x09304B04098027FCULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.dura.gad"},
	{0x0930A331EDB4037AULL, "6163280_m.fxe"},
	{0x0930B81F05231213ULL, "ss_circlemages_f_1.wbd"},
	{0x093133282E56F44CULL, "134779904.ogg"},
	{0x093146B6AE7FC6ECULL, "vxht_keep_63z_0_0.vlm"},
	{0x0931913F7DC3E3B8ULL, "6092128_m.fxe"},
	{0x093269E4C70B8623ULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench1f_sebastianintro.ani"},
	{0x0932AE148F48B8F5ULL, "492131645.ogg"},
	{0x0932BD01D51C799CULL, "6067810_m.fxe"},
	{0x093327891123DB07ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_sepulcher_n.dds"},
	{0x09334C9662E492BDULL, "hlac_ht_0_hightown_a3_-65536_0y.dds"},
	{0x09334F2604B1DD84ULL, "hlac_lt_0_lowtown_n_196607_0y.dds"},
	{0x093350034A6013F7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scraplight_dp.dds"},
	{0x0933586A9FFD5511ULL, "hlac_ht_keep_3732_16x.dds"},
	{0x093399ED33AB7F2AULL, "static_static_p18.psh.fxo"},
	{0x0933BB6069D49A0CULL, "vxht_0_hightown_a3_bi9_0_0.vlm"},
	{0x093419526302D12EULL, "277671422.ogg"},
	{0x0934D2AF8333AA0EULL, "fen210_dying_magister6108617.cl"},
	{0x09354C06C7933A43ULL, "gen00fl_aveline6139519.cl"},
	{0x093562222FEB2F00ULL, "6172279_m.fxe"},
	{0x093663E10F24A0DFULL, "87844684.ogg"},
	{0x09374A04098644FAULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.dura.ani"},
	{0x0937784675EC83A6ULL, "6125703_m.fxe"},
	{0x0937806F4E74B711ULL, "vxw_sundermount_25a_0_0.vlm"},
	{0x0937C4F37462E24AULL, "zz_mcr_debug6088425.cl"},
	{0x0937FE69F863063EULL, "vxdr_0_deeproads_a27_0_0.vlm"},
	{0x09382B39962FEDCAULL, "6157193_m.fxe"},
	{0x093890478465264AULL, "277609582.ogg"},
	{0x0938C9DB571203ABULL, "vxht_bartrand_n_68l_0_0.vlm"},
	{0x09392CB8B7FBC49CULL, "881418220.ogg"},
	{0x0939BFE96F29212BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_010a_0t.dds"},
	{0x0939D8FED66C898DULL, "468557509.ogg"},
	{0x093A89F76C013EE7ULL, "vxw_sundermount_2dp_0_0.vlm"},
	{0x093AB078F709FF82ULL, "hlac_ht_chantry_n_2534_42z.dds"},
	{0x093B8E9012C2C1C3ULL, "6149371_m.fxe"},
	{0x093BA856F8679C59ULL, "vxlt_0_lowtown_n_h29_0_0.vlm"},
	{0x093BE53CF33F1F5EULL, "gen00fl_aveline6120819.cl"},
	{0x093C5C498A44BCEEULL, "vxlt_undercityencounter_rr_0_0.vlm"},
	{0x093C7DCC59BE2FC5ULL, "6120857_m.fxe"},
	{0x093CA45158D5725AULL, "vxht_playershouse_n_1qk_0_0.vlm"},
	{0x093D4D7DF6401732ULL, "dae000cr_grand_cleric.utc"},
	{0x093DF691DBB39A1DULL, "vxht_bartrand_n_3i4_0_0.vlm"},
	{0x093DFB7393D174E0ULL, "dae210_gate_guard6084319.cl"},
	{0x093E42175CCFD9A4ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb106.mao"},
	{0x093E64D77A91180CULL, "w_bonepitdestroyed_a3_66_0.msh"},
	{0x093E6AD77A9122FEULL, "w_bonepitdestroyed_a3_66_0.mmh"},
	{0x093E76D77A913695ULL, "w_bonepitdestroyed_a3_66_0.mao"},
	{0x093F68977E890CD2ULL, "6115509_f.fxe"},
	{0x0940031C6FB9FC38ULL, "6097228_m.fxe"},
	{0x09406F5EC0348183ULL, "cd_mage110cs_cut_abomination.wbd"},
	{0x094173AAF0043D70ULL, "hlac_ht_brothel_n_3369_0z.dds"},
	{0x0942CC1FE2635FAAULL, "383236737.ogg"},
	{0x0942E03DD629FBF8ULL, "hlac_ca_caverns1_24425_8z.dds"},
	{0x0942ED5E747F37D0ULL, "vxlt_darktown_2xx_0_0.vlm"},
	{0x09438CBC0A8A1B58ULL, "hlac_ca_caverns2_5802_3z.dds"},
	{0x0943CE49EBEFBB23ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_appd_0s.dds"},
	{0x09446B3F9CE4A5DAULL, "162602421.ogg"},
	{0x09451A72D256BE0FULL, "l_qun101ar_talvashoth_cave.bnk"},
	{0x09453221EAD8641BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_loop.ani"},
	{0x09455DCEA01A9FA0ULL, "vxht_chantry_n_27q_0_0.vlm"},
	{0x09455E43B3146D07ULL, "6106936_m.fxe"},
	{0x09455F9630667FC2ULL, "zz_followers_banter6158937.cl"},
	{0x09456F3A6C272C51ULL, "vxht_bartrand_n_6ii_0_0.vlm"},
	{0x0945BC7CCD3215B0ULL, "6120015_f.fxe"},
	{0x0945E0C67A986A9EULL, "6072119_m.fxe"},
	{0x09463EA7681E4996ULL, "follower_barks6153821.cl"},
	{0x09464C67B5E005C6ULL, "6162254_m.fxe"},
	{0x09464F8F8CAFDC7CULL, "6167215_m.fxe"},
	{0x0946A2D692E83609ULL, "vxdr_primevaltomb_86g_0_0.vlm"},
	{0x094758961D61F27BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cookerpots.msh"},
	{0x09476A961D6210C6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cookerpots.mao"},
	{0x094776961D622565ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cookerpots.mmh"},
	{0x094780E4C71E090DULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench1f_sebastianintro.gad"},
	{0x094793DA80105E24ULL, "vxlt_0_lowtown_a2_fip_0_0.vlm"},
	{0x09481FA3A170784BULL, "art\\levels\\races\\design\\_textures\\prp_herbs_dp.dds"},
	{0x09483CF82670E58AULL, "6189370_m.fxe"},
	{0x09485F17FAAA44D9ULL, "605983637.ogg"},
	{0x0948678B045E13BCULL, "hlac_ca_caverns1_24426_11y.dds"},
	{0x0948FBD45DF75767ULL, "vxw_sundermount_2is_0_0.vlm"},
	{0x09494844A792FDFBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_basic.phy"},
	{0x0949489BA03A8451ULL, "static_static_p4.psh.fxo.dx11.assembly"},
	{0x09497760E60A722EULL, "vxlt_darktown_5bw_0_0.vlm"},
	{0x0949A4DDC7DC78C2ULL, "vxdr_primevaltomb_vz_0_0.vlm"},
	{0x094A4125FD614E1DULL, "vxht_0_hightown_a3_cbh_0_0.vlm"},
	{0x094A53913EC395F7ULL, "6140615_m.fxe"},
	{0x094B162E4ACB8ADAULL, "vxht_playershouse_n_1td_0_0.vlm"},
	{0x094B46194D377268ULL, "309207643.ogg"},
	{0x094B69AD9268DBE0ULL, "hlac_lt_0_lowtown_a2_-65537_2z.dds"},
	{0x094CA362970ED71EULL, "vxht_playershouse_n_1o5_0_0.vlm"},
	{0x094CBAC91441DA9BULL, "character_character_p105.psh.fxo.dx11.assembly"},
	{0x094CC2FE71F23702ULL, "229209327.ogg"},
	{0x094D186BFA341D89ULL, "vxlt_gallowscourtyard_7hr_0_0.vlm"},
	{0x094D9C4CD6929788ULL, "857409106.ogg"},
	{0x094DD51128DC6C1BULL, "hlac_lt_0_lowtown_docks_196607_2z.dds"},
	{0x094DEDC749CFA753ULL, "gen00fl_aveline6132864.cl"},
	{0x094E8006A5BEB57DULL, "vxht_bartrand_n_5pi_0_0.vlm"},
	{0x094EA6624C9D0CA7ULL, "art\\vfx\\combat\\vfx_w_wns_left_b_close1_c.anb"},
	{0x094F5B0AC3198F21ULL, "421065555.ogg"},
	{0x094F7A078C8BFDF4ULL, "sky_w_blightlands.mao"},
	{0x09513F2D79A09949ULL, "6041543_m.fxe"},
	{0x09514704D4C5B219ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.impa.ani"},
	{0x095195171BE0BCB6ULL, "48012353.ogg"},
	{0x0952042F9F725F1CULL, "llac_lt_0_lowtown_n_0_4.dds"},
	{0x0952B8187C6CA60BULL, "6134825_m.fxe"},
	{0x0952F9532667FC54ULL, "6126977_m.fxe"},
	{0x0952FD1338263FE1ULL, "116379404.ogg"},
	{0x09536009FA21CEE6ULL, "1008723500.ogg"},
	{0x0953D8802E107614ULL, "vxdr_0_deeproads_cxq_0_0.vlm"},
	{0x0953E1175E1D44D1ULL, "vxht_bartrand_n_6rh_0_0.vlm"},
	{0x09542CA9D85704D1ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.impa.ani"},
	{0x0954664E6F340611ULL, "art\\characters\\playercharacter\\shared\\textures\\bdy_bnb_0d.dds"},
	{0x0954E3B12931E0E1ULL, "vxht_0_hightown_n_a2l_0_0.vlm"},
	{0x09550731D6D3CFEFULL, "seb321_leliana6180966.cl"},
	{0x095553657AC8C01FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_enter.gad"},
	{0x09555D32604B3AEDULL, "vrc321_varric.bnk"},
	{0x0955686FF43933B4ULL, "451044958.ogg"},
	{0x09556B12B776E717ULL, "6121760_m.fxe"},
	{0x0955966942BC451FULL, "dae100_bran6118307.cl"},
	{0x095674A125F8020FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.dura.gad"},
	{0x09576DB14FED6FE7ULL, "vxw_sundermount_2rr_0_0.vlm"},
	{0x0957C75C1A3480E5ULL, "qun230ar_cavern.are"},
	{0x09582204D4CB925BULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.impa.gad"},
	{0x095843CB301FECEAULL, "zz_lgt_1_debug6140405.cl"},
	{0x0958647AD1FBD057ULL, "6177889_f.fxe"},
	{0x09586DAD61F4E53EULL, "gen00fl_bethany6102771.cl"},
	{0x09587F32604DAA55ULL, "vrc321_varric.cnv"},
	{0x0958D1E7D4857F09ULL, "vxlt_0_lowtown_docks_bat_0_0.vlm"},
	{0x095901F1D96BDA0FULL, "dae000_arishok6067668.cl"},
	{0x09594CE8CC5A2419ULL, "hlac_lt_gallowscourtyard_n_0_1y.dds"},
	{0x09595921EAE94FA5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_loop.gad"},
	{0x095A34DED322D747ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medj_0e.dds"},
	{0x095A37A9D85B8373ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.impa.gad"},
	{0x095AF70759364410ULL, "hlac_w_bonepit_262149_0z.dds"},
	{0x095B06B8836CE88CULL, "vxht_chantry_n_2te_0_1.vlm"},
	{0x095B1C484E91AC65ULL, "vxlt_0_lowtown_docks_a3_c1n_0_0.vlm"},
	{0x095B35F1F28FB7CEULL, "zz_approval_debug6133717.cl"},
	{0x095B792F1CC9722FULL, "follower_banter6172323.cl"},
	{0x095BF092C64AB8DDULL, "6123221_m.fxe"},
	{0x095C120EE88405DCULL, "561135746.ogg"},
	{0x095C3BC7BC7402B9ULL, "llac_ca_caverns1_mines_24425_13.dds"},
	{0x095C48657ACECC0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_enter.evt"},
	{0x095C7C81D3B18979ULL, "pro000_colonel_ewald6040605.cl"},
	{0x095C85B4CE9BAE7CULL, "139084830.ogg"},
	{0x095C89A125FC91BCULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.dura.evt"},
	{0x095CB5B2F699529FULL, "qun100_mage_party_comment6019114.cl"},
	{0x095CEB923637FB96ULL, "vxw_bonepitdestroyed_a3_2a5_0_0.vlm"},
	{0x095DBF670B215889ULL, "ser100_party_comments6195135.cl"},
	{0x095DCB9ACB1C7361ULL, "dae000_gamlen6121565.cl"},
	{0x095E0C4A177BB403ULL, "mrl000_marethari6129365.cl"},
	{0x095F13C566DFE509ULL, "6147123_f.fxe"},
	{0x095F1752B721CCC9ULL, "700348072.ogg"},
	{0x095F3F04D4D1E240ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.impa.evt"},
	{0x095F8BBC8148C740ULL, "vxlt_gamlenshouse_n_31_0_0.vlm"},
	{0x095F91EF63F668FAULL, "vxlt_0_lowtown_a2_hfe_0_0.vlm"},
	{0x095FD381929B53D9ULL, "6156629_m.fxe"},
	{0x09613164CE86EE0EULL, "vxlt_0_lowtown_a3_n_c_hjm_0_0.vlm"},
	{0x096134A9D8619CD8ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.impa.evt"},
	{0x0961A28A69FE3F14ULL, "primitives_primitive_p6.psh.fxo.dx11.assembly"},
	{0x0962557B6CBDA052ULL, "6153535_m.fxe"},
	{0x096255AD908FF959ULL, "genip_sarcophagous.utp"},
	{0x0962B023485BEFDDULL, "hlac_lt_0_lowtown_docks_131070_1x.dds"},
	{0x0962C85C1A3E89CBULL, "qun230ar_cavern.ncs"},
	{0x09635E895A5BFD9CULL, "hlac_w_blightlands_327679_0y.dds"},
	{0x096360E3C4B54BD1ULL, "w_bonepitdestroyed_a3_209_0.msh"},
	{0x09636EE3C4B563ECULL, "w_bonepitdestroyed_a3_209_0.mao"},
	{0x09637AE3C4B5778FULL, "w_bonepitdestroyed_a3_209_0.mmh"},
	{0x0963BA99A0B3FE71ULL, "vxlt_gallowscourtyard_74w_0_1.vlm"},
	{0x0963CACA1C5751A0ULL, "318326508.ogg"},
	{0x0963DEBB268E07D1ULL, "12470751.ogg"},
	{0x096400E4A3A2460EULL, "vxw_encounter2_r8_0_0.vlm"},
	{0x096461B9E2409B55ULL, "302829900.ogg"},
	{0x096545C1D89E646DULL, "vxlt_gallowstemplar_n_3qt_0_0.vlm"},
	{0x0965615EC0538631ULL, "cd_mage110cs_cut_abomination.bnk"},
	{0x0965936D497BD1C0ULL, "hlac_lt_darktown_-1_47y.dds"},
	{0x0965E353319A50C5ULL, "docks_male_human_worker_sta6175994.cl"},
	{0x096641B7926E33A3ULL, "vxht_mansion2_31f_0_0.vlm"},
	{0x0966CBCFF5E767A3ULL, "364560151.ogg"},
	{0x0967B4B747AC5597ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgtp_0n.dds"},
	{0x0967C8893111742FULL, "hlac_lt_0_lowtown_n_-65537_6x.dds"},
	{0x0967CEC340CB291AULL, "mrl150_merrill6077735.cl"},
	{0x0967FE3FA83980E3ULL, "cod_hst_woundedcoast.utp"},
	{0x0968098F0503A207ULL, "vxlt_backalley_n_1u4_0_0.vlm"},
	{0x096819CD15FEE920ULL, "vxdr_slavers_1t8_0_1.vlm"},
	{0x09682240F2B12C1FULL, "ash_01_l2.mao"},
	{0x0968860F0F379008ULL, "mag291_demon_marethari6027972.cl"},
	{0x09688A7E370EED6EULL, "1021285711.ogg"},
	{0x0968A693007B8579ULL, "6208951_m.fxe"},
	{0x0968C85C1A42F512ULL, "qun230ar_cavern.lst"},
	{0x0968D6A024CF10E3ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_stars.mao"},
	{0x09692FA96B93A5AAULL, "6135092_m.fxe"},
	{0x0969517DF6655963ULL, "dae000cr_grand_cleric.bnk"},
	{0x096959E5AD1CA1C8ULL, "533095993.ogg"},
	{0x096979732770B526ULL, "hightown_male_elfservnt_mob6170078.cl"},
	{0x0969CDB5BB838245ULL, "hlac_lt_undercityserialkiller_-1_36x.dds"},
	{0x096A10DA2710E905ULL, "vxlt_0_lowtown_docks_a3_n_d56_0_0.vlm"},
	{0x096A30657ADAE165ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_enter.ani"},
	{0x096A321C88F0EC9BULL, "921153275.ogg"},
	{0x096A626FB840D15DULL, "6128321_m.fxe"},
	{0x096A91A12608DCD5ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.dura.ani"},
	{0x096AE609E425BF51ULL, "ser221cr_gascard.utc"},
	{0x096B333CB42246B9ULL, "vxht_chantry_n_2av_0_0.vlm"},
	{0x096B45EF6657FE31ULL, "ave200_donnic6112157.cl"},
	{0x096BCB608489EB61ULL, "w_encounter2_camp_12_0_lc.mmh"},
	{0x096C2D45DBE20649ULL, "6033762_m.fxe"},
	{0x096C337DF6675B93ULL, "dae000cr_grand_cleric.cnv"},
	{0x096C3EDC7CECFB64ULL, "dae200_cullen6140270.cl"},
	{0x096CD932EE6CF419ULL, "vxdr_primevaltomb_6po_0_0.vlm"},
	{0x096D91DCBB2204A6ULL, "vxca_caverns1_mines_e65_0_0.vlm"},
	{0x096D9B066C091DCBULL, "llac_lt_gallowscourtyard_131071_0.dds"},
	{0x096DC78E4DC8C864ULL, "vxlt_0_lowtown_n_eix_0_0.vlm"},
	{0x096E24C32C2BCFBFULL, "544562243.ogg"},
	{0x096E4F0CE9392C90ULL, "w_bonepitmining_a2_242_0.msh"},
	{0x096E5D0CE93944AAULL, "w_bonepitmining_a2_242_0.mmh"},
	{0x096E610CE9394BE1ULL, "w_bonepitmining_a2_242_0.mao"},
	{0x096EDC05BBEB0BF6ULL, "127586829.ogg"},
	{0x096FD509F40FC0ACULL, "6161598_m.fxe"},
	{0x09704C42032D0569ULL, "dae000ip_crowd_mf5_talk17.utp"},
	{0x0970510266B87667ULL, "6170616_m.fxe"},
	{0x09706BD16B4E8B5EULL, "vxlt_gallowscourtyard_4xb_0_0.vlm"},
	{0x0970A0EAA08A7D99ULL, "hlac_lt_0_lowtown_65538_3x.dds"},
	{0x0970A1BE7336FDAEULL, "1015487318.ogg"},
	{0x0970A8F1AD69047DULL, "609194303.ogg"},
	{0x0970BC3A11FCA714ULL, "vxdr_0_deeproads_crw_0_0.vlm"},
	{0x097104AEB1406F67ULL, "6085906_f.fxe"},
	{0x09716075D339AB9AULL, "6114088_m.fxe"},
	{0x0972B1B75C0B753FULL, "vxlt_gallowstemplar_f_43c_0_0.vlm"},
	{0x0972C25812F60511ULL, "dae000_cerimon6163678.cl"},
	{0x09737047B1BE7015ULL, "gen_im_arm_hlm_lgt_act3_01.uti"},
	{0x09737FEA1FAEFDDFULL, "6117682_m.fxe"},
	{0x097430E00702DD54ULL, "6129550_m.fxe"},
	{0x0974516B33D9A0E7ULL, "vxw_sundermount_2xl_0_0.vlm"},
	{0x09751C378B829985ULL, "6137035_m.fxe"},
	{0x09752F44A7B80D9AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_basic.msh"},
	{0x0975340C6AAB5881ULL, "test_2da_str.ncs"},
	{0x09753D44A7B825D3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_basic.mao"},
	{0x09754144A7B82CE0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_basic.mmh"},
	{0x0975BCDA3005C861ULL, "vxht_0_hightown_a2_be1_0_0.vlm"},
	{0x09762401239BAEA1ULL, "flare01.mao"},
	{0x097663C55242625BULL, "vxlt_gallowsprison_n_6pd_0_0.vlm"},
	{0x0976A00E01C4062AULL, "dae000_sabina6161456.cl"},
	{0x0976F0C2DB8815D4ULL, "309893194.ogg"},
	{0x0977A3624CC00CA8ULL, "art\\vfx\\combat\\vfx_w_wns_left_b_close1_c.mmh"},
	{0x097882B719070185ULL, "vxlt_0_lowtown_docks_a3_n_d05_0_0.vlm"},
	{0x0978E21496E01A3FULL, "vxht_keep_69w_0_0.vlm"},
	{0x097905B1E6219DE3ULL, "818266309.ogg"},
	{0x09795B45F8A4F78EULL, "vxlt_undercitydungeon_1yy_0_0.vlm"},
	{0x0979D59D7BA0FEFDULL, "vxht_bartrand_n_5af_0_0.vlm"},
	{0x097A4C7C8F915C97ULL, "6033598_f.fxe"},
	{0x097A52E4F777C4F9ULL, "303509584.ogg"},
	{0x097B4642751E64C7ULL, "377560973.ogg"},
	{0x097B4B0FE0630C99ULL, "vxdr_primevaltomb_6wn_0_0.vlm"},
	{0x097B60FE7BD37F36ULL, "6075697_m.fxe"},
	{0x097B710D345C1448ULL, "m_mcr371ar_galows_templr_nt.bnk"},
	{0x097B8CA8FE09A08DULL, "lt_0_lowtown_docks_a3_43_0.rml"},
	{0x097C75A947E299FAULL, "vxlt_0_lowtown_a2_hhg_0_0.vlm"},
	{0x097D28D26590CF11ULL, "738668704.ogg"},
	{0x097D435A49B3EBC5ULL, "722036753.ogg"},
	{0x097DBA2A1EC243AEULL, "lowtown_male_drunk_mob6151166.cl"},
	{0x097E0A6E1990784DULL, "911196998.ogg"},
	{0x097E183D084FF928ULL, "vxca_cave_2e4_0_0.vlm"},
	{0x097E449ED0B39FD0ULL, "vxlt_undercityencounter_1fg_0_0.vlm"},
	{0x097F1F1973DD6308ULL, "zz_ave_debug6064627.cl"},
	{0x097F458FF15A7448ULL, "vxht_0_hightown_a2_be4_0_0.vlm"},
	{0x097F58FDF6F0D731ULL, "hlac_dr_slavers_3207_3x.dds"},
	{0x097F67E055961657ULL, "vxht_brothel_n_3wd_0_0.vlm"},
	{0x097F9DEBDA5F9812ULL, "vxw_encounter2_camp_qp_0_1.vlm"},
	{0x098013BAA6FF0CDBULL, "6113985_m.fxe"},
	{0x0980C87B4DC20DA3ULL, "vxlt_warehouse_qw_0_0.vlm"},
	{0x09810B369733FA9FULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.mmh"},
	{0x09811ACCF4B6574EULL, "hlac_ht_mansion2b_818_49y.dds"},
	{0x09811BA32A3EB613ULL, "6156308_m.fxe"},
	{0x09812320E5FCA69FULL, "187465545.ogg"},
	{0x098138FFF47FD185ULL, "dre000_sibling_has_blight6036457.cl"},
	{0x098244ED834E35CBULL, "6149199_m.fxe"},
	{0x09825C0120929D0CULL, "framebuffereffects_radialblur_p0.psh.fxo.dx11"},
	{0x0982AD43BC45632FULL, "615507067.ogg"},
	{0x0982CCFD23CB6DA3ULL, "6072320_f.fxe"},
	{0x09834FCAB36F934CULL, "vxca_caverns2prc_aaq_0_0.vlm"},
	{0x09839C0A31DBB725ULL, "6013409_m.fxe"},
	{0x0983A3CC18EE0983ULL, "vxht_brothel_n_5l3_0_0.vlm"},
	{0x0984316C3D40C1CDULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_8.pwk"},
	{0x0984B79FADE47A7DULL, "6103560_m.fxe"},
	{0x0984D672AAA76294ULL, "gen00fl_dog6165737.cl"},
	{0x0984FED3399C68E5ULL, "hlac_lt_0_lowtown_a3_n_c_131070_0x.dds"},
	{0x09850B5EDB097B4EULL, "1024737114.ogg"},
	{0x09851A95507DF590ULL, "vxht_0_hightown_a2_av2_0_0.vlm"},
	{0x098520249A69980FULL, "hlac_lt_0_lowtown_a2_n_131072_6y.dds"},
	{0x09854361AB8D456BULL, "mcr310_sibling6076270.cl"},
	{0x0985F1C8E6EAB906ULL, "dae011_drunk2_sta.cnv"},
	{0x09860D568F4B670BULL, "666388170.ogg"},
	{0x09860F75E1B271BBULL, "251747403.ogg"},
	{0x098653324AFC8C6AULL, "w_bonepitdestroyed_a3_243_0.mao"},
	{0x098657324AFC92B9ULL, "w_bonepitdestroyed_a3_243_0.mmh"},
	{0x098661324AFCA3A7ULL, "w_bonepitdestroyed_a3_243_0.msh"},
	{0x0986B50B423A577BULL, "narrative_varric_demo6170807.cl"},
	{0x098816301DACB64DULL, "vxlt_gallowstemplar_f_660_0_0.vlm"},
	{0x0988D9A5B7F1101DULL, "vxlt_gallowstemplar_n_5c8_0_0.vlm"},
	{0x0988F1F73F85772CULL, "blood_conj_08.mao"},
	{0x0988F3C8E6ECF2C6ULL, "dae011_drunk2_sta.bnk"},
	{0x098947F59BFAFDF5ULL, "zz_fol_debug6056998.cl"},
	{0x09895E35E7678733ULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunaricontroller_sa.dds"},
	{0x09896F2DE0BFCE0CULL, "mrl000_marethari6069864.cl"},
	{0x09899907E1C724E3ULL, "118467572.ogg"},
	{0x0989D6EF451F5C2EULL, "65524077.ogg"},
	{0x098A380C078B178DULL, "175363961.ogg"},
	{0x098A4D23DF6023B8ULL, "295840479.ogg"},
	{0x098A8993686AF845ULL, "vxlt_0_lowtown_a2_hfn_0_0.vlm"},
	{0x098AADD9C5EEA4CBULL, "l_ser111ar_foundry.wbd"},
	{0x098ACCA80610BECDULL, "vxlt_0_lowtown_a2_gx3_0_0.vlm"},
	{0x098B3B33BD711376ULL, "6125435_m.fxe"},
	{0x098B42E37D8FCA73ULL, "696277062.ogg"},
	{0x098BD2749E1B4080ULL, "zz_approval_debug6003702.cl"},
	{0x098C3C250706FA4EULL, "541878720.ogg"},
	{0x098E2439CE6945E8ULL, "6167000_m.fxe"},
	{0x098E3F4335CDEC6DULL, "vxlt_0_lowtown_n_g93_0_0.vlm"},
	{0x098E625343D11803ULL, "art\\vfx\\combat\\vfx_glb_stun_c.dura.ani"},
	{0x098E67988DDBA3FDULL, "6125106_m.fxe"},
	{0x098E79109404BAB4ULL, "zz_approval_debug6003698.cl"},
	{0x098E8D1173B9FFE7ULL, "lowtown_fem_alienage1_sta6162565.cl"},
	{0x098EA26BC1D0391AULL, "6092085_m.fxe"},
	{0x098F27E08D129236ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpouta.mao"},
	{0x098F33E08D12A6D5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpouta.mmh"},
	{0x098F35E08D12AA4BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpouta.msh"},
	{0x098F4DB66DE62E1EULL, "6118440_m.fxe"},
	{0x098F4F961D9F9812ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cookerpots.phy"},
	{0x098FB7F45F4497D3ULL, "6084245_m.fxe"},
	{0x098FE060648F8D85ULL, "art\\vfx\\combat\\vfx_gib_rob_appc.mao"},
	{0x098FEEA50526B50EULL, "6152778_m.fxe"},
	{0x098FF3AE821B6BC1ULL, "6115960_f.fxe"},
	{0x0990029B84665DC8ULL, "mer240_jansen6050328.cl"},
	{0x09900AF74750609DULL, "102546494.ogg"},
	{0x0990377EDB9B288BULL, "766986214.ogg"},
	{0x0990FFBA26E780EBULL, "69222022.ogg"},
	{0x09914689750CCE0AULL, "6104188_m.fxe"},
	{0x09916045D052A9ABULL, "6076868_m.fxe"},
	{0x099171D65F69ECF4ULL, "vxlt_0_lowtown_docks_a3_n_a0m_0_0.vlm"},
	{0x0991F35015B730C7ULL, "hlac_ca_caverns1_mines_24423_10y.dds"},
	{0x09920DE723D1CFA5ULL, "6018509_m.fxe"},
	{0x099267833A5BC509ULL, "vxca_cavern1_haunted_i1k_0_0.vlm"},
	{0x09927DC93B527D38ULL, "vxlt_gallowstemplar_f_6qd_0_0.vlm"},
	{0x0992E4C00D72F67AULL, "thefarcliffskrkwl.dds"},
	{0x099314B561413138ULL, "132738383.ogg"},
	{0x0993297C9FDA92FDULL, "6104460_m.fxe"},
	{0x09933751F7222B77ULL, "186363587.ogg"},
	{0x09938C7242740E10ULL, "vxht_0_hightown_a2_aq3_0_0.vlm"},
	{0x09939B711F452F60ULL, "gen00fl_isabela6079053.cl"},
	{0x099409F67F5B8943ULL, "14733315.ogg"},
	{0x099438C13D6E9E9DULL, "6094512_m.fxe"},
	{0x0994684C8FFE3B61ULL, "737596721.ogg"},
	{0x0994B4A5BC1682F8ULL, "6104087_m.fxe"},
	{0x0995766A893485B5ULL, "509716245.ogg"},
	{0x09957BAF3D47AD9EULL, "6149288_m.fxe"},
	{0x099610E4747EA66AULL, "hlac_w_encounter3_-196608_0z.dds"},
	{0x09961B3A438C03CCULL, "lgt160_ntb_elf_conflict6163374.cl"},
	{0x09962E3DCF1FB65BULL, "396785630.ogg"},
	{0x099687C4B9581CC7ULL, "6075612_m.fxe"},
	{0x0996B9575F8D2FFDULL, "vxht_bartrand_n_5od_0_0.vlm"},
	{0x0997118B9B9C49ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_crack_knuckles.ani"},
	{0x099718BEC0B9EF0EULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.mmh"},
	{0x099767DDAEC20D26ULL, "vxlt_hangedman_1vf_0_0.vlm"},
	{0x0998AA7A0AD4BDDFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicechair_0.pwk"},
	{0x0998B9D735ACA239ULL, "33386345.ogg"},
	{0x0998D7EEF3505640ULL, "hlac_ht_0_hightown_0_3x.dds"},
	{0x0998FB705A6110C5ULL, "vxlt_0_lowtown_a2_hcm_0_0.vlm"},
	{0x0999781FF93CC49AULL, "vxlt_warehouse_n_12e_0_0.vlm"},
	{0x0999CC8D37EEB813ULL, "6087378_m.fxe"},
	{0x0999D46CD53370F6ULL, "vxw_encounter2_camp_nl_0_0.vlm"},
	{0x0999F987C6CD8F63ULL, "gen_im_acc_blt_act3_05.uti"},
	{0x099A263920C1BDA1ULL, "one300ip_worldmap.utp"},
	{0x099A78209586F6E7ULL, "6042049_m.fxe"},
	{0x099A946381F14806ULL, "410388215.ogg"},
	{0x099AB035CB41C8B4ULL, "393120644.ogg"},
	{0x099B7A5343DBCBFAULL, "art\\vfx\\combat\\vfx_glb_stun_c.dura.evt"},
	{0x099B9165B419413AULL, "zz_mrl_debug6109314.cl"},
	{0x099C171AD1B7E629ULL, "hlac_w_encounter3_-131070_0y.dds"},
	{0x099C4DBEA8066EA8ULL, "gen00fl_dog.utc"},
	{0x099C53F8381CAE28ULL, "mcr370_meredith6083042.cl"},
	{0x099CC0866570C19BULL, "6156379_m.fxe"},
	{0x099CC2161A68FAFAULL, "909532822.ogg"},
	{0x099CCB39CFA799CBULL, "vxlt_0_lowtown_a2_cmn_0_0.vlm"},
	{0x099CF9FF2D96D99FULL, "vxht_mansion_n_5i7_0_0.vlm"},
	{0x099D1E8B9BA0CB2EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_crack_knuckles.gad"},
	{0x099E191EB13C49DBULL, "vxht_mansion_n_4bg_0_0.vlm"},
	{0x099F1B757EC6BC9BULL, "dae000_gen_turn_in_16178392.cl"},
	{0x099F3A011B2C2E7CULL, "lgt120_orz_bln_official6140172.cl"},
	{0x099F5C97D772A5BDULL, "mer140_bonepit_comments6062991.cl"},
	{0x099F86888C8BD623ULL, "hlac_lt_0_lowtown_n_-262144_1x.dds"},
	{0x099F96F036279303ULL, "962727408.ogg"},
	{0x099FC64473207999ULL, "dae000_merchant_armor_hm6164276.cl"},
	{0x099FCB29EFF3C81BULL, "962646565.ogg"},
	{0x099FEBA96C213A6EULL, "835922782.ogg"},
	{0x09A0630D347B18F2ULL, "m_mcr371ar_galows_templr_nt.wbd"},
	{0x09A095777060FBC3ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga101a_0t.dds"},
	{0x09A0AD6681F1FC33ULL, "qcr220_meredith6185152.cl"},
	{0x09A0AFAD49E05B14ULL, "ceil_clo.mao"},
	{0x09A0C4B3451934BCULL, "vxdr_primevaltomb_6m1_0_0.vlm"},
	{0x09A12120F136231FULL, "6125593_m.fxe"},
	{0x09A128E7F362F690ULL, "llac_lt_0_lowtown_a2_n_-65535_1.dds"},
	{0x09A19BF35C79EB72ULL, "con_ico_romance.dds"},
	{0x09A1B14EFFE1D8A4ULL, "llac_lt_0_lowtown_a2_n_c_131072_4.dds"},
	{0x09A20C90B8352606ULL, "361764089.ogg"},
	{0x09A25133C41F7FC9ULL, "441012331.ogg"},
	{0x09A2795343E1E8EDULL, "art\\vfx\\combat\\vfx_glb_stun_c.dura.gad"},
	{0x09A339A018465CF3ULL, "6109252_m.fxe"},
	{0x09A3456BCD9CB64AULL, "vxht_keep_n_3hk_0_0.vlm"},
	{0x09A36A99A88DF504ULL, "6182009_m.fxe"},
	{0x09A396145CC7E025ULL, "w_woundedcoast_85_0.rml"},
	{0x09A3EA85085E5EBDULL, "vxlt_backalley_n_1bo_0_0.vlm"},
	{0x09A415700AD9A7E5ULL, "follower_banter6163446.cl"},
	{0x09A4198B9BA6E10DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_crack_knuckles.evt"},
	{0x09A52DDDE00E670CULL, "vxca_caverns2prc_9pz_0_0.vlm"},
	{0x09A537BEC0C6610DULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.anb"},
	{0x09A55F5E2CC2E865ULL, "llac_lt_0_lowtown_a2_n_c_65536_2.dds"},
	{0x09A5739B91D7686EULL, "one280_iwan6096894.cl"},
	{0x09A6472E3881EC39ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_standb_1.mao"},
	{0x09A666FB80B50A53ULL, "1014972876.ogg"},
	{0x09A70A5728F871B2ULL, "346591727.ogg"},
	{0x09A70CA854ED76A0ULL, "6034375_m.fxe"},
	{0x09A75EF4A8F82072ULL, "6185385_m.fxe"},
	{0x09A792E3027EAFDDULL, "ser131ip_to_darktown.utp"},
	{0x09A7D8DD8EE324FDULL, "hlac_ca_cavern1_haunted_24423_10y.dds"},
	{0x09A84649C7298976ULL, "vxw_encounter2_camp_im_0_0.vlm"},
	{0x09A84E0B11E6E179ULL, "257292565.ogg"},
	{0x09A8CF80159F850CULL, "500226627.ogg"},
	{0x09A8E4B5416D886DULL, "6087857_m.fxe"},
	{0x09A9C1AF27AAA975ULL, "6112889_m.fxe"},
	{0x09AA0C36975701DCULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.anb"},
	{0x09AAA092E66CF9ACULL, "hlac_w_bonepitdestroyed_a3_262147_0x.dds"},
	{0x09AB1E94F4C5BF27ULL, "893028342.ogg"},
	{0x09AB40EB28D30D75ULL, "vrc221_varric6048575.cl"},
	{0x09AB456EDB505F48ULL, "6167476_f.fxe"},
	{0x09ABF8E8633846ABULL, "hlac_lt_0_lowtown_n_1_0z.dds"},
	{0x09AC49A03FBF4BFDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutcn.mao"},
	{0x09AC4DA03FBF5236ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutcn.mmh"},
	{0x09AC57A03FBF6334ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutcn.msh"},
	{0x09AC869175FD92E1ULL, "vxlt_0_lowtown_a2_he7_0_0.vlm"},
	{0x09AC8C2F748785D5ULL, "6171882_m.fxe"},
	{0x09ACFE3088B4E826ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.anb"},
	{0x09AD68451007699DULL, "hlac_lt_0_lowtown_-1_2x.dds"},
	{0x09AE4C44EA40519AULL, "187843324.ogg"},
	{0x09AE606EAB2A4BF1ULL, "vxht_chantry_n_2zk_0_0.vlm"},
	{0x09AE955442E848D4ULL, "6034076_m.fxe"},
	{0x09AEED2E3BCC2701ULL, "6148851_m.fxe"},
	{0x09AF3690370F4D3CULL, "vxdr_primevaltomb_6p6_0_0.vlm"},
	{0x09AFB6542F58D70EULL, "vxht_0_hightown_a2_c4e_0_1.vlm"},
	{0x09B05CC7A410BC2CULL, "mer240_jansen6050412.cl"},
	{0x09B08720F39EA5F5ULL, "6146790_m.fxe"},
	{0x09B08FD9C60F4099ULL, "l_ser111ar_foundry.bnk"},
	{0x09B0C7737FF801ADULL, "art\\levels\\ztools\\default\\generic\\0di_angle.mmh"},
	{0x09B0CB737FF808FEULL, "art\\levels\\ztools\\default\\generic\\0di_angle.mao"},
	{0x09B0D9737FF82033ULL, "art\\levels\\ztools\\default\\generic\\0di_angle.msh"},
	{0x09B11CE17B586470ULL, "w_weapon.wbd"},
	{0x09B1ED0E1D472733ULL, "hlac_dr_primevalentrance_1726_8y.dds"},
	{0x09B1FAC92880E7DEULL, "6164775_m.fxe"},
	{0x09B2343DB6DFFF72ULL, "one110_pryce6088868.cl"},
	{0x09B23B96BF069BDFULL, "hlac_lt_0_lowtown_a2_65538_0y.dds"},
	{0x09B2592D3E5087A0ULL, "6118811_m.fxe"},
	{0x09B2DE60AECADB0FULL, "vxlt_0_lowtown_a2_n_c_ghh_0_0.vlm"},
	{0x09B3034014B9E70CULL, "follower_banter6141931.cl"},
	{0x09B3252A9F1B0183ULL, "436820091.ogg"},
	{0x09B3569C87BFAA07ULL, "ts_kirk_guards_03.gda"},
	{0x09B439B0D201B438ULL, "6040163_m.fxe"},
	{0x09B44B9792AE3E26ULL, "vxlt_hangedman_1xh_0_0.vlm"},
	{0x09B4AD8C823337EFULL, "fen110_anso6094957.cl"},
	{0x09B5251303CFB8FFULL, "hlac_lt_hangedman_n_-1_29z.dds"},
	{0x09B575C16C9FEABEULL, "354328553.ogg"},
	{0x09B5B35B798159B7ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_1y.dds"},
	{0x09B5BD0B0CC1308DULL, "lav_lakemidtone.mao"},
	{0x09B60F4737523EC8ULL, "vxlt_0_lowtown_a2_he2_0_0.vlm"},
	{0x09B6672C844A4570ULL, "dae000_flemeth6000680.cl"},
	{0x09B6BA13D17B6A51ULL, "dae000_bartrand6064088.cl"},
	{0x09B73F289F9C4546ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravea_d.dds"},
	{0x09B794238C32865CULL, "ht_0_hightownexplosion_n_20_0_ll.mmh"},
	{0x09B7BE419980AEEEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masf_0s.dds"},
	{0x09B7E7A9A07DB273ULL, "dae000_gamlen6121545.cl"},
	{0x09B83FA2190B3E24ULL, "850986701.ogg"},
	{0x09B9924ED8F3EC34ULL, "vxdr_0_deeproads_eud_0_0.vlm"},
	{0x09B9D44050E8460EULL, "6083549_f.fxe"},
	{0x09B9D909E4697D1FULL, "ser221cr_gascard.ncs"},
	{0x09BA1202ABA6AD14ULL, "680059501.ogg"},
	{0x09BA31BE2298FA09ULL, "vxdr_primevaltomb_8no_0_0.vlm"},
	{0x09BA5A23686ABCFEULL, "ave310_brennan6136885.cl"},
	{0x09BA95E9B1F44CD1ULL, "544583811.ogg"},
	{0x09BAAA9A81E2AA26ULL, "vxlt_gallowsprison_n_5rl_0_0.vlm"},
	{0x09BAF86E67F3AB61ULL, "vxlt_0_lowtown_a2_hb6_0_0.vlm"},
	{0x09BB013088C12AD1ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.mmh"},
	{0x09BB1AD289AC1BB9ULL, "lt_0_lowtown_docks_a3_n_117_0.msh"},
	{0x09BB24D289AC2CA7ULL, "lt_0_lowtown_docks_a3_n_117_0.mmh"},
	{0x09BB28D289AC33F4ULL, "lt_0_lowtown_docks_a3_n_117_0.mao"},
	{0x09BB315E1597392EULL, "vxca_caverns1_mines_jg1_0_0.vlm"},
	{0x09BB35BEA820F6F9ULL, "gen00fl_dog.bnk"},
	{0x09BBDC9084642B1AULL, "dae300_cullen6140611.cl"},
	{0x09BC2E6863B56FBBULL, "fen110_anso6095015.cl"},
	{0x09BC703994375347ULL, "327751947.ogg"},
	{0x09BD67168A8B1E1AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_toweratall.msh"},
	{0x09BD75168A8B3653ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_toweratall.mao"},
	{0x09BD79168A8B3D60ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_toweratall.mmh"},
	{0x09BDA86D290565BCULL, "vxdr_primevaltomb_6w7_0_0.vlm"},
	{0x09BE30FF482FEBB5ULL, "6063842_m.fxe"},
	{0x09BE57BEA8236659ULL, "gen00fl_dog.cnv"},
	{0x09BE6BD053032811ULL, "dae000_madam_lusine6151215.cl"},
	{0x09BEF14D6054561EULL, "6078976_f.fxe"},
	{0x09C00100ECB43D0BULL, "295707740.ogg"},
	{0x09C042A962F6E9EAULL, "sound_settings.gda"},
	{0x09C05D397096E636ULL, "w_encounter3_85_0.msh"},
	{0x09C067397096F734ULL, "w_encounter3_85_0.mmh"},
	{0x09C06B397096FE47ULL, "w_encounter3_85_0.mao"},
	{0x09C0CB0A3046A020ULL, "6117911_m.fxe"},
	{0x09C17BD8A430FBE8ULL, "vxht_0_hightown_a2_c6y_0_1.vlm"},
	{0x09C1C0738006696AULL, "art\\levels\\ztools\\default\\generic\\0di_angle.phy"},
	{0x09C1D16F97A3584BULL, "6168136_m.fxe"},
	{0x09C2BD7484A456A6ULL, "vxlt_hangedman_1gi_0_0.vlm"},
	{0x09C3C6749B226C94ULL, "vxht_chantry_2sz_0_0.vlm"},
	{0x09C3CEB604E8E362ULL, "6070997_m.fxe"},
	{0x09C4ADCFC6C084B4ULL, "178286303.ogg"},
	{0x09C534488F394888ULL, "follower_banter6163229.cl"},
	{0x09C585381CDDC5DAULL, "121412595.ogg"},
	{0x09C59B4BA210D0E5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzarda_sa.dds"},
	{0x09C6FA1A094B4644ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_nud_0t.dds"},
	{0x09C71C016584F432ULL, "6068466_m.fxe"},
	{0x09C7871E1F0C1FF7ULL, "vxht_0_hightown_avm_0_0.vlm"},
	{0x09C7A131B5F3B425ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_6x.dds"},
	{0x09C7EE20FB0E0580ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\large_d.dds"},
	{0x09C83BACE108CFD2ULL, "vxlt_hangedman_n_24x_0_0.vlm"},
	{0x09C8A39B148F1289ULL, "vxdr_primevaltomb_8in_0_0.vlm"},
	{0x09C8C5C40101665EULL, "vxlt_0_lowtown_a2_fab_0_0.vlm"},
	{0x09C8C64C4E28E23CULL, "mag110_samson6018709.cl"},
	{0x09C976145CE8794EULL, "w_woundedcoast_85_0.mmh"},
	{0x09C982145CE88DE5ULL, "w_woundedcoast_85_0.mao"},
	{0x09C990145CE8A53CULL, "w_woundedcoast_85_0.msh"},
	{0x09C9D80C72902C47ULL, "vis121_saemus6118459.cl"},
	{0x09CB44288F167CF1ULL, "vxht_chantry_n_2pi_0_0.vlm"},
	{0x09CB5D25156CE471ULL, "6070113_f.fxe"},
	{0x09CB935DB779A0CAULL, "vxlt_0_lowtown_gbv_0_0.vlm"},
	{0x09CBA868B290B009ULL, "6096319_f.fxe"},
	{0x09CBF9B3FB053DB8ULL, "6089902_m.fxe"},
	{0x09CC83F0A91FAFAFULL, "hlac_ht_keep_4519_14z.dds"},
	{0x09CC8B6338BBC9A8ULL, "vxw_bonepitdestroyed_a3_7a_0_0.vlm"},
	{0x09CCC89D709AC0F5ULL, "386044264.ogg"},
	{0x09CD071E752178ADULL, "gen00fl_isabela_init6077303.cl"},
	{0x09CD9AC03E3A75D9ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.dura.ani"},
	{0x09CDCC80D5180997ULL, "6204966_m.fxe"},
	{0x09CDD6EFA79A57B1ULL, "mcr370_meredith6085928.cl"},
	{0x09CDDB266AA20260ULL, "hlac_ht_playershouse_n_855_31y.dds"},
	{0x09CE783E43842CCFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerc_d.dds"},
	{0x09CE95870A88EFE1ULL, "639377611.ogg"},
	{0x09CE95A2A762D8AFULL, "hlac_lt_0_lowtown_a2_n_65534_4x.dds"},
	{0x09CEA0D57931E8DBULL, "6064597_m.fxe"},
	{0x09CEA7C23DA37DC4ULL, "vxdr_primevaltomb_6rd_0_0.vlm"},
	{0x09CF42A3BDE62F60ULL, "gen00fl_isabela6097013.cl"},
	{0x09CF87D38BFFDA25ULL, "vxw_encounter2_camp_p5_0_0.vlm"},
	{0x09CFA6788448300AULL, "vxca_caverns1_coi_0_0.vlm"},
	{0x09CFD598867BF9CAULL, "vxlt_0_lowtown_a2_n_c_i1d_0_0.vlm"},
	{0x09CFEF10FD3DDC39ULL, "6070424_m.fxe"},
	{0x09D02E4278C4E181ULL, "lt_0_lowtown_docks_a3_n_37_0.rml"},
	{0x09D052550DE4566CULL, "6177913_m.fxe"},
	{0x09D0695CEA7E2649ULL, "vxw_woundedcoast_1de_0_0.vlm"},
	{0x09D0DA46564477BAULL, "vrc000ev_item_acquired.ncs"},
	{0x09D10F4C3C3F5ABAULL, "20615965.ogg"},
	{0x09D11E47C4185844ULL, "dae000_arishok6068501.cl"},
	{0x09D13543AC2CE9E5ULL, "6165182_m.fxe"},
	{0x09D1B0BCD27F7FA3ULL, "693807084.ogg"},
	{0x09D244DF8B7AFADEULL, "687858095.ogg"},
	{0x09D2552CEBB40BFEULL, "494538517.ogg"},
	{0x09D2797EB32570C8ULL, "dae100_bran6118298.cl"},
	{0x09D2A1776D67B06EULL, "gen_im_npc_wep_sta_gena.uti"},
	{0x09D312A3EBE30EE7ULL, "162537775.ogg"},
	{0x09D39A37912D5A3DULL, "hlac_lt_0_lowtown_a2_131070_6z.dds"},
	{0x09D3A560A294C69BULL, "art\\characters\\creatures\\ragedemon\\c_ragedemon.mao"},
	{0x09D3BB5F0F7F4AE5ULL, "llac_w_woundedcoast_0_11.dds"},
	{0x09D3EDA32A9B42D6ULL, "vxlt_0_lowtown_e28_0_0.vlm"},
	{0x09D447D22B22CD5DULL, "ss_lothering.dds"},
	{0x09D475C03E40551BULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.dura.gad"},
	{0x09D4AF09DC923D90ULL, "hlac_ht_keep_3739_1z.dds"},
	{0x09D4C04C5E37B20EULL, "vxlt_0_lowtown_a3_n_c_h2e_0_0.vlm"},
	{0x09D5F196847684B8ULL, "6110442_m.fxe"},
	{0x09D60DED51763EC8ULL, "299841296.ogg"},
	{0x09D60E42CFB49A48ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grill_n.dds"},
	{0x09D63F7F95D87C6FULL, "hlac_w_sundermount_262148_0y.dds"},
	{0x09D6819E0B6DF3C3ULL, "zz_vault_debug6203265.cl"},
	{0x09D7157806852B09ULL, "vxdr_primevaltomb_8dm_0_0.vlm"},
	{0x09D72454235ECBB2ULL, "vxca_caverns2_8nz_0_0.vlm"},
	{0x09D738A420313B0EULL, "203133136.ogg"},
	{0x09D746784D0A2750ULL, "6137481_m.fxe"},
	{0x09D84A7F54CEEBF5ULL, "ch2notnt_s2.psh.fxo"},
	{0x09D8F89ABB4C1E3DULL, "montage_subs_demohack6217120.cl"},
	{0x09D936D8A6387E8BULL, "vxlt_0_lowtown_a3_n_c_byw_0_0.vlm"},
	{0x09D9372A27FC6741ULL, "6015955_m.fxe"},
	{0x09D93910A5359722ULL, "6027540_m.fxe"},
	{0x09DA23B1A5463727ULL, "qun221_viscount6084697.cl"},
	{0x09DA2B3E9B175120ULL, "6119460_f.fxe"},
	{0x09DA43AB3951B71BULL, "vxw_sundermount_35f_0_0.vlm"},
	{0x09DA816AA46C84C1ULL, "vxca_caverns1_igx_0_0.vlm"},
	{0x09DA9500FB236456ULL, "pro000_wesley6007723.cl"},
	{0x09DB43B471DF85C3ULL, "695589296.ogg"},
	{0x09DB4BC68518B1F1ULL, "6070378_m.fxe"},
	{0x09DB92C03E46A500ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.dura.evt"},
	{0x09DBB7DE2ACCF123ULL, "vxw_encounter3_6f_0_0.vlm"},
	{0x09DBD0D30FE9B860ULL, "lt_0_lowtown_a2_n_38_0.anb"},
	{0x09DBD16815DB1F7AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_cara_s.dds"},
	{0x09DC2AAE7CC7EAB6ULL, "vxlt_gamlenshouse_n_5e_0_0.vlm"},
	{0x09DC38B696205720ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scroll_vt.mao"},
	{0x09DCACD85AE6C44EULL, "744812639.ogg"},
	{0x09DD21F5D895ADD0ULL, "119167267.ogg"},
	{0x09DD49FA9E0E92D1ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_sepulcher_sa.dds"},
	{0x09DD95AA57289C72ULL, "vxht_brothel_n_53z_0_1.vlm"},
	{0x09DDF9B07DF5F2A5ULL, "vxw_encounter2_camp_k4_0_0.vlm"},
	{0x09DDFFF9DFF6F068ULL, "730295781.ogg"},
	{0x09DE6FD28A639F49ULL, "dae000ip_crowd_mf1_talk13.utp"},
	{0x09DE97A16915FEADULL, "vxht_0_hightown_a3_bcz_0_0.vlm"},
	{0x09DEAF30929EF0B8ULL, "dae021_sister_samea_sta6142178.cl"},
	{0x09DEB6B149525DFDULL, "273715226.ogg"},
	{0x09DED68082FFB3C4ULL, "mag291_demon_marethari6166176.cl"},
	{0x09DEDB39DC743EC9ULL, "vxw_woundedcoast_1od_0_0.vlm"},
	{0x09DEEEA113D02619ULL, "dae000_gamlen6121558.cl"},
	{0x09DF020BB02022A9ULL, "vxht_keep_n_70h_0_0.vlm"},
	{0x09DF696534360763ULL, "gen00fl_dog6117791.cl"},
	{0x09DFBE7D1250BB12ULL, "vxca_caverns1_mines_hw3_0_0.vlm"},
	{0x09DFBF44EB120E93ULL, "vxlt_warehouse_n_eb_0_0.vlm"},
	{0x09DFE5447EC17464ULL, "vxdr_primevaltomb_7zs_0_2.vlm"},
	{0x09E06400B78D190CULL, "6096831_f.fxe"},
	{0x09E07D6516083BB2ULL, "6065933_m.fxe"},
	{0x09E0E5CEB2C5E7ADULL, "zz_qcr_debug6215491.cl"},
	{0x09E14641A00F7F3AULL, "6103787_m.fxe"},
	{0x09E19E80B37A982EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleon.msh"},
	{0x09E1AC80B37AB07FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleon.mao"},
	{0x09E1ACF5F24B0BF0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_post.mao"},
	{0x09E1B0F5F24B1203ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_post.mmh"},
	{0x09E1B880B37AC41CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleon.mmh"},
	{0x09E1BEF5F24B2A5DULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_post.msh"},
	{0x09E209BB8B37142AULL, "gen00fl_dog6115080.cl"},
	{0x09E2C2FB81C96794ULL, "zz_one_debug6074013.cl"},
	{0x09E2DB5A054FE59FULL, "character_character_p138.psh.fxo.dx11.assembly"},
	{0x09E35FAFA13F5006ULL, "lgt311_shw_mer_mettin6156550.cl"},
	{0x09E3A2B062CE813AULL, "484726519.ogg"},
	{0x09E3BA40B00FA747ULL, "qun110_petrice6042305.cl"},
	{0x09E3C17FD7AD2006ULL, "6166294_m.fxe"},
	{0x09E4421BD13735C8ULL, "530771290.ogg"},
	{0x09E45D2BDB7FFEFDULL, "gen_im_trs_chn_light_doc_04.uti"},
	{0x09E47E68EED0FBA4ULL, "hlac_ca_caverns2_5802_58z.dds"},
	{0x09E55CC81D538F7EULL, "hlac_w_encounter2_131072_0x.dds"},
	{0x09E5862C114C8F66ULL, "756587550.ogg"},
	{0x09E5F2C51D9B2E97ULL, "dae000_flemeth6000726.cl"},
	{0x09E5F34042329C49ULL, "6118490_m.fxe"},
	{0x09E5FC57D86F90D3ULL, "ht_0_hightown_a2_36_0_ll.mmh"},
	{0x09E6C26B4DFE2B32ULL, "zz_one_debug6099942.cl"},
	{0x09E6EDCB859B235CULL, "hlac_ht_0_hightown_a3_-196607_0x.dds"},
	{0x09E70DD289D140DCULL, "lt_0_lowtown_docks_a3_n_117_0.phy"},
	{0x09E759AE7BC0BEF6ULL, "hlac_ht_mansion_n_3374_0y.dds"},
	{0x09E7A3BC8F5870DAULL, "6070442_f.fxe"},
	{0x09E7E2D6E5F16B28ULL, "hlac_lt_0_lowtown_docks_a3_-4_3y.dds"},
	{0x09E809F7E79A74EFULL, "prcscr_one.gda"},
	{0x09E82A8D990E501AULL, "vxht_0_hightown_a2_9zd_0_0.vlm"},
	{0x09E858E603A28BB7ULL, "676716628.ogg"},
	{0x09E8733563458572ULL, "671268109.ogg"},
	{0x09E8980278EBC4BCULL, "61233724.ogg"},
	{0x09E8A314749BC253ULL, "vxht_bartrand_n_4hv_0_0.vlm"},
	{0x09E8B00EFC06343DULL, "1054512199.ogg"},
	{0x09E8C4D1C4C0FE6EULL, "6033699_m.fxe"},
	{0x09E8D2999BD92A32ULL, "vxht_0_hightown_a2_bx5_0_0.vlm"},
	{0x09E9057A9581E6B0ULL, "930170233.ogg"},
	{0x09E918DFD062AD69ULL, "vxw_sundermount_36t_0_1.vlm"},
	{0x09E93AB757BFB8D8ULL, "vxdr_primevalentrance_1k8_0_0.vlm"},
	{0x09E9427AD04592B5ULL, "559240996.ogg"},
	{0x09E97834D4A796FDULL, "partypicker.gda"},
	{0x09E97A34D4A79A7AULL, "partypicker.gfx"},
	{0x09E9BDA3770ECA71ULL, "6075078_m.fxe"},
	{0x09E9CC59A74CA043ULL, "vxlt_gallowstemplar_n_3ir_0_0.vlm"},
	{0x09E9E167607879D8ULL, "blurring_additive.mao"},
	{0x09E9E24AFE08324DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3.mmh"},
	{0x09EA29BB1CC309A3ULL, "vxw_encounter3_3i_0_0.vlm"},
	{0x09EA727B08079DDAULL, "llac_lt_0_lowtown_docks_n_65534_6.dds"},
	{0x09EB520D2A31FD4BULL, "vxdr_primevaltomb_4xo_0_0.vlm"},
	{0x09EB82DC43D02AD8ULL, "and111_lirene6084547.cl"},
	{0x09EB98AC1593B672ULL, "173277269.ogg"},
	{0x09EBD43A1503C896ULL, "vxlt_0_lowtown_a2_eny_0_1.vlm"},
	{0x09EC23D5C22CD93DULL, "vxlt_backalley_n_1yj_0_0.vlm"},
	{0x09EC545DC67DC4D1ULL, "hlac_lt_0_lowtown_-65535_0y.dds"},
	{0x09EC6B8D6FEC0B25ULL, "vxw_encounter2_camp_f3_0_0.vlm"},
	{0x09ECE770A823F532ULL, "841637227.ogg"},
	{0x09ED98B27FC178C2ULL, "6101411_m.fxe"},
	{0x09EDFD294D0AB8DCULL, "gen00fl_varric6099451.cl"},
	{0x09EEB7A04A6C8BCEULL, "gen00fl_sebastian6195857.cl"},
	{0x09EF0D12FF383389ULL, "498190422.ogg"},
	{0x09EF2F39210A0D68ULL, "one300ip_worldmap.ncs"},
	{0x09F0080002060598ULL, "pro000_refugee_angry6052170.cl"},
	{0x09F032EE8A447E5EULL, "hlac_lt_0_lowtown_n_-65536_3y.dds"},
	{0x09F04AFBAD871B2FULL, "pro000cr_guard.utc"},
	{0x09F0D534D898C247ULL, "vxlt_gallowstemplar_n_31v_0_0.vlm"},
	{0x09F0ECCC4942CC91ULL, "851042184.ogg"},
	{0x09F1091D310C3B8CULL, "vxdr_primevaltomb_84d_0_0.vlm"},
	{0x09F10BC4DAD32BA9ULL, "follower_banter6169831.cl"},
	{0x09F1734F36BA3580ULL, "llac_lt_0_lowtown_65536_0.dds"},
	{0x09F176B79C312FE4ULL, "333140006.ogg"},
	{0x09F1D6AA1C3A257BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabsnb.mmh"},
	{0x09F1E2AA1C3A3918ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabsnb.mao"},
	{0x09F1F4AA1C3A5865ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabsnb.msh"},
	{0x09F20CFCB5F17D80ULL, "6065957_m.fxe"},
	{0x09F28B1BC8B5EF08ULL, "llac_lt_0_lowtown_a2_n_-131073_2.dds"},
	{0x09F2B1F5F2598728ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp_post.phy"},
	{0x09F2CDC515DDEFFDULL, "star_yellow.dds"},
	{0x09F2D18E28F1D96EULL, "gen00fl_merrill6121651.cl"},
	{0x09F33E4624E1E1D1ULL, "263945091.ogg"},
	{0x09F34146815DB737ULL, "6088703_f.fxe"},
	{0x09F363B24D956B0BULL, "vxw_blightlands_w9_0_0.vlm"},
	{0x09F37D7450FFD9AAULL, "441521427.ogg"},
	{0x09F38A49102CD854ULL, "vxlt_hangedman_n_2em_0_0.vlm"},
	{0x09F3B6D841D55B0CULL, "6087661_m.fxe"},
	{0x09F3C1A3DA1157BAULL, "439536242.ogg"},
	{0x09F3C5384C16EF97ULL, "zz_ser_debug6078560.cl"},
	{0x09F3C6FC33E1554FULL, "6148350_m.fxe"},
	{0x09F4080E074AFCB2ULL, "vxca_caverns2_8pd_0_0.vlm"},
	{0x09F4A41C36E33D5AULL, "635240222.ogg"},
	{0x09F579C91D3649DCULL, "vxca_caverns1_j28_0_1.vlm"},
	{0x09F5947D6AD99DDDULL, "hlac_lt_darktown_-1_74z.dds"},
	{0x09F5CAF7479EC2AAULL, "ser221_party_comment_found.stg"},
	{0x09F5EB733237B148ULL, "6171847_m.fxe"},
	{0x09F639EE780E51A9ULL, "dae000ip_crowd_mf2_talk3.utp"},
	{0x09F6541A7DF0A96CULL, "122265499.ogg"},
	{0x09F6AAFD9E210AA4ULL, "aip_raiderarcher_base.gda"},
	{0x09F6C58E7C646AD2ULL, "6034367_f.fxe"},
	{0x09F6E96CB587054CULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_rtl.anb"},
	{0x09F7306990AF2673ULL, "6066036_m.fxe"},
	{0x09F7B1A6DED5DF26ULL, "786306303.ogg"},
	{0x09F7BACE9E2CF206ULL, "hlac_ht_playershouse_n_855_36z.dds"},
	{0x09F82837B1433375ULL, "179117506.ogg"},
	{0x09F8339E3DC7D313ULL, "6156119_m.fxe"},
	{0x09F89694D188C108ULL, "6091688_m.fxe"},
	{0x09F8F2E17B95D232ULL, "w_weapon.bnk"},
	{0x09F9DDF42DD8B570ULL, "995976751.ogg"},
	{0x09FA1A3C345C5244ULL, "905875303.ogg"},
	{0x09FA7303AD22DB0DULL, "hlac_lt_0_lowtown_a2_n_c_-131072_0x.dds"},
	{0x09FAA4F26EAE4DE3ULL, "vxlt_warehouse_n_11m_0_0.vlm"},
	{0x09FADB4AFE16990AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3.phy"},
	{0x09FADD6A61E223A5ULL, "vxw_encounter2_camp_a2_0_0.vlm"},
	{0x09FAEB85CE9B841AULL, "909503101.ogg"},
	{0x09FC8BCF33C1CAC1ULL, "zz_followers_banter6158943.cl"},
	{0x09FD1E375E65FCBDULL, "302676281.ogg"},
	{0x09FD33DA0871E99CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appa_0t.dds"},
	{0x09FE3AFD13EB0C92ULL, "6033542_f.fxe"},
	{0x09FE8F7D2C54700FULL, "656316680.ogg"},
	{0x09FE938DF4CD462CULL, "vxlt_gallowstemplar_5cz_0_0.vlm"},
	{0x09FF0D8ED16DC8E4ULL, "6166726_m.fxe"},
	{0x09FF4C24B8BEEB83ULL, "mer120_hubert6173905.cl"},
	{0x0A00718D10196592ULL, "zz_fol_debug6057008.cl"},
	{0x0A011758E0435A02ULL, "790830175.ogg"},
	{0x0A013772B5A66AD1ULL, "vxht_bartrand_n_66d_0_0.vlm"},
	{0x0A016C96A4BAAEF4ULL, "vxw_bonepitdestroyed_a3_ki_0_0.vlm"},
	{0x0A021E25ED84A07AULL, "hm_danarius.mor"},
	{0x0A026B0EDC677489ULL, "vxdr_primevaltomb_8ur_0_0.vlm"},
	{0x0A03208F575F7AA6ULL, "vxw_encounter2_camp_5a_0_0.vlm"},
	{0x0A0338DCD3D17719ULL, "113623799.ogg"},
	{0x0A03B01ABC6700C9ULL, "165395044.ogg"},
	{0x0A03BB9B6794DBF1ULL, "qcr220_cut_distraction6042317.cl"},
	{0x0A03C3D26A5723EEULL, "hlac_lt_0_lowtown_n_65537_0z.dds"},
	{0x0A04BB3EEB3045CAULL, "hlac_dr_primevaltomb_9621_5y.dds"},
	{0x0A04FD553CE08077ULL, "hlac_ht_0_hightown_a2_-65537_4y.dds"},
	{0x0A0505882DD177EBULL, "6139130_m.fxe"},
	{0x0A05508F35166FCCULL, "vxca_caverns2prc_a4n_0_0.vlm"},
	{0x0A066DE849190A67ULL, "36627740.ogg"},
	{0x0A06D12FDF201BBCULL, "vxlt_0_lowtown_n_f0t_0_0.vlm"},
	{0x0A072405908578F5ULL, "6161278_m.fxe"},
	{0x0A078734D4C0ADDBULL, "partypicker.ncs"},
	{0x0A07873FF7463A99ULL, "6036512_m.fxe"},
	{0x0A078DE87D070FDFULL, "w_encounter2_9_0.msh"},
	{0x0A0793E87D071981ULL, "w_encounter2_9_0.mmh"},
	{0x0A079FE87D072E22ULL, "w_encounter2_9_0.mao"},
	{0x0A0828F00B2771ADULL, "vxca_caverns1_mines_d5z_0_0.vlm"},
	{0x0A0916CF60A46663ULL, "vxlt_warehouse_n_16l_0_0.vlm"},
	{0x0A0918DEB22F1486ULL, "527915924.ogg"},
	{0x0A095DC8AE14953EULL, "zz_and_debug6178602.cl"},
	{0x0A09E68C40905412ULL, "night_probe_3654.mtx"},
	{0x0A09F2A8DBEB366CULL, "vxht_keep_6fu_0_0.vlm"},
	{0x0A0A30D0B2568849ULL, "vxw_woundedcoast_1pi_0_0.vlm"},
	{0x0A0A3A2C5D4EA6A9ULL, "203975519.ogg"},
	{0x0A0A470CEB2A7C0BULL, "vfx_vfx_p4.psh.fxo.dx11"},
	{0x0A0AC838D47C3C56ULL, "vxlt_0_lowtown_hat_0_0.vlm"},
	{0x0A0AE51BEB8E5A22ULL, "155516531.ogg"},
	{0x0A0AFCACAD51D8F5ULL, "hlac_lt_hangedman_n_-1_22y.dds"},
	{0x0A0B1F172C629032ULL, "vxlt_gallowstemplar_5an_0_0.vlm"},
	{0x0A0B59B3E40CA763ULL, "142089744.ogg"},
	{0x0A0BB1DDF85CA94FULL, "mag210_arianni6056861.cl"},
	{0x0A0BE094C9CB4587ULL, "hlac_lt_0_lowtown_a2_n_c_131073_2x.dds"},
	{0x0A0C28DBD6ABACF5ULL, "6035859_f.fxe"},
	{0x0A0C52C89ABDEF6BULL, "vxw_sundermount_370_0_0.vlm"},
	{0x0A0CDCB0DF56CF93ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrlb.phy"},
	{0x0A0D37740DB1337AULL, "gen00cs_cutscene_end.ncs"},
	{0x0A0DB3FCB5DF6C03ULL, "6114456_m.fxe"},
	{0x0A0F341618CD756CULL, "vxht_0_hightown_n_buc_0_0.vlm"},
	{0x0A0F4F0958BEA1C6ULL, "vxht_mansion2_3c1_0_0.vlm"},
	{0x0A0F86C753F63F04ULL, "llac_lt_0_lowtown_-196608_0.dds"},
	{0x0A1027B161EDC739ULL, "vxca_caverns2prc_aba_0_0.vlm"},
	{0x0A106DB89ED9CAEDULL, "266163076.ogg"},
	{0x0A10E2723EDB6132ULL, "vxht_0_hightown_bub_0_0.vlm"},
	{0x0A10E4884B77D513ULL, "6087129_m.fxe"},
	{0x0A1151422707F173ULL, "hlac_ca_caverns2_5802_47y.dds"},
	{0x0A11573FFB2B6C0EULL, "vxw_encounter2_64_0_0.vlm"},
	{0x0A11B270E83F1BD3ULL, "hlac_lt_0_lowtown_a2_n_131071_2x.dds"},
	{0x0A11E35A6C3B07FBULL, "vxht_keep_57z_0_0.vlm"},
	{0x0A11FE633D575F78ULL, "vxht_mansion2b_32h_0_0.vlm"},
	{0x0A1231853B151D5FULL, "6087600_m.fxe"},
	{0x0A13CC1A3DC33C40ULL, "follower_banter6141870.cl"},
	{0x0A147C1812432CA2ULL, "6163235_m.fxe"},
	{0x0A1538F26FFCD6DBULL, "lt_0_lowtown_docks_a3_117_0.phy"},
	{0x0A15613940E81010ULL, "dae000_mother6099211.cl"},
	{0x0A159F772E6BB036ULL, "vxlt_0_lowtown_docks_a3_9fj_0_0.vlm"},
	{0x0A16A41EDF630031ULL, "195044059.ogg"},
	{0x0A16C456ACB6BD2EULL, "6053697_m.fxe"},
	{0x0A16CBC37739CD31ULL, "sst_ss_sebastian.gda"},
	{0x0A17E4A07B1EB8C3ULL, "one100dg_elren6014256.cl"},
	{0x0A184B55EA35DDDFULL, "121974320.ogg"},
	{0x0A186485CDE14EECULL, "vxht_keep_6cj_0_0.vlm"},
	{0x0A18DE2FB9586C11ULL, "lgt210pt_frame_follow.ncs"},
	{0x0A18EFA4A8820F8AULL, "vxw_encounter2_jd_0_0.vlm"},
	{0x0A19643187FEC48CULL, "var_creature.gda"},
	{0x0A1975E05A6B71E2ULL, "plt_cod_itm_vague_blade.plo"},
	{0x0A1986B8B2EA7093ULL, "vxlt_warehouse_n_yf_0_0.vlm"},
	{0x0A1990F41E58A8B2ULL, "vxlt_gallowstemplar_5jo_0_0.vlm"},
	{0x0A19C05C18ACCE2AULL, "838152494.ogg"},
	{0x0A19E535004B2491ULL, "100095833.ogg"},
	{0x0A1A002E441304E2ULL, "6103888_m.fxe"},
	{0x0A1A3C5E8770AFF2ULL, "250358216.ogg"},
	{0x0A1AB633325885B8ULL, "lgt110_gxa_vgk_silver_order6117038.cl"},
	{0x0A1AC4A58CB407EBULL, "vxw_sundermount_323_0_0.vlm"},
	{0x0A1AD5F3715E25FBULL, "vxca_cavern1_haunted_j2a_0_2.vlm"},
	{0x0A1B7747D8B9772CULL, "vxlt_gallowstemplar_5ix_0_0.vlm"},
	{0x0A1B7BFBB827368FULL, "lt_0_lowtown_docks_n_39_0_lc.mmh"},
	{0x0A1BAFD3101F6388ULL, "lt_0_lowtown_a2_n_38_0.rml"},
	{0x0A1BB994A5EEA3D4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_nud_0d.dds"},
	{0x0A1BC0A4A4DB0FF9ULL, "vxlt_0_lowtown_a2_n_c_gy9_0_0.vlm"},
	{0x0A1BFDD1D1D67AC5ULL, "6115343_m.fxe"},
	{0x0A1C663970E5542FULL, "w_encounter3_85_0.phy"},
	{0x0A1D1CA3F8D957ECULL, "578594154.ogg"},
	{0x0A1D663EBE4A90AFULL, "8129522.ogg"},
	{0x0A1D726665146954ULL, "vxht_bartrand_n_47h_0_0.vlm"},
	{0x0A1DC0E64AB4BA46ULL, "vxht_mansion2_3f2_0_0.vlm"},
	{0x0A1E346D141DC793ULL, "vxlt_darktown_4un_0_0.vlm"},
	{0x0A1FC91CED21848EULL, "vxw_encounter2_15_0_0.vlm"},
	{0x0A20F42F525E009FULL, "6143218_m.fxe"},
	{0x0A21218D0CB42A2EULL, "gen00fl_anders6103781.cl"},
	{0x0A212EE5FE3DD835ULL, "vxht_keep_3il_0_0.vlm"},
	{0x0A2136AD321ABFE0ULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_ulri.phy"},
	{0x0A217E762C9B4312ULL, "6172282_m.fxe"},
	{0x0A21E04ECA58FC25ULL, "lowtown_guard_chase6169518.cl"},
	{0x0A21F93251F14665ULL, "1028545341.ogg"},
	{0x0A22029640AE15A9ULL, "6204807_m.fxe"},
	{0x0A2269EBF6901DCFULL, "hlac_lt_0_lowtown_docks_n_131071_4y.dds"},
	{0x0A22896A7D97BA82ULL, "vxht_brothel_n_3s9_0_0.vlm"},
	{0x0A229A0D63B18C32ULL, "vxht_0_hightown_a2_bl1_0_0.vlm"},
	{0x0A22E50C9762005EULL, "art\\characters\\creatures\\mubari\\animation\\com_pee.ani"},
	{0x0A22ED182ACCBD9DULL, "vxht_0_hightown_a3_c9y_0_0.vlm"},
	{0x0A233F203801072BULL, "255566682.ogg"},
	{0x0A238F7804364537ULL, "596938124.ogg"},
	{0x0A239119CCC58196ULL, "art\\characters\\weapons\\longbows\\w_lbw_001a.msh"},
	{0x0A239B19CCC59294ULL, "art\\characters\\weapons\\longbows\\w_lbw_001a.mmh"},
	{0x0A239F19CCC599A7ULL, "art\\characters\\weapons\\longbows\\w_lbw_001a.mao"},
	{0x0A23EB04BDE0F833ULL, "llac_lt_0_lowtown_131070_0.dds"},
	{0x0A23EBF8B693FE08ULL, "hlac_ht_0_hightown_n_-65536_4x.dds"},
	{0x0A244B4BAC50E835ULL, "716372601.ogg"},
	{0x0A2477B848097995ULL, "cry_val_b02.mao"},
	{0x0A24CBB6BFEA6616ULL, "147902997.ogg"},
	{0x0A24CECD47950B09ULL, "6053863_m.fxe"},
	{0x0A24D05544439A97ULL, "hlac_ca_cavern1_haunted_24424_18x.dds"},
	{0x0A25CCF542FE52E0ULL, "vxlt_0_lowtown_docks_a3_n_cx3_0_0.vlm"},
	{0x0A261259B529A550ULL, "6188714_f.fxe"},
	{0x0A262F4EB63CEEE2ULL, "6141503_m.fxe"},
	{0x0A267978F672C23DULL, "gen00fl_dog6167668.cl"},
	{0x0A26E6D1DF784379ULL, "glow.mao"},
	{0x0A2727929788CBCAULL, "vxht_chantry_n_31t_0_0.vlm"},
	{0x0A277236A494AD32ULL, "character_rendertodepth_p6.psh.fxo.dx11"},
	{0x0A278E67AE04F63DULL, "261125107.ogg"},
	{0x0A28C3781DC59273ULL, "6102143_m.fxe"},
	{0x0A28CE5F886194A6ULL, "vxca_caverns1_j2c_0_0.vlm"},
	{0x0A290E6CE88D64DDULL, "gen00fl_anders6096600.cl"},
	{0x0A2936827EAA206BULL, "vxw_sundermount_392_0_0.vlm"},
	{0x0A2A3343D6EF2B6EULL, "gen00fl_anders6125622.cl"},
	{0x0A2A6FAA06109A13ULL, "6057357_m.fxe"},
	{0x0A2AB0A41DE72386ULL, "and311ar_andersshop.are"},
	{0x0A2AC41824BDEF6AULL, "dae000_flemeth6000963.cl"},
	{0x0A2B554A6C8A6275ULL, "vxlt_darktown_2uz_0_0.vlm"},
	{0x0A2B9D435967F590ULL, "6027722_m.fxe"},
	{0x0A2BEBF011BC87ACULL, "face0_ds2.psh.fxo"},
	{0x0A2C73E87D25FFAAULL, "w_encounter2_9_0.rml"},
	{0x0A2C9F3D8B178820ULL, "6161086_m.fxe"},
	{0x0A2D1A45DFEC5AD2ULL, "430055694.ogg"},
	{0x0A2D3DB0154DC21AULL, "6156873_m.fxe"},
	{0x0A2D85DBC3F8019EULL, "m_dae310ar_docks.bnk"},
	{0x0A2DFA2CAE7A9B3EULL, "art\\vfx\\combat\\vfx_glb_ele_e_major_c.impa.ani"},
	{0x0A2E043882F9C39CULL, "hlac_ht_0_hightown_a3_-1_2x.dds"},
	{0x0A2E93C5DBDC9994ULL, "fxp_pillar_sym05.dds"},
	{0x0A2F3349E59C7108ULL, "990223814.ogg"},
	{0x0A2FFD0C976CB32FULL, "art\\characters\\creatures\\mubari\\animation\\com_pee.evt"},
	{0x0A304AFD3DAADF7CULL, "751881617.ogg"},
	{0x0A305DA8B7F232C4ULL, "lt_0_lowtown_docks_a3_n_233_0.msh"},
	{0x0A306FA8B7F2512DULL, "lt_0_lowtown_docks_a3_n_233_0.mao"},
	{0x0A3073A8B7F25866ULL, "lt_0_lowtown_docks_a3_n_233_0.mmh"},
	{0x0A315EF51CC2D61DULL, "vxht_0_hightown_a3_c2x_0_0.vlm"},
	{0x0A32198698FE8554ULL, "552911049.ogg"},
	{0x0A323A001F6654B9ULL, "art\\levels\\ztools\\default\\generic\\9cc_beam.msh"},
	{0x0A3244001F6665A7ULL, "art\\levels\\ztools\\default\\generic\\9cc_beam.mmh"},
	{0x0A3248001F666CF4ULL, "art\\levels\\ztools\\default\\generic\\9cc_beam.mao"},
	{0x0A324EE6ABF187D4ULL, "gen_im_arm_cht_nobd_t2.uti"},
	{0x0A326C34336BF838ULL, "324063312.ogg"},
	{0x0A3274603E7F3D35ULL, "vxdr_primevaltomb_50p_0_0.vlm"},
	{0x0A32AAA3F90C4ED3ULL, "6001008_m.fxe"},
	{0x0A32D4B12F3841DCULL, "6114849_f.fxe"},
	{0x0A32E90E00819E8FULL, "docks_male_human_worker_sta6176038.cl"},
	{0x0A3394E87D2C56FEULL, "w_encounter2_9_0.phy"},
	{0x0A339B653636BA7DULL, "40792108.ogg"},
	{0x0A33B19FC803300CULL, "637940195.ogg"},
	{0x0A340AA4C45018E9ULL, "vxht_brothel_n_6bp_0_0.vlm"},
	{0x0A3424FA46CDE976ULL, "lt_gallowscourtyard_n_mmi.gff"},
	{0x0A34817817BFB8D4ULL, "489252282.ogg"},
	{0x0A34AA4CE05F58BFULL, "vxht_keep_6fz_0_0.vlm"},
	{0x0A358B98588F36D7ULL, "189528705.ogg"},
	{0x0A358E5DFDC537B2ULL, "llac_lt_0_lowtown_a3_n_c_131073_0.dds"},
	{0x0A35A4D641BC7E9FULL, "mcr310_argument6127560.cl"},
	{0x0A35D35E8C6E408AULL, "vxw_encounter2_pf_0_0.vlm"},
	{0x0A35E72A8ACB95B4ULL, "vxlt_darktown_514_0_0.vlm"},
	{0x0A362AA201664AC0ULL, "vxlt_undercityserialkiller_1le_0_0.vlm"},
	{0x0A3674AE0244D9B2ULL, "vxlt_gallowstemplar_5pi_0_0.vlm"},
	{0x0A36BBF7B3E4BD08ULL, "hlac_ca_caverns1_24425_9y.dds"},
	{0x0A36E7CF3EE1AFA9ULL, "vxlt_0_lowtown_a2_n_c_e8l_0_0.vlm"},
	{0x0A36FE0C9772D3D8ULL, "art\\characters\\creatures\\mubari\\animation\\com_pee.gad"},
	{0x0A3713F20A16F183ULL, "6121759_m.fxe"},
	{0x0A37958212215624ULL, "979312894.ogg"},
	{0x0A37C60679C90116ULL, "2760946.ogg"},
	{0x0A38328F7F2F07DFULL, "hlac_ht_keep_n_3731_0y.dds"},
	{0x0A3852B3A255D9CAULL, "781781235.ogg"},
	{0x0A38F7AA1C765FB0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabsnb.phy"},
	{0x0A397F65DA1DFC23ULL, "vxht_0_hightown_n_b8c_0_0.vlm"},
	{0x0A398F5E770F8922ULL, "one300cr_ggt_thug5.utc"},
	{0x0A39DF9F78B52902ULL, "323845168.ogg"},
	{0x0A3A924BFFCE37CDULL, "65532006.ogg"},
	{0x0A3AC971A3A9A23FULL, "vxlt_gallowstemplar_n_45m_0_0.vlm"},
	{0x0A3B122CAE854E0FULL, "art\\vfx\\combat\\vfx_glb_ele_e_major_c.impa.evt"},
	{0x0A3B5860A000F202ULL, "character_character_p94.psh.fxo.dx11.assembly"},
	{0x0A3B72BDAD3FE7F0ULL, "one220_yevhen6102812.cl"},
	{0x0A3C76BE9D7B3EFFULL, "1002404453.ogg"},
	{0x0A3CFC92013E107BULL, "6111764_f.fxe"},
	{0x0A3D38641326EF90ULL, "253786197.ogg"},
	{0x0A3D5303E7C61ED5ULL, "vxht_0_hightown_a3_621_0_0.vlm"},
	{0x0A3D86FC1A00673BULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernlavaedge_e.dds"},
	{0x0A3E2FF2372BB0FBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutc.mmh"},
	{0x0A3E3BF2372BC498ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutc.mao"},
	{0x0A3E4DF2372BE3E5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutc.msh"},
	{0x0A3E7B1140834B00ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_0.pwk"},
	{0x0A3EC9F37448B3C2ULL, "vxw_sundermount_20m_0_0.vlm"},
	{0x0A3ED7940AD9E375ULL, "w_bonepitmining_a2_254_0.msh"},
	{0x0A3ED8A3660A1C19ULL, "art\\vfx\\combat\\vfx_glb_general_imp_c.impa.ani"},
	{0x0A3ED9940AD9E6EBULL, "w_bonepitmining_a2_254_0.mmh"},
	{0x0A3EDB972C45FA8DULL, "673834792.ogg"},
	{0x0A3EE5940AD9FA88ULL, "w_bonepitmining_a2_254_0.mao"},
	{0x0A3EF5D1DF8CEE07ULL, "glow.dds"},
	{0x0A3EFDBDC4226FB9ULL, "w_bonepit_137_0.msh"},
	{0x0A3F07BDC42280A7ULL, "w_bonepit_137_0.mmh"},
	{0x0A3F0BBDC42287F4ULL, "w_bonepit_137_0.mao"},
	{0x0A3F54DFCD9313B2ULL, "6080508_f.fxe"},
	{0x0A3FD0D20EB8EE9DULL, "vxht_0_hightown_a3_c7w_0_0.vlm"},
	{0x0A4055F4B9BF1105ULL, "vxht_mansion2b_32q_0_0.vlm"},
	{0x0A40A119AF3A0A5AULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobe.phy"},
	{0x0A411FF27021E77AULL, "lt_0_lowtown_docks_a3_117_0.msh"},
	{0x0A412DF27021FEB3ULL, "lt_0_lowtown_docks_a3_117_0.mao"},
	{0x0A4131F2702205C0ULL, "lt_0_lowtown_docks_a3_117_0.mmh"},
	{0x0A41FFDD2F3ED7DEULL, "vxw_encounter1_cn_0_0.vlm"},
	{0x0A42089435983EDEULL, "vxlt_hangedman_n_rr_0_0.vlm"},
	{0x0A42132CAE8B6EB8ULL, "art\\vfx\\combat\\vfx_glb_ele_e_major_c.impa.gad"},
	{0x0A427C81B6463169ULL, "vxht_brothel_n_6ow_0_0.vlm"},
	{0x0A42B8D85CA6064CULL, "art\\vfx\\cinematic\\vfx_lightbeam02_p.dura.ani"},
	{0x0A431C29D255713FULL, "vxht_keep_6ce_0_0.vlm"},
	{0x0A433E18CE10643CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_left_1.ani"},
	{0x0A436E2334589C1BULL, "vxlt_darktown_2qp_0_0.vlm"},
	{0x0A43AAD386114316ULL, "w_encounter3_35_0.rml"},
	{0x0A446C38222CDC01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v3.ani"},
	{0x0A459AC5DBEF84C2ULL, "fxp_pillar_sym05.mao"},
	{0x0A45B3A3660FFC5BULL, "art\\vfx\\combat\\vfx_glb_general_imp_c.impa.gad"},
	{0x0A46356802182CACULL, "6020445_f.fxe"},
	{0x0A46729A12E9F458ULL, "vxlt_warehouse_n_12w_0_0.vlm"},
	{0x0A46CED181DA3D18ULL, "mag360_alain6044657.cl"},
	{0x0A46EF650BBEF90EULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsml.mmh"},
	{0x0A46F3EE92E617E6ULL, "6111289_m.fxe"},
	{0x0A46FB650BBF0DA5ULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsml.mao"},
	{0x0A4709650BBF25FCULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsml.msh"},
	{0x0A4753AD323BC075ULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_ulri.msh"},
	{0x0A4755AD323BC3EBULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_ulri.mmh"},
	{0x0A476EDD513666BFULL, "681601438.ogg"},
	{0x0A482AF2D41E6689ULL, "art\\characters\\playercharacter\\elffemale\\ef_boo_lgte.phy"},
	{0x0A484D49D926F237ULL, "art\\vfx\\models\\vfx_arm_shield.mmh"},
	{0x0A486349D92717A9ULL, "art\\vfx\\models\\vfx_arm_shield.msh"},
	{0x0A48920ACE9CA84CULL, "793811975.ogg"},
	{0x0A48D256C83CF171ULL, "hlac_lt_0_lowtown_n_-3_3z.dds"},
	{0x0A4926303C7D2FD0ULL, "llac_lt_gallowscourtyard_n_131071_2.dds"},
	{0x0A4940F969F2C424ULL, "6002796_m.fxe"},
	{0x0A496BEA20117D14ULL, "539622371.ogg"},
	{0x0A49C5D85CAC3A8EULL, "art\\vfx\\cinematic\\vfx_lightbeam02_p.dura.gad"},
	{0x0A4A1F76F26DD3BCULL, "ser270_templar6165681.cl"},
	{0x0A4A4B18CE16997EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_left_1.gad"},
	{0x0A4A773822315AA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v3.gad"},
	{0x0A4A8412F9D0DAF4ULL, "410329322.ogg"},
	{0x0A4AA3CFB8125073ULL, "vxlt_foundry_n_v5_0_0.vlm"},
	{0x0A4AA3D38617563AULL, "w_encounter3_35_0.phy"},
	{0x0A4ACB19C7EF65F6ULL, "373068092.ogg"},
	{0x0A4AF1E7B7B33E32ULL, "vxlt_0_lowtown_a2_hg2_0_0.vlm"},
	{0x0A4B0AADAC918E16ULL, "dae300_alistair_king6133824.cl"},
	{0x0A4C8DBC1FC7B0F9ULL, "llac_w_encounter2_-65537_0.dds"},
	{0x0A4C94741E5638BBULL, "6071524_m.fxe"},
	{0x0A4CD0A366164C40ULL, "art\\vfx\\combat\\vfx_glb_general_imp_c.impa.evt"},
	{0x0A4D3BD0663ECC42ULL, "vxw_sundermount_25n_0_0.vlm"},
	{0x0A4D402DBE583DECULL, "6094904_m.fxe"},
	{0x0A4DB3392E72E386ULL, "hlac_lt_undercitydungeon_-1_41x.dds"},
	{0x0A4DBA13513FB8FFULL, "hlac_lt_0_lowtown_docks_a3_n_65533_1x.dds"},
	{0x0A4DE9AE6136FC18ULL, "llac_lt_0_lowtown_196607_0.dds"},
	{0x0A4E319E40698C6EULL, "gen00fl_fenris_base_max.stg"},
	{0x0A4E7F61AA4D4853ULL, "fen210_orana6109753.cl"},
	{0x0A4EB4C0A240F8BFULL, "hlac_dr_dungeon_3216_12z.dds"},
	{0x0A4F437D21E06250ULL, "vxlt_0_lowtown_docks_n_cq3_0_0.vlm"},
	{0x0A4F4DB483477C38ULL, "vxlt_gallowscourtyard_75t_0_0.vlm"},
	{0x0A4F797B7CAFC537ULL, "vxlt_gallowscourtyard_n_62d_0_0.vlm"},
	{0x0A4FAD6AEF22B924ULL, "vxlt_0_lowtown_cmq_0_0.vlm"},
	{0x0A4FC0D85CB09E6DULL, "art\\vfx\\cinematic\\vfx_lightbeam02_p.dura.evt"},
	{0x0A5004A472E282CEULL, "zz_prmpt_debug6198217.cl"},
	{0x0A5071BA2134F05EULL, "vxw_encounter1_ho_0_0.vlm"},
	{0x0A50B7A41E07FF71ULL, "and311ar_andersshop.lst"},
	{0x0A50C3BB5FFCF9C0ULL, "llac_w_encounter2_327684_0.dds"},
	{0x0A50ED4887F9AFF8ULL, "6208464_m.fxe"},
	{0x0A50EE5EA83C49E9ULL, "vxht_brothel_n_6hv_0_0.vlm"},
	{0x0A50EEE8980DB240ULL, "6125333_f.fxe"},
	{0x0A51198CBDE47178ULL, "vxlt_0_lowtown_docks_a3_n_cwr_0_0.vlm"},
	{0x0A512618CE1C789DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_left_1.evt"},
	{0x0A516E26FE688220ULL, "6147932_m.fxe"},
	{0x0A517279E753B2CEULL, "w_sundermount_217_0.phy"},
	{0x0A5197437F3D5670ULL, "one100dg_elren6058864.cl"},
	{0x0A51AA3F6E3B8FB9ULL, "vxlt_0_lowtown_a2_n_c_hsx_0_0.vlm"},
	{0x0A51E000264EB49BULL, "vxlt_darktown_2ts_0_0.vlm"},
	{0x0A51E9D2AF28DE6EULL, "995692729.ogg"},
	{0x0A51F6FCD3E39415ULL, "778360758.ogg"},
	{0x0A52BE36B21E191FULL, "vxht_0_hightown_a2_brj_0_0.vlm"},
	{0x0A537C84CDF2258EULL, "111183094.ogg"},
	{0x0A53E544A53CD069ULL, "llac_lt_0_lowtown_n_-1_7.dds"},
	{0x0A54A744F40E452CULL, "6027545_f.fxe"},
	{0x0A54B7630B103EA5ULL, "6015803_m.fxe"},
	{0x0A54C7ADF12E33E4ULL, "1048997341.ogg"},
	{0x0A54C8CB0C6FE5CDULL, "store_dae310_weapons.utm"},
	{0x0A54F3E2B0E02384ULL, "185208169.ogg"},
	{0x0A5617D91826ED10ULL, "tent_n.dds"},
	{0x0A5720EA5AAB4099ULL, "hlac_ca_caverns2_5802_46x.dds"},
	{0x0A574863FFED593FULL, "6077947_f.fxe"},
	{0x0A574C2D35BA2EF3ULL, "nat380_sibling6207560.cl"},
	{0x0A577FEBC70E2041ULL, "6121205_m.fxe"},
	{0x0A5789E674F707EDULL, "vxdr_primevaltomb_75a_0_0.vlm"},
	{0x0A57E8D2A682BBB1ULL, "art\\vfx\\combat\\material__5.mao"},
	{0x0A5803BDE80B223EULL, "vxlt_hangedman_n_2xa_0_1.vlm"},
	{0x0A587492BB3749E7ULL, "461559694.ogg"},
	{0x0A58B7A18D73E859ULL, "984695978.ogg"},
	{0x0A595CD4D28B7758ULL, "one100dg_kelder6060541.cl"},
	{0x0A5A8CADFE63B6B6ULL, "464352406.ogg"},
	{0x0A5AEAFE6693F417ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chaircutup_0.pwk"},
	{0x0A5AF3A02DB52FFCULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_vial_sa.dds"},
	{0x0A5B3EFB266F90C3ULL, "lowtown_male_alienage1_sta.bnk"},
	{0x0A5B474AEB24F39AULL, "gen00cs_cs_end_music_int_1.ncs"},
	{0x0A5B6CD7BE750C9CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgtp_0d.dds"},
	{0x0A5B9DA914776806ULL, "hlac_lt_gallowstemplar_n_2383_10z.dds"},
	{0x0A5BBB53143D63FBULL, "214262493.ogg"},
	{0x0A5C21DE99470A09ULL, "481626891.ogg"},
	{0x0A5C24DDECA62497ULL, "zz_mag_debug6053423.cl"},
	{0x0A5C34155984B016ULL, "6043380_m.fxe"},
	{0x0A5C822F26136141ULL, "vxca_caverns1_i2w_0_0.vlm"},
	{0x0A5CAB4792BC7B6AULL, "zz_dre_debug6069193.cl"},
	{0x0A5CB97DA1004011ULL, "416003643.ogg"},
	{0x0A5CFFD39358B41FULL, "lt_0_lowtown_docks_n_104_0.phy"},
	{0x0A5DA0463F02ECF4ULL, "vxht_bartrand_n_5w1_0_0.vlm"},
	{0x0A5DA09C1A00C1CBULL, "fxa_evl_dusts.mao"},
	{0x0A5DF635EA045451ULL, "6180675_m.fxe"},
	{0x0A5E96B8A0772ABBULL, "hlac_lt_0_lowtown_docks_-131073_3z.dds"},
	{0x0A5EA72A7A4E2F8FULL, "vxlt_gallowscourtyard_n_6xr_0_0.vlm"},
	{0x0A5F20FB267346F3ULL, "lowtown_male_alienage1_sta.cnv"},
	{0x0A5F4B70B74626C4ULL, "hlac_lt_0_lowtown_docks_65534_5x.dds"},
	{0x0A5F7B37C9530CFDULL, "784097550.ogg"},
	{0x0A5FD4958189B24DULL, "200528055.ogg"},
	{0x0A5FE372295620FEULL, "6043897_f.fxe"},
	{0x0A6111E2F87F681EULL, "vxlt_warehouse_11o_0_0.vlm"},
	{0x0A6134F33B01BCFDULL, "t1_mul_re2.tnt"},
	{0x0A61D1FA5E310B57ULL, "zz_rdr_debug6066631.cl"},
	{0x0A61D229243F820AULL, "l_dae211ar_gamlens_house_nt.wbd"},
	{0x0A61D5B851F6A49EULL, "6037388_f.fxe"},
	{0x0A621669DA787CE9ULL, "gen00fl_merrill6162123.cl"},
	{0x0A6270EE5B9007EDULL, "6040830_m.fxe"},
	{0x0A62C1843E8A2E26ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_act_c.impa.ani"},
	{0x0A62F630A21E3165ULL, "839946880.ogg"},
	{0x0A63022986448B35ULL, "vxw_bonepit_yr_0_0.vlm"},
	{0x0A63287B7B2E2F02ULL, "343617888.ogg"},
	{0x0A646355CEE93506ULL, "vxht_0_hightown_a3_bwa_0_0.vlm"},
	{0x0A6464FFAE01D77DULL, "hm_nathaniel.mor"},
	{0x0A647F8CFE6607DFULL, "follower_banter6165662.cl"},
	{0x0A64F916A6FA5630ULL, "clf260_harley6069679.cl"},
	{0x0A659B4DA7A1A676ULL, "plt_cod_itm_last_descent.plo"},
	{0x0A65AF4699D178F0ULL, "dae210_gate_guard6085082.cl"},
	{0x0A66A9B7FD6604EBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keepthin_1.mao"},
	{0x0A66AB7205DB5D9AULL, "leave_f.mao"},
	{0x0A672D9E7547E349ULL, "571489586.ogg"},
	{0x0A6766B0D10C2479ULL, "gen_im_trs_qun_epic_sword.uti"},
	{0x0A676ADBE3CF2EFEULL, "vxht_0_hightown_a3_boq_0_0.vlm"},
	{0x0A67A31D9A797705ULL, "vxca_caverns1_mines_cn5_0_1.vlm"},
	{0x0A687F1F76DF0043ULL, "gen00fl_dog6166430.cl"},
	{0x0A687F640C8622FBULL, "art\\characters\\weapons\\staff\\textures\\w_stf_010a_0d.dds"},
	{0x0A68A1AC9C772DF4ULL, "6068496_f.fxe"},
	{0x0A68B80C97616C17ULL, "vxht_0_hightown_9ux_0_0.vlm"},
	{0x0A699C7875E75902ULL, "vxht_bartrand_n_6rg_0_0.vlm"},
	{0x0A69C28CE86E59B5ULL, "6040889_m.fxe"},
	{0x0A69FB874E78F96AULL, "vxlt_0_lowtown_a2_n_c_fga_0_0.vlm"},
	{0x0A6A504DA92F730BULL, "vxlt_darktown_43d_0_0.vlm"},
	{0x0A6A721ED584DEF1ULL, "6005030_f.fxe"},
	{0x0A6A75EB970E6247ULL, "gen00fl_dog6107350.cl"},
	{0x0A6AB3B706CA9361ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\patrolling.gad"},
	{0x0A6AF0BDC44794DCULL, "w_bonepit_137_0.phy"},
	{0x0A6B4069A9F11183ULL, "llac_w_encounter3_-131075_0.dds"},
	{0x0A6C122330F90574ULL, "vxht_bartrand_n_5z6_0_0.vlm"},
	{0x0A6C24A01B7E55E1ULL, "llac_ht_0_hightown_n_-65534_1.dds"},
	{0x0A6C5DC696AC134CULL, "vxw_woundedcoast_1ge_0_0.vlm"},
	{0x0A6D15DFF5A18090ULL, "vxw_blightlands_by_0_0.vlm"},
	{0x0A6D59CEBB160167ULL, "6161531_m.fxe"},
	{0x0A6D6B6C1745D860ULL, "dae200cr_viscount6087233.cl"},
	{0x0A6DED236DC79A22ULL, "6132493_m.fxe"},
	{0x0A6E00D393672924ULL, "lt_0_lowtown_docks_n_104_0.mmh"},
	{0x0A6E04D393673077ULL, "lt_0_lowtown_docks_n_104_0.mao"},
	{0x0A6E16D393674EC6ULL, "lt_0_lowtown_docks_n_104_0.msh"},
	{0x0A6E57DBC42F4AA4ULL, "m_dae310ar_docks.wbd"},
	{0x0A6EA2BD46A3A671ULL, "768846778.ogg"},
	{0x0A6ED5B0DFA98E08ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrlb.mmh"},
	{0x0A6EE1B0DFA9A2ABULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrlb.mao"},
	{0x0A6EF3B0DFA9C132ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrlb.msh"},
	{0x0A6F24D413EFF49AULL, "hlac_w_woundedcoast_-131068_0y.dds"},
	{0x0A6F93DBA3EFECDFULL, "6030493_m.fxe"},
	{0x0A6FA9843E948F97ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_act_c.impa.evt"},
	{0x0A6FEE2E7E8AD70AULL, "vxw_blightlands_6o_0_0.vlm"},
	{0x0A70B4B706CF000FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\patrolling.ani"},
	{0x0A70EA7B46DD7E21ULL, "6119710_m.fxe"},
	{0x0A713C4C55B3EB65ULL, "vxdr_0_deeproads_exb_0_0.vlm"},
	{0x0A716FC2EF3037C8ULL, "hlac_lt_gallowstemplar_2391_7x.dds"},
	{0x0A717D9C22D95BCBULL, "594828869.ogg"},
	{0x0A7215C7A57F2219ULL, "llac_ht_mansion2_818_38.dds"},
	{0x0A7261B2792305CFULL, "277415962.ogg"},
	{0x0A72B39E5D1064D6ULL, "fen110_fenris6118126.cl"},
	{0x0A72D5F95A3F9EC7ULL, "vxlt_gallowstemplar_n_3di_0_0.vlm"},
	{0x0A72F84C0FDBE9FFULL, "6071949_f.fxe"},
	{0x0A72F978F3B7E82CULL, "gen00fl_dog6165728.cl"},
	{0x0A73507AF3AEAECFULL, "408993787.ogg"},
	{0x0A7376EF732A5085ULL, "591604512.ogg"},
	{0x0A73D97E3DC9607CULL, "lgt211cr_alistair_drunk.utc"},
	{0x0A7413F2D4437AAEULL, "art\\characters\\playercharacter\\elffemale\\ef_boo_lgte.mmh"},
	{0x0A742DF2D443A69CULL, "art\\characters\\playercharacter\\elffemale\\ef_boo_lgte.msh"},
	{0x0A743E7EFC852943ULL, "409657783.ogg"},
	{0x0A74455C05DE59E1ULL, "hlac_lt_0_lowtown_-65537_6y.dds"},
	{0x0A744B783BDAFD09ULL, "vxw_woundedcoast_nz_0_0.vlm"},
	{0x0A7462F291D75C00ULL, "mag170_recruit016094518.cl"},
	{0x0A746A883ED9167DULL, "hlac_lt_0_lowtown_a2_-196609_1y.dds"},
	{0x0A750D9093032A0CULL, "llac_lt_0_lowtown_docks_a3_-3_1.dds"},
	{0x0A751E67BAB4E2F5ULL, "vxlt_undercityencounter_1ff_0_0.vlm"},
	{0x0A75532FA14C72D9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basic.msh"},
	{0x0A755D2FA14C83C7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basic.mmh"},
	{0x0A75612FA14C8A14ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basic.mao"},
	{0x0A75BA9C681480E1ULL, "47635176.ogg"},
	{0x0A75BBC21EB5AC65ULL, "6172861_m.fxe"},
	{0x0A75E94E7787A753ULL, "390526853.ogg"},
	{0x0A760F1F5596C366ULL, "hightown_male_noble2_mob6170930.cl"},
	{0x0A761475137FAC4EULL, "zz_rogue_dragon.cut"},
	{0x0A76820FDD69D222ULL, "vxlt_gallowscourtyard_7en_0_0.vlm"},
	{0x0A769C852147E89FULL, "hlac_w_encounter2_camp_262144_0z.dds"},
	{0x0A76AA843E9AB020ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_act_c.impa.gad"},
	{0x0A76E4AA7FBA3C7DULL, "671824650.ogg"},
	{0x0A773D930436F4C0ULL, "6148681_m.fxe"},
	{0x0A779AF5A8667ABFULL, "zz_qcr_debug6045069.cl"},
	{0x0A77B5680623F377ULL, "6130959_m.fxe"},
	{0x0A7836676EA73A36ULL, "vxdr_0_deeproads_euv_0_0.vlm"},
	{0x0A786B1C2E22E68BULL, "6018815_m.fxe"},
	{0x0A788CE122D2F4C1ULL, "vxw_encounter3_33_0_0.vlm"},
	{0x0A78973DC9ABA01BULL, "hlac_lt_0_lowtown_a2_n_c_131070_6y.dds"},
	{0x0A7934E4D292F836ULL, "6098960_m.fxe"},
	{0x0A7956945D55AB7AULL, "gen00fl_merrill6116150.cl"},
	{0x0A7970FE37C2E26AULL, "vxw_woundedcoast_5h_0_0.vlm"},
	{0x0A797B97C68BE417ULL, "690572462.ogg"},
	{0x0A798249F74183AFULL, "6130101_f.fxe"},
	{0x0A7997E73AAD8240ULL, "dae000cr_grand_cleric6148129.cl"},
	{0x0A79C6EE1421CF13ULL, "v_c_blood_sh.mao"},
	{0x0A7A4504C4E17230ULL, "vxlt_darktown_4zd_0_0.vlm"},
	{0x0A7AB7AA4F0481E7ULL, "405312866.ogg"},
	{0x0A7ABE3047778B26ULL, "265450046.ogg"},
	{0x0A7AC7F1435F65C6ULL, "6138792_m.fxe"},
	{0x0A7ADB7291E44759ULL, "zz_dae_debug6029511.cl"},
	{0x0A7AE04986034E2BULL, "vxht_mansion2_2f6_0_0.vlm"},
	{0x0A7B0C58D813EE7FULL, "vxlt_gallowscourtyard_n_6y1_0_0.vlm"},
	{0x0A7B20184F73DB13ULL, "6129900_m.fxe"},
	{0x0A7B6A10BD0C8374ULL, "mag100_karras6034709.cl"},
	{0x0A7B7DAFE95F339AULL, "6038687_m.fxe"},
	{0x0A7C8B0C387E6A5DULL, "vxdr_0_deeproads_bo0_0_0.vlm"},
	{0x0A7C8DCD5E84BE0FULL, "llac_w_encounter1_131070_0.dds"},
	{0x0A7CAAE3B29FA57DULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga103a_0d.dds"},
	{0x0A7CACF41D1CF2B9ULL, "mrl150_merrill6070991.cl"},
	{0x0A7CAEF6FFC444E9ULL, "6117629_f.fxe"},
	{0x0A7CD02030659545ULL, "6217304_m.fxe"},
	{0x0A7CDE62F2C1EE8BULL, "878723620.ogg"},
	{0x0A7D281822BF09FCULL, "hlac_ht_0_hightown_n_-65536_3y.dds"},
	{0x0A7D94D945F3CD57ULL, "gen_im_trs_dwv_junk2_01.uti"},
	{0x0A7DA1CCB99ED87AULL, "91555014.ogg"},
	{0x0A7DA6638AA8C456ULL, "837047100.ogg"},
	{0x0A7E515B7E3D2EA7ULL, "w_encounter2_42_0_ll.mmh"},
	{0x0A7E562D1BA16B61ULL, "6072259_m.fxe"},
	{0x0A7E9164120D20DCULL, "prty_carver_armor_2_templar.dds"},
	{0x0A7EAC0F50D1A8D3ULL, "128816296.ogg"},
	{0x0A7F3AE964F55D97ULL, "237525966.ogg"},
	{0x0A7F8EB364FD287AULL, "vxdr_primevalentrance_1a8_0_0.vlm"},
	{0x0A7F9B4CD00273B3ULL, "gen00fl_aveline6063555.cl"},
	{0x0A7FCC6D1FA5977DULL, "496463722.ogg"},
	{0x0A805326CE1B994EULL, "vxht_0_hightown_a3_4kg_0_0.vlm"},
	{0x0A80B2B2A13AABF3ULL, "green.dds"},
	{0x0A8118C2F0E6B977ULL, "13993929.ogg"},
	{0x0A81470FB2D56606ULL, "vxht_0_hightown_a3_bic_0_0.vlm"},
	{0x0A8163DD44E589DAULL, "follower_banter6155640.cl"},
	{0x0A817254897B5F1CULL, "qun200_party_comment6090455.cl"},
	{0x0A82280EB1F1EDE6ULL, "vxht_playershouse_n_1q7_0_0.vlm"},
	{0x0A82BD552DD11589ULL, "vxw_woundedcoast_ky_0_0.vlm"},
	{0x0A8366BBD4934466ULL, "vxlt_0_lowtown_a2_n_c_eqd_0_0.vlm"},
	{0x0A8374BADB6C0784ULL, "465924356.ogg"},
	{0x0A83854B52C7B66CULL, "6091891_m.fxe"},
	{0x0A83EFEDC23E2DC4ULL, "gen00fl_aveline6132780.cl"},
	{0x0A8459A463ECF293ULL, "dae111_gamlen_intro_2.cut"},
	{0x0A847508AA0287C0ULL, "6133012_m.fxe"},
	{0x0A84801C748A9ACBULL, "art\\vfx\\creatures\\fxc_arc_shd_c.dura.evt"},
	{0x0A84C52FF11D5F46ULL, "vxw_encounter3_d1_0_0.vlm"},
	{0x0A85100C9FC77119ULL, "hlac_lt_blackemporium_1055_20y.dds"},
	{0x0A85453AD11AA8B1ULL, "619071971.ogg"},
	{0x0A857ED42C35F87FULL, "and121im_alrik_letter.uti"},
	{0x0A85847205F5CC4CULL, "leave_f.dds"},
	{0x0A861F6DC43284CCULL, "126994267.ogg"},
	{0x0A869F0BE24AFB15ULL, "vxht_0_hightown_a3_ds_0_0.vlm"},
	{0x0A86A9A7A9502A30ULL, "zz_dog_debug6204094.cl"},
	{0x0A8713B653C7D15CULL, "6108142_m.fxe"},
	{0x0A872221C898A335ULL, "vxlt_0_lowtown_n_f1v_0_0.vlm"},
	{0x0A87244FC420228FULL, "lgt220ip_shp_mag_suspect.ncs"},
	{0x0A876E381D321309ULL, "460785571.ogg"},
	{0x0A878B12F72AE248ULL, "638230059.ogg"},
	{0x0A87C42784AF5955ULL, "332040683.ogg"},
	{0x0A87CFBCD6784D4DULL, "426668809.ogg"},
	{0x0A87E2DB29B8FAEAULL, "vxw_woundedcoast_2i_0_0.vlm"},
	{0x0A8804C519949BFBULL, "319603203.ogg"},
	{0x0A8818CC7FBB03A2ULL, "330891370.ogg"},
	{0x0A8828C26FD0E02BULL, "6096839_m.fxe"},
	{0x0A888083F0736004ULL, "6004376_m.fxe"},
	{0x0A88D0E1CFBD1D5EULL, "vxht_mansion_n_5cx_0_0.vlm"},
	{0x0A88E78A26918C03ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_darren.ani"},
	{0x0A8954BA24589951ULL, "40947416.ogg"},
	{0x0A8988AA59685136ULL, "rev_twi.mao"},
	{0x0A8A928069349C4AULL, "1067818744.ogg"},
	{0x0A8B76D6AC505ADCULL, "follower_banter6151332.cl"},
	{0x0A8B7AA3F64275D0ULL, "gen00fl_aveline6106880.cl"},
	{0x0A8B93F32D15F96AULL, "dae000_madam_lusine6152080.cl"},
	{0x0A8B991C7490E38CULL, "art\\vfx\\creatures\\fxc_arc_shd_c.dura.gad"},
	{0x0A8BB190E5590EDBULL, "zz_fen_debug6165492.cl"},
	{0x0A8C1001D45B7AAAULL, "plt_cod_itm_finesse.plo"},
	{0x0A8C7FA3700D896DULL, "hlac_w_encounter1_131070_0z.dds"},
	{0x0A8C9067FD50D39CULL, "vxlt_0_lowtown_env_0_0.vlm"},
	{0x0A8C99B6FE9FB201ULL, "lt_0_lowtown_a2_n_53_0.rml"},
	{0x0A8C9F7F57C8CA17ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_com_c.impa.gad"},
	{0x0A8D30C562564127ULL, "vxca_caverns2_71i_0_0.vlm"},
	{0x0A8D98AB6394FB99ULL, "355521385.ogg"},
	{0x0A8DCE352AC9AF21ULL, "6113110_m.fxe"},
	{0x0A8E6A5768CEC9FCULL, "follower_banter6170724.cl"},
	{0x0A8E736F14AF9D66ULL, "pro000_friedrich6050171.cl"},
	{0x0A8E8A41266AF66FULL, "134959899.ogg"},
	{0x0A8F0016D6358C14ULL, "vxlt_gallowsprison_n_6o6_0_0.vlm"},
	{0x0A8FC32953A30650ULL, "703572349.ogg"},
	{0x0A90998B26EE3C07ULL, "leave_01.mao"},
	{0x0A90AD68A9B14C22ULL, "vxlt_0_lowtown_docks_bcu_0_0.vlm"},
	{0x0A90BDB3A0D4654FULL, "6217135_m.fxe"},
	{0x0A91048965941045ULL, "vxw_sundermount_315_0_1.vlm"},
	{0x0A9114785A601F28ULL, "206116431.ogg"},
	{0x0A914BFC338ABDB6ULL, "vxlt_warehouse_104_0_0.vlm"},
	{0x0A914C56880F8DD7ULL, "6104017_m.fxe"},
	{0x0A917179E7899491ULL, "w_sundermount_217_0.mmh"},
	{0x0A917D79E789A932ULL, "w_sundermount_217_0.mao"},
	{0x0A918B79E789C14FULL, "w_sundermount_217_0.msh"},
	{0x0A91981C74954D8AULL, "art\\vfx\\creatures\\fxc_arc_shd_c.dura.ani"},
	{0x0A923862BCA5F52CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_gesture_right.gad"},
	{0x0A93B00B691BA0B6ULL, "follower_banter6141850.cl"},
	{0x0A93B47F57CF0CA4ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_com_c.impa.evt"},
	{0x0A93BCC0AF429FDDULL, "6119310_m.fxe"},
	{0x0A93FC7AF4931573ULL, "mag311_nyssa6042233.cl"},
	{0x0A941014CA401546ULL, "vxca_caverns1_inx_0_0.vlm"},
	{0x0A9474834783B3A1ULL, "follower_banter6154048.cl"},
	{0x0A94B03137966846ULL, "fx_rage_flip.mao"},
	{0x0A94D771654394F1ULL, "vxlt_0_lowtown_a2_n_c_hlm_0_0.vlm"},
	{0x0A94EB6328387224ULL, "ser121_jethann6104426.cl"},
	{0x0A9531D6F4B31181ULL, "follower_barks6171983.cl"},
	{0x0A95487C455F49CCULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masf_0d.dds"},
	{0x0A9548AFCD0DA996ULL, "vxca_caverns1_mines_efu_0_0.vlm"},
	{0x0A954D50A30F5E3EULL, "lt_0_lowtown_a2_n_38_0_ll.mmh"},
	{0x0A9593FEBA8EBBB5ULL, "vxlt_0_lowtown_n_f4u_0_0.vlm"},
	{0x0A95DF501285DF7FULL, "hlac_ht_0_hightown_n_-1_1z.dds"},
	{0x0A95F8F0DC308C30ULL, "6115474_m.fxe"},
	{0x0A966C1AA8BA5A8DULL, "739674241.ogg"},
	{0x0A96B54E6EE7D2B8ULL, "mrl000_marethari6129254.cl"},
	{0x0A96BFC93D7EBF71ULL, "dre000_dougal6129740.cl"},
	{0x0A96C48898A17CC1ULL, "6094673_m.fxe"},
	{0x0A972231B7981BDAULL, "lgt160_ntb_elf_conflict6163369.cl"},
	{0x0A979DFBF84E2ECFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medd_0t.dds"},
	{0x0A9822567BE98B5FULL, "6092512_m.fxe"},
	{0x0A9876D41E8A9D2DULL, "follower_banter6163723.cl"},
	{0x0A9916FD0407F99AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkl.gad"},
	{0x0A993762BCAC122AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_gesture_right.ani"},
	{0x0A9960E2E8E04760ULL, "gen00fl_anders6125224.cl"},
	{0x0A99A997590224E5ULL, "pro000_athenril6048725.cl"},
	{0x0A99CCED3F63BC44ULL, "hlac_ht_mansion2_818_26z.dds"},
	{0x0A99FAF076818E32ULL, "hlac_ca_cavern1_haunted_24424_28z.dds"},
	{0x0A9AA416D94494AAULL, "vxlt_undercityserialkiller_1iq_0_1.vlm"},
	{0x0A9AB0921F31FD2CULL, "art\\characters\\creatures\\buzzard\\textures\\c_buzzard_0n.dds"},
	{0x0A9B0244EF46EC1CULL, "vxlt_0_lowtown_eiw_0_0.vlm"},
	{0x0A9B27F924931A3FULL, "vxca_cave_23h_0_1.vlm"},
	{0x0A9B71FD1EDD4BF3ULL, "271315077.ogg"},
	{0x0A9B80199ACDC5ECULL, "hla_ht_0_hightown_n_1_0z.dds"},
	{0x0A9BEF72A75A05BEULL, "335256989.ogg"},
	{0x0A9C02E0EFB005CAULL, "testhurlock.dds"},
	{0x0A9C196C981750C3ULL, "zz_isa_debug6133963.cl"},
	{0x0A9C2C2E23E61B19ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb101a_0d.dds"},
	{0x0A9C9A19AF885C83ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobe.msh"},
	{0x0A9CA711031E8621ULL, "vxlt_darktown_52t_0_0.vlm"},
	{0x0A9CA819AF8874DDULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobe.mmh"},
	{0x0A9CAC19AF887BEEULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobe.mao"},
	{0x0A9CB17A7260486DULL, "582283085.ogg"},
	{0x0A9CD44809DA9B2BULL, "vxht_0_hightown_a3_anh_0_0.vlm"},
	{0x0A9CE89B21559206ULL, "6210162_m.fxe"},
	{0x0A9CFE8A26A25CEDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_darren.gad"},
	{0x0A9D57B1FB023CD8ULL, "vxlt_backalley_n_1e7_0_0.vlm"},
	{0x0A9D71F3C82BA494ULL, "vxlt_gallowsprison_n_6h7_0_0.vlm"},
	{0x0A9DD63DF3EBDF1CULL, "vxca_cavern1_haunted_eh5_0_0.vlm"},
	{0x0A9F1F459BA764A2ULL, "vxlt_0_lowtown_docks_bfv_0_0.vlm"},
	{0x0A9FB372A90AB521ULL, "nat380_nathaniel6066493.cl"},
	{0x0A9FE2D3C8E19BC7ULL, "zz_charge_blight2.cut"},
	{0x0AA011B046D0115EULL, "590462237.ogg"},
	{0x0AA019FD040E1DB1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkl.evt"},
	{0x0AA032C930C3AAF1ULL, "6155286_m.fxe"},
	{0x0AA04E300E9A480FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_right_hand.gad"},
	{0x0AA06E3560FCC329ULL, "hlac_lt_0_lowtown_docks_-65535_0x.dds"},
	{0x0AA07C34C0021021ULL, "var_trigger.gda"},
	{0x0AA0CA3E8304877FULL, "6154207_m.fxe"},
	{0x0AA0CC7F57D9BF9DULL, "art\\vfx\\combat\\vfx_w_2nd_atk_com_c.impa.ani"},
	{0x0AA10588490AD48FULL, "6208558_f.fxe"},
	{0x0AA12AF11B70D6C3ULL, "314456149.ogg"},
	{0x0AA162966099F4F4ULL, "docks_fem_sailor_mob6176421.cl"},
	{0x0AA17CFC558800DBULL, "vxlt_0_lowtown_a3_n_c_e2e_0_0.vlm"},
	{0x0AA1C4E9BA46712FULL, "6179444_m.fxe"},
	{0x0AA1EE5CAA12EFF5ULL, "6201371_f.fxe"},
	{0x0AA2074B572E8871ULL, "gen00fl_fenris6126952.cl"},
	{0x0AA2462FA1734AFCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basic.phy"},
	{0x0AA26865576831ACULL, "pro000_colonel_ewald6047377.cl"},
	{0x0AA2D7CEDA1ECAA6ULL, "gen_im_trs_drk_two_2hs_r2.uti"},
	{0x0AA34A2EE32C2695ULL, "373793276.ogg"},
	{0x0AA3826C0EB00EA2ULL, "6068362_f.fxe"},
	{0x0AA4A454031CBDC1ULL, "vxht_keep_n_6y6_0_0.vlm"},
	{0x0AA51D81E399DD2FULL, "6107236_m.fxe"},
	{0x0AA5B49BB3A94E5EULL, "vxht_mansion_n_5iz_0_0.vlm"},
	{0x0AA5D8C63F15B953ULL, "vxlt_hangedman_n_2dm_0_0.vlm"},
	{0x0AA5E11392489BCFULL, "dre000_vampire6034329.cl"},
	{0x0AA5E2B45954ACDEULL, "vxlt_hangedman_210_0_0.vlm"},
	{0x0AA5E2DBA8A25F17ULL, "vxht_chantry_n_23v_0_0.vlm"},
	{0x0AA663300E9ED7BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_right_hand.evt"},
	{0x0AA69596B39D20E7ULL, "art\\vfx\\combat\\vfx_sg_quake_act_c.anb"},
	{0x0AA69CD38665A963ULL, "w_encounter3_35_0.msh"},
	{0x0AA6AAD38665C0BDULL, "w_encounter3_35_0.mmh"},
	{0x0AA6AED38665C7CEULL, "w_encounter3_35_0.mao"},
	{0x0AA719D15013A827ULL, "hlac_ca_caverns1_mines_24424_0y.dds"},
	{0x0AA7443D9C1AB408ULL, "6083805_m.fxe"},
	{0x0AA7A08B2701275DULL, "leave_01.dds"},
	{0x0AA7B89F3230B00BULL, "dae021_sister_colinda_sta6150203.cl"},
	{0x0AA8B029247B4B20ULL, "l_dae211ar_gamlens_house_nt.bnk"},
	{0x0AA91BBC24FFDBEAULL, "6051329_m.fxe"},
	{0x0AA95CFC2FCF5829ULL, "dae000_sabina6162342.cl"},
	{0x0AA9BDB1E4E16187ULL, "6071955_m.fxe"},
	{0x0AA9C29F0C956BD3ULL, "6001219_m.fxe"},
	{0x0AAA707BF1EFD4ADULL, "vxht_keep_5xa_0_0.vlm"},
	{0x0AAA86576971B7C5ULL, "vxdr_primevalentrance_1a3_0_0.vlm"},
	{0x0AAA8F15573DCEC7ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_0z.dds"},
	{0x0AAA9BEEE76E50C7ULL, "gen00fl_varric6076062.cl"},
	{0x0AAAA60E5D307F18ULL, "6155733_m.fxe"},
	{0x0AAAD77842A852D7ULL, "6176005_m.fxe"},
	{0x0AAB18EDF5149EA1ULL, "vxlt_darktown_57k_0_0.vlm"},
	{0x0AAB4F9F2AFFF342ULL, "176277273.ogg"},
	{0x0AAB76819D40E9AEULL, "qun220cr_javaris6071521.cl"},
	{0x0AABA83DAD1C561BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv1.ani"},
	{0x0AABC13334E12D66ULL, "113061345.ogg"},
	{0x0AAC5FDCB00E7200ULL, "6110172_m.fxe"},
	{0x0AAC8983169B837BULL, "zz_dae_debug6047614.cl"},
	{0x0AACB7120A462F09ULL, "6080510_f.fxe"},
	{0x0AAD65F674FFEFF2ULL, "vxdr_0_deeproads_fhz_0_0.vlm"},
	{0x0AAD6798820BA496ULL, "vxht_0_hightown_a2_bzw_0_0.vlm"},
	{0x0AADB82FE7309B14ULL, "lgt311_shw_mer_agatha6156498.cl"},
	{0x0AADEA1C350E9576ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbonepile.mao"},
	{0x0AADF61C350EA915ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbonepile.mmh"},
	{0x0AADF81C350EAC8BULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbonepile.msh"},
	{0x0AAE31FD041A8380ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkl.ani"},
	{0x0AAE571E51C1F34AULL, "gen00fl_dog6107335.cl"},
	{0x0AAEDF840DBFA879ULL, "gen00fl_dog6141138.cl"},
	{0x0AAF62B8DFC0AD47ULL, "6101003_m.fxe"},
	{0x0AAFFA6AF19E355EULL, "vxca_caverns1_mines_fms_0_0.vlm"},
	{0x0AB02F654BC79B70ULL, "hlac_lt_hangedman_n_-1_11z.dds"},
	{0x0AB0576F17D2CD13ULL, "hm_cgn_005.mop"},
	{0x0AB0776678BF6F97ULL, "261470896.ogg"},
	{0x0AB0F1BCE44DAC75ULL, "lit_aoe_05.mao"},
	{0x0AB13017D8556489ULL, "vxw_bonepit_1vp_0_0.vlm"},
	{0x0AB165793854CC29ULL, "pro000ar_blightlands_real.are"},
	{0x0AB1DE780BD106A3ULL, "gen00fl_fenris6126521.cl"},
	{0x0AB1E85FDD6394F9ULL, "ht_0_hightown_45_0.anb"},
	{0x0AB29F4687A9DBB4ULL, "6016641_m.fxe"},
	{0x0AB2C6B5969F1BD4ULL, "hlac_ht_0_hightown_n_-65537_3z.dds"},
	{0x0AB3A0376A03471AULL, "143197887.ogg"},
	{0x0AB3F680022D7272ULL, "340531701.ogg"},
	{0x0AB40AF487838F3AULL, "hf_zmother_h2s1.mor"},
	{0x0AB45F4DD6724CFEULL, "523177612.ogg"},
	{0x0AB4637217C1F77AULL, "6078802_f.fxe"},
	{0x0AB46B300EAB22D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_right_hand.ani"},
	{0x0AB4FC43E239F207ULL, "6100040_m.fxe"},
	{0x0AB5B27EA38A8F40ULL, "63277414.ogg"},
	{0x0AB5C8D86101F0E4ULL, "30078549.ogg"},
	{0x0AB5E5FFCEDF17B6ULL, "445758111.ogg"},
	{0x0AB67170C1B1BB10ULL, "vxw_sundermount_2ja_0_0.vlm"},
	{0x0AB69E3A42A39193ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_reaction.evt"},
	{0x0AB6B570352F7394ULL, "vxht_mansion_n_4w7_0_0.vlm"},
	{0x0AB6F7E9AE1DBC98ULL, "330492809.ogg"},
	{0x0AB7B2DA7864241CULL, "vxlt_0_lowtown_docks_a3_n_cz0_0_1.vlm"},
	{0x0AB7C390CED1944BULL, "lgt250_way_dalish_conflict6147152.cl"},
	{0x0AB7DCB2A16A353AULL, "green.tnt"},
	{0x0AB826A9756E4D0CULL, "6127039_m.fxe"},
	{0x0AB8B03DAD26EE12ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv1.evt"},
	{0x0AB8B5DE0A54725BULL, "dre000_cut_thewayout.bnk"},
	{0x0AB92EF9149DFDE6ULL, "gen00fl_isabela6075248.cl"},
	{0x0AB9765F26B31B55ULL, "vxht_bartrand_n_5sx_0_0.vlm"},
	{0x0AB99440FC9FED48ULL, "hlac_lt_gallowstemplar_f_7198_6x.dds"},
	{0x0AB9B801EDC6CC2BULL, "vxht_0_hightown_a3_apj_0_0.vlm"},
	{0x0AB9BB81AA8C504DULL, "gen00fl_aveline6152529.cl"},
	{0x0AB9DA87F94FCC09ULL, "498140500.ogg"},
	{0x0ABA20189B4C0451ULL, "6170427_m.fxe"},
	{0x0ABA3B6BDEEE6DD8ULL, "vxlt_backalley_n_1k5_0_0.vlm"},
	{0x0ABA44B60504E1EBULL, "vxw_bonepitdestroyed_a3_by_0_0.vlm"},
	{0x0ABA7A905327D668ULL, "character_character_p254.psh.fxo.dx11"},
	{0x0ABAAE8964087708ULL, "plt_dae010pt_amb_lowtown.plo"},
	{0x0ABB4434A010E602ULL, "hlac_ca_caverns1_24425_0x.dds"},
	{0x0ABB89222EB3F810ULL, "one300_zevran6076551.cl"},
	{0x0ABC647C4AD01E5BULL, "hlac_ht_keep_3732_16z.dds"},
	{0x0ABC973A42A7F1D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_reaction.gad"},
	{0x0ABCB7DE0A585E1BULL, "dre000_cut_thewayout.cnv"},
	{0x0ABCFC7541D48E2FULL, "6045982_f.fxe"},
	{0x0ABD079B604FFA56ULL, "vxht_chantry_n_2rx_0_0.vlm"},
	{0x0ABDE29C98DB30E4ULL, "dae000_quintus6154725.cl"},
	{0x0ABE4851B28FAB98ULL, "dae000ip_crowd_mf8_talk15.utp"},
	{0x0ABE60B6397431DBULL, "vxlt_0_lowtown_a3_n_c_e8g_0_0.vlm"},
	{0x0ABE6D3F92E9AD89ULL, "vxca_caverns2_4x7_0_0.vlm"},
	{0x0ABEA77F47D9665BULL, "993190787.ogg"},
	{0x0ABECDEF26D0FD36ULL, "hlac_lt_0_lowtown_docks_65535_1y.dds"},
	{0x0ABEF5F002216603ULL, "follower_banter6170642.cl"},
	{0x0ABF1EF7D7921B93ULL, "mag101_aombanter.bnk"},
	{0x0ABF6AEA00C69698ULL, "gen00fl_dog6109285.cl"},
	{0x0ABF78B6FECAD391ULL, "lt_0_lowtown_a2_n_53_0.anb"},
	{0x0ABFAB8E73ABD806ULL, "vxlt_undercitydungeon_1ip_0_1.vlm"},
	{0x0ABFCF3DAD2D41A5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv1.gad"},
	{0x0ABFE0BE885D2335ULL, "hlac_lt_0_lowtown_docks_a3_n_0_1x.dds"},
	{0x0AC02ECFAA3BE83BULL, "lgt310_shw_ors_informant6156340.cl"},
	{0x0AC135EEF63FB660ULL, "hlac_w_woundedcoast_-131065_0x.dds"},
	{0x0AC17EE9639C969AULL, "vxca_cavern1_haunted_gif_0_0.vlm"},
	{0x0AC240F7D7948BE3ULL, "mag101_aombanter.cnv"},
	{0x0AC25CFA98A30F41ULL, "dre000_dougal6130957.cl"},
	{0x0AC265DF122F9C5DULL, "subareaid.gda"},
	{0x0AC267E1E4E98DA1ULL, "mag291_demon_marethari6028848.cl"},
	{0x0AC2ACE4907808EAULL, "836032149.ogg"},
	{0x0AC2FACADAC66287ULL, "qun221_viscount6074078.cl"},
	{0x0AC39AD0AB62FA54ULL, "761268501.ogg"},
	{0x0AC3A63A42AE29B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_reaction.ani"},
	{0x0AC424E010157311ULL, "6136230_m.fxe"},
	{0x0AC47C8EF2C9EF82ULL, "6019116_m.fxe"},
	{0x0AC4D584A5E997D4ULL, "hlac_ht_0_hightown_131071_1y.dds"},
	{0x0AC4E34DB3A7D390ULL, "vxw_sundermount_2on_0_0.vlm"},
	{0x0AC5A5DB6AF2A857ULL, "follower_barks6153729.cl"},
	{0x0AC5E1B58A761649ULL, "6119177_f.fxe"},
	{0x0AC61E027EF8C518ULL, "w_woundedcoast_35_0.rml"},
	{0x0AC66397DA707B71ULL, "vxdr_primevaltomb_8mc_0_1.vlm"},
	{0x0AC67E4A15C9D3CEULL, "766476004.ogg"},
	{0x0AC6A2A1630C1241ULL, "llac_lt_warehouse_-65535_1.dds"},
	{0x0AC6B18B56F0BDCBULL, "vxht_chantry_n_20k_0_0.vlm"},
	{0x0AC6C95D7E0397A1ULL, "6099542_m.fxe"},
	{0x0AC6EF9CA4BC1838ULL, "art\\vfx\\materials\\vfx_d_shockwave.mao"},
	{0x0AC7958E395814F4ULL, "929772460.ogg"},
	{0x0AC838245A019377ULL, "6095148_m.fxe"},
	{0x0AC8627C1B1698F9ULL, "w_encounter2_49_0.msh"},
	{0x0AC86C7C1B16A9E7ULL, "w_encounter2_49_0.mmh"},
	{0x0AC86D2F38731C79ULL, "narrative_flemeth6002799.cl"},
	{0x0AC8707C1B16B034ULL, "w_encounter2_49_0.mao"},
	{0x0AC89A26990A1995ULL, "6154506_m.fxe"},
	{0x0AC8AA34CCE18732ULL, "6071257_m.fxe"},
	{0x0AC99AECF32E6627ULL, "6119900_f.fxe"},
	{0x0AC9B472355D6970ULL, "hlac_lt_0_lowtown_n_-65535_5x.dds"},
	{0x0AC9CD23B8634BC6ULL, "206556495.ogg"},
	{0x0AC9D2A33D737E41ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.dura.ani"},
	{0x0ACA05E2688617B7ULL, "6121104_m.fxe"},
	{0x0ACA4F003E265F10ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05b.mmh"},
	{0x0ACA5D003E26772AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05b.msh"},
	{0x0ACAC09510E25B9CULL, "6036988_m.fxe"},
	{0x0ACBC222D7FE499AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic.mmh"},
	{0x0ACBC622D7FE50D1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic.mao"},
	{0x0ACBD422D7FE68E0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic.msh"},
	{0x0ACC51484097D656ULL, "vxca_cavern1_haunted_edk_0_0.vlm"},
	{0x0ACC6D38280B559BULL, "6137283_f.fxe"},
	{0x0ACC6F41FF647802ULL, "vxca_caverns2prc_7tw_0_1.vlm"},
	{0x0ACDA230390FE16FULL, "vxlt_0_lowtown_a2_n_e84_0_0.vlm"},
	{0x0ACDB33EDA5BE8EBULL, "6136957_m.fxe"},
	{0x0ACE1F1F689F86CAULL, "hlac_ca_caverns2_5802_53y.dds"},
	{0x0ACE5FA25F21D640ULL, "vxlt_warehouse_19h_0_1.vlm"},
	{0x0ACE9296B3BE6DE8ULL, "art\\vfx\\combat\\vfx_sg_quake_act_c.mmh"},
	{0x0ACE9D457254A2ABULL, "6076150_m.fxe"},
	{0x0ACE9EE52D1BF6F1ULL, "vxlt_0_lowtown_a2_n_c_hxa_0_0.vlm"},
	{0x0ACFB12FB5C232A3ULL, "lgt210_eau_symp_templar6154717.cl"},
	{0x0AD01F92D5B7B563ULL, "qun221_petrice6076156.cl"},
	{0x0AD0285959A6E31EULL, "and121_karl6093471.cl"},
	{0x0AD036B803F36F1BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mbn.mmh"},
	{0x0AD042B803F383B8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mbn.mao"},
	{0x0AD049F1A7068D18ULL, "439181655.ogg"},
	{0x0AD054B803F3A205ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mbn.msh"},
	{0x0AD0DDA33D79AFE3ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.dura.gad"},
	{0x0AD1231ADF95BBF8ULL, "vxca_caverns1_mines_i8m_0_0.vlm"},
	{0x0AD16AB3AADAC09FULL, "6133806_m.fxe"},
	{0x0AD173459DF8BCFCULL, "6093843_f.fxe"},
	{0x0AD19DC6F12D4358ULL, "llac_lt_undercityserialkiller_-1_33.dds"},
	{0x0AD25C241B9FABBFULL, "543869611.ogg"},
	{0x0AD2A02125811166ULL, "lgt220_shp_mag_guard6175909.cl"},
	{0x0AD3552AA59DEC10ULL, "vxw_sundermount_2po_0_0.vlm"},
	{0x0AD3650FA862AE3DULL, "6153381_m.fxe"},
	{0x0AD373D031345B24ULL, "vxdr_primevalentrance_28f_0_0.vlm"},
	{0x0AD3EA5DDAA7C5A3ULL, "hlac_w_encounter1_0_3y.dds"},
	{0x0AD42D483644803FULL, "889293996.ogg"},
	{0x0AD4EE9CB756B8EFULL, "vxca_caverns1_mines_j5j_0_0.vlm"},
	{0x0AD4FE0F2EBF3B62ULL, "llac_lt_undercitydungeon_-1_43.dds"},
	{0x0AD50B09920CA577ULL, "vxdr_primevaltomb_7km_0_0.vlm"},
	{0x0AD51570D64BB1BCULL, "vxlt_0_lowtown_a2_n_c_i0q_0_0.vlm"},
	{0x0AD5208805A2A830ULL, "491236661.ogg"},
	{0x0AD5CC4DA474BD29ULL, "vxlt_0_lowtown_a2_n_c_eoy_0_0.vlm"},
	{0x0AD61BD0B482383AULL, "6201959_m.fxe"},
	{0x0AD628AD71E43DABULL, "6133312_m.fxe"},
	{0x0AD6D23A68207593ULL, "mrl150_merrill6093770.cl"},
	{0x0AD756793875827FULL, "pro000ar_blightlands_real.ncs"},
	{0x0AD79672A7A5715CULL, "6083953_m.fxe"},
	{0x0AD7A82D1655C36AULL, "vxca_caverns1_e8n_0_0.vlm"},
	{0x0AD7A940F5A37EA1ULL, "6103308_m.fxe"},
	{0x0AD7BAA33D7F92C8ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.dura.evt"},
	{0x0AD7E6FF4B52C6AEULL, "6071089_f.fxe"},
	{0x0AD80EFA059867F6ULL, "gen00fl_aveline6132809.cl"},
	{0x0AD818BE5575C634ULL, "vxca_caverns1_ei6_0_1.vlm"},
	{0x0AD81EF91131790FULL, "6055610_m.fxe"},
	{0x0AD83D1472C2E664ULL, "vxca_caverns2_aid_0_0.vlm"},
	{0x0AD848957F7B5269ULL, "zz_vault_debug6203278.cl"},
	{0x0AD84A1A40261B26ULL, "static_static_p72.psh.fxo.dx11"},
	{0x0AD858FA63D17074ULL, "vxht_mansion2b_1bv_0_0.vlm"},
	{0x0AD87CF9EE1CA143ULL, "6072248_m.fxe"},
	{0x0AD8B3B829609F41ULL, "hlac_w_blightlands_262141_0z.dds"},
	{0x0AD915C22330FE6CULL, "6152075_m.fxe"},
	{0x0AD91DB72855C5BCULL, "face_face_p3.vsh.fxo.dx11"},
	{0x0AD97A1D020A1C4CULL, "6116080_m.fxe"},
	{0x0ADA3A4671749FB6ULL, "hlac_w_encounter1_1_0z.dds"},
	{0x0ADA6F9BBEC36432ULL, "114320933.ogg"},
	{0x0ADAABC38F83955EULL, "6043876_m.fxe"},
	{0x0ADAD69CDF40010AULL, "450163166.ogg"},
	{0x0ADB1D93D89DED7EULL, "hlac_lt_gallowsprison_n_-1_3x.dds"},
	{0x0ADB4A44AC8988BBULL, "6053857_f.fxe"},
	{0x0ADBCC8184CAEA26ULL, "vxw_encounter2_camp_jp_0_0.vlm"},
	{0x0ADC191A19F156F1ULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_exit.ani"},
	{0x0ADC2318101A510CULL, "hlac_w_sundermount_-65542_0x.dds"},
	{0x0ADC33E51A6905ACULL, "llac_w_encounter2_-65536_0.dds"},
	{0x0ADCAB89642558C7ULL, "plt_dae010pt_amb_lowtown.nss"},
	{0x0ADDAE71D9DBC142ULL, "vxw_sundermount_2kp_0_0.vlm"},
	{0x0ADE7679387BD76EULL, "pro000ar_blightlands_real.lst"},
	{0x0ADEE51E817921BAULL, "art\\characters\\weapons\\staff\\w_stf_dssta.msh"},
	{0x0ADEF31E817939F3ULL, "art\\characters\\weapons\\staff\\w_stf_dssta.mao"},
	{0x0ADEF71E81794000ULL, "art\\characters\\weapons\\staff\\w_stf_dssta.mmh"},
	{0x0ADEFB16E3BF4F74ULL, "6148388_m.fxe"},
	{0x0ADF29F3010C1F23ULL, "752287635.ogg"},
	{0x0AE05D7560E8F458ULL, "llac_lt_0_lowtown_docks_a3_-65538_3.dds"},
	{0x0AE060D49BA9A0A6ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_002a_d.dds"},
	{0x0AE07B43388229A3ULL, "vxht_keep_n_63z_0_0.vlm"},
	{0x0AE07D66B4CF500EULL, "dre000_varric6027749.cl"},
	{0x0AE0F3198B4D3C22ULL, "vxlt_0_lowtown_a3_n_c_ev4_0_0.vlm"},
	{0x0AE11E3ABDF1611FULL, "w_bonepitmining_a2_212_0.msh"},
	{0x0AE1243ABDF16BC1ULL, "w_bonepitmining_a2_212_0.mmh"},
	{0x0AE1303ABDF18062ULL, "w_bonepitmining_a2_212_0.mao"},
	{0x0AE17C09C789F287ULL, "6166574_m.fxe"},
	{0x0AE1B3973BEE3A50ULL, "rdr111dg_martin6015059.cl"},
	{0x0AE1D6EC9A58C6BDULL, "6156081_m.fxe"},
	{0x0AE1E89B60F6A441ULL, "vxdr_primevalentrance_319_0_0.vlm"},
	{0x0AE26CAB7A6598B7ULL, "fen110_fenris6100322.cl"},
	{0x0AE2F7C5B5556807ULL, "cd_nat380_sibling.wbd"},
	{0x0AE3241A19F78793ULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_exit.gad"},
	{0x0AE329B9E3F7668AULL, "vxw_encounter2_4z_0_0.vlm"},
	{0x0AE354CF3EEAE7ECULL, "6088325_m.fxe"},
	{0x0AE384AF2B386511ULL, "vxlt_gallowstemplar_f_5pm_0_0.vlm"},
	{0x0AE385F2120756AAULL, "boulder_mac_so.mao"},
	{0x0AE3AD1761EFC8A1ULL, "6093742_m.fxe"},
	{0x0AE51C027F1372F5ULL, "w_woundedcoast_35_0.msh"},
	{0x0AE51E027F13766BULL, "w_woundedcoast_35_0.mmh"},
	{0x0AE52A027F138A08ULL, "w_woundedcoast_35_0.mao"},
	{0x0AE547D96746CBE3ULL, "llac_ca_caverns1_mines_24425_11.dds"},
	{0x0AE59BA82D179B69ULL, "vxw_sundermount_2co_0_0.vlm"},
	{0x0AE5A65F536E96A9ULL, "144074700.ogg"},
	{0x0AE60F8CCD3576B8ULL, "llac_lt_0_lowtown_n_-3_4.dds"},
	{0x0AE61B1DE7999721ULL, "6104208_m.fxe"},
	{0x0AE63794EBBA7F0CULL, "hla_ht_0_hightown_a2_65535_0x.dds"},
	{0x0AE6452716649536ULL, "zz_qun_debug6028227.cl"},
	{0x0AE6AE2E418062FEULL, "967360694.ogg"},
	{0x0AE6E37824A37390ULL, "one130_thug6133091.cl"},
	{0x0AE7679FE8D6CE8FULL, "l_env_wal_lowtowndocks_cough.wbd"},
	{0x0AE958E9299F99E6ULL, "black.psh.fxo"},
	{0x0AE975C6E479ED4DULL, "lt_backalley_g_mmap.dds"},
	{0x0AE98E8BEC22E5FCULL, "w_sundermount_134_0.mao"},
	{0x0AE99A8BEC22F99FULL, "w_sundermount_134_0.mmh"},
	{0x0AE9A08BEC230441ULL, "w_sundermount_134_0.msh"},
	{0x0AEA211A19FDA1F8ULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_exit.evt"},
	{0x0AEA3E5E76C102A6ULL, "vxw_encounter2_camp_eq_0_0.vlm"},
	{0x0AEA602982B5645FULL, "car131ip_chest_gift.utp"},
	{0x0AEA8A0841BFF841ULL, "300651420.ogg"},
	{0x0AEACB221FAC443EULL, "713029642.ogg"},
	{0x0AEAD22AF678839FULL, "91184444.ogg"},
	{0x0AEAE95FDD935BF9ULL, "ht_0_hightown_45_0.rml"},
	{0x0AEAF5AEAA23D342ULL, "gen00fl_anders6119702.cl"},
	{0x0AEB15AE9891A2F0ULL, "6141112_m.fxe"},
	{0x0AEB1F5B6DA1684AULL, "179820971.ogg"},
	{0x0AEB571D9743EAE6ULL, "6129426_m.fxe"},
	{0x0AEB790D780BCE4BULL, "llac_lt_0_lowtown_docks_a3_n_65534_6.dds"},
	{0x0AEB9DE8B1FD5409ULL, "6175147_m.fxe"},
	{0x0AEBBF5D1C29DE04ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\blendtree\\doorsimple.anb"},
	{0x0AEC204ECBD1D9C2ULL, "vxw_sundermount_2lq_0_0.vlm"},
	{0x0AEC70AACFD6DEACULL, "306938669.ogg"},
	{0x0AED861443E3B232ULL, "vxht_brothel_n_4a8_0_0.vlm"},
	{0x0AEE06D4C381ECF8ULL, "vxca_caverns1_mines_i2k_0_0.vlm"},
	{0x0AEE3E9E27C92F78ULL, "mcr371_hostile_fenris6082231.cl"},
	{0x0AEF545AD2C0999FULL, "6085695_m.fxe"},
	{0x0AEFABB049838BADULL, "6137596_m.fxe"},
	{0x0AF050137EA0194BULL, "healertacticpreset.gda"},
	{0x0AF0ADB6310AA461ULL, "953697932.ogg"},
	{0x0AF0E122D81D9AA5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic.phy"},
	{0x0AF0FF021E6A6264ULL, "90634769.ogg"},
	{0x0AF1068AB1EC16C6ULL, "hlac_lt_0_lowtown_a2_n_65533_3y.dds"},
	{0x0AF13B855CAA733FULL, "vxht_mansion_n_60s_0_0.vlm"},
	{0x0AF28A2DE039D8B1ULL, "dre180ip_to_primeval_crypt.utp"},
	{0x0AF28C8BDC5560B8ULL, "vxca_caverns2prc_9pn_0_0.vlm"},
	{0x0AF3353478440024ULL, "vxht_0_hightown_n_a7j_0_0.vlm"},
	{0x0AF351684C0AB7AEULL, "vxlt_warehouse_b7_0_0.vlm"},
	{0x0AF374CAF9E18035ULL, "vxw_bonepit_7x_0_0.vlm"},
	{0x0AF40D851F0DB3E9ULL, "vxw_sundermount_2dn_0_0.vlm"},
	{0x0AF41978DE8DCBAAULL, "dae200_saemus6142455.cl"},
	{0x0AF42C7BE317F453ULL, "mcr310_sibling6076241.cl"},
	{0x0AF42F517A8BFAA2ULL, "vxw_bonepitdestroyed_a3_1no_0_0.vlm"},
	{0x0AF4557C1B3BBD1CULL, "w_encounter2_49_0.phy"},
	{0x0AF53CB447BDA174ULL, "vxht_mansion2b_1xt_0_0.vlm"},
	{0x0AF5C1133DF52CFEULL, "vxca_caverns2prc_9x0_0_0.vlm"},
	{0x0AF5E57D946455F0ULL, "lt_0_lowtown_docks_250_0.mao"},
	{0x0AF5E97D94645C03ULL, "lt_0_lowtown_docks_250_0.mmh"},
	{0x0AF5F77D9464745DULL, "lt_0_lowtown_docks_250_0.msh"},
	{0x0AF6153ABB99110BULL, "vxlt_warehouse_n_qf_0_0.vlm"},
	{0x0AF707D3FE8F54F0ULL, "hlac_lt_darktown_-1_0z.dds"},
	{0x0AF74600226D8177ULL, "dae000ip_crowd_mf2_talk5.utp"},
	{0x0AF77A876C6D2694ULL, "zz_followers_banter6158148.cl"},
	{0x0AF8CF8486C6B368ULL, "vxlt_0_lowtown_docks_a3_n_c14_0_0.vlm"},
	{0x0AF8DD84E9AE9565ULL, "234972057.ogg"},
	{0x0AF91C325CB8E778ULL, "vxht_mansion2b_3bx_0_0.vlm"},
	{0x0AF92817C6154C1FULL, "mer220_hubert6041210.cl"},
	{0x0AFA922BBDC7F242ULL, "vxw_sundermount_2qr_0_0.vlm"},
	{0x0AFAC190409164E6ULL, "6032897_f.fxe"},
	{0x0AFAD9CFB128C165ULL, "6098083_f.fxe"},
	{0x0AFB1A1BE5AD14D8ULL, "gen_im_acc_rng_act3_14.uti"},
	{0x0AFB6C7C1B420344ULL, "w_encounter2_49_0.rml"},
	{0x0AFBC088D2CEEC80ULL, "ht_0_hightown_n_49_0.rml"},
	{0x0AFBF7F135D9CAB2ULL, "vxht_brothel_n_4f9_0_0.vlm"},
	{0x0AFCA6CD4B502352ULL, "vxlt_darktown_5gt_0_0.vlm"},
	{0x0AFCB63CAC47C4F4ULL, "ss_probabilitymodif.gda"},
	{0x0AFCCE0095F27DA7ULL, "mcr370_meredith6083582.cl"},
	{0x0AFCD1C5F63497F4ULL, "vxlt_0_lowtown_a2_n_gvc_0_0.vlm"},
	{0x0AFD1C77B3986D49ULL, "ave100_donnic6058924.cl"},
	{0x0AFD74E575D00812ULL, "gen00fl_sebastian6177817.cl"},
	{0x0AFD85840424CD94ULL, "65534320.ogg"},
	{0x0AFE20C1DB8A83B3ULL, "plt_lgt260pt_way_blood_mage.nss"},
	{0x0AFE7D13D1476D10ULL, "181444053.ogg"},
	{0x0AFEBB47E1307552ULL, "60507516.ogg"},
	{0x0AFEEC91FD56EC59ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_back_loop.gad"},
	{0x0AFF72DF3174FBACULL, "6156242_m.fxe"},
	{0x0AFFAB844108F181ULL, "follower_banter6155472.cl"},
	{0x0AFFC4E14DDC0A01ULL, "329491201.ogg"},
	{0x0AFFC913262182ABULL, "hlac_lt_gallowstemplar_n_2395_1z.dds"},
	{0x0B006EEA02C4F71BULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_possessionsw.mmh"},
	{0x0B007AEA02C50BB8ULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_possessionsw.mao"},
	{0x0B008CDD2F597E79ULL, "dae300_vashothlead6062785.cl"},
	{0x0B008CEA02C52A05ULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_possessionsw.msh"},
	{0x0B01E3EB93B346D7ULL, "pro000_cut2d_real_story6001241.cl"},
	{0x0B022B7C442FD6C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_1.ani"},
	{0x0B027F621103CC69ULL, "vxw_sundermount_2im_0_0.vlm"},
	{0x0B02F5A88FB6D4B6ULL, "242270726.ogg"},
	{0x0B034780D4FF1DB2ULL, "39951816.ogg"},
	{0x0B0399CE5FDC662EULL, "hlac_lt_0_lowtown_docks_n_-131071_0y.dds"},
	{0x0B03AE9139B3B9F4ULL, "vxht_mansion2b_1su_0_0.vlm"},
	{0x0B0402A026F4EA19ULL, "art\\characters\\playercharacter\\humanmale\\hlm_tmpb.msh"},
	{0x0B040CA026F4FB07ULL, "art\\characters\\playercharacter\\humanmale\\hlm_tmpb.mmh"},
	{0x0B0432F02FEB457EULL, "vxca_caverns2prc_9g1_0_0.vlm"},
	{0x0B045042A170F839ULL, "vxlt_hangedman_n_2vu_0_0.vlm"},
	{0x0B048717AD8F298BULL, "vxlt_warehouse_n_ty_0_0.vlm"},
	{0x0B04A1D3B0498A1CULL, "cd_qcr221_cut_heros_reward.wbd"},
	{0x0B0548C995F816FBULL, "ttx_window_small_b_1.msh"},
	{0x0B0566C995F849E5ULL, "ttx_window_small_b_1.mmh"},
	{0x0B0599D7955C380DULL, "dae000_male_templar_mob6174369.cl"},
	{0x0B05FD91FD5D27A7ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_back_loop.ani"},
	{0x0B0609F1D74B1030ULL, "vxlt_warehouse_n_8b_0_0.vlm"},
	{0x0B0628A174ACA1EEULL, "hlac_lt_undercitydungeon_-1_7z.dds"},
	{0x0B0661DBF270CC63ULL, "dae021_sister_lorena_sta6142647.cl"},
	{0x0B067A1AA10FA696ULL, "6109217_m.fxe"},
	{0x0B070F404705C5F7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01b.phy"},
	{0x0B0778FA0CD0B521ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_wkf.ani"},
	{0x0B0817998443C3ADULL, "vxht_brothel_n_5fo_0_0.vlm"},
	{0x0B08357D01CBB530ULL, "llac_lt_0_lowtown_a2_-65536_7.dds"},
	{0x0B08962647ED0CBBULL, "vxlt_0_lowtown_a3_n_c_gnc_0_0.vlm"},
	{0x0B090408AFBE0AC2ULL, "vxw_sundermount_2zs_0_0.vlm"},
	{0x0B09287C4435F026ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_1.gad"},
	{0x0B094AA3086710A4ULL, "6156545_m.fxe"},
	{0x0B09E76A76764EECULL, "gen_im_trs_crc_sta_nat_r1.uti"},
	{0x0B09F11BA60EDA56ULL, "fx_c_fresnel_add.mao"},
	{0x0B0A61BD8724AB6FULL, "6172988_m.fxe"},
	{0x0B0A73C97584CD47ULL, "773478355.ogg"},
	{0x0B0A9A6F2FE0E188ULL, "mag360_samson6042051.cl"},
	{0x0B0B064D2A503CE0ULL, "vxlt_house_dx_0_0.vlm"},
	{0x0B0B08346A118B90ULL, "hlac_w_woundedcoast_327684_0y.dds"},
	{0x0B0B165D60AFC1F8ULL, "lt_hangedman_3567_ll.mmh"},
	{0x0B0B30365FDD9282ULL, "hlac_w_encounter1_-65535_0z.dds"},
	{0x0B0B56003E5E022BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05b.phy"},
	{0x0B0BEE0EBC71F239ULL, "6155523_m.fxe"},
	{0x0B0BFC8436061E2BULL, "hlac_ht_brothel_n_3364_1y.dds"},
	{0x0B0BFCD1B25ABFE8ULL, "6066182_f.fxe"},
	{0x0B0C5ABC661A9DC3ULL, "one300_nuncio6073805.cl"},
	{0x0B0CCFD6F133D689ULL, "vxht_0_hightown_a2_ca0_0_0.vlm"},
	{0x0B0D4186FFE7462CULL, "567238867.ogg"},
	{0x0B0DC13E9ADED75AULL, "mer230ip_to_arlan_hideout.utp"},
	{0x0B0DDE05AC6FD153ULL, "513741480.ogg"},
	{0x0B0E6E18B60A5024ULL, "llac_w_bonepitmining_a2_65542_0.dds"},
	{0x0B0E83FA0CD6E6C3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_wkf.gad"},
	{0x0B0ED2ACE41DE48AULL, "vxw_blightlands_ib_0_0.vlm"},
	{0x0B0F2AA262DA1278ULL, "zz_vault_debug6203354.cl"},
	{0x0B0F7045C04191B8ULL, "vxca_caverns2prc_9zl_0_0.vlm"},
	{0x0B0FAE44386DCC6EULL, "6087209_m.fxe"},
	{0x0B10237C443C0605ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_1.evt"},
	{0x0B10F097653C1570ULL, "gencr_apostate.utc"},
	{0x0B10F13F02F9E4E9ULL, "vxw_sundermount_2rl_0_0.vlm"},
	{0x0B11898D20AD290AULL, "vxlt_0_lowtown_a3_n_c_by2_0_0.vlm"},
	{0x0B11B3F8D08EB0AAULL, "vxca_caverns1_int_0_0.vlm"},
	{0x0B11E84023AE4D32ULL, "6110365_m.fxe"},
	{0x0B12206E2BA9D274ULL, "vxht_mansion2b_1vz_0_0.vlm"},
	{0x0B13C2AA91EE998CULL, "6035938_m.fxe"},
	{0x0B13E93E1675CB35ULL, "141351815.ogg"},
	{0x0B13F420BE74CA9EULL, "6148702_m.fxe"},
	{0x0B1475A74AA0E022ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st02a_0s.dds"},
	{0x0B14BCAF47D715A8ULL, "dae000_keep_servant_1_sta6156894.cl"},
	{0x0B14DB9B3460453AULL, "52458453.ogg"},
	{0x0B15AB3C4899EA2EULL, "6154765_m.fxe"},
	{0x0B15EEE969DE4E09ULL, "405398035.ogg"},
	{0x0B15EF5FF4F5D3EFULL, "dae000ip_crowd_ff1_talk1.utp"},
	{0x0B15F103E59D3135ULL, "590412145.ogg"},
	{0x0B1647F62F5F9BE2ULL, "6105706_f.fxe"},
	{0x0B1651577478175FULL, "hlac_lt_shop_-1_4y.dds"},
	{0x0B1684585771D76EULL, "hlac_lt_0_lowtown_a3_n_c_-65537_1y.dds"},
	{0x0B1689767639DC2DULL, "vxht_brothel_n_5in_0_0.vlm"},
	{0x0B16B44326F0EB25ULL, "qun221_petrice6083820.cl"},
	{0x0B16BA94E2C7AB9BULL, "ave331_post_fight.bnk"},
	{0x0B16FF9A9AECF7A1ULL, "292491156.ogg"},
	{0x0B17284D082E9DC9ULL, "313331627.ogg"},
	{0x0B174E1EB27E6D0EULL, "941860372.ogg"},
	{0x0B17653CB78280AEULL, "6098831_f.fxe"},
	{0x0B177102163B25A2ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_2z.dds"},
	{0x0B17F8FC3F77C48BULL, "hlac_ht_mansion2b_818_37y.dds"},
	{0x0B183918F7494C6FULL, "llac_lt_0_lowtown_a3_n_c_65534_3.dds"},
	{0x0B1842485541B320ULL, "vxlt_hangedman_n_24z_0_0.vlm"},
	{0x0B1852840F4866BBULL, "hlac_lt_0_lowtown_docks_a3_131071_2y.dds"},
	{0x0B19BC94E2C9E55BULL, "ave331_post_fight.cnv"},
	{0x0B1A1DA5374F2F98ULL, "vxdr_primevaltomb_9d9_0_0.vlm"},
	{0x0B1A503DCBA187F5ULL, "6025981_m.fxe"},
	{0x0B1B2ED6605BCD7EULL, "298212969.ogg"},
	{0x0B1B357F4D43E9D9ULL, "vxht_mansion_n_6hm_0_0.vlm"},
	{0x0B1B5FF56A6869FDULL, "llac_ca_caverns2prc_5802_38.dds"},
	{0x0B1B6584A148CF8DULL, "character_character_p129.psh.fxo.dx11.assembly"},
	{0x0B1C10E336AA9890ULL, "hlac_lt_0_lowtown_n_-2_3x.dds"},
	{0x0B1C5E663FE99E85ULL, "vxca_caverns1_mines_i2r_0_0.vlm"},
	{0x0B1C81AB8E34915BULL, "106967439.ogg"},
	{0x0B1CF5EDDA319034ULL, "lgt000cr_epic_merchant_3.utc"},
	{0x0B1D4489D613FD0AULL, "vxw_blightlands_rc_0_0.vlm"},
	{0x0B1D6B0B101BFA1BULL, "lt_0_lowtown_docks_a3_226_0.mmh"},
	{0x0B1D7408A1FD1162ULL, "789051745.ogg"},
	{0x0B1D770B101C0EB8ULL, "lt_0_lowtown_docks_a3_226_0.mao"},
	{0x0B1D890B101C2D05ULL, "lt_0_lowtown_docks_a3_226_0.msh"},
	{0x0B1E1C2B29B076FCULL, "vxlt_0_lowtown_a2_n_c_ekl_0_0.vlm"},
	{0x0B1E5ECF219450C2ULL, "457476779.ogg"},
	{0x0B1E9E3CC0720026ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.impa.evt"},
	{0x0B1EBBE715064C47ULL, "gen00fl_aveline6120768.cl"},
	{0x0B1ECA682A7BFADDULL, "qcr220_grey_wardens6060406.cl"},
	{0x0B1FEFCB93808482ULL, "w_bonepitdestroyed_a3_3_0_ll.mmh"},
	{0x0B200BC91E3900E3ULL, "vxlt_darktown_6lg_0_0.vlm"},
	{0x0B2026750322700AULL, "6155490_m.fxe"},
	{0x0B2055AB109D4EE7ULL, "vxlt_gallowstemplar_f_55p_0_0.vlm"},
	{0x0B2090B19CF67F42ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_cth_comx_0n.dds"},
	{0x0B210B8DC2B026A8ULL, "hlac_lt_warehouse_0_3x.dds"},
	{0x0B215E9598DAAC2EULL, "lla_ht_0_hightown_a2_-131073_0.dds"},
	{0x0B217DD998C266D6ULL, "vxlt_0_lowtown_a2_n_c_h1u_0_0.vlm"},
	{0x0B21CA5365CCF286ULL, "vxht_keep_5ay_0_0.vlm"},
	{0x0B21CDAAD5898E0AULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzsteps_0.pwk"},
	{0x0B21FB8D161712A4ULL, "6087575_m.fxe"},
	{0x0B2264329564E711ULL, "847041324.ogg"},
	{0x0B22887A14CA947BULL, "6076280_m.fxe"},
	{0x0B22A2C4E888ECBBULL, "zz_and_debug6178599.cl"},
	{0x0B22ABBBD90E1FE3ULL, "vxca_cavern1_haunted_ghl_0_0.vlm"},
	{0x0B22C61956085383ULL, "qcr000_arishok6059320.cl"},
	{0x0B2302D125EBDD3AULL, "vxlt_0_lowtown_a2_n_c_hqy_0_0.vlm"},
	{0x0B2320E4FCC7C2A3ULL, "gen00fl_aveline6122766.cl"},
	{0x0B23A698CA2BB11BULL, "hlac_lt_0_lowtown_131071_1x.dds"},
	{0x0B23DA880DADDE2BULL, "pro000_merchant6051630.cl"},
	{0x0B23F09CBCF360DFULL, "ser221_party_comments6166113.cl"},
	{0x0B2401C5B58D103DULL, "cd_nat380_sibling.bnk"},
	{0x0B241467151D4F12ULL, "314295815.ogg"},
	{0x0B241D68B8CD757DULL, "lgt321pt_shw_ors_cache.ncs"},
	{0x0B245F745E845F77ULL, "mag311_nyssa6041944.cl"},
	{0x0B247076AAB240FAULL, "lgt321pt_alistair_king.ncs"},
	{0x0B2494D8691F3602ULL, "ht_0_hightown_a3_28_0_lc.mmh"},
	{0x0B249D6DF2AC57D4ULL, "296437179.ogg"},
	{0x0B24A0797127190EULL, "338715270.ogg"},
	{0x0B24B53CC0769371ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.impa.gad"},
	{0x0B25632F5B047E16ULL, "vxca_caverns2prc_8j7_0_0.vlm"},
	{0x0B2564B0BF669782ULL, "vxht_bartrand_n_63t_0_0.vlm"},
	{0x0B2579E02BD93DBBULL, "vxlt_0_lowtown_a3_n_c_gxe_0_0.vlm"},
	{0x0B2593E7BF312961ULL, "6054295_m.fxe"},
	{0x0B263E95B0A406A6ULL, "llac_ca_cavern1_haunted_24424_20.dds"},
	{0x0B2695FB3E591F98ULL, "mer141_jansen6070135.cl"},
	{0x0B26DEA25F4E380AULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkf.gad"},
	{0x0B273D4FAF51A21CULL, "zz_prmpt_debug6200963.cl"},
	{0x0B2773C3D9209905ULL, "vxht_mansion2b_3ba_0_0.vlm"},
	{0x0B27819FE90CDF65ULL, "l_env_wal_lowtowndocks_cough.bnk"},
	{0x0B27AE6A7FEED617ULL, "471884814.ogg"},
	{0x0B281D2194FA2E57ULL, "w_bonepitmining_a2_249_0.msh"},
	{0x0B282F2194FA4CBAULL, "w_bonepitmining_a2_249_0.mao"},
	{0x0B28332194FA53C9ULL, "w_bonepitmining_a2_249_0.mmh"},
	{0x0B288DF6E0B88ED6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantrya_sa.dds"},
	{0x0B28E2DB3A42793EULL, "269095643.ogg"},
	{0x0B2A66FCBFB71148ULL, "6105901_m.fxe"},
	{0x0B2A87BEF54EA24FULL, "dae000_mother6136944.cl"},
	{0x0B2AF822DFAF05C4ULL, "387598950.ogg"},
	{0x0B2BA5BD19486621ULL, "reticle.vsh.fxo"},
	{0x0B2BB63CC07CB31FULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.impa.ani"},
	{0x0B2BC71FE41D9178ULL, "vxlt_gallowstemplar_f_39e_0_0.vlm"},
	{0x0B2C22E6B9247E5AULL, "6149693_m.fxe"},
	{0x0B2C53FFA42DC2B8ULL, "vxca_caverns2prc_9lj_0_0.vlm"},
	{0x0B2C9C1BB4ED861DULL, "mrl000_marethari6129239.cl"},
	{0x0B2D405AB427B40BULL, "122372749.ogg"},
	{0x0B2DA71C1BB71714ULL, "vxlt_shop_dp_0_0.vlm"},
	{0x0B2DC1A25F5425A1ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkf.evt"},
	{0x0B2DD4F8E6E615E9ULL, "vxw_sundermount_2xr_0_0.vlm"},
	{0x0B2E57A7EC737B2BULL, "follower_banter6160962.cl"},
	{0x0B2EE4EE7D4BD137ULL, "372929401.ogg"},
	{0x0B2F4CD158C318AFULL, "hf_madam_lusine.mor"},
	{0x0B2F8DA82BFAA699ULL, "vxlt_0_lowtown_docks_a3_btp_0_0.vlm"},
	{0x0B2FF5A0271A0F3CULL, "art\\characters\\playercharacter\\humanmale\\hlm_tmpb.phy"},
	{0x0B3046B589CD9E1FULL, "follower_banter6141581.cl"},
	{0x0B307C92BCE021C6ULL, "vxlt_0_lowtown_a2_eq9_0_0.vlm"},
	{0x0B309FD3B06EC196ULL, "cd_qcr221_cut_heros_reward.bnk"},
	{0x0B311D98CB043863ULL, "vxca_cavern1_haunted_geo_0_0.vlm"},
	{0x0B31C5EF53AAB9CDULL, "6043787_m.fxe"},
	{0x0B31F3C2BE953E55ULL, "6105921_f.fxe"},
	{0x0B32A4DFA4F211D9ULL, "w_bonepitdestroyed_a3_155_0.msh"},
	{0x0B32AEDFA4F222C7ULL, "w_bonepitdestroyed_a3_155_0.mmh"},
	{0x0B32B2DFA4F22914ULL, "w_bonepitdestroyed_a3_155_0.mao"},
	{0x0B33809DC7F4E439ULL, "327523708.ogg"},
	{0x0B33F5AA53B72F8EULL, "190461508.ogg"},
	{0x0B342D1239CB748EULL, "art\\characters\\creatures\\dragon\\c_dragonb_eyes.mao"},
	{0x0B3462B4C05EC7EFULL, "vxlt_warehouse_ac_0_0.vlm"},
	{0x0B34B1C33FB14AECULL, "6000974_m.fxe"},
	{0x0B353C084FF8BE89ULL, "610767424.ogg"},
	{0x0B35658E3C0E8BEAULL, "696512283.ogg"},
	{0x0B3580D59D48882CULL, "hlac_lt_backalley_g_65536_5z.dds"},
	{0x0B35C18E22B3EE20ULL, "hlac_lt_0_lowtown_a2_n_-196609_4z.dds"},
	{0x0B35D357C660013CULL, "hlac_ht_keep_n_3735_0z.dds"},
	{0x0B35F2593E4821A0ULL, "vxlt_backalley_n_207_0_0.vlm"},
	{0x0B35F5EC2B9848AAULL, "vxlt_gallowscourtyard_65c_0_0.vlm"},
	{0x0B364456BAD9E157ULL, "hlac_lt_0_lowtown_-1_2z.dds"},
	{0x0B36530B53675A66ULL, "184579480.ogg"},
	{0x0B37AB6919EB8CF5ULL, "vxlt_0_lowtown_docks_n_cv3_0_0.vlm"},
	{0x0B3842F1DF86F0C1ULL, "dae000cr_grand_cleric6124775.cl"},
	{0x0B384449D04E524BULL, "vxlt_0_lowtown_ex3_0_0.vlm"},
	{0x0B385359CBBFC19BULL, "923173248.ogg"},
	{0x0B387F57E5739CB7ULL, "6126952_m.fxe"},
	{0x0B394DE0AC755600ULL, "dae200cr_brothel_gamlen.utc"},
	{0x0B3953A1DA49DFEAULL, "gen00fl_fenris6115371.cl"},
	{0x0B399FAA1C6905C8ULL, "6062112_m.fxe"},
	{0x0B39C48D3A044DFBULL, "vxht_brothel_n_5gz_0_0.vlm"},
	{0x0B3A35BD500854A4ULL, "dae000_katriela6166395.cl"},
	{0x0B3A6CA9D33D46F7ULL, "vxlt_0_lowtown_a3_n_c_gnw_0_0.vlm"},
	{0x0B3A989601ECCE7FULL, "zz_followers_banter6158484.cl"},
	{0x0B3AB9A25F5EA2F0ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkf.ani"},
	{0x0B3B0FBF25BD5BCAULL, "vxht_brothel_n_3es_0_0.vlm"},
	{0x0B3B4D23BFA4FD68ULL, "6147698_f.fxe"},
	{0x0B3B594F8E8A14A8ULL, "6125257_m.fxe"},
	{0x0B3BCD63F8899F5DULL, "matdef_dca_diffuse.dds"},
	{0x0B3C056E6C45BEF4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsaltar.phy"},
	{0x0B3CF19487F311A4ULL, "w_bonepitdestroyed_a3_44_0.msh"},
	{0x0B3D039487F3300DULL, "w_bonepitdestroyed_a3_44_0.mao"},
	{0x0B3D079487F33746ULL, "w_bonepitdestroyed_a3_44_0.mmh"},
	{0x0B3ECDA3C55695C3ULL, "6112111_m.fxe"},
	{0x0B3F2ED463CABE0DULL, "vxht_bartrand_n_32u_0_0.vlm"},
	{0x0B3F80BC4ACDA8B1ULL, "qun211_party_comment6072256.cl"},
	{0x0B3FEEC42D6E5A1FULL, "377830166.ogg"},
	{0x0B401C8887F38163ULL, "6054827_f.fxe"},
	{0x0B408DC8DCC8F31CULL, "hlac_lt_0_lowtown_a2_n_-2_5x.dds"},
	{0x0B418C3527FFDC8DULL, "hlac_lt_0_lowtown_docks_n_-131074_1z.dds"},
	{0x0B422A65AA294420ULL, "vxht_bartrand_n_4hq_0_0.vlm"},
	{0x0B422BE42D639BB7ULL, "630793870.ogg"},
	{0x0B42486AA352C882ULL, "vxht_bartrand_n_69v_0_0.vlm"},
	{0x0B424D595DBE35ECULL, "gen00fl_isabela6135667.cl"},
	{0x0B42A86D53EEE903ULL, "933621371.ogg"},
	{0x0B42DBF591566D7EULL, "66959214.ogg"},
	{0x0B42E69D8DED7E80ULL, "962974294.ogg"},
	{0x0B433298E6FCB5C9ULL, "6099577_m.fxe"},
	{0x0B4428E41BD6FD2FULL, "vxht_mansion2b_2uk_0_0.vlm"},
	{0x0B451908D032B5B5ULL, "art\\vfx\\models\\vfx_gui_brittle.msh"},
	{0x0B451B08D032B92BULL, "art\\vfx\\models\\vfx_gui_brittle.mmh"},
	{0x0B45352B7601F157ULL, "vxht_mansion2b_21o_0_0.vlm"},
	{0x0B45A7D8019E411FULL, "vxlt_0_lowtown_docks_cyg_0_0.vlm"},
	{0x0B45C0F495695623ULL, "945065132.ogg"},
	{0x0B45E26E6C4DD7F1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsaltar.msh"},
	{0x0B45F06E6C4DEF0CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsaltar.mao"},
	{0x0B45FC6E6C4E03AFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsaltar.mmh"},
	{0x0B461D460BE1A575ULL, "vxlt_0_lowtown_docks_n_cq2_0_0.vlm"},
	{0x0B4638C15B5E9200ULL, "43766156.ogg"},
	{0x0B469D4DB758C03AULL, "vxca_caverns2prc_a9o_0_0.vlm"},
	{0x0B46A2A754145800ULL, "ico_gloves_medium_1.dds"},
	{0x0B46B626C2446ACBULL, "vxlt_0_lowtown_es2_0_0.vlm"},
	{0x0B46F18C390317FAULL, "gen00fl_sebastian6174687.cl"},
	{0x0B47670ECD0F3D14ULL, "gen00fl_anders6196824.cl"},
	{0x0B477A22AE08B6F2ULL, "902663523.ogg"},
	{0x0B478B6E9E03856AULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02bricks.mao"},
	{0x0B478F6E9E038BB9ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02bricks.mmh"},
	{0x0B47996E9E039CA7ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02bricks.msh"},
	{0x0B48069175D37165ULL, "hlac_ht_brothel_n_3381_0y.dds"},
	{0x0B480F1D2CB66E42ULL, "gen00fl_aveline6197228.cl"},
	{0x0B48A01A189FC165ULL, "6065388_f.fxe"},
	{0x0B4912E9D91F2B5FULL, "ave200_aveline_bar6116376.cl"},
	{0x0B4950D4CED38B6EULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.cess.evt"},
	{0x0B4988E6F2A86E67ULL, "vxlt_gallowstemplar_n_5by_0_0.vlm"},
	{0x0B49EADA9FF43F91ULL, "vxlt_gallowscourtyard_n_3x1_0_0.vlm"},
	{0x0B4A038DF9E7687DULL, "hlac_lt_0_lowtown_a3_n_c_131071_1x.dds"},
	{0x0B4A315F974835F4ULL, "pro000_opening6178521.cl"},
	{0x0B4B297FBC853B75ULL, "329159013.ogg"},
	{0x0B4B70F478AF562FULL, "vxht_bartrand_n_653_0_0.vlm"},
	{0x0B4B8010E5396BF0ULL, "ser211ip_note_scribbles_3.utp"},
	{0x0B4BBDBE7391A887ULL, "vxdr_0_deeproads_ctz_0_0.vlm"},
	{0x0B4BE142BCADE5D2ULL, "gen00fl_sebastian6198195.cl"},
	{0x0B4C2BA580DC415DULL, "vxht_mansion2b_23g_0_0.vlm"},
	{0x0B4CA38BEC7786E4ULL, "w_sundermount_134_0.phy"},
	{0x0B4CA57E7888DC35ULL, "6108749_m.fxe"},
	{0x0B4CD5DD19B195EEULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_combat_runf.evt"},
	{0x0B4E07948801AA0DULL, "w_bonepitdestroyed_a3_44_0.rml"},
	{0x0B4E156D8E574FD1ULL, "art\\levels\\4cinematics\\dwarf\\prp_dwarfchair.msh"},
	{0x0B4E2F6D8E577B8FULL, "art\\levels\\4cinematics\\dwarf\\prp_dwarfchair.mmh"},
	{0x0B4E6989B7DEB8D5ULL, "dae000_arishok6067767.cl"},
	{0x0B4E73EB059B85F0ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl2_c.mmh"},
	{0x0B4EC0128E86AA5DULL, "184842275.ogg"},
	{0x0B4EC9C07E679397ULL, "860633416.ogg"},
	{0x0B4ECFC06761EFCAULL, "431782801.ogg"},
	{0x0B4EDECFA99CB53AULL, "6098847_m.fxe"},
	{0x0B4EE88E11F6A079ULL, "hlac_lt_0_lowtown_a2_65533_1y.dds"},
	{0x0B4EF040474346ECULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01b.mmh"},
	{0x0B4EF640474350DEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01b.msh"},
	{0x0B4F172632A99948ULL, "ser221_party_comments6209855.cl"},
	{0x0B4F47D4CED7E719ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.cess.gad"},
	{0x0B4F4B7D0A8F0D8AULL, "6122441_m.fxe"},
	{0x0B4F60DF4575CF85ULL, "177888284.ogg"},
	{0x0B50449F325C1C8CULL, "additivedot01.mao"},
	{0x0B505FE0F9298830ULL, "hlac_lt_0_lowtown_a3_n_c_-196607_0y.dds"},
	{0x0B507E8070D4DD56ULL, "follower_banter6160789.cl"},
	{0x0B508F67EC2BDD7AULL, "vxlt_backalley_n_1u5_0_0.vlm"},
	{0x0B524B6FE975E424ULL, "196488332.ogg"},
	{0x0B529A8BEC7BE36CULL, "w_sundermount_134_0.rml"},
	{0x0B52A8CE392FE21DULL, "llac_ca_caverns1_mines_24422_7.dds"},
	{0x0B534DFCE98CEB8AULL, "gen_im_npc_elf_arm_cht_lgtc.uti"},
	{0x0B53CF82EF632ADEULL, "one200dg_solivitus6060325.cl"},
	{0x0B5462F40E29E0D4ULL, "vxca_caverns1_mines_iib_0_0.vlm"},
	{0x0B549C64BB9AB238ULL, "zz_qun_debug6045390.cl"},
	{0x0B550F2AA94ED8BAULL, "vxca_caverns2prc_a4h_0_0.vlm"},
	{0x0B5572BDF3560C52ULL, "ht_keep_3741.anb"},
	{0x0B558002CE0D356FULL, "680031250.ogg"},
	{0x0B5658D4CEDE2367ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.cess.ani"},
	{0x0B56FA4F9A1984B9ULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash.impa.ani"},
	{0x0B5702A8C7B8A824ULL, "6101457_m.fxe"},
	{0x0B575063B72977F7ULL, "vxlt_0_lowtown_a3_n_c_gxq_0_0.vlm"},
	{0x0B57B0E1D5FA9DB3ULL, "vxht_0_hightown_n_7g3_0_0.vlm"},
	{0x0B57D6428FB48F0FULL, "gen_im_arm_cht_nobb_t3.uti"},
	{0x0B5801705A6D6B3FULL, "270179991.ogg"},
	{0x0B5810CAA6C5C83DULL, "hlac_w_encounter2_camp_65538_0z.dds"},
	{0x0B581B7CCF710A46ULL, "6180607_m.fxe"},
	{0x0B5823E7661C5F2EULL, "mrl310_merrill6173877.cl"},
	{0x0B583B74EF26E6A4ULL, "6141619_m.fxe"},
	{0x0B58523F15FEABA4ULL, "vxht_0_hightown_n_ajm_0_0.vlm"},
	{0x0B5879942AEF389BULL, "art\\vfx\\combat\\vfx_w_bsk_regain_sta_imp_c.impa.ani"},
	{0x0B59C1149D0FC0D5ULL, "6141867_m.fxe"},
	{0x0B5A99D419583FBAULL, "vxlt_warehouse_157_0_0.vlm"},
	{0x0B5AB0607F661157ULL, "450056270.ogg"},
	{0x0B5B51F6299737FAULL, "hlac_ca_caverns1_mines_24425_9x.dds"},
	{0x0B5B64C5E0EB8238ULL, "vxht_keep_5u3_0_0.vlm"},
	{0x0B5B87A00ED51C8BULL, "t3_eey_br2.tnt"},
	{0x0B5BD39ADBF732FDULL, "vxlt_gallowstemplar_5au_0_0.vlm"},
	{0x0B5C1F2B2BCC6513ULL, "10225711.ogg"},
	{0x0B5C8646BB530EFDULL, "root_mage_primal2.dds"},
	{0x0B5C96EB05A7FE0FULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl2_c.anb"},
	{0x0B5D0F144CCC154BULL, "vxca_cave_1b1_0_0.vlm"},
	{0x0B5D8655F798D1B7ULL, "6137068_m.fxe"},
	{0x0B5DB596C0E224FBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f_hand_gesture_other.ani"},
	{0x0B5DD54F9A1F64FBULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash.impa.gad"},
	{0x0B5E6E4F0B1D1A21ULL, "dae221_viscount_chantry6154266.cl"},
	{0x0B5EA3520CB43BC5ULL, "vxlt_backalley_n_1o0_0_0.vlm"},
	{0x0B5F8F1984FE6466ULL, "lowtown_male_lowlife2_mob.cnv"},
	{0x0B5F97DFA518E9FCULL, "w_bonepitdestroyed_a3_155_0.phy"},
	{0x0B5FB84B9641116FULL, "vxlt_warehouse_pl_0_0.vlm"},
	{0x0B5FBD1420D4D33FULL, "503869286.ogg"},
	{0x0B601652CAE8E6C9ULL, "6093777_m.fxe"},
	{0x0B6040E6A4714282ULL, "6075532_m.fxe"},
	{0x0B60615092DA8FA8ULL, "ave321cr_gen_guardsman.utc"},
	{0x0B6071C81D94EC72ULL, "qun120cr_javaris6042975.cl"},
	{0x0B6094133DC935BCULL, "6188645_m.fxe"},
	{0x0B60B1A1F6FCECABULL, "6047248_m.fxe"},
	{0x0B60C52304CA23A9ULL, "6105940_m.fxe"},
	{0x0B60ED2256CB64B5ULL, "dae300_vashothlead6065065.cl"},
	{0x0B610843CBCC3E02ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_cwk.gad"},
	{0x0B615B11DD310FC6ULL, "plt_cod_bks_black_fox.plo"},
	{0x0B621F3383B17D6CULL, "llac_lt_0_lowtown_a3_n_c_65537_3.dds"},
	{0x0B6237AC0F16FBEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightdisengage.gad"},
	{0x0B629DA9CEBD57B5ULL, "vxht_keep_n_3zv_0_0.vlm"},
	{0x0B62B458D50F3526ULL, "6129071_m.fxe"},
	{0x0B62D238AA1EBAA3ULL, "6079667_m.fxe"},
	{0x0B62FD4D67FBA4F5ULL, "57487671.ogg"},
	{0x0B6301297136B22BULL, "vxlt_gallowstemplar_45r_0_0.vlm"},
	{0x0B636A5F6CC4F87FULL, "art\\characters\\playercharacter\\humanfemale\\rob_appb.phy"},
	{0x0B6381079B44F13AULL, "vxca_caverns2prc_a3i_0_0.vlm"},
	{0x0B63911985025026ULL, "lowtown_male_lowlife2_mob.bnk"},
	{0x0B6399E0A6309BCBULL, "vxlt_0_lowtown_ei4_0_0.vlm"},
	{0x0B63B309ECEC98F3ULL, "vxlt_0_lowtown_docks_n_bay_0_0.vlm"},
	{0x0B63CC77A33D324CULL, "vxlt_0_lowtown_docks_a3_n_d08_0_0.vlm"},
	{0x0B63F24F9A2401E0ULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash.impa.evt"},
	{0x0B641BC1DBE12E1CULL, "plt_lgt260pt_way_blood_mage.plo"},
	{0x0B644E4155AFC2D6ULL, "6187945_m.fxe"},
	{0x0B648D7C221546C6ULL, "vxht_0_hightown_n_b95_0_0.vlm"},
	{0x0B64F0E1B5904BECULL, "94414545.ogg"},
	{0x0B64FC21156D7494ULL, "zz_fen_debug6134363.cl"},
	{0x0B6581942AF9D092ULL, "art\\vfx\\combat\\vfx_w_bsk_regain_sta_imp_c.impa.evt"},
	{0x0B659609BC83CB4BULL, "vxht_chantry_n_2yd_0_0.vlm"},
	{0x0B65DF894488BFB3ULL, "vxdr_0_deeproads_fip_0_0.vlm"},
	{0x0B65E9E2825D41FBULL, "hlac_lt_0_lowtown_a2_-131073_2x.dds"},
	{0x0B65FD05A3157AFDULL, "ser210_witness6168177.cl"},
	{0x0B660C3E842F45A9ULL, "gen00fl_varric6103557.cl"},
	{0x0B66875AA9CA40C2ULL, "dae000_bodhan6129078.cl"},
	{0x0B66E1263E93DD21ULL, "vxlt_darktown_5tf_0_0.vlm"},
	{0x0B67357830008C13ULL, "6036925_m.fxe"},
	{0x0B679AF554D69296ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.anb"},
	{0x0B67B37706CFC3A5ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_infb_0n.dds"},
	{0x0B68207AEF5D4F91ULL, "552102237.ogg"},
	{0x0B682B43CBD298F9ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_cwk.evt"},
	{0x0B683B3E22F32113ULL, "vxht_mansion2_1wb_0_0.vlm"},
	{0x0B684F3E8546CBCFULL, "49794167.ogg"},
	{0x0B68CC6DADB5D954ULL, "600069353.ogg"},
	{0x0B691AAC0F1CE881ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightdisengage.evt"},
	{0x0B6974FC1E007CC1ULL, "gen00fl_varric6093616.cl"},
	{0x0B69C1F14154AA71ULL, "6101231_f.fxe"},
	{0x0B6A5BA534C058F6ULL, "vxlt_0_lowtown_a3_n_c_i4f_0_0.vlm"},
	{0x0B6A843B56D57314ULL, "art\\levels\\races\\design\\animation\\beartrap.activetoinactive.ani"},
	{0x0B6B284648137925ULL, "vxw_bonepitdestroyed_a3_4q_0_0.vlm"},
	{0x0B6B3CF8414A76D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_basic.msh"},
	{0x0B6B46F8414A87D4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_basic.mmh"},
	{0x0B6B4AF8414A8EE7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_basic.mao"},
	{0x0B6B516B6CBE1F9EULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.dura.ani"},
	{0x0B6B5F21D51FB27DULL, "llac_ca_cavern1_haunted_24426_10.dds"},
	{0x0B6B73D74C6C6C02ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.dura.gad"},
	{0x0B6B88CD88019B38ULL, "55983099.ogg"},
	{0x0B6C86AE9EA2EEEFULL, "vxlt_0_lowtown_a2_n_es9_0_0.vlm"},
	{0x0B6CA0942B002425ULL, "art\\vfx\\combat\\vfx_w_bsk_regain_sta_imp_c.impa.gad"},
	{0x0B6CC57216907426ULL, "6139275_f.fxe"},
	{0x0B6CD9E39FF15D8BULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.cess.ani"},
	{0x0B6CEEFFCD9D1B17ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur1_flemeth_real.gad"},
	{0x0B6D56B7FBE9FA01ULL, "vxw_bonepitdestroyed_a3_9n_0_0.vlm"},
	{0x0B6DA5DE96E9C9C9ULL, "gen00fl_aveline6113993.cl"},
	{0x0B6DE69A6C98412EULL, "6043884_f.fxe"},
	{0x0B6E2A28883729EFULL, "vxlt_warehouse_uo_0_0.vlm"},
	{0x0B6E3D62B1CF94EFULL, "llac_lt_undercityencounter_-1_18.dds"},
	{0x0B6EA2D94A21E39DULL, "vxht_mansion_n_5a5_0_0.vlm"},
	{0x0B6EE2AA00ABE327ULL, "character_character_p97.psh.fxo"},
	{0x0B6F7E7AF1B946AFULL, "vxht_mansion2b_2dt_0_0.vlm"},
	{0x0B6F94ED3E0D3165ULL, "6093526_m.fxe"},
	{0x0B6FC7B879484260ULL, "6100259_f.fxe"},
	{0x0B6FEEB0EF1EA55EULL, "vxht_mansion_n_53h_0_0.vlm"},
	{0x0B70FB2B37AA51FBULL, "llac_lt_0_lowtown_n_65534_2.dds"},
	{0x0B7127F57EA27CC3ULL, "vxca_cavern1_haunted_gkp_0_1.vlm"},
	{0x0B7166F9D222F962ULL, "6161431_m.fxe"},
	{0x0B71671C4A279B29ULL, "6055627_m.fxe"},
	{0x0B7173B31EB23372ULL, "gen00flst_isa_init_2.stg"},
	{0x0B71813B56DB8D76ULL, "art\\levels\\races\\design\\animation\\beartrap.activetoinactive.gad"},
	{0x0B71AA5DAC10BEEBULL, "gen_im_wep_war_shd_lore_01.uti"},
	{0x0B71DC96C0F30F85ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f_hand_gesture_other.gad"},
	{0x0B72374CC8184223ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.impa.evt"},
	{0x0B7296D74C72C6F9ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.dura.evt"},
	{0x0B72C51EBBA47A4AULL, "6071508_f.fxe"},
	{0x0B72F77D528239B1ULL, "256887159.ogg"},
	{0x0B734E38ABBA3180ULL, "vxca_caverns1_mines_hu7_0_0.vlm"},
	{0x0B7357E48EA080CDULL, "6087058_f.fxe"},
	{0x0B7407E6AE79E3CBULL, "vxht_chantry_n_2tg_0_0.vlm"},
	{0x0B746B5F6CD36C84ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appb.mmh"},
	{0x0B746F5F6CD373D7ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appb.mao"},
	{0x0B74815F6CD39226ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appb.msh"},
	{0x0B7481F94F197B95ULL, "vxlt_gallowscourtyard_n_774_0_0.vlm"},
	{0x0B74BF432D343F30ULL, "6064654_f.fxe"},
	{0x0B74E41F59890468ULL, "dae000_arishok6080482.cl"},
	{0x0B75015C5A246F82ULL, "6042884_m.fxe"},
	{0x0B757910865AD0A3ULL, "vxca_cavern1_haunted_hz1_0_0.vlm"},
	{0x0B759DF554E2D501ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.mmh"},
	{0x0B75CA7D92861370ULL, "vxca_caverns1_mines_gk6_0_0.vlm"},
	{0x0B75D5F00BAABB98ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1_tool.phy"},
	{0x0B762E9495B7162AULL, "505876861.ogg"},
	{0x0B763343CBDEE3E8ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_cwk.ani"},
	{0x0B76E581D878539EULL, "zz_mrl_debug6185299.cl"},
	{0x0B7712AC0F2918D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightdisengage.ani"},
	{0x0B77A725F8FB6DBAULL, "hlac_lt_0_lowtown_-65536_6y.dds"},
	{0x0B7812F70AB90F1BULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplarfemale3_.ani"},
	{0x0B789C3B56E1D9D5ULL, "art\\levels\\races\\design\\animation\\beartrap.activetoinactive.evt"},
	{0x0B78B754BFE363FDULL, "vxlt_gallowstemplar_5ws_0_0.vlm"},
	{0x0B79104CC81E1EA4ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.impa.gad"},
	{0x0B7940BC06DB8559ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_wic.mmh"},
	{0x0B794ABC06DB9647ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_wic.msh"},
	{0x0B79696B6CCA866FULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.dura.evt"},
	{0x0B797C5A20441F56ULL, "art\\characters\\creatures\\mubari\\animation\\drnk_ptn.ani"},
	{0x0B799F011C445D16ULL, "6036329_f.fxe"},
	{0x0B7A55F0C3CBF21DULL, "vxlt_hangedman_1w6_0_0.vlm"},
	{0x0B7A5CA362521A49ULL, "vxlt_0_lowtown_a2_n_e15_0_0.vlm"},
	{0x0B7A6EB414B17F54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_x_pos.gad"},
	{0x0B7AB815B93E3723ULL, "vxca_caverns1_ckp_0_0.vlm"},
	{0x0B7AC1E39FFD7202ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.cess.evt"},
	{0x0B7ACEFDF7165654ULL, "vrc321st_bartrand_papers.stg"},
	{0x0B7B3E47AD4CD23FULL, "1060156255.ogg"},
	{0x0B7C72F4A8D74150ULL, "mrl000pt_gift_ring.ncs"},
	{0x0B7C7E680F38B242ULL, "ach_comp_level_2.dds"},
	{0x0B7CBFBA5B211C6DULL, "648648962.ogg"},
	{0x0B7CC358D552518FULL, "6110948_m.fxe"},
	{0x0B7D219A93121662ULL, "6153236_m.fxe"},
	{0x0B7D5486F2A1FBA7ULL, "character_character_p40.psh.fxo"},
	{0x0B7DCE1B4066C0CCULL, "971939987.ogg"},
	{0x0B7DE262EF306AD1ULL, "sse_ss_crow_m_1.gda"},
	{0x0B7DE3CDAC0CC509ULL, "293477834.ogg"},
	{0x0B7DF057E3AF5F2FULL, "vxht_mansion2b_2aw_0_0.vlm"},
	{0x0B7E22B38FC66D16ULL, "484098421.ogg"},
	{0x0B7E4103B76D6845ULL, "vxw_blightlands_13f_0_0.vlm"},
	{0x0B7F4A788F5547DCULL, "m_dae240ar_bone_pit.wbd"},
	{0x0B7F6A6B6CCEF318ULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.dura.gad"},
	{0x0B7FA6A944732071ULL, "hlac_lt_backalley_g_0_14z.dds"},
	{0x0B7FBFA7CEA4FD61ULL, "vxw_sundermount_34w_0_0.vlm"},
	{0x0B801F4CC8245682ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.impa.ani"},
	{0x0B8039871CDCEB64ULL, "6156310_m.fxe"},
	{0x0B805699CD5A3596ULL, "866606914.ogg"},
	{0x0B809ED74C7F11E8ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.dura.ani"},
	{0x0B80B3A0AC70611CULL, "vxca_caverns2prc_6hu_0_0.vlm"},
	{0x0B80B6BEEDD99B77ULL, "gen00fl_anders6111280.cl"},
	{0x0B80BB143F3E7816ULL, "vxlt_0_lowtown_docks_95h_0_0.vlm"},
	{0x0B80EF351FC51C08ULL, "l_env_wal_gallows_callout.wbd"},
	{0x0B8113CFE70137D6ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbed_loop.ani"},
	{0x0B811BFFCDAE109DULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur1_flemeth_real.ani"},
	{0x0B81342EC4F9D874ULL, "6085551_m.fxe"},
	{0x0B814165DA5EA01DULL, "vxw_bonepitdestroyed_a3_25n_0_0.vlm"},
	{0x0B814172846A347EULL, "6109711_f.fxe"},
	{0x0B816A2EDB1FCA2DULL, "zz_followers_banter6158002.cl"},
	{0x0B817DB414B7B732ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_x_pos.ani"},
	{0x0B81C0E3A0038F15ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.cess.gad"},
	{0x0B81E01BCF366BF4ULL, "zz_mag_debug6039790.cl"},
	{0x0B81E43F18CB3D07ULL, "vxht_chantry_2cu_0_0.vlm"},
	{0x0B82471916E69CF1ULL, "775130811.ogg"},
	{0x0B82A8E19325CBB2ULL, "lgt120_orz_bln_official6176122.cl"},
	{0x0B82F9B7C64B2416ULL, "667564765.ogg"},
	{0x0B83301A61218F9DULL, "6105830_m.fxe"},
	{0x0B8345474B060EDFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_hangtarp_new_d.dds"},
	{0x0B8469D87912D72AULL, "vxw_encounter3_96_0_0.vlm"},
	{0x0B847B2D4A4F6EF7ULL, "hlac_ca_caverns2_5802_64z.dds"},
	{0x0B84BEC79739D89AULL, "vxht_mansion2_3gh_0_1.vlm"},
	{0x0B84EF52A5B0AD81ULL, "6103458_m.fxe"},
	{0x0B852F9B9CFA1590ULL, "6121565_m.fxe"},
	{0x0B85AC0633B6BFAEULL, "vxht_0_hightown_n_9pa_0_0.vlm"},
	{0x0B8606780CFE278FULL, "art\\characters\\creatures\\mubari\\animation\\hv_dmg.gad"},
	{0x0B86122565651476ULL, "vxca_caverns1_gkm_0_0.vlm"},
	{0x0B8624FB111DC6A2ULL, "zz_approval_debug6133708.cl"},
	{0x0B864CD752DFD956ULL, "lt_0_lowtown_docks_n_206_0.mao"},
	{0x0B8658D752DFEDF5ULL, "lt_0_lowtown_docks_n_206_0.mmh"},
	{0x0B865AD752DFF16BULL, "lt_0_lowtown_docks_n_206_0.msh"},
	{0x0B86845A204EB747ULL, "art\\characters\\creatures\\mubari\\animation\\drnk_ptn.evt"},
	{0x0B86C63F4BF6E96EULL, "6087018_m.fxe"},
	{0x0B86EC57EDFEA95FULL, "vxlt_hangedman_n_27l_0_0.vlm"},
	{0x0B87893B267890DAULL, "6125835_m.fxe"},
	{0x0B8797214FCEF535ULL, "6163355_m.fxe"},
	{0x0B87BE88C5D15360ULL, "vxlt_gallowstemplar_38u_0_0.vlm"},
	{0x0B87D9C3FDCDA120ULL, "qun221cr_mob_leader.utc"},
	{0x0B8850162080F28CULL, "w_encounter3_173_0.msh"},
	{0x0B8856162080FD7EULL, "w_encounter3_173_0.mmh"},
	{0x0B88621620811115ULL, "w_encounter3_173_0.mao"},
	{0x0B8875BDF3816A8EULL, "ht_keep_3741.rml"},
	{0x0B89449BDAC7D434ULL, "vxdr_primevaltomb_2gr_0_1.vlm"},
	{0x0B89485D8FF9A948ULL, "vxw_sundermount_34r_0_0.vlm"},
	{0x0B8985A87F2D3317ULL, "771886767.ogg"},
	{0x0B8A8B69163592C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_knife_to_throat_loop.gad"},
	{0x0B8AC5C598E97795ULL, "dae000ip_ambient_liecudlf_s.utp"},
	{0x0B8B38F2EC9D8760ULL, "mag100_thrask6031998.cl"},
	{0x0B8B3EA408AA82ECULL, "6219277_m.fxe"},
	{0x0B8B7401F1B688CFULL, "dae000_hate_follower_react6103218.cl"},
	{0x0B8B95FDE5C1A747ULL, "character_character_p59.psh.fxo.dx11"},
	{0x0B8BC011954BD455ULL, "vxht_keep_493_0_0.vlm"},
	{0x0B8BC663E4981427ULL, "character_character_p31.psh.fxo"},
	{0x0B8C1B780D02B73CULL, "art\\characters\\creatures\\mubari\\animation\\hv_dmg.evt"},
	{0x0B8C2E3E9B71622BULL, "732782075.ogg"},
	{0x0B8C39F70AC9FAA5ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplarfemale3_.gad"},
	{0x0B8C7F5FAD4FFF3FULL, "6270314.ogg"},
	{0x0B8CD26AD30AD65EULL, "vxht_mansion_n_59j_0_0.vlm"},
	{0x0B8CFCCFFEC621A5ULL, "vfx_chunkyparticle_p4.psh.fxo.dx11"},
	{0x0B8D3BF632551E8DULL, "vxlt_gallowscourtyard_4qj_0_0.vlm"},
	{0x0B8D6E226E7B69C0ULL, "6028411_m.fxe"},
	{0x0B8DA55A20550D90ULL, "art\\characters\\creatures\\mubari\\animation\\drnk_ptn.gad"},
	{0x0B8DABEEB26BF6B9ULL, "vrc221tr_first_corpses.utt"},
	{0x0B8DB3F7043156DDULL, "798892543.ogg"},
	{0x0B8E0B66DA37E4D9ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax101a_0t.dds"},
	{0x0B8E3184C09B15E1ULL, "vxw_sundermount_33v_0_0.vlm"},
	{0x0B8F96CB8AE55912ULL, "vxlt_hangedman_28g_0_0.vlm"},
	{0x0B8FD6A6BA197237ULL, "fen110_anso6094988.cl"},
	{0x0B8FFBC4984CEB44ULL, "conversation_ani_variations.gda"},
	{0x0B8FFD03F7AC6495ULL, "vxca_caverns1_mines_gey_0_0.vlm"},
	{0x0B9031FE1CD1F729ULL, "follower_banter6165365.cl"},
	{0x0B90348511C696C2ULL, "llac_lt_0_lowtown_1_4.dds"},
	{0x0B910A6BE6D7F234ULL, "vxw_bonepitmining_a2_1x0_0_1.vlm"},
	{0x0B913119C41BA36BULL, "vxht_mansion2b_1ip_0_0.vlm"},
	{0x0B918C69163BB36FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_knife_to_throat_loop.ani"},
	{0x0B91DCB7909CB64EULL, "182929414.ogg"},
	{0x0B91E525E9EF5717ULL, "hlac_ht_0_hightown_a2_131071_1x.dds"},
	{0x0B91E766A401158FULL, "717188120.ogg"},
	{0x0B9200BCB3EE002DULL, "98479718.ogg"},
	{0x0B9205A0DB73B9FFULL, "637780716.ogg"},
	{0x0B920CA4E708AB1BULL, "vxlt_0_lowtown_a2_n_gz4_0_0.vlm"},
	{0x0B923A62AC1CDA4AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglebody_n.dds"},
	{0x0B92DBB56B08EFAAULL, "vxw_encounter3_27_0_0.vlm"},
	{0x0B93522D99DA9BC4ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgtb_0n.dds"},
	{0x0B93B1FBCA23936FULL, "lt_0_lowtown_10_0.rml"},
	{0x0B9418FA21676EA0ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_orsa_0d.dds"},
	{0x0B9478C72DAB1FE3ULL, "gencr_dalish_warrior_2w.utc"},
	{0x0B94CD89B16E01F7ULL, "1017516622.ogg"},
	{0x0B95138E8468E4F4ULL, "lgt220_shp_mag_guard6149889.cl"},
	{0x0B953CCFE7122610ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbed_loop.gad"},
	{0x0B955E34DFF4C1DFULL, "vxlt_hangedman_n_22o_0_0.vlm"},
	{0x0B95A3EBE04CA3D4ULL, "art\\vfx\\combat\\vfx_r_brd_psong_act_c.anb"},
	{0x0B95E1EA3681219DULL, "follower_banter6141680.cl"},
	{0x0B96239E77505C1AULL, "llac_lt_0_lowtown_n_-3_6.dds"},
	{0x0B9648817AE3D77AULL, "llac_lt_backalley_n_0_5.dds"},
	{0x0B96C4C6A86806C6ULL, "vxlt_0_lowtown_ewg_0_0.vlm"},
	{0x0B96FC1066075C98ULL, "vxlt_0_lowtown_f4c_0_0.vlm"},
	{0x0B97BA3A81EFC1C8ULL, "vxw_sundermount_33s_0_0.vlm"},
	{0x0B980F7B82EF30C5ULL, "zz_followers_banter6159276.cl"},
	{0x0B987D263D4E0D4DULL, "art\\vfx\\models\\vfx_branch.mmh"},
	{0x0B988A3AF7E0B8ADULL, "vxht_brothel_n_54e_0_0.vlm"},
	{0x0B988F263D4E2BD3ULL, "art\\vfx\\models\\vfx_branch.msh"},
	{0x0B99F3AD610BAC00ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brie_d.dds"},
	{0x0B9A23780D0F0255ULL, "art\\characters\\creatures\\mubari\\animation\\hv_dmg.ani"},
	{0x0B9AFC1F5C2E2451ULL, "vxlt_gallowstemplar_4up_0_0.vlm"},
	{0x0B9B7A79036139E6ULL, "dae011_badpoet_sta6145209.cl"},
	{0x0B9BC8FB733DDC79ULL, "vxlt_0_lowtown_a2_gmp_0_0.vlm"},
	{0x0B9BE7E79C41D77DULL, "gen00fl_merrill6145212.cl"},
	{0x0B9C338EEC4A73FCULL, "445473560.ogg"},
	{0x0B9C56AECDE8D3C8ULL, "6060467_f.fxe"},
	{0x0B9C662A8F9AB217ULL, "648664676.ogg"},
	{0x0B9CB892858FD1BEULL, "6107530_m.fxe"},
	{0x0B9D41CB1DF4A584ULL, "fighting.mao"},
	{0x0B9D4BD8EF323003ULL, "hlac_w_bonepitdestroyed_a3_131074_0y.dds"},
	{0x0B9D98AD96EC8F1DULL, "vxlt_gallowsprison_n_5s0_0_0.vlm"},
	{0x0B9E4982C346CA16ULL, "gen00fl_anders6062178.cl"},
	{0x0B9EC05DAA848222ULL, "685383070.ogg"},
	{0x0B9ED9B7595C3DE8ULL, "lgt110cr_gxa_vgk_cnspr_1.utc"},
	{0x0B9EDBCD2B352879ULL, "6028797_m.fxe"},
	{0x0B9FB74AC4FB3AD3ULL, "ave121cr_brennan.utc"},
	{0x0B9FDE0DF66B7B21ULL, "451010169.ogg"},
	{0x0B9FE1AB72444391ULL, "qun111_ketojan6041040.cl"},
	{0x0B9FE2E733DE86B8ULL, "vxca_caverns2prc_94b_0_0.vlm"},
	{0x0BA07E81D8FEC39BULL, "vxlt_0_lowtown_a2_n_gw7_0_0.vlm"},
	{0x0BA16B90CB0BA3DAULL, "345215941.ogg"},
	{0x0BA195D0749A4CEAULL, "vxlt_undercityserialkiller_18c_0_0.vlm"},
	{0x0BA1B49BE9CB2E5AULL, "mag320cr_templar_1.utc"},
	{0x0BA282EBE056F597ULL, "art\\vfx\\combat\\vfx_r_brd_psong_act_c.mmh"},
	{0x0BA28920FEC7382CULL, "423320165.ogg"},
	{0x0BA28D35F3EA01B7ULL, "6031313_m.fxe"},
	{0x0BA2B35E10EB098FULL, "llac_lt_0_lowtown_docks_a3_-65537_3.dds"},
	{0x0BA2DB29B71ADFE8ULL, "40014436.ogg"},
	{0x0BA3047480687CDBULL, "vxlt_0_lowtown_gpi_0_0.vlm"},
	{0x0BA376872F4DC8CCULL, "573927652.ogg"},
	{0x0BA401BAE7A860C8ULL, "hlac_lt_0_lowtown_n_131070_2z.dds"},
	{0x0BA4986B820E8D14ULL, "6135659_f.fxe"},
	{0x0BA49CF24FB089D4ULL, "6050085_m.fxe"},
	{0x0BA4AA5848A1FEEEULL, "6019211_m.fxe"},
	{0x0BA4B208D712AF4DULL, "w_bonepitdestroyed_a3_180_0.mmh"},
	{0x0BA4B608D712B59EULL, "w_bonepitdestroyed_a3_180_0.mao"},
	{0x0BA4C408D712CDD3ULL, "w_bonepitdestroyed_a3_180_0.msh"},
	{0x0BA536A39A5E1F46ULL, "vxlt_0_lowtown_erh_0_0.vlm"},
	{0x0BA5728817324B63ULL, "6164561_m.fxe"},
	{0x0BA5FB90CAC585CDULL, "hlac_lt_0_lowtown_docks_a3_-65538_6x.dds"},
	{0x0BA5FD0F4BEB0FBCULL, "911200841.ogg"},
	{0x0BA61560D0DAEEC6ULL, "209521803.ogg"},
	{0x0BA6A83608FA385AULL, "fen210_tevinter_ambusher6108302.cl"},
	{0x0BA6B52260CBAB10ULL, "mcr371cr_corpse.ncs"},
	{0x0BA6B99D62A28957ULL, "vxlt_hangedman_w1_0_0.vlm"},
	{0x0BA6D7044C55482BULL, "6205648_m.fxe"},
	{0x0BA6DFFF469A534EULL, "vxht_bartrand_n_67t_0_0.vlm"},
	{0x0BA7A4ED653A9445ULL, "114406979.ogg"},
	{0x0BA81D60CCBA6339ULL, "zze_roguevshighdragon_white.cut"},
	{0x0BA83AAA73780E85ULL, "art\\vfx\\combat\\vfx_w_wns_left_close1_c.mmh"},
	{0x0BA8DE504B1D70F5ULL, "llac_lt_backalley_g_0_7.dds"},
	{0x0BA9D063EE213494ULL, "dae000_orsino6154306.cl"},
	{0x0BA9DA6DF349CE4DULL, "vxw_bonepit_la_0_0.vlm"},
	{0x0BAA1BD17049F712ULL, "700866720.ogg"},
	{0x0BAA29497C9F6D9AULL, "27641974.ogg"},
	{0x0BAA551BA79D0A27ULL, "hlac_dr_0_deeproads_5842_17z.dds"},
	{0x0BAADCCD832A173FULL, "595556839.ogg"},
	{0x0BAB27A753318991ULL, "hlac_lt_0_lowtown_131072_0x.dds"},
	{0x0BAB48788F7A7F56ULL, "m_dae240ar_bone_pit.bnk"},
	{0x0BAB7FABA82D421AULL, "228815538.ogg"},
	{0x0BABACEEB284F67CULL, "vrc221tr_first_corpses.ncs"},
	{0x0BAC26F91106CA22ULL, "vxw_woundedcoast_hw_0_0.vlm"},
	{0x0BAC3B1825883FA2ULL, "630835171.ogg"},
	{0x0BAC7A856ED18A12ULL, "vxlt_hangedman_22a_0_0.vlm"},
	{0x0BAC9A7F3633A8F1ULL, "vxlt_hangedman_n_2xc_0_0.vlm"},
	{0x0BACCB9808AAFD4AULL, "354948399.ogg"},
	{0x0BACD614BFEA2772ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\blendtree\\doorsimple.anb"},
	{0x0BAD289EDDBDB46AULL, "art\\vfx\\combat\\vfx_m_pri_wintgrsp_imp_p.impa.gad"},
	{0x0BAD29284A58D6ACULL, "vxlt_warehouse_n_12c_0_0.vlm"},
	{0x0BAD3FCC2EA0A53BULL, "6169500_m.fxe"},
	{0x0BAD3FCECA74DC88ULL, "119827350.ogg"},
	{0x0BAD5E011B173EF8ULL, "dre000_dougal6131131.cl"},
	{0x0BADF30424DE119EULL, "vxlt_darktown_588_0_1.vlm"},
	{0x0BAE14D3A807D46BULL, "vxht_mansion2b_1gr_0_0.vlm"},
	{0x0BAEA9C08E320BEBULL, "6144329_m.fxe"},
	{0x0BAEE35DCD57561AULL, "6075770_m.fxe"},
	{0x0BAF3157609D0DD1ULL, "vxht_0_hightown_cds_0_0.vlm"},
	{0x0BAF3F4C792F4E72ULL, "6176299_m.fxe"},
	{0x0BAFCAAF401F3CADULL, "384741376.ogg"},
	{0x0BAFE2D4C4183D54ULL, "hlac_w_encounter2_196610_0y.dds"},
	{0x0BAFF5BB96E4792FULL, "hlac_lt_0_lowtown_a3_n_c_0_4z.dds"},
	{0x0BAFFC63E2FBB9A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_order_1.ani"},
	{0x0BB00B5A329E2D8EULL, "164664799.ogg"},
	{0x0BB00DFA966D4217ULL, "dre000_hireling16038627.cl"},
	{0x0BB0B7C06FDB7CFEULL, "cd_mrl250_pol.bnk"},
	{0x0BB15E2A356CF9D2ULL, "llac_lt_undercitydungeon_-1_50.dds"},
	{0x0BB1ECBDBE67482EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appb_0s.dds"},
	{0x0BB24DC610C9196DULL, "gen00fl_anders6103735.cl"},
	{0x0BB25FA4D73F6F65ULL, "676346168.ogg"},
	{0x0BB28DAA4B63576AULL, "gen00fl_isabela6096764.cl"},
	{0x0BB28E8A47BCF2A1ULL, "vxdr_primevaltomb_8px_0_1.vlm"},
	{0x0BB3505E60790971ULL, "gen00fl_merrill6152667.cl"},
	{0x0BB35DF0F29E60CBULL, "6077078_m.fxe"},
	{0x0BB378A8A2AB7349ULL, "l_env_wal_hightown_int_calm_p2p.bnk"},
	{0x0BB3D27F7E0DFC91ULL, "and121_karl6149823.cl"},
	{0x0BB400F1531DB996ULL, "6177234_m.fxe"},
	{0x0BB40B9EDDC3A101ULL, "art\\vfx\\combat\\vfx_m_pri_wintgrsp_imp_p.impa.evt"},
	{0x0BB42CF436B3A411ULL, "lt_0_lowtown_docks_n_29_0_ll.mmh"},
	{0x0BB4608A328963EBULL, "6127591_m.fxe"},
	{0x0BB4D42F3C501813ULL, "546927435.ogg"},
	{0x0BB51AE531482168ULL, "806540607.ogg"},
	{0x0BB537AA73829452ULL, "art\\vfx\\combat\\vfx_w_wns_left_close1_c.anb"},
	{0x0BB55F8CE2FD8301ULL, "6069646_m.fxe"},
	{0x0BB58F64EB81FE23ULL, "vxw_encounter2_cr_0_0.vlm"},
	{0x0BB5C3EA04FE43B2ULL, "llac_lt_0_lowtown_docks_a3_n_-65537_3.dds"},
	{0x0BB742F1ABE7960FULL, "pro000_merchant6051810.cl"},
	{0x0BB7BB057682B5CAULL, "6109767_m.fxe"},
	{0x0BB84500FDCF94B4ULL, "hightown_market_guard_sta6169631.cl"},
	{0x0BB88F45FFE37BF0ULL, "hlac_ht_mansion2_818_59x.dds"},
	{0x0BB899AC0FE49182ULL, "dae210_gate_guard6085048.cl"},
	{0x0BB8D11A45C8B746ULL, "qcr000_arishok6060150.cl"},
	{0x0BB94A48728BD7F8ULL, "9302965.ogg"},
	{0x0BB954A10610B588ULL, "vrc321_varric6173851.cl"},
	{0x0BB967F1952CA981ULL, "qcr220st_orsino.stg"},
	{0x0BB9B6AEFD23754BULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_2z.dds"},
	{0x0BB9DB18DCA44472ULL, "follower_banter6172795.cl"},
	{0x0BBA0321F69CC39EULL, "vxlt_0_lowtown_docks_a3_n_d4v_0_0.vlm"},
	{0x0BBA686DE1018D41ULL, "vxlt_undercitydungeon_11r_0_0.vlm"},
	{0x0BBA8037AD30CA2BULL, "vxdr_primevaltomb_1cc_0_0.vlm"},
	{0x0BBA91CF1B67C21AULL, "414128224.ogg"},
	{0x0BBAB69A9E52B404ULL, "gencr_dalish_assassin.utc"},
	{0x0BBAF97B75FAFAB0ULL, "183141752.ogg"},
	{0x0BBB6EBA41055C0EULL, "6143713_f.fxe"},
	{0x0BBB838BB537AC91ULL, "character_character_p147.psh.fxo.dx11"},
	{0x0BBB8F9124E5C9C6ULL, "vxca_cave_1ae_0_0.vlm"},
	{0x0BBBC551BBAB2A7BULL, "vxht_brothel_n_52u_0_0.vlm"},
	{0x0BBC138D698DF16CULL, "vxht_mansion_n_6km_0_0.vlm"},
	{0x0BBC1C620AE47C3AULL, "vxw_woundedcoast_1gc_0_0.vlm"},
	{0x0BBC406AD4DF3D7EULL, "vxht_brothel_n_5ko_0_0.vlm"},
	{0x0BBD64282A7E7BD0ULL, "784818604.ogg"},
	{0x0BBD64836CF92DCCULL, "vxlt_0_lowtown_docks_a3_n_c3j_0_0.vlm"},
	{0x0BBDBC3ADC03A268ULL, "vxlt_gallowscourtyard_n_3vj_0_0.vlm"},
	{0x0BBEAE646DA024CEULL, "hlac_lt_0_lowtown_a3_n_c_65536_6x.dds"},
	{0x0BBF0A78047E6893ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv2.ani"},
	{0x0BBF8D8B4833F990ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screenb.mao"},
	{0x0BBF918B483400A3ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screenb.mmh"},
	{0x0BBF9F8B483418FDULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screenb.msh"},
	{0x0BBFE327A5AA5E91ULL, "pro000_athenril6049469.cl"},
	{0x0BBFED87AD3AECFCULL, "vxlt_0_lowtown_a2_n_c_hm3_0_0.vlm"},
	{0x0BC05107AC28AD51ULL, "mag211_marethari6025326.cl"},
	{0x0BC08CF01C0C4516ULL, "vxht_brothel_n_4dc_0_0.vlm"},
	{0x0BC0C989AF5857D4ULL, "hlac_lt_0_lowtown_docks_-196607_0x.dds"},
	{0x0BC16B61FFF275EAULL, "hlac_ca_caverns1_24424_1x.dds"},
	{0x0BC1CFCDE494794BULL, "6078178_m.fxe"},
	{0x0BC2039EDDCFD150ULL, "art\\vfx\\combat\\vfx_m_pri_wintgrsp_imp_p.impa.ani"},
	{0x0BC21E9401F77F79ULL, "dae000_fenarel6111552.cl"},
	{0x0BC2314A89169A42ULL, "gen00fl_isabela6135643.cl"},
	{0x0BC243016F21F564ULL, "261596254.ogg"},
	{0x0BC28C03A23E8157ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocksmall.msh"},
	{0x0BC29B75D95BF86FULL, "mrl150_hunter6071003.cl"},
	{0x0BC29E03A23E9FBAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocksmall.mao"},
	{0x0BC2A203A23EA6C9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocksmall.mmh"},
	{0x0BC30E97B7885AA6ULL, "gen00fl_sebastian6180158.cl"},
	{0x0BC3E563E30C3BA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_order_1.gad"},
	{0x0BC40141DD7816A3ULL, "vxw_encounter2_nu_0_0.vlm"},
	{0x0BC4B1584A3F06C7ULL, "43993034.ogg"},
	{0x0BC4D4CC6DF1788FULL, "hlac_ht_keep_3732_12z.dds"},
	{0x0BC511E6FA6AE966ULL, "vxht_bartrand_n_73a_0_0.vlm"},
	{0x0BC53BC873A86C09ULL, "vxw_bonepitmining_a2_1nz_0_0.vlm"},
	{0x0BC5404EAD9906A6ULL, "6086259_m.fxe"},
	{0x0BC62CE26878CE4AULL, "6102667_m.fxe"},
	{0x0BC63A9E72D39430ULL, "zz_ave_debug6174720.cl"},
	{0x0BC643033F29C17CULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.impa.ani"},
	{0x0BC6447C176448EBULL, "qun230tr_contact_ambient.utt"},
	{0x0BC66BF67BB5F8A0ULL, "and111_lirene6084536.cl"},
	{0x0BC68CA265FA6D1BULL, "fx_fl_smoke.dds"},
	{0x0BC68D60360128BCULL, "192539577.ogg"},
	{0x0BC699CF4BC71976ULL, "vxlt_0_lowtown_docks_a3_d0i_0_0.vlm"},
	{0x0BC6AD2B2B064C73ULL, "970669075.ogg"},
	{0x0BC6B7E8B04EFC04ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\wkf.gad"},
	{0x0BC7450869EC69F8ULL, "525077066.ogg"},
	{0x0BC745F84198E4CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_basic.phy"},
	{0x0BC7554E26DEA3FBULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_atlas_jimmy_0n.dds"},
	{0x0BC76C4DE51F0FFBULL, "hlac_w_sundermount_0_5y.dds"},
	{0x0BC76C5896004214ULL, "6179528_m.fxe"},
	{0x0BC7AAC1B1EF5C06ULL, "vxht_brothel_n_5fp_0_0.vlm"},
	{0x0BC7B0AC104DCB99ULL, "llac_ht_keep_3732_7.dds"},
	{0x0BC7E7E0D74D4EE2ULL, "vxlt_warehouse_n_dd_0_0.vlm"},
	{0x0BC805669B074D3FULL, "vxht_keep_n_607_0_0.vlm"},
	{0x0BC8153520015FCAULL, "l_env_wal_gallows_callout.bnk"},
	{0x0BC8DF8FEB553B3CULL, "6154257_m.fxe"},
	{0x0BC90AB2F4F2FB22ULL, "vxw_woundedcoast_bq_0_0.vlm"},
	{0x0BC91F6A9E09AB4EULL, "913571004.ogg"},
	{0x0BCA737CCC093FD2ULL, "dae000_gen_player_turn_in6177857.cl"},
	{0x0BCA76C02D79F05EULL, "vxw_encounter2_camp_rp_0_0.vlm"},
	{0x0BCA856A5B8409ECULL, "vxht_mansion_n_6nb_0_0.vlm"},
	{0x0BCAE9C64084AFA3ULL, "vxlt_undercitydungeon_1j5_0_0.vlm"},
	{0x0BCB7F4350F79034ULL, "zz_one_debug6077409.cl"},
	{0x0BCBAEC17981DF1CULL, "vxlt_0_lowtown_a2_n_c_hr2_0_0.vlm"},
	{0x0BCBE00CE35CED8CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerb_sa.dds"},
	{0x0BCC27873FD8299AULL, "lt_0_lowtown_docks_a3_n_242_0.mmh"},
	{0x0BCC2B873FD830D1ULL, "lt_0_lowtown_docks_a3_n_242_0.mao"},
	{0x0BCC39873FD848E0ULL, "lt_0_lowtown_docks_a3_n_242_0.msh"},
	{0x0BCC43291BF9A6FFULL, "754096776.ogg"},
	{0x0BCC81A27CAE79F4ULL, "575406623.ogg"},
	{0x0BCD0278048A98CAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv2.evt"},
	{0x0BCD50033F2FF5BEULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.impa.gad"},
	{0x0BCD5585C8B9E914ULL, "vxw_bonepitdestroyed_a3_19l_0_0.vlm"},
	{0x0BCDC6E8B05534E2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\wkf.ani"},
	{0x0BCE644D94E14FFEULL, "qun110_petrice6042287.cl"},
	{0x0BCEA9DFF85C42EEULL, "ave110tr_donnic_bandits.utt"},
	{0x0BCED434F04B8CD6ULL, "vxlt_0_lowtown_a2_n_c_hua_0_0.vlm"},
	{0x0BCF5F4DD58495EDULL, "924997889.ogg"},
	{0x0BCF845537FDAFB4ULL, "12321380.ogg"},
	{0x0BCFCA253C994F37ULL, "6025338_m.fxe"},
	{0x0BCFDE2E7DC996C5ULL, "one300_zevran6076956.cl"},
	{0x0BD0196F96167FF1ULL, "hlac_w_sundermount_65536_0y.dds"},
	{0x0BD0928B484275C8ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screenb.phy"},
	{0x0BD09A804BF4DC9CULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_die_v1.ani"},
	{0x0BD0CFDFF896C4ADULL, "jh.cut"},
	{0x0BD117E031FEC54AULL, "w_bonepitmining_a2_186_0.mao"},
	{0x0BD11BE031FECB99ULL, "w_bonepitmining_a2_186_0.mmh"},
	{0x0BD125E031FEDC87ULL, "w_bonepitmining_a2_186_0.msh"},
	{0x0BD165E8534AABC9ULL, "llac_ht_0_hightown_0_0.dds"},
	{0x0BD1C2F00BF8FA2DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1_tool.msh"},
	{0x0BD1D0F00BF91260ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1_tool.mao"},
	{0x0BD1D4F00BF918B3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1_tool.mmh"},
	{0x0BD266F50C333C91ULL, "6083072_f.fxe"},
	{0x0BD2731ECF6E2F23ULL, "vxw_encounter2_it_0_0.vlm"},
	{0x0BD2BD31BB73279CULL, "6155979_m.fxe"},
	{0x0BD2EDCBF53D2C10ULL, "668140735.ogg"},
	{0x0BD2FDAC1DD54888ULL, "6191248_f.fxe"},
	{0x0BD343128A955E5BULL, "vxht_chantry_n_30r_0_0.vlm"},
	{0x0BD35E3F1DE4B10BULL, "vxht_0_hightown_a3_cf2_0_0.vlm"},
	{0x0BD35E7A64829510ULL, "6150976_m.fxe"},
	{0x0BD3746A1129675EULL, "vxca_caverns2prc_8bg_0_0.vlm"},
	{0x0BD37E950831723CULL, "dae011_gossips6143025.cl"},
	{0x0BD3D23B1393B8CCULL, "6208647_f.fxe"},
	{0x0BD3E17804907F3DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv2.gad"},
	{0x0BD423DCCBF7CF56ULL, "vxlt_hangedman_n_20p_0_0.vlm"},
	{0x0BD42B033F35D5DDULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.impa.evt"},
	{0x0BD4D0B99EF25A34ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_5y.dds"},
	{0x0BD538AFD806A6D7ULL, "497449884.ogg"},
	{0x0BD6182BE59A7C20ULL, "6028514_m.fxe"},
	{0x0BD61C9EA3E57486ULL, "vxht_brothel_n_5iq_0_0.vlm"},
	{0x0BD6AD76FA3C1C8FULL, "177606026.ogg"},
	{0x0BD6F620134294ECULL, "24349950.ogg"},
	{0x0BD7517C1772D19AULL, "qun230tr_contact_ambient.ncs"},
	{0x0BD7A7804BFB11DEULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_die_v1.gad"},
	{0x0BD7C51384C51E62ULL, "hlac_ht_0_hightown_a3_65533_2z.dds"},
	{0x0BD7E6DED50BBC6AULL, "6054511_m.fxe"},
	{0x0BD81E8165DE0AEFULL, "6114856_m.fxe"},
	{0x0BD8365BB13A6C72ULL, "6110062_m.fxe"},
	{0x0BD83A148E14F34FULL, "vxlt_0_lowtown_a2_n_dej_0_0.vlm"},
	{0x0BD86C5A905590D9ULL, "vxlt_undercitydungeon_1ir_0_0.vlm"},
	{0x0BD8774595588B8DULL, "follower_banter6156239.cl"},
	{0x0BD893F264375C12ULL, "hlac_dr_0_deeproads_5845_13x.dds"},
	{0x0BD8BFF46A794936ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_cobble_n.dds"},
	{0x0BD8E89D1F7008DEULL, "vxw_encounter2_camp_mq_0_0.vlm"},
	{0x0BD8F84C514BD0D0ULL, "651977225.ogg"},
	{0x0BD90A8EA8B3E6FEULL, "llac_ht_keep_n_3731_2.dds"},
	{0x0BD96A6A7DEA1DEBULL, "vxht_mansion2b_1pw_0_0.vlm"},
	{0x0BDA0F5C947FF55CULL, "vxca_caverns1_mines_huk_0_0.vlm"},
	{0x0BDA47E20E6AACD0ULL, "784278604.ogg"},
	{0x0BDA99BBBB8B1922ULL, "vxw_bonepit_1sh_0_0.vlm"},
	{0x0BDB51BA118E193BULL, "abi_m_glyph_repulsion.bnk"},
	{0x0BDB5628C3293DE3ULL, "758374841.ogg"},
	{0x0BDBA82B834E24F6ULL, "zz_dog_debug6131978.cl"},
	{0x0BDC41BC07300446ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_wic.phy"},
	{0x0BDC6CF54B57A14AULL, "mag110_arianni6016827.cl"},
	{0x0BDD4611E241A556ULL, "vxlt_0_lowtown_a2_n_c_hpb_0_0.vlm"},
	{0x0BDE097463F5ED75ULL, "llac_ca_caverns2_5802_24.dds"},
	{0x0BDE12477BCBCE0AULL, "hlac_w_sundermount_-65536_0y.dds"},
	{0x0BDE82804C00F1FDULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_die_v1.evt"},
	{0x0BDEC374ABB8CA47ULL, "291488537.ogg"},
	{0x0BDEF4F25CC5F063ULL, "mag360_thrask6046979.cl"},
	{0x0BE07B4BA15A6FC2ULL, "479717577.ogg"},
	{0x0BE0CC2D7D50C4C6ULL, "art\\vfx\\combat\\vfx_d_teleport_start_p.anb"},
	{0x0BE0F0763E3504E0ULL, "lt_0_lowtown_n_21_0_ll.mmh"},
	{0x0BE13BBECEBDFFA7ULL, "t1_mue_pu3.tnt"},
	{0x0BE178EAE4B39C4EULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\disengage.ani"},
	{0x0BE1DCA2C2D8A915ULL, "175807580.ogg"},
	{0x0BE25865470019BCULL, "848630544.ogg"},
	{0x0BE2C19B58EB9E45ULL, "vxca_caverns2_62q_0_2.vlm"},
	{0x0BE2CA75D3C7F50AULL, "gen00fl_merrill6142083.cl"},
	{0x0BE2E9C9B2BD1403ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_sworda.msh"},
	{0x0BE2F7C9B2BD2C5DULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_sworda.mmh"},
	{0x0BE2FBC9B2BD336EULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_sworda.mao"},
	{0x0BE368506C2590E4ULL, "vxlt_gamlenshouse_n_b6_0_0.vlm"},
	{0x0BE3EBA5F469B985ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc107a_0n.dds"},
	{0x0BE41D8F2BE9DD28ULL, "117785788.ogg"},
	{0x0BE431604A5884CCULL, "155335925.ogg"},
	{0x0BE4501279EC7314ULL, "6177728_m.fxe"},
	{0x0BE59CDFF86F0B17ULL, "ave110tr_donnic_bandits.ncs"},
	{0x0BE5A3ECDCCA821FULL, "hlac_lt_gallowstemplar_n_2394_2x.dds"},
	{0x0BE5BE706E170F6DULL, "mag100_karras6052953.cl"},
	{0x0BE5EE6CD8DF2C22ULL, "vxw_woundedcoast_ts_0_0.vlm"},
	{0x0BE5FBDE1CABB466ULL, "gen00flst_beth_carv_will.stg"},
	{0x0BE668F6ECD0A4FBULL, "vxht_0_hightown_a2_c6o_0_0.vlm"},
	{0x0BE68C75D7FA4855ULL, "208571404.ogg"},
	{0x0BE6905E57D4236FULL, "6113956_m.fxe"},
	{0x0BE6A0C3CACFC452ULL, "683282465.ogg"},
	{0x0BE6CB316E0213BFULL, "vxht_mansion_n_6kb_0_0.vlm"},
	{0x0BE6DE37824BA959ULL, "vxlt_undercitydungeon_1lq_0_0.vlm"},
	{0x0BE6FE1598BD2F92ULL, "317845451.ogg"},
	{0x0BE75A7A1166215EULL, "vxw_encounter2_camp_hr_0_0.vlm"},
	{0x0BE79FF3C30F2896ULL, "6054865_m.fxe"},
	{0x0BE80F21C91C3EF7ULL, "514691210.ogg"},
	{0x0BE84604D1B28736ULL, "vxca_caverns1_mines_jgi_0_0.vlm"},
	{0x0BE87705E3CA161DULL, "zz_qun_debug6028891.cl"},
	{0x0BE87E644E876FF5ULL, "t3_wdg_gw1.tnt"},
	{0x0BE8D35549AED866ULL, "vxlt_undercitydungeon_xw_0_0.vlm"},
	{0x0BE8F9A301ECBDDFULL, "zz_leaphit.cut"},
	{0x0BE92544A9372EA1ULL, "dae210_gate_guard6085122.cl"},
	{0x0BE978C2B2F5E2ACULL, "6119096_m.fxe"},
	{0x0BE990D5B8CBD8D4ULL, "llac_lt_backalley_g_65537_8.dds"},
	{0x0BE9DF670C6C0A6CULL, "dae000_bartrand6064061.cl"},
	{0x0BEA4101EC796802ULL, "gen00fl_sebastian6189023.cl"},
	{0x0BEB32338782CAC6ULL, "one100it_fang.uti"},
	{0x0BEBA2891783FB14ULL, "vxca_cavern1_haunted_ikh_0_0.vlm"},
	{0x0BEBAA2EFC4097C9ULL, "vxlt_0_lowtown_docks_a3_n_cxf_0_0.vlm"},
	{0x0BEC82985F3C3BA5ULL, "mag370_meredith6046906.cl"},
	{0x0BECCDC15019BB51ULL, "lt_0_lowtown_docks_n_222_0.msh"},
	{0x0BECDBC15019D36CULL, "lt_0_lowtown_docks_n_222_0.mao"},
	{0x0BECE7C15019E70FULL, "lt_0_lowtown_docks_n_222_0.mmh"},
	{0x0BED077AD79CC44DULL, "195736837.ogg"},
	{0x0BED0C28FCFDC373ULL, "qun231_party_comment6077130.cl"},
	{0x0BED181F2C0E8084ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_act_c.mmh"},
	{0x0BED353F26C9DE61ULL, "1072076198.ogg"},
	{0x0BED4FB216085B6EULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl_c.impa.evt"},
	{0x0BED5AC7073CE814ULL, "6067115_m.fxe"},
	{0x0BED678B6334034FULL, "ctplt_rock.mao"},
	{0x0BED797DD6AD1400ULL, "6134941_m.fxe"},
	{0x0BEDCF2D7D5B5471ULL, "art\\vfx\\combat\\vfx_d_teleport_start_p.mmh"},
	{0x0BEDE7259910282EULL, "vxht_keep_n_6i5_0_0.vlm"},
	{0x0BEDF48E2BC3CB74ULL, "vxlt_undercityencounter_1c1_0_1.vlm"},
	{0x0BEE1E06BB182D40ULL, "vxht_keep_61l_0_0.vlm"},
	{0x0BEF5BF609CBF6F9ULL, "one280_iwan6096752.cl"},
	{0x0BEF70EAE4BFCB9FULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\disengage.evt"},
	{0x0BEFA005E90B1D0CULL, "llac_lt_gallowscourtyard_65534_1.dds"},
	{0x0BF00256EAFB55BCULL, "6084387_m.fxe"},
	{0x0BF03BA28BB8E2E8ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.anb"},
	{0x0BF0746DBDBDA180ULL, "197504679.ogg"},
	{0x0BF0ACA96A9A220CULL, "gen00fl_aveline6128413.cl"},
	{0x0BF0FE9C5CD352DBULL, "vxca_cavern1_haunted_j5f_0_0.vlm"},
	{0x0BF12F86FFDECD71ULL, "6133668_m.fxe"},
	{0x0BF189C07012C504ULL, "cd_mrl250_pol.wbd"},
	{0x0BF1A056FDEF7ECEULL, "dae000_arishok6067708.cl"},
	{0x0BF1A846D21AD047ULL, "hlac_lt_0_lowtown_docks_n_-131074_1x.dds"},
	{0x0BF1BDC882B9766CULL, "gen_im_trs_qun_hvy_hlm_r0.uti"},
	{0x0BF1E658C857376EULL, "hlac_lt_0_lowtown_a2_n_65537_2z.dds"},
	{0x0BF20E36577A3467ULL, "vxht_0_hightown_a3_c39_0_0.vlm"},
	{0x0BF27BE8B800BE88ULL, "6067137_m.fxe"},
	{0x0BF33D77AD2F9862ULL, "vxlt_warehouse_n_ug_0_0.vlm"},
	{0x0BF346B2160CB719ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl_c.impa.gad"},
	{0x0BF3953DA0743891ULL, "gen_im_trs_dwv_junk3_04.uti"},
	{0x0BF3E186C9D17146ULL, "dae000_bodhan6129126.cl"},
	{0x0BF40CB6DBCBDC9DULL, "110144642.ogg"},
	{0x0BF4CBE430AC8FC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_loop.gad"},
	{0x0BF5106616018296ULL, "14498824.ogg"},
	{0x0BF53D0E5FF82C3FULL, "vxht_mansion_n_6nm_0_0.vlm"},
	{0x0BF550147441C1D9ULL, "vxlt_undercitydungeon_1wp_0_0.vlm"},
	{0x0BF57D2AA71DD486ULL, "lgt110_gxa_vgk_silver_order6117025.cl"},
	{0x0BF58BCC93FC3860ULL, "vxht_0_hightown_a88_0_0.vlm"},
	{0x0BF58CC01683894CULL, "6127325_m.fxe"},
	{0x0BF5E9F80BE83BE9ULL, "art\\vfx\\materials\\chunky\\vfx_c_falloff_invert_blend.mao"},
	{0x0BF633AA0DC8F20EULL, "w_blightlands_75_0.phy"},
	{0x0BF651EAE4C5B588ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\disengage.gad"},
	{0x0BF67EB7306D42A7ULL, "6103631_f.fxe"},
	{0x0BF7DD032B8A9D84ULL, "206073554.ogg"},
	{0x0BF86A5FA65B39EDULL, "7471020.ogg"},
	{0x0BF8FDDA2C9BA7DAULL, "vxlt_gallowstemplar_f_43j_0_0.vlm"},
	{0x0BF9F82DF4D85DAEULL, "vxlt_0_lowtown_a2_en0_0_0.vlm"},
	{0x0BFA11E1C4535F65ULL, "259909773.ogg"},
	{0x0BFA20731D7B6308ULL, "6041538_f.fxe"},
	{0x0BFA265BDD55AC1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction.evt"},
	{0x0BFA56A8A2E73B9BULL, "l_env_wal_hightown_int_calm_p2p.wbd"},
	{0x0BFA57B21612F367ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl_c.impa.ani"},
	{0x0BFA8C8B1A423705ULL, "6135770_m.fxe"},
	{0x0BFACEF5FA697EF8ULL, "232094725.ogg"},
	{0x0BFB2697B60C653AULL, "gen_im_trs_krk_hvy_glv_r0.uti"},
	{0x0BFBDB95EA2A622CULL, "vxht_0_hightown_a2_9zb_0_0.vlm"},
	{0x0BFBEEE430B2E9B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_loop.evt"},
	{0x0BFC998229A3E37CULL, "hlac_lt_darktown_-1_68z.dds"},
	{0x0BFCE965DAA79102ULL, "6036951_m.fxe"},
	{0x0BFCF31357D3414FULL, "gen00fl_aveline6120863.cl"},
	{0x0BFD49AE90CE912AULL, "hlac_lt_undercityserialkiller_-1_41y.dds"},
	{0x0BFD57EA04CC16B4ULL, "gen_im_trs_rdr_hvy_hlm_r0.uti"},
	{0x0BFE585DF94E35D5ULL, "6077090_m.fxe"},
	{0x0BFF01E1FD506334ULL, "vxht_mansion2_3ei_0_0.vlm"},
	{0x0C001E52C9073DA5ULL, "706737645.ogg"},
	{0x0C0036B01D82E130ULL, "vxht_0_hightown_a3_c54_0_0.vlm"},
	{0x0C003D5BDD5A3F49ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction.gad"},
	{0x0C00801349704CE7ULL, "vxht_0_hightown_a3_c48_0_0.vlm"},
	{0x0C00FE99DE3AC662ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamb.msh"},
	{0x0C010099DE3AC918ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamb.mmh"},
	{0x0C010C99DE3ADDBBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamb.mao"},
	{0x0C01614C5B496A58ULL, "307117832.ogg"},
	{0x0C01C04A0C7147D2ULL, "vxca_caverns1_mines_imx_0_0.vlm"},
	{0x0C01E239C2DFC51CULL, "6122146_m.fxe"},
	{0x0C01E8CADBA6848FULL, "vxht_0_hightown_a3_b8e_0_0.vlm"},
	{0x0C027E9634F40604ULL, "zz_arshcockforestcu_v2.cut"},
	{0x0C028E73C783D9AEULL, "166053379.ogg"},
	{0x0C02B3915900E807ULL, "character_character_p167.psh.fxo.dx11"},
	{0x0C0306BE6E042274ULL, "follower_banter6149175.cl"},
	{0x0C032C466471DF08ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml103a_0n.dds"},
	{0x0C034CB0D0BCD5FBULL, "vxht_0_hightown_a2_c8a_0_0.vlm"},
	{0x0C036D293D6FCE86ULL, "116951836.ogg"},
	{0x0C039D91CE2725E5ULL, "mag110_vincento6017904.cl"},
	{0x0C03DBA69999ABE1ULL, "gen_im_cns_tom_rog_assassin.uti"},
	{0x0C040FCEF82C813DULL, "art\\characters\\weapons\\dagger\\w_dgr_103a.mao"},
	{0x0C0452E6F2A5D434ULL, "llac_ca_caverns1_mines_24421_5.dds"},
	{0x0C04C00BCB6780EEULL, "dae000ip_crowd_ff6_talk18.utp"},
	{0x0C04C6E5985EE830ULL, "572362461.ogg"},
	{0x0C04F3399356C394ULL, "art\\characters\\creatures\\spider\\animation\\cwkb.ani"},
	{0x0C054D5C2ABA965AULL, "vxlt_gallowstemplar_n_4tw_0_0.vlm"},
	{0x0C05F1E1BB6ADF54ULL, "vxlt_0_lowtown_a2_n_eje_0_0.vlm"},
	{0x0C06A722751405F5ULL, "6061769_m.fxe"},
	{0x0C070E5BDD600D97ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction.ani"},
	{0x0C071450C642D169ULL, "6164772_m.fxe"},
	{0x0C072A7F38C717AEULL, "vxdr_dungeon_20z_0_0.vlm"},
	{0x0C0826D86709677CULL, "vxlt_0_lowtown_a2_n_c_ht4_0_0.vlm"},
	{0x0C088CC5E060B9BEULL, "6199281_m.fxe"},
	{0x0C08F6E430BD81A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_loop.ani"},
	{0x0C0981F9BD61CC3DULL, "6123395_m.fxe"},
	{0x0C09BF4D47782FE0ULL, "vxlt_gallowstemplar_3mz_0_0.vlm"},
	{0x0C0A3FB6BA8CB626ULL, "t1_skn_006.tnt"},
	{0x0C0A574A15C5B771ULL, "6064519_m.fxe"},
	{0x0C0A79B8AC9EA829ULL, "vxlt_undercityserialkiller_1a7_0_0.vlm"},
	{0x0C0B5FFA91D76C0EULL, "hlac_w_bonepit_458754_0y.dds"},
	{0x0C0BBC729D6B799DULL, "vxca_caverns1_j59_0_0.vlm"},
	{0x0C0BCD2769460EC2ULL, "571770070.ogg"},
	{0x0C0BD85179FA1F57ULL, "6067758_m.fxe"},
	{0x0C0BEA42306182F0ULL, "vxw_bonepitmining_a2_ow_0_0.vlm"},
	{0x0C0BF039935CDDF6ULL, "art\\characters\\creatures\\spider\\animation\\cwkb.gad"},
	{0x0C0C2FAD1F42E69EULL, "848808727.ogg"},
	{0x0C0C3CAB53F43CA6ULL, "llac_lt_0_lowtown_a2_n_c_1_7.dds"},
	{0x0C0CAC2579371547ULL, "gen_im_npc_arm_hum_cht_slva.uti"},
	{0x0C0D5B3CD3A68453ULL, "6016589_m.fxe"},
	{0x0C0D828A2547DCBFULL, "gen00fl_sebastian6174334.cl"},
	{0x0C0E239FCAE52B67ULL, "458411958.ogg"},
	{0x0C0E55DA79E9DAD1ULL, "vxht_0_hightown_n_c3p_0_1.vlm"},
	{0x0C0E83614022B331ULL, "486835061.ogg"},
	{0x0C0EB4832C6A734CULL, "6072654_m.fxe"},
	{0x0C0EDB87F8BA2D10ULL, "gen00fl_sebastian6180840.cl"},
	{0x0C0F36177E263992ULL, "fer_ter_grass_001_diff.dds"},
	{0x0C0F69F6547E06F2ULL, "524344053.ogg"},
	{0x0C0F74EB6D031880ULL, "622151340.ogg"},
	{0x0C0FE9F8C78F2B9BULL, "w_woundedcoast_220_0.mmh"},
	{0x0C0FEA958AC793A9ULL, "art\\characters\\weapons\\dagger\\w_dgr_002a.phy"},
	{0x0C0FF5F8C78F4038ULL, "w_woundedcoast_220_0.mao"},
	{0x0C1007F8C78F5E85ULL, "w_woundedcoast_220_0.msh"},
	{0x0C103226FE676052ULL, "vxca_caverns1_mines_ijy_0_0.vlm"},
	{0x0C109C75ACB507AEULL, "mcr371cr_harvester.ncs"},
	{0x0C10B0E623232ABFULL, "676169642.ogg"},
	{0x0C10EBEDFAD1FF10ULL, "549455327.ogg"},
	{0x0C115BC16FE3D616ULL, "lt_0_lowtown_n_53_0_lc.mmh"},
	{0x0C1226D2FBC4D0CEULL, "6088226_m.fxe"},
	{0x0C1228512DEDB0BDULL, "fen110_fenris6118039.cl"},
	{0x0C1234D76DEF4511ULL, "gen00fl_anders6119408.cl"},
	{0x0C12450F60C4F294ULL, "vxca_caverns2_5aa_0_1.vlm"},
	{0x0C12C0EB2DBBD8EFULL, "664819166.ogg"},
	{0x0C12D1366E693754ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpb.phy"},
	{0x0C130B3993632A55ULL, "art\\characters\\creatures\\spider\\animation\\cwkb.evt"},
	{0x0C133B4B939BD524ULL, "6207814_m.fxe"},
	{0x0C133DD0F402D28CULL, "vxlt_0_lowtown_a2_n_c_fj4_0_0.vlm"},
	{0x0C137B19175E007BULL, "6175093_m.fxe"},
	{0x0C139B9BA794D0B6ULL, "vxca_caverns1_mines_jtl_0_0.vlm"},
	{0x0C13A80B06B118B5ULL, "dre000_vampire6034334.cl"},
	{0x0C140A2765CAB0D3ULL, "255326345.ogg"},
	{0x0C144E61FE57B951ULL, "vxht_0_hightown_c9t_0_0.vlm"},
	{0x0C1468AAAEB6A22AULL, "vxlt_0_lowtown_ewk_0_0.vlm"},
	{0x0C146E30CAE8A8B8ULL, "w_bonepitdestroyed_a3_27_0.rml"},
	{0x0C149FFE56EBF720ULL, "6090345_f.fxe"},
	{0x0C14B09A351D364FULL, "6093627_m.fxe"},
	{0x0C14D2E274A1F9EFULL, "6072569_m.fxe"},
	{0x0C151B1F2C2FD87BULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_act_c.anb"},
	{0x0C159B752F90A163ULL, "vxlt_darktown_67v_0_0.vlm"},
	{0x0C160DD2650CE779ULL, "vfx_vfx_p8.vsh.fxo.dx11.assembly"},
	{0x0C169180F7773F85ULL, "vfx_chunkyparticle_p4.psh.fxo.dx11.assembly"},
	{0x0C16FA4786A6F01BULL, "vxlt_0_lowtown_a2_n_bxg_0_0.vlm"},
	{0x0C17F3D6AF57E4BDULL, "6128295_m.fxe"},
	{0x0C18BDA1321E42B1ULL, "vxlt_0_lowtown_a2_gvi_0_0.vlm"},
	{0x0C1932B1F6C8E841ULL, "sky_dr_0_deeproads.mao"},
	{0x0C193AA28BDBE583ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.mmh"},
	{0x0C19A6A968FA9AA2ULL, "gen00fl_sebastian6195848.cl"},
	{0x0C19C03F6CD98C48ULL, "head_flip_rev.dds"},
	{0x0C1A0A2574C61A97ULL, "art\\vfx\\creatures\\vfx_c_hd_fswipe_c.impa.gad"},
	{0x0C1A111E7503D445ULL, "486172654.ogg"},
	{0x0C1A5363D756EDCBULL, "6108098_f.fxe"},
	{0x0C1A5C1F22579B70ULL, "vxw_bonepitmining_a2_jt_0_0.vlm"},
	{0x0C1A9F0D30D30185ULL, "6118771_m.fxe"},
	{0x0C1AAC6617739885ULL, "hlac_ht_keep_3732_19y.dds"},
	{0x0C1BF64F70869312ULL, "vxlt_foundry_n_12s_0_0.vlm"},
	{0x0C1BF67A1060D3E0ULL, "qun160_vashothlead6053911.cl"},
	{0x0C1C5078500C0F28ULL, "follower_banter6156254.cl"},
	{0x0C1C6120404A39FEULL, "gen00fl_aveline6110489.cl"},
	{0x0C1CC7B76BDFF351ULL, "vxht_0_hightown_n_c8q_0_1.vlm"},
	{0x0C1D98342CB235CEULL, "art\\vfx\\combat\\vfx_itm_myt_favor_aoe_p.mmh"},
	{0x0C1D9B350B110E2AULL, "6155595_m.fxe"},
	{0x0C1DAE366E730351ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpb.msh"},
	{0x0C1DB8C6DD435330ULL, "6134848_m.fxe"},
	{0x0C1DC8366E732F0FULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpb.mmh"},
	{0x0C1E131EE69D1E49ULL, "371708524.ogg"},
	{0x0C1E464DF59BD1A4ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightstreaks.mmh"},
	{0x0C1EA403F05D78D2ULL, "vxca_caverns1_mines_igz_0_0.vlm"},
	{0x0C1F04596A749DA4ULL, "6092612_m.fxe"},
	{0x0C1F7A1B407BF725ULL, "6118973_m.fxe"},
	{0x0C2038ABC83FD4D3ULL, "vxht_mansion2b_31n_0_0.vlm"},
	{0x0C20694C62991FBFULL, "hlac_ht_mansion2b_818_33y.dds"},
	{0x0C2098AFEDBAE94EULL, "6083326_m.fxe"},
	{0x0C211F2574CC5D24ULL, "art\\vfx\\creatures\\vfx_c_hd_fswipe_c.impa.evt"},
	{0x0C2158F97F1472A8ULL, "mcr371cr_merrill_cut.utc"},
	{0x0C2172E1828E285EULL, "m_isa121ar_chantry.bnk"},
	{0x0C2194568F930AF6ULL, "vxlt_undercityencounter_1dw_0_0.vlm"},
	{0x0C219F6AE6E939BAULL, "docks_male_elven_worker_sta6175792.cl"},
	{0x0C21C2D4326B74EFULL, "hlac_lt_0_lowtown_docks_a3_131071_6y.dds"},
	{0x0C223FBA11C9FD39ULL, "abi_m_glyph_repulsion.wbd"},
	{0x0C22835C4EBA390AULL, "264106326.ogg"},
	{0x0C22D30C755993A2ULL, "vxw_bonepit_1hu_0_0.vlm"},
	{0x0C2457C74229C23BULL, "6169711_m.fxe"},
	{0x0C246026D5AF1225ULL, "204312777.ogg"},
	{0x0C247D819409EF4BULL, "6152582_m.fxe"},
	{0x0C248994DF335F86ULL, "one100dg_lia6013849.cl"},
	{0x0C251C1469F27D46ULL, "6198708_f.fxe"},
	{0x0C265940A67A5A23ULL, "vxht_brothel_n_6eg_0_0.vlm"},
	{0x0C272238EBB64FF5ULL, "6088672_m.fxe"},
	{0x0C272F7E24145B31ULL, "vxlt_0_lowtown_a2_gmh_0_0.vlm"},
	{0x0C27309193D1A6E7ULL, "hlac_dr_primevaltomb_659_15y.dds"},
	{0x0C275E9CD609A7E3ULL, "342187277.ogg"},
	{0x0C27831E8141C236ULL, "6177861_m.fxe"},
	{0x0C27A6C0ECF06B98ULL, "vxlt_0_lowtown_n_gkx_0_0.vlm"},
	{0x0C27BFF8B4766682ULL, "ht_0_hightown_a2_ll.mmh"},
	{0x0C27ECE415A3A267ULL, "mag311_emile_de_launcet6043701.cl"},
	{0x0C28130DE10081CFULL, "hlac_dr_0_deeproads_5840_2z.dds"},
	{0x0C28873A7A418D74ULL, "vxca_caverns2_8b8_0_0.vlm"},
	{0x0C2890143FFD9FFFULL, "hlac_ht_0_hightown_n_-65534_0y.dds"},
	{0x0C2910EA22C91A05ULL, "6117671_m.fxe"},
	{0x0C2971A8A576F6A3ULL, "vxw_blightlands_7q_0_0.vlm"},
	{0x0C298D8DF8AECF64ULL, "6057325_m.fxe"},
	{0x0C29B96189947352ULL, "vxht_mansion2b_31k_0_0.vlm"},
	{0x0C29CF8B478FAAF3ULL, "gen00fl_fenris6127096.cl"},
	{0x0C2A776354A356E7ULL, "6041573_f.fxe"},
	{0x0C2A7ECABFB464AAULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.anb"},
	{0x0C2AB7342CBCF4CDULL, "art\\vfx\\combat\\vfx_itm_myt_favor_aoe_p.anb"},
	{0x0C2AE56527DF9354ULL, "l_dae421ar_player_home_nt.bnk"},
	{0x0C2AF303276AED17ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack01.gad"},
	{0x0C2B455E5808EB3CULL, "6106658_m.fxe"},
	{0x0C2B505A2A0A81CDULL, "gen00fl_varric6118759.cl"},
	{0x0C2B5B7D6136D74CULL, "gen00fl_isabela6135595.cl"},
	{0x0C2C6C384BD8018AULL, "6135849_m.fxe"},
	{0x0C2C82DD6C209DE9ULL, "vxlt_gallowstemplar_f_5yb_0_0.vlm"},
	{0x0C2E372574D7101DULL, "art\\vfx\\creatures\\vfx_c_hd_fswipe_c.impa.ani"},
	{0x0C2EF8720D64386FULL, "vxht_keep_n_6fa_0_0.vlm"},
	{0x0C2F180532FC77F1ULL, "gen00fl_aveline6114295.cl"},
	{0x0C2F204A0F56DAC4ULL, "6185387_m.fxe"},
	{0x0C2FBE1C72ACA1D3ULL, "vxlt_gallowstemplar_6sn_0_0.vlm"},
	{0x0C2FCF526490017EULL, "vxht_brothel_n_53g_0_0.vlm"},
	{0x0C31183A6874926EULL, "gen_im_trs_slv_sta_ele_r1.uti"},
	{0x0C31766064C3AADAULL, "548440124.ogg"},
	{0x0C32080327712FA4ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack01.evt"},
	{0x0C320871E83741EAULL, "vxlt_undercityserialkiller_1ju_0_0.vlm"},
	{0x0C32585F51AED2E8ULL, "vxlt_gallowstemplar_653_0_0.vlm"},
	{0x0C326C30CB01A295ULL, "w_bonepitdestroyed_a3_27_0.msh"},
	{0x0C326E30CB01A60BULL, "w_bonepitdestroyed_a3_27_0.mmh"},
	{0x0C327A30CB01BAA8ULL, "w_bonepitdestroyed_a3_27_0.mao"},
	{0x0C32D6016865724DULL, "lt_0_lowtown_docks_n_158_0.msh"},
	{0x0C32E40168658980ULL, "lt_0_lowtown_docks_n_158_0.mao"},
	{0x0C32E801686590D3ULL, "lt_0_lowtown_docks_n_158_0.mmh"},
	{0x0C32F12C1983E295ULL, "fx_circle_distor.mao"},
	{0x0C331200ECE3E03FULL, "dae000_orsino6081382.cl"},
	{0x0C331CE032519786ULL, "w_bonepitmining_a2_186_0.phy"},
	{0x0C3371D8AAB6E48BULL, "141445750.ogg"},
	{0x0C33B15E3927C82BULL, "303562853.ogg"},
	{0x0C33C9299638B892ULL, "docks_male_human_worker_sta6175960.cl"},
	{0x0C340B8DD5B702D1ULL, "llac_ht_0_hightown_a3_2_1.dds"},
	{0x0C344F8414BF5B91ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_masd.mao"},
	{0x0C34772E882E45C7ULL, "304234957.ogg"},
	{0x0C34A2AF1C4E9C97ULL, "6116003_m.fxe"},
	{0x0C359415DDAC6875ULL, "6083772_m.fxe"},
	{0x0C3604F1A2537451ULL, "llac_lt_gallowsprison_n_-1_15.dds"},
	{0x0C362095CB4D81B9ULL, "gen00fl_varric6105476.cl"},
	{0x0C3624BD1C7DE33AULL, "478888072.ogg"},
	{0x0C36351B03215B0CULL, "881053287.ogg"},
	{0x0C36DD91DB8ADD54ULL, "165834481.ogg"},
	{0x0C3732AA0E0087D1ULL, "w_blightlands_75_0.mmh"},
	{0x0C373EAA0E009C72ULL, "w_blightlands_75_0.mao"},
	{0x0C374CAA0E00B38FULL, "w_blightlands_75_0.msh"},
	{0x0C3782C714BF3285ULL, "6112571_m.fxe"},
	{0x0C3799FADB3736DFULL, "vxw_woundedcoast_1d3_0_0.vlm"},
	{0x0C37B63097CD0E3BULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedenter.ani"},
	{0x0C37E385976D0F23ULL, "vxw_blightlands_8p_0_0.vlm"},
	{0x0C38F1D53E1527E2ULL, "rdr111dg_martin6054582.cl"},
	{0x0C3933752B34652AULL, "vxca_cave_1ai_0_0.vlm"},
	{0x0C393ED234014052ULL, "vxlt_gallowstemplar_6sk_0_0.vlm"},
	{0x0C3957EE5282E3A8ULL, "693260994.ogg"},
	{0x0C39F3325768743BULL, "3300113.ogg"},
	{0x0C3A8B697B206688ULL, "6055060_m.fxe"},
	{0x0C3B80C32D8B68D4ULL, "vxca_caverns1_mines_i9r_0_0.vlm"},
	{0x0C3B87BDD449A9D2ULL, "vxca_caverns1_mines_iyd_0_0.vlm"},
	{0x0C3BD3958AECA7CEULL, "art\\characters\\weapons\\dagger\\w_dgr_002a.mmh"},
	{0x0C3BDF958AECBC65ULL, "art\\characters\\weapons\\dagger\\w_dgr_002a.mao"},
	{0x0C3BDFA4DE39C068ULL, "follower_banter6167997.cl"},
	{0x0C3BED958AECD3BCULL, "art\\characters\\weapons\\dagger\\w_dgr_002a.msh"},
	{0x0C3C5833374B9376ULL, "659475424.ogg"},
	{0x0C3C5EFC1A580873ULL, "vxlt_0_lowtown_docks_n_b4f_0_0.vlm"},
	{0x0C3C7C1E9EF3571FULL, "gen00fl_anders6125885.cl"},
	{0x0C3C817675042635ULL, "hlac_lt_backalley_n_0_1y.dds"},
	{0x0C3D234ED237BF0CULL, "llac_ht_0_hightown_n_65537_0.dds"},
	{0x0C3D724C0FF77EFCULL, "zz_followers_banter6159165.cl"},
	{0x0C3D9C40D9A3314CULL, "6171076_m.fxe"},
	{0x0C3DBEDDCBB46F2EULL, "6125581_m.fxe"},
	{0x0C3E4E602087932AULL, "6125426_m.fxe"},
	{0x0C3E9064EA5BD39FULL, "combat_smaoe_r2_a3.gda"},
	{0x0C3EAAD1FF356C98ULL, "slime.mao"},
	{0x0C3EBCFDBDCB305AULL, "hlac_lt_0_lowtown_docks_a3_65533_2y.dds"},
	{0x0C3EC0C2E85CF875ULL, "and111_lirene6044962.cl"},
	{0x0C3EC25AD5A37994ULL, "vxdr_primevaltomb_85s_0_0.vlm"},
	{0x0C3F2003277BE29DULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack01.ani"},
	{0x0C3F367CF6F90C55ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamloop_10s.mmh"},
	{0x0C3FA612819BB25FULL, "hlac_ht_0_hightown_n_65537_1z.dds"},
	{0x0C404289D967A216ULL, "follower_banter6173129.cl"},
	{0x0C405DD090C2EEFDULL, "hlac_lt_0_lowtown_n_65538_4x.dds"},
	{0x0C41308470023E95ULL, "6062791_m.fxe"},
	{0x0C41994737D612B5ULL, "mcr310_sibling6076346.cl"},
	{0x0C41B55CE34054A2ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_4z.dds"},
	{0x0C424EB280F18D52ULL, "6169332_m.fxe"},
	{0x0C42B8DEB4184807ULL, "lgt000ip_ultra_mage_remains.ncs"},
	{0x0C42CCD30CE011CCULL, "1007726605.ogg"},
	{0x0C42D6472B5BDEB3ULL, "86702611.ogg"},
	{0x0C42E75F84DE3926ULL, "zz_mcr_debug6080861.cl"},
	{0x0C431311013699E7ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_006a_0t.dds"},
	{0x0C433CFA8A668B23ULL, "vxht_brothel_n_6ki_0_0.vlm"},
	{0x0C435450D1CA4DFAULL, "gen00fl_isabela6135658.cl"},
	{0x0C43F7CABE981856ULL, "368979431.ogg"},
	{0x0C441713A7D55B29ULL, "car100pt_gift.ncs"},
	{0x0C4437B3FBE7EB6BULL, "6076507_m.fxe"},
	{0x0C44454B63D912B2ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_anders_chant_ex2.gad"},
	{0x0C446D8480021750ULL, "168044928.ogg"},
	{0x0C44BE3097D7A632ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedenter.evt"},
	{0x0C45367FA1D43F8FULL, "728832249.ogg"},
	{0x0C459A866FB694ABULL, "gen00fl_fenris6126459.cl"},
	{0x0C45EAF6D82D8703ULL, "vxlt_undercityserialkiller_120_0_0.vlm"},
	{0x0C45F16928EF51F8ULL, "6149303_m.fxe"},
	{0x0C460BD7CD2D4F5FULL, "vxw_woundedcoast_1o2_0_0.vlm"},
	{0x0C4632073866A89DULL, "hla_ht_0_hightown_n_65538_1x.dds"},
	{0x0C46A82DAA2C0DC8ULL, "dre000_varric6208467.cl"},
	{0x0C46C2FB6F449818ULL, "6066772_m.fxe"},
	{0x0C472DEEAD66664EULL, "6108944_m.fxe"},
	{0x0C47B8BD2289E84FULL, "446293436.ogg"},
	{0x0C48134A50DC8C9AULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.anb"},
	{0x0C4827E098380B23ULL, "gen00fl_sebastian6188706.cl"},
	{0x0C4865CA179EFB4BULL, "lgt360_mag_orlanna6155314.cl"},
	{0x0C4869801029D626ULL, "dae000_paivel6111933.cl"},
	{0x0C49BED28402D70EULL, "credits.gfx"},
	{0x0C49C4BED5C66EF2ULL, "dae000ip_crowd_lm2_drunk2.utp"},
	{0x0C49CA67D35BFC5FULL, "fen210_hadriana.bnk"},
	{0x0C4A4FAAF76DD768ULL, "6071916_m.fxe"},
	{0x0C4AFE4380E071A4ULL, "dae000_gamlen6165971.cl"},
	{0x0C4BDD3097DDF9C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedenter.gad"},
	{0x0C4BE643D188D9FBULL, "art\\levels\\races\\proxy\\props\\_textures\\woodrope_n.dds"},
	{0x0C4C0D0C769D144BULL, "vxlt_0_lowtown_a2_n_c_hkz_0_0.vlm"},
	{0x0C4C337CF70391E2ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamloop_10s.anb"},
	{0x0C4CAC67D35DFF67ULL, "fen210_hadriana.cnv"},
	{0x0C4CD3E92ED661CCULL, "gen00fl_merrill6106434.cl"},
	{0x0C4D01DECDD22730ULL, "llac_w_bonepitdestroyed_a3_0_2.dds"},
	{0x0C4D11A1CF54D735ULL, "vxht_brothel_n_33z_0_0.vlm"},
	{0x0C4D2B0AD171719EULL, "6092240_f.fxe"},
	{0x0C4D96205CE1843FULL, "819065408.ogg"},
	{0x0C4DAE255DA3CF24ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_dis_aoe_p.mmh"},
	{0x0C4DB393ADE5741CULL, "vxca_caverns2_937_0_1.vlm"},
	{0x0C4DCE0D1A6DD20BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_reaction.evt"},
	{0x0C4E0C3F680F71E8ULL, "402175030.ogg"},
	{0x0C4E5971350A4582ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v3.gad"},
	{0x0C4EDC26D924A307ULL, "hlac_lt_0_lowtown_a2_131071_0y.dds"},
	{0x0C4F25FE94FC4587ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_014a_0s.dds"},
	{0x0C4F7A0687F32DBFULL, "art\\vfx\\cinematic\\vfx_spark_imp_p.mmh"},
	{0x0C4F7FC3D0A716AFULL, "vxlt_gallowstemplar_f_2mw_0_0.vlm"},
	{0x0C4FF95D7B2D438AULL, "vxlt_darktown_4w5_0_0.vlm"},
	{0x0C5005C006AA3215ULL, "6166932_m.fxe"},
	{0x0C505E25044F528DULL, "vxw_encounter1_eo_0_0.vlm"},
	{0x0C50676CCD7DADFCULL, "9307980.ogg"},
	{0x0C5068A409C3F6D6ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs102a_0n.dds"},
	{0x0C508EBF71E2B3BEULL, "vxlt_undercitydungeon_205_0_0.vlm"},
	{0x0C50C3619747197EULL, "lgt210_magu_selby6142015.cl"},
	{0x0C50E3EAE23818D4ULL, "548404038.ogg"},
	{0x0C527588A508EA88ULL, "dae000cr_grand_cleric6124682.cl"},
	{0x0C529778A96825A2ULL, "vxca_caverns1_edr_0_0.vlm"},
	{0x0C52C3647C9A8A38ULL, "6111346_m.fxe"},
	{0x0C52CAC08FFE07CDULL, "vxht_keep_n_3ik_0_0.vlm"},
	{0x0C52DEEDC24FA663ULL, "llac_ca_cavern1_haunted_24425_8.dds"},
	{0x0C5303FE44E8A38AULL, "mag100_cullen6093538.cl"},
	{0x0C530BB323CD59AEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperb.msh"},
	{0x0C53136CEDFE3CF2ULL, "101105837.ogg"},
	{0x0C5319B323CD71FFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperb.mao"},
	{0x0C531E1F3FE99A5DULL, "dae000_orana6149021.cl"},
	{0x0C5325B323CD859CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperb.mmh"},
	{0x0C533BDED02B4EBDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_d.dds"},
	{0x0C53699D647EAD90ULL, "6163277_m.fxe"},
	{0x0C53733F6D43ED00ULL, "vxlt_0_lowtown_gof_0_1.vlm"},
	{0x0C5381CABFD76E9DULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.mmh"},
	{0x0C53E0719EA16DEDULL, "dae000_bartrand6205977.cl"},
	{0x0C545CD3CA239F83ULL, "vxlt_undercityserialkiller_173_0_0.vlm"},
	{0x0C54B5CE8F36C22FULL, "lit_aoe_03.mao"},
	{0x0C54E70D1A741BCCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_reaction.gad"},
	{0x0C55E0997022D7A7ULL, "965624728.ogg"},
	{0x0C564901A7F0BC91ULL, "vxht_bartrand_n_33v_0_0.vlm"},
	{0x0C57B4FE2464AB6DULL, "6078851_m.fxe"},
	{0x0C57C1433D138F83ULL, "ms_clo.mao"},
	{0x0C57D8744202E769ULL, "vxlt_gallowstemplar_f_5jg_0_0.vlm"},
	{0x0C58304B63E99858ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_anders_chant_ex2.ani"},
	{0x0C58A81E0DFCE0A4ULL, "198980892.ogg"},
	{0x0C590C4BC44F9BF1ULL, "146585284.ogg"},
	{0x0C595C4F9654EDA6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft_180.evt"},
	{0x0C596952E17E7B0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrike.gad"},
	{0x0C5A3A05C9FEF890ULL, "542586280.ogg"},
	{0x0C5A62E7D27ED54EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmld.msh"},
	{0x0C5A70E7D27EEC9FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmld.mao"},
	{0x0C5A7CE7D27F013CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmld.mmh"},
	{0x0C5A904059F34AF1ULL, "242723182.ogg"},
	{0x0C5ADF0A6BE23C22ULL, "6136963_m.fxe"},
	{0x0C5B1CFC3B672275ULL, "follower_banter6171038.cl"},
	{0x0C5B494DF905B1EEULL, "6160696_m.fxe"},
	{0x0C5BE60D1A7A38CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_reaction.ani"},
	{0x0C5BFC5D570ECC5CULL, "zz_approval_debug6003931.cl"},
	{0x0C5C25709FDB8C9CULL, "vxca_caverns2_946_0_1.vlm"},
	{0x0C5CA117364A7E89ULL, "follower_banter6160674.cl"},
	{0x0C5D0836AC1D9AE2ULL, "vxlt_darktown_4jv_0_0.vlm"},
	{0x0C5D48D9D482F69CULL, "vxht_0_hightown_a3_9ar_0_0.vlm"},
	{0x0C5E06C50062C241ULL, "vxlt_0_lowtown_a3_n_c_hcq_0_0.vlm"},
	{0x0C5E256B5D854F77ULL, "6167069_m.fxe"},
	{0x0C5ECDE5B4EE24B8ULL, "one200_gold_followers6120150.cl"},
	{0x0C5ECE2DF1B76D2BULL, "ave100_donnic6105996.cl"},
	{0x0C5ED001F6456B0DULL, "vxw_encounter1_jn_0_0.vlm"},
	{0x0C5F0189825073C7ULL, "character_character_p163.psh.fxo.dx11.assembly"},
	{0x0C5F424F7D3A3197ULL, "31616404.ogg"},
	{0x0C5F734F965980F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft_180.gad"},
	{0x0C5F7E52E1830ABCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrike.evt"},
	{0x0C5F82A0FD43A51CULL, "lgt210_shp_kill_qunari2.cnv"},
	{0x0C5FA4CD763610F0ULL, "hlac_lt_darktown_-1_90y.dds"},
	{0x0C6002271889C4AEULL, "gen00fl_isabela6100074.cl"},
	{0x0C601EE641B34EB7ULL, "prx_upperblank.dds"},
	{0x0C606A7AF33CC2E0ULL, "vxlt_gallowstemplar_3sd_0_0.vlm"},
	{0x0C60A4D2F402A07FULL, "qun100ip_compound_gate.ncs"},
	{0x0C60C301DE15F3FDULL, "6104957_f.fxe"},
	{0x0C6144E182C3BE64ULL, "m_isa121ar_chantry.wbd"},
	{0x0C61A0C92BC6AD48ULL, "gen_im_wep_war_one_act1_02.uti"},
	{0x0C61AD7A0F8DF28BULL, "vxht_mansion2_1ud_0_0.vlm"},
	{0x0C6245DCE86B30ABULL, "pro000_varric6170664.cl"},
	{0x0C627045FA9FFF9FULL, "dae000_meredith.bnk"},
	{0x0C62A4A0FD46156CULL, "lgt210_shp_kill_qunari2.bnk"},
	{0x0C63119EE1C32B80ULL, "zz_followers_banter6157329.cl"},
	{0x0C638471351CEB68ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v3.ani"},
	{0x0C64B5B78C7C03F8ULL, "art\\characters\\creatures\\desiredemon\\animation\\blendtree\\desiredemon_movement.anb"},
	{0x0C64B98D526519D3ULL, "gen00fl_merrill6108257.cl"},
	{0x0C656AFC564E176CULL, "art\\levels\\races\\design\\animation\\floortrapunactivate.ani"},
	{0x0C6587DD981B30B6ULL, "mag371_elsa6064671.cl"},
	{0x0C65A0E114DDDAB0ULL, "gen00fl_aveline6122068.cl"},
	{0x0C65F2F80BCB6A11ULL, "220209167.ogg"},
	{0x0C662D37C6E7DABFULL, "vxlt_undercityencounter_13v_0_1.vlm"},
	{0x0C665245FAA3B5A7ULL, "dae000_meredith.cnv"},
	{0x0C66744F965FA09FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft_180.ani"},
	{0x0C67100C577EC4EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_reaction_left_shoulder.ani"},
	{0x0C671505A39B7AD8ULL, "vxdr_0_deeproads_f53_0_0.vlm"},
	{0x0C674FFDB8BF19F8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_filler.phy"},
	{0x0C67913B24411768ULL, "vxlt_undercityserialkiller_1ic_0_1.vlm"},
	{0x0C67BEC9388AB341ULL, "vxlt_undercitydungeon_1mn_0_0.vlm"},
	{0x0C683A09BEB20702ULL, "hlac_ht_0_hightown_a2_-131073_0y.dds"},
	{0x0C689DA43E033C45ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_3z.dds"},
	{0x0C68B21507EC1973ULL, "plt_cod_ltr_forgamlen.nss"},
	{0x0C68BD483463D2C6ULL, "dae100_bran6118348.cl"},
	{0x0C690941436620B8ULL, "atl_load_town_dxt5_00.dds"},
	{0x0C69319561E4DFD0ULL, "pro000_merchant6051587.cl"},
	{0x0C694B45C83B4CEDULL, "gen00fl_isabela_init6075998.cl"},
	{0x0C69FD60B79F79D1ULL, "qun120cr_javaris6042831.cl"},
	{0x0C6A55786BD35A17ULL, "vxlt_hangedman_n_25r_0_0.vlm"},
	{0x0C6A6896B5DF4EA0ULL, "hlac_lt_darktown_-1_20z.dds"},
	{0x0C6AA0C61FEE383CULL, "framebuffereffects_hslmatrixfade_p0.psh.fxo.dx11"},
	{0x0C6ABBE3A244A9D9ULL, "gen00fl_aveline6132682.cl"},
	{0x0C6ABFD0D5F5E8A1ULL, "633730133.ogg"},
	{0x0C6ACE544F15FF88ULL, "lt_0_lowtown_docks_17_0_lc.mmh"},
	{0x0C6AD8F0DC6D7D18ULL, "gen_im_trs_rdr_hvy_cht_r2.uti"},
	{0x0C6ADF7FB9AA8DE1ULL, "6072471_m.fxe"},
	{0x0C6AFF652815A306ULL, "l_dae421ar_player_home_nt.wbd"},
	{0x0C6B6408352D1BBAULL, "gen00fl_sebastian6189038.cl"},
	{0x0C6B77FC565298AEULL, "art\\levels\\races\\design\\animation\\floortrapunactivate.gad"},
	{0x0C6BEB06BDF7837EULL, "487551970.ogg"},
	{0x0C6CD3DBD0657E46ULL, "follower_banter6173330.cl"},
	{0x0C6CDD175F19748AULL, "vxlt_darktown_4y7_0_0.vlm"},
	{0x0C6D8652E18F55D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrike.ani"},
	{0x0C6DD16E3DA6BFFEULL, "vxw_sundermount_1zh_0_0.vlm"},
	{0x0C6DEE4C45903D33ULL, "hm_templar_captain.mor"},
	{0x0C6E884D06746462ULL, "rot_clo.mao"},
	{0x0C6EA5AC28002C4FULL, "lt_0_lowtown_docks_187_0.mao"},
	{0x0C6EB1AC280040ECULL, "lt_0_lowtown_docks_187_0.mmh"},
	{0x0C6EB7AC28004ADEULL, "lt_0_lowtown_docks_187_0.msh"},
	{0x0C6F10DF609A80ACULL, "6154730_m.fxe"},
	{0x0C6F92852CC1FC08ULL, "6173702_m.fxe"},
	{0x0C6FA1109B31D74AULL, "832724451.ogg"},
	{0x0C6FA9861FB2CF33ULL, "vxht_0_hightown_a3_b09_0_0.vlm"},
	{0x0C706D0ED2BBB58CULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs101a.phy"},
	{0x0C707449499F012FULL, "dae011_gossips6149769.cl"},
	{0x0C70772F96F69E27ULL, "zz_seb_debug6214542.cl"},
	{0x0C70DC043D4814BFULL, "hlac_w_bonepit_131075_0z.dds"},
	{0x0C71164A50FF974DULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.mmh"},
	{0x0C7228B49919F63DULL, "stone.dds"},
	{0x0C723174D8E0342AULL, "wm_ico_doorway_roll.dds"},
	{0x0C7268583DECF515ULL, "vxlt_gallowscourtyard_6yl_0_0.vlm"},
	{0x0C7272FC5658AE8DULL, "art\\levels\\races\\design\\animation\\floortrapunactivate.evt"},
	{0x0C727899C9E9840CULL, "vxca_caverns1_e88_0_0.vlm"},
	{0x0C72B78FE665630CULL, "6125363_m.fxe"},
	{0x0C7329FD4AA60570ULL, "hlac_lt_0_lowtown_a2_2_1x.dds"},
	{0x0C733B10D8C3014BULL, "gen_im_wep_mag_sta_lore_03.uti"},
	{0x0C73A6CAE0BCA500ULL, "1066377583.ogg"},
	{0x0C74080C5789413FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_reaction_left_shoulder.evt"},
	{0x0C740BC44FEF1506ULL, "tbe100_xenon6190980.cl"},
	{0x0C74780895AF20B4ULL, "vxw_encounter3_bi_0_0.vlm"},
	{0x0C747D5D8FD95448ULL, "mag131_keran6044523.cl"},
	{0x0C74D2543A2C1E75ULL, "gen_im_arm_cht_nobb_t1.uti"},
	{0x0C76384067D68100ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean_3.mao"},
	{0x0C76B1255DC6D91BULL, "art\\vfx\\combat\\vfx_r_shd_decoy_dis_aoe_p.anb"},
	{0x0C76D203A6CD2CF6ULL, "vxlt_0_lowtown_docks_a3_n_aor_0_0.vlm"},
	{0x0C76DBE0560FE8DDULL, "844483686.ogg"},
	{0x0C773A65A644EEDAULL, "239454452.ogg"},
	{0x0C777B57DC6ACCD3ULL, "6146781_m.fxe"},
	{0x0C787B06881634FCULL, "art\\vfx\\cinematic\\vfx_spark_imp_p.anb"},
	{0x0C78EF025335D847ULL, "6172953_m.fxe"},
	{0x0C78FF6ED731B49CULL, "565380498.ogg"},
	{0x0C795F25B7B03B14ULL, "w_bonepitdestroyed_a3_11_0_lc.mmh"},
	{0x0C798D25B904346FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_meat_d.dds"},
	{0x0C79B0FFFCA5B1EDULL, "610192664.ogg"},
	{0x0C7A0C2785A27887ULL, "85544317.ogg"},
	{0x0C7A41C2A205BC6AULL, "vxlt_undercityserialkiller_1sr_0_0.vlm"},
	{0x0C7A4E2398AEB8FDULL, "6176725_f.fxe"},
	{0x0C7AE90C578F2B28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_reaction_left_shoulder.gad"},
	{0x0C7B3A4E0082B6FCULL, "651634276.ogg"},
	{0x0C7B7E3FFC3A1A8EULL, "166715699.ogg"},
	{0x0C7C1C86CC8BFDACULL, "511649613.ogg"},
	{0x0C7C80D669457FE9ULL, "622530748.ogg"},
	{0x0C7D64CBA5B64E8AULL, "6124048_f.fxe"},
	{0x0C7D671A05EF7147ULL, "576938896.ogg"},
	{0x0C7D8CBE5366A1FDULL, "vxdr_primevaltomb_6yx_0_0.vlm"},
	{0x0C7D9F78F2865F8BULL, "6036679_m.fxe"},
	{0x0C7DCAD19D47E7E2ULL, "238227332.ogg"},
	{0x0C7E205765E0514DULL, "vxht_keep_n_3xh_0_0.vlm"},
	{0x0C7E4D3A4903BA5EULL, "146701919.ogg"},
	{0x0C7E522C1AB696C9ULL, "hlac_lt_0_lowtown_a2_n_c_2_1z.dds"},
	{0x0C7E7DA8ABD3672BULL, "vxht_0_hightown_a2_c81_0_0.vlm"},
	{0x0C7E82358D2E654FULL, "lt_0_lowtown_docks_n_34_0.rml"},
	{0x0C7E9133F37A238BULL, "vxht_mansion2_1cf_0_0.vlm"},
	{0x0C7EDC45C4825A4AULL, "vxht_mansion2b_2nc_0_0.vlm"},
	{0x0C7F5BE7D29DE507ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmld.phy"},
	{0x0C7FA799DE89D7DFULL, "one220_yevhen6103018.cl"},
	{0x0C7FEEA76196B0BAULL, "6053180_f.fxe"},
	{0x0C8031F9FD636473ULL, "llac_ht_0_hightown_0_2.dds"},
	{0x0C8053A1E2395DCAULL, "hlac_w_bonepitdestroyed_a3_327681_0y.dds"},
	{0x0C8102A67EF73FB4ULL, "vxht_mansion2_30v_0_0.vlm"},
	{0x0C8132BEFA843FE7ULL, "vxht_keep_73l_0_0.vlm"},
	{0x0C8164ACCE651EF3ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv4.evt"},
	{0x0C818038F366F4C3ULL, "sf_textcolormult.psh.fxo"},
	{0x0C81CD0ACDFAB4DFULL, "149033742.ogg"},
	{0x0C81E676A160DB71ULL, "6087220_m.fxe"},
	{0x0C821D8A5E9237F2ULL, "dae000cr_grand_cleric6187948.cl"},
	{0x0C826F2AE2EFDECCULL, "store_dae210_weapons.utm"},
	{0x0C82842A64E837DEULL, "6031631_m.fxe"},
	{0x0C833C742AABFE6BULL, "40033380.ogg"},
	{0x0C8347B80AA8EB11ULL, "hlac_lt_0_lowtown_docks_-196611_1y.dds"},
	{0x0C83BC67629C428CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.activetoinactive.ani"},
	{0x0C83C44AF699A7E6ULL, "hlac_lt_0_lowtown_a2_n_c_65536_9z.dds"},
	{0x0C8445D133F7F866ULL, "mag210_arianni6021222.cl"},
	{0x0C84720D5D9CEFD4ULL, "6187910_m.fxe"},
	{0x0C850ED045B859DFULL, "824303711.ogg"},
	{0x0C859BDB70EE7A35ULL, "6155664_m.fxe"},
	{0x0C8619EA75C92751ULL, "816561455.ogg"},
	{0x0C869DC0F11EA5FAULL, "lt_0_lowtown_docks_n_102_0.mmh"},
	{0x0C86A1C0F11EAC31ULL, "lt_0_lowtown_docks_n_102_0.mao"},
	{0x0C86AAA30BA2F13AULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.cess.gad"},
	{0x0C86AFC0F11EC440ULL, "lt_0_lowtown_docks_n_102_0.msh"},
	{0x0C875B9CB3350853ULL, "dae000_mother6099197.cl"},
	{0x0C87B28F389BB7A3ULL, "6114816_m.fxe"},
	{0x0C885DACCE6B3134ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv4.gad"},
	{0x0C88D8C487E4A8A9ULL, "6116324_m.fxe"},
	{0x0C8909B00ED1B162ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelopen.msh"},
	{0x0C890BB00ED1B418ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelopen.mmh"},
	{0x0C8917B00ED1C8BBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelopen.mao"},
	{0x0C89B13E451660C5ULL, "zz_fol_debug6096543.cl"},
	{0x0C89B2639A46AE8CULL, "gen00fl_anders6107655.cl"},
	{0x0C89C729991CA044ULL, "237618843.ogg"},
	{0x0C89CC67A5DFDC18ULL, "300461011.ogg"},
	{0x0C89E2A6F29FC0EDULL, "6111664_m.fxe"},
	{0x0C8A87077D170894ULL, "vxca_cavern1_haunted_i2w_0_0.vlm"},
	{0x0C8AC96762A277CEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.activetoinactive.gad"},
	{0x0C8AEECCD96FEC1DULL, "6060740_m.fxe"},
	{0x0C8B24281A21B8CBULL, "vxlt_hangedman_n_26o_0_0.vlm"},
	{0x0C8B8DD63BA45389ULL, "6204964_m.fxe"},
	{0x0C8B8E470789AA9CULL, "ser121_jethann6164727.cl"},
	{0x0C8C3D6A4FA5FAE1ULL, "hlac_lt_0_lowtown_0_6x.dds"},
	{0x0C8C801E249B7D62ULL, "6067868_f.fxe"},
	{0x0C8C9E1331C427E5ULL, "568672513.ogg"},
	{0x0C8D09BD081ED774ULL, "518043731.ogg"},
	{0x0C8D2E8B7C508A13ULL, "rune_armor_iron.dds"},
	{0x0C8D3E98DF08D399ULL, "vxht_bartrand_n_6pn_0_0.vlm"},
	{0x0C8D973FBA114848ULL, "art\\characters\\creatures\\strider\\cn_hed_str.msh"},
	{0x0C8DA93FBA1166D9ULL, "art\\characters\\creatures\\strider\\cn_hed_str.mao"},
	{0x0C8DADA30BA91551ULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.cess.evt"},
	{0x0C8DB53FBA117B72ULL, "art\\characters\\creatures\\strider\\cn_hed_str.mmh"},
	{0x0C8DBA2CEEE9AEA8ULL, "static_static_p10.psh.fxo.dx11.assembly"},
	{0x0C8DD61570860F28ULL, "llac_w_woundedcoast_-196600_0.dds"},
	{0x0C8E6CACCE6FB612ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv4.ani"},
	{0x0C8FA354E4A5F43EULL, "6130976_m.fxe"},
	{0x0C8FBA7B021DB3D6ULL, "6128393_m.fxe"},
	{0x0C90014FAA42E91EULL, "240271662.ogg"},
	{0x0C918C2EFE577FE9ULL, "585289145.ogg"},
	{0x0C9192CB1ADA6BEFULL, "dae000_mother6004098.cl"},
	{0x0C91C46762A88DADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.activetoinactive.evt"},
	{0x0C927C1A92FFFF1BULL, "lowtown_fem_alienage1_sta6162594.cl"},
	{0x0C92C3A245388D13ULL, "pro000_merchant6051601.cl"},
	{0x0C92D7FCC92CDA00ULL, "vxlt_0_lowtown_docks_n_coy_0_0.vlm"},
	{0x0C9312B00EDA14A3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelopen.phy"},
	{0x0C93B43A8BF6F914ULL, "gen00fl_fenris6108076.cl"},
	{0x0C93CCB9BF5732B4ULL, "872871300.ogg"},
	{0x0C93E1D1818801B1ULL, "hlac_lt_0_lowtown_docks_n_65534_6y.dds"},
	{0x0C9439ED966786E1ULL, "6135169_m.fxe"},
	{0x0C943C0A31651B7CULL, "gen_im_wep_rog_dua_act2_07.uti"},
	{0x0C943D95C0BAEB17ULL, "vxht_bartrand_n_4gw_0_0.vlm"},
	{0x0C945EC7B9D5CB25ULL, "gen00fl_carver6104459.cl"},
	{0x0C9521466895BF54ULL, "llac_ht_0_hightown_a3_-131071_0.dds"},
	{0x0C95409D87D44ABCULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_stub_0n.dds"},
	{0x0C95B5C41772B44AULL, "ser200_aveline6201493.cl"},
	{0x0C96246C2A91D023ULL, "6113916_m.fxe"},
	{0x0C9624C832133E54ULL, "vxlt_gallowstemplar_f_38z_0_0.vlm"},
	{0x0C969939D2C915DDULL, "38569578.ogg"},
	{0x0C969A2103D1BD4BULL, "zz_followers_banter6157931.cl"},
	{0x0C969FFD9CD5E90BULL, "vxdr_primevalentrance_15p_0_1.vlm"},
	{0x0C97A5DEB460685AULL, "lgt000ip_ultra_mage_remains.utp"},
	{0x0C97B6092C6F5860ULL, "vxht_0_hightown_a2_awq_0_0.vlm"},
	{0x0C9832AE34FBBE0AULL, "vxlt_darktown_4n2_0_0.vlm"},
	{0x0C985A7535D3E754ULL, "dae200cr_viscount6087228.cl"},
	{0x0C987723DA8980A9ULL, "lt_gamlenshouse_n_probe_6l.dds"},
	{0x0C990D15A03A4617ULL, "ts_apostates_prog_junk2.gda"},
	{0x0C99399CB527E9C6ULL, "250438602.ogg"},
	{0x0C99C82596243165ULL, "vxht_mansion2_3do_0_0.vlm"},
	{0x0C9AC5A30BB3C820ULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.cess.ani"},
	{0x0C9AD7BAEF577B7DULL, "hlac_lt_0_lowtown_docks_a3_-65539_4y.dds"},
	{0x0C9B05DF924E363FULL, "follower_banter6147631.cl"},
	{0x0C9B75B1FBC1CC99ULL, "dae021cr_sister_colinda.utc"},
	{0x0C9B91DC362875FFULL, "6037399_m.fxe"},
	{0x0C9BD147CCAEC549ULL, "6177408_m.fxe"},
	{0x0C9C18FA7B6863C9ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_0x.dds"},
	{0x0C9CC4B0BE83E292ULL, "6126367_f.fxe"},
	{0x0C9D047FFC820F5DULL, "6037317_m.fxe"},
	{0x0C9D3847E37A1E48ULL, "6079682_m.fxe"},
	{0x0C9D3DE6BC8067DAULL, "w_encounter2_camp_25_0.rml"},
	{0x0C9DA31F941A4C30ULL, "326138717.ogg"},
	{0x0C9DD1FB24268FEFULL, "vxht_0_hightown_a2_626_0_0.vlm"},
	{0x0C9E38266456764EULL, "hlac_lt_0_lowtown_docks_a3_n_65534_1x.dds"},
	{0x0C9E64446B397B84ULL, "6161046_m.fxe"},
	{0x0C9F259C7B949566ULL, "llac_w_bonepitmining_a2_0_4.dds"},
	{0x0C9F41E4B8202128ULL, "6098971_m.fxe"},
	{0x0C9FA7F46A9599FDULL, "vxlt_0_lowtown_a2_h97_0_0.vlm"},
	{0x0C9FAB3525671077ULL, "hlac_lt_0_lowtown_docks_-4_1y.dds"},
	{0x0C9FD1BE3C746160ULL, "gen00fl_fenris6112022.cl"},
	{0x0C9FF23F5AC56BB0ULL, "hlac_lt_0_lowtown_-65539_0x.dds"},
	{0x0CA00711A9058177ULL, "hlac_dr_primevaltomb_663_3y.dds"},
	{0x0CA19B7B9FAB8615ULL, "6169744_m.fxe"},
	{0x0CA231EE825618FFULL, "hlac_lt_0_lowtown_docks_a3_n_196607_1y.dds"},
	{0x0CA2EC8B719D02ABULL, "6146998_m.fxe"},
	{0x0CA41EE6BC8651CEULL, "w_encounter2_camp_25_0.phy"},
	{0x0CA4651A70E090E4ULL, "6097321_m.fxe"},
	{0x0CA47410A6B47642ULL, "vxlt_gallowstemplar_n_5xb_0_0.vlm"},
	{0x0CA5A6BDD62734B8ULL, "205226737.ogg"},
	{0x0CA5E059C356FDDEULL, "art\\vfx\\creatures\\vfx_c_corpse_c.dura.evt"},
	{0x0CA60D07C1D0FE9BULL, "cod_bks_fifth_blight_2.utp"},
	{0x0CA642F649FDAA20ULL, "6119584_m.fxe"},
	{0x0CA64854204C7B2EULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\pine_dp.dds"},
	{0x0CA6798DA8B58768ULL, "6087180_m.fxe"},
	{0x0CA6E9F96F1DE234ULL, "vxht_0_hightown_n_bca_0_0.vlm"},
	{0x0CA6F7EB9B886989ULL, "204163388.ogg"},
	{0x0CA705B14676DB94ULL, "246493224.ogg"},
	{0x0CA7199DB77E28DCULL, "648564533.ogg"},
	{0x0CA7E473E93E8389ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_cleanse_area.ani"},
	{0x0CA7E4D35B4A20EBULL, "353036889.ogg"},
	{0x0CA8569673871FC0ULL, "6076346_m.fxe"},
	{0x0CA90AAD6C88D65AULL, "vxlt_0_lowtown_a2_e3c_0_0.vlm"},
	{0x0CA9BB1B0C124290ULL, "92256969.ogg"},
	{0x0CA9D2462F90B071ULL, "llac_w_blightlands_196611_0.dds"},
	{0x0CA9D96271548E26ULL, "mcr310_loyalist_templars6076901.cl"},
	{0x0CA9DF9F6D82900AULL, "hlac_lt_warehouse_0_3z.dds"},
	{0x0CAA71C17DDAD668ULL, "seb221_flora_harimann6177688.cl"},
	{0x0CAA8522FD736E33ULL, "6027592_m.fxe"},
	{0x0CAA91606975DB5EULL, "365429249.ogg"},
	{0x0CAB1C6FC44FCF68ULL, "hlac_lt_0_lowtown_131070_3z.dds"},
	{0x0CAB50D732ACA72EULL, "951703720.ogg"},
	{0x0CABA7DA1110730EULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_meda_n.dds"},
	{0x0CABAF028DB8D978ULL, "llac_lt_gallowstemplar_n_2383_1.dds"},
	{0x0CABBCC0F13DF605ULL, "lt_0_lowtown_docks_n_102_0.phy"},
	{0x0CAC3C237BC1E9A5ULL, "239022784.ogg"},
	{0x0CAC3E316E3AF86FULL, "6134537_m.fxe"},
	{0x0CAC4B374C83AC7EULL, "6055112_m.fxe"},
	{0x0CACC25CBAE5292AULL, "vxca_cave_19a_0_0.vlm"},
	{0x0CACC6DB7796174EULL, "dae000_arishok6067821.cl"},
	{0x0CACE59B664FE45FULL, "vxlt_darktown_5cy_0_0.vlm"},
	{0x0CACF759C35D4309ULL, "art\\vfx\\creatures\\vfx_c_corpse_c.dura.gad"},
	{0x0CAEFF73E944D00BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_cleanse_area.gad"},
	{0x0CAFD5CC74C6574DULL, "6040975_m.fxe"},
	{0x0CB02A6F3F58082AULL, "252786503.ogg"},
	{0x0CB03FD624DABA99ULL, "mcr371st_hostile_aveline.stg"},
	{0x0CB0450B2E941018ULL, "21576616.ogg"},
	{0x0CB045736A1E215CULL, "59407539.ogg"},
	{0x0CB07478660A1131ULL, "hlac_w_encounter2_camp_-2_0z.dds"},
	{0x0CB109BDA77B4496ULL, "6162315_m.fxe"},
	{0x0CB1214FA4A71C17ULL, "vxht_bartrand_n_4qq_0_0.vlm"},
	{0x0CB1440ED2F30757ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs101a.mmh"},
	{0x0CB1480ED2F30E64ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs101a.mao"},
	{0x0CB15A0ED2F32CC9ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs101a.msh"},
	{0x0CB15E6863931B2BULL, "6143698_m.fxe"},
	{0x0CB1E8D15DE5DB1BULL, "59411719.ogg"},
	{0x0CB214233EA34087ULL, "one280_merin6120918.cl"},
	{0x0CB23F88CB93BC0BULL, "dae000_worthy6166611.cl"},
	{0x0CB2C87397FD0E04ULL, "vxht_keep_61h_0_0.vlm"},
	{0x0CB34F37B9B24652ULL, "vxlt_0_lowtown_a2_n_efg_0_0.vlm"},
	{0x0CB3683B502D7569ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02b.phy"},
	{0x0CB3C859C3631257ULL, "art\\vfx\\creatures\\vfx_c_corpse_c.dura.ani"},
	{0x0CB4072307DFB073ULL, "6106947_m.fxe"},
	{0x0CB43365540AA539ULL, "hlac_lt_gallowstemplar_f_2394_10y.dds"},
	{0x0CB437E780DA0230ULL, "gen00fl_anders6125691.cl"},
	{0x0CB441B47187343DULL, "6132937_m.fxe"},
	{0x0CB452A2000FE0B4ULL, "dae160st_saemus_1_alt.stg"},
	{0x0CB4D4FEDA544CCEULL, "414054435.ogg"},
	{0x0CB50254EB4677D7ULL, "6108696_f.fxe"},
	{0x0CB5FC73E94AEA70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_cleanse_area.evt"},
	{0x0CB648266DB76668ULL, "10513482.ogg"},
	{0x0CB65054F2012306ULL, "zz_one_debug6076181.cl"},
	{0x0CB67ADB86710982ULL, "vxht_0_hightown_cas_0_0.vlm"},
	{0x0CB6C375075948C7ULL, "lt_0_lowtown_a3_n_c_46_0_lc.mmh"},
	{0x0CB72207C1DF95BEULL, "cod_bks_fifth_blight_2.ncs"},
	{0x0CB73C45D4C48582ULL, "6143174_m.fxe"},
	{0x0CB830CC009D8B7BULL, "811996742.ogg"},
	{0x0CB878794CD6AF2DULL, "vxdr_dungeon_11x_0_1.vlm"},
	{0x0CB88F3C7029D5AAULL, "hlac_w_encounter3_0_5x.dds"},
	{0x0CB8A38C22BB88BCULL, "hlac_lt_gallowstemplar_2391_0y.dds"},
	{0x0CB8C60075077BD3ULL, "6120409_f.fxe"},
	{0x0CB9B5048F164B6DULL, "619271542.ogg"},
	{0x0CB9D1119935C04BULL, "171447487.ogg"},
	{0x0CBACDC903D20BCAULL, "lgt110cr_gxa_vgk_cnspr_3.utc"},
	{0x0CBAF66FB0EF8B5AULL, "vxht_keep_6fs_0_0.vlm"},
	{0x0CBB279333E9BE82ULL, "96642979.ogg"},
	{0x0CBC25DF2AE3BAA2ULL, "505974316.ogg"},
	{0x0CBC876CD1A462D1ULL, "883417293.ogg"},
	{0x0CBCDC4CC6C44F45ULL, "hlac_w_sundermount_131073_0y.dds"},
	{0x0CBD034CA25B67CAULL, "337761487.ogg"},
	{0x0CBD93DD363A0790ULL, "l_dae321ar_keep.wbd"},
	{0x0CBDD30CEA3E27C1ULL, "gen_im_flw_dua_isabela_01.uti"},
	{0x0CBEDF2C840E3BDCULL, "6098778_f.fxe"},
	{0x0CBEE4C4D7FF4FD3ULL, "6101465_m.fxe"},
	{0x0CBEEE49BDD0D1F9ULL, "742254234.ogg"},
	{0x0CBEFAA897E094A5ULL, "6085179_f.fxe"},
	{0x0CBF12A3519391F7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_exit.gad"},
	{0x0CBF994D2EA9E766ULL, "6000965_m.fxe"},
	{0x0CC1DB20BB7E657AULL, "ripple_sliver2.dds"},
	{0x0CC1F223B67F82DFULL, "aip_apostate_base.gda"},
	{0x0CC23CFDB90BA48DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_filler.msh"},
	{0x0CC24AFDB90BBCC0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_filler.mao"},
	{0x0CC24EFDB90BC313ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_filler.mmh"},
	{0x0CC27C7DCB5E0637ULL, "661277529.ogg"},
	{0x0CC34EE2AF720921ULL, "ser221_gamlen6131907.cl"},
	{0x0CC3CDB3530A1334ULL, "vxht_0_hightown_n_big_0_0.vlm"},
	{0x0CC432373B9D90E6ULL, "vxlt_0_lowtown_docks_a3_n_d1u_0_0.vlm"},
	{0x0CC4641AC1F049CCULL, "6051910_m.fxe"},
	{0x0CC4703610821C27ULL, "894305433.ogg"},
	{0x0CC4D20AD6558A0DULL, "hlac_dr_0_deeproads_5843_6y.dds"},
	{0x0CC4EF9AC32D0A73ULL, "zz_ave_debug6056685.cl"},
	{0x0CC53F30B57F77D0ULL, "222902201.ogg"},
	{0x0CC59578B55B5F15ULL, "995805711.ogg"},
	{0x0CC5C37A109496C4ULL, "gen00fl_anders6113100.cl"},
	{0x0CC5C81409060140ULL, "vxca_caverns1_ijr_0_0.vlm"},
	{0x0CC5F568AF698DB3ULL, "blood_conj_08_l2.mao"},
	{0x0CC627A35199D384ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_exit.evt"},
	{0x0CC6D51D342CDF3CULL, "hlac_ht_mansion2b_818_64z.dds"},
	{0x0CC766AC76A5F668ULL, "72702693.ogg"},
	{0x0CC781B6FC2819DFULL, "hlac_ht_chantry_2534_18y.dds"},
	{0x0CC7D39AFAF6B498ULL, "vxca_caverns2prc_8s2_0_0.vlm"},
	{0x0CC7E93DA0AB1AE3ULL, "qun221_viscount6070092.cl"},
	{0x0CC7EA6624070ED4ULL, "289166105.ogg"},
	{0x0CC8119867727A4FULL, "vxlt_hangedman_286_0_0.vlm"},
	{0x0CC836D67F45CF56ULL, "6032263_m.fxe"},
	{0x0CC8554C71BB7863ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_0y.dds"},
	{0x0CC8BCA064AB5D99ULL, "plt_lgt110pt_pier_pressure.plo"},
	{0x0CC94191E96ECAB3ULL, "vxht_0_hightown_a3_a7g_0_0.vlm"},
	{0x0CC9684CA2E5A3DAULL, "vxht_keep_6cl_0_0.vlm"},
	{0x0CC9BEEBC3A696FEULL, "llac_lt_0_lowtown_docks_a3_n_-1_0.dds"},
	{0x0CC9F47330D85F08ULL, "269278853.ogg"},
	{0x0CC9F72C306F86CAULL, "489682425.ogg"},
	{0x0CCA2AAEE5041ECAULL, "427560635.ogg"},
	{0x0CCA413CC12C57C0ULL, "6090392_f.fxe"},
	{0x0CCA82781B217E9DULL, "200058222.ogg"},
	{0x0CCAA36FECD1C825ULL, "759648248.ogg"},
	{0x0CCB7B412E6E8778ULL, "6142012_m.fxe"},
	{0x0CCBE58D714F901AULL, "vxht_bartrand_n_64u_0_0.vlm"},
	{0x0CCC2110EEE05F71ULL, "hlac_lt_undercityserialkiller_-1_39y.dds"},
	{0x0CCD12A5B3AA0E36ULL, "mnd_con_custic.mao"},
	{0x0CCD15FB3DDD207EULL, "zz_one_debug6073991.cl"},
	{0x0CCD17C1EC577D42ULL, "316126267.ogg"},
	{0x0CCDC71591D62C7BULL, "vis121_saemus6118420.cl"},
	{0x0CCDCC58CBEB004BULL, "gen00fl_anders6119877.cl"},
	{0x0CCE23EE147BFBBBULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf05.phy"},
	{0x0CCE4922AAC045FCULL, "292747142.ogg"},
	{0x0CCEA9B9DFC7E2DAULL, "vxht_0_hightown_n_bpf_0_0.vlm"},
	{0x0CCEAD150842C3DCULL, "plt_cod_ltr_forgamlen.plo"},
	{0x0CCEDE815DD456FFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medd_0d.dds"},
	{0x0CCF4E217D5D9C15ULL, "vxw_encounter3_f0_0_0.vlm"},
	{0x0CCF5982D14174D0ULL, "llac_lt_0_lowtown_a2_1_8.dds"},
	{0x0CCF66926CCAC846ULL, "vxht_brothel_n_630_0_0.vlm"},
	{0x0CCF861B1D4B50B2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_xenon_sa.dds"},
	{0x0CCF8D6A0B65DDB3ULL, "gen00fl_aveline6137481.cl"},
	{0x0CCFAB64F4244150ULL, "vxht_0_hightown_b94_0_0.vlm"},
	{0x0CD010D76DD3DFBAULL, "gen00fl_isabela6102454.cl"},
	{0x0CD065A2E86DE76FULL, "character_character_p24.psh.fxo.dx11"},
	{0x0CD0D68C2B0D7E7BULL, "vxdr_primevaltomb_26g_0_0.vlm"},
	{0x0CD0F88D4434008FULL, "art\\vfx\\materials\\vfx_cyc_add.mao"},
	{0x0CD159C1273924FCULL, "6090411_f.fxe"},
	{0x0CD179AA0B2182C0ULL, "vxht_mansion2_311_0_0.vlm"},
	{0x0CD1A02090767641ULL, "vxht_brothel_n_6k3_0_0.vlm"},
	{0x0CD1D66B8732B252ULL, "vxlt_warehouse_v7_0_0.vlm"},
	{0x0CD215D985590370ULL, "6094734_f.fxe"},
	{0x0CD25D1340F42DB7ULL, "6152199_m.fxe"},
	{0x0CD29792CE504B38ULL, "character_character_p245.psh.fxo.dx11.assembly"},
	{0x0CD2AD7F5947AEB9ULL, "6123411_f.fxe"},
	{0x0CD2D2ABB4961911ULL, "128801892.ogg"},
	{0x0CD2D76167D8D2CCULL, "27937349.ogg"},
	{0x0CD2D85CE3501970ULL, "vxca_caverns1_ii7_0_0.vlm"},
	{0x0CD2D9CE6D4CF53CULL, "107331289.ogg"},
	{0x0CD32216E9C509C0ULL, "233106016.ogg"},
	{0x0CD33FA351A4877DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_exit.ani"},
	{0x0CD3CFFCAC0CFE36ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1j.mao"},
	{0x0CD3DBFCAC0D12D5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1j.mmh"},
	{0x0CD3DDFCAC0D164BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1j.msh"},
	{0x0CD429057446E2AFULL, "6111257_m.fxe"},
	{0x0CD439F0FAFC19C0ULL, "vxca_caverns1_ims_0_0.vlm"},
	{0x0CD49E28561CBF76ULL, "dae000_merchant_hm_mob6174220.cl"},
	{0x0CD4AEE181742E96ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb103a_0s.dds"},
	{0x0CD4D5E11214A02FULL, "150529595.ogg"},
	{0x0CD5205FCE7BBE80ULL, "vxlt_0_lowtown_a2_n_gmj_0_0.vlm"},
	{0x0CD52418CE3FF153ULL, "vxlt_0_lowtown_a2_n_eqg_0_0.vlm"},
	{0x0CD5297A7CDC617FULL, "hlac_lt_warehouse_n_-65536_1x.dds"},
	{0x0CD59248EE21F05BULL, "dae300_vashothlead6062776.cl"},
	{0x0CD641F456FD8A73ULL, "llac_ht_mansion2_818_29.dds"},
	{0x0CD68375596892CFULL, "vxlt_hangedman_251_0_0.vlm"},
	{0x0CD683DF0D1C6F72ULL, "502288018.ogg"},
	{0x0CD78BA3AE28214FULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.mmh"},
	{0x0CD7B6DDCFD299E9ULL, "vxlt_gallowstemplar_n_5qb_0_0.vlm"},
	{0x0CD7C1C15A9BF27BULL, "885942524.ogg"},
	{0x0CD7D00B876BEB57ULL, "vxlt_gallowstemplar_478_0_0.vlm"},
	{0x0CD7DA2994DBBC5AULL, "vxht_keep_6lm_0_0.vlm"},
	{0x0CD7DEFD1FA59840ULL, "6160636_m.fxe"},
	{0x0CD831D1DD140557ULL, "hlac_ht_chantry_2534_10y.dds"},
	{0x0CD8D21D0CF4CDD8ULL, "176252087.ogg"},
	{0x0CD960611FF1AFD4ULL, "6152358_m.fxe"},
	{0x0CD976A70748A723ULL, "follower_banter6170237.cl"},
	{0x0CD9E4E376385C85ULL, "dae000cr_grand_cleric6148096.cl"},
	{0x0CDADC591A583929ULL, "hlac_lt_0_lowtown_a2_n_c_65536_12y.dds"},
	{0x0CDB2340FA9A4399ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagkeep.msh"},
	{0x0CDB2D40FA9A5487ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagkeep.mmh"},
	{0x0CDB3140FA9A5BD4ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagkeep.mao"},
	{0x0CDB444CBD4F4ECCULL, "377102022.ogg"},
	{0x0CDB84DFE40259D7ULL, "llac_ca_caverns1_mines_24422_5.dds"},
	{0x0CDBDC77F2F2E269ULL, "712909146.ogg"},
	{0x0CDC0A090F1B14C7ULL, "llac_w_bonepitdestroyed_a3_1_0.dds"},
	{0x0CDC4F884048440FULL, "386383903.ogg"},
	{0x0CDC94FA1F2AF5D8ULL, "305657488.ogg"},
	{0x0CDD09072AF8F60FULL, "vxlt_0_lowtown_a2_n_hib_0_0.vlm"},
	{0x0CDE44E9943A94D5ULL, "vxht_0_hightown_a2_bi9_0_0.vlm"},
	{0x0CDE9463DC43D447ULL, "6088245_m.fxe"},
	{0x0CDE990656805215ULL, "vxca_caverns1_ihy_0_0.vlm"},
	{0x0CDEE1729F948CC3ULL, "gen00fl_aveline6132774.cl"},
	{0x0CDF2690E8E2875FULL, "primitives_primitive_p3.vsh.fxo.dx11"},
	{0x0CDF513B5052888EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02b.mmh"},
	{0x0CDF6B3B5052B57CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02b.msh"},
	{0x0CDF946421756C91ULL, "vxw_encounter2_camp_p8_0_1.vlm"},
	{0x0CE02ACD04147B13ULL, "vxca_caverns1_mines_e7s_0_0.vlm"},
	{0x0CE09B96257ADBF1ULL, "llac_lt_0_lowtown_docks_a3_n_131072_1.dds"},
	{0x0CE0C7EBF38E25E3ULL, "character_character_p134.psh.fxo.dx11.assembly"},
	{0x0CE11486F785708DULL, "art\\characters\\playercharacter\\humanmale\\boo_medd.mmh"},
	{0x0CE11886F78577DEULL, "art\\characters\\playercharacter\\humanmale\\boo_medd.mao"},
	{0x0CE12686F7858F13ULL, "art\\characters\\playercharacter\\humanmale\\boo_medd.msh"},
	{0x0CE155AAB6D09CE0ULL, "152332884.ogg"},
	{0x0CE160EED677D2A2ULL, "vxlt_backalley_g_1u8_0_0.vlm"},
	{0x0CE1A9CDDBF45DB2ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb106a_0t.dds"},
	{0x0CE1F9606FB78077ULL, "hlac_lt_gallowscourtyard_-131073_0y.dds"},
	{0x0CE1FB6E195D93F2ULL, "816261673.ogg"},
	{0x0CE25AA675E9FCA3ULL, "1041266337.ogg"},
	{0x0CE29FFD6ED9E69DULL, "201770493.ogg"},
	{0x0CE2AF46C80801FDULL, "6051036_m.fxe"},
	{0x0CE2FDC8B715DB52ULL, "dae000_gen_turn_in_26178484.cl"},
	{0x0CE3291F1B2DA0AAULL, "vxlt_undercitydungeon_15q_0_0.vlm"},
	{0x0CE32B02541EDD79ULL, "vxlt_gallowstemplar_n_6tb_0_0.vlm"},
	{0x0CE3923CC071D700ULL, "vxlt_0_lowtown_a2_n_gjk_0_0.vlm"},
	{0x0CE41DE6BCBC3391ULL, "w_encounter2_camp_25_0.mmh"},
	{0x0CE429E6BCBC4832ULL, "w_encounter2_camp_25_0.mao"},
	{0x0CE430B5D2C3B510ULL, "152598240.ogg"},
	{0x0CE437E6BCBC604FULL, "w_encounter2_camp_25_0.msh"},
	{0x0CE4616BFE9604DCULL, "6095333_f.fxe"},
	{0x0CE463ACA7664E72ULL, "vxht_mansion2_2yf_0_0.vlm"},
	{0x0CE466B6D1443BC3ULL, "lt_0_lowtown_docks_n_1_0.rml"},
	{0x0CE46A2191064658ULL, "6104843_m.fxe"},
	{0x0CE4857AEB24D2E4ULL, "6160795_m.fxe"},
	{0x0CE4F71BA8254595ULL, "w_sundermount_197_0.phy"},
	{0x0CE53FA162E501EAULL, "6152327_m.fxe"},
	{0x0CE59037091392B0ULL, "vxca_caverns1_mines_e89_0_0.vlm"},
	{0x0CE599946BAECEA6ULL, "6088781_m.fxe"},
	{0x0CE5F99BDEA0AA88ULL, "636336917.ogg"},
	{0x0CE61135EA4A0D2AULL, "llac_ca_cavern1_haunted_24427_3.dds"},
	{0x0CE6544F2BD56B09ULL, "vxlt_blackemporium_tf_0_0.vlm"},
	{0x0CE65C210137F523ULL, "vxw_sundermount_2ga_0_0.vlm"},
	{0x0CE6AD0F2E28465FULL, "vxlt_darktown_5we_0_0.vlm"},
	{0x0CE6B70C00662688ULL, "657684970.ogg"},
	{0x0CE7F5CBBEBF3DE7ULL, "mag120_macha6091369.cl"},
	{0x0CE7FFD23DA74788ULL, "dae000_madam_lusine6154912.cl"},
	{0x0CE8B43FBA5E256DULL, "art\\characters\\creatures\\strider\\cn_hed_str.phy"},
	{0x0CE92B074F98D7BDULL, "mrl350ip_at_demoncave.utp"},
	{0x0CE95CF07603FAE4ULL, "777374292.ogg"},
	{0x0CE9AF61C9ADA8A7ULL, "vxlt_warehouse_18a_0_0.vlm"},
	{0x0CE9E98D3CC439B7ULL, "vxht_keep_5hc_0_0.vlm"},
	{0x0CEA087A1006E05BULL, "703839708.ogg"},
	{0x0CEAB7A064C83BAEULL, "plt_lgt110pt_pier_pressure.nss"},
	{0x0CEAC46CE6339E4CULL, "vxca_caverns1_mines_i8y_0_0.vlm"},
	{0x0CEAC8D8521191C7ULL, "6125653_f.fxe"},
	{0x0CEB48C7BA8AD3DAULL, "467616385.ogg"},
	{0x0CEC64EF851DA310ULL, "hlac_w_woundedcoast_-65532_0y.dds"},
	{0x0CECC3FD359E01D3ULL, "451246302.ogg"},
	{0x0CED0640CE39ECC7ULL, "6083345_m.fxe"},
	{0x0CED27F2CD6811E0ULL, "6077983_m.fxe"},
	{0x0CED7AEE70A53674ULL, "zz_vault_debug6177774.cl"},
	{0x0CED8B721635A011ULL, "llac_lt_foundry_n_320_22.dds"},
	{0x0CED902939B6B391ULL, "da2terrain_rendertodepth_p0.psh.fxo"},
	{0x0CEE6371F9A161EBULL, "hlac_lt_0_lowtown_n_2_1x.dds"},
	{0x0CEEFD985E21D820ULL, "hlac_ca_caverns2_5802_18z.dds"},
	{0x0CEF0A1E347832D6ULL, "qun160_party_comment_26061239.cl"},
	{0x0CEF65DD97936FCDULL, "hlac_lt_darktown_-1_85z.dds"},
	{0x0CEFD1039B72417CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_enter.gad"},
	{0x0CEFD2CBC86DEB22ULL, "vxlt_backalley_g_1r9_0_0.vlm"},
	{0x0CF074425EC909D1ULL, "6071510_f.fxe"},
	{0x0CF0BFFCD839169AULL, "277687265.ogg"},
	{0x0CF108D30D8D1F6AULL, "gen00fl_varric6087601.cl"},
	{0x0CF1147907EB4820ULL, "394453729.ogg"},
	{0x0CF19CDF4614F5F9ULL, "vxlt_gallowstemplar_n_6qa_0_0.vlm"},
	{0x0CF20D86F793D84AULL, "art\\characters\\playercharacter\\humanmale\\boo_medd.phy"},
	{0x0CF23087DBBDE566ULL, "335939910.ogg"},
	{0x0CF2F9ED9875384BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junke_n.dds"},
	{0x0CF348BE4765A5A4ULL, "vfx_cone_lightning_f_01.mao"},
	{0x0CF436F84CE57714ULL, "fxc_lactate_add.mao"},
	{0x0CF49493D97D538BULL, "gen00fl_varric6076516.cl"},
	{0x0CF49E851352C9E3ULL, "zz_followers_banter6159111.cl"},
	{0x0CF4B38BB748E0C2ULL, "195685682.ogg"},
	{0x0CF4C4AA13E12A03ULL, "hlac_w_bonepitdestroyed_a3_0_5x.dds"},
	{0x0CF4CDEB7C3996A1ULL, "vxlt_gallowscourtyard_5a3_0_0.vlm"},
	{0x0CF5745525BE494DULL, "209214089.ogg"},
	{0x0CF594CF9E755787ULL, "vxca_caverns1_mines_dpc_0_0.vlm"},
	{0x0CF598DA6D3E76CBULL, "6154988_m.fxe"},
	{0x0CF63469E94048A1ULL, "vxht_mansion2_325_0_0.vlm"},
	{0x0CF678CD1755FD34ULL, "vxw_encounter3_7v_0_0.vlm"},
	{0x0CF6868A3BDC2C7AULL, "1029820033.ogg"},
	{0x0CF68923581898BBULL, "qun211_braggart6075703.cl"},
	{0x0CF6D0039B785E7AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_enter.ani"},
	{0x0CF6F2F2D65114E7ULL, "vxw_encounter1_kc_0_0.vlm"},
	{0x0CF79ED08AC95705ULL, "llac_ht_keep_n_3734_0.dds"},
	{0x0CF7C7318E756871ULL, "art\\vfx\\combat\\vfx_c_rw_s1_s2_trans_c.impa.ani"},
	{0x0CF8AE64C49FAC30ULL, "298076912.ogg"},
	{0x0CF8FE71B4E9F871ULL, "6087071_m.fxe"},
	{0x0CF987C06E52CFB7ULL, "hlac_w_blightlands_262147_0z.dds"},
	{0x0CF9AF07A2B8222EULL, "mag160_feynriel6016574.cl"},
	{0x0CF9C1E4D3FD9F43ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\idletorelax.ani"},
	{0x0CFA0AEE14A10C5AULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf05.msh"},
	{0x0CFA1CEE14A12AA0ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf05.mmh"},
	{0x0CFA200709687A26ULL, "vxlt_0_lowtown_docks_a3_d3l_0_0.vlm"},
	{0x0CFA469AB55C3925ULL, "hlac_lt_0_lowtown_a2_n_c_65537_7x.dds"},
	{0x0CFA90E826B04B14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\defend.ani"},
	{0x0CFAF190C0B933C6ULL, "creature_core.ncs"},
	{0x0CFAFD6F1CCE220EULL, "hlac_lt_0_lowtown_a2_-65537_6x.dds"},
	{0x0CFB3DB04553AE79ULL, "vxht_0_hightown_a3_bu4_0_0.vlm"},
	{0x0CFB7C7694810061ULL, "combat_smaoe_r2_a5.gda"},
	{0x0CFC32B2885BE33BULL, "pou_rock_sm.mao"},
	{0x0CFC411D4E6A6253ULL, "vxca_caverns1_mines_i7o_0_0.vlm"},
	{0x0CFCF3630FF13C9FULL, "vxlt_0_lowtown_docks_a3_d13_0_0.vlm"},
	{0x0CFD0CA05668F046ULL, "qcr000_arishok6042510.cl"},
	{0x0CFD174948D43922ULL, "vxw_bonepitmining_a2_1hw_0_0.vlm"},
	{0x0CFD3AEDDB5386A6ULL, "6216173_m.fxe"},
	{0x0CFE0EE959DE1B53ULL, "318313198.ogg"},
	{0x0CFE45EAF56082A8ULL, "qun110_petrice6042249.cl"},
	{0x0CFED2318E7B9913ULL, "art\\vfx\\combat\\vfx_c_rw_s1_s2_trans_c.impa.gad"},
	{0x0CFF15146BF78934ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_3y.dds"},
	{0x0CFF29D8F6AB3967ULL, "435466421.ogg"},
	{0x0CFF760F0A6DEFC9ULL, "291178791.ogg"},
	{0x0D0008F67266A297ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_lava1_d.dds"},
	{0x0D0032100FA7E164ULL, "420266332.ogg"},
	{0x0D008CA3AE4B278CULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.anb"},
	{0x0D00BB096EB5FB39ULL, "w_sundermount_0_0.msh"},
	{0x0D00C5096EB60C27ULL, "w_sundermount_0_0.mmh"},
	{0x0D00C80A302B3E3CULL, "hlac_lt_gallowstemplar_f_7198_1y.dds"},
	{0x0D00C9096EB61374ULL, "w_sundermount_0_0.mao"},
	{0x0D00FF6D08CFCA65ULL, "6176065_m.fxe"},
	{0x0D018DE826B66576ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\defend.gad"},
	{0x0D02116D195C9097ULL, "1011338781.ogg"},
	{0x0D027093390FB3BDULL, "6125502_m.fxe"},
	{0x0D027D4E4F9AFFA6ULL, "6086981_m.fxe"},
	{0x0D028BF2A5BAD286ULL, "gen00fl_merrill6128757.cl"},
	{0x0D02E4C7B31401C5ULL, "205206019.ogg"},
	{0x0D033E56ECE82896ULL, "vxw_blightlands_10t_0_0.vlm"},
	{0x0D033FDAE5242623ULL, "vxw_sundermount_2mc_0_0.vlm"},
	{0x0D0355BA37316685ULL, "503480832.ogg"},
	{0x0D038F48886E95FFULL, "301302406.ogg"},
	{0x0D04051605147F3DULL, "llac_lt_0_lowtown_a2_0_8.dds"},
	{0x0D0427FD9A228D1DULL, "hlac_w_woundedcoast_262144_0y.dds"},
	{0x0D043E11923A8B6CULL, "hm_patron1.mor"},
	{0x0D04620A5C0911E0ULL, "6097872_f.fxe"},
	{0x0D046252846EA635ULL, "one140cr_charade6128483.cl"},
	{0x0D0503C8A2A24E6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_down.gad"},
	{0x0D0539C590CCD5E0ULL, "zz_vault_debug6178670.cl"},
	{0x0D0569DD36777652ULL, "l_dae321ar_keep.bnk"},
	{0x0D05C1D30FBF00D2ULL, "vxca_caverns1_mines_i7j_0_0.vlm"},
	{0x0D05CF318E81B378ULL, "art\\vfx\\combat\\vfx_c_rw_s1_s2_trans_c.impa.evt"},
	{0x0D05DB66ADC83814ULL, "933479902.ogg"},
	{0x0D065A643E4624A1ULL, "llac_lt_undercitydungeon_-1_26.dds"},
	{0x0D06931BAD99D9A7ULL, "vxlt_warehouse_12c_0_0.vlm"},
	{0x0D069CC3E83C46D5ULL, "vxlt_0_lowtown_a2_hd8_0_0.vlm"},
	{0x0D06D54BACF03146ULL, "and100_anders6095923.cl"},
	{0x0D071640FABF68BCULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagkeep.phy"},
	{0x0D07D9E4D40A053AULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\idletorelax.evt"},
	{0x0D0808C5E269C823ULL, "6162511_f.fxe"},
	{0x0D0826390D3CF427ULL, "gen00fl_merrill6129551.cl"},
	{0x0D084C2A35B99801ULL, "mcr370_meredith6084422.cl"},
	{0x0D0891E3FB5E92A6ULL, "vxlt_0_lowtown_docks_a3_d4m_0_0.vlm"},
	{0x0D08A8E826BCB1D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\defend.evt"},
	{0x0D08B04CF1DFE939ULL, "art\\vfx\\cinematic\\vfx_orsino_swirly.mmh"},
	{0x0D08D59FB5DBA08AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_gong.mao"},
	{0x0D08D99FB5DBA7D9ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_gong.mmh"},
	{0x0D08E39FB5DBB8C7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_gong.msh"},
	{0x0D091F880A3752EFULL, "zz_varricarrows.cut"},
	{0x0D09377839AE5B13ULL, "vxlt_gallowstemplar_26e_0_0.vlm"},
	{0x0D0951EFE86933A7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishok_overlay.anb"},
	{0x0D099FF1717ECD80ULL, "6034353_m.fxe"},
	{0x0D09D1872FA2E537ULL, "vxht_0_hightown_c78_0_0.vlm"},
	{0x0D09F4A22CFCADA9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_en.ani"},
	{0x0D0A0001B1068394ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sword_r.dds"},
	{0x0D0A38FA882D9CFEULL, "vxlt_0_lowtown_a2_n_c_cw0_0_0.vlm"},
	{0x0D0A989E90B274E2ULL, "mag360_alain6065045.cl"},
	{0x0D0ABC01BD2EE65EULL, "6111263_m.fxe"},
	{0x0D0ABC511D4C8FEFULL, "122302588.ogg"},
	{0x0D0B6C6FD35D5FE2ULL, "dre000_varric6027752.cl"},
	{0x0D0B9F63BE3FD53BULL, "lowtown_male_beggar_sta6156759.cl"},
	{0x0D0BE6C8A2A83B01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_down.evt"},
	{0x0D0BE79F134B571AULL, "lt_0_lowtown_docks_161_0.msh"},
	{0x0D0BF59F134B6F53ULL, "lt_0_lowtown_docks_161_0.mao"},
	{0x0D0BF99F134B7660ULL, "lt_0_lowtown_docks_161_0.mmh"},
	{0x0D0C0218860B8C94ULL, "mcr371cr_varric_cut.utc"},
	{0x0D0C1DB66CBFBC02ULL, "qun120_arishok_guard6060605.cl"},
	{0x0D0C23D75930C8F9ULL, "vxlt_0_lowtown_a2_n_hd3_0_0.vlm"},
	{0x0D0C5A7A514A8FFDULL, "zz_qun_debug6028288.cl"},
	{0x0D0CA533EE0257EFULL, "282353507.ogg"},
	{0x0D0CCA0A8D74072EULL, "6130006-mrl3_mir_humanswer2.cut"},
	{0x0D0CDC7BA3761144ULL, "vxht_0_hightown_n_7g0_0_0.vlm"},
	{0x0D0D0972A222F02FULL, "286875252.ogg"},
	{0x0D0DAB963E626DB7ULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypoucha.msh"},
	{0x0D0DBD963E628C1AULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypoucha.mao"},
	{0x0D0DC1963E629329ULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypoucha.mmh"},
	{0x0D0DC748811B41FFULL, "6163436_m.fxe"},
	{0x0D0DD8E4D40E6F2DULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\idletorelax.gad"},
	{0x0D0DE93187696700ULL, "follower_banter6155862.cl"},
	{0x0D0DF57DF1C11F7DULL, "zz_mcr_debug6071547.cl"},
	{0x0D0E22966158C1E4ULL, "gen00fl_anders6201414.cl"},
	{0x0D0E35BB938C8EAFULL, "ave330ip_chest.utp"},
	{0x0D0E7EFEA3DF7DE5ULL, "l_ser211ar_dungeon.wbd"},
	{0x0D0F3E4C2E82EEE5ULL, "vxw_encounter3_co_0_0.vlm"},
	{0x0D0F6C2F7D7AF5CCULL, "vxht_0_hightown_n_9i4_0_0.vlm"},
	{0x0D0F8DF1E1541BBCULL, "804901655.ogg"},
	{0x0D0FB88B2B4DFE82ULL, "lt_shop_3567_lc.mmh"},
	{0x0D0FF4C138041FABULL, "6182015_m.fxe"},
	{0x0D0FF7AE9F8D19E3ULL, "ico_helm_medium_4.dds"},
	{0x0D100892AD2DD95AULL, "470623730.ogg"},
	{0x0D10B673C226477AULL, "15970770.ogg"},
	{0x0D110FA22D02FA2BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_en.gad"},
	{0x0D111938F52D27E5ULL, "vxht_keep_6bt_0_0.vlm"},
	{0x0D113565E40EC037ULL, "llac_lt_0_lowtown_docks_-131074_2.dds"},
	{0x0D114CD6A3800295ULL, "vxlt_gallowscourtyard_66q_0_0.vlm"},
	{0x0D11B1B7D71A3EA3ULL, "vxw_sundermount_2vb_0_0.vlm"},
	{0x0D11C5AFC1233D3FULL, "zz_fen_debug6134199.cl"},
	{0x0D11CA19ED7D3705ULL, "90488000.ogg"},
	{0x0D11DF7B383024E4ULL, "292289617.ogg"},
	{0x0D121C12170F8A20ULL, "457077796.ogg"},
	{0x0D1265CE46EA39B6ULL, "gen_im_trs_slv_hvy_glv_r1.uti"},
	{0x0D12BBFEEBAC016CULL, "vxw_sundermount_24g_0_0.vlm"},
	{0x0D12BD846356C9CCULL, "vxht_mansion2b_3bl_0_0.vlm"},
	{0x0D12E42AA53F7D03ULL, "mag360_thrask6043657.cl"},
	{0x0D132ED1A8F76A9EULL, "vxht_keep_6a2_0_0.vlm"},
	{0x0D13439608AA7B9BULL, "vxht_mansion_n_4xu_0_0.vlm"},
	{0x0D13976CD6CB185FULL, "art\\vfx\\combat\\vfx_glb_phys_imp_maj_c.mmh"},
	{0x0D13BF5D709B5640ULL, "6172644_m.fxe"},
	{0x0D1422B0C37EA0E8ULL, "gen00fl_varric6093592.cl"},
	{0x0D143F1EB4886D08ULL, "674967087.ogg"},
	{0x0D148B4C03E8DD17ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_wkf.gad"},
	{0x0D14A98800D25980ULL, "fx_snow_wisp.mao"},
	{0x0D14EB82ACB75EAFULL, "dae000ip_crowd_mf3_talk13.utp"},
	{0x0D155B159037894DULL, "518721564.ogg"},
	{0x0D15913E403E8802ULL, "gen_im_trs_krk_low_junk1_03.uti"},
	{0x0D15B1886763F9F4ULL, "zz_followers_banter6159396.cl"},
	{0x0D15E88F9BEA9088ULL, "leap_dot.mao"},
	{0x0D16014DBD190D1DULL, "vxht_0_hightown_n_arx_0_0.vlm"},
	{0x0D163F70F5FF772FULL, "hlac_w_bonepitmining_a2_0_9x.dds"},
	{0x0D166F6ACD619ACEULL, "vxht_0_hightown_a3_bg9_0_0.vlm"},
	{0x0D16A3DF392B9C5EULL, "fen110_fenris6100299.cl"},
	{0x0D16D2BAC8915D53ULL, "art\\vfx\\materials\\vfx_fire_whispy.mao"},
	{0x0D16E3D37F92DD9BULL, "vxlt_gallowsprison_n_6mw_0_0.vlm"},
	{0x0D170CA22D076090ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_en.evt"},
	{0x0D170DD9BA4FFD9FULL, "dre000_bethany6037327.cl"},
	{0x0D171A7D6F423107ULL, "vxlt_shop_ex_0_0.vlm"},
	{0x0D177CDA240777E8ULL, "155744020.ogg"},
	{0x0D178E015FD749CDULL, "410680325.ogg"},
	{0x0D17E75379850F74ULL, "pro000_captain_wright6047467.cl"},
	{0x0D17FAABF99D0E57ULL, "6085743_m.fxe"},
	{0x0D1831D4BCC61274ULL, "6164694_m.fxe"},
	{0x0D196474F105F8F9ULL, "hlac_lt_0_lowtown_n_-65537_4x.dds"},
	{0x0D19DEC8A2B46B50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_down.ani"},
	{0x0D1A506C0181558FULL, "500984108.ogg"},
	{0x0D1ADDA898F597A5ULL, "llac_lt_backalley_n_1_7.dds"},
	{0x0D1B0B2312043344ULL, "vxdr_0_deeproads_fis_0_0.vlm"},
	{0x0D1B1D47BE7D22F1ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrock_darelief.dds"},
	{0x0D1B22E82778D18FULL, "wm_ico_lowtowndocks_roll.dds"},
	{0x0D1B6A59E5504E97ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_ceiling_n.dds"},
	{0x0D1BA04C03EF1FA4ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_wkf.evt"},
	{0x0D1BC32FDFF19766ULL, "6050133_m.fxe"},
	{0x0D1C0330596C9F01ULL, "vxlt_0_lowtown_docks_n_cot_0_0.vlm"},
	{0x0D1C6F642F577B4FULL, "29526030.ogg"},
	{0x0D1CBA33243C6C3BULL, "art\\vfx\\combat\\vfx_w_jug_tremor_lar_aoe_p.impa.ani"},
	{0x0D1CBF6F018A528EULL, "6155380_m.fxe"},
	{0x0D1CF17A10B586ADULL, "hlac_ht_chantry_2534_11x.dds"},
	{0x0D1DC6B01E2FCB3CULL, "mer240_jansen6081436.cl"},
	{0x0D1E5887C397EDE4ULL, "vxdr_0_deeproads_es2_0_0.vlm"},
	{0x0D1F6E2E6404535EULL, "6121035_m.fxe"},
	{0x0D202F6B6171E334ULL, "vxlt_0_lowtown_a2_ewq_0_0.vlm"},
	{0x0D2066525FD283A6ULL, "62735463.ogg"},
	{0x0D2072144AB4693BULL, "34952475.ogg"},
	{0x0D20F1E601164537ULL, "vxlt_gallowstemplar_456_0_0.vlm"},
	{0x0D212DDBDDA219ECULL, "vxw_sundermount_29d_0_0.vlm"},
	{0x0D21716EFBC09FECULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobe_0n.dds"},
	{0x0D21A0AE9AED831EULL, "vxht_keep_6j3_0_0.vlm"},
	{0x0D220B82B85A1822ULL, "78022508.ogg"},
	{0x0D222448FC0D6C3EULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.mmh"},
	{0x0D222C1F3DE2CDD3ULL, "vxlt_blackemporium_152_0_0.vlm"},
	{0x0D22A94469611370ULL, "770749277.ogg"},
	{0x0D22EE8C7EB055A4ULL, "w_bonepitdestroyed_a3_192_0.mmh"},
	{0x0D22F28C7EB05CF7ULL, "w_bonepitdestroyed_a3_192_0.mao"},
	{0x0D23048C7EB07B46ULL, "w_bonepitdestroyed_a3_192_0.msh"},
	{0x0D2324345B4652A4ULL, "vxca_caverns2prc_8kw_0_1.vlm"},
	{0x0D232CA5D0D0D1D5ULL, "isa311_isabela6091720.cl"},
	{0x0D24924BD54B6B7FULL, "166663230.ogg"},
	{0x0D24EBBCD4D33D6BULL, "gen_im_trs_apo_sta_cld_r0.uti"},
	{0x0D25229D6CF5EC54ULL, "vxca_caverns1_mines_i1k_0_1.vlm"},
	{0x0D2549B6D17BA13FULL, "lt_0_lowtown_docks_n_1_0.anb"},
	{0x0D2555D5D88C5E4AULL, "vxlt_0_lowtown_a2_n_f1i_0_0.vlm"},
	{0x0D259C3E6B132040ULL, "571184297.ogg"},
	{0x0D25BAEBE0D7D70FULL, "vxht_0_hightown_a3_au9_0_0.vlm"},
	{0x0D2685E2C9622B3CULL, "dae300_petrice6097866.cl"},
	{0x0D2726104203E8CDULL, "zz_isa_debug6133994.cl"},
	{0x0D272F77CC9A8FAAULL, "and100_anders6095879.cl"},
	{0x0D27770C4FA06059ULL, "zz_qun_debug6028275.cl"},
	{0x0D28B84C03F9D29DULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_wkf.ani"},
	{0x0D2922B9350EFD7EULL, "6120357_m.fxe"},
	{0x0D293279268BC2A5ULL, "vxht_mansion2b_1kh_0_0.vlm"},
	{0x0D298C28AD7FF550ULL, "vxw_bonepit_zn_0_0.vlm"},
	{0x0D29C23324470432ULL, "art\\vfx\\combat\\vfx_w_jug_tremor_lar_aoe_p.impa.evt"},
	{0x0D29D40EF94BB649ULL, "follower_banter6147413.cl"},
	{0x0D2A1A9335C72533ULL, "gen00fl_anders6114481.cl"},
	{0x0D2AAEE7E22BCC64ULL, "llac_ht_chantry_n_2534_35.dds"},
	{0x0D2BDDFFD8FE4561ULL, "vxca_caverns1_mines_djf_0_1.vlm"},
	{0x0D2BF5ED9334FA0DULL, "weaponparts.gda"},
	{0x0D2C0858C44D0356ULL, "6132871_m.fxe"},
	{0x0D2C0A1BA86168B0ULL, "w_sundermount_197_0.msh"},
	{0x0D2C181BA86180CAULL, "w_sundermount_197_0.mmh"},
	{0x0D2C1C1BA8618701ULL, "w_sundermount_197_0.mao"},
	{0x0D2D1E3DC4CF4F73ULL, "hlac_lt_0_lowtown_a2_n_c_2_1x.dds"},
	{0x0D2D225195FC3A35ULL, "6207415_m.fxe"},
	{0x0D2D25ABFA91D666ULL, "6096226_f.fxe"},
	{0x0D2DC1D1ADB6DD74ULL, "6088836_f.fxe"},
	{0x0D2DC7788D993FE0ULL, "6111654_m.fxe"},
	{0x0D2DFCF2D91958E5ULL, "vxht_keep_6hr_0_0.vlm"},
	{0x0D2DFFAD23E8AF89ULL, "529531967.ogg"},
	{0x0D2E03F62F95FD69ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scroll_sa.dds"},
	{0x0D2E04912080EFACULL, "pro000_opening6207382.cl"},
	{0x0D2E1DE545597CB1ULL, "vxlt_0_lowtown_a2_d66_0_0.vlm"},
	{0x0D2E3AA5FD636A23ULL, "6126810_f.fxe"},
	{0x0D2E53AB123B764AULL, "372827005.ogg"},
	{0x0D2E71769A39B107ULL, "277212910.ogg"},
	{0x0D2E7348A886A49CULL, "dae000_mother6099233.cl"},
	{0x0D2EAE5ABB99EF7DULL, "llac_dr_slavers_3215_6.dds"},
	{0x0D2EC78D570721C2ULL, "vxht_keep_7ac_0_0.vlm"},
	{0x0D2F01C5FE2A0B3FULL, "vxw_sundermount_23i_0_0.vlm"},
	{0x0D300348FC1971FDULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.anb"},
	{0x0D30AD4CF201362EULL, "art\\vfx\\cinematic\\vfx_orsino_swirly.anb"},
	{0x0D30E133244D57C5ULL, "art\\vfx\\combat\\vfx_w_jug_tremor_lar_aoe_p.impa.gad"},
	{0x0D31AA8165994604ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\_1_-_default__3ww_basic_.mao"},
	{0x0D31C07137816CBAULL, "fen110_fenris6100264.cl"},
	{0x0D326F41CE3A76D1ULL, "187159217.ogg"},
	{0x0D327ECF61B36432ULL, "dae000_arishok6075347.cl"},
	{0x0D3330EEEE08479FULL, "mag291_torpor6026991.cl"},
	{0x0D334237087E2207ULL, "qcr220st_meredith.stg"},
	{0x0D33C5096EE16584ULL, "w_sundermount_0_0.rml"},
	{0x0D33E11A26693A74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion3.ani"},
	{0x0D3433B8FE0452DBULL, "hlac_lt_0_lowtown_131072_0z.dds"},
	{0x0D3449ADB67DD0A4ULL, "static_static_p49.psh.fxo"},
	{0x0D34F1E9FEB24ACEULL, "vxht_keep_586_0_0.vlm"},
	{0x0D356C41B153DCF1ULL, "vxht_mansion2b_30g_0_0.vlm"},
	{0x0D366719CAB8D845ULL, "hlac_dr_dungeon_3216_9z.dds"},
	{0x0D3686105484A39CULL, "llac_w_encounter3_0_5.dds"},
	{0x0D36995B4BEDF31BULL, "hlac_ca_caverns1_24421_7x.dds"},
	{0x0D387B6F7C074EFEULL, "vxlt_0_lowtown_a2_n_c_fek_0_0.vlm"},
	{0x0D38882B1B13338DULL, "6064835_f.fxe"},
	{0x0D3889FEDD2D288FULL, "mcr310_argument6127609.cl"},
	{0x0D38A091417956DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep_3.mao"},
	{0x0D3956FB98B6C540ULL, "vxca_caverns1_i2z_0_0.vlm"},
	{0x0D399AEFCCAC545FULL, "322226114.ogg"},
	{0x0D39DE1A266DA0D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion3.gad"},
	{0x0D3B986CD6EC6B9CULL, "art\\vfx\\combat\\vfx_glb_phys_imp_maj_c.anb"},
	{0x0D3BC39E0A965C94ULL, "352690038.ogg"},
	{0x0D3BCC759D4A4DA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fa_transition.ani"},
	{0x0D3BDC9024CF4100ULL, "mag291_demon_marethari6029062.cl"},
	{0x0D3D1A7795D98910ULL, "follower_banter6148455.cl"},
	{0x0D3D73A2F02023BFULL, "vxw_sundermount_24h_0_0.vlm"},
	{0x0D3D97C7E8F96220ULL, "635637565.ogg"},
	{0x0D3DA421577D632FULL, "vxlt_0_lowtown_n_exh_0_0.vlm"},
	{0x0D3DB88ABFA18C6DULL, "atl_newcontent_dxt5_dat.xml"},
	{0x0D3E819785B82366ULL, "vxlt_gallowstemplar_f_436_0_0.vlm"},
	{0x0D3F1758F6DF43F5ULL, "165323823.ogg"},
	{0x0D3FA6BF14AA8DB6ULL, "vxht_mansion2_30d_0_0.vlm"},
	{0x0D3FE15FDB5E36ECULL, "llac_ht_chantry_n_2534_6.dds"},
	{0x0D40207A21CF747CULL, "828964641.ogg"},
	{0x0D40B4D642382E0EULL, "one280_merin6104375.cl"},
	{0x0D40E495E8549C96ULL, "zz_mag_debug6039837.cl"},
	{0x0D40F91A2673ED35ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion3.evt"},
	{0x0D40FFD93F01CD83ULL, "ser221_gamlen6133082.cl"},
	{0x0D41391A00D69FA2ULL, "zz_isa_debug6133173.cl"},
	{0x0D41C9759D4EB406ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fa_transition.gad"},
	{0x0D4204503C4805D6ULL, "544623994.ogg"},
	{0x0D420D72229148B9ULL, "vxlt_0_lowtown_a2_n_ehr_0_0.vlm"},
	{0x0D4263BBE4AB4B22ULL, "vxw_bonepitmining_a2_204_0_0.vlm"},
	{0x0D42B5744CAFFC7AULL, "vxdr_0_deeproads_cqt_0_0.vlm"},
	{0x0D42D5E362C63C4BULL, "rock.dds"},
	{0x0D42ECBB9E8888D9ULL, "44934890.ogg"},
	{0x0D436261A89EDE94ULL, "vxca_cavern1_haunted_j28_0_0.vlm"},
	{0x0D43EA34D56EF033ULL, "14909910.ogg"},
	{0x0D43F04C80B0061BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_enco_0s.dds"},
	{0x0D4439B831A47CD8ULL, "w_encounter2_camp_7_0.phy"},
	{0x0D445281DEA8E23CULL, "fen110_fenris6100356.cl"},
	{0x0D4471DE54D1D6CEULL, "28864596.ogg"},
	{0x0D447CFDCAA3CD46ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fin_c.impa.ani"},
	{0x0D4481729AF73B3DULL, "6067671_m.fxe"},
	{0x0D4483CFF62ECA95ULL, "291056774.ogg"},
	{0x0D44B1DFC13B4CCEULL, "vxht_0_hightown_a3_aut_0_0.vlm"},
	{0x0D451CED9E0C1E99ULL, "219253494.ogg"},
	{0x0D455B7D56216867ULL, "gen_im_trs_apo_lgt_cht_r1.uti"},
	{0x0D45A3582F8F44F7ULL, "6016642_m.fxe"},
	{0x0D45A8E888B45943ULL, "6028852_m.fxe"},
	{0x0D45FC574F922B48ULL, "mer320_hubert6090371.cl"},
	{0x0D46462168E8650AULL, "825896411.ogg"},
	{0x0D46EFD0A3804040ULL, "vxw_encounter3_61_0_0.vlm"},
	{0x0D475A4AC096BDB5ULL, "hlac_w_bonepit_393222_0x.dds"},
	{0x0D477124C9FFC710ULL, "vxlt_gallowstemplar_n_53a_0_0.vlm"},
	{0x0D487AA5C0E14B16ULL, "892301746.ogg"},
	{0x0D48860CFE8056A8ULL, "vxlt_gallowstemplar_5mp_0_0.vlm"},
	{0x0D48997CE36045D6ULL, "cn_bdy_abma_0t.dds"},
	{0x0D48C4759D54CAE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fa_transition.evt"},
	{0x0D48E73C1DFC053AULL, "hm_qmq_visguard2.mor"},
	{0x0D498C577498F854ULL, "759558047.ogg"},
	{0x0D49CA84FC782780ULL, "702204230.ogg"},
	{0x0D49D3E90B1B2EECULL, "6106508_m.fxe"},
	{0x0D49F1B0F7B343DDULL, "w_encounter2.arl"},
	{0x0D4A1AEEDBEA24DCULL, "6116754_m.fxe"},
	{0x0D4AE50C25811C73ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_0z.dds"},
	{0x0D4B0DD676FE6EECULL, "vxw_encounter2_53_0_0.vlm"},
	{0x0D4B38B831AA99F0ULL, "w_encounter2_camp_7_0.rml"},
	{0x0D4B8CF4B834A603ULL, "513759809.ogg"},
	{0x0D4BCC6CDA83626EULL, "314276122.ogg"},
	{0x0D4BFD50E38FCD56ULL, "hlac_ht_brothel_n_3370_1y.dds"},
	{0x0D4D000272B673AAULL, "vxlt_0_lowtown_a3_n_c_hsv_0_0.vlm"},
	{0x0D4D097EE01C4593ULL, "hlac_ht_0_hightown_n_65536_4z.dds"},
	{0x0D4D146D23762047ULL, "hlac_dr_primevaltomb_659_4y.dds"},
	{0x0D4D5BB898A017B2ULL, "87059104.ogg"},
	{0x0D4D6453912F1105ULL, "lt_0_lowtown_a2_n_c_mmap.dds"},
	{0x0D4E13E8CBB42738ULL, "art\\characters\\playercharacter\\humanmale\\hlm_wic.mao"},
	{0x0D4E1724AEEDAECDULL, "875590014.ogg"},
	{0x0D4E58FEA415208FULL, "l_ser211ar_dungeon.bnk"},
	{0x0D4EA4842063BD23ULL, "gen00fl_varric6118832.cl"},
	{0x0D4F044A525400C1ULL, "dr_0_deeproads_5842.anb"},
	{0x0D4FB5691632788BULL, "84502893.ogg"},
	{0x0D501576DFBC085EULL, "vxw_bonepit_220_0_0.vlm"},
	{0x0D504723236BA6F2ULL, "vxht_0_hightown_c35_0_0.vlm"},
	{0x0D5063DA0F2E6162ULL, "vxht_mansion2_1vd_0_0.vlm"},
	{0x0D50CF18F55F5D6FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containerb_sa.dds"},
	{0x0D510184AFEEAA25ULL, "and100_anders6095957.cl"},
	{0x0D51288E2256F75CULL, "gen00fl_varric6103266.cl"},
	{0x0D51433EB91490D0ULL, "gen00fl_merrill6108436.cl"},
	{0x0D516756EE71A4C6ULL, "vxht_brothel_n_6d9_0_0.vlm"},
	{0x0D51A611B14885D6ULL, "6040984_f.fxe"},
	{0x0D51AD23DF26C553ULL, "367348365.ogg"},
	{0x0D51D4D60B336D69ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsrock.phy"},
	{0x0D521F3BFD8E420AULL, "6017899_m.fxe"},
	{0x0D5264FDCAAFE1B7ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fin_c.impa.evt"},
	{0x0D5301A904DBDFBAULL, "152333328.ogg"},
	{0x0D53666E5759E220ULL, "6080206_m.fxe"},
	{0x0D536DE8BFACF1AEULL, "vxw_encounter2_g7_0_0.vlm"},
	{0x0D538C5AAC5D2610ULL, "710937593.ogg"},
	{0x0D53F490D6816BD0ULL, "441054822.ogg"},
	{0x0D5418BFAE78BFDDULL, "vxlt_0_lowtown_n_e3g_0_0.vlm"},
	{0x0D54457951E438B3ULL, "one300_zevran6089732.cl"},
	{0x0D545764E00573FAULL, "vxht_0_hightown_n_7lo_0_0.vlm"},
	{0x0D549F76D3A929A2ULL, "6148799_m.fxe"},
	{0x0D54C42AACA067B9ULL, "gen00fl_anders6111251.cl"},
	{0x0D54E1BF83623ED0ULL, "vxw_bonepit_km_0_0.vlm"},
	{0x0D551A415C715A36ULL, "5547625.ogg"},
	{0x0D55568443FDE5B1ULL, "qun221_viscount6084732.cl"},
	{0x0D555AB8224D350BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right.ani"},
	{0x0D5568CA2264D281ULL, "679419009.ogg"},
	{0x0D55C90A1F0D8876ULL, "6067979_m.fxe"},
	{0x0D55F5B1BFD1902BULL, "6071249_m.fxe"},
	{0x0D560F4A8E2F51C8ULL, "32106170.ogg"},
	{0x0D563AB57CA2F640ULL, "vxca_caverns1_i8x_0_0.vlm"},
	{0x0D56504650E6A2B7ULL, "hlac_ca_caverns1_24426_6z.dds"},
	{0x0D56B2D4E7D57D9BULL, "134875555.ogg"},
	{0x0D571C3512DB17B6ULL, "vxht_0_hightown_n_bxr_0_0.vlm"},
	{0x0D578B24D37B95F1ULL, "mrl000_marethari6129200.cl"},
	{0x0D578FD7814D017AULL, "204321691.ogg"},
	{0x0D57F44F8326575AULL, "hlac_ca_caverns1_24425_8x.dds"},
	{0x0D5837F1F82B959AULL, "680205625.ogg"},
	{0x0D5888CDAA01BB67ULL, "trap_impact.ncs"},
	{0x0D588EE8A9FEBC3FULL, "w_encounter3_26_0.msh"},
	{0x0D5894E8A9FEC6E1ULL, "w_encounter3_26_0.mmh"},
	{0x0D58A02BB1BC7B97ULL, "gen_im_cns_tom_war_templar.uti"},
	{0x0D58A0E8A9FEDA82ULL, "w_encounter3_26_0.mao"},
	{0x0D58C0650B2B8374ULL, "354240015.ogg"},
	{0x0D5915DD2A5807C1ULL, "lgt210tr_frame_flower.ncs"},
	{0x0D592AD72C9754BFULL, "mcr371_hostile_fenris6082245.cl"},
	{0x0D5965FDCAB60240ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fin_c.impa.gad"},
	{0x0D59961A778C3416ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_wallcap.msh"},
	{0x0D59A01A778C4514ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_wallcap.mmh"},
	{0x0D59A41A778C4C27ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_wallcap.mao"},
	{0x0D59CF21EB6DD9CAULL, "gen00fl_merrill6129880.cl"},
	{0x0D5A2C45E7C568C0ULL, "dae000_gamlen6145090.cl"},
	{0x0D5AC5B1351C35A0ULL, "847318740.ogg"},
	{0x0D5ACF3CD223DBBFULL, "12239770.ogg"},
	{0x0D5B1C1A4426BE25ULL, "326540867.ogg"},
	{0x0D5BD23F40F8A523ULL, "6134577_m.fxe"},
	{0x0D5BDC1BD365913FULL, "564612085.ogg"},
	{0x0D5C1DFDE1CA25D0ULL, "vxlt_0_lowtown_n_f5c_0_0.vlm"},
	{0x0D5C4EE2475922B9ULL, "dre000_bethany6037314.cl"},
	{0x0D5C57515C6D167FULL, "lowtown_fem_prostitute_sta6152368.cl"},
	{0x0D5C6C3E298F80BDULL, "vxht_mansion2_22l_0_0.vlm"},
	{0x0D5CF8A20A8DC4ADULL, "6095577_m.fxe"},
	{0x0D5D0A988109389FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage4_.gad"},
	{0x0D5D38B38FA8EDA1ULL, "vxw_encounter3_86_0_0.vlm"},
	{0x0D5DC7E1442D30CAULL, "6171287_m.fxe"},
	{0x0D5E47EED1A3ED67ULL, "vxlt_gallowsprison_n_5rq_0_0.vlm"},
	{0x0D5E4B935EBB8381ULL, "vxlt_0_lowtown_a2_n_gmg_0_0.vlm"},
	{0x0D5EAA7967F6E50DULL, "vxlt_hangedman_16y_0_0.vlm"},
	{0x0D5EBCC0B1BD5221ULL, "588099965.ogg"},
	{0x0D5EF71F20736525ULL, "fen210_fenris6109438.cl"},
	{0x0D5F1CEA10089884ULL, "vxca_caverns1_iys_0_0.vlm"},
	{0x0D5FFAB28ED5C5A7ULL, "hlac_w_encounter2_-2_0z.dds"},
	{0x0D611A7F02B0110AULL, "420251112.ogg"},
	{0x0D622A6C0C38FEC0ULL, "200890685.ogg"},
	{0x0D62A20E8830887DULL, "qun231_party_comment6077006.cl"},
	{0x0D62CC5153988A3AULL, "gen_im_npc_arm_hum_glv_rdra.uti"},
	{0x0D6315CBAF16BC33ULL, "vxht_mansion2_2lp_0_0.vlm"},
	{0x0D6336AB56BFDAD0ULL, "mrl210_merrill6102102.cl"},
	{0x0D6342B822594982ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right.evt"},
	{0x0D6349B6054C2E31ULL, "vxlt_gallowscourtyard_72t_0_0.vlm"},
	{0x0D63535CF33A3066ULL, "mag101_grace6170862.cl"},
	{0x0D63E49506AF1959ULL, "vxw_woundedcoast_109_0_0.vlm"},
	{0x0D6472A3B03A7B64ULL, "51228501.ogg"},
	{0x0D655211F875E4E6ULL, "llac_lt_0_lowtown_n_-131073_2.dds"},
	{0x0D6589F033E19821ULL, "491364466.ogg"},
	{0x0D65902BABC375D8ULL, "w_encounter2_56_0.phy"},
	{0x0D660FB398448300ULL, "553369298.ogg"},
	{0x0D6627059AEFAE89ULL, "mag360_samson6044920.cl"},
	{0x0D66535C229EC164ULL, "6036917_m.fxe"},
	{0x0D6691D87D666864ULL, "vxdr_0_deeproads_eh7_0_0.vlm"},
	{0x0D67C0F9AD200F48ULL, "hlac_w_encounter3_-196607_0y.dds"},
	{0x0D683F56F63B779CULL, "hlac_lt_warehouse_0_4y.dds"},
	{0x0D68811EA86FE388ULL, "6179999_m.fxe"},
	{0x0D691BB9FB1D3F35ULL, "166962717.ogg"},
	{0x0D6975A9FBFF76D7ULL, "32185918.ogg"},
	{0x0D69AD9135462D84ULL, "204559564.ogg"},
	{0x0D69EEC3B46F0E9BULL, "vxht_mansion_n_4fk_0_0.vlm"},
	{0x0D6A28297AB085E5ULL, "vxdr_primevalentrance_3df_0_0.vlm"},
	{0x0D6A41B8225F6695ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right.gad"},
	{0x0D6A6A17FFF3787AULL, "vxht_chantry_2ct_0_0.vlm"},
	{0x0D6A8DA4FC097891ULL, "llac_ca_caverns1_24424_0.dds"},
	{0x0D6A8EFE9CDB1B25ULL, "vxlt_0_lowtown_a2_n_gwy_0_0.vlm"},
	{0x0D6B39FDD6E8D7BFULL, "mag110_arianni6206382.cl"},
	{0x0D6B7D4DE9507A85ULL, "1055589583.ogg"},
	{0x0D6BB8A2EA8A5E12ULL, "root_rog_general4.dds"},
	{0x0D6BDA9FB63026C6ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_gong.phy"},
	{0x0D6C201B646CF536ULL, "hlac_lt_gallowstemplar_f_2394_3z.dds"},
	{0x0D6C8F2BABC992F0ULL, "w_encounter2_56_0.rml"},
	{0x0D6CA160971FC009ULL, "1021593448.ogg"},
	{0x0D6CF3549F7A7B0EULL, "art\\vfx\\combat\\vfx_sta_rng_atk_fin_c.impa.evt"},
	{0x0D6DC45199122502ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\custom enter\\second_enter_lp.gad"},
	{0x0D6DFA07CE0EC245ULL, "gen00fl_fenris6126800.cl"},
	{0x0D6DFC38C6F0682EULL, "88363150.ogg"},
	{0x0D6E51ABDAA95229ULL, "hlac_w_bonepitmining_a2_65539_0y.dds"},
	{0x0D6E95969D9BA29BULL, "616687708.ogg"},
	{0x0D6E9A67679A62F7ULL, "6074475_m.fxe"},
	{0x0D6EAA9F64965C20ULL, "vxca_caverns1_mines_igx_0_0.vlm"},
	{0x0D6EFF37B42D414FULL, "hlac_lt_0_lowtown_a2_n_c_131070_2y.dds"},
	{0x0D6F7C084B82E1EBULL, "gen_im_trs_gen_med_hlm_r2.uti"},
	{0x0D6FA53B1DCD39FDULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\jagged_sa.dds"},
	{0x0D6FC783481B542EULL, "vxht_0_hightown_a2_ass_0_0.vlm"},
	{0x0D6FE2C406011A53ULL, "vxlt_house_ep_0_0.vlm"},
	{0x0D703EF7264D8786ULL, "vxw_woundedcoast_2e_0_0.vlm"},
	{0x0D70A24B36AAEB46ULL, "357755680.ogg"},
	{0x0D70A66EED9C2640ULL, "vxlt_0_lowtown_a2_ep0_0_0.vlm"},
	{0x0D70D6ABE7604D5AULL, "humanfemale_b.fxa"},
	{0x0D71B70677D6E5A8ULL, "eye.mat"},
	{0x0D71E798811B59E5ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage4_.ani"},
	{0x0D724A1669FADDDDULL, "vxht_0_hightown_91n_0_0.vlm"},
	{0x0D72E641A4FA7237ULL, "vxht_brothel_n_5gn_0_0.vlm"},
	{0x0D7309884ECA4EC0ULL, "mag311_keran6054391.cl"},
	{0x0D7309E191030AEAULL, "vxca_caverns1_mines_gif_0_0.vlm"},
	{0x0D73D0DFE507ED86ULL, "612909311.ogg"},
	{0x0D73DD784061E9F9ULL, "vxca_cavern1_haunted_imo_0_0.vlm"},
	{0x0D73EA549F808AB9ULL, "art\\vfx\\combat\\vfx_sta_rng_atk_fin_c.impa.gad"},
	{0x0D73FFEEF6C748B6ULL, "vxht_0_hightown_n_bnt_0_0.vlm"},
	{0x0D742A564FB212E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_malepc.gad"},
	{0x0D748305D78A4511ULL, "hlac_ca_caverns2prc_5802_30x.dds"},
	{0x0D74B5B0FCB773E0ULL, "6046284_m.fxe"},
	{0x0D74C7A120DB8710ULL, "6054752_f.fxe"},
	{0x0D74CC58548EC279ULL, "vxht_0_hightown_n_4of_0_0.vlm"},
	{0x0D7539BD15F9A7F5ULL, "752986954.ogg"},
	{0x0D75F16074253C67ULL, "6163133_m.fxe"},
	{0x0D75FB0981CC0C04ULL, "6142849_m.fxe"},
	{0x0D762D8502D4484CULL, "638432666.ogg"},
	{0x0D76A78E949755A5ULL, "qcr220_orsino.bnk"},
	{0x0D76EA3D016DBFBAULL, "llac_lt_0_lowtown_a2_-196608_4.dds"},
	{0x0D76FAB0F7DA408BULL, "w_encounter2.rml"},
	{0x0D773BB853BE4F9CULL, "vxw_woundedcoast_19o_0_0.vlm"},
	{0x0D7787A82D3AA5DAULL, "vxht_mansion_n_60z_0_0.vlm"},
	{0x0D7817681C58E765ULL, "6176216_m.fxe"},
	{0x0D7860B1E1571D49ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv1.ani"},
	{0x0D78E0078AFCBFD8ULL, "vxht_bartrand_n_64g_0_0.vlm"},
	{0x0D78E1CE69C4EF65ULL, "gen_im_cns_cft_arn_03.uti"},
	{0x0D790857E554C23FULL, "gen00fl_dog6165769.cl"},
	{0x0D794FF80D7BB1BDULL, "vxht_mansion2_28j_0_0.vlm"},
	{0x0D796379C755B8D2ULL, "vxlt_house_eu_0_0.vlm"},
	{0x0D796FC4509B0451ULL, "427850575.ogg"},
	{0x0D798841E7EEFA6DULL, "547799581.ogg"},
	{0x0D79888E4C98D0BDULL, "6125713_m.fxe"},
	{0x0D79898E949958DDULL, "qcr220_orsino.cnv"},
	{0x0D79BFFCFDCD56A0ULL, "vxht_mansion2b_31i_0_0.vlm"},
	{0x0D7A4D564FB6B9D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_malepc.evt"},
	{0x0D7A78260F284901ULL, "dae000_mother6035922.cl"},
	{0x0D7A8FD19E7FD20EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_buff.ani"},
	{0x0D7AA861D5CE7D58ULL, "849173354.ogg"},
	{0x0D7AD8D271667083ULL, "docks_male_elven_worker_sta6175809.cl"},
	{0x0D7AFB549F86C607ULL, "art\\vfx\\combat\\vfx_sta_rng_atk_fin_c.impa.ani"},
	{0x0D7B5E50B82DAABAULL, "vxht_mansion2b_32z_0_0.vlm"},
	{0x0D7B726B92D53DCAULL, "6031994_m.fxe"},
	{0x0D7BB0C79CFCA7E4ULL, "6145164_f.fxe"},
	{0x0D7BB970E510AAE2ULL, "vxht_mansion2_1gg_0_0.vlm"},
	{0x0D7C00A3F3F4C984ULL, "vxca_caverns1_igq_0_0.vlm"},
	{0x0D7C0AC5AA2C31B9ULL, "qun111_ketojan6041031.cl"},
	{0x0D7C181D9488A40BULL, "6110851_f.fxe"},
	{0x0D7C56C5E6E1F7AFULL, "472868899.ogg"},
	{0x0D7CE5B79A34D16EULL, "6035203_f.fxe"},
	{0x0D7D74E8AA1DAC0AULL, "w_encounter3_26_0.rml"},
	{0x0D7DBDD60B58808EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsrock.mmh"},
	{0x0D7DC9D60B589525ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsrock.mao"},
	{0x0D7DD7D60B58AD7CULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsrock.msh"},
	{0x0D7E1E1B5350F1CFULL, "6103610_m.fxe"},
	{0x0D7E6D42C7EDF9C5ULL, "art\\vfx\\environment\\vfx_tiki_fire.anb"},
	{0x0D7E7BB1E15BB6CBULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv1.gad"},
	{0x0D7E9FC26D281E2BULL, "438560726.ogg"},
	{0x0D7ECE7F52D44791ULL, "llac_w_sundermount_-196614_0.dds"},
	{0x0D7F59CECC1D9670ULL, "playersoundsets.gda"},
	{0x0D7F682A2BE12DE1ULL, "6131579_f.fxe"},
	{0x0D7F68F13E143640ULL, "qun110_petrice6042252.cl"},
	{0x0D7F9B8450341DB9ULL, "885229209.ogg"},
	{0x0D80318E67575D79ULL, "hlac_lt_0_lowtown_a2_n_65534_6x.dds"},
	{0x0D8060432EC35693ULL, "6053470_enter_keep.cut"},
	{0x0D812559D76CB68DULL, "157354112.ogg"},
	{0x0D81581E96F08AB7ULL, "vxht_brothel_n_5ji_0_0.vlm"},
	{0x0D81A4D0C4B998FCULL, "339382718.ogg"},
	{0x0D8271CBE8BD6136ULL, "vxht_0_hightown_n_biu_0_0.vlm"},
	{0x0D82BC17DA9DEB6EULL, "vxw_sundermount_30y_0_1.vlm"},
	{0x0D82D80E56261CFBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1h.mmh"},
	{0x0D82E40E56263098ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1h.mao"},
	{0x0D82EF519924CAE8ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\custom enter\\second_enter_lp.ani"},
	{0x0D82F60E56264FE5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1h.msh"},
	{0x0D8307F7A4D222E0ULL, "vxca_caverns1_eib_0_1.vlm"},
	{0x0D83278DEEAD8C60ULL, "6041384_m.fxe"},
	{0x0D832A83DFE1B230ULL, "art\\characters\\playercharacter\\humanmale\\arm_medd.mao"},
	{0x0D832E83DFE1B943ULL, "art\\characters\\playercharacter\\humanmale\\arm_medd.mmh"},
	{0x0D833C83DFE1D09DULL, "art\\characters\\playercharacter\\humanmale\\arm_medd.msh"},
	{0x0D84093FDD011B98ULL, "vxw_encounter2_ox_0_1.vlm"},
	{0x0D8528A6691B056CULL, "vxw_blightlands_m8_0_0.vlm"},
	{0x0D8540DE5061FE50ULL, "199531163.ogg"},
	{0x0D8577240BE701EAULL, "963662978.ogg"},
	{0x0D8578B1E161D030ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv1.evt"},
	{0x0D860E3086A7EAB9ULL, "vxca_caverns1_mines_i0x_0_0.vlm"},
	{0x0D861BB5D442B285ULL, "1063252336.ogg"},
	{0x0D8640DC4A4296F5ULL, "6107182_m.fxe"},
	{0x0D86EF51D9C4D3A1ULL, "vxlt_0_lowtown_a2_er7_0_0.vlm"},
	{0x0D874DD1E3DFA97AULL, "vxht_chantry_2yv_0_0.vlm"},
	{0x0D8787D19E8A4F5FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_buff.evt"},
	{0x0D879AE6F9B582C2ULL, "gen00fl_merrill6115872.cl"},
	{0x0D87B5046D31ED48ULL, "6072719_m.fxe"},
	{0x0D87D39E6EE51A40ULL, "vxht_0_hightown_n_b26_0_0.vlm"},
	{0x0D87F270955EC50FULL, "6029060_m.fxe"},
	{0x0D87FB1F288885F4ULL, "nat300ip_lore_item_chest.utp"},
	{0x0D8855564FC304C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_malepc.ani"},
	{0x0D886D93CFF7228EULL, "435005434.ogg"},
	{0x0D8889A7E23CAE28ULL, "qun200_party_comment6095254.cl"},
	{0x0D88E54A52854451ULL, "dr_0_deeproads_5842.rml"},
	{0x0D88EFBFAACB93B3ULL, "6071822_f.fxe"},
	{0x0D89456DA83A23A0ULL, "vxlt_gallowstemplar_6si_0_0.vlm"},
	{0x0D898ADCA65FF9B5ULL, "dae000_meredith6085276.cl"},
	{0x0D89AE8DD284A5C7ULL, "277073272.ogg"},
	{0x0D89B0125F4BB894ULL, "193564916.ogg"},
	{0x0D8A7280E5EAE204ULL, "vxca_caverns1_ijv_0_0.vlm"},
	{0x0D8A7ABEB48ED6A5ULL, "6094872_f.fxe"},
	{0x0D8AFCFFB1F73D44ULL, "822259431.ogg"},
	{0x0D8B2EF9EA93524CULL, "344569801.ogg"},
	{0x0D8B8887A1C2AD7EULL, "6116554_m.fxe"},
	{0x0D8CCCF8A1954A6EULL, "vxht_0_hightown_a3_c0y_0_0.vlm"},
	{0x0D8D26F89D783D96ULL, "llac_ca_caverns1_mines_24421_7.dds"},
	{0x0D8D50DDF1B321D9ULL, "vxca_caverns1_fmi_0_0.vlm"},
	{0x0D8E1B424B565988ULL, "vxdr_primevaltomb_531_0_1.vlm"},
	{0x0D8E68D19E903948ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_buff.gad"},
	{0x0D8F0C18EB2CCE46ULL, "vxht_mansion2b_1po_0_1.vlm"},
	{0x0D8F25B11D6F71CBULL, "vxlt_hangedman_n_2hq_0_0.vlm"},
	{0x0D8F7CA84515662FULL, "vxht_0_hightown_a2_57f_0_0.vlm"},
	{0x0D8FC9FB88E6A337ULL, "vxht_brothel_n_5mh_0_0.vlm"},
	{0x0D8FE0522ABF29ECULL, "hlac_lt_undercityencounter_-1_35z.dds"},
	{0x0D8FE9690220A5B0ULL, "6077297_m.fxe"},
	{0x0D901ECDE50290F4ULL, "hlac_ht_mansion_n_3364_1y.dds"},
	{0x0D90C132244E1AF9ULL, "vxca_cavern1_haunted_igm_0_0.vlm"},
	{0x0D90F0C6042C9BBDULL, "hlac_lt_hangedman_n_-1_21x.dds"},
	{0x0D915FD517D578D0ULL, "vxlt_warehouse_n_109_0_0.vlm"},
	{0x0D91A300EB341141ULL, "vxw_sundermount_2m9_0_0.vlm"},
	{0x0D92650076E714F0ULL, "6177859_f.fxe"},
	{0x0D927C5F6FDE551FULL, "91257650.ogg"},
	{0x0D92C63192AF9117ULL, "6061851_m.fxe"},
	{0x0D92F97F84D59EE3ULL, "lt_0_lowtown_n_33_0_ll.mmh"},
	{0x0D92FC9089A2F6BFULL, "vxw_blightlands_g5_0_0.vlm"},
	{0x0D936247742AEA1CULL, "framebuffereffects.mat.dx11"},
	{0x0D93698F6B5B9C25ULL, "llac_ca_caverns2_5802_37.dds"},
	{0x0D937003A475AF38ULL, "6201265_m.fxe"},
	{0x0D939A835B111DECULL, "vxw_blightlands_v9_0_0.vlm"},
	{0x0D93BD561C65803BULL, "6034681_f.fxe"},
	{0x0D93E04AC622AD21ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv1.ani"},
	{0x0D942F83DFF02E68ULL, "art\\characters\\playercharacter\\humanmale\\arm_medd.phy"},
	{0x0D947F35340101EFULL, "vxw_encounter2_dc_0_0.vlm"},
	{0x0D95612ECBBAEC21ULL, "vxlt_0_lowtown_a2_eu6_0_0.vlm"},
	{0x0D959B3642369985ULL, "vxlt_0_lowtown_a2_n_c_db3_0_0.vlm"},
	{0x0D95E49572BD64A5ULL, "vxlt_0_lowtown_a2_n_gff_0_0.vlm"},
	{0x0D962416E8066384ULL, "vxht_mansion2_315_0_0.vlm"},
	{0x0D965D7F67160A2CULL, "llac_dr_slavers_3217_5.dds"},
	{0x0D969D79A68FB347ULL, "mcr370_assignments6081788.cl"},
	{0x0D972480E3B505ACULL, "6125554_m.fxe"},
	{0x0D97A37CD2134C12ULL, "ss_matttest.cut"},
	{0x0D9819B39299F213ULL, "fen210st_hadriana.stg"},
	{0x0D985CB9A9A26868ULL, "hlac_lt_0_lowtown_docks_a3_-2_2x.dds"},
	{0x0D989D2AC8FCDBE2ULL, "vxht_mansion2_1ma_0_0.vlm"},
	{0x0D98AF1A5893257CULL, "vxlt_0_lowtown_docks_a3_n_d3q_0_0.vlm"},
	{0x0D98E45DD7E0FA84ULL, "vxca_caverns1_imw_0_0.vlm"},
	{0x0D991E33F1B821EBULL, "vxlt_0_lowtown_n_fip_0_0.vlm"},
	{0x0D99DB5E3179017AULL, "6076497_f.fxe"},
	{0x0D9A31E8BE1DB2C4ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_newmetal_d.dds"},
	{0x0D9A69B5970ECF5CULL, "qun110_petrice6018578.cl"},
	{0x0D9AB7CD66BAEA2BULL, "hlac_lt_gallowscourtyard_-131072_2x.dds"},
	{0x0D9ACC8AECB33F4CULL, "391114744.ogg"},
	{0x0D9AEB4AC628DEC3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv1.gad"},
	{0x0D9B3ED5938B62EEULL, "vxht_0_hightown_a3_c5z_0_0.vlm"},
	{0x0D9B7A21E40EE879ULL, "mag211_marethari6025317.cl"},
	{0x0D9BB373A336AB70ULL, "crt_pride_demon.bnk"},
	{0x0D9C1AD4A3B0B5C9ULL, "hla_ht_0_hightown_a2_1_0z.dds"},
	{0x0D9CA422D417E441ULL, "257050491.ogg"},
	{0x0D9CDD3F76EF1E33ULL, "vxht_mansion2_2xt_0_0.vlm"},
	{0x0D9D14EE35276FEBULL, "qcr210cr_elven_supporter_rg.utc"},
	{0x0D9D23992E4DFF13ULL, "vxlt_0_lowtown_docks_a3_n_d0b_0_1.vlm"},
	{0x0D9D2FC257D2AD91ULL, "dae000_bartrand6206910.cl"},
	{0x0D9DBF88E4EE727DULL, "art\\characters\\weapons\\unique\\aveline\\w_shd_ave.mao"},
	{0x0D9DC388E4EE78B6ULL, "art\\characters\\weapons\\unique\\aveline\\w_shd_ave.mmh"},
	{0x0D9DCD88E4EE89B4ULL, "art\\characters\\weapons\\unique\\aveline\\w_shd_ave.msh"},
	{0x0D9E551384917401ULL, "hlac_lt_0_lowtown_a2_n_131069_0y.dds"},
	{0x0D9E854C69524C49ULL, "vxht_0_hightown_a2o_0_0.vlm"},
	{0x0D9ECF99F38D7173ULL, "art\\vfx\\materials\\vfx_d_impact_add.mao"},
	{0x0D9ED41F9836B429ULL, "dae000_friend_turn_in6179269.cl"},
	{0x0D9ED4B56464D2E2ULL, "terrain_blend6.psh.fxo"},
	{0x0D9EE8BFDAB47736ULL, "pro000cr_refugee5.utc"},
	{0x0D9F680030D62362ULL, "mer221_sabin6040021.cl"},
	{0x0D9F956B11D4FDB9ULL, "vxht_mansion2b_3gn_0_0.vlm"},
	{0x0D9FC2740C0D9D5DULL, "hm_ave3_insurgent1.mor"},
	{0x0DA026B831F2BB6DULL, "w_encounter2_camp_7_0.msh"},
	{0x0DA0285341AF9EE9ULL, "152313039.ogg"},
	{0x0DA02BD810032574ULL, "ht_brothel_n_mmi.gff"},
	{0x0DA034B831F2D2A0ULL, "w_encounter2_camp_7_0.mao"},
	{0x0DA038B831F2D9F3ULL, "w_encounter2_camp_7_0.mmh"},
	{0x0DA11EDD2A95CCD4ULL, "lgt210tr_frame_flower.utt"},
	{0x0DA14A102839F86AULL, "clf260_fell_orden.cnv"},
	{0x0DA14E37695EEC35ULL, "318366800.ogg"},
	{0x0DA16E6D7B990F3FULL, "vxw_blightlands_h4_0_0.vlm"},
	{0x0DA1C84AC62EC1A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv1.evt"},
	{0x0DA20783712831D2ULL, "hlac_lt_0_lowtown_docks_n_-262144_0y.dds"},
	{0x0DA20D02EF3E60BBULL, "6133060_m.fxe"},
	{0x0DA2C31971EDB2A4ULL, "908252040.ogg"},
	{0x0DA2F11225F71A6FULL, "vxw_encounter2_ob_0_0.vlm"},
	{0x0DA33FF43349E5ABULL, "gen00fl_anders6183809.cl"},
	{0x0DA3D30BBDB104A1ULL, "vxlt_0_lowtown_a2_ex5_0_0.vlm"},
	{0x0DA42C10283BFA92ULL, "clf260_fell_orden.bnk"},
	{0x0DA445010972E293ULL, "darktown_fem_fereldan_sta6152801.cl"},
	{0x0DA4567264B37D25ULL, "vxlt_0_lowtown_a2_n_gce_0_0.vlm"},
	{0x0DA45FCDDA0CFD88ULL, "665826962.ogg"},
	{0x0DA472719EF9F034ULL, "6103327_m.fxe"},
	{0x0DA5D5D76A4FD46DULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ci_basic.msh"},
	{0x0DA5E3D76A4FEBA0ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ci_basic.mao"},
	{0x0DA5E7D76A4FF2F3ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ci_basic.mmh"},
	{0x0DA701A617A930D9ULL, "937334260.ogg"},
	{0x0DA76E42C8110006ULL, "art\\vfx\\environment\\vfx_tiki_fire.mmh"},
	{0x0DA7798C0639C29EULL, "dreev_item_acquired.ncs"},
	{0x0DA82FC915F1C00DULL, "llac_ht_chantry_2534_15.dds"},
	{0x0DA847156F870047ULL, "218805669.ogg"},
	{0x0DA91CE918CE0E1FULL, "gen00fl_anders6119464.cl"},
	{0x0DA91F406765AFE4ULL, "vxlt_darktown_58c_0_0.vlm"},
	{0x0DAA527188A63023ULL, "sse_ss_ambient_f.gda"},
	{0x0DAAC62F5A47F35DULL, "6116056_m.fxe"},
	{0x0DAB78B8F1E6EE76ULL, "seb221_flora_harimann6177695.cl"},
	{0x0DABF2F15C6FD5F2ULL, "mer320_hubert6089685.cl"},
	{0x0DAC2D275180A40DULL, "mag101_aombanter6052746.cl"},
	{0x0DAD00ADC8CA4BD8ULL, "560433250.ogg"},
	{0x0DAD3275C810C617ULL, "vxlt_darktown_5ky_0_0.vlm"},
	{0x0DAD6E59E67B2E39ULL, "leave_07_l3.mao"},
	{0x0DAE525BFEEDC7D4ULL, "vxca_caverns1_mines_cp3_0_0.vlm"},
	{0x0DAE61C8AA334863ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_enter.ani"},
	{0x0DAE820CB8F72CE7ULL, "gen00fl_sebastian6188618.cl"},
	{0x0DAE8AF806B450FBULL, "gen00fl_varric6076079.cl"},
	{0x0DAEC5C8D9FEF333ULL, "6077812_m.fxe"},
	{0x0DAF0E93133AF811ULL, "llac_w_bonepit_458752_0.dds"},
	{0x0DAF111A89DD6085ULL, "31334373.ogg"},
	{0x0DAF29D63A5FAFD9ULL, "hlac_lt_0_lowtown_n_1_10y.dds"},
	{0x0DAF58268157B5DBULL, "hlac_ht_0_hightown_a3_-196609_0y.dds"},
	{0x0DAF5ED6B2E36518ULL, "vxw_encounter2_pu_0_1.vlm"},
	{0x0DAFE04A6D8F27BFULL, "vxw_blightlands_m7_0_0.vlm"},
	{0x0DB04B8AE1A3EFC0ULL, "vxdr_primevalentrance_347_0_0.vlm"},
	{0x0DB06307F8A4AA97ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_plaster.dds"},
	{0x0DB0CB50826089E2ULL, "lgt200cr_pride_abomination.utc"},
	{0x0DB0FA9C2FC6B28AULL, "lgt210cr_qun_blk_qunari.utc"},
	{0x0DB113498A077976ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb7.msh"},
	{0x0DB11D498A078A74ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb7.mmh"},
	{0x0DB14E8AAF35A224ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_escape.ani"},
	{0x0DB162EF17ED32EFULL, "vxw_encounter2_jm_0_0.vlm"},
	{0x0DB1722CBB2D5317ULL, "233758685.ogg"},
	{0x0DB191848A10F969ULL, "vxht_0_hightown_n_5yi_0_0.vlm"},
	{0x0DB1B1806C7B9214ULL, "6107100_f.fxe"},
	{0x0DB1F3F9EF64A4C8ULL, "99495132.ogg"},
	{0x0DB21013B103D2DFULL, "zz_ave_debug6134008.cl"},
	{0x0DB297C2B7D4340BULL, "gen00fl_isabela6090879.cl"},
	{0x0DB2C0ACEFD20DB6ULL, "vxht_0_hightown_n_9p9_0_0.vlm"},
	{0x0DB3F9676DB4D57EULL, "zz_mer_debug6028951.cl"},
	{0x0DB3FB40745A7229ULL, "vxht_keep_619_0_0.vlm"},
	{0x0DB48D82DB961E98ULL, "382057214.ogg"},
	{0x0DB4BD4076E741BAULL, "6105518_m.fxe"},
	{0x0DB4D40B22DA246CULL, "llac_lt_0_lowtown_a3_n_c_-196609_3.dds"},
	{0x0DB50DF10CFAA523ULL, "6206952_f.fxe"},
	{0x0DB51CE5AD094B36ULL, "vxw_encounter3_7d_0_0.vlm"},
	{0x0DB5280F0562323BULL, "vxht_bartrand_n_5vy_0_0.vlm"},
	{0x0DB59940FCD35141ULL, "hlac_ht_0_hightown_a2_-65537_2y.dds"},
	{0x0DB59F1A77DAA20FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_wallcap.phy"},
	{0x0DB5A39336EFEEECULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_0x.dds"},
	{0x0DB5ECF9FA84EA8CULL, "6066394_m.fxe"},
	{0x0DB65EB8A97B866FULL, "59345423.ogg"},
	{0x0DB689ACB4B37B35ULL, "lt_0_lowtown_docks_29_0_lc.mmh"},
	{0x0DB6AC01D4381C51ULL, "6061277_m.fxe"},
	{0x0DB6C560ACDE3406ULL, "llac_lt_0_lowtown_a2_n_c_131072_6.dds"},
	{0x0DB74B8AAF3A0886ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_escape.gad"},
	{0x0DB760E92304E9C8ULL, "t3_shd_lst.tnt"},
	{0x0DB793D5EF7CD0BBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11b.phy"},
	{0x0DB7A412E3274BA4ULL, "6129824_m.fxe"},
	{0x0DB8A866A1F89577ULL, "gen_im_trs_cot_dua_dag_r1.uti"},
	{0x0DB9FC6E629DA511ULL, "vxlt_hangedman_1tg_0_0.vlm"},
	{0x0DBA3F3556FC47C9ULL, "gen00fl_aveline6139265.cl"},
	{0x0DBB79C8AA3DFB5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_enter.evt"},
	{0x0DBBA452BA06DE97ULL, "vxlt_darktown_5lx_0_0.vlm"},
	{0x0DBBFA1331819A43ULL, "6140354_m.fxe"},
	{0x0DBC4C0F04DAF1BCULL, "mcr370_samson6087533.cl"},
	{0x0DBCBB33DAE9AF11ULL, "vxht_mansion_n_5me_0_0.vlm"},
	{0x0DBCDB8FB1D01B8CULL, "gen00fl_fenris6103327.cl"},
	{0x0DBD34B16BA376F8ULL, "6147661_f.fxe"},
	{0x0DBE253C302E1E83ULL, "lt_0_lowtown_a2_n_c_35_0_ll.mmh"},
	{0x0DBE468AAF401F65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_escape.evt"},
	{0x0DBE52275F85403FULL, "vxw_blightlands_v6_0_0.vlm"},
	{0x0DBEBD67D39A0840ULL, "vxdr_primevalentrance_314_0_0.vlm"},
	{0x0DBEC0E4601278CBULL, "850502517.ogg"},
	{0x0DBEF0BCD32A91BBULL, "6139660_m.fxe"},
	{0x0DBF18B0519422DAULL, "vxca_caverns2_3cf_0_0.vlm"},
	{0x0DBF97083BDBD0E7ULL, "dae000_hate_follower_react6103203.cl"},
	{0x0DBFC3896880C010ULL, "mag360_samson6042060.cl"},
	{0x0DC0044912EA5006ULL, "vxca_caverns1_e8b_0_0.vlm"},
	{0x0DC0099BCC350C3BULL, "zz_vrc_debug6200225.cl"},
	{0x0DC09573A355943AULL, "crt_pride_demon.wbd"},
	{0x0DC0A54354917278ULL, "224198037.ogg"},
	{0x0DC0E95E05ACDD53ULL, "17514919.ogg"},
	{0x0DC1341A1F59F4A2ULL, "6093587_m.fxe"},
	{0x0DC17A111AD67512ULL, "mag101_grace6034201.cl"},
	{0x0DC17D2BAC11B46DULL, "w_encounter2_56_0.msh"},
	{0x0DC18B2BAC11CBA0ULL, "w_encounter2_56_0.mao"},
	{0x0DC18F2BAC11D2F3ULL, "w_encounter2_56_0.mmh"},
	{0x0DC2466860EAE733ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand_2.evt"},
	{0x0DC26325BB15FA47ULL, "380216399.ogg"},
	{0x0DC278C8AA44184DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_enter.gad"},
	{0x0DC287D07EE8AF72ULL, "106416055.ogg"},
	{0x0DC2E772ECBD89ABULL, "157220389.ogg"},
	{0x0DC301986EBBE2EFULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.impa.gad"},
	{0x0DC37F582DFB206FULL, "6013267_m.fxe"},
	{0x0DC399EBF7584ABBULL, "vxht_bartrand_n_5yx_0_0.vlm"},
	{0x0DC3B10189988CE5ULL, "vxht_0_hightown_n_b5n_0_0.vlm"},
	{0x0DC3EBABBDBF20DFULL, "vxca_caverns1_mines_gk4_0_1.vlm"},
	{0x0DC42422DFC896DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_urgent_walk.anb"},
	{0x0DC476DC86B217FDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_quill_sa.dds"},
	{0x0DC47857E5FC72FFULL, "hlac_w_sundermount_0_0z.dds"},
	{0x0DC4CCF0255098BEULL, "6179254_m.fxe"},
	{0x0DC570623796382EULL, "902786864.ogg"},
	{0x0DC59D8B67F140EAULL, "fx_air_twirl.mao"},
	{0x0DC65A69FA818B1FULL, "6088766_m.fxe"},
	{0x0DC678B262638216ULL, "6069861_m.fxe"},
	{0x0DC6852DDE258475ULL, "vxht_0_hightown_b95_0_0.vlm"},
	{0x0DC6A802FDD71C17ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medd_0t.dds"},
	{0x0DC720FC9AF3897FULL, "339850988.ogg"},
	{0x0DC727EE14965169ULL, "22043727.ogg"},
	{0x0DC7799C66AAC14AULL, "sto_fl_light.mao"},
	{0x0DC86E4B5493BD91ULL, "vxlt_hangedman_1sf_0_0.vlm"},
	{0x0DC8B3BD1AD29072ULL, "gr_clu.mao"},
	{0x0DC93F6860F0FA74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand_2.gad"},
	{0x0DCA16986EC2249CULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.impa.evt"},
	{0x0DCA5EC44383412DULL, "vxlt_undercityserialkiller_18p_0_0.vlm"},
	{0x0DCA94855C20EC5DULL, "vxlt_0_lowtown_n_h87_0_0.vlm"},
	{0x0DCAF7BD3BC25234ULL, "gallows_avatar.cut"},
	{0x0DCAFFF42C10558FULL, "gen00fl_anders6111381.cl"},
	{0x0DCB4B7250A9DB35ULL, "plt_lgt000pt_ultra_5_01.plo"},
	{0x0DCB8474B2FD4286ULL, "6107439_m.fxe"},
	{0x0DCBCF8260F0228EULL, "hightown_fem_cityguard_mob6168529.cl"},
	{0x0DCBEA7C76CD8245ULL, "850422248.ogg"},
	{0x0DCCCBE55EF895D6ULL, "hf_arthuris.mor"},
	{0x0DCCD0BED83DF42FULL, "vrc321_varric6087124.cl"},
	{0x0DCD9E94A619F3C0ULL, "mag211_marethari6025152.cl"},
	{0x0DCE18F5B4399D7FULL, "6161610_m.fxe"},
	{0x0DCE249A3D22A2A6ULL, "lgt220ip_shp_mag_devrona.utp"},
	{0x0DCE3FF83054CE08ULL, "vxdr_0_deeproads_d3o_0_0.vlm"},
	{0x0DCE46A8FBD963EFULL, "vxw_encounter2_po_0_0.vlm"},
	{0x0DCE9F76A9C3C615ULL, "hlac_lt_gallowscourtyard_n_-2_1x.dds"},
	{0x0DCEE2D20B5D39DFULL, "vxca_caverns1_mines_iko_0_0.vlm"},
	{0x0DCF0B3AF4C0402AULL, "vxht_brothel_n_6d5_0_0.vlm"},
	{0x0DCF0E18E3392057ULL, "hla_ht_0_hightown_n_65538_1z.dds"},
	{0x0DCFF096A3867096ULL, "6074489_m.fxe"},
	{0x0DCFF4822C00C797ULL, "6097898_m.fxe"},
	{0x0DD024F8E84DEE11ULL, "6083057_m.fxe"},
	{0x0DD04E6860F73252ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand_2.ani"},
	{0x0DD12EF71FF5DF93ULL, "pro000_varric6170655.cl"},
	{0x0DD1998066435399ULL, "fadetosepia.mao"},
	{0x0DD1F810A77EB6DEULL, "875149338.ogg"},
	{0x0DD20AB608D4B608ULL, "art\\vfx\\textures\\vfx_vollight_dust.dds"},
	{0x0DD222DE7B8EA565ULL, "vxht_0_hightown_n_b0m_0_0.vlm"},
	{0x0DD26CF5E950597BULL, "6103716_f.fxe"},
	{0x0DD2C9D6C76E3DAEULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressureup.ani"},
	{0x0DD39BD6AA2BDA4BULL, "224711022.ogg"},
	{0x0DD4184BF731F463ULL, "gen00fl_aveline6106948.cl"},
	{0x0DD43BC59651ABB6ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_lute.msh"},
	{0x0DD445C59651BCB4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_lute.mmh"},
	{0x0DD448E2F09A1BFFULL, "vxht_0_hightown_c5n_0_0.vlm"},
	{0x0DD449C59651C3C7ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_lute.mao"},
	{0x0DD467A6F3C4ACDFULL, "6066335_m.fxe"},
	{0x0DD46ADEFFB7333AULL, "6149427_f.fxe"},
	{0x0DD494BD6CCEB322ULL, "vxdr_0_deeproads_buu_0_0.vlm"},
	{0x0DD49B3196528C9CULL, "one300_zevran6076573.cl"},
	{0x0DD5064ABFC2B5A1ULL, "vxdr_primevalentrance_333_0_0.vlm"},
	{0x0DD55103B64DBA48ULL, "l_mag211ar_ariannis_house.bnk"},
	{0x0DD556D1A8983C54ULL, "hlac_lt_ruinedhouse_n_-1_6z.dds"},
	{0x0DD56B7D8AE824B0ULL, "6089490_m.fxe"},
	{0x0DD62E6E3F2F9AA2ULL, "349863308.ogg"},
	{0x0DD6752277D477B7ULL, "820105412.ogg"},
	{0x0DD6AC1609C4CEB4ULL, "245824188.ogg"},
	{0x0DD6C9AA33C53FE8ULL, "vxlt_0_lowtown_fgn_0_0.vlm"},
	{0x0DD6D0153E39ECEDULL, "153550143.ogg"},
	{0x0DD71E986ECCBCB5ULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.impa.ani"},
	{0x0DD75853B06F5D37ULL, "dae000ip_crowd_mf5_talk11.utp"},
	{0x0DD8FB6F4E9A2E77ULL, "lt_0_lowtown_docks_173_0.msh"},
	{0x0DD90D6F4E9A4CDAULL, "lt_0_lowtown_docks_173_0.mao"},
	{0x0DD9116F4E9A53E9ULL, "lt_0_lowtown_docks_173_0.mmh"},
	{0x0DD9315607386642ULL, "vxlt_0_lowtown_docks_n_b94_0_0.vlm"},
	{0x0DD93A16780FB2FCULL, "118104706.ogg"},
	{0x0DD95D0DC997360FULL, "gen00fl_varric6093636.cl"},
	{0x0DD9F8790B26DF39ULL, "llac_lt_0_lowtown_docks_n_0_0.dds"},
	{0x0DDA512AC1CEA4C9ULL, "gen_im_arm_glv_med_act3_02.uti"},
	{0x0DDA67B8265E4275ULL, "963778195.ogg"},
	{0x0DDAAAE148ECEADFULL, "707113905.ogg"},
	{0x0DDAC2F98C46B1C6ULL, "llac_ht_chantry_n_2534_37.dds"},
	{0x0DDB03D45785EB4EULL, "art\\characters\\playercharacter\\children\\mt_bas_kf_scalpm1.mao"},
	{0x0DDB4B2232F48C63ULL, "861694627.ogg"},
	{0x0DDB5651AAE7C4E8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_lgtd_n.dds"},
	{0x0DDB7B0118D73840ULL, "477297282.ogg"},
	{0x0DDB86404B8999CDULL, "vxca_cavern1_haunted_jg6_0_0.vlm"},
	{0x0DDBAC7328C76416ULL, "vxlt_0_lowtown_a2_n_c_gij_0_0.vlm"},
	{0x0DDBFACC0C545A2CULL, "383420846.ogg"},
	{0x0DDBFCB2AB888FB4ULL, "6081689_m.fxe"},
	{0x0DDC8F9E43777771ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\stun.ani"},
	{0x0DDC9F51C996B857ULL, "lt_0_lowtown_docks_201_0.msh"},
	{0x0DDCB151C996D6BAULL, "lt_0_lowtown_docks_201_0.mao"},
	{0x0DDCB551C996DDC9ULL, "lt_0_lowtown_docks_201_0.mmh"},
	{0x0DDE6273957C8916ULL, "6079789_m.fxe"},
	{0x0DDEC55F30F1E928ULL, "vxca_caverns2_80n_0_0.vlm"},
	{0x0DE0DED2DB4671FBULL, "6104616_f.fxe"},
	{0x0DE121EC1438DDA4ULL, "186334698.ogg"},
	{0x0DE1B9CCBF8602C7ULL, "vxlt_0_lowtown_docks_a3_n_b9b_0_0.vlm"},
	{0x0DE1C5D14D756E6FULL, "cth_cm2.tnt"},
	{0x0DE2391B1405E06EULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_lfoot.ani"},
	{0x0DE2A30744BE9C75ULL, "w_sundermount_140_0.phy"},
	{0x0DE3069A5EC4CBA2ULL, "vxdr_0_deeproads_bpv_0_0.vlm"},
	{0x0DE33A1C7EEB92E1ULL, "6024989_f.fxe"},
	{0x0DE33E23DE6DBC1FULL, "6086966_m.fxe"},
	{0x0DE345E24A005E6CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgertes.msh"},
	{0x0DE34BE24A00685EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgertes.mmh"},
	{0x0DE357E24A007CF5ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgertes.mao"},
	{0x0DE37AD5EFA1E15AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11b.msh"},
	{0x0DE38CD5EFA1FFA0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11b.mmh"},
	{0x0DE39A9E437DA813ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\stun.gad"},
	{0x0DE3EA960FB22069ULL, "81570327.ogg"},
	{0x0DE40ED3E10B7525ULL, "vxlt_0_lowtown_docks_n_c4y_0_0.vlm"},
	{0x0DE40F113E1220EEULL, "6035297_m.fxe"},
	{0x0DE455EE3E0A4E6BULL, "917328117.ogg"},
	{0x0DE45A5874A30EBEULL, "mag291_torpor6116237.cl"},
	{0x0DE4F90E795CB991ULL, "vxlt_0_lowtown_f24_0_0.vlm"},
	{0x0DE6FE9290B26C77ULL, "vxht_bartrand_n_5vm_0_0.vlm"},
	{0x0DE711FC08FA62DDULL, "6074495_m.fxe"},
	{0x0DE750D13F112C9BULL, "6188612_f.fxe"},
	{0x0DE75401DD15C0F9ULL, "llac_ht_chantry_2534_2.dds"},
	{0x0DE772B334FCBD70ULL, "6095345_m.fxe"},
	{0x0DE7A2D6C78057E8ULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressureup.gad"},
	{0x0DE7A332F92E7EC2ULL, "vxlt_0_lowtown_docks_n_b65_0_0.vlm"},
	{0x0DE7D6D720982FB3ULL, "vxca_caverns1_hm9_0_0.vlm"},
	{0x0DE857E204654C12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_damage_reduction.gad"},
	{0x0DE8C8D411D5D131ULL, "6162565_m.fxe"},
	{0x0DE96789F05B17C4ULL, "35611378.ogg"},
	{0x0DE9C40744C4F2A1ULL, "w_sundermount_140_0.rml"},
	{0x0DEA2043B4C87704ULL, "qcr220cr_noblewoman3.utc"},
	{0x0DEA979E4383C278ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\stun.evt"},
	{0x0DEAA5073A9C76E1ULL, "hlac_dr_primevaltomb_664_6y.dds"},
	{0x0DEB86A88D6E83EDULL, "529878776.ogg"},
	{0x0DEBADE249C74293ULL, "dr_dungeon_mmi.gff"},
	{0x0DEBDDC610142576ULL, "ave200_donnic6110730.cl"},
	{0x0DEBEEF4D8AC712AULL, "vxht_brothel_n_6j7_0_0.vlm"},
	{0x0DEC053EF2F5457BULL, "6136445_m.fxe"},
	{0x0DEC28EDFC576265ULL, "vxdr_primevalentrance_39m_0_0.vlm"},
	{0x0DECAE9200AAEE64ULL, "hlac_w_encounter2_262142_0y.dds"},
	{0x0DECAF7D60BBA2BCULL, "254329449.ogg"},
	{0x0DED62031B5EE995ULL, "zz_dae_debug6184889.cl"},
	{0x0DED9846A694C4CBULL, "394993736.ogg"},
	{0x0DEDF82ECCDF00D3ULL, "art\\levels\\ztools\\default\\generic\\0wi_angle.phy"},
	{0x0DEE2BD26BB8523BULL, "231059595.ogg"},
	{0x0DEE458A7670E8FFULL, "t3_wdg_bn1.tnt"},
	{0x0DEF311B14105CBFULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_lfoot.evt"},
	{0x0DEF5AE2046B7049ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_damage_reduction.evt"},
	{0x0DEFED3B8B7D0BE5ULL, "llac_lt_0_lowtown_a2_n_c_65535_5.dds"},
	{0x0DF06F466BA83FB9ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb1.msh"},
	{0x0DF079466BA850A7ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb1.mmh"},
	{0x0DF1A4137638B120ULL, "character_character_p191.psh.fxo.dx11.assembly"},
	{0x0DF1C0BA5FA41247ULL, "6120416_f.fxe"},
	{0x0DF1FA9DA8A82906ULL, "259967446.ogg"},
	{0x0DF220EE0585410AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_standing_cheer.gad"},
	{0x0DF24F376ED455B0ULL, "6087690_m.fxe"},
	{0x0DF285BFCB0DCAEFULL, "dae000_gamlen6166382.cl"},
	{0x0DF2B296138630B0ULL, "6140933_m.fxe"},
	{0x0DF33FB7B5BD34B4ULL, "mer141_jansen6085166.cl"},
	{0x0DF343492994859BULL, "art\\vfx\\materials\\vfx_fire_link_blend.mao"},
	{0x0DF3C09466092FB9ULL, "prm000im_hindsight_belt.uti"},
	{0x0DF3F383B0E97B42ULL, "684785639.ogg"},
	{0x0DF4F6A4B93C90FCULL, "343331210.ogg"},
	{0x0DF5F04BAE9C55A2ULL, "2222095.ogg"},
	{0x0DF6121B141646A8ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_lfoot.gad"},
	{0x0DF6DF2541A32D0AULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masa.mao"},
	{0x0DF6E32541A33459ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masa.mmh"},
	{0x0DF6ED2541A34547ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masa.msh"},
	{0x0DF78A1ABA614262ULL, "llac_ca_caverns1_24423_4.dds"},
	{0x0DF7ADF9AAC9E737ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml_b.dura.gad"},
	{0x0DF85D05FA80446CULL, "zz_seb_debug6214536.cl"},
	{0x0DF876C4B70A2DA9ULL, "6114872_m.fxe"},
	{0x0DF8A09FC619634AULL, "vxht_0_hightown_a3_am8_0_0.vlm"},
	{0x0DF8AA7DB074C9F5ULL, "946067711.ogg"},
	{0x0DF9002939586643ULL, "gen00fl_bethany6101218.cl"},
	{0x0DF903EE058B2EA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_standing_cheer.evt"},
	{0x0DF91EFD6E2AB603ULL, "gen00fl_sebastian6180492.cl"},
	{0x0DF990C8D98CD80CULL, "786607815.ogg"},
	{0x0DF9927A9BCD5706ULL, "hlac_ca_caverns2_5802_58x.dds"},
	{0x0DFA39755CB91CA9ULL, "hlac_lt_0_lowtown_-65537_5x.dds"},
	{0x0DFA4303B66CBEF2ULL, "l_mag211ar_ariannis_house.wbd"},
	{0x0DFA838725482577ULL, "gen00fl_fenris6110784.cl"},
	{0x0DFA9ACAEE4D7AE5ULL, "vxdr_primevalentrance_32l_0_0.vlm"},
	{0x0DFAE4B3ED8B843EULL, "749601575.ogg"},
	{0x0DFB1CA2FC658A96ULL, "837006204.ogg"},
	{0x0DFB9DEF8223700BULL, "hlac_lt_0_lowtown_a3_n_c_1_8y.dds"},
	{0x0DFB9FB2152C0DCAULL, "vxlt_gallowstemplar_n_2jc_0_0.vlm"},
	{0x0DFB9FC368824DA8ULL, "vxht_0_hightown_b5h_0_0.vlm"},
	{0x0DFBAE7F24B69382ULL, "6139971_m.fxe"},
	{0x0DFBC358550CD91FULL, "6164781_m.fxe"},
	{0x0DFC96B0E5D9DB6EULL, "gencr_qunari_captain.utc"},
	{0x0DFCD1D931C63411ULL, "character_character_p280.psh.fxo.dx11.assembly"},
	{0x0DFCEFDD10BB1C0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small.ani"},
	{0x0DFD402EDAE0BACBULL, "288343672.ogg"},
	{0x0DFD42E2047784B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_damage_reduction.ani"},
	{0x0DFDC2F9AACE76C4ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml_b.dura.evt"},
	{0x0DFDC51A687D0BEAULL, "gen00fl_merrill6162034.cl"},
	{0x0DFE20E629220292ULL, "mcr371_hostile_aveline6082179.cl"},
	{0x0DFE5BA9DEE7AE3CULL, "6167245_f.fxe"},
	{0x0DFEB6FEEA84DD0DULL, "hlac_dr_dungeon_3216_1z.dds"},
	{0x0DFF1BD1EA32E9F7ULL, "722667049.ogg"},
	{0x0DFF43FAF13868B1ULL, "vxca_cavern1_haunted_imw_0_0.vlm"},
	{0x0DFF4EE27BAE3EAAULL, "6087533_m.fxe"},
	{0x0DFFC173EC401BDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_stumble.evt"},
	{0x0E00008AD61AC1B7ULL, "6060500_m.fxe"},
	{0x0E002AEC84D8CB10ULL, "mrl150_merrill6078329.cl"},
	{0x0E00665D227E23DFULL, "ht_0_hightownexplosion_n_20_0_lc.mmh"},
	{0x0E0204D669D4F62AULL, "vxht_0_hightown_a3_9bi_0_0.vlm"},
	{0x0E0220EA786D6F97ULL, "vxw_woundedcoast_161_0_0.vlm"},
	{0x0E02C506F7E19241ULL, "qun111_ketojan6120459.cl"},
	{0x0E02C855139B149EULL, "930359926.ogg"},
	{0x0E03632FC2526565ULL, "gen00fl_isabela6135619.cl"},
	{0x0E03ED5EB55FA3E5ULL, "hlac_w_bonepitdestroyed_a3_3_0x.dds"},
	{0x0E03FCDD10C1514EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small.gad"},
	{0x0E042C8E50051330ULL, "vxlt_gallowstemplar_2pd_0_0.vlm"},
	{0x0E043A3BECC9331BULL, "hlac_dr_0_deeproads_5843_1z.dds"},
	{0x0E046D0FFDCA99B5ULL, "vxht_mansion2b_2pr_0_0.vlm"},
	{0x0E04714820087AD3ULL, "w_blightlands_166_0.phy"},
	{0x0E0484201DAFD89AULL, "6124775_m.fxe"},
	{0x0E04EDBC4B4C032AULL, "ts_raiders_wpn_table.gda"},
	{0x0E05BB9A46EE45E6ULL, "vxht_0_hightown_n_bu9_0_0.vlm"},
	{0x0E05C6F8027A05C2ULL, "character_character_p171.psh.fxo.dx11.assembly"},
	{0x0E05FBEE0595ABF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_standing_cheer.ani"},
	{0x0E061882B1260695ULL, "vxca_cave_1cl_0_0.vlm"},
	{0x0E062DD016783933ULL, "564513357.ogg"},
	{0x0E0638D9802FDF75ULL, "vxlt_warehouse_n_1c1_0_0.vlm"},
	{0x0E06701E3EDFB9CBULL, "lt_gallowsprison_n_mmap_ph.dds"},
	{0x0E0689FBA7E52D92ULL, "ico_staff_1.dds"},
	{0x0E069951CD565066ULL, "vxlt_gallowstemplar_n_458_0_0.vlm"},
	{0x0E06C7C3ED594002ULL, "6160847_m.fxe"},
	{0x0E06D873EC466109ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_stumble.gad"},
	{0x0E072651C1382361ULL, "vxlt_0_lowtown_a2_gpo_0_0.vlm"},
	{0x0E083E7250DD7312ULL, "plt_lgt000pt_ultra_5_01.nss"},
	{0x0E08570619BB6C36ULL, "6042295_m.fxe"},
	{0x0E086178EDF5FA6FULL, "vxw_blightlands_6f_0_0.vlm"},
	{0x0E0A8124317B6A12ULL, "97213169.ogg"},
	{0x0E0ABD1BAE564793ULL, "vxht_0_hightown_a2_bjf_0_0.vlm"},
	{0x0E0AF7DD10C7672DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small.evt"},
	{0x0E0B63419522EB50ULL, "6190424_m.fxe"},
	{0x0E0B6541EC980D34ULL, "vxlt_0_lowtown_byp_0_0.vlm"},
	{0x0E0B70E5CAB55361ULL, "mag360_thrask6043893.cl"},
	{0x0E0B9A3D80652104ULL, "vxw_encounter3_65_0_0.vlm"},
	{0x0E0BC508C413EBCDULL, "370560824.ogg"},
	{0x0E0BDAF9AADADCBDULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml_b.dura.ani"},
	{0x0E0C184A0E15DBD9ULL, "vxlt_foundry_n_wd_0_0.vlm"},
	{0x0E0C3249DE332F3BULL, "hlac_w_sundermount_131066_0y.dds"},
	{0x0E0C488672F7C258ULL, "6078128_m.fxe"},
	{0x0E0C67C7F47D5A1BULL, "face2_s1.psh.fxo"},
	{0x0E0C6E57AF9AAB0CULL, "6053329_m.fxe"},
	{0x0E0CCDF8E9C44B77ULL, "367748839.ogg"},
	{0x0E0CDFC43A5C4BCEULL, "one110_pryce6088370.cl"},
	{0x0E0D171D3A6D2AB8ULL, "147729235.ogg"},
	{0x0E0DA879F7604756ULL, "vxca_caverns1_mines_iip_0_0.vlm"},
	{0x0E0DA973EC4C3057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_stumble.ani"},
	{0x0E0DB5D7E32E8131ULL, "vxca_cavern1_haunted_ijv_0_0.vlm"},
	{0x0E0DC475B170111BULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed.dura.evt"},
	{0x0E0E054AC4EE3C5FULL, "6059148_m.fxe"},
	{0x0E0E33B3DEDF902EULL, "llac_dr_dungeon_3214_2.dds"},
	{0x0E0E4F356D8D5A34ULL, "773858585.ogg"},
	{0x0E0E528E1163E051ULL, "475887682.ogg"},
	{0x0E0E7E211020EBA1ULL, "6181926_m.fxe"},
	{0x0E0EBA57730B4C22ULL, "315887770.ogg"},
	{0x0E0EEC09FC021F66ULL, "vxht_keep_6c0_0_0.vlm"},
	{0x0E0F20F822E3D5FBULL, "887837806.ogg"},
	{0x0E0F7FCD4A14D75EULL, "426507163.ogg"},
	{0x0E0F9B6D0256B5AAULL, "mag360_samson6044886.cl"},
	{0x0E0FD551424FA560ULL, "vxca_caverns2_62p_0_2.vlm"},
	{0x0E0FE7304E1279CAULL, "cd_mag291_torpor.wbd"},
	{0x0E1092C76A638817ULL, "vxw_woundedcoast_110_0_0.vlm"},
	{0x0E10A79C0E623811ULL, "vxlt_hangedman_1jy_0_0.vlm"},
	{0x0E10AF07828CCF48ULL, "vxlt_0_lowtown_a2_gpj_0_0.vlm"},
	{0x0E112E94CF8DEB5DULL, "vxht_0_hightown_9zc_0_0.vlm"},
	{0x0E116A37492F3FC6ULL, "vxw_sundermount_2f3_0_0.vlm"},
	{0x0E118DA5B0FD0CB2ULL, "vxlt_undercityencounter_142_0_1.vlm"},
	{0x0E11C797AC8E4A07ULL, "vxlt_gallowscourtyard_7br_0_0.vlm"},
	{0x0E11DCC8B8BFA761ULL, "6202067_m.fxe"},
	{0x0E11E07CAFC59595ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_0y.dds"},
	{0x0E12554F7D6DAB58ULL, "follower_banter6141913.cl"},
	{0x0E12CFE6CF264F0DULL, "llac_w_bonepitmining_a2_196609_0.dds"},
	{0x0E13BDE5870D8D20ULL, "gen00fl_isabela6096858.cl"},
	{0x0E144A1CA7185830ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys.mmh"},
	{0x0E1465C78A41F11DULL, "mrl310_merrill6119663.cl"},
	{0x0E147FA1D67683AAULL, "llac_ca_cave_2194_12.dds"},
	{0x0E148102F0571F2FULL, "hlac_lt_hangedman_n_-1_28y.dds"},
	{0x0E1499CC12AD164EULL, "vxht_brothel_n_5q1_0_0.vlm"},
	{0x0E14BD75B176245CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed.dura.gad"},
	{0x0E151CF8FA298FCEULL, "136166894.ogg"},
	{0x0E1523A3051FEA7EULL, "6100114_f.fxe"},
	{0x0E16DF698BA5225EULL, "vxlt_darktown_2vh_0_0.vlm"},
	{0x0E173A7BDD300353ULL, "vfx_chunkyparticle_p0.psh.fxo"},
	{0x0E173A80E865254AULL, "vxht_0_hightown_cb2_0_0.vlm"},
	{0x0E176F0198D3CC9DULL, "553780562.ogg"},
	{0x0E17D4131577DA09ULL, "snow_t_f.mao"},
	{0x0E1801C19CDB7BB6ULL, "284856955.ogg"},
	{0x0E184A3A1FE5643FULL, "dre000_dougal6131145.cl"},
	{0x0E186C8E4A41D111ULL, "6042120_m.fxe"},
	{0x0E189AF7BEFD9336ULL, "vxlt_0_lowtown_a2_ekg_0_0.vlm"},
	{0x0E1967DB9D893755ULL, "vxlt_0_lowtown_docks_c51_0_0.vlm"},
	{0x0E197394EB51D419ULL, "6115609_f.fxe"},
	{0x0E1996A0BEA81753ULL, "vxht_keep_n_676_0_0.vlm"},
	{0x0E1A371401CBEAA3ULL, "dre000_varric6042938.cl"},
	{0x0E1A989964F7A175ULL, "317704018.ogg"},
	{0x0E1ABC75B17A8E5AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed.dura.ani"},
	{0x0E1B89F032D93994ULL, "vxw_bonepitmining_a2_2a1_0_0.vlm"},
	{0x0E1CE2379846A7AEULL, "llac_lt_0_lowtown_a3_n_c_-131072_2.dds"},
	{0x0E1D252BB56F61FEULL, "llac_ht_0_hightown_a3_65533_2.dds"},
	{0x0E1D3C77976D98A3ULL, "6103236_m.fxe"},
	{0x0E1D3EC7314EDF47ULL, "ht_0_hightown_n_19_0.rml"},
	{0x0E1D5DE6EDF837E6ULL, "vxht_keep_6l1_0_0.vlm"},
	{0x0E1D72D865AC0143ULL, "6102908_f.fxe"},
	{0x0E1E9AFA68D9BA18ULL, "6165147_m.fxe"},
	{0x0E1F0220653EA738ULL, "6101941_m.fxe"},
	{0x0E1F20E47482E7C8ULL, "vxlt_0_lowtown_a2_gwk_0_0.vlm"},
	{0x0E1F22B222B2EBD3ULL, "hlac_ht_mansion2b_818_18z.dds"},
	{0x0E1F52B9F9A3427DULL, "hlac_ht_mansion2_818_10z.dds"},
	{0x0E1F5E1656A1A1E9ULL, "gen00fl_varric6093627.cl"},
	{0x0E1FD27D49470305ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslb_0n.dds"},
	{0x0E1FDC143B255846ULL, "vxw_sundermount_2k4_0_0.vlm"},
	{0x0E2039749E846287ULL, "vxlt_gallowscourtyard_7gm_0_0.vlm"},
	{0x0E2065F197813875ULL, "vxw_bonepit_zo_0_0.vlm"},
	{0x0E20909F9357AED9ULL, "6079338_f.fxe"},
	{0x0E20CFEB514DCB7CULL, "295444963.ogg"},
	{0x0E20EC54691E075FULL, "sky_blank_exterior.mao"},
	{0x0E213A93CE54E729ULL, "6029057_m.fxe"},
	{0x0E216D1CA7231E4FULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys.anb"},
	{0x0E21B1D91CD7D062ULL, "845972467.ogg"},
	{0x0E21CC0D5F114A76ULL, "w_blightlands_49_0.msh"},
	{0x0E21D60D5F115B74ULL, "w_blightlands_49_0.mmh"},
	{0x0E21DA0D5F116187ULL, "w_blightlands_49_0.mao"},
	{0x0E21DD06FB6B95A0ULL, "vxlt_warehouse_v5_0_0.vlm"},
	{0x0E2265CF1A49A113ULL, "ser270_templar6127488.cl"},
	{0x0E22AAF0344C4C7BULL, "gen00fl_aveline6120777.cl"},
	{0x0E22AB28162DCF42ULL, "art\\characters\\creatures\\ragedemon\\textures\\c_ragedmon_0s.dds"},
	{0x0E22E57F0EC624A4ULL, "gencr_dalish_warrior_ws.utc"},
	{0x0E2317567FFCB5D2ULL, "vxht_keep_n_673_0_0.vlm"},
	{0x0E2322E24A3606F9ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgertes.phy"},
	{0x0E23A6EE167C7C9AULL, "308533721.ogg"},
	{0x0E23CF8C62CA3795ULL, "dae000_mother6002007.cl"},
	{0x0E240A0BA5245461ULL, "vxlt_0_lowtown_a2_gjm_0_0.vlm"},
	{0x0E240A410BC3AC9CULL, "vxht_mansion2b_333_0_0.vlm"},
	{0x0E240D2670F34BCCULL, "gen00fl_aveline6140907.cl"},
	{0x0E2551467D9B3ADEULL, "vxlt_darktown_2yi_0_0.vlm"},
	{0x0E2552F46678402FULL, "zz_isabellakiss.cut"},
	{0x0E25931FBA400DA8ULL, "6102764_m.fxe"},
	{0x0E25B5200B3B1291ULL, "vxw_encounter2_camp_7l_0_0.vlm"},
	{0x0E25D2781BF9E3F7ULL, "ser270_templar.bnk"},
	{0x0E2647B4AC7F07C9ULL, "vxlt_gallowstemplar_3s1_0_0.vlm"},
	{0x0E265EBA69966000ULL, "vxlt_gallowstemplar_5cv_0_0.vlm"},
	{0x0E26B5D8A94F7966ULL, "293216412.ogg"},
	{0x0E26CBCFD96E6FDEULL, "215262589.ogg"},
	{0x0E26D7166FEF7C27ULL, "framebuffereffects_distortion_p0.psh.fxo.dx11.assembly"},
	{0x0E27A5AC9FBE48CDULL, "253518188.ogg"},
	{0x0E28087DB09E2FD3ULL, "vxht_keep_n_629_0_0.vlm"},
	{0x0E28A82E2B78B7ACULL, "136340555.ogg"},
	{0x0E28B4781BFBE60FULL, "ser270_templar.cnv"},
	{0x0E2927A8F5F8C1D0ULL, "vxht_keep_n_3yb_0_0.vlm"},
	{0x0E294AAA36BF413FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackleft.gad"},
	{0x0E29AD13EB91F9ABULL, "vxlt_0_lowtown_a2_n_c_euq_0_0.vlm"},
	{0x0E29B60744FABE90ULL, "w_sundermount_140_0.msh"},
	{0x0E29C40744FAD6AAULL, "w_sundermount_140_0.mmh"},
	{0x0E29C80744FADDE1ULL, "w_sundermount_140_0.mao"},
	{0x0E2A8C33DB4C7856ULL, "vxca_caverns1_mines_icr_0_0.vlm"},
	{0x0E2AE7EEA0F3E066ULL, "93924218.ogg"},
	{0x0E2AEDED470A729EULL, "182923908.ogg"},
	{0x0E2AFBEDC2AB7D71ULL, "vxht_mansion2b_3gv_0_0.vlm"},
	{0x0E2B0D4D77C6224AULL, "art\\characters\\playercharacter\\shared\\textures\\uv_color_guide.dds"},
	{0x0E2B107959A9CD19ULL, "vxht_0_hightown_a3_a2n_0_0.vlm"},
	{0x0E2BAE548963B123ULL, "6104336_m.fxe"},
	{0x0E2BCAF446D32822ULL, "lgt310cr_mag_melson.utc"},
	{0x0E2D2FBC05EBD3CAULL, "vxw_encounter2_p7_0_1.vlm"},
	{0x0E2DB319625795BBULL, "lgt220_mag_apost_apost16156739.cl"},
	{0x0E2E124EB37A1C5DULL, "vxht_0_hightown_9pa_0_0.vlm"},
	{0x0E2E4DF12D1B70C6ULL, "vxw_sundermount_2l5_0_0.vlm"},
	{0x0E2F37E34A80E6F7ULL, "vxht_bartrand_n_5mh_0_0.vlm"},
	{0x0E2F6153CC24FC2CULL, "88256946.ogg"},
	{0x0E2F6F78729C8354ULL, "art\\vfx\\combat\\vfx_r_arc_burst_sh_aoe_p.mmh"},
	{0x0E303FAA36C54D2CULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackleft.evt"},
	{0x0E3044C596A019AFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_lute.phy"},
	{0x0E306562F5054D60ULL, "vxlt_darktown_41e_0_0.vlm"},
	{0x0E307ED68204E1E7ULL, "art\\vfx\\combat\\vfx_m_cre_conj_c.impa.gad"},
	{0x0E308E5BC0806476ULL, "947021027.ogg"},
	{0x0E309B28D8C3D90DULL, "gen00fl_anders6111080.cl"},
	{0x0E315696EBE25726ULL, "211630191.ogg"},
	{0x0E31C1C165DBF026ULL, "art\\vfx\\combat\\vfx_c_mer_burst_c.impa.evt"},
	{0x0E31DE24343C5C31ULL, "vxca_caverns2prc_9xc_0_0.vlm"},
	{0x0E32170028FB4DF3ULL, "6105379_m.fxe"},
	{0x0E3224285BD4F270ULL, "isa211_isabela6202182.cl"},
	{0x0E325144742DDDF1ULL, "vxlt_gallowstemplar_f_35d_0_0.vlm"},
	{0x0E32BDA11773C0A6ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.dura.ani"},
	{0x0E32C2D0E2FAB12EULL, "zz_one000_barber6179405.cl"},
	{0x0E33F0816F2238CAULL, "hlac_lt_0_lowtown_131070_3x.dds"},
	{0x0E340F6D5816724FULL, "6094966_f.fxe"},
	{0x0E3426FCFD312B11ULL, "vxw_encounter2_camp_2c_0_0.vlm"},
	{0x0E3455FE2D4FD42AULL, "6101491_m.fxe"},
	{0x0E34794FF2A79F6DULL, "vrc321ip_spooky5_obj_3.utp"},
	{0x0E34B4770DCA7AC6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_extreme_v1.evt"},
	{0x0E35007780383FA1ULL, "vxlt_gallowstemplar_39k_0_0.vlm"},
	{0x0E351F584BB4BB35ULL, "gen00fl_anders6096455.cl"},
	{0x0E3550DBCA810704ULL, "vxlt_0_lowtown_a2_ep4_0_0.vlm"},
	{0x0E35714776DF35BCULL, "6016902_m.fxe"},
	{0x0E35A191748372D6ULL, "vxht_mansion2b_35c_0_0.vlm"},
	{0x0E35A692284CEC72ULL, "zz_slate_3_baddy.cut"},
	{0x0E35C747E3A44B4DULL, "vxlt_backalley_g_1e3_0_0.vlm"},
	{0x0E35DC99E96BF5BCULL, "6163484_m.fxe"},
	{0x0E3646A253ADAB3AULL, "464368739.ogg"},
	{0x0E3671A2AE20CC4FULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_fema_0d.dds"},
	{0x0E3673D682093AF4ULL, "art\\vfx\\combat\\vfx_m_cre_conj_c.impa.evt"},
	{0x0E36A7991B24AC07ULL, "vxdr_primevaltomb_81x_0_0.vlm"},
	{0x0E377B5850CA6C0EULL, "w_sundermount_78_0_lc.mmh"},
	{0x0E37AC1885DB3DD2ULL, "235794163.ogg"},
	{0x0E37D8C165E08371ULL, "art\\vfx\\combat\\vfx_c_mer_burst_c.impa.gad"},
	{0x0E3969A2A985055BULL, "vxht_mansion2b_2jz_0_0.vlm"},
	{0x0E3991E1DB02A4E4ULL, "vxlt_darktown_5ni_0_0.vlm"},
	{0x0E39A21F70BCF372ULL, "vxht_0_hightown_n_c7b_0_0.vlm"},
	{0x0E3A8351AC9EE3D5ULL, "341358766.ogg"},
	{0x0E3AB64CAD90B4E9ULL, "vxht_keep_73r_0_0.vlm"},
	{0x0E3ABE9FE98D0D78ULL, "6117263_m.fxe"},
	{0x0E3B036839691351ULL, "53954294.ogg"},
	{0x0E3B365A2A144DBCULL, "6050412_m.fxe"},
	{0x0E3B5C9C34FBB3F1ULL, "520490748.ogg"},
	{0x0E3BAD847A87227AULL, "554188506.ogg"},
	{0x0E3BCB770DD0C011ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_extreme_v1.gad"},
	{0x0E3BD0ED1F163AD8ULL, "6166458_m.fxe"},
	{0x0E3BD5A0CD432C1EULL, "6096452_m.fxe"},
	{0x0E3BF9A536344EE0ULL, "vxw_bonepitmining_a2_1xd_0_1.vlm"},
	{0x0E3C4D8B42E5469FULL, "w_bonepit_42_0.rml"},
	{0x0E3C6E653BB2FA37ULL, "vxht_0_hightown_a2_c8u_0_0.vlm"},
	{0x0E3C9A158510C2F4ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgertes.phy"},
	{0x0E3CA4092125AD0FULL, "hlac_lt_0_lowtown_131072_4z.dds"},
	{0x0E3CBFCE1F118946ULL, "vxw_sundermount_2q6_0_0.vlm"},
	{0x0E3D8E1781D71CF5ULL, "logpile_d.dds"},
	{0x0E3DF7FF2173F981ULL, "869504405.ogg"},
	{0x0E3E27AA36D16185ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackleft.ani"},
	{0x0E3EA83E6EAD1C7EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\hu_reaction_impale_en.evt"},
	{0x0E3ED9C165E6A31FULL, "art\\vfx\\combat\\vfx_c_mer_burst_c.impa.ani"},
	{0x0E3EDB0778D606ADULL, "zz_approval_debug6003680.cl"},
	{0x0E3F6CDA83A06F26ULL, "6038618_m.fxe"},
	{0x0E3FA5A1177E2217ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.dura.evt"},
	{0x0E3FB553A8BCB1D1ULL, "6150685_m.fxe"},
	{0x0E3FC1037E30AE8BULL, "zz_rdr_debug6066648.cl"},
	{0x0E3FC6AF309586BEULL, "rdr111dg_martin6015062.cl"},
	{0x0E4039C684D031FEULL, "dre000_varric6208061.cl"},
	{0x0E40895BD1856683ULL, "698310132.ogg"},
	{0x0E40F9F9595C9BF3ULL, "zz_followers_banter6157881.cl"},
	{0x0E41488C647915BEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_paintings_backtoback_da.dds"},
	{0x0E4164438F97480EULL, "hlac_ca_caverns1_24422_0y.dds"},
	{0x0E4186D2F44E1213ULL, "vxlt_0_lowtown_a2_hfh_0_0.vlm"},
	{0x0E418CC43A01ECFFULL, "gen00fl_aveline6139762.cl"},
	{0x0E41B30C2A5F1D73ULL, "6161445_m.fxe"},
	{0x0E42018EA826921CULL, "vxlt_gallowstemplar_6t0_0_0.vlm"},
	{0x0E42877E85297468ULL, "vxlt_gallowstemplar_n_4sq_0_1.vlm"},
	{0x0E42CC770DD6E0BFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_extreme_v1.ani"},
	{0x0E42D9D11522B7E4ULL, "one350_variel6085544.cl"},
	{0x0E4308D6498846E5ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_qu101a_0d.dds"},
	{0x0E4321AE267E0A10ULL, "llac_w_bonepitmining_a2_0_2.dds"},
	{0x0E4342744D829100ULL, "vxlt_gallowstemplar_5it_0_0.vlm"},
	{0x0E4442B2B6636CA6ULL, "6113648_f.fxe"},
	{0x0E44503F0B0BD434ULL, "zz_qun_debug6028218.cl"},
	{0x0E446BD682156A2DULL, "art\\vfx\\combat\\vfx_m_cre_conj_c.impa.ani"},
	{0x0E44F1620195F0D5ULL, "6051875_f.fxe"},
	{0x0E44F3BA54804C27ULL, "dae000_madam_lusine6151389.cl"},
	{0x0E464D798E4BC253ULL, "zz_hawke_hightownnt.cut"},
	{0x0E4677158518DBF1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgertes.msh"},
	{0x0E4685158518F30CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgertes.mao"},
	{0x0E469115851907AFULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgertes.mmh"},
	{0x0E46A6A1178442A0ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.dura.gad"},
	{0x0E46BC2A78EFBADDULL, "6080176_f.fxe"},
	{0x0E4756C294439C00ULL, "706631450.ogg"},
	{0x0E4808A4F3489B56ULL, "6163284_m.fxe"},
	{0x0E4869A2F91BAADAULL, "art\\characters\\creatures\\arh_abm_rev\\c_abomination.phy"},
	{0x0E48B55F80AEB2D4ULL, "6082272_f.fxe"},
	{0x0E48D6AA5CC1F507ULL, "vxlt_0_lowtown_docks_bde_0_0.vlm"},
	{0x0E49043CB9A4CDA6ULL, "hlac_lt_warehouse_n_0_4z.dds"},
	{0x0E49312D453BBA64ULL, "6060737_f.fxe"},
	{0x0E496B78CFAEB919ULL, "hightown_male_noble_gossip6171386.cl"},
	{0x0E49C9702293292CULL, "hlac_ca_caverns1_24427_3y.dds"},
	{0x0E49E7B8390729C9ULL, "mcr370_guardsman6087667.cl"},
	{0x0E4A11F80108DCFAULL, "zz_vault_debug6177615.cl"},
	{0x0E4A742419A91F87ULL, "vxht_0_hightown_n_cb0_0_0.vlm"},
	{0x0E4AAEEDF4E7F39EULL, "hlac_w_encounter2_327679_0x.dds"},
	{0x0E4AE6E64DC96E73ULL, "hla_ht_0_hightown_a2_1_0x.dds"},
	{0x0E4AF644B23A1F09ULL, "rdr110dg_isabela6098928.cl"},
	{0x0E4B31AB1107A1C6ULL, "vxw_sundermount_2z7_0_0.vlm"},
	{0x0E4B3D3F17CC9638ULL, "432694206.ogg"},
	{0x0E4B62F03C3D2C91ULL, "vxca_caverns1_j4b_0_0.vlm"},
	{0x0E4B68DCCCD75EFCULL, "zz_mcr_debug6071698.cl"},
	{0x0E4B9D35F7386BCAULL, "6100104_m.fxe"},
	{0x0E4BEF2327E44C7BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1a.mmh"},
	{0x0E4BFB2327E46018ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1a.mao"},
	{0x0E4C0D2327E47F65ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1a.msh"},
	{0x0E4C68F24E4FEEF3ULL, "6037852_m.fxe"},
	{0x0E4C7E0B4BC9FB04ULL, "vxht_0_hightown_n_b22_0_0.vlm"},
	{0x0E4CB2876DCB9717ULL, "spit_spit.mao"},
	{0x0E4D180C18ECB126ULL, "hlac_ca_caverns2prc_5802_46z.dds"},
	{0x0E4D6DACEB4E3A73ULL, "6153260_m.fxe"},
	{0x0E4DA6F4D84FCDD4ULL, "mag360_alain6044617.cl"},
	{0x0E4E6C7859C9B467ULL, "vxca_caverns1_jgq_0_0.vlm"},
	{0x0E4E9C811CE19827ULL, "hlac_lt_gallowstemplar_2394_0y.dds"},
	{0x0E4F688B8C37BFF8ULL, "hlac_lt_0_lowtown_n_-65535_6y.dds"},
	{0x0E4F9EA4F482D707ULL, "vxht_keep_n_4an_0_0.vlm"},
	{0x0E4FF8AFE6442A93ULL, "vxlt_0_lowtown_a2_hck_0_0.vlm"},
	{0x0E50356307A58E24ULL, "420180243.ogg"},
	{0x0E504B0DCE7A0792ULL, "corpses_n.dds"},
	{0x0E50697AE4F44843ULL, "345473490.ogg"},
	{0x0E50F12ECD337248ULL, "art\\levels\\ztools\\default\\generic\\0wi_angle.mmh"},
	{0x0E50FD2ECD3386EBULL, "art\\levels\\ztools\\default\\generic\\0wi_angle.mao"},
	{0x0E510F2ECD33A572ULL, "art\\levels\\ztools\\default\\generic\\0wi_angle.msh"},
	{0x0E518C34F6118F97ULL, "hlac_lt_0_lowtown_docks_131070_1z.dds"},
	{0x0E51B4513F78A980ULL, "vxlt_gallowstemplar_5ru_0_0.vlm"},
	{0x0E51B6489D8E1253ULL, "follower_barks6153215.cl"},
	{0x0E527E054BF4EC0DULL, "gen00fl_sebastian6174703.cl"},
	{0x0E52A176F573FFAFULL, "6116559_m.fxe"},
	{0x0E52DACCECF9A85FULL, "vxlt_0_lowtown_docks_a3_n_cx0_0_0.vlm"},
	{0x0E52EC12181D612FULL, "166000882.ogg"},
	{0x0E52F7C6CBCB68F5ULL, "vxlt_0_lowtown_n_f5b_0_0.vlm"},
	{0x0E532C6D09D1115AULL, "ht_0_hightown_n_50_0.rml"},
	{0x0E5346666C4E3892ULL, "1066076473.ogg"},
	{0x0E53510B8C72340CULL, "11568203.ogg"},
	{0x0E53614A625AF76FULL, "llac_lt_0_lowtown_n_-2_6.dds"},
	{0x0E5380519272F955ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_short.phy"},
	{0x0E53D420B76F4FACULL, "hlac_lt_gallowscourtyard_n_-65538_1x.dds"},
	{0x0E543DEE473DD650ULL, "6110477_m.fxe"},
	{0x0E55333E3041F2F7ULL, "769325577.ogg"},
	{0x0E553756E2E81DB9ULL, "6076263_f.fxe"},
	{0x0E55854A5B14BE7DULL, "dre000_bodahn6031357.cl"},
	{0x0E559C445C7EE0A3ULL, "crate_03.dds"},
	{0x0E55B9D488581D54ULL, "787879008.ogg"},
	{0x0E55F6B248156010ULL, "vxw_encounter1_cp_0_0.vlm"},
	{0x0E55FF694E6EC710ULL, "vxlt_hangedman_n_rl_0_0.vlm"},
	{0x0E562162DB43DDEDULL, "6101225_m.fxe"},
	{0x0E563BAFD66A9BB5ULL, "967752300.ogg"},
	{0x0E56AD47F538E592ULL, "vxw_bonepit_21a_0_0.vlm"},
	{0x0E56D4C6729D0A7BULL, "art\\vfx\\combat\\vfx_q_ari_heal_act_c.mmh"},
	{0x0E57410674E7BFC3ULL, "658160999.ogg"},
	{0x0E5766C23E58383AULL, "vxca_caverns1_mines_i2y_0_0.vlm"},
	{0x0E5786F05AAD25B4ULL, "dae000cr_grand_cleric6148136.cl"},
	{0x0E579A0663965DE2ULL, "hlac_lt_0_lowtown_docks_-131075_4x.dds"},
	{0x0E57BC8B7BA9FA41ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall.phy"},
	{0x0E57C5304E4FF5E0ULL, "cd_mag291_torpor.bnk"},
	{0x0E584BDAC00CB3C7ULL, "llac_ht_chantry_2534_17.dds"},
	{0x0E58727872BF8D0BULL, "art\\vfx\\combat\\vfx_r_arc_burst_sh_aoe_p.anb"},
	{0x0E58880E0AC0CC6EULL, "ico_crossbow_5.dds"},
	{0x0E58C3D5481115D3ULL, "hlac_lt_0_lowtown_a2_n_c_0_8y.dds"},
	{0x0E58CD4624CDD4E7ULL, "jagged_d.dds"},
	{0x0E590FE0D9323A70ULL, "llac_ht_0_hightown_a2_-65536_1.dds"},
	{0x0E591BBB4033DA87ULL, "163082636.ogg"},
	{0x0E592DDF7AFA778AULL, "mag291cr_torpor.utc"},
	{0x0E592E76D8C45746ULL, "6134590_m.fxe"},
	{0x0E59AE7277EA89EBULL, "6205367_m.fxe"},
	{0x0E59D0092610DCDFULL, "mag111pt_act_of_mercy.ncs"},
	{0x0E59E0986E4199CDULL, "ca_caverns1_24425_lc.mmh"},
	{0x0E59E42541F7B346ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masa.phy"},
	{0x0E5A0B3D43D69210ULL, "llac_lt_0_lowtown_docks_n_131070_1.dds"},
	{0x0E5A4D84B2D18AC9ULL, "vxht_0_hightown_aqz_0_0.vlm"},
	{0x0E5AD7465CAF17F1ULL, "565526286.ogg"},
	{0x0E5AF0F06B65B934ULL, "hlac_lt_0_lowtown_docks_-1_2x.dds"},
	{0x0E5B80C71590170AULL, "106538625.ogg"},
	{0x0E5BCD5383F9AD28ULL, "vxdr_primevaltomb_6vt_0_0.vlm"},
	{0x0E5BCEDC3F431052ULL, "lt_gallowscourtyard_n_50_0_lc.mmh"},
	{0x0E5BFD2204BA1613ULL, "water_water_p2.vsh.fxo.dx11.assembly"},
	{0x0E5C9B1717AE7A32ULL, "vxht_0_hightown_n_7g6_0_0.vlm"},
	{0x0E5CD31CCC7B90CEULL, "vxht_brothel_n_5h6_0_0.vlm"},
	{0x0E5D66C6022ED171ULL, "mag160_feynriel6037376.cl"},
	{0x0E5DC66B1725E852ULL, "300635177.ogg"},
	{0x0E5E684FA10C4FC0ULL, "w_sundermount_88_0.phy"},
	{0x0E5E6A8CD83A4313ULL, "vxlt_0_lowtown_a2_hhj_0_0.vlm"},
	{0x0E5EE1179DAD85A0ULL, "6096007_m.fxe"},
	{0x0E5F4B2914EE94EEULL, "24470962.ogg"},
	{0x0E5FC326D62CE19EULL, "hlac_lt_hangedman_n_-1_15x.dds"},
	{0x0E5FF616947F6FABULL, "fen210_fenris6109403.cl"},
	{0x0E604287D2A78828ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose4.anb"},
	{0x0E608257F94B34FCULL, "vxht_mansion2b_315_0_0.vlm"},
	{0x0E60910C6A1BA776ULL, "vxca_caverns2_8o1_0_0.vlm"},
	{0x0E6109EAC5C8A2A1ULL, "llac_ht_keep_n_4519_0.dds"},
	{0x0E613C8DF19BD235ULL, "matdef_dc_normal.dds"},
	{0x0E613D8C854C5E30ULL, "390386713.ogg"},
	{0x0E61765E6BD809B8ULL, "6140951_m.fxe"},
	{0x0E6196C185ED055BULL, "art\\levels\\races\\elf\\props\\dalish\\prp_dream.msh"},
	{0x0E61A8C185ED23A6ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_dream.mao"},
	{0x0E61B4C185ED3845ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_dream.mmh"},
	{0x0E627C1FD78F32BAULL, "vxht_mansion2b_3bj_0_0.vlm"},
	{0x0E62C0397806EEE4ULL, "lt_warehouse_ll.mmh"},
	{0x0E62CC195C8278A8ULL, "hlac_ht_mansion_n_3373_0z.dds"},
	{0x0E62D5495A6D55C4ULL, "hlac_dr_primevaltomb_661_3x.dds"},
	{0x0E632A1FC169E459ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_injured_pose_loop.gad"},
	{0x0E63B9ECECF2EE23ULL, "gen00fl_merrill6119611.cl"},
	{0x0E63DC1808356C1AULL, "6165545_m.fxe"},
	{0x0E640463FE7A5CD6ULL, "vxca_caverns2_801_0_1.vlm"},
	{0x0E642A28EC723CE8ULL, "zz_followers_banter6159459.cl"},
	{0x0E64688F3A0B7890ULL, "vxw_encounter1_hq_0_0.vlm"},
	{0x0E64874FA110F078ULL, "w_sundermount_88_0.rml"},
	{0x0E64DA919C4A5FC4ULL, "6013358_m.fxe"},
	{0x0E64EA6B2AB69FD8ULL, "zz_qun_debug6028302.cl"},
	{0x0E64F5C672A97FE0ULL, "art\\vfx\\combat\\vfx_q_ari_heal_act_c.anb"},
	{0x0E64F827D0883115ULL, "hlac_w_encounter3_-262148_0z.dds"},
	{0x0E662A46D3447296ULL, "dae200cr_viscount6087374.cl"},
	{0x0E662D91F06391DCULL, "6127659_m.fxe"},
	{0x0E664C4145B4DA7AULL, "6056618_m.fxe"},
	{0x0E673C4BA76234E8ULL, "vxlt_0_lowtown_f5d_0_0.vlm"},
	{0x0E6767EB99C58208ULL, "vxht_0_hightown_a3_c7f_0_0.vlm"},
	{0x0E676A48205CEC48ULL, "w_blightlands_166_0.mmh"},
	{0x0E677648205D00EBULL, "w_blightlands_166_0.mao"},
	{0x0E678848205D1F72ULL, "w_blightlands_166_0.msh"},
	{0x0E6789E929A07857ULL, "285747765.ogg"},
	{0x0E67A17B4A1137BEULL, "6162346_m.fxe"},
	{0x0E67D2176C736101ULL, "583368163.ogg"},
	{0x0E67F73F8F8F04CAULL, "161735566.ogg"},
	{0x0E68204F69E0A26BULL, "6208467_m.fxe"},
	{0x0E683D35BC23DFDCULL, "gen00fl_anders6111304.cl"},
	{0x0E684CAEB2932898ULL, "vxw_sundermount_351_0_0.vlm"},
	{0x0E69277B28288B90ULL, "vrc321st_spooky_scene_1.stg"},
	{0x0E698074B150E226ULL, "358607580.ogg"},
	{0x0E69CB55D278832BULL, "ser270_emeric6113588.cl"},
	{0x0E69E5A60E32A7CAULL, "vxlt_0_lowtown_n_exa_0_0.vlm"},
	{0x0E69F8121CA89037ULL, "6187939_m.fxe"},
	{0x0E6A3B1FC1701FA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_injured_pose_loop.ani"},
	{0x0E6A6306A409F2E8ULL, "597288201.ogg"},
	{0x0E6A64B928DBAE6BULL, "6013322_m.fxe"},
	{0x0E6AC9BE863C9C32ULL, "vxdr_0_deeproads_fiu_0_0.vlm"},
	{0x0E6ADF739AAF3EA2ULL, "w_encounter1_47_0.msh"},
	{0x0E6AE1739AAF4258ULL, "w_encounter1_47_0.mmh"},
	{0x0E6AE2FF66BC8DC7ULL, "888904479.ogg"},
	{0x0E6AED739AAF56FBULL, "w_encounter1_47_0.mao"},
	{0x0E6B2344B4E58B43ULL, "gen00fl_aveline6132899.cl"},
	{0x0E6B6D952A501A9AULL, "vxht_0_hightown_ccm_0_0.vlm"},
	{0x0E6B9E23606EBFE6ULL, "6122368_m.fxe"},
	{0x0E6BEC33AE62180FULL, "vxlt_undercitydungeon_1iu_0_1.vlm"},
	{0x0E6C4FF003C5BB8DULL, "lit_up_05_l3.mao"},
	{0x0E6C54EC9D4AFB12ULL, "zz_one000_merchant6179348.cl"},
	{0x0E6C5EBDA077598EULL, "6005764_m.fxe"},
	{0x0E6CEACE26374C0BULL, "578937536.ogg"},
	{0x0E6D00BBCF41BAF7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffc_sa.dds"},
	{0x0E6D17543E92465AULL, "vxht_mansion_n_6kk_0_0.vlm"},
	{0x0E6D380F95724AA8ULL, "450525049.ogg"},
	{0x0E6D6F248A46F2E9ULL, "lt_0_lowtown_docks_n_25_0.rml"},
	{0x0E6E00F2EC37A529ULL, "zz_neel_test.cut"},
	{0x0E6E0B6EF7D92661ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_strt.ani"},
	{0x0E6E1A641D23358BULL, "539043476.ogg"},
	{0x0E6E5F7B56154852ULL, "hlac_ca_caverns1_24424_6y.dds"},
	{0x0E6EF993AFB5F376ULL, "vxht_0_hightown_be6_0_0.vlm"},
	{0x0E6F027C0AF58B39ULL, "vxlt_gallowstemplar_n_37b_0_0.vlm"},
	{0x0E6F02E95C11BFF6ULL, "vxca_caverns2_8j2_0_0.vlm"},
	{0x0E6F247BEBD5897FULL, "hlac_lt_undercitydungeon_-1_39z.dds"},
	{0x0E6F5247C9E670DDULL, "rdr111dg_martin6015184.cl"},
	{0x0E6FE0FA3F88CE66ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgtp_0n.dds"},
	{0x0E7007C8A3B801FCULL, "vxlt_gallowstemplar_6s5_0_0.vlm"},
	{0x0E70DEB98C8AB30EULL, "vxw_bonepitdestroyed_a3_233_0_0.vlm"},
	{0x0E70F7F8E29614E7ULL, "vxw_encounter2_camp_or_0_0.vlm"},
	{0x0E7124D6FECCB539ULL, "vxlt_gallowstemplar_f_3sn_0_0.vlm"},
	{0x0E71793B54AD31F0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_basic.mao"},
	{0x0E717D3B54AD3803ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_basic.mmh"},
	{0x0E718414DD514E84ULL, "6132926_f.fxe"},
	{0x0E718B3B54AD505DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_basic.msh"},
	{0x0E72252864EAFDA1ULL, "vxlt_0_lowtown_bz0_0_0.vlm"},
	{0x0E7391DE0AD4B40AULL, "hlac_ht_keep_3732_1y.dds"},
	{0x0E74447851DCD28DULL, "6080654_m.fxe"},
	{0x0E7472243D85A372ULL, "vxht_0_hightown_a2_c1w_0_0.vlm"},
	{0x0E75166EF7DF5703ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_strt.gad"},
	{0x0E756E9A2AC5AC46ULL, "533343636.ogg"},
	{0x0E7613583C07503EULL, "6167046_m.fxe"},
	{0x0E769783C108695BULL, "llac_lt_foundry_n_320_20.dds"},
	{0x0E76D60D5F599B77ULL, "w_blightlands_49_0.rml"},
	{0x0E77160689F9CA25ULL, "6101966_m.fxe"},
	{0x0E771C5CCA684FCFULL, "art\\levels\\atmosphere\\sky\\cloudmaterial.mao"},
	{0x0E7742441DD61F03ULL, "art\\characters\\creatures\\desiredemon\\animation\\kiss.ani"},
	{0x0E776140F98AB94CULL, "463900396.ogg"},
	{0x0E78294ADDE6A21AULL, "vxht_0_hightown_a3_asv_0_0.vlm"},
	{0x0E78403599678125ULL, "6168912_m.fxe"},
	{0x0E7857830028C04AULL, "vxlt_0_lowtown_n_eub_0_0.vlm"},
	{0x0E78605F53D10F17ULL, "6134292_m.fxe"},
	{0x0E794833BFC2AF9EULL, "hlac_lt_0_lowtown_n_-65535_1z.dds"},
	{0x0E797CE7FA250673ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardleft.ani"},
	{0x0E79DD9B45ADA00FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fr.gad"},
	{0x0E7B06229D1D3956ULL, "item_variations_col.gda"},
	{0x0E7B893130885EDAULL, "vxht_mansion_n_6nd_0_0.vlm"},
	{0x0E7C5B084EFD0FC9ULL, "t3_arm_klr.tnt"},
	{0x0E7CA9FCCC446560ULL, "6104277_f.fxe"},
	{0x0E7D8A2C844F8184ULL, "vxlt_0_lowtown_a2_eo9_0_0.vlm"},
	{0x0E7DD5B317FF1B7CULL, "vxca_caverns1_mines_j7c_0_0.vlm"},
	{0x0E7E6C8142D0A306ULL, "gen00tr_talk.ncs"},
	{0x0E7E9FB8A078226FULL, "art\\toolset\\tools_lightprobe.mao"},
	{0x0E7EABB8A078360CULL, "art\\toolset\\tools_lightprobe.mmh"},
	{0x0E7EB1B8A07840FEULL, "art\\toolset\\tools_lightprobe.msh"},
	{0x0E7F69D5D48C2D67ULL, "vxw_encounter2_camp_jm_0_0.vlm"},
	{0x0E7FF29B45B22FBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fr.evt"},
	{0x0E80985C7F1C146EULL, "vxca_caverns1_mines_e8w_0_0.vlm"},
	{0x0E8162D047D6C91AULL, "vxw_encounter2_qp_0_1.vlm"},
	{0x0E81651D98A4C571ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_litwindows.msh"},
	{0x0E81731D98A4DC8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_litwindows.mao"},
	{0x0E817F1D98A4F12FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_litwindows.mmh"},
	{0x0E818435E223B82EULL, "gen00fl_isabela_init6076477.cl"},
	{0x0E81E2D63A5E9C95ULL, "che_wispy.mao"},
	{0x0E827E3B54BBAD28ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_basic.phy"},
	{0x0E829EEABFBF42DFULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_lp.gad"},
	{0x0E82A98713AAF4CCULL, "mer110_jansen6032610.cl"},
	{0x0E8367D4B07EE904ULL, "vxdr_primevalentrance_310_0_0.vlm"},
	{0x0E83854FA12B9D55ULL, "w_sundermount_88_0.msh"},
	{0x0E83874FA12BA0CBULL, "w_sundermount_88_0.mmh"},
	{0x0E83934FA12BB568ULL, "w_sundermount_88_0.mao"},
	{0x0E841189C8C14B5AULL, "6127309_m.fxe"},
	{0x0E845A441DE0D2FAULL, "art\\characters\\creatures\\desiredemon\\animation\\kiss.evt"},
	{0x0E85005EDD1FDD26ULL, "676193633.ogg"},
	{0x0E856DF1BB6D68D3ULL, "w_woundedcoast_29_0.anb"},
	{0x0E8601EFD271D3B0ULL, "hlac_lt_0_lowtown_a3_n_c_-65539_0z.dds"},
	{0x0E8617BB672D0C60ULL, "dae000_mother6103123.cl"},
	{0x0E8626FC8D027FADULL, "vxlt_undercityserialkiller_1yc_0_0.vlm"},
	{0x0E8636E34AB9E033ULL, "gen00fl_merrill6147025.cl"},
	{0x0E864C41695CB017ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a_2.mao"},
	{0x0E8674E7FA2F82AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardleft.evt"},
	{0x0E8682B697A7D779ULL, "vxlt_0_lowtown_docks_b61_0_0.vlm"},
	{0x0E86AF7160CF3A8AULL, "6054371_m.fxe"},
	{0x0E874AD369D15D99ULL, "dae100_bran6072449.cl"},
	{0x0E88399E01D6BBF5ULL, "vxlt_warehouse_n_108_0_0.vlm"},
	{0x0E8845533FA0CF7AULL, "gen00fl_anders6125640.cl"},
	{0x0E88BF7B28574210ULL, "143308887.ogg"},
	{0x0E88CAE2B9DD4E75ULL, "hlac_lt_0_lowtown_n_131071_2x.dds"},
	{0x0E89F1EC712ED986ULL, "vxlt_undercityserialkiller_18o_0_0.vlm"},
	{0x0E8A418B4327538FULL, "w_bonepit_42_0.mao"},
	{0x0E8A4D8B4327682CULL, "w_bonepit_42_0.mmh"},
	{0x0E8A538B4327721EULL, "w_bonepit_42_0.msh"},
	{0x0E8A623D6F9CD6EFULL, "vxw_blightlands_ci_0_0.vlm"},
	{0x0E8A7593D619113AULL, "6031987_m.fxe"},
	{0x0E8AA85F97B816DBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbasic_0.pwk"},
	{0x0E8B45E7929B6579ULL, "dae150_chandan6148637.cl"},
	{0x0E8B55EB7D217ECCULL, "227185722.ogg"},
	{0x0E8B59441DE6EFEDULL, "art\\characters\\creatures\\desiredemon\\animation\\kiss.gad"},
	{0x0E8B6B3D5751B532ULL, "hlac_lt_0_lowtown_65534_5y.dds"},
	{0x0E8B96B5B0A073F7ULL, "hlac_w_woundedcoast_-131071_0z.dds"},
	{0x0E8BB5850FA55583ULL, "mag200_feynriel6217339.cl"},
	{0x0E8C7E596CE0C07BULL, "vxlt_undercityserialkiller_1iu_0_0.vlm"},
	{0x0E8CD655C5C8F655ULL, "qun211ar_poison_encounter.are"},
	{0x0E8D53E7FA35691DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardleft.gad"},
	{0x0E8D872C601D63D7ULL, "vxca_cavern1_haunted_fm6_0_0.vlm"},
	{0x0E8DDBB2C68245E7ULL, "vxw_encounter2_camp_el_0_0.vlm"},
	{0x0E8DF45720129001ULL, "vxlt_0_lowtown_docks_n_8pz_0_0.vlm"},
	{0x0E8DFA9B45BE7AD5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fr.ani"},
	{0x0E8E88B7278E75A3ULL, "vxw_woundedcoast_25_0_1.vlm"},
	{0x0E8ED4479D7A4B06ULL, "one300_murder_followers6089903.cl"},
	{0x0E8F1604937CBFF3ULL, "dae000_arishok6080476.cl"},
	{0x0E8F3F0DC9D52A50ULL, "142464900.ogg"},
	{0x0E8F74D949B07287ULL, "vxw_bonepitdestroyed_a3_1gm_0_0.vlm"},
	{0x0E8FA0FDC8B672A8ULL, "6051147_m.fxe"},
	{0x0E901E1044081E99ULL, "follower_banter6166966.cl"},
	{0x0E90274D153A79C9ULL, "6120822_m.fxe"},
	{0x0E90484D47F7C1CDULL, "dae000ip_lean_back_d.utp"},
	{0x0E9064DF5509EED3ULL, "blood_clu.mao"},
	{0x0E912B5750751E5CULL, "lowtown_male_lowlife1_sta6165030.cl"},
	{0x0E91684FA137A930ULL, "w_sundermount_88_0.anb"},
	{0x0E9188B07AFE1653ULL, "17732377.ogg"},
	{0x0E91AD010AF0A491ULL, "6045966_f.fxe"},
	{0x0E91D0DDBF2D01F1ULL, "696701112.ogg"},
	{0x0E92E1258EDE7361ULL, "6102570_m.fxe"},
	{0x0E92F36BDF4788E8ULL, "hlac_lt_0_lowtown_docks_65532_0y.dds"},
	{0x0E93218EC80BD5FDULL, "vxlt_0_lowtown_a2_n_c_gv8_0_0.vlm"},
	{0x0E9370F1BB79AB4EULL, "w_woundedcoast_29_0.msh"},
	{0x0E937EF1BB79C29FULL, "w_woundedcoast_29_0.mao"},
	{0x0E9386E90B03039DULL, "vxw_sundermount_30o_0_0.vlm"},
	{0x0E938AF1BB79D73CULL, "w_woundedcoast_29_0.mmh"},
	{0x0E93B9A686108ED7ULL, "lgt311_shw_mer_mettin6156548.cl"},
	{0x0E93BA9372606F39ULL, "hlac_lt_0_lowtown_docks_n_-3_0z.dds"},
	{0x0E93CE735FD3F0C1ULL, "6048419_f.fxe"},
	{0x0E9459A0A3AE15A3ULL, "dae011cr_swindler.utc"},
	{0x0E947C7916F44C46ULL, "gen00fl_anders6125121.cl"},
	{0x0E949D2BB3A3FDE3ULL, "6015971_m.fxe"},
	{0x0E94AE4F3A1FF78AULL, "357835676.ogg"},
	{0x0E950033C869DCD2ULL, "vxlt_gallowstemplar_3fu_0_0.vlm"},
	{0x0E95590BA7995B7FULL, "6097151_m.fxe"},
	{0x0E955BEFEDBC6A9DULL, "220712573.ogg"},
	{0x0E959A5DE74AF6B3ULL, "6170708_m.fxe"},
	{0x0E959B3F3920EE44ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_lgta_0n.dds"},
	{0x0E95A9F1CDBBFF87ULL, "art\\vfx\\combat\\vfx_sg_hurl_act_c.anb"},
	{0x0E964DAA5670ED7BULL, "650403436.ogg"},
	{0x0E967BEABFCFB025ULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_lp.ani"},
	{0x0E9690F5813AD66AULL, "llac_ht_mansion_n_3370_1.dds"},
	{0x0E96F755C5D182BBULL, "qun211ar_poison_encounter.ncs"},
	{0x0E9749DF6FC4A5B0ULL, "6076122_m.fxe"},
	{0x0E974B4C4EB07CE5ULL, "1050182349.ogg"},
	{0x0E9756E3C463C7A9ULL, "vxht_mansion2_31d_0_0.vlm"},
	{0x0E97A19C8F4D1EA1ULL, "vxht_0_hightown_a2_a2s_0_0.vlm"},
	{0x0E980109B6B1C09FULL, "vxdr_0_deeproads_cxf_0_0.vlm"},
	{0x0E989169D9FE2836ULL, "186338304.ogg"},
	{0x0E98A27075E15EA4ULL, "hlac_lt_gallowstemplar_2383_1y.dds"},
	{0x0E98A769E876D780ULL, "hlac_lt_0_lowtown_a2_131072_6y.dds"},
	{0x0E98D41A6192EF6FULL, "vxw_blightlands_dh_0_0.vlm"},
	{0x0E98F93CD6F17981ULL, "100564434.ogg"},
	{0x0E99133A83EB8DD9ULL, "6016947_m.fxe"},
	{0x0E9995C52F6DDA2FULL, "art\\levels\\races\\elf\\props\\_textures\\prp_evilchya_sa.dds"},
	{0x0E9A184D2B4F6226ULL, "6053161_m.fxe"},
	{0x0E9A935192AF1C70ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_short.msh"},
	{0x0E9AA15192AF338AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_short.mmh"},
	{0x0E9AA55192AF3AC1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_short.mao"},
	{0x0E9AF0365ED6D8FBULL, "vxlt_undercityserialkiller_1nt_0_0.vlm"},
	{0x0E9AF908724DC975ULL, "vxlt_0_lowtown_g9l_0_0.vlm"},
	{0x0E9BC4A3B8DF5787ULL, "vxdr_primevaltomb_8nq_0_0.vlm"},
	{0x0E9C47764B284F19ULL, "6108943_m.fxe"},
	{0x0E9C6107889589D5ULL, "vxht_0_hightown_a3_c3r_0_1.vlm"},
	{0x0E9C768C54F31FFFULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgtb.phy"},
	{0x0E9D1930BF825EF6ULL, "6024568_f.fxe"},
	{0x0E9D7FF84173F3AAULL, "vxw_sundermount_2k8_0_0.vlm"},
	{0x0E9D8A7918CDD0F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_action.ani"},
	{0x0E9D94FB60268235ULL, "858244772.ogg"},
	{0x0E9DD755C5D76B62ULL, "qun211ar_poison_encounter.lst"},
	{0x0E9DE6B63BA68B07ULL, "vxw_bonepitdestroyed_a3_1hl_0_0.vlm"},
	{0x0E9E2FA29D7598BFULL, "6139817_m.fxe"},
	{0x0E9E6EE3F4E0D694ULL, "94625899.ogg"},
	{0x0E9EA35826A3937EULL, "llac_lt_0_lowtown_n_-65535_5.dds"},
	{0x0E9F169931483CEAULL, "746608687.ogg"},
	{0x0E9F1DB552E85975ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gear_n.dds"},
	{0x0E9F273806D399ADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galofliplo.mmh"},
	{0x0E9F2B3806D3A0FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galofliplo.mao"},
	{0x0E9F393806D3B833ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galofliplo.msh"},
	{0x0E9FBA2021E3F8CEULL, "209383272.ogg"},
	{0x0EA02C70A7A1EB4DULL, "417638619.ogg"},
	{0x0EA06E248A724A09ULL, "lt_0_lowtown_docks_n_25_0.anb"},
	{0x0EA09B4B4329469BULL, "6005400_m.fxe"},
	{0x0EA0CE5CABA0CEC3ULL, "6128958_m.fxe"},
	{0x0EA0E744E5EBA484ULL, "6089573_m.fxe"},
	{0x0EA168C84A71E8BCULL, "vxlt_hangedman_tc_0_0.vlm"},
	{0x0EA183600E54645BULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masb_0s.dds"},
	{0x0EA18D225F7471AEULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v2.mmh"},
	{0x0EA1C194FFEE82A3ULL, "6042979_m.fxe"},
	{0x0EA270869BD8C123ULL, "hlac_lt_0_lowtown_n_-65537_4z.dds"},
	{0x0EA2A910DA7F9DA9ULL, "157399160.ogg"},
	{0x0EA2DBDAC065741EULL, "vfx_distortionmask_p0.psh.fxo"},
	{0x0EA2F88122A12E53ULL, "hlac_dr_slavers_3216_9z.dds"},
	{0x0EA30AB670EEB0ADULL, "vxlt_undercityserialkiller_1sa_0_0.vlm"},
	{0x0EA3766A8FAD23DFULL, "738537602.ogg"},
	{0x0EA37B6119973993ULL, "zz_dae_debug6029849.cl"},
	{0x0EA380B41BB02DC6ULL, "6103062_m.fxe"},
	{0x0EA3877918D23756ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_action.gad"},
	{0x0EA3D5AC8DFB8CB7ULL, "plt_vrc300pt_n_rival_max.nss"},
	{0x0EA3E1F3E71DB0D6ULL, "118740193.ogg"},
	{0x0EA441D52572A615ULL, "m_dae321ar_chantry.bnk"},
	{0x0EA462A2F969FD03ULL, "art\\characters\\creatures\\arh_abm_rev\\c_abomination.msh"},
	{0x0EA470A2F96A155DULL, "art\\characters\\creatures\\arh_abm_rev\\c_abomination.mmh"},
	{0x0EA49C36904A70E8ULL, "hightown_market_guard_sta6174611.cl"},
	{0x0EA5B3641A1028A5ULL, "zz_followers_banter6158103.cl"},
	{0x0EA5C8C0B659E029ULL, "vxht_mansion2_34k_0_0.vlm"},
	{0x0EA60D6F0B4ED177ULL, "dae000_bodhan6146809.cl"},
	{0x0EA66FB5FA102227ULL, "392129461.ogg"},
	{0x0EA672E6A8A7D91FULL, "vxdr_0_deeproads_cua_0_0.vlm"},
	{0x0EA675D08BA3B6F6ULL, "6123389_m.fxe"},
	{0x0EA69B9588F49B1EULL, "hlac_lt_darktown_-1_4x.dds"},
	{0x0EA6C74A59139171ULL, "qun221_viscount6095594.cl"},
	{0x0EA6E0E8D9B614C9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gibbet_dp.dds"},
	{0x0EA70D262EF7A0F0ULL, "bet131ip_chest_gift.utp"},
	{0x0EA7A3574F18D9F8ULL, "art\\characters\\creatures\\textures\\cn_gut_0n.dds"},
	{0x0EA800F93B658D99ULL, "vxht_mansion2_2eu_0_0.vlm"},
	{0x0EA80E15C44699D4ULL, "vxht_mansion_n_288_0_0.vlm"},
	{0x0EA88A3844B9D85FULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.mmh"},
	{0x0EA90CF6E518FDC5ULL, "ht_0_hightown_a2_26_0_lc.mmh"},
	{0x0EA92572E4430E46ULL, "hlac_w_bonepitdestroyed_a3_393220_0y.dds"},
	{0x0EA92FAC062E173BULL, "llac_lt_0_lowtown_docks_a3_0_4.dds"},
	{0x0EA953BD885E09C4ULL, "dae000_hate_enemies6167576.cl"},
	{0x0EA98DC1862AABF2ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_dream.phy"},
	{0x0EA9C9080EF6E7B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_dmg_bonus_stagger.evt"},
	{0x0EAA3680AAD57007ULL, "vxdr_primevaltomb_8ip_0_0.vlm"},
	{0x0EAAA27918D883B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_action.evt"},
	{0x0EAB24EDACEEC43DULL, "792711765.ogg"},
	{0x0EAB4570275B05AEULL, "6133146_m.fxe"},
	{0x0EABF1D5336A0C2AULL, "vxw_sundermount_2l9_0_0.vlm"},
	{0x0EAC19B37B09FE55ULL, "llac_dr_primevaltomb_9621_7.dds"},
	{0x0EAC8F6FF09B12D9ULL, "gen00fl_anders6113002.cl"},
	{0x0EACC71118159221ULL, "art\\vfx\\combat\\vfx_gib_hm_meda_rtl.anb"},
	{0x0EACF6582A5FE36CULL, "vxlt_foundry_n_zj_0_0.vlm"},
	{0x0EACFA0E39E50B81ULL, "gen00fl_fenris6126624.cl"},
	{0x0EAD44B0207AB78AULL, "6179917_m.fxe"},
	{0x0EAD633CCDF63621ULL, "fxe_caustics_dif.dds"},
	{0x0EAD778C55019404ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgtb.mmh"},
	{0x0EAD7B8C55019B57ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgtb.mao"},
	{0x0EAD8D8C5501B9A6ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgtb.msh"},
	{0x0EADF3B4FA7CA087ULL, "vxlt_0_lowtown_docks_b9b_0_0.vlm"},
	{0x0EAE33E4551865D2ULL, "6133821_m.fxe"},
	{0x0EAE612FBFB5386BULL, "vxht_keep_n_36t_0_0.vlm"},
	{0x0EAEA271749E2D22ULL, "vxlt_blackemporium_8m_0_0.vlm"},
	{0x0EAEC453BB9B0E1BULL, "vxlt_warehouse_n_115_0_0.vlm"},
	{0x0EAED310949FD0C7ULL, "hlac_dr_dungeon_3216_1x.dds"},
	{0x0EAFAC225F80E3ADULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v2.anb"},
	{0x0EAFB4BCE6EA6A7FULL, "lt_0_lowtown_a2_n_25_0_lc.mmh"},
	{0x0EB0000F7DD2528BULL, "vxlt_gallowstemplar_2xs_0_0.vlm"},
	{0x0EB00C352F8AE41EULL, "6041139_m.fxe"},
	{0x0EB017BA23E8FAAEULL, "dae300_cullen6140646.cl"},
	{0x0EB0203806E2016AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galofliplo.phy"},
	{0x0EB042CAA96505E7ULL, "gen_im_npc_arm_hum_glv_krkb.uti"},
	{0x0EB0777C4F0CBAB8ULL, "6149450_f.fxe"},
	{0x0EB0837ED810C8D7ULL, "93258327.ogg"},
	{0x0EB0C0080EFCF741ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_dmg_bonus_stagger.gad"},
	{0x0EB0E1F044CE82D7ULL, "aip_spidercorrupted_base.gda"},
	{0x0EB17340C4E3C1CCULL, "hlac_w_encounter3_-131076_0z.dds"},
	{0x0EB18AF1BB92D48FULL, "w_woundedcoast_29_0.rml"},
	{0x0EB1CB83F4EC5723ULL, "zz_fol_debug6057010.cl"},
	{0x0EB1D69AFD162A78ULL, "6117432_m.fxe"},
	{0x0EB1DDA77FB7A1A9ULL, "vxlt_warehouse_qu_0_0.vlm"},
	{0x0EB1EDF376F3AD3BULL, "403058786.ogg"},
	{0x0EB248FB1E1ABC0BULL, "vxca_caverns1_mines_edb_0_0.vlm"},
	{0x0EB3ABB635EC7D4BULL, "22027622.ogg"},
	{0x0EB42753CFFA32E6ULL, "259643081.ogg"},
	{0x0EB430BD493BABE3ULL, "27529987.ogg"},
	{0x0EB45407CDC323EDULL, "dae000ip_crowd_mf4_talk5.utp"},
	{0x0EB522ED5B17740BULL, "vxdr_primevalentrance_230_0_0.vlm"},
	{0x0EB54F045BF8B3F3ULL, "188562319.ogg"},
	{0x0EB5B7D4457F206FULL, "vxw_blightlands_rj_0_0.vlm"},
	{0x0EB66D15AFD1A91EULL, "6122764_m.fxe"},
	{0x0EB672D62D5BA619ULL, "vxht_mansion2_2nt_0_0.vlm"},
	{0x0EB67B8FE87F0CB7ULL, "hlac_lt_undercityserialkiller_-1_9z.dds"},
	{0x0EB6822E52D86F72ULL, "vxca_caverns2_4zz_0_0.vlm"},
	{0x0EB6BB4715E9F445ULL, "6028843_m.fxe"},
	{0x0EB6CF05F066D376ULL, "6135180_m.fxe"},
	{0x0EB6FDA4057321FAULL, "vxlt_0_lowtown_a2_n_h2j_0_0.vlm"},
	{0x0EB76AEF35F5C3B8ULL, "wm_ico_keep_roll.dds"},
	{0x0EB7B1D97083B2B2ULL, "6118192_m.fxe"},
	{0x0EB7C1080F0317EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_dmg_bonus_stagger.ani"},
	{0x0EB819CB17D6116EULL, "273448024.ogg"},
	{0x0EB8652F9B52BA63ULL, "vxlt_gallowstemplar_4qf_0_0.vlm"},
	{0x0EB86C9050771CD4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v1.ani"},
	{0x0EB96541B25B6098ULL, "hlac_ht_0_hightown_-65534_0y.dds"},
	{0x0EB9861A10AFB5B1ULL, "mag121_viveka6083365.cl"},
	{0x0EB9C58B7BFCD3B6ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall.mmh"},
	{0x0EBAACD51C3E1B63ULL, "6043868_m.fxe"},
	{0x0EBB5AADC9EC89A6ULL, "6113059_f.fxe"},
	{0x0EBB68351C55FBECULL, "vxlt_foundry_n_uk_0_0.vlm"},
	{0x0EBB8C6374B15BA6ULL, "lgt210_qun_blk_qunari6142328.cl"},
	{0x0EBBB18FB93C66E3ULL, "6107276_m.fxe"},
	{0x0EBBD032DDC432A6ULL, "6107440_m.fxe"},
	{0x0EBC2A50076EEB60ULL, "vxlt_warehouse_n_oc_0_0.vlm"},
	{0x0EBCF6BCD4E9DEBFULL, "809287636.ogg"},
	{0x0EBD13A8B99B5C1DULL, "one300_zevran6089110.cl"},
	{0x0EBD3630AD91269BULL, "vxlt_warehouse_n_164_0_0.vlm"},
	{0x0EBD78E10ADE8530ULL, "vxht_0_hightown_a2_b7f_0_0.vlm"},
	{0x0EBDCAFEC9D7D584ULL, "6087773_m.fxe"},
	{0x0EBE02189D3B4AAEULL, "vxw_sundermount_2w8_0_0.vlm"},
	{0x0EBEA6F1CDDEFF88ULL, "art\\vfx\\combat\\vfx_sg_hurl_act_c.mmh"},
	{0x0EBEC03EC86FAF6CULL, "vxlt_gallowstemplar_f_5qc_0_0.vlm"},
	{0x0EBF4E5B1AA03B1AULL, "6028163_m.fxe"},
	{0x0EBF6990507D3636ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v1.gad"},
	{0x0EC035A4C0891F4EULL, "vxht_0_hightown_n_c0k_0_0.vlm"},
	{0x0EC051D3EB1D97AAULL, "6077670_f.fxe"},
	{0x0EC069FB049C2953ULL, "hlac_ht_0_hightown_-65536_4x.dds"},
	{0x0EC0D047FACE033FULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.cess.gad"},
	{0x0EC178EAE705F01BULL, "zz_and_debug6134047.cl"},
	{0x0EC1DF3EDEED5315ULL, "dae300_bran6143453.cl"},
	{0x0EC2C3E5E7CAD2F1ULL, "vxlt_gallowstemplar_f_3gr_0_0.vlm"},
	{0x0EC2DBB395D028D7ULL, "6103980_m.fxe"},
	{0x0EC2EFA4B8C9B92FULL, "6032828_m.fxe"},
	{0x0EC32178DB627C98ULL, "6172824_m.fxe"},
	{0x0EC33C9A8FF50A2FULL, "523705500.ogg"},
	{0x0EC346E6AD8B445DULL, "359531418.ogg"},
	{0x0EC3CE78078D0A33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_behind.ani"},
	{0x0EC3DD97304217B7ULL, "300919988.ogg"},
	{0x0EC3F1400A74BD90ULL, "vxca_caverns1_mines_fmm_0_0.vlm"},
	{0x0EC490D3FBFA7A80ULL, "llac_ht_keep_3732_2.dds"},
	{0x0EC4B89CC027E22FULL, "dae000_loyal_mage.bnk"},
	{0x0EC4E035DB691348ULL, "vxht_chantry_n_2ta_0_1.vlm"},
	{0x0EC4FC159557A975ULL, "hlac_lt_undercitydungeon_-1_32y.dds"},
	{0x0EC546F1CEEDA053ULL, "vxlt_gallowstemplar_3rt_0_0.vlm"},
	{0x0EC584905081CF95ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v1.evt"},
	{0x0EC623B66279CB32ULL, "6117092_m.fxe"},
	{0x0EC641705A1B7D60ULL, "llac_lt_gallowstemplar_2391_9.dds"},
	{0x0EC66589CF31E49DULL, "ave121_brennan6105817.cl"},
	{0x0EC6C51474441751ULL, "vxlt_0_lowtown_a2_n_c_eht_0_0.vlm"},
	{0x0EC7012A96354403ULL, "hlac_w_sundermount_-196610_0z.dds"},
	{0x0EC702821B96EE2EULL, "w_bonepit_51_0.msh"},
	{0x0EC70AFFB1D681E1ULL, "llac_ca_caverns1_24424_13.dds"},
	{0x0EC710821B97067FULL, "w_bonepit_51_0.mao"},
	{0x0EC71C821B971A1CULL, "w_bonepit_51_0.mmh"},
	{0x0EC75BA76AFF4FD2ULL, "6064390_m.fxe"},
	{0x0EC76E139D59A32CULL, "6133849_m.fxe"},
	{0x0EC7C547FAD40F2CULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.cess.evt"},
	{0x0EC7DA9CC02A5197ULL, "dae000_loyal_mage.cnv"},
	{0x0EC827349CFE8854ULL, "hlac_lt_0_lowtown_docks_131071_7x.dds"},
	{0x0EC8BDB45D190B71ULL, "hlac_lt_0_lowtown_a2_1_9x.dds"},
	{0x0ECB04118FA3C904ULL, "and321tr_distract.utt"},
	{0x0ECBEABDFCD49DB0ULL, "vxht_0_hightown_a2_b8g_0_0.vlm"},
	{0x0ECC44A59A8F71D5ULL, "vxlt_backalley_n_1fk_0_0.vlm"},
	{0x0ECC6DA7DDC25601ULL, "vxlt_gallowstemplar_5iy_0_0.vlm"},
	{0x0ECC73F58F31632EULL, "vxw_sundermount_2x9_0_0.vlm"},
	{0x0ECD242B1FB19FC6ULL, "136513656.ogg"},
	{0x0ECD477E21538F75ULL, "dae321ar_player_home_nt.are"},
	{0x0ECD5C4D737A500DULL, "llac_lt_blackemporium_1055_16.dds"},
	{0x0ECD88236DC2189AULL, "6054069_m.fxe"},
	{0x0ECD8BE9636F2207ULL, "215779781.ogg"},
	{0x0ECDA2305D7CAE8CULL, "6162528_m.fxe"},
	{0x0ECDA94496975E56ULL, "vxlt_0_lowtown_a2_n_eox_0_0.vlm"},
	{0x0ECDFCFB81FD9F07ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.impa.gad"},
	{0x0ECE25A17968CB80ULL, "ch2_dy.psh.fxo"},
	{0x0ECEB74A0DCCF636ULL, "w_encounter2_camp_19_0.msh"},
	{0x0ECEC14A0DCD0734ULL, "w_encounter2_camp_19_0.mmh"},
	{0x0ECEC54A0DCD0E47ULL, "w_encounter2_camp_19_0.mao"},
	{0x0ECEC7A790423ED2ULL, "vxlt_gallowstemplar_3rq_0_0.vlm"},
	{0x0ECFB7CF7A67BA42ULL, "vxlt_hangedman_1th_0_0.vlm"},
	{0x0ED00963B4E1ED07ULL, "gen_im_cns_tom_attribute_01.uti"},
	{0x0ED017C0E1CD10EAULL, "vxca_caverns1_iog_0_0.vlm"},
	{0x0ED037B50A9B9CD7ULL, "vxlt_0_lowtown_n_by8_0_0.vlm"},
	{0x0ED067D52598219FULL, "m_dae321ar_chantry.wbd"},
	{0x0ED08B3844DB2B9CULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.anb"},
	{0x0ED08EF07A209E88ULL, "w_encounter3_250_0.mmh"},
	{0x0ED099FD3DBC148BULL, "vxlt_hangedman_2hj_0_0.vlm"},
	{0x0ED09AF07A20B32BULL, "w_encounter3_250_0.mao"},
	{0x0ED0ACF07A20D1B2ULL, "w_encounter3_250_0.msh"},
	{0x0ED1C469A41CDCCDULL, "llac_lt_0_lowtown_a3_n_c_1_10.dds"},
	{0x0ED1C67807993A6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_behind.evt"},
	{0x0ED2107FE66DDCB0ULL, "376412146.ogg"},
	{0x0ED22C7CB32BA036ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_cenotaphfire_day.mmh"},
	{0x0ED235ECE66D8551ULL, "496269977.ogg"},
	{0x0ED2503E5E1F9ABDULL, "546050065.ogg"},
	{0x0ED25356B4274F12ULL, "qun110_petrice6042298.cl"},
	{0x0ED27694F2B3C442ULL, "vxht_mansion_n_5mj_0_0.vlm"},
	{0x0ED29515FA75F2FAULL, "684513353.ogg"},
	{0x0ED328DA315B4DF0ULL, "vxca_caverns2prc_aie_0_0.vlm"},
	{0x0ED33E2D488875AFULL, "956066678.ogg"},
	{0x0ED3BD84A9F971C2ULL, "ss_templars_f_1.bnk"},
	{0x0ED419489155ED19ULL, "6152162_m.fxe"},
	{0x0ED4218E667A74E0ULL, "w_sundermount_152_0.phy"},
	{0x0ED429AC86C17F47ULL, "vxht_0_hightown_a3_c84_0_0.vlm"},
	{0x0ED43D92D884FFFAULL, "gen00fl_aveline6053326.cl"},
	{0x0ED4AC8EEFB0780EULL, "pro000st_colonel_ewald_1.stg"},
	{0x0ED4B7AA26AF3AFBULL, "vxlt_undercityserialkiller_1rh_0_0.vlm"},
	{0x0ED4BB72FA13CB66ULL, "6106511_m.fxe"},
	{0x0ED4F1FB8203AB14ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.impa.evt"},
	{0x0ED51F7B47AC496FULL, "vxw_sundermount_36r_0_1.vlm"},
	{0x0ED5693C3DA6292BULL, "819379552.ogg"},
	{0x0ED58C1780B7B987ULL, "vxdr_primevaltomb_8zm_0_0.vlm"},
	{0x0ED5AD47FAE02385ULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.cess.ani"},
	{0x0ED6672BCBE2FFB0ULL, "vxht_0_hightown_n_ap7_0_0.vlm"},
	{0x0ED6B17D5DB9741FULL, "sib111_mother6121433.cl"},
	{0x0ED735B501AAE5E9ULL, "6121325_f.fxe"},
	{0x0ED73B4606BDDB6DULL, "llac_lt_0_lowtown_a2_n_c_-3_5.dds"},
	{0x0ED7A9800D7E175FULL, "llac_ca_caverns1_mines_24426_14.dds"},
	{0x0ED7ADFC2ED405BFULL, "vxlt_foundry_n_ze_0_0.vlm"},
	{0x0ED84E210F1812DBULL, "vxht_mansion2b_25w_0_0.vlm"},
	{0x0ED8A578079F20DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_behind.gad"},
	{0x0ED8B0D26695CEEEULL, "vxlt_hangedman_20s_0_0.vlm"},
	{0x0ED8E608CDD4C872ULL, "779426117.ogg"},
	{0x0ED8F948EE67CE9EULL, "llac_dr_primevaltomb_663_1.dds"},
	{0x0ED90E09EB5B1C60ULL, "vxlt_warehouse_n_ya_0_0.vlm"},
	{0x0ED924A44FA61717ULL, "follower_banter6161024.cl"},
	{0x0ED93898010CAF54ULL, "6119333_m.fxe"},
	{0x0ED9BE4C7B8B4AB2ULL, "hlac_ht_bartrand_n_3371_2y.dds"},
	{0x0ED9EC31CB066A12ULL, "qun131_party_comment6117263.cl"},
	{0x0EDA67ED312D94F9ULL, "6171791_f.fxe"},
	{0x0EDABFB1619BFC8AULL, "docks_male_elven_worker_sta6175846.cl"},
	{0x0EDACFC552A5930AULL, "vxw_encounter1_do_0_0.vlm"},
	{0x0EDAE6AC73934CA2ULL, "6108954_f.fxe"},
	{0x0EDAFA847C890389ULL, "gen00fl_isabela_init6076616.cl"},
	{0x0EDBD8E7BD31BD84ULL, "6022567_m.fxe"},
	{0x0EDBDA2F220D1AC9ULL, "808085176.ogg"},
	{0x0EDBE8C9B7B1EECFULL, "vxlt_0_lowtown_a3_n_c_csh_0_0.vlm"},
	{0x0EDC60FF46124E13ULL, "190449068.ogg"},
	{0x0EDCD71A8C8559ACULL, "hlac_w_woundedcoast_0_3z.dds"},
	{0x0EDD0E1A1E725F1DULL, "vxlt_0_lowtown_docks_a3_d20_0_0.vlm"},
	{0x0EDD664EB4A03EA1ULL, "follower_banter6156200.cl"},
	{0x0EDDBD9B827C0DFCULL, "art\\vfx\\materials\\vfx_bub_burst.mao"},
	{0x0EDE1AA8859D92FAULL, "vxca_caverns1_mines_dpb_0_0.vlm"},
	{0x0EDE468DB47A44AEULL, "801688610.ogg"},
	{0x0EDE487E21620382ULL, "dae321ar_player_home_nt.lst"},
	{0x0EDE99DE98699538ULL, "939745582.ogg"},
	{0x0EDECFB08CC2B523ULL, "vxlt_gallowscourtyard_n_4uw_0_0.vlm"},
	{0x0EDEFFCC044FC88EULL, "hlac_w_encounter3_-327680_0x.dds"},
	{0x0EDF0BDA2FB22D0BULL, "vxlt_hangedman_2em_0_0.vlm"},
	{0x0EDF7E05E3548D2CULL, "and111_lirene6084518.cl"},
	{0x0EDF9695C26BA22CULL, "vxht_bartrand_n_64s_0_0.vlm"},
	{0x0EDFD2F4B3B2EA0BULL, "pro000_cut2d_real_story6001238.cl"},
	{0x0EDFFAA7CC719C60ULL, "755812824.ogg"},
	{0x0EE08A8BBD3FDF48ULL, "6103060_m.fxe"},
	{0x0EE0D3711D07466BULL, "vxlt_warehouse_n_13o_0_0.vlm"},
	{0x0EE128969EE48433ULL, "940652375.ogg"},
	{0x0EE15C2FF2E312A2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw3sm.msh"},
	{0x0EE15E2FF2E31658ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw3sm.mmh"},
	{0x0EE15E61073D03D7ULL, "vxlt_0_lowtown_h8r_0_0.vlm"},
	{0x0EE16A2FF2E32AFBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw3sm.mao"},
	{0x0EE19AB723516670ULL, "vxca_caverns2prc_adb_0_0.vlm"},
	{0x0EE1D7C528BAB8F2ULL, "vxca_caverns2_4ke_0_0.vlm"},
	{0x0EE1F54809E7B145ULL, "vxlt_0_lowtown_a2_n_h2a_0_0.vlm"},
	{0x0EE26ED55D288D8FULL, "vxw_woundedcoast_fd_0_0.vlm"},
	{0x0EE26F3C0F8FA79DULL, "pro_mother_lookatvillage.cut"},
	{0x0EE2A1C368259161ULL, "vxw_bonepitdestroyed_a3_1wj_0_1.vlm"},
	{0x0EE2CECB78D9AAD6ULL, "gen00fl_isabela6101455.cl"},
	{0x0EE2E9FB820FDB4DULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.impa.ani"},
	{0x0EE385C5D52E3354ULL, "8930040.ogg"},
	{0x0EE406995831CAA8ULL, "hlac_lt_0_lowtown_a2_n_c_65535_8z.dds"},
	{0x0EE409E5F00095F2ULL, "6101235_m.fxe"},
	{0x0EE4632A6EE4FEB4ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_wkf.ani"},
	{0x0EE464E53AA199A8ULL, "814672642.ogg"},
	{0x0EE4708F188F3D97ULL, "hlac_lt_darktown_-1_74x.dds"},
	{0x0EE4A013BFD8D3BFULL, "vxht_chantry_147_0_0.vlm"},
	{0x0EE5226BE9C683E7ULL, "vxht_0_hightown_n_at8_0_0.vlm"},
	{0x0EE52E9ACFD745FDULL, "6040666_f.fxe"},
	{0x0EE5B822AC312859ULL, "6149070_m.fxe"},
	{0x0EE61C821BB1CB6FULL, "w_bonepit_51_0.rml"},
	{0x0EE6DBC54E1B5EC4ULL, "art\\levels\\races\\vfx_models\\prp_camplogmed.msh"},
	{0x0EE6E4CD0ECF707CULL, "6129355_f.fxe"},
	{0x0EE6EDC54E1B7D2DULL, "art\\levels\\races\\vfx_models\\prp_camplogmed.mao"},
	{0x0EE6F1C54E1B8466ULL, "art\\levels\\races\\vfx_models\\prp_camplogmed.mmh"},
	{0x0EE7455830BFF7FDULL, "900741820.ogg"},
	{0x0EE74C8E030A576BULL, "vxht_0_hightown_a2_9bi_0_0.vlm"},
	{0x0EE7D424744E376FULL, "llac_lt_0_lowtown_a3_n_c_-65535_3.dds"},
	{0x0EE7FC7C8D8D1B88ULL, "hlac_ht_chantry_n_2534_19x.dds"},
	{0x0EE820D03D5305C6ULL, "481458918.ogg"},
	{0x0EE837786A2C45ECULL, "807079634.ogg"},
	{0x0EE865B2A72264C1ULL, "6209919_f.fxe"},
	{0x0EE8C371C542521CULL, "6179829_m.fxe"},
	{0x0EE8D96BF3236BB6ULL, "vxlt_0_lowtown_docks_bdy_0_1.vlm"},
	{0x0EE9285F7E7BA2D5ULL, "vxlt_backalley_n_1li_0_0.vlm"},
	{0x0EE93A3C3A8F076DULL, "llac_lt_0_lowtown_n_-131072_4.dds"},
	{0x0EE941A2449BAB8AULL, "vxw_encounter1_ih_0_0.vlm"},
	{0x0EE977E2CCE3D1BFULL, "vxlt_gallowstemplar_f_5ql_0_0.vlm"},
	{0x0EEA8CFE7A838F56ULL, "vxlt_0_lowtown_a2_n_euz_0_0.vlm"},
	{0x0EEA92AF713BF3F3ULL, "88084391.ogg"},
	{0x0EEB06515EA11D5FULL, "gen_im_wep_war_two_act2_02.uti"},
	{0x0EEB1996EDB4A325ULL, "character_character_p187.psh.fxo.dx11.assembly"},
	{0x0EEB2875E679FA15ULL, "hightown_male_elfservnt_mob.bnk"},
	{0x0EEB602A6EEB1816ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_wkf.gad"},
	{0x0EEB926F6B621E37ULL, "follower_banter6149291.cl"},
	{0x0EEBD6575C8B3D9AULL, "art\\characters\\weapons\\longswords\\w_lsw_ls101.mao"},
	{0x0EEC2D672970CE7AULL, "vxht_mansion2_216_0_0.vlm"},
	{0x0EED7DB721A8458BULL, "vxlt_hangedman_2bl_0_0.vlm"},
	{0x0EED942AFE3BE37CULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_stn_0n.dds"},
	{0x0EEE6DE2B42902E7ULL, "365332191.ogg"},
	{0x0EEEB8AC8E3B4698ULL, "plt_vrc300pt_n_rival_max.plo"},
	{0x0EEEBFC9D65EA04BULL, "hlac_w_blightlands_262141_0x.dds"},
	{0x0EEF454E0EFD5EEBULL, "vxlt_warehouse_n_18n_0_0.vlm"},
	{0x0EEF4A75E67E1D6DULL, "hightown_male_elfservnt_mob.cnv"},
	{0x0EEF65BE28F7F6E2ULL, "vxdr_dungeon_1mf_0_0.vlm"},
	{0x0EEFA34EC6CAB3BFULL, "911924409.ogg"},
	{0x0EEFA88F16558856ULL, "6125817_m.fxe"},
	{0x0EF00C9415477EF0ULL, "vxca_caverns2prc_acc_0_0.vlm"},
	{0x0EF01B118FC30CB1ULL, "and321tr_distract.ncs"},
	{0x0EF0E0B24F1EA60FULL, "vxw_woundedcoast_cg_0_0.vlm"},
	{0x0EF0E8C54E23C931ULL, "art\\levels\\races\\vfx_models\\prp_camplogmed.phy"},
	{0x0EF0EB40467DDDA7ULL, "6059984_m.fxe"},
	{0x0EF2009F3EDFE228ULL, "91551225.ogg"},
	{0x0EF2064EC448623AULL, "vxlt_0_lowtown_docks_a3_n_czf_0_1.vlm"},
	{0x0EF21BB24045C333ULL, "vxht_0_hightown_n_c6x_0_0.vlm"},
	{0x0EF2476AA77663A3ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_maga_0s.dds"},
	{0x0EF27B2A6EF16575ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_wkf.evt"},
	{0x0EF2C3AF3CDEBC03ULL, "6054688_m.fxe"},
	{0x0EF30397DAFEDD6FULL, "42246380.ogg"},
	{0x0EF3403AFE7B134EULL, "gen_im_wep_war_shd_act1_03.uti"},
	{0x0EF399B6A6DC41DBULL, "llac_ca_caverns1_24424_2.dds"},
	{0x0EF3F89DE7023EBBULL, "mag120cr_macha.utc"},
	{0x0EF488BFB18FC569ULL, "vxlt_hangedman_1ve_0_0.vlm"},
	{0x0EF53B3FFF0D45AFULL, "hlac_ht_0_hightown_n_-1_0y.dds"},
	{0x0EF579EA28525CB4ULL, "188453346.ogg"},
	{0x0EF58DBD284CF049ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose2left.anb"},
	{0x0EF5A5EB4DC1FD26ULL, "6143637_m.fxe"},
	{0x0EF5B4A967224338ULL, "hlac_lt_0_lowtown_docks_n_0_3x.dds"},
	{0x0EF60B144D849746ULL, "vxlt_0_lowtown_a2_n_c_bxp_0_0.vlm"},
	{0x0EF630F70E8506CCULL, "vxw_sundermount_318_0_1.vlm"},
	{0x0EF67423DEEAFAE1ULL, "49671202.ogg"},
	{0x0EF6892EAE1424D5ULL, "6104180_f.fxe"},
	{0x0EF6D787174AA5DFULL, "202839850.ogg"},
	{0x0EF6F98C0F170943ULL, "6046830_m.fxe"},
	{0x0EF79A3C7071BB55ULL, "vxlt_backalley_n_1wh_0_0.vlm"},
	{0x0EF7C44A0DF011EBULL, "w_encounter2_camp_19_0.anb"},
	{0x0EF7FAABB3B1AFA2ULL, "vxlt_0_lowtown_docks_a3_n_d07_0_1.vlm"},
	{0x0EF81F1AEA1A38A6ULL, "383956914.ogg"},
	{0x0EF81FEDF9B80611ULL, "262962501.ogg"},
	{0x0EF922F7CA660432ULL, "zz_and_debug6134145.cl"},
	{0x0EF980BA170333B0ULL, "6118019_m.fxe"},
	{0x0EF9A784AA1A1BE0ULL, "ss_templars_f_1.wbd"},
	{0x0EFA32137BCB08E3ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm01.mmh"},
	{0x0EFA3E8E669B7575ULL, "w_sundermount_152_0.msh"},
	{0x0EFA40137BCB203DULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm01.msh"},
	{0x0EFA408E669B78EBULL, "w_sundermount_152_0.mmh"},
	{0x0EFA4C8E669B8C88ULL, "w_sundermount_152_0.mao"},
	{0x0EFACAAB06766C6BULL, "662825978.ogg"},
	{0x0EFADBEE2A921341ULL, "6119890_f.fxe"},
	{0x0EFB2B7C00CAA7AFULL, "zz_isa_debug6133945.cl"},
	{0x0EFB464CB8539426ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_exit.evt"},
	{0x0EFB686A53AE2302ULL, "6052089_m.fxe"},
	{0x0EFB7B6721718667ULL, "gen00fl_anders6096475.cl"},
	{0x0EFBDEBA12AB26D4ULL, "418233689.ogg"},
	{0x0EFBE0369DED0F9AULL, "vxht_0_hightown_c5g_0_0.vlm"},
	{0x0EFBF57B42CC3CE1ULL, "follower_banter6167961.cl"},
	{0x0EFC2AB576F814FCULL, "6163981_m.fxe"},
	{0x0EFC41C3730770D2ULL, "222087197.ogg"},
	{0x0EFC47FADC6C08C5ULL, "hlac_w_encounter1_5_0y.dds"},
	{0x0EFC8B59AFA334F1ULL, "vxlt_gallowstemplar_f_3sv_0_0.vlm"},
	{0x0EFCE00B2A26ACCAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_keep.mao"},
	{0x0EFCE40B2A26B319ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_keep.mmh"},
	{0x0EFCEE0B2A26C407ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_keep.msh"},
	{0x0EFD1C9F639E0453ULL, "vxw_encounter2_b1_0_0.vlm"},
	{0x0EFD8F17825D0E5BULL, "hlac_ca_caverns2prc_5802_30z.dds"},
	{0x0EFE4F72F6C6E808ULL, "6072342_m.fxe"},
	{0x0EFE5F3D7A4ED152ULL, "art\\vfx\\textures\\vfx_gaping_wound.dds"},
	{0x0EFF528F4114BE8FULL, "vxw_woundedcoast_xf_0_0.vlm"},
	{0x0EFFA056566A6C7AULL, "vxlt_shop_ey_0_0.vlm"},
	{0x0F0050E8A54B421EULL, "hlac_ht_0_hightown_a3_1_1z.dds"},
	{0x0F00D2E94569CAE7ULL, "6128777_f.fxe"},
	{0x0F015D4CB8582771ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_exit.gad"},
	{0x0F01697609EF456CULL, "vxca_cavern1_haunted_j2b_0_2.vlm"},
	{0x0F0175CABE121E3CULL, "hlac_ht_brothel_n_3372_0z.dds"},
	{0x0F01F852CC1712B7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_2windows.phy"},
	{0x0F02389C82862C6BULL, "6197880_m.fxe"},
	{0x0F026F80F062EF81ULL, "art\\vfx\\materials\\weapontrail\\weapontrail_streaks.mao"},
	{0x0F0304DD9326D7EEULL, "t3_com_c00.tnt"},
	{0x0F03413E5E2166D4ULL, "w_encounter2_41_0.phy"},
	{0x0F036620F682373BULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintinga.phy"},
	{0x0F03FEED668C40A5ULL, "79713091.ogg"},
	{0x0F042F579AF1190BULL, "vxht_0_hightown_a2_bhx_0_0.vlm"},
	{0x0F044DA803BD3A7DULL, "6102941_m.fxe"},
	{0x0F04EDBD6F11D99EULL, "237036029.ogg"},
	{0x0F054F32C07A61C0ULL, "889945597.ogg"},
	{0x0F05823884CECB8AULL, "zz_spider_stomp_v2.cut"},
	{0x0F05B8C1F9823809ULL, "6016679_m.fxe"},
	{0x0F05F0E5FE0798F3ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_p_wkf.ani"},
	{0x0F060ABF52BE1F78ULL, "165074452.ogg"},
	{0x0F060DA7C8167EB8ULL, "6076100_m.fxe"},
	{0x0F068AFD640D8670ULL, "zz_jp_spike_head.cut"},
	{0x0F069D5524F2A2D2ULL, "vxw_encounter2_b4_0_0.vlm"},
	{0x0F07303233683C85ULL, "vxht_mansion2_1ge_0_1.vlm"},
	{0x0F0786A0DF4A910FULL, "ser121_jethann6104618.cl"},
	{0x0F078B707081D1CAULL, "739763591.ogg"},
	{0x0F0820689C6D2C27ULL, "94398666.ogg"},
	{0x0F08400B13A7A0D8ULL, "vxht_0_hightown_a3_aif_0_0.vlm"},
	{0x0F085513C65DF57DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsstair.mao"},
	{0x0F085913C65DFBB6ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsstair.mmh"},
	{0x0F085E4CB85E471FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_exit.ani"},
	{0x0F086313C65E0CB4ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsstair.msh"},
	{0x0F089F61F6E973FCULL, "vxlt_warehouse_vi_0_0.vlm"},
	{0x0F08B11C375929B3ULL, "art\\characters\\heraldry\\hld_alienage.dds"},
	{0x0F08CAF0338A9BF0ULL, "character_character_p10.psh.fxo.dx11"},
	{0x0F090FFC8F6DA8E2ULL, "m_rai301ar_caverns.wbd"},
	{0x0F09124C8A122245ULL, "vxca_caverns1_mines_dpi_0_0.vlm"},
	{0x0F093DA0122A25A3ULL, "hlac_lt_0_lowtown_a2_n_65534_6z.dds"},
	{0x0F094093A5A72321ULL, "character_character_p190.psh.fxo.dx11"},
	{0x0F096599F44D1912ULL, "879917676.ogg"},
	{0x0F0988FF1B719806ULL, "zz_dae_debug6029077.cl"},
	{0x0F09C7672EA3F79EULL, "ser270_emeric6111484.cl"},
	{0x0F0A383E5E27767CULL, "w_encounter2_41_0.rml"},
	{0x0F0A69A20A8D39B7ULL, "6119411_m.fxe"},
	{0x0F0A9AA597CF6BB9ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkright.ani"},
	{0x0F0B33137BD97D08ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm01.phy"},
	{0x0F0B733306B1FC95ULL, "vxlt_0_lowtown_a3_n_c_htx_0_0.vlm"},
	{0x0F0B80D91B33F20BULL, "38807922.ogg"},
	{0x0F0B8E7C55941CD3ULL, "vxw_encounter2_m0_0_0.vlm"},
	{0x0F0CCD0A5CC28529ULL, "vxw_encounter3_6d_0_0.vlm"},
	{0x0F0D516123B0A5D6ULL, "vxlt_0_lowtown_n_f9h_0_1.vlm"},
	{0x0F0D62A145C42167ULL, "6169438_m.fxe"},
	{0x0F0D7F5043F514A0ULL, "948066913.ogg"},
	{0x0F0D8F2B36787F6FULL, "6128012_f.fxe"},
	{0x0F0E1E3E5E2B32D1ULL, "w_encounter2_41_0.msh"},
	{0x0F0E24668577DC70ULL, "vxlt_0_lowtown_a3_n_c_hp5_0_0.vlm"},
	{0x0F0E2C3E5E2B4AECULL, "w_encounter2_41_0.mao"},
	{0x0F0E383E5E2B5E8FULL, "w_encounter2_41_0.mmh"},
	{0x0F0E98C99C2F70A7ULL, "dae000_pc_home_ambients6208687.cl"},
	{0x0F0EFF6C2431F433ULL, "vxht_0_hightown_n_c8z_0_0.vlm"},
	{0x0F0F43BD1757467FULL, "one200_gold_followers6117306.cl"},
	{0x0F0F52FA815F5E6DULL, "6113987_m.fxe"},
	{0x0F0F7CCEB3A919ABULL, "character_character_p95.psh.fxo.dx11.assembly"},
	{0x0F10087E90E33A4EULL, "6110619_m.fxe"},
	{0x0F10E8E6B181CC3BULL, "55454190.ogg"},
	{0x0F1147B26BD0A96CULL, "vxdr_0_deeproads_a6y_0_0.vlm"},
	{0x0F116C79957BF669ULL, "vxlt_hangedman_1xk_0_0.vlm"},
	{0x0F1175A597D54BFBULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkright.gad"},
	{0x0F117C14C45FBE22ULL, "578816827.ogg"},
	{0x0F1181265EAD4417ULL, "dae000_tomwise6167343.cl"},
	{0x0F11980315E420DEULL, "dae340st_cowan.stg"},
	{0x0F12E8E5FE12152AULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_p_wkf.evt"},
	{0x0F1325E06FE591E5ULL, "888840423.ogg"},
	{0x0F13D041AB8B9D63ULL, "zz_vault_debug6177793.cl"},
	{0x0F13FD2A67039936ULL, "vxca_cavern1_haunted_i1i_0_1.vlm"},
	{0x0F1430431B15BC07ULL, "1058424968.ogg"},
	{0x0F145C3B32342759ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_right_down_loop.gad"},
	{0x0F14BA5144340910ULL, "6175906_m.fxe"},
	{0x0F150F3216E8BB52ULL, "vxw_encounter2_m5_0_0.vlm"},
	{0x0F15D66EF23D4D64ULL, "6207397_m.fxe"},
	{0x0F163C0450E41ABCULL, "def_dirt_n.dds"},
	{0x0F163E9C1CDA1B18ULL, "vxw_bonepitdestroyed_a3_dx_0_0.vlm"},
	{0x0F16B1E8059DB958ULL, "vxht_0_hightown_a3_avg_0_0.vlm"},
	{0x0F16E8B20252F1A9ULL, "6103021_m.fxe"},
	{0x0F1795E5F36CDEE7ULL, "docks_male_laborer_mob6175405.cl"},
	{0x0F17D124DA825866ULL, "hlac_lt_0_lowtown_a2_n_131070_6z.dds"},
	{0x0F18B8D531213B3FULL, "vxca_caverns2_8l0_0_1.vlm"},
	{0x0F18F9F4DA041221ULL, "mag360_thrask6053254.cl"},
	{0x0F197F617F54AF63ULL, "776926097.ogg"},
	{0x0F19C7E5FE17FB9DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_p_wkf.gad"},
	{0x0F1A0059478A3553ULL, "vxw_encounter2_h3_0_0.vlm"},
	{0x0F1A378C9811FC38ULL, "vxlt_undercityserialkiller_1j7_0_0.vlm"},
	{0x0F1AA8FCB043EB9BULL, "vxlt_0_lowtown_a3_n_c_ere_0_0.vlm"},
	{0x0F1AC095E6F7BA4BULL, "hlac_w_sundermount_-196611_0y.dds"},
	{0x0F1AD0D47ED6330CULL, "hlac_lt_0_lowtown_a2_n_c_1_10x.dds"},
	{0x0F1B14865D35C030ULL, "dae000_mother6156396.cl"},
	{0x0F1B3EE74EB89DA9ULL, "vxw_encounter3_3k_0_0.vlm"},
	{0x0F1B44CC1C92B311ULL, "vxw_encounter2_camp_bs_0_0.vlm"},
	{0x0F1B6D3B323A62A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_right_down_loop.ani"},
	{0x0F1B94ED0629E814ULL, "mag320cr_circle_mage.utc"},
	{0x0F1C623C21696B46ULL, "root_rog_duelist.dds"},
	{0x0F1D1050BFFD4B12ULL, "lgt261cr_archer.utc"},
	{0x0F1DB93E80101EBFULL, "987619100.ogg"},
	{0x0F1E07FF6AD4BDEFULL, "follower_banter6170624.cl"},
	{0x0F1E17EA0FF58674ULL, "362391790.ogg"},
	{0x0F1E3C5EEEE74E2BULL, "lgt000im_ultra_shawl.uti"},
	{0x0F1E4E076C9DAC69ULL, "dae000ip_crowd_ff3_talk9.utp"},
	{0x0F1ECBD319F67CEEULL, "707707055.ogg"},
	{0x0F1F7220E935F0C3ULL, "hlac_w_blightlands_65536_1z.dds"},
	{0x0F1FDE5687720EE9ULL, "vxlt_hangedman_1gj_0_0.vlm"},
	{0x0F1FE76A464CFA41ULL, "795165451.ogg"},
	{0x0F2041781733336BULL, "75531064.ogg"},
	{0x0F20952EDC4F982EULL, "617091763.ogg"},
	{0x0F210A8D7C0A4B98ULL, "hlac_lt_0_lowtown_a2_n_c_-1_6z.dds"},
	{0x0F2130CB5474D1CAULL, "hlac_lt_0_lowtown_docks_a3_-2_2z.dds"},
	{0x0F21527179838293ULL, "vxht_keep_n_5zv_0_0.vlm"},
	{0x0F2185CC2829C2E6ULL, "594547529.ogg"},
	{0x0F219E6B71A5623FULL, "gen00flst_varric_posthaunt.stg"},
	{0x0F21A7DB10667E59ULL, "671612925.ogg"},
	{0x0F21F1189FA4119CULL, "llac_lt_0_lowtown_docks_n_131071_4.dds"},
	{0x0F22390558A39F44ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.mmh"},
	{0x0F22CA0584F979E0ULL, "hlac_w_bonepit_196612_0y.dds"},
	{0x0F22EFD3465404D5ULL, "vxlt_backalley_n_1xu_0_0.vlm"},
	{0x0F234C0C5A4AB8FAULL, "dog221ip_table_leg.utp"},
	{0x0F23680CD7EB165DULL, "gen00fl_anders6112966.cl"},
	{0x0F23810F08DED3D2ULL, "vxw_encounter2_h6_0_0.vlm"},
	{0x0F23AB36F27D526EULL, "w_blightlands_67_0.msh"},
	{0x0F23B936F27D69BFULL, "w_blightlands_67_0.mao"},
	{0x0F23C14A0E154737ULL, "w_encounter2_camp_19_0.rml"},
	{0x0F23C536F27D7E5CULL, "w_blightlands_67_0.mmh"},
	{0x0F23EF64C5A758BCULL, "mag130_danzig6018820.cl"},
	{0x0F243DCBAEC23396ULL, "art\\levels\\ztools\\default\\generic\\0co_basic.phy"},
	{0x0F24706D541D6B27ULL, "579770097.ogg"},
	{0x0F247916BB1A1174ULL, "6128926_m.fxe"},
	{0x0F24B0790ED03398ULL, "vxw_bonepitdestroyed_a3_oy_0_0.vlm"},
	{0x0F24D9F9CB528EB3ULL, "vxw_woundedcoast_o7_0_0.vlm"},
	{0x0F24E2BDF05FC0ECULL, "6075084_m.fxe"},
	{0x0F2503E480D4B975ULL, "6110267_m.fxe"},
	{0x0F2537ECF06A6789ULL, "6069848_m.fxe"},
	{0x0F2549A43FA6A6B5ULL, "mag160_varian_ilithis6016511.cl"},
	{0x0F259F8209900DA0ULL, "6018746_m.fxe"},
	{0x0F25D6B24C0F13DAULL, "hlac_w_bonepitmining_a2_458754_0y.dds"},
	{0x0F25D95189AF7A21ULL, "vxht_brothel_n_628_0_0.vlm"},
	{0x0F2621984253DB98ULL, "6033099_m.fxe"},
	{0x0F264E47AF95E5F0ULL, "vxlt_undercitydungeon_1c9_0_0.vlm"},
	{0x0F2662FD262C4D42ULL, "vxlt_hangedman_1jv_0_0.vlm"},
	{0x0F26A87089B94895ULL, "6064726_m.fxe"},
	{0x0F272460A817EE1AULL, "follower_banter6163421.cl"},
	{0x0F272AF6493404E5ULL, "follower_banter6136358.cl"},
	{0x0F27422C36C19050ULL, "9244160.ogg"},
	{0x0F27AFDBC741E640ULL, "vxlt_undercitydungeon_1ge_0_0.vlm"},
	{0x0F28723639804DD3ULL, "vxw_encounter2_s2_0_0.vlm"},
	{0x0F2875C748CC0B3DULL, "follower_banter6151353.cl"},
	{0x0F28A9698A0814B8ULL, "vxlt_undercityserialkiller_1o4_0_0.vlm"},
	{0x0F291AD9A23A041BULL, "vxlt_0_lowtown_a3_n_c_eud_0_0.vlm"},
	{0x0F2A97FA5ADEFBC5ULL, "vxlt_shop_er_0_0.vlm"},
	{0x0F2AC04A0E1B642FULL, "w_encounter2_camp_19_0.phy"},
	{0x0F2AC6ED9AE95396ULL, "follower_banter6160608.cl"},
	{0x0F2AE64AFDA851C6ULL, "singleroofa_n.dds"},
	{0x0F2AFF24B92D1E2BULL, "vxca_caverns1_mines_im4_0_1.vlm"},
	{0x0F2B1914961D90A2ULL, "zz_lgt_2_debug6206577.cl"},
	{0x0F2B2F2A7EFEAA9EULL, "685525186.ogg"},
	{0x0F2B6BA9E5EABE75ULL, "vxlt_0_lowtown_gkr_0_0.vlm"},
	{0x0F2B92DB81F9BC4EULL, "6084712_m.fxe"},
	{0x0F2C1668AD642812ULL, "vxlt_0_lowtown_docks_a3_aoi_0_0.vlm"},
	{0x0F2C795F9E01ECA7ULL, "6082235_f.fxe"},
	{0x0F2C9E3A44E2424BULL, "6110427_m.fxe"},
	{0x0F2CD8CCC9A554F3ULL, "lgt120cr_orz_bln_servant.utc"},
	{0x0F2CF0B46CC3B64CULL, "6059015_m.fxe"},
	{0x0F2D497FA76A74F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_loop.ani"},
	{0x0F2D87086E73542BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_loop.ani"},
	{0x0F2DC34E1060AE61ULL, "dae100cr_mother.utc"},
	{0x0F2DD81692946F6CULL, "172714367.ogg"},
	{0x0F2E246E3C8B3CF0ULL, "lt_0_lowtown_docks_a3_150_0.mao"},
	{0x0F2E24F991721C5DULL, "gen00fl_varric6119002.cl"},
	{0x0F2E26CBAECA6137ULL, "art\\levels\\ztools\\default\\generic\\0co_basic.msh"},
	{0x0F2E286E3C8B4303ULL, "lt_0_lowtown_docks_a3_150_0.mmh"},
	{0x0F2E366E3C8B5B5DULL, "lt_0_lowtown_docks_a3_150_0.msh"},
	{0x0F2E38CBAECA7F9AULL, "art\\levels\\ztools\\default\\generic\\0co_basic.mao"},
	{0x0F2E3CCBAECA86A9ULL, "art\\levels\\ztools\\default\\generic\\0co_basic.mmh"},
	{0x0F2E503379682769ULL, "vxlt_hangedman_1bi_0_0.vlm"},
	{0x0F2E81BC9310E13AULL, "313276275.ogg"},
	{0x0F2EED8B257A7064ULL, "6031699_m.fxe"},
	{0x0F2F349F598BB152ULL, "gen00fl_isabela6101547.cl"},
	{0x0F2F4D20F6A747DAULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintinga.msh"},
	{0x0F2F4F2E3E8363C5ULL, "6107344_m.fxe"},
	{0x0F2F5B20F6A75F13ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintinga.mao"},
	{0x0F2F5F20F6A76620ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintinga.mmh"},
	{0x0F2F6DD1CD4DA4FFULL, "139513733.ogg"},
	{0x0F2FF38302D87393ULL, "556935112.ogg"},
	{0x0F30430889D1635CULL, "6036742_m.fxe"},
	{0x0F3074A9016CB793ULL, "prp_testladder.mao"},
	{0x0F30ECEC442831B6ULL, "hlac_lt_0_lowtown_a2_65536_11z.dds"},
	{0x0F3154E51C419ED1ULL, "6149154_f.fxe"},
	{0x0F317BF00CFA15C4ULL, "6189912_m.fxe"},
	{0x0F31BA3F00B6686CULL, "llac_lt_0_lowtown_a3_n_c_-65536_3.dds"},
	{0x0F31E04186EA2CA5ULL, "6043758_m.fxe"},
	{0x0F32768392F8BD91ULL, "dre180ip_deep_roads_chest_2.utp"},
	{0x0F32B3EF78438DA4ULL, "801723807.ogg"},
	{0x0F32D3B2B1941D02ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockxxl.msh"},
	{0x0F32D5B2B19420B8ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockxxl.mmh"},
	{0x0F32E1B2B194355BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockxxl.mao"},
	{0x0F331E614330B474ULL, "vxca_cavern1_haunted_gkb_0_0.vlm"},
	{0x0F33467FA76EDB56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_loop.gad"},
	{0x0F334BD6BD48A733ULL, "vxw_woundedcoast_d6_0_0.vlm"},
	{0x0F3367485B083F24ULL, "one280_iwan6096874.cl"},
	{0x0F338D30ECA1595FULL, "lgt000ip_toe_crate.utp"},
	{0x0F341FBB49A7FECAULL, "vxlt_0_lowtown_n_e2w_0_0.vlm"},
	{0x0F34372ACBFBCE42ULL, "6142634_m.fxe"},
	{0x0F343F568360950AULL, "vxw_bonepitdestroyed_a3_26x_0_0.vlm"},
	{0x0F346E384CA28DF5ULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gax_002a_n.dds"},
	{0x0F34B10EF5C04AE4ULL, "193064054.ogg"},
	{0x0F34CC200BFF0266ULL, "gen00fl_aveline6114845.cl"},
	{0x0F357C32F7867FEEULL, "hlac_lt_0_lowtown_docks_n_-131072_0x.dds"},
	{0x0F35E7891A47282CULL, "6136693_m.fxe"},
	{0x0F3621B8B937FEC0ULL, "vxlt_undercitydungeon_1jb_0_0.vlm"},
	{0x0F362E001F728813ULL, "vxdr_primevalentrance_1r0_0_0.vlm"},
	{0x0F365D3BA924516CULL, "6106720_m.fxe"},
	{0x0F366F27C1ADB199ULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_lp.gad"},
	{0x0F36825D89FFCAB3ULL, "follower_banter6156260.cl"},
	{0x0F371B467BFE2D38ULL, "vxlt_undercityserialkiller_1d5_0_0.vlm"},
	{0x0F374BAAA1F8E053ULL, "ach_seb_retribution.dds"},
	{0x0F378CB694301C9BULL, "vxlt_0_lowtown_a3_n_c_exg_0_0.vlm"},
	{0x0F37B77F0D84EECBULL, "6149515_m.fxe"},
	{0x0F38384D866A9D28ULL, "vxht_keep_n_3c8_0_0.vlm"},
	{0x0F383BBD0D04DFE7ULL, "lt_0_lowtown_a2_n_mmi.gff"},
	{0x0F3852C9B466C187ULL, "dre000_varric6033307.cl"},
	{0x0F38DC9F7EE0F3DFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb05.phy"},
	{0x0F38F51F9D040DE1ULL, "gen00fl_dog6176631.cl"},
	{0x0F3953A95EAA5D9DULL, "6103731_m.fxe"},
	{0x0F39B162ECD187D2ULL, "vxca_caverns2_62p_0_0.vlm"},
	{0x0F3A747D8443EFA5ULL, "416150262.ogg"},
	{0x0F3B41495AEF802CULL, "6089601_m.fxe"},
	{0x0F3B7081230527C2ULL, "vxw_encounter2_camp_7c_0_0.vlm"},
	{0x0F3B8B79DC67D27AULL, "w_bonepitdestroyed_a3_194_0.mmh"},
	{0x0F3B8F79DC67D8B1ULL, "w_bonepitdestroyed_a3_194_0.mao"},
	{0x0F3B9D79DC67F0C0ULL, "w_bonepitdestroyed_a3_194_0.msh"},
	{0x0F3BAC95C1EAD9FFULL, "305417067.ogg"},
	{0x0F3BB96D9784D205ULL, "llac_ht_0_hightown_a3_65536_0.dds"},
	{0x0F3BFF567044CBBFULL, "vxdr_0_deeproads_a6v_0_0.vlm"},
	{0x0F3C1B6BFE8E8035ULL, "zz_dre_debug6168233.cl"},
	{0x0F3C5CD67EA88466ULL, "6049452_m.fxe"},
	{0x0F3C6D6D37AD0355ULL, "vxca_caverns1_dq7_0_0.vlm"},
	{0x0F3CF295E74C2ED7ULL, "vxht_chantry_n_2ip_0_0.vlm"},
	{0x0F3D10FA18928B60ULL, "327284048.ogg"},
	{0x0F3D197007235403ULL, "385805222.ogg"},
	{0x0F3D8027C1B3EDE7ULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_lp.ani"},
	{0x0F3EC335661D0C55ULL, "321886379.ogg"},
	{0x0F3F296E3C99B828ULL, "lt_0_lowtown_docks_a3_150_0.phy"},
	{0x0F3F690DD9FD7E70ULL, "625297623.ogg"},
	{0x0F3FD483E97BD406ULL, "6066385_m.fxe"},
	{0x0F3FE8FB8A621142ULL, "641855487.ogg"},
	{0x0F40294F2F7267E3ULL, "vxht_0_hightown_c8m_0_0.vlm"},
	{0x0F40C410ED2EB1E5ULL, "351914352.ogg"},
	{0x0F40C70109295611ULL, "689196802.ogg"},
	{0x0F410649957D41D7ULL, "6135655_m.fxe"},
	{0x0F411A2F93036634ULL, "vxht_brothel_n_6bq_0_0.vlm"},
	{0x0F414E690F0A4071ULL, "117155377.ogg"},
	{0x0F41BDB3AF3EBFB3ULL, "vxw_woundedcoast_a9_0_0.vlm"},
	{0x0F42612CDF72634BULL, "vxlt_0_lowtown_docks_n_aoh_0_0.vlm"},
	{0x0F426E086E8585B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_loop.gad"},
	{0x0F427E4EF78C598EULL, "vxlt_0_lowtown_n_gby_0_0.vlm"},
	{0x0F429397D740967FULL, "gen00fl_aveline6043100.cl"},
	{0x0F4303A23EA1871DULL, "6125317_m.fxe"},
	{0x0F439092290B2B60ULL, "fx_arrow.dds"},
	{0x0F43EF8912575495ULL, "6165267_m.fxe"},
	{0x0F448FEE8820CFD4ULL, "lgt000im_azurejamos.uti"},
	{0x0F44C7D14E3C62E0ULL, "441208645.ogg"},
	{0x0F44CCDCE1266A36ULL, "6185235_m.fxe"},
	{0x0F44F7F7E6D644A1ULL, "vxht_0_hightown_a2_avo_0_0.vlm"},
	{0x0F451B666245A157ULL, "zz_fol_debug6098073.cl"},
	{0x0F45897A6131DA0BULL, "6111536_m.fxe"},
	{0x0F458CF75524DFDAULL, "one120_athenril6088904.cl"},
	{0x0F458D236DF445B8ULL, "vxlt_undercityserialkiller_1y2_0_0.vlm"},
	{0x0F465E65B980B253ULL, "character_character_p28.psh.fxo.dx11"},
	{0x0F46B51C39763F64ULL, "6127090_f.fxe"},
	{0x0F46E650E3367245ULL, "ab_cyc.mao"},
	{0x0F480D2B4A1F99A9ULL, "dae011_conspiracy_mob6141231.cl"},
	{0x0F4875C079802525ULL, "360945303.ogg"},
	{0x0F48B71AFCE5B095ULL, "163526990.ogg"},
	{0x0F48CD52CC52CC0FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_2windows.mao"},
	{0x0F48D6ED6DDD760CULL, "85178065.ogg"},
	{0x0F48D952CC52E0ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_2windows.mmh"},
	{0x0F48DF52CC52EA9EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_2windows.msh"},
	{0x0F49771E981FC28DULL, "839778586.ogg"},
	{0x0F49DD9F7EEF68E4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb05.mmh"},
	{0x0F49F39F7EEF8D86ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb05.msh"},
	{0x0F4A6621C5580B47ULL, "dae000_flemeth6013380.cl"},
	{0x0F4A9C1A65F77FF9ULL, "face_face_p2.psh.fxo.dx11.assembly"},
	{0x0F4B3C0558C6A93BULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.anb"},
	{0x0F4B57CC37676092ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medg_0d.dds"},
	{0x0F4B6472D9424757ULL, "vxht_chantry_n_2ds_0_0.vlm"},
	{0x0F4BECD7C4D20344ULL, "llac_dr_0_deeproads_17038_1.dds"},
	{0x0F4C99226ABFA545ULL, "vxlt_gallowscourtyard_77w_0_0.vlm"},
	{0x0F4CE99A1035508AULL, "6096863_m.fxe"},
	{0x0F4CF670286F770BULL, "dae000_flemeth6013418.cl"},
	{0x0F4D51455390C6CBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightlighta_sa.dds"},
	{0x0F4DD22CF5A2C0FAULL, "hlac_lt_gallowstemplar_n_2383_7z.dds"},
	{0x0F4E872EEA0302EDULL, "gen00fl_bethany6097595.cl"},
	{0x0F4E9266C5A3F4B4ULL, "269454286.ogg"},
	{0x0F4F8C0C84F97EB4ULL, "vxht_brothel_n_6ov_0_0.vlm"},
	{0x0F4FC4E574FAED59ULL, "703482465.ogg"},
	{0x0F4FEC677C3FAA17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_enter.gad"},
	{0x0F501B4AA595D38EULL, "l_dae110ar_lowtown_nt.wbd"},
	{0x0F501DFC8FA9C298ULL, "m_rai301ar_caverns.bnk"},
	{0x0F5067265A74442EULL, "hlac_w_blightlands_-131070_0x.dds"},
	{0x0F50BC043DE70647ULL, "lit_aoe_03_l3.mao"},
	{0x0F51E8B82C83D362ULL, "gen_im_cns_tom_rog_shadow.uti"},
	{0x0F52DF60669AE5F3ULL, "dm_dwarven_scout.mor"},
	{0x0F52F3903D7B1400ULL, "vxlt_warehouse_n_1ef_0_0.vlm"},
	{0x0F53065B23D2942EULL, "6076854_m.fxe"},
	{0x0F535020A6C0D94BULL, "6015845_m.fxe"},
	{0x0F53FE71D8B737AAULL, "6072289_m.fxe"},
	{0x0F5431F6F9743852ULL, "269498730.ogg"},
	{0x0F548BCA03B402EAULL, "follower_banter6156054.cl"},
	{0x0F54A2ABE68A6890ULL, "character_character_p67.psh.fxo.dx11.assembly"},
	{0x0F54C6C5BCCDBDD4ULL, "vxlt_0_lowtown_a3_n_c_g60_0_0.vlm"},
	{0x0F5570C575AAB334ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbodyhang.phy"},
	{0x0F55DE5DD2215A87ULL, "vxw_woundedcoast_nd_0_0.vlm"},
	{0x0F55FDDDA7B4FF71ULL, "vxht_0_hightown_a2_a81_0_0.vlm"},
	{0x0F56197E74027544ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.mmh"},
	{0x0F562ACAC4F354C9ULL, "narrative_varric6172515.cl"},
	{0x0F56456E988487F2ULL, "vxlt_0_lowtown_docks_a3_ctn_0_0.vlm"},
	{0x0F5679B7B20F32E5ULL, "544989242.ogg"},
	{0x0F56C8D71A1B7D97ULL, "plt_cod_ltr_otl_emile.plo"},
	{0x0F56F1E822443984ULL, "ch1armtnt_s2y.psh.fxo"},
	{0x0F5701677C45ECA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_enter.evt"},
	{0x0F57093893F3E004ULL, "isa211_isabela6202285.cl"},
	{0x0F578D453D422D51ULL, "457946509.ogg"},
	{0x0F58E1C410DBC71AULL, "mrl000_marethari6068818.cl"},
	{0x0F59202CB845C14FULL, "ser221_gascard6133878.cl"},
	{0x0F592D8F5AADB45AULL, "t1_mub_re3.tnt"},
	{0x0F594F750D420CA8ULL, "893009599.ogg"},
	{0x0F59B19BAFBEB287ULL, "6166880_m.fxe"},
	{0x0F5A4CAB963D6A9CULL, "hf_mage_1.mor"},
	{0x0F5A790719C928C0ULL, "ser120_ghyslain6104994.cl"},
	{0x0F5A84E0A44A0384ULL, "vxlt_0_lowtown_docks_a3_n_czh_0_1.vlm"},
	{0x0F5B0AFF5CB5BDC5ULL, "vxlt_gallowscourtyard_78v_0_0.vlm"},
	{0x0F5B219E3EA02AF2ULL, "dae000ip_crowd_mf2_merchant.utp"},
	{0x0F5B3F9C18DB5A10ULL, "hlac_lt_0_lowtown_docks_a3_262143_0y.dds"},
	{0x0F5B9557CEE48002ULL, "93682556.ogg"},
	{0x0F5B9C837BB56512ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\exit.gad"},
	{0x0F5BB9461C15E027ULL, "1052649284.ogg"},
	{0x0F5C5540FC08CDD3ULL, "32214188.ogg"},
	{0x0F5CD62DF02F596AULL, "619188711.ogg"},
	{0x0F5D0D09135E98E3ULL, "vxht_0_hightown_c2o_0_0.vlm"},
	{0x0F5D75C155304B40ULL, "llac_ca_caverns2_5802_69.dds"},
	{0x0F5DF01B13DD3F07ULL, "vxht_0_hightown_a2_c74_0_0.vlm"},
	{0x0F5DFDE976EF9734ULL, "vxht_brothel_n_6hw_0_0.vlm"},
	{0x0F5EE50B2A797F06ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rw_keep.phy"},
	{0x0F5F2B8603D2F335ULL, "658655956.ogg"},
	{0x0F5F44E6C35E944BULL, "vxlt_0_lowtown_docks_n_aqj_0_0.vlm"},
	{0x0F5F4DC575B2CC31ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbodyhang.msh"},
	{0x0F5F5BC575B2E44CULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbodyhang.mao"},
	{0x0F5F67C575B2F8EFULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbodyhang.mmh"},
	{0x0F5FA6CD2CE86814ULL, "665474121.ogg"},
	{0x0F5FD5895AF0CFE9ULL, "6111264_m.fxe"},
	{0x0F602EF9B9094DBDULL, "zz_lgt_1_debug6136993.cl"},
	{0x0F6054F36FF9F6A8ULL, "docks_male_overseer_mob6176575.cl"},
	{0x0F6087AEFEF25906ULL, "6098836_m.fxe"},
	{0x0F610AE5780B76DFULL, "zz_followers_banter6158556.cl"},
	{0x0F611E51540741C2ULL, "vxca_caverns1_j4m_0_0.vlm"},
	{0x0F618396F554D193ULL, "vxdr_primevalentrance_1a5_0_0.vlm"},
	{0x0F618864FE78724BULL, "shelf_sa.dds"},
	{0x0F61DBB1CAC275A1ULL, "vxht_0_hightown_a2_alm_0_0.vlm"},
	{0x0F629F837BBB8949ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\exit.evt"},
	{0x0F62AB859BE7D738ULL, "lla_ht_0_hightown_a3_65536_0.dds"},
	{0x0F63048AB5F9A863ULL, "llac_lt_0_lowtown_docks_n_0_2.dds"},
	{0x0F630A46A576B9FCULL, "blood_conj_06a_l2.mao"},
	{0x0F635119057C4C62ULL, "6101132_m.fxe"},
	{0x0F6389E99DB18BF6ULL, "6068376_m.fxe"},
	{0x0F638E98257998B0ULL, "6045858_m.fxe"},
	{0x0F63B55697B84112ULL, "mcr371_hostile_merrill6082145.cl"},
	{0x0F63C81AE7F464E4ULL, "vxht_mansion2_3g7_0_0.vlm"},
	{0x0F6419677C509F9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_enter.ani"},
	{0x0F642DF0DA970F8BULL, "hlac_lt_0_lowtown_a2_n_131074_0y.dds"},
	{0x0F648355C64AF512ULL, "hlac_w_encounter3_0_2y.dds"},
	{0x0F64EDCF882C675EULL, "146336760.ogg"},
	{0x0F65C170A2D36A74ULL, "vxdr_primevaltomb_6up_0_1.vlm"},
	{0x0F65C29C8DFF3390ULL, "vxca_caverns1_mines_ec8_0_0.vlm"},
	{0x0F65FE1073AD526AULL, "72752030.ogg"},
	{0x0F66310669962F53ULL, "gen00fl_sebastian6180298.cl"},
	{0x0F66EBAA2D47D2B9ULL, "6146857_m.fxe"},
	{0x0F66FA9974C85DCBULL, "174184466.ogg"},
	{0x0F66FAB79FCFF5BBULL, "425195658.ogg"},
	{0x0F6736CE39E2EE5CULL, "bkl_star.mao"},
	{0x0F675E6B092446D7ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4_tool.phy"},
	{0x0F676CA1FA902D36ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v3.ani"},
	{0x0F687EBDC64A6957ULL, "60034220.ogg"},
	{0x0F688A9276676A03ULL, "gen00fl_fenris6115354.cl"},
	{0x0F68D06538A66893ULL, "560566663.ogg"},
	{0x0F69F5429AE3BCB4ULL, "1045647137.ogg"},
	{0x0F6A1C24E20DB875ULL, "vxht_chantry_n_36v_0_0.vlm"},
	{0x0F6A4FEB166A92CCULL, "331867574.ogg"},
	{0x0F6A94920C139029ULL, "vxht_0_hightown_n_b8a_0_0.vlm"},
	{0x0F6AD1704BB02FC6ULL, "694873346.ogg"},
	{0x0F6AD7D71A2C400CULL, "plt_cod_ltr_otl_emile.nss"},
	{0x0F6B1060832E02EBULL, "lgt260_eau_tmp_roderick6143718.cl"},
	{0x0F6BE0929B45978CULL, "art\\vfx\\combat\\material__28.mao"},
	{0x0F6C4802220C5770ULL, "6020420_m.fxe"},
	{0x0F6C6FC668E5AFB4ULL, "vxht_brothel_n_6ut_0_0.vlm"},
	{0x0F6CCCCA911F3D99ULL, "vxht_0_hightown_n_cas_0_0.vlm"},
	{0x0F6CF4C3CDFFD5C1ULL, "224700743.ogg"},
	{0x0F6D6EF36ACEF39DULL, "follower_banter6172981.cl"},
	{0x0F6DF7E64B7CCEF1ULL, "vxlt_gallowstemplar_n_3or_0_0.vlm"},
	{0x0F6E24E3006C7C76ULL, "vxht_keep_3zr_0_0.vlm"},
	{0x0F6E639824B8F2B0ULL, "vxlt_0_lowtown_a3_n_c_cmq_0_0.vlm"},
	{0x0F6E7F11F2A9666DULL, "prcscr_col.gda"},
	{0x0F6F73692293A9E3ULL, "fen110_fenris6118060.cl"},
	{0x0F6F9EFD16F7D2EEULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_urn.msh"},
	{0x0F6FACFD16F7EA3FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_urn.mao"},
	{0x0F6FB8FD16F7FEDCULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_urn.mmh"},
	{0x0F70307659C30EB3ULL, "vxw_bonepit_1x0_0_0.vlm"},
	{0x0F70601387E88923ULL, "llac_ht_chantry_2534_0.dds"},
	{0x0F7087837BC79DB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\exit.ani"},
	{0x0F70AA6E7A240C34ULL, "vxw_sundermount_26a_0_0.vlm"},
	{0x0F716F7DAD70D976ULL, "958788315.ogg"},
	{0x0F724719232B2DDEULL, "hlac_ca_cavern1_haunted_24424_21y.dds"},
	{0x0F724FA0BECCB420ULL, "500851506.ogg"},
	{0x0F72C3C6CB5A0082ULL, "dae320ip_at_chantry.utp"},
	{0x0F731DF1F4359920ULL, "vxht_keep_n_671_0_0.vlm"},
	{0x0F733A57313FF928ULL, "vxlt_0_lowtown_a2_n_c_i16_0_0.vlm"},
	{0x0F7395102B6063ECULL, "vxht_0_hightown_a3_ca4_0_0.vlm"},
	{0x0F7487B601C63CE9ULL, "6121032_m.fxe"},
	{0x0F75581C7D349C3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v5.ani"},
	{0x0F7574A1FA9C7827ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v3.evt"},
	{0x0F75A6784ACA60E3ULL, "gen00fl_merrill6127294.cl"},
	{0x0F761C1032E2A019ULL, "618145372.ogg"},
	{0x0F7628C49AE634A9ULL, "vxdr_primevalentrance_34b_0_0.vlm"},
	{0x0F766027E380459DULL, "dae000_sandal6147803.cl"},
	{0x0F76A09423DE47A4ULL, "6082122_f.fxe"},
	{0x0F76D9229EE4C5A3ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.impa.evt"},
	{0x0F77654D5ABF4B1FULL, "909190760.ogg"},
	{0x0F779726320EF7E2ULL, "563947505.ogg"},
	{0x0F78010B5F448C5AULL, "6083703_f.fxe"},
	{0x0F78044CC728ADE4ULL, "6068877_m.fxe"},
	{0x0F783627BA9911B6ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.dura.evt"},
	{0x0F783F2942842440ULL, "tal_thre_clu.mao"},
	{0x0F78681895B45E1AULL, "vxw_sundermount_2ak_0_1.vlm"},
	{0x0F7896D087992208ULL, "art\\vfx\\combat\\vfx_m_pri_wintgrsp_imp_p.anb"},
	{0x0F78A30CC98F8233ULL, "359300357.ogg"},
	{0x0F7939AB821B5BCDULL, "114078512.ogg"},
	{0x0F7A799932B72A0BULL, "812927460.ogg"},
	{0x0F7AB9DF14026FF0ULL, "6027520_m.fxe"},
	{0x0F7AD3D4F7C97007ULL, "vxht_0_hightown_a2_c96_0_0.vlm"},
	{0x0F7AD922BA221E9FULL, "6053186_m.fxe"},
	{0x0F7B272DDC30466DULL, "6165036_m.fxe"},
	{0x0F7B281B50E409A7ULL, "827421384.ogg"},
	{0x0F7B2FCBFA88A3F8ULL, "755393826.ogg"},
	{0x0F7B8855F886D573ULL, "6084515_m.fxe"},
	{0x0F7B95A1FAA11C70ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v3.gad"},
	{0x0F7BA70BC275EE96ULL, "6071907_m.fxe"},
	{0x0F7BB4022CE4CE33ULL, "lt_0_lowtown_a2_n_c_45_0_lc.mmh"},
	{0x0F7BE31326B406A7ULL, "mag160_feynriel6020055.cl"},
	{0x0F7C114AA5BAFDBCULL, "l_dae110ar_lowtown_nt.bnk"},
	{0x0F7C13314C7FC223ULL, "6151028_m.fxe"},
	{0x0F7C1C3AC362953CULL, "mag291_ntbanter6116213.cl"},
	{0x0F7D07E085A90F35ULL, "811332124.ogg"},
	{0x0F7D37DAF4568244ULL, "gen00fl_fenris6106683.cl"},
	{0x0F7D3922E33FB6E0ULL, "gen00fl_isabela6090250.cl"},
	{0x0F7D785F1D9543C7ULL, "llac_lt_blackemporium_1055_14.dds"},
	{0x0F7DAAB2B689FF3BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_inst.ani"},
	{0x0F7DB2229EEAA224ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.impa.gad"},
	{0x0F7DBD83A70FD7B6ULL, "vxht_brothel_n_61z_0_0.vlm"},
	{0x0F7E34BEF3D19F0AULL, "6065270_m.fxe"},
	{0x0F7EBC4D907BE798ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinswall.phy"},
	{0x0F7F1C7E74257F3BULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.anb"},
	{0x0F7F2D27BA9F2141ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.dura.gad"},
	{0x0F7F6D9CD4A3CF69ULL, "leave_03.mao"},
	{0x0F7F875E47C44D01ULL, "vxht_0_hightown_b1q_0_0.vlm"},
	{0x0F7FAE0393ADC7A5ULL, "6107783_f.fxe"},
	{0x0F8025EF8271C632ULL, "6089571_m.fxe"},
	{0x0F80586F548E2B43ULL, "mag121_eau_banter6107384.cl"},
	{0x0F80828F7BC83FB6ULL, "vxw_encounter2_go_0_0.vlm"},
	{0x0F80B637DFCFEE4FULL, "vxw_bonepit_261_0_0.vlm"},
	{0x0F80C59F7D2F3E2EULL, "981778884.ogg"},
	{0x0F80DF326B72C590ULL, "vxht_0_hightown_n_bns_0_1.vlm"},
	{0x0F813FE07517560CULL, "196580357.ogg"},
	{0x0F81AFFC34A2D919ULL, "vxlt_gallowstemplar_432_0_0.vlm"},
	{0x0F81EA0BDDD50CF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_enter.gad"},
	{0x0F826EEEB86157BAULL, "275211974.ogg"},
	{0x0F82701C7D3F4F0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v5.evt"},
	{0x0F8293FD3ECB761AULL, "6127147_m.fxe"},
	{0x0F82AFEDE30AEE13ULL, "art\\vfx\\textures\\vfx_crater.dds"},
	{0x0F82EF32CD65DCC9ULL, "615382288.ogg"},
	{0x0F83119B3DE5503EULL, "vxw_bonepitdestroyed_a3_e1_0_0.vlm"},
	{0x0F83C299F72B1D42ULL, "mage_sleep.dds"},
	{0x0F83E58607F52D90ULL, "vxht_mansion_n_53v_0_0.vlm"},
	{0x0F84139EADA08C10ULL, "vxlt_hangedman_2o0_0_0.vlm"},
	{0x0F84B52A09BE4149ULL, "vxdr_primevaltomb_7bx_0_0.vlm"},
	{0x0F84C1229EF0DA02ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.impa.ani"},
	{0x0F850E8DD1E80B02ULL, "isa211_isabela6202148.cl"},
	{0x0F8534876F0CEF0EULL, "gen00fl_merrill6141240.cl"},
	{0x0F858B4A4C641D42ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_charge.gad"},
	{0x0F85A5AB1D8ABC89ULL, "vxlt_0_lowtown_n_h7j_0_0.vlm"},
	{0x0F85D9C323D15352ULL, "52710470.ogg"},
	{0x0F862E27BAA541EFULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.dura.ani"},
	{0x0F86808C74353095ULL, "876925927.ogg"},
	{0x0F8687BEB7D0D805ULL, "6204960_m.fxe"},
	{0x0F86917113B8907FULL, "fen210_orana6109775.cl"},
	{0x0F8699E5505B7AE1ULL, "vxlt_gallowsprison_n_64r_0_0.vlm"},
	{0x0F86CF7689AA68F7ULL, "w_bonepitmining_a2_70_0.msh"},
	{0x0F86E17689AA875AULL, "w_bonepitmining_a2_70_0.mao"},
	{0x0F86E57689AA8E69ULL, "w_bonepitmining_a2_70_0.mmh"},
	{0x0F871658FE860A0CULL, "vxw_bonepitmining_a2_gs_0_0.vlm"},
	{0x0F873DE271DEE066ULL, "6178360_f.fxe"},
	{0x0F8743844DFE3A86ULL, "275712774.ogg"},
	{0x0F876638E8979937ULL, "261532022.ogg"},
	{0x0F876C489CB35491ULL, "vxlt_0_lowtown_a3_n_c_gv9_0_0.vlm"},
	{0x0F884C7BC0519D76ULL, "6143057_m.fxe"},
	{0x0F88ED0BDDDB3029ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_enter.evt"},
	{0x0F88FD8A8EF7B905ULL, "character_character_p262.psh.fxo.dx11.assembly"},
	{0x0F89711C7D456FB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v5.gad"},
	{0x0F89ABD09A4EE570ULL, "llac_lt_0_lowtown_docks_a3_-65539_2.dds"},
	{0x0F8A012FAFAA88CFULL, "cd_mer221_sabin.wbd"},
	{0x0F8A04A88D2CA05FULL, "365404519.ogg"},
	{0x0F8A4D795DA1613FULL, "544871257.ogg"},
	{0x0F8A56A32051953DULL, "vxht_0_hightown_n_c7y_0_0.vlm"},
	{0x0F8A6E0A95A90534ULL, "170829027.ogg"},
	{0x0F8AB2B2B6949732ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_inst.evt"},
	{0x0F8AFE2A16C39A26ULL, "mrl310_merrill6144516.cl"},
	{0x0F8C5FA243271863ULL, "748056564.ogg"},
	{0x0F8CAE4A4C6A7739ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_charge.evt"},
	{0x0F8CAEA295799480ULL, "420918080.ogg"},
	{0x0F8D07AAC87E0081ULL, "gen_im_wep_rog_dua_act3_01.uti"},
	{0x0F8D7E9A9A735454ULL, "hlac_lt_gallowstemplar_2384_1x.dds"},
	{0x0F8E00AD15DC0EA9ULL, "6103252_m.fxe"},
	{0x0F8E01AD59C79258ULL, "6130080_m.fxe"},
	{0x0F8E2143C1E096E0ULL, "vxlt_0_lowtown_docks_a3_n_b11_0_0.vlm"},
	{0x0F8E6AB77932A703ULL, "211370078.ogg"},
	{0x0F8E84A115C373EBULL, "gen00fl_aveline6139296.cl"},
	{0x0F8EAD54646C5031ULL, "203133711.ogg"},
	{0x0F8EC36201256796ULL, "236539776.ogg"},
	{0x0F8EF0BCC5A4463BULL, "45555694.ogg"},
	{0x0F8F2814D1C606CFULL, "vxw_bonepit_210_0_0.vlm"},
	{0x0F8FA3883BB11F64ULL, "hlac_ca_cavern1_haunted_24421_2y.dds"},
	{0x0F8FE31493ABF436ULL, "llac_ht_chantry_n_2534_24.dds"},
	{0x0F90250EA572B695ULL, "gen_im_trs_cot_arc_lbw_r2.uti"},
	{0x0F90343B4AFDA533ULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.dura.ani"},
	{0x0F90E3237075C04AULL, "212092150.ogg"},
	{0x0F90FB83E3281C36ULL, "521503042.ogg"},
	{0x0F91D1B2B69AEAC5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_inst.gad"},
	{0x0F91E9E24A3A2CAFULL, "vxca_caverns2_7jl_0_0.vlm"},
	{0x0F920FD5457942A2ULL, "12758294.ogg"},
	{0x0F92857B9F96A490ULL, "vxlt_hangedman_2d1_0_0.vlm"},
	{0x0F92C2FD6E9BC207ULL, "vxw_bonepit_1vn_0_0.vlm"},
	{0x0F931B6CD6B7604BULL, "vxlt_0_lowtown_a2_n_erf_0_0.vlm"},
	{0x0F93277FE404AC73ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_018a_0d.dds"},
	{0x0F93CCA9FA26F075ULL, "m_dae211ar_gamlens_house_nt.bnk"},
	{0x0F9417880F80D509ULL, "vxlt_0_lowtown_n_h2i_0_0.vlm"},
	{0x0F94554879AF0CD7ULL, "6067985_m.fxe"},
	{0x0F94604EAAC98792ULL, "zz_mrl_debug6086558.cl"},
	{0x0F9510EFA2431C5DULL, "dae011_corff_sta6142475.cl"},
	{0x0F9549D60203F4ECULL, "vxlt_gallowstemplar_n_5jf_0_0.vlm"},
	{0x0F958835F07C228CULL, "vxw_bonepitmining_a2_bp_0_0.vlm"},
	{0x0F95A1E4592849E4ULL, "mag311_nyssa6042401.cl"},
	{0x0F9658645268FE1CULL, "gen00fl_isabela6101057.cl"},
	{0x0F967C9CD4B6C717ULL, "leave_03.dds"},
	{0x0F96A168DE31D3B0ULL, "711211728.ogg"},
	{0x0F96D50BDDE74498ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_enter.ani"},
	{0x0F9705F81E542A24ULL, "vxlt_0_lowtown_a2_gpf_0_0.vlm"},
	{0x0F97241E8EDC54A0ULL, "6066497_m.fxe"},
	{0x0F9727381064508EULL, "hlac_ht_0_hightown_a2_-131074_0x.dds"},
	{0x0F975F1DFB371987ULL, "zz_qun_debug6028259.cl"},
	{0x0F98539C80423BDBULL, "358227305.ogg"},
	{0x0F991C897F464BF3ULL, "6156858_m.fxe"},
	{0x0F99B64A4C750F28ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_charge.ani"},
	{0x0F99C27D07BD096AULL, "one280_iwan6096915.cl"},
	{0x0F99DBA10F339A49ULL, "6095933_m.fxe"},
	{0x0F9A7A73B6BA9BCAULL, "lt_0_lowtown_docks_n_109_0.mao"},
	{0x0F9A7E73B6BAA219ULL, "lt_0_lowtown_docks_n_109_0.mmh"},
	{0x0F9A8873B6BAB307ULL, "lt_0_lowtown_docks_n_109_0.msh"},
	{0x0F9B860D2FA55833ULL, "vxw_bonepit_1k5_0_0.vlm"},
	{0x0F9BCC507A25AB27ULL, "vxht_keep_6f2_0_0.vlm"},
	{0x0F9BD29F12721258ULL, "395625042.ogg"},
	{0x0F9BF0487AB1E195ULL, "vxht_brothel_n_6f8_0_0.vlm"},
	{0x0F9D19A97665D410ULL, "hlac_ht_mansion2_818_39z.dds"},
	{0x0F9D50C9B7A40931ULL, "1222583.ogg"},
	{0x0F9DA8028F15A934ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_deadmeredith_n.dds"},
	{0x0F9E17AE8BEFD507ULL, "vxw_woundedcoast_wc_0_0.vlm"},
	{0x0F9E1E6F2151DA8DULL, "6136106_m.fxe"},
	{0x0F9E2C3B4B09D56AULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.dura.evt"},
	{0x0F9E35BB7BDCF19EULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.cess.ani"},
	{0x0F9E41745EE29CB1ULL, "llac_lt_undercityencounter_-1_16.dds"},
	{0x0F9F5CE2862AF07BULL, "vxlt_gallowstemplar_f_4qe_0_0.vlm"},
	{0x0F9FF55521D1813EULL, "vxw_bonepitdestroyed_a3_k3_0_0.vlm"},
	{0x0FA095D087BA72A3ULL, "art\\vfx\\combat\\vfx_m_pri_wintgrsp_imp_p.mmh"},
	{0x0FA0C93FEBE15E90ULL, "vxht_mansion_n_59t_0_0.vlm"},
	{0x0FA18D49C8AD78CBULL, "vxlt_0_lowtown_a2_n_ewy_0_0.vlm"},
	{0x0FA1B463A5CC22E0ULL, "and111_lirene6084499.cl"},
	{0x0FA3425417F03C7AULL, "140754036.ogg"},
	{0x0FA36B1AB8844051ULL, "llac_ca_caverns1_24424_26.dds"},
	{0x0FA38522A66CE1D2ULL, "llac_w_woundedcoast_131080_0.dds"},
	{0x0FA3BBB2F3FA0D6CULL, "vxlt_gallowstemplar_n_5mg_0_0.vlm"},
	{0x0FA47BB4905F8320ULL, "w_encounter2_camp_41_0.phy"},
	{0x0FA4927B2CFE2468ULL, "vxlt_0_lowtown_a3_n_c_gjx_0_0.vlm"},
	{0x0FA50B3B4B0FBBDDULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.dura.gad"},
	{0x0FA537842D203112ULL, "603705575.ogg"},
	{0x0FA575C38AC51DC3ULL, "vxlt_hangedman_282_0_0.vlm"},
	{0x0FA57D454EEBC55CULL, "hlac_ht_chantry_2534_45x.dds"},
	{0x0FA5C08FDD5C1A7AULL, "632524987.ogg"},
	{0x0FA5C202D4221E52ULL, "246769183.ogg"},
	{0x0FA5C8BB113D289AULL, "hlac_lt_0_lowtown_docks_n_0_3z.dds"},
	{0x0FA5D50F35241395ULL, "lgt260_eau_tmp_conrad6143789.cl"},
	{0x0FA64BEE05149BF9ULL, "zz_followers_banter6158547.cl"},
	{0x0FA66B7B30F2A79CULL, "zz_ave_debug6059191.cl"},
	{0x0FA675A8840D5264ULL, "6164598_f.fxe"},
	{0x0FA6DB2654D4457CULL, "758325967.ogg"},
	{0x0FA73A5D043DC63DULL, "vxht_0_hightown_n_c9w_0_0.vlm"},
	{0x0FA75B039D5A805CULL, "6036535_m.fxe"},
	{0x0FA76E3B7CD5F063ULL, "6113128_f.fxe"},
	{0x0FA7D315F898C276ULL, "vxdr_primevaltomb_6s9_0_0.vlm"},
	{0x0FA7DA160B5656A1ULL, "vxht_brothel_n_6g7_0_0.vlm"},
	{0x0FA7EA5E99FD3F57ULL, "1025809327.ogg"},
	{0x0FA84F61BD37AD6AULL, "isa321_velasco6091861.cl"},
	{0x0FA879E4402CC610ULL, "730765068.ogg"},
	{0x0FA8DAB0B79A3F58ULL, "vxht_0_hightown_c5u_0_0.vlm"},
	{0x0FA9532C48E36A89ULL, "gen00fl_isabela6106415.cl"},
	{0x0FA958659DC61CCEULL, "6189666_m.fxe"},
	{0x0FA989C75C318F84ULL, "db_dot.mao"},
	{0x0FAA43BDD080AD69ULL, "llac_w_sundermount_65535_0.dds"},
	{0x0FAAA0BA6EA000D2ULL, "6200879_m.fxe"},
	{0x0FAB5D73FA08282CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_gesture_2.ani"},
	{0x0FAB8942B8F2F5C3ULL, "vxht_0_hightown_c3q_0_1.vlm"},
	{0x0FAB9AB49065D6D8ULL, "w_encounter2_camp_41_0.rml"},
	{0x0FAC4DBB7BE9586FULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.cess.evt"},
	{0x0FADADC21EE7F31BULL, "hlac_lt_0_lowtown_0_7y.dds"},
	{0x0FADBD73A23AAFB5ULL, "hlac_ca_caverns1_mines_24424_22y.dds"},
	{0x0FADCEBF782108FBULL, "vxlt_gallowstemplar_f_4vd_0_0.vlm"},
	{0x0FADE4AAEAADB06FULL, "hlac_dr_primevaltomb_659_12x.dds"},
	{0x0FAE50554A7AB44AULL, "17196619.ogg"},
	{0x0FAE7EC00E10B289ULL, "isa211ar_foundry.are"},
	{0x0FAEF31ACB6E111EULL, "6120407_m.fxe"},
	{0x0FAF336B0961B32FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4_tool.mao"},
	{0x0FAF3F6B0961C7CCULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4_tool.mmh"},
	{0x0FAF456B0961D1BEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4_tool.msh"},
	{0x0FAFA2E55FE9B645ULL, "406457087.ogg"},
	{0x0FAFADBB68F5860FULL, "208170311.ogg"},
	{0x0FAFFF26BAA3914BULL, "vxlt_0_lowtown_a2_n_exx_0_0.vlm"},
	{0x0FB091C6CD40F156ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.inactivetoactive.evt"},
	{0x0FB10CD7FE0B4E3DULL, "282013755.ogg"},
	{0x0FB10FE1498393EBULL, "214847128.ogg"},
	{0x0FB17D3BF397B117ULL, "rdr110dg_guard_warehouse6016696.cl"},
	{0x0FB18061578FD2AFULL, "955539371.ogg"},
	{0x0FB1C6C6ED4B6D07ULL, "dae000cr_shopper_f.utc"},
	{0x0FB1E586CBF10CF7ULL, "vxca_caverns2_8mj_0_0.vlm"},
	{0x0FB208A4CC1AA457ULL, "vxlt_darktown_52s_0_1.vlm"},
	{0x0FB21785248EE076ULL, "mer141cr_jansen.utc"},
	{0x0FB24EBB7BEDC518ULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.cess.gad"},
	{0x0FB26A73FA0E5D6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_gesture_2.gad"},
	{0x0FB284531DBE77BAULL, "zz_party.ncs"},
	{0x0FB29B1197AEC1E6ULL, "vxht_mansion_n_6k7_0_0.vlm"},
	{0x0FB31E3C98AE6514ULL, "vxlt_blackemporium_zm_0_0.vlm"},
	{0x0FB3E7A07CBB3643ULL, "vxlt_hangedman_255_0_0.vlm"},
	{0x0FB3FEAD40A5EEDEULL, "llac_lt_0_lowtown_a2_n_-3_0.dds"},
	{0x0FB4771A59EF86D2ULL, "54522659.ogg"},
	{0x0FB4CC606FCCA8DFULL, "vxlt_gallowstemplar_38v_0_0.vlm"},
	{0x0FB4CD268266E081ULL, "hlac_w_sundermount_131068_0y.dds"},
	{0x0FB5359EBC051C69ULL, "6041734_m.fxe"},
	{0x0FB53691F4B93E89ULL, "w_bonepitmining_a2_149_0.mao"},
	{0x0FB54291F4B95322ULL, "w_bonepitmining_a2_149_0.mmh"},
	{0x0FB54491F4B956D8ULL, "w_bonepitmining_a2_149_0.msh"},
	{0x0FB5890A1D59CD88ULL, "t3_wdg_w01.tnt"},
	{0x0FB6344CE68DDE72ULL, "6118358_m.fxe"},
	{0x0FB63C5F56FC8BBAULL, "6056856_m.fxe"},
	{0x0FB6AE3105AF17E4ULL, "6136741_m.fxe"},
	{0x0FB6D78363A924FAULL, "follower_barks6173416.cl"},
	{0x0FB732DDD9409B41ULL, "6020992_f.fxe"},
	{0x0FB77A808371BC14ULL, "one220_yevhen6103110.cl"},
	{0x0FB788C6CD4700E1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.inactivetoactive.gad"},
	{0x0FB8353D7E6288BBULL, "ip_restrict_style_dualwield.ncs"},
	{0x0FB86158813ED3F6ULL, "mag210_arianni6030493.cl"},
	{0x0FB8841C88C7FDABULL, "t3_grl_g03.tnt"},
	{0x0FB91E43D2B925E2ULL, "vxht_0_hightown_a2_bwd_0_0.vlm"},
	{0x0FB9C8C2744F545AULL, "dae000_female_tranquil6165136.cl"},
	{0x0FBB17A277C1EE87ULL, "hlac_lt_0_lowtown_docks_a3_-65538_6z.dds"},
	{0x0FBB7E4235D13EFDULL, "vxdr_0_deeproads_fin_0_0.vlm"},
	{0x0FBBA40A82D43E33ULL, "6101546_m.fxe"},
	{0x0FBBEC8F6EFBA2F9ULL, "6111668_m.fxe"},
	{0x0FBC1C53122B14D6ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.dura.ani"},
	{0x0FBC33A5C1A9C3B3ULL, "263571850.ogg"},
	{0x0FBC4839A58F4337ULL, "vxlt_0_lowtown_a2_n_c_ef8_0_0.vlm"},
	{0x0FBD0EFCE97AFC45ULL, "hlac_lt_0_lowtown_docks_n_-3_1y.dds"},
	{0x0FBD35446E590A29ULL, "mcr370_meredith6084413.cl"},
	{0x0FBD89C6CD4B6D8FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.inactivetoactive.ani"},
	{0x0FBD91A88EA4D13BULL, "6120603_f.fxe"},
	{0x0FBDD51CEA1AD20CULL, "6144527_f.fxe"},
	{0x0FBDE0C8D22E1EEDULL, "vxlt_0_lowtown_n_cmk_0_0.vlm"},
	{0x0FBE3DF26DEC860EULL, "65274583.ogg"},
	{0x0FBEA26C0A9DC437ULL, "dae200_saemus6142812.cl"},
	{0x0FBF6C2B0027F58AULL, "877543222.ogg"},
	{0x0FBF7A9B86B76CEAULL, "vxlt_hangedman_n_1r9_0_0.vlm"},
	{0x0FBFF2A9FA4C6BFFULL, "m_dae211ar_gamlens_house_nt.wbd"},
	{0x0FC0017A0678173FULL, "vxlt_gallowstemplar_n_5ji_0_0.vlm"},
	{0x0FC030C1660DF265ULL, "lt_0_lowtown_docks_n_21_0.rml"},
	{0x0FC0BAD706D365FAULL, "492974546.ogg"},
	{0x0FC10CEE89A4DA66ULL, "vxht_mansion_n_6n8_0_0.vlm"},
	{0x0FC13E77D4392952ULL, "llac_ht_playershouse_n_855_14.dds"},
	{0x0FC1B4EDB7E12929ULL, "708673468.ogg"},
	{0x0FC223D580365918ULL, "lt_0_lowtown_docks_12_0_ll.mmh"},
	{0x0FC2597D6EB14EC3ULL, "vxlt_hangedman_224_0_0.vlm"},
	{0x0FC25AA9C29128A8ULL, "llac_lt_0_lowtown_docks_131070_4.dds"},
	{0x0FC2758FCBDEE9FDULL, "hlac_lt_backalley_n_0_2x.dds"},
	{0x0FC2CB4E16CCDB52ULL, "mrl351_phantomelves6170344.cl"},
	{0x0FC321AB432653CCULL, "6094489_m.fxe"},
	{0x0FC3BFC5C0B0565AULL, "vxw_sundermount_24a_0_0.vlm"},
	{0x0FC3DBBF3DBF3C28ULL, "zz_prmpt_debug6198224.cl"},
	{0x0FC40D597D45323CULL, "vxlt_gallowstemplar_43k_0_0.vlm"},
	{0x0FC418F5EA625B6EULL, "vxht_0_hightown_n_a4m_0_0.vlm"},
	{0x0FC42489FB828C0EULL, "593822890.ogg"},
	{0x0FC43163B257A029ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02brgdebris.mao"},
	{0x0FC43D63B257B4C2ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02brgdebris.mmh"},
	{0x0FC43F63B257B878ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02brgdebris.msh"},
	{0x0FC4C10839629A85ULL, "6086365_f.fxe"},
	{0x0FC4EBBD6988D598ULL, "41336708.ogg"},
	{0x0FC50282DAE705BDULL, "da_defaultcc.dds"},
	{0x0FC50A7AEECDCAD2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medg_0n.dds"},
	{0x0FC526A7ED1A453DULL, "vxht_0_hightown_a2_c1l_0_0.vlm"},
	{0x0FC53F6F3068939FULL, "384000354.ogg"},
	{0x0FC5BE8A8812A504ULL, "vxlt_0_lowtown_a3_n_c_hwt_0_0.vlm"},
	{0x0FC5F1C4D4027770ULL, "871939010.ogg"},
	{0x0FC676884622B9D2ULL, "qun110_petrice6041358.cl"},
	{0x0FC6B00A923841BCULL, "dae000cr_grand_cleric6148161.cl"},
	{0x0FC70466459CF3ADULL, "follower_banter6162104.cl"},
	{0x0FC714EAF28B6741ULL, "dae000_bartrand6206859.cl"},
	{0x0FC79020C4AF3E62ULL, "vxht_0_hightown_a2_bxe_0_0.vlm"},
	{0x0FC7CAE233AB3707ULL, "752234.ogg"},
	{0x0FC84248CFA3A02CULL, "hlac_lt_0_lowtown_docks_a3_n_-3_0z.dds"},
	{0x0FC8AA4E48D7E9CBULL, "vxca_caverns1_mines_i1q_0_0.vlm"},
	{0x0FC924531235ACC7ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.dura.evt"},
	{0x0FC93197221A1EC1ULL, "ch2emnotnt.psh.fxo"},
	{0x0FC93815C054EEBAULL, "6117947_f.fxe"},
	{0x0FC998B4907ED0B5ULL, "w_encounter2_camp_41_0.msh"},
	{0x0FC99AB4907ED42BULL, "w_encounter2_camp_41_0.mmh"},
	{0x0FC9A6B4907EE8C8ULL, "w_encounter2_camp_41_0.mao"},
	{0x0FC9C7DEE435A8BEULL, "6166301_f.fxe"},
	{0x0FCA0F242FDC915AULL, "6019152_m.fxe"},
	{0x0FCA1CF2BC418C92ULL, "vxlt_0_lowtown_a2_n_c_gz7_0_0.vlm"},
	{0x0FCA466407E1CDC5ULL, "vxw_bonepitdestroyed_a3_18v_0_0.vlm"},
	{0x0FCACC648E5D09E4ULL, "vxw_encounter3_g6_0_0.vlm"},
	{0x0FCB1B2FAFE24BA5ULL, "cd_mer221_sabin.bnk"},
	{0x0FCB1E6AC6A623FEULL, "mag100_cullen6148681.cl"},
	{0x0FCB2B480096F5B5ULL, "nat380_sibling6207557.cl"},
	{0x0FCC185D426FA6CDULL, "gen_im_wep_mag_sta_act3_08.uti"},
	{0x0FCC6E7942455C28ULL, "zz_ave_debug6056756.cl"},
	{0x0FCC8A7136742407ULL, "vxw_bonepit_1jj_0_0.vlm"},
	{0x0FCCFDE0236E9CC4ULL, "vxlt_0_lowtown_docks_bg7_0_0.vlm"},
	{0x0FCD533BAEB1EA10ULL, "109137364.ogg"},
	{0x0FCD6E9639E11BE5ULL, "vxht_brothel_n_6fh_0_0.vlm"},
	{0x0FCE09E9DCFEE8A3ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc103a_0t.dds"},
	{0x0FCE5C82D65225E2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_ric_0s.dds"},
	{0x0FCEC27D1FA88A5BULL, "546733285.ogg"},
	{0x0FCEE2B257A151C9ULL, "253801431.ogg"},
	{0x0FCF99059F273CEBULL, "gen_im_trs_dem_sta_fir_r0.uti"},
	{0x0FD01993100A4013ULL, "hlac_lt_0_lowtown_n_0_0y.dds"},
	{0x0FD04553123C0310ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.dura.gad"},
	{0x0FD0D308EFD000E2ULL, "zz_followers_banter6157738.cl"},
	{0x0FD131874E759A78ULL, "tanner_deep_roads_v1.cut"},
	{0x0FD1ADC2A11F4F56ULL, "657757005.ogg"},
	{0x0FD1C1A8CA9B1FCEULL, "hlac_ht_keep_3732_5y.dds"},
	{0x0FD27F366F3B4ABCULL, "vxlt_gallowstemplar_46h_0_0.vlm"},
	{0x0FD2ACE27038D7AEULL, "9150258.ogg"},
	{0x0FD2C595302110E2ULL, "vxca_caverns2_8p4_0_0.vlm"},
	{0x0FD36FC00E2FB5DFULL, "isa211ar_foundry.ncs"},
	{0x0FD3B25F0C5E8856ULL, "fx_dust05_a.dds"},
	{0x0FD3CD141FB5B9D3ULL, "llac_lt_0_lowtown_a2_n_c_-1_5.dds"},
	{0x0FD3E5787C199601ULL, "428692525.ogg"},
	{0x0FD48875629CABB9ULL, "lgt210_eau_keran_father6144906.cl"},
	{0x0FD490402FFEF98FULL, "hlac_lt_0_lowtown_a2_131071_7x.dds"},
	{0x0FD4D8C029B35056ULL, "6122095_f.fxe"},
	{0x0FD4F454BF21F080ULL, "vxlt_warehouse_n_12o_0_0.vlm"},
	{0x0FD5620E96E61908ULL, "gen_im_npc_elf_arm_cht_lgta.uti"},
	{0x0FD5BC0160295849ULL, "6085741_m.fxe"},
	{0x0FD5F4D44CBBCB66ULL, "gen00fl_aveline6112704.cl"},
	{0x0FD6C9C4B581D74CULL, "and111_lirene6085100.cl"},
	{0x0FD6F58A938BDB03ULL, "428179558.ogg"},
	{0x0FD771483DFE1AFFULL, "6118855_f.fxe"},
	{0x0FD7D0EE73BD9E96ULL, "6087195_m.fxe"},
	{0x0FD7FE816E802E0FULL, "vxw_woundedcoast_3w_0_0.vlm"},
	{0x0FD83BC7339D8705ULL, "6034679_m.fxe"},
	{0x0FD8419A576B1464ULL, "art\\vfx\\combat\\vfx_w_2nd_whirl_c.mmh"},
	{0x0FD87060C71FA740ULL, "zz_followers_banter6158382.cl"},
	{0x0FD90E2173A3A24EULL, "vxlt_0_lowtown_a2_n_c_ff2_0_0.vlm"},
	{0x0FD9F26453C8A015ULL, "vxw_sundermount_277_0_0.vlm"},
	{0x0FDA6C0BAA8B2DF4ULL, "hlac_lt_0_lowtown_docks_n_-4_4z.dds"},
	{0x0FDA8FC00E360ACEULL, "isa211ar_foundry.lst"},
	{0x0FDA94E13A90DE48ULL, "192513964.ogg"},
	{0x0FDAA94D90CA262DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinswall.msh"},
	{0x0FDAB74D90CA3E60ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinswall.mao"},
	{0x0FDABB4D90CA44B3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinswall.mmh"},
	{0x0FDB06A016BC0559ULL, "hlac_ht_keep_4519_2y.dds"},
	{0x0FDBA8DE8D400D7AULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_mag_0d.dds"},
	{0x0FDBD84C1C6DA2C4ULL, "6075215_m.fxe"},
	{0x0FDC69E9504710BBULL, "hlac_lt_0_lowtown_a2_n_c_-65537_2x.dds"},
	{0x0FDC6F61398511F3ULL, "gen00fl_merrill6151105.cl"},
	{0x0FDC77E64CA6CAFEULL, "ave200_donnic6112145.cl"},
	{0x0FDCB8AAAA3E60D5ULL, "amulet_of_ashes.dds"},
	{0x0FDD061091A1A787ULL, "vxlt_undercityserialkiller_1j5_0_1.vlm"},
	{0x0FDD265B034BD6AEULL, "1009000110.ogg"},
	{0x0FDD81F27D69AEB5ULL, "vxlt_hangedman_4n_0_0.vlm"},
	{0x0FDD8944D79D08B7ULL, "433271960.ogg"},
	{0x0FDDB7755DC3C144ULL, "6034817_f.fxe"},
	{0x0FDDC30CD21BFF51ULL, "vxdr_0_deeproads_esk_0_0.vlm"},
	{0x0FDE4F2B981A4BB4ULL, "follower_banter6149214.cl"},
	{0x0FDE6D9546507890ULL, "vxw_encounter2_camp_rn_0_0.vlm"},
	{0x0FDE73D36E90AE94ULL, "vxlt_blackemporium_up_0_0.vlm"},
	{0x0FDE74E1E1C330DFULL, "314517347.ogg"},
	{0x0FDE858F89B5854CULL, "276485947.ogg"},
	{0x0FDE9082DD2C065CULL, "hlac_lt_0_lowtown_docks_a3_-2_5y.dds"},
	{0x0FDED4DC89BE48E0ULL, "hlac_lt_0_lowtown_docks_n_65533_1x.dds"},
	{0x0FDF5F5D1F86E0BBULL, "vxlt_gallowstemplar_5uo_0_0.vlm"},
	{0x0FDF85710E4964D4ULL, "llac_ca_caverns2_5802_54.dds"},
	{0x0FE02B0FE737DA0BULL, "vxdr_primevaltomb_3gq_0_0.vlm"},
	{0x0FE0FD0D1A0DC094ULL, "zz_dae_debug6142967.cl"},
	{0x0FE118F0B35BCBAEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtopoc.mmh"},
	{0x0FE124F0B35BE045ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtopoc.mao"},
	{0x0FE132F0B35BF79CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtopoc.msh"},
	{0x0FE17187D076861BULL, "522738497.ogg"},
	{0x0FE1E13054EC0891ULL, "gen00fl_anders6201084.cl"},
	{0x0FE1E5CC717A1AB4ULL, "art\\vfx\\combat\\vfx_c_pride_cast_f_c.mmh"},
	{0x0FE24EEC27CFE77CULL, "gen00fl_varric6076229.cl"},
	{0x0FE3B1AB83E94622ULL, "pro000ip_corpse_weapons3.utp"},
	{0x0FE3C50216382296ULL, "hlac_lt_0_lowtown_docks_-1_2z.dds"},
	{0x0FE419F451E0FA14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_2h_sw.ani"},
	{0x0FE4448CC60300D4ULL, "399293560.ogg"},
	{0x0FE4471FAFE75568ULL, "6102409_m.fxe"},
	{0x0FE504C2419869B5ULL, "6042267_m.fxe"},
	{0x0FE528693990B824ULL, "6070528_m.fxe"},
	{0x0FE563143A08D865ULL, "character_character_p28.psh.fxo"},
	{0x0FE5817C1B5F3E50ULL, "hlac_lt_gallowsprison_n_-1_7z.dds"},
	{0x0FE5CFD4A19997F3ULL, "6045798_m.fxe"},
	{0x0FE5EA1A9DA987DAULL, "151553231.ogg"},
	{0x0FE6949DD3EF5B76ULL, "vrc221_varric6043508.cl"},
	{0x0FE73C4FDBB561F2ULL, "hlac_w_encounter2_131070_0z.dds"},
	{0x0FE740B717C63315ULL, "648229699.ogg"},
	{0x0FE7EA2848777285ULL, "57915477.ogg"},
	{0x0FE824D2A2E21467ULL, "vxw_bonepit_243_0_0.vlm"},
	{0x0FE84E5F9EE0EB3FULL, "gen00fl_merrill6107454.cl"},
	{0x0FE859445ECF2B6AULL, "564301498.ogg"},
	{0x0FE8644145BEB895ULL, "vxw_sundermount_286_0_0.vlm"},
	{0x0FE888D69181F3AAULL, "llac_lt_blackemporium_1055_0.dds"},
	{0x0FE9A756E764E4D2ULL, "160353541.ogg"},
	{0x0FE9FC4A616A3BDDULL, "llac_ht_0_hightown_-1_0.dds"},
	{0x0FEAC7CD1FED372DULL, "hlac_ca_cavern1_haunted_24425_4z.dds"},
	{0x0FEADDA85673BE0FULL, "6035905_f.fxe"},
	{0x0FEAECED1D593A37ULL, "18362339.ogg"},
	{0x0FEB16F451E71476ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_2h_sw.gad"},
	{0x0FEB1EB533C726CBULL, "1008875558.ogg"},
	{0x0FEB9E3824564761ULL, "mage_crushingprison.dds"},
	{0x0FEBA02B0754E20AULL, "hlac_ht_mansion_n_3373_0x.dds"},
	{0x0FEBD228231C8B12ULL, "6018902_m.fxe"},
	{0x0FEBFF185F0F700BULL, "vxht_brothel_n_4gs_0_0.vlm"},
	{0x0FEC2D0018B81034ULL, "6041512_f.fxe"},
	{0x0FEC3F224CF15098ULL, "vxlt_0_lowtown_docks_n_cuo_0_0.vlm"},
	{0x0FEC5A48A426C704ULL, "vxlt_undercitydungeon_1ga_0_0.vlm"},
	{0x0FECDF7238469110ULL, "vxw_encounter2_camp_mo_0_0.vlm"},
	{0x0FED13DF19C4080AULL, "444961707.ogg"},
	{0x0FED2AFFDBF0DC98ULL, "vxht_mansion2b_1kq_0_0.vlm"},
	{0x0FEE4058A208C7A7ULL, "vxht_0_hightown_n_9bb_0_0.vlm"},
	{0x0FEED48E901AECEFULL, "930376762.ogg"},
	{0x0FEF9FF967B17562ULL, "vxht_0_hightown_bu2_0_0.vlm"},
	{0x0FEFF52E13487370ULL, "6174875_m.fxe"},
	{0x0FF03396774763C2ULL, "938662822.ogg"},
	{0x0FF18D3376F16541ULL, "6134764_m.fxe"},
	{0x0FF1C4C2475711F4ULL, "894926718.ogg"},
	{0x0FF1CAECF73E0CC5ULL, "782868473.ogg"},
	{0x0FF1F2E99966FC61ULL, "211654374.ogg"},
	{0x0FF231F451ED60D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\charge_2h_sw.evt"},
	{0x0FF2F723BA0770E5ULL, "gencr_circle_mage_f.utc"},
	{0x0FF2FFA79F39D953ULL, "vxw_blightlands_a7_0_0.vlm"},
	{0x0FF325A0B907CE5DULL, "969669369.ogg"},
	{0x0FF32FC16639498DULL, "lt_0_lowtown_docks_n_21_0.anb"},
	{0x0FF42258C7F38E92ULL, "78170993.ogg"},
	{0x0FF4421DB82F9D30ULL, "art\\vfx\\models\\vfx_slash.mmh"},
	{0x0FF446EE3E324B7BULL, "isa321cr_velasco.utc"},
	{0x0FF4501DB82FB54AULL, "art\\vfx\\models\\vfx_slash.msh"},
	{0x0FF45226C6EAE9F7ULL, "vxlt_gallowstemplar_5oy_0_0.vlm"},
	{0x0FF49FCCD6679B0DULL, "6143151_m.fxe"},
	{0x0FF4AB8C5AAB66B6ULL, "al_warrior_t06.gda"},
	{0x0FF4E4B7483B5F95ULL, "189472556.ogg"},
	{0x0FF4EE831D0448EAULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_c.anb"},
	{0x0FF50193C5D80F39ULL, "742246714.ogg"},
	{0x0FF523040DDF5BC1ULL, "6098259_m.fxe"},
	{0x0FF58D8FDE0B580BULL, "201251057.ogg"},
	{0x0FF5F2C96022C146ULL, "55696102.ogg"},
	{0x0FF6FAC85435ABAAULL, "hlac_dr_primevaltomb_658_7z.dds"},
	{0x0FF70EEE3C9E6257ULL, "mcr310_argument6074431.cl"},
	{0x0FF757341D322920ULL, "237833891.ogg"},
	{0x0FF7787A54635B8AULL, "dae000_arishok6067740.cl"},
	{0x0FF7E70228892320ULL, "narrative_varric_demo6170819.cl"},
	{0x0FF8646AAB02D796ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_kick_feet_shrug.ani"},
	{0x0FF8B221F9EA182BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent.ani"},
	{0x0FF929141188F2D3ULL, "gen00fl_aveline6197230.cl"},
	{0x0FF9672B33D6011DULL, "hlac_w_bonepit_0_0y.dds"},
	{0x0FFA49DA167C9DE0ULL, "204915911.ogg"},
	{0x0FFA4D7093B6857AULL, "vxlt_gallowscourtyard_7bs_0_0.vlm"},
	{0x0FFA6DE9F7298E65ULL, "435845770.ogg"},
	{0x0FFABFBC235E00AAULL, "764311949.ogg"},
	{0x0FFADCCB55EA6007ULL, "spt_fronds.psh.fxo"},
	{0x0FFB4BF06B6C5CC3ULL, "art\\characters\\creatures\\dragon\\animation\\cwkb.evt"},
	{0x0FFB514F2A3CA990ULL, "vxw_encounter2_camp_hl_0_0.vlm"},
	{0x0FFB9DB579C8B2DDULL, "433171578.ogg"},
	{0x0FFBA4BE72465D2DULL, "622642780.ogg"},
	{0x0FFBE77EC782370AULL, "vxdr_primevalentrance_26v_0_0.vlm"},
	{0x0FFC119D4C02C3DAULL, "vxw_encounter2_55_0_0.vlm"},
	{0x0FFC1F91137FB2A9ULL, "396437647.ogg"},
	{0x0FFC805D608E77D2ULL, "vxw_blightlands_a2_0_0.vlm"},
	{0x0FFC841FB5AA4118ULL, "ico_crossbow_3.dds"},
	{0x0FFCEE1CC507754BULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v3.evt"},
	{0x0FFD5A258A588E99ULL, "6219135_m.fxe"},
	{0x0FFDE84D3970FFB5ULL, "vxlt_0_lowtown_n_ews_0_0.vlm"},
	{0x0FFEF001D12CD308ULL, "fen000im_gft_book.uti"},
	{0x0FFF85FE4FEB0704ULL, "vxlt_0_lowtown_a3_n_c_hcx_0_0.vlm"},
	{0x0FFF98578D686C2FULL, "602272725.ogg"},
	{0x0FFF9EC41FAA6928ULL, "vxlt_darktown_6oa_0_0.vlm"},
	{0x0FFFBE4828B6B436ULL, "vxht_brothel_n_6bc_0_0.vlm"},
	{0x1000D9D2D29C7F6CULL, "zz_vis_debug6050354.cl"},
	{0x1000F9054B7EDEFBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblepile.msh"},
	{0x10010B054B7EFD46ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblepile.mao"},
	{0x100117054B7F11E5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblepile.mmh"},
	{0x10011D0FEDC10134ULL, "vxw_sundermount_2do_0_0.vlm"},
	{0x1001449A578E1E5BULL, "art\\vfx\\combat\\vfx_w_2nd_whirl_c.anb"},
	{0x10017184912FF1D3ULL, "vxw_blightlands_j6_0_0.vlm"},
	{0x1001DC436B08AEFEULL, "abi_m_fireball.bnk"},
	{0x100202B034119D04ULL, "445069167.ogg"},
	{0x100224F06B723944ULL, "art\\characters\\creatures\\dragon\\animation\\cwkb.gad"},
	{0x1002A9BF53D4D2C0ULL, "897744690.ogg"},
	{0x1002C41B5BD6F522ULL, "llac_ht_0_hightown_n_-65536_0.dds"},
	{0x100310B3A1F92070ULL, "hlac_lt_0_lowtown_a2_n_65535_8x.dds"},
	{0x1004071CC50DBE0CULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v3.gad"},
	{0x1005448B23390B2CULL, "ts_kirk_hightown_prog_junk1.gda"},
	{0x1005AA965E7A7EC2ULL, "vxw_encounter2_camp_qq_0_0.vlm"},
	{0x1005FF67A57E688FULL, "hlac_ca_caverns2prc_5802_34z.dds"},
	{0x1006205F0D69AC2BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v1.evt"},
	{0x10066C6AAB0F2287ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_kick_feet_shrug.evt"},
	{0x10069A21F9F62CA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent.evt"},
	{0x10071C4A7704599EULL, "665595950.ogg"},
	{0x1007D47A7BCA4C42ULL, "hlac_ht_keep_3731_3z.dds"},
	{0x1007F0060CD62969ULL, "vxca_caverns1_jgo_0_0.vlm"},
	{0x100833F06B76BE22ULL, "art\\characters\\creatures\\dragon\\animation\\cwkb.ani"},
	{0x100840A62127AC5FULL, "lgt110_gxa_coa_meeting6117497.cl"},
	{0x100860E2509D10DFULL, "hlac_lt_0_lowtown_n_1_4z.dds"},
	{0x1008B484BD493886ULL, "hlac_lt_0_lowtown_docks_n_65533_0y.dds"},
	{0x1008BF4D85AC9DFAULL, "vxlt_gallowscourtyard_7gl_0_0.vlm"},
	{0x10090FB0F0ECFE72ULL, "6165004_m.fxe"},
	{0x100918A3A7FE48D1ULL, "vxdr_0_deeproads_ebh_0_0.vlm"},
	{0x1009AAA5DE3AC6D9ULL, "nosynccombatanims.gda"},
	{0x100A061CC512280AULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v3.ani"},
	{0x100A17162103CC30ULL, "mag211_marethari6021347.cl"},
	{0x100AE8CC719D256BULL, "art\\vfx\\combat\\vfx_c_pride_cast_f_c.anb"},
	{0x100AF23A52849052ULL, "vxw_blightlands_j3_0_0.vlm"},
	{0x100AF9643E348873ULL, "llac_w_bonepit_327681_0.dds"},
	{0x100B04E4A6FB7A57ULL, "ave200_aveline_bar6116444.cl"},
	{0x100B60B2C88C1B44ULL, "6163765_m.fxe"},
	{0x100B63C90F830CFFULL, "6019126_m.fxe"},
	{0x100B70B212F7DA65ULL, "vxw_sundermount_27g_0_0.vlm"},
	{0x100B9AD0C8B18382ULL, "dae000_bartrand6207004.cl"},
	{0x100BE45F2CD8DD1BULL, "mcr310_sibling6076379.cl"},
	{0x100D395F0D6FF5ECULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v1.gad"},
	{0x100D8D6AAB1579D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_kick_feet_shrug.gad"},
	{0x100D98C82146DB95ULL, "hlac_w_encounter2_camp_-131072_0z.dds"},
	{0x100D9921F9FC49B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent.gad"},
	{0x100DCD5AE5BF0161ULL, "hlac_lt_0_lowtown_a2_n_c_131072_2y.dds"},
	{0x100E83361A8A0FAAULL, "llac_lt_0_lowtown_docks_-1_2.dds"},
	{0x100EAF4407BB94B2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wrba_0d.dds"},
	{0x100EC55890AF22FBULL, "6177891_m.fxe"},
	{0x100F7BF70F7FB09DULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_onyxstatuette_sa.dds"},
	{0x100F8EECDFB719B4ULL, "vxw_sundermount_2il_0_0.vlm"},
	{0x100FE36183260A53ULL, "vxw_blightlands_o9_0_0.vlm"},
	{0x100FF566249DD24FULL, "252866714.ogg"},
	{0x1010267C93C37818ULL, "mag110_vincento6017675.cl"},
	{0x1010BFC672B76D04ULL, "122308552.ogg"},
	{0x1011488D19A96936ULL, "vxw_blightlands_9n_0_0.vlm"},
	{0x1011757BD6DFDBFAULL, "858950708.ogg"},
	{0x1011906F6A77BDA9ULL, "vxht_keep_n_63x_0_0.vlm"},
	{0x1011E0BB38D669C5ULL, "6128945_m.fxe"},
	{0x1011FE22D090EDFAULL, "96274457.ogg"},
	{0x101224E9AA89298CULL, "dae000_mother6004110.cl"},
	{0x101282EEFE9B8CD9ULL, "zz_followers_banter6159711.cl"},
	{0x1012D8681C4D8E33ULL, "6127511_m.fxe"},
	{0x101344139796156BULL, "462500170.ogg"},
	{0x1013828663EE41DBULL, "465245935.ogg"},
	{0x10141C7350709742ULL, "vxw_encounter2_camp_lr_0_0.vlm"},
	{0x10142957D7A59A05ULL, "6016055_m.fxe"},
	{0x1014385F0D7612EAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v1.ani"},
	{0x1014E93B43ABDD39ULL, "hlac_lt_0_lowtown_n_0_1x.dds"},
	{0x1015044A2460FEC4ULL, "388691627.ogg"},
	{0x1016669B372D81DAULL, "6129912_m.fxe"},
	{0x1016A067B02915CEULL, "vxlt_0_lowtown_docks_a3_coc_0_0.vlm"},
	{0x1016CBD46EBF848DULL, "vxlt_darktown_2qv_0_0.vlm"},
	{0x1017256D61D1764CULL, "6108519_m.fxe"},
	{0x1017312A77A2B67AULL, "vxlt_gallowscourtyard_7hm_0_0.vlm"},
	{0x101754AF34F1B98BULL, "vxht_brothel_n_4vt_0_0.vlm"},
	{0x1017C549BABA13D4ULL, "fen210_orana6174969.cl"},
	{0x1017DE6C3A4504E0ULL, "873881610.ogg"},
	{0x10180FAA444DA6EDULL, "6053569_f.fxe"},
	{0x101847BEE321E1EDULL, "6024563_f.fxe"},
	{0x101862F4442DF2C1ULL, "gen_im_arm_glv_med_start.uti"},
	{0x1019129EE9AADC7BULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_3y.dds"},
	{0x1019E28F04EDF2E5ULL, "vxw_sundermount_28f_0_0.vlm"},
	{0x101A63DADFB91D56ULL, "696004131.ogg"},
	{0x101A8592450855D7ULL, "134964688.ogg"},
	{0x101A9641F85C52B8ULL, "221381274.ogg"},
	{0x101AD2FB0F334B6AULL, "449371631.ogg"},
	{0x101AF6819E8B5D1AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodburn_d.dds"},
	{0x101B12913943FC8FULL, "vxlt_gamlenshouse_n_n5_0_0.vlm"},
	{0x101B46F82C2ED164ULL, "cod_bks_first_children.utp"},
	{0x101B7D4552204317ULL, "vxlt_house_fy_0_0.vlm"},
	{0x101BFF7979003090ULL, "w_woundedcoast_228_0.mao"},
	{0x101C0379790037A3ULL, "w_woundedcoast_228_0.mmh"},
	{0x101C117979004FFDULL, "w_woundedcoast_228_0.msh"},
	{0x101C391DB78C061CULL, "vxca_caverns1_mines_cny_0_0.vlm"},
	{0x101C8B126D13A893ULL, "llac_lt_0_lowtown_docks_n_-3_8.dds"},
	{0x101CC6A51D333863ULL, "hlac_lt_0_lowtown_docks_n_-3_0x.dds"},
	{0x101D08539B62B71EULL, "350141820.ogg"},
	{0x101DF1831D2752DDULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_c.mmh"},
	{0x101E00C9D1AD3234ULL, "vxw_sundermount_2rm_0_0.vlm"},
	{0x101E021345D9E010ULL, "gen_im_trs_rdr_arc_lbw_r1.uti"},
	{0x101E09B5189DB2D6ULL, "6070240_f.fxe"},
	{0x101ECE43047B2220ULL, "vxlt_gallowstemplar_3rs_0_0.vlm"},
	{0x101EDC425D711647ULL, "405573371.ogg"},
	{0x101EDC830877EBF9ULL, "lt_0_lowtown_docks_n_110_0.msh"},
	{0x101EE6830877FCE7ULL, "lt_0_lowtown_docks_n_110_0.mmh"},
	{0x101EEA8308780334ULL, "lt_0_lowtown_docks_n_110_0.mao"},
	{0x101F2D72024CE77AULL, "vxdr_primevaltomb_81y_0_0.vlm"},
	{0x101F6A18B500EC44ULL, "6050183_f.fxe"},
	{0x101F81656363788FULL, "6072324_f.fxe"},
	{0x101FAB15F3647D74ULL, "hlac_lt_0_lowtown_n_196607_1z.dds"},
	{0x102015A8604A1C98ULL, "vxlt_0_lowtown_a2_n_gzc_0_0.vlm"},
	{0x10202E4B7106C506ULL, "426564933.ogg"},
	{0x10209EC2AB5D4381ULL, "612201241.ogg"},
	{0x1021476E862E4A02ULL, "ser221ip_note_shipment.utp"},
	{0x10216DEE34D6266AULL, "396991481.ogg"},
	{0x10217B1137170BCCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_rdy.gad"},
	{0x1021E6E6B1826E96ULL, "64280875.ogg"},
	{0x102204B716D32D83ULL, "vxw_encounter2_y_0_0.vlm"},
	{0x102281D62B98BDACULL, "hlac_w_woundedcoast_6_0y.dds"},
	{0x10228E504266AFC2ULL, "vxw_encounter2_camp_gs_0_0.vlm"},
	{0x1022EAA651F417FAULL, "vxlt_0_lowtown_docks_bak_0_0.vlm"},
	{0x1022F8B038DE1FE1ULL, "6086573_m.fxe"},
	{0x102310500D3E2E83ULL, "vxca_caverns2prc_8h1_0_1.vlm"},
	{0x10237F798D8A2094ULL, "6125330_m.fxe"},
	{0x102380BD0BD1562DULL, "vxw_bonepitmining_a2_b5_0_0.vlm"},
	{0x1023A6C25EF37C00ULL, "vxht_0_hightown_a2_9zn_0_0.vlm"},
	{0x1023E7BEF0C81E10ULL, "vxdr_slavers_2hf_0_0.vlm"},
	{0x1024521A6FD1B009ULL, "vxlt_0_lowtown_e1v_0_0.vlm"},
	{0x10247302F33552DBULL, "vxlt_0_lowtown_n_etx_0_0.vlm"},
	{0x102488A0A65C237DULL, "vxlt_0_lowtown_a2_db6_0_0.vlm"},
	{0x10253DB160B59D0DULL, "vxlt_darktown_2tu_0_0.vlm"},
	{0x10254514982B14C5ULL, "vxlt_gallowscourtyard_7bx_0_0.vlm"},
	{0x10256FC487D7544DULL, "796202674.ogg"},
	{0x102583DEDE4CF9F2ULL, "438428311.ogg"},
	{0x10270E233F071519ULL, "vxlt_0_lowtown_a2_n_c_eox_0_1.vlm"},
	{0x102750AD91362984ULL, "mcr371_hostile_fenris6082213.cl"},
	{0x10278E225329B47CULL, "6087531_m.fxe"},
	{0x1027C28F5B63CCE0ULL, "vxht_mansion2_1we_0_0.vlm"},
	{0x102829EDBD400F6BULL, "follower_banter6136363.cl"},
	{0x10286B1E1298EFD0ULL, "6127685_f.fxe"},
	{0x10286C1E4EC22319ULL, "w_encounter3_204_0.msh"},
	{0x1028761E4EC23407ULL, "w_encounter3_204_0.mmh"},
	{0x10287A11371D28CAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_rdy.ani"},
	{0x10287A1E4EC23B54ULL, "w_encounter3_204_0.mao"},
	{0x1028D388543E5F63ULL, "6081692_f.fxe"},
	{0x1028F1412AA104A5ULL, "vxlt_0_lowtown_a3_n_c_e38_0_0.vlm"},
	{0x10293600881D06EEULL, "vxht_0_hightown_n_akd_0_0.vlm"},
	{0x1029AF6EE0895262ULL, "vxlt_0_lowtown_a2_hgb_0_0.vlm"},
	{0x1029BCE6E0554F88ULL, "dae000ip_crowd_ff5_talk17.utp"},
	{0x102A084E22724503ULL, "vxlt_hangedman_n_1xr_0_0.vlm"},
	{0x102A2F6D733F206BULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_crown_n.dds"},
	{0x102A7C246A92D5E1ULL, "6171510_m.fxe"},
	{0x102B13DEFE98FDB6ULL, "vxht_brothel_n_6uf_0_0.vlm"},
	{0x102B15196149E8E1ULL, "zz_isa_debug6133987.cl"},
	{0x102B1844357CC1F4ULL, "6179511_m.fxe"},
	{0x102B4506055C650BULL, "dae000ip_crowd_ff4_talk4.utp"},
	{0x102BFCE9CDBA6A8EULL, "106942336.ogg"},
	{0x102CD4BF4C98F7F2ULL, "vxht_playershouse_n_1qc_0_0.vlm"},
	{0x102D0479790EACC8ULL, "w_woundedcoast_228_0.phy"},
	{0x102D04E88FBAF623ULL, "vxht_keep_70e_0_0.vlm"},
	{0x102D35D640AE3BDCULL, "vxlt_shop_cl_0_1.vlm"},
	{0x102D8D9606EC7EE3ULL, "334576850.ogg"},
	{0x102E124C12F77EBCULL, "40744779.ogg"},
	{0x102E622A664EA710ULL, "llac_ht_0_hightown_n_2_0.dds"},
	{0x102E72EBEC5F01EEULL, "llac_dr_0_deeproads_5845_11.dds"},
	{0x102EC88C2AFF00B4ULL, "754198095.ogg"},
	{0x102EE289E19D116EULL, "llac_lt_backalley_g_0_8.dds"},
	{0x102F133E4F01B633ULL, "6067674_f.fxe"},
	{0x102F2A21D3BE2D2FULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb102a_0s.dds"},
	{0x102F4675E2C3DB48ULL, "hlac_ht_mansion2_818_64x.dds"},
	{0x102F9C2D89EC1F88ULL, "zz_mer_debug6039335.cl"},
	{0x10303AA75E70A550ULL, "llac_ca_cavern1_haunted_24424_26.dds"},
	{0x10304B3BCBBAAC85ULL, "mag320st_orsino.stg"},
	{0x1030654D435B692CULL, "art\\characters\\playercharacter\\humanfemale\\hf_arm_flma.mao"},
	{0x10307AAF451BD6E9ULL, "445873987.ogg"},
	{0x103092552617F184ULL, "221181258.ogg"},
	{0x103103BD7464F99AULL, "llac_w_blightlands_393213_0.dds"},
	{0x10314D5F80EE26EEULL, "vxht_keep_n_55r_0_0.vlm"},
	{0x10315757B5D72F3BULL, "6172676_m.fxe"},
	{0x10315C8343EA307AULL, "vxlt_0_lowtown_docks_bdd_0_0.vlm"},
	{0x1031DA87A7D9A1ECULL, "390229168.ogg"},
	{0x1031F299FDC76EADULL, "vxw_bonepitmining_a2_i4_0_0.vlm"},
	{0x10324BAE8E445220ULL, "gen00fl_varric6104708.cl"},
	{0x1032763202DBD39BULL, "art\\vfx\\cinematic\\vfx_chantry_int_exp_p.impa.ani"},
	{0x1032844B7C838693ULL, "vxlt_backalley_n_1u8_0_0.vlm"},
	{0x1032B27CE82ADA27ULL, "343238617.ogg"},
	{0x1032F9FD00D15AFAULL, "vxht_0_hightown_n_cb1_0_0.vlm"},
	{0x10334212B4096AEDULL, "zz_mag_debug6039788.cl"},
	{0x1033B6F18A212D45ULL, "vxlt_gallowscourtyard_7gg_0_0.vlm"},
	{0x1034B4C7C996CD56ULL, "w_bonepitmining_a2_195_0.mao"},
	{0x1034C0C7C996E1F5ULL, "w_bonepitmining_a2_195_0.mmh"},
	{0x1034C2C7C996E56BULL, "w_bonepitmining_a2_195_0.msh"},
	{0x1035934C4E1E154FULL, "km_pryce.mor"},
	{0x10362C4C30339894ULL, "mrl310_party_comments6142848.cl"},
	{0x10366D21C6A7CC19ULL, "6093253_f.fxe"},
	{0x1036E850EBD61C9BULL, "149281381.ogg"},
	{0x10371A95EB80A01AULL, "141009153.ogg"},
	{0x1037B6846693F6A9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cot.phy"},
	{0x1037B741F192535AULL, "art\\levels\\races\\design\\animation\\beartrap.inactivetoactive.gad"},
	{0x1038247DDBAB127AULL, "vxht_keep_n_4ao_0_0.vlm"},
	{0x103851BE5EDB5079ULL, "vxlt_backalley_g_1kq_0_0.vlm"},
	{0x10387A2B14685D83ULL, "vxlt_hangedman_n_1uu_0_0.vlm"},
	{0x1038C60E21F5C0E7ULL, "6089668_m.fxe"},
	{0x1038E72CA0265652ULL, "584914368.ogg"},
	{0x1038ED638798A269ULL, "vxw_encounter2_camp_js_0_0.vlm"},
	{0x103921084CD1FE7EULL, "951668140.ogg"},
	{0x1039541E8D8FC4E2ULL, "91101451.ogg"},
	{0x103AE483B5996334ULL, "vxw_sundermount_2xs_0_0.vlm"},
	{0x103B01701FF06CFAULL, "367044552.ogg"},
	{0x103B467E26A997D0ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfireout.mmh"},
	{0x103B469C3E8F1072ULL, "vxht_playershouse_n_1tl_0_0.vlm"},
	{0x103B544C602B6984ULL, "dae000cr_grand_cleric6185504.cl"},
	{0x103BB578A7E7B45CULL, "vxlt_warehouse_n_123_0_0.vlm"},
	{0x103BDFDE32AD2A12ULL, "gen00fl_fenris6126874.cl"},
	{0x103BF0D5E8B2D8B4ULL, "genip_exploding_crate_lg.utp"},
	{0x103C8E0F57E7BC36ULL, "art\\levels\\ztools\\helpers\\texels\\_textures\\128m.dds"},
	{0x103D072817BFB4ACULL, "vxht_mansion2b_78_0_0.vlm"},
	{0x103D7F05AF725117ULL, "6130137_m.fxe"},
	{0x103DB3A51B7B2227ULL, "plt_cod_prc_brotherjoffrey.plo"},
	{0x103E219779E222DDULL, "vxlt_hangedman_n_27v_0_0.vlm"},
	{0x103E34A878E977F2ULL, "isa211_isabela6070110.cl"},
	{0x103E4F068E5C4A97ULL, "349068568.ogg"},
	{0x103E68E0EA1B38D7ULL, "vrc221_varric6044050.cl"},
	{0x103EACEEA59B9C14ULL, "gen00fl_isabela6096821.cl"},
	{0x103EBA41F1987771ULL, "art\\levels\\races\\design\\animation\\beartrap.inactivetoactive.evt"},
	{0x103F7E3202E66B92ULL, "art\\vfx\\cinematic\\vfx_chantry_int_exp_p.impa.evt"},
	{0x104002E6BF9181B6ULL, "vxca_caverns2_4x5_0_1.vlm"},
	{0x1040A24FFD6EF525ULL, "6172256_m.fxe"},
	{0x1040AAED9957D5D8ULL, "6015184_f.fxe"},
	{0x104189C9BE91607EULL, "vxca_caverns1_mines_gkh_0_0.vlm"},
	{0x1041DA697FB11564ULL, "mag160_varian_ilithis6171471.cl"},
	{0x1042221DA0F9330EULL, "6097468_m.fxe"},
	{0x10423C8A4CEDDB98ULL, "gen00fl_aveline6053337.cl"},
	{0x10428C60A8CC1994ULL, "ser121_jethann6104549.cl"},
	{0x10428D620657C06DULL, "vxca_caverns1_mines_icl_0_1.vlm"},
	{0x1042A31893371E11ULL, "26299951.ogg"},
	{0x1042AE436B3FF704ULL, "abi_m_fireball.wbd"},
	{0x1042B3063B168DC6ULL, "vxlt_0_lowtown_docks_n_bfa_0_0.vlm"},
	{0x1042FB75EF8FEA4EULL, "lgt260_eau_tmp_flowergirl6143867.cl"},
	{0x1043042D30E92EAAULL, "gen00fl_anders6111319.cl"},
	{0x1043A6A6E3F19E5DULL, "458395588.ogg"},
	{0x1043AD23400D6643ULL, "6105876_m.fxe"},
	{0x1043ADEA777511D0ULL, "1026023958.ogg"},
	{0x1043C5D09AE83270ULL, "355552108.ogg"},
	{0x10442F80E00332D7ULL, "vxht_bartrand_n_4aa_0_0.vlm"},
	{0x104434F9C31E236DULL, "gen_im_fla_carver_02.uti"},
	{0x1044B3C5FBE95C01ULL, "6029062_m.fxe"},
	{0x1044E2E9222FA4C8ULL, "fx_lightning_2.mao"},
	{0x10452ED0AF39F432ULL, "243222676.ogg"},
	{0x10453F01E3699DDFULL, "vxlt_gallowstemplar_3jx_0_0.vlm"},
	{0x1045758FBC9B0D10ULL, "vxlt_0_lowtown_docks_cs6_0_0.vlm"},
	{0x1045DFB992F37023ULL, "15464333.ogg"},
	{0x10466BEB35E30976ULL, "dre000_dougal6130978.cl"},
	{0x10467A0ECE76181FULL, "vxlt_shop_kz_0_0.vlm"},
	{0x10469D3202ECBF25ULL, "art\\vfx\\cinematic\\vfx_chantry_int_exp_p.impa.gad"},
	{0x1046C39B50D168F9ULL, "vxlt_backalley_g_1pp_0_0.vlm"},
	{0x10475F40798EBAE9ULL, "vxw_encounter2_camp_er_0_0.vlm"},
	{0x10482D7EB4F29780ULL, "457197641.ogg"},
	{0x10484E34AB1AECB4ULL, "m_mag201ar_caverns.wbd"},
	{0x1048697E26B45DEFULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfireout.anb"},
	{0x10487B620B9D8ACEULL, "vxlt_gamlenshouse_n_ay_0_0.vlm"},
	{0x10488F17FA6B3DFCULL, "483427831.ogg"},
	{0x1048F0054BBC8492ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblepile.phy"},
	{0x10497E5398B21FA3ULL, "gen00fl_anders6119125.cl"},
	{0x10498CC345F06FCDULL, "art\\characters\\creatures\\arh_abm_rev\\c_rev_head.mao"},
	{0x1049CED7C350CAAEULL, "6096823_m.fxe"},
	{0x104A514DABD35441ULL, "6104712_m.fxe"},
	{0x104A7823DEF7C377ULL, "hlac_lt_0_lowtown_a2_n_c_0_3x.dds"},
	{0x104B89C795691580ULL, "str_creater_half.mao"},
	{0x104BB4ED09A97387ULL, "mag370_meredith6064603.cl"},
	{0x104BD241F1A32A40ULL, "art\\levels\\races\\design\\animation\\beartrap.inactivetoactive.ani"},
	{0x104C6E1CE8B92769ULL, "gen00fl_merrill6111165.cl"},
	{0x104C93746BD83B5DULL, "vxlt_hangedman_n_22u_0_0.vlm"},
	{0x104C98F565561967ULL, "gen_im_trs_dal_amu_r2.uti"},
	{0x104CED3DD17A59E8ULL, "6109318_m.fxe"},
	{0x104DC4269635D40BULL, "vxlt_0_lowtown_docks_c1c_0_0.vlm"},
	{0x104DC85E253E7094ULL, "zz_followers_banter6157642.cl"},
	{0x104E0473B174A08CULL, "vxca_caverns1_ia3_0_0.vlm"},
	{0x104E3F15DB3A3421ULL, "hlac_ht_keep_3733_1y.dds"},
	{0x104F77806D33C432ULL, "217548164.ogg"},
	{0x104FCC020EC6F8A8ULL, "vrc221cr_guard_fake.utc"},
	{0x104FD88EC392A0C1ULL, "hlac_ht_keep_3732_10z.dds"},
	{0x104FF00DA4BE9486ULL, "6041148_m.fxe"},
	{0x104FF10E34191567ULL, "vxlt_0_lowtown_a3_n_c_i0w_0_0.vlm"},
	{0x104FF2D9B32DFFEAULL, "336466569.ogg"},
	{0x10500BC19EE0E978ULL, "6028806_m.fxe"},
	{0x105016383D1B85E5ULL, "hlac_ht_mansion2b_818_5x.dds"},
	{0x105042DF78E22D8FULL, "vxw_bonepitmining_a2_1xf_0_0.vlm"},
	{0x10504720E559E141ULL, "hlac_lt_0_lowtown_docks_-4_7y.dds"},
	{0x1050B6913F34C6C2ULL, "952998857.ogg"},
	{0x105124E32D0CA646ULL, "vxlt_0_lowtown_docks_n_bcb_0_0.vlm"},
	{0x105133DBACFA0FADULL, "150529302.ogg"},
	{0x10515ABBC08E6BD7ULL, "isa311_castillon6092116.cl"},
	{0x10522E9C9566F901ULL, "gen00fl_anders6125308.cl"},
	{0x1052EBCAFF812817ULL, "622740422.ogg"},
	{0x1052F7C41B4FD75AULL, "w_woundedcoast_160_0.msh"},
	{0x105305C41B4FEE93ULL, "w_woundedcoast_160_0.mao"},
	{0x105309C41B4FF5A0ULL, "w_woundedcoast_160_0.mmh"},
	{0x10536DEF800F88AFULL, "art\\vfx\\creatures\\fxc_shade_c.dura.gad"},
	{0x10537C531714B3FBULL, "6175804_m.fxe"},
	{0x105399E4EA123244ULL, "genip_trap_door.utp"},
	{0x1053EA99B080C599ULL, "vxht_0_hightown_n_c1c_0_0.vlm"},
	{0x1053ED6F30978634ULL, "mcr310_argument6075802.cl"},
	{0x1054511B077F3DB2ULL, "808917618.ogg"},
	{0x10549158CEB4D24BULL, "store_and211cr_lirene.utm"},
	{0x10555DE4F8548E83ULL, "vxlt_hangedman_n_1ow_0_0.vlm"},
	{0x10558AA1468EF1A8ULL, "w_bonepitmining_a2_176_0.mmh"},
	{0x105596A1468F064BULL, "w_bonepitmining_a2_176_0.mao"},
	{0x1055994960B73A3DULL, "vxht_0_hightown_a2_ccb_0_0.vlm"},
	{0x1055A8A1468F24D2ULL, "w_bonepitmining_a2_176_0.msh"},
	{0x1055B3EF97B3C7E1ULL, "vxlt_gallowscourtyard_7f8_0_0.vlm"},
	{0x1055DFF98DC9438CULL, "6108887_m.fxe"},
	{0x1056202067DE1EDDULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_bigtreasure_dp.dds"},
	{0x1056A3F0992B8620ULL, "vxw_encounter2_b6_0_0.vlm"},
	{0x1056C1CBDCAE85B1ULL, "6170504_m.fxe"},
	{0x1056F8857B60F9B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tw_tal_crtstrk.ani"},
	{0x10572B5ECE995CADULL, "6170514_f.fxe"},
	{0x105789BA4C739806ULL, "6162531_m.fxe"},
	{0x1057F1EE5BAE0640ULL, "243928189.ogg"},
	{0x105A0E488B2E9EE1ULL, "vxca_caverns2prc_9ne_0_0.vlm"},
	{0x105A0F7C00B09BABULL, "vxca_caverns1_mines_jfo_0_0.vlm"},
	{0x105A47EC0B99C9B3ULL, "dae000_gen_turn_in_46198681.cl"},
	{0x105A6BA591FE34B7ULL, "104386566.ogg"},
	{0x105A82EF8015CB5CULL, "art\\vfx\\creatures\\fxc_shade_c.dura.evt"},
	{0x105A8775D9642AEFULL, "qun231tr_varnell_amb.ncs"},
	{0x105AA46437D427C2ULL, "dae000_gen_turn_in_56178423.cl"},
	{0x105AB6477960C121ULL, "37218068.ogg"},
	{0x105AF5F9E7C50D5AULL, "125269205.ogg"},
	{0x105B793375D3CA92ULL, "vxht_chantry_23l_0_0.vlm"},
	{0x105B8089D8E23A4DULL, "995277998.ogg"},
	{0x105B8CF1C6CB4BFCULL, "2522359.ogg"},
	{0x105BB79901A53C13ULL, "mag121_viveka6082603.cl"},
	{0x105C5427485EBFC5ULL, "vxlt_0_lowtown_docks_bdo_0_0.vlm"},
	{0x105C953E04CB49FCULL, "6101493_m.fxe"},
	{0x105D24700AD6446BULL, "6149300_m.fxe"},
	{0x105D6C67CE3D4387ULL, "hlac_ht_mansion2_818_44y.dds"},
	{0x105E1F884F61B581ULL, "vxlt_warehouse_n_12b_0_0.vlm"},
	{0x105E9CC752D0CE90ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun_2.mao"},
	{0x105F1606D24C3490ULL, "vfx_vfx_p10.vsh.fxo.dx11.assembly"},
	{0x105F4BC37835F038ULL, "6103583_m.fxe"},
	{0x105F57FDF6F62BDCULL, "hlac_lt_0_lowtown_a2_n_c_65536_1x.dds"},
	{0x105FABA3726110B3ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_008a_0d.dds"},
	{0x105FF269B0591660ULL, "fake_im_arm_hlm_hvy_champ.uti"},
	{0x10602EEE38E4E41DULL, "zz_lgt_3_debug6140424.cl"},
	{0x1060482B1657E641ULL, "353764939.ogg"},
	{0x10608F271310C303ULL, "10781822.ogg"},
	{0x1060AFB50FB9AF4AULL, "6057078_f.fxe"},
	{0x1060F115BAC6B7E7ULL, "hlac_lt_darktown_-1_11y.dds"},
	{0x10612F0B355DF3CDULL, "647651750.ogg"},
	{0x1061754E5B3D47FFULL, "art\\characters\\playercharacter\\humanfemale\\hf_boo_lgte.mmh"},
	{0x10617B4E5B3D51A1ULL, "art\\characters\\playercharacter\\humanfemale\\hf_boo_lgte.msh"},
	{0x106197AC8D110E54ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_3y.dds"},
	{0x106296991484760EULL, "follower_banter6165409.cl"},
	{0x10629E0A9C0A555BULL, "lt_gallowscourtyard_44_0.rml"},
	{0x1062DA5E9359492FULL, "256050399.ogg"},
	{0x10631C21E01FA1C5ULL, "vxht_keep_n_4ad_0_0.vlm"},
	{0x10635ACD6818904DULL, "genip_pickaxe.utp"},
	{0x10635DC8B262491FULL, "vxlt_shop_ad_0_0.vlm"},
	{0x1063641815E914A1ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage1_.ani"},
	{0x106377641F6DC8E0ULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_art_hm00.phy"},
	{0x10639F8466B90ACEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cot.mmh"},
	{0x1063A75534BD99F9ULL, "vxlt_backalley_g_1zv_0_0.vlm"},
	{0x1063AB8466B91F65ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cot.mao"},
	{0x1063B98466B936BCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cot.msh"},
	{0x1063F0C01D4B0B73ULL, "mcr371_db_harvester_warr.cut"},
	{0x10641499DB495591ULL, "llac_dr_0_deeproads_5849_3.dds"},
	{0x1064604E5583EB43ULL, "6111749_m.fxe"},
	{0x106493EC563259E1ULL, "123525061.ogg"},
	{0x106500857B6D44A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tw_tal_crtstrk.evt"},
	{0x106515CD8B219EA0ULL, "vxw_encounter2_m7_0_0.vlm"},
	{0x106523CFDCC4FF49ULL, "vxlt_warehouse_n_e0_0_0.vlm"},
	{0x10653FE75609B71EULL, "vxw_encounter2_qt_0_1.vlm"},
	{0x106571728F41AC39ULL, "combat_smcf_range.gda"},
	{0x1066477803ABC7EEULL, "821351248.ogg"},
	{0x106662ED30F81AC7ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_0y.dds"},
	{0x106684736164046AULL, "llac_ht_0_hightown_a3_-65536_4.dds"},
	{0x1067792747258FD0ULL, "6015679_m.fxe"},
	{0x1067B30DA80D7FD9ULL, "498695169.ogg"},
	{0x1067E3CCA9C5A7F8ULL, "285269219.ogg"},
	{0x1067FD71F598F247ULL, "80568563.ogg"},
	{0x10681F8563B8D1D2ULL, "hlac_lt_backalley_g_1_1z.dds"},
	{0x10684096E2B748BBULL, "qun221_viscount6070481.cl"},
	{0x10685229E1A98FFCULL, "vxht_keep_n_62b_0_0.vlm"},
	{0x10685C81BCF3F198ULL, "6190969_m.fxe"},
	{0x106877554B04EAD5ULL, "380178987.ogg"},
	{0x10688AEF80221675ULL, "art\\vfx\\creatures\\fxc_shade_c.dura.ani"},
	{0x1068AE776D2D599FULL, "6084205_m.fxe"},
	{0x10693AB6EACC3896ULL, "577221824.ogg"},
	{0x106988D53EE2957AULL, "vxw_bonepitdestroyed_a3_1bs_0_0.vlm"},
	{0x10699C31FD13E837ULL, "679104789.ogg"},
	{0x1069C847866F0BFDULL, "309037319.ogg"},
	{0x106A6F1815EF4643ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage1_.gad"},
	{0x106A9879A03DA025ULL, "vxlt_0_lowtown_a2_bz0_0_0.vlm"},
	{0x106AE42C269DBF86ULL, "6107058_f.fxe"},
	{0x106B21857B71E8F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tw_tal_crtstrk.gad"},
	{0x106B72E3163B3722ULL, "1000599855.ogg"},
	{0x106BCD338B726B32ULL, "6084693_m.fxe"},
	{0x106C0FBF0D9973BFULL, "6165122_m.fxe"},
	{0x106C2305FBCA6114ULL, "vxdr_0_deeproads_cvr_0_1.vlm"},
	{0x106CBAB00BF138BCULL, "6097835_m.fxe"},
	{0x106CCE22CF620DE7ULL, "swashbklr_expertise.dds"},
	{0x106CDAAB0304340AULL, "hlac_lt_0_lowtown_a2_n_c_65535_8x.dds"},
	{0x106CE5B96E768D5CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2.ani"},
	{0x106CED2C14127E90ULL, "gen00fl_merrill6162018.cl"},
	{0x106D03EC33E0CF77ULL, "vxdr_primevalentrance_146_0_0.vlm"},
	{0x106D38ECC600D721ULL, "qun160_vashothbandits6069277.cl"},
	{0x106D588D3DCA124DULL, "274586344.ogg"},
	{0x106D7DD2A45686E3ULL, "vxdr_primevaltomb_4h7_0_0.vlm"},
	{0x106DC1C45ED3F4A1ULL, "gen00fl_fenris6126390.cl"},
	{0x106DD29AFC0402A6ULL, "darktown_male_sickly_sta6175325.cl"},
	{0x106DF990311F3EBCULL, "hlac_lt_0_lowtown_a2_0_1z.dds"},
	{0x106E089D10F8D746ULL, "vxlt_0_lowtown_docks_n_bud_0_0.vlm"},
	{0x106E2072E737BA11ULL, "813578535.ogg"},
	{0x106E77CD5DD77540ULL, "vxw_sundermount_2m4_0_0.vlm"},
	{0x106EFC499DCF9F26ULL, "vxlt_shop_f8_0_0.vlm"},
	{0x106F3A6E867055DFULL, "ser221ip_note_shipment.ncs"},
	{0x106F3BF9E9BA8CEDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkendb.mmh"},
	{0x106F3FF9E9BA933EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkendb.mao"},
	{0x106F4DF9E9BAAB73ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkendb.msh"},
	{0x10716F5141EB816EULL, "vxht_0_hightown_n_apy_0_0.vlm"},
	{0x1071CFA5A458619FULL, "vxlt_shop_dg_0_0.vlm"},
	{0x1071DDA1A1468138ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryb.phy"},
	{0x10720730AD7205FFULL, "w_bonepitdestroyed_a3_0_0.rml"},
	{0x107288C8B5115865ULL, "gen00fl_anders6125216.cl"},
	{0x1072F2B96E7B0E9EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2.gad"},
	{0x1073501D582364EFULL, "292896713.ogg"},
	{0x107369AE783C60F1ULL, "ts_generic_acc_props.gda"},
	{0x1073B004649DB95FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2a.phy"},
	{0x1073DD7A37721C50ULL, "w_bonepit_23_0.mao"},
	{0x1073E17A37722363ULL, "w_bonepit_23_0.mmh"},
	{0x1073EF7A37723ABDULL, "w_bonepit_23_0.msh"},
	{0x1073F9F4655BA3DBULL, "vxw_bonepitmining_a2_p9_0_0.vlm"},
	{0x10743C34AB40094EULL, "m_mag201ar_caverns.bnk"},
	{0x10749880724E7E31ULL, "vxlt_0_lowtown_a2_n_c_fe8_0_0.vlm"},
	{0x1074A8FD259D442BULL, "909335204.ogg"},
	{0x107581991EA4CD4FULL, "follower_banter6143095.cl"},
	{0x10766EC7F156ADB2ULL, "6110784_f.fxe"},
	{0x1076BE85093E2D55ULL, "vxca_caverns1_j25_0_1.vlm"},
	{0x1076E0478DEF4AC5ULL, "707820498.ogg"},
	{0x107704B3A4C8141FULL, "781239137.ogg"},
	{0x1077A51503398113ULL, "dae100_alistair_drunk6135129.cl"},
	{0x1077B0FBA193BD69ULL, "dre000_sibling_has_blight6036567.cl"},
	{0x1077FAB230D8ADFAULL, "vxw_bonepitdestroyed_a3_1gl_0_0.vlm"},
	{0x107813AE64D246AAULL, "6127662_m.fxe"},
	{0x10782F4EA9DA3BCAULL, "87738496.ogg"},
	{0x10784DB87490617FULL, "vxlt_0_lowtown_a2_n_c_eoo_0_0.vlm"},
	{0x10787475D97D0796ULL, "qun231tr_varnell_amb.utt"},
	{0x1078891DA4092F08ULL, "218777929.ogg"},
	{0x1078D1F652B7D4CCULL, "pro000_cut2d_thefirstyear6015285.cl"},
	{0x107905B1E8F20DC0ULL, "dae000_hate_enemies6167600.cl"},
	{0x10792E6C6AEC634DULL, "hlac_lt_0_lowtown_n_-196608_1y.dds"},
	{0x107965EC4C7987AFULL, "645210087.ogg"},
	{0x10799CA9CBB1C102ULL, "109165710.ogg"},
	{0x1079A0D0D48CC1CEULL, "vxht_playershouse_n_1lh_0_0.vlm"},
	{0x1079CDB96E80EEBDULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2.evt"},
	{0x107A1C116E4FC542ULL, "art\\toolset\\characters\\playercharacter\\elfmale\\mt_art_em00.msh"},
	{0x107A1E116E4FC8F8ULL, "art\\toolset\\characters\\playercharacter\\elfmale\\mt_art_em00.mmh"},
	{0x107A251D346E4849ULL, "llac_ca_caverns1_24421_5.dds"},
	{0x107A2E52370AA1F0ULL, "ca_caverns1_24425.rml"},
	{0x107A6849B5FE3C72ULL, "vxlt_0_lowtown_ekf_0_1.vlm"},
	{0x107A7369F0729D40ULL, "343792752.ogg"},
	{0x107AEF7EE9F06328ULL, "zz_area_signage1.utp"},
	{0x107B19FCA6ED192EULL, "6065230_m.fxe"},
	{0x107B8C03B48DAD1DULL, "ser270_templar6175481.cl"},
	{0x107BB25AB5E8FF3FULL, "ico_mace_2.dds"},
	{0x107C77DD69E9B7AFULL, "gen00fl_varric6152698.cl"},
	{0x107C7A7A02EEEFC6ULL, "vxlt_0_lowtown_docks_n_bre_0_0.vlm"},
	{0x107D0383D0BE25EEULL, "vxdr_primevalentrance_2dx_0_0.vlm"},
	{0x107D3AA9636592B8ULL, "gen_im_trs_krk_hvy_glv_r2.uti"},
	{0x107D68763D0C9BACULL, "vxht_0_hightown_a3_avs_0_0.vlm"},
	{0x107DD8721A7B9B34ULL, "6060203_f.fxe"},
	{0x107DD9D80F0E69D9ULL, "character_character_p257.psh.fxo.dx11"},
	{0x107DF16616644B05ULL, "gen00fl_fenris6103359.cl"},
	{0x107E2AEC35391394ULL, "6108609_m.fxe"},
	{0x107E43646EFCE44CULL, "44168757.ogg"},
	{0x107E512555F15A61ULL, "vxlt_0_lowtown_docks_be0_0_0.vlm"},
	{0x107F868E85EC2AE0ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_2z.dds"},
	{0x107FE12E33E199EEULL, "vxht_0_hightown_n_auz_0_0.vlm"},
	{0x107FEA1A4072FBB2ULL, "hlac_ht_0_hightown_a3_65535_2x.dds"},
	{0x108010069FA204BCULL, "6032011_f.fxe"},
	{0x10801255A2183DAAULL, "cod_hst_kirk_arch_1a_5.utp"},
	{0x1080330EF9A926A8ULL, "6119275_f.fxe"},
	{0x108034F9E9C8F3AAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkendb.phy"},
	{0x10803A2AC807EE00ULL, "hla_ht_0_hightown_n_-131074_0x.dds"},
	{0x10805B06A1B8EF61ULL, "6031342_m.fxe"},
	{0x10808BBA9F7090C7ULL, "darktown_male_elf_sta6174545.cl"},
	{0x1080AC7A31A91F91ULL, "mrl150_merrill6150734.cl"},
	{0x10810D6BCAA3F398ULL, "6045919_m.fxe"},
	{0x1081DA9DA2D544B8ULL, "dae150_pol.cnv"},
	{0x1081F34E9420FAD5ULL, "446293522.ogg"},
	{0x1081F9876F0DCFA0ULL, "vxw_encounter2_s5_0_0.vlm"},
	{0x10820789C0B13049ULL, "vxlt_warehouse_n_o6_0_0.vlm"},
	{0x10821EAB3A8A2D08ULL, "mag291_vincento6027348.cl"},
	{0x1082CA8ADA66435EULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv4.ani"},
	{0x1082D4B83AAD1757ULL, "723864490.ogg"},
	{0x1082EF2A296AF94BULL, "6169423_m.fxe"},
	{0x108325A838756C26ULL, "llac_lt_0_lowtown_a2_n_c_-262144_0.dds"},
	{0x1084230929C7AA93ULL, "gen00fl_merrill6126316.cl"},
	{0x108425116E582883ULL, "art\\toolset\\characters\\playercharacter\\elfmale\\mt_art_em00.phy"},
	{0x10844A7CA00792FAULL, "vxdr_primevaltomb_8np_0_0.vlm"},
	{0x10847BF7D58CFD3CULL, "rdr111dg_martin6014874.cl"},
	{0x10847CA242ADCD16ULL, "hla_ht_0_hightown_a2_131073_0y.dds"},
	{0x1084B10464AC2E64ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2a.mmh"},
	{0x1084B50464AC34B7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2a.mao"},
	{0x1084BC9DA2D747C0ULL, "dae150_pol.bnk"},
	{0x1084C0B04A0022A1ULL, "vxw_sundermount_2o3_0_0.vlm"},
	{0x1084C70464AC5306ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2a.msh"},
	{0x108503C87A067D96ULL, "vxlt_hangedman_1vg_0_1.vlm"},
	{0x10854E403DDDD4E3ULL, "287886985.ogg"},
	{0x10855EF0C242F29FULL, "vxlt_gallowscourtyard_73y_0_0.vlm"},
	{0x10857E2643129884ULL, "llac_lt_hangedman_n_-1_11.dds"},
	{0x1085A539F8497DF4ULL, "6121154_m.fxe"},
	{0x1085C34F74DBB006ULL, "follower_banter6143113.cl"},
	{0x1085E16577D9F287ULL, "572985416.ogg"},
	{0x1085FF3FB307690BULL, "279238501.ogg"},
	{0x10863F17BBF7D6E3ULL, "lgt100cr_gang_3_lt.utc"},
	{0x10866A22B4DABDFAULL, "gen00fl_isabela6075235.cl"},
	{0x10867DA437D34F0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_exit.ani"},
	{0x1086B2B0721A9B8FULL, "lgt160_ntb_elf_conflict6172100.cl"},
	{0x108796AD0348F363ULL, "940104672.ogg"},
	{0x1087CF3720193ADDULL, "1006468383.ogg"},
	{0x1087D9DB17460648ULL, "vxlt_0_lowtown_docks_be5_0_0.vlm"},
	{0x10892CB756DC9708ULL, "w_bonepitdestroyed_a3_9_0_ll.mmh"},
	{0x10897038A54DD4AFULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml101a_0s.dds"},
	{0x10898798059DEAFEULL, "hlac_lt_0_lowtown_a3_n_c_-65536_8y.dds"},
	{0x108994641F8EC975ULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_art_hm00.msh"},
	{0x108996641F8ECCEBULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_art_hm00.mmh"},
	{0x1089AABE5AADCB16ULL, "hlac_w_woundedcoast_196608_0y.dds"},
	{0x108A3CD560F4DDCCULL, "gen00fl_sebastian6178335.cl"},
	{0x108A9B9ADAA8A40CULL, "260990103.ogg"},
	{0x108B042FE4CD73B4ULL, "mrl000_marethari6129355.cl"},
	{0x108B0AB572BD1808ULL, "hlac_lt_0_lowtown_docks_a3_-131074_6x.dds"},
	{0x108B195FFC9552D5ULL, "333330169.ogg"},
	{0x108B35DBCAD87387ULL, "weapon_weapon_p2.psh.fxo.dx11"},
	{0x108B3CDA938CC227ULL, "vxlt_warehouse_5w_0_0.vlm"},
	{0x108B41AA601F88A1ULL, "6148419_m.fxe"},
	{0x108BDA532F02B42CULL, "vxht_0_hightown_a3_asp_0_0.vlm"},
	{0x108BE17A3786B480ULL, "w_bonepit_23_0.rml"},
	{0x108BF14340960C34ULL, "gen_im_trs_chn_light_doc_01.uti"},
	{0x108C56B1C78DE829ULL, "69958103.ogg"},
	{0x108D78529D38185FULL, "vxlt_gallowstemplar_3sg_0_0.vlm"},
	{0x108DB0F6DB68C064ULL, "vxht_0_hightown_n_b44_0_0.vlm"},
	{0x108E0BC0BA3A3DBAULL, "gen00fl_sebastian6191246.cl"},
	{0x108E0F9FBF8CE296ULL, "vxca_caverns1_mines_dqa_0_0.vlm"},
	{0x108E3E32B0B1F1B6ULL, "6004411_m.fxe"},
	{0x108E8DCD79A04F62ULL, "vxca_caverns2_839_0_0.vlm"},
	{0x108EDDA318E830D3ULL, "fen310_fenris6118192.cl"},
	{0x108F354D4FBD7726ULL, "290392438.ogg"},
	{0x108F4D392916C314ULL, "gen_im_trs_slv_lgt_hlm_r1.uti"},
	{0x108FE28ADA70F62FULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv4.evt"},
	{0x1090755C09C4E027ULL, "6174044_f.fxe"},
	{0x10907966B2A748C9ULL, "vxlt_warehouse_n_r5_0_0.vlm"},
	{0x109093A69530D253ULL, "follower_banter6173016.cl"},
	{0x109121311F9FE7C9ULL, "13072554.ogg"},
	{0x10922DB7138783EFULL, "6174818_m.fxe"},
	{0x1092BC5991FDAB7AULL, "vxdr_primevaltomb_8iq_0_0.vlm"},
	{0x1092C4D36192522DULL, "219334717.ogg"},
	{0x1093328D3BF63B21ULL, "vxw_sundermount_2p2_0_0.vlm"},
	{0x109375A437DDCC5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_exit.evt"},
	{0x1093D0CDB4390B1FULL, "vxlt_gallowscourtyard_74x_0_0.vlm"},
	{0x10941252EB350B9FULL, "6045372_m.fxe"},
	{0x10943330ABFA8018ULL, "750152552.ogg"},
	{0x10943FE78E745DC3ULL, "vxht_0_hightown_a2_7vs_0_0.vlm"},
	{0x1094533F7991B2A3ULL, "vxw_encounter1_eq_0_0.vlm"},
	{0x1095222AAD1607F2ULL, "hlac_lt_0_lowtown_a2_n_-2_1z.dds"},
	{0x1095F49A39B93954ULL, "hlac_lt_0_lowtown_a2_1_11z.dds"},
	{0x109601E70AC57414ULL, "232900200.ogg"},
	{0x109611A95D06AC68ULL, "llac_ca_cave_2193_0.dds"},
	{0x109618571AAD95A5ULL, "946422068.ogg"},
	{0x1096E38ADA7716D8ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv4.gad"},
	{0x1097382C76D1F2E0ULL, "6019095_m.fxe"},
	{0x109771D3373D6D9CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_idle.ani"},
	{0x109776B1AD8A7DC9ULL, "464670217.ogg"},
	{0x1097FA28CFF13C2AULL, "91504283.ogg"},
	{0x109804DFCB95F335ULL, "136837153.ogg"},
	{0x1098F82BF593BAF6ULL, "w_blightlands_162_0.msh"},
	{0x1099022BF593CBF4ULL, "w_blightlands_162_0.mmh"},
	{0x1099062BF593D207ULL, "w_blightlands_162_0.mao"},
	{0x109A56A437E3B648ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_exit.gad"},
	{0x109AC67C51624231ULL, "176577738.ogg"},
	{0x109AD12656C9C919ULL, "hlac_lt_blackemporium_1055_0y.dds"},
	{0x109B5A84734B6A3EULL, "vxlt_0_lowtown_a2_n_gv9_0_0.vlm"},
	{0x109D2073A2D9C561ULL, "6206863_m.fxe"},
	{0x109D27A9B47D69C2ULL, "vxlt_0_lowtown_a3_n_c_gv6_0_0.vlm"},
	{0x109D8470E2EE8C1DULL, "981780945.ogg"},
	{0x109DBD5B5929BD67ULL, "vxlt_0_lowtown_docks_a3_d3q_0_0.vlm"},
	{0x109DEC4E4D4BDB60ULL, "285572359.ogg"},
	{0x109E7ED33743A2DEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_idle.gad"},
	{0x109EEB43A49D6149ULL, "vxlt_warehouse_n_y4_0_0.vlm"},
	{0x109F5C11F3660E62ULL, "vxca_caverns2prc_9pu_0_1.vlm"},
	{0x109F80539885BE00ULL, "gen00fl_isabela6079282.cl"},
	{0x10A0476D8509E4A7ULL, "985261941.ogg"},
	{0x10A0720174DC228FULL, "893574924.ogg"},
	{0x10A1586AED6244E6ULL, "6070395_m.fxe"},
	{0x10A1A46A2DEC53A1ULL, "vxw_sundermount_2u1_0_0.vlm"},
	{0x10A1B9461AC92349ULL, "hlac_w_woundedcoast_-196606_0y.dds"},
	{0x10A21E232882C38DULL, "ts_circleofmagi_02.gda"},
	{0x10A27A4139DCA2A7ULL, "hlac_lt_gallowstemplar_f_2383_4z.dds"},
	{0x10A2A10A9C403E27ULL, "lt_gallowscourtyard_44_0.anb"},
	{0x10A2A83DEC930F0DULL, "vxdr_0_deeproads_6qw_0_1.vlm"},
	{0x10A2BB8D1C3069B3ULL, "qun200_party_comment6095260.cl"},
	{0x10A2BFB43BD78760ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be01a.phy"},
	{0x10A2C51C6B87CB23ULL, "vxw_encounter1_jp_0_0.vlm"},
	{0x10A2F256354D3D45ULL, "vxw_bonepitdestroyed_a3_1gg_0_0.vlm"},
	{0x10A3676D6DB706A3ULL, "vxdr_primevaltomb_9dv_0_0.vlm"},
	{0x10A3F53A438A1E6EULL, "gen00fl_aveline6114834.cl"},
	{0x10A42A6579091446ULL, "6179269_m.fxe"},
	{0x10A42C798F3DF715ULL, "vxw_sundermount_2y9_0_0.vlm"},
	{0x10A45E138C5788E7ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.ces.gad"},
	{0x10A57E3B0DCAC005ULL, "190789917.ogg"},
	{0x10A5845ECA4DA58EULL, "mag210_arianni6030488.cl"},
	{0x10A597FBB6AE65E2ULL, "hlac_ht_bartrand_n_3373_3x.dds"},
	{0x10A5A72D06B7CA5FULL, "291578766.ogg"},
	{0x10A5DC743C549CF3ULL, "zz_dragon_debug.ncs"},
	{0x10A6220788A1405AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_questioning.gad"},
	{0x10A6650B2D55181CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicerchair_sa.dds"},
	{0x10A69797A2DD8AC4ULL, "mag100_cullen6094605.cl"},
	{0x10A6CEAE09EEC987ULL, "vxw_bonepitmining_a2_1nd_0_0.vlm"},
	{0x10A7529533B5A085ULL, "6069924_f.fxe"},
	{0x10A7AB0EF4545138ULL, "742461215.ogg"},
	{0x10A7AC560EEC7D36ULL, "69155451.ogg"},
	{0x10A7B61C1DD003BAULL, "hlac_lt_gallowscourtyard_n_65535_0y.dds"},
	{0x10A8885EB72965CFULL, "ci_smoke05.dds"},
	{0x10A8AD7743EA45BBULL, "gen_im_trs_rdr_junk1_01.uti"},
	{0x10A8DE20E1074CABULL, "vxlt_gallowstemplar_n_4o3_0_0.vlm"},
	{0x10A95C993961DEB9ULL, "llac_ca_caverns1_24425_10.dds"},
	{0x10AA5530788AE7E5ULL, "vxw_sundermount_2nl_0_0.vlm"},
	{0x10AB30172A17212AULL, "follower_banter6141450.cl"},
	{0x10AB37BEB3BDB515ULL, "zz_dre_debug6056958.cl"},
	{0x10ABB1D701CE9409ULL, "897966864.ogg"},
	{0x10ABCD41EB33CB29ULL, "vxca_caverns1_ckr_0_0.vlm"},
	{0x10AC2F384B1FD5E7ULL, "vxlt_0_lowtown_docks_a3_d4p_0_0.vlm"},
	{0x10AC5ADE4229E281ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_a_c.impa.ani"},
	{0x10AC859C7A3D2343ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_3craggyrock.mao"},
	{0x10AC899C7A3D2990ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_3craggyrock.mmh"},
	{0x10AC979C7A3D41AAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_3craggyrock.msh"},
	{0x10ACCE5BBB9C1641ULL, "mcr310_sibling6103598.cl"},
	{0x10ACD1F5EF334101ULL, "vxht_0_hightown_a2_9zc_0_0.vlm"},
	{0x10AD250788A76471ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_questioning.evt"},
	{0x10ADEFF0E96D12F6ULL, "vrc321tr_spooky_scenes_3.ncs"},
	{0x10AE240DA5FAD7F8ULL, "476874127.ogg"},
	{0x10AEA08A34A46CC5ULL, "hlac_ht_keep_n_4519_8x.dds"},
	{0x10AF4697584A35F5ULL, "zz_qcr_debug6056621.cl"},
	{0x10AF7E4A73EC2DDCULL, "vxlt_backalley_n_1kq_0_0.vlm"},
	{0x10AF9A2D7A73091DULL, "one220_yevhen6116245.cl"},
	{0x10AFB336C85D567BULL, "dae111st_dog_family.stg"},
	{0x10AFB98A0EF958CAULL, "vxca_caverns2_7ud_0_0.vlm"},
	{0x10B010D94E1AE080ULL, "dae000ip_crowd_em1_talk8.utp"},
	{0x10B0AB8D7D530BD7ULL, "739366632.ogg"},
	{0x10B11C3070C85721ULL, "mag101_aombanter6052759.cl"},
	{0x10B16433274355C5ULL, "vxw_bonepitdestroyed_a3_1hf_0_0.vlm"},
	{0x10B1C516E6AB7EB5ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyn.phy"},
	{0x10B1FDB602EC17E2ULL, "vxdr_primevaltomb_3hm_0_0.vlm"},
	{0x10B22989547CF472ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_p.impa.gad"},
	{0x10B265DE422E6123ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_a_c.impa.gad"},
	{0x10B32A9AB4DA4CF0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_ornate.mao"},
	{0x10B32E9AB4DA5303ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_ornate.mmh"},
	{0x10B33C9AB4DA6B5DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_ornate.msh"},
	{0x10B4247D2B92FDC5ULL, "6172421_m.fxe"},
	{0x10B44EDBB0985DA4ULL, "334587398.ogg"},
	{0x10B468B552D7216DULL, "follower_banter6149187.cl"},
	{0x10B48E2243966455ULL, "6122449_f.fxe"},
	{0x10B50B87A5775228ULL, "gen_im_cutscene_dagger.uti"},
	{0x10B537F1784ED70CULL, "6141497_m.fxe"},
	{0x10B5408AFBE4E207ULL, "vxw_bonepitmining_a2_1kg_0_0.vlm"},
	{0x10B5432010D1D751ULL, "lgt160_ntb_elf_conflict6163346.cl"},
	{0x10B580C06F6473C0ULL, "ser221_gascard6134540.cl"},
	{0x10B5A30129AF4870ULL, "6191294_m.fxe"},
	{0x10B5A9992EDEB557ULL, "dae000cr_grand_cleric6185387.cl"},
	{0x10B5B928FFB74F44ULL, "vxht_0_hightown_n_c3x_0_0.vlm"},
	{0x10B5E2B375386C8EULL, "hlac_dr_primevaltomb_9621_1y.dds"},
	{0x10B666B73239A977ULL, "zz_vault_debug6214944.cl"},
	{0x10B7E787769D26A4ULL, "542504273.ogg"},
	{0x10B84796481AE141ULL, "441464200.ogg"},
	{0x10B84B138C68112DULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.ces.ani"},
	{0x10B8722AD71F31A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_lhand.evt"},
	{0x10B8C70D6A810065ULL, "vxw_sundermount_2sk_0_0.vlm"},
	{0x10B92C89548317A9ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_p.impa.evt"},
	{0x10B942DE42344408ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_a_c.impa.evt"},
	{0x10B9766F5E92FF19ULL, "llac_ca_cavern1_haunted_24424_16.dds"},
	{0x10B99121BE65B07FULL, "art\\characters\\playercharacter\\humanmale\\arm_bda.phy"},
	{0x10B9E54A096A2FE5ULL, "lt_gallowscourtyard_59_0_ll.mmh"},
	{0x10BA26D046900E49ULL, "follower_banter6172295.cl"},
	{0x10BA29898865D801ULL, "197412335.ogg"},
	{0x10BA3D0788B21740ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_questioning.ani"},
	{0x10BA9EFE5AED48C8ULL, "6054849_m.fxe"},
	{0x10BAC5D478809EFEULL, "vxw_bonepitmining_a2_e7_0_0.vlm"},
	{0x10BAE5C9AD3336F0ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.mmh"},
	{0x10BB3BD214F5C548ULL, "vxlt_0_lowtown_a3_n_c_env_0_0.vlm"},
	{0x10BC155CBF967DB3ULL, "vxlt_0_lowtown_docks_a3_c1h_0_0.vlm"},
	{0x10BC250A221B3977ULL, "fen110_fenris6096216.cl"},
	{0x10BCB70D58412F3BULL, "6096765_m.fxe"},
	{0x10BDF02765E2465CULL, "vxlt_backalley_n_1vv_0_0.vlm"},
	{0x10BE892AD723C4F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_lhand.gad"},
	{0x10BE93D4DFBE2358ULL, "vxht_0_hightown_a2_c42_0_0.vlm"},
	{0x10BEF7BF924FBC61ULL, "vxca_caverns1_k2u_0_0.vlm"},
	{0x10BF121BFFD3A9CEULL, "6027557_m.fxe"},
	{0x10BF41A62A1A2D1FULL, "mag291_ntbanter6056619.cl"},
	{0x10BF6929C3A1C331ULL, "mcr370_meredith6083625.cl"},
	{0x10BF961F02840170ULL, "119650611.ogg"},
	{0x10BFD61019396E45ULL, "vxw_bonepitdestroyed_a3_1me_0_0.vlm"},
	{0x10BFF0580A3CE315ULL, "hlac_lt_0_lowtown_65535_2x.dds"},
	{0x10BFFB30ADB412EFULL, "w_bonepitdestroyed_a3_0_0.mao"},
	{0x10C00730ADB4268CULL, "w_bonepitdestroyed_a3_0_0.mmh"},
	{0x10C00D30ADB4317EULL, "w_bonepitdestroyed_a3_0_0.msh"},
	{0x10C024F9459731E1ULL, "vxlt_0_lowtown_a3_n_c_eqr_0_0.vlm"},
	{0x10C0252734F403FFULL, "hlac_lt_0_lowtown_a2_65537_4x.dds"},
	{0x10C0746C370FCCA4ULL, "hlac_ca_cavern1_haunted_24421_12x.dds"},
	{0x10C0C3143AB534DAULL, "llac_lt_gallowstemplar_n_2383_3.dds"},
	{0x10C12C34C496A18AULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.anb"},
	{0x10C17131E6196B45ULL, "vxlt_0_lowtown_docks_b9h_0_0.vlm"},
	{0x10C21434A02FD99FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_basic.msh"},
	{0x10C21A34A02FE441ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_basic.mmh"},
	{0x10C22634A02FF8E2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_basic.mao"},
	{0x10C2736BAD2039D6ULL, "402752921.ogg"},
	{0x10C3064B3B582AB9ULL, "193990240.ogg"},
	{0x10C33A1E0B1AF2C7ULL, "vxca_cavern1_haunted_j26_0_1.vlm"},
	{0x10C3B8410FB1CF31ULL, "vxlt_backalley_g_1ki_0_0.vlm"},
	{0x10C3DD26CFBFF418ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue.anb"},
	{0x10C42B05F1AD67C4ULL, "vxht_0_hightown_n_c8y_0_0.vlm"},
	{0x10C42DCF2B9BB1E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_enter.ani"},
	{0x10C42F9AB4E8C828ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_ornate.phy"},
	{0x10C47E4E5B91D444ULL, "art\\characters\\playercharacter\\humanfemale\\hf_boo_lgte.phy"},
	{0x10C4C3E0D66AADDFULL, "vxht_0_hightown_a2_awr_0_0.vlm"},
	{0x10C51EF0FCD00668ULL, "349728597.ogg"},
	{0x10C58A2AD729E49FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_lhand.ani"},
	{0x10C5DAFE2B9E5B6FULL, "6017810_m.fxe"},
	{0x10C60380D01F9542ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_nud_0s.dds"},
	{0x10C645C4B98C6787ULL, "s_ballista_04_l3.mao"},
	{0x10C6FA60DAD514A8ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_dssta_0t.dds"},
	{0x10C71489548F2C18ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_p.impa.ani"},
	{0x10C738EA5C7718E5ULL, "vxw_sundermount_2tj_0_0.vlm"},
	{0x10C75AD1ADE50632ULL, "6110769_m.fxe"},
	{0x10C7EA610AA6B07FULL, "vxlt_gamlenshouse_n_3r_0_0.vlm"},
	{0x10C854A80B6455BAULL, "6148194_f.fxe"},
	{0x10C875FEFE05C5BBULL, "754900286.ogg"},
	{0x10C8807553A46848ULL, "vxca_caverns1_k2p_0_0.vlm"},
	{0x10C8DCB43BF887F5ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be01a.msh"},
	{0x10C8DEB43BF88B6BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be01a.mmh"},
	{0x10C8DF7D2F93E439ULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_0y.dds"},
	{0x10C908C9AD3FAF0FULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.anb"},
	{0x10C910DB4CE36148ULL, "6051749_m.fxe"},
	{0x10C9124CB06175A0ULL, "hlac_lt_0_lowtown_a3_n_c_65536_2z.dds"},
	{0x10C9ADAF06EBDDC8ULL, "vxlt_0_lowtown_a3_n_c_eqw_0_0.vlm"},
	{0x10C9B76E70637B37ULL, "plt_cod_crt_shade.plo"},
	{0x10CA3F9C5E2AFEB1ULL, "follower_banter6163280.cl"},
	{0x10CA9221BE742484ULL, "art\\characters\\playercharacter\\humanmale\\arm_bda.mmh"},
	{0x10CA9621BE742BD7ULL, "art\\characters\\playercharacter\\humanmale\\arm_bda.mao"},
	{0x10CAA821BE744A26ULL, "art\\characters\\playercharacter\\humanmale\\arm_bda.msh"},
	{0x10CAB572B432154FULL, "170224692.ogg"},
	{0x10CACC2DB57AFF1FULL, "tm_gwp_h01.tnt"},
	{0x10CACCBCC0C2C6CEULL, "6075154_m.fxe"},
	{0x10CAEA556085FD4CULL, "125469942.ogg"},
	{0x10CB2ACF2BA1CB46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_enter.gad"},
	{0x10CB2F2740BA2E1EULL, "224008536.ogg"},
	{0x10CB866232E222AEULL, "6150219_m.fxe"},
	{0x10CBA76FA36040FDULL, "follower_banter6167877.cl"},
	{0x10CBBE1E03074AA3ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v5.ani"},
	{0x10CC02C7ACCDEAA9ULL, "art\\characters\\weapons\\shields\\w_shd_eln.mao"},
	{0x10CC439B6CB9C662ULL, "vxht_brothel_n_4ah_0_0.vlm"},
	{0x10CC83CD59D60D7AULL, "vxdr_primevaltomb_8um_0_0.vlm"},
	{0x10CC94BB5E294B8EULL, "6175386_m.fxe"},
	{0x10CCC95C5EA120A5ULL, "vxw_bonepit_85_0_0.vlm"},
	{0x10CD05B1D1B43BD8ULL, "vxht_0_hightown_a2_c33_0_0.vlm"},
	{0x10CDCAA1A194BFCDULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryb.msh"},
	{0x10CDCE41CF9FA839ULL, "hlac_w_sundermount_-65538_0x.dds"},
	{0x10CDD8A1A194D700ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryb.mao"},
	{0x10CDDCA1A194DE53ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryb.mmh"},
	{0x10CDDDA0B95ECDD7ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01deadtree.phy"},
	{0x10CEBA866507A589ULL, "hlac_w_bonepit_0_9z.dds"},
	{0x10CEC8C7335B41AAULL, "vxlt_0_lowtown_docks_n_bcn_0_0.vlm"},
	{0x10CEDC2550CAFB89ULL, "6054691_m.fxe"},
	{0x10CF399FC372A517ULL, "23301679.ogg"},
	{0x10CF43988AB23AABULL, "hlac_lt_0_lowtown_a2_n_c_-65535_3y.dds"},
	{0x10CF555A5842A11FULL, "6149421_m.fxe"},
	{0x10CFBD26D42E9502ULL, "102077245.ogg"},
	{0x10D148A9D7DAC789ULL, "hlac_lt_gallowscourtyard_n_131071_0y.dds"},
	{0x10D225CF2BA7E125ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_enter.evt"},
	{0x10D2C8867A5076CFULL, "hlac_dr_primevaltomb_659_3x.dds"},
	{0x10D2DBAE69F8C190ULL, "gen00fl_aveline6139165.cl"},
	{0x10D2FAF0E98C4107ULL, "vrc321tr_spooky_scenes_3.utt"},
	{0x10D315B296AA9094ULL, "art\\characters\\playercharacter\\humanfemale\\arm_mede.phy"},
	{0x10D4313AC802042AULL, "art\\characters\\weapons\\staff\\textures\\w_stf_016a_s.dds"},
	{0x10D43C6B1E5F616BULL, "6149111_m.fxe"},
	{0x10D46DF153F73527ULL, "hlac_lt_0_lowtown_docks_a3_196606_0z.dds"},
	{0x10D47506FACD03B8ULL, "vxw_bonepitmining_a2_gg_0_0.vlm"},
	{0x10D4ABA4C66CCF30ULL, "6167304_f.fxe"},
	{0x10D58355F431DFCCULL, "115476244.ogg"},
	{0x10D5AAC74E6D3165ULL, "vxw_sundermount_2yi_0_0.vlm"},
	{0x10D64A1FC29EAB9AULL, "dae000_gamlen6005774.cl"},
	{0x10D727B967EC1E4CULL, "75657167.ogg"},
	{0x10D7575370F801E1ULL, "1027612044.ogg"},
	{0x10D7E198BAC8E676ULL, "vxlt_0_lowtown_gik_0_0.vlm"},
	{0x10D81F8BF8E1F648ULL, "vxlt_0_lowtown_a3_n_c_ett_0_0.vlm"},
	{0x10D84C5794272A4CULL, "hlac_ht_playershouse_n_855_5z.dds"},
	{0x10D87012CCFC49A9ULL, "960859523.ogg"},
	{0x10D9A7701B2C7DF4ULL, "6152800_m.fxe"},
	{0x10D9AE41D5FD73D0ULL, "6112930_m.fxe"},
	{0x10D9CF58708C7ED4ULL, "544175002.ogg"},
	{0x10D9D61E0313B09AULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v5.evt"},
	{0x10D9F3C6F4B7A1A5ULL, "projdecal_projdecal_p0.vsh.fxo.dx11.assembly"},
	{0x10DA059398CE0244ULL, "vxw_woundedcoast_o4_0_0.vlm"},
	{0x10DA94D93F19DCBBULL, "624105322.ogg"},
	{0x10DAB5785EAFDEE2ULL, "vxht_brothel_n_4fi_0_0.vlm"},
	{0x10DAD7C08AA6964FULL, "6174915_m.fxe"},
	{0x10DB1D371B7AD2B0ULL, "llac_lt_0_lowtown_-262144_1.dds"},
	{0x10DB4ED49C78B65DULL, "6151630_f.fxe"},
	{0x10DB6345251CFE2EULL, "nat320_delilah6065924.cl"},
	{0x10DB6C8A138E01EBULL, "191217051.ogg"},
	{0x10DBC40DF23EA9DCULL, "370393524.ogg"},
	{0x10DC235F833CABD5ULL, "6173581_m.fxe"},
	{0x10DC2D63B50C761CULL, "vxht_mansion_n_2br_0_0.vlm"},
	{0x10DC7D75B622E8B6ULL, "qcr000_varric6208237.cl"},
	{0x10DCB48F83A0CFF4ULL, "art\\characters\\creatures\\dragon\\animation\\hv_dmg.ani"},
	{0x10DCBC5549972224ULL, "123514954.ogg"},
	{0x10DCF2B296B2A991ULL, "art\\characters\\playercharacter\\humanfemale\\arm_mede.msh"},
	{0x10DD00B296B2C1ACULL, "art\\characters\\playercharacter\\humanfemale\\arm_mede.mao"},
	{0x10DD0CB296B2D64FULL, "art\\characters\\playercharacter\\humanfemale\\arm_mede.mmh"},
	{0x10DD139251F2CD86ULL, "6167460_m.fxe"},
	{0x10DDC7C25B04F480ULL, "6147608_m.fxe"},
	{0x10DDDB887104FE5CULL, "hlac_ht_0_hightown_n_131070_0z.dds"},
	{0x10DE190EC96EC867ULL, "gen00fl_fenris6105845.cl"},
	{0x10DE201E1241F5FDULL, "471135860.ogg"},
	{0x10DE567AF1DEFC9BULL, "vxht_brothel_n_3wp_0_0.vlm"},
	{0x10DF39F77ECBDBC0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edgebase.phy"},
	{0x10DF54D3EA357234ULL, "hlac_ht_0_hightown_a3_131072_0x.dds"},
	{0x10DFFFC0D6AD0806ULL, "106766541.ogg"},
	{0x10E033E94316BBD6ULL, "60194282.ogg"},
	{0x10E03B3B3CB4915BULL, "vxlt_0_lowtown_n_e5w_0_0.vlm"},
	{0x10E080F1D812862DULL, "gen_im_delete_me_03.uti"},
	{0x10E0CFAC3A4F8059ULL, "418332583.ogg"},
	{0x10E0D51E0319CD8DULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v5.gad"},
	{0x10E17672154BBCF9ULL, "6118202_f.fxe"},
	{0x10E22C5882674582ULL, "6083591_m.fxe"},
	{0x10E280574E4C1ED4ULL, "6037330_m.fxe"},
	{0x10E284433443849BULL, "6126736_m.fxe"},
	{0x10E2B18F83A53656ULL, "art\\characters\\creatures\\dragon\\animation\\hv_dmg.gad"},
	{0x10E2E6E3ECC31C38ULL, "vxw_bonepitmining_a2_bd_0_0.vlm"},
	{0x10E2F507CFCEA3B2ULL, "vxlt_0_lowtown_docks_bs5_0_0.vlm"},
	{0x10E3527341CF1A55ULL, "llac_lt_0_lowtown_docks_-131075_1.dds"},
	{0x10E36EA637531E67ULL, "gen00fl_merrill6129472.cl"},
	{0x10E446D4DEE823FCULL, "hlac_lt_0_lowtown_docks_a3_-131074_1y.dds"},
	{0x10E48178C690F8D7ULL, "vxlt_hangedman_n_2ur_0_0.vlm"},
	{0x10E523207B0BAA23ULL, "docks_male_human_worker_sta6175978.cl"},
	{0x10E541C0190EF528ULL, "6148960_m.fxe"},
	{0x10E5B405B4E27CCFULL, "vxlt_0_lowtown_a2_n_c_gdd_0_0.vlm"},
	{0x10E5E3DFD75DAEA9ULL, "6085936_m.fxe"},
	{0x10E6A4911D779229ULL, "vxw_encounter2_cp_0_0.vlm"},
	{0x10E6D519841AE4A9ULL, "216359917.ogg"},
	{0x10E6F4E5A51B02D4ULL, "vxlt_darktown_4ii_0_0.vlm"},
	{0x10E73F8E65AB5C19ULL, "hlac_lt_0_lowtown_docks_-65537_5y.dds"},
	{0x10E76F3C0EB5111FULL, "90748317.ogg"},
	{0x10E7D4A9205029FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_2windows.mao"},
	{0x10E7E0A920503D9FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_2windows.mmh"},
	{0x10E7E6A920504841ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_2windows.msh"},
	{0x10E7F3FDFF369213ULL, "mcr370_cut_player_survives6098269.cl"},
	{0x10E7F953F1358318ULL, "plt_cod_hst_dalish_culture.plo"},
	{0x10E8496193E90ED2ULL, "mag291_torpor6116228.cl"},
	{0x10E85D326F6E27C4ULL, "350498975.ogg"},
	{0x10E85E67EC167A73ULL, "framebuffereffects_cocdownsample_p0.psh.fxo.dx11"},
	{0x10E8E54801E4EAA4ULL, "llac_lt_0_lowtown_a2_n_131072_6.dds"},
	{0x10E909946C548445ULL, "vxdr_primevaltomb_8zg_0_0.vlm"},
	{0x10E92E6798A95BCFULL, "19824812.ogg"},
	{0x10E92F34C4B7F97DULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.mmh"},
	{0x10E99A3A7843A096ULL, "vxlt_gallowsprison_n_649_0_0.vlm"},
	{0x10E9B36F5DD0F02AULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.cess.gad"},
	{0x10E9CC8F83AB82B5ULL, "art\\characters\\creatures\\dragon\\animation\\hv_dmg.evt"},
	{0x10EA615D4FD9722FULL, "6125214_m.fxe"},
	{0x10EAA4A5DB0B8FDCULL, "qun221_petrice6082796.cl"},
	{0x10EABD1349E96C63ULL, "6111217_m.fxe"},
	{0x10EBF65DAB5F5154ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.impa.ani"},
	{0x10EBFDF386C2A67CULL, "6127109_m.fxe"},
	{0x10EC30D28AFD4E58ULL, "gen00fl_bethany6005722.cl"},
	{0x10ECBA6A1C2874E4ULL, "6002778_m.fxe"},
	{0x10ECD766CAEE3829ULL, "6151080_m.fxe"},
	{0x10ED92F2255698E6ULL, "6051384_f.fxe"},
	{0x10EDBEB15E6CB505ULL, "hlac_w_sundermount_131076_0z.dds"},
	{0x10EE1B34A05520BEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_basic.phy"},
	{0x10EE1F6CDF19769AULL, "w_bonepit_255_0.msh"},
	{0x10EE2D6CDF198ED3ULL, "w_bonepit_255_0.mao"},
	{0x10EE316CDF1995E0ULL, "w_bonepit_255_0.mmh"},
	{0x10EF137601C41354ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masa_0s.dds"},
	{0x10EF1B0E6DBA378BULL, "86074426.ogg"},
	{0x10EF703A949BBCE5ULL, "gen_im_cns_pot_stamina_01.uti"},
	{0x10EF764FE226BE27ULL, "566244196.ogg"},
	{0x10F0623A3EFBC620ULL, "6115636_m.fxe"},
	{0x10F094417C917465ULL, "6043453_m.fxe"},
	{0x10F0966F5DD6DDC1ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.cess.evt"},
	{0x10F0AB36A57A3AF9ULL, "oil_loop_a.dds"},
	{0x10F1002820AD5E47ULL, "dre000_sandal6033756.cl"},
	{0x10F1F56DDA00D338ULL, "watertoolset.psh.fxo"},
	{0x10F26B4F2504FFACULL, "mag131_keran6086244.cl"},
	{0x10F2F35DAB656AB6ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.impa.gad"},
	{0x10F373EBD83B53D7ULL, "6128275_m.fxe"},
	{0x10F3C02F0E17E033ULL, "vxht_playershouse_n_1uz_0_0.vlm"},
	{0x10F3F6BC43202078ULL, "301765833.ogg"},
	{0x10F3F9F657DCBBEEULL, "6149051_m.fxe"},
	{0x10F46F4A9DAC73A5ULL, "gen00fl_aveline6121897.cl"},
	{0x10F4F14225780E88ULL, "vxdr_slavers_14a_0_0.vlm"},
	{0x10F5166E0F6DAAA9ULL, "vxw_encounter2_nw_0_0.vlm"},
	{0x10F54149A5FC838FULL, "zz_followers_banter6158031.cl"},
	{0x10F54BAD7B701436ULL, "6155898_m.fxe"},
	{0x10F57ECB6D457A37ULL, "956844289.ogg"},
	{0x10F64CA5FAE84D50ULL, "vxht_mansion2b_2bm_0_0.vlm"},
	{0x10F6C842D3644FBEULL, "follower_banter6164730.cl"},
	{0x10F77508610C2590ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_local_ogre_frozen.anb"},
	{0x10F7B3E08B96CFF0ULL, "6033705_m.fxe"},
	{0x10F7DD52C333C900ULL, "322572922.ogg"},
	{0x10F7E32BD27A98EEULL, "gen00fl_isabela6015840.cl"},
	{0x10F8760E6B514B29ULL, "256466217.ogg"},
	{0x10F89E8BE9F1903AULL, "lt_0_lowtown_docks_n.rml"},
	{0x10F8E9A3144A7236ULL, "dae000_hate_enemies6167620.cl"},
	{0x10F90E5DAB6A0415ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.impa.evt"},
	{0x10F927905A6E27F7ULL, "vxdr_primevaltomb_6qb_0_0.vlm"},
	{0x10F92D5ED414A1AEULL, "gen00fl_aveline6140127.cl"},
	{0x10F930327AEB958DULL, "488251618.ogg"},
	{0x10F977AA6AF38540ULL, "6149202_m.fxe"},
	{0x10F9D816E6E954D0ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyn.msh"},
	{0x10F9E616E6E96CEAULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyn.mmh"},
	{0x10F9EA16E6E97321ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyn.mao"},
	{0x10FA021F98ADF355ULL, "hlac_lt_gallowstemplar_7198_7y.dds"},
	{0x10FA21D93E8082E5ULL, "gen00fl_aveline6115476.cl"},
	{0x10FA37EE9FFCD324ULL, "character_character_p9.psh.fxo"},
	{0x10FA55236CE13F4EULL, "6163204_m.fxe"},
	{0x10FAA235D74AD44FULL, "hlac_dr_0_deeproads_5843_5z.dds"},
	{0x10FB228B1A3C6F99ULL, "6081204_m.fxe"},
	{0x10FB800B8DFA4826ULL, "273331502.ogg"},
	{0x10FBB6F80D52C46CULL, "lt_0_lowtown_docks_n_186_0.msh"},
	{0x10FBBCF80D52CE5EULL, "lt_0_lowtown_docks_n_186_0.mmh"},
	{0x10FBC8F80D52E2F5ULL, "lt_0_lowtown_docks_n_186_0.mao"},
	{0x10FBFEF2727A43A9ULL, "vxlt_undercitydungeon_1j7_0_0.vlm"},
	{0x10FC1EC6D6AD68CAULL, "framebuffereffects_twopassblurcochorizontl_p0.psh.fxo.dx11.assembly"},
	{0x10FCAA041B34F396ULL, "pro000_soldier16040818.cl"},
	{0x10FCD50EE03E66DBULL, "6048425_m.fxe"},
	{0x10FCFB1E2981C280ULL, "vxca_cavern1_haunted_gil_0_0.vlm"},
	{0x10FD1EF920CC5B14ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_c.impa.ani"},
	{0x10FD427E97D52D47ULL, "gen_im_trs_krk_shd_met_r0.uti"},
	{0x10FD8E6F5DE15A10ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.cess.ani"},
	{0x10FDA11BDB2F3707ULL, "gen_im_wep_rog_dua_act1_05.uti"},
	{0x10FDADE7FD98BF8AULL, "6059336_m.fxe"},
	{0x10FDCF05B2249AB6ULL, "vxht_chantry_1yi_0_0.vlm"},
	{0x10FE94C049E81656ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexwild.msh"},
	{0x10FE9A3080725BF0ULL, "hlac_ca_caverns1_mines_24425_0y.dds"},
	{0x10FE9EC049E82754ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexwild.mmh"},
	{0x10FEA2C049E82E67ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexwild.mao"},
	{0x10FEEC99B280CDBDULL, "vxca_caverns2_8xf_0_1.vlm"},
	{0x10FF3CD903321D43ULL, "523387906.ogg"},
	{0x10FF6F3FF0BED963ULL, "6121009_m.fxe"},
	{0x10FFB012FF0A7618ULL, "gen00fl_aveline6132928.cl"},
	{0x10FFB7DF760C3541ULL, "ht_bartrand_n_3364.rml"},
	{0x10FFBF60AFADA866ULL, "978304317.ogg"},
	{0x110038D658DC58A6ULL, "vxw_bonepitdestroyed_a3_163_0_0.vlm"},
	{0x1100C83A0D922214ULL, "hlac_lt_gallowscourtyard_n_-65538_2y.dds"},
	{0x1100D77ADA42F47EULL, "6108752_f.fxe"},
	{0x1100D8A8AC22B592ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\shielddefence_override.anb"},
	{0x11010C322B7E7A23ULL, "411382895.ogg"},
	{0x11017EC6D3E3BFC4ULL, "64580335.ogg"},
	{0x1101E0B7D3D4EE8AULL, "6188566_m.fxe"},
	{0x1102210304B89EF7ULL, "6208687_m.fxe"},
	{0x1102328DDF519B8EULL, "plt_cod_mgc_andraste.plo"},
	{0x110253062BBA0843ULL, "169528174.ogg"},
	{0x11027E5D91551470ULL, "llac_lt_0_lowtown_-131071_1.dds"},
	{0x1102937B0F1489FAULL, "pet_so.mao"},
	{0x110297BF98CEADCFULL, "vxlt_0_lowtown_a2_n_c_gzf_0_0.vlm"},
	{0x1102B20AB6DA86D2ULL, "6014029_m.fxe"},
	{0x1103884B0163C329ULL, "vxw_encounter2_iv_0_0.vlm"},
	{0x1103BA627A1E3DEFULL, "54209894.ogg"},
	{0x1103CD682975889CULL, "hlac_lt_0_lowtown_docks_n_-262143_0z.dds"},
	{0x1103DB34B2C6859CULL, "129680488.ogg"},
	{0x11041BF920D27576ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_c.impa.gad"},
	{0x110456F77EEB2955ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edgebase.msh"},
	{0x110458F77EEB2CCBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edgebase.mmh"},
	{0x11045DAA03C85851ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargegreen_0.pwk"},
	{0x110464F77EEB4168ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edgebase.mao"},
	{0x110481EF448FD887ULL, "hlac_lt_darktown_-1_85x.dds"},
	{0x1104C245953288BFULL, "406155329.ogg"},
	{0x1104D9C581695362ULL, "107559896.ogg"},
	{0x1104FD13AC0A0F92ULL, "6048187_m.fxe"},
	{0x1105691FEAD72DECULL, "qun111_ketojan6039017.cl"},
	{0x11063E1B7FC4F695ULL, "dae000cr_grand_cleric6124700.cl"},
	{0x1106765C56B518C6ULL, "dre000_varric6207965.cl"},
	{0x11068ECB7A869FFEULL, "gen00fl_fenris6115326.cl"},
	{0x1106FC7FBE05F91CULL, "zz_mag_debug6051239.cl"},
	{0x1107CB2B95CE0B7AULL, "218850730.ogg"},
	{0x1107CBCD9E233881ULL, "hlac_ht_0_hightown_-131074_0y.dds"},
	{0x11081D4A6398B274ULL, "146025936.ogg"},
	{0x1108522EEF4F444CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3a.phy"},
	{0x1108E489FC8C7423ULL, "285320014.ogg"},
	{0x11092369EAFC7B02ULL, "6164652_m.fxe"},
	{0x1109CA06287BD90BULL, "dae000_bodhan_letter6144309.cl"},
	{0x110A44424B9AB58CULL, "hlac_ht_playershouse_n_855_14x.dds"},
	{0x110A484F6E5D6676ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.impa.evt"},
	{0x110A5B391F0D6FB8ULL, "hlac_w_encounter2_262146_0y.dds"},
	{0x110B36F920D8C1D5ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_c.impa.evt"},
	{0x110B6DA91B65CB78ULL, "122509621.ogg"},
	{0x110B9989B715E01CULL, "vxca_caverns2_9fi_0_0.vlm"},
	{0x110BF337972305CCULL, "dae000ip_crowd_lm5_talk1.utp"},
	{0x110C75C86607EC49ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\additive\\arm_r_tochest.ani"},
	{0x110CB6AD660C23E3ULL, "gen00fl_bethany6102776.cl"},
	{0x110D24F83CDA88DFULL, "6042122_m.fxe"},
	{0x110D905FA18C69E4ULL, "392628389.ogg"},
	{0x110E5D58548E9804ULL, "597023176.ogg"},
	{0x110E90405B30CD39ULL, "lla_ht_0_hightown_a2_-65535_0.dds"},
	{0x110EB5A78183C156ULL, "vxht_0_hightown_av0_0_0.vlm"},
	{0x110F3C12A87483F6ULL, "482833819.ogg"},
	{0x110F6A7FC605321AULL, "6081208_f.fxe"},
	{0x110F776E6D0E642BULL, "llac_w_blightlands_-65538_0.dds"},
	{0x110F913EEF0921ACULL, "vxht_0_hightown_c5a_0_0.vlm"},
	{0x110FB29A9EA0C1C4ULL, "gen00fl_aveline6139716.cl"},
	{0x11103F4F6E61C201ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.impa.gad"},
	{0x111049061158C603ULL, "6201514_m.fxe"},
	{0x111064EE7233EF6BULL, "s_ballista_03_l2.mao"},
	{0x1110935A91F530AFULL, "cli600cr_hurlock_ranged.utc"},
	{0x1110B608101C4303ULL, "6135224_m.fxe"},
	{0x1110EC766864C0EAULL, "dae000_gamlen6088135.cl"},
	{0x11113473048894E9ULL, "hlac_lt_backalley_g_0_9z.dds"},
	{0x111143F6EA74D488ULL, "mcr371_orsino.cnv"},
	{0x111158C45E43A558ULL, "gen00fl_aveline6122110.cl"},
	{0x11119644FC061E4BULL, "6124679_m.fxe"},
	{0x1111AACD932F5ECFULL, "ach_kill_rock.dds"},
	{0x1111C8F0F121CC3BULL, "vxca_caverns1_gk5_0_1.vlm"},
	{0x111201DFD592F6BFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_angrydebate_1_darren.gad"},
	{0x111236E93C77CB8CULL, "6031985_m.fxe"},
	{0x111290C8660C85CBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\additive\\arm_r_tochest.gad"},
	{0x1112EB24C4C7E8A6ULL, "792789914.ogg"},
	{0x1112F737ED7F8601ULL, "6107559_m.fxe"},
	{0x1113272A9B3AAB34ULL, "56470116.ogg"},
	{0x1113B08FE316BE14ULL, "mcr371_hostile_anders6082105.cl"},
	{0x1113B7D15B102606ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v2.ani"},
	{0x11140D26899693D1ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.cess.ani"},
	{0x1114AA7D71730A90ULL, "nat380_sibling6103577.cl"},
	{0x1114DA59B77BC43EULL, "vxlt_0_lowtown_docks_n_c59_0_0.vlm"},
	{0x1114F4DCEC8A8B91ULL, "6040834_m.fxe"},
	{0x111515DFA6D823D5ULL, "vxdr_0_deeproads_cf6_0_0.vlm"},
	{0x111565F6EA78F7F0ULL, "mcr371_orsino.bnk"},
	{0x11157670168FF6ACULL, "919856686.ogg"},
	{0x1115B2A0B99C3A2FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01deadtree.mao"},
	{0x1115BEA0B99C4ECCULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01deadtree.mmh"},
	{0x1115C4A0B99C58BEULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01deadtree.msh"},
	{0x1115E020CD72E751ULL, "vxlt_darktown_5t6_0_0.vlm"},
	{0x11162C1D3C48B3DDULL, "65163093.ogg"},
	{0x111638BD01C0DB50ULL, "llac_lt_0_lowtown_a2_n_c_1_1.dds"},
	{0x11163E8056D02504ULL, "vxw_sundermount_2g2_0_0.vlm"},
	{0x11166A4A53A60516ULL, "vxca_caverns1_j2b_0_1.vlm"},
	{0x1116B7F125031AE9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls106a_0d.dds"},
	{0x11172831DAF70748ULL, "seb321_leliana6180967.cl"},
	{0x1117404F6E67E2AFULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.impa.ani"},
	{0x11176A8F58552035ULL, "650696229.ogg"},
	{0x111822AD7F477208ULL, "739854136.ogg"},
	{0x111869FFB726BBDCULL, "w_woundedcoast_39_0.mao"},
	{0x111875FFB726D07FULL, "w_woundedcoast_39_0.mmh"},
	{0x1118792331963DE6ULL, "6120762_m.fxe"},
	{0x11187BFFB726DA21ULL, "w_woundedcoast_39_0.msh"},
	{0x1118FA929EF96E8BULL, "mag170_thrask6020545.cl"},
	{0x11192855F8D96D8AULL, "6068422_m.fxe"},
	{0x11193B1C75E75CB2ULL, "421903520.ogg"},
	{0x11198DC866129F30ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\additive\\arm_r_tochest.evt"},
	{0x1119D1CF7D632C70ULL, "hlac_lt_0_lowtown_a2_-1_3x.dds"},
	{0x1119E57252C835D4ULL, "hlac_lt_0_lowtown_docks_a3_-131075_1z.dds"},
	{0x111A1826899B1273ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.cess.gad"},
	{0x111A30D054903264ULL, "llac_ht_0_hightown_-65534_1.dds"},
	{0x111ABA61AFF44B2AULL, "ave331_jeven6152174.cl"},
	{0x111AE2875786293CULL, "mrl351cr_fakemaleelf.utc"},
	{0x111B42D37F9A5854ULL, "hlac_lt_undercityencounter_-1_6z.dds"},
	{0x111B5EF86881FD46ULL, "vxlt_0_lowtown_docks_n_b9p_0_0.vlm"},
	{0x111B92A201E4BB2AULL, "ser270_emeric6149324.cl"},
	{0x111BBBC9AD2C41ADULL, "w_blightlands_48_0.mmh"},
	{0x111BBFC9AD2C48FEULL, "w_blightlands_48_0.mao"},
	{0x111BCDC9AD2C6033ULL, "w_blightlands_48_0.msh"},
	{0x111BE54DD89F46CDULL, "6079280_f.fxe"},
	{0x111C02B9DD4191E8ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall_tool.msh"},
	{0x111C06DC638B6825ULL, "vxlt_0_lowtown_gj5_0_0.vlm"},
	{0x111C14B9DD41B079ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall_tool.mao"},
	{0x111C20B9DD41C412ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall_tool.mmh"},
	{0x111C2BE204BCEB71ULL, "6079019_f.fxe"},
	{0x111C44C222BF9C68ULL, "vxw_bonepitdestroyed_a3_9k_0_0.vlm"},
	{0x111C6B7F286BCC67ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sidetable_n.dds"},
	{0x111C888BF357C94BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_loop.ani"},
	{0x111D0163A49316DFULL, "684258576.ogg"},
	{0x111D1A7D11D5B1E2ULL, "rdr110dg_guard_warehouse6016707.cl"},
	{0x111D568D7B718959ULL, "28056367.ogg"},
	{0x111D79DF0CC414B5ULL, "zz_vault_debug6177759.cl"},
	{0x111DD9994CDF7D61ULL, "6118080_f.fxe"},
	{0x111DE447A03C789AULL, "vxlt_0_lowtown_a2_n_c_e2m_0_0.vlm"},
	{0x111E30F3B0CE150CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkb_n.dds"},
	{0x111E365D2EE229C1ULL, "vxlt_0_lowtown_ffx_0_0.vlm"},
	{0x111E4E338A468930ULL, "179973343.ogg"},
	{0x111E83D280148695ULL, "292595847.ogg"},
	{0x111EE04ADAB749FDULL, "vxht_0_hightown_a3_a7y_0_0.vlm"},
	{0x111FD8A1BCEC7C00ULL, "6045999_m.fxe"},
	{0x11200035B64692D0ULL, "6169868_m.fxe"},
	{0x1120179451A32D93ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em02.phy"},
	{0x1120468966FAAE16ULL, "vxht_0_hightown_a3_bfn_0_0.vlm"},
	{0x112093403224F6D3ULL, "dae100_viscount6118510.cl"},
	{0x1121049497218FF7ULL, "02_fire.dds"},
	{0x1121152689A12BD8ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.cess.evt"},
	{0x11212966DB36A06AULL, "one100_solivitus6048107.cl"},
	{0x11213160A2AEB4DEULL, "ip_restrict_style_staff.ncs"},
	{0x11219FD15B1C3B77ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v2.evt"},
	{0x11223EE4529C7CC4ULL, "vxw_woundedcoast_p9_0_0.vlm"},
	{0x1122AD2D456B72A2ULL, "character_character_p285.psh.fxo.dx11"},
	{0x1122C5625987D22BULL, "vxlt_gallowscourtyard_n_7gj_0_0.vlm"},
	{0x112389AEE3D2A1C8ULL, "178410830.ogg"},
	{0x112465BB34C053D7ULL, "tbe100_xenon6190998.cl"},
	{0x112496107329BE2BULL, "6042144_m.fxe"},
	{0x1124D67AD8120098ULL, "follower_barks6173407.cl"},
	{0x1124E1161A0786BCULL, "6189031_m.fxe"},
	{0x1124EB85D14C651FULL, "6076153_f.fxe"},
	{0x1124EF4CBC56877EULL, "322260179.ogg"},
	{0x11253120369B4CE7ULL, "355318965.ogg"},
	{0x11255C10273E8244ULL, "vxw_bonepit_1x3_0_0.vlm"},
	{0x1125BBC9AD349626ULL, "w_blightlands_48_0.rml"},
	{0x1125DA1BACB047C7ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.cess.gad"},
	{0x1126088972A72F66ULL, "686449346.ogg"},
	{0x11261905EE10A48BULL, "tbe000im_ancient_chest.uti"},
	{0x112673004762447EULL, "376905316.ogg"},
	{0x1126DEDFD5A51705ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_angrydebate_1_darren.ani"},
	{0x11274773F1CC7D9BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantsml.mmh"},
	{0x11275373F1CC9238ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantsml.mao"},
	{0x11276573F1CCB085ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantsml.msh"},
	{0x11276D5AE9EC4508ULL, "395378432.ogg"},
	{0x1127A0D15B20A800ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v2.gad"},
	{0x1127C79627830872ULL, "6127486_m.fxe"},
	{0x1127E676D49B97ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_shield.msh"},
	{0x1127EC76D49BA19EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_shield.mmh"},
	{0x1127F876D49BB635ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_shield.mao"},
	{0x11283E6E36DD224BULL, "zz_amb_beggar.utc"},
	{0x1128478467E104EAULL, "vxlt_0_lowtown_a3_n_c_hdc_0_0.vlm"},
	{0x1128E047CBA99779ULL, "dae000_tranquil_1.bnk"},
	{0x1128F12EDE8700F3ULL, "llac_ca_caverns1_24421_7.dds"},
	{0x1129233AB8F89956ULL, "t3_shl_grn.tnt"},
	{0x1129499092AEC055ULL, "984315696.ogg"},
	{0x1129A5715FCA5746ULL, "dre000_varric6029149.cl"},
	{0x1129D0D55A7815C6ULL, "vxlt_0_lowtown_docks_n_b6q_0_0.vlm"},
	{0x112A27DBCBC6AE91ULL, "hlac_dr_0_deeproads_5845_5z.dds"},
	{0x112A5DFB8A78D987ULL, "vis121_saemus6089667.cl"},
	{0x112B141EC9A6008DULL, "575656980.ogg"},
	{0x112BB7AEFFE30AA6ULL, "vxw_bonepit_nd_0_0.vlm"},
	{0x112BC282435A2C27ULL, "6135762_f.fxe"},
	{0x112C0247CBAC06D9ULL, "dae000_tranquil_1.cnv"},
	{0x112C9B0ED9DEE39BULL, "w_bonepitdestroyed_a3_52_0.mmh"},
	{0x112CA70ED9DEF838ULL, "w_bonepitdestroyed_a3_52_0.mao"},
	{0x112CB90ED9DF1685ULL, "w_bonepitdestroyed_a3_52_0.msh"},
	{0x112CCF1BACB653D4ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.cess.evt"},
	{0x112CF7B6999C8326ULL, "vxca_cavern1_haunted_ii0_0_0.vlm"},
	{0x112D62473EB8D5C1ULL, "506594307.ogg"},
	{0x112DE601B2E58311ULL, "vxlt_darktown_2rw_0_0.vlm"},
	{0x112E21DE58465596ULL, "6036991_m.fxe"},
	{0x112E53892044E93DULL, "lgt220_shp_mag_guard6149521.cl"},
	{0x112F1E17E9C3172FULL, "135874654.ogg"},
	{0x112F2945A87D3997ULL, "dae000_epic_merchant_36179506.cl"},
	{0x112F5A32AD5A2DCBULL, "364452715.ogg"},
	{0x112F6AC16E9CFAD1ULL, "vxlt_gallowscourtyard_n_76k_0_0.vlm"},
	{0x112F78A3D515E153ULL, "vxlt_warehouse_n_sx_0_0.vlm"},
	{0x112FE27C9B593116ULL, "sp_module_item_lost.ncs"},
	{0x11306F8BF36847D5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_loop.gad"},
	{0x1130AF66F1DC9F2CULL, "ser121_jethann6104552.cl"},
	{0x1130CF389B005B2FULL, "50291781.ogg"},
	{0x1131738CEA5563A9ULL, "6018670_f.fxe"},
	{0x11318C9627C33D7EULL, "lt_0_lowtown_53_0_ll.mmh"},
	{0x1131CF639FC46A9CULL, "plt_fen100pt_recruited.nss"},
	{0x1131D1F5CCE7DA8FULL, "llac_lt_0_lowtown_docks_a3_-4_5.dds"},
	{0x1131ECF0505374F5ULL, "gen00fl_aveline6105437.cl"},
	{0x1132253C586DBDF8ULL, "fighting70a.mao"},
	{0x11327F8BEA22D39CULL, "lt_0_lowtown_docks_n.arl"},
	{0x113296DF763756D1ULL, "ht_bartrand_n_3364.anb"},
	{0x1132A046681255AFULL, "vxlt_warehouse_10a_0_0.vlm"},
	{0x1132F6F069E1B332ULL, "6175070_m.fxe"},
	{0x1133223A3ABC5604ULL, "vxw_sundermount_2m0_0_0.vlm"},
	{0x11332A852B4965A6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen2.evt"},
	{0x113349CA84E05871ULL, "hlac_w_blightlands_65539_1y.dds"},
	{0x113622CA4B99FCFEULL, "vxca_caverns1_mines_icx_0_2.vlm"},
	{0x11364EB4B4802A4DULL, "vxlt_0_lowtown_n_eoc_0_0.vlm"},
	{0x113672D9C724708EULL, "468377537.ogg"},
	{0x1136A4B75C324E67ULL, "dae150_chandan6148796.cl"},
	{0x1136B9A3D1E9CFE5ULL, "6209853_m.fxe"},
	{0x1136C5DA357316BEULL, "6116231_m.fxe"},
	{0x1137193258BC7B76ULL, "w_bonepitmining_a2_16_0.rml"},
	{0x11372D2F437D790CULL, "430181720.ogg"},
	{0x1137510DCC5B75CAULL, "6108746_m.fxe"},
	{0x113770309E092283ULL, "vxlt_undercitydungeon_190_0_0.vlm"},
	{0x1137AAF332266ADFULL, "dre000_vampire6167443.cl"},
	{0x1137D14CE1825A9DULL, "vxlt_0_lowtown_a2_gne_0_0.vlm"},
	{0x1137F249C30C006CULL, "1034369215.ogg"},
	{0x11385665AF7989E3ULL, "vxca_cavern1_haunted_fmz_0_0.vlm"},
	{0x1138AB54CD865F18ULL, "dae000_bartrand6027550.cl"},
	{0x1138E18E3D2D7080ULL, "46845371.ogg"},
	{0x11392CA9A2605A1CULL, "6059428_m.fxe"},
	{0x113941852B4DF8F1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen2.gad"},
	{0x113973E894BAFA4EULL, "vxw_bonepitmining_a2_169_0_0.vlm"},
	{0x1139BCCE9E8DDA66ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg102a_0d.dds"},
	{0x1139C71BACC0D00DULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.cess.ani"},
	{0x113A1561545817E2ULL, "173653223.ogg"},
	{0x113B0A7AB3E9609CULL, "follower_banter6161090.cl"},
	{0x113BA605E407BA65ULL, "475161832.ogg"},
	{0x113BB7BEA3B48E02ULL, "vxdr_0_deeproads_eha_0_0.vlm"},
	{0x113C2D3D6DBAF377ULL, "plt_cod_ltr_orsino_thank.nss"},
	{0x113C9CF2D269E743ULL, "hlac_lt_0_lowtown_131073_0y.dds"},
	{0x113CFCD74A775445ULL, "6042488_f.fxe"},
	{0x113D664B94A9647CULL, "vxw_encounter2_bj_0_0.vlm"},
	{0x113D674E565EFC8CULL, "vxlt_hangedman_n_1xi_0_0.vlm"},
	{0x113E16A7EBF38E21ULL, "vxlt_warehouse_n_ad_0_0.vlm"},
	{0x113E6436B94995FAULL, "vxw_woundedcoast_ne_0_0.vlm"},
	{0x113FD645CAD5B6FCULL, "vxht_mansion2_1bt_0_2.vlm"},
	{0x113FFAD2730387AEULL, "81754052.ogg"},
	{0x11401D28AAFDB8E3ULL, "6041747_m.fxe"},
	{0x114042852B54189FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen2.ani"},
	{0x114056FFBEB58CCFULL, "dae000ip_crowd_ff8_talk4.utp"},
	{0x1140F125CA703005ULL, "llac_lt_0_lowtown_a2_n_c_-1_3.dds"},
	{0x1141745F2995B37FULL, "follower_banter6172972.cl"},
	{0x114176FFB749D6BCULL, "w_woundedcoast_39_0.anb"},
	{0x1141955AA51F3F66ULL, "vxw_encounter2_5i_0_0.vlm"},
	{0x1142542531EFD27BULL, "vxlt_warehouse_r6_0_0.vlm"},
	{0x11428AEABE8887ABULL, "gen00fl_sebastian6177919.cl"},
	{0x1143B64F0AD5F0EFULL, "follower_barks6153439.cl"},
	{0x11442B4AEEFC558DULL, "6160866_m.fxe"},
	{0x11445D2DE106C433ULL, "756817694.ogg"},
	{0x114504D19025E253ULL, "732050374.ogg"},
	{0x114552913D81B948ULL, "em_orsino.mor"},
	{0x114574215A77568BULL, "vxca_caverns1_dnh_0_0.vlm"},
	{0x1145B9A75923B76CULL, "mer341_earl6054259.cl"},
	{0x1145DBF333727849ULL, "158126299.ogg"},
	{0x11466E91AC443A94ULL, "hlac_lt_0_lowtown_docks_a3_131070_3z.dds"},
	{0x114675F3FB057A7AULL, "vxht_0_hightown_a2_c75_0_0.vlm"},
	{0x11476FBB6DE565D8ULL, "vxca_cavern1_haunted_ck2_0_0.vlm"},
	{0x1148292EEF84E217ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3a.mmh"},
	{0x11482D2EEF84E924ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3a.mao"},
	{0x11483F2EEF850789ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3a.msh"},
	{0x11484DB5F24D34CDULL, "363188930.ogg"},
	{0x114855DF4955AAD4ULL, "496896455.ogg"},
	{0x114877035AB95349ULL, "pro000_friedrich6050185.cl"},
	{0x1148BEB188BE9A6CULL, "6018558_f.fxe"},
	{0x1148E8DE04B9B174ULL, "ave200_donnic6110769.cl"},
	{0x114A30C0B2D1F6AEULL, "vxht_0_hightown_a3_c3l_0_0.vlm"},
	{0x114AA90CEDF7C260ULL, "6119698_m.fxe"},
	{0x114AB27B7FBE5031ULL, "6059991_m.fxe"},
	{0x114ACFB000A416CDULL, "467915841.ogg"},
	{0x114AE9A920A4CAE4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_2windows.phy"},
	{0x114BA874CFAAAC62ULL, "6147909_m.fxe"},
	{0x114BB983FD3F99F3ULL, "6050179_f.fxe"},
	{0x114BD828869F7CFCULL, "vxw_encounter2_mk_0_0.vlm"},
	{0x114C05BA21CEA2CAULL, "hlac_ht_0_hightown_a2_65535_0y.dds"},
	{0x114C52C88060B840ULL, "hlac_w_encounter3_-131074_0x.dds"},
	{0x114CD613AB3FAE7AULL, "vxw_woundedcoast_kf_0_0.vlm"},
	{0x114CD93D4C073730ULL, "vxlt_0_lowtown_a2_n_c_gcn_0_0.vlm"},
	{0x114D5C0704E8EAC0ULL, "107933667.ogg"},
	{0x114D7845CFA567BCULL, "follower_banter6149247.cl"},
	{0x114E2CBABD6550FFULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extfloor.dds"},
	{0x114F84BA6E5325C6ULL, "6129721_m.fxe"},
	{0x114FD5C894234B77ULL, "gen00fl_merrill6130135.cl"},
	{0x1150124C72978CF1ULL, "396806395.ogg"},
	{0x1150434F07ACF2AAULL, "follower_banter6155514.cl"},
	{0x1150A31AB93F946CULL, "405690248.ogg"},
	{0x1150AE1177DD3740ULL, "lowtown_fem_alienage1_sta6162564.cl"},
	{0x1150B1A6FD20CBC4ULL, "vxw_bonepit_1k6_0_0.vlm"},
	{0x1150CC7BFA62EF63ULL, "472364353.ogg"},
	{0x1150EBE119B98DACULL, "867814080.ogg"},
	{0x1151B9768C3F5E0FULL, "6041219_m.fxe"},
	{0x1152070B5E8CF6D4ULL, "zz_dae_debug6029905.cl"},
	{0x11520B97B2FAD128ULL, "hlac_lt_foundry_n_320_14x.dds"},
	{0x11521172C8ADCCD7ULL, "5952820.ogg"},
	{0x11522AFC04FA89E2ULL, "952238458.ogg"},
	{0x1152509AD783D13BULL, "vxlt_0_lowtown_docks_a3_d01_0_1.vlm"},
	{0x11528EF3B56C7B1EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgtb_0s.dds"},
	{0x1153185E37EDA7C6ULL, "1071436600.ogg"},
	{0x1153D5F49B62C41BULL, "dae000_meredith6085249.cl"},
	{0x1154A654F5723863ULL, "fen200ar_ambush_exterior.are"},
	{0x1154C02C6F327917ULL, "6017667_m.fxe"},
	{0x115565D151E3B4D3ULL, "859068641.ogg"},
	{0x1155B022498CCB33ULL, "hm_gen_high_mage.mor"},
	{0x1155D91ADE7B1B05ULL, "6217700_m.fxe"},
	{0x11561544473BEA58ULL, "vxht_keep_n_70e_0_0.vlm"},
	{0x1156904A5B445419ULL, "llac_ht_bartrand_n_3366_0.dds"},
	{0x1157356D6CCA8260ULL, "zz_ser_debug6085774.cl"},
	{0x1157565D24864BC0ULL, "402958725.ogg"},
	{0x11576E6CD86FEE29ULL, "vxht_brothel_n_6ee_0_0.vlm"},
	{0x1157A7461598EFF7ULL, "6139674_f.fxe"},
	{0x1157DF4640BDE11DULL, "ser191_emeric6120416.cl"},
	{0x115818808932747FULL, "752111606.ogg"},
	{0x1159245871B468B1ULL, "6056891_m.fxe"},
	{0x1159C4AB887952CDULL, "hlac_lt_gallowstemplar_2388_4z.dds"},
	{0x1159D6373A2FD22EULL, "dae000cr_grand_cleric6182015.cl"},
	{0x1159DD78867FAB1AULL, "hlac_ht_0_hightown_a2_-65536_2y.dds"},
	{0x115A0F83D9E742D4ULL, "6050409_m.fxe"},
	{0x115A4A057895957CULL, "vxw_encounter2_hh_0_0.vlm"},
	{0x115A4B083A4B2D8CULL, "vxlt_hangedman_n_1ro_0_0.vlm"},
	{0x115A74CCBF4BCE1DULL, "325821336.ogg"},
	{0x115A86D4D76C8AA9ULL, "vxw_blightlands_7s_0_0.vlm"},
	{0x115A8AC7DB8A07D2ULL, "vxlt_darktown_4qg_0_0.vlm"},
	{0x115A8D5DFA39A3C0ULL, "885573967.ogg"},
	{0x115A9DC04A36844FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexwild.phy"},
	{0x115AF8D5C6BD73A0ULL, "vxw_blightlands_j1_0_0.vlm"},
	{0x115B15BF399BF77EULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.dura.ani"},
	{0x115B2C8C80D31FCBULL, "674832085.ogg"},
	{0x115B34CEDE86CF5CULL, "vxdr_primevaltomb_82c_0_1.vlm"},
	{0x115B8913513063A2ULL, "art\\characters\\playercharacter\\elfmale\\har_fls.phy"},
	{0x115C8780235A2836ULL, "pro000_colonel_ewald6047380.cl"},
	{0x115D2460E7590DD7ULL, "vxlt_darktown_5ak_0_0.vlm"},
	{0x115D314B2BE3FBF0ULL, "623505397.ogg"},
	{0x115D547885B179D0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1b.mao"},
	{0x115D587885B180E3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1b.mmh"},
	{0x115D667885B1983DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1b.msh"},
	{0x115DAFB6034CB820ULL, "vxlt_0_lowtown_a3_n_c_bza_0_0.vlm"},
	{0x115DCF2653E38EA6ULL, "6088855_m.fxe"},
	{0x115DFA1EFF64385AULL, "vxlt_foundry_n_zl_0_0.vlm"},
	{0x115E9E3DF665ED60ULL, "vxlt_0_lowtown_a2_n_d06_0_0.vlm"},
	{0x115EBD5A88BE407BULL, "hlac_lt_0_lowtown_docks_a3_n_-131072_0z.dds"},
	{0x115EDBE0AD26EDB4ULL, "6173618_m.fxe"},
	{0x115F2383EF16E444ULL, "vxw_bonepit_1l7_0_0.vlm"},
	{0x115F54115A753DE2ULL, "vxdr_primevaltomb_34q_0_0.vlm"},
	{0x115FC82323B9C7C7ULL, "186872759.ogg"},
	{0x115FEFCACA3FCCD8ULL, "gen00fl_anders6125769.cl"},
	{0x1160202C6CAB5980ULL, "art\\vfx\\materials\\vfx_d_shockwave_blend.mao"},
	{0x1160803B6C251935ULL, "character_character_p150.psh.fxo.dx11.assembly"},
	{0x11610CA368FEE632ULL, "778434797.ogg"},
	{0x11611345D82499DAULL, "framebuffereffects_downsample_p0.psh.fxo.dx11.assembly"},
	{0x11611705AE6081D8ULL, "genip_cin_arrow.utp"},
	{0x1161362CB8C790F5ULL, "6077230_m.fxe"},
	{0x116178794FEB3430ULL, "zz_prmpt_debug6200934.cl"},
	{0x1161B99DE545F664ULL, "vxlt_0_lowtown_bz9_0_0.vlm"},
	{0x116257DB3E63878BULL, "vxca_caverns1_dpj_0_0.vlm"},
	{0x116274EFFBBBB5A3ULL, "6142763_m.fxe"},
	{0x1163108E63B4D344ULL, "226865289.ogg"},
	{0x1163401D555D9756ULL, "hlac_lt_0_lowtown_docks_n_-4_4x.dds"},
	{0x116359ADDEF1AB7AULL, "vxht_0_hightown_a2_c97_0_0.vlm"},
	{0x1163AC35A119D333ULL, "22830499.ogg"},
	{0x1163F10C5DEF197DULL, "art\\characters\\heraldry\\hld_seekers.dds"},
	{0x1164107B2A804F73ULL, "follower_barks6153563.cl"},
	{0x11641ADAF2335DDDULL, "vxlt_darktown_5cc_0_0.vlm"},
	{0x11646ABF258FCFEAULL, "153050626.ogg"},
	{0x116491D12480BEB2ULL, "115604014.ogg"},
	{0x1164E7FF10B634B4ULL, "vxlt_gallowstemplar_f_5jf_0_0.vlm"},
	{0x1164F24F96A2FA0DULL, "hairalpha_hair_p0.psh.fxo"},
	{0x11655A11B75572C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_eb_transition.evt"},
	{0x11656A56216F938DULL, "vxlt_0_lowtown_a2_n_gz2_0_0.vlm"},
	{0x116699017FE6E9B1ULL, "1044349752.ogg"},
	{0x11677224A81F5B83ULL, "604170302.ogg"},
	{0x1167A399639A3935ULL, "dae000_unhappy_mage.bnk"},
	{0x1167E2C120FB5316ULL, "ico_helm_heavy_5.dds"},
	{0x1167E765795EF22AULL, "art\\characters\\playercharacter\\humanmale\\_hm_hlf_masb.mao"},
	{0x11683E759C0133B5ULL, "115176259.ogg"},
	{0x1168574D4E93D281ULL, "6053489_m.fxe"},
	{0x11687D199FD532AEULL, "6205674_m.fxe"},
	{0x1168C376D4D2F339ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_shield.phy"},
	{0x1168F8B1C962A329ULL, "vxw_blightlands_8r_0_0.vlm"},
	{0x116919FC1B5E5204ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls103a_0d.dds"},
	{0x11692DBF39A85D4FULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.dura.evt"},
	{0x11695BDABDBE2545ULL, "vxw_woundedcoast_nn_0_0.vlm"},
	{0x1169823339335241ULL, "6208471_f.fxe"},
	{0x1169830A24647C18ULL, "ave200_donnic6110633.cl"},
	{0x1169AFB4DDAB1E52ULL, "6068695_m.fxe"},
	{0x1169BCF72FF36830ULL, "vxlt_0_lowtown_a2_n_c_gyl_0_0.vlm"},
	{0x116AC599639CA88DULL, "dae000_unhappy_mage.cnv"},
	{0x116B963DD94F2657ULL, "vxlt_darktown_5jj_0_0.vlm"},
	{0x116C009F5D93139AULL, "lav_river03a.dds"},
	{0x116C3311B75B4F44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_eb_transition.gad"},
	{0x116C410345D9A726ULL, "6083955_m.fxe"},
	{0x116CEE53261535E3ULL, "6148372_m.fxe"},
	{0x116CF91159325A2BULL, "vxht_keep_6v6_0_0.vlm"},
	{0x116CFFDC877D25ABULL, "6013843_m.fxe"},
	{0x116D392CDF99076DULL, "785948541.ogg"},
	{0x116D3E4DA545BD66ULL, "vxw_blightlands_ha_0_0.vlm"},
	{0x116D71C3ECF8C547ULL, "6084340_m.fxe"},
	{0x116DC0DC829F7428ULL, "vxht_bartrand_n_6qt_0_0.vlm"},
	{0x116EF2E15263355AULL, "character_character_p166.psh.fxo.dx11.assembly"},
	{0x116F17660A37813DULL, "vxht_playershouse_n_1zy_0_0.vlm"},
	{0x116F853E3984D003ULL, "vxca_cavern1_haunted_e8n_0_0.vlm"},
	{0x116F8DCEF53FA6D1ULL, "vxw_encounter3_fm_0_1.vlm"},
	{0x116FA3236C83ADA3ULL, "ave310_brennan6136882.cl"},
	{0x116FA809AABDA975ULL, "6078330_m.fxe"},
	{0x116FBE979511F372ULL, "vxht_bartrand_n_735_0_0.vlm"},
	{0x116FBFE16A243ABBULL, "mer240cr_miner3.utc"},
	{0x116FC4059D74045AULL, "vxlt_gallowstemplar_f_5qe_0_0.vlm"},
	{0x116FE3ED793BF0F5ULL, "57148215.ogg"},
	{0x11702EBF39AE7DF8ULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.dura.gad"},
	{0x11708DE8D8485A19ULL, "vxlt_0_lowtown_docks_a3_czg_0_0.vlm"},
	{0x11709920917739A9ULL, "hlac_dr_primevaltomb_664_5x.dds"},
	{0x1170E1731B0CC917ULL, "vxlt_gallowstemplar_3qu_0_0.vlm"},
	{0x1171243312DC1A39ULL, "hlac_lt_undercitydungeon_-1_24z.dds"},
	{0x117154813E042AAFULL, "6173019_m.fxe"},
	{0x1171583EED56AFA4ULL, "zz_eric_soundtst.cut"},
	{0x11715ABF6B3D3D9CULL, "mcr371_scared_mage_m6086908.cl"},
	{0x1171C1C65AD80B22ULL, "hlac_dr_0_deeproads_5838_0z.dds"},
	{0x117210FD7BB546CEULL, "gen00fl_isabela6096801.cl"},
	{0x11723A862DC58C99ULL, "6081415_m.fxe"},
	{0x11724211B75FD422ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_eb_transition.ani"},
	{0x1172F5CEAB8CA4DFULL, "art\\characters\\playercharacter\\humanfemale\\boo_medd.phy"},
	{0x1172FE9AA90D6DCEULL, "vxht_0_hightown_a2_cbc_0_1.vlm"},
	{0x117359DC02AC4D34ULL, "vxlt_gallowstemplar_f_5mg_0_0.vlm"},
	{0x11745226BC5C1F29ULL, "vxht_brothel_n_6kk_0_0.vlm"},
	{0x117457CFB91EC029ULL, "llac_lt_0_lowtown_-1_2.dds"},
	{0x11745FC895CFFB27ULL, "dae000_arishok6072820.cl"},
	{0x1174B25FCB38FE0EULL, "vxw_sundermount_371_0_0.vlm"},
	{0x1174E4325AF49712ULL, "498147362.ogg"},
	{0x1174F9BC7119CF4EULL, "vxdr_0_deeproads_ewd_0_0.vlm"},
	{0x11755ED1653C9883ULL, "gen00fl_anders6125478.cl"},
	{0x11756489553EBB49ULL, "hlac_lt_0_lowtown_docks_a3_65532_1x.dds"},
	{0x1175834FFC2E8049ULL, "266667253.ogg"},
	{0x11761F18FAEA3C0AULL, "61332704.ogg"},
	{0x117676AFFDDB148DULL, "vxca_caverns2_8g7_0_1.vlm"},
	{0x1176B3442B0CE6ACULL, "art\\cinematics\\additional animation\\cs_femnoblei_cinematicanim.gad"},
	{0x1176E7BBE914F4EFULL, "188291091.ogg"},
	{0x1177A21D1F3A3B42ULL, "ser270_cut_aftermath6162279.cl"},
	{0x1177CDB7AFB43DC5ULL, "vxw_woundedcoast_km_0_0.vlm"},
	{0x1177D44299569A11ULL, "designertags.gda"},
	{0x1177F40A2C2EBF61ULL, "gen_im_npc_wep_shd_gena.uti"},
	{0x117800FD81BF4990ULL, "6104098_m.fxe"},
	{0x11784D6A7C8F29D0ULL, "vxht_mansion2b_27r_0_0.vlm"},
	{0x117888BC91BC1D98ULL, "hlac_ca_caverns1_24424_27z.dds"},
	{0x1178AE6AB3194421ULL, "6117284_m.fxe"},
	{0x1178C4351DB1FEE4ULL, "vxht_0_hightown_a2_a7u_0_0.vlm"},
	{0x1178E52D646F5F66ULL, "ser221_gascard6149263.cl"},
	{0x1178F6522D4F76A8ULL, "lgt150_nat_were_supporter6111009.cl"},
	{0x117969FADA6DE59AULL, "6110612_f.fxe"},
	{0x117A081ACB453ED7ULL, "vxlt_darktown_5oe_0_0.vlm"},
	{0x117A9B0EDA210528ULL, "w_bonepitdestroyed_a3_52_0.rml"},
	{0x117B06EE59F7C2E3ULL, "vxlt_0_lowtown_n_byt_0_0.vlm"},
	{0x117B3B0C2DDAA9BDULL, "hlac_lt_0_lowtown_docks_n_-2_5z.dds"},
	{0x117B47B34AC56F7CULL, "dae011_badpoet_sta6145184.cl"},
	{0x117B48D655C3FD7AULL, "vxw_bonepit_1vo_0_0.vlm"},
	{0x117B58380C9113B7ULL, "zz_rain_o_arrows.cut"},
	{0x117BB02A973BD5E6ULL, "vxw_blightlands_mb_0_0.vlm"},
	{0x117C9875D56E5F58ULL, "gen00fl_isabela6090320.cl"},
	{0x117C99D43288845BULL, "6005822_m.fxe"},
	{0x117D8942FC2D99BDULL, "vxht_playershouse_n_1qx_0_0.vlm"},
	{0x117DB2442B1303AAULL, "art\\cinematics\\additional animation\\cs_femnoblei_cinematicanim.ani"},
	{0x117DB280147FECE0ULL, "mage_petrify.dds"},
	{0x117E45268DB3D582ULL, "vxlt_darktown_585_0_0.vlm"},
	{0x117E97616450CA89ULL, "lgt160cr_ntb_elf_dalish.utc"},
	{0x117EFE512994736FULL, "6041549_m.fxe"},
	{0x117F2796D519BFD8ULL, "6031743_m.fxe"},
	{0x117F4693F2A7AD8EULL, "hlac_ca_caverns2_5802_57y.dds"},
	{0x117F591C179DF60FULL, "607565516.ogg"},
	{0x117F6A6808638372ULL, "6076353_m.fxe"},
	{0x117F73A2379DFE06ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_c.anb"},
	{0x1180178857BC01A6ULL, "vxdr_primevaltomb_877_0_0.vlm"},
	{0x118019064F19FF22ULL, "ser221_gamlen6135300.cl"},
	{0x11801BBA7D2291B0ULL, "one000_ilen6205660.cl"},
	{0x11809693C43ABDA9ULL, "mag110_cut_abomination6037102.cl"},
	{0x1180AF54F5978248ULL, "fen200ar_ambush_exterior.lst"},
	{0x1181207EC3A94B9FULL, "6155529_m.fxe"},
	{0x1181280892DE1730ULL, "qun221_viscount6069183.cl"},
	{0x11817013515108F5ULL, "art\\characters\\playercharacter\\elfmale\\har_fls.mmh"},
	{0x1181721351510C6BULL, "art\\characters\\playercharacter\\elfmale\\har_fls.msh"},
	{0x1182109451F5EC08ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em02.mmh"},
	{0x11822E9451F61F32ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em02.msh"},
	{0x11825E6393106DA1ULL, "6135662_m.fxe"},
	{0x118275FFB7816C4CULL, "w_woundedcoast_39_0.rml"},
	{0x11829064A10A390CULL, "6087763_f.fxe"},
	{0x1182CAFDC2238708ULL, "art\\levels\\4cinematics\\qunari\\prp_qunariamulet.mmh"},
	{0x1182D6FDC2239BABULL, "art\\levels\\4cinematics\\qunari\\prp_qunariamulet.mao"},
	{0x1182D818A8B7F14CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_1_step.ani"},
	{0x1182E22304E7E300ULL, "148742565.ogg"},
	{0x1182E8FDC223BA32ULL, "art\\levels\\4cinematics\\qunari\\prp_qunariamulet.msh"},
	{0x1183037BD25CA996ULL, "mag170_thrask6054079.cl"},
	{0x11830731F3B2F5EBULL, "hlac_lt_0_lowtown_a2_65535_6x.dds"},
	{0x1183243CBD2F168EULL, "vxw_sundermount_322_0_0.vlm"},
	{0x11836B99630FE7CEULL, "vxdr_0_deeproads_epe_0_0.vlm"},
	{0x1183E279AAB82F37ULL, "hlac_lt_0_lowtown_a2_65536_5x.dds"},
	{0x1183F6CEAB9B19E4ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medd.mmh"},
	{0x1183FACEAB9B2037ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medd.mao"},
	{0x118408C83F5964C8ULL, "6162339_m.fxe"},
	{0x11840CCEAB9B3E86ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medd.msh"},
	{0x11843CF35063A5DFULL, "6042373_m.fxe"},
	{0x11845638D6FEB7E4ULL, "6179854_m.fxe"},
	{0x118538FDBB6AEBF4ULL, "6177849_m.fxe"},
	{0x11858223606AD34FULL, "vxht_chantry_262_0_0.vlm"},
	{0x11858F9ED4D0F681ULL, "dae000_orsino6085515.cl"},
	{0x118620FBF3061C11ULL, "vxlt_undercityencounter_1fj_0_0.vlm"},
	{0x118642B53964CED7ULL, "357787736.ogg"},
	{0x1186580069F805A5ULL, "dae000_mother6101038.cl"},
	{0x11866BAE02EA53ACULL, "hlac_w_bonepit_393221_0x.dds"},
	{0x11869D877318107AULL, "vxw_woundedcoast_wb_0_0.vlm"},
	{0x1186BF476E854250ULL, "vxht_mansion2b_20s_0_0.vlm"},
	{0x1187103D6DFAAD58ULL, "plt_cod_ltr_orsino_thank.plo"},
	{0x1187310325BE2C7EULL, "6098937_m.fxe"},
	{0x1187AF54F59DA099ULL, "fen200ar_ambush_exterior.ncs"},
	{0x118819424D777F6BULL, "1024576961.ogg"},
	{0x11882B4B579FBFF9ULL, "6101229_m.fxe"},
	{0x11886E2CDF54BE5BULL, "dae150_harshal6148875.cl"},
	{0x1188E302C9743436ULL, "hlac_lt_undercityencounter_-1_50y.dds"},
	{0x1189E063A00EFDC7ULL, "plt_fen100pt_recruited.plo"},
	{0x1189E518A8BE258EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_1_step.gad"},
	{0x118A22078931EE66ULL, "vxw_blightlands_vc_0_0.vlm"},
	{0x118A28DF34D82250ULL, "184899696.ogg"},
	{0x118AC96E8F993AFBULL, "mcr310_argument6075090.cl"},
	{0x118B37A1F6033003ULL, "hlac_w_bonepitmining_a2_196612_0z.dds"},
	{0x118B493ABF037C6CULL, "6097905_m.fxe"},
	{0x118B7FACE052AD47ULL, "6027574_m.fxe"},
	{0x118BD111761A29C7ULL, "6108497_m.fxe"},
	{0x118D2A125A2B9444ULL, "587919227.ogg"},
	{0x118D2EFB0E970640ULL, "47340202.ogg"},
	{0x118D76A237AA40B1ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_c.mmh"},
	{0x118D8D9B71F1F7D7ULL, "533844651.ogg"},
	{0x118D8E051C81BD36ULL, "6106612_m.fxe"},
	{0x118DCFCE32DF86EAULL, "follower_banter6172728.cl"},
	{0x118E896549B21A26ULL, "vxdr_primevaltomb_828_0_0.vlm"},
	{0x118EF77676240EB8ULL, "332993883.ogg"},
	{0x118F0EE5EE528B49ULL, "6104882_m.fxe"},
	{0x118F526A9E7ACDFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\pp_male_mage.anb"},
	{0x118F59B7D2A07661ULL, "lgt320_shw_ors_marlein6156376.cl"},
	{0x118FC9BD4C8D628DULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypouchb_d.dds"},
	{0x118FE018A8C2896DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_1_step.evt"},
	{0x119025E5180CED91ULL, "119174024.ogg"},
	{0x11902A1B3F2B800FULL, "llac_ht_brothel_n_2995_3.dds"},
	{0x11904E05C2D0AFA0ULL, "zz_vault_debug6203178.cl"},
	{0x11911B4041533FD2ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.mmh"},
	{0x1191732F2FED3C9AULL, "vxlt_0_lowtown_a2_n_c_eje_0_0.vlm"},
	{0x1191C7CE9F9B108EULL, "6030491_f.fxe"},
	{0x11922F8637241A2BULL, "vxlt_warehouse_ce_0_0.vlm"},
	{0x11929D22246190B5ULL, "645747309.ogg"},
	{0x1192E7CAC14BE785ULL, "one280_emrys6099750.cl"},
	{0x11935A789038A54DULL, "llac_dr_slavers_3216_10.dds"},
	{0x11939F7C09BEBF1BULL, "one200_gold_followers6101136.cl"},
	{0x11944B8358AB8FD7ULL, "plt_cod_ltr_sk_starkhaven.plo"},
	{0x119566B68EA04A60ULL, "6076319_m.fxe"},
	{0x119637CEC76C5D41ULL, "vxht_0_hightown_4n2_0_0.vlm"},
	{0x11972E3E5BA4B965ULL, "6135635_m.fxe"},
	{0x11977ABAF7B7E622ULL, "vxht_0_hightown_6pm_0_0.vlm"},
	{0x1197B079A80FA10BULL, "vxw_bonepit_2al_0_0.vlm"},
	{0x1197C46DA1708599ULL, "hlac_ca_caverns1_mines_24426_10z.dds"},
	{0x1197D49606A0595DULL, "hlac_lt_0_lowtown_docks_a3_n_-1_3x.dds"},
	{0x1198B255E19D616EULL, "328205121.ogg"},
	{0x11990B32590F2DC3ULL, "w_bonepitmining_a2_16_0.msh"},
	{0x119915BC8B297F16ULL, "dae000ip_crowd_lm5_talk18.utp"},
	{0x11991932590F451DULL, "w_bonepitmining_a2_16_0.mmh"},
	{0x11991D32590F4C2EULL, "w_bonepitmining_a2_16_0.mao"},
	{0x11994362D403E3D3ULL, "vxlt_darktown_4hi_0_0.vlm"},
	{0x1199456063597D84ULL, "lt_0_lowtown_docks_n_210_0.mmh"},
	{0x11994960635984D7ULL, "lt_0_lowtown_docks_n_210_0.mao"},
	{0x11995B606359A326ULL, "lt_0_lowtown_docks_n_210_0.msh"},
	{0x11997A8B2E5889D0ULL, "rdr111dg_martin6053592.cl"},
	{0x1199BC19C24FEEBCULL, "hlac_lt_undercityserialkiller_-1_17x.dds"},
	{0x1199EC698BD5E06FULL, "6033768_m.fxe"},
	{0x119A4F39E20F589CULL, "follower_banter6170197.cl"},
	{0x119A61A766277867ULL, "100455776.ogg"},
	{0x119AB43A26E60E25ULL, "533247878.ogg"},
	{0x119AB6771F770680ULL, "fxa_rocka.dds"},
	{0x119ADAD50F671983ULL, "vxca_cavern1_haunted_e7s_0_0.vlm"},
	{0x119AF9C948D29364ULL, "395336632.ogg"},
	{0x119B19607F23A7B1ULL, "follower_banter6141738.cl"},
	{0x119B9022FA756C86ULL, "vxca_caverns1_iai_0_0.vlm"},
	{0x119BA62635EAB525ULL, "isa311_isabela6092512.cl"},
	{0x119BE80FD2D6EEB4ULL, "nat380_nathaniel_banter6090043.cl"},
	{0x119C8B6A3A4B7FDEULL, "vxlt_gallowscourtyard_745_0_0.vlm"},
	{0x119C8CDA56AB329CULL, "6083760_m.fxe"},
	{0x119C9C9405D1179DULL, "hlac_w_blightlands_327682_0z.dds"},
	{0x119CA491ED43A978ULL, "718719559.ogg"},
	{0x119CDC5F1090653BULL, "llac_lt_gallowstemplar_n_2388_1.dds"},
	{0x119D463AA3D98568ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge03_sa.dds"},
	{0x119D9022CEAC8C9FULL, "w_blightlands_171_0.phy"},
	{0x119DD79E2E76AF4BULL, "6155312_m.fxe"},
	{0x119DF54CEE7EA663ULL, "hlac_lt_0_lowtown_n_0_1z.dds"},
	{0x119EB7C783958695ULL, "234260086.ogg"},
	{0x119EB88EA80DA2E7ULL, "199565837.ogg"},
	{0x119EDCE55578356DULL, "zz_vault_debug6177764.cl"},
	{0x119F00EECBF90480ULL, "gen00fl_merrill6106993.cl"},
	{0x119F155C92ED0A58ULL, "lt_0_lowtown_docks_a3_n_16_0_ll.mmh"},
	{0x119F1D8AA5E4C608ULL, "zz_otl_debug6060956.cl"},
	{0x119F307056DC21F5ULL, "6166600_m.fxe"},
	{0x119FB21E3C8526E0ULL, "vxca_caverns1_mines_e7t_0_0.vlm"},
	{0x119FEB405ACA72AEULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_lp.ani"},
	{0x11A03D00A705EE87ULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash.anb"},
	{0x11A05F263BB7B1C4ULL, "441791271.ogg"},
	{0x11A07A2524EAA019ULL, "179972604.ogg"},
	{0x11A08D85CD654152ULL, "lgt310_mis_burn_murder26152766.cl"},
	{0x11A09FF192F0B759ULL, "vxdr_primevaltomb_706_0_0.vlm"},
	{0x11A0A163291A32ABULL, "vxlt_warehouse_dd_0_0.vlm"},
	{0x11A1AF03FE2069EFULL, "vxlt_gallowscourtyard_4wb_0_0.vlm"},
	{0x11A1D8CE2895CA36ULL, "one110cr_harimann_guard_2.utc"},
	{0x11A1F50D8FEE2C4EULL, "6125897_m.fxe"},
	{0x11A25E1BED7F5049ULL, "6174243_m.fxe"},
	{0x11A29F93DD394523ULL, "vxlt_hangedman_16g_0_0.vlm"},
	{0x11A31A17C79B803DULL, "349671030.ogg"},
	{0x11A3347ADB51612CULL, "6096872_m.fxe"},
	{0x11A335F63719335EULL, "mag360_samson6044903.cl"},
	{0x11A36C972DE189A8ULL, "441868887.ogg"},
	{0x11A4E743215F4C17ULL, "vxht_0_hightown_as2_0_0.vlm"},
	{0x11A50B2DD643404FULL, "mag101_grace6034376.cl"},
	{0x11A517BE6AE2AEB0ULL, "6206088_m.fxe"},
	{0x11A51900061E9865ULL, "pro000_refugee_angry6052173.cl"},
	{0x11A5254332DA4434ULL, "6113130_f.fxe"},
	{0x11A6407A5A388CC5ULL, "vxw_bonepit_1vd_0_0.vlm"},
	{0x11A695B1823A14E4ULL, "vxlt_gallowstemplar_3nk_0_0.vlm"},
	{0x11A7031CA611496EULL, "6070980_m.fxe"},
	{0x11A731EC174E97EBULL, "6106882_m.fxe"},
	{0x11A78A06B6011E16ULL, "vxlt_0_lowtown_docks_bqw_0_0.vlm"},
	{0x11A79888B90AA2D7ULL, "557596028.ogg"},
	{0x11A7C0A40575FF1EULL, "vxlt_foundry_n_x2_0_0.vlm"},
	{0x11A8288659E17E4CULL, "w_bonepitmining_a2_2_0_lc.mmh"},
	{0x11A83C92AF3CF0E1ULL, "vxw_woundedcoast_g5_0_0.vlm"},
	{0x11A855884AF8E1EEULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.cess.ani"},
	{0x11A85A8358BC524CULL, "plt_cod_ltr_sk_starkhaven.nss"},
	{0x11A87F279E36E6D6ULL, "hlac_lt_0_lowtown_n_196607_1x.dds"},
	{0x11A89C0EA717C720ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_5z.dds"},
	{0x11A8B4CB5A065AC7ULL, "6106697_m.fxe"},
	{0x11A9670F4FA10534ULL, "524036172.ogg"},
	{0x11A9B40BF8E62325ULL, "hlac_w_sundermount_-196613_0y.dds"},
	{0x11A9C07D6804C74AULL, "vxlt_gamlenshouse_n_59_0_0.vlm"},
	{0x11AA555CA020BC19ULL, "vxlt_0_lowtown_docks_a3_cnk_0_0.vlm"},
	{0x11AA5D1E85D9F68FULL, "character_character_p246.psh.fxo.dx11.assembly"},
	{0x11AA85CD2D891C2EULL, "vxlt_0_lowtown_a2_ff7_0_0.vlm"},
	{0x11AAD080001342C3ULL, "hf_flora_harimann.mor"},
	{0x11AAD2D24DF10169ULL, "seb201cr_lady_harimann.utc"},
	{0x11AAFD472C41985EULL, "vxlt_gallowscourtyard_796_0_0.vlm"},
	{0x11AAFF182DB9E683ULL, "vxlt_undercitydungeon_1a8_0_0.vlm"},
	{0x11AB327580E0F1EDULL, "zz_mrl_debug6184213.cl"},
	{0x11AB43D429A0D028ULL, "6113278_m.fxe"},
	{0x11ABB56500F0CE7BULL, "vxw_encounter3_48_0_0.vlm"},
	{0x11ABFA55361C28B6ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01b.msh"},
	{0x11AC0455361C39B4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01b.mmh"},
	{0x11AC8E7037D2283CULL, "dae000cr_grand_cleric6173686.cl"},
	{0x11AC9FE321F0B411ULL, "zz_lgt_1_debug6114618.cl"},
	{0x11ACAD8420F83125ULL, "vxlt_darktown_6gp_0_0.vlm"},
	{0x11ACE24918073173ULL, "mag170_thrask6054014.cl"},
	{0x11ACF7D2DEF94982ULL, "rock_aa.dds"},
	{0x11AD66973B3A2E17ULL, "qcr000_arishok6042508.cl"},
	{0x11ADFB07D5FD7758ULL, "vxlt_0_lowtown_a2_n_c_e8a_0_0.vlm"},
	{0x11AE4C7FD6D1A019ULL, "6104306_f.fxe"},
	{0x11AE5A5DF93CFF7CULL, "gen_im_trs_qun_two_2ha_r1.uti"},
	{0x11AE9122CEBB01A4ULL, "w_blightlands_171_0.mmh"},
	{0x11AE9522CEBB08F7ULL, "w_blightlands_171_0.mao"},
	{0x11AEA722CEBB2746ULL, "w_blightlands_171_0.msh"},
	{0x11AEA8FEA27F1EC1ULL, "vxlt_0_lowtown_f4v_0_0.vlm"},
	{0x11AF29C2394C38FCULL, "zz_followers_banter6157361.cl"},
	{0x11AF4D76DBCD5025ULL, "vxca_cavern1_haunted_gey_0_0.vlm"},
	{0x11AF9111CA2B4990ULL, "zz_dae_debug6049947.cl"},
	{0x11B002B0A3732D3FULL, "hlac_w_bonepitmining_a2_1_0x.dds"},
	{0x11B0628DB6EE411FULL, "vxw_bonepitdestroyed_a3_1pl_0_0.vlm"},
	{0x11B0DC96F197A714ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_9.pwk"},
	{0x11B1C54870919CC8ULL, "vxw_woundedcoast_g0_0_0.vlm"},
	{0x11B207FE2AFD08D1ULL, "lgt260_way_blood_mage6148375.cl"},
	{0x11B20C2F6E77A0D7ULL, "lt_0_lowtown_docks_a3_n_186_0.msh"},
	{0x11B21B726EBB6DB1ULL, "dae000cr_grand_cleric6187930.cl"},
	{0x11B21E2F6E77BF3AULL, "lt_0_lowtown_docks_a3_n_186_0.mao"},
	{0x11B2222F6E77C649ULL, "lt_0_lowtown_docks_a3_n_186_0.mmh"},
	{0x11B2E5E98A8C1157ULL, "gen00fl_merrill6128729.cl"},
	{0x11B3D036477B67A3ULL, "vxht_mansion_n_54h_0_0.vlm"},
	{0x11B40006BA674A4DULL, "6042102_m.fxe"},
	{0x11B41CE843A1D4E5ULL, "vxlt_foundry_n_vs_0_0.vlm"},
	{0x11B427774BAA58E1ULL, "484021365.ogg"},
	{0x11B4708A574E754AULL, "hlac_lt_0_lowtown_docks_n_-196609_0y.dds"},
	{0x11B4C4405ADC8CE8ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_lp.gad"},
	{0x11B54D884B035E3FULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.cess.evt"},
	{0x11B55740CE633321ULL, "hlac_lt_darktown_-1_39y.dds"},
	{0x11B56D2E8219D293ULL, "6071890_m.fxe"},
	{0x11B694E058884392ULL, "6115832_m.fxe"},
	{0x11B6FE87863EF688ULL, "hlac_ht_bartrand_n_3369_0x.dds"},
	{0x11B7A2400B407A5AULL, "zz_dae_debug6029806.cl"},
	{0x11B873DCDE619D86ULL, "vxca_caverns1_iok_0_0.vlm"},
	{0x11B8B444BF9BED96ULL, "vxlt_gallowscourtyard_7fs_0_0.vlm"},
	{0x11B9756591DEAE6FULL, "6086287_m.fxe"},
	{0x11B9C4CB069D2B48ULL, "694234596.ogg"},
	{0x11B9DAC84522F8F6ULL, "hlac_lt_gallowstemplar_2383_0x.dds"},
	{0x11BA1A4041764379ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.anb"},
	{0x11BA56497794AF0EULL, "6092510_f.fxe"},
	{0x11BA93EE426E0E70ULL, "vxlt_0_lowtown_docks_a3_n_cor_0_0.vlm"},
	{0x11BB333D1A684DF7ULL, "w_bonepit_247_0.msh"},
	{0x11BB453D1A686C5AULL, "w_bonepit_247_0.mao"},
	{0x11BB493D1A687369ULL, "w_bonepit_247_0.mmh"},
	{0x11BB701376A4A410ULL, "711099403.ogg"},
	{0x11BB757E1F01785AULL, "6059039_m.fxe"},
	{0x11BC122ADEACB2AFULL, "6018598_m.fxe"},
	{0x11BC2E884B094828ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.cess.gad"},
	{0x11BC58593C1E1FF1ULL, "141054431.ogg"},
	{0x11BC5DC9ED006DACULL, "6088729_m.fxe"},
	{0x11BCD7539122FACCULL, "dae000_worthy6166571.cl"},
	{0x11BD2E28F2B01A94ULL, "204816395.ogg"},
	{0x11BD53B2A906DEFAULL, "hlac_ca_caverns2_5802_23z.dds"},
	{0x11BDC0EFD07388DFULL, "dae000_quintus6165250.cl"},
	{0x11BDF51300574D02ULL, "764821954.ogg"},
	{0x11BE985812685C24ULL, "vxca_caverns2prc_9yi_0_0.vlm"},
	{0x11BED976B8036294ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls103a_0t.dds"},
	{0x11BFF4104B5F1A35ULL, "vxlt_0_lowtown_a3_n_c_fi5_0_0.vlm"},
	{0x11C06AB32D5D3C52ULL, "6141240_m.fxe"},
	{0x11C0D6D47A746690ULL, "845713336.ogg"},
	{0x11C1619579098D63ULL, "weapon_weapon_p0.vsh.fxo.dx11"},
	{0x11C1FF9C6C8DEAE1ULL, "918031519.ogg"},
	{0x11C24B73FF69D43CULL, "339475308.ogg"},
	{0x11C2C56EA94598E4ULL, "ser120_ghyslain6118443.cl"},
	{0x11C3114702F5D549ULL, "6142992_m.fxe"},
	{0x11C382270F9277FAULL, "vxw_bonepit_1od_0_0.vlm"},
	{0x11C39BFB257B40F5ULL, "6087633_m.fxe"},
	{0x11C406CA63FCA6B3ULL, "llac_lt_0_lowtown_docks_a3_n_196608_1.dds"},
	{0x11C4185195535F44ULL, "fx_explosion02_flipb.dds"},
	{0x11C448884DEA995FULL, "170373646.ogg"},
	{0x11C49CBCC9102E66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_medium.evt"},
	{0x11C4D08E98537AEAULL, "81389740.ogg"},
	{0x11C510BB382B8F97ULL, "vxht_0_hightown_a2_aob_0_0.vlm"},
	{0x11C5A203411FF3A9ULL, "vxlt_0_lowtown_docks_a3_n_cz9_0_1.vlm"},
	{0x11C5CEE221C5A0E6ULL, "615852746.ogg"},
	{0x11C5F6A6A119E0D3ULL, "618268947.ogg"},
	{0x11C6AC46E004732EULL, "vxht_0_hightown_n_8rr_0_0.vlm"},
	{0x11C6DB171ADBBFBFULL, "vxw_bonepitdestroyed_a3_272_0_0.vlm"},
	{0x11C70FF8F7585FBEULL, "vxca_caverns2_8bb_0_0.vlm"},
	{0x11C71C0623F2490CULL, "vxca_cavern1_haunted_egx_0_0.vlm"},
	{0x11C72661925553FDULL, "vxht_mansion2_2im_0_0.vlm"},
	{0x11C72878A2FAC86FULL, "weapon_rendertodepth_p0.psh.fxo.dx11"},
	{0x11C73A0A703A3047ULL, "353087041.ogg"},
	{0x11C7E4C34C70652BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookstand_d.dds"},
	{0x11C85E17FBB0550DULL, "zz_fen_debug6134359.cl"},
	{0x11C90D3BBFB6B099ULL, "121017869.ogg"},
	{0x11C93A00A728EE88ULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash.mmh"},
	{0x11CA149359497A2BULL, "573796998.ogg"},
	{0x11CA8EE63FC3479CULL, "tm_hlk_b01.tnt"},
	{0x11CAD3A3B619C479ULL, "6088247_m.fxe"},
	{0x11CADE1C11368000ULL, "976821850.ogg"},
	{0x11CB3650E988C9CBULL, "6087473_m.fxe"},
	{0x11CB37F3B1590662ULL, "569593425.ogg"},
	{0x11CBB3BCC91673B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_medium.gad"},
	{0x11CBF6F9FEFC7C2BULL, "vxlt_warehouse_wy_0_0.vlm"},
	{0x11CC60904E0F2944ULL, "616798476.ogg"},
	{0x11CD364037953911ULL, "llac_dr_primevaltomb_659_14.dds"},
	{0x11CD469A2A18EE0CULL, "6153458_m.fxe"},
	{0x11CDBE7A099F1899ULL, "vxlt_gamlenshouse_n_mn_0_0.vlm"},
	{0x11CDD04A81DD522EULL, "vxht_keep_4mp_0_0.vlm"},
	{0x11CE20F4E779AA87ULL, "643091602.ogg"},
	{0x11CE5E69029147C4ULL, "llac_ca_cavern1_haunted_24423_9.dds"},
	{0x11CE7B48E23197FEULL, "art\\vfx\\combat\\vfx_w_bsk_regain_sta_imp_c.mmh"},
	{0x11CE84F63C89B274ULL, "art\\vfx\\materials\\chunky\\vfx_c_moving_cloud.mao"},
	{0x11CFF5B7417CCE9DULL, "430607087.ogg"},
	{0x11D008E16F1C95F3ULL, "vxlt_0_lowtown_docks_n_cnm_0_0.vlm"},
	{0x11D025600E629F1CULL, "hlac_ca_caverns1_24427_2z.dds"},
	{0x11D045EEA5852479ULL, "944003737.ogg"},
	{0x11D09202A3359BB6ULL, "lowtown_fem_alienage1_sta6162544.cl"},
	{0x11D0BDCF9E69D076ULL, "dae200_cullen6140315.cl"},
	{0x11D175BBD1C7000FULL, "vxht_chantry_36o_0_0.vlm"},
	{0x11D19D2763E38657ULL, "lt_0_lowtown_docks_a3_n_247_0.mmh"},
	{0x11D1A12763E38D64ULL, "lt_0_lowtown_docks_a3_n_247_0.mao"},
	{0x11D1B32763E3ABC9ULL, "lt_0_lowtown_docks_a3_n_247_0.msh"},
	{0x11D1C62E291FD348ULL, "vxw_bonepit_1q8_0_0.vlm"},
	{0x11D214BF90535D63ULL, "378635256.ogg"},
	{0x11D26EFD97487698ULL, "llac_dr_0_deeproads_5845_17.dds"},
	{0x11D2B4BCC91C945FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_medium.ani"},
	{0x11D34244EDDFB715ULL, "plt_lgt250pt_way_dlsh_cnflc.nss"},
	{0x11D360FA2CB4DB60ULL, "6165969_m.fxe"},
	{0x11D3D73DA47085C9ULL, "gen00fl_isabela_init6110277.cl"},
	{0x11D50AEE6E9D0761ULL, "6041411_m.fxe"},
	{0x11D52080D40F1781ULL, "fxa_evl_aura2.mao"},
	{0x11D56C034B0CB47BULL, "vxlt_0_lowtown_a2_ev9_0_0.vlm"},
	{0x11D581D5E94E783EULL, "vxca_caverns2_8ic_0_0.vlm"},
	{0x11D58F0BF2ADE8F1ULL, "6115295_m.fxe"},
	{0x11D5AFBF0B03E747ULL, "vxht_0_hightown_n_av6_0_0.vlm"},
	{0x11D5ED09867C385DULL, "vxw_woundedcoast_149_0_0.vlm"},
	{0x11D6BD446A901750ULL, "228598428.ogg"},
	{0x11D777A8265A3F70ULL, "vxlt_0_lowtown_docks_a3_n_cup_0_0.vlm"},
	{0x11D7F8445C28EB4AULL, "hlac_ht_chantry_2534_26y.dds"},
	{0x11D87E96F12114B7ULL, "vxlt_gallowstemplar_f_4qq_0_0.vlm"},
	{0x11D8C60CF39E6517ULL, "6171866_m.fxe"},
	{0x11D8CD9F1514A52EULL, "mcr371cr_anders_cut.utc"},
	{0x11D90E3E63665787ULL, "clf260_harley6173865.cl"},
	{0x11D914B63D57AC55ULL, "dae300_alistair_king6133818.cl"},
	{0x11D91D2AC8783490ULL, "vxca_caverns1_gjl_0_0.vlm"},
	{0x11D93F7FF84F75F1ULL, "418657927.ogg"},
	{0x11D9554955DC022EULL, "vxht_keep_739_0_0.vlm"},
	{0x11D9C9E806693BBCULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\armoire.closing.ani"},
	{0x11D9DB782B3DD6ECULL, "6165279_m.fxe"},
	{0x11DA68D6F0F294ABULL, "vxlt_warehouse_xx_0_0.vlm"},
	{0x11DA747F9EF1F970ULL, "304208144.ogg"},
	{0x11DA765B6C49C9DBULL, "762856043.ogg"},
	{0x11DB39ED33507020ULL, "171137149.ogg"},
	{0x11DBE8350B2D2309ULL, "ht_0_hightown_a3_33_0_ll.mmh"},
	{0x11DC5A48E23D9CBDULL, "art\\vfx\\combat\\vfx_w_bsk_regain_sta_imp_c.anb"},
	{0x11DD1E985D8CA7DCULL, "llac_lt_0_lowtown_a2_n_c_0_4.dds"},
	{0x11DD8A9B51D49B24ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutd.msh"},
	{0x11DD9C9B51D4B98DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutd.mao"},
	{0x11DDA09B51D4C0C6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutd.mmh"},
	{0x11DDAE0FC9D62B88ULL, "6001289_m.fxe"},
	{0x11DDEFF9A778478AULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_bdy_snga_0n.dds"},
	{0x11DE530BB8B407CAULL, "pro000_cut2d_thefirstyear6005295.cl"},
	{0x11DEB491F2F377F5ULL, "vxht_mansion2b_2im_0_0.vlm"},
	{0x11DEF70D46C5A64BULL, "hlac_w_bonepitmining_a2_196613_0y.dds"},
	{0x11DFBE56DDE2B983ULL, "789231335.ogg"},
	{0x11DFCBB85DE2ECCFULL, "plt_zz_apppt_debug.plo"},
	{0x11E0380B1B15EBC8ULL, "vxw_bonepit_1z9_0_0.vlm"},
	{0x11E073194B36EA44ULL, "gen00fl_fenris6126739.cl"},
	{0x11E08BF1AB5C32C5ULL, "6148240_m.fxe"},
	{0x11E0D6E8066F70FEULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\armoire.closing.gad"},
	{0x11E119DC2F8452CDULL, "vxht_mansion_n_6zz_0_1.vlm"},
	{0x11E22A209FCFCB10ULL, "vxw_blightlands_128_0_0.vlm"},
	{0x11E2B2202E628566ULL, "one280_emrys6120059.cl"},
	{0x11E2CA88C59DD1E5ULL, "vxlt_gallowstemplar_f_5cx_0_0.vlm"},
	{0x11E300F23E1741B9ULL, "711839004.ogg"},
	{0x11E30985E0B93208ULL, "mcr370_messenger6068983.cl"},
	{0x11E346774E3B2CD9ULL, "vxlt_undercityencounter_1bw_0_1.vlm"},
	{0x11E4C6689E85E734ULL, "vxlt_gallowstemplar_n_5qc_0_0.vlm"},
	{0x11E4CBD0E8F70EEAULL, "vxw_woundedcoast_1n8_0_0.vlm"},
	{0x11E4E06300F1B7B8ULL, "gen_im_npc_arm_hum_glv_rdrc.uti"},
	{0x11E5021955810479ULL, "875458192.ogg"},
	{0x11E58A611798375BULL, "vxht_0_hightown_9cn_0_0.vlm"},
	{0x11E59395F4CFCEA4ULL, "vxlt_gallowscourtyard_7f1_0_0.vlm"},
	{0x11E5A4EB6CC9A346ULL, "zz_fen_debug6165671.cl"},
	{0x11E5F2F1BA85284EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkd_sa.dds"},
	{0x11E5F8EFCCE573CEULL, "hlac_w_woundedcoast_0_12z.dds"},
	{0x11E65CA391941A75ULL, "546726120.ogg"},
	{0x11E66ED56792D617ULL, "gen_im_trs_qun_lgt_boo_r2.uti"},
	{0x11E67547901A8EF2ULL, "mcr370_samson.cnv"},
	{0x11E687447F7CA773ULL, "fen210_fenris6108721.cl"},
	{0x11E6E8C3925D1134ULL, "vxlt_gallowstemplar_f_55o_0_0.vlm"},
	{0x11E71E7E91587D79ULL, "dae100_gamlen_letter6146995.cl"},
	{0x11E75C534AF490B6ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wrba.msh"},
	{0x11E766534AF4A1B4ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wrba.mmh"},
	{0x11E76A534AF4A8C7ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wrba.mao"},
	{0x11E7B09C0F75B3A0ULL, "54157363.ogg"},
	{0x11E7F479D1863F8BULL, "517550101.ogg"},
	{0x11E804ABAA6AAAE5ULL, "hlac_w_bonepitdestroyed_a3_0_4y.dds"},
	{0x11E84C4C231C4383ULL, "vxca_caverns2prc_a5f_0_0.vlm"},
	{0x11E8A5AA769B59BEULL, "6119753_m.fxe"},
	{0x11E8D45731CEF46BULL, "6133019_m.fxe"},
	{0x11E93EF1F7602CB3ULL, "zz_approval_debug6133710.cl"},
	{0x11E99747901CFE4AULL, "mcr370_samson.bnk"},
	{0x11E9C317C4D7CD28ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_quanarithrone_d.dds"},
	{0x11E9F75642AD6FF8ULL, "764350333.ogg"},
	{0x11EA335823487A48ULL, "870339646.ogg"},
	{0x11EA3AE187DE5E0BULL, "vxlt_0_lowtown_a2_n_c_gnv_0_0.vlm"},
	{0x11EB9719A1DD2C6FULL, "vxlt_0_lowtown_a2_fec_0_0.vlm"},
	{0x11EC7B66C84AECC2ULL, "6039392_m.fxe"},
	{0x11ECEDDA2B8909C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_neck.ani"},
	{0x11ED20AB861C1EDBULL, "llac_dr_0_deeproads_5849_1.dds"},
	{0x11ED266EE4E99075ULL, "vxht_mansion2b_2bl_0_0.vlm"},
	{0x11ED4131F718227BULL, "static_static_p13.psh.fxo.dx11"},
	{0x11ED9FC914A611C9ULL, "vxht_mansion2_1w0_0_0.vlm"},
	{0x11EE8BDEB0F8FFD1ULL, "533578402.ogg"},
	{0x11EEAA4F1A066CF2ULL, "123707200.ogg"},
	{0x11EEE25EED6743FFULL, "lt_0_lowtown_docks_230_0.msh"},
	{0x11EEE85EED674DA1ULL, "lt_0_lowtown_docks_230_0.mmh"},
	{0x11EEF45EED676242ULL, "lt_0_lowtown_docks_230_0.mao"},
	{0x11EEFDCE9D524B45ULL, "6145140_m.fxe"},
	{0x11F06A3A0AF107D7ULL, "471204751.ogg"},
	{0x11F09982D6FE2F57ULL, "vxw_encounter3_g8_0_1.vlm"},
	{0x11F0AC8E3F64FCECULL, "vxdr_primevaltomb_912_0_0.vlm"},
	{0x11F1D80D67260717ULL, "pro000_colonel_ewald6040544.cl"},
	{0x11F1D8D97CEF7DE9ULL, "900815244.ogg"},
	{0x11F24A0AB95B459EULL, "art\\vfx\\combat\\vfx_g_dual_att_back_c.impa.ani"},
	{0x11F28546A56ABB93ULL, "6053685_m.fxe"},
	{0x11F33DADDAED276AULL, "vxw_woundedcoast_1i9_0_0.vlm"},
	{0x11F37246991117BAULL, "mag131_keran6086259.cl"},
	{0x11F3AF50B57706B1ULL, "236263254.ogg"},
	{0x11F3F343CBA415D5ULL, "6136689_m.fxe"},
	{0x11F54144EDFC9C76ULL, "plt_lgt250pt_way_dlsh_cnflc.plo"},
	{0x11F54746B24C273CULL, "623827057.ogg"},
	{0x11F579DB7831A938ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_dssta_0d.dds"},
	{0x11F5918508AA4F8AULL, "ico_tactic.dds"},
	{0x11F600543F1299DFULL, "6089509_m.fxe"},
	{0x11F63847D0895201ULL, "vxdr_0_deeproads_cg4_0_1.vlm"},
	{0x11F6BE2915125C03ULL, "vxca_caverns2prc_a0i_0_0.vlm"},
	{0x11F72A35AA440281ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_mag.mao"},
	{0x11F75F9952F6E0DBULL, "vxht_chantry_n_28c_0_0.vlm"},
	{0x11F816B2FB74E878ULL, "qun100_mage_party_comment6019115.cl"},
	{0x11F8C68A25F87F05ULL, "hlac_lt_0_lowtown_n_-262144_0y.dds"},
	{0x11F8CD056D95BE30ULL, "1063743620.ogg"},
	{0x11F8DCAB239FA2EFULL, "llac_ca_cavern1_haunted_24421_0.dds"},
	{0x11F96A46C88FDFA5ULL, "6093573-ancienthistory.cut"},
	{0x11F96B533CFF9829ULL, "vxw_blightlands_nx_0_0.vlm"},
	{0x11FA195876032A40ULL, "lgt150_nat_were_supporter6111012.cl"},
	{0x11FA4FDC81BD7DB1ULL, "dae200_cullen6140273.cl"},
	{0x11FB05DA2B956FBAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_neck.evt"},
	{0x11FB13FAD93F2641ULL, "916053005.ogg"},
	{0x11FBF5F15CD3A505ULL, "6082142_m.fxe"},
	{0x11FC11A6069C2A49ULL, "vxht_mansion2_1z7_0_0.vlm"},
	{0x11FC859C3DA6C2B9ULL, "llac_lt_darktown_-1_83.dds"},
	{0x11FC937643BBD237ULL, "6016008_m.fxe"},
	{0x11FC9A1294E5B416ULL, "vxw_bonepitdestroyed_a3_1wp_0_0.vlm"},
	{0x11FCF5F386742EA6ULL, "vxlt_0_lowtown_gcm_0_0.vlm"},
	{0x11FD02113938D453ULL, "83507978.ogg"},
	{0x11FD6C7429B1DF76ULL, "gen00fl_isabela6135690.cl"},
	{0x11FE237FF17DA2FFULL, "6143104_m.fxe"},
	{0x11FE45CF7716F2C2ULL, "371239155.ogg"},
	{0x11FE5A9CFD460EC9ULL, "6201167_m.fxe"},
	{0x11FE64010170731FULL, "hlac_lt_0_lowtown_a2_n_c_-3_7x.dds"},
	{0x11FE98B6983522EDULL, "11143498.ogg"},
	{0x11FEB19B1DDF39E2ULL, "6076058_m.fxe"},
	{0x120032BE33E53B48ULL, "814677454.ogg"},
	{0x1200620AB967AC6FULL, "art\\vfx\\combat\\vfx_g_dual_att_back_c.impa.evt"},
	{0x1200908BFFFED2F9ULL, "vxht_chantry_350_0_0.vlm"},
	{0x120104DA2B99D9ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_neck.gad"},
	{0x12017EFB700914A9ULL, "dae000_mother6135745.cl"},
	{0x1201D674B6F1C7EBULL, "hlac_w_encounter2_-65531_0x.dds"},
	{0x1201E943B5E1824CULL, "6170642_m.fxe"},
	{0x12028DF59A4B6131ULL, "272713885.ogg"},
	{0x120318BC5337A0FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_wep_n.dds"},
	{0x12032086FFF46ADDULL, "dae000_bodhan6146850.cl"},
	{0x1203226EB71CE0FBULL, "vxht_keep_n_6m0_0_0.vlm"},
	{0x12040B42AC8337C6ULL, "381756675.ogg"},
	{0x120428B8B04C29C1ULL, "384646107.ogg"},
	{0x12042A50AB69D856ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightc.phy"},
	{0x12043D0F6E086D59ULL, "6218024_m.fxe"},
	{0x1204622C7B3F913AULL, "follower_banter6148699.cl"},
	{0x1204FD2EF9869316ULL, "vxht_keep_n_4hx_0_0.vlm"},
	{0x120511185D14784CULL, "413297073.ogg"},
	{0x1205B81115BB256BULL, "6139619_m.fxe"},
	{0x12064760B78A9836ULL, "6142797_m.fxe"},
	{0x1206630AB96C1918ULL, "art\\vfx\\combat\\vfx_g_dual_att_back_c.impa.gad"},
	{0x120667F5755D4FE9ULL, "187920341.ogg"},
	{0x12070CC042FE3460ULL, "6094305_f.fxe"},
	{0x12074B8D3E8232BDULL, "6130965_f.fxe"},
	{0x1207637195720C9EULL, "6106866_f.fxe"},
	{0x120785C1531C5AB5ULL, "6068029_m.fxe"},
	{0x12080355366A96AFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01b.phy"},
	{0x12082C7C51492397ULL, "lt_0_lowtown_docks_a3_39_0_lc.mmh"},
	{0x12094609BD9F55C3ULL, "gen00fl_anders6119271.cl"},
	{0x120963B5576554EEULL, "6141514_m.fxe"},
	{0x120968F04BBD3A48ULL, "ico_boots_med_1.dds"},
	{0x1209749BE7258A69ULL, "vxlt_0_lowtown_a3_n_c_i0i_0_0.vlm"},
	{0x1209B46E3E7185D1ULL, "6129094_m.fxe"},
	{0x120A1E01305C5CCDULL, "6129084_f.fxe"},
	{0x120A67CE4EC9BD85ULL, "6085042_m.fxe"},
	{0x120A97965BA40609ULL, "follower_banter6167037.cl"},
	{0x120AF54271B571B6ULL, "lt_0_lowtown_docks_a3_101_0.phy"},
	{0x120B16B42A933176ULL, "6041690_m.fxe"},
	{0x120B9B3C91C16B15ULL, "hlac_w_blightlands_327677_0z.dds"},
	{0x120BC32A2392B893ULL, "lowtown_male_drunk_mob6151161.cl"},
	{0x120BF4F5B99C9012ULL, "6045962_m.fxe"},
	{0x120D6DD4D6A25625ULL, "one100dg_kelder6060542.cl"},
	{0x120DD8F63AD8F857ULL, "274626356.ogg"},
	{0x120E301978AF9955ULL, "zz_ave_debug6064624.cl"},
	{0x120EBD084F74D9A0ULL, "hlac_lt_0_lowtown_a3_n_c_65538_1y.dds"},
	{0x120F1350AB73B8F7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightc.msh"},
	{0x120F2550AB73D75AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightc.mao"},
	{0x120F2950AB73DE69ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightc.mmh"},
	{0x120F628C9A68756CULL, "vxw_encounter2_camp_3a_0_0.vlm"},
	{0x120F76D9FD945557ULL, "vrc321ip_haunted_door.utp"},
	{0x12117240601EF4C0ULL, "63231342.ogg"},
	{0x12118CD2C020E7CCULL, "18406681.ogg"},
	{0x1211944BA912F97BULL, "vxht_keep_n_6h3_0_0.vlm"},
	{0x12119BE6278E9E75ULL, "6108918_m.fxe"},
	{0x1211E04F2026794EULL, "follower_banter6166462.cl"},
	{0x121226909F0C2A7EULL, "vxca_caverns1_i1a_0_0.vlm"},
	{0x12123E0A8C5E2925ULL, "gen_im_trs_krk_low_junk2_01.uti"},
	{0x12128735F0D68D6FULL, "gen_im_trs_gen_two_2hm_r0.uti"},
	{0x12129036CADAC96CULL, "6071173_f.fxe"},
	{0x1212E40E22FECADFULL, "6087709_m.fxe"},
	{0x1212FC4A03191D47ULL, "6149612_m.fxe"},
	{0x121358E7121DD763ULL, "191537953.ogg"},
	{0x1213F9A2018A6C26ULL, "hlac_lt_0_lowtown_a2_n_-65535_3z.dds"},
	{0x12142BBB30C38C87ULL, "art\\effects\\cloudshadows\\w_woundedcoast\\cloud1.dds"},
	{0x12149E991A4B53B9ULL, "vxlt_darktown_4lw_0_0.vlm"},
	{0x1214C74C2F59253BULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ca_basic.mmh"},
	{0x1214D34C2F5939D8ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ca_basic.mao"},
	{0x1214DE4271BD9F57ULL, "lt_0_lowtown_docks_a3_101_0.msh"},
	{0x1214E54C2F595825ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ca_basic.msh"},
	{0x1214F04271BDBDBAULL, "lt_0_lowtown_docks_a3_101_0.mao"},
	{0x1214F44271BDC4C9ULL, "lt_0_lowtown_docks_a3_101_0.mmh"},
	{0x12156DC70A0A72E4ULL, "vxht_brothel_n_621_0_0.vlm"},
	{0x1215C45F01F70EDEULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrysmelt.dura.evt"},
	{0x1215E5E01734C97BULL, "gen_im_cns_cft_arn_05.uti"},
	{0x12161EA8EA883F7FULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_6z.dds"},
	{0x1216394101B41649ULL, "6135333_f.fxe"},
	{0x121655767C27C0C2ULL, "gen00fl_varric6078522.cl"},
	{0x12168E487AA1223DULL, "6165555-mrl351-pridedemon-a.cut"},
	{0x1217714D332D8929ULL, "vxw_sundermount_2gc_0_0.vlm"},
	{0x12178503829F3723ULL, "gen_im_delete_me_05.uti"},
	{0x1217EFE3984304A4ULL, "6139238_m.fxe"},
	{0x1217F5D38E7CBBA2ULL, "vxlt_0_lowtown_enx_0_0.vlm"},
	{0x12180CB221195C99ULL, "vxdr_primevaltomb_2xc_0_0.vlm"},
	{0x1218F9B4E629A34EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev2_b.ani"},
	{0x12195AC4043DE6A8ULL, "6128382_m.fxe"},
	{0x121A38ABD76DA85EULL, "vxw_bonepitdestroyed_a3_196_0_0.vlm"},
	{0x121A690E65081933ULL, "gen00fl_sebastian6178341.cl"},
	{0x121B1737DA289A28ULL, "vxht_bartrand_n_65h_0_0.vlm"},
	{0x121B3C38B28D6273ULL, "6136782_m.fxe"},
	{0x121BA441C4B283B8ULL, "vxlt_0_lowtown_a3_n_c_fhk_0_0.vlm"},
	{0x121C088E85C5C92BULL, "506201614.ogg"},
	{0x121CCA2277CCCB23ULL, "vxw_encounter2_camp_pc_0_0.vlm"},
	{0x121CDB5F01FD5409ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrysmelt.dura.gad"},
	{0x121DF5987B410F63ULL, "vxlt_0_lowtown_a2_n_c_exn_0_0.vlm"},
	{0x121E3E3748159945ULL, "vxca_caverns1_eaj_0_0.vlm"},
	{0x121EED886C6180FAULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyn_n.dds"},
	{0x121F7C1F44AD4BF9ULL, "152554573.ogg"},
	{0x122006289B0911FBULL, "vxht_keep_n_6c2_0_0.vlm"},
	{0x122080EFCE238704ULL, "lla_ht_0_hightown_n_65538_0.dds"},
	{0x1220F4418FAA970AULL, "mrl000_marethari6129379.cl"},
	{0x12219A300FC84803ULL, "gen00fl_sebastian6188650.cl"},
	{0x1221EC7124161ECDULL, "gen00fl_isabela6079050.cl"},
	{0x122241046F4659A8ULL, "vxht_mansion2b_2n1_0_0.vlm"},
	{0x12224144C84EF708ULL, "319182747.ogg"},
	{0x1222F9DB426CE9AAULL, "770751128.ogg"},
	{0x1223254E6D7F5796ULL, "vxlt_0_lowtown_a3_n_c_eq9_0_0.vlm"},
	{0x1223329FF6787651ULL, "vxht_mansion2_3gn_0_0.vlm"},
	{0x122332FEE312256CULL, "plt_vrc300pt_n_friend_max.plo"},
	{0x122371E11A81ABD7ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.dura.gad"},
	{0x1223794DBE535ED4ULL, "6166779_m.fxe"},
	{0x1223AC5F02032357ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrysmelt.dura.ani"},
	{0x1223D0A7A122A5C9ULL, "89522764.ogg"},
	{0x1223D447D283D92FULL, "hlac_w_sundermount_0_1y.dds"},
	{0x1224309AFF5773F3ULL, "hlac_lt_0_lowtown_docks_a3_65532_1z.dds"},
	{0x1224841AB40AEC7CULL, "vxw_encounter2_2z_0_0.vlm"},
	{0x122498D70064B8DAULL, "vxht_0_hightown_a3_ca2_0_0.vlm"},
	{0x1224AA57DD07A593ULL, "vxdr_0_deeproads_cqy_0_0.vlm"},
	{0x1224B6703C1DF9E6ULL, "vxca_caverns1_fn3_0_0.vlm"},
	{0x12255A9A7A3DFAF9ULL, "zz_approval_debug6003060.cl"},
	{0x1225E32A2523A1A9ULL, "vxw_sundermount_2hb_0_0.vlm"},
	{0x1225F36807833515ULL, "6143407_m.fxe"},
	{0x12266C1BCD3792E6ULL, "522881149.ogg"},
	{0x1226E65CDC075FABULL, "vxw_encounter3_4h_0_0.vlm"},
	{0x122815055D4C6C64ULL, "29893430.ogg"},
	{0x122855464789F112ULL, "hlac_lt_0_lowtown_docks_n_-65537_4z.dds"},
	{0x1228D8C19AB5837DULL, "hlac_lt_backalley_g_0_0y.dds"},
	{0x1229064885ED48C5ULL, "6030467_f.fxe"},
	{0x1229101966B009FEULL, "vxca_caverns1_mines_gk9_0_1.vlm"},
	{0x12297963CA199524ULL, "vxw_bonepitdestroyed_a3_1w2_0_0.vlm"},
	{0x1229C42F0D066B32ULL, "vxw_woundedcoast_o2_0_0.vlm"},
	{0x1229EAB85E2159B4ULL, "plt_zz_apppt_debug.nss"},
	{0x122A13D23441A1DEULL, "vxca_cavern1_haunted_in1_0_0.vlm"},
	{0x122A17FEE31815A3ULL, "plt_vrc300pt_n_friend_max.nss"},
	{0x122A2302AE9292D0ULL, "300651937.ogg"},
	{0x122A620D2DF09BF4ULL, "m_dae321ar_player_home_nt.bnk"},
	{0x122A64F83CED3175ULL, "mcr370_meredith6083041.cl"},
	{0x122A86E11A87EE64ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.dura.evt"},
	{0x122ABDD945F391D9ULL, "175217507.ogg"},
	{0x122AD63EA47BAA90ULL, "vxlt_darktown_2vv_0_0.vlm"},
	{0x122AF0D064C5B62AULL, "hlac_lt_0_lowtown_a2_n_-65536_5y.dds"},
	{0x122AF5E3D09B7144ULL, "vxw_bonepitmining_a2_1o4_0_0.vlm"},
	{0x122B3BFF69C2E3A3ULL, "vxw_encounter2_camp_kb_0_0.vlm"},
	{0x122B9B81E53CFB82ULL, "vxlt_darktown_5t9_0_0.vlm"},
	{0x122BA853ACE67F3FULL, "vxw_encounter2_camp_8o_0_0.vlm"},
	{0x122BC665CD2037B5ULL, "8312115.ogg"},
	{0x122CAEE1364D9BB2ULL, "gen_im_trs_gen_junk1_02.uti"},
	{0x122CDA4899F7A752ULL, "585500759.ogg"},
	{0x122CFC225F347129ULL, "6171907_m.fxe"},
	{0x122D122C37456CA8ULL, "304798667.ogg"},
	{0x122DD2B4E63BBC88ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev2_b.gad"},
	{0x122DEB76EE09A90CULL, "vxlt_undercitydungeon_135_0_0.vlm"},
	{0x122E013C8EC0234CULL, "lt_0_lowtown_docks_a3_n_152_0.msh"},
	{0x122E073C8EC02D3EULL, "lt_0_lowtown_docks_a3_n_152_0.mmh"},
	{0x122E133C8EC041D5ULL, "lt_0_lowtown_docks_a3_n_152_0.mao"},
	{0x122E2B11CBAC8337ULL, "lgt220_mag_tmplr_templar6140250.cl"},
	{0x122E713D19B44759ULL, "gen00fl_carver6102567.cl"},
	{0x122F54538B2E9063ULL, "vxlt_0_lowtown_a2_n_c_h29_0_0.vlm"},
	{0x122F5786BC4FA797ULL, "42368138.ogg"},
	{0x1230960D52C42B58ULL, "392922957.ogg"},
	{0x1230A0EC92906C49ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.mmh"},
	{0x12325180EDF6A3E4ULL, "vxht_brothel_n_687_0_0.vlm"},
	{0x123268AAE434A7E3ULL, "llac_ca_caverns1_24425_12.dds"},
	{0x12334953D045397CULL, "vxw_blightlands_al_0_0.vlm"},
	{0x1233589803B7BDF8ULL, "hm_carver_h2s2.mor"},
	{0x1233E6D62560445AULL, "6143307_m.fxe"},
	{0x1234405E7BBDEAD6ULL, "6171070_m.fxe"},
	{0x123455071719BA29ULL, "vxw_sundermount_2ma_0_0.vlm"},
	{0x1234FE554A3244BEULL, "lgt000cr_mystery_friend.utc"},
	{0x1235435058A4AAFCULL, "hlac_lt_0_lowtown_a2_n_-65537_5x.dds"},
	{0x12354F5C6DD421DBULL, "527692914.ogg"},
	{0x12355839CDFD782BULL, "vxw_encounter3_1k_0_0.vlm"},
	{0x1235FEF0E041DB3AULL, "336978370.ogg"},
	{0x12361E7B02D72196ULL, "6155982_m.fxe"},
	{0x12365C8933F40497ULL, "one300dg_solivitus6048297.cl"},
	{0x1236690BC1D28383ULL, "6017674_m.fxe"},
	{0x1236A7C1F44A1B01ULL, "hlac_lt_0_lowtown_n_65535_6y.dds"},
	{0x1236FCFC9E22208FULL, "dae000_bartrand6206875.cl"},
	{0x12370EFF9207C1C4ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_min_c.impa.ani"},
	{0x123768A6338A7FBDULL, "qun110_petrice6018555.cl"},
	{0x1237D508E57079EBULL, "6170681_m.fxe"},
	{0x1237DBBFE512315AULL, "vxlt_gallowstemplar_n_5wg_0_0.vlm"},
	{0x1237EB40BC0FADA4ULL, "vxw_bonepitdestroyed_a3_1x3_0_0.vlm"},
	{0x123813EE30F863EDULL, "vxlt_backalley_n_1zg_0_0.vlm"},
	{0x1238360BFEFC83B2ULL, "vxw_woundedcoast_d3_0_0.vlm"},
	{0x12383BC2BAB94BD1ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_die_lp_v1.ani"},
	{0x123885AF2637BA5EULL, "vxca_cavern1_haunted_ik2_0_0.vlm"},
	{0x12389EE11A94545DULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.dura.ani"},
	{0x1238A1F432F4A247ULL, "29877048.ogg"},
	{0x123996BCD11DEF56ULL, "gen_im_trs_dwv_med_hlm_r0.uti"},
	{0x1239ADDC5BB8FC23ULL, "vxw_encounter2_camp_fe_0_0.vlm"},
	{0x1239E264090B7F69ULL, "hlac_ht_0_hightown_-196608_0y.dds"},
	{0x123A0A13F7715A02ULL, "1019407233.ogg"},
	{0x123A1F72A4ED0182ULL, "vxlt_backalley_n_1y0_0_1.vlm"},
	{0x123A41393D979B47ULL, "496003021.ogg"},
	{0x123A50F5548FCC05ULL, "gen00fl_merrill6121254.cl"},
	{0x123A7D97DC4DCE5EULL, "mer140_bonepit_comments6062990.cl"},
	{0x123A912BB68BCF1FULL, "495355860.ogg"},
	{0x123B1C4C9A8463C9ULL, "vxlt_0_lowtown_docks_a3_n_bat_0_0.vlm"},
	{0x123B520E2506C748ULL, "vxca_caverns1_eg1_0_0.vlm"},
	{0x123BE74477FD553AULL, "dae000_merchant_armor_hm6164277.cl"},
	{0x123BFF0E261FF194ULL, "hla_ht_0_hightown_a3_131073_0z.dds"},
	{0x123C0CBD97AC6FEDULL, "cracked.mao"},
	{0x123C9B0E6DC9205AULL, "829107087.ogg"},
	{0x123CC934DA06AE74ULL, "6105431_m.fxe"},
	{0x123CF7F4DD574628ULL, "hlac_lt_0_lowtown_a2_n_0_4x.dds"},
	{0x123D054D83541333ULL, "mrl150_hunter6071285.cl"},
	{0x123E0BFF920DDB26ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_min_c.impa.gad"},
	{0x123E46C2BABDCA73ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_die_lp_v1.gad"},
	{0x123E691678518DB4ULL, "6174512_m.fxe"},
	{0x123EB081A6F25D4BULL, "dae030cr_sickly_fereldan_m.utc"},
	{0x123ED963794D81DBULL, "blood_loop2.dds"},
	{0x123EDB806F8E1731ULL, "dae000ip_crowd_em3_talk3.utp"},
	{0x123F1ABAD0E2453DULL, "vxdr_0_deeproads_evp_0_0.vlm"},
	{0x123F3C7F78AB2534ULL, "default_flatnm.dds"},
	{0x123F8807196C218EULL, "vxw_bonepitdestroyed_a3_hz_0_0.vlm"},
	{0x123FE96662CFA097ULL, "379569424.ogg"},
	{0x1240A3A03FB2BEAEULL, "vxw_encounter1_f5_0_0.vlm"},
	{0x1240ADBC6E8760D0ULL, "6117649_m.fxe"},
	{0x1240E1D5C7043FC2ULL, "vxlt_0_lowtown_docks_a3_d6w_0_0.vlm"},
	{0x1241347B423D1AEBULL, "dae000_flemeth6001008.cl"},
	{0x124167D497F71D7CULL, "vxw_encounter2_8x_0_0.vlm"},
	{0x12416D242E0A2C1DULL, "761797875.ogg"},
	{0x1241BB30C23B51FCULL, "vxw_blightlands_jm_0_0.vlm"},
	{0x1241CF09D2EC14D3ULL, "6004201_m.fxe"},
	{0x12426AF789799F18ULL, "6095244_m.fxe"},
	{0x1242AB863F687323ULL, "vxca_caverns2prc_8nz_0_0.vlm"},
	{0x1242B23B6DB40356ULL, "6174370_m.fxe"},
	{0x1242CBB712AF194AULL, "6132747_m.fxe"},
	{0x1242F74941AFB6B2ULL, "dae000ip_crowd_lm5_talk7.utp"},
	{0x12434BD77A429B07ULL, "gen_im_wep_rog_dua_act3_10.uti"},
	{0x124365534B42FEAFULL, "art\\characters\\playercharacter\\humanfemale\\arm_wrba.phy"},
	{0x124401A96FFC77E7ULL, "lgt110_urn_guard6170673.cl"},
	{0x12443942BA0E9BE2ULL, "886503370.ogg"},
	{0x1244A2F0B08A2B78ULL, "6028360_f.fxe"},
	{0x1244D3634EB42E96ULL, "vxw_bonepitdestroyed_a3_1lu_0_0.vlm"},
	{0x1244E1053411D1DCULL, "vxlt_gallowscourtyard_n_6yn_0_0.vlm"},
	{0x1244F7DD0BF7294DULL, "570691144.ogg"},
	{0x124506FF9213F105ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_min_c.impa.evt"},
	{0x12459CB5128E50E4ULL, "6107070_f.fxe"},
	{0x1245A9E698052CECULL, "644964483.ogg"},
	{0x1245E1946B2EE02CULL, "142172629.ogg"},
	{0x12465CF7883A81FBULL, "and110_refugee6111607.cl"},
	{0x12467F9724035F6EULL, "115238835.ogg"},
	{0x124684623712CC63ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_hed_suca_0t.dds"},
	{0x12471A6B1ADC6161ULL, "70235325.ogg"},
	{0x12473EB2F4590446ULL, "dae000_bartrand6206933.cl"},
	{0x1247AF1E084B4816ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayin.msh"},
	{0x1247B91E084B5914ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayin.mmh"},
	{0x1247BD1E084B6027ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayin.mao"},
	{0x12481FB94DAF14A3ULL, "vxw_encounter2_camp_ad_0_0.vlm"},
	{0x12482379116A54D0ULL, "hlac_ht_bartrand_n_3374_1y.dds"},
	{0x12494B2F512358E3ULL, "vxlt_0_lowtown_a2_n_c_ews_0_0.vlm"},
	{0x12499A681F599BD0ULL, "root_war_offense.dds"},
	{0x124AC01D06ABE631ULL, "w_bonepitmining_a2_85_0.mmh"},
	{0x124ACC1D06ABFAD2ULL, "w_bonepitmining_a2_85_0.mao"},
	{0x124ADA1D06AC12EFULL, "w_bonepitmining_a2_85_0.msh"},
	{0x124AF02EEC326C3DULL, "hlac_lt_undercitydungeon_-1_31x.dds"},
	{0x124AF7904F9E4D5AULL, "hlac_lt_hangedman_n_-1_16y.dds"},
	{0x124C5EFB904B2193ULL, "vxht_chantry_2cy_0_0.vlm"},
	{0x124C71EFCE9046B3ULL, "ave200_donnic6060809.cl"},
	{0x124C8957E63173D7ULL, "art\\characters\\weapons\\mauls\\w_mal_ml101a.phy"},
	{0x124CDA56286FD874ULL, "art\\vfx\\combat\\vfx_m_cre_heal_imp_c.anb"},
	{0x124CDF863B9084D8ULL, "vxlt_0_lowtown_a2_n_c_eot_0_0.vlm"},
	{0x124CEA401CF0E0CFULL, "w_blightlands_78_0.mao"},
	{0x124CEBA77B16039FULL, "vxht_chantry_n_351_0_0.vlm"},
	{0x124CF6401CF0F56CULL, "w_blightlands_78_0.mmh"},
	{0x124CFC401CF0FF5EULL, "w_blightlands_78_0.msh"},
	{0x124D076652249C6BULL, "102027797.ogg"},
	{0x124D19CE299742EBULL, "92352842.ogg"},
	{0x124D4F3A6AA6256CULL, "425854356.ogg"},
	{0x124DCB78E9194634ULL, "dae000_dalish_apprentices6147568.cl"},
	{0x124EF9926C680B69ULL, "262835223.ogg"},
	{0x124F157D31A8D72EULL, "vxw_encounter1_k6_0_0.vlm"},
	{0x124F709C34648877ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexb.phy"},
	{0x124FC866C2800BA2ULL, "difficulty.gda"},
	{0x12500DA00C33BAFAULL, "6125909_m.fxe"},
	{0x12502D0DB4316A7CULL, "vxw_blightlands_ob_0_0.vlm"},
	{0x12503B59C312C32CULL, "art\\vfx\\combat\\vfx_f_mer_creep_act_c.impa.ani"},
	{0x1251C652F0444031ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv1.ani"},
	{0x1251F10EDD93EB94ULL, "and100_anders6206127.cl"},
	{0x1251F1A6CB73FFEBULL, "6117260_m.fxe"},
	{0x1251FE16DD7D99BFULL, "6208672_m.fxe"},
	{0x1252133FBCA5B0C2ULL, "vxlt_darktown_2uy_0_0.vlm"},
	{0x125232C7AEF7DB1CULL, "354607999.ogg"},
	{0x1252B28ADA148023ULL, "lgt110_orz_harr_carta6109039.cl"},
	{0x1252DB934592FB1CULL, "vxdr_primevaltomb_7ra_0_0.vlm"},
	{0x125338745425C55FULL, "lgt210ip_ingredient2.utp"},
	{0x125381E0E2C9D3BBULL, "380886882.ogg"},
	{0x125396D5B1C1B2C5ULL, "vxw_bonepit_12h_0_0.vlm"},
	{0x1253CC81B55870D6ULL, "art\\characters\\creatures\\dragon\\animation\\tal_slam.ani"},
	{0x125473AC748D3B8FULL, "523023480.ogg"},
	{0x125481C178B3CA08ULL, "art\\characters\\weapons\\greatswords\\w_gsw_012a.msh"},
	{0x125493C178B3E899ULL, "art\\characters\\weapons\\greatswords\\w_gsw_012a.mao"},
	{0x12549FC178B3FD32ULL, "art\\characters\\weapons\\greatswords\\w_gsw_012a.mmh"},
	{0x1254B540AB32B685ULL, "330251924.ogg"},
	{0x1254CEFA9FFBDEA4ULL, "vxw_bonepitdestroyed_a3_1f1_0_0.vlm"},
	{0x1254DFE0AB5FBE2BULL, "vxht_mansion_n_610_0_0.vlm"},
	{0x1255027FAAADE7DCULL, "ser221_gascard6122459.cl"},
	{0x1255592ABC0EE8CCULL, "dre000_dougal6131162.cl"},
	{0x12558F167E52E299ULL, "vxht_playershouse_n_1va_0_0.vlm"},
	{0x12561D7AF8B64CACULL, "mag311_nyssa6042232.cl"},
	{0x12562A38B0584F19ULL, "686661532.ogg"},
	{0x12569A0AA92BC9BAULL, "dae000_gamlen6048463.cl"},
	{0x1256DA08BF649247ULL, "sib111st_mother.stg"},
	{0x1256DA537A727163ULL, "hlac_w_sundermount_-65538_0z.dds"},
	{0x1256E5B4950D34D8ULL, "art\\characters\\weapons\\greataxes\\w_gax_006a.phy"},
	{0x125740E90C363269ULL, "vxlt_0_lowtown_docks_a3_d3o_0_0.vlm"},
	{0x12574859C318F86EULL, "art\\vfx\\combat\\vfx_f_mer_creep_act_c.impa.gad"},
	{0x1257740DFFBB4766ULL, "955858981.ogg"},
	{0x125792941EEDD46EULL, "6170921_m.fxe"},
	{0x1257D00D051AA386ULL, "6037039_m.fxe"},
	{0x1258A5F4DF2885C6ULL, "llac_lt_gallowscourtyard_-1_4.dds"},
	{0x1258BA8DEA9DCC7EULL, "qcr000_cut2d_thechampion6025124.cl"},
	{0x1258D152F04A71D3ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv1.gad"},
	{0x1259513ECB6E9095ULL, "dwarfmale.fxa"},
	{0x12599216E8087F30ULL, "follower_banter6151462.cl"},
	{0x12599DEC92B36C7EULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.anb"},
	{0x1259F5BB94A98203ULL, "196241679.ogg"},
	{0x125A1EC2957FCBC6ULL, "vxlt_hangedman_1o1_0_0.vlm"},
	{0x125A5B13790EF163ULL, "6117989_m.fxe"},
	{0x125AB956287BDD37ULL, "art\\vfx\\combat\\vfx_m_cre_heal_imp_c.mmh"},
	{0x125AC7AFF9FE35B1ULL, "6063837_m.fxe"},
	{0x125C1C590A8C0367ULL, "qun110tr_petrice_mugging.ncs"},
	{0x125C57E438993ADCULL, "vxdr_0_deeproads_f5w_0_0.vlm"},
	{0x125CECE5C7612703ULL, "hlac_w_sundermount_-393211_0z.dds"},
	{0x125D12D87F76747AULL, "hlac_lt_0_lowtown_docks_a3_-4_2x.dds"},
	{0x125D706E80A56846ULL, "gen00fl_fenris6126467.cl"},
	{0x125D773C479FA86FULL, "vxlt_gallowscourtyard_46q_0_0.vlm"},
	{0x125DC1BE2F29A34CULL, "gen00fl_varric6088831.cl"},
	{0x125DCB6076A7FF83ULL, "vxw_woundedcoast_1cr_0_0.vlm"},
	{0x125E1A63032291C3ULL, "vxw_bonepit_265_0_0.vlm"},
	{0x125E4359C31F0E4DULL, "art\\vfx\\combat\\vfx_f_mer_creep_act_c.impa.evt"},
	{0x125E4755BE763DC5ULL, "dae000_tranquil_16165099.cl"},
	{0x125E5807E40534E8ULL, "ts_kirkwall_03.gda"},
	{0x125E767C94A95124ULL, "vxlt_0_lowtown_a3_n_c_etx_0_0.vlm"},
	{0x125E9EEAA62782FCULL, "vxw_blightlands_pc_0_0.vlm"},
	{0x125ECE52F04ED838ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv1.evt"},
	{0x125EE1CB1EC79A4AULL, "vxlt_warehouse_10h_0_0.vlm"},
	{0x12605BC17F371BC3ULL, "hlac_lt_0_lowtown_a2_-65536_0x.dds"},
	{0x12606383BD6A186BULL, "6112160_m.fxe"},
	{0x1260B27821C4D944ULL, "vxht_playershouse_n_1jt_0_0.vlm"},
	{0x1260D481B56308C7ULL, "art\\characters\\creatures\\dragon\\animation\\tal_slam.evt"},
	{0x12611F2AB444A169ULL, "gen_im_trs_rdr_hvy_boo_r2.uti"},
	{0x12622B4D5E9E65A2ULL, "vxht_chantry_n_2e5_0_0.vlm"},
	{0x1262A16D013577D3ULL, "mag320_evelina.bnk"},
	{0x126340D791F1F724ULL, "vxw_bonepitdestroyed_a3_1k6_0_0.vlm"},
	{0x1263554A48E8CED0ULL, "vxht_0_hightown_ca1_0_0.vlm"},
	{0x1263CB9B5A893F6CULL, "vxdr_primevalentrance_277_0_0.vlm"},
	{0x12646ED40CFBCB48ULL, "dae200cr_captain_wright.cnv"},
	{0x126472E3AE52C156ULL, "vxdr_primevaltomb_7p5_0_0.vlm"},
	{0x12650C44F6A83AD9ULL, "mrl150_hunter6071298.cl"},
	{0x1265726FA130EB5EULL, "6016319_m.fxe"},
	{0x1265C36D0137E723ULL, "mag320_evelina.cnv"},
	{0x1265E6E2677470C0ULL, "6164571_m.fxe"},
	{0x12662309C01B42A4ULL, "820605670.ogg"},
	{0x12663228C0CFCAF1ULL, "6149078_m.fxe"},
	{0x1266A22C87BBE13DULL, "vxlt_warehouse_121_0_0.vlm"},
	{0x1266FF5E3241C180ULL, "icicle01.dds"},
	{0x12671F0E7B967EF7ULL, "hlac_w_bonepitmining_a2_131078_0y.dds"},
	{0x126790D40CFE3AB0ULL, "dae200cr_captain_wright.bnk"},
	{0x1267AD7A7D84F464ULL, "dae011_corff_sta6149635.cl"},
	{0x1267F581B5695F10ULL, "art\\characters\\creatures\\dragon\\animation\\tal_slam.gad"},
	{0x12680FB40E168B72ULL, "gen_im_acc_amu_act3_01.uti"},
	{0x126828E59507DB96ULL, "hlac_ht_0_hightown_a3_131072_0z.dds"},
	{0x1268D66B6C7EAC18ULL, "587532888.ogg"},
	{0x126904C0BF441D20ULL, "zz_mcr_debug6088441.cl"},
	{0x12693C493BEC21F6ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg102a_0t.dds"},
	{0x1269B8C22955C889ULL, "lt_0_lowtown_docks_181_0.mao"},
	{0x1269C4C22955DD22ULL, "lt_0_lowtown_docks_181_0.mmh"},
	{0x1269C6C22955E0D8ULL, "lt_0_lowtown_docks_181_0.msh"},
	{0x1269CE5EF935C9D1ULL, "6076380_f.fxe"},
	{0x1269D99F3653E864ULL, "dae021_sister_colinda_sta6150202.cl"},
	{0x126A6FA5C93D1364ULL, "6076049_m.fxe"},
	{0x126A7C0D2E26ABA6ULL, "m_dae321ar_player_home_nt.wbd"},
	{0x126A993372E04228ULL, "hlac_lt_foundry_n_320_7x.dds"},
	{0x126A9A9E09F4DB79ULL, "follower_banter6173029.cl"},
	{0x126C3D3D689E1803ULL, "vxw_woundedcoast_1nu_0_0.vlm"},
	{0x126C8C3FF518AA43ULL, "vxw_bonepit_214_0_0.vlm"},
	{0x126C93C4ABBD25EBULL, "hlac_ht_keep_4519_16x.dds"},
	{0x126DE0FFF2FD4E8DULL, "6103977_m.fxe"},
	{0x126E44CED284FF32ULL, "811550749.ogg"},
	{0x126E550B6487A78DULL, "ht_0_hightownexplosion_n_27_0.rml"},
	{0x126EDE3C8EF77ED9ULL, "lt_0_lowtown_docks_a3_n_152_0.phy"},
	{0x126F273366906CF5ULL, "vxht_mansion2b_27s_0_0.vlm"},
	{0x126F568678AF4D95ULL, "ht_0_hightown_16_0_ll.mmh"},
	{0x126FBA7CBCFCB10FULL, "vxdr_0_deeproads_7tw_0_1.vlm"},
	{0x126FDA13A1DBCD22ULL, "6142065_m.fxe"},
	{0x12702E04DC493F58ULL, "269569018.ogg"},
	{0x1270323F7ABE3171ULL, "character_character_p134.psh.fxo"},
	{0x1270C9B99CB16C70ULL, "gen00fl_isabela6090863.cl"},
	{0x127140F68D11F144ULL, "vxlt_0_lowtown_docks_a3_c1k_0_0.vlm"},
	{0x127164EC9E22CC41ULL, "mag100_cullen6093557.cl"},
	{0x1271B2B483E80FA4ULL, "vxw_bonepitdestroyed_a3_1l7_0_0.vlm"},
	{0x1271C41BF1767516ULL, "vxlt_0_lowtown_docks_b7e_0_0.vlm"},
	{0x1271F5002A9DF6EBULL, "664394397.ogg"},
	{0x1271FFE83EE77902ULL, "vxca_caverns1_mines_jlp_0_0.vlm"},
	{0x127236471BE0D922ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attk_v2.gad"},
	{0x127313D6EA3657E7ULL, "6095903_f.fxe"},
	{0x12732F348A69EBC4ULL, "vxw_bonepitmining_a2_1p9_0_0.vlm"},
	{0x1273430FC210B66FULL, "vxw_encounter1_df_0_0.vlm"},
	{0x1273493E529DB393ULL, "gen_im_trs_gen_blt_r2.uti"},
	{0x127400B1D5F47E1CULL, "zz_dae_debug6036592.cl"},
	{0x12747BEA04DF2965ULL, "gen00fl_dog6109286.cl"},
	{0x12747FE1DF19C1E7ULL, "hlac_lt_hangedman_n_-1_0y.dds"},
	{0x1274A3712B6AE947ULL, "vxdr_primevalentrance_1ia_0_0.vlm"},
	{0x1274A405B2C5E371ULL, "6142178_m.fxe"},
	{0x12750DEF56D2A839ULL, "vxlt_warehouse_n_5p_0_0.vlm"},
	{0x12751AAB9B76EB32ULL, "vxw_bonepit_1x5_0_0.vlm"},
	{0x12754859E428E7C8ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn8plank.mmh"},
	{0x12755459E428FC6BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn8plank.mao"},
	{0x12755FCF3EB85051ULL, "124516217.ogg"},
	{0x12756659E4291AF2ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn8plank.msh"},
	{0x12757B71D6AC09C8ULL, "sp_module_travel.ncs"},
	{0x12762F56B353CE2BULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.impa.ani"},
	{0x12767A3F164F2C3CULL, "hlac_ht_mansion2_818_63y.dds"},
	{0x1276A333863822B2ULL, "6018784_m.fxe"},
	{0x1276D1BBF5D04789ULL, "vxlt_ruinedhouse_n_hm_0_0.vlm"},
	{0x1276F93A77186E61ULL, "cs_rom000cs_fenris.wbd"},
	{0x1277A634A54A9ADDULL, "vxca_caverns1_mines_inv_0_0.vlm"},
	{0x1277B1E9A327E3DAULL, "584160763.ogg"},
	{0x1277B492662D6B13ULL, "vxht_chantry_2rz_0_0.vlm"},
	{0x1278352E5628458DULL, "824811395.ogg"},
	{0x1278D7CA14B2FA54ULL, "art\\vfx\\creatures\\fxc_ogre_db_d_c.impa.ani"},
	{0x1278F0B3F2951874ULL, "fx_dust_sms01.dds"},
	{0x127959471BE73319ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attk_v2.evt"},
	{0x1279C4F92477D749ULL, "vxht_bartrand_n_5qz_0_0.vlm"},
	{0x1279DA1DF3FCC868ULL, "vxlt_0_lowtown_a2_n_db4_0_0.vlm"},
	{0x127A72D266BA16E8ULL, "131322417.ogg"},
	{0x127A7B351EA01367ULL, "vrc221_varric6040821.cl"},
	{0x127AAF1A5A943083ULL, "vxw_woundedcoast_1it_0_0.vlm"},
	{0x127B49BE49B59B7DULL, "hlac_lt_warehouse_-65536_0z.dds"},
	{0x127BBD5A627D6DACULL, "6047283_m.fxe"},
	{0x127C24F47EE92FB6ULL, "6132673_m.fxe"},
	{0x127C403A910ABD03ULL, "art\\vfx\\combat\\vfx_str_att_front_l_leg_c.impa.ani"},
	{0x127C43A9D88BA081ULL, "528537087.ogg"},
	{0x127CABD6AE151343ULL, "mcr310_sibling6095347.cl"},
	{0x127CC94AE5738733ULL, "hla_ht_0_hightown_a2_2_0y.dds"},
	{0x127D58C248C56C42ULL, "hlac_lt_0_lowtown_docks_n_-65540_2z.dds"},
	{0x127D7682B85F3FEAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgb_1.mao"},
	{0x127D9F09A544E55CULL, "514738075.ogg"},
	{0x127DC02EBE35201DULL, "6085235_m.fxe"},
	{0x127E68BAC0AEA78CULL, "101748896.ogg"},
	{0x127F3353C2635B9DULL, "vxlt_gallowstemplar_n_6sw_0_0.vlm"},
	{0x127F7A4DC52EB570ULL, "vxht_mansion2b_3dv_0_0.vlm"},
	{0x127FCDFAECA2697CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkbackright.ani"},
	{0x127FD4CA14B913B6ULL, "art\\vfx\\creatures\\fxc_ogre_db_d_c.impa.gad"},
	{0x127FDDB51E0D93C5ULL, "art\\vfx\\combat\\vfx_glb_sleep_c.anb"},
	{0x12804AF023E025B8ULL, "hlac_lt_0_lowtown_n_-65536_2x.dds"},
	{0x12806EA8ECF29C67ULL, "6068694_m.fxe"},
	{0x128074A45CC70987ULL, "6116134_f.fxe"},
	{0x1280D1349BA4CF4EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_wallcap.msh"},
	{0x1280DF349BA4E69FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_wallcap.mao"},
	{0x1280EB349BA4FB3CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_wallcap.mmh"},
	{0x12817B624E4AC3ECULL, "6127404_m.fxe"},
	{0x128187DBE594877CULL, "6152729_m.fxe"},
	{0x1281B4ECB406CEEFULL, "vxw_encounter1_ia_0_0.vlm"},
	{0x1282BDEB439DED21ULL, "art\\levels\\races\\proxy\\props\\_textures\\tables_sa.dds"},
	{0x1282C2654706DEB8ULL, "vxw_encounter2_v_0_0.vlm"},
	{0x1282CBB550F6F8BFULL, "ave331cr_insurgent_guard2.utc"},
	{0x1282D803685F2D5BULL, "mcr371_hostile_aveline6132464.cl"},
	{0x1283015DF3DD5FDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_1.evt"},
	{0x12837FCC48C8C0B9ULL, "vxlt_warehouse_n_8w_0_0.vlm"},
	{0x1283D808765A8C0CULL, "eye.mat.dx11"},
	{0x12841756B35FE2A2ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.impa.evt"},
	{0x1284B4DB31F86391ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngan.mmh"},
	{0x1284C02DE5ED8FB9ULL, "mrl000_marethari6069867.cl"},
	{0x1284C0DB31F87832ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngan.mao"},
	{0x1284CEDB31F8904FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngan.msh"},
	{0x1284F195A2536BA1ULL, "zz_prmpt_debug6198270.cl"},
	{0x12852BAF2FAD0A56ULL, "6050159_m.fxe"},
	{0x12853D6E97DBF83AULL, "lt_0_lowtown_46_0.rml"},
	{0x128591F300551310ULL, "6095266_m.fxe"},
	{0x1285CC4F20A7BFDBULL, "310671805.ogg"},
	{0x1285EFCA14BDAD15ULL, "art\\vfx\\creatures\\fxc_ogre_db_d_c.impa.evt"},
	{0x128636B6288363DBULL, "zz_dre_debug6056963.cl"},
	{0x128661471BF1CB08ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attk_v2.ani"},
	{0x1286A4C0D6E602A2ULL, "vxlt_hangedman_n_19d_0_0.vlm"},
	{0x1286B6669BF6DE5EULL, "6152764_f.fxe"},
	{0x1286DAFAECA89DBEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkbackright.gad"},
	{0x128706208B466B9CULL, "def_i_clo.mao"},
	{0x12875F5C5831EB41ULL, "dae340st_jansen.stg"},
	{0x128836D6166DEFC9ULL, "vxht_bartrand_n_5ty_0_0.vlm"},
	{0x12886386AC97764DULL, "28253414.ogg"},
	{0x128891430A1271C0ULL, "91555962.ogg"},
	{0x128899E77DDAE9BDULL, "lt_0_lowtown_docks_192_0.mao"},
	{0x12889DE77DDAF0F6ULL, "lt_0_lowtown_docks_192_0.mmh"},
	{0x1288A7E77DDB01F4ULL, "lt_0_lowtown_docks_192_0.msh"},
	{0x1288D21C22BCC6D4ULL, "au_cl.mao"},
	{0x1288F7DB60CF7853ULL, "576824515.ogg"},
	{0x12894AADD124D473ULL, "71864058.ogg"},
	{0x1289583A911570FAULL, "art\\vfx\\combat\\vfx_str_att_front_l_leg_c.impa.evt"},
	{0x1289A7E92A5DABE7ULL, "255337295.ogg"},
	{0x1289B301CB41E621ULL, "vxca_cave_1dt_0_1.vlm"},
	{0x1289C2AF40A54D7CULL, "990326025.ogg"},
	{0x1289F055C34A6A04ULL, "dre000_dougal6142216.cl"},
	{0x1289FA5DF3E3721CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_1.gad"},
	{0x1289FFB8C3175853ULL, "386279978.ogg"},
	{0x128A9A690576D211ULL, "mag100_thrask6032066.cl"},
	{0x128AB3BA24A18BD9ULL, "hlac_lt_0_lowtown_a2_n_131072_0y.dds"},
	{0x128B1656B365FFB5ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.impa.gad"},
	{0x128B2FB904CA8B64ULL, "hla_ht_0_hightown_-65536_0x.dds"},
	{0x128B3C66B69F309CULL, "6153887_m.fxe"},
	{0x128C94DC9931510DULL, "art\\levels\\races\\proxy\\props\\_textures\\cheese_d.dds"},
	{0x128C9F063714AA53ULL, "vxht_mansion2_1pw_0_0.vlm"},
	{0x128CC54384298AABULL, "vxht_keep_n_6hc_0_0.vlm"},
	{0x128CC8D14181B800ULL, "ht_0_hightown_42_0_ll.mmh"},
	{0x128CCF62AF654A9CULL, "486026381.ogg"},
	{0x128CDC097A494E1DULL, "206873559.ogg"},
	{0x128D53E96597C560ULL, "gen00fl_anders6125985.cl"},
	{0x128D6A8018DA18C2ULL, "vxlt_undercityencounter_1id_0_0.vlm"},
	{0x128D72288C8BE7ABULL, "art\\vfx\\combat\\vfx_w_wns_front_close5_c.mmh"},
	{0x128D85A2331285E2ULL, "follower_banter6165412.cl"},
	{0x128D867FF1B7159EULL, "hla_ht_0_hightown_a3_65535_0y.dds"},
	{0x128E6C3C05573A6BULL, "follower_banter6172266.cl"},
	{0x128E833F5EFD61BFULL, "vxdr_primevalentrance_278_0_0.vlm"},
	{0x128F0F5AF52CFA0CULL, "hlac_lt_gallowstemplar_f_2382_0z.dds"},
	{0x128F5486F11704FAULL, "vxw_bonepitmining_a2_1ne_0_0.vlm"},
	{0x1290145DFBFD3B84ULL, "6021690_m.fxe"},
	{0x1290573A911B8DEDULL, "art\\vfx\\combat\\vfx_str_att_front_l_leg_c.impa.gad"},
	{0x12906C3A48B81279ULL, "lgt160_ntb_elf_conflict6163377.cl"},
	{0x1290B3F0D012CD08ULL, "6126414_fen_from_you.cut"},
	{0x1290F5584820B703ULL, "137980717.ogg"},
	{0x1290F6344153F63CULL, "zz_one_debug6078372.cl"},
	{0x1290F95DF3E98F1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_1.ani"},
	{0x12915ACCE6F3A991ULL, "745741067.ogg"},
	{0x12917ECDBF1BC592ULL, "6040811_m.fxe"},
	{0x1291CFFF026606DAULL, "hlac_lt_darktown_-1_7y.dds"},
	{0x1291E30F7AB68BEFULL, "zz_dre_debug6054675.cl"},
	{0x1291FD18A75B5B08ULL, "6167633_m.fxe"},
	{0x129273E5B5D5A01AULL, "68575720.ogg"},
	{0x12930237CDE436B9ULL, "lt_0_lowtown_a3_n_c_8_0.msh"},
	{0x12930C37CDE447A7ULL, "lt_0_lowtown_a3_n_c_8_0.mmh"},
	{0x12931037CDE44EF4ULL, "lt_0_lowtown_a3_n_c_8_0.mao"},
	{0x1293383BADB6AC55ULL, "vxht_mansion2_23w_0_0.vlm"},
	{0x12934264253A4506ULL, "mag360_samson6042105.cl"},
	{0x12935D6D0435D7C5ULL, "ht_mansion_n_3364_lc.mmh"},
	{0x129388D40D23684DULL, "dae200cr_captain_wright.utc"},
	{0x129398399F0DC9B4ULL, "vfx_distortionmask_p0.psh.fxo.dx11.assembly"},
	{0x1293D48310D3B327ULL, "hlac_lt_gallowstemplar_n_7198_4y.dds"},
	{0x129419FA945AFA81ULL, "dae340_db_drag_rog.cut"},
	{0x12946A57E66EF4CCULL, "art\\characters\\weapons\\mauls\\w_mal_ml101a.mmh"},
	{0x12946DC0BE42FB7BULL, "6164356_m.fxe"},
	{0x12947057E66EFEBEULL, "art\\characters\\weapons\\mauls\\w_mal_ml101a.msh"},
	{0x1294A01ACDF4E5CBULL, "6139703_m.fxe"},
	{0x1294C8B07B69AEB6ULL, "590390989.ogg"},
	{0x129504C995E5534DULL, "vxca_caverns2_70x_0_0.vlm"},
	{0x1295061026AACFEDULL, "vxlt_0_lowtown_docks_a3_bge_0_0.vlm"},
	{0x1295399C1EBA0FDDULL, "fx_smokelink01.dds"},
	{0x12953F8BE3357FA3ULL, "hlac_lt_gallowscourtyard_n_-1_3y.dds"},
	{0x1296149A77D27B2EULL, "vxlt_hangedman_27f_0_0.vlm"},
	{0x1296310F3A71D32BULL, "vxlt_warehouse_9w_0_0.vlm"},
	{0x1296814C56E3E78AULL, "one300_nuncio6075044.cl"},
	{0x12969F43D0B17746ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_exit.ani"},
	{0x1296A0E6D0EA4709ULL, "6176472_m.fxe"},
	{0x1297459C34A1F4CFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexb.mao"},
	{0x1297519C34A2096CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexb.mmh"},
	{0x1297579C34A2135EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexb.msh"},
	{0x12976178D5F00CF3ULL, "hlac_lt_0_lowtown_65537_1y.dds"},
	{0x129789E7CB81D1D6ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgc.msh"},
	{0x129793E7CB81E2D4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgc.mmh"},
	{0x129797E7CB81E9E7ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgc.mao"},
	{0x1297EC73C6F53BBCULL, "vxht_playershouse_n_1v8_0_0.vlm"},
	{0x1297F8F0126C4650ULL, "212835216.ogg"},
	{0x1297FDBE61433703ULL, "ht_keep_n_mmi.gff"},
	{0x1298C280DEEEA45DULL, "vxlt_0_lowtown_a2_n_e1y_0_0.vlm"},
	{0x1298D64BDFF1C02BULL, "vxlt_0_lowtown_a2_eml_0_0.vlm"},
	{0x129A1C6D04763741ULL, "ht_0_hightown_16_0.rml"},
	{0x129A2418445E6F93ULL, "197408385.ogg"},
	{0x129A38C5F344897CULL, "6083759_m.fxe"},
	{0x129A789C89F6A745ULL, "hlac_lt_undercitydungeon_-1_25y.dds"},
	{0x129B361E45D95397ULL, "638160452.ogg"},
	{0x129B3720761FA32BULL, "vxht_keep_n_6cb_0_0.vlm"},
	{0x129B53288C97F010ULL, "art\\vfx\\combat\\vfx_w_wns_front_close5_c.anb"},
	{0x129BADD4B94CE315ULL, "vxlt_0_lowtown_docks_n_bao_0_0.vlm"},
	{0x129BDC5D0AD03142ULL, "vxlt_undercityencounter_1fe_0_0.vlm"},
	{0x129C1B3A7737C463ULL, "cs_rom000cs_fenris.bnk"},
	{0x129CD1CDB6A8C614ULL, "6090391_m.fxe"},
	{0x129CE5B606810FEFULL, "vfx_env_dust_01.mao"},
	{0x129D1CD2BADE3F5CULL, "mag101_alain6032127.cl"},
	{0x129DC663E30D1D7AULL, "vxw_bonepitmining_a2_1kf_0_0.vlm"},
	{0x129E5370A0BF991FULL, "art\\characters\\creatures\\mubari\\animation\\eattwitch.gad"},
	{0x129F25C06EF55CD1ULL, "hlac_ht_playershouse_n_855_29y.dds"},
	{0x129F6750A9A6E05EULL, "6083714_f.fxe"},
	{0x129F7E490F0ABEE5ULL, "dae320ip_at_brothel.utp"},
	{0x12A0015B37BD3782ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb5.mmh"},
	{0x12A0035B37BD3B38ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb5.msh"},
	{0x12A01E43997665C5ULL, "6030636_f.fxe"},
	{0x12A043C1351A59E7ULL, "hlac_lt_0_lowtown_a2_n_c_65538_2x.dds"},
	{0x12A052EE3C4DBC7CULL, "6127660_m.fxe"},
	{0x12A0965138BF41B7ULL, "vxlt_gallowstemplar_n_4ss_0_0.vlm"},
	{0x12A0A5C57F75B738ULL, "6134949_m.fxe"},
	{0x12A16984ECED3D8CULL, "w_encounter2_camp_20_0_lc.mmh"},
	{0x12A1B308A0314DE8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robb_0s.dds"},
	{0x12A1E36107A0988DULL, "6149867_m.fxe"},
	{0x12A1F967DD6E27E1ULL, "6126675_m.fxe"},
	{0x12A1FD5BE2D6137BULL, "ttx_blocker_b_1.mmh"},
	{0x12A20F6AB2B9860CULL, "883507157.ogg"},
	{0x12A21B5BE2D64665ULL, "ttx_blocker_b_1.msh"},
	{0x12A21D1A42942BADULL, "follower_banter6141873.cl"},
	{0x12A3233B4FC85D12ULL, "vxlt_gallowscourtyard_75n_0_0.vlm"},
	{0x12A328372E0957CFULL, "dae360ar_wounded_coast.are"},
	{0x12A32E36A99D6B83ULL, "zz_vault_debug6203382.cl"},
	{0x12A33C5FFF9626FFULL, "812221063.ogg"},
	{0x12A36A8134393536ULL, "vxlt_gallowstemplar_n_5qq_0_0.vlm"},
	{0x12A37DDC5880B3E6ULL, "vxlt_foundry_n_z0_0_0.vlm"},
	{0x12A3B81E0899B60FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayin.phy"},
	{0x12A415BC0969944FULL, "hlac_lt_0_lowtown_0_3y.dds"},
	{0x12A4605BF63DF012ULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_c.impa.gad"},
	{0x12A48743D0BD8BB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_exit.evt"},
	{0x12A4D4874506A844ULL, "402881676.ogg"},
	{0x12A509E257006E0CULL, "587780692.ogg"},
	{0x12A53364781F52D1ULL, "vxht_mansion2_32e_0_0.vlm"},
	{0x12A54870A0C5A50CULL, "art\\characters\\creatures\\mubari\\animation\\eattwitch.evt"},
	{0x12A57477C4628924ULL, "ts_kirk_darktown_prog_junk1.gda"},
	{0x12A58761D296FBADULL, "hlac_ht_0_hightown_65536_1z.dds"},
	{0x12A5A09D3CA3477FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocklarge.msh"},
	{0x12A5A69D3CA35121ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocklarge.mmh"},
	{0x12A5B29D3CA365C2ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrocklarge.mao"},
	{0x12A5CA349BC3DF07ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_wallcap.phy"},
	{0x12A5F9BF662C0913ULL, "6097270_m.fxe"},
	{0x12A6B2090968024EULL, "99400037.ogg"},
	{0x12A6D372FE9518E1ULL, "1775957.ogg"},
	{0x12A8937DA301ADCEULL, "w_sundermount_128_0.msh"},
	{0x12A8A17DA301C51FULL, "w_sundermount_128_0.mao"},
	{0x12A8AD7DA301D9BCULL, "w_sundermount_128_0.mmh"},
	{0x12A8DEB51E309A06ULL, "art\\vfx\\combat\\vfx_glb_sleep_c.mmh"},
	{0x12A8E52C9CE6CED8ULL, "follower_banter6149306.cl"},
	{0x12A95815C8F93E24ULL, "vxdr_primevaltomb_8v6_0_0.vlm"},
	{0x12A9DE49693B6B17ULL, "vxw_sundermount_33q_0_1.vlm"},
	{0x12AB2D487DDCB641ULL, "vxw_encounter2_camp_p9_0_0.vlm"},
	{0x12AB635BF6441449ULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_c.impa.evt"},
	{0x12AB8843D0C3AC40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_exit.gad"},
	{0x12ABF99C740073F6ULL, "605283767.ogg"},
	{0x12AC08A27657E346ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_exit.evt"},
	{0x12AC63D03AD4417FULL, "96736152.ogg"},
	{0x12AD6404A8B821BDULL, "372110609.ogg"},
	{0x12AD6E6A5F587123ULL, "mcr310_cut_chantry_blows_up6087566.cl"},
	{0x12AE076475C6CA3BULL, "6103363_f.fxe"},
	{0x12AE19622365550DULL, "one000_ilen6037075.cl"},
	{0x12AE39D54353D10CULL, "vxca_cavern1_haunted_e7h_0_0.vlm"},
	{0x12AE47A76DA8629BULL, "follower_barks6153826.cl"},
	{0x12AE90EBDE938320ULL, "6164434_m.fxe"},
	{0x12AEE21F634F4D32ULL, "vxw_bonepit_1l1_0_0.vlm"},
	{0x12AF71231AD86BEEULL, "hlac_lt_undercityserialkiller_-1_45y.dds"},
	{0x12AFBE37DF79707EULL, "vxlt_0_lowtown_euy_0_0.vlm"},
	{0x12B01006E3B51E5CULL, "6128774_f.fxe"},
	{0x12B035E4E94E122DULL, "6076062_m.fxe"},
	{0x12B0CF2F02692545ULL, "face_face_p6.psh.fxo.dx11.assembly"},
	{0x12B0DB5F50096A68ULL, "433946854.ogg"},
	{0x12B1065409D839D7ULL, "vrc321st_haunted_vase.stg"},
	{0x12B109590AD422BEULL, "qun110tr_petrice_mugging.utt"},
	{0x12B1BEE5F5C7F48CULL, "6147658_f.fxe"},
	{0x12B1E427C936774CULL, "w_bonepit_172_0.msh"},
	{0x12B1EA27C936813EULL, "w_bonepit_172_0.mmh"},
	{0x12B1F627C93695D5ULL, "w_bonepit_172_0.mao"},
	{0x12B27D4D9E5DD42EULL, "lgt111pt_gxa_coa.ncs"},
	{0x12B2D2B4955B736DULL, "art\\characters\\weapons\\greataxes\\w_gax_006a.msh"},
	{0x12B2E0B4955B8AA0ULL, "art\\characters\\weapons\\greataxes\\w_gax_006a.mao"},
	{0x12B2E4B4955B91F3ULL, "art\\characters\\weapons\\greataxes\\w_gax_006a.mmh"},
	{0x12B31FA2765E2891ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_exit.gad"},
	{0x12B33070A0D1BA65ULL, "art\\characters\\creatures\\mubari\\animation\\eattwitch.ani"},
	{0x12B333ED769977DBULL, "hlac_dr_0_deeproads_5845_5x.dds"},
	{0x12B37D3209085275ULL, "t3_wdg_w04.tnt"},
	{0x12B3BCBA25069E9BULL, "rdr111dg_assistant6016650.cl"},
	{0x12B421372E17BE8CULL, "dae360ar_wounded_coast.lst"},
	{0x12B4A1EB2E0D2776ULL, "vrc221_varric6048574.cl"},
	{0x12B547C2F6907FE6ULL, "vxlt_gallowstemplar_f_5q9_0_0.vlm"},
	{0x12B5CEB0ACB654FEULL, "w_encounter2_32_0.rml"},
	{0x12B5DB561D9300C6ULL, "6105749_m.fxe"},
	{0x12B61B9328883FB8ULL, "vxlt_hangedman_n_1ox_0_0.vlm"},
	{0x12B7F49D0CF6F3D3ULL, "vxht_mansion2_1ax_0_0.vlm"},
	{0x12B82BA182031BCAULL, "6075253_m.fxe"},
	{0x12B88650B8F77DBBULL, "dae000_arishok6067714.cl"},
	{0x12B94B5BF65028B8ULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_c.impa.ani"},
	{0x12B9506DF5980B11ULL, "424691670.ogg"},
	{0x12B9577CC602EB0EULL, "water_water_p1.vsh.fxo"},
	{0x12B99F256FD2CEC1ULL, "vxw_encounter2_camp_k8_0_0.vlm"},
	{0x12B9D8FAD2A4AAFEULL, "6166954_f.fxe"},
	{0x12BA20A27664493FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_exit.ani"},
	{0x12BA6EEC463E6F83ULL, "6171817_m.fxe"},
	{0x12BAC373240E3F4FULL, "168920665.ogg"},
	{0x12BB01372E1DA735ULL, "dae360ar_wounded_coast.ncs"},
	{0x12BB0EA02A30360BULL, "6077987_f.fxe"},
	{0x12BB3D48770581A1ULL, "hlac_dr_primevalentrance_1726_1x.dds"},
	{0x12BB58E3D532ECEFULL, "mans_best_friend.dds"},
	{0x12BC0A329924C8D8ULL, "lt_0_lowtown_docks_n_121_0.phy"},
	{0x12BC144197CE852AULL, "vxw_encounter3_m_0_0.vlm"},
	{0x12BC5537BC67539EULL, "vxdr_0_deeproads_d3i_0_0.vlm"},
	{0x12BC7F5D8E4C28E8ULL, "vxlt_0_lowtown_docks_n_ctp_0_0.vlm"},
	{0x12BCE7B0ACBC9DC2ULL, "w_encounter2_32_0.phy"},
	{0x12BCFB0A4A278D2DULL, "zz_qun_debug6083719.cl"},
	{0x12BD08A26CD9978EULL, "6124050_m.fxe"},
	{0x12BDC191369043D6ULL, "zz_ave_debug6056783.cl"},
	{0x12BDCD2960019DB2ULL, "character_character_p5.psh.fxo.dx11.assembly"},
	{0x12BE15E141837C88ULL, "w_bonepitmining_a2_235_0.mmh"},
	{0x12BE1F9369DFA8C7ULL, "357040422.ogg"},
	{0x12BE21E14183912BULL, "w_bonepitmining_a2_235_0.mao"},
	{0x12BE33E14183AFB2ULL, "w_bonepitmining_a2_235_0.msh"},
	{0x12BE45330816FBFBULL, "w_blightlands_92_0.phy"},
	{0x12BE754BB4080706ULL, "lt_0_lowtown_docks_n_33_0_ll.mmh"},
	{0x12BEBA45868604E5ULL, "vxlt_0_lowtown_docks_n_bd0_0_0.vlm"},
	{0x12BEF537CE095BDCULL, "lt_0_lowtown_a3_n_c_8_0.phy"},
	{0x12BF1A6E980D3546ULL, "lt_0_lowtown_46_0.anb"},
	{0x12BF20A193557259ULL, "300261665.ogg"},
	{0x12BFBA2833732CD1ULL, "vxlt_warehouse_tv_0_0.vlm"},
	{0x12BFF576820FA3A6ULL, "background_icons.gda"},
	{0x12C0437A078D757DULL, "6116148_m.fxe"},
	{0x12C04E3B18256636ULL, "vxlt_gallowstemplar_n_5os_0_0.vlm"},
	{0x12C05BA6FC8D196DULL, "vxlt_0_lowtown_docks_a3_bvb_0_0.vlm"},
	{0x12C07A8AA627F907ULL, "594570805.ogg"},
	{0x12C08522B4854AC3ULL, "hlac_ca_caverns2_5802_15z.dds"},
	{0x12C092CE90C1C861ULL, "vxlt_backalley_g_1jn_0_0.vlm"},
	{0x12C11471A4CF595FULL, "mag210_arianni6057343.cl"},
	{0x12C1343EE93B709DULL, "vxdr_primevaltomb_508_0_0.vlm"},
	{0x12C13E997C0EF596ULL, "255743107.ogg"},
	{0x12C25ED17A005761ULL, "mcr370_cut_marriage6098259.cl"},
	{0x12C275975F538005ULL, "6055568_m.fxe"},
	{0x12C2CE45500CC310ULL, "278188923.ogg"},
	{0x12C317F11B74C58CULL, "vxlt_0_lowtown_a3_n_c_hd5_0_0.vlm"},
	{0x12C3670C76544314ULL, "6135933_m.fxe"},
	{0x12C4724C60AC69D0ULL, "vxlt_0_lowtown_n_eva_0_0.vlm"},
	{0x12C4EFCA56FDF08DULL, "569715233.ogg"},
	{0x12C63F6A7F2E2DABULL, "isa211_isabela6202107.cl"},
	{0x12C692A123D4A0A1ULL, "vxht_mansion2_1sd_0_0.vlm"},
	{0x12C6AD7DA31AD70FULL, "w_sundermount_128_0.rml"},
	{0x12C706D012E7469AULL, "501070193.ogg"},
	{0x12C7A7CB76C6D541ULL, "hlac_lt_0_lowtown_131072_6z.dds"},
	{0x12C7D73994D32515ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fin_c.mmh"},
	{0x12C7FA3924A1FC54ULL, "llac_w_encounter2_131076_0.dds"},
	{0x12C8556E3E28BAF8ULL, "aip_coteriearcher_base.gda"},
	{0x12C8BAD79981236EULL, "dae000_male_templar_mob6174368.cl"},
	{0x12C8BE07E781ACC5ULL, "vxw_bonepitmining_a2_1km_0_0.vlm"},
	{0x12C8CE32F7503468ULL, "gen00fl_varric6201276.cl"},
	{0x12CA1B8452167448ULL, "vxlt_backalley_g_1jk_0_0.vlm"},
	{0x12CA6DE87F16172DULL, "hlac_lt_0_lowtown_docks_262143_0y.dds"},
	{0x12CAA334B038DDBBULL, "6072801_f.fxe"},
	{0x12CAF13A80424168ULL, "vxlt_0_lowtown_docks_n_coq_0_0.vlm"},
	{0x12CAFDB806F3DE88ULL, "hlac_ca_caverns1_24425_11z.dds"},
	{0x12CB251D82AE14C3ULL, "gen_im_cft_rec_wrn_03.uti"},
	{0x12CB3150A9304965ULL, "7147448.ogg"},
	{0x12CC2C836F0D9CFAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_right.mmh"},
	{0x12CC30836F0DA331ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_right.mao"},
	{0x12CC3E836F0DBB40ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_right.msh"},
	{0x12CC6CA427A216A3ULL, "isa311_castillon6094907.cl"},
	{0x12CC920B0287BA4EULL, "689148244.ogg"},
	{0x12CC97D0011763D7ULL, "6106086_m.fxe"},
	{0x12CCB2A0F2CBB86FULL, "vxw_bonepitdestroyed_a3_26q_0_0.vlm"},
	{0x12CCD356794D352DULL, "sst_ss_corpse_1.gda"},
	{0x12CCFB6D04A158D1ULL, "ht_0_hightown_16_0.anb"},
	{0x12CCFB8F649021B5ULL, "hlac_lt_darktown_-1_36x.dds"},
	{0x12CD47AEAE18EB44ULL, "fen310_fenris6144530.cl"},
	{0x12CD720E971B5575ULL, "one000_ilen6059251.cl"},
	{0x12CDB7089E9F166CULL, "combat_dw_r1_a3.gda"},
	{0x12CE2C0525694551ULL, "vxlt_warehouse_yu_0_0.vlm"},
	{0x12CE806BD85CEFAFULL, "hlac_lt_undercitydungeon_-1_38y.dds"},
	{0x12CEC0180A1B7EB6ULL, "vxlt_gallowstemplar_n_5bt_0_0.vlm"},
	{0x12CF04AB82B7E0E1ULL, "vxlt_backalley_g_1wm_0_0.vlm"},
	{0x12CF8174BEC6E9B0ULL, "47050387.ogg"},
	{0x12CFCD624A92B3D6ULL, "gen00fl_aveline6139493.cl"},
	{0x12D019FADE185FF0ULL, "hightown_male_noble_gossip6171406.cl"},
	{0x12D0351297EBEAB2ULL, "mer210_coterie_contact6079356.cl"},
	{0x12D03E4E912ED383ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wmda_0s.dds"},
	{0x12D05499A9CBE1ACULL, "6208183_m.fxe"},
	{0x12D06BC7B6002DCFULL, "6060508_m.fxe"},
	{0x12D0766C274A4A61ULL, "lowtown_fem_prostitute_sta6152357.cl"},
	{0x12D10D2911E97DFFULL, "con_ico_endromance.dds"},
	{0x12D154118C7545EAULL, "494535334.ogg"},
	{0x12D1B374BDA20450ULL, "gen00fl_varric6118811.cl"},
	{0x12D2C96FA3EC0FBEULL, "qun230cr_guild_contact6068843.cl"},
	{0x12D30B1A0F46CB52ULL, "299393292.ogg"},
	{0x12D37A8471281FD5ULL, "599489243.ogg"},
	{0x12D38AF7429D04ADULL, "gen00fl_anders6119192.cl"},
	{0x12D3CA3BF3C11879ULL, "w_bonepitdestroyed_a3_182_0.msh"},
	{0x12D3D43BF3C12967ULL, "w_bonepitdestroyed_a3_182_0.mmh"},
	{0x12D3D83BF3C12FB4ULL, "w_bonepitdestroyed_a3_182_0.mao"},
	{0x12D3F0E9FC962CCCULL, "vxht_0_hightown_a3_c83_0_0.vlm"},
	{0x12D4640F47D99E08ULL, "w_bonepit_235_0.msh"},
	{0x12D4760F47D9BC99ULL, "w_bonepit_235_0.mao"},
	{0x12D4774D4343C347ULL, "6075843_m.fxe"},
	{0x12D4820F47D9D132ULL, "w_bonepit_235_0.mmh"},
	{0x12D4A506CCC9DE2DULL, "6139460_m.fxe"},
	{0x12D4EF95329C7492ULL, "vxht_0_hightown_a3_bi7_0_0.vlm"},
	{0x12D54B5B0FBF9FAFULL, "6000784_m.fxe"},
	{0x12D5A3D0D2A89342ULL, "vxlt_undercityencounter_1ra_0_0.vlm"},
	{0x12D5BAE38963E7E9ULL, "one220_yevhen6102910.cl"},
	{0x12D5D43994DF5DA2ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fin_c.anb"},
	{0x12D5E5C61C8DBC1BULL, "6080227_m.fxe"},
	{0x12D5EF2B17D389BCULL, "mcr310_argument6136206.cl"},
	{0x12D631E0081B4C87ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_facem1.mao"},
	{0x12D639D81255594EULL, "744659038.ogg"},
	{0x12D6C64897F2DF67ULL, "6086940_m.fxe"},
	{0x12D6E1E7E16CDACDULL, "zz_followers_banter6157779.cl"},
	{0x12D6E431D89992D0ULL, "vxht_0_hightown_c99_0_0.vlm"},
	{0x12D78DD7AAE57F7AULL, "vxw_bonepitmining_a2_1wb_0_0.vlm"},
	{0x12D868DB2E7430ECULL, "vxca_cavern1_haunted_k2u_0_0.vlm"},
	{0x12D88D61440C8CC8ULL, "vxlt_backalley_g_1wh_0_0.vlm"},
	{0x12D8BA5A2EBDEFB1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_noba_0s.dds"},
	{0x12D8C7B7ED4B0DADULL, "vxlt_undercitydungeon_19n_0_0.vlm"},
	{0x12D8F15B932C205BULL, "6121935_m.fxe"},
	{0x12D8F2E9808AA16EULL, "620644956.ogg"},
	{0x12D959EB4750B1A6ULL, "6036347-carvertakecareb.cut"},
	{0x12DA022FDEE5C1ADULL, "vxca_cavern1_haunted_hkl_0_0.vlm"},
	{0x12DADFA95DCD3A8AULL, "hlac_lt_foundry_n_320_14z.dds"},
	{0x12DAEE7DA022D6BBULL, "ser100_party_comments6120595.cl"},
	{0x12DB247DE4C1D0EFULL, "vxw_bonepitdestroyed_a3_21p_0_0.vlm"},
	{0x12DB547FBD4B220BULL, "6172827_m.fxe"},
	{0x12DB94A671C395B5ULL, "6143458_m.fxe"},
	{0x12DBCB47A269B3D7ULL, "6162602_m.fxe"},
	{0x12DC545956CFC1DDULL, "6162262_m.fxe"},
	{0x12DCA4FF04DA84F4ULL, "6030667_m.fxe"},
	{0x12DCB70AF9A3BB3AULL, "gen00fl_dog6167694.cl"},
	{0x12DCD66AFEBE62ACULL, "6149160_f.fxe"},
	{0x12DCE77E1538E3A7ULL, "vxw_encounter2_5t_0_0.vlm"},
	{0x12DD2F52A65E4D94ULL, "gen00fl_isabela6088105.cl"},
	{0x12DD312109935C19ULL, "hlac_lt_0_lowtown_65537_0x.dds"},
	{0x12DDC8A199FDCC35ULL, "debug_golem.ncs"},
	{0x12DEB8EF9CD8AF2AULL, "6154048_m.fxe"},
	{0x12DF0648B1BBDA1FULL, "vxw_bonepit_1rj_0_0.vlm"},
	{0x12DF29AAD5D9FE77ULL, "276072029.ogg"},
	{0x12DFAD9FCBA8FCF4ULL, "6148083_m.fxe"},
	{0x12DFD304399B2840ULL, "6118222_m.fxe"},
	{0x12E16BD764BECB11ULL, "art\\levels\\races\\design\\prp_bodybagc.mmh"},
	{0x12E177D764BEDFB2ULL, "art\\levels\\races\\design\\prp_bodybagc.mao"},
	{0x12E185D764BEF7CFULL, "art\\levels\\races\\design\\prp_bodybagc.msh"},
	{0x12E1C2B0ACDB7B76ULL, "w_encounter2_32_0.mao"},
	{0x12E1CEB0ACDB8F15ULL, "w_encounter2_32_0.mmh"},
	{0x12E1D0B0ACDB928BULL, "w_encounter2_32_0.msh"},
	{0x12E1FA56BE63F8E3ULL, "vxlt_hangedman_n_212_0_0.vlm"},
	{0x12E27656512A5A68ULL, "6014874_m.fxe"},
	{0x12E2A0ED8EE2F07BULL, "nat380_party_banter6090055.cl"},
	{0x12E2FA940C502689ULL, "mag370_meredith6046876.cl"},
	{0x12E33B5082864572ULL, "qun131_taskmaster6039717.cl"},
	{0x12E386243F3C4C41ULL, "ht_0_hightownexplosion_n_34_0_ll.mmh"},
	{0x12E3DDA7946326D4ULL, "372474317.ogg"},
	{0x12E46FA4A60FE815ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_5y.dds"},
	{0x12E5007E42C5D64AULL, "662411208.ogg"},
	{0x12E52715A89EDE9FULL, "static_static_p8.psh.fxo"},
	{0x12E575C27CBAB497ULL, "236188467.ogg"},
	{0x12E5D1DEC472DAC8ULL, "vxw_bonepitmining_a2_1m2_0_0.vlm"},
	{0x12E5E6AB5CD15136ULL, "art\\vfx\\combat\\vfx_m_bld_wound_aoe_imp_p.impa.ani"},
	{0x12E64E0199D20A04ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.mmh"},
	{0x12E698B0C62FC2ACULL, "lt_0_lowtown_docks_a3_n_134_0.msh"},
	{0x12E69EB0C62FCC9EULL, "lt_0_lowtown_docks_a3_n_134_0.mmh"},
	{0x12E6A1474311640DULL, "vxht_mansion_n_4xs_0_0.vlm"},
	{0x12E6AAB0C62FE135ULL, "lt_0_lowtown_docks_a3_n_134_0.mao"},
	{0x12E6C5D8C884302FULL, "6173420_m.fxe"},
	{0x12E6DD33D8502BFEULL, "zz_mer_debug6062706.cl"},
	{0x12E7A84FD01BF9EBULL, "939571445.ogg"},
	{0x12E7B4082E21944AULL, "6163723_m.fxe"},
	{0x12E8554F34DAEC0FULL, "428368297.ogg"},
	{0x12E92E48205F2D72ULL, "vxlt_gallowstemplar_f_43b_0_0.vlm"},
	{0x12E945E241E02108ULL, "499775407.ogg"},
	{0x12EA2C33083C0B9AULL, "w_blightlands_92_0.msh"},
	{0x12EA32A46BEE222DULL, "267954871.ogg"},
	{0x12EA3A33083C23D3ULL, "w_blightlands_92_0.mao"},
	{0x12EA3E33083C2AE0ULL, "w_blightlands_92_0.mmh"},
	{0x12EA4184B9F9C60DULL, "vxlt_gallowsprison_n_6mq_0_0.vlm"},
	{0x12EABB05F5144761ULL, "vxw_bonepitmining_a2_1j6_0_0.vlm"},
	{0x12EACAD2A8CCCF85ULL, "vxdr_primevaltomb_34s_0_1.vlm"},
	{0x12EBE86566A411E1ULL, "vxlt_backalley_g_1yc_0_0.vlm"},
	{0x12EBEAD8A8362EA4ULL, "gen00fl_merrill6121820.cl"},
	{0x12EC0B5C0B33D365ULL, "175497992.ogg"},
	{0x12EC41DE96297320ULL, "gen00fl_aveline6122143.cl"},
	{0x12ED32BDC90C4684ULL, "dae210_gate_guard.stg"},
	{0x12EDDC2A96B478CBULL, "6149475_f.fxe"},
	{0x12EE1F9A982AF028ULL, "mag101tr_aom_firstfight2.utt"},
	{0x12EE596DC64E9054ULL, "qcr000_arishok6084275.cl"},
	{0x12EFD0CF8212B492ULL, "316014140.ogg"},
	{0x12EFD5494B8483EFULL, "llac_lt_0_lowtown_docks_n_-65539_3.dds"},
	{0x12F09692BAB7869CULL, "en_fl_half.mao"},
	{0x12F0D4A3E0825DCCULL, "vxht_0_hightown_a3_c61_0_0.vlm"},
	{0x12F14B836F2CED05ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_right.phy"},
	{0x12F18E44598C2421ULL, "vis100ip_bounty_poster.ncs"},
	{0x12F1C407C42C323EULL, "prx_lowerblank.dds"},
	{0x12F1D34F1688A592ULL, "vxht_0_hightown_a3_bc1_0_0.vlm"},
	{0x12F2661D5B7E388DULL, "6042036_m.fxe"},
	{0x12F2B3D8B40D71A5ULL, "403079819.ogg"},
	{0x12F2C127C96DD2D9ULL, "w_bonepit_172_0.phy"},
	{0x12F2CAB4099165A0ULL, "vxlt_darktown_4hn_0_0.vlm"},
	{0x12F3776E0D392F9BULL, "6097862_f.fxe"},
	{0x12F392E7CBD03FCFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgc.phy"},
	{0x12F3EEAB5CDD9C27ULL, "art\\vfx\\combat\\vfx_m_bld_wound_aoe_imp_p.impa.evt"},
	{0x12F4139EBD63F645ULL, "vxw_bonepitmining_a2_1zj_0_0.vlm"},
	{0x12F5711B27F8BDC8ULL, "vxlt_backalley_g_1yf_0_0.vlm"},
	{0x12F5DFDC751CF02EULL, "gen00fl_anders6125745.cl"},
	{0x12F5F01834B3D05EULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_c.impa.ani"},
	{0x12F6204542760270ULL, "6148739_m.fxe"},
	{0x12F7AA04A360595FULL, "vxlt_warehouse_n_rc_0_0.vlm"},
	{0x12F7AFE312A777EEULL, "6156370_m.fxe"},
	{0x12F8184A81CEEE4DULL, "mag291_vincento6056853.cl"},
	{0x12F845436BF6C33DULL, "342346902.ogg"},
	{0x12F8E091AE434D64ULL, "qun120cr_javaris6042546.cl"},
	{0x12F990B0F496B20EULL, "vxht_mansion2b_1iq_0_0.vlm"},
	{0x12FA0FAB5CE24070ULL, "art\\vfx\\combat\\vfx_m_bld_wound_aoe_imp_p.impa.gad"},
	{0x12FA3044BDAEE363ULL, "hlac_lt_undercitydungeon_-1_24x.dds"},
	{0x12FA4B9DD84DC26DULL, "6042703_f.fxe"},
	{0x12FA6F581974E483ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb7.msh"},
	{0x12FA7D581974FCDDULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb7.mmh"},
	{0x12FAA0DF442B374AULL, "692621569.ogg"},
	{0x12FAA336C8EE5F88ULL, "6162762_m.fxe"},
	{0x12FACB819DECDDFEULL, "vxca_cavern1_haunted_ee0_0_0.vlm"},
	{0x12FB3079C5887414ULL, "972547347.ogg"},
	{0x12FB405B15FE9B09ULL, "6130135_m.fxe"},
	{0x12FB7D961CFA5BC3ULL, "zz_arishock_blightcu.cut"},
	{0x12FB9279DCD8A2DEULL, "vxca_caverns2prc_908_0_0.vlm"},
	{0x12FBBEA3C374FEAAULL, "fx_ring01.mao"},
	{0x12FBC70FC3691691ULL, "6077955_m.fxe"},
	{0x12FBDB19BA994026ULL, "gen00fl_sebastian6180510.cl"},
	{0x12FC0B4DD0EA9FF1ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.impa.ani"},
	{0x12FC9D885BF3168EULL, "gen00fl_varric6103435.cl"},
	{0x12FCB029A53005DCULL, "6174699_m.fxe"},
	{0x12FCB6BE1D875940ULL, "6112022_m.fxe"},
	{0x12FCFABE2202F3F3ULL, "959124834.ogg"},
	{0x12FE1CA38611437DULL, "1015866860.ogg"},
	{0x12FEC7C0939B0D68ULL, "vxlt_0_lowtown_a2_n_gpe_0_0.vlm"},
	{0x12FF476CC74A0114ULL, "6109721_m.fxe"},
	{0x12FFBD3BF3E63C9CULL, "w_bonepitdestroyed_a3_182_0.phy"},
	{0x12FFF54EE874507FULL, "134213562.ogg"},
	{0x1300F68A40C28173ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_2.ani"},
	{0x13010E242CBEBDE0ULL, "vxlt_0_lowtown_a2_hfo_0_0.vlm"},
	{0x1301AD02B85AD284ULL, "766736869.ogg"},
	{0x130209DB8071DBC3ULL, "t3_dog_d10.tnt"},
	{0x1302865CF7F8773EULL, "95537133.ogg"},
	{0x1302C15EC8580D9FULL, "1058125968.ogg"},
	{0x1302CC01A89974A2ULL, "20959133.ogg"},
	{0x1303081834BE832FULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_c.impa.evt"},
	{0x1303164DD0F0D093ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.impa.gad"},
	{0x1303CDB356D36EDEULL, "6103408_m.fxe"},
	{0x1303DE8A7D89ECBEULL, "vxlt_undercityserialkiller_xd_0_0.vlm"},
	{0x13045607F7CEF44EULL, "dae011_conspiracy_mob6141278.cl"},
	{0x1304F806D1E04224ULL, "rog_pinpointstrike.dds"},
	{0x1305B9C6AB9F7A19ULL, "6020450_m.fxe"},
	{0x1305C9DB4AC58DEDULL, "402074336.ogg"},
	{0x1306099B00937BF4ULL, "233737951.ogg"},
	{0x13061BE1955671DFULL, "vxlt_warehouse_n_yb_0_0.vlm"},
	{0x1306D6A851550F6FULL, "gen00fl_carver6060548.cl"},
	{0x1307526780E3D788ULL, "cnvprev_human_female.utc"},
	{0x13077BE5319B6798ULL, "69411641.ogg"},
	{0x130788D13DF4F5EFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottlewine_n.dds"},
	{0x1309779C077E221BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.resetup.ani"},
	{0x1309FEC42C0EC160ULL, "ico_health_1.dds"},
	{0x130A091834C4A3D8ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_c.impa.gad"},
	{0x130A134DD0F6EAF8ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.impa.evt"},
	{0x130A1D164EFDEA93ULL, "6081743_m.fxe"},
	{0x130A6CB29FC0C5B1ULL, "189638923.ogg"},
	{0x130A76D4E12FA9A0ULL, "cod_mgc_chant_maker.utp"},
	{0x130A98D4D96A40C3ULL, "zz_follower_test.stg"},
	{0x130AAE7D2788289BULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbuckfruit.phy"},
	{0x130B81164578F832ULL, "6136462_m.fxe"},
	{0x130BC3830BC0AD2DULL, "6126734_m.fxe"},
	{0x130C37AE1E87F7D1ULL, "vxw_encounter3_8f_0_0.vlm"},
	{0x130D632167DD65B4ULL, "llac_dr_0_deeproads_5839_3.dds"},
	{0x130DEE8A40CCFDAAULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_2.evt"},
	{0x130E510199F361FBULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.anb"},
	{0x130EE64EF99F4071ULL, "vxlt_warehouse_n_8o_0_0.vlm"},
	{0x130EF0350883C12EULL, "zz_lgt_2_debug6211861.cl"},
	{0x130F018AC8FDFB6FULL, "vfx_distortionmask_p1.vsh.fxo.dx11.assembly"},
	{0x130F80011EB4D660ULL, "vxlt_0_lowtown_a2_hcl_0_0.vlm"},
	{0x1310556F8401F400ULL, "653518387.ogg"},
	{0x131072C19CCBE3ECULL, "ser221_gascard6165986.cl"},
	{0x1310FEC25716CD0DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magb_0s.dds"},
	{0x131162C79BB6E6CCULL, "640675673.ogg"},
	{0x13119401148743AEULL, "dr_primevaltomb_662.rml"},
	{0x1311C5BC14F6C88FULL, "gen00fl_isabela6101470.cl"},
	{0x131200EDC70EB011ULL, "gen00fl_aveline6132783.cl"},
	{0x1312E83D1C0BBAB6ULL, "425404337.ogg"},
	{0x1312FBF5AD420F98ULL, "zz_qcr_debug6045068.cl"},
	{0x1313D6BA397A0A4BULL, "gen00fl_merrill6116056.cl"},
	{0x131402416B7D9CAEULL, "vxlt_foundry_n_10i_0_0.vlm"},
	{0x13142BA39D959299ULL, "6027550_m.fxe"},
	{0x13147AB4203942B8ULL, "vxca_cavern1_haunted_egl_0_0.vlm"},
	{0x1314CD8A40D2E41DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_2.gad"},
	{0x1316746AD882E30EULL, "vxht_mansion2b_1gs_0_0.vlm"},
	{0x13167F9C0788BA12ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.resetup.evt"},
	{0x13174A7B79F3F648ULL, "243552557.ogg"},
	{0x1317562BF46FBB13ULL, "dae000_merchant_generic_hm6172644.cl"},
	{0x1317F7329973076DULL, "lt_0_lowtown_docks_n_121_0.msh"},
	{0x1318053299731EA0ULL, "lt_0_lowtown_docks_n_121_0.mao"},
	{0x13180932997325F3ULL, "lt_0_lowtown_docks_n_121_0.mmh"},
	{0x13181C89DB519534ULL, "vxw_sundermount_2zd_0_1.vlm"},
	{0x13182FC58CAC2ED8ULL, "llac_dr_dungeon_3214_4.dds"},
	{0x131984127DEC1102ULL, "cs_pro000_db_ogre_mage.wbd"},
	{0x1319924B5125FA2EULL, "534075107.ogg"},
	{0x131A263DA0A4BA47ULL, "one120_athenril6088323.cl"},
	{0x131A91355A1FAEEDULL, "614814324.ogg"},
	{0x131AA98B107E1051ULL, "vxw_encounter3_5e_0_0.vlm"},
	{0x131AB18223B2BB24ULL, "vxlt_0_lowtown_df3_0_0.vlm"},
	{0x131AD7A3C38FD9BCULL, "fx_ring01.dds"},
	{0x131AF2C1BA570C49ULL, "gen00fl_fenris6107225.cl"},
	{0x131B38DBFFB065E7ULL, "ts_beasts_drag_prog_eye.gda"},
	{0x131B58AD60FD3D28ULL, "6155889_m.fxe"},
	{0x131BD8B1D112333EULL, "art\\vfx\\combat\\vfx_c_pride_cast_f2_c.impa.evt"},
	{0x131CBC6F19807F6BULL, "pro000_friedrich6050176.cl"},
	{0x131CEE6F60966377ULL, "6109371_m.fxe"},
	{0x131D5C5823DBB8D7ULL, "932125173.ogg"},
	{0x131D8AFE4A4D2246ULL, "6006251_m.fxe"},
	{0x131D9E9C078F0DA5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.resetup.gad"},
	{0x131DB726FF9E1B56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_exit.ani"},
	{0x131DE94F836E686AULL, "24007542.ogg"},
	{0x131DF1DE10AAEEE0ULL, "vxlt_0_lowtown_a2_hhm_0_0.vlm"},
	{0x131F7D4BE680434BULL, "6076811_m.fxe"},
	{0x131F91C1F85D7275ULL, "630158989.ogg"},
	{0x131F9FDF29BFB2E5ULL, "vxht_0_hightown_n_5jd_0_0.vlm"},
	{0x13202ABDC937928FULL, "dae210_gate_guard.bnk"},
	{0x13205BF066853268ULL, "6127134_m.fxe"},
	{0x132088B93596C63DULL, "238126236.ogg"},
	{0x1320C24DD66858F4ULL, "art\\vfx\\creatures\\fxc_spider_dppr_c.mmh"},
	{0x1320D1BFA396204AULL, "6070342_m.fxe"},
	{0x1321119344E2A4D6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_basic.phy"},
	{0x1321121EAF956BB4ULL, "llac_lt_0_lowtown_n_65533_2.dds"},
	{0x13215B8C12590A5FULL, "vfx_blt_lava_volume.bnk"},
	{0x1322292DF104AFD2ULL, "dae000ip_crowd_ff4_talk1.utp"},
	{0x132247EB48659F41ULL, "6034019_m.fxe"},
	{0x13227106F3001132ULL, "vxw_bonepit_149_0_0.vlm"},
	{0x1322E833BF8AE881ULL, "6051713_m.fxe"},
	{0x1322EFB1D1187969ULL, "art\\vfx\\combat\\vfx_c_pride_cast_f2_c.impa.gad"},
	{0x13232FCA28729062ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\runfront.gad"},
	{0x1323BCD327A503A8ULL, "869865176.ogg"},
	{0x13244CBDC93BB5F7ULL, "dae210_gate_guard.cnv"},
	{0x13247442F7443653ULL, "mag211_marethari6025359.cl"},
	{0x132504E16BBE69B7ULL, "853168424.ogg"},
	{0x132578A5B0A38F57ULL, "hlac_w_blightlands_327682_0x.dds"},
	{0x1325BCECD5A2C986ULL, "vxlt_undercityencounter_13c_0_1.vlm"},
	{0x1325D7DA69260E30ULL, "lt_0_lowtown_docks_178_0.msh"},
	{0x1325E5DA6926264AULL, "lt_0_lowtown_docks_178_0.mmh"},
	{0x1325E9DA69262C81ULL, "lt_0_lowtown_docks_178_0.mao"},
	{0x1326262FDED2112EULL, "918848273.ogg"},
	{0x13264AAD45B1C383ULL, "lt_0_lowtown_docks_a3_9_0.rml"},
	{0x132655FF3CB57DE4ULL, "462381158.ogg"},
	{0x1326A7C17789B65CULL, "lt_0_lowtown_docks_210_0.mao"},
	{0x1326B3C17789CAFFULL, "lt_0_lowtown_docks_210_0.mmh"},
	{0x1326B50A4DD58DD2ULL, "471246693.ogg"},
	{0x1326B84904D7660FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.mmh"},
	{0x1326B9C17789D4A1ULL, "lt_0_lowtown_docks_210_0.msh"},
	{0x1326DE4946134658ULL, "llac_lt_0_lowtown_a3_n_c_-131072_4.dds"},
	{0x132775B0C6671E39ULL, "lt_0_lowtown_docks_a3_n_134_0.phy"},
	{0x1327B149D5BD6113ULL, "6153826_m.fxe"},
	{0x1327D8329256A01DULL, "plt_qcr000pt_main.plo"},
	{0x13282D3E88527D0AULL, "gen00fl_varric6103556.cl"},
	{0x132845B6D5FBF711ULL, "40207303.ogg"},
	{0x13285BE11684A50BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentflameblast.evt"},
	{0x1328E2A1289CF8ADULL, "vfx_distortionmask_p1.psh.fxo.dx11"},
	{0x132952CA28773759ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\runfront.evt"},
	{0x1329C0B1D11E47B7ULL, "art\\vfx\\combat\\vfx_c_pride_cast_f2_c.impa.ani"},
	{0x1329DF35F48C544EULL, "gen_im_trs_prf_sta_phy_r1.uti"},
	{0x132A08E3F1FEE056ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd03.msh"},
	{0x132A12E3F1FEF154ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd03.mmh"},
	{0x132A8F86774D0F64ULL, "llac_lt_backalley_n_65536_4.dds"},
	{0x132ABF26FFA8B347ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_exit.evt"},
	{0x132AC85D1A88CCDDULL, "one100dg_kelder6055630.cl"},
	{0x132B014ECF92A9E6ULL, "lowtown_guard_chase6169519.cl"},
	{0x132BFA9344EC8577ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_basic.msh"},
	{0x132C0C9344ECA3DAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_basic.mao"},
	{0x132C109344ECAAE9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_basic.mmh"},
	{0x132C2A2C53333420ULL, "gen_im_wep_rog_arc_act1_04.uti"},
	{0x132C43F1A5C89702ULL, "vxlt_gallowscourtyard_n_7ft_0_0.vlm"},
	{0x132C48C9E03B3913ULL, "zz_dae_debug6029117.cl"},
	{0x132D2A5F74146E85ULL, "hlac_w_encounter1_0_0z.dds"},
	{0x132D3BD01EDCF582ULL, "6092126_m.fxe"},
	{0x132D8C214DA6F1ACULL, "72458574.ogg"},
	{0x132DFA644B3ABD54ULL, "crt_aoe_02.mao"},
	{0x132E04C9C07A9F87ULL, "6083727_m.fxe"},
	{0x132E42383C0A0704ULL, "hlac_ht_bartrand_n_3371_0y.dds"},
	{0x132EA4F02D6712FDULL, "hlac_lt_0_lowtown_-3_2z.dds"},
	{0x132EDB32925CC44AULL, "plt_qcr000pt_main.nss"},
	{0x132F013807CC4241ULL, "tools_icon.mao"},
	{0x132F0B839A41047AULL, "vfx_lavacom.dds"},
	{0x132F74E1168AEECCULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentflameblast.gad"},
	{0x13305F01CDFB0B1AULL, "hlac_lt_0_lowtown_n_-65536_2z.dds"},
	{0x13307C2540ABC4D0ULL, "llac_ht_0_hightown_n_131070_0.dds"},
	{0x1330ABA349BE7F12ULL, "vxca_cavern1_haunted_iob_0_0.vlm"},
	{0x1330B08539A761D7ULL, "vxlt_0_lowtown_a2_n_env_0_0.vlm"},
	{0x1330E37E33D7180AULL, "6004107_m.fxe"},
	{0x1331201700250DE4ULL, "644550017.ogg"},
	{0x133178A70F7299E5ULL, "da2terrain_rendertodepth_p2.psh.fxo.dx11"},
	{0x13319204D0A91C85ULL, "wm_ico_sundermount.dds"},
	{0x1331E026FFAF0990ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_exit.gad"},
	{0x1331F056F1774A93ULL, "6089956_m.fxe"},
	{0x133212C0651192EAULL, "vxlt_0_lowtown_a2_n_c_byo_0_0.vlm"},
	{0x133238AE42E7344BULL, "524263746.ogg"},
	{0x133280F96E5D2EB4ULL, "lt_gallowscourtyard_n_mmap.dds"},
	{0x1332A5AD69C52AB5ULL, "hlac_lt_0_lowtown_n_-131071_1z.dds"},
	{0x133381F83B622FD7ULL, "gen00fl_fenris6126778.cl"},
	{0x133388433BD2D5C6ULL, "6162113_m.fxe"},
	{0x1333B18CF915737DULL, "hlac_ca_caverns1_mines_24424_21x.dds"},
	{0x1334044836F3BF19ULL, "vxdr_0_deeproads_ego_0_0.vlm"},
	{0x13342CBE52DDC82DULL, "vxlt_0_lowtown_a3_n_c_hrr_0_0.vlm"},
	{0x13343D54F2D16A76ULL, "6156411_m.fxe"},
	{0x1335A9D2C283B9A3ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushringglw.msh"},
	{0x1335B7D2C283D1FDULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushringglw.mmh"},
	{0x1335BBD2C283D80EULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushringglw.mao"},
	{0x1335CC387473B0CBULL, "1035267920.ogg"},
	{0x1335EE4C68A3DDD1ULL, "vxlt_0_lowtown_a2_erg_0_0.vlm"},
	{0x13360BCCB94529DBULL, "6024539_m.fxe"},
	{0x133673E116910BCAULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentflameblast.ani"},
	{0x1336927EAD076016ULL, "vxw_woundedcoast_va_0_0.vlm"},
	{0x133693DF052D64C5ULL, "vxlt_0_lowtown_a2_n_c_f54_0_0.vlm"},
	{0x133728C0C615A426ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed_v2.evt"},
	{0x13375ACA28838248ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\runfront.ani"},
	{0x1337957D27AEEC3AULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbuckfruit.msh"},
	{0x13379A69ED11648DULL, "659439009.ogg"},
	{0x1337A37D27AF0473ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbuckfruit.mao"},
	{0x1337A77D27AF0A80ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbuckfruit.mmh"},
	{0x1337C79350A3B4CFULL, "500575575.ogg"},
	{0x1338974459C83630ULL, "vis100ip_bounty_poster.utp"},
	{0x1338EE010E428A82ULL, "vxht_mansion2_3ga_0_0.vlm"},
	{0x1338F306C2227F40ULL, "165731847.ogg"},
	{0x1339166C61EE63FBULL, "903634620.ogg"},
	{0x1339AFAA81A96596ULL, "vxw_bonepit_1zw_0_0.vlm"},
	{0x1339DBD9E9F9199DULL, "445006653.ogg"},
	{0x133A3D2BEEC451E4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sackclose.msh"},
	{0x133A4F2BEEC4704DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sackclose.mao"},
	{0x133A532BEEC47686ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sackclose.mmh"},
	{0x133AB5CE97BEAF82ULL, "vxlt_gallowscourtyard_n_7iu_0_0.vlm"},
	{0x133ABA8505A9214AULL, "hlac_lt_0_lowtown_a3_n_c_65533_0y.dds"},
	{0x133ACB065E94C693ULL, "dre000_sandal6028411.cl"},
	{0x133B00F20C4F5D74ULL, "character_character_p227.psh.fxo.dx11.assembly"},
	{0x133B2A03AC6CB2E5ULL, "6123315_m.fxe"},
	{0x133B43CAAEE570C6ULL, "hla_ht_0_hightown_-65536_0z.dds"},
	{0x133B70FB5341424EULL, "128341674.ogg"},
	{0x133C76E494EF6E8BULL, "vxlt_0_lowtown_a2_n_by8_0_0.vlm"},
	{0x133C889DA19E2587ULL, "llac_ht_0_hightown_-196608_1.dds"},
	{0x133CCD1A3980A5F9ULL, "hlac_lt_blackemporium_1055_18z.dds"},
	{0x133CCD5734DC7F98ULL, "vxlt_warehouse_n_11b_0_0.vlm"},
	{0x133D3FC0C61A3771ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed_v2.gad"},
	{0x133E2C84A83F2679ULL, "hlac_lt_0_lowtown_docks_n_-196607_0x.dds"},
	{0x133E423B7C43CE30ULL, "1039111589.ogg"},
	{0x133E424D21298D71ULL, "6065013_m.fxe"},
	{0x133E795582E171E1ULL, "vxw_bonepitmining_a2_29p_0_0.vlm"},
	{0x133E80EDA43FD0D6ULL, "431854937.ogg"},
	{0x133E94C5D8B4F4B5ULL, "6020483_m.fxe"},
	{0x133EF7A00030A07DULL, "70595397.ogg"},
	{0x133F7AA8F83518E0ULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_0y.dds"},
	{0x133F8439750BCE49ULL, "1018085152.ogg"},
	{0x13407581CC592A70ULL, "vxlt_gallowstemplar_n_39e_0_0.vlm"},
	{0x134082D4D997AAC4ULL, "zz_follower_test.cnv"},
	{0x1340B777B06C96FCULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.impa.ani"},
	{0x1341E3E1A245DAC0ULL, "744591349.ogg"},
	{0x1341E8495DBE804AULL, "6169772_m.fxe"},
	{0x13422EAC3E247CD6ULL, "art\\cinematics\\props\\prp_barge.msh"},
	{0x134238AC3E248DD4ULL, "art\\cinematics\\props\\prp_barge.mmh"},
	{0x13423CAC3E2494E7ULL, "art\\cinematics\\props\\prp_barge.mao"},
	{0x134295B9416FB0FFULL, "6139923_m.fxe"},
	{0x13429E0100AD7D5BULL, "pro000_wesley6007724.cl"},
	{0x1342AB1077A23557ULL, "6044821_m.fxe"},
	{0x1342B47251879C61ULL, "833901515.ogg"},
	{0x1342CEB8F5324E8BULL, "278048303.ogg"},
	{0x1342F7EFAD2F2392ULL, "mcr370_meredith6085929.cl"},
	{0x1342FD046FFEB024ULL, "hlac_w_encounter3_-4_0z.dds"},
	{0x134440C0C620571FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed_v2.ani"},
	{0x13444A2BEECCBC51ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sackclose.phy"},
	{0x134460295A99F651ULL, "vxlt_0_lowtown_a2_euf_0_0.vlm"},
	{0x1344CDE5EB8FD0A7ULL, "mag131_keran6086191.cl"},
	{0x1344EEC9453A47A0ULL, "nat380_temmerin6120951.cl"},
	{0x1345045B9EFD7896ULL, "vxw_woundedcoast_sb_0_0.vlm"},
	{0x134612079267AF0BULL, "199413000.ogg"},
	{0x13466212A8D64192ULL, "vxca_caverns2prc_9pt_0_0.vlm"},
	{0x1346850C6A6E1303ULL, "llac_ht_0_hightown_n_65536_2.dds"},
	{0x134691C9DFB74F1CULL, "packages_types.gda"},
	{0x13470B4E33E1EF1EULL, "vxht_0_hightown_a3_c2o_0_0.vlm"},
	{0x1347BA6B318E427DULL, "vxht_0_hightown_n_c6f_0_0.vlm"},
	{0x1347C477B072CB3EULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.impa.gad"},
	{0x1347C5D77C440234ULL, "vxht_keep_73s_0_0.vlm"},
	{0x1347E2A27C95A045ULL, "al_fenris_s01.gda"},
	{0x1348020B44361DC8ULL, "vxw_bonepitmining_a2_29u_0_0.vlm"},
	{0x134893B9A5EF75E7ULL, "hlac_lt_gallowstemplar_n_2384_6y.dds"},
	{0x1348C54DD689AFABULL, "art\\vfx\\creatures\\fxc_spider_dppr_c.anb"},
	{0x13491925EA017CF5ULL, "21390875.ogg"},
	{0x13495F135CDC230EULL, "vxlt_0_lowtown_a2_n_gna_0_0.vlm"},
	{0x13496AC3C975C12CULL, "lt_0_lowtown_docks_a3_199_0.msh"},
	{0x134970C3C975CB1EULL, "lt_0_lowtown_docks_a3_199_0.mmh"},
	{0x13497CC3C975DFB5ULL, "lt_0_lowtown_docks_a3_199_0.mao"},
	{0x13497DCFEB680124ULL, "436171930.ogg"},
	{0x1349B53B18B99EEAULL, "152920010.ogg"},
	{0x134A0E9A98793255ULL, "mag101tr_aom_firstfight2.ncs"},
	{0x134B13FB8753FB21ULL, "zz_one_debug6074010.cl"},
	{0x134B3D447762E1DCULL, "6001887_m.fxe"},
	{0x134B518E2B7F2B2AULL, "vxlt_hangedman_175_0_0.vlm"},
	{0x134B7BB29376D169ULL, "140086046.ogg"},
	{0x134BD432A68DCCA6ULL, "vxlt_0_lowtown_docks_b63_0_1.vlm"},
	{0x134BF39C9DABB331ULL, "dae000_quintus6154726.cl"},
	{0x134C33576CAEF455ULL, "259507124.ogg"},
	{0x134C6F403E12BBF3ULL, "vxht_0_hightown_bnv_0_0.vlm"},
	{0x134C7895EA6E8873ULL, "vxca_caverns2_8qr_0_0.vlm"},
	{0x134C88DAE1C3E7DCULL, "845760631.ogg"},
	{0x134CA520B1A6415EULL, "295751993.ogg"},
	{0x134CAF5669273294ULL, "hightown_male_cityguard_mob6169972.cl"},
	{0x134CEE3F00816C76ULL, "6087641_m.fxe"},
	{0x134CFF7810AC2F5CULL, "gen00fl_fenris6126520.cl"},
	{0x134D05F7EE1C41C5ULL, "hf_guard_1.mor"},
	{0x134D06A2CAAB0D35ULL, "6027583_m.fxe"},
	{0x134D418C127E1895ULL, "vfx_blt_lava_volume.wbd"},
	{0x134D5A09415F90ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_loop.ani"},
	{0x134E107A2A19E72FULL, "6195135_f.fxe"},
	{0x134E9F77B078AB5DULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.impa.evt"},
	{0x134EB94904F8BA4CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.anb"},
	{0x134EDD4426DFCFE1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobd_0d.dds"},
	{0x134F0073F735EAD5ULL, "vxht_mansion2_2pz_0_0.vlm"},
	{0x134FFEF6756974FEULL, "isa211_isabela6202203.cl"},
	{0x1350CD01DF6491CDULL, "vxca_caverns2_7qk_0_0.vlm"},
	{0x1351107836C9F92DULL, "vxlt_0_lowtown_a3_n_c_hdp_0_0.vlm"},
	{0x13515F379C3596EBULL, "hlac_lt_0_lowtown_a2_131074_0z.dds"},
	{0x13522D989FE23544ULL, "zz_fen_debug6215345.cl"},
	{0x1352D2064C900ED1ULL, "vxlt_0_lowtown_a2_exe_0_0.vlm"},
	{0x1352F83A4962DF70ULL, "258855182.ogg"},
	{0x135367094164122EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_loop.gad"},
	{0x13540C0874E2E2BFULL, "6040869_m.fxe"},
	{0x135426FD8540F554ULL, "6214707_f.fxe"},
	{0x135443146EDC0B8BULL, "vxht_chantry_2zj_0_0.vlm"},
	{0x13549B60734B5F6EULL, "6161422_m.fxe"},
	{0x1354C5ADE7023E9FULL, "6147944_m.fxe"},
	{0x1354F73D95990ACDULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyo_0d.dds"},
	{0x13550A47AE6B9999ULL, "dre000_sandal6028358.cl"},
	{0x13562ADA42B49620ULL, "hlac_w_encounter3_-65537_0x.dds"},
	{0x13562C4823845AFDULL, "vxht_0_hightown_n_c3e_0_0.vlm"},
	{0x13563A1A1D30DCCEULL, "narrative_varric_demo6170822.cl"},
	{0x13564251E268025BULL, "llac_dr_primevaltomb_659_16.dds"},
	{0x13564A15B1BA2E46ULL, "vxw_encounter2_camp_l6_0_0.vlm"},
	{0x135695A9C59CCBE1ULL, "742181611.ogg"},
	{0x13572E24190FD786ULL, "6153441_m.fxe"},
	{0x1357960062D5E64EULL, "192817745.ogg"},
	{0x1357BE1D63E2B7CBULL, "zz_seb_debug6189581.cl"},
	{0x1357E318F15FDE79ULL, "vfx_vfx_p0.vsh.fxo.dx11"},
	{0x1358011AD482EED6ULL, "vxlt_darktown_4vv_0_0.vlm"},
	{0x135801D9BEC95D50ULL, "mer240_jansen6081443.cl"},
	{0x13580ADB0975FA34ULL, "6148476_m.fxe"},
	{0x13581D1EC4E72E24ULL, "vxw_bonepit_1q4_0_0.vlm"},
	{0x13583FE965F0C0FAULL, "6170437_m.fxe"},
	{0x135842586DBF098BULL, "6091852_m.fxe"},
	{0x13586B80F9D6F051ULL, "lowtown_fem_prostitute_sta6161614.cl"},
	{0x1359B6B8B5B22ECEULL, "hla_ht_0_hightown_131070_0x.dds"},
	{0x135A1E470845F571ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_04tent.msh"},
	{0x135A2C4708460C8CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_04tent.mao"},
	{0x135A2F1332EA11F5ULL, "vxht_0_hightown_ca0_0_0.vlm"},
	{0x135A38470846212FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_04tent.mmh"},
	{0x135A60511914CD77ULL, "llac_w_encounter1_-131068_0.dds"},
	{0x135A974E6CA47EB7ULL, "w_bonepit_90_0.msh"},
	{0x135A9F1C2533771BULL, "6067818_f.fxe"},
	{0x135AA94E6CA49D1AULL, "w_bonepit_90_0.mao"},
	{0x135AAD4E6CA4A429ULL, "w_bonepit_90_0.mmh"},
	{0x135ABD2F82FA0FADULL, "hlac_lt_gallowscourtyard_131071_1y.dds"},
	{0x135AE11D3008D473ULL, "vxht_0_hightown_bky_0_0.vlm"},
	{0x135B5A0AB027A0F5ULL, "hlac_lt_0_lowtown_n_131072_5x.dds"},
	{0x135B6E218CA59836ULL, "6170727_m.fxe"},
	{0x135BE0A49EF07ADAULL, "ave110_ambush_2.cnv"},
	{0x135C1C88ED40CFAEULL, "dae000_keep_bureaucrat_1_st6165348.cl"},
	{0x135C20DAFC0875FFULL, "6165711_m.fxe"},
	{0x135C6A0E5C83FA5CULL, "framebuffereffects_linearizedepth_p0.psh.fxo"},
	{0x135CB74BCF93BA5CULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.dura.ani"},
	{0x135D7678FBAD0A5BULL, "6020074_m.fxe"},
	{0x135D9928D9E3A257ULL, "6034480_m.fxe"},
	{0x135DDE01ED81E015ULL, "hlac_dr_0_deeproads_19793_0x.dds"},
	{0x135DF6B7830B0541ULL, "6025332_m.fxe"},
	{0x135E32215565AA70ULL, "6080141_m.fxe"},
	{0x135E695E6A06DB42ULL, "vxlt_0_lowtown_docks_n_cv0_0_0.vlm"},
	{0x135E7DFA9D7FEAE2ULL, "dre000_dougal6130956.cl"},
	{0x135F02A49EF2EA22ULL, "ave110_ambush_2.bnk"},
	{0x135F2E5E41DFAB12ULL, "158633264.ogg"},
	{0x135F3820D7C900DDULL, "art\\levels\\ztools\\helpers\\character\\_textures\\prp_zzdwarfface_d.dds"},
	{0x135F4DC0625DB3ABULL, "gen00fl_anders6128330.cl"},
	{0x135F8A9C1E506942ULL, "6147170_m.fxe"},
	{0x135F9FBAED4AE408ULL, "vxlt_undercityencounter_1j3_0_0.vlm"},
	{0x135FDB17FDA0161BULL, "vrc221_varric6044182.cl"},
	{0x135FE03DF1463E21ULL, "1044147723.ogg"},
	{0x13600B6956FBE68FULL, "6067779_f.fxe"},
	{0x136024EB04B92425ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bigwheel_dp.dds"},
	{0x13604EAFB347D22EULL, "vxlt_hangedman_2mt_0_0.vlm"},
	{0x1360605A43BF8346ULL, "zz_approval_debug6003551.cl"},
	{0x13607BC35A427BE6ULL, "vxlt_gallowstemplar_n_5y9_0_0.vlm"},
	{0x136083B64EC3588BULL, "6104295_m.fxe"},
	{0x1360D9D26A1E40AEULL, "6110361_m.fxe"},
	{0x13612385A3BF967FULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_whippingpost.phy"},
	{0x13615FF6A3E3851BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_enter.evt"},
	{0x136167FD36AC8745ULL, "root_swashbuckler.dds"},
	{0x13616954BEF164AAULL, "6110286_m.fxe"},
	{0x136192127E29DEB8ULL, "cs_pro000_db_ogre_mage.bnk"},
	{0x13619D1CE62396E9ULL, "290359771.ogg"},
	{0x1361A9ACCFDBDDE6ULL, "6072758_f.fxe"},
	{0x1361C6DB6FCF83F0ULL, "follower_barks6153728.cl"},
	{0x1361C9E29B8D5E4DULL, "402486412.ogg"},
	{0x1361DF7E55F83E0AULL, "gen00fl_aveline6139691.cl"},
	{0x13627E5E7302D0CEULL, "6064062_m.fxe"},
	{0x1362C44BCF983B9EULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.dura.gad"},
	{0x1362C7798087AFA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\holysmite.ani"},
	{0x1362D60775A13054ULL, "6174307_m.fxe"},
	{0x13630D3D654177EEULL, "6166522_m.fxe"},
	{0x1363964205D10D88ULL, "mapvisbrush.mao"},
	{0x136424ECEF2DDBACULL, "combat_deathblow_01.gda"},
	{0x136446C10EB72CDCULL, "vxlt_0_lowtown_a2_dej_0_0.vlm"},
	{0x13645A4FB5BD0C08ULL, "w_woundedcoast_94_0.mmh"},
	{0x1364664FB5BD20ABULL, "w_woundedcoast_94_0.mao"},
	{0x1364784FB5BD3F32ULL, "w_woundedcoast_94_0.msh"},
	{0x13649E25157A737DULL, "vxht_0_hightown_n_c8d_0_0.vlm"},
	{0x1364BBF2A3B046C6ULL, "vxw_encounter2_camp_g7_0_0.vlm"},
	{0x13657874CE6BB7A2ULL, "118310066.ogg"},
	{0x13659E7DC778B312ULL, "6020530_m.fxe"},
	{0x1365D9B2089E527BULL, "vxlt_0_lowtown_a3_n_c_hqy_0_0.vlm"},
	{0x1365E348D86502E4ULL, "vxw_sundermount_2j5_0_0.vlm"},
	{0x1366430196FBE21FULL, "6182016_f.fxe"},
	{0x13668EFBB6DD46A4ULL, "vxw_bonepit_1z5_0_0.vlm"},
	{0x13668F226BC01E98ULL, "93139391.ogg"},
	{0x1366D71EC838C88FULL, "qun211_braggart6095455.cl"},
	{0x1366EAC8E6809975ULL, "vxht_0_hightown_a3_aq7_0_0.vlm"},
	{0x13672DAD45E929FFULL, "lt_0_lowtown_docks_a3_9_0.anb"},
	{0x136810CA74DBD0C6ULL, "946219720.ogg"},
	{0x136858F6A3E9985CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_enter.gad"},
	{0x1368F86E35AD83C8ULL, "6041528_m.fxe"},
	{0x13699686C374916BULL, "w_bonepit_mmi.gff"},
	{0x13699F4BCF9E1BBDULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.dura.evt"},
	{0x1369C22937A4E893ULL, "6071159_m.fxe"},
	{0x136A1C9459813466ULL, "vxht_0_hightown_a3_can_0_0.vlm"},
	{0x136A5316D545CCD5ULL, "gen00fl_sebastian6175003.cl"},
	{0x136A7B9C00A616ACULL, "w_bonepit_114_0.msh"},
	{0x136A819C00A6209EULL, "w_bonepit_114_0.mmh"},
	{0x136A8D9C00A63535ULL, "w_bonepit_114_0.mao"},
	{0x136ADC95C7A2E9E9ULL, "zz_ser_debug6085539.cl"},
	{0x136AF40D87B7662AULL, "vxht_mansion_n_501_0_0.vlm"},
	{0x136C5DE8568C35E9ULL, "6103627_m.fxe"},
	{0x136C960101E6E0C1ULL, "hightown_market_guard_sta6169632.cl"},
	{0x136CDB3B5BFCF3C2ULL, "vxlt_0_lowtown_docks_n_cq1_0_0.vlm"},
	{0x136D7FAE7D812D74ULL, "vxht_0_hightown_c3c_0_0.vlm"},
	{0x136DF4321AB62A2DULL, "vxlt_0_lowtown_a3_n_c_hnn_0_0.vlm"},
	{0x136E30C3E56CD096ULL, "649671120.ogg"},
	{0x136E57F6A3EE025AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_enter.ani"},
	{0x136FF1C70E8A11A7ULL, "6207475_m.fxe"},
	{0x137062F987A86273ULL, "6124712_m.fxe"},
	{0x1370DF798094159AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\holysmite.evt"},
	{0x1370F9395F95F936ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masc_0s.dds"},
	{0x1371183DB5665FD3ULL, "vxlt_0_lowtown_a2_fi3_0_0.vlm"},
	{0x13712A60AA94B70CULL, "vxw_woundedcoast_1ci_0_0.vlm"},
	{0x13713121E008FF32ULL, "6151487_m.fxe"},
	{0x13714E487C27AD52ULL, "996067935.ogg"},
	{0x1371653FF5C2D588ULL, "l_lgt261ar_tome_cave.wbd"},
	{0x137166C571CE685DULL, "art\\characters\\creatures\\shade\\textures\\c_shade_0i.dds"},
	{0x1371BE8CDE578998ULL, "llac_lt_gallowstemplar_f_2394_12.dds"},
	{0x13722485A3CE0A84ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_whippingpost.mmh"},
	{0x13722885A3CE11D7ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_whippingpost.mao"},
	{0x13723A85A3CE3026ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_whippingpost.msh"},
	{0x13728F647DD4C1BAULL, "lgt220_shp_mag_guard6179475.cl"},
	{0x13729380608D6ABCULL, "vxlt_warehouse_14r_0_0.vlm"},
	{0x137297DDAFDE311DULL, "307344686.ogg"},
	{0x1373D9EDB54CE9CDULL, "6096801_f.fxe"},
	{0x1373E28E0A422E3CULL, "6155940_m.fxe"},
	{0x1373F298A80F5767ULL, "vxlt_gallowsprison_n_6ag_0_0.vlm"},
	{0x1373F62A05908259ULL, "49013833.ogg"},
	{0x1375534CEE6B603FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_cheers.gad"},
	{0x1375711F52F616BDULL, "mcr370_moira6087498.cl"},
	{0x13759290F5F1E455ULL, "gen00fl_dog6141097.cl"},
	{0x13759760EF76C3C1ULL, "hlac_w_sundermount_3_0y.dds"},
	{0x1375BA6AD16CBFD6ULL, "bflipgood.dds"},
	{0x1375BFA276F47379ULL, "998175541.ogg"},
	{0x13762B277E9DB053ULL, "hlac_lt_0_lowtown_docks_a3_-65537_1z.dds"},
	{0x13763D3E7AA5E8C6ULL, "1071196146.ogg"},
	{0x1376521A53869AFEULL, "6174932_m.fxe"},
	{0x13769ED05617A7AFULL, "mag371_elsa6041259.cl"},
	{0x1376EC799D8C6F2DULL, "one110_gustav6089809.cl"},
	{0x13771B0ECFAFC2FBULL, "vxw_encounter2_f5_0_0.vlm"},
	{0x137745676EED44B6ULL, "vxht_0_hightown_a3_c34_0_0.vlm"},
	{0x13775ACADC69F72FULL, "157758961.ogg"},
	{0x13778177898F2130ULL, "vxlt_gallowstemplar_f_6s6_0_0.vlm"},
	{0x1377DE79809A328DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\holysmite.gad"},
	{0x1377E9BF94B4C190ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgtb_0n.dds"},
	{0x13782B9E21E339F6ULL, "lgt251cr_stone_golem.utc"},
	{0x137843B901C49300ULL, "dae000ip_crowd_mf8_merchant.utp"},
	{0x1378465E715BBE8BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_come_0d.dds"},
	{0x1378D99AB3D3ED37ULL, "352320051.ogg"},
	{0x13793DA37BA1F6EEULL, "vxht_chantry_28u_0_0.vlm"},
	{0x13798F1E78993A96ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyo_0t.dds"},
	{0x1379980A75D9BFF3ULL, "dae000_elegant6166896.cl"},
	{0x137AB47F44D3DA32ULL, "vxw_bonepitmining_a2_1o2_0_0.vlm"},
	{0x137AD36E57A3B1F1ULL, "ef_radha.mor"},
	{0x137AF66DDA35931BULL, "vxlt_gallowstemplar_4t1_0_0.vlm"},
	{0x137B05E76DF77553ULL, "256177979.ogg"},
	{0x137B7FDC84E36665ULL, "291325193.ogg"},
	{0x137B9A360F316CFDULL, "mag121_eau_banter6121089.cl"},
	{0x137BB3DB48F1E2A6ULL, "hlac_lt_backalley_n_0_13y.dds"},
	{0x137BB5C905CBF050ULL, "278079623.ogg"},
	{0x137BD8EE9BFF87F6ULL, "lt_0_lowtown_docks_38_0.anb"},
	{0x137C0779228CBF59ULL, "ht_mansion_n_3365.anb"},
	{0x137C484CEE716C2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_cheers.evt"},
	{0x137C660F0CAC42ADULL, "vxlt_0_lowtown_a3_n_c_hkm_0_0.vlm"},
	{0x137C83548EA8B449ULL, "qun200_party_comment6090456.cl"},
	{0x137CB42DB0072B5BULL, "w_sundermount_97_0.mmh"},
	{0x137CC02DB0073FF8ULL, "w_sundermount_97_0.mao"},
	{0x137CD22DB0075E45ULL, "w_sundermount_97_0.msh"},
	{0x137DE9108D0ED4F9ULL, "6068793_m.fxe"},
	{0x137E37585CAF2C43ULL, "llac_lt_0_lowtown_a2_-131072_3.dds"},
	{0x137E6C720684B5EAULL, "284132015.ogg"},
	{0x137EDEE31DDFF959ULL, "art\\vfx\\creatures\\fxc_succubus_c.dura.ani"},
	{0x137F3B5AD10EFC89ULL, "qun221_petrice6077315.cl"},
	{0x137F79598DC8EAB8ULL, "twopassblurshadows.psh.fxo"},
	{0x137F9C3D9C8ACF8CULL, "vxw_woundedcoast_1nn_0_0.vlm"},
	{0x137FB9243D8B22FEULL, "vxca_cavern1_haunted_hma_0_0.vlm"},
	{0x137FBDBE3810D11EULL, "28545367.ogg"},
	{0x137FC576FBA0DCCEULL, "6165361_m.fxe"},
	{0x13802ED58999E56CULL, "gen00fl_aveline6105401.cl"},
	{0x13803EC2A30B674DULL, "6050383_m.fxe"},
	{0x1380BD0E2C55B0BEULL, "lt_0_lowtown_docks_a3_3_0_lc.mmh"},
	{0x1380C59471316DAFULL, "hlac_ca_caverns1_mines_24424_7x.dds"},
	{0x13811E8C4B87BE20ULL, "hlac_lt_0_lowtown_a2_n_65533_2x.dds"},
	{0x1381674205E9BBEAULL, "mapvisbrush.dds"},
	{0x1381749CC3616E9DULL, "6094013_m.fxe"},
	{0x1381DDC6D918A83FULL, "363606306.ogg"},
	{0x1382D63C9421CF89ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxa.phy"},
	{0x1383AD4D98FA40A3ULL, "qun110_petrice6042280.cl"},
	{0x1383FCB1C2A6FBDFULL, "w_encounter3_16_0.rml"},
	{0x13840099F52D1419ULL, "478645368.ogg"},
	{0x138403C5B88C5950ULL, "gen_im_acc_amu_act3_03.uti"},
	{0x13858CEBC1A5DB7BULL, "vxw_encounter2_a4_0_0.vlm"},
	{0x138591ADE8798BE3ULL, "llac_lt_darktown_-1_81.dds"},
	{0x138598E99019FB74ULL, "hlac_ht_keep_n_4519_14x.dds"},
	{0x1385B74460E35D36ULL, "vxht_0_hightown_a3_c45_0_0.vlm"},
	{0x1385B9E31DE5D89BULL, "art\\vfx\\creatures\\fxc_succubus_c.dura.gad"},
	{0x138611E3F24D4E4FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd03.phy"},
	{0x138686FC11408D52ULL, "vxlt_0_lowtown_docks_a3_n_d0g_0_1.vlm"},
	{0x1386FC4D6F0CC877ULL, "6054976_m.fxe"},
	{0x13877EE6CE02F544ULL, "vfx_bird_raven01.dds"},
	{0x13879C9F33B76A16ULL, "484674995.ogg"},
	{0x1387BE42E276D2ACULL, "qun220cr_javaris6072368.cl"},
	{0x1388726A5E7A54C4ULL, "623553798.ogg"},
	{0x1388A7C1D879A71BULL, "6136364_m.fxe"},
	{0x1388AA1E6A026DC6ULL, "vxlt_0_lowtown_a3_n_c_ghl_0_0.vlm"},
	{0x13890D4B9D37B5FAULL, "92596078.ogg"},
	{0x13897B576DFC1E29ULL, "follower_banter6170727.cl"},
	{0x13898E7D88769BCFULL, "vxlt_0_lowtown_a2_d3s_0_0.vlm"},
	{0x1389A3D3725E2EADULL, "vxca_caverns2prc_a58_0_0.vlm"},
	{0x1389D6C7474332C8ULL, "gen00fl_dog6170601.cl"},
	{0x1389EE6093E623D4ULL, "hlac_ht_0_hightown_a2_-196608_0z.dds"},
	{0x138A304CEE7D8085ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_cheers.ani"},
	{0x138B670CFEFB9B24ULL, "aip_qunariarishock_plot.gda"},
	{0x138C99A9D5111F51ULL, "6088764_m.fxe"},
	{0x138CC839A0CD749AULL, "272761559.ogg"},
	{0x138CD6E31DEC2880ULL, "art\\vfx\\creatures\\fxc_succubus_c.dura.evt"},
	{0x138D0F4ACA5B4AE6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_movement.anb"},
	{0x138DD7F1275116E3ULL, "6093601_m.fxe"},
	{0x138E0E1A8E80E80CULL, "vxw_woundedcoast_1io_0_0.vlm"},
	{0x138E24CBAFD9FCE0ULL, "6062825_f.fxe"},
	{0x138E35352DEB9363ULL, "104283188.ogg"},
	{0x138E474946928B4FULL, "vxlt_shop_j9_0_0.vlm"},
	{0x138E4C9EF177214AULL, "dae000ip_crowd_mf5_talk18.utp"},
	{0x138E8A1516D49FAFULL, "92925068.ogg"},
	{0x138EA9E9A2B68943ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_com.mao"},
	{0x138EAC619C21C0D0ULL, "vxlt_0_lowtown_n_e0s_0_0.vlm"},
	{0x138F0B0E69F768FEULL, "601706154.ogg"},
	{0x138F5441DEEB8545ULL, "root_mage_entropy.dds"},
	{0x138F5E6EED2195D3ULL, "art\\vfx\\combat\\vfx_glb_crust_c.impa.ani"},
	{0x138F85CED5C9DF48ULL, "407697944.ogg"},
	{0x13903068AE71F0C9ULL, "974229622.ogg"},
	{0x13904587B7529DDEULL, "vxca_cavern1_haunted_j2b_0_0.vlm"},
	{0x13907DE90ED7FAD9ULL, "art\\levels\\races\\design\\prp_bodybaga.msh"},
	{0x139087E90ED80BC7ULL, "art\\levels\\races\\design\\prp_bodybaga.mmh"},
	{0x13908BE90ED81214ULL, "art\\levels\\races\\design\\prp_bodybaga.mao"},
	{0x13908CBC58645C7BULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.impa.ani"},
	{0x13908CF5C7748101ULL, "vxht_bartrand_n_64r_0_0.vlm"},
	{0x139100E627737473ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzstage.phy"},
	{0x13923D51F9D41612ULL, "vxdr_primevaltomb_84j_0_0.vlm"},
	{0x1392736F6CC747EEULL, "gen_im_trs_slv_junk1_03.uti"},
	{0x1392E7E843AEB22DULL, "w_blightlands_163_0.mmh"},
	{0x1392EBE843AEB97EULL, "w_blightlands_163_0.mao"},
	{0x1392F9E843AED0B3ULL, "w_blightlands_163_0.msh"},
	{0x139316D60250FF07ULL, "261349989.ogg"},
	{0x13931F18786AEE6BULL, "6048704_m.fxe"},
	{0x1393C1743C99C429ULL, "hlac_lt_0_lowtown_a2_n_c_131072_1x.dds"},
	{0x1393FEC8B39BF3FBULL, "vxw_encounter2_l7_0_0.vlm"},
	{0x13943D158C799354ULL, "hlac_lt_0_lowtown_a2_n_c_65534_4y.dds"},
	{0x13952056E26DCF2EULL, "292271292.ogg"},
	{0x13953570355C22FBULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_wkf.ani"},
	{0x139538CEAC682A24ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutcn.mmh"},
	{0x13953AD04642CB03ULL, "w_woundedcoast_58_0.mao"},
	{0x13953CCEAC683177ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutcn.mao"},
	{0x13953ED04642D250ULL, "w_woundedcoast_58_0.mmh"},
	{0x13954CD04642EA6AULL, "w_woundedcoast_58_0.msh"},
	{0x13954ECEAC684FC6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutcn.msh"},
	{0x13956E2A6102E0F7ULL, "6051676_m.fxe"},
	{0x1395F90AF43F370FULL, "llac_w_encounter1_-1_0.dds"},
	{0x1396BA7F529CCEF1ULL, "872300986.ogg"},
	{0x1396D1F993D398F9ULL, "vxw_bonepitmining_a2_ja_0_0.vlm"},
	{0x1396E2E3432C77E1ULL, "vxdr_0_deeproads_bb3_0_0.vlm"},
	{0x13970E9350F20E89ULL, "910090309.ogg"},
	{0x13977319FEB77FEAULL, "6097837_m.fxe"},
	{0x1397C3A867227D3FULL, "308641782.ogg"},
	{0x1397DB47E5E8119FULL, "6101425_m.fxe"},
	{0x13981D1A3E0E3A9EULL, "1007235316.ogg"},
	{0x13982540ADE4C7B6ULL, "6163446_m.fxe"},
	{0x139836A781891E4BULL, "6032994_m.fxe"},
	{0x13983CDB086E4540ULL, "vxht_keep_702_0_0.vlm"},
	{0x1398AA8083D84317ULL, "follower_banter6153982.cl"},
	{0x1398BA26B82F8609ULL, "81130576.ogg"},
	{0x1398F28795CACB13ULL, "vxlt_0_lowtown_a2_n_h2g_0_0.vlm"},
	{0x13993F60E686221CULL, "hlac_lt_foundry_n_320_13y.dds"},
	{0x139962EB246AB553ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\aveline_exp_movement.anb"},
	{0x13999268ECEC5314ULL, "vxlt_hangedman_n_1te_0_0.vlm"},
	{0x1399D150B7F3035FULL, "869733476.ogg"},
	{0x139A54A034E5B85FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_tallwin.msh"},
	{0x139A5AA034E5C201ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_tallwin.mmh"},
	{0x139A66A034E5D6A2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_tallwin.mao"},
	{0x139A82FBDF7A8104ULL, "588062548.ogg"},
	{0x139AB3F269332F5BULL, "plt_rumor_mage_escort.nss"},
	{0x139B060CDD4838D5ULL, "6054394_m.fxe"},
	{0x139B45462D27B0B1ULL, "llac_lt_0_lowtown_docks_n_196608_0.dds"},
	{0x139B725BB512D08FULL, "vxht_mansion2_3c4_0_0.vlm"},
	{0x139BA93384BF76BBULL, "690289061.ogg"},
	{0x139C379BA831958AULL, "6108131_m.fxe"},
	{0x139C52B0ABBC1640ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bookcase_1.mao"},
	{0x139C566EED2C120AULL, "art\\vfx\\combat\\vfx_glb_crust_c.impa.evt"},
	{0x139CAEA8A34ED174ULL, "6128492_m.fxe"},
	{0x139D3A293E8EF58FULL, "vxlt_gallowscourtyard_n_7e8_0_0.vlm"},
	{0x139DB9655C95F359ULL, "pro000_colonel_ewald6047374.cl"},
	{0x139DD55EDC7891B2ULL, "follower_banter6143131.cl"},
	{0x139E0FB6CA1F2763ULL, "propalpha1_ds1y.psh.fxo"},
	{0x139E4E5D6AA8E12BULL, "133019110.ogg"},
	{0x139E93AE49ABB2EBULL, "vxlt_gallowstemplar_4sh_0_0.vlm"},
	{0x139E94BC5870A772ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.impa.evt"},
	{0x139EA98FD59EE195ULL, "gen00fl_aveline6002585.cl"},
	{0x139F07530B70391DULL, "721396326.ogg"},
	{0x139F4E8D4FD9A1D3ULL, "llac_lt_0_lowtown_a3_n_c_-3_4.dds"},
	{0x139F815F82C9876EULL, "374952790.ogg"},
	{0x13A032C9751A76CEULL, "zz_qun_debug6045402.cl"},
	{0x13A1236D75E7B53AULL, "hlac_lt_0_lowtown_docks_65532_1x.dds"},
	{0x13A123B43B84C01AULL, "ave200_donnic6059622.cl"},
	{0x13A190BB33C97C09ULL, "147769999.ogg"},
	{0x13A190F56A6106EBULL, "6043604_m.fxe"},
	{0x13A1A6DE5BFB10FAULL, "146675407.ogg"},
	{0x13A1A7DAEFE7FD33ULL, "vxlt_gallowstemplar_f_42x_0_0.vlm"},
	{0x13A233ABC91A81C1ULL, "293405963.ogg"},
	{0x13A25750E6379C3CULL, "120022973.ogg"},
	{0x13A2C3644CD18549ULL, "6123377_m.fxe"},
	{0x13A3356EED31F97DULL, "art\\vfx\\combat\\vfx_glb_crust_c.impa.gad"},
	{0x13A33D7035686DF2ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_wkf.evt"},
	{0x13A486572C9B5039ULL, "hightown_fem_elfservant_mob6169048.cl"},
	{0x13A4956642E695D2ULL, "96638512.ogg"},
	{0x13A49AFC66E70284ULL, "36476268.ogg"},
	{0x13A49CBEC384E071ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobd_0t.dds"},
	{0x13A4B33145F2CDAEULL, "one220_yevhen6200145.cl"},
	{0x13A4B3BC58754705ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.impa.gad"},
	{0x13A5243781094EACULL, "vxlt_0_lowtown_a2_n_c_ejc_0_0.vlm"},
	{0x13A5254E9C280A11ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_onyxstatue_0.pwk"},
	{0x13A58DD84DEE9EC6ULL, "vxlt_0_lowtown_a3_n_c_gbn_0_0.vlm"},
	{0x13A5C0C5CB63823BULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v2_c.mmh"},
	{0x13A623B2C8A697C2ULL, "965743493.ogg"},
	{0x13A7C60A757CDB99ULL, "lt_gallowscourtyard_36_0_lc.mmh"},
	{0x13A80445DEE26B94ULL, "vxlt_hangedman_n_1qj_0_0.vlm"},
	{0x13A80E3AABBDD0F6ULL, "vxht_mansion2b_2ab_0_0.vlm"},
	{0x13A89183434B00FBULL, "6110643_m.fxe"},
	{0x13A8F4609E5119FBULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheesebit.msh"},
	{0x13A8FFB030390D43ULL, "916373570.ogg"},
	{0x13A906609E513846ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheesebit.mao"},
	{0x13A90ECA6D8D259CULL, "257139456.ogg"},
	{0x13A912609E514CE5ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheesebit.mmh"},
	{0x13A92599C26649FEULL, "character_character_p162.psh.fxo.dx11.assembly"},
	{0x13A95C70356D0D85ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_wkf.gad"},
	{0x13A9E438A708E90FULL, "vxht_mansion2_3f7_0_0.vlm"},
	{0x13AA171051D9F23BULL, "hlac_lt_hangedman_-1_13z.dds"},
	{0x13AB26142DCD0EEDULL, "vxdr_0_deeproads_f0i_0_0.vlm"},
	{0x13AB4E55497BE5B0ULL, "6067930_f.fxe"},
	{0x13AC1B7619E62A22ULL, "zz_followers_banter6157697.cl"},
	{0x13AC86A91F25B646ULL, "901339757.ogg"},
	{0x13ACA56C408F6413ULL, "593073325.ogg"},
	{0x13ACD0CA900233AAULL, "one280_iwan6120122.cl"},
	{0x13ACE0D0B258769BULL, "667773047.ogg"},
	{0x13AD349177452B0DULL, "350465295.ogg"},
	{0x13AD62041CD405EAULL, "gen00fl_varric6076212.cl"},
	{0x13AE2305556B5E19ULL, "757137164.ogg"},
	{0x13AE57257A2A1657ULL, "415256888.ogg"},
	{0x13AE642318946E78ULL, "ave260tr_long_road_beasts.ncs"},
	{0x13AEADF77F50772FULL, "mer340_cowan6054129.cl"},
	{0x13AEB331E954D7F8ULL, "vxlt_0_lowtown_docks_cp3_0_0.vlm"},
	{0x13AEBBEE9C2AAF8AULL, "lt_0_lowtown_docks_38_0.rml"},
	{0x13AEBF3C9446E3AEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxa.mmh"},
	{0x13AECB3C9446F845ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxa.mao"},
	{0x13AED93C94470F9CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxa.msh"},
	{0x13AEE6D19DD67CA2ULL, "vxht_0_hightown_a3_9up_0_0.vlm"},
	{0x13AF21037A44AB59ULL, "697714968.ogg"},
	{0x13AF360F0570F3F4ULL, "llac_ht_keep_n_3736_3.dds"},
	{0x13AFBD8E4B52C097ULL, "vxlt_warehouse_n_ct_0_0.vlm"},
	{0x13AFC0D2A060E51BULL, "553944030.ogg"},
	{0x13AFD42199CF5FA7ULL, "qcr220_meredith6041546.cl"},
	{0x13B025A02568F484ULL, "6131919_m.fxe"},
	{0x13B0684AE1954FF2ULL, "vxht_0_hightown_n_bum_0_0.vlm"},
	{0x13B06D0DA94403FAULL, "vxlt_blackemporium_ty_0_0.vlm"},
	{0x13B093F3D2CD51DBULL, "mag320_orsino6071231.cl"},
	{0x13B0A4B0F89BF0ECULL, "w_bonepitdestroyed_a3_110_0.msh"},
	{0x13B0AAB0F89BFADEULL, "w_bonepitdestroyed_a3_110_0.mmh"},
	{0x13B0B6B0F89C0F75ULL, "w_bonepitdestroyed_a3_110_0.mao"},
	{0x13B1460267FD5A72ULL, "vxlt_gallowstemplar_n_45l_0_0.vlm"},
	{0x13B165D6816E676CULL, "818576603.ogg"},
	{0x13B23F1E0CF8BA3AULL, "llac_lt_gallowstemplar_n_7198_2.dds"},
	{0x13B27C89100FA50BULL, "vxlt_0_lowtown_a2_go4_0_0.vlm"},
	{0x13B290740E1F3161ULL, "411962066.ogg"},
	{0x13B2B22243B2C00FULL, "zz_approval_debug6003673.cl"},
	{0x13B2E1C5CB6E44A0ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v2_c.anb"},
	{0x13B34F46A0C1DE53ULL, "vxw_bonepitdestroyed_a3_2a4_0_0.vlm"},
	{0x13B430FCC4EBFD4BULL, "dae000_gamlen6162113.cl"},
	{0x13B47144FC33FCA1ULL, "465227982.ogg"},
	{0x13B4E628C87F86E9ULL, "vxht_chantry_211_0_0.vlm"},
	{0x13B4F1DF66D9F9F3ULL, "zz_dog_debug.cnv"},
	{0x13B55D990F6AE396ULL, "71689391.ogg"},
	{0x13B5621E146BB229ULL, "6167250_m.fxe"},
	{0x13B5DBD3BCEF63CAULL, "vxca_caverns1_mines_jg5_0_0.vlm"},
	{0x13B6087922BE3959ULL, "ht_mansion_n_3365.rml"},
	{0x13B62A76E5165F12ULL, "216388905.ogg"},
	{0x13B656974A3FAE00ULL, "llac_ca_cavern1_haunted_24424_13.dds"},
	{0x13B6B5983448B5B5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_hightown_windows_e.dds"},
	{0x13B735BCFB8ED8F3ULL, "6085783_m.fxe"},
	{0x13B74304BD3A877CULL, "275801359.ogg"},
	{0x13B7CA0CBAFA6877ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.impa.gad"},
	{0x13B7DCFB393A9FF5ULL, "hlac_lt_warehouse_n_1_1y.dds"},
	{0x13B7FEDCEE5606C3ULL, "371882129.ogg"},
	{0x13B8123347DDCD3EULL, "vxlt_warehouse_n_158_0_0.vlm"},
	{0x13B841C9E541A9EFULL, "hlac_ca_caverns1_mines_24424_28y.dds"},
	{0x13B88B3FF5FF194AULL, "l_lgt261ar_tome_cave.bnk"},
	{0x13B8B240679D4913ULL, "1002361758.ogg"},
	{0x13B8E072241CCF5DULL, "ht_0_hightown_34_0_ll.mmh"},
	{0x13B96B508679EB7AULL, "6116389_m.fxe"},
	{0x13B998BA96A92972ULL, "vxht_keep_6cd_0_0.vlm"},
	{0x13B99D996F138C65ULL, "hlac_w_sundermount_327677_0x.dds"},
	{0x13BA156014C8E66EULL, "788632998.ogg"},
	{0x13BA694F9576464EULL, "vxht_mansion2_3cy_0_0.vlm"},
	{0x13BA6AC8B639CFECULL, "873407469.ogg"},
	{0x13BA8B2685618F49ULL, "character_character_p35.psh.fxo.dx11"},
	{0x13BAEEC58FE96702ULL, "vxht_mansion2_32j_0_0.vlm"},
	{0x13BBA43D99BAD20FULL, "hlac_dr_0_deeproads_5845_9x.dds"},
	{0x13BBE7283928A629ULL, "llac_w_bonepit_393222_0.dds"},
	{0x13BC2BA169012FA2ULL, "229371666.ogg"},
	{0x13BC311D1EF26B4EULL, "vxlt_gallowscourtyard_n_7eu_0_0.vlm"},
	{0x13BC53C61DC85209ULL, "vxw_woundedcoast_10i_0_0.vlm"},
	{0x13BC9704CAE7270BULL, "1053720588.ogg"},
	{0x13BCC30089625D00ULL, "6055644_m.fxe"},
	{0x13BCCFFC62167CD2ULL, "vxw_bonepitdestroyed_a3_2a1_0_0.vlm"},
	{0x13BDC455546B6BADULL, "hlac_ht_keep_n_3732_13z.dds"},
	{0x13BDF843DD1D83ACULL, "6146953_m.fxe"},
	{0x13BE8E35B7620C48ULL, "6042290_f.fxe"},
	{0x13BEBEA975B126AAULL, "100309676.ogg"},
	{0x13BEDF0CBB00AA04ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.impa.evt"},
	{0x13BEE7195AE52F3CULL, "qcr000_arishok6059321.cl"},
	{0x13BF1CF15819E5C4ULL, "75630494.ogg"},
	{0x13BF4F280012D303ULL, "hlac_lt_0_lowtown_-131073_0x.dds"},
	{0x13C0519CC1D0A8B8ULL, "ser221_party_comments6166112.cl"},
	{0x13C06B60B3441950ULL, "follower_banter6155538.cl"},
	{0x13C0DF2640C0A1A0ULL, "llac_ht_chantry_n_2534_22.dds"},
	{0x13C0EE660205BD8BULL, "vxlt_0_lowtown_a2_gb7_0_0.vlm"},
	{0x13C147BA8E01F03DULL, "775862284.ogg"},
	{0x13C17E724D59168FULL, "hlac_w_bonepitmining_a2_327681_0x.dds"},
	{0x13C2A7CFD2370D5CULL, "hlac_lt_0_lowtown_docks_a3_-3_7x.dds"},
	{0x13C2F0D838C730B9ULL, "6163816_m.fxe"},
	{0x13C300E5CC1A24B8ULL, "ht_0_hightown_a2_35_0.rml"},
	{0x13C332F199039518ULL, "510702760.ogg"},
	{0x13C38CD920A45DD6ULL, "zz_strider_stomp_v2.cut"},
	{0x13C4F3090B90B775ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2e.msh"},
	{0x13C4F5090B90BAEBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2e.mmh"},
	{0x13C501090B90CE88ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2e.mao"},
	{0x13C5027271EB0A31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chant2door.msh"},
	{0x13C5107271EB224CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chant2door.mao"},
	{0x13C51C7271EB36EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chant2door.mmh"},
	{0x13C595BA0D718E18ULL, "vxlt_0_lowtown_a3_n_c_ex0_0_0.vlm"},
	{0x13C59AA81B6199ECULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spike_d.dds"},
	{0x13C5CC068829AF8AULL, "hlac_lt_0_lowtown_a2_n_0_4z.dds"},
	{0x13C65BA0350AFF7EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_tallwin.phy"},
	{0x13C67B2318A92775ULL, "ave260tr_long_road_beasts.utt"},
	{0x13C688AEF6DF65E6ULL, "hlac_ht_0_hightown_-65537_0z.dds"},
	{0x13C702C780E6D1C7ULL, "face_rendertodepth_p1.vsh.fxo"},
	{0x13C703AC59472882ULL, "vxlt_warehouse_sx_0_0.vlm"},
	{0x13C70BA12297C91BULL, "1043326484.ogg"},
	{0x13C70E5F65E5FD19ULL, "9668053.ogg"},
	{0x13C72E102C727E7AULL, "w_blightlands_177_0.mmh"},
	{0x13C732102C7284B1ULL, "w_blightlands_177_0.mao"},
	{0x13C740102C729CC0ULL, "w_blightlands_177_0.msh"},
	{0x13C7CB678EB8F709ULL, "art\\characters\\playercharacter\\elfmale\\glv_lgte.phy"},
	{0x13C7EFA37878A668ULL, "w_encounter3_197_0.mmh"},
	{0x13C7FBA37878BA0BULL, "w_encounter3_197_0.mao"},
	{0x13C80DA37878D892ULL, "w_encounter3_197_0.msh"},
	{0x13C88966ACC9E920ULL, "1015176672.ogg"},
	{0x13C8DB2C876C5ECEULL, "vxht_mansion2_3fz_0_0.vlm"},
	{0x13C91C580B6C28C2ULL, "hlac_ht_chantry_n_2534_1y.dds"},
	{0x13C9956370E8ED7FULL, "vxca_caverns1_mines_j4i_0_0.vlm"},
	{0x13C9CAC687E0E371ULL, "6170908_m.fxe"},
	{0x13C9D60DCCF78005ULL, "6054695_m.fxe"},
	{0x13CA4413081DF45BULL, "vxlt_hangedman_n_25f_0_0.vlm"},
	{0x13CABA6341C69C7EULL, "mer221_sabin6044072.cl"},
	{0x13CABEF2695C65F4ULL, "plt_rumor_mage_escort.plo"},
	{0x13CAE823BF4C3B3CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1_0.pwk"},
	{0x13CB2FDABF8E82D3ULL, "default_blackzero.dds"},
	{0x13CB34DD7B587580ULL, "6052744_m.fxe"},
	{0x13CB3AE7AB1BF567ULL, "6110535_m.fxe"},
	{0x13CB42B35F73C1B2ULL, "446900160.ogg"},
	{0x13CB7B266DE49A13ULL, "gen_im_arm_cht_hvy_act3_03.uti"},
	{0x13CBB42DB04AFFE8ULL, "w_sundermount_97_0.rml"},
	{0x13CBF70CBB0B5DFDULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.impa.ani"},
	{0x13CC0B434B7E25FFULL, "llac_dr_primevaltomb_664_7.dds"},
	{0x13CCC0469F59EC5DULL, "dae210_gate_guard6085081.cl"},
	{0x13CD41648F9AA906ULL, "6037422_m.fxe"},
	{0x13CDAA5B9272C800ULL, "gen_im_arm_cht_nobe_t5.uti"},
	{0x13CE27E3B35FD9F4ULL, "zz_qun_debug6085862.cl"},
	{0x13CE3ED046729713ULL, "w_woundedcoast_58_0.rml"},
	{0x13CE9317403FCBF1ULL, "vxlt_backalley_g_1r4_0_2.vlm"},
	{0x13CE96688AEACC88ULL, "dae011_drunk1_sta6149685.cl"},
	{0x13CEEC8006CE4118ULL, "mag311_emile_de_launcet6043865.cl"},
	{0x13CF04ACE2D32E16ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundenter.ani"},
	{0x13CFB2322AA83938ULL, "vxlt_ruinedhouse_n_5s_0_0.vlm"},
	{0x13D005F8EA99B124ULL, "6141703_m.fxe"},
	{0x13D042EC69B88138ULL, "vxca_cavern1_haunted_e8a_0_0.vlm"},
	{0x13D08C3DF9A11F4BULL, "vxlt_warehouse_n_di_0_0.vlm"},
	{0x13D0E8233FF61F70ULL, "vxlt_gallowstemplar_n_3kk_0_0.vlm"},
	{0x13D196353986EE90ULL, "vxdr_0_deeproads_cr3_0_0.vlm"},
	{0x13D1D512B34F3B0BULL, "6029136_m.fxe"},
	{0x13D2094BD566D0E6ULL, "vxdr_0_deeproads_ax1_0_0.vlm"},
	{0x13D24AEC02DE4942ULL, "mcr371_scared_mage_f.cnv"},
	{0x13D2BC0FDBFBB5EAULL, "llac_w_bonepitmining_a2_393218_0.dds"},
	{0x13D2BF00CEADC29DULL, "vxca_caverns1_k6m_0_0.vlm"},
	{0x13D2EF6997849A17ULL, "6120125_m.fxe"},
	{0x13D2F0B1C2EABBCFULL, "w_encounter3_16_0.mao"},
	{0x13D2FCB1C2EAD06CULL, "w_encounter3_16_0.mmh"},
	{0x13D302B1C2EADA5EULL, "w_encounter3_16_0.msh"},
	{0x13D3D577543C3818ULL, "6174585_m.fxe"},
	{0x13D3E128E1CC40E1ULL, "dre120tr_meet_varric.ncs"},
	{0x13D3F1253D8DC1F6ULL, "hlac_ca_cavern1_haunted_24424_1y.dds"},
	{0x13D44E4493E64DD7ULL, "zz_one_debug6025814.cl"},
	{0x13D4639243F06708ULL, "fx_st_cloud_a.dds"},
	{0x13D4B7DB932ED152ULL, "292322181.ogg"},
	{0x13D52CEC02E04C7AULL, "mcr371_scared_mage_f.bnk"},
	{0x13D555D772143072ULL, "gen00fl_anders6119409.cl"},
	{0x13D575894B3D4102ULL, "vxlt_warehouse_ty_0_0.vlm"},
	{0x13D5C80B6B1DA337ULL, "6116058_m.fxe"},
	{0x13D5E71730D2DC6DULL, "hlac_lt_0_lowtown_a2_n_-131072_0z.dds"},
	{0x13D601F058CA87FFULL, "844261008.ogg"},
	{0x13D634E98E8AA999ULL, "997537315.ogg"},
	{0x13D692C127D683BAULL, "6000738_m.fxe"},
	{0x13D7454393A753EDULL, "865139742.ogg"},
	{0x13D78C602FE3EFAEULL, "vxw_encounter2_camp_j8_0_0.vlm"},
	{0x13D79A73E9A46A36ULL, "vxlt_0_lowtown_a2_ffo_0_0.vlm"},
	{0x13D8665F0535214EULL, "lt_0_lowtown_docks_185_0.mmh"},
	{0x13D86757C8300E55ULL, "vxlt_0_lowtown_docks_bb9_0_0.vlm"},
	{0x13D8725F053535E5ULL, "lt_0_lowtown_docks_185_0.mao"},
	{0x13D8805F05354D3CULL, "lt_0_lowtown_docks_185_0.msh"},
	{0x13D8C1A7E364FAE6ULL, "6072549_f.fxe"},
	{0x13D8E280D4F1634EULL, "hlac_lt_0_lowtown_docks_n_65535_2y.dds"},
	{0x13D95F928E70BA86ULL, "nat380_nathaniel6070262.cl"},
	{0x13D9E10D635EA357ULL, "vxlt_house_gj_0_0.vlm"},
	{0x13D9E9B12B71EAABULL, "ave200_donnic6111983.cl"},
	{0x13DA14DF66F95148ULL, "zz_dog_debug.ncs"},
	{0x13DA5CC984D87BD3ULL, "follower_banter6171054.cl"},
	{0x13DA691E14702CB5ULL, "llac_ht_0_hightown_n_65536_4.dds"},
	{0x13DA7546DD65988CULL, "102558571.ogg"},
	{0x13DA85AE3CAE440FULL, "llac_dr_dungeon_3215_6.dds"},
	{0x13DAA3982167F4C6ULL, "llac_lt_backalley_n_65536_6.dds"},
	{0x13DAB73611F6E112ULL, "hlac_lt_gallowscourtyard_n_-131073_0y.dds"},
	{0x13DAD4DEEA72A5CFULL, "6119419_m.fxe"},
	{0x13DB64B1ADB89345ULL, "vxlt_blackemporium_tr_0_0.vlm"},
	{0x13DC70C6058341F5ULL, "lt_0_lowtown_docks_101_0.phy"},
	{0x13DC9A06A5D4F727ULL, "vis121_saemus6100087.cl"},
	{0x13DD0CACE2DF7907ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundenter.evt"},
	{0x13DD71A5A60360BAULL, "abilities.gfx"},
	{0x13DDE020B200EC01ULL, "6162363_m.fxe"},
	{0x13DEC57F7A1E5AA6ULL, "6116752_m.fxe"},
	{0x13DFF15C988B339AULL, "281647208.ogg"},
	{0x13DFFBEABC215143ULL, "vxlt_0_lowtown_a2_n_c_gzb_0_0.vlm"},
	{0x13E03F4AA034BAAEULL, "vxht_keep_n_53d_0_0.vlm"},
	{0x13E062689C11CFADULL, "art\\vfx\\materials\\chunky\\vfx_shield_blend.mao"},
	{0x13E0AFF467D9669EULL, "dae300_alistair_king6134959.cl"},
	{0x13E0D9C2B1F42E60ULL, "vxht_keep_n_5zq_0_0.vlm"},
	{0x13E0DB6DA3B5BF65ULL, "88087187.ogg"},
	{0x13E0F072970A8DDDULL, "6033659_f.fxe"},
	{0x13E16471573D691DULL, "vxca_caverns2prc_92u_0_0.vlm"},
	{0x13E17E8CEF0126A3ULL, "253671173.ogg"},
	{0x13E19539E6C21593ULL, "vxlt_shop_et_0_0.vlm"},
	{0x13E214CB39474C4FULL, "vxht_0_hightown_a3_bxe_0_0.vlm"},
	{0x13E3196DA8E1F703ULL, "6000798_f.fxe"},
	{0x13E3208125179A0FULL, "hlac_ht_mansion2_818_43x.dds"},
	{0x13E3673376C6FF6FULL, "vxht_0_hightown_a2_bm7_0_0.vlm"},
	{0x13E36E54200A6130ULL, "dae000_elegant6166916.cl"},
	{0x13E3E226F3B1C116ULL, "676766697.ogg"},
	{0x13E3E7663D335982ULL, "vxlt_warehouse_yz_0_0.vlm"},
	{0x13E3F0D1E3BC7FB1ULL, "6035157_m.fxe"},
	{0x13E42DACE2E5D050ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundenter.gad"},
	{0x13E4ED93BF537EC1ULL, "prm_col_place_promo.ncs"},
	{0x13E4FABEEDBA9C48ULL, "llac_lt_0_lowtown_a2_n_-3_6.dds"},
	{0x13E5C00F08CC5A89ULL, "vxca_caverns1_fm9_0_0.vlm"},
	{0x13E626576CA22C20ULL, "vxht_mansion2_1pp_0_0.vlm"},
	{0x13E6371E39ABFC41ULL, "vxdr_slavers_210_0_0.vlm"},
	{0x13E70605D9328B96ULL, "vxw_bonepit_16c_0_0.vlm"},
	{0x13E7112F2D23D559ULL, "gen_im_cft_rec_wrn_01.uti"},
	{0x13E7741F96DFCCE5ULL, "hlac_w_bonepitmining_a2_262146_0z.dds"},
	{0x13E7A52F11A2C16FULL, "hlac_lt_darktown_-1_16x.dds"},
	{0x13E7BE00A88E6C23ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_2y.dds"},
	{0x13E7FC008A34CFB3ULL, "vxlt_gallowscourtyard_6w8_0_0.vlm"},
	{0x13E80D3EE09B143AULL, "6173494_m.fxe"},
	{0x13E8898120FAEB35ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_app_aoe_p.mmh"},
	{0x13E8B44AF37D757FULL, "str_slime_d.dds"},
	{0x13E8C1BB02545DC1ULL, "197664093.ogg"},
	{0x13E98502132B2297ULL, "vxlt_warehouse_n_wp_0_0.vlm"},
	{0x13E9B19CBBEFB949ULL, "6032127_m.fxe"},
	{0x13E9BC82D393F2A1ULL, "t3_wdg_br3.tnt"},
	{0x13EA22060CD3AE76ULL, "dae000_mother6177016.cl"},
	{0x13EA2FBEA96DB1F2ULL, "vxht_0_hightown_n_bai_0_0.vlm"},
	{0x13EA645BA5EC29EBULL, "340501057.ogg"},
	{0x13EA8090BBD8CA98ULL, "mcr371_post_harvester6083682.cl"},
	{0x13EAA06F025ADBBBULL, "hlac_lt_gallowstemplar_f_2384_3z.dds"},
	{0x13EAAEF793FD59F4ULL, "lowtown_male_beggar_sta6151027.cl"},
	{0x13EAF4A7E5D30CD3ULL, "vxdr_primevaltomb_9e5_0_0.vlm"},
	{0x13EB3CB666A18A14ULL, "vxlt_0_lowtown_docks_b5h_0_1.vlm"},
	{0x13EC0507618DA10EULL, "268537991.ogg"},
	{0x13EC1206FD02789FULL, "78091235.ogg"},
	{0x13EC170687D2714FULL, "hlac_lt_0_lowtown_a2_n_c_-65536_7x.dds"},
	{0x13ED5A520A85BD25ULL, "6082164_m.fxe"},
	{0x13EE8649BE7E726AULL, "595585751.ogg"},
	{0x13EEAA2942E15AFFULL, "hlac_lt_0_lowtown_docks_n_-65538_2x.dds"},
	{0x13EEE1E5CC3F2B70ULL, "ht_0_hightown_a2_35_0.anb"},
	{0x13EEEA75D3BDCF83ULL, "6105955_m.fxe"},
	{0x13EEFB367D8522C3ULL, "mag160_feynriel6020095.cl"},
	{0x13F0374817962A9BULL, "hlac_w_sundermount_196605_0y.dds"},
	{0x13F03DB8EF13CB96ULL, "709274962.ogg"},
	{0x13F086A82B3D64CFULL, "vxht_0_hightown_a3_bwd_0_0.vlm"},
	{0x13F0D1B39BF4FAAAULL, "vxw_encounter2_camp_b4_0_0.vlm"},
	{0x13F15CD92187AD7CULL, "hlac_lt_undercityencounter_-1_40z.dds"},
	{0x13F1622D5482BADEULL, "838503544.ogg"},
	{0x13F162B2E3A651AEULL, "883909045.ogg"},
	{0x13F210804046B731ULL, "6024109_f.fxe"},
	{0x13F21E01EC1223F6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_octwindow.mao"},
	{0x13F22A01EC123795ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_octwindow.mmh"},
	{0x13F22C01EC123B0BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_octwindow.msh"},
	{0x13F24C06AAC6542BULL, "vxw_encounter2_fe_0_0.vlm"},
	{0x13F2E4724A05F235ULL, "vxlt_warehouse_15l_0_0.vlm"},
	{0x13F2F9E627C632E8ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzstage.mmh"},
	{0x13F305E627C6468BULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzstage.mao"},
	{0x13F317E627C66512ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzstage.msh"},
	{0x13F36D451DB2AB8AULL, "hlac_lt_foundry_n_320_7z.dds"},
	{0x13F3AFEFDF3C4255ULL, "vxw_bonepitdestroyed_a3_1v4_0_0.vlm"},
	{0x13F3B4678EDE0B2EULL, "art\\characters\\playercharacter\\elfmale\\glv_lgte.mmh"},
	{0x13F3C0678EDE1FC5ULL, "art\\characters\\playercharacter\\elfmale\\glv_lgte.mao"},
	{0x13F3CE678EDE371CULL, "art\\characters\\playercharacter\\elfmale\\glv_lgte.msh"},
	{0x13F3DFB0B8F5C95EULL, "641904176.ogg"},
	{0x13F430EC02FB030FULL, "mcr371_scared_mage_f.utc"},
	{0x13F444198EE2512AULL, "ts_kirkwall_01.gda"},
	{0x13F448B07C5F281BULL, "vxht_0_hightown_c85_0_0.vlm"},
	{0x13F4D4935822E6FFULL, "dae000_meredith6092609.cl"},
	{0x13F656C77C9E377BULL, "vxht_brothel_n_6c9_0_0.vlm"},
	{0x13F68681210723C2ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_app_aoe_p.anb"},
	{0x13F6BCA657CA1118ULL, "hlac_lt_0_lowtown_n_-2_0y.dds"},
	{0x13F6E2E56EC7CE31ULL, "vxw_bonepitmining_a2_ed_0_0.vlm"},
	{0x13F7D9B381982FABULL, "dae011_corff_sta6149641.cl"},
	{0x13F809FF6CB2CAAAULL, "docks_male_human_worker_sta6175937.cl"},
	{0x13F8253341A27A8DULL, "6087039_m.fxe"},
	{0x13F91292FDFC8EDBULL, "mer240_jansen6050332.cl"},
	{0x13F93B877D0E3F0AULL, "quickbar.gfx"},
	{0x13F97D443BA6BC39ULL, "plt_lgt120pt_bhelen_lt.plo"},
	{0x13F9953F5B25E076ULL, "mcr310_loyalist_templars.stg"},
	{0x13F9F28DDB70BCDEULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.cess.evt"},
	{0x13FA5AF72254DE2BULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwounddeathblow.evt"},
	{0x13FAB74BF2AA4FD2ULL, "follower_banter6164709.cl"},
	{0x13FAC13A072902B3ULL, "vxht_mansion2_3di_0_0.vlm"},
	{0x13FB0B29A1858DC0ULL, "vxlt_0_lowtown_ffu_0_0.vlm"},
	{0x13FB140F9EEA9441ULL, "hlac_lt_0_lowtown_a2_n_c_0_1x.dds"},
	{0x13FBE1D4CF8368CBULL, "vxlt_warehouse_n_uj_0_0.vlm"},
	{0x13FCAE9EEBBDC246ULL, "plt_cod_crt_wolf.plo"},
	{0x13FCDFA4A00D8243ULL, "vxlt_0_lowtown_a2_n_c_gpd_0_0.vlm"},
	{0x13FD0C909D74186DULL, "w_encounter3_255_0.mmh"},
	{0x13FD10909D741EBEULL, "w_encounter3_255_0.mao"},
	{0x13FD1E909D7436F3ULL, "w_encounter3_255_0.msh"},
	{0x13FDACB171CFFD21ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_combat_idle.ani"},
	{0x13FDBB7487376291ULL, "vxht_mansion_n_4yt_0_0.vlm"},
	{0x13FDC97C6ADEB14CULL, "114686891.ogg"},
	{0x13FE462329FAF58BULL, "1053896777.ogg"},
	{0x13FE6321BF7EAD43ULL, "mag121_idunna6084237.cl"},
	{0x13FEC64E7471F905ULL, "mrl000_marethari6129257.cl"},
	{0x13FF30B4DDE1A1DBULL, "6044296_m.fxe"},
	{0x13FF55B60800FFBAULL, "6166957_m.fxe"},
	{0x140033E3FB8BEDE1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\aveline_injured_movement.anb"},
	{0x1400BDE39CBC6CABULL, "vxw_encounter2_ad_0_0.vlm"},
	{0x1400F8F8F1823FFFULL, "gen00fl_merrill6107032.cl"},
	{0x1401098DDB770209ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.cess.gad"},
	{0x14013B1FCC1C0AF6ULL, "934630494.ogg"},
	{0x140173F7225B27ECULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwounddeathblow.gad"},
	{0x1401984A804312B4ULL, "vxlt_hangedman_1vd_0_0.vlm"},
	{0x1401BA563D69B047ULL, "283499189.ogg"},
	{0x1401C528A967C0A0ULL, "lt_0_lowtown_a2_n_30_0.rml"},
	{0x1401DD7EC8D00335ULL, "gen00fl_merrill6112339.cl"},
	{0x140211913BF3E522ULL, "llac_lt_darktown_-1_26.dds"},
	{0x14023A8C2AB08B65ULL, "vxca_caverns2prc_8kr_0_1.vlm"},
	{0x140283EDCB15B95EULL, "hlac_ca_caverns1_mines_24424_15x.dds"},
	{0x14029E0D4EFB9DDBULL, "vxht_bartrand_n_4lb_0_0.vlm"},
	{0x14048E30C61ADA1EULL, "298542002.ogg"},
	{0x1404B7B171D62EC3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_combat_idle.gad"},
	{0x1405009ACEAE829BULL, "6114091_f.fxe"},
	{0x14051B2C9C485861ULL, "6113244_m.fxe"},
	{0x140595FC22DD5862ULL, "gen00fl_varric6093617.cl"},
	{0x14060C619B600A23ULL, "lgt210_magu_selby6142012.cl"},
	{0x140669F011F75036ULL, "vxht_keep_73a_0_0.vlm"},
	{0x1407141E3F9AFF5CULL, "6088202_m.fxe"},
	{0x14073437A4204EF6ULL, "or_db_smk.mao"},
	{0x14073B2B29279038ULL, "zz_fol_debug6103800.cl"},
	{0x14076AEEF2D0FDA1ULL, "vxlt_hangedman_n_w2_0_0.vlm"},
	{0x1407C980C3B6D8C7ULL, "6117732_m.fxe"},
	{0x1407DA8DDB7CD157ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.cess.ani"},
	{0x140816DCC637AE2BULL, "hlac_lt_0_lowtown_docks_-131073_5x.dds"},
	{0x140840CE33EDC6F3ULL, "vxca_caverns2_80a_0_0.vlm"},
	{0x140872F7226144EAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwounddeathblow.ani"},
	{0x1408BD21992E1630ULL, "llac_ht_mansion2b_818_8.dds"},
	{0x1408C553AD91DEE6ULL, "1008757749.ogg"},
	{0x1409130D3FD71171ULL, "vxht_0_hightown_a3_c59_0_0.vlm"},
	{0x14094C43D7BACE1CULL, "6154015_m.fxe"},
	{0x1409814914F8596AULL, "344622902.ogg"},
	{0x140993424EB1F3DEULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluvian_sa.dds"},
	{0x140A53B1C179814BULL, "vxlt_warehouse_n_xm_0_0.vlm"},
	{0x140A550F0BBDEB1AULL, "zz_slate5_track.cut"},
	{0x140A597CA873A19AULL, "art\\vfx\\combat\\vfx_r_ibl_appear_imp_p.anb"},
	{0x140ABFA5DD20D554ULL, "6044623_f.fxe"},
	{0x140AD609FBA3ABE1ULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.anb"},
	{0x140B2E29C4776EDEULL, "6102793_m.fxe"},
	{0x140B94B171DC11A8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_combat_idle.evt"},
	{0x140BB8DC761D1577ULL, "vrc221_bartrand6065289.cl"},
	{0x140BBCC65DED97B8ULL, "vxlt_undercitydungeon_1a7_0_0.vlm"},
	{0x140C3260D5A4FAD3ULL, "plt_mrl300pt_n_firsttalk.plo"},
	{0x140C5D0E7578BAB7ULL, "6051063_m.fxe"},
	{0x140D4A64B8F6A728ULL, "vxlt_0_lowtown_n_ex3_0_0.vlm"},
	{0x140D6A620F2995CFULL, "vxht_0_hightown_a3_bif_0_0.vlm"},
	{0x140DC9F2AF11F3A5ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.anb"},
	{0x140DDCA11DBD2C4DULL, "llac_lt_0_lowtown_docks_-196611_2.dds"},
	{0x140F2FC08EB2852BULL, "vxw_encounter2_lg_0_0.vlm"},
	{0x140F9A358E348844ULL, "follower_banter6172906.cl"},
	{0x140FD6802F6EB648ULL, "6066911_f.fxe"},
	{0x140FE4DCBEB84929ULL, "vxlt_gallowscourtyard_n_4uu_0_0.vlm"},
	{0x141015AF105FBAA6ULL, "mag360ip_barrier1.utp"},
	{0x1410E8824D718E0FULL, "vxw_encounter3_d4_0_0.vlm"},
	{0x1410F3E5680149F1ULL, "hlac_ht_keep_n_4519_7y.dds"},
	{0x1410FEA6F157A59FULL, "mag330_walter.bnk"},
	{0x14115D2EB933C5C8ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_2x.dds"},
	{0x14115E10341A5E76ULL, "6127019_m.fxe"},
	{0x14119885E532A0EBULL, "l_one100ar_ironwood.wbd"},
	{0x1411A492D8C0AACCULL, "6154943_m.fxe"},
	{0x1411A499FB0BCEACULL, "vxlt_gamlenshouse_n_6r_0_0.vlm"},
	{0x1412083EE898D793ULL, "vxlt_gallowscourtyard_78p_0_0.vlm"},
	{0x141268890350B2B6ULL, "hlac_ht_chantry_2534_39y.dds"},
	{0x1413C69F52B3FF31ULL, "vxw_bonepitmining_a2_oz_0_0.vlm"},
	{0x1413DC1BEEFFDE95ULL, "vxlt_0_lowtown_fdz_0_0.vlm"},
	{0x1414B73A56DA6C3BULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p.mmh"},
	{0x1414E0A6F15B5BA7ULL, "mag330_walter.cnv"},
	{0x14159B40BFF2AF2DULL, "vxw_bonepitdestroyed_a3_1wn_0_1.vlm"},
	{0x1416078FDB14970AULL, "monster3.cut"},
	{0x14162E92D0B69B90ULL, "vxht_chantry_2v1_0_0.vlm"},
	{0x1416AC0CDD76F2AFULL, "6082276_m.fxe"},
	{0x1416F4D3E5B59FA6ULL, "919881411.ogg"},
	{0x1417276F5BA7BF0EULL, "hla_ht_0_hightown_a2_-131072_0y.dds"},
	{0x1417A0ACF4AF88D9ULL, "plt_cod_cha_fenris3.plo"},
	{0x1417B709FBAE015AULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.mmh"},
	{0x141916D03984D201ULL, "hlac_dr_0_deeproads_5840_0z.dds"},
	{0x14192173F9C499F0ULL, "vxlt_gallowstemplar_n_3rt_0_0.vlm"},
	{0x1419A2BC9EAF5102ULL, "673997328.ogg"},
	{0x141A2EA34FE3B038ULL, "vxlt_undercitydungeon_1d4_0_0.vlm"},
	{0x141AEA28E20852F4ULL, "dre120tr_meet_varric.utt"},
	{0x141AEB59329E36BAULL, "lgt220_mag_tmplr_templar6140352.cl"},
	{0x141B3392644A9C77ULL, "hlac_lt_0_lowtown_docks_a3_-65538_7y.dds"},
	{0x141B78443BC39A4EULL, "plt_lgt120pt_bhelen_lt.nss"},
	{0x141B9D223B6E94F6ULL, "vxht_mansion2b_29j_0_0.vlm"},
	{0x141BDB40ACD9B23EULL, "6122104_m.fxe"},
	{0x141C20DBAA0A3078ULL, "717351400.ogg"},
	{0x141CBF22DC1D37E0ULL, "6071480_m.fxe"},
	{0x141D5E6A35AF073CULL, "623214152.ogg"},
	{0x141D970AE280A873ULL, "773084745.ogg"},
	{0x141E429565FD71BDULL, "pro000_merchant6051584.cl"},
	{0x141E7C04642F43B4ULL, "vxlt_hangedman_1xj_0_0.vlm"},
	{0x141F5A5F3F67A68FULL, "vxw_encounter3_a7_0_0.vlm"},
	{0x141F81786A07559BULL, "69075396.ogg"},
	{0x142084E0AF164678ULL, "vxdr_slavers_1t1_0_0.vlm"},
	{0x1420E4B532B0ACB2ULL, "vrc321_varric6086999.cl"},
	{0x14210AE82DC57D60ULL, "vxdr_primevalentrance_1a2_0_0.vlm"},
	{0x1421B61F952F6615ULL, "812944964.ogg"},
	{0x1421C3694D52D696ULL, "mrl310_merrill6144358.cl"},
	{0x1421D83A56E52EA0ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p.anb"},
	{0x1421F30F36520C02ULL, "vxw_encounter3_8i_0_0.vlm"},
	{0x14222E97D6B2BE66ULL, "6118032_f.fxe"},
	{0x14224DF8E0F5F715ULL, "vxlt_0_lowtown_fiy_0_0.vlm"},
	{0x14232F6F26DC433DULL, "vxht_keep_63o_0_0.vlm"},
	{0x14233233E9E58B76ULL, "vxlt_0_lowtown_a2_n_f1u_0_0.vlm"},
	{0x142383C605BF6410ULL, "lt_0_lowtown_docks_101_0.msh"},
	{0x1423890EBFAA2C8DULL, "6048742_m.fxe"},
	{0x14239158FE955E0AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\runforward.gad"},
	{0x142391C605BF7C2AULL, "lt_0_lowtown_docks_101_0.mmh"},
	{0x142395C605BF8361ULL, "lt_0_lowtown_docks_101_0.mao"},
	{0x1423AF8570CE95BAULL, "vxht_0_hightown_a3_c85_0_0.vlm"},
	{0x1423BB0C6D16A1C5ULL, "sib131_slaver_lord6165211.cl"},
	{0x14240DFE62180E88ULL, "331092541.ogg"},
	{0x1424DB075363AFB4ULL, "vxlt_gallowsprison_n_5rn_0_0.vlm"},
	{0x1424E486A78415EDULL, "llac_w_encounter2_camp_196609_0.dds"},
	{0x142563EB56702692ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdga_0t.dds"},
	{0x14258D89B04675C7ULL, "follower_banter6165424.cl"},
	{0x1425C75897358BD2ULL, "star.mao"},
	{0x1425D8E0CD1E810CULL, "d15_b_shot2.dds"},
	{0x1425DF85DF8E07B0ULL, "vxlt_gallowscourtyard_729_0_0.vlm"},
	{0x1425F4D771803DC5ULL, "vxlt_undercityserialkiller_1m7_0_0.vlm"},
	{0x14263F761722400AULL, "159134166.ogg"},
	{0x1426532E66125C0CULL, "849206139.ogg"},
	{0x142669F1030CCBBCULL, "37020626.ogg"},
	{0x14268E8955FEF4C1ULL, "aip_coteriealchemist_base.gda"},
	{0x1426D1092D12A73AULL, "6207405_m.fxe"},
	{0x1426D63EE3EA7909ULL, "static_rendertodepth_p0.psh.fxo.dx11.assembly"},
	{0x1427701E47BD0886ULL, "6129888_m.fxe"},
	{0x1428172F77F144FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_trn_lft_180.gad"},
	{0x14285AEE652EAAEDULL, "438798853.ogg"},
	{0x14286BCF8E5E3BD6ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_c.impa.ani"},
	{0x1428A08041D9C8B8ULL, "vxlt_undercitydungeon_1o5_0_0.vlm"},
	{0x142A2E1E9CE2D828ULL, "vxlt_0_lowtown_n_en1_0_0.vlm"},
	{0x142A7458FE9B4BA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\runforward.evt"},
	{0x142AD674F6E3FFBFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupleft.msh"},
	{0x142ADC74F6E40A61ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupleft.mmh"},
	{0x142AE874F6E41E02ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupleft.mao"},
	{0x142B408512F3607FULL, "6136305_f.fxe"},
	{0x142B6AF0FB0C27C5ULL, "vfx_bugs_flies01.dds"},
	{0x142B6C6FB4EDBDBCULL, "w_woundedcoast_134_0.mao"},
	{0x142B786FB4EDD25FULL, "w_woundedcoast_134_0.mmh"},
	{0x142B7E6FB4EDDC01ULL, "w_woundedcoast_134_0.msh"},
	{0x142BDB0513E246BEULL, "vxlt_backalley_n_1hz_0_0.vlm"},
	{0x142BECBA19E89E70ULL, "780122527.ogg"},
	{0x142C137A729EB62BULL, "vxw_encounter2_ry_0_0.vlm"},
	{0x142C20B9E1E66C1CULL, "mer221stg_sabin.stg"},
	{0x142C42572335612CULL, "follower_banter6173145.cl"},
	{0x142C993F5B51401DULL, "mcr310_loyalist_templars.bnk"},
	{0x142CFF6F215BC2E5ULL, "vxlt_hangedman_n_tk_0_0.vlm"},
	{0x142D363DFD92631EULL, "gen_im_wep_rog_arc_act1_02.uti"},
	{0x142D7763A714A455ULL, "vxw_bonepitdestroyed_a3_1j0_0_0.vlm"},
	{0x142D875A4861C286ULL, "vxlt_0_lowtown_a3_n_c_hnq_0_0.vlm"},
	{0x142DAB1D6C4E86E4ULL, "vxlt_warehouse_n_am_0_0.vlm"},
	{0x142DF9E1B3E29749ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_act_c.impa.ani"},
	{0x142E0C2F77F59DECULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_trn_lft_180.evt"},
	{0x142E5F555A2C66B6ULL, "313973538.ogg"},
	{0x142E61D932CB4DDFULL, "392898071.ogg"},
	{0x142EE05BCBC5E54BULL, "6113286_m.fxe"},
	{0x142EFB5BAD8AC91EULL, "hlac_w_woundedcoast_-262142_0y.dds"},
	{0x142F30AFE92C0044ULL, "6093462_m.fxe"},
	{0x142F4DB504D4FD2BULL, "ht_0_hightown_a3_45_0_lc.mmh"},
	{0x142FDF762DD503CEULL, "vxw_encounter3_dy_0_0.vlm"},
	{0x142FE8DC8F3485A8ULL, "hlac_w_sundermount_393212_0y.dds"},
	{0x14303269344168E1ULL, "699905926.ogg"},
	{0x143064EC28482482ULL, "vxw_encounter3_5j_0_0.vlm"},
	{0x14307B3F5B54F625ULL, "mcr310_loyalist_templars.cnv"},
	{0x14307B4402686C3FULL, "hlac_w_encounter3_-393211_0z.dds"},
	{0x1430BD2B5C220061ULL, "lt_0_lowtown_docks_a3_n_33_0_ll.mmh"},
	{0x1430DD21899FCD5DULL, "gencr_talvashoth_archer.utc"},
	{0x14312754DF3735B0ULL, "vxlt_gallowstemplar_n_6t7_0_0.vlm"},
	{0x14313B92B283D8E0ULL, "zz_vrc_debug6200237.cl"},
	{0x143152B88637E3BAULL, "w_encounter3_245_0.mmh"},
	{0x143156B88637EAF1ULL, "w_encounter3_245_0.mao"},
	{0x143164B886380200ULL, "w_encounter3_245_0.msh"},
	{0x1431E752F776D0DEULL, "hlac_lt_darktown_-1_45y.dds"},
	{0x14320E6C1EE95570ULL, "lt_0_lowtown_docks_n_1_0_lc.mmh"},
	{0x14332C33DB6FFDE6ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_4x.dds"},
	{0x14332E624B2692E1ULL, "434312705.ogg"},
	{0x14335C7CA896AC4DULL, "art\\vfx\\combat\\vfx_r_ibl_appear_imp_p.mmh"},
	{0x143414E1B3E730CBULL, "art\\vfx\\combat\\vfx_w_def_ele_def_act_c.impa.gad"},
	{0x143435AA430DA8BBULL, "734627440.ogg"},
	{0x14345B65A5E294B8ULL, "6051280_m.fxe"},
	{0x1434C9AACFC178EBULL, "vxht_0_hightown_c1n_0_0.vlm"},
	{0x1434FC67868B7BDCULL, "718977746.ogg"},
	{0x143560F8E949CC79ULL, "vxdr_primevaltomb_7zr_0_2.vlm"},
	{0x143573CF8E68D3C7ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_c.impa.evt"},
	{0x1436699D611C35B0ULL, "2246486.ogg"},
	{0x1436B338E2B25657ULL, "narrative_banter6165527.cl"},
	{0x1436CAF2AF34FAE6ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.mmh"},
	{0x14370291C5EC107DULL, "6076087_f.fxe"},
	{0x1437126BF846BD63ULL, "6095976_f.fxe"},
	{0x143715F004E465F2ULL, "6074370_f.fxe"},
	{0x14376C58FEA5C8F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\runforward.ani"},
	{0x14377A85E5533CB9ULL, "l_one100ar_ironwood.bnk"},
	{0x1437D5E9136E2718ULL, "189361909.ogg"},
	{0x1437F14992838DC9ULL, "llac_lt_warehouse_0_0.dds"},
	{0x143808E7A11EB21EULL, "gen00fl_merrill6145213.cl"},
	{0x14381814C9EA0EA0ULL, "482444631.ogg"},
	{0x14385D3274B61C3DULL, "77110775.ogg"},
	{0x1438761D366EB66FULL, "6126545_m.fxe"},
	{0x14390F741D93A64EULL, "hlac_lt_foundry_n_320_18z.dds"},
	{0x143928130F1F16D4ULL, "hlac_ca_caverns1_24426_9x.dds"},
	{0x1439662D9AC3D06CULL, "6065892_m.fxe"},
	{0x1439A696FF1D68D8ULL, "513423547.ogg"},
	{0x1439D84E947F6BFFULL, "vxht_mansion2_31r_0_0.vlm"},
	{0x143B11E1B3ED4A30ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_act_c.impa.evt"},
	{0x143B5FBE481B74B4ULL, "vxlt_hangedman_1bh_0_0.vlm"},
	{0x143B66737D3031E9ULL, "6135269_f.fxe"},
	{0x143B714C1351DB65ULL, "vxlt_hangedman_n_sj_0_0.vlm"},
	{0x143BE8F2B5D08F3FULL, "6110041_m.fxe"},
	{0x143BF42F7801B245ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_trn_lft_180.ani"},
	{0x143BF9373A57DB06ULL, "vxlt_0_lowtown_a3_n_c_hkr_0_0.vlm"},
	{0x143C1960D5CDF470ULL, "plt_mrl300pt_n_firsttalk.nss"},
	{0x143C4134F705A135ULL, "hlac_ht_0_hightown_n_-131069_0z.dds"},
	{0x143C94CF8E6F2A10ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_c.impa.gad"},
	{0x143CB25A9404B8CEULL, "vxlt_0_lowtown_a2_e8f_0_0.vlm"},
	{0x143D1F40C6E133B9ULL, "6043739_f.fxe"},
	{0x143D9E5962EBF699ULL, "gen00fl_isabela6135664.cl"},
	{0x143E3929F8CDC268ULL, "189114019.ogg"},
	{0x143E3B4A2332DB49ULL, "vxca_caverns1_mines_eb2_0_0.vlm"},
	{0x143E51531FCB1C4EULL, "vxw_encounter3_az_0_0.vlm"},
	{0x143E7DB8F24C8259ULL, "vxlt_undercityencounter_13a_0_0.vlm"},
	{0x143EA158047400F2ULL, "6131176_f.fxe"},
	{0x143EA704C892BC5EULL, "6116129_m.fxe"},
	{0x143F2A5B06F7B9CEULL, "vxca_caverns1_in0_0_0.vlm"},
	{0x143F2E7D5132CCE8ULL, "411948953.ogg"},
	{0x143FE2408F400355ULL, "plt_lgt360pt_nt_rescue.nss"},
	{0x144013290AC8743DULL, "vxht_keep_69m_0_0.vlm"},
	{0x1440933F54BAC6BAULL, "vxht_0_hightown_a3_c67_0_0.vlm"},
	{0x1440D4870FE1632BULL, "follower_banter6136701.cl"},
	{0x1441140C09B14F05ULL, "demoral_02.mao"},
	{0x14411EB600F37A93ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.anb"},
	{0x1441326FADBCA17EULL, "zz_vault_debug6203338.cl"},
	{0x14417ABC5D5B55B9ULL, "hlac_ca_caverns2prc_5802_20y.dds"},
	{0x14418A8AD38749BEULL, "gen00fl_anders6142416.cl"},
	{0x144197DE25C68764ULL, "448273938.ogg"},
	{0x1441E815C995E8BCULL, "6071798_m.fxe"},
	{0x1441E926713AD6C0ULL, "hlac_lt_0_lowtown_a2_n_131070_1y.dds"},
	{0x14427B089DA11356ULL, "257976785.ogg"},
	{0x1442A437F7E14AE4ULL, "856363502.ogg"},
	{0x1442D891556C6EC5ULL, "vxlt_undercityserialkiller_1g5_0_0.vlm"},
	{0x1442FAF728F50921ULL, "mag100cr_templar3.utc"},
	{0x1443113695678A11ULL, "one280_emrys6104258.cl"},
	{0x14433B87C1B7916BULL, "vxht_0_hightown_c4a_0_0.vlm"},
	{0x1443E3D81B000053ULL, "326461621.ogg"},
	{0x1444A50D523FC30BULL, "vxht_0_hightown_a2_c3e_0_1.vlm"},
	{0x1444B5A8FD6BFD79ULL, "hf_bethany_h2s2.mor"},
	{0x1445C9643C207E80ULL, "vxlt_0_lowtown_docks_n_9rd_0_0.vlm"},
	{0x1445F4B37AD39CE0ULL, "mcr310_argument6072689.cl"},
	{0x1445F9DEF359CEDBULL, "792784509.ogg"},
	{0x1446C3528A1BCF24ULL, "6112295_m.fxe"},
	{0x1446E8C444757914ULL, "vxlt_hangedman_n_18q_0_0.vlm"},
	{0x144748EC16625D1BULL, "181313815.ogg"},
	{0x144779E1EFDB6E19ULL, "store_one200cr_ilen.utm"},
	{0x1447B728A6B0200BULL, "vxht_0_hightown_a3_b9z_0_0.vlm"},
	{0x1447ECDC1572BFA3ULL, "6148714_m.fxe"},
	{0x14481D21CEB70A5CULL, "root_mage_spirithealer.dds"},
	{0x1448390B0839A1D5ULL, "vxdr_primevaltomb_8f3_0_0.vlm"},
	{0x14486234879FF505ULL, "vxlt_0_lowtown_a3_n_c_csz_0_0.vlm"},
	{0x1449BE725A3FA88AULL, "6170316_m.fxe"},
	{0x1449C2DD51FDEA7CULL, "vxlt_gallowscourtyard_n_4lk_0_0.vlm"},
	{0x144A1B555A331D6BULL, "dae000ip_exit_player_house.utp"},
	{0x144A2B9683419BE6ULL, "6147704_f.fxe"},
	{0x144ABD78583B261EULL, "153638543.ogg"},
	{0x144B0DF81E70D207ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexwild_d.dds"},
	{0x144B56EBC88C3493ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_loop_c.impa.ani"},
	{0x144BA9AD806DF202ULL, "vxlt_0_lowtown_a2_erh_0_0.vlm"},
	{0x144BC77F3F40F0DCULL, "pro000pt_main.ncs"},
	{0x144C10AAFB12C4EEULL, "171155833.ogg"},
	{0x144C62CD3C78F64DULL, "6047252_f.fxe"},
	{0x144CAD271528F3C9ULL, "vxca_caverns1_mines_eg1_0_0.vlm"},
	{0x144CB630204B924BULL, "6141278_m.fxe"},
	{0x144CBA2BD4279E6AULL, "hlac_lt_0_lowtown_a2_131070_3z.dds"},
	{0x144CC13362DDB65CULL, "498928543.ogg"},
	{0x144D3BB6C6B835B2ULL, "llac_lt_0_lowtown_docks_a3_1_0.dds"},
	{0x144D3E427F40FF47ULL, "gen00fl_dog6107718.cl"},
	{0x144D7E3B0CF6B5F2ULL, "223234611.ogg"},
	{0x144DCC6D413F9990ULL, "dae200_bran6144586.cl"},
	{0x144E3BB600FE36FCULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.mmh"},
	{0x144E8D512F1F5D60ULL, "gen00fl_fenris6152646.cl"},
	{0x144F27943C66DF34ULL, "6135681_m.fxe"},
	{0x144FC3AD6CE76899ULL, "hlac_w_sundermount_-327678_0x.dds"},
	{0x1450025EADBEAB58ULL, "6089813_f.fxe"},
	{0x1450058F9981EF6BULL, "hlac_lt_0_lowtown_a3_n_c_0_6x.dds"},
	{0x1450176371D5018CULL, "hlac_lt_undercitydungeon_-1_49z.dds"},
	{0x1450291EE0F7801CULL, "734870179.ogg"},
	{0x14505EA97AAFCD57ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he2a.phy"},
	{0x145075548CC8E601ULL, "6171932_m.fxe"},
	{0x145076413A2B95E0ULL, "plt_gen00pt_party.plo"},
	{0x14507804BA987E5EULL, "mrl150_merrill_pt1.stg"},
	{0x1450BB4BEE477F95ULL, "vxht_mansion_n_5b4_0_0.vlm"},
	{0x14511494574A28BEULL, "701054463.ogg"},
	{0x145118D564F4FAACULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\blendtree\\templarseeker_override.anb"},
	{0x1451B5D81A5AE38DULL, "273899289.ogg"},
	{0x1451FD40BFCC43C3ULL, "vxlt_gallowscourtyard_n_6rh_0_0.vlm"},
	{0x14520F600EF22211ULL, "hlac_ht_chantry_n_2534_33y.dds"},
	{0x14528D7DD3EE4720ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basic.phy"},
	{0x145421CBCA516637ULL, "vxlt_warehouse_93_0_0.vlm"},
	{0x14545F124FD345FFULL, "vxlt_warehouse_qc_0_0.vlm"},
	{0x145467003B922EB5ULL, "vxw_woundedcoast_1g8_0_0.vlm"},
	{0x1454809FF354E465ULL, "112330234.ogg"},
	{0x1454AD654A67388DULL, "1007993044.ogg"},
	{0x1454C3993E850001ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_maj_c.impa.ani"},
	{0x1454D067055F225FULL, "vxht_mansion2_1wf_0_0.vlm"},
	{0x1455111C5316FFD9ULL, "qun160_vashothlead6053893.cl"},
	{0x145519C63EF52FE7ULL, "gen00fl_anders6107694.cl"},
	{0x1455377BFA85E9C6ULL, "gen_im_npc_arm_hum_boo_tmpc.uti"},
	{0x1455F728834293E6ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopeningmage.ani"},
	{0x1456356D70837022ULL, "6117445_m.fxe"},
	{0x1456419358FB5544ULL, "w_encounter1_mmi.gff"},
	{0x14564F9E2D5355C5ULL, "mcr371_hostile_fenris6082232.cl"},
	{0x1456605EE44DD2DFULL, "vxht_0_hightown_a3_61z_0_0.vlm"},
	{0x1456712FE190E32FULL, "vxw_encounter3_71_0_0.vlm"},
	{0x1456A7AFEC5FF7D7ULL, "mag370_meredith6064971.cl"},
	{0x1456DAAACAD279F4ULL, "vxht_0_hightown_n_c74_0_0.vlm"},
	{0x1456DD74F70946DEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupleft.phy"},
	{0x14594EEBC89864CAULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_loop_c.impa.evt"},
	{0x1459829B03B7D171ULL, "6112033_f.fxe"},
	{0x14598E26A327375CULL, "vxht_mansion2_20w_0_0.vlm"},
	{0x1459A22D7E5678D7ULL, "mcr310_party_banter6087804.cl"},
	{0x1459B5F2EB7F8644ULL, "640628499.ogg"},
	{0x1459BB454B82B9D4ULL, "lt_0_lowtown_docks_n_118_0.phy"},
	{0x1459F32BB1134173ULL, "vxdr_primevaltomb_6iv_0_0.vlm"},
	{0x1459F3F2B01483E5ULL, "31892206.ogg"},
	{0x145A2D9C3C266183ULL, "hlac_ht_mansion2b_818_55y.dds"},
	{0x145AC0EEE401E4A3ULL, "vxlt_darktown_2qh_0_0.vlm"},
	{0x145ACE993E897EA3ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_maj_c.impa.gad"},
	{0x145AFD1DA5FA5BEEULL, "177487769.ogg"},
	{0x145B5FEF79E56791ULL, "pro000_merchant6051579.cl"},
	{0x145B763889A46E71ULL, "6153019_m.fxe"},
	{0x145B9916DD620ADAULL, "vxw_bonepitdestroyed_a3_224_0_0.vlm"},
	{0x145BA369325AC764ULL, "892491335.ogg"},
	{0x145BB0E618550EDEULL, "6085324_m.fxe"},
	{0x145BC67C1E099F7BULL, "gen00fl_dog6145226.cl"},
	{0x145BCE9CB08FB571ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_runf.ani"},
	{0x145BE71FF3174136ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wc_n.dds"},
	{0x145C41EF6ADB3599ULL, "1030333705.ogg"},
	{0x145CE747E5532604ULL, "vxht_keep_706_0_0.vlm"},
	{0x145DA06F2D020D47ULL, "6007723_m.fxe"},
	{0x145DE5044D46E400ULL, "552163839.ogg"},
	{0x145E3BE12F485E6BULL, "qun231_varnell6099594.cl"},
	{0x145ED9C0AF8D2365ULL, "418518431.ogg"},
	{0x145EE451DFD3E7A4ULL, "vxlt_backalley_g_1wd_0_0.vlm"},
	{0x145F466A09D7F436ULL, "vxw_encounter1_al_0_0.vlm"},
	{0x145FDF9071A56346ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robb_n.dds"},
	{0x14602DEBC89E4B3DULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_loop_c.impa.gad"},
	{0x1460F551ABDB231AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_ltintdub.phy"},
	{0x1461049B54991C4CULL, "one110ip_ggt_warehouse.utp"},
	{0x14617A5EF7F2AB1DULL, "gen00fl_aveline6139210.cl"},
	{0x14618975338E177FULL, "hlac_lt_0_lowtown_a2_-65539_0y.dds"},
	{0x14619B5EABD506DAULL, "vxlt_0_lowtown_a2_f16_0_0.vlm"},
	{0x1461AB993E8F6188ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_maj_c.impa.evt"},
	{0x1461B4A0F4368D44ULL, "6028819_m.fxe"},
	{0x1461C583AD91A7B3ULL, "vxw_encounter3_ff_0_0.vlm"},
	{0x1461E1408F5CE7B6ULL, "plt_lgt360pt_nt_rescue.plo"},
	{0x146205EAE7F8D807ULL, "6136319_m.fxe"},
	{0x14625F7E1C02A455ULL, "694708445.ogg"},
	{0x1462D0EF41C95E7FULL, "vxlt_warehouse_zb_0_0.vlm"},
	{0x1462D99CB095E613ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_runf.gad"},
	{0x146322B469938A7EULL, "6041401_m.fxe"},
	{0x14632BA52C7F988AULL, "character_character_p133.psh.fxo"},
	{0x14638C830B4C74B2ULL, "boulder_mac_02.mao"},
	{0x14640A0FB991A1C2ULL, "w_encounter3_45_0.msh"},
	{0x14640C0FB991A578ULL, "w_encounter3_45_0.mmh"},
	{0x1464180FB991B91BULL, "w_encounter3_45_0.mao"},
	{0x146457149A715FDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_collapse_help_up_malepc.ani"},
	{0x1464877B252F1489ULL, "vxlt_foundry_n_w4_0_0.vlm"},
	{0x1464959628F030B5ULL, "vxlt_darktown_59y_0_0.vlm"},
	{0x146498454B8C85D1ULL, "lt_0_lowtown_docks_n_118_0.msh"},
	{0x1464A6454B8C9DECULL, "lt_0_lowtown_docks_n_118_0.mao"},
	{0x1464B2454B8CB18FULL, "lt_0_lowtown_docks_n_118_0.mmh"},
	{0x146534610646C5FDULL, "ave260cr_bandit.utc"},
	{0x14654E6420080940ULL, "6166901_f.fxe"},
	{0x146589F40BEFF34FULL, "68216972.ogg"},
	{0x1465973DE2C31234ULL, "753696363.ogg"},
	{0x1465CF88A8FBED4FULL, "6128327_m.fxe"},
	{0x1465E4A07090A1CBULL, "hlac_ht_keep_3732_10x.dds"},
	{0x1465FA299A5F4766ULL, "attack_pressed.cur"},
	{0x1466003690BE2651ULL, "vxlt_gallowstemplar_3n2_0_0.vlm"},
	{0x146624A73FA848D3ULL, "6166611_m.fxe"},
	{0x146642BAF98275ACULL, "vxlt_undercityserialkiller_141_0_0.vlm"},
	{0x1466C3833D5EC039ULL, "mag100_thrask6032057.cl"},
	{0x1466E7020432D2DCULL, "183324492.ogg"},
	{0x1467A0037FE0C328ULL, "43009103.ogg"},
	{0x1467A1BBBD61FAD0ULL, "follower_banter6160736.cl"},
	{0x1467A5CB1870D48CULL, "6075674_f.fxe"},
	{0x1467B81620FE2B72ULL, "vxht_mansion_n_60r_0_0.vlm"},
	{0x1467CE399BE07D4AULL, "fxe_water_v_sp.dds"},
	{0x14680C0FB9958D8BULL, "w_encounter3_45_0.rml"},
	{0x1468471BC95548E7ULL, "6104890_f.fxe"},
	{0x14686D16418CF8DEULL, "vxw_bonepitmining_a2_2ak_0_0.vlm"},
	{0x14688D67645A2302ULL, "vxlt_0_lowtown_a2_exj_0_0.vlm"},
	{0x1468C842EF888999ULL, "vxlt_gallowscourtyard_n_74i_0_0.vlm"},
	{0x1468FD117294CB4EULL, "164532151.ogg"},
	{0x1469058A6F144FC7ULL, "vxlt_0_lowtown_a2_n_c_gwy_0_0.vlm"},
	{0x146906A5D99153A9ULL, "409402262.ogg"},
	{0x14692E4890AF6E73ULL, "142921860.ogg"},
	{0x1469502129C23AB2ULL, "dae000_brothel_gamlen.cnv"},
	{0x1469C2E2728CE63FULL, "6071520_m.fxe"},
	{0x1469D69CB09C0078ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_runf.evt"},
	{0x146A64149A75E11EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_collapse_help_up_malepc.gad"},
	{0x146AE0288354C8E0ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopeningmage.gad"},
	{0x146AE04059E7A695ULL, "vxlt_hangedman_1q4_0_0.vlm"},
	{0x146B091C38214B9FULL, "74154529.ogg"},
	{0x146B5F9C23A25F5DULL, "1028825308.ogg"},
	{0x146C722129C4AA0AULL, "dae000_brothel_gamlen.bnk"},
	{0x146C96A712B92A13ULL, "qun120_arishok6051491.cl"},
	{0x146D50FC28E08034ULL, "vxw_encounter2_camp_5c_0_0.vlm"},
	{0x146D5E11982F81CDULL, "364024855.ogg"},
	{0x146DB846FBCE0CB6ULL, "vxw_encounter1_fm_0_0.vlm"},
	{0x146DC390053AABEFULL, "mag210_arianni6024103.cl"},
	{0x146DEA8358B40F25ULL, "hlac_w_bonepitmining_a2_196611_0y.dds"},
	{0x146E0E2D2D6211B2ULL, "qcr220_grey_wardens6092033.cl"},
	{0x146E3B3BE8AC3A93ULL, "follower_banter6161297.cl"},
	{0x146E3C91FEC852DBULL, "vxlt_darktown_50p_0_0.vlm"},
	{0x146F59E8679A048BULL, "zz_and_debug6134168.cl"},
	{0x146F7D6C7AD5F98EULL, "335897690.ogg"},
	{0x146FBC2D0872C536ULL, "450931309.ogg"},
	{0x1470085ECDF2AE7DULL, "vxw_woundedcoast_48_0_0.vlm"},
	{0x1470242152DF3661ULL, "520876304.ogg"},
	{0x147037609F87C033ULL, "vxw_encounter3_ci_0_0.vlm"},
	{0x14713F149A7BC13DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_collapse_help_up_malepc.evt"},
	{0x14718BDCF2F08194ULL, "198755203.ogg"},
	{0x1471AA0109B1462AULL, "art\\vfx\\combat\\vfx_a_stealth_act_c.impa.gad"},
	{0x1471DC705B3AD234ULL, "qun231_varnell6078091.cl"},
	{0x14720A97303AB802ULL, "lt_0_lowtown_a2_51_0.rml"},
	{0x1472124D6631F933ULL, "vxht_keep_6br_0_0.vlm"},
	{0x1472128CA003CE4AULL, "vxca_caverns2prc_8d0_0_1.vlm"},
	{0x1472BBA262D24958ULL, "vxht_bartrand_n_221_0_0.vlm"},
	{0x1472CF974B4BEF7DULL, "vxlt_0_lowtown_docks_a3_cxa_0_0.vlm"},
	{0x1472D1C6A1CECF13ULL, "vxw_bonepitdestroyed_a3_17q_0_0.vlm"},
	{0x1472D842CF4865DDULL, "6135625_f.fxe"},
	{0x1472F631956499E5ULL, "6119224_m.fxe"},
	{0x14731CB8DFA0F0EAULL, "mag131_taraneh6094542.cl"},
	{0x147371A71BCAE583ULL, "822748631.ogg"},
	{0x14745A075784BB2AULL, "art\\vfx\\combat\\vfx_glb_ele_s_minor_c.impa.gad"},
	{0x14753539128B2847ULL, "6177706_f.fxe"},
	{0x1475A878E1F73487ULL, "lgt220cr_mag_tmplr_hunt.utc"},
	{0x1476C3391778404CULL, "vxlt_0_lowtown_docks_b5p_0_1.vlm"},
	{0x14773B0D9ED14841ULL, "hlac_w_blightlands_196604_0z.dds"},
	{0x1477984D1841B8CFULL, "vxca_cavern1_haunted_id2_0_0.vlm"},
	{0x1477A2835916FD82ULL, "ser221_quentin6123387.cl"},
	{0x1477AA7DD40D94B5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basic.msh"},
	{0x1477AC7DD40D982BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basic.mmh"},
	{0x1477B87DD40DACC8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basic.mao"},
	{0x147860C49E338553ULL, "mer200_ambush_leader.bnk"},
	{0x14788D0109B733C1ULL, "art\\vfx\\combat\\vfx_a_stealth_act_c.impa.evt"},
	{0x1478AED6E9212737ULL, "ave200_donnic6112160.cl"},
	{0x1478C41CAE92E2F7ULL, "hlac_dr_slavers_3217_4z.dds"},
	{0x1478E747D9DCAC83ULL, "ch1notnt_vl.psh.fxo"},
	{0x14790C32886A7103ULL, "vxlt_0_lowtown_docks_a3_n_cys_0_0.vlm"},
	{0x14794FF3DA5EC3AAULL, "834785063.ogg"},
	{0x1479777EC9F08AD3ULL, "6177841_f.fxe"},
	{0x147999E48E84EA9DULL, "llac_lt_0_lowtown_n_65537_3.dds"},
	{0x1479B6EC08838421ULL, "vxlt_hangedman_21s_0_0.vlm"},
	{0x1479DA1AB3040772ULL, "dae000_templar_knight.cnv"},
	{0x147A7E34FD3F223BULL, "6122128_f.fxe"},
	{0x147AABCCB4A49AB9ULL, "655670893.ogg"},
	{0x147AF1B5F55681DBULL, "vxw_woundedcoast_1f5_0_0.vlm"},
	{0x147B3522F9F62DE7ULL, "zz_followers_banter6157834.cl"},
	{0x147B3D07578AA8C1ULL, "art\\vfx\\combat\\vfx_glb_ele_s_minor_c.impa.evt"},
	{0x147B59B76D137D06ULL, "6092492_m.fxe"},
	{0x147B82C49E35F4A3ULL, "mer200_ambush_leader.cnv"},
	{0x147BC80BC3C018A4ULL, "vxlt_backalley_g_1yj_0_0.vlm"},
	{0x147BED44E3087DCDULL, "hlac_lt_darktown_-1_65x.dds"},
	{0x147C2A23EDC42536ULL, "vxw_encounter1_kn_0_0.vlm"},
	{0x147C885F40BE7BF7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbig.mmh"},
	{0x147C8C5F40BE8204ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbig.mao"},
	{0x147C9E5F40BEA169ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbig.msh"},
	{0x147CC2CD3330A88BULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_enter.evt"},
	{0x147CFC1AB30676CAULL, "dae000_templar_knight.bnk"},
	{0x147D65085CF180F9ULL, "865522809.ogg"},
	{0x147E7A3BBFE8C6FDULL, "vxw_woundedcoast_17_0_0.vlm"},
	{0x147FF638EF88B61DULL, "hlac_lt_gallowstemplar_7198_4x.dds"},
	{0x147FFECD4F7CFEEEULL, "6081672_m.fxe"},
	{0x1480EEC390204E39ULL, "gen00fl_fenris6126699.cl"},
	{0x14810EF0C7F42F14ULL, "fen110_fenris6097041.cl"},
	{0x148143A393C4E793ULL, "vxw_bonepitdestroyed_a3_18p_0_0.vlm"},
	{0x1481680E875AB265ULL, "6114524_m.fxe"},
	{0x14817222A9D7C1DFULL, "6078102_f.fxe"},
	{0x1481D1BC807E30A0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_hightown_fabric_n.dds"},
	{0x14822204F601F68DULL, "vxlt_warehouse_n_113_0_0.vlm"},
	{0x14823999AD76B9E5ULL, "vxht_mansion_n_5bd_0_0.vlm"},
	{0x14823B2A402C24D4ULL, "vxlt_gallowstemplar_5o6_0_0.vlm"},
	{0x1482D8ED991E2456ULL, "204645513.ogg"},
	{0x1482FEE7B8F3C8FBULL, "789997756.ogg"},
	{0x14830176984468B0ULL, "vxht_0_hightown_a3_bam_0_0.vlm"},
	{0x14832BBCA8D73357ULL, "hlac_w_sundermount_-65540_0y.dds"},
	{0x14833B008BD8BAB1ULL, "mag291st_anders.stg"},
	{0x1483437F140E1BF7ULL, "6086395_m.fxe"},
	{0x14835E8350799DFFULL, "hlac_w_encounter3_-262142_0z.dds"},
	{0x1483732B3EE58C63ULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_ex.evt"},
	{0x1483DBCD3336F24CULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_enter.gad"},
	{0x14842BBE10E2971CULL, "hlac_lt_0_lowtown_a2_n_0_3y.dds"},
	{0x14843701E4B7F149ULL, "461590035.ogg"},
	{0x14849805E8A6F1BEULL, "zz_qun_debug6028890.cl"},
	{0x1484CD73745EB66EULL, "vxlt_0_lowtown_a2_gmk_0_0.vlm"},
	{0x14853D33FF07431CULL, "6170008_m.fxe"},
	{0x14857F401530005AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_ground.anb"},
	{0x1485850109C1B010ULL, "art\\vfx\\combat\\vfx_a_stealth_act_c.impa.ani"},
	{0x1485862A862303F5ULL, "vxlt_0_lowtown_n_e0r_0_0.vlm"},
	{0x1485937A572F5B7BULL, "508128890.ogg"},
	{0x1485EF01C93AACC9ULL, "vxlt_0_lowtown_n_csi_0_0.vlm"},
	{0x1485F8A714D26FA1ULL, "309914571.ogg"},
	{0x148672EC53528DA5ULL, "6139382_f.fxe"},
	{0x14867B5D49AC3CB3ULL, "185169153.ogg"},
	{0x1486B10019FCEAE8ULL, "6068510_m.fxe"},
	{0x14876FECD4334409ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containera.phy"},
	{0x148782CEE4D4CF16ULL, "vxw_bonepitmining_a2_1va_0_0.vlm"},
	{0x1487BAB50A6BD3A8ULL, "6122906_m.fxe"},
	{0x1487C57E4170C319ULL, "vxlt_0_lowtown_a2_epd_0_1.vlm"},
	{0x1487FEA334256E37ULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_hed_gib_0s.dds"},
	{0x1488350757952510ULL, "art\\vfx\\combat\\vfx_glb_ele_s_minor_c.impa.ani"},
	{0x14884FE8CDE98480ULL, "vxht_0_hightown_a3_alq_0_0.vlm"},
	{0x1488A7285A3390FBULL, "dae000_merchant_hm_mob6174227.cl"},
	{0x1488BDA9990914FCULL, "lt_0_lowtown_a2_n_c_27_0_ll.mmh"},
	{0x1488BDBF9EB2D32FULL, "genip_trap_fire_jet_long.utp"},
	{0x1488E64200A36AA6ULL, "qun221cr_petrice.utc"},
	{0x1489472770A01573ULL, "vxlt_undercityserialkiller_wt_0_0.vlm"},
	{0x14896B1DE338C46BULL, "isa311_isabela6095634.cl"},
	{0x148994F7E5477126ULL, "mag111st_cassandra1.stg"},
	{0x148A4C2B3EEB68E4ULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_ex.gad"},
	{0x148ACBCD7B7133C3ULL, "626478095.ogg"},
	{0x148ADACD333D0F4AULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_enter.ani"},
	{0x148B09F2D576AC63ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcmale_fenrisrom.ani"},
	{0x148B7397B55FDE3BULL, "1053555292.ogg"},
	{0x148BA155C159999DULL, "6060649_m.fxe"},
	{0x148BF5C046FC3E34ULL, "766084524.ogg"},
	{0x148C18BE88529953ULL, "hlac_dr_primevalentrance_1724_1x.dds"},
	{0x148D0753055B8136ULL, "577351730.ogg"},
	{0x148D1DDED1257CC4ULL, "6082218_f.fxe"},
	{0x148D2F77608C0412ULL, "lt_gallowstemplar_2394.rml"},
	{0x148D403984776157ULL, "hla_ht_0_hightown_a2_65537_0y.dds"},
	{0x148D60D6E220074FULL, "hlac_w_bonepitmining_a2_327682_0y.dds"},
	{0x148D7444C70E7F78ULL, "llac_lt_0_lowtown_a2_-65535_0.dds"},
	{0x148D816FB540ABA4ULL, "w_woundedcoast_134_0.phy"},
	{0x148DD79FEC027A67ULL, "6149880_f.fxe"},
	{0x148E279FC3B87B3BULL, "6149021_m.fxe"},
	{0x148E6B793BA0F9ABULL, "vxht_brothel_n_6ik_0_0.vlm"},
	{0x148E8E7F74FCC26FULL, "6176749_f.fxe"},
	{0x148E96CC33EE0ECDULL, "vxlt_0_lowtown_a2_n_c_hl9_0_0.vlm"},
	{0x148EC94FC88F85C9ULL, "hlac_w_bonepitdestroyed_a3_458755_0y.dds"},
	{0x148ECE254252034EULL, "1019839838.ogg"},
	{0x148EF6074A1E2A33ULL, "vxht_keep_6ht_0_0.vlm"},
	{0x148F43F8415DE2ABULL, "hlac_ca_caverns1_24424_16y.dds"},
	{0x148F7FB3B8C0540DULL, "vxca_cavern1_haunted_gkn_0_1.vlm"},
	{0x1490555D15E8E767ULL, "vxdr_primevaltomb_56x_0_0.vlm"},
	{0x14906AC42A932373ULL, "ico_longbow_4.dds"},
	{0x149093E1E7F80F0DULL, "vxlt_warehouse_n_162_0_0.vlm"},
	{0x149100329E0CCE9BULL, "6058877_m.fxe"},
	{0x14910329FB1F90C1ULL, "hlac_ca_caverns2prc_5802_36z.dds"},
	{0x14915B2B3EF1A0C2ULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_ex.ani"},
	{0x14915C3945DE7940ULL, "qun160_party_comment_26065000.cl"},
	{0x1492772C663A319BULL, "llac_ca_caverns1_24424_24.dds"},
	{0x1492AD90F3FA15ECULL, "56523850.ogg"},
	{0x1492CA18CC8F6C3BULL, "hlac_lt_warehouse_65537_4x.dds"},
	{0x1492FECC59E29A34ULL, "vxca_caverns1_j2a_0_0.vlm"},
	{0x14930F9E87275A50ULL, "narrative_flemeth6002802.cl"},
	{0x149343429740EE75ULL, "6109051_f.fxe"},
	{0x14937CFF3572FE03ULL, "vrc321_spooky_scenes6215550.cl"},
	{0x1493E2FE3A3C6C82ULL, "vxlt_0_lowtown_a2_eiu_0_0.vlm"},
	{0x14940D54ED2B76EFULL, "vxdr_primevaltomb_3i3_0_0.vlm"},
	{0x1494435F82ABE7CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_enter.gad"},
	{0x1494CD57F2C0EBA2ULL, "710445962.ogg"},
	{0x14950CCE1742F33BULL, "vxht_bartrand_n_67e_0_0.vlm"},
	{0x1495A6BA44CEA580ULL, "mag100_cullen6095127.cl"},
	{0x1495C7E3A3EA9E79ULL, "vxlt_0_lowtown_a3_n_c_fdz_0_0.vlm"},
	{0x1495F4ABD6CAE796ULL, "vxw_bonepitmining_a2_1sb_0_0.vlm"},
	{0x1496373312AFCF16ULL, "llac_dr_0_deeproads_5839_1.dds"},
	{0x149643D2D06A4F82ULL, "vxht_keep_n_6gs_0_0.vlm"},
	{0x149660BD730F6BA1ULL, "6125748_m.fxe"},
	{0x1496EDBEF17DB91EULL, "vxw_bonepitdestroyed_a3_25y_0_0.vlm"},
	{0x1496FEF3A82ED47AULL, "6092360_m.fxe"},
	{0x14976B010E0E59C1ULL, "vxca_caverns1_gi8_0_0.vlm"},
	{0x1497D56FD942B2DBULL, "vxw_woundedcoast_1l7_0_0.vlm"},
	{0x14983FA97AED4E4CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he2a.mmh"},
	{0x149845A97AED583EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he2a.msh"},
	{0x149861B068416F38ULL, "692688346.ogg"},
	{0x1498A5EFA0C67912ULL, "6101064_m.fxe"},
	{0x149930494C95DD5BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_exit.evt"},
	{0x14993B306BF8B34EULL, "vxdr_primevalentrance_3bc_0_0.vlm"},
	{0x14995BA9418583B5ULL, "m_qun100ar_plot_start.bnk"},
	{0x1499C3A9AF99B7D3ULL, "6023880_m.fxe"},
	{0x149A1991881D7EBFULL, "6048435_f.fxe"},
	{0x149A2F0C189D017EULL, "hlac_ht_playershouse_n_855_33y.dds"},
	{0x149A3890141E2ACFULL, "primitives_primitive_p5.psh.fxo.dx11"},
	{0x149B1FF5C27ABD62ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_flat_1.mao"},
	{0x149B585F82B22A7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_enter.evt"},
	{0x149C0782E9B3F92EULL, "vxht_keep_n_5pq_0_0.vlm"},
	{0x149C5E8E1916E0E5ULL, "vxlt_hangedman_1qd_0_0.vlm"},
	{0x149C6C87DF88F4C5ULL, "lt_0_lowtown_docks_a3_mmap_ph.dds"},
	{0x149C81BF839465BBULL, "llac_lt_warehouse_n_0_9.dds"},
	{0x149C997CB5AE93BBULL, "6142121_m.fxe"},
	{0x149CA9976B56783CULL, "6120464_f.fxe"},
	{0x149CBB9AD57D8A7DULL, "6054538_m.fxe"},
	{0x149CCD7861B4A3E7ULL, "art\\vfx\\combat\\vfx_glb_general_imp_c.anb"},
	{0x149CFC072162A504ULL, "861647206.ogg"},
	{0x149D2CCDAC028171ULL, "664712461.ogg"},
	{0x149D9A4B58FFB884ULL, "art\\characters\\playercharacter\\elfmale\\glv_nuda.msh"},
	{0x149D9F015BFB7B20ULL, "gen00flst_fenris_romance_sh.stg"},
	{0x149DAC4B58FFD7EDULL, "art\\characters\\playercharacter\\elfmale\\glv_nuda.mao"},
	{0x149DB04B58FFDE26ULL, "art\\characters\\playercharacter\\elfmale\\glv_nuda.mmh"},
	{0x149DD3A92F967127ULL, "108205726.ogg"},
	{0x149E164C9C269FFDULL, "248658776.ogg"},
	{0x149ED389D252FEB6ULL, "hlac_ca_cavern1_haunted_24421_3x.dds"},
	{0x149EE5C185B7F879ULL, "vxlt_gallowscourtyard_698_0_0.vlm"},
	{0x149F1D539162EAE5ULL, "vxht_mansion_n_5hb_0_0.vlm"},
	{0x149F20F2D5877C4DULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcmale_fenrisrom.gad"},
	{0x149F4064DD855459ULL, "vxlt_0_lowtown_docks_a3_n_bz5_0_0.vlm"},
	{0x149F85E8D43D82F8ULL, "hlac_lt_0_lowtown_a2_-1_0y.dds"},
	{0x14A029494C9BEF9CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_exit.gad"},
	{0x14A04276E0B00727ULL, "6101212_m.fxe"},
	{0x14A055BC02213590ULL, "6149127_m.fxe"},
	{0x14A0AF3C0CCCC032ULL, "332765806.ogg"},
	{0x14A0F317112BF822ULL, "6078126_m.fxe"},
	{0x14A116951BF510FBULL, "dae000_mother6121012.cl"},
	{0x14A1775EB0387BB6ULL, "dae000_merchant_generic_hm6172629.cl"},
	{0x14A18DD7A979F7ACULL, "hlac_lt_0_lowtown_docks_a3_65535_3z.dds"},
	{0x14A1A60DAD137112ULL, "gen00fl_dog6167089.cl"},
	{0x14A1B055146951DAULL, "vxdr_primevaltomb_914_0_0.vlm"},
	{0x14A1B5A02214AE7FULL, "vxca_caverns1_i29_0_0.vlm"},
	{0x14A222FA7800770FULL, "6119734_m.fxe"},
	{0x14A2393983828D13ULL, "hlac_w_blightlands_-3_0z.dds"},
	{0x14A295DD715663E7ULL, "w_encounter2_15_0.rml"},
	{0x14A2F4ADCEDE4485ULL, "711064626.ogg"},
	{0x14A3496276FE3A15ULL, "142594390.ogg"},
	{0x14A355E457356A5EULL, "vxht_brothel_n_3uw_0_0.vlm"},
	{0x14A37085AC151955ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg04a.phy"},
	{0x14A379E8D263A790ULL, "ca_caverns2_mmap_ph.dds"},
	{0x14A37EAB09390BBBULL, "vxht_bartrand_n_60d_0_0.vlm"},
	{0x14A41EEA2F716092ULL, "6095407_f.fxe"},
	{0x14A44765A4B2654EULL, "vxca_caverns1_i37_0_0.vlm"},
	{0x14A4782241978149ULL, "6074368_f.fxe"},
	{0x14A4A7B421C00895ULL, "vxlt_hangedman_1m0_0_0.vlm"},
	{0x14A4CEDFB51C0402ULL, "hlac_w_sundermount_327678_0z.dds"},
	{0x14A4E23F369126D4ULL, "vxht_0_hightown_by0_0_0.vlm"},
	{0x14A4F16CB42BE05AULL, "289883260.ogg"},
	{0x14A4F62F9E36F745ULL, "one101st_charade_cavern.stg"},
	{0x14A576F1DFF69B3CULL, "38705166.ogg"},
	{0x14A58D7DE5077424ULL, "vxlt_undercityserialkiller_22d_0_0.vlm"},
	{0x14A59AA8ABA0D9D0ULL, "6139299_f.fxe"},
	{0x14A5A06E0FC19C22ULL, "ser221_gamlen6134822.cl"},
	{0x14A61CE04D404BF0ULL, "hlac_lt_gallowstemplar_2395_0y.dds"},
	{0x14A66F5A1BBF5E0DULL, "831347805.ogg"},
	{0x14A705D90EB8C05BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_come_0t.dds"},
	{0x14A728494CA20C9AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_exit.ani"},
	{0x14A7A74B590823F1ULL, "art\\characters\\playercharacter\\elfmale\\glv_nuda.phy"},
	{0x14A7D20DF1F65EEEULL, "6089827_m.fxe"},
	{0x14A8605F82BCC195ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_enter.ani"},
	{0x14A890CBB4AA01BCULL, "w_sundermount_8_0.rml"},
	{0x14A89944AC61E945ULL, "644700235.ogg"},
	{0x14A973D31D77D1AEULL, "hlac_lt_0_lowtown_docks_a3_n_131071_4y.dds"},
	{0x14A994DD715C80FFULL, "w_encounter2_15_0.phy"},
	{0x14A9A067D52C0776ULL, "6116693_m.fxe"},
	{0x14A9E6305A67D516ULL, "llac_lt_0_lowtown_n_65533_0.dds"},
	{0x14AAAC97EFF258C8ULL, "52640092.ogg"},
	{0x14AAAE2B5675FC77ULL, "6082618_m.fxe"},
	{0x14AAC6432152FE72ULL, "t3_app.tnt"},
	{0x14AAE3A09FA71575ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_003a_0t.dds"},
	{0x14AB151D71417027ULL, "character_character_p145.psh.fxo.dx11.assembly"},
	{0x14AB25A0382AE2BCULL, "vxlt_gallowscourtyard_n_740_0_0.vlm"},
	{0x14AB30983BA3B765ULL, "6125793_m.fxe"},
	{0x14AB76CBB4AC0B11ULL, "w_sundermount_8_0.msh"},
	{0x14AB84CBB4AC232CULL, "w_sundermount_8_0.mao"},
	{0x14AB90CBB4AC37CFULL, "w_sundermount_8_0.mmh"},
	{0x14ABEC81E3A2CF1AULL, "art\\levels\\races\\design\\prp_tableleg.msh"},
	{0x14ABFA81E3A2E753ULL, "art\\levels\\races\\design\\prp_tableleg.mao"},
	{0x14ABFE81E3A2EE60ULL, "art\\levels\\races\\design\\prp_tableleg.mmh"},
	{0x14AC8A8824E6425CULL, "vxlt_0_lowtown_docks_n_9r8_0_0.vlm"},
	{0x14ACE485059BAB64ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_partycall_0.pwk"},
	{0x14AD2DE8C1B6B342ULL, "6080565_m.fxe"},
	{0x14AD581C479BE612ULL, "6155616_m.fxe"},
	{0x14AF4E752CDE297FULL, "vxw_encounter3_6r_0_0.vlm"},
	{0x14AF745C3E1D8750ULL, "vxht_keep_n_734_0_0.vlm"},
	{0x14B03EDB937CE826ULL, "vxht_mansion2b_2qb_0_0.vlm"},
	{0x14B094D769F68F8FULL, "6114034_m.fxe"},
	{0x14B0E45C7DDEECD2ULL, "252724688.ogg"},
	{0x14B108EBBE92AE4DULL, "mag360_alain6044625.cl"},
	{0x14B1CCFCA66B931CULL, "hlac_lt_foundry_n_320_0y.dds"},
	{0x14B1E03AAF7BE04EULL, "vxw_encounter3_9r_0_0.vlm"},
	{0x14B1E3DF9B5825D3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_minehammer.phy"},
	{0x14B1F4767D76B1DAULL, "vxw_encounter2_camp_8f_0_0.vlm"},
	{0x14B20CA081FD4659ULL, "vxlt_undercityencounter_1ki_0_0.vlm"},
	{0x14B20CA98268908AULL, "mcr370_cut_player_survives6171844.cl"},
	{0x14B297882A948A9CULL, "vxlt_0_lowtown_docks_a3_n_bsk_0_0.vlm"},
	{0x14B2E9FF338D99C9ULL, "6079068_f.fxe"},
	{0x14B34C186F04168EULL, "gst_standing_forward.gda"},
	{0x14B358ECD458582EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containera.mmh"},
	{0x14B35F3AC196DB64ULL, "hlac_dr_dungeon_3207_6x.dds"},
	{0x14B364ECD4586CC5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containera.mao"},
	{0x14B372ECD458841CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containera.msh"},
	{0x14B3876074FD006FULL, "hlac_lt_hangedman_-1_17z.dds"},
	{0x14B3D56E49B5C8E6ULL, "fake_im_arm_glv_lgt_champ.uti"},
	{0x14B3F034A97F0527ULL, "592003352.ogg"},
	{0x14B424B9BF20C390ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ca_basic.mao"},
	{0x14B428B9BF20CAA3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ca_basic.mmh"},
	{0x14B436B9BF20E2FDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ca_basic.msh"},
	{0x14B4391B0B02B229ULL, "vxht_bartrand_n_32i_0_0.vlm"},
	{0x14B4A1BC50627492ULL, "qun211_party_comment6072257.cl"},
	{0x14B5C1069D8EED2FULL, "6051315_m.fxe"},
	{0x14B5FDCF7C64CEBAULL, "mer141_jansen6017761.cl"},
	{0x14B63877FB9F18B1ULL, "203711021.ogg"},
	{0x14B64A26FD6DC185ULL, "m_qcr220ar_hightown_nt.wbd"},
	{0x14B65675D0BFAE6EULL, "w_bonepitmining_a2_36_0_ll.mmh"},
	{0x14B6D1898B840DC4ULL, "lgt320_shw_ors_black_op6156321.cl"},
	{0x14B6FD03190502FBULL, "art\\vfx\\combat\\vfx_r_arc_critical_shot_c.mmh"},
	{0x14B77B631B78EBDCULL, "506590922.ogg"},
	{0x14B85A6F5710440BULL, "113940438.ogg"},
	{0x14B88776512FDBDBULL, "llac_w_woundedcoast_4_0.dds"},
	{0x14B8E2FD9C02F691ULL, "llac_ht_keep_4519_5.dds"},
	{0x14B93B179C33EE3AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestold.mmh"},
	{0x14B93F179C33F571ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestold.mao"},
	{0x14B94D179C340C80ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestold.msh"},
	{0x14B9524E451CA182ULL, "nat320_delilah6065915.cl"},
	{0x14B95729B41FE1DAULL, "6141824_m.fxe"},
	{0x14B98CB59E3AC3D0ULL, "312527420.ogg"},
	{0x14BA2BDF55A66CF5ULL, "6106484_m.fxe"},
	{0x14BA76C2CDCE3C5BULL, "798187957.ogg"},
	{0x14BA89C91D7ABCADULL, "299288238.ogg"},
	{0x14BA952B96EDC4DAULL, "6042053_m.fxe"},
	{0x14BA95DD716AF504ULL, "w_encounter2_15_0.mmh"},
	{0x14BA99DD716AFC57ULL, "w_encounter2_15_0.mao"},
	{0x14BAA9B178DC62D8ULL, "6076733_m.fxe"},
	{0x14BAABDD716B1AA6ULL, "w_encounter2_15_0.msh"},
	{0x14BB6335F0E18821ULL, "870524092.ogg"},
	{0x14BB6D546377BAD7ULL, "t1_mue_ro3.tnt"},
	{0x14BB83F8A59BC721ULL, "dae000_bartrand6031129.cl"},
	{0x14BBDF572069B76CULL, "6143419_m.fxe"},
	{0x14BC02637C25A1DBULL, "mrl150_merrill6085805.cl"},
	{0x14BC522A42452AF1ULL, "6066881_m.fxe"},
	{0x14BCEE51AC297643ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_ltintdub.msh"},
	{0x14BCFC51AC298D9DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_ltintdub.mmh"},
	{0x14BD0051AC2994AEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_ltintdub.mao"},
	{0x14BD12E26CB4ADBCULL, "vxlt_0_lowtown_docks_a3_n_d4d_0_0.vlm"},
	{0x14BD2AE1307A29A8ULL, "lt_0_lowtown_docks_a3_n_149_0.msh"},
	{0x14BD3CE1307A4839ULL, "lt_0_lowtown_docks_a3_n_149_0.mao"},
	{0x14BD3F39AA6BFD75ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubrush_sa.dds"},
	{0x14BD48E1307A5CD2ULL, "lt_0_lowtown_docks_a3_n_149_0.mmh"},
	{0x14BD56F5163789ADULL, "hlac_ca_cavern1_haunted_24426_3z.dds"},
	{0x14BD6BCB2D06E9A0ULL, "vxlt_gallowscourtyard_3vy_0_0.vlm"},
	{0x14BDBAFB374DC0F2ULL, "vxca_caverns2_3ho_0_0.vlm"},
	{0x14BE14547C1F0F8CULL, "hlac_ca_caverns1_24422_7z.dds"},
	{0x14BE7BA91A5C48FCULL, "mag320cr_templar_archer.utc"},
	{0x14BEB0B8857300A6ULL, "vxht_mansion2b_2jc_0_0.vlm"},
	{0x14BEB60FF3958306ULL, "vxw_woundedcoast_1c3_0_0.vlm"},
	{0x14BEC65273176B01ULL, "dae000cr_grand_cleric6147673.cl"},
	{0x14BEC66BF8F4A495ULL, "6016943_m.fxe"},
	{0x14BEC7AA7CC7EED6ULL, "w_woundedcoast_6_0.anb"},
	{0x14BF32D3E384B471ULL, "666955110.ogg"},
	{0x14BF3C0D56DE484CULL, "dae210cr_qunari_guard.utc"},
	{0x14BF5C3BBC32758CULL, "100185377.ogg"},
	{0x14BF9D850AD31ECAULL, "6133043_m.fxe"},
	{0x14C05217A171F8CEULL, "vxw_encounter3_2s_0_0.vlm"},
	{0x14C06055B33A62C9ULL, "gen00fl_anders6121205.cl"},
	{0x14C086BFACEE79EAULL, "176267838.ogg"},
	{0x14C0C8568D955771ULL, "gen_im_trs_rdr_one_mac_r0.uti"},
	{0x14C109651C8AA31CULL, "vxlt_0_lowtown_docks_a3_n_bvh_0_0.vlm"},
	{0x14C22254EE83FA69ULL, "227132813.ogg"},
	{0x14C2BB97F216A919ULL, "hlac_w_bonepitdestroyed_a3_131077_0z.dds"},
	{0x14C2D87C5B3D4D61ULL, "vxlt_gallowscourtyard_6zx_0_1.vlm"},
	{0x14C38F3E9CBAB66EULL, "vxlt_backalley_g_1vp_0_1.vlm"},
	{0x14C393317ED1BFA0ULL, "dae000_keep_guard_1_sta6163717.cl"},
	{0x14C45B6C0831FFE0ULL, "lt_gallowsprison_n_mmi.gff"},
	{0x14C4CA7861D5F0E8ULL, "art\\vfx\\combat\\vfx_glb_general_imp_c.mmh"},
	{0x14C51E0319117860ULL, "art\\vfx\\combat\\vfx_r_arc_critical_shot_c.anb"},
	{0x14C581A941AAFF3FULL, "m_qun100ar_plot_start.wbd"},
	{0x14C5D92BE4536E23ULL, "hlac_lt_blackemporium_1055_18x.dds"},
	{0x14C64BA807B9B4E2ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_dirt_n.dds"},
	{0x14C6D392DC3E6E71ULL, "6161407_m.fxe"},
	{0x14C6E351A2932D77ULL, "vxht_bartrand_n_67q_0_0.vlm"},
	{0x14C738965311EEA3ULL, "hlac_lt_0_lowtown_docks_n_-196607_0z.dds"},
	{0x14C8031FCA53D546ULL, "6119748_m.fxe"},
	{0x14C822E1E39DF70BULL, "hlac_dr_0_deeproads_5840_0x.dds"},
	{0x14C85E3FFBC670CDULL, "vxlt_0_lowtown_a2_n_c_hx5_0_0.vlm"},
	{0x14C8B1E7A004D1BDULL, "6053315_m.fxe"},
	{0x14C8B54F4B382DF9ULL, "6110674_m.fxe"},
	{0x14C8B8C993F33714ULL, "6108606_f.fxe"},
	{0x14C8C4CA6C0C444BULL, "674016425.ogg"},
	{0x14C92192BCFCE7BEULL, "ht_0_hightown_a2_45_0_lc.mmh"},
	{0x14C9659107C8729EULL, "hlac_lt_0_lowtown_a2_-1_7z.dds"},
	{0x14C9906B7CA29F6FULL, "6177402_m.fxe"},
	{0x14C9B7CE155FD74CULL, "873515443.ogg"},
	{0x14C9EEB286FF0459ULL, "vxw_bonepitdestroyed_a3_iv_0_0.vlm"},
	{0x14C9F0A4E30A42B5ULL, "follower_banter6167994.cl"},
	{0x14CA3C6DC934BA71ULL, "vxlt_0_lowtown_a2_n_d5t_0_0.vlm"},
	{0x14CACF5B9177450AULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgtp_0d.dds"},
	{0x14CBA2119586DD95ULL, "202804210.ogg"},
	{0x14CBAA597807646CULL, "llac_lt_0_lowtown_-65535_6.dds"},
	{0x14CBAA68E3724F5DULL, "796697827.ogg"},
	{0x14CBAEE20EB47161ULL, "llac_w_bonepit_196609_0.dds"},
	{0x14CBC4AA7CD2739FULL, "w_woundedcoast_6_0.msh"},
	{0x14CBCAAA7CD27E41ULL, "w_woundedcoast_6_0.mmh"},
	{0x14CBD6AA7CD292E2ULL, "w_woundedcoast_6_0.mao"},
	{0x14CC486FB463B6FEULL, "qun221_viscount6096280.cl"},
	{0x14CC790AEDAE703AULL, "gen_im_trs_gem_rosediam_02.uti"},
	{0x14CC98E0595837E0ULL, "1058343396.ogg"},
	{0x14CCC679A576E094ULL, "590362616.ogg"},
	{0x14CCECB90934C85EULL, "zz_approval_debug6003788.cl"},
	{0x14CCFAEB98D1C938ULL, "vxw_woundedcoast_1nz_0_0.vlm"},
	{0x14CD229577691926ULL, "vxht_mansion2b_2gd_0_0.vlm"},
	{0x14CDEF4D273F1A60ULL, "lt_0_lowtown_docks_135_0.phy"},
	{0x14CE105D176EFE73ULL, "llac_lt_0_lowtown_a2_65536_9.dds"},
	{0x14CE4CFDB6CCBCD9ULL, "6139936_m.fxe"},
	{0x14CE58FBDFFD78A4ULL, "vxw_bonepitmining_a2_29y_0_0.vlm"},
	{0x14CE96C2D3071568ULL, "6083764_m.fxe"},
	{0x14CFAF439B14168AULL, "6161000_m.fxe"},
	{0x14CFFECEFB3F3D99ULL, "vxlt_0_lowtown_a2_eoi_0_1.vlm"},
	{0x14D00CC1D2F87D94ULL, "vxlt_ruinedhouse_n_6r_0_0.vlm"},
	{0x14D0C8D5AD671723ULL, "follower_banter6164496.cl"},
	{0x14D10677C3632D3EULL, "vxlt_0_lowtown_docks_a3_brq_0_0.vlm"},
	{0x14D16D282036F6F7ULL, "6174511_m.fxe"},
	{0x14D1A71FC04786BEULL, "vxca_cavern1_haunted_gk8_0_0.vlm"},
	{0x14D1AFFDF2B4A299ULL, "hlac_ht_0_hightown_a3_65536_3y.dds"},
	{0x14D208DA7BE40E28ULL, "ss_qunari_mage.bnk"},
	{0x14D28640F7550AB4ULL, "ht_0_hightown_43_0_lc.mmh"},
	{0x14D2C4CE57B18F9AULL, "vxlt_gallowscourtyard_n_7au_0_0.vlm"},
	{0x14D2C73905B4888EULL, "llac_ca_cavern1_haunted_24425_1.dds"},
	{0x14D33D972F484CBFULL, "cd_clf260_harley.bnk"},
	{0x14D3C65D399D54FCULL, "one280_merin6116767.cl"},
	{0x14D3EAA6C7EAE765ULL, "fx_dis_sphere.mao"},
	{0x14D411A58E251A87ULL, "6053087_m.fxe"},
	{0x14D429836E25ABEDULL, "6144851_m.fxe"},
	{0x14D44436A440415AULL, "1415162.ogg"},
	{0x14D4B92DAEFC9015ULL, "dre000_varric6208464.cl"},
	{0x14D4CD7D9E10BE05ULL, "zz_personality_debug6004633.cl"},
	{0x14D54B73FF9B2983ULL, "6136885_m.fxe"},
	{0x14D5887DE40C2806ULL, "hlac_ca_cavern1_haunted_24421_12z.dds"},
	{0x14D67F9D682CB866ULL, "hlac_lt_gallowstemplar_f_2383_12x.dds"},
	{0x14D6CA1386F916F5ULL, "follower_banter6148755.cl"},
	{0x14D71F96086C9932ULL, "vxdr_0_deeproads_exa_0_0.vlm"},
	{0x14D810677EFC5147ULL, "unlockables_col.gda"},
	{0x14D8901D105AC553ULL, "6108909_m.fxe"},
	{0x14D8AFAA7107E0D9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7oc.msh"},
	{0x14D8B9AA7107F1C7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7oc.mmh"},
	{0x14D8BDAA7107F814ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7oc.mao"},
	{0x14D8D450C8F417C4ULL, "w_encounter3_119_0.msh"},
	{0x14D8D85C0F1FDB97ULL, "llac_ht_0_hightown_-1_2.dds"},
	{0x14D8E650C8F4362DULL, "w_encounter3_119_0.mao"},
	{0x14D8EA50C8F43D66ULL, "w_encounter3_119_0.mmh"},
	{0x14D91037B81D24A1ULL, "hlac_lt_0_lowtown_docks_-2_6x.dds"},
	{0x14D932500474315FULL, "framebuffereffects_motionblur_p0.psh.fxo.dx11.assembly"},
	{0x14D9401807CD6CBEULL, "378281023.ogg"},
	{0x14D9FA3095D346BAULL, "33963352.ogg"},
	{0x14DA8F9235C96C49ULL, "6119206_m.fxe"},
	{0x14DA9EB51B39F1F2ULL, "vxca_caverns2_3ri_0_0.vlm"},
	{0x14DAAE6E76D140D6ULL, "730889061.ogg"},
	{0x14DADE6B5B25B436ULL, "dae000_bartrand6205969.cl"},
	{0x14DB4E3B64A46D50ULL, "vxht_0_hightown_c3v_0_1.vlm"},
	{0x14DB96D8F997C64EULL, "vxlt_0_lowtown_a2_eqq_0_0.vlm"},
	{0x14DB99C9D781B406ULL, "vxw_woundedcoast_1i5_0_0.vlm"},
	{0x14DBD1807A110A9DULL, "vxht_brothel_n_332_0_0.vlm"},
	{0x14DC0BA05D44577CULL, "307930170.ogg"},
	{0x14DCE390B79083A2ULL, "41872026.ogg"},
	{0x14DCEE6FCEFCBFEBULL, "836929266.ogg"},
	{0x14DD201CF9E291B6ULL, "vxlt_hangedman_1xx_0_0.vlm"},
	{0x14DD40D29BA36D51ULL, "333464572.ogg"},
	{0x14DD71AC429FF092ULL, "llac_lt_darktown_-1_53.dds"},
	{0x14DD895626C674E7ULL, "409781989.ogg"},
	{0x14DDF4679F5AB009ULL, "6034796_m.fxe"},
	{0x14DE0617D355E433ULL, "14567639.ogg"},
	{0x14DE740FF73E34ECULL, "706451119.ogg"},
	{0x14DE932C86A009F5ULL, "and230st_base_dissent_end.stg"},
	{0x14DEA92FB1E0658EULL, "mag110_arianni_intro_end.ncs"},
	{0x14DEDCA5E4B4F472ULL, "llac_lt_0_lowtown_a3_n_c_65536_4.dds"},
	{0x14DF65CE65F9CDFEULL, "dre000_dougal6131195.cl"},
	{0x14DF94D0F8B4BB8DULL, "vxlt_0_lowtown_a3_n_c_eod_0_0.vlm"},
	{0x14E0348617129562ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls101a_0t.dds"},
	{0x14E050D44F1AC2C7ULL, "6043529_m.fxe"},
	{0x14E09F81D17DE7E1ULL, "ss_cityelf_m_1.wbd"},
	{0x14E0DAA686DC946EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archuprmost.phy"},
	{0x14E11493C3FEB539ULL, "146989374.ogg"},
	{0x14E138D51147A6D9ULL, "6068876_m.fxe"},
	{0x14E139B04398CC36ULL, "vxht_0_hightown_4of_0_0.vlm"},
	{0x14E2A9615752344BULL, "6117915_f.fxe"},
	{0x14E2B17D757D62F0ULL, "hlac_ca_caverns1_mines_24426_4x.dds"},
	{0x14E2EB2FAC99629CULL, "art\\levels\\ztools\\default\\generic\\2co_basic.mao"},
	{0x14E2F72FAC99773FULL, "art\\levels\\ztools\\default\\generic\\2co_basic.mmh"},
	{0x14E2FD2FAC9981E1ULL, "art\\levels\\ztools\\default\\generic\\2co_basic.msh"},
	{0x14E3CF17C9ED33C9ULL, "vxlt_0_lowtown_a2_hh8_0_0.vlm"},
	{0x14E3D0F7599A0A15ULL, "691019331.ogg"},
	{0x14E45390441B9A4AULL, "58122901.ogg"},
	{0x14E479C201A7CFEAULL, "363422897.ogg"},
	{0x14E4869812053A33ULL, "hlac_w_bonepit_0_9x.dds"},
	{0x14E4C9C1AD524078ULL, "qun221_petrice6076812.cl"},
	{0x14E4E03B3B449A51ULL, "6045789_m.fxe"},
	{0x14E50EC2EFE6CDC1ULL, "6031812_m.fxe"},
	{0x14E530671199CB19ULL, "dae000_bartrand6064062.cl"},
	{0x14E571E9CE612E2DULL, "589578799.ogg"},
	{0x14E5A84841AA6A11ULL, "6172959_m.fxe"},
	{0x14E62C4904DF5A0EULL, "986211423.ogg"},
	{0x14E6658ECDE4EB63ULL, "6099838_m.fxe"},
	{0x14E6BB3D06F61252ULL, "vxca_caverns1_mines_eap_0_0.vlm"},
	{0x14E6BD5B3C9C4673ULL, "llac_lt_warehouse_0_2.dds"},
	{0x14E70BEA727CAED7ULL, "vxlt_gallowstemplar_3ok_0_0.vlm"},
	{0x14E7452065237067ULL, "vxlt_0_lowtown_a3_n_c_e8c_0_0.vlm"},
	{0x14E74D817DFFC59BULL, "aip_strider_plot.gda"},
	{0x14E77EC5F08A9A49ULL, "vxht_brothel_n_435_0_0.vlm"},
	{0x14E7BDCC311835E0ULL, "343334781.ogg"},
	{0x14E7DBB2BC1DF4ABULL, "vxw_encounter2_1t_0_0.vlm"},
	{0x14E80A2B092C6323ULL, "6126505_m.fxe"},
	{0x14E8457A02279356ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v2.evt"},
	{0x14E85EBF4314048BULL, "729313741.ogg"},
	{0x14E92696F798AE31ULL, "vxca_caverns1_il9_0_0.vlm"},
	{0x14E99A7E799C2B37ULL, "vxht_keep_n_618_0_0.vlm"},
	{0x14E9DB99AC42F7DAULL, "gen00fl_merrill6133146.cl"},
	{0x14E9E4FD211EF9C0ULL, "dr_primevalentrance_1726_ll.mmh"},
	{0x14EA08B5EB8DDECEULL, "vxlt_0_lowtown_a2_etr_0_0.vlm"},
	{0x14EA0A5476F37177ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_gallowsalcov.msh"},
	{0x14EA0BA6C977CC86ULL, "vxw_woundedcoast_1t6_0_0.vlm"},
	{0x14EA1C5476F38FDAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_gallowsalcov.mao"},
	{0x14EA205476F396E9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_gallowsalcov.mmh"},
	{0x14EA8385AC513C70ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg04a.msh"},
	{0x14EA9185AC51538AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg04a.mmh"},
	{0x14EADE99A51F7B47ULL, "757501978.ogg"},
	{0x14EB4E56998E9990ULL, "101231114.ogg"},
	{0x14EBAC6437978C53ULL, "vxca_caverns1_mines_eft_0_0.vlm"},
	{0x14EC4236F02B7ECBULL, "gen_im_trs_gen_arc_lbw_r0.uti"},
	{0x14EC76D137EC3FE7ULL, "230523101.ogg"},
	{0x14ECE3C4D6102AE1ULL, "zz_one_debug6113371.cl"},
	{0x14ED1AE94B2387F7ULL, "795583048.ogg"},
	{0x14ED42D2FD5697ADULL, "streak01.mao"},
	{0x14ED80C87F4B5855ULL, "vxht_0_hightown_a2_c5h_0_0.vlm"},
	{0x14ED8F8B5B340DCEULL, "tbe100ip_shop_codex_12.utp"},
	{0x14EDFBDDF362E24AULL, "vxlt_warehouse_b3_0_0.vlm"},
	{0x14EE02647D56FEDDULL, "vxlt_gallowstemplar_3mc_0_0.vlm"},
	{0x14EE0B271D5A3993ULL, "gen00fl_isabela6100073.cl"},
	{0x14EE20307A81522DULL, "941223079.ogg"},
	{0x14EE4EF74A90B2F9ULL, "402689086.ogg"},
	{0x14EE6C942B67C2F1ULL, "vxca_caverns1_mines_e84_0_0.vlm"},
	{0x14EE9DDC72034901ULL, "hlac_ca_cave_2194_2x.dds"},
	{0x14EF078092DE26D6ULL, "genip_node_h1.utp"},
	{0x14EF190FA6B607BAULL, "plt_cod_hst_mag_deathroot.plo"},
	{0x14EF3C7A022DA2E1ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v2.gad"},
	{0x14EF3F1F44C675FEULL, "dae000_orana6149020.cl"},
	{0x14EF825D1C4A5BC4ULL, "6091640_m.fxe"},
	{0x14EF9561295AAF52ULL, "gen_im_acc_rng_act3_01.uti"},
	{0x14EFC0B384F6C7E5ULL, "cod_bks_fifth_blight.ncs"},
	{0x14EFD24788B4869AULL, "gen00fl_isabela6079024.cl"},
	{0x14F0B0CE7E69758FULL, "vxht_0_hightown_n_b90_0_0.vlm"},
	{0x14F0C22DF1928169ULL, "mag160_feynriel6020088.cl"},
	{0x14F0DCF55B466CF7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeenter.gad"},
	{0x14F12AFA2321CB7CULL, "llac_ht_0_hightown_-65535_0.dds"},
	{0x14F181FA10499E85ULL, "bld_dri_mo.mao"},
	{0x14F1AAAA7CF3176AULL, "w_woundedcoast_6_0.rml"},
	{0x14F1BA3941776E44ULL, "plt_tbe100pt_emporium.plo"},
	{0x14F1DEBB02A3AE7DULL, "mer320_hubert.stg"},
	{0x14F22493D660FE67ULL, "hlac_ht_0_hightown_a2_-131072_1x.dds"},
	{0x14F24389822E2FDBULL, "gen00flst_post_romance.stg"},
	{0x14F25F197A8286DAULL, "233214579.ogg"},
	{0x14F2A71EAC6B1523ULL, "303649554.ogg"},
	{0x14F3274E218ABDF0ULL, "zz_dae.ncs"},
	{0x14F40C4D27601AF5ULL, "lt_0_lowtown_docks_135_0.msh"},
	{0x14F40E4D27601E6BULL, "lt_0_lowtown_docks_135_0.mmh"},
	{0x14F410BF257DE9EFULL, "hlac_lt_gallowstemplar_f_2384_7z.dds"},
	{0x14F41A4D27603208ULL, "lt_0_lowtown_docks_135_0.mao"},
	{0x14F428AC03C6CABCULL, "gen00fl_varric6080080.cl"},
	{0x14F438778232C04DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masa_0d.dds"},
	{0x14F4BDFA458E94A6ULL, "6083788_m.fxe"},
	{0x14F4C5947BC12DE0ULL, "vxw_blightlands_13g_0_0.vlm"},
	{0x14F53D7A02320F8FULL, "art\\characters\\creatures\\mubari\\animation\\atk_v2.ani"},
	{0x14F5936B01EA5AA4ULL, "hlac_ht_keep_n_3741_0y.dds"},
	{0x14F5B941515C9D83ULL, "vxht_mansion2b_3ch_0_0.vlm"},
	{0x14F653F30DA814A2ULL, "gen00fl_sebastian6187771.cl"},
	{0x14F67856D7A340ADULL, "204331600.ogg"},
	{0x14F72426FDA5182FULL, "m_qcr220ar_hightown_nt.bnk"},
	{0x14F7489367C4A6C4ULL, "251333102.ogg"},
	{0x14F7F1F55B4CAE84ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeenter.evt"},
	{0x14F7FADA7C04C5D2ULL, "ss_qunari_mage.wbd"},
	{0x14F88ACEA18C0642ULL, "mcr370_meredith6083093.cl"},
	{0x14F9411A7C7A680AULL, "vxlt_0_lowtown_a2_gbr_0_0.vlm"},
	{0x14F9A562CFBE35E5ULL, "vxw_encounter2_camp_9z_0_0.vlm"},
	{0x14FA3ADF8C8536DAULL, "vxlt_0_lowtown_docks_a3_cz9_0_1.vlm"},
	{0x14FAABCE93C65D7FULL, "6141818_m.fxe"},
	{0x14FB040C7ECD0A48ULL, "6185230_f.fxe"},
	{0x14FB05FC7541622CULL, "one300dg_lia6064533.cl"},
	{0x14FB41545D54C6A8ULL, "94677661.ogg"},
	{0x14FB7AA3F156CA6DULL, "dae000_mother6121076.cl"},
	{0x14FBA03FBDCC0E5FULL, "230387732.ogg"},
	{0x14FBA056CCA2C0C6ULL, "298543830.ogg"},
	{0x14FBE090AEE8B075ULL, "vxht_playershouse_n_208_0_0.vlm"},
	{0x14FBF2A5714170D5ULL, "vxht_0_hightown_a2_c0w_0_0.vlm"},
	{0x14FBF312E7EB0F14ULL, "vxca_cavern1_haunted_e6g_0_0.vlm"},
	{0x14FBF4D0D2B864BEULL, "llac_ht_0_hightown_196607_0.dds"},
	{0x14FC1E725A1FF6DDULL, "6141896_m.fxe"},
	{0x14FC51798B6FB9DFULL, "plt_vrc300pt_n_hntd_start.nss"},
	{0x14FC74416F4D175DULL, "vxlt_gallowstemplar_3jb_0_0.vlm"},
	{0x14FC788ADCA0EC8DULL, "vxlt_0_lowtown_a3_n_c_eub_0_0.vlm"},
	{0x14FC81DAAEAFE572ULL, "dae000_merchant_mage_hf6135891.cl"},
	{0x14FCC6274456D197ULL, "685046576.ogg"},
	{0x14FD67E886DCC301ULL, "rdr111dg_harbormaster6015965.cl"},
	{0x14FD75F156769BE3ULL, "6091373_m.fxe"},
	{0x14FDF959AEE14606ULL, "llac_lt_0_lowtown_a2_n_131072_4.dds"},
	{0x14FE4BEEEFB5EC10ULL, "vxlt_gallowsprison_n_6hs_0_0.vlm"},
	{0x14FE6F15EBC64101ULL, "fen110_fenris6118088.cl"},
	{0x14FF23972F6D5BF5ULL, "cd_clf260_harley.wbd"},
	{0x14FF5529EE09360DULL, "6095634_m.fxe"},
	{0x15001E4DE8194633ULL, "vxht_0_hightown_aok_0_1.vlm"},
	{0x150039B182D31113ULL, "vxca_caverns1_hvl_0_0.vlm"},
	{0x15009501F74B54A8ULL, "dae000_hate_follower_react6103219.cl"},
	{0x1500B1B38505213CULL, "cod_bks_fifth_blight.utp"},
	{0x150143519E091451ULL, "hlac_lt_0_lowtown_docks_a3_-196610_0z.dds"},
	{0x150160318EB03001ULL, "art\\levels\\races\\proxy\\props\\_textures\\fires_sa.dds"},
	{0x150175C9DC776614ULL, "ss_player_male.wbd"},
	{0x1501E3E0A6F94B5FULL, "gen00fl_merrill6106426.cl"},
	{0x1501FA861791FC3DULL, "vxlt_gallowstemplar_n_45w_0_0.vlm"},
	{0x1501FE674F712710ULL, "cod_hst_kirk_arch_1a_7.utp"},
	{0x150254C454BFF70BULL, "tbe100_xenon6190987.cl"},
	{0x15028CAB5BA94E0AULL, "6087515_m.fxe"},
	{0x150292765B8CD2A4ULL, "hlac_lt_backalley_g_65537_4z.dds"},
	{0x1502BEE711523244ULL, "636981647.ogg"},
	{0x1502F2EEC32935D1ULL, "vxw_encounter2_kp_0_0.vlm"},
	{0x15032FD73784AD26ULL, "6086888_m.fxe"},
	{0x15037D404FF75588ULL, "638207611.ogg"},
	{0x1503AFEC0FC57EABULL, "6027471_m.fxe"},
	{0x15042B1E4352B603ULL, "vxht_mansion2b_3fk_0_0.vlm"},
	{0x1504DECC72E873C5ULL, "319058816.ogg"},
	{0x150509F55B57627DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeenter.ani"},
	{0x150559B92A580A26ULL, "hlac_lt_0_lowtown_docks_a3_196608_0y.dds"},
	{0x15056EB7C99AD8A7ULL, "vxht_0_hightown_a3_cas_0_0.vlm"},
	{0x15059C947E099E83ULL, "t3_wdg_br1.tnt"},
	{0x1505A2AA712EB8FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7oc.phy"},
	{0x1505B74F377AEAE3ULL, "6104732_m.fxe"},
	{0x1505C181D19D3DE3ULL, "ss_cityelf_m_1.bnk"},
	{0x1505E50006180D87ULL, "vxlt_0_lowtown_e1h_0_0.vlm"},
	{0x1505EF7E1C2ED86EULL, "static_static_p25.psh.fxo.dx11"},
	{0x1506338608B46CCDULL, "ht_0_hightown_n_30_0_lc.mmh"},
	{0x1506543EBD2D22BDULL, "gen00fl_merrill6108435.cl"},
	{0x1506DCEF14397A47ULL, "103892226.ogg"},
	{0x1507385D800423D6ULL, "l_mcr310ar_docks_nt.wbd"},
	{0x1507AC8C04848CD5ULL, "plt_lgt310pt_shw_mer_main.nss"},
	{0x1507E91F31E27F74ULL, "cd_dre000_cut_rock_wraith.bnk"},
	{0x150853E33B784FE8ULL, "qcr220_orsino6041839.cl"},
	{0x1508A5516C3E6191ULL, "6101221_m.fxe"},
	{0x1509A858801808BEULL, "gen00fl_aveline6139790.cl"},
	{0x150A463061731AD0ULL, "vxca_caverns2_7wd_0_0.vlm"},
	{0x150A6CE264845F17ULL, "6055427_m.fxe"},
	{0x150AE61E61432FDDULL, "vxlt_gallowstemplar_3ga_0_0.vlm"},
	{0x150AEA67CE97050DULL, "vxlt_0_lowtown_a3_n_c_exa_0_0.vlm"},
	{0x150AF0B48170BE99ULL, "lowlevelloadhints.gda"},
	{0x150B24D953D44AA4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wic_0s.dds"},
	{0x150C15175FB75E26ULL, "vxlt_backalley_g_1pr_0_1.vlm"},
	{0x150C729EFA941805ULL, "llac_lt_0_lowtown_a3_n_c_-3_2.dds"},
	{0x150CA7B9EF92BBC0ULL, "6040624_m.fxe"},
	{0x150CD697D64F6020ULL, "vxw_bonepitdestroyed_a3_2a3_0_0.vlm"},
	{0x150CEE4E4314F640ULL, "6120792_f.fxe"},
	{0x150D15FCDC2F347FULL, "vxht_0_hightown_n_b8w_0_0.vlm"},
	{0x150D2110645B605BULL, "vxlt_darktown_5km_0_0.vlm"},
	{0x150D62BAC7DEB0B2ULL, "6079648_m.fxe"},
	{0x150E10936C1AA34BULL, "6036697_m.fxe"},
	{0x150E1BDE49218EEBULL, "779463234.ogg"},
	{0x150E2570743C1D6AULL, "dae200cr_viscount6087389.cl"},
	{0x150E671FE128EBCDULL, "vxlt_warehouse_13n_0_0.vlm"},
	{0x150E6CFB3AEC64D6ULL, "hlac_ht_keep_n_4519_14z.dds"},
	{0x150EE4B22006459EULL, "6128920_m.fxe"},
	{0x150F48954257F772ULL, "193699335.ogg"},
	{0x150F7FFA3914BE61ULL, "6132850_m.fxe"},
	{0x150F8EAB83200A3DULL, "vxht_0_hightown_n_bpu_0_0.vlm"},
	{0x150FA4F4B589A93FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mbn.msh"},
	{0x150FAAF4B589B3E1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mbn.mmh"},
	{0x150FB6F4B589C782ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mbn.mao"},
	{0x150FFE1AF7ECF2DCULL, "6125478_f.fxe"},
	{0x151037E6D97A4F18ULL, "6070814_m.fxe"},
	{0x1510B46F56B6F33DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_boardsxa.phy"},
	{0x15112C1E689307C0ULL, "hlac_lt_0_lowtown_docks_a3_-3_8y.dds"},
	{0x151144798B81FFF0ULL, "plt_vrc300pt_n_hntd_start.plo"},
	{0x151185C4CBFC5E87ULL, "llac_lt_0_lowtown_docks_65535_5.dds"},
	{0x1511C8157DF43F64ULL, "hlac_dr_slavers_3214_7y.dds"},
	{0x1512899CFF97A2EEULL, "vxlt_gallowstemplar_f_6sx_0_0.vlm"},
	{0x15128ED138C8DA6BULL, "mag210_arianni6021225.cl"},
	{0x15131658B40A8CB0ULL, "cs_qcr221cs_isabellas_return.bnk"},
	{0x1513BA9D6E4373A8ULL, "828560490.ogg"},
	{0x1513D436C9FDC64CULL, "1038907774.ogg"},
	{0x15146588040C790BULL, "vxlt_0_lowtown_a2_n_hcp_0_0.vlm"},
	{0x1514867916D71AB9ULL, "qun220cr_javaris6071519.cl"},
	{0x1514CF622F8D945AULL, "vxdr_primevalentrance_271_0_0.vlm"},
	{0x1514DCDF9BAC9748ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_minehammer.mmh"},
	{0x1514E8DF9BACABEBULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_minehammer.mao"},
	{0x1514FADF9BACCA72ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_minehammer.msh"},
	{0x1514FF90DB8976B4ULL, "vxca_caverns1_jgn_0_0.vlm"},
	{0x15152D5BD500547FULL, "vxht_keep_n_6je_0_0.vlm"},
	{0x151552E7DDFDD7DCULL, "follower_banter6161133.cl"},
	{0x15156F0120A16545ULL, "dre000_dougal6131132.cl"},
	{0x151584B32A2BA2BDULL, "275192780.ogg"},
	{0x151847AB5AD86CBDULL, "mrl210_merrill6102101.cl"},
	{0x15187A7C729DB754ULL, "vxca_caverns2_864_0_0.vlm"},
	{0x1519320B00B3BEC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_cheers.gad"},
	{0x151947E130C8B9CDULL, "lt_0_lowtown_docks_a3_n_149_0.phy"},
	{0x151951115294C444ULL, "party_member.cur"},
	{0x1519774EBA2538A4ULL, "6040967_m.fxe"},
	{0x1519BEF2CDA1E616ULL, "461626991.ogg"},
	{0x1519C1236B23278FULL, "vfx_vfx_p6.vsh.fxo.dx11"},
	{0x151B7EF3E9CC70E3ULL, "vxca_caverns1_mines_ih2_0_0.vlm"},
	{0x151B92ED565178DBULL, "vxlt_darktown_5ll_0_0.vlm"},
	{0x151C0A3DE60088DEULL, "dae150ip_at_scaverns_b.utp"},
	{0x151C1CCACAF122DCULL, "vxht_0_hightown_cf8_0_0.vlm"},
	{0x151C35474870A8C6ULL, "6110955_f.fxe"},
	{0x151C95CB8A6A479CULL, "vxht_bartrand_n_5pl_0_0.vlm"},
	{0x151C9C9C19802158ULL, "vxht_keep_647_0_0.vlm"},
	{0x151CA7D853E98014ULL, "vxlt_hangedman_n_228_0_0.vlm"},
	{0x151CDDD51C554F68ULL, "llac_lt_0_lowtown_docks_n_-196610_1.dds"},
	{0x151D5FB41C750FE0ULL, "mag130cr_slaver3.utc"},
	{0x151D7E6A7EA8B2BBULL, "vxht_0_hightown_a3_b78_0_0.vlm"},
	{0x151DA7A4D337E365ULL, "gen00_merrillact2_pt3.stg"},
	{0x151E0BF4BC28C2A9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoorbird_sa.dds"},
	{0x151EEAAE97BA871CULL, "182346040.ogg"},
	{0x151F939D3A7E60BBULL, "hlac_ht_mansion2_818_1x.dds"},
	{0x151FD6A8A71566D1ULL, "vxw_encounter2_qv_0_0.vlm"},
	{0x151FE2B94FCE10D9ULL, "6136740_m.fxe"},
	{0x152022BB83BD990CULL, "hlac_lt_0_lowtown_-65536_4y.dds"},
	{0x1520550B00BA18B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_cheers.evt"},
	{0x15208ED62E1C559AULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgtp_0t.dds"},
	{0x1520D8869DECE572ULL, "ser120_ghyslain6118438.cl"},
	{0x152151A6BB1E4F03ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_meat_sa.dds"},
	{0x152157EA6E593847ULL, "6126937_f.fxe"},
	{0x1521738EA890B60BULL, "286333686.ogg"},
	{0x15219D33F143471BULL, "vxdr_primevaltomb_4ho_0_0.vlm"},
	{0x1521A96F56C553A9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_boardsxa.mao"},
	{0x1521B56F56C56842ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_boardsxa.mmh"},
	{0x1521B573D3E875BAULL, "hlac_ca_cavern1_haunted_24424_8z.dds"},
	{0x1521B76F56C56BF8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_boardsxa.msh"},
	{0x1521C6F3A3F0C2FEULL, "follower_banter6156021.cl"},
	{0x1521D9A687142931ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archuprmost.mmh"},
	{0x1521E5A687143DD2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archuprmost.mao"},
	{0x1521F3A6871455EFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archuprmost.msh"},
	{0x1521FC6938F01EC2ULL, "313541342.ogg"},
	{0x152252432A459762ULL, "art\\vfx\\textures\\vfx_dust.dds"},
	{0x15229C5CF731F91FULL, "vxlt_blackemporium_ug_0_0.vlm"},
	{0x1522D447EE871F29ULL, "llac_lt_0_lowtown_a2_n_c_-131073_1.dds"},
	{0x1522E81904BC358FULL, "vxlt_0_lowtown_ewb_0_0.vlm"},
	{0x1523890F47F85667ULL, "hlac_dr_0_deeproads_17038_1y.dds"},
	{0x15238C99F30947EBULL, "411115822.ogg"},
	{0x152394D20EDE6FD6ULL, "6059013_m.fxe"},
	{0x1523FBD147F01B4EULL, "hlac_lt_0_lowtown_a2_n_0_8z.dds"},
	{0x15248D4BC518DD50ULL, "43557914.ogg"},
	{0x1524AA15FA9D5E33ULL, "vxht_0_hightown_n_b5h_0_0.vlm"},
	{0x152564CE5E88EC24ULL, "6164359_m.fxe"},
	{0x1525738FCBE8C212ULL, "gen00fl_isabela6099343.cl"},
	{0x1525C2E43817BBFBULL, "118712740.ogg"},
	{0x15270D47600CAA21ULL, "vxlt_hangedman_2uo_0_0.vlm"},
	{0x15271C0C5CF7F17FULL, "vxdr_primevalentrance_31w_0_0.vlm"},
	{0x1527611D6ABF103CULL, "mag131_keran6121099.cl"},
	{0x1527A742113EBFF6ULL, "gen00fl_merrill6108676.cl"},
	{0x1527CC82D99B98FCULL, "qcr000_varric6065827.cl"},
	{0x1527E4BB02D147CAULL, "mer320_hubert.cnv"},
	{0x1528687524D1DE3FULL, "461295255.ogg"},
	{0x15289A8CEED27440ULL, "6156897_m.fxe"},
	{0x1528EB30D340E873ULL, "dae000_junar6147871.cl"},
	{0x152944EE7CE2CEDAULL, "w_bonepit_9_0.rml"},
	{0x15298F0E4456BE66ULL, "6119387_f.fxe"},
	{0x1529AB8C04A17136ULL, "plt_lgt310pt_shw_mer_main.plo"},
	{0x1529AC3BC4E4E014ULL, "vxht_bartrand_n_4d1_0_0.vlm"},
	{0x1529E36F3825DB26ULL, "fx_darkbolt.dds"},
	{0x152A578B8D49A373ULL, "dae210_gate_guard6084325.cl"},
	{0x152AA46650548E26ULL, "hlac_ht_chantry_n_2534_43y.dds"},
	{0x152AB374D59570D6ULL, "qcr270_fake_meredith6066343.cl"},
	{0x152AC6BB02D34AF2ULL, "mer320_hubert.bnk"},
	{0x152B0E790B7639D8ULL, "vxht_keep_614_0_0.vlm"},
	{0x152B2AD7BBE72D2FULL, "hm_refugee_boy.mor"},
	{0x152BF514BE93CE36ULL, "vxw_encounter2_camp_5q_0_0.vlm"},
	{0x152C1F51FC392CE8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobb_0n.dds"},
	{0x152C36C8E12E2739ULL, "940986117.ogg"},
	{0x152C62C8397736CAULL, "56547094.ogg"},
	{0x152D1CE7D173A2E8ULL, "6092199_m.fxe"},
	{0x152D5D0B00C4B0A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_cheers.ani"},
	{0x152D63C9DC9C82AEULL, "ss_player_male.bnk"},
	{0x152D788809F36754ULL, "vxdr_primevaltomb_3pc_0_0.vlm"},
	{0x152DB0A0511C61AEULL, "vxw_encounter2_camp_h_0_0.vlm"},
	{0x152DD705D6959F1DULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant_1.mao"},
	{0x152DDE6B9699F1E2ULL, "6151278_m.fxe"},
	{0x152DEB8A5951DD41ULL, "vxw_bonepitdestroyed_a3_dm_0_0.vlm"},
	{0x152DFB5DF75A4B91ULL, "one110ip_ggt_dog_pile.ncs"},
	{0x152E2D1BF32A0DD6ULL, "vxht_bartrand_n_5v0_0_0.vlm"},
	{0x152E41AD8E735CFEULL, "144037993.ogg"},
	{0x152E4B9E785DBA9DULL, "zz_ave_debug6174723.cl"},
	{0x152E549641C42959ULL, "6131640_m.fxe"},
	{0x152E7168D06737CDULL, "hlac_lt_blackemporium_1055_11y.dds"},
	{0x152E75E86D17BC05ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe5_p.mmh"},
	{0x152ECA3384A511C4ULL, "655963350.ogg"},
	{0x152ED52359C74639ULL, "llac_lt_undercityencounter_-1_38.dds"},
	{0x152EF6FB12260ED3ULL, "dre000_varric6036991.cl"},
	{0x152F2A181ECCA0CAULL, "6076490_m.fxe"},
	{0x152F78CEAC160222ULL, "1041459677.ogg"},
	{0x153014FD80750B81ULL, "vxht_0_hightown_ccj_0_0.vlm"},
	{0x15306760E6844BDBULL, "6086886_f.fxe"},
	{0x15309F9001437B89ULL, "gen00fl_isabela6090815.cl"},
	{0x15313E850222C45FULL, "vxht_keep_4np_0_0.vlm"},
	{0x1531568C1402E72FULL, "95135981.ogg"},
	{0x15316885CFD27191ULL, "vxw_bonepit_zc_0_0.vlm"},
	{0x1531A1AB15516BE1ULL, "damagetypes.gda"},
	{0x1531AFAF63F8306AULL, "227065611.ogg"},
	{0x1531CC5A77D754B1ULL, "one120st_magistrate.stg"},
	{0x1531D5800C43BE2EULL, "vxw_bonepit_29e_0_0.vlm"},
	{0x1532081F5D3D7FE7ULL, "plt_cod_hst_dwarven_carta.plo"},
	{0x1532302C91894004ULL, "vxca_caverns1_mines_fo7_0_0.vlm"},
	{0x15329246340B5A21ULL, "vxlt_hangedman_136_0_0.vlm"},
	{0x1532C6E8B8967462ULL, "follower_banter6173095.cl"},
	{0x15333166F3EA4345ULL, "355644002.ogg"},
	{0x15335E5D80299F44ULL, "l_mcr310ar_docks_nt.bnk"},
	{0x15339C40C1CE5096ULL, "dae011_gossips6149775.cl"},
	{0x1533A84BFB1DA051ULL, "240578263.ogg"},
	{0x15360883D6A463DCULL, "llac_lt_0_lowtown_n_65535_8.dds"},
	{0x153686F948A9D013ULL, "6083812_f.fxe"},
	{0x1536950DAC1BD225ULL, "vxlt_gallowstemplar_4qp_0_0.vlm"},
	{0x15375C6B63D23B80ULL, "vxht_0_hightown_c5m_0_0.vlm"},
	{0x153788FEFF2BE4C5ULL, "w_blightlands_50_0.rml"},
	{0x15378B116455D734ULL, "vxw_encounter2_camp_om_0_0.vlm"},
	{0x15380A20B0435D56ULL, "llac_ht_keep_n_3736_1.dds"},
	{0x153822627625702BULL, "mcr310_loyalist_templars6076906.cl"},
	{0x153839878333D317ULL, "vxht_mansion2_1qt_0_0.vlm"},
	{0x1538F858B42B297AULL, "cs_qcr221cs_isabellas_return.wbd"},
	{0x1539B0885C0613A1ULL, "hlac_lt_0_lowtown_docks_a3_n_-262143_0x.dds"},
	{0x153AAAB0CC828B7BULL, "llac_lt_0_lowtown_docks_n_196607_1.dds"},
	{0x153ACA7A57A9D6BCULL, "141189871.ogg"},
	{0x153B5D72EEBAE7CAULL, "dae011_order_drink.cnv"},
	{0x153BB585F7690DC5ULL, "gen00fl_varric6087451.cl"},
	{0x153C2950849015D6ULL, "art\\characters\\weapons\\shields\\w_shd_006a.msh"},
	{0x153C3350849026D4ULL, "art\\characters\\weapons\\shields\\w_shd_006a.mmh"},
	{0x153C375084902DE7ULL, "art\\characters\\weapons\\shields\\w_shd_006a.mao"},
	{0x153C5D674B47F5C1ULL, "vxw_bonepitdestroyed_a3_ol_0_0.vlm"},
	{0x153C78932BE6B357ULL, "833635296.ogg"},
	{0x153C80552FEEF6E0ULL, "mage_bloodcontrol.dds"},
	{0x153C9EF8E5202656ULL, "vxht_bartrand_n_5y1_0_0.vlm"},
	{0x153CC5E31B0CFD46ULL, "135433737.ogg"},
	{0x153CF597895E15BCULL, "hlac_ca_cavern1_haunted_24426_10y.dds"},
	{0x153D04C6E45C28A3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_basic_1.mao"},
	{0x153D3A67D794AA6AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfallr_1.mao"},
	{0x153D688A6A5D54B7ULL, "and230_mage_girl6109104.cl"},
	{0x153D94A3511AEAF6ULL, "496697805.ogg"},
	{0x153DC9D40FEBA4A2ULL, "art\\vfx\\textures\\vfx_lightning.dds"},
	{0x153E3F72EEBCEAF2ULL, "dae011_order_drink.bnk"},
	{0x153E4EE556C7FEC1ULL, "llac_lt_0_lowtown_a2_n_c_-65536_1.dds"},
	{0x153E752548220713ULL, "vxca_cavern1_haunted_i1y_0_0.vlm"},
	{0x153E81F75F5B0572ULL, "vxht_keep_n_606_0_0.vlm"},
	{0x153E8A5F1CC194ABULL, "6167894_m.fxe"},
	{0x153E94F2F9172D78ULL, "496778205.ogg"},
	{0x153F9A89C0B28D6DULL, "vrc321_party_barks6095742.cl"},
	{0x153F9FC72A120625ULL, "hlac_w_sundermount_-393212_0y.dds"},
	{0x153FCBD2E8A8668FULL, "vxlt_0_lowtown_eml_0_0.vlm"},
	{0x153FD690F3DED1F4ULL, "dae000ip_crowd_lm2_talk17.utp"},
	{0x1540339B3A3AD5E4ULL, "dae300_bran6143487.cl"},
	{0x1540CB49C134428DULL, "hlac_lt_0_lowtown_a2_196606_0z.dds"},
	{0x1540FFE3A7E9B978ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_006a_n.dds"},
	{0x15410DCC6AF07DA8ULL, "w_bonepit_169_0.msh"},
	{0x154116E82798EAB6ULL, "vxlt_foundry_n_101_0_0.vlm"},
	{0x1541186294B1B2D1ULL, "6053184_m.fxe"},
	{0x15411FCC6AF09C39ULL, "w_bonepit_169_0.mao"},
	{0x15412594CDEEAF45ULL, "gen_im_arm_hlm_med_act3_02.uti"},
	{0x15412BCC6AF0B0D2ULL, "w_bonepit_169_0.mmh"},
	{0x15414A5417BFB438ULL, "mag320_orsino6071076.cl"},
	{0x154165AC8EA35174ULL, "405659139.ogg"},
	{0x1541FBBE764B92BBULL, "vxdr_primevaltomb_6vb_0_1.vlm"},
	{0x154269A34B5108BFULL, "6132768_m.fxe"},
	{0x154342E1D9BBF154ULL, "mag311_nyssa6042368.cl"},
	{0x154399EFB2CC94A0ULL, "mag100_karras6034749.cl"},
	{0x1543A6438434AFA4ULL, "vxdr_0_deeproads_bx8_0_0.vlm"},
	{0x1543F2F2A69D27A9ULL, "hlac_w_encounter1_65534_0y.dds"},
	{0x15440292F077C4CCULL, "gen00fl_isabela6099420.cl"},
	{0x15440D4029F2E97AULL, "qcr000_arishok6057093.cl"},
	{0x154451B9B8EDDC14ULL, "87709494.ogg"},
	{0x154455C977F9F9EAULL, "llac_ht_mansion2b_818_22.dds"},
	{0x15448D93ED3A9C67ULL, "mer340_cowan6083547.cl"},
	{0x1544DF659846A762ULL, "vxht_0_hightown_a3_b5c_0_0.vlm"},
	{0x1544EDE98B023A36ULL, "zz_qun_debug6085730.cl"},
	{0x15455C371C4D734FULL, "vxht_brothel_n_5cj_0_0.vlm"},
	{0x1545FCEE564BEFB4ULL, "vxw_encounter2_camp_jr_0_0.vlm"},
	{0x1546AB647529EB97ULL, "vxht_mansion2_1tw_0_0.vlm"},
	{0x1546C37880A26B83ULL, "vxdr_0_deeproads_4xm_0_0.vlm"},
	{0x1548031F32188F26ULL, "cd_dre000_cut_rock_wraith.wbd"},
	{0x154922DC9F68ACB7ULL, "6034831_m.fxe"},
	{0x15494286CBA511E2ULL, "plt_lgt310pt_fixer_1.nss"},
	{0x15495EE7FF605F7DULL, "vxht_playershouse_n_1ud_0_0.vlm"},
	{0x1549C36B9064BEBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward.ani"},
	{0x154B0D1D156692E3ULL, "vxlt_0_lowtown_a2_n_e80_0_0.vlm"},
	{0x154B505D5B415020ULL, "6122800_f.fxe"},
	{0x154B6CC7D6F05B73ULL, "583917420.ogg"},
	{0x154BD5B148234DB5ULL, "vxlt_0_lowtown_docks_a3_n_d0e_0_0.vlm"},
	{0x154BDD020DC9DA6FULL, "art\\vfx\\combat\\vfx_gib_qunc_ltrb.anb"},
	{0x154BFCD4487F24A8ULL, "6037370_m.fxe"},
	{0x154C73F4367F073CULL, "824508621.ogg"},
	{0x154CBC8631CA3981ULL, "6105795_m.fxe"},
	{0x154CF6690A591A42ULL, "77311081.ogg"},
	{0x154D22026D5C33B3ULL, "vxdr_primevalentrance_39k_0_0.vlm"},
	{0x154D22EEEBAC8B79ULL, "vxlt_hangedman_n_1mo_0_0.vlm"},
	{0x154DEC7A6FECC12EULL, "803116953.ogg"},
	{0x154DF42DA20C7970ULL, "llac_w_blightlands_0_7.dds"},
	{0x154E135E97BC49A7ULL, "vxlt_0_lowtown_a2_n_c_eov_0_1.vlm"},
	{0x154E1E44856DEC99ULL, "vxht_mansion2_3bm_0_0.vlm"},
	{0x154E3DAFDA9E7F0FULL, "vxlt_0_lowtown_eho_0_0.vlm"},
	{0x154ED6CAD48DBB67ULL, "mag210_arianni6031651.cl"},
	{0x154EF539DAEF3F00ULL, "51332602.ogg"},
	{0x154F8DD5DBF53D55ULL, "hla_ht_0_hightown_0_0y.dds"},
	{0x154FC4DE1FA11A8DULL, "418124740.ogg"},
	{0x15502C4A1EF5645DULL, "303392151.ogg"},
	{0x155072A89EB80965ULL, "171768007.ogg"},
	{0x1550A78714DB72F0ULL, "6042837_m.fxe"},
	{0x1551360CF755C707ULL, "vxht_mansion_n_621_0_0.vlm"},
	{0x1551564E7648B0A5ULL, "6174395_m.fxe"},
	{0x1551BAE0FE02D253ULL, "6111386_f.fxe"},
	{0x1551F6B5FF98B81FULL, "vxht_mansion2b_262_0_0.vlm"},
	{0x15520A0898C5C36AULL, "llac_lt_0_lowtown_a2_n_c_131074_0.dds"},
	{0x1552484DD9A956EFULL, "qcr000_arishok6165085.cl"},
	{0x1552598993AD6A63ULL, "vxca_caverns1_mines_j5n_0_0.vlm"},
	{0x155262DE35EEF3A1ULL, "vxlt_hangedman_2dl_0_0.vlm"},
	{0x155298E0F2FFEA15ULL, "llac_ht_brothel_n_3373_0.dds"},
	{0x1552A2503F0AC590ULL, "combat_2h_r1_a2.gda"},
	{0x155351428A3CBFE2ULL, "vxht_0_hightown_a3_b0l_0_0.vlm"},
	{0x1553C1E09A9A7AEDULL, "art\\characters\\weapons\\axes\\w_axe101a.mmh"},
	{0x1553D3E09A9A9973ULL, "art\\characters\\weapons\\axes\\w_axe101a.msh"},
	{0x15541C0AB7641BE5ULL, "6046876_m.fxe"},
	{0x15546ECB48420834ULL, "vxw_encounter2_camp_es_0_0.vlm"},
	{0x1554CEABE7250CC2ULL, "6155757_m.fxe"},
	{0x1554F12598EDCB5AULL, "295091138.ogg"},
	{0x155500E5052CA0CAULL, "vxht_0_hightown_n_9aj_0_0.vlm"},
	{0x15551D0A0EADF81BULL, "lt_0_lowtown_docks_n_219_0.msh"},
	{0x15552F0A0EAE1766ULL, "lt_0_lowtown_docks_n_219_0.mao"},
	{0x15553B0A0EAE2B05ULL, "lt_0_lowtown_docks_n_219_0.mmh"},
	{0x15559F7BDD9A6BDBULL, "mag121_idunna6083948.cl"},
	{0x1555BE0C3FB6DE1AULL, "vxca_caverns1_ij8_0_0.vlm"},
	{0x1555CC6CB6837432ULL, "672203314.ogg"},
	{0x1555CF602A45C8BBULL, "hlac_lt_backalley_n_1_7z.dds"},
	{0x15566702D31BDFB1ULL, "330369928.ogg"},
	{0x1556CB6B906F56B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward.evt"},
	{0x1556D9109A3560D1ULL, "vxht_brothel_n_6gg_0_0.vlm"},
	{0x15579844DD9BD8F3ULL, "485610988.ogg"},
	{0x1557C3D9F5BED318ULL, "vxlt_darktown_67y_0_0.vlm"},
	{0x1557EF42F1D59435ULL, "6169616_m.fxe"},
	{0x1558A092BBFD9AE6ULL, "114337083.ogg"},
	{0x1558BC91253A89B6ULL, "113424885.ogg"},
	{0x1558EB71F2FC06A6ULL, "6126503_f.fxe"},
	{0x15590837318231CBULL, "vxht_mansion2_1vi_0_0.vlm"},
	{0x1559176C9391A4C9ULL, "lla_ht_0_hightown_a2_65536_0.dds"},
	{0x15592083BC5C7A31ULL, "hlac_ht_0_hightown_a3_-65536_1z.dds"},
	{0x15592897C3D5431BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic_2.mao"},
	{0x15596C48E0184D37ULL, "fx_arrows.mao"},
	{0x1559A6E9A2FBFDABULL, "root_creeper.dds"},
	{0x1559D9BFC02F2524ULL, "llac_lt_undercitydungeon_-1_30.dds"},
	{0x1559EE2BA10ADEF8ULL, "ts_dwarven_prog_junk2.gda"},
	{0x155A3AF96949A9B7ULL, "463292865.ogg"},
	{0x155A65028135F472ULL, "mcr370_negotiation6078753.cl"},
	{0x155ADF7073ABDB4EULL, "mag100_karras6052952.cl"},
	{0x155B827E20485C5EULL, "dae000_flemeth6013332.cl"},
	{0x155B86CC9EAB1496ULL, "hightown_male_noble_gossip6171325.cl"},
	{0x155BF18AF1C36615ULL, "gen00fl_anders6103995.cl"},
	{0x155BF648EFB3A1B3ULL, "671471917.ogg"},
	{0x155C369C26B27CE7ULL, "6084227_m.fxe"},
	{0x155C8ADFBF2C38F8ULL, "cst_cloud.mao"},
	{0x155CC4E759BC1C33ULL, "cd_pro000ar_gallow_court.bnk"},
	{0x155CEAFB3F509569ULL, "ht_0_hightown_lc.mmh"},
	{0x155D1718533D1FC8ULL, "hlac_lt_0_lowtown_a3_n_c_-3_6y.dds"},
	{0x155D354E5118EC1DULL, "vxca_cavern1_haunted_jlk_0_0.vlm"},
	{0x155D47C23255CBF2ULL, "vxht_mansion_n_6kc_0_0.vlm"},
	{0x155D68FEFF4C7DEEULL, "w_blightlands_50_0.mmh"},
	{0x155D6A720E54DABDULL, "art\\levels\\races\\design\\trap_bolt.mao"},
	{0x155D6E720E54E1F6ULL, "art\\levels\\races\\design\\trap_bolt.mmh"},
	{0x155D74FEFF4C9185ULL, "w_blightlands_50_0.mao"},
	{0x155D78720E54F2F4ULL, "art\\levels\\races\\design\\trap_bolt.msh"},
	{0x155D82FEFF4CA9DCULL, "w_blightlands_50_0.msh"},
	{0x155D9536EB060E26ULL, "vxht_mansion2b_25n_0_0.vlm"},
	{0x155DEA6B9075AA45ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward.gad"},
	{0x155DF047B27C1763ULL, "lgt320_mis_resp_swordhunter6150277.cl"},
	{0x155E1A14C1B08A29ULL, "vxht_keep_70g_0_0.vlm"},
	{0x155E2E1A40549C82ULL, "vxlt_0_lowtown_a2_n_f9h_0_1.vlm"},
	{0x155F307616F73C2CULL, "745474391.ogg"},
	{0x155F68666C8EEE08ULL, "vxlt_0_lowtown_a2_gnt_0_0.vlm"},
	{0x155FD9AF05B7B191ULL, "6127298_m.fxe"},
	{0x156001C01B717E0AULL, "cod_bks_death_templar.utp"},
	{0x156001D47CADAE03ULL, "vxdr_0_deeproads_9iv_0_1.vlm"},
	{0x156050C25BF798F2ULL, "dae000_bartrand6206911.cl"},
	{0x1560F51F9C59EB8AULL, "dae000_friend_turn_in6179268.cl"},
	{0x15610BA3C62B6D83ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialg.mao"},
	{0x15610C85D075060CULL, "6044790_m.fxe"},
	{0x15610FA3C62B74D0ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialg.mmh"},
	{0x1561120271392106ULL, "lgt110_gxa_coa_spy6117404.cl"},
	{0x15611DA3C62B8CEAULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialg.msh"},
	{0x156127D386ABE78BULL, "hm_circlemage02.mor"},
	{0x1561B3BC89667AF3ULL, "vxht_keep_n_3c7_0_0.vlm"},
	{0x15620DD97DCF4FD2ULL, "628587732.ogg"},
	{0x15623FF10039A44FULL, "vxht_brothel_n_5it_0_0.vlm"},
	{0x156285331B936210ULL, "106944053.ogg"},
	{0x1562BD49E055DF40ULL, "vxca_caverns2prc_7px_0_0.vlm"},
	{0x1563318B9543EB13ULL, "art\\vfx\\creatures\\fxc_mrtorch_c.dura.ani"},
	{0x1563734C6BB1C0C6ULL, "hlac_dr_dungeon_3207_6z.dds"},
	{0x15638F1E59161C97ULL, "vxht_mansion2_1bq_0_0.vlm"},
	{0x15642FE931ACF69AULL, "vxca_caverns1_im9_0_0.vlm"},
	{0x156440752E479368ULL, "673989736.ogg"},
	{0x1564BAE09AA8E1AAULL, "art\\characters\\weapons\\axes\\w_axe101a.phy"},
	{0x1564CF1C4438C21CULL, "vxht_bartrand_n_5oa_0_0.vlm"},
	{0x1565D8817205116AULL, "vxlt_0_lowtown_docks_n_b5h_0_0.vlm"},
	{0x1565EE90CDBD0B83ULL, "6168680_m.fxe"},
	{0x15668765FEA14C2AULL, "art\\vfx\\cinematic\\vfx_ship_front_p.anb"},
	{0x1566A4194E6208A5ULL, "hlac_lt_0_lowtown_a2_-65536_1y.dds"},
	{0x156773B6DA5ACF44ULL, "6065812_m.fxe"},
	{0x15677A1423784A4BULL, "vxht_mansion2_1yh_0_0.vlm"},
	{0x15687DB0EF82F988ULL, "759388048.ogg"},
	{0x1569174A923388C1ULL, "mag130_danzig6018812.cl"},
	{0x156984549348A2B4ULL, "53636377.ogg"},
	{0x15699CE414E3351DULL, "meleering6.gda"},
	{0x1569A0EC670B415FULL, "cloth2.psh.fxo"},
	{0x1569E157298FD78DULL, "good.mao"},
	{0x156A06A8CF98BC79ULL, "vxlt_hangedman_n_1gm_0_0.vlm"},
	{0x156AC93ACFBEB015ULL, "6142199_f.fxe"},
	{0x156ACF25D475E2C5ULL, "hlac_ca_caverns2prc_5802_27x.dds"},
	{0x156AE1895EC37E63ULL, "amb111cs_bar_fight.cut"},
	{0x156B5DA4B6D2F32FULL, "vxlt_0_lowtown_by7_0_0.vlm"},
	{0x156BB99F244BE472ULL, "vxht_mansion_n_6nl_0_0.vlm"},
	{0x156CBF532C5F5508ULL, "gen_im_trs_slv_hvy_cht_r1.uti"},
	{0x156CF6E8299DC8DBULL, "459269543.ogg"},
	{0x156D1C979B3BEB2FULL, "284862197.ogg"},
	{0x156D245BBCE131CAULL, "6165040_m.fxe"},
	{0x156EC9F07DF34C81ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_elna_0d.dds"},
	{0x156EE0BD3575BB87ULL, "hlac_lt_gallowstemplar_2388_4x.dds"},
	{0x156EE6A316E03979ULL, "vxlt_0_lowtown_ep1_0_0.vlm"},
	{0x156F4E2755A58F6EULL, "mag101_aombanter6052747.cl"},
	{0x156F6B50C9C439D2ULL, "vxlt_0_lowtown_a2_hex_0_0.vlm"},
	{0x156F8FDEC3FFE298ULL, "6039404_m.fxe"},
	{0x156FA8ED00620EA0ULL, "365940902.ogg"},
	{0x156FCE6882337EECULL, "vxlt_undercityencounter_1bo_0_0.vlm"},
	{0x157024EE7D1E9A91ULL, "w_bonepit_9_0.mmh"},
	{0x15702BCCDABCB69BULL, "384630746.ogg"},
	{0x157030EE7D1EAF32ULL, "w_bonepit_9_0.mao"},
	{0x15703EEE7D1EC74FULL, "w_bonepit_9_0.msh"},
	{0x15708A996B6F1E6DULL, "gen_im_trs_slv_one_axe_r0.uti"},
	{0x1570F7EB846DF558ULL, "w_bonepitdestroyed_a3_173_0.phy"},
	{0x1571233BCCCA2541ULL, "6135745_m.fxe"},
	{0x1571298B95501B4AULL, "art\\vfx\\creatures\\fxc_mrtorch_c.dura.evt"},
	{0x15714281689A32ADULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_mushroompink_d.dds"},
	{0x15718ECE1352B77EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffb_d.dds"},
	{0x1572183CF1291C04ULL, "6047470_m.fxe"},
	{0x15723B6B24033026ULL, "cod_hst_orz_deeproads.utp"},
	{0x1572462003FAC988ULL, "vxlt_0_lowtown_a3_n_c_f6p_0_1.vlm"},
	{0x15735FD0477464A7ULL, "nat380_nathaniel_banter6067282.cl"},
	{0x15737AB65A770AC7ULL, "6093595_m.fxe"},
	{0x1573C665918584F8ULL, "gen00fl_aveline6116719.cl"},
	{0x1574451655CD3E39ULL, "6092554_m.fxe"},
	{0x1574B3250850068AULL, "vxlt_gallowsprison_n_6lm_0_0.vlm"},
	{0x1574BDE99BA9A59AULL, "ip_restrict_class.ncs"},
	{0x157502A65773D757ULL, "vxht_0_hightown_a2_aul_0_0.vlm"},
	{0x1575AF07FC5C5BD9ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsb_0d.dds"},
	{0x1575BE91C7BC1384ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngbn.msh"},
	{0x1575D091C7BC32EDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngbn.mao"},
	{0x1575D491C7BC3926ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngbn.mmh"},
	{0x1575F33C0D4DF691ULL, "586729631.ogg"},
	{0x1576780E6924EF83ULL, "107324990.ogg"},
	{0x1576B25A39204D78ULL, "171363810.ogg"},
	{0x1577341E3B961328ULL, "6207573_m.fxe"},
	{0x1577DCAE57125D43ULL, "art\\characters\\creatures\\strider\\animation\\walkfrontright.ani"},
	{0x1578088B955601BDULL, "art\\vfx\\creatures\\fxc_mrtorch_c.dura.gad"},
	{0x15781E7E464EA275ULL, "590737264.ogg"},
	{0x15786E7171FC8513ULL, "vxlt_0_lowtown_docks_b9n_0_0.vlm"},
	{0x15792F86CBCE1565ULL, "plt_lgt310pt_fixer_1.plo"},
	{0x15793EBD608E632FULL, "6129373_m.fxe"},
	{0x157954EDBCC1E7C5ULL, "6104600_m.fxe"},
	{0x15795890598BF329ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_reaction.ani"},
	{0x15797CC76AD7B970ULL, "178469632.ogg"},
	{0x1579ADB67AE7FD7BULL, "bla_dot.mao"},
	{0x1579ADD4EA3FB9FAULL, "mcr370_meredith6083088.cl"},
	{0x157A46F7BD1074C3ULL, "6116800_f.fxe"},
	{0x157AE0E32ABF1A2AULL, "mcr370_meredith6136133.cl"},
	{0x157B8AE4F4F48FE4ULL, "llac_lt_0_lowtown_docks_-196610_6.dds"},
	{0x157B912F5007F84BULL, "82479047.ogg"},
	{0x157BDE0338953C43ULL, "341219913.ogg"},
	{0x157BF665FD26399AULL, "vxlt_warehouse_qj_0_0.vlm"},
	{0x157C3D39A00D712CULL, "dae370ip_weapon_store.utp"},
	{0x157D328746B9FE7AULL, "6181749_m.fxe"},
	{0x157D8788981FD176ULL, "hlac_ca_caverns2_5802_34z.dds"},
	{0x157E1117C8FB3D3BULL, "hlac_lt_0_lowtown_docks_n_-1_0z.dds"},
	{0x157E8075967AB3FDULL, "ser221_gamlen6135169.cl"},
	{0x157E818845FCB23EULL, "853534068.ogg"},
	{0x157FA1CBDDA813F1ULL, "6092028_m.fxe"},
	{0x158072D83D024D97ULL, "vxht_mansion2_1hs_0_0.vlm"},
	{0x1580739059923FABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_reaction.gad"},
	{0x1580877881090CF3ULL, "lt_0_lowtown_n_24_0.rml"},
	{0x1581B8083FFF0840ULL, "dae000_hate_follower_react6103202.cl"},
	{0x15825A4E4B9FEAC0ULL, "vxca_caverns1_jl7_0_0.vlm"},
	{0x1583046623E64E95ULL, "lt_warehouse_n_5_0_lc.mmh"},
	{0x15831B46C393EE39ULL, "557999203.ogg"},
	{0x15832501FA461F0AULL, "vxlt_gallowsprison_n_6in_0_0.vlm"},
	{0x15832FA582018208ULL, "6153443_m.fxe"},
	{0x15834200CC546703ULL, "open_inventory_pressed.cur"},
	{0x15845DCE07647B4BULL, "vxht_mansion2_1gj_0_0.vlm"},
	{0x158509E01985B381ULL, "134765287.ogg"},
	{0x158588533A5AC6BBULL, "vxht_0_hightown_a3_a1i_0_0.vlm"},
	{0x1585F4AE571EC33AULL, "art\\characters\\creatures\\strider\\animation\\walkfrontright.evt"},
	{0x15861F572151E30AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxb.mao"},
	{0x158623572151EA59ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxb.mmh"},
	{0x15862D572151FB47ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxb.msh"},
	{0x158670905996A610ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_reaction.evt"},
	{0x158692D633B39F49ULL, "6095888_f.fxe"},
	{0x1586961CCC1AB2A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanmale_ambient_cheerwinning_2_ian.gad"},
	{0x15869F470CB6FFC9ULL, "ser121_jethann6164724.cl"},
	{0x1587BFFA5EE4248AULL, "892946288.ogg"},
	{0x1587F0C43712BDA4ULL, "6043171_m.fxe"},
	{0x15880214F0801A40ULL, "vxw_encounter2_camp_p4_0_0.vlm"},
	{0x1588166B541C2D9BULL, "111594767.ogg"},
	{0x1588F7A5F554494CULL, "818455153.ogg"},
	{0x15896CA2037885DAULL, "vxca_cavern1_haunted_k2s_0_0.vlm"},
	{0x158984872BF67A09ULL, "vxlt_0_lowtown_docks_n_b5v_0_1.vlm"},
	{0x158999FF713C9965ULL, "6182291_m.fxe"},
	{0x158A6842EF1C521AULL, "vxlt_warehouse_zk_0_0.vlm"},
	{0x158A813D4A86D050ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_06skeleton.msh"},
	{0x158A8F3D4A86E86AULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_06skeleton.mmh"},
	{0x158A933D4A86EEA1ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_06skeleton.mao"},
	{0x158BCA5CFACC6A79ULL, "vxlt_0_lowtown_ef7_0_0.vlm"},
	{0x158BF3AE57232D2DULL, "art\\characters\\creatures\\strider\\animation\\walkfrontright.gad"},
	{0x158BF97CED578ECDULL, "vxlt_undercitydungeon_wu_0_0.vlm"},
	{0x158C31316299CC57ULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_lp.gad"},
	{0x158C3338092420E1ULL, "6103564_m.fxe"},
	{0x158C6B32BC57288CULL, "vxlt_0_lowtown_docks_a3_n_cyh_0_0.vlm"},
	{0x158C77C6E8214884ULL, "924584875.ogg"},
	{0x158CBB42DFB9F666ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down.ani"},
	{0x158D43D3D1519832ULL, "zz_followers_banter6158577.cl"},
	{0x158D4CDF81DE0ECCULL, "vxdr_primevalentrance_19v_0_0.vlm"},
	{0x158D976C9AF56D07ULL, "qcr000_varric6208249.cl"},
	{0x158E811743415E90ULL, "vxlt_shop_d4_0_0.vlm"},
	{0x158E8C768A8AF3D3ULL, "hlac_lt_0_lowtown_n_-65537_5y.dds"},
	{0x158ED75C57F63CA0ULL, "vxw_bonepitdestroyed_a3_248_0_0.vlm"},
	{0x158EF1BEDC612B88ULL, "vrc321_varric6087125.cl"},
	{0x158F48C041DC4D6CULL, "art\\characters\\creatures\\crowds\\elfmale\\textures\\prp_hightown_jimmy_dp.dds"},
	{0x158F5F7BD14A3E93ULL, "vxlt_0_lowtown_d5u_0_0.vlm"},
	{0x158F8A65FEC4561DULL, "art\\vfx\\cinematic\\vfx_ship_front_p.mmh"},
	{0x158FC47054E622AAULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magicloker_d.dds"},
	{0x158FEB93BD0065D7ULL, "398195774.ogg"},
	{0x158FFA255952F082ULL, "vxca_caverns2_6t6_0_0.vlm"},
	{0x1590722D133BA3CEULL, "crate_dust.mao"},
	{0x1590B7DD455AE69CULL, "6166209_m.fxe"},
	{0x1590D830DD160FA9ULL, "mrl150_merrill6070498.cl"},
	{0x159165CD4099D2A1ULL, "w_bonepitmining_a2_24_0.rml"},
	{0x15916EBA698EE39BULL, "79836567.ogg"},
	{0x159196DEEC3C378AULL, "vxlt_gallowsprison_n_6ro_0_0.vlm"},
	{0x15919ACF4D16CB85ULL, "6176476_m.fxe"},
	{0x1591D5984A555C8BULL, "vxca_caverns2prc_9u6_0_0.vlm"},
	{0x1591F179190A1318ULL, "hlac_ht_0_hightown_n_-65535_1z.dds"},
	{0x1592688FEC674BE0ULL, "263344451.ogg"},
	{0x159273985997C2F1ULL, "791161437.ogg"},
	{0x15927E219FE8A9BAULL, "226028626.ogg"},
	{0x15927F8000873595ULL, "hlac_lt_0_lowtown_65536_5x.dds"},
	{0x1592AA235CF57454ULL, "qun211_braggart6075702.cl"},
	{0x1592AAF744886DE6ULL, "vxw_bonepitdestroyed_a3_27g_0_0.vlm"},
	{0x1593463162A00EE4ULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_lp.evt"},
	{0x15938AC5959DC54DULL, "zz_vault_debug6178673.cl"},
	{0x159542C70DAE03D5ULL, "zz_one_debug6075013.cl"},
	{0x159673F1E27632C0ULL, "vxw_encounter2_camp_k5_0_0.vlm"},
	{0x159678863BB2F98CULL, "seb221_flora_harimann6177679.cl"},
	{0x1596ACF18511AE91ULL, "zz_vrc_debug6064296.cl"},
	{0x1596E4BAFB4E3499ULL, "6076956_m.fxe"},
	{0x159714B59C14FC51ULL, "1136587.ogg"},
	{0x1597622BAB377ED9ULL, "877070141.ogg"},
	{0x159765CEC83FF859ULL, "llac_lt_gallowstemplar_f_2384_7.dds"},
	{0x1598429621E62C34ULL, "qun160_vashothlead6064636.cl"},
	{0x15986DB200663121ULL, "dae000_arishok6086036.cl"},
	{0x159871D714CCD8B3ULL, "follower_banter6170466.cl"},
	{0x1598AEADB0CF2C48ULL, "204841782.ogg"},
	{0x15990C3B9E01C4DDULL, "pro000_friedrich6051036.cl"},
	{0x15991153292EDA70ULL, "6013310_m.fxe"},
	{0x1599317A7D6BEC5AULL, "6153508_m.fxe"},
	{0x15993C323A27F511ULL, "vxlt_warehouse_n_104_0_0.vlm"},
	{0x1599535CA6B780D4ULL, "hlac_lt_undercityserialkiller_-1_6x.dds"},
	{0x15995E6BF70DA71AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomea_sa.dds"},
	{0x15998F32492DFEC2ULL, "817945064.ogg"},
	{0x15999C7641F7CB64ULL, "6151269_m.fxe"},
	{0x1599D5D4D799C15EULL, "gen_im_trs_apo_sta_fir_r1.uti"},
	{0x159A34448CE29066ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_act_c.impa.ani"},
	{0x159A40A760EBC5C4ULL, "vxht_bartrand_n_2ny_0_0.vlm"},
	{0x159AA342DFC60AD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down.evt"},
	{0x159AAB8E282177A4ULL, "76870015.ogg"},
	{0x159AC11CCC2BA488ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanmale_ambient_cheerwinning_2_ian.ani"},
	{0x159ADA551578BB39ULL, "6133376_m.fxe"},
	{0x159B7FBB7F33EF33ULL, "vrc321_varric6084442.cl"},
	{0x159B8AF1769840DEULL, "648897321.ogg"},
	{0x159B91ABDE61F2C6ULL, "t1_tat_pnk.tnt"},
	{0x159C0E9C31EF254EULL, "dae_rogue_modal.ncs"},
	{0x159C245BC621E006ULL, "follower_banter6170696.cl"},
	{0x159C687E6749EE0AULL, "vxht_mansion_n_4vn_0_0.vlm"},
	{0x159C7CC972120924ULL, "6164148_m.fxe"},
	{0x159CA6DE2718B01BULL, "crt_cycle_2.dds"},
	{0x159CCD888E51A74BULL, "143167728.ogg"},
	{0x159CD2FAB9192ED4ULL, "hlac_ca_caverns2prc_5802_0x.dds"},
	{0x159D2ACC6B3F0DCDULL, "w_bonepit_169_0.phy"},
	{0x159D466697186B6EULL, "13135928.ogg"},
	{0x159D83AB46708B46ULL, "313247626.ogg"},
	{0x159D9BB691F3D44FULL, "6114416_m.fxe"},
	{0x159DA9EE1C1C6E0BULL, "hlac_ca_cave_2194_2z.dds"},
	{0x159DE38CC7FBB595ULL, "224928479.ogg"},
	{0x159DF57375FB1FCEULL, "hlac_lt_0_lowtown_n_65536_3x.dds"},
	{0x159E1D8F4C97FF5CULL, "vxht_0_hightown_c51_0_0.vlm"},
	{0x159E273137E910D8ULL, "vxca_caverns1_mines_j4r_0_0.vlm"},
	{0x159E4AF7DCA8C7A1ULL, "vxw_encounter2_camp_r3_0_0.vlm"},
	{0x159E699B1F85A8D4ULL, "vxht_brothel_n_4wh_0_0.vlm"},
	{0x159EA2AB70953E02ULL, "455918833.ogg"},
	{0x159ED84E856C4E54ULL, "6016916_f.fxe"},
	{0x159F09C0275D85C8ULL, "zz_amb_test6105060.cl"},
	{0x159F3C1FBB0A8794ULL, "og_flsh_01.dds"},
	{0x159F4317F004D27FULL, "6077049_m.fxe"},
	{0x15A0092E6D63B063ULL, "120387130.ogg"},
	{0x15A14187EB50AC4BULL, "vxht_mansion2_1ml_0_0.vlm"},
	{0x15A15E3162AC74DDULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_lp.ani"},
	{0x15A1A442DFCC2B60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down.gad"},
	{0x15A2541AC2210C14ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_dsmca_0n.dds"},
	{0x15A256B35A6CF4F6ULL, "eyecr_wyverna.utc"},
	{0x15A36D11E2FBDE89ULL, "gencr_spider_monstrous.utc"},
	{0x15A3E2E759F85191ULL, "cd_pro000ar_gallow_court.wbd"},
	{0x15A439E6359271AAULL, "narrative_banter6136836.cl"},
	{0x15A4554E466018FAULL, "6051839_m.fxe"},
	{0x15A4C054D813FFBFULL, "zz_prima.cut"},
	{0x15A4E5CED46C4B40ULL, "vxw_encounter2_camp_f2_0_0.vlm"},
	{0x15A588EDF47523A5ULL, "mag121_idunna6084859.cl"},
	{0x15A5BE59C3BCE5E1ULL, "6173336_m.fxe"},
	{0x15A62505D7B7ED6DULL, "98096223.ogg"},
	{0x15A64E25281ECA75ULL, "vxht_keep_n_735_0_0.vlm"},
	{0x15A685B7716CA9CEULL, "31601126.ogg"},
	{0x15A6B3F828CB4D91ULL, "sky_sky2_p0.vsh.fxo"},
	{0x15A6F5B9A8C6E939ULL, "222883363.ogg"},
	{0x15A703C2C2286512ULL, "vxw_woundedcoast_87_0_0.vlm"},
	{0x15A71F1C04F18266ULL, "vxlt_0_lowtown_a2_f1j_0_0.vlm"},
	{0x15A788DBA58CA98DULL, "vxlt_darktown_6l9_0_0.vlm"},
	{0x15A7AC811D1941CFULL, "llac_lt_gallowscourtyard_n_-65538_0.dds"},
	{0x15A7B6326FC53590ULL, "6148692_m.fxe"},
	{0x15A7C5832A0CAF3AULL, "hf_refugee01.mor"},
	{0x15A7FE79FAEC1649ULL, "6167500_m.fxe"},
	{0x15A81C448CEEA4D7ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_act_c.impa.evt"},
	{0x15A85C56985B0B6DULL, "vxht_mansion2_25i_0_0.vlm"},
	{0x15A8768A3D350E07ULL, "llac_dr_slavers_3216_12.dds"},
	{0x15A87B7A56276A9EULL, "zz_qun_debug6028289.cl"},
	{0x15A8DE69F37C0D68ULL, "vxlt_0_lowtown_a3_n_c_ci5_0_0.vlm"},
	{0x15AA28CE6061C88DULL, "vxca_cavern1_haunted_gh2_0_0.vlm"},
	{0x15AADD1F9C247747ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firecoal_n.dds"},
	{0x15AC00C7390FCF06ULL, "dm_lgt_harrowmont.mor"},
	{0x15AC26E3A4E50627ULL, "hlac_lt_0_lowtown_a2_n_131071_1y.dds"},
	{0x15AC696A3E19385DULL, "vxlt_0_lowtown_docks_a3_d32_0_1.vlm"},
	{0x15AC990E29DF2958ULL, "vxca_caverns1_mines_j3s_0_0.vlm"},
	{0x15AC9C0DFFCB31A8ULL, "46685764.ogg"},
	{0x15ACBCD4CE9EE021ULL, "vxw_encounter2_camp_m2_0_0.vlm"},
	{0x15AD57B4FABDA2B9ULL, "gen00fl_aveline6122135.cl"},
	{0x15ADBF54A716B3BEULL, "sto_cl_white.mao"},
	{0x15ADE22F3F81DD67ULL, "hlac_lt_0_lowtown_docks_a3_-131071_0z.dds"},
	{0x15AE00F249A9E101ULL, "142700483.ogg"},
	{0x15AECE9626746587ULL, "6085880_meredith_surgery.cut"},
	{0x15AF1D448CF4C560ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_act_c.impa.gad"},
	{0x15AF693A4C2BFF46ULL, "dae300_bran6143085.cl"},
	{0x15AFE4206A500718ULL, "art\\vfx\\materials\\decal\\vfx_pd_blood_human.mao"},
	{0x15AFECD3D4A47644ULL, "vxht_mansion2_3dj_0_0.vlm"},
	{0x15B04D0C1C4B4D4AULL, "dae000cr_grand_cleric6124723.cl"},
	{0x15B0BE68C111183FULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_001a_0t.dds"},
	{0x15B0F4FCBBC5597BULL, "lt_0_lowtown_docks_n_38_0.rml"},
	{0x15B10FA3DE1FCF06ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyo_0d.dds"},
	{0x15B1CDA19CB57963ULL, "vxlt_warehouse_ag_0_0.vlm"},
	{0x15B1E701AFF6CFADULL, "6136366_m.fxe"},
	{0x15B3709E180CF845ULL, "6119448_m.fxe"},
	{0x15B38395A0102E19ULL, "vxlt_0_lowtown_a2_e1j_0_0.vlm"},
	{0x15B411DA3A9361FFULL, "vxht_keep_n_63n_0_0.vlm"},
	{0x15B4A6BFCF310248ULL, "dae000_gamlen6166383.cl"},
	{0x15B4BFA50B2FD0EFULL, "6080685_m.fxe"},
	{0x15B5583BCE1F7AAEULL, "vxca_caverns1_gh8_0_0.vlm"},
	{0x15B595A65AC2DACFULL, "1028220.ogg"},
	{0x15B618751518180BULL, "vxlt_0_lowtown_docks_bxe_0_0.vlm"},
	{0x15B62716A42F9964ULL, "vxht_mansion2_1sm_0_0.vlm"},
	{0x15B6431FBB1D72C2ULL, "og_flsh_01.mao"},
	{0x15B7303AD09145A3ULL, "6119147_m.fxe"},
	{0x15B797E610C4B999ULL, "6064065_m.fxe"},
	{0x15B7B1D54B347DFDULL, "20837582.ogg"},
	{0x15B80F54EA68B9EEULL, "6070446_f.fxe"},
	{0x15B842D5FAFA9013ULL, "6136650_m.fxe"},
	{0x15B8675E97AB2695ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut2flat.phy"},
	{0x15B86C99D092CD3DULL, "vxht_mansion_n_60i_0_0.vlm"},
	{0x15B931C1D1762861ULL, "hla_ht_0_hightown_n_-2_0y.dds"},
	{0x15BAD36EB9C72639ULL, "479960581.ogg"},
	{0x15BAF34BA5684630ULL, "vxht_brothel_n_5kq_0_0.vlm"},
	{0x15BB2EB1C094F8A1ULL, "vxw_encounter2_camp_h1_0_0.vlm"},
	{0x15BB55F893AF5AAEULL, "fxe_ripple_add.mao"},
	{0x15BB75790FB5FBC1ULL, "6102813_m.fxe"},
	{0x15BBBE0290AEB07AULL, "hlac_w_encounter2_camp_131076_0z.dds"},
	{0x15BC3F4F1D3E7959ULL, "art\\characters\\creatures\\mubari\\animation\\growl.ani"},
	{0x15BC9BCD6ACCE8CAULL, "art\\levels\\races\\design\\prp_embrium.mao"},
	{0x15BC9BDB4B24DEC9ULL, "hlac_lt_0_lowtown_n_65535_5x.dds"},
	{0x15BC9FCD6ACCEF19ULL, "art\\levels\\races\\design\\prp_embrium.mmh"},
	{0x15BCA9CD6ACD0007ULL, "art\\levels\\races\\design\\prp_embrium.msh"},
	{0x15BD1E4F6A57A2CBULL, "797685126.ogg"},
	{0x15BD464F32FB1368ULL, "6153883_m.fxe"},
	{0x15BDB6C121E69C15ULL, "vxw_encounter2_camp_d9_0_0.vlm"},
	{0x15BE49C092905DB7ULL, "450175825.ogg"},
	{0x15BEE39E4F2BF514ULL, "vxw_woundedcoast_1dd_0_0.vlm"},
	{0x15BF7EFB71719437ULL, "15975206.ogg"},
	{0x15C01051DA7A1D18ULL, "vxw_sundermount_31l_0_1.vlm"},
	{0x15C054417D1206DAULL, "6087058_m.fxe"},
	{0x15C0D3C06BCE7586ULL, "93356536.ogg"},
	{0x15C0E7AB5596896EULL, "vxlt_gallowscourtyard_72w_0_0.vlm"},
	{0x15C0FE0ADF4B8AD2ULL, "gen00fl_aveline6113677.cl"},
	{0x15C16D3D169903F7ULL, "879037337.ogg"},
	{0x15C24EA54F6C5494ULL, "6111993_m.fxe"},
	{0x15C26B116C0EBA4DULL, "dae000ip_crowd_ff8_talk6.utp"},
	{0x15C2CFC6A8550B01ULL, "6041514_m.fxe"},
	{0x15C30BC32AADF1E0ULL, "vxw_sundermount_2z0_0_1.vlm"},
	{0x15C31A4F1D44589BULL, "art\\characters\\creatures\\mubari\\animation\\growl.gad"},
	{0x15C32B45CFBED5C9ULL, "695264892.ogg"},
	{0x15C331BE2C2C77F7ULL, "ht_0_hightown_n_mmi.gff"},
	{0x15C3896B1A96AA11ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_elna_0t.dds"},
	{0x15C3946B1DCCAA19ULL, "vxw_encounter3_em_0_0.vlm"},
	{0x15C43AEA0220397FULL, "vxlt_0_lowtown_bxp_0_0.vlm"},
	{0x15C45EB793059897ULL, "vxht_bartrand_n_5s2_0_0.vlm"},
	{0x15C4662434CAA2B6ULL, "dae300_alistair_king6133781.cl"},
	{0x15C4ED9928059926ULL, "6036744_m.fxe"},
	{0x15C5102C9D19D586ULL, "dae000_serendipity6161407.cl"},
	{0x15C53208A38D4B30ULL, "hlac_lt_0_lowtown_docks_-65539_0y.dds"},
	{0x15C5686BAB8746A9ULL, "vxw_encounter1_es_0_0.vlm"},
	{0x15C569A5CE4DBAC8ULL, "hlac_lt_0_lowtown_n_-131073_0z.dds"},
	{0x15C61D61E533BDD2ULL, "hlac_w_bonepitdestroyed_a3_131078_0x.dds"},
	{0x15C64DC44F580FD2ULL, "dae120cr_stat_guard.utc"},
	{0x15C6664F81848A25ULL, "follower_banter6141910.cl"},
	{0x15C67624B60689A8ULL, "6139393_m.fxe"},
	{0x15C6A2A5B4C9E19FULL, "vxlt_gallowstemplar_f_6s4_0_1.vlm"},
	{0x15C6A38AD64BCC5EULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_statue_c.mmh"},
	{0x15C6DEE5BA784B05ULL, "one200_gold_followers6120153.cl"},
	{0x15C7F1E7CA19064EULL, "art\\characters\\playercharacter\\humanmale\\rob_magc.msh"},
	{0x15C7FFE7CA191D9FULL, "art\\characters\\playercharacter\\humanmale\\rob_magc.mao"},
	{0x15C80BE7CA19323CULL, "art\\characters\\playercharacter\\humanmale\\rob_magc.mmh"},
	{0x15C81C0534D184DCULL, "54033674.ogg"},
	{0x15C84196039E7202ULL, "hlac_ht_0_hightown_-131071_1z.dds"},
	{0x15C8C140F42E3551ULL, "llac_dr_dungeon_3214_10.dds"},
	{0x15C95076AE0F38B5ULL, "6117785_m.fxe"},
	{0x15C99448E828DE07ULL, "vxca_cave_1dc_0_0.vlm"},
	{0x15C9A08EB28B1121ULL, "vxw_encounter2_camp_c0_0_0.vlm"},
	{0x15C9A24280B52F9AULL, "dae200cr_viscount6087244.cl"},
	{0x15C9B3A802E0D50FULL, "vxca_cave_14k_0_0.vlm"},
	{0x15C9B90247CA8431ULL, "6207442_m.fxe"},
	{0x15CA374F1D4AA880ULL, "art\\characters\\creatures\\mubari\\animation\\growl.evt"},
	{0x15CA9AC33B898FCBULL, "6096307_m.fxe"},
	{0x15CAF9FB187273B2ULL, "6139688_f.fxe"},
	{0x15CB5DF46E6981CFULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.impa.gad"},
	{0x15CB8E7CA7C12C99ULL, "vxlt_gallowstemplar_n_47a_0_0.vlm"},
	{0x15CC94814E2F90AEULL, "6161040_m.fxe"},
	{0x15CCBFF5E6807C5FULL, "vxca_caverns1_mines_e7w_0_0.vlm"},
	{0x15CCDAF5468BE097ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_basic.phy"},
	{0x15CCDB249ED942DBULL, "6120486_m.fxe"},
	{0x15CCE4EB84BC33EDULL, "w_bonepitdestroyed_a3_173_0.msh"},
	{0x15CCF2EB84BC4B20ULL, "w_bonepitdestroyed_a3_173_0.mao"},
	{0x15CCF6EB84BC5273ULL, "w_bonepitdestroyed_a3_173_0.mmh"},
	{0x15CD281525CB6333ULL, "mag291_torpor6025987.cl"},
	{0x15CD557B41220D94ULL, "vxw_woundedcoast_1oe_0_0.vlm"},
	{0x15CD6DB12B85ED05ULL, "231462528.ogg"},
	{0x15CD7B0786DB04D7ULL, "599473068.ogg"},
	{0x15CE3F68FD395D80ULL, "hlac_ht_bartrand_n_3374_0z.dds"},
	{0x15CE47DA09AE1CFBULL, "x_narrativeroom_0.rml"},
	{0x15CEB0A0E1DA2749ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.dura.ani"},
	{0x15CF13B6885E78B6ULL, "art\\characters\\creatures\\strider\\animation\\turnleft30.ani"},
	{0x15CF4B6799DFF87FULL, "mer221_sabin6044765.cl"},
	{0x15CF61662946528FULL, "6172786_m.fxe"},
	{0x15CFBDD6470A56F3ULL, "one280_merin6104372.cl"},
	{0x15CFD36422DB13D3ULL, "root_marksman.dds"},
	{0x15D066AADD751183ULL, "221278968.ogg"},
	{0x15D08B7BFFCFE8F2ULL, "6149217_m.fxe"},
	{0x15D0A4A203A867F5ULL, "art\\characters\\playercharacter\\humanmale\\boo_masf.phy"},
	{0x15D10DA60DC1F67CULL, "6001177_m.fxe"},
	{0x15D14F5C03BDDCE7ULL, "6142503_m.fxe"},
	{0x15D157CD40CF9E90ULL, "w_bonepitmining_a2_24_0.msh"},
	{0x15D163FE366B31E3ULL, "vxdr_primevaltomb_2wr_0_2.vlm"},
	{0x15D165CD40CFB6AAULL, "w_bonepitmining_a2_24_0.mmh"},
	{0x15D169CD40CFBDE1ULL, "w_bonepitmining_a2_24_0.mao"},
	{0x15D182A2AECFFE82ULL, "6117812_m.fxe"},
	{0x15D1AF70DAC06C4DULL, "vxlt_hangedman_n_2i4_0_0.vlm"},
	{0x15D1E57DAC9D9A10ULL, "203642908.ogg"},
	{0x15D25154FD29A565ULL, "vxw_encounter2_camp_nk_0_0.vlm"},
	{0x15D272F46E6FC47CULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.impa.evt"},
	{0x15D297702688B91EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sluce_sa.dds"},
	{0x15D325AC23EAB578ULL, "6141578_m.fxe"},
	{0x15D3828AD6561E1DULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_statue_c.anb"},
	{0x15D3CC2C333C36DAULL, "vxca_cave_21z_0_0.vlm"},
	{0x15D3DA489D7D5F29ULL, "vxw_encounter1_jr_0_0.vlm"},
	{0x15D3E134D66CFB99ULL, "vxht_keep_6ay_0_0.vlm"},
	{0x15D40845272A4C42ULL, "hlac_lt_0_lowtown_n_65537_1y.dds"},
	{0x15D413F4B47D76A3ULL, "6113747_m.fxe"},
	{0x15D47C999FAC9AA9ULL, "vxdr_primevaltomb_9dt_0_0.vlm"},
	{0x15D47F0922C02FB0ULL, "vxlt_0_lowtown_docks_b6d_0_0.vlm"},
	{0x15D4CBA0E1DEC0CBULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.dura.gad"},
	{0x15D4DEFE9962C599ULL, "6165366_m.fxe"},
	{0x15D4E689DC31713DULL, "6108323_m.fxe"},
	{0x15D54157ACE84B56ULL, "gen00fl_merrill6151003.cl"},
	{0x15D5703392785DF9ULL, "hlac_w_sundermount_0_7y.dds"},
	{0x15D5BA8A304A6D83ULL, "vxlt_gallowstemplar_f_3rk_0_0.vlm"},
	{0x15D5DDA8C56E00FEULL, "6187916_m.fxe"},
	{0x15D689D05A3B1844ULL, "mag270cr_recruit01.utc"},
	{0x15D6ECE700A83493ULL, "473762534.ogg"},
	{0x15D71EC8A9C85AD3ULL, "vxlt_0_lowtown_docks_95i_0_0.vlm"},
	{0x15D72C8983D6A6F0ULL, "mag170_thrask6020557.cl"},
	{0x15D74D569A9E8E9AULL, "character_character_p142.psh.fxo.dx11"},
	{0x15D7F7A8C79C1832ULL, "lgt250_way_dalish_conflict6147128.cl"},
	{0x15D89A011C6C72DBULL, "6064958_f.fxe"},
	{0x15D900EA97583DB2ULL, "6144214_m.fxe"},
	{0x15D92F54CDABCDCDULL, "287239217.ogg"},
	{0x15D93A29481DDAC4ULL, "and111_lirene6090216.cl"},
	{0x15D993E383BFDC7EULL, "vxlt_undercitydungeon_1c6_0_1.vlm"},
	{0x15D9E799A356C6ACULL, "gen00fl_aveline6152571.cl"},
	{0x15DA1FBE4E723A4AULL, "vxht_keep_n_6n0_0_0.vlm"},
	{0x15DA79B2E2E96897ULL, "sst_ss_slavers_m_1.gda"},
	{0x15DAA28A0DEE27D9ULL, "vxht_mansion2b_1ta_0_0.vlm"},
	{0x15DB1FC4FEB3ECFAULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_lp.gad"},
	{0x15DB3CE72427B0A5ULL, "vxht_0_hightown_c2y_0_0.vlm"},
	{0x15DB5E35E5AD14BAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avea_1.mao"},
	{0x15DB61270335E8ADULL, "28576205.ogg"},
	{0x15DBC75833182614ULL, "vxw_woundedcoast_1jj_0_0.vlm"},
	{0x15DBC8A0E1E4DA30ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.dura.evt"},
	{0x15DBF23D1AD8D9D1ULL, "vxw_blightlands_bt_0_0.vlm"},
	{0x15DD166257FCB057ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_lsw_reva_0d.dds"},
	{0x15DD1BB6886AC3A7ULL, "art\\characters\\creatures\\strider\\animation\\turnleft30.evt"},
	{0x15DD23387297C2E3ULL, "vxlt_warehouse_ph_0_0.vlm"},
	{0x15DD781D37E1EDF6ULL, "74404863.ogg"},
	{0x15DDAE1C95C78DE4ULL, "w_sundermount_30_0.msh"},
	{0x15DDC01C95C7AC4DULL, "w_sundermount_30_0.mao"},
	{0x15DDC41C95C7B286ULL, "w_sundermount_30_0.mmh"},
	{0x15DE6BB9EBDD68BBULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs103a.msh"},
	{0x15DE7DB9EBDD8706ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs103a.mao"},
	{0x15DE89B9EBDD9BA5ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs103a.mmh"},
	{0x15DEDADB19F38673ULL, "fer_ter_rock_001.mao"},
	{0x15DEF82423701BEFULL, "vxw_sundermount_31v_0_0.vlm"},
	{0x15DF17BDD501E07BULL, "6149627_m.fxe"},
	{0x15DF4E4F876F0386ULL, "6047383_f.fxe"},
	{0x15DF62B56208CA2DULL, "6151242_m.fxe"},
	{0x15DF7AF46E7A5B95ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.impa.ani"},
	{0x15DF8F6B47B62265ULL, "vxca_cavern1_haunted_k6t_0_0.vlm"},
	{0x15E0269A64239878ULL, "351644836.ogg"},
	{0x15E053C17289EB14ULL, "6059965_m.fxe"},
	{0x15E078DD1C4F6FABULL, "1055440974.ogg"},
	{0x15E0C331EF1FBDE5ULL, "vxw_encounter2_camp_ij_0_0.vlm"},
	{0x15E0C5C34A2269D9ULL, "sg7_blur_dot.mao"},
	{0x15E0CCF305773D89ULL, "6129128_m.fxe"},
	{0x15E122C4FEB85D11ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_lp.evt"},
	{0x15E1640E4DDF6539ULL, "vxw_bonepitdestroyed_a3_1xc_0_1.vlm"},
	{0x15E18381791FF240ULL, "91466347.ogg"},
	{0x15E1D54A02733028ULL, "334842354.ogg"},
	{0x15E219F1930FC5ADULL, "6164967_m.fxe"},
	{0x15E27E622ACE314BULL, "220788464.ogg"},
	{0x15E2FF1DD014C4D6ULL, "w_bonepitmining_a2_172_0.msh"},
	{0x15E301082ADBFBC0ULL, "ave100ar_patrol.are"},
	{0x15E305FAA07DADB9ULL, "184898067.ogg"},
	{0x15E30681B18A941EULL, "vxdr_slavers_14g_0_0.vlm"},
	{0x15E3091DD014D5D4ULL, "w_bonepitmining_a2_172_0.mmh"},
	{0x15E30D1DD014DCE7ULL, "w_bonepitmining_a2_172_0.mao"},
	{0x15E336B07F3D1B8FULL, "vxlt_undercitydungeon_12y_0_0.vlm"},
	{0x15E33CB6886F67F0ULL, "art\\characters\\creatures\\strider\\animation\\turnleft30.gad"},
	{0x15E3DB478221F7CEULL, "vxht_mansion_n_6jl_0_0.vlm"},
	{0x15E3F2DB0E86A6BEULL, "76579394.ogg"},
	{0x15E4266D029F188BULL, "vxlt_gallowscourtyard_7ec_0_0.vlm"},
	{0x15E4EC6669367D3BULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.cess.ani"},
	{0x15E52AFD1D6C29E3ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.cess.ani"},
	{0x15E55088BC895F4AULL, "6019013_m.fxe"},
	{0x15E557A5CC27359AULL, "vxw_encounter3_3t_0_0.vlm"},
	{0x15E59BA3D184AFEEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_chest_out.ani"},
	{0x15E5DD1374024F37ULL, "llac_lt_gallowscourtyard_-65535_0.dds"},
	{0x15E5F8082ADE2242ULL, "ave100ar_patrol.ncs"},
	{0x15E68D65E5187038ULL, "vxca_caverns2_8kj_0_1.vlm"},
	{0x15E6B1FFA517D5A1ULL, "vxdr_slavers_19s_0_0.vlm"},
	{0x15E6C27D2BB5E91BULL, "263282488.ogg"},
	{0x15E6FB4E8E7922B1ULL, "vxlt_hangedman_n_1jv_0_0.vlm"},
	{0x15E73893488D5BACULL, "hlac_lt_0_lowtown_docks_a3_65533_0y.dds"},
	{0x15E87D45EC96582DULL, "dae000_gamlen6145093.cl"},
	{0x15E8919B406852CAULL, "vxht_keep_n_6i1_0_0.vlm"},
	{0x15E8A30A90384475ULL, "ser191_emeric6108784.cl"},
	{0x15E9108BC9C80579ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_justbuttress.msh"},
	{0x15E91A8BC9C81667ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_justbuttress.mmh"},
	{0x15E91E8BC9C81CB4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_justbuttress.mao"},
	{0x15E9245721A66946ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxb.phy"},
	{0x15E95351FA77180EULL, "6083072_m.fxe"},
	{0x15E9AA32B9C36017ULL, "6116821_m.fxe"},
	{0x15EAC3AFD45E2C9AULL, "genip_door_dwarven_normal.utp"},
	{0x15EB1DC3EBC279F8ULL, "6137531_m.fxe"},
	{0x15EB2118856855B4ULL, "6084480_m.fxe"},
	{0x15EB9515648DDB63ULL, "vxlt_warehouse_uk_0_0.vlm"},
	{0x15EBD1F159CD7096ULL, "vxht_chantry_32f_0_0.vlm"},
	{0x15EBD32B1C7167F1ULL, "gen_im_acc_rng_act2_01.uti"},
	{0x15EC15C1B7246983ULL, "6055434_m.fxe"},
	{0x15EC6F395930A0A2ULL, "640261142.ogg"},
	{0x15EC8C86E8FC60A7ULL, "vxht_0_hightown_a3_c1c_0_0.vlm"},
	{0x15ECEAE7CA381607ULL, "art\\characters\\playercharacter\\humanmale\\rob_magc.phy"},
	{0x15ED10082AE469D3ULL, "ave100ar_patrol.lst"},
	{0x15ED43232E40D2C2ULL, "vxht_0_hightown_b96_0_0.vlm"},
	{0x15EDA6EA215E6D58ULL, "zz_dae_debug6029299.cl"},
	{0x15EDB7BAEA9E2D4AULL, "175259717.ogg"},
	{0x15EE5988FE0B9100ULL, "523956277.ogg"},
	{0x15EEC41C95D6264DULL, "w_sundermount_30_0.rml"},
	{0x15EF0E2301CA12ADULL, "vxlt_0_lowtown_docks_a3_n_cv0_0_0.vlm"},
	{0x15EF224E090E7939ULL, "1049799327.ogg"},
	{0x15EF3AC4FEC4C3E0ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cnjure_lp.ani"},
	{0x15EF641FFF1B4F96ULL, "w_encounter3_193_0.msh"},
	{0x15EF6E1FFF1B6094ULL, "w_encounter3_193_0.mmh"},
	{0x15EF721FFF1B67A7ULL, "w_encounter3_193_0.mao"},
	{0x15EFD0CBE4C991E1ULL, "hlac_dr_slavers_3207_5z.dds"},
	{0x15EFE3D654282E8CULL, "679057231.ogg"},
	{0x15F06058F9EA2EF4ULL, "285991275.ogg"},
	{0x15F0E75B6B4F3647ULL, "hlac_lt_0_lowtown_a2_196606_0x.dds"},
	{0x15F119DB3CE87044ULL, "6174029_f.fxe"},
	{0x15F13DC9B10BFF9DULL, "vxlt_backalley_g_1xd_0_0.vlm"},
	{0x15F15BD1235A8647ULL, "ico_mm_ally.dds"},
	{0x15F198C150806E23ULL, "vxlt_0_lowtown_a2_n_gc3_0_0.vlm"},
	{0x15F1CA43C01DA23DULL, "6106123_m.fxe"},
	{0x15F1F46669411532ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.cess.evt"},
	{0x15F1F7FCBBFCF547ULL, "lt_0_lowtown_docks_n_38_0.anb"},
	{0x15F20011C248D42CULL, "mcr370_negotiation6080176.cl"},
	{0x15F216397B46494DULL, "vxdr_0_deeproads_adc_0_1.vlm"},
	{0x15F235C0EFCFCD62ULL, "vxht_0_hightown_a3_byw_0_0.vlm"},
	{0x15F242FD1D76DCDAULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.cess.evt"},
	{0x15F24D247418104EULL, "vxht_mansion_n_6mm_0_0.vlm"},
	{0x15F293A3D18F2C3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_chest_out.evt"},
	{0x15F29616D2EBE698ULL, "mag160_feynriel6020436.cl"},
	{0x15F29A195BE30B38ULL, "mag211_marethari6025307.cl"},
	{0x15F346AF89799E79ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.cess.ani"},
	{0x15F3595F3E32AAABULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockdown_v1.evt"},
	{0x15F375931963A162ULL, "550161460.ogg"},
	{0x15F398F2D1CD3E7CULL, "vxw_bonepitdestroyed_a3_2ao_0_0.vlm"},
	{0x15F3BC28DCE8C46EULL, "gen00fl_anders6111081.cl"},
	{0x15F4095E074A2C5FULL, "llac_dr_primevalentrance_1724_5.dds"},
	{0x15F426BADD710364ULL, "dae000cr_grand_cleric6215878.cl"},
	{0x15F46463E8EEF252ULL, "393891937.ogg"},
	{0x15F4E2572DCCA8ABULL, "w_encounter2_camp_36_0_lc.mmh"},
	{0x15F4E5E9EE861F93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_exit.ani"},
	{0x15F5BBFE518E348FULL, "400147470.ogg"},
	{0x15F5C612A8AC3E62ULL, "306085080.ogg"},
	{0x15F5E5911728FC5EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xb_1.mao"},
	{0x15F63A43922B1F85ULL, "6130131_m.fxe"},
	{0x15F63F03C76B4710ULL, "gen00flst_varric_nterrors.stg"},
	{0x15F6E11560FD43A1ULL, "vxlt_0_lowtown_docks_cvt_0_0.vlm"},
	{0x15F73245A801241AULL, "zz_car_debug6122475.cl"},
	{0x15F75EC12468CA61ULL, "467490724.ogg"},
	{0x15F77C6E6EEA6E5EULL, "qcr220_grey_wardens6135947.cl"},
	{0x15F7A8B98155E07FULL, "hlac_lt_foundry_n_320_23y.dds"},
	{0x15F7B2A072F3436CULL, "one220cr_yevhen.utc"},
	{0x15F7F91FCB128C71ULL, "clf260_harley6069665.cl"},
	{0x15F824A4631A4641ULL, "6187933_m.fxe"},
	{0x15F837C573C6F7CFULL, "24089152.ogg"},
	{0x15F841145119406FULL, "cs_rom000cs_merril.wbd"},
	{0x15F86FE24C35FE5AULL, "dre000_bethany6037315.cl"},
	{0x15F8A6B0875ACE4AULL, "6143960_f.fxe"},
	{0x15F8AB1217045714ULL, "vxw_woundedcoast_1ph_0_0.vlm"},
	{0x15F910430187EEB5ULL, "zz_isa_debug6133932.cl"},
	{0x15F91366694768C5ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.cess.gad"},
	{0x15F941FD1D7CF9CDULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.cess.gad"},
	{0x15F974A3D1951628ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_chest_out.gad"},
	{0x15F9CD5C20236C28ULL, "llac_lt_0_lowtown_a3_n_c_65535_1.dds"},
	{0x15F9F0FBC8376044ULL, "6070788_m.fxe"},
	{0x15FA21AF897F7DBBULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.cess.gad"},
	{0x15FA725F3E38F46CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockdown_v1.gad"},
	{0x15FB33FB35F5CB6FULL, "hlac_lt_hangedman_n_-1_7x.dds"},
	{0x15FBB6D4F74A729DULL, "gen00fl_fenris6127115.cl"},
	{0x15FC277CB6449CE7ULL, "vxht_chantry_n_30v_0_0.vlm"},
	{0x15FCB012AFB5AC9DULL, "6219254_m.fxe"},
	{0x15FD247E7E3EAC1EULL, "vxca_caverns1_mines_ei5_0_1.vlm"},
	{0x15FD342EA489D456ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_exit.evt"},
	{0x15FE51C5CEE6928BULL, "vxlt_0_lowtown_docks_bcx_0_0.vlm"},
	{0x15FE613FDEB56A45ULL, "119339663.ogg"},
	{0x15FEBB3447EC3FC9ULL, "811823743.ogg"},
	{0x15FEC30E8D0D631EULL, "qun231_party_comment6077007.cl"},
	{0x15FEF156A7F7E6DFULL, "6078857_m.fxe"},
	{0x15FF4E030E10F067ULL, "920244253.ogg"},
	{0x15FF7A5E97E749B0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut2flat.msh"},
	{0x15FF885E97E761CAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut2flat.mmh"},
	{0x15FF8AF90EAD8D60ULL, "690717152.ogg"},
	{0x15FF8C5E97E76801ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut2flat.mao"},
	{0x16000162A762245EULL, "vxlt_blackemporium_z7_0_0.vlm"},
	{0x1600342608E3968FULL, "6044050_m.fxe"},
	{0x16006ADDD120E532ULL, "6047413_f.fxe"},
	{0x16006C65DE2E7F49ULL, "6044823_m.fxe"},
	{0x1600CF22E7912925ULL, "vxlt_gallowstemplar_47b_0_0.vlm"},
	{0x160117926F9B7448ULL, "isa211_isabela6069419.cl"},
	{0x16011FF94B745DF5ULL, "vxca_caverns2_7we_0_0.vlm"},
	{0x16013EAF8985CDA0ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.cess.evt"},
	{0x16015B83727DE710ULL, "da2terrain_terrain_p0.psh.fxo.dx11.assembly"},
	{0x1601715F3E3F116AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockdown_v1.ani"},
	{0x1601AC64A0ACA456ULL, "6162232_m.fxe"},
	{0x160269BB774C1727ULL, "approval_data.gda"},
	{0x160275F639576257ULL, "llac_lt_0_lowtown_n_65537_1.dds"},
	{0x1602BA2F689682B1ULL, "1066356411.ogg"},
	{0x1602CFA59E1835DBULL, "gen00fl_carver6097344.cl"},
	{0x160322505F32910BULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.dura.evt"},
	{0x160327E516E6B7A8ULL, "dae000_epic_merchant_46179515.cl"},
	{0x160372E93247CAAAULL, "vxlt_darktown_2r0_0_0.vlm"},
	{0x1603EC10896912C1ULL, "vxlt_0_lowtown_docks_n_cuz_0_0.vlm"},
	{0x16042B2EA48FE3E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_exit.gad"},
	{0x1604333B603B605FULL, "follower_banter6136684.cl"},
	{0x160496C614CB7137ULL, "lt_0_lowtown_docks_n_248_0.mmh"},
	{0x16049AC614CB7844ULL, "lt_0_lowtown_docks_n_248_0.mao"},
	{0x1604ACC614CB96A9ULL, "lt_0_lowtown_docks_n_248_0.msh"},
	{0x1604D7EE18689EC1ULL, "vxht_mansion2b_1kd_0_0.vlm"},
	{0x1604F2F40E0FF196ULL, "6088451_m.fxe"},
	{0x1605020A994E63EBULL, "211352342.ogg"},
	{0x16055AD67A4C9D6AULL, "vxlt_0_lowtown_fef_0_0.vlm"},
	{0x160581FAC0DE7A79ULL, "ave200_donnic6110612.cl"},
	{0x1605DF2AA6DC84F5ULL, "vxdr_primevaltomb_7e8_0_1.vlm"},
	{0x1605F1C8666F033CULL, "vxw_encounter3_e4_0_0.vlm"},
	{0x1605F598084A78F8ULL, "407509353.ogg"},
	{0x160660A11E71D29CULL, "hlac_w_sundermount_-196608_0x.dds"},
	{0x16070BE86E63A18CULL, "6155694_m.fxe"},
	{0x16078B19B357BC62ULL, "6108085_m.fxe"},
	{0x16080583392BFD53ULL, "art\\characters\\weapons\\unique\\meredith\\textures\\w_gsw_014a_0s.dds"},
	{0x16082D2302FE7FD4ULL, "t3_wpn_gld.tnt"},
	{0x1608D24A9A5B2DD7ULL, "hlac_lt_gallowstemplar_7198_4z.dds"},
	{0x16090D715588AAD2ULL, "811670287.ogg"},
	{0x160918418549550CULL, "vxht_mansion2b_3cs_0_0.vlm"},
	{0x160959A1B5AB3265ULL, "vxdr_slavers_2id_0_0.vlm"},
	{0x1609B2538D7F6302ULL, "vxht_0_hightown_a2_av0_0_0.vlm"},
	{0x1609BCE9EE98363DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_exit.gad"},
	{0x1609FB7101652247ULL, "gencr_dalish_warrior_2h_f.utc"},
	{0x160A22A282E36C18ULL, "vxw_bonepitmining_a2_242_0_0.vlm"},
	{0x160A2C2EA494508FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_exit.ani"},
	{0x160A389EDDCEE0E6ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st01a_0d.dds"},
	{0x160A3B505F38DACCULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.dura.gad"},
	{0x160AD3E5F11F0F08ULL, "246739759.ogg"},
	{0x160AED4A3F3F524BULL, "607840932.ogg"},
	{0x160B2036F1207B7EULL, "hlac_w_woundedcoast_65543_0x.dds"},
	{0x160BBB729F426461ULL, "crater_half.dds"},
	{0x160BD7F05EC471C1ULL, "dae000cr_grand_cleric6148135.cl"},
	{0x160C227AE9A2FF2BULL, "870178697.ogg"},
	{0x160C61354008EA2CULL, "vxca_caverns1_mines_gj9_0_0.vlm"},
	{0x160D0D92763757E8ULL, "vxlt_warehouse_n_12g_0_0.vlm"},
	{0x160D59358D912554ULL, "vfx_vfx_p7.vsh.fxo.dx11"},
	{0x160DB8603ABC7996ULL, "vxlt_blackemporium_3t_0_0.vlm"},
	{0x160DEBD9F06385BCULL, "vxlt_gallowstemplar_n_478_0_0.vlm"},
	{0x160E1A290C81A3FCULL, "w_woundedcoast_219_0.mao"},
	{0x160E25F1CDE024E6ULL, "83277300.ogg"},
	{0x160E26290C81B79FULL, "w_woundedcoast_219_0.mmh"},
	{0x160E2C290C81C241ULL, "w_woundedcoast_219_0.msh"},
	{0x160E454553A07082ULL, "w_encounter3_180_0.msh"},
	{0x160E474553A07438ULL, "w_encounter3_180_0.mmh"},
	{0x160E534553A088DBULL, "w_encounter3_180_0.mao"},
	{0x160E6A1C5B7F8F58ULL, "vxht_mansion2_31i_0_0.vlm"},
	{0x160E72DA529DE675ULL, "6155826_m.fxe"},
	{0x160F66365739D72CULL, "6106324_m.fxe"},
	{0x161050213EF779B6ULL, "hlac_ht_chantry_n_2534_8z.dds"},
	{0x16106403E48CBE11ULL, "6064591_m.fxe"},
	{0x1610885468C26FC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanleft.gad"},
	{0x1610EC3563E6D787ULL, "sst_ss_coterie_mage_m.gda"},
	{0x16113A505F3EF7CAULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.dura.ani"},
	{0x16113AE16AEF1F9BULL, "mrl150_merrill6093965.cl"},
	{0x161155727D23BA0CULL, "712345505.ogg"},
	{0x1611E4C6243DE32AULL, "vxlt_darktown_2u1_0_0.vlm"},
	{0x16122E38318C0C5EULL, "73341516.ogg"},
	{0x16127ED447AA2756ULL, "follower_banter6169816.cl"},
	{0x161312E3DDF55636ULL, "vxw_encounter2_camp_ea_0_0.vlm"},
	{0x1613ACCE92B15E92ULL, "vxca_caverns1_e86_0_0.vlm"},
	{0x1613AFF546C79AEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_basic.mao"},
	{0x1613BBF546C7AE8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_basic.mmh"},
	{0x1613C1F546C7B97EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_basic.msh"},
	{0x1615038BC9ED299CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_justbuttress.phy"},
	{0x1615A49305E62841ULL, "llac_ht_playershouse_n_855_8.dds"},
	{0x16161479E2DA43BDULL, "646436879.ogg"},
	{0x16163753504DECD5ULL, "6188587_m.fxe"},
	{0x16169AA7E70D3175ULL, "qun200_party_comment6095257.cl"},
	{0x1616CD025932209BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04b.phy"},
	{0x1616F0A4869627D3ULL, "vxht_brothel_n_4db_0_0.vlm"},
	{0x1616F11D7B0D1B44ULL, "vxw_encounter3_fe_0_0.vlm"},
	{0x1616F9D276434C3CULL, "docks_male_elven_worker_sta6175808.cl"},
	{0x16172B04913092E6ULL, "lt_0_lowtown_a3_n_c_35_0_ll.mmh"},
	{0x16177D5468C87BD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanleft.evt"},
	{0x1617B7A203E48A10ULL, "art\\characters\\playercharacter\\humanmale\\boo_masf.msh"},
	{0x1617C5A203E4A22AULL, "art\\characters\\playercharacter\\humanmale\\boo_masf.mmh"},
	{0x1617C9A203E4A961ULL, "art\\characters\\playercharacter\\humanmale\\boo_masf.mao"},
	{0x16180126B2CAF502ULL, "579098222.ogg"},
	{0x16180FABAE710A5BULL, "6042759_m.fxe"},
	{0x161812A92F0AD2E2ULL, "foodstorage_n.dds"},
	{0x16183EE7C359B6F4ULL, "405025909.ogg"},
	{0x16184E4553A8D4C3ULL, "w_encounter3_180_0.phy"},
	{0x161863A1E6406A1AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsb.msh"},
	{0x161871A1E6408253ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsb.mao"},
	{0x161875A1E6408960ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsb.mmh"},
	{0x1618971C96841751ULL, "115950413.ogg"},
	{0x1619D76E233D830AULL, "sse_ss_torpor.gda"},
	{0x1619DE5CDB237A10ULL, "hla_ht_0_hightown_a3_-196608_0z.dds"},
	{0x1619E30688829167ULL, "vxca_cavern1_haunted_j5b_0_0.vlm"},
	{0x161A05E6AE23200EULL, "one140cr_charade6137617.cl"},
	{0x161A4EB862C3323BULL, "6112050_m.fxe"},
	{0x161A81430A67DE8FULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.cess.gad"},
	{0x161ADDCD6B4C17C0ULL, "975240601.ogg"},
	{0x161B8334C3956CB2ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage3_.gad"},
	{0x161B93C04396F2C0ULL, "t3_hlk_h01.tnt"},
	{0x161C710C6DAE5AECULL, "art\\levels\\races\\ferelden\\props\\mining\\_2_-_default.mao"},
	{0x161D2EF7A6227CB5ULL, "zz_one_debug6077389.cl"},
	{0x161DC0CA581E1BF1ULL, "crt_beeup2.mao"},
	{0x161E01C5514463F0ULL, "6096469_m.fxe"},
	{0x161E0AF7AF35D410ULL, "6117274_m.fxe"},
	{0x161E87F4E9A5FCF1ULL, "w_bonepit_65_0.mmh"},
	{0x161E906AF2308326ULL, "vxlt_gallowstemplar_n_2pa_0_0.vlm"},
	{0x161E93F4E9A61092ULL, "w_bonepit_65_0.mao"},
	{0x161EA1F4E9A628AFULL, "w_bonepit_65_0.msh"},
	{0x161ED464633B3D99ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_3.pwk"},
	{0x161F43893F9D43E8ULL, "6107194_m.fxe"},
	{0x161F8A843A4FC691ULL, "vxlt_0_lowtown_gax_0_0.vlm"},
	{0x162022F8AE5BCA68ULL, "6129030_f.fxe"},
	{0x162056A31633FBAAULL, "vxlt_darktown_2x2_0_0.vlm"},
	{0x1620591A49A13F13ULL, "vxw_woundedcoast_nh_0_0.vlm"},
	{0x1620715A47EAC652ULL, "vxht_brothel_n_4dg_0_0.vlm"},
	{0x162096430A6C6E3CULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.cess.evt"},
	{0x162136FC7EF7C32FULL, "follower_banter6163163.cl"},
	{0x162137D3CDC5FCA1ULL, "252215987.ogg"},
	{0x162146F1069C7581ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_7y.dds"},
	{0x1621903FD08F9618ULL, "6128643_m.fxe"},
	{0x1621FD7A4A93A58BULL, "vxht_keep_n_4ng_0_0.vlm"},
	{0x16221690398E6AF8ULL, "dae000ip_crowd_lm4_drunk2.utp"},
	{0x16221EFFF4B049DEULL, "llac_w_sundermount_-196604_0.dds"},
	{0x16223B468CCAE853ULL, "one300_zevran6075740.cl"},
	{0x162279A4C4A19BA1ULL, "hlac_ht_0_hightown_a2_-65538_1y.dds"},
	{0x1622AC327430CD61ULL, "6090329_m.fxe"},
	{0x1624397B44BA0483ULL, "hlac_lt_0_lowtown_a2_n_1_3y.dds"},
	{0x16246ED39743ED62ULL, "6106285_m.fxe"},
	{0x1624755468D2F80DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanleft.ani"},
	{0x1624AAF81EF131F0ULL, "370239659.ogg"},
	{0x16254C3C2C7542FDULL, "776852240.ogg"},
	{0x16254DCDE6E91451ULL, "vxw_sundermount_2jl_0_0.vlm"},
	{0x162562B9EC195C52ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs103a.phy"},
	{0x162562FA6D0333C4ULL, "vxw_encounter3_cj_0_0.vlm"},
	{0x1625C23A9BD387D0ULL, "ht_0_hightown_a3_26_0_lc.mmh"},
	{0x1625D4FF70EE6B55ULL, "w_woundedcoast_197_0.phy"},
	{0x1626390768FF5F30ULL, "6039273_m.fxe"},
	{0x16264904B6F8EAF6ULL, "6027777_m.fxe"},
	{0x1626FFB318AA9A36ULL, "6087848_f.fxe"},
	{0x162710626805BBE2ULL, "6071104_f.fxe"},
	{0x1627214F958E2AFFULL, "lt_0_lowtown_docks_a3_130_0.msh"},
	{0x1627274F958E34A1ULL, "lt_0_lowtown_docks_a3_130_0.mmh"},
	{0x1627334F958E4942ULL, "lt_0_lowtown_docks_a3_130_0.mao"},
	{0x1627352344380D60ULL, "one280_merin6120919.cl"},
	{0x16273DE733AD6CC4ULL, "vxht_keep_6bq_0_0.vlm"},
	{0x16280C3AD7E8DF0BULL, "vxw_bonepitdestroyed_a3_1zn_0_0.vlm"},
	{0x16284B5AF5B183D8ULL, "373982101.ogg"},
	{0x16286AD78B5D2393ULL, "vxht_0_hightown_a2_c78_0_0.vlm"},
	{0x1628DB7EBFB5589AULL, "art\\characters\\playercharacter\\humanfemale\\boo_magc.mmh"},
	{0x1628DF7EBFB55FD1ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magc.mao"},
	{0x1628ED7EBFB577E0ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magc.msh"},
	{0x1628F0E016D36958ULL, "vxlt_warehouse_qx_0_0.vlm"},
	{0x16291A3322A646F3ULL, "vxlt_darktown_6on_0_0.vlm"},
	{0x162941A82FCDD730ULL, "6141097_m.fxe"},
	{0x16294CA3AB47CF28ULL, "dae000_sandal6148184.cl"},
	{0x16295F33A43D9F68ULL, "dae200_bran6191225.cl"},
	{0x16296EA02D612974ULL, "vxdr_primevaltomb_2u5_0_0.vlm"},
	{0x1629758442C6C00BULL, "vxca_cavern1_haunted_i6h_0_0.vlm"},
	{0x1629ECAFE36BA0C0ULL, "6202329_m.fxe"},
	{0x162A6953BE702DDDULL, "830752051.ogg"},
	{0x162B63C76979B01BULL, "gen00fl_fenris6110814.cl"},
	{0x162B9EF69F0F7546ULL, "llac_lt_0_lowtown_docks_-196610_4.dds"},
	{0x162BA30C92CBD311ULL, "lla_ht_0_hightown_-131074_0.dds"},
	{0x162BFD34B7A82F62ULL, "vxht_0_hightown_a3_bes_0_0.vlm"},
	{0x162BFEC76A8CA901ULL, "1057466567.ogg"},
	{0x162C12661A2D44FCULL, "313058169.ogg"},
	{0x162C7DE17CA06645ULL, "vxlt_gallowscourtyard_6tk_0_0.vlm"},
	{0x162CA8B997F9A846ULL, "137435908.ogg"},
	{0x162D552655D42523ULL, "807342339.ogg"},
	{0x162D84CBAAD216B6ULL, "263796946.ogg"},
	{0x162DAAC34F3D6414ULL, "6131909_player_ease_pain.cut"},
	{0x162DB425E13278EEULL, "gen00fl_isabela6106230.cl"},
	{0x162E17DFE3CFBDC9ULL, "145147667.ogg"},
	{0x162E235E684E9682ULL, "6101767_m.fxe"},
	{0x162E823335190A30ULL, "vxht_brothel_n_53y_0_0.vlm"},
	{0x162E912B69A77CB1ULL, "24561560.ogg"},
	{0x162E9E430A78B955ULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.cess.ani"},
	{0x162EFEB2A27517CFULL, "follower_banter6163637.cl"},
	{0x162F277DA31B24F2ULL, "and111cr_lirene.utc"},
	{0x162F6E34C3A5F258ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage3_.ani"},
	{0x1630A9C2B778D79CULL, "hlac_lt_0_lowtown_-262143_0x.dds"},
	{0x1630FE248F08F335ULL, "262398072.ogg"},
	{0x163103B45F215525ULL, "878048141.ogg"},
	{0x1631FC195B012BF6ULL, "vxlt_0_lowtown_n_e2k_0_0.vlm"},
	{0x1632A4822344C54BULL, "hlac_lt_0_lowtown_a2_n_c_131073_1y.dds"},
	{0x16330D8CC2F8647FULL, "502496810.ogg"},
	{0x163350A9C04EE17AULL, "w_blightlands_13_0.rml"},
	{0x1633BFAAD8DF2CD1ULL, "vxw_sundermount_2oc_0_0.vlm"},
	{0x1633DC448F86C732ULL, "hlac_lt_0_lowtown_131070_4y.dds"},
	{0x1634590D08269258ULL, "131811601.ogg"},
	{0x1634A82E9A647151ULL, "vxlt_gallowstemplar_f_5uw_0_1.vlm"},
	{0x1635281A46472987ULL, "venge_quickaction.dds"},
	{0x16357C83A984E7BDULL, "6042564_m.fxe"},
	{0x16359209BEDF7B02ULL, "vxw_encounter2_qy_0_0.vlm"},
	{0x1635CDEE580A79A5ULL, "6127301_f.fxe"},
	{0x1635F9B1AE5E11AAULL, "lgt120_orz_bln_official6140202.cl"},
	{0x163630B91A264045ULL, "t3_abm_h01.tnt"},
	{0x16363751E1B0AEBDULL, "hlac_lt_0_lowtown_a2_n_-65539_0x.dds"},
	{0x16365F5B755CE2D5ULL, "vxw_bonepitmining_a2_1l8_0_0.vlm"},
	{0x16368F7188D7CB57ULL, "6075202_f.fxe"},
	{0x1636EAEB7877B23DULL, "6206931_m.fxe"},
	{0x1636F9A1EA9C7CE3ULL, "vxlt_blackemporium_13h_0_0.vlm"},
	{0x1638030A6DEF3CCAULL, "w_bonepitmining_a2_28_0_lc.mmh"},
	{0x16382B21DC20D71EULL, "vxht_mansion_n_53y_0_1.vlm"},
	{0x16385B14514F5045ULL, "cs_rom000cs_merril.bnk"},
	{0x16388828E5F20CE2ULL, "fen110_fenris6117981.cl"},
	{0x16388F2C286264CFULL, "hm_friedrich_thug3.mor"},
	{0x16398AA8561AF0CFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad_12.pwk"},
	{0x163AE2AD92BDF8EDULL, "770679625.ogg"},
	{0x163B5C080589C04DULL, "m_dae221ar_player_home_nt.wbd"},
	{0x163B98B9C423093DULL, "510227565.ogg"},
	{0x163C05943B8EB626ULL, "6036533_m.fxe"},
	{0x163C304F5D841266ULL, "gen00fl_dog6165775.cl"},
	{0x163C5B9622E7D5ECULL, "llac_lt_0_lowtown_a2_n_c_131069_0.dds"},
	{0x163C999AB2F9D17DULL, "mrl150_merrill6093534.cl"},
	{0x163CC4AF8253481AULL, "vxht_keep_n_6jl_0_0.vlm"},
	{0x163D62D44F5C3ED7ULL, "816508193.ogg"},
	{0x163DA8558CF58C07ULL, "ser221_gamlen6134814.cl"},
	{0x163E2D406210743AULL, "6076483_m.fxe"},
	{0x163EA82455AC69CEULL, "6024556_m.fxe"},
	{0x163F081DD06332CFULL, "w_bonepitmining_a2_172_0.phy"},
	{0x163F412931EE66C9ULL, "329825530.ogg"},
	{0x164021271A2E4FC3ULL, "hlac_dr_slavers_3215_8x.dds"},
	{0x16404CFBA385EA1BULL, "vxlt_0_lowtown_h8f_0_0.vlm"},
	{0x16415B39009467DDULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_genbenchxa.phy"},
	{0x1641EF0F46D5BFDBULL, "llac_ht_keep_4519_7.dds"},
	{0x164220546CA43810ULL, "561424974.ogg"},
	{0x16423187CAD54551ULL, "vxw_sundermount_2pb_0_0.vlm"},
	{0x164292AE5F867E4FULL, "zz_ave_debug6062035.cl"},
	{0x1643542189A510D1ULL, "152489378.ogg"},
	{0x1643A5D39AC52D57ULL, "761624357.ogg"},
	{0x1643B4025958E43AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04b.msh"},
	{0x1643B89E16CB9CD4ULL, "6059042_m.fxe"},
	{0x1643C60259590280ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04b.mmh"},
	{0x164421A117999DC4ULL, "vxht_keep_6hw_0_0.vlm"},
	{0x16445F040050A10DULL, "zz_dae_debug6142955.cl"},
	{0x16456AC0B224C326ULL, "6107178_m.fxe"},
	{0x16459978A0123315ULL, "353459145.ogg"},
	{0x16470953713FAE0DULL, "27019192.ogg"},
	{0x1647AD50DA107184ULL, "w_bonepitmining_a2_161_0.mmh"},
	{0x1647B150DA1078D7ULL, "w_bonepitmining_a2_161_0.mao"},
	{0x1647C350DA109726ULL, "w_bonepitmining_a2_161_0.msh"},
	{0x164857E34EDD186CULL, "gen00fl_merrill6147024.cl"},
	{0x1648D3BA97B86E1AULL, "379304141.ogg"},
	{0x1648D5B7DE1B7D9AULL, "vxlt_0_lowtown_a2_eow_0_0.vlm"},
	{0x1648DB15907C0C87ULL, "93071449.ogg"},
	{0x164933084D82550EULL, "vxlt_warehouse_n_13n_0_0.vlm"},
	{0x16495CEFDF6D671AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_crumble_1.mao"},
	{0x164960986E524E39ULL, "vxht_0_hightown_a2_b8r_0_0.vlm"},
	{0x1649995064DC45B4ULL, "lt_0_lowtown_n_46_0_lc.mmh"},
	{0x164B4BF7C94FD4F4ULL, "vxlt_0_lowtown_a2_n_c_eqf_0_0.vlm"},
	{0x164B814DA33DF7FFULL, "fen110_anso6095028.cl"},
	{0x164C398E27844B89ULL, "gen00fl_varric6103265.cl"},
	{0x164C4386684017BEULL, "vxdr_0_deeproads_fbt_0_0.vlm"},
	{0x164C4BE08E7AEFB3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardb.phy"},
	{0x164CAC81CD64FB04ULL, "vxw_encounter2_camp_p0_0_0.vlm"},
	{0x164D40C80D0C9B99ULL, "vxlt_undercitydungeon_1j6_0_1.vlm"},
	{0x164D4D21FB8B08D4ULL, "6085883_m.fxe"},
	{0x164D9630EA5697BFULL, "llac_w_encounter3_0_11.dds"},
	{0x164DD999EE43DCDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_thrtn.evt"},
	{0x164DFA7EBFD4A9A5ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magc.phy"},
	{0x164E40685B8F1220ULL, "hlac_lt_undercityserialkiller_-1_18y.dds"},
	{0x164E4EACBB4F6F7FULL, "vxca_caverns1_ckd_0_0.vlm"},
	{0x164E6DF51DB39C02ULL, "6111348_m.fxe"},
	{0x164F07E681FFC7CFULL, "338236176.ogg"},
	{0x164F2B75DE4B1AF8ULL, "vxlt_0_lowtown_a2_n_c_gwj_0_0.vlm"},
	{0x164F411E006FAED4ULL, "hlac_ht_mansion2_818_35z.dds"},
	{0x164FFD8295149862ULL, "llac_w_woundedcoast_-196606_0.dds"},
	{0x16501D593FA7A36AULL, "6208667_f.fxe"},
	{0x16506E416B10E5A2ULL, "pro000st_meeran.stg"},
	{0x16507D7AF4B1BDD7ULL, "344275495.ogg"},
	{0x1650A364BCCB5DD1ULL, "vxw_sundermount_2ua_0_0.vlm"},
	{0x1650B7F86CD709E3ULL, "129166266.ogg"},
	{0x1650B8184E6E89D3ULL, "vxht_brothel_n_4xf_0_0.vlm"},
	{0x1650DA15649C5765ULL, "6074392_m.fxe"},
	{0x1650EBBE7C782952ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_tk_camp_fire.gad"},
	{0x16513D7EA78F7146ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernrock_n.dds"},
	{0x1651A10289B00E0EULL, "follower_banter6171026.cl"},
	{0x1651E1B9D9AEFEFBULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firelogs_n.dds"},
	{0x1651F66C4FC69E2AULL, "zz_followers_banter6157594.cl"},
	{0x1652167DE442E818ULL, "spt_leafmeshes.psh.fxo"},
	{0x1652503900A2C849ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_genbenchxa.mao"},
	{0x16525C3900A2DCE2ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_genbenchxa.mmh"},
	{0x16525E3900A2DF98ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_genbenchxa.msh"},
	{0x1652FE791518701AULL, "follower_banter6143140.cl"},
	{0x1653106EF29AB8B4ULL, "gen_im_npc_elf_arm_cht_dalc.uti"},
	{0x1653C3CA772849B4ULL, "888879369.ogg"},
	{0x1653F57EA7B8115BULL, "869533038.ogg"},
	{0x1653FA34B91239A9ULL, "6038604_m.fxe"},
	{0x16543056B46AC3DBULL, "616344452.ogg"},
	{0x1654A65080AE0FC1ULL, "vxlt_0_lowtown_a2_n_glt_0_0.vlm"},
	{0x1654ADAD161C8416ULL, "follower_banter6151363.cl"},
	{0x1654F099EE4A2209ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_thrtn.gad"},
	{0x1654F4749E267506ULL, "6070584_m.fxe"},
	{0x1654F50C449981B8ULL, "6048303_m.fxe"},
	{0x1654F53C5AAB24F7ULL, "vxht_keep_583_0_0.vlm"},
	{0x16556379D9791E7EULL, "mcr310_argument6182276.cl"},
	{0x1655C82D28192750ULL, "qcr221_cut_qunari_defeated6042785.cl"},
	{0x1656D44F98842522ULL, "art\\vfx\\materials\\vfx_markerlight.mao"},
	{0x1656EEBE7C7C9989ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_tk_camp_fire.evt"},
	{0x16570C7AB44F3402ULL, "6112921_m.fxe"},
	{0x16574DEDFFEE9AF6ULL, "matdef_dcb_diffuse.dds"},
	{0x1657A4E53F786D8EULL, "vxlt_warehouse_n_18o_0_0.vlm"},
	{0x165966B12D7655E2ULL, "demo_weapon_swap6207186.cl"},
	{0x165A0C9B4DB7B198ULL, "6032593_m.fxe"},
	{0x165A1C7EB7646F57ULL, "505670950.ogg"},
	{0x165A208E1179A113ULL, "vxw_woundedcoast_zl_0_0.vlm"},
	{0x165A531F88AA0FE6ULL, "vxdr_primevalentrance_27m_0_0.vlm"},
	{0x165A878EC2907304ULL, "vxht_brothel_n_2ny_0_0.vlm"},
	{0x165A9DFE6B62448BULL, "6095247_m.fxe"},
	{0x165AA27928CB4ED8ULL, "494225866.ogg"},
	{0x165AE87DEEF0177AULL, "553191877.ogg"},
	{0x165B0964AB464971ULL, "lgt000cr_baz_den_rev.utc"},
	{0x165B1E5EBF5B1384ULL, "vxw_encounter2_camp_k1_0_0.vlm"},
	{0x165B2A1970773C0AULL, "vxlt_0_lowtown_a2_n_hf6_0_0.vlm"},
	{0x165B89E4C91EF70FULL, "vxht_brothel_n_635_0_0.vlm"},
	{0x165BC0001608E368ULL, "vxht_0_hightown_a2_9zf_0_0.vlm"},
	{0x165BC199EE4FF157ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_thrtn.ani"},
	{0x165BC21E5F45E322ULL, "mrl000_marethari6142493.cl"},
	{0x165BEF94AAEAE32DULL, "mag211_marethari6025151.cl"},
	{0x165CE1F04182C3EEULL, "vxca_cavern1_haunted_imt_0_0.vlm"},
	{0x165D3DB9E61BA693ULL, "vxw_bonepit_1vb_0_0.vlm"},
	{0x165D9D52D0413378ULL, "vxlt_0_lowtown_a2_n_c_gpk_0_0.vlm"},
	{0x165DBE1D3A189421ULL, "6055072_m.fxe"},
	{0x165DE8E16DCF73E2ULL, "dre000_sibling_has_blight6036557.cl"},
	{0x165DFEEB74DDD73FULL, "vxht_mansion2b_1hs_0_0.vlm"},
	{0x165E109194CD63F9ULL, "vxca_caverns2prc_adw_0_0.vlm"},
	{0x165E3A79EB2D5735ULL, "968778888.ogg"},
	{0x165E69EF7A4CA0BFULL, "6163515_m.fxe"},
	{0x165EA10B067C361DULL, "mrl150_merrill6070599.cl"},
	{0x165ED8A41BE7CA62ULL, "hlac_ht_playershouse_n_855_10z.dds"},
	{0x165F3FB04D68D6EFULL, "hlac_lt_backalley_n_1_3z.dds"},
	{0x165F4BF256926FE5ULL, "6079092_m.fxe"},
	{0x165FB19C1CFA0D8BULL, "vxdr_primevaltomb_8lo_0_0.vlm"},
	{0x165FC93D592A0660ULL, "gen_im_trs_dem_amu_r2.uti"},
	{0x166022976C3E24DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleft.phy"},
	{0x16603F5534AF084CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_001a_0t.dds"},
	{0x166064C781B2DBF1ULL, "148316895.ogg"},
	{0x1660E6202C90CC93ULL, "art\\characters\\playercharacter\\shared\\textures\\default_black.dds"},
	{0x1661419EC7252F04ULL, "772755999.ogg"},
	{0x1661BE4534D6E745ULL, "6029376_m.fxe"},
	{0x1661D3AE9FC1410BULL, "vxw_bonepitdestroyed_a3_1nb_0_0.vlm"},
	{0x16623239143BA412ULL, "6172934_m.fxe"},
	{0x16625EEEE163DE92ULL, "143243418.ogg"},
	{0x1662C856F63E4319ULL, "6112662_f.fxe"},
	{0x1663182D72A42841ULL, "vxlt_0_lowtown_a2_n_gik_0_0.vlm"},
	{0x166346DDE41728FEULL, "hlac_ht_0_hightown_a2_65533_1x.dds"},
	{0x166347A4EB7A6511ULL, "sse_ss_fereldan_m_1.gda"},
	{0x16641BCAD1D87A7EULL, "vxht_mansion2b_348_0_0.vlm"},
	{0x1664A7FE9670E56BULL, "6172467_m.fxe"},
	{0x1664B5321D3426F7ULL, "vxht_0_hightown_a3_auq_0_0.vlm"},
	{0x1664D6BE7C88AEF8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_tk_camp_fire.ani"},
	{0x1664DAA4E4311836ULL, "35959878.ogg"},
	{0x1664FAFCDB1BDA52ULL, "vxlt_0_lowtown_a2_h2i_0_0.vlm"},
	{0x1665AD588EB6B86FULL, "6044223_m.fxe"},
	{0x16668431F63DF48DULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb103a.mmh"},
	{0x16669631F63E1313ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb103a.msh"},
	{0x166727DA617E027CULL, "vxw_bonepitdestroyed_a3_29g_0_0.vlm"},
	{0x1667585EDF46474DULL, "vxlt_0_lowtown_docks_a3_n_d2w_0_0.vlm"},
	{0x16677BF29E162036ULL, "vxdr_primevalentrance_2e3_0_0.vlm"},
	{0x1667EECB8F47AE60ULL, "gen00fl_sebastian6189421.cl"},
	{0x166848554335C5C2ULL, "6058955_m.fxe"},
	{0x166858192652ACAFULL, "gen00fl_anders6119293.cl"},
	{0x1668926B036FB993ULL, "vxw_woundedcoast_wo_0_0.vlm"},
	{0x1668BC7DCD77A177ULL, "art\\characters\\playercharacter\\elfmale\\em_boo_lgte.phy"},
	{0x166927404C2C48BEULL, "vxdr_0_deeproads_fhv_0_0.vlm"},
	{0x1669903BB1512C04ULL, "vxw_encounter2_camp_f6_0_0.vlm"},
	{0x166A560AF2BB8B84ULL, "gen00fl_carver6102582.cl"},
	{0x166A5619A8BA848DULL, "6040824_f.fxe"},
	{0x166A63BDFCEAABDDULL, "hlac_lt_0_lowtown_docks_a3_n_-4_6x.dds"},
	{0x166AAA9B04923DB7ULL, "6187816_m.fxe"},
	{0x166ABBBEBB1B7022ULL, "6140333_m.fxe"},
	{0x166ACF2B49258393ULL, "419164725.ogg"},
	{0x166B53CD3378DC6EULL, "vxca_cavern1_haunted_iju_0_0.vlm"},
	{0x166B67ECF53D9773ULL, "hlac_lt_0_lowtown_n_65535_5z.dds"},
	{0x166BA644DD9B6A76ULL, "vxht_keep_n_7a8_0_0.vlm"},
	{0x166BB6625D6DE9AEULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromopen.ani"},
	{0x166BF337EB6AF5A8ULL, "hlac_lt_0_lowtown_a2_-196608_3y.dds"},
	{0x166C1BEFC6864AD5ULL, "vxht_playershouse_n_1ix_0_0.vlm"},
	{0x166C2FFA2C0EACA1ULL, "6052172_m.fxe"},
	{0x166C826E86C37C79ULL, "vxca_caverns2prc_acv_0_0.vlm"},
	{0x166C9471B1FF7502ULL, "vxht_brothel_n_6gh_0_0.vlm"},
	{0x166CE7FF712A8E70ULL, "w_woundedcoast_197_0.msh"},
	{0x166CF5FF712AA58AULL, "w_woundedcoast_197_0.mmh"},
	{0x166CF9FF712AACC1ULL, "w_woundedcoast_197_0.mao"},
	{0x166D235ABCF63B94ULL, "art\\levels\\races\\proxy\\props\\door\\animation\\doorclosing.ani"},
	{0x166D24DDF16DB881ULL, "582634511.ogg"},
	{0x166D844C85151C67ULL, "135371726.ogg"},
	{0x166DA782E8D24CB6ULL, "lt_warehouse_n_3_0.rml"},
	{0x166E072B1D0D8254ULL, "6166776_f.fxe"},
	{0x166F03F1E84B5795ULL, "6170904_m.fxe"},
	{0x166F1871C91881FAULL, "llac_dr_0_deeproads_5842_3.dds"},
	{0x166F2D67D42B14A3ULL, "hlac_lt_warehouse_n_-65536_2y.dds"},
	{0x166F561BC39D07D3ULL, "441855844.ogg"},
	{0x166FB4E959A31F11ULL, "art\\vfx\\cinematic\\vfx_lyrium_glow_c.impa.ani"},
	{0x1670343DCA361E02ULL, "6087490_m.fxe"},
	{0x16703B0EF57FF76CULL, "6140905_f.fxe"},
	{0x1670A80BAA613D5AULL, "471817543.ogg"},
	{0x1670E1EA7A53CA12ULL, "mnd_pro_02.mao"},
	{0x16710F43FFA39D53ULL, "173560247.ogg"},
	{0x16712F290CD644E4ULL, "w_woundedcoast_219_0.phy"},
	{0x16724AD7C1590C5EULL, "523634771.ogg"},
	{0x1672797DBD7ED042ULL, "vxw_bonepit_ko_0_0.vlm"},
	{0x167294E720251D34ULL, "vxdr_slavers_20z_0_0.vlm"},
	{0x1672A24A612461DBULL, "6107176_f.fxe"},
	{0x1672CCF694E5C9A2ULL, "vxlt_0_lowtown_a2_n_ex5_0_0.vlm"},
	{0x1673BE7C6552C2B0ULL, "gen00fl_merrill6128669.cl"},
	{0x1673CC99FB231FB7ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_noise2_d.dds"},
	{0x167415F555BAF313ULL, "dae000ip_crowd_em3_talk10.utp"},
	{0x1674205ABCFC55F6ULL, "art\\levels\\races\\proxy\\props\\door\\animation\\doorclosing.gad"},
	{0x16743DCCD808947DULL, "dae000_gamlen6149475.cl"},
	{0x16746AAEA13F96CCULL, "vxdr_primevalentrance_1if_0_0.vlm"},
	{0x1674A082E8D85FDAULL, "lt_warehouse_n_3_0.phy"},
	{0x16757A98D70FA5AFULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_7y.dds"},
	{0x167599B753741AFCULL, "vxw_bonepitdestroyed_a3_24d_0_0.vlm"},
	{0x1675B21A6B870A29ULL, "6128765_m.fxe"},
	{0x16760AFE584ED506ULL, "6169174_m.fxe"},
	{0x16760D80DF1CBFB8ULL, "161709884.ogg"},
	{0x1676215D03C3ADA5ULL, "vxlt_0_lowtown_n_e15_0_0.vlm"},
	{0x1676BFE959A950B3ULL, "art\\vfx\\cinematic\\vfx_lyrium_glow_c.impa.gad"},
	{0x1676CA0AC8B008C7ULL, "vxlt_gallowstemplar_f_6q7_0_0.vlm"},
	{0x1677E0A147898EF8ULL, "615181923.ogg"},
	{0x1677E2DB1E465509ULL, "74848727.ogg"},
	{0x16780218A3474484ULL, "vxw_encounter2_camp_a7_0_0.vlm"},
	{0x1678072078E24819ULL, "vxht_0_hightown_n_b7l_0_0.vlm"},
	{0x16785C4F226CAFA6ULL, "233549820.ogg"},
	{0x1678632AACA8380FULL, "vxlt_0_lowtown_a3_n_c_hey_0_0.vlm"},
	{0x1678A0B82F674361ULL, "6133753_f.fxe"},
	{0x1679AE625D7A19FFULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromopen.evt"},
	{0x1679B110CE314B2AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_cutoff_d.dds"},
	{0x1679B5BC1BE6B742ULL, "vxlt_0_lowtown_n_f5a_0_0.vlm"},
	{0x1679C87EA95AFF78ULL, "hlac_lt_0_lowtown_a3_n_c_-2_4x.dds"},
	{0x167A73494A3C0F6DULL, "zz_isa_debug6133929.cl"},
	{0x167ABFC6632E3698ULL, "6028420_m.fxe"},
	{0x167B30A9C08C6031ULL, "w_blightlands_13_0.mmh"},
	{0x167B3CA9C08C74D2ULL, "w_blightlands_13_0.mao"},
	{0x167B4AA9C08C8CEFULL, "w_blightlands_13_0.msh"},
	{0x167B80982A382F6CULL, "vxht_mansion2b_1wq_0_0.vlm"},
	{0x167BBF05E072562BULL, "gen_im_arm_hlm_med_act1_02.uti"},
	{0x167C660805C16877ULL, "m_dae221ar_player_home_nt.bnk"},
	{0x167C988B1D12D7D7ULL, "vxlt_0_lowtown_a3_n_c_bys_0_0.vlm"},
	{0x167CD2FF46AC0343ULL, "6090338_m.fxe"},
	{0x167D2840DA354059ULL, "gen00fl_fenris6108048.cl"},
	{0x167D30C9DBBEC91AULL, "gen00fl_aveline6115411.cl"},
	{0x167DBCE959AF6A18ULL, "art\\vfx\\cinematic\\vfx_lyrium_glow_c.impa.evt"},
	{0x167DE71C2F4CDEA6ULL, "6059228_m.fxe"},
	{0x167E283C0566A135ULL, "118104682.ogg"},
	{0x167E316C8271A9DBULL, "6168154_m.fxe"},
	{0x167FA9D47A650221ULL, "vxht_keep_n_5vh_0_0.vlm"},
	{0x167FE5887358FB15ULL, "llac_ca_caverns2_5802_51.dds"},
	{0x167FFBE756905941ULL, "vxlt_0_lowtown_a2_n_gci_0_0.vlm"},
	{0x168034AD388733F5ULL, "vxlt_0_lowtown_a2_n_c_eh8_0_0.vlm"},
	{0x1680870A0508DE80ULL, "6072566_m.fxe"},
	{0x16808F625D8003E8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromopen.gad"},
	{0x1680C3328D14839EULL, "6169326_f.fxe"},
	{0x168132F94214FEBDULL, "6120479_m.fxe"},
	{0x1681536EF625B89BULL, "6110917_m.fxe"},
	{0x1681596ACA6BE053ULL, "6040955_m.fxe"},
	{0x16815D4E6B499C49ULL, "combat_smr_only_range.gda"},
	{0x1681BB1C77E92258ULL, "small_n.dds"},
	{0x1681C5E05C3A9CE7ULL, "6134919_m.fxe"},
	{0x1682009060C1FD99ULL, "vxht_mansion_n_66s_0_0.vlm"},
	{0x1682E2546255A798ULL, "w_bonepitmining_a2_36_0.rml"},
	{0x1683190F3C6A0F04ULL, "hlac_ca_caverns1_24423_6z.dds"},
	{0x16838FE0A5B15E74ULL, "6173355_m.fxe"},
	{0x1683B8CE2F49E15FULL, "one000_ilen6036913.cl"},
	{0x1683E042F3DE4CD8ULL, "vxw_encounter3_6i_0_0.vlm"},
	{0x1683FCBF0795E0F3ULL, "6152744_m.fxe"},
	{0x1684AB915950BEF2ULL, "202426068.ogg"},
	{0x1684B52CE0FE2034ULL, "hlac_lt_gallowstemplar_2390_1y.dds"},
	{0x1684B63814764385ULL, "fen210_orana6109702.cl"},
	{0x16851F326B81E24EULL, "302030875.ogg"},
	{0x168585431CDB8344ULL, "mag291_torpor6021667.cl"},
	{0x1685B8F8CA1D526BULL, "ts_merchant_tables.gda"},
	{0x16860F2028F61576ULL, "lowtown_male_drunk_mob.cnv"},
	{0x1686B0A7B455F917ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_3z.dds"},
	{0x1686B115D0C8D890ULL, "6118507_m.fxe"},
	{0x1686C38C2A3D8C8DULL, "qun211cr_elf_warrior6072202.cl"},
	{0x1686C8D748F4E675ULL, "6070208_m.fxe"},
	{0x16870660200F0413ULL, "6133132_m.fxe"},
	{0x1687C0AB1522DADBULL, "vxlt_0_lowtown_docks_a3_n_cq3_0_0.vlm"},
	{0x1687F0184C845710ULL, "vxlt_0_lowtown_a2_byh_0_0.vlm"},
	{0x16883B351FFCBE47ULL, "static_passthrough_p0.vsh.fxo.dx11.assembly"},
	{0x1688CC9D3CD32062ULL, "m_fen110ar_abandoned_house.bnk"},
	{0x1688D89102B0C2E9ULL, "vxw_bonepit_ng_0_0.vlm"},
	{0x1688E8419B15954EULL, "vxlt_0_lowtown_a3_n_c_hx3_0_0.vlm"},
	{0x168925FBED93367FULL, "vxw_encounter2_cf_0_0.vlm"},
	{0x1689B9269D6CAA8BULL, "235128398.ogg"},
	{0x1689C4A256D0B408ULL, "mrl150_merrill6085743.cl"},
	{0x168A112028FA0156ULL, "lowtown_male_drunk_mob.bnk"},
	{0x168AF0A162104EFCULL, "dae000ip_crowd_em1_talk11.utp"},
	{0x168B04E15B0FB79CULL, "vxlt_gallowstemplar_56e_0_0.vlm"},
	{0x168BE3248313EB9CULL, "6117681_m.fxe"},
	{0x168C0F9B1AED699AULL, "qcr220_meredith6054752.cl"},
	{0x168C933F98542DC9ULL, "vxlt_backalley_n_1de_0_0.vlm"},
	{0x168DFD8183AB3999ULL, "hlac_lt_0_lowtown_docks_n_131071_2y.dds"},
	{0x168E752F1FB9FF69ULL, "6025307_m.fxe"},
	{0x168E7C6023F5C2E9ULL, "llac_ca_cavern1_haunted_24424_3.dds"},
	{0x168EB618B2289CF9ULL, "31746453.ogg"},
	{0x168EC4674283878BULL, "vxlt_0_lowtown_docks_b5r_0_0.vlm"},
	{0x168F85041DF18309ULL, "hlac_ht_0_hightown_a2_-65537_1x.dds"},
	{0x16901261E3A5D8AAULL, "hlac_lt_0_lowtown_docks_65535_2x.dds"},
	{0x16901BC87BFA6870ULL, "fx_c_cloud_add.mao"},
	{0x16901D63ECB727E4ULL, "1022607033.ogg"},
	{0x16914494BA943C7DULL, "hlac_dr_primevaltomb_658_11z.dds"},
	{0x1691A10B6F9A10F8ULL, "6208352_m.fxe"},
	{0x16922F76D87CEE22ULL, "190592927.ogg"},
	{0x169253C9D8331E71ULL, "mcr310_argument6074392.cl"},
	{0x1692C538EBBA29A1ULL, "6173714_m.fxe"},
	{0x1692EEB83C3B0606ULL, "6163374_m.fxe"},
	{0x16931C4C2DA00BA0ULL, "6121507_m.fxe"},
	{0x1694217465CAA33EULL, "vxlt_0_lowtown_a2_n_fi5_0_0.vlm"},
	{0x16942B5D8D460A59ULL, "371401841.ogg"},
	{0x16950DF477F3F203ULL, "6113963_m.fxe"},
	{0x169522F2C2BEF110ULL, "6117407_m.fxe"},
	{0x16952D94D5F857BEULL, "6095945_m.fxe"},
	{0x1695B65BD64A0183ULL, "one220_yevhen6121663.cl"},
	{0x1695C1C13C4C92B6ULL, "6019156_m.fxe"},
	{0x1695FE2D24BD0C4DULL, "6149415_f.fxe"},
	{0x169627521F46CF40ULL, "mer210_coterie_contact6036768.cl"},
	{0x169632880718F35BULL, "vxlt_0_lowtown_docks_a3_n_cz2_0_0.vlm"},
	{0x16964394B649B43CULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtb.mao"},
	{0x16964CE6CF34A433ULL, "6170601_m.fxe"},
	{0x16964F94B649C8DFULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtb.mmh"},
	{0x16965594B649D281ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtb.msh"},
	{0x1696BAA412AE2B77ULL, "llac_lt_0_lowtown_-65536_6.dds"},
	{0x1696C04F2994F20FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.open.gad"},
	{0x1697052DADF40893ULL, "vxw_bonepit_1jf_0_0.vlm"},
	{0x169710C8A09E88ADULL, "vxht_mansion2b_3c6_0_0.vlm"},
	{0x16975A925A076F37ULL, "vxw_encounter2_61_0_0.vlm"},
	{0x1697649754FAA988ULL, "gen00fl_varric6191283.cl"},
	{0x169795CD025EEE92ULL, "oil_proj_dot_g.mao"},
	{0x169797D8DF894EFFULL, "vxw_encounter2_na_0_0.vlm"},
	{0x1697B00FF33F8E47ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_6y.dds"},
	{0x1697B6CA657CC57DULL, "6051810_m.fxe"},
	{0x1697B7485798FCF3ULL, "257130584.ogg"},
	{0x1698874BBE2F371BULL, "113295998.ogg"},
	{0x1698F3C37B6F8C2BULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.impa.evt"},
	{0x16998E431676790DULL, "mrl310_party_comments6142850.cl"},
	{0x1699B303B5BB1674ULL, "cod_bks_lite_vials_3.utp"},
	{0x1699C7D3B7B27FFDULL, "llac_lt_0_lowtown_a2_65536_12.dds"},
	{0x169A6B2A247D129AULL, "6166621_m.fxe"},
	{0x169BE1308395151BULL, "vxht_chantry_n_2id_0_0.vlm"},
	{0x169C06E5555D71B9ULL, "dae110ip_at_gallows.utp"},
	{0x169C3DCD72F86B21ULL, "6089485_f.fxe"},
	{0x169CDE09BA155974ULL, "mcr370_meredith6083598.cl"},
	{0x169D2F9DCDA26C5CULL, "vxdr_primevaltomb_130_0_0.vlm"},
	{0x169D54C36829B02AULL, "dae151ar_mount_caverns.ncs"},
	{0x169D5E6E94985302ULL, "6169879_m.fxe"},
	{0x169D98D854F188D6ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.anb"},
	{0x169DD756D7B0C615ULL, "749165030.ogg"},
	{0x169DE46D072797AFULL, "mag360_samson6044889.cl"},
	{0x169DEDF6767CCFEEULL, "vxca_caverns2prc_9ga_0_0.vlm"},
	{0x169EA3B13376A785ULL, "715822589.ogg"},
	{0x169ED8FB9A4DCB67ULL, "mag110_arianni6016833.cl"},
	{0x169F020D43E8F918ULL, "lt_gallowstemplar_8165_ll.mmh"},
	{0x169F351C4F3D4462ULL, "vxlt_hangedman_30v_0_0.vlm"},
	{0x169F62EEFC299554ULL, "hlac_ca_caverns2prc_5802_51x.dds"},
	{0x16A00CC37B75D5ECULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.impa.gad"},
	{0x16A034BA891A0C66ULL, "525194712.ogg"},
	{0x16A05378C48C940FULL, "107402403.ogg"},
	{0x16A05B5A70F079FBULL, "vxlt_gallowscourtyard_n_4v6_0_0.vlm"},
	{0x16A09700CB507246ULL, "vxlt_0_lowtown_docks_n_cvt_0_0.vlm"},
	{0x16A0D4D749B34BA7ULL, "root_mage_spirit.dds"},
	{0x16A0EBC1C31E01D2ULL, "6117962_m.fxe"},
	{0x16A0F928200C5A93ULL, "docks_male_overseer_mob6176611.cl"},
	{0x16A112925D3E4C2FULL, "vxht_brothel_n_6b6_0_0.vlm"},
	{0x16A1E05462705575ULL, "w_bonepitmining_a2_36_0.msh"},
	{0x16A1E254627058EBULL, "w_bonepitmining_a2_36_0.mmh"},
	{0x16A1EE5462706C88ULL, "w_bonepitmining_a2_36_0.mao"},
	{0x16A1F2FB0F30009AULL, "6089416_m.fxe"},
	{0x16A2058B68BE3961ULL, "6107999_m.fxe"},
	{0x16A25890C420DFE9ULL, "6082223_m.fxe"},
	{0x16A3518F61E744ABULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_higharmoire_d.dds"},
	{0x16A351E67D3AF009ULL, "vxht_0_hightown_a6c_0_0.vlm"},
	{0x16A3683E06578043ULL, "6135956_m.fxe"},
	{0x16A3A826DCE5EF1EULL, "hlac_lt_0_lowtown_a3_n_c_-2_5y.dds"},
	{0x16A3B414ECB01A2EULL, "6169905_m.fxe"},
	{0x16A3EB86F342FEDFULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown3.msh"},
	{0x16A3F186F3430881ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown3.mmh"},
	{0x16A41B89841D8369ULL, "6051611_f.fxe"},
	{0x16A43CCA426A5202ULL, "qun210cr_guard6072102.cl"},
	{0x16A44CFA7318D839ULL, "168068112.ogg"},
	{0x16A46CC3682FF7BBULL, "dae151ar_mount_caverns.lst"},
	{0x16A48F98C9AD38C0ULL, "vxdr_primevalentrance_2zi_0_1.vlm"},
	{0x16A51F92F2E23CB9ULL, "ser270_emeric6113348.cl"},
	{0x16A56BE813EBEBDFULL, "vxca_caverns1_mines_ebd_0_0.vlm"},
	{0x16A57E56C27DC372ULL, "vxw_sundermount_23h_0_0.vlm"},
	{0x16A582A59294A12DULL, "vxht_mansion2b_3f5_0_0.vlm"},
	{0x16A589479333ABFEULL, "283489211.ogg"},
	{0x16A5B1E6C727B8CAULL, "llac_lt_0_lowtown_docks_n_-196610_3.dds"},
	{0x16A5CBFB7F01C64EULL, "vxlt_0_lowtown_a3_n_c_hb5_0_0.vlm"},
	{0x16A6016D6E8D7D14ULL, "vxw_woundedcoast_14t_0_0.vlm"},
	{0x16A6125136813CD8ULL, "1005704892.ogg"},
	{0x16A614B0A8A1E2E7ULL, "392454273.ogg"},
	{0x16A61A3E3F3DC93FULL, "hlac_ca_cave_2194_6z.dds"},
	{0x16A65DA1F53D0C73ULL, "6065316_m.fxe"},
	{0x16A6B29D25BFF5DDULL, "llac_lt_0_lowtown_n_-65536_5.dds"},
	{0x16A6BF62F24B861BULL, "vxlt_0_lowtown_a2_n_c_ei5_0_0.vlm"},
	{0x16A6CA6D4D6FEF08ULL, "vxlt_gallowstemplar_n_6sf_0_0.vlm"},
	{0x16A6DCABE4F16D8FULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_hed_0n.dds"},
	{0x16A708BF13742DC2ULL, "vxdr_primevaltomb_5ap_0_0.vlm"},
	{0x16A70BC37B7BF2EAULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.impa.ani"},
	{0x16A75B38D6052788ULL, "6092490_m.fxe"},
	{0x16A789EAA5C6B2D0ULL, "590369200.ogg"},
	{0x16A80A3D4E3756F4ULL, "6111381_f.fxe"},
	{0x16A8D77BD2AFC8B6ULL, "llac_lt_0_lowtown_a2_n_65536_7.dds"},
	{0x16A9087D4338C1CEULL, "433869993.ogg"},
	{0x16A976F97C405EC9ULL, "vxlt_backalley_n_1jk_0_0.vlm"},
	{0x16A9A1926B53B271ULL, "230525901.ogg"},
	{0x16A9AB4D34CB7BE3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1b.mao"},
	{0x16A9AF4D34CB8230ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1b.mmh"},
	{0x16A9BD4D34CB9A4AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1b.msh"},
	{0x16A9E59AAB50B5B6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen.ani"},
	{0x16AA1D48DCA727ABULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d01_0t.dds"},
	{0x16AA1EAF59F75BDAULL, "art\\characters\\weapons\\longbows\\w_lbw_003a.mmh"},
	{0x16AA22AF59F76211ULL, "art\\characters\\weapons\\longbows\\w_lbw_003a.mao"},
	{0x16AA30AF59F77A20ULL, "art\\characters\\weapons\\longbows\\w_lbw_003a.msh"},
	{0x16AA403AA6DC0840ULL, "6075790_m.fxe"},
	{0x16AA431AD904D049ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small.phy"},
	{0x16AA9BD854FC1841ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.mmh"},
	{0x16AADD4F29A5CCD5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.open.ani"},
	{0x16AB2758F8393597ULL, "1023483976.ogg"},
	{0x16AB49BF84F628D3ULL, "lt_gallowscourtyard_probe_7az.dds"},
	{0x16AB975055478E1BULL, "938887730.ogg"},
	{0x16ABFE6B7AA436BEULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_0z.dds"},
	{0x16ACA3AAFC3480A8ULL, "ht_0_hightown_a2_42_0_ll.mmh"},
	{0x16AD2346A8181494ULL, "vxdr_0_deeproads_bpa_0_1.vlm"},
	{0x16AD4F5492B7F47CULL, "combat_smaoe_r1_a1.gda"},
	{0x16AD77A150AD5ECFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\vst_hightown_sa.dds"},
	{0x16ADB69D3CF21680ULL, "m_fen110ar_abandoned_house.wbd"},
	{0x16ADCED8263F2AF4ULL, "gen00fl_anders6111186.cl"},
	{0x16AE2A00D3A99A1FULL, "lgt000cr_elf_turn_in_1.utc"},
	{0x16AE3AC735FBAEE4ULL, "6163606_m.fxe"},
	{0x16AE3F3092C16E1FULL, "gen00fl_varric6118784.cl"},
	{0x16AE5E4CBF276E89ULL, "6126412_f.fxe"},
	{0x16AF08DDBD468AC6ULL, "vxlt_0_lowtown_docks_n_cqu_0_0.vlm"},
	{0x16AF3858536A3DEDULL, "6033555_m.fxe"},
	{0x16AF44E08ECF6128ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardb.mmh"},
	{0x16AF62E08ECF9452ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardb.msh"},
	{0x16AFABDAF3D5764BULL, "vxca_caverns2_83e_0_1.vlm"},
	{0x16B031A122234758ULL, "hlac_blank_exterior_0_0x.dds"},
	{0x16B09D7DCDB5226CULL, "art\\characters\\playercharacter\\elfmale\\em_boo_lgte.mmh"},
	{0x16B0A37DCDB52C5EULL, "art\\characters\\playercharacter\\elfmale\\em_boo_lgte.msh"},
	{0x16B0B0D8A1B2FDFFULL, "6134826_f.fxe"},
	{0x16B1440860948FEAULL, "hlac_lt_0_lowtown_docks_a3_n_131071_0y.dds"},
	{0x16B196C43FF530CCULL, "hlac_lt_0_lowtown_docks_a3_-4_4x.dds"},
	{0x16B1BFDDE88F0774ULL, "28367279.ogg"},
	{0x16B1D9C6C336AB58ULL, "vxht_mansion_n_631_0_0.vlm"},
	{0x16B1DCD2DB99C38DULL, "183758058.ogg"},
	{0x16B21EE8D3AFD44EULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_altar.phy"},
	{0x16B220B3A88A860AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfireout.dura.gad"},
	{0x16B240E1FD1F0968ULL, "llac_lt_0_lowtown_0_7.dds"},
	{0x16B2A42CCDCBE7FCULL, "dae000st_arrivalprimeval.stg"},
	{0x16B3AB3A24C0F918ULL, "dre000_dougal6131144.cl"},
	{0x16B3D6B374BBE35CULL, "vxht_0_hightown_a2_c4v_0_0.vlm"},
	{0x16B3DDC505E2045FULL, "vxca_caverns1_mines_egg_0_0.vlm"},
	{0x16B3F033B473DBF2ULL, "vxw_sundermount_24i_0_0.vlm"},
	{0x16B4644B075319ABULL, "6163797_m.fxe"},
	{0x16B47310B637AF1DULL, "hlac_lt_0_lowtown_a3_n_c_131070_3y.dds"},
	{0x16B518A414D8D9F5ULL, "v_c_blood_05.dds"},
	{0x16B547EB2E92BBEFULL, "6125441_m.fxe"},
	{0x16B602EE48819CF6ULL, "6080988_f.fxe"},
	{0x16B62FF660360FA5ULL, "gen00fl_dog6107446.cl"},
	{0x16B64523AAF7575DULL, "dae011_wench_mob6144869.cl"},
	{0x16B6F9F2FC08F5B9ULL, "ht_0_hightownexplosion_n_38_0.rml"},
	{0x16B742FAC2A2A5D7ULL, "mag100_karras6034692.cl"},
	{0x16B75D0F0A0846E9ULL, "mcr370_samson6087530.cl"},
	{0x16B7ED9AAB5D00A7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen.evt"},
	{0x16B82BF5C11966A6ULL, "vxlt_0_lowtown_n_eqb_0_0.vlm"},
	{0x16B83ABA0D7D5061ULL, "gen00fl_aveline6139070.cl"},
	{0x16B86B1C0640AD06ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen_v2.ani"},
	{0x16B88A41FE1147E6ULL, "6108438_m.fxe"},
	{0x16B903B3A89073A1ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfireout.dura.evt"},
	{0x16B9B50971C459F9ULL, "vxca_caverns2prc_6t8_0_0.vlm"},
	{0x16BA0C6876DB27EAULL, "ts_generic_prog_junk2.gda"},
	{0x16BA18F826BF324BULL, "art\\characters\\weapons\\spears\\textures\\w_spr_001a_0t.dds"},
	{0x16BA1C9CAF9C1719ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_03a_0d.dds"},
	{0x16BA647DC184A13CULL, "vxht_0_hightown_n_b75_0_0.vlm"},
	{0x16BAD53BDBB10263ULL, "vxht_keep_n_6ad_0_0.vlm"},
	{0x16BB384E1EC39416ULL, "one280_iwan6101112.cl"},
	{0x16BB50B69FBDA20AULL, "art\\vfx\\combat\\vfx_gib_hm_hed.mao"},
	{0x16BB7ACF5B4ABD25ULL, "6043260_m.fxe"},
	{0x16BBAACE7E771CD4ULL, "gen_im_trs_dwv_med_hlm_r2.uti"},
	{0x16BBD63533D25231ULL, "vxca_cavern1_haunted_hm2_0_0.vlm"},
	{0x16BC1B976C8C7703ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleft.msh"},
	{0x16BC29976C8C8F5DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleft.mmh"},
	{0x16BC2D976C8C966EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleft.mao"},
	{0x16BC6E22D3E0863DULL, "vxht_mansion_n_6nw_0_0.vlm"},
	{0x16BD53981F346934ULL, "gen00fl_aveline6139899.cl"},
	{0x16BD9FEAFD4F8C44ULL, "95129279.ogg"},
	{0x16BDF64C412A7D2FULL, "vxht_brothel_n_6h0_0_0.vlm"},
	{0x16BE0E9AAB61A4F0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen.gad"},
	{0x16BE6992C61E2E58ULL, "vxht_keep_n_6yc_0_0.vlm"},
	{0x16BF029BD0DCE79AULL, "388175680.ogg"},
	{0x16BF24B0B6C917E7ULL, "6111086_m.fxe"},
	{0x16BF84DF790697B5ULL, "art\\vfx\\combat\\vfx_m_cre_heal_act_c.mmh"},
	{0x16BFB448D2E85141ULL, "vxlt_0_lowtown_docks_n_c4u_0_0.vlm"},
	{0x16C01B42C273C117ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masb_0s.dds"},
	{0x16C0CBD0E7CB2613ULL, "zz_one000_barber6179402.cl"},
	{0x16C0CC86054A9582ULL, "hlac_ht_mansion2b_818_26z.dds"},
	{0x16C0ED7F9B764D24ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehangb.msh"},
	{0x16C0FF7F9B766B8DULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehangb.mao"},
	{0x16C1037F9B7672C6ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehangb.mmh"},
	{0x16C11B2A9C729AA1ULL, "vxlt_0_lowtown_docks_c0f_0_0.vlm"},
	{0x16C11B6A6C404DEDULL, "6148781_m.fxe"},
	{0x16C1358F572B5291ULL, "follower_banter6155748.cl"},
	{0x16C1B28EF8A4B7E2ULL, "vxlt_0_lowtown_docks_bse_0_0.vlm"},
	{0x16C203D8225A01C5ULL, "lt_backalley_n_1_0.rml"},
	{0x16C2489066B1FBDCULL, "vxht_0_hightown_a2_c3w_0_0.vlm"},
	{0x16C26210A669F472ULL, "vxw_sundermount_29j_0_0.vlm"},
	{0x16C319A878D667F7ULL, "76837259.ogg"},
	{0x16C360937BB5C5CEULL, "gen00fl_anders6125330.cl"},
	{0x16C383C4DACB86F3ULL, "one100_ht_banter.bnk"},
	{0x16C3895CB737774BULL, "6090190_m.fxe"},
	{0x16C4130B8018B21EULL, "6152806_m.fxe"},
	{0x16C419E70971BC3CULL, "hlac_w_bonepitdestroyed_a3_196608_0z.dds"},
	{0x16C45151505F1A06ULL, "dae300cr_meredith.utc"},
	{0x16C5C18506F22E89ULL, "724743422.ogg"},
	{0x16C5E2080EE4C826ULL, "vxht_chantry_352_0_1.vlm"},
	{0x16C5FBB3A89AF0F0ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfireout.dura.ani"},
	{0x16C5FC8C2818EDF5ULL, "fen210_hadriana6110372.cl"},
	{0x16C6531C064CC277ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen_v2.evt"},
	{0x16C6553E2D32B571ULL, "vxht_0_hightown_a2_b7k_0_0.vlm"},
	{0x16C67D57B09B2346ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls108a_0e.dds"},
	{0x16C69DD2B30F7F26ULL, "vxlt_0_lowtown_n_ejc_0_0.vlm"},
	{0x16C6A5902A04C97CULL, "qun120_arishok_guard6054006.cl"},
	{0x16C6DD97601C6E2BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexwild_sa.dds"},
	{0x16C6E7BDD2F818BFULL, "ef_dead_3.mor"},
	{0x16C7117542AE40FEULL, "hlac_lt_0_lowtown_docks_n_65532_1y.dds"},
	{0x16C7973E7221CD0FULL, "hlac_ca_cavern1_haunted_24425_10y.dds"},
	{0x16C7A5C4DACFA943ULL, "one100_ht_banter.cnv"},
	{0x16C7BF6E6AD6C1F2ULL, "360615469.ogg"},
	{0x16C7F13BFC1BE0F2ULL, "hlac_w_blightlands_262146_0x.dds"},
	{0x16C8A0044F1730BEULL, "899567237.ogg"},
	{0x16C964FB12E324DBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_baserocks.mmh"},
	{0x16C970FB12E33978ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_baserocks.mao"},
	{0x16C982FB12E357C5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_baserocks.msh"},
	{0x16C9AFC6ACF2540AULL, "703979764.ogg"},
	{0x16C9B48CE8D05B91ULL, "618721211.ogg"},
	{0x16C9D41967347154ULL, "lgt220_mag_apost_apost16156738.cl"},
	{0x16C9E4927BE838FAULL, "6169303_m.fxe"},
	{0x16CA27F3B3814C9DULL, "154591385.ogg"},
	{0x16CAF4B7719FC748ULL, "310504649.ogg"},
	{0x16CB06224C91950CULL, "hlac_dr_slavers_3214_11x.dds"},
	{0x16CBC50B824B5C7DULL, "353435136.ogg"},
	{0x16CC541C06512F00ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen_v2.gad"},
	{0x16CC61A3E530254DULL, "6155619_m.fxe"},
	{0x16CC6829332095AFULL, "vxht_brothel_n_6u3_0_0.vlm"},
	{0x16CC97D96C762C76ULL, "313694284.ogg"},
	{0x16CCFEB323A13E4FULL, "6016000_m.fxe"},
	{0x16CD35A45E3D166CULL, "vxw_sundermount_32q_0_1.vlm"},
	{0x16CD81DF7912D042ULL, "art\\vfx\\combat\\vfx_m_cre_heal_act_c.anb"},
	{0x16CDBF49A6B05967ULL, "art\\characters\\creatures\\spider\\animation\\blendtree\\spider_overlay.anb"},
	{0x16CE0A2C946E819AULL, "6023892_m.fxe"},
	{0x16CE0D1B488B9839ULL, "6048258_m.fxe"},
	{0x16CE7D48C0B51F10ULL, "vxca_cavern1_haunted_i85_0_0.vlm"},
	{0x16CEC1298B391FB8ULL, "gen00fl_anders6116158.cl"},
	{0x16CF47D2FCCB7F6FULL, "hlac_lt_gallowstemplar_n_2384_1x.dds"},
	{0x16CF7639C4FD5DAAULL, "6078419_f.fxe"},
	{0x16CFA8130C8C1845ULL, "rdr300st_martin.utm"},
	{0x16D07DC368554FF8ULL, "dae151ar_mount_caverns.are"},
	{0x16D09982E926B203ULL, "lt_warehouse_n_3_0.msh"},
	{0x16D0A782E926CA5DULL, "lt_warehouse_n_3_0.mmh"},
	{0x16D0AB82E926D16EULL, "lt_warehouse_n_3_0.mao"},
	{0x16D0D4747FAF68AAULL, "llac_lt_0_lowtown_docks_a3_-196610_5.dds"},
	{0x16D0EAD119F33A31ULL, "one350_variel6085547.cl"},
	{0x16D175624C2405B7ULL, "vxw_blightlands_l8_0_0.vlm"},
	{0x16D1D850C5C5150BULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.cess.evt"},
	{0x16D1FD921059098FULL, "vxlt_undercityencounter_13f_0_1.vlm"},
	{0x16D281AFCB67111BULL, "777164960.ogg"},
	{0x16D2C7158DBCEFB0ULL, "qcr221_cut_isabelas_return.cnv"},
	{0x16D31C9924841E85ULL, "29561499.ogg"},
	{0x16D3393AD8489630ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_posta.mao"},
	{0x16D33D3AD8489D43ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_posta.mmh"},
	{0x16D34B3AD848B49DULL, "art\\levels\\races\\elf\\props\\dalish\\prp_posta.msh"},
	{0x16D3CEE5951C9382ULL, "gen00fl_anders6125756.cl"},
	{0x16D3EB990FBBE426ULL, "hlac_w_encounter1_65540_0y.dds"},
	{0x16D423BB973B319BULL, "310935790.ogg"},
	{0x16D4C3BDB37E7FAFULL, "344544528.ogg"},
	{0x16D4C71B1F28CDF1ULL, "vxht_0_hightown_a2_b8j_0_0.vlm"},
	{0x16D4CB152A62D2B7ULL, "264322503.ogg"},
	{0x16D4DC340BCAF9A9ULL, "vxca_caverns2prc_87s_0_0.vlm"},
	{0x16D5D62DDCCCAD18ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screenb_1.mao"},
	{0x16D5E9158DBF5F08ULL, "qcr221_cut_isabelas_return.bnk"},
	{0x16D603C64149E6EFULL, "llac_w_encounter2_65536_0.dds"},
	{0x16D63EBAED8BC99EULL, "and100_anders6096027.cl"},
	{0x16D66D14317FCA42ULL, "hlac_lt_undercitydungeon_-1_6y.dds"},
	{0x16D6ABD174FEB86AULL, "gen00fl_fenris6116543.cl"},
	{0x16D72C1AD92B976EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small.mmh"},
	{0x16D7381AD92BAB05ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small.mao"},
	{0x16D7461AD92BC35CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small.msh"},
	{0x16D82D5AC4F3C79EULL, "vxdr_slavers_2hi_0_1.vlm"},
	{0x16D8528D806B9F4BULL, "358174977.ogg"},
	{0x16D85AA846DD10F7ULL, "fxa_web_flare.dds"},
	{0x16D898DFA1198C5AULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scl_a.mao"},
	{0x16D8F150C5CB5ECCULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.cess.gad"},
	{0x16D90BC86560B569ULL, "670746908.ogg"},
	{0x16D91C3E53407500ULL, "hlac_lt_warehouse_n_0_3y.dds"},
	{0x16D91D9424AE289EULL, "208681852.ogg"},
	{0x16D94D85AFF85622ULL, "art\\characters\\weapons\\dagger\\w_dger_102a.msh"},
	{0x16D94F85AFF859D8ULL, "art\\characters\\weapons\\dagger\\w_dger_102a.mmh"},
	{0x16D9B6E5F7901640ULL, "docks_male_laborer_mob6175404.cl"},
	{0x16D9F6E20557F486ULL, "vxlt_0_lowtown_docks_a3_n_cy2_0_0.vlm"},
	{0x16DA1BAA8F3FC564ULL, "vxlt_hangedman_n_m5_0_0.vlm"},
	{0x16DA852B86AC1239ULL, "vxht_brothel_n_4yw_0_0.vlm"},
	{0x16DB088ACCDD2128ULL, "251009890.ogg"},
	{0x16DB7E933CF15DC2ULL, "476492587.ogg"},
	{0x16DB86C2CD9ED911ULL, "vxlt_darktown_6c0_0_0.vlm"},
	{0x16DBA92DE7E6559AULL, "vxht_keep_n_63g_0_0.vlm"},
	{0x16DBEC497D9C5DE9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medd_0t.dds"},
	{0x16DC551FD40E28ACULL, "6016586_m.fxe"},
	{0x16DC622AD6F765EBULL, "vxht_chantry_n_2b1_0_0.vlm"},
	{0x16DC650BE6C39A6DULL, "659326365.ogg"},
	{0x16DCC27A27F7B9B2ULL, "character_character_p69.psh.fxo.dx11.assembly"},
	{0x16DCC2BE7A485E6EULL, "zz_pro_debug6059743.cl"},
	{0x16DCEBF575EC9E17ULL, "vxlt_hangedman_26l_0_0.vlm"},
	{0x16DD300DB2885880ULL, "hlac_ca_caverns2prc_5802_41y.dds"},
	{0x16DDB3DD23C4D64BULL, "hlac_lt_0_lowtown_131072_6x.dds"},
	{0x16DDD52473616E7AULL, "6150691_m.fxe"},
	{0x16DE01E81B9B7018ULL, "follower_banter6141896.cl"},
	{0x16DE0C8A6F726EA9ULL, "444953745.ogg"},
	{0x16DE516294A146F0ULL, "face_face_p21.psh.fxo.dx11"},
	{0x16DE7DE82A838C43ULL, "133529745.ogg"},
	{0x16DED55F89ED1042ULL, "341751717.ogg"},
	{0x16DEFC75FFCE9541ULL, "fen320ip_to_base.utp"},
	{0x16DF2235EF035430ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzcart.mao"},
	{0x16DF2635EF035B43ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzcart.mmh"},
	{0x16DF3435EF03729DULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzcart.msh"},
	{0x16DF489E574D3E43ULL, "6109788_m.fxe"},
	{0x16DF7DE35E6E190FULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_001a_0d.dds"},
	{0x16DFF050C5D17BCAULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.cess.ani"},
	{0x16E03F4DC896EC51ULL, "llac_lt_0_lowtown_docks_a3_n_-3_5.dds"},
	{0x16E0D2C551974C39ULL, "gen00fl_varric6076015.cl"},
	{0x16E0EE322EAAF7D5ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drgb_0s.dds"},
	{0x16E16D6FFD531DA3ULL, "977917551.ogg"},
	{0x16E17F84F2B7AE4BULL, "300265933.ogg"},
	{0x16E18DA06F17574BULL, "vxdr_primevaltomb_3m2_0_1.vlm"},
	{0x16E195D746E4CCEBULL, "6091412_m.fxe"},
	{0x16E1A7CAA32F57D8ULL, "vxht_0_hightown_n_b8l_0_0.vlm"},
	{0x16E1D6885EBAACEEULL, "gen00fl_sebastian6195808.cl"},
	{0x16E23267203DEE98ULL, "qcr220_grey_wardens6207475.cl"},
	{0x16E3125E08CD14EDULL, "llac_lt_hangedman_-1_13.dds"},
	{0x16E36F70A54658EEULL, "w_encounter3_72_0.mmh"},
	{0x16E373B4D81B509BULL, "vxlt_0_lowtown_a2_n_c_h2a_0_0.vlm"},
	{0x16E37B70A5466C85ULL, "w_encounter3_72_0.mao"},
	{0x16E38970A54684DCULL, "w_encounter3_72_0.msh"},
	{0x16E3B7FD563D9C67ULL, "782069911.ogg"},
	{0x16E3BF3AE7B48017ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\harvester_overlay.anb"},
	{0x16E3C7C4B1E18CD6ULL, "zz_followers_banter6157987.cl"},
	{0x16E43E3AD8571268ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_posta.phy"},
	{0x16E45685B0026D63ULL, "art\\characters\\weapons\\dagger\\w_dger_102a.phy"},
	{0x16E4635B7B33C75CULL, "seb321_leliana6180938.cl"},
	{0x16E4E5627970FBA9ULL, "vxht_mansion_n_54j_0_0.vlm"},
	{0x16E58C78D48B94BAULL, "hightown_male_noble_gossip6171387.cl"},
	{0x16E59485042C0C5FULL, "llac_lt_0_lowtown_n_-65537_6.dds"},
	{0x16E5AD488EF38117ULL, "vxdr_dungeon_1sq_0_0.vlm"},
	{0x16E6C8CB1444FFC2ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_2x.dds"},
	{0x16E6ED1984636ACCULL, "6131182_m.fxe"},
	{0x16E7AFEA87A96AF0ULL, "crt_spblm_dot.mao"},
	{0x16E8019B9231B664ULL, "256287905.ogg"},
	{0x16E80950E6991F32ULL, "hlac_ht_chantry_2534_41z.dds"},
	{0x16E855BBB3D4BA78ULL, "262309141.ogg"},
	{0x16E899359507CAC9ULL, "6052147_m.fxe"},
	{0x16E89D013382A7D7ULL, "344076715.ogg"},
	{0x16E913B570FC7C8EULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viale.mmh"},
	{0x16E9198A6437250CULL, "vxlt_gallowstemplar_f_3rp_0_0.vlm"},
	{0x16E91FB570FC9125ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viale.mao"},
	{0x16E92DB570FCA97CULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viale.msh"},
	{0x16E9B64D3501724BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1b.phy"},
	{0x16EA8BE1C6819898ULL, "mi_dot.mao"},
	{0x16EB37B7E1D6CFF6ULL, "1038908690.ogg"},
	{0x16EB9439E88B0D64ULL, "fen200ip_exit_quest.utp"},
	{0x16EBF82E19361A45ULL, "483588988.ogg"},
	{0x16EC065A4367F8A5ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_chg_p.mmh"},
	{0x16EC712A90495FD3ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_onyxstatue.phy"},
	{0x16EC904358FCB806ULL, "6109428_m.fxe"},
	{0x16ED15A27AD29851ULL, "vxlt_warehouse_n_a4_0_0.vlm"},
	{0x16ED5DF1A130C8BEULL, "hlac_w_woundedcoast_-262137_0y.dds"},
	{0x16EDB565E02D4871ULL, "vxht_0_hightown_n_au9_0_0.vlm"},
	{0x16EDBA7B494356C3ULL, "6102688_m.fxe"},
	{0x16EDF06C90AB9ECFULL, "vxw_woundedcoast_5a_0_0.vlm"},
	{0x16EE1389F43A500DULL, "llac_lt_0_lowtown_65536_9.dds"},
	{0x16EE6AD8ED4048FAULL, "vxlt_0_lowtown_e1i_0_0.vlm"},
	{0x16EE773FAF7934DDULL, "llac_lt_gallowstemplar_n_2384_3.dds"},
	{0x16EEF743B5C07102ULL, "vxht_0_hightown_c3t_0_1.vlm"},
	{0x16EF858602E9E125ULL, "209070140.ogg"},
	{0x16EFCB75322299FEULL, "ter_stoney_s.dds"},
	{0x16F007B438DAE56BULL, "6096512_m.fxe"},
	{0x16F02735EF11D068ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzcart.phy"},
	{0x16F02BEE6572C819ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb107a_0t.dds"},
	{0x16F04F26AEDAC982ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.mmh"},
	{0x16F08AC72B0E1C4AULL, "combat_ws_r2_a1.gda"},
	{0x16F0D03EB09A141FULL, "202445116.ogg"},
	{0x16F15965FB5DA197ULL, "follower_banter6164023.cl"},
	{0x16F1A0D9C7E1DC8BULL, "follower_banter6170259.cl"},
	{0x16F21DE8D3E5B611ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_altar.mmh"},
	{0x16F229E8D3E5CAB2ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_altar.mao"},
	{0x16F237E8D3E5E2CFULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_altar.msh"},
	{0x16F28E95B9B56849ULL, "ser200_aveline6201588.cl"},
	{0x16F2E08373CD39A3ULL, "6175479_m.fxe"},
	{0x16F2EA5F77D9F215ULL, "static1_punch.psh.fxo"},
	{0x16F33F7F13AE223AULL, "6165242_m.fxe"},
	{0x16F3AE05FFAE0519ULL, "zz_seb_debug6214535.cl"},
	{0x16F502D8228559EDULL, "lt_backalley_n_1_0.anb"},
	{0x16F54ED9663AD87FULL, "476750350.ogg"},
	{0x16F5C30320975096ULL, "zz_dae_debug6184888.cl"},
	{0x16F5D031F7C8AD58ULL, "vxlt_0_lowtown_a2_eoe_0_0.vlm"},
	{0x16F63EA366B96D1EULL, "vxca_cave_2b7_0_0.vlm"},
	{0x16F671BDBA11AEACULL, "6077367_m.fxe"},
	{0x16F7012ACFE9E322ULL, "ser191_emeric6108824.cl"},
	{0x16F7E8F13799FB31ULL, "vxca_caverns2prc_acn_0_0.vlm"},
	{0x16F80A1B4B4BC74BULL, "vxdr_dungeon_1mk_0_0.vlm"},
	{0x16F84C9324BBDD84ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube.msh"},
	{0x16F85894B69CA224ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtb.phy"},
	{0x16F85E9324BBFCEDULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube.mao"},
	{0x16F8629324BC0326ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube.mmh"},
	{0x16F8AB168A75E59EULL, "744408540.ogg"},
	{0x16F9035A43727E72ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_chg_p.anb"},
	{0x16F914102E68AE56ULL, "vxca_caverns2_8m7_0_0.vlm"},
	{0x16F970CF0CECD4D6ULL, "6105631_m.fxe"},
	{0x16F97310C1FB241FULL, "267966087.ogg"},
	{0x16F9BB2B90C6B69EULL, "vxlt_0_lowtown_docks_cs9_0_1.vlm"},
	{0x16F9EFD7A88B927FULL, "vxht_brothel_n_6es_0_0.vlm"},
	{0x16FA25799B3CC2C0ULL, "lt_0_lowtown_a2_n_c_38_0_ll.mmh"},
	{0x16FA6F32CDDE60ABULL, "lt_undercityencounter_3567.anb"},
	{0x16FAE37B82AA6CC2ULL, "6139896_m.fxe"},
	{0x16FB2423B9282126ULL, "hlac_w_encounter1_-65535_1y.dds"},
	{0x16FB2DED3D078E70ULL, "6115351_m.fxe"},
	{0x16FB7A02C07D19E2ULL, "vxlt_0_lowtown_docks_bga_0_0.vlm"},
	{0x16FBE2FA4249C90EULL, "210778434.ogg"},
	{0x16FC2B7833198A35ULL, "1034799015.ogg"},
	{0x16FC819D2B29494EULL, "vxht_brothel_n_6dq_0_0.vlm"},
	{0x16FCDD3E5EE1C27CULL, "6163888_m.fxe"},
	{0x16FD083FA7882EFFULL, "vxw_blightlands_7e_0_0.vlm"},
	{0x16FDE6CB6B12709BULL, "935171214.ogg"},
	{0x16FEB92BA73558A8ULL, "6127563_m.fxe"},
	{0x16FF79FF156B0F3BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_walllcap.msh"},
	{0x16FF7F67705B95BCULL, "713020706.ogg"},
	{0x16FF8BFF156B2D86ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_walllcap.mao"},
	{0x16FF97FF156B4225ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_walllcap.mmh"},
	{0x16FFAB6F48DCDF32ULL, "vxht_mansion2_3dl_0_0.vlm"},
	{0x16FFB986BF89E137ULL, "30118811.ogg"},
	{0x16FFE9E203AD76B4ULL, "358047434.ogg"},
	{0x170051E80DCB7ED4ULL, "255776989.ogg"},
	{0x1700C86DD1944861ULL, "art\\characters\\playercharacter\\humanmale\\bdy_rob_magc.mao"},
	{0x1700FE3AC5E2EB56ULL, "92347052.ogg"},
	{0x1701078A250822D5ULL, "132665196.ogg"},
	{0x1701454238127AF3ULL, "201032571.ogg"},
	{0x170146A9F2529C6CULL, "67904057.ogg"},
	{0x170167F49FCC392EULL, "hlac_lt_backalley_n_0_16z.dds"},
	{0x170204411231AEBFULL, "w_encounter3_96_0.phy"},
	{0x17048C6147A4EBE3ULL, "153701365.ogg"},
	{0x1704C5A41FD4989EULL, "mag371_elsa6065016.cl"},
	{0x1704F8A6F3F272C4ULL, "ser120_ghyslain6104334.cl"},
	{0x17052F870FD1DF7DULL, "default_miss01.mao"},
	{0x170548B393025DE9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beampile.mao"},
	{0x170554B393027182ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beampile.mmh"},
	{0x170556B393027538ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beampile.msh"},
	{0x17057BC72FF16C30ULL, "mcr370_desten.cnv"},
	{0x1706547C1DF3DB68ULL, "combat.arl"},
	{0x1706609F4D0B324EULL, "6071510_m.fxe"},
	{0x1706B16EBF64C03DULL, "qun120_arishok6051310.cl"},
	{0x1706B2D02FE47006ULL, "mcr310_argument6075141.cl"},
	{0x170764EF6F85B446ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_3y.dds"},
	{0x170806BDFCD3AD32ULL, "193525148.ogg"},
	{0x17089DC72FF3DB88ULL, "mcr370_desten.bnk"},
	{0x1708B8C302BA2408ULL, "6068373_f.fxe"},
	{0x170931C390B76A7DULL, "6216710_m.fxe"},
	{0x17095ED2549F81E8ULL, "art\\vfx\\models\\vfx_sphere_deform.msh"},
	{0x17097CD2549FB412ULL, "art\\vfx\\models\\vfx_sphere_deform.mmh"},
	{0x1709924A865C7C20ULL, "mag321cr_servant_distract.utc"},
	{0x170AC2581E49D736ULL, "732876344.ogg"},
	{0x170ACAFA2A04992BULL, "6110697_f.fxe"},
	{0x170AE855C29AFE19ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowflat.msh"},
	{0x170AF255C29B0F07ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowflat.mmh"},
	{0x170AF37A1D1F61CEULL, "vxht_brothel_n_6ar_0_0.vlm"},
	{0x170AF655C29B1654ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowflat.mao"},
	{0x170B096D5BF81300ULL, "488591445.ogg"},
	{0x170B15B5362B6D47ULL, "6024535_m.fxe"},
	{0x170B7A1C997E477FULL, "vxw_blightlands_8d_0_0.vlm"},
	{0x170BBE79AC247F20ULL, "mcr370_assignments6081789.cl"},
	{0x170C40B0465D62F1ULL, "uh_hed_olda_0n.dds"},
	{0x170C8B70681DA785ULL, "940035127.ogg"},
	{0x170CBFF51DEEE9D2ULL, "6161093_m.fxe"},
	{0x170CDEBB1D47B97EULL, "601083167.ogg"},
	{0x170CF20B0A5D4653ULL, "234683670.ogg"},
	{0x170D7B6A51EB4785ULL, "6017661_m.fxe"},
	{0x170DCA656576A617ULL, "gen00fl_fenris6126419.cl"},
	{0x170DEAC414F44B9FULL, "mrl000_marethari6068817.cl"},
	{0x170E8FC9D0733F0BULL, "204991989.ogg"},
	{0x170ECE015226DAD1ULL, "llac_lt_darktown_-1_36.dds"},
	{0x170F2BFEAFEFDAFBULL, "cod_ltr_flora_diary_0.utp"},
	{0x171185F056D1C53EULL, "605160857.ogg"},
	{0x17120E49D3D9D25CULL, "6081436_m.fxe"},
	{0x171293B5034917D9ULL, "592217058.ogg"},
	{0x1712B4C3833105BAULL, "statecnt_bodybag.gda"},
	{0x1712B5ABA5142873ULL, "vrc221_hugin6042388.cl"},
	{0x17130523D1104EB2ULL, "450118304.ogg"},
	{0x17130541124023C4ULL, "w_encounter3_96_0.mmh"},
	{0x1713094112402A17ULL, "w_encounter3_96_0.mao"},
	{0x17131B4112404966ULL, "w_encounter3_96_0.msh"},
	{0x171445DAA2722D4DULL, "6075221_m.fxe"},
	{0x171487B94FEECAC0ULL, "6034515_m.fxe"},
	{0x17149C7B641B3B6BULL, "vxdr_0_deeproads_es8_0_1.vlm"},
	{0x1714AF2434D181E1ULL, "237319778.ogg"},
	{0x1714C17DBF1354EEULL, "6125494_f.fxe"},
	{0x1714E03F8A2C32E6ULL, "hlac_lt_gallowstemplar_f_2394_9x.dds"},
	{0x1714E0DEBE7A884BULL, "442622757.ogg"},
	{0x171600C3A7CA58B1ULL, "vxlt_0_lowtown_a3_n_c_hp8_0_0.vlm"},
	{0x17161B57D2294E66ULL, "vxdr_primevalentrance_2vz_0_0.vlm"},
	{0x17168426B5D8D7B4ULL, "vxlt_0_lowtown_a3_n_c_i0h_0_0.vlm"},
	{0x171750ADAF0A27E1ULL, "qun120cr_javaris6042884.cl"},
	{0x17176624B3F5D7C5ULL, "449726243.ogg"},
	{0x1717732E65BF8345ULL, "gen00fl_anders6113086.cl"},
	{0x171797C74A9B118AULL, "159451973.ogg"},
	{0x1717EAFE3FECC772ULL, "vxw_blightlands_h5_0_0.vlm"},
	{0x171835BC2BA4DCAFULL, "6033381_f.fxe"},
	{0x17184E26AEFC1A29ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.anb"},
	{0x1719360D084A463EULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkf.ani"},
	{0x171965570F157A4EULL, "vxht_brothel_n_6js_0_0.vlm"},
	{0x1719CB56363C108BULL, "vxw_woundedcoast_hz_0_0.vlm"},
	{0x1719EC37EC11DD62ULL, "6078544_m.fxe"},
	{0x171A66010F09624CULL, "darktown_fem_fereldan_sta6152800.cl"},
	{0x171A9170A1F4C887ULL, "6056726_f.fxe"},
	{0x171B10EF34726E84ULL, "6108145_f.fxe"},
	{0x171B31D20FE50252ULL, "6166193_m.fxe"},
	{0x171B9D039DB97058ULL, "257354415.ogg"},
	{0x171BC981283A9BE2ULL, "316113058.ogg"},
	{0x171C2066DE0467A5ULL, "572176513.ogg"},
	{0x171C86415135CD9CULL, "6147882_m.fxe"},
	{0x171D1D47A24B2D0BULL, "one300_murder_followers6089904.cl"},
	{0x171D385BA4A0B1F4ULL, "hlac_ht_playershouse_n_855_17y.dds"},
	{0x171E3AE4D19F0DAAULL, "vxlt_0_lowtown_docks_n_cvx_0_0.vlm"},
	{0x171EA7FD50536134ULL, "hlac_lt_hangedman_-1_25x.dds"},
	{0x171F11DAF205E70CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_homa_0s.dds"},
	{0x171F484FB5A1341FULL, "6110864_f.fxe"},
	{0x17206C370A6D8824ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_loop_elf_male.gad"},
	{0x1720EC3E30FD11D5ULL, "gen_im_trs_rdr_two_2hm_r0.uti"},
	{0x1721546E1BA74EC2ULL, "follower_banter6136661.cl"},
	{0x1721C2F1CC10D74BULL, "1007147542.ogg"},
	{0x17226BEBF1D8788CULL, "vxca_caverns2_96i_0_0.vlm"},
	{0x1722904A44B132CFULL, "371357081.ogg"},
	{0x17229C2D5EC311FEULL, "908215234.ogg"},
	{0x1723C8C9F4877AB4ULL, "vxca_caverns1_ghr_0_0.vlm"},
	{0x1723E9AE8EEADFA1ULL, "vxw_bonepit_22s_0_0.vlm"},
	{0x1724E9EFDBDFCC4AULL, "ts_generic_01.gda"},
	{0x172505D5F584ED6CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02a.phy"},
	{0x17250BF8C03F876CULL, "follower_banter6148723.cl"},
	{0x1725623BBEE863F0ULL, "6095994_m.fxe"},
	{0x1725A82A5BA91AA9ULL, "98256225.ogg"},
	{0x1725FA246630272CULL, "6161037_m.fxe"},
	{0x17264E0D0854F90FULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkf.evt"},
	{0x17265CDB31E2DFF2ULL, "vxw_blightlands_m6_0_0.vlm"},
	{0x1726957F92D9CE6EULL, "art\\levels\\races\\proxy\\props\\door\\door.phy"},
	{0x1726EF1328E3B703ULL, "6044137_m.fxe"},
	{0x17276C32CE0549F7ULL, "lt_undercityencounter_3567.rml"},
	{0x17277B370A73C002ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_loop_elf_male.ani"},
	{0x172787E92E25FEDBULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.dura.evt"},
	{0x172848443DFF5DAFULL, "hlac_lt_gallowstemplar_2394_7x.dds"},
	{0x172877E8245C16DEULL, "6152367_m.fxe"},
	{0x17289CCD4863740CULL, "6082099_m.fxe"},
	{0x1728FD1701FD48E9ULL, "vxca_cavern1_haunted_i08_0_0.vlm"},
	{0x1729A2A0BF28F7D0ULL, "lt_shop_mmap_ph.dds"},
	{0x1729E95FC0A2C51CULL, "vxlt_gallowstemplar_5ap_0_0.vlm"},
	{0x1729F1AF62A9EA38ULL, "bedsingle_n.dds"},
	{0x172A4FAA180AC324ULL, "lt_0_lowtown_docks_n_173_0.msh"},
	{0x172A5C90198608BBULL, "llac_ht_keep_3732_18.dds"},
	{0x172A61AA180AE18DULL, "lt_0_lowtown_docks_n_173_0.mao"},
	{0x172A65AA180AE8C6ULL, "lt_0_lowtown_docks_n_173_0.mmh"},
	{0x172A9661B853C423ULL, "vxht_mansion_n_4xm_0_0.vlm"},
	{0x172BAB92781704DFULL, "hlac_lt_0_lowtown_a2_n_c_-65535_7y.dds"},
	{0x172BCDBA47107524ULL, "920592487.ogg"},
	{0x172BDD34F54E75FFULL, "33347128.ogg"},
	{0x172BF7169725E1B8ULL, "mcr370_assignments6182062.cl"},
	{0x172C11DC1249EF4EULL, "vxw_sundermount_20a_0_0.vlm"},
	{0x172C5FE6421BD8EFULL, "6125636_m.fxe"},
	{0x172CBC44BA74BB54ULL, "gen00fl_aveline_post_board.stg"},
	{0x172D4F0D085B19B8ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkf.gad"},
	{0x172D5A2354D40A4DULL, "6041392_m.fxe"},
	{0x172E019D0817F195ULL, "vxw_bonepitdestroyed_a3_7x_0_0.vlm"},
	{0x172E1BFBD6BA135FULL, "vxlt_0_lowtown_a2_hfl_0_0.vlm"},
	{0x172E2AA4D1E86231ULL, "vxlt_undercitydungeon_1c4_0_0.vlm"},
	{0x172E369F2F3C2623ULL, "vxlt_gallowsprison_n_6mo_0_0.vlm"},
	{0x172E64236B61C663ULL, "vxht_keep_n_69l_0_0.vlm"},
	{0x172E80E92E2C111CULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.dura.gad"},
	{0x172EC74B1C869293ULL, "6076860_m.fxe"},
	{0x172F590E1769ACBFULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden5.msh"},
	{0x172F5F0E1769B761ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden5.mmh"},
	{0x172FA4EB25AA849DULL, "t3_eye_lbr.tnt"},
	{0x173011B31DF9635EULL, "6033152_m.fxe"},
	{0x17311D2E022A622AULL, "art\\characters\\creatures\\desiredemon\\animation\\blendtree\\desiredemon.anb"},
	{0x17314151FCB65336ULL, "dae150_chandan6148557.cl"},
	{0x173142E5C3CEB18BULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_act_c.impa.ani"},
	{0x17314CCF77CEFD11ULL, "dae000_mother6004122.cl"},
	{0x1731917D83DD9B72ULL, "llac_lt_0_lowtown_a3_n_c_-196608_4.dds"},
	{0x1731A9572F0E9474ULL, "lgt200cr_tome_desire.utc"},
	{0x1731FDB37142E3A5ULL, "lgt320_mis_resp_swordhunter6150280.cl"},
	{0x173261803125F279ULL, "hlac_ca_caverns1_mines_24424_5x.dds"},
	{0x17326EECCC7A785AULL, "mag110_samson6049811.cl"},
	{0x1732746925C20AE8ULL, "vxlt_darktown_6km_0_0.vlm"},
	{0x173282AC83DD17B1ULL, "gen00fl_fenris6128064.cl"},
	{0x173292C461C0340CULL, "hlac_dr_primevaltomb_9621_6z.dds"},
	{0x1732BAA607996426ULL, "8386207.ogg"},
	{0x1732F4C956B87565ULL, "6095872_m.fxe"},
	{0x173393E72AED2DE6ULL, "gen00fl_sebastian6189078.cl"},
	{0x1733E54910387EFCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\runleanr.evt"},
	{0x1733FF6283150E27ULL, "vxlt_0_lowtown_a2_f4t_0_0.vlm"},
	{0x173427AB5E6B53C2ULL, "mag100_thrask6037385.cl"},
	{0x17346B9710E3A2BAULL, "107001520.ogg"},
	{0x1734CEB823D8F872ULL, "vxw_blightlands_v7_0_0.vlm"},
	{0x173519760F910DAFULL, "6039181_f.fxe"},
	{0x173545E55219DBC2ULL, "vxlt_0_lowtown_gaw_0_0.vlm"},
	{0x17357FE92E322E1AULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.dura.ani"},
	{0x1735A50F7C4BE69AULL, "vxlt_undercitydungeon_146_0_0.vlm"},
	{0x1735CDAE605E9488ULL, "gen00fl_merrill6130105.cl"},
	{0x173627EE2E2C0456ULL, "6138730_m.fxe"},
	{0x1736DB55C2C0233CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowflat.phy"},
	{0x173724B9F24F8634ULL, "hlac_ht_brothel_n_3382_0y.dds"},
	{0x17372E504C9057B0ULL, "vxlt_0_lowtown_a2_n_c_feu_0_0.vlm"},
	{0x17384D8DF0B384F1ULL, "26654453.ogg"},
	{0x17385B3CB298DD9CULL, "vxlt_gallowstemplar_5jq_0_0.vlm"},
	{0x17385C3FF827F21FULL, "6169918_f.fxe"},
	{0x173877F9F9C114BCULL, "6152523_m.fxe"},
	{0x1738EB3F9CC96415ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_8x.dds"},
	{0x17396DE1628905DCULL, "310269560.ogg"},
	{0x17397FC81AA7E288ULL, "214297090.ogg"},
	{0x1739BBE5DE7E027AULL, "vxht_mansion_n_620_0_0.vlm"},
	{0x1739D93745CA4100ULL, "354673884.ogg"},
	{0x173A83B9044007CEULL, "vxw_sundermount_25b_0_0.vlm"},
	{0x173AEFB989BC73B2ULL, "ht_keep_mmap.dds"},
	{0x173B092EFEB32882ULL, "vxw_sundermount_2jc_0_0.vlm"},
	{0x173B5F7AFF4772E1ULL, "vxlt_0_lowtown_e23_0_0.vlm"},
	{0x173B70BCCF34985EULL, "753063392.ogg"},
	{0x173BE192B636D494ULL, "111475230.ogg"},
	{0x173BEDC9DAE70EE7ULL, "hlac_lt_0_lowtown_a2_n_1_8x.dds"},
	{0x173BF3D242D92DACULL, "6089708_m.fxe"},
	{0x173C8DD8C8B02BDFULL, "vxlt_0_lowtown_a2_hco_0_0.vlm"},
	{0x173CE72C11A8D7A0ULL, "6002293_m.fxe"},
	{0x173CFAB3FC50D472ULL, "6164502_m.fxe"},
	{0x173DBF6BFB60CD57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_mark_activate.gad"},
	{0x173DD114901562BDULL, "hlac_lt_warehouse_n_1_2x.dds"},
	{0x173DDB50B7BA9E7EULL, "vxlt_gallowstemplar_686_0_0.vlm"},
	{0x173E256F027ED002ULL, "zz_followers_banter6157389.cl"},
	{0x173E845D0EBF6F18ULL, "395650020.ogg"},
	{0x173EB5BDF0FE204FULL, "lt_warehouse_n_2_0_ll.mmh"},
	{0x173EE40769D93C5EULL, "753990155.ogg"},
	{0x173F125897B90A21ULL, "6092597_f.fxe"},
	{0x173F2AE5C3DAC602ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_act_c.impa.evt"},
	{0x173FFBF1A9E6A7ABULL, "636465397.ogg"},
	{0x17405E4472918821ULL, "904515315.ogg"},
	{0x1740E14D7AACE238ULL, "t1_mue_re2.tnt"},
	{0x1740EF603FC522A8ULL, "397909850.ogg"},
	{0x174189DFC6E042FBULL, "6044534_f.fxe"},
	{0x174192426CF3FB53ULL, "lt_0_lowtown_docks_a3_1_0_ll.mmh"},
	{0x1741DF97A078D35CULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_bdy_0s.dds"},
	{0x1742A467226BA452ULL, "vxw_sundermount_348_0_0.vlm"},
	{0x1742D9DBE59D9033ULL, "1055296016.ogg"},
	{0x1743A38A5A27183FULL, "vxlt_0_lowtown_a2_gx9_0_0.vlm"},
	{0x1743F928C2DEE2C8ULL, "qcr000_arishok6059303.cl"},
	{0x174426FF182E6515ULL, "llac_w_bonepitmining_a2_393222_0.dds"},
	{0x17446990D1CD9599ULL, "gen00fl_isabela6118382.cl"},
	{0x1744D46BFB670FE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_mark_activate.evt"},
	{0x1744E830C09C1EC8ULL, "vxlt_0_lowtown_e26_0_0.vlm"},
	{0x17452178CD02A092ULL, "zz_vault_debug6203301.cl"},
	{0x1745689E5DF1B4E4ULL, "899391423.ogg"},
	{0x1745A02D3E867030ULL, "vxlt_0_lowtown_a2_n_c_fhr_0_0.vlm"},
	{0x1745BA79217074B8ULL, "719732636.ogg"},
	{0x17460D88DE1A7E8CULL, "gol_smk.mao"},
	{0x17462395C33A91B5ULL, "hlac_lt_0_lowtown_a2_65538_1x.dds"},
	{0x174629E5C3E0E315ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_act_c.impa.gad"},
	{0x17464A5B932BC08AULL, "6061766_m.fxe"},
	{0x17465DC742F7A46EULL, "570370714.ogg"},
	{0x1746CD19A48EF61CULL, "vxlt_gallowstemplar_5wv_0_0.vlm"},
	{0x1747224C22A595ADULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcrate_0.pwk"},
	{0x17475F9270CD0DADULL, "fen210_hadriana6110369.cl"},
	{0x1747E44F1C975067ULL, "13149056.ogg"},
	{0x174890BF75D5FC1BULL, "vxlt_0_lowtown_a2_n_c_hof_0_0.vlm"},
	{0x17489A5E2332B4C4ULL, "light.mao"},
	{0x1748C73061885D54ULL, "tv_door_04.dds"},
	{0x17494DA21A20BA6CULL, "zz_vrc_debug6200259.cl"},
	{0x17497B0BF0A94102ULL, "vxw_sundermount_2ol_0_0.vlm"},
	{0x1749A41808F4F19AULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_stocks.phy"},
	{0x1749F320B12BA27EULL, "797073263.ogg"},
	{0x1749F7D7D7432584ULL, "llac_ht_bartrand_n_2995_6.dds"},
	{0x174A306A72068F0DULL, "6070508_f.fxe"},
	{0x174A44CCD8D65C67ULL, "docks_scene_dockworkers6176515.cl"},
	{0x174A501A5345FC58ULL, "gen_im_trs_dal_dua_dag_r1.uti"},
	{0x174A59DCA37786F3ULL, "art\\characters\\creatures\\spider\\animation\\dppr.ani"},
	{0x174A720537026F99ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml104a_0t.dds"},
	{0x174A80E8327D909EULL, "93215475.ogg"},
	{0x174A8318FBEA5D24ULL, "hlac_lt_0_lowtown_docks_n_-262144_2y.dds"},
	{0x174ACD9C864BD1A0ULL, "45368466.ogg"},
	{0x174AD3005A53DF12ULL, "dae000_worthy6166599.cl"},
	{0x174AFFB5BAA6445FULL, "vxlt_0_lowtown_a2_hhn_0_0.vlm"},
	{0x174B9036FFB34430ULL, "6190983_m.fxe"},
	{0x174C279B6F3E7FCFULL, "vxca_caverns1_e8g_0_0.vlm"},
	{0x174C2D84BE82AD12ULL, "262890001.ogg"},
	{0x174C330639CD64D5ULL, "hlac_w_encounter2_camp_393214_0y.dds"},
	{0x174C398EFF46ED64ULL, "vxht_mansion_n_5c2_0_0.vlm"},
	{0x174C422D84F9137FULL, "character_rendertodepth_p5.psh.fxo.dx11"},
	{0x174C56985E0B0267ULL, "6039951_m.fxe"},
	{0x174C9312A3AEF999ULL, "6128798_m.fxe"},
	{0x174D1B1A6027770EULL, "vxht_mansion_n_540_0_1.vlm"},
	{0x174D335298CABE08ULL, "vxca_caverns2prc_9ss_0_0.vlm"},
	{0x174D5FAF98BDD590ULL, "6057093_m.fxe"},
	{0x174DDF4EA9C25F29ULL, "vxw_encounter2_camp_pa_0_0.vlm"},
	{0x174E0B89F6408152ULL, "zz_followers_banter6158223.cl"},
	{0x174E185E67051C4AULL, "vxw_encounter2_g3_0_0.vlm"},
	{0x174E2A6FB0B974C0ULL, "mer320_hubert6053132.cl"},
	{0x174EFFD58097BBD5ULL, "one300ar_crow_camp.are"},
	{0x174F6A2A909DD148ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_onyxstatue.mmh"},
	{0x174F762A909DE5EBULL, "art\\levels\\races\\primeval\\props\\relics\\prp_onyxstatue.mao"},
	{0x174F847D5EEE0737ULL, "vxlt_0_lowtown_a2_n_cmu_0_0.vlm"},
	{0x174F882A909E0472ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_onyxstatue.msh"},
	{0x174FDD37A162DEA4ULL, "gen00fl_bethany6101361.cl"},
	{0x174FE982D44605C6ULL, "46669665.ogg"},
	{0x175048E1A8A77D08ULL, "mer320_hubert6065779.cl"},
	{0x17508403D362D0BFULL, "701221235.ogg"},
	{0x175116441461BCD2ULL, "vxw_sundermount_339_0_0.vlm"},
	{0x17513E0FB369C9B1ULL, "6127272_m.fxe"},
	{0x1751D9931326E636ULL, "93946948.ogg"},
	{0x1751F1358841F792ULL, "llac_lt_0_lowtown_docks_a3_n_-65538_8.dds"},
	{0x1752A084C4E5AA40ULL, "933700749.ogg"},
	{0x1752C7143133CACCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_rightrailing.msh"},
	{0x1752CD143133D4BEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_rightrailing.mmh"},
	{0x1752D9143133E955ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_rightrailing.mao"},
	{0x1752EC6BFB7375DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_mark_activate.ani"},
	{0x175331D075DFF01EULL, "scaleform_sf_solid_p0.psh.fxo.dx11.assembly"},
	{0x17533CE4EB5B3590ULL, "24887311.ogg"},
	{0x17533F350319DF3EULL, "6054014_m.fxe"},
	{0x175380F13215306DULL, "hlac_ht_0_hightown_196607_0y.dds"},
	{0x17541405478F88EFULL, "694246514.ogg"},
	{0x1754638A9D41E33BULL, "187792740.ogg"},
	{0x1754745B04F1511BULL, "548476861.ogg"},
	{0x1755906A69FFFC79ULL, "vxw_woundedcoast_1ej_0_0.vlm"},
	{0x17559B33F80A5EA4ULL, "6121561_m.fxe"},
	{0x1755D39CBBE4478BULL, "m_fen201ar_slavers_dungeon.wbd"},
	{0x17562C757F51F814ULL, "vxca_caverns1_mines_ik8_0_0.vlm"},
	{0x175676EF81904EB8ULL, "vxht_mansion2b_3cg_0_0.vlm"},
	{0x1756B231A918F477ULL, "6068579_m.fxe"},
	{0x1756B39E6649C7F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_1.mao"},
	{0x175747DB0384B71EULL, "383971218.ogg"},
	{0x175751DCA382032AULL, "art\\characters\\creatures\\spider\\animation\\dppr.evt"},
	{0x1757561715AEE662ULL, "atl_warninglig_dxt5_dat.xml"},
	{0x1757B611C9DFB781ULL, "da2terrain_rendertodepth_p0.vsh.fxo.dx11"},
	{0x1757ECE8E29F5982ULL, "vxw_sundermount_2pm_0_0.vlm"},
	{0x1757F221E1AB6FD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_up.ani"},
	{0x1757FB24154EDFC9ULL, "llac_ca_caverns1_mines_24426_7.dds"},
	{0x175811161CFB0B86ULL, "hlac_w_encounter3_-4_0x.dds"},
	{0x17589CF74E4F76E1ULL, "36911364.ogg"},
	{0x1758E3685334C4D7ULL, "6036486_m.fxe"},
	{0x175920D580A0483BULL, "one300ar_crow_camp.ncs"},
	{0x175927AD062D7304ULL, "character_rendertodepth_p6.psh.fxo.dx11.assembly"},
	{0x175932423FE52E25ULL, "gen00fl_aveline6139328.cl"},
	{0x17596E070EEF8AA3ULL, "6180519_f.fxe"},
	{0x175A16D58941E6E9ULL, "vxlt_0_lowtown_gcn_0_0.vlm"},
	{0x175A55DF35CDB55CULL, "6040581_m.fxe"},
	{0x175A5B31CBE0CB90ULL, "888307004.ogg"},
	{0x175A7FD94FC4DE13ULL, "one3510cr_dead_fighter_4.utc"},
	{0x175AD98496E86464ULL, "vxlt_hangedman_24u_0_0.vlm"},
	{0x175AFF1D94B812B1ULL, "c_dog_tatc.dds"},
	{0x175B0729E649578EULL, "one200pt_herbalist.ncs"},
	{0x175B6578B5565C60ULL, "plt_cod_cha_sebastian1.plo"},
	{0x175B86D8F7DC2454ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\tre_bark_d.dds"},
	{0x175B9A87A4324AA3ULL, "633203408.ogg"},
	{0x175CA483AC085CE6ULL, "2869755.ogg"},
	{0x175CB3943842872AULL, "6116684_m.fxe"},
	{0x175D1E0CC23CF475ULL, "zz_approval_debug6133727.cl"},
	{0x175D212AB1914ADCULL, "ser270_templar6149984.cl"},
	{0x175D368E919599A4ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.impa.ani"},
	{0x175D99EAEBE2CBB4ULL, "zz_and_debug6134046.cl"},
	{0x175DA5FA89BBDEFCULL, "hm_icarver_h3s2.mor"},
	{0x175DE0F4E8E0C45DULL, "dr_0_deeproads_20354_lc.mmh"},
	{0x175E30DCA387E99DULL, "art\\characters\\creatures\\spider\\animation\\dppr.gad"},
	{0x175E4810E2D20D70ULL, "vxdr_primevaltomb_7bl_0_1.vlm"},
	{0x175E6953D98885FFULL, "6156227_m.fxe"},
	{0x175EE6B21FB09D5CULL, "lt_0_lowtown_docks_a3_110_0.mao"},
	{0x175EF2B21FB0B1FFULL, "lt_0_lowtown_docks_a3_110_0.mmh"},
	{0x175EF8B21FB0BBA1ULL, "lt_0_lowtown_docks_a3_110_0.msh"},
	{0x175F4AA394CE546DULL, "966277581.ogg"},
	{0x175F9C35B80C360CULL, "llac_lt_0_lowtown_docks_a3_n_-131076_1.dds"},
	{0x176000D580A630E2ULL, "one300ar_crow_camp.lst"},
	{0x1760970D5E37C562ULL, "vxlt_0_lowtown_docks_b8h_0_0.vlm"},
	{0x1760C88F0B645FF7ULL, "vxlt_0_lowtown_docks_n_b4k_0_1.vlm"},
	{0x1760CE30619CFA82ULL, "tv_door_04.mao"},
	{0x1761441C9F6217A6ULL, "800350662.ogg"},
	{0x17616D54E12BB047ULL, "6042841_m.fxe"},
	{0x1761DB4441187504ULL, "305007757.ogg"},
	{0x1762389E8647421FULL, "6096053_m.fxe"},
	{0x176322C118C6A9D2ULL, "vxw_blightlands_10p_0_0.vlm"},
	{0x1763267F374782DEULL, "6005778_f.fxe"},
	{0x1763338E919A0006ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.impa.gad"},
	{0x17634C27BFF2744EULL, "563639057.ogg"},
	{0x1763CD4C7773B6B3ULL, "6077070_m.fxe"},
	{0x1763D0EB5AA779D4ULL, "follower_banter6169160.cl"},
	{0x17643C5C5A8662D3ULL, "6108695_m.fxe"},
	{0x17645073DAE97FB4ULL, "vxlt_0_lowtown_docks_a3_d3n_0_0.vlm"},
	{0x17645ACAC8C5ED5DULL, "vxdr_0_deeproads_cpe_0_1.vlm"},
	{0x1764EA21E1B5EC0AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_up.evt"},
	{0x17650046885FCB22ULL, "fen000im_gft_sword.uti"},
	{0x1765026BEE305394ULL, "llac_lt_0_lowtown_a2_n_0_6.dds"},
	{0x17651D8CD170550FULL, "mag330_walter6045630.cl"},
	{0x17658305279EA10DULL, "772155361.ogg"},
	{0x1765A5BE6E46BC76ULL, "6053309_m.fxe"},
	{0x1765DCD5F5BC3E37ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02a.mmh"},
	{0x1765F2D5F5BC63A9ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02a.msh"},
	{0x176613E7FCD97422ULL, "llac_ca_caverns2prc_5802_20.dds"},
	{0x1766962527B208C8ULL, "45085884.ogg"},
	{0x1766AFB8EF458432ULL, "vxw_encounter3_fc_0_0.vlm"},
	{0x176729B74C444418ULL, "la_spla.mao"},
	{0x1767947F93116331ULL, "art\\levels\\races\\proxy\\props\\door\\door.mmh"},
	{0x1767A07F931177D2ULL, "art\\levels\\races\\proxy\\props\\door\\door.mao"},
	{0x1767AE7F93118FEFULL, "art\\levels\\races\\proxy\\props\\door\\door.msh"},
	{0x1768085C9E8F3AA3ULL, "830383932.ogg"},
	{0x17688619CCFB7150ULL, "character_character_p58.psh.fxo"},
	{0x1768B4DF0F28CB68ULL, "6086770_m.fxe"},
	{0x1768ED26C6111FEFULL, "6107222_f.fxe"},
	{0x17691D48E3331E64ULL, "vxht_mansion_n_5i0_0_0.vlm"},
	{0x17699D1919FF0090ULL, "vxht_chantry_n_35b_0_0.vlm"},
	{0x1769A1422732D6FDULL, "vxca_caverns1_mines_gh2_0_0.vlm"},
	{0x1769BF558B634F27ULL, "945593152.ogg"},
	{0x176A2E8E91A016E5ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.impa.evt"},
	{0x176A708FADCAA867ULL, "dae000ip_crowd_em2_talk13.utp"},
	{0x176A89EE3EB522A4ULL, "combat_st_r5_combo.gda"},
	{0x176AA94264141E72ULL, "llac_ht_mansion2b_818_66.dds"},
	{0x176AC3088DAE9029ULL, "vxw_encounter2_camp_fg_0_0.vlm"},
	{0x176ACB35FCB56DE8ULL, "282033990.ogg"},
	{0x176AD92790C6B604ULL, "ico_amulet_1.dds"},
	{0x176B390A3DB6855BULL, "803499073.ogg"},
	{0x176B4F9CD69945E5ULL, "114605657.ogg"},
	{0x176BC921E1BBD37DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_up.gad"},
	{0x176C3AC9A1A77312ULL, "6076403_m.fxe"},
	{0x176C68A7581E41D1ULL, "dae010_shopper_m1.utc"},
	{0x176C69C683FCABB3ULL, "vxw_sundermount_27a_0_0.vlm"},
	{0x176CD777F23E5ADEULL, "305402158.ogg"},
	{0x176D421C95C9647FULL, "art\\vfx\\environment\\vfx_dark_fog.mmh"},
	{0x176D7A7E97EC00A3ULL, "6120860_m.fxe"},
	{0x176D9551650C1389ULL, "vxca_caverns2_9ei_0_0.vlm"},
	{0x176D9D8C2BABB888ULL, "character_character_p1.psh.fxo.dx11"},
	{0x176DD1C2786625DFULL, "dae000_female_tranquil6165139.cl"},
	{0x176E107038417927ULL, "ms_wispy_darker.mao"},
	{0x176E6492367E5277ULL, "vxht_brothel_n_5e0_0_0.vlm"},
	{0x176E8327397F3378ULL, "vxlt_gallowscourtyard_n_63x_0_0.vlm"},
	{0x176E88F8D15AA394ULL, "mag100_karras6034756.cl"},
	{0x176EAEBBA6CF7835ULL, "hlac_lt_gallowscourtyard_-65537_3y.dds"},
	{0x176EE7FC3CD9D550ULL, "899440947.ogg"},
	{0x176F52ABEAF85AE2ULL, "art\\levels\\4cinematics\\ferelden\\prp_arrow.msh"},
	{0x176F54ABEAF85D98ULL, "art\\levels\\4cinematics\\ferelden\\prp_arrow.mmh"},
	{0x176F60ABEAF8723BULL, "art\\levels\\4cinematics\\ferelden\\prp_arrow.mao"},
	{0x177000E3D24FEDE0ULL, "6039707_m.fxe"},
	{0x177077F5BC23A9A0ULL, "vxht_brothel_n_4de_0_0.vlm"},
	{0x1771496A816EAE13ULL, "vxw_bonepitmining_a2_1nh_0_0.vlm"},
	{0x17714FCD72724D54ULL, "vxlt_0_lowtown_a2_n_c_gon_0_0.vlm"},
	{0x177231A168E41A94ULL, "6032474_m.fxe"},
	{0x177238F41506AEEBULL, "6021231_m.fxe"},
	{0x177274244DEC2D79ULL, "vxw_woundedcoast_1kh_0_0.vlm"},
	{0x1772C250CCDF9834ULL, "vxlt_0_lowtown_docks_a3_d4o_0_0.vlm"},
	{0x1772CEDCC5E2D026ULL, "173412165.ogg"},
	{0x17733024F7D2B0DAULL, "framebuffereffects_motionblurcombine_p0.psh.fxo.dx11.assembly"},
	{0x1773631821280564ULL, "vxca_caverns1_mines_fm9_0_0.vlm"},
	{0x177390C5FB9B74D9ULL, "589104015.ogg"},
	{0x17739E2F39F6DFEDULL, "zz_followers_banter6159441.cl"},
	{0x1773C0B2715E0729ULL, "vxca_caverns2prc_8nx_0_0.vlm"},
	{0x1773D087511F86CDULL, "qun221_viscount6070119.cl"},
	{0x1773D82CE56F37C7ULL, "vxht_0_hightown_a3_bdy_0_0.vlm"},
	{0x1774A69E40BC0DA5ULL, "486592308.ogg"},
	{0x1774C219D757CAEAULL, "323384672.ogg"},
	{0x17752195E13B9CB2ULL, "vxw_encounter3_cl_0_0.vlm"},
	{0x17753C0D655E06C6ULL, "651075987.ogg"},
	{0x177542C27271BF3CULL, "6129856_f.fxe"},
	{0x1775E60068A2631AULL, "vxca_caverns1_ckm_0_0.vlm"},
	{0x1776AEB81B92C8E1ULL, "26305434.ogg"},
	{0x1776FC82A7E5D5B2ULL, "vxht_keep_6bw_0_0.vlm"},
	{0x177712117F8C58ADULL, "vxlt_gallowstemplar_f_3r5_0_0.vlm"},
	{0x17771FE4F664A496ULL, "em_vintner.mor"},
	{0x17773865A5799E64ULL, "6077455_m.fxe"},
	{0x1777C75EFD97FF95ULL, "6153226_m.fxe"},
	{0x1778B1E45C432D35ULL, "6103691_f.fxe"},
	{0x17791B6A099621F5ULL, "llac_lt_0_lowtown_a2_-131072_5.dds"},
	{0x17792C518A4E6672ULL, "hlac_lt_warehouse_n_0_8z.dds"},
	{0x177934E57FA4A8A9ULL, "vxw_encounter2_camp_af_0_0.vlm"},
	{0x17797F47E1D1B48FULL, "56769923.ogg"},
	{0x177A38D0E0EDD053ULL, "tbe100_xenon.bnk"},
	{0x177AB59CBC033159ULL, "m_fen201ar_slavers_dungeon.bnk"},
	{0x177B52E9EB4A843FULL, "lt_0_lowtown_a3_n_c_21_0.rml"},
	{0x177B7767C9A7D01BULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedidle.evt"},
	{0x177B78266614C582ULL, "6078183_f.fxe"},
	{0x177B8DB2ADD122DEULL, "dae300_alistair_king6133525.cl"},
	{0x177B9FFF6704A181ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_1y.dds"},
	{0x177BC2C66A7D1848ULL, "vxlt_0_lowtown_haz_0_0.vlm"},
	{0x177BC61B12D29049ULL, "6104939_f.fxe"},
	{0x177C24BDBCFED856ULL, "llac_ht_0_hightown_a3_65534_1.dds"},
	{0x177C6A4E4816148CULL, "739043482.ogg"},
	{0x177CB34204A0435AULL, "468176972.ogg"},
	{0x177CD5CC12F8FB9FULL, "335539317.ogg"},
	{0x177CD66F28746AF7ULL, "vxht_brothel_n_5h3_0_0.vlm"},
	{0x177CFA8718D70379ULL, "mer110_jansen6032613.cl"},
	{0x177D5AD0E0F03FA3ULL, "tbe100_xenon.cnv"},
	{0x177DC321439FE72DULL, "hlac_lt_0_lowtown_a2_-262144_4z.dds"},
	{0x177DE4FB11BB72A3ULL, "48915228.ogg"},
	{0x177E73CAB0D8079EULL, "zz_fampt_debug.ncs"},
	{0x177E79FB19F48A53ULL, "6080591_m.fxe"},
	{0x177EA8AF7190E4B4ULL, "vxca_caverns1_j1y_0_1.vlm"},
	{0x177F09E8009122E6ULL, "6179052_m.fxe"},
	{0x178118E1292BFC38ULL, "6135274_m.fxe"},
	{0x1781418F64185723ULL, "vxht_mansion_n_4fs_0_0.vlm"},
	{0x17815D755D49FADEULL, "228151641.ogg"},
	{0x178160ACCEABEBC7ULL, "conversation_dof_values.gda"},
	{0x17820FDFB880A442ULL, "vxlt_hangedman_1vs_0_1.vlm"},
	{0x17821E74CB582A9AULL, "maa.cut"},
	{0x17827067C9ADE35CULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedidle.gad"},
	{0x17827634E5BDD2A0ULL, "dae000_flemeth6000815.cl"},
	{0x17828B513B50B821ULL, "6139535_f.fxe"},
	{0x1782A16DCAF5D58AULL, "llac_lt_0_lowtown_a3_n_c_65535_3.dds"},
	{0x178311657711CBA6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_back_angle_up.ani"},
	{0x178409077ABAFDA1ULL, "589037087.ogg"},
	{0x178431927BD3BFCFULL, "174736106.ogg"},
	{0x17845A57ADBF6141ULL, "176897366.ogg"},
	{0x1784644731A82FC3ULL, "6015773_m.fxe"},
	{0x178484880983069AULL, "6064835_m.fxe"},
	{0x17850866DAC8B0A7ULL, "hlac_ht_keep_n_3732_8x.dds"},
	{0x17855CB960975B25ULL, "hlac_lt_backalley_n_65536_6y.dds"},
	{0x1785613641361A45ULL, "6085276_f.fxe"},
	{0x1785A04960E55E72ULL, "210822915.ogg"},
	{0x17860219640FB699ULL, "vxht_mansion_n_6la_0_0.vlm"},
	{0x17864400548F2458ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_basic_2.mao"},
	{0x17864A343B2804FFULL, "gen_im_trs_chn_amu_r0.uti"},
	{0x1786624049AC423AULL, "211138741.ogg"},
	{0x178672084F35E45DULL, "6121897_f.fxe"},
	{0x1786B0E75109F336ULL, "245433119.ogg"},
	{0x1786DDCF8B11E1D9ULL, "316512160.ogg"},
	{0x1786FF28FAD13B52ULL, "625341430.ogg"},
	{0x1787280B32FDB005ULL, "671827673.ogg"},
	{0x178747764ED3F729ULL, "ser221_quentin6128909.cl"},
	{0x178758CF71DBD90BULL, "919966780.ogg"},
	{0x178865AD68FF6262ULL, "6075972_m.fxe"},
	{0x178867FCAE795EF4ULL, "vxca_caverns2_7a7_0_1.vlm"},
	{0x17886F67C9B24D5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedidle.ani"},
	{0x178948DD7F3CF908ULL, "855759265.ogg"},
	{0x178A35F9C44CF5B4ULL, "104947223.ogg"},
	{0x178A36C363D84AC8ULL, "hlac_lt_0_lowtown_docks_-65538_1y.dds"},
	{0x178A3CFF86EE461AULL, "892689921.ogg"},
	{0x178AE3F49B2FB2B6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medh_0s.dds"},
	{0x178B37B0150E688CULL, "qcr000_arishok6059970.cl"},
	{0x178B99CFA4E67D95ULL, "45087061.ogg"},
	{0x178C6B098DDE30AFULL, "vxw_sundermount_2w5_0_0.vlm"},
	{0x178CA8C3BF78E3F4ULL, "zz_suhwan.cut"},
	{0x178D7AE11B2523FFULL, "vxw_blightlands_eo_0_0.vlm"},
	{0x178DFA8F6DC8EA4FULL, "zz_approval_debug6003754.cl"},
	{0x178E1B66A01D3CE7ULL, "hlac_lt_0_lowtown_docks_n_131070_4z.dds"},
	{0x178EE4D421E60836ULL, "560585049.ogg"},
	{0x178EE585989EB627ULL, "lowtown_fem_lowlife_mob6168984.cl"},
	{0x178F4E872432A3CFULL, "isa311_isabela6175514.cl"},
	{0x178F57DE31D85E79ULL, "vxw_woundedcoast_1qn_0_0.vlm"},
	{0x178F940FDB9C0192ULL, "6126478_m.fxe"},
	{0x178FA76E4E424BACULL, "vxlt_warehouse_ql_0_0.vlm"},
	{0x1790288996794F69ULL, "771496148.ogg"},
	{0x17905F8DE39ADEDAULL, "vxlt_undercityencounter_10w_0_0.vlm"},
	{0x179077699A8EDA0BULL, "vxw_bonepit_1hx_0_0.vlm"},
	{0x1790DA1A322906C2ULL, "vxht_chantry_n_36u_0_0.vlm"},
	{0x17916202F988F6D5ULL, "273410170.ogg"},
	{0x179170E0F7391871ULL, "6070415_m.fxe"},
	{0x1791814AD6573385ULL, "6171860_m.fxe"},
	{0x1791C0E650BFBBA3ULL, "ave200_donnic6112142.cl"},
	{0x1791DE5D43124BA7ULL, "gen00fl_sebastian6179333.cl"},
	{0x1792897F3F616C25ULL, "zz_dae_debug6143031.cl"},
	{0x1792963ADDC6BE1EULL, "llac_ca_caverns1_24422_1.dds"},
	{0x1792A02B9C3197C1ULL, "follower_banter6149217.cl"},
	{0x1792BD8DADCFC5B7ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isaa_0d.dds"},
	{0x1792D3B8CE0B2360ULL, "126852598.ogg"},
	{0x1793A414316B2659ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_rightrailing.phy"},
	{0x1793DF1987835DF1ULL, "vxdr_primevaltomb_39m_0_0.vlm"},
	{0x1793E03C8BD206B2ULL, "vxht_keep_6hq_0_0.vlm"},
	{0x179401500B476093ULL, "llac_lt_undercityencounter_-1_29.dds"},
	{0x17946DB1F5923F68ULL, "hlac_ht_brothel_n_3370_2z.dds"},
	{0x17954F3D8E47401BULL, "448750623.ogg"},
	{0x17956CA352993888ULL, "vxca_caverns2prc_9jl_0_0.vlm"},
	{0x17956EE2D501350EULL, "qun221_viscount6074042.cl"},
	{0x179619B80B50B91BULL, "vxlt_hangedman_n_21j_0_0.vlm"},
	{0x1796431C95EC6ABCULL, "art\\vfx\\environment\\vfx_dark_fog.anb"},
	{0x179661AE5AD4C570ULL, "hlac_lt_undercityencounter_-1_22z.dds"},
	{0x1796FA6577224DA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_back_angle_up.gad"},
	{0x1797088B3D2ED7B3ULL, "667317038.ogg"},
	{0x1797373A53F741C6ULL, "590421276.ogg"},
	{0x17986D5D0E541C43ULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_highway.msh"},
	{0x17987198ED3D2B4FULL, "6037311_m.fxe"},
	{0x17987B5D0E54339DULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_highway.mmh"},
	{0x17987F5D0E543AAEULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_highway.mao"},
	{0x1798CCB7C9205CD8ULL, "vxlt_0_lowtown_bxk_0_0.vlm"},
	{0x179976ED32736A57ULL, "vxlt_warehouse_n_qb_0_0.vlm"},
	{0x179A863EA1D2BEB0ULL, "6048844_m.fxe"},
	{0x179AAE37BACEB9ABULL, "6079670_m.fxe"},
	{0x179B0A5A3131A9E2ULL, "6169822_m.fxe"},
	{0x179B2064DD1C2B5EULL, "vxht_mansion2b_2ic_0_1.vlm"},
	{0x179B50BE29A83667ULL, "70926645.ogg"},
	{0x179B68E30D12180EULL, "gen00fl_merrill6130006.cl"},
	{0x179B8F0596A81292ULL, "and121cr_karl.utc"},
	{0x179BBC16AD80F75EULL, "200856098.ogg"},
	{0x179BECBE0D1B3C7FULL, "vxw_blightlands_nn_0_0.vlm"},
	{0x179BEDA1E47D53E6ULL, "6177252_m.fxe"},
	{0x179C9F869F589E23ULL, "6131093_m.fxe"},
	{0x179CF56F133F7FDBULL, "qun131_taskmaster6039691.cl"},
	{0x179D15EEF07E0FF6ULL, "vxlt_0_lowtown_docks_a3_cv0_0_0.vlm"},
	{0x179D74D8B744C174ULL, "dae011_corff_sta6142519.cl"},
	{0x179D7A1F80547FEEULL, "241469633.ogg"},
	{0x179D8D89A5836E8FULL, "vxw_sundermount_2f6_0_0.vlm"},
	{0x179DD763AB7B7083ULL, "hightown_fem_elfservant_mob.bnk"},
	{0x179E194B4038642CULL, "vxlt_warehouse_zm_0_0.vlm"},
	{0x179E4786DFCF2C13ULL, "6014002_m.fxe"},
	{0x179E4844D127AF37ULL, "310033085.ogg"},
	{0x179E619DCE3A279DULL, "6108727_m.fxe"},
	{0x179E8B289AE6E85EULL, "6034908_m.fxe"},
	{0x179E9D2786D998B3ULL, "52889300.ogg"},
	{0x179EC19E6687956FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_1.pwk"},
	{0x179EEA25FD3E1A15ULL, "6087116_m.fxe"},
	{0x179F1D061E8092AEULL, "649792227.ogg"},
	{0x179F38A410FD3B9CULL, "926690868.ogg"},
	{0x179F8150083032EBULL, "6145106_m.fxe"},
	{0x179F91F71980FD70ULL, "dae000_keep_bureaucrat_2_st6165392.cl"},
	{0x179F9C7CAB630D78ULL, "6164751_m.fxe"},
	{0x179FE2BDE92F30F1ULL, "6075715_m.fxe"},
	{0x179FFFEA6D63CCC4ULL, "zz_mag_debug6044329.cl"},
	{0x17A02B432DE05E0CULL, "vxlt_0_lowtown_fe0_0_0.vlm"},
	{0x17A051084D03CEDCULL, "vxlt_0_lowtown_a2_n_c_fim_0_0.vlm"},
	{0x17A06D673D4DBA5DULL, "vxlt_warehouse_n_oz_0_0.vlm"},
	{0x17A0B963AB7D73B3ULL, "hightown_fem_elfservant_mob.cnv"},
	{0x17A0F3F146BE9797ULL, "vxht_0_hightown_a2_bpt_0_0.vlm"},
	{0x17A0FA84F62BBE0CULL, "hlac_w_blightlands_131069_0y.dds"},
	{0x17A157433A20C16CULL, "6149084_m.fxe"},
	{0x17A1643170D66CB3ULL, "6069983_m.fxe"},
	{0x17A1E4098EA385B2ULL, "hlac_lt_0_lowtown_a2_n_c_65536_5z.dds"},
	{0x17A2BF67B9DF398CULL, "vxlt_gallowstemplar_n_3qq_0_0.vlm"},
	{0x17A327883CFD7CCBULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_108a_n.dds"},
	{0x17A33DB47D8C6CE1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_cara_d.dds"},
	{0x17A3957F6015F56CULL, "574122784.ogg"},
	{0x17A4052237D26660ULL, "lt_0_lowtown_a3_n_c_19_0_ll.mmh"},
	{0x17A4A722957CB66BULL, "738823561.ogg"},
	{0x17A55614CFCB5A43ULL, "llac_lt_undercitydungeon_-1_19.dds"},
	{0x17A59D18094344C3ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_stocks.msh"},
	{0x17A5AB1809435C1DULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_stocks.mmh"},
	{0x17A5AF180943632EULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_stocks.mao"},
	{0x17A5F29228F5CD6DULL, "sst_ss_bartrand.gda"},
	{0x17A69329D52DD3FDULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunaricontroller_n.dds"},
	{0x17A740B67E19BD59ULL, "742477622.ogg"},
	{0x17A7536F8088A909ULL, "vxht_0_hightown_alq_0_0.vlm"},
	{0x17A7780E97F9B440ULL, "987268749.ogg"},
	{0x17A842601FACA9A3ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgtb.mmh"},
	{0x17A850601FACC1FDULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgtb.msh"},
	{0x17A8FF32EC8EB957ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf03.phy"},
	{0x17A9334222221C9EULL, "vxw_sundermount_30x_0_0.vlm"},
	{0x17A9AFB128C8CF5BULL, "dae000_bartrand6027586.cl"},
	{0x17A9F4D9D8DBE397ULL, "isa311_isabela6091829.cl"},
	{0x17AA5E9AFF1154FFULL, "vxw_blightlands_si_0_0.vlm"},
	{0x17AA771166409B2EULL, "llac_blank_exterior_0_0z.dds"},
	{0x17AB10DE49471013ULL, "vxw_bonepitmining_a2_1zl_0_0.vlm"},
	{0x17AB3329599CF948ULL, "llac_w_encounter3_-131076_0.dds"},
	{0x17AB876AE1E5CA1EULL, "art\\characters\\creatures\\mubari\\animation\\walk_enter.ani"},
	{0x17ABC92B59BCA5B4ULL, "vxht_mansion_n_57o_0_1.vlm"},
	{0x17ABFF669779870FULL, "vxw_sundermount_2k1_0_0.vlm"},
	{0x17AC488721520D16ULL, "fx_cloudbrown.dds"},
	{0x17AC77C9BF883292ULL, "6128678_m.fxe"},
	{0x17ACC50E4857C490ULL, "crack.dds"},
	{0x17ACDB2DA1D3D429ULL, "zz_pro_cutscenes6007952.cl"},
	{0x17AD35947759461DULL, "and111_lirene6169991.cl"},
	{0x17ADDD67B6B43892ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_up_to_down.gad"},
	{0x17AE074F22F5EE44ULL, "881099168.ogg"},
	{0x17AE0DDF60452C3EULL, "6141140_f.fxe"},
	{0x17AE1DBC8CDC4F37ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left_1.mao"},
	{0x17AE62739131BF3BULL, "vxht_0_hightown_n_c0z_0_0.vlm"},
	{0x17AE8141AC0AE54BULL, "541489978.ogg"},
	{0x17AEB850D59EAB71ULL, "435807233.ogg"},
	{0x17AEBDD0457B88FAULL, "118131542.ogg"},
	{0x17AEDF442F43D2DDULL, "vxlt_warehouse_n_ry_0_0.vlm"},
	{0x17AFC62DF29DBC77ULL, "6149264_f.fxe"},
	{0x17B0AD70416C6EE5ULL, "llac_lt_0_lowtown_docks_65536_0.dds"},
	{0x17B0B8913E6B68F4ULL, "vxlt_undercitydungeon_xu_0_0.vlm"},
	{0x17B0F1B76D11BBF4ULL, "6125819_m.fxe"},
	{0x17B161F2CED36DABULL, "542953564.ogg"},
	{0x17B16C3F582F51A7ULL, "455107689.ogg"},
	{0x17B18FF86E42319BULL, "vxht_bartrand_n_4ft_0_0.vlm"},
	{0x17B206F67888DB07ULL, "6039854_m.fxe"},
	{0x17B22AD760CA1F75ULL, "hlac_lt_warehouse_1_2y.dds"},
	{0x17B24A3F36BA71B6ULL, "66240781.ogg"},
	{0x17B2876AB1617871ULL, "6169245_m.fxe"},
	{0x17B2BDDF51DA83BAULL, "hlac_lt_undercityserialkiller_-1_52y.dds"},
	{0x17B41AF4DEDF3AC2ULL, "mag360_thrask6053255.cl"},
	{0x17B43C7C470C3B63ULL, "127036679.ogg"},
	{0x17B5B75D354A8F77ULL, "hlac_w_encounter2_-131070_0x.dds"},
	{0x17B5BCA6CE928F17ULL, "lt_0_lowtown_18_0.rml"},
	{0x17B5CA7C0DB2872EULL, "105875086.ogg"},
	{0x17B68643C042A1A5ULL, "vxca_cavern1_haunted_fml_0_0.vlm"},
	{0x17B6E6D5CE778BF2ULL, "6118237_m.fxe"},
	{0x17B7496DBB33ADC8ULL, "ser221_party_comments6164309.cl"},
	{0x17B87547B7D6DF15ULL, "6057320_m.fxe"},
	{0x17B8D077F1076D7FULL, "vxw_blightlands_th_0_0.vlm"},
	{0x17B8D22415E39A57ULL, "ht_0_hightown_n_37_0_ll.mmh"},
	{0x17B8E9134369431BULL, "hlac_ht_0_hightown_a2_131071_2y.dds"},
	{0x17B8F987BE63AF41ULL, "vxht_mansion_n_4x7_0_0.vlm"},
	{0x17B943601FBB1EC8ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgtb.phy"},
	{0x17B9571B6648BC7AULL, "w_encounter3_31_0.rml"},
	{0x17B982BB3B3D2893ULL, "vxw_bonepitmining_a2_1wo_0_0.vlm"},
	{0x17B99F6AE1F230EFULL, "art\\characters\\creatures\\mubari\\animation\\walk_enter.evt"},
	{0x17B9C55F1EF97E90ULL, "6092397_m.fxe"},
	{0x17B9F2B803492D7FULL, "vxw_sundermount_2gu_0_0.vlm"},
	{0x17BA64C09A15CED1ULL, "713649326.ogg"},
	{0x17BA7143896F9F8FULL, "vxw_sundermount_2l0_0_0.vlm"},
	{0x17BAF410AF8AECADULL, "ser270_templar6127406.cl"},
	{0x17BB085B7792F257ULL, "vxlt_darktown_4wt_0_0.vlm"},
	{0x17BB34E6C2BD2149ULL, "hlac_lt_0_lowtown_docks_196607_1y.dds"},
	{0x17BB4557B226589DULL, "6106527_m.fxe"},
	{0x17BB75E9EB80A3A3ULL, "lt_0_lowtown_a3_n_c_21_0.anb"},
	{0x17BC07C512489E76ULL, "vxca_caverns2_7jy_0_0.vlm"},
	{0x17BC432331244D2EULL, "vxdr_0_deeproads_cmt_0_0.vlm"},
	{0x17BCFDAF9CEAE454ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_0z.dds"},
	{0x17BD033D865AC00EULL, "231300733.ogg"},
	{0x17BD51212139EB5DULL, "vxlt_warehouse_n_yx_0_0.vlm"},
	{0x17BD82E54E2D49D5ULL, "fx_smoketrail01.mao"},
	{0x17BDBB337FE63F95ULL, "hlac_lt_0_lowtown_docks_a3_65534_4y.dds"},
	{0x17BE0AE04B7A2B76ULL, "fex_ruin_arch_lrg_a.mmh"},
	{0x17BE14E04B7A3C74ULL, "fex_ruin_arch_lrg_a.msh"},
	{0x17BE8F0D8D5B48E5ULL, "vxca_caverns1_gkj_0_0.vlm"},
	{0x17BEE80456315676ULL, "mag160_feynriel6020074.cl"},
	{0x17BF08B83E9BF5AAULL, "mcr370_guardsman6087666.cl"},
	{0x17BF2F2C8CDC429AULL, "vxw_encounter2_nh_0_0.vlm"},
	{0x17BFA06AE1F69D98ULL, "art\\characters\\creatures\\mubari\\animation\\walk_enter.gad"},
	{0x17C001D560384A1BULL, "vxht_bartrand_n_4kw_0_0.vlm"},
	{0x17C02585B7D267FDULL, "zz_dae_debug6050739.cl"},
	{0x17C18C815773BE5EULL, "6201616_m.fxe"},
	{0x17C1C2A0F50CF64BULL, "vxht_0_hightown_a2_bwi_0_0.vlm"},
	{0x17C1FED5826D425DULL, "vxlt_darktown_4up_0_0.vlm"},
	{0x17C2204E7C600D7CULL, "vxht_0_hightown_n_aju_0_0.vlm"},
	{0x17C22ACB852A5FC2ULL, "hlac_lt_0_lowtown_a2_n_65537_3y.dds"},
	{0x17C2C867B6C67138ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_up_to_down.ani"},
	{0x17C3F22DE73EAF9EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_da_transition.ani"},
	{0x17C44F872166AB44ULL, "fx_cloudbrown.mao"},
	{0x17C530A114586518ULL, "6160661_m.fxe"},
	{0x17C5821AE998A1C6ULL, "895006459.ogg"},
	{0x17C66E3DEEB8FD91ULL, "ef_ineria.mor"},
	{0x17C6A82952D68DDDULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kita_0s.dds"},
	{0x17C6F5E4EFAF45A1ULL, "isa311_castillon6092051.cl"},
	{0x17C7140D007B982EULL, "vxht_chantry_2ir_0_0.vlm"},
	{0x17C7171BBCF8320AULL, "vxlt_0_lowtown_docks_b57_0_0.vlm"},
	{0x17C78337D1D43AEAULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_crnf.gad"},
	{0x17C82E37CB10A909ULL, "6055425_m.fxe"},
	{0x17C86494F53F45FFULL, "vxw_sundermount_2ht_0_0.vlm"},
	{0x17C8E3207B65B80FULL, "vxw_sundermount_2q3_0_0.vlm"},
	{0x17C90F830EC735A2ULL, "648122007.ogg"},
	{0x17C96B1C481F9CB6ULL, "216443127.ogg"},
	{0x17C9A54A0DA60DDBULL, "hlac_dr_primevalentrance_1726_0y.dds"},
	{0x17C9C13337E417CBULL, "t3_wdg_bw2.tnt"},
	{0x17C9D07FE9D51861ULL, "42551130.ogg"},
	{0x17CA0413B5BEA850ULL, "vxlt_darktown_5sz_0_0.vlm"},
	{0x17CA0A7C26923960ULL, "hm_hayder.mor"},
	{0x17CA2AB3C1A433CAULL, "ave310_donnic6134723.cl"},
	{0x17CA9C8BA202A1F2ULL, "6208366_f.fxe"},
	{0x17CB2EB7FD197271ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamfast.anb"},
	{0x17CB60634E2D5C82ULL, "gen00fl_varric6118761.cl"},
	{0x17CB7F9B22860086ULL, "zz_followers_banter6157297.cl"},
	{0x17CBCF7E243D6D41ULL, "hlac_lt_0_lowtown_docks_a3_-65538_1y.dds"},
	{0x17CC1A8E8362FDB5ULL, "hlac_ht_keep_n_3731_1x.dds"},
	{0x17CC63B5E5F9F671ULL, "gen_im_trs_qun_hvy_boo_r2.uti"},
	{0x17CC6A06A7D0B95DULL, "413829710.ogg"},
	{0x17CD07B6706E4A60ULL, "nat380_sibling6207662.cl"},
	{0x17CD5D9EC733D5D5ULL, "vxlt_undercityserialkiller_1zn_0_0.vlm"},
	{0x17CDA1097ED25B1AULL, "vxw_encounter2_ii_0_0.vlm"},
	{0x17CDA52F3C4EC12DULL, "68314526.ogg"},
	{0x17CE32F33D8BED7BULL, "vxlt_0_lowtown_a2_fg5_0_0.vlm"},
	{0x17CE6637D1DA2781ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_crnf.evt"},
	{0x17CEB2002366B8EFULL, "zz_dae_debug6049968.cl"},
	{0x17CF53AD67B0560AULL, "6150170_m.fxe"},
	{0x17CFAFF396670E9EULL, "zz_ave_debug6207729.cl"},
	{0x17CFF78458FBC59DULL, "6154555_m.fxe"},
	{0x17D0347DE7030ECBULL, "vxht_0_hightown_a2_bxh_0_0.vlm"},
	{0x17D09E3456FE506AULL, "vxlt_0_lowtown_a2_eqd_0_1.vlm"},
	{0x17D16A4680030677ULL, "6119015_m.fxe"},
	{0x17D1781B03D3040FULL, "302258885.ogg"},
	{0x17D1A92267B10DFAULL, "lt_0_lowtown_docks_24_0_lc.mmh"},
	{0x17D20A2DE74B166FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_da_transition.evt"},
	{0x17D23929EE53B3F9ULL, "143850293.ogg"},
	{0x17D323A7738581EAULL, "vxlt_0_lowtown_a3_n_c_gll_0_0.vlm"},
	{0x17D4B286D4AD17A9ULL, "47233592.ogg"},
	{0x17D4C5A7671BB216ULL, "chardebug.psh.fxo"},
	{0x17D52125EAD7FA5BULL, "vxdr_primevaltomb_3vs_0_1.vlm"},
	{0x17D549E54E417AB3ULL, "fx_smoketrail01.dds"},
	{0x17D617169A15EE84ULL, "fen210_fenris6109402.cl"},
	{0x17D61E4BB0568F0CULL, "ach_seb_cloak.dds"},
	{0x17D6B90C07D23666ULL, "llac_lt_0_lowtown_a2_n_-131072_5.dds"},
	{0x17D6D671E7355E7FULL, "vxw_sundermount_2mw_0_0.vlm"},
	{0x17D754FD6D5BD08FULL, "vxw_sundermount_2z2_0_0.vlm"},
	{0x17D80B2DE74F8318ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_da_transition.gad"},
	{0x17D942CE1B4D0B34ULL, "art\\levels\\races\\design\\prp_largetreasur_0.pwk"},
	{0x17D94FB7FD25E7EAULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamfast.mmh"},
	{0x17D95ECBC15027C4ULL, "vxht_mansion_n_6oy_0_0.vlm"},
	{0x17D9E8AEB562D592ULL, "hlac_lt_undercitydungeon_-1_51y.dds"},
	{0x17DA11013ECDF43CULL, "370210188.ogg"},
	{0x17DA2044986D87C8ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medp.msh"},
	{0x17DA247060EE6694ULL, "vxlt_0_lowtown_docks_a3_n_cxg_0_0.vlm"},
	{0x17DA3244986DA659ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medp.mao"},
	{0x17DA3E44986DBAF2ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medp.mmh"},
	{0x17DA53F3916C1496ULL, "vxdr_primevaltomb_56d_0_1.vlm"},
	{0x17DAA4843AF01197ULL, "vxca_caverns1_k2r_0_1.vlm"},
	{0x17DACA65094A913AULL, "905178697.ogg"},
	{0x17DB69F25274E14BULL, "vxca_caverns2_8p9_0_0.vlm"},
	{0x17DB7CDCFC58CB78ULL, "6176507_m.fxe"},
	{0x17DC41C448DDE121ULL, "vxlt_darktown_2s4_0_0.vlm"},
	{0x17DC5E37D1E657D0ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180l_crnf.ani"},
	{0x17DC6F156329F121ULL, "vxlt_0_lowtown_docks_a3_aru_0_0.vlm"},
	{0x17DCD1B95D3DF314ULL, "zz_followers_banter6158445.cl"},
	{0x17DCF36E8921F83EULL, "204192732.ogg"},
	{0x17DCFDFC2083B42CULL, "vxca_caverns1_i26_0_0.vlm"},
	{0x17DE1288926E73F8ULL, "555691951.ogg"},
	{0x17DE8AED538244EFULL, "sg5_blur_dot.mao"},
	{0x17DEB7E7BB929CEDULL, "lgt320_shw_ors_marlein6156318.cl"},
	{0x17DED172CCD37B9FULL, "vxlt_0_lowtown_n_h8m_0_0.vlm"},
	{0x17DEDE5C04E645CAULL, "vxw_blightlands_92_0_0.vlm"},
	{0x17DF24F33D57B1BFULL, "sst_ss_raider_m_2.gda"},
	{0x17DF2D2E3CADBC88ULL, "vxht_keep_n_3f3_0_0.vlm"},
	{0x17DF54DB532CD262ULL, "642922346.ogg"},
	{0x17DF71FF481EEDB5ULL, "llac_lt_0_lowtown_a3_n_c_131074_0.dds"},
	{0x17DFEC55D80F0204ULL, "ser270_emeric6113589.cl"},
	{0x17E038F71C81F977ULL, "vxht_0_hightown_n_c0n_0_0.vlm"},
	{0x17E0BE743508A234ULL, "weapon_distortionmask_p0.vsh.fxo.dx11.assembly"},
	{0x17E0D35252CE1F1BULL, "hlac_w_sundermount_-196606_0y.dds"},
	{0x17E120BB1C3E8ABBULL, "vxw_woundedcoast_i9_0_0.vlm"},
	{0x17E19584657B9A6AULL, "vxlt_0_lowtown_a3_n_c_gcm_0_0.vlm"},
	{0x17E22B09FD6C1261ULL, "vxca_cave_1o4_0_0.vlm"},
	{0x17E2ED6B10EF343DULL, "966042352.ogg"},
	{0x17E34DC01055B4B8ULL, "6110913_f.fxe"},
	{0x17E354D7853FCD12ULL, "dre000_dougal6131186.cl"},
	{0x17E3B58CB7BA66C1ULL, "gen00fl_aveline6128528.cl"},
	{0x17E4E9CC5A441731ULL, "vxht_0_hightown_a2_c73_0_1.vlm"},
	{0x17E4F3E8D26C2D15ULL, "wardogtacticspreset.gda"},
	{0x17E500C2D389CF69ULL, "w_sundermount_218_0.phy"},
	{0x17E53174F50433FEULL, "693038120.ogg"},
	{0x17E579891501460BULL, "gen_im_acc_rng_act3_04.uti"},
	{0x17E57DB0139A6430ULL, "dae000_meredith6092550.cl"},
	{0x17E584AFD00A04E3ULL, "vxca_caverns1_cod_0_0.vlm"},
	{0x17E5C8F838D93061ULL, "vxlt_0_lowtown_h9w_0_0.vlm"},
	{0x17E65042DDDD9886ULL, "montage_subs_demohack.cnv"},
	{0x17E65DF24D753BD7ULL, "vxlt_darktown_4dq_0_0.vlm"},
	{0x17E6731B2AC0556DULL, "vxht_brothel_n_52s_0_0.vlm"},
	{0x17E71155D691C98EULL, "6083057_f.fxe"},
	{0x17E78FA6483168F8ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgta_0s.dds"},
	{0x17E792FF140139BAULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.cess.gad"},
	{0x17E821951CB5C27BULL, "hlac_dr_slavers_3215_6z.dds"},
	{0x17E82D3460247C37ULL, "hlac_lt_gallowstemplar_n_2391_7x.dds"},
	{0x17E8ADFD21F6C963ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_imp_p.impa.evt"},
	{0x17E8D872C41C331BULL, "art\\characters\\weapons\\spears\\textures\\w_spr_001a_0d.dds"},
	{0x17E930F504A81628ULL, "vxdr_0_deeproads_eux_0_0.vlm"},
	{0x17E95FCC779F83E2ULL, "vxht_keep_n_40t_0_0.vlm"},
	{0x17E9D556D69F7F3FULL, "247958192.ogg"},
	{0x17EA0424A2C73161ULL, "hlac_lt_darktown_-1_76x.dds"},
	{0x17EA06812EB53632ULL, "hlac_w_blightlands_262145_0y.dds"},
	{0x17EA938164D6F66DULL, "w_bonepitmining_a2_230_0.mmh"},
	{0x17EA978164D6FCBEULL, "w_bonepitmining_a2_230_0.mao"},
	{0x17EA9D45B813D789ULL, "vxdr_primevaltomb_77g_0_0.vlm"},
	{0x17EAA58164D714F3ULL, "w_bonepitmining_a2_230_0.msh"},
	{0x17EAE7464AC05E89ULL, "gen00fl_aveline6121927.cl"},
	{0x17EB3882A0B3333CULL, "vxlt_undercityencounter_uu_0_0.vlm"},
	{0x17EB552A5B19D9C9ULL, "991921174.ogg"},
	{0x17EB6A4CA71454AAULL, "lt_0_lowtown_51_0.rml"},
	{0x17EB76EA12147B37ULL, "6038574_m.fxe"},
	{0x17EBB3BFBEC0BE48ULL, "vxca_cave_1o1_0_0.vlm"},
	{0x17EBFB3BACCDEAAEULL, "6087775_m.fxe"},
	{0x17EC087E3624DD8BULL, "6174586_m.fxe"},
	{0x17ED31354BF61412ULL, "w_bonepit_17_0_lc.mmh"},
	{0x17ED497C462D4506ULL, "mcr310cr_templar_trooper.utc"},
	{0x17EDAB4471185F58ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_olda_0d.dds"},
	{0x17EDB434D59F32CFULL, "art\\vfx\\combat\\vfx_gib_hlk_p_1_1.mao"},
	{0x17EDFADC29E47BC3ULL, "6111271_m.fxe"},
	{0x17EE65F374CA79F5ULL, "docks_male_overseer_mob6176576.cl"},
	{0x17EE95FF14075DD1ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.cess.evt"},
	{0x17EEFB4E622B3307ULL, "533548292.ogg"},
	{0x17EF1D572896F55AULL, "180553844.ogg"},
	{0x17EF35BB099DC9EDULL, "6036299_m.fxe"},
	{0x17EF51ADFA2DDC48ULL, "vxlt_0_lowtown_h9r_0_0.vlm"},
	{0x17EF86FD21FCA5E4ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_imp_p.impa.gad"},
	{0x17F0485ECB4D6C4EULL, "llac_ht_chantry_2534_9.dds"},
	{0x17F0980596F06248ULL, "and121cr_karl.ncs"},
	{0x17F0CDBF56BD1B11ULL, "llac_dr_primevaltomb_658_4.dds"},
	{0x17F0E032ECCC3A4CULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf03.mmh"},
	{0x17F0E632ECCC443EULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf03.msh"},
	{0x17F17E7CA720A03EULL, "479685191.ogg"},
	{0x17F1AB9123105F73ULL, "mgn.cut"},
	{0x17F22FFF0F42D1CEULL, "gen00fl_aveline6120752.cl"},
	{0x17F23150A6EEF44AULL, "6068026_m.fxe"},
	{0x17F2861FBA171913ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_exit.ani"},
	{0x17F2BFCF36251CBFULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_up_c.mmh"},
	{0x17F2CC3C3380DA7BULL, "6168770_m.fxe"},
	{0x17F3625BC82EFBF6ULL, "zz_qunpt_debug.ncs"},
	{0x17F3AD98E72A4225ULL, "llac_dr_0_deeproads_5843_1.dds"},
	{0x17F3D658D676FC75ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantryb_n.dds"},
	{0x17F4215698B61A90ULL, "llac_ht_0_hightown_-131074_2.dds"},
	{0x17F42947932196CCULL, "dae011_drunk2_sta6144218.cl"},
	{0x17F44CCF13707E24ULL, "6075123_m.fxe"},
	{0x17F47E6338C3F803ULL, "vxca_caverns2prc_7u4_0_0.vlm"},
	{0x17F4EC5C60E8A2EDULL, "33749515.ogg"},
	{0x17F4F03D42C3EA62ULL, "hlac_lt_0_lowtown_a2_-131072_8x.dds"},
	{0x17F5151906682022ULL, "248739463.ogg"},
	{0x17F51D4E9D6C98BBULL, "196268087.ogg"},
	{0x17F551A805DA1B01ULL, "zz_followers_banter6157335.cl"},
	{0x17F5A542F8C3B427ULL, "vxdr_primevalentrance_27p_0_0.vlm"},
	{0x17F620C67D74B699ULL, "242732786.ogg"},
	{0x17F632DA7212653DULL, "vxw_sundermount_23s_0_0.vlm"},
	{0x17F69415DACB754CULL, "839967967.ogg"},
	{0x17F695FD2202DDC2ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_imp_p.impa.ani"},
	{0x17F6CA0F08BF17A5ULL, "hlac_ht_playershouse_n_855_26x.dds"},
	{0x17F75CDFD318D9CCULL, "6127989_m.fxe"},
	{0x17F7DEB818BF89BDULL, "6170575_m.fxe"},
	{0x17F7F143E5BC0C8AULL, "563230410.ogg"},
	{0x17F8FD2F1A317D71ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_cara_t.dds"},
	{0x17F903D379AC8CA7ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev5_b.gad"},
	{0x17F908AE1D4347ACULL, "vxw_encounter3_3r_0_0.vlm"},
	{0x17F90F22AA09F009ULL, "vxdr_primevaltomb_70f_0_0.vlm"},
	{0x17FA832AF262C42BULL, "nat320_delilah6065952.cl"},
	{0x17FA8BC007CA0CE2ULL, "rdr111dg_martin6054894.cl"},
	{0x17FA8D946180EC18ULL, "vxlt_gallowstemplar_4qi_0_0.vlm"},
	{0x17FB60D9C3EEFEFBULL, "528985949.ogg"},
	{0x17FBA54FC25B7E4CULL, "gen_im_trs_cot_med_glv_r1.uti"},
	{0x17FBADFF141210A0ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.cess.ani"},
	{0x17FBB7BA50A4D710ULL, "vxw_bonepit_1r9_0_0.vlm"},
	{0x17FBF6A4DF2F750CULL, "hlac_ca_caverns2_5802_50z.dds"},
	{0x17FBF9F49C59CADBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjutted.mmh"},
	{0x17FC05F49C59DF78ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjutted.mao"},
	{0x17FC17F49C59FDC5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjutted.msh"},
	{0x17FC4D058104E386ULL, "mag131cr_taraneh.utc"},
	{0x17FCDD7D420FD9A2ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravec_n.dds"},
	{0x17FD12F0610080B5ULL, "504196116.ogg"},
	{0x17FD2D5699DFAE01ULL, "6081680_m.fxe"},
	{0x17FD40F8267719A5ULL, "vxlt_0_lowtown_a2_n_c_e1s_0_0.vlm"},
	{0x17FDDAA6FEC2AC31ULL, "vxw_woundedcoast_1kp_0_0.vlm"},
	{0x17FE160968A83C1FULL, "gen00fl_anders6125593.cl"},
	{0x17FE3405BB9489ECULL, "llac_lt_0_lowtown_a2_n_c_65534_1.dds"},
	{0x17FF16B773BD4DE5ULL, "vxlt_0_lowtown_docks_a3_d0n_0_0.vlm"},
	{0x17FF5BA40F368CB7ULL, "gen00fl_isabela6088351.cl"},
	{0x17FFD75585BF2311ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_gguard_meredacti.gad"},
	{0x1800658C43659CA2ULL, "169857772.ogg"},
	{0x18007E1FBA23494AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_exit.evt"},
	{0x1800CF4AC1350AC9ULL, "hlac_lt_0_lowtown_65536_9x.dds"},
	{0x1800F2D869828D4BULL, "vxlt_gallowstemplar_5j2_0_0.vlm"},
	{0x1801371B66863B31ULL, "w_encounter3_31_0.mmh"},
	{0x1801431B66864FD2ULL, "w_encounter3_31_0.mao"},
	{0x1801511B668667EFULL, "w_encounter3_31_0.msh"},
	{0x1801653820CFC48DULL, "317805878.ogg"},
	{0x1801ACB4A86A8B99ULL, "vxca_caverns2prc_1z9_0_0.vlm"},
	{0x18025F1ABCC6D538ULL, "hlac_ca_cavern1_haunted_24424_20x.dds"},
	{0x1802A15AC7F5412DULL, "552760613.ogg"},
	{0x1802AD08D0B9E8C3ULL, "6121023_m.fxe"},
	{0x1802FC5E9F271622ULL, "zz_dae_debug6036576.cl"},
	{0x18030E6D6EB4A210ULL, "vxw_encounter2_camp_ox_0_1.vlm"},
	{0x180406512A3F9E0BULL, "vxw_bonepit_10p_0_0.vlm"},
	{0x18047C87FD39799EULL, "lgt360_mag_orlanna6155393.cl"},
	{0x180499514976848EULL, "vxca_cavern1_haunted_jw1_0_0.vlm"},
	{0x1804A4B764087DBDULL, "vxw_sundermount_24r_0_0.vlm"},
	{0x1804EB4147385069ULL, "6129340_m.fxe"},
	{0x180538A6EE639AF0ULL, "6107694_f.fxe"},
	{0x1805B870D32694FDULL, "vxlt_gallowstemplar_f_47e_0_0.vlm"},
	{0x1806548D2FE3D369ULL, "6174034_m.fxe"},
	{0x18067E60D76124B8ULL, "651525411.ogg"},
	{0x1806C735BF679873ULL, "llac_ca_caverns1_mines_24426_5.dds"},
	{0x1806D85585C543BFULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_gguard_meredacti.ani"},
	{0x18071AEF0CD922CCULL, "w_bonepitmining_a2_59_0.msh"},
	{0x180720EF0CD92CBEULL, "w_bonepitmining_a2_59_0.mmh"},
	{0x18072CEF0CD94155ULL, "w_bonepitmining_a2_59_0.mao"},
	{0x1807433C8F253106ULL, "886056490.ogg"},
	{0x18075D1FBA292FBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_exit.gad"},
	{0x18082FDFEE718A64ULL, "vxlt_hangedman_2xa_0_0.vlm"},
	{0x18086A07EE71F316ULL, "152338480.ogg"},
	{0x180871C8CEB48CD7ULL, "418740470.ogg"},
	{0x180881B635DC7F62ULL, "vxw_woundedcoast_ob_0_0.vlm"},
	{0x180A17FAEE24E0DFULL, "14125949.ogg"},
	{0x180A7B34FFB2F8D7ULL, "6125333_m.fxe"},
	{0x180B48DAF9270491ULL, "gen00fl_fenris6106680.cl"},
	{0x180B95A62348A450ULL, "256541696.ogg"},
	{0x180C08CD960052AEULL, "amb_scene_alleydrunks.wbd"},
	{0x180C0CCAFD99D58EULL, "gen00fl_varric6037017.cl"},
	{0x180C4B576572747EULL, "follower_banter6141700.cl"},
	{0x180CDF9EA00EFFD4ULL, "115609877.ogg"},
	{0x180CEB1B3B5DE3EAULL, "vxlt_0_lowtown_a3_n_c_gph_0_0.vlm"},
	{0x180D3CE37ECF3C6FULL, "vxlt_0_lowtown_a2_n_byt_0_0.vlm"},
	{0x180DBF005002C0C2ULL, "hlac_lt_0_lowtown_docks_-65538_8x.dds"},
	{0x180DE79A5BEF83DFULL, "vxht_keep_n_5zr_0_0.vlm"},
	{0x180DF0D379BEC8EDULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev5_b.ani"},
	{0x180E35CD997A1319ULL, "vxht_brothel_n_60n_0_0.vlm"},
	{0x180E8F1CECDC5ECAULL, "gen00fl_merrill6111164.cl"},
	{0x180E92D324DBCC2DULL, "w_sundermount_81_0_ll.mmh"},
	{0x180F8FC12D635FFCULL, "6156408_m.fxe"},
	{0x1810BA512890C117ULL, "zz_approval_debug6003567.cl"},
	{0x1810E59A31EB68FAULL, "llac_lt_0_lowtown_a2_n_131073_1.dds"},
	{0x1810E9C2D3AEE28EULL, "w_sundermount_218_0.mmh"},
	{0x1810F5C2D3AEF725ULL, "w_sundermount_218_0.mao"},
	{0x1810FC14CCA3C278ULL, "6017872_f.fxe"},
	{0x181103C2D3AF0F7CULL, "w_sundermount_218_0.msh"},
	{0x181116C05BE5CFEDULL, "vxht_0_hightown_a2_c6i_0_0.vlm"},
	{0x181120214BE8954BULL, "hlac_lt_0_lowtown_a2_n_c_0_1z.dds"},
	{0x1811B6975033AD86ULL, "95086310.ogg"},
	{0x1811B8EFA2B24E2CULL, "vxlt_gallowstemplar_6s4_0_1.vlm"},
	{0x1811E9503ECA1BFDULL, "qcr220_grey_wardens6060459.cl"},
	{0x1811EA221057B363ULL, "vxw_encounter2_dg_0_0.vlm"},
	{0x181213EC79A5A0E9ULL, "art\\vfx\\combat\\vfx_gib_ogre.gib.gad"},
	{0x1812782E1C35B68BULL, "vxw_bonepit_15s_0_0.vlm"},
	{0x1812BD6BC3D29A55ULL, "570478187.ogg"},
	{0x18137D79771AA312ULL, "8999308.ogg"},
	{0x181392C3A891EBD0ULL, "6082198_m.fxe"},
	{0x1813F6E6B9896EEAULL, "follower_banter6166677.cl"},
	{0x1813FE976EC3228FULL, "hlac_lt_0_lowtown_n_-196608_2z.dds"},
	{0x18143D8773DF17F3ULL, "gen00fl_merrill6141247.cl"},
	{0x18144099B7050ACCULL, "6129789_m.fxe"},
	{0x18145B709866C447ULL, "vxht_keep_5dg_0_0.vlm"},
	{0x18147A9C938D1B45ULL, "6135947_m.fxe"},
	{0x1814BCA4CC366A00ULL, "follower_banter6163885.cl"},
	{0x1814FF4AF0232016ULL, "6143707_m.fxe"},
	{0x18150EED5F088368ULL, "370019167.ogg"},
	{0x181539649933DA4EULL, "393867792.ogg"},
	{0x18158D3A8C95C7E9ULL, "6031827_m.fxe"},
	{0x1815B5EEEEE2994EULL, "llac_ht_brothel_n_3367_0.dds"},
	{0x181656B127924688ULL, "6108259_f.fxe"},
	{0x1816A1BCE067A2E4ULL, "vxlt_hangedman_2uf_0_0.vlm"},
	{0x1816C9797D29F190ULL, "6165903_m.fxe"},
	{0x1816DAC83A1C6DA2ULL, "6149569_m.fxe"},
	{0x1816DB10E0D773A4ULL, "hlac_ca_caverns1_24424_4y.dds"},
	{0x1816F39327D297E2ULL, "vxw_woundedcoast_dc_0_0.vlm"},
	{0x1816F553EC7DFBE0ULL, "and121_karl6093868.cl"},
	{0x18176C342B929C0BULL, "hlac_lt_0_lowtown_docks_n_-131074_2y.dds"},
	{0x1817D1C276202EB5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_glv_cas_0s.dds"},
	{0x1818268C7DE0A8C3ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxb_n.dds"},
	{0x1818713D28F23622ULL, "art\\characters\\creatures\\spider\\animation\\blendtree\\spider.anb"},
	{0x18189DDB71D29FA2ULL, "411867564.ogg"},
	{0x18189F98B5C0952AULL, "qun111_ketojan6040785.cl"},
	{0x1818B4F07571DE31ULL, "890998372.ogg"},
	{0x1818B6299FC9A01CULL, "dre000_varric6207989.cl"},
	{0x1818E4EC79AB6F37ULL, "art\\vfx\\combat\\vfx_gib_ogre.gib.ani"},
	{0x181AA392E8C3408AULL, "vxca_cavern1_haunted_j6j_0_0.vlm"},
	{0x181ABE60E2AEDD31ULL, "vxw_woundedcoast_1qv_0_0.vlm"},
	{0x181AD8632BD48422ULL, "hlac_ht_playershouse_n_855_1x.dds"},
	{0x181AE85B7FE46AAEULL, "llac_lt_0_lowtown_a3_n_c_131071_2.dds"},
	{0x181B0392C85D8B93ULL, "555796598.ogg"},
	{0x181B0C48516D1088ULL, "hlac_lt_gallowstemplar_n_2388_2z.dds"},
	{0x181B56132A71388BULL, "746750060.ogg"},
	{0x181B5CF82D53FC6AULL, "vxlt_0_lowtown_a3_n_c_gwi_0_0.vlm"},
	{0x181BC0CF364823FCULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_up_c.anb"},
	{0x181BE7B009A977BBULL, "mag110_cut_abomination6016220.cl"},
	{0x181C048C47CD2C83ULL, "hlac_w_sundermount_65530_0y.dds"},
	{0x181C25FFE4CFBA83ULL, "6103243_m.fxe"},
	{0x181C27891300B933ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopeningmage.ani"},
	{0x181C796F596B06FCULL, "mag121_eau_banter6107385.cl"},
	{0x181D4033EEB2B751ULL, "192250020.ogg"},
	{0x181D7147CF86DE93ULL, "vxdr_0_deeproads_bfe_0_0.vlm"},
	{0x181E682EA6F5B16BULL, "6126357_f.fxe"},
	{0x181E6C2AC7500155ULL, "hlac_lt_0_lowtown_docks_-196611_0z.dds"},
	{0x181EB52D3CFC326AULL, "vxdr_primevalentrance_30h_0_1.vlm"},
	{0x181EDA57015F3E50ULL, "vxca_caverns1_mines_jl2_0_0.vlm"},
	{0x181F48DE0BD8C2DEULL, "dae000_sandal6147756.cl"},
	{0x181FDA1133EF57E7ULL, "llac_lt_0_lowtown_docks_n_-65536_0.dds"},
	{0x18212DC3E1FEBDA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_right.ani"},
	{0x1821872B55DE861AULL, "vxht_brothel_n_6ez_0_0.vlm"},
	{0x1821BC2D84917375ULL, "vxlt_undercityencounter_13d_0_1.vlm"},
	{0x1822E58C6B51D71DULL, "hlac_w_bonepit_262148_0x.dds"},
	{0x1822E6BB36BB119DULL, "6049446_m.fxe"},
	{0x1822F8B2CAB99507ULL, "llac_lt_0_lowtown_docks_-196611_0.dds"},
	{0x1822FE97658616C9ULL, "gen00fl_sebastian6180821.cl"},
	{0x18232AA31907492DULL, "hlac_w_encounter2_camp_-131070_0x.dds"},
	{0x18239E4AA15B2E53ULL, "vxlt_0_lowtown_a2_n_c_glh_0_0.vlm"},
	{0x1823AFB525CACF8BULL, "vxlt_backalley_g_1f0_0_0.vlm"},
	{0x1823B6BDC91DC0A3ULL, "6102874_f.fxe"},
	{0x1823D535BF8EF6A4ULL, "6163273_m.fxe"},
	{0x1823F1EA1F89A7E2ULL, "vxw_sundermount_1yy_0_0.vlm"},
	{0x1823FAF6E411CD7EULL, "6125130_m.fxe"},
	{0x18244B514BDA751CULL, "vxlt_undercityserialkiller_ye_0_0.vlm"},
	{0x18249F9354DB229AULL, "vxw_blightlands_7l_0_0.vlm"},
	{0x1824B454C13C20C8ULL, "w_sundermount_70_0.msh"},
	{0x1824C12CB57E295AULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockjagged.msh"},
	{0x1824C654C13C3F59ULL, "w_sundermount_70_0.mao"},
	{0x1824CF2CB57E4093ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockjagged.mao"},
	{0x1824D254C13C53F2ULL, "w_sundermount_70_0.mmh"},
	{0x1824D32CB57E47A0ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockjagged.mmh"},
	{0x1824E16A4C32DCC4ULL, "25296484.ogg"},
	{0x1825657019C8B062ULL, "vxw_woundedcoast_al_0_0.vlm"},
	{0x18258FC7BC04C913ULL, "6096831_m.fxe"},
	{0x1825D6CAFF042CF3ULL, "6059294_m.fxe"},
	{0x182636A87D87F712ULL, "ser211ip_book.ncs"},
	{0x1826442C7400F6A1ULL, "153981040.ogg"},
	{0x18269AA1893919D5ULL, "gen00fl_fenris6112119.cl"},
	{0x1826D9971C28BFF6ULL, "6162548_m.fxe"},
	{0x182734D03EF2E29EULL, "6105437_m.fxe"},
	{0x1827B409DF11A1F9ULL, "428563241.ogg"},
	{0x182831FECD7BEA33ULL, "vxw_sundermount_2tl_0_0.vlm"},
	{0x182840AB3B1361FCULL, "6087658_m.fxe"},
	{0x182849FA08CACA28ULL, "6043895_f.fxe"},
	{0x1828EF136213566EULL, "lighttest.arl"},
	{0x1828EF5D439CC7E2ULL, "llac_lt_0_lowtown_n_131070_4.dds"},
	{0x1828FF77B48E51BFULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boiler.msh"},
	{0x18290577B48E5C61ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boiler.mmh"},
	{0x18291177B48E7002ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boiler.mao"},
	{0x1829F2C3C8B90159ULL, "vxht_keep_3hk_0_0.vlm"},
	{0x182AD067C9B7F01BULL, "696303277.ogg"},
	{0x182AD8B8B4E2E849ULL, "llac_w_encounter2_-2_0.dds"},
	{0x182B20105A65A8D9ULL, "cod_ltr_flora_diary_2.utp"},
	{0x182B378131B6FEF9ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_evilchya_n.dds"},
	{0x182C109F42D857CDULL, "vxlt_gallowstemplar_f_350_0_0.vlm"},
	{0x182CC7FD6B79523FULL, "549269996.ogg"},
	{0x182CCD989E158077ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown1.mmh"},
	{0x182CE3989E15A5E9ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown1.msh"},
	{0x182D02F81B4F6BA3ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_s_p.impa.evt"},
	{0x182DA41A24D3DA26ULL, "133747056.ogg"},
	{0x182DCFE0321B32FFULL, "art\\vfx\\creatures\\fxc_bm_voml_c.impa.gad"},
	{0x182DF8B708EF802CULL, "vxca_caverns1_mines_j2a_0_1.vlm"},
	{0x182DFA7A3FD200EDULL, "vxht_0_hightown_a2_c8g_0_0.vlm"},
	{0x182E23CA493591ADULL, "vxlt_0_lowtown_feu_0_0.vlm"},
	{0x182ECDDBF443E463ULL, "vxw_encounter2_ji_0_0.vlm"},
	{0x182F3989EE2584E7ULL, "6153691_m.fxe"},
	{0x182F45C3E20B239AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_right.evt"},
	{0x182F58EB3885AC8EULL, "6042120_f.fxe"},
	{0x182FBCE86866C604ULL, "6045612_m.fxe"},
	{0x18300C5F7716A316ULL, "674134269.ogg"},
	{0x1830E75B219AAFC3ULL, "351546128.ogg"},
	{0x1830FE891312CFDDULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopeningmage.gad"},
	{0x18310F49C662F88EULL, "qun120_arishok6049642.cl"},
	{0x1831323F31962F43ULL, "vxht_brothel_n_5fq_0_0.vlm"},
	{0x18316A45BE7690A8ULL, "vxdr_0_deeproads_eju_0_0.vlm"},
	{0x1831F0C61F2C8067ULL, "6166192_m.fxe"},
	{0x18321027935146D3ULL, "vxlt_0_lowtown_a2_n_c_gik_0_0.vlm"},
	{0x183253CC02DD22CBULL, "vxlt_0_lowtown_a2_hgo_0_0.vlm"},
	{0x18327369D3585822ULL, "6015839_m.fxe"},
	{0x1833117046D13B1AULL, "vxw_blightlands_8m_0_0.vlm"},
	{0x1833195F13460C55ULL, "hlac_ht_mansion_n_3372_0x.dds"},
	{0x1833A58DBA5FB751ULL, "llac_ht_0_hightown_a3_65535_3.dds"},
	{0x1833C4E0321F8BECULL, "art\\vfx\\creatures\\fxc_bm_voml_c.impa.evt"},
	{0x1833CB539BB2C87DULL, "6109232_m.fxe"},
	{0x1833DBF81B554824ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_s_p.impa.gad"},
	{0x183427C47092E036ULL, "dae000_gamlen6121511.cl"},
	{0x183431FC8B2C2ED0ULL, "195827893.ogg"},
	{0x1834D58F7D36CCDBULL, "lt_0_lowtown_a2_33_0_lc.mmh"},
	{0x18353594AA1F39BCULL, "372516668.ogg"},
	{0x18353D4498BA64EDULL, "art\\characters\\playercharacter\\humanfemale\\arm_medp.phy"},
	{0x18355005F81CF1F7ULL, "6142117_f.fxe"},
	{0x1835DBD1AA281EC3ULL, "darktown_fem_lowlife_mob.bnk"},
	{0x1836176BFF10A70CULL, "6127237_m.fxe"},
	{0x18363B8EA7301530ULL, "160725605.ogg"},
	{0x183644C3E211408DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_right.gad"},
	{0x18367838607E1EB8ULL, "vxlt_gallowstemplar_f_3rd_0_0.vlm"},
	{0x18367A84BB78361AULL, "118471664.ogg"},
	{0x1836A3DBBF7202B3ULL, "vxw_sundermount_2yo_0_0.vlm"},
	{0x1836BCF9215BC7ABULL, "lgt260_eau_tmp_roderick6143695.cl"},
	{0x1836C83CBE59FDD5ULL, "6001226_m.fxe"},
	{0x1837196995A24BC2ULL, "hlac_w_bonepit_327686_0x.dds"},
	{0x183783A255EB7DA7ULL, "m_qun231ar_varnell_hideout.bnk"},
	{0x18379A861EFBB1E1ULL, "mag170_hugh6206065.cl"},
	{0x1837FE2ED23FDB9DULL, "art\\vfx\\textures\\potion\\vfx_potion_d.dds"},
	{0x183806282E475332ULL, "vxht_0_hightown_n_b0n_0_0.vlm"},
	{0x18380E74E5930114ULL, "follower_barks6153579.cl"},
	{0x18384865E03F8BCEULL, "zz_int.arl"},
	{0x1838FECD96272FDCULL, "amb_scene_alleydrunks.bnk"},
	{0x183916105AD3D8FEULL, "6114435_m.fxe"},
	{0x18392F2251C86798ULL, "zz_area_signage_19.utp"},
	{0x1839BDD1AA2BD4F3ULL, "darktown_fem_lowlife_mob.cnv"},
	{0x183A083517FA98EAULL, "6110072_f.fxe"},
	{0x183AC0DDF3DD36CFULL, "6207787_m.fxe"},
	{0x183AEAF81B5B8002ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_s_p.impa.ani"},
	{0x183B14F3A7F172CAULL, "llac_lt_0_lowtown_0_5.dds"},
	{0x183B1A57DC22F1FFULL, "llac_ca_caverns1_24426_0.dds"},
	{0x183BB1F7A50F94E3ULL, "6085336_m.fxe"},
	{0x183C4692DE0F18FFULL, "gen00fl_aveline6053329.cl"},
	{0x183CB159954BC6D4ULL, "vxht_0_hightown_a3_4n2_0_0.vlm"},
	{0x183D11312AE346DBULL, "593277990.ogg"},
	{0x183D4A9EF0C0769DULL, "vxw_bonepitmining_a2_8g_0_0.vlm"},
	{0x183D4BD5D698B248ULL, "vxca_cavern1_haunted_ede_0_0.vlm"},
	{0x183DC512F62435D2ULL, "gen00fl_varric6107273.cl"},
	{0x183DDB8F19222F56ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkforward.ani"},
	{0x183DDFA8A08CFA3DULL, "6085529_m.fxe"},
	{0x183DE7B830E845A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_down.ani"},
	{0x183E015511A7CE57ULL, "6133012_f.fxe"},
	{0x183E25B6EB8CFE5EULL, "279467708.ogg"},
	{0x183E37EC1CFDE0DDULL, "6041944_f.fxe"},
	{0x183E4DAD8F66544DULL, "vxht_0_hightown_n_9zh_0_0.vlm"},
	{0x183FA03533E7BEF1ULL, "1002582263.ogg"},
	{0x183FA41C238C47C3ULL, "vxht_brothel_n_5ip_0_0.vlm"},
	{0x184073C09E25E777ULL, "rdr111dg_assistant6050919.cl"},
	{0x18415D2D18E85B61ULL, "vxht_chantry_n_2hq_0_0.vlm"},
	{0x18416AD5A36552CFULL, "vxlt_undercityserialkiller_17w_0_0.vlm"},
	{0x1841A5FFE7E28E7CULL, "1036366362.ogg"},
	{0x1841ACE0322BA045ULL, "art\\vfx\\creatures\\fxc_bm_voml_c.impa.ani"},
	{0x1841EA28E7767110ULL, "follower_banter6160699.cl"},
	{0x18423C9A5044488BULL, "vxca_caverns1_e7v_0_0.vlm"},
	{0x18423D308DA8E0FDULL, "6102332_m.fxe"},
	{0x1842BA84E2F05DF3ULL, "6053094_m.fxe"},
	{0x1842E3125669E2C0ULL, "500174254.ogg"},
	{0x18433B663D9510BEULL, "combat_smaoe_r1_a3.gda"},
	{0x18435AF18AF8C843ULL, "6041721_m.fxe"},
	{0x1843701362D9DF1BULL, "vxlt_hangedman_n_2uf_0_0.vlm"},
	{0x18439D2F2C894E16ULL, "403502336.ogg"},
	{0x184409B2A93FB7AFULL, "darktown_male_fereldan_sta6175153.cl"},
	{0x18441AF49C97B810ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjutted.phy"},
	{0x18445C6C3E466F91ULL, "38983097.ogg"},
	{0x184470493D07773DULL, "hlac_ht_mansion_n_2995_3y.dds"},
	{0x18447C6C03C6E736ULL, "zz_dre_debug6168232.cl"},
	{0x184492A6DD2477F3ULL, "vxht_bartrand_n_65g_0_0.vlm"},
	{0x1844F2B830EE7743ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_down.gad"},
	{0x1844F4EFCCB12FE2ULL, "one140cr_charade6137608.cl"},
	{0x18453A19B0501655ULL, "6004126_m.fxe"},
	{0x18455B16B941FBC8ULL, "383936495.ogg"},
	{0x184566EDDE3B278AULL, "6036731_m.fxe"},
	{0x1846521F04A9D51AULL, "6172295_m.fxe"},
	{0x18467199139F0937ULL, "277716052.ogg"},
	{0x1846E2386D5340BDULL, "306441739.ogg"},
	{0x184760610594026FULL, "535159586.ogg"},
	{0x18476508CDB99684ULL, "6151009_m.fxe"},
	{0x184787ED4CC9F17EULL, "6113535_m.fxe"},
	{0x18484DD2A08761E7ULL, "6147926_m.fxe"},
	{0x18485E58DA9E3A2CULL, "vxht_0_hightown_n_b8x_0_0.vlm"},
	{0x1849270C7302B64EULL, "hf_qmq_minion4.mor"},
	{0x184936047CDE8228ULL, "344375763.ogg"},
	{0x184963EB18A27EABULL, "vxlt_0_lowtown_a2_fge_0_0.vlm"},
	{0x18498B12B6FC44E6ULL, "art\\characters\\creatures\\dragon\\animation\\ctrn180r.evt"},
	{0x1849D4C11AB83381ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\cuddle_lying_male_loop.gad"},
	{0x184A02BA469BFDD2ULL, "vxlt_0_lowtown_a2_n_c_gbo_0_0.vlm"},
	{0x184A278EE5321592ULL, "17884963.ogg"},
	{0x184A29B4A9589742ULL, "6115775_m.fxe"},
	{0x184A7F7E7B13C66AULL, "llac_dr_dungeon_3207_6.dds"},
	{0x184AE166FEDF0C6CULL, "vxdr_0_deeproads_9gr_0_0.vlm"},
	{0x184AE38F192CC747ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkforward.evt"},
	{0x184AE5E2DA3D0748ULL, "vxht_chantry_n_2ht_0_0.vlm"},
	{0x184B07177A191820ULL, "6067375_m.fxe"},
	{0x184BA441DC704517ULL, "vxlt_0_lowtown_docks_cnj_0_0.vlm"},
	{0x184BB195D8301563ULL, "vxw_encounter2_pk_0_0.vlm"},
	{0x184BCFB830F45A28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_down.evt"},
	{0x184BF7ECE8482C5AULL, "268900311.ogg"},
	{0x184C29A87DA8B0F7ULL, "ser211ip_book.utp"},
	{0x184CCB959F8F62E0ULL, "dae200_saemus6142703.cl"},
	{0x184CDCC22BC0CF9AULL, "vxht_brothel_n_6tu_0_0.vlm"},
	{0x184D01E196F82AEDULL, "dae200_bran6083828.cl"},
	{0x184D370EEF16F2C0ULL, "131142581.ogg"},
	{0x184DA6FEA4702E6EULL, "259111234.ogg"},
	{0x184DB4FA795D699EULL, "vxlt_undercityencounter_1j5_0_0.vlm"},
	{0x184DEDBD0F9A01A6ULL, "vxht_0_hightown_9uu_0_0.vlm"},
	{0x184E18BFD7C0D2DFULL, "vxdr_primevalentrance_1a1_0_0.vlm"},
	{0x184E19CCEB7A3CE7ULL, "hlac_dr_0_deeproads_5841_3z.dds"},
	{0x184E2A8CF327BDF2ULL, "vxlt_foundry_n_zd_0_0.vlm"},
	{0x184E349D6E999E06ULL, "vxlt_0_lowtown_docks_n_b5d_0_0.vlm"},
	{0x184E5BEE14FD4CE8ULL, "lgt110_arl_bann_teagan.cnv"},
	{0x184E625809FFB1FAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_4.gad"},
	{0x184F054BFBAD190BULL, "vxlt_backalley_g_1u5_0_0.vlm"},
	{0x184F3785E6C953CBULL, "vxlt_0_lowtown_a2_hia_0_0.vlm"},
	{0x184FCF0A0ADE73E1ULL, "vxht_chantry_n_2cp_0_0.vlm"},
	{0x185053453D9F32F3ULL, "417423897.ogg"},
	{0x18505AFCE482D945ULL, "6109795_m.fxe"},
	{0x18505FF2F87181FDULL, "vxht_mansion2_3dw_0_0.vlm"},
	{0x185075B7D36F5A2CULL, "vxht_keep_n_6jj_0_0.vlm"},
	{0x1850A212B7028A31ULL, "art\\characters\\creatures\\dragon\\animation\\ctrn180r.gad"},
	{0x1850BB06EFAAF9E2ULL, "vxw_woundedcoast_po_0_0.vlm"},
	{0x1850D5C11ABE542FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\cuddle_lying_male_loop.ani"},
	{0x1851540051C9A18FULL, "hlac_lt_0_lowtown_docks_a3_n_131073_0z.dds"},
	{0x18517DEE14FFBC50ULL, "lgt110_arl_bann_teagan.bnk"},
	{0x1851FBA8C7F1A77EULL, "hlac_lt_0_lowtown_a3_n_c_65535_1x.dds"},
	{0x1852048F19331D90ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkforward.gad"},
	{0x18520C37AF101032ULL, "vxdr_primevalentrance_39n_0_0.vlm"},
	{0x18524B7F75937645ULL, "vxlt_hangedman_n_14p_0_0.vlm"},
	{0x185339489820338EULL, "6115836_m.fxe"},
	{0x1853BE6DF15D2948ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_dksa_0s.dds"},
	{0x1853D832C44CFF62ULL, "vxw_bonepit_1xe_0_0.vlm"},
	{0x185437C4A09660F2ULL, "dae000_gamlen6102527.cl"},
	{0x185465580A042211ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_4.evt"},
	{0x18550677B4B398DEULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boiler.phy"},
	{0x1855B46CB2AA81E7ULL, "hlac_lt_undercityencounter_-1_39y.dds"},
	{0x1855D68976DFD4B8ULL, "plt_seb120pt_duty_letter.plo"},
	{0x1855DEE95DBF777BULL, "vxw_encounter1_b0_0_0.vlm"},
	{0x1855E0587B22E9ACULL, "6092609_m.fxe"},
	{0x1856044F67D4DDE9ULL, "vxht_brothel_n_32i_0_0.vlm"},
	{0x1856482AB6CCBE38ULL, "792813240.ogg"},
	{0x1856509340ED6BC9ULL, "vxdr_0_deeproads_f4i_0_0.vlm"},
	{0x18571C75A8829E51ULL, "mag200_feynriel6034480.cl"},
	{0x18574AF68E8C4A69ULL, "cr_center.mao"},
	{0x1857A312B708AADFULL, "art\\characters\\creatures\\dragon\\animation\\ctrn180r.ani"},
	{0x1857AF2108BC0F56ULL, "vxht_brothel_n_4wz_0_0.vlm"},
	{0x18581DA1020A1B9FULL, "follower_banter6155724.cl"},
	{0x18583978824737CEULL, "hlac_w_bonepit_327681_0y.dds"},
	{0x185855689C7EA471ULL, "5087117.ogg"},
	{0x18586AB5ACCCB2F8ULL, "6095510_m.fxe"},
	{0x18589B919B4EAFC2ULL, "6110875_m.fxe"},
	{0x1858B5511F66FDBFULL, "vxdr_0_deeproads_9is_0_0.vlm"},
	{0x1859155117EAF1D6ULL, "art\\vfx\\models\\vfx_target_cone_small.msh"},
	{0x18591F5117EB02D4ULL, "art\\vfx\\models\\vfx_target_cone_small.mmh"},
	{0x1859235117EB09E7ULL, "art\\vfx\\models\\vfx_target_cone_small.mao"},
	{0x185957BFCC331FC8ULL, "vxht_chantry_n_2cu_0_0.vlm"},
	{0x185A00BBB0796ACEULL, "3358254.ogg"},
	{0x185A113EB7893F65ULL, "vxlt_undercityencounter_1dp_0_0.vlm"},
	{0x185A3139B3F138ABULL, "gen00fl_aveline6139313.cl"},
	{0x185A4BC20BFA3A43ULL, "hlac_ht_keep_n_3732_3y.dds"},
	{0x185AFA43ABB36B86ULL, "bet100pt_gift.ncs"},
	{0x185B5D0BF8801E41ULL, "6114414_m.fxe"},
	{0x185BC2B6ADE64B10ULL, "lt_0_lowtown_docks_a3_201_0.msh"},
	{0x185BD0B6ADE6632AULL, "lt_0_lowtown_docks_a3_201_0.mmh"},
	{0x185BD4B6ADE66A61ULL, "lt_0_lowtown_docks_a3_201_0.mao"},
	{0x185C1BE5DCD7D04BULL, "6115514_m.fxe"},
	{0x185C26D76B53821EULL, "vxlt_undercityencounter_1g6_0_0.vlm"},
	{0x185C5D73E793B5DCULL, "character_character_p1.vsh.fxo.dx11.assembly"},
	{0x185C5F4D0D6C879CULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_floortile_n.dds"},
	{0x185C73C9166216FAULL, "383575845.ogg"},
	{0x185C86C02F378FACULL, "vxlt_0_lowtown_a2_eoq_0_0.vlm"},
	{0x185C924C87F68F88ULL, "llac_ca_caverns1_24422_7.dds"},
	{0x185D3C1E554F253DULL, "atl_dialoguepo_dxt5_dat.xml"},
	{0x185D3E84A7471EC1ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_varr.phy"},
	{0x185D65BE69339053ULL, "vxlt_0_lowtown_a2_n_c_gxl_0_0.vlm"},
	{0x185D75312832042FULL, "mag121_idunna6084211.cl"},
	{0x185D7728EDA3318BULL, "vxlt_backalley_g_1r4_0_0.vlm"},
	{0x185D8D8B725DC2F7ULL, "llac_lt_warehouse_65537_0.dds"},
	{0x185DB986DD421EF6ULL, "41285800.ogg"},
	{0x185DF1AA4E1D5226ULL, "vxht_mansion_n_332_0_0.vlm"},
	{0x185EFD523C348822ULL, "gen_im_acc_amu_act2_06.uti"},
	{0x185F5D4050A87B1EULL, "zz_prmpt_debug6200943.cl"},
	{0x185F6944CB8BA9D1ULL, "6137577_f.fxe"},
	{0x185FF473913789F2ULL, "vxlt_gallowstemplar_f_5qm_0_0.vlm"},
	{0x18607E14A10628B2ULL, "vxdr_primevalentrance_32o_0_0.vlm"},
	{0x18622A9BE707B887ULL, "dae200cr_viscount6098517.cl"},
	{0x18623830991C1AC6ULL, "6125422_m.fxe"},
	{0x186264685B66F72CULL, "vxdr_primevalentrance_31x_0_0.vlm"},
	{0x18627D580A1088E0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_4.ani"},
	{0x186299948CCEE088ULL, "qun110_petrice6018538.cl"},
	{0x1862B2FC335DDDEBULL, "vxht_bartrand_n_74a_0_0.vlm"},
	{0x1862F36D0AC05FF5ULL, "6110890_m.fxe"},
	{0x186335EAF76842E2ULL, "dae000_bartrand6206858.cl"},
	{0x1863756BBD7159BCULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magicloker_sa.dds"},
	{0x186399D6ABE86284ULL, "follower_banter6167229.cl"},
	{0x1863AF94A979BC3FULL, "vxlt_0_lowtown_docks_a3_cz0_0_1.vlm"},
	{0x18643DD4518CAD6BULL, "gen00fl_aveline6112703.cl"},
	{0x186450C64FB58FFBULL, "vxw_encounter1_g3_0_0.vlm"},
	{0x18645AB5B0F97418ULL, "vxlt_0_lowtown_a2_gob_0_1.vlm"},
	{0x186479A256125ADDULL, "m_qun231ar_varnell_hideout.wbd"},
	{0x1864808AEDB82FCDULL, "289115133.ogg"},
	{0x1864E1437E139599ULL, "gen00fl_isabela6099535.cl"},
	{0x18652331D7366D59ULL, "120794323.ogg"},
	{0x18652D4849E71C40ULL, "hlac_w_bonepitmining_a2_0_12y.dds"},
	{0x18664583AB85C1B5ULL, "hlac_lt_undercityserialkiller_-1_38z.dds"},
	{0x1866BEAE5824BBC4ULL, "6142315_m.fxe"},
	{0x18673ECB7D14C701ULL, "839814307.ogg"},
	{0x186762F1126A1411ULL, "dae000_flemeth6000892.cl"},
	{0x18679DFF7C7BDD5AULL, "vxlt_0_lowtown_docks_a3_n_aps_0_0.vlm"},
	{0x1867FCDA49A5F21EULL, "gen00fl_anders6142544.cl"},
	{0x18686596CC57B978ULL, "hlac_ca_caverns1_mines_24426_7y.dds"},
	{0x18687CA0AEFFF0ACULL, "dae010_alienage_elf_m1.utc"},
	{0x186887FBC05C7617ULL, "vxlt_0_lowtown_docks_ctd_0_0.vlm"},
	{0x1868A68845D68672ULL, "dae000_fenarel6133346.cl"},
	{0x186944280653428FULL, "lt_0_lowtown_docks_248_0.mao"},
	{0x186950280653572CULL, "lt_0_lowtown_docks_248_0.mmh"},
	{0x186956280653611EULL, "lt_0_lowtown_docks_248_0.msh"},
	{0x1869A421CA71ED49ULL, "plt_cod_cha_anders1.plo"},
	{0x1869CEE8EA7636C1ULL, "6113514_m.fxe"},
	{0x186A010B7698BAA1ULL, "283437820.ogg"},
	{0x186ABBA66E097F8BULL, "vxw_bonepitmining_a2_1hz_0_0.vlm"},
	{0x186B4A5307C8414EULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masb.phy"},
	{0x186B8E59F57D59B8ULL, "579802450.ogg"},
	{0x186BC039F5EAF88DULL, "6041086_m.fxe"},
	{0x186C29A6A6897258ULL, "vxw_encounter2_nz_0_0.vlm"},
	{0x186C48D5E4B60A6BULL, "6142309_m.fxe"},
	{0x186C56264B5FC356ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_024a_0s.dds"},
	{0x186C7B3E8E07477CULL, "6161136_m.fxe"},
	{0x186C8B33636FF53CULL, "gen00fl_merrill6150976.cl"},
	{0x186CEEF77E5B99ADULL, "zz_pro_debug6059737.cl"},
	{0x186D832C457973B2ULL, "dae000cr_grand_cleric6156775.cl"},
	{0x186DA77B34FB1887ULL, "ht_mansion_n_9156.rml"},
	{0x186DB148BD281D99ULL, "mcr372ar_gallows_templar_nt.are"},
	{0x186DFFB854EDF484ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_maga_0t.dds"},
	{0x186E168622173FB5ULL, "6057319_m.fxe"},
	{0x186E34DBD91632BCULL, "1047997085.ogg"},
	{0x186E59F68E9F4217ULL, "cr_center.dds"},
	{0x186F01C4949CEC0AULL, "vxw_woundedcoast_b9_0_0.vlm"},
	{0x186F737D4C001B54ULL, "6069276_m.fxe"},
	{0x186F7BA0A1678D6BULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.dura.ani"},
	{0x186FF79B3AD94486ULL, "vxca_caverns2_963_0_0.vlm"},
	{0x18700ECB7547D124ULL, "6149628_m.fxe"},
	{0x187107A41B323B1EULL, "isa211_isabela6069435.cl"},
	{0x1871181DB6C457FFULL, "demoral_04.mao"},
	{0x18716E26A565ADE9ULL, "51254565.ogg"},
	{0x187225C0605BB74EULL, "art\\levels\\races\\design\\prp_prpbodybagb.mmh"},
	{0x18723FC0605BE33CULL, "art\\levels\\races\\design\\prp_prpbodybagb.msh"},
	{0x18726974CC85A8EAULL, "478863123.ogg"},
	{0x18729B68005AC10EULL, "vxlt_darktown_60z_0_0.vlm"},
	{0x1873A88D460B2B46ULL, "vxht_0_hightown_bxe_0_0.vlm"},
	{0x1873C071741E598EULL, "1044185793.ogg"},
	{0x1874230A3B9612FFULL, "6092116_m.fxe"},
	{0x187474C0DB6EC8CCULL, "6085685_m.fxe"},
	{0x18748E3B29323FF0ULL, "t3_glk_h02.tnt"},
	{0x1874D0AB07F95A6FULL, "vxlt_0_lowtown_a2_ghy_0_0.vlm"},
	{0x1875848BDD4913F9ULL, "53157974.ogg"},
	{0x1876BCE94DB843C2ULL, "hlac_lt_0_lowtown_docks_-3_0z.dds"},
	{0x1876F9D8B2528E97ULL, "vxlt_0_lowtown_docks_cqg_0_0.vlm"},
	{0x18770CE7CA409E62ULL, "fx_c_distortion.mao"},
	{0x18772D3AC88FEA69ULL, "mag291_ntbanter6116210.cl"},
	{0x187748E320B2E189ULL, "293731374.ogg"},
	{0x1877615BB9CF5580ULL, "art\\vfx\\combat\\vfx_gib_hlk_p_v2.anb"},
	{0x18777DFA7E09859DULL, "6206346_m.fxe"},
	{0x187784D631034013ULL, "962729334.ogg"},
	{0x187789A7D6FF5836ULL, "394373298.ogg"},
	{0x18779A047E9B0E2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_chest_out.gad"},
	{0x18780D161EDEFCA2ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.impa.gad"},
	{0x18785AC3F404BEF3ULL, "gen00fl_aveline6128801.cl"},
	{0x18788E5A3A2344A2ULL, "hlac_ht_0_hightown_-1_2y.dds"},
	{0x1878BD854A61D1F9ULL, "vxlt_0_lowtown_docks_a3_n_cvm_0_1.vlm"},
	{0x1878E1EB8690FEB0ULL, "664539825.ogg"},
	{0x18791A3348906A99ULL, "6123653_f.fxe"},
	{0x1879200554F3F639ULL, "llac_lt_0_lowtown_a2_-196609_2.dds"},
	{0x187A41A540F7CFAFULL, "6059345_f.fxe"},
	{0x187AACFBE5DB7D79ULL, "844040672.ogg"},
	{0x187B0E0E6639DA96ULL, "933256776.ogg"},
	{0x187B8400055504A4ULL, "cod_hst_city_elf_cultur.utp"},
	{0x187BD441EF94A1ADULL, "ser200_aveline6201522.cl"},
	{0x187BE51991344D54ULL, "l_and341ar_mines.bnk"},
	{0x187BE9D89866A6DCULL, "vxlt_0_lowtown_a2_n_c_es4_0_0.vlm"},
	{0x187C255014F68A96ULL, "vxdr_primevaltomb_829_0_1.vlm"},
	{0x187C570F258872C6ULL, "vxlt_gallowstemplar_5qj_0_0.vlm"},
	{0x187C63A0A171EEE2ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.dura.evt"},
	{0x187C96889FC3E843ULL, "gen_im_trs_slv_shd_wdn_r0.uti"},
	{0x187CDE483E63D173ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier1_epilogue.ani"},
	{0x187D0A329B7D9516ULL, "43898464.ogg"},
	{0x187D8AD93E8C34CBULL, "vxht_0_hightown_a2_b4d_0_0.vlm"},
	{0x187DDA5A133F50C9ULL, "6105409_m.fxe"},
	{0x187E197F9BD6A976ULL, "210160536.ogg"},
	{0x187E252A2D4DBEF7ULL, "6084483_m.fxe"},
	{0x187E7CEA9AD7CC83ULL, "gen_im_cns_tom_talent_01.uti"},
	{0x187EAF047EA150DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_chest_out.evt"},
	{0x187EDD751D22C114ULL, "6108700_m.fxe"},
	{0x187F30161EE55699ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.impa.evt"},
	{0x187F32AEF3F56383ULL, "ht_0_hightown_28_0_lc.mmh"},
	{0x187F96DD4E65E7E0ULL, "vxca_caverns1_mines_dph_0_0.vlm"},
	{0x187FD154C188FDEDULL, "w_sundermount_70_0.phy"},
	{0x187FF812AC397CDCULL, "llac_dr_0_deeproads_17039_0.dds"},
	{0x1880340696EF3F2CULL, "gen_im_trs_prf_rng_r0.uti"},
	{0x188251B3F5A09998ULL, "vxht_0_hightown_c8b_0_0.vlm"},
	{0x1882D6396D3DC6B3ULL, "6170869_m.fxe"},
	{0x1882F8EC05CF83AFULL, "w_sundermount_245_0.mao"},
	{0x188304EC05CF984CULL, "w_sundermount_245_0.mmh"},
	{0x18830AEC05CFA23EULL, "w_sundermount_245_0.msh"},
	{0x18832A84F270A4BEULL, "1040086175.ogg"},
	{0x188331DA8ED25231ULL, "3531144.ogg"},
	{0x188362A0A1780BF5ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.dura.gad"},
	{0x1883681B6330A12AULL, "278686795.ogg"},
	{0x1883A8BEFD35A4DAULL, "w_encounter1_water_8.mmh"},
	{0x1883ACBEFD35AB11ULL, "w_encounter1_water_8.mao"},
	{0x1883BABEFD35C320ULL, "w_encounter1_water_8.msh"},
	{0x1883BBC1733D9811ULL, "hlac_w_bonepit_65538_0z.dds"},
	{0x1883C6D8D4CD54E7ULL, "34352389.ogg"},
	{0x18841C333B74319BULL, "leave_05_l3.mao"},
	{0x188496D31AADB89BULL, "static_static_p24.psh.fxo"},
	{0x1884AEF48A0E0E1BULL, "6167307_f.fxe"},
	{0x1884E8706F8D1F19ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_eye_flma_0d.dds"},
	{0x188543012994E75EULL, "31410207.ogg"},
	{0x1885533FF792517AULL, "6143224_m.fxe"},
	{0x1885557A24F7B943ULL, "6171268_m.fxe"},
	{0x1885CC65CA4CABE3ULL, "vxw_blightlands_6b_0_0.vlm"},
	{0x1886D254C18F1D99ULL, "w_sundermount_70_0.rml"},
	{0x18870D95AC808252ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintinga_sa.dds"},
	{0x18871B59CEE02B82ULL, "framebuffereffects_screencolorclear_p0.psh.fxo"},
	{0x18874BA5F1FC01D5ULL, "949491601.ogg"},
	{0x188753989434521EULL, "6026954_m.fxe"},
	{0x188795D0CD06CDF1ULL, "zz_otl_debug6060670.cl"},
	{0x1887BBEEFCD8C186ULL, "6139294_m.fxe"},
	{0x1887FF650AA0407FULL, "gen00fl_mrl_act3_mirror_mh.stg"},
	{0x18880E6C67466487ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_anders_fury.gad"},
	{0x188881781E4728B6ULL, "6143134_m.fxe"},
	{0x1888B773DF406366ULL, "seb300im_maura_gloves.uti"},
	{0x1888C04DA9FE074BULL, "vxca_caverns2_845_0_0.vlm"},
	{0x18891FB8E631120EULL, "hlac_ht_chantry_2534_22y.dds"},
	{0x188975B48200D99DULL, "hlac_ht_0_hightown_a3_-131074_2z.dds"},
	{0x1889DED146144C39ULL, "6155721_m.fxe"},
	{0x188A61D0F36FF0D1ULL, "6117943_m.fxe"},
	{0x188B6F6851AFD845ULL, "pro000_captain_wright6004335.cl"},
	{0x188BFCB630824D4BULL, "vxht_0_hightown_a2_b9g_0_0.vlm"},
	{0x188C38161EEFEE88ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.impa.ani"},
	{0x188C3F9FEB2E059AULL, "6036429_m.fxe"},
	{0x188C5669BF84DDB9ULL, "hlac_lt_0_lowtown_docks_a3_131071_0y.dds"},
	{0x188C5F02472FA010ULL, "800367129.ogg"},
	{0x188CB7047EAD9BF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_chest_out.ani"},
	{0x188CBF40E9EC9FEEULL, "6060699_m.fxe"},
	{0x188CE2A57C31982AULL, "780843646.ogg"},
	{0x188D471F4BCF494BULL, "hlac_w_blightlands_196604_0x.dds"},
	{0x188DC9A8FE2A3DCFULL, "sib131_slaver_lord6165190.cl"},
	{0x188DDE6DCFB09E7DULL, "llac_lt_0_lowtown_a2_n_c_131071_6.dds"},
	{0x188E0B7A37E4F8F8ULL, "159780477.ogg"},
	{0x188E0C4654ED0CF9ULL, "lt_0_lowtown_n_40_0.rml"},
	{0x188E394A609B79F8ULL, "1012026444.ogg"},
	{0x188E8D9ACD224876ULL, "6139805_m.fxe"},
	{0x188F036C674C7094ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_anders_fury.evt"},
	{0x188F14A204054D72ULL, "dae000_fenarel6128126.cl"},
	{0x188F64EE58C667A5ULL, "137812368.ogg"},
	{0x188F6AFF02EAB92FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_charge.gad"},
	{0x188FDC596CCC8C1FULL, "zz_firestaff3.cut"},
	{0x188FFB91DBF38CAEULL, "ca_cave_2193.anb"},
	{0x189074B4223BB8E3ULL, "6025103_m.fxe"},
	{0x18908354C80653D8ULL, "662835028.ogg"},
	{0x18909EB97BE7F30BULL, "ave200_donnic6111640.cl"},
	{0x1890B5483E74341DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier1_epilogue.gad"},
	{0x1890D7EFD91085EDULL, "w_woundedcoast_91_0.mmh"},
	{0x1890DBEFD9108C3EULL, "w_woundedcoast_91_0.mao"},
	{0x1890E9EFD910A473ULL, "w_woundedcoast_91_0.msh"},
	{0x1890F023C486D5ADULL, "109644458.ogg"},
	{0x189109569004E687ULL, "hlac_lt_darktown_-1_65z.dds"},
	{0x1892453EFFE2A91EULL, "hlac_ca_caverns1_mines_24426_0z.dds"},
	{0x189261F1399BAD13ULL, "vxlt_blackemporium_tt_0_0.vlm"},
	{0x189262EAA832486BULL, "vxlt_hangedman_n_2y2_0_0.vlm"},
	{0x18929D1F26A73306ULL, "6171548_m.fxe"},
	{0x1892FB27F6AF4DFCULL, "llac_ht_0_hightown_a2_2_1.dds"},
	{0x189308B00CA3D11BULL, "static_static_p55.psh.fxo"},
	{0x1893686F8A9C81BCULL, "ico_ring_1.dds"},
	{0x18938CDD0D89919AULL, "vxdr_0_deeproads_cu9_0_1.vlm"},
	{0x1893B49DA19A92ADULL, "lt_0_lowtown_docks_ll.mmh"},
	{0x1893C1ABF36E5299ULL, "6024551_m.fxe"},
	{0x189445F6B3ECA740ULL, "llac_lt_gallowstemplar_f_2384_2.dds"},
	{0x1894E9E808A35D87ULL, "6153794_m.fxe"},
	{0x1895038C5637DD0CULL, "vxdr_primevaltomb_75d_0_0.vlm"},
	{0x1895AD1BFEDFD015ULL, "275506891.ogg"},
	{0x1895AE7800E0A58BULL, "vxlt_undercityserialkiller_205_0_0.vlm"},
	{0x1895D261CEB36144ULL, "373329027.ogg"},
	{0x18961626C09BB77AULL, "378154503.ogg"},
	{0x18963C8EE7E0E630ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersml.mao"},
	{0x1896408EE7E0ED43ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersml.mmh"},
	{0x18964E8EE7E1049DULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersml.msh"},
	{0x18967FFF02F0FBDCULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_charge.evt"},
	{0x189699BEF5735B40ULL, "173360255.ogg"},
	{0x1896AF09226D97A7ULL, "hla_ht_0_hightown_a2_-65534_0y.dds"},
	{0x1896B739B898011DULL, "9630360.ogg"},
	{0x1896CB8C633AEF44ULL, "follower_banter6173044.cl"},
	{0x1896F05A91745424ULL, "456465826.ogg"},
	{0x1896FADF2CCE841FULL, "461918745.ogg"},
	{0x1896FEF54841C822ULL, "vxca_caverns2_9fg_0_0.vlm"},
	{0x18970C892F3B448DULL, "hlac_lt_gallowstemplar_f_2388_0y.dds"},
	{0x1897322A9BF41FCBULL, "vxca_caverns2_834_0_0.vlm"},
	{0x1897F1C7A91E8E2FULL, "6188633_m.fxe"},
	{0x18981E16DFE6B8D0ULL, "396480983.ogg"},
	{0x189841F839AA12BBULL, "6028922_m.fxe"},
	{0x18988D6818603C32ULL, "dae200_cullen6140489.cl"},
	{0x189972AE3C9F38AAULL, "hlac_ht_mansion2_818_15z.dds"},
	{0x1899C248BD4D75BEULL, "mcr372ar_gallows_templar_nt.lst"},
	{0x189A1F0ECCCB6CDCULL, "follower_banter6171914.cl"},
	{0x189A575B6A7F358AULL, "vxw_woundedcoast_q4_0_0.vlm"},
	{0x189A878BF2FD325EULL, "w_blightlands_40_0.rml"},
	{0x189ABFFDA1900C52ULL, "6067367_f.fxe"},
	{0x189B16DB17A5EF38ULL, "vxht_0_hightown_a4t_0_0.vlm"},
	{0x189B647040F4BE3FULL, "dae000_gamlen6162232.cl"},
	{0x189BBBBE8315A4E9ULL, "hlac_dr_0_deeproads_5849_9z.dds"},
	{0x189BC3F34AD29BFCULL, "vxlt_0_lowtown_a2_h27_0_0.vlm"},
	{0x189C25768C0A2725ULL, "vxlt_0_lowtown_a2_n_c_exx_0_0.vlm"},
	{0x189CC4CC4AC6976AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killervase_dp.dds"},
	{0x189CE0C690965C33ULL, "lgt251ip_chest_disturbed.utp"},
	{0x189CFB6C6758A0CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_anders_fury.ani"},
	{0x189D88F2B0802DDBULL, "279032013.ogg"},
	{0x189DBEE893E66593ULL, "static_static_p19.psh.fxo"},
	{0x189DE6ABB40EAF87ULL, "vxw_woundedcoast_10w_0_0.vlm"},
	{0x189DF5D2DBC70557ULL, "dae000cr_grand_cleric6173647.cl"},
	{0x189E399722EC6EFAULL, "vxht_0_hightown_a2_c3p_0_1.vlm"},
	{0x189E5FD0272F9896ULL, "vxlt_0_lowtown_e2x_0_0.vlm"},
	{0x189E68CA818A10E2ULL, "mi_cyc.mao"},
	{0x189E75C5F9C01CBEULL, "vxlt_gallowscourtyard_n_6zx_0_1.vlm"},
	{0x189F92384A3F2AC5ULL, "320213019.ogg"},
	{0x18A024F67E92A73EULL, "vxht_mansion2_2yj_0_0.vlm"},
	{0x18A0DDB938607945ULL, "6082304_m.fxe"},
	{0x18A10EFC189D4B86ULL, "6174648_m.fxe"},
	{0x18A11C8B1F32C160ULL, "vxlt_shop_es_0_0.vlm"},
	{0x18A14591CE3CA8C6ULL, "6198699_m.fxe"},
	{0x18A2613C63C6534EULL, "lt_0_lowtown_docks_129_0.msh"},
	{0x18A26BB4EFF80ABBULL, "447559139.ogg"},
	{0x18A26F3C63C66A9FULL, "lt_0_lowtown_docks_129_0.mao"},
	{0x18A27B3C63C67F3CULL, "lt_0_lowtown_docks_129_0.mmh"},
	{0x18A28856740ADADAULL, "llac_lt_0_lowtown_a2_-65535_2.dds"},
	{0x18A30004739FC643ULL, "6031977_m.fxe"},
	{0x18A307FDF6996A64ULL, "6139726_m.fxe"},
	{0x18A4131C20441B79ULL, "vxlt_0_lowtown_docks_a3_n_cyp_0_1.vlm"},
	{0x18A4743299FB602BULL, "207593690.ogg"},
	{0x18A487FF02FD46F5ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_charge.ani"},
	{0x18A58FA63B7655C0ULL, "gen00fl_merrill6129473.cl"},
	{0x18A610F13D314C37ULL, "6117182_f.fxe"},
	{0x18A6D0D8D6C60CCEULL, "969519311.ogg"},
	{0x18A6F920492FFC34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_major_v1.ani"},
	{0x18A70F3BA81D91CBULL, "hlac_ca_caverns2prc_5802_36x.dds"},
	{0x18A77C0EFB25CA96ULL, "hlac_lt_hangedman_-1_25z.dds"},
	{0x18A78A0C3209DAD3ULL, "781171428.ogg"},
	{0x18A7B6E9DED154F6ULL, "6109770_m.fxe"},
	{0x18A82DC44B060472ULL, "vxlt_gallowstemplar_f_5hj_0_0.vlm"},
	{0x18A83005AC835402ULL, "glo_plc_doors.bnk"},
	{0x18A874E9062A40ACULL, "334175584.ogg"},
	{0x18A8995DF934800EULL, "6045966_m.fxe"},
	{0x18A8C1E6A7A17AADULL, "122907594.ogg"},
	{0x18AA2596929D1C49ULL, "883039594.ogg"},
	{0x18AA32CDA082A875ULL, "hlac_lt_0_lowtown_n_-131072_0x.dds"},
	{0x18AA395678AC0D35ULL, "hlac_lt_shop_-1_5x.dds"},
	{0x18AA3D289ED45FEFULL, "hlac_dr_0_deeproads_17038_6x.dds"},
	{0x18AA6B9CEE957B46ULL, "643420508.ogg"},
	{0x18AAB4AE2364EC10ULL, "646849379.ogg"},
	{0x18AAC32740EFF164ULL, "vxw_encounter2_camp_h8_0_0.vlm"},
	{0x18AB495307FE2311ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masb.mmh"},
	{0x18AB635307FE4FCFULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masb.msh"},
	{0x18ABCA01F2931547ULL, "vxht_brothel_n_5fm_0_0.vlm"},
	{0x18ABD49F373BD3FCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2_b.ani"},
	{0x18ABDF47405346BCULL, "256495759.ogg"},
	{0x18AC10D259797F7EULL, "zz_isa_debug6175287.cl"},
	{0x18AC1D0957A28550ULL, "730342661.ogg"},
	{0x18AC4B94BAD52B47ULL, "267425711.ogg"},
	{0x18ACF0F199C0A678ULL, "mrl310_merrill6143821.cl"},
	{0x18AD2BDD0D6EACB4ULL, "_global.bnk"},
	{0x18AD2F7E98DE3401ULL, "gen_im_npc_dwf_arm_cht_cxt2.uti"},
	{0x18ADC8F65F77DBF0ULL, "326521801.ogg"},
	{0x18ADF62049361596ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_major_v1.gad"},
	{0x18AE77B161B9DF72ULL, "llac_lt_0_lowtown_a3_n_c_-65537_4.dds"},
	{0x18AEFAD342F0E00CULL, "dae000_bartrand_a.stg"},
	{0x18AF04C62F1619BEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.evt"},
	{0x18AF4EEBCEC36165ULL, "vxlt_0_lowtown_docks_a3_n_aou_0_0.vlm"},
	{0x18AF9B5C6B4DC373ULL, "hlac_lt_0_lowtown_65536_9z.dds"},
	{0x18B0587FA72EE4AEULL, "hlac_ht_chantry_n_2534_46z.dds"},
	{0x18B2732C66E1BA9AULL, "hlac_ca_cavern1_haunted_24424_20z.dds"},
	{0x18B27BE7349883F2ULL, "vxdr_0_deeproads_a6w_0_0.vlm"},
	{0x18B28F29969C53E5ULL, "hlac_lt_0_lowtown_-131073_1y.dds"},
	{0x18B2B44CBC2F246CULL, "469029820.ogg"},
	{0x18B2B81157388C91ULL, "6068094_m.fxe"},
	{0x18B2E19F3742083EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2_b.gad"},
	{0x18B2F4F727D7FA0BULL, "vxw_bonepitmining_a2_1qq_0_0.vlm"},
	{0x18B3295C8CFF8D20ULL, "vxw_blightlands_10r_0_0.vlm"},
	{0x18B4C3BB62CE798FULL, "dre000_varric6029403.cl"},
	{0x18B4CB4CA8CF494FULL, "6082135_m.fxe"},
	{0x18B4D75AF3892E89ULL, "vxdr_primevaltomb_7qq_0_0.vlm"},
	{0x18B4E6910A9BBC55ULL, "vxlt_gallowstemplar_f_4sq_0_0.vlm"},
	{0x18B51120493C62F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_major_v1.evt"},
	{0x18B51234C878179AULL, "vxw_blightlands_ef_0_0.vlm"},
	{0x18B561AD9606ACD1ULL, "mcr371_hostile_fenris6082210.cl"},
	{0x18B5F26254A74A40ULL, "654828379.ogg"},
	{0x18B61BC62F1C5FE9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.gad"},
	{0x18B628C6D0C76D76ULL, "6102670_m.fxe"},
	{0x18B65F7A3514F16DULL, "6150502_m.fxe"},
	{0x18B68F5304F299DFULL, "6113981_m.fxe"},
	{0x18B69AC3B40EF176ULL, "vxlt_backalley_n_1of_0_1.vlm"},
	{0x18B6AAF1A89953E1ULL, "gen_im_wep_war_one_act0_04.uti"},
	{0x18B6E4D75D3BC37AULL, "160537581.ogg"},
	{0x18B6EBB65C95009DULL, "llac_w_bonepitmining_a2_262148_0.dds"},
	{0x18B6F5CA5C95E6B7ULL, "594502545.ogg"},
	{0x18B7105BF780E1D7ULL, "258572651.ogg"},
	{0x18B722C744304DFBULL, "899823011.ogg"},
	{0x18B7643C9EDA26FDULL, "vxw_encounter3_fx_0_0.vlm"},
	{0x18B7664308B97596ULL, "688292030.ogg"},
	{0x18B7FDC2BA0F7C9DULL, "vxlt_0_lowtown_n_ef8_0_0.vlm"},
	{0x18B860494CFE1C4BULL, "162803025.ogg"},
	{0x18B8980532145B6DULL, "466068638.ogg"},
	{0x18B909306FF65825ULL, "vxlt_0_lowtown_a2_n_c_erf_0_0.vlm"},
	{0x18B9350432E609E4ULL, "vxw_encounter2_camp_c9_0_0.vlm"},
	{0x18B981825D1556EBULL, "769512182.ogg"},
	{0x18B9BC9F3747E85DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2_b.evt"},
	{0x18B9BDB6AC7138B7ULL, "mcr370_messenger6071834.cl"},
	{0x18BA353DBDEE71EBULL, "hlac_lt_0_lowtown_docks_n_-196611_1z.dds"},
	{0x18BA3BDEE4892DC7ULL, "vxht_brothel_n_5il_0_0.vlm"},
	{0x18BABEC5D1E25EAEULL, "6084697_m.fxe"},
	{0x18BAC0C4336681F9ULL, "dae021_sister_lorena_sta6153108.cl"},
	{0x18BAC3965C1D49A1ULL, "poses.gda"},
	{0x18BB6D91DBF8BAA3ULL, "hlac_ca_caverns1_mines_24424_5z.dds"},
	{0x18BBB15F4ED52BCDULL, "fen110_anso6095004.cl"},
	{0x18BBE34AA859D671ULL, "100473770.ogg"},
	{0x18BBFF19916A5D06ULL, "l_and341ar_mines.wbd"},
	{0x18BC42BFE4786024ULL, "art\\characters\\creatures\\spider\\animation\\electr.ani"},
	{0x18BC62BD7670D3E0ULL, "83154137.ogg"},
	{0x18BCAE230B1BD406ULL, "748362277.ogg"},
	{0x18BCECC62F222E37ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.ani"},
	{0x18BCF57B6F6E2D81ULL, "6044052_m.fxe"},
	{0x18BD00DB33D8029CULL, "creeper_bloodmagic.dds"},
	{0x18BD677E54B7A6FFULL, "925939657.ogg"},
	{0x18BDFDE8F0EB738CULL, "252927537.ogg"},
	{0x18BE1A34BC27DE6EULL, "hm_conspiracy.mor"},
	{0x18BED35F471E54DAULL, "33864985.ogg"},
	{0x18BF4784A799F836ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_varr.mmh"},
	{0x18BF5184A79A0934ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_varr.msh"},
	{0x18BF8CB4B72448C7ULL, "blood_cycle.mao"},
	{0x18BFBB13CA06B179ULL, "vxlt_0_lowtown_a2_n_c_bz9_0_0.vlm"},
	{0x18BFC6D102419599ULL, "vxw_sundermount_21e_0_0.vlm"},
	{0x18BFF101E3979F20ULL, "6039549_m.fxe"},
	{0x18C007168CC85C45ULL, "hlac_lt_0_lowtown_a2_n_c_1_2z.dds"},
	{0x18C13139363FE66CULL, "6042785_defeated.cut"},
	{0x18C1D1E424403F5EULL, "6097603_f.fxe"},
	{0x18C1D43459190266ULL, "vxw_sundermount_38h_0_0.vlm"},
	{0x18C1F5B39732D434ULL, "vxht_chantry_210_0_0.vlm"},
	{0x18C1FFD0369EED07ULL, "549217382.ogg"},
	{0x18C23FBFE47CC686ULL, "art\\characters\\creatures\\spider\\animation\\electr.gad"},
	{0x18C25E671D04EE68ULL, "6069382_m.fxe"},
	{0x18C2AAD7BD7C307BULL, "6091838_m.fxe"},
	{0x18C2FE91DC1EEB12ULL, "ca_cave_2193.rml"},
	{0x18C31DB32F136AE0ULL, "vxlt_gallowscourtyard_77v_0_0.vlm"},
	{0x18C38411BA6E301AULL, "vxw_blightlands_ng_0_0.vlm"},
	{0x18C3BF32F1930514ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_maga_0d.dds"},
	{0x18C4618F55626791ULL, "vxht_mansion2b_2vl_0_0.vlm"},
	{0x18C46E043D30CC27ULL, "326037349.ogg"},
	{0x18C4EE9E7789A21FULL, "6110218_m.fxe"},
	{0x18C50CA0A60509F6ULL, "vxlt_backalley_n_1ji_0_1.vlm"},
	{0x18C51E17C1D4513AULL, "w_woundedcoast_89_0.mmh"},
	{0x18C52217C1D45871ULL, "w_woundedcoast_89_0.mao"},
	{0x18C53017C1D46F80ULL, "w_woundedcoast_89_0.msh"},
	{0x18C569C4625A8425ULL, "gen00fl_aveline6122113.cl"},
	{0x18C5D61990D03F7DULL, "vxw_encounter3_cw_0_0.vlm"},
	{0x18C61E2C5EB73891ULL, "follower_banter6141604.cl"},
	{0x18C6AF2177FCDCFDULL, "vxca_caverns1_ih1_0_0.vlm"},
	{0x18C7109A69C6B91EULL, "6030661_m.fxe"},
	{0x18C72CD79C90437AULL, "hlac_ht_chantry_2534_30z.dds"},
	{0x18C761290ED72D57ULL, "bink_bink_p0.vsh.fxo.dx11"},
	{0x18C771B260CA7DDEULL, "869407783.ogg"},
	{0x18C77B0D61EC70A5ULL, "vxlt_0_lowtown_a2_n_c_ewe_0_0.vlm"},
	{0x18C77B8BF3240BD6ULL, "w_blightlands_40_0.mao"},
	{0x18C7878BF3242075ULL, "w_blightlands_40_0.mmh"},
	{0x18C7898BF32423EBULL, "w_blightlands_40_0.msh"},
	{0x18C7B106577A787DULL, "vxht_keep_6bl_0_0.vlm"},
	{0x18C8348B07EDF03FULL, "6028441_m.fxe"},
	{0x18C88167CB361F75ULL, "hlac_lt_0_lowtown_-196609_3y.dds"},
	{0x18C8D7F891D61729ULL, "hlac_lt_0_lowtown_2_0x.dds"},
	{0x18C8E0EE10E12B54ULL, "6105934_m.fxe"},
	{0x18C93ABFE482DD65ULL, "art\\characters\\creatures\\spider\\animation\\electr.evt"},
	{0x18C9E100F310CD0DULL, "vxht_0_hightown_c3o_0_1.vlm"},
	{0x18CACD81052486EFULL, "qun110_petrice6042150.cl"},
	{0x18CAD3A2C7CEF985ULL, "866971895.ogg"},
	{0x18CBEB062C9F1164ULL, "dae000_bodhan_letter6144308.cl"},
	{0x18CC69229A18DFF8ULL, "vxlt_shop_j2_0_0.vlm"},
	{0x18CD1A05ACA24B20ULL, "glo_plc_doors.wbd"},
	{0x18CD228107D2BFF1ULL, "hlac_ca_caverns2prc_5802_28y.dds"},
	{0x18CD22D79FD7DB16ULL, "mag360_samson6054685.cl"},
	{0x18CD6FDEB66D2E71ULL, "6155574_m.fxe"},
	{0x18CDFBE8C06A25C6ULL, "435742067.ogg"},
	{0x18CE1C9948E63081ULL, "mag291_demon_marethari6029058.cl"},
	{0x18CE3F8AF5B32CEFULL, "w_bonepit_68_0.mao"},
	{0x18CE4B8AF5B3408CULL, "w_bonepit_68_0.mmh"},
	{0x18CE518AF5B34B7EULL, "w_bonepit_68_0.msh"},
	{0x18CE81A56F8BB5D8ULL, "vxht_brothel_n_6eh_0_0.vlm"},
	{0x18CEA55C55F4C4AAULL, "vxlt_gallowstemplar_5nk_0_0.vlm"},
	{0x18CF40F74486FE6AULL, "75161370.ogg"},
	{0x18CF4C07A3D80D39ULL, "1056808957.ogg"},
	{0x18CFC137B8EEBD6BULL, "195983402.ogg"},
	{0x18CFCED61C125CCFULL, "hlac_lt_0_lowtown_docks_a3_n_262143_0y.dds"},
	{0x18D02F96880C1393ULL, "vxlt_0_lowtown_docks_a3_n_95i_0_0.vlm"},
	{0x18D11AC4BAADE5F9ULL, "and111_lirene6085103.cl"},
	{0x18D14D2D35BA10AFULL, "gen00fl_anders6111316.cl"},
	{0x18D1DD77170247CFULL, "vxlt_undercityserialkiller_1yy_0_0.vlm"},
	{0x18D208825602B941ULL, "ch1_s1y.psh.fxo"},
	{0x18D2159A30C7680FULL, "hlac_dr_primevalentrance_1726_4y.dds"},
	{0x18D3485C0699AA2AULL, "vxht_mansion_n_4y3_0_0.vlm"},
	{0x18D3A9FA22A7C141ULL, "874356648.ogg"},
	{0x18D3DF9ED79BAB8BULL, "885233639.ogg"},
	{0x18D3E03F5D7F052DULL, "gen_im_cft_rec_psn_01.uti"},
	{0x18D5225712EBDE87ULL, "6054878_f.fxe"},
	{0x18D537578A459200ULL, "204735256.ogg"},
	{0x18D53A65D1109216ULL, "vxlt_0_lowtown_ha4_0_0.vlm"},
	{0x18D55223D002D735ULL, "mag100_cullen6044513.cl"},
	{0x18D64D69322B3006ULL, "gen00fl_sebastian6180446.cl"},
	{0x18D65600132E6AEFULL, "cod_bks_lite_foreshadow.utp"},
	{0x18D672C9A7B50290ULL, "hlac_lt_0_lowtown_docks_n_196608_1z.dds"},
	{0x18D684D8BCBCF9C3ULL, "one121_trap6112685.cl"},
	{0x18D6D5F0C1F80D09ULL, "725705674.ogg"},
	{0x18D78E2728AE30FCULL, "gen00fl_anders6121178.cl"},
	{0x18D8010898802D0CULL, "art\\vfx\\textures\\vfx_gui_mov_target.dds"},
	{0x18D9D3D6CC9C29AAULL, "hlac_ca_caverns2_5802_29x.dds"},
	{0x18D9E77ADC2A9365ULL, "follower_barks6173404.cl"},
	{0x18D9FED945FA84CFULL, "dae300_cullen6135887.cl"},
	{0x18D9FF5264FE8266ULL, "vxht_0_hightown_ds_0_0.vlm"},
	{0x18DA23D5BA36ED59ULL, "hlac_lt_0_lowtown_a2_131071_5x.dds"},
	{0x18DA2E6508E5AE05ULL, "dae100_bran6086639.cl"},
	{0x18DA89E0EEF81470ULL, "vrc221_varric6044051.cl"},
	{0x18DB1B92A31CA6E4ULL, "mag170_thrask6020544.cl"},
	{0x18DB629110C3594FULL, "ht_0_hightown_a3_water_15346.mao"},
	{0x18DB6E9110C36DECULL, "ht_0_hightown_a3_water_15346.mmh"},
	{0x18DB749110C377DEULL, "ht_0_hightown_a3_water_15346.msh"},
	{0x18DBBB22F2F9A6D2ULL, "6027966_m.fxe"},
	{0x18DC1483D1C895C4ULL, "vxht_brothel_n_57p_0_1.vlm"},
	{0x18DC7BC95CAB98EBULL, "ave121_jeven6105710.cl"},
	{0x18DD1844EECF90EEULL, "hlac_lt_0_lowtown_a2_n_-65536_1y.dds"},
	{0x18DD6CCC6F9E59CAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexc.mao"},
	{0x18DD70CC6F9E6019ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexc.mmh"},
	{0x18DD7ACC6F9E7107ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexc.msh"},
	{0x18DD9B26ED69A802ULL, "static_static_p32.psh.fxo.dx11"},
	{0x18DDD6A96BADD8C4ULL, "6142104_m.fxe"},
	{0x18DE0687B938F21DULL, "lgt250_way_dalish_conflict6147169.cl"},
	{0x18DE9B07482ABFADULL, "366388142.ogg"},
	{0x18DF0B89A2394220ULL, "6139186_m.fxe"},
	{0x18DF3D5A667D4809ULL, "6128682_m.fxe"},
	{0x18DF81BE2ACC212BULL, "vxw_encounter1_gc_0_0.vlm"},
	{0x18DF841667FC281CULL, "mrl150_merrill_pt3.stg"},
	{0x18E067CB9E5A611AULL, "vxw_blightlands_ta_0_0.vlm"},
	{0x18E09679C02ACA33ULL, "54162878.ogg"},
	{0x18E1149530EC8B08ULL, "33973474.ogg"},
	{0x18E134EF48ABA23BULL, "art\\vfx\\cinematic\\vfx_orsino_swirly.impa.evt"},
	{0x18E16B4D3762AD9FULL, "101786501.ogg"},
	{0x18E1ACBA8B0C02C0ULL, "vxht_mansion2b_1ki_0_0.vlm"},
	{0x18E27F14F9B57ECFULL, "6074493_m.fxe"},
	{0x18E280CB630B2AB8ULL, "45991608.ogg"},
	{0x18E3004AFC699D2DULL, "482104893.ogg"},
	{0x18E324190D311BD6ULL, "487303452.ogg"},
	{0x18E359D58103AF4BULL, "vxht_brothel_n_4fd_0_0.vlm"},
	{0x18E3875B786BD67BULL, "209638607.ogg"},
	{0x18E40D18F9ED8E5DULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_5y.dds"},
	{0x18E494C03B66A97DULL, "vxht_keep_6hj_0_0.vlm"},
	{0x18E4E1F9A4430C1EULL, "270695562.ogg"},
	{0x18E64DA30C63D429ULL, "309782138.ogg"},
	{0x18E687F8E103B5CEULL, "vxlt_gallowstemplar_f_5pn_0_0.vlm"},
	{0x18E6C334B48A60C9ULL, "vxdr_0_deeproads_fjw_0_0.vlm"},
	{0x18E6CE7EED2A8768ULL, "507292821.ogg"},
	{0x18E70F87F1DCB164ULL, "hlac_dr_0_deeproads_5845_15x.dds"},
	{0x18E72DEF48B0027CULL, "art\\vfx\\cinematic\\vfx_orsino_swirly.impa.gad"},
	{0x18E732D1A17118B9ULL, "vxlt_0_lowtown_a2_n_c_gyy_0_0.vlm"},
	{0x18E7569203EB671AULL, "87883464.ogg"},
	{0x18E7689259843D9BULL, "vxlt_darktown_5pp_0_0.vlm"},
	{0x18E7C28C6999E8E4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_har_flma_0s.dds"},
	{0x18E7E6C23E6EF2A2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curver.msh"},
	{0x18E7E8C23E6EF658ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curver.mmh"},
	{0x18E7F4C23E6F0AFBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curver.mao"},
	{0x18E804D4BE8E1E35ULL, "3823298.ogg"},
	{0x18E85FDF9D8B377AULL, "1068810909.ogg"},
	{0x18E8D9AC2A7BDD5EULL, "vxht_0_hightown_627_0_0.vlm"},
	{0x18E967E098A1F1A8ULL, "dae300_petrice6098089.cl"},
	{0x18E9683D74E80583ULL, "art\\vfx\\materials\\vfx_distortion_ball.mao"},
	{0x18E9E98C2791B93BULL, "6218905_m.fxe"},
	{0x18EA65B6B21FE407ULL, "gen00fl_anders6103751.cl"},
	{0x18EA8A3571A1ABEAULL, "6129141_m.fxe"},
	{0x18EAEA848F9B73ADULL, "qcr000_varric6208210.cl"},
	{0x18EBA5F82FCF2481ULL, "hlac_dr_0_deeproads_5843_7z.dds"},
	{0x18EBAD2F63047F96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\exit.ani"},
	{0x18EBC4AB522D205DULL, "mag160_feynriel6016584.cl"},
	{0x18EC4E7D8CC65304ULL, "vxw_bonepitdestroyed_a3_jj_0_0.vlm"},
	{0x18EC53E40A129E74ULL, "500227075.ogg"},
	{0x18ECDE69A8CF5B21ULL, "6035913_m.fxe"},
	{0x18ED3B20E87D7D38ULL, "395312326.ogg"},
	{0x18EDC7F1001714DDULL, "hlac_w_bonepit_262147_0y.dds"},
	{0x18EE1743C5BE4B66ULL, "310741031.ogg"},
	{0x18EE2CEF48B61F7AULL, "art\\vfx\\cinematic\\vfx_orsino_swirly.impa.ani"},
	{0x18EE45DD0DA67066ULL, "_global.wbd"},
	{0x18EEABCFCA9406A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_forward.gad"},
	{0x18EEAD338F8A8C6FULL, "143815762.ogg"},
	{0x18EEF25DA5E3F686ULL, "227687943.ogg"},
	{0x18EF5A36390267ACULL, "vxht_keep_n_63a_0_0.vlm"},
	{0x18EF984A8A0046DAULL, "965092986.ogg"},
	{0x18F0F0F1EBAB974FULL, "6079793_m.fxe"},
	{0x18F11E17C1F98C51ULL, "w_woundedcoast_89_0.rml"},
	{0x18F1F3CE32505A70ULL, "mag120_macha6080630.cl"},
	{0x18F20201AAA7AE80ULL, "666081078.ogg"},
	{0x18F24C96E1594C07ULL, "6141128_m.fxe"},
	{0x18F2B5D49A763212ULL, "6113089_m.fxe"},
	{0x18F2C0AE48D138E7ULL, "vxht_mansion2b_28p_0_0.vlm"},
	{0x18F2E9F1EB1454CEULL, "6071778_m.fxe"},
	{0x18F2EFC23E7909E3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curver.phy"},
	{0x18F3444A6A219A06ULL, "dae000cr_mother.utc"},
	{0x18F42908B0A5B449ULL, "zz_dae_debug6029313.cl"},
	{0x18F4C9BDFA3AADE9ULL, "199490043.ogg"},
	{0x18F55F63C5A2B280ULL, "6095327_m.fxe"},
	{0x18F568C917357140ULL, "477829818.ogg"},
	{0x18F5CECFCA9A6099ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_forward.evt"},
	{0x18F5FD4FDAF20B86ULL, "6107800_m.fxe"},
	{0x18F6F4802E41404CULL, "70586424.ogg"},
	{0x18F79571CC7DEE48ULL, "vxlt_0_lowtown_a2_n_c_hnr_0_0.vlm"},
	{0x18F7D546181126C6ULL, "6163987_m.fxe"},
	{0x18F828E00F91E63AULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintinga_n.dds"},
	{0x18F851E592EF3D9BULL, "6117112_m.fxe"},
	{0x18F8BFB58BC5CD1BULL, "139955854.ogg"},
	{0x18F907F12FB93B63ULL, "vxw_encounter2_4w_0_0.vlm"},
	{0x18F91173CF9C7C77ULL, "vxlt_0_lowtown_a2_n_c_ff7_0_0.vlm"},
	{0x18F941236BFB7DBFULL, "t3_eey_gn3.tnt"},
	{0x18F976FC2337005EULL, "6048288_m.fxe"},
	{0x18F99034ADF4D5E4ULL, "gen00fl_dog6146965.cl"},
	{0x18F9B52F6310CA87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\exit.evt"},
	{0x18F9C03229F6ACF6ULL, "dre120tr_hurlock_team.ncs"},
	{0x18F9FAF32BD70E2AULL, "vxlt_gallowstemplar_5qf_0_0.vlm"},
	{0x18FA5CB0DAE4E1E1ULL, "454006219.ogg"},
	{0x18FA625ADD33857EULL, "one100dg_vanard6048250.cl"},
	{0x18FACD88F568226EULL, "246207852.ogg"},
	{0x18FC1B9B00D4E4ABULL, "darktown_male_sickly_sta6175322.cl"},
	{0x18FC4FEE3DC1BFBEULL, "zz_lgt_3_debug6140425.cl"},
	{0x18FCCD22E4A242BBULL, "841152110.ogg"},
	{0x18FE116247341FE2ULL, "vxw_woundedcoast_4s_0_0.vlm"},
	{0x18FE1F73232A380CULL, "434677005.ogg"},
	{0x18FE67BF9791305BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_rhand.ani"},
	{0x18FEAEE4A8A93B09ULL, "move_icon.mao"},
	{0x18FEBF50FA6925F6ULL, "butterfly01.dds"},
	{0x18FF3E1C83DAA444ULL, "hlac_lt_darktown_-1_59z.dds"},
	{0x18FF64BFA68D0222ULL, "6021245_m.fxe"},
	{0x18FF9A3C67AA59B2ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_04a_0t.dds"},
	{0x19002EFFAE852B2FULL, "6114438_m.fxe"},
	{0x1900D62F631721D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\exit.gad"},
	{0x1901328B3AC75167ULL, "vxht_mansion2b_25s_0_0.vlm"},
	{0x190185CE4F255851ULL, "989553460.ogg"},
	{0x1901C1620824C88AULL, "vxca_cavern1_haunted_jfz_0_0.vlm"},
	{0x190219B706447324ULL, "vxlt_0_lowtown_hav_0_0.vlm"},
	{0x19023C59545719AEULL, "vxht_mansion_n_4h4_0_0.vlm"},
	{0x1902CD2FD81BAC04ULL, "443668890.ogg"},
	{0x1902D49FA32E5B16ULL, "6110484_m.fxe"},
	{0x1902D6CFCAA4F888ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_forward.ani"},
	{0x1903408CE7EF9559ULL, "seb221_ruxton_harimann6177218.cl"},
	{0x19035FD4E5EFC24FULL, "hlac_ht_brothel_n_3373_2x.dds"},
	{0x19036033BDCAFFCDULL, "zz_ser_debug6078157.cl"},
	{0x1903E0139C4E1666ULL, "hlac_lt_0_lowtown_n_-1_0z.dds"},
	{0x1904834037A74414ULL, "109386995.ogg"},
	{0x1904A7E14036F151ULL, "6166172_m.fxe"},
	{0x1904CD3680D1B5BBULL, "vxlt_0_lowtown_docks_a3_cor_0_0.vlm"},
	{0x1904DADC5CB7E393ULL, "art\\vfx\\combat\\vfx_w_wns_front_close_c.impa.ani"},
	{0x19051C6DD72575DFULL, "6064593_m.fxe"},
	{0x19052C87F5EC8A80ULL, "256630679.ogg"},
	{0x19054C112F76D0B0ULL, "qcr000_arishok6060149.cl"},
	{0x1905D0277722461EULL, "6064388_f.fxe"},
	{0x1905D4AF4FC04893ULL, "978414037.ogg"},
	{0x1906322FD2BFF78DULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1achar.mmh"},
	{0x1906362FD2BFFEDEULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1achar.mao"},
	{0x19063D2D44E724A6ULL, "vxlt_hangedman_n_1i0_0_0.vlm"},
	{0x1906442FD2C01613ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1achar.msh"},
	{0x190666D37A74B3F4ULL, "vxlt_gallowstemplar_f_434_0_0.vlm"},
	{0x19068A857F8A1B99ULL, "vxw_encounter2_30_0_0.vlm"},
	{0x1906CB5D7E35D13BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\root.evt"},
	{0x190736F91A3197B0ULL, "6167866_m.fxe"},
	{0x190756B1EDC2FD2DULL, "dae000_hate_enemies6167603.cl"},
	{0x1907CD2A4BF38863ULL, "vxw_blightlands_cm_0_0.vlm"},
	{0x1907E8D9152D18DEULL, "6043388_m.fxe"},
	{0x190859ECCADBFFC2ULL, "qun160_vashothbandits6069276.cl"},
	{0x1908BF3C449B571AULL, "art\\characters\\weapons\\shields\\w_shd_004a.mmh"},
	{0x1908C33C449B5E51ULL, "art\\characters\\weapons\\shields\\w_shd_004a.mao"},
	{0x1908D13C449B7660ULL, "art\\characters\\weapons\\shields\\w_shd_004a.msh"},
	{0x190940CF690FF704ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgta_0s.dds"},
	{0x19094CBCBC8415C0ULL, "lgt260_way_blood_mage6148404.cl"},
	{0x1909666BCC6C1790ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic_1.mao"},
	{0x190991FF3646C2BBULL, "vxca_caverns1_e65_0_0.vlm"},
	{0x190A2D2C45263CA7ULL, "6180840_m.fxe"},
	{0x190A313289DE20B0ULL, "6076886_f.fxe"},
	{0x190A339DA33C7754ULL, "llac_lt_0_lowtown_docks_65534_3.dds"},
	{0x190AAAC76A7A202AULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbucket.mao"},
	{0x190AAEC76A7A2779ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbucket.mmh"},
	{0x190AB8C76A7A3867ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbucket.msh"},
	{0x190B529A0663E13BULL, "vxlt_gamlenshouse_n_pw_0_0.vlm"},
	{0x190BD72DF26CE4F1ULL, "180866475.ogg"},
	{0x190C086CFA97FFBEULL, "vxht_0_hightown_c39_0_0.vlm"},
	{0x190C6FBF979D7B52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_rhand.evt"},
	{0x190C735B93913C01ULL, "hlac_w_bonepitdestroyed_a3_262150_0y.dds"},
	{0x190CB51999FC1F8FULL, "lgt210cr_debt_thug2.utc"},
	{0x190CC45D7E3A317CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\root.gad"},
	{0x190CDD8E4861AF55ULL, "6039980_m.fxe"},
	{0x190CE7D60DD70A30ULL, "hlac_ca_caverns1_mines_24424_11z.dds"},
	{0x190D6B874C5417AFULL, "6061399_f.fxe"},
	{0x190DAEB3BFD66412ULL, "qun110_petrice6042135.cl"},
	{0x190DFD59677BE820ULL, "mrl150_merrill6084461.cl"},
	{0x190E063C25F5FE31ULL, "da2_prc_seb_de-de.cre"},
	{0x190E07CACDE5E715ULL, "abi_m_force_feild.bnk"},
	{0x190FA0A51B50535AULL, "vxht_mansion2b_1yu_0_0.vlm"},
	{0x190FA7A49C24190CULL, "vxht_brothel_n_2b5_0_0.vlm"},
	{0x190FF4E70F2F519DULL, "6151299_m.fxe"},
	{0x1910A8E3D702F732ULL, "167905446.ogg"},
	{0x191166B2C1616EADULL, "hlac_ht_0_hightown_a3_-131072_0z.dds"},
	{0x19116E7E0C138FA2ULL, "follower_banter6161214.cl"},
	{0x191190802772ACBBULL, "pro000_colonel_ewald6047387.cl"},
	{0x1911DD8FB6E5FF4EULL, "vxlt_gallowstemplar_f_5aa_0_0.vlm"},
	{0x1912020E4BC1B608ULL, "6001336_f.fxe"},
	{0x1912D2DC5CC413CAULL, "art\\vfx\\combat\\vfx_w_wns_front_close_c.impa.evt"},
	{0x19138109785E6457ULL, "vxht_mansion2_1xc_0_0.vlm"},
	{0x19138EBF97A3CEE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_rhand.gad"},
	{0x1913C35D7E404E7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\root.ani"},
	{0x1913F774A81B3168ULL, "610691236.ogg"},
	{0x191425E03F2F9938ULL, "228094470.ogg"},
	{0x19145AEC2F907088ULL, "dae000ip_crowd_ff2_talk1.utp"},
	{0x19148287DE7533C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy_ent.evt"},
	{0x1914C4541E1A76C9ULL, "fadeeffectmask.dds"},
	{0x191515FCB4A6A633ULL, "vxlt_foundry_n_vu_0_0.vlm"},
	{0x191556E22D4C85DAULL, "gen00fl_sebastian6174016.cl"},
	{0x1915B691ABA9E79CULL, "831138808.ogg"},
	{0x1915F6911E2BE61DULL, "vxw_bonepitmining_a2_mx_0_0.vlm"},
	{0x1917391AFE8B5481ULL, "vxlt_undercityserialkiller_140_0_0.vlm"},
	{0x1917433A7A524C2BULL, "vxlt_0_lowtown_a3_n_c_d06_0_0.vlm"},
	{0x1917721126671474ULL, "6174116_m.fxe"},
	{0x1917A68808892E06ULL, "hlac_lt_backalley_g_65537_4x.dds"},
	{0x19187D27908FEDEFULL, "vxht_0_hightown_a3_b96_0_0.vlm"},
	{0x1918AC47937FCE66ULL, "scaleform_depth.mao"},
	{0x1918B08DACA532FAULL, "hlac_lt_0_lowtown_docks_65533_2z.dds"},
	{0x1918F9919446C10CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsb.msh"},
	{0x1918FF919446CBFEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsb.mmh"},
	{0x19190B919446DF95ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsb.mao"},
	{0x19194498600F8D86ULL, "vxlt_0_lowtown_n_g9l_0_0.vlm"},
	{0x191946F0A2FE58F1ULL, "33120996.ogg"},
	{0x1919B1DC5CC9FA3DULL, "art\\vfx\\combat\\vfx_w_wns_front_close_c.impa.gad"},
	{0x1919BFE17FDBD2F8ULL, "vxlt_gallowscourtyard_754_0_0.vlm"},
	{0x1919F7C594117FE9ULL, "framebuffereffects_diffusiondof_p9.psh.fxo.dx11"},
	{0x191A0BAB4A6E11A3ULL, "vxw_bonepit_8c_0_0.vlm"},
	{0x191A38C3899645EAULL, "6162074_m.fxe"},
	{0x191A55929A5BE90DULL, "mrl150_hunter6071759.cl"},
	{0x191AAC23561465C5ULL, "36133166.ogg"},
	{0x191B5698BAE7C6D0ULL, "6144586_m.fxe"},
	{0x191B9987DE7B7911ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy_ent.gad"},
	{0x191C41D4659634B4ULL, "vxdr_0_deeproads_cp9_0_0.vlm"},
	{0x191C9E88AB00358BULL, "pro000_refugee_male.bnk"},
	{0x191CB2FF0A767499ULL, "778248654.ogg"},
	{0x191D41C3A064A8CAULL, "hlac_ht_brothel_n_3370_2x.dds"},
	{0x191DD79357B6487FULL, "6183812_m.fxe"},
	{0x191DFD7E56D4F26EULL, "zz_followers_banter6158151.cl"},
	{0x191E54446C5265D1ULL, "gen00fl_varric6076446.cl"},
	{0x191E641BEA516D27ULL, "6143470_m.fxe"},
	{0x191E940A287D0EE5ULL, "ave200_donnic6110630.cl"},
	{0x191E949A7F3EB651ULL, "vxlt_gallowscourtyard_4iv_0_0.vlm"},
	{0x191EA2F96FD95A9FULL, "204318944.ogg"},
	{0x191ECB322A15DB07ULL, "dre120tr_hurlock_team.utt"},
	{0x191ED8B668D56CD3ULL, "546964430.ogg"},
	{0x191EE08369333DFFULL, "follower_barks6173419.cl"},
	{0x191F9DB11274E99DULL, "hlac_lt_gallowscourtyard_65536_0z.dds"},
	{0x191FBEBEC0809863ULL, "hlac_lt_0_lowtown_a2_n_c_1_1y.dds"},
	{0x191FC61AD237BB67ULL, "6139431_m.fxe"},
	{0x191FCB8088FC4FA1ULL, "one220_yevhen6103113.cl"},
	{0x191FDC844EB3F73FULL, "hlac_lt_0_lowtown_docks_n_-131074_6y.dds"},
	{0x19201DFFF8EA6436ULL, "gen00fl_fenris6105864.cl"},
	{0x192036A48CCDA0F8ULL, "vxht_0_hightown_a3_bxn_0_0.vlm"},
	{0x19204EE43F7E1B1BULL, "6140660_m.fxe"},
	{0x19208166D5A5AEB4ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.mmh"},
	{0x1920A088AB04226BULL, "pro000_refugee_male.cnv"},
	{0x1920AC034AD84A4EULL, "6042736_m.fxe"},
	{0x19212BA5FED2292FULL, "llac_ca_caverns2prc_5802_47.dds"},
	{0x192156365FFC2A6EULL, "vxlt_backalley_g_1kj_0_0.vlm"},
	{0x1921645FF808FE1FULL, "162518431.ogg"},
	{0x1921956051781F44ULL, "vxw_sundermount_27b_0_0.vlm"},
	{0x1921AE48668A905BULL, "str_cracked01_d.dds"},
	{0x1921D36C25BC835FULL, "vxw_blightlands_13d_0_0.vlm"},
	{0x1921F2E66A547CD7ULL, "vxht_mansion2_1cb_0_0.vlm"},
	{0x192263865FC381DFULL, "6165030_m.fxe"},
	{0x19229A87DE8199BFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy_ent.ani"},
	{0x192310DBA80D0019ULL, "6080449_m.fxe"},
	{0x19236E3F63764C99ULL, "vxw_encounter2_96_0_0.vlm"},
	{0x192389EB6DF71BA3ULL, "vxlt_0_lowtown_a3_n_c_eoz_0_0.vlm"},
	{0x192389FD16BD85A9ULL, "llac_ht_mansion_n_3372_0.dds"},
	{0x19239D7A57129F03ULL, "330475993.ogg"},
	{0x1924041F8ACBD113ULL, "475285442.ogg"},
	{0x19240B40A5662C4AULL, "628388436.ogg"},
	{0x192440EFA830841AULL, "839921893.ogg"},
	{0x1924686E1021FE9DULL, "vxw_bonepitmining_a2_hw_0_0.vlm"},
	{0x19249F1119E620CFULL, "532221514.ogg"},
	{0x19254011DE6E3119ULL, "clf260_harley6089577.cl"},
	{0x1925A424FB3BBE22ULL, "686684682.ogg"},
	{0x1926317B2EA386ACULL, "follower_barks6153562.cl"},
	{0x192649E0FABFABFAULL, "vxlt_undercitydungeon_1z8_0_0.vlm"},
	{0x19268FB721635FAFULL, "172250419.ogg"},
	{0x1926EBC125632F0BULL, "dae000ip_crowd_ff6_talk11.utp"},
	{0x19278DB775B481C7ULL, "character_character_p129.psh.fxo"},
	{0x19280C2DD8D5B6F2ULL, "vxlt_gallowstemplar_n_5wo_0_0.vlm"},
	{0x192829E5C86EA117ULL, "hlac_w_woundedcoast_3_0y.dds"},
	{0x19282C767C7B1479ULL, "vxht_store_c5_0_0.vlm"},
	{0x19288D9F41D7C659ULL, "hlac_dr_dungeon_3217_9y.dds"},
	{0x1928AAA07B8C5E6AULL, "6167374_m.fxe"},
	{0x1928CA8EDA58671DULL, "lt_0_lowtown_a2_53_0_lc.mmh"},
	{0x1928D9ED21F8CD71ULL, "art\\characters\\creatures\\ragedemon\\c_ragedmon.mmh"},
	{0x1928F3ED21F8F92FULL, "art\\characters\\creatures\\ragedemon\\c_ragedmon.msh"},
	{0x1929091A4FA92777ULL, "hlac_lt_0_lowtown_1_3z.dds"},
	{0x1929136C0ED61051ULL, "6087545_m.fxe"},
	{0x19292B2DCBBB4631ULL, "dre180ip_primeval_chest_2.utp"},
	{0x1929C451E4B0D289ULL, "hlac_ht_keep_3732_15y.dds"},
	{0x1929C875ACDDDF50ULL, "6141686_m.fxe"},
	{0x1929F3127DFD8582ULL, "249129461.ogg"},
	{0x192A32E406781309ULL, "hlac_lt_0_lowtown_docks_-4_4x.dds"},
	{0x192B24845F5EC617ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.cess.gad"},
	{0x192B3F20344D5CEDULL, "hlac_lt_0_lowtown_n_131071_1y.dds"},
	{0x192BED667F1D17FFULL, "llac_ht_0_hightown_a3_-196608_2.dds"},
	{0x192C4B3DCF8701F5ULL, "lt_0_lowtown_docks_a3_235_0.msh"},
	{0x192C4D3DCF87056BULL, "lt_0_lowtown_docks_a3_235_0.mmh"},
	{0x192C593DCF871908ULL, "lt_0_lowtown_docks_a3_235_0.mao"},
	{0x192CC61830DFB17BULL, "vfx_rendertoshadowmap_p1.psh.fxo.dx11.assembly"},
	{0x192D7775489E0E6DULL, "6150713_m.fxe"},
	{0x192D7A44FE49902AULL, "94177572.ogg"},
	{0x192D8BDF4C773CF0ULL, "zz_and_debug6134110.cl"},
	{0x192DF1434AF57B5EULL, "static_static_p43.psh.fxo.dx11"},
	{0x192E261EC652EF5EULL, "vxht_0_hightown_9zt_0_0.vlm"},
	{0x192E91326B6D1A7BULL, "vxlt_undercitydungeon_22j_0_0.vlm"},
	{0x192EC755FEEDF85BULL, "vxca_cavern1_haunted_dp7_0_0.vlm"},
	{0x192F1A4437DED994ULL, "550749743.ogg"},
	{0x192F70C58F8A69C6ULL, "vxht_0_hightown_b9r_0_0.vlm"},
	{0x192FC81351F242EEULL, "vxlt_backalley_g_1pk_0_0.vlm"},
	{0x1930073D436E37C4ULL, "vxw_sundermount_28c_0_0.vlm"},
	{0x1930A975D9853E25ULL, "gen00fl_isabela6090323.cl"},
	{0x1930CEA547EB6823ULL, "mag291_torpor6027013.cl"},
	{0x19317FB15CC5EDD9ULL, "186046259.ogg"},
	{0x1931FDA9F12395B1ULL, "vxlt_0_lowtown_docks_a3_n_c1a_0_0.vlm"},
	{0x193239845F6508A4ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.cess.evt"},
	{0x19325917DA3F7AE0ULL, "vxw_bonepitdestroyed_a3_17v_0_0.vlm"},
	{0x193282A42E4EFEE5ULL, "446285846.ogg"},
	{0x193322C121D5D1E3ULL, "vxw_blightlands_rn_0_0.vlm"},
	{0x19335A7254F94FD5ULL, "nat380cr_explosive_barrel.utc"},
	{0x1933C6AA1C05688AULL, "mer210stg_smith.stg"},
	{0x193400BDBF079F44ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_ltr.ani"},
	{0x19340B075026C9C3ULL, "3386069.ogg"},
	{0x19345DCB1B480A45ULL, "vxlt_undercitydungeon_1l5_0_0.vlm"},
	{0x1934DBC8D6B5116FULL, "vxht_mansion_n_4vg_0_0.vlm"},
	{0x193563FC040B8BEFULL, "943467051.ogg"},
	{0x1935849B26E39834ULL, "vxht_chantry_2i8_0_0.vlm"},
	{0x19363DEF8AF11574ULL, "gen00fl_anders6119241.cl"},
	{0x1936460CC4F7BCE1ULL, "vxlt_0_lowtown_a2_n_c_huj_0_0.vlm"},
	{0x19367E0ACACBCF72ULL, "vxlt_gallowstemplar_n_5jh_0_0.vlm"},
	{0x1936A3BA0C21EFF7ULL, "vxlt_0_lowtown_docks_a3_cof_0_0.vlm"},
	{0x193709C8B3EA0BF9ULL, "vxlt_0_lowtown_docks_a3_n_cog_0_0.vlm"},
	{0x1937F9AC42EC056CULL, "vxw_bonepit_ld_0_0.vlm"},
	{0x1939781D7B63BA2EULL, "l_and111ar_andersshop.wbd"},
	{0x193A043912205C62ULL, "hm_carver.mor"},
	{0x193A2DCACE0B629FULL, "abi_m_force_feild.wbd"},
	{0x193A3EE55D8E2C3FULL, "vxht_playershouse_n_1s5_0_1.vlm"},
	{0x193AFDBDBF0DB8A6ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_ltr.gad"},
	{0x193B1BA762E6E6F8ULL, "hlac_lt_0_lowtown_a2_131072_5x.dds"},
	{0x193B66AF0F5B67B6ULL, "follower_banter6149199.cl"},
	{0x193BA7422EE04C87ULL, "mag291_torpor6026977.cl"},
	{0x193C067B39C6F812ULL, "hlac_lt_0_lowtown_docks_65535_5y.dds"},
	{0x193C3CA366300A47ULL, "vxht_brothel_n_54g_0_0.vlm"},
	{0x193C4E645AE7C84AULL, "dae000ip_crowd_ff3_talk6.utp"},
	{0x193C82A89AA6D0D4ULL, "hlac_ht_0_hightown_n_131071_0y.dds"},
	{0x193C83703D75E91EULL, "vxht_mansion2b_1a9_0_0.vlm"},
	{0x193D648E2141D0AAULL, "hlac_lt_gallowstemplar_f_2394_2y.dds"},
	{0x193D85C220EC6662ULL, "6089481_m.fxe"},
	{0x193D98BA02A9C5F3ULL, "383041624.ogg"},
	{0x193E28B32BFDFDD5ULL, "6002793_m.fxe"},
	{0x193EE693FF9D7C96ULL, "21164860.ogg"},
	{0x193EF39345EC5E99ULL, "429411112.ogg"},
	{0x193F51845F6FBB9DULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.cess.ani"},
	{0x193F54E3F41D7A3AULL, "6026204_m.fxe"},
	{0x193F64D741DACFE2ULL, "vxca_caverns1_e3w_0_0.vlm"},
	{0x193F7AFFEE4A2368ULL, "vxht_bartrand_n_64w_0_0.vlm"},
	{0x193FC492C6643A49ULL, "215493184.ogg"},
	{0x193FCEC2864C68C8ULL, "vxlt_0_lowtown_a2_n_c_huo_0_0.vlm"},
	{0x193FED9FC38FC72DULL, "gen00fl_merrill6129931.cl"},
	{0x1940348F54555180ULL, "966068003.ogg"},
	{0x1940CAF4CC359360ULL, "vxw_bonepitdestroyed_a3_18w_0_0.vlm"},
	{0x1941005975B35A13ULL, "art\\vfx\\combat\\vfx_f_mer_emove_app_p.anb"},
	{0x194157327A9B7BE5ULL, "724500843.ogg"},
	{0x1941C79583E87219ULL, "vxw_sundermount_2bl_0_0.vlm"},
	{0x1941F60F75FBB98BULL, "6002805_m.fxe"},
	{0x1942696AB0C710F1ULL, "t3_shk_h03.tnt"},
	{0x194279F2138C2868ULL, "var_treasure.gda"},
	{0x1942B793C85DF50AULL, "mag110_cut_abomination6037103.cl"},
	{0x1942B983A3AF438EULL, "mer241ar_caverns.ncs"},
	{0x19435D943FC18229ULL, "dre000_sibling_has_blight6036480.cl"},
	{0x19438C1ED1B5F347ULL, "246343047.ogg"},
	{0x1943C5676A2F2121ULL, "hlac_ht_chantry_2534_16y.dds"},
	{0x1944322818796D0CULL, "1028204715.ogg"},
	{0x1944B7E9B6EDD561ULL, "vxlt_0_lowtown_a2_n_c_hpi_0_0.vlm"},
	{0x1944ED911532DD25ULL, "940349460.ogg"},
	{0x194534B5203DD2E3ULL, "mag370_meredith6046838.cl"},
	{0x194583426913D7D0ULL, "llac_w_bonepitdestroyed_a3_458753_0.dds"},
	{0x1946697EE9593CA5ULL, "hlac_dr_slavers_3215_9y.dds"},
	{0x194741420A83C74EULL, "6149154_m.fxe"},
	{0x194788797F5F87BCULL, "tbe100im_anderfel_2h.uti"},
	{0x1947BB222930D18BULL, "6048515_m.fxe"},
	{0x1947C06B102A1050ULL, "rog_challenge.dds"},
	{0x1948E7E2C82C74BCULL, "vxw_encounter2_3i_0_0.vlm"},
	{0x1948EF5BC078F1E2ULL, "mcr310_sibling6103599.cl"},
	{0x19494304B2D9CA8BULL, "703769296.ogg"},
	{0x19495617CD67F8ACULL, "vxlt_undercitydungeon_140_0_0.vlm"},
	{0x19498466D5C8B96BULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.anb"},
	{0x1949A7FA63DEDDA8ULL, "6109713_f.fxe"},
	{0x1949B183A3B55567ULL, "mer241ar_caverns.lst"},
	{0x194A6AEE39488B15ULL, "895911796.ogg"},
	{0x194AB07C2C7BF431ULL, "character_character_p183.psh.fxo.dx11"},
	{0x194ACAE296AA610AULL, "vxw_bonepit_156_0_0.vlm"},
	{0x194C3D3075A37FC2ULL, "mag101_aombanter6052758.cl"},
	{0x194C9AEA112524C6ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.impa.ani"},
	{0x194CAA9616881EC1ULL, "lt_0_lowtown_n_45_0_lc.mmh"},
	{0x194CB8299347FBA9ULL, "6125612_m.fxe"},
	{0x194D6D9DD2AFE265ULL, "hlac_ht_mansion2b_818_56z.dds"},
	{0x194DBD0E31C10D53ULL, "6070219_m.fxe"},
	{0x194DF3FD541D1DDBULL, "6173925_m.fxe"},
	{0x194E1D5975BE167CULL, "art\\vfx\\combat\\vfx_f_mer_emove_app_p.mmh"},
	{0x194E55889F9339DDULL, "hla_ht_0_hightown_65536_0x.dds"},
	{0x194E7818B7474E4BULL, "dae000ip_crowd_mf8_talk12.utp"},
	{0x194E8B8B32711BBDULL, "rdr111dg_martin6053591.cl"},
	{0x194E9215DD1A07B4ULL, "gen00fl_sebastian6188606.cl"},
	{0x194EA83D718A8EE7ULL, "nat380_nathaniel_banter6067303.cl"},
	{0x194EAAE4D75D750FULL, "6095966_m.fxe"},
	{0x194ED959B5816490ULL, "1035834118.ogg"},
	{0x194F0C2AD426CFCFULL, "gen00fl_isabela6075377.cl"},
	{0x194F5190B1DCCCCEULL, "565373667.ogg"},
	{0x1950155FEBBDEBDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_drink.evt"},
	{0x1950B8268D37CAB2ULL, "473207003.ogg"},
	{0x19514184FF343B45ULL, "vxlt_undercitydungeon_1z3_0_0.vlm"},
	{0x1951E4D4AF325797ULL, "222990375.ogg"},
	{0x19527685A392FB97ULL, "738107689.ogg"},
	{0x1952F6DA2DCE1F12ULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v2.gad"},
	{0x1952FC5929C8E884ULL, "457371728.ogg"},
	{0x195334C7F85FD36EULL, "455843283.ogg"},
	{0x1953616F829A6CEBULL, "vxlt_0_lowtown_docks_a3_n_d3r_0_0.vlm"},
	{0x195397D86B928F7AULL, "hlac_ca_cavern1_haunted_24421_8y.dds"},
	{0x19542AD484E794D2ULL, "blood_conj_04.dds"},
	{0x19545AF703C5FFCBULL, "llac_lt_0_lowtown_a2_n_c_-65536_3.dds"},
	{0x1954635E1A40D76AULL, "414106615.ogg"},
	{0x19547A45715441C1ULL, "vxlt_warehouse_n_11w_0_0.vlm"},
	{0x1954D4F2AF8024BEULL, "gen00fl_anders6119526.cl"},
	{0x1955090DF19C9A9EULL, "1391347.ogg"},
	{0x19551610C3F20ADDULL, "699700394.ogg"},
	{0x1955B31EFC79DFCEULL, "6142054_m.fxe"},
	{0x1955B59A6E5B6771ULL, "vxlt_0_lowtown_a2_n_c_gcc_0_0.vlm"},
	{0x19562135C47C34F7ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01gravecairn_1.mao"},
	{0x19564F228EA11C24ULL, "82028297.ogg"},
	{0x19566DEB7B49A094ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_meda_0s.dds"},
	{0x1956892D20676388ULL, "isa120_hayder6068800.cl"},
	{0x19572C5FEBC43109ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_drink.gad"},
	{0x195750B2423CBCB8ULL, "6044996_m.fxe"},
	{0x19582457CCAA0A1AULL, "t3_cth_cm3.tnt"},
	{0x195824A34D898AA8ULL, "vxht_mansion2_6s_0_0.vlm"},
	{0x195838C431A64E0EULL, "hlac_lt_warehouse_65536_0z.dds"},
	{0x1958503F2C47105FULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_greyrubble_n.dds"},
	{0x19585452888696D8ULL, "6085296_m.fxe"},
	{0x195880D9332E720EULL, "259372253.ogg"},
	{0x19594279FF884AFFULL, "vxdr_primevaltomb_50s_0_1.vlm"},
	{0x195972066DA9EE62ULL, "vxw_bonepitmining_a2_1ob_0_0.vlm"},
	{0x1959F9DA2DD44349ULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v2.evt"},
	{0x1959FB6CAB036ED5ULL, "vxht_keep_60j_0_0.vlm"},
	{0x195A005887B122F8ULL, "plt_cod_crt_golem.plo"},
	{0x195A06AFF08440F9ULL, "324230504.ogg"},
	{0x195A280870AED716ULL, "vxlt_0_lowtown_eqe_0_0.vlm"},
	{0x195A82EA11313937ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.impa.evt"},
	{0x195AA2A62EF5C2F8ULL, "mag291_ntbanter6056618.cl"},
	{0x195AB283A3C3C99CULL, "mer241ar_caverns.are"},
	{0x195B6B69D2C8DE6FULL, "6137657_m.fxe"},
	{0x195B7A26EACD1B19ULL, "6071517_m.fxe"},
	{0x195B9813F86419A8ULL, "hlac_dr_dungeon_3214_9y.dds"},
	{0x195B9CDDA20A81F8ULL, "vxht_keep_53a_0_0.vlm"},
	{0x195BEBDAD5227AB5ULL, "6080474_m.fxe"},
	{0x195C36FA8D8B0395ULL, "1034684483.ogg"},
	{0x195C65DA4613365BULL, "6176625_m.fxe"},
	{0x195C731501E9566EULL, "321521004.ogg"},
	{0x195C93526DE99729ULL, "6173606_m.fxe"},
	{0x195CAA07137696A4ULL, "dae340ar_bone_pit.are"},
	{0x195CD0B3C933BD40ULL, "vxlt_0_lowtown_a2_n_gcd_0_0.vlm"},
	{0x195CD379C1456747ULL, "199003517.ogg"},
	{0x195D0012FB2B6B05ULL, "vxht_0_hightown_a3_bwv_0_0.vlm"},
	{0x195D86AEDCB6C787ULL, "199950583.ogg"},
	{0x195DB2F935B1AE2DULL, "vxlt_0_lowtown_n_gca_0_0.vlm"},
	{0x195DFD5FEBCA0057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_drink.ani"},
	{0x195E2A408A3F6259ULL, "vxlt_0_lowtown_a2_n_c_i0h_0_0.vlm"},
	{0x195E4002FE34C2FDULL, "6094268_m.fxe"},
	{0x195EEC007F67462CULL, "6057112_m.fxe"},
	{0x195F8E6995109B34ULL, "6173733_m.fxe"},
	{0x19605F11C99792BCULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench3_arishokaddress.ani"},
	{0x1960D43F65135FA4ULL, "llac_w_encounter2_camp_327682_0.dds"},
	{0x1960D6999EE89B91ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.dura.ani"},
	{0x196105137A906EB6ULL, "stormcloud_small.dds"},
	{0x196156093C20C4CDULL, "static_static_p82.psh.fxo.dx11"},
	{0x196183EA113759C0ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.impa.gad"},
	{0x1961D6FA341F42E7ULL, "6097560_m.fxe"},
	{0x19622BB171C4085AULL, "rdr111dg_raider_ambusher6055075.cl"},
	{0x19629AAE068813FDULL, "389974458.ogg"},
	{0x1962A312C35E43AFULL, "6106644_m.fxe"},
	{0x1963041DB8EEFE77ULL, "6179758_m.fxe"},
	{0x19641302CE4E260CULL, "681598184.ogg"},
	{0x1964172436E80DA1ULL, "435684559.ogg"},
	{0x19644D601B77E7FEULL, "vxlt_0_lowtown_f5b_0_0.vlm"},
	{0x1964A211CE43DC7DULL, "zz_dae_debug6049944.cl"},
	{0x1964FB4F9671D69EULL, "hlac_lt_0_lowtown_a2_131072_4y.dds"},
	{0x196543148F933C47ULL, "361995253.ogg"},
	{0x1965CB9CAC18A5BCULL, "vxw_encounter2_9o_0_0.vlm"},
	{0x19666E1D7B8A975CULL, "l_and111ar_andersshop.bnk"},
	{0x196688FB30F99E04ULL, "6137688_f.fxe"},
	{0x196746B9D907F267ULL, "807594147.ogg"},
	{0x19676C11C99DC7FEULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench3_arishokaddress.gad"},
	{0x196783F380D67E07ULL, "zz_dae_debug6029162.cl"},
	{0x1967E1999EEECD33ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.dura.gad"},
	{0x1967E1DA2DE057B8ULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v2.ani"},
	{0x19688C40FB67C272ULL, "zz_followers_banter6158616.cl"},
	{0x196892BF08C96307ULL, "6133374_m.fxe"},
	{0x1968EAE2DBBD177DULL, "hlac_lt_0_lowtown_n_-3_7z.dds"},
	{0x196902CAB70FBEBBULL, "ch1alpharmtnt_vl.psh.fxo"},
	{0x19696C4529ED7313ULL, "nat320_delilah6065923.cl"},
	{0x196A368A244ED49EULL, "6126694_m.fxe"},
	{0x196A98FA4F45EAF6ULL, "docks_scene_dockworkers6176471.cl"},
	{0x196ABC2848E8A4AEULL, "vxlt_shop_fp_0_0.vlm"},
	{0x196BA994E1B74BAAULL, "dae021_sister_colinda_sta6148308.cl"},
	{0x196BAB400F574BDFULL, "zz_dae_debug6029809.cl"},
	{0x196BDC88B299C8EFULL, "llac_lt_0_lowtown_docks_-2_4.dds"},
	{0x196BE21E77B36FC5ULL, "vxlt_ruinedhouse_n_hy_0_0.vlm"},
	{0x196BF9D484FBD350ULL, "blood_conj_04.mao"},
	{0x196C389FEE94077AULL, "6165620_m.fxe"},
	{0x196C7E5D6FDA8C23ULL, "vxlt_0_lowtown_docks_n_b54_0_1.vlm"},
	{0x196CFCB14767DF17ULL, "6135804_m.fxe"},
	{0x196D20FF6643C702ULL, "6172008_m.fxe"},
	{0x196D46CA3A543A8DULL, "429101073.ogg"},
	{0x196D9483345CD19CULL, "hlac_lt_0_lowtown_docks_a3_65533_1z.dds"},
	{0x196DAFC76ACEA666ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbucket.phy"},
	{0x196DD13571828258ULL, "hlac_w_sundermount_-196603_0y.dds"},
	{0x196DE30BB3785957ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v2.gad"},
	{0x196E61356BA10AABULL, "vxlt_0_lowtown_cif_0_0.vlm"},
	{0x196ED397EBE96D53ULL, "6169049_m.fxe"},
	{0x196EDE999EF4E698ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.dura.evt"},
	{0x196EEDD16D2B8086ULL, "llac_lt_undercitydungeon_-1_32.dds"},
	{0x196EF10AD920C25BULL, "lgt200ip_eau_tome_final.utp"},
	{0x196EF3098FAC6FC7ULL, "513336433.ogg"},
	{0x196F24713194C4C6ULL, "vxlt_warehouse_n_1cv_0_0.vlm"},
	{0x196F84F722276772ULL, "6097356_m.fxe"},
	{0x196FA31D3CBFA16EULL, "stormcloud.dds"},
	{0x196FF1DA13456A9BULL, "909111572.ogg"},
	{0x19707AB1FE10AE64ULL, "llac_lt_backalley_g_0_17.dds"},
	{0x1971247B2105CCB6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closedtoopen.evt"},
	{0x19714E8F27C97F4DULL, "431459274.ogg"},
	{0x19715669E4EC5271ULL, "hlac_lt_gallowstemplar_2382_0x.dds"},
	{0x19716A1CBD4FB7A2ULL, "gen00fl_anders6128382.cl"},
	{0x197176B153E5D668ULL, "187715548.ogg"},
	{0x19717BC638CC962AULL, "vxht_mansion2b_1gf_0_1.vlm"},
	{0x1971F8AFCFA94DCCULL, "6106948_m.fxe"},
	{0x19722816268DF29DULL, "mag211_marethari6021344.cl"},
	{0x19727C330786077BULL, "vxlt_0_lowtown_n_ghv_0_0.vlm"},
	{0x19729D1AB35D227BULL, "vxlt_hangedman_109_0_0.vlm"},
	{0x1972BA9BC7159846ULL, "vxdr_primevaltomb_6xd_0_0.vlm"},
	{0x1972DBBB212E14CCULL, "art\\characters\\creatures\\mubari\\animation\\com_dig.ani"},
	{0x19741B4A4FF8C56FULL, "891683011.ogg"},
	{0x19748FDDEBF30CA1ULL, "vxdr_0_deeproads_cxh_0_0.vlm"},
	{0x1974F80BB37E9BE4ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v2.evt"},
	{0x19754F8FFB77AD86ULL, "art\\vfx\\combat\\vfx_glb_ele_e_extreme_c.impa.evt"},
	{0x197558384A469421ULL, "193354884.ogg"},
	{0x1975AB0F04E1DDA1ULL, "art\\characters\\creatures\\dragon\\animation\\p.ani"},
	{0x1975C4E53BA454A9ULL, "hlac_dr_slavers_3207_0y.dds"},
	{0x19761A55A3B844ABULL, "777202385.ogg"},
	{0x1976713F5725CF6EULL, "vxlt_undercityserialkiller_25k_0_0.vlm"},
	{0x1976A6E13402CAE8ULL, "382402407.ogg"},
	{0x19771DC7416398FFULL, "rog_obscuredstealth.dds"},
	{0x19781B7B210BDC41ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closedtoopen.gad"},
	{0x197896D516C8E788ULL, "t3_nbl_c01.tnt"},
	{0x19789EA84D6C9DA5ULL, "6204803_m.fxe"},
	{0x1978B8F94D75A6C0ULL, "3747595.ogg"},
	{0x1978C7D63CFEBF34ULL, "dae000ip_ambient_sitcudlm_s.utp"},
	{0x1978D517E78594BBULL, "6126494_f.fxe"},
	{0x1978E56FAB86AE03ULL, "lt_0_lowtown_docks_a3_n_154_0.mao"},
	{0x1978E96FAB86B550ULL, "lt_0_lowtown_docks_a3_n_154_0.mmh"},
	{0x1978F76FAB86CD6AULL, "lt_0_lowtown_docks_a3_n_154_0.msh"},
	{0x19790FDFFB77533CULL, "vxlt_0_lowtown_a2_n_c_d7q_0_0.vlm"},
	{0x19794F204429455AULL, "art\\characters\\weapons\\longswords\\w_lsw_013a.phy"},
	{0x19796D5D922265FCULL, "vxw_sundermount_34f_0_0.vlm"},
	{0x19798C2C22A691F4ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_basic_2.mao"},
	{0x1979B5EF8A78FF6EULL, "dae000_bartrand6031137.cl"},
	{0x1979D9D1018FE45BULL, "llac_dr_primevaltomb_658_6.dds"},
	{0x1979E6ED0CBBA258ULL, "6093641_f.fxe"},
	{0x1979E8BB2134490EULL, "art\\characters\\creatures\\mubari\\animation\\com_dig.gad"},
	{0x197A48CEEBE5DD7DULL, "vxht_0_hightown_n_b5v_0_0.vlm"},
	{0x197AACB505965ED7ULL, "1031323101.ogg"},
	{0x197AB51DB2BBAB10ULL, "llac_lt_0_lowtown_a2_n_-131072_3.dds"},
	{0x197B00D47A0CDC8CULL, "6165240_m.fxe"},
	{0x197B92DC5839DF82ULL, "6175108_m.fxe"},
	{0x197BB10713915346ULL, "dae340ar_bone_pit.ncs"},
	{0x197C5F2B16CD2DA3ULL, "340954107.ogg"},
	{0x197C668FFB7DF3D1ULL, "art\\vfx\\combat\\vfx_glb_ele_e_extreme_c.impa.gad"},
	{0x197CB60F04E80F43ULL, "art\\characters\\creatures\\dragon\\animation\\p.gad"},
	{0x197D259BD6E8B020ULL, "vxlt_0_lowtown_a2_n_c_glo_0_0.vlm"},
	{0x197D3A8355BFD3FBULL, "hlac_lt_undercityserialkiller_-1_25z.dds"},
	{0x197DA14DB2EE91C9ULL, "886040601.ogg"},
	{0x197E4D3659A54A3FULL, "ht_0_hightown_a3_22_0.rml"},
	{0x197E554851CDD640ULL, "ser120_ghyslain6119718.cl"},
	{0x197EF068C6117858ULL, "vxw_blightlands_sr_0_0.vlm"},
	{0x197F1C7B2111FCEFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closedtoopen.ani"},
	{0x197FE3BB2138ACEDULL, "art\\characters\\creatures\\mubari\\animation\\com_dig.evt"},
	{0x19817572553B9E9BULL, "nat380cr_explosive_barrel.ncs"},
	{0x1981A9071395B13FULL, "dae340ar_bone_pit.lst"},
	{0x1981BA821EB04D07ULL, "6064845_m.fxe"},
	{0x1981D1BF4147D0EDULL, "zz_arishock_spin.cut"},
	{0x1981DBD62B31379EULL, "739699636.ogg"},
	{0x1981E882B051A6AEULL, "6175074_m.fxe"},
	{0x19820F2A614CC7A9ULL, "704423777.ogg"},
	{0x1982678FFB82617FULL, "art\\vfx\\combat\\vfx_glb_ele_e_extreme_c.impa.ani"},
	{0x1982B9AA4F2623C6ULL, "53580757.ogg"},
	{0x198301BADDE92521ULL, "vxdr_0_deeproads_cuo_0_0.vlm"},
	{0x19830BB61CE0E8C7ULL, "6115114_m.fxe"},
	{0x1983100BB38B01DDULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v2.ani"},
	{0x1983902D9FD1A89EULL, "zz_mrl_debug6086518.cl"},
	{0x1983930F04EDF228ULL, "art\\characters\\creatures\\dragon\\animation\\p.evt"},
	{0x1983D31E2C1D2DC4ULL, "hlac_ca_caverns2_5802_61z.dds"},
	{0x198424F2CC8ACB3CULL, "vxw_sundermount_2b5_0_0.vlm"},
	{0x198454435B33680DULL, "530383078.ogg"},
	{0x198468C413E424F8ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_hed_0s.dds"},
	{0x1984C6CFA280A1FBULL, "dae200_cullen6140312.cl"},
	{0x1984ED792CCFA382ULL, "hlac_lt_gallowscourtyard_-131071_1y.dds"},
	{0x1984EF7C384D535FULL, "6042218_m.fxe"},
	{0x1985090EA3537FC4ULL, "493051491.ogg"},
	{0x19852B3ABBBB4CE9ULL, "6033155_m.fxe"},
	{0x19857678ECB59407ULL, "6139574_m.fxe"},
	{0x1985EB86A3549974ULL, "6177940_m.fxe"},
	{0x1985EC2BD74B0DD3ULL, "gen00fl_isabela6015847.cl"},
	{0x1985FAD95BC37A13ULL, "80891519.ogg"},
	{0x1985FCAF8C290A2CULL, "fen210_dying_magister6108611.cl"},
	{0x198612661B9C45C6ULL, "gen00fl_fenris6103358.cl"},
	{0x19862E71DB3BB07FULL, "dae000_male_commoner1_mob6174464.cl"},
	{0x198648F92D399C41ULL, "vxht_0_hightown_a2_9ur_0_0.vlm"},
	{0x19869E99FD1FF18DULL, "435207226.ogg"},
	{0x19871EB7FC53F997ULL, "vxw_bonepit_25l_0_0.vlm"},
	{0x198746610F4FB917ULL, "851485565.ogg"},
	{0x198779C85CE45016ULL, "vxht_chantry_n_2r8_0_0.vlm"},
	{0x1987DF3A84187E7CULL, "vxw_sundermount_33g_0_0.vlm"},
	{0x198834A7BB80135EULL, "vxht_chantry_n_26l_0_0.vlm"},
	{0x198855A9D10DB485ULL, "vxht_0_hightown_a3_bds_0_0.vlm"},
	{0x1988DDD43F896D79ULL, "6101522_m.fxe"},
	{0x1988F50261111F9BULL, "6147131_m.fxe"},
	{0x19891B52389E5150ULL, "lgt210_magu_selby6142031.cl"},
	{0x19895C9D9461F60CULL, "hlac_w_encounter1_-65537_0z.dds"},
	{0x19897403F0179176ULL, "578791276.ogg"},
	{0x19897E633E7F57D1ULL, "6148190_m.fxe"},
	{0x19899708B9BE752CULL, "vxca_caverns1_ckk_0_0.vlm"},
	{0x198A12522414A1A5ULL, "vxht_bartrand_n_4k9_0_0.vlm"},
	{0x198AC677AC9E26CEULL, "173861145.ogg"},
	{0x198BCDEA3903A01AULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.cess.gad"},
	{0x198CFF8DCE001B14ULL, "vxht_0_hightown_a2n_0_0.vlm"},
	{0x198D6245B80790D8ULL, "vxw_blightlands_ts_0_0.vlm"},
	{0x198E09A0408384C6ULL, "6034657_f.fxe"},
	{0x198E6390524CCC13ULL, "252701345.ogg"},
	{0x198F0004C5CADB78ULL, "6126883_m.fxe"},
	{0x198F9857D728F589ULL, "6104147_m.fxe"},
	{0x199048252C6F8472ULL, "dae380st_nathaniel.stg"},
	{0x199054E86DC94496ULL, "6071896_f.fxe"},
	{0x1990711ADB67F636ULL, "llac_lt_0_lowtown_a2_n_c_-65538_0.dds"},
	{0x199088A296E39B06ULL, "379652901.ogg"},
	{0x19909A8A7106DAE0ULL, "6020086_m.fxe"},
	{0x19909F816A759796ULL, "t3_mag.tnt"},
	{0x1990C1D4FFAB1ED0ULL, "vxlt_warehouse_y8_0_0.vlm"},
	{0x1990C2F1852835CCULL, "732176675.ogg"},
	{0x1990E8ED1877211DULL, "w_encounter3_21_0.rml"},
	{0x19913E41ACF09437ULL, "hlac_ht_brothel_n_3368_0z.dds"},
	{0x19914A18D4D6962BULL, "hightown_male_noble_gossip6174062.cl"},
	{0x19916BDBA1005A7BULL, "6175655_m.fxe"},
	{0x1991BBF944B5B1CEULL, "vxlt_gallowstemplar_n_5xn_0_0.vlm"},
	{0x19920C3EF292B0E1ULL, "164782395.ogg"},
	{0x19924AE3AA2204BBULL, "gen00fl_sebastian6187754.cl"},
	{0x1992D0EA3909C431ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.cess.evt"},
	{0x199478A272BA6AC3ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_0x.dds"},
	{0x1994BF5C5B85FACBULL, "dre000_varric6207962.cl"},
	{0x19950E4BBD62DFA8ULL, "hlac_ca_cavern1_haunted_24423_0y.dds"},
	{0x199545472565DCB7ULL, "vxlt_gallowscourtyard_n_471_0_0.vlm"},
	{0x19957DFF1344CCCDULL, "vxlt_undercityencounter_wf_0_0.vlm"},
	{0x19959094EE4A1217ULL, "vxw_bonepit_20o_0_0.vlm"},
	{0x19962187C1455F77ULL, "6118244_f.fxe"},
	{0x1996E18D680354A6ULL, "440032638.ogg"},
	{0x199723999BF796C6ULL, "hlac_dr_0_deeproads_5845_15z.dds"},
	{0x199736BC8110E073ULL, "6149757_m.fxe"},
	{0x19973AF04DD2C885ULL, "475756968.ogg"},
	{0x19974243E5506B15ULL, "885268004.ogg"},
	{0x19982AD6A114615DULL, "vxht_bartrand_n_5q6_0_0.vlm"},
	{0x19982FC2D4AD40F4ULL, "162018628.ogg"},
	{0x1998829A4A39A8D9ULL, "gen_im_trs_slv_shd_wdn_r2.uti"},
	{0x1998842F160ABA25ULL, "vxht_bartrand_n_4p8_0_0.vlm"},
	{0x1998A06A50811591ULL, "evl_lit.mao"},
	{0x1998EF50388E24DFULL, "vxlt_0_lowtown_docks_a3_bf8_0_0.vlm"},
	{0x19994114DCD6FC34ULL, "gen00fl_isabela6109804.cl"},
	{0x199968FF6B0F4C8AULL, "611426763.ogg"},
	{0x1999A42330859BF7ULL, "vxht_playershouse_n_1lm_0_0.vlm"},
	{0x199A0955BAD4E120ULL, "vxlt_0_lowtown_a2_n_c_gbm_0_0.vlm"},
	{0x199A3067629CBB80ULL, "hlac_lt_0_lowtown_a2_n_-4_0y.dds"},
	{0x199A4BD811249B75ULL, "dae000_gamlen6061607.cl"},
	{0x199A93FF66BA9614ULL, "hlac_ht_0_hightown_n_-65536_0x.dds"},
	{0x199A9C1A4A1963D6ULL, "gen00fl_isabela6085614.cl"},
	{0x199AB209D9E8498BULL, "hlac_dr_0_deeproads_5843_7x.dds"},
	{0x199AE30A24B93317ULL, "256375966.ogg"},
	{0x199B0597739954B1ULL, "llac_ht_mansion2b_818_43.dds"},
	{0x199B0F99E0180C58ULL, "llac_ht_chantry_n_2534_53.dds"},
	{0x199BC9A55C7EDA60ULL, "vxlt_gallowscourtyard_7by_0_0.vlm"},
	{0x199C8915458BB540ULL, "vxlt_0_lowtown_docks_n_c4x_0_0.vlm"},
	{0x199C953C5E47D12CULL, "6061799_m.fxe"},
	{0x199CA6DF767F33E8ULL, "vxw_sundermount_376_0_1.vlm"},
	{0x199CB4EA4A621BD1ULL, "da2terrain_terrain_p1.psh.fxo.dx11"},
	{0x199CF662BC3F6958ULL, "vxw_sundermount_2ho_0_0.vlm"},
	{0x199D2E1908982C1AULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.impa.gad"},
	{0x199D3E8A967579B0ULL, "one200dg_solivitus6048181.cl"},
	{0x199D41D22B25021AULL, "w_encounter3_166_0.msh"},
	{0x199D4AEDC2D4DB44ULL, "follower_banter6136362.cl"},
	{0x199D4FD22B251A53ULL, "w_encounter3_166_0.mao"},
	{0x199D53D22B252160ULL, "w_encounter3_166_0.mmh"},
	{0x199D87EE9C09FF31ULL, "6164192_m.fxe"},
	{0x199DF0DD48D4450DULL, "208873988.ogg"},
	{0x199E6D7BA922BBD6ULL, "6033681_f.fxe"},
	{0x199E9E0CF0AE1C92ULL, "762020238.ogg"},
	{0x199F884708283A50ULL, "gen00fl_aveline6139248.cl"},
	{0x199FE574C1D55621ULL, "vxdr_0_deeproads_com_0_0.vlm"},
	{0x19A038DBC81FA9B1ULL, "vxht_mansion2b_34k_0_0.vlm"},
	{0x19A074C47B8E6463ULL, "6031832_m.fxe"},
	{0x19A0C8C8EAF38AF1ULL, "art\\characters\\creatures\\mubari\\animation\\cwkbl.ani"},
	{0x19A0E8EA39162A00ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.cess.ani"},
	{0x19A18A716CA39524ULL, "6139120_m.fxe"},
	{0x19A19AD48DA08176ULL, "gen00fl_sebastian6188569.cl"},
	{0x19A1AB57277868E2ULL, "vxw_bonepitmining_a2_1po_0_0.vlm"},
	{0x19A203F36C029BF7ULL, "407292324.ogg"},
	{0x19A24C396A53DAB9ULL, "859737534.ogg"},
	{0x19A25A4D72A6F6C4ULL, "6106993_m.fxe"},
	{0x19A2D2D5664EB115ULL, "5091455.ogg"},
	{0x19A2D865001E150FULL, "6174519_m.fxe"},
	{0x19A35287989DB9D5ULL, "art\\vfx\\combat\\vfx_m_sta_ele_range_c.anb"},
	{0x19A43119089E5031ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.impa.evt"},
	{0x19A49AA1AD60207FULL, "6135595_f.fxe"},
	{0x19A4B3D8666ABC17ULL, "117919230.ogg"},
	{0x19A50D2FF4EB7D6CULL, "6156307_m.fxe"},
	{0x19A56B49D8F15784ULL, "vxlt_0_lowtown_docks_n_cus_0_0.vlm"},
	{0x19A5A8997306F8F3ULL, "6142657_m.fxe"},
	{0x19A634271D44C1CFULL, "art\\vfx\\materials\\vfx_fire_ground_fb.mao"},
	{0x19A69CB3930A79DDULL, "vxht_bartrand_n_5t5_0_0.vlm"},
	{0x19A7226CCEFF7EE8ULL, "distortion.psh.fxo"},
	{0x19A73B0A5D773949ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_4x.dds"},
	{0x19A7484979D1F4F8ULL, "vxca_cavern1_haunted_io8_0_0.vlm"},
	{0x19A774F0DA1182C4ULL, "6065081_m.fxe"},
	{0x19A7D3C8EAF9BB93ULL, "art\\characters\\creatures\\mubari\\animation\\cwkbl.gad"},
	{0x19A85C8338775F06ULL, "6126943_f.fxe"},
	{0x19A86ED1507DFCC1ULL, "vxw_bonepit_89_0_0.vlm"},
	{0x19A8D7AD6AE8FE9CULL, "gen00fl_bethany6102777.cl"},
	{0x19A905BDC2985CE4ULL, "vxht_keep_n_5je_0_0.vlm"},
	{0x19A98D91553907E6ULL, "758884334.ogg"},
	{0x19A99A3ED40B976EULL, "180022663.ogg"},
	{0x19AA3B824E74F2E0ULL, "vxlt_gallowscourtyard_7gf_0_0.vlm"},
	{0x19AAEFE555D6D905ULL, "883430816.ogg"},
	{0x19AB18BC68754C68ULL, "vxw_sundermount_325_0_1.vlm"},
	{0x19AB683FAE3581D8ULL, "vxw_sundermount_2ml_0_0.vlm"},
	{0x19ABAD18D42DDB42ULL, "ts_reactive_mana.gda"},
	{0x19ABC19E16244ED7ULL, "hlac_w_bonepit_262148_0z.dds"},
	{0x19AC006B7BD010D1ULL, "mer320_hubert6089947.cl"},
	{0x19ACD71755D3BABEULL, "206433677.ogg"},
	{0x19AD0F1B8C512863ULL, "docks_fem_sailor_mob6176309.cl"},
	{0x19AD604C161CF4D9ULL, "llac_lt_0_lowtown_1_3.dds"},
	{0x19AE6466ADC36C56ULL, "6092364_m.fxe"},
	{0x19AED0C8EAFFD5F8ULL, "art\\characters\\creatures\\mubari\\animation\\cwkbl.evt"},
	{0x19AEDAA58EEFD505ULL, "152264548.ogg"},
	{0x19AF5689A8FB13C8ULL, "922445107.ogg"},
	{0x19AFD3E56440806AULL, "vxht_0_hightown_a3_5h3_0_0.vlm"},
	{0x19B0114A89690A30ULL, "6177188_m.fxe"},
	{0x19B01BB832B9F925ULL, "gen00fl_sebastian6191254.cl"},
	{0x19B0A8F740CC2BBDULL, "vxlt_gallowstemplar_f_5qv_0_0.vlm"},
	{0x19B13298509ACB7DULL, "vxdr_primevalentrance_32t_0_0.vlm"},
	{0x19B194E1CB4084C8ULL, "6095958_m.fxe"},
	{0x19B1A4A635DF5932ULL, "499219116.ogg"},
	{0x19B20801C5151444ULL, "vxw_sundermount_2eh_0_0.vlm"},
	{0x19B21B2F8FDE5A4FULL, "crt_shade.wbd"},
	{0x19B23B2ABBA38919ULL, "dae000_tomwise6167279.cl"},
	{0x19B2491908AAB600ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.impa.ani"},
	{0x19B251D87A79883DULL, "follower_banter6161112.cl"},
	{0x19B2706C9FA1C6DFULL, "qun120_arishok6051907.cl"},
	{0x19B29757B96A9771ULL, "vxdr_primevaltomb_7ud_0_1.vlm"},
	{0x19B3600D1942937DULL, "hlac_dr_0_deeproads_5849_0y.dds"},
	{0x19B374487F3A3571ULL, "llac_lt_backalley_n_0_2.dds"},
	{0x19B3AA2C8932B770ULL, "gen_im_fla_varric_01.uti"},
	{0x19B4591992183124ULL, "vxht_chantry_n_2rz_0_0.vlm"},
	{0x19B471B9BA06104CULL, "hlac_ht_keep_3732_2z.dds"},
	{0x19B4B977D86F739FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curvel.msh"},
	{0x19B4BF77D86F7E41ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curvel.mmh"},
	{0x19B4CB77D86F92E2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curvel.mao"},
	{0x19B4F7916D4E6BF5ULL, "32052106.ogg"},
	{0x19B5383F1E11EC30ULL, "dae100_bran6118351.cl"},
	{0x19B543B7B4B3FF5FULL, "lt_0_lowtown_docks_40_0.rml"},
	{0x19B5E928D677CC3EULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_003a_0n.dds"},
	{0x19B5EB190824F74DULL, "60048531.ogg"},
	{0x19B6C4B35DA48331ULL, "hlac_lt_hangedman_n_-1_26y.dds"},
	{0x19B6ED0F9EC11220ULL, "vxlt_0_lowtown_a2_n_c_gxk_0_0.vlm"},
	{0x19B8190862F23364ULL, "6147811_m.fxe"},
	{0x19B866014BD5D105ULL, "lgt160st_elf_conflict.stg"},
	{0x19B8DA9344588182ULL, "dae000_pc_home_ambients6208349.cl"},
	{0x19B8DD0E0C627BE9ULL, "372303764.ogg"},
	{0x19B8F06FABBCA56BULL, "lt_0_lowtown_docks_a3_n_154_0.phy"},
	{0x19B9DA1CA02B9A58ULL, "vxw_sundermount_2vm_0_0.vlm"},
	{0x19BA0F2533CAD8E2ULL, "6005774_f.fxe"},
	{0x19BACF59B479456FULL, "6042284-pc_choice2.cut"},
	{0x19BB1C79F266B261ULL, "6088395_m.fxe"},
	{0x19BB20A87ED7188BULL, "665814090.ogg"},
	{0x19BB84E0FA9CD00AULL, "743453679.ogg"},
	{0x19BC33083FC419B8ULL, "hlac_lt_0_lowtown_a3_n_c_-3_7z.dds"},
	{0x19BC741CDEDFF04DULL, "gen00fl_carver6102686.cl"},
	{0x19BC8A8726F5254FULL, "zz_and_debug6178681.cl"},
	{0x19BCA0852AFAD6C7ULL, "vis121_saemus6167897.cl"},
	{0x19BD26FC4A723CD7ULL, "dae000_gamlen6048514.cl"},
	{0x19BDBE1A029B7F6CULL, "6087377_m.fxe"},
	{0x19BDF89D0C827738ULL, "vrc321cr_fake_cin_guard.utc"},
	{0x19BE249F6E5FDE21ULL, "335526144.ogg"},
	{0x19BE5EB67FA9C1E7ULL, "sky_lt_gallowscourtyard.mao"},
	{0x19BEAF8EF0AFB154ULL, "340648580.ogg"},
	{0x19BEF19C44A05B34ULL, "6053911_f.fxe"},
	{0x19BFC102BB43BCE0ULL, "vxca_caverns1_hvk_0_0.vlm"},
	{0x19BFFE8A55B4C2D8ULL, "42716266.ogg"},
	{0x19BFFEF5EE079086ULL, "6127204_m.fxe"},
	{0x19C014C7C2BC5D16ULL, "556776486.ogg"},
	{0x19C021287137E6ACULL, "9092150.ogg"},
	{0x19C079DEB70B2CC4ULL, "vxw_sundermount_2ni_0_0.vlm"},
	{0x19C07AA8AC4E64DCULL, "586570968.ogg"},
	{0x19C0F1F20F6949D2ULL, "6093882_m.fxe"},
	{0x19C126E412A9B0F7ULL, "6067646_m.fxe"},
	{0x19C1D6A0658118B7ULL, "70400035.ogg"},
	{0x19C2FC7D0D574516ULL, "rdr111dg_martin6015084.cl"},
	{0x19C31D7BDA47C77CULL, "art\\vfx\\textures\\vfx_cyc_dark.dds"},
	{0x19C32230D93E29FFULL, "mag110_arianni6016914.cl"},
	{0x19C3A2C672600D24ULL, "vxca_cavern1_haunted_edi_0_0.vlm"},
	{0x19C3B015EB8EC91DULL, "6082127_f.fxe"},
	{0x19C40FC7C4381829ULL, "w_encounter3_10_0.rml"},
	{0x19C45A4CF4DB0C16ULL, "6179786_m.fxe"},
	{0x19C51C076F76F19AULL, "mag131_keran6091394.cl"},
	{0x19C580CD73D94073ULL, "vxht_0_hightown_a3_ame_0_0.vlm"},
	{0x19C58F67FF5B59C2ULL, "318624631.ogg"},
	{0x19C6DDE797611A12ULL, "471754684.ogg"},
	{0x19C739354253425AULL, "vxlt_gallowstemplar_f_663_0_0.vlm"},
	{0x19C77305989CF096ULL, "468092753.ogg"},
	{0x19C7A661B6C6108DULL, "vxht_0_hightown_c83_0_0.vlm"},
	{0x19C89E8A3F5E9CF3ULL, "134842261.ogg"},
	{0x19C902238C1D0C91ULL, "dre000_cut_bartrandbetrayal6029355.cl"},
	{0x19C971B7B7CE0149ULL, "vxht_0_hightown_n_car_0_1.vlm"},
	{0x19CA99A853CEB138ULL, "6092589_f.fxe"},
	{0x19CB538798BF0D16ULL, "art\\vfx\\combat\\vfx_m_sta_ele_range_c.mmh"},
	{0x19CB5497B3FE4243ULL, "vxw_woundedcoast_5e_0_0.vlm"},
	{0x19CB71AB09C9FE6DULL, "341431483.ogg"},
	{0x19CBFC4572AEDC5BULL, "261105859.ogg"},
	{0x19CC33A3C163B8EDULL, "vxdr_primevalentrance_14u_0_1.vlm"},
	{0x19CC4B1906831174ULL, "hlac_lt_0_lowtown_65534_6z.dds"},
	{0x19CC621BF41AB2A5ULL, "cs_pro000cs_flemeth_real.wbd"},
	{0x19CE03C7C4405809ULL, "w_encounter3_10_0.mao"},
	{0x19CE049606FD3AAEULL, "gen00fl_anders6111100.cl"},
	{0x19CE0FC7C4406CA2ULL, "w_encounter3_10_0.mmh"},
	{0x19CE11B6019A053BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrlb_0t.dds"},
	{0x19CE11C7C4407058ULL, "w_encounter3_10_0.msh"},
	{0x19CE95E35F189998ULL, "vxca_caverns2_78q_0_0.vlm"},
	{0x19CECC1491440F54ULL, "vxht_0_hightown_a3_b8b_0_0.vlm"},
	{0x19CED658976ABC8FULL, "vxlt_0_lowtown_docks_n_bfd_0_0.vlm"},
	{0x19CF4EC0DA129587ULL, "gen00fl_carver6097227.cl"},
	{0x19D0A09D7E239114ULL, "art\\levels\\races\\proxy\\props\\_textures\\breadoven_n.dds"},
	{0x19D0D5B5F2DA8668ULL, "6119198_f.fxe"},
	{0x19D0FD3F27AB4001ULL, "vxlt_0_lowtown_n_eqh_0_1.vlm"},
	{0x19D10D5F9D83DAF1ULL, "673323821.ogg"},
	{0x19D13CD376046224ULL, "vxht_chantry_n_2hx_0_0.vlm"},
	{0x19D2D8B80CB28560ULL, "vxlt_0_lowtown_docks_bdn_0_0.vlm"},
	{0x19D2E034DDF854ACULL, "vxw_encounter2_nn_0_0.vlm"},
	{0x19D3278C8D394142ULL, "6099538_m.fxe"},
	{0x19D36B96F85DFDF7ULL, "vxht_playershouse_n_1pi_0_0.vlm"},
	{0x19D3A7693C58E9A4ULL, "6170798_m.fxe"},
	{0x19D3DBE51F85E258ULL, "rdr111dg_martin6014614.cl"},
	{0x19D4DCDBA7242C62ULL, "art\\vfx\\combat\\vfx_m_bld_sac_act_c.mmh"},
	{0x19D5435E08DA02BAULL, "6143422_m.fxe"},
	{0x19D5482044779783ULL, "art\\characters\\weapons\\longswords\\w_lsw_013a.msh"},
	{0x19D556204477AFDDULL, "art\\characters\\weapons\\longswords\\w_lsw_013a.mmh"},
	{0x19D55A204477B6EEULL, "art\\characters\\weapons\\longswords\\w_lsw_013a.mao"},
	{0x19D63D0761992AEDULL, "vxlt_0_lowtown_a2_n_c_f4o_0_0.vlm"},
	{0x19D6628450B8A0F0ULL, "262347943.ogg"},
	{0x19D669B14E708231ULL, "55940284.ogg"},
	{0x19D6BB566091789AULL, "809315404.ogg"},
	{0x19D6D145F0BEAFF4ULL, "vxht_keep_6c2_0_0.vlm"},
	{0x19D6DA6E3CABA45DULL, "6155484_m.fxe"},
	{0x19D78106BA813DF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_welcome.evt"},
	{0x19D80CF636DBE336ULL, "6160807_m.fxe"},
	{0x19D8B65339E9152FULL, "6044627_m.fxe"},
	{0x19D8EA4BB0360C48ULL, "340685136.ogg"},
	{0x19D9C674A5F45AC3ULL, "vxw_woundedcoast_2d_0_0.vlm"},
	{0x19D9E8F0925F238AULL, "vxw_bonepitmining_a2_25c_0_0.vlm"},
	{0x19DA50A6FAB8277AULL, "hlac_w_sundermount_262146_0z.dds"},
	{0x19DA92CD41485D19ULL, "hlac_lt_0_lowtown_a2_n_c_131070_4y.dds"},
	{0x19DAD17EF0BB0982ULL, "llac_dr_0_deeproads_17639_0.dds"},
	{0x19DB394DDC342E72ULL, "6015973_f.fxe"},
	{0x19DBFE8E16AE753DULL, "vxlt_gallowstemplar_f_5bs_0_0.vlm"},
	{0x19DC3F9339621D97ULL, "vxlt_backalley_g_1ji_0_1.vlm"},
	{0x19DC72B7207AD506ULL, "6082018_f.fxe"},
	{0x19DC797CB374F1B4ULL, "ser270_emeric6113750.cl"},
	{0x19DCB6BDBD6B69E2ULL, "300636020.ogg"},
	{0x19DCDAA2A14EABFAULL, "dae000ip_crowd_lm2_talk11.utp"},
	{0x19DD3DF1833A27D4ULL, "vxht_0_hightown_a3_b7c_0_0.vlm"},
	{0x19DD5D989AF75DC4ULL, "vxw_sundermount_2to_0_0.vlm"},
	{0x19DD7806BA859981ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_welcome.gad"},
	{0x19DD87A5559E2F2DULL, "213434866.ogg"},
	{0x19DE1101447341E8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho02.msh"},
	{0x19DE2F0144737412ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho02.mmh"},
	{0x19DE2FE159D122ECULL, "ht_0_hightownexplosion_n_25_0_lc.mmh"},
	{0x19DE6A064437B7F5ULL, "vxlt_gallowscourtyard_n_7i6_0_0.vlm"},
	{0x19DF4974C89C1FAEULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend03_da.dds"},
	{0x19DF4EF412A49105ULL, "vxw_bonepit_1xg_0_1.vlm"},
	{0x19DFA87AEC0783D4ULL, "794104506.ogg"},
	{0x19DFAEB067FA7AA4ULL, "vxht_chantry_n_2cy_0_0.vlm"},
	{0x19DFDFBA20B002EAULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_009a_0d.dds"},
	{0x19E015C8EC56C249ULL, "art\\vfx\\combat\\vfx_str_att_back_act_c.mmh"},
	{0x19E05A4D45AD0562ULL, "6092593_m.fxe"},
	{0x19E0C077D894BABEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curvel.phy"},
	{0x19E0E9FEDFA03EECULL, "vxht_keep_n_3in_0_0.vlm"},
	{0x19E13E06D8C73D16ULL, "6177586_m.fxe"},
	{0x19E14240CFC61FF9ULL, "6033559_m.fxe"},
	{0x19E15211CFEE6D2CULL, "vxw_encounter2_io_0_0.vlm"},
	{0x19E17BB7C252C692ULL, "6095394_m.fxe"},
	{0x19E1BEDD9C57B2FCULL, "mag100_cullen6095167.cl"},
	{0x19E1C5EAB271E144ULL, "6039747_m.fxe"},
	{0x19E1DD73EA541677ULL, "vxht_playershouse_n_1wh_0_0.vlm"},
	{0x19E1ECA2DB6E1B2AULL, "vxdr_primevaltomb_6uo_0_0.vlm"},
	{0x19E204C181A1DEC5ULL, "shock_projectile_f_s.dds"},
	{0x19E248C47E67221BULL, "vxlt_0_lowtown_a2_n_e2v_0_0.vlm"},
	{0x19E25E8DC0F396EFULL, "6149898_m.fxe"},
	{0x19E2623A527ED6B8ULL, "vxdr_primevaltomb_75p_0_0.vlm"},
	{0x19E2DBC7EA1D94DCULL, "zz_spy130_darktown_res_26055327.cl"},
	{0x19E3012511FD93FFULL, "239106189.ogg"},
	{0x19E33A3F08DAEC04ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv5.ani"},
	{0x19E3D49B833E8DBBULL, "470520410.ogg"},
	{0x19E3DA63DE2048C4ULL, "vxdr_0_deeproads_cvq_0_2.vlm"},
	{0x19E3E553125BA2B7ULL, "aip_hurlockemissary_base.gda"},
	{0x19E406FDDE75C554ULL, "llac_lt_0_lowtown_docks_a3_196607_0.dds"},
	{0x19E41AAE96250273ULL, "vxht_0_hightown_cbo_0_0.vlm"},
	{0x19E41FFDAA4820B2ULL, "6121545_m.fxe"},
	{0x19E46F7F2634D89CULL, "993393282.ogg"},
	{0x19E47906BA8BBA2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_welcome.ani"},
	{0x19E4CE8B79405944ULL, "6141563_m.fxe"},
	{0x19E5B9EBB29E025DULL, "dae300_alistair_king6133599.cl"},
	{0x19E5DD8E34C8A27CULL, "yellow.dds"},
	{0x19E62D16F656C4D6ULL, "6149878_f.fxe"},
	{0x19E69B9C1497408DULL, "vis121_saemus6103170.cl"},
	{0x19E6FE68F239A506ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkback.ani"},
	{0x19E711DB27EDA571ULL, "vxht_chantry_286_0_0.vlm"},
	{0x19E746EFC744FF10ULL, "character_character_p278.psh.fxo.dx11.assembly"},
	{0x19E7DF373F004713ULL, "dae000cr_grand_cleric6182012.cl"},
	{0x19E810242A19ED9DULL, "949765096.ogg"},
	{0x19E8A7649D5B3F43ULL, "93470437.ogg"},
	{0x19E8FAA55ACB69F8ULL, "6070498_m.fxe"},
	{0x19E919650C9F32C7ULL, "gen_im_acc_blt_act3_01.uti"},
	{0x19EA373F08E10666ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv5.gad"},
	{0x19EA723A6CD0A12FULL, "seb321_leliana6180979.cl"},
	{0x19EA7BCE76BF04ABULL, "261780667.ogg"},
	{0x19EABDD4BFEF07C8ULL, "633167758.ogg"},
	{0x19EB47ABA24CA48AULL, "vxw_bonepitmining_a2_1q4_0_0.vlm"},
	{0x19EBCF758CED7644ULL, "vxw_sundermount_2yl_0_0.vlm"},
	{0x19EC1C8EF47EA9F5ULL, "6144330_m.fxe"},
	{0x19EC1E761C304D3EULL, "372532493.ogg"},
	{0x19EC2FB7A903BBF1ULL, "llac_ca_caverns2prc_5802_49.dds"},
	{0x19EC90369A7DBC40ULL, "qcr220_orsino6061815.cl"},
	{0x19ECA101BF885BF3ULL, "90366678.ogg"},
	{0x19ECE4ED18C5793DULL, "w_encounter3_21_0.mao"},
	{0x19ECE8ED18C58076ULL, "w_encounter3_21_0.mmh"},
	{0x19ECF2ED18C59174ULL, "w_encounter3_21_0.msh"},
	{0x19ED0E34D5D6CC35ULL, "one120cr_athenril.utc"},
	{0x19ED19942B574DC7ULL, "6042755_m.fxe"},
	{0x19ED6201F872215FULL, "fen310_fenris6118246.cl"},
	{0x19EDBC4F2A32C159ULL, "mag131_keran6086247.cl"},
	{0x19EDC407225C6FA4ULL, "6117519_m.fxe"},
	{0x19EDC680CF99B418ULL, "471085519.ogg"},
	{0x19EDCE128E73B3ABULL, "vxlt_hangedman_10i_0_0.vlm"},
	{0x19EE040B27CD7F1EULL, "544081651.ogg"},
	{0x19EE3054EC900E3FULL, "vxlt_0_lowtown_a3_n_c_ef5_0_0.vlm"},
	{0x19EECE47BDFE356CULL, "vxlt_0_lowtown_a3_n_c_ew9_0_0.vlm"},
	{0x19EF037F689FE04FULL, "218581987.ogg"},
	{0x19EF0A0BEC5B3EADULL, "6089499_m.fxe"},
	{0x19EF180574BAC2DFULL, "follower_banter6173342.cl"},
	{0x19EF29AA6C852DC6ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb104a_0d.dds"},
	{0x19F00B4671257242ULL, "6093738_m.fxe"},
	{0x19F0323F08E56945ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv5.evt"},
	{0x19F04F634BBF059BULL, "hlac_lt_0_lowtown_docks_a3_-196610_0x.dds"},
	{0x19F05AF309035B07ULL, "dae021_sister_lorena_sta6150059.cl"},
	{0x19F05E7FCD6433AAULL, "vxdr_primevaltomb_6xh_0_0.vlm"},
	{0x19F0832091798BFFULL, "gen00fl_isabela6088931.cl"},
	{0x19F0969495BF987AULL, "gen00fl_sebastian6179767.cl"},
	{0x19F0B98E289D9D22ULL, "6124685_m.fxe"},
	{0x19F1133F05ADF220ULL, "157347085.ogg"},
	{0x19F16D75DE02946FULL, "87771454.ogg"},
	{0x19F3352F90161D25ULL, "crt_shade.bnk"},
	{0x19F3B4FFD4AAE0F4ULL, "vxht_keep_6i0_0_0.vlm"},
	{0x19F3B6E4B88C32CBULL, "6144574_m.fxe"},
	{0x19F431A20A7FA317ULL, "359163789.ogg"},
	{0x19F4A4FB1E02B98CULL, "986369689.ogg"},
	{0x19F4E668F245BA77ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkback.evt"},
	{0x19F5090F915B449CULL, "376052991.ogg"},
	{0x19F5636947E94C3EULL, "vxht_0_hightown_n_c7n_0_0.vlm"},
	{0x19F596289358A9B9ULL, "56930648.ogg"},
	{0x19F5ACBAA5055CA0ULL, "darktown_male_elf_sta6174544.cl"},
	{0x19F63F9681811540ULL, "605454781.ogg"},
	{0x19F6728CBC5D3080ULL, "gen00fl_fenris6113924.cl"},
	{0x19F6BED7F4B2BA9DULL, "hm_mage_1.mor"},
	{0x19F700AA5118ADC2ULL, "vxht_mansion2b_2lm_0_0.vlm"},
	{0x19F81476F901DF5BULL, "702960241.ogg"},
	{0x19F825E99B18FDC6ULL, "vrc221cr_crazed_archer.utc"},
	{0x19F8E9A8D5D8A786ULL, "372122247.ogg"},
	{0x19F93C2B50476AA9ULL, "242380273.ogg"},
	{0x19F93EABAEA54747ULL, "crt_revenant.wbd"},
	{0x19F952008017F82FULL, "576145092.ogg"},
	{0x19F9A78D308C42B1ULL, "94527273.ogg"},
	{0x19F9FF8CF4361B55ULL, "6089770_m.fxe"},
	{0x19FA2BEF6D4D060FULL, "vxlt_0_lowtown_docks_n_bua_0_0.vlm"},
	{0x19FAB45B95CDFD25ULL, "vxht_bartrand_n_72y_0_0.vlm"},
	{0x19FAE768F24A2700ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkback.gad"},
	{0x19FB08BA5834C674ULL, "zz_followers_banter6159033.cl"},
	{0x19FB6B3EBEFFFF7DULL, "art\\vfx\\materials\\vfx_dot_blur_add.mao"},
	{0x19FB7D16473450F9ULL, "zz_start_dre6080521.cl"},
	{0x19FB8464383D99E1ULL, "562106715.ogg"},
	{0x19FBBEB29C408124ULL, "6133553_m.fxe"},
	{0x19FBE49C360F39BDULL, "6063820_m.fxe"},
	{0x19FBF02836AAB2F9ULL, "vxht_mansion2b_3dc_0_0.vlm"},
	{0x19FBF129582C6208ULL, "vxlt_backalley_g_1vw_0_0.vlm"},
	{0x19FC1F07D492B12BULL, "vxlt_0_lowtown_n_gm9_0_0.vlm"},
	{0x19FC24D3B239725DULL, "vxca_cavern1_haunted_icl_0_1.vlm"},
	{0x19FC7F69A90DED6FULL, "follower_banter6163264.cl"},
	{0x19FCA231DE8626BFULL, "vxlt_0_lowtown_a3_n_c_ei4_0_0.vlm"},
	{0x19FCC85AE5FD0203ULL, "w_bonepit_174_0.mao"},
	{0x19FCCC5AE5FD0950ULL, "w_bonepit_174_0.mmh"},
	{0x19FCDA5AE5FD216AULL, "w_bonepit_174_0.msh"},
	{0x19FD01DAB7447248ULL, "hlac_w_sundermount_65534_0z.dds"},
	{0x19FD05DE08A3D53BULL, "hlac_ca_cave_2194_3y.dds"},
	{0x19FD15039368AB4DULL, "dae200_saemus6091682.cl"},
	{0x19FD37B28C579D15ULL, "6155901_m.fxe"},
	{0x19FDB0769EE59DE7ULL, "w_encounter3_5_0.rml"},
	{0x19FDC9F9A95580CCULL, "gen_im_npc_arm_hum_glv_tmpb.uti"},
	{0x19FDDBDBA7472F09ULL, "art\\vfx\\combat\\vfx_m_bld_sac_act_c.anb"},
	{0x19FE165C7084F8B7ULL, "ht_mansion_n_mmap.dds"},
	{0x19FED65B60C58E2EULL, "6096041_m.fxe"},
	{0x19FEECCBB7A4343AULL, "gen_im_cns_flu_varric_04.uti"},
	{0x19FF2C7E6253531BULL, "vxlt_0_lowtown_a2_n_e8h_0_0.vlm"},
	{0x1A00356632B74338ULL, "vxlt_0_lowtown_n_gi2_0_0.vlm"},
	{0x1A00FC777CEBF5EBULL, "yellow50.dds"},
	{0x1A021B56443228E4ULL, "220048071.ogg"},
	{0x1A0228C1AA824B4BULL, "6141674_m.fxe"},
	{0x1A022C16FFC6BF63ULL, "llac_lt_0_lowtown_a2_-196609_0.dds"},
	{0x1A02469037F0B367ULL, "763152791.ogg"},
	{0x1A02B61B8912CA6FULL, "vxlt_gallowsprison_n_69w_0_0.vlm"},
	{0x1A0384FA8564EC1BULL, "w_bonepitmining_a2_168_0.msh"},
	{0x1A0396FA85650B66ULL, "w_bonepitmining_a2_168_0.mao"},
	{0x1A03A2FA85651F05ULL, "w_bonepitmining_a2_168_0.mmh"},
	{0x1A03FEA31E7CFC8CULL, "fen310_fenris6118193.cl"},
	{0x1A0422E705FF90C2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rm_basic.phy"},
	{0x1A053C8CA3913417ULL, "vxw_blightlands_11z_0_0.vlm"},
	{0x1A05698227FCF4D5ULL, "6077358_f.fxe"},
	{0x1A059E1E7F61F17AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha04.mmh"},
	{0x1A05B01E7F620FC0ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha04.msh"},
	{0x1A0678EAE09A2A9BULL, "vxca_caverns1_mines_j56_0_0.vlm"},
	{0x1A0679D00749D652ULL, "493971702.ogg"},
	{0x1A06945AAAF12707ULL, "150015803.ogg"},
	{0x1A070CD2AB622F72ULL, "48760465.ogg"},
	{0x1A072E2D6940416BULL, "ser221_gascard6149264.cl"},
	{0x1A07C179F0C08AFAULL, "941438445.ogg"},
	{0x1A07D9148DD5A3C9ULL, "82237724.ogg"},
	{0x1A083EF143B65A9DULL, "103149689.ogg"},
	{0x1A088DFA8B1731F1ULL, "249642262.ogg"},
	{0x1A08D23CAADA830BULL, "vxht_chantry_30u_0_0.vlm"},
	{0x1A08FEDE27077BFBULL, "45185004.ogg"},
	{0x1A09086363E3A2D8ULL, "289370975.ogg"},
	{0x1A0912C8EC79C27EULL, "art\\vfx\\combat\\vfx_str_att_back_act_c.anb"},
	{0x1A092BED6DF301D0ULL, "6117003_m.fxe"},
	{0x1A0974A2AC164023ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_buttress.mao"},
	{0x1A0978A2AC164770ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_buttress.mmh"},
	{0x1A0986A2AC165E8AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_buttress.msh"},
	{0x1A098FBB441BC335ULL, "conversation_dof_parameters.gda"},
	{0x1A0A0350E0A3DC2CULL, "vxht_keep_n_62c_0_1.vlm"},
	{0x1A0A47129B0441F4ULL, "fen110_fenris6067187.cl"},
	{0x1A0A87C8E692C446ULL, "6156507_m.fxe"},
	{0x1A0AFF77A5868B23ULL, "6137617_m.fxe"},
	{0x1A0B621470DF5AF2ULL, "hlac_w_blightlands_0_12z.dds"},
	{0x1A0B740F8BB86932ULL, "1001961394.ogg"},
	{0x1A0BA1A2E414613FULL, "vxht_keep_n_3ia_0_0.vlm"},
	{0x1A0BA4140DCAFED1ULL, "vxht_mansion_n_5ck_0_0.vlm"},
	{0x1A0BB99DB2252332ULL, "6075374_m.fxe"},
	{0x1A0BEDC5D0476FADULL, "6087699_m.fxe"},
	{0x1A0C3F95B582886CULL, "vxht_keep_n_3xm_0_0.vlm"},
	{0x1A0C53D6D97AF1E9ULL, "132735370.ogg"},
	{0x1A0CC7D31E10615BULL, "hlac_w_bonepit_65538_0x.dds"},
	{0x1A0D2F06F10CE037ULL, "578039821.ogg"},
	{0x1A0D3C1BF452094FULL, "cs_pro000cs_flemeth_real.bnk"},
	{0x1A0D5FBD964AA069ULL, "vxht_mansion2b_2qa_0_0.vlm"},
	{0x1A0DCFE41EA6843EULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig2_n.dds"},
	{0x1A0E3862FE24C892ULL, "860485898.ogg"},
	{0x1A0EE4F4A3F7FAFEULL, "6000680_m.fxe"},
	{0x1A0F356D00289033ULL, "zz_followers_banter6157661.cl"},
	{0x1A0F8EF40940E8ADULL, "hlac_w_encounter2_camp_196607_0z.dds"},
	{0x1A101BE56300ECBDULL, "vxlt_0_lowtown_docks_a3_czj_0_1.vlm"},
	{0x1A103D0E918F8AD9ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_1z.dds"},
	{0x1A106289743F43F1ULL, "llac_w_woundedcoast_-262139_0.dds"},
	{0x1A11250CA54E653CULL, "t3_shd_slr.tnt"},
	{0x1A11CB25D1825BDEULL, "vxlt_hangedman_n_1k2_0_0.vlm"},
	{0x1A11E3BF62668A42ULL, "785428961.ogg"},
	{0x1A1251C62CD35157ULL, "hlac_ht_0_hightown_a3_-131074_2x.dds"},
	{0x1A12A8B64C6C2462ULL, "hlac_ca_caverns1_24422_3x.dds"},
	{0x1A1465DBE1163C80ULL, "625609109.ogg"},
	{0x1A14951FA07D5F8BULL, "405031914.ogg"},
	{0x1A14C1F6112CD694ULL, "vxlt_gallowstemplar_3s0_0_0.vlm"},
	{0x1A14FC5FEE9B49B9ULL, "vxlt_gallowscourtyard_n_66n_0_0.vlm"},
	{0x1A153B076E4D84A0ULL, "tbe100st_crafted_a2.utm"},
	{0x1A15B0769EFA2F04ULL, "w_encounter3_5_0.mmh"},
	{0x1A15B4769EFA3657ULL, "w_encounter3_5_0.mao"},
	{0x1A15B9F9B446F3D4ULL, "6161590_m.fxe"},
	{0x1A15C6769EFA54A6ULL, "w_encounter3_5_0.msh"},
	{0x1A15DF77A0CD9CCDULL, "gen00fl_isabela6079793.cl"},
	{0x1A1625305C88CD0DULL, "vxca_caverns2_71z_0_1.vlm"},
	{0x1A162C179002ECFBULL, "art\\characters\\creatures\\mubari\\animation\\rdy_crnf.ani"},
	{0x1A1646BBD3832357ULL, "lt_0_lowtown_3_0.rml"},
	{0x1A16A4BDB9C10EC9ULL, "6036484_m.fxe"},
	{0x1A17CF4F479D45A0ULL, "53990350.ogg"},
	{0x1A17FA1DDE0B8F28ULL, "1037515653.ogg"},
	{0x1A181A46B1E64B49ULL, "w_bonepitdestroyed_a3_6_0.rml"},
	{0x1A1899B1A374DCAFULL, "vxht_0_hightown_a3_c3a_0_0.vlm"},
	{0x1A18A04712305ADEULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.impa.ani"},
	{0x1A18C2A73610B4F2ULL, "6140553_m.fxe"},
	{0x1A18CF57CA8A9DA4ULL, "vxlt_0_lowtown_a2_n_c_h1w_0_0.vlm"},
	{0x1A18D18BB4DD3D09ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit.mao"},
	{0x1A18D4E33C189308ULL, "vxlt_backalley_g_1xu_0_0.vlm"},
	{0x1A18DD8BB4DD51A2ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit.mmh"},
	{0x1A18DF8BB4DD5558ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit.msh"},
	{0x1A1976E6F9A102E0ULL, "vxw_bonepitdestroyed_a3_1gf_0_0.vlm"},
	{0x1A19CFE4D30E82DAULL, "325592648.ogg"},
	{0x1A1A8ABC142B6C47ULL, "6132864_f.fxe"},
	{0x1A1A9E282ABCA6F4ULL, "hlac_ht_bartrand_n_3371_1z.dds"},
	{0x1A1B0F6FE0A0EF68ULL, "ttx_roof_int_c_1.mmh"},
	{0x1A1B2D6FE0A12192ULL, "ttx_roof_int_c_1.msh"},
	{0x1A1B34AD23527C6BULL, "vxht_mansion2b_316_0_0.vlm"},
	{0x1A1B3773EE0C3BDBULL, "lt_0_lowtown_docks_a3_n_166_0.mmh"},
	{0x1A1B4373EE0C5078ULL, "lt_0_lowtown_docks_a3_n_166_0.mao"},
	{0x1A1B505384B84A6DULL, "ch1armtnt_s1y.psh.fxo"},
	{0x1A1B5573EE0C6EC5ULL, "lt_0_lowtown_docks_a3_n_166_0.msh"},
	{0x1A1B60AE087B3B44ULL, "lgt210_eau_senestra6156307.cl"},
	{0x1A1B809F678C376AULL, "343702324.ogg"},
	{0x1A1BE26EDCA2072BULL, "vxht_mansion_n_58l_0_0.vlm"},
	{0x1A1BF04F7F8F4DA5ULL, "671131826.ogg"},
	{0x1A1C0E7CA856119DULL, "331587545.ogg"},
	{0x1A1C1ADC6AD6403AULL, "vxlt_gallowscourtyard_n_6ru_0_0.vlm"},
	{0x1A1C88ED47871D39ULL, "tv_wispy_smoke.mao"},
	{0x1A1D192016A37438ULL, "vxlt_0_lowtown_n_gc0_0_0.vlm"},
	{0x1A1D29092C15356EULL, "6150336_m.fxe"},
	{0x1A1D64294A19E4CFULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_1x.dds"},
	{0x1A1DCA5C8F641AC8ULL, "vxlt_hangedman_n_2eq_0_0.vlm"},
	{0x1A1E5712877DBBC9ULL, "vxlt_0_lowtown_a2_n_h25_0_0.vlm"},
	{0x1A1E73D266E04C39ULL, "ht_brothel_n_4152.rml"},
	{0x1A1E7B8F611B5354ULL, "840525837.ogg"},
	{0x1A1E9107ED01DC29ULL, "6217375_m.fxe"},
	{0x1A1EC15ED7FD0707ULL, "6120012_f.fxe"},
	{0x1A1FF716591B10D7ULL, "661509859.ogg"},
	{0x1A208EC3A82B93C6ULL, "llac_lt_backalley_g_0_15.dds"},
	{0x1A20E45E54809985ULL, "6069945_m.fxe"},
	{0x1A20F81714961EB6ULL, "hlac_dr_slavers_3214_6x.dds"},
	{0x1A220E46B1EE8B29ULL, "w_bonepitdestroyed_a3_6_0.mao"},
	{0x1A221A46B1EE9FC2ULL, "w_bonepitdestroyed_a3_6_0.mmh"},
	{0x1A221C46B1EEA378ULL, "w_bonepitdestroyed_a3_6_0.msh"},
	{0x1A2283F0DC8F01AEULL, "vxlt_0_lowtown_a3_n_c_e24_0_0.vlm"},
	{0x1A22A701D4D9FF20ULL, "6175960_m.fxe"},
	{0x1A2378234FDB7EF8ULL, "561051978.ogg"},
	{0x1A23809B08C89059ULL, "371548438.ogg"},
	{0x1A2381E096B2696AULL, "vxw_bonepitdestroyed_a3_lr_0_0.vlm"},
	{0x1A243417900F37F2ULL, "art\\characters\\creatures\\mubari\\animation\\rdy_crnf.evt"},
	{0x1A24F0F52329EF11ULL, "one200_gold_followers6120131.cl"},
	{0x1A251DBE1D29CB05ULL, "vxca_caverns2prc_9l3_0_0.vlm"},
	{0x1A2558E3AC2D162BULL, "479572034.ogg"},
	{0x1A25B847123B0DAFULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.impa.evt"},
	{0x1A25C1E9F735C688ULL, "w_bonepitdestroyed_a3_253_0.msh"},
	{0x1A25C6B168CFE7E0ULL, "vxdr_primevaltomb_8nz_0_0.vlm"},
	{0x1A25D3E9F735E519ULL, "w_bonepitdestroyed_a3_253_0.mao"},
	{0x1A25DADF11FC7BB6ULL, "zz_vault_debug6177758.cl"},
	{0x1A25DFE9F735F9B2ULL, "w_bonepitdestroyed_a3_253_0.mmh"},
	{0x1A2621A6774D313EULL, "isa311_isabela6116378.cl"},
	{0x1A262BB4EC4C7440ULL, "gen00fl_fenris6127040.cl"},
	{0x1A262C6887F98194ULL, "158755690.ogg"},
	{0x1A2648FAFE11E0F4ULL, "334364467.ogg"},
	{0x1A26C388163FE304ULL, "vxlt_backalley_n_1wl_0_1.vlm"},
	{0x1A26EA6E94761594ULL, "mcr310_argument6075091.cl"},
	{0x1A26EF8640F0FE3AULL, "and331st_romance.stg"},
	{0x1A27168497042A79ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_1y.dds"},
	{0x1A2778D5C5BD906CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic_1.mao"},
	{0x1A277EEB73344669ULL, "ave200_donnic6111808.cl"},
	{0x1A2820F4F3D9E359ULL, "gen00fl_isabela6096839.cl"},
	{0x1A2896F1CF31E4B4ULL, "826052775.ogg"},
	{0x1A28B79544B588AEULL, "vxca_cavern1_haunted_iix_0_0.vlm"},
	{0x1A28FDE7061E6E76ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rm_basic.mao"},
	{0x1A2909E7061E8215ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rm_basic.mmh"},
	{0x1A290BE7061E858BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rm_basic.msh"},
	{0x1A293283B9BA2493ULL, "508882776.ogg"},
	{0x1A2983E2D6587C3AULL, "xx_stone.dds"},
	{0x1A298BC692912343ULL, "vxca_caverns1_e8c_0_0.vlm"},
	{0x1A29C9A7032B4A5EULL, "llac_ca_caverns2_5802_49.dds"},
	{0x1A2A3767E51285DEULL, "6034694_m.fxe"},
	{0x1A2A53179013D785ULL, "art\\characters\\creatures\\mubari\\animation\\rdy_crnf.gad"},
	{0x1A2ABA1DCDBF496BULL, "vxlt_gallowstemplar_6s6_0_0.vlm"},
	{0x1A2ABD1E7F814185ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha04.phy"},
	{0x1A2AE96E8F8871DBULL, "6098266_m.fxe"},
	{0x1A2B0BA1EAD2919FULL, "710545317.ogg"},
	{0x1A2B8AB8BE287834ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_ash.mao"},
	{0x1A2B95B1AE3C11EDULL, "qun221_viscount6084695.cl"},
	{0x1A2BC9E733EB2C8BULL, "gen00fl_sebastian6180009.cl"},
	{0x1A2BE37A8B627F0FULL, "vxht_bartrand_n_349_0_0.vlm"},
	{0x1A2C1B24009ECD3CULL, "316830760.ogg"},
	{0x1A2C812C94EFDDA5ULL, "396189825.ogg"},
	{0x1A2CB94712412E58ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.impa.gad"},
	{0x1A2DEA28CF806ACEULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transitiontostage2.evt"},
	{0x1A2E46E26158FF31ULL, "vrc221_bartrand6044996.cl"},
	{0x1A2EA08829048D28ULL, "vxht_mansion2_3eu_0_0.vlm"},
	{0x1A2F47D04C24DA2AULL, "follower_banter6172294.cl"},
	{0x1A2F5382100AF188ULL, "qun120_arishok6051898.cl"},
	{0x1A2FC07C0E9B9AB4ULL, "one200_gold_followers6101137.cl"},
	{0x1A30314EE7FA2C34ULL, "993118403.ogg"},
	{0x1A30957867E22F37ULL, "186284442.ogg"},
	{0x1A30ABBCF5BDCFF7ULL, "vxlt_undercityserialkiller_191_0_0.vlm"},
	{0x1A30EC3B5035CB2EULL, "hlac_ca_caverns1_24425_4x.dds"},
	{0x1A31089587AD4F01ULL, "519850223.ogg"},
	{0x1A31244EB0D49F9BULL, "1827261.ogg"},
	{0x1A315C6344E4477EULL, "383036003.ogg"},
	{0x1A31689193A70B40ULL, "6099411_m.fxe"},
	{0x1A3179D7FC222626ULL, "ico_mm_mappin.dds"},
	{0x1A31C2C6E69B6194ULL, "299829401.ogg"},
	{0x1A322D8E3AC57724ULL, "6165165_m.fxe"},
	{0x1A325A7668A095AEULL, "gui_guipixelshader_p2.psh.fxo"},
	{0x1A3297C26AF91564ULL, "sound_projects_to_load.gda"},
	{0x1A32DEA00FF45A53ULL, "1046409216.ogg"},
	{0x1A338725EA3143D2ULL, "6080213_m.fxe"},
	{0x1A33DCB0AAAE427AULL, "324990302.ogg"},
	{0x1A33E0FA13D22D94ULL, "lt_0_lowtown_a3_n_c_0_0.phy"},
	{0x1A33EBF2B61D2FFEULL, "6037336_m.fxe"},
	{0x1A34381F25C62EA6ULL, "vxlt_0_lowtown_ep3_0_1.vlm"},
	{0x1A3448CFCF75F178ULL, "vxht_keep_5xp_0_0.vlm"},
	{0x1A347D8C59F3EEC1ULL, "hlac_w_encounter3_-262141_0y.dds"},
	{0x1A348A29C9368F12ULL, "mcr370_meredith6083624.cl"},
	{0x1A34AC64A0EBFC50ULL, "vxht_keep_n_6l0_0_0.vlm"},
	{0x1A34E128CF867A79ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transitiontostage2.gad"},
	{0x1A357BE11B2FD368ULL, "6153757_m.fxe"},
	{0x1A359F34BD55137AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_009a_0t.dds"},
	{0x1A361B7AB916B5C9ULL, "follower_banter6161093.cl"},
	{0x1A36F9AAE3AD4851ULL, "vxht_mansion_n_5rh_0_0.vlm"},
	{0x1A370906DE96B4CCULL, "lt_0_lowtown_a2_40_0.rml"},
	{0x1A371FA5C5B5CE5BULL, "152533367.ogg"},
	{0x1A3765A2AA830667ULL, "94112745.ogg"},
	{0x1A3771F4380C1CF4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinscol.phy"},
	{0x1A37965E5692126AULL, "llac_ht_0_hightown_65537_1.dds"},
	{0x1A37C5320649D391ULL, "hlac_lt_blackemporium_1055_16x.dds"},
	{0x1A37F5C2AA6D30E0ULL, "vxlt_0_lowtown_docks_b9i_0_0.vlm"},
	{0x1A384417BDC7A52CULL, "137089273.ogg"},
	{0x1A38509BA5F734ACULL, "vxw_blightlands_7j_0_0.vlm"},
	{0x1A3866F1BD3E07E6ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_mindblast.ani"},
	{0x1A38A8B3714D2FACULL, "lla_ht_0_hightown_a3_-196608_0.dds"},
	{0x1A390EC2341D2AC3ULL, "293472512.ogg"},
	{0x1A3948ABAEDB3C7DULL, "crt_revenant.bnk"},
	{0x1A397A742A08C1CFULL, "6142895_m.fxe"},
	{0x1A39B44D32C4A0C6ULL, "art\\vfx\\combat\\vfx_sg_blast_act_c.impa.ani"},
	{0x1A3A1F9371D36532ULL, "ball_blue.mao"},
	{0x1A3A2E0144C1D10DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho02.phy"},
	{0x1A3A55577D58978FULL, "vxht_bartrand_n_338_0_0.vlm"},
	{0x1A3AF228CF8B02C7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transitiontostage2.ani"},
	{0x1A3B4DF0558D8EF6ULL, "gen00fl_aveline6105436.cl"},
	{0x1A3C6CB11E701032ULL, "zz_tanner.cut"},
	{0x1A3CFF1A9FC7930FULL, "vxlt_gallowstemplar_n_521_0_0.vlm"},
	{0x1A3D0032EB204857ULL, "356162198.ogg"},
	{0x1A3DBDFA13DA4691ULL, "lt_0_lowtown_a3_n_c_0_0.msh"},
	{0x1A3DCBFA13DA5EACULL, "lt_0_lowtown_a3_n_c_0_0.mao"},
	{0x1A3DD7FA13DA734FULL, "lt_0_lowtown_a3_n_c_0_0.mmh"},
	{0x1A3E38046F4F8766ULL, "389982449.ogg"},
	{0x1A3EAFA1A682A094ULL, "6037409_f.fxe"},
	{0x1A3EE5C6820941D7ULL, "gen_im_cns_tom_mag_sphealer.uti"},
	{0x1A3F06064D5B3D30ULL, "genip_door_cave.utp"},
	{0x1A3F609C63BFCA92ULL, "follower_banter6163281.cl"},
	{0x1A3F67AAC07B32AEULL, "vxlt_0_lowtown_a3_n_c_e86_0_0.vlm"},
	{0x1A3FA41ED0930B17ULL, "vxlt_0_lowtown_docks_a3_aqj_0_0.vlm"},
	{0x1A4074F017C4D1E0ULL, "vxlt_0_lowtown_d8u_0_0.vlm"},
	{0x1A409E38FD86BE51ULL, "fire_02.mao"},
	{0x1A40BBE221980546ULL, "zz_vrc_debug6064279.cl"},
	{0x1A40BD6CFCB431D0ULL, "vxlt_0_lowtown_a2_ei7_0_0.vlm"},
	{0x1A40D2F9ECD45223ULL, "162401669.ogg"},
	{0x1A412F813931CBD0ULL, "hm_warden.mor"},
	{0x1A414EF4381435F1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinscol.msh"},
	{0x1A415CF438144D0CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinscol.mao"},
	{0x1A4168F4381461AFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinscol.mmh"},
	{0x1A41FE349EBBC9F6ULL, "vxlt_darktown_5bo_0_0.vlm"},
	{0x1A42172A50864E4BULL, "plt_cod_hst_mgc_spindlweed.plo"},
	{0x1A4221A61EFF4DE0ULL, "6015958_m.fxe"},
	{0x1A42B6E04BB6BE3EULL, "6020656_f.fxe"},
	{0x1A42FF3A043B6EC4ULL, "dae011_conspiracy_mob6149575.cl"},
	{0x1A4315A8B6A1BF7AULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_eyes_frown.gad"},
	{0x1A432C352F214C57ULL, "zz_fol_debug6096556.cl"},
	{0x1A435ED0B1E646C8ULL, "6089735_m.fxe"},
	{0x1A43B1412E4DE613ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc.phy"},
	{0x1A4514F8B9854FC9ULL, "vxht_mansion2_24j_0_0.vlm"},
	{0x1A453A0D662500ACULL, "ht_brothel_n_9156_ll.mmh"},
	{0x1A459B4F28A1B9AEULL, "vxca_cavern1_haunted_icz_0_0.vlm"},
	{0x1A45EC5A74FB2847ULL, "vxw_woundedcoast_5y_0_0.vlm"},
	{0x1A464EF1BD4A1C57ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_mindblast.evt"},
	{0x1A4685F632CAF484ULL, "6051569_m.fxe"},
	{0x1A46C27897ED4D2CULL, "vxw_blightlands_8k_0_0.vlm"},
	{0x1A46D2C6A39BFE73ULL, "508150242.ogg"},
	{0x1A46F9E533FC7B6BULL, "vxca_caverns1_mines_j2c_0_0.vlm"},
	{0x1A479C4D32D0B537ULL, "art\\vfx\\combat\\vfx_sg_blast_act_c.impa.evt"},
	{0x1A480B5A2B45317CULL, "6122149_f.fxe"},
	{0x1A488945D4D2BCE9ULL, "follower_banter6149244.cl"},
	{0x1A48976174DC2B54ULL, "6087085_m.fxe"},
	{0x1A48F78E1BC53234ULL, "vxlt_gallowsprison_n_6jy_0_0.vlm"},
	{0x1A4918A8B6A62F91ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_eyes_frown.evt"},
	{0x1A493999F05870F9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffa_sa.dds"},
	{0x1A494363EF62C1F2ULL, "oil_proj_glow.mao"},
	{0x1A49454C7884F4BEULL, "gen00fl_varric6033624.cl"},
	{0x1A496FC13E14D546ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml107a_0d.dds"},
	{0x1A49ACFC884F69DDULL, "lgt000cr_hate_conspirator.utc"},
	{0x1A4A4C6A9DA3B8B8ULL, "567632010.ogg"},
	{0x1A4A50FA08F3BA81ULL, "87664299.ogg"},
	{0x1A4A7BFA85A0DFB2ULL, "w_bonepitmining_a2_168_0.phy"},
	{0x1A4A7FA2AC4DE98BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_buttress.phy"},
	{0x1A4B6EC3F1959B98ULL, "gen_im_arm_cht_nobd_t6.uti"},
	{0x1A4BB193B6F2751DULL, "285270519.ogg"},
	{0x1A4BDA59BB292138ULL, "6118016_f.fxe"},
	{0x1A4BEA6289C3E1D5ULL, "6101452_m.fxe"},
	{0x1A4D38957C9D9BCAULL, "lt_0_lowtown_docks_a3_n_34_0.rml"},
	{0x1A4D4FF1BD503CE0ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_mindblast.gad"},
	{0x1A4D7DA5F8A1739EULL, "vxlt_0_lowtown_a2_gnr_0_0.vlm"},
	{0x1A4D8F76D9AA00F7ULL, "vxlt_undercityserialkiller_133_0_0.vlm"},
	{0x1A4E1F908FC4F6DDULL, "mrl150_merrill6077752.cl"},
	{0x1A4E4C4B77933C40ULL, "6093611_m.fxe"},
	{0x1A4E9D4D32D6D5C0ULL, "art\\vfx\\combat\\vfx_sg_blast_act_c.impa.gad"},
	{0x1A4F42A24F908E2BULL, "275534849.ogg"},
	{0x1A4F5C810BBD947FULL, "zz_vrc_debug6200218.cl"},
	{0x1A4F6BE3CDCD9947ULL, "6198489_m.fxe"},
	{0x1A4FAA6378BD6041ULL, "dae150_chandan6148808.cl"},
	{0x1A500FD8DE8F4313ULL, "219811535.ogg"},
	{0x1A502A45D82F1067ULL, "t3_wdg_dr3.tnt"},
	{0x1A50C8908311C5C7ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkright.gad"},
	{0x1A50D66EAE161B31ULL, "ser120_ghyslain6118440.cl"},
	{0x1A50F02A7CE6607DULL, "6171983_m.fxe"},
	{0x1A513D593A41CB9DULL, "432081135.ogg"},
	{0x1A516374FD5CE646ULL, "gen00fl_fenris6126980.cl"},
	{0x1A52191EEC222C84ULL, "vxlt_backalley_n_1xq_0_1.vlm"},
	{0x1A5238607D9DBC33ULL, "6020777_m.fxe"},
	{0x1A5241E3B0611BA9ULL, "642453557.ogg"},
	{0x1A52926CCCC9838DULL, "6076707_m.fxe"},
	{0x1A52B80373C5A8E3ULL, "vxlt_warehouse_n_t1_0_0.vlm"},
	{0x1A52F8C54F4F72AEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofb.mmh"},
	{0x1A5304C54F4F8745ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofb.mao"},
	{0x1A5312C54F4F9E9CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofb.msh"},
	{0x1A5333BA0C493592ULL, "vxht_mansion_n_34c_0_0.vlm"},
	{0x1A5339BB9E5AEFF2ULL, "6107299_m.fxe"},
	{0x1A536B914B1504F3ULL, "vxht_brothel_n_57s_0_1.vlm"},
	{0x1A5386D5AB7B6849ULL, "vxht_mansion2_21i_0_0.vlm"},
	{0x1A541A2084CA62BAULL, "mrl251ip_bosshorde_a.utp"},
	{0x1A54C03406EB45B6ULL, "6177393_m.fxe"},
	{0x1A55786330AAA409ULL, "265864284.ogg"},
	{0x1A56071C078FF1F3ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_4z.dds"},
	{0x1A5612D39579C439ULL, "vxlt_gallowstemplar_f_686_0_0.vlm"},
	{0x1A5619DF089BE235ULL, "gen_im_arm_cht_elf_comc_t2.uti"},
	{0x1A56F330E5E346E5ULL, "6103739_m.fxe"},
	{0x1A5730A8B6B29660ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_eyes_frown.ani"},
	{0x1A57BD908317D1D4ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkright.evt"},
	{0x1A587D0310BE82B4ULL, "6179986_m.fxe"},
	{0x1A588781D4C6DFEDULL, "w_blightlands_194_0.msh"},
	{0x1A589581D4C6F720ULL, "w_blightlands_194_0.mao"},
	{0x1A589981D4C6FE73ULL, "w_blightlands_194_0.mmh"},
	{0x1A58B4486D6F186BULL, "779450108.ogg"},
	{0x1A5904697708EEACULL, "t3_cth_cm5.tnt"},
	{0x1A59C7C8B708AE9CULL, "zz_and_debug6178604.cl"},
	{0x1A5A29148C42B723ULL, "hlac_lt_0_lowtown_a2_65537_7y.dds"},
	{0x1A5B1DDA2B1631D5ULL, "6161256_m.fxe"},
	{0x1A5B1E706F10C3E4ULL, "745623072.ogg"},
	{0x1A5B41222B8D611BULL, "155976617.ogg"},
	{0x1A5B69B6005C0FB5ULL, "hlac_ht_chantry_2534_19x.dds"},
	{0x1A5BDCF7A47E27BDULL, "vxlt_gallowstemplar_n_5yv_0_0.vlm"},
	{0x1A5C35A29E71E876ULL, "6174190_m.fxe"},
	{0x1A5C5C19C55C7B99ULL, "995203434.ogg"},
	{0x1A5C674568F70BC5ULL, "6088301_f.fxe"},
	{0x1A5C6DFF9483B5C2ULL, "gen00fl_sebastian6178360.cl"},
	{0x1A5CF7A2FC03F53FULL, "6175772_m.fxe"},
	{0x1A5D0206BF87C8E8ULL, "6140986_m.fxe"},
	{0x1A5E188744C17BD4ULL, "art\\levels\\ztools\\default\\generic\\9ff_bookcase.phy"},
	{0x1A5E3EBD3BE09E64ULL, "627537781.ogg"},
	{0x1A5ED3AC8036EE50ULL, "760947047.ogg"},
	{0x1A5FB9F6344CFAF9ULL, "6093597_m.fxe"},
	{0x1A6006290F8ABDE9ULL, "vxlt_blackemporium_y5_0_0.vlm"},
	{0x1A608DCA7CDCE1ACULL, "vxht_brothel_n_6ts_0_0.vlm"},
	{0x1A6099EF01CF2A26ULL, "678878895.ogg"},
	{0x1A613EC5C535FAB9ULL, "6124295_m.fxe"},
	{0x1A616D53F352FBCBULL, "and200pt_dissent.ncs"},
	{0x1A616F4B45300A4DULL, "471915417.ogg"},
	{0x1A61E34A0E3E1176ULL, "fa_dust.mao"},
	{0x1A62A0614BDDA305ULL, "vxht_keep_5xi_0_0.vlm"},
	{0x1A6456250BCD4331ULL, "mcr371_orsino6083356.cl"},
	{0x1A647B69EB59F588ULL, "6207436_m.fxe"},
	{0x1A64B59083224E0DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkright.ani"},
	{0x1A65248A12C49C1DULL, "30572480.ogg"},
	{0x1A652F3446242635ULL, "6143160_m.fxe"},
	{0x1A654EA9D407E984ULL, "hlac_w_bonepit_458756_0y.dds"},
	{0x1A65807492BB5BABULL, "mag291_ntbanter6166156.cl"},
	{0x1A6624F8891A98DEULL, "6016030_m.fxe"},
	{0x1A668E3E99C74C5CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkd_d.dds"},
	{0x1A66933949DD7D9EULL, "dae000_mother6099213.cl"},
	{0x1A669A568BE796F8ULL, "vxht_brothel_n_5uc_0_0.vlm"},
	{0x1A66DA457B0C602AULL, "vxdr_primevaltomb_5a8_0_0.vlm"},
	{0x1A66E0C3F8BFD155ULL, "hla_ht_0_hightown_a3_1_0y.dds"},
	{0x1A66EEF4B692E8EDULL, "557329332.ogg"},
	{0x1A67090A7B9156FFULL, "dre000_varric6036973.cl"},
	{0x1A67631B14582788ULL, "genip_cin_barge.utp"},
	{0x1A678FA6D2E2A195ULL, "vxca_caverns2_8of_0_0.vlm"},
	{0x1A68730732EB9019ULL, "vxlt_foundry_n_xy_0_0.vlm"},
	{0x1A68A350A474C561ULL, "6050178_m.fxe"},
	{0x1A68F58744CB47D1ULL, "art\\levels\\ztools\\default\\generic\\9ff_bookcase.msh"},
	{0x1A69038744CB5FECULL, "art\\levels\\ztools\\default\\generic\\9ff_bookcase.mao"},
	{0x1A690F8744CB738FULL, "art\\levels\\ztools\\default\\generic\\9ff_bookcase.mmh"},
	{0x1A6979B1B6153A9BULL, "hlac_lt_0_lowtown_n_196606_0y.dds"},
	{0x1A697F7ABD9AE16AULL, "hm_keran.mor"},
	{0x1A698FB71D0C4A55ULL, "6166356_m.fxe"},
	{0x1A6A3A76B48D949BULL, "vxlt_0_lowtown_docks_n_api_0_0.vlm"},
	{0x1A6A55A26673D977ULL, "w_encounter3_152_0.msh"},
	{0x1A6A67A26673F7DAULL, "w_encounter3_152_0.mao"},
	{0x1A6A6BA26673FEE9ULL, "w_encounter3_152_0.mmh"},
	{0x1A6B4B45B03A0324ULL, "6076904_m.fxe"},
	{0x1A6B570FC4C95319ULL, "invalid_pressed.cur"},
	{0x1A6B66C7BB2BE394ULL, "6108528_m.fxe"},
	{0x1A6B962A4EABEB8DULL, "llac_ht_mansion2_818_34.dds"},
	{0x1A6BBB70CB2BF776ULL, "hlac_lt_0_lowtown_docks_n_-262144_3x.dds"},
	{0x1A6C4719E9DEFF1AULL, "hlac_lt_0_lowtown_a3_n_c_-3_7x.dds"},
	{0x1A6C4F70582FC859ULL, "hlac_w_encounter2_camp_196610_0z.dds"},
	{0x1A6C64DE3082F1B3ULL, "6075472_m.fxe"},
	{0x1A6CAD8D1F1EC816ULL, "6154066_m.fxe"},
	{0x1A6D7FD3C5B67145ULL, "6171406_m.fxe"},
	{0x1A6D85F1B70A81A8ULL, "1041156145.ogg"},
	{0x1A6EA8279E81B510ULL, "vxht_mansion2b_26a_0_0.vlm"},
	{0x1A6F51BEBC3383CFULL, "dae000_mother6121045.cl"},
	{0x1A6F6AA1984E2B5DULL, "gen00pt_attributes.ncs"},
	{0x1A6FE3AB8F5ECD1FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_walltypexa.phy"},
	{0x1A7048C5B4BE62D0ULL, "gen00fl_fenris6126883.cl"},
	{0x1A70E6CC424A3487ULL, "m_dae210ar_docks.bnk"},
	{0x1A70F56EDEE021BDULL, "zz_tbe_debug.ncs"},
	{0x1A7153FBC5B08832ULL, "vxw_sundermount_27d_0_0.vlm"},
	{0x1A718D186C7417BEULL, "llac_lt_0_lowtown_a2_n_c_-196609_0.dds"},
	{0x1A71986833B887BBULL, "vxdr_0_deeproads_ept_0_0.vlm"},
	{0x1A721D400234C775ULL, "6185382_m.fxe"},
	{0x1A725CABAA15847DULL, "zz_followers_banter6158868.cl"},
	{0x1A72604FF8856A62ULL, "458870028.ogg"},
	{0x1A7294F44CC75A8CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesm_0.pwk"},
	{0x1A737A4FAD5D0777ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb03a.phy"},
	{0x1A739B378686538CULL, "781424431.ogg"},
	{0x1A73EDEB719A854BULL, "zz_fen_debug6165676.cl"},
	{0x1A7400BF3056CC00ULL, "596328380.ogg"},
	{0x1A743D9D7FC1BEEAULL, "llac_lt_0_lowtown_docks_196607_1.dds"},
	{0x1A7550DA086F447BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock1a.msh"},
	{0x1A7562DA086F62C6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock1a.mao"},
	{0x1A756EDA086F7765ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock1a.mmh"},
	{0x1A75C4CA7B861035ULL, "t3_eye_amb.tnt"},
	{0x1A75D4B0579E8279ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masb_0s.dds"},
	{0x1A761E8C0AC721BFULL, "face2_ds2y.psh.fxo"},
	{0x1A76314823CFE2A4ULL, "vxca_caverns2prc_aee_0_0.vlm"},
	{0x1A76567DE7B595F6ULL, "gen00fl_fenris6113904.cl"},
	{0x1A7658B34FF2C3A9ULL, "dae011_badpoet_sta6145187.cl"},
	{0x1A7716F804BF2332ULL, "6027526_m.fxe"},
	{0x1A780AE0753F0359ULL, "851668722.ogg"},
	{0x1A78C0B1886A58BDULL, "vxlt_gallowstemplar_n_5wt_0_0.vlm"},
	{0x1A7915FE5CA3C4A0ULL, "ave331_post_fight6138827.cl"},
	{0x1A793B71489AEC9AULL, "mag200_feynriel6030438.cl"},
	{0x1A79D1EAE51D337DULL, "vxht_mansion_n_6od_0_0.vlm"},
	{0x1A79E1F4E5DE684BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_mer_tw.ani"},
	{0x1A7A093025798A09ULL, "vxlt_undercityencounter_193_0_0.vlm"},
	{0x1A7A7ECA75A7BB98ULL, "vxca_cavern1_haunted_fmu_0_0.vlm"},
	{0x1A7AD85DE1382B96ULL, "6018578_f.fxe"},
	{0x1A7B06617D0C24E2ULL, "llac_lt_gallowscourtyard_n_-65536_3.dds"},
	{0x1A7B107F6880C886ULL, "6034235_f.fxe"},
	{0x1A7B8EC0CAB738B7ULL, "6122871_m.fxe"},
	{0x1A7BD1F854DD0992ULL, "dae000_hate_follower_react6167484.cl"},
	{0x1A7BF1B0B7AC89C5ULL, "6174706_m.fxe"},
	{0x1A7C0FBB7E51A43FULL, "dae000_flemeth6001358.cl"},
	{0x1A7C69D314C39131ULL, "llac_lt_0_lowtown_a2_n_-1_5.dds"},
	{0x1A7CA37AC4C2FCFBULL, "vxht_brothel_n_29w_0_0.vlm"},
	{0x1A7D27429F85E194ULL, "dr_0_deeproads_mmap_ph.dds"},
	{0x1A7D4CA152BBA4C3ULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v3.ani"},
	{0x1A7EE83D85CA9203ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkleft.ani"},
	{0x1A7EF1AF167C03FAULL, "116160704.ogg"},
	{0x1A7F098AF8B53933ULL, "6169342_m.fxe"},
	{0x1A7F664BB32ED2D6ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_dothalf.mao"},
	{0x1A7FB58FF18DAF11ULL, "6149618_m.fxe"},
	{0x1A7FC5D8B7A6A0B2ULL, "vxw_sundermount_28e_0_0.vlm"},
	{0x1A7FF0A48D7EA2F5ULL, "751766232.ogg"},
	{0x1A80E4AB8F6D4224ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_walltypexa.mmh"},
	{0x1A80E8AB8F6D4977ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_walltypexa.mao"},
	{0x1A80FAAB8F6D67C6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_walltypexa.msh"},
	{0x1A81EBBDC11CE3C1ULL, "gen_im_trs_slv_sta_phy_r1.uti"},
	{0x1A822C0D5C0D475AULL, "dae010_urchin_f1.utc"},
	{0x1A8337AD9DA244D8ULL, "mer320_hubert6052979.cl"},
	{0x1A8390A9175C74EAULL, "fen200ar_slaver_ambush2.ncs"},
	{0x1A83CD8928D19E57ULL, "6090392_m.fxe"},
	{0x1A83F5357263F5DDULL, "vxca_caverns1_mines_e7m_0_0.vlm"},
	{0x1A847360B6CED295ULL, "vxca_caverns2_8qd_0_0.vlm"},
	{0x1A84E69EA1835CE6ULL, "seb201_lady_harimann6177396.cl"},
	{0x1A84F4E132197A02ULL, "atl_deathscree_dxt1_dat.xml"},
	{0x1A85B110A96DB420ULL, "hla_ht_0_hightown_-65534_0z.dds"},
	{0x1A862378477E4BA5ULL, "vxlt_darktown_5a1_0_0.vlm"},
	{0x1A866C849B36EAFAULL, "vxw_woundedcoast_10v_0_0.vlm"},
	{0x1A8691FD6CCA5944ULL, "vxlt_0_lowtown_a2_n_c_eo9_0_1.vlm"},
	{0x1A86C474C1BA963DULL, "gen00fl_varric6118812.cl"},
	{0x1A87126FA8034E63ULL, "qun230cr_guild_contact6068844.cl"},
	{0x1A879B9DE9AC61F5ULL, "vxlt_warehouse_y9_0_0.vlm"},
	{0x1A87A133A1595035ULL, "atl_chargen_dxt5_00.dds"},
	{0x1A87B7F83B272524ULL, "870843754.ogg"},
	{0x1A87C9F4E5EA7CC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_mer_tw.evt"},
	{0x1A88578C1269FA47ULL, "dae000_male_commoner1_mob6174455.cl"},
	{0x1A88603050058384ULL, "hlac_lt_0_lowtown_131070_6y.dds"},
	{0x1A888F02DF2B7C64ULL, "vxca_caverns1_gj8_0_0.vlm"},
	{0x1A88C388CEDB26D5ULL, "hlac_lt_backalley_n_65537_4x.dds"},
	{0x1A891D01574B707BULL, "297369463.ogg"},
	{0x1A8923D1D6CFA902ULL, "6024112_m.fxe"},
	{0x1A89D7FBE693486EULL, "mcr371cr_carver_cut.utc"},
	{0x1A8A1CA7AD36775FULL, "312714225.ogg"},
	{0x1A8A2C79B32186FAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_mattock.mmh"},
	{0x1A8A3079B3218D31ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_mattock.mao"},
	{0x1A8A3E79B321A540ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_mattock.msh"},
	{0x1A8A4D45EF80E6B9ULL, "llac_lt_0_lowtown_docks_a3_-3_8.dds"},
	{0x1A8A7565860B9782ULL, "hf_qun_barker.mor"},
	{0x1A8AA8A91762BC7BULL, "fen200ar_slaver_ambush2.lst"},
	{0x1A8AF2EBCDAA2C41ULL, "vxca_cavern1_haunted_geu_0_0.vlm"},
	{0x1A8B153C50910E54ULL, "865403125.ogg"},
	{0x1A8B2B13AD9DAB09ULL, "350166943.ogg"},
	{0x1A8BD559C69D45E4ULL, "6064891_m.fxe"},
	{0x1A8C003D85D545FAULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkleft.evt"},
	{0x1A8C7DBFC16AFBC8ULL, "qun160_vashothlead6064643.cl"},
	{0x1A8D1908C6600ACEULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_6z.dds"},
	{0x1A8DC8F4E5EEE6D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_mer_tw.gad"},
	{0x1A8EDC9056575ADAULL, "hlac_lt_0_lowtown_a3_n_c_-2_4z.dds"},
	{0x1A8F1C073277AA14ULL, "6180000_f.fxe"},
	{0x1A8F1DE844F71006ULL, "art\\characters\\creatures\\dragon\\animation\\wk_p_rf.evt"},
	{0x1A903515F04D836EULL, "llac_w_sundermount_-327686_0.dds"},
	{0x1A914299BF84831BULL, "vxlt_warehouse_p0_0_0.vlm"},
	{0x1A91609C9E26E1A8ULL, "vxlt_warehouse_ba_0_0.vlm"},
	{0x1A9163A152CC74ADULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v3.gad"},
	{0x1A919CFC15CCB937ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_cst.gad"},
	{0x1A922DAA4C695790ULL, "vxw_woundedcoast_u_0_0.vlm"},
	{0x1A92CFDFE7C86DBFULL, "963977239.ogg"},
	{0x1A92DBE7FC691E95ULL, "vxht_0_hightown_bko_0_0.vlm"},
	{0x1A92FF3D85DB62EDULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkleft.gad"},
	{0x1A937C55B96E95A3ULL, "985783425.ogg"},
	{0x1A93A30C57CA1ABDULL, "vxdr_0_deeproads_adb_0_0.vlm"},
	{0x1A94955539746425ULL, "vxlt_darktown_5j0_0_0.vlm"},
	{0x1A94D14A4A53BC18ULL, "llac_lt_0_lowtown_n_131073_1.dds"},
	{0x1A94D28EE7017476ULL, "6170569_m.fxe"},
	{0x1A94E5CD9AF86EB8ULL, "zz_mer_debug6073702.cl"},
	{0x1A94FD0ECD12A977ULL, "vxdr_0_deeproads_ewa_0_0.vlm"},
	{0x1A94FD9B7CAE2AF3ULL, "vxlt_0_lowtown_docks_aos_0_0.vlm"},
	{0x1A9506475E1ACD98ULL, "228349262.ogg"},
	{0x1A953203E2A2490BULL, "mag100_karras6034689.cl"},
	{0x1A954E9A186A8232ULL, "385258520.ogg"},
	{0x1A956039E73CADC9ULL, "follower_banter6170194.cl"},
	{0x1A95D06EDEFEFF5AULL, "zz_tbe_debug.cnv"},
	{0x1A95E528B6ABBB42ULL, "vxht_mansion2b_27p_0_0.vlm"},
	{0x1A9634E844FD5651ULL, "art\\characters\\creatures\\dragon\\animation\\wk_p_rf.gad"},
	{0x1A9655CAB7B29E7FULL, "vxlt_gallowscourtyard_4va_0_0.vlm"},
	{0x1A96C2A110C3711DULL, "272826963.ogg"},
	{0x1A9778CBE6161EE1ULL, "40964961.ogg"},
	{0x1A97B1FC15D148C4ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_cst.evt"},
	{0x1A98B74DCAE820AAULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_brd_0d.dds"},
	{0x1A98E578271BB103ULL, "gen00fl_aveline6002631.cl"},
	{0x1A98EB36C1D683F8ULL, "zz_isa_debug6133225.cl"},
	{0x1A990DF49211DBE5ULL, "vxca_caverns2_8o6_0_0.vlm"},
	{0x1A999BC8020C8E34ULL, "6087669_m.fxe"},
	{0x1A9A484ABB345BF5ULL, "zz_fen_debug6134337.cl"},
	{0x1A9ADCB75ED3D122ULL, "519539613.ogg"},
	{0x1A9B12213B61900BULL, "vxw_bonepitmining_a2_70_0_0.vlm"},
	{0x1A9B8B7045946134ULL, "gen00fl_anders6096485.cl"},
	{0x1A9B9AF989BE1035ULL, "genip_cin_book_inquisition2.utp"},
	{0x1A9BED44DCA16A33ULL, "6163142_m.fxe"},
	{0x1A9BEEF1208098DEULL, "hlac_lt_0_lowtown_docks_n_-196610_3z.dds"},
	{0x1A9C35E84501C3FFULL, "art\\characters\\creatures\\dragon\\animation\\wk_p_rf.ani"},
	{0x1A9CDCCC426F5EBDULL, "m_dae210ar_docks.wbd"},
	{0x1A9CEDFF099AD13BULL, "vxdr_0_deeproads_eci_0_0.vlm"},
	{0x1A9CF1C36B1D3FCAULL, "qun221_petrice6082945.cl"},
	{0x1A9E1558139642BEULL, "6123381_m.fxe"},
	{0x1A9E70501A53D173ULL, "484005695.ogg"},
	{0x1A9EECD2041B9869ULL, "gen00fl_sebastian6180674.cl"},
	{0x1A9F1A5F28828FDBULL, "w_bonepit_146_0.mmh"},
	{0x1A9F265F2882A478ULL, "w_bonepit_146_0.mao"},
	{0x1A9F385F2882C2C5ULL, "w_bonepit_146_0.msh"},
	{0x1A9F4E547FC643D7ULL, "vxw_bonepitdestroyed_a3_269_0_0.vlm"},
	{0x1A9F7EFB900DA660ULL, "vis121_saemus6089666.cl"},
	{0x1A9FAEF2A8C3CCD1ULL, "6028520_m.fxe"},
	{0x1A9FBF3557F3467DULL, "hlac_lt_backalley_g_0_10z.dds"},
	{0x1AA049DD98BA738FULL, "mer200im_arlan_note.uti"},
	{0x1AA09999456CBF86ULL, "834512601.ogg"},
	{0x1AA0A32CDD726713ULL, "6018697_m.fxe"},
	{0x1AA1828313572036ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_ext_c.impa.evt"},
	{0x1AA1DEA20B747F49ULL, "mer_oil.mao"},
	{0x1AA22770C6212002ULL, "69504848.ogg"},
	{0x1AA2CEBEEAB1E1AFULL, "dae000_keep_male_servant_mo6161479.cl"},
	{0x1AA309DDC1F32595ULL, "vrc321ip_spooky6_obj_4.utp"},
	{0x1AA3261F4CFA5C1BULL, "art\\effects\\cloudshadows\\w_blightlands\\cloud1.dds"},
	{0x1AA411A39B261380ULL, "hlac_lt_0_lowtown_a2_n_-65535_4y.dds"},
	{0x1AA416AEC1634AA8ULL, "vxw_encounter3_bu_0_0.vlm"},
	{0x1AA45705A8A1D3C2ULL, "vxht_mansion2b_20q_0_0.vlm"},
	{0x1AA4C7263B2462E6ULL, "isa311_isabela6092513.cl"},
	{0x1AA55115CAEF17B3ULL, "hlac_ht_0_hightown_a2_-65537_1z.dds"},
	{0x1AA56709C27C317CULL, "gen00fl_anders6119270.cl"},
	{0x1AA5AA412EA0A488ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc.mmh"},
	{0x1AA5B6412EA0B92BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc.mao"},
	{0x1AA5C55F29086E28ULL, "plt_zz_serpt_debug.plo"},
	{0x1AA5C8412EA0D7B2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc.msh"},
	{0x1AA5C9FC15DDAEBDULL, "art\\characters\\creatures\\desiredemon\\animation\\a_cst.ani"},
	{0x1AA60FA617A4D8B6ULL, "gen00fl_fenris6127060.cl"},
	{0x1AA6209E22492CA4ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_altar_0.pwk"},
	{0x1AA63444E2D892FDULL, "llac_ca_caverns1_24427_3.dds"},
	{0x1AA665361E345664ULL, "6150275_m.fxe"},
	{0x1AA7767BD7AC27E5ULL, "vxht_0_hightown_be1_0_0.vlm"},
	{0x1AA77FD18407F465ULL, "vxca_caverns2_8j5_0_0.vlm"},
	{0x1AA79F703CFF7D69ULL, "dae000cr_grand_cleric6173685.cl"},
	{0x1AA7BE3A467EE82FULL, "vxht_keep_734_0_0.vlm"},
	{0x1AA7D1D0B3546F07ULL, "qun120_arishok6051424.cl"},
	{0x1AA8595BB8BC70A0ULL, "hlac_lt_0_lowtown_n_65536_7z.dds"},
	{0x1AA85D460302417AULL, "517525218.ogg"},
	{0x1AA865EBD75788D9ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sib2_flemeth_real.gad"},
	{0x1AA87983135D2FC1ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_ext_c.impa.gad"},
	{0x1AA879C2BD476157ULL, "hlac_lt_gallowscourtyard_65536_0x.dds"},
	{0x1AA8AABCFA3954DDULL, "hlac_lt_0_lowtown_131071_5z.dds"},
	{0x1AA8B4DC4D20B766ULL, "and111_lirene6084545.cl"},
	{0x1AA903AB4A93897BULL, "6103193_m.fxe"},
	{0x1AA945894E4E9424ULL, "6165334_m.fxe"},
	{0x1AA94BC25EDFE4A3ULL, "llac_ht_0_hightown_a3_-131069_0.dds"},
	{0x1AA94F105B7E842EULL, "vxw_bonepitdestroyed_a3_1rw_0_0.vlm"},
	{0x1AA9852E410C26FAULL, "478843172.ogg"},
	{0x1AAA62C7E87D1B57ULL, "6135930_m.fxe"},
	{0x1AAA6E1937D43FF2ULL, "6121242_m.fxe"},
	{0x1AAA91882675D357ULL, "ef_sabina.mor"},
	{0x1AAAB7D1C10113E7ULL, "hm_cgn_009.mop"},
	{0x1AAAEB6C8617B87EULL, "358338651.ogg"},
	{0x1AAB8A5FBDA9CDC9ULL, "llac_lt_0_lowtown_n_131072_5.dds"},
	{0x1AAB8F2191BAA4E8ULL, "6037594_m.fxe"},
	{0x1AACB961A2F0FB7AULL, "hlac_lt_0_lowtown_docks_n_65537_0y.dds"},
	{0x1AACCF42725EC757ULL, "6129676_m.fxe"},
	{0x1AAD351B10FEA2A8ULL, "6059002_m.fxe"},
	{0x1AAD3A000B569226ULL, "pro000_refugee_angry6052172.cl"},
	{0x1AAD92463804B9C2ULL, "hlac_w_blightlands_131071_0z.dds"},
	{0x1AAD9415D07AF58CULL, "6110690_f.fxe"},
	{0x1AADA3CDB45CEBEDULL, "gen00fl_carver6126338.cl"},
	{0x1AADC03171BC5C57ULL, "vxw_bonepitdestroyed_a3_218_0_0.vlm"},
	{0x1AADE44FA7270DA2ULL, "art\\cinematics\\additional animation\\cs_qun_crossedarms_cinematicanim.gad"},
	{0x1AAE8E7803D451A8ULL, "6143125_m.fxe"},
	{0x1AAEE8E36A412556ULL, "139378935.ogg"},
	{0x1AAF3D41947B790FULL, "mrl000_marethari6129376.cl"},
	{0x1AAF4E790D70942AULL, "hlac_w_encounter2_65540_0x.dds"},
	{0x1AAF769537113AA3ULL, "dae000_orsino6081377.cl"},
	{0x1AAF76EBD75DC427ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sib2_flemeth_real.ani"},
	{0x1AAF7A831363506FULL, "art\\vfx\\combat\\vfx_glb_phys_imp_ext_c.impa.ani"},
	{0x1AAFDB6837A1A60DULL, "gen_im_trs_qun_hvy_glv_r1.uti"},
	{0x1AB161AEC581967BULL, "465418139.ogg"},
	{0x1AB1D5A1EC3B7985ULL, "vxht_brothel_n_5cx_0_0.vlm"},
	{0x1AB234329658D8BEULL, "vxht_0_hightown_a2_cdt_0_0.vlm"},
	{0x1AB291979B7AD3F9ULL, "docks_male_elven_worker_sta6176121.cl"},
	{0x1AB2C4F2C5DF736EULL, "lla_ht_0_hightown_a2_131073_1.dds"},
	{0x1AB2E2FA60A5F836ULL, "6069289_m.fxe"},
	{0x1AB2FCA47E010920ULL, "6131189_f.fxe"},
	{0x1AB31D2D1FD87616ULL, "w_bonepit_59_0.msh"},
	{0x1AB3272D1FD88714ULL, "w_bonepit_59_0.mmh"},
	{0x1AB32B2D1FD88E27ULL, "w_bonepit_59_0.mao"},
	{0x1AB3BD7065CA6C08ULL, "6139296_m.fxe"},
	{0x1AB3D0F993D5F91DULL, "vxlt_0_lowtown_a3_n_c_f4v_0_0.vlm"},
	{0x1AB51892BBFC3884ULL, "47485334.ogg"},
	{0x1AB5AD7B6160D7BDULL, "t1_mub_ta2.tnt"},
	{0x1AB5F1AE75FE0CE5ULL, "vxca_caverns2_8q4_0_0.vlm"},
	{0x1AB5F427B6CC448FULL, "vxlt_0_lowtown_docks_n_b6t_0_0.vlm"},
	{0x1AB65F40D8FD2339ULL, "4442223.ogg"},
	{0x1AB6B9A9178813B8ULL, "fen200ar_slaver_ambush2.are"},
	{0x1AB72C22DBF5EB6DULL, "vxlt_undercityserialkiller_1is_0_0.vlm"},
	{0x1AB73512F2BC4DA3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf_l.ani"},
	{0x1AB751D2B01AE339ULL, "gen00fl_player6207707.cl"},
	{0x1AB7BB3FCF4A5BD1ULL, "vxca_caverns1_mines_ei7_0_0.vlm"},
	{0x1AB7CD52A1E4269BULL, "llac_dr_dungeon_3214_12.dds"},
	{0x1AB7E43F11C6484AULL, "gen00fl_fenris6126557.cl"},
	{0x1AB801746F6ECBFCULL, "841540392.ogg"},
	{0x1AB839A15EA80E4AULL, "666615013.ogg"},
	{0x1AB8ABEAC41F0684ULL, "gen00fl_sebastian6177918.cl"},
	{0x1AB9B663DB6F4F8EULL, "lt_gallowscourtyard_56_0.rml"},
	{0x1ABB0D20DDC16911ULL, "ht_0_hightownexplosion_n_44_0_lc.mmh"},
	{0x1ABB5B4FAD9A886CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb03a.mmh"},
	{0x1ABB614FAD9A925EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb03a.msh"},
	{0x1ABCC1D26D937F49ULL, "1032296789.ogg"},
	{0x1ABCDD17EE391B1DULL, "zz_followers_banter6157928.cl"},
	{0x1ABD47DA08ACEA12ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock1a.phy"},
	{0x1ABD57954511C901ULL, "vxca_caverns1_mines_gj8_0_0.vlm"},
	{0x1ABDAEAF9D0C1EDBULL, "vxlt_darktown_49b_0_0.vlm"},
	{0x1ABDD12C049B0358ULL, "llac_w_encounter2_0_0.dds"},
	{0x1ABDDD6C2039FF2EULL, "dae200cr_viscount6087231.cl"},
	{0x1ABE1530CEA27B41ULL, "lowtown_male_alienage1_sta6162773.cl"},
	{0x1ABE317ED24B6815ULL, "vxht_0_hightown_bzl_0_0.vlm"},
	{0x1ABE478504055ADCULL, "608058696.ogg"},
	{0x1ABE7429A9C9BD44ULL, "gen00fl_anders6125615.cl"},
	{0x1ABEC5917EB89D42ULL, "gen00fl_sebastian6195831.cl"},
	{0x1ABF8D40998B0982ULL, "dae000_paivel6134507.cl"},
	{0x1ABF9641A35767ABULL, "vxca_cave_253_0_0.vlm"},
	{0x1ABFB9A017822FF3ULL, "296926616.ogg"},
	{0x1AC052A5A2F4F2F7ULL, "vxdr_0_deeproads_efb_0_0.vlm"},
	{0x1AC06DBF931FD348ULL, "dae011_corff_sta6142380.cl"},
	{0x1AC10A011D4D8D61ULL, "6201347_m.fxe"},
	{0x1AC1397EDBDC3398ULL, "vxlt_0_lowtown_a2_n_c_e1j_0_0.vlm"},
	{0x1AC1CA1EE6EE91B4ULL, "333881859.ogg"},
	{0x1AC20F4FA737FF88ULL, "art\\cinematics\\additional animation\\cs_qun_crossedarms_cinematicanim.ani"},
	{0x1AC2181B6C8E37D5ULL, "vxca_caverns1_i9f_0_0.vlm"},
	{0x1AC2763E8AC9D699ULL, "vxlt_gallowstemplar_f_5nk_0_0.vlm"},
	{0x1AC2BDA598B3F7AAULL, "vxw_bonepit_99_0_0.vlm"},
	{0x1AC32820473D2D85ULL, "6055421_m.fxe"},
	{0x1AC351B720F0BB6BULL, "vxlt_warehouse_vj_0_0.vlm"},
	{0x1AC38D2E2AB72D42ULL, "6053566_f.fxe"},
	{0x1AC485A373796E2CULL, "146610350.ogg"},
	{0x1AC4862533813A2DULL, "180649571.ogg"},
	{0x1AC486B933993EC6ULL, "l_dae210ar_lowtown.bnk"},
	{0x1AC4A8CF3B3B0701ULL, "737733069.ogg"},
	{0x1AC554122A5C6D6EULL, "dae000ip_crowd_ff6_talk3.utp"},
	{0x1AC5AF0FBF05F450ULL, "lowtown_fem_urchin_mob6165903.cl"},
	{0x1AC5CEF506D68B95ULL, "237998043.ogg"},
	{0x1AC60C8B9DC807BDULL, "vxdr_primevaltomb_2wo_0_1.vlm"},
	{0x1AC625B32213C3A4ULL, "vxlt_0_lowtown_a2_n_c_huc_0_0.vlm"},
	{0x1AC627B5E1BFFAEEULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_tw.anb"},
	{0x1AC6793EA539A1CEULL, "vxlt_0_lowtown_docks_n_b98_0_0.vlm"},
	{0x1AC77C56801368C8ULL, "6092242_f.fxe"},
	{0x1AC7C25F29254FE7ULL, "plt_zz_serpt_debug.nss"},
	{0x1AC8032E56E62325ULL, "6055623_m.fxe"},
	{0x1AC811057EC43796ULL, "dae000_sabina6161445.cl"},
	{0x1AC916ACC58D8164ULL, "vxlt_0_lowtown_docks_a3_d14_0_0.vlm"},
	{0x1AC96AF3EB78F797ULL, "mag110_samson6167352.cl"},
	{0x1AC9F1A062102706ULL, "vxlt_0_lowtown_a3_n_c_gca_0_0.vlm"},
	{0x1ACA104D622D6A97ULL, "art\\vfx\\cinematic\\vfx_m_sta_fire_close_imp_p.impa.gad"},
	{0x1ACA4196822219C4ULL, "vxlt_foundry_n_vv_0_0.vlm"},
	{0x1ACB328A33F2AF25ULL, "193081652.ogg"},
	{0x1ACB5D392FCB4079ULL, "vxlt_gallowstemplar_n_34w_0_0.vlm"},
	{0x1ACB755AFDC00624ULL, "vxlt_hangedman_n_20r_0_0.vlm"},
	{0x1ACC06DE51252AA7ULL, "ico_gloves_light_3.dds"},
	{0x1ACC4C12F2CED08DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf_l.gad"},
	{0x1ACC6745B07763BAULL, "157988310.ogg"},
	{0x1ACC6885F63B6EB8ULL, "556543280.ogg"},
	{0x1ACC86ACEBB2915EULL, "6206002_m.fxe"},
	{0x1ACCAD72F2A19BDCULL, "character_character_p32.psh.fxo.dx11"},
	{0x1ACD19D927C3CCD7ULL, "6139879_m.fxe"},
	{0x1ACD584EE329051EULL, "6113265_m.fxe"},
	{0x1ACE017F6006155EULL, "art\\levels\\races\\darkspawn\\props\\_textures\\prp_ogreskull_n.dds"},
	{0x1ACE55FF1CC75B98ULL, "vxlt_0_lowtown_h81_0_0.vlm"},
	{0x1ACF17C4B8C6606FULL, "vxlt_0_lowtown_docks_bc4_0_0.vlm"},
	{0x1ACF56EBB2C1C9E2ULL, "pet_02.mao"},
	{0x1ACF7488A2E3990FULL, "vxca_caverns1_mines_coe_0_0.vlm"},
	{0x1ACFE4B5777E823CULL, "454970375.ogg"},
	{0x1AD027A190DCFD3BULL, "vxht_chantry_330_0_0.vlm"},
	{0x1AD0B27DD29CD6FEULL, "art\\vfx\\combat\\vfx_w_wns_front_close4_c.impa.evt"},
	{0x1AD0CA7F3D606223ULL, "mag311_nyssa6062378.cl"},
	{0x1AD1254D6233AD24ULL, "art\\vfx\\cinematic\\vfx_m_sta_fire_close_imp_p.impa.evt"},
	{0x1AD140846613189CULL, "6136231_m.fxe"},
	{0x1AD17F208631E614ULL, "358490306.ogg"},
	{0x1AD183B6EFD562AFULL, "21303158.ogg"},
	{0x1AD190E695DD54A1ULL, "696231981.ogg"},
	{0x1AD1A64EFFB4E39EULL, "6076871_f.fxe"},
	{0x1AD1DB9F45C6C5F3ULL, "qun221_viscount6070140.cl"},
	{0x1AD21495746216D9ULL, "6071162_m.fxe"},
	{0x1AD2CD2E29970225ULL, "isa121cr_hayder.utc"},
	{0x1AD2F835E6637D3BULL, "hlac_lt_undercityencounter_-1_32x.dds"},
	{0x1AD307BC93729DC6ULL, "vxht_chantry_n_361_0_0.vlm"},
	{0x1AD37285C04C21B2ULL, "vxdr_primevaltomb_2ur_0_1.vlm"},
	{0x1AD3AFE26FEEAE1FULL, "978906796.ogg"},
	{0x1AD496FDEE179C8BULL, "hlac_dr_dungeon_3216_2y.dds"},
	{0x1AD49EF9B3368CD1ULL, "vxca_caverns1_mines_ec5_0_0.vlm"},
	{0x1AD4EB1B972FBA4EULL, "vxlt_0_lowtown_docks_n_b69_0_0.vlm"},
	{0x1AD4FB08F2BD9D8AULL, "ave310_brennan6137231.cl"},
	{0x1AD5578FB4B38BADULL, "6139529_m.fxe"},
	{0x1AD563048BF8F7F7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolemb_combat.anb"},
	{0x1AD5CC05A2AA8526ULL, "hlac_lt_0_lowtown_docks_a3_n_65532_1z.dds"},
	{0x1AD65C6F2CF5D8B0ULL, "vxw_bonepitmining_a2_o7_0_0.vlm"},
	{0x1AD697EC1387BF92ULL, "vxw_encounter2_l_0_0.vlm"},
	{0x1AD6A5D258E34467ULL, "vxw_bonepitmining_a2_l9_0_0.vlm"},
	{0x1AD6B40CAD4E39A4ULL, "6172802_m.fxe"},
	{0x1AD7351A0B8A422CULL, "vxw_blightlands_na_0_0.vlm"},
	{0x1AD7827FB5F967DDULL, "dae000_worthy6166608.cl"},
	{0x1AD7C238831CF1B3ULL, "6027004_m.fxe"},
	{0x1AD7C97DD2A31C29ULL, "art\\vfx\\combat\\vfx_w_wns_front_close4_c.impa.gad"},
	{0x1AD7CE99F16BF9F8ULL, "dae000_fenarel6133326.cl"},
	{0x1AD84E6B06640C2AULL, "gen00fl_sebastian6179877.cl"},
	{0x1AD87F07CAF53658ULL, "597717479.ogg"},
	{0x1AD964B6EF883194ULL, "6082046_m.fxe"},
	{0x1AD96E356B00B14EULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v8.ani"},
	{0x1AD97F27F5008766ULL, "667104741.ogg"},
	{0x1AD9B23783C035C5ULL, "zz_dae_debug6029814.cl"},
	{0x1AD9D38D7FC47620ULL, "llac_lt_0_lowtown_a2_n_65536_1.dds"},
	{0x1AD9E42FFDF3E279ULL, "hlac_lt_gallowstemplar_2394_5x.dds"},
	{0x1AD9E57BAC77DCBFULL, "6096761_m.fxe"},
	{0x1ADA1E49CFAF6A61ULL, "6219078_m.fxe"},
	{0x1ADA323D769485E0ULL, "vxht_brothel_n_597_0_0.vlm"},
	{0x1ADA707A4D669615ULL, "hlac_lt_0_lowtown_-65535_1z.dds"},
	{0x1ADA8E708516CD82ULL, "vxw_encounter1_cr_0_0.vlm"},
	{0x1ADA97278B703482ULL, "vxlt_hangedman_n_rn_0_0.vlm"},
	{0x1ADAB0B25DBA6067ULL, "hlac_lt_0_lowtown_a2_n_c_131071_2z.dds"},
	{0x1ADAD557D529E726ULL, "6093180_f.fxe"},
	{0x1ADB4033203872ECULL, "cn_hed_glka_0t.dds"},
	{0x1ADBACA3FF36305EULL, "gen00fl_aveline6106882.cl"},
	{0x1ADC2390EE4D3581ULL, "zz_fen_debug6165490.cl"},
	{0x1ADC24485D2F0078ULL, "w_bonepitmining_a2_10_0.rml"},
	{0x1ADC4B8E749EC2C0ULL, "dre000_bodahn6031295.cl"},
	{0x1ADC5267918D1671ULL, "hlac_w_bonepit_4_0y.dds"},
	{0x1ADC9EACE2DD6DEAULL, "vxlt_0_lowtown_docks_n_cwk_0_0.vlm"},
	{0x1ADCB5E456E0FCB1ULL, "art\\vfx\\combat\\vfx_m_sph_revival_cast_c.anb"},
	{0x1ADD2B38C21DC305ULL, "vxht_brothel_n_5le_0_0.vlm"},
	{0x1ADD39DA17D8C258ULL, "mag360_alain6044648.cl"},
	{0x1ADD4CE068EB1CE8ULL, "vxht_0_hightown_a3_avw_0_0.vlm"},
	{0x1ADD89A1AABC78EFULL, "vxlt_0_lowtown_docks_bf7_0_0.vlm"},
	{0x1ADD951CB82B082CULL, "122668812.ogg"},
	{0x1ADE10EB8BC13189ULL, "hla_ht_0_hightown_a3_2_0x.dds"},
	{0x1ADE3D4D623E601DULL, "art\\vfx\\cinematic\\vfx_m_sta_fire_close_imp_p.impa.ani"},
	{0x1ADE9A7DD2A8EB77ULL, "art\\vfx\\combat\\vfx_w_wns_front_close4_c.impa.ani"},
	{0x1ADE9B831F26D712ULL, "vxlt_hangedman_n_1xg_0_0.vlm"},
	{0x1ADEC08EFAEC0F11ULL, "vxht_0_hightown_c94_0_0.vlm"},
	{0x1ADF0307A74AADE8ULL, "vxht_0_hightown_cco_0_0.vlm"},
	{0x1ADF414332C42DBDULL, "6098299_m.fxe"},
	{0x1ADFB2615809311CULL, "6131331_m.fxe"},
	{0x1ADFE06B8211EAE0ULL, "vxw_woundedcoast_no_0_0.vlm"},
	{0x1AE083C3208BC01BULL, "187360389.ogg"},
	{0x1AE10C73EE09A32EULL, "445441075.ogg"},
	{0x1AE179998568B646ULL, "vxht_chantry_n_332_0_0.vlm"},
	{0x1AE1BE1849C4E003ULL, "41361000.ogg"},
	{0x1AE2738C06C55155ULL, "dae000cr_grand_cleric6164469.cl"},
	{0x1AE27DF78D175C94ULL, "and110_refugee6111606.cl"},
	{0x1AE28F795BAA5BE8ULL, "6092515_m.fxe"},
	{0x1AE2EBE9F509BB73ULL, "gen00fl_aveline6121994.cl"},
	{0x1AE3C3BA95B4E095ULL, "vxlt_gallowstemplar_5yu_0_0.vlm"},
	{0x1AE3F9771F185D49ULL, "vxht_mansion2_2ow_0_0.vlm"},
	{0x1AE5300703AB2DD7ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_ashcover_d.dds"},
	{0x1AE594D202F427AEULL, "57656070.ogg"},
	{0x1AE59EC5960EE384ULL, "6088689_f.fxe"},
	{0x1AE5A6F6FD805AACULL, "vxw_blightlands_sf_0_0.vlm"},
	{0x1AE71F1CD11C3FF4ULL, "15263074.ogg"},
	{0x1AE73769AF922CA3ULL, "vxht_brothel_n_3hi_0_0.vlm"},
	{0x1AE73EDC44B7256FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maso_0s.dds"},
	{0x1AE766356B0CE09FULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v8.evt"},
	{0x1AE77A244EBE5799ULL, "6125457_f.fxe"},
	{0x1AE7D693E17E4A14ULL, "6085146_m.fxe"},
	{0x1AE89026C1A582E1ULL, "6212178_m.fxe"},
	{0x1AE89E7D466B3D16ULL, "6170011_m.fxe"},
	{0x1AE8B3024FAA3D57ULL, "351853275.ogg"},
	{0x1AE9004D770CE602ULL, "vxw_encounter1_hs_0_0.vlm"},
	{0x1AE9D2E2F1D46E2EULL, "gen00fl_anders6125226.cl"},
	{0x1AE9D6E456EBBF2AULL, "art\\vfx\\combat\\vfx_m_sph_revival_cast_c.mmh"},
	{0x1AE9EC717F01D165ULL, "vxlt_gallowstemplar_5v8_0_0.vlm"},
	{0x1AE9FFF1866637F5ULL, "hlac_lt_0_lowtown_a2_n_2_1x.dds"},
	{0x1AEA084CA2648C95ULL, "6066672_m.fxe"},
	{0x1AEA939172ECC9E7ULL, "609765897.ogg"},
	{0x1AEA9C7B7D3420B6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwielddeselectedexit.ani"},
	{0x1AEAFE2583A420D8ULL, "6163978_m.fxe"},
	{0x1AEB02D931D1121EULL, "6093972_m.fxe"},
	{0x1AEB190337893D1BULL, "vxlt_blackemporium_130_0_0.vlm"},
	{0x1AEB2EA2B4205E1BULL, "mag291_vincento6027352.cl"},
	{0x1AEB534CAB25D06AULL, "vxw_sundermount_2zj_0_1.vlm"},
	{0x1AEBABF22EC2ACDFULL, "hlac_ca_caverns1_24424_12y.dds"},
	{0x1AEBD0FA1C1B1BA0ULL, "vxlt_0_lowtown_docks_a3_n_cy5_0_1.vlm"},
	{0x1AEBDFF7BBEC58F4ULL, "769740056.ogg"},
	{0x1AEC3828730A885AULL, "vxlt_0_lowtown_n_fhz_0_0.vlm"},
	{0x1AEC579843F95D8AULL, "follower_banter6161243.cl"},
	{0x1AEC65CE72977B70ULL, "17521772.ogg"},
	{0x1AED8A706A4EDAE9ULL, "vxht_mansion2b_3go_0_1.vlm"},
	{0x1AEE47356B12CA88ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v8.gad"},
	{0x1AEE524874080360ULL, "vxw_woundedcoast_kl_0_0.vlm"},
	{0x1AEE7F8D7CA22885ULL, "hlac_lt_gallowscourtyard_n_-1_0z.dds"},
	{0x1AEFE88971DF0B0FULL, "hlac_lt_0_lowtown_docks_-2_3y.dds"},
	{0x1AF0077B9BC8B5CBULL, "vxlt_0_lowtown_a2_n_c_gjz_0_0.vlm"},
	{0x1AF0088CF0C4EB4CULL, "6105376_m.fxe"},
	{0x1AF056C3085C5362ULL, "6175320_m.fxe"},
	{0x1AF08279C12A4DE8ULL, "hla_ht_0_hightown_a2_131072_0x.dds"},
	{0x1AF0ED5A97323C55ULL, "369875485.ogg"},
	{0x1AF10AD322D7B991ULL, "6166782_f.fxe"},
	{0x1AF2E70DB792A388ULL, "112998011.ogg"},
	{0x1AF37969F0D5ACC3ULL, "vxlt_undercityserialkiller_18n_0_0.vlm"},
	{0x1AF37AE570D6C5D2ULL, "vxht_0_hightown_2xa_0_0.vlm"},
	{0x1AF380D4D6FA2FF9ULL, "vxlt_0_lowtown_docks_n_b08_0_0.vlm"},
	{0x1AF3894E282E877DULL, "vxlt_0_lowtown_a3_n_c_ht0_0_0.vlm"},
	{0x1AF418D3EF76732CULL, "vxw_blightlands_tg_0_0.vlm"},
	{0x1AF4A2D4FDAA3BFDULL, "hlac_lt_gallowscourtyard_-65537_0x.dds"},
	{0x1AF4D084B2F89021ULL, "character_character_p235.psh.fxo.dx11.assembly"},
	{0x1AF4FAE52BAA2F8DULL, "vxdr_primevaltomb_4hi_0_0.vlm"},
	{0x1AF53B1401B8332CULL, "vxw_sundermount_2gz_0_0.vlm"},
	{0x1AF5473737F27086ULL, "vxlt_0_lowtown_a3_n_c_gpd_0_0.vlm"},
	{0x1AF5560804F42C64ULL, "6147749_m.fxe"},
	{0x1AF55829F8DDF303ULL, "one100dg_vanard6059699.cl"},
	{0x1AF578DF7542815DULL, "6207006_m.fxe"},
	{0x1AF5B095D92A4BE8ULL, "hlac_w_woundedcoast_196609_0x.dds"},
	{0x1AF5BC37206DE67BULL, "660937419.ogg"},
	{0x1AF5BE7A825576B1ULL, "dae011_corff_sta6149636.cl"},
	{0x1AF5D92DC1753E85ULL, "6143602_f.fxe"},
	{0x1AF5DF31283EA671ULL, "559899509.ogg"},
	{0x1AF62A0703142091ULL, "dae000ip_crowd_em3_talk12.utp"},
	{0x1AF663B8CB7F3211ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxb_1.mao"},
	{0x1AF6D4FE4A0745F9ULL, "110172294.ogg"},
	{0x1AF6E20BEB66B637ULL, "vxw_bonepitdestroyed_a3_236_0_0.vlm"},
	{0x1AF73D1EA97D09EBULL, "vxca_caverns1_mines_i77_0_0.vlm"},
	{0x1AF755C0C4150C8DULL, "zz_mcr_debug6088442.cl"},
	{0x1AF785AD423D6AAFULL, "w_woundedcoast_111_0.mao"},
	{0x1AF791AD423D7F4CULL, "w_woundedcoast_111_0.mmh"},
	{0x1AF797AD423D893EULL, "w_woundedcoast_111_0.msh"},
	{0x1AF7A44058E12F91ULL, "vxlt_0_lowtown_n_eia_0_0.vlm"},
	{0x1AF7EA9F3B246AB1ULL, "dae021_sister_colinda_sta6150201.cl"},
	{0x1AF826DC46483754ULL, "492896060.ogg"},
	{0x1AF841B5B034A98BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mase_0n.dds"},
	{0x1AF85DB540110BF8ULL, "6111203_m.fxe"},
	{0x1AF8A0E4B8BB2A9DULL, "rdr110dg_dockworker.bnk"},
	{0x1AF8A47B7D406BA7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwielddeselectedexit.evt"},
	{0x1AF9350306D38D75ULL, "mrl150cr_arcane_horror.utc"},
	{0x1AF98AE0297F559BULL, "vxlt_blackemporium_163_0_0.vlm"},
	{0x1AF99ABB18BA0C59ULL, "6040773_m.fxe"},
	{0x1AFA23001D56A358ULL, "mer221_sabin6048283.cl"},
	{0x1AFA59E2CE289F68ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masa_0d.dds"},
	{0x1AFA9F1B51BAD5FCULL, "pro000_merchant6051842.cl"},
	{0x1AFB1FEF2E65265EULL, "98633407.ogg"},
	{0x1AFB22485D49AD55ULL, "w_bonepitmining_a2_10_0.msh"},
	{0x1AFB24485D49B0CBULL, "w_bonepitmining_a2_10_0.mmh"},
	{0x1AFB2DDE7F314E34ULL, "vxlt_gallowscourtyard_761_0_0.vlm"},
	{0x1AFB30485D49C568ULL, "w_bonepitmining_a2_10_0.mao"},
	{0x1AFB7F3D03130812ULL, "vxlt_hangedman_n_1ra_0_0.vlm"},
	{0x1AFBE52DD3E3615AULL, "vxdr_0_deeproads_9gt_0_0.vlm"},
	{0x1AFC82E4B8BEE0A5ULL, "rdr110dg_dockworker.cnv"},
	{0x1AFC85606B11E1DFULL, "ter_bigrock01_d.dds"},
	{0x1AFEC57B7D450FF0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwielddeselectedexit.gad"},
	{0x1AFEDE893FA37DC7ULL, "hlac_lt_0_lowtown_-2_4y.dds"},
	{0x1AFF7FC06AF22119ULL, "432694001.ogg"},
	{0x1AFFA406C0EBAEE5ULL, "sib131_slave_a6165260.cl"},
	{0x1B0030C5A5C93A2FULL, "6067740_f.fxe"},
	{0x1B008B3D7A54139BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_basic.phy"},
	{0x1B00C6A384EB34C3ULL, "267817884.ogg"},
	{0x1B0136E252E88F29ULL, "6198472_m.fxe"},
	{0x1B01925BCD0A400EULL, "vxw_bonepitmining_a2_57_0_0.vlm"},
	{0x1B019418BDDF6549ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_2x.dds"},
	{0x1B019B70F18AE07CULL, "6027762_m.fxe"},
	{0x1B01C69D394D7D32ULL, "vxw_sundermount_2en_0_0.vlm"},
	{0x1B02315F00FD0D2AULL, "hlac_lt_0_lowtown_a2_65536_10y.dds"},
	{0x1B030C27D7CD3F9FULL, "vxw_woundedcoast_1ds_0_0.vlm"},
	{0x1B033DE8143EF291ULL, "686364640.ogg"},
	{0x1B038948ABA826AAULL, "vxlt_0_lowtown_gl8_0_0.vlm"},
	{0x1B03F7351CFC99C8ULL, "gen00fl_isabela_init6110267.cl"},
	{0x1B0400D18B2AC9A6ULL, "mag360_alain6044655.cl"},
	{0x1B04205A17D5D3C8ULL, "907549464.ogg"},
	{0x1B04775AD26E427DULL, "addtacticpreset.ncs"},
	{0x1B0533DB69E46732ULL, "hlac_lt_0_lowtown_docks_a3_-3_3z.dds"},
	{0x1B0547F6798265A3ULL, "isa211_isabela6202204.cl"},
	{0x1B0588B933D0D96CULL, "l_dae210ar_lowtown.wbd"},
	{0x1B06762E2BC6E36FULL, "hlac_ca_cave_2194_7y.dds"},
	{0x1B06F11FC377ECC0ULL, "lgt110cr_coa_captain_1.utc"},
	{0x1B06F3457A26C18EULL, "vxlt_gallowstemplar_4si_0_0.vlm"},
	{0x1B075176FBBE604AULL, "ser270_templar6165683.cl"},
	{0x1B07F067E85F441FULL, "llac_lt_0_lowtown_a2_n_c_-196608_4.dds"},
	{0x1B08230555A70ED5ULL, "6207811_m.fxe"},
	{0x1B08979CD59E68A6ULL, "6152143_m.fxe"},
	{0x1B08F7ADE60623AEULL, "22891327.ogg"},
	{0x1B093666F86E0250ULL, "vxht_0_hightown_a2_a2n_0_0.vlm"},
	{0x1B0A06CF497D3967ULL, "368381857.ogg"},
	{0x1B0A223842D93958ULL, "vfx_chunkyparticle_p1.vsh.fxo"},
	{0x1B0AA2F26EE1DD1AULL, "skel_f2.dds"},
	{0x1B0AD4E6BD89D6CDULL, "891207493.ogg"},
	{0x1B0B747C7209299BULL, "hlac_lt_darktown_-1_77y.dds"},
	{0x1B0C2B6D7A9BBED0ULL, "6142789_f.fxe"},
	{0x1B0C50F1DFA9A075ULL, "follower_banter6170491.cl"},
	{0x1B0C70E5AC549D57ULL, "lgt260_way_blood_mage6148382.cl"},
	{0x1B0D70FEA10DC894ULL, "408767767.ogg"},
	{0x1B0DC0DBB8BA5CF8ULL, "381510901.ogg"},
	{0x1B0DEFFD51CEFD8BULL, "827895031.ogg"},
	{0x1B0E779412638BD6ULL, "381609070.ogg"},
	{0x1B0E89ABD2FE1FFAULL, "hlac_w_bonepitmining_a2_65541_0y.dds"},
	{0x1B0EA8494F1D209BULL, "422083280.ogg"},
	{0x1B0EF613759ED660ULL, "vxlt_0_lowtown_a3_n_c_goj_0_1.vlm"},
	{0x1B0F0C913055A4AFULL, "vxw_encounter1_f8_0_0.vlm"},
	{0x1B0F736A83D3C666ULL, "vxlt_0_lowtown_a2_gxl_0_0.vlm"},
	{0x1B0F8A2BFB36BB6DULL, "6165651_f.fxe"},
	{0x1B0FA59D1118B7D1ULL, "vxht_mansion_n_59y_0_0.vlm"},
	{0x1B1008C959F84DEBULL, "972119170.ogg"},
	{0x1B10387A2B4395B2ULL, "vxw_sundermount_2no_0_0.vlm"},
	{0x1B109DB21280BA2AULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtb.mao"},
	{0x1B10A1B21280C179ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtb.mmh"},
	{0x1B10ABB21280D267ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtb.msh"},
	{0x1B1170AD39B5C3E0ULL, "zz_cinetest.lst"},
	{0x1B117E04C9C3581FULL, "vxw_woundedcoast_1or_0_0.vlm"},
	{0x1B11BF5C380A714DULL, "6113086_m.fxe"},
	{0x1B1206B86FB0E0BEULL, "vxlt_0_lowtown_docks_a3_ctb_0_0.vlm"},
	{0x1B121BC3B07EC670ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1a.msh"},
	{0x1B121ECDE5A4642CULL, "vxw_sundermount_2mx_0_0.vlm"},
	{0x1B1229C3B07EDD8AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1a.mmh"},
	{0x1B122DC3B07EE4C1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1a.mao"},
	{0x1B131F4878764DADULL, "6118010_m.fxe"},
	{0x1B139853F105DB3DULL, "58028077.ogg"},
	{0x1B14BE2D3F0B27D9ULL, "nat380_sibling6207562.cl"},
	{0x1B14D24EDFCD6184ULL, "6033701_m.fxe"},
	{0x1B1543D2A339D679ULL, "vxca_caverns2_95i_0_0.vlm"},
	{0x1B1673F35923679DULL, "vxlt_0_lowtown_h7m_0_0.vlm"},
	{0x1B168ED4DBDC03E6ULL, "one100dg_kelder6060543.cl"},
	{0x1B16D1649FFF17A5ULL, "w_encounter3_14_0.rml"},
	{0x1B16D50008C0FB22ULL, "narrative_varric6172473.cl"},
	{0x1B16ED62C7A22B70ULL, "vxdr_primevalentrance_15n_0_0.vlm"},
	{0x1B17797CDB438565ULL, "hlac_lt_0_lowtown_a2_n_1_0z.dds"},
	{0x1B17E39045A2365EULL, "357864227.ogg"},
	{0x1B1862F6E6FF3912ULL, "vxlt_hangedman_n_1dc_0_0.vlm"},
	{0x1B1869DCCB891BFFULL, "6117822_m.fxe"},
	{0x1B18EED2F8839CDFULL, "w_bonepit_10_0.msh"},
	{0x1B18F4D2F883A681ULL, "w_bonepit_10_0.mmh"},
	{0x1B1900D2F883BB22ULL, "w_bonepit_10_0.mao"},
	{0x1B19025CFE366B70ULL, "504563359.ogg"},
	{0x1B19A7DF49EA4CE0ULL, "vxw_woundedcoast_zk_0_0.vlm"},
	{0x1B1A0615FCCF0FF6ULL, "vxht_0_hightown_a3_7fw_0_0.vlm"},
	{0x1B1A2C2DDBD80C28ULL, "mag101_grace6034377.cl"},
	{0x1B1B07D54BC265AEULL, "6132667_f.fxe"},
	{0x1B1B75C701966E53ULL, "6115781_m.fxe"},
	{0x1B1BB23BF8C6DF47ULL, "llac_ht_mansion2_818_36.dds"},
	{0x1B1BC22506896729ULL, "6120115_f.fxe"},
	{0x1B1C34CA7BBFC31DULL, "vxca_cavern1_haunted_eic_0_0.vlm"},
	{0x1B1C67AD39BFBBABULL, "zz_cinetest.are"},
	{0x1B1CC50B1E8C5260ULL, "vxw_bonepit_1ve_0_0.vlm"},
	{0x1B1CCCC52C82C270ULL, "dae000ip_crowd_mf6_talk2.utp"},
	{0x1B1CE971786179C7ULL, "6103953_m.fxe"},
	{0x1B1D1F7738D4B92AULL, "mag370_meredith6172392.cl"},
	{0x1B1D282098F5756BULL, "vxlt_blackemporium_15j_0_0.vlm"},
	{0x1B1D6C31EA62858CULL, "gen00fl_aveline6128798.cl"},
	{0x1B1DA92509E1C256ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb104a_0t.dds"},
	{0x1B1DBEC2E2621124ULL, "vxlt_ruinedhouse_n_i3_0_0.vlm"},
	{0x1B1DCFA27E7664DAULL, "dae000_fenarel6133315.cl"},
	{0x1B1DF32D3DB6122EULL, "dae000_fenarel6147437.cl"},
	{0x1B1E4A977391D527ULL, "6144866_m.fxe"},
	{0x1B1E5D1F4FBD4373ULL, "6109237_f.fxe"},
	{0x1B1EAA571D39AE32ULL, "vxw_sundermount_2sh_0_0.vlm"},
	{0x1B205157A073F9DCULL, "vxht_keep_61p_0_0.vlm"},
	{0x1B209FE241F5F819ULL, "vxht_0_hightown_a3_cb4_0_0.vlm"},
	{0x1B20FBE1489D8EBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_reaction.ani"},
	{0x1B215F7525951302ULL, "vxht_mansion_n_5cd_0_0.vlm"},
	{0x1B2169AB0F728BE4ULL, "gen00fl_merrill_pchouseact3.stg"},
	{0x1B218A0501B0AD91ULL, "6070908_f.fxe"},
	{0x1B218B642A0B270BULL, "mag360_samson6042102.cl"},
	{0x1B218FFBBC2E2083ULL, "lt_0_lowtown_docks_a3_n_230_0.msh"},
	{0x1B2191256A6C662DULL, "6117110_m.fxe"},
	{0x1B219DFBBC2E38DDULL, "lt_0_lowtown_docks_a3_n_230_0.mmh"},
	{0x1B21A1FBBC2E3FEEULL, "lt_0_lowtown_docks_a3_n_230_0.mao"},
	{0x1B21C9C55D0FCB23ULL, "6054266_m.fxe"},
	{0x1B21FED143E3B608ULL, "6107802_m.fxe"},
	{0x1B2217162FCE8BB3ULL, "1030240760.ogg"},
	{0x1B221AD1C7A3A4A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_berserk.ani"},
	{0x1B22405ED0BA5D41ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medb_0t.dds"},
	{0x1B22A474B2A02E71ULL, "hlac_ca_cavern1_haunted_24423_11z.dds"},
	{0x1B22ADED82760229ULL, "vxlt_0_lowtown_a2_n_gc1_0_0.vlm"},
	{0x1B22D65C68D70956ULL, "qun120cr_javaris6042988.cl"},
	{0x1B22D821B0D117F7ULL, "602810860.ogg"},
	{0x1B238E58FAA2AC20ULL, "199189968.ogg"},
	{0x1B240B69076F7DAEULL, "vxlt_0_lowtown_gp8_0_0.vlm"},
	{0x1B240DB0AE13C3FAULL, "455399681.ogg"},
	{0x1B247F609564B136ULL, "vxlt_gallowsprison_n_6pm_0_0.vlm"},
	{0x1B24E4F82C00BA54ULL, "hightown_fem_noble2_mob6169542.cl"},
	{0x1B24E6A27BBE06B0ULL, "217753436.ogg"},
	{0x1B24F66D15DC8ED8ULL, "6174790_m.fxe"},
	{0x1B250DEC749276F6ULL, "llac_lt_undercitydungeon_-1_47.dds"},
	{0x1B256FD01262B01CULL, "art\\characters\\creatures\\rubblegolem\\textures\\c_rbg_0da.dds"},
	{0x1B25EA5E4680A267ULL, "185498279.ogg"},
	{0x1B2605D7E086502BULL, "376688805.ogg"},
	{0x1B261773430A2728ULL, "vxw_encounter3_7j_0_0.vlm"},
	{0x1B2733DC33C2C011ULL, "hlac_lt_0_lowtown_-1_5y.dds"},
	{0x1B2747F9444C834FULL, "gen00flst_sebactthreeb.stg"},
	{0x1B275A4A5F7363DCULL, "lt_0_lowtown_docks_139_0.mao"},
	{0x1B27664A5F73787FULL, "lt_0_lowtown_docks_139_0.mmh"},
	{0x1B276C4A5F738221ULL, "lt_0_lowtown_docks_139_0.msh"},
	{0x1B2774FEB819ADD7ULL, "gen00fl_fenris6126837.cl"},
	{0x1B2776BCDA85A180ULL, "gen00fl_anders6114435.cl"},
	{0x1B27812CB4EC4FAAULL, "water.mat.dx11"},
	{0x1B27F4B30F8028E2ULL, "dae000ip_crowd_em1_talk17.utp"},
	{0x1B2819BC3BE06560ULL, "vxw_woundedcoast_wh_0_0.vlm"},
	{0x1B286C7C82C89154ULL, "6154910_m.fxe"},
	{0x1B28EC9581CCE034ULL, "6216117_m.fxe"},
	{0x1B28FD6C045C28FEULL, "mcr371_orsino6083469.cl"},
	{0x1B291E5855B6F453ULL, "mrl310_party_comments6142760.cl"},
	{0x1B292558F4BF5C75ULL, "vxlt_0_lowtown_a2_ef7_0_0.vlm"},
	{0x1B2925D1C7A9D643ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_berserk.gad"},
	{0x1B295AC55F18D833ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv1.ani"},
	{0x1B29A9BD5EA9295DULL, "static_static_p13.psh.fxo"},
	{0x1B2A0252413129BEULL, "vxlt_0_lowtown_n_bz0_0_0.vlm"},
	{0x1B2AD6624EDF39FBULL, "363122200.ogg"},
	{0x1B2AE71F4DF15B4DULL, "6080737_m.fxe"},
	{0x1B2B29C42C718799ULL, "hlac_w_sundermount_-393216_0y.dds"},
	{0x1B2B3F40497C574CULL, "art\\levels\\races\\natural\\proxy\\foliage\\gra_zzgrasssml.msh"},
	{0x1B2B4540497C613EULL, "art\\levels\\races\\natural\\proxy\\foliage\\gra_zzgrasssml.mmh"},
	{0x1B2B5140497C75D5ULL, "art\\levels\\races\\natural\\proxy\\foliage\\gra_zzgrasssml.mao"},
	{0x1B2B862D8AED3F75ULL, "vxht_keep_n_6l1_0_0.vlm"},
	{0x1B2B90908305AD59ULL, "810602811.ogg"},
	{0x1B2C309FD45829A4ULL, "vxlt_ruinedhouse_n_l0_0_0.vlm"},
	{0x1B2CA8EBE4210D00ULL, "vxca_caverns1_mines_j4j_0_0.vlm"},
	{0x1B2D1C340F2FC6B2ULL, "vxw_sundermount_2ti_0_0.vlm"},
	{0x1B2D2538313CD523ULL, "923910416.ogg"},
	{0x1B2D723D7A7AD73AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_basic.msh"},
	{0x1B2D803D7A7AEF73ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_basic.mao"},
	{0x1B2D843D7A7AF580ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_basic.mmh"},
	{0x1B2D9C9BD92B6A4EULL, "6118574_m.fxe"},
	{0x1B2E03E148A826B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_reaction.evt"},
	{0x1B2EB108D6F97CE9ULL, "1036647772.ogg"},
	{0x1B2EB7ACE65FA477ULL, "hlac_w_bonepitmining_a2_0_1x.dds"},
	{0x1B2EBC0A74F1F517ULL, "941714234.ogg"},
	{0x1B2F0F3D3A99E7A3ULL, "dae000ip_crowd_ff1_talk5.utp"},
	{0x1B2FBA6E46D0E14AULL, "458944473.ogg"},
	{0x1B2FF4B9E3E419E6ULL, "6156334_m.fxe"},
	{0x1B3002D1C7AFB928ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_berserk.evt"},
	{0x1B301DC279F65E26ULL, "6141792_m.fxe"},
	{0x1B31621F6CB653E8ULL, "llac_lt_gallowstemplar_f_2383_9.dds"},
	{0x1B318B63E773D729ULL, "ser200_aveline.bnk"},
	{0x1B322D1777C90228ULL, "vxht_0_hightown_n_9u4_0_0.vlm"},
	{0x1B335C4ECAE8EEA9ULL, "isa211_isabela6071022.cl"},
	{0x1B339EAB835B063DULL, "fen110_fenris6100320.cl"},
	{0x1B33C9A111D2907BULL, "hm_lgt_apostate01.mor"},
	{0x1B33E2CB26A37099ULL, "vxlt_gallowstemplar_n_5rg_0_0.vlm"},
	{0x1B33ECD69E6F0484ULL, "6133746_m.fxe"},
	{0x1B3461A596673E32ULL, "gen00fl_anders6125903.cl"},
	{0x1B34C8666B8CF798ULL, "vxlt_0_lowtown_a2_n_c_eib_0_0.vlm"},
	{0x1B34EFDDDABB17F6ULL, "6156738_m.fxe"},
	{0x1B34F0767C09F1DEULL, "6069731_m.fxe"},
	{0x1B3522E148AE7A45ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_reaction.gad"},
	{0x1B3537E1576ACFAFULL, "6125622_f.fxe"},
	{0x1B355890D4738E97ULL, "184675875.ogg"},
	{0x1B356D63E7778D69ULL, "ser200_aveline.cnv"},
	{0x1B36B715632AF07AULL, "cod_bks_lite_vials_5.utp"},
	{0x1B373BD13FE0905EULL, "229471414.ogg"},
	{0x1B3752C55F25086AULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv1.evt"},
	{0x1B376F271F59F554ULL, "zz_qun_debug6028221.cl"},
	{0x1B379735E6B574F5ULL, "vxlt_0_lowtown_a2_ei6_0_0.vlm"},
	{0x1B381368B0190DF6ULL, "hlac_ca_caverns1_24424_5x.dds"},
	{0x1B381B9A509F41DDULL, "static_static_p40.psh.fxo"},
	{0x1B381EF0B7A4E23DULL, "hlac_lt_warehouse_1_1x.dds"},
	{0x1B385C10EECC9F18ULL, "vxht_chantry_n_2d0_0_0.vlm"},
	{0x1B388D4CE1E6D1F2ULL, "vxlt_0_lowtown_a3_n_c_d6y_0_0.vlm"},
	{0x1B38BFD05A3ADF08ULL, "mag371_elsa6041258.cl"},
	{0x1B395A0E98FD532FULL, "432847885.ogg"},
	{0x1B39D7F6936B834CULL, "vxht_0_hightown_a3_bxz_0_0.vlm"},
	{0x1B3A003F991CDF26ULL, "vxw_woundedcoast_1r0_0_0.vlm"},
	{0x1B3A7B6B1C82382EULL, "364233106.ogg"},
	{0x1B3A7F7014E283ABULL, "follower_banter6163440.cl"},
	{0x1B3AD0E64213C0FEULL, "gen00fl_isabela6098216.cl"},
	{0x1B3B19C40CAAD2CEULL, "sky_w_encounter1.mao"},
	{0x1B3B1C4725052B2DULL, "6147762_m.fxe"},
	{0x1B3B8E110125DF32ULL, "vxw_sundermount_2yj_0_0.vlm"},
	{0x1B3BBD8C04163A7AULL, "153283202.ogg"},
	{0x1B3BCDF723B71CECULL, "6084545_m.fxe"},
	{0x1B3C1C2D3B4E8402ULL, "6052913_m.fxe"},
	{0x1B3CB164A01FB0CEULL, "w_encounter3_14_0.mmh"},
	{0x1B3CBD64A01FC565ULL, "w_encounter3_14_0.mao"},
	{0x1B3CCB64A01FDCBCULL, "w_encounter3_14_0.msh"},
	{0x1B3CE2CA09AC2ED3ULL, "vxht_0_hightown_a3_bzk_0_0.vlm"},
	{0x1B3D57A7DF2F77F2ULL, "vxw_bonepitdestroyed_a3_273_0_0.vlm"},
	{0x1B3D8AD477CD33B6ULL, "562547603.ogg"},
	{0x1B3DB90772E7573FULL, "gen_im_cns_flu_sebastian_02.uti"},
	{0x1B3DD4D2F8A28CAAULL, "w_bonepit_10_0.rml"},
	{0x1B3DDBBF1BA65FE4ULL, "vxw_sundermount_34o_0_1.vlm"},
	{0x1B3E31C55F2AEEDDULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv1.gad"},
	{0x1B3E3DA5E7DE93A0ULL, "419927789.ogg"},
	{0x1B3E9DC347538F5DULL, "mag170_recruit026080922.cl"},
	{0x1B3ED5CB67AB8744ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.dura.ani"},
	{0x1B3F4F36AE7A473CULL, "zz_vault_debug6203383.cl"},
	{0x1B3FFDEEE2CD8353ULL, "vxlt_gallowsprison_n_648_0_0.vlm"},
	{0x1B403B99004E1115ULL, "vxca_caverns2_80w_0_0.vlm"},
	{0x1B406A5840690AA4ULL, "mrl150_dalish_guard6170405.cl"},
	{0x1B4124F1254F2D8AULL, "vxht_chantry_300_0_0.vlm"},
	{0x1B41FAEECBB9F6E8ULL, "dae000_pc_home_ambients6208471.cl"},
	{0x1B4201BA9D747570ULL, "mag370_meredith6037404.cl"},
	{0x1B42F6E6E2872355ULL, "vxht_mansion_n_3v6_0_0.vlm"},
	{0x1B4302B0F8EE59BBULL, "art\\vfx\\combat\\vfx_m_blood_change_c.dura.evt"},
	{0x1B4324BEFFE0123DULL, "llac_lt_0_lowtown_a3_n_c_1_9.dds"},
	{0x1B4330032CE33A3AULL, "683220959.ogg"},
	{0x1B43A4D0FDEC7F53ULL, "hlac_ht_mansion2_818_21x.dds"},
	{0x1B43FB0E448E90A9ULL, "mcr310st_argument_blast.stg"},
	{0x1B4506ED0406B90FULL, "one220_yevhen6104147.cl"},
	{0x1B4578B1AFFCDA07ULL, "265401672.ogg"},
	{0x1B458566D9E7B29AULL, "137282368.ogg"},
	{0x1B45D2CB67B1A0A6ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.dura.gad"},
	{0x1B45F56FE53A0B9EULL, "hlac_lt_undercityencounter_-1_26x.dds"},
	{0x1B468B5647CADC7CULL, "vxlt_gallowscourtyard_n_59u_0_0.vlm"},
	{0x1B47068239D462F4ULL, "vxht_mansion2_3gv_0_1.vlm"},
	{0x1B47C1392D83E2BDULL, "309203691.ogg"},
	{0x1B480FED7060FC28ULL, "art\\characters\\weapons\\mauls\\w_mal_ml103a.msh"},
	{0x1B482DED70612F52ULL, "art\\characters\\weapons\\mauls\\w_mal_ml103a.mmh"},
	{0x1B48786BB2612B61ULL, "6187825_m.fxe"},
	{0x1B4888FE47762B02ULL, "vxca_caverns1_mines_fmo_0_0.vlm"},
	{0x1B48BB82C3B3128FULL, "6171116_m.fxe"},
	{0x1B48FBB0F8F2B9FCULL, "art\\vfx\\combat\\vfx_m_blood_change_c.dura.gad"},
	{0x1B494DC39478E856ULL, "isa211_party_reaction6202332.cl"},
	{0x1B49739CAF75BCDBULL, "vxlt_warehouse_n_wd_0_0.vlm"},
	{0x1B499C4AA38E82E3ULL, "vxlt_gallowscourtyard_n_4qk_0_0.vlm"},
	{0x1B49E63B30031CF4ULL, "708151631.ogg"},
	{0x1B4C4D9C0D9C7864ULL, "vxw_sundermount_33n_0_1.vlm"},
	{0x1B4C84F0BBE795F3ULL, "257661352.ogg"},
	{0x1B4C9B97C31E683DULL, "vxht_0_hightown_a3_c11_0_0.vlm"},
	{0x1B4CB3A4F063E3D8ULL, "mer320_hubert6053015.cl"},
	{0x1B4CCDCB67B7B685ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.dura.evt"},
	{0x1B4D34F31CD7231DULL, "dae000_sabina6151869.cl"},
	{0x1B4D49F26B87D1D1ULL, "vxw_encounter2_camp_rc_0_0.vlm"},
	{0x1B4D4CAFAE169839ULL, "86173912.ogg"},
	{0x1B4DCB4D369F3E1DULL, "ser221_gamlen6134683.cl"},
	{0x1B4E5E79D999ADA1ULL, "vxlt_shop_di_0_0.vlm"},
	{0x1B4E7987877F429BULL, "mag291cr_crowd_elf_3.utc"},
	{0x1B4FA526A42E8431ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sidetable_sa.dds"},
	{0x1B4FAA2AC13AF779ULL, "dre000_dougal6131161.cl"},
	{0x1B4FC679720EF36EULL, "6096207_m.fxe"},
	{0x1B4FCE118D260D28ULL, "vxlt_0_lowtown_a2_ewm_0_0.vlm"},
	{0x1B4FFAB0F8F8D6FAULL, "art\\vfx\\combat\\vfx_m_blood_change_c.dura.ani"},
	{0x1B5045A671AF4DFEULL, "6082112_m.fxe"},
	{0x1B50D090825B3584ULL, "6139146_m.fxe"},
	{0x1B51722484E5DFA0ULL, "6172738_m.fxe"},
	{0x1B518D4517BA8CA4ULL, "rdr110dg_guard_warehouse6016684.cl"},
	{0x1B51A1A2F6A2B500ULL, "one300_zevran6089764.cl"},
	{0x1B523C5A4FCA4815ULL, "606429178.ogg"},
	{0x1B528B2A2FFCB953ULL, "plt_mag200pt_nt_ambient.nss"},
	{0x1B52C53DC302CE52ULL, "146594708.ogg"},
	{0x1B52D05CD298E5D8ULL, "6015011_m.fxe"},
	{0x1B538958E2B436DAULL, "6136642_m.fxe"},
	{0x1B53AE0F68186A31ULL, "6091703_f.fxe"},
	{0x1B541E3427FAAEF1ULL, "6076040_m.fxe"},
	{0x1B54202A0AA0FB38ULL, "6131977_m.fxe"},
	{0x1B548F5BC7963A14ULL, "hlac_lt_0_lowtown_a3_n_c_65535_9z.dds"},
	{0x1B551F2AB1557AD6ULL, "hlac_lt_0_lowtown_65534_6x.dds"},
	{0x1B55A83732B46184ULL, "hlac_ca_caverns1_24426_3z.dds"},
	{0x1B55AAAD94F51557ULL, "vxca_caverns1_mines_edf_0_0.vlm"},
	{0x1B55FA0AEAF16BD6ULL, "6057074_f.fxe"},
	{0x1B568C7EE664B460ULL, "vxw_bonepit_1ja_0_0.vlm"},
	{0x1B56BBB07757B44CULL, "vxht_0_hightown_a3_brx_0_0.vlm"},
	{0x1B56C3BBE0A1BF31ULL, "vxlt_gallowstemplar_n_34o_0_0.vlm"},
	{0x1B576E0B3AFD96BFULL, "464987058.ogg"},
	{0x1B57E579A16BD55BULL, "vxlt_warehouse_n_zg_0_0.vlm"},
	{0x1B57F416E780E672ULL, "6080820_m.fxe"},
	{0x1B58849FD1F1CD57ULL, "vxdr_primevalentrance_234_0_0.vlm"},
	{0x1B588BA1D2E36128ULL, "vxw_sundermount_2zx_0_1.vlm"},
	{0x1B58CB3C0CE008D0ULL, "6043027_m.fxe"},
	{0x1B5908694D220145ULL, "dae100_bran6118305.cl"},
	{0x1B593FCFAAA04B97ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_6z.dds"},
	{0x1B59DBE0171104ECULL, "6027339_m.fxe"},
	{0x1B59ECACF525D2CCULL, "6083610_m.fxe"},
	{0x1B5A267488D7F234ULL, "hlac_lt_0_lowtown_a3_n_c_65536_11x.dds"},
	{0x1B5A72A88BE8EBAEULL, "6136466_m.fxe"},
	{0x1B5A9CB8B589F5D7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock3a.msh"},
	{0x1B5AAEB8B58A143AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock3a.mao"},
	{0x1B5AB2B8B58A1B49ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock3a.mmh"},
	{0x1B5AEBD9D0D40407ULL, "366646017.ogg"},
	{0x1B5B192170A726DEULL, "vxlt_hangedman_n_27a_0_0.vlm"},
	{0x1B5B7B469E9B30BFULL, "mag131_keran6086256.cl"},
	{0x1B5BBBCF5D7DEA51ULL, "vxw_encounter2_camp_mb_0_0.vlm"},
	{0x1B5C50FFD97DA5EEULL, "lgt111tr_gxa_coa_spy.utt"},
	{0x1B5C5A8A1D22F876ULL, "6152275_m.fxe"},
	{0x1B5C83593A30A9FFULL, "mcr310_argument6182464.cl"},
	{0x1B5D72AF9C2A1C1FULL, "225912486.ogg"},
	{0x1B5D7571A9ACA138ULL, "mag210_arianni6057342.cl"},
	{0x1B5DB178463B8921ULL, "76893862.ogg"},
	{0x1B5DF19569167758ULL, "6034322_m.fxe"},
	{0x1B5EB36E2065322BULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_p.mmh"},
	{0x1B5EF5E2A5FD5CE7ULL, "797847078.ogg"},
	{0x1B5F3861FC85BA19ULL, "vxlt_gallowstemplar_n_5cd_0_0.vlm"},
	{0x1B601DFD416F4118ULL, "vxlt_0_lowtown_a2_n_c_exa_0_0.vlm"},
	{0x1B6027B300FF0EC5ULL, "qun100_mage_party_comment6019116.cl"},
	{0x1B603E455E9656D4ULL, "721834164.ogg"},
	{0x1B609419DB252725ULL, "hlac_ht_keep_n_3732_0z.dds"},
	{0x1B610B8AE28D0D3FULL, "115565554.ogg"},
	{0x1B61D2FD7E61957DULL, "123389375.ogg"},
	{0x1B620090A6599A23ULL, "lt_0_lowtown_docks_a3_n_173_0.msh"},
	{0x1B620E90A659B27DULL, "lt_0_lowtown_docks_a3_n_173_0.mmh"},
	{0x1B621290A659B88EULL, "lt_0_lowtown_docks_a3_n_173_0.mao"},
	{0x1B6228A907496515ULL, "202539644.ogg"},
	{0x1B62CCB02FC33DC7ULL, "180397339.ogg"},
	{0x1B62FD3F8A98513CULL, "vxht_0_hightown_a3_cbm_0_0.vlm"},
	{0x1B6328F5F1C800A1ULL, "vxca_cave_2d7_0_1.vlm"},
	{0x1B638EB05A43068FULL, "gen00fl_anders6125438.cl"},
	{0x1B63B3C1C89472DBULL, "mag291_torpor6172571.cl"},
	{0x1B63F0BAFA5123D5ULL, "21314088.ogg"},
	{0x1B63F49DB18F5941ULL, "6116052_m.fxe"},
	{0x1B643028A6B82127ULL, "341843765.ogg"},
	{0x1B64506DE541683AULL, "52561094.ogg"},
	{0x1B647D8B6875612FULL, "vxlt_hangedman_27k_0_0.vlm"},
	{0x1B64F38FC468F83FULL, "art\\levels\\races\\natural\\highlands\\props\\prp_caveent.msh"},
	{0x1B64F98FC46902E1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_caveent.mmh"},
	{0x1B64FE5BD85ACCE0ULL, "vxw_bonepit_1on_0_0.vlm"},
	{0x1B65058FC4691682ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_caveent.mao"},
	{0x1B652D8D694DCCCCULL, "vxht_0_hightown_a3_biy_0_0.vlm"},
	{0x1B6583BA0029330DULL, "762017495.ogg"},
	{0x1B658A46E39E3735ULL, "6108997_f.fxe"},
	{0x1B65A98921A3E4EEULL, "one100_party_comments6106778.cl"},
	{0x1B6791C1E7EFD150ULL, "6122923_f.fxe"},
	{0x1B67E9D67BA2EAFFULL, "vxw_encounter1_ee_0_0.vlm"},
	{0x1B685E89E71BEB4CULL, "6086910_m.fxe"},
	{0x1B68E1AAC710CCDAULL, "6079280_m.fxe"},
	{0x1B69A57E9B2BD0ECULL, "vxca_cavern1_haunted_jg3_0_0.vlm"},
	{0x1B69CD01E90AB442ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_reaction.gad"},
	{0x1B69E3A2A485D4DBULL, "6070798_m.fxe"},
	{0x1B69ECDAD698856DULL, "17870215.ogg"},
	{0x1B6A1FE04E61D5D0ULL, "weapon_distortionmask_p0.psh.fxo.dx11"},
	{0x1B6A2DAC4F7402D1ULL, "vxw_encounter2_camp_ha_0_0.vlm"},
	{0x1B6A31D48FBFE27CULL, "zz_followers_banter6159042.cl"},
	{0x1B6A6B3C28ED27BDULL, "vxlt_hangedman_n_4s_0_0.vlm"},
	{0x1B6A779A91A17FB1ULL, "gen00fl_carver6097430.cl"},
	{0x1B6A87E62E7A6FA6ULL, "6116579_f.fxe"},
	{0x1B6BFAAE6AF7045CULL, "art\\levels\\races\\proxy\\_textures\\ter_4_d.dds"},
	{0x1B6C946E20713A90ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_p.anb"},
	{0x1B6C976C2C272502ULL, "lowtown_fem_prostitute_sta6152356.cl"},
	{0x1B6D804631F178AFULL, "hlac_lt_gallowstemplar_n_7198_3x.dds"},
	{0x1B6D878D3A8CD55AULL, "203867530.ogg"},
	{0x1B6D990288A6BACCULL, "vxlt_0_lowtown_enw_0_1.vlm"},
	{0x1B6E75EF0CBC5125ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_010a_0d.dds"},
	{0x1B6F1238D92441A9ULL, "757624550.ogg"},
	{0x1B6F6A4CD1516F3CULL, "64671903.ogg"},
	{0x1B6F70C832369541ULL, "vxlt_darktown_67l_0_0.vlm"},
	{0x1B6F75A91AE7A2E8ULL, "vxht_0_hightown_c5e_0_0.vlm"},
	{0x1B6F75B0539A28F8ULL, "6111012_m.fxe"},
	{0x1B6F7CD547EAA40AULL, "698814554.ogg"},
	{0x1B701FB6466A6289ULL, "vxht_bartrand_n_5ro_0_0.vlm"},
	{0x1B70397FE05CE3A8ULL, "mag131_taraneh6094539.cl"},
	{0x1B70F001E9110E39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_reaction.evt"},
	{0x1B70F5CD67E1574EULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180l_rnf.ani"},
	{0x1B717CB8515CC218ULL, "vxlt_0_lowtown_a2_n_c_h26_0_0.vlm"},
	{0x1B7191E9D9E26C44ULL, "sb03_bcloud.dds"},
	{0x1B720176EA8F6743ULL, "hlac_lt_gamlenshouse_n_-1_0y.dds"},
	{0x1B720590E3DA53AEULL, "vxca_cavern1_haunted_j57_0_0.vlm"},
	{0x1B725803C8AA9426ULL, "536990374.ogg"},
	{0x1B72789317679963ULL, "6043763_m.fxe"},
	{0x1B72E22F49C25372ULL, "6070481_m.fxe"},
	{0x1B732F6B6E6C1014ULL, "vxdr_primevaltomb_7by_0_0.vlm"},
	{0x1B733E23687B423AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_enter.gad"},
	{0x1B7343FFD9906E17ULL, "lgt111tr_gxa_coa_spy.ncs"},
	{0x1B736E838E557289ULL, "plt_cod_bks_death_templar.plo"},
	{0x1B736F506AC49A67ULL, "llac_ca_cavern1_haunted_24425_4.dds"},
	{0x1B7396654CD6CE3CULL, "vxht_bartrand_n_3s7_0_0.vlm"},
	{0x1B73A2B212D54066ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtb.phy"},
	{0x1B73AE2B086CCEFBULL, "6208412_m.fxe"},
	{0x1B748216585A26BDULL, "714091249.ogg"},
	{0x1B74846FE30D82DCULL, "96590597.ogg"},
	{0x1B748CB29D459558ULL, "418712690.ogg"},
	{0x1B749174677819FDULL, "mag311_nyssa6041946.cl"},
	{0x1B750B742AD2FA76ULL, "6083425_m.fxe"},
	{0x1B7557D9C558DE99ULL, "lowtown_argument6169405.cl"},
	{0x1B756A9636A24339ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masc_0n.dds"},
	{0x1B756ED2CF84D576ULL, "6144364_m.fxe"},
	{0x1B75F294C767DAFBULL, "6013331_m.fxe"},
	{0x1B765BB36D99037FULL, "vxw_encounter1_jd_0_0.vlm"},
	{0x1B768F61D3989F5EULL, "vxht_bartrand_n_4dk_0_0.vlm"},
	{0x1B769B653E403749ULL, "gen00fl_dog6117793.cl"},
	{0x1B76AA3DD18490D3ULL, "vxht_0_hightown_a3_bfo_0_0.vlm"},
	{0x1B76FE046FC83EFFULL, "vxdr_primevaltomb_9db_0_0.vlm"},
	{0x1B770F7DA4FFB254ULL, "ser100_party_comments6120594.cl"},
	{0x1B77156AB197D99BULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.cess.ani"},
	{0x1B77EF3BDB7169D6ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml107a_0t.dds"},
	{0x1B77F82718804C7BULL, "111656805.ogg"},
	{0x1B77FFD96D5F6DD1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medb_0d.dds"},
	{0x1B7857C862E5ED1FULL, "vxlt_darktown_5c9_0_0.vlm"},
	{0x1B7900A6F46DF15BULL, "178322002.ogg"},
	{0x1B797179034AE4A4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_basic.msh"},
	{0x1B798379034B030DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_basic.mao"},
	{0x1B798779034B0A46ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_basic.mmh"},
	{0x1B798CDA290D666DULL, "6071759_m.fxe"},
	{0x1B79ECB4DB593239ULL, "gen00fl_aveline6139742.cl"},
	{0x1B7A412368816651ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_enter.evt"},
	{0x1B7A804D8C6B5274ULL, "vxht_mansion_n_6k5_0_0.vlm"},
	{0x1B7A92DE9AFA628DULL, "gen00fl_aveline6122140.cl"},
	{0x1B7B6B71CB1462F4ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_2_darren.ani"},
	{0x1B7BB14CF4FE3B69ULL, "vxlt_gallowstemplar_n_2pb_0_0.vlm"},
	{0x1B7C3B733BE928D7ULL, "vxht_0_hightown_a2_byc_0_0.vlm"},
	{0x1B7C4C5006656104ULL, "da2terrain_rendertodepth_p1.psh.fxo.dx11"},
	{0x1B7CE5A717B2590CULL, "additivemagicplate01.mao"},
	{0x1B7D071EF2B5022FULL, "vxca_caverns1_mines_fn7_0_0.vlm"},
	{0x1B7DF4C665022686ULL, "802285539.ogg"},
	{0x1B7DF801E91BA628ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_reaction.ani"},
	{0x1B7E919338607B09ULL, "vxht_bartrand_n_5un_0_0.vlm"},
	{0x1B7EA6D87D62D170ULL, "440403658.ogg"},
	{0x1B7ED17EDB706FA1ULL, "vxlt_gallowscourtyard_79u_0_0.vlm"},
	{0x1B7EEDCD67ED869FULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180l_rnf.evt"},
	{0x1B7F188BBDF27BC8ULL, "gen00fl_varric6102755.cl"},
	{0x1B7F44BD49C4982BULL, "llac_lt_0_lowtown_a3_n_c_65536_9.dds"},
	{0x1B7FDB0C1947D2A7ULL, "260000656.ogg"},
	{0x1B8001518D663B0BULL, "6113093_m.fxe"},
	{0x1B8046C66ACA23C4ULL, "272115008.ogg"},
	{0x1B80A352DB523CD0ULL, "mag111ar_warehouse.are"},
	{0x1B80ED0883056042ULL, "vxht_0_hightown_ca3_0_0.vlm"},
	{0x1B8137EDD940305BULL, "pro000_colonel_ewald6039980.cl"},
	{0x1B81624C17AF77BDULL, "seb321_leliana6180915.cl"},
	{0x1B816871CB18C956ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_2_darren.gad"},
	{0x1B8176178D6A963DULL, "6173667_f.fxe"},
	{0x1B81CC9613DD246DULL, "lgt000ip_store_epic2_2.utp"},
	{0x1B81D1BF2DEEE36FULL, "263329677.ogg"},
	{0x1B81E747BAEB6CE2ULL, "llac_lt_gamlenshouse_n_-1_2.dds"},
	{0x1B8235F393B879C2ULL, "vxlt_gallowstemplar_n_2mn_0_0.vlm"},
	{0x1B826D8CA68C8657ULL, "vxlt_hangedman_2ei_0_0.vlm"},
	{0x1B8378E3B778D801ULL, "zz_qun_debug6085861.cl"},
	{0x1B8383B534ED29E9ULL, "one280_emrys6116789.cl"},
	{0x1B83867493E52988ULL, "w_encounter3_102_0.msh"},
	{0x1B83987493E54819ULL, "w_encounter3_102_0.mao"},
	{0x1B83A47493E55CB2ULL, "w_encounter3_102_0.mmh"},
	{0x1B83BA52DB5499D2ULL, "mag111ar_warehouse.ncs"},
	{0x1B83E8DC79ED6513ULL, "gen00fl_anders6125742.cl"},
	{0x1B841D6AB1A27192ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.cess.evt"},
	{0x1B8425EC363F834CULL, "hlac_lt_0_lowtown_docks_a3_-1_1x.dds"},
	{0x1B8468720268A6DAULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_ready.gad"},
	{0x1B846A7BEA1921C4ULL, "art\\vfx\\combat\\vfx_r_dwp_left_close1_c.impa.ani"},
	{0x1B847E7903550211ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_basic.phy"},
	{0x1B85CECD67F37088ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180l_rnf.gad"},
	{0x1B85D96E2694266EULL, "41568260.ogg"},
	{0x1B86210520A57CF9ULL, "and231ip_loot_1.utp"},
	{0x1B8651A1A9D7E802ULL, "vxlt_0_lowtown_coy_0_0.vlm"},
	{0x1B868B0160DE2D7BULL, "536717703.ogg"},
	{0x1B86A9B5ED33A37DULL, "6086036_m.fxe"},
	{0x1B86F191B313CFB1ULL, "qun120cr_javaris6042545.cl"},
	{0x1B86F5C3C2AAD3A7ULL, "587261510.ogg"},
	{0x1B86F81555D410D8ULL, "vxlt_gallowscourtyard_n_4ux_0_0.vlm"},
	{0x1B872985B3B64BF9ULL, "mer320_hubert6089914.cl"},
	{0x1B8737FE6D757704ULL, "248087744.ogg"},
	{0x1B875923688C1920ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_enter.ani"},
	{0x1B87829B2B32328FULL, "1020384373.ogg"},
	{0x1B8854EFA1F6C19AULL, "151531253.ogg"},
	{0x1B8961D877BB0BC4ULL, "vxlt_gallowstemplar_f_51k_0_0.vlm"},
	{0x1B8994071A7CD11CULL, "vxdr_0_deeproads_fik_0_0.vlm"},
	{0x1B89958554FE6F28ULL, "vxlt_0_lowtown_a2_eki_0_0.vlm"},
	{0x1B8A2419BF6A222BULL, "gen00fl_sebastian6180517.cl"},
	{0x1B8A6F4A5FC804C4ULL, "lt_0_lowtown_docks_139_0.phy"},
	{0x1B8AB252DB5AAAC3ULL, "mag111ar_warehouse.lst"},
	{0x1B8AEA7BE0834EFAULL, "w_bonepit_153_0.phy"},
	{0x1B8B3C6AB1A8C525ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.cess.gad"},
	{0x1B8B677BEA1F3B26ULL, "art\\vfx\\combat\\vfx_r_dwp_left_close1_c.impa.gad"},
	{0x1B8B6B72026ECAF1ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_ready.evt"},
	{0x1B8BDCB94F33901EULL, "828539267.ogg"},
	{0x1B8BF659EB2C5228ULL, "vxht_keep_n_6ic_0_0.vlm"},
	{0x1B8C964DFC3CC834ULL, "23080168.ogg"},
	{0x1B8D03702A569389ULL, "vxht_bartrand_n_5xm_0_0.vlm"},
	{0x1B8D2935BC850DD7ULL, "vxlt_gallowstemplar_2ds_0_0.vlm"},
	{0x1B8D8BCC5C66126BULL, "vxlt_warehouse_8x_0_0.vlm"},
	{0x1B8DA8BE369BE7C2ULL, "vxht_0_hightown_a3_aq4_0_0.vlm"},
	{0x1B8DABB6FA3DEF39ULL, "6110843_m.fxe"},
	{0x1B8F4474A880FCCEULL, "zz_approval_debug6003700.cl"},
	{0x1B8F4E4D3D938891ULL, "vxlt_0_lowtown_a2_n_c_enw_0_1.vlm"},
	{0x1B8F7C965F2FBFA2ULL, "6137560_m.fxe"},
	{0x1B90093417667114ULL, "6067098_m.fxe"},
	{0x1B902A0165672021ULL, "6107032_m.fxe"},
	{0x1B9030BD24129452ULL, "gen00fl_varric6078676.cl"},
	{0x1B90FA8FC48E3F5EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_caveent.phy"},
	{0x1B9130CF8D02C058ULL, "145453097.ogg"},
	{0x1B913A41059C0A00ULL, "vxca_caverns1_ij2_0_0.vlm"},
	{0x1B9198A244A17887ULL, "vxlt_0_lowtown_a3_n_c_db4_0_0.vlm"},
	{0x1B91FE0BAFF8A861ULL, "vxca_caverns1_mines_icy_0_0.vlm"},
	{0x1B92627BEA255105ULL, "art\\vfx\\combat\\vfx_r_dwp_left_close1_c.impa.evt"},
	{0x1B92A721F8CCAC66ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_bigrock01_d.dds"},
	{0x1B93721810CAFF16ULL, "330537957.ogg"},
	{0x1B938CC39CCA9BD4ULL, "hlac_ca_caverns1_mines_24421_3x.dds"},
	{0x1B93D2CCF34084C4ULL, "gen00flst_isa_act3_pchouse.stg"},
	{0x1B93FB15931A0138ULL, "675717256.ogg"},
	{0x1B943B783C2D7534ULL, "560652114.ogg"},
	{0x1B944BF3E4117296ULL, "730945664.ogg"},
	{0x1B94C2C80C695EE0ULL, "vxca_caverns1_eak_0_0.vlm"},
	{0x1B958EAED3759597ULL, "llac_lt_0_lowtown_n_-65536_7.dds"},
	{0x1B95974E6B409779ULL, "vxlt_gallowstemplar_n_3ra_0_0.vlm"},
	{0x1B959A0A4C92B8CCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_basic.msh"},
	{0x1B95A00A4C92C2BEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_basic.mmh"},
	{0x1B95A6562711667FULL, "hlac_lt_0_lowtown_a2_n_c_-196607_0y.dds"},
	{0x1B95AA0AEF03F607ULL, "gen_im_trs_krk_hgh_junk2_02.uti"},
	{0x1B95AC0A4C92D755ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_basic.mao"},
	{0x1B968476901F3717ULL, "character_character_p11.psh.fxo.dx11.assembly"},
	{0x1B96FCEAB18AE815ULL, "465084785.ogg"},
	{0x1B9728236B1D88B4ULL, "vxca_cavern1_haunted_ii2_0_0.vlm"},
	{0x1B976C36ADD68081ULL, "vxlt_gallowscourtyard_n_7ar_0_0.vlm"},
	{0x1B97922630B80FD4ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l.ani"},
	{0x1B97BED9890153BBULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.dura.ani"},
	{0x1B97E69F75DD4303ULL, "vxlt_0_lowtown_a2_n_gx2_0_0.vlm"},
	{0x1B98837202797DC0ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_ready.ani"},
	{0x1B9890C86B6AA339ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_6z.dds"},
	{0x1B98C5C16466F368ULL, "plt_dre120pt_reqs_money.plo"},
	{0x1B98E652DC3A72C1ULL, "997079633.ogg"},
	{0x1B991C167F387EE3ULL, "6180376_m.fxe"},
	{0x1B9966C387B4DC47ULL, "vxlt_undercityserialkiller_17o_0_0.vlm"},
	{0x1B99860A7A78BEFAULL, "vxht_bartrand_n_756_0_1.vlm"},
	{0x1B9A0FAA4B0C1ED3ULL, "6165117_m.fxe"},
	{0x1B9A37BE03030A7FULL, "6089660_m.fxe"},
	{0x1B9A46E0FB03909DULL, "gen00fl_aveline6139658.cl"},
	{0x1B9A4B9486D69254ULL, "w_blightlands_187_0.phy"},
	{0x1B9A536B82DBF4C9ULL, "vxht_0_hightown_n_cbn_0_0.vlm"},
	{0x1B9A89A0BBEA70C1ULL, "6051069_m.fxe"},
	{0x1B9A9F6CDFE1830AULL, "cod_mgc_circle_heirarch.utp"},
	{0x1B9AF07332092076ULL, "one200dg_solivitus6066518.cl"},
	{0x1B9B754D1C4CAC09ULL, "vxht_bartrand_n_5cl_0_0.vlm"},
	{0x1B9BCB181C256513ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_fail.ani"},
	{0x1B9BD22A25FF48F8ULL, "233408589.ogg"},
	{0x1B9BDD1A9DFA7E03ULL, "dae000_fenarel6134577.cl"},
	{0x1B9C5AC9A845B019ULL, "6149153_m.fxe"},
	{0x1B9C5E4DE3C2DF62ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containera_n.dds"},
	{0x1B9C6A608456ACB3ULL, "vxlt_hangedman_n_sl_0_0.vlm"},
	{0x1B9CEBD707B78B80ULL, "vxht_mansion2_31q_0_0.vlm"},
	{0x1B9CF9350D543613ULL, "zz_lgt_2_debug6211866.cl"},
	{0x1B9D2942003C878EULL, "vxht_0_hightown_c1o_0_0.vlm"},
	{0x1B9D4A89DFE46F30ULL, "vxca_caverns1_iiw_0_0.vlm"},
	{0x1B9D5427106E8DAFULL, "6207769_m.fxe"},
	{0x1B9D8A9CEDD2C2ADULL, "character_rendertodepth_p1.psh.fxo"},
	{0x1B9DF3B4D108517FULL, "6112973_m.fxe"},
	{0x1B9E30794BCDBC67ULL, "94863272.ogg"},
	{0x1B9E3E24576AD032ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track_2.mao"},
	{0x1B9E441E37E308D1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cenotaph_s.dds"},
	{0x1B9E8F2630BE2936ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l.gad"},
	{0x1B9FAC1DF7922280ULL, "vxca_caverns1_im3_0_0.vlm"},
	{0x1B9FF70931C5BB8CULL, "94580392.ogg"},
	{0x1B9FFFF798165801ULL, "lowtown_male_beggar_sta6151024.cl"},
	{0x1BA0406A6041C8C1ULL, "6074491_m.fxe"},
	{0x1BA0AA5554105531ULL, "vxca_caverns2_95q_0_0.vlm"},
	{0x1BA0AE1F2D4A9501ULL, "511064307.ogg"},
	{0x1BA0DA03323F7426ULL, "vxlt_darktown_5lu_0_0.vlm"},
	{0x1BA1938E97E50709ULL, "hlac_lt_0_lowtown_131072_3y.dds"},
	{0x1BA1BF0AB1D422E5ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lrgchair_n.dds"},
	{0x1BA1C0F3980C5A68ULL, "llac_lt_0_lowtown_docks_n_-131072_0.dds"},
	{0x1BA1CFC008B30021ULL, "fen310_fenris.bnk"},
	{0x1BA1E00CB62B9443ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls108.mao"},
	{0x1BA2224A69E4A784ULL, "270460597.ogg"},
	{0x1BA224E4713A3772ULL, "vxw_encounter2_camp_8n_0_0.vlm"},
	{0x1BA25BA906955135ULL, "293803864.ogg"},
	{0x1BA265709C7C2E85ULL, "6116827_m.fxe"},
	{0x1BA2978BE9C775F4ULL, "strcave01_d.dds"},
	{0x1BA2A1CD5B0B082AULL, "vxlt_darktown_6hv_0_0.vlm"},
	{0x1BA2ACBFB110B8E3ULL, "gen_im_cns_flu_anders_03.uti"},
	{0x1BA2E56F0A3ED04DULL, "dae000_bartrand6027500.cl"},
	{0x1BA324EFBFC52AB2ULL, "6033722_f.fxe"},
	{0x1BA3545379641F05ULL, "663310518.ogg"},
	{0x1BA37239D58F1056ULL, "hlac_ht_bartrand_n_3371_1x.dds"},
	{0x1BA3EEDC042D2458ULL, "594071356.ogg"},
	{0x1BA409E9F7298CC7ULL, "100522416.ogg"},
	{0x1BA42CED70AF8C4DULL, "art\\characters\\weapons\\mauls\\w_mal_ml103a.phy"},
	{0x1BA49B12D4D422E3ULL, "6041685_f.fxe"},
	{0x1BA4AA2630C2C295ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l.evt"},
	{0x1BA4C6D9890BEBB2ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.dura.evt"},
	{0x1BA4F43182421C8DULL, "228526446.ogg"},
	{0x1BA4F9DC2DFB6BE5ULL, "6107065_m.fxe"},
	{0x1BA5289486E05E51ULL, "w_blightlands_187_0.msh"},
	{0x1BA5369486E0766CULL, "w_blightlands_187_0.mao"},
	{0x1BA5429486E08A0FULL, "w_blightlands_187_0.mmh"},
	{0x1BA5498A4986C654ULL, "6116853_m.fxe"},
	{0x1BA55990F1CDEF99ULL, "art\\cinematics\\_crit path\\followerintros\\cs_player_merillintro.gad"},
	{0x1BA5AE9A5E0E6004ULL, "220514525.ogg"},
	{0x1BA5F1C008B72291ULL, "fen310_fenris.cnv"},
	{0x1BA61337586A3E62ULL, "ts_kirkwall_arm_table.gda"},
	{0x1BA660C2FAE0B62FULL, "6120037_m.fxe"},
	{0x1BA69C1F0A844EC8ULL, "vxht_mansion2b_35m_0_0.vlm"},
	{0x1BA6A1CB2862B0E5ULL, "vxht_keep_3zu_0_0.vlm"},
	{0x1BA86901F48B1505ULL, "hlac_lt_hangedman_n_-1_29x.dds"},
	{0x1BA89D9B9C3CA90CULL, "one280_iwan6096892.cl"},
	{0x1BA8B3F03AFC3B19ULL, "war_chancetostagger.dds"},
	{0x1BA904A34782FCC8ULL, "119383267.ogg"},
	{0x1BA9C3181C31954AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_fail.evt"},
	{0x1BA9D2B8A80AD9D2ULL, "sib131_slaver_lord6098064.cl"},
	{0x1BAA46C7591034FFULL, "fxa_cen_exo_smoke_blend.dds"},
	{0x1BABD48D87DCF397ULL, "dae000ip_crowd_ff5_talk5.utp"},
	{0x1BABE5D989123F45ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.dura.gad"},
	{0x1BAC5941424379A5ULL, "vxlt_0_lowtown_a2_n_e28_0_0.vlm"},
	{0x1BAC6A90F1D42BE7ULL, "art\\cinematics\\_crit path\\followerintros\\cs_player_merillintro.ani"},
	{0x1BACCD8B82FE2BC4ULL, "vxlt_undercityserialkiller_1lp_0_1.vlm"},
	{0x1BAD1DFCA3B8A068ULL, "dae000_bartrand6206874.cl"},
	{0x1BAD6CD532F4D975ULL, "99370450.ogg"},
	{0x1BADDB414A22920BULL, "6076312_m.fxe"},
	{0x1BAE87841CA5C25DULL, "vxht_0_hightown_a2_bjx_0_0.vlm"},
	{0x1BAEDB81A9FEAA8BULL, "lgt220_shp_mag_guard6149909.cl"},
	{0x1BAF4BE024358CA6ULL, "vxlt_darktown_5qv_0_0.vlm"},
	{0x1BB0602A67F81DF3ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_2z.dds"},
	{0x1BB096C163304FF2ULL, "vxw_encounter2_camp_3o_0_0.vlm"},
	{0x1BB0A2181C377BBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_fail.gad"},
	{0x1BB11FFD942C1C39ULL, "267241133.ogg"},
	{0x1BB1370F349F80EAULL, "6156275_m.fxe"},
	{0x1BB1A2BDE936D727ULL, "ca_caverns1_24426_ll.mmh"},
	{0x1BB1BE8E7F572515ULL, "vxdr_primevaltomb_526_0_0.vlm"},
	{0x1BB1E872C72D3F9FULL, "6164402_m.fxe"},
	{0x1BB24286E0B3B0ADULL, "6187784_m.fxe"},
	{0x1BB2EF876C0A3DFCULL, "729192710.ogg"},
	{0x1BB39D4CE6E187CEULL, "6153562_m.fxe"},
	{0x1BB4BA1579F65B11ULL, "vxht_0_hightown_90e_0_0.vlm"},
	{0x1BB4BAB25EB8E0B5ULL, "6160835_m.fxe"},
	{0x1BB50974D30CFE0AULL, "hlac_lt_0_lowtown_docks_-1_3y.dds"},
	{0x1BB53433A7281350ULL, "hlac_ca_caverns1_24425_1y.dds"},
	{0x1BB5D41F7460D201ULL, "vxca_caverns1_mines_j4g_0_0.vlm"},
	{0x1BB5F550CCAE61B6ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_wkf.ani"},
	{0x1BB64232B38E72F3ULL, "325694549.ogg"},
	{0x1BB6BB5363EADB7BULL, "one110cr_coterie_thug_1.utc"},
	{0x1BB75D63936539CEULL, "hlac_lt_0_lowtown_docks_a3_n_131070_0z.dds"},
	{0x1BB7CD0ED58DACBEULL, "dre000_vampire6034289.cl"},
	{0x1BB7E66A806246B2ULL, "one300_nuncio6084004.cl"},
	{0x1BB82616CA66914FULL, "vxca_caverns1_i7v_0_0.vlm"},
	{0x1BB8862A305363BCULL, "plt_mag200pt_nt_ambient.plo"},
	{0x1BB8D05AF439DC7AULL, "qcr000_varric6208264.cl"},
	{0x1BB903B83565CAB2ULL, "hlac_ht_store_187_3z.dds"},
	{0x1BB932FABCD694D8ULL, "513411807.ogg"},
	{0x1BB934D59F529DEDULL, "6025088_m.fxe"},
	{0x1BB97949D689D1F3ULL, "mrl150_merrill6070799.cl"},
	{0x1BB9B10B69CEF75BULL, "docks_male_overseer_mob6176549.cl"},
	{0x1BBA138CF6FB57CBULL, "6043226_m.fxe"},
	{0x1BBA175CA98C3AC8ULL, "6174228_m.fxe"},
	{0x1BBA8C6ED101C7A5ULL, "753180513.ogg"},
	{0x1BBAC2C16483D427ULL, "plt_dre120pt_reqs_money.nss"},
	{0x1BBAD93EA024991FULL, "llac_lt_undercityserialkiller_-1_52.dds"},
	{0x1BBB23B60E3FEAECULL, "mag101_grace6034027.cl"},
	{0x1BBC0C3D88A08780ULL, "6062053_m.fxe"},
	{0x1BBC3A98B66306B6ULL, "skin_depth.vsh.fxo"},
	{0x1BBC51F9C77C43C0ULL, "zz_mcr_debug6088434.cl"},
	{0x1BBC6D957471C5A3ULL, "830666570.ogg"},
	{0x1BBCD6566FA6F879ULL, "static.mat.dx11"},
	{0x1BBCF9610E9BDADDULL, "vxht_0_hightown_a2_bow_0_0.vlm"},
	{0x1BBCFFB81868F5CEULL, "vxlt_hangedman_1td_0_0.vlm"},
	{0x1BBDA2AACC81B68AULL, "pro000_colonel_ewald6040587.cl"},
	{0x1BBE1AD1CAF82659ULL, "vxlt_0_lowtown_a2_n_fgt_0_0.vlm"},
	{0x1BBE73793675C54AULL, "vxlt_0_lowtown_docks_b5f_0_1.vlm"},
	{0x1BBEC4DBB6CC3AC2ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumtile_r.dds"},
	{0x1BBF196F2492666EULL, "combat_lc_r1_a5.gda"},
	{0x1BBF68C671AFEBB2ULL, "726276213.ogg"},
	{0x1BBF86DA92242C73ULL, "414769251.ogg"},
	{0x1BBFBD9F051D38E6ULL, "6087133_m.fxe"},
	{0x1BBFBE7D90B4FFCEULL, "vxht_mansion_n_5mf_0_0.vlm"},
	{0x1BBFC56F7DF573D8ULL, "follower_banter6165472.cl"},
	{0x1BC055799342BEDEULL, "hlac_ht_chantry_2534_48y.dds"},
	{0x1BC0D143B11C9CDBULL, "hlac_lt_blackemporium_1055_16z.dds"},
	{0x1BC0D55A9B54F47AULL, "6100074_m.fxe"},
	{0x1BC2F27DB1B17DCCULL, "w_encounter1_9_0.msh"},
	{0x1BC2F87DB1B187BEULL, "w_encounter1_9_0.mmh"},
	{0x1BC3047DB1B19C55ULL, "w_encounter1_9_0.mao"},
	{0x1BC3401CD039833BULL, "vxlt_gallowscourtyard_n_75t_0_0.vlm"},
	{0x1BC372A29E2A8727ULL, "27657077.ogg"},
	{0x1BC3E1AB179D4AEDULL, "llac_lt_gallowstemplar_2391_0.dds"},
	{0x1BC3E6863346826CULL, "vxlt_backalley_g_1e6_0_0.vlm"},
	{0x1BC3FD50CCBAACA7ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_wkf.evt"},
	{0x1BC443BE1F2B5170ULL, "vxdr_primevalentrance_1qr_0_0.vlm"},
	{0x1BC4C9EB0408AAE1ULL, "vxlt_0_lowtown_docks_n_b7t_0_1.vlm"},
	{0x1BC4FE9F41ED40D0ULL, "vxht_0_hightown_a2_aqs_0_0.vlm"},
	{0x1BC5E477B5017E00ULL, "vxlt_warehouse_za_0_0.vlm"},
	{0x1BC6487DD8785F5CULL, "6168593_m.fxe"},
	{0x1BC64FE3B2BD1F3AULL, "vxlt_gallowstemplar_f_6q6_0_0.vlm"},
	{0x1BC711AAE8A2D763ULL, "191994936.ogg"},
	{0x1BC71F875256CC00ULL, "2182542.ogg"},
	{0x1BC7E53846414F85ULL, "gen00fl_anders6096503.cl"},
	{0x1BC7F75EAA77FBBEULL, "vxdr_primevaltomb_2uo_0_0.vlm"},
	{0x1BC81AE30D279EEAULL, "w_blightlands_145_0.mao"},
	{0x1BC81EE30D27A539ULL, "w_blightlands_145_0.mmh"},
	{0x1BC828E30D27B627ULL, "w_blightlands_145_0.msh"},
	{0x1BC8AED6A74987B6ULL, "1001227717.ogg"},
	{0x1BC8E59D13E59F29ULL, "6079693_f.fxe"},
	{0x1BC92E15A1CB88E8ULL, "art\\levels\\ztools\\default\\generic\\3wo_basic.msh"},
	{0x1BC94015A1CBA779ULL, "art\\levels\\ztools\\default\\generic\\3wo_basic.mao"},
	{0x1BC943EC0A15DE90ULL, "llac_lt_foundry_n_320_9.dds"},
	{0x1BC94C15A1CBBB12ULL, "art\\levels\\ztools\\default\\generic\\3wo_basic.mmh"},
	{0x1BC9DD9F77246BD2ULL, "llac_ht_0_hightown_a2_-196608_2.dds"},
	{0x1BCA1E50CCBF50F0ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_wkf.gad"},
	{0x1BCAA4B2CCA23A20ULL, "186523032.ogg"},
	{0x1BCAC7087B357514ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_basic.phy"},
	{0x1BCAD3724B3B16D4ULL, "hlac_ca_caverns1_24426_12x.dds"},
	{0x1BCB19F7D6725494ULL, "557411393.ogg"},
	{0x1BCB362F3B83CBF1ULL, "6076651_m.fxe"},
	{0x1BCB71950A5F0E4EULL, "vxlt_hangedman_1se_0_0.vlm"},
	{0x1BCD06D8090A008DULL, "334211160.ogg"},
	{0x1BCD32D5FEF71DC9ULL, "vxlt_0_lowtown_docks_a3_czf_0_1.vlm"},
	{0x1BCD76A0E7147002ULL, "vxca_caverns1_mines_ei8_0_0.vlm"},
	{0x1BCDC7CD314F02C5ULL, "6148161_m.fxe"},
	{0x1BCDCB12F206409BULL, "zz_lgt_1_debug6188680.cl"},
	{0x1BCF09B3A3FC1466ULL, "6189011_m.fxe"},
	{0x1BCF4B5F764CDD9BULL, "llac_lt_0_lowtown_docks_a3_n_-3_7.dds"},
	{0x1BCFB4FC908782ADULL, "6014259_m.fxe"},
	{0x1BD00E07F806E8FAULL, "6067849_m.fxe"},
	{0x1BD043F7320600E0ULL, "1003980801.ogg"},
	{0x1BD083F043206E58ULL, "hlac_lt_0_lowtown_a2_-131072_0z.dds"},
	{0x1BD08D37CCA5BEA4ULL, "6093541_m.fxe"},
	{0x1BD0DC4D8CFCFE10ULL, "6068526_f.fxe"},
	{0x1BD1311E7F886D73ULL, "gen00fl_isabela_init6077305.cl"},
	{0x1BD1859421FB1A76ULL, "one300_zevran6089744.cl"},
	{0x1BD1A4AE8C241694ULL, "isa211_walleyedsam.cnv"},
	{0x1BD1C6D93BFF8E58ULL, "526497526.ogg"},
	{0x1BD3521BD82C45A2ULL, "zz_mag_debug6039792.cl"},
	{0x1BD364D5B5D63A8AULL, "vxw_bonepitmining_a2_7u_0_0.vlm"},
	{0x1BD3707C33E35950ULL, "vxht_0_hightown_a2_alp_0_0.vlm"},
	{0x1BD3A89844257C13ULL, "hlac_dr_primevaltomb_659_11y.dds"},
	{0x1BD4077C29C708E9ULL, "44140783.ogg"},
	{0x1BD4A4087B3D8E11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_basic.msh"},
	{0x1BD4B2087B3DA62CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_basic.mao"},
	{0x1BD4BE087B3DBACFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_basic.mmh"},
	{0x1BD4DEF3C50649C4ULL, "vxlt_0_lowtown_docks_n_bd5_0_0.vlm"},
	{0x1BD4EA7A0DBCEA2AULL, "15530231.ogg"},
	{0x1BD5867692DC5E42ULL, "llac_lt_gallowscourtyard_-65538_4.dds"},
	{0x1BD586AE8C27CCD4ULL, "isa211_walleyedsam.bnk"},
	{0x1BD5DDB23E362FADULL, "mag110_samson.bnk"},
	{0x1BD5E185FC8E29BEULL, "6119252_m.fxe"},
	{0x1BD6770A4CCA1459ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_basic.phy"},
	{0x1BD72C3F603244E5ULL, "448289729.ogg"},
	{0x1BD7B452A6FD35F8ULL, "zz_followers_banter6158676.cl"},
	{0x1BD7B8A2E1ED8616ULL, "969677769.ogg"},
	{0x1BD7BD1C7FF56B1EULL, "6132879_m.fxe"},
	{0x1BD8CAE19E812EADULL, "mag321ip_launcet_to_hitown.utp"},
	{0x1BD8FFB23E389F15ULL, "mag110_samson.cnv"},
	{0x1BDA0D374888C487ULL, "vxdr_primevaltomb_6qp_0_2.vlm"},
	{0x1BDA242C31CF383EULL, "vxht_playershouse_n_1zn_0_0.vlm"},
	{0x1BDA628B36A569C6ULL, "259842574.ogg"},
	{0x1BDA76082FC937D3ULL, "lowtown_male_urchin_sta6164216.cl"},
	{0x1BDA9CAF08E7171BULL, "art\\vfx\\combat\\vfx_gib_qunw_rtlb.anb"},
	{0x1BDAC013E687A24EULL, "6087032_m.fxe"},
	{0x1BDAD0D848F1C31AULL, "vxlt_gallowscourtyard_4lj_0_0.vlm"},
	{0x1BDB44D00522EAEEULL, "vxca_caverns1_mines_gk8_0_0.vlm"},
	{0x1BDC629AC5B1CC03ULL, "montage_subs_demohack6217126.cl"},
	{0x1BDCF3BF43C8B9DEULL, "6120003_m.fxe"},
	{0x1BDD074E1138F7BFULL, "hlac_dr_dungeon_3216_6y.dds"},
	{0x1BDD53515D2ED497ULL, "llac_lt_gallowstemplar_n_2384_1.dds"},
	{0x1BDD5DD9969A82FCULL, "59262110.ogg"},
	{0x1BDDD04B6B65C1A9ULL, "207366320.ogg"},
	{0x1BDE70371EDA3830ULL, "vxlt_gallowscourtyard_62l_0_0.vlm"},
	{0x1BDE9429060C3DFAULL, "168636819.ogg"},
	{0x1BDE96F82C86FCEDULL, "dae000_arishok6067671.cl"},
	{0x1BDEE3438EAEDC7AULL, "6161448_m.fxe"},
	{0x1BDF5420B2E347C9ULL, "vxw_bonepitdestroyed_a3_1g3_0_0.vlm"},
	{0x1BDF5E1259559F3BULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_foam02.mao"},
	{0x1BDF64B9A7126AEBULL, "835073557.ogg"},
	{0x1BDF98B95BD1B87BULL, "hlac_lt_backalley_g_65536_2x.dds"},
	{0x1BDFB0FF55BE84AEULL, "qun110_petrice6037399.cl"},
	{0x1BDFE14FD760E839ULL, "spk_spk.mao"},
	{0x1BE0F7A96D2A458FULL, "art\\characters\\creatures\\pridedemon\\animation\\blendtree\\pridedemon_overlay.anb"},
	{0x1BE1277803178270ULL, "vxdr_primevalentrance_1kp_0_0.vlm"},
	{0x1BE152E602F27813ULL, "mcr310cr_confused_guard_ldr.utc"},
	{0x1BE21C0E5CBE6BF7ULL, "one100dg_vanard6047410.cl"},
	{0x1BE26B325C2082AFULL, "vxlt_foundry_n_10d_0_0.vlm"},
	{0x1BE2AE20691E8286ULL, "llac_w_encounter1_65538_0.dds"},
	{0x1BE2CD7859C9AD1CULL, "18006731.ogg"},
	{0x1BE332AE1E7B8B33ULL, "vxht_mansion_n_5bb_0_0.vlm"},
	{0x1BE3363157AE9CFDULL, "136037636.ogg"},
	{0x1BE3939725D5E839ULL, "vxht_0_hightown_a3_boz_0_0.vlm"},
	{0x1BE3CA6C81F2A22CULL, "isa000pt_main.ncs"},
	{0x1BE4431A220151B3ULL, "narrative_varric_demo6170825.cl"},
	{0x1BE48857C56BB99DULL, "plt_cod_bks_chantry_hist_3.plo"},
	{0x1BE5513B629D8CDDULL, "vxlt_0_lowtown_a2_hfv_0_0.vlm"},
	{0x1BE5898BD014A02FULL, "dae000_madam_lusine6152164.cl"},
	{0x1BE5C203C2F25D23ULL, "643342405.ogg"},
	{0x1BE5CF376F9CD26BULL, "204837515.ogg"},
	{0x1BE5D58070295FAAULL, "360406526.ogg"},
	{0x1BE5E37BE0CFEE23ULL, "w_bonepit_153_0.msh"},
	{0x1BE5F17BE0D0067DULL, "w_bonepit_153_0.mmh"},
	{0x1BE5F57BE0D00C8EULL, "w_bonepit_153_0.mao"},
	{0x1BE64BEA3F753BE3ULL, "follower_banter6141686.cl"},
	{0x1BE66BAF50E59B1CULL, "mag100_cullen6140917.cl"},
	{0x1BE6A1F289B781D2ULL, "115285992.ogg"},
	{0x1BE6C922BA45BB17ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_8.dds"},
	{0x1BE7FC582F280A4EULL, "vxlt_0_lowtown_f27_0_0.vlm"},
	{0x1BE8960923C550BEULL, "vxht_playershouse_n_1qo_0_0.vlm"},
	{0x1BE8D46F96900944ULL, "6092054_m.fxe"},
	{0x1BE8FD7317F05A5AULL, "vxw_bonepit_lb_0_0.vlm"},
	{0x1BE8FFD435BFA7F7ULL, "character_rendertodepth_p9.psh.fxo.dx11.assembly"},
	{0x1BE952B3505F44B8ULL, "hlac_lt_0_lowtown_docks_-65538_0z.dds"},
	{0x1BEA29F6F72EC79FULL, "vxw_woundedcoast_14c_0_0.vlm"},
	{0x1BEA47F5293D0734ULL, "vxdr_primevaltomb_875_0_0.vlm"},
	{0x1BEA9155201BDE6FULL, "vxw_woundedcoast_h6_0_0.vlm"},
	{0x1BEAADAF35F5D0E0ULL, "dae000_bartrand6030680.cl"},
	{0x1BEAB278B9FB5800ULL, "6041546_m.fxe"},
	{0x1BEAFA354272E70AULL, "657870111.ogg"},
	{0x1BEB1E354FF96DF1ULL, "vxlt_0_lowtown_a2_d1w_0_0.vlm"},
	{0x1BEBA45E68FDA464ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench1_arishokaddress.gad"},
	{0x1BEBDE19BBF663EFULL, "782320085.ogg"},
	{0x1BEC2BF032FCEC22ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_enter.gad"},
	{0x1BEC399D0260857DULL, "hlac_lt_undercityserialkiller_-1_35y.dds"},
	{0x1BECE8EF2C879BFEULL, "333240250.ogg"},
	{0x1BECFB2B76033CB7ULL, "dre180ar_ruins_entrance.ncs"},
	{0x1BEE9E2A37BAA4BFULL, "vxlt_backalley_g_1e9_0_0.vlm"},
	{0x1BEEA95A4890654BULL, "zz_approval_debug6003556.cl"},
	{0x1BEF09FE1EC24860ULL, "llac_lt_undercitydungeon_-1_41.dds"},
	{0x1BEF361F62961FF1ULL, "hlac_lt_0_lowtown_a2_-196609_0z.dds"},
	{0x1BEF5984E4A9D250ULL, "hlac_w_sundermount_65540_0y.dds"},
	{0x1BF0F561AA449F1BULL, "vxht_keep_n_647_0_0.vlm"},
	{0x1BF1007B919CBE15ULL, "llac_lt_0_lowtown_a2_131072_2.dds"},
	{0x1BF16768138C84EDULL, "rog_accuracymode.dds"},
	{0x1BF23BAA5ED56ECBULL, "character_rendertodepth_p1.psh.fxo.dx11"},
	{0x1BF2927B3CF60477ULL, "6173730_m.fxe"},
	{0x1BF29E658B07B59CULL, "vxlt_0_lowtown_a2_db3_0_0.vlm"},
	{0x1BF2AFE3DE3FBDEBULL, "6176504_m.fxe"},
	{0x1BF2B35E6903DC42ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench1_arishokaddress.ani"},
	{0x1BF34EF033034619ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_enter.evt"},
	{0x1BF37325DA6078ACULL, "hlac_lt_hangedman_n_-1_14y.dds"},
	{0x1BF3C3185493A55DULL, "vxlt_0_lowtown_a2_hcu_0_0.vlm"},
	{0x1BF3DC6F9E806BFFULL, "vxlt_0_lowtown_f48_0_0.vlm"},
	{0x1BF3FB2B76095B06ULL, "dre180ar_ruins_entrance.lst"},
	{0x1BF45B0070D4EDEAULL, "hlac_lt_0_lowtown_docks_n_-2_0z.dds"},
	{0x1BF4A8A13D485797ULL, "vxdr_0_deeproads_cwc_0_0.vlm"},
	{0x1BF4FCE04E80CD83ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_1.ani"},
	{0x1BF5E2D751965A64ULL, "mag120_macha6080649.cl"},
	{0x1BF60283DAA87002ULL, "194590756.ogg"},
	{0x1BF653800EBD7E6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_behind.ani"},
	{0x1BF66C162C310818ULL, "hlac_ht_playershouse_n_855_9z.dds"},
	{0x1BF67E2E364BC69AULL, "6066712_m.fxe"},
	{0x1BF6A1CBF767FB91ULL, "ave121_brennan6105893.cl"},
	{0x1BF6B858110C2B16ULL, "326843509.ogg"},
	{0x1BF6EAE42597A0D8ULL, "184747834.ogg"},
	{0x1BF7C867D514DAE6ULL, "hlac_ht_keep_3731_2y.dds"},
	{0x1BF8CE777F5FA8CAULL, "833304805.ogg"},
	{0x1BF903321211F6EFULL, "vxw_woundedcoast_e1_0_0.vlm"},
	{0x1BF9D116D90A828BULL, "vxlt_gallowscourtyard_6xq_0_0.vlm"},
	{0x1BFADA5D4762728DULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_bdy_suca_0n.dds"},
	{0x1BFADC9FCEA0B937ULL, "6062818_m.fxe"},
	{0x1BFADFD9EE3A18A7ULL, "art\\levels\\races\\design\\_textures\\plc_ballista_01_0_sa.dds"},
	{0x1BFBC192B5244305ULL, "atmosphere.gda"},
	{0x1BFC37DA96CF78C9ULL, "vxw_bonepitdestroyed_a3_1m1_0_0.vlm"},
	{0x1BFC3F80C41EE912ULL, "6149635_m.fxe"},
	{0x1BFCD2EA84463C8FULL, "485309453.ogg"},
	{0x1BFD0F853812431CULL, "77574390.ogg"},
	{0x1BFDDA12FFDCCC1BULL, "llac_lt_darktown_-1_34.dds"},
	{0x1BFE0A2B7611C911ULL, "dre180ar_ruins_entrance.are"},
	{0x1BFE5163FC798492ULL, "one100dg_vanard6048269.cl"},
	{0x1BFE83CE256F87C7ULL, "vxlt_undercityserialkiller_1jd_0_0.vlm"},
	{0x1BFED4DAA32D8650ULL, "397768956.ogg"},
	{0x1BFEEA735A126976ULL, "one110_gustav6089813.cl"},
	{0x1BFEF836ED2FD0D0ULL, "6201539_m.fxe"},
	{0x1BFFB42C9F527E06ULL, "281068296.ogg"},
	{0x1C00102FE26F4575ULL, "vxht_0_hightown_a2_a2o_0_0.vlm"},
	{0x1C0016680267BC33ULL, "vxht_mansion_n_5hd_0_0.vlm"},
	{0x1C00397DEA45957AULL, "llac_ca_caverns1_mines_24424_6.dds"},
	{0x1C0056F0330DDE08ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_enter.ani"},
	{0x1C005B840A442BFFULL, "ht_0_hightown_21_0.rml"},
	{0x1C00775109C21939ULL, "vxht_0_hightown_a3_bax_0_0.vlm"},
	{0x1C00A492A8BC91B7ULL, "vxca_caverns2prc_7ve_0_0.vlm"},
	{0x1C00E58B660712D4ULL, "6044192_m.fxe"},
	{0x1C01C2506958D24AULL, "6100287_m.fxe"},
	{0x1C01D3FD4787489DULL, "vxlt_0_lowtown_a2_n_c_hjo_0_0.vlm"},
	{0x1C0214E04E8B817AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_1.evt"},
	{0x1C033B800EC7DFE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_behind.evt"},
	{0x1C034C48FE06E9BDULL, "ht_0_hightown_a2_16_0_ll.mmh"},
	{0x1C03AE67B688CD6EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door.phy"},
	{0x1C03CF1833B41933ULL, "follower_banter6141549.cl"},
	{0x1C03CF7DB1E8D959ULL, "w_encounter1_9_0.phy"},
	{0x1C04363DD6603415ULL, "vxlt_backalley_n_1w9_0_1.vlm"},
	{0x1C04B93F38AEDF0CULL, "gen00fl_varric6102801.cl"},
	{0x1C05523B67202D6BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrlb_0d.dds"},
	{0x1C056705D974AB83ULL, "6129078_m.fxe"},
	{0x1C0579F92BDE6E0EULL, "llac_lt_darktown_-1_88.dds"},
	{0x1C059173E192C3FFULL, "art\\levels\\races\\elf\\props\\alienage\\prp_fire.msh"},
	{0x1C059773E192CDA1ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_fire.mmh"},
	{0x1C05A373E192E242ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_fire.mao"},
	{0x1C061C22926A60E2ULL, "vxca_caverns1_mines_gez_0_0.vlm"},
	{0x1C066D41E057E65CULL, "vxca_caverns1_imo_0_0.vlm"},
	{0x1C06806C3D66F5CEULL, "199095296.ogg"},
	{0x1C071513D094DB95ULL, "155697750.ogg"},
	{0x1C074865FA79F8E9ULL, "6109218_m.fxe"},
	{0x1C076AC40C0F94E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_wall_move_item.ani"},
	{0x1C081306EBFB62EEULL, "6013446_m.fxe"},
	{0x1C087D4444F69B02ULL, "vxht_mansion_n_53t_0_0.vlm"},
	{0x1C08AB5CEAA1F982ULL, "vxlt_hangedman_2o2_0_0.vlm"},
	{0x1C08D80238C6140FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_down.gad"},
	{0x1C08E8974C02D403ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90pause.ani"},
	{0x1C08EF590741AC15ULL, "592577596.ogg"},
	{0x1C0913E04E919E6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_1.gad"},
	{0x1C098ED9F55C0F99ULL, "vrc221_varric6043934.cl"},
	{0x1C09D87DB1ED5475ULL, "w_encounter1_9_0.rml"},
	{0x1C0A395EE4A82D94ULL, "vxdr_0_deeproads_gc4_0_0.vlm"},
	{0x1C0A3A800ECDFCF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_behind.gad"},
	{0x1C0A4DCB47CA62E4ULL, "6065917_m.fxe"},
	{0x1C0B19D366349B85ULL, "6153389_m.fxe"},
	{0x1C0B655742DC8B22ULL, "art\\characters\\creatures\\spider\\textures\\c_poispidra_0t.dds"},
	{0x1C0BC3C1A1F9A499ULL, "ser221_gascard6165985.cl"},
	{0x1C0BCBDCE28D8A00ULL, "hlac_lt_backalley_n_0_12x.dds"},
	{0x1C0BE342A516385AULL, "1015453830.ogg"},
	{0x1C0C5320256C292FULL, "6085739_f.fxe"},
	{0x1C0C7B8DCD9B8508ULL, "6108999_m.fxe"},
	{0x1C0CD5CD2FD4963DULL, "6095974_f.fxe"},
	{0x1C0CE324E739E991ULL, "vxht_playershouse_n_204_0_0.vlm"},
	{0x1C0CFFBEAF824DF2ULL, "hlac_w_woundedcoast_65541_0z.dds"},
	{0x1C0D04C92E5F168EULL, "dae010_chase_thief.utc"},
	{0x1C0D5A3FC872A427ULL, "vxht_mansion2b_1w6_0_0.vlm"},
	{0x1C0D70AC5CD1FB6DULL, "6168156_m.fxe"},
	{0x1C0D828BA2E74E1FULL, "hlac_lt_0_lowtown_docks_a3_196607_1z.dds"},
	{0x1C0E2223AF645909ULL, "l_dae280ar_deep_roads.bnk"},
	{0x1C0EED0238CAA3BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_down.evt"},
	{0x1C0F17F92A25D053ULL, "vxlt_gallowscourtyard_7fr_0_0.vlm"},
	{0x1C0F27A7B4707944ULL, "one100st_elren.stg"},
	{0x1C0FB01D918F2ADBULL, "6015823_m.fxe"},
	{0x1C0FF9B23E67D7C6ULL, "mag110_samson.stg"},
	{0x1C106AE53B2CBD66ULL, "6208692_f.fxe"},
	{0x1C10921F57D13F5EULL, "mcr370_moira6087499.cl"},
	{0x1C10B6547529425FULL, "hlac_lt_0_lowtown_a3_n_c_131073_1x.dds"},
	{0x1C1107318936D7AEULL, "llac_lt_0_lowtown_docks_a3_n_131071_5.dds"},
	{0x1C1127A55E953BD3ULL, "6165324_m.fxe"},
	{0x1C11567F5ADEB5B5ULL, "sto_sm.mao"},
	{0x1C11DC352DA7BD27ULL, "302857433.ogg"},
	{0x1C1261FDA0164900ULL, "vxw_encounter3_6q_0_0.vlm"},
	{0x1C12A2CB35750BACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\originals\\enter_walk.ani"},
	{0x1C130C8FDDA34260ULL, "dae00cr_madam_lusine.utc"},
	{0x1C132524F3B31DD6ULL, "656663673.ogg"},
	{0x1C132C8B891A3BEFULL, "hlac_w_bonepitdestroyed_a3_327683_0z.dds"},
	{0x1C13AAE5C42D88CEULL, "vxlt_hangedman_1jz_0_0.vlm"},
	{0x1C1419DFD21166B7ULL, "dae000_hate_follower_react6167472.cl"},
	{0x1C145D3F979C3E2DULL, "vfx_dlc2_lightning_ball.wbd"},
	{0x1C14FF515CCE432AULL, "474715581.ogg"},
	{0x1C154CD7C3E71EE9ULL, "vxlt_0_lowtown_n_eqa_0_0.vlm"},
	{0x1C1552C40C1BA957ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_wall_move_item.evt"},
	{0x1C15D9532D1776D4ULL, "vxlt_0_lowtown_docks_b4j_0_0.vlm"},
	{0x1C1600974C0D87FAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90pause.evt"},
	{0x1C166105C2427ADFULL, "85166735.ogg"},
	{0x1C16B281F956135DULL, "hlac_lt_0_lowtown_a2_n_c_65537_4y.dds"},
	{0x1C16E03B529036D3ULL, "hlac_lt_0_lowtown_docks_-131074_0z.dds"},
	{0x1C171D39DC981202ULL, "vxlt_hangedman_2d3_0_0.vlm"},
	{0x1C176F4F801AA7D9ULL, "6163607_m.fxe"},
	{0x1C17BB36140E479EULL, "mag121_eau_banter6121088.cl"},
	{0x1C17E859682122C8ULL, "55120702.ogg"},
	{0x1C17FD823437BCF2ULL, "185638252.ogg"},
	{0x1C181BE60AA99D24ULL, "6208635_m.fxe"},
	{0x1C182A0886C2320DULL, "vfx_sparkae_02.dds"},
	{0x1C1876FFD21A8832ULL, "340668896.ogg"},
	{0x1C18784077C71939ULL, "hlac_dr_primevaltomb_659_10x.dds"},
	{0x1C1898AEEB7A6ED2ULL, "vxlt_gallowscourtyard_7fw_0_0.vlm"},
	{0x1C18AB3BD69E4614ULL, "vxdr_0_deeproads_gh5_0_0.vlm"},
	{0x1C18C5FF04D7C459ULL, "t3_eye_gry.tnt"},
	{0x1C18CF40DB8E107FULL, "664435847.ogg"},
	{0x1C1944084736C88BULL, "6172614_m.fxe"},
	{0x1C1955D696D54F7EULL, "vxw_bonepitdestroyed_a3_9m_0_0.vlm"},
	{0x1C197F66215411E3ULL, "768029270.ogg"},
	{0x1C19AFCB357B40EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\originals\\enter_walk.gad"},
	{0x1C19DB7FFD60B6D0ULL, "hlac_ht_keep_3743_0x.dds"},
	{0x1C19F1386227AE38ULL, "llac_lt_backalley_g_65537_4.dds"},
	{0x1C1A0031F65A82B2ULL, "vxlt_foundry_n_vp_0_0.vlm"},
	{0x1C1B744D7741E2BFULL, "6078128_f.fxe"},
	{0x1C1C53C40C21C9E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_wall_move_item.gad"},
	{0x1C1C5D44F08BD397ULL, "vxht_chantry_2ec_0_0.vlm"},
	{0x1C1CF50238D6EED5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_down.ani"},
	{0x1C1CFF974C13A4EDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90pause.gad"},
	{0x1C1D33E5313E6BAEULL, "dae000_orsino6085459.cl"},
	{0x1C1DCFE9110B79D2ULL, "dae000ip_crowd_ff6_talk14.utp"},
	{0x1C1DF9FF43ADFCD9ULL, "dae200_bran6144556.cl"},
	{0x1C1E0A32E1C84D43ULL, "297194201.ogg"},
	{0x1C1E993CF039E4F4ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityexit.ani"},
	{0x1C1F19F0DA771B63ULL, "vxlt_gallowscourtyard_4wf_0_0.vlm"},
	{0x1C1F449BB9A39C21ULL, "mrl150_hunter6071760.cl"},
	{0x1C1FEEE3F9B85DDEULL, "718206460.ogg"},
	{0x1C2044037EA61F67ULL, "vxht_chantry_n_28g_0_0.vlm"},
	{0x1C205C784C583321ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_down.ani"},
	{0x1C20AACB358156CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\originals\\enter_walk.evt"},
	{0x1C20FDD11C171631ULL, "vxlt_gallowstemplar_n_3ry_0_0.vlm"},
	{0x1C2146E9668C59E6ULL, "llac_lt_0_lowtown_a2_n_-131071_2.dds"},
	{0x1C21C584CEF384DCULL, "9503072.ogg"},
	{0x1C228E620F712176ULL, "vxht_0_hightown_n_bx2_0_0.vlm"},
	{0x1C22F519A52EBD9BULL, "6092195_f.fxe"},
	{0x1C232A379E194839ULL, "607053523.ogg"},
	{0x1C23454AC1B4BB3EULL, "6154057_m.fxe"},
	{0x1C235F940030CAF2ULL, "t1_skn_002.tnt"},
	{0x1C247F618210A371ULL, "killallhostiles.ncs"},
	{0x1C24963CF03E4B56ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityexit.gad"},
	{0x1C249D7996F49A53ULL, "vxca_caverns1_hua_0_0.vlm"},
	{0x1C254B15A21A180DULL, "art\\levels\\ztools\\default\\generic\\3wo_basic.phy"},
	{0x1C2560FE28E2CC02ULL, "vxht_mansion_n_59v_0_0.vlm"},
	{0x1C256B5EE9EA8EC9ULL, "vxdr_primevaltomb_8z3_0_0.vlm"},
	{0x1C25CA913C1A5CDBULL, "zz_ave_debug6056784.cl"},
	{0x1C25DD4735E0EBFCULL, "llac_lt_0_lowtown_docks_a3_n_-65538_6.dds"},
	{0x1C2662904DD649BCULL, "6170053_m.fxe"},
	{0x1C270BDE0B6DBCE2ULL, "art\\vfx\\models\\vfx_sphere.mmh"},
	{0x1C270DDE0B6DBF98ULL, "art\\vfx\\models\\vfx_sphere.msh"},
	{0x1C27516A7E1EFF57ULL, "gen_im_trs_cot_med_hlm_r2.uti"},
	{0x1C2767784C5E64C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_down.gad"},
	{0x1C27C04793A412F0ULL, "611295130.ogg"},
	{0x1C27F62940F96A59ULL, "gen00fl_varric6081259.cl"},
	{0x1C281D6E8900128FULL, "6068571_m.fxe"},
	{0x1C284F0A48289DEAULL, "6156024_m.fxe"},
	{0x1C287BABA89D3D0CULL, "llac_w_bonepitmining_a2_2_0.dds"},
	{0x1C28B869C106634EULL, "vxw_encounter2_camp_7o_0_0.vlm"},
	{0x1C295A1ACAB54DA5ULL, "dae300_cullen6135342.cl"},
	{0x1C2A0B1449A11668ULL, "hlac_lt_0_lowtown_n_-65536_8z.dds"},
	{0x1C2ABE06ADF42909ULL, "6135806_m.fxe"},
	{0x1C2B40B6B6F57EAEULL, "w_bonepitmining_a2_228_0.mmh"},
	{0x1C2B4CB6B6F59345ULL, "w_bonepitmining_a2_228_0.mao"},
	{0x1C2B5AB6B6F5AA9CULL, "w_bonepitmining_a2_228_0.msh"},
	{0x1C2B9FE72CB35BA5ULL, "dae000_hate_enemies6167500.cl"},
	{0x1C2BB13CF04497B5ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityexit.evt"},
	{0x1C2C1245E431F378ULL, "vxw_woundedcoast_5j_0_0.vlm"},
	{0x1C2C251F8FD54F49ULL, "6098813_m.fxe"},
	{0x1C2D506D460E1951ULL, "vxca_caverns1_gkg_0_1.vlm"},
	{0x1C2D5D43E1546C0CULL, "gen00fl_anders6125624.cl"},
	{0x1C2D97CAC6F0284AULL, "vxht_keep_40s_0_0.vlm"},
	{0x1C2DFFE705E2F23EULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_roba_0t.dds"},
	{0x1C2E1E2F584938D2ULL, "vxca_caverns1_hud_0_0.vlm"},
	{0x1C2E2A457467BDDFULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02oldgrave.msh"},
	{0x1C2E30457467C781ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02oldgrave.mmh"},
	{0x1C2E3C457467DC22ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02oldgrave.mao"},
	{0x1C2E4A2E9DF9CE03ULL, "vxlt_0_lowtown_docks_n_c14_0_0.vlm"},
	{0x1C2EB6F93B2A1F9FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc02a.phy"},
	{0x1C2EBBE301E237F1ULL, "vxlt_gallowscourtyard_724_0_0.vlm"},
	{0x1C2EEBF552F6B72EULL, "follower_banter6172694.cl"},
	{0x1C2F5343A7E9E396ULL, "w_bonepitmining_a2_238_0.mao"},
	{0x1C2F5F43A7E9F835ULL, "w_bonepitmining_a2_238_0.mmh"},
	{0x1C2F6143A7E9FBABULL, "w_bonepitmining_a2_238_0.msh"},
	{0x1C2F890A17DA3775ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyn_0n.dds"},
	{0x1C2F92F5CC01F3E9ULL, "gen_im_trs_slv_lgt_cht_r2.uti"},
	{0x1C2FE0F684426573ULL, "628259533.ogg"},
	{0x1C2FFD7FAD7E9125ULL, "vxlt_warehouse_aq_0_0.vlm"},
	{0x1C300F3B1EE15732ULL, "6139144_m.fxe"},
	{0x1C30AE5F6A37DE65ULL, "vxw_bonepit_mf_0_0.vlm"},
	{0x1C30D0BBB9DF59E0ULL, "vxw_bonepitmining_a2_1no_0_0.vlm"},
	{0x1C31003F016739F6ULL, "vxht_0_hightown_n_bc3_0_0.vlm"},
	{0x1C314F11E3F7A5C0ULL, "vxlt_warehouse_n_11z_0_0.vlm"},
	{0x1C31593572799CF2ULL, "6111215_f.fxe"},
	{0x1C31ABC24861AE0FULL, "430529840.ogg"},
	{0x1C31CA954B9902BAULL, "6170705_m.fxe"},
	{0x1C321C0A4FBC590EULL, "zz_qun_debug6083718.cl"},
	{0x1C329873E1B9BD1EULL, "art\\levels\\races\\elf\\props\\alienage\\prp_fire.phy"},
	{0x1C32CF8B390C4C7DULL, "hlac_ca_caverns1_mines_24421_11z.dds"},
	{0x1C3311BB25076089ULL, "vxht_mansion2b_1t1_0_0.vlm"},
	{0x1C334B6CA22180A9ULL, "zz_isa_debug6133961.cl"},
	{0x1C336BE96DD427DBULL, "6179596_m.fxe"},
	{0x1C33B7B40740CF30ULL, "ave100_donnic6106004.cl"},
	{0x1C33BCF8C80F5B80ULL, "220065795.ogg"},
	{0x1C33DC7C04259645ULL, "6050174_m.fxe"},
	{0x1C33DD3BDBE0A749ULL, "vxdr_primevaltomb_8u2_0_0.vlm"},
	{0x1C33E3ADC4D614E8ULL, "lt_0_lowtown_a3_n_c_44_0.anb"},
	{0x1C33FA09DFA0294FULL, "vxlt_hangedman_n_1r0_0_0.vlm"},
	{0x1C34020D5A6A35BDULL, "follower_banter6143192.cl"},
	{0x1C34D804B45A4F8CULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv6.ani"},
	{0x1C351B1C6B438E61ULL, "vxlt_0_lowtown_a2_n_eos_0_0.vlm"},
	{0x1C35A7A534CA7413ULL, "vxht_mansion_n_3ow_0_0.vlm"},
	{0x1C35F3550ED35CE6ULL, "flame_alpha.dds"},
	{0x1C36DAEE1311D27BULL, "hlac_ht_mansion2b_818_10z.dds"},
	{0x1C371D8BC9489087ULL, "279246236.ogg"},
	{0x1C373E7076B4BBA7ULL, "493395260.ogg"},
	{0x1C37A87A832B2D36ULL, "vxw_bonepit_xr_0_0.vlm"},
	{0x1C37B52EE86362D3ULL, "vxlt_0_lowtown_docks_bez_0_0.vlm"},
	{0x1C38215EC920219FULL, "276013727.ogg"},
	{0x1C38E105ECBE3063ULL, "zz_qun_debug6028897.cl"},
	{0x1C39103BA05E3968ULL, "885073018.ogg"},
	{0x1C3A3A3744E8D9F5ULL, "llac_lt_gallowstemplar_n_2395_1.dds"},
	{0x1C3A4B948D28322CULL, "hlac_lt_darktown_-1_5y.dds"},
	{0x1C3A8422D6280BF8ULL, "vxw_woundedcoast_2k_0_0.vlm"},
	{0x1C3AD59E159C5EC0ULL, "6093879_m.fxe"},
	{0x1C3AFC42ABE570D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_silence.gad"},
	{0x1C3B03E8184C3728ULL, "vxlt_0_lowtown_byl_0_0.vlm"},
	{0x1C3B4E15E70AE94AULL, "vxw_encounter1_f1_0_0.vlm"},
	{0x1C3BAE8A8FDA135BULL, "469047456.ogg"},
	{0x1C3BE504B46084CEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv6.gad"},
	{0x1C3BFDD879D5C22FULL, "6166772_m.fxe"},
	{0x1C3C900C4A3F5152ULL, "vxca_caverns1_hze_0_0.vlm"},
	{0x1C3C958D0465034FULL, "374118634.ogg"},
	{0x1C3CC4F59A356E2BULL, "vxw_sundermount_37p_0_0.vlm"},
	{0x1C3D58E48E8C5C37ULL, "vxca_caverns2prc_8k9_0_0.vlm"},
	{0x1C3D893DE9F0FBB9ULL, "dre000_hireling26038673.cl"},
	{0x1C3DA56E74914110ULL, "hlac_lt_0_lowtown_docks_a3_-262144_3x.dds"},
	{0x1C3DFCF3BBC072A2ULL, "vxlt_0_lowtown_docks_a3_n_bse_0_0.vlm"},
	{0x1C3E06BD44ABAA4FULL, "dae100_bran6144430.cl"},
	{0x1C3EA3D22C983A48ULL, "vxlt_0_lowtown_a2_n_eov_0_0.vlm"},
	{0x1C3EA4BDB7A536FEULL, "one220_yevhen6102810.cl"},
	{0x1C3EA6B00E84ED93ULL, "lt_0_lowtown_docks_n_9_0_ll.mmh"},
	{0x1C3EFADCA205A5F4ULL, "6056612_m.fxe"},
	{0x1C3EFB6381373041ULL, "vxlt_0_lowtown_a2_n_fha_0_0.vlm"},
	{0x1C3F13161EF13C44ULL, "593402999.ogg"},
	{0x1C3F4298ABD57260ULL, "vxw_bonepitmining_a2_1kl_0_0.vlm"},
	{0x1C3F53F4D66467FEULL, "927917500.ogg"},
	{0x1C3F598DB07C69DEULL, "6116060_m.fxe"},
	{0x1C3FB7F93B3894A4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc02a.mmh"},
	{0x1C3FCDF93B38BA46ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc02a.msh"},
	{0x1C3FD6EF1A842D59ULL, "art\\characters\\creatures\\ragedemon\\textures\\c_ragedmon_0t.dds"},
	{0x1C409DCE534B2A81ULL, "vxlt_warehouse_qm_0_0.vlm"},
	{0x1C40A75731782BDAULL, "hightown_fem_elfservant_mob6169049.cl"},
	{0x1C40AD045D008C07ULL, "vxht_0_hightown_aup_0_0.vlm"},
	{0x1C40FE0241FB1505ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_1z.dds"},
	{0x1C4101652412AB06ULL, "74533242.ogg"},
	{0x1C417E840A7BFE63ULL, "ht_0_hightown_21_0.anb"},
	{0x1C418BAD090EF92BULL, "zz_t-trailer_shot_2_pt2.cut"},
	{0x1C41ACE5081FF04BULL, "937599838.ogg"},
	{0x1C421142ABEBB364ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_silence.evt"},
	{0x1C4232208AD0BD3EULL, "6083836_m.fxe"},
	{0x1C425714DC0B4577ULL, "hlac_lt_0_lowtown_-65534_0x.dds"},
	{0x1C42968186B3FBEDULL, "vxlt_0_lowtown_a2_n_c_gob_0_1.vlm"},
	{0x1C42E004B4669AADULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv6.evt"},
	{0x1C42F8AF92FE08C5ULL, "hlac_dr_dungeon_3216_12x.dds"},
	{0x1C4310C872C10AC3ULL, "6119554_m.fxe"},
	{0x1C437B50446AF7EBULL, "6116585_f.fxe"},
	{0x1C437FFB28CDBFFEULL, "art\\vfx\\combat\\vfx_c_pride_cast_p_c.impa.evt"},
	{0x1C438180E50FA920ULL, "vxht_0_hightown_2xc_0_0.vlm"},
	{0x1C438770E35F5E03ULL, "6167245_m.fxe"},
	{0x1C445C40A2EDE512ULL, "dae000ip_crowd_mf8_talk17.utp"},
	{0x1C44A3D42C9FAF4AULL, "104714832.ogg"},
	{0x1C44AD67B6C06231ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door.mmh"},
	{0x1C44B967B6C076D2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door.mao"},
	{0x1C44BD68F49B2A84ULL, "6107179_f.fxe"},
	{0x1C44C767B6C08EEFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door.msh"},
	{0x1C457911142E0206ULL, "1025681405.ogg"},
	{0x1C458E3825B8165EULL, "vxw_woundedcoast_1p2_0_0.vlm"},
	{0x1C46BD1EDF3E0247ULL, "vxlt_undercityserialkiller_1sc_0_0.vlm"},
	{0x1C46E15227348538ULL, "312248992.ogg"},
	{0x1C475DB7D4722289ULL, "6170038_m.fxe"},
	{0x1C479D3483A20CFAULL, "6086912_m.fxe"},
	{0x1C47FAC231C46AF1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_bark01_d.dds"},
	{0x1C47FDFE97E0D908ULL, "vxw_bonepitmining_a2_6x_0_0.vlm"},
	{0x1C48107815D98389ULL, "gen00fl_fenris6126523.cl"},
	{0x1C48758EF3817E71ULL, "vxlt_gallowstemplar_n_6q9_0_0.vlm"},
	{0x1C48C68EC5D8B631ULL, "917857538.ogg"},
	{0x1C48CDC3CA0D8C8CULL, "6180810_m.fxe"},
	{0x1C48D119A3DE76CAULL, "6101558_m.fxe"},
	{0x1C490B9416E5C6A1ULL, "vxlt_gallowscourtyard_7cg_0_0.vlm"},
	{0x1C4926F000F0D395ULL, "6129182_m.fxe"},
	{0x1C492719FD37E247ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01bargewreck_1.mao"},
	{0x1C4998D972171C58ULL, "6039000_m.fxe"},
	{0x1C49BFF2D90101CAULL, "vxw_encounter1_k2_0_0.vlm"},
	{0x1C49F2E86A5BC980ULL, "6182940_m.fxe"},
	{0x1C4A0D4E0155C31DULL, "vxlt_0_lowtown_a2_n_c_hqj_0_0.vlm"},
	{0x1C4A96FB28D40529ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p_c.impa.gad"},
	{0x1C4AF460774FBC05ULL, "vxlt_0_lowtown_a2_n_c_fgz_0_0.vlm"},
	{0x1C4B36D28C2B86ABULL, "vxw_sundermount_32s_0_0.vlm"},
	{0x1C4B47EE646B4E2DULL, "883844705.ogg"},
	{0x1C4B5CB57BB46532ULL, "402243762.ogg"},
	{0x1C4C2D9D01501EA6ULL, "w_woundedcoast_17_0_ll.mmh"},
	{0x1C4C43299F03B769ULL, "lt_0_lowtown_docks_a3_164_0.mao"},
	{0x1C4C49C78536FE66ULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_act_c.impa.ani"},
	{0x1C4C4AD4659B5DADULL, "vxw_bonepitdestroyed_a3_la_0_0.vlm"},
	{0x1C4C4F299F03CB02ULL, "lt_0_lowtown_docks_a3_164_0.mmh"},
	{0x1C4C51299F03CEB8ULL, "lt_0_lowtown_docks_a3_164_0.msh"},
	{0x1C4C8EBB110FAEB0ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiremed.mmh"},
	{0x1C4C95DC69CF8D30ULL, "hlac_lt_0_lowtown_a2_n_-65538_0z.dds"},
	{0x1C4CD6CEAE39CDFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkbackright.gad"},
	{0x1C4E2F10ADBE3364ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallowbird.msh"},
	{0x1C4E4110ADBE51CDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallowbird.mao"},
	{0x1C4E4510ADBE5806ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallowbird.mmh"},
	{0x1C4E6639F2087D4EULL, "vxca_caverns1_j4a_0_0.vlm"},
	{0x1C4E75812210E23DULL, "hlac_lt_0_lowtown_-196609_0x.dds"},
	{0x1C4EC262F1FC44D9ULL, "gen_im_wep_war_two_act0_02.uti"},
	{0x1C4F0FAB45414301ULL, "vxlt_warehouse_zl_0_0.vlm"},
	{0x1C4F34ADB6F882B4ULL, "dae300_alistair_king6133822.cl"},
	{0x1C4F6E0A0C1BAB92ULL, "vxca_cavern1_haunted_e7f_0_0.vlm"},
	{0x1C502942ABF8195DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_silence.ani"},
	{0x1C503EEA175C446CULL, "vxw_bonepitmining_a2_219_0_0.vlm"},
	{0x1C5167FB28D9D477ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p_c.impa.ani"},
	{0x1C519A1EB4125B13ULL, "6053329_f.fxe"},
	{0x1C51FED64F2FBF61ULL, "vxlt_0_lowtown_a2_n_euq_0_0.vlm"},
	{0x1C524BFB6D338D67ULL, "hlac_lt_0_lowtown_a2_-65535_5z.dds"},
	{0x1C527B20E8E21FFDULL, "vxw_sundermount_2eu_0_0.vlm"},
	{0x1C52C5428FE760F3ULL, "hlac_ca_caverns1_24421_8y.dds"},
	{0x1C52DB4670D3BA4CULL, "6089887_m.fxe"},
	{0x1C52FDC6FF6F96D3ULL, "vxw_bonepitdestroyed_a3_1wq_0_0.vlm"},
	{0x1C5336543C4F9C50ULL, "vxht_brothel_n_6s6_0_0.vlm"},
	{0x1C539E114172CE14ULL, "vxlt_warehouse_n_e1_0_0.vlm"},
	{0x1C53D5CEAE3FEAFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkbackright.ani"},
	{0x1C5480DABF12296EULL, "vxw_bonepitmining_a2_eg_0_0.vlm"},
	{0x1C548C3467175E36ULL, "vxw_bonepit_nt_0_0.vlm"},
	{0x1C54C94389A2A57EULL, "one100dg_elren6058866.cl"},
	{0x1C54E9F0B38C5C56ULL, "823374265.ogg"},
	{0x1C55003E343DDE71ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_011a_0n.dds"},
	{0x1C55673F97D3E657ULL, "vfx_dlc2_lightning_ball.bnk"},
	{0x1C557C8793062538ULL, "dae000_party_night_comment6201612.cl"},
	{0x1C55A838AB3F6A00ULL, "6084700_f.fxe"},
	{0x1C560023AFA1D55BULL, "l_dae280ar_deep_roads.wbd"},
	{0x1C566FBC336EFBD2ULL, "vxht_keep_n_4ms_0_0.vlm"},
	{0x1C56B2EC4872DFACULL, "one110_meeran_st.stg"},
	{0x1C571A5E1ECDE2DAULL, "6129084_m.fxe"},
	{0x1C575CE39774CDC4ULL, "551690750.ogg"},
	{0x1C576386727F60E5ULL, "100596167.ogg"},
	{0x1C5784AB3D61C7EDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medf_0d.dds"},
	{0x1C588CEB9BEFF1BCULL, "vxht_0_hightown_a3_c7r_0_0.vlm"},
	{0x1C5937D4B2576FDDULL, "vxlt_0_lowtown_docks_cxj_0_0.vlm"},
	{0x1C597BEFF48B6C14ULL, "672678189.ogg"},
	{0x1C59B1BB111A74CFULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiremed.anb"},
	{0x1C5A31C7854312D7ULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_act_c.impa.evt"},
	{0x1C5A69D76099A505ULL, "vxw_woundedcoast_5s_0_0.vlm"},
	{0x1C5AA0F70FA4BED5ULL, "6061787_m.fxe"},
	{0x1C5AAE18AE3E7037ULL, "lt_0_lowtown_docks_a3_148_0.phy"},
	{0x1C5AB3B19799A03AULL, "6141907_m.fxe"},
	{0x1C5B15FC12B218CDULL, "llac_ca_caverns2_5802_60.dds"},
	{0x1C5B60E3641075D3ULL, "vxht_keep_n_4hy_0_0.vlm"},
	{0x1C5B9F2EFF195A7DULL, "827850795.ogg"},
	{0x1C5BAB8DEA495829ULL, "vxht_mansion_n_4xo_0_0.vlm"},
	{0x1C5C7E7CC0C43552ULL, "vxw_bonepitdestroyed_a3_1wt_0_0.vlm"},
	{0x1C5D003D22365EE1ULL, "6115326_f.fxe"},
	{0x1C5D6D2C3F4F2B22ULL, "tacticstargettypes.gda"},
	{0x1C5DEC505041FA31ULL, "571379089.ogg"},
	{0x1C5E07A2967FC109ULL, "35933895.ogg"},
	{0x1C5EAC73715EF3C7ULL, "mcr310_argument6182455.cl"},
	{0x1C5EFAB2463FC27EULL, "721153597.ogg"},
	{0x1C5F2DA5930A39DDULL, "vxdr_primevaltomb_3vt_0_0.vlm"},
	{0x1C5F4BCB6131BA29ULL, "vxlt_gallowsprison_n_6mm_0_0.vlm"},
	{0x1C5F7162FAD443D3ULL, "424208252.ogg"},
	{0x1C5FF78655662111ULL, "959351023.ogg"},
	{0x1C600D8801AF9A9EULL, "gen00fl_sebastian6180842.cl"},
	{0x1C600EB5855CC520ULL, "vxlt_0_lowtown_docks_n_b5c_0_1.vlm"},
	{0x1C6056DDF70D269DULL, "zz_mag_debug6053421.cl"},
	{0x1C609616DCA77FB5ULL, "follower_banter6171535.cl"},
	{0x1C60ECFDDAD8387DULL, "vxw_sundermount_2nt_0_0.vlm"},
	{0x1C6132C785493360ULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_act_c.impa.gad"},
	{0x1C616B34555B6EA5ULL, "llac_lt_foundry_n_320_13.dds"},
	{0x1C618CB69BEA0087ULL, "281952646.ogg"},
	{0x1C61B08BEAEECE9FULL, "llac_lt_blackemporium_1055_9.dds"},
	{0x1C62D9DFE4AF09D9ULL, "64716076.ogg"},
	{0x1C62F041A8C6AAA3ULL, "hlac_lt_gallowstemplar_2394_5z.dds"},
	{0x1C6305538351E602ULL, "vxw_encounter2_camp_rl_0_0.vlm"},
	{0x1C632568309BD3AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_loop.gad"},
	{0x1C636BE4791E8644ULL, "6126362_m.fxe"},
	{0x1C63FC310D9334CAULL, "vxca_caverns1_mines_e6y_0_0.vlm"},
	{0x1C6424536C4E0564ULL, "vxdr_0_deeproads_cxa_0_0.vlm"},
	{0x1C64D58F9871CA7CULL, "6129117_f.fxe"},
	{0x1C64E1B87BCD0A08ULL, "vxw_bonepitmining_a2_8v_0_0.vlm"},
	{0x1C6549A9DDCD47A7ULL, "634800589.ogg"},
	{0x1C664CA1C63F66A0ULL, "mcr370_varric6208282.cl"},
	{0x1C67266057F53434ULL, "vxlt_0_lowtown_gco_0_0.vlm"},
	{0x1C67ADDA7922E602ULL, "gencr_bloodmage_f.utc"},
	{0x1C67CDA8D2B1D405ULL, "1011985107.ogg"},
	{0x1C67F2BBB6151A87ULL, "gen00fl_anders6196685.cl"},
	{0x1C67F2FBB1FE7D03ULL, "6086250_m.fxe"},
	{0x1C683D223BBCF2CDULL, "199039565.ogg"},
	{0x1C68C75F4A7C3A3EULL, "6108254_m.fxe"},
	{0x1C68D0EE9A2239BCULL, "39169299.ogg"},
	{0x1C68DBB4528FBD85ULL, "vxw_woundedcoast_2r_0_0.vlm"},
	{0x1C6902067E33DDE3ULL, "vxlt_0_lowtown_a2_feg_0_0.vlm"},
	{0x1C69445478DD4B73ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_rnf.evt"},
	{0x1C69A7B591FFEB66ULL, "242280655.ogg"},
	{0x1C69F60E328E650AULL, "74941629.ogg"},
	{0x1C6A06B35CD7F435ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_brazier_s.dds"},
	{0x1C6A3A6830A2165CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_loop.evt"},
	{0x1C6A982F81B7BBE0ULL, "vxw_bonepitmining_a2_1zk_0_0.vlm"},
	{0x1C6AEA77C22A3B38ULL, "6052564_m.fxe"},
	{0x1C6B7F0FA545408FULL, "blood_cycle_l3.mao"},
	{0x1C6B8A4C7E1291DCULL, "vxca_caverns1_i2f_0_0.vlm"},
	{0x1C6BF3CEFFE9C784ULL, "vxht_chantry_2t2_0_1.vlm"},
	{0x1C6C6D1C9481DDCDULL, "hlac_lt_gallowscourtyard_n_131071_1z.dds"},
	{0x1C6DD8172C701E51ULL, "vxlt_0_lowtown_docks_a3_d0c_0_1.vlm"},
	{0x1C6DEB95857C28A1ULL, "vxdr_primevaltomb_80i_0_0.vlm"},
	{0x1C6E81B46CD8323AULL, "6163424_m.fxe"},
	{0x1C6F1F8795A7CDEDULL, "vxht_keep_62l_0_0.vlm"},
	{0x1C6F63EF67AAC55AULL, "328311228.ogg"},
	{0x1C6F9DBC8EB3A56BULL, "6095513_m.fxe"},
	{0x1C701A252B51A4D1ULL, "vxlt_0_lowtown_docks_c06_0_0.vlm"},
	{0x1C702526AB316751ULL, "plt_mag000pt_otl_debug.nss"},
	{0x1C7029854F675400ULL, "vxht_0_hightown_n_b8t_0_0.vlm"},
	{0x1C703D5478E35DB4ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_rnf.gad"},
	{0x1C706F4498C32970ULL, "zz_one_debug6025815.cl"},
	{0x1C707B4A666E1A4BULL, "load_controller.gfx"},
	{0x1C707C11386FFC7BULL, "hlac_lt_0_lowtown_a2_n_c_0_0y.dds"},
	{0x1C7081CB255EFF14ULL, "vxlt_warehouse_n_o7_0_0.vlm"},
	{0x1C714D20639588C7ULL, "llac_ht_keep_3732_9.dds"},
	{0x1C716494A2FE5A6EULL, "vxw_bonepitmining_a2_oy_0_0.vlm"},
	{0x1C7177307547FE82ULL, "vxw_encounter2_camp_mm_0_0.vlm"},
	{0x1C719A1186A9E849ULL, "hlac_dr_0_deeproads_5843_2y.dds"},
	{0x1C719FAA0519E34BULL, "311310781.ogg"},
	{0x1C722AC25599A24CULL, "hlac_ht_mansion_n_3371_0x.dds"},
	{0x1C722D761E0BEADEULL, "6069869_m.fxe"},
	{0x1C7296305E441DE4ULL, "vxdr_0_deeproads_cuf_0_0.vlm"},
	{0x1C72FEEB53F04A64ULL, "hlac_w_encounter3_0_0y.dds"},
	{0x1C73178A6836077CULL, "6155805_m.fxe"},
	{0x1C73D38B3ED20890ULL, "bee_glow.dds"},
	{0x1C73DC52EC69950CULL, "506267983.ogg"},
	{0x1C73E2ADC50BF6A0ULL, "lt_0_lowtown_a3_n_c_44_0.rml"},
	{0x1C73F8BA3F35AA04ULL, "vxlt_warehouse_n_1dq_0_0.vlm"},
	{0x1C7432B39EB8C3C8ULL, "vxca_caverns1_mines_i6o_0_0.vlm"},
	{0x1C7557B4B768D4E4ULL, "llac_w_bonepit_65537_0.dds"},
	{0x1C75AA1562571609ULL, "vxdr_0_deeproads_6z7_0_0.vlm"},
	{0x1C75F3D0112255F6ULL, "95154044.ogg"},
	{0x1C76362ADF4A7D7FULL, "vxlt_hangedman_16s_0_0.vlm"},
	{0x1C763C47D3C546C6ULL, "vxht_bartrand_n_6rc_0_0.vlm"},
	{0x1C764C5478E7E292ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_rnf.ani"},
	{0x1C767257705F1682ULL, "llac_lt_backalley_n_0_16.dds"},
	{0x1C76BF1FFEFEEB6FULL, "w_blightlands_243_0.mao"},
	{0x1C76CB1FFEFEFF0CULL, "w_blightlands_243_0.mmh"},
	{0x1C76D11FFEFF09FEULL, "w_blightlands_243_0.msh"},
	{0x1C76E8083E436AFEULL, "6072663_m.fxe"},
	{0x1C77E7491B6D4A17ULL, "gen00fl_bethany6097881.cl"},
	{0x1C78426830AE6175ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_loop.ani"},
	{0x1C78609A65F208CEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a.msh"},
	{0x1C786E9A65F2201FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a.mao"},
	{0x1C787A9A65F234BCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a.mmh"},
	{0x1C78EC8A3F638826ULL, "plt_cod_ltr_emeric_note.nss"},
	{0x1C78FB1E172BF381ULL, "6179886_m.fxe"},
	{0x1C790A0C73ADD460ULL, "vxw_bonepitmining_a2_1wh_0_0.vlm"},
	{0x1C7934479BEFA29DULL, "dre000_stroud6036706.cl"},
	{0x1C79568B6BFCB74AULL, "hla_ht_0_hightown_a2_131072_0z.dds"},
	{0x1C79737BDDBCF98FULL, "vxht_mansion_n_508_0_0.vlm"},
	{0x1C7A1D142CB26E8DULL, "vxlt_0_lowtown_a2_n_c_ei3_0_0.vlm"},
	{0x1C7A4146C25DE782ULL, "vxca_caverns1_jtm_0_0.vlm"},
	{0x1C7A8B9B7EDAD873ULL, "vxlt_0_lowtown_docks_a3_cyp_0_0.vlm"},
	{0x1C7A9D0E7B56E583ULL, "6016902_f.fxe"},
	{0x1C7A9D6E3B517401ULL, "3723660.ogg"},
	{0x1C7AC5CF0CE99016ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_inst.ani"},
	{0x1C7AF68E1BD066BFULL, "vxw_bonepitmining_a2_216_0_0.vlm"},
	{0x1C7B793E7BED8426ULL, "hlac_lt_hangedman_-1_3z.dds"},
	{0x1C7B9C456428B546ULL, "774679176.ogg"},
	{0x1C7BA55A5F2ED95EULL, "6074481_m.fxe"},
	{0x1C7C2063615F6185ULL, "6103043_m.fxe"},
	{0x1C7C3162D3CC9E64ULL, "lt_0_lowtown_docks_a3_n_101_0.msh"},
	{0x1C7C4362D3CCBCCDULL, "lt_0_lowtown_docks_a3_n_101_0.mao"},
	{0x1C7C4762D3CCC306ULL, "lt_0_lowtown_docks_a3_n_101_0.mmh"},
	{0x1C7C5DEABBBEE300ULL, "6060231_m.fxe"},
	{0x1C7C99B7BCB935F3ULL, "vxca_caverns2_7u9_0_0.vlm"},
	{0x1C7D9F0CA55548C3ULL, "6093843_m.fxe"},
	{0x1C7DD0B7BEC4697DULL, "vxw_sundermount_2tr_0_0.vlm"},
	{0x1C7DF7C81B5BE7FBULL, "vxw_bonepitdestroyed_a3_mh_0_0.vlm"},
	{0x1C7E83DBB7994053ULL, "root_extraschool.dds"},
	{0x1C7E84A783FCB54AULL, "hlac_w_woundedcoast_196609_0z.dds"},
	{0x1C7EF3A817551794ULL, "vxlt_warehouse_n_r4_0_0.vlm"},
	{0x1C7F591AA96204C1ULL, "110591604.ogg"},
	{0x1C7F64AF746EB977ULL, "390208122.ogg"},
	{0x1C7FBABBE777F24EULL, "6076380_m.fxe"},
	{0x1C7FD67194F472EEULL, "vxw_bonepitmining_a2_jz_0_0.vlm"},
	{0x1C7FE90D673E1702ULL, "vxw_encounter2_camp_hn_0_0.vlm"},
	{0x1C8079BEAA1591C0ULL, "art\\characters\\creatures\\corpse\\textures\\cn_hed_cor_0d.dds"},
	{0x1C807C1F9862C3F9ULL, "ef_gen_dalish_assassin.mor"},
	{0x1C80E31E4F5FC26AULL, "787052438.ogg"},
	{0x1C81080D503A3664ULL, "vxdr_0_deeproads_crg_0_0.vlm"},
	{0x1C811264E8F861DEULL, "vxlt_darktown_5ct_0_0.vlm"},
	{0x1C812898B988094FULL, "6066182_m.fxe"},
	{0x1C8198F1FA0F8304ULL, "dae000_keep_noble_sta6161534.cl"},
	{0x1C81BE5862930454ULL, "vxht_keep_n_42h_0_0.vlm"},
	{0x1C81E6CB59139725ULL, "521865194.ogg"},
	{0x1C827F61A285E28EULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_roba_0d.dds"},
	{0x1C82FD3536CE322AULL, "hlac_dr_0_deeproads_5839_1z.dds"},
	{0x1C838400F5D807E5ULL, "dae000_orsino6081380.cl"},
	{0x1C85301686B1BA13ULL, "gen00fl_dog6176628.cl"},
	{0x1C85439DC32160C0ULL, "vxw_bonepit_84_0_0.vlm"},
	{0x1C8587B07836CC5FULL, "vxlt_0_lowtown_a2_n_h2c_0_0.vlm"},
	{0x1C85A8353ADD157FULL, "llac_ca_caverns1_24425_7.dds"},
	{0x1C85AB1AA4BD93F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkbackleft.ani"},
	{0x1C86243AC0EBCDDEULL, "hlac_ht_chantry_2534_5z.dds"},
	{0x1C863E62D3D508D1ULL, "lt_0_lowtown_docks_a3_n_101_0.phy"},
	{0x1C872726320D48E2ULL, "hlac_dr_primevaltomb_9621_2x.dds"},
	{0x1C87D41396A67ED2ULL, "vxw_bonepitdestroyed_a3_1fw_0_0.vlm"},
	{0x1C87E6BC1A8D4868ULL, "gen00fl_isabela6101471.cl"},
	{0x1C880E7015CDA02EULL, "hlac_dr_slavers_3216_11x.dds"},
	{0x1C88119A7A7A75D6ULL, "vxht_0_hightown_a3_b8p_0_0.vlm"},
	{0x1C881E4491394906ULL, "gen_im_arm_cht_elf_coma.uti"},
	{0x1C885F5BD0599930ULL, "hlac_lt_0_lowtown_a2_n_-3_3y.dds"},
	{0x1C888EF11EA8870DULL, "vxlt_0_lowtown_a2_n_c_en2_0_0.vlm"},
	{0x1C88C443CA92573BULL, "llac_w_blightlands_-1_0.dds"},
	{0x1C88CDCF0CF5DB07ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_inst.evt"},
	{0x1C88D6DB88EF086AULL, "771824380.ogg"},
	{0x1C88E65516CF0687ULL, "vxht_0_hightown_alo_0_0.vlm"},
	{0x1C88FC614588AF87ULL, "874542739.ogg"},
	{0x1C8909FCEA8295C9ULL, "548729574.ogg"},
	{0x1C89233F03413B85ULL, "hlac_ht_0_hightown_n_-1_1x.dds"},
	{0x1C89D9954F8AC7E7ULL, "565075254.ogg"},
	{0x1C89E5C97A812ED1ULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_pcmale.gad"},
	{0x1C8A266CB0A320CCULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_anders_andersintro.gad"},
	{0x1C8A4974C2C4D478ULL, "vxca_caverns1_e8l_0_0.vlm"},
	{0x1C8A4DB8C63B7F29ULL, "hlac_lt_0_lowtown_docks_n_65534_3z.dds"},
	{0x1C8A8F15E5069DF8ULL, "6121940_f.fxe"},
	{0x1C8ADBA7B3B72C3EULL, "zz_dog_debug6204096.cl"},
	{0x1C8B4B1420A8B911ULL, "6140207_f.fxe"},
	{0x1C8B7E5EF44A0B12ULL, "hm_icarver_h2s1.mor"},
	{0x1C8BA81AA4C1FA56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkbackleft.gad"},
	{0x1C8BB8A48784EE3BULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.dura.evt"},
	{0x1C8C410EA0F6783BULL, "648735512.ogg"},
	{0x1C8C4294B0BA81FDULL, "vxw_sundermount_2yq_0_0.vlm"},
	{0x1C8C5511AC5E9DD4ULL, "6061232_m.fxe"},
	{0x1C8CA4B6DE5E2FF6ULL, "vxht_0_hightown_n_9uz_0_0.vlm"},
	{0x1C8CF161EC2F8D21ULL, "vxw_bonepitdestroyed_a3_1pb_0_0.vlm"},
	{0x1C8D018A3F745561ULL, "plt_cod_ltr_emeric_note.plo"},
	{0x1C8D4FF4E1EF1D33ULL, "607473857.ogg"},
	{0x1C8D6585094B3014ULL, "vxlt_warehouse_n_y5_0_0.vlm"},
	{0x1C8D7622345A1AF9ULL, "6041049_f.fxe"},
	{0x1C8DF45C258C9301ULL, "vxca_caverns1_i27_0_0.vlm"},
	{0x1C8ED8090DBBCD55ULL, "gen00fl_isabela6097560.cl"},
	{0x1C8F2929354AB906ULL, "isa211_isabela6202269.cl"},
	{0x1C8F60CAD5E9C957ULL, "vxht_brothel_n_4gw_0_0.vlm"},
	{0x1C8F812A28F5DE9AULL, "vxw_encounter1_el_0_0.vlm"},
	{0x1C8F8441DAEE7A5EULL, "vxlt_darktown_5du_0_0.vlm"},
	{0x1C8FD362ED0E0D00ULL, "437559542.ogg"},
	{0x1C8FE6C97A859C7FULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_pcmale.ani"},
	{0x1C8FEECF0CFC3250ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_inst.gad"},
	{0x1C903FEAA8EC06F2ULL, "6131636_m.fxe"},
	{0x1C907B15322A5B89ULL, "6149280_f.fxe"},
	{0x1C90BA7997D68B2EULL, "mag101_grace6053093.cl"},
	{0x1C90BF101F2AA5CAULL, "vxlt_hangedman_27b_0_0.vlm"},
	{0x1C90D0DEACF0D4C3ULL, "vrc221_bartrand6043278.cl"},
	{0x1C91256CB0A93DCAULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_anders_andersintro.ani"},
	{0x1C9177B1E669664DULL, "lgt320_shw_ors_secret_nob6156350.cl"},
	{0x1C91B1A487894E7CULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.dura.gad"},
	{0x1C91E431161D94B4ULL, "707286592.ogg"},
	{0x1C91EDC5B4A493DAULL, "vxht_keep_n_3ih_0_0.vlm"},
	{0x1C933B6740CCE1D4ULL, "llac_lt_gallowscourtyard_-2_1.dds"},
	{0x1C934634E770B67CULL, "691268909.ogg"},
	{0x1C9364B4D1864F28ULL, "631259055.ogg"},
	{0x1C9397FBA64D7B41ULL, "80745056.ogg"},
	{0x1C9406B46F48BD36ULL, "gen00fl_aveline6152546.cl"},
	{0x1C9475F412DF084EULL, "follower_banter6168024.cl"},
	{0x1C94C05D1E5461A3ULL, "4653562.ogg"},
	{0x1C94DE90CFBABB3EULL, "6162658_m.fxe"},
	{0x1C953CF2403E03BEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sack_sa.dds"},
	{0x1C956CE3044B6B4FULL, "llac_lt_0_lowtown_a3_n_c_65533_0.dds"},
	{0x1C95D6BA36725F74ULL, "art\\characters\\creatures\\crowds\\elfmale\\textures\\prp_hightown_jimmy_0t.dds"},
	{0x1C961660CF5507F9ULL, "6013346_m.fxe"},
	{0x1C963FF182F4FE02ULL, "gen00fl_aveline6132830.cl"},
	{0x1C966A1B3F39B0A0ULL, "vxht_0_hightown_a3_bzl_0_0.vlm"},
	{0x1C96AD0538B1D230ULL, "6113933_m.fxe"},
	{0x1C974ED00FFCEFB4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masc_0n.dds"},
	{0x1C97525057E9BDB1ULL, "840951992.ogg"},
	{0x1C985067E1442348ULL, "6129913_f.fxe"},
	{0x1C9894EAC38AB019ULL, "hlac_ca_caverns1_24421_9x.dds"},
	{0x1C98B0A4878F6B7AULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.dura.ani"},
	{0x1C98F78584EA96D0ULL, "16176588.ogg"},
	{0x1C990E9148D05633ULL, "6182446_m.fxe"},
	{0x1C992F03DBEFB041ULL, "6072606_m.fxe"},
	{0x1C9966C4FA7A2A1AULL, "hlac_lt_0_lowtown_docks_-65538_0x.dds"},
	{0x1C9997CA209468F1ULL, "lgt360_mag_orlanna6155316.cl"},
	{0x1C99BD5435AF545CULL, "hlac_ht_0_hightown_a2_-65536_1z.dds"},
	{0x1C9A2F006220FC72ULL, "6101190_m.fxe"},
	{0x1C9A34AC778FEBEFULL, "vxca_cavern1_haunted_jfs_0_0.vlm"},
	{0x1C9A7CC71ECFF6BCULL, "224193421.ogg"},
	{0x1C9AC0FA41E557A5ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_1y.dds"},
	{0x1C9B4AE95D73954DULL, "follower_banter6148702.cl"},
	{0x1C9B592F718EB372ULL, "6168164_m.fxe"},
	{0x1C9B76A59304FF7EULL, "face_rendertodepth_p2.psh.fxo.dx11"},
	{0x1C9B8957BDAE5A3AULL, "6121555_f.fxe"},
	{0x1C9D599A66111887ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a.phy"},
	{0x1C9D59F8F65F87D8ULL, "gen00fl_merrill6107033.cl"},
	{0x1C9E691F38B1D5DBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\runforward.evt"},
	{0x1C9E880CF51AB350ULL, "vxw_blightlands_b9_0_0.vlm"},
	{0x1C9EECF224207409ULL, "6142380_f.fxe"},
	{0x1C9F0259C4644AC1ULL, "hlac_lt_0_lowtown_n_-196608_0z.dds"},
	{0x1C9F9A61CEC96F02ULL, "atl_chanters_dxt5_dat.xml"},
	{0x1C9FE145605F59CDULL, "6065840_m.fxe"},
	{0x1C9FF88F0FC5DA30ULL, "hlac_lt_gallowstemplar_2394_11z.dds"},
	{0x1CA06A7111BF1731ULL, "vxlt_0_lowtown_f51_0_0.vlm"},
	{0x1CA06EBB314EA848ULL, "dae000_gamlen6149418.cl"},
	{0x1CA075441B7D0729ULL, "742452072.ogg"},
	{0x1CA0D04B9C62EEC8ULL, "1068358946.ogg"},
	{0x1CA0E04220CFFFD4ULL, "vxw_woundedcoast_ib_0_0.vlm"},
	{0x1CA146C4DE5EE5F3ULL, "llac_w_bonepitdestroyed_a3_393220_0.dds"},
	{0x1CA18318AE7A298FULL, "lt_0_lowtown_docks_a3_148_0.mao"},
	{0x1CA18F18AE7A3E2CULL, "lt_0_lowtown_docks_a3_148_0.mmh"},
	{0x1CA19518AE7A481EULL, "lt_0_lowtown_docks_a3_148_0.msh"},
	{0x1CA2061B679258A8ULL, "plt_lgt000pt_tombs_first.nss"},
	{0x1CA20C6F39EB7ED6ULL, "mag360_samson6042053.cl"},
	{0x1CA2A0DF1A388E7EULL, "751871869.ogg"},
	{0x1CA2F4A70A771375ULL, "6110065_m.fxe"},
	{0x1CA31DE1AAE9F47DULL, "6172088_f.fxe"},
	{0x1CA3DAFD603EC67CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_threat_decay.ani"},
	{0x1CA3FC2A71ECD01BULL, "6084422_f.fxe"},
	{0x1CA4A8C212496965ULL, "333193762.ogg"},
	{0x1CA4B7CD7A92AFD2ULL, "vxw_bonepitdestroyed_a3_1lq_0_0.vlm"},
	{0x1CA4C24FC530713CULL, "mag320_orsino6071146.cl"},
	{0x1CA50BC6E34E1033ULL, "790751606.ogg"},
	{0x1CA5621F38B7E81CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\runforward.gad"},
	{0x1CA572AB0294B80DULL, "vxlt_0_lowtown_a2_n_c_ex0_0_0.vlm"},
	{0x1CA5DEE9EF5ECFCEULL, "240785110.ogg"},
	{0x1CA5EF6D953C532BULL, "vxlt_darktown_2wj_0_0.vlm"},
	{0x1CA699FF117485DBULL, "vxht_mansion2_1tc_0_0.vlm"},
	{0x1CA69B4379EF4B26ULL, "hlac_ht_mansion2_818_55z.dds"},
	{0x1CA6BDEF745D7021ULL, "6162503_m.fxe"},
	{0x1CA72F9FF918343BULL, "vxlt_gallowscourtyard_50r_0_0.vlm"},
	{0x1CA7A9B940BADA31ULL, "766036630.ogg"},
	{0x1CA7CBF6F55F25ECULL, "361971090.ogg"},
	{0x1CA7FB371B962FA2ULL, "vxht_bartrand_n_5pb_0_0.vlm"},
	{0x1CA82B49CE4A6135ULL, "6106813_m.fxe"},
	{0x1CA85E29A470CC87ULL, "64718421.ogg"},
	{0x1CA8961E661EE63BULL, "vxht_mansion2_3ch_0_0.vlm"},
	{0x1CA90F1B679886BBULL, "plt_lgt000pt_tombs_first.plo"},
	{0x1CA9FF0E224C2681ULL, "vxw_encounter3_3s_0_0.vlm"},
	{0x1CAA407401C9A3C0ULL, "llac_lt_0_lowtown_1_6.dds"},
	{0x1CAA883975EB5DE9ULL, "6206214_m.fxe"},
	{0x1CAAA6081A77E6FFULL, "hairalpha_rendertodepth_p0.psh.fxo"},
	{0x1CAAAA0D45AF6A6CULL, "art\\characters\\weapons\\longbows\\w_lbw_lw103a.msh"},
	{0x1CAAB00D45AF745EULL, "art\\characters\\weapons\\longbows\\w_lbw_lw103a.mmh"},
	{0x1CAAC58DF3ED6AA7ULL, "vxlt_0_lowtown_a2_gxq_0_0.vlm"},
	{0x1CAAD8160978C401ULL, "vxca_caverns1_i85_0_0.vlm"},
	{0x1CAAE7FD6044FABEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_threat_decay.gad"},
	{0x1CAB3E653AF99F21ULL, "vxca_cavern1_haunted_i2n_0_0.vlm"},
	{0x1CABCAB7F6E34EE6ULL, "vxlt_gallowsprison_n_62c_0_0.vlm"},
	{0x1CAC0EABBA9E5D93ULL, "dae000cr_cassandra.utc"},
	{0x1CAC32CF62C09B81ULL, "hlac_lt_0_lowtown_docks_n_-3_8z.dds"},
	{0x1CAC441FD63E5242ULL, "vxlt_0_lowtown_n_e0q_0_0.vlm"},
	{0x1CAC611F38BE051AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\runforward.ani"},
	{0x1CAC67FBBEDAAB5EULL, "vxlt_darktown_5rw_0_0.vlm"},
	{0x1CAC6B1046868AB0ULL, "mcr310_cut_chantry_blows_up6116283.cl"},
	{0x1CAC8BCCA5328555ULL, "6129821_m.fxe"},
	{0x1CACA4B4F8613D5FULL, "vxca_caverns1_mines_dpk_0_0.vlm"},
	{0x1CACE4E7E14C8F7AULL, "lowtown_fem_alienage1_sta6162531.cl"},
	{0x1CAD186FC9812A60ULL, "vxlt_0_lowtown_a2_n_c_f55_0_0.vlm"},
	{0x1CAD3C26AB653C3AULL, "plt_mag000pt_otl_debug.plo"},
	{0x1CAD667DE7767DC5ULL, "157396785.ogg"},
	{0x1CAD668B39F86B75ULL, "vxht_0_hightown_a2_avs_0_0.vlm"},
	{0x1CAD994C1A61E247ULL, "m_and111ar_andersshop.wbd"},
	{0x1CADCF627D101449ULL, "vxht_0_hightown_a2_c7j_0_0.vlm"},
	{0x1CAE36EFAE17C957ULL, "150282135.ogg"},
	{0x1CAE390457C12B6AULL, "296530643.ogg"},
	{0x1CAE7118B0027930ULL, "vxlt_0_lowtown_eug_0_0.vlm"},
	{0x1CAF2D57730A41FBULL, "art\\levels\\ztools\\default\\generic\\0cc_basic.msh"},
	{0x1CAF3F57730A6046ULL, "art\\levels\\ztools\\default\\generic\\0cc_basic.mao"},
	{0x1CAF4B57730A74E5ULL, "art\\levels\\ztools\\default\\generic\\0cc_basic.mmh"},
	{0x1CB01AAC663FD80CULL, "hlac_lt_gallowstemplar_2383_9y.dds"},
	{0x1CB01DE7B1D1D372ULL, "gen00fl_aveline6114023.cl"},
	{0x1CB063F50155C971ULL, "hlac_ht_mansion_n_3381_1y.dds"},
	{0x1CB078EAC6D246CFULL, "6147698_m.fxe"},
	{0x1CB0BCB21C049662ULL, "w_woundedcoast_99_0.phy"},
	{0x1CB0FB832102F1A1ULL, "zz_dae_debug6047616.cl"},
	{0x1CB1554F8C7F9D23ULL, "342196965.ogg"},
	{0x1CB16E1231E004ACULL, "6208320_f.fxe"},
	{0x1CB1A744CFA81E42ULL, "351928419.ogg"},
	{0x1CB1C2FD604ADADDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_threat_decay.evt"},
	{0x1CB256BB960DEB29ULL, "vxht_mansion_n_4fq_0_0.vlm"},
	{0x1CB308D2BC402510ULL, "zz_followers_banter6159693.cl"},
	{0x1CB310C5DD4AD0D3ULL, "6144321_m.fxe"},
	{0x1CB315FC6A760474ULL, "rdr111dg_harbormaster6016145.cl"},
	{0x1CB3D84BA3E9822CULL, "6136451_m.fxe"},
	{0x1CB4672E127CEE2FULL, "vxw_encounter2_camp_e4_0_0.vlm"},
	{0x1CB48ED242C622F3ULL, "6127963_f.fxe"},
	{0x1CB4DB00DA5E7DE6ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityenter.ani"},
	{0x1CB52FCC665388A1ULL, "6167603_m.fxe"},
	{0x1CB61CA4BE275803ULL, "6174454_m.fxe"},
	{0x1CB656E609457094ULL, "52286656.ogg"},
	{0x1CB65779526AC12AULL, "432063367.ogg"},
	{0x1CB6886320EFAD19ULL, "gen00fl_fenris6113970.cl"},
	{0x1CB6C1B0D5C491A0ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliff02_da.dds"},
	{0x1CB6D1661282390DULL, "vxlt_0_lowtown_a2_n_c_h2g_0_0.vlm"},
	{0x1CB707FB5814FEBBULL, "vxht_mansion2_3fk_0_0.vlm"},
	{0x1CB750B960700391ULL, "vxht_keep_n_739_0_0.vlm"},
	{0x1CB8006F78EE5BE1ULL, "hlac_lt_gallowstemplar_f_2393_0y.dds"},
	{0x1CB824A2F6921E9BULL, "art\\characters\\heraldry\\hld_inquisition.dds"},
	{0x1CB846F8C211D6A1ULL, "vxw_bonepitdestroyed_a3_1cg_0_0.vlm"},
	{0x1CB852D84756613EULL, "lgt110_gxa_coa_spy6117453.cl"},
	{0x1CB8CFC8E19123BBULL, "vxlt_gallowscourtyard_n_7hg_0_0.vlm"},
	{0x1CB8E79EEC69E0D3ULL, "gen00fl_bethany6109229.cl"},
	{0x1CB952E506D03F89ULL, "gen00fl_aveline6122764.cl"},
	{0x1CB9ACE3F53D62E2ULL, "gen00fl_sebastian6188587.cl"},
	{0x1CB9D2913055C277ULL, "6085431_m.fxe"},
	{0x1CBA031C5A5DDB4CULL, "vxht_brothel_n_5ik_0_0.vlm"},
	{0x1CBA450CECD97655ULL, "345393033.ogg"},
	{0x1CBAB0BD130014FBULL, "hlac_dr_slavers_3216_1z.dds"},
	{0x1CBACA180760410BULL, "docks_male_laborer_mob6175399.cl"},
	{0x1CBAFCE1915F5565ULL, "623051783.ogg"},
	{0x1CBB7691C1F42023ULL, "vxht_mansion2b_21c_0_0.vlm"},
	{0x1CBB9D7F7BFB96DDULL, "hlac_ca_caverns1_24421_3z.dds"},
	{0x1CBBBA04651A8398ULL, "dae300cr_orsino.utc"},
	{0x1CBBFA23A17C4835ULL, "vxht_0_hightown_a3_c8n_0_0.vlm"},
	{0x1CBC4C856C8C7951ULL, "338903218.ogg"},
	{0x1CBC7199C6B1C1A7ULL, "860318427.ogg"},
	{0x1CBC80D8D685E2ECULL, "vxht_mansion2b_2of_0_0.vlm"},
	{0x1CBC878AC054A7D4ULL, "gen00fl_anders6113030.cl"},
	{0x1CBCE2F5A1F891B0ULL, "vxlt_0_lowtown_epd_0_0.vlm"},
	{0x1CBCF34F2AAB663CULL, "6136684_m.fxe"},
	{0x1CBD1DF7F2993381ULL, "ash_dust.mao"},
	{0x1CBD3DDBD95997A4ULL, "follower_banter6173336.cl"},
	{0x1CBD78E4AA9B2A1AULL, "aip_follower_isabella.gda"},
	{0x1CBE265FDCF8AD57ULL, "vxht_keep_n_5xl_0_0.vlm"},
	{0x1CBE828E8D326F57ULL, "113995487.ogg"},
	{0x1CBE9D50E356BF5AULL, "ser121_jethann6106567.cl"},
	{0x1CBEE7B2BF366374ULL, "6214612_m.fxe"},
	{0x1CBEEAC7B8C85F4FULL, "6140798_m.fxe"},
	{0x1CBF06D4616E963CULL, "rdr111dg_assistant6016660.cl"},
	{0x1CBF50B92DC75C1CULL, "336014997.ogg"},
	{0x1CBF86035D9BD888ULL, "6036459_f.fxe"},
	{0x1CC09A097C3C4FF0ULL, "vxlt_backalley_n_1vz_0_0.vlm"},
	{0x1CC0D282484A7AF8ULL, "vxdr_0_deeproads_alj_0_1.vlm"},
	{0x1CC10E232DFFC467ULL, "823655015.ogg"},
	{0x1CC1475583C68160ULL, "35755373.ogg"},
	{0x1CC1E92F9FEC78EDULL, "zz_seb_debug6214540.cl"},
	{0x1CC1FF5491127838ULL, "hlac_lt_darktown_-1_44x.dds"},
	{0x1CC2182027F04DC1ULL, "6061804_m.fxe"},
	{0x1CC2590CB760A9ACULL, "vxlt_gallowscourtyard_n_4bi_0_0.vlm"},
	{0x1CC29D7835C1488DULL, "hf_circlemage01.mor"},
	{0x1CC2C300DA6A9257ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityenter.evt"},
	{0x1CC35E05CF7C4E3AULL, "vxht_0_hightown_a3_bdx_0_0.vlm"},
	{0x1CC399CD473926FBULL, "vrc321_varric6094458.cl"},
	{0x1CC3A7CB9C440599ULL, "follower_banter6163738.cl"},
	{0x1CC40642069EA34CULL, "plt_lgt000pt_ultra_2_05.plo"},
	{0x1CC475C458E32F64ULL, "tbe100_xenon6190986.cl"},
	{0x1CC4CF251DA471F1ULL, "hlac_lt_0_lowtown_a3_n_c_131071_6y.dds"},
	{0x1CC4DEF9393D3086ULL, "6084749_m.fxe"},
	{0x1CC4ED7B0D3E6AC3ULL, "vxca_caverns1_mines_gik_0_0.vlm"},
	{0x1CC5A06AE769E364ULL, "vxht_chantry_2ve_0_0.vlm"},
	{0x1CC5E278E3B736B2ULL, "vxht_bartrand_n_3ka_0_0.vlm"},
	{0x1CC63EE55E3BA0EDULL, "6099317_m.fxe"},
	{0x1CC6B58112AFB966ULL, "6084536_m.fxe"},
	{0x1CC6B8D5B407EF21ULL, "vxw_bonepitdestroyed_a3_1df_0_0.vlm"},
	{0x1CC776B1A4503A5BULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_disap_p.mmh"},
	{0x1CC827D05C0057D3ULL, "vxlt_0_lowtown_docks_b7d_0_0.vlm"},
	{0x1CC88E07F42451F2ULL, "gen_im_trs_dem_lgt_glv_r2.uti"},
	{0x1CC94193D43DF70FULL, "hlac_lt_blackemporium_1055_12z.dds"},
	{0x1CC94A4FBE3B7E73ULL, "vxlt_gamlenshouse_n_ix_0_0.vlm"},
	{0x1CC98CFEF6A372D7ULL, "191256629.ogg"},
	{0x1CC99034DCC291DDULL, "6162103_m.fxe"},
	{0x1CC9B980C109E4E9ULL, "hlac_ht_chantry_2534_15x.dds"},
	{0x1CC9C400DA70B2E0ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityenter.gad"},
	{0x1CC9F5DE5FFEE153ULL, "6096905_m.fxe"},
	{0x1CCA054A0C42939AULL, "llac_lt_backalley_g_65537_6.dds"},
	{0x1CCA4A451DE49C75ULL, "vxht_0_hightown_a2_alq_0_0.vlm"},
	{0x1CCA95327F15A187ULL, "43962742.ogg"},
	{0x1CCAB627EDCE7271ULL, "vxca_cavern1_haunted_im7_0_0.vlm"},
	{0x1CCADC34BF661F3BULL, "15730203.ogg"},
	{0x1CCAEB4206A49383ULL, "plt_lgt000pt_ultra_2_05.nss"},
	{0x1CCB15B6662C8F63ULL, "lgt000im_ultra_knife.uti"},
	{0x1CCB3DAAFE5036C7ULL, "71251470.ogg"},
	{0x1CCB482E224B167FULL, "zz_qcr_debug6074955.cl"},
	{0x1CCBA3BCF4BEC062ULL, "vxlt_gallowscourtyard_6wm_0_0.vlm"},
	{0x1CCBD9FFFD080D99ULL, "dae000ip_crowd_mf3_merchant.utp"},
	{0x1CCC081742D56155ULL, "6088675_f.fxe"},
	{0x1CCC457ADC2551C6ULL, "777779521.ogg"},
	{0x1CCCC468632A9E5AULL, "552832384.ogg"},
	{0x1CCCDCB430308ECFULL, "hlac_lt_gallowstemplar_f_2388_3z.dds"},
	{0x1CCCDD07E4D623A4ULL, "one140cr_charade6137040.cl"},
	{0x1CCD6E3D4185BAB5ULL, "hlac_w_blightlands_65539_0z.dds"},
	{0x1CCE23262CA32DD0ULL, "6207707_f.fxe"},
	{0x1CCE2A62C92E16DFULL, "vxlt_shop_ep_0_0.vlm"},
	{0x1CCE2D6D98A2EE42ULL, "1072729765.ogg"},
	{0x1CCEE373CE060EDFULL, "6084348_m.fxe"},
	{0x1CCF0E2B90219C84ULL, "llac_lt_gallowscourtyard_-65537_1.dds"},
	{0x1CD0D257002C21B7ULL, "llac_ht_mansion2_818_25.dds"},
	{0x1CD0DB9672ECA903ULL, "300486473.ogg"},
	{0x1CD10FCE3243EAE2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_rough.msh"},
	{0x1CD111CE3243ED98ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_rough.mmh"},
	{0x1CD1120B1002043BULL, "dre000_vampire6034332.cl"},
	{0x1CD11DCE3244023BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_rough.mao"},
	{0x1CD14AAD743691A8ULL, "follower_banter6167988.cl"},
	{0x1CD1A8861D54F652ULL, "vxlt_0_lowtown_docks_b7a_0_0.vlm"},
	{0x1CD1E9877EB4C1B8ULL, "vxw_bonepitdestroyed_a3_p4_0_0.vlm"},
	{0x1CD22D9BBFDCA8DFULL, "519786155.ogg"},
	{0x1CD2408E700A6617ULL, "19158078.ogg"},
	{0x1CD29180450E57C7ULL, "6048726_m.fxe"},
	{0x1CD2E79E9DB78C59ULL, "626960547.ogg"},
	{0x1CD308A02336D421ULL, "vxdr_primevaltomb_8mb_0_0.vlm"},
	{0x1CD335D7CEF9082FULL, "vxlt_0_lowtown_docks_coq_0_0.vlm"},
	{0x1CD350CDF1787922ULL, "vxht_bartrand_n_5am_0_0.vlm"},
	{0x1CD39259640C91FCULL, "and121_karl6093477.cl"},
	{0x1CD3EAA5ADCB7F19ULL, "816859198.ogg"},
	{0x1CD3F140F3BCBD9BULL, "vxht_0_hightown_a2_aux_0_0.vlm"},
	{0x1CD3FB8B1C18448FULL, "vxw_encounter2_camp_q0_0_0.vlm"},
	{0x1CD4811DC80BA840ULL, "vxlt_0_lowtown_gj4_0_0.vlm"},
	{0x1CD4EC85FC79545FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medb_0t.dds"},
	{0x1CD50586E925198FULL, "421566731.ogg"},
	{0x1CD52D4D58CC51D8ULL, "vxht_keep_70j_0_0.vlm"},
	{0x1CD597B1A45CAFC0ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_disap_p.anb"},
	{0x1CD597B21C237316ULL, "w_woundedcoast_99_0.mao"},
	{0x1CD5A3B21C2387B5ULL, "w_woundedcoast_99_0.mmh"},
	{0x1CD5A5B21C238B2BULL, "w_woundedcoast_99_0.msh"},
	{0x1CD5AFD7EC791783ULL, "ser211ip_note_scribbles_4.utp"},
	{0x1CD671BAEFC0331DULL, "vxlt_darktown_533_0_0.vlm"},
	{0x1CD6889E7B1E86C1ULL, "6174328_m.fxe"},
	{0x1CD6A3A17FEE6D0DULL, "w_bonepitmining_a2_209_0.mmh"},
	{0x1CD6A7A17FEE745EULL, "w_bonepitmining_a2_209_0.mao"},
	{0x1CD6B5A17FEE8B93ULL, "w_bonepitmining_a2_209_0.msh"},
	{0x1CD7026BBEADB1C5ULL, "6102949_m.fxe"},
	{0x1CD81AA6F3F49D16ULL, "vxca_caverns1_mines_ii0_0_0.vlm"},
	{0x1CD88285AD04892EULL, "6133458_m.fxe"},
	{0x1CD8D30D6DB2BF0FULL, "6053915_m.fxe"},
	{0x1CD8DDDD85687935ULL, "vxht_0_hightown_a3_c6l_0_0.vlm"},
	{0x1CD907567EF67FC8ULL, "6127664_m.fxe"},
	{0x1CD92804DFC48AF1ULL, "vxca_cavern1_haunted_ij6_0_0.vlm"},
	{0x1CD95903DDA26564ULL, "197489811.ogg"},
	{0x1CD986D9B0DDC544ULL, "w_encounter3_230_0.msh"},
	{0x1CD998D9B0DDE3ADULL, "w_encounter3_230_0.mao"},
	{0x1CD99CD9B0DDEAE6ULL, "w_encounter3_230_0.mmh"},
	{0x1CDA07B9275C2F10ULL, "sst_ss_qunari_m_2.gda"},
	{0x1CDA46E19DD5108DULL, "458505604.ogg"},
	{0x1CDAA02D0BED3A4EULL, "lt_0_lowtown_docks_a3_229_0.msh"},
	{0x1CDAA7B5E8A861D4ULL, "vxw_woundedcoast_uf_0_0.vlm"},
	{0x1CDAAE2D0BED519FULL, "lt_0_lowtown_docks_a3_229_0.mao"},
	{0x1CDABA2D0BED663CULL, "lt_0_lowtown_docks_a3_229_0.mmh"},
	{0x1CDB07D86280276CULL, "6075703_m.fxe"},
	{0x1CDB10EDFD6FD8BEULL, "hlac_lt_0_lowtown_a3_n_c_65538_3y.dds"},
	{0x1CDB2429736D0296ULL, "927558202.ogg"},
	{0x1CDB3935C5A34423ULL, "967130120.ogg"},
	{0x1CDB820797B39FBEULL, "6016463_m.fxe"},
	{0x1CDC346F7AEB4577ULL, "vxlt_gallowscourtyard_n_7ep_0_0.vlm"},
	{0x1CDC5D2EBFA3DE52ULL, "vxdr_primevaltomb_8vp_0_0.vlm"},
	{0x1CDC925E10A67B04ULL, "mcr371_orsino6083320.cl"},
	{0x1CDCD2ECD39497EBULL, "25929642.ogg"},
	{0x1CDDFA3BA5979174ULL, "hlac_dr_slavers_3214_12y.dds"},
	{0x1CDE440F86E9C42EULL, "lt_0_lowtown_docks_a3_157_0.msh"},
	{0x1CDE520F86E9DC7FULL, "lt_0_lowtown_docks_a3_157_0.mao"},
	{0x1CDE5E0F86E9F01CULL, "lt_0_lowtown_docks_a3_157_0.mmh"},
	{0x1CDEC49B29B19207ULL, "mag160_varian_ilithis6016508.cl"},
	{0x1CDEFCD8A2259AA8ULL, "static_staticreliefmap_p1.gsh.fxo.dx11"},
	{0x1CDF0375D7129CE4ULL, "lt_0_lowtown_docks_a3_186_0.msh"},
	{0x1CDF045BD8407774ULL, "gen00fl_aveline6139914.cl"},
	{0x1CDF1575D712BB4DULL, "lt_0_lowtown_docks_a3_186_0.mao"},
	{0x1CDF1975D712C186ULL, "lt_0_lowtown_docks_a3_186_0.mmh"},
	{0x1CDF62ED60ABF31AULL, "art\\vfx\\combat\\vfx_g_dual_mode_c.anb"},
	{0x1CDFBF7145BA43F7ULL, "lt_0_lowtown_docks_n_12_0.anb"},
	{0x1CE07016AA08D0E3ULL, "1036416640.ogg"},
	{0x1CE18117B217FD20ULL, "vxlt_shop_1w_0_0.vlm"},
	{0x1CE1B651279AF83AULL, "955447475.ogg"},
	{0x1CE1BFAA57EB5DE5ULL, "6126837_m.fxe"},
	{0x1CE1C2AAE36E91A2ULL, "vxht_bartrand_n_5dn_0_0.vlm"},
	{0x1CE1CB0289795FF5ULL, "hlac_w_sundermount_65535_0x.dds"},
	{0x1CE1F163E98F3200ULL, "gen_im_acc_amu_act2_04.uti"},
	{0x1CE26D680E0E5D0FULL, "vxw_encounter2_camp_l3_0_0.vlm"},
	{0x1CE2B3FC1B1A14D7ULL, "vxlt_0_lowtown_docks_c4d_0_0.vlm"},
	{0x1CE2ED3D4130A49BULL, "vxw_encounter2_j1_0_0.vlm"},
	{0x1CE2FA8633AA5B96ULL, "hlac_lt_0_lowtown_a3_n_c_65536_11z.dds"},
	{0x1CE326DEE64E621DULL, "vxw_woundedcoast_1gp_0_0.vlm"},
	{0x1CE437987B9CE145ULL, "mag311_huon6042705.cl"},
	{0x1CE45A730D04B668ULL, "6143679_m.fxe"},
	{0x1CE4F98F76AFF812ULL, "249953327.ogg"},
	{0x1CE513771E26AF9EULL, "6173142_m.fxe"},
	{0x1CE55949233FD41DULL, "547958517.ogg"},
	{0x1CE56E4EE81DC9AAULL, "vxht_bartrand_n_6in_0_0.vlm"},
	{0x1CE591632B8EF886ULL, "417467580.ogg"},
	{0x1CE5BBF3DD4F06C0ULL, "916167596.ogg"},
	{0x1CE7387CDAFA053FULL, "vxht_mansion2b_2oi_0_0.vlm"},
	{0x1CE7418B50E0A3D9ULL, "gen00fl_fenris6127094.cl"},
	{0x1CE7669D70577F6FULL, "6202267_m.fxe"},
	{0x1CE773A54A2E82D2ULL, "hlac_w_bonepitmining_a2_2_0x.dds"},
	{0x1CE7BD27E1A2C104ULL, "vxlt_0_lowtown_a2_bz5_0_0.vlm"},
	{0x1CE7FB6E7F823C83ULL, "vxlt_0_lowtown_n_cmu_0_0.vlm"},
	{0x1CE82A27D1C4D610ULL, "929139939.ogg"},
	{0x1CE88776D8AAF162ULL, "vxlt_gallowscourtyard_6yo_0_0.vlm"},
	{0x1CE89EED7C66699EULL, "vxca_caverns2prc_97a_0_0.vlm"},
	{0x1CE8B416AF4F8DD9ULL, "hlac_ht_mansion2_818_41x.dds"},
	{0x1CE91992DA9E7A54ULL, "vxw_woundedcoast_rg_0_0.vlm"},
	{0x1CE97BF6B2DAF6D7ULL, "vxht_keep_n_5ki_0_0.vlm"},
	{0x1CEA131A8B400F63ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlm_meda_0s.dds"},
	{0x1CEA2BC72EB4D034ULL, "projdecal_projdecal_p1.psh.fxo"},
	{0x1CEA4325B66464BAULL, "llac_lt_0_lowtown_docks_a3_n_-4_1.dds"},
	{0x1CEA67B78DA6D8C4ULL, "hlac_ht_0_hightown_a3_65537_0z.dds"},
	{0x1CEA6871AB66E3E4ULL, "327680251.ogg"},
	{0x1CEAA64C6CE15DF7ULL, "vxlt_gallowscourtyard_n_7hs_0_0.vlm"},
	{0x1CEAEEBCCCDC11B3ULL, "6133833_m.fxe"},
	{0x1CEB771DD4D105C4ULL, "one100dg_nabil6033137.cl"},
	{0x1CEBA8B556808015ULL, "6124739_f.fxe"},
	{0x1CEBB490C608E13FULL, "6074332_m.fxe"},
	{0x1CEBBFDD0CE67815ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb105a.msh"},
	{0x1CEBC1DD0CE67B8BULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb105a.mmh"},
	{0x1CEBECC06511475FULL, "95714115.ogg"},
	{0x1CEC6992EDB8DBE7ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_aoe_p.anb"},
	{0x1CECB32573181039ULL, "182947008.ogg"},
	{0x1CECB79B6EEA7F76ULL, "263681624.ogg"},
	{0x1CECFD9AE8FBD51FULL, "art\\characters\\creatures\\mubari\\animation\\cwkf_old.gad"},
	{0x1CED00A90905EB09ULL, "vxlt_gallowstemplar_55u_0_0.vlm"},
	{0x1CED0174B1EF7EBFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07b.mmh"},
	{0x1CED0774B1EF8961ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07b.msh"},
	{0x1CED3D537D8E8F91ULL, "435992276.ogg"},
	{0x1CED6F32533C35F2ULL, "mcr370cr_samson.utc"},
	{0x1CEDA34C1A97D77DULL, "m_and111ar_andersshop.bnk"},
	{0x1CEDDFBDF6106B9AULL, "6149867_f.fxe"},
	{0x1CEDFB7606754E25ULL, "717761554.ogg"},
	{0x1CEE0AAA825EAA04ULL, "vxdr_primevaltomb_5b3_0_0.vlm"},
	{0x1CEF6720CE6106E5ULL, "129960349.ogg"},
	{0x1CEFAD9A04B9DCC2ULL, "em_gen_dalish_assassin.mor"},
	{0x1CF01991B2E5392FULL, "vxlt_0_lowtown_docks_cus_0_0.vlm"},
	{0x1CF04362C9AB2E2DULL, "character_character_p26.psh.fxo.dx11"},
	{0x1CF043BE4568882EULL, "gen00fl_fenris6112020.cl"},
	{0x1CF060B979DE037FULL, "vxw_encounter2_camp_pw_0_0.vlm"},
	{0x1CF085038CF9495CULL, "98768367.ogg"},
	{0x1CF0ADFC9E6767E3ULL, "vxw_encounter1_db_0_0.vlm"},
	{0x1CF0CBC932AB9FA5ULL, "vxht_keep_n_6zw_0_0.vlm"},
	{0x1CF0DF450004758FULL, "vxw_encounter2_camp_g2_0_0.vlm"},
	{0x1CF1135C13E885A8ULL, "6125918_m.fxe"},
	{0x1CF125467C8D9477ULL, "hlac_lt_darktown_-1_64y.dds"},
	{0x1CF1295D39FC6BEFULL, "qun160ip_chest_2.utp"},
	{0x1CF17D1D11C91BA8ULL, "vxw_encounter2_ga_0_0.vlm"},
	{0x1CF1A454943D802AULL, "qun200_party_comment6090457.cl"},
	{0x1CF1D10274C82B60ULL, "hlac_ht_0_hightown_a3_65533_2x.dds"},
	{0x1CF23C069F0E6FABULL, "6128711_f.fxe"},
	{0x1CF2D96820CED93DULL, "vxw_encounter2_camp_8u_0_0.vlm"},
	{0x1CF2FEDEADBBE527ULL, "vxlt_0_lowtown_a2_gcl_0_0.vlm"},
	{0x1CF3E02BDA13E22AULL, "vxht_bartrand_n_6ro_0_0.vlm"},
	{0x1CF3E2F5F6155C3CULL, "6077600_m.fxe"},
	{0x1CF3F29AE901E10CULL, "art\\characters\\creatures\\mubari\\animation\\cwkf_old.evt"},
	{0x1CF4490A1654F4ACULL, "597674592.ogg"},
	{0x1CF4903F682485F5ULL, "7138558.ogg"},
	{0x1CF4C08348A33950ULL, "dae000_oldmangossip6174902.cl"},
	{0x1CF59662E8996B87ULL, "zz_dre_debug6168226.cl"},
	{0x1CF5C32D83335470ULL, "mcr310_party_banter6087805.cl"},
	{0x1CF6FEF09F0D92B8ULL, "6114851_m.fxe"},
	{0x1CF724577347E792ULL, "art\\levels\\ztools\\default\\generic\\0cc_basic.phy"},
	{0x1CF76A484AB36C5FULL, "269770264.ogg"},
	{0x1CF7B1434CBD0F19ULL, "hlac_lt_0_lowtown_-65537_3z.dds"},
	{0x1CF818A0BEB0ABEBULL, "vxw_encounter2_bi_0_0.vlm"},
	{0x1CF88C73D3C84373ULL, "hlac_dr_primevaltomb_659_0y.dds"},
	{0x1CF89326472F1EB8ULL, "828229355.ogg"},
	{0x1CF8DFA97195277AULL, "750484756.ogg"},
	{0x1CF8F32B146FFEC5ULL, "mag100_karras6035367.cl"},
	{0x1CF954B8DFA71901ULL, "vxht_0_hightown_n_b8y_0_0.vlm"},
	{0x1CF9630E149DCCFEULL, "6076589_f.fxe"},
	{0x1CFA1AD5F16F3724ULL, "414908291.ogg"},
	{0x1CFA266DB7FEF9BFULL, "6079032_m.fxe"},
	{0x1CFAF41CC0465E97ULL, "horizontalblur7.vsh.fxo"},
	{0x1CFB33ADF29A8B64ULL, "llac_lt_0_lowtown_docks_a3_n_-131073_4.dds"},
	{0x1CFB4C46D0F424CDULL, "hlac_w_encounter1_65538_0y.dds"},
	{0x1CFB6653B51AB042ULL, "120655579.ogg"},
	{0x1CFC2AC2F1ADE3FBULL, "and111_lirene6044964.cl"},
	{0x1CFC7ED54891F7F0ULL, "hlac_lt_0_lowtown_a2_n_c_65535_0z.dds"},
	{0x1CFCF0DB8089C46CULL, "dae000_arishok6067827.cl"},
	{0x1CFDA1C9EE31BEE2ULL, "6146963_f.fxe"},
	{0x1CFDDA7D3516F057ULL, "vxdr_slavers_1jh_0_0.vlm"},
	{0x1CFDE6A7F398C25FULL, "6144102_f.fxe"},
	{0x1CFE918A6C536165ULL, "1063368231.ogg"},
	{0x1CFE9D67CB04D8DFULL, "vxlt_gallowscourtyard_78t_0_0.vlm"},
	{0x1CFED2966BD41BFFULL, "vxw_encounter2_camp_kv_0_0.vlm"},
	{0x1CFED53E6571A8F3ULL, "6063952_m.fxe"},
	{0x1CFED6A153D3940EULL, "6172930_m.fxe"},
	{0x1CFEF7B5AB93FF73ULL, "285204693.ogg"},
	{0x1CFF03474126FE3BULL, "mcr310_sibling6076340.cl"},
	{0x1CFF1FD9905D8063ULL, "vxw_encounter1_ie_0_0.vlm"},
	{0x1CFFD0F7251CD59BULL, "vxw_encounter2_p3_0_0.vlm"},
	{0x1CFFE0BBAC3537F8ULL, "6080627_f.fxe"},
	{0x1D00095CE3849FF6ULL, "hlac_ht_mansion_n_3370_1z.dds"},
	{0x1D00144E0E42F264ULL, "w_bonepit_121_0.msh"},
	{0x1D00264E0E4310CDULL, "w_bonepit_121_0.mao"},
	{0x1D002A4E0E431706ULL, "w_bonepit_121_0.mmh"},
	{0x1D003F930E2F13A1ULL, "500569143.ogg"},
	{0x1D008ABCAEA492E5ULL, "6135918_m.fxe"},
	{0x1D00AE3902787558ULL, "vxlt_0_lowtown_a3_n_c_df3_0_0.vlm"},
	{0x1D014B4512C4F1BDULL, "vxw_encounter2_camp_3t_0_0.vlm"},
	{0x1D0151121E15CC26ULL, "6135114_f.fxe"},
	{0x1D016C17D8783901ULL, "vxht_keep_n_6jk_0_0.vlm"},
	{0x1D0182D5F7E3449FULL, "638273606.ogg"},
	{0x1D01B805EF43BA5BULL, "814713607.ogg"},
	{0x1D01DA9AE90DF665ULL, "art\\characters\\creatures\\mubari\\animation\\cwkf_old.ani"},
	{0x1D01ECA3F0C90265ULL, "vxlt_0_lowtown_gbm_0_0.vlm"},
	{0x1D027188D48776B1ULL, "dae000_worthy6166613.cl"},
	{0x1D02E024C383E125ULL, "vxlt_house_g0_0_0.vlm"},
	{0x1D02EAA1448B6529ULL, "396339554.ogg"},
	{0x1D041C36BEE6363FULL, "vxht_mansion2b_2ek_0_0.vlm"},
	{0x1D04F9F829287E24ULL, "vxca_caverns1_mines_iir_0_0.vlm"},
	{0x1D050BC16D98A190ULL, "w_woundedcoast_80_0.rml"},
	{0x1D054BDA95374000ULL, "hlac_w_bonepit_458753_0z.dds"},
	{0x1D056A9AE0126470ULL, "hlac_w_blightlands_0_1x.dds"},
	{0x1D05901B4993C30AULL, "vxlt_0_lowtown_a2_cmk_0_0.vlm"},
	{0x1D068A7DB0A6C46BULL, "vxw_encounter2_mh_0_0.vlm"},
	{0x1D069C2EA0DAF378ULL, "132707428.ogg"},
	{0x1D0837B1EC2880B9ULL, "6099528_m.fxe"},
	{0x1D0865ED60CEFDCDULL, "art\\vfx\\combat\\vfx_g_dual_mode_c.mmh"},
	{0x1D086F9712824EA9ULL, "dae300_alistair_king6134895.cl"},
	{0x1D087628EA996CB1ULL, "530159866.ogg"},
	{0x1D087D3733410CDCULL, "pro000_cut_flemeth_rescue.cnv"},
	{0x1D08D03CD29DB82FULL, "vxdr_primevalentrance_2ef_0_0.vlm"},
	{0x1D093B567600F1D7ULL, "6177918_m.fxe"},
	{0x1D097A9699DBCFF4ULL, "vxw_encounter2_5k_0_0.vlm"},
	{0x1D09AF4F746D02B8ULL, "zz_rogueflips.cut"},
	{0x1D0A0DBC03BC62C6ULL, "mer320_hubert6053747.cl"},
	{0x1D0A214E0E4B5CD1ULL, "w_bonepit_121_0.phy"},
	{0x1D0AAB5BC4FE83DAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba02a.mmh"},
	{0x1D0ABD5BC4FEA220ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba02a.msh"},
	{0x1D0B6272E3E6B4FFULL, "6076806_m.fxe"},
	{0x1D0B9F3733437C2CULL, "pro000_cut_flemeth_rescue.bnk"},
	{0x1D0BB0FB468D23B8ULL, "vxw_bonepitdestroyed_a3_l0_0_0.vlm"},
	{0x1D0BF0F6EED0E640ULL, "6095433_m.fxe"},
	{0x1D0BFA650E272AF8ULL, "275705417.ogg"},
	{0x1D0C08B7D9A422FEULL, "22352525.ogg"},
	{0x1D0C3C396343FC97ULL, "hlac_w_woundedcoast_-262138_0y.dds"},
	{0x1D0CA1B438038856ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_exit.ani"},
	{0x1D0CBC7145E12C0BULL, "lt_0_lowtown_docks_n_12_0.rml"},
	{0x1D0CD013EB0F3621ULL, "vxdr_primevaltomb_8yf_0_0.vlm"},
	{0x1D0D44735DCA347FULL, "vxw_encounter2_camp_fq_0_0.vlm"},
	{0x1D0D7D2034406E81ULL, "vxlt_0_lowtown_a2_eop_0_0.vlm"},
	{0x1D0DE9E75DCC1041ULL, "19616031.ogg"},
	{0x1D0DF28141D431B7ULL, "6154939_m.fxe"},
	{0x1D0DFB6F80C127BBULL, "vxlt_0_lowtown_a2_e2u_0_0.vlm"},
	{0x1D0E05CA41373697ULL, "6134941_f.fxe"},
	{0x1D0E39BA605F0F0AULL, "zz_firestaff2_v5.cut"},
	{0x1D0EA272E6943D21ULL, "6106178_m.fxe"},
	{0x1D0EACB712D5C74AULL, "vxlt_foundry_n_10m_0_0.vlm"},
	{0x1D0ECE4BE8366A0CULL, "387994958.ogg"},
	{0x1D0F50BFA65E51CAULL, "6170932_m.fxe"},
	{0x1D0F80E3A05CA32AULL, "6165123_m.fxe"},
	{0x1D105B9B486760D4ULL, "follower_banter6169865.cl"},
	{0x1D10812906DF30F8ULL, "clf221_jalen6089770.cl"},
	{0x1D10BEE8E5621633ULL, "vxca_caverns1_jdh_0_0.vlm"},
	{0x1D10F7CAD8B0F2C0ULL, "mag210_arianni6031650.cl"},
	{0x1D1161CB6BB3DF1FULL, "458082895.ogg"},
	{0x1D11B35AF47CB32EULL, "ser270_emeric6111398.cl"},
	{0x1D125E95735C9192ULL, "vxw_woundedcoast_1cg_0_0.vlm"},
	{0x1D131FA1B6071B55ULL, "lt_0_lowtown_a2_n_53_0_ll.mmh"},
	{0x1D1339A4F2DA6DF0ULL, "6080923_m.fxe"},
	{0x1D135AC8606FD601ULL, "vxdr_primevalentrance_31y_0_0.vlm"},
	{0x1D13A508516C0842ULL, "215224598.ogg"},
	{0x1D13B349DA5A779BULL, "629867979.ogg"},
	{0x1D13BBCD3BBDC8EDULL, "gen00fl_bethany6121324.cl"},
	{0x1D13C9DD17928073ULL, "cod_shw_mer_notes_2.utp"},
	{0x1D13D1842F554B46ULL, "ht_mansion2b_mmap.dds"},
	{0x1D140F9FAE064D62ULL, "art\\vfx\\combat\\vfx_glb_ele_e_extreme_c.mmh"},
	{0x1D143A185DA09E8DULL, "570460915.ogg"},
	{0x1D146692EDDA28E8ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_aoe_p.mmh"},
	{0x1D1489BC9B43FFD8ULL, "gen00fl_aveline6131895.cl"},
	{0x1D14FC5AA29CDCEBULL, "vxw_encounter2_hk_0_0.vlm"},
	{0x1D156014B12473FCULL, "llac_ca_caverns2prc_5802_13.dds"},
	{0x1D15A1E5A94E3D83ULL, "100410177.ogg"},
	{0x1D163D263CC29827ULL, "6039762_m.fxe"},
	{0x1D16936D0A301907ULL, "6075348_m.fxe"},
	{0x1D16998521D70A4EULL, "128105236.ogg"},
	{0x1D16AF1C397191F6ULL, "follower_banter6149238.cl"},
	{0x1D16F57A198A0472ULL, "gen00fl_anders6113102.cl"},
	{0x1D172562ADCB8180ULL, "vxht_keep_n_63m_0_0.vlm"},
	{0x1D17BD316CF677A9ULL, "6086894_m.fxe"},
	{0x1D180BD9F9044AF3ULL, "15044209.ogg"},
	{0x1D18B88D5CD6947AULL, "6084518_m.fxe"},
	{0x1D18C1899A9CA8FFULL, "follower_banter6154051.cl"},
	{0x1D1920E965E84D81ULL, "pro000ip_wesleys_ring.uti"},
	{0x1D19898AF8725426ULL, "665855140.ogg"},
	{0x1D1993DD5153A80EULL, "861289422.ogg"},
	{0x1D19970B18775C9DULL, "vxlt_hangedman_n_2zm_0_0.vlm"},
	{0x1D1999446CF654F4ULL, "588978672.ogg"},
	{0x1D19A9B4380E2047ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_exit.evt"},
	{0x1D19EF79A8435D29ULL, "fen110_party_comment6095225.cl"},
	{0x1D1A4A388AA40A5CULL, "tent_sa.dds"},
	{0x1D1ABFEC0AE3DAC0ULL, "gen00fl_aveline6139525.cl"},
	{0x1D1AC45C5B8914B4ULL, "6078790_f.fxe"},
	{0x1D1AE78D7E77A76CULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_dieloop.ani"},
	{0x1D1AFB809A6BBF4AULL, "6118273_m.fxe"},
	{0x1D1B1D8A938018E8ULL, "hlac_w_sundermount_-131067_0x.dds"},
	{0x1D1BB0AAA8531B06ULL, "hlac_ht_0_hightown_a3_65533_3y.dds"},
	{0x1D1BB6504FC04CFFULL, "vxw_encounter2_camp_ap_0_0.vlm"},
	{0x1D1C5DDA96F62508ULL, "6042280_m.fxe"},
	{0x1D1C994B875C266BULL, "508056647.ogg"},
	{0x1D1CB9376F6A75FCULL, "gen00fl_varric6093668.cl"},
	{0x1D1CC94CCCDD8EFFULL, "vxht_mansion2b_2wp_0_1.vlm"},
	{0x1D1CF686226C4411ULL, "6121916_f.fxe"},
	{0x1D1CFB87241BD7E5ULL, "519428387.ogg"},
	{0x1D1D0FC394FD29DAULL, "gen00fl_fenris6126698.cl"},
	{0x1D1D160BF38A8A6CULL, "art\\characters\\playercharacter\\humanfemale\\mt_eye_flma.mao"},
	{0x1D1D3C3C2B615989ULL, "hlac_w_encounter3_-65533_0y.dds"},
	{0x1D1D761EE7891655ULL, "hm_amb_darktown_lowlife.mor"},
	{0x1D1E03DF56E9F6C8ULL, "gencr_trap_explosives_2.utc"},
	{0x1D1E0C44FE7DA3BBULL, "vxw_encounter3_dh_0_0.vlm"},
	{0x1D1E4CF9E98FB11FULL, "vxlt_0_lowtown_docks_a3_n_cxp_0_0.vlm"},
	{0x1D1E5938E83913DFULL, "llac_lt_0_lowtown_docks_0_1.dds"},
	{0x1D1ED1A80B787C62ULL, "gen_im_trs_gen_hvy_glv_r0.uti"},
	{0x1D1EE726EE0D6E0AULL, "898080580.ogg"},
	{0x1D1EFE58D4E06DF1ULL, "gen00fl_anders6119875.cl"},
	{0x1D1F192D7DC01705ULL, "hlac_lt_blackemporium_1055_19y.dds"},
	{0x1D1F9C0E0B9C0FACULL, "dae011_gossips6149799.cl"},
	{0x1D1FAD394AAF68ADULL, "qun160_party_comment_26065003.cl"},
	{0x1D20CAB438147690ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_exit.gad"},
	{0x1D20F48D7E7C28AEULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_dieloop.gad"},
	{0x1D2184B23B542F6DULL, "hlac_dr_dungeon_3214_10x.dds"},
	{0x1D2189D2085A6DBAULL, "ht_bartrand_n_3374.rml"},
	{0x1D2196AF71F6CB62ULL, "art\\levels\\races\\natural\\proxy\\foliage\\pine.mao"},
	{0x1D219B75FC156264ULL, "510066921.ogg"},
	{0x1D219CDA163D6C63ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_6x.dds"},
	{0x1D21DDB20D14AF24ULL, "vxca_caverns1_mines_icp_0_0.vlm"},
	{0x1D21FC6E5A843123ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge13.ani"},
	{0x1D22490CFC131912ULL, "hlac_ht_mansion2b_818_39z.dds"},
	{0x1D22B19F6815A89BULL, "llac_ht_0_hightown_a3_65535_1.dds"},
	{0x1D2327619F5BA4DBULL, "599722250.ogg"},
	{0x1D2357DD8112CA09ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medd.phy"},
	{0x1D2392A32B4E6CC6ULL, "vxht_bartrand_n_66o_0_0.vlm"},
	{0x1D23C14CD0D89583ULL, "hlac_ht_0_hightown_n_65536_2x.dds"},
	{0x1D2459478937C353ULL, "framebuffereffects_distortion_p0.psh.fxo.dx11"},
	{0x1D249D8F393EB2B2ULL, "vxlt_0_lowtown_docks_n_bd3_0_0.vlm"},
	{0x1D24C260AFA81E8FULL, "t3_wpn_ash.tnt"},
	{0x1D24C8285F106B94ULL, "dae000_merchant_hm_mob6174226.cl"},
	{0x1D24E46D2DB10958ULL, "llac_lt_0_lowtown_a3_n_c_131071_4.dds"},
	{0x1D25A243FEFFDC7BULL, "gen00fl_merrill6150317.cl"},
	{0x1D26B1ACA80C3574ULL, "vxw_blightlands_gb_0_0.vlm"},
	{0x1D276E3E78FB2DA8ULL, "501527332.ogg"},
	{0x1D280C881814E051ULL, "pro000_merchant6051632.cl"},
	{0x1D2858A54C3BB27AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad_12.pwk"},
	{0x1D28629CC75ACF05ULL, "ser221_party_comments6166111.cl"},
	{0x1D286D7A27430EF1ULL, "6146888_m.fxe"},
	{0x1D2A470DB238A8D8ULL, "hlac_ht_playershouse_n_855_32x.dds"},
	{0x1D2A82E45B196508ULL, "189999092.ogg"},
	{0x1D2A933CFA8401A5ULL, "vxht_keep_n_6fs_0_0.vlm"},
	{0x1D2AA9E085FADC0AULL, "6083371_m.fxe"},
	{0x1D2B45DC9E09584BULL, "hlac_lt_darktown_-1_86y.dds"},
	{0x1D2B7FE16C3E4C3EULL, "vxht_bartrand_n_739_0_0.vlm"},
	{0x1D2BBCB8C9BBFFE2ULL, "225115846.ogg"},
	{0x1D2BC8BE21319A14ULL, "6165271_m.fxe"},
	{0x1D2C3790F97213EBULL, "451673427.ogg"},
	{0x1D2C47C779DCA5A1ULL, "6167474_m.fxe"},
	{0x1D2C7E21F073BC3BULL, "vxw_encounter3_ak_0_0.vlm"},
	{0x1D2D5DBA925CB838ULL, "lgt310ip_body_2.utp"},
	{0x1D2DCFED2451FC5AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chaircutup.mmh"},
	{0x1D2DD3ED24520291ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chaircutup.mao"},
	{0x1D2DDCC0DFFF9A05ULL, "dae000_arishok6068487.cl"},
	{0x1D2DE1ED24521AA0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chaircutup.msh"},
	{0x1D2E0B5FF25B7363ULL, "dae000ip_crowd_mf3_talk17.utp"},
	{0x1D2E7002FB44B2BAULL, "6062987_m.fxe"},
	{0x1D2EB3B6C23FE208ULL, "259363381.ogg"},
	{0x1D2F1BCDB7426108ULL, "vxw_bonepitmining_a2_fh_0_0.vlm"},
	{0x1D2F424F5748C292ULL, "vxw_woundedcoast_1ia_0_0.vlm"},
	{0x1D2F797B8CF9BF8BULL, "zz_followers_banter6159270.cl"},
	{0x1D303AA42F31D998ULL, "6048479_m.fxe"},
	{0x1D30CA5BC51F87E5ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba02a.phy"},
	{0x1D318FE1DFADD40DULL, "tbe100im_magister_staff.uti"},
	{0x1D31D372BDBC9776ULL, "hlac_lt_backalley_g_1_0y.dds"},
	{0x1D3204801D448546ULL, "vxht_bartrand_n_63p_0_0.vlm"},
	{0x1D3223CB537F2D0AULL, "vxlt_gallowscourtyard_6x4_0_0.vlm"},
	{0x1D3242825B39EE1CULL, "vxht_0_hightown_a3_by9_0_0.vlm"},
	{0x1D32A5C5507AD8E5ULL, "6162727_m.fxe"},
	{0x1D32BA0D091FC4F1ULL, "clf360_evets6072220.cl"},
	{0x1D334A696E2E76B3ULL, "vxw_encounter2_camp_nm_0_0.vlm"},
	{0x1D341CA6AE38DACDULL, "llac_ht_0_hightown_n_-131071_1.dds"},
	{0x1D34BA9187D50AE7ULL, "579859413.ogg"},
	{0x1D3523899A024DF4ULL, "vxw_blightlands_hc_0_0.vlm"},
	{0x1D352D2B347FC38FULL, "fen210_orana6109681.cl"},
	{0x1D3591B0584C901CULL, "658136087.ogg"},
	{0x1D35BB2FA2D5CB2AULL, "static_static_p29.psh.fxo.dx11"},
	{0x1D35D6D7285410C9ULL, "856200675.ogg"},
	{0x1D364CC0E3960AFAULL, "6156855_m.fxe"},
	{0x1D37136E5A96B40DULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge13.gad"},
	{0x1D37C522E64132C1ULL, "art\\vfx\\textures\\vfx_glyph_repul.dds"},
	{0x1D3852AF55ECF0EEULL, "vxca_caverns1_ii9_0_0.vlm"},
	{0x1D38692616E5F6B9ULL, "vxht_brothel_n_5kd_0_0.vlm"},
	{0x1D38A94BBEBF5EF4ULL, "hlac_w_encounter2_131073_0y.dds"},
	{0x1D38EFA298EFCB12ULL, "hlac_lt_0_lowtown_a2_65534_6z.dds"},
	{0x1D390519EC7A1A25ULL, "vxht_keep_n_6ar_0_0.vlm"},
	{0x1D3908BFD9BBDB3BULL, "vxht_mansion2_35b_0_0.vlm"},
	{0x1D393E26C51F2E2DULL, "vxlt_0_lowtown_a2_n_gxl_0_0.vlm"},
	{0x1D398EF379298185ULL, "mcr310_argument6074341.cl"},
	{0x1D3B06AA81A90309ULL, "vxca_caverns2_8ap_0_0.vlm"},
	{0x1D3B12F668998BB6ULL, "103072698.ogg"},
	{0x1D3B283E177299DAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\til_lowtown_main_drelief.dds"},
	{0x1D3B64EC9841C224ULL, "vxht_mansion2b_38f_0_0.vlm"},
	{0x1D3BE6FDC8EC2461ULL, "dae100cr_qunari_warrior.utc"},
	{0x1D3CEB78E89D3D0AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutan.mao"},
	{0x1D3CEF78E89D4459ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutan.mmh"},
	{0x1D3CF978E89D5547ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutan.msh"},
	{0x1D3D0E9FAE295009ULL, "art\\vfx\\combat\\vfx_glb_ele_e_extreme_c.anb"},
	{0x1D3D8DAAA9387988ULL, "vxw_bonepitmining_a2_mi_0_0.vlm"},
	{0x1D3DF09C3E1DC854ULL, "dae_warrior_singletarget.ncs"},
	{0x1D3E26164435E0E1ULL, "6042794_f.fxe"},
	{0x1D3E4203F0AD0949ULL, "vxht_0_hightown_a2_cap_0_0.vlm"},
	{0x1D3E5C1C076B9299ULL, "hlac_dr_primevaltomb_659_1x.dds"},
	{0x1D3EA2CF777D3756ULL, "413376064.ogg"},
	{0x1D3EE74CE9E17C91ULL, "89234087.ogg"},
	{0x1D3EECEE799B1022ULL, "zz_vault_debug6177776.cl"},
	{0x1D3F853DCE3B50C7ULL, "vxca_caverns1_mines_giw_0_0.vlm"},
	{0x1D3FFEF5B375A7B3ULL, "hlac_lt_0_lowtown_docks_-4_4z.dds"},
	{0x1D4003CDAC1F2280ULL, "fx_icicle.mao"},
	{0x1D40DB58488A8A88ULL, "isa321_velasco6094486.cl"},
	{0x1D41B1244DDFF24CULL, "pro000_merchant6051248.cl"},
	{0x1D41BC4660248F33ULL, "vxw_encounter2_camp_il_0_0.vlm"},
	{0x1D421CF3DF297790ULL, "vxht_0_hightown_9zj_0_0.vlm"},
	{0x1D427D5CAC5D4CC0ULL, "llac_dr_slavers_3217_9.dds"},
	{0x1D4299C862B2F3B6ULL, "481965100.ogg"},
	{0x1D42B0010ECC461AULL, "hlac_ca_caverns1_mines_24424_16y.dds"},
	{0x1D433F8B4FC2F19FULL, "6048252_m.fxe"},
	{0x1D43FB4F1ED05322ULL, "6164700_m.fxe"},
	{0x1D44A64532351756ULL, "static_static_p61.psh.fxo.dx11"},
	{0x1D4501958EB68D99ULL, "vxht_mansion_n_5ai_0_0.vlm"},
	{0x1D4528DD2976322FULL, "6037024_f.fxe"},
	{0x1D455E8B45DAC430ULL, "llac_ca_caverns1_mines_24423_7.dds"},
	{0x1D456D40181B9862ULL, "vxw_encounter3_f3_0_0.vlm"},
	{0x1D46618C74C642C7ULL, "prm000_varric_crest6207405.cl"},
	{0x1D46DA9A46D2AC18ULL, "547673042.ogg"},
	{0x1D473C06077DB7F0ULL, "6046833_m.fxe"},
	{0x1D47F7D6636B206AULL, "6084477_m.fxe"},
	{0x1D4899969DB104B8ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_hcliff2_d.dds"},
	{0x1D48C0450E146B49ULL, "hlac_lt_0_lowtown_-65535_5z.dds"},
	{0x1D497887739F1B89ULL, "vxca_caverns2_8lw_0_0.vlm"},
	{0x1D49D9B7C8040B41ULL, "vxht_mansion2b_219_0_0.vlm"},
	{0x1D49EE0AA0064184ULL, "vxw_bonepit_80_0_0.vlm"},
	{0x1D4A02560F373418ULL, "lt_0_lowtown_docks_a3_46_0.rml"},
	{0x1D4A1D8DCF2269E3ULL, "vxw_bonepitdestroyed_a3_26u_0_0.vlm"},
	{0x1D4A9DEA320C3DC6ULL, "plt_cod_bks_lite_tombs_2.plo"},
	{0x1D4B7118818D1A3AULL, "vxht_mansion2_26b_0_1.vlm"},
	{0x1D4BCC53DA54A855ULL, "combat_smaoe_r2_a1.gda"},
	{0x1D4BD8F923682898ULL, "gen_im_trs_chn_blt_r0.uti"},
	{0x1D4BFF879B2E9208ULL, "vxw_bonepitmining_a2_hf_0_0.vlm"},
	{0x1D4CA48D49E3A23DULL, "32890698.ogg"},
	{0x1D4D0A1A3D8F5491ULL, "6043523_m.fxe"},
	{0x1D4D76A05F5D09A4ULL, "qcr000_arishok6042516.cl"},
	{0x1D4D83C2BA9A77CBULL, "vxht_0_hightown_a3_b5n_0_0.vlm"},
	{0x1D4D84E9E3869985ULL, "ave200_aveline_bar6116374.cl"},
	{0x1D4D88DB5D9D7AD0ULL, "1006272440.ogg"},
	{0x1D4D9A7284AF578EULL, "genip_trap_lever.utp"},
	{0x1D4DEA987DB9F0BBULL, "202503012.ogg"},
	{0x1D4E3D981F1766D2ULL, "vxw_woundedcoast_mx_0_0.vlm"},
	{0x1D4EE83A0130B646ULL, "vxht_bartrand_n_69r_0_0.vlm"},
	{0x1D4F0A74B2425804ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07b.phy"},
	{0x1D4F40DD8137DE2EULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medd.mmh"},
	{0x1D4F5ADD81380A1CULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medd.msh"},
	{0x1D502ABF9D4DC97EULL, "hlac_lt_gallowstemplar_2383_2z.dds"},
	{0x1D502FB90FE3425AULL, "hlac_lt_0_lowtown_a2_131072_5z.dds"},
	{0x1D50540B80832EECULL, "zz_lgt_2_debug6206580.cl"},
	{0x1D508B06AF7C89E2ULL, "hlac_ca_caverns2_5802_54x.dds"},
	{0x1D50E6CD2ABF816AULL, "144285428.ogg"},
	{0x1D5101E48550A4C0ULL, "follower_banter6160611.cl"},
	{0x1D51067DC5072451ULL, "vxw_blightlands_125_0_0.vlm"},
	{0x1D510D7830E223BDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wmda_0d.dds"},
	{0x1D51E942720C58E0ULL, "vxlt_blackemporium_ts_0_0.vlm"},
	{0x1D51F4E3943DE039ULL, "vxlt_0_lowtown_docks_b6q_0_0.vlm"},
	{0x1D5207437DEE7EF4ULL, "vxw_blightlands_va_0_0.vlm"},
	{0x1D525A1DE68487F9ULL, "vxlt_0_lowtown_a3_n_c_gj5_0_0.vlm"},
	{0x1D52F5F2AEAEECE4ULL, "gen00fl_merrill6128751.cl"},
	{0x1D532763A2F9D934ULL, "867097513.ogg"},
	{0x1D5368CC87BA5C1DULL, "598232390.ogg"},
	{0x1D539734779A75ABULL, "hlac_lt_hangedman_-1_15x.dds"},
	{0x1D53B52F6E406125ULL, "mag321cr_servant_gossip_1b.utc"},
	{0x1D53EEED24730065ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chaircutup.phy"},
	{0x1D5401F0A6899458ULL, "6140133_m.fxe"},
	{0x1D54115FFA586808ULL, "6205980_m.fxe"},
	{0x1D54371821095DF8ULL, "vxlt_undercityserialkiller_18a_0_0.vlm"},
	{0x1D5470A4CEC616CCULL, "lgt120_orz_bln_official6140183.cl"},
	{0x1D54B824F13E4E62ULL, "vxca_caverns1_ihz_0_0.vlm"},
	{0x1D54BFF9D7D6CED6ULL, "vxlt_0_lowtown_a2_n_hbv_0_0.vlm"},
	{0x1D55282F47425D03ULL, "vxca_caverns2_70w_0_1.vlm"},
	{0x1D55AB405CB7598DULL, "llac_lt_gallowstemplar_7198_0.dds"},
	{0x1D55ABC599C0FCAEULL, "zz_vault_debug6178672.cl"},
	{0x1D55DCB201F27119ULL, "276350794.ogg"},
	{0x1D566C304557BB53ULL, "81075383.ogg"},
	{0x1D5723638100CA74ULL, "mrl150_merrill6085804.cl"},
	{0x1D573F4BB5E44AA4ULL, "and100_anders6095925.cl"},
	{0x1D57750C561882A0ULL, "vxlt_0_lowtown_docks_bqq_0_0.vlm"},
	{0x1D5782A7448B4AE8ULL, "279042448.ogg"},
	{0x1D57A0221EF34A19ULL, "vxw_bonepitdestroyed_a3_252_0_0.vlm"},
	{0x1D57B2863855DE5BULL, "vxht_keep_4bp_0_0.vlm"},
	{0x1D57DCDA13C26077ULL, "vxlt_gallowstemplar_n_46h_0_0.vlm"},
	{0x1D57E10EB9396789ULL, "vxht_0_hightown_bdz_0_0.vlm"},
	{0x1D57EA6465953409ULL, "vxca_caverns2_8kv_0_0.vlm"},
	{0x1D5856D69A2EA6F6ULL, "vxlt_gallowstemplar_34g_0_0.vlm"},
	{0x1D5878F93C29BB21ULL, "6110808_f.fxe"},
	{0x1D588714E9A8563FULL, "mag101_grace6034192.cl"},
	{0x1D58BB15F3AC59F4ULL, "1000659405.ogg"},
	{0x1D58D5EA53BED2AEULL, "hlac_ht_0_hightown_a2_-1_1z.dds"},
	{0x1D590B045C151A98ULL, "453611350.ogg"},
	{0x1D5914139B7078EDULL, "781596191.ogg"},
	{0x1D59250B228AECD2ULL, "hlac_dr_slavers_3217_11y.dds"},
	{0x1D5958E41F1E996DULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robb.mmh"},
	{0x1D595CE41F1E9FBEULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robb.mao"},
	{0x1D596AE41F1EB7F3ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robb.msh"},
	{0x1D5998391632CEEDULL, "gen00fl_merrill6129553.cl"},
	{0x1D59DCFA06DB6A67ULL, "zz_dae_debug6184890.cl"},
	{0x1D59F9C16DE0C30DULL, "w_woundedcoast_80_0.msh"},
	{0x1D5A07C16DE0DB40ULL, "w_woundedcoast_80_0.mao"},
	{0x1D5A0BC16DE0E193ULL, "w_woundedcoast_80_0.mmh"},
	{0x1D5A7023676E6FDEULL, "hlac_lt_0_lowtown_-196608_4x.dds"},
	{0x1D5AE75277F446A2ULL, "dae000cr_grand_cleric6147672.cl"},
	{0x1D5AEC541991843CULL, "346179154.ogg"},
	{0x1D5AF9255A9FFCCAULL, "vxlt_hangedman_2mp_0_0.vlm"},
	{0x1D5B3EE10CD9C08BULL, "hlac_lt_0_lowtown_docks_n_-3_8x.dds"},
	{0x1D5B66D2088BAAC6ULL, "ht_bartrand_n_3374.anb"},
	{0x1D5B8CCDBC144401ULL, "6033668_f.fxe"},
	{0x1D5C32E0F8257BDBULL, "6163681_m.fxe"},
	{0x1D5C5F910AB2F850ULL, "hlac_lt_0_lowtown_a3_n_c_65535_5z.dds"},
	{0x1D5CA6144E7B17E0ULL, "974635334.ogg"},
	{0x1D5CB5ACEFFD92F7ULL, "vxlt_0_lowtown_a2_e8c_0_0.vlm"},
	{0x1D5E9011AF0EE4EDULL, "vxlt_warehouse_i1_0_0.vlm"},
	{0x1D5E9812ABDB314AULL, "6117921_f.fxe"},
	{0x1D5F2689FA56B499ULL, "vxlt_hangedman_1ri_0_0.vlm"},
	{0x1D5F2B107751A264ULL, "90912009.ogg"},
	{0x1D5F2DAD62F093F7ULL, "vxca_caverns1_in5_0_0.vlm"},
	{0x1D5F66EF843AD05FULL, "vxw_bonepitdestroyed_a3_17u_0_0.vlm"},
	{0x1D5FA5F489EC51C7ULL, "dae300_vashothlead.bnk"},
	{0x1D5FB72DAE42AE3DULL, "mag311_nyssa6171519.cl"},
	{0x1D60261A1B9864BAULL, "vxca_caverns2prc_9l8_0_0.vlm"},
	{0x1D608C04BE5A8492ULL, "hlac_w_encounter2_-65536_0x.dds"},
	{0x1D60AC5E5D7C1BE4ULL, "338995991.ogg"},
	{0x1D60DB7EC26C83BCULL, "171368231.ogg"},
	{0x1D614110E895EB20ULL, "vxca_caverns1_hkx_0_0.vlm"},
	{0x1D61B5B3B42299C6ULL, "hlac_dr_0_deeproads_5838_1y.dds"},
	{0x1D62245F4837255DULL, "hm_amb_shopper.mor"},
	{0x1D625B614DCA41E6ULL, "vxca_caverns2prc_9fv_0_0.vlm"},
	{0x1D62BAADED9B50ECULL, "vxlt_gallowstemplar_f_2pc_0_0.vlm"},
	{0x1D62C7F489EEC13FULL, "dae300_vashothlead.cnv"},
	{0x1D62F584CB3AC3B5ULL, "6042065_m.fxe"},
	{0x1D6348CA3FCBDC9AULL, "6099163_m.fxe"},
	{0x1D63F5F723EA127AULL, "follower_banter6141506.cl"},
	{0x1D6454D14196D100ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubenw.phy"},
	{0x1D6479614FDF71A1ULL, "6016569_m.fxe"},
	{0x1D6527C585787140ULL, "vxlt_darktown_6gq_0_0.vlm"},
	{0x1D658C4A77A088F6ULL, "gen00fl_merrill6150691.cl"},
	{0x1D65943741F2B9B3ULL, "tal_tau_02.mao"},
	{0x1D65D6DD0B9ABF90ULL, "749852100.ogg"},
	{0x1D65F6C5F517CCD9ULL, "6078745_m.fxe"},
	{0x1D662E66CBE3F2E7ULL, "l_dae130ar_undercity_dt.bnk"},
	{0x1D66BB74628C7B30ULL, "hlac_ca_cavern1_haunted_24425_9x.dds"},
	{0x1D6729EF9F483862ULL, "6062688_m.fxe"},
	{0x1D677EDB7AFAFF8FULL, "6201833_m.fxe"},
	{0x1D67913699051C25ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dsla_0n.dds"},
	{0x1D67C7B8404D9040ULL, "vxca_cavern1_haunted_gex_0_0.vlm"},
	{0x1D67D49309BA4F89ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock2a.phy"},
	{0x1D67FABBFC12992DULL, "207017141.ogg"},
	{0x1D682ECE23CE8BF3ULL, "vxlt_gamlenshouse_n_2e_0_0.vlm"},
	{0x1D686E211FD330C2ULL, "zz_fen_debug6134361.cl"},
	{0x1D689F4A5FB35220ULL, "vxlt_gallowscourtyard_7fu_0_0.vlm"},
	{0x1D68A7A7C6457E33ULL, "6036526_f.fxe"},
	{0x1D68B676CFC5D10BULL, "vxlt_0_lowtown_exs_0_0.vlm"},
	{0x1D68ED9C1B261DC3ULL, "6044519_m.fxe"},
	{0x1D69776945B7A18DULL, "vxlt_hangedman_n_21l_0_0.vlm"},
	{0x1D69E00707B43DFCULL, "6139814_m.fxe"},
	{0x1D6A350BBF9B7017ULL, "zz_followers_banter6159627.cl"},
	{0x1D6A51E41F2D002AULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robb.phy"},
	{0x1D6AB4B5A5EF257DULL, "vxlt_foundry_n_vk_0_0.vlm"},
	{0x1D6AEEED1C3DC289ULL, "hlac_ca_caverns2prc_5802_33y.dds"},
	{0x1D6AF02202BC8402ULL, "dae000_junar6147850.cl"},
	{0x1D6AF89AE26AC4D3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v1.ani"},
	{0x1D6B2152030397D2ULL, "vxw_woundedcoast_gz_0_0.vlm"},
	{0x1D6B7871A717656AULL, "vxht_0_hightown_b8i_0_0.vlm"},
	{0x1D6B7F542F0A03B0ULL, "809649138.ogg"},
	{0x1D6BB97534DCCC52ULL, "6134877_m.fxe"},
	{0x1D6BD5C02E0ED9E6ULL, "w_weapon_foley.bnk"},
	{0x1D6BE445BCCF153BULL, "6093645_m.fxe"},
	{0x1D6D01EEA104FD6DULL, "vxlt_warehouse_r0_0_0.vlm"},
	{0x1D6D46CEDB0FDEB0ULL, "w_bonepitdestroyed_a3_84_0.msh"},
	{0x1D6D54CEDB0FF6CAULL, "w_bonepitdestroyed_a3_84_0.mmh"},
	{0x1D6D58CEDB0FFD01ULL, "w_bonepitdestroyed_a3_84_0.mao"},
	{0x1D6D9866EC4CCD19ULL, "vxlt_hangedman_1yh_0_0.vlm"},
	{0x1D6DD8CC7630E8DFULL, "vxw_bonepitdestroyed_a3_18t_0_0.vlm"},
	{0x1D6DE50391F5AA40ULL, "mrl150_dalish_guard6068270.cl"},
	{0x1D6E09BF89892886ULL, "chanters.gfx"},
	{0x1D6E1C0D180A42E2ULL, "469456217.ogg"},
	{0x1D6F16E5C5A54758ULL, "gen00fl_anders6112895.cl"},
	{0x1D6F60F91583F683ULL, "gen00fl_anders6144062.cl"},
	{0x1D6F74A9A49B2CDCULL, "vxht_keep_n_680_0_0.vlm"},
	{0x1D70B8D5BEB79028ULL, "vxht_brothel_n_6bm_0_0.vlm"},
	{0x1D71D30A18439F22ULL, "252578298.ogg"},
	{0x1D71EA4354FF3D62ULL, "mer110_jansen6032933.cl"},
	{0x1D7213663B2D5D9AULL, "hlac_lt_darktown_-1_44z.dds"},
	{0x1D725C3B01CA46EFULL, "vxw_bonepitmining_a2_25j_0_0.vlm"},
	{0x1D7286A139675FD1ULL, "genim_arm_hlm_inq.uti"},
	{0x1D72D066C6F508C0ULL, "vxca_caverns2prc_91t_0_1.vlm"},
	{0x1D73059A8CF7902DULL, "vxlt_0_lowtown_a2_n_gdh_0_0.vlm"},
	{0x1D730B9FC83AD4D9ULL, "royalarcher_combobreaker.dds"},
	{0x1D732D31F5D77661ULL, "6142032_f.fxe"},
	{0x1D734547E76561C5ULL, "vxca_caverns1_mines_cmm_0_0.vlm"},
	{0x1D734A9A5E09B408ULL, "lt_0_lowtown_docks_194_0.mmh"},
	{0x1D73569A5E09C8ABULL, "lt_0_lowtown_docks_194_0.mao"},
	{0x1D73689A5E09E732ULL, "lt_0_lowtown_docks_194_0.msh"},
	{0x1D73732637FB9759ULL, "vxht_mansion2_2e5_0_0.vlm"},
	{0x1D739C1632778019ULL, "585079661.ogg"},
	{0x1D73C37B3F5EBDC7ULL, "gen_im_trs_krk_hgh_junk1_03.uti"},
	{0x1D73E2DC714686F7ULL, "vxht_0_hightown_n_bze_0_0.vlm"},
	{0x1D7478E5FFD80275ULL, "6189917_m.fxe"},
	{0x1D74C4C89D259889ULL, "vxht_0_hightown_bjx_0_0.vlm"},
	{0x1D750A15341CB93CULL, "6026933_f.fxe"},
	{0x1D758A24AC3C1078ULL, "6145224_m.fxe"},
	{0x1D758DC4F2843698ULL, "6104924_m.fxe"},
	{0x1D7643D66B28F621ULL, "6049837_m.fxe"},
	{0x1D76884752EA0A03ULL, "vxht_0_hightown_a2_4c7_0_0.vlm"},
	{0x1D773F2B9DE2B652ULL, "llac_ht_bartrand_n_3365_0.dds"},
	{0x1D7750104AF79513ULL, "zz_isa_debug6133992.cl"},
	{0x1D77F09AE275410AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v1.evt"},
	{0x1D78002C8219F239ULL, "zz_lgt_2_debug6211859.cl"},
	{0x1D788B76E78E90EEULL, "lt_0_lowtown_docks_a3_25_0_lc.mmh"},
	{0x1D78D3DFB54A8DDBULL, "mag131_taraneh6095692.cl"},
	{0x1D79CF26359D5B34ULL, "469165016.ogg"},
	{0x1D7A11C4F627000DULL, "526046235.ogg"},
	{0x1D7A1E7B2BC9B3FAULL, "vxlt_0_lowtown_a3_n_c_db5_0_0.vlm"},
	{0x1D7A1EDB7884B0AAULL, "gen_im_trs_gen_sta_spi_r2.uti"},
	{0x1D7A2558BA288F57ULL, "6100141_m.fxe"},
	{0x1D7BF88549BF7E4AULL, "518268127.ogg"},
	{0x1D7C7A38A791F80EULL, "185549886.ogg"},
	{0x1D7CC151C953B7A4ULL, "dae000_oldmangossip6174766.cl"},
	{0x1D7CD4CCABE228C2ULL, "dae000_arishok6067803.cl"},
	{0x1D7DE6869691455CULL, "vxht_keep_n_631_0_0.vlm"},
	{0x1D7E24CACC821C20ULL, "vxca_caverns1_huf_0_0.vlm"},
	{0x1D7E2BA3E8E68282ULL, "vxht_keep_n_5j3_0_0.vlm"},
	{0x1D7E8456259B2A53ULL, "vxw_woundedcoast_ya_0_0.vlm"},
	{0x1D7EC3DF6FBE2283ULL, "174281455.ogg"},
	{0x1D7ECF9AE27B287DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v1.gad"},
	{0x1D7F24F8FC2673E6ULL, "clf260cr_guard3.utc"},
	{0x1D7F548AF28D4C72ULL, "334267768.ogg"},
	{0x1D7F8A75B9B6CC1DULL, "hlac_ht_mansion2b_818_6y.dds"},
	{0x1D7FFE021AF044D3ULL, "6208647_m.fxe"},
	{0x1D803F78427E9A58ULL, "hlac_lt_undercitydungeon_-1_45z.dds"},
	{0x1D80479B13D8E37CULL, "vxlt_gallowsprison_n_64c_0_0.vlm"},
	{0x1D806B325D142DEBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityexit.evt"},
	{0x1D8077BD8760FB6AULL, "72471278.ogg"},
	{0x1D80AD6A56A68646ULL, "35123817.ogg"},
	{0x1D8141634EA7FEA0ULL, "smoketempadd.mao"},
	{0x1D81B51AD86CEC4BULL, "vxlt_0_lowtown_docks_b84_0_1.vlm"},
	{0x1D822A2800E5F31AULL, "24399489.ogg"},
	{0x1D822DAA5A1C6EA8ULL, "6170673_m.fxe"},
	{0x1D827D4C06F281EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\longrangeattack.evt"},
	{0x1D828EA99D710F85ULL, "vxlt_undercityserialkiller_18x_0_0.vlm"},
	{0x1D82D2DF15B92549ULL, "122123332.ogg"},
	{0x1D83000C05E956FAULL, "6045795_m.fxe"},
	{0x1D8369761B490822ULL, "75689063.ogg"},
	{0x1D847679768B370DULL, "6164844_m.fxe"},
	{0x1D84B3EECE22E21DULL, "darktown_male_fer_urch_sta6174957.cl"},
	{0x1D85D586431BB902ULL, "vxlt_0_lowtown_docks_c09_0_0.vlm"},
	{0x1D8684325D18C3ACULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityexit.gad"},
	{0x1D86D0690301E3E1ULL, "vxca_caverns1_mines_ck7_0_0.vlm"},
	{0x1D86FA7748DDB754ULL, "vxlt_0_lowtown_a2_n_c_ffx_0_0.vlm"},
	{0x1D870D4FAF402CD9ULL, "6161375_m.fxe"},
	{0x1D873F254118F77DULL, "lt_0_lowtown_a3_n_c_38_0.rml"},
	{0x1D8767FF3CA1533BULL, "vxdr_primevalentrance_3br_0_0.vlm"},
	{0x1D876BF53E3DFEE2ULL, "vxlt_gallowscourtyard_66r_0_0.vlm"},
	{0x1D878D7A596A8388ULL, "w_bonepitmining_a2_26_0_lc.mmh"},
	{0x1D87F86397D97387ULL, "vxdr_primevaltomb_7tl_0_0.vlm"},
	{0x1D88050BE6EFC8D2ULL, "vxw_woundedcoast_yd_0_0.vlm"},
	{0x1D881EA974A4BF5AULL, "gen00fl_aveline6128411.cl"},
	{0x1D886A983A531D63ULL, "6162808_m.fxe"},
	{0x1D88744C06F6DD99ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\longrangeattack.gad"},
	{0x1D88A88DCDBE7BBEULL, "vxht_mansion2_ad_0_0.vlm"},
	{0x1D88AE169A300E23ULL, "hlac_lt_0_lowtown_docks_n_-65538_1y.dds"},
	{0x1D88C1D6DEA6BEC4ULL, "6202358_m.fxe"},
	{0x1D8925072B5E316CULL, "360714636.ogg"},
	{0x1D8954BA92AB07D9ULL, "lgt310ip_body_2.ncs"},
	{0x1D8971D141B61E95ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubenw.msh"},
	{0x1D8973D141B6220BULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubenw.mmh"},
	{0x1D897FD141B636A8ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubenw.mao"},
	{0x1D89D654B1C1F14FULL, "one280_merin6116757.cl"},
	{0x1D8A564498F7F210ULL, "vxht_0_hightown_9aw_0_0.vlm"},
	{0x1D8A707E4501886FULL, "hlac_lt_gallowstemplar_n_2394_11x.dds"},
	{0x1D8ADFD873CBAC3CULL, "l_dae320ar_hightown.wbd"},
	{0x1D8AF07935F509FDULL, "vxlt_gallowstemplar_f_5xa_0_0.vlm"},
	{0x1D8B12F8AF352E7DULL, "6092082_m.fxe"},
	{0x1D8B2237BB91CE52ULL, "vxw_bonepit_1qr_0_0.vlm"},
	{0x1D8B40BED7BCEB51ULL, "vxlt_darktown_2sd_0_0.vlm"},
	{0x1D8B9742AC757790ULL, "279769969.ogg"},
	{0x1D8BED56C29B04ABULL, "6181825_m.fxe"},
	{0x1D8C53933E9A67CAULL, "vxht_brothel_n_757_0_0.vlm"},
	{0x1D8D1CAD5EA47229ULL, "13274824.ogg"},
	{0x1D8D28AA6B2FDFD8ULL, "and100_anders6205531.cl"},
	{0x1D8D3E38970190EEULL, "vxlt_0_lowtown_a3_n_c_fab_0_0.vlm"},
	{0x1D8D83325D1EE0AAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityexit.ani"},
	{0x1D8DB1E8CB2CADD6ULL, "vxht_keep_55m_0_0.vlm"},
	{0x1D8E4AFE499163B4ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_lrg.dura.ani"},
	{0x1D8E59678A8FE016ULL, "6140997_m.fxe"},
	{0x1D8EA9A9C37DFEB9ULL, "gen00fl_anders6121186.cl"},
	{0x1D8EFDFF30FCE836ULL, "hlac_lt_0_lowtown_docks_n_-65539_2z.dds"},
	{0x1D8F0F58D731A7ACULL, "w_encounter3_212_0.msh"},
	{0x1D8F1558D731B19EULL, "w_encounter3_212_0.mmh"},
	{0x1D8F2158D731C635ULL, "w_encounter3_212_0.mao"},
	{0x1D8F854C06FD19E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\longrangeattack.ani"},
	{0x1D8F96D27F6EAD82ULL, "qcr000_arishok6057109.cl"},
	{0x1D90135EEC334853ULL, "vxw_bonepit_1zv_0_0.vlm"},
	{0x1D90442F57D98E1EULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean_1.mao"},
	{0x1D90591EC4568FC8ULL, "vxca_caverns1_mines_ck2_0_0.vlm"},
	{0x1D9107D46E1DFCB0ULL, "gen00fl_sebastian6174995.cl"},
	{0x1D914F627B9E2333ULL, "sib100_father_contact6108498.cl"},
	{0x1D91882696087630ULL, "vxca_cave_1fs_0_1.vlm"},
	{0x1D922466CC091C1DULL, "l_dae130ar_undercity_dt.wbd"},
	{0x1D92A3B3B26D6671ULL, "follower_banter6155616.cl"},
	{0x1D92FF43A5EEBFFCULL, "163193911.ogg"},
	{0x1D93098479386585ULL, "6108306_f.fxe"},
	{0x1D935F869A9F328DULL, "vxht_bartrand_n_4kq_0_0.vlm"},
	{0x1D93BD9309DF63AEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock2a.mmh"},
	{0x1D93C99309DF7845ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock2a.mao"},
	{0x1D93D79309DF8F9CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock2a.msh"},
	{0x1D93E7146BD84FA0ULL, "640529170.ogg"},
	{0x1D940C0DA5A81A8BULL, "vxlt_0_lowtown_eit_0_0.vlm"},
	{0x1D942093281B5F2BULL, "hlac_lt_gallowstemplar_f_2384_5x.dds"},
	{0x1D9486DA808B3E3BULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm06.phy"},
	{0x1D9523EF16B046FAULL, "strmt01_n.dds"},
	{0x1D9547FE49977D16ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_lrg.dura.gad"},
	{0x1D9561D5DF1BF675ULL, "vxw_blightlands_b8_0_0.vlm"},
	{0x1D9571BAC3F3B180ULL, "hlac_lt_0_lowtown_docks_a3_196608_1x.dds"},
	{0x1D957EEA324BF405ULL, "plt_cod_bks_lite_tombs_2.nss"},
	{0x1D957F2CA1364559ULL, "6166275_m.fxe"},
	{0x1D9640B150C201FBULL, "vxlt_0_lowtown_a2_n_gnp_0_0.vlm"},
	{0x1D966A4089CF8C07ULL, "vxdr_primevaltomb_7qo_0_0.vlm"},
	{0x1D9676E8D8E5E152ULL, "vxw_woundedcoast_ve_0_0.vlm"},
	{0x1D96BE1E620A5D83ULL, "6083688_m.fxe"},
	{0x1D96E006AA4D0D8BULL, "vxca_cavern1_haunted_eh6_0_0.vlm"},
	{0x1D976F4FBA13C32AULL, "zz_prmpt_debug6200961.cl"},
	{0x1D98246EC352E259ULL, "6111170_m.fxe"},
	{0x1D985E47EBF6FEC4ULL, "vxht_mansion_n_5ba_0_0.vlm"},
	{0x1D98AA8ECF3FCD81ULL, "441615479.ogg"},
	{0x1D99110A83937C44ULL, "683943803.ogg"},
	{0x1D992B8AA4F08904ULL, "vxlt_0_lowtown_gj0_0_0.vlm"},
	{0x1D9959E079AF2AD5ULL, "6175099_m.fxe"},
	{0x1D999414AD87E6D2ULL, "vxw_bonepit_1zs_0_0.vlm"},
	{0x1D99CC51BBF78015ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_chestb_n.dds"},
	{0x1D99FD7F6795A33CULL, "vxw_bonepitdestroyed_a3_25k_0_0.vlm"},
	{0x1D9A10301EA3BFFCULL, "vxht_0_hightown_a3_bl6_0_0.vlm"},
	{0x1D9A247C578AF442ULL, "dae000_quintus6154943.cl"},
	{0x1D9AE6C877C47CD7ULL, "6152209_m.fxe"},
	{0x1D9B0D3A9010EB20ULL, "hla_ht_0_hightown_a2_-196608_0y.dds"},
	{0x1D9B1E0A8580327FULL, "407491055.ogg"},
	{0x1D9B681009875B53ULL, "vxw_woundedcoast_sc_0_0.vlm"},
	{0x1D9B78931D78B678ULL, "6093619_m.fxe"},
	{0x1D9C0E6C9499D9A8ULL, "vxht_brothel_n_6uh_0_0.vlm"},
	{0x1D9C448EA7E90D85ULL, "dae000ip_crowd_em1_talk1.utp"},
	{0x1D9C62FE499DCA75ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_lrg.dura.evt"},
	{0x1D9C6E70790CFF6FULL, "dae200cr_viscount6087386.cl"},
	{0x1D9D2D745E11B2BFULL, "6130080_f.fxe"},
	{0x1D9DCD7E8C9D711AULL, "vxlt_hangedman_16z_0_0.vlm"},
	{0x1D9E1298FF03E67EULL, "6216627_m.fxe"},
	{0x1D9E1962B29E2DF4ULL, "220040440.ogg"},
	{0x1D9E8DED82C1D4EEULL, "qun230cr_carta_thug.utc"},
	{0x1D9F1A379FF5A1B6ULL, "vxca_cavern1_haunted_j5o_0_0.vlm"},
	{0x1D9F2AD2AAD6B560ULL, "6137463_m.fxe"},
	{0x1D9F51E7A6A78543ULL, "gen00fl_merrill6145214.cl"},
	{0x1D9F7263815D4085ULL, "vxlt_undercityserialkiller_12f_0_0.vlm"},
	{0x1D9FA3DEB20C69A7ULL, "63327262.ogg"},
	{0x1D9FFA0387FE8EB0ULL, "vxca_cave_1ar_0_1.vlm"},
	{0x1D9FFDEA7F5E314DULL, "ave100_ambush_16106194.cl"},
	{0x1DA00D098024803BULL, "vxw_encounter3_9c_0_0.vlm"},
	{0x1DA0A6724376003DULL, "6137692_f.fxe"},
	{0x1DA0C39D03E5C28FULL, "6071191_m.fxe"},
	{0x1DA11F9961B50D25ULL, "429975849.ogg"},
	{0x1DA13687319D082AULL, "vxht_bartrand_n_66c_0_0.vlm"},
	{0x1DA16F991CDC3347ULL, "598436390.ogg"},
	{0x1DA174C041420067ULL, "6125341_f.fxe"},
	{0x1DA1753EC2084B5EULL, "gen00fl_merrill6108434.cl"},
	{0x1DA183E742F90DBCULL, "vxlt_hangedman_1r0_0_0.vlm"},
	{0x1DA184522299E263ULL, "hlac_dr_primevaltomb_659_10z.dds"},
	{0x1DA1D228F2FDD1A2ULL, "638513936.ogg"},
	{0x1DA2117EE5BAEBA2ULL, "64152855.ogg"},
	{0x1DA28606C35347CEULL, "6088764_f.fxe"},
	{0x1DA2995607B08A4BULL, "766738811.ogg"},
	{0x1DA2C79353872158ULL, "607957415.ogg"},
	{0x1DA3152AF4F15AE1ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_mag_0s.dds"},
	{0x1DA3B4A36A2BE47EULL, "gen00fl_fenris6128054.cl"},
	{0x1DA3D4E3EA1F636CULL, "mag371_elsa6064668.cl"},
	{0x1DA3EB0036CA7E46ULL, "2022808.ogg"},
	{0x1DA418647746912FULL, "6097843_m.fxe"},
	{0x1DA4295B7C6504D1ULL, "153154598.ogg"},
	{0x1DA42B1AEE8F7303ULL, "lt_0_lowtown_docks_158_0.mao"},
	{0x1DA42F1AEE8F7A50ULL, "lt_0_lowtown_docks_158_0.mmh"},
	{0x1DA432563BAF4D84ULL, "gen_im_arm_cht_elf_comc.uti"},
	{0x1DA437E7A22F412FULL, "vxht_keep_n_5dx_0_0.vlm"},
	{0x1DA43D1AEE8F926AULL, "lt_0_lowtown_docks_158_0.msh"},
	{0x1DA4E1CB241DB8D4ULL, "vxw_woundedcoast_3p_0_0.vlm"},
	{0x1DA4E8C5CADBF9D2ULL, "vxw_woundedcoast_sf_0_0.vlm"},
	{0x1DA4ED04F5613A9BULL, "6125576_f.fxe"},
	{0x1DA5016E487F9064ULL, "641390821.ogg"},
	{0x1DA52078DA6EFF5DULL, "llac_lt_warehouse_n_1_6.dds"},
	{0x1DA52FFB54007603ULL, "6086988_m.fxe"},
	{0x1DA60223914CB17BULL, "441449104.ogg"},
	{0x1DA6330BA7B1CDEBULL, "6126880_m.fxe"},
	{0x1DA65EC79E22C53CULL, "225768907.ogg"},
	{0x1DA67657A7A7DF20ULL, "vxht_keep_n_4mq_0_0.vlm"},
	{0x1DA6C3A3099B4DD8ULL, "498685958.ogg"},
	{0x1DA6CAC3F8387614ULL, "art\\cinematics\\deathblows\\meatgolem\\cs_gibhed_meatwarr.gad"},
	{0x1DA86208251FF871ULL, "lgt160_ntb_elf_conflict6163391.cl"},
	{0x1DA86DFC9DC3F43BULL, "6119947_m.fxe"},
	{0x1DA89A9333AC0B3CULL, "6139762_m.fxe"},
	{0x1DA8A46905C32DF6ULL, "6199195_m.fxe"},
	{0x1DA8AFE205D1A6DEULL, "mag371_elsa6041234.cl"},
	{0x1DA8C4E47063092AULL, "hlac_lt_0_lowtown_-3_7z.dds"},
	{0x1DA99CD7358B7C65ULL, "mcr371_hostile_fenris6082247.cl"},
	{0x1DA9AFA4CFCE63B7ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_wf.gad"},
	{0x1DA9C6988AC8A42AULL, "38465682.ogg"},
	{0x1DAA830218C8B9D4ULL, "vxw_bonepit_1ws_0_1.vlm"},
	{0x1DAB0E9F1267B9E9ULL, "vxht_0_hightown_9uv_0_0.vlm"},
	{0x1DAB4984BC4F5C4CULL, "isa211_isabela6202151.cl"},
	{0x1DAC53A88D46B667ULL, "dm__merin.mor"},
	{0x1DAC789DB4D4EE1EULL, "vxlt_hangedman_n_301_0_0.vlm"},
	{0x1DAC851834FD18A0ULL, "vxw_bonepitdestroyed_a3_1wv_0_0.vlm"},
	{0x1DAC94017C2ADE25ULL, "vxht_keep_n_69z_0_0.vlm"},
	{0x1DACAE499AD65D66ULL, "72113075.ogg"},
	{0x1DACD7C02E46738CULL, "w_weapon_foley.wbd"},
	{0x1DACE6A5F1A4696BULL, "701561487.ogg"},
	{0x1DACFAB3151E1FD6ULL, "mer120_hubert6070508.cl"},
	{0x1DAD4B0B0F0C072BULL, "6017865_m.fxe"},
	{0x1DAD8AEF3ED1652AULL, "8008411.ogg"},
	{0x1DADD9C3F83EAEF2ULL, "art\\cinematics\\deathblows\\meatgolem\\cs_gibhed_meatwarr.ani"},
	{0x1DADE44073535905ULL, "vxlt_undercityserialkiller_17e_0_0.vlm"},
	{0x1DAE7EE6721A98BBULL, "vxw_encounter3_2b_0_0.vlm"},
	{0x1DAE8F52F6381718ULL, "6040546_f.fxe"},
	{0x1DAF7459DB9C2048ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_fereldencoin_n.dds"},
	{0x1DAFAB72551471D0ULL, "vxlt_warehouse_n_ay_0_0.vlm"},
	{0x1DAFCFC10D44E556ULL, "57408743.ogg"},
	{0x1DAFE69D333818E7ULL, "6122241_f.fxe"},
	{0x1DAFFF69817BF718ULL, "6005818_m.fxe"},
	{0x1DB0A7791BB3F65AULL, "qun220cr_javaris6071518.cl"},
	{0x1DB162352E878F00ULL, "vxca_caverns1_ckg_0_0.vlm"},
	{0x1DB2A8F9A93CC13FULL, "mcr371_magesurrender_st.stg"},
	{0x1DB2DF25F4737FCAULL, "hlac_lt_0_lowtown_n_-65536_8x.dds"},
	{0x1DB368AB5FB3955EULL, "mrl210_merrill6102100.cl"},
	{0x1DB54201CFE32FC4ULL, "vxht_mansion_n_5hg_0_0.vlm"},
	{0x1DB57315818F930EULL, "344467037.ogg"},
	{0x1DB5BA0AD5B3A85EULL, "vxlt_0_lowtown_a2_n_e1n_0_0.vlm"},
	{0x1DB6BEECFF92F41AULL, "vxca_caverns1_icq_0_2.vlm"},
	{0x1DB6DDD873F0E3B6ULL, "l_dae320ar_hightown.bnk"},
	{0x1DB6F5871577E204ULL, "follower_banter6136700.cl"},
	{0x1DB7AF5232658873ULL, "735370363.ogg"},
	{0x1DB876C9BB608282ULL, "6034355_m.fxe"},
	{0x1DB88470F6005CE3ULL, "6204971_m.fxe"},
	{0x1DB8B91A96CCDBFFULL, "62228522.ogg"},
	{0x1DB8C1A070C06E19ULL, "6087264_m.fxe"},
	{0x1DB99764FDF6A885ULL, "dae000_pickpocket_scene6175092.cl"},
	{0x1DB9AAACA4B5F24FULL, "vxlt_0_lowtown_a3_n_c_hnt_0_0.vlm"},
	{0x1DBA337DC70F0BA2ULL, "zz_approval_debug6003739.cl"},
	{0x1DBA3E2541444E95ULL, "lt_0_lowtown_a3_n_c_38_0.anb"},
	{0x1DBA67FC51EADE39ULL, "vxca_cave_18x_0_1.vlm"},
	{0x1DBABC67E9F03580ULL, "hlac_ht_chantry_n_2534_42x.dds"},
	{0x1DBB2967B47FF871ULL, "hlac_lt_0_lowtown_n_65538_3y.dds"},
	{0x1DBB443D65D12154ULL, "14445086.ogg"},
	{0x1DBB9291C45C53CCULL, "hlac_lt_0_lowtown_docks_n_65535_5z.dds"},
	{0x1DBB951CE6FD2621ULL, "vxw_bonepit_1rd_0_0.vlm"},
	{0x1DBBA5BD6C6E9EF4ULL, "6206376_m.fxe"},
	{0x1DBBEC6DEF4D7087ULL, "vxca_cave_240_0_0.vlm"},
	{0x1DBBF204222C4761ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffc_n.dds"},
	{0x1DBC11D1CB707E2CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_oldmonument.phy"},
	{0x1DBC6D5EB0FE0999ULL, "vxw_encounter2_camp_6g_0_0.vlm"},
	{0x1DBCB4639C1D3152ULL, "483343849.ogg"},
	{0x1DBCFA7FC647ECD9ULL, "6207947_m.fxe"},
	{0x1DBDDCA4CFDF593DULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_wf.ani"},
	{0x1DBDE9F31E36A695ULL, "vxht_brothel_n_361_0_0.vlm"},
	{0x1DBED389C2A138DBULL, "dae000_arishok6067761.cl"},
	{0x1DBF6BC2D20586BFULL, "961294525.ogg"},
	{0x1DBF7B71FE9B20CFULL, "6071916_f.fxe"},
	{0x1DC06DDA80B04EDAULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm06.msh"},
	{0x1DC07FDA80B06D20ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm06.mmh"},
	{0x1DC0BEE7F80DF11FULL, "362758637.ogg"},
	{0x1DC0D64AA3BE6FDAULL, "6096801_m.fxe"},
	{0x1DC12FD8FB45A0ADULL, "vxlt_0_lowtown_docks_n_c6k_0_0.vlm"},
	{0x1DC132A6C5AAEAEAULL, "mag100_cullen6140928.cl"},
	{0x1DC1A4652BF5461AULL, "626765978.ogg"},
	{0x1DC1C46D4EE8C3C4ULL, "fen210_tevinter_ambusher.cnv"},
	{0x1DC2E58379EADFE5ULL, "6031661_m.fxe"},
	{0x1DC3372B6507D304ULL, "6172770_m.fxe"},
	{0x1DC34EB4302EE376ULL, "gen_im_arm_boo_hvy_act2_02.uti"},
	{0x1DC424FBFDBF0F86ULL, "330940506.ogg"},
	{0x1DC44AB800AB3F77ULL, "vxca_caverns1_i32_0_0.vlm"},
	{0x1DC4CE64C65CD346ULL, "zz_qun_debug6045392.cl"},
	{0x1DC4FAC2C85F9524ULL, "vxlt_0_lowtown_a2_n_eoz_0_0.vlm"},
	{0x1DC4FAC8DF71FA86ULL, "vxlt_0_lowtown_docks_n_cwg_0_0.vlm"},
	{0x1DC55E0E981C65B8ULL, "426774877.ogg"},
	{0x1DC5E66D4EECE624ULL, "fen210_tevinter_ambusher.bnk"},
	{0x1DC62BFA8029C999ULL, "280130894.ogg"},
	{0x1DC733257A20A7C7ULL, "vxw_bonepitmining_a2_b7_0_0.vlm"},
	{0x1DC7421126BC7594ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottle_sa.dds"},
	{0x1DC79EDEC65E1C3DULL, "460166923.ogg"},
	{0x1DC8024EF457F47EULL, "vxlt_0_lowtown_docks_n_cow_0_0.vlm"},
	{0x1DC81C8996AC0ACFULL, "vxlt_0_lowtown_a3_n_c_hkw_0_0.vlm"},
	{0x1DC872120832D4E6ULL, "vxlt_0_lowtown_a3_n_c_eia_0_0.vlm"},
	{0x1DC8CA59653C7E6DULL, "vxca_caverns1_i0y_0_0.vlm"},
	{0x1DC8D77D067A2FDFULL, "vxlt_gallowscourtyard_7bz_0_0.vlm"},
	{0x1DC92B1267DEE183ULL, "902153414.ogg"},
	{0x1DC92C961D8C27C7ULL, "823464039.ogg"},
	{0x1DC951863AD837D4ULL, "vxca_caverns1_e6n_0_0.vlm"},
	{0x1DC9B054A4ED53EBULL, "llac_w_sundermount_-262140_0.dds"},
	{0x1DC9DAD2B40BAA53ULL, "vxw_bonepit_1nz_0_0.vlm"},
	{0x1DCA3FC4473F595EULL, "216539491.ogg"},
	{0x1DCA5ED454895E25ULL, "6051890_m.fxe"},
	{0x1DCAAA0117B2D8D6ULL, "757867671.ogg"},
	{0x1DCADF3BA2F42219ULL, "vxw_encounter2_camp_1f_0_0.vlm"},
	{0x1DCB1FD08F7BAB0CULL, "31091358.ogg"},
	{0x1DCBE81F738738B2ULL, "6043559_m.fxe"},
	{0x1DCC820009D853CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_ex.evt"},
	{0x1DCC899CD963D6C1ULL, "56672676.ogg"},
	{0x1DCCE1C1DE9D27A3ULL, "204542004.ogg"},
	{0x1DCD0C1A783A18C2ULL, "vxht_keep_n_736_0_0.vlm"},
	{0x1DCDBF9AC837DE1FULL, "follower_banter6161323.cl"},
	{0x1DCE885C488D9DA1ULL, "mag291_ntbanter6056828.cl"},
	{0x1DCED75D3ECAA929ULL, "one280_merin6116764.cl"},
	{0x1DCF2650DD97D25BULL, "6034277_m.fxe"},
	{0x1DCF78C57F12485FULL, "409712185.ogg"},
	{0x1DCF8442C0DEDFA2ULL, "914626154.ogg"},
	{0x1DD082F1D1D327D2ULL, "6115623_m.fxe"},
	{0x1DD0C492EC1C1B6BULL, "vxw_encounter2_7f_0_0.vlm"},
	{0x1DD0E620CD182061ULL, "332502611.ogg"},
	{0x1DD122461F77704CULL, "6034516_f.fxe"},
	{0x1DD1698C745EA3A8ULL, "llac_lt_0_lowtown_docks_n_-65538_1.dds"},
	{0x1DD1E38D0B74BA77ULL, "vxw_encounter3_9w_0_0.vlm"},
	{0x1DD3178770CA22FBULL, "326839318.ogg"},
	{0x1DD3246D05090D2BULL, "plt_seb120pt_duty.plo"},
	{0x1DD394AB56E37117ULL, "vxlt_gallowscourtyard_71k_0_0.vlm"},
	{0x1DD39B0009DE9C8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_ex.gad"},
	{0x1DD3C02C5F5B2310ULL, "hlac_ht_0_hightown_65535_3x.dds"},
	{0x1DD3DC40360D5DE6ULL, "380897401.ogg"},
	{0x1DD4B8AB4DB07DE4ULL, "6050099_f.fxe"},
	{0x1DD563895AE2DFAFULL, "6064636_m.fxe"},
	{0x1DD5A2237CDEB92BULL, "qun221_viscount6092395.cl"},
	{0x1DD5A5026C16C047ULL, "vxw_bonepitmining_a2_i6_0_0.vlm"},
	{0x1DD5FCEC99442E39ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chac_s.dds"},
	{0x1DD70516131E3D32ULL, "gen00fl_isabela6096795.cl"},
	{0x1DD74959F870485FULL, "vxlt_gallowscourtyard_7ge_0_0.vlm"},
	{0x1DD79A580FCC575DULL, "lt_0_lowtown_docks_a3_n_19_0_ll.mmh"},
	{0x1DD7DAAF0ADF6220ULL, "vxw_bonepitdestroyed_a3_1fu_0_0.vlm"},
	{0x1DD8375AEEC187DBULL, "zz_fol_debug6062216.cl"},
	{0x1DD860B6AB0EEEFCULL, "vxlt_undercityserialkiller_1c5_0_1.vlm"},
	{0x1DD8D0457CCC799BULL, "vxw_blightlands_i7_0_0.vlm"},
	{0x1DD8D2E0FB655A8EULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpa.phy"},
	{0x1DD8F3C675C8AFC5ULL, "one220_yevhen6115741.cl"},
	{0x1DD94456E87C647CULL, "lgt160pt_ntb_elves.ncs"},
	{0x1DD97AF35D7763A9ULL, "6036758_m.fxe"},
	{0x1DD99A0009E3068AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_ex.ani"},
	{0x1DD9A40E1156330CULL, "vxht_0_hightown_a2_7fx_0_0.vlm"},
	{0x1DD9C9BCF42000D3ULL, "517540315.ogg"},
	{0x1DD9EC65FCBF236BULL, "715334101.ogg"},
	{0x1DDA556F145BF336ULL, "6153135_m.fxe"},
	{0x1DDAE412F88E997DULL, "82771844.ogg"},
	{0x1DDB8725B95BBE1BULL, "vxlt_0_lowtown_a3_n_c_byg_0_0.vlm"},
	{0x1DDBE5AA098BF399ULL, "6078132_m.fxe"},
	{0x1DDC2C67E982B4A1ULL, "seb221ip_at_hightown.utp"},
	{0x1DDD409D3E4B939CULL, "6083056_m.fxe"},
	{0x1DDD5B6645145FF1ULL, "vxlt_0_lowtown_docks_b6i_0_0.vlm"},
	{0x1DDE6D8A58E69CB7ULL, "vxw_bonepitdestroyed_a3_8o_0_0.vlm"},
	{0x1DDEC0445E7BFED5ULL, "llac_lt_0_lowtown_a3_n_c_65536_12.dds"},
	{0x1DDF366FDE1233EBULL, "vxw_encounter2_2y_0_0.vlm"},
	{0x1DDFD6F555BD4F28ULL, "mag110_arianni6016829.cl"},
	{0x1DDFE510DB434F3CULL, "seb221st_harimann.stg"},
	{0x1DE05569FD6AD2F7ULL, "vxw_encounter3_2v_0_0.vlm"},
	{0x1DE061630B598C69ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ds001a_0n.dds"},
	{0x1DE1479C135D33EEULL, "6115476_m.fxe"},
	{0x1DE1DE7CAC4BC624ULL, "vxlt_0_lowtown_a2_n_eux_0_0.vlm"},
	{0x1DE21B2EE2002399ULL, "vxca_caverns1_j5e_0_0.vlm"},
	{0x1DE272B441AE9FE9ULL, "6041262_m.fxe"},
	{0x1DE3948B3C3FCEE8ULL, "hlac_ca_caverns1_mines_24423_2y.dds"},
	{0x1DE3B7E2D2B2B7CAULL, "dae300_petrice6097864.cl"},
	{0x1DE3BC755CE357B7ULL, "276696362.ogg"},
	{0x1DE482199E9A1FD0ULL, "gen_im_trs_dem_lgt_glv_r0.uti"},
	{0x1DE53FEB70EF98C7ULL, "lowtown_argument6169429.cl"},
	{0x1DE58B9CC29E5C36ULL, "w_woundedcoast_21_0.msh"},
	{0x1DE5959CC29E6D34ULL, "w_woundedcoast_21_0.mmh"},
	{0x1DE5999CC29E7447ULL, "w_woundedcoast_21_0.mao"},
	{0x1DE59E4D6F5C00D0ULL, "p_armoire_01d.dds"},
	{0x1DE645D047E4D7D9ULL, "vxlt_hangedman_5e_0_0.vlm"},
	{0x1DE64C8BFCD57AA0ULL, "vxw_bonepitdestroyed_a3_1kr_0_0.vlm"},
	{0x1DE67E2AA6CE7ACFULL, "293178597.ogg"},
	{0x1DE73A55DDB31E8FULL, "llac_lt_0_lowtown_a3_n_c_131070_6.dds"},
	{0x1DE742226EC2921BULL, "vxw_blightlands_r6_0_0.vlm"},
	{0x1DE79704A3E206F7ULL, "zz_fol_debug6066151.cl"},
	{0x1DE7A19AA160333AULL, "6092265_m.fxe"},
	{0x1DE7B624A1B1201CULL, "223066599.ogg"},
	{0x1DE7F5DAEEE18754ULL, "vxlt_gamlenshouse_n_n2_0_0.vlm"},
	{0x1DE82A1BD9D8C2EEULL, "972834656.ogg"},
	{0x1DE8396E23A8ACE9ULL, "vxht_chantry_n_23x_0_0.vlm"},
	{0x1DE858DB9DDBB543ULL, "vxdr_0_deeproads_cfr_0_2.vlm"},
	{0x1DE91E09688CC4A8ULL, "hlac_ht_0_hightown_a3_-131071_1y.dds"},
	{0x1DE92543AACEAB67ULL, "896165277.ogg"},
	{0x1DE92747C697F395ULL, "tv_door_01.dds"},
	{0x1DE95FF0CD7EC2A1ULL, "fen110_fenris6097042.cl"},
	{0x1DE978847D6BBA61ULL, "l_lgt110ar_gang_hideout_d.bnk"},
	{0x1DE9FA297FAD7939ULL, "vxw_sundermount_1zc_0_0.vlm"},
	{0x1DE9FEF3F802791DULL, "662956340.ogg"},
	{0x1DEA07148A17CC61ULL, "865447469.ogg"},
	{0x1DEB010F251B3BFAULL, "6208205_f.fxe"},
	{0x1DEB05733EDA91E7ULL, "vxdr_primevaltomb_53r_0_1.vlm"},
	{0x1DEBEAAE7035D7FEULL, "876590201.ogg"},
	{0x1DEBF298A05669E6ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_bdy_hlk_0s.dds"},
	{0x1DEC398460CB5600ULL, "vxw_encounter2_ce_0_0.vlm"},
	{0x1DEC4AB01F6F87C4ULL, "vxlt_hangedman_1me_0_0.vlm"},
	{0x1DEC9BAC30DA8A74ULL, "vxca_caverns1_fn1_0_0.vlm"},
	{0x1DECF8DD99756BABULL, "vxht_mansion2b_1z5_0_0.vlm"},
	{0x1DED4DB1684C21C4ULL, "gen_im_trs_bst_drg_gland_02.uti"},
	{0x1DEDFBA8FA4C80EBULL, "vxw_blightlands_ao_0_0.vlm"},
	{0x1DEE81E09A0DD52CULL, "vxlt_gallowscourtyard_4ll_0_0.vlm"},
	{0x1DEF006FD401EB26ULL, "6040587_m.fxe"},
	{0x1DEF39AC08F41FE9ULL, "gen00fl_varric6080083.cl"},
	{0x1DEF760F37D34B01ULL, "hlac_lt_0_lowtown_a2_65536_7x.dds"},
	{0x1DEFB61B32A7F64AULL, "vxht_0_hightown_n_aol_0_0.vlm"},
	{0x1DEFE0A80CCFF890ULL, "follower_banter6167056.cl"},
	{0x1DF061BEB37B9E01ULL, "6180459_m.fxe"},
	{0x1DF085DD891041F0ULL, "hlac_lt_0_lowtown_-131072_3y.dds"},
	{0x1DF1707EA830F895ULL, "character_character_p121.psh.fxo.dx11"},
	{0x1DF3287D33CD32E7ULL, "vxlt_hangedman_n_25b_0_0.vlm"},
	{0x1DF38B4E054D6820ULL, "hlac_dr_dungeon_3207_1y.dds"},
	{0x1DF3E2CC88E8B7ECULL, "6101112_m.fxe"},
	{0x1DF422CFD7A28AA2ULL, "hlac_lt_0_lowtown_docks_a3_-4_0z.dds"},
	{0x1DF4608B1C8B72CDULL, "vxht_0_hightown_a3_c0n_0_0.vlm"},
	{0x1DF4BE68EECB9320ULL, "vxw_bonepitdestroyed_a3_1ls_0_0.vlm"},
	{0x1DF62CA821077C7CULL, "6083547_f.fxe"},
	{0x1DF6986BED0DD394ULL, "zz_vault_debug6203282.cl"},
	{0x1DF6E5DC89AA1C7CULL, "dae111_gamlen_intro.cut"},
	{0x1DF74C53035C8EF6ULL, "lgt200_eau_tome6171792.cl"},
	{0x1DF78C56B82D23F3ULL, "hlac_lt_0_lowtown_-65535_5x.dds"},
	{0x1DF79936A002603EULL, "vxlt_gallowstemplar_f_4wr_0_0.vlm"},
	{0x1DF7FD5C4B9DF431ULL, "hlac_ht_0_hightown_-196610_0x.dds"},
	{0x1DF82EFE0C0A7DACULL, "6110987_m.fxe"},
	{0x1DF8490E15F44E43ULL, "819845961.ogg"},
	{0x1DF9ABFD42B4A7B6ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.impa.evt"},
	{0x1DFA4C77D270D781ULL, "vxlt_undercitydungeon_141_0_0.vlm"},
	{0x1DFA4FEB55AFEE9CULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax102a_0n.dds"},
	{0x1DFAA384B0210DC4ULL, "6070483_m.fxe"},
	{0x1DFAAB6152C16E80ULL, "vxw_encounter2_nb_0_0.vlm"},
	{0x1DFAB5B22A150137ULL, "vxw_sundermount_374_0_0.vlm"},
	{0x1DFAD72238F07962ULL, "vxlt_0_lowtown_docks_n_b51_0_1.vlm"},
	{0x1DFADA1E12239DCBULL, "vxht_0_hightown_a3_byz_0_0.vlm"},
	{0x1DFB130D3FDA23D2ULL, "hlac_lt_0_lowtown_-65536_1x.dds"},
	{0x1DFB5A6EB36EF40CULL, "vxlt_0_lowtown_n_cmn_0_0.vlm"},
	{0x1DFB7DF438832D19ULL, "746036269.ogg"},
	{0x1DFB9A2B5590A077ULL, "vxlt_0_lowtown_a2_eqt_0_0.vlm"},
	{0x1DFBD171D0C09790ULL, "one100dg_vanard6017506.cl"},
	{0x1DFBD8D507B0C07EULL, "vxlt_0_lowtown_a2_n_gpc_0_0.vlm"},
	{0x1DFBEF294C88CCECULL, "dre000_sandal.stg"},
	{0x1DFC0044D1929A20ULL, "z_dust2_b.dds"},
	{0x1DFC19190D6FC614ULL, "655129594.ogg"},
	{0x1DFC6D85EC42996BULL, "vxw_blightlands_jn_0_0.vlm"},
	{0x1DFCD19433F7E7D8ULL, "754158705.ogg"},
	{0x1DFCE8D1CBA7CFF7ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_oldmonument.mmh"},
	{0x1DFCECD1CBA7D604ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_oldmonument.mao"},
	{0x1DFCFED1CBA7F569ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_oldmonument.msh"},
	{0x1DFDAC0AD4EA5407ULL, "6139880_m.fxe"},
	{0x1DFDCE482A97AF6DULL, "hlac_lt_0_lowtown_n_131072_6y.dds"},
	{0x1DFDD4912F758B3BULL, "vxht_0_hightown_n_b9l_0_0.vlm"},
	{0x1DFDF7D2F495DC7AULL, "dae000_arishok6067818.cl"},
	{0x1DFE085E132D4481ULL, "fx_glyph.dds"},
	{0x1DFE8C9C9923A961ULL, "412174699.ogg"},
	{0x1DFEC80B70E23AE9ULL, "l_mag121ar_store.bnk"},
	{0x1DFEE70754B16FD9ULL, "234783213.ogg"},
	{0x1DFF738156C5EC22ULL, "follower_banter6173110.cl"},
	{0x1DFF775C85AE4CCAULL, "vxlt_0_lowtown_docks_cox_0_0.vlm"},
	{0x1DFF9C46338F5042ULL, "6115379_m.fxe"},
	{0x1DFFE611F1F1C296ULL, "mag291_vincento6027728.cl"},
	{0x1E003E70FCE6A42FULL, "vxw_bonepit_29h_0_0.vlm"},
	{0x1E00842A389C2000ULL, "1072129246.ogg"},
	{0x1E00909C28031AAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_informal.evt"},
	{0x1E0097AF7E21C943ULL, "6173402_m.fxe"},
	{0x1E00A2FD42BAB741ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.impa.gad"},
	{0x1E00D920E09CA6D6ULL, "457756871.ogg"},
	{0x1E017D6399F90945ULL, "7959609.ogg"},
	{0x1E0184097395055FULL, "vxht_0_hightown_n_cb8_0_0.vlm"},
	{0x1E01BB3362579512ULL, "1025769185.ogg"},
	{0x1E01DC765C8ACACFULL, "hlac_lt_darktown_-1_73y.dds"},
	{0x1E025C48D1982C5EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pota_n.dds"},
	{0x1E02BAFAEE11618BULL, "vxlt_hangedman_n_16h_0_0.vlm"},
	{0x1E02D2680E818B4DULL, "vxht_0_hightown_a3_c5m_0_0.vlm"},
	{0x1E0300E3EE928213ULL, "6144527_m.fxe"},
	{0x1E031FF4A96EDAD2ULL, "vxca_caverns1_ebd_0_0.vlm"},
	{0x1E036D63C6E5DB70ULL, "dae000ip_crowd_ff3_talk11.utp"},
	{0x1E03AC0099D6552FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medb_0d.dds"},
	{0x1E03ADAD59BE2E68ULL, "isa111_isabela6164601.cl"},
	{0x1E03D2228280638EULL, "follower_banter6149205.cl"},
	{0x1E041ABE0C4C17B4ULL, "gen00fl_fenris6127039.cl"},
	{0x1E0444EE640348B7ULL, "hlac_w_sundermount_-65535_0y.dds"},
	{0x1E045CB23F5E57EDULL, "vxht_mansion_n_61f_0_0.vlm"},
	{0x1E0485271119DBBFULL, "vxlt_hangedman_rv_0_0.vlm"},
	{0x1E0508A1ACE7CFE8ULL, "vxlt_0_lowtown_a2_n_c_ejg_0_0.vlm"},
	{0x1E0549ED96DC0201ULL, "ser191_emeric6104793.cl"},
	{0x1E05535916FAF34BULL, "vxht_0_hightown_n_c3r_0_1.vlm"},
	{0x1E058B039237616BULL, "6167490_m.fxe"},
	{0x1E059BCDF4C74176ULL, "hla_ht_0_hightown_a3_-131070_0x.dds"},
	{0x1E05A59B0399A916ULL, "1046679915.ogg"},
	{0x1E05B98DB53DE0D7ULL, "vxca_caverns1_dpn_0_0.vlm"},
	{0x1E05C75206D24D47ULL, "llac_lt_gallowstemplar_7198_2.dds"},
	{0x1E05D61951FED70AULL, "406313948.ogg"},
	{0x1E05E70CBE4DF024ULL, "distor_si.dds"},
	{0x1E05EF941A7BDB89ULL, "1022114316.ogg"},
	{0x1E0673543F25BBA5ULL, "6097042_f.fxe"},
	{0x1E06BC105E5A98D8ULL, "vxlt_undercitydungeon_247_0_0.vlm"},
	{0x1E0780CFCEBEF979ULL, "644623037.ogg"},
	{0x1E07879C28092A59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_informal.gad"},
	{0x1E07A3FD42C0D7EFULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.impa.ani"},
	{0x1E07DC9E12ADC86BULL, "mer240_jansen6050444.cl"},
	{0x1E07E93A1878F4FEULL, "gen00fl_carver6108937.cl"},
	{0x1E084047C6B2CE9FULL, "tv_door_01.mao"},
	{0x1E08DC74D9FFA7EFULL, "6108735_f.fxe"},
	{0x1E091D3E44B78700ULL, "vxw_encounter2_ic_0_0.vlm"},
	{0x1E09278F1C0B19B7ULL, "vxw_sundermount_327_0_0.vlm"},
	{0x1E09C1B7CE96443AULL, "vxht_brothel_n_5im_0_0.vlm"},
	{0x1E09F8CCB601A401ULL, "vxca_cavern1_haunted_hi6_0_0.vlm"},
	{0x1E0A52AC289560EFULL, "bld_pro_02.dds"},
	{0x1E0ADF62DE38B1EBULL, "vxw_blightlands_oa_0_0.vlm"},
	{0x1E0AE61952737439ULL, "6115857_m.fxe"},
	{0x1E0B4E3A19BB8F79ULL, "llac_dr_0_deeproads_5841_3.dds"},
	{0x1E0B82AD94DD8CFCULL, "vxca_caverns2prc_9ws_0_0.vlm"},
	{0x1E0C69741068791EULL, "hlac_w_sundermount_131072_0y.dds"},
	{0x1E0D09C48D9683D2ULL, "dae000_meredith6166282.cl"},
	{0x1E0D38695252F06DULL, "454516833.ogg"},
	{0x1E0DB869CA467AF5ULL, "795725404.ogg"},
	{0x1E0DF1728D6CECCBULL, "801441971.ogg"},
	{0x1E0E5A2676FE9DB9ULL, "6135284_m.fxe"},
	{0x1E0E989C280F65A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_informal.ani"},
	{0x1E0E9EBD96BEAB67ULL, "vxht_bartrand_n_4bd_0_0.vlm"},
	{0x1E0EB04DEEDCBCAFULL, "vxw_bonepit_24k_0_0.vlm"},
	{0x1E0EF1B4A205AE7DULL, "215945962.ogg"},
	{0x1E0FE8186E503AD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\swoverlay.anb"},
	{0x1E10090844CFF7ADULL, "dae000_hate_follower_react6103201.cl"},
	{0x1E100B01D2A1ECA5ULL, "vxlt_gallowstemplar_3o1_0_0.vlm"},
	{0x1E1063651D5576F5ULL, "6070110_m.fxe"},
	{0x1E1063E7E32B2C09ULL, "follower_banter6161130.cl"},
	{0x1E106A358E115134ULL, "vxht_brothel_n_3vt_0_0.vlm"},
	{0x1E10F58971762D9EULL, "mag360_samson6042062.cl"},
	{0x1E11147DFDCD3FB0ULL, "and100_anders6206343.cl"},
	{0x1E115309D546F323ULL, "471809884.ogg"},
	{0x1E11E6C3CA8CF2CCULL, "con_ico_investigate.dds"},
	{0x1E12AC6407A3F3F3ULL, "vxdr_0_deeproads_euw_0_0.vlm"},
	{0x1E133D449F97C120ULL, "vxlt_0_lowtown_docks_b0b_0_0.vlm"},
	{0x1E1352B441578A1DULL, "llac_lt_gallowstemplar_2382_0.dds"},
	{0x1E137A7E9EDDE868ULL, "vxlt_0_lowtown_a2_n_c_eod_0_0.vlm"},
	{0x1E13EEB51C33C201ULL, "6174214_m.fxe"},
	{0x1E1494D3F1FEA863ULL, "404876248.ogg"},
	{0x1E15A352131484F1ULL, "hlac_lt_0_lowtown_docks_a3_-65539_5z.dds"},
	{0x1E15C422725107A8ULL, "mag100_karras6034723.cl"},
	{0x1E1647B92A231255ULL, "6096064_m.fxe"},
	{0x1E164FDA5872E535ULL, "vxlt_hangedman_25s_0_0.vlm"},
	{0x1E168982DBDC1FEBULL, "370570939.ogg"},
	{0x1E16C9804632AAF6ULL, "814950793.ogg"},
	{0x1E170AF79FB25E68ULL, "427158076.ogg"},
	{0x1E1734C6566B9AA8ULL, "vxlt_0_lowtown_a2_n_c_d66_0_0.vlm"},
	{0x1E1742496B0450ECULL, "gen00fl_fenris6108054.cl"},
	{0x1E17DAA0B03E483BULL, "vxdr_primevalentrance_30l_0_0.vlm"},
	{0x1E183D38D7FF84C7ULL, "chu_02.dds"},
	{0x1E183E270E6651DFULL, "180873440.ogg"},
	{0x1E197BCFAA87383CULL, "524460166.ogg"},
	{0x1E19D1E0FB9CF051ULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpa.mmh"},
	{0x1E19D608BFEBE96AULL, "621725164.ogg"},
	{0x1E19EBE0FB9D1C0FULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpa.msh"},
	{0x1E1A31D686107149ULL, "gen00fl_dog6166807.cl"},
	{0x1E1A42EF2B696D6EULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medb.msh"},
	{0x1E1A5CEF2B69995CULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medb.mmh"},
	{0x1E1AA3AD1841AF0FULL, "art\\levels\\4cinematics\\ferelden\\prp_barge.mao"},
	{0x1E1AA945703025DAULL, "vxca_cavern1_haunted_jg5_0_0.vlm"},
	{0x1E1AA9BDDFA178A9ULL, "6104175_f.fxe"},
	{0x1E1AAFAD1841C3ACULL, "art\\levels\\4cinematics\\ferelden\\prp_barge.mmh"},
	{0x1E1AB5AD1841CD9EULL, "art\\levels\\4cinematics\\ferelden\\prp_barge.msh"},
	{0x1E1B048BB5E96733ULL, "pro000cr_hurlock_firewall.utc"},
	{0x1E1B543E9513F7FFULL, "6088748_m.fxe"},
	{0x1E1BA554AFE2D0CFULL, "6087180_f.fxe"},
	{0x1E1BD20EB3A8C8D5ULL, "follower_banter6160645.cl"},
	{0x1E1BD868691EEC5FULL, "249585321.ogg"},
	{0x1E1C195E13465F0BULL, "fx_glyph.mao"},
	{0x1E1C4CA5CFF3C9ADULL, "hlac_lt_0_lowtown_docks_a3_n_0_2y.dds"},
	{0x1E1C5B16699A7DCAULL, "vxlt_0_lowtown_docks_cuz_0_0.vlm"},
	{0x1E1C664E76437AB6ULL, "6071886_m.fxe"},
	{0x1E1CBE320DDD6F4FULL, "follower_banter6169990.cl"},
	{0x1E1D025AF6388647ULL, "6140321_m.fxe"},
	{0x1E1D26292E7FB854ULL, "vxw_bonepitmining_a2_1xa_0_0.vlm"},
	{0x1E1D3771CD35A352ULL, "gen_im_wep_rog_dua_act3_06.uti"},
	{0x1E1D9B1781CFAFB7ULL, "terrain_blend3.psh.fxo"},
	{0x1E1DC510A58D9AB0ULL, "6103213_m.fxe"},
	{0x1E1E3FA8A2ADD1C5ULL, "gen00fl_merrill6107805.cl"},
	{0x1E1E7DCD7277AE4DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grindstone_n.dds"},
	{0x1E1ED391DF4DC01FULL, "6084186_m.fxe"},
	{0x1E1F9FE6F74B9D89ULL, "6167276_m.fxe"},
	{0x1E2003AE8D5B0BD2ULL, "vxca_caverns1_ehf_0_0.vlm"},
	{0x1E20CA48D59029AFULL, "hlac_dr_0_deeproads_17039_1z.dds"},
	{0x1E211B3E4F20EF8BULL, "zz_fol_debug6096545.cl"},
	{0x1E2170F66DD9DFEFULL, "vxdr_0_deeproads_bp7_0_0.vlm"},
	{0x1E21732803041D05ULL, "vxlt_undercityserialkiller_1om_0_0.vlm"},
	{0x1E21AF21918DD9A0ULL, "vxlt_0_lowtown_docks_b7c_0_0.vlm"},
	{0x1E21FB5A5E0870C3ULL, "vxca_caverns1_ioj_0_0.vlm"},
	{0x1E23AC1F26A5BE70ULL, "93321741.ogg"},
	{0x1E23FD1B81D0A38EULL, "w_bonepit_24_0.msh"},
	{0x1E240B1B81D0BBDFULL, "w_bonepit_24_0.mao"},
	{0x1E24171B81D0D07CULL, "w_bonepit_24_0.mmh"},
	{0x1E24C1B74A68FDB5ULL, "vxlt_hangedman_22r_0_0.vlm"},
	{0x1E24C3CAA78393E6ULL, "ca_caverns2prc_mmi.gff"},
	{0x1E250C3CED28C58DULL, "49697905.ogg"},
	{0x1E250F8846D2963EULL, "520042872.ogg"},
	{0x1E25A8AEE8AEF8B7ULL, "llac_dr_0_deeproads_5845_1.dds"},
	{0x1E268C4CF6A46413ULL, "gen00fl_isabela6096950.cl"},
	{0x1E26A77B3B540430ULL, "556042349.ogg"},
	{0x1E26A7D118B6F315ULL, "452623993.ogg"},
	{0x1E26D30F1A0D9331ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sitting_table_clasped_1_loop.gad"},
	{0x1E26DD813E12E9CEULL, "244783893.ogg"},
	{0x1E2751DC7B391C2FULL, "gen00fl_aveline6122757.cl"},
	{0x1E2783042409F79FULL, "784960283.ogg"},
	{0x1E2785485786DCBDULL, "character_rendertodepth_p7.psh.fxo.dx11"},
	{0x1E278FA8DEAB7524ULL, "vxlt_0_lowtown_docks_a3_n_d38_0_1.vlm"},
	{0x1E281EA08DE97A2AULL, "6083436_m.fxe"},
	{0x1E285CD2388F5497ULL, "417969856.ogg"},
	{0x1E2863AC28AE7B61ULL, "bld_pro_02.mao"},
	{0x1E2932DD4428C77AULL, "vxht_0_hightown_auq_0_0.vlm"},
	{0x1E29367E810F88C4ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.anb"},
	{0x1E2A0B9BEDF2230BULL, "263391894.ogg"},
	{0x1E2A2C3DDC08F017ULL, "6034353_f.fxe"},
	{0x1E2A653B5718EB1BULL, "233417094.ogg"},
	{0x1E2A79192A75E408ULL, "gen00fl_anders6119292.cl"},
	{0x1E2ACCBE33E582AEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basicx.msh"},
	{0x1E2ADABE33E59AFFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basicx.mao"},
	{0x1E2AE6BE33E5AE9CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basicx.mmh"},
	{0x1E2B07AAF453A4B5ULL, "zz_followers_banter6159012.cl"},
	{0x1E2B2184450A1DD8ULL, "6148854_m.fxe"},
	{0x1E2B8BAAE3E44D32ULL, "vxht_keep_78h_0_0.vlm"},
	{0x1E2B98062075D0D4ULL, "vxw_bonepitmining_a2_1uf_0_0.vlm"},
	{0x1E2BC1F8CC877773ULL, "vxlt_warehouse_105_0_0.vlm"},
	{0x1E2C42BB82B695DAULL, "463787960.ogg"},
	{0x1E2C46EAE57D3EF6ULL, "6110057_m.fxe"},
	{0x1E2C63CA33DCC1A0ULL, "vxdr_primevaltomb_8vr_0_0.vlm"},
	{0x1E2CE30A4E74CD78ULL, "vxlt_undercityserialkiller_1sp_0_0.vlm"},
	{0x1E2CEC7523A22812ULL, "mcr371_orsino6083472.cl"},
	{0x1E2D1AC02C2E0815ULL, "zz_amb_test6105063.cl"},
	{0x1E2D6177340C0981ULL, "hlac_lt_0_lowtown_n_0_9x.dds"},
	{0x1E2D766C715038B7ULL, "isa321_velasco6117629.cl"},
	{0x1E2D77591F7CC8D5ULL, "vxw_woundedcoast_127_0_0.vlm"},
	{0x1E2DD40F1A13B3DFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sitting_table_clasped_1_loop.ani"},
	{0x1E2DF815330694C0ULL, "hlac_lt_0_lowtown_n_-1_7y.dds"},
	{0x1E2E5A35CA69D3E7ULL, "dae300_bran6143446.cl"},
	{0x1E2E93358DA61E97ULL, "vxht_mansion2_2a7_0_0.vlm"},
	{0x1E2EC08C8472F603ULL, "dae200cr_viscount6087188.cl"},
	{0x1E2EC94A168F86B2ULL, "654785117.ogg"},
	{0x1E2EF7294CB433D7ULL, "dre000_sandal.bnk"},
	{0x1E2F734FB5E7CF7DULL, "6110564_m.fxe"},
	{0x1E2F811267FE8DE4ULL, "vxlt_shop_ac_0_0.vlm"},
	{0x1E2FD99B84498B19ULL, "6087035_m.fxe"},
	{0x1E2FDEA1C0BE7B6AULL, "lgt000cr_hate_templar.utc"},
	{0x1E30276467004D75ULL, "6085791_f.fxe"},
	{0x1E307951F07D1C6AULL, "667743122.ogg"},
	{0x1E3086847DA7D413ULL, "l_lgt110ar_gang_hideout_d.wbd"},
	{0x1E30BBB619752B28ULL, "vxw_sundermount_2ds_0_0.vlm"},
	{0x1E30E4D8F7EAC192ULL, "122733457.ogg"},
	{0x1E316BDED93FA796ULL, "vxlt_0_lowtown_docks_n_bsb_0_0.vlm"},
	{0x1E31D9294CB636EFULL, "dre000_sandal.cnv"},
	{0x1E32EA0172570F65ULL, "hlac_ht_0_hightown_n_65537_1x.dds"},
	{0x1E33427A5220D0E6ULL, "6180979_m.fxe"},
	{0x1E3400727791F2E0ULL, "679184650.ogg"},
	{0x1E3492E96A2FF8B1ULL, "hlac_lt_0_lowtown_n_131071_6x.dds"},
	{0x1E34C4587D97281CULL, "mag291_torpor6116231.cl"},
	{0x1E34DBC70B1D281AULL, "vxw_encounter2_camp_fx_0_0.vlm"},
	{0x1E350984C99DBC0CULL, "vxw_bonepit_1wk_0_1.vlm"},
	{0x1E356FC364E3FA63ULL, "6206267_m.fxe"},
	{0x1E3590F30F7E9642ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyo_0n.dds"},
	{0x1E35A59947768BDAULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown2.mmh"},
	{0x1E35B7994776AA20ULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown2.msh"},
	{0x1E3612957D05A530ULL, "gen00pt_ai.ncs"},
	{0x1E361804709986C6ULL, "590149574.ogg"},
	{0x1E36247999A705FDULL, "vxlt_gallowstemplar_n_5pa_0_0.vlm"},
	{0x1E363E210185ECCCULL, "vxlt_0_lowtown_a2_n_c_fgw_0_0.vlm"},
	{0x1E36557E811A47C7ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.mmh"},
	{0x1E366C5D006D8397ULL, "gen00fl_fenris6103340.cl"},
	{0x1E3677A1D9A5CA42ULL, "pro000ip_gallows_gate.utp"},
	{0x1E36D8446ABFD67CULL, "vxca_caverns2prc_9dv_0_0.vlm"},
	{0x1E37FE3D8E5F9DFCULL, "1050856618.ogg"},
	{0x1E38294A4C7293DAULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killervase.phy"},
	{0x1E3837F8790028FFULL, "6086948_m.fxe"},
	{0x1E383EF78DB54759ULL, "vxht_0_hightown_n_ca3_0_0.vlm"},
	{0x1E38440F54536A84ULL, "vxlt_darktown_6bz_0_0.vlm"},
	{0x1E389F84013E1B43ULL, "hlac_lt_gallowstemplar_n_2394_1y.dds"},
	{0x1E38CE8A390B59CFULL, "761725582.ogg"},
	{0x1E38F4ACA44103CEULL, "ave321tr_aveline_final.utt"},
	{0x1E390F56AFE3A46BULL, "gen_im_trs_slv_lgt_boo_r1.uti"},
	{0x1E3926066E521964ULL, "677186395.ogg"},
	{0x1E39385DD0036311ULL, "496436838.ogg"},
	{0x1E393B6D055FE7F8ULL, "plt_seb120pt_duty.nss"},
	{0x1E3942827C32AAABULL, "149483169.ogg"},
	{0x1E395971357B9B92ULL, "mrl000pt_main.ncs"},
	{0x1E3A75CF1CEA3B2FULL, "vxht_mansion2_1yu_0_1.vlm"},
	{0x1E3A8D68BEC75401ULL, "vxca_caverns1_ckj_0_0.vlm"},
	{0x1E3A959CC2E6AD37ULL, "w_woundedcoast_21_0.rml"},
	{0x1E3AC5B9B5EE4A40ULL, "gen_im_trs_gen_hvy_glv_r2.uti"},
	{0x1E3AF6A9BA943FC7ULL, "zz_mer_debug6039437.cl"},
	{0x1E3B122C1EE9FA57ULL, "vxw_bonepit_2ah_0_0.vlm"},
	{0x1E3BD3C41EAB134DULL, "llac_lt_gallowscourtyard_-131072_4.dds"},
	{0x1E3C30E6E20A43CCULL, "one300dg_solivitus6056840.cl"},
	{0x1E3C42151828246DULL, "vxht_keep_n_6h5_0_0.vlm"},
	{0x1E3C743CE3FFB3D6ULL, "270690248.ogg"},
	{0x1E3CDA901237385BULL, "vxlt_hangedman_26x_0_0.vlm"},
	{0x1E3D319A4D48D997ULL, "hla_ht_0_hightown_65536_0z.dds"},
	{0x1E3D32CFC6196E28ULL, "hlac_w_sundermount_-131068_0y.dds"},
	{0x1E3D69BA43BC6E53ULL, "vxw_bonepit_16b_0_0.vlm"},
	{0x1E3D7099DF7C3982ULL, "genip_treasure_pile.utp"},
	{0x1E3F2D930B6B43A8ULL, "vxw_sundermount_2ip_0_0.vlm"},
	{0x1E3F3BEF2B887D27ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medb.phy"},
	{0x1E3F457FC54F625EULL, "ca_caverns1_mines_24421_lc.mmh"},
	{0x1E3F49B0C5231010ULL, "dre000_dougal6129794.cl"},
	{0x1E3FD955916FEFBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_loop.gad"},
	{0x1E41E47796FE0DBDULL, "6177831_m.fxe"},
	{0x1E42864607CF0EC0ULL, "6126982_m.fxe"},
	{0x1E42B7BFD4018495ULL, "dae000_gamlen6166380.cl"},
	{0x1E42CD3F7C9ED748ULL, "gen00fl_sebastian6188673.cl"},
	{0x1E43171B81EB80CFULL, "w_bonepit_24_0.rml"},
	{0x1E434DA3FD13409AULL, "vxw_encounter2_camp_ay_0_0.vlm"},
	{0x1E4357E576200185ULL, "gen_im_wep_war_shd_act3_02.uti"},
	{0x1E43AA8AEA715374ULL, "vxw_encounter2_qn_0_1.vlm"},
	{0x1E43EE666F317FE3ULL, "vxht_0_hightown_a2_bg1_0_0.vlm"},
	{0x1E44898F0A1A8BAFULL, "hlac_dr_slavers_3215_2z.dds"},
	{0x1E45479F17CAD1DAULL, "704859681.ogg"},
	{0x1E45A60B711E033BULL, "l_mag121ar_store.wbd"},
	{0x1E4684BEA39C585FULL, "vxw_bonepitdestroyed_a3_1ge_0_0.vlm"},
	{0x1E46CDFE29A20048ULL, "w_sundermount_223_0.mmh"},
	{0x1E46D85591760CBAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_loop.ani"},
	{0x1E46D9FE29A214EBULL, "w_sundermount_223_0.mao"},
	{0x1E46E3A5EE07D912ULL, "6007744_m.fxe"},
	{0x1E46EA7005110CD2ULL, "vxw_bonepit_16g_0_0.vlm"},
	{0x1E46EBFE29A23372ULL, "w_sundermount_223_0.msh"},
	{0x1E47515F8F459A04ULL, "zz_mcr_debug6080867.cl"},
	{0x1E4765CF2F047C97ULL, "gen00fl_aveline6114073.cl"},
	{0x1E4794C82A9969FDULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls107a_0e.dds"},
	{0x1E4799F800347DF6ULL, "llac_ca_caverns1_24423_0.dds"},
	{0x1E47BD0A6196157BULL, "ht_0_hightownexplosion_n_10_0.phy"},
	{0x1E47BE50DC31AED8ULL, "gen00fl_isabela6135656.cl"},
	{0x1E47D6B9DF30FBBDULL, "ico_dagger_4.dds"},
	{0x1E48BE6B61108153ULL, "vxw_woundedcoast_7o_0_0.vlm"},
	{0x1E496684C00A1895ULL, "vxht_chantry_2tg_0_1.vlm"},
	{0x1E4978B4FF9A7E5AULL, "gen00fl_aveline6122134.cl"},
	{0x1E4A1F719CD177E1ULL, "dae000ip_crowd_mf3_talk15.utp"},
	{0x1E4A3D2D00550B84ULL, "765651394.ogg"},
	{0x1E4A56A62252D7CFULL, "6171857_m.fxe"},
	{0x1E4A5B130368F9D5ULL, "vxw_woundedcoast_185_0_0.vlm"},
	{0x1E4AB3F20A1E3CEDULL, "vxht_keep_n_6c4_0_0.vlm"},
	{0x1E4AB5872E3BDFFDULL, "gen00fl_fenris6110786.cl"},
	{0x1E4AC7F8A64AA0D8ULL, "vxlt_hangedman_16h_0_0.vlm"},
	{0x1E4AD9624FE9B2CFULL, "vxw_woundedcoast_1c6_0_0.vlm"},
	{0x1E4AFE2E550C023DULL, "ser191_emeric6120449.cl"},
	{0x1E4B20D77C63A5A9ULL, "vxw_bonepit_8a_0_0.vlm"},
	{0x1E4B5450F94C5528ULL, "llac_lt_0_lowtown_a2_n_c_65537_2.dds"},
	{0x1E4BA1087DF8986AULL, "49919439.ogg"},
	{0x1E4CDAC1B2DC66C5ULL, "qun221_petrice6076811.cl"},
	{0x1E4D21675EAE0AE3ULL, "gen00fl_aveline6121969.cl"},
	{0x1E4D22293138E2D9ULL, "vxlt_gallowstemplar_43r_0_0.vlm"},
	{0x1E4D9F6FFD615C28ULL, "vxw_sundermount_2rq_0_0.vlm"},
	{0x1E4E1171D35E1A23ULL, "vxlt_gallowstemplar_474_0_0.vlm"},
	{0x1E4F2FB4FC1055DDULL, "mag110_arianni6016765.cl"},
	{0x1E4F30438B4773D2ULL, "dae000_gamlen6165973.cl"},
	{0x1E4F349CCC08E3CFULL, "t3_enc.tnt"},
	{0x1E4FD6B49DA1F63FULL, "6015184_m.fxe"},
	{0x1E500B8F9CE563D7ULL, "293779268.ogg"},
	{0x1E504AD0546731D8ULL, "hlac_lt_0_lowtown_-65536_8y.dds"},
	{0x1E508FB63270D66EULL, "574331209.ogg"},
	{0x1E50A4D04A627CBDULL, "mind_02_glow.mao"},
	{0x1E50C5BE34064667ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basicx.phy"},
	{0x1E50C893B98AD141ULL, "96498425.ogg"},
	{0x1E50E7ACA4557FF7ULL, "ave321tr_aveline_final.ncs"},
	{0x1E51892A11BD7271ULL, "qun110_petrice6018652.cl"},
	{0x1E51F57B6D3F2DCAULL, "vxdr_primevalentrance_2oi_0_0.vlm"},
	{0x1E51FBCC29E85EA6ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_001a_0t.dds"},
	{0x1E523CF9E136E9F0ULL, "6121169_m.fxe"},
	{0x1E523F2122651FD2ULL, "vxw_woundedcoast_7j_0_0.vlm"},
	{0x1E5265DA1976F716ULL, "6127060_m.fxe"},
	{0x1E52D48912CB3D5FULL, "vxdr_primevaltomb_8ny_0_0.vlm"},
	{0x1E52FAA8499119DCULL, "vxw_encounter2_c9_0_0.vlm"},
	{0x1E53456ABB9DF5EFULL, "24135848.ogg"},
	{0x1E542A81489D56C5ULL, "vxht_0_hightown_auz_0_0.vlm"},
	{0x1E547D1A9DDF1CE4ULL, "follower_banter6161195.cl"},
	{0x1E549F179FECAFA9ULL, "vxlt_0_lowtown_a3_n_c_eox_0_0.vlm"},
	{0x1E56681C161BDDBDULL, "hlac_lt_gallowscourtyard_n_-131072_1z.dds"},
	{0x1E5742C00536ED86ULL, "540866788.ogg"},
	{0x1E576DFE4F500568ULL, "mag100_cullen6093536.cl"},
	{0x1E57E8705605D9FDULL, "vrc321st_bartrand_item1.stg"},
	{0x1E57F07F311EA363ULL, "6049714_m.fxe"},
	{0x1E58977700EA0809ULL, "6129212_m.fxe"},
	{0x1E594B3F41DFCB4FULL, "vxw_woundedcoast_1n1_0_0.vlm"},
	{0x1E5988A90E02A979ULL, "89297167.ogg"},
	{0x1E599665708158ACULL, "1048336900.ogg"},
	{0x1E59AC5907179D49ULL, "llac_w_blightlands_-65536_0.dds"},
	{0x1E59E8CC63886817ULL, "vxht_mansion2_2p0_0_0.vlm"},
	{0x1E59F81AE7968DA9ULL, "6107164_m.fxe"},
	{0x1E5A251B8FF98CC3ULL, "w_sundermount_61_0.rml"},
	{0x1E5ABD2AD2B8D01AULL, "vxca_caverns2prc_8ta_0_0.vlm"},
	{0x1E5B13E2A680BCB4ULL, "zz_stever_test6144484.cl"},
	{0x1E5B2BB0544A1A38ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_bonepile_n.dds"},
	{0x1E5B3A9BCADC7F05ULL, "vxlt_undercityserialkiller_1si_0_0.vlm"},
	{0x1E5BDAD256FF4CF6ULL, "llac_w_sundermount_-196613_0.dds"},
	{0x1E5BF89B108CC358ULL, "vxht_mansion_n_54g_0_0.vlm"},
	{0x1E5D03601BC3B200ULL, "vxht_brothel_n_6ep_0_0.vlm"},
	{0x1E5D0B136E4A68E2ULL, "zz_qun_debug6083721.cl"},
	{0x1E5D1B0CB7FC29CAULL, "6019109_m.fxe"},
	{0x1E5DA019BC463784ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof3extrude.msh"},
	{0x1E5DB219BC4656EDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof3extrude.mao"},
	{0x1E5DB619BC465D26ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof3extrude.mmh"},
	{0x1E5E9D05666BDA5CULL, "vxlt_0_lowtown_a3_n_c_hfc_0_0.vlm"},
	{0x1E5EB6377A242218ULL, "ave331_jeven6138730.cl"},
	{0x1E5F001E28771A5FULL, "6151135_f.fxe"},
	{0x1E5F0A4C57D98FA6ULL, "279923080.ogg"},
	{0x1E601BC81AC04E80ULL, "vxw_blightlands_7f_0_0.vlm"},
	{0x1E60AC01BA2B4C1FULL, "926809236.ogg"},
	{0x1E60CE89A754CA3FULL, "ser221_quentin6123393.cl"},
	{0x1E60D1CB747693BFULL, "hlac_lt_0_lowtown_docks_131072_1x.dds"},
	{0x1E612C1B8FFFB73BULL, "w_sundermount_61_0.phy"},
	{0x1E6165B1BC32FBE2ULL, "543388957.ogg"},
	{0x1E619E6E5CF1DE6BULL, "follower_rival.dds"},
	{0x1E625EEE300EA44BULL, "vxht_mansion_n_70f_0_0.vlm"},
	{0x1E62836145E4C594ULL, "llac_ht_chantry_2534_32.dds"},
	{0x1E62BF46F6DA1D3BULL, "6208261_m.fxe"},
	{0x1E62FD8516D11146ULL, "18142271.ogg"},
	{0x1E635E0F47B4DCE9ULL, "vxlt_hangedman_n_1i3_0_0.vlm"},
	{0x1E636E79E7B2531EULL, "34909799.ogg"},
	{0x1E63AF312FFB1BBFULL, "hlac_lt_0_lowtown_docks_n_-3_4x.dds"},
	{0x1E641BF363A70019ULL, "vxca_caverns1_jfl_0_0.vlm"},
	{0x1E64FAE882593674ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower_mcr370_playerflees.gad"},
	{0x1E65039A5468865FULL, "vxlt_0_lowtown_docks_b9j_0_0.vlm"},
	{0x1E651FF32E0524C5ULL, "hlac_w_bonepit_131075_0x.dds"},
	{0x1E6568B71039C57CULL, "hlac_w_encounter2_131072_0z.dds"},
	{0x1E659273A3C0D674ULL, "ss_qunari_m_2.wbd"},
	{0x1E65C4AE6C1BFD23ULL, "239160008.ogg"},
	{0x1E671C5F35796068ULL, "6021085_m.fxe"},
	{0x1E6757C27A7BC7A1ULL, "tbe100_xenon6189671.cl"},
	{0x1E67AD19BC4EA2F1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof3extrude.phy"},
	{0x1E67BD1C33D5E3CFULL, "vxw_woundedcoast_1i0_0_0.vlm"},
	{0x1E67E9A4584AB728ULL, "290895883.ogg"},
	{0x1E684D5452176EB7ULL, "fxa_web_02.mao"},
	{0x1E684E5700CCABF6ULL, "vxht_0_hightown_a2_ask_0_0.vlm"},
	{0x1E68590E1425006CULL, "hlac_ht_0_hightown_a2_-65536_0y.dds"},
	{0x1E68D317457A47D0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brk_1.mao"},
	{0x1E68D5D971A67C91ULL, "142248440.ogg"},
	{0x1E693CFCF1839421ULL, "llac_ca_cavern1_haunted_24427_4.dds"},
	{0x1E697E2E414C16B7ULL, "6133540_f.fxe"},
	{0x1E698119D6CE76DDULL, "65083755.ogg"},
	{0x1E699F29F0625541ULL, "6096892_f.fxe"},
	{0x1E6A8329E14D8D28ULL, "vxw_sundermount_2xo_0_0.vlm"},
	{0x1E6AFCE81389CB04ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_floor_d.dds"},
	{0x1E6B63184D3D4474ULL, "vxlt_foundry_n_z2_0_0.vlm"},
	{0x1E6BFF44F7943D8AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_deadfish_d.dds"},
	{0x1E6C09E8825F6E52ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower_mcr370_playerflees.ani"},
	{0x1E6CE2AFB40D6AE0ULL, "377504204.ogg"},
	{0x1E6D4E155A206C40ULL, "931475830.ogg"},
	{0x1E6D724AFC5023FFULL, "dae210_gate_guard6085137.cl"},
	{0x1E6D881B394D5F7FULL, "hlac_lt_0_lowtown_a2_n_c_0_5z.dds"},
	{0x1E6DB7C4D3DCAE19ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar10_.gad"},
	{0x1E6DE589A9632BC6ULL, "qun221_petrice6077417.cl"},
	{0x1E6DFD16F91BE39FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_en.gad"},
	{0x1E6E1761E16D2E12ULL, "vxlt_0_lowtown_a2_n_c_by7_0_0.vlm"},
	{0x1E6E8DA50CB66700ULL, "vxw_blightlands_8g_0_0.vlm"},
	{0x1E6EAA8C0D221679ULL, "6180584_f.fxe"},
	{0x1E6EC24E1305B569ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar9_.gad"},
	{0x1E6EC993482576A8ULL, "vxht_chantry_1yg_0_0.vlm"},
	{0x1E6ECE083F92C998ULL, "gen00fl_sebastian6189036.cl"},
	{0x1E6EF1B784542B43ULL, "llac_w_bonepit_196614_0.dds"},
	{0x1E6F4166ECF7CAB2ULL, "6076551_m.fxe"},
	{0x1E6F856BAE19FF9AULL, "6048742_f.fxe"},
	{0x1E6FC8165F95BB4FULL, "gen00fl_varric6093629.cl"},
	{0x1E704EE38041394BULL, "dae000cr_grand_cleric6148090.cl"},
	{0x1E70A641AAE3646DULL, "lt_0_lowtown_a2_n_c_51_0.rml"},
	{0x1E70B64C692AF54FULL, "127505739.ogg"},
	{0x1E70C13EFB674F0BULL, "228292158.ogg"},
	{0x1E70C41C45AA53B1ULL, "hlac_lt_undercitydungeon_-1_36y.dds"},
	{0x1E70CB456EDD381AULL, "dae000_bartrand6027574.cl"},
	{0x1E71097BD0D1360BULL, "ser120tr_ghyslain_ambient.ncs"},
	{0x1E7113D53D348F2AULL, "305735319.ogg"},
	{0x1E716C2DFDF741FAULL, "vxht_0_hightown_aln_0_0.vlm"},
	{0x1E717B0E24BC6EEEULL, "art\\characters\\creatures\\hurlock\\textures\\w_lsw_hlk_0d.dds"},
	{0x1E71C1CB1CAA87BDULL, "6135698_f.fxe"},
	{0x1E7271518225BAB3ULL, "779820280.ogg"},
	{0x1E73A202C9BA0883ULL, "957321833.ogg"},
	{0x1E73A40A61BB251AULL, "ht_0_hightownexplosion_n_10_0.msh"},
	{0x1E73B20A61BB3D53ULL, "ht_0_hightownexplosion_n_10_0.mao"},
	{0x1E73B60A61BB4460ULL, "ht_0_hightownexplosion_n_10_0.mmh"},
	{0x1E73F78C99574D22ULL, "173410585.ogg"},
	{0x1E7448550B38A5F9ULL, "6076748_f.fxe"},
	{0x1E745090B35270E2ULL, "vxlt_0_lowtown_eok_0_0.vlm"},
	{0x1E7453D67FCA13F9ULL, "art\\characters\\playercharacter\\qunarimale\\uni_com.mmh"},
	{0x1E745DD67FCA24E7ULL, "art\\characters\\playercharacter\\qunarimale\\uni_com.msh"},
	{0x1E74BA5BD23E7749ULL, "llac_lt_0_lowtown_n_65536_6.dds"},
	{0x1E74C8C4D3E2EA67ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar10_.ani"},
	{0x1E74EA4D084AD75AULL, "vxlt_backalley_g_1e0_0_0.vlm"},
	{0x1E74F216F921EF8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_en.evt"},
	{0x1E74F6403E82264BULL, "vxw_bonepit_1x9_0_1.vlm"},
	{0x1E75934E130B83B7ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar9_.ani"},
	{0x1E75E877B90A61FEULL, "320987805.ogg"},
	{0x1E762EF925CBFC4FULL, "vxw_woundedcoast_1t3_0_0.vlm"},
	{0x1E76B1177B1BBE94ULL, "vxht_bartrand_n_5tx_0_0.vlm"},
	{0x1E76DCFA3366BCA0ULL, "ht_0_hightown_a3_43_0.rml"},
	{0x1E76F1F1058FB582ULL, "zz_followers_banter6157741.cl"},
	{0x1E77640F21BD235FULL, "one220_yevhen6098686.cl"},
	{0x1E78195C2B0E6B22ULL, "leap_lit.mao"},
	{0x1E7837BC6F12A6ACULL, "6095522_f.fxe"},
	{0x1E7849CEB9BA5425ULL, "hlac_lt_gamlenshouse_n_-1_3z.dds"},
	{0x1E790501673051D0ULL, "art\\vfx\\combat\\vfx_w_gn1_build_up_c.mmh"},
	{0x1E797E0F0E2B7E4AULL, "mcr370_samson6087531.cl"},
	{0x1E799143D7285136ULL, "dre000_stroud6036673.cl"},
	{0x1E79B52A351C14B1ULL, "6131895_m.fxe"},
	{0x1E79D31C84E0F3F7ULL, "qcr220_grey_wardens6060383.cl"},
	{0x1E79D44FCAECF8B9ULL, "follower_banter6172952.cl"},
	{0x1E79DD3F985E910CULL, "hlac_ht_0_hightown_a2_65535_3z.dds"},
	{0x1E7A7C34EB3868ECULL, "lav_river01a.dds"},
	{0x1E7B0EF13BFD40F4ULL, "6094481_m.fxe"},
	{0x1E7B450443995306ULL, "llac_w_blightlands_196605_0.dds"},
	{0x1E7BFCF0C0E8B827ULL, "hlac_lt_0_lowtown_a3_n_c_65534_5z.dds"},
	{0x1E7C1E7045CB4054ULL, "373048975.ogg"},
	{0x1E7C363658C92885ULL, "6067679_m.fxe"},
	{0x1E7C398C72D41A65ULL, "mag101_grace6034059.cl"},
	{0x1E7D1112421BF3CEULL, "vxlt_0_lowtown_n_gos_0_1.vlm"},
	{0x1E7D3DCFF552F86EULL, "cd_dre000_vampire.wbd"},
	{0x1E7D85882E95B0C3ULL, "6090411_m.fxe"},
	{0x1E7D97B1D5EB897BULL, "861665825.ogg"},
	{0x1E7DC3A515C8E2F0ULL, "497095288.ogg"},
	{0x1E7EC2C0F78683B2ULL, "341785082.ogg"},
	{0x1E7ECAE5BCC2E5C5ULL, "hf_zmother_h1s1.mor"},
	{0x1E7F1B3BED6F2F0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_halfthinorn.mmh"},
	{0x1E7F1F3BED6F365EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_halfthinorn.mao"},
	{0x1E7F2D3BED6F4D93ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_halfthinorn.msh"},
	{0x1E7F4A2C9567D7F0ULL, "qcr220cr_carver_templar.utc"},
	{0x1E7F71D247B1E4D6ULL, "dae000_madam_lusine6154910.cl"},
	{0x1E7FB595D7AB9442ULL, "hlac_lt_undercityserialkiller_-1_46x.dds"},
	{0x1E809B89BF286AF7ULL, "409904845.ogg"},
	{0x1E80A4ECB0A74CE3ULL, "cod_bks_lite_tombs_2.utp"},
	{0x1E80C30128F13906ULL, "vxlt_0_lowtown_docks_n_c6t_0_0.vlm"},
	{0x1E80EB3070CCF281ULL, "ht_playershouse_n_855_ll.mmh"},
	{0x1E814F6906FAF077ULL, "6072370_m.fxe"},
	{0x1E8155ED0F16871BULL, "ts_demonic_acc_props.gda"},
	{0x1E81C528E1B93953ULL, "gen00fl_anders6111086.cl"},
	{0x1E81EA9C7C2B2598ULL, "hlac_ht_mansion2b_818_42z.dds"},
	{0x1E8242824F577158ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_boilerbase_n.dds"},
	{0x1E824BC7CBE90C4EULL, "222650443.ogg"},
	{0x1E82DA16F92E04E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_en.ani"},
	{0x1E82F3672073D89AULL, "mag121_viveka6082583.cl"},
	{0x1E83374693FFF04CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a_2.mao"},
	{0x1E8338635C2E00FDULL, "lt_0_lowtown_docks_237_0.mao"},
	{0x1E833C635C2E0736ULL, "lt_0_lowtown_docks_237_0.mmh"},
	{0x1E8346635C2E1834ULL, "lt_0_lowtown_docks_237_0.msh"},
	{0x1E834B04A0B566F1ULL, "6151260_m.fxe"},
	{0x1E835F5F8C01AEA0ULL, "953142277.ogg"},
	{0x1E836B234C0538D1ULL, "llac_lt_0_lowtown_a2_n_c_-3_1.dds"},
	{0x1E8383F13B0729CFULL, "655634626.ogg"},
	{0x1E839DBC2AE04C7BULL, "102920395.ogg"},
	{0x1E83B03A65BA6261ULL, "llac_ht_0_hightown_a3_-65536_3.dds"},
	{0x1E83D694E3013381ULL, "vxw_encounter2_no_0_0.vlm"},
	{0x1E8422A8768E1BE1ULL, "86602852.ogg"},
	{0x1E843826D32AED61ULL, "6044931_m.fxe"},
	{0x1E8492A2967DD44FULL, "hlac_lt_0_lowtown_-2_3x.dds"},
	{0x1E84D75E03EE7944ULL, "6149488_m.fxe"},
	{0x1E85AEC712EAE8E1ULL, "pro000cs_voyageend.ncs"},
	{0x1E85D2A6C602CBAFULL, "hlac_lt_0_lowtown_a2_n_131071_6x.dds"},
	{0x1E862801673B17EFULL, "art\\vfx\\combat\\vfx_w_gn1_build_up_c.anb"},
	{0x1E8659CADABE161BULL, "6113813_m.fxe"},
	{0x1E874611731BF38DULL, "30040859.ogg"},
	{0x1E877815022E8742ULL, "815292896.ogg"},
	{0x1E87A228C36F2060ULL, "zz_one_debug6077966.cl"},
	{0x1E87D86AEB6C313DULL, "922019701.ogg"},
	{0x1E880353083BD780ULL, "1028269475.ogg"},
	{0x1E88BA85FB0A5D5FULL, "dre000_varric6207935.cl"},
	{0x1E8931070B9A6F6DULL, "6144073_m.fxe"},
	{0x1E893F0802D0DF4BULL, "6118077_f.fxe"},
	{0x1E8A512AD25740EFULL, "mag161ar_cave.are"},
	{0x1E8B4BD7057499BDULL, "32008440.ogg"},
	{0x1E8B4F77BC374F4BULL, "mrl351_phantomelves6170336.cl"},
	{0x1E8B9EDBDD4328F4ULL, "vxlt_gallowstemplar_f_5t8_0_0.vlm"},
	{0x1E8BEC7A8F57D4A9ULL, "vxlt_0_lowtown_gob_0_0.vlm"},
	{0x1E8C1AE61765FD76ULL, "llac_w_bonepit_458755_0.dds"},
	{0x1E8C485FB63209A4ULL, "243047718.ogg"},
	{0x1E8C56B09A5DCEFFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha05a.phy"},
	{0x1E8CDDC251D32B3BULL, "hlac_ca_caverns1_24421_10z.dds"},
	{0x1E8D131B9024C7DAULL, "w_sundermount_61_0.msh"},
	{0x1E8D211B9024DF13ULL, "w_sundermount_61_0.mao"},
	{0x1E8D251B9024E620ULL, "w_sundermount_61_0.mmh"},
	{0x1E8D982617C1DD25ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax105a_0d.dds"},
	{0x1E8DA901299FC678ULL, "pro000_merchant6051554.cl"},
	{0x1E8DF1F51075B8C5ULL, "vxht_0_hightown_aif_0_0.vlm"},
	{0x1E8E02BD11EC47B6ULL, "307318330.ogg"},
	{0x1E8E1D287C6AB5F4ULL, "309914875.ogg"},
	{0x1E8E3C354E73FA3EULL, "vxht_keep_63x_0_0.vlm"},
	{0x1E8E43571FF2363AULL, "lt_hangedman_3567.anb"},
	{0x1E8E6FC875A6BF98ULL, "104047277.ogg"},
	{0x1E8EB06A5E58CA1FULL, "vxlt_gallowstemplar_351_0_0.vlm"},
	{0x1E8F0373C1F205C0ULL, "art\\vfx\\combat\\vfx_glb_foot_dust_c.mmh"},
	{0x1E8F34DE1AE75186ULL, "vxlt_0_lowtown_docks_n_c1u_0_0.vlm"},
	{0x1E8F5ED5FDF3470FULL, "6119275_m.fxe"},
	{0x1E8F70ACBE08DC8DULL, "vxlt_darktown_5ay_0_0.vlm"},
	{0x1E900E7BD0EBEFF2ULL, "ser120tr_ghyslain_ambient.utt"},
	{0x1E90143BED7D96CAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_halfthinorn.phy"},
	{0x1E901A002120CB25ULL, "vxw_encounter2_dq_0_0.vlm"},
	{0x1E9030AF3989A01CULL, "rdr111dg_martin6015064.cl"},
	{0x1E904B503515D9C4ULL, "287442447.ogg"},
	{0x1E90A3C68DC44B5CULL, "dre000_varric6208067.cl"},
	{0x1E916C78E6930CD2ULL, "6018847_m.fxe"},
	{0x1E918073A3E5F20EULL, "ss_qunari_m_2.bnk"},
	{0x1E91C0D4068C9337ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basic.phy"},
	{0x1E91F2D7217340B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_drinking.anb"},
	{0x1E9230A1D4AEAB7AULL, "ser211cr_demon_leandra.utc"},
	{0x1E92351AD2139789ULL, "6174445_m.fxe"},
	{0x1E923EF8F33D660EULL, "1047993995.ogg"},
	{0x1E924871D4F74C01ULL, "vxw_encounter2_in_0_0.vlm"},
	{0x1E92B47B105C50A1ULL, "608722518.ogg"},
	{0x1E92E2B8D41470B2ULL, "vxlt_hangedman_n_th_0_0.vlm"},
	{0x1E93C945E3DD7962ULL, "871161116.ogg"},
	{0x1E93EF9D387414F0ULL, "6105875_m.fxe"},
	{0x1E94224A4CC0E603ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killervase.msh"},
	{0x1E94304A4CC0FE5DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killervase.mmh"},
	{0x1E94344A4CC1056EULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killervase.mao"},
	{0x1E945DB5B66B2F94ULL, "llac_lt_0_lowtown_a2_65535_5.dds"},
	{0x1E94625D629E65FCULL, "117983612.ogg"},
	{0x1E9481EE59DB9C18ULL, "vxca_caverns1_mines_j5a_0_0.vlm"},
	{0x1E9492EBC09DAB79ULL, "6060361_f.fxe"},
	{0x1E956A197FD664D1ULL, "llac_lt_0_lowtown_n_-131072_0.dds"},
	{0x1E96E21A157720B7ULL, "zz_seb_debug6183091.cl"},
	{0x1E97263A965CB542ULL, "gen00fl_fenris6108074.cl"},
	{0x1E97DDC2A67CB5EEULL, "6174983_m.fxe"},
	{0x1E98634EA6E54357ULL, "vxdr_0_deeproads_cuh_0_1.vlm"},
	{0x1E986D9F8B181843ULL, "843667034.ogg"},
	{0x1E995F2A2AEA7C5AULL, "6047252_m.fxe"},
	{0x1E99E8038DF47CDBULL, "492310868.ogg"},
	{0x1E99FEA232185BC1ULL, "vxht_0_hightown_c8w_0_0.vlm"},
	{0x1E9A5E57814DED29ULL, "vxlt_0_lowtown_gja_0_0.vlm"},
	{0x1E9A861D5B109EE0ULL, "151979839.ogg"},
	{0x1E9B4A2AD265A7ACULL, "mag161ar_cave.lst"},
	{0x1E9B5E161D2C28D6ULL, "6124403_m.fxe"},
	{0x1E9B706BCBD5E18DULL, "follower_barks6153503.cl"},
	{0x1E9BA540C758699BULL, "dae011_gossips6149772.cl"},
	{0x1E9BDFBA285EEC46ULL, "plt_mag120pt_enemies.plo"},
	{0x1E9CE673C1FE11DFULL, "art\\vfx\\combat\\vfx_glb_foot_dust_c.anb"},
	{0x1E9D57B09A6C4304ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha05a.mmh"},
	{0x1E9D6DB09A6C68A6ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha05a.msh"},
	{0x1E9DE289AFFEF50DULL, "vxlt_darktown_5jx_0_0.vlm"},
	{0x1E9E01BDB7991525ULL, "vxht_0_hightown_a2_au6_0_0.vlm"},
	{0x1E9E37C22CB719A2ULL, "vxca_caverns2_8pt_0_0.vlm"},
	{0x1E9E3C92EBDBA702ULL, "vxlt_0_lowtown_docks_a3_d1b_0_0.vlm"},
	{0x1E9E443393504A20ULL, "vxw_woundedcoast_mz_0_0.vlm"},
	{0x1E9E8220E1EC700BULL, "hlac_lt_0_lowtown_a2_65536_7z.dds"},
	{0x1E9EB24CA99E8024ULL, "392195986.ogg"},
	{0x1E9FF57E8ED53D5CULL, "106695722.ogg"},
	{0x1EA0943B6518A838ULL, "follower_banter6136685.cl"},
	{0x1EA0F3DBD5FAA44BULL, "6201485_f.fxe"},
	{0x1EA1092B4A3A7B8AULL, "vxlt_0_lowtown_ex6_0_0.vlm"},
	{0x1EA1301633208969ULL, "mrl350ip_grave_b.utp"},
	{0x1EA15DBFF45382FCULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.cess.ani"},
	{0x1EA1705673ED7569ULL, "6166287_f.fxe"},
	{0x1EA2258070472C22ULL, "534103039.ogg"},
	{0x1EA22A2AD26B9055ULL, "mag161ar_cave.ncs"},
	{0x1EA232F40E8FF231ULL, "6077111_m.fxe"},
	{0x1EA26600CA92C463ULL, "gen00fl_merrill6127513.cl"},
	{0x1EA268CD9397909DULL, "vxht_mansion2b_32i_0_0.vlm"},
	{0x1EA3030C03EEA3C7ULL, "6189397_m.fxe"},
	{0x1EA332327A11F0ACULL, "vxw_encounter1_ej_0_0.vlm"},
	{0x1EA3A80554EA4C53ULL, "gen00fl_sebastian6174705.cl"},
	{0x1EA3D6202105C067ULL, "vxw_woundedcoast_1f1_0_0.vlm"},
	{0x1EA3E1641E1F0B3FULL, "592387458.ogg"},
	{0x1EA3EAC6C962B6CCULL, "vxca_caverns1_e8x_0_0.vlm"},
	{0x1EA3F01041E12C22ULL, "vxlt_gallowstemplar_2e5_0_0.vlm"},
	{0x1EA3F19C3E52824AULL, "hlac_w_sundermount_-131067_0z.dds"},
	{0x1EA42350BA7DB271ULL, "dre000_stroud6036719.cl"},
	{0x1EA4670F1DDEB58CULL, "zz_followers_banter6158508.cl"},
	{0x1EA4F2292EF6396AULL, "dre000_cut2d_ragstoriches6025082.cl"},
	{0x1EA5BF21A5C2A37BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_1.evt"},
	{0x1EA6853B3F15B4F5ULL, "vxlt_warehouse_n_ax_0_0.vlm"},
	{0x1EA6957D8F561035ULL, "llac_ca_caverns1_24426_11.dds"},
	{0x1EA6B43649075845ULL, "335672698.ogg"},
	{0x1EA72D15CCBB192BULL, "662836130.ogg"},
	{0x1EA746A93E1ED409ULL, "6090872_f.fxe"},
	{0x1EA8158C00F4A5D8ULL, "sse_ss_cityelf_m_1.gda"},
	{0x1EA8310DEE182A84ULL, "docks_male_overseer_mob6176621.cl"},
	{0x1EA83888DB2AF670ULL, "dae000_orsino6085488.cl"},
	{0x1EA86ABFF459B73EULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.cess.gad"},
	{0x1EA933CFF578219CULL, "cd_dre000_vampire.bnk"},
	{0x1EA93583B86F38A9ULL, "739233231.ogg"},
	{0x1EA9AB3685590556ULL, "99066622.ogg"},
	{0x1EAA481955E0C425ULL, "216046488.ogg"},
	{0x1EAA4CD8F632D4DAULL, "vxht_0_hightown_a3_cbj_0_1.vlm"},
	{0x1EAA94B0CD878A36ULL, "gen00fl_varric6093590.cl"},
	{0x1EAB1FEF32602B3EULL, "vxht_keep_69z_0_0.vlm"},
	{0x1EAC28D343EBFCD5ULL, "170913525.ogg"},
	{0x1EAC5466A1F50D8DULL, "vxlt_darktown_5ow_0_0.vlm"},
	{0x1EAC6AB13D07C56DULL, "zz_followers_banter6159009.cl"},
	{0x1EAC739AA98F2DA5ULL, "vxht_0_hightown_a2_ap5_0_0.vlm"},
	{0x1EACB821A5C8B5BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_1.gad"},
	{0x1EAD39E26D6E0993ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_exit.ani"},
	{0x1EAD5BE16FCBFB34ULL, "mrl150_merrill6093964.cl"},
	{0x1EAD9FC548DAEE52ULL, "vxw_bonepitmining_a2_1jy_0_0.vlm"},
	{0x1EADAA6B91259B6EULL, "306557673.ogg"},
	{0x1EAE1AEAB4A6EDF6ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiresml.dura.ani"},
	{0x1EAE89FD554FBBD2ULL, "dr_primevaltomb_664_ll.mmh"},
	{0x1EAF45BFF45F975DULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.cess.evt"},
	{0x1EAF657E5D852A89ULL, "lt_0_lowtown_a2_34_0.rml"},
	{0x1EAF7B083C30940AULL, "vxlt_0_lowtown_es7_0_0.vlm"},
	{0x1EAFC1057037BA3EULL, "460881186.ogg"},
	{0x1EB182876059BF34ULL, "hlac_lt_0_lowtown_docks_a3_-4_7y.dds"},
	{0x1EB1A40F6C08092CULL, "vxw_encounter1_jk_0_0.vlm"},
	{0x1EB1B580E2465EABULL, "6072136_m.fxe"},
	{0x1EB1BE60EBD6089DULL, "464578663.ogg"},
	{0x1EB1E6771068E883ULL, "6097835_f.fxe"},
	{0x1EB22CF702E87961ULL, "476668587.ogg"},
	{0x1EB247FD12FBD8E7ULL, "vxw_woundedcoast_1a0_0_0.vlm"},
	{0x1EB2BFC7B5DAC47BULL, "6106891_m.fxe"},
	{0x1EB3B721A5CED2BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_1.ani"},
	{0x1EB448B97CCBFF12ULL, "6074459_f.fxe"},
	{0x1EB44DB31F11011FULL, "vxca_caverns2_91s_0_0.vlm"},
	{0x1EB460B26AC74D5BULL, "555754835.ogg"},
	{0x1EB4BB89C647594EULL, "vrc321_party_barks6095743.cl"},
	{0x1EB58AB7A020D8D4ULL, "6178493_m.fxe"},
	{0x1EB5FC16122C48ECULL, "vxlt_0_lowtown_n_era_0_0.vlm"},
	{0x1EB68B68C4C870A4ULL, "198481980.ogg"},
	{0x1EB6CC4702956F6FULL, "6177723_m.fxe"},
	{0x1EB6E25C16048CC1ULL, "vxht_0_hightown_c2u_0_0.vlm"},
	{0x1EB76A3ED94777A9ULL, "llac_ht_playershouse_n_855_13.dds"},
	{0x1EB796778DF4838EULL, "lt_0_lowtown_docks_n_26_0.anb"},
	{0x1EB860641F6D97D3ULL, "qun211cr_commoner2.utc"},
	{0x1EB869E78B41043EULL, "gen00fl_anders6125693.cl"},
	{0x1EB89F20CD0E546EULL, "122400490.ogg"},
	{0x1EB95F0563D64FC8ULL, "dae000ip_crowd_ff7_talk8.utp"},
	{0x1EB9D2AF09149A7CULL, "vxlt_0_lowtown_a3_n_c_eqc_0_0.vlm"},
	{0x1EBA4B76900894A4ULL, "vxlt_0_lowtown_n_el9_0_2.vlm"},
	{0x1EBAC64393EB260DULL, "vxlt_darktown_5pv_0_0.vlm"},
	{0x1EBAFF1EC4B9E2DDULL, "98392136.ogg"},
	{0x1EBB0007102F440AULL, "vxlt_0_lowtown_h1x_0_0.vlm"},
	{0x1EBB22EAB4B185E7ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiresml.dura.evt"},
	{0x1EBB27ED773C7B20ULL, "vxw_woundedcoast_gx_0_0.vlm"},
	{0x1EBB31E26D7A39CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_exit.evt"},
	{0x1EBBBC17F4D3ACD2ULL, "dre000_vampire6034292.cl"},
	{0x1EBBF776DFDEF6C9ULL, "gen_im_trs_gem_cub_zirc_02.uti"},
	{0x1EBC41AEF100DD98ULL, "follower_banter6163984.cl"},
	{0x1EBCDEE8551E65A0ULL, "6067840_f.fxe"},
	{0x1EBD0BD3363940C7ULL, "fen310_fenris6118664.cl"},
	{0x1EBDDFD9D3733082ULL, "vxlt_darktown_2yu_0_0.vlm"},
	{0x1EBDECE52E26AC8AULL, "vxlt_0_lowtown_en0_0_0.vlm"},
	{0x1EBFA10A5C72BE4BULL, "llac_w_woundedcoast_196608_0.dds"},
	{0x1EBFEB19C790C8C4ULL, "art\\levels\\races\\design\\prp_rubble.msh"},
	{0x1EBFFD19C790E72DULL, "art\\levels\\races\\design\\prp_rubble.mao"},
	{0x1EC00119C790EE66ULL, "art\\levels\\races\\design\\prp_rubble.mmh"},
	{0x1EC045E33601D345ULL, "vxdr_primevaltomb_7yl_0_1.vlm"},
	{0x1EC06B27396E9B3EULL, "6076934_f.fxe"},
	{0x1EC0B9DA04F1F167ULL, "vxw_woundedcoast_1l3_0_0.vlm"},
	{0x1EC0BC47D2DC3D23ULL, "rdr111dg_martin6015182.cl"},
	{0x1EC0F2DC963201F5ULL, "6170197_m.fxe"},
	{0x1EC138292C7364E5ULL, "art\\vfx\\combat\\vfx_g_spider_mode_c.anb"},
	{0x1EC14657201D9566ULL, "lt_hangedman_3567.rml"},
	{0x1EC16483A5EF6646ULL, "vxht_0_hightown_a3_ato_0_0.vlm"},
	{0x1EC1A56D6CC3839FULL, "lt_0_lowtown_a2_18_0.rml"},
	{0x1EC1B445267FC389ULL, "144700193.ogg"},
	{0x1EC210E26D80203DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_exit.gad"},
	{0x1EC243EAB4B7DC30ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiresml.dura.gad"},
	{0x1EC2922B23B0D3B3ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.anb"},
	{0x1EC3102B3C65BE83ULL, "vxlt_warehouse_n_142_0_0.vlm"},
	{0x1EC31607B00420B9ULL, "vxlt_0_lowtown_a2_n_c_fhg_0_0.vlm"},
	{0x1EC31AAAE44EF746ULL, "vxht_0_hightown_caw_0_0.vlm"},
	{0x1EC3A8272AEC2D26ULL, "633226989.ogg"},
	{0x1EC3B15411BECE90ULL, "644443230.ogg"},
	{0x1EC404F88AEC0C56ULL, "vxht_brothel_n_5hn_0_0.vlm"},
	{0x1EC463F3DA21D425ULL, "vxw_sundermount_33k_0_1.vlm"},
	{0x1EC4F28740DE26D8ULL, "vxw_encounter2_camp_pl_0_0.vlm"},
	{0x1EC4FCCCFACD7C34ULL, "llac_lt_gallowstemplar_2383_4.dds"},
	{0x1EC53BCDA419B483ULL, "6032011_m.fxe"},
	{0x1EC575A3CEF324F5ULL, "6163531_m.fxe"},
	{0x1EC5B924D21622FFULL, "fen110_fenris6101963.cl"},
	{0x1EC627252560F6E8ULL, "vxlt_undercityserialkiller_145_0_0.vlm"},
	{0x1EC6857C020B6EC0ULL, "509164306.ogg"},
	{0x1EC707AE8E93D8ABULL, "hlac_lt_0_lowtown_a2_n_c_65538_1y.dds"},
	{0x1EC7522BFA176C1BULL, "759133313.ogg"},
	{0x1EC86D696D97BD85ULL, "hlac_lt_warehouse_n_-65536_1z.dds"},
	{0x1EC9211334C2F4DFULL, "590161340.ogg"},
	{0x1EC924C3815E2774ULL, "art\\levels\\races\\qunari\\props\\military\\prp_quanarithron.phy"},
	{0x1EC9927079D2DF63ULL, "370870110.ogg"},
	{0x1EC9940B10208E92ULL, "6043502_m.fxe"},
	{0x1EC9F819C7993331ULL, "art\\levels\\races\\design\\prp_rubble.phy"},
	{0x1ECA794F6254F46BULL, "gen00fl_dog6165772.cl"},
	{0x1ECAB2D95023F287ULL, "589507571.ogg"},
	{0x1ECB7132DE7D0726ULL, "vxca_caverns2_6hf_0_0.vlm"},
	{0x1ECB903460D37859ULL, "ave310cr_bandit_m.utc"},
	{0x1ECD0D881261AA7AULL, "6084217_m.fxe"},
	{0x1ECD2E5D61D2A06AULL, "zz_approval_debug6003933.cl"},
	{0x1ECDB77DD1CC3EA9ULL, "6179767_m.fxe"},
	{0x1ECE0B27F1C22194ULL, "lt_gallowstemplar_2389.rml"},
	{0x1ECE2DEA08835EF0ULL, "6145134_m.fxe"},
	{0x1ECE5AEB366FF48AULL, "vxht_chantry_n_2ya_0_0.vlm"},
	{0x1ECEAB209E40008FULL, "vxca_caverns1_ge2_0_0.vlm"},
	{0x1ECEFFE5BFB045C6ULL, "one200_gold_followers6120152.cl"},
	{0x1ECF12B35AA86374ULL, "6108511_m.fxe"},
	{0x1ECF1ED9164DDD56ULL, "vxlt_0_lowtown_docks_b4x_0_0.vlm"},
	{0x1ECF32BB24C1B80AULL, "vxca_cavern1_haunted_ehs_0_0.vlm"},
	{0x1ECF371651FB9C15ULL, "vxht_bartrand_n_6qm_0_0.vlm"},
	{0x1ECF74A65D689953ULL, "vxw_bonepitmining_a2_1ya_0_0.vlm"},
	{0x1ECFAF2B23BB8F1CULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.mmh"},
	{0x1ECFD66097E57EC6ULL, "vxht_0_hightown_a3_aqp_0_0.vlm"},
	{0x1ED0725653A1727DULL, "zz_dae_debug6029490.cl"},
	{0x1ED0D01218D2ED95ULL, "41921826.ogg"},
	{0x1ED16C9B93B12337ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_mag_0s.dds"},
	{0x1ED182082E5BD703ULL, "vxlt_warehouse_n_195_0_0.vlm"},
	{0x1ED18B3DBED940EBULL, "6103640_f.fxe"},
	{0x1ED1BB8E47F4459CULL, "308427035.ogg"},
	{0x1ED1F4EE9852EE18ULL, "follower_banter6141517.cl"},
	{0x1ED256DB58D642DCULL, "6013383_m.fxe"},
	{0x1ED29B7DDAF12237ULL, "6117779_f.fxe"},
	{0x1ED2F681A988190FULL, "268671805.ogg"},
	{0x1ED301C381664071ULL, "art\\levels\\races\\qunari\\props\\military\\prp_quanarithron.msh"},
	{0x1ED30FC38166578CULL, "art\\levels\\races\\qunari\\props\\military\\prp_quanarithron.mao"},
	{0x1ED313F6B95C0BD9ULL, "magma.dds"},
	{0x1ED31BC381666C2FULL, "art\\levels\\races\\qunari\\props\\military\\prp_quanarithron.mmh"},
	{0x1ED348D2BF07FF35ULL, "w_bonepitmining_a2_36_0_lc.mmh"},
	{0x1ED3646432D43F58ULL, "vxw_encounter2_camp_km_0_0.vlm"},
	{0x1ED36C5675FA17BFULL, "llac_ht_keep_4519_12.dds"},
	{0x1ED392015F850B64ULL, "gen00fl_fenris6126766.cl"},
	{0x1ED3B65EA644FEE8ULL, "al_rogue_t04.gda"},
	{0x1ED3D7127F94681CULL, "6053270_f.fxe"},
	{0x1ED40BE34554FFFBULL, "23534714.ogg"},
	{0x1ED422F0AD75F0CBULL, "vxca_caverns1_hma_0_0.vlm"},
	{0x1ED4231C7E3BAC22ULL, "prp_hmbdynud.mao"},
	{0x1ED4B9650EF72424ULL, "hlac_w_blightlands_327676_0x.dds"},
	{0x1ED52B8D5D27A739ULL, "gen00fl_merrill6108255.cl"},
	{0x1ED551C6DB02D0BBULL, "6103676_m.fxe"},
	{0x1ED5AB0E26E38B55ULL, "vxlt_foundry_n_yz_0_0.vlm"},
	{0x1ED5E69A2B2D2420ULL, "t1_skn_h08.tnt"},
	{0x1ED5EA6554611D24ULL, "646948942.ogg"},
	{0x1ED6986541A35FACULL, "6177885_m.fxe"},
	{0x1ED6A876A2410A5BULL, "vxdr_primevaltomb_8pz_0_1.vlm"},
	{0x1ED754A2419A676DULL, "lt_0_lowtown_docks_a3_178_0.mmh"},
	{0x1ED754D6801E92E6ULL, "art\\characters\\playercharacter\\qunarimale\\uni_com.phy"},
	{0x1ED758A2419A6DBEULL, "lt_0_lowtown_docks_a3_178_0.mao"},
	{0x1ED766A2419A85F3ULL, "lt_0_lowtown_docks_a3_178_0.msh"},
	{0x1ED7857F8FBDA5BAULL, "6140283_m.fxe"},
	{0x1ED804BE0EA1E66AULL, "6096752_f.fxe"},
	{0x1ED80BA75B28AC20ULL, "vxw_woundedcoast_yf_0_0.vlm"},
	{0x1ED821EBA35EC4B7ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.dura.gad"},
	{0x1ED8671F420E210BULL, "darktown_fem_lowlife_mob6176165.cl"},
	{0x1ED895D406C84C8FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basic.mao"},
	{0x1ED8A1D406C8612CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basic.mmh"},
	{0x1ED8A7D406C86B1EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3ww_basic.msh"},
	{0x1ED8C8C90CEEA0CBULL, "vxlt_0_lowtown_docks_bvk_0_0.vlm"},
	{0x1ED8CB6D61A5136AULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredidies.gad"},
	{0x1ED91AC9FAA56B2CULL, "637090243.ogg"},
	{0x1ED95B566D95F56EULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_f_c.impa.ani"},
	{0x1ED96729C5DB1CC9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowon.mao"},
	{0x1ED97329C5DB3162ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowon.mmh"},
	{0x1ED97529C5DB3418ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowon.msh"},
	{0x1ED9D1135EF4C140ULL, "6025364_m.fxe"},
	{0x1ED9EDAEAB8AC8DDULL, "llac_ht_keep_n_3732_11.dds"},
	{0x1ED9F191963DBC7AULL, "lt_0_lowtown_a2_ll.mmh"},
	{0x1ED9F5405703490DULL, "gen00fl_merrill6120972.cl"},
	{0x1EDABD275EA5786AULL, "vxlt_backalley_n_1z4_0_0.vlm"},
	{0x1EDB219CD531F070ULL, "929868251.ogg"},
	{0x1EDC1FF0B9C6BEFDULL, "lt_0_lowtown_docks_a3_29_0.anb"},
	{0x1EDC6D88DE252E8BULL, "hlac_lt_0_lowtown_n_0_9z.dds"},
	{0x1EDC6E57599B1496ULL, "mer320_hubert6090373.cl"},
	{0x1EDCC698420EBBA6ULL, "vxht_0_hightown_a73_0_0.vlm"},
	{0x1EDCE33A20056FD8ULL, "378075947.ogg"},
	{0x1EDCE87081B48FE1ULL, "6105610_m.fxe"},
	{0x1EDCFF43CE70A6DDULL, "one220_yevhen6098785.cl"},
	{0x1EDD3217F1AFCF82ULL, "vxht_keep_n_60w_0_1.vlm"},
	{0x1EDD38F65097CD4DULL, "669922692.ogg"},
	{0x1EDD6BB8060BE6BBULL, "llac_ht_chantry_2534_13.dds"},
	{0x1EDE0AE0695C1810ULL, "hlac_ht_0_hightown_n_-131074_1z.dds"},
	{0x1EDE29CED306E6C1ULL, "t1_mub_ro2.tnt"},
	{0x1EDE36EBA3635444ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.dura.evt"},
	{0x1EDE5B8EDB6F155FULL, "fxe_snow_tiny_bld.mao"},
	{0x1EDE85C665D496EAULL, "6163640_m.fxe"},
	{0x1EDEC23A30C8CD96ULL, "vxlt_0_lowtown_docks_n_b7n_0_0.vlm"},
	{0x1EDEDA156876BF84ULL, "269958623.ogg"},
	{0x1EDF3E47A66E6564ULL, "one300_murder_followers6089905.cl"},
	{0x1EDF5398A6527541ULL, "hlac_w_bonepitmining_a2_0_3x.dds"},
	{0x1EDFFB273F36EBC3ULL, "6109681_m.fxe"},
	{0x1EE0760C9ECA6DCAULL, "zz_arishock.cut"},
	{0x1EE10D8BC9289E19ULL, "dae000_arishok6078341.cl"},
	{0x1EE112C2E99B8356ULL, "283485996.ogg"},
	{0x1EE1233984EFFEFAULL, "hlac_lt_0_lowtown_a2_-196608_2x.dds"},
	{0x1EE16932FBF6A9DFULL, "plt_cod_cha_isabela1.plo"},
	{0x1EE183F71424A29DULL, "vxlt_0_lowtown_docks_a3_n_cws_0_0.vlm"},
	{0x1EE1D64124CA57D8ULL, "vxw_encounter2_camp_fj_0_0.vlm"},
	{0x1EE260A54C681AC7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v1.gad"},
	{0x1EE2647E5DB082A9ULL, "lt_0_lowtown_a2_34_0.anb"},
	{0x1EE28DC4CA1E0A3EULL, "t1_skn_e03.tnt"},
	{0x1EE2FAEFDD16D92BULL, "219314964.ogg"},
	{0x1EE32F47D475F359ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_nud_0t.dds"},
	{0x1EE34FB034828B97ULL, "vxlt_0_lowtown_docks_bbs_0_0.vlm"},
	{0x1EE37DEFBECDBF24ULL, "vxlt_gallowstemplar_n_35e_0_0.vlm"},
	{0x1EE3EDF390F5F6D6ULL, "one140cr_charade6166017.cl"},
	{0x1EE40AFD56E78E09ULL, "hlac_lt_backalley_n_65537_8x.dds"},
	{0x1EE41AF085FAEBC1ULL, "art\\vfx\\textures\\vfx_blood_splatter_d.dds"},
	{0x1EE488AE0BEE70DDULL, "322437664.ogg"},
	{0x1EE4D82D81147A80ULL, "6044004_m.fxe"},
	{0x1EE4DE1DAEF4A9C4ULL, "qun120_arishok6053229.cl"},
	{0x1EE4E6791FE865A4ULL, "gen00fl_anders6125127.cl"},
	{0x1EE521F788B0E6FBULL, "6109053_f.fxe"},
	{0x1EE54C6C78151F97ULL, "follower_banter6164742.cl"},
	{0x1EE5FF2E7068B191ULL, "follower_barks6153226.cl"},
	{0x1EE62D99C1C122EBULL, "6096263_f.fxe"},
	{0x1EE637BD2985B343ULL, "hlac_w_bonepitmining_a2_393222_0z.dds"},
	{0x1EE64856E17733BDULL, "6068545_f.fxe"},
	{0x1EE653566DA071BFULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_f_c.impa.evt"},
	{0x1EE67D844D1EC4A0ULL, "vxw_woundedcoast_vg_0_0.vlm"},
	{0x1EE6C0BA289EA285ULL, "plt_mag120pt_enemies.nss"},
	{0x1EE6C228CA7E0E7BULL, "6144981_m.fxe"},
	{0x1EE6F5EB36F0AA38ULL, "gen00fl_anders6119339.cl"},
	{0x1EE72ABD2A8F8DFBULL, "6148699_m.fxe"},
	{0x1EE73A359878333FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25_tool.phy"},
	{0x1EE7673910B35052ULL, "vxw_bonepitmining_a2_1ve_0_0.vlm"},
	{0x1EE81CE3602F67B2ULL, "vxht_mansion_n_5bg_0_0.vlm"},
	{0x1EE86B84B9F939EBULL, "and100_anders6095951.cl"},
	{0x1EE876033BA9EA44ULL, "vxw_encounter2_camp_nn_0_0.vlm"},
	{0x1EE892C306A0898AULL, "static_static_p76.psh.fxo.dx11"},
	{0x1EE893F9D2813BD8ULL, "zz_one_debug6099760.cl"},
	{0x1EE8AE352114FD81ULL, "6098062_m.fxe"},
	{0x1EE946FBAB001381ULL, "vxw_blightlands_7k_0_0.vlm"},
	{0x1EE955A54C6E26D4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v1.evt"},
	{0x1EE99AB021C0CA20ULL, "vxw_bonepit_1zq_0_0.vlm"},
	{0x1EE9ED2D0731DAE2ULL, "235510288.ogg"},
	{0x1EEA0668E43EE79EULL, "vxlt_backalley_g_1vq_0_0.vlm"},
	{0x1EEA39292C966B26ULL, "art\\vfx\\combat\\vfx_g_spider_mode_c.mmh"},
	{0x1EEA5FEECC54408BULL, "lgt260_way_blood_mage6148399.cl"},
	{0x1EEA99778E1FE2F2ULL, "lt_0_lowtown_docks_n_26_0.rml"},
	{0x1EEAA2B8FEEBFAF5ULL, "pro000ip_merchant_chest.utp"},
	{0x1EEAE0CD83A3CC82ULL, "follower_banter6161342.cl"},
	{0x1EEB96A4F9602954ULL, "ter_sand01_n.dds"},
	{0x1EEBA7E8D60CFBA4ULL, "plt_lgt000pt_lite_key_02.plo"},
	{0x1EEBC1743DA6EE4FULL, "hlac_lt_gallowstemplar_f_2390_1y.dds"},
	{0x1EEC07D59AC1EC8AULL, "199630787.ogg"},
	{0x1EEC2769D15592BEULL, "cntr_cl_f.mao"},
	{0x1EEC4EEBA36FBA3DULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.dura.ani"},
	{0x1EECE0A2E5633F17ULL, "hlac_lt_0_lowtown_docks_a3_65536_0y.dds"},
	{0x1EED34566DA65BA8ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_f_c.impa.gad"},
	{0x1EED3D939A3FC20FULL, "6088911_m.fxe"},
	{0x1EED44053A6FEAE8ULL, "hlac_lt_0_lowtown_a2_-3_6y.dds"},
	{0x1EEDA66D61B73050ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredidies.ani"},
	{0x1EEDFA515CA7890DULL, "6040001_m.fxe"},
	{0x1EEE2F6852CC4B70ULL, "hla_ht_0_hightown_a2_-65535_0y.dds"},
	{0x1EEEA41710FFEE3FULL, "animation_tuning.gda"},
	{0x1EEF2597089683BFULL, "vxlt_0_lowtown_n_eoi_0_0.vlm"},
	{0x1EEFEFE5B11131C0ULL, "plt_and200pt_dissent.nss"},
	{0x1EF0481E16C07058ULL, "vxw_encounter2_camp_ak_0_0.vlm"},
	{0x1EF04E94E8FAE2DFULL, "6002009_m.fxe"},
	{0x1EF05D65FB8D8A85ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\prp_ccliff02_sa.dds"},
	{0x1EF08C7A659B292CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright.ani"},
	{0x1EF08E698E5D4380ULL, "vxw_blightlands_el_0_0.vlm"},
	{0x1EF0C0733E0717E8ULL, "dae000_fem_commoner1_mob6174482.cl"},
	{0x1EF1EED2AC8BE14DULL, "545674080.ogg"},
	{0x1EF22A0BA92F9762ULL, "hlac_w_sundermount_-327686_0z.dds"},
	{0x1EF289585505A6BBULL, "gen00fl_anders6096453.cl"},
	{0x1EF36BCF995D301EULL, "6171210_m.fxe"},
	{0x1EF3D179C7624C47ULL, "116475054.ogg"},
	{0x1EF3D9B713402B3BULL, "art\\characters\\playercharacter\\humanmale\\las_viscount.mao"},
	{0x1EF3DCFD38BEC633ULL, "hla_ht_0_hightown_a3_2_0z.dds"},
	{0x1EF3F60662E3AB18ULL, "vxlt_warehouse_ah_0_0.vlm"},
	{0x1EF41EE3670F2872ULL, "vxlt_undercityencounter_1hg_0_0.vlm"},
	{0x1EF4658516B71DAFULL, "404897025.ogg"},
	{0x1EF4EF613F14DD20ULL, "vxw_woundedcoast_sd_0_0.vlm"},
	{0x1EF5306DBF77F361ULL, "6077793_m.fxe"},
	{0x1EF53C415C0ECAD4ULL, "128697650.ogg"},
	{0x1EF56F3B7040C92AULL, "6208673_m.fxe"},
	{0x1EF586A1424E858BULL, "gen00fl_sebastian6174860.cl"},
	{0x1EF58F3F6159C612ULL, "zz_ser_debug6078000.cl"},
	{0x1EF5B3C2D310BE67ULL, "6110922_m.fxe"},
	{0x1EF5CDAFFE64AE4BULL, "419692475.ogg"},
	{0x1EF5F114F3237E32ULL, "vxlt_0_lowtown_a3_n_c_d5t_0_0.vlm"},
	{0x1EF63FCB94189DCDULL, "gen00fl_sebastian6189422.cl"},
	{0x1EF64DA54C78A30DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v1.ani"},
	{0x1EF6E7E02DA002C4ULL, "vxw_encounter2_camp_io_0_0.vlm"},
	{0x1EF776C23C40557BULL, "6104804_m.fxe"},
	{0x1EF7997A65A15E6EULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright.gad"},
	{0x1EF7B8D89CF62C01ULL, "vxw_blightlands_8j_0_0.vlm"},
	{0x1EF7CD9CBDF795B9ULL, "dae000_mother6099195.cl"},
	{0x1EF829ED22B71AEEULL, "489346523.ogg"},
	{0x1EF83B359886A844ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25_tool.mmh"},
	{0x1EF83F359886AE97ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25_tool.mao"},
	{0x1EF851359886CDE6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25_tool.msh"},
	{0x1EF86587CEBC509BULL, "vxlt_warehouse_n_qv_0_0.vlm"},
	{0x1EF8BD9D1BABDE00ULL, "660505448.ogg"},
	{0x1EF9FB23627E54E1ULL, "qun211_braggart6075701.cl"},
	{0x1EFA160408A746B1ULL, "llac_lt_0_lowtown_a2_-196608_3.dds"},
	{0x1EFA360F213E2B84ULL, "vxht_brothel_n_3f4_0_0.vlm"},
	{0x1EFA94DB3C31F2F6ULL, "art\\characters\\playercharacter\\humanmale\\boo_urva.phy"},
	{0x1EFAB65C2DFC0508ULL, "art\\characters\\playercharacter\\elffemale\\animation\\blendtree\\ef_exp_movement.anb"},
	{0x1EFB00D14A3670CAULL, "159213546.ogg"},
	{0x1EFB0F6CAC2217BBULL, "382568340.ogg"},
	{0x1EFB25CDF3C4FE2DULL, "6101514_f.fxe"},
	{0x1EFB4DEA90E160DCULL, "6071578_m.fxe"},
	{0x1EFBAF708C35DA8FULL, "6083811_m.fxe"},
	{0x1EFBE2365DB5B582ULL, "dae000_fenarel6147428.cl"},
	{0x1EFC5456B5CB5055ULL, "gen00fl_carver6102332.cl"},
	{0x1EFD527563C89837ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_6x.dds"},
	{0x1EFDAD84BDE3E627ULL, "vxca_caverns2prc_9fk_0_0.vlm"},
	{0x1EFDCF1164FDF75EULL, "vxlt_house_iu_0_0.vlm"},
	{0x1EFE09DB3A1E445DULL, "character_character_p0.psh.fxo"},
	{0x1EFE34260560037BULL, "vxlt_gallowsprison_n_6k1_0_0.vlm"},
	{0x1EFE947A65A7744DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright.evt"},
	{0x1EFEE3226E9C5843ULL, "mcr370_negotiation6080227.cl"},
	{0x1EFF004680535C00ULL, "vxw_blightlands_nm_0_0.vlm"},
	{0x1EFF4F8450064C2BULL, "hlac_lt_backalley_n_1_1x.dds"},
	{0x1EFFF08352100F47ULL, "vxlt_0_lowtown_a3_n_c_fh8_0_0.vlm"},
	{0x1F002CD836BD5917ULL, "197372770.ogg"},
	{0x1F003BD36B82FED0ULL, "gen00fl_fenris6107208.cl"},
	{0x1F006D087663F576ULL, "632529290.ogg"},
	{0x1F007460A2A7C59EULL, "6129404_m.fxe"},
	{0x1F0095EDD961F8A5ULL, "6036335_f.fxe"},
	{0x1F011D659B90365FULL, "llac_dr_slavers_3216_3.dds"},
	{0x1F01714306C055B6ULL, "zz_isa_debug6133933.cl"},
	{0x1F01DBCA4B07137BULL, "258581688.ogg"},
	{0x1F01DDAC8B53489EULL, "6174110_m.fxe"},
	{0x1F02013942D9009FULL, "dr_0_deeproads_5840.rml"},
	{0x1F023F51EE098E1FULL, "vxlt_gallowstemplar_3m9_0_0.vlm"},
	{0x1F0241D7CBCF8EB2ULL, "vxlt_hangedman_1qg_0_0.vlm"},
	{0x1F02428DAD47EA69ULL, "vxlt_hangedman_n_108_0_0.vlm"},
	{0x1F028979EA1E75F7ULL, "mcr371_hostile_anders6082096.cl"},
	{0x1F031BC53A4B323BULL, "148014414.ogg"},
	{0x1F03C3EB5B715939ULL, "vxlt_backalley_g_1k1_0_0.vlm"},
	{0x1F03DE99DE1AD1F7ULL, "vxht_mansion_n_6ji_0_0.vlm"},
	{0x1F03FC43DB15C3A5ULL, "vxw_blightlands_6t_0_0.vlm"},
	{0x1F040E582B28C154ULL, "fen110ip_smuggler_chest.ncs"},
	{0x1F040FFC15C52C6DULL, "578092297.ogg"},
	{0x1F041E5FE2D0EA4BULL, "vxlt_0_lowtown_docks_bgl_0_0.vlm"},
	{0x1F0427911CA9D99AULL, "fen110_fenris6098895.cl"},
	{0x1F047DDB3C3A1F97ULL, "art\\characters\\playercharacter\\humanmale\\boo_urva.msh"},
	{0x1F0493DB3C3A4509ULL, "art\\characters\\playercharacter\\humanmale\\boo_urva.mmh"},
	{0x1F05009F37A367AAULL, "43694664.ogg"},
	{0x1F0549E736323CFCULL, "436659340.ogg"},
	{0x1F054FBCA0F47A3FULL, "sb03_perlin.dds"},
	{0x1F05A0B584BC507BULL, "6072266_m.fxe"},
	{0x1F05CFC9D2B8D98CULL, "hlac_ht_chantry_2534_25z.dds"},
	{0x1F064D5212E677D6ULL, "6041703_kerankill.cut"},
	{0x1F068F4C64D4D2DCULL, "801957381.ogg"},
	{0x1F06EA22C82B189EULL, "vxlt_backalley_g_1xs_0_0.vlm"},
	{0x1F07F24E7B988874ULL, "vxca_caverns1_j2c_0_2.vlm"},
	{0x1F0844BB83B20EFBULL, "51172401.ogg"},
	{0x1F087DCE42FDEC92ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\exit.gad"},
	{0x1F0A24B3D164D62CULL, "155442962.ogg"},
	{0x1F0A40C9E7EBFEBDULL, "vxca_cave_2jx_0_0.vlm"},
	{0x1F0AC3AEE967C70DULL, "679366059.ogg"},
	{0x1F0B0DEEAC9E4967ULL, "331345576.ogg"},
	{0x1F0B1F45EEA805DBULL, "zz_area_signage6.utp"},
	{0x1F0B8E331A252378ULL, "826766908.ogg"},
	{0x1F0BD19978DE76A3ULL, "619238983.ogg"},
	{0x1F0BF2FC0A926346ULL, "20792642.ogg"},
	{0x1F0C56F0E21F9318ULL, "6121969_m.fxe"},
	{0x1F0C65CCDCF993C0ULL, "docks_scene_dockworkers6176514.cl"},
	{0x1F0CF2DF5A55B136ULL, "6097266_m.fxe"},
	{0x1F0D722372497480ULL, "vxw_blightlands_sj_0_0.vlm"},
	{0x1F0D7DA8C28F7663ULL, "one300_zevran6089116.cl"},
	{0x1F0D87F679563A53ULL, "gen00fl_sebastian6178378.cl"},
	{0x1F0DAD3FDDD4F43AULL, "hlac_ca_caverns1_mines_24422_5y.dds"},
	{0x1F0DE2C3CE99E33DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_basic.phy"},
	{0x1F0DF50366317EC5ULL, "578800122.ogg"},
	{0x1F0E3D0C1F35D5DEULL, "hlac_lt_0_lowtown_a2_n_131070_3y.dds"},
	{0x1F0E549DAA85A767ULL, "472437613.ogg"},
	{0x1F0E736EA862B917ULL, "vxca_cavern1_haunted_i7c_0_0.vlm"},
	{0x1F0F0046545C6F96ULL, "856957923.ogg"},
	{0x1F0F09E628BCE3A4ULL, "276777899.ogg"},
	{0x1F0F32CA84A2058FULL, "vxlt_0_lowtown_docks_a3_d66_0_0.vlm"},
	{0x1F0F5468D4FF4A5BULL, "vxht_0_hightown_a2_buc_0_0.vlm"},
	{0x1F0F816DBC464EFBULL, "642249110.ogg"},
	{0x1F0FA9519684B019ULL, "hlac_lt_0_lowtown_0_5y.dds"},
	{0x1F0FAC6287D41744ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikib.msh"},
	{0x1F0FBE6287D435ADULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikib.mao"},
	{0x1F0FC26287D43CE6ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikib.mmh"},
	{0x1F0FD7195B32DDD0ULL, "491337012.ogg"},
	{0x1F0FDD59A217791DULL, "211236349.ogg"},
	{0x1F10D9C046CFDC18ULL, "vxlt_warehouse_of_0_0.vlm"},
	{0x1F1118EB5B785A0AULL, "dae200_bran6094894.cl"},
	{0x1F115F7DFC26CD43ULL, "zz_mcr_debug6071541.cl"},
	{0x1F117841C9822863ULL, "vxdr_primevaltomb_3i7_0_0.vlm"},
	{0x1F11842A81E5C081ULL, "vxht_brothel_n_6tr_0_0.vlm"},
	{0x1F119609AA644F60ULL, "llac_ca_caverns1_24423_6.dds"},
	{0x1F11AF3984ABFF75ULL, "6149108_m.fxe"},
	{0x1F11B616395C922EULL, "536773238.ogg"},
	{0x1F11F257884D5E0FULL, "vxlt_darktown_2r9_0_0.vlm"},
	{0x1F125076D010EA77ULL, "vxht_mansion_n_6mh_0_0.vlm"},
	{0x1F137379ADA4D8A2ULL, "6076390_f.fxe"},
	{0x1F13BE188EE0C350ULL, "character_character_p146.psh.fxo.dx11.assembly"},
	{0x1F141F39C28B7971ULL, "w_woundedcoast_248_0.mmh"},
	{0x1F142B39C28B8D12ULL, "w_woundedcoast_248_0.mao"},
	{0x1F143939C28BA52FULL, "w_woundedcoast_248_0.msh"},
	{0x1F147371CD1C9C40ULL, "6174963_m.fxe"},
	{0x1F14F18808C477E3ULL, "6069705_m.fxe"},
	{0x1F1520F0B9F685F5ULL, "lt_0_lowtown_docks_a3_29_0.rml"},
	{0x1F15A4C6C548A616ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_sw.phy"},
	{0x1F160DB151B91B50ULL, "6054640_f.fxe"},
	{0x1F16383C1AAD9C2FULL, "440403600.ogg"},
	{0x1F16CF89D825FEE3ULL, "ave121_brennan6105811.cl"},
	{0x1F191EB2F119FEA4ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm02.msh"},
	{0x1F1934B2F11A2446ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm02.mmh"},
	{0x1F199F8FA2F3BF61ULL, "6138751_m.fxe"},
	{0x1F19B96287DC81B1ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikib.phy"},
	{0x1F19E58259BD0C55ULL, "plt_cod_bks_fifth_blight.plo"},
	{0x1F1A7600BF296A9FULL, "vxlt_0_lowtown_ekg_0_1.vlm"},
	{0x1F1A7FD9C4B5B9C5ULL, "dre000_bethany6037325.cl"},
	{0x1F1B17DFE94C347BULL, "vxlt_gallowsprison_n_6q3_0_0.vlm"},
	{0x1F1BBDFE5D642952ULL, "dae000_bodhan6129089.cl"},
	{0x1F1BE400643F8D00ULL, "vxw_blightlands_tk_0_0.vlm"},
	{0x1F1BED582B3D1A85ULL, "fen110ip_smuggler_chest.utp"},
	{0x1F1C27A686F73448ULL, "art\\characters\\weapons\\spears\\w_spr_003a_p.msh"},
	{0x1F1C39A686F752D9ULL, "art\\characters\\weapons\\spears\\w_spr_003a_p.mao"},
	{0x1F1C45A686F76772ULL, "art\\characters\\weapons\\spears\\w_spr_003a_p.mmh"},
	{0x1F1C6608F72ECEC3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8ma.mao"},
	{0x1F1C6A08F72ED510ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8ma.mmh"},
	{0x1F1C7808F72EED2AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8ma.msh"},
	{0x1F1C7D41992F3D98ULL, "w_bonepitdestroyed_a3_41_0.rml"},
	{0x1F1CD03BE2FAE6AEULL, "51151616.ogg"},
	{0x1F1D064076814D00ULL, "vxw_sundermount_2gv_0_0.vlm"},
	{0x1F1D68CE43102538ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\exit.ani"},
	{0x1F1D6FAF7168CA7FULL, "788702366.ogg"},
	{0x1F1DE4921B153ACBULL, "672601245.ogg"},
	{0x1F1DE82A0DBC13EDULL, "503776270.ogg"},
	{0x1F1E055093001C07ULL, "6133778_m.fxe"},
	{0x1F1E0E08ECC7A9B0ULL, "vxht_0_hightown_a3_91n_0_0.vlm"},
	{0x1F1ED7C3CEA843A9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_basic.mao"},
	{0x1F1EE3C3CEA85842ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_basic.mmh"},
	{0x1F1EE5C3CEA85BF8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_basic.msh"},
	{0x1F1EE7B501B13EC4ULL, "6079336_m.fxe"},
	{0x1F1F1C31F1E5FD79ULL, "hlac_lt_gallowstemplar_n_7198_1x.dds"},
	{0x1F1F4B9D38C5F498ULL, "vxlt_warehouse_pg_0_0.vlm"},
	{0x1F1F8821FE601A33ULL, "vxca_caverns1_gkl_0_0.vlm"},
	{0x1F1F8DC6C550D3B7ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_sw.msh"},
	{0x1F1FA3C6C550F929ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_sw.mmh"},
	{0x1F1FEA1EBB7840E3ULL, "vxdr_primevaltomb_3b6_0_0.vlm"},
	{0x1F203FD72BEF25CCULL, "qcr000_cut2d_thechampion.cnv"},
	{0x1F2060B2E3189F1FULL, "amb_scene_violentattack.bnk"},
	{0x1F2064347A43768FULL, "vxlt_darktown_2u8_0_0.vlm"},
	{0x1F2089057A8D59FFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02a.phy"},
	{0x1F208D2275C69156ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a_2.mao"},
	{0x1F2189E388B2CC36ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.dura.evt"},
	{0x1F21DC3DCA45A562ULL, "889649532.ogg"},
	{0x1F221C6B34072C66ULL, "zz_qun_debug6028300.cl"},
	{0x1F2265D7B7A1AF94ULL, "501512422.ogg"},
	{0x1F22AE8E2B1A3148ULL, "447594962.ogg"},
	{0x1F2312486B7AD8B8ULL, "vxlt_0_lowtown_docks_a3_czk_0_1.vlm"},
	{0x1F231A84D7ED6DC4ULL, "6168166_m.fxe"},
	{0x1F2321D72BF128FCULL, "qcr000_cut2d_thechampion.bnk"},
	{0x1F23866D7C572FF3ULL, "llac_lt_0_lowtown_n_65536_4.dds"},
	{0x1F238ED87AD216D1ULL, "vxdr_0_deeproads_cx8_0_0.vlm"},
	{0x1F242BB2F1241C11ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm02.phy"},
	{0x1F243406A0474C7EULL, "pro000_athenril6049428.cl"},
	{0x1F2442E8C8CD6B6BULL, "6121012_m.fxe"},
	{0x1F24CFBD1A53AED1ULL, "tattoo_tints.gda"},
	{0x1F2514AECB12D9F4ULL, "pro000cr_refugee7.utc"},
	{0x1F25601D0C6F9FF2ULL, "6093662_m.fxe"},
	{0x1F258FF7AB5AE3B6ULL, "zz_one_debug6077388.cl"},
	{0x1F25F674D66D099FULL, "56155399.ogg"},
	{0x1F264A4624DBD8D2ULL, "zz_otl_debug.cnv"},
	{0x1F2662F50E066B63ULL, "em_mrl150cr_dalish_guard1.mor"},
	{0x1F267E0114EE7636ULL, "6027654_m.fxe"},
	{0x1F26B4BFF09D08C0ULL, "6149790_m.fxe"},
	{0x1F26EF2CF98C6AA7ULL, "hlac_ht_chantry_2534_53x.dds"},
	{0x1F26F40FE1CDFFCCULL, "6125486_m.fxe"},
	{0x1F27B113276096B6ULL, "hla_ht_0_hightown_a3_-131073_0y.dds"},
	{0x1F27CA145E7DFD4BULL, "vxht_keep_n_4jk_0_0.vlm"},
	{0x1F27F2AE1085356CULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_deda_0d.dds"},
	{0x1F280EBB1FC111E4ULL, "6074429_m.fxe"},
	{0x1F283D935049E2A8ULL, "mag291_vincento6027371.cl"},
	{0x1F285B15F8748C4AULL, "vxw_encounter2_camp_l_0_0.vlm"},
	{0x1F28672016741F83ULL, "vxlt_backalley_n_1nu_0_0.vlm"},
	{0x1F2880E388B8DBC1ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.dura.gad"},
	{0x1F28BCE8D640CDEBULL, "plt_lgt000pt_lite_key_02.nss"},
	{0x1F28E12556F40634ULL, "vxht_mansion2b_2ja_0_0.vlm"},
	{0x1F29C5DCD1D3BAB6ULL, "34813993.ogg"},
	{0x1F2A1E2EB6E7ADA5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portetm_d.dds"},
	{0x1F2A4040EBB59F3CULL, "6096847_m.fxe"},
	{0x1F2A918ADE2F197EULL, "406483333.ogg"},
	{0x1F2ADB9F3572E5B2ULL, "llac_w_sundermount_0_7.dds"},
	{0x1F2ADFAE8FBE4436ULL, "675052115.ogg"},
	{0x1F2B0883610CF154ULL, "6129756_m.fxe"},
	{0x1F2B38C4C18BBD8EULL, "vxlt_0_lowtown_a2_n_c_ht6_0_0.vlm"},
	{0x1F2B55C9A5770A8DULL, "993978927.ogg"},
	{0x1F2B781D68776580ULL, "vxw_sundermount_2hw_0_0.vlm"},
	{0x1F2C1A004127D218ULL, "vxlt_0_lowtown_a2_n_es2_0_0.vlm"},
	{0x1F2C1B7CC4646D22ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb01.msh"},
	{0x1F2C1D7CC46470D8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb01.mmh"},
	{0x1F2C7B858C502A30ULL, "406937649.ogg"},
	{0x1F2C99C102F59A6EULL, "qun111_ketojan6019013.cl"},
	{0x1F2CDEAD7851F1D8ULL, "6041484_f.fxe"},
	{0x1F2D9F67851F0312ULL, "vxlt_0_lowtown_docks_a3_n_cyf_0_0.vlm"},
	{0x1F2DBD7A2ABC0D18ULL, "vxlt_warehouse_ud_0_0.vlm"},
	{0x1F2DE8E5B1468683ULL, "plt_and200pt_dissent.plo"},
	{0x1F2E16932FD42529ULL, "706929309.ogg"},
	{0x1F2EACB9B6E49360ULL, "mrl150_dalish_guard6068744.cl"},
	{0x1F2F0CFD12E184DDULL, "dae000_bodhan_letter6144310.cl"},
	{0x1F2F81E388BEFC6FULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.dura.ani"},
	{0x1F2FBF6D9F0744D9ULL, "6140910_m.fxe"},
	{0x1F303FBA9C39F451ULL, "one220_yevhen6135790.cl"},
	{0x1F305B30CB003379ULL, "313153412.ogg"},
	{0x1F306630EB362B2DULL, "6048433_f.fxe"},
	{0x1F307DE393148BC3ULL, "6082210_m.fxe"},
	{0x1F3084C59BF835BAULL, "459570112.ogg"},
	{0x1F30968ABA47E92AULL, "vxht_0_hightown_a3_aob_0_0.vlm"},
	{0x1F30E94B68B0D3CEULL, "vxlt_darktown_2rt_0_0.vlm"},
	{0x1F3129D95EAB775AULL, "hlac_lt_undercityencounter_-1_25y.dds"},
	{0x1F314DCADBBBB1A0ULL, "683474161.ogg"},
	{0x1F318A057A9BCE04ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02a.mmh"},
	{0x1F31A0057A9BF3A6ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02a.msh"},
	{0x1F3201D04E3DC4F4ULL, "vxht_0_hightown_a3_cal_0_0.vlm"},
	{0x1F3228E2350965E5ULL, "6060450_m.fxe"},
	{0x1F32501AF8BD1098ULL, "987190439.ogg"},
	{0x1F32B3E9FA26059CULL, "vxht_0_hightown_n_c8a_0_0.vlm"},
	{0x1F32C79B7AFDB876ULL, "6132750_m.fxe"},
	{0x1F331CEB27CEF325ULL, "art\\vfx\\creatures\\fxc_ash_ww_p.anb"},
	{0x1F3365CE9A6BC150ULL, "271156370.ogg"},
	{0x1F34E894E8B42563ULL, "qcr000_varric6066265.cl"},
	{0x1F34FD0414C29253ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_1x.dds"},
	{0x1F35C3B3C8AFA7FFULL, "mag131_keran6086177.cl"},
	{0x1F35D180C4B3D5AEULL, "188491315.ogg"},
	{0x1F36A2EEF86E03C5ULL, "mag291_torpor6026993.cl"},
	{0x1F36DAAFF7C8099BULL, "vxlt_darktown_4yb_0_0.vlm"},
	{0x1F37247CC46E8463ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb01.phy"},
	{0x1F373F4624EA3955ULL, "zz_otl_debug.ncs"},
	{0x1F37530248EA1EB4ULL, "vxht_mansion2b_2gf_0_0.vlm"},
	{0x1F3789C75D5A16F7ULL, "zz_and_debug6134128.cl"},
	{0x1F37C98047FEF18EULL, "llac_lt_undercityencounter_-1_33.dds"},
	{0x1F38276B8CF58450ULL, "vxht_mansion2_1sy_0_0.vlm"},
	{0x1F38A0099740BF64ULL, "vxlt_gallowscourtyard_7cn_0_0.vlm"},
	{0x1F38A48DF88CB69DULL, "gen00fl_isabela6101083.cl"},
	{0x1F38AEC78F0EFE59ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbuckfruit_0.pwk"},
	{0x1F38E20F865D0060ULL, "457693257.ogg"},
	{0x1F38EFFFA0EC8FA3ULL, "vxdr_primevaltomb_4hw_0_0.vlm"},
	{0x1F398782B1C46643ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he01.mmh"},
	{0x1F399582B1C47D9DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he01.msh"},
	{0x1F3998FED17253CAULL, "114150226.ogg"},
	{0x1F39C4979D7716BCULL, "11202963.ogg"},
	{0x1F39E9FA5A6D7E00ULL, "vxw_sundermount_2mt_0_0.vlm"},
	{0x1F3A1604FDB8B1E5ULL, "lt_0_lowtown_34_0_ll.mmh"},
	{0x1F3A617804F02E2FULL, "458277974.ogg"},
	{0x1F3AE40EB313314DULL, "vxdr_primevaltomb_53p_0_1.vlm"},
	{0x1F3B0F7369988BD4ULL, "hlac_lt_gallowstemplar_2391_3x.dds"},
	{0x1F3B502392F71C00ULL, "hlac_ca_caverns1_24425_0z.dds"},
	{0x1F3B7B419949EB75ULL, "w_bonepitdestroyed_a3_41_0.msh"},
	{0x1F3B7D419949EEEBULL, "w_bonepitdestroyed_a3_41_0.mmh"},
	{0x1F3B8941994A0288ULL, "w_bonepitdestroyed_a3_41_0.mao"},
	{0x1F3BDB79B4E92C1BULL, "face_rendertoshadowmap_p0.vsh.fxo.dx11"},
	{0x1F3C1663CF60460CULL, "llac_lt_0_lowtown_docks_a3_-131075_5.dds"},
	{0x1F3C4EA6B98E478CULL, "6118697_m.fxe"},
	{0x1F3C583420DDA2F2ULL, "vxlt_undercityencounter_1ql_0_0.vlm"},
	{0x1F3CCAFA2FB4D883ULL, "6053168_m.fxe"},
	{0x1F3D2F443C4628BCULL, "vxht_0_hightown_n_awr_0_0.vlm"},
	{0x1F3D562C7524E635ULL, "carpetold_n.dds"},
	{0x1F3D715626B5C1C0ULL, "889630750.ogg"},
	{0x1F3DADF6774142A1ULL, "6116719_m.fxe"},
	{0x1F3E3D89895F1A5DULL, "mag170_thrask6020554.cl"},
	{0x1F3E5187E34BFF76ULL, "cheese_d.dds"},
	{0x1F3E61F5FE2940E3ULL, "ht_0_hightownexplosion_n_29_0.rml"},
	{0x1F3EBF82025574CAULL, "619809401.ogg"},
	{0x1F3EED8B216C1847ULL, "6070794_m.fxe"},
	{0x1F3F4E902F34FD4EULL, "mag291_demon_marethari6029060.cl"},
	{0x1F3F5B285AA6EC4EULL, "vxlt_darktown_2uu_0_0.vlm"},
	{0x1F3FA3B8FF343C44ULL, "pro000ip_merchant_chest.ncs"},
	{0x1F3FBB3886B44EEDULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_roba_n.dds"},
	{0x1F3FD69EC5C6B2D8ULL, "933210022.ogg"},
	{0x1F3FDFF56592FAB7ULL, "zz_dae_debug6029896.cl"},
	{0x1F400AAD6EE487F1ULL, "vxlt_0_lowtown_a2_n_c_fex_0_0.vlm"},
	{0x1F404464E18BF74DULL, "vxlt_hangedman_20d_0_0.vlm"},
	{0x1F40922A82AB9D33ULL, "dae300_alistair_king6133799.cl"},
	{0x1F411847EF26568EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_lean_forward_look_right.ani"},
	{0x1F4293C6D6EB3612ULL, "one220_yevhen6102829.cl"},
	{0x1F42B0D6F5C052A6ULL, "vxlt_undercityencounter_zb_0_0.vlm"},
	{0x1F42E3C99D308E66ULL, "256800636.ogg"},
	{0x1F432ACAD80DE474ULL, "6120746_m.fxe"},
	{0x1F4477578200B767ULL, "vxlt_darktown_5lh_0_0.vlm"},
	{0x1F448DC68873F83EULL, "6087649_m.fxe"},
	{0x1F44B4B32BB1A5A0ULL, "vxca_caverns1_egy_0_0.vlm"},
	{0x1F44DFC709C9ECFAULL, "plt_vrc200pt_n_romance.nss"},
	{0x1F46B5B6F3E5A25FULL, "vxw_woundedcoast_zh_0_0.vlm"},
	{0x1F46D199055D3A17ULL, "vxw_bonepitdestroyed_a3_1ou_0_0.vlm"},
	{0x1F4711E68936D7E4ULL, "vxlt_gallowscourtyard_7do_0_0.vlm"},
	{0x1F47150995EE3DEEULL, "vxca_caverns2prc_adl_0_0.vlm"},
	{0x1F4864317B89EA4FULL, "hlac_ht_0_hightown_a3_196607_0x.dds"},
	{0x1F48FDBA25140318ULL, "vxlt_0_lowtown_a2_n_ey0_0_0.vlm"},
	{0x1F491BDC9CD7AC5BULL, "vxht_0_hightown_a2_big_0_0.vlm"},
	{0x1F49D2E2C887A7DFULL, "vxw_bonepit_1vf_0_0.vlm"},
	{0x1F4A3543E2B591B6ULL, "gen_im_arm_boo_hvy_act1_03.uti"},
	{0x1F4A80B4EB4E51F8ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em03.phy"},
	{0x1F4A8882B1D2DB68ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he01.phy"},
	{0x1F4B10B9AF84B207ULL, "character_character_p259.psh.fxo.dx11"},
	{0x1F4B876B1CC60E81ULL, "ring_hawk_family.dds"},
	{0x1F4B9A22C6A9136EULL, "hlac_ht_mansion2_818_19z.dds"},
	{0x1F4BCDC1CE785F96ULL, "hlac_lt_0_lowtown_a2_n_-2_0y.dds"},
	{0x1F4BF9FDAA9CA4CFULL, "6122879_m.fxe"},
	{0x1F4C46B2E33DAE55ULL, "amb_scene_violentattack.wbd"},
	{0x1F4C58E7C6878A1AULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_movement_explore.anb"},
	{0x1F4C86EAD703F2E5ULL, "6139658_f.fxe"},
	{0x1F4C88049CCDB8D9ULL, "dae000_arishok6080474.cl"},
	{0x1F4CF243FB987E87ULL, "6115741_m.fxe"},
	{0x1F4D28C0F18C412AULL, "6047462_f.fxe"},
	{0x1F4DCD054C9D04CEULL, "vxlt_darktown_2xv_0_0.vlm"},
	{0x1F4DE2BC770A25B9ULL, "143910290.ogg"},
	{0x1F4E5BE0D11FD88FULL, "6162633_m.fxe"},
	{0x1F4E64AC4E975514ULL, "693095426.ogg"},
	{0x1F4F306BDC93AB2AULL, "vxht_0_hightown_cdt_0_0.vlm"},
	{0x1F4F32DB4419518EULL, "913309939.ogg"},
	{0x1F4FA8F024E1C27AULL, "llac_dr_primevaltomb_11985_0.dds"},
	{0x1F4FD8286746C1C4ULL, "218598421.ogg"},
	{0x1F4FE063E58FE89AULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.cess.gad"},
	{0x1F50099327C220F6ULL, "366863095.ogg"},
	{0x1F50A0E5C79E06DAULL, "vxht_mansion2b_2yf_0_0.vlm"},
	{0x1F50F3733901138FULL, "87411934.ogg"},
	{0x1F50FFB9369E015FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_beckon.gad"},
	{0x1F510ED9750E2965ULL, "art\\vfx\\environment\\vfx_lyrium_aurora.mmh"},
	{0x1F51C6D1F5A089C1ULL, "vxlt_0_lowtown_n_e19_0_0.vlm"},
	{0x1F51DF7241651FEAULL, "vxca_caverns1_j1w_0_1.vlm"},
	{0x1F5206859CC3A080ULL, "lowtown_fem_lowlife_mob6168985.cl"},
	{0x1F5306EC13630189ULL, "367245914.ogg"},
	{0x1F537FCE7C5042F1ULL, "6115453_m.fxe"},
	{0x1F53CAD31D0FD612ULL, "68453653.ogg"},
	{0x1F53D515A921B84FULL, "lowtown_male_lowlife1_sta6164964.cl"},
	{0x1F53F206D571C942ULL, "w_blightlands_134_0.phy"},
	{0x1F53FA2B771EACEDULL, "956083079.ogg"},
	{0x1F54173237DF93F3ULL, "qun211_braggart6075721.cl"},
	{0x1F54E6F1C8049F3DULL, "6088063_m.fxe"},
	{0x1F54F147EF36BDC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_lean_forward_look_right.gad"},
	{0x1F552793E5DBBADFULL, "vxw_woundedcoast_wk_0_0.vlm"},
	{0x1F556C3D9362CA7BULL, "llac_w_encounter2_131071_0.dds"},
	{0x1F5586E687E4566EULL, "vxca_caverns2prc_acm_0_0.vlm"},
	{0x1F55DE6506B01A7EULL, "145716893.ogg"},
	{0x1F560B663B672B35ULL, "388709912.ogg"},
	{0x1F56172393455EA6ULL, "vxlt_gallowstemplar_3qx_0_0.vlm"},
	{0x1F5652A28AE4250BULL, "vxht_0_hightown_a3_9aq_0_0.vlm"},
	{0x1F56E363E5960CB1ULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.cess.evt"},
	{0x1F571934E0417086ULL, "418174500.ogg"},
	{0x1F572976DC49E858ULL, "hlac_lt_0_lowtown_docks_n_65532_0x.dds"},
	{0x1F57299CFC775D07ULL, "gen00fl_fenris6127078.cl"},
	{0x1F574F8D7723075CULL, "vxw_blightlands_e0_0_0.vlm"},
	{0x1F5762B040EF7B93ULL, "6031985_f.fxe"},
	{0x1F57D60AD94982DAULL, "w_sundermount_40_0_ll.mmh"},
	{0x1F57EF64CE7DFE7CULL, "726423497.ogg"},
	{0x1F57F4B936A40D4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_beckon.evt"},
	{0x1F582B9C4DA0681AULL, "592133696.ogg"},
	{0x1F583B6C30616847ULL, "1013204022.ogg"},
	{0x1F592F191C55FC79ULL, "vxca_caverns1_mines_hur_0_0.vlm"},
	{0x1F595C9178862661ULL, "6122824_m.fxe"},
	{0x1F59D7AB1A71B476ULL, "vxht_0_hightown_a2_car_0_1.vlm"},
	{0x1F59E18BEA817F9FULL, "hlac_lt_warehouse_n_65536_1y.dds"},
	{0x1F59EF8B39987B93ULL, "vxw_woundedcoast_15x_0_0.vlm"},
	{0x1F59F2FF2465B264ULL, "zz_dae_debug6029071.cl"},
	{0x1F59FFFA76365822ULL, "vxca_caverns2_83y_0_0.vlm"},
	{0x1F5A25C80A066656ULL, "vxw_woundedcoast_ip_0_0.vlm"},
	{0x1F5A3D395AEEE3C5ULL, "pro000_athenril6049447.cl"},
	{0x1F5A3D9E1F310D0AULL, "llac_lt_0_lowtown_docks_n_-65538_3.dds"},
	{0x1F5AAB19616D3185ULL, "mag211_marethari6025304.cl"},
	{0x1F5ACAF8A0FD2709ULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.mmh"},
	{0x1F5B2982CB84237CULL, "hlac_lt_gallowsprison_n_-1_3z.dds"},
	{0x1F5B61589B0B9B2CULL, "face_face_p10.psh.fxo"},
	{0x1F5B8229306CCA8EULL, "lgt351ip_shw_mer_at_smount.utp"},
	{0x1F5B890C94C17A7EULL, "vxlt_gallowstemplar_5cx_0_0.vlm"},
	{0x1F5C11BD512A796CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wickermat.phy"},
	{0x1F5C1DEB27F1FA66ULL, "art\\vfx\\creatures\\fxc_ash_ww_p.mmh"},
	{0x1F5C2FDD78451DD9ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_chantryboard.msh"},
	{0x1F5C39DD78452EC7ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_chantryboard.mmh"},
	{0x1F5C3DDD78453514ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_chantryboard.mao"},
	{0x1F5C6FE20834DD7DULL, "vxlt_0_lowtown_docks_n_b48_0_0.vlm"},
	{0x1F5CB4ADFCFFC308ULL, "6038458_m.fxe"},
	{0x1F5F0BD9751A6132ULL, "art\\vfx\\environment\\vfx_lyrium_aurora.anb"},
	{0x1F5F1A25E3BEB34AULL, "114596383.ogg"},
	{0x1F5F3B67BF24930FULL, "llac_ht_chantry_2534_44.dds"},
	{0x1F5F88AC8276E3B2ULL, "vxw_encounter2_camp_9y_0_0.vlm"},
	{0x1F5F993B079A4ADCULL, "lt_0_lowtown_docks_a3_239_0.mao"},
	{0x1F5FA53B079A5F7FULL, "lt_0_lowtown_docks_a3_239_0.mmh"},
	{0x1F5FAB3B079A6921ULL, "lt_0_lowtown_docks_a3_239_0.msh"},
	{0x1F5FAD10D0F86B62ULL, "hlac_lt_0_lowtown_docks_n_65533_4y.dds"},
	{0x1F5FC56E7474F383ULL, "qcr220_grey_wardens6135940.cl"},
	{0x1F600FCBE4C21F33ULL, "vxlt_0_lowtown_docks_a3_d0h_0_0.vlm"},
	{0x1F602D02C94CA166ULL, "vxca_cavern1_haunted_jgi_0_0.vlm"},
	{0x1F60DA82EC387502ULL, "412731820.ogg"},
	{0x1F615ED702A1A084ULL, "art\\vfx\\combat\\vfx_r_dwp_left_close1_c.anb"},
	{0x1F618D8535D5ED1DULL, "vxca_caverns1_mines_i8l_0_0.vlm"},
	{0x1F61CA12D63C0517ULL, "362156174.ogg"},
	{0x1F630D255E86F2F0ULL, "6181757_f.fxe"},
	{0x1F642BF49458A44EULL, "dae000_brothel_orana6160798.cl"},
	{0x1F643CBEBC3D7CCAULL, "6027377_m.fxe"},
	{0x1F6464B784008489ULL, "6088791_m.fxe"},
	{0x1F647501081C2D09ULL, "337228453.ogg"},
	{0x1F64BADCE4EB009CULL, "hlac_w_encounter2_327679_0z.dds"},
	{0x1F64CA036F926EDFULL, "gen00fl_varric6078851.cl"},
	{0x1F64DCB936AE6EA5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_beckon.ani"},
	{0x1F64F0D1830C4AEBULL, "6041750_m.fxe"},
	{0x1F64FB63E5A27280ULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.cess.ani"},
	{0x1F65736C521FB82BULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.dura.ani"},
	{0x1F65C16A69191FDCULL, "vxw_blightlands_n1_0_0.vlm"},
	{0x1F65F4E3A924B2BBULL, "6173095_m.fxe"},
	{0x1F66E8277BDF69BEULL, "vxht_keep_6fw_0_0.vlm"},
	{0x1F66F0A325113670ULL, "vxlt_0_lowtown_a2_def_0_0.vlm"},
	{0x1F674D6DFAC6B968ULL, "plt_genpt_crafting.plo"},
	{0x1F677BD22976DEE1ULL, "lgt311_shw_mer_intro6156404.cl"},
	{0x1F6807E5FC6105ADULL, "docks_male_laborer_mob6175407.cl"},
	{0x1F6892F2D0E76F6DULL, "6095612_m.fxe"},
	{0x1F6897A4FBFC7ED6ULL, "vxw_woundedcoast_fq_0_0.vlm"},
	{0x1F69104E08ED1C81ULL, "framebuffereffects_diffusiondof_p1.psh.fxo.dx11"},
	{0x1F693B6A808B95CBULL, "vxlt_0_lowtown_docks_b8e_0_0.vlm"},
	{0x1F6A4F890CF4BC37ULL, "579896416.ogg"},
	{0x1F6AC7F630D4FAD8ULL, "fen110_fenris6117965.cl"},
	{0x1F6AFCEBFA54BF21ULL, "3327057.ogg"},
	{0x1F6B5C1FC1433111ULL, "llac_lt_gallowstemplar_n_2391_9.dds"},
	{0x1F6BB66AE21F8F45ULL, "mcr310_argument6075799.cl"},
	{0x1F6C3275410BEC8BULL, "pausemenu.gfx"},
	{0x1F6D1A1FD0A75852ULL, "clf260_harley6069664.cl"},
	{0x1F6D5A3E30A73C63ULL, "6041259_m.fxe"},
	{0x1F6D67201618B462ULL, "follower_banter6143373.cl"},
	{0x1F6D7F669BE5A7C6ULL, "61567991.ogg"},
	{0x1F6DC5F569120484ULL, "6042833_m.fxe"},
	{0x1F6DD86A27A7404CULL, "vxlt_undercityserialkiller_18u_0_0.vlm"},
	{0x1F6DFA89746CFC32ULL, "vxw_encounter2_camp_4z_0_0.vlm"},
	{0x1F6E9756880C9BDFULL, "6136979_m.fxe"},
	{0x1F6F69B4CC3EF5D0ULL, "322116274.ogg"},
	{0x1F6F6BD4EF217C4EULL, "ave200_donnic6110712.cl"},
	{0x1F6F7DD702AE1287ULL, "art\\vfx\\combat\\vfx_r_dwp_left_close1_c.mmh"},
	{0x1F700734D6BCA8CDULL, "6171792_m.fxe"},
	{0x1F700A4A0193EF20ULL, "vxca_caverns1_ehd_0_0.vlm"},
	{0x1F70667308C26F6FULL, "6145199_m.fxe"},
	{0x1F707E3C7F01AEFAULL, "vxdr_primevaltomb_7tm_0_0.vlm"},
	{0x1F7082DA0A5BA925ULL, "follower_banter6164564.cl"},
	{0x1F708F2A8354ED0EULL, "6104386_m.fxe"},
	{0x1F70BAB832A319BBULL, "6164002_m.fxe"},
	{0x1F70EF3C62575BC7ULL, "mrl000_marethari6129444.cl"},
	{0x1F70F578C6F1416FULL, "l_mrl251ar_elf_cave.wbd"},
	{0x1F7173B6F0D95E5BULL, "vxlt_0_lowtown_a2_hdf_0_0.vlm"},
	{0x1F71A5F553996AE0ULL, "hlac_lt_0_lowtown_docks_-131075_4z.dds"},
	{0x1F71AA0E0B576445ULL, "6173413_m.fxe"},
	{0x1F71E0E9BFE8EE68ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sword_n.dds"},
	{0x1F735B6C522BCCA2ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.dura.evt"},
	{0x1F7433475B0F385CULL, "vxw_blightlands_s6_0_0.vlm"},
	{0x1F7437E22ECD86C0ULL, "vxlt_backalley_g_1pd_0_1.vlm"},
	{0x1F74D20CE0E0E2A3ULL, "gen00fl_anders6112960.cl"},
	{0x1F74FB689AB1A604ULL, "lgt231ar_tome_final.are"},
	{0x1F7509FD586A873DULL, "632146449.ogg"},
	{0x1F750C1CBA5C917CULL, "dre000_sibling_has_blight6036528.cl"},
	{0x1F7546E56BF035B5ULL, "vxht_0_hightown_n_aia_0_0.vlm"},
	{0x1F755A046DD5823EULL, "vxht_keep_6ch_0_0.vlm"},
	{0x1F75C4E0698E32ABULL, "vxht_mansion_n_4zo_0_0.vlm"},
	{0x1F76BA5E414A83FDULL, "872698944.ogg"},
	{0x1F76C60BF3FB8D0EULL, "6127298_f.fxe"},
	{0x1F770981EDF29756ULL, "vxw_woundedcoast_cr_0_0.vlm"},
	{0x1F7737CF40825325ULL, "vxw_encounter2_4a_0_0.vlm"},
	{0x1F78177031BCE3B7ULL, "20305221.ogg"},
	{0x1F7835EECAB4202EULL, "hlac_w_bonepitdestroyed_a3_327680_0z.dds"},
	{0x1F786CC678ADAB7EULL, "vxlt_gallowstemplar_5iz_0_0.vlm"},
	{0x1F78CD06D590A6F6ULL, "w_blightlands_134_0.mao"},
	{0x1F78D906D590BA95ULL, "w_blightlands_134_0.mmh"},
	{0x1F78DB06D590BE0BULL, "w_blightlands_134_0.msh"},
	{0x1F79252D2069288AULL, "mag311_nyssa6172421.cl"},
	{0x1F793243A084286AULL, "6133858_m.fxe"},
	{0x1F79B99D1E9D0881ULL, "vxw_blightlands_ea_0_0.vlm"},
	{0x1F79CF15C62FEE9CULL, "hlac_lt_hangedman_n_-1_15z.dds"},
	{0x1F7A5A6C5231E9B5ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.dura.gad"},
	{0x1F7B40BF60E78532ULL, "art\\vfx\\combat\\vfx_m_pri_fb_cast_c.impa.gad"},
	{0x1F7C85D76C8A85E4ULL, "vxw_bonepitdestroyed_a3_1pk_0_0.vlm"},
	{0x1F7C9BCA5A9FC430ULL, "6151041_m.fxe"},
	{0x1F7D8740B61994C7ULL, "vrc321tr_bartrands_items.ncs"},
	{0x1F7D8D62A5F41D58ULL, "art\\characters\\creatures\\dog\\textures\\c_dog_tatc.dds"},
	{0x1F7D91882F7C9EA6ULL, "270889833.ogg"},
	{0x1F7DD6A99BA38D7BULL, "vxlt_backalley_n_1u2_0_2.vlm"},
	{0x1F7E196AE91E0F41ULL, "591771182.ogg"},
	{0x1F7E3EC0D25A0FEEULL, "6169853_m.fxe"},
	{0x1F7E713F19C21E1DULL, "vxca_caverns1_mines_i2j_0_0.vlm"},
	{0x1F7E7911C8B2C14DULL, "6174492_m.fxe"},
	{0x1F7E8679FEE288C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\runforward.ani"},
	{0x1F7EB364E9A9887FULL, "one100dg_kelder6045192.cl"},
	{0x1F7ED22FAFD237CFULL, "6105915_m.fxe"},
	{0x1F7EF01970F7C77AULL, "vxdr_primevaltomb_7qn_0_0.vlm"},
	{0x1F7F0F614471373AULL, "888014973.ogg"},
	{0x1F7F3A5BD3027949ULL, "vxlt_undercityserialkiller_1ga_0_0.vlm"},
	{0x1F7F94035372B537ULL, "vxht_mansion2b_1it_0_0.vlm"},
	{0x1F7FDEE9E54DB7EBULL, "6033953_m.fxe"},
	{0x1F80798110A5CA23ULL, "vxht_0_hightown_9xg_0_0.vlm"},
	{0x1F807B46C744F336ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_001a_0d.dds"},
	{0x1F80E02AAA95998DULL, "6051487_m.fxe"},
	{0x1F80F4C15116ECACULL, "6083364_m.fxe"},
	{0x1F817876BEC10DB3ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\playercharacter_.ani"},
	{0x1F81911CDC73AA21ULL, "hlac_lt_gallowscourtyard_131071_0z.dds"},
	{0x1F81F8FD0FA9FA34ULL, "6167620_m.fxe"},
	{0x1F8243BF60EDA969ULL, "art\\vfx\\combat\\vfx_m_pri_fb_cast_c.impa.evt"},
	{0x1F8265768E0B4A9EULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrainexit.ani"},
	{0x1F827247E1AF3483ULL, "490821874.ogg"},
	{0x1F82A5244D0550DCULL, "vxw_blightlands_t7_0_0.vlm"},
	{0x1F82BB01B26B50D3ULL, "art\\vfx\\combat\\vfx_glb_ele_e_additive_c.impa.ani"},
	{0x1F82C2770DC17960ULL, "pro000_colonel_ewald6047399.cl"},
	{0x1F8333F177084154ULL, "hlac_lt_gallowstemplar_f_7198_2x.dds"},
	{0x1F83C7F8A120273EULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.anb"},
	{0x1F857B5EDFE8AFD6ULL, "vxw_woundedcoast_xs_0_0.vlm"},
	{0x1F85AB4C439A6193ULL, "6101810_m.fxe"},
	{0x1F86102E7C024815ULL, "hlac_ht_keep_n_3736_2x.dds"},
	{0x1F867FA0366DD82BULL, "6111855_m.fxe"},
	{0x1F868A264C44C202ULL, "6117030_m.fxe"},
	{0x1F86C70E0778DB84ULL, "724133729.ogg"},
	{0x1F86EEFE4B0E89FBULL, "hlac_w_woundedcoast_0_4x.dds"},
	{0x1F8742F5F74D2EEFULL, "llac_lt_backalley_g_65536_6.dds"},
	{0x1F8790A12A67AA14ULL, "fence_d.dds"},
	{0x1F8801E5BB81D8F8ULL, "6093871_m.fxe"},
	{0x1F88A6B1E2DDABF6ULL, "567126768.ogg"},
	{0x1F8922DD786BF5FCULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_chantryboard.phy"},
	{0x1F892445A8D16CCEULL, "vxlt_undercityencounter_1fi_0_0.vlm"},
	{0x1F89869CB2F9189DULL, "vxht_mansion2b_3by_0_0.vlm"},
	{0x1F8987967F9D5BA5ULL, "6095173_m.fxe"},
	{0x1F8A68598F7668DAULL, "vxht_mansion2b_2eb_0_0.vlm"},
	{0x1F8A84668AD1FD57ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_zbm_0d.dds"},
	{0x1F8A8993BC181031ULL, "6092222_m.fxe"},
	{0x1F8A99553190A8D8ULL, "hlac_lt_0_lowtown_docks_a3_131072_1z.dds"},
	{0x1F8AF7B45E809E64ULL, "vxw_bonepitdestroyed_a3_1uh_0_0.vlm"},
	{0x1F8B1D711BA9EAF0ULL, "w_woundedcoast_174_0.mao"},
	{0x1F8B21711BA9F103ULL, "w_woundedcoast_174_0.mmh"},
	{0x1F8B2F711BAA095DULL, "w_woundedcoast_174_0.msh"},
	{0x1F8B3C2552C3C880ULL, "hlac_w_encounter1_-65535_0x.dds"},
	{0x1F8BC89A2232BD6EULL, "vxlt_undercitydungeon_1c7_0_0.vlm"},
	{0x1F8C6E79FEEE9D37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\runforward.evt"},
	{0x1F8CD91C438FABA2ULL, "one350_vinell6106484.cl"},
	{0x1F8DDC3113906790ULL, "279663199.ogg"},
	{0x1F8E2B415C621CF8ULL, "vxht_keep_66q_0_0.vlm"},
	{0x1F8E7E52750E0FDAULL, "6053194_m.fxe"},
	{0x1F8E9F45F70DBDDBULL, "671211849.ogg"},
	{0x1F8F1C77DF1A9E10ULL, "vxht_keep_n_8b0_0_0.vlm"},
	{0x1F8F29D34CFFBE70ULL, "fxa_shield_spot.mao"},
	{0x1F8F2BBF60F80AD8ULL, "art\\vfx\\combat\\vfx_m_pri_fb_cast_c.impa.ani"},
	{0x1F8FB301B275CD0AULL, "art\\vfx\\combat\\vfx_glb_ele_e_additive_c.impa.evt"},
	{0x1F8FDD91B2D3E9ECULL, "6135246_f.fxe"},
	{0x1F8FE728956E7254ULL, "vxht_chantry_33k_0_0.vlm"},
	{0x1F90062FC5B58F91ULL, "6080676_m.fxe"},
	{0x1F90D39952D520F1ULL, "vxca_caverns1_e4m_0_0.vlm"},
	{0x1F912022A82F42C9ULL, "6170597_m.fxe"},
	{0x1F91F04E047F87EDULL, "gen_im_trs_gem_onyx_03.uti"},
	{0x1F9205FC8376FF90ULL, "follower_banter6143238.cl"},
	{0x1F92291B30989A15ULL, "vxht_mansion2b_2k7_0_0.vlm"},
	{0x1F92DAAF582FE869ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleftcrn.mao"},
	{0x1F92DBF4F1FC2B36ULL, "872296219.ogg"},
	{0x1F92E6AF582FFC02ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleftcrn.mmh"},
	{0x1F92E8AF582FFFB8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleftcrn.msh"},
	{0x1F9309A6F8C2F511ULL, "ser120_ghyslain6104337.cl"},
	{0x1F935422AFC5BC9CULL, "hlac_lt_0_lowtown_n_-65535_1x.dds"},
	{0x1F936F79FEF4BDC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\runforward.gad"},
	{0x1F937CCCE7E40D51ULL, "vxht_chantry_2s3_0_0.vlm"},
	{0x1F93ED3BD1DEC856ULL, "vxw_woundedcoast_ut_0_0.vlm"},
	{0x1F9402689ACC62A6ULL, "lgt231ar_tome_final.ncs"},
	{0x1F942F345F76242BULL, "vxw_bonepitdestroyed_a3_1x9_0_1.vlm"},
	{0x1F945A4BC48E57A3ULL, "llac_dr_0_deeproads_5841_1.dds"},
	{0x1F946EE54EB2B8A5ULL, "vxw_blightlands_dz_0_0.vlm"},
	{0x1F9486A53C6A5EA8ULL, "vxlt_gallowstemplar_f_5jz_0_0.vlm"},
	{0x1F94F096DF1F15D0ULL, "gy_smo.mao"},
	{0x1F9572335F083EBAULL, "vxw_woundedcoast_5x_0_0.vlm"},
	{0x1F95771D2C8F905BULL, "vxht_keep_n_57y_0_0.vlm"},
	{0x1F95D8ABAC9E6ADAULL, "lgt220cr_shp_mag_victim.utc"},
	{0x1F960069E47F937AULL, "vxca_cave_230_0_0.vlm"},
	{0x1F9648D7C462C5ABULL, "vxlt_gallowsprison_n_6qc_0_0.vlm"},
	{0x1F964F76BED3245DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\playercharacter_.gad"},
	{0x1F967E768E1C1E18ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrainexit.gad"},
	{0x1F969201B27BB47DULL, "art\\vfx\\combat\\vfx_glb_ele_e_additive_c.impa.gad"},
	{0x1F969D5702893981ULL, "vxw_blightlands_sg_0_0.vlm"},
	{0x1F96BFCE19EBDBDBULL, "gen00fl_fenris6106703.cl"},
	{0x1F96D7F1244A3C01ULL, "20245260.ogg"},
	{0x1F96F10B7949F020ULL, "vxw_bonepit_16e_0_0.vlm"},
	{0x1F977CF109215BA0ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_2x.dds"},
	{0x1F9786CEA7EEF497ULL, "hlac_lt_0_lowtown_131071_5x.dds"},
	{0x1F97935131EF9F38ULL, "668505251.ogg"},
	{0x1F97F2AAEC5300B4ULL, "6083975_m.fxe"},
	{0x1F982EFF053D77E3ULL, "907468696.ogg"},
	{0x1F983019ED9E495CULL, "559486782.ogg"},
	{0x1F992E00FD8989CCULL, "vxlt_undercityserialkiller_17h_0_0.vlm"},
	{0x1F99928F347ED9B2ULL, "6121949_m.fxe"},
	{0x1F99B5A8F85BA276ULL, "6070144_m.fxe"},
	{0x1F99B90038AD83B4ULL, "6126843_m.fxe"},
	{0x1F99EE87CA236174ULL, "vxdr_0_deeproads_ax3_0_0.vlm"},
	{0x1F9A8D9D82913772ULL, "gen_im_wep_war_one_act3_02.uti"},
	{0x1F9ACE71AB9F2B62ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf03a.msh"},
	{0x1F9AD071AB9F2E18ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf03a.mmh"},
	{0x1F9AD2DABB5497E4ULL, "vxca_cavern1_haunted_i2g_0_0.vlm"},
	{0x1F9AFA689AD2739FULL, "lgt231ar_tome_final.lst"},
	{0x1F9B1C928F0FFA8BULL, "hightown_fem_courtesan_mob6168593.cl"},
	{0x1F9B622C9216BC9FULL, "vxht_0_hightown_ato_0_0.vlm"},
	{0x1F9B9EAF8DD580C6ULL, "65375824.ogg"},
	{0x1F9BAE4FB0C27170ULL, "6060705_m.fxe"},
	{0x1F9BD48EBE6E0CA6ULL, "vxht_0_hightown_a2_atn_0_0.vlm"},
	{0x1F9BE1CDCE09FBB9ULL, "6081385_m.fxe"},
	{0x1F9C1E15B6EEAA49ULL, "vxlt_undercityserialkiller_1qg_0_0.vlm"},
	{0x1F9C2EC5EC2A01D7ULL, "llac_lt_gallowstemplar_2382_2.dds"},
	{0x1F9C5E9918D0223DULL, "dae000_templar_recruit_16164432.cl"},
	{0x1F9C77BD375EE637ULL, "vxht_mansion2b_1gv_0_0.vlm"},
	{0x1F9CE20DC8C632FDULL, "6070477_m.fxe"},
	{0x1F9CE8BD5161CA37ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wickermat.mmh"},
	{0x1F9CECBD5161D144ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wickermat.mao"},
	{0x1F9CFEBD5161EFA9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wickermat.msh"},
	{0x1F9D055C10B809BDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scale.mao"},
	{0x1F9D095C10B810F6ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scale.mmh"},
	{0x1F9D135C10B821F4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scale.msh"},
	{0x1F9E590587648AD4ULL, "vxht_chantry_36h_0_0.vlm"},
	{0x1F9E5DDDFE58F181ULL, "hlac_w_woundedcoast_327687_0y.dds"},
	{0x1F9EA548B348C5AAULL, "dae000_mother6099231.cl"},
	{0x1F9EBB504DB05227ULL, "6080454_m.fxe"},
	{0x1F9EDDC31A461789ULL, "ser270_emeric6149369.cl"},
	{0x1F9FE2D15FB29553ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_db_transition.ani"},
	{0x1FA04F05934CBAF8ULL, "6164038_m.fxe"},
	{0x1FA0BAC8B8D33AE4ULL, "lt_0_lowtown_docks_a3_39_0_ll.mmh"},
	{0x1FA0FC1C23AC69C4ULL, "6072828_f.fxe"},
	{0x1FA129099BA6AE4FULL, "gen00fl_merrill6106963.cl"},
	{0x1FA1C8D83EC4EDEDULL, "vxw_encounter2_f3_0_0.vlm"},
	{0x1FA1EB9A468631A1ULL, "art\\vfx\\combat\\vfx_gib_glv_tmpa.mao"},
	{0x1FA245BC969E0320ULL, "vxw_woundedcoast_7h_0_0.vlm"},
	{0x1FA246D6F11FD3F3ULL, "117620129.ogg"},
	{0x1FA25F18C3D4E0D6ULL, "vxw_woundedcoast_ru_0_0.vlm"},
	{0x1FA27A34C101E369ULL, "339528380.ogg"},
	{0x1FA2E0C240A8D125ULL, "vxw_blightlands_iy_0_0.vlm"},
	{0x1FA47246D675ABFAULL, "vxca_cave_241_0_0.vlm"},
	{0x1FA4AC9314359B8FULL, "dae000ip_crowd_ff7_talk16.utp"},
	{0x1FA4D771ABA78EA3ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf03a.phy"},
	{0x1FA50F33F47F5201ULL, "vxw_blightlands_tf_0_0.vlm"},
	{0x1FA532EF7E288BBDULL, "gen_im_trs_chn_junk2_01.uti"},
	{0x1FA56DB4EB9ADC8DULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em03.msh"},
	{0x1FA57FB4EB9AFB13ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em03.mmh"},
	{0x1FA597139C981AC9ULL, "zz_mer_debug6016968.cl"},
	{0x1FA59C3722FD47D5ULL, "og_flsh_04.dds"},
	{0x1FA5C0F1BBCFC5EEULL, "art\\characters\\playercharacter\\humanmale\\hoo_chac.msh"},
	{0x1FA5DAF1BBCFF1DCULL, "art\\characters\\playercharacter\\humanmale\\hoo_chac.mmh"},
	{0x1FA6285237DF0F3FULL, "351985077.ogg"},
	{0x1FA6625940F9AFA1ULL, "6153473_m.fxe"},
	{0x1FA6922EBDFA4F4FULL, "vxlt_0_lowtown_a2_n_c_byf_0_0.vlm"},
	{0x1FA6F271050DCB3FULL, "hlac_lt_0_lowtown_a2_65536_3z.dds"},
	{0x1FA73B389CE79AB2ULL, "isa211_isabela6202287.cl"},
	{0x1FA769F0D05B5890ULL, "vxht_bartrand_n_4ns_0_0.vlm"},
	{0x1FA78981CAAAED53ULL, "hlac_ca_caverns1_mines_24424_4y.dds"},
	{0x1FA7A84E371F2845ULL, "art\\vfx\\cinematic\\vfx_mother_tears.mmh"},
	{0x1FA7DB6E426CCF64ULL, "vxw_bonepitdestroyed_a3_1cn_0_0.vlm"},
	{0x1FA8EB93F70E6F77ULL, "vxlt_0_lowtown_a3_n_c_hx6_0_0.vlm"},
	{0x1FA940DD489282E5ULL, "6060641_m.fxe"},
	{0x1FA9435550545FF8ULL, "495573490.ogg"},
	{0x1FA963CD73FF8ADDULL, "58062934.ogg"},
	{0x1FA9C43EB298EE6DULL, "one100dg_kelder6060641.cl"},
	{0x1FA9E7BD756C56C5ULL, "vxdr_primevaltomb_7qe_0_0.vlm"},
	{0x1FAA75F7CB1885A7ULL, "138220326.ogg"},
	{0x1FAA93B5084C0D95ULL, "llac_lt_0_lowtown_docks_65534_6.dds"},
	{0x1FAA942B0FF7BB7DULL, "6037325_f.fxe"},
	{0x1FAAE31A8E82ABBCULL, "w_blightlands_ll.mmh"},
	{0x1FAAFCC70A20D16DULL, "plt_vrc200pt_n_romance.plo"},
	{0x1FAB185824A6AC35ULL, "hlac_ht_keep_n_4519_6z.dds"},
	{0x1FAB1EF209085334ULL, "mcr370_meredithrally_st.stg"},
	{0x1FABFC3C2F2078A4ULL, "llac_lt_0_lowtown_n_65538_5.dds"},
	{0x1FAC765F72913A52ULL, "one140cr_charade6128507.cl"},
	{0x1FAC8EE92333CA45ULL, "gen00fl_anders6119466.cl"},
	{0x1FAC8F7C794C5A30ULL, "hlac_w_bonepitdestroyed_a3_0_11x.dds"},
	{0x1FACDAD15FBD118AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_db_transition.evt"},
	{0x1FACDAFCABA4E7F6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02tree.mao"},
	{0x1FACE6FCABA4FB95ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02tree.mmh"},
	{0x1FACE8FCABA4FF0BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02tree.msh"},
	{0x1FAD2B5BF425E631ULL, "zz_one_debug6099969.cl"},
	{0x1FAD6EF24A6DA0F1ULL, "gen_im_trs_dal_junk1_02.uti"},
	{0x1FAD7582F075D4FAULL, "btn_a.dds"},
	{0x1FAD989325404764ULL, "6079578_m.fxe"},
	{0x1FAE10F697F330F2ULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunarirelic_sa.dds"},
	{0x1FAE119840918A73ULL, "vxca_caverns1_mines_jtm_0_0.vlm"},
	{0x1FAF0CD51484CB15ULL, "vxht_mansion2b_2a5_0_0.vlm"},
	{0x1FAF6F90413EB6A4ULL, "238636000.ogg"},
	{0x1FAF79FD5E0F9C36ULL, "follower_banter6155796.cl"},
	{0x1FAF964D86CC6B9BULL, "art\\characters\\playercharacter\\humanmale\\bdy_medd.phy"},
	{0x1FAF9B962C20B21FULL, "6070040_m.fxe"},
	{0x1FAFB0BCDD065D22ULL, "vxlt_undercityserialkiller_yk_0_0.vlm"},
	{0x1FAFC07E0310F584ULL, "6172518_m.fxe"},
	{0x1FB03AB530BB066DULL, "vxw_encounter2_a2_0_0.vlm"},
	{0x1FB064E082FC4858ULL, "follower_banter6164844.cl"},
	{0x1FB0B28D49CD3ABBULL, "6018514_m.fxe"},
	{0x1FB10F78C7275145ULL, "l_mrl251ar_elf_cave.bnk"},
	{0x1FB1234EE7F31521ULL, "dae000ip_crowd_ff1_talk7.utp"},
	{0x1FB12514A77DDC89ULL, "6127009_m.fxe"},
	{0x1FB1529F329EE9A5ULL, "vxw_blightlands_rx_0_0.vlm"},
	{0x1FB2C955928A58E0ULL, "ser221_gamlen6134815.cl"},
	{0x1FB2DEC202E0D925ULL, "art\\vfx\\cinematic\\vfx_anders_intro_c.anb"},
	{0x1FB30C694E86B7D7ULL, "vxdr_0_deeproads_cvl_0_0.vlm"},
	{0x1FB3B9D15FC2F8FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_db_transition.gad"},
	{0x1FB3BD587D2AF660ULL, "666492258.ogg"},
	{0x1FB3EB57BE6A6004ULL, "6208433_m.fxe"},
	{0x1FB3FA3E864C3765ULL, "mer210_smith6051406.cl"},
	{0x1FB44D767A4E56AAULL, "vxht_0_hightown_c9s_0_0.vlm"},
	{0x1FB461E13721F3DDULL, "6059326_m.fxe"},
	{0x1FB4A350F8B72A81ULL, "vxw_sundermount_2hz_0_0.vlm"},
	{0x1FB4D4A360983B4AULL, "gen00fl_sebastian6177897.cl"},
	{0x1FB524943B638056ULL, "gen00fl_aveline6139130.cl"},
	{0x1FB5358BFDD1BBE5ULL, "vxht_mansion2b_2nx_0_0.vlm"},
	{0x1FB536A2794DCAAFULL, "526251976.ogg"},
	{0x1FB571E7B35B66B1ULL, "132543001.ogg"},
	{0x1FB578A0E715D059ULL, "dae000_keep_bureaucrat_1_st6165312.cl"},
	{0x1FB57A13A58BAB4FULL, "745625327.ogg"},
	{0x1FB5A54E372B6012ULL, "art\\vfx\\cinematic\\vfx_mother_tears.anb"},
	{0x1FB5DBCDC2517110ULL, "vxht_bartrand_n_4sp_0_0.vlm"},
	{0x1FB5FAD41E9FE384ULL, "312861054.ogg"},
	{0x1FB6291D27C531E8ULL, "w_bonepitmining_a2_213_0.mmh"},
	{0x1FB6351D27C5458BULL, "w_bonepitmining_a2_213_0.mao"},
	{0x1FB6471D27C56412ULL, "w_bonepitmining_a2_213_0.msh"},
	{0x1FB64D4B3462E7E4ULL, "vxw_bonepitdestroyed_a3_1do_0_0.vlm"},
	{0x1FB694B70655480CULL, "6161604_m.fxe"},
	{0x1FB69971A1589B09ULL, "vxlt_0_lowtown_a2_n_c_i08_0_0.vlm"},
	{0x1FB70F2503A75044ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc04a.msh"},
	{0x1FB7252503A775E6ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc04a.mmh"},
	{0x1FB7320D80FFB930ULL, "w_encounter3_159_0.msh"},
	{0x1FB7400D80FFD14AULL, "w_encounter3_159_0.mmh"},
	{0x1FB7440D80FFD781ULL, "w_encounter3_159_0.mao"},
	{0x1FB7A41BEC8E0F53ULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.cess.ani"},
	{0x1FB7BC46D6B7CF05ULL, "hlac_w_sundermount_0_0x.dds"},
	{0x1FB8575BC6A94FE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_warrior.gad"},
	{0x1FB858D424A9A032ULL, "664253527.ogg"},
	{0x1FB8B568AE2DF48FULL, "dae000ip_crowd_ff4_talk8.utp"},
	{0x1FB8B848A25A3DA6ULL, "vxht_0_hightown_a2_ajp_0_0.vlm"},
	{0x1FB9104564B6B766ULL, "hlac_ht_chantry_n_2534_2x.dds"},
	{0x1FB9662519ED1511ULL, "6075575_m.fxe"},
	{0x1FBA6F9B2DEAF08FULL, "6113278_f.fxe"},
	{0x1FBA7F267AFC351AULL, "gen00fl_aveline6140905.cl"},
	{0x1FBA8C825378F203ULL, "hlac_ca_caverns2_5802_6z.dds"},
	{0x1FBB0A53288E10B3ULL, "vxlt_undercityencounter_1dv_0_0.vlm"},
	{0x1FBB9D2B98B09453ULL, "fen310_fenris6118219.cl"},
	{0x1FBBF7E09A10061DULL, "6106628_m.fxe"},
	{0x1FBC5E9D2E5CA971ULL, "335979269.ogg"},
	{0x1FBD3A6E06A4FEA1ULL, "w_sundermount_179_0.phy"},
	{0x1FBD803625BEB035ULL, "vxht_0_hightown_n_avl_0_0.vlm"},
	{0x1FBDBAFDFBD206ACULL, "character_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0x1FBE1318C1A72CA2ULL, "ht_0_hightown_a2_59_0.rml"},
	{0x1FBE5EA727A34154ULL, "art\\vfx\\creatures\\fxc_reven_pull_p.impa.ani"},
	{0x1FBE7A5BC6ADF6D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_warrior.evt"},
	{0x1FBEAC9222B11EEDULL, "vxw_encounter2_l1_0_0.vlm"},
	{0x1FBEE9C5DA3776DBULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.impa.ani"},
	{0x1FBF37A17B497CB3ULL, "dae000_gen_turn_in_26198702.cl"},
	{0x1FBFC47C24950225ULL, "vxw_blightlands_wg_0_0.vlm"},
	{0x1FC0A06F02F4F071ULL, "hlac_lt_0_lowtown_docks_a3_n_0_5x.dds"},
	{0x1FC0A78478F7A81CULL, "hlac_lt_darktown_-1_4z.dds"},
	{0x1FC11C2503AFBAB1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc04a.phy"},
	{0x1FC17E46407CD057ULL, "vxdr_0_deeproads_cso_0_0.vlm"},
	{0x1FC25CFCD9159DC8ULL, "6110280_m.fxe"},
	{0x1FC2986EAB886D8EULL, "mcr370_varric.cnv"},
	{0x1FC3152DEAAD4301ULL, "vxw_sundermount_2my_0_0.vlm"},
	{0x1FC3508FDCC2D194ULL, "6187774_m.fxe"},
	{0x1FC37B9BD69AC861ULL, "zz_vrc_debug6200227.cl"},
	{0x1FC3B53723166FDFULL, "og_flsh_04.mao"},
	{0x1FC3FAA4DB723450ULL, "6136868_m.fxe"},
	{0x1FC44E95961905BFULL, "405245168.ogg"},
	{0x1FC4902B62B510F2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen2.gad"},
	{0x1FC4927599DE8D29ULL, "dae000_flemeth6013328.cl"},
	{0x1FC49C1BEC988B8AULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.cess.evt"},
	{0x1FC551E70081110EULL, "6164151_m.fxe"},
	{0x1FC55BA727A95AB6ULL, "art\\vfx\\creatures\\fxc_reven_pull_p.impa.gad"},
	{0x1FC5638070718C9FULL, "95845413.ogg"},
	{0x1FC5BA6EAB8ADDFEULL, "mcr370_varric.bnk"},
	{0x1FC6B3451B9C0370ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basicnb.mao"},
	{0x1FC6B7451B9C0983ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basicnb.mmh"},
	{0x1FC6C5451B9C21DDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basicnb.msh"},
	{0x1FC6CA27EC745FE3ULL, "dae000_sandal6147805.cl"},
	{0x1FC7BC2BA581E913ULL, "6087763_m.fxe"},
	{0x1FC7C5A3BC8713B4ULL, "vxht_mansion2b_25l_0_0.vlm"},
	{0x1FC7CD540D412FB7ULL, "vxht_mansion2b_1ps_0_0.vlm"},
	{0x1FC876965CA7105BULL, "vxlt_0_lowtown_docks_ctp_0_0.vlm"},
	{0x1FC87F829D173C32ULL, "6103170_m.fxe"},
	{0x1FC8904BF3242888ULL, "250605509.ogg"},
	{0x1FC89D2CFB7A5B7BULL, "gen00fl_anders6103669.cl"},
	{0x1FC8DFE7535D6B66ULL, "vxlt_backalley_g_1bm_0_0.vlm"},
	{0x1FC8E469F27D8969ULL, "root_rog_general3.dds"},
	{0x1FC9DCD88DEC39D4ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_killdagger_sa.dds"},
	{0x1FC9E9C2D7F3CAD5ULL, "6110260_f.fxe"},
	{0x1FCA4A64515F0CA7ULL, "mrl210_merrill6166844.cl"},
	{0x1FCAB9F1BBEED5A7ULL, "art\\characters\\playercharacter\\humanmale\\hoo_chac.phy"},
	{0x1FCB0C5A4924B417ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_c.anb"},
	{0x1FCB3B323F9A67DFULL, "6040017_f.fxe"},
	{0x1FCB55CAB93975CBULL, "370234592.ogg"},
	{0x1FCB56FA6F81546EULL, "mag360_samson6062121.cl"},
	{0x1FCB70B9A73B95A9ULL, "art\\characters\\creatures\\spider\\textures\\c_giaspidra_0d.dds"},
	{0x1FCB76A727ADF415ULL, "art\\vfx\\creatures\\fxc_reven_pull_p.impa.evt"},
	{0x1FCB7B1BEC9E72FDULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.cess.gad"},
	{0x1FCB932B62BB3429ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen2.evt"},
	{0x1FCBB2D200011823ULL, "w_bonepit_237_0.mao"},
	{0x1FCBB6D200011F70ULL, "w_bonepit_237_0.mmh"},
	{0x1FCBC4D20001368AULL, "w_bonepit_237_0.msh"},
	{0x1FCBF6B11513F5C8ULL, "18626036.ogg"},
	{0x1FCC20A6EFF7B983ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit_2.mao"},
	{0x1FCC496D0E42FA31ULL, "seb221cr_flora_harimann.utc"},
	{0x1FCC638AD7B16019ULL, "116433933.ogg"},
	{0x1FCC825BC6BA41C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_warrior.ani"},
	{0x1FCCF1C5DA43C1D2ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.impa.evt"},
	{0x1FCD55132FA9E16DULL, "mag160_feynriel6020057.cl"},
	{0x1FCDF24CD4883076ULL, "271488237.ogg"},
	{0x1FCEBFF566DDC54BULL, "97273496.ogg"},
	{0x1FCECA721580FD00ULL, "dae300_bran6143520.cl"},
	{0x1FCF09D79480026EULL, "831933937.ogg"},
	{0x1FCF31F4367757B5ULL, "gen00fl_anders6111383.cl"},
	{0x1FCF6265BBB82637ULL, "vxlt_0_lowtown_a2_n_gnd_0_0.vlm"},
	{0x1FCFC54C850E7B3EULL, "llac_w_bonepitmining_a2_196611_0.dds"},
	{0x1FCFE07414E66EA5ULL, "hlac_w_encounter2_camp_262144_0x.dds"},
	{0x1FCFF0EC977CEDF1ULL, "131307343.ogg"},
	{0x1FD01E6A3D9475D6ULL, "291330246.ogg"},
	{0x1FD05F581C2C2595ULL, "ave331_jeven6137264.cl"},
	{0x1FD0714B957B6477ULL, "llac_lt_0_lowtown_a2_131070_6.dds"},
	{0x1FD10B8AFE7D8D8FULL, "897636552.ogg"},
	{0x1FD12F2C4CB32C37ULL, "16941152.ogg"},
	{0x1FD176EE13BF5915ULL, "6053693_f.fxe"},
	{0x1FD1F4915FC4593EULL, "lgt210_qun_blk_qunari.cnv"},
	{0x1FD20A28C7AF6515ULL, "qcr000_arishok6059300.cl"},
	{0x1FD27440B661B41EULL, "vrc321tr_bartrands_items.utt"},
	{0x1FD2B5B303318467ULL, "1017295114.ogg"},
	{0x1FD31A925AED11BCULL, "6080361_m.fxe"},
	{0x1FD388D5B4DB20F7ULL, "623155452.ogg"},
	{0x1FD38E70200E4F09ULL, "follower_banter6155559.cl"},
	{0x1FD3B79D1D9D7EBCULL, "582220906.ogg"},
	{0x1FD410C5DA4A1565ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.impa.gad"},
	{0x1FD42862A41EBE8AULL, "llac_lt_0_lowtown_a2_n_c_65537_0.dds"},
	{0x1FD46260DA64AE42ULL, "gen00fl_anders6095499.cl"},
	{0x1FD47B7C39318F72ULL, "vxht_mansion2b_1hr_0_0.vlm"},
	{0x1FD4E085BA753149ULL, "440291665.ogg"},
	{0x1FD52D2AF5234A17ULL, "mag311_keran6041491.cl"},
	{0x1FD591AD877F5E07ULL, "379080787.ogg"},
	{0x1FD5D6915FC80F4EULL, "lgt210_qun_blk_qunari.bnk"},
	{0x1FD629126165437EULL, "ts_beasts_prog_pelt.gda"},
	{0x1FD69248135BE1B4ULL, "hlac_w_bonepitmining_a2_262147_0x.dds"},
	{0x1FD6E8734E9D28DBULL, "vxlt_0_lowtown_docks_cqs_0_0.vlm"},
	{0x1FD707EEC93E1C22ULL, "vxdr_primevaltomb_6ys_0_0.vlm"},
	{0x1FD7E554C6C522F8ULL, "sst_ss_fereldan_m_1.gda"},
	{0x1FD830FD3327B799ULL, "6178886_f.fxe"},
	{0x1FD834BC9EEA8C61ULL, "mcr310_argument6072695.cl"},
	{0x1FD902EEA641F3B8ULL, "vxdr_primevaltomb_2wn_0_1.vlm"},
	{0x1FD97B2B62C74898ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closedtoopen2.ani"},
	{0x1FD9A27D43249D2FULL, "hlac_lt_undercityserialkiller_-1_21z.dds"},
	{0x1FDA0F1A3D007E00ULL, "6103110_m.fxe"},
	{0x1FDA7A7CE4BD877EULL, "vxht_0_hightown_a2_4kf_0_0.vlm"},
	{0x1FDA7E15CD40E043ULL, "1019660822.ogg"},
	{0x1FDAFA185BBF9ACAULL, "6096485_m.fxe"},
	{0x1FDBAE54611DADA3ULL, "gen_im_flw_sta_bethany_03.uti"},
	{0x1FDBB1BC3A92BB7CULL, "vxw_woundedcoast_s8_0_0.vlm"},
	{0x1FDBDFC20303E066ULL, "art\\vfx\\cinematic\\vfx_anders_intro_c.mmh"},
	{0x1FDC7D4D86F32F3AULL, "art\\characters\\playercharacter\\humanmale\\bdy_medd.msh"},
	{0x1FDC8B4D86F34773ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medd.mao"},
	{0x1FDC8F4D86F34D80ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medd.mmh"},
	{0x1FDCA517B4CF7803ULL, "vxca_cavern1_haunted_gj2_0_0.vlm"},
	{0x1FDCAEF37F56F69DULL, "846618941.ogg"},
	{0x1FDCF579453F2FEFULL, "zz_ave_debug6174985.cl"},
	{0x1FDD4B545286B6E4ULL, "vxlt_0_lowtown_d9b_0_0.vlm"},
	{0x1FDE0DF5BA410E81ULL, "gen00fl_dog6106117.cl"},
	{0x1FDE790223F0F56AULL, "vxca_caverns2prc_3yi_0_0.vlm"},
	{0x1FDED7C0037E4A17ULL, "dae000_gamlen6048499.cl"},
	{0x1FDF5A059E8A82C7ULL, "404920284.ogg"},
	{0x1FDF676A61BC01A3ULL, "757786087.ogg"},
	{0x1FDF95974D81048CULL, "6101549_m.fxe"},
	{0x1FE06940023A70C9ULL, "261755205.ogg"},
	{0x1FE08B22D3B40565ULL, "vxht_mansion2b_2ae_0_0.vlm"},
	{0x1FE0A914A1D73F39ULL, "dre000_vampire6034411.cl"},
	{0x1FE12164CEF7E54AULL, "mag130_danzig6018822.cl"},
	{0x1FE13A9856E0FAC0ULL, "plt_cod_itm_sataareth.plo"},
	{0x1FE18B99297BA274ULL, "21911736.ogg"},
	{0x1FE2083B068CD175ULL, "262198250.ogg"},
	{0x1FE2089F5792322EULL, "452716478.ogg"},
	{0x1FE232F3C93209ECULL, "isa311_castillon6092070.cl"},
	{0x1FE36681BB7A1076ULL, "355400898.ogg"},
	{0x1FE409C1620EFFAEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_booster_sa.dds"},
	{0x1FE4559D514F1D71ULL, "320598367.ogg"},
	{0x1FE529F0A18BEE05ULL, "739674357.ogg"},
	{0x1FE5383A8813BAB1ULL, "gen_im_wep_mag_sta_act3_04.uti"},
	{0x1FE53DCB012B14CEULL, "6021274_m.fxe"},
	{0x1FE5585927E336D0ULL, "6060828_m.fxe"},
	{0x1FE5587A2F43515DULL, "vxdr_0_deeproads_cke_0_0.vlm"},
	{0x1FE588056BCA6EE1ULL, "618288099.ogg"},
	{0x1FE58ADBF7E6F205ULL, "423155153.ogg"},
	{0x1FE5F909BE8085E3ULL, "art\\characters\\creatures\\spider\\animation\\wk_p_lf.ani"},
	{0x1FE6C2EFC7547AD1ULL, "109546927.ogg"},
	{0x1FE73FAD7DC691F7ULL, "mer240_jansen6088672.cl"},
	{0x1FE75AC731D08A4BULL, "972459995.ogg"},
	{0x1FE7ABB91D71AD6AULL, "6136686_m.fxe"},
	{0x1FE7B4587CDB0307ULL, "vxlt_0_lowtown_docks_a3_n_bqz_0_0.vlm"},
	{0x1FE8373A74C6C0FEULL, "250215744.ogg"},
	{0x1FE83D65C5161E5FULL, "85831347.ogg"},
	{0x1FE861D37B1AC391ULL, "393578600.ogg"},
	{0x1FE8FE3C6A0D7AA7ULL, "62929788.ogg"},
	{0x1FEAAB3E3D187AB4ULL, "hlac_lt_warehouse_0_7x.dds"},
	{0x1FEAF6E1AB4BF5EEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanl.evt"},
	{0x1FEB463B57944D7BULL, "w_encounter3_48_0.msh"},
	{0x1FEB583B57946BC6ULL, "w_encounter3_48_0.mao"},
	{0x1FEB643B57948065ULL, "w_encounter3_48_0.mmh"},
	{0x1FEB8780F24F7018ULL, "6045577_m.fxe"},
	{0x1FEB8FBFA2FBB110ULL, "hlac_ca_cavern1_haunted_24421_5x.dds"},
	{0x1FEBD19F1905F44FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reactio.gad"},
	{0x1FEC3843514E71A8ULL, "vxht_mansion_n_57s_0_1.vlm"},
	{0x1FECC461985F2B9AULL, "vxw_sundermount_35c_0_0.vlm"},
	{0x1FED33D2CB993855ULL, "hlac_ht_chantry_n_2534_32z.dds"},
	{0x1FED40470A87C3D8ULL, "6164939_m.fxe"},
	{0x1FEE50500E78F4B3ULL, "6083653_m.fxe"},
	{0x1FEF3483CB1DB920ULL, "vxw_bonepitmining_a2_1mz_0_0.vlm"},
	{0x1FEF8C02AB4BAB6EULL, "hlac_lt_gallowstemplar_f_2394_6y.dds"},
	{0x1FF08B7BE5B159ACULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_basic.msh"},
	{0x1FF0917BE5B1639EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_basic.mmh"},
	{0x1FF09D7BE5B17835ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_basic.mao"},
	{0x1FF0EAAD5BA0016DULL, "vxlt_undercitydungeon_1yn_0_0.vlm"},
	{0x1FF0EDE1AB505199ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanl.gad"},
	{0x1FF148826A246E1CULL, "212744811.ogg"},
	{0x1FF1E86172963518ULL, "6126322_m.fxe"},
	{0x1FF24C888C4C6782ULL, "hlac_lt_0_lowtown_docks_-131072_0y.dds"},
	{0x1FF27C96A4479624ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\additive01.ani"},
	{0x1FF2B14033BBC0E8ULL, "245755897.ogg"},
	{0x1FF2E69F190C36FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reactio.evt"},
	{0x1FF30537DE6ADB9CULL, "llac_dr_primevaltomb_663_3.dds"},
	{0x1FF31109BE8B38DAULL, "art\\characters\\creatures\\spider\\animation\\wk_p_lf.evt"},
	{0x1FF4095A4947B418ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_c.mmh"},
	{0x1FF46472E05C76B4ULL, "6109718_m.fxe"},
	{0x1FF4CCEF71B38B90ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01oldgrave.msh"},
	{0x1FF4DAEF71B3A3AAULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01oldgrave.mmh"},
	{0x1FF4DEEF71B3AAE1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01oldgrave.mao"},
	{0x1FF500E8EC62A4AEULL, "vxw_bonepitmining_a2_ph_0_0.vlm"},
	{0x1FF509E210BC3ED9ULL, "6064890_m.fxe"},
	{0x1FF556CDB6451BA9ULL, "lt_0_lowtown_docks_46_0.rml"},
	{0x1FF57EEC42688207ULL, "vxw_bonepitdestroyed_a3_n9_0_0.vlm"},
	{0x1FF58937E757F8D9ULL, "883767691.ogg"},
	{0x1FF5B0603D9A97A6ULL, "6124238_m.fxe"},
	{0x1FF5C953FE98A0F4ULL, "368403350.ogg"},
	{0x1FF6C66DC0A00B5EULL, "vxht_0_hightown_a86_0_0.vlm"},
	{0x1FF6FA6AC298C8DDULL, "vxht_0_hightown_n_7fz_0_0.vlm"},
	{0x1FF74B1023E1555DULL, "110574964.ogg"},
	{0x1FF7894D0B23EB02ULL, "6083602_f.fxe"},
	{0x1FF78B8B4ADB12CFULL, "6141890_m.fxe"},
	{0x1FF7A4C6E5B664D7ULL, "vxht_chantry_2zn_0_0.vlm"},
	{0x1FF7F47A4071CB1CULL, "llac_w_bonepit_6_0.dds"},
	{0x1FF7FEE1AB568DE7ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanl.ani"},
	{0x1FF8279FD5337741ULL, "gen_im_flw_arc_varric_03.uti"},
	{0x1FF838DC68C89F27ULL, "face_face_p4.psh.fxo.dx11"},
	{0x1FF87996A44BFC86ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\additive01.gad"},
	{0x1FF8A26EABB60E01ULL, "mcr370_varric.stg"},
	{0x1FF8CD9C150EF333ULL, "ts_gems_03.gda"},
	{0x1FF970AFBAF137BAULL, "6070610_m.fxe"},
	{0x1FF9856C07DF6BBBULL, "zz_dre_debug6168235.cl"},
	{0x1FF995AD71324A14ULL, "6180821_m.fxe"},
	{0x1FF9EBECF6FB6B09ULL, "gen00fl_merrill6119613.cl"},
	{0x1FFA1009BE9155CDULL, "art\\characters\\creatures\\spider\\animation\\wk_p_lf.gad"},
	{0x1FFA1044FADC4C0CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masb_0t.dds"},
	{0x1FFB22EE513E0F02ULL, "dae000_orsino6085442.cl"},
	{0x1FFB26F0BA25ABB7ULL, "and100_anders6111673.cl"},
	{0x1FFBAD6DB45B7A3EULL, "ceil_dust.mao"},
	{0x1FFC4AD9C49734C3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mb.msh"},
	{0x1FFC58D9C4974C1DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mb.mmh"},
	{0x1FFC5CD9C497532EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mb.mao"},
	{0x1FFC8F61E268DBBFULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.cess.gad"},
	{0x1FFCCA0AAE8FEB35ULL, "hlac_lt_0_lowtown_a2_65536_8y.dds"},
	{0x1FFCFDA0A1DF3418ULL, "vxht_keep_n_6ak_0_0.vlm"},
	{0x1FFD0B3EAA57B80AULL, "vxlt_0_lowtown_a2_n_c_hrt_0_0.vlm"},
	{0x1FFD5B4C94D402FBULL, "6028409_m.fxe"},
	{0x1FFD8A91ACD1EBF1ULL, "lt_0_lowtown_docks_223_0.mmh"},
	{0x1FFD9691ACD1FF92ULL, "lt_0_lowtown_docks_223_0.mao"},
	{0x1FFDA491ACD217AFULL, "lt_0_lowtown_docks_223_0.msh"},
	{0x1FFDDFDF81C95B81ULL, "6081789_m.fxe"},
	{0x1FFDF5855353E08BULL, "585082791.ogg"},
	{0x1FFE104446C0C6DFULL, "6101528_m.fxe"},
	{0x1FFE4485820B6AD1ULL, "art\\characters\\weapons\\spears\\w_qui_qspr01a.msh"},
	{0x1FFE4B0B9B44354CULL, "vxlt_undercityserialkiller_1jc_0_0.vlm"},
	{0x1FFE5E85820B968FULL, "art\\characters\\weapons\\spears\\w_qui_qspr01a.mmh"},
	{0x1FFE916E16E5ACD1ULL, "883040622.ogg"},
	{0x1FFEC86C85B19985ULL, "dae000_bartrand6204163.cl"},
	{0x1FFF2B312C734521ULL, "vxlt_blackemporium_ui_0_0.vlm"},
	{0x1FFF7496A4521365ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\additive01.evt"},
	{0x1FFFC75201AA0285ULL, "vxlt_0_lowtown_eoi_0_1.vlm"},
	{0x1FFFEE9F1916CE15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reactio.ani"},
	{0x1FFFEFA86EDBDD48ULL, "mer240cr_miner4.utc"},
	{0x20002A4BB1560F8CULL, "fx_blood05.mao"},
	{0x2000E07F40CA8492ULL, "sky_lt_gallowstemplar_f.mao"},
	{0x20013A6620F44365ULL, "ave200_aveline_bar6115007.cl"},
	{0x20013BB3E83C4738ULL, "art\\levels\\atmosphere\\sky\\sb_day.mmh"},
	{0x200145D77ED4A263ULL, "444066906.ogg"},
	{0x20016C3D9D585E9EULL, "687480494.ogg"},
	{0x2001B2BC22E3B014ULL, "hlac_lt_0_lowtown_docks_a3_-131074_2x.dds"},
	{0x2001E25D9A9D7943ULL, "vxlt_undercitydungeon_19p_0_0.vlm"},
	{0x20029064AE4EE98FULL, "genpt_party_rand_banter.ncs"},
	{0x2002983C639476BBULL, "llac_lt_blackemporium_1055_10.dds"},
	{0x20037AF6B3210A77ULL, "gen00fl_aveline6122788.cl"},
	{0x20038461E26EE7ACULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.cess.evt"},
	{0x2003BC82F03D6603ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_exit.evt"},
	{0x2003D993F6354AF5ULL, "140149156.ogg"},
	{0x2003ED871B3B0D52ULL, "1059392142.ogg"},
	{0x2003F0C9345E9A87ULL, "vxw_bonepitdestroyed_a3_i8_0_0.vlm"},
	{0x2005EE003EBEC492ULL, "hlac_lt_0_lowtown_a3_n_c_1_6z.dds"},
	{0x20060725B34BF824ULL, "vxht_0_hightown_av2_0_0.vlm"},
	{0x20062AB2AD62EF08ULL, "darktown_male_fereldan_sta6175152.cl"},
	{0x20079225DFBE1A3BULL, "vxht_brothel_n_6jb_0_0.vlm"},
	{0x20081F4A0C3056A8ULL, "atl_shared_dxt1_02.dds"},
	{0x20082D8B60EAF782ULL, "hlac_ca_caverns2prc_5802_60z.dds"},
	{0x2008676CA3884F35ULL, "663048284.ogg"},
	{0x20088B588A328014ULL, "mag210_arianni6030495.cl"},
	{0x200A5D63AC6686D6ULL, "plt_cod_cha_leliana_new.plo"},
	{0x200A9582F0434284ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_exit.gad"},
	{0x200A96068CBE0FA1ULL, "752187430.ogg"},
	{0x200C23FD3C85D238ULL, "99613235.ogg"},
	{0x200D0D1DE286CD5DULL, "vxht_chantry_2ye_0_0.vlm"},
	{0x200D3B9635D66836ULL, "535428197.ogg"},
	{0x200D691C6B1A8D13ULL, "hlac_lt_darktown_-1_15y.dds"},
	{0x200EF1C0739E77C8ULL, "951660689.ogg"},
	{0x200F63FDA5FC2A82ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_elftarp_new_d.dds"},
	{0x200F99BE68534F59ULL, "6150274_m.fxe"},
	{0x200FB85530626F4DULL, "w_sundermount_200_0.mmh"},
	{0x200FBC553062759EULL, "w_sundermount_200_0.mao"},
	{0x200FCA5530628DD3ULL, "w_sundermount_200_0.msh"},
	{0x200FE5D3B7A53A38ULL, "llac_ca_caverns2_5802_58.dds"},
	{0x201048003FC832B3ULL, "vxlt_0_lowtown_docks_a3_n_aos_0_0.vlm"},
	{0x20107AB853D87F11ULL, "sib100_father_contact6110075.cl"},
	{0x2010CCC377FA130AULL, "vxlt_hangedman_n_28d_0_0.vlm"},
	{0x2011261358304FCBULL, "vxw_woundedcoast_oo_0_0.vlm"},
	{0x20116C61E27AFC05ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.cess.ani"},
	{0x2011A482F0497B62ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_exit.ani"},
	{0x2011DAF3E14A3DF3ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_4y.dds"},
	{0x2011EAF5CCFA9B47ULL, "vxlt_darktown_4u3_0_1.vlm"},
	{0x2011EDBA93F6260DULL, "ef_ath_entourage.mor"},
	{0x2011FF2A3EAA53EBULL, "262127386.ogg"},
	{0x2012963145A94F02ULL, "371934791.ogg"},
	{0x2012E614B7D8DF13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_small.evt"},
	{0x20147902A54210A4ULL, "vxht_0_hightown_ay3_0_0.vlm"},
	{0x201494AA5E05FD6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\sebastian_bow_idle.gad"},
	{0x2015144F1B523F73ULL, "357255192.ogg"},
	{0x20159860A9353D2FULL, "ave331_jeven6137279.cl"},
	{0x20168769BE113B28ULL, "vxlt_gallowstemplar_f_55s_0_0.vlm"},
	{0x2016952A7D46183EULL, "gen_im_arm_glv_hvy_act1_02.uti"},
	{0x2016DECAD2025568ULL, "dae300_alistair_king6134922.cl"},
	{0x2017120E60C3E7A1ULL, "mag360_samson6171448.cl"},
	{0x20181099BC7DBB85ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_002a_0d.dds"},
	{0x20184C95E187C0A9ULL, "vxht_brothel_n_3hk_0_0.vlm"},
	{0x2018DF14B7DD3F54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_small.gad"},
	{0x20190EF130B24864ULL, "vxht_chantry_n_356_0_0.vlm"},
	{0x20196BCB66121B5EULL, "vxlt_gallowstemplar_3ju_0_0.vlm"},
	{0x201972C35811E72FULL, "492216564.ogg"},
	{0x201986C82559BEBFULL, "gen00fl_isabela6106170.cl"},
	{0x2019874650D15D65ULL, "vxca_caverns2_7em_0_2.vlm"},
	{0x2019C0050F61AE53ULL, "274992400.ogg"},
	{0x201A3EAAE2F28F3BULL, "vxw_sundermount_20p_0_0.vlm"},
	{0x201A5CD7954E4F1CULL, "hlac_ht_0_hightown_a3_1_1x.dds"},
	{0x201A93B085CC1095ULL, "art\\vfx\\creatures\\vfx_c_drag_roar_c.anb"},
	{0x201AD23E7766EA97ULL, "art\\characters\\playercharacter\\qunarimale\\har_stn.phy"},
	{0x201C3B1E467AA7F2ULL, "6208246_m.fxe"},
	{0x201C84EF87222377ULL, "vxht_brothel_n_6dt_0_0.vlm"},
	{0x201CC8778E743CDFULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinstower.phy"},
	{0x201D058CC8EBDD59ULL, "6043278_f.fxe"},
	{0x201D6C979D28FAADULL, "710641844.ogg"},
	{0x201E1ED7D3650628ULL, "sst_ss_dalish_m_1.gda"},
	{0x201E20FF7A1EB32EULL, "6080492_f.fxe"},
	{0x201F3F6E06F7D1DDULL, "w_sundermount_179_0.mao"},
	{0x201F436E06F7D816ULL, "w_sundermount_179_0.mmh"},
	{0x201F4D6E06F7E914ULL, "w_sundermount_179_0.msh"},
	{0x201F5A9F3D36EFE3ULL, "6116072_m.fxe"},
	{0x201F8C3235846F3AULL, "w_bonepitmining_a2_141_0.mmh"},
	{0x201F903235847671ULL, "w_bonepitmining_a2_141_0.mao"},
	{0x201F9E3235848D80ULL, "w_bonepitmining_a2_141_0.msh"},
	{0x201FAC222F1C1B8FULL, "gen_im_arm_hlm_hvy_act3_01.uti"},
	{0x201FEE14B7E37732ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_small.ani"},
	{0x2020567FC244EE2EULL, "vxw_bonepitmining_a2_ak_0_0.vlm"},
	{0x20205CF8C56D4819ULL, "follower_banter6148720.cl"},
	{0x2020B1553070D60AULL, "w_sundermount_200_0.phy"},
	{0x2020BF64B42E6C02ULL, "qun231_varnell6098600.cl"},
	{0x2021B5CDD557EAF4ULL, "follower_banter6169753.cl"},
	{0x20222BCD3E45E29CULL, "follower_banter6160844.cl"},
	{0x20230B21027EBB87ULL, "dae000_serendipity6161372.cl"},
	{0x20236B75FDC5E19BULL, "lowtown_male_cityguard_mob6150498.cl"},
	{0x202379D2786C0258ULL, "395536627.ogg"},
	{0x2023BC6FB2059AE3ULL, "655572360.ogg"},
	{0x202411F77CE7A326ULL, "vxdr_primevaltomb_767_0_0.vlm"},
	{0x2024FCD9E3288D78ULL, "vxca_cave_23s_0_1.vlm"},
	{0x202566EFAF5D0F65ULL, "283638258.ogg"},
	{0x20263E601331F06EULL, "llac_lt_undercitydungeon_-1_10.dds"},
	{0x20264D2C7B256D7AULL, "mcr310_sibling6076314.cl"},
	{0x20268D3144F883A7ULL, "hlac_dr_primevalentrance_1724_2y.dds"},
	{0x20269677397B55A2ULL, "vxca_caverns2prc_938_0_1.vlm"},
	{0x2026AE843AA76F40ULL, "vxht_mansion2b_214_0_0.vlm"},
	{0x202704E4E3539C6AULL, "qun120_arishok_guard6049150.cl"},
	{0x202737372B92ED83ULL, "66110119.ogg"},
	{0x20275043183AD8BFULL, "111370774.ogg"},
	{0x20275060D2D0998EULL, "6064591_f.fxe"},
	{0x2027835F7051B4B1ULL, "193795638.ogg"},
	{0x2027F23CF45F0A50ULL, "ave310_donnic.cnv"},
	{0x2028AE1F795508A8ULL, "qcr220_orsino6060702.cl"},
	{0x2028B087D4E8A7BBULL, "vxw_sundermount_25s_0_0.vlm"},
	{0x2028BC3B68848674ULL, "6125903_m.fxe"},
	{0x20296FAA5E181A50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\sebastian_bow_idle.ani"},
	{0x20299C5C0A5C2606ULL, "625509555.ogg"},
	{0x2029D45F2D2A24C8ULL, "mag320_orsino6071160.cl"},
	{0x202A57F23504686AULL, "gen00fl_merrill6106270.cl"},
	{0x202A58C5C67D8476ULL, "6002141_m.fxe"},
	{0x202A61D8FC1E06E2ULL, "w_bonepit_30_0_ll.mmh"},
	{0x202A68F2784BDFAAULL, "700823262.ogg"},
	{0x202A83F1168EFF72ULL, "dae000_flemeth6000893.cl"},
	{0x202AF6CC79183BF7ULL, "vxht_brothel_n_6aw_0_0.vlm"},
	{0x202B143CF46179A8ULL, "ave310_donnic.bnk"},
	{0x202B21EBA59FA914ULL, "hlac_ca_cave_2193_0z.dds"},
	{0x202B553054FF4B7EULL, "6094450_m.fxe"},
	{0x202B64A748BB3ACDULL, "108030245.ogg"},
	{0x202B663F3B67B402ULL, "hlac_w_bonepitdestroyed_a3_65542_0x.dds"},
	{0x202B8CE971F890A6ULL, "6125875_m.fxe"},
	{0x202BA915784D092CULL, "vxw_encounter2_camp_px_0_0.vlm"},
	{0x202BC060FAA1E641ULL, "vxht_0_hightown_9v3_0_0.vlm"},
	{0x202C643B57CC176EULL, "w_encounter3_48_0.rml"},
	{0x202C964741DD784BULL, "vxw_sundermount_1yt_0_0.vlm"},
	{0x202CAD1A5F35AAE7ULL, "vxca_caverns1_mines_ilb_0_0.vlm"},
	{0x202CC2920E2E8790ULL, "and100_anders6095806.cl"},
	{0x202D2CABF6164418ULL, "6067387_f.fxe"},
	{0x202D3C03465C5897ULL, "vxw_bonepit_1pt_0_0.vlm"},
	{0x202D51BF93562C7CULL, "43187630.ogg"},
	{0x202D56A9D06B9A7AULL, "vxca_cavern1_haunted_j2d_0_2.vlm"},
	{0x202DA52E56897D10ULL, "376164665.ogg"},
	{0x202DC9778E82B1E4ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinstower.mmh"},
	{0x202DCD778E82B837ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinstower.mao"},
	{0x202DDF778E82D686ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinstower.msh"},
	{0x202E09CD3C1C80CBULL, "vxw_woundedcoast_aa_0_0.vlm"},
	{0x202E652304A80AC1ULL, "dre000_vampire6034349.cl"},
	{0x202E8610C465BB8CULL, "6195856_m.fxe"},
	{0x202EC85CB43B06AEULL, "vxw_bonepitmining_a2_ll_0_0.vlm"},
	{0x202F013476F6C775ULL, "vxht_mansion2_1sx_0_0.vlm"},
	{0x202F1476A3D08B77ULL, "atl_saveload_dxt5_01.dds"},
	{0x202F7A6AFE7A037BULL, "w_encounter2_camp_51_0.rml"},
	{0x202FA5B05A38BD39ULL, "134820476.ogg"},
	{0x202FC4E85DF07E69ULL, "6097870_m.fxe"},
	{0x203021F55C756B2CULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_hed_gib_0d.dds"},
	{0x20302E467839801AULL, "spy121_jalen.stg"},
	{0x2030C89E91AB221AULL, "6045975_m.fxe"},
	{0x20314291F935E334ULL, "6016506_m.fxe"},
	{0x2031687BE5E8B539ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_basic.phy"},
	{0x203193E2EC3B79D8ULL, "6069643_m.fxe"},
	{0x2031E5DF8ED5B0E7ULL, "57596618.ogg"},
	{0x2032040A7F040E69ULL, "w_bonepit_35_0.rml"},
	{0x20326C2F7B7B9C21ULL, "vxdr_primevaltomb_6pv_0_1.vlm"},
	{0x2032737F3E01691CULL, "vxlt_0_lowtown_docks_a3_cug_0_0.vlm"},
	{0x20331633D32B7362ULL, "787749978.ogg"},
	{0x203338F01FE66D5BULL, "6168972_m.fxe"},
	{0x20333D3B57D1F312ULL, "w_encounter3_48_0.phy"},
	{0x2033B8B5DBB9D932ULL, "328822818.ogg"},
	{0x2034A1E878EDA1DEULL, "hlac_lt_0_lowtown_a2_n_-1_2z.dds"},
	{0x203598F2B4CB16D6ULL, "vxlt_0_lowtown_docks_a3_cp6_0_0.vlm"},
	{0x2035C4604CA1D877ULL, "hlac_ht_chantry_n_2534_12z.dds"},
	{0x2035C46F7E25F2FEULL, "126603017.ogg"},
	{0x2035C84A15FE53BBULL, "218110628.ogg"},
	{0x2036816AFE802DD3ULL, "w_encounter2_camp_51_0.phy"},
	{0x20369D4A8EA60236ULL, "6025798_f.fxe"},
	{0x2036ADE1A5E1A7BCULL, "gen00fl_fenris6110824.cl"},
	{0x20375133450D5826ULL, "hlac_lt_gallowscourtyard_-1_0z.dds"},
	{0x20376DD484EC33A0ULL, "vxw_bonepitmining_a2_1vg_0_0.vlm"},
	{0x2037DF06A2E88ACFULL, "art\\vfx\\creatures\\fxc_rage_burst_p.cess.gad"},
	{0x2038349EAFE25332ULL, "vxw_encounter2_camp_nh_0_0.vlm"},
	{0x2038A8EA02D7AD84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\crushing_prison_reaction_db_loop.gad"},
	{0x2038AF4E3A22D881ULL, "and100_anders6095881.cl"},
	{0x2038C05FA7D512E5ULL, "gen00fl_merrill6107456.cl"},
	{0x20392C2AFAEE11A3ULL, "mag210_arianni6196196.cl"},
	{0x20394A39929C2B02ULL, "vxdr_0_deeproads_cx7_0_0.vlm"},
	{0x203968A96B0E5477ULL, "vxht_brothel_n_6jv_0_0.vlm"},
	{0x203A1AF26A4321ACULL, "vxw_encounter2_camp_ky_0_0.vlm"},
	{0x203A3CA11B50B16BULL, "hlac_lt_0_lowtown_a2_n_1_10y.dds"},
	{0x203A4576985213C2ULL, "vxlt_darktown_57t_0_0.vlm"},
	{0x203A99E7BC2A9CA6ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualmode.ani"},
	{0x203AB10FB70456EAULL, "89265460.ogg"},
	{0x203AB572AAC8A9B5ULL, "hlac_lt_0_lowtown_n_0_6y.dds"},
	{0x203AFA90F13B7E61ULL, "dae000_madam_lusine6151488.cl"},
	{0x203B126CE0BFD188ULL, "526329892.ogg"},
	{0x203B224F8FB05659ULL, "zz_firestaff2_forest_v2.cut"},
	{0x203B276879D6C294ULL, "vxca_caverns1_mines_eg0_0_0.vlm"},
	{0x203BC9140D29CA8DULL, "187009574.ogg"},
	{0x203BF80A7F0C4E49ULL, "w_bonepit_35_0.mao"},
	{0x203C040A7F0C62E2ULL, "w_bonepit_35_0.mmh"},
	{0x203C060A7F0C6598ULL, "w_bonepit_35_0.msh"},
	{0x203C4A214BDF1E72ULL, "538660293.ogg"},
	{0x203C6897C0B21A5CULL, "gen_im_wep_rog_arc_lore_03.uti"},
	{0x203CBD68079BFC35ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamdrain.anb"},
	{0x203CDF5122DEFD4AULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsack.mao"},
	{0x203CE35122DF0399ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsack.mmh"},
	{0x203CED5122DF1487ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsack.msh"},
	{0x203D3A39A6311F2EULL, "vxw_bonepitmining_a2_km_0_0.vlm"},
	{0x203DE67A42D5A66BULL, "zz_mag_debug6054589.cl"},
	{0x203E0E3212BCC159ULL, "vxca_caverns2_7th_0_0.vlm"},
	{0x203E823EC9D82788ULL, "319504477.ogg"},
	{0x203EF406A2EECD7CULL, "art\\vfx\\creatures\\fxc_rage_burst_p.cess.evt"},
	{0x203F02A1775C86B9ULL, "350052882.ogg"},
	{0x203FB7EA02DDE662ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\crushing_prison_reaction_db_loop.ani"},
	{0x203FBAA5A01C5AA8ULL, "vxlt_gallowstemplar_n_5bz_0_0.vlm"},
	{0x204016F1BF5B2D6EULL, "546966271.ogg"},
	{0x20401BCA581B00B6ULL, "fen210_dying_magister6108626.cl"},
	{0x2040444E7C7A1572ULL, "hlac_ht_0_hightown_a2_-196607_0y.dds"},
	{0x2040E55C2FF7819CULL, "vxlt_0_lowtown_docks_a3_cxd_0_0.vlm"},
	{0x2040F3E3493C7711ULL, "6044290_m.fxe"},
	{0x20416A49951D68AEULL, "430854587.ogg"},
	{0x2041C128C30DC2D5ULL, "one120_athenril6113296.cl"},
	{0x2042668B4B3CB342ULL, "718285568.ogg"},
	{0x20430507833CFB73ULL, "zz_approval_debug6003686.cl"},
	{0x20430D07147F67B4ULL, "6049814_m.fxe"},
	{0x2043706134FE457FULL, "mag121_idunna6084480.cl"},
	{0x204394B085EF17D6ULL, "art\\vfx\\creatures\\vfx_c_drag_roar_c.mmh"},
	{0x2044503276C08DB4ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.anb"},
	{0x204460C2FE09F3D6ULL, "vxw_bonepit_11v_0_0.vlm"},
	{0x20447B60E7B784F2ULL, "6134723_m.fxe"},
	{0x2044AB3235A3C045ULL, "w_bonepitmining_a2_141_0.phy"},
	{0x2044BB17796264FFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphanga.msh"},
	{0x2044C11779626EA1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphanga.mmh"},
	{0x2044CD1779628342ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphanga.mao"},
	{0x20454DEEB1351735ULL, "833291832.ogg"},
	{0x20463116EC91D2FBULL, "vxlt_0_lowtown_a3_n_c_e10_0_0.vlm"},
	{0x2046791B568DCBA3ULL, "vxlt_0_lowtown_n_by4_0_0.vlm"},
	{0x2046845C5512AFCCULL, "vxlt_undercityserialkiller_1sd_0_0.vlm"},
	{0x2046A67BA1D86BB2ULL, "vxw_encounter2_camp_ii_0_0.vlm"},
	{0x2046C66E457D14E6ULL, "vxw_bonepitmining_a2_21c_0_0.vlm"},
	{0x204781E7BC34FE17ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualmode.evt"},
	{0x2047BAE87F5EB9A2ULL, "art\\characters\\weapons\\spears\\w_spr_002a.msh"},
	{0x2047BC1684924382ULL, "vxdr_0_deeproads_cu0_0_0.vlm"},
	{0x2047BCE87F5EBD58ULL, "art\\characters\\weapons\\spears\\w_spr_002a.mmh"},
	{0x2047C8E87F5ED1FBULL, "art\\characters\\weapons\\spears\\w_spr_002a.mao"},
	{0x2047FD83FEF4D409ULL, "zz_fol_debug6057012.cl"},
	{0x2048B7538A482C42ULL, "vxlt_darktown_58u_0_0.vlm"},
	{0x2048EDB7DD3A908BULL, "351416746.ogg"},
	{0x2048EE82C83A073CULL, "6205694_m.fxe"},
	{0x2048EF120736AC97ULL, "566705435.ogg"},
	{0x20494EBFA5265FF0ULL, "630939615.ogg"},
	{0x204975C3D5BC91F1ULL, "732227475.ogg"},
	{0x204B4067DF48879DULL, "hlac_dr_0_deeproads_5845_1z.dds"},
	{0x204BC05778C1877AULL, "432133783.ogg"},
	{0x204BD0C4B76BA5DBULL, "zz_followers_banter6157980.cl"},
	{0x204BD5222610329EULL, "hlac_w_encounter1_4_0y.dds"},
	{0x204BF3885415700AULL, "143343772.ogg"},
	{0x204BFC06A2F96495ULL, "art\\vfx\\creatures\\fxc_rage_burst_p.cess.ani"},
	{0x204C0487EA2A77A4ULL, "vxht_chantry_n_12m_0_0.vlm"},
	{0x204C152E6E42C977ULL, "vxw_sundermount_20d_0_0.vlm"},
	{0x204C5371600B1B72ULL, "6104730_f.fxe"},
	{0x204C9000A494591CULL, "vxw_woundedcoast_12_0_0.vlm"},
	{0x204CEA9DBE699326ULL, "6077835_m.fxe"},
	{0x204D04A005982632ULL, "6042831_m.fxe"},
	{0x204D10E4022B2D80ULL, "29943281.ogg"},
	{0x204D87B00CFD7865ULL, "1008397360.ogg"},
	{0x204D8D05A60BDFE4ULL, "155528011.ogg"},
	{0x204DD0DEA59FE596ULL, "llac_lt_gallowstemplar_2383_6.dds"},
	{0x204E0E60176F17EDULL, "237784123.ogg"},
	{0x204E23C694FA1BEAULL, "6161332_m.fxe"},
	{0x204E5C2195A418E5ULL, "llac_ca_cavern1_haunted_24421_11.dds"},
	{0x204E82E7BC3B1EA0ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualmode.gad"},
	{0x204EAE6340D96276ULL, "dae300_bran6143500.cl"},
	{0x204EC7E7A9F95076ULL, "6163503_m.fxe"},
	{0x204ED8D4938562FEULL, "sum_pro_04.mao"},
	{0x204F0B7762033BF3ULL, "980872551.ogg"},
	{0x204F4CE8512D7E5CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1_loop.ani"},
	{0x204FB09F1E261E7BULL, "mcr370cr_corpse_3.utc"},
	{0x20501182EE869C62ULL, "vxw_sundermount_274_0_0.vlm"},
	{0x2050449231011451ULL, "vxlt_gallowsprison_n_6bx_0_0.vlm"},
	{0x2050A489DD840669ULL, "vxlt_darktown_50j_0_0.vlm"},
	{0x2050F5888BA0DE3BULL, "one220_yevhen6119967.cl"},
	{0x2050F8093AC475CDULL, "hlac_ht_0_hightown_a2_65536_1x.dds"},
	{0x205118F4E2B58902ULL, "mag360_alain6044615.cl"},
	{0x2051B29BDCB0C007ULL, "vxht_bartrand_n_5rq_0_0.vlm"},
	{0x2051CAF449FA8C39ULL, "w_bonepitmining_a2_171_0.msh"},
	{0x2051D3F8A1A87DB6ULL, "llac_ht_0_hightown_n_-65536_4.dds"},
	{0x2051D4F449FA9D27ULL, "w_bonepitmining_a2_171_0.mmh"},
	{0x2051D8F449FAA474ULL, "w_bonepitmining_a2_171_0.mao"},
	{0x20522F3276CC9377ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.mmh"},
	{0x205238C49EBC2A39ULL, "hlac_lt_darktown_-1_14x.dds"},
	{0x2052C3E87F68D0E3ULL, "art\\characters\\weapons\\spears\\w_spr_002a.phy"},
	{0x2052F97377B06191ULL, "hlac_ht_0_hightown_n_196607_0z.dds"},
	{0x20531000AAAAEC3AULL, "vxht_0_hightown_a2_cbw_0_1.vlm"},
	{0x2053D48DA3ABF206ULL, "vxht_0_hightown_b8e_0_0.vlm"},
	{0x2053FBAFA65EACD9ULL, "vxlt_0_lowtown_n_bxq_0_0.vlm"},
	{0x205504D6EF3E462BULL, "368124735.ogg"},
	{0x205520CF38480568ULL, "llac_lt_gallowscourtyard_0_0.dds"},
	{0x205559E85131FF9EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1_loop.gad"},
	{0x20559870CDEAC79CULL, "6110712_m.fxe"},
	{0x2055B3854864FCEBULL, "vxdr_0_deeproads_d2w_0_0.vlm"},
	{0x2055CFEB84C06D5FULL, "pro000_guard6005643.cl"},
	{0x2055DE3B86E9FE57ULL, "vxlt_0_lowtown_a2_go0_0_0.vlm"},
	{0x2055EED99366109AULL, "vxlt_darktown_2qu_0_0.vlm"},
	{0x20562DF376885C02ULL, "vxdr_0_deeproads_cr1_0_0.vlm"},
	{0x20562E6E07F5530DULL, "vxlt_warehouse_o7_0_0.vlm"},
	{0x205674B0D6FD795FULL, "6033553_m.fxe"},
	{0x2056CCA4FD11CD7CULL, "vxca_caverns1_eh8_0_0.vlm"},
	{0x2057915E7760787FULL, "vxlt_0_lowtown_docks_n_cp7_0_0.vlm"},
	{0x2057AA9C14ED8D19ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_5x.dds"},
	{0x2057B62967C9A377ULL, "seb221cr_ruxton_harimann.utc"},
	{0x20581BF72C006AF3ULL, "vxht_mansion2_3ez_0_0.vlm"},
	{0x2058322AB6BE9F09ULL, "ser270_templar6149987.cl"},
	{0x20585CEF383410E7ULL, "771771728.ogg"},
	{0x20589E31BA529781ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\textures\\windowglowcombined_n.dds"},
	{0x2058EF4A657A6C43ULL, "dre000_bodahn6031351.cl"},
	{0x2059167D9B9A1CFCULL, "llac_ht_keep_n_3731_0.dds"},
	{0x2059339C667741ABULL, "6043543_m.fxe"},
	{0x20594A7299CE1838ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_bdy_hvyo.mao"},
	{0x20595F6411FECA4BULL, "vxw_woundedcoast_pb_0_0.vlm"},
	{0x205A010194FF1253ULL, "dae000cr_grand_cleric6066109.cl"},
	{0x205A4E5931BD1A3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon.ani"},
	{0x205A77AF6DF44BE0ULL, "6134769_m.fxe"},
	{0x205AF1EBF6A8F259ULL, "vxca_caverns2_7nn_0_0.vlm"},
	{0x205AF473E5534028ULL, "697083036.ogg"},
	{0x205B17647EF7DCAAULL, "158028842.ogg"},
	{0x205B789C4F0D076EULL, "vxw_woundedcoast_1ks_0_0.vlm"},
	{0x205B93E859EDC6EAULL, "zz_dre_debug6056986.cl"},
	{0x205C4685BCAF429EULL, "zz_dae_debug6050738.cl"},
	{0x205C6310FFC6CD63ULL, "vxw_sundermount_31r_0_0.vlm"},
	{0x205C7C8FE6A0CFCBULL, "vxw_bonepit_1xh_0_0.vlm"},
	{0x205C95A386F0346AULL, "6166232_m.fxe"},
	{0x205D35CDF539D774ULL, "6144959_m.fxe"},
	{0x205D38636168B1EDULL, "6189399_m.fxe"},
	{0x205D74D91933AF0AULL, "gen00fl_fenris6110839.cl"},
	{0x205D833ACB37DFD1ULL, "vxdr_primevaltomb_6v0_0_1.vlm"},
	{0x205E083CF48CBE9FULL, "ave310_donnic.stg"},
	{0x205E835FE07CB4E2ULL, "vxw_sundermount_285_0_0.vlm"},
	{0x205E9719DAA67D15ULL, "vxht_0_hightown_a3_a6n_0_0.vlm"},
	{0x205ED157609CD886ULL, "plt_lgt000pt_tombs_third.plo"},
	{0x20602478CEA6D887ULL, "vxht_bartrand_n_5up_0_0.vlm"},
	{0x206075F1E3BDEF8CULL, "llac_lt_0_lowtown_a2_n_c_-131072_1.dds"},
	{0x20607A86A9978B04ULL, "w_encounter3_78_0.mmh"},
	{0x20607E86A9979257ULL, "w_encounter3_78_0.mao"},
	{0x20609086A997B0A6ULL, "w_encounter3_78_0.msh"},
	{0x2061A73E77A2A4EFULL, "art\\characters\\playercharacter\\qunarimale\\har_stn.mao"},
	{0x2061B33E77A2B88CULL, "art\\characters\\playercharacter\\qunarimale\\har_stn.mmh"},
	{0x2061B93E77A2C37EULL, "art\\characters\\playercharacter\\qunarimale\\har_stn.msh"},
	{0x206204B5D1AF3D73ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_look_around_sway.ani"},
	{0x2062303BA317708AULL, "6087390_f.fxe"},
	{0x2062686ADC90E585ULL, "hlac_lt_undercitydungeon_-1_39x.dds"},
	{0x20628BE7555AE1E5ULL, "vxht_mansion2b_22l_0_0.vlm"},
	{0x206361EB31A8CF3BULL, "qun111_ketojan6019043.cl"},
	{0x2063C2BE2616D6FEULL, "320377534.ogg"},
	{0x2063FDC722B1321BULL, "vxht_mansion2_1ut_0_0.vlm"},
	{0x20641C14A5465C53ULL, "vxlt_blackemporium_3u_0_0.vlm"},
	{0x206460B6855C291AULL, "vxlt_darktown_2tv_0_0.vlm"},
	{0x206467B28BA7A9ADULL, "vxlt_warehouse_n_14l_0_0.vlm"},
	{0x2064A04AF9EB6B8DULL, "vxlt_warehouse_p6_0_0.vlm"},
	{0x2064C279CEC0066FULL, "gen00fl_fenris6128000.cl"},
	{0x206538A238BBF637ULL, "vxlt_0_lowtown_a2_gy2_0_0.vlm"},
	{0x206567202FC0732EULL, "6043151_m.fxe"},
	{0x206591F7CBA55EBDULL, "darktown_male_lowlife_mob6176180.cl"},
	{0x206593CCB8A3B01DULL, "6179046_m.fxe"},
	{0x2065BE6807BF0376ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamdrain.mmh"},
	{0x20661A3461AC244FULL, "vxlt_0_lowtown_docks_a3_n_cy7_0_0.vlm"},
	{0x20666B1EE9225A1BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_rubblelarge.msh"},
	{0x20667D1EE9227966ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_rubblelarge.mao"},
	{0x2066891EE9228D05ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_rubblelarge.mmh"},
	{0x20670BB6282176C9ULL, "6111992_m.fxe"},
	{0x206725454A740D24ULL, "263115705.ogg"},
	{0x2067665931C7CD0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon.evt"},
	{0x2068920A7C390F8CULL, "hlac_lt_0_lowtown_a3_n_c_65533_3z.dds"},
	{0x20698C6AC40CC556ULL, "6208394_m.fxe"},
	{0x2069A27051221B2AULL, "al_rogue_t06.gda"},
	{0x2069A6ACFFE16A14ULL, "746376423.ogg"},
	{0x2069E5A54F97AD99ULL, "vxw_sundermount_303_0_0.vlm"},
	{0x2069FA9F45EBCF29ULL, "971465294.ogg"},
	{0x206A5446DDAB6734ULL, "dae200cr_viscount6087372.cl"},
	{0x206A5ACA7C6B254DULL, "716295551.ogg"},
	{0x206ABF78B0F8465AULL, "526878556.ogg"},
	{0x206AC6F53C17BCB5ULL, "724036294.ogg"},
	{0x206AE0CF71AABAADULL, "vxht_0_hightown_a2_9ar_0_0.vlm"},
	{0x206B5973A29D44E6ULL, "lt_gallowscourtyard_27_0.rml"},
	{0x206B6ABE02899A2AULL, "dre000_varric6208058.cl"},
	{0x206BE3403A3A62CDULL, "character_character_p114.psh.fxo.dx11.assembly"},
	{0x206CB85EBA914151ULL, "hlac_lt_0_lowtown_65535_6x.dds"},
	{0x206D0D788EDBE2F0ULL, "lgt220_shp_mag_guard6149913.cl"},
	{0x206D2407F77363FDULL, "6068898_m.fxe"},
	{0x206D89B6AC13492AULL, "990175911.ogg"},
	{0x206D9464D44985AAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev3.gad"},
	{0x206D9CCA669AFAD2ULL, "vxlt_blackemporium_3p_0_0.vlm"},
	{0x206DA16B6210B4D4ULL, "hlac_lt_0_lowtown_n_131070_6z.dds"},
	{0x206E10EE817B2C19ULL, "hlac_dr_primevaltomb_664_3z.dds"},
	{0x206E675931CDEDB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon.gad"},
	{0x206E9655C09CF107ULL, "vxht_bartrand_n_5xs_0_0.vlm"},
	{0x206EAE8F57C10630ULL, "zz_dae_debug6029426.cl"},
	{0x206ECBE29B435DEBULL, "53053654.ogg"},
	{0x206EFCB5D1B9B9AAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_look_around_sway.evt"},
	{0x206F068762ED8D5AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2craggyrock.phy"},
	{0x206F50ABD940A149ULL, "828215599.ogg"},
	{0x206F62F007BEED8CULL, "6092267_m.fxe"},
	{0x2070320DBFAE8187ULL, "llac_ca_caverns2_5802_62.dds"},
	{0x207045810A1F7E0BULL, "vxw_woundedcoast_19l_0_0.vlm"},
	{0x207055E09048A029ULL, "89923918.ogg"},
	{0x2070C1545D317EA4ULL, "6149123_m.fxe"},
	{0x2071431DEB05FA2FULL, "862611788.ogg"},
	{0x20725A2C8E2CD07BULL, "hlac_lt_0_lowtown_docks_a3_n_-3_7x.dds"},
	{0x20726FA414A74A9BULL, "vxht_mansion2_1xw_0_0.vlm"},
	{0x207341D298B6D1E7ULL, "llac_w_encounter3_-262140_0.dds"},
	{0x20734E6FD37F5E0EULL, "dae000_dalish_apprentices6147571.cl"},
	{0x2073586ABAE9AA5FULL, "6039753_m.fxe"},
	{0x2073921C098A5D42ULL, "6060192_m.fxe"},
	{0x207469432DE87B94ULL, "vxlt_0_lowtown_n_csh_0_0.vlm"},
	{0x2074890F49A76CFEULL, "scaleform_sf_main_p0.psh.fxo.dx11"},
	{0x2074B9C363390154ULL, "dae000ip_crowd_mf4_drunk2.utp"},
	{0x2074DDFC12E47692ULL, "229455504.ogg"},
	{0x20757554002AD317ULL, "vxw_bonepit_1is_0_0.vlm"},
	{0x2075DBB5D1BFA01DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_look_around_sway.gad"},
	{0x2076CB2DFCA885A9ULL, "mag330_walter6045786.cl"},
	{0x2076CE8A8B13EA40ULL, "vxlt_0_lowtown_docks_n_9bt_0_0.vlm"},
	{0x20785BE1FED8CEF6ULL, "w_sundermount_227_0.msh"},
	{0x20785C5632F9386EULL, "vxw_woundedcoast_1qu_0_0.vlm"},
	{0x207865E1FED8DFF4ULL, "w_sundermount_227_0.mmh"},
	{0x207869E1FED8E607ULL, "w_sundermount_227_0.mao"},
	{0x207A36280E1D983AULL, "876835096.ogg"},
	{0x207AE28816E4FBD1ULL, "6018826_m.fxe"},
	{0x207B40083F49A7C0ULL, "126431232.ogg"},
	{0x207BD02C30599F41ULL, "64468249.ogg"},
	{0x207BFE9C7FAA3CE0ULL, "vxca_caverns2_89l_0_1.vlm"},
	{0x207C10DA961E290BULL, "412214034.ogg"},
	{0x207C75571BE040E9ULL, "6071908_m.fxe"},
	{0x207C7A205B9A5F47ULL, "vxlt_undercitydungeon_19l_0_0.vlm"},
	{0x207CF60798C32BFFULL, "vxlt_0_lowtown_a2_n_gz8_0_0.vlm"},
	{0x207D3CB37B01E0E7ULL, "6109736_f.fxe"},
	{0x207DAEFCC9AE5579ULL, "gen00fl_sebastian6174718.cl"},
	{0x207DB36674C7CE7CULL, "vxw_bonepit_169_0_0.vlm"},
	{0x207DBDF44A1FB15CULL, "w_bonepitmining_a2_171_0.phy"},
	{0x207DD4CCF9B5060CULL, "6097167_m.fxe"},
	{0x207DE1EA6B32FBF2ULL, "vxlt_0_lowtown_a2_n_bz0_0_0.vlm"},
	{0x207EE1E9D387A41AULL, "1051225406.ogg"},
	{0x207F2EBC511D836BULL, "6117992_m.fxe"},
	{0x207F59FD34FE48CDULL, "81503167.ogg"},
	{0x207FBF414FA11606ULL, "6016975_m.fxe"},
	{0x207FF81B54A20E80ULL, "vxca_caverns2prc_8w7_0_1.vlm"},
	{0x208015A6F9981C61ULL, "425053899.ogg"},
	{0x20802A93C22DD606ULL, "6163433_m.fxe"},
	{0x20809773F92EEF80ULL, "zz_jepler_mocap_test.cut"},
	{0x2080E181069D631BULL, "vxht_mansion2_1cv_0_0.vlm"},
	{0x20810E3D29793D6DULL, "gen00fl_anders6144107.cl"},
	{0x20816F64D459EF90ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev3.ani"},
	{0x2081892E417E4687ULL, "hlac_lt_gallowscourtyard_n_131071_1x.dds"},
	{0x20820636C11601CFULL, "6016826_m.fxe"},
	{0x208277869C8FE11DULL, "6177246_m.fxe"},
	{0x2082B9088E3C21F8ULL, "vxca_caverns1_ioe_0_0.vlm"},
	{0x20831E215BFF8503ULL, "llac_lt_0_lowtown_-196607_0.dds"},
	{0x2083D8AD1F0C579FULL, "hlac_lt_darktown_-1_21z.dds"},
	{0x2083EA553BB511BFULL, "756500778.ogg"},
	{0x2084B487941FD853ULL, "mag320cr_crowd_4.utc"},
	{0x2085AE35EC8AACCCULL, "gen00fl_isabela_init6076471.cl"},
	{0x2085B9319BAAEF03ULL, "dae110cr_human_f_01.utc"},
	{0x208608030CE166C3ULL, "6171321_m.fxe"},
	{0x20862E6E12F564EFULL, "6023831_m.fxe"},
	{0x20863D6AB9F6ED9CULL, "vxca_caverns1_mines_j2b_0_2.vlm"},
	{0x208688B01A3C2A39ULL, "qcr000_arishok6059973.cl"},
	{0x2086A71EF1803843ULL, "llac_ca_caverns1_mines_24424_23.dds"},
	{0x208719459E94D291ULL, "6206690_m.fxe"},
	{0x208768DF478E92D0ULL, "vxw_encounter1_c0_0_0.vlm"},
	{0x2087A89FE6AFECE4ULL, "6141444_m.fxe"},
	{0x2087D226BC83194BULL, "vxw_bonepit_1km_0_0.vlm"},
	{0x20882291AD2076B8ULL, "gen00fl_fenris6126375.cl"},
	{0x208832F627480740ULL, "vxlt_hangedman_wj_0_0.vlm"},
	{0x20884D97E752DA60ULL, "6108307_f.fxe"},
	{0x20888922C4FA4E0CULL, "qcr220_meredith6060308.cl"},
	{0x2088954809E7E13BULL, "nat380_sibling6207551.cl"},
	{0x2088C0D7C3B188B4ULL, "vxlt_gallowstemplar_n_2pc_0_0.vlm"},
	{0x2088D79EBB98926AULL, "62623755.ogg"},
	{0x2088E4985CE044B2ULL, "ts_demonic_03.gda"},
	{0x208904CC8384EF6CULL, "t3_com_c02.tnt"},
	{0x20890817921BE17CULL, "vxw_woundedcoast_74_0_0.vlm"},
	{0x208AAC1421C7332FULL, "vxdr_0_deeproads_cmy_0_0.vlm"},
	{0x208B06C93DE730DBULL, "vxht_keep_n_5hh_0_0.vlm"},
	{0x208B1EB436ECBAF8ULL, "hlac_w_blightlands_0_2y.dds"},
	{0x208B9CF41438F1FDULL, "512712924.ogg"},
	{0x208BFBD508309AB6ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgtp_0s.dds"},
	{0x208C4536A215D55EULL, "llac_ca_cave_2194_5.dds"},
	{0x208C7B4DD6C72C70ULL, "darktown_fem_lowlife_mob6174153.cl"},
	{0x208E8BFCBB6DAABDULL, "dae211ip_at_gamlen_to_ltown.utp"},
	{0x208F535DF8937B9BULL, "vxht_mansion2_1fi_0_0.vlm"},
	{0x208FA4DFF0EC7D5FULL, "hlac_w_bonepitdestroyed_a3_0_1x.dds"},
	{0x208FAF4546024515ULL, "315444981.ogg"},
	{0x208FC603D871AC3EULL, "61705690.ogg"},
	{0x20901A5BE0F7E588ULL, "6075189_m.fxe"},
	{0x20902C83C72E95A0ULL, "1068511253.ogg"},
	{0x20913BC8AEECBC9EULL, "710279071.ogg"},
	{0x2091707AF1E01DDBULL, "vxdr_primevalentrance_1o3_0_0.vlm"},
	{0x209172363F99C80EULL, "lt_gallowsprison_n_3567_ll.mmh"},
	{0x2091B87382850118ULL, "art\\characters\\creatures\\arh_abm_rev\\c_rev_shield.mao"},
	{0x2092A47F52D6A1DDULL, "wm_ico_lowtowndocks.dds"},
	{0x2092D9C0363C7B18ULL, "qun111_ketojan6040967.cl"},
	{0x2093101126751427ULL, "follower_banter6151236.cl"},
	{0x20931094162BFC43ULL, "6129355_m.fxe"},
	{0x209382AF3D3525C9ULL, "6056864_m.fxe"},
	{0x2093A98692DBC419ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_aoe_p.impa.ani"},
	{0x2094C8E55D1E9C7FULL, "mag100_cullen6093793.cl"},
	{0x2095093A01CEEA10ULL, "vxlt_darktown_5cj_0_0.vlm"},
	{0x209538B8D083C73FULL, "513935719.ogg"},
	{0x20953EE41DCF3988ULL, "vxca_caverns2prc_a9m_0_0.vlm"},
	{0x20956E6C19EB8DB2ULL, "6013316_m.fxe"},
	{0x209582C7AE1727C1ULL, "6162266_m.fxe"},
	{0x2095C8F9A3759912ULL, "227513183.ogg"},
	{0x2095DABC3984AB50ULL, "vxw_encounter1_h1_0_0.vlm"},
	{0x209650424F522CA3ULL, "vxht_chantry_n_2sz_0_0.vlm"},
	{0x20970ED7C80895C3ULL, "vxlt_0_lowtown_a3_n_c_hnp_0_0.vlm"},
	{0x209721219BEA9769ULL, "vxw_woundedcoast_1r3_0_0.vlm"},
	{0x20974D9F5130E317ULL, "vxht_0_hightown_a2_cb7_0_0.vlm"},
	{0x209836ECB5DF333FULL, "ave331_post_fight6138808.cl"},
	{0x2098DED5207D9F49ULL, "6037407_m.fxe"},
	{0x2099565140B86723ULL, "6206858_m.fxe"},
	{0x2099619AF57B0A8DULL, "958555150.ogg"},
	{0x209977096D83D1F8ULL, "gen00fl_anders6125592.cl"},
	{0x20997A6AFED49F48ULL, "w_encounter2_camp_51_0.mmh"},
	{0x2099866AFED4B3EBULL, "w_encounter2_camp_51_0.mao"},
	{0x2099986AFED4D272ULL, "w_encounter2_camp_51_0.msh"},
	{0x2099BEDBB20C5B58ULL, "vxw_sundermount_35q_0_0.vlm"},
	{0x209A743B1321299FULL, "vxdr_primevaltomb_7ll_0_0.vlm"},
	{0x209A848692E1A45BULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_aoe_p.impa.gad"},
	{0x209B2FAC3532ABFDULL, "vxlt_hangedman_1hw_0_0.vlm"},
	{0x209B9E3CFF9BBEFAULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_meda_n.dds"},
	{0x209BB1813BE908B5ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_nud.mao"},
	{0x209BBC1BF2A18202ULL, "one111_gamlen.cnv"},
	{0x209C3C832131124DULL, "708840855.ogg"},
	{0x209D5F20ABA4AF75ULL, "6133611_m.fxe"},
	{0x209E4644782D0538ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hugekeg_sa.dds"},
	{0x209E5C12095937DAULL, "279089021.ogg"},
	{0x209EE4512331CF86ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsack.phy"},
	{0x209F9E1BF2A5383AULL, "one111_gamlen.bnk"},
	{0x209FDE6AAA1AEE9DULL, "190865547.ogg"},
	{0x20A0A18316719B32ULL, "258670468.ogg"},
	{0x20A0A963CA57DB68ULL, "mag160_feynriel6020483.cl"},
	{0x20A0ABD6D73C9919ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_exit.ani"},
	{0x20A16738D5F22A13ULL, "224911486.ogg"},
	{0x20A18C6C661218CFULL, "dae000_gamlen6102704.cl"},
	{0x20A1A18692E7F440ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_aoe_p.impa.evt"},
	{0x20A1C1323C806541ULL, "hlac_lt_darktown_-1_62y.dds"},
	{0x20A1CE957675E47FULL, "qun110tr_arishok_ambient.utt"},
	{0x20A1DD22EE60AC92ULL, "vxlt_0_lowtown_docks_n_b50_0_0.vlm"},
	{0x20A2FA8D6AE824F1ULL, "vxlt_undercityserialkiller_15s_0_0.vlm"},
	{0x20A352DB9F425CF2ULL, "6113378_f.fxe"},
	{0x20A3B0C10FC55208ULL, "vxca_caverns2prc_a4j_0_0.vlm"},
	{0x20A3EB1CA9BB0367ULL, "gen00fl_varric6093654.cl"},
	{0x20A4067618717AFEULL, "art\\vfx\\combat\\vfx_str_stomp_act_c.impa.ani"},
	{0x20A42FBE0734E9A7ULL, "hlac_lt_0_lowtown_a2_65535_2x.dds"},
	{0x20A51BB141D5E52EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_throat_cut_threat.evt"},
	{0x20A580B4B9FEAE43ULL, "vxlt_0_lowtown_a3_n_c_hks_0_0.vlm"},
	{0x20A5B53ABEFD43C0ULL, "plt_cod_ltr_olivia_wayward.nss"},
	{0x20A5BB41455B9F09ULL, "hlac_lt_0_lowtown_docks_a3_-65538_2x.dds"},
	{0x20A60C4E1E7B2F9CULL, "449581180.ogg"},
	{0x20A617E41DA929DFULL, "vxw_bonepitmining_a2_1wk_0_0.vlm"},
	{0x20A64CFB39253E45ULL, "6016158_m.fxe"},
	{0x20A667F4DBC3FD4AULL, "191379668.ogg"},
	{0x20A73E01BBDC6038ULL, "549738213.ogg"},
	{0x20A75EF3B43727D1ULL, "mag160_feynriel6020101.cl"},
	{0x20A786D6D742795BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_exit.gad"},
	{0x20A7FDE68698BA3CULL, "vxlt_0_lowtown_a3_n_c_i17_0_0.vlm"},
	{0x20A828439CB8C5A3ULL, "hlac_lt_gallowstemplar_n_7198_1z.dds"},
	{0x20A8444D9200B21FULL, "hlac_lt_0_lowtown_a2_-65536_4x.dds"},
	{0x20A84DCFFD6F8F40ULL, "449260863.ogg"},
	{0x20A8563EE5D828E1ULL, "236539612.ogg"},
	{0x20A8E62E5E1B76DCULL, "6162296_m.fxe"},
	{0x20A93B6A7FCEB908ULL, "vxlt_warehouse_155_0_0.vlm"},
	{0x20A9B25760DC8FC5ULL, "plt_lgt000pt_tombs_third.nss"},
	{0x20AAC4CC22344F53ULL, "hlac_dr_0_deeproads_5842_13z.dds"},
	{0x20AADF4AA55B2A43ULL, "vxw_sundermount_36v_0_1.vlm"},
	{0x20AB0241B595CD49ULL, "zz_vault_debug6177791.cl"},
	{0x20AB5626155FEFE4ULL, "6059672_f.fxe"},
	{0x20AB8C5D1BAB39E0ULL, "qcr221_cut_qunari_defeated.cnv"},
	{0x20AC08344AB45405ULL, "968311672.ogg"},
	{0x20AC1296D2F73E1EULL, "hlac_lt_0_lowtown_a2_n_c_65535_4x.dds"},
	{0x20AC12B141DBF4D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_throat_cut_threat.gad"},
	{0x20AC4302983A06BCULL, "vxw_sundermount_30j_0_0.vlm"},
	{0x20AC8CFCBB86AA7CULL, "dae211ip_at_gamlen_to_ltown.ncs"},
	{0x20AC8D12DF390664ULL, "36533783.ogg"},
	{0x20AC9A1628D81F8AULL, "ico_1h_sword_4.dds"},
	{0x20AD1ED108ED337BULL, "6043613_m.fxe"},
	{0x20AD21965FCDDDB1ULL, "983247552.ogg"},
	{0x20AD6FE74659F900ULL, "vxlt_gallowscourtyard_n_3w1_0_0.vlm"},
	{0x20AD722B9E492AE8ULL, "gen_im_trs_qun_horn_02.uti"},
	{0x20ADDE08D03A6FE0ULL, "follower_banter6141566.cl"},
	{0x20AE7E31541852B3ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medp.phy"},
	{0x20AE858977234E10ULL, "gen00fl_fenris6106621.cl"},
	{0x20AEA3D6D748C940ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_exit.evt"},
	{0x20AF45380AF67698ULL, "fen110_party_comment.cnv"},
	{0x20AF5811BDEC635DULL, "art\\vfx\\materials\\chunky\\vfx_c_crsh_pris.mao"},
	{0x20AF6E5D1BAEEF18ULL, "qcr221_cut_qunari_defeated.bnk"},
	{0x20AF86974BD03D21ULL, "vxlt_0_lowtown_a2_n_c_elj_0_2.vlm"},
	{0x20AF99CEE3E49AD0ULL, "ss_merrill.bnk"},
	{0x20B1ECF3E5BB1B10ULL, "vxlt_darktown_5ih_0_0.vlm"},
	{0x20B21E76187DE0CFULL, "art\\vfx\\combat\\vfx_str_stomp_act_c.impa.evt"},
	{0x20B227380AF879A0ULL, "fen110_party_comment.bnk"},
	{0x20B26D9D0A240BD6ULL, "927480132.ogg"},
	{0x20B282E8C0847563ULL, "zz_pro_debug6007866.cl"},
	{0x20B2D74D17256DE2ULL, "dae000_bartrand6016853.cl"},
	{0x20B2E39576847A9EULL, "qun110tr_arishok_ambient.ncs"},
	{0x20B2EA3CBECF6ADBULL, "237414639.ogg"},
	{0x20B3156C907C115FULL, "6005826_m.fxe"},
	{0x20B323B141E23027ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_throat_cut_threat.ani"},
	{0x20B41488C8C49609ULL, "ser270_emeric6201571.cl"},
	{0x20B441BA2E4FEF4CULL, "dae300_cullen6140640.cl"},
	{0x20B48F2DF7160CC7ULL, "vxca_cavern1_haunted_e82_0_0.vlm"},
	{0x20B4FE5C6A77AA9EULL, "hlac_w_blightlands_0_5z.dds"},
	{0x20B50E6B71624BCBULL, "hlac_lt_0_lowtown_n_-196608_0x.dds"},
	{0x20B5A9E8A1919A98ULL, "434763494.ogg"},
	{0x20B5BF1A9AE75E56ULL, "ht_mansion2b_818.rml"},
	{0x20B6006175119D50ULL, "6175890_m.fxe"},
	{0x20B626726BCEF5A1ULL, "669342533.ogg"},
	{0x20B63AB8BD676B55ULL, "6110995_m.fxe"},
	{0x20B65276972A579AULL, "6040824_m.fxe"},
	{0x20B757666E64586EULL, "862465087.ogg"},
	{0x20B757F4F70D5A9FULL, "vxdr_primevaltomb_7bn_0_0.vlm"},
	{0x20B7580B50118F5CULL, "6167196_m.fxe"},
	{0x20B81A051EE98C71ULL, "mrl310_party_comments6142788.cl"},
	{0x20B855DA6DBD9F7DULL, "6116046_f.fxe"},
	{0x20B91F7618840178ULL, "art\\vfx\\combat\\vfx_str_stomp_act_c.impa.gad"},
	{0x20B9232948859CBBULL, "vxw_sundermount_2km_0_0.vlm"},
	{0x20B95108684D3E06ULL, "6120090_m.fxe"},
	{0x20B9C95D9FC277A4ULL, "lt_gallowscourtyard_43_0.anb"},
	{0x20B9E569BA737B70ULL, "6137711_m.fxe"},
	{0x20BA201B1E7AD072ULL, "vxlt_0_lowtown_a2_gx8_0_0.vlm"},
	{0x20BAC76D12873294ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_imp_p.impa.ani"},
	{0x20BB9024A1D14666ULL, "6124735_m.fxe"},
	{0x20BBC106B3157B8EULL, "vxlt_0_lowtown_docks_a3_n_d3s_0_0.vlm"},
	{0x20BBE9AF9ACC78FCULL, "vxca_caverns1_e53_0_0.vlm"},
	{0x20BC317887954191ULL, "gen000fl_merrill_pchouse.stg"},
	{0x20BCFAEBE35A603FULL, "gen00fl_anders6125724.cl"},
	{0x20BDB7AF6F845FD1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_9.pwk"},
	{0x20BE76B471F0258DULL, "vxlt_blackemporium_136_0_0.vlm"},
	{0x20BEAF0D24019EFDULL, "vxw_encounter2_camp_4a_0_0.vlm"},
	{0x20BEBF11C394FEFDULL, "46594960.ogg"},
	{0x20BF16F232C831E7ULL, "vxlt_0_lowtown_docks_a3_n_cz6_0_0.vlm"},
	{0x20BF8942D9125233ULL, "bloomblend.psh.fxo"},
	{0x20BFB0ADFB39024BULL, "mag100ip_gravestone.ncs"},
	{0x20BFF5F4A33D8F95ULL, "6089947_m.fxe"},
	{0x20C02AAEE7A314DDULL, "dae000ip_crowd_em2_talk2.utp"},
	{0x20C05ED0D7B13390ULL, "vxlt_darktown_5ri_0_0.vlm"},
	{0x20C08B32C8861765ULL, "555406952.ogg"},
	{0x20C0C0A4BEABA00DULL, "dae000ip_crowd_ff7_talk14.utp"},
	{0x20C0FF698CCD3C0EULL, "vxht_keep_n_db_0_0.vlm"},
	{0x20C19F9B8B7DF6D1ULL, "lt_gallowscourtyard_33_0.rml"},
	{0x20C1C14BD6268AA0ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_4x.dds"},
	{0x20C1C46D128D4CF6ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_imp_p.impa.gad"},
	{0x20C1EFDCC7ACAA75ULL, "67621056.ogg"},
	{0x20C2171B69C511CFULL, "mcr370_meredith6083573.cl"},
	{0x20C24AD1872818C5ULL, "6139070_f.fxe"},
	{0x20C276049298A825ULL, "vxlt_warehouse_n_q8_0_0.vlm"},
	{0x20C2C30A4958A889ULL, "plt_cod_ltr_flora_diary.plo"},
	{0x20C31FCE51EEA800ULL, "vfx0_atm.psh.fxo"},
	{0x20C3A020BAA05312ULL, "vxca_caverns2_96g_0_0.vlm"},
	{0x20C3AA55A560B55AULL, "6153917_m.fxe"},
	{0x20C3D77C3CA4EA14ULL, "vxlt_0_lowtown_a2_hc8_0_0.vlm"},
	{0x20C3FBBE847D4B46ULL, "404789431.ogg"},
	{0x20C4EA24CECF2F4AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_exit.gad"},
	{0x20C50668ACE657BBULL, "710129957.ogg"},
	{0x20C5993B476DFD92ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_ext_c.mmh"},
	{0x20C7911F60800A89ULL, "dae000_sandal6147818.cl"},
	{0x20C795063A7BB53BULL, "vxw_sundermount_2ll_0_0.vlm"},
	{0x20C8DF6D12939955ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_imp_p.impa.evt"},
	{0x20C8E74A39F69C71ULL, "vxlt_backalley_n_1pp_0_1.vlm"},
	{0x20C95FE23163575FULL, "gen00fl_sebastian6174019.cl"},
	{0x20C9D0603C5CAB78ULL, "vxht_mansion_n_349_0_0.vlm"},
	{0x20C9EDD72CAED695ULL, "6110407_m.fxe"},
	{0x20C9F76FCE87EA35ULL, "art\\vfx\\textures\\vfx_half_dot.dds"},
	{0x20C9FCAB20673D3AULL, "826135287.ogg"},
	{0x20CA141E3C088FE4ULL, "509919272.ogg"},
	{0x20CA7FBADB2B0433ULL, "lt_0_lowtown_a2_n_42_0_ll.mmh"},
	{0x20CA80133E9F218FULL, "gen00fl_varric6105485.cl"},
	{0x20CAFF87633BDF83ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2craggyrock.msh"},
	{0x20CB0D87633BF7DDULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2craggyrock.mmh"},
	{0x20CB1187633BFEEEULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2craggyrock.mao"},
	{0x20CB21EB60305A61ULL, "follower_banter6169163.cl"},
	{0x20CB2E5D5B87A266ULL, "fx_blood03.mao"},
	{0x20CB7A89820851BAULL, "qun160_vashothlead6053931.cl"},
	{0x20CB8CC3EC35BBFCULL, "251393175.ogg"},
	{0x20CBC62E742691D5ULL, "739395441.ogg"},
	{0x20CC02B8FC94DBB9ULL, "vxht_chantry_35p_0_0.vlm"},
	{0x20CCE89163E63E0DULL, "vxlt_blackemporium_165_0_0.vlm"},
	{0x20CD35576971DF44ULL, "hlac_ca_caverns1_24427_0x.dds"},
	{0x20CD587C9D1404A6ULL, "mag110_vincento6017677.cl"},
	{0x20CD921DC45F2873ULL, "vxw_encounter3_bz_0_0.vlm"},
	{0x20CDE069728193ABULL, "qun221_petrice6075979.cl"},
	{0x20CEF479D0D05935ULL, "vxw_bonepit_21r_0_0.vlm"},
	{0x20CF8601D17C1CC2ULL, "574403371.ogg"},
	{0x20D0A4F141669009ULL, "35420913.ogg"},
	{0x20D104CC7AD368F0ULL, "1007330512.ogg"},
	{0x20D107FA744A0CC0ULL, "ip_restrict_status.ncs"},
	{0x20D1A0D88F276989ULL, "ser120_ghyslain6108669.cl"},
	{0x20D20BE20DACB320ULL, "t1_mul_pu1.tnt"},
	{0x20D249592E9B0294ULL, "vxlt_0_lowtown_a2_hh9_0_0.vlm"},
	{0x20D2C829816D2C2BULL, "882704497.ogg"},
	{0x20D2C8B86CCCC4E5ULL, "540024521.ogg"},
	{0x20D340A4015A63DBULL, "vxw_bonepitmining_a2_1wv_0_1.vlm"},
	{0x20D354DD1428C0D1ULL, "mer231_brekker.bnk"},
	{0x20D3C0B3B3476139ULL, "251302778.ogg"},
	{0x20D3CDFC4A29A575ULL, "6165003_m.fxe"},
	{0x20D4B269A5CCFFF0ULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_bdy_ogra_0s.dds"},
	{0x20D4B3BBCBDB8DC4ULL, "vxca_caverns1_gko_0_0.vlm"},
	{0x20D4DC415C9DB6C1ULL, "vxlt_0_lowtown_n_byn_0_0.vlm"},
	{0x20D5330696180A20ULL, "6209852_m.fxe"},
	{0x20D58F0DC7D5FB5DULL, "6128880_m.fxe"},
	{0x20D5A28B629C4259ULL, "6147159_m.fxe"},
	{0x20D606E32C71CDBBULL, "vxw_sundermount_2qo_0_0.vlm"},
	{0x20D636DD142AC3E1ULL, "mer231_brekker.cnv"},
	{0x20D66BAF17B61ED6ULL, "vxw_bonepitmining_a2_1xs_0_0.vlm"},
	{0x20D6D9F8D6E53721ULL, "mag100_karras6034755.cl"},
	{0x20D6E337E71B7C4AULL, "6018538_f.fxe"},
	{0x20D70FB9942E64F4ULL, "mcr370_meredith6136188.cl"},
	{0x20D7579C06219DD3ULL, "6070101_m.fxe"},
	{0x20D7A43F0C90CD7BULL, "sst_ss_ragedemon_1.gda"},
	{0x20D7D2487D30716BULL, "853444029.ogg"},
	{0x20D7F1E761546390ULL, "hlac_w_sundermount_-131076_0z.dds"},
	{0x20D8581FCF988065ULL, "gen00fl_fenris6108088.cl"},
	{0x20D8DB66AEBAF1F3ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_lyrium.phy"},
	{0x20D98CDC51491A69ULL, "47539586.ogg"},
	{0x20D9B9B9A07C3DCFULL, "6069910_m.fxe"},
	{0x20D9C524CEE14C30ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_exit.ani"},
	{0x20D9E613EFB06DCCULL, "6147972_m.fxe"},
	{0x20D9ED1EDBC16819ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplintera.mmh"},
	{0x20D9F71EDBC17907ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplintera.msh"},
	{0x20DA2647737C7FDFULL, "vxlt_0_lowtown_a2_n_c_f56_0_0.vlm"},
	{0x20DA3A05CA303C04ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_mage_meredithsintro.gad"},
	{0x20DB5F1A3786A082ULL, "42767930.ogg"},
	{0x20DB762FEDC19B62ULL, "mrl000_marethari6129357.cl"},
	{0x20DB94C0A3010F10ULL, "rdr111dg_assistant6050918.cl"},
	{0x20DBA6377B20D24DULL, "120462965.ogg"},
	{0x20DC314CC0D59967ULL, "6085267_f.fxe"},
	{0x20DC7964698D6F85ULL, "6129065_m.fxe"},
	{0x20DCD9D97FAE4F50ULL, "vxlt_hangedman_24a_0_0.vlm"},
	{0x20DD26536A523D85ULL, "266463488.ogg"},
	{0x20DD3D7795164054ULL, "498611591.ogg"},
	{0x20DDB5ADFB520836ULL, "mag100ip_gravestone.utp"},
	{0x20DDD246731AD771ULL, "6075966_m.fxe"},
	{0x20DE0D943053CB4BULL, "6116366_m.fxe"},
	{0x20DEDEA3C8AF80EFULL, "hlac_ht_0_hightown_a2_65537_0z.dds"},
	{0x20DF29B16C035E68ULL, "docks_male_elven_worker_sta6175848.cl"},
	{0x20DF648486F0656FULL, "gen00fl_isabela_init6076618.cl"},
	{0x20DFB16DF7C1404AULL, "hlac_ht_keep_n_4519_15y.dds"},
	{0x20E001A55AEDF333ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopeb.phy"},
	{0x20E0EEE0BE62B671ULL, "ms_wispy.mao"},
	{0x20E14905CA3674E2ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_mage_meredithsintro.ani"},
	{0x20E1780A1E4BC0D7ULL, "vxlt_gallowstemplar_n_653_0_0.vlm"},
	{0x20E1AB2F1959F41BULL, "vxlt_gallowscourtyard_4r1_0_0.vlm"},
	{0x20E2AD8BCCEFA41FULL, "vxdr_primevaltomb_7uk_0_0.vlm"},
	{0x20E2FA4EF026A5F3ULL, "549034992.ogg"},
	{0x20E2FA8A132E925AULL, "6149415_m.fxe"},
	{0x20E3234B552573A2ULL, "22170439.ogg"},
	{0x20E3AE3ABF329883ULL, "plt_cod_ltr_olivia_wayward.plo"},
	{0x20E3D6B2B35BA803ULL, "dae300_alistair_king6133522.cl"},
	{0x20E3FC8820F79400ULL, "zz_ave_debug6056774.cl"},
	{0x20E46E355C1490BBULL, "ss_slavers_m_1.bnk"},
	{0x20E478C01E67E63BULL, "vxw_sundermount_2zn_0_0.vlm"},
	{0x20E487D7CC8B3E2AULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyn_0s.dds"},
	{0x20E497164BB44E4DULL, "vxdr_primevalentrance_2do_0_0.vlm"},
	{0x20E56365E75837B3ULL, "6205639_m.fxe"},
	{0x20E563D4395513BDULL, "isa111_isabela6067118.cl"},
	{0x20E634BE579FB257ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urva_0t.dds"},
	{0x20E65F51D1C9810DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal.mmh"},
	{0x20E66351D1C9885EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal.mao"},
	{0x20E67151D1C99F93ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal.msh"},
	{0x20E6ABBFF1189C96ULL, "vxdr_primevaltomb_7kp_0_0.vlm"},
	{0x20E6D43C68D111CAULL, "qcr220_meredith6041538.cl"},
	{0x20E70D52989E93DBULL, "vxw_bonepit_25x_0_0.vlm"},
	{0x20E720074159256DULL, "vxlt_0_lowtown_docks_b4f_0_1.vlm"},
	{0x20E774C0AED7C24DULL, "vxht_0_hightown_n_aum_0_0.vlm"},
	{0x20E7D73D2EEA835CULL, "6071710_m.fxe"},
	{0x20E85847532DB594ULL, "gen_im_arm_cht_hvy_act0_01.uti"},
	{0x20E86596FC617AB6ULL, "6101337_f.fxe"},
	{0x20E93281CD487883ULL, "127430703.ogg"},
	{0x20E973D2C0DA6D56ULL, "art\\levels\\races\\design\\blendtree\\gasbarrel.anb"},
	{0x20EA28EE1A119D73ULL, "376273018.ogg"},
	{0x20EAF9ACD3D5D6F7ULL, "vxw_sundermount_2ky_0_0.vlm"},
	{0x20EB0FDD591A6DC6ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_dothard.dds"},
	{0x20EB4BB671A467D0ULL, "vxlt_hangedman_21n_0_0.vlm"},
	{0x20EB58F1178C5004ULL, "vxht_mansion2b_210_0_0.vlm"},
	{0x20EBA7CAF451292BULL, "213137884.ogg"},
	{0x20EBEAC0E87E7CE7ULL, "478734637.ogg"},
	{0x20EBFE1284344144ULL, "815959227.ogg"},
	{0x20EC74CB96F03C55ULL, "plt_zz_mer130pt_when_and_wh.plo"},
	{0x20EC7BF26A1277FBULL, "zz_tanner_docks.cut"},
	{0x20ECFD97A8B3E140ULL, "6135771_m.fxe"},
	{0x20EE1018F0211348ULL, "652569269.ogg"},
	{0x20EE983B47910139ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_ext_c.anb"},
	{0x20EEEC61E426D012ULL, "vxht_0_hightown_a4n_0_0.vlm"},
	{0x20EEFE44A2FF8C76ULL, "vxlt_gallowstemplar_f_5j4_0_0.vlm"},
	{0x20EF0F4F109C1240ULL, "vxlt_gallowstemplar_4qq_0_0.vlm"},
	{0x20EF8908431713E9ULL, "vxlt_darktown_5kw_0_0.vlm"},
	{0x20EFBC1A9B18D10AULL, "ht_mansion2b_818.anb"},
	{0x20EFBFA6A71CF013ULL, "tbe300im_nowish_ring.uti"},
	{0x20EFE7062FE81624ULL, "dae000_arishok6095660.cl"},
	{0x20F11CB11C04EFA6ULL, "hlac_lt_0_lowtown_a2_n_65535_4z.dds"},
	{0x20F16F66D9B59478ULL, "6103464_m.fxe"},
	{0x20F19231ABD7D173ULL, "vxht_0_hightown_n_bxs_0_0.vlm"},
	{0x20F1A41B64A20544ULL, "1038205409.ogg"},
	{0x20F1E39E381D791EULL, "54985670.ogg"},
	{0x20F2511B49737260ULL, "follower_banter6148482.cl"},
	{0x20F295118E34BE78ULL, "pn_eye_nrma_0d.dds"},
	{0x20F2D9FAD562C7D1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basic.mmh"},
	{0x20F2E5FAD562DC72ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basic.mao"},
	{0x20F2F3FAD562F38FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basic.msh"},
	{0x20F308F33DAA66CDULL, "vxdr_primevalentrance_2kn_0_0.vlm"},
	{0x20F34F68FBA24FD6ULL, "vxw_bonepitmining_a2_1ru_0_0.vlm"},
	{0x20F39753BD1931AEULL, "gen00fl_fenris6113953.cl"},
	{0x20F3F0C60485C878ULL, "572708308.ogg"},
	{0x20F434314C00A4D5ULL, "qun210wp_arishok_marker.utp"},
	{0x20F46CA55B57A839ULL, "35937103.ogg"},
	{0x20F472499D8304E3ULL, "6055579_m.fxe"},
	{0x20F4DB32A8E0D8CFULL, "307059874.ogg"},
	{0x20F541B224302527ULL, "vxlt_0_lowtown_a2_n_c_csi_0_0.vlm"},
	{0x20F5CA33ABA9623CULL, "vxlt_0_lowtown_docks_a3_d25_0_0.vlm"},
	{0x20F5FE6E7D383460ULL, "gen00fl_isabela6123224.cl"},
	{0x20F61A97996B3537ULL, "406301933.ogg"},
	{0x20F673A45A158E95ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow1win.phy"},
	{0x20F6EAACFA3FA493ULL, "6147658_m.fxe"},
	{0x20F6FDC551B36A22ULL, "hlac_ht_keep_3732_6x.dds"},
	{0x20F6FF7CD48A0270ULL, "778314068.ogg"},
	{0x20F79FE2E9F46D1AULL, "mag210_arianni6021248.cl"},
	{0x20F80A935353A275ULL, "6118528_m.fxe"},
	{0x20F8965B83F550C7ULL, "85696841.ogg"},
	{0x20F8D591ADBEB82EULL, "1073638611.ogg"},
	{0x20F931D24370615CULL, "art\\vfx\\creatures\\fxc_ft_dirt_l_c.impa.ani"},
	{0x20F96B89C5CBEF77ULL, "vxw_sundermount_2lx_0_0.vlm"},
	{0x20F9AA5565FE79E3ULL, "art\\characters\\playercharacter\\elffemale\\rob_appb.mao"},
	{0x20F9ADC3DD7ACFA4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerf.msh"},
	{0x20F9AE5565FE8030ULL, "art\\characters\\playercharacter\\elffemale\\rob_appb.mmh"},
	{0x20F9BC5565FE984AULL, "art\\characters\\playercharacter\\elffemale\\rob_appb.msh"},
	{0x20F9BFC3DD7AEE0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerf.mao"},
	{0x20F9C3C3DD7AF546ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerf.mmh"},
	{0x20F9C7343685C825ULL, "ht_brothel_n_3368.rml"},
	{0x20F9EB47E4075160ULL, "ht_0_hightown_44_0.rml"},
	{0x20F9FB75E1884A56ULL, "vxdr_primevaltomb_2t4_0_0.vlm"},
	{0x20FAE4E324ACBDFDULL, "gen00fl_fenris6116568.cl"},
	{0x20FAE85D9FFA4304ULL, "lt_gallowscourtyard_43_0.rml"},
	{0x20FB610E4C33DD4BULL, "vxw_bonepit_13e_0_0.vlm"},
	{0x20FB6A810C856EA0ULL, "ave310_donnic6134783.cl"},
	{0x20FB7F26396645A7ULL, "vxca_cavern1_haunted_jgt_0_0.vlm"},
	{0x20FBB78BBEE0C037ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagviscount_sa.dds"},
	{0x20FBCA7C2A464506ULL, "6064039_m.fxe"},
	{0x20FBD6021FDD290CULL, "vxlt_darktown_6i7_0_0.vlm"},
	{0x20FCBB96C2D61C26ULL, "character_character_p155.psh.fxo.dx11"},
	{0x20FD67DE460FC262ULL, "vxw_sundermount_2s8_0_0.vlm"},
	{0x20FD6A1B3131B8BAULL, "dae021_sister_alema_mob6144915.cl"},
	{0x20FE94A9E588FD7AULL, "art\\characters\\creatures\\dragonling\\animation\\blendtree\\dragonling.anb"},
	{0x20FEECCC24947196ULL, "128511556.ogg"},
	{0x20FF3ED24374E29EULL, "art\\vfx\\creatures\\fxc_ft_dirt_l_c.impa.gad"},
	{0x20FFBC9280719AE9ULL, "gen00fl_fenris6115356.cl"},
	{0x20FFF088693E0598ULL, "6094676_m.fxe"},
	{0x2100128A4AEADAC5ULL, "6118589_f.fxe"},
	{0x21009D78B78934F2ULL, "mag131_keran6094506.cl"},
	{0x2100BB1AB30B8B81ULL, "mag170_hugh6095618.cl"},
	{0x2100D317639A6C2CULL, "plt_prm_cod_evra_belt.plo"},
	{0x2100D8A8C8B15DE8ULL, "6088152_m.fxe"},
	{0x2101E3AA386170FFULL, "vxlt_0_lowtown_a2_n_bxk_0_0.vlm"},
	{0x2101EA9CB74AEF65ULL, "371680273.ogg"},
	{0x21020F8D99D26770ULL, "1017207099.ogg"},
	{0x2102C50A29062FE8ULL, "gen_im_trs_gen_hvy_cht_r0.uti"},
	{0x2102E953AD134058ULL, "vxlt_darktown_2qg_0_0.vlm"},
	{0x2103D2E70E861D10ULL, "6121222_m.fxe"},
	{0x21041270EEF779BAULL, "881352768.ogg"},
	{0x21051DC8D4A92896ULL, "lt_0_lowtown_docks_a3_n_133_0.phy"},
	{0x21055C89EC35BD13ULL, "6140758_m.fxe"},
	{0x2105EB089A6561B4ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\runforward.ani"},
	{0x210619D2437AC2BDULL, "art\\vfx\\creatures\\fxc_ft_dirt_l_c.impa.evt"},
	{0x2106231137337E3AULL, "287885031.ogg"},
	{0x2106B3D11D90FBB4ULL, "vxdr_0_deeproads_8wc_0_0.vlm"},
	{0x21079A2049BE9D74ULL, "646056887.ogg"},
	{0x2107AC4362FD9961ULL, "vxlt_darktown_4ip_0_0.vlm"},
	{0x2107F0D6795D3C54ULL, "cd_mcr371_intro.bnk"},
	{0x210821979460D79CULL, "33123945.ogg"},
	{0x21086DAEC74EE11FULL, "hlac_ca_caverns2_5802_19z.dds"},
	{0x21087607C719E075ULL, "206635595.ogg"},
	{0x2109916EDC79455DULL, "6162990_m.fxe"},
	{0x2109ED4FC91DF8D3ULL, "qcr000_arishok6166989.cl"},
	{0x210A1EA6EA4D0AFEULL, "vxlt_warehouse_n_12a_0_0.vlm"},
	{0x210ABDF8715AF588ULL, "vxw_woundedcoast_1ga_0_0.vlm"},
	{0x210BD84337BEF412ULL, "gen00fl_carver6108928.cl"},
	{0x210BD9BB3805DAE2ULL, "vxw_sundermount_2t9_0_0.vlm"},
	{0x210BED75322FA315ULL, "vxht_0_hightown_a3_ajr_0_0.vlm"},
	{0x210C6CC2270344E9ULL, "vxlt_darktown_5qu_0_0.vlm"},
	{0x210C89E8A04DDFB3ULL, "6001898_f.fxe"},
	{0x210C90209846C992ULL, "6080617_m.fxe"},
	{0x210CE52BEE0A0A51ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw4sm.msh"},
	{0x210CE8089A6B7B16ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\runforward.gad"},
	{0x210CF32BEE0A226CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw4sm.mao"},
	{0x210CFF2BEE0A360FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw4sm.mmh"},
	{0x210D31673909A67CULL, "ser270_emeric6111482.cl"},
	{0x210D3407055015AAULL, "6071521_m.fxe"},
	{0x210D4936CF13A4D4ULL, "zz_fab_jump.ncs"},
	{0x210E75EB8FC40273ULL, "vxht_0_hightown_n_bnu_0_0.vlm"},
	{0x210EE65DFDBA90B5ULL, "6075463_m.fxe"},
	{0x210F034C745F55B5ULL, "gen00fl_aveline6139939.cl"},
	{0x210F06C8D4B15637ULL, "lt_0_lowtown_docks_a3_n_133_0.msh"},
	{0x210F18C8D4B1749AULL, "lt_0_lowtown_docks_a3_n_133_0.mao"},
	{0x210F1CC8D4B17BA9ULL, "lt_0_lowtown_docks_a3_n_133_0.mmh"},
	{0x210F39BCD983AFC6ULL, "vxht_keep_n_73r_0_0.vlm"},
	{0x210F8B2606F542B8ULL, "6090034_m.fxe"},
	{0x210FBA21D6343276ULL, "lt_gallowscourtyard_n_35_0.rml"},
	{0x21102BD31B01BB1FULL, "lt_0_lowtown_a2_n_28_0.rml"},
	{0x21110E9AEA27B258ULL, "dae130ip_at_darktown.utp"},
	{0x211134F924524548ULL, "vxlt_darktown_4iu_0_0.vlm"},
	{0x21117731546CC428ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medp.mmh"},
	{0x211179E89818281DULL, "w_woundedcoast_108_0.mao"},
	{0x21117DE898182F56ULL, "w_woundedcoast_108_0.mmh"},
	{0x21118331546CD8CBULL, "art\\characters\\playercharacter\\humanfemale\\boo_medp.mao"},
	{0x211187E898184054ULL, "w_woundedcoast_108_0.msh"},
	{0x21119531546CF752ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medp.msh"},
	{0x2111A69A89057BC7ULL, "vxlt_0_lowtown_a3_n_c_hnl_0_0.vlm"},
	{0x2111F6B18085CC3EULL, "6036454_m.fxe"},
	{0x21120156C6FAE616ULL, "vxdr_primevaltomb_7zs_0_0.vlm"},
	{0x211208DE942218B7ULL, "6200108_m.fxe"},
	{0x2112295233CEB984ULL, "6088072_f.fxe"},
	{0x2112D08340DE1113ULL, "590861332.ogg"},
	{0x2112EC3CF7A89059ULL, "vrc321_spooky_scenes6216683.cl"},
	{0x21131B99B8DC3A0BULL, "gen_im_wep_rog_dua_act3_03.uti"},
	{0x211361C9122D445EULL, "390550372.ogg"},
	{0x2113CDBBE82CB2AEULL, "hlac_w_woundedcoast_65539_0y.dds"},
	{0x21145736F8120D16ULL, "6124021_m.fxe"},
	{0x2114D36C8A966767ULL, "6072598_m.fxe"},
	{0x2114D56EEB1CF762ULL, "772624594.ogg"},
	{0x2114D57EDBB77138ULL, "8017109.ogg"},
	{0x211530C9217FBA40ULL, "one220_yevhen6103381.cl"},
	{0x21153B65B23D92C4ULL, "vxlt_0_lowtown_docks_a3_d0k_0_0.vlm"},
	{0x211551DFA4362ED3ULL, "6163771_m.fxe"},
	{0x21159494F41898C9ULL, "mcr310_argument6072725.cl"},
	{0x211623DFE640DEC5ULL, "6125928_m.fxe"},
	{0x21164F43A9B82077ULL, "vxw_sundermount_2zz_0_0.vlm"},
	{0x21167449570F0A23ULL, "lt_0_lowtown_n_51_0.rml"},
	{0x21167A9B418537B7ULL, "6167025_m.fxe"},
	{0x2116D3C2CEDCDBB9ULL, "lav_river01mat.mao"},
	{0x2116FAF5A283C789ULL, "985403725.ogg"},
	{0x211774ACE40BF19CULL, "vxht_playershouse_n_1ua_0_0.vlm"},
	{0x21178A6F5E985B29ULL, "mag121_eau_banter6107386.cl"},
	{0x2117A81A10474788ULL, "6044682_m.fxe"},
	{0x2117C58619773FF6ULL, "6077736_m.fxe"},
	{0x21180B00B5ABF6B2ULL, "6120843_m.fxe"},
	{0x21189E74D7F184F1ULL, "6111178_m.fxe"},
	{0x21189F46E8E74B2EULL, "378487544.ogg"},
	{0x2118B3B3F6DF2AECULL, "vxca_caverns2_7eb_0_0.vlm"},
	{0x21190CF7C4247B13ULL, "870868853.ogg"},
	{0x211938162FEAE403ULL, "764140823.ogg"},
	{0x21195C22253D27A0ULL, "mcr371_scared_mage_m6087697.cl"},
	{0x2119FEF5D028E151ULL, "879051744.ogg"},
	{0x211A5F522425BB95ULL, "vxht_0_hightown_a3_awq_0_0.vlm"},
	{0x211A9A205AD59F56ULL, "vxht_playershouse_n_1p4_0_0.vlm"},
	{0x211B034337037658ULL, "llac_lt_0_lowtown_docks_a3_n_131071_3.dds"},
	{0x211B43516272A97CULL, "vxw_bonepitmining_a2_27c_0_0.vlm"},
	{0x211B81F6B13AC440ULL, "6117670_m.fxe"},
	{0x211B8475FF826BFBULL, "6135156_m.fxe"},
	{0x211BAC0E7DF3679EULL, "6077335_f.fxe"},
	{0x211BC39839991992ULL, "6020093_m.fxe"},
	{0x211C5CAED45E984FULL, "vxw_encounter2_qi_0_1.vlm"},
	{0x211C67F70002C3BAULL, "lt_0_lowtown_a2_28_0.anb"},
	{0x211D1409C224DB0BULL, "vxht_mansion2b_330_0_0.vlm"},
	{0x211D93493FBF2D4EULL, "35376206.ogg"},
	{0x211E0F7E891C5162ULL, "6130001-mrl3_mir_humanswer1.cut"},
	{0x211EA1FA535CBC7BULL, "docks_scene_dockworkers6176476.cl"},
	{0x211EC6ACE14B2C70ULL, "vxht_0_hightown_cf4_0_0.vlm"},
	{0x211F3C93683E9D46ULL, "699928506.ogg"},
	{0x211F7ABD481F9B17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_exit.gad"},
	{0x21200D1B393EB7B0ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg02a.mmh"},
	{0x212018777AFB9447ULL, "vxlt_0_lowtown_a3_n_c_hko_0_0.vlm"},
	{0x21201B1B393ECFCAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg02a.msh"},
	{0x21207333B8F0FE96ULL, "vxdr_primevaltomb_7wt_0_0.vlm"},
	{0x2121217F2A422647ULL, "dae000_flemeth6013267.cl"},
	{0x2121E0696BCD5940ULL, "hlac_ht_keep_3731_3x.dds"},
	{0x2121FAE7BBB7E722ULL, "442230179.ogg"},
	{0x212361652B82FF84ULL, "97603148.ogg"},
	{0x2123A177AEF95A6BULL, "mi_gla_add.mao"},
	{0x2123CA7AAB660C24ULL, "6070832_m.fxe"},
	{0x2123FADAB9534206ULL, "vxlt_warehouse_n_14s_0_0.vlm"},
	{0x21249F819A549785ULL, "vxca_caverns1_i7d_0_0.vlm"},
	{0x2124C506B2531C4EULL, "501835150.ogg"},
	{0x2124DC8C680B7191ULL, "follower_banter6173047.cl"},
	{0x2124E2E1DAEF9393ULL, "dae121_bran_st.stg"},
	{0x2124FECBAB33E628ULL, "llac_ca_cavern1_haunted_24423_5.dds"},
	{0x212579E7919E106AULL, "ht_0_hightown_a3_49_0.rml"},
	{0x2125E294EBC49446ULL, "28385862.ogg"},
	{0x21260EA51B679A10ULL, "lt_0_lowtown_docks_n_183_0.mao"},
	{0x212612A51B67A123ULL, "lt_0_lowtown_docks_n_183_0.mmh"},
	{0x212620A51B67B97DULL, "lt_0_lowtown_docks_n_183_0.msh"},
	{0x2126C3D427120C90ULL, "458940480.ogg"},
	{0x21273F954C48A56DULL, "fen110_fenris6118138.cl"},
	{0x2127ACD30964B057ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwstraight_1.mao"},
	{0x2127D7FC73734E0BULL, "one350_variel6087037.cl"},
	{0x2127F768FB08D4B8ULL, "14914270.ogg"},
	{0x212828DFACABBB6CULL, "266548742.ogg"},
	{0x21283D8D733C5643ULL, "vxw_woundedcoast_1c2_0_0.vlm"},
	{0x212881A1BE14AD7AULL, "6117953_m.fxe"},
	{0x2128FE62E6764678ULL, "vxca_cave_2ia_0_1.vlm"},
	{0x21290BFD4CCBB7D6ULL, "vxht_playershouse_n_1w5_0_0.vlm"},
	{0x212956F049E3EF0BULL, "150804377.ogg"},
	{0x212967CB9723D432ULL, "plt_zz_mer130pt_when_and_wh.nss"},
	{0x2129FB2DD8C4A8C5ULL, "199301365.ogg"},
	{0x212A4A5876CE4B69ULL, "gen_im_trs_tmp_one_axe_r2.uti"},
	{0x212A732D4BA4F525ULL, "fxa_pois_smoke.mao"},
	{0x212A73F79F565306ULL, "vxht_0_hightown_a2_9am_0_0.vlm"},
	{0x212A90394A1CA2AFULL, "123731518.ogg"},
	{0x212AB077D2CB4AF5ULL, "6143732_f.fxe"},
	{0x212AE28ED1575BABULL, "qun111ip_at_lowtown.utp"},
	{0x212B0B5DF15A6C8DULL, "hlac_lt_gallowscourtyard_n_-2_2y.dds"},
	{0x212B3B09C8480F02ULL, "plt_one100pt_loose_end_athe.plo"},
	{0x212B5C355C5074B9ULL, "ss_slavers_m_1.wbd"},
	{0x212B619DE7835B6FULL, "870293875.ogg"},
	{0x212B88046CDB234CULL, "ave310_brennan.stg"},
	{0x212BD041652E7E95ULL, "vxlt_gallowstemplar_n_2e5_0_0.vlm"},
	{0x212C567887F1610BULL, "mag320_evelina6051877.cl"},
	{0x212CA20C72602A7CULL, "vxw_bonepitmining_a2_1s8_0_0.vlm"},
	{0x212CC63436B1204DULL, "ht_brothel_n_3368.anb"},
	{0x212D8D916A17B997ULL, "139435150.ogg"},
	{0x212DC624F3A5C24FULL, "gen_im_trs_tmp_temp_armor.uti"},
	{0x212E02051CF369E3ULL, "850049883.ogg"},
	{0x212F4727B4F1470EULL, "lowtown_male_alienage1_sta6162769.cl"},
	{0x21304B838F310762ULL, "641457320.ogg"},
	{0x2130CE3F849084B4ULL, "vxlt_0_lowtown_n_eja_0_0.vlm"},
	{0x2130D053FA212A51ULL, "229716499.ogg"},
	{0x2130FD64CDFC9870ULL, "6137500_m.fxe"},
	{0x21319503F05D65DCULL, "276214943.ogg"},
	{0x21324568687F5B88ULL, "llac_lt_0_lowtown_a2_-4_0.dds"},
	{0x2132493EE9AFD31BULL, "dae300_bran6143455.cl"},
	{0x2132C15E6C29642EULL, "6095347_m.fxe"},
	{0x2132FFC6CF4D9A2EULL, "gen_im_arm_glv_med_champion.uti"},
	{0x2133A1DCDABC49BDULL, "vxht_mansion2b_1ch_0_0.vlm"},
	{0x2133A7BD4830909DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_exit.ani"},
	{0x213446195E87F697ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16ma.msh"},
	{0x213458195E8815FAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16ma.mao"},
	{0x21345C195E881C09ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16ma.mmh"},
	{0x21345EAB64AD01DCULL, "826379056.ogg"},
	{0x2134C84AD2847B4BULL, "873170786.ogg"},
	{0x2134D475B8B92C68ULL, "hlac_lt_warehouse_-65535_0z.dds"},
	{0x213552576920B38EULL, "l_dae120ar_hightown.wbd"},
	{0x21362C523D9C9A0AULL, "6121581_m.fxe"},
	{0x213651BE71012EF8ULL, "root_guardian.dds"},
	{0x21368F2B48120C6AULL, "64289542.ogg"},
	{0x2136AF6A65326EC3ULL, "vxw_woundedcoast_1n5_0_0.vlm"},
	{0x2136FF3ABD781DA1ULL, "838304636.ogg"},
	{0x21374F60C597C4A2ULL, "337563533.ogg"},
	{0x213783AD722ED1E9ULL, "mrl150tr_guard_talk.ncs"},
	{0x2137C227F91DE66CULL, "376576705.ogg"},
	{0x21389C753BE95711ULL, "307177022.ogg"},
	{0x2138FC87E3C12D7BULL, "988032010.ogg"},
	{0x213970A0A832D1E6ULL, "46907884.ogg"},
	{0x2139B5556634704BULL, "art\\characters\\playercharacter\\elffemale\\rob_appb.phy"},
	{0x213A138973F591D4ULL, "art\\vfx\\combat\\vfx_r_var_rain_arrow_act_c.anb"},
	{0x213A9B55F615E889ULL, "6107472_m.fxe"},
	{0x213AD466AF0DB068ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_lyrium.mmh"},
	{0x213AE066AF0DC40BULL, "art\\levels\\races\\primeval\\props\\relics\\prp_lyrium.mao"},
	{0x213AF266AF0DE292ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_lyrium.msh"},
	{0x213B83005837E47AULL, "follower_banter6172805.cl"},
	{0x213BEE1EDC143406ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplintera.phy"},
	{0x213BEF8589A9359BULL, "6190988_m.fxe"},
	{0x213C05F03F83A18DULL, "mag330st_walter_3.stg"},
	{0x213C4A3AB05D3222ULL, "follower_banter6155859.cl"},
	{0x213C840FEAA07679ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_partition.msh"},
	{0x213C8E0FEAA08767ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_partition.mmh"},
	{0x213C920FEAA08DB4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_partition.mao"},
	{0x213CA833D993AAF5ULL, "gen_im_arm_hlm_hvy_act3_03.uti"},
	{0x213D2B233C31A8ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\wideslash.evt"},
	{0x213D4126EF9BF371ULL, "hlac_lt_backalley_n_0_5y.dds"},
	{0x213D86A45A51B1B0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow1win.msh"},
	{0x213D94A45A51C9CAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow1win.mmh"},
	{0x213D98A45A51D001ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow1win.mao"},
	{0x213D9DE2A600FD6EULL, "6087692_m.fxe"},
	{0x213E0D86773D9459ULL, "610369109.ogg"},
	{0x213E16EAD6F6BE34ULL, "626979881.ogg"},
	{0x213E2C523CB6E33DULL, "lgt210_magu_selby6142032.cl"},
	{0x213E6EC4A9D90935ULL, "6163153_m.fxe"},
	{0x213E789E804DBD57ULL, "ico_1h_axe_2.dds"},
	{0x213E976B9F2CD4C9ULL, "237986255.ogg"},
	{0x213EE7DBD0690825ULL, "hlac_lt_gallowstemplar_n_2394_2z.dds"},
	{0x213F0AA40092AE32ULL, "gen00fl_varric6088864.cl"},
	{0x213FC1B105495303ULL, "dae000_pc_home_ambients6208588.cl"},
	{0x21409281ACEA1392ULL, "6046927_m.fxe"},
	{0x21409F339A11AC7FULL, "w_blightlands_157_0.phy"},
	{0x2140A99A56267630ULL, "dae000_gamlen6144350.cl"},
	{0x214118D0D076C6BFULL, "hlac_lt_undercitydungeon_-1_20x.dds"},
	{0x214147B3E1AF6017ULL, "vxw_bonepitmining_a2_1lr_0_0.vlm"},
	{0x2141744B4AB1A599ULL, "hlac_w_sundermount_65533_0z.dds"},
	{0x2141A6CB5A64BC9FULL, "dae_warrior_aoe_duration.ncs"},
	{0x214213B9CCB2623DULL, "vxht_mansion2b_1fg_0_0.vlm"},
	{0x2142372EAE4D0D59ULL, "6096323_m.fxe"},
	{0x214280AA131EF584ULL, "6095990_m.fxe"},
	{0x2142AD65EF684CD6ULL, "hlac_w_bonepit_262146_0y.dds"},
	{0x2142FAA55B4264A8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopeb.mmh"},
	{0x2142FC6C1E2B6BE1ULL, "hlac_ht_bartrand_n_3370_1x.dds"},
	{0x214306A55B42794BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopeb.mao"},
	{0x214318A55B4297D2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopeb.msh"},
	{0x21436B57FB534D3FULL, "vxca_caverns2_7em_0_0.vlm"},
	{0x2143D64EA935B874ULL, "661519967.ogg"},
	{0x214444233C37F26CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\wideslash.gad"},
	{0x21448E5BEE0FCF1FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7ic.msh"},
	{0x2144945BEE0FD9C1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7ic.mmh"},
	{0x2144A05BEE0FEE62ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7ic.mao"},
	{0x2145214757288743ULL, "vxw_woundedcoast_1i4_0_0.vlm"},
	{0x214536FAC643FA59ULL, "lt_0_lowtown_a2_n_c_21_0_lc.mmh"},
	{0x2145BFF17E0D0254ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_rblt.anb"},
	{0x214659E867C5CF93ULL, "gen00fl_aveline6132828.cl"},
	{0x21465A40D4AC22E8ULL, "vxlt_gallowscourtyard_n_7aw_0_0.vlm"},
	{0x2146F28973FFE397ULL, "art\\vfx\\combat\\vfx_r_var_rain_arrow_act_c.mmh"},
	{0x21473242DC58691CULL, "follower_banter6164736.cl"},
	{0x2147CC2745FE317BULL, "6160608_m.fxe"},
	{0x21480AD679934C06ULL, "cd_mcr371_intro.wbd"},
	{0x2148186A920C1B25ULL, "lt_0_lowtown_docks_42_0.rml"},
	{0x214822A37461BA74ULL, "775571794.ogg"},
	{0x21484D666CAAF05DULL, "zz_fol_debug6098071.cl"},
	{0x21486CAD723D1DCCULL, "mrl150tr_guard_talk.utt"},
	{0x2148982DD253AF84ULL, "6050200_f.fxe"},
	{0x2148B721D663F22AULL, "lt_gallowscourtyard_n_35_0.anb"},
	{0x2149379589E3937DULL, "554068500.ogg"},
	{0x2149CADAB68CCCFCULL, "llac_lt_0_lowtown_docks_a3_n_-1_2.dds"},
	{0x2149E4FEEF2F486CULL, "6027453_f.fxe"},
	{0x214A036E30927F49ULL, "hlac_w_bonepitmining_a2_327680_0y.dds"},
	{0x214A10904C9693A9ULL, "vxlt_gallowscourtyard_7ad_0_1.vlm"},
	{0x214A13A31D3FD254ULL, "dae000_hate_enemies6167626.cl"},
	{0x214A5E691DFE0AECULL, "llac_lt_backalley_n_0_18.dds"},
	{0x214AD11A0A1961DEULL, "243827347.ogg"},
	{0x214AFFAABC5C36BAULL, "816343371.ogg"},
	{0x214B101026BA9EE6ULL, "vxlt_0_lowtown_a2_n_f51_0_0.vlm"},
	{0x214B43233C3E0F6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\wideslash.ani"},
	{0x214C0D6032493E16ULL, "vxlt_0_lowtown_a3_n_c_gnj_0_0.vlm"},
	{0x214D588DCB3ED194ULL, "dae000_maren6148416.cl"},
	{0x214DBE74AF18A795ULL, "ser200_aveline6201548.cl"},
	{0x214E09A70F7C9E3AULL, "vxlt_undercityserialkiller_1je_0_0.vlm"},
	{0x214E09E948A2FAF0ULL, "dae100cr_bartrand.utc"},
	{0x214E161370B5A10AULL, "6155829_m.fxe"},
	{0x214E2DA747F98C02ULL, "lt_gallowscourtyard_n_43_0_lc.mmh"},
	{0x214E52E3F2920BC8ULL, "185893822.ogg"},
	{0x214E9EF0FF7FD9BEULL, "art\\levels\\races\\qunari\\props\\military\\prp_quanarithron_0.pwk"},
	{0x214ED4A8D2E9F22FULL, "art\\cinematics\\additional animation\\cs_femnobleii_cinematicanim.gad"},
	{0x214F045E9EE02C12ULL, "6041827_m.fxe"},
	{0x214F448D164E9E7FULL, "vxht_0_hightown_8q1_0_0.vlm"},
	{0x214F4BD753BBE833ULL, "hlac_lt_0_lowtown_a2_131071_4y.dds"},
	{0x214F6AF7002E22E6ULL, "lt_0_lowtown_a2_28_0.rml"},
	{0x21504ED31B37DA83ULL, "lt_0_lowtown_a2_n_28_0.anb"},
	{0x215090CA97408C56ULL, "6083616_m.fxe"},
	{0x2151799CF2580D95ULL, "dre000_varric6208019.cl"},
	{0x2151A0339A202084ULL, "w_blightlands_157_0.mmh"},
	{0x2151A2F3A9116345ULL, "228480544.ogg"},
	{0x2151A4339A2027D7ULL, "w_blightlands_157_0.mao"},
	{0x2151B12EF46843B3ULL, "gen00fl_bethany6097593.cl"},
	{0x2151B6339A204626ULL, "w_blightlands_157_0.msh"},
	{0x2151E2FBD98E7E2EULL, "hlac_w_woundedcoast_-196603_0y.dds"},
	{0x2151FA55E0BBCA7BULL, "22604223.ogg"},
	{0x215225B59B229CA6ULL, "gen_im_npc_wep_one_macb.uti"},
	{0x21524E6B9F347A7CULL, "713394331.ogg"},
	{0x21526E12F763B64FULL, "dae000_arishok6067624.cl"},
	{0x2152F6F807DF8F9EULL, "vxw_bonepitmining_a2_8p_0_0.vlm"},
	{0x21539324491E9FC3ULL, "vxw_woundedcoast_1t7_0_0.vlm"},
	{0x2153FF528718BA5BULL, "vxlt_backalley_n_1l7_0_0.vlm"},
	{0x21541263043579A9ULL, "6094772_m.fxe"},
	{0x2154455629DF1F41ULL, "hlac_lt_blackemporium_1055_10z.dds"},
	{0x215457739B334AE6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic_1.mao"},
	{0x2154CC1DC6A23B68ULL, "vxlt_gallowscourtyard_n_7dt_0_0.vlm"},
	{0x2154EB83E1B58D22ULL, "zz_vault_debug6203257.cl"},
	{0x2155122767400B0FULL, "art\\vfx\\combat\\vfx_glb_ele_p_extreme_c.mmh"},
	{0x215524E1A346A9A0ULL, "315477252.ogg"},
	{0x21557B48F3BCB1B6ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.dura.evt"},
	{0x2155BDC579A4EE4AULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_bdy_0d.dds"},
	{0x215685D8817799BCULL, "189526933.ogg"},
	{0x2156F35200CE3C6CULL, "1003673028.ogg"},
	{0x2156F8CB837AB95CULL, "gen00fl_fenris6115320.cl"},
	{0x215739FE158700F0ULL, "dae000ip_crowd_ff4_talk10.utp"},
	{0x2158749776B89721ULL, "6196711_m.fxe"},
	{0x2159129F04C99F24ULL, "mag360_thrask6053525.cl"},
	{0x215942321BBC2166ULL, "vxlt_gallowstemplar_f_2p9_0_0.vlm"},
	{0x2159912657F452FEULL, "llac_lt_gallowstemplar_2393_1.dds"},
	{0x2159A23403E62CA5ULL, "192118986.ogg"},
	{0x2159B188F93657AEULL, "vxht_mansion_n_33z_0_0.vlm"},
	{0x215A04E444C83E87ULL, "hm_apostate1.mor"},
	{0x215A65FBFE85B146ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_flourish_additive.evt"},
	{0x215A90333944FE74ULL, "vxw_bonepitdestroyed_a3_27e_0_0.vlm"},
	{0x215AE73199FAD241ULL, "vxlt_0_lowtown_a2_n_c_env_0_0.vlm"},
	{0x215B3974DC9CAF0CULL, "hlac_lt_undercityserialkiller_-1_50y.dds"},
	{0x215B3B1AF8FBB58AULL, "ch1notnt_s1y.psh.fxo"},
	{0x215B4DBF47C9D976ULL, "zz_prmpt_debug6198226.cl"},
	{0x215B8982225A7C06ULL, "171749694.ogg"},
	{0x215C1D4B9FD35622ULL, "w_encounter3_63_0.msh"},
	{0x215C1F4B9FD359D8ULL, "w_encounter3_63_0.mmh"},
	{0x215C2B4B9FD36E7BULL, "w_encounter3_63_0.mao"},
	{0x215C5E05598EC67BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_kd_enter.ani"},
	{0x215C7248F3C2C141ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.dura.gad"},
	{0x215CFD3C090119D1ULL, "6039386_m.fxe"},
	{0x215D15D93B274B5EULL, "943174817.ogg"},
	{0x215D454870CF4E15ULL, "hlac_ca_caverns2_5802_12y.dds"},
	{0x215D814F47B571F3ULL, "hlac_ht_keep_n_3734_1y.dds"},
	{0x215E0A09D04D98BAULL, "6208601_f.fxe"},
	{0x215E5C2D6D8ED226ULL, "6060726_f.fxe"},
	{0x215E90046D068A37ULL, "ave310_brennan.bnk"},
	{0x215ED6BE5A253911ULL, "vxlt_warehouse_e1_0_0.vlm"},
	{0x215F6659BE2E4B16ULL, "hlac_w_bonepitmining_a2_262147_0z.dds"},
	{0x2160141B3974A7CBULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg02a.phy"},
	{0x21604CB7B4B71100ULL, "docks_male_elven_worker_sta6175835.cl"},
	{0x21606F6FC1BA68CDULL, "6141136_f.fxe"},
	{0x2160AB050550DF3EULL, "25364787.ogg"},
	{0x216103E893FFF504ULL, "382253577.ogg"},
	{0x216108465056DE6AULL, "gen00fl_aveline6121926.cl"},
	{0x216148576945DDBCULL, "l_dae120ar_hightown.bnk"},
	{0x216165427419A4D2ULL, "gen_im_trs_rdr_one_lsw_r2.uti"},
	{0x21617CFBFE8BF691ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_flourish_additive.gad"},
	{0x2162712F790ED2DBULL, "vxlt_backalley_n_1w6_0_0.vlm"},
	{0x216272046D0A404FULL, "ave310_brennan.cnv"},
	{0x216273072244212DULL, "183745478.ogg"},
	{0x216312E3B1DABC47ULL, "llac_ca_cavern1_haunted_24424_18.dds"},
	{0x21637348F3C8E1EFULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.dura.ani"},
	{0x216398F9C36EFB83ULL, "zz_lgt_1_debug6136995.cl"},
	{0x2163C5814EB28A83ULL, "6125412_m.fxe"},
	{0x2163D9F48444B6DDULL, "vxca_caverns1_mines_dpq_0_0.vlm"},
	{0x2163F1A8D2FC7FF5ULL, "art\\cinematics\\additional animation\\cs_femnobleii_cinematicanim.ani"},
	{0x2164AD393CFA455FULL, "6156393_m.fxe"},
	{0x2164C2299B367468ULL, "vxca_caverns1_mines_j4b_0_0.vlm"},
	{0x21652B273FDBCE5BULL, "vxw_blightlands_11n_0_0.vlm"},
	{0x21656C883576A311ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.impa.ani"},
	{0x21658D86E0194898ULL, "52075555.ogg"},
	{0x2165C2CD7979679FULL, "llac_lt_0_lowtown_docks_-4_6.dds"},
	{0x2165DC7D7A68781EULL, "nat380_sibling6103575.cl"},
	{0x21664070EE718410ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_roba_0d.dds"},
	{0x21668667DA668126ULL, "gen00fl_varric6077879.cl"},
	{0x2166FFDCE7309DDAULL, "vxlt_0_lowtown_n_erg_0_0.vlm"},
	{0x21672ACC05D93F6FULL, "cd_ser270_cut_motherreveal.wbd"},
	{0x2167686233A2BEB3ULL, "lt_0_lowtown_docks_a3_n_26_0_ll.mmh"},
	{0x216803E7785D3167ULL, "one220_yevhen6098776.cl"},
	{0x2168770FEAC59A9CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_partition.phy"},
	{0x21687DFBFE92173FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_flourish_additive.ani"},
	{0x2168B4B89335079AULL, "ht_0_hightownexplosion_n_26_0_ll.mmh"},
	{0x2168F11A16356F16ULL, "vxlt_0_lowtown_a3_n_c_gxl_0_0.vlm"},
	{0x216902102B3B16F4ULL, "vxw_bonepitdestroyed_a3_22j_0_0.vlm"},
	{0x21692C92A7ED2931ULL, "mag170_thrask6020547.cl"},
	{0x2169335BD2521731ULL, "6024559_m.fxe"},
	{0x216991B090780455ULL, "193922335.ogg"},
	{0x2169FE9ABE676ED1ULL, "lla_ht_0_hightown_a2_-196608_0.dds"},
	{0x216A0B1124294939ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.mmh"},
	{0x216A102E9DCDCD9FULL, "1009636673.ogg"},
	{0x216A24660A19E4F9ULL, "art\\characters\\creatures\\mubari\\animation\\rdy_old.gad"},
	{0x216A77D1C456EF36ULL, "24151663.ogg"},
	{0x216B20362E22E983ULL, "vxlt_gallowscourtyard_n_62h_0_0.vlm"},
	{0x216B6283562313D5ULL, "llac_lt_0_lowtown_a2_n_0_3.dds"},
	{0x216BB49D98BB7AA1ULL, "llac_lt_blackemporium_1055_7.dds"},
	{0x216C7788357CD4B3ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.impa.gad"},
	{0x216CF9413D01020AULL, "317921582.ogg"},
	{0x216D44723F4CE2E5ULL, "6196724_m.fxe"},
	{0x216D4EE37F4A8EF2ULL, "vxdr_primevalentrance_f7_0_1.vlm"},
	{0x216D51EACC74D872ULL, "335792094.ogg"},
	{0x216D6950A294ABBDULL, "vxht_mansion2b_1wd_0_0.vlm"},
	{0x216D7EF9D2769456ULL, "vxht_0_hightown_a3_c3e_0_1.vlm"},
	{0x216DA18702A93A02ULL, "hla_ht_0_hightown_n_-131071_0y.dds"},
	{0x216E1559C37BDCC8ULL, "lt_0_lowtown_n_27_0.anb"},
	{0x216E2B022C917A61ULL, "tbe300im_final_staff.uti"},
	{0x216E7C7F1791A99FULL, "273741733.ogg"},
	{0x216EF10ED2DBB32AULL, "756727132.ogg"},
	{0x216F4D724A82DFD3ULL, "6094489_f.fxe"},
	{0x216F66BE08B25032ULL, "character_character_p276.psh.fxo.dx11.assembly"},
	{0x216F72B5C88AB751ULL, "primitive_wd.mao"},
	{0x216FFD524BE521D3ULL, "6142208_m.fxe"},
	{0x2170352A18504AD9ULL, "157665524.ogg"},
	{0x217035660A1E6D47ULL, "art\\characters\\creatures\\mubari\\animation\\rdy_old.ani"},
	{0x21708505599FB105ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_kd_enter.gad"},
	{0x2170955BEE35163EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7ic.phy"},
	{0x2170CF4C0077C0D3ULL, "and231cr_lyrium_smuggler2.utc"},
	{0x2170F640BA721618ULL, "vxw_encounter2_jf_0_0.vlm"},
	{0x217172B8749840D2ULL, "hlac_dr_primevaltomb_665_1x.dds"},
	{0x2171AB13E264349CULL, "6121697_m.fxe"},
	{0x21724C5C8B73A210ULL, "6126894_m.fxe"},
	{0x21726D9A7E1D44F0ULL, "vxht_keep_5gu_0_0.vlm"},
	{0x2172A5D8C199D57CULL, "one121_trap6112684.cl"},
	{0x217332C257915791ULL, "hlac_w_bonepit_0_1z.dds"},
	{0x217334068D2C8CE8ULL, "vxca_caverns1_mines_j3c_0_0.vlm"},
	{0x217360308195CE5FULL, "mag291_vincento6027680.cl"},
	{0x217374883582EE18ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.impa.evt"},
	{0x21737F4FE7EAE416ULL, "hlac_lt_warehouse_0_7z.dds"},
	{0x21739D0431D1E6DBULL, "vxw_blightlands_14a_0_0.vlm"},
	{0x217400E08D958C53ULL, "cod_hst_kirk_hightown.utp"},
	{0x21742D76872C982AULL, "w_sundermount_114_0.mao"},
	{0x21743176872C9F79ULL, "w_sundermount_114_0.mmh"},
	{0x21743B76872CB067ULL, "w_sundermount_114_0.msh"},
	{0x2174E1E376753A05ULL, "6085468_m.fxe"},
	{0x217597330D7DF035ULL, "6089918_m.fxe"},
	{0x2176C7825CA673D4ULL, "vxht_mansion2_3f0_0_0.vlm"},
	{0x2176D36EA4BBBFC0ULL, "vxht_0_hightown_c8z_0_0.vlm"},
	{0x217715B9BADAD0CDULL, "6118201_m.fxe"},
	{0x2177CAEB7DE70341ULL, "vxlt_0_lowtown_a2_n_c_ext_0_0.vlm"},
	{0x2177DA4658357156ULL, "vxlt_undercityserialkiller_19l_0_0.vlm"},
	{0x2177FB9A8D5E03E9ULL, "w_bonepit_2_0.rml"},
	{0x217949809021EAF3ULL, "vxlt_0_lowtown_a2_ewb_0_0.vlm"},
	{0x2179731938667343ULL, "hlac_lt_0_lowtown_a2_65535_9y.dds"},
	{0x217A56804345905AULL, "6041392_f.fxe"},
	{0x217B3119B38602AFULL, "719848248.ogg"},
	{0x217B3A0B99421F1EULL, "dae000_light_tombs_boss6179471.cl"},
	{0x217B47C39D5212B8ULL, "221088438.ogg"},
	{0x217B7EFF850FFBA5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_avea_d.dds"},
	{0x217B89B137061C65ULL, "6093499_f.fxe"},
	{0x217B907BCDB08478ULL, "hlac_lt_warehouse_n_0_0x.dds"},
	{0x217B919250FD8057ULL, "llac_lt_gallowstemplar_f_2383_11.dds"},
	{0x217C1F1E00549080ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_cracked01_sa.dds"},
	{0x217CF3D8D7F1A09FULL, "w_sundermount_212_0.phy"},
	{0x217D132767615F4CULL, "art\\vfx\\combat\\vfx_glb_ele_p_extreme_c.anb"},
	{0x217D4E8B22C55776ULL, "vxlt_gallowstemplar_n_5t2_0_0.vlm"},
	{0x217D74C6C2BC148BULL, "6148758_m.fxe"},
	{0x217DC42A7E9241B2ULL, "106823042.ogg"},
	{0x217DDA65B1D7503FULL, "vxht_chantry_n_2p7_0_0.vlm"},
	{0x217DFC9F0411632DULL, "vxca_cavern1_haunted_gjl_0_0.vlm"},
	{0x217E42A8318FD5F5ULL, "vxw_encounter1_c1_0_0.vlm"},
	{0x217E8442312EC561ULL, "pro000cr_friedrich_thug_3.utc"},
	{0x217EA7C9E7FF9CC4ULL, "vxlt_undercityencounter_1az_0_0.vlm"},
	{0x217EF71A1C07AE5FULL, "vxlt_blackemporium_tp_0_0.vlm"},
	{0x217F2F121077364CULL, "654088139.ogg"},
	{0x217FA44B2CF867F8ULL, "hlac_ca_caverns2prc_5802_42x.dds"},
	{0x217FBA9919650901ULL, "hlac_w_woundedcoast_-65531_0y.dds"},
	{0x2180DF7770135D70ULL, "vxht_keep_5jr_0_0.vlm"},
	{0x218167F924A8593EULL, "gen_im_wep_rog_dua_act2_05.uti"},
	{0x2181B23B6ABE3A7DULL, "5104339.ogg"},
	{0x2181EF9A8D6643C9ULL, "w_bonepit_2_0.mao"},
	{0x2181FB9A8D665862ULL, "w_bonepit_2_0.mmh"},
	{0x2181FD9A8D665B18ULL, "w_bonepit_2_0.msh"},
	{0x21820C1A4910660DULL, "586411967.ogg"},
	{0x21822EECEA9229B6ULL, "hlac_w_encounter3_0_1x.dds"},
	{0x2183227295EB9FA4ULL, "6039422_m.fxe"},
	{0x2183636BAB9D0EE2ULL, "sky_ht_0_hightown_a2.mao"},
	{0x218386B8D9227F7AULL, "dre000_arrival_deeproads.stg"},
	{0x2183944453CD8B0EULL, "vxht_mansion2b_317_0_0.vlm"},
	{0x2183B1B892D97133ULL, "mag100_thrask6034817.cl"},
	{0x2183E396CB1CCEDAULL, "vxlt_0_lowtown_n_eha_0_0.vlm"},
	{0x218472A832C26CF1ULL, "6163867_m.fxe"},
	{0x218483F5BE86D8D0ULL, "6099077_m.fxe"},
	{0x2184BCF9034FAFA1ULL, "6020552_f.fxe"},
	{0x2184C5A75ACB0DE5ULL, "w_bonepit_31_0.rml"},
	{0x2185475FCD8E79ABULL, "art\\characters\\playercharacter\\elffemale\\glv_mrl.mao"},
	{0x21855FA25511905DULL, "vxlt_shop_ej_0_0.vlm"},
	{0x21858796ACD0E94DULL, "6117301_m.fxe"},
	{0x2185EA6B06D99510ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_pailempty_0.pwk"},
	{0x2186A6737F31F1D4ULL, "6084407_f.fxe"},
	{0x21877B64772348F6ULL, "hlac_w_sundermount_393211_0z.dds"},
	{0x21877E14E41A9208ULL, "tbe200im_fallen_dagger.uti"},
	{0x2187811D810F33B7ULL, "vxlt_0_lowtown_n_fit_0_0.vlm"},
	{0x2187E5AB5709FBFEULL, "mag160_feynriel6016585.cl"},
	{0x2188E4FF1F59D58CULL, "436062003.ogg"},
	{0x2188EEB4ED70D4D5ULL, "6042407_m.fxe"},
	{0x21894723642E3A1EULL, "6201362_m.fxe"},
	{0x2189D98F7CEB7E48ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_blueceiling_sa.dds"},
	{0x218A1C321845E673ULL, "vxht_keep_n_6xk_0_0.vlm"},
	{0x218A49B9EC2DF103ULL, "art\\levels\\atmosphere\\sky\\sb03_bcloud.dds"},
	{0x218A575F9B65B56AULL, "239955470.ogg"},
	{0x218A7AE4D8977F7DULL, "1039595816.ogg"},
	{0x218AC047EDAFD1AFULL, "6102420_m.fxe"},
	{0x218AEFE618BDB1EAULL, "vxht_bartrand_n_68i_0_0.vlm"},
	{0x218B024293C0CC01ULL, "6094545_m.fxe"},
	{0x218B127D703A91B8ULL, "zz_approval_debug6003823.cl"},
	{0x218B3DCE103CE08FULL, "6085439_m.fxe"},
	{0x218CB4852385EE75ULL, "vxw_encounter1_h0_0_0.vlm"},
	{0x218CDC904CF3304CULL, "w_bonepitdestroyed_a3_208_0.msh"},
	{0x218CE2904CF33A3EULL, "w_bonepitdestroyed_a3_208_0.mmh"},
	{0x218CEE904CF34ED5ULL, "w_bonepitdestroyed_a3_208_0.mao"},
	{0x218D2F034E401C05ULL, "6174441_m.fxe"},
	{0x218D9E74562DD7CDULL, "vxdr_primevaltomb_399_0_0.vlm"},
	{0x218DBC50BF3F19A7ULL, "hlac_ht_keep_4519_12x.dds"},
	{0x218DD9FA9E5E4718ULL, "vxw_encounter2_pd_0_0.vlm"},
	{0x218DDDC9F3A8D333ULL, "lt_0_lowtown_a2_n_37_0.anb"},
	{0x218DF4D8D80015A4ULL, "w_sundermount_212_0.mmh"},
	{0x218DF8D8D8001CF7ULL, "w_sundermount_212_0.mao"},
	{0x218E0AD8D8003B46ULL, "w_sundermount_212_0.msh"},
	{0x218E875B61822C0CULL, "lgt160_ntb_elf_conflict6163338.cl"},
	{0x218E905AA92D94F7ULL, "ser221_party_comments6166193.cl"},
	{0x218E92283A6E3320ULL, "dae300_bran6143674.cl"},
	{0x218EAD878F333387ULL, "11311840.ogg"},
	{0x218ED3BF43763230ULL, "965372969.ogg"},
	{0x218F84F302CEF2FCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose3left.anb"},
	{0x218F9B9F416DF0EEULL, "vxlt_0_lowtown_docks_a3_n_c1b_0_0.vlm"},
	{0x219000ED91057D66ULL, "6209893_f.fxe"},
	{0x2190272BCE359B2BULL, "gen00fl_aveline6122098.cl"},
	{0x21903DB88EE22944ULL, "6099460_m.fxe"},
	{0x2190530BA0E5768FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_loop.gad"},
	{0x2190882303AD1EB0ULL, "hlac_ht_chantry_2534_1x.dds"},
	{0x2190C20BB8F8E115ULL, "180731250.ogg"},
	{0x2190F35BB702C00AULL, "atl_shared_dxt1_00.dds"},
	{0x21913409C89EB6C9ULL, "plt_one100pt_loose_end_athe.nss"},
	{0x21917E835EB0A38FULL, "gen_im_acc_blt_act1_03.uti"},
	{0x2191B6A12028C211ULL, "gen00fl_aveline6139294.cl"},
	{0x21920811244A962EULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.anb"},
	{0x2192E42BE126647DULL, "302259259.ogg"},
	{0x2192E9B40F27AA37ULL, "w_bonepit_113_0.msh"},
	{0x2192FBB40F27C89AULL, "w_bonepit_113_0.mao"},
	{0x2192FFB40F27CFA9ULL, "w_bonepit_113_0.mmh"},
	{0x2193155DC5F1E422ULL, "gen_im_arm_cht_comc.uti"},
	{0x2193B72888A7F0C0ULL, "vxht_0_hightown_c2x_0_0.vlm"},
	{0x219411CD4BA0DB88ULL, "gen00fl_varric.cnv"},
	{0x2194224719D2F282ULL, "320153031.ogg"},
	{0x21943C3D0ECBE509ULL, "vxht_bartrand_n_6jx_0_0.vlm"},
	{0x2194942EB2E05AC4ULL, "llac_lt_0_lowtown_a2_n_65534_3.dds"},
	{0x2194AEA561D33441ULL, "vxlt_0_lowtown_a2_n_c_erj_0_0.vlm"},
	{0x2194F33EC38B5626ULL, "vxlt_warehouse_n_cy_0_0.vlm"},
	{0x2195222F8BFFCEF1ULL, "172159789.ogg"},
	{0x21952F2BF9F206FBULL, "lt_0_lowtown_a3_n_c_25_0_ll.mmh"},
	{0x2195300ED1F8C109ULL, "follower_banter6171917.cl"},
	{0x21956C881D7D25C7ULL, "374232051.ogg"},
	{0x21962738AA0372FDULL, "vxca_cavern1_haunted_k6l_0_0.vlm"},
	{0x21962CC76076621AULL, "6070508_m.fxe"},
	{0x219642F7C94B18BAULL, "vxlt_undercityserialkiller_1sb_0_0.vlm"},
	{0x219674F817DD0027ULL, "659452488.ogg"},
	{0x219697C325024124ULL, "6072269_m.fxe"},
	{0x21975AFF516D0E7DULL, "vxw_encounter2_camp_ir_0_0.vlm"},
	{0x21978549BEF7737AULL, "hm_nat_elf_were.mor"},
	{0x21979DBCC155052DULL, "lgt260_way_blood_mage6148407.cl"},
	{0x219833CD4BA4FEF0ULL, "gen00fl_varric.bnk"},
	{0x21989356BB49B35AULL, "ht_0_hightownexplosion_n_45_0_lc.mmh"},
	{0x21989E3221FD9B11ULL, "vxlt_warehouse_y5_0_0.vlm"},
	{0x2198A211C92D8DD0ULL, "vxlt_hangedman_2ur_0_0.vlm"},
	{0x2199810E1E995A8BULL, "914991949.ogg"},
	{0x2199A2E4C5C32661ULL, "hlac_lt_0_lowtown_a2_n_-196607_0y.dds"},
	{0x2199DF7C04EE7A9DULL, "llac_ht_chantry_n_2534_16.dds"},
	{0x219A6F049D533A81ULL, "vxlt_gallowstemplar_f_59z_0_0.vlm"},
	{0x219ABE1F95C3813FULL, "vxht_chantry_n_2f1_0_0.vlm"},
	{0x219B1CF12E268592ULL, "6175423_m.fxe"},
	{0x219B2800318562B7ULL, "119436009.ogg"},
	{0x219BA867C0121BCBULL, "ser221_party_comments6111559.cl"},
	{0x219BC6DE1B6D4E52ULL, "gen00fl_fenris6127125.cl"},
	{0x219CB9D16AF2493FULL, "mrl150_dalish_guard6068297.cl"},
	{0x219D6A7C5E528D63ULL, "6113884_m.fxe"},
	{0x219D7CD44A560D0FULL, "art\\vfx\\combat\\vfx_glb_ele_c_extreme_c.impa.gad"},
	{0x219D8F2A987D7CB3ULL, "vxlt_gallowstemplar_522_0_0.vlm"},
	{0x219E0F17791F9DD0ULL, "6029041_m.fxe"},
	{0x219E314CC90E6963ULL, "6099193_m.fxe"},
	{0x219E540EC5C5131BULL, "vxlt_0_lowtown_docks_cw4_0_1.vlm"},
	{0x219E8251B048B53DULL, "6108648_m.fxe"},
	{0x219EBDD00432A33DULL, "462518018.ogg"},
	{0x21A007C89D1705FDULL, "gen00fl_merrill6130137.cl"},
	{0x21A13DAE5B62E72CULL, "508694367.ogg"},
	{0x21A1E2EB5DA16ADAULL, "6134903_m.fxe"},
	{0x21A2B0EB8E22B53DULL, "gen_im_trs_dwv_junk2_03.uti"},
	{0x21A3208253C94CC1ULL, "vxlt_0_lowtown_a2_n_c_ewi_0_0.vlm"},
	{0x21A350F77B58C119ULL, "hlac_ht_keep_n_3734_0x.dds"},
	{0x21A391D44A5A9CBCULL, "art\\vfx\\combat\\vfx_glb_ele_c_extreme_c.impa.evt"},
	{0x21A3A5E9E02CA0AEULL, "233572066.ogg"},
	{0x21A3F4FD7B07056CULL, "llac_lt_0_lowtown_-131072_2.dds"},
	{0x21A4504A1A1A1BE7ULL, "vxlt_house_fi_0_0.vlm"},
	{0x21A4700BA0F65155ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_loop.ani"},
	{0x21A48D5B46161F9AULL, "vxlt_0_lowtown_a2_n_gz1_0_0.vlm"},
	{0x21A561FD04E378C4ULL, "315099159.ogg"},
	{0x21A57024013B741EULL, "hlac_lt_warehouse_n_0_1y.dds"},
	{0x21A5CC7D1345FFD1ULL, "hlac_w_woundedcoast_1_0y.dds"},
	{0x21A5CCDC436326FDULL, "vxw_encounter2_camp_dq_0_0.vlm"},
	{0x21A6B47B388062B8ULL, "w_bonepitmining_a2_5_0.rml"},
	{0x21A6B543EBB99021ULL, "l_mag100ar_eau_road.bnk"},
	{0x21A6BBAF9E28309BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ext.ani"},
	{0x21A6E801D2C2E47AULL, "vrc221_bartrand6042887.cl"},
	{0x21A7354472FD4AA4ULL, "862564337.ogg"},
	{0x21A744CC060F4F45ULL, "cd_ser270_cut_motherreveal.bnk"},
	{0x21A745A52F05272AULL, "gen00fl_isabela6110152.cl"},
	{0x21A7A76D75BEA92EULL, "zz_ser_debug6085776.cl"},
	{0x21A80CDACE2BD1E9ULL, "6104275_m.fxe"},
	{0x21A82C547D9B5880ULL, "qun221cr_saemus.utc"},
	{0x21A8EF785F600204ULL, "6149376_m.fxe"},
	{0x21A983F36083579EULL, "hlac_ca_caverns2prc_5802_43y.dds"},
	{0x21A996E9B493C6DAULL, "dae000_mother6004112.cl"},
	{0x21A9A5A75AE9F30EULL, "w_bonepit_31_0.mmh"},
	{0x21A9B1A75AEA07A5ULL, "w_bonepit_31_0.mao"},
	{0x21A9BFA75AEA1FFCULL, "w_bonepit_31_0.msh"},
	{0x21AA2BE3E28B877DULL, "mag321_comtesse_de_launcet6064887.cl"},
	{0x21AA4FB449CC394BULL, "vxw_bonepitmining_a2_1pb_0_0.vlm"},
	{0x21AA799129B13697ULL, "hlac_ca_caverns1_24421_3x.dds"},
	{0x21ABC1546BA112EBULL, "hlac_lt_0_lowtown_a2_n_-196609_1z.dds"},
	{0x21AC67BA7ABB4E27ULL, "isa211_party_reaction6202326.cl"},
	{0x21AC933A70F3D888ULL, "seb321_leliana6180978.cl"},
	{0x21AD217272CE5A44ULL, "6093660_m.fxe"},
	{0x21AD6C83BAFA01A3ULL, "vxlt_hangedman_n_21r_0_0.vlm"},
	{0x21AD8FBC5655F5F2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scale_sa.dds"},
	{0x21AD9C192F634157ULL, "zz_vault_debug6200343.cl"},
	{0x21AE4B47FBD485EBULL, "vxw_bonepitdestroyed_a3_2al_0_0.vlm"},
	{0x21AE8252CBCACAFAULL, "6156561_m.fxe"},
	{0x21AEB0C9397821E5ULL, "vxlt_0_lowtown_n_euy_0_0.vlm"},
	{0x21AEC5879A215616ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_8.pwk"},
	{0x21AF009C70CC3F9AULL, "401535053.ogg"},
	{0x21AF1459C3B37180ULL, "lt_0_lowtown_n_27_0.rml"},
	{0x21AF597C64778A0EULL, "vxca_caverns1_mines_j2b_0_0.vlm"},
	{0x21AF99697321BDF6ULL, "6133054_m.fxe"},
	{0x21AFED88F8997166ULL, "362820003.ogg"},
	{0x21B053CABE73DAB3ULL, "dae000_gamlen6121509.cl"},
	{0x21B059BA5C4DC481ULL, "zz_followers_banter6159030.cl"},
	{0x21B09B70BF84602CULL, "one300cr_ggt_thug3.utc"},
	{0x21B0C117B4F2D252ULL, "ht_keep_n_3731.anb"},
	{0x21B0C2807236B336ULL, "6133810_m.fxe"},
	{0x21B199D44A66E7D5ULL, "art\\vfx\\combat\\vfx_glb_ele_c_extreme_c.impa.ani"},
	{0x21B259BC61167934ULL, "658690411.ogg"},
	{0x21B26D3BC1DB13FEULL, "6062370_m.fxe"},
	{0x21B310F457FA4CF3ULL, "vxlt_0_lowtown_a2_ekf_0_0.vlm"},
	{0x21B32F69EE1F4C78ULL, "ht_0_hightownexplosion_n_51_0_ll.mmh"},
	{0x21B3C3AF9E32C892ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ext.evt"},
	{0x21B41F79F3CAA271ULL, "26932535.ogg"},
	{0x21B42B958DC98FA2ULL, "dog321cr_closet_man.utc"},
	{0x21B43D9356B28858ULL, "281372847.ogg"},
	{0x21B47A23846194ABULL, "41450544.ogg"},
	{0x21B4C4F2DC45AC56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_berserk.ani"},
	{0x21B4FAA354AB9FCBULL, "6121503_f.fxe"},
	{0x21B5AF3BFDFEB31DULL, "rdr110dg_guard_warehouse6016694.cl"},
	{0x21B6573E4109D86EULL, "6043717_m.fxe"},
	{0x21B6ED98D87C77CAULL, "vxdr_0_deeproads_cmp_0_0.vlm"},
	{0x21B6F668991EFB8DULL, "gen00fl_merrillact2_pt1.stg"},
	{0x21B77463F9CEB47FULL, "82722129.ogg"},
	{0x21B7D4C99C13237FULL, "hlac_lt_0_lowtown_a2_n_-131073_1x.dds"},
	{0x21B7D5B417F9F154ULL, "gen_im_arm_cht_noba_t5.uti"},
	{0x21B835835CE900A6ULL, "hlac_ht_0_hightown_-131071_0y.dds"},
	{0x21B8E8088C7F2E1AULL, "zz_rdr_debug.ncs"},
	{0x21B8F30BE1776483ULL, "lgt310_mag_melson6152133.cl"},
	{0x21B8FA3624E78477ULL, "dae000_arishok6082027.cl"},
	{0x21B949644C116940ULL, "vxlt_gallowstemplar_47c_0_0.vlm"},
	{0x21B9D0B5711EB5A7ULL, "331489533.ogg"},
	{0x21B9FAC9F3CE3FEFULL, "lt_0_lowtown_a2_n_37_0.rml"},
	{0x21BA90F523C07FD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small_trim1.gad"},
	{0x21BAE2AF9E391C25ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ext.gad"},
	{0x21BB1FCCC5CDE826ULL, "499560687.ogg"},
	{0x21BB2845833A5E2BULL, "vxlt_backalley_g_1wn_0_1.vlm"},
	{0x21BB41836E1085D8ULL, "follower_barks6173418.cl"},
	{0x21BB72267C654B0DULL, "6118628_f.fxe"},
	{0x21BBBFEB8B162860ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_roba_0t.dds"},
	{0x21BC0BCC4BA3655AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_ruin.mmh"},
	{0x21BC0FCC4BA36B91ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_ruin.mao"},
	{0x21BC1DCC4BA383A0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_ruin.msh"},
	{0x21BCA13AAC3EA57BULL, "lt_gallowstemplar_f_mmap_ph.dds"},
	{0x21BD11B6563D947BULL, "982386616.ogg"},
	{0x21BD22A62B6E3A65ULL, "vxlt_0_lowtown_n_enx_0_0.vlm"},
	{0x21BD6888E1432A98ULL, "dae000_meredith6154285.cl"},
	{0x21BD88D894A3F3A6ULL, "lgt160_wk_soph_cultist6119763.cl"},
	{0x21BDA26F22C3969BULL, "fx_snowparticals.dds"},
	{0x21BDA365DAD3DE20ULL, "vxlt_blackemporium_3r_0_0.vlm"},
	{0x21BDB285009819B9ULL, "vxca_caverns1_mines_gkc_0_0.vlm"},
	{0x21BF24061A88AA07ULL, "gen00fl_anders6125715.cl"},
	{0x21BFA3EAA6B172E0ULL, "hlac_ht_bartrand_n_3373_3z.dds"},
	{0x21BFB6957AEF2184ULL, "mag370_meredith6039460.cl"},
	{0x21C0698F499DF432ULL, "llac_lt_0_lowtown_a2_-65536_5.dds"},
	{0x21C16111E34B0CBAULL, "clf260_harley6089576.cl"},
	{0x21C1895BD0226748ULL, "gen00fl_isabela6101066.cl"},
	{0x21C1A5F523C6C264ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small_trim1.evt"},
	{0x21C1B45D037AC27EULL, "hlac_lt_0_lowtown_docks_n_196606_0y.dds"},
	{0x21C1BE82A7D3C2CAULL, "vxht_chantry_2iv_0_0.vlm"},
	{0x21C1CCF2DC504447ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_berserk.evt"},
	{0x21C23AFD84A8F3ECULL, "gen00fl_isabela6096807.cl"},
	{0x21C257315A8BCC24ULL, "art\\vfx\\materials\\vfx_link_smo_add.mao"},
	{0x21C27D432CC0343CULL, "fonts_it-it.gfx"},
	{0x21C2DBB060B58F1FULL, "qcr000_aveline.bnk"},
	{0x21C2EC420E8032B8ULL, "vxlt_backalley_n_1zv_0_0.vlm"},
	{0x21C319435B3A0506ULL, "vxlt_backalley_n_1z8_0_0.vlm"},
	{0x21C35134692E0253ULL, "6108519_f.fxe"},
	{0x21C35D13D9D84AECULL, "104648410.ogg"},
	{0x21C3D6CFCE98A96FULL, "w_woundedcoast_245_0.mao"},
	{0x21C3E2CFCE98BD0CULL, "w_woundedcoast_245_0.mmh"},
	{0x21C3E4F56AEEBC6CULL, "hm_quintus.mor"},
	{0x21C3E8CFCE98C7FEULL, "w_woundedcoast_245_0.msh"},
	{0x21C3FB7D6702439EULL, "llac_lt_0_lowtown_n_131071_1.dds"},
	{0x21C3FC328074DAE6ULL, "hlac_lt_0_lowtown_n_65537_2x.dds"},
	{0x21C42BCD4BCA2B8DULL, "gen00fl_varric.utc"},
	{0x21C45312E4AB878DULL, "vxht_keep_n_641_0_0.vlm"},
	{0x21C4B27B38995C95ULL, "w_bonepitmining_a2_5_0.msh"},
	{0x21C4B47B3899600BULL, "w_bonepitmining_a2_5_0.mmh"},
	{0x21C4C07B389974A8ULL, "w_bonepitmining_a2_5_0.mao"},
	{0x21C55F75CA72904AULL, "vxdr_0_deeproads_cjq_0_0.vlm"},
	{0x21C5739A0D43A644ULL, "vxlt_0_lowtown_docks_a3_n_aop_0_0.vlm"},
	{0x21C58115DC8CC7BCULL, "w_sundermount_54_0_lc.mmh"},
	{0x21C6265F9D09EBA7ULL, "vxw_encounter2_camp_pn_0_1.vlm"},
	{0x21C6BDB060B94527ULL, "qcr000_aveline.cnv"},
	{0x21C6C9131E426347ULL, "follower_banter6149220.cl"},
	{0x21C6FA62EFF5554CULL, "vxw_bonepitmining_a2_b0_0_0.vlm"},
	{0x21C70E2D93F6BCD6ULL, "zz_mer_debug6039337.cl"},
	{0x21C7333CD97ACCFAULL, "6087551_m.fxe"},
	{0x21C7D33D81C5C7F5ULL, "55716894.ogg"},
	{0x21C7F4CA907ADD98ULL, "vxw_blightlands_6m_0_0.vlm"},
	{0x21C84E3A8C5A8FE7ULL, "6045866_m.fxe"},
	{0x21C860C26885FD33ULL, "6088237_f.fxe"},
	{0x21C8B069DCC01EEEULL, "llac_w_sundermount_393215_0.dds"},
	{0x21C8D01039E73BE0ULL, "vxlt_hangedman_n_14q_0_0.vlm"},
	{0x21C8E2C41DEB7A23ULL, "vxht_bartrand_n_4fl_0_0.vlm"},
	{0x21C8EDF2DC569A90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_berserk.gad"},
	{0x21C9234C98139C8AULL, "6027562_m.fxe"},
	{0x21C9B0DFEE8A3090ULL, "vxlt_0_lowtown_a2_n_e1p_0_0.vlm"},
	{0x21C9FE513B9687C1ULL, "6138768_m.fxe"},
	{0x21CA3A71BE1B6518ULL, "llac_lt_undercitydungeon_-1_16.dds"},
	{0x21CB6EBF315E8A62ULL, "169845670.ogg"},
	{0x21CB94831D6452E5ULL, "vxlt_0_lowtown_n_ekg_0_0.vlm"},
	{0x21CBB48BA53273EEULL, "dae000_arishok6068587.cl"},
	{0x21CBE612D8F7B54FULL, "vxlt_0_lowtown_docks_n_b8f_0_0.vlm"},
	{0x21CBE68042DA176DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldbase.mmh"},
	{0x21CBEA8042DA1DBEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldbase.mao"},
	{0x21CBF88042DA35F3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldbase.msh"},
	{0x21CC0E28CA220279ULL, "327846777.ogg"},
	{0x21CC448A1BCF8F6BULL, "110976854.ogg"},
	{0x21CD077DAEB41D06ULL, "740835016.ogg"},
	{0x21CD4AB3E7918C92ULL, "6015949_m.fxe"},
	{0x21CDC2B0BFB75A7DULL, "llac_ht_keep_3734_1.dds"},
	{0x21CDCF4CBD45D674ULL, "vxw_encounter2_camp_39_0_0.vlm"},
	{0x21CDFC89D86ADE39ULL, "vxlt_0_lowtown_docks_n_bgl_0_1.vlm"},
	{0x21CE95E8AC16F2D1ULL, "zz_neel_test2.cut"},
	{0x21CF0F8E347A7FF5ULL, "stoney2_d.dds"},
	{0x21CF7FCB123282E4ULL, "526117604.ogg"},
	{0x21CFBDF523D3285DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small_trim1.ani"},
	{0x21CFEC5F5916EE8AULL, "zz_hawkecu.cut"},
	{0x21CFEDBF2965CF03ULL, "zz_vault_debug6203229.cl"},
	{0x21D011093F3C9C56ULL, "337140952.ogg"},
	{0x21D0305F99C9DB4AULL, "vxht_chantry_2bw_0_0.vlm"},
	{0x21D03176877AFDD2ULL, "w_sundermount_114_0.rml"},
	{0x21D0A54A44F9FFAFULL, "vxht_mansion_n_4h9_0_0.vlm"},
	{0x21D0C81591E922FAULL, "hlac_dr_dungeon_3214_8x.dds"},
	{0x21D1A45475F93590ULL, "6167484_m.fxe"},
	{0x21D1FF76C824F78BULL, "vxlt_0_lowtown_a2_n_ew9_0_0.vlm"},
	{0x21D27280DE91CCF6ULL, "6089789_f.fxe"},
	{0x21D2A0093B7A20D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\left_down_lhand.ani"},
	{0x21D309C2D2429EA2ULL, "137165726.ogg"},
	{0x21D33050599DC7B2ULL, "ave310_brennan6136773.cl"},
	{0x21D397F691184A8BULL, "vxht_mansion_n_541_0_1.vlm"},
	{0x21D3D152BC68A8CAULL, "vxdr_0_deeproads_cgr_0_0.vlm"},
	{0x21D3D56087CB1FE2ULL, "6043065_m.fxe"},
	{0x21D4165DB8401E39ULL, "434622781.ogg"},
	{0x21D41C798A1AFF57ULL, "hlac_dr_0_deeproads_5845_1x.dds"},
	{0x21D47648CC7D108FULL, "vxht_0_hightown_big_0_0.vlm"},
	{0x21D4DDED5902B98AULL, "6075107_m.fxe"},
	{0x21D50514BECCF2E8ULL, "vxht_mansion2_31y_0_0.vlm"},
	{0x21D5B255C58BA964ULL, "static1_punch_nolit.psh.fxo"},
	{0x21D630CD9E8CE5A7ULL, "vxlt_undercityserialkiller_24m_0_0.vlm"},
	{0x21D685604BE29899ULL, "6136003_m.fxe"},
	{0x21D6961B39835943ULL, "662383816.ogg"},
	{0x21D6F070E0A37786ULL, "457195142.ogg"},
	{0x21D70D8BD1BAF977ULL, "368979548.ogg"},
	{0x21D7327687811E66ULL, "w_sundermount_114_0.phy"},
	{0x21D74462C5B8CDA9ULL, "hlac_ht_0_hightown_n_-3_0y.dds"},
	{0x21D754A10FE192A3ULL, "vxht_bartrand_n_4ko_0_0.vlm"},
	{0x21D761D2E547E25CULL, "360259276.ogg"},
	{0x21D77643EEFB43B8ULL, "qcr000_arishok6060123.cl"},
	{0x21D7A92FBAFF5BD9ULL, "hlac_dr_primevalentrance_1726_2y.dds"},
	{0x21D7BEDC4A1D72A6ULL, "zz_flyover_test.cut"},
	{0x21D82B7C781311A3ULL, "hlac_w_encounter3_-131069_0y.dds"},
	{0x21D86A95FC4EE2E5ULL, "fen310_varania6118571.cl"},
	{0x21D8AB093B7E9F73ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\left_down_lhand.gad"},
	{0x21D8C70EBB8C5CACULL, "follower_banter6169959.cl"},
	{0x21D9E1DCE2121F67ULL, "vxht_mansion2b_2er_0_1.vlm"},
	{0x21D9EE047E6EBF91ULL, "vxw_encounter2_pp_0_1.vlm"},
	{0x21DA0BE6355B22CFULL, "606646533.ogg"},
	{0x21DA32CC66FB8D06ULL, "6207770_f.fxe"},
	{0x21DA57EFCAEDCDCFULL, "vxlt_0_lowtown_docks_n_b5a_0_0.vlm"},
	{0x21DADD7C623BA783ULL, "6125603_m.fxe"},
	{0x21DAF78384B0C330ULL, "llac_lt_0_lowtown_a2_-1_5.dds"},
	{0x21DB44D6498EF363ULL, "hlac_lt_darktown_-1_14z.dds"},
	{0x21DB7B6F22DC525DULL, "fx_snowparticals.mao"},
	{0x21DB810C83E61E7AULL, "lt_gallowscourtyard_n_44_0.rml"},
	{0x21DBB0A24999EBBDULL, "6110627_m.fxe"},
	{0x21DBBB0493444941ULL, "7628147.ogg"},
	{0x21DBF3B2C1A00552ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_bck.gad"},
	{0x21DC054A75FD3A4DULL, "vxw_bonepit_28e_0_0.vlm"},
	{0x21DC058522832ADBULL, "hlac_ht_0_hightown_n_196607_0x.dds"},
	{0x21DC292DF2997D7FULL, "hm_ser_edgert.mor"},
	{0x21DC689666CE0DF2ULL, "vxlt_0_lowtown_n_fhr_0_0.vlm"},
	{0x21DC98989360D306ULL, "follower_banter6160943.cl"},
	{0x21DCD8B9288BE7D0ULL, "mag360st_samson.stg"},
	{0x21DD4A7B63500CA8ULL, "247884489.ogg"},
	{0x21DD6E45CB1737B8ULL, "class_abilities_varric.gda"},
	{0x21DD9D99AE67DB16ULL, "w_bonepitmining_a2_217_0.msh"},
	{0x21DDA799AE67EC14ULL, "w_bonepitmining_a2_217_0.mmh"},
	{0x21DDAB99AE67F327ULL, "w_bonepitmining_a2_217_0.mao"},
	{0x21DDCBCC030F18D4ULL, "vxw_encounter3_d3_0_0.vlm"},
	{0x21DE04C52990BD7AULL, "vxw_bonepitdestroyed_a3_n8_0_0.vlm"},
	{0x21DE5DBE21B99DCFULL, "6154931_m.fxe"},
	{0x21DE72A14E543AA0ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgtp_0s.dds"},
	{0x21DE9CD8D6D17C91ULL, "lt_0_lowtown_docks_a3_n_199_0.mmh"},
	{0x21DEA8D8D6D19132ULL, "lt_0_lowtown_docks_a3_n_199_0.mao"},
	{0x21DEAC0E55F956FAULL, "vxlt_0_lowtown_docks_a3_n_btt_0_0.vlm"},
	{0x21DEB2D491CEE8A8ULL, "vxw_encounter1_fc_0_0.vlm"},
	{0x21DEB6D8D6D1A94FULL, "lt_0_lowtown_docks_a3_n_199_0.msh"},
	{0x21DF59A40E28C727ULL, "follower_banter6172770.cl"},
	{0x21DFA8093B84B8D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\left_down_lhand.evt"},
	{0x21E004B3185004A4ULL, "6118370_m.fxe"},
	{0x21E026591D4CD9ECULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur2_flemeth_real.ani"},
	{0x21E07153BA1B100BULL, "vxlt_0_lowtown_a2_n_ex8_0_0.vlm"},
	{0x21E1007B284DBA36ULL, "vxlt_0_lowtown_n_eqr_0_0.vlm"},
	{0x21E19493FBCC552AULL, "hlac_ht_chantry_n_2534_9y.dds"},
	{0x21E1A18BB7074505ULL, "framebuffereffects_fadedistortion_p0.psh.fxo.dx11.assembly"},
	{0x21E1ECC1A957DDFAULL, "art\\levels\\ztools\\default\\generic\\9ff_rocktall.mmh"},
	{0x21E1F0C1A957E431ULL, "art\\levels\\ztools\\default\\generic\\9ff_rocktall.mao"},
	{0x21E1F6B2C1A47589ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_bck.evt"},
	{0x21E1FB57FEB06963ULL, "115218353.ogg"},
	{0x21E1FEC1A957FC40ULL, "art\\levels\\ztools\\default\\generic\\9ff_rocktall.msh"},
	{0x21E2087AE1628D26ULL, "follower_barks6173405.cl"},
	{0x21E2257F5BA5AAF5ULL, "vxlt_hangedman_21o_0_0.vlm"},
	{0x21E22ACC4BC46965ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_ruin.phy"},
	{0x21E279008532C4A7ULL, "6066627_m.fxe"},
	{0x21E304C883AD49F2ULL, "6095400_m.fxe"},
	{0x21E36198F1AC0848ULL, "plt_lgt320pt_shw_ors_main.nss"},
	{0x21E3A62121FA6158ULL, "6024961_m.fxe"},
	{0x21E3C417B51E308EULL, "ht_keep_n_3731.rml"},
	{0x21E4117C12B2E6C1ULL, "one200_gold_followers6101134.cl"},
	{0x21E41344ED513E3AULL, "art\\characters\\creatures\\mubari\\animation\\dg_sniff_ex.gad"},
	{0x21E446BBD7CBE02EULL, "zz_and_debug6178581.cl"},
	{0x21E46F0D24737CC2ULL, "zz_dae_debug6142965.cl"},
	{0x21E49C255ED580AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_pinning_shot.ani"},
	{0x21E4BB6D64CDE338ULL, "6017817_m.fxe"},
	{0x21E4D5088CA4497DULL, "zz_rdr_debug.cnv"},
	{0x21E4F73D3DD8B119ULL, "6131103_m.fxe"},
	{0x21E541525A1D8C74ULL, "491864453.ogg"},
	{0x21E568F14BCAACFBULL, "ht_0_hightown_n_42_0.rml"},
	{0x21E61A7B06425F4EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_basic.msh"},
	{0x21E6249D31E4BD5BULL, "732836305.ogg"},
	{0x21E6287B0642769FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_basic.mao"},
	{0x21E633591D515B2EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur2_flemeth_real.gad"},
	{0x21E6347B06428B3CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_basic.mmh"},
	{0x21E66AF9C3BB89DFULL, "vxlt_undercityserialkiller_1g7_0_0.vlm"},
	{0x21E7BA854F149C6EULL, "gen00fl_fenris6110893.cl"},
	{0x21E89FBA0F0DE40CULL, "vxdr_0_deeproads_f0l_0_0.vlm"},
	{0x21E97FB3BAFA74ACULL, "hlac_lt_gallowstemplar_n_2392_1x.dds"},
	{0x21EA6A98F1B2365BULL, "plt_lgt320pt_shw_ors_main.plo"},
	{0x21EB1644ED576251ULL, "art\\characters\\creatures\\mubari\\animation\\dg_sniff_ex.evt"},
	{0x21EB46E008F46F3AULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st105a_0e.dds"},
	{0x21EBF79730856963ULL, "6064676_m.fxe"},
	{0x21EC3DA8F5053154ULL, "vxw_encounter3_a0_0_0.vlm"},
	{0x21EC6D0EB9E9500AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\custom\\crushing_prison_reaction_db_loop.gad"},
	{0x21EC76A21B86D5FAULL, "vxw_bonepitdestroyed_a3_i9_0_0.vlm"},
	{0x21EC8BBDF3E9B429ULL, "vxht_mansion2b_21a_0_0.vlm"},
	{0x21ECCD0B8572555AULL, "fen210_fenris6108757.cl"},
	{0x21ED0A84611E87FEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.ani"},
	{0x21ED16B0DC31CED4ULL, "vxca_caverns1_inz_0_0.vlm"},
	{0x21ED3673228B02D9ULL, "vxlt_warehouse_11d_0_0.vlm"},
	{0x21ED6A7B031B21F2ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolemb.anb"},
	{0x21EDE2E430773805ULL, "art\\characters\\creatures\\shade\\textures\\c_shadeb_0s.dds"},
	{0x21EE76900A461D24ULL, "6117270_m.fxe"},
	{0x21EEC343EBF75DD3ULL, "l_mag100ar_eau_road.wbd"},
	{0x21EF3E7D904243F5ULL, "427685236.ogg"},
	{0x21EF52E7950DC058ULL, "lit_up_01_l2.mao"},
	{0x21EF8BD87A20CCE8ULL, "vxlt_warehouse_qh_0_0.vlm"},
	{0x21EFAAC20FEE03C0ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.mmh"},
	{0x21EFDEB2C1B08AF8ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_bck.ani"},
	{0x21F00218857D4DF8ULL, "vxlt_undercityencounter_13m_0_1.vlm"},
	{0x21F01CD698E2AD67ULL, "llac_w_encounter3_-393213_0.dds"},
	{0x21F0F541F5296D8EULL, "ser200_aveline6201523.cl"},
	{0x21F129D5A74F7899ULL, "art\\characters\\playercharacter\\children\\mt_bas_kf_lashesm1.mao"},
	{0x21F21FFC1DADA067ULL, "art\\vfx\\combat\\vfx_c_profane_crust_c.anb"},
	{0x21F285E180BFAF75ULL, "6127481_m.fxe"},
	{0x21F294255EE1B0FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_pinning_shot.evt"},
	{0x21F2F5F8D6339A49ULL, "vxht_keep_6eu_0_0.vlm"},
	{0x21F3056ADEC6238BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_roba_0n.dds"},
	{0x21F32D4DFE204042ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v3.gad"},
	{0x21F337C64E246E3DULL, "w_bonepit_17_0.rml"},
	{0x21F34B5B6F18C8CAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_ex.gad"},
	{0x21F3E04042B51E36ULL, "6139316_f.fxe"},
	{0x21F4808091B89C30ULL, "llac_lt_0_lowtown_docks_131071_5.dds"},
	{0x21F4FD21F93CD737ULL, "182677480.ogg"},
	{0x21F5109EDF97977DULL, "vxlt_0_lowtown_a3_n_c_gz9_0_0.vlm"},
	{0x21F5126629F7478EULL, "body_tints.gda"},
	{0x21F5DD54D6D6EBA6ULL, "dae000_bartrand6027552.cl"},
	{0x21F6CC7571BBB051ULL, "vxlt_0_lowtown_a2_bye_0_0.vlm"},
	{0x21F7690DA4309A3BULL, "6025356_m.fxe"},
	{0x21F77644F167B61AULL, "334643205.ogg"},
	{0x21F80A8EA5C4D0D1ULL, "vxlt_gallowscourtyard_7c7_0_0.vlm"},
	{0x21F80D303BB64F3CULL, "6071237_m.fxe"},
	{0x21F82E44ED621520ULL, "art\\characters\\creatures\\mubari\\animation\\dg_sniff_ex.ani"},
	{0x21F867A0CAFDAAF6ULL, "931749272.ogg"},
	{0x21F8828B86F0C4B1ULL, "6080441_m.fxe"},
	{0x21F8C01D0C2DE01FULL, "6114944_m.fxe"},
	{0x21F969532AEABE00ULL, "vxlt_gallowscourtyard_4vb_0_0.vlm"},
	{0x21F975255EE79AE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_pinning_shot.gad"},
	{0x21F9D0F5ACD15A10ULL, "805145926.ogg"},
	{0x21F9F294E6882DAFULL, "dae021_sister_colinda_sta6148307.cl"},
	{0x21FA07090316AD33ULL, "vxca_caverns2_8o0_0_0.vlm"},
	{0x21FA2E5B6F1EB661ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_ex.evt"},
	{0x21FA40EE4703B15DULL, "mcr310_argument6074433.cl"},
	{0x21FA504DFE269A39ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v3.evt"},
	{0x21FAAE28AA265EB6ULL, "w_sundermount_120_0.msh"},
	{0x21FAB828AA266FB4ULL, "w_sundermount_120_0.mmh"},
	{0x21FABC28AA2676C7ULL, "w_sundermount_120_0.mao"},
	{0x21FB2284612AEDCFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.evt"},
	{0x21FBA4D141586DE5ULL, "hlac_lt_0_lowtown_n_1_4x.dds"},
	{0x21FBA85014811B59ULL, "vxlt_warehouse_14k_0_0.vlm"},
	{0x21FC1CEBC07B588BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_drink.evt"},
	{0x21FC3A588B1C627AULL, "6101437_m.fxe"},
	{0x21FC87D42498A4A1ULL, "231558889.ogg"},
	{0x21FCE86CFC3C35A4ULL, "6112170_m.fxe"},
	{0x21FD35D36AEB7A19ULL, "vxht_mansion2b_1ar_0_0.vlm"},
	{0x21FD4A79F4CAF3D7ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.dura.gad"},
	{0x21FD67DE1A2792B3ULL, "6108959_f.fxe"},
	{0x21FD8DC20FFA0FDFULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.anb"},
	{0x21FDF05DA113D244ULL, "6116989_m.fxe"},
	{0x21FDFDB56C16E568ULL, "vxlt_warehouse_zi_0_0.vlm"},
	{0x21FE2CF8BA64FC7FULL, "vxht_mansion_n_4xy_0_0.vlm"},
	{0x21FF5AE7E2A09924ULL, "vxca_caverns2_62g_0_1.vlm"},
	{0x21FF6890B1D139B6ULL, "6101727_m.fxe"},
	{0x21FF6DDA3D82E6CAULL, "6168120_m.fxe"},
	{0x22000A5E878E9F3BULL, "vxlt_gallowstemplar_f_5ap_0_0.vlm"},
	{0x2200480EB9F9BAF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\custom\\crushing_prison_reaction_db_loop.ani"},
	{0x22007569E97B3DACULL, "vxw_sundermount_35e_0_0.vlm"},
	{0x2200BEBE3D02B34EULL, "vxht_mansion_n_4yw_0_0.vlm"},
	{0x2200CF2DBAF8AA21ULL, "one300_zevran6077757.cl"},
	{0x2200F4AD0B14CDFDULL, "6124772_m.fxe"},
	{0x22018C403D2092EAULL, "hlac_lt_0_lowtown_docks_n_-131073_5x.dds"},
	{0x2201B630FDAEE73BULL, "133092249.ogg"},
	{0x2202238461310E78ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.gad"},
	{0x220268CE88A3E47CULL, "ht_0_hightownexplosion_n_36_0_lc.mmh"},
	{0x220335D01C8A739AULL, "lt_backalley_g_0_0_lc.mmh"},
	{0x220335EBC081A24CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_drink.gad"},
	{0x220422F2A3619745ULL, "6129362_m.fxe"},
	{0x22045756CF31DE19ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangbar_sa.dds"},
	{0x22045F79F4D13664ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.dura.evt"},
	{0x220466540521320AULL, "6163257_m.fxe"},
	{0x2204D78F61E5F14EULL, "988712588.ogg"},
	{0x2204F0B0299CA6ABULL, "fx_lightning_5.mao"},
	{0x2204F3BCE644166BULL, "234637101.ogg"},
	{0x220527FB94530642ULL, "vxlt_gallowscourtyard_n_7i5_0_0.vlm"},
	{0x220528EF94285718ULL, "hlac_lt_0_lowtown_a3_n_c_65538_2x.dds"},
	{0x22059DAACDA79AB4ULL, "140750838.ogg"},
	{0x22060134EC0DA1C9ULL, "art\\characters\\weapons\\staff\\w_stf_002a.mao"},
	{0x22060D34EC0DB662ULL, "art\\characters\\weapons\\staff\\w_stf_002a.mmh"},
	{0x22060F34EC0DB918ULL, "art\\characters\\weapons\\staff\\w_stf_002a.msh"},
	{0x22068B4B3CFC8FF7ULL, "vxlt_gallowstemplar_f_5pk_0_0.vlm"},
	{0x2206B33045D1474AULL, "genip_blocker_small.utp"},
	{0x2206BA656AD3B564ULL, "ts_raiders_acc_table.gda"},
	{0x2206BE85896AF8F1ULL, "gen00fl_merrill6128675.cl"},
	{0x220707AC8A6D196EULL, "6139522_m.fxe"},
	{0x22070BC1A9772E05ULL, "art\\levels\\ztools\\default\\generic\\9ff_rocktall.phy"},
	{0x22073B9D31630D8EULL, "18090859.ogg"},
	{0x2207584DFE313228ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v3.ani"},
	{0x22077312CFE5B066ULL, "905741758.ogg"},
	{0x2208265B6F2AE5B0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_ex.ani"},
	{0x22085FC2E9D64966ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\w_small_dagger_0_n.dds"},
	{0x22087A6181DCA18FULL, "vxlt_gallowstemplar_5qo_0_0.vlm"},
	{0x2208FC692E054CC5ULL, "vxw_bonepitdestroyed_a3_n3_0_0.vlm"},
	{0x2209882E38629E4CULL, "rock_body.dds"},
	{0x2209BCA4094A0BE6ULL, "451260923.ogg"},
	{0x2209BE1423AE223AULL, "1029003342.ogg"},
	{0x220A052590470755ULL, "vxht_0_hightown_ak0_0_0.vlm"},
	{0x220A34EBC087BF4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_drink.ani"},
	{0x220A3E4D56E7E8FAULL, "hlac_ca_cavern1_haunted_24423_1x.dds"},
	{0x220B137B06616F07ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_basic.phy"},
	{0x220B97E098A4C7FDULL, "663071594.ogg"},
	{0x220BA7B05CE19299ULL, "vxht_mansion2b_1dq_0_0.vlm"},
	{0x220BC7332FF23DE7ULL, "6140044_m.fxe"},
	{0x220BEB1A02375DD4ULL, "6066875_m.fxe"},
	{0x220BEFF3C52D75ABULL, "6113750_m.fxe"},
	{0x220C3F9ADA322AF9ULL, "vxlt_gallowstemplar_5c3_0_0.vlm"},
	{0x220CE520EF345F41ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_overlay.anb"},
	{0x220D012202FE8DC8ULL, "6078353_m.fxe"},
	{0x220E2AA347C82CFDULL, "fen110_party_comment6095250.cl"},
	{0x220E5E533D84C2AEULL, "290740958.ogg"},
	{0x220E90CE1C354059ULL, "isa211_isabela6072764.cl"},
	{0x220F8840AC3DF27AULL, "art\\levels\\4cinematics\\ferelden\\prp_cupwood.msh"},
	{0x220F9640AC3E09B3ULL, "art\\levels\\4cinematics\\ferelden\\prp_cupwood.mao"},
	{0x220F9A40AC3E10C0ULL, "art\\levels\\4cinematics\\ferelden\\prp_cupwood.mmh"},
	{0x22107F348F691F0FULL, "dre000_carver6104924.cl"},
	{0x2210B40339299414ULL, "313254152.ogg"},
	{0x2210D36D76BCCD5CULL, "82801317.ogg"},
	{0x2210E3AECEF3B62BULL, "vxca_cavern1_haunted_j4e_0_0.vlm"},
	{0x22111E33356108FAULL, "6077461_f.fxe"},
	{0x22112938FF882A42ULL, "186796234.ogg"},
	{0x221213816EA2989CULL, "ser100_party_comments6120600.cl"},
	{0x22127779F4DD9C5DULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.dura.ani"},
	{0x221318D58042CC02ULL, "vxlt_shop_d6_0_0.vlm"},
	{0x2213AEDC62D94BA4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha03a.msh"},
	{0x2213C18E13275A3FULL, "vxht_0_hightown_a3_cbc_0_1.vlm"},
	{0x2213C4DC62D97146ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha03a.mmh"},
	{0x22147AB143D42603ULL, "363604951.ogg"},
	{0x2214EE4889B101D1ULL, "vxlt_gallowscourtyard_7i5_0_0.vlm"},
	{0x2214FD0EB78F6377ULL, "6167623_m.fxe"},
	{0x22154DAE99286687ULL, "6190975_m.fxe"},
	{0x2215B6EC52454C92ULL, "6120894_m.fxe"},
	{0x2215C7E94D83700BULL, "6090820_m.fxe"},
	{0x2215DB86EE282ED6ULL, "107777906.ogg"},
	{0x221650C60E9B21D3ULL, "vxlt_gallowscourtyard_6pa_0_0.vlm"},
	{0x22169C2F101E49DDULL, "vxw_bonepitdestroyed_a3_17o_0_0.vlm"},
	{0x2216EAC2E702DE33ULL, "vxca_caverns2_8q2_0_0.vlm"},
	{0x2217014E67621665ULL, "hlac_ht_0_hightown_n_65536_3y.dds"},
	{0x2218BFE08C068A4BULL, "gen00fl_varric6075768.cl"},
	{0x221948CA27992CD2ULL, "bld_dri_mei.mao"},
	{0x22196822E5548169ULL, "hlac_ht_keep_n_4519_2z.dds"},
	{0x221A1CFC1DCEED68ULL, "art\\vfx\\combat\\vfx_c_profane_crust_c.mmh"},
	{0x221A26E9E00DE497ULL, "vxlt_blackemporium_ik_0_0.vlm"},
	{0x221A27CF0FBDF377ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_sitting.anb"},
	{0x221A4623F8FA99B9ULL, "art\\characters\\creatures\\dragonling\\mesh03.mmh"},
	{0x221A5023F8FAAAA7ULL, "art\\characters\\creatures\\dragonling\\mesh03.msh"},
	{0x221A84B462BE795AULL, "1038932611.ogg"},
	{0x221B4CA47B58139EULL, "426276984.ogg"},
	{0x221CEE186B7AD03BULL, "vxlt_gallowstemplar_f_5or_0_0.vlm"},
	{0x221D310D4354E69EULL, "vxw_bonepitmining_a2_fn_0_0.vlm"},
	{0x221D52980D640EBEULL, "atl_debug_dxt5_00.dds"},
	{0x221D54F59E8EE924ULL, "982016743.ogg"},
	{0x221E0CAD6330549CULL, "zz_blightcu.cut"},
	{0x221E979A7D7C9A1BULL, "vxdr_slavers_1t7_0_1.vlm"},
	{0x221E9911EA12F863ULL, "lt_gallowscourtyard_n_35_0_lc.mmh"},
	{0x221EA49B11AE8D32ULL, "greentest.tnt"},
	{0x221EBBDC62E36911ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha03a.phy"},
	{0x221EDAB087AD4FD7ULL, "6166901_m.fxe"},
	{0x221EF7CA04A6D71FULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv2.gad"},
	{0x221F02D0101A93C8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2b.mmh"},
	{0x221F0ED0101AA86BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2b.mao"},
	{0x221F20D0101AC6F2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2b.msh"},
	{0x221F3D0AFE8E8062ULL, "atl_worldmap_dxt5_01.dds"},
	{0x221FB0D5D5130F55ULL, "hlac_ca_cavern1_haunted_24422_1x.dds"},
	{0x221FB49EE1FE2EC0ULL, "vxca_caverns1_mines_gk7_0_1.vlm"},
	{0x2220712F0C3E8280ULL, "vxw_sundermount_314_0_1.vlm"},
	{0x22208713E00ECD17ULL, "follower_banter6141651.cl"},
	{0x2220F7E88F6D9713ULL, "hlac_lt_hangedman_n_-1_4y.dds"},
	{0x2221281EB212E005ULL, "703328572.ogg"},
	{0x2221A405310E7888ULL, "character_character_p50.psh.fxo"},
	{0x2221DA2F8B8549EBULL, "vxw_bonepitdestroyed_a3_29d_0_0.vlm"},
	{0x2222158FA87DF2B6ULL, "vxw_bonepitmining_a2_p0_0_0.vlm"},
	{0x2222385B7D66A074ULL, "vxdr_primevalentrance_27o_0_0.vlm"},
	{0x22228F71E018F858ULL, "dae000_male_commoner1_mob6174465.cl"},
	{0x2222BABFFF83EC89ULL, "572209323.ogg"},
	{0x2223B0CA4ECE4F9BULL, "gen_im_wep_rog_arc_act2_02.uti"},
	{0x22243931C83708CDULL, "6087383_m.fxe"},
	{0x22246A4FAF6E9D83ULL, "ht_keep_n_4519_ll.mmh"},
	{0x222472574013F6B2ULL, "vxca_caverns1_gki_0_0.vlm"},
	{0x22249164A4241790ULL, "899366968.ogg"},
	{0x22250E0C0214625DULL, "vxw_bonepitdestroyed_a3_18n_0_0.vlm"},
	{0x222559C5B8D6F4BDULL, "gen00fl_fenris6126880.cl"},
	{0x2225794A124F7C1EULL, "533036259.ogg"},
	{0x2225C7BEBF8C66A2ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyn_0n.dds"},
	{0x2225ECCA04ACE30CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv2.evt"},
	{0x22261CE5BA9D26AFULL, "dae131ip_at_to_darktown.utp"},
	{0x22264CC1C05AFDA7ULL, "1067369689.ogg"},
	{0x22265228C71577F7ULL, "llac_ca_caverns2_5802_71.dds"},
	{0x22266BF14C0248C7ULL, "ht_0_hightown_n_42_0.anb"},
	{0x2226AA711C10FD26ULL, "102291359.ogg"},
	{0x2226B16486AADAF6ULL, "6133623_m.fxe"},
	{0x22273B9B35933E20ULL, "hlac_dr_0_deeproads_5845_12y.dds"},
	{0x2227E48CCD4895A9ULL, "dae000ip_crowd_mf7_talk2.utp"},
	{0x22288B6A40CDC399ULL, "vxht_mansion2b_1rw_0_0.vlm"},
	{0x222A1E68CA040ABCULL, "126778876.ogg"},
	{0x222A428773487B93ULL, "vxca_caverns1_mines_cmk_0_0.vlm"},
	{0x222A5F7DEBCC677BULL, "gen00fl_fenris6113903.cl"},
	{0x222A6FF08C926C71ULL, "579182395.ogg"},
	{0x222AB33C4795722FULL, "vxlt_shop_cr_0_0.vlm"},
	{0x222AB7F0BAECF7B9ULL, "hlac_lt_0_lowtown_65535_8z.dds"},
	{0x222AD807AEFBF1E7ULL, "art\\vfx\\combat\\vfx_v_fir_imm_act_c.impa.gad"},
	{0x222BA2EA354AFF1EULL, "vxw_bonepitmining_a2_mo_0_0.vlm"},
	{0x222BB14E516BCA0EULL, "vxlt_warehouse_vk_0_0.vlm"},
	{0x222BEF96548E1F6EULL, "qcr221_cut_diplomatic_sol6201294.cl"},
	{0x222C48B84B74E1F8ULL, "6116105_m.fxe"},
	{0x222C552F1E797ACFULL, "docks_male_human_worker_sta6176074.cl"},
	{0x222C7BCCFEDD5B01ULL, "vxca_caverns2_6wa_0_1.vlm"},
	{0x222C7F0D7C3C921AULL, "vxlt_0_lowtown_docks_n_c42_0_0.vlm"},
	{0x222CBD11E092ECF7ULL, "6126840_m.fxe"},
	{0x222CE47899504544ULL, "6208055_m.fxe"},
	{0x222D6DA6BF765A0AULL, "vxca_caverns1_mines_ioj_0_0.vlm"},
	{0x222E2AAA92149C44ULL, "179509179.ogg"},
	{0x22302B14558A14FBULL, "arm_thg.tnt"},
	{0x22304C0C7D7B626BULL, "vxw_bonepitdestroyed_a3_24g_0_0.vlm"},
	{0x22308E9F107DC1D5ULL, "6088294_m.fxe"},
	{0x2230CD07AF004AF4ULL, "art\\vfx\\combat\\vfx_v_fir_imm_act_c.impa.evt"},
	{0x2230E73B610FB1C5ULL, "ht_mansion_n_3372.anb"},
	{0x2231021C8B2BA7AFULL, "setappearance.ncs"},
	{0x2231D5337290207AULL, "517125045.ogg"},
	{0x22326EA025E193EBULL, "6208049_m.fxe"},
	{0x2232A6E9F1F310A7ULL, "outside_day.msh"},
	{0x22334EEC4F8E5900ULL, "6095936_m.fxe"},
	{0x22337C969EE97E67ULL, "bu_wh.mao"},
	{0x2233D4CA04B8F865ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv2.ani"},
	{0x2233E4AF2DC75FF0ULL, "combat_deathblow_12.gda"},
	{0x223423CEC9EEA528ULL, "vxlt_hangedman_270_0_0.vlm"},
	{0x22347B4CF9DAB6AEULL, "root_war_general2.dds"},
	{0x22349886605F72FDULL, "289949745.ogg"},
	{0x2235374AC12FB5D7ULL, "vxca_caverns2prc_9u2_0_0.vlm"},
	{0x2235D6A55B603A2FULL, "925310629.ogg"},
	{0x223608FA6079F92BULL, "168029935.ogg"},
	{0x223637C11EFFB792ULL, "vxdr_primevaltomb_75j_0_0.vlm"},
	{0x2236B71A8F79197DULL, "228860458.ogg"},
	{0x2236BA8E22526E06ULL, "48316.ogg"},
	{0x2236FD4732C3DC19ULL, "vxht_mansion2b_1uv_0_0.vlm"},
	{0x22376E9285F75079ULL, "hlac_lt_0_lowtown_a2_n_131071_4x.dds"},
	{0x22379531B0147479ULL, "vxlt_gallowstemplar_5r0_0_0.vlm"},
	{0x2238029331576A8FULL, "dae000_keep_servant_1_sta6160890.cl"},
	{0x223823C0DB5CDE05ULL, "557298708.ogg"},
	{0x223941C13BE4930AULL, "lt_0_lowtown_n_34_0_ll.mmh"},
	{0x2239CA75DEBEEBE6ULL, "gen00fl_isabela6090322.cl"},
	{0x2239E9D4F249D99FULL, "hlac_w_blightlands_65533_0z.dds"},
	{0x223A3874C3D8FB7AULL, "vxht_bartrand_n_5rp_0_0.vlm"},
	{0x223A79D91BAB505EULL, "mag110_samson6016158.cl"},
	{0x223A853588C6E17AULL, "324876121.ogg"},
	{0x223B29EA4CD6D40EULL, "517968987.ogg"},
	{0x223B3055F5E76D18ULL, "vxw_encounter2_4x_0_0.vlm"},
	{0x223BA74506D7EC80ULL, "127099855.ogg"},
	{0x223BF05C7B0E9751ULL, "vxw_bonepitdestroyed_a3_1p2_0_0.vlm"},
	{0x223CE2664C623568ULL, "vxca_caverns1_i22_0_0.vlm"},
	{0x223CF494672B6D6EULL, "hlac_w_sundermount_-65537_0z.dds"},
	{0x223EC507AF0C7A2DULL, "art\\vfx\\combat\\vfx_v_fir_imm_act_c.impa.ani"},
	{0x223EDC0F79536DD7ULL, "vxht_chantry_n_358_0_1.vlm"},
	{0x223F5701B5037596ULL, "vxlt_darktown_4x8_0_0.vlm"},
	{0x223F5D261755EEBFULL, "6126387_m.fxe"},
	{0x223F63DAA725B555ULL, "6144275_m.fxe"},
	{0x223F78BD7C0F1A33ULL, "6088066_f.fxe"},
	{0x223F976BEE8FE2AEULL, "hlac_lt_0_lowtown_docks_a3_n_65535_2z.dds"},
	{0x223FD0D7F22E7A42ULL, "6076331_m.fxe"},
	{0x2240EC7843D36031ULL, "343788516.ogg"},
	{0x22410DF71C0DC844ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbed_0.pwk"},
	{0x22413B4889124663ULL, "6109765_f.fxe"},
	{0x22417D2762128830ULL, "vxlt_0_lowtown_docks_n_b0m_0_0.vlm"},
	{0x22421EEEB0015842ULL, "gen00fl_isabela6096823.cl"},
	{0x22423FEF4E3B51A3ULL, "art\\vfx\\models\\vfx_gui_disorient.mao"},
	{0x224243EF4E3B58F0ULL, "art\\vfx\\models\\vfx_gui_disorient.mmh"},
	{0x224251EF4E3B700AULL, "art\\vfx\\models\\vfx_gui_disorient.msh"},
	{0x22427DBDECE708C6ULL, "dre000_dougal.cnv"},
	{0x2242B7946B834E76ULL, "vxht_0_hightown_a3_c3t_0_0.vlm"},
	{0x2242F76E83B7367BULL, "6119280_m.fxe"},
	{0x2242F941DE963B01ULL, "vxht_bartrand_n_32q_0_0.vlm"},
	{0x2243266F5BEC765CULL, "6144933_m.fxe"},
	{0x2244E481546C9B21ULL, "vxlt_0_lowtown_n_e33_0_0.vlm"},
	{0x22450058BB7D0B90ULL, "6034506_m.fxe"},
	{0x22454B34DB738251ULL, "hm_bolund.mor"},
	{0x2245774659507009ULL, "w_encounter2_camp_13_0_ll.mmh"},
	{0x22458B512905B327ULL, "llac_w_blightlands_262145_0.dds"},
	{0x2245A30501053681ULL, "vxlt_gallowstemplar_n_51z_0_0.vlm"},
	{0x2245A58910833494ULL, "art\\vfx\\combat\\vfx_glb_cold_explo_p.anb"},
	{0x22464CBC874E6515ULL, "802817195.ogg"},
	{0x22466649743376FFULL, "vxht_mansion_n_4cd_0_0.vlm"},
	{0x22467FBDECEAF486ULL, "dre000_dougal.bnk"},
	{0x22473525F815BD0EULL, "art\\characters\\playercharacter\\humanmale\\glv_masf.phy"},
	{0x22476DAE8804683FULL, "vxlt_0_lowtown_a2_n_c_hl3_0_0.vlm"},
	{0x2247E3AA1ABEED88ULL, "follower_banter6160758.cl"},
	{0x224834A989FD113BULL, "699931093.ogg"},
	{0x2248AA51B5CF13FAULL, "vxht_bartrand_n_5uq_0_0.vlm"},
	{0x2248D40D32054CC1ULL, "vxht_bartrand_n_4a7_0_0.vlm"},
	{0x22492903B0E8E782ULL, "mcr310_argument.cnv"},
	{0x2249301B912E031CULL, "docks_fem_sailor_mob6176308.cl"},
	{0x224950A8D5DE472DULL, "714999619.ogg"},
	{0x22497F6412EE73D6ULL, "6089881_m.fxe"},
	{0x2249A5E7DD28BFF7ULL, "600261060.ogg"},
	{0x2249F58F1221BA18ULL, "vxw_blightlands_cb_0_0.vlm"},
	{0x224A26A6DC10644EULL, "art\\cinematics\\props\\letterback.mao"},
	{0x224A3DA7D59FD3BEULL, "seb321_leliana6197877.cl"},
	{0x224A5160A3628B0AULL, "vxca_caverns1_mines_iad_0_0.vlm"},
	{0x224A62396D04AFD1ULL, "vxw_bonepitdestroyed_a3_1u1_0_0.vlm"},
	{0x224A7B3B572BB068ULL, "vxw_encounter3_6y_0_0.vlm"},
	{0x224AB5D6013F72B9ULL, "zz_spir.arl"},
	{0x224AF1FDCA79BAA5ULL, "vxht_mansion2_1c8_0_0.vlm"},
	{0x224B937586A34592ULL, "6121219_m.fxe"},
	{0x224BF35240377C07ULL, "vxlt_backalley_n_1nx_0_1.vlm"},
	{0x224C3A49766A423CULL, "477095264.ogg"},
	{0x224CC8D55B72222AULL, "6070175_m.fxe"},
	{0x224D0B03B0EC9DBAULL, "mcr310_argument.bnk"},
	{0x224D6B267E603C36ULL, "vxw_bonepitmining_a2_a3_0_0.vlm"},
	{0x224DB233E92AA2D6ULL, "6156851_m.fxe"},
	{0x224DD81291D9C567ULL, "6040937_m.fxe"},
	{0x224E33C64E71135DULL, "w_bonepit_17_0.mao"},
	{0x224E37C64E711996ULL, "w_bonepit_17_0.mmh"},
	{0x224E41C64E712A94ULL, "w_bonepit_17_0.msh"},
	{0x224E67AD558C2380ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend4_da.dds"},
	{0x224E6C2A2A4CBA0BULL, "vxlt_0_lowtown_docks_n_bqq_0_0.vlm"},
	{0x224EC669F75EB62EULL, "one300_zevran6089713.cl"},
	{0x224F269ED5B67879ULL, "234050726.ogg"},
	{0x224F2734C9304E25ULL, "abi_w_charge.wbd"},
	{0x224FB828AA6EAFB7ULL, "w_sundermount_120_0.rml"},
	{0x22502C384F982A22ULL, "vxlt_gallowscourtyard_n_6q8_0_0.vlm"},
	{0x225035159F0E1B37ULL, "dre000_varric6034661.cl"},
	{0x225088A89F4FC586ULL, "6076591_m.fxe"},
	{0x2250F13CFF614506ULL, "6072689_m.fxe"},
	{0x2251989834675330ULL, "668976741.ogg"},
	{0x2252D72B1AA1F6F2ULL, "vxht_mansion_n_435_0_0.vlm"},
	{0x225336C470BDD28CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangbar_1.mao"},
	{0x225344F5B4B3310DULL, "6076008_m.fxe"},
	{0x22535EF08C01F5A7ULL, "6079353_m.fxe"},
	{0x2253670C588AF0CBULL, "mer340_bonepit_comments6083158.cl"},
	{0x225377469C5FBAD6ULL, "6110541_m.fxe"},
	{0x22538489108F3A57ULL, "art\\vfx\\combat\\vfx_glb_cold_explo_p.mmh"},
	{0x2253F9753B392AF8ULL, "88775119.ogg"},
	{0x22545D0B44BEF736ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugsmall.phy"},
	{0x22545D4F198BB650ULL, "art\\levels\\races\\vfx_models\\prp_brazlogsml.msh"},
	{0x22546B4F198BCE6AULL, "art\\levels\\races\\vfx_models\\prp_brazlogsml.mmh"},
	{0x22546F4F198BD4A1ULL, "art\\levels\\races\\vfx_models\\prp_brazlogsml.mao"},
	{0x2254776C4F0FBED3ULL, "706248527.ogg"},
	{0x2254BC25320DEBC6ULL, "mer320_hubert6083760.cl"},
	{0x2254D82666298F7FULL, "vxht_mansion_n_4fg_0_0.vlm"},
	{0x2255C4476F4AE2C9ULL, "vxlt_0_lowtown_e16_0_0.vlm"},
	{0x225657AE336BF607ULL, "vxht_0_hightown_bjf_0_0.vlm"},
	{0x225670DDEC543B29ULL, "ss_pride_demon.bnk"},
	{0x22567D7E4B68A7ECULL, "vxlt_0_lowtown_a2_n_c_hx0_0_0.vlm"},
	{0x22569E8BFCE76347ULL, "seb000pt_endstage.ncs"},
	{0x2256B728AA74CCAFULL, "w_sundermount_120_0.phy"},
	{0x2256EE8EC55E3775ULL, "gen00fl_isabela6087679.cl"},
	{0x22570BA710FCC691ULL, "vxlt_0_lowtown_a2_el9_0_0.vlm"},
	{0x22571C2EA7C52C7AULL, "vxht_bartrand_n_5xr_0_0.vlm"},
	{0x225745EA23FB6541ULL, "vxht_bartrand_n_4f6_0_0.vlm"},
	{0x22574A159B9F316BULL, "vxlt_0_lowtown_csh_0_0.vlm"},
	{0x225791C0623E24A7ULL, "6168103_m.fxe"},
	{0x22585E34719FD798ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thin.phy"},
	{0x225878CD918EC2DCULL, "ico_boots_med_5.dds"},
	{0x22589C4EAAAD0288ULL, "480160492.ogg"},
	{0x2258AFC1F94089C7ULL, "6143957_m.fxe"},
	{0x2258C876526C46F6ULL, "fxe_water_pen_d.mao"},
	{0x225947FC4F4F1870ULL, "dae000_gamlen6048515.cl"},
	{0x225968870798D218ULL, "742804399.ogg"},
	{0x22596C7BA72AA991ULL, "vxht_keep_n_6b3_0_0.vlm"},
	{0x22597254040214C1ULL, "lt_0_lowtown_docks_a3_137_0.phy"},
	{0x2259829F4D86EFF0ULL, "253529611.ogg"},
	{0x22599070CF7AA440ULL, "6180388_m.fxe"},
	{0x225A8EA7E00F1A50ULL, "hlac_lt_gallowstemplar_2383_6x.dds"},
	{0x225AE8EB167B91EBULL, "gen00fl_merrill6114061.cl"},
	{0x225AF27CFB0B8F84ULL, "233129767.ogg"},
	{0x225AF354C2525142ULL, "hlac_w_encounter3_-262144_1z.dds"},
	{0x225B2641496C342FULL, "vxlt_gallowscourtyard_76v_0_0.vlm"},
	{0x225B3AB24D683AAAULL, "6075275_m.fxe"},
	{0x225B5AF1B65CE3DBULL, "vxlt_0_lowtown_docks_a3_a0m_0_0.vlm"},
	{0x225BDD03705654B6ULL, "vxw_bonepitmining_a2_l4_0_0.vlm"},
	{0x225C382F86C34998ULL, "6099504_f.fxe"},
	{0x225C3C193E948A73ULL, "ca_caverns1_24422.rml"},
	{0x225C7CBBDC9491EDULL, "6094787_m.fxe"},
	{0x225CDAE1494870DDULL, "ht_0_hightown_45_0_ll.mmh"},
	{0x225CE441BC2B44CBULL, "6099027_m.fxe"},
	{0x225D03182A8BF95CULL, "scaleform_solid.mao"},
	{0x225DF45EF57190D2ULL, "295147057.ogg"},
	{0x225E460B44C724D7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugsmall.msh"},
	{0x225E580B44C7433AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugsmall.mao"},
	{0x225E5C0B44C74A49ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugsmall.mmh"},
	{0x225E80F62D754729ULL, "llac_w_blightlands_327680_0.dds"},
	{0x225E933CCEEDE12BULL, "hlac_ht_0_hightown_a2_-131074_1z.dds"},
	{0x225F2E25B208CBBFULL, "6089813_m.fxe"},
	{0x225F2F444CC7A6FBULL, "805945176.ogg"},
	{0x225F9CB5D9F8188EULL, "vxca_caverns1_mines_i76_0_0.vlm"},
	{0x225FCC9084B5F554ULL, "vxw_encounter3_98_0_0.vlm"},
	{0x22604CA4F59A617DULL, "6165542_m.fxe"},
	{0x2260E7E887E5BC59ULL, "b_gr_dis.mao"},
	{0x22610DD2083ECFCAULL, "gen00fl_sebastian6180675.cl"},
	{0x2261E52E71907FEFULL, "vxlt_0_lowtown_h6p_0_0.vlm"},
	{0x2261F2B30249252CULL, "6115953_m.fxe"},
	{0x2261FC6A240ABC74ULL, "hlac_lt_0_lowtown_docks_a3_-2_6x.dds"},
	{0x2262603A99DC5F17ULL, "vxlt_blackemporium_vf_0_0.vlm"},
	{0x2262D7A900269A1AULL, "w_bonepitmining_a2_224_0.mmh"},
	{0x2262DBA90026A151ULL, "w_bonepitmining_a2_224_0.mao"},
	{0x2262E9A90026B960ULL, "w_bonepitmining_a2_224_0.msh"},
	{0x226379774C163729ULL, "al_mage_attributes.gda"},
	{0x22638E3556249830ULL, "6095819_m.fxe"},
	{0x22645FF1377AA729ULL, "zz_lgt_1_debug6136988.cl"},
	{0x2264C98B25620E87ULL, "vxht_0_hightown_bwa_0_0.vlm"},
	{0x2264D171949B4A43ULL, "follower_banter6155694.cl"},
	{0x2264D20C8F87E86FULL, "vxlt_0_lowtown_ei8_0_0.vlm"},
	{0x2264D2E0D1BDDB07ULL, "vxca_caverns2_8xe_0_0.vlm"},
	{0x2264D40BAA2FC0B0ULL, "hlac_lt_0_lowtown_-196608_0z.dds"},
	{0x2264FA012675FBB2ULL, "vxlt_0_lowtown_docks_a3_d0m_0_0.vlm"},
	{0x2265AE2F2DBCD224ULL, "582804036.ogg"},
	{0x2265C790C30F3439ULL, "hlac_lt_hangedman_n_-1_5x.dds"},
	{0x226633EDF088FB89ULL, "423580616.ogg"},
	{0x2266CE35A5DB17EFULL, "vxht_chantry_n_2yh_0_0.vlm"},
	{0x2267A1A8A300407DULL, "fxa_shield_spk_mm.mao"},
	{0x2267A4E0C8728D87ULL, "gen_im_delete_me_01.uti"},
	{0x2268024DC170BCC0ULL, "6185088_m.fxe"},
	{0x22680882371551F8ULL, "397317514.ogg"},
	{0x226825CF2F205946ULL, "6161130_m.fxe"},
	{0x226875DFCA1E2750ULL, "vxlt_0_lowtown_docks_cvi_0_0.vlm"},
	{0x2268C097BFD1450EULL, "581824044.ogg"},
	{0x2268DF2FA038F5DDULL, "703059216.ogg"},
	{0x2269390AA80E64C4ULL, "450179794.ogg"},
	{0x22699FE1E48222ACULL, "vxlt_darktown_2qs_0_0.vlm"},
	{0x226A11FF1C8B7DEEULL, "48000093.ogg"},
	{0x226A4EE0624C6D36ULL, "vxw_bonepitmining_a2_k5_0_0.vlm"},
	{0x226A790182F76065ULL, "vxlt_0_lowtown_n_e2l_0_0.vlm"},
	{0x226AE83B61412BCDULL, "ht_mansion_n_3372.rml"},
	{0x226B0E027A8A01EEULL, "334830249.ogg"},
	{0x226B661DF6A670D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ext.evt"},
	{0x226BDEC7FADBA991ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_ave_flemeth_real.ani"},
	{0x226BF7CAA86F9083ULL, "6170348_m.fxe"},
	{0x226CDA765DAD3C89ULL, "m_mag221ar_store.bnk"},
	{0x226CF176B1B86A77ULL, "vxlt_gallowscourtyard_n_3x7_0_0.vlm"},
	{0x226D5395176B9849ULL, "937424578.ogg"},
	{0x226DEAE806FEAE4EULL, "vxht_mansion2b_1xo_0_1.vlm"},
	{0x226E3E6D76AC0DD4ULL, "vxw_encounter3_29_0_0.vlm"},
	{0x226E5386C3C7AF9FULL, "gen_im_npc_arm_hum_cht_tmpc.uti"},
	{0x226E97A48DC9842EULL, "gen00fl_aveline6053344.cl"},
	{0x226F252B4F3F6C3CULL, "6071066_m.fxe"},
	{0x226F6A5344496414ULL, "hlac_w_encounter2_327678_0y.dds"},
	{0x226F6FFC4E824ECCULL, "6172392_m.fxe"},
	{0x226F88D474270C02ULL, "34707027.ogg"},
	{0x226FB17D8BA034BEULL, "vxw_sundermount_24e_0_0.vlm"},
	{0x226FDECBB4BD0249ULL, "745327203.ogg"},
	{0x2270284DEB2AEF4AULL, "art\\characters\\playercharacter\\humanfemale\\arm_urvb.mao"},
	{0x227051E1F383662CULL, "817484996.ogg"},
	{0x2270826D487AAD00ULL, "205652572.ogg"},
	{0x2271161E9DD83C84ULL, "hla_ht_0_hightown_a2_-131070_0y.dds"},
	{0x22720D63B58E4548ULL, "hlac_w_encounter2_131077_0y.dds"},
	{0x227216712D09CB13ULL, "gen00fl_isabela6079056.cl"},
	{0x22723C9B066B6384ULL, "darktown_male_sickly_sta6175323.cl"},
	{0x22725F1DF6AC8314ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ext.gad"},
	{0x2272E9C7FAE1DB33ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_ave_flemeth_real.gad"},
	{0x22731F8117A861BFULL, "gen_im_trs_chn_dua_dag_r0.uti"},
	{0x2273791236701F09ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztablelong.phy"},
	{0x2273A1F5BA43A345ULL, "vxht_bartrand_n_5uz_0_0.vlm"},
	{0x227400680FC552A4ULL, "6112703_m.fxe"},
	{0x22742DCF7F8B626BULL, "vxlt_0_lowtown_cmj_0_0.vlm"},
	{0x2274A91E0CB96137ULL, "6013307_f.fxe"},
	{0x2275285BAB2E1117ULL, "zz_qun_debug6045386.cl"},
	{0x22754B25E8040398ULL, "vxw_blightlands_ra_0_0.vlm"},
	{0x227555A1B9CDD24EULL, "6137577_m.fxe"},
	{0x2275B7D042E6F951ULL, "vxw_bonepitdestroyed_a3_1d6_0_0.vlm"},
	{0x2275D06E8CFFD811ULL, "sse_ss_anders.gda"},
	{0x22765DE16FDEA257ULL, "vxlt_0_lowtown_docks_brc_0_0.vlm"},
	{0x2276BA916CEE7946ULL, "527250997.ogg"},
	{0x2276DAD44C2F4931ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_action.ani"},
	{0x227811BED6783B2CULL, "vxlt_darktown_2tp_0_0.vlm"},
	{0x22781C0D9060A1C4ULL, "hlac_dr_slavers_3217_1z.dds"},
	{0x227825683A09C305ULL, "llac_lt_gallowstemplar_2390_1.dds"},
	{0x227866FF048654E7ULL, "8881871.ogg"},
	{0x2278C4EEACBF0E55ULL, "vfx_water_splash_s_p.bnk"},
	{0x22796E1DF6B2BBF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ext.ani"},
	{0x22797CD417A41D45ULL, "art\\vfx\\materials\\vfx_dot.mao"},
	{0x2279B11A0DBEA2F5ULL, "6160624_m.fxe"},
	{0x2279E28D351EC8F9ULL, "605809694.ogg"},
	{0x227A49C6AF7D3B04ULL, "337971086.ogg"},
	{0x227A6F5D66DE64E2ULL, "6135243_m.fxe"},
	{0x227AFA45FC9E6673ULL, "551307743.ogg"},
	{0x227B0D80FD01FC94ULL, "vxlt_backalley_n_1dd_0_0.vlm"},
	{0x227B228AB0FC8091ULL, "ef_slave01.mor"},
	{0x227B47A3F4F010F7ULL, "rdr110dg_dockworker6015934.cl"},
	{0x227B51DAFC125099ULL, "gen00fl_sebastian6189400.cl"},
	{0x227B6A60BB2B9C28ULL, "w_encounter1_16_0.msh"},
	{0x227B7C60BB2BBAB9ULL, "w_encounter1_16_0.mao"},
	{0x227B8860BB2BCF52ULL, "w_encounter1_16_0.mmh"},
	{0x227BB69EB076DE80ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\gibgolem.anb"},
	{0x227C77B1F357C90EULL, "dae000_hate_enemies6167602.cl"},
	{0x227D3F04F2EF563BULL, "vxht_bartrand_n_3uv_0_0.vlm"},
	{0x227DE5D44C357AD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_action.gad"},
	{0x227E235A7D964D3EULL, "vxw_sundermount_29f_0_0.vlm"},
	{0x227F26739123D1D4ULL, "6086921_m.fxe"},
	{0x227F334628368F5DULL, "6173659_m.fxe"},
	{0x2280273A08695FBCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_basic.mao"},
	{0x2280333A0869745FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_basic.mmh"},
	{0x2280393A08697E01ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_basic.msh"},
	{0x22817120AFAB42CAULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic_4.mao"},
	{0x22819D66DC4639B4ULL, "6156750_m.fxe"},
	{0x2281CA03079C7DC3ULL, "6034934_m.fxe"},
	{0x2281DAEBB77ADDFEULL, "dae300_alistair_king6133598.cl"},
	{0x228213D2AC39BBC5ULL, "vxht_bartrand_n_5xy_0_0.vlm"},
	{0x228220A1328541D3ULL, "vrc321_party_barks6096257.cl"},
	{0x22828B38E2C33FD0ULL, "dae000_madam_lusine6152235.cl"},
	{0x22829EFD8D68DE88ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaricontro.mmh"},
	{0x2282AAFD8D68F32BULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaricontro.mao"},
	{0x2282BCFD8D6911B2ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaricontro.msh"},
	{0x2282C338C844EB7CULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_lucky_.gad"},
	{0x2282E638A5FD050EULL, "llac_ht_mansion2_818_46.dds"},
	{0x22835DF200DCD1CFULL, "6054749_m.fxe"},
	{0x228396E73F1AE0A8ULL, "framebuffereffects_radialblur_p0.psh.fxo"},
	{0x2283E2D44C39E138ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_grab_action.evt"},
	{0x2283E9415256ED8CULL, "6027111_m.fxe"},
	{0x2284B850E743CA1DULL, "645045122.ogg"},
	{0x2284BE99B84E475FULL, "6080674_m.fxe"},
	{0x2284DA2856803D91ULL, "822245901.ogg"},
	{0x22850AE129FBE6FEULL, "6180636_m.fxe"},
	{0x228587B7C970840EULL, "vxlt_blackemporium_15k_0_0.vlm"},
	{0x22858855C68F53B6ULL, "dae000_gen_turn_in_16198708.cl"},
	{0x2285C143BC39C1FDULL, "84896949.ogg"},
	{0x2285E1B491FBD95BULL, "471027563.ogg"},
	{0x2285EFB839C4287DULL, "942153489.ogg"},
	{0x2286BFB0896154F1ULL, "6112565_f.fxe"},
	{0x2286E0C95E916C1DULL, "6161652_m.fxe"},
	{0x228713DCCE0D9883ULL, "6071798_f.fxe"},
	{0x22874E7F95105D0CULL, "dae210ip_store_armor.utp"},
	{0x22876D1032C7F5D6ULL, "770243982.ogg"},
	{0x22882E22FBE623B2ULL, "6149374_m.fxe"},
	{0x22883425F84D52D1ULL, "art\\characters\\playercharacter\\humanmale\\glv_masf.mmh"},
	{0x22884025F84D6772ULL, "art\\characters\\playercharacter\\humanmale\\glv_masf.mao"},
	{0x22884E25F84D7E8FULL, "art\\characters\\playercharacter\\humanmale\\glv_masf.msh"},
	{0x228883EB67B9808CULL, "hlac_lt_0_lowtown_docks_-196608_0y.dds"},
	{0x2288D15ECBD9B1D7ULL, "6170119_m.fxe"},
	{0x2289250A5F59C70BULL, "mag100_karras6034928.cl"},
	{0x22894FEC9E04833BULL, "lav_lakedark.mao"},
	{0x2289C238C84B087AULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_lucky_.ani"},
	{0x228A4A8641800062ULL, "6147637_m.fxe"},
	{0x228AA7D05B25EDE3ULL, "vxlt_0_lowtown_a2_n_byp_0_0.vlm"},
	{0x228B1DE73A39363BULL, "gen_im_acc_rng_lore_01.uti"},
	{0x228BCDA73789E733ULL, "673797868.ogg"},
	{0x228C05F97027E8E9ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffland.ani"},
	{0x228C2B23B24FE5FDULL, "pro000_friedrich6051069.cl"},
	{0x228C96BA109A5DE7ULL, "ave310_donnic6134737.cl"},
	{0x228D48748819F697ULL, "zz_personality_debug6004626.cl"},
	{0x228D8F990C73E871ULL, "art\\vfx\\textures\\vfx_smoke_thick.dds"},
	{0x228DF78F11321508ULL, "llac_lt_0_lowtown_n_131071_7.dds"},
	{0x228E525A3519F3A4ULL, "6028406_m.fxe"},
	{0x228ECAD4C98009D7ULL, "301523148.ogg"},
	{0x228EEEB8B79B26F2ULL, "framebuffereffects_bloomblend_p0.psh.fxo.dx11.assembly"},
	{0x228EF86ECCFB2447ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_aoe_p.anb"},
	{0x228F1A99017405D7ULL, "gen00fl_aveline6132679.cl"},
	{0x228F485B47334F88ULL, "6169631_m.fxe"},
	{0x22900134C967A4CFULL, "abi_w_charge.bnk"},
	{0x229085AF9E2FD445ULL, "vxht_bartrand_n_5cx_0_0.vlm"},
	{0x22908BD948895EABULL, "gen00fl_aveline6115470.cl"},
	{0x22909FBF0F9278BBULL, "fx_fireringfb_add.mao"},
	{0x2291017A9E2AC458ULL, "6217617_m.fxe"},
	{0x229137AB56644651ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_1y.dds"},
	{0x229175C0C49FEB98ULL, "gen00fl_sebastian6191248.cl"},
	{0x2291AD1752B56AECULL, "446801049.ogg"},
	{0x2291D36BCC19959DULL, "character_character_p26.psh.fxo.dx11.assembly"},
	{0x22925A77EFFAB440ULL, "69608191.ogg"},
	{0x2292F184F27B6339ULL, "896435.ogg"},
	{0x229320F9702E356BULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffland.gad"},
	{0x229342E800769BAFULL, "hlac_ht_mansion2b_818_14z.dds"},
	{0x2293FC8A88249F17ULL, "llac_lt_warehouse_n_1_4.dds"},
	{0x229482135AD47FB3ULL, "6071296_m.fxe"},
	{0x2294EF47ABD5A99DULL, "6069831_m.fxe"},
	{0x2295433168C45D9AULL, "llac_lt_0_lowtown_a2_n_-196607_0.dds"},
	{0x22958BC9A80E3CCDULL, "375539579.ogg"},
	{0x229595C75DB50929ULL, "54681823.ogg"},
	{0x2295D67BE2649346ULL, "one100dg_nabil6173811.cl"},
	{0x2295EEF2F45376EFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stocks_d.dds"},
	{0x2295F5798A4A49A8ULL, "268154709.ogg"},
	{0x2295FA9D60B93468ULL, "mcr310_party_banter6087791.cl"},
	{0x22964BBEAEF22A9BULL, "w_blightlands_53_0.msh"},
	{0x22965DBEAEF249E6ULL, "w_blightlands_53_0.mao"},
	{0x229669BEAEF25D85ULL, "w_blightlands_53_0.mmh"},
	{0x2296E67C0EB5F40FULL, "6122906_f.fxe"},
	{0x22971A580FD792F4ULL, "llac_lt_warehouse_n_65536_1.dds"},
	{0x22975FC94C0F5066ULL, "vfx_sparkae2.dds"},
	{0x22978BE9C019A024ULL, "vxlt_darktown_4iy_0_0.vlm"},
	{0x2297A8B2FD4D1DA4ULL, "dae000_bartrand6206935.cl"},
	{0x2297BB9E3A818BF5ULL, "vxht_0_hightown_a2_brx_0_0.vlm"},
	{0x2298012D00F595CFULL, "lt_0_lowtown_mmap_ph.dds"},
	{0x2298578CACDF6EFAULL, "lgt000ip_altar_static.utp"},
	{0x2298E3E903B6D90EULL, "6149631_m.fxe"},
	{0x2298F25E4C94F74AULL, "dae000_quintus6156659.cl"},
	{0x22995842A3BB5A3AULL, "6066617_m.fxe"},
	{0x2299673497AB5C0DULL, "vxw_encounter2_d9_0_0.vlm"},
	{0x229A1DF970344ED0ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffland.evt"},
	{0x229A54FE837D2CA1ULL, "258759494.ogg"},
	{0x229A90F123B8B454ULL, "gen00fl_isabela6099666.cl"},
	{0x229AD25B54B5F19BULL, "qun111_ketojan6059230.cl"},
	{0x229C3B77295206C3ULL, "plt_cod_bks_tranquil.plo"},
	{0x229C6FA44F938016ULL, "qun100pt_mage_escort.ncs"},
	{0x229C865E30CB22ECULL, "vxlt_warehouse_13k_0_0.vlm"},
	{0x229D311AA85597FCULL, "llac_ht_0_hightown_a3_65533_0.dds"},
	{0x229DDA01177EA46BULL, "gen00fl_anders6119459.cl"},
	{0x229E2F457B4CC9A3ULL, "6126686_m.fxe"},
	{0x229E4D34B03EC40FULL, "dre000_vampire6034365.cl"},
	{0x229E4EDDEC91B77BULL, "ss_pride_demon.wbd"},
	{0x229EE60042A5C61BULL, "6095034_m.fxe"},
	{0x229F62123695332EULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztablelong.mmh"},
	{0x229F626CFD8E05DCULL, "714626734.ogg"},
	{0x229F6E12369547C5ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztablelong.mao"},
	{0x229F78409F16B401ULL, "vxlt_gallowscourtyard_4lm_0_0.vlm"},
	{0x229F7C1236955F1CULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztablelong.msh"},
	{0x229FC72B156B0E5CULL, "6170624_m.fxe"},
	{0x229FC83A815D2DEDULL, "lt_0_lowtown_docks_191_0.mmh"},
	{0x229FCC3A815D343EULL, "lt_0_lowtown_docks_191_0.mao"},
	{0x229FD29FD0D76FF8ULL, "fen110_fenris6174755.cl"},
	{0x229FDA3A815D4C73ULL, "lt_0_lowtown_docks_191_0.msh"},
	{0x22A063A388540BE3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06b.mmh"},
	{0x22A071A38854233DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06b.msh"},
	{0x22A0E18A559337AEULL, "vxlt_hangedman_n_10s_0_0.vlm"},
	{0x22A1155460889C49ULL, "vxht_mansion_n_62o_0_0.vlm"},
	{0x22A1C796C1AFCEF1ULL, "6147818_m.fxe"},
	{0x22A29291B047548DULL, "art\\characters\\playercharacter\\humanfemale\\glv_ric.mmh"},
	{0x22A2A491B0477313ULL, "art\\characters\\playercharacter\\humanfemale\\glv_ric.msh"},
	{0x22A2B88914DB72AFULL, "dae000_mother6006254.cl"},
	{0x22A2F253A2EB7A4EULL, "gen00fl_isabela6079280.cl"},
	{0x22A350252D2CA13AULL, "6139957_m.fxe"},
	{0x22A36A1B9D95D77FULL, "w_bonepitmining_a2_12_0_lc.mmh"},
	{0x22A40419F1228BDDULL, "vxca_caverns1_hvr_0_0.vlm"},
	{0x22A40E64B53F2224ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basic.msh"},
	{0x22A42064B53F408DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basic.mao"},
	{0x22A42464B53F47C6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basic.mmh"},
	{0x22A4636F22CF6E35ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf1a.phy"},
	{0x22A48D674C7A0B92ULL, "126455180.ogg"},
	{0x22A4AB3A3E6C8951ULL, "641666277.ogg"},
	{0x22A4EAEEACE489DFULL, "vfx_water_splash_s_p.wbd"},
	{0x22A4ECBE00F40B5DULL, "vxdr_primevaltomb_81j_0_0.vlm"},
	{0x22A5A38E9686EDB9ULL, "6028357_m.fxe"},
	{0x22A5E947814CD2F2ULL, "fen210_orana6109721.cl"},
	{0x22A61BDC8B2BD837ULL, "gen00fl_anders6125805.cl"},
	{0x22A624999A7BB22DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_windtree.mmh"},
	{0x22A628999A7BB97EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_windtree.mao"},
	{0x22A636999A7BD0B3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_windtree.msh"},
	{0x22A69FA107CEEA14ULL, "6016955_m.fxe"},
	{0x22A79CCBF0688F45ULL, "839964295.ogg"},
	{0x22A7D7C7CA578F95ULL, "mer220cr_hubert.utc"},
	{0x22A8035516F2A23EULL, "vxw_encounter2_51_0_0.vlm"},
	{0x22A86D7DB909B11DULL, "dae000_bodhan6129138.cl"},
	{0x22A8A9F2509B3BDDULL, "vxlt_0_lowtown_docks_bax_0_0.vlm"},
	{0x22A91315C21636B0ULL, "lgt331ip_shw_mer_to_dtown.utp"},
	{0x22A9419FDE79C51CULL, "ht_bartrand_n_mmi.gff"},
	{0x22A96E355693041DULL, "pro000cr_flemeth_ogre.utc"},
	{0x22A981F7CCA0C018ULL, "follower_banter6172836.cl"},
	{0x22A98DEDFE3CE161ULL, "vxca_caverns2_8mm_0_1.vlm"},
	{0x22A9B1A9061ACC6DULL, "532106327.ogg"},
	{0x22A9BF2139ED3E8EULL, "qun221_viscount6095612.cl"},
	{0x22AA34E057522984ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_cast_c.impa.ani"},
	{0x22AAF83B22C13B6CULL, "vxlt_warehouse_16h_0_0.vlm"},
	{0x22AB260E595E8C8AULL, "vxlt_0_lowtown_docks_n_aom_0_0.vlm"},
	{0x22AC1C7C6BDEE8B1ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_016a_d.dds"},
	{0x22ACD1921CE88493ULL, "6075674_m.fxe"},
	{0x22ACEA8BB8F3A633ULL, "atl_fonttest_dxt1_dat.xml"},
	{0x22ACEB076E2EFF90ULL, "ht_mansion_n_mmap_ph.dds"},
	{0x22ADC8483F184BFFULL, "6039959_m.fxe"},
	{0x22ADD2DD56064F8AULL, "llac_ca_cavern1_haunted_24423_7.dds"},
	{0x22AE6969F5CF4DBAULL, "6160825_m.fxe"},
	{0x22AE7161D3013F57ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\spit.gad"},
	{0x22AE7A95092832F5ULL, "gen_im_acc_blt_act1_01.uti"},
	{0x22AECE4B21C70FC0ULL, "qcr220_grey_wardens6135909.cl"},
	{0x22AF1B64B5493F91ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basic.phy"},
	{0x22AFB7B4CFEBB895ULL, "6170092_m.fxe"},
	{0x22B00CE7C05FE684ULL, "hlac_ca_caverns1_mines_24421_9z.dds"},
	{0x22B03973B3A5E771ULL, "6140918_m.fxe"},
	{0x22B05A21FF4DDB68ULL, "vxht_keep_n_6jn_0_0.vlm"},
	{0x22B0C0735E952426ULL, "hlac_ht_mansion2b_818_27y.dds"},
	{0x22B0C84234751960ULL, "mag291_torpor6026976.cl"},
	{0x22B0F0C9FE99D5F9ULL, "vxca_caverns2_810_0_0.vlm"},
	{0x22B131E0575843E6ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_cast_c.impa.gad"},
	{0x22B164A388628008ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06b.phy"},
	{0x22B1A2BD16BAA731ULL, "hlac_lt_0_lowtown_a2_n_c_-65539_0z.dds"},
	{0x22B27B491848FB8CULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc107a_0s.dds"},
	{0x22B2C03D1AF8B5F9ULL, "6135370_m.fxe"},
	{0x22B33D867A493773ULL, "112788899.ogg"},
	{0x22B3691625AF6005ULL, "hlac_lt_0_lowtown_a2_65537_4z.dds"},
	{0x22B3719AB0B39BEFULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_m_p.mmh"},
	{0x22B38B91B055BC4AULL, "art\\characters\\playercharacter\\humanfemale\\glv_ric.phy"},
	{0x22B3905CF02A27F7ULL, "6206906_m.fxe"},
	{0x22B3BB948AE335DAULL, "48926348.ogg"},
	{0x22B3C5EE6FFBB0BDULL, "plt_cod_mcg_resolutionists.plo"},
	{0x22B41AFFA4ED7EDEULL, "hlac_lt_0_lowtown_docks_a3_n_-65535_0y.dds"},
	{0x22B424B5FBB056ECULL, "ht_0_hightownexplosion_n_44_0.rml"},
	{0x22B44B3471EE162DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thin.msh"},
	{0x22B4593471EE2E60ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thin.mao"},
	{0x22B45D3471EE34B3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_thin.mmh"},
	{0x22B49948313DFB9BULL, "vxca_caverns1_mines_edr_0_0.vlm"},
	{0x22B4B8765DEAB8DBULL, "m_mag221ar_store.wbd"},
	{0x22B4DCCCA30B33C1ULL, "551225126.ogg"},
	{0x22B4E5DE4C66F4CBULL, "587332144.ogg"},
	{0x22B50C2F2AF746D6ULL, "follower_banter6172320.cl"},
	{0x22B53ED44390631AULL, "art\\characters\\creatures\\buzzard\\c_buzzardb.mmh"},
	{0x22B550D443908260ULL, "art\\characters\\creatures\\buzzard\\c_buzzardb.msh"},
	{0x22B58661D30781E4ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\spit.evt"},
	{0x22B58C23B538B1A6ULL, "589481724.ogg"},
	{0x22B628C0A00963C9ULL, "hlac_ht_0_hightown_n_65532_0x.dds"},
	{0x22B62CEAA4EEFDBDULL, "ht_playershouse_n_855.rml"},
	{0x22B63BFC87A78B3AULL, "6068817_m.fxe"},
	{0x22B64AEE7B978D0DULL, "vxw_encounter2_j7_0_0.vlm"},
	{0x22B680958B15ECF0ULL, "vxht_0_hightown_a2_c4z_0_0.vlm"},
	{0x22B6F867663FAB21ULL, "370858073.ogg"},
	{0x22B71BCF4291545DULL, "vxlt_0_lowtown_docks_bdw_0_0.vlm"},
	{0x22B72CE0575CA6C5ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_cast_c.impa.evt"},
	{0x22B792B55D3C6233ULL, "follower_banter6149181.cl"},
	{0x22B7AEC1DA7C8485ULL, "mag210_arianni6021209.cl"},
	{0x22B7BBB57EE53830ULL, "6174426_m.fxe"},
	{0x22B7E54148989176ULL, "vxht_keep_4hg_0_0.vlm"},
	{0x22B7F56ECD1E2448ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_aoe_p.mmh"},
	{0x22B81E29E11B0734ULL, "vxht_0_hightown_9uw_0_0.vlm"},
	{0x22B8DB993944045DULL, "dae000cr_grand_cleric6185385.cl"},
	{0x22B9509610B57DE7ULL, "lt_0_lowtown_n_19_0_ll.mmh"},
	{0x22B96C64FF75261BULL, "character_character_p4.psh.fxo.dx11.assembly"},
	{0x22B9B03BDD36B3BBULL, "vxlt_gallowstemplar_n_5ls_0_0.vlm"},
	{0x22B9B66434D2C831ULL, "6131002_m.fxe"},
	{0x22BA13CAEB48922BULL, "6018419_m.fxe"},
	{0x22BA76C0C839F7EEULL, "zz_mcr_debug6088443.cl"},
	{0x22BA9510D88DE82AULL, "345874216.ogg"},
	{0x22BADFEC50F2A226ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_cassandra_.ani"},
	{0x22BB12FF5F38FAA3ULL, "6128486_m.fxe"},
	{0x22BB6A6EC947852EULL, "vxht_0_hightown_a2_c70_0_1.vlm"},
	{0x22BB6D6BD810DA33ULL, "statecnt_store.gda"},
	{0x22BB77540454E7FDULL, "lt_0_lowtown_docks_a3_137_0.mao"},
	{0x22BB7B540454EE36ULL, "lt_0_lowtown_docks_a3_137_0.mmh"},
	{0x22BB85540454FF34ULL, "lt_0_lowtown_docks_a3_137_0.msh"},
	{0x22BBCF0E79EE4051ULL, "vxdr_primevaltomb_8y6_0_0.vlm"},
	{0x22BBD21BF9B8C59CULL, "6087091_f.fxe"},
	{0x22BC2529BC422B11ULL, "6198483_m.fxe"},
	{0x22BC6B2A5B1610E8ULL, "vxlt_0_lowtown_a2_eou_0_0.vlm"},
	{0x22BCB05DDFD808A9ULL, "mrl210_merrill6185711.cl"},
	{0x22BD4470C91BE48DULL, "6123412_m.fxe"},
	{0x22BD69DEDD4D10DFULL, "dae000_gamlen6102519.cl"},
	{0x22BD941834070054ULL, "vxht_0_hightown_n_b97_0_0.vlm"},
	{0x22BDA887638B95CAULL, "hlac_lt_warehouse_-65535_0x.dds"},
	{0x22BE15827B11813EULL, "one280_merin.cnv"},
	{0x22BE4653F445DE1BULL, "vxht_0_hightown_a2_bsu_0_0.vlm"},
	{0x22BE6559C7AB00D7ULL, "vxca_cavern1_haunted_i9m_0_0.vlm"},
	{0x22BEBDE605371829ULL, "6083318_m.fxe"},
	{0x22BF055E686A6DB0ULL, "6115819_f.fxe"},
	{0x22BF1D2262215AB6ULL, "dae000cr_grand_cleric6181928.cl"},
	{0x22BF483930F89AC9ULL, "llac_lt_0_lowtown_n_65535_1.dds"},
	{0x22BF763EE51ACC13ULL, "6169184_m.fxe"},
	{0x22BF7D05F4409FFBULL, "vxlt_darktown_60k_0_0.vlm"},
	{0x22BFC4AA3F13C837ULL, "gen00fl_varric6033542.cl"},
	{0x22BFF4886A41ED32ULL, "vxlt_undercityencounter_1ds_0_0.vlm"},
	{0x22C156A9DE1B544BULL, "6208733_m.fxe"},
	{0x22C1DAAFAF03ECB1ULL, "lgt311_shw_mer_mettin6156555.cl"},
	{0x22C1F7827B15374EULL, "one280_merin.bnk"},
	{0x22C2978476E23C7EULL, "qun221_petrice6077361.cl"},
	{0x22C2DBC8C0408EC9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\mov_p.ani"},
	{0x22C39E61D313E7DDULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\spit.ani"},
	{0x22C3EDAD4DE7D78DULL, "6117015_m.fxe"},
	{0x22C3FA84461D21F4ULL, "vxht_0_hightown_n_amx_0_0.vlm"},
	{0x22C445F3A94D30F3ULL, "mrl310_party_comments6142904.cl"},
	{0x22C446A3A11B4C77ULL, "6016505_m.fxe"},
	{0x22C450F97D47D042ULL, "v_cage_04.dds"},
	{0x22C4548C8799DCBAULL, "hlac_lt_0_lowtown_docks_a3_0_0y.dds"},
	{0x22C4EDB2ACA4EAD5ULL, "6141706_m.fxe"},
	{0x22C53E5F2542ABDEULL, "lt_0_lowtown_docks_n_35_0.rml"},
	{0x22C53F338038AD21ULL, "6169160_m.fxe"},
	{0x22C54E885F7525B5ULL, "vxht_chantry_26p_0_0.vlm"},
	{0x22C55F2CC16DFED7ULL, "vxht_keep_n_4nc_0_0.vlm"},
	{0x22C5D236C62D0E25ULL, "964863922.ogg"},
	{0x22C5E68D3D25A987ULL, "vxlt_0_lowtown_n_fj1_0_0.vlm"},
	{0x22C5F28615294995ULL, "gen00fl_varric6021118.cl"},
	{0x22C6571E3A8EA9F6ULL, "vxht_keep_4mh_0_0.vlm"},
	{0x22C661018ACA58E6ULL, "hlac_lt_undercitydungeon_-1_5x.dds"},
	{0x22C67A9AAABA3CD7ULL, "6120792_m.fxe"},
	{0x22C6BF98AABF444AULL, "vxlt_gallowsprison_n_6a6_0_0.vlm"},
	{0x22C6F4CB124E2B65ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemb_hed_0d.dds"},
	{0x22C6FCEDA436049FULL, "hlac_lt_backalley_n_65536_3x.dds"},
	{0x22C70A773D695FBAULL, "zz_t_trailer_cut_aveline.cut"},
	{0x22C73E02353F453FULL, "vxlt_warehouse_13d_0_0.vlm"},
	{0x22C7854D119A61D1ULL, "ico_mm_plotarrow.dds"},
	{0x22C8345D27B010B3ULL, "cd_mag160_varian_ilithis.bnk"},
	{0x22C8401D75910C42ULL, "zz_airshock_angry_v1.cut"},
	{0x22C8F6C8C045284BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\mov_p.gad"},
	{0x22C905BAEAE3252FULL, "181324156.ogg"},
	{0x22C90CAAE2B48985ULL, "vxca_cavern1_haunted_e88_0_0.vlm"},
	{0x22C9284AC2D9A4D4ULL, "vxdr_primevalentrance_376_0_0.vlm"},
	{0x22C9A46E457C680EULL, "6068094_f.fxe"},
	{0x22CA39D984142929ULL, "916230031.ogg"},
	{0x22CABD3769250365ULL, "w_bonepit_6_0.rml"},
	{0x22CAF4DD72E7D697ULL, "seal_pls.mao"},
	{0x22CAFD5EEEDB0A80ULL, "vxw_encounter1_ef_0_0.vlm"},
	{0x22CB084ED40B6DF6ULL, "w_encounter2_camp_55_0.msh"},
	{0x22CB0923E005A4EDULL, "vxlt_gallowscourtyard_n_4v0_0_0.vlm"},
	{0x22CB124ED40B7EF4ULL, "w_encounter2_camp_55_0.mmh"},
	{0x22CB164ED40B8507ULL, "w_encounter2_camp_55_0.mao"},
	{0x22CB6DED4EC33857ULL, "vxw_bonepitdestroyed_a3_1zj_0_0.vlm"},
	{0x22CBB00270AB9215ULL, "dae011_badpoet_sta6146919.cl"},
	{0x22CBD6CE2608D6DEULL, "vxlt_warehouse_n_rn_0_0.vlm"},
	{0x22CC3CB64E163F44ULL, "6134721_m.fxe"},
	{0x22CC3D928584AC7DULL, "vxlt_0_lowtown_a2_eqg_0_1.vlm"},
	{0x22CCE40B25D4C5AFULL, "hlac_lt_0_lowtown_a2_n_c_0_4y.dds"},
	{0x22CD0EF42407882DULL, "214166018.ogg"},
	{0x22CD461EEB937FB9ULL, "6196688_f.fxe"},
	{0x22CD84A23586C672ULL, "204897233.ogg"},
	{0x22CDE6A9CEF1C6D0ULL, "152454092.ogg"},
	{0x22CDEB30EDC4EAB9ULL, "256313472.ogg"},
	{0x22CE20D4C2461653ULL, "6033636_m.fxe"},
	{0x22CEC8EC51032420ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_cassandra_.gad"},
	{0x22CED0964BD74FC8ULL, "208601850.ogg"},
	{0x22CF0E2AE1D9B4CCULL, "vxlt_gallowscourtyard_61y_0_0.vlm"},
	{0x22CF116FADC5EEC3ULL, "follower_banter6167871.cl"},
	{0x22CF14C5281B0126ULL, "mag311_huon_aoe.ncs"},
	{0x22CF20965984803EULL, "34360749.ogg"},
	{0x22CF365AAC5E5837ULL, "vxlt_warehouse_n_13k_0_0.vlm"},
	{0x22CFF3C8C04B41B0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\mov_p.evt"},
	{0x22D00692629908A0ULL, "gen_im_trs_cot_blt_r1.uti"},
	{0x22D19A305E65C990ULL, "637096120.ogg"},
	{0x22D1E5939F0F05B3ULL, "zz_ave_debug6062000.cl"},
	{0x22D276FC45AF2903ULL, "vxht_mansion2_25w_0_0.vlm"},
	{0x22D32EA85F83BE0DULL, "vxw_encounter2_p5_0_0.vlm"},
	{0x22D38B29A984B96EULL, "6175311_m.fxe"},
	{0x22D3AC481719203EULL, "vxw_blightlands_h9_0_0.vlm"},
	{0x22D3B3082C389779ULL, "hlac_lt_0_lowtown_docks_-4_2z.dds"},
	{0x22D40E626A20F93AULL, "lt_0_lowtown_docks_189_0.mmh"},
	{0x22D40EE4AFD70879ULL, "hlac_dr_primevaltomb_663_0z.dds"},
	{0x22D412626A210071ULL, "lt_0_lowtown_docks_189_0.mao"},
	{0x22D420626A211780ULL, "lt_0_lowtown_docks_189_0.msh"},
	{0x22D42925A1E3A04AULL, "vxht_chantry_n_2t3_0_1.vlm"},
	{0x22D4383B38B8EDF6ULL, "6165220_m.fxe"},
	{0x22D468955E3EA299ULL, "vxlt_0_lowtown_a2_gz5_0_0.vlm"},
	{0x22D4AABC08B1DC09ULL, "ca_caverns1_mines_24424_lc.mmh"},
	{0x22D4B3CDCC3149DFULL, "6150715_m.fxe"},
	{0x22D4BF4B64F51A10ULL, "mag121_idunna6084243.cl"},
	{0x22D4EA2DEEE13CDFULL, "mrl000_marethari6069869.cl"},
	{0x22D572DF64F06204ULL, "vxdr_primevaltomb_1mv_0_0.vlm"},
	{0x22D61118BC64D0F7ULL, "6049817_m.fxe"},
	{0x22D669BEAF28418EULL, "w_blightlands_53_0.rml"},
	{0x22D67577FF619458ULL, "vxlt_0_lowtown_a2_n_bxp_0_0.vlm"},
	{0x22D6A8B631778A81ULL, "zz_dre_debug6056961.cl"},
	{0x22D6FC5FF57D2AEEULL, "295028695.ogg"},
	{0x22D78760BB7A2C4DULL, "w_encounter1_16_0.phy"},
	{0x22D7BA6FFB19EDD4ULL, "vxht_keep_6h6_0_1.vlm"},
	{0x22D83F6366BACEE6ULL, "71701161.ogg"},
	{0x22D855D4CB0E3638ULL, "6086280_m.fxe"},
	{0x22D888DD9DF83F6FULL, "123467488.ogg"},
	{0x22D8B4316381EE28ULL, "dae011_gossips6149751.cl"},
	{0x22D8EF820A6DC1CDULL, "226506064.ogg"},
	{0x22D908DDBC9B82C0ULL, "6175142_m.fxe"},
	{0x22D919E847F986C0ULL, "6201409_m.fxe"},
	{0x22D92DE10BB3DFF2ULL, "vxw_bonepit_lj_0_0.vlm"},
	{0x22D96F3BE0D12300ULL, "vxw_encounter1_jg_0_0.vlm"},
	{0x22D9800B923D0836ULL, "vxca_caverns2_8kt_0_1.vlm"},
	{0x22D9E9D2D6929838ULL, "art\\levels\\races\\design\\_textures\\prp_chestb_n.dds"},
	{0x22DA118CE3005E80ULL, "vxdr_primevaltomb_9da_0_0.vlm"},
	{0x22DA48AB17FEEF5EULL, "vxlt_warehouse_n_yo_0_0.vlm"},
	{0x22DAE8CC145015F8ULL, "vxlt_0_lowtown_a3_n_c_hda_0_0.vlm"},
	{0x22DB08AEFB78713DULL, "171071233.ogg"},
	{0x22DB2A0DD8320F1BULL, "vxht_0_hightown_a2_byw_0_0.vlm"},
	{0x22DBA5C3A659527AULL, "372895075.ogg"},
	{0x22DBFA1B4876D23BULL, "vxlt_0_lowtown_docks_n_bf8_0_0.vlm"},
	{0x22DC1C20DB1E14F5ULL, "dm_bodahn.mor"},
	{0x22DC3D0B282052F9ULL, "vxht_0_hightown_b19_0_0.vlm"},
	{0x22DC45EDE6CCC008ULL, "6207708_m.fxe"},
	{0x22DC61A25B0117F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right.phy"},
	{0x22DC729AB0D6A22CULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_m_p.anb"},
	{0x22DCA4F3866B4AE0ULL, "zz_dae_debug6029163.cl"},
	{0x22DCCFDDA1850729ULL, "mag100_cullen6095164.cl"},
	{0x22DD78CF08809C8DULL, "6147007_m.fxe"},
	{0x22DD7B27782D457CULL, "gen00fl_merrill6130088.cl"},
	{0x22DD86722FB06387ULL, "176547304.ogg"},
	{0x22DD8860BB7E99F9ULL, "w_encounter1_16_0.rml"},
	{0x22DDA8379E5470B7ULL, "vxlt_warehouse_n_18j_0_0.vlm"},
	{0x22DDB756CD684003ULL, "324680967.ogg"},
	{0x22DDC42F2E3A109AULL, "6111670_m.fxe"},
	{0x22DDC94D15EA8108ULL, "hlac_ht_keep_n_3743_1x.dds"},
	{0x22DDDFBD37021050ULL, "10640591.ogg"},
	{0x22DE7E4E50FCAED1ULL, "6140068_m.fxe"},
	{0x22DEC85BB9EC9392ULL, "gen00fl_aveline6140804.cl"},
	{0x22DEFD7965CA0350ULL, "mag320_orsino6071133.cl"},
	{0x22DF08B31DB6BDB3ULL, "vxht_0_hightown_n_c3j_0_1.vlm"},
	{0x22DF561B168C0A5FULL, "plt_cod_itm_tranquility.plo"},
	{0x22DF61FB7B48F689ULL, "vfx_lightning_f_02.dds"},
	{0x22E0685CFECC736BULL, "hlac_lt_darktown_-1_25z.dds"},
	{0x22E0C1DE0747D362ULL, "146219760.ogg"},
	{0x22E1B64107420A7EULL, "dre000st_bartrandbetrayal.stg"},
	{0x22E1D55E6FCE2803ULL, "6120464_m.fxe"},
	{0x22E1D7A425415523ULL, "vxlt_darktown_5jf_0_0.vlm"},
	{0x22E23C3A08BC4DA4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_basic.phy"},
	{0x22E23F8DB67181DDULL, "vxca_cavern1_haunted_i1g_0_0.vlm"},
	{0x22E266AF3C073560ULL, "vxlt_ruinedhouse_n_hx_0_0.vlm"},
	{0x22E31EF266E7AE34ULL, "zz_followers_banter6158430.cl"},
	{0x22E390A332263BB7ULL, "880006210.ogg"},
	{0x22E393A0F658EFB0ULL, "vxlt_gallowstemplar_n_3r4_0_0.vlm"},
	{0x22E3DD0422465B67ULL, "vxlt_gallowstemplar_n_3s8_0_0.vlm"},
	{0x22E3F9C84EE771D6ULL, "gen_im_trs_prf_rocks_03.uti"},
	{0x22E4F4A037CB8CEEULL, "vxht_mansion2b_3dx_0_0.vlm"},
	{0x22E4F75A29239F2AULL, "6169714_m.fxe"},
	{0x22E4FB26BC985FA5ULL, "6124421_f.fxe"},
	{0x22E57C1D8D9D2BF0ULL, "6164760_m.fxe"},
	{0x22E58CA5EB516787ULL, "isa120_hayder6066911.cl"},
	{0x22E5BCA96CCE1B1CULL, "popups.gda"},
	{0x22E6643CB0F3D069ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_lose.ani"},
	{0x22E6EB53312C7251ULL, "vxht_chantry_n_32n_0_0.vlm"},
	{0x22E7CEF8463E1CFBULL, "mcr370_meredith6083047.cl"},
	{0x22E8D9637A43F233ULL, "6149472_m.fxe"},
	{0x22E917297BF41809ULL, "760728209.ogg"},
	{0x22E92BEAA51A55D5ULL, "ht_playershouse_n_855.anb"},
	{0x22E9605642D37681ULL, "vxlt_0_lowtown_a2_fhn_0_0.vlm"},
	{0x22E9CD76BBF37F86ULL, "hlac_w_blightlands_327676_0z.dds"},
	{0x22E9D46FD87CE88DULL, "art\\vfx\\models\\vfx_blood_drop.mmh"},
	{0x22E9DE56B9B61409ULL, "lt_0_lowtown_docks_a3_n_ll.mmh"},
	{0x22E9E66FD87D0713ULL, "art\\vfx\\models\\vfx_blood_drop.msh"},
	{0x22EA215BF8370054ULL, "894421594.ogg"},
	{0x22EAF8C8C0190C74ULL, "t3_arm_pl4.tnt"},
	{0x22EB114492966E16ULL, "mag291_demon_marethari6028881.cl"},
	{0x22EB4C42CDE03942ULL, "vxlt_blackemporium_108_0_0.vlm"},
	{0x22EB512032383C88ULL, "gen00fl_anders6136017.cl"},
	{0x22EB7E69B42D51D5ULL, "vxca_caverns2prc_7od_0_0.vlm"},
	{0x22EC18846A05EB3DULL, "6189423_m.fxe"},
	{0x22EC1FB5B0B5C629ULL, "plt_cod_cha_zevran.plo"},
	{0x22EC766F230D4450ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf1a.msh"},
	{0x22EC846F230D5C6AULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf1a.mmh"},
	{0x22EC9BED0841CF0BULL, "228208131.ogg"},
	{0x22ED564C547F0B91ULL, "812830263.ogg"},
	{0x22ED6E715EAD710AULL, "vxlt_0_lowtown_a2_n_eib_0_0.vlm"},
	{0x22ED7F3CB0FA1CEBULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_lose.gad"},
	{0x22EE7CCE61758D68ULL, "hlac_ht_playershouse_n_855_18x.dds"},
	{0x22EE9319D4774AF3ULL, "vxca_caverns1_cjo_0_0.vlm"},
	{0x22EF9D376943E88EULL, "w_bonepit_6_0.mmh"},
	{0x22EFA9376943FD25ULL, "w_bonepit_6_0.mao"},
	{0x22EFB7376944157CULL, "w_bonepit_6_0.msh"},
	{0x22EFC52FB363978AULL, "685608357.ogg"},
	{0x22EFCB62BA59C054ULL, "vxlt_0_lowtown_ewe_0_0.vlm"},
	{0x22F01914781E4CFEULL, "follower_banter6143261.cl"},
	{0x22F03D25678B10CFULL, "vxca_cavern1_haunted_dr1_0_0.vlm"},
	{0x22F0498117376DA3ULL, "vxlt_darktown_5oi_0_0.vlm"},
	{0x22F092AE0C944BE4ULL, "w_bonepitmining_a2_114_0.mmh"},
	{0x22F096AE0C945237ULL, "w_bonepitmining_a2_114_0.mao"},
	{0x22F0A8AE0C947086ULL, "w_bonepitmining_a2_114_0.msh"},
	{0x22F0EB19984D54EBULL, "fen310_fenris6117658.cl"},
	{0x22F0ECE1CD31E7AFULL, "vxlt_0_lowtown_a3_n_c_e29_0_0.vlm"},
	{0x22F16722EB260F81ULL, "6078796_m.fxe"},
	{0x22F1887243225BF9ULL, "709803705.ogg"},
	{0x22F1AFBEA91FF0D4ULL, "zz_mag_debug6054500.cl"},
	{0x22F1C48CD2DFFB3AULL, "hlac_ca_caverns1_mines_24423_3x.dds"},
	{0x22F1D10A475A531DULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrl.phy"},
	{0x22F1DC3819723620ULL, "61868487.ogg"},
	{0x22F2E9C26150E89BULL, "6109384_m.fxe"},
	{0x22F2FCBEC2687125ULL, "qun211cr_henchman6088119.cl"},
	{0x22F310C376B51067ULL, "311462464.ogg"},
	{0x22F33A7E0F7D19BBULL, "w_bonepit_43_0.msh"},
	{0x22F3460B4528CE84ULL, "vxlt_0_lowtown_n_e10_0_0.vlm"},
	{0x22F34C7E0F7D3806ULL, "w_bonepit_43_0.mao"},
	{0x22F3587E0F7D4CA5ULL, "w_bonepit_43_0.mmh"},
	{0x22F37B711AA2D81FULL, "hlac_w_bonepit_65536_0z.dds"},
	{0x22F3867ABEB11959ULL, "zz_facefx_test6088660.cl"},
	{0x22F47C3CB1003650ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_lose.evt"},
	{0x22F481CD3AFB1C81ULL, "378711422.ogg"},
	{0x22F49509348A38F9ULL, "gen00fl_merrill6126314.cl"},
	{0x22F4A64620289268ULL, "hlac_lt_0_lowtown_docks_-65539_3x.dds"},
	{0x22F4C38491B97B53ULL, "vxlt_0_lowtown_e2y_0_0.vlm"},
	{0x22F4EE8DB264A3FBULL, "gen00fl_carver6097393.cl"},
	{0x22F50BC056434B13ULL, "hlac_lt_gallowstemplar_n_2384_2y.dds"},
	{0x22F52881E1536014ULL, "dae000_bartrand6064050.cl"},
	{0x22F52B4BBBFA5A85ULL, "1054539750.ogg"},
	{0x22F55FAA5350764BULL, "hlac_w_bonepitmining_a2_0_3z.dds"},
	{0x22F57894C7E642BFULL, "vxlt_0_lowtown_n_ekg_0_2.vlm"},
	{0x22F5BF7296651F6DULL, "6031324_m.fxe"},
	{0x22F5EB638CE1210EULL, "vxlt_warehouse_8y_0_0.vlm"},
	{0x22F60890ECDDD9E4ULL, "6147640_m.fxe"},
	{0x22F640666C907266ULL, "vxdr_0_deeproads_cfs_0_2.vlm"},
	{0x22F6C297A687891FULL, "14867935.ogg"},
	{0x22F7208635586EAFULL, "vxca_cavern1_haunted_iiu_0_0.vlm"},
	{0x22F73AFA1495D40EULL, "519090729.ogg"},
	{0x22F76A8967E4D014ULL, "190920634.ogg"},
	{0x22F81F3548225ECCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11a.msh"},
	{0x22F82535482268BEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11a.mmh"},
	{0x22F84E5CA2213C56ULL, "583632810.ogg"},
	{0x22F92A7ABA7A1C90ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_005a_s.dds"},
	{0x22F9444D27D6C3EEULL, "one280_merin6105457.cl"},
	{0x22F9455627B0EEA5ULL, "956765038.ogg"},
	{0x22F97F9B6CF44777ULL, "4581199.ogg"},
	{0x22F9AA6A8E3B6FD5ULL, "hf_bloodmage1.mor"},
	{0x22F9D70F03E522B3ULL, "hlac_lt_backalley_n_65537_8z.dds"},
	{0x22FA2458BCF37C25ULL, "rage_tend_flip6x6.dds"},
	{0x22FA475A7D098E7BULL, "zz_mcr_debug6091895.cl"},
	{0x22FACFBDDE185DF9ULL, "6147223_m.fxe"},
	{0x22FB37996C0480DBULL, "647646705.ogg"},
	{0x22FB4174626B1DA0ULL, "81901252.ogg"},
	{0x22FC3ED4026420DBULL, "hlac_w_bonepit_0_1x.dds"},
	{0x22FD91BBF058388CULL, "6120855_f.fxe"},
	{0x22FDA7871B7E8664ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_2.ani"},
	{0x22FDDC3B420D1515ULL, "6001198_m.fxe"},
	{0x22FDDE10CE73D73BULL, "hlac_lt_0_lowtown_a2_65536_6y.dds"},
	{0x22FE2509E1BB03D2ULL, "1048158575.ogg"},
	{0x22FE3D3FAC4FD8D4ULL, "vxlt_0_lowtown_erj_0_0.vlm"},
	{0x22FEB9A7769C8A41ULL, "follower_barks6153824.cl"},
	{0x22FEBB5E092D8623ULL, "vxlt_darktown_5ph_0_0.vlm"},
	{0x22FEDDEE2369C70FULL, "vxdr_primevaltomb_6xa_0_0.vlm"},
	{0x22FEFFC7CD6F437FULL, "6170388_m.fxe"},
	{0x22FF5E03D8199FFFULL, "6034387_m.fxe"},
	{0x22FF779FBFCF2A35ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend3_da.dds"},
	{0x23001CF3371B97B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverloop.evt"},
	{0x2300218F3AC5A170ULL, "33747527.ogg"},
	{0x23017450D21DD379ULL, "llac_lt_0_lowtown_a2_65538_1.dds"},
	{0x23019F11D72B40CFULL, "743420584.ogg"},
	{0x230205D4BE6DDB66ULL, "gen00fl_varric6076032.cl"},
	{0x2302C60A4768B389ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrl.mao"},
	{0x2302CD16D319954CULL, "mag170_hugh.cnv"},
	{0x2302D20A4768C822ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrl.mmh"},
	{0x2302D40A4768CBD8ULL, "art\\characters\\playercharacter\\elffemale\\bdy_mrl.msh"},
	{0x2302E66A0243CADDULL, "957221612.ogg"},
	{0x2303200497B7A266ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_loop.ani"},
	{0x23035B166214C6E2ULL, "vxlt_0_lowtown_docks_n_bdc_0_0.vlm"},
	{0x23037BBF78EFDC41ULL, "6170485_m.fxe"},
	{0x23042EBA2DFAC641ULL, "rdr111dg_assistant6016652.cl"},
	{0x230433F81D8A1FA3ULL, "524148810.ogg"},
	{0x2304A4871B849FC6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_2.gad"},
	{0x2304AEDA2143B259ULL, "vxlt_gallowstemplar_f_4o1_0_0.vlm"},
	{0x2304D3FD71C0CDEBULL, "hlac_lt_0_lowtown_a2_-65536_8x.dds"},
	{0x23051CAF42D404E3ULL, "gen_im_arm_cht_comc_t6.uti"},
	{0x23052A7238721A1FULL, "lgt260_eau_tmp_roderick6143732.cl"},
	{0x23053561169B9A57ULL, "vxw_bonepitdestroyed_a3_1nf_0_0.vlm"},
	{0x230535F115A46632ULL, "llac_w_bonepitdestroyed_a3_0_0.dds"},
	{0x23055672CE6CF5F2ULL, "71364317.ogg"},
	{0x2305AF16D31B987CULL, "mag170_hugh.bnk"},
	{0x230609D1BBBF39F7ULL, "gen00fl_aveline6122043.cl"},
	{0x2306FB73BC540396ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldidle.evt"},
	{0x230713F33721A741ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverloop.gad"},
	{0x23072211607A4B31ULL, "hlac_lt_0_lowtown_n_131072_1x.dds"},
	{0x23074F5DC1B2E07CULL, "300654237.ogg"},
	{0x230766FE6174B40DULL, "ave331_post_fight6138824.cl"},
	{0x23083C27E062E137ULL, "mrl350ip_grave_d.utp"},
	{0x23087F3D76F58675ULL, "403732842.ogg"},
	{0x23088F86491DB093ULL, "mag320_orsino6071095.cl"},
	{0x23097ECB2A5285D1ULL, "6048250_m.fxe"},
	{0x2309B1F2CD83BD96ULL, "dae000_arishok6072874.cl"},
	{0x230A09C89EAEB6DDULL, "vxdr_primevaltomb_8nc_0_0.vlm"},
	{0x230B0F554E169C42ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1a_1.mao"},
	{0x230B771A2DB33DA2ULL, "6148113_m.fxe"},
	{0x230B9F871B8AB5A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_2.evt"},
	{0x230BF2E5CCCFF37BULL, "25726072.ogg"},
	{0x230C96BF694AF91EULL, "202373416.ogg"},
	{0x230CAF1C9E45F154ULL, "vxlt_0_lowtown_emk_0_0.vlm"},
	{0x230CDCBE725E8DCDULL, "vxht_mansion2b_2or_0_1.vlm"},
	{0x230DC5EFBD8EE502ULL, "vxlt_gallowscourtyard_7c8_0_0.vlm"},
	{0x230DCB4CFA614D61ULL, "300581587.ogg"},
	{0x230DE3DB423456A5ULL, "l_mer200ar_exterior.wbd"},
	{0x230DF273BC5A1321ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldidle.gad"},
	{0x230E14F33727C7EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverloop.ani"},
	{0x230E300926170499ULL, "vxlt_0_lowtown_a2_gn9_0_0.vlm"},
	{0x230E710280C6BA3AULL, "6168133_m.fxe"},
	{0x230F525D27EC4611ULL, "cd_mag160_varian_ilithis.wbd"},
	{0x230F90885A170A49ULL, "hlac_lt_undercitydungeon_-1_29y.dds"},
	{0x230FE942248AC84CULL, "vxca_caverns1_i7i_0_0.vlm"},
	{0x231044CA34751E40ULL, "combat_deathblow_05.gda"},
	{0x2310AD5BEBF710E1ULL, "6216714_m.fxe"},
	{0x2311080497C3B6D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_loop.evt"},
	{0x231145C497A4DA5BULL, "1059268912.ogg"},
	{0x23116C95FBCBD390ULL, "wos_02.mao"},
	{0x2311CCF3540ADF62ULL, "vxlt_0_lowtown_docks_n_bal_0_0.vlm"},
	{0x2312107447A82220ULL, "615435620.ogg"},
	{0x23124FE32FCE8CCEULL, "vxdr_primevaltomb_54c_0_0.vlm"},
	{0x23129128A0171777ULL, "art\\vfx\\cinematic\\vfx_chest_pierce_blood_p.anb"},
	{0x2312A3ED7922C887ULL, "dae000_tomwise.bnk"},
	{0x2312E9B0AE83D963ULL, "56852230.ogg"},
	{0x231310871E1C3A91ULL, "character_character_p103.psh.fxo.dx11.assembly"},
	{0x23134EDFC786BC19ULL, "hm_gen_templar_seeker.mor"},
	{0x2314414C5C4A50CFULL, "mag110_samson6018704.cl"},
	{0x2314446D7BD67CECULL, "270403608.ogg"},
	{0x23145ACE40B4EAA1ULL, "6075288_f.fxe"},
	{0x231491FC5697C695ULL, "6078550_m.fxe"},
	{0x2314962D36D78354ULL, "vxca_cave_1ag_0_0.vlm"},
	{0x2314F373BC6033CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldidle.ani"},
	{0x2314F972ABA8B922ULL, "vxdr_0_deeproads_fie_0_0.vlm"},
	{0x23153235817C0F32ULL, "vxlt_0_lowtown_docks_a3_n_aov_0_0.vlm"},
	{0x231542E66DED348BULL, "vxca_caverns1_mines_cnz_0_0.vlm"},
	{0x2315996CF65E59DCULL, "gen00fl_anders6096607.cl"},
	{0x2315B63E09D67521ULL, "dae000_flemeth6000809.cl"},
	{0x2315C5ED792538FFULL, "dae000_tomwise.cnv"},
	{0x2316B8FE642DBE3AULL, "vxw_bonepitmining_a2_b6_0_0.vlm"},
	{0x2317189F4CEC7B76ULL, "93676722.ogg"},
	{0x231718B70064DE2FULL, "hlac_dr_slavers_3216_5z.dds"},
	{0x2317228FED7C7A42ULL, "mcr371_hostile_anders6082107.cl"},
	{0x23175A2B372469F0ULL, "193972982.ogg"},
	{0x2317B59908ACA96BULL, "vxht_0_hightown_n_a7q_0_0.vlm"},
	{0x231801D793B8EFAEULL, "lt_0_lowtown_docks_a3_160_0.mmh"},
	{0x2318090497C9D760ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_loop.gad"},
	{0x23180DD793B90445ULL, "lt_0_lowtown_docks_a3_160_0.mao"},
	{0x23181BD793B91B9CULL, "lt_0_lowtown_docks_a3_160_0.msh"},
	{0x231839DBAF72A98CULL, "113885.ogg"},
	{0x2318B9ABEDC0ABFDULL, "mrl250cr_thief_c.utc"},
	{0x23196513113905CFULL, "gen_im_acc_blt_act2_02.uti"},
	{0x23199A179CAB7272ULL, "6126997_m.fxe"},
	{0x2319E8F71FA95622ULL, "6145213_m.fxe"},
	{0x231A0FFBDCEC801AULL, "6075231_m.fxe"},
	{0x231A548B6B828E7DULL, "vxdr_primevalentrance_30k_0_1.vlm"},
	{0x231A8EBFC63B7E15ULL, "qun160_vashothlead6064640.cl"},
	{0x231AB4516B0A2818ULL, "6148357_m.fxe"},
	{0x231AC06451621FB9ULL, "vxca_caverns1_iib_0_0.vlm"},
	{0x231AC5DDD08536FEULL, "32662939.ogg"},
	{0x231AF88C14C4F973ULL, "vxw_encounter2_gn_0_0.vlm"},
	{0x231C146484AD5496ULL, "lt_0_lowtown_docks_a3_170_0.mao"},
	{0x231C206484AD6935ULL, "lt_0_lowtown_docks_a3_170_0.mmh"},
	{0x231C226484AD6CABULL, "lt_0_lowtown_docks_a3_170_0.msh"},
	{0x231C38D9E04BFFDDULL, "vxlt_0_lowtown_docks_b9p_0_0.vlm"},
	{0x231CFE1184114218ULL, "679667339.ogg"},
	{0x231D4522B1B8038AULL, "dm_surfacer.mor"},
	{0x231D493F239C129DULL, "dae100_bran6118352.cl"},
	{0x231D6F37FB6C68A6ULL, "vxht_mansion2_1qy_0_0.vlm"},
	{0x231D944FE40286B4ULL, "6043891_f.fxe"},
	{0x231D9A25E1D73592ULL, "vxlt_0_lowtown_a3_n_c_gc5_0_0.vlm"},
	{0x231E19FFB534FCE4ULL, "vxht_mansion2b_265_0_0.vlm"},
	{0x231EAF471989C97CULL, "llac_lt_0_lowtown_n_-65535_7.dds"},
	{0x231EDC06286564BBULL, "6041480_m.fxe"},
	{0x231EE7FF6764A6D6ULL, "lt_0_lowtown_docks_16_0.anb"},
	{0x231F153B7EB0C470ULL, "596174343.ogg"},
	{0x231FE665934F57BDULL, "vxdr_0_deeproads_9ii_0_0.vlm"},
	{0x231FE937F5367225ULL, "dae011_conspiracy_mob6142182.cl"},
	{0x2320124ED453BEF7ULL, "w_encounter2_camp_55_0.rml"},
	{0x232037C36FCA9810ULL, "vxlt_house_dh_0_0.vlm"},
	{0x232049C34F4969C5ULL, "mrl150_merrill6077738.cl"},
	{0x232049F85C71A3D7ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.anb"},
	{0x23207CF14ABAD9E2ULL, "hlac_w_encounter3_-65531_0x.dds"},
	{0x23208F2DB2E43651ULL, "6040917_m.fxe"},
	{0x232139B6DDFEB10EULL, "hlac_lt_0_lowtown_docks_-131075_0x.dds"},
	{0x232152CE544D7895ULL, "22711493.ogg"},
	{0x23216A3EBAD5D105ULL, "vxlt_undercitydungeon_1ol_0_0.vlm"},
	{0x2321833ADF856FE4ULL, "gen_im_cns_cft_pot_02.uti"},
	{0x2321E574C695BEDEULL, "gen00fl_varric6118813.cl"},
	{0x232203DFF69A7CE1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic_2.mao"},
	{0x23226565EF013334ULL, "626246518.ogg"},
	{0x23229B2A4BC8BFACULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglebody_sa.dds"},
	{0x2322BB8DAFC0F257ULL, "llac_ht_chantry_n_2534_14.dds"},
	{0x2322EBA6E6FCA92AULL, "696765971.ogg"},
	{0x2322F005E46AC294ULL, "wpnstand_02.mao"},
	{0x232341689C34C813ULL, "148727981.ogg"},
	{0x2324042408512607ULL, "nat380_nathaniel6068350.cl"},
	{0x232425BE7E53638AULL, "follower_banter6171101.cl"},
	{0x232434143A2F2E82ULL, "vxdr_primevaltomb_855_0_0.vlm"},
	{0x23245EEAF8D6873FULL, "hlac_w_woundedcoast_262149_0x.dds"},
	{0x2324A0683825F5C3ULL, "320781747.ogg"},
	{0x23252ADB5623D6BAULL, "vxw_bonepitmining_a2_i7_0_0.vlm"},
	{0x2325C44C67B7F01AULL, "vxht_mansion_n_4xp_0_0.vlm"},
	{0x2325F0C55A45D50FULL, "art\\characters\\weapons\\staff\\w_stf_013a.mao"},
	{0x2325FCC55A45E9ACULL, "art\\characters\\weapons\\staff\\w_stf_013a.mmh"},
	{0x2325FFEED400C13FULL, "314445710.ogg"},
	{0x232602C55A45F39EULL, "art\\characters\\weapons\\staff\\w_stf_013a.msh"},
	{0x2326114ED45828EFULL, "w_encounter2_camp_55_0.phy"},
	{0x2326241774A3F885ULL, "vxlt_0_lowtown_a3_n_c_hnf_0_0.vlm"},
	{0x23263238B84833A5ULL, "547564064.ogg"},
	{0x232688717AB78B90ULL, "99689117.ogg"},
	{0x23269AA5A11296C3ULL, "vxlt_0_lowtown_docks_a3_n_cs1_0_0.vlm"},
	{0x2326ED82829AE7DDULL, "vxdr_primevaltomb_8da_0_0.vlm"},
	{0x2326FA2A21C4BBBBULL, "dae000_serendipity6161369.cl"},
	{0x23273D994C3CA7A6ULL, "mcr310_argument6072654.cl"},
	{0x2327FE808FF450ADULL, "hlac_ht_0_hightown_-196608_1z.dds"},
	{0x23280C6E25AAB308ULL, "81403402.ogg"},
	{0x2328CAF469132C16ULL, "81841712.ogg"},
	{0x2328E069176DECD7ULL, "6116191_m.fxe"},
	{0x2328F09E77A7C577ULL, "dae000ip_crowd_mf7_talk4.utp"},
	{0x2329324143583839ULL, "vxca_caverns1_ila_0_0.vlm"},
	{0x232937A32432E49CULL, "6101008_m.fxe"},
	{0x23296489DEA0521AULL, "vxlt_gallowsprison_n_6mr_0_0.vlm"},
	{0x2329977B40A9647DULL, "gen_im_trs_dwv_junk5_02.uti"},
	{0x232A04B89593AC07ULL, "472135616.ogg"},
	{0x232A1596F3BB52E8ULL, "pn_eye_nrma_0t.dds"},
	{0x232A895F2B57499DULL, "llac_lt_0_lowtown_131071_7.dds"},
	{0x232ADF52BB8EF77CULL, "656071726.ogg"},
	{0x232AFD738DE998FAULL, "nat380_sibling6067985.cl"},
	{0x232B99A835DB08CDULL, "6080618_m.fxe"},
	{0x232B9E1A3B9A74D3ULL, "vxlt_0_lowtown_ha5_0_0.vlm"},
	{0x232B9F2A28686D4AULL, "vxlt_shop_rx_0_0.vlm"},
	{0x232BE114ED628126ULL, "vxht_mansion2_1tz_0_0.vlm"},
	{0x232C06F2BA1C4038ULL, "hlac_lt_0_lowtown_docks_n_-196610_4y.dds"},
	{0x232CBD978D86BD77ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_004a_s.dds"},
	{0x232D66C3F41FA0DEULL, "llac_lt_0_lowtown_docks_a3_n_-131075_4.dds"},
	{0x232DAC67F193F7B0ULL, "mag100_karras6034532.cl"},
	{0x232DC27C3318C0DEULL, "6129109_m.fxe"},
	{0x232DCFFB94DE95CDULL, "vis121_saemus6089665.cl"},
	{0x232FD49E199B5B0AULL, "zz_vault_debug6203266.cl"},
	{0x2330043D3756A783ULL, "6151873_m.fxe"},
	{0x23306031A6A2E4BDULL, "6125199_m.fxe"},
	{0x233094EBDEF2992AULL, "599795833.ogg"},
	{0x2330FBC40EF7B15CULL, "6117494_m.fxe"},
	{0x2331812DF7E21B5EULL, "hlac_w_bonepitmining_a2_0_10y.dds"},
	{0x2331C185308FA2A0ULL, "vis121_saemus6167896.cl"},
	{0x23322619B95DE0D7ULL, "vxlt_0_lowtown_docks_b5v_0_0.vlm"},
	{0x23324395D026B0B8ULL, "hlac_w_bonepitdestroyed_a3_0_12y.dds"},
	{0x233249CBFEA99230ULL, "epilogue_text_positions.gda"},
	{0x233257BD7E7FFB3AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1c.phy"},
	{0x2332913B5919B04DULL, "vxlt_0_lowtown_a2_n_c_gm9_0_0.vlm"},
	{0x2332AD103ECDF3CBULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_abma_0d.dds"},
	{0x23330454749312F2ULL, "770008185.ogg"},
	{0x23334E09BFC834F3ULL, "lgt320cr_shw_ors_sympathize.utc"},
	{0x2333587E0FB330AEULL, "w_bonepit_43_0.rml"},
	{0x23348A30E25FD92FULL, "hlac_ht_0_hightown_-65536_0x.dds"},
	{0x2334995B497924EFULL, "sb_fade_clouda.dds"},
	{0x233506782BF88CBCULL, "gen00fl_aveline6002630.cl"},
	{0x23351ECFFCEF1352ULL, "vxlt_0_lowtown_ha0_0_0.vlm"},
	{0x23356936C3E14C7FULL, "mag101_aombanter6052762.cl"},
	{0x23359700D98E3476ULL, "dae300st_solivitus.utm"},
	{0x2335A725456EBB4CULL, "hlac_lt_0_lowtown_a2_-196608_4x.dds"},
	{0x2335BD45A93CD0B2ULL, "dae000_keep_noble_sta.cnv"},
	{0x2335C6E64FC87A71ULL, "6092565_m.fxe"},
	{0x2335FA19B15FC832ULL, "crt_aoe_cycle.mao"},
	{0x2336880769743021ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_modal_activ.ani"},
	{0x23368F8AAFEF6356ULL, "zz_otl_debug6060954.cl"},
	{0x2336A1463E047B34ULL, "122874169.ogg"},
	{0x2336FDDC51F1996BULL, "and111_lirene6084542.cl"},
	{0x23374EA25B4DA28DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right.msh"},
	{0x23375CA25B4DBAC0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right.mao"},
	{0x233760A25B4DC113ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right.mmh"},
	{0x2337A602673AD616ULL, "835340128.ogg"},
	{0x2337F8D8B33EAB19ULL, "docks_male_overseer_mob6176525.cl"},
	{0x23380A6455FD440BULL, "6099345_f.fxe"},
	{0x2338DF45A93F400AULL, "dae000_keep_noble_sta.bnk"},
	{0x2338FC354859BA59ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11a.phy"},
	{0x2339DB6889E4E839ULL, "hlac_lt_gallowstemplar_n_2384_3x.dds"},
	{0x233A2474CD3FA8E6ULL, "532165262.ogg"},
	{0x233A305A4575E794ULL, "fen210_hadriana6110418.cl"},
	{0x233A32A0588DF999ULL, "gen00fl_sebastian6195852.cl"},
	{0x233A5A5272687B92ULL, "follower_barks6154158.cl"},
	{0x233AE0DFD61FC2ACULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isaa_0s.dds"},
	{0x233B8E28A03A1778ULL, "art\\vfx\\cinematic\\vfx_chest_pierce_blood_p.mmh"},
	{0x233BC978CAAC95BBULL, "hlac_lt_0_lowtown_n_0_8y.dds"},
	{0x233C12C851013777ULL, "vxht_keep_7aj_0_0.vlm"},
	{0x233C51463BA9BEDBULL, "art\\vfx\\cinematic\\vfx_dragon_saliva_c.impa.ani"},
	{0x233CAD40D8E6FC67ULL, "895329034.ogg"},
	{0x233CB894AA75056CULL, "6095482_m.fxe"},
	{0x233CCA1F773B88BDULL, "vxdr_0_deeproads_9gg_0_0.vlm"},
	{0x233CFA7A5343A5C2ULL, "zz_seb_debug6185069.cl"},
	{0x233D634FE35CECDEULL, "combat_smce_range.gda"},
	{0x233D9307697A61C3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_modal_activ.gad"},
	{0x233ED20E9898FD3DULL, "6122099_m.fxe"},
	{0x233FDDF7629ECA09ULL, "6125407_m.fxe"},
	{0x233FEEF6D7DBD4CAULL, "6085392_f.fxe"},
	{0x23403C0917C8BE94ULL, "666351456.ogg"},
	{0x23404F72635F54B4ULL, "gen_im_npc_arm_hum_boo_gwdb.uti"},
	{0x234093EBDADA353BULL, "6111121_m.fxe"},
	{0x2340B4A9154CD91AULL, "lt_0_lowtown_docks_n_144_0.msh"},
	{0x2340B609FAA92BFDULL, "6083538_m.fxe"},
	{0x2340C2A9154CF153ULL, "lt_0_lowtown_docks_n_144_0.mao"},
	{0x2340C6A9154CF860ULL, "lt_0_lowtown_docks_n_144_0.mmh"},
	{0x234149129193C833ULL, "vxdr_primevaltomb_6s8_0_0.vlm"},
	{0x234234734FBAF3A2ULL, "mag170_thrask6054040.cl"},
	{0x2343494EDFE832F9ULL, "t3_wdg_tw2.tnt"},
	{0x234363227206FB82ULL, "vxw_bonepit_4b_0_0.vlm"},
	{0x23437E5F84FEC7C3ULL, "vxlt_0_lowtown_docks_a3_n_cy3_0_0.vlm"},
	{0x2343C7F9E1CD8EB9ULL, "6148640_m.fxe"},
	{0x234446B662174B9BULL, "6181915_m.fxe"},
	{0x23447007698044A8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_modal_activ.evt"},
	{0x234476AF7D920014ULL, "2539636.ogg"},
	{0x23449E37073DA0A8ULL, "vxca_caverns1_jgr_0_0.vlm"},
	{0x2345671515495FA7ULL, "6137486_m.fxe"},
	{0x2345D4A175CD1C1CULL, "and100_anders6095828.cl"},
	{0x234631B92CA36F74ULL, "987475378.ogg"},
	{0x234653C9ADD07E94ULL, "232377389.ogg"},
	{0x2346C08E4A250531ULL, "6040565_f.fxe"},
	{0x234846F85C92F0D8ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.mmh"},
	{0x23489747727D2690ULL, "gen00fl_anders6111366.cl"},
	{0x2348EAA18ECB0596ULL, "follower_banter6160745.cl"},
	{0x23493867A06FAA8CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_san01.msh"},
	{0x23493E67A06FB57EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_san01.mmh"},
	{0x23498239D997EA1CULL, "6111661_m.fxe"},
	{0x23499862890AF89AULL, "6163741_m.fxe"},
	{0x2349FEAD1E82CC7BULL, "hlac_lt_gallowscourtyard_-65538_1x.dds"},
	{0x234A00D3FA4EE0BDULL, "vxlt_0_lowtown_n_ffg_0_0.vlm"},
	{0x234A59463BB609D2ULL, "art\\vfx\\cinematic\\vfx_dragon_saliva_c.impa.evt"},
	{0x234A6D1687C37980ULL, "ser221_gamlen6162346.cl"},
	{0x234B4B078F24F52EULL, "vxlt_hangedman_n_257_0_0.vlm"},
	{0x234B920ECD872C18ULL, "hlac_w_bonepitdestroyed_a3_393221_0x.dds"},
	{0x234BE66C250A7413ULL, "dae200cr_viscount6087236.cl"},
	{0x234C088BC68A66FAULL, "6054494_m.fxe"},
	{0x234C8022E499B7E9ULL, "hlac_w_blightlands_131072_0x.dds"},
	{0x234C8529AE9A3F91ULL, "gen00fl_anders6125616.cl"},
	{0x234CA70B1AF2B6EDULL, "hlac_lt_0_lowtown_n_-131072_3y.dds"},
	{0x234D3C008B1F14FCULL, "mcr370_cut_player_rules6098284.cl"},
	{0x234D45385FF460EFULL, "vxlt_0_lowtown_a2_n_c_hkv_0_0.vlm"},
	{0x234D4A5B16BBCCE1ULL, "358727450.ogg"},
	{0x234D85276DF586AEULL, "6149648_m.fxe"},
	{0x234DA0A0E4B987D1ULL, "139803131.ogg"},
	{0x234DAB8DABDDCA57ULL, "93134913.ogg"},
	{0x234E28344AA482CAULL, "zz_one_debug6078370.cl"},
	{0x234E4FD45494E289ULL, "6122507_m.fxe"},
	{0x234EB297A057AE9AULL, "docks_male_elven_worker_sta6176120.cl"},
	{0x234EBDDB426BAD4FULL, "l_mer200ar_exterior.bnk"},
	{0x234EBF866DC30425ULL, "ch2armtnt_s1.psh.fxo"},
	{0x234F0C319778DD53ULL, "hlac_lt_gallowstemplar_2394_4y.dds"},
	{0x234F3EA064E0E8E8ULL, "vxht_keep_n_63e_0_0.vlm"},
	{0x234F5E2ABB6E7415ULL, "6112962_m.fxe"},
	{0x234F74F53D05E14DULL, "vxlt_0_lowtown_a2_n_c_gc7_0_0.vlm"},
	{0x234FC11939DA37ECULL, "narrative_varric6124289.cl"},
	{0x234FC2DDD99D871EULL, "llac_lt_0_lowtown_docks_-262144_2.dds"},
	{0x2350C7E113E3E397ULL, "vxlt_0_lowtown_a2_n_eot_0_1.vlm"},
	{0x235124C483E0E9E2ULL, "dae000_loyal_mage6165229.cl"},
	{0x235178463BBC5D65ULL, "art\\vfx\\cinematic\\vfx_dragon_saliva_c.impa.gad"},
	{0x2351B2A173B46515ULL, "hlac_ht_brothel_n_3371_2x.dds"},
	{0x2351C6D57B9525FEULL, "m_dae320ar_hightown_nt.bnk"},
	{0x2351CAFF678FCF6AULL, "lt_0_lowtown_docks_16_0.rml"},
	{0x23524285866FEFABULL, "vxlt_backalley_g_1dv_0_0.vlm"},
	{0x23524BC0FF73868AULL, "vxlt_gallowstemplar_f_34p_0_0.vlm"},
	{0x235258B4FB09B386ULL, "t1_mul_ro3.tnt"},
	{0x2352A7CDA1814933ULL, "vxdr_primevaltomb_57c_0_0.vlm"},
	{0x2352BAC465F8F044ULL, "vxlt_gallowstemplar_521_0_0.vlm"},
	{0x23536E09AE4DC9F9ULL, "vxlt_backalley_n_1nv_0_1.vlm"},
	{0x23537E3D456F815BULL, "6052952_m.fxe"},
	{0x2353AD7583EE9A06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_enter.evt"},
	{0x235426234A7E77FBULL, "art\\vfx\\textures\\vfx_leaf_fb.dds"},
	{0x235428927F1ACF81ULL, "vxw_encounter1_ek_0_0.vlm"},
	{0x2354D8C8BC3603C9ULL, "zz_and_debug6178607.cl"},
	{0x235533A325E586EAULL, "444673895.ogg"},
	{0x23554E3785C4E578ULL, "52319134.ogg"},
	{0x2355790B68F2B202ULL, "zz_dae_debug6029907.cl"},
	{0x2355CC00AA23ABE1ULL, "8632330.ogg"},
	{0x2355E1E3BC2411ECULL, "6089959_m.fxe"},
	{0x2356DAD6287ED25DULL, "6077344_m.fxe"},
	{0x23579ABA52B4A84CULL, "813688455.ogg"},
	{0x2357A29BF219E6FEULL, "gen_im_cns_flu_merril_02.uti"},
	{0x23582651A297AAB2ULL, "hlac_lt_0_lowtown_-65539_0z.dds"},
	{0x23589F949B49B17FULL, "gen00fl_sebastian6179768.cl"},
	{0x235A4CDFD2179431ULL, "gen00fl_anders6125542.cl"},
	{0x235AC47583F4E051ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_enter.gad"},
	{0x235B3C5094928A3FULL, "34248712.ogg"},
	{0x235B6947CEDF1B0AULL, "vxlt_0_lowtown_docks_n_bb9_0_0.vlm"},
	{0x235BA075205EF5AFULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldb.mao"},
	{0x235BAC75205F0A4CULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldb.mmh"},
	{0x235BB275205F143EULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldb.msh"},
	{0x235C559CF21B80E7ULL, "vxht_0_hightown_a2_ajm_0_0.vlm"},
	{0x235CF06F291D181FULL, "205182614.ogg"},
	{0x235D0CAB07498EDBULL, "618567042.ogg"},
	{0x235D710A4B0D5ADEULL, "6162034_m.fxe"},
	{0x235D73528EF46818ULL, "152644972.ogg"},
	{0x235DD3D7CD2690E9ULL, "vxlt_house_fw_0_0.vlm"},
	{0x235E552B36D0C342ULL, "vxlt_0_lowtown_a2_ei5_0_0.vlm"},
	{0x235E6075CEAFDB8DULL, "504559885.ogg"},
	{0x235EFD8DA33C89D5ULL, "mag131_taraneh6088696.cl"},
	{0x235EFEEFD0C0C922ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp03.msh"},
	{0x235F00EFD0C0CCD8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp03.mmh"},
	{0x235F43C75C24D99BULL, "310624779.ogg"},
	{0x235F445133345690ULL, "gen00fl_isabela6085858.cl"},
	{0x235F6825A8EA3CDBULL, "vxht_chantry_2ou_0_0.vlm"},
	{0x235FA7BD3AB9A572ULL, "hlac_ht_brothel_n_3369_0x.dds"},
	{0x235FC1665B800BCEULL, "vxw_bonepit_kc_0_0.vlm"},
	{0x2360AC1F29F1F9D0ULL, "6013295_m.fxe"},
	{0x2360B5FFCCFEA642ULL, "vxht_keep_n_6g3_0_0.vlm"},
	{0x2360C57583F94DFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_enter.ani"},
	{0x2360E2E82BCEA045ULL, "6068004_f.fxe"},
	{0x236181907C5057E1ULL, "llac_ht_mansion2b_818_25.dds"},
	{0x23629A6F7110E801ULL, "vxw_encounter1_jj_0_0.vlm"},
	{0x2362AF114D0DA19BULL, "471264032.ogg"},
	{0x2362EEBBC1BD7144ULL, "mag100_cullen6093741.cl"},
	{0x23630918CB890A5BULL, "vxdr_0_deeproads_clv_0_0.vlm"},
	{0x2363A501E98D4A31ULL, "hlac_lt_warehouse_n_1_5y.dds"},
	{0x23644F2EBD06A9CCULL, "6065319_m.fxe"},
	{0x236460480D3A669EULL, "vxlt_0_lowtown_a3_n_c_fij_0_0.vlm"},
	{0x2364CA890C82DD53ULL, "6112168_m.fxe"},
	{0x2364F4D24A221C73ULL, "hlac_ht_0_hightown_n_65532_0z.dds"},
	{0x23651EAC3398855EULL, "terrain_types.gda"},
	{0x236545AAACE75F7BULL, "mcr310cr_templar_lyl_ldr.utc"},
	{0x2365EC58D334F299ULL, "6094726_f.fxe"},
	{0x23660C8CC3AAA427ULL, "hla_ht_0_hightown_-65535_0z.dds"},
	{0x23668B4D9250BF4BULL, "vxw_encounter3_fo_0_1.vlm"},
	{0x2366F463F606E2C6ULL, "mage_arcanemastery.dds"},
	{0x2366FDA4B9E7F2CEULL, "vxlt_0_lowtown_n_f5m_0_0.vlm"},
	{0x236729B7900A0128ULL, "vxw_encounter2_camp_oq_0_0.vlm"},
	{0x23685A5F58DEC119ULL, "6105820_f.fxe"},
	{0x2369131DFD11A292ULL, "zz_dog_debug6204126.cl"},
	{0x236966946F57C0B2ULL, "347896816.ogg"},
	{0x23696EE91770D6D9ULL, "gen00pt_generic_actions.ncs"},
	{0x2369B215A15B34EDULL, "6118951_m.fxe"},
	{0x236A07EFD0CAE063ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp03.phy"},
	{0x236A653AD2CEC218ULL, "6099349_m.fxe"},
	{0x236AD6683F701FEDULL, "25459028.ogg"},
	{0x236B0CBC99B16F1FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge_1.mao"},
	{0x236B28E075698B60ULL, "zz_qun_debug6085729.cl"},
	{0x236B498A495414B3ULL, "506163192.ogg"},
	{0x236BE2E73D03735EULL, "6167334_m.fxe"},
	{0x236C8732967A4FCBULL, "6101394_f.fxe"},
	{0x236D735BC4A470AAULL, "lgt220st_mag_templar_ambush.stg"},
	{0x236D8ADAA1CE40EFULL, "6106402_m.fxe"},
	{0x236DDA029AE0555BULL, "vxht_chantry_2ht_0_0.vlm"},
	{0x236E144ADB115373ULL, "llac_lt_0_lowtown_n_65535_3.dds"},
	{0x236E1826A839658FULL, "qcr220_grey_wardens6060484.cl"},
	{0x236E9552E99894C0ULL, "zz_seb_debug6215742.cl"},
	{0x236EA5AE42016008ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_beardm1.mao"},
	{0x236EDDDAAF307FA5ULL, "vxw_encounter1_dt_0_0.vlm"},
	{0x236EF153586EFDD7ULL, "277460759.ogg"},
	{0x236F008524C2E1C0ULL, "vxca_cavern1_haunted_fmm_0_0.vlm"},
	{0x236F00E5378C8FC8ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb01a.msh"},
	{0x236F1EE5378CC2F2ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb01a.mmh"},
	{0x2370165E428EF5BCULL, "6115571_m.fxe"},
	{0x23712E1BAD03E327ULL, "hlac_lt_0_lowtown_a3_n_c_0_2x.dds"},
	{0x23724B9E4CF3C62DULL, "w_sundermount_226_0.mmh"},
	{0x23724D7698629407ULL, "vxdr_primevaltomb_6pa_0_0.vlm"},
	{0x23724F9E4CF3CD7EULL, "w_sundermount_226_0.mao"},
	{0x23725D9E4CF3E4B3ULL, "w_sundermount_226_0.msh"},
	{0x23727734F9BB2A1BULL, "llac_lt_0_lowtown_a2_n_c_-3_3.dds"},
	{0x23731EC2A2D9DEF2ULL, "vxlt_0_lowtown_a3_n_c_ei5_0_0.vlm"},
	{0x237363027A328E04ULL, "470406403.ogg"},
	{0x2373DFA55066467EULL, "vxlt_backalley_n_1vu_0_1.vlm"},
	{0x23745B396C481D94ULL, "fen200ip_second_ambush_exit.ncs"},
	{0x2374E6432BBB5CDFULL, "603364737.ogg"},
	{0x237502F7FE72CF4AULL, "mcr370_cut_player_rules6098299.cl"},
	{0x237526DAEFA8997DULL, "vxca_caverns1_gkr_0_0.vlm"},
	{0x23753324E40C862EULL, "vxht_bartrand_n_4g2_0_0.vlm"},
	{0x2375483A69FAC98DULL, "hlac_w_bonepitdestroyed_a3_4_0y.dds"},
	{0x23759B94820019A8ULL, "vxw_encounter2_camp_jn_0_0.vlm"},
	{0x2376A81B8A87769BULL, "dae000cr_grand_cleric6124706.cl"},
	{0x23775C83C4C564CDULL, "367338571.ogg"},
	{0x2377B8014C216A2FULL, "vxw_bonepitdestroyed_a3_1rz_0_0.vlm"},
	{0x2377DBE2F6A4E313ULL, "gen00fl_anders6125221.cl"},
	{0x2378420590B1F574ULL, "6036311_m.fxe"},
	{0x2378C2D3EF362491ULL, "6099313_m.fxe"},
	{0x2379307C585091E8ULL, "6115816_m.fxe"},
	{0x23793FDEF3602871ULL, "6139507_m.fxe"},
	{0x2379454541769540ULL, "341436748.ogg"},
	{0x2379C66B42A1690DULL, "1676946.ogg"},
	{0x2379F95C57740A1FULL, "650361341.ogg"},
	{0x237A3C14AF5891CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v6.evt"},
	{0x237AE50571DDABF0ULL, "llac_w_bonepitmining_a2_327686_0.dds"},
	{0x237AEF9F5033F488ULL, "6166717_m.fxe"},
	{0x237B379F1681D365ULL, "framebuffereffects_distortion_p0.psh.fxo"},
	{0x237B9A49DAAD092CULL, "mrl150_merrill6070798.cl"},
	{0x237BF2AA281136E2ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_019a_0s.dds"},
	{0x237C7067D479234CULL, "hm_qmq_minion5.mor"},
	{0x237CBB811477DB5CULL, "mag100ip_corpse_1.utp"},
	{0x237CD2A3A66754EBULL, "vxht_0_hightown_n_ajv_0_0.vlm"},
	{0x237CF2394A35CD8EULL, "vxht_keep_n_67o_0_0.vlm"},
	{0x237D015468BAF502ULL, "vxw_bonepitdestroyed_a3_1c8_0_0.vlm"},
	{0x237D48880F8FD00FULL, "vxlt_warehouse_n_10z_0_0.vlm"},
	{0x237D4FB7A1269825ULL, "vxw_encounter1_is_0_0.vlm"},
	{0x237D6A21F35FDA1FULL, "6099522_m.fxe"},
	{0x237D8CF5DD17C158ULL, "614192885.ogg"},
	{0x237DC2214E129CAAULL, "vxdr_0_deeproads_esl_0_0.vlm"},
	{0x237DC9B56F624EE3ULL, "701816789.ogg"},
	{0x237E5AB475AA33B5ULL, "6092258_f.fxe"},
	{0x237EA372298B43AAULL, "6075139_f.fxe"},
	{0x237EACA71D4758C1ULL, "vxlt_darktown_5o3_0_0.vlm"},
	{0x237F269E0553AE29ULL, "vxlt_gallowstemplar_476_0_0.vlm"},
	{0x237F303B0C8A08FCULL, "lt_hangedman_n_3567_lc.mmh"},
	{0x237F9A6E681A53BEULL, "vxlt_0_lowtown_ekj_0_1.vlm"},
	{0x23809D64B0D70EC6ULL, "6116680_m.fxe"},
	{0x23810617C74F5E28ULL, "hlac_ca_caverns1_mines_24425_3x.dds"},
	{0x23810E5287C62C3FULL, "llac_lt_0_lowtown_docks_n_65533_4.dds"},
	{0x238120A2F85A4CF5ULL, "hm_ser_emeric.mor"},
	{0x23815514AF5EDA8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v6.gad"},
	{0x2381FC96463A9FC3ULL, "512600070.ogg"},
	{0x238285CB37E17CA2ULL, "llac_lt_hangedman_-1_25.dds"},
	{0x2382A7B84A10F500ULL, "character_character_p155.psh.fxo.dx11.assembly"},
	{0x2382BAC6B29195B3ULL, "6121092_m.fxe"},
	{0x2383A501D6029EAEULL, "vxht_bartrand_n_4l3_0_0.vlm"},
	{0x2383CF2AEA609FE1ULL, "21646549.ogg"},
	{0x2383D4EA2146141DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_basic.phy"},
	{0x23840D7173F63228ULL, "vxw_encounter2_camp_eo_0_0.vlm"},
	{0x238411AF56A411B4ULL, "hlac_lt_0_lowtown_docks_a3_-1_2y.dds"},
	{0x238438FD7BCED708ULL, "w_encounter3_209_0.msh"},
	{0x23844AFD7BCEF599ULL, "w_encounter3_209_0.mao"},
	{0x238456FD7BCF0A32ULL, "w_encounter3_209_0.mmh"},
	{0x23846A9BDF0B608FULL, "702483803.ogg"},
	{0x2384762B2D8C561BULL, "llac_lt_0_lowtown_n_-131072_2.dds"},
	{0x2385AACD928B9889ULL, "6138805_m.fxe"},
	{0x23860D75D0697079ULL, "zz_followers_banter6158169.cl"},
	{0x2386839ABA38DC03ULL, "240095450.ogg"},
	{0x238694ABF5A83B54ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_c.anb"},
	{0x2386E54E7C8BEBACULL, "6148116_f.fxe"},
	{0x23870CAC27CCC2EFULL, "vxlt_0_lowtown_a2_n_c_hwr_0_0.vlm"},
	{0x23875414AF63448AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v6.ani"},
	{0x2387ADD68B6E227FULL, "hlac_lt_darktown_-1_82y.dds"},
	{0x2388141A57969E71ULL, "730331420.ogg"},
	{0x23882514C6884754ULL, "vxca_cave_19o_0_0.vlm"},
	{0x238847B871E582CFULL, "428893218.ogg"},
	{0x238880692FBB351BULL, "vxht_0_hightown_a2_b5c_0_0.vlm"},
	{0x23889A42790DCF8EULL, "tow_ex02.mao"},
	{0x2389093215126F41ULL, "6062812_m.fxe"},
	{0x23891114463E8003ULL, "737010331.ogg"},
	{0x23893B28EF65D801ULL, "6095857_f.fxe"},
	{0x238997AAACD3F1E5ULL, "91904631.ogg"},
	{0x2389A2B766D60B91ULL, "6103959_m.fxe"},
	{0x2389AC5C8748F90EULL, "art\\characters\\weapons\\axes\\w_axe_04a.msh"},
	{0x2389BA5C8749115FULL, "art\\characters\\weapons\\axes\\w_axe_04a.mao"},
	{0x2389C65C874925FCULL, "art\\characters\\weapons\\axes\\w_axe_04a.mmh"},
	{0x2389EB5AD497FF85ULL, "517134512.ogg"},
	{0x238A1567A0A70619ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_san01.phy"},
	{0x238AABD06BA73E7AULL, "qun111_ketojan6019031.cl"},
	{0x238AE3B59084BC64ULL, "art\\cinematics\\deathblows\\rockgolem\\db_pc_rockrogue.gad"},
	{0x238AE4F5EB91C346ULL, "vxlt_darktown_581_0_0.vlm"},
	{0x238AF0C8776CFBA3ULL, "hlac_w_bonepitmining_a2_393216_0y.dds"},
	{0x238B08DD0A95C5E4ULL, "240744292.ogg"},
	{0x238B73315AB10D82ULL, "vxw_bonepitdestroyed_a3_1d9_0_0.vlm"},
	{0x238C3A396C5C77C5ULL, "fen200ip_second_ambush_exit.utp"},
	{0x238D114FD964B25CULL, "182497483.ogg"},
	{0x238D1E840F3D7141ULL, "vxlt_darktown_5p2_0_0.vlm"},
	{0x238D41DCF5DE2E15ULL, "hlac_ht_playershouse_n_855_28z.dds"},
	{0x238D7B07CE3339DEULL, "hlac_lt_0_lowtown_a2_n_-196608_3z.dds"},
	{0x238DCE9BB96E315DULL, "25827388.ogg"},
	{0x238DF0F4DA544918ULL, "1060805022.ogg"},
	{0x238E4F4BE112E3B9ULL, "vxca_caverns1_i1j_0_0.vlm"},
	{0x238E50BD7ECE4E63ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1c.msh"},
	{0x238E5EBD7ECE65BDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1c.mmh"},
	{0x238E62BD7ECE6CCEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1c.mao"},
	{0x238ECDE7444411E6ULL, "386553471.ogg"},
	{0x238F08A32FD1BB51ULL, "6043141_m.fxe"},
	{0x238F1D805235FE67ULL, "zz_dae_debug.cnv"},
	{0x238F20D3C3905A8BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brklarge_1.mao"},
	{0x238F31307C4EC507ULL, "vxdr_primevaltomb_6zc_0_0.vlm"},
	{0x238F4A3CD347310DULL, "vxw_blightlands_c3_0_0.vlm"},
	{0x238F5C3521DC3B03ULL, "6198463_m.fxe"},
	{0x238FAFB6848DB416ULL, "lgt110cr_coa_sailor2.utc"},
	{0x2390D0834B32B641ULL, "782708485.ogg"},
	{0x2391296339D4C50DULL, "143270158.ogg"},
	{0x2391653B2CE3EE5DULL, "vxht_0_hightown_n_c53_0_0.vlm"},
	{0x2391F2B5908AF442ULL, "art\\cinematics\\deathblows\\rockgolem\\db_pc_rockrogue.ani"},
	{0x23923936AEB0D34DULL, "585486308.ogg"},
	{0x239298D57BCC6E04ULL, "m_dae320ar_hightown_nt.wbd"},
	{0x2392D4209798B285ULL, "fen310ip_to_hanged_man.utp"},
	{0x239373ABF5B28D17ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_c.mmh"},
	{0x2393EC19447C873FULL, "94620855.ogg"},
	{0x2394C9EA21547489ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_basic.mao"},
	{0x2394D519C693635DULL, "6166114_m.fxe"},
	{0x2394D5EA21548922ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_basic.mmh"},
	{0x2394D7EA21548CD8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_basic.msh"},
	{0x2395CC89DBB548A2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masd_0n.dds"},
	{0x2395E8146CE69F63ULL, "vxht_0_hightown_a3_b92_0_0.vlm"},
	{0x239696F1B87E5FD4ULL, "vxca_cave_14l_0_0.vlm"},
	{0x2397223E402E9A49ULL, "hlac_w_encounter2_-131067_0z.dds"},
	{0x2397525154B4071EULL, "hlac_lt_0_lowtown_a2_n_131073_1z.dds"},
	{0x239764B138C87940ULL, "hlac_lt_0_lowtown_docks_n_-65540_2x.dds"},
	{0x2397E1B4EF27D8EEULL, "562452260.ogg"},
	{0x239894E0E5009038ULL, "1063984234.ogg"},
	{0x2398EA80158B79A5ULL, "hlac_w_encounter2_camp_-131068_0y.dds"},
	{0x23994880523E9BD4ULL, "zz_dae_debug.ncs"},
	{0x2399AAC510D77CC4ULL, "6093808_m.fxe"},
	{0x239A02FCDCAD3C7EULL, "6087083_m.fxe"},
	{0x239A3444C1B36DF3ULL, "vxht_brothel_n_6bb_0_0.vlm"},
	{0x239AADD236085657ULL, "990609343.ogg"},
	{0x239AED0CD1D20179ULL, "6121814_m.fxe"},
	{0x239C23812D62AB3EULL, "hlac_ca_cavern1_haunted_24421_1z.dds"},
	{0x239C3F7BEFF345CEULL, "vxlt_warehouse_n_107_0_0.vlm"},
	{0x239D0DEA51BD5870ULL, "lt_0_lowtown_docks_a3_38_0_lc.mmh"},
	{0x239DBC19C53D498DULL, "vxw_blightlands_d2_0_0.vlm"},
	{0x239DF15CAE6C5A66ULL, "lt_0_lowtown_a2_n_18_0.rml"},
	{0x239E06D5616F006AULL, "vxht_0_hightown_c7y_0_0.vlm"},
	{0x239E9ABF81D90BA5ULL, "qcr000_arishok6059991.cl"},
	{0x239EDCB915040F78ULL, "6174053_m.fxe"},
	{0x239EE3022199E6C8ULL, "vxw_bonepit_1wo_0_1.vlm"},
	{0x239FED76AA97D1EEULL, "403994407.ogg"},
	{0x239FFE40BFCCEC2DULL, "960077130.ogg"},
	{0x23A07099135EFB9BULL, "6177183_m.fxe"},
	{0x23A08A2F22714765ULL, "722322803.ogg"},
	{0x23A08FBA8782805DULL, "6156517_m.fxe"},
	{0x23A0998C0714CC0DULL, "dae000_mother6121029.cl"},
	{0x23A0FCE310D9DB36ULL, "205708264.ogg"},
	{0x23A2395095469AA9ULL, "dae300_bran6143439.cl"},
	{0x23A270E527A20C21ULL, "vxht_mansion_n_3hh_0_0.vlm"},
	{0x23A2F35BA43C1CC4ULL, "297910462.ogg"},
	{0x23A306876F6609E5ULL, "632781910.ogg"},
	{0x23A35F98A8D7A2F2ULL, "zz_fen_debug6215347.cl"},
	{0x23A3A458BE728CF1ULL, "hlac_lt_backalley_g_0_1z.dds"},
	{0x23A4469F94303D45ULL, "hlac_w_bonepitmining_a2_1_0z.dds"},
	{0x23A543417EAD1C29ULL, "23627047.ogg"},
	{0x23A55EC923813546ULL, "961600257.ogg"},
	{0x23A57B3496192FB6ULL, "vxht_chantry_n_28j_0_0.vlm"},
	{0x23A59EF61E9F94ACULL, "6070007_m.fxe"},
	{0x23A5C4FA43C89B27ULL, "6042005_m.fxe"},
	{0x23A5E0ED3A7F0799ULL, "vxlt_0_lowtown_a3_n_c_eoy_0_1.vlm"},
	{0x23A5EFD43EF601EEULL, "vxlt_gallowstemplar_n_39k_0_0.vlm"},
	{0x23A626E702389F71ULL, "vxca_caverns1_iiz_0_0.vlm"},
	{0x23A62B9A1F5507D8ULL, "lt_gallowstemplar_n_probe_2qy.dds"},
	{0x23A691C635B7927AULL, "w_sundermount_214_0.mmh"},
	{0x23A695C635B798B1ULL, "w_sundermount_214_0.mao"},
	{0x23A6A3C635B7B0C0ULL, "w_sundermount_214_0.msh"},
	{0x23A6E9E2E3F36D83ULL, "zz_dae_debug6206729.cl"},
	{0x23A730CEC15013ACULL, "zz_qcr_debug6215496.cl"},
	{0x23A7A17662B8B75BULL, "vxht_chantry_2th_0_0.vlm"},
	{0x23A7C1C4C4B6CAB3ULL, "6134928_m.fxe"},
	{0x23A7FFF98D1D4A18ULL, "vxlt_darktown_4w7_0_0.vlm"},
	{0x23A83D2DE0A88F75ULL, "mag101_grace6034374.cl"},
	{0x23A84DB32EF4E92FULL, "6170469_m.fxe"},
	{0x23A8658E503FE6E4ULL, "fx_darkbolt_blend.mao"},
	{0x23A8ADB78C9FD752ULL, "vxlt_0_lowtown_h98_0_0.vlm"},
	{0x23A917B823F4E62AULL, "vxdr_0_deeproads_ebo_0_0.vlm"},
	{0x23A93A81F94679E8ULL, "vxlt_undercitydungeon_144_0_0.vlm"},
	{0x23A95746AE6FFF1DULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_004a_0n.dds"},
	{0x23AA51C0A1F3052BULL, "gen00fl_merrill6128938.cl"},
	{0x23AA93DD0212F382ULL, "hlac_w_blightlands_196610_0x.dds"},
	{0x23AA9D9FF6985982ULL, "w_bonepitmining_a2_201_0.phy"},
	{0x23ABBE89B2A622F3ULL, "vxw_blightlands_9o_0_0.vlm"},
	{0x23ABD5607A3EBE21ULL, "dre000_varric6027779.cl"},
	{0x23ABF6E49C017FC9ULL, "rob_cot.tnt"},
	{0x23AC2DF6B733620DULL, "vxw_blightlands_i1_0_0.vlm"},
	{0x23AC560E67599899ULL, "mag210_arianni6031705.cl"},
	{0x23AC673691513D4EULL, "one200dg_solivitus6048200.cl"},
	{0x23AC67C411C6F09FULL, "vxht_bartrand_n_5ql_0_0.vlm"},
	{0x23AC7B839FC85CA4ULL, "6135159_m.fxe"},
	{0x23AC7B8C15DE738BULL, "860327876.ogg"},
	{0x23ACD3C0E7F74EDAULL, "vxca_caverns2_7kj_0_0.vlm"},
	{0x23AD4028AB31E04EULL, "dae000cr_grand_cleric6181915.cl"},
	{0x23AD43B335EA2507ULL, "gen00fl_sebastian6179984.cl"},
	{0x23AD57C899745EE1ULL, "6131162_f.fxe"},
	{0x23ADD443FEA1169BULL, "23013888.ogg"},
	{0x23AE5C758B4632B3ULL, "zz_mrl_debug6184215.cl"},
	{0x23AEB78E20FCD2BAULL, "1023821111.ogg"},
	{0x23AF32A2D09220C4ULL, "vxca_caverns2_8o3_0_0.vlm"},
	{0x23AF845C5ADD7402ULL, "vxlt_0_lowtown_ghl_0_0.vlm"},
	{0x23AFA55C8769BCC7ULL, "art\\characters\\weapons\\axes\\w_axe_04a.phy"},
	{0x23AFC0081DA98A50ULL, "6095161_m.fxe"},
	{0x23AFF4C83FD5410EULL, "109751588.ogg"},
	{0x23AFFF2E64FF3FDFULL, "hlac_lt_hangedman_-1_19x.dds"},
	{0x23B08161DBD15A91ULL, "vxw_encounter2_e2_0_0.vlm"},
	{0x23B08BBCDB38B9D3ULL, "vxlt_0_lowtown_eqd_0_0.vlm"},
	{0x23B08EBBF5FC6717ULL, "vxca_caverns2_78p_0_2.vlm"},
	{0x23B0AFFDC4EC3DB8ULL, "wm_ico_freemarches.dds"},
	{0x23B0CA5A4CB39CA4ULL, "zz_approval_debug6003557.cl"},
	{0x23B1BD3E409CE4D6ULL, "hlac_ht_chantry_2534_40y.dds"},
	{0x23B1C39E5EBC2743ULL, "6188615_m.fxe"},
	{0x23B1F7E4E8A25EC4ULL, "6164023_m.fxe"},
	{0x23B21EB8DE74EE66ULL, "hlac_lt_0_lowtown_a2_n_65537_0x.dds"},
	{0x23B26C5067ED6189ULL, "t3_arm_hst.tnt"},
	{0x23B286664B3B2295ULL, "vxlt_0_lowtown_docks_a3_cyf_0_0.vlm"},
	{0x23B29B9B3181F22EULL, "589354723.ogg"},
	{0x23B3B2A4E010F3B7ULL, "hlac_lt_warehouse_65537_0x.dds"},
	{0x23B3B898272BBDF0ULL, "gen00fl_sebastian6174818.cl"},
	{0x23B3B8D9E7A6A62DULL, "vxht_bartrand_n_68z_0_0.vlm"},
	{0x23B3C40265BFC0E3ULL, "hlac_lt_0_lowtown_65535_8x.dds"},
	{0x23B450D2795250F8ULL, "vxw_woundedcoast_1t8_0_0.vlm"},
	{0x23B461B130EC1A6EULL, "vxlt_gallowstemplar_n_34l_0_0.vlm"},
	{0x23B4AF25BC911015ULL, "lt_0_lowtown_docks_n_11_0_lc.mmh"},
	{0x23B4F82F46441178ULL, "narrative_flemeth6002796.cl"},
	{0x23B5273432E2606AULL, "vxlt_gallowstemplar_45w_0_0.vlm"},
	{0x23B5792AF6C2E53CULL, "fen310_varania.cnv"},
	{0x23B58FF9C3468132ULL, "6137638_m.fxe"},
	{0x23B5AE84D60A9E7DULL, "vxlt_0_lowtown_docks_a3_d0v_0_0.vlm"},
	{0x23B5BB398AF759C0ULL, "vxlt_0_lowtown_a2_n_c_e1r_0_0.vlm"},
	{0x23B61CD7DC9AD625ULL, "475750733.ogg"},
	{0x23B6BE740F35BD4CULL, "437965123.ogg"},
	{0x23B725C62478D01BULL, "vxdr_primevaltomb_3tl_0_0.vlm"},
	{0x23B7AB08D7FD3818ULL, "6217132_m.fxe"},
	{0x23B7DEC82CC776ECULL, "zz_mcr_debug6091902.cl"},
	{0x23B83E63973231ACULL, "vxlt_0_lowtown_a2_n_gz7_0_0.vlm"},
	{0x23B86E35E952EDE4ULL, "hlac_ht_0_hightown_-1_1z.dds"},
	{0x23B86E520449A7DBULL, "176073043.ogg"},
	{0x23B896BE90A89C7FULL, "6095244_f.fxe"},
	{0x23B8CC3019D278E9ULL, "gen00fl_sebastian6188652.cl"},
	{0x23B8F88E71C744A1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb03.phy"},
	{0x23B97B4A146150D8ULL, "vxlt_gallowsprison_n_6gb_0_0.vlm"},
	{0x23B99B2AF6C7078CULL, "fen310_varania.bnk"},
	{0x23BA0C729C8D5852ULL, "vxlt_0_lowtown_eqa_0_0.vlm"},
	{0x23BA9FD3A9297A8DULL, "vxw_blightlands_r0_0_0.vlm"},
	{0x23BAD9A103BD091FULL, "vxht_bartrand_n_5to_0_0.vlm"},
	{0x23BADA2F2FFB5C56ULL, "gen_im_npc_arm_hum_cht_krka.uti"},
	{0x23BB0986A780FC4DULL, "rune_armor_balance.dds"},
	{0x23BB1CAF22A60E1AULL, "gen_im_arm_boo_med_act2_03.uti"},
	{0x23BBA1E03B764CC0ULL, "1047328489.ogg"},
	{0x23BC11E7753211B3ULL, "art\\characters\\creatures\\strider\\animation\\groundstompexit.ani"},
	{0x23BD5034185027D6ULL, "6154163_m.fxe"},
	{0x23BD9B2A162C6CC4ULL, "vxht_0_hightown_be4_0_0.vlm"},
	{0x23BDE584CE4376D1ULL, "dae030cr_fem_elf_sta.utc"},
	{0x23BE222E6C40D2B1ULL, "cyc_cloud2.dds"},
	{0x23BE3375D0136B2DULL, "388470368.ogg"},
	{0x23BF12A56FD1A5F9ULL, "dae000ip_crowd_mf5_talk4.utp"},
	{0x23BF8B6BCABE7894ULL, "67325748.ogg"},
	{0x23C05FA5541A6FDBULL, "llac_w_sundermount_-393213_0.dds"},
	{0x23C07AA430CA18A3ULL, "hlac_lt_0_lowtown_a2_n_131071_4z.dds"},
	{0x23C0A98C01A36D7DULL, "hlac_lt_0_lowtown_n_0_5x.dds"},
	{0x23C0C517C3226A37ULL, "mcr370_meredith6083111.cl"},
	{0x23C0CC60F645D387ULL, "cd_seb201_lady_harimann.wbd"},
	{0x23C1ECF340AEE89CULL, "215833099.ogg"},
	{0x23C277CD32DB58AFULL, "884217416.ogg"},
	{0x23C2CB82654171F0ULL, "vxlt_0_lowtown_a2_n_c_e27_0_0.vlm"},
	{0x23C3AA8E43B3584FULL, "93774780.ogg"},
	{0x23C42382B7F0B9DDULL, "vxw_woundedcoast_nv_0_0.vlm"},
	{0x23C44BBD886DD50BULL, "6060608_f.fxe"},
	{0x23C45D5B63285286ULL, "ht_0_hightown_n_probe_9x4.dds"},
	{0x23C45D6CA5784D9EULL, "811273057.ogg"},
	{0x23C4B10D26A74BDEULL, "561047940.ogg"},
	{0x23C4E3B371097B18ULL, "vxlt_darktown_4y5_0_0.vlm"},
	{0x23C4EB3D2BC26CE7ULL, "vxlt_0_lowtown_n_eop_0_1.vlm"},
	{0x23C5473B348A1A67ULL, "l_fen200ar_slavers_exterior.bnk"},
	{0x23C5486124069CA9ULL, "character_character_p154.psh.fxo.dx11.assembly"},
	{0x23C597A3166EE89BULL, "vxdr_primevaltomb_3io_0_0.vlm"},
	{0x23C5CAA129F6EB13ULL, "hlac_ht_0_hightown_n_-65538_1z.dds"},
	{0x23C617F83202D751ULL, "zz_bet_debug6113387.cl"},
	{0x23C6FE857B22C220ULL, "83943599.ogg"},
	{0x23C79798DEC1E29CULL, "gen00fl_fenris6106603.cl"},
	{0x23C84D6BDFEB81A5ULL, "vxw_bonepitdestroyed_a3_26c_0_0.vlm"},
	{0x23C88B83BE598BCCULL, "6095430_f.fxe"},
	{0x23C8A989FB56364FULL, "6072965_m.fxe"},
	{0x23C8BBA39CC50221ULL, "vxlt_hangedman_n_1qs_0_0.vlm"},
	{0x23C8C9C059406897ULL, "llac_ht_keep_n_3732_13.dds"},
	{0x23C8F441BF28AE9AULL, "6139218_m.fxe"},
	{0x23C9A145AD1FD587ULL, "w_encounter2_camp_20_0.rml"},
	{0x23C9C483A8D413F3ULL, "vxw_sundermount_26r_0_0.vlm"},
	{0x23CA09E7753E41EAULL, "art\\characters\\creatures\\strider\\animation\\groundstompexit.evt"},
	{0x23CA1DE537D96CEDULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb01a.phy"},
	{0x23CAEE96DE03C0FAULL, "hlac_ht_0_hightown_n_1_1z.dds"},
	{0x23CB3030B576D2EBULL, "712103142.ogg"},
	{0x23CBB0C635D6E285ULL, "w_sundermount_214_0.phy"},
	{0x23CBC336169C451AULL, "zz_ave_debug6174537.cl"},
	{0x23CBDDE45E569AB8ULL, "vxlt_gallowscourtyard_n_62g_0_0.vlm"},
	{0x23CBE2574F14D009ULL, "vxht_mansion2_2ee_0_0.vlm"},
	{0x23CC165CB47E51C4ULL, "vxca_caverns2_8q1_0_0.vlm"},
	{0x23CC3A737D05423DULL, "hlac_w_blightlands_-65536_0x.dds"},
	{0x23CD2299E717B7A7ULL, "vxw_bonepitmining_a2_4q_0_0.vlm"},
	{0x23CD558F270330C6ULL, "amb_scene_dogbark.bnk"},
	{0x23CD651BBFBD8B91ULL, "vxw_encounter2_k0_0_0.vlm"},
	{0x23CDC66293A5D82AULL, "313479333.ogg"},
	{0x23CE041759549959ULL, "lt_0_lowtown_38_0_ll.mmh"},
	{0x23CFA045AD243F9FULL, "w_encounter2_camp_20_0.phy"},
	{0x23CFAD5F2A5DD0DCULL, "6087016_m.fxe"},
	{0x23CFE405824AFD31ULL, "648610219.ogg"},
	{0x23D0080841B3B7DCULL, "gen00fl_varric6081977.cl"},
	{0x23D00BEED08A7935ULL, "dae000_pc_home_ambients6208472.cl"},
	{0x23D01049EAEF815FULL, "vxlt_gallowscourtyard_6ti_0_0.vlm"},
	{0x23D0257BA9844C95ULL, "zz_mer_debug6073558.cl"},
	{0x23D03057FBEC5110ULL, "gen00fl_aveline6121946.cl"},
	{0x23D0789FF6B8EA36ULL, "w_bonepitmining_a2_201_0.mao"},
	{0x23D0849FF6B8FED5ULL, "w_bonepitmining_a2_201_0.mmh"},
	{0x23D0869FF6B9024BULL, "w_bonepitmining_a2_201_0.msh"},
	{0x23D09AC2DB9CCD19ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.open.ani"},
	{0x23D0E8E77544285DULL, "art\\characters\\creatures\\strider\\animation\\groundstompexit.gad"},
	{0x23D18A6F91B85482ULL, "vxdr_primevaltomb_8y9_0_0.vlm"},
	{0x23D19E590A28AF3CULL, "gen00fl_merrill6131348.cl"},
	{0x23D20F257A5EA1F9ULL, "412766997.ogg"},
	{0x23D294B928F3C124ULL, "llac_lt_0_lowtown_a2_0_5.dds"},
	{0x23D2955FA9E6D25DULL, "vxw_woundedcoast_ku_0_0.vlm"},
	{0x23D2BEC681651FD8ULL, "vxht_mansion_n_4xb_0_0.vlm"},
	{0x23D2E77CA78269E8ULL, "360197594.ogg"},
	{0x23D34AA4134388E7ULL, "hlac_ht_bartrand_n_3366_1y.dds"},
	{0x23D3664976E9E348ULL, "552190569.ogg"},
	{0x23D3827278C816CEULL, "104731428.ogg"},
	{0x23D38E43C921DE91ULL, "512677553.ogg"},
	{0x23D39B9907371699ULL, "361132227.ogg"},
	{0x23D3A98541AC0484ULL, "dae130ip_at_to_sewers1.utp"},
	{0x23D429D091154DEBULL, "405208754.ogg"},
	{0x23D43763ADAE7D3DULL, "ave200_aveline_bar6177331.cl"},
	{0x23D4E933343E6769ULL, "dae_item_aoe_instant.ncs"},
	{0x23D5087D91E7E338ULL, "zz_dae_debug6049994.cl"},
	{0x23D50FC75354AA60ULL, "one300_zevran6089096.cl"},
	{0x23D5372C6B6DA48BULL, "vxht_mansion_n_2bq_0_0.vlm"},
	{0x23D63B856C2AC1C3ULL, "hlac_ca_caverns2prc_5802_18x.dds"},
	{0x23D6BF48D1E19A25ULL, "vxw_bonepitdestroyed_a3_21b_0_0.vlm"},
	{0x23D6C4D1926CE374ULL, "vxca_cave_2j5_0_0.vlm"},
	{0x23D74E6B28EED02FULL, "mrl000_marethari6142173.cl"},
	{0x23D775C2DBA2AD5BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.open.gad"},
	{0x23D7D9494AE64397ULL, "gen00fl_aveline6033190.cl"},
	{0x23D82C6FEC4FDC10ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6_tool.mao"},
	{0x23D8306FEC4FE323ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6_tool.mmh"},
	{0x23D83E6FEC4FFB7DULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6_tool.msh"},
	{0x23DA10C291029AD7ULL, "vxw_woundedcoast_bx_0_0.vlm"},
	{0x23DA5434410AE889ULL, "vxht_mansion2_2nd_0_0.vlm"},
	{0x23DAA9B7C54139D2ULL, "914824770.ogg"},
	{0x23DABEC34C306AFEULL, "mag170_recruit026080923.cl"},
	{0x23DAFC5E06B0C690ULL, "w_bonepitdestroyed_a3_115_0.mao"},
	{0x23DB005E06B0CDA3ULL, "w_bonepitdestroyed_a3_115_0.mmh"},
	{0x23DB0E5E06B0E5FDULL, "w_bonepitdestroyed_a3_115_0.msh"},
	{0x23DC3377E4CD3BB4ULL, "llac_lt_0_lowtown_131072_5.dds"},
	{0x23DC4FDB91D0350BULL, "6041839_m.fxe"},
	{0x23DC984B53EFAEBBULL, "vxca_caverns1_mines_co9_0_0.vlm"},
	{0x23DCF99C1443D9A6ULL, "vxht_keep_4bi_0_0.vlm"},
	{0x23DD9F00DB01BBA2ULL, "nat380_sibling6068067.cl"},
	{0x23DDDD7CC76E32D3ULL, "vxht_chantry_n_2r9_0_0.vlm"},
	{0x23DDFC15371BF1E4ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_knckbck.ani"},
	{0x23DDFE5CC8E5EA60ULL, "gen00fl_aveline6101430.cl"},
	{0x23DEA5AC4D963983ULL, "483481458.ogg"},
	{0x23DEB5D350D98EB1ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_megatiki_n.dds"},
	{0x23DEC8D773E7EF2BULL, "6067725_m.fxe"},
	{0x23DEDF50BAF5786FULL, "vxht_0_hightown_a3_9x6_0_0.vlm"},
	{0x23DF049663B1C7F1ULL, "6147740_m.fxe"},
	{0x23DF1799F3AC8661ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ca_transition.ani"},
	{0x23E0521257251EC7ULL, "mag320_mage_conspirator6040764.cl"},
	{0x23E06420062557DCULL, "tbe000im_sorrow_dagger.uti"},
	{0x23E09E451C8CC2F1ULL, "rdr110dg_guard_warehouse6016687.cl"},
	{0x23E0A145AD32B4A4ULL, "w_encounter2_camp_20_0.mmh"},
	{0x23E0A545AD32BBF7ULL, "w_encounter2_camp_20_0.mao"},
	{0x23E0A647A72F98E6ULL, "scaleform_sf_main_p2.vsh.fxo"},
	{0x23E0B745AD32DA46ULL, "w_encounter2_camp_20_0.msh"},
	{0x23E0D135F6981F4DULL, "vxht_chantry_28b_0_0.vlm"},
	{0x23E0D7E13852C8F5ULL, "hightown_male_elfservnt_mob6170080.cl"},
	{0x23E22D4FDA1749D0ULL, "gen00fl_aveline6131226.cl"},
	{0x23E273493382B327ULL, "6125727_f.fxe"},
	{0x23E2D4CBEAB2DB0FULL, "llac_dr_slavers_3217_11.dds"},
	{0x23E30AD71EB90DAAULL, "hlac_lt_0_lowtown_n_-3_2z.dds"},
	{0x23E3189BB166AA76ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock_sa.dds"},
	{0x23E404B40A6A5CE5ULL, "vxlt_0_lowtown_docks_a3_cy6_0_0.vlm"},
	{0x23E40626E558A54BULL, "6133782_f.fxe"},
	{0x23E44D9A8A613422ULL, "hlac_lt_0_lowtown_-131072_8z.dds"},
	{0x23E46D50563AC923ULL, "6119459_m.fxe"},
	{0x23E472BEC1C84FA8ULL, "dae000_mother6121044.cl"},
	{0x23E4997FDDA758AFULL, "6016808_m.fxe"},
	{0x23E4F91537220B46ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_knckbck.gad"},
	{0x23E512D1B9DD5451ULL, "6084184_m.fxe"},
	{0x23E5D5FC799BE08CULL, "vxht_mansion2_25l_0_0.vlm"},
	{0x23E62299F3B2B703ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ca_transition.gad"},
	{0x23E67FCDA75F8966ULL, "143774011.ogg"},
	{0x23E712AE2E0ACF59ULL, "lgt210_qun_blk_qunari6142398.cl"},
	{0x23E73511D83A3AA4ULL, "6165907_m.fxe"},
	{0x23E7A70D53FDD1A3ULL, "hlac_ht_0_hightown_a3_-65534_0y.dds"},
	{0x23E7AB8F35619466ULL, "999204270.ogg"},
	{0x23E7C842667423A9ULL, "235384837.ogg"},
	{0x23E8AEB2EA1A2BA4ULL, "art\\characters\\creatures\\stonegolem\\animation\\hv_dmg.ani"},
	{0x23E9316FEC5E5848ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6_tool.phy"},
	{0x23E9D14195CAE268ULL, "92155662.ogg"},
	{0x23E9EC4D3B7C19BEULL, "ser221_gamlen6134682.cl"},
	{0x23EA3F2E173E619CULL, "70263033.ogg"},
	{0x23EACA89D398AF39ULL, "249923590.ogg"},
	{0x23EAD07BCEDD25D6ULL, "hlac_lt_0_lowtown_docks_a3_-2_6z.dds"},
	{0x23EB9FCB579AB8D7ULL, "vxw_bonepit_1sa_0_0.vlm"},
	{0x23EBF41537282125ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_knckbck.evt"},
	{0x23EC5E75C5555D18ULL, "309647755.ogg"},
	{0x23ECFF99F3B89AE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ca_transition.evt"},
	{0x23ED767355A7E071ULL, "6140640_m.fxe"},
	{0x23EE7F016BB3EB10ULL, "dre000_cut_sibling_dies6036341.cl"},
	{0x23EEABB2EA1E9206ULL, "art\\characters\\creatures\\stonegolem\\animation\\hv_dmg.gad"},
	{0x23EEAEBF8F0198FBULL, "hla_ht_0_hightown_n_-65536_0x.dds"},
	{0x23EED3A26DE1FD63ULL, "hlac_lt_hangedman_n_-1_5z.dds"},
	{0x23EF504852C2AD1AULL, "ht_0_hightown_n_21_0.anb"},
	{0x23F05CC97E0E23ECULL, "zz_qun_debug6045404.cl"},
	{0x23F0EFE850722DB5ULL, "6081849_f.fxe"},
	{0x23F13D3B34AF439DULL, "l_fen200ar_slavers_exterior.wbd"},
	{0x23F25957FAE05F66ULL, "gen00fl_anders6119096.cl"},
	{0x23F2996606D0C8E7ULL, "6155745_m.fxe"},
	{0x23F3929C03D48A5AULL, "vxw_sundermount_30m_0_1.vlm"},
	{0x23F3C554DF2AC0B1ULL, "551835859.ogg"},
	{0x23F3EDBDBBBE27A3ULL, "one220_yevhen6102817.cl"},
	{0x23F478D8218632A9ULL, "vxlt_0_lowtown_n_e1a_0_0.vlm"},
	{0x23F49F722FF442C8ULL, "236504882.ogg"},
	{0x23F4A114CB956B1FULL, "vxht_bartrand_n_5hk_0_0.vlm"},
	{0x23F4F0552613F67CULL, "dae000_arishok6068453.cl"},
	{0x23F5779D2EE87E6CULL, "vxlt_0_lowtown_docks_a3_n_d2r_0_0.vlm"},
	{0x23F57C1EFC16FED9ULL, "hlac_lt_0_lowtown_docks_-2_5y.dds"},
	{0x23F5A6B2EA24A8E5ULL, "art\\characters\\creatures\\stonegolem\\animation\\hv_dmg.evt"},
	{0x23F5ACCCC24677BDULL, "dr_0_deeproads.arl"},
	{0x23F5CA8254C2F02EULL, "lt_0_lowtown_docks_a3_123_0.phy"},
	{0x23F5D647C1948B27ULL, "mag101st_grace_inside.stg"},
	{0x23F60A0D718D091DULL, "pro000_colonel_ewald6040546.cl"},
	{0x23F6A694053F5318ULL, "6060604_m.fxe"},
	{0x23F6F47C74EECBD7ULL, "vxw_woundedcoast_tz_0_0.vlm"},
	{0x23F76741BB29D7F6ULL, "vxw_encounter2_camp_51_0_0.vlm"},
	{0x23F7A0A56C7455C7ULL, "weapon2.vsh.fxo"},
	{0x23F7A93FE14899F9ULL, "llac_lt_0_lowtown_a2_n_131071_0.dds"},
	{0x23F89130D4668E66ULL, "art\\characters\\creatures\\ragedemon\\animation\\die_v1.ani"},
	{0x23F89F48EFBE6396ULL, "vxht_bartrand_n_5vp_0_0.vlm"},
	{0x23F9958073E77BA4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopen.ani"},
	{0x23F9B76BE60067B6ULL, "vxlt_0_lowtown_a3_n_c_f56_0_0.vlm"},
	{0x23FA7DF294B44777ULL, "hlac_w_bonepitdestroyed_a3_327682_0y.dds"},
	{0x23FABFF0BEAAF472ULL, "6031165_m.fxe"},
	{0x23FB2A6F2068EA75ULL, "gen00fl_isabela6097173.cl"},
	{0x23FB5FFEBCC2462CULL, "rock_02.mao"},
	{0x23FB7406D110173AULL, "hlac_lt_0_lowtown_a2_-3_7x.dds"},
	{0x23FB7FE9B7101876ULL, "6093495_m.fxe"},
	{0x23FBB2DBCC59EB12ULL, "753986246.ogg"},
	{0x23FBC0EDBAC065B2ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb101a_0s.dds"},
	{0x23FBF8AA7C5D82AEULL, "vxw_woundedcoast_1fd_0_0.vlm"},
	{0x23FC3FAFCD1B577FULL, "61177289.ogg"},
	{0x23FCE9359FBB9B76ULL, "6042981_m.fxe"},
	{0x23FCFE3AFF6E1434ULL, "follower_banter6169472.cl"},
	{0x23FD093FF1749116ULL, "1015532821.ogg"},
	{0x23FD788C17FEC620ULL, "dae000_male_commoner1_mob6174454.cl"},
	{0x23FD7B37E82667DDULL, "6161205_m.fxe"},
	{0x23FDC89185796C72ULL, "6149781_m.fxe"},
	{0x23FDEAF67FC91BDDULL, "vxw_woundedcoast_zr_0_0.vlm"},
	{0x23FE06664FE5D29CULL, "vxw_woundedcoast_1sx_0_1.vlm"},
	{0x23FE484AA45AB06EULL, "vxca_caverns2_95r_0_0.vlm"},
	{0x23FEB94C44FDCF31ULL, "hlac_ht_0_hightown_a2_-196610_0z.dds"},
	{0x23FF163BC2E8C353ULL, "597762450.ogg"},
	{0x23FF74F380140D93ULL, "gen00fl_aveline6115441.cl"},
	{0x23FF928073EBE206ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopen.gad"},
	{0x23FF990514420325ULL, "957001775.ogg"},
	{0x24008047EEB8E70BULL, "44438378.ogg"},
	{0x2400C902B9736339ULL, "6128651_m.fxe"},
	{0x2401D660F67D7BBDULL, "cd_seb201_lady_harimann.bnk"},
	{0x24024A6252F1A2E0ULL, "6034227_m.fxe"},
	{0x24026D2C3B7E1D0DULL, "670725437.ogg"},
	{0x24028271F4B955B6ULL, "gen00fl_merrill6105671.cl"},
	{0x2402A0F050DE7340ULL, "62584438.ogg"},
	{0x240340A7B58B0B4EULL, "6144135_m.fxe"},
	{0x240385ADBB0FCEC1ULL, "dae300_alistair_king6133821.cl"},
	{0x240394FA80ABC4D0ULL, "494352405.ogg"},
	{0x2403AF570C391D3BULL, "209242286.ogg"},
	{0x2403BA87D438158BULL, "6045574_m.fxe"},
	{0x2404A3526D770E8BULL, "vxca_caverns2_9fj_0_0.vlm"},
	{0x2404F2AEA564CE63ULL, "6031996_f.fxe"},
	{0x2405206B88A347E9ULL, "zz_dragon_male.cut"},
	{0x2405665966E4E457ULL, "vxw_woundedcoast_qu_0_0.vlm"},
	{0x24059047EE375E03ULL, "vxht_keep_5ho_0_0.vlm"},
	{0x2405976E01DFA066ULL, "180518786.ogg"},
	{0x24064BD98E6FB649ULL, "llac_ht_0_hightown_a2_-131071_0.dds"},
	{0x2406506F6C475BC5ULL, "290221717.ogg"},
	{0x24067930D472A2D7ULL, "art\\characters\\creatures\\ragedemon\\animation\\die_v1.evt"},
	{0x24071125E1B47C16ULL, "vxht_bartrand_n_5yq_0_0.vlm"},
	{0x240731BA9D50FC14ULL, "26176457.ogg"},
	{0x24074E60F196E56FULL, "hlac_lt_0_lowtown_a2_65536_2y.dds"},
	{0x24076D412F017156ULL, "270023802.ogg"},
	{0x2407C506C623A8A6ULL, "sib131_slave_a6165261.cl"},
	{0x240883AFF59C3397ULL, "6153408_m.fxe"},
	{0x2408C0D91964F6F8ULL, "6054575_m.fxe"},
	{0x24096297DB439B85ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_6x.dds"},
	{0x2409E4552CF1751BULL, "15367104.ogg"},
	{0x2409E8FFF2E8AD03ULL, "6137680_m.fxe"},
	{0x240A64644D325A37ULL, "6166596_m.fxe"},
	{0x240A64E4B208056EULL, "6150075_m.fxe"},
	{0x240A6A876E539B2EULL, "vxw_woundedcoast_1ae_0_0.vlm"},
	{0x240B198F1046DB86ULL, "6070607_m.fxe"},
	{0x240B5568384CEFDCULL, "qcr220_grey_wardens6060401.cl"},
	{0x240BEEFEE02E1156ULL, "vxht_bartrand_n_2k4_0_0.vlm"},
	{0x240C5CD371BF345DULL, "vxw_woundedcoast_wq_0_0.vlm"},
	{0x240D298BC2C2FE15ULL, "gen00fl_varric6102756.cl"},
	{0x240D7A30D478C360ULL, "art\\characters\\creatures\\ragedemon\\animation\\die_v1.gad"},
	{0x240D96E73C54B3E0ULL, "6118983_m.fxe"},
	{0x240DC879509BB077ULL, "lt_0_lowtown_docks_n_130_0.msh"},
	{0x240DDA79509BCEDAULL, "lt_0_lowtown_docks_n_130_0.mao"},
	{0x240DDE79509BD5E9ULL, "lt_0_lowtown_docks_n_130_0.mmh"},
	{0x240DE5A95619677FULL, "676479680.ogg"},
	{0x240E17FDB786BFA1ULL, "llac_ht_0_hightown_a3_-131070_0.dds"},
	{0x240E2E522C96E379ULL, "241956864.ogg"},
	{0x240E578F273ACB6CULL, "amb_scene_dogbark.wbd"},
	{0x240F52187BFB37B4ULL, "vxlt_hangedman_n_109_0_0.vlm"},
	{0x240F7F70E70DD0EDULL, "6076050_f.fxe"},
	{0x241013420E6328CDULL, "oil_loop.dds"},
	{0x2410584B37E57406ULL, "245415753.ogg"},
	{0x2410945451543D45ULL, "l_env_wal_gallows_hvycough.wbd"},
	{0x24110F5DD98BE9F0ULL, "hlac_ht_keep_3736_0x.dds"},
	{0x24111A96BAD7BD45ULL, "gen00fl_anders6116078.cl"},
	{0x241165D6EAD5281FULL, "6093573_m.fxe"},
	{0x2411D3496DEDFBDCULL, "514623894.ogg"},
	{0x2412BEAC8057B28AULL, "6165105_m.fxe"},
	{0x24146DCE26743C27ULL, "momnt_glow.mao"},
	{0x241490A4CF9869BCULL, "127520096.ogg"},
	{0x2414A3953739E024ULL, "6172635_m.fxe"},
	{0x2414D8C9ED1EDD4BULL, "art\\vfx\\combat\\vfx_r_brd_shout_act_c.mmh"},
	{0x2414DA7A905B84A4ULL, "6166813_f.fxe"},
	{0x2414F02DFA179ACAULL, "vxca_cavern1_haunted_ilc_0_0.vlm"},
	{0x2415098AD09E0F12ULL, "gen00fl_fenris6113881.cl"},
	{0x2415E6A8BDD33687ULL, "351668899.ogg"},
	{0x2416496DD8579FCCULL, "lt_0_lowtown_docks_n_223_0.msh"},
	{0x24164F6DD857A9BEULL, "lt_0_lowtown_docks_n_223_0.mmh"},
	{0x24165B6DD857BE55ULL, "lt_0_lowtown_docks_n_223_0.mao"},
	{0x24176287795EF4A9ULL, "hlac_w_blightlands_131075_0y.dds"},
	{0x2417CAEF67D6C722ULL, "w_sundermount_249_0.msh"},
	{0x2417CCEF67D6CAD8ULL, "w_sundermount_249_0.mmh"},
	{0x2417CE92869A571CULL, "6133409_m.fxe"},
	{0x2417D8EF67D6DF7BULL, "w_sundermount_249_0.mao"},
	{0x2418676C7C086F16ULL, "dae000_meredith6080973.cl"},
	{0x2418AE207F2B7AB7ULL, "148515397.ogg"},
	{0x2418D64EAA12F53FULL, "1017106898.ogg"},
	{0x2418D77A95E4E999ULL, "hlac_ht_bartrand_n_3365_0y.dds"},
	{0x2418DC646049B3AEULL, "vxw_woundedcoast_1lf_0_0.vlm"},
	{0x2418FD8602695427ULL, "vxlt_hangedman_62_0_0.vlm"},
	{0x2419A4FC4814F91BULL, "vxht_brothel_n_4vd_0_0.vlm"},
	{0x2419B5CE91E96371ULL, "vxca_caverns1_i1r_0_0.vlm"},
	{0x241A78FEBCDD213EULL, "rock_02.dds"},
	{0x241B018E721A1E16ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb03.mmh"},
	{0x241B097655956ACBULL, "25664415.ogg"},
	{0x241B0B8E721A2F14ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb03.msh"},
	{0x241B74867DC2F32AULL, "dae000_meredith6092585.cl"},
	{0x241B76F16BB39564ULL, "1005535326.ogg"},
	{0x241B900518EA970FULL, "w_woundedcoast_mmap_ph.dds"},
	{0x241BB5BD6863E2BEULL, "vxlt_gallowstemplar_f_5qa_0_0.vlm"},
	{0x241BE05B8DF7D00EULL, "556304623.ogg"},
	{0x241BE0EF9738D82BULL, "6041116_m.fxe"},
	{0x241C34FF494CBAD2ULL, "mag101tr_aom_undead.ncs"},
	{0x241C523054EC059BULL, "w_encounter3_71_0.msh"},
	{0x241C5C3F51D3898DULL, "hlac_lt_0_lowtown_65535_1y.dds"},
	{0x241C643054EC24E6ULL, "w_encounter3_71_0.mao"},
	{0x241C703054EC3885ULL, "w_encounter3_71_0.mmh"},
	{0x241CD1B51CE81C8AULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5.gad"},
	{0x241D1511EC2BF1F6ULL, "vxca_caverns1_j23_0_0.vlm"},
	{0x241D2E6683E72C4AULL, "vxlt_0_lowtown_a3_n_c_e20_0_0.vlm"},
	{0x241D47CEC8E8A756ULL, "vxlt_gallowscourtyard_6qs_0_0.vlm"},
	{0x241D4D74AD5171B3ULL, "zz_approval_debug6003707.cl"},
	{0x241E75B7DD1D141CULL, "vxht_mansion2_3dr_0_0.vlm"},
	{0x241E791EA41B4BBEULL, "6162540_m.fxe"},
	{0x241EFB9488CF0723ULL, "6093964_m.fxe"},
	{0x241F4023F8D1E4F6ULL, "216542992.ogg"},
	{0x241F73ED4531FF97ULL, "20885619.ogg"},
	{0x241FEE69BAA9FC56ULL, "llac_lt_warehouse_n_65536_3.dds"},
	{0x24213766C720036AULL, "6068444_m.fxe"},
	{0x24220322E303CCFAULL, "vxca_caverns2_8ki_0_0.vlm"},
	{0x242241D0177F9D2BULL, "6122783_m.fxe"},
	{0x2422534852EE0C46ULL, "ht_0_hightown_n_21_0.rml"},
	{0x2422761945EA9D7FULL, "vxlt_0_lowtown_a3_n_c_exk_0_0.vlm"},
	{0x2422B448954B902EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3pil.mmh"},
	{0x2422B91D2F9F24F0ULL, "6164838_m.fxe"},
	{0x2422B9C9ED2AE5B0ULL, "art\\vfx\\combat\\vfx_r_brd_shout_act_c.anb"},
	{0x2422C048954BA4C5ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3pil.mao"},
	{0x2422CE48954BBC1CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3pil.msh"},
	{0x24233D625D284277ULL, "305196801.ogg"},
	{0x24237414B967EAABULL, "34053440.ogg"},
	{0x2423A040E2F0D9EBULL, "hlac_lt_0_lowtown_a2_n_c_196607_1y.dds"},
	{0x2423B4B51CEE0A21ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5.evt"},
	{0x24248AD69485E1FEULL, "386865569.ogg"},
	{0x24254DA15FD060EEULL, "570471060.ogg"},
	{0x24257841E1160983ULL, "vxht_0_hightown_c74_0_0.vlm"},
	{0x24257E9A24F244A4ULL, "vxht_bartrand_n_5v2_0_0.vlm"},
	{0x2425EF72B081B159ULL, "lt_warehouse_n_ll.mmh"},
	{0x24263FAD5AD2F359ULL, "340280130.ogg"},
	{0x2426CE25328A93C2ULL, "vxht_0_hightown_a2_a2l_0_0.vlm"},
	{0x2426D8033132A750ULL, "6095330_m.fxe"},
	{0x2426E18AE856F4EBULL, "6118774_f.fxe"},
	{0x2427171F368616CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_reaction.gad"},
	{0x242862BA7E5991D4ULL, "vxlt_0_lowtown_a3_n_c_hsx_0_0.vlm"},
	{0x24286D2E7EBC81D4ULL, "vxht_brothel_n_632_0_0.vlm"},
	{0x242882312EAD4A00ULL, "6104887_m.fxe"},
	{0x24289214E14ABF69ULL, "vxca_caverns2_6he_0_0.vlm"},
	{0x2429AE696246D3B8ULL, "936440705.ogg"},
	{0x2429C9929764D4E4ULL, "nat380_nathaniel6070264.cl"},
	{0x2429CEABEABB5ECAULL, "218979062.ogg"},
	{0x2429E264BB4881BDULL, "vxw_bonepit_lq_0_0.vlm"},
	{0x242A587793298B3BULL, "vxht_0_hightown_7g0_0_0.vlm"},
	{0x242AA814048084BBULL, "6118742_m.fxe"},
	{0x242B2CCAFB420879ULL, "6070796_m.fxe"},
	{0x242BCE68926455D3ULL, "dae200cr_viscount.bnk"},
	{0x242C204C7042809EULL, "vxlt_0_lowtown_h7z_0_0.vlm"},
	{0x242CBFC63F1FCA90ULL, "719870033.ogg"},
	{0x242CC5A79AC3A325ULL, "981049521.ogg"},
	{0x242DB3A2ECF11343ULL, "6152729_f.fxe"},
	{0x242DE197D22B0052ULL, "lt_backalley_n_2_0.anb"},
	{0x242DFA1F368C0461ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_reaction.evt"},
	{0x242E0C06AECAD1EDULL, "vis121_saemus6100085.cl"},
	{0x242E7CE7029A25D3ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_exp_p.anb"},
	{0x242E937B6D4A84CBULL, "lgt000ip_store_epic1_3.utp"},
	{0x242E97FD3FDC55B9ULL, "994826894.ogg"},
	{0x242E9E06FBC234F9ULL, "vxlt_backalley_n_1kh_0_0.vlm"},
	{0x242EF0689266C523ULL, "dae200cr_viscount.cnv"},
	{0x242EFA0652120DAEULL, "463069639.ogg"},
	{0x242F2946AF173528ULL, "1053063198.ogg"},
	{0x242F7C1F5849CC5AULL, "vxht_chantry_n_2a9_0_0.vlm"},
	{0x24301A05844E509BULL, "dae000_sabina6161442.cl"},
	{0x24305AD7715B4B23ULL, "709624508.ogg"},
	{0x24306BAA289E18FAULL, "vxht_0_hightown_baf_0_0.vlm"},
	{0x24309A2C2C7410ABULL, "373543497.ogg"},
	{0x2430ACB51CF88770ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5.ani"},
	{0x243108EBA59908EEULL, "gen00fl_dog6107353.cl"},
	{0x243155D696AC937CULL, "49343292.ogg"},
	{0x24317B4A095C4641ULL, "166407526.ogg"},
	{0x24318C6F04C3828DULL, "vxlt_darktown_44s_0_0.vlm"},
	{0x24319791C9D3F602ULL, "6151465_m.fxe"},
	{0x2431B551D8B57E47ULL, "follower_banter6168342.cl"},
	{0x2431C28A1E0A8F61ULL, "413920062.ogg"},
	{0x2432AF2FEE7364DFULL, "ip_restrict_spec.ncs"},
	{0x2432C49F09D6D5E8ULL, "vxw_encounter2_nj_0_0.vlm"},
	{0x2432FE56A9904F03ULL, "hlac_w_woundedcoast_262146_0y.dds"},
	{0x2433AAF201A7C284ULL, "vxca_cavern1_haunted_fmi_0_0.vlm"},
	{0x2433BDD8519D9A08ULL, "6097156_m.fxe"},
	{0x2433E7D9A62C4E5EULL, "hlac_ht_0_hightown_a2_0_3x.dds"},
	{0x2433F07716E85D24ULL, "vxht_bartrand_n_5y3_0_0.vlm"},
	{0x24340A9C338C24DCULL, "llac_lt_0_lowtown_a2_n_-3_2.dds"},
	{0x2435BE6F46CEC02BULL, "gen00fl_merrill6124022.cl"},
	{0x2435C01E4435E4AEULL, "vxw_woundedcoast_1rx_0_0.vlm"},
	{0x2435C98254F8D2F1ULL, "lt_0_lowtown_docks_a3_123_0.mmh"},
	{0x2435D58254F8E692ULL, "lt_0_lowtown_docks_a3_123_0.mao"},
	{0x2435E38254F8FEAFULL, "lt_0_lowtown_docks_a3_123_0.msh"},
	{0x24361F8D0A88E8D3ULL, "485376882.ogg"},
	{0x2436DE6D2DEF5A3CULL, "zz_followers_banter6159129.cl"},
	{0x243851E72E3FBCE5ULL, "fx_ice_dis.mao"},
	{0x24388BC9DA2EC9E2ULL, "681563665.ogg"},
	{0x243984E21298E377ULL, "gen00fl_varric6076123.cl"},
	{0x24399EA7ED3820F9ULL, "6075496_m.fxe"},
	{0x2439A8DE1E629CBCULL, "ts_profane_arm_table.gda"},
	{0x2439DDE07995A1FAULL, "6083739_m.fxe"},
	{0x2439E77A44DC73E9ULL, "vxht_0_hightown_a70_0_0.vlm"},
	{0x243A07CB556671EAULL, "vxca_caverns2_488_0_0.vlm"},
	{0x243A122067D35D4AULL, "vxlt_0_lowtown_a3_n_c_e82_0_0.vlm"},
	{0x243A44E489D6DE5CULL, "gen_im_cft_rec_wrn_08.uti"},
	{0x243A4C0615C75A94ULL, "dae000_mother6177010.cl"},
	{0x243A7391F5712853ULL, "vxca_cave_18k_0_0.vlm"},
	{0x243AB67BF556AF13ULL, "hlac_lt_gallowstemplar_n_2382_1z.dds"},
	{0x243AC3310EEDDA47ULL, "vxht_mansion2b_3c4_0_0.vlm"},
	{0x243B0556793E90A7ULL, "hlac_lt_0_lowtown_n_-196609_3z.dds"},
	{0x243BAD00F5F855F4ULL, "151035854.ogg"},
	{0x243BB77B9A5421FCULL, "337080517.ogg"},
	{0x243BB8BB8076D69AULL, "87035784.ogg"},
	{0x243BF21F369833B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_reaction.ani"},
	{0x243C0E8255742D1EULL, "lgt000ip_ancient_stone_door.utp"},
	{0x243C8C14625FD67EULL, "vxht_bartrand_n_64q_0_0.vlm"},
	{0x243C99E702A6943CULL, "art\\vfx\\cinematic\\vfx_meredith_swd_exp_p.mmh"},
	{0x243D8D2376DAA46EULL, "crt_corpse.wbd"},
	{0x243DBDBC645DA72BULL, "741935716.ogg"},
	{0x243E05DBFFE9E0E4ULL, "dae000cr_grand_cleric6173677.cl"},
	{0x243E1D374BCD3F41ULL, "gen00fl_aveline6107062.cl"},
	{0x243E617088F3A40BULL, "vxht_0_hightown_aob_0_0.vlm"},
	{0x243E88E9F58243C5ULL, "vxca_caverns2_8lc_0_0.vlm"},
	{0x243E890FD9EBA02EULL, "qcr000_arishok.cnv"},
	{0x243EB5015EFE1089ULL, "vxlt_hangedman_55_0_0.vlm"},
	{0x243EDB5A1BEA75E9ULL, "6044239_f.fxe"},
	{0x243FB2F3E1F22E06ULL, "360523393.ogg"},
	{0x243FC11F862F340BULL, "ht_0_hightownexplosion_n_mmap_ph.dds"},
	{0x244004E9CEF0788BULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_006a_0n.dds"},
	{0x24407C294DFFE837ULL, "hlac_ht_mansion2_818_5x.dds"},
	{0x2440936FC5733BAFULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_noise_d.dds"},
	{0x2440A842EDEA2CAFULL, "408943966.ogg"},
	{0x2441073C85B6E798ULL, "481411209.ogg"},
	{0x244127FF496BC0B3ULL, "mag101tr_aom_undead.utt"},
	{0x2441367BFBCCEE68ULL, "vxw_encounter2_ik_0_0.vlm"},
	{0x24419B1291E25430ULL, "281654127.ogg"},
	{0x2441AB0FD9EE0F9EULL, "qcr000_arishok.bnk"},
	{0x244251BC3DB50D06ULL, "1069064789.ogg"},
	{0x2442F125634A26C5ULL, "mrl000_marethari6069874.cl"},
	{0x2442FCF9936EFA42ULL, "llac_w_woundedcoast_0_1.dds"},
	{0x2443028BC2C376AEULL, "6033707_m.fxe"},
	{0x24438D5A5A83EF65ULL, "61850845.ogg"},
	{0x244392367D55E222ULL, "hlac_w_encounter3_-65532_0y.dds"},
	{0x24439A8E196E11F5ULL, "42341858.ogg"},
	{0x2443C544CC4325EBULL, "6112574_f.fxe"},
	{0x2443E8070A49652EULL, "hlac_ht_0_hightown_a2_65534_0z.dds"},
	{0x2443F447B6C5C6D2ULL, "vxca_cave_18n_0_0.vlm"},
	{0x244512E55996A560ULL, "stone_star.mao"},
	{0x2445EE4FEA4752F3ULL, "hlac_w_encounter2_-131067_0x.dds"},
	{0x244600DD6BD125B6ULL, "hlac_lt_0_lowtown_docks_n_-65537_5y.dds"},
	{0x24463BB7E0943A31ULL, "vxlt_warehouse_n_122_0_0.vlm"},
	{0x24466DFD3822D58CULL, "hlac_ht_0_hightown_-131073_1z.dds"},
	{0x24467176DE356CC3ULL, "hlac_lt_0_lowtown_a3_n_c_0_9y.dds"},
	{0x2446EF1CEC31A572ULL, "fen310st_vanrania.stg"},
	{0x2447338077CFC38BULL, "6140512_m.fxe"},
	{0x244756AD662E3CB0ULL, "233930505.ogg"},
	{0x24487B4753A59F89ULL, "hlac_ht_chantry_n_2534_36z.dds"},
	{0x2448EC42A5575924ULL, "art\\vfx\\cinematic\\vfx_lightbeam01_p.anb"},
	{0x2448ED04C4E8558BULL, "art\\levels\\races\\elf\\props\\_textures\\prp_dream_dp.dds"},
	{0x244912A0ADF23A52ULL, "vxca_caverns2prc_62e_0_0.vlm"},
	{0x2449350E00E3F2C7ULL, "vxht_mansion2b_3f7_0_0.vlm"},
	{0x244941730E399BDDULL, "vxw_bonepit_1ow_0_0.vlm"},
	{0x244990027AB99DC3ULL, "zz_ave_debug6174808.cl"},
	{0x244A44390CD1170DULL, "165825271.ogg"},
	{0x244A4552EDCD33E8ULL, "llac_ht_0_hightown_a3_-1_0.dds"},
	{0x244A7031EE539CD8ULL, "81136122.ogg"},
	{0x244A8B36CEE9918CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pry_open.ani"},
	{0x244B19DE030B89D6ULL, "6149437_m.fxe"},
	{0x244B519F435FA9FFULL, "vxlt_0_lowtown_gqv_0_0.vlm"},
	{0x244C5CB02C9397DEULL, "6111673_f.fxe"},
	{0x244C6B574A332705ULL, "6077949_f.fxe"},
	{0x244C7EB543AA4843ULL, "6127660_f.fxe"},
	{0x244C825279278A2EULL, "vxw_bonepitmining_a2_1wg_0_1.vlm"},
	{0x244CD6F3992DE34EULL, "lt_0_lowtown_docks_n_water_16559.msh"},
	{0x244CE4F3992DFA9FULL, "lt_0_lowtown_docks_n_water_16559.mao"},
	{0x244CF0F3992E0F3CULL, "lt_0_lowtown_docks_n_water_16559.mmh"},
	{0x244CFAC6E7785C45ULL, "vxca_caverns2_8kb_0_0.vlm"},
	{0x244D017F9555EDA7ULL, "dae000cr_grand_cleric6124698.cl"},
	{0x244EAE32585AD6C7ULL, "74559553.ogg"},
	{0x244EB5DC3DCA66DBULL, "lt_0_lowtown_a3_n_c_water_21475.mmh"},
	{0x244EB8009E5C271CULL, "268958778.ogg"},
	{0x244EB86D7927B5E1ULL, "tal_thre_ske.mao"},
	{0x244EC1DC3DCA7B78ULL, "lt_0_lowtown_a3_n_c_water_21475.mao"},
	{0x244EC929F102FD6FULL, "6201650_f.fxe"},
	{0x244ED3DC3DCA99C5ULL, "lt_0_lowtown_a3_n_c_water_21475.msh"},
	{0x244F01506CAED36EULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_3z.dds"},
	{0x244F5439AABA6AA3ULL, "863129479.ogg"},
	{0x244F826C7FC672A8ULL, "38276992.ogg"},
	{0x244FB397E6919376ULL, "character_character_p67.psh.fxo.dx11"},
	{0x24505E303867C322ULL, "mag311_emile_de_launcet.cnv"},
	{0x24506D7B2CDE6638ULL, "hlac_ht_chantry_2534_49x.dds"},
	{0x2451216E81C56DB6ULL, "llac_lt_darktown_-1_22.dds"},
	{0x24515F9825E40C43ULL, "hlac_lt_0_lowtown_a2_n_c_-3_4y.dds"},
	{0x24516AE4B0CE3015ULL, "vxlt_0_lowtown_docks_a3_c2k_0_0.vlm"},
	{0x24516E54518B93EFULL, "l_env_wal_gallows_hvycough.bnk"},
	{0x24519836CEEFC6CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pry_open.gad"},
	{0x2451B4BE7176C935ULL, "follower_banter6169772.cl"},
	{0x2452083010E2646BULL, "450028947.ogg"},
	{0x24522B20917F0BD3ULL, "llac_dr_primevaltomb_664_3.dds"},
	{0x24528A824E9927EEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12b.msh"},
	{0x2452A4824E9953DCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12b.mmh"},
	{0x2452D2C387664973ULL, "6083499_m.fxe"},
	{0x2453382D27A92A5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_dw.anb"},
	{0x245340303869C65AULL, "mag311_emile_de_launcet.bnk"},
	{0x2453540C29DADED1ULL, "462230093.ogg"},
	{0x24537441ED162300ULL, "vxw_encounter2_camp_pt_0_0.vlm"},
	{0x2453C0C9ED3F9801ULL, "hlac_lt_gallowstemplar_n_2391_5x.dds"},
	{0x2454624A17BA367BULL, "172660210.ogg"},
	{0x24549FB7F08A6740ULL, "vxlt_0_lowtown_a2_n_c_exy_0_0.vlm"},
	{0x2454A98B757BB020ULL, "299459859.ogg"},
	{0x2454D20B6EAA1FF4ULL, "docks_male_overseer_mob6176548.cl"},
	{0x2454E57CEE75599FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillara.mmh"},
	{0x2454EB7CEE756441ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillara.msh"},
	{0x24559E306B323EDCULL, "6092304_m.fxe"},
	{0x2455D65E36E4A593ULL, "870497099.ogg"},
	{0x2455F57661A6BBC7ULL, "6125910_f.fxe"},
	{0x245614F2BE6456CFULL, "vxlt_0_lowtown_a3_n_c_gcd_0_0.vlm"},
	{0x2456764874D9E36BULL, "vxlt_0_lowtown_a2_h24_0_0.vlm"},
	{0x2456DA5A8C292871ULL, "w_bonepit_180_0.mmh"},
	{0x2456E65A8C293C12ULL, "w_bonepit_180_0.mao"},
	{0x2456F45A8C29542FULL, "w_bonepit_180_0.msh"},
	{0x24570B42A563CB27ULL, "art\\vfx\\cinematic\\vfx_lightbeam01_p.mmh"},
	{0x245744CAAAC871DCULL, "fx_smoke_cig.mao"},
	{0x24589336CEF5DCADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pry_open.evt"},
	{0x2458AEAC0422D7A5ULL, "219969357.ogg"},
	{0x2458D5836BF929F6ULL, "vxlt_hangedman_2ml_0_0.vlm"},
	{0x2458F48528D9069FULL, "6208607_m.fxe"},
	{0x24598BBAF501A256ULL, "6142537_m.fxe"},
	{0x2459E4433FC9296FULL, "6098819_m.fxe"},
	{0x245A642BA305EE62ULL, "hlac_ht_chantry_2534_21x.dds"},
	{0x245A80C7596F105AULL, "gen00fl_aveline6132867.cl"},
	{0x245B634E2D12A845ULL, "vxht_0_hightown_bam_0_0.vlm"},
	{0x245BEE31149BA552ULL, "hlac_lt_0_lowtown_a3_n_c_-1_3z.dds"},
	{0x245C3EE2B11B83A4ULL, "547101309.ogg"},
	{0x245C486BA3F4DD50ULL, "vxlt_darktown_2v6_0_0.vlm"},
	{0x245C600D56C9F564ULL, "vxca_caverns1_j1z_0_2.vlm"},
	{0x245C849243509400ULL, "6142997_m.fxe"},
	{0x245CBF19D70B5FA3ULL, "hlac_lt_0_lowtown_docks_-4_2x.dds"},
	{0x245D1AF65AA9D0A3ULL, "hlac_dr_primevaltomb_663_0x.dds"},
	{0x245D3B17C4D3123EULL, "6142220_f.fxe"},
	{0x245DAB4400C7D20FULL, "502414427.ogg"},
	{0x245DBC959805EDBDULL, "vxca_caverns2prc_9f9_0_0.vlm"},
	{0x245DC2A999FB25E8ULL, "637711946.ogg"},
	{0x245EA4601F67885DULL, "hlac_lt_darktown_-1_29x.dds"},
	{0x245F403301591DB6ULL, "gen00fl_varric6201274.cl"},
	{0x245F460DECCAA6A4ULL, "vxht_bartrand_n_5j6_0_0.vlm"},
	{0x245F9D110A94E8E3ULL, "vxht_mansion_n_42o_0_0.vlm"},
	{0x245FA7DD98C3313AULL, "vxca_caverns1_i7o_0_0.vlm"},
	{0x245FBB8463391F91ULL, "vfx_chunkyparticle_p0.psh.fxo.dx11"},
	{0x245FC340027D4331ULL, "6165762_m.fxe"},
	{0x2460158FF90E74A1ULL, "692191630.ogg"},
	{0x24602CD56702B67CULL, "634538876.ogg"},
	{0x24603573A099C73BULL, "lt_0_lowtown_docks_a3_n_140_0.mmh"},
	{0x246037AF55F8847BULL, "dae000_keep_servant_1_sta6156891.cl"},
	{0x24604173A099DBD8ULL, "lt_0_lowtown_docks_a3_n_140_0.mao"},
	{0x24605373A099FA25ULL, "lt_0_lowtown_docks_a3_n_140_0.msh"},
	{0x2460A346843D8C96ULL, "vxlt_gallowscourtyard_3x5_0_0.vlm"},
	{0x2460AE960C5C37ACULL, "dae000_quintus6154699.cl"},
	{0x2460BE329AA49DE7ULL, "6050117_f.fxe"},
	{0x2460E497D2565E8EULL, "lt_backalley_n_2_0.rml"},
	{0x2460E4FD014F7093ULL, "vxdr_primevaltomb_7qc_0_0.vlm"},
	{0x246158B30C4AEB0DULL, "plt_cod_cha_merrill3.plo"},
	{0x24616C8ADC2AF49BULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_abma_0t.dds"},
	{0x2461A8CA27BE614DULL, "122409797.ogg"},
	{0x2461E61EDF0C3B80ULL, "vxw_encounter2_camp_ku_0_0.vlm"},
	{0x24621D13816519AEULL, "6133740_m.fxe"},
	{0x24622969525BAE06ULL, "dae100_bran6118304.cl"},
	{0x24622DFD41B9CF41ULL, "728474919.ogg"},
	{0x24624DB8A6E7FC49ULL, "vxlt_undercitydungeon_1ib_0_0.vlm"},
	{0x24626C4A6CE1EC63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v1.ani"},
	{0x2462B86382286550ULL, "6085906_m.fxe"},
	{0x2462C4A44FE001DEULL, "vxht_mansion2_1ko_0_0.vlm"},
	{0x24632483DCDC1066ULL, "150443795.ogg"},
	{0x24635C441DF002F3ULL, "dae200_saemus6142368.cl"},
	{0x246387998CF09A32ULL, "follower_banter6141784.cl"},
	{0x2463F3BFAA99460CULL, "vxlt_0_lowtown_eod_0_1.vlm"},
	{0x24646CC3ED6D641BULL, "hlac_ht_bartrand_n_3370_0y.dds"},
	{0x24648BBBA49A0ED5ULL, "233057849.ogg"},
	{0x2464F6BC9898CF80ULL, "337838231.ogg"},
	{0x2465B4BACA6423CEULL, "6176549_m.fxe"},
	{0x2465E6B93110B1FDULL, "vxlt_0_lowtown_docks_a3_n_aom_0_0.vlm"},
	{0x2466141AE7C0DE87ULL, "hlac_ht_0_hightown_a2_65536_1z.dds"},
	{0x2466B0D6A702D266ULL, "hlac_lt_0_lowtown_docks_-3_1y.dds"},
	{0x24671156E5C9B201ULL, "w_sundermount_78_0.rml"},
	{0x24672BA98DA79EA7ULL, "551177540.ogg"},
	{0x246746F3C4782235ULL, "hlac_ca_caverns2prc_5802_29z.dds"},
	{0x24676FB2CC965A57ULL, "41191863.ogg"},
	{0x24678B6649999850ULL, "595087550.ogg"},
	{0x2467B73BE2080FB4ULL, "6036557_m.fxe"},
	{0x2467D6C0FB42BD0CULL, "vxht_mansion2_2fe_0_0.vlm"},
	{0x24681657C2E9499EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\dazed.ani"},
	{0x246819D2C058B7FDULL, "zz_followers_banter6159690.cl"},
	{0x24683645010D7C7FULL, "mrl150_hunter6071296.cl"},
	{0x24688247936DD346ULL, "hlac_ht_0_hightown_-1_1x.dds"},
	{0x2468A0F72296F9DDULL, "561248513.ogg"},
	{0x2468F3F75434DFA3ULL, "w_bonepit_28_0_ll.mmh"},
	{0x2468F7D29856F2C9ULL, "art\\characters\\weapons\\axes\\w_axe_105a.mao"},
	{0x24695A2EC114D009ULL, "606267156.ogg"},
	{0x2469832376FFCD9CULL, "crt_corpse.bnk"},
	{0x2469EC12F6E31C34ULL, "zz_lgt_1_debug6188681.cl"},
	{0x2469F345F78CF18EULL, "352581732.ogg"},
	{0x246A4436967F5EB1ULL, "hlac_ht_mansion_n_2995_2z.dds"},
	{0x246A472A05FA036BULL, "6016336_m.fxe"},
	{0x246A9CFD243784CDULL, "6121216_m.fxe"},
	{0x246ABA4895EAF5D0ULL, "vxlt_darktown_2y7_0_0.vlm"},
	{0x246B386759B98743ULL, "149035766.ogg"},
	{0x246B612E5F3DE46FULL, "mag101_alain6053000.cl"},
	{0x246B79F8690982FDULL, "2288474.ogg"},
	{0x246B9546E2309262ULL, "6034702_m.fxe"},
	{0x246B97F6E8A74B56ULL, "hlac_lt_0_lowtown_a2_n_c_65536_4y.dds"},
	{0x246BA0427CE7BF01ULL, "art\\characters\\weapons\\promo\\shields\\textures\\w_shd_011a_0n.dds"},
	{0x246BD356E192CBBEULL, "dae000_arishok6072777.cl"},
	{0x246BEA68929A6922ULL, "dae200cr_viscount.utc"},
	{0x246C13AABCB413ABULL, "rdr111dg_martin6055142.cl"},
	{0x246C42193080A283ULL, "w_bonepitdestroyed_a3_229_0.msh"},
	{0x246C50193080BADDULL, "w_bonepitdestroyed_a3_229_0.mmh"},
	{0x246C54193080C1EEULL, "w_bonepitdestroyed_a3_229_0.mao"},
	{0x246D09BEFBAA309EULL, "gen00fl_anders6111283.cl"},
	{0x246D95C4EE6C7D6EULL, "vxlt_0_lowtown_fg8_0_0.vlm"},
	{0x246D9CEBE1BCE50FULL, "vxht_chantry_n_337_0_0.vlm"},
	{0x246E897CCA8DC6B2ULL, "6118314_m.fxe"},
	{0x246E959FF89B5ABBULL, "w_encounter3_27_0.rml"},
	{0x246EEA14E121D30EULL, "vxw_encounter2_mi_0_0.vlm"},
	{0x246EEE991EC89591ULL, "6098249_f.fxe"},
	{0x246F513DB035D962ULL, "6167873_m.fxe"},
	{0x246F7184B7C51D18ULL, "525352556.ogg"},
	{0x246FBCAE2E85C4CBULL, "dae000_hate_enemies6167599.cl"},
	{0x246FC6EBD9939CC8ULL, "vxlt_gallowstemplar_2rv_0_0.vlm"},
	{0x24702FE74009D5A3ULL, "war_secondwind.dds"},
	{0x24705250FC933047ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avel_1.mao"},
	{0x247057FBD1025400ULL, "vxw_encounter2_camp_fr_0_0.vlm"},
	{0x2470D0DED6AD157BULL, "6105894_m.fxe"},
	{0x24712EEEF5F8F76EULL, "gen00fl_varric6076061.cl"},
	{0x2471579553C3EC33ULL, "vxlt_0_lowtown_gij_0_0.vlm"},
	{0x24718371D4769840ULL, "vxlt_0_lowtown_a2_n_c_erg_0_0.vlm"},
	{0x2471A03FFCEA84E2ULL, "vxlt_hangedman_n_t8_0_0.vlm"},
	{0x24723EDF9F073DF5ULL, "ch1_dy.psh.fxo"},
	{0x2472A3CB092C5B7FULL, "vxlt_0_lowtown_docks_a3_n_d0f_0_1.vlm"},
	{0x2472DFFF05F5985EULL, "6069920_m.fxe"},
	{0x24730893E1329091ULL, "288440928.ogg"},
	{0x24734C0215523E86ULL, "vxht_bartrand_n_68e_0_0.vlm"},
	{0x247371F8AFC3B866ULL, "243839036.ogg"},
	{0x2473C983E5D10350ULL, "follower_banter6163254.cl"},
	{0x247404223D3238C3ULL, "one300_zevran6076554.cl"},
	{0x247407A4931CB834ULL, "946323719.ogg"},
	{0x24741E9B514A1563ULL, "486756906.ogg"},
	{0x24743B4539074B08ULL, "vxlt_gallowstemplar_f_3kk_0_0.vlm"},
	{0x2474930FDA1940A1ULL, "qcr000_arishok.stg"},
	{0x2474D69BE8F43CABULL, "hm_ser_witness.mor"},
	{0x24758B8DBF353791ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_in_c.anb"},
	{0x24759BF138B71E10ULL, "vxlt_gallowstemplar_3st_0_0.vlm"},
	{0x2475EBDCE092FF7FULL, "27769592.ogg"},
	{0x24762E57C2F5B06FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\dazed.evt"},
	{0x24765B7AAC25FB84ULL, "vxlt_0_lowtown_n_byg_0_0.vlm"},
	{0x24765DCAAAE34CCEULL, "fx_smoke_cig.dds"},
	{0x2476834A6CF2BC4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v1.gad"},
	{0x2476F2D5989A0A31ULL, "256758091.ogg"},
	{0x247750E00C47F6CAULL, "hlac_ht_playershouse_n_855_18z.dds"},
	{0x247783824EB837A7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12b.phy"},
	{0x2478470810FED945ULL, "vxht_0_hightown_bwk_0_0.vlm"},
	{0x24785ABF5C450FA7ULL, "hlac_lt_0_lowtown_docks_a3_n_131072_1z.dds"},
	{0x247877250A03D648ULL, "vxht_0_hightown_bg2_0_0.vlm"},
	{0x2478B67CE11ADAEAULL, "store_dae210_armor.utm"},
	{0x2478C081B7FAA60AULL, "hlac_lt_hangedman_-1_24y.dds"},
	{0x24795528E8AAA114ULL, "vfx_blt_fire_wall_p_dur_placed.bnk"},
	{0x247A0158A29319A7ULL, "930742691.ogg"},
	{0x247A0AC9BC91080CULL, "vxht_keep_57y_0_0.vlm"},
	{0x247A636133DB8A61ULL, "art\\vfx\\materials\\vfx_smoke_thick.mao"},
	{0x247A65A667DC3A84ULL, "vxlt_0_lowtown_a2_n_c_e1v_0_0.vlm"},
	{0x247AA04F7BF21EBDULL, "vxca_caverns2prc_9h7_0_0.vlm"},
	{0x247AB8CA18790310ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_3y.dds"},
	{0x247B00C8C165FFC2ULL, "vxlt_undercityencounter_xl_0_0.vlm"},
	{0x247B41A8C0485689ULL, "mag320_noblewoman_16060828.cl"},
	{0x247C19196435BBCAULL, "qcr000_arishok6059323.cl"},
	{0x247C2F57C2FA1D18ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\dazed.gad"},
	{0x247C352A2E5546F9ULL, "lowtown_male_drunk_mob6151163.cl"},
	{0x247C415DD16F0CCCULL, "vxlt_undercitydungeon_19k_0_0.vlm"},
	{0x247CCDA706192B29ULL, "vxdr_0_deeproads_bhh_0_1.vlm"},
	{0x247CD87DF91DF755ULL, "6043174_m.fxe"},
	{0x247CEB180B837964ULL, "docks_male_laborer_mob6175398.cl"},
	{0x247D0BCE2E691D8FULL, "6125360_m.fxe"},
	{0x247D2649BDB0F4F7ULL, "6060481_m.fxe"},
	{0x247D5A6EC55126EAULL, "6208335_f.fxe"},
	{0x247D5BF1D317EB8EULL, "vxw_encounter2_hj_0_0.vlm"},
	{0x247D7A57CAFAFBCAULL, "hlac_lt_0_lowtown_docks_-65539_3z.dds"},
	{0x247E082D5CDED79BULL, "6182828_m.fxe"},
	{0x247E38C8CB89B548ULL, "vxlt_gallowstemplar_2mw_0_0.vlm"},
	{0x247E51FCB6C6928BULL, "6179466_m.fxe"},
	{0x247E79A12312AF98ULL, "dae000_gamlen6121557.cl"},
	{0x247E9A198372195BULL, "zz_ave_debug6064622.cl"},
	{0x247E9BC66C61716EULL, "vxlt_gallowstemplar_n_3rz_0_0.vlm"},
	{0x247EC9D8C2F86C80ULL, "vxw_encounter2_camp_as_0_0.vlm"},
	{0x247EFEE63680971CULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gax_002a_s.dds"},
	{0x247F00CD65514AAEULL, "6139140_m.fxe"},
	{0x247F0A1574E1715CULL, "1008677715.ogg"},
	{0x247F1FC6CA6C5393ULL, "859975639.ogg"},
	{0x247F2B2F85954532ULL, "787487245.ogg"},
	{0x247F3F42D25E7C0EULL, "6118611_m.fxe"},
	{0x247F58A0C6AE9301ULL, "vxlt_hangedman_18s_0_0.vlm"},
	{0x247FAF1705EC8FAAULL, "gen00fl_varric6118992.cl"},
	{0x247FF40FB6F91A59ULL, "6015087_m.fxe"},
	{0x2480121CEEE09D62ULL, "vxlt_hangedman_n_s9_0_0.vlm"},
	{0x2480170D30CF7AA5ULL, "vxlt_hangedman_2v6_0_0.vlm"},
	{0x24808CB2CCD4A26DULL, "vxw_encounter1_b6_0_0.vlm"},
	{0x248183B732170B25ULL, "vxdr_primevalentrance_23n_0_0.vlm"},
	{0x24818B309E672DF5ULL, "llac_ca_caverns1_mines_24424_25.dds"},
	{0x2481AD92BB77FFAEULL, "zz_vrc_debug6200235.cl"},
	{0x2481EC0014D0A5EFULL, "vxlt_hangedman_n_2mr_0_0.vlm"},
	{0x24820691D8EFD18FULL, "gen00fl_aveline6104960.cl"},
	{0x24826BEC3424F1F1ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_102a_n.dds"},
	{0x248276616FBC9969ULL, "682629576.ogg"},
	{0x2482AC8DBF3FFA0AULL, "art\\vfx\\combat\\vfx_c_rw_teleport_in_c.mmh"},
	{0x2482D00A4EBD2B20ULL, "llac_ht_0_hightown_n_-65536_2.dds"},
	{0x2482FAA7736613CFULL, "510985449.ogg"},
	{0x248347C92F22A2CBULL, "140678021.ogg"},
	{0x2483F9F6957C7DE4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_lgta_0s.dds"},
	{0x2484186712E85FBBULL, "975523095.ogg"},
	{0x2484D68320B44AABULL, "6115349_m.fxe"},
	{0x2485132654803433ULL, "1000428541.ogg"},
	{0x2485256B7127F6A0ULL, "vxlt_0_lowtown_ha2_0_0.vlm"},
	{0x2485324C5CABDCA8ULL, "follower_banter6155835.cl"},
	{0x24858CBBF2535273ULL, "vxw_sundermount_2wa_0_0.vlm"},
	{0x2486010A6720195FULL, "516934744.ogg"},
	{0x2486976BFA96EDEDULL, "6076627_f.fxe"},
	{0x2486F257A855BB48ULL, "vxca_caverns2_8w7_0_0.vlm"},
	{0x2486F3449B913DA8ULL, "qun221cr_petrice_mob3.utc"},
	{0x2487188AEDD0D034ULL, "716819930.ogg"},
	{0x24874F68DBE9EEACULL, "llac_lt_0_lowtown_a2_n_131070_0.dds"},
	{0x2487C782A6A7E380ULL, "vxca_caverns2prc_8sj_0_0.vlm"},
	{0x2487F63665928E96ULL, "26815510.ogg"},
	{0x2487FEB6342055C1ULL, "vxlt_0_lowtown_a2_n_e24_0_0.vlm"},
	{0x248820541B4DF64BULL, "6095135_m.fxe"},
	{0x24883B798A4012B0ULL, "6075612_f.fxe"},
	{0x248927729D6FF9EAULL, "750025960.ogg"},
	{0x24897A7AE2C7C9CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardright.ani"},
	{0x2489D6A793E0CF8AULL, "6165314_m.fxe"},
	{0x2489F11C96BFB779ULL, "602638941.ogg"},
	{0x248A06ED78431B91ULL, "88381500.ogg"},
	{0x248A80627CF09BA3ULL, "llac_lt_0_lowtown_a2_65538_3.dds"},
	{0x248BDB7ED8F727E1ULL, "vxca_caverns2_8r1_0_0.vlm"},
	{0x248C1E97AA9CED74ULL, "782401572.ogg"},
	{0x248C2BD5666EECD0ULL, "6042049_f.fxe"},
	{0x248C4438CD6EDFF1ULL, "llac_dr_0_deeproads_5842_4.dds"},
	{0x248C9A77747FFB7FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier3_epilogue.gad"},
	{0x248D06A856C04576ULL, "791333075.ogg"},
	{0x248D685E183E3960ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_pot_d.dds"},
	{0x248DAD631D4EAD98ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmb.phy"},
	{0x248DC94720E85729ULL, "dae150_harshal6148824.cl"},
	{0x248E30A96EA6B2AAULL, "llac_lt_gallowscourtyard_n_-1_1.dds"},
	{0x248E4BEED70A3777ULL, "zz_isapt_debug.ncs"},
	{0x248E86D8C98ED43BULL, "dae000ip_crowd_mf1_drunk2.utp"},
	{0x248E8CFC566D9683ULL, "vxht_brothel_n_5gb_0_0.vlm"},
	{0x248EA4E8F5A1396CULL, "6115441_m.fxe"},
	{0x248EFE8FBECABAEDULL, "vxw_encounter1_g5_0_0.vlm"},
	{0x248F967172D36793ULL, "408642490.ogg"},
	{0x248FA478C6A759E5ULL, "410167222.ogg"},
	{0x248FF594240D23A5ULL, "vxdr_primevalentrance_26m_0_0.vlm"},
	{0x24903236C624C253ULL, "391123499.ogg"},
	{0x24906A6AD9593CF8ULL, "hlac_ht_keep_n_3739_1y.dds"},
	{0x2490AE28A4CE8009ULL, "vxht_0_hightown_n_cac_0_0.vlm"},
	{0x2490CB4ED6DA4971ULL, "128232838.ogg"},
	{0x24915576348DB9C4ULL, "prp_str_branch_d.dds"},
	{0x24918409EC239A7BULL, "6111727_m.fxe"},
	{0x2491CB87130B3226ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackright.ani"},
	{0x24920A2AF7AAEF26ULL, "562915198.ogg"},
	{0x2492928A1B937088ULL, "mer221_sabin6045323.cl"},
	{0x249341A2DDB8920EULL, "llac_lt_0_lowtown_docks_a3_-3_3.dds"},
	{0x2493484B867B4E1FULL, "genip_pp_drk_standing.utp"},
	{0x2493A9A4F0BD9201ULL, "6153730_m.fxe"},
	{0x2494555BDC597581ULL, "gen00fl_aveline6139917.cl"},
	{0x249474AC0758ABA2ULL, "347280826.ogg"},
	{0x2494C4E65CA86F1CULL, "vxca_caverns1_ih4_0_0.vlm"},
	{0x24951C9DD2D91968ULL, "vxht_brothel_n_6ex_0_0.vlm"},
	{0x24959708A6C49638ULL, "one100dg_lia6049228.cl"},
	{0x2495AF95EE112B73ULL, "6048191_m.fxe"},
	{0x2495BAACAFF5353BULL, "6101951_m.fxe"},
	{0x2495C6CB3C260A9CULL, "vxht_keep_6bi_0_0.vlm"},
	{0x2495F65A04C0F922ULL, "705343457.ogg"},
	{0x24961A78A67C8F0FULL, "309168220.ogg"},
	{0x249718061BD1FEF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_loop.gad"},
	{0x249747004C68A9DAULL, "hlac_lt_0_lowtown_a2_n_c_65535_7y.dds"},
	{0x2497627AE2D3DE42ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardright.evt"},
	{0x249792FB2DD01A99ULL, "vxdr_primevaltomb_50d_0_0.vlm"},
	{0x24980A3B6AB2A9E8ULL, "6189512_m.fxe"},
	{0x24980DAD936DE97FULL, "624871461.ogg"},
	{0x24983505D1D5B5E8ULL, "vxw_blightlands_7n_0_0.vlm"},
	{0x24983AEA3C644A80ULL, "6135652_m.fxe"},
	{0x24983C9C7110228CULL, "vxht_0_hightown_a3_ak8_0_0.vlm"},
	{0x249845024B215473ULL, "vxht_chantry_1yh_0_0.vlm"},
	{0x24984C1A538F06F1ULL, "qcr000_arishok6060155.cl"},
	{0x249890D8C70C7EFBULL, "mag321_comtesse_de_launcet6042805.cl"},
	{0x24994A9FBF7CD44FULL, "dae010_lowlife_f1.utc"},
	{0x2499F056E5F4D391ULL, "w_sundermount_78_0.anb"},
	{0x2499F36FDC19AAAEULL, "hlac_ca_caverns2_5802_31y.dds"},
	{0x249A0D7F0C8A6C14ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_mad_man.ani"},
	{0x249A76F0893F7C01ULL, "vxca_caverns2prc_7z7_0_0.vlm"},
	{0x249AC46CF9BF6772ULL, "dae231ip_at_to_darktown2.utp"},
	{0x249C99CA5BC7F3AFULL, "and100_anders6096002.cl"},
	{0x249C9D4EB4C646FBULL, "6148215_m.fxe"},
	{0x249CFED94863AF03ULL, "vxht_brothel_n_5je_0_0.vlm"},
	{0x249D20CFC8AC82B1ULL, "vxlt_backalley_n_1er_0_0.vlm"},
	{0x249D29061BD68747ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_loop.ani"},
	{0x249D617AE2D84855ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardright.gad"},
	{0x249DD3958E0437A8ULL, "zz_vault_debug6203277.cl"},
	{0x249E70FC48230E23ULL, "vxlt_0_lowtown_docks_n_azs_0_0.vlm"},
	{0x249E94AA5FD1686AULL, "431353454.ogg"},
	{0x249EB38713159397ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackright.evt"},
	{0x249EC02724E2D5E1ULL, "zz_qun_debug6028222.cl"},
	{0x249ECB832EE05779ULL, "6175543_m.fxe"},
	{0x249EEF53C066608AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_8.pwk"},
	{0x249F097BFA0B641BULL, "art\\characters\\weapons\\shields\\w_shd_twrc.msh"},
	{0x249F1B7BFA0B8366ULL, "art\\characters\\weapons\\shields\\w_shd_twrc.mao"},
	{0x249F277BFA0B9705ULL, "art\\characters\\weapons\\shields\\w_shd_twrc.mmh"},
	{0x249F567166C3E0BEULL, "6109778_m.fxe"},
	{0x249F5E783321B443ULL, "one130_thug6133094.cl"},
	{0x249FD1B8F88E0AE9ULL, "vxlt_undercityencounter_za_0_0.vlm"},
	{0x24A007FA96FA8F21ULL, "308025043.ogg"},
	{0x24A0147B70C0699BULL, "mag110_ws_banter6190053.cl"},
	{0x24A034D54E979CE4ULL, "6027443_m.fxe"},
	{0x24A038713EF0F873ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxb_0.pwk"},
	{0x24A045898F6A107FULL, "vfx_distortionmask_p5.vsh.fxo.dx11"},
	{0x24A07777749068C5ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier3_epilogue.ani"},
	{0x24A10A7F0C908676ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_mad_man.gad"},
	{0x24A1576F0FBA3B34ULL, "545206343.ogg"},
	{0x24A1E59A388AA50BULL, "gen00fl_merrill6127197.cl"},
	{0x24A2795FDA315932ULL, "vxlt_gallowstemplar_527_0_0.vlm"},
	{0x24A3EE9BA16A6AB9ULL, "one280_iwan6096891.cl"},
	{0x24A3F98602D6AECAULL, "vxw_bonepitdestroyed_a3_1rs_0_0.vlm"},
	{0x24A43E24E264A721ULL, "dae000_merchant_access_hf6162973.cl"},
	{0x24A4BC262C06267AULL, "llac_w_woundedcoast_327687_0.dds"},
	{0x24A51ACF78C8229BULL, "847771377.ogg"},
	{0x24A52683AFA3BEE6ULL, "793960163.ogg"},
	{0x24A53A8A9C0E66A5ULL, "vxht_0_hightown_a2_by9_0_0.vlm"},
	{0x24A55646D18A760BULL, "6096408_m.fxe"},
	{0x24A58BC647E956ADULL, "gen00fl_anders6107696.cl"},
	{0x24A5A40B7C0B27C9ULL, "t3_grl_g01.tnt"},
	{0x24A5AB3A6F0F3617ULL, "dre000_cut_sibling_dies6036335.cl"},
	{0x24A5B236C1DD7DB6ULL, "6088770_m.fxe"},
	{0x24A5B487131BB420ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackright.gad"},
	{0x24A5E6F6897C4853ULL, "and111_lirene6084533.cl"},
	{0x24A630DBDF523A82ULL, "combat_deathblow_07.gda"},
	{0x24A6646A176D2F91ULL, "vxw_blightlands_b4_0_0.vlm"},
	{0x24A6A65DF27C135DULL, "1031442284.ogg"},
	{0x24A6A6E2C3CBCE68ULL, "vxw_blightlands_8o_0_0.vlm"},
	{0x24A6AE7963063B0CULL, "vxht_0_hightown_a3_ap9_0_0.vlm"},
	{0x24A6E9E0BA667D0FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_med.mmh"},
	{0x24A70356E5FF7EF0ULL, "w_sundermount_78_0.msh"},
	{0x24A71156E5FF960AULL, "w_sundermount_78_0.mmh"},
	{0x24A7119E906FFE49ULL, "art\\vfx\\textures\\vfx_blood_splatter01.dds"},
	{0x24A71556E5FF9D41ULL, "w_sundermount_78_0.mao"},
	{0x24A76803A119F757ULL, "vxw_bonepit_10t_0_0.vlm"},
	{0x24A7BE0F1AC69332ULL, "477061863.ogg"},
	{0x24A8257F0C96D2D5ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_mad_man.evt"},
	{0x24A9098C0CC3E242ULL, "fen110_fenris6100233.cl"},
	{0x24A912BAA2FCE8DDULL, "mag370_meredith6037407.cl"},
	{0x24AA02EB16DE2D9AULL, "6101257_m.fxe"},
	{0x24AAF626CFE27E8BULL, "640969735.ogg"},
	{0x24AB0F2BA6BC5F7BULL, "6145115_m.fxe"},
	{0x24AC09C6E6513077ULL, "820059458.ogg"},
	{0x24AC248744BDA6F6ULL, "22989578.ogg"},
	{0x24AC2EDC96535E41ULL, "vxca_caverns1_mines_j5t_0_0.vlm"},
	{0x24AC4A77F627CE06ULL, "6107667_f.fxe"},
	{0x24AC52D73FC937AEULL, "hlac_lt_0_lowtown_docks_n_-196608_0z.dds"},
	{0x24AD32E848B97259ULL, "489529587.ogg"},
	{0x24AD5CB7ABA2B942ULL, "hlac_lt_0_lowtown_a2_n_-65536_2x.dds"},
	{0x24ADC53EC4F91DD9ULL, "vxht_keep_6j8_0_0.vlm"},
	{0x24AE0705184314B0ULL, "lowtown_fem_cityguard_mob6164831.cl"},
	{0x24AE74BBD2D3A2DDULL, "454560375.ogg"},
	{0x24AE7D241CD27CF1ULL, "hlac_ca_caverns1_mines_24424_26y.dds"},
	{0x24AE805993436552ULL, "hlac_lt_0_lowtown_docks_n_-131073_2y.dds"},
	{0x24AE845B6356F3C3ULL, "127302541.ogg"},
	{0x24AEA03BBA10EC10ULL, "sse_ss_city_guard_f_1.gda"},
	{0x24AF2EA819765AB0ULL, "dre000_carver6037630.cl"},
	{0x24AFB6F86B1CC8A5ULL, "vxht_0_hightown_n_a1i_0_0.vlm"},
	{0x24B03ADFD6EE4250ULL, "dae000_hate_follower_react6167473.cl"},
	{0x24B057678C591508ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_3z.dds"},
	{0x24B0A7B7C9B85A5EULL, "6088459_m.fxe"},
	{0x24B0BB0A61F4836AULL, "hlac_w_blightlands_196611_1x.dds"},
	{0x24B0E330325A838AULL, "496548765.ogg"},
	{0x24B100EF37898602ULL, "lt_0_lowtown_docks_244_0.msh"},
	{0x24B102EF378989B8ULL, "lt_0_lowtown_docks_244_0.mmh"},
	{0x24B10EEF37899E5BULL, "lt_0_lowtown_docks_244_0.mao"},
	{0x24B12D1AA62947B6ULL, "zz_followers_banter6157716.cl"},
	{0x24B15C093BC2C294ULL, "tbe200im_stealth_boots.uti"},
	{0x24B18EAD429B5E29ULL, "vxht_0_hightown_9xe_0_0.vlm"},
	{0x24B1E492A2E9A72AULL, "llac_w_bonepitdestroyed_a3_65540_0.dds"},
	{0x24B21EEE3A82D0EBULL, "vxw_sundermount_3aj_0_0.vlm"},
	{0x24B258038840FBF9ULL, "6031751_m.fxe"},
	{0x24B2AA8520123B9CULL, "vxht_keep_6ho_0_0.vlm"},
	{0x24B3326D718EC8A8ULL, "895532602.ogg"},
	{0x24B36570D629C157ULL, "llac_lt_0_lowtown_131071_5.dds"},
	{0x24B3B535F933A10FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagkeep_d.dds"},
	{0x24B4566B6CEABA4CULL, "vxca_cavern1_haunted_e85_0_0.vlm"},
	{0x24B4631EF74FA3BDULL, "qun221_petrice6082676.cl"},
	{0x24B467DBE190E10DULL, "w_bonepitmining_a2_148_0.msh"},
	{0x24B46B26C85F07FBULL, "890236357.ogg"},
	{0x24B47178D88221AFULL, "6105471_f.fxe"},
	{0x24B475DBE190F940ULL, "w_bonepitmining_a2_148_0.mao"},
	{0x24B479DBE190FF93ULL, "w_bonepitmining_a2_148_0.mmh"},
	{0x24B4F9188824465AULL, "gen_im_trs_qun_amu_r2.uti"},
	{0x24B5A21E28C919BFULL, "gen_im_arm_cht_comb_t1.uti"},
	{0x24B5D9E093100FD7ULL, "vxw_bonepit_15w_0_0.vlm"},
	{0x24B6E586263F4670ULL, "hlac_lt_0_lowtown_docks_n_-131076_1y.dds"},
	{0x24B6F1F75C7D24E0ULL, "gen00fl_fenris6148977.cl"},
	{0x24B7BC1AEE407143ULL, "vxlt_0_lowtown_docks_n_b5e_0_0.vlm"},
	{0x24B7EBF25CF47A9CULL, "vxht_0_hightown_n_bge_0_0.vlm"},
	{0x24B7F9AF947DE4CFULL, "synclargecreatures4.gda"},
	{0x24B81CF3DD5B8610ULL, "vxca_caverns2_9a0_0_0.vlm"},
	{0x24B8A64FFC4E4CF7ULL, "517933357.ogg"},
	{0x24B913A93B271E39ULL, "hlac_lt_0_lowtown_n_65535_3z.dds"},
	{0x24B91AFF4889257AULL, "dae200_bran6144557.cl"},
	{0x24B979051354B3AAULL, "w_bonepitmining_a2_133_0.mao"},
	{0x24B97D051354BAF9ULL, "w_bonepitmining_a2_133_0.mmh"},
	{0x24B97E4EBF94979EULL, "6108517_m.fxe"},
	{0x24B987051354CBE7ULL, "w_bonepitmining_a2_133_0.msh"},
	{0x24B98E8801840CA7ULL, "6085612_f.fxe"},
	{0x24B9912423C6D5BBULL, "hla_ht_0_hightown_n_-65535_0y.dds"},
	{0x24B9FA2B0BFBA6CCULL, "hlac_w_encounter2_camp_65539_0x.dds"},
	{0x24BA0489AC98B3B1ULL, "vxlt_backalley_n_1kp_0_0.vlm"},
	{0x24BA1BDE61E5802CULL, "279919894.ogg"},
	{0x24BA316038BEB02EULL, "801888344.ogg"},
	{0x24BA659BBE7EC4C2ULL, "mrl150_hunter6071761.cl"},
	{0x24BA6F28E8E264C6ULL, "vfx_blt_fire_wall_p_dur_placed.wbd"},
	{0x24BACC6BF063AC81ULL, "6087660_m.fxe"},
	{0x24BB1BD54E149E54ULL, "mag311_emile_de_launcet6043726.cl"},
	{0x24BB2573B65C3466ULL, "lt_0_lowtown_a3_n_c_44_0_ll.mmh"},
	{0x24BB8752F25933B3ULL, "hlac_lt_0_lowtown_docks_a3_-65538_2z.dds"},
	{0x24BBF850C35D39E1ULL, "dae000_arishok6067716.cl"},
	{0x24BC3A2B4AB011BEULL, "6107978_m.fxe"},
	{0x24BCC67966E2F5C4ULL, "vxht_keep_3zp_0_0.vlm"},
	{0x24BCF8986B8B3970ULL, "6202310_m.fxe"},
	{0x24BD41484F2A82C8ULL, "llac_ca_cave_2194_3.dds"},
	{0x24BD600CE3A70BABULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_rage.ani"},
	{0x24BD8684BC6988DBULL, "595800651.ogg"},
	{0x24BDA7EDEC65D622ULL, "vxht_0_hightown_a3_by7_0_0.vlm"},
	{0x24BDE07688B4DF01ULL, "hlac_lt_gallowstemplar_f_2388_1z.dds"},
	{0x24BEE2BBFF6758AEULL, "vxdr_0_deeproads_bal_0_1.vlm"},
	{0x24BF148D1CE041FFULL, "vxlt_darktown_6l3_0_0.vlm"},
	{0x24BF179461F77888ULL, "288315786.ogg"},
	{0x24BF93917C4AFAB5ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_6x.dds"},
	{0x24BF93EAA4253395ULL, "88075322.ogg"},
	{0x24BFBEBBEFA71077ULL, "follower_banter6164694.cl"},
	{0x24C01EB71A30D4C7ULL, "dae000ip_crowd_mf5_talk2.utp"},
	{0x24C07234A8BB62E8ULL, "vxht_brothel_n_6tw_0_0.vlm"},
	{0x24C098607C8705A4ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.cess.ani"},
	{0x24C09A78CE9DFFB6ULL, "hlac_dr_primevaltomb_658_3z.dds"},
	{0x24C128F2624EF8DDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medc_0t.dds"},
	{0x24C1450A1AA8829AULL, "vxlt_warehouse_n_110_0_0.vlm"},
	{0x24C1B47FD7B560FFULL, "vxca_cavern1_haunted_gh8_0_0.vlm"},
	{0x24C20CCD66BFB55FULL, "w_blightlands_88_0.msh"},
	{0x24C212CD66BFBF01ULL, "w_blightlands_88_0.mmh"},
	{0x24C21ECD66BFD3A2ULL, "w_blightlands_88_0.mao"},
	{0x24C221A5EE2533A6ULL, "vxlt_hangedman_26a_0_0.vlm"},
	{0x24C2284CB0BE47E1ULL, "ene_f.mao"},
	{0x24C295F0A5C9AEB6ULL, "6086970_m.fxe"},
	{0x24C2C81201021F70ULL, "337360482.ogg"},
	{0x24C2CB29C971826AULL, "9904563.ogg"},
	{0x24C2E77A34B7B163ULL, "hlac_lt_gallowstemplar_n_2384_3z.dds"},
	{0x24C2F9AB3C30BEC9ULL, "qun220cr_javaris6071854.cl"},
	{0x24C3086EA576C8D0ULL, "hlac_w_encounter3_-65538_0y.dds"},
	{0x24C3172945D492FAULL, "gen00fl_varric6081258.cl"},
	{0x24C3CA42F9156E58ULL, "vxca_caverns2prc_4x3_0_0.vlm"},
	{0x24C3E8385F888433ULL, "6034376_m.fxe"},
	{0x24C417A6AA9419B5ULL, "lt_0_lowtown_docks_a3_n_175_0.msh"},
	{0x24C419A6AA941D2BULL, "lt_0_lowtown_docks_a3_n_175_0.mmh"},
	{0x24C425A6AA9431C8ULL, "lt_0_lowtown_docks_a3_n_175_0.mao"},
	{0x24C48F65EC14C1A9ULL, "llac_ht_0_hightown_65535_0.dds"},
	{0x24C4AE0DEBC284B5ULL, "hlac_w_encounter2_327683_0z.dds"},
	{0x24C4E8C4A157E8CDULL, "hlac_ht_keep_3732_14z.dds"},
	{0x24C500CA537664B5ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffy_n.dds"},
	{0x24C548AE4B630275ULL, "gen00fl_anders6125366.cl"},
	{0x24C627E06CCF329FULL, "weapon2_s2.psh.fxo"},
	{0x24C6353BC0C8749DULL, "hslmatrix.psh.fxo"},
	{0x24C695607C8B6C06ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.cess.gad"},
	{0x24C6DA6A89057BE2ULL, "vxht_mansion_n_5b7_0_0.vlm"},
	{0x24C6E9BDCC15DEBEULL, "vxlt_gallowstemplar_n_5ya_0_0.vlm"},
	{0x24C711679CA528D7ULL, "vxlt_backalley_g_1f4_0_0.vlm"},
	{0x24C751ACE44F8F54ULL, "vxht_brothel_n_6j9_0_0.vlm"},
	{0x24C7541A6005CC08ULL, "lt_house_mmap_ph.dds"},
	{0x24C7A01E996DA1E6ULL, "6095497_m.fxe"},
	{0x24C820C475956899ULL, "vxw_bonepit_yf_0_0.vlm"},
	{0x24C876669E8ECC31ULL, "vxlt_backalley_n_1vw_0_0.vlm"},
	{0x24C88D46DC3DDCABULL, "229733298.ogg"},
	{0x24C8A0BFC09D95B7ULL, "llac_lt_0_lowtown_a2_65536_5.dds"},
	{0x24C9CB3955D5E490ULL, "vxht_0_hightown_a2_amo_0_0.vlm"},
	{0x24C9DE7AB382CD2FULL, "430445618.ogg"},
	{0x24CB480CE3B32022ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_rage.evt"},
	{0x24CBC3390923390DULL, "vxlt_warehouse_n_qp_0_0.vlm"},
	{0x24CC66599DEB7F97ULL, "additivegradient01.mao"},
	{0x24CCD013B1051A57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v2.gad"},
	{0x24CD3C1754F978B0ULL, "6112097_m.fxe"},
	{0x24CD90607C9182E5ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.cess.evt"},
	{0x24CDC6CA5791A649ULL, "6062782_m.fxe"},
	{0x24CE6377967D0372ULL, "hlac_lt_hangedman_n_-1_11x.dds"},
	{0x24CEC367235120CBULL, "atl_graphicsde_dxt5_00.dds"},
	{0x24CEE0E20FC73F2CULL, "lt_0_lowtown_docks_n_205_0.msh"},
	{0x24CEE6E20FC7491EULL, "lt_0_lowtown_docks_n_205_0.mmh"},
	{0x24CEE87851D307AFULL, "m_mag101ar_caverns.wbd"},
	{0x24CEEAE0BA87D14CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_med.anb"},
	{0x24CEF2E20FC75DB5ULL, "lt_0_lowtown_docks_n_205_0.mao"},
	{0x24CEF649BB2656C0ULL, "29812095.ogg"},
	{0x24CF083B19843C76ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_mora_d.dds"},
	{0x24CF34BD48EB9533ULL, "6118812_m.fxe"},
	{0x24CF69D5EA93982AULL, "vxlt_gallowscourtyard_n_76l_0_0.vlm"},
	{0x24CFB6E70C9E9B1AULL, "vxlt_warehouse_n_161_0_0.vlm"},
	{0x24D00115F84C2B93ULL, "hlac_dr_0_deeproads_17038_5y.dds"},
	{0x24D006384B794A46ULL, "gen00fl_anders6096502.cl"},
	{0x24D08A8ECB54973EULL, "vxht_0_hightown_a3_ca6_0_0.vlm"},
	{0x24D0A3B36D7F156AULL, "vxht_0_hightown_a2_c84_0_0.vlm"},
	{0x24D101252CD64B3AULL, "hlac_lt_0_lowtown_a2_n_c_-2_3z.dds"},
	{0x24D118C39913FB5FULL, "gen00fl_fenris6126697.cl"},
	{0x24D12422EA6570A7ULL, "vxw_bonepitmining_a2_nc_0_0.vlm"},
	{0x24D1637CB96271E1ULL, "287219628.ogg"},
	{0x24D1A9BA1447EF13ULL, "hla_ht_0_hightown_a3_-1_0x.dds"},
	{0x24D1FFC4E2F8D9BFULL, "lowtown_male_beggar_sta6151040.cl"},
	{0x24D2470CE3B93D35ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_rage.gad"},
	{0x24D2F7CD9E9FECBDULL, "612966540.ogg"},
	{0x24D3133279398F5AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_small.mmh"},
	{0x24D3173279399591ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_small.mao"},
	{0x24D325327939ADA0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_small.msh"},
	{0x24D3E513B10B5CE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v2.evt"},
	{0x24D4A2657B8FB4FBULL, "470176985.ogg"},
	{0x24D53C57B0709A55ULL, "lt_0_lowtown_a2_n_c_4_0.phy"},
	{0x24D569C85A10C386ULL, "mag360_thrask6043526.cl"},
	{0x24D5825FEC1D6D1DULL, "907593255.ogg"},
	{0x24D5A5AEDDA6A7CFULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_c.impa.gad"},
	{0x24D5D122B31E9C6AULL, "6052993_m.fxe"},
	{0x24D5DA58CC5E0D16ULL, "6170510_m.fxe"},
	{0x24D600D521A46877ULL, "ht_0_hightown_n_46_0.rml"},
	{0x24D66208A380C89EULL, "6106717_m.fxe"},
	{0x24D67A0A76395085ULL, "731164623.ogg"},
	{0x24D6B4574A318BFFULL, "6201392_m.fxe"},
	{0x24D6C0BFA3BDEF0AULL, "w_sundermount_132_0.mao"},
	{0x24D6C4BFA3BDF659ULL, "w_sundermount_132_0.mmh"},
	{0x24D6CEBFA3BE0747ULL, "w_sundermount_132_0.msh"},
	{0x24D6D320D78643CDULL, "73268092.ogg"},
	{0x24D7562A67D81A8DULL, "w_bonepitmining_a2_102_0.mmh"},
	{0x24D75A2A67D821DEULL, "w_bonepitmining_a2_102_0.mao"},
	{0x24D7682A67D83913ULL, "w_bonepitmining_a2_102_0.msh"},
	{0x24D7959FF8F44388ULL, "w_encounter3_27_0.mmh"},
	{0x24D7A19FF8F4582BULL, "w_encounter3_27_0.mao"},
	{0x24D7B39FF8F476B2ULL, "w_encounter3_27_0.msh"},
	{0x24D7D839476CA072ULL, "6174045_f.fxe"},
	{0x24D82EEAB68E97BBULL, "6140941_m.fxe"},
	{0x24D851F546D3FE17ULL, "ser270_emeric6149381.cl"},
	{0x24D8DEF279934E8BULL, "6121203_m.fxe"},
	{0x24D97F18BA8017C3ULL, "vxca_caverns2_962_0_0.vlm"},
	{0x24D9AA3DEECDD75AULL, "dre000_hireling26038672.cl"},
	{0x24D9AE39129D7A6AULL, "136817615.ogg"},
	{0x24DA3515FB19518DULL, "vxlt_warehouse_n_to_0_0.vlm"},
	{0x24DA3C5ACCEEE387ULL, "mag170_thrask6054036.cl"},
	{0x24DACEA8C2C3F507ULL, "darktown_fem_elf_sta6174535.cl"},
	{0x24DAE783DC09DBC1ULL, "ht_0_hightown_a2_52_0.rml"},
	{0x24DB02D00CF0203DULL, "1013390228.ogg"},
	{0x24DB77DCBBA53264ULL, "6163394_m.fxe"},
	{0x24DBC675E35389AEULL, "vxdr_0_deeproads_bwn_0_1.vlm"},
	{0x24DBCA6E5701A73AULL, "hlac_w_sundermount_-262149_0z.dds"},
	{0x24DC38A74987BEC9ULL, "6069682_m.fxe"},
	{0x24DC6ADD7C53192AULL, "546867745.ogg"},
	{0x24DC98C41CC62D75ULL, "framebuffereffects_cocmask_p0.psh.fxo.dx11"},
	{0x24DC9C05EE3999C9ULL, "6166903_m.fxe"},
	{0x24DC9F757D55E801ULL, "vxw_encounter2_camp_py_0_0.vlm"},
	{0x24DCB5884CA32A27ULL, "573235363.ogg"},
	{0x24DCBAAEDDACEA7CULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_c.impa.evt"},
	{0x24DCE120AE284B3BULL, "vxlt_gallowsprison_n_5wj_0_0.vlm"},
	{0x24DD030A7461CFF7ULL, "vxlt_hangedman_1ta_0_0.vlm"},
	{0x24DD678E75781454ULL, "vxht_0_hightown_a3_9bg_0_0.vlm"},
	{0x24DDDBB2DC89B0AAULL, "vxlt_gallowscourtyard_n_79m_0_0.vlm"},
	{0x24DDE9DF1AB90300ULL, "6054229_m.fxe"},
	{0x24DE104652A29F8AULL, "56618662.ogg"},
	{0x24DE5691505600A8ULL, "vxw_bonepit_240_0_0.vlm"},
	{0x24DE607DAA8892E1ULL, "ser100_party_comments6120597.cl"},
	{0x24DE8A0C836A29CEULL, "vxlt_darktown_6c3_0_0.vlm"},
	{0x24DEE88415BE843CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_109a_sa.dds"},
	{0x24DF008C987897B7ULL, "hlac_dr_0_deeproads_5849_10x.dds"},
	{0x24DF15F2C8748307ULL, "184095462.ogg"},
	{0x24DF5C3E8AA47F5AULL, "hlac_lt_0_lowtown_a3_n_c_196606_0y.dds"},
	{0x24DFD8653C2A6942ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_getupappear.gad"},
	{0x24E006093AE5164FULL, "396481868.ogg"},
	{0x24E08D42E433F10EULL, "936080000.ogg"},
	{0x24E0C897AE5A1D30ULL, "6090885_m.fxe"},
	{0x24E17C1A54B11373ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.cess.evt"},
	{0x24E17CAF5612F049ULL, "mag100_cullen6140914.cl"},
	{0x24E1FD13B117C2DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v2.ani"},
	{0x24E2341A043ADE96ULL, "dae000cr_grand_cleric6148141.cl"},
	{0x24E28263198E7E9FULL, "hlac_lt_0_lowtown_a3_n_c_65536_3z.dds"},
	{0x24E2A2937AD80B4EULL, "6095556_m.fxe"},
	{0x24E2B9F70908A94BULL, "vxlt_gallowscourtyard_6rc_0_0.vlm"},
	{0x24E2C08E59D1AB18ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh04.phy"},
	{0x24E301114BE95026ULL, "combat_lmm_r1_a4.gda"},
	{0x24E316799BBF0AAEULL, "vxw_woundedcoast_16t_0_0.vlm"},
	{0x24E33961C7AA11DDULL, "hlac_ca_caverns2_5802_11x.dds"},
	{0x24E3418932D6C41CULL, "vxht_0_hightown_n_bhx_0_0.vlm"},
	{0x24E36D3B5979089FULL, "gen00fl_anders6125636.cl"},
	{0x24E3967DBDE36CB9ULL, "mag320_orsino6071088.cl"},
	{0x24E3A3B30A748634ULL, "vxht_bartrand_n_4gx_0_0.vlm"},
	{0x24E3E5A14F1DB1E6ULL, "vxca_caverns2_7v5_0_0.vlm"},
	{0x24E3F7E61F64C002ULL, "161889533.ogg"},
	{0x24E4185EDB101B3BULL, "w_bonepit_160_0.mmh"},
	{0x24E4245EDB102FD8ULL, "w_bonepit_160_0.mao"},
	{0x24E432B1E0B69F2AULL, "zz_dae_debug6036590.cl"},
	{0x24E4365EDB104E25ULL, "w_bonepit_160_0.msh"},
	{0x24E4B4081180A3F8ULL, "vxlt_0_lowtown_a2_n_c_bym_0_0.vlm"},
	{0x24E55E4598B76AE4ULL, "dae000cr_grand_cleric6147713.cl"},
	{0x24E580A4BB8C113FULL, "6136715_m.fxe"},
	{0x24E59E4CC71EA1D7ULL, "vxca_caverns1_e7r_0_0.vlm"},
	{0x24E66ED5A2AE9BCAULL, "vxca_caverns2prc_96j_0_0.vlm"},
	{0x24E69319C8BB5D67ULL, "vxlt_darktown_49f_0_0.vlm"},
	{0x24E6FB653C30C339ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_getupappear.evt"},
	{0x24E710E6A41E94F2ULL, "6031830_f.fxe"},
	{0x24E712ED994A5EA1ULL, "nat380_party_banter6090057.cl"},
	{0x24E7ADEE69448A38ULL, "zz_fen_debug6134385.cl"},
	{0x24E7E8848012D172ULL, "hlac_lt_0_lowtown_n_-65535_5z.dds"},
	{0x24E8027ED414B03FULL, "vxht_mansion2b_309_0_0.vlm"},
	{0x24E84F2A67E6824AULL, "w_bonepitmining_a2_102_0.phy"},
	{0x24E8751A54B725B4ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.cess.gad"},
	{0x24E97865FCEDA76DULL, "qun131_taskmaster6039688.cl"},
	{0x24E99A631D9CEC2DULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmb.msh"},
	{0x24E9AC631D9D0AB3ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmb.mmh"},
	{0x24E9C2AEDDB78195ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_c.impa.ani"},
	{0x24EA10C439DDF5CAULL, "vxdr_primevaltomb_8os_0_0.vlm"},
	{0x24EA4769350ADDDBULL, "6033979_m.fxe"},
	{0x24EA92BE2B248E20ULL, "target_cone_invalid.cur"},
	{0x24EAA56A51B80B0EULL, "534113218.ogg"},
	{0x24EAB64AC81A392DULL, "lt_0_lowtown_docks_4_0.rml"},
	{0x24EB11526F4C0081ULL, "vxw_encounter2_camp_kx_0_0.vlm"},
	{0x24EB24371886E06AULL, "ser221_gascard6148839.cl"},
	{0x24EB89E920961580ULL, "6067982_m.fxe"},
	{0x24EC391E941DBD35ULL, "128113845.ogg"},
	{0x24EC40BCB7E98099ULL, "vxlt_0_lowtown_n_f1b_0_0.vlm"},
	{0x24EC5B8095F3DF96ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_grand.msh"},
	{0x24EC658095F3F094ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_grand.mmh"},
	{0x24EC698095F3F7A7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_grand.mao"},
	{0x24EC6E6B04507764ULL, "6045018_m.fxe"},
	{0x24ECAAF53D3CDA45ULL, "zz_slate5_onwhite.cut"},
	{0x24ECFDF90E58409CULL, "dae110ip_at_ws_warehouse.utp"},
	{0x24ED54A74E42C10DULL, "vxlt_darktown_4wf_0_0.vlm"},
	{0x24EDA805638C3188ULL, "art\\vfx\\materials\\vfx_smoke_puffy2.mao"},
	{0x24EE0C4A663501C0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_couldron_sa.dds"},
	{0x24EE13CD66E4FC7EULL, "w_blightlands_88_0.phy"},
	{0x24EE62206D26450CULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslb_0s.dds"},
	{0x24EE841A54BBAA92ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.cess.ani"},
	{0x24EEDF225310C13CULL, "lgt110cr_coa_sailor.utc"},
	{0x24EF3A74A05035B0ULL, "6095785_m.fxe"},
	{0x24EF5D3FAE3107B6ULL, "166008960.ogg"},
	{0x24EF9D51C45DAB5BULL, "vxht_mansion2_266_0_0.vlm"},
	{0x24EFD045611FD0CCULL, "vxlt_undercitydungeon_1ac_0_0.vlm"},
	{0x24F01BA207CF8A91ULL, "logpile_sa.dds"},
	{0x24F060646B23A8AAULL, "gen00fl_anders6096638.cl"},
	{0x24F0D66A4677E0BEULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.dura.ani"},
	{0x24F0E4098ED6CB94ULL, "45968653.ogg"},
	{0x24F1458029823B7AULL, "hlac_lt_0_lowtown_a3_n_c_-262144_1z.dds"},
	{0x24F23D1673E43A28ULL, "6106939_m.fxe"},
	{0x24F23F54A1F8283EULL, "vxlt_gallowstemplar_n_5jd_0_0.vlm"},
	{0x24F403653C3B5B28ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_getupappear.ani"},
	{0x24F40624F7BCC309ULL, "ch1armtnt_ds1y.psh.fxo"},
	{0x24F46FF196DAB09EULL, "6163583_m.fxe"},
	{0x24F49E63FF6130DBULL, "vxlt_shop_ip_0_0.vlm"},
	{0x24F4EB32727537F5ULL, "fxe_sunshine.dds"},
	{0x24F610E2FD2AF125ULL, "98339944.ogg"},
	{0x24F71EC19AC570D1ULL, "llac_w_bonepitmining_a2_458756_0.dds"},
	{0x24F736EA4EE7F4C4ULL, "242306014.ogg"},
	{0x24F74C0328205C63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_around.ani"},
	{0x24F7B20C13585EE8ULL, "qun110_petrice6042205.cl"},
	{0x24F82D41C4A4596AULL, "hlac_ca_cavern1_haunted_24424_2x.dds"},
	{0x24F8C453B2B3D585ULL, "qun110_petrice6018681.cl"},
	{0x24F8E6265D2C2F54ULL, "hlac_ht_mansion2_818_60x.dds"},
	{0x24F8F570C13BC4E9ULL, "vxht_0_hightown_a2_atm_0_0.vlm"},
	{0x24F93232795A9365ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_small.phy"},
	{0x24F9E5CDD4E9F327ULL, "6164626_m.fxe"},
	{0x24FA572A5ACFB90EULL, "hlac_w_sundermount_-65542_0z.dds"},
	{0x24FABC1BD6E411B4ULL, "mcr370_meredith6085871.cl"},
	{0x24FB33FD639C0181ULL, "938957693.ogg"},
	{0x24FC979501467827ULL, "218219367.ogg"},
	{0x24FCFA838163BFDCULL, "vxca_caverns2prc_8t7_0_0.vlm"},
	{0x24FD53BA970779F6ULL, "vxlt_0_lowtown_docks_cod_0_0.vlm"},
	{0x24FDEE6A4682938FULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.dura.evt"},
	{0x24FE08666209726EULL, "vxlt_0_lowtown_f52_0_0.vlm"},
	{0x24FE1E9D5C09ADA2ULL, "hlac_lt_0_lowtown_docks_n_65535_1x.dds"},
	{0x24FE294F14426787ULL, "61896580.ogg"},
	{0x24FE7A73F46AA652ULL, "vxlt_hangedman_n_204_0_0.vlm"},
	{0x24FE8969B383421DULL, "plt_dae021pt_amb_chantry.plo"},
	{0x24FED16804692B16ULL, "6116789_m.fxe"},
	{0x24FEE5D295723101ULL, "art\\levels\\ztools\\default\\generic\\2wr_basic.phy"},
	{0x24FEF13E44CA89B2ULL, "vxca_caverns2_8o5_0_0.vlm"},
	{0x24FF5129A7E6D620ULL, "dae020cr_hm_noble3.utc"},
	{0x2500164A417D0516ULL, "llac_ht_keep_n_4519_18.dds"},
	{0x25003B03F3434DB0ULL, "art\\levels\\races\\proxy\\props\\_textures\\altar_n.dds"},
	{0x2500818CD6F94B9AULL, "hlac_ht_chantry_2534_49z.dds"},
	{0x25009FC34ABF5E7BULL, "170911217.ogg"},
	{0x250124FA6D7335FEULL, "zz_rdr_debug6066632.cl"},
	{0x250132EBC7865A93ULL, "mag360_alain6044623.cl"},
	{0x25016E7FFDD7669EULL, "hlac_lt_0_lowtown_a2_n_-131072_9z.dds"},
	{0x2501E7CCB881AB1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforward.gad"},
	{0x2502BEDD3E50FBB6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_loop.evt"},
	{0x2502CCDB9758BD0BULL, "hlac_lt_gallowstemplar_n_2391_5z.dds"},
	{0x2502CDB9AA11C7C1ULL, "gen00fl_anders6112988.cl"},
	{0x25035FFD954C8DE1ULL, "mag311_keran6041578.cl"},
	{0x250361295E35FC5CULL, "vxca_cave_22u_0_0.vlm"},
	{0x2503A9477C90931DULL, "vxca_caverns1_mines_hmf_0_0.vlm"},
	{0x2504419768767ABEULL, "lgt310pt_fixer_2.ncs"},
	{0x25046403282B0F5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_around.evt"},
	{0x25046BA7366FE70EULL, "hla_ht_0_hightown_a2_65535_0z.dds"},
	{0x2504EF6A4688B438ULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.dura.gad"},
	{0x25054435D45FE52DULL, "vxlt_gallowstemplar_4xs_0_0.vlm"},
	{0x250557267A2F578CULL, "6109596_m.fxe"},
	{0x25058C69B389664AULL, "plt_dae021pt_amb_chantry.nss"},
	{0x2506A36C9029245CULL, "269552801.ogg"},
	{0x2506FFE3104F5174ULL, "gen00fl_varric6075646.cl"},
	{0x2507674DB331DD69ULL, "vxht_0_hightown_a2_aol_0_0.vlm"},
	{0x25078145206CFCDAULL, "vxlt_0_lowtown_a2_n_c_hx6_0_0.vlm"},
	{0x2507F730111B950EULL, "6166782_m.fxe"},
	{0x2507FFAA8B20E477ULL, "vxlt_0_lowtown_f22_0_0.vlm"},
	{0x250801B2EF6F21BCULL, "gen00fl_anders6125853.cl"},
	{0x250845F25141ACDFULL, "vxht_mansion_n_629_0_0.vlm"},
	{0x250990C4CB238877ULL, "hm_amb_keep_guard4.mor"},
	{0x2509B5DD3E570B41ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_loop.gad"},
	{0x2509C9D942CAAB0EULL, "559143700.ogg"},
	{0x2509DA297221C78AULL, "hlac_ca_caverns1_mines_24425_3z.dds"},
	{0x250A130E10C63BA0ULL, "vxlt_0_lowtown_eqc_0_0.vlm"},
	{0x250A2A656D68967BULL, "197407713.ogg"},
	{0x250A3861322EF20DULL, "vxlt_darktown_4yd_0_0.vlm"},
	{0x250AE106536728D1ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_4z.dds"},
	{0x250B630328312C4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_around.gad"},
	{0x250B7C548565FB82ULL, "6137234_m.fxe"},
	{0x250B914015892B80ULL, "hlac_lt_hangedman_-1_4y.dds"},
	{0x250BF1CC41CC0761ULL, "6188107_m.fxe"},
	{0x250C319DF138784DULL, "vxlt_0_lowtown_a2_df2_0_0.vlm"},
	{0x250CB3FF450C01CCULL, "vxlt_undercitydungeon_1oa_0_0.vlm"},
	{0x250D0C22A6FD350EULL, "165050733.ogg"},
	{0x250D59C58A64D5B2ULL, "vxht_0_hightown_be2_0_0.vlm"},
	{0x250DBB408B6D75AAULL, "zz_alpha.ncs"},
	{0x250DE4F2AB7F790FULL, "6127563_f.fxe"},
	{0x250E32D659963C07ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v6.gad"},
	{0x250E63D655CD1C92ULL, "vxht_bartrand_n_54n_0_0.vlm"},
	{0x250EBD85B2ABAFC1ULL, "769460353.ogg"},
	{0x250EE1131B2D6D98ULL, "684718666.ogg"},
	{0x250F192DDEFFB379ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_right_walk.ani"},
	{0x250F27B9A61952F7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potb_d.dds"},
	{0x250F3B3824FFFB66ULL, "vxca_caverns2_7e8_0_0.vlm"},
	{0x250F5F0196C4A265ULL, "501453867.ogg"},
	{0x250FA9F1FEE00651ULL, "dae000_keep_noble_sta6161537.cl"},
	{0x25100278520ACA85ULL, "m_mag101ar_caverns.bnk"},
	{0x25108DC0D8147303ULL, "vxht_brothel_n_52m_0_0.vlm"},
	{0x2510B6DD3E5D2BEFULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_loop.ani"},
	{0x2510E957F84B6AEDULL, "hlac_dr_slavers_3207_1z.dds"},
	{0x25116DD7FEDA294CULL, "vxlt_0_lowtown_a3_n_c_hnk_0_0.vlm"},
	{0x2511AD005CCA5B59ULL, "gen00fl_aveline6132813.cl"},
	{0x2512381F4AB0CE52ULL, "mag211_marethari6021356.cl"},
	{0x251246134644A785ULL, "847131144.ogg"},
	{0x2512C479935A7257ULL, "6165272_m.fxe"},
	{0x25130B998DE85585ULL, "vxlt_0_lowtown_a2_n_c_byt_0_0.vlm"},
	{0x25142BCEC1A75177ULL, "6115438_f.fxe"},
	{0x2514591731E732EAULL, "mcr371cr_harvester2.utc"},
	{0x25145D9E47A710A9ULL, "46118432.ogg"},
	{0x2514B7B708639D4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_2arm_20.ani"},
	{0x2514C483B921F87EULL, "212284681.ogg"},
	{0x2514D0BE2BEDC8B9ULL, "hlac_dr_0_deeproads_17038_4x.dds"},
	{0x251513D2EFBE28DDULL, "vxw_bonepitmining_a2_1nv_0_0.vlm"},
	{0x2515185F4BDACFA2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldrailb.msh"},
	{0x25151A5F4BDAD358ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldrailb.mmh"},
	{0x2515265F4BDAE7FBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldrailb.mao"},
	{0x251527D6599C4814ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v6.evt"},
	{0x25153BE64944BCF2ULL, "hlac_lt_darktown_-1_0x.dds"},
	{0x2515469662CC5F74ULL, "pro000_captain_wright6026084.cl"},
	{0x251547D59ECA7936ULL, "165411532.ogg"},
	{0x2515D92AA527F5E9ULL, "vxht_0_hightown_a2_ajs_0_0.vlm"},
	{0x2515F0F774DCB988ULL, "follower_banter6169704.cl"},
	{0x2515F42DDF0592BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_right_walk.gad"},
	{0x25160E9D70723A56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_2.evt"},
	{0x25163C3C5A877538ULL, "hlac_ht_chantry_2534_2y.dds"},
	{0x25164E115AF47C9BULL, "sst_ss_fereldan_f_1.gda"},
	{0x2516E3D0B247885DULL, "follower_banter6170240.cl"},
	{0x251702CCB8943500ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforward.ani"},
	{0x2517144AAD929F6BULL, "6104567_m.fxe"},
	{0x2517AE4A9F3121FBULL, "94280783.ogg"},
	{0x2517B127CFD261A3ULL, "195289351.ogg"},
	{0x2518238F4B26AB0BULL, "6094675_m.fxe"},
	{0x2518513195B0B998ULL, "hlac_lt_0_lowtown_docks_a3_-262144_0y.dds"},
	{0x2518728F0B60D98AULL, "220849434.ogg"},
	{0x25188B2FEF2D3DD2ULL, "526117369.ogg"},
	{0x25193AC85274A767ULL, "w_encounter2_camp_mmi.gff"},
	{0x25194B99A967E95EULL, "6005015_f.fxe"},
	{0x2519DF299AEF3CAAULL, "hlac_lt_backalley_g_0_15z.dds"},
	{0x251A37747AF3AAF6ULL, "vxlt_0_lowtown_docks_cuf_0_0.vlm"},
	{0x251A9BF5FFD00783ULL, "mag100_cullen6205415.cl"},
	{0x251AC479EDDBAA47ULL, "vxlt_gallowstemplar_f_3r7_0_0.vlm"},
	{0x251AE52C2F1BC8E7ULL, "follower_banter6171094.cl"},
	{0x251B73409E72DCC0ULL, "mrl150_marethari.stg"},
	{0x251BD4F828B6BAB2ULL, "vxca_caverns2_8q7_0_0.vlm"},
	{0x251BFDFA0AFEA1C0ULL, "zz_dae_debug6184891.cl"},
	{0x251C0D900165AC90ULL, "487465705.ogg"},
	{0x251C10E066F69342ULL, "539564815.ogg"},
	{0x251C4F57B0ACBD70ULL, "lt_0_lowtown_a2_n_c_4_0.msh"},
	{0x251C5D57B0ACD48AULL, "lt_0_lowtown_a2_n_c_4_0.mmh"},
	{0x251C6157B0ACDBC1ULL, "lt_0_lowtown_a2_n_c_4_0.mao"},
	{0x251C87425FC21863ULL, "vxca_caverns1_hi5_0_0.vlm"},
	{0x251C9680DED48B13ULL, "236800312.ogg"},
	{0x251D059D707849E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_2.gad"},
	{0x251D112DDF0BE2A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_right_walk.evt"},
	{0x251D982D83E8B956ULL, "6125480_m.fxe"},
	{0x251E10ECA08532B4ULL, "6104784_m.fxe"},
	{0x251EBA358859D818ULL, "6099420_m.fxe"},
	{0x251EC3799CA70013ULL, "mag101_grace6053094.cl"},
	{0x251ED79262DA2710ULL, "l_env_wal_lowtowndocks_wal_calm_p2p.wbd"},
	{0x251F4E2D52E266E9ULL, "129341113.ogg"},
	{0x251F5EE22059C661ULL, "llac_lt_gallowscourtyard_-65536_1.dds"},
	{0x251FBC2C82FBCB5BULL, "gen00flst_varric_maxappriv.stg"},
	{0x251FDFB4F0D041CCULL, "vxlt_0_lowtown_a3_n_c_hkh_0_0.vlm"},
	{0x251FE32D6D968899ULL, "zz_qun_debug6028239.cl"},
	{0x251FEE6F23E5CD91ULL, "pro000_friedrich6050174.cl"},
	{0x2520215F4BE4E6E3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldrailb.phy"},
	{0x252031D84BD0332CULL, "vxlt_gallowsprison_n_6gv_0_0.vlm"},
	{0x2520F0D1751335C8ULL, "hlac_ht_chantry_n_2534_26z.dds"},
	{0x252172D557ABF751ULL, "93118306.ogg"},
	{0x2521AA80A35711BCULL, "gen00fl_varric6033536.cl"},
	{0x2521EA8759A0E714ULL, "hm_lgt_captain01.mor"},
	{0x25223CCE1736755FULL, "vxht_mansion_n_3os_0_0.vlm"},
	{0x25225DB86CB29F86ULL, "6043286_f.fxe"},
	{0x2522AB95D91E76B8ULL, "gen00fl_varric6105479.cl"},
	{0x2522AFB7086FCC9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_2arm_20.evt"},
	{0x2522AFB7B5924118ULL, "806328373.ogg"},
	{0x2522D5FCEBF1C245ULL, "6188622_m.fxe"},
	{0x2522DA86ABDEB035ULL, "cod_hst_free_marches.utp"},
	{0x2522DD91E4EB2020ULL, "w_bonepit_155_0.phy"},
	{0x2523069D707CB68FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_2.ani"},
	{0x25231FD659A8784DULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v6.ani"},
	{0x25236FA87BF8A2DFULL, "hlac_lt_0_lowtown_a2_n_c_65538_5y.dds"},
	{0x252402835690FB09ULL, "6078460_m.fxe"},
	{0x2524671EE4F04FC1ULL, "54830551.ogg"},
	{0x2524FA4419CE99E5ULL, "6085125_m.fxe"},
	{0x252533B190F79D84ULL, "683358553.ogg"},
	{0x2525375A05CF2C98ULL, "993088702.ogg"},
	{0x2525571583A51BBBULL, "vxw_encounter2_camp_2q_0_0.vlm"},
	{0x2525C4B1D6C16FE0ULL, "1045824833.ogg"},
	{0x2525E48A68EF899EULL, "264223959.ogg"},
	{0x2526AED42030A3BEULL, "hlac_w_bonepitmining_a2_393220_0y.dds"},
	{0x2526ED6399D03039ULL, "dae200cr_viscount6087209.cl"},
	{0x2527667022A951A5ULL, "mag101cr_apostate2.utc"},
	{0x2528774A6F708F06ULL, "6131074_m.fxe"},
	{0x25288BC0AEE7A3CEULL, "mcr371_alain6087581.cl"},
	{0x2528A09DA03874FAULL, "485099852.ogg"},
	{0x2528F37807754508ULL, "mrl150_dalish_guard6068159.cl"},
	{0x252923C23F678EF3ULL, "hlac_ht_brothel_n_3373_1y.dds"},
	{0x25297D36EE7B2A62ULL, "hlac_lt_warehouse_65536_1y.dds"},
	{0x252990B70875B688ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_2arm_20.gad"},
	{0x252997B928F832CCULL, "vxlt_undercitydungeon_1ug_0_0.vlm"},
	{0x2529A5E63595C435ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava2_e.dds"},
	{0x2529D3A11EBAB93AULL, "207117633.ogg"},
	{0x2529D8654F0B6777ULL, "gen00fl_anders6119147.cl"},
	{0x252A2353B07531B9ULL, "6119660_m.fxe"},
	{0x252AD963A2E477FEULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.mmh"},
	{0x252B39607BAF92A3ULL, "gen_im_trs_tmp_med_hlm_r1.uti"},
	{0x252C60B82EEF1EA2ULL, "6093820_m.fxe"},
	{0x252CA29250FC6D0CULL, "vxht_0_hightown_n_cfc_0_0.vlm"},
	{0x252CCC36E4CDD8E6ULL, "727684569.ogg"},
	{0x252D0B1AD840F0A7ULL, "6150368_m.fxe"},
	{0x252D3F5C98B24E7DULL, "6083710_m.fxe"},
	{0x252E1B721A2D1569ULL, "primitives.mat.dx11"},
	{0x252EF97CCD8F4479ULL, "character_character_p280.psh.fxo.dx11"},
	{0x2530A3B5A0BF6195ULL, "6166688_m.fxe"},
	{0x25316A7C4384830CULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.dura.ani"},
	{0x25317691B6EDBC8AULL, "6127587_m.fxe"},
	{0x2531796324F156BAULL, "qun110_mugger.cnv"},
	{0x253210265CDC7BD9ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_slaterock_s.dds"},
	{0x253253DDAF3D5747ULL, "vxlt_0_lowtown_docks_n_b5y_0_0.vlm"},
	{0x25326F9E44E5A0E9ULL, "6094307_m.fxe"},
	{0x2532A46EE550F65FULL, "412276117.ogg"},
	{0x2532B092691D0FE0ULL, "rog_assassinstaminarecovery.dds"},
	{0x2532C4BFA40C54B2ULL, "w_sundermount_132_0.rml"},
	{0x2532E5C886726578ULL, "60619541.ogg"},
	{0x2532FE6B67F2585AULL, "668873956.ogg"},
	{0x25331535BA94CDBEULL, "404784296.ogg"},
	{0x253334AA75E2DA38ULL, "vxht_mansion2_25x_0_0.vlm"},
	{0x253361ABC29CAC86ULL, "vxht_mansion2_256_0_0.vlm"},
	{0x25339E9AFC46F522ULL, "vxht_mansion_n_754_0_0.vlm"},
	{0x2533A23C0BD12A3CULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extceiling.dds"},
	{0x2533DEA6F8D0297DULL, "6146822_m.fxe"},
	{0x25340388F110C226ULL, "w_bonepitmining_a2_mmap_ph.dds"},
	{0x25341FB2BC29AD6BULL, "vxw_sundermount_34e_0_0.vlm"},
	{0x253464E0867897F8ULL, "qun220cr_javaris6071493.cl"},
	{0x25348B810959372CULL, "573133389.ogg"},
	{0x2534B40F99D7B281ULL, "801792194.ogg"},
	{0x2534B69DEA34D5C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_faint_enter.ani"},
	{0x2534DDBEDE0526A9ULL, "vxca_cavern1_haunted_fm8_0_0.vlm"},
	{0x25356024FC6A5B53ULL, "449056108.ogg"},
	{0x25356AE514324131ULL, "llac_lt_gallowstemplar_n_7198_5.dds"},
	{0x25358DF808113B8DULL, "vxlt_darktown_4na_0_0.vlm"},
	{0x25359B6324F57902ULL, "qun110_mugger.bnk"},
	{0x25359EA6BD296DB8ULL, "6032063_m.fxe"},
	{0x2535E60D67AE44EBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_countthrone_n.dds"},
	{0x2536B345138AF280ULL, "hlac_ht_mansion2_818_52y.dds"},
	{0x2536BFA810E5CB21ULL, "6102714_m.fxe"},
	{0x2536EF5BABF993C2ULL, "hlac_ht_brothel_n_2995_6z.dds"},
	{0x253719843768C368ULL, "vxw_blightlands_ne_0_0.vlm"},
	{0x253759FCC30EEB95ULL, "hlac_lt_0_lowtown_a3_n_c_65536_8y.dds"},
	{0x25387661EFF4890EULL, "6070908_m.fxe"},
	{0x2538777C438AB84EULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.dura.gad"},
	{0x25389093134C9967ULL, "vxlt_warehouse_n_cd_0_0.vlm"},
	{0x2538B863A2F07CBDULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.anb"},
	{0x2538E3C103A684C6ULL, "vxlt_gallowscourtyard_n_7fp_0_0.vlm"},
	{0x2538EA21C72A2E10ULL, "6037675_m.fxe"},
	{0x25390E930E4E95D4ULL, "350701503.ogg"},
	{0x25391B0AC145DA98ULL, "vxlt_gallowscourtyard_n_45x_0_0.vlm"},
	{0x2539242A1C972A0EULL, "vxw_encounter2_7g_0_0.vlm"},
	{0x25397554B8D4022CULL, "vxht_mansion_n_4xv_0_0.vlm"},
	{0x2539C5BFA4127546ULL, "w_sundermount_132_0.phy"},
	{0x2539E94FABD86A8EULL, "w_sundermount_98_0.msh"},
	{0x2539F74FABD882DFULL, "w_sundermount_98_0.mao"},
	{0x253A034FABD8977CULL, "w_sundermount_98_0.mmh"},
	{0x253A1E4C0F379D77ULL, "mcr370_negotiation6080252.cl"},
	{0x253A4CAA7FB78C58ULL, "547671747.ogg"},
	{0x253A51F937FA0A15ULL, "745509757.ogg"},
	{0x253A874A0AAF15CBULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_encn_0t.dds"},
	{0x253A8EE6C984D58DULL, "follower_banter6156161.cl"},
	{0x253ADA0F53B2305CULL, "hlac_lt_0_lowtown_docks_-65539_4y.dds"},
	{0x253B497A277252A2ULL, "224011556.ogg"},
	{0x253B9B900147789AULL, "vxlt_0_lowtown_a3_n_c_eua_0_0.vlm"},
	{0x253C071F624BA991ULL, "6202024_f.fxe"},
	{0x253C9308C2B11D3BULL, "vxca_caverns1_j4p_0_0.vlm"},
	{0x253CC6A72AFFFF97ULL, "vxht_0_hightown_65g_0_0.vlm"},
	{0x253D15922FBC642CULL, "vxlt_gallowsprison_n_6mt_0_0.vlm"},
	{0x253D195DF02AC204ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.dura.ani"},
	{0x253E13955D324AE9ULL, "876526988.ogg"},
	{0x253E36021A4412B2ULL, "hlac_lt_0_lowtown_a3_n_c_-65539_0x.dds"},
	{0x253E44584015FA89ULL, "550362947.ogg"},
	{0x253E58529AF52DBCULL, "vxca_caverns2prc_9sg_0_0.vlm"},
	{0x253EAD8E5A1FE9ADULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh04.msh"},
	{0x253EBF8E5A200833ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh04.mmh"},
	{0x253F05D0BBAEF2B0ULL, "hlac_lt_0_lowtown_a2_n_-1_6x.dds"},
	{0x253F0988DF991D01ULL, "672336901.ogg"},
	{0x253F0B740F08CE56ULL, "dae010cr_worker_elf_m.utc"},
	{0x253F29DE4127882FULL, "392050044.ogg"},
	{0x253F4786B0DB93D1ULL, "hlac_lt_0_lowtown_65536_1x.dds"},
	{0x253F727C4390CE2DULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.dura.evt"},
	{0x253F977D0F4C5D3AULL, "6149346_f.fxe"},
	{0x253FA91F4DBA8F5CULL, "dae000_orana6149026.cl"},
	{0x253FAE15B23CC905ULL, "853974445.ogg"},
	{0x253FD21E60F82475ULL, "gen00fl_dog6107338.cl"},
	{0x2540768D2B554A11ULL, "ser270_templar6164083.cl"},
	{0x25412C2DFA869ACFULL, "mag160_feynriel6020086.cl"},
	{0x254155266AFA0E2CULL, "mag160st_feynriel_sund.stg"},
	{0x2541CF4B89CAFE40ULL, "1031450794.ogg"},
	{0x254237D6515BE786ULL, "mcr310_argument6127563.cl"},
	{0x25428C938894E1C0ULL, "vxht_bartrand_n_4k8_0_0.vlm"},
	{0x2542918FAE1FC5EBULL, "vxw_sundermount_33d_0_0.vlm"},
	{0x2542CE9DEA413BBAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_faint_enter.evt"},
	{0x254303AA4E6F24B8ULL, "233175649.ogg"},
	{0x2543704040C54BFEULL, "678336027.ogg"},
	{0x254396C522596290ULL, "vxdr_primevaltomb_747_0_0.vlm"},
	{0x2543A6912AE319E3ULL, "lt_0_lowtown_docks_39_0_ll.mmh"},
	{0x2544165DF030DC66ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.dura.gad"},
	{0x25441C1D3AA83393ULL, "58701787.ogg"},
	{0x254449EFDBD8FDE2ULL, "139185469.ogg"},
	{0x254462154B6D4CFAULL, "ser120cr_shade_cut2.utc"},
	{0x25447F33D37BFDDBULL, "gen00fl_isabela6102488.cl"},
	{0x254530C97F099116ULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg_reduced.mao"},
	{0x25453CC97F09A5B5ULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg_reduced.mmh"},
	{0x25453EC97F09A92BULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg_reduced.msh"},
	{0x25458B61295EDBE8ULL, "vxw_blightlands_sb_0_0.vlm"},
	{0x25458DD06C207846ULL, "vxht_mansion2_3ge_0_0.vlm"},
	{0x25459F290B29E4FAULL, "qun231_party_comment6077133.cl"},
	{0x254622CCE18D01A8ULL, "vxlt_0_lowtown_a3_n_c_i0t_0_0.vlm"},
	{0x2546352B629522C9ULL, "art\\vfx\\textures\\vfx_link_smo_fb.dds"},
	{0x2546458AD523182BULL, "dae000_meredith6092612.cl"},
	{0x254685E1159622F8ULL, "hlac_lt_0_lowtown_a2_n_-65535_7x.dds"},
	{0x2546D30DA8B82344ULL, "vxw_bonepit_mc_0_0.vlm"},
	{0x2547013C0002D4C7ULL, "zz_mcr_debug6080787.cl"},
	{0x2547559DF59C9D46ULL, "vxlt_gallowscourtyard_n_7iq_0_0.vlm"},
	{0x254796070E8D428EULL, "vxw_encounter2_2x_0_0.vlm"},
	{0x2547ABD7CD1AFEBFULL, "mcr370_cut_marriage6098262.cl"},
	{0x2547AE8411DBBBAEULL, "6136905_m.fxe"},
	{0x2547B52EE9A9F62CULL, "zz_followers_banter6158005.cl"},
	{0x2547D7CAD618573CULL, "556155517.ogg"},
	{0x2547FA91E50A6DB5ULL, "w_bonepit_155_0.msh"},
	{0x2547FC91E50A712BULL, "w_bonepit_155_0.mmh"},
	{0x25480891E50A85C8ULL, "w_bonepit_155_0.mao"},
	{0x25480CF1E20679C2ULL, "gen00fl_isabela6086448.cl"},
	{0x254825A182957907ULL, "llac_dr_primevalentrance_1725_4.dds"},
	{0x2548648096424D8FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_grand.phy"},
	{0x2548BBBC1CFF3D77ULL, "vxw_encounter2_camp_7j_0_0.vlm"},
	{0x2548CD9DEA45A5ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_faint_enter.gad"},
	{0x2548D9FB10CB0D28ULL, "6129091_m.fxe"},
	{0x2548ECF655301E43ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_intimidate_pose_ex.evt"},
	{0x254A115DF0353F45ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.dura.evt"},
	{0x254A6F9D5FE32943ULL, "791123961.ogg"},
	{0x254AA72ED4043E15ULL, "vxdr_0_deeproads_eua_0_0.vlm"},
	{0x254AF29DBEBF69B2ULL, "llac_lt_0_lowtown_a2_n_1_9.dds"},
	{0x254B1EE288FD10F4ULL, "vxlt_undercityencounter_1t5_0_0.vlm"},
	{0x254B3D054E605740ULL, "gen00fl_isabela6110293.cl"},
	{0x254B4E3153F96621ULL, "ht_0_hightownexplosion_n_30_0.rml"},
	{0x254BA5BB680327FBULL, "hlac_lt_0_lowtown_docks_a3_65535_4x.dds"},
	{0x254BE55BFBF4298BULL, "vxdr_primevaltomb_7rb_0_0.vlm"},
	{0x254C1C8325DE1A42ULL, "zz_dae_debug6047617.cl"},
	{0x254C82B1951D11E3ULL, "vxlt_0_lowtown_docks_bc0_0_0.vlm"},
	{0x254D8A2505CF77DAULL, "vxlt_warehouse_13m_0_0.vlm"},
	{0x254D910E7C3A9733ULL, "852708069.ogg"},
	{0x254DC3C4C6DB451FULL, "zz_followers_banter6157243.cl"},
	{0x254E2CB0F4E5D3D2ULL, "seb321_leliana6197868.cl"},
	{0x254E4D453B1C1BE8ULL, "hm_am_commoner2.mor"},
	{0x254EDB46B7968ADDULL, "vxw_bonepitmining_a2_1zr_0_0.vlm"},
	{0x254EF9DA7572ED21ULL, "vxht_keep_n_8bm_0_0.vlm"},
	{0x254F9C02862FD8E4ULL, "504801419.ogg"},
	{0x254FEF98912AD4E4ULL, "llac_ca_caverns2prc_5802_35.dds"},
	{0x254FFB0BB8A13262ULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v1.gad"},
	{0x25502273AF50D223ULL, "vxlt_house_gf_0_0.vlm"},
	{0x25502BB0003FAC1EULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_male.evt"},
	{0x2550619FC12424F1ULL, "906243852.ogg"},
	{0x2550EC7FAD494FEBULL, "mer140ip_item_rubble.utp"},
	{0x2550FE707A8AFA40ULL, "vxht_bartrand_n_4p9_0_0.vlm"},
	{0x2551077643010210ULL, "4955786.ogg"},
	{0x25520871FE90C299ULL, "186238779.ogg"},
	{0x255211D99C6BDB2EULL, "719129742.ogg"},
	{0x25528260A1159EB8ULL, "6038663_m.fxe"},
	{0x2552B115B8C974F8ULL, "34900602.ogg"},
	{0x255322348DF62075ULL, "vxlt_darktown_2v7_0_0.vlm"},
	{0x255351406D87A8B8ULL, "216163418.ogg"},
	{0x255356B17DFE1355ULL, "614366416.ogg"},
	{0x2553FD3E1B54F468ULL, "vxw_blightlands_tc_0_0.vlm"},
	{0x25540BF3BB05BEA2ULL, "t1_mul_pu3.tnt"},
	{0x255493B1D3587EA0ULL, "801700606.ogg"},
	{0x2555729F14105B95ULL, "6140914_m.fxe"},
	{0x25561E0BB8A5D959ULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v1.evt"},
	{0x255642B000443F49ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_male.gad"},
	{0x25565B402AC78F8EULL, "vxw_blightlands_an_0_0.vlm"},
	{0x2556A38BA485B962ULL, "co_distor.mao"},
	{0x2556D191D57FA955ULL, "591002022.ogg"},
	{0x25572D990EF555F7ULL, "vxw_encounter2_camp_2e_0_0.vlm"},
	{0x25573E6FF687FD43ULL, "vxlt_0_lowtown_fek_0_0.vlm"},
	{0x2557568C6F9E3E38ULL, "6126542_m.fxe"},
	{0x2557771519769926ULL, "606380591.ogg"},
	{0x2558151FED261E83ULL, "vxht_0_hightown_a2_c8y_0_0.vlm"},
	{0x2558F3F2BBCBDA66ULL, "vxca_cavern1_haunted_efu_0_0.vlm"},
	{0x2559034FABF347CFULL, "w_sundermount_98_0.rml"},
	{0x255932412A415EE6ULL, "ca_caverns1_24427.rml"},
	{0x2559594EA21F27A0ULL, "lt_0_lowtown_a2_20_0_ll.mmh"},
	{0x2559C5FE5AF691ECULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_whoresign_n.dds"},
	{0x255A1258890C221CULL, "gen00fl_aveline6139796.cl"},
	{0x255A4BAD1A3F0075ULL, "6131158_m.fxe"},
	{0x255A737128AD8C97ULL, "vxht_keep_6m3_0_0.vlm"},
	{0x255AD2B6A88D3900ULL, "gen00fl_anders6125422.cl"},
	{0x255AF48E87132A63ULL, "vxlt_0_lowtown_docks_bf3_0_0.vlm"},
	{0x255AFD16D63C84C3ULL, "41284107.ogg"},
	{0x255B1261B9ECB15AULL, "fen210_orana6109750.cl"},
	{0x255BC4706847329BULL, "hlac_lt_0_lowtown_65535_6z.dds"},
	{0x255BEE07E9A859F1ULL, "one140cr_charade6137043.cl"},
	{0x255C837A2C65814BULL, "vxw_encounter3_c1_0_0.vlm"},
	{0x255CB312F0C95C51ULL, "6121551_m.fxe"},
	{0x255D13B0004A0D97ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_male.ani"},
	{0x255D2453041370A9ULL, "hm_adriano.mor"},
	{0x255D4D23A98CA35DULL, "vxw_bonepitmining_a2_1wq_0_0.vlm"},
	{0x255D5C8213921ECBULL, "vxca_caverns1_ihw_0_0.vlm"},
	{0x255D7F5520C218A1ULL, "6113904_f.fxe"},
	{0x255DB414AC3753B4ULL, "vxlt_darktown_50k_0_0.vlm"},
	{0x255DDC1848D488B3ULL, "hlac_ht_mansion2_818_40y.dds"},
	{0x255DDF34231B5AB0ULL, "6140270_m.fxe"},
	{0x255E36B94313951EULL, "lt_0_lowtown_a2_20_0.rml"},
	{0x255E4B90A4ABF0A2ULL, "hlac_lt_foundry_n_320_19y.dds"},
	{0x255E8F707D3036C8ULL, "dae200cr_viscount6087387.cl"},
	{0x255EE925E2866EA4ULL, "6128979_m.fxe"},
	{0x255F6F4D9DDD31D2ULL, "vxht_0_hightown_a3_9wz_0_0.vlm"},
	{0x255FBACD4C160194ULL, "vrc321_varric6094459.cl"},
	{0x255FC1CFB97E7F02ULL, "lgt310_shw_ors_informant6156343.cl"},
	{0x255FE24FABF92E47ULL, "w_sundermount_98_0.phy"},
	{0x25600E0EC89FA012ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_ex.gad"},
	{0x25602ED0AA484CB9ULL, "6080711_m.fxe"},
	{0x25606A6DD56E3EA5ULL, "character_character_p64.psh.fxo"},
	{0x2560DB88F24FB6CFULL, "hlac_lt_gallowscourtyard_n_-2_1z.dds"},
	{0x2560F31CC8EC5DF1ULL, "dae000ip_crowd_ff7_talk5.utp"},
	{0x2561382C52257085ULL, "lt_0_lowtown_a2_30_0.rml"},
	{0x25614255A0A54D18ULL, "vxlt_gallowscourtyard_4wi_0_0.vlm"},
	{0x256194117FEC38F5ULL, "vxlt_darktown_2y6_0_0.vlm"},
	{0x256199853AC85CABULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_rnf.ani"},
	{0x2561DF3030BE82D8ULL, "llac_ca_caverns1_mines_24421_12.dds"},
	{0x2561E9B28CE7CACBULL, "cd_dre000_bethany.wbd"},
	{0x2561EAD295C6B73DULL, "art\\levels\\ztools\\default\\generic\\2wr_basic.mao"},
	{0x2561EED295C6BE76ULL, "art\\levels\\ztools\\default\\generic\\2wr_basic.mmh"},
	{0x2561F8D295C6CF74ULL, "art\\levels\\ztools\\default\\generic\\2wr_basic.msh"},
	{0x25621EEA838168AEULL, "ave100_ambush_16106195.cl"},
	{0x2562387FE25A5689ULL, "40982420.ogg"},
	{0x25637FC06CC4B5D1ULL, "gen00fl_anders6128332.cl"},
	{0x2563BAA228AF15E6ULL, "vxht_brothel_n_5kf_0_1.vlm"},
	{0x2563DC5FA1A41168ULL, "52495603.ogg"},
	{0x2564260BB8B22448ULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v1.ani"},
	{0x256479C0F279738EULL, "vxw_encounter2_8z_0_0.vlm"},
	{0x2564AE45FE9EDB81ULL, "fen310_varania6154441.cl"},
	{0x2564B0791FCAC7DFULL, "qun220cr_javaris6071517.cl"},
	{0x2564CD1D1CBDA80EULL, "vxw_blightlands_jo_0_0.vlm"},
	{0x256511C7375A7DB6ULL, "6047256_m.fxe"},
	{0x25651ABAFC0E5CF2ULL, "vxlt_backalley_g_1e8_0_0.vlm"},
	{0x25655A0F5C6F7236ULL, "117507229.ogg"},
	{0x2565A95D7A64A797ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_c.impa.gad"},
	{0x2565BE35C23724E2ULL, "145658508.ogg"},
	{0x2565C16ADB966DE4ULL, "llac_lt_0_lowtown_docks_n_-131073_5.dds"},
	{0x2565C869619456B8ULL, "6123642_m.fxe"},
	{0x2566158A3373FE26ULL, "gen00fl_sebastian6174337.cl"},
	{0x25661FA8BDEA824CULL, "195939596.ogg"},
	{0x256650AFF5D7D5D8ULL, "vxdr_primevaltomb_5bo_0_0.vlm"},
	{0x2566AD92631795D2ULL, "l_env_wal_lowtowndocks_wal_calm_p2p.bnk"},
	{0x2567110EC8A5C449ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_ex.evt"},
	{0x25672BC8BA761E79ULL, "6155475_m.fxe"},
	{0x2567CB785E2D7DFBULL, "follower_banter6156251.cl"},
	{0x25683510B2838E85ULL, "gen00fl_aveline6114862.cl"},
	{0x256858C9D5D330EEULL, "152870517.ogg"},
	{0x25689CCB4867CDD7ULL, "art\\characters\\creatures\\stonegolem\\animation\\enter_old.gad"},
	{0x2569427E195C558AULL, "232239706.ogg"},
	{0x2569452C7180E7FAULL, "vxdr_primevaltomb_6wa_0_0.vlm"},
	{0x25698CD54F4F2383ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmwht.mao"},
	{0x256990D54F4F2AD0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmwht.mmh"},
	{0x25699ED54F4F42EAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmwht.msh"},
	{0x2569A7530CB600CEULL, "hlac_w_encounter3_-131076_0x.dds"},
	{0x256A5167D6DD204BULL, "hlac_lt_blackemporium_1055_10x.dds"},
	{0x256A70D35B20F954ULL, "vxw_sundermount_2f1_0_0.vlm"},
	{0x256AA35E1576FE51ULL, "mcr371_orsino6083323.cl"},
	{0x256AA7494BD34D22ULL, "vxlt_hangedman_306_0_0.vlm"},
	{0x256ACBC728094DABULL, "sky_w_bonepitmining_a2.mao"},
	{0x256AEDF3DD9D755CULL, "331841295.ogg"},
	{0x256AF384EA845453ULL, "1062248927.ogg"},
	{0x256BCA1B521FB9BDULL, "489553968.ogg"},
	{0x256BE16A92796698ULL, "400200936.ogg"},
	{0x256BE7C66CA5D4FCULL, "vxlt_0_lowtown_han_0_0.vlm"},
	{0x256C325C6311AACFULL, "ach_kill_wtf.dds"},
	{0x256C6D1A2B710FF9ULL, "art\\vfx\\textures\\vfx_glyph_impotence.dds"},
	{0x256CA864FEBBB234ULL, "hlac_w_encounter2_camp_196613_0z.dds"},
	{0x256CBE5D7A6AEA24ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_c.impa.evt"},
	{0x256CCBA51590A26EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_quick_modal_activate.ani"},
	{0x256CEC8832CE040BULL, "hlac_lt_gallowstemplar_f_2388_1x.dds"},
	{0x256DB44CE53ED110ULL, "hlac_dr_slavers_3214_0x.dds"},
	{0x256DFFF95A4F50BDULL, "hlac_w_encounter3_-327677_0y.dds"},
	{0x256E698C4E015777ULL, "vxca_caverns1_j4d_0_0.vlm"},
	{0x256E7FC7AE7E0B43ULL, "llac_ca_cavern1_haunted_24422_3.dds"},
	{0x256EB852635F6869ULL, "isa321_velasco6091846.cl"},
	{0x256EF36A730ADE19ULL, "hlac_ht_brothel_n_3373_0x.dds"},
	{0x256F48BF9D859BB0ULL, "175663084.ogg"},
	{0x256F780145AE59A6ULL, "vxlt_hangedman_2zm_0_0.vlm"},
	{0x256F81853AD47122ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_rnf.evt"},
	{0x256FA751B8161301ULL, "96639246.ogg"},
	{0x256FB1CB486E1064ULL, "art\\characters\\creatures\\stonegolem\\animation\\enter_old.evt"},
	{0x2570193A60FB043AULL, "337863354.ogg"},
	{0x25703AE6FA38A3F3ULL, "9491369.ogg"},
	{0x2570607D89A8E1CFULL, "lt_0_lowtown_docks_n_16_0_ll.mmh"},
	{0x25706298584C372FULL, "vxht_brothel_n_3v3_0_0.vlm"},
	{0x257065894921129EULL, "hlac_lt_0_lowtown_a2_n_-65535_6y.dds"},
	{0x257086CB32903706ULL, "6143101_m.fxe"},
	{0x2570B4D283E68177ULL, "hlac_lt_gallowstemplar_f_2391_1z.dds"},
	{0x25710427A0750B73ULL, "plt_cod_aow_codex.plo"},
	{0x25717479034AEB7FULL, "gen00fl_dog6165725.cl"},
	{0x2571B59CC3BE88F8ULL, "6000826_m.fxe"},
	{0x257230F3ACE4DC5CULL, "follower_banter6156027.cl"},
	{0x25723647D5A0AA6CULL, "vxlt_0_lowtown_a2_n_ev9_0_0.vlm"},
	{0x2572C4E41281F0A1ULL, "vxlt_ruinedhouse_n_ab_0_0.vlm"},
	{0x257347BF62F77A3DULL, "ht_0_hightown_38_0.rml"},
	{0x257421406E39452DULL, "vxlt_0_lowtown_docks_n_95h_0_0.vlm"},
	{0x25743A46538A6F38ULL, "6128742_m.fxe"},
	{0x2574E166861A3226ULL, "hlac_lt_gallowscourtyard_-131071_2x.dds"},
	{0x2574F90EC8B1D8B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_ex.ani"},
	{0x257538FA27EEDDADULL, "character_character_p240.psh.fxo.dx11.assembly"},
	{0x2575803BB259EC16ULL, "vxlt_gallowstemplar_51z_0_0.vlm"},
	{0x2575DAD606B04651ULL, "6070042_m.fxe"},
	{0x25767EA9191F5DD5ULL, "vxht_0_hightown_a3_5w5_0_0.vlm"},
	{0x257680853ADA8E35ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_rnf.gad"},
	{0x2576A37D6F4ADA3AULL, "hlac_w_sundermount_65541_0x.dds"},
	{0x2576D67D6F58451FULL, "gen00fl_isabela6135598.cl"},
	{0x257798C61F51926CULL, "gen00fl_anders6103732.cl"},
	{0x2577B7096377007AULL, "vxdr_primevaltomb_6zb_0_0.vlm"},
	{0x2577CBF7C4300809ULL, "vxht_0_hightown_n_c1s_0_0.vlm"},
	{0x2577E37C3C62E15AULL, "6080737_f.fxe"},
	{0x25785FAE4B404336ULL, "vxlt_hangedman_n_25o_0_0.vlm"},
	{0x25788E293B445642ULL, "6052108_m.fxe"},
	{0x2578E2B04D1711D4ULL, "vxw_sundermount_2k6_0_0.vlm"},
	{0x2579881DD9A18811ULL, "one100dg_nabil6033134.cl"},
	{0x2579C3A5159B1EBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_quick_modal_activate.evt"},
	{0x2579C52D1D45C8D9ULL, "6141654_m.fxe"},
	{0x2579D65D7A759D1DULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_c.impa.ani"},
	{0x2579F060CE600DB0ULL, "dae010pt_amb_tavern.ncs"},
	{0x257A1D2F7A8DCB4DULL, "vxlt_gallowstemplar_f_6so_0_0.vlm"},
	{0x257A224DCE7D82E8ULL, "darktown_male_lowlife_mob.cnv"},
	{0x257A7D64B57A3C6FULL, "art\\vfx\\combat\\vfx_w_2nd_whirl_c.impa.gad"},
	{0x257B17AADD659E67ULL, "904874175.ogg"},
	{0x257B9F9DBDB8E231ULL, "114351057.ogg"},
	{0x257C50DE9309F2FAULL, "6131203_m.fxe"},
	{0x257C85B1478087D3ULL, "6172647_m.fxe"},
	{0x257D444DCE7FF250ULL, "darktown_male_lowlife_mob.bnk"},
	{0x257D8DEDEE7B0B3BULL, "w_encounter2_19_0_ll.mmh"},
	{0x257D9D3E4F73A935ULL, "vxca_cavern1_haunted_ixt_0_0.vlm"},
	{0x257DA3C0FE38C483ULL, "6152523_f.fxe"},
	{0x257DC9CB487A765DULL, "art\\characters\\creatures\\stonegolem\\animation\\enter_old.ani"},
	{0x257E08DB747ED076ULL, "isa211_isabela6069848.cl"},
	{0x257E3BF36EB2347BULL, "hlac_ca_caverns2prc_5802_16z.dds"},
	{0x257E4CBE4A38FD13ULL, "gen00fl_fenris6112027.cl"},
	{0x257EB59E8151D4E3ULL, "zz_ave_debug6174725.cl"},
	{0x257ED03FBE94ABC1ULL, "qcr220_orsino6061809.cl"},
	{0x257F485D35465D9CULL, "vxdr_primevalentrance_32q_0_0.vlm"},
	{0x257FE4A892CCD0E3ULL, "vxlt_0_lowtown_a3_n_c_eho_0_0.vlm"},
	{0x2580305471EEA461ULL, "6143241_m.fxe"},
	{0x2580A4A515A108A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_quick_modal_activate.gad"},
	{0x2580B5518C1B6223ULL, "llac_lt_0_lowtown_a2_n_131071_2.dds"},
	{0x2580C9E3CD1B13E7ULL, "vxlt_warehouse_n_zc_0_0.vlm"},
	{0x25810F34467C4D8EULL, "41088694.ogg"},
	{0x258132D97FB55B69ULL, "vxdr_primevaltomb_764_0_0.vlm"},
	{0x25819264B5807E1CULL, "art\\vfx\\combat\\vfx_w_2nd_whirl_c.impa.evt"},
	{0x2581AEA572A27CACULL, "vxht_mansion_n_4ck_0_0.vlm"},
	{0x2581B0D700A9D90EULL, "vxw_blightlands_pa_0_0.vlm"},
	{0x2581DE315327EE6AULL, "one110_gustav6173925.cl"},
	{0x2582BB0CD1313064ULL, "w_blightlands_210_0.mmh"},
	{0x2582BF0CD13136B7ULL, "w_blightlands_210_0.mao"},
	{0x2582CB5FF0538810ULL, "951731176.ogg"},
	{0x2582D10CD1315506ULL, "w_blightlands_210_0.msh"},
	{0x2582F432FDC9CCBEULL, "508345092.ogg"},
	{0x25835423A9E1BF1EULL, "vxca_caverns2prc_a9k_0_0.vlm"},
	{0x25838DAB1EF32D0CULL, "6074061_m.fxe"},
	{0x2583919C34489DD4ULL, "hlac_lt_0_lowtown_a2_n_-65534_0z.dds"},
	{0x2583C560713897EEULL, "hlac_w_woundedcoast_-196600_0x.dds"},
	{0x2584257E1514EEBCULL, "6169925_m.fxe"},
	{0x25842E0474FD8893ULL, "171115656.ogg"},
	{0x2584D7C81557CBCEULL, "art\\characters\\weapons\\greataxes\\wgaxga105a.phy"},
	{0x2585836A7A14CAA5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_staircap_1.mao"},
	{0x2585913252F2C382ULL, "6087563_f.fxe"},
	{0x25862D73A424ECABULL, "dae000ip_crowd_ff2_talk15.utp"},
	{0x2586973E8EC90F3BULL, "vxht_mansion_n_50d_0_0.vlm"},
	{0x2586CA1A2C9CB51FULL, "ave321cr_guardsman2.utc"},
	{0x2586D4E9D5C0DE6CULL, "qun111_ketojan6040938.cl"},
	{0x25874B679BCF6DAAULL, "6165219_m.fxe"},
	{0x2587523206716498ULL, "390389213.ogg"},
	{0x2587548D3F0D2A54ULL, "vxw_sundermount_2l7_0_0.vlm"},
	{0x2587CBB28D086699ULL, "cd_dre000_bethany.bnk"},
	{0x2587D928343A92D2ULL, "938428758.ogg"},
	{0x2588457D2F6DA36EULL, "art\\levels\\4cinematics\\ferelden\\prp_bottlewine.mmh"},
	{0x2588517D2F6DB705ULL, "art\\levels\\4cinematics\\ferelden\\prp_bottlewine.mao"},
	{0x25885F7D2F6DCF5CULL, "art\\levels\\4cinematics\\ferelden\\prp_bottlewine.msh"},
	{0x258889439CCBDFF0ULL, "gencr_revenant.utc"},
	{0x2588AF0B74C848BAULL, "hlac_lt_0_lowtown_a3_n_c_131072_0y.dds"},
	{0x2588ED4C1E18EC8FULL, "zz_followers_banter6159168.cl"},
	{0x25892788C21D25EEULL, "art\\vfx\\environment\\vfx_trap_bolt_p.mmh"},
	{0x25898DA6E577BD1BULL, "6179494_f.fxe"},
	{0x2589CC0159A74705ULL, "265606088.ogg"},
	{0x2589D6E34872843DULL, "87729541.ogg"},
	{0x258AB820ABDB65B3ULL, "art\\characters\\weapons\\staff\\w_stf_004a.phy"},
	{0x258B686BC777D274ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sword_sa.dds"},
	{0x258B7F53658AF288ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right_1.mao"},
	{0x258D3D8A2EB6002EULL, "716101484.ogg"},
	{0x258D4A8A88D49E20ULL, "hf_gen_templar_archer.mor"},
	{0x258DA0623B619E65ULL, "6082256_m.fxe"},
	{0x258DB28CE1E490A5ULL, "6032000_m.fxe"},
	{0x258E9A64B58B1635ULL, "art\\vfx\\combat\\vfx_w_2nd_whirl_c.impa.ani"},
	{0x258ED42A026C26D4ULL, "genip_node_k2.utp"},
	{0x258F02106F70E533ULL, "961970889.ogg"},
	{0x258F9AC378DE4C3CULL, "gen00fl_aveline6115273.cl"},
	{0x258FD3E1BB6244B9ULL, "vxlt_0_lowtown_a3_n_c_hld_0_0.vlm"},
	{0x2590007B2D18556DULL, "vxca_caverns2prc_806_0_0.vlm"},
	{0x259020826498952CULL, "vxht_mansion_n_4fh_0_0.vlm"},
	{0x259022B3F29FF18EULL, "vxw_blightlands_ub_0_0.vlm"},
	{0x2590A48D7AACDFDAULL, "hlac_lt_warehouse_n_0_0z.dds"},
	{0x2590CB0E6A7D7B4CULL, "305669588.ogg"},
	{0x2590CEF4ED4270C1ULL, "vxlt_blackemporium_13r_0_0.vlm"},
	{0x2590EB4EE5E008D3ULL, "qun111_ketojan_post_thug.stg"},
	{0x259171C555CC3882ULL, "910063667.ogg"},
	{0x259247CE9B48D494ULL, "313583669.ogg"},
	{0x259309B7FAB60B5DULL, "vxht_playershouse_n_1r1_0_0.vlm"},
	{0x2593536A31EAFFFBULL, "286162255.ogg"},
	{0x259386B58C5CD1ABULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.cess.ani"},
	{0x2593A4D451824CBFULL, "6052097_m.fxe"},
	{0x2593EF1A069DA9A8ULL, "vxlt_0_lowtown_docks_a3_d3r_0_0.vlm"},
	{0x2593FAE32AFEAA20ULL, "vxca_cave_18l_0_0.vlm"},
	{0x25943CD075F57745ULL, "vxdr_primevaltomb_6wj_0_0.vlm"},
	{0x2594A8BD393A8EADULL, "6051627_m.fxe"},
	{0x2594AFA0E230F68DULL, "265067003.ogg"},
	{0x2594B85CD9F4C35AULL, "hlac_ca_caverns2prc_5802_42z.dds"},
	{0x2595304E7E7AD5CBULL, "mrl000_marethari6129251.cl"},
	{0x25959521C988DD29ULL, "mag121_idunna6084235.cl"},
	{0x2595BCB3D4D79217ULL, "hlac_dr_dungeon_3214_11y.dds"},
	{0x2595C66A310342D4ULL, "vxw_sundermount_2q4_0_0.vlm"},
	{0x2595E340FEFD3F6FULL, "6135938_m.fxe"},
	{0x25960F6FAAF73B14ULL, "6189000_m.fxe"},
	{0x25962E5E48CE109BULL, "6038839_m.fxe"},
	{0x25964688C227E4EDULL, "art\\vfx\\environment\\vfx_trap_bolt_p.anb"},
	{0x2596E110D286827FULL, "qun160_vashothlead6055570.cl"},
	{0x25970A53C4F77043ULL, "hlac_lt_0_lowtown_a2_n_c_-1_3z.dds"},
	{0x2597206EEEDCB071ULL, "841862350.ogg"},
	{0x259750AA0BE13E9EULL, "vxlt_warehouse_153_0_0.vlm"},
	{0x25978A5F8C227186ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.impa.ani"},
	{0x25979AFA0119FA0FULL, "vxht_0_hightown_a3_4of_0_0.vlm"},
	{0x2597A5A211FD457BULL, "llac_lt_0_lowtown_docks_65535_1.dds"},
	{0x259833B9434508C2ULL, "lt_0_lowtown_a2_20_0.anb"},
	{0x2598562CAFF5715FULL, "vxdr_primevaltomb_7zt_0_2.vlm"},
	{0x259875F24B8330ADULL, "vxht_keep_5e2_0_0.vlm"},
	{0x2598F1924C009F4FULL, "gen_im_trs_gen_two_2hs_r2.uti"},
	{0x2599193C222B1DA0ULL, "vxca_caverns2prc_62g_0_0.vlm"},
	{0x2599534C55CB1B6CULL, "vxca_caverns1_mines_i01_0_0.vlm"},
	{0x259956149F9BE0E1ULL, "vxdr_primevaltomb_6j5_0_0.vlm"},
	{0x259AC268EE3322E9ULL, "6167473_f.fxe"},
	{0x259AF0C331D4B6BCULL, "vxw_bonepit_75_0_0.vlm"},
	{0x259C123F2D57B6E5ULL, "6085356_m.fxe"},
	{0x259C9B24301146BFULL, "505106842.ogg"},
	{0x259CBAC9EAFDC779ULL, "zz_dae_debug6029115.cl"},
	{0x259CE0025CA60171ULL, "vxht_0_hightown_n_cbw_0_1.vlm"},
	{0x259CEDEBDA14CCBFULL, "vxlt_0_lowtown_a2_n_ev6_0_0.vlm"},
	{0x259D32233E656B14ULL, "vrc321_varric6086940.cl"},
	{0x259D8126B4BABB98ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_higharmoire.phy"},
	{0x259EF5EC09B718A2ULL, "6080271_m.fxe"},
	{0x259F000FE32CDE3BULL, "163796135.ogg"},
	{0x259F48CADD81E22DULL, "mag210_arianni6031653.cl"},
	{0x259FA5223CC95D75ULL, "226946547.ogg"},
	{0x259FBA24BDE52D42ULL, "hlac_w_encounter3_-131071_0y.dds"},
	{0x259FD34FCA5DC669ULL, "mag320_orsino6071145.cl"},
	{0x25A005BAA0B17AF6ULL, "ico_armor_light_3.dds"},
	{0x25A15D63E91FD124ULL, "6167589_m.fxe"},
	{0x25A16EB58C68E622ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.cess.evt"},
	{0x25A1AEFF5436C410ULL, "hlac_lt_0_lowtown_docks_-65538_9y.dds"},
	{0x25A24887512472D2ULL, "6133793_m.fxe"},
	{0x25A25F8CE78DCD24ULL, "vxlt_gallowstemplar_518_0_0.vlm"},
	{0x25A260F6F893C228ULL, "vxlt_0_lowtown_docks_a3_d4s_0_0.vlm"},
	{0x25A2B229ADD71D46ULL, "882913003.ogg"},
	{0x25A2C46A37198CC9ULL, "396386787.ogg"},
	{0x25A2DECA60F08CC8ULL, "vxdr_primevaltomb_6j0_0_0.vlm"},
	{0x25A2E9F1B473E7D3ULL, "hlac_lt_gallowstemplar_f_2383_0z.dds"},
	{0x25A3CA69A45D105CULL, "art\\characters\\weapons\\longswords\\w_lsw_ls103.mao"},
	{0x25A41E7795783717ULL, "glow4.dds"},
	{0x25A430AC6A9DE4B4ULL, "vxw_woundedcoast_1r2_0_0.vlm"},
	{0x25A4384722F95B54ULL, "vxw_sundermount_2z5_0_0.vlm"},
	{0x25A52592C8953C9EULL, "6105639_m.fxe"},
	{0x25A52993DE050D29ULL, "dspwn_02.mao"},
	{0x25A5725F8C2E86F7ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.impa.evt"},
	{0x25A6349768A17F84ULL, "pro000_athenril6048722.cl"},
	{0x25A66E9A0EB87149ULL, "ht_0_hightown_27_0.rml"},
	{0x25A6E5948F9B7B6BULL, "6126989_m.fxe"},
	{0x25A718E51564C55BULL, "ser221_party_comments6163598.cl"},
	{0x25A71D40AB20D05EULL, "zz_vrc_debug.cnv"},
	{0x25A71F890C747001ULL, "577100287.ogg"},
	{0x25A8125215CABA54ULL, "566682694.ogg"},
	{0x25A85AE8CEC0BA9CULL, "vxca_caverns2_3te_0_0.vlm"},
	{0x25A86DB58C6F0335ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.cess.gad"},
	{0x25A898728700D52DULL, "llac_ht_0_hightown_a3_1_2.dds"},
	{0x25A910E8A39026D6ULL, "vxlt_0_lowtown_eww_0_0.vlm"},
	{0x25A93F750D86BBA6ULL, "vxlt_hangedman_2fq_0_0.vlm"},
	{0x25A9425FECD34F23ULL, "dae000cr_porfiria.utc"},
	{0x25A997D54F851AEBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmwht.phy"},
	{0x25A9C895B9874B88ULL, "6019121_m.fxe"},
	{0x25A9D0A609FF27D0ULL, "6171911_m.fxe"},
	{0x25A9E426737340FCULL, "genip_pp_gli_sustained.utp"},
	{0x25A9F181ACD3ECF9ULL, "qun220cr_javaris6071524.cl"},
	{0x25A9FBE5282330F7ULL, "vxht_mansion_n_57q_0_0.vlm"},
	{0x25AA64CEC14B8022ULL, "fex_spt_cn.msh"},
	{0x25AA66CEC14B83D8ULL, "fex_spt_cn.mmh"},
	{0x25AAF4BC826FC064ULL, "vxlt_0_lowtown_n_h8j_0_0.vlm"},
	{0x25AB2A6B44620F98ULL, "vxlt_0_lowtown_a2_feh_0_0.vlm"},
	{0x25AB414D8693AF89ULL, "mcr371_hostile_merrill6082171.cl"},
	{0x25AB735F8C32F380ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.impa.gad"},
	{0x25AC921A12891D46ULL, "vxw_encounter3_gd_0_0.vlm"},
	{0x25AD714B6AB169E0ULL, "579324464.ogg"},
	{0x25ADBA45368B9CFEULL, "vxht_keep_511_0_0.vlm"},
	{0x25ADCFD2155A4399ULL, "hlac_ca_cavern1_haunted_24424_19y.dds"},
	{0x25AE6088C2C48725ULL, "99930816.ogg"},
	{0x25AF0979292B6DB1ULL, "6149863_m.fxe"},
	{0x25AF1E8B8880A349ULL, "6054551_f.fxe"},
	{0x25AF6A7294596125ULL, "6126561_m.fxe"},
	{0x25AF91A703D24FA2ULL, "6208304_f.fxe"},
	{0x25AFABC27B534CF5ULL, "6021228_m.fxe"},
	{0x25B09500397683A0ULL, "static_static_p63.psh.fxo.dx11"},
	{0x25B15948D203A552ULL, "qun111cr_ketojan.utc"},
	{0x25B167701DBD1B3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\inferno.gad"},
	{0x25B194EE9A25FB94ULL, "gen_im_npc_wep_two_qune.uti"},
	{0x25B19F61DDC9CAAFULL, "mcr310_cut_chantry_blows_up6087557.cl"},
	{0x25B1B8954BA9CEC0ULL, "lgt320_shw_ors_black_op.cnv"},
	{0x25B1CA00E496AFF3ULL, "hm_ser_ghyslain.mor"},
	{0x25B2AB5674F27009ULL, "gen00fl_aveline6139229.cl"},
	{0x25B2D514379D2AE1ULL, "237480922.ogg"},
	{0x25B309D0D5519984ULL, "6028855_m.fxe"},
	{0x25B32E9CFA4E3D40ULL, "vxht_bartrand_n_72x_0_0.vlm"},
	{0x25B3307A43B1DD9DULL, "104417808.ogg"},
	{0x25B36007F0C7EE93ULL, "6168419_m.fxe"},
	{0x25B3B4779EA341FCULL, "vxht_0_hightown_a2_du_0_0.vlm"},
	{0x25B405B90235DB92ULL, "isa111_isabela6067054.cl"},
	{0x25B49A954BABD1F8ULL, "lgt320_shw_ors_black_op.bnk"},
	{0x25B549183A6F4349ULL, "war_defensedmgupgrade.dds"},
	{0x25B5B0F4389F8E1BULL, "788900366.ogg"},
	{0x25B68C8D1A499501ULL, "424361890.ogg"},
	{0x25B6AA9A72A98B7AULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.anb"},
	{0x25B6C6E87A95020CULL, "611104861.ogg"},
	{0x25B6E4224E19C235ULL, "zz_approval_debug6003671.cl"},
	{0x25B76D577893184BULL, "vxdr_primevalentrance_3d8_0_0.vlm"},
	{0x25B782C595863F56ULL, "vxlt_0_lowtown_erx_0_0.vlm"},
	{0x25B787CF7766EE11ULL, "vxlt_warehouse_n_at_0_0.vlm"},
	{0x25B79EF489EBD094ULL, "475308658.ogg"},
	{0x25B7B8D3DEB147B0ULL, "6119122_m.fxe"},
	{0x25B7E1D0F046811CULL, "vxdr_0_deeproads_a24_0_1.vlm"},
	{0x25B7F757F1CC47A7ULL, "238976352.ogg"},
	{0x25B7FD9C43F644A3ULL, "324356453.ogg"},
	{0x25B85C701DC3272CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\inferno.evt"},
	{0x25B85DEC3E61DC2DULL, "vxht_0_hightown_c0k_0_0.vlm"},
	{0x25B875682761BE43ULL, "763717978.ogg"},
	{0x25B89A523E43F0EFULL, "6139871_m.fxe"},
	{0x25B9C37C5D7B1C49ULL, "606746554.ogg"},
	{0x25BA0DB65A87636DULL, "vxlt_0_lowtown_n_gbn_0_0.vlm"},
	{0x25BA2FDD927A8EECULL, "vxlt_gallowstemplar_f_355_0_0.vlm"},
	{0x25BA331F53614966ULL, "pro000tr_barrier_tutorial.utt"},
	{0x25BA453632FC0957ULL, "vxdr_0_deeproads_bdc_0_0.vlm"},
	{0x25BA6310CDB82774ULL, "hlac_lt_0_lowtown_a2_n_65536_10x.dds"},
	{0x25BA6E8976E96F03ULL, "6129856_m.fxe"},
	{0x25BAB679AB404E95ULL, "dae270ip_mage_store.utp"},
	{0x25BB4EA1C4165C10ULL, "6119900_m.fxe"},
	{0x25BB8171C41ABB3CULL, "gen_im_trs_qun_sta_fir_r1.uti"},
	{0x25BB82B69965B02CULL, "6133737_m.fxe"},
	{0x25BC1212E934A3D5ULL, "water_water_p0.vsh.fxo.dx11.assembly"},
	{0x25BD176F94909AF1ULL, "hla_ht_0_hightown_a3_-131071_0x.dds"},
	{0x25BD7491D166D66EULL, "hlac_w_encounter2_camp_196608_0x.dds"},
	{0x25BE5172321068CCULL, "follower_banter6167979.cl"},
	{0x25BEC35375492FB6ULL, "llac_lt_0_lowtown_a2_n_-65536_8.dds"},
	{0x25BF4F323C398205ULL, "llac_dr_primevaltomb_664_5.dds"},
	{0x25C041384288A50FULL, "6066397_m.fxe"},
	{0x25C088EC02D11125ULL, "749146730.ogg"},
	{0x25C0FC6EDAC55CE9ULL, "6167333_m.fxe"},
	{0x25C1239289D35A1CULL, "mag291_vincento6166418.cl"},
	{0x25C151A88BBC5FE7ULL, "face_face_p0.psh.fxo.dx11.assembly"},
	{0x25C1FA91955ED537ULL, "27054382.ogg"},
	{0x25C20C251044D01FULL, "hlac_lt_0_lowtown_docks_a3_n_65534_4z.dds"},
	{0x25C2292942F1849AULL, "607711981.ogg"},
	{0x25C321AB76C74881ULL, "dae300cr_viscount.utc"},
	{0x25C331B4726F13AAULL, "vxht_mansion2_3gi_0_0.vlm"},
	{0x25C3A4673ECCC7F2ULL, "6201252_m.fxe"},
	{0x25C3C360104EDEE8ULL, "qun120cr_javaris6043233.cl"},
	{0x25C3DA7B0C44FF68ULL, "character_character_p188.psh.fxo.dx11"},
	{0x25C400C1BBE7D8F9ULL, "vxlt_gallowscourtyard_n_6yw_0_0.vlm"},
	{0x25C411693D9044F2ULL, "6135702_m.fxe"},
	{0x25C43E67627A952BULL, "ico_shortbow_2.dds"},
	{0x25C4519E70696770ULL, "llac_lt_0_lowtown_docks_a3_n_131070_2.dds"},
	{0x25C4A64BC153F609ULL, "hlac_ht_0_hightown_n_-131071_1x.dds"},
	{0x25C4B7A12F03EE5BULL, "6015001_m.fxe"},
	{0x25C4D6C8158DAD91ULL, "art\\characters\\weapons\\greataxes\\wgaxga105a.mmh"},
	{0x25C4F0C8158DDA4FULL, "art\\characters\\weapons\\greataxes\\wgaxga105a.msh"},
	{0x25C4F80B2226652DULL, "w_encounter2_camp_54_0.mmh"},
	{0x25C4FC0B22266C7EULL, "w_encounter2_camp_54_0.mao"},
	{0x25C50A0B222683B3ULL, "w_encounter2_camp_54_0.msh"},
	{0x25C5287B8326CC62ULL, "284326643.ogg"},
	{0x25C5E0F12EE17D6FULL, "vxca_caverns1_mines_ioc_0_0.vlm"},
	{0x25C5F4A2877C57D6ULL, "vxlt_0_lowtown_emy_0_0.vlm"},
	{0x25C644701DCF3B85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\inferno.ani"},
	{0x25C69F498AF4FA74ULL, "6177276_m.fxe"},
	{0x25C6CFC93057F4ADULL, "vxht_0_hightown_c7j_0_0.vlm"},
	{0x25C6E6EE48334989ULL, "6037072_m.fxe"},
	{0x25C6FC0A85DBE795ULL, "vxlt_darktown_511_0_0.vlm"},
	{0x25C731531F135591ULL, "493926510.ogg"},
	{0x25C7568807386DC3ULL, "gen00fl_sebastian6180845.cl"},
	{0x25C7656E8147C0A9ULL, "vxht_chantry_n_2sx_0_0.vlm"},
	{0x25C766EFE19B1BEDULL, "260859393.ogg"},
	{0x25C78E94EDC754C6ULL, "vxht_mansion2_32n_0_0.vlm"},
	{0x25C7E7D1750C3FF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.evt"},
	{0x25C7FBB0C120DA77ULL, "vfx_chunkyparticle_p0.vsh.fxo.dx11"},
	{0x25C839E0FADAAF1CULL, "hlac_ca_caverns1_mines_24425_4y.dds"},
	{0x25C858E47E0DD166ULL, "6125589_m.fxe"},
	{0x25C93FF8E3B7395AULL, "78209754.ogg"},
	{0x25C97C9E40289A8DULL, "6069677_m.fxe"},
	{0x25CACCE3984953DEULL, "642411534.ogg"},
	{0x25CB1C8211FE1ADAULL, "vxlt_undercitydungeon_wv_0_0.vlm"},
	{0x25CB47EE50DAC023ULL, "520389907.ogg"},
	{0x25CBB25037087758ULL, "vxlt_0_lowtown_a3_n_c_eou_0_0.vlm"},
	{0x25CBCB2E0ACC7F1AULL, "plt_mag320pt_showdown.nss"},
	{0x25CBFFF945A775BAULL, "vxlt_undercitydungeon_19m_0_0.vlm"},
	{0x25CC4C54996435E5ULL, "10490463.ogg"},
	{0x25CC6CEBDF088B6CULL, "vxht_0_hightown_n_9zm_0_0.vlm"},
	{0x25CD5603F8A38FC9ULL, "207030371.ogg"},
	{0x25CDDBB3200DE9D6ULL, "vxca_cave_1au_0_0.vlm"},
	{0x25CDDED175109B81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.gad"},
	{0x25CEA26CD0B51BA0ULL, "321905523.ogg"},
	{0x25CEA9CA22301E95ULL, "mag291_torpor6056815.cl"},
	{0x25CEC0B863F7CC90ULL, "llac_w_woundedcoast_65543_0.dds"},
	{0x25CEF80B222EB9A6ULL, "w_encounter2_camp_54_0.rml"},
	{0x25CF08B617CA961AULL, "6165026_m.fxe"},
	{0x25CF6175DE383714ULL, "65799161.ogg"},
	{0x25CF7D8ADF1A45DBULL, "vxca_cavern1_haunted_ihz_0_0.vlm"},
	{0x25CF9AA52A6C018DULL, "hlac_w_encounter2_camp_327681_0z.dds"},
	{0x25CFB2622170F52DULL, "362120924.ogg"},
	{0x25CFBCE9880FD706ULL, "843465048.ogg"},
	{0x25D03868FF0E0140ULL, "vxlt_undercitydungeon_1c8_0_1.vlm"},
	{0x25D0500B3D9386DBULL, "6051056_m.fxe"},
	{0x25D06716F85A02C6ULL, "643849479.ogg"},
	{0x25D0BD7AEF893226ULL, "hlac_ca_caverns1_mines_24426_8z.dds"},
	{0x25D18B44C8070322ULL, "604177687.ogg"},
	{0x25D27C02E3130445ULL, "mcr371cr_fenris.utc"},
	{0x25D2CFBAE7F58D48ULL, "413020204.ogg"},
	{0x25D30A40AB45EAB9ULL, "zz_vrc_debug.ncs"},
	{0x25D31964989F9D4AULL, "llac_ht_0_hightown_a3_-1_2.dds"},
	{0x25D367D6A0EE2386ULL, "6180900_m.fxe"},
	{0x25D3B2B5E143455AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01b.mmh"},
	{0x25D3C4B5E14363A0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01b.msh"},
	{0x25D40C22432504DEULL, "6060152_m.fxe"},
	{0x25D428DF1A272387ULL, "lowtown_male_beggar_sta6151071.cl"},
	{0x25D4667F79727056ULL, "vxlt_0_lowtown_ehz_0_0.vlm"},
	{0x25D471BF69DBFE69ULL, "6124758_m.fxe"},
	{0x25D4A6C12A0EDE70ULL, "6015770_m.fxe"},
	{0x25D4DFD17516BC2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.ani"},
	{0x25D52D63A8681D5FULL, "gen00fl_anders6107658.cl"},
	{0x25D53B408D5598B4ULL, "llac_ht_0_hightown_n_-196609_0.dds"},
	{0x25D59BF5CA2C36EEULL, "store_gallows_prologue.utm"},
	{0x25D5F10B2234CCEAULL, "w_encounter2_camp_54_0.phy"},
	{0x25D5FEF8E2987EB7ULL, "gen_im_npc_arm_hum_hlm_cotb.uti"},
	{0x25D6423D61C771ACULL, "art\\vfx\\combat\\vfx_gib_cora_ltrb.gad"},
	{0x25D693EB6B2246F5ULL, "lit_bt_02.mao"},
	{0x25D6C3584A98E3F2ULL, "hlac_ht_mansion2_818_59z.dds"},
	{0x25D6CF2E41B22040ULL, "6052601_m.fxe"},
	{0x25D6F04C77C111CEULL, "hlac_lt_hangedman_-1_20y.dds"},
	{0x25D707A59B97A68FULL, "314468607.ogg"},
	{0x25D72DF6D5DF7EA7ULL, "173630904.ogg"},
	{0x25D7566EC81AD606ULL, "vxht_0_hightown_a3_ana_0_0.vlm"},
	{0x25D821825CC41723ULL, "645793229.ogg"},
	{0x25D827DF08D3A239ULL, "310090387.ogg"},
	{0x25D82F31C9BFEF22ULL, "dae000cr_grand_cleric6181906.cl"},
	{0x25D87197AD0F0367ULL, "llac_lt_undercityencounter_-1_36.dds"},
	{0x25D8D236033963A9ULL, "w_blightlands_245_0.mao"},
	{0x25D8DE3603397842ULL, "w_blightlands_245_0.mmh"},
	{0x25D8E03603397BF8ULL, "w_blightlands_245_0.msh"},
	{0x25D96FE85FE7BF2EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_ric_0n.dds"},
	{0x25DA5DF4DD00CFFEULL, "vxht_0_hightown_a3_avq_0_0.vlm"},
	{0x25DA84AC454697E3ULL, "896514585.ogg"},
	{0x25DAA443510CA1D0ULL, "vxw_sundermount_36q_0_1.vlm"},
	{0x25DAE7130E9413A6ULL, "mag100_cullen.cnv"},
	{0x25DB0011B032EE95ULL, "6092569_m.fxe"},
	{0x25DB77A0C06F41F9ULL, "vxdr_0_deeproads_f5n_0_0.vlm"},
	{0x25DB8AEEB1B2CE01ULL, "dae000ip_crowd_ff8_talk2.utp"},
	{0x25DBE1E87C8D747CULL, "394093260.ogg"},
	{0x25DC3B2A4EC93D9AULL, "248521046.ogg"},
	{0x25DC84CAE56B48DDULL, "fx_aoe_radius_no.mao"},
	{0x25DC9251A3C68598ULL, "926811018.ogg"},
	{0x25DD413D61CD8EA1ULL, "art\\vfx\\combat\\vfx_gib_cora_ltrb.anb"},
	{0x25DD413D61CD8EAAULL, "art\\vfx\\combat\\vfx_gib_cora_ltrb.ani"},
	{0x25DD7208C647922EULL, "vxlt_0_lowtown_docks_a3_n_cno_0_0.vlm"},
	{0x25DDC0AD142FAEFEULL, "6061234_m.fxe"},
	{0x25DDE713910E7626ULL, "763540410.ogg"},
	{0x25DDE9130E964D66ULL, "mag100_cullen.bnk"},
	{0x25DE373C723B08CFULL, "688546682.ogg"},
	{0x25DE3C6C9211A7FBULL, "lgt300cr_gang_3_docks.utc"},
	{0x25DE8BE9EC728EA7ULL, "vxlt_0_lowtown_a3_n_c_fek_0_0.vlm"},
	{0x25DF0772F4623F14ULL, "vxht_0_hightown_a3_b19_0_0.vlm"},
	{0x25DF725DA4C15236ULL, "88973382.ogg"},
	{0x25DF76ADFC5698FCULL, "vxlt_0_lowtown_h9f_0_0.vlm"},
	{0x25DF9230E1837130ULL, "dae000ip_crowd_lm5_talk16.utp"},
	{0x25DFAD9A72CC952DULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.mmh"},
	{0x25E002CE8DDD3FB3ULL, "170680005.ogg"},
	{0x25E011C9F285143BULL, "art\\vfx\\combat\\vfx_a_stealth_att_2_c.impa.evt"},
	{0x25E042C136E970D8ULL, "hlac_lt_gallowstemplar_2383_5y.dds"},
	{0x25E11F4DCD245A3BULL, "vxw_encounter2_camp_ql_0_0.vlm"},
	{0x25E125555D7653E2ULL, "672067509.ogg"},
	{0x25E14230E60A290FULL, "mrl150_merrill6070496.cl"},
	{0x25E172F0A4606271ULL, "vxlt_0_lowtown_a3_n_c_hpx_0_0.vlm"},
	{0x25E17C379F4938C6ULL, "34640317.ogg"},
	{0x25E1EE18339CB705ULL, "hlac_lt_0_lowtown_docks_n_-65538_2z.dds"},
	{0x25E2506486C83F21ULL, "face1.vsh.fxo"},
	{0x25E2520EF444A02CULL, "hlac_ht_0_hightown_n_65533_0z.dds"},
	{0x25E26812B241AB45ULL, "hlac_lt_0_lowtown_n_65535_2y.dds"},
	{0x25E315E83CDFCAD8ULL, "mag291_demon_marethari6028850.cl"},
	{0x25E356F83873532FULL, "hlac_w_woundedcoast_0_8x.dds"},
	{0x25E3717FF7CB7F9CULL, "hlac_lt_0_lowtown_a2_-1_7x.dds"},
	{0x25E3968D1E33D2B7ULL, "6188578_m.fxe"},
	{0x25E3B4C59E917193ULL, "zz_vault_debug6178675.cl"},
	{0x25E411F88329DF44ULL, "gen_im_cns_tom_rog_duelist.uti"},
	{0x25E423EB54210F72ULL, "6118761_m.fxe"},
	{0x25E4EAAF33737994ULL, "gen_im_npc_wep_arc_sbwa.uti"},
	{0x25E5C84BBA10EE86ULL, "vxht_0_hightown_a3_akb_0_0.vlm"},
	{0x25E6015D48E70D31ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flaggallows_d.dds"},
	{0x25E60AC9F289747CULL, "art\\vfx\\combat\\vfx_a_stealth_att_2_c.impa.gad"},
	{0x25E629F2CCDFC790ULL, "720270888.ogg"},
	{0x25E6BCC7BEF6BC22ULL, "vxlt_0_lowtown_docks_apk_0_0.vlm"},
	{0x25E6F2AAD9AC2C7DULL, "vxw_sundermount_2zi_0_1.vlm"},
	{0x25E725773D1DEB25ULL, "vxht_keep_n_5x6_0_0.vlm"},
	{0x25E73F353407088AULL, "785353416.ogg"},
	{0x25E76C84A9135280ULL, "gen00fl_aveline6139688.cl"},
	{0x25E79EC646195406ULL, "6143310_m.fxe"},
	{0x25E7BB6E8F9BC337ULL, "6103918_f.fxe"},
	{0x25E80755D3AD3E89ULL, "w_blightlands_18_0.rml"},
	{0x25E8153889CF1C08ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_meat.msh"},
	{0x25E8273889CF3A99ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_meat.mao"},
	{0x25E8333889CF4F32ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_meat.mmh"},
	{0x25E870DAB5D22738ULL, "6177405_m.fxe"},
	{0x25E8CE6AC41A3D38ULL, "qun131_taskmaster6039728.cl"},
	{0x25E8CFD1CEF6E87EULL, "vxht_0_hightown_a3_asr_0_0.vlm"},
	{0x25E9297E39A3EDA4ULL, "llac_ht_0_hightown_n_131074_0.dds"},
	{0x25E9788EB02B9EF1ULL, "vxca_cavern1_haunted_e53_0_0.vlm"},
	{0x25E98DD7BD127ADDULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portrait.mao"},
	{0x25E991D7BD128116ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portrait.mmh"},
	{0x25E99BD7BD129214ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portrait.msh"},
	{0x25E9A11BE0197895ULL, "hlac_ht_mansion2b_818_61x.dds"},
	{0x25EA0922667EFB42ULL, "6217008_m.fxe"},
	{0x25EA3B275E45C00AULL, "6205155_m.fxe"},
	{0x25EA647E7B9D6DF8ULL, "606336398.ogg"},
	{0x25EA763BA6F79123ULL, "pro000_friedrich6051030.cl"},
	{0x25EA81EE73A86210ULL, "6126748_m.fxe"},
	{0x25EB203EE221E9F2ULL, "guitypes_seb.gda"},
	{0x25EBF661B5AF26D1ULL, "vxht_mansion2b_23c_0_0.vlm"},
	{0x25EC0B645B9249B7ULL, "art\\vfx\\creatures\\fxc_arcane_proj_p.mmh"},
	{0x25EC8E5BCF15FA64ULL, "follower_banter6170690.cl"},
	{0x25ED09C9F28F917AULL, "art\\vfx\\combat\\vfx_a_stealth_att_2_c.impa.ani"},
	{0x25EDB120AC2FD728ULL, "art\\characters\\weapons\\staff\\w_stf_004a.mmh"},
	{0x25EDB99E375DE48BULL, "mcr371_hostile_fenris6082234.cl"},
	{0x25EDBD20AC2FEBCBULL, "art\\characters\\weapons\\staff\\w_stf_004a.mao"},
	{0x25EDCF20AC300A52ULL, "art\\characters\\weapons\\staff\\w_stf_004a.msh"},
	{0x25EE0768B296AFB9ULL, "88756441.ogg"},
	{0x25EE1AA3056F8617ULL, "rdr111dg_martin6014695.cl"},
	{0x25EE4CCF8AA88D3FULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v3.gad"},
	{0x25EEC2308FD194F3ULL, "gen_im_arm_boo_med_set_med.uti"},
	{0x25EECA034A02437EULL, "mer340_bonepit_comments6090838.cl"},
	{0x25EF8AA624DEE5B0ULL, "6121983_m.fxe"},
	{0x25EF912ABF1A72BBULL, "vxw_encounter2_camp_lo_0_0.vlm"},
	{0x25F057313F6EE396ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_reaction.evt"},
	{0x25F0658E6030E524ULL, "vxca_caverns2_8m5_0_0.vlm"},
	{0x25F0E5F865FD9D4FULL, "6207000_m.fxe"},
	{0x25F0FEEB10146594ULL, "6129907_m.fxe"},
	{0x25F1499B9686F0C5ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend02_da.dds"},
	{0x25F1731EE0241EF2ULL, "vxw_bonepitmining_a2_217_0_0.vlm"},
	{0x25F1CDB54DBE4120ULL, "hlac_dr_0_deeproads_5838_0x.dds"},
	{0x25F1F2EB7B6866E2ULL, "art\\characters\\weapons\\longswords\\w_lsw_021a.msh"},
	{0x25F1F4EB7B686998ULL, "art\\characters\\weapons\\longswords\\w_lsw_021a.mmh"},
	{0x25F200EB7B687E3BULL, "art\\characters\\weapons\\longswords\\w_lsw_021a.mao"},
	{0x25F295C8461727F2ULL, "6117661_m.fxe"},
	{0x25F2FB55D3B73269ULL, "w_blightlands_18_0.mao"},
	{0x25F30755D3B74602ULL, "w_blightlands_18_0.mmh"},
	{0x25F30955D3B749B8ULL, "w_blightlands_18_0.msh"},
	{0x25F311BD9E72F5A1ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medb.phy"},
	{0x25F44A16B071791CULL, "lt_0_lowtown_docks_n_ll.mmh"},
	{0x25F45AB0C93BA2FDULL, "dre000_dougal6129797.cl"},
	{0x25F48ECB2F48CBAAULL, "6136933_f.fxe"},
	{0x25F49685470A08D0ULL, "vxdr_0_deeproads_cur_0_0.vlm"},
	{0x25F4F184EF24BE61ULL, "1001940020.ogg"},
	{0x25F541CF8AAE992CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v3.evt"},
	{0x25F55EDF9EEFFDB6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_carcass.msh"},
	{0x25F568DF9EF00EB4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_carcass.mmh"},
	{0x25F56CDF9EF015C7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_carcass.mao"},
	{0x25F66DD2D239DB1CULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm2.gad"},
	{0x25F7248FE37CADBFULL, "vxht_0_hightown_n_9zb_0_0.vlm"},
	{0x25F7343A191421CFULL, "hlac_w_encounter3_-262148_0x.dds"},
	{0x25F74E313F74F321ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_reaction.gad"},
	{0x25F7BDB076FE6D1AULL, "vxca_cavern1_haunted_ge2_0_0.vlm"},
	{0x25F7C6D4FA6AC541ULL, "vxht_keep_n_6kx_0_0.vlm"},
	{0x25F841DB85B78519ULL, "dae000_arishok6067824.cl"},
	{0x25F870B84C461735ULL, "vxca_caverns2prc_9lc_0_0.vlm"},
	{0x25F87A58450B21E5ULL, "vxlt_darktown_51a_0_0.vlm"},
	{0x25F87CBCE6187EFBULL, "vxht_0_hightown_c4q_0_0.vlm"},
	{0x25F8C09EE2909CC1ULL, "1008166634.ogg"},
	{0x25F8D337FD6D7B1EULL, "vxw_woundedcoast_1gg_0_0.vlm"},
	{0x25F9285625495D75ULL, "hlac_lt_0_lowtown_n_-65537_3y.dds"},
	{0x25F929A8C85D20E8ULL, "hm_carver_h3s3.mor"},
	{0x25F93E330AEA317AULL, "w_encounter2_camp_26_0.mmh"},
	{0x25F942330AEA37B1ULL, "w_encounter2_camp_26_0.mao"},
	{0x25F950330AEA4FC0ULL, "w_encounter2_camp_26_0.msh"},
	{0x25F96E26B508FA2DULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_higharmoire.msh"},
	{0x25F97C26B5091260ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_higharmoire.mao"},
	{0x25F98026B50918B3ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_higharmoire.mmh"},
	{0x25F996EEC6F164A0ULL, "294375039.ogg"},
	{0x25F9C1D013440C91ULL, "hlac_lt_gallowstemplar_7198_3y.dds"},
	{0x25F9D1B5E1644965ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01b.phy"},
	{0x25FA10AB2F22287EULL, "104781984.ogg"},
	{0x25FA55C2AA33C32EULL, "vxlt_0_lowtown_docks_a3_n_cta_0_0.vlm"},
	{0x25FBCF50E0488DA5ULL, "gen00fl_isabela6135655.cl"},
	{0x25FBCF810A24F757ULL, "follower_banter6141772.cl"},
	{0x25FBFBEB7B70CA23ULL, "art\\characters\\weapons\\longswords\\w_lsw_021a.phy"},
	{0x25FC0BB38BFF31D1ULL, "dae011_corff_sta6149643.cl"},
	{0x25FC344CAF4CDD6EULL, "43809097.ogg"},
	{0x25FC739D8666E433ULL, "558133620.ogg"},
	{0x25FC7EB5F0FABCA1ULL, "vxw_woundedcoast_1js_0_0.vlm"},
	{0x25FCC2D4E85248DBULL, "cs_mcr370cs_animate_guardian.bnk"},
	{0x25FCC819E693AB3FULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlogpile.phy"},
	{0x25FD6CD2D23FF81AULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm2.ani"},
	{0x25FDAB677A35CE0DULL, "38527068.ogg"},
	{0x25FDD17A5B4BFFA0ULL, "combat_smr3_only_range.gda"},
	{0x25FE0307B1108B3BULL, "vxw_encounter2_camp_gn_0_0.vlm"},
	{0x25FE27032C0B0A55ULL, "vxht_bartrand_n_5z3_0_0.vlm"},
	{0x25FE4F313F7B13CFULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_reaction.ani"},
	{0x25FE9451CD29D37FULL, "6119594_m.fxe"},
	{0x25FF1E2141CA4B55ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.mmh"},
	{0x25FF5EFD6BD8A110ULL, "vxdr_primevaltomb_7u8_0_0.vlm"},
	{0x25FFA232E2EB6BFDULL, "gen00fl_fenris6126901.cl"},
	{0x25FFFF18543DF9DCULL, "796028206.ogg"},
	{0x260003AA3B45BA46ULL, "llac_ca_caverns2prc_5802_37.dds"},
	{0x26010A0771A0B0D1ULL, "vxw_woundedcoast_x_0_0.vlm"},
	{0x26020B682356C7FEULL, "hlac_lt_0_lowtown_docks_n_-262144_1z.dds"},
	{0x2602142B19A9AB86ULL, "mag100_karras6035366.cl"},
	{0x26023D0EC9F837E6ULL, "881127177.ogg"},
	{0x26025F9A3019758FULL, "vxht_bartrand_n_6rf_0_0.vlm"},
	{0x2603086239002150ULL, "vxdr_0_deeproads_crs_0_0.vlm"},
	{0x260329CF8ABAAD85ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v3.ani"},
	{0x26034851082EE68DULL, "203235507.ogg"},
	{0x26034DCBB71CD82AULL, "m_var221ar_mansion_real.wbd"},
	{0x2603A081A2E2ED22ULL, "vxlt_0_lowtown_docks_azu_0_0.vlm"},
	{0x2603D668071CDA4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_hurt_wesley.gad"},
	{0x26041506E123233AULL, "vxca_cavern1_haunted_e83_0_0.vlm"},
	{0x260434F22EED4864ULL, "art\\vfx\\materials\\vfx_link_smo_blend.mao"},
	{0x26048276F2652AA1ULL, "102152477.ogg"},
	{0x2606D5624579CD6CULL, "hlac_lt_0_lowtown_-262144_1z.dds"},
	{0x2606E2953E3C2FB5ULL, "vxca_caverns2prc_9kb_0_0.vlm"},
	{0x2606F26B68B68325ULL, "745028896.ogg"},
	{0x26071FF110563A2FULL, "1029482083.ogg"},
	{0x2607C4884EE96639ULL, "gen00fl_merrill6106033.cl"},
	{0x2608261F53A3543FULL, "pro000tr_barrier_tutorial.ncs"},
	{0x260AB9680722C7E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_hurt_wesley.evt"},
	{0x260B0A7003C079B4ULL, "vxlt_darktown_5lw_0_0.vlm"},
	{0x260B36B7D1DA3CF4ULL, "241056963.ogg"},
	{0x260B3D2BBA7A183FULL, "hlac_lt_gallowstemplar_n_2391_1z.dds"},
	{0x260B72B7F86DD504ULL, "hlac_ht_0_hightown_-131072_0z.dds"},
	{0x260B81F91A60D23CULL, "gen00fl_anders6144063.cl"},
	{0x260BD251CE25EDF1ULL, "dae000_oldmangossip6174765.cl"},
	{0x260C1B2141D4D0E2ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.anb"},
	{0x260C74E4A306A3BBULL, "vxw_encounter2_camp_ba_0_0.vlm"},
	{0x260C8185DBBF593DULL, "91858413.ogg"},
	{0x260C8514DB1B5EB2ULL, "vxht_keep_3zv_0_0.vlm"},
	{0x260C9B444DF5AE42ULL, "dae150_pol6148881.cl"},
	{0x260C9BE7AEC3A8AAULL, "theirons.dds"},
	{0x260CD153EC2B8A94ULL, "vxlt_0_lowtown_a2_n_h24_0_0.vlm"},
	{0x260DBB7FF9901001ULL, "dr_primevaltomb_659_lc.mmh"},
	{0x260DC919E6A22044ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlogpile.mmh"},
	{0x260DCD19E6A22697ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlogpile.mao"},
	{0x260DD41833C0850EULL, "llac_lt_gallowscourtyard_65534_3.dds"},
	{0x260DD99DE0669522ULL, "823549205.ogg"},
	{0x260DDF19E6A245E6ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlogpile.msh"},
	{0x260E9EF1A20020F2ULL, "vxlt_0_lowtown_a2_n_feu_0_0.vlm"},
	{0x260EA5A2726EC78BULL, "zz_vault_debug6203351.cl"},
	{0x260F0D72EE6EDB42ULL, "mcr370cr_gate_guardian_top.utc"},
	{0x260F332B15F168A1ULL, "6127210_m.fxe"},
	{0x260F396BCA745BE2ULL, "atl_itemupgrad_dxt5_01.dds"},
	{0x26100A4FCD6D9E49ULL, "57124046.ogg"},
	{0x2610951DCE98945AULL, "art\\vfx\\combat\\vfx_glb_ele_c_minor_c.impa.gad"},
	{0x2610D87D39DCFF99ULL, "515681519.ogg"},
	{0x2611115BAE306EA4ULL, "one350_variel6085179.cl"},
	{0x261123E3F5BA5BEBULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_ric_0d.dds"},
	{0x261171901869B4D8ULL, "vxlt_0_lowtown_gjl_0_0.vlm"},
	{0x2611FF6B7D815E1FULL, "follower_barks6173420.cl"},
	{0x261236DA20D40616ULL, "640480079.ogg"},
	{0x26127B0E130034A5ULL, "vxht_keep_n_5k3_0_0.vlm"},
	{0x26128BA8D1C20C44ULL, "97937520.ogg"},
	{0x2612B14C73374757ULL, "6152317_m.fxe"},
	{0x2612F5D158749477ULL, "vxw_encounter2_camp_qx_0_0.vlm"},
	{0x26133D8EB1B11666ULL, "vxlt_0_lowtown_a2_n_c_hlf_0_0.vlm"},
	{0x26136A52A630E6C2ULL, "atl_achievemen_dxt5_00.dds"},
	{0x2613C7377E3CB4E5ULL, "ave331_jeven6138733.cl"},
	{0x2613E311ECDAB54AULL, "290458392.ogg"},
	{0x26147C104C10B11DULL, "mcr310_cut_chantry_blows_up6116280.cl"},
	{0x261495B8FF1014D0ULL, "1030307620.ogg"},
	{0x261590243DC002D4ULL, "dae300_alistair_king6133787.cl"},
	{0x2615C0E0C8D1A53FULL, "566946937.ogg"},
	{0x2616096F734C8E00ULL, "gen_im_arm_cht_coma.uti"},
	{0x2616F225D8B86762ULL, "vxw_encounter2_camp_n8_0_0.vlm"},
	{0x2617323C08458D49ULL, "vxca_caverns1_fmy_0_0.vlm"},
	{0x2617397C8E1FF42EULL, "vxlt_0_lowtown_docks_a3_n_czc_0_0.vlm"},
	{0x2617981DCE9EB871ULL, "art\\vfx\\combat\\vfx_glb_ele_c_minor_c.impa.evt"},
	{0x2617BC11A3D737C8ULL, "476185696.ogg"},
	{0x2618544D46A9EDCCULL, "gen_im_wep_war_shd_act1_01.uti"},
	{0x26185AE0FB36BCF1ULL, "968176211.ogg"},
	{0x2618B168072EF730ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_hurt_wesley.ani"},
	{0x2618E85E272CB26EULL, "lt_hangedman_mmi.gff"},
	{0x26192B71AD29E6ECULL, "dae000_bartrand6205970.cl"},
	{0x2619AD413CD5BFE0ULL, "703278755.ogg"},
	{0x2619F4515F31C866ULL, "vxca_caverns2prc_abr_0_0.vlm"},
	{0x261A6420588FC6ACULL, "6080274_f.fxe"},
	{0x261BAB482D632534ULL, "hlac_lt_foundry_n_320_10x.dds"},
	{0x261BADC81B34945FULL, "658052575.ogg"},
	{0x261BD32612F7B09FULL, "one350_variel6087042.cl"},
	{0x261C582D41EAA1B6ULL, "vxlt_darktown_50y_0_0.vlm"},
	{0x261D070DE765EFB2ULL, "6176659_m.fxe"},
	{0x261D68F1CC5F7FF0ULL, "hlac_lt_0_lowtown_docks_-262144_2x.dds"},
	{0x261D829AD1592BFAULL, "zz_ave_debug6056686.cl"},
	{0x261DC8193291EAACULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.cess.ani"},
	{0x261DDBD0B714FDA0ULL, "vxlt_0_lowtown_a2_n_c_hd8_0_0.vlm"},
	{0x261DF19A1C6410BFULL, "art\\levels\\races\\proxy\\props\\_textures\\bedsingle_sa.dds"},
	{0x261E5067C915CE48ULL, "681178717.ogg"},
	{0x261E5D330B098185ULL, "w_encounter2_camp_26_0.phy"},
	{0x261ED16DC8BC7CDFULL, "hlac_lt_0_lowtown_a2_-65537_3z.dds"},
	{0x261F29D708941474ULL, "321121297.ogg"},
	{0x261F4E004AF4600BULL, "mag170_hugh6080760.cl"},
	{0x261FBC3C25CE5062ULL, "vxca_caverns1_jw8_0_0.vlm"},
	{0x2620043520007543ULL, "lt_0_lowtown_docks_n_156_0.msh"},
	{0x2620123520008C9DULL, "lt_0_lowtown_docks_n_156_0.mmh"},
	{0x26201635200093AEULL, "lt_0_lowtown_docks_n_156_0.mao"},
	{0x262032AAB8575BD0ULL, "additivering.mao"},
	{0x2620D607AA50D460ULL, "6173782-anderswearylook.cut"},
	{0x2620F916901B18F1ULL, "pro000_colonel_ewald6040559.cl"},
	{0x262199B3BF66D31EULL, "6122805_m.fxe"},
	{0x26237EE387E53F65ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_008a_0d.dds"},
	{0x2623AF285500FABAULL, "hlac_w_encounter1_-65538_0y.dds"},
	{0x2623DF0843AB5C65ULL, "gen00fl_sebastian6189035.cl"},
	{0x2623ECC10C012A8CULL, "vxht_brothel_n_52v_0_0.vlm"},
	{0x2624B01DCEA96B40ULL, "art\\vfx\\combat\\vfx_glb_ele_c_minor_c.impa.ani"},
	{0x2624BF32F3257C45ULL, "hlac_w_encounter3_-393211_0x.dds"},
	{0x2624C1FFFCFE8064ULL, "hlac_lt_gallowstemplar_n_2383_8y.dds"},
	{0x2624D51932981FEEULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.cess.gad"},
	{0x2624ECBD616D4502ULL, "6172981_m.fxe"},
	{0x2624F47ABABB5745ULL, "mag100_thrask6032044.cl"},
	{0x2625212E0041A6B2ULL, "ave100_donnic6105995.cl"},
	{0x26253E330B0F6B91ULL, "w_encounter2_camp_26_0.rml"},
	{0x26256402CAAE7FE2ULL, "vxw_encounter2_camp_i9_0_0.vlm"},
	{0x26261EF12B7632C8ULL, "dae_mage_singletarget.ncs"},
	{0x2626377409CDB976ULL, "vxw_encounter2_camp_p_0_0.vlm"},
	{0x2626F99F1FF48525ULL, "6107458_m.fxe"},
	{0x2627E8C22849B660ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_mist_4x16.mmh"},
	{0x262840965E445780ULL, "vxw_sundermount_35i_0_0.vlm"},
	{0x262841FD7D19471DULL, "one280ip_sealed_door.utp"},
	{0x262898DCF6F56A32ULL, "pro000_varric6170667.cl"},
	{0x262899F21C5BE316ULL, "vxht_0_hightown_byb_0_0.vlm"},
	{0x2628C23C517D69C8ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_arh_0s.dds"},
	{0x2628CF62DAB70870ULL, "llac_lt_0_lowtown_docks_a3_-131073_3.dds"},
	{0x2628E99B6B22020DULL, "vxht_mansion2_1rs_0_0.vlm"},
	{0x2629300FFA73A426ULL, "vxht_keep_n_58a_0_0.vlm"},
	{0x26296776AC26C82DULL, "hlac_lt_darktown_-1_35y.dds"},
	{0x2629B43D7F7830AAULL, "6101795-mrl250_poll_wall2v2.cut"},
	{0x262A0352E3F9A6D7ULL, "dae000_bartrand6030630.cl"},
	{0x262A1FC937E5A97DULL, "449001753.ogg"},
	{0x262A27F65D9BA742ULL, "209531309.ogg"},
	{0x262A35FE18D7B8AAULL, "vxw_encounter3_gh_0_0.vlm"},
	{0x262A7FB7D31D3898ULL, "sst_ss_athenril.gda"},
	{0x262A831D9D59D342ULL, "6095853_m.fxe"},
	{0x262B4FFA4A41FCA9ULL, "gen00fl_sebastian6174651.cl"},
	{0x262B69DFF085DCCBULL, "6080181_f.fxe"},
	{0x262BD019329E35CDULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.cess.evt"},
	{0x262BF0AD0A2E1C17ULL, "vxlt_hangedman_n_1vl_0_0.vlm"},
	{0x262CF6311C79085CULL, "6035299_m.fxe"},
	{0x262D70D5D8C16EFCULL, "vxca_cavern1_haunted_edq_0_0.vlm"},
	{0x262DCE9880E961C0ULL, "6041249_m.fxe"},
	{0x262E05E81F146D88ULL, "gen_im_trs_bst_spd_gland_04.uti"},
	{0x262E06C5B9067E69ULL, "llac_w_blightlands_65533_0.dds"},
	{0x262E089EF1FEAC8CULL, "gen00fl_bethany6109228.cl"},
	{0x262E92DE942FF9C6ULL, "vxw_encounter3_8m_0_0.vlm"},
	{0x262F72CDCD95C819ULL, "w_encounter3_51_0.msh"},
	{0x262F7CCDCD95D907ULL, "w_encounter3_51_0.mmh"},
	{0x262F80CDCD95E054ULL, "w_encounter3_51_0.mao"},
	{0x262FD7E8ABB0BE16ULL, "hlac_lt_warehouse_n_0_9y.dds"},
	{0x262FECA72946FFCBULL, "6093523_f.fxe"},
	{0x26304CC1C104C316ULL, "vxlt_undercityencounter_zc_0_1.vlm"},
	{0x2630B83A95C6B5EFULL, "6085255_m.fxe"},
	{0x2630BE58A9F81FECULL, "31770570.ogg"},
	{0x2630C6C9B81149B1ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingb_n.dds"},
	{0x2630E82E0B21B08DULL, "plt_mag320pt_showdown.plo"},
	{0x2631996CA22B9464ULL, "dae100_viscount6103308.cl"},
	{0x26326FE3846470A4ULL, "dae000cr_grand_cleric6148091.cl"},
	{0x2632D66164959998ULL, "vxw_encounter1_dm_0_0.vlm"},
	{0x26333A432F975791ULL, "6135378_m.fxe"},
	{0x263376E1D216389EULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_stomp.ani"},
	{0x2633C9B7546191E9ULL, "6165568_m.fxe"},
	{0x2633D965FE360E41ULL, "6218190_m.fxe"},
	{0x26341E6F1FEC5399ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatikib.msh"},
	{0x263426B6C038518DULL, "zz_followers_banter6158976.cl"},
	{0x2634286F1FEC6487ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatikib.mmh"},
	{0x26342C6F1FEC6BD4ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatikib.mao"},
	{0x2634861AD2186DB2ULL, "721061656.ogg"},
	{0x263499935A64F289ULL, "6170579_m.fxe"},
	{0x2634F454C0B7AF2FULL, "6156036_m.fxe"},
	{0x263500B7C7E48D3AULL, "hlac_lt_gallowscourtyard_n_-65537_4x.dds"},
	{0x263514E1641A4D3AULL, "6033407_f.fxe"},
	{0x2635236FBB0B83BDULL, "694328704.ogg"},
	{0x26354416F4F5722AULL, "563167921.ogg"},
	{0x2635F35B7AF9EF35ULL, "vxht_keep_5x9_0_0.vlm"},
	{0x2635FDC0E53794C6ULL, "dae000_arishok6068486.cl"},
	{0x26362C59A7EB27DDULL, "vxlt_blackemporium_tj_0_0.vlm"},
	{0x2636605CDC554F0BULL, "qun231cr_varnell_minion5.utc"},
	{0x2636748CB56F7AAEULL, "hlac_ht_mansion2b_818_22z.dds"},
	{0x2636AEDB74E29AAEULL, "gen_im_trs_trs_two_2hs_r2.uti"},
	{0x2637042A1B4D3A07ULL, "vxlt_darktown_618_0_0.vlm"},
	{0x2637190F1A7BAB70ULL, "ico_shield_4.dds"},
	{0x26375B785D181A8DULL, "vxht_mansion2_1ur_0_0.vlm"},
	{0x263775C10D14EA8FULL, "6043895_m.fxe"},
	{0x2637DBE1E800D81BULL, "dae000_denier6162688.cl"},
	{0x2637E102BCBD5C40ULL, "220908818.ogg"},
	{0x26383472ADC1F40AULL, "gen00fl_aveline6132777.cl"},
	{0x26385617D4DB4546ULL, "gen_im_trs_krk_arc_lbw_r1.uti"},
	{0x26396E1BFA39F4CBULL, "vxdr_primevalentrance_393_0_0.vlm"},
	{0x2639BD0B5B662950ULL, "hlac_w_bonepitmining_a2_458756_0y.dds"},
	{0x263A55E2D6172351ULL, "vxlt_0_lowtown_n_ekj_0_1.vlm"},
	{0x263A5B147F384C46ULL, "822527385.ogg"},
	{0x263AD7DFDECB4E85ULL, "6125403_m.fxe"},
	{0x263B03611052F129ULL, "786471071.ogg"},
	{0x263C2D7F72A9A449ULL, "mag101_grace6035845.cl"},
	{0x263C71A04D07C954ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st104a_0n.dds"},
	{0x263C72BEF135E6DEULL, "ht_chantry_n_2534.rml"},
	{0x263CCAB875B53B18ULL, "ht_0_hightownexplosion_n_22_0.rml"},
	{0x263CFFDB0DC843D9ULL, "996012727.ogg"},
	{0x263D04BB86261246ULL, "vxw_encounter3_5n_0_0.vlm"},
	{0x263D55C5C3943B0FULL, "llac_dr_primevaltomb_9621_5.dds"},
	{0x263DA63F379E183CULL, "549705763.ogg"},
	{0x263DC98684065D96ULL, "382128165.ogg"},
	{0x263E3D56224B2A70ULL, "6111272_m.fxe"},
	{0x263E7FE64ABF2367ULL, "6079564_m.fxe"},
	{0x263E9DB9ABFE8BC0ULL, "vxlt_darktown_5a0_0_0.vlm"},
	{0x263ED4ED5B4EE33BULL, "vxlt_0_lowtown_a3_n_c_gb7_0_0.vlm"},
	{0x263F4904AC88B320ULL, "6105839_m.fxe"},
	{0x263F875FAEAB98B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_idle.ani"},
	{0x263FB6A8F325E72EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_roof.msh"},
	{0x263FC4A8F325FF7FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_roof.mao"},
	{0x263FD0A8F326131CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_roof.mmh"},
	{0x2640694F836A55E3ULL, "6177367_m.fxe"},
	{0x2641483E568BB218ULL, "vxw_encounter1_ij_0_0.vlm"},
	{0x26418EE1D2229F6FULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_stomp.evt"},
	{0x26421FBAE5F9E763ULL, "hlac_lt_0_lowtown_n_65535_3x.dds"},
	{0x26428D4353D7BBF5ULL, "t1_mue_ro1.tnt"},
	{0x2642B37D9D3A8F30ULL, "6128874_m.fxe"},
	{0x2642C07813A69705ULL, "mag100_cullen6206261.cl"},
	{0x26432816A391B173ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_2y.dds"},
	{0x2643C289DA345ECFULL, "6083764_f.fxe"},
	{0x2643D89A7F6F86DDULL, "cod_hst_kirk_lowtown.utp"},
	{0x2643DE604BE239C8ULL, "6098878_m.fxe"},
	{0x2644B0D4E88FDFD9ULL, "cs_mcr370cs_animate_guardian.wbd"},
	{0x264506871A486551ULL, "follower_banter6136703.cl"},
	{0x2645580800DE7FB4ULL, "24797108.ogg"},
	{0x2645774ED1BAEC23ULL, "6135970_m.fxe"},
	{0x264596D93A7FA890ULL, "vxlt_darktown_41u_0_0.vlm"},
	{0x26460E76787CAFA0ULL, "6173746_m.fxe"},
	{0x26461ED08D366A22ULL, "248752091.ogg"},
	{0x2646625FAEB178FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_idle.gad"},
	{0x26468E8379897C91ULL, "6142494_f.fxe"},
	{0x2646B8DCF0B1106DULL, "mrl310_merrill6120115.cl"},
	{0x26478FE1D2270C18ULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_stomp.gad"},
	{0x2647DFF8EC300D4BULL, "vxdr_primevalentrance_322_0_0.vlm"},
	{0x2647E4F308426EACULL, "674587214.ogg"},
	{0x2647E61874B53851ULL, "166450720.ogg"},
	{0x2647E7AFC9C8C59CULL, "vxht_mansion_n_6oa_0_0.vlm"},
	{0x2647F91B42411629ULL, "pro000_wesley6007717.cl"},
	{0x264815085EF75335ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keep_1.mao"},
	{0x26482758ABE95827ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_03oldgrave_1.mao"},
	{0x264937EDCF0865A2ULL, "hlac_lt_0_lowtown_a2_n_0_9y.dds"},
	{0x264949BCD0C16705ULL, "6122503_m.fxe"},
	{0x26497CBFCE5BDC7CULL, "vxht_chantry_n_2ca_0_0.vlm"},
	{0x264A2269BB843A93ULL, "follower_barks6153259.cl"},
	{0x264A2BCBB758A140ULL, "m_var221ar_mansion_real.bnk"},
	{0x264AA0D1482133E4ULL, "vrc321_varric6086799.cl"},
	{0x264AA88C7F225C80ULL, "6128272_m.fxe"},
	{0x264B0D2340927356ULL, "vxht_mansion_n_6j4_0_0.vlm"},
	{0x264B35833C8B0466ULL, "face_rendertodepth_p0.vsh.fxo"},
	{0x264B60A1400F8871ULL, "6155772_m.fxe"},
	{0x264BAC95F2D24620ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_d.dds"},
	{0x264BD5BAC16AB0CDULL, "vxht_0_hightown_a3_bu8_0_0.vlm"},
	{0x264C5DF45A28144AULL, "6164466_f.fxe"},
	{0x264C63BE6914134CULL, "164466281.ogg"},
	{0x264CCB81D6B72D46ULL, "625059531.ogg"},
	{0x264D086C3C4BA41EULL, "568295193.ogg"},
	{0x264D0F969DF4A440ULL, "vxlt_darktown_5j1_0_0.vlm"},
	{0x264D43C09C14EAADULL, "lt_gallowscourtyard_n_77_0_lc.mmh"},
	{0x264D9226423C9B9BULL, "ser221_party_comments6209850.cl"},
	{0x264DBDAC38950BA6ULL, "vxlt_0_lowtown_docks_cti_0_0.vlm"},
	{0x264DE03CFB9FD95EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_exit.ani"},
	{0x264E36C1572D4D70ULL, "vxht_playershouse_n_1ri_0_1.vlm"},
	{0x264E90663D26A50CULL, "vxht_0_hightown_a2_c6l_0_0.vlm"},
	{0x264E9A5032D86D76ULL, "203749522.ogg"},
	{0x264EC1A4D3F22579ULL, "lgt120_orz_bln_official6140180.cl"},
	{0x264F11B05F7650D4ULL, "vxlt_0_lowtown_a2_n_c_bxr_0_0.vlm"},
	{0x264F67446CBE57B1ULL, "vxlt_gallowscourtyard_n_6yo_0_0.vlm"},
	{0x264F7BA6CA7BCCEFULL, "mag100_cullen6140927.cl"},
	{0x2650603BAC7B5A51ULL, "228127104.ogg"},
	{0x265060FAA5B11FBEULL, "dre000_hireling16038624.cl"},
	{0x2650A96F07B1AE38ULL, "240230186.ogg"},
	{0x2650CDBCD8D77B1BULL, "6083997_f.fxe"},
	{0x2651475FBD73B1AAULL, "vxlt_warehouse_sp_0_0.vlm"},
	{0x265167DF9F3E6BAFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_carcass.phy"},
	{0x26516F23929C3C4CULL, "al_bethany_talents.gda"},
	{0x26517EEB5954CCB6ULL, "621169791.ogg"},
	{0x2651B3A63E0107A4ULL, "444777010.ogg"},
	{0x2651C1EBF6D99367ULL, "6055593_m.fxe"},
	{0x2652EF82C9209770ULL, "6121020_m.fxe"},
	{0x26534DF0A1D4D9D2ULL, "int_test.arl"},
	{0x26534E67FD322321ULL, "403792743.ogg"},
	{0x26543F3241044B8DULL, "vxht_mansion2_1cp_0_0.vlm"},
	{0x265467122D6E6FD2ULL, "1029996996.ogg"},
	{0x2654E69C98273921ULL, "vxlt_darktown_5c7_0_0.vlm"},
	{0x265501D34BD38A51ULL, "517382591.ogg"},
	{0x265516BD9EC5C8DDULL, "art\\characters\\playercharacter\\humanfemale\\boo_medb.mao"},
	{0x26551ABD9EC5CF16ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medb.mmh"},
	{0x265524BD9EC5E014ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medb.msh"},
	{0x26553BE0E8FF479FULL, "418502575.ogg"},
	{0x26557F8FFA4C8FE3ULL, "584435222.ogg"},
	{0x2655A5E9E8BE9446ULL, "ave200_aveline_bar6116375.cl"},
	{0x26563E1E1F68F640ULL, "652613467.ogg"},
	{0x265731586864B11BULL, "meleering8.gda"},
	{0x26578F978D2DD684ULL, "hlac_lt_gallowstemplar_2391_9z.dds"},
	{0x26579202E33525B2ULL, "vxdr_primevaltomb_57f_0_0.vlm"},
	{0x2657DC192C74F6B8ULL, "vxht_keep_5ec_0_0.vlm"},
	{0x26597F0032888BD6ULL, "vxht_mansion_n_6m5_0_0.vlm"},
	{0x265A3F20B63D78BCULL, "dae200_cullen.cnv"},
	{0x265A4797B360C94DULL, "vxht_0_hightown_a3_bp7_0_0.vlm"},
	{0x265A5CBED1138A96ULL, "6169915_f.fxe"},
	{0x265A843A417034F0ULL, "cheat.ncs"},
	{0x265AC437B1221FF3ULL, "vxlt_gallowstemplar_n_5or_0_0.vlm"},
	{0x265AF83CFBAA8C2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_exit.evt"},
	{0x265B57611CED270EULL, "llac_ht_0_hightown_n_65537_2.dds"},
	{0x265B65CDCDBAED3CULL, "w_encounter3_51_0.phy"},
	{0x265B76DF6B7F90E6ULL, "vxlt_0_lowtown_a2_n_c_hsk_0_0.vlm"},
	{0x265B8CDB0851D1FBULL, "38267698.ogg"},
	{0x265C1C6DD3337DD8ULL, "vxca_cavern1_haunted_icm_0_1.vlm"},
	{0x265C3946473FB68FULL, "vxlt_0_lowtown_docks_a3_8rn_0_0.vlm"},
	{0x265C5C78301C7A82ULL, "and230st_base_dissent_start.stg"},
	{0x265CB4F7B2A8385EULL, "vxca_caverns1_mines_ehw_0_0.vlm"},
	{0x265CD42D88BBC7DDULL, "mcr310_party_banter6087806.cl"},
	{0x265CE790A5E0D113ULL, "dae000_pc_home_ambients6208636.cl"},
	{0x265D065CFF9B9A6FULL, "hf_qmq_patronf1.mor"},
	{0x265D99F4428DAFFFULL, "gen00fl_carver6103683.cl"},
	{0x265DB50D9F4DBD72ULL, "vxht_mansion2b_2oh_0_0.vlm"},
	{0x265E11F0055B640FULL, "fx_rip_flipbook.mao"},
	{0x265E36FAB027208BULL, "hlac_lt_0_lowtown_a2_n_c_1_9y.dds"},
	{0x265E4E0B0211CDFAULL, "hlac_ht_0_hightown_a3_-131071_0x.dds"},
	{0x265E6120B6419B0CULL, "dae200_cullen.bnk"},
	{0x265E86E2E7D959DBULL, "vxw_bonepitdestroyed_a3_1vj_0_0.vlm"},
	{0x265EC2391766A54CULL, "hlac_ca_cavern1_haunted_24423_7x.dds"},
	{0x265F8E25FCF2E2A1ULL, "gencr_high_dragon_flemeth.utc"},
	{0x265F9DB0D1A00EBBULL, "gen00fl_varric6093597.cl"},
	{0x265FAC99EAF55DDDULL, "lt_0_lowtown_docks_a3_n_33_0.rml"},
	{0x265FB93CAF69CA2AULL, "vxlt_warehouse_tq_0_0.vlm"},
	{0x265FC62F34FFC004ULL, "dae000_gamlen6005754.cl"},
	{0x265FF56F9B9AC1F5ULL, "hlac_lt_0_lowtown_a2_n_65534_1y.dds"},
	{0x2660116F201178BCULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatikib.phy"},
	{0x2660387C28710DA1ULL, "gen00fl_dog6145224.cl"},
	{0x26604D7E98D8E4E5ULL, "6077130_f.fxe"},
	{0x266092A84F9C64A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_action.ani"},
	{0x26612C737312923AULL, "vxlt_0_lowtown_a3_n_c_hnm_0_0.vlm"},
	{0x266157483DEE4C56ULL, "gen00fl_anders6125796.cl"},
	{0x2661A956DA1BBB1CULL, "gen00fl_anders6111348.cl"},
	{0x2661F93CFBB0ACD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_exit.gad"},
	{0x2662327A09BF7677ULL, "6111364_f.fxe"},
	{0x26627CCDCDC13364ULL, "w_encounter3_51_0.rml"},
	{0x2662B10F32FA640DULL, "vxht_mansion2_1fo_0_0.vlm"},
	{0x266307C5279534D0ULL, "6017808_m.fxe"},
	{0x26637552D14FD84BULL, "6126528_f.fxe"},
	{0x2663AF39F0C65593ULL, "mag160_varian_ilithis6018876.cl"},
	{0x266476DC12903086ULL, "919314922.ogg"},
	{0x26648A51E5C538C7ULL, "altar_d.dds"},
	{0x266507F91F71D4EBULL, "gen_im_npc_wep_shd_genc.uti"},
	{0x2665AFA8F346AAE7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_roof.phy"},
	{0x2665B820D2067E17ULL, "vxlt_hangedman_n_1bh_0_0.vlm"},
	{0x266600E95551E921ULL, "gen00fl_anders6095713.cl"},
	{0x266695BBD7E4E7E8ULL, "844716312.ogg"},
	{0x2666DE237145C0B1ULL, "vxdr_0_deeproads_f5v_0_0.vlm"},
	{0x2667A399F0EF6F0FULL, "hightown_fem_noble2_mob.bnk"},
	{0x2667AB457F6A2461ULL, "llac_dr_dungeon_3207_1.dds"},
	{0x2667ADA84FA2B12BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_action.gad"},
	{0x2667D5D422DC57F6ULL, "327194837.ogg"},
	{0x266903403672DE9BULL, "sg7_cl.mao"},
	{0x2669209ACD1373F8ULL, "follower_banter6161322.cl"},
	{0x266922972F7902E8ULL, "vxlt_0_lowtown_a2_n_c_eki_0_1.vlm"},
	{0x26693614A3183873ULL, "vxlt_gallowstemplar_n_5bu_0_0.vlm"},
	{0x266946C4A80C169BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_encn_0d.dds"},
	{0x2669A95C4D68C642ULL, "mag291_ntbanter6056829.cl"},
	{0x266A01CC4AE14C60ULL, "ico_armor_light_5.dds"},
	{0x266A1830BBCB96F7ULL, "6064971_m.fxe"},
	{0x266A21E0731405B0ULL, "dae000_gen_turn_in_4.cnv"},
	{0x266A32262372EC93ULL, "710676206.ogg"},
	{0x266A337118AEB3D9ULL, "881104762.ogg"},
	{0x266B2FE6B718802CULL, "one140cr_charade6137611.cl"},
	{0x266B377DA8199ACBULL, "zz_personality_debug6004635.cl"},
	{0x266B74202112D60CULL, "vxht_0_hightown_a2_c8b_0_0.vlm"},
	{0x266BAEFA1F8596DDULL, "mer210_coterie_contact6079362.cl"},
	{0x266BC599F0F39277ULL, "hightown_fem_noble2_mob.cnv"},
	{0x266BED4F7075608DULL, "mag210_arianni6030467.cl"},
	{0x266C372AB36CE54AULL, "mag360_thrask6043654.cl"},
	{0x266C86BD8EFD55FEULL, "vxht_0_hightown_c5c_0_0.vlm"},
	{0x266CAC9CFAE34711ULL, "6098260_m.fxe"},
	{0x266D03BA95EE52B6ULL, "art\\vfx\\creatures\\fxc_bm_vomr_c.impa.evt"},
	{0x266D43E073167508ULL, "dae000_gen_turn_in_4.bnk"},
	{0x266DAAA84FA71790ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_sweet_action.evt"},
	{0x266DEF53563C2E0EULL, "ca_cavern1_haunted_26035_ll.mmh"},
	{0x266E042CD95522B2ULL, "751890107.ogg"},
	{0x266E2B19A15FE2AAULL, "vxlt_warehouse_yr_0_0.vlm"},
	{0x266EAFBFDC2D7DBDULL, "t3_shd_hrn.tnt"},
	{0x266EC4F8B266548DULL, "l_pro000ar_blightlands_real.bnk"},
	{0x266F5B86D4502180ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p2_c.mmh"},
	{0x266F9E506508AABAULL, "vxlt_0_lowtown_a3_n_c_hkn_0_0.vlm"},
	{0x266FD60C95136CC1ULL, "976080278.ogg"},
	{0x267113C3FFBBB3A8ULL, "6067858_m.fxe"},
	{0x267152AEF5197065ULL, "follower_banter6163987.cl"},
	{0x2671571D1370FEF9ULL, "vxlt_gallowscourtyard_n_65c_0_0.vlm"},
	{0x2671BE58B6EB6FD0ULL, "proberender.sub.xml"},
	{0x26721C8DC1F29AD5ULL, "6075315_m.fxe"},
	{0x2672AC1E109D0FC9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1a.phy"},
	{0x26734898C857BA44ULL, "6174952_m.fxe"},
	{0x267356D8084CF10CULL, "4379513.ogg"},
	{0x26735A16B70D2708ULL, "572046493.ogg"},
	{0x267367AAC96A5BF3ULL, "102974026.ogg"},
	{0x26736F31D2A391ACULL, "649032086.ogg"},
	{0x2673D5F8C7DA6FE4ULL, "6189036_f.fxe"},
	{0x2673FABA95F46241ULL, "art\\vfx\\creatures\\fxc_bm_vomr_c.impa.gad"},
	{0x26746D2A71A62F0DULL, "6018703_m.fxe"},
	{0x2674D95E4DCD2A94ULL, "gen00fl_isabela6099561.cl"},
	{0x26752D36C23810C6ULL, "vxlt_0_lowtown_a2_exn_0_0.vlm"},
	{0x267596401A2263B7ULL, "6119190_m.fxe"},
	{0x2676630445476442ULL, "mag101_grace6034380.cl"},
	{0x26766355FB1AF0C2ULL, "395160507.ogg"},
	{0x26766E1EEE68A863ULL, "vxw_woundedcoast_e5_0_0.vlm"},
	{0x26766FBEF1675982ULL, "ht_chantry_n_2534.anb"},
	{0x2676A2724134FA05ULL, "hlac_w_encounter3_-262142_0x.dds"},
	{0x267749A4A84DAA55ULL, "92709884.ogg"},
	{0x26776D2B33E9B146ULL, "zz_fol_debug6103802.cl"},
	{0x26778913DB4E54A3ULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90l.ani"},
	{0x2677D15D10072556ULL, "vxlt_undercitydungeon_15m_0_0.vlm"},
	{0x267811847C42B1E3ULL, "431204636.ogg"},
	{0x26784C1D929E01BCULL, "947562496.ogg"},
	{0x26789A76D0B1AB1BULL, "6114838_m.fxe"},
	{0x267949D1C6BB7333ULL, "qun100ip_at_exit.utp"},
	{0x267987517C97FD20ULL, "6145178_m.fxe"},
	{0x2679C6ADA69F829BULL, "944338178.ogg"},
	{0x267A0EB570E7932AULL, "661454020.ogg"},
	{0x267A26D4649E7000ULL, "75370862.ogg"},
	{0x267A2E4199FD3906ULL, "vxht_0_hightown_n_cf9_0_0.vlm"},
	{0x267A5806CDD97B0CULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_win.ani"},
	{0x267A7B1CC6B1FE65ULL, "vxlt_darktown_5bh_0_0.vlm"},
	{0x267ABD3E267C887CULL, "vxca_cave_180_0_0.vlm"},
	{0x267AFBBA95FA82EFULL, "art\\vfx\\creatures\\fxc_bm_vomr_c.impa.ani"},
	{0x267B79CD852DBC75ULL, "static_passthrough_p0.psh.fxo.dx11.assembly"},
	{0x267B8C2255F3CCE8ULL, "6111665_m.fxe"},
	{0x267C3E86D45A7A9FULL, "art\\vfx\\combat\\vfx_c_pride_cast_p2_c.anb"},
	{0x267C59B48108D1A5ULL, "6053728_m.fxe"},
	{0x267C92D44C761A73ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightpath_darelief.dds"},
	{0x267CBF953737C30FULL, "hlac_w_encounter2_camp_-131071_0y.dds"},
	{0x267CD7D5DD4093FAULL, "103289720.ogg"},
	{0x267D01306E48E49AULL, "art\\characters\\creatures\\dragon\\animation\\blendtree\\dragon.anb"},
	{0x267D2A2CDC0A1198ULL, "1052906777.ogg"},
	{0x267D2B9CC7C73A9DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capleft.mao"},
	{0x267D2F9CC7C741D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capleft.mmh"},
	{0x267D399CC7C752D4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capleft.msh"},
	{0x267D403123D68D67ULL, "zz_approval_debug6003651.cl"},
	{0x267DE55BB9EE6D21ULL, "hlac_ht_0_hightown_a3_-131074_0x.dds"},
	{0x267E0D9075573770ULL, "hlac_dr_primevalentrance_1725_2y.dds"},
	{0x267E7FF6E5C94A79ULL, "hlac_lt_0_lowtown_a2_n_c_0_2y.dds"},
	{0x267E99BB62E9BBB7ULL, "844625840.ogg"},
	{0x267EA2D2EC45AB4EULL, "126104273.ogg"},
	{0x267F1FDB51C197F9ULL, "bur_smoke.mao"},
	{0x267FB6B6D486F547ULL, "hlac_w_encounter2_camp_327681_0x.dds"},
	{0x267FC38B9A1477C9ULL, "zz_followers_banter6159492.cl"},
	{0x26800EB11B77E0A7ULL, "pro000_colonel_ewald6040593.cl"},
	{0x26803C336E0982A1ULL, "vxlt_darktown_5r4_0_0.vlm"},
	{0x2680663B5D0431AAULL, "ts_slavers_wpn_table.gda"},
	{0x2680818E3725205CULL, "6046882_m.fxe"},
	{0x26815813B415A628ULL, "woodrope_sa.dds"},
	{0x26816506CDDFB04EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_win.gad"},
	{0x2681DA76D23C989EULL, "hla_ht_0_hightown_a2_-65536_0z.dds"},
	{0x2682165055D39F58ULL, "6130956_m.fxe"},
	{0x268230F207A393CBULL, "6171945_m.fxe"},
	{0x26828FB25E6351E9ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdga_0s.dds"},
	{0x26830CD9425570D5ULL, "145928615.ogg"},
	{0x268364868196F9A4ULL, "clf260_fell_orden6070489.cl"},
	{0x26841A5274ACD8FFULL, "w_woundedcoast_194_0.msh"},
	{0x2684205274ACE2A1ULL, "w_woundedcoast_194_0.mmh"},
	{0x26842C5274ACF742ULL, "w_woundedcoast_194_0.mao"},
	{0x268498D411618AFCULL, "rock_jagged_specular.dds"},
	{0x2684CA082565590BULL, "757953278.ogg"},
	{0x2684DFFBE05EC0E3ULL, "vxw_woundedcoast_b4_0_0.vlm"},
	{0x2684F4976EE14894ULL, "vxdr_dungeon_1m1_0_1.vlm"},
	{0x2685091B9BF28E3DULL, "295478271.ogg"},
	{0x2685511042504CF5ULL, "vxlt_backalley_n_1n1_0_2.vlm"},
	{0x268564795E9B12FEULL, "400920436.ogg"},
	{0x2685A113DB5ABA9AULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90l.evt"},
	{0x2685C349D2B1E0EEULL, "hm_ser_theodore.mor"},
	{0x2685D254F92707ECULL, "hlac_w_sundermount_-327684_0z.dds"},
	{0x2685ED63DAD80C19ULL, "195457868.ogg"},
	{0x2686433A01FD3DD6ULL, "vxlt_undercitydungeon_18n_0_0.vlm"},
	{0x268662F15E038D06ULL, "vxht_0_hightown_c7u_0_0.vlm"},
	{0x268663ADE0A94C7FULL, "6070814_f.fxe"},
	{0x2686A337DEB6329AULL, "gen00fl_merrill6141176.cl"},
	{0x2686FB1EEE08D4EAULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs08a_0e.dds"},
	{0x268705EE9C6B80E5ULL, "follower_banter6141514.cl"},
	{0x26876D0EE05CB90FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_1groupcrag.mao"},
	{0x2687790EE05CCDACULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_1groupcrag.mmh"},
	{0x268779FF260C7CE0ULL, "dae000_ineria6136319.cl"},
	{0x26877F0EE05CD79EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_1groupcrag.msh"},
	{0x2687E13B875526CEULL, "655510501.ogg"},
	{0x26880A8E888A81DEULL, "vxca_caverns1_mines_egj_0_0.vlm"},
	{0x26886006CDE5C62DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_win.evt"},
	{0x268864C02FD18DF7ULL, "gen00fl_merrill6121814.cl"},
	{0x268883CF3EEBA35CULL, "6169363_m.fxe"},
	{0x26888556A8BF0987ULL, "6033409_m.fxe"},
	{0x2688B6B9FF74AB86ULL, "character_character_p236.psh.fxo.dx11.assembly"},
	{0x2688DCCD415294CEULL, "gen00fl_bethany6121325.cl"},
	{0x2688F7BED7350099ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_3x.dds"},
	{0x26892711CB850E64ULL, "578655930.ogg"},
	{0x2689471900C6D9AAULL, "and100_anders6095981.cl"},
	{0x268964E2EDC4EC94ULL, "vfx_vfx_p9.psh.fxo.dx11"},
	{0x2689A2223C1A4500ULL, "gen00fl_aveline6120840.cl"},
	{0x2689DC79BDBBA35BULL, "vxw_bonepitdestroyed_a3_1ao_0_0.vlm"},
	{0x268B66CC09F80107ULL, "95664494.ogg"},
	{0x268B9C98295A19D0ULL, "dae000_meredith6092569.cl"},
	{0x268C45F295509B98ULL, "vxw_bonepitdestroyed_a3_26z_0_0.vlm"},
	{0x268CA013DB60D78DULL, "art\\characters\\creatures\\dragon\\animation\\ctrn90l.gad"},
	{0x268CEC4620A26392ULL, "hm_lgt_senestra.mor"},
	{0x268CEC67F41CC45FULL, "180840846.ogg"},
	{0x268D1928649AFF21ULL, "dae000_merchant_hm_mob6174225.cl"},
	{0x268D3B61E5C92138ULL, "6103118_m.fxe"},
	{0x268D728EB8ABD95FULL, "gen00fl_isabela_init6175652.cl"},
	{0x268D87B582698AB0ULL, "202167123.ogg"},
	{0x268D8A0E1055E716ULL, "mag210_arianni6024548.cl"},
	{0x268DBE994E7C1D0BULL, "lgt310_mis_burn_murder2.bnk"},
	{0x268EA11ECEDCF2C8ULL, "312687175.ogg"},
	{0x268EE6A5F1469CD5ULL, "vxlt_0_lowtown_a2_n_gvf_0_0.vlm"},
	{0x268F01BA470A1B5CULL, "vxw_sundermount_355_0_0.vlm"},
	{0x268F06CD7D5C7A63ULL, "art\\characters\\creatures\\stonegolem\\animation\\k_kb.evt"},
	{0x268F35253E18702AULL, "hlac_lt_0_lowtown_docks_-3_8z.dds"},
	{0x268F423B4E147251ULL, "llac_lt_hangedman_-1_17.dds"},
	{0x268FE8577A958112ULL, "437956466.ogg"},
	{0x269002ED5B507F0BULL, "6105586_m.fxe"},
	{0x26906F79345405F9ULL, "6034279_m.fxe"},
	{0x2690FF9F9779C1D7ULL, "ave310_brennan6137001.cl"},
	{0x269138E5463415B8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexcity_n.dds"},
	{0x269151C3DD25A283ULL, "266224423.ogg"},
	{0x26915E0F955B6D72ULL, "hlac_lt_0_lowtown_a2_2_1z.dds"},
	{0x269177FC7F4A4C7AULL, "one300dg_lia6064531.cl"},
	{0x2691C0994E8009EBULL, "lgt310_mis_burn_murder2.cnv"},
	{0x2691D08E41072B06ULL, "vxht_keep_5y1_0_0.vlm"},
	{0x2691FDE45B29AAB0ULL, "zz_isa_debug6134417.cl"},
	{0x269279F9FE0E94B7ULL, "6113790_m.fxe"},
	{0x2692AB99EB20B5F5ULL, "lt_0_lowtown_docks_a3_n_33_0.anb"},
	{0x2692B040DD818F6FULL, "619001790.ogg"},
	{0x2692DB2287528C73ULL, "follower_banter6149202.cl"},
	{0x26930DE25EEBF366ULL, "vxlt_warehouse_131_0_0.vlm"},
	{0x26935148968D3FB7ULL, "482684803.ogg"},
	{0x2693562F63643B6BULL, "16702681.ogg"},
	{0x26935A1A636D906FULL, "hlac_w_sundermount_196602_0x.dds"},
	{0x2694299667CBAD70ULL, "character_character_p201.psh.fxo.dx11.assembly"},
	{0x26945398C00EACEFULL, "vxlt_0_lowtown_a2_n_erj_0_0.vlm"},
	{0x269471C53C5224C6ULL, "vxht_keep_n_6lv_0_0.vlm"},
	{0x2695168F720EE33FULL, "545009725.ogg"},
	{0x2695890409460153ULL, "zz_dae_debug6142953.cl"},
	{0x26958F720CED2930ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_0y.dds"},
	{0x2695C37134C5BED5ULL, "vxht_mansion2b_1sp_0_0.vlm"},
	{0x2695DFCD7D6256E4ULL, "art\\characters\\creatures\\stonegolem\\animation\\k_kb.gad"},
	{0x2695F9A824A84749ULL, "vxlt_foundry_n_wt_0_0.vlm"},
	{0x269691A91CF08C32ULL, "vxw_bonepit_me_0_0.vlm"},
	{0x2697D1F12A02F8C0ULL, "qcr000_arishok6060093.cl"},
	{0x2698179A79DD5576ULL, "692928570.ogg"},
	{0x2698618A2A8144F2ULL, "6053852_f.fxe"},
	{0x26986355BB3B0E98ULL, "zz_followers_banter6159330.cl"},
	{0x2698CABC55235F3FULL, "crs_clo.mao"},
	{0x26992C5FCDE8919EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_darren.ani"},
	{0x26997F7DB57BECFBULL, "280537545.ogg"},
	{0x269A3993D79566FBULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_act_c.impa.ani"},
	{0x269A41440F3795E2ULL, "gen_im_trs_dem_rng_r0.uti"},
	{0x269AA8261497CDC8ULL, "606468550.ogg"},
	{0x269B6D74D6614DB9ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_017_0d.dds"},
	{0x269BB8440526BC8AULL, "rev_twi_a.mao"},
	{0x269BBAF8B28D31B7ULL, "l_pro000ar_blightlands_real.wbd"},
	{0x269CA4F9D6981AA5ULL, "zz_one_debug6099763.cl"},
	{0x269CEECD7D688EC2ULL, "art\\characters\\creatures\\stonegolem\\animation\\k_kb.ani"},
	{0x269D4637A549AA99ULL, "art\\characters\\playercharacter\\humanmale\\glv_masd.msh"},
	{0x269D5037A549BB87ULL, "art\\characters\\playercharacter\\humanmale\\glv_masd.mmh"},
	{0x269D5437A549C2D4ULL, "art\\characters\\playercharacter\\humanmale\\glv_masd.mao"},
	{0x269D564EE2BD1AB8ULL, "vxlt_gallowstemplar_455_0_0.vlm"},
	{0x269D9767B3F1D170ULL, "vxca_caverns1_mines_gje_0_0.vlm"},
	{0x269DDCCFD6C091E3ULL, "hlac_dr_0_deeproads_17038_4z.dds"},
	{0x269E3E7EAA5468F4ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm01.phy"},
	{0x269F951E10C3D6EEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1a.mmh"},
	{0x269F9719EAF6D68EULL, "dre000_vampire6034357.cl"},
	{0x269FA11E10C3EA85ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1a.mao"},
	{0x269FAF1E10C402DCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1a.msh"},
	{0x269FBDED1DB7C166ULL, "528797989.ogg"},
	{0x26A016A77BC7D269ULL, "vxlt_darktown_49x_0_0.vlm"},
	{0x26A1359BA765BCA2ULL, "vxht_0_hightown_n_c6m_0_0.vlm"},
	{0x26A17FBF50E20BE6ULL, "vxlt_warehouse_162_0_0.vlm"},
	{0x26A1B87FD296A7AFULL, "vxlt_blackemporium_nv_0_0.vlm"},
	{0x26A1C3A51F3668EAULL, "hlac_ca_caverns2_5802_35y.dds"},
	{0x26A1C3B5C44AF1E3ULL, "vxw_woundedcoast_t6_0_0.vlm"},
	{0x26A20C55FF43F6B5ULL, "vxht_brothel_n_5i6_0_0.vlm"},
	{0x26A221854E9ABA89ULL, "dae000ip_crowd_ff2_talk17.utp"},
	{0x26A227D8A12F5A55ULL, "m_qun210ar_javaris_encounter.bnk"},
	{0x26A237710EEC948EULL, "6140207_m.fxe"},
	{0x26A258684908392EULL, "vxht_mansion2b_3ek_0_0.vlm"},
	{0x26A2C575B204C56FULL, "vxlt_0_lowtown_a2_n_ewu_0_0.vlm"},
	{0x26A2C76BB39991CAULL, "841566101.ogg"},
	{0x26A2E3A22E483D46ULL, "vxht_keep_n_6gw_0_0.vlm"},
	{0x26A2F213FEAE3BE6ULL, "art\\levels\\races\\proxy\\_textures\\default_grey.dds"},
	{0x26A3D52277218AF5ULL, "mag100_karras6034720.cl"},
	{0x26A40DCFFB6BBA0DULL, "306429988.ogg"},
	{0x26A4D99C67E1012FULL, "6061239_m.fxe"},
	{0x26A4E2E74E7B1678ULL, "vxca_caverns2prc_43g_0_1.vlm"},
	{0x26A5678922D7D45CULL, "6163163_m.fxe"},
	{0x26A67B6403ECD35AULL, "vxlt_0_lowtown_docks_a3_n_d2t_0_0.vlm"},
	{0x26A6970361CAC40DULL, "vxlt_gallowscourtyard_4w8_0_0.vlm"},
	{0x26A6BD7F231B795FULL, "6069665_f.fxe"},
	{0x26A6C033A1A7D45BULL, "vxw_bonepitdestroyed_a3_1oa_0_0.vlm"},
	{0x26A70C7F5202F5FCULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_run.ani"},
	{0x26A70D8B914929FFULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\death.gad"},
	{0x26A738E52C0B7407ULL, "6101267_m.fxe"},
	{0x26A74CB0B1FA93B6ULL, "6146986_m.fxe"},
	{0x26A7868FE1148FD7ULL, "rdr111dg_harbormaster6016232.cl"},
	{0x26A7CF8F919FE47CULL, "6139667_m.fxe"},
	{0x26A7E305E0F810D5ULL, "mag320_nobleman_16060840.cl"},
	{0x26A81B7EAA5C81F1ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm01.msh"},
	{0x26A8357EAA5CADAFULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm01.mmh"},
	{0x26A84193D7A1B1F2ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_act_c.impa.evt"},
	{0x26A852AA81049B8BULL, "plt_zz_propt_debug.plo"},
	{0x26A9202E748D9CF2ULL, "follower_barks6153227.cl"},
	{0x26A9BBA44B02A875ULL, "6142310_m.fxe"},
	{0x26AB099156C96C8DULL, "static_static_p0.vsh.fxo"},
	{0x26ABC9260D04C814ULL, "6027331_m.fxe"},
	{0x26ABCA5FD532CDD5ULL, "vxlt_0_lowtown_a2_n_ghd_0_0.vlm"},
	{0x26ABFB177B903890ULL, "out_of_memory.dds"},
	{0x26AC33EF21B4DC57ULL, "vxht_chantry_30q_0_0.vlm"},
	{0x26AC691A176C84F1ULL, "vxw_woundedcoast_1e2_0_0.vlm"},
	{0x26ACD9012AABF40BULL, "dre000_dougal6131134.cl"},
	{0x26AD028B914D82ECULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\death.evt"},
	{0x26AD455FCDF96518ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_darren.gad"},
	{0x26AD780DB351B2E7ULL, "6939338.ogg"},
	{0x26ADF3EEA54DD09AULL, "zz_mer_debug.ncs"},
	{0x26AE197F52092A3EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_run.gad"},
	{0x26AE6093D7A65185ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_act_c.impa.gad"},
	{0x26AE6BEF807A9B98ULL, "hlac_ht_0_hightown_65535_0y.dds"},
	{0x26AE83AA8DA60B60ULL, "vxht_0_hightown_c9x_0_1.vlm"},
	{0x26AE93161A26982AULL, "6093486_m.fxe"},
	{0x26AEB2CC7E1B498DULL, "698546900.ogg"},
	{0x26AED9BDD0635E2CULL, "492471535.ogg"},
	{0x26AEFFE116ECB4DBULL, "ch1_ds2.psh.fxo"},
	{0x26AF75E912161520ULL, "vxht_0_hightown_a3_9wx_0_0.vlm"},
	{0x26AF91FA59C08BABULL, "zz_neel_test4.cut"},
	{0x26AFA778995BD522ULL, "vxht_0_hightown_n_c3n_0_0.vlm"},
	{0x26B07E32F13A0F35ULL, "vxht_brothel_n_5l5_0_0.vlm"},
	{0x26B0D0EB04B27C05ULL, "217191806.ogg"},
	{0x26B0DFAEB0230C4EULL, "991554847.ogg"},
	{0x26B0F545A45BD6DEULL, "mer120_hubert6017819.cl"},
	{0x26B1215274D3D21EULL, "w_woundedcoast_194_0.phy"},
	{0x26B13752A3FADDEFULL, "vxlt_0_lowtown_a2_n_ext_0_0.vlm"},
	{0x26B1401B55173487ULL, "905688795.ogg"},
	{0x26B16BC9EE841C81ULL, "vxw_sundermount_35d_0_0.vlm"},
	{0x26B1D25FF334E4A4ULL, "6074375_m.fxe"},
	{0x26B1E54827565E2DULL, "vxht_brothel_n_523_0_0.vlm"},
	{0x26B2A7F994D0362AULL, "art\\vfx\\combat\\vfx_glb_lit_imp_ext_c.impa.gad"},
	{0x26B3A244CFA40288ULL, "isa311_isabela6092494.cl"},
	{0x26B406EE08B152C5ULL, "6177215_m.fxe"},
	{0x26B4EDA3F6767AB5ULL, "gen_im_trs_gen_two_2hs_r0.uti"},
	{0x26B4F47F520F0A5DULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_run.evt"},
	{0x26B4F90482824506ULL, "6139926_m.fxe"},
	{0x26B534210FFC2696ULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef_scalpm1.mao"},
	{0x26B57A803FBC45DFULL, "plt_nat300pt_find_nathaniel.nss"},
	{0x26B5BF18E48813F1ULL, "art\\characters\\creatures\\desiredemon\\animation\\dazed.ani"},
	{0x26B64B6013DE73F5ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_stonetree.phy"},
	{0x26B6B12751AF7053ULL, "209346003.ogg"},
	{0x26B6BC6A823FE3B3ULL, "306323611.ogg"},
	{0x26B6BDE1D9547C28ULL, "one200dg_solivitus6060297.cl"},
	{0x26B7928F6AD3C2F7ULL, "6041494_f.fxe"},
	{0x26B7B1BF6AD0D963ULL, "dae021_fem_worshipper_sta6150354.cl"},
	{0x26B7CD323BDAFC56ULL, "w_bonepitdestroyed_a3_166_0.phy"},
	{0x26B850191D1FFAE4ULL, "vxlt_hangedman_n_1qz_0_0.vlm"},
	{0x26B860CB983BD52EULL, "gen00fl_sebastian6189423.cl"},
	{0x26B8BBC5B1106CEBULL, "373466554.ogg"},
	{0x26B8CAC9D208604DULL, "qun221_viscount6069382.cl"},
	{0x26B8E6826E4B261BULL, "mrl251cr_varterral.utc"},
	{0x26B90C938FF8D816ULL, "vxht_0_hightown_b7l_0_0.vlm"},
	{0x26B95DC7F2C47FA4ULL, "6131173_m.fxe"},
	{0x26B987A9B1C16969ULL, "6062827_m.fxe"},
	{0x26B98AF994D623C1ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_ext_c.impa.evt"},
	{0x26B98FA6276E842AULL, "vxlt_blackemporium_17q_0_0.vlm"},
	{0x26B9CBEED162863FULL, "103811843.ogg"},
	{0x26BA3B6C6339C374ULL, "6092365_f.fxe"},
	{0x26BA880AFBAF4632ULL, "gen00fl_carver6102580.cl"},
	{0x26BAA89F8C74ADAAULL, "vxw_encounter3_5b_0_0.vlm"},
	{0x26BAC62ED7FCFA3DULL, "dre000st_bethany.stg"},
	{0x26BAEA8B91599745ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\death.ani"},
	{0x26BB2B02DF90902EULL, "400372727.ogg"},
	{0x26BC9688A963AD71ULL, "6062690_m.fxe"},
	{0x26BCB25209789C0BULL, "dae000_orana6149049.cl"},
	{0x26BCCA18E48E4493ULL, "art\\characters\\creatures\\desiredemon\\animation\\dazed.gad"},
	{0x26BD0D312174ADDEULL, "vxht_0_hightown_a2_ban_0_0.vlm"},
	{0x26BDCE88510C6C43ULL, "vfx_chunkyparticle_p4.vsh.fxo.dx11"},
	{0x26BE19558B51EDA2ULL, "vxht_0_hightown_n_c8o_0_0.vlm"},
	{0x26BE27D27998B40FULL, "vxht_bartrand_n_63u_0_0.vlm"},
	{0x26BF4DD98B92E9E9ULL, "424561566.ogg"},
	{0x26BF5FE170F0E016ULL, "gen00fl_sebastian6180693.cl"},
	{0x26BFF9A8B9E86693ULL, "vxw_bonepitdestroyed_a3_n5_0_0.vlm"},
	{0x26C00F6C062CD0E2ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.cess.gad"},
	{0x26C03299B07E4F43ULL, "6120011_m.fxe"},
	{0x26C19F130CDA199CULL, "vxlt_0_lowtown_a3_n_c_ht5_0_0.vlm"},
	{0x26C1B4F5FB2333D6ULL, "6089755_m.fxe"},
	{0x26C1B865032EA346ULL, "dae000_pickpocket_scene6175093.cl"},
	{0x26C2584EF069D655ULL, "hlac_lt_0_lowtown_-1_4z.dds"},
	{0x26C2B6323BE4DCF7ULL, "w_bonepitdestroyed_a3_166_0.msh"},
	{0x26C2C8323BE4FB5AULL, "w_bonepitdestroyed_a3_166_0.mao"},
	{0x26C2CC323BE50269ULL, "w_bonepitdestroyed_a3_166_0.mmh"},
	{0x26C2FC1407B1B4D3ULL, "ts_raiders_01.gda"},
	{0x26C3AA418F79171FULL, "dae011_corff_sta6152744.cl"},
	{0x26C3D5F1BB86FF36ULL, "pro000_merchant6051813.cl"},
	{0x26C3E58CD31C6382ULL, "vxlt_blackemporium_u6_0_0.vlm"},
	{0x26C4486CEC40C33FULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzcoins.msh"},
	{0x26C44E6CEC40CDE1ULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzcoins.mmh"},
	{0x26C45A6CEC40E182ULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzcoins.mao"},
	{0x26C47582BC78EC4DULL, "vxca_cavern1_haunted_imc_0_0.vlm"},
	{0x26C4770CBB672C24ULL, "6168045_m.fxe"},
	{0x26C4DC29D0E4B737ULL, "6118510_f.fxe"},
	{0x26C5285C50319EF2ULL, "6034918_f.fxe"},
	{0x26C537F4FD3D0C6CULL, "vxlt_0_lowtown_n_d06_0_0.vlm"},
	{0x26C6326C063177D9ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.cess.evt"},
	{0x26C682F994E0A010ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_ext_c.impa.ani"},
	{0x26C6B8CB2C7013A1ULL, "vxw_woundedcoast_14m_0_0.vlm"},
	{0x26C764D90B2DB84EULL, "vxht_mansion2b_2vo_0_0.vlm"},
	{0x26C7674D9D0A4BDBULL, "w_bonepitdestroyed_a3_123_0.phy"},
	{0x26C77E7081EEF096ULL, "vxht_0_hightown_b0m_0_0.vlm"},
	{0x26C7C55329578754ULL, "art\\levels\\ztools\\default\\generic\\0dd_basic.phy"},
	{0x26C7E4DBDCFD2F06ULL, "6068866_m.fxe"},
	{0x26C837F40EB39033ULL, "486815050.ogg"},
	{0x26C8DD2E0D878B81ULL, "vxca_caverns1_gix_0_0.vlm"},
	{0x26C8ED97EB089E48ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgta_0n.dds"},
	{0x26C93937A56ECFBCULL, "art\\characters\\playercharacter\\humanmale\\glv_masd.phy"},
	{0x26C94CD3FB58B5F1ULL, "vxw_woundedcoast_1k0_0_0.vlm"},
	{0x26C98FB43A4E130BULL, "544392017.ogg"},
	{0x26C9EE484E205F95ULL, "27271000.ogg"},
	{0x26CA12B7AEA6D983ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_maso_0n.dds"},
	{0x26CA1FFBD9625401ULL, "3571041.ogg"},
	{0x26CA57E89D254FFAULL, "art\\characters\\playercharacter\\humanfemale\\glv_avea.mmh"},
	{0x26CA5BE89D255631ULL, "art\\characters\\playercharacter\\humanfemale\\glv_avea.mao"},
	{0x26CA69E89D256E40ULL, "art\\characters\\playercharacter\\humanfemale\\glv_avea.msh"},
	{0x26CA6D803FCE8BF0ULL, "plt_nat300pt_find_nathaniel.plo"},
	{0x26CA94765DD47B9FULL, "isa211st_isabela.stg"},
	{0x26CAD28247973F81ULL, "cd_qun221_viscount.wbd"},
	{0x26CBADD1AB5088D6ULL, "zz_followers_banter6157341.cl"},
	{0x26CBBF36A76AA6A8ULL, "6088761_m.fxe"},
	{0x26CD31922E94D3A5ULL, "dae300_vashothlead6062806.cl"},
	{0x26CD662D384B789EULL, "435151960.ogg"},
	{0x26CE4DD8A154D5DFULL, "m_qun210ar_javaris_encounter.wbd"},
	{0x26CE4EA3FDEE8FABULL, "vxca_cave_1o7_0_1.vlm"},
	{0x26CE63600843DE7EULL, "hlac_ca_caverns1_24424_7z.dds"},
	{0x26CE6B85ABDE7F13ULL, "vxw_bonepitdestroyed_a3_i4_0_0.vlm"},
	{0x26CF7E693D23C467ULL, "vxht_mansion2_1tg_0_0.vlm"},
	{0x26D039FFDDAFB3F9ULL, "art\\characters\\playercharacter\\qunarimale\\hed_mag.mao"},
	{0x26D045137C42EDE4ULL, "6139274_m.fxe"},
	{0x26D0A05737DFC74AULL, "character_character_p16.psh.fxo.dx11"},
	{0x26D0DE3F816F5995ULL, "gen00fl_sebastian6188670.cl"},
	{0x26D18CE874DA839FULL, "gen00fl_sebastian6179885.cl"},
	{0x26D1DA6B768E4A23ULL, "6082315_m.fxe"},
	{0x26D273B883C62C14ULL, "vxht_mansion2_2ov_0_0.vlm"},
	{0x26D28D5D054A5F30ULL, "gen00fl_fenris6103341.cl"},
	{0x26D2A25329615351ULL, "art\\levels\\ztools\\default\\generic\\0dd_basic.msh"},
	{0x26D2B05329616B6CULL, "art\\levels\\ztools\\default\\generic\\0dd_basic.mao"},
	{0x26D2BC5329617F0FULL, "art\\levels\\ztools\\default\\generic\\0dd_basic.mmh"},
	{0x26D2E75FAE6F04CDULL, "vxca_cavern1_haunted_ijb_0_0.vlm"},
	{0x26D31107331641BCULL, "427006375.ogg"},
	{0x26D345EB705B3D38ULL, "174818580.ogg"},
	{0x26D3A014445F58B8ULL, "6080682_m.fxe"},
	{0x26D43A6C063DC2C8ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.cess.ani"},
	{0x26D44C2AD605484AULL, "6097239_m.fxe"},
	{0x26D474F7C018D876ULL, "framebuffereffects_twopassblur_p0.psh.fxo"},
	{0x26D4E40DAC12ADF5ULL, "hlac_w_bonepit_327685_0z.dds"},
	{0x26D4F6126BC494ACULL, "vxlt_warehouse_n_116_0_0.vlm"},
	{0x26D51D2C10ECBC63ULL, "791906317.ogg"},
	{0x26D5625F94161D51ULL, "zz_mcr_debug6080864.cl"},
	{0x26D5C96D3BA2EC26ULL, "hlac_dr_primevaltomb_657_0z.dds"},
	{0x26D5EB35445821E7ULL, "6139192_m.fxe"},
	{0x26D625942B66B6F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_shffl.ani"},
	{0x26D6335647D4B442ULL, "one280_merin6105448.cl"},
	{0x26D77C79E37C1808ULL, "287298352.ogg"},
	{0x26D785DAB7F02A39ULL, "509494314.ogg"},
	{0x26D79431E861C905ULL, "6088849_m.fxe"},
	{0x26D7F0338997255FULL, "vxlt_undercitydungeon_1l7_0_0.vlm"},
	{0x26D7F8F565D66E9BULL, "6042550_f.fxe"},
	{0x26D805EEB437BAFDULL, "540877392.ogg"},
	{0x26D861B5C76B421BULL, "character_character_p68.psh.fxo.dx11.assembly"},
	{0x26D8686A091115A7ULL, "621374488.ogg"},
	{0x26D8CFFEA2626E5BULL, "lt_0_lowtown_a2_42_0_lc.mmh"},
	{0x26D9A7F64BAA1311ULL, "250748548.ogg"},
	{0x26D9C1AD7D121191ULL, "vxlt_gallowscourtyard_n_4w7_0_0.vlm"},
	{0x26D9E0EEA572EBFDULL, "zz_mer_debug.cnv"},
	{0x26D9FEB00857B87AULL, "hlac_lt_gallowscourtyard_n_-65535_0z.dds"},
	{0x26DA7492FA80AE1AULL, "gen00fl_isabela6099422.cl"},
	{0x26DA92EF108FE525ULL, "vxw_sundermount_31d_0_0.vlm"},
	{0x26DACCFDDFDB3F9FULL, "art\\characters\\creatures\\gibgolem\\bdy_gib.mao"},
	{0x26DB0B8C5D84E50FULL, "vxht_bartrand_n_69w_0_0.vlm"},
	{0x26DC83DD3B55AEA5ULL, "hlac_ca_caverns2prc_5802_19y.dds"},
	{0x26DC8C3AF499FDD7ULL, "vxlt_blackemporium_6f_0_0.vlm"},
	{0x26DCAE921ACCBB78ULL, "180686900.ogg"},
	{0x26DD00942B6C963BULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_shffl.gad"},
	{0x26DD1EC659FC29CEULL, "vxht_bartrand_n_638_0_0.vlm"},
	{0x26DD45B8A76D0250ULL, "fx_ice_crust.mao"},
	{0x26DD8141D8EBC74FULL, "dae000_hate_follower_react6187870.cl"},
	{0x26DE1943BA9269A7ULL, "hlac_ht_mansion2_818_22y.dds"},
	{0x26DE231273D94FB5ULL, "art\\vfx\\materials\\vfx_smo_thick_threshold.mao"},
	{0x26DE81AFECEF95A9ULL, "vxlt_hangedman_n_21p_0_0.vlm"},
	{0x26DEEC2C4E06863BULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_scra_0d.dds"},
	{0x26DF5A24B62E7705ULL, "zz_pose_squat_test.cut"},
	{0x26DFC2DEEE7D1BD6ULL, "6028141_m.fxe"},
	{0x26DFCDA33EDF5F10ULL, "vxlt_warehouse_n_rp_0_0.vlm"},
	{0x26E06DA9C136BE56ULL, "915035067.ogg"},
	{0x26E087E5E0E8616FULL, "w_bonepitdestroyed_a3_73_0.mao"},
	{0x26E093E5E0E8750CULL, "w_bonepitdestroyed_a3_73_0.mmh"},
	{0x26E099E5E0E87FFEULL, "w_bonepitdestroyed_a3_73_0.msh"},
	{0x26E1532693DA0AA5ULL, "l_rdr301ar_caverns.wbd"},
	{0x26E1A840678FF856ULL, "vxlt_0_lowtown_a3_n_c_hsj_0_0.vlm"},
	{0x26E2480CAEE0D4BEULL, "6087242_f.fxe"},
	{0x26E2E9091BEF112DULL, "6149395_m.fxe"},
	{0x26E3485E45C898ABULL, "mrl310_merrill6120359.cl"},
	{0x26E405D1BDDEB93FULL, "isa311_castillon6117916.cl"},
	{0x26E41D942B72E620ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_shffl.evt"},
	{0x26E45F3DD69093AAULL, "vxlt_0_lowtown_a2_eua_0_0.vlm"},
	{0x26E487FCC4701B16ULL, "lt_0_lowtown_docks_199_0.mao"},
	{0x26E493FCC4702FB5ULL, "lt_0_lowtown_docks_199_0.mmh"},
	{0x26E495FCC470332BULL, "lt_0_lowtown_docks_199_0.msh"},
	{0x26E499630CB801EBULL, "hlac_dr_slavers_3215_0x.dds"},
	{0x26E4B1DAF8613DFAULL, "hlac_w_woundedcoast_65536_0x.dds"},
	{0x26E4F73F8FF72330ULL, "6172550_m.fxe"},
	{0x26E5524AE339C5C4ULL, "325671925.ogg"},
	{0x26E5B6FD08E03506ULL, "675450867.ogg"},
	{0x26E5BBA5CB2EDFFFULL, "6039404_f.fxe"},
	{0x26E5F747101B57FEULL, "arm_cot.tnt"},
	{0x26E6014B34D61978ULL, "188215948.ogg"},
	{0x26E6261CA2545DF9ULL, "hlac_ht_0_hightown_-65536_2x.dds"},
	{0x26E62A1A7B7C6A18ULL, "rdr111st_assistant.stg"},
	{0x26E6308DDF44E6F1ULL, "vxw_woundedcoast_1q6_0_0.vlm"},
	{0x26E65C6AD118347DULL, "586848662.ogg"},
	{0x26E6982E63862C2EULL, "858402645.ogg"},
	{0x26E6B040DEA787B9ULL, "hlac_ht_0_hightown_a3_-65537_0z.dds"},
	{0x26E7BA71CEA6AAFCULL, "vxlt_0_lowtown_a2_n_c_hnf_0_0.vlm"},
	{0x26E7C756C8C2549EULL, "zz_cave_accolades_2_cinetes.cut"},
	{0x26E82F6FCBF40616ULL, "fxa_power_dirty.dds"},
	{0x26E8326C01553890ULL, "6103631_m.fxe"},
	{0x26E835A24932511AULL, "mag100_karras6053498.cl"},
	{0x26E86936F318E9A5ULL, "fen210_hadriana6110453.cl"},
	{0x26E93488D57BB7D5ULL, "6075126_m.fxe"},
	{0x26E94D93E70F9D72ULL, "6053873_m.fxe"},
	{0x26E9C9D8ABB92760ULL, "vxca_caverns1_mines_cml_0_0.vlm"},
	{0x26E9EB5B3785D94FULL, "593234623.ogg"},
	{0x26EABF0B6EDEA9EFULL, "vxlt_gallowstemplar_f_34y_0_0.vlm"},
	{0x26EB704E310B4BF5ULL, "vxdr_0_deeproads_cus_0_0.vlm"},
	{0x26EBF688BC334993ULL, "vxlt_0_lowtown_docks_a3_n_bez_0_0.vlm"},
	{0x26EC4E1D30B5B991ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel_1.mao"},
	{0x26EC57A3085D356DULL, "6176528_m.fxe"},
	{0x26EC5A4EA5754144ULL, "682253781.ogg"},
	{0x26EC604F8231EADDULL, "art\\levels\\ztools\\default\\generic\\3ow_basic.phy"},
	{0x26EC771BF073C7E5ULL, "6072084_m.fxe"},
	{0x26EC9790EA20BC4EULL, "plt_lgt110pt_urn_miracle.plo"},
	{0x26ECD771D8AF1B5BULL, "vxdr_primevaltomb_8ra_0_0.vlm"},
	{0x26ECE1EA3B67B93CULL, "static1_punch_ds1.psh.fxo"},
	{0x26EDA12E5B3E675DULL, "226277818.ogg"},
	{0x26EDADE740CE8A77ULL, "fen210_hadriana6110600.cl"},
	{0x26EDF522D7EE8365ULL, "cinedesign_global_events.wbd"},
	{0x26EE3F8030D57790ULL, "vxlt_warehouse_n_yq_0_0.vlm"},
	{0x26EE66FDEB454F40ULL, "6082143_m.fxe"},
	{0x26EEDF5E3FE88945ULL, "292554769.ogg"},
	{0x26EF577267B25D14ULL, "vxht_mansion2_2ut_0_0.vlm"},
	{0x26EF76E89D44A005ULL, "art\\characters\\playercharacter\\humanfemale\\glv_avea.phy"},
	{0x26EFDE8E42BD7D9DULL, "6129725_f.fxe"},
	{0x26F01A1D598610D6ULL, "vxlt_0_lowtown_a3_n_c_hxk_0_0.vlm"},
	{0x26F04EC3DFB2519CULL, "747940137.ogg"},
	{0x26F04F987120B4C5ULL, "gen_im_npc_arm_hum_cht_tmpa.uti"},
	{0x26F07F190587868DULL, "vxlt_0_lowtown_a2_fe9_0_0.vlm"},
	{0x26F0F48247B84883ULL, "cd_qun221_viscount.bnk"},
	{0x26F0F4F77AF00708ULL, "384851250.ogg"},
	{0x26F13943ED1B967CULL, "258960237.ogg"},
	{0x26F1410C1C2E2C32ULL, "hightown_market_guard_sta6169589.cl"},
	{0x26F143C18770EE0FULL, "mag100ip_bones_1.utp"},
	{0x26F1676847FDE82EULL, "6172776_m.fxe"},
	{0x26F2053CE5525BE5ULL, "dae000_keep_maid_sta6162412.cl"},
	{0x26F21420C6590E69ULL, "vxlt_warehouse_n_cz_0_0.vlm"},
	{0x26F2474EBBAB846BULL, "zz_isa_debug6133219.cl"},
	{0x26F25EFFDF6512D9ULL, "zz_followers_banter6157722.cl"},
	{0x26F260BAA05D2BB2ULL, "one220_yevhen6135791.cl"},
	{0x26F2ED349208CE90ULL, "vxlt_darktown_4uy_0_0.vlm"},
	{0x26F2F3BBF03589B3ULL, "6028517_m.fxe"},
	{0x26F34164DC7EBE48ULL, "114954941.ogg"},
	{0x26F34E4D9D2F5C7AULL, "w_bonepitdestroyed_a3_123_0.msh"},
	{0x26F35C4D9D2F73B3ULL, "w_bonepitdestroyed_a3_123_0.mao"},
	{0x26F3604D9D2F7AC0ULL, "w_bonepitdestroyed_a3_123_0.mmh"},
	{0x26F38CF847852874ULL, "338055545.ogg"},
	{0x26F3DC923E4A3E6FULL, "vxlt_0_lowtown_docks_n_bb0_0_0.vlm"},
	{0x26F455DC75BB4852ULL, "6172871_m.fxe"},
	{0x26F52247BC864749ULL, "lgt320_mis_resp_swordhunter6150275.cl"},
	{0x26F567F0A139D81EULL, "450038239.ogg"},
	{0x26F57C76A98E1B96ULL, "hlac_w_encounter2_camp_196613_0x.dds"},
	{0x26F59577A457C9C3ULL, "follower_banter6148450.cl"},
	{0x26F5BCC0C161A8F2ULL, "vxlt_0_lowtown_a2_n_f5e_0_0.vlm"},
	{0x26F5E950823E5A4CULL, "hlac_lt_0_lowtown_docks_-1_0z.dds"},
	{0x26F678B74ABE1938ULL, "dre000_vampire6034399.cl"},
	{0x26F6BF4DB7EDE499ULL, "351906391.ogg"},
	{0x26F7167CEA36BEE7ULL, "plt_cod_crt_revenant.plo"},
	{0x26F7A41133513E6DULL, "hlac_ht_keep_n_3732_6x.dds"},
	{0x26F7B7E7460FA80AULL, "pro000_aveline.cnv"},
	{0x26F7EBF64FB5F061ULL, "zz_rooftop.cut"},
	{0x26F875DACE239CC1ULL, "511025843.ogg"},
	{0x26F9258A9371E183ULL, "933911144.ogg"},
	{0x26F9575CF8B09F72ULL, "469703319.ogg"},
	{0x26F96A3DC5543136ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta.msh"},
	{0x26F9743DC5544234ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta.mmh"},
	{0x26F9E22B23016475ULL, "vxdr_0_deeproads_crr_0_0.vlm"},
	{0x26F9F7E22519960DULL, "dre000_dougal6130962.cl"},
	{0x26FABB4D89ED78F7ULL, "6036554_m.fxe"},
	{0x26FB0DEFA8B10FEAULL, "327622734.ogg"},
	{0x26FB99E746135E32ULL, "pro000_aveline.bnk"},
	{0x26FC44C15A6FED84ULL, "hm_spotter.mor"},
	{0x26FCD8DADD3B67C0ULL, "6085043_m.fxe"},
	{0x26FD14E01CFF7FC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_left.ani"},
	{0x26FD554F82404B49ULL, "art\\levels\\ztools\\default\\generic\\3ow_basic.mao"},
	{0x26FD5E60141A9610ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_stonetree.msh"},
	{0x26FD614F82405FE2ULL, "art\\levels\\ztools\\default\\generic\\3ow_basic.mmh"},
	{0x26FD634F82406298ULL, "art\\levels\\ztools\\default\\generic\\3ow_basic.msh"},
	{0x26FD6C60141AAE2AULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_stonetree.mmh"},
	{0x26FD7060141AB561ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_stonetree.mao"},
	{0x26FDBC8778976038ULL, "vxht_brothel_n_5jj_0_0.vlm"},
	{0x26FEA4A3FBBC0B33ULL, "dae000_mother6121070.cl"},
	{0x26FEC410E890A49EULL, "da2terrain_rendertodepth_p1.vsh.fxo.dx11"},
	{0x26FEDE3E8B8B177AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_wsl.mmh"},
	{0x26FEF03E8B8B35C0ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_wsl.msh"},
	{0x270021AEE6E75DECULL, "196430898.ogg"},
	{0x270051421A5FA14AULL, "513216305.ogg"},
	{0x270057AD0EED4883ULL, "vxlt_0_lowtown_a2_n_c_f4q_0_0.vlm"},
	{0x270080CB04C3CB5EULL, "hlac_lt_0_lowtown_docks_n_0_5x.dds"},
	{0x27008890EA314BDDULL, "plt_lgt110pt_urn_miracle.nss"},
	{0x27012C4CD3061A4EULL, "vxht_mansion2b_2bc_0_0.vlm"},
	{0x2701599E865FB170ULL, "llac_lt_0_lowtown_a3_n_c_-4_0.dds"},
	{0x27016561A400FFC7ULL, "vxlt_0_lowtown_n_gcc_0_0.vlm"},
	{0x270180B04801D660ULL, "6119416_m.fxe"},
	{0x2701DC0227DF13A6ULL, "vxht_keep_n_5gp_0_0.vlm"},
	{0x270223549D4D04E0ULL, "hlac_w_encounter2_-131072_0y.dds"},
	{0x270248C5596290DEULL, "vxlt_0_lowtown_a2_hfa_0_0.vlm"},
	{0x270254551476B153ULL, "zz_followers_banter6157658.cl"},
	{0x2703448B2CE4B103ULL, "vxlt_0_lowtown_a2_n_c_eu9_0_0.vlm"},
	{0x2703C3D0185CAB2FULL, "lgt300cr_gang_1_ht.utc"},
	{0x270411E01D059926ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_left.gad"},
	{0x27043D6D87F02FCFULL, "gen00fl_aveline6139977.cl"},
	{0x270447C3B76C72AFULL, "532462966.ogg"},
	{0x2704844FEAEE3C05ULL, "gen00fl_varric6077823.cl"},
	{0x27049E2BB292DBFCULL, "vxlt_0_lowtown_a2_n_c_hdd_0_0.vlm"},
	{0x2705A3F71B0D9EFAULL, "fen110_fenris6097058.cl"},
	{0x2706397A66286FD4ULL, "532276052.ogg"},
	{0x27065AA284BF4BC6ULL, "vxlt_gallowstemplar_390_0_0.vlm"},
	{0x270697A557BF9A34ULL, "hlac_lt_0_lowtown_a2_n_0_0x.dds"},
	{0x27078ACEA851E93EULL, "vxw_bonepitdestroyed_a3_220_0_0.vlm"},
	{0x2707C8ABB824F4A6ULL, "vxdr_slavers_1ap_0_0.vlm"},
	{0x270827C687523723ULL, "vxht_0_hightown_a2_akb_0_0.vlm"},
	{0x27084D65172E4467ULL, "6132716_f.fxe"},
	{0x2708667B3576F6D5ULL, "vxlt_0_lowtown_docks_n_c5f_0_0.vlm"},
	{0x270869CD41D726A9ULL, "vxht_bartrand_n_4km_0_0.vlm"},
	{0x2708AF5F96F6D876ULL, "lla_ht_0_hightown_-131073_0.dds"},
	{0x270945DD04FC2667ULL, "vxht_mansion2_1hc_0_0.vlm"},
	{0x270986136772AA76ULL, "gen00fl_aveline6120860.cl"},
	{0x2709BE3EC792D083ULL, "gen00fl_merrill6108433.cl"},
	{0x270AB91D7D505D44ULL, "6111315_m.fxe"},
	{0x270B0BB94B995521ULL, "vxht_0_hightown_n_c4e_0_1.vlm"},
	{0x270B0CE01D0BAF05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_left.evt"},
	{0x270C2E646A8D78B8ULL, "vxht_brothel_n_5mk_0_0.vlm"},
	{0x270C451A669C94E2ULL, "gen_im_acc_blt_act3_08.uti"},
	{0x270D590E15AB9F6DULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_002a_0s.dds"},
	{0x270E5BE92052924FULL, "vrc321ip_golem_proj_target.utp"},
	{0x270E69AA815B7658ULL, "plt_zz_propt_debug.nss"},
	{0x270E785FAC5FDCF4ULL, "6133878_m.fxe"},
	{0x270ED33F81B90ACEULL, "349864059.ogg"},
	{0x270EDC6283929D0CULL, "ring.mao"},
	{0x270EE6D9C371DDE0ULL, "vxht_0_hightown_a6v_0_0.vlm"},
	{0x270F297A4AFE7AA4ULL, "338580557.ogg"},
	{0x27101F6990728465ULL, "346126223.ogg"},
	{0x2710470D75B29A36ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.inactive.ani"},
	{0x2710BAA24B58A95EULL, "vxlt_0_lowtown_a2_hcb_0_0.vlm"},
	{0x271189EF27D88958ULL, "348497375.ogg"},
	{0x2711BA037AD98504ULL, "vxlt_darktown_5j5_0_0.vlm"},
	{0x2711BB6407E9FE73ULL, "vxht_0_hightown_a3_c35_0_0.vlm"},
	{0x2711FB03EACF9E1BULL, "vxdr_0_deeproads_cvx_0_0.vlm"},
	{0x27126935EB31A700ULL, "112275983.ogg"},
	{0x2712934970321199ULL, "gen00fl_fenris6108057.cl"},
	{0x2712B86CE009FDA9ULL, "6173383_m.fxe"},
	{0x271335F64CB8F1DDULL, "hlac_ht_playershouse_n_855_25y.dds"},
	{0x2713984D84336559ULL, "38273653.ogg"},
	{0x27139FCB90B0FA56ULL, "6101433_m.fxe"},
	{0x2713AFF9E9DAF462ULL, "mcr310_argument6074524.cl"},
	{0x2713B7E5E30CFC13ULL, "511776246.ogg"},
	{0x27140418C96BD618ULL, "73188515.ogg"},
	{0x271410AA934D564CULL, "6134822_m.fxe"},
	{0x27142774CE25D20DULL, "primitives_primitive_p1.vsh.fxo.dx11.assembly"},
	{0x2714CED667958B00ULL, "1036044770.ogg"},
	{0x271536A319A007ACULL, "vxht_0_hightown_a2_9p0_0_0.vlm"},
	{0x2715F54FCFC9D45AULL, "follower_banter6172953.cl"},
	{0x2716B65874391DACULL, "hlac_lt_darktown_-1_51z.dds"},
	{0x2716C568896AC502ULL, "vxw_woundedcoast_w_0_0.vlm"},
	{0x2716D858276D0F55ULL, "vxlt_0_lowtown_docks_n_c0e_0_0.vlm"},
	{0x271760FF8598D39EULL, "76158980.ogg"},
	{0x27179F9953E69D63ULL, "vxca_cavern1_haunted_jfw_0_0.vlm"},
	{0x2717F51898271E27ULL, "6013418_m.fxe"},
	{0x27181BAD28F76DB2ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallceltic_0.pwk"},
	{0x27182954AEE90D7CULL, "lt_0_lowtown_docks_n_190_0.mao"},
	{0x27183554AEE9211FULL, "lt_0_lowtown_docks_n_190_0.mmh"},
	{0x27183B54AEE92BC1ULL, "lt_0_lowtown_docks_n_190_0.msh"},
	{0x271879C91E742278ULL, "vxlt_0_lowtown_docks_n_b5j_0_0.vlm"},
	{0x2718A1BC52DF7A13ULL, "134366095.ogg"},
	{0x2718E8DC5E88E1B3ULL, "scaleform_textcolormultiply.mao"},
	{0x2719937177E6F98AULL, "165015143.ogg"},
	{0x2719AD0EA3AD3CFEULL, "llac_lt_0_lowtown_65537_7.dds"},
	{0x271A9641E98B0181ULL, "vxlt_darktown_2qr_0_0.vlm"},
	{0x271AE4A2BB060019ULL, "hlac_ht_0_hightown_a2_131071_0y.dds"},
	{0x271B5E23570977FDULL, "qun211cr_merc_loot.ncs"},
	{0x271C2A5ABE31F82DULL, "gen_im_trs_crc_junk1_01.uti"},
	{0x271C2B18258B4CD0ULL, "6041573_m.fxe"},
	{0x271CA0B83780018EULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv1.evt"},
	{0x271D10ACAF5D40A3ULL, "w_bonepitdestroyed_a3_251_0.mao"},
	{0x271D14ACAF5D47F0ULL, "w_bonepitdestroyed_a3_251_0.mmh"},
	{0x271D22ACAF5D5F0AULL, "w_bonepitdestroyed_a3_251_0.msh"},
	{0x271D2520B6CB7ADAULL, "6136325_m.fxe"},
	{0x271D29B4C9CCB765ULL, "145995651.ogg"},
	{0x271D2BF392AB452FULL, "clf260_harley6069960.cl"},
	{0x271DFB1F84A16AC8ULL, "lit_up_04.mao"},
	{0x271E1D8870B78B3EULL, "6153430_m.fxe"},
	{0x271E2CD4AFC540B8ULL, "gen00fl_merrill6127228.cl"},
	{0x271EBEB575D15B90ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_deadmeredith_sa.dds"},
	{0x271F2C7F3D4EC1DEULL, "vxlt_0_lowtown_a2_hhc_0_0.vlm"},
	{0x271F836539385581ULL, "t1_tat_drd.tnt"},
	{0x27202D40F9E016F3ULL, "vxht_0_hightown_a3_c44_0_0.vlm"},
	{0x2720427B1CD149DEULL, "hlac_w_encounter2_camp_131072_0z.dds"},
	{0x2720509415BA4458ULL, "w_encounter3_99_0.phy"},
	{0x2720A38A7B445FE6ULL, "vxlt_backalley_g_1n0_0_1.vlm"},
	{0x2720D39F5E8BC575ULL, "6142101_m.fxe"},
	{0x27212951DDAC1180ULL, "gen00fl_carver6103400.cl"},
	{0x2721455E8CBDD865ULL, "water_water_p0.psh.fxo"},
	{0x2721CFB0FCB958FAULL, "233173076.ogg"},
	{0x27222D269411614FULL, "l_rdr301ar_caverns.bnk"},
	{0x272397B837861139ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv1.gad"},
	{0x27239D3D62E703ECULL, "vxlt_undercityencounter_wc_0_0.vlm"},
	{0x2723AB0864A1FCA4ULL, "vxht_brothel_n_4hu_0_0.vlm"},
	{0x2723FD3E8BAA6785ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_wsl.phy"},
	{0x2724700D75C38970ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.inactive.gad"},
	{0x27247886381A3A8EULL, "73890960.ogg"},
	{0x2724A07D60CCA5B3ULL, "vxht_mansion2b_2ka_0_0.vlm"},
	{0x27250F93000CE092ULL, "106972407.ogg"},
	{0x27253440B942D3ACULL, "224728388.ogg"},
	{0x2725C0B8B98FE619ULL, "69707422.ogg"},
	{0x272760A8A7E77E86ULL, "gen00fl_merrill6107804.cl"},
	{0x27283D844343C303ULL, "hlac_w_bonepit_2_0y.dds"},
	{0x2728610B3176E644ULL, "hlac_lt_0_lowtown_1_6z.dds"},
	{0x272885287F37595EULL, "t3_shd_gld.tnt"},
	{0x272891E4EEC02079ULL, "one100dg_kelder6055421.cl"},
	{0x2728BD4335095438ULL, "448012887.ogg"},
	{0x272901D5CFE63E59ULL, "6071837_f.fxe"},
	{0x2729081EDB811A01ULL, "vxlt_darktown_2tq_0_0.vlm"},
	{0x27294DAF1CBDF2FAULL, "6110224_m.fxe"},
	{0x2729C92F6F38087EULL, "6089961_f.fxe"},
	{0x272AA02536EAC5A0ULL, "gen00fl_aveline6137686.cl"},
	{0x272AA8B8378C4C87ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv1.ani"},
	{0x272AF056346ED47BULL, "vfx_vfx_p2.vsh.fxo.dx11"},
	{0x272C98DE2378780CULL, "vxlt_0_lowtown_a2_n_c_fdz_0_0.vlm"},
	{0x272CF0F2FAF2D213ULL, "hlac_ca_caverns1_24423_1x.dds"},
	{0x272D015DE092372EULL, "229126255.ogg"},
	{0x272D7A59DC54172FULL, "6154733_m.fxe"},
	{0x272DCF22D824260FULL, "cinedesign_global_events.bnk"},
	{0x272DDD74DFFAB274ULL, "qcr270_fake_meredith6066345.cl"},
	{0x272E7E5B5E7E7FD3ULL, "vxca_caverns1_mines_ii1_0_0.vlm"},
	{0x272EF963268CF12AULL, "vxht_keep_n_6bt_0_0.vlm"},
	{0x272F5DE24EEA9382ULL, "6072560_m.fxe"},
	{0x27301D2AB4FC8B01ULL, "6103784_m.fxe"},
	{0x27303F80B59E17CCULL, "256411540.ogg"},
	{0x27309A83D6EB7DF9ULL, "vxlt_0_lowtown_a2_eqh_0_2.vlm"},
	{0x2730B140886BAE1AULL, "6080723_m.fxe"},
	{0x2730CCC879890273ULL, "6208297_f.fxe"},
	{0x2731214C0DAAC7EAULL, "572961301.ogg"},
	{0x2731BD0DC9857EE3ULL, "251795361.ogg"},
	{0x2731F306E9E8CEEAULL, "hlac_w_blightlands_327681_0x.dds"},
	{0x2732E7584CF2CD8BULL, "6148354_m.fxe"},
	{0x2733056EF416FFD5ULL, "351694902.ogg"},
	{0x27331838EDC81BBAULL, "977510340.ogg"},
	{0x27334056436F969FULL, "hlac_lt_gallowscourtyard_-65536_0y.dds"},
	{0x27336ACB802BF415ULL, "154164077.ogg"},
	{0x2733940A95AD9619ULL, "vxca_cavern1_haunted_jl7_0_0.vlm"},
	{0x2733E2E310B01F8EULL, "6121916_m.fxe"},
	{0x273438E3ED489634ULL, "hlac_lt_foundry_n_320_3x.dds"},
	{0x27347D55E0DF77FAULL, "ca_caverns1_mines_24426.rml"},
	{0x27349F5A4A64C669ULL, "follower_banter6163200.cl"},
	{0x2735436A642167FCULL, "vxlt_hangedman_n_20j_0_0.vlm"},
	{0x273583122E8C0ADEULL, "art\\vfx\\combat\\vfx_r_asn_as_arch_act_c.impa.ani"},
	{0x2735ED2593D950C3ULL, "6016624_m.fxe"},
	{0x273627A8A86009DAULL, "vxht_0_hightown_a3_7vk_0_0.vlm"},
	{0x273711F1945F232DULL, "554486522.ogg"},
	{0x2737466BCD7F794DULL, "follower_banter6155421.cl"},
	{0x273752ADF5CAE5C4ULL, "gen_im_trs_gen_sta_nat_r1.uti"},
	{0x27376A99B6628BF2ULL, "748751235.ogg"},
	{0x273799EAE669174CULL, "w_encounter2_40_0.msh"},
	{0x27379FEAE669213EULL, "w_encounter2_40_0.mmh"},
	{0x2737ABEAE66935D5ULL, "w_encounter2_40_0.mao"},
	{0x2737FF111FD31E52ULL, "vxca_caverns1_mines_ii4_0_0.vlm"},
	{0x2738B0CE3554CDE7ULL, "dae270ar_gallows.ncs"},
	{0x27398852F65534C7ULL, "vxca_caverns2prc_7uh_0_0.vlm"},
	{0x2739AFA79E131A1BULL, "vxht_chantry_2dx_0_0.vlm"},
	{0x273A508E130B12C1ULL, "34806255.ogg"},
	{0x273A8AB32A8B519BULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5.mmh"},
	{0x273A936D50711213ULL, "hlac_lt_0_lowtown_docks_n_131070_0z.dds"},
	{0x273A9D734A931BE9ULL, "llac_lt_darktown_-1_54.dds"},
	{0x273AF47F98706F0DULL, "318590591.ogg"},
	{0x273B00AE919DFAD0ULL, "dae000_mother6103062.cl"},
	{0x273B66E419B5C6ECULL, "314456933.ogg"},
	{0x273B8AE4C97F1BF4ULL, "36431122.ogg"},
	{0x273C5462808C7251ULL, "mcr310_loyalist_templars6076904.cl"},
	{0x273CB801548C1D4DULL, "hlac_lt_0_lowtown_131072_2z.dds"},
	{0x273D24DF5467087AULL, "art\\characters\\weapons\\maces\\textures\\w_mce_001a_0n.dds"},
	{0x273D52FC45BB0DDCULL, "367051754.ogg"},
	{0x273D5F459D95D6D2ULL, "gen_im_npc_dwf_arm_cht_cxt5.uti"},
	{0x273DDF5498360318ULL, "nat320_delilah6065904.cl"},
	{0x273DF10ACFAC2262ULL, "497216656.ogg"},
	{0x273E3B638FA293BEULL, "845862300.ogg"},
	{0x273E659F7A6B3DFCULL, "vxlt_0_lowtown_a2_n_c_hpx_0_0.vlm"},
	{0x273E73D3B23BE957ULL, "vxw_bonepitmining_a2_74_0_0.vlm"},
	{0x273E8700CF6F9F1CULL, "gen00fl_merrill6127512.cl"},
	{0x273ED0884D10D879ULL, "w_sundermount_202_0.msh"},
	{0x273EDA884D10E967ULL, "w_sundermount_202_0.mmh"},
	{0x273EDE884D10EFB4ULL, "w_sundermount_202_0.mao"},
	{0x273F1212F94E9A6BULL, "hlac_lt_0_lowtown_a3_n_c_65536_7z.dds"},
	{0x273F908E55EE52A1ULL, "6042292_m.fxe"},
	{0x273F90CE355AB596ULL, "dae270ar_gallows.lst"},
	{0x273FE3A8607D5703ULL, "6098284_m.fxe"},
	{0x274003C8AF480091ULL, "vxht_mansion2_1st_0_0.vlm"},
	{0x2740DABB606D722CULL, "vxlt_0_lowtown_a3_n_c_erf_0_0.vlm"},
	{0x2741408B70CDC0ADULL, "879329827.ogg"},
	{0x2741529425C89E47ULL, "vxht_brothel_n_2br_0_0.vlm"},
	{0x27416CB4902AF584ULL, "and330st_base_justice_start.stg"},
	{0x2741843744B8D6B3ULL, "vxht_mansion2b_2ac_0_0.vlm"},
	{0x274247C744C9EDB1ULL, "1008314282.ogg"},
	{0x27429B122E96BDAFULL, "art\\vfx\\combat\\vfx_r_asn_as_arch_act_c.impa.evt"},
	{0x2742ADEA93202DEAULL, "lt_gallowscourtyard_n_57_0.rml"},
	{0x2742C489CB17CE33ULL, "vrc321_party_barks6095744.cl"},
	{0x27431D8E3679F2B4ULL, "418521441.ogg"},
	{0x274321E393EF8AD0ULL, "497982500.ogg"},
	{0x274326ECD1A9937DULL, "372540006.ogg"},
	{0x27433722788A90D6ULL, "hlac_lt_0_lowtown_a2_n_65536_10z.dds"},
	{0x27435A1B04D0C652ULL, "11616403.ogg"},
	{0x2743659B44A02412ULL, "dae300_bran6143485.cl"},
	{0x27438F05400312B9ULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation4.ani"},
	{0x2743C65C2C8254D5ULL, "180678913.ogg"},
	{0x2743EA2C313F8153ULL, "areadata_one.gda"},
	{0x27469A6349CDC94DULL, "490291158.ogg"},
	{0x2746A54498C0C026ULL, "6070359_m.fxe"},
	{0x2746B6B2437BABFBULL, "84917464.ogg"},
	{0x2746D15A9ADBD405ULL, "vxlt_0_lowtown_docks_n_b5s_0_0.vlm"},
	{0x2747E456668D5B66ULL, "vxlt_hangedman_n_2bj_0_0.vlm"},
	{0x2747F03EA90BD569ULL, "w_sundermount_81_0.rml"},
	{0x27487F68C18D1145ULL, "vxlt_0_lowtown_docks_a3_btt_0_0.vlm"},
	{0x274908A202F77CD8ULL, "508239964.ogg"},
	{0x27499C122E9CDE58ULL, "art\\vfx\\combat\\vfx_r_asn_as_arch_act_c.impa.gad"},
	{0x27499FCE35632381ULL, "dae270ar_gallows.are"},
	{0x2749CF51290A5B4CULL, "vrc221_bartrand6043024.cl"},
	{0x2749D9F4D3BA959EULL, "vxlt_0_lowtown_docks_a3_n_c3h_0_0.vlm"},
	{0x274A51CDFE9C8548ULL, "gen00fl_dog6166817.cl"},
	{0x274A5BB0F297C5E1ULL, "932755255.ogg"},
	{0x274A5E619B998930ULL, "gen00fl_anders6113135.cl"},
	{0x274A5FED26BEB789ULL, "vxdr_primevaltomb_2wr_0_0.vlm"},
	{0x274A6A054008F2FBULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation4.gad"},
	{0x274B18BFD939EB96ULL, "dae000_gamlen6166381.cl"},
	{0x274B2058B3CF89CFULL, "ht_0_hightown_a2_51_0_lc.mmh"},
	{0x274BA163D7F1982AULL, "w_blightlands_237_0.mao"},
	{0x274BA563D7F19F79ULL, "w_blightlands_237_0.mmh"},
	{0x274BAF63D7F1B067ULL, "w_blightlands_237_0.msh"},
	{0x274BE3F8C9BDF963ULL, "6167326_m.fxe"},
	{0x274BFA7456A4730BULL, "120472698.ogg"},
	{0x274BFB3D4F90D1D2ULL, "790825991.ogg"},
	{0x274C029BEE45C526ULL, "6167190_m.fxe"},
	{0x274C146EE68BC261ULL, "6106963_m.fxe"},
	{0x274C7E489807297BULL, "sse_ss_cityelf_f_1.gda"},
	{0x274CF7EB21AA2B2AULL, "follower_banter6156018.cl"},
	{0x274D10E96BBDD62CULL, "art\\levels\\4cinematics\\ferelden\\letter.mmh"},
	{0x274D16E96BBDE01EULL, "art\\levels\\4cinematics\\ferelden\\letter.msh"},
	{0x274DA20CDF59303EULL, "vxdr_primevaltomb_910_0_0.vlm"},
	{0x274DC720164F5212ULL, "6055436_f.fxe"},
	{0x274E54E1675B263FULL, "vxlt_undercityencounter_wl_0_0.vlm"},
	{0x274E96EBE2E97D09ULL, "narrative_varric6172559.cl"},
	{0x274EA2FC58882069ULL, "283968445.ogg"},
	{0x274ECE0FD02D327AULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgs.mao"},
	{0x274F0E69B909860FULL, "6162758_m.fxe"},
	{0x274F38F0E2B88589ULL, "172973064.ogg"},
	{0x274F4C9852638AACULL, "vxlt_0_lowtown_a3_n_c_eug_0_0.vlm"},
	{0x274F8683760AB20FULL, "one350tr_strider.ncs"},
	{0x27505EB6B482B5CEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet03a.msh"},
	{0x275078B6B482E1BCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet03a.mmh"},
	{0x27509E526FB21F79ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kitb_0n.dds"},
	{0x2750E56A2A490099ULL, "6089806_m.fxe"},
	{0x2751AE0B54990DDFULL, "6018711_f.fxe"},
	{0x2751E43EA9141549ULL, "w_sundermount_81_0.mao"},
	{0x2751F03EA91429E2ULL, "w_sundermount_81_0.mmh"},
	{0x2751F23EA9142C98ULL, "w_sundermount_81_0.msh"},
	{0x27521C2880D9FE9CULL, "vxht_brothel_n_6fu_0_0.vlm"},
	{0x2752373CF0761D63ULL, "vxw_bonepit_108_0_0.vlm"},
	{0x27538310F42C212AULL, "612250886.ogg"},
	{0x2753EE1AD9723D98ULL, "vxdr_primevalentrance_26t_0_0.vlm"},
	{0x2753FCEA6BB8EAD0ULL, "6039570_m.fxe"},
	{0x2754169C00153CADULL, "6164601_f.fxe"},
	{0x2754607DA628F8AAULL, "one110_pryce6088420.cl"},
	{0x27547E89A025C699ULL, "47827876.ogg"},
	{0x2754BB9A3FD7DD7BULL, "hlac_w_bonepitmining_a2_0_2y.dds"},
	{0x2755B0BB372CBF73ULL, "79053801.ogg"},
	{0x2755FBC1B8161386ULL, "qun221_petrice6076810.cl"},
	{0x27562AA6FCE7E072ULL, "ser120_ghyslain6104336.cl"},
	{0x2756936181FF4B1BULL, "vxht_chantry_2zz_0_0.vlm"},
	{0x275694A38C2E519CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep_1.mao"},
	{0x2756BDB512A095F2ULL, "zz_ave_debug6059247.cl"},
	{0x2757A7EFF50EF925ULL, "hlac_lt_0_lowtown_a2_n_c_-3_7z.dds"},
	{0x2757BB8900BDA4C4ULL, "905204891.ogg"},
	{0x275850C73EC177FAULL, "29863130.ogg"},
	{0x275857BB87E86FDBULL, "vxht_0_hightown_a2_c56_0_0.vlm"},
	{0x2758D3C16828FE0FULL, "6109713_m.fxe"},
	{0x2758F43E2A340D93ULL, "mer341_earl.bnk"},
	{0x275907CA0596DAE8ULL, "901959135.ogg"},
	{0x2759556337B4928AULL, "167842277.ogg"},
	{0x2759A996F4351444ULL, "6169882_m.fxe"},
	{0x275A55D7002F2136ULL, "6071275_m.fxe"},
	{0x275A8F1F8BD83FD0ULL, "lt_0_lowtown_docks_a3_n_141_0.msh"},
	{0x275A9D1F8BD857EAULL, "lt_0_lowtown_docks_a3_n_141_0.mmh"},
	{0x275AA11F8BD85E21ULL, "lt_0_lowtown_docks_a3_n_141_0.mao"},
	{0x275B7986998D291AULL, "fen310_varania6118592.cl"},
	{0x275BEF61C4ECC2B0ULL, "6206606_m.fxe"},
	{0x275C163E2A367DE3ULL, "mer341_earl.cnv"},
	{0x275C334C71AF808BULL, "490544357.ogg"},
	{0x275C80A99F00EAE7ULL, "art\\vfx\\combat\\vfx_glb_ele_s_additive_c.impa.gad"},
	{0x275CC6BE59513EBFULL, "vxlt_undercityencounter_ro_0_0.vlm"},
	{0x275D075120C90FBFULL, "vxw_sundermount_30d_0_1.vlm"},
	{0x275D39988AB0017DULL, "vxht_keep_3zm_0_0.vlm"},
	{0x275D82BD45C9B79FULL, "45265425.ogg"},
	{0x275DD04F8B8F19EBULL, "follower_banter6141916.cl"},
	{0x275DE62E4866903EULL, "vxw_encounter2_camp_8b_0_0.vlm"},
	{0x275DF2DE111CC9B1ULL, "hlac_lt_warehouse_1_6y.dds"},
	{0x275E48C3894E9D47ULL, "264506705.ogg"},
	{0x275E62D5A1344E48ULL, "llac_lt_0_lowtown_docks_a3_n_-131075_2.dds"},
	{0x275EAB5CE1258C65ULL, "gen00fl_aveline6139805.cl"},
	{0x275ECEB4B79943C3ULL, "vxlt_0_lowtown_a3_n_c_gpe_0_0.vlm"},
	{0x275EE96BF2986721ULL, "zz_vault_debug6203281.cl"},
	{0x275EF20F4B845ECEULL, "164652743.ogg"},
	{0x275F53755194C2BDULL, "vxlt_0_lowtown_a2_n_fen_0_0.vlm"},
	{0x275FEC7E2AAFBD3CULL, "dae000_flemeth6013334.cl"},
	{0x2760016564CB931FULL, "hm_gxa_vgk_cnspr_1.mor"},
	{0x276023EB75928B21ULL, "fx_flamefb03.dds"},
	{0x27608D982906E196ULL, "15573890.ogg"},
	{0x2760F6B3945D9343ULL, "gen00fl_fenris6115318.cl"},
	{0x27613457113284C7ULL, "dae120cr_amb_dwarf_5.utc"},
	{0x27618CA1F2B47C80ULL, "6209368_m.fxe"},
	{0x27619A0F5282FD42ULL, "637166944.ogg"},
	{0x27619C1A648A13CDULL, "6217120_m.fxe"},
	{0x2761B32011B58152ULL, "follower_banter6151459.cl"},
	{0x276275A99F0543F4ULL, "art\\vfx\\combat\\vfx_glb_ele_s_additive_c.impa.evt"},
	{0x27634F23574713D7ULL, "qun211cr_merc_loot.utc"},
	{0x2763C4B07BB2531FULL, "802175312.ogg"},
	{0x2763E203D6422004ULL, "ser221_gascard6165447.cl"},
	{0x276411437F0F3825ULL, "vxw_bonepit_2a2_0_0.vlm"},
	{0x2764227D6CD31D22ULL, "ser120_ghyslain6104826.cl"},
	{0x276475E087132725ULL, "follower_banter6164847.cl"},
	{0x2764A6B6496F920AULL, "vxlt_warehouse_n_uo_0_0.vlm"},
	{0x2764D382A6C3B9C0ULL, "vxlt_0_lowtown_a2_n_e29_0_0.vlm"},
	{0x2765D829EDFEB2D0ULL, "891547185.ogg"},
	{0x2765D83D6DA25875ULL, "art\\characters\\playercharacter\\humanmale\\hlf_lgtd.phy"},
	{0x2765F8094560A6ECULL, "6208544_f.fxe"},
	{0x27660DD9FC409F54ULL, "gen_im_trs_prf_rocks_01.uti"},
	{0x2766A885C6BB5394ULL, "vxht_keep_5ic_0_0.vlm"},
	{0x276743FA7E02BF84ULL, "llac_lt_0_lowtown_a2_65537_3.dds"},
	{0x2767C1D8BC4C8171ULL, "strmt02_n.dds"},
	{0x2767CFEDA381B032ULL, "vxlt_0_lowtown_gbn_0_0.vlm"},
	{0x2767E749DFEB2D39ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunklow_s.dds"},
	{0x2768EA0C162D6B0DULL, "lt_0_lowtown_docks_180_0.msh"},
	{0x2768F80C162D8340ULL, "lt_0_lowtown_docks_180_0.mao"},
	{0x2768FC0C162D8993ULL, "lt_0_lowtown_docks_180_0.mmh"},
	{0x27696FAA8A079FF0ULL, "lt_0_lowtown_a2_n_mmap_ph.dds"},
	{0x2769A720E27154C6ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrdb_0s.dds"},
	{0x276A833A303A9DD1ULL, "hlac_lt_0_lowtown_docks_a3_65534_0y.dds"},
	{0x276AA434ADABAE83ULL, "869944403.ogg"},
	{0x276AC3884D35FC9CULL, "w_sundermount_202_0.phy"},
	{0x276AFE8862508769ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_evilchya.phy"},
	{0x276C2F623CF2F743ULL, "mrl000_marethari6142410.cl"},
	{0x276C5468748C8F26ULL, "vxw_bonepit_25a_0_0.vlm"},
	{0x276C580B3A5CA8BEULL, "vxw_encounter2_camp_3c_0_0.vlm"},
	{0x276C647C42CAB3FEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobc_0s.dds"},
	{0x276C677B9884F951ULL, "664706763.ogg"},
	{0x276CB70DD841FEB9ULL, "vxlt_0_lowtown_docks_a3_n_8uf_0_0.vlm"},
	{0x276D166D58E1C3EDULL, "qunev_item_acquired.ncs"},
	{0x276DC89DA70A467BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabsnb.msh"},
	{0x276DDA9DA70A64C6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabsnb.mao"},
	{0x276DE69DA70A7965ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabsnb.mmh"},
	{0x276E2C82B2E70C99ULL, "6140587_m.fxe"},
	{0x276E7383762542B6ULL, "one350tr_strider.utt"},
	{0x276EFB18CB770B8DULL, "6151400_m.fxe"},
	{0x276FDDD96BFD778DULL, "dae000cr_grand_cleric6124780.cl"},
	{0x27706DA99F11732DULL, "art\\vfx\\combat\\vfx_glb_ele_s_additive_c.impa.ani"},
	{0x277084930F59CC5AULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_arishok_arishokaddress.gad"},
	{0x2770D68B54AEB05AULL, "w_bonepitmining_a2_12_0.rml"},
	{0x277109C3B8E04C39ULL, "hlac_ca_cave_2194_1y.dds"},
	{0x27712EFFBBED4422ULL, "6018731_m.fxe"},
	{0x27714C8D620482DAULL, "gen00fl_merrill6108254.cl"},
	{0x2772606869E7375DULL, "qcr220_meredith6054731.cl"},
	{0x2772C5D6E91F55DBULL, "246054491.ogg"},
	{0x2772E6BD2C0E2E89ULL, "zz_followers_banter6158352.cl"},
	{0x277318933B65AA8AULL, "vxlt_warehouse_n_xh_0_0.vlm"},
	{0x2773F77924BA9BF1ULL, "gen00fl_anders6125124.cl"},
	{0x277402DD76242B00ULL, "6119203_m.fxe"},
	{0x2774104FE8C446F0ULL, "hlac_dr_dungeon_3214_1y.dds"},
	{0x2774206920F457F1ULL, "615526190.ogg"},
	{0x27745DD7702804B6ULL, "524512023.ogg"},
	{0x27748126630442FCULL, "249283901.ogg"},
	{0x27750CB84E5CC75EULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.cess.evt"},
	{0x277557B6B4A1C587ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet03a.phy"},
	{0x277589E2B08274FBULL, "w_bonepit_28_0.rml"},
	{0x2775DDCE4133335AULL, "6096027_m.fxe"},
	{0x27764174257C4857ULL, "796118375.ogg"},
	{0x277642460D7FCB26ULL, "vxca_caverns2_812_0_1.vlm"},
	{0x27764D668D6344B6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_ruin.msh"},
	{0x277657668D6355B4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_ruin.mmh"},
	{0x27765B668D635CC7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_ruin.mao"},
	{0x2776C2F791A85142ULL, "combat_dw_r1_a5.gda"},
	{0x27777BF647EA3947ULL, "dae011_badpoet_sta6147006.cl"},
	{0x2778281E4FD8972CULL, "vxlt_0_lowtown_gjx_0_0.vlm"},
	{0x27783E8976FF00C3ULL, "mag360_samson6042065.cl"},
	{0x27785BA9934E71BEULL, "vxlt_undercitydungeon_23x_0_0.vlm"},
	{0x277876EAE6A072D9ULL, "w_encounter2_40_0.phy"},
	{0x2778E4930BEAD760ULL, "hlac_ht_playershouse_n_855_10x.dds"},
	{0x277962BB9AADB678ULL, "dae000_orsino6085422.cl"},
	{0x2779707492022509ULL, "vxht_playershouse_n_1vq_0_0.vlm"},
	{0x277986AABFC38953ULL, "359882677.ogg"},
	{0x2779CAA7623F89E9ULL, "hlac_lt_gallowstemplar_2382_3y.dds"},
	{0x2779D056F25C0F8AULL, "393916948.ogg"},
	{0x277A02648E0DA908ULL, "dae000_arishok6068435.cl"},
	{0x277AC2AF7784546CULL, "hlac_lt_0_lowtown_a3_n_c_-3_1x.dds"},
	{0x277AC6456682A7A6ULL, "vxw_bonepit_20b_0_0.vlm"},
	{0x277AD842DA1E0D92ULL, "vxht_0_hightown_a2_7fv_0_0.vlm"},
	{0x277AEF3EA9372C89ULL, "w_sundermount_81_0.anb"},
	{0x277B01B7D56D93F2ULL, "vxlt_hangedman_rw_0_0.vlm"},
	{0x277BBA2CF1841F4DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_chab_s.dds"},
	{0x277BF4A15754BEEAULL, "6042501_m.fxe"},
	{0x277C23B84E630C89ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.cess.gad"},
	{0x277C3D94160882EDULL, "w_encounter3_99_0.msh"},
	{0x277C4B9416089A20ULL, "w_encounter3_99_0.mao"},
	{0x277C4F941608A173ULL, "w_encounter3_99_0.mmh"},
	{0x277CBE4032EC1CDEULL, "6081737_m.fxe"},
	{0x277D4F4D9C026594ULL, "931341920.ogg"},
	{0x277D7382F09637C1ULL, "6049654_f.fxe"},
	{0x277DD73A2D938382ULL, "124282445.ogg"},
	{0x277E17D4F4A71AD8ULL, "mcr370_meredith6083086.cl"},
	{0x277E428632CC15A7ULL, "gen00fl_aveline6128979.cl"},
	{0x277E56CFA0483B77ULL, "659999891.ogg"},
	{0x277E7FEAE6A4EDF5ULL, "w_encounter2_40_0.rml"},
	{0x277EA70150A9D7A2ULL, "dae000_arishok6067649.cl"},
	{0x277ED17342D79A35ULL, "dae000_fem_commoner1_mob6174481.cl"},
	{0x277F1E7E826CC6C9ULL, "cs_dae340_db_drag_mage.wbd"},
	{0x277F82358C1C710FULL, "ht_keep_3739.anb"},
	{0x277FD2FAC3FBEEACULL, "595704514.ogg"},
	{0x27804DA0ED5A998AULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.cess.gad"},
	{0x27809B4BD20F3FFFULL, "vxht_mansion2b_24t_0_0.vlm"},
	{0x2781ABB32AC78CD0ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5.phy"},
	{0x278219914AEF7884ULL, "pro000_colonel_ewald6040624.cl"},
	{0x27826B5205ABB354ULL, "dae150_chandan6148551.cl"},
	{0x27827C650715395FULL, "llac_w_bonepitmining_a2_65541_0.dds"},
	{0x2782B83B035313A5ULL, "gen00fl_carver6104882.cl"},
	{0x2782F4B84E68DBD7ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.cess.ani"},
	{0x27831A5D9C2CFF47ULL, "6181750_f.fxe"},
	{0x27832FCE1E9E0C98ULL, "99370367.ogg"},
	{0x278393FE7968AA3EULL, "vxca_caverns1_j7d_0_0.vlm"},
	{0x27849B01BE8B5A2CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_concrete_n.dds"},
	{0x27849F930F6AA340ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_arishok_arishokaddress.ani"},
	{0x2784D90E26E80DEFULL, "6129456_m.fxe"},
	{0x2785D2F113E4C54DULL, "361010079.ogg"},
	{0x278699FB41CEAFACULL, "vxlt_0_lowtown_gqy_0_0.vlm"},
	{0x278730A0ED608721ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.cess.evt"},
	{0x2787E4E5B48FD6B5ULL, "6205606_m.fxe"},
	{0x2788190F9B6C09C1ULL, "vxht_0_hightown_a2_ap9_0_0.vlm"},
	{0x2788745AEB8A3A96ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.dura.ani"},
	{0x2788A32EF6E9D841ULL, "vxw_encounter2_js_0_0.vlm"},
	{0x2788CDACE789C394ULL, "vxht_0_hightown_n_cbo_0_0.vlm"},
	{0x27890AD8CBA69C37ULL, "dae260st_harley1.stg"},
	{0x278912B41105C2BFULL, "6093641_m.fxe"},
	{0x2789605006627E3FULL, "639146655.ogg"},
	{0x278965B1A558F686ULL, "dae000_arishok6078421.cl"},
	{0x278A36AEC3C3C94BULL, "vxlt_0_lowtown_a3_n_c_htv_0_0.vlm"},
	{0x278BB4C692F19F89ULL, "dre000_varric6208064.cl"},
	{0x278BD3A8E0CB9769ULL, "6060462_f.fxe"},
	{0x278C5D232D83002BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srda_0d.dds"},
	{0x278C6F25D1A36621ULL, "wm_ico_shop_roll.dds"},
	{0x278CEA5D597C26C4ULL, "6148036_m.fxe"},
	{0x278CF697D902C4B7ULL, "vxlt_gallowstemplar_4sl_0_0.vlm"},
	{0x278D10DA88B514C7ULL, "w_bonepitdestroyed_a3_11_0_ll.mmh"},
	{0x278D6328219CF7B7ULL, "hlac_lt_ruinedhouse_n_-1_2y.dds"},
	{0x278DD5122DC84799ULL, "vxlt_backalley_g_1l4_0_0.vlm"},
	{0x278E24908CC7832AULL, "gen00fl_fenris6105561.cl"},
	{0x278E446C8EE34B87ULL, "6076424_m.fxe"},
	{0x278E521D9552F742ULL, "699829883.ogg"},
	{0x278EAA5859E28254ULL, "gen00fl_anders6096452.cl"},
	{0x278EC45ACFED7FE3ULL, "vxlt_backalley_g_1hz_0_0.vlm"},
	{0x278F7A59646FA6DBULL, "6118171_m.fxe"},
	{0x278FD72EBB2F53DBULL, "dr_slavers_3207_lc.mmh"},
	{0x27900BA3D75FBEA0ULL, "follower_banter6154003.cl"},
	{0x279016C1F860B515ULL, "vxht_mansion2_3el_0_0.vlm"},
	{0x27901E4459F0BD1CULL, "vxw_sundermount_27z_0_0.vlm"},
	{0x27908069DBF9694EULL, "hlac_ht_playershouse_n_855_5x.dds"},
	{0x2790CF98935DEEEFULL, "lgt220_shp_mag_guard6149501.cl"},
	{0x2790D8E281C9670EULL, "hlac_ht_keep_n_4519_11y.dds"},
	{0x279185D76AF62E3DULL, "vxht_mansion2b_2yu_0_0.vlm"},
	{0x2791B44F0193A894ULL, "hlac_lt_gallowstemplar_n_2388_6z.dds"},
	{0x27920B53A3CF8047ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_exit_right.gad"},
	{0x2792159874B3E6A7ULL, "gen00fl_fenris6112100.cl"},
	{0x27929E7B91D4995BULL, "908311254.ogg"},
	{0x2792A18AB83A864EULL, "229819478.ogg"},
	{0x2792FF42E0B7D74FULL, "209347457.ogg"},
	{0x279343B7D0BA6AD6ULL, "vxw_sundermount_209_0_0.vlm"},
	{0x27935149634D11B2ULL, "llac_lt_0_lowtown_-262144_3.dds"},
	{0x2793CC104A90B567ULL, "campaign_base_fr-fr.cre"},
	{0x2793F8169B50EC06ULL, "gen00fl_anders6201687.cl"},
	{0x279428A0ED6B0470ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.cess.ani"},
	{0x279444338B6AF853ULL, "hlac_lt_gallowstemplar_n_7198_0y.dds"},
	{0x279496AA12FD1339ULL, "gen00fl_merrill6130076.cl"},
	{0x27951FA5753AAABFULL, "gen00fl_anders6111121.cl"},
	{0x27954EBFB58913CFULL, "6129030_m.fxe"},
	{0x27957DD3D5D6E1B5ULL, "lt_0_lowtown_docks_a3_25_0_ll.mmh"},
	{0x27961C23675B2F82ULL, "qun211_braggart6075700.cl"},
	{0x279652278225BBFCULL, "49315336.ogg"},
	{0x27967C5AEB968587ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.dura.evt"},
	{0x27968AEC8D622241ULL, "vxht_0_hightown_a2_ak8_0_0.vlm"},
	{0x2796C37730664078ULL, "hlac_ht_keep_3736_3y.dds"},
	{0x2796E78862759A8EULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_evilchya.mmh"},
	{0x2796F3886275AF25ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_evilchya.mao"},
	{0x279701886275C77CULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_evilchya.msh"},
	{0x2797320C0F0457DDULL, "348636742.ogg"},
	{0x279738E63D94912CULL, "320972955.ogg"},
	{0x27976EDDE04D573DULL, "6157193_f.fxe"},
	{0x2797A0D234732D12ULL, "6095677_m.fxe"},
	{0x2797BC06D7BE937FULL, "6128643_f.fxe"},
	{0x2798ECB3FA1D9F7AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.open.gad"},
	{0x27990053A3D58C54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_exit_right.evt"},
	{0x279930548AC85B2AULL, "765496153.ogg"},
	{0x2799DACC20E6A236ULL, "668373089.ogg"},
	{0x2799FEC619C1E9FDULL, "6154003_m.fxe"},
	{0x279B4B5A1C7BA72CULL, "hlac_ca_caverns2_5802_69z.dds"},
	{0x279B9C7AD49D0B5BULL, "vxw_bonepit_1uc_0_0.vlm"},
	{0x279BB9A24D084754ULL, "vxlt_0_lowtown_docks_n_bfc_0_0.vlm"},
	{0x279BCFC148084455ULL, "558774056.ogg"},
	{0x279BEF14396D9EB8ULL, "vxht_0_hightown_a2_c6x_0_0.vlm"},
	{0x279D9D5AEB9CDCD0ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.dura.gad"},
	{0x279DCDF963F01130ULL, "323677675.ogg"},
	{0x279E550DB8D17D38ULL, "hlac_lt_0_lowtown_a2_n_131070_2x.dds"},
	{0x279F2204ED2F15EDULL, "6096176_m.fxe"},
	{0x279F9D33230AEC6FULL, "art\\vfx\\creatures\\fxc_footstep_l_c.mmh"},
	{0x279FD4DA69D2CC4FULL, "hlac_w_encounter2_camp_-131072_0x.dds"},
	{0x27A07908224AFC61ULL, "vxlt_0_lowtown_a2_n_c_ghv_0_0.vlm"},
	{0x27A0ED3A965371B2ULL, "gen00fl_aveline6122808.cl"},
	{0x27A11122CACFDF25ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb105a_0t.dds"},
	{0x27A119A29FAAD8F6ULL, "6077447_f.fxe"},
	{0x27A14BB9582ED181ULL, "ht_0_hightown_n_53_0.anb"},
	{0x27A1AAD25F81CDF2ULL, "6172057_m.fxe"},
	{0x27A2122B6A74E3EAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_8.pwk"},
	{0x27A2A3E591315B16ULL, "6087765_f.fxe"},
	{0x27A2C43B3F75C706ULL, "hlac_ht_playershouse_n_855_11y.dds"},
	{0x27A31BDE727B4B0AULL, "vxlt_darktown_4ec_0_0.vlm"},
	{0x27A320F0510976DEULL, "gen00fl_aveline6113978.cl"},
	{0x27A36C4343F7595CULL, "6155430_m.fxe"},
	{0x27A37CAF72296C57ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.impa.gad"},
	{0x27A38EA2CA9C4F24ULL, "hlac_lt_0_lowtown_-65536_7x.dds"},
	{0x27A3E531C11EC854ULL, "390770954.ogg"},
	{0x27A4434A478C0F6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_action.gad"},
	{0x27A47F59D8358E96ULL, "hlac_lt_foundry_n_320_10z.dds"},
	{0x27A4F07E828D47ABULL, "cs_dae340_db_drag_mage.bnk"},
	{0x27A5215B1AC72BD2ULL, "hlac_ca_cavern1_haunted_24424_5y.dds"},
	{0x27A540D58273F0B5ULL, "53419544.ogg"},
	{0x27A574B140ECF5FAULL, "6045338_f.fxe"},
	{0x27A586E8DAD60941ULL, "vxw_encounter2_pq_0_0.vlm"},
	{0x27A5F853A3E0088DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_exit_right.ani"},
	{0x27A81EBCE90A983BULL, "dae000_pc_home_ambients6208319.cl"},
	{0x27A8341454AD5F46ULL, "gencr_rubble_golem_a.utc"},
	{0x27A84DDB11C0DAE1ULL, "102242891.ogg"},
	{0x27A8648606FE950BULL, "hlac_w_encounter2_camp_65537_0y.dds"},
	{0x27A893CA6F5D052DULL, "ser200_aveline6201485.cl"},
	{0x27A89AE634A8C03DULL, "hlac_ht_0_hightown_a2_-65535_0y.dds"},
	{0x27A904994B1D2E27ULL, "248229402.ogg"},
	{0x27A9116EE219DA84ULL, "6044880_m.fxe"},
	{0x27A92B22797909CCULL, "6201222_f.fxe"},
	{0x27A9457E4997587FULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_alpha_a.dds"},
	{0x27AA01BDE39FA848ULL, "vxlt_0_lowtown_a2_n_c_ghs_0_0.vlm"},
	{0x27AA46C1EC089948ULL, "dae000_merchant_armor_hm6164284.cl"},
	{0x27AA7E86EC5F83E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_give.ani"},
	{0x27AA91AF722FAEE4ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.impa.evt"},
	{0x27AA97CD35C5AB7FULL, "507880768.ogg"},
	{0x27AACAD2D86B74A4ULL, "gen00fl_sebastian6177921.cl"},
	{0x27AAF89B7AEDB102ULL, "467681770.ogg"},
	{0x27AB584A4792511CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_action.evt"},
	{0x27AB5F358C4170B3ULL, "ht_keep_3739.rml"},
	{0x27AC7BAA06AD3E60ULL, "348390766.ogg"},
	{0x27AC94216093FE70ULL, "734012856.ogg"},
	{0x27ACBD1EF151CC9EULL, "lt_0_lowtown_a2_n_10_0.rml"},
	{0x27ACEB3D6DDE7A90ULL, "art\\characters\\playercharacter\\humanmale\\hlf_lgtd.msh"},
	{0x27ACF93D6DDE92AAULL, "art\\characters\\playercharacter\\humanmale\\hlf_lgtd.mmh"},
	{0x27ACFD3D6DDE99E1ULL, "art\\characters\\playercharacter\\humanmale\\hlf_lgtd.mao"},
	{0x27AD07B3FA2E7660ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.open.ani"},
	{0x27AD8687ADC30CF0ULL, "hlac_ca_cavern1_haunted_24423_8y.dds"},
	{0x27AE28FED77EC370ULL, "1056594125.ogg"},
	{0x27AE69914EE25F3DULL, "vxht_mansion2b_2os_0_0.vlm"},
	{0x27AEA428555E0E60ULL, "6080390_m.fxe"},
	{0x27AF20606749D985ULL, "vxlt_gallowstemplar_n_3e0_0_0.vlm"},
	{0x27AFA285BAB1C556ULL, "mag100_cullen6140968.cl"},
	{0x27AFBC4567CBDEDBULL, "293763487.ogg"},
	{0x27B04A3825013FCFULL, "vxlt_0_lowtown_a2_n_gos_0_0.vlm"},
	{0x27B0DE3E609D72A5ULL, "959367624.ogg"},
	{0x27B19DB2C8D7DA07ULL, "6126453_m.fxe"},
	{0x27B1BFCF589CE02DULL, "llac_ca_caverns1_mines_24424_14.dds"},
	{0x27B2F31A3B3DB520ULL, "mrl310_party_comments6143054.cl"},
	{0x27B48D1F70D08E71ULL, "789428896.ogg"},
	{0x27B4E652C25C9123ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyn_0n.dds"},
	{0x27B4E6A3EDC52A40ULL, "6055569_m.fxe"},
	{0x27B5131ED4699AB3ULL, "vxht_mansion2b_29k_0_0.vlm"},
	{0x27B519EB8022C4EFULL, "gen_im_trs_bst_drg_scale_03.uti"},
	{0x27B523C542AAFCB6ULL, "llac_ca_caverns2prc_5802_24.dds"},
	{0x27B5B0F166AC50C6ULL, "255172839.ogg"},
	{0x27B5BF9DA747EC12ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabsnb.phy"},
	{0x27B5D189A37FEEABULL, "mag210_arianni6057318.cl"},
	{0x27B63555B9AD63C0ULL, "gen_im_trs_dem_rng_r2.uti"},
	{0x27B66349FE9FBFB7ULL, "259822119.ogg"},
	{0x27B68CE2B0BA10C7ULL, "w_bonepit_28_0.anb"},
	{0x27B7327320D9AEFFULL, "6096035_m.fxe"},
	{0x27B74717488C8632ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightl_1.mao"},
	{0x27B79686EC6A36DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_give.evt"},
	{0x27B7B68B54EA7C11ULL, "w_bonepitmining_a2_12_0.mmh"},
	{0x27B7C28B54EA90B2ULL, "w_bonepitmining_a2_12_0.mao"},
	{0x27B7D08B54EAA8CFULL, "w_bonepitmining_a2_12_0.msh"},
	{0x27B7FC1E26BBAE78ULL, "vxlt_0_lowtown_fed_0_0.vlm"},
	{0x27B84352D2EAE4B0ULL, "vxlt_0_lowtown_docks_n_cth_0_0.vlm"},
	{0x27B854BB59060861ULL, "119980531.ogg"},
	{0x27B8604A479CE935ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_action.ani"},
	{0x27B8A9AF723C14DDULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.impa.ani"},
	{0x27B8C5C8F209B888ULL, "437893600.ogg"},
	{0x27B8D2CE1D59CFB8ULL, "vxht_0_hightown_a2_c8v_0_0.vlm"},
	{0x27B8FFCF6A13A206ULL, "vxht_0_hightown_a2_c88_0_0.vlm"},
	{0x27B9AE49306E3557ULL, "287275149.ogg"},
	{0x27BA132DD2EBD18BULL, "vxlt_0_lowtown_docks_a3_n_bvk_0_0.vlm"},
	{0x27BA9E7AF1AAEF56ULL, "1047659220.ogg"},
	{0x27BA9F588A1DE40CULL, "hlac_lt_0_lowtown_a2_n_65538_1x.dds"},
	{0x27BAA4E6EA989800ULL, "vxlt_0_lowtown_docks_n_cp4_0_0.vlm"},
	{0x27BB025345ED589DULL, "dre000_varric6207959.cl"},
	{0x27BC3D351AFB03E0ULL, "vxht_0_hightown_arz_0_0.vlm"},
	{0x27BCA83082E09133ULL, "vxlt_0_lowtown_a2_n_f1t_0_0.vlm"},
	{0x27BCFDB9BBB582CDULL, "mrl150_dalish_guard6068747.cl"},
	{0x27BDC674A05E006EULL, "vfx_vfx_p9.vsh.fxo.dx11"},
	{0x27BDD8EACF2D5122ULL, "vxlt_0_lowtown_docks_a3_cui_0_0.vlm"},
	{0x27BE792CA485F363ULL, "6174559_m.fxe"},
	{0x27BE9586EC7053CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_give.gad"},
	{0x27BEBBBD13572DB9ULL, "zz_vault_debug6178668.cl"},
	{0x27BEC28656C45FA3ULL, "hm_ave3_insurgent3.mor"},
	{0x27BF360A85A3471FULL, "vxlt_backalley_n_1ds_0_0.vlm"},
	{0x27BF617ED03B18D4ULL, "hlac_ca_cavern1_haunted_24426_13x.dds"},
	{0x27BFF34442BA9150ULL, "vxlt_shop_dt_0_0.vlm"},
	{0x27C05464A80AC884ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_dwpa_0d.dds"},
	{0x27C0C119FA7D72BEULL, "vxdr_primevalentrance_275_0_0.vlm"},
	{0x27C0D150DEEB0663ULL, "6108744_f.fxe"},
	{0x27C16E64F869E82BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv1.evt"},
	{0x27C23762D65F144CULL, "56078167.ogg"},
	{0x27C25E8B0F7D0EA7ULL, "6110470_m.fxe"},
	{0x27C3113A791E6F95ULL, "833510913.ogg"},
	{0x27C31E865498B7B3ULL, "6119587_m.fxe"},
	{0x27C3BEB83447B2E2ULL, "427141542.ogg"},
	{0x27C4564565B73AD0ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean.msh"},
	{0x27C4592B55442330ULL, "6069571_m.fxe"},
	{0x27C4644565B752EAULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean.mmh"},
	{0x27C4684565B75921ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean.mao"},
	{0x27C47A1E2C4F4423ULL, "vxw_encounter2_47_0_0.vlm"},
	{0x27C4BD01C483BB2AULL, "dae000ip_crowd_lm1_talk3.utp"},
	{0x27C59E1CD8A9806DULL, "172199028.ogg"},
	{0x27C5A44641CF8CADULL, "rune_armor_fire.dds"},
	{0x27C60222C841AB23ULL, "gen00fl_anders6202357.cl"},
	{0x27C60A5A6AA743BCULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_0x.dds"},
	{0x27C62512EC27FC78ULL, "6204969_m.fxe"},
	{0x27C6B52FC4E0FD30ULL, "vxlt_0_lowtown_docks_n_coi_0_0.vlm"},
	{0x27C6BE713D174DF9ULL, "gen00fl_carver6103883.cl"},
	{0x27C6EA062ED60B04ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyo_0d.dds"},
	{0x27C72AD981B888F5ULL, "341035241.ogg"},
	{0x27C76CF2F626210BULL, "mcr310_argument6156722.cl"},
	{0x27C88764F87031ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv1.gad"},
	{0x27C89E33232DF2ACULL, "art\\vfx\\creatures\\fxc_footstep_l_c.anb"},
	{0x27C9A36A3213EB07ULL, "vxca_caverns1_mines_k6n_0_0.vlm"},
	{0x27C9D0634C189739ULL, "art\\levels\\ztools\\default\\generic\\prp_joe_2.mao"},
	{0x27C9D989A2DCECC9ULL, "10830829.ogg"},
	{0x27CA118E5A9F4C62ULL, "w_bonepit_64_0.msh"},
	{0x27CA138E5A9F4F18ULL, "w_bonepit_64_0.mmh"},
	{0x27CA1DEFA78D752EULL, "vxca_caverns2_99f_0_0.vlm"},
	{0x27CA1F8E5A9F63BBULL, "w_bonepit_64_0.mao"},
	{0x27CA3C06390725F9ULL, "vxht_mansion2_27o_0_0.vlm"},
	{0x27CB4D4B32CE903DULL, "vxht_mansion2b_2eq_0_0.vlm"},
	{0x27CB7C2FDE75A260ULL, "lt_foundry_n_mmap_ph.dds"},
	{0x27CC900C320E95A3ULL, "c_wldsylvna_0t.dds"},
	{0x27CD091C120DF4B4ULL, "vxdr_0_deeproads_es3_0_1.vlm"},
	{0x27CDBB3FDF27EEF1ULL, "hlac_lt_darktown_-1_32x.dds"},
	{0x27CDC2E4D20980D0ULL, "6112686_m.fxe"},
	{0x27CDCE6217911694ULL, "vxw_bonepitdestroyed_a3_1g2_0_0.vlm"},
	{0x27CE380A436F68D1ULL, "6099159_m.fxe"},
	{0x27CE8377995003EEULL, "6115735_m.fxe"},
	{0x27CEE0F24DB49069ULL, "610885090.ogg"},
	{0x27CEE7C8C3E82CD6ULL, "hlac_lt_0_lowtown_docks_a3_-3_2y.dds"},
	{0x27CF027A05B396CFULL, "docks_male_sailor_mob6176647.cl"},
	{0x27CF8664F8764EEAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv1.ani"},
	{0x27D0325C5354A7E8ULL, "m_fen200ar_ambush_exterior.wbd"},
	{0x27D04818AD2BCA30ULL, "6088245_f.fxe"},
	{0x27D0C02B12FD8395ULL, "6068466_f.fxe"},
	{0x27D0E64EC128A5B0ULL, "vxlt_undercityencounter_1c5_0_1.vlm"},
	{0x27D19063468ED033ULL, "6114287_m.fxe"},
	{0x27D1FCB27C202559ULL, "vxw_encounter2_3p_0_0.vlm"},
	{0x27D222D0553E790DULL, "hlac_lt_ruinedhouse_n_-1_3x.dds"},
	{0x27D256668DB1B2AFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_ruin.phy"},
	{0x27D2B15B78128A50ULL, "mag121_viveka6156510.cl"},
	{0x27D39BB857FD6A72ULL, "sib100_father_contact6110074.cl"},
	{0x27D3EB60E6EBA776ULL, "gen00fl_anders6096469.cl"},
	{0x27D3F31790D01B3BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\vst_hightown_d.dds"},
	{0x27D411519580D2A5ULL, "95132076.ogg"},
	{0x27D5166DB034A397ULL, "art\\characters\\playercharacter\\humanfemale\\boo_maga.phy"},
	{0x27D52178E2CFA58CULL, "hf_shopgirl.mor"},
	{0x27D563EE9C756D5BULL, "vxw_bonepit_1ig_0_0.vlm"},
	{0x27D66D960C30FCBFULL, "and111_donation_box6200723.cl"},
	{0x27D673C83D4490ADULL, "llac_lt_0_lowtown_a2_n_65536_8.dds"},
	{0x27D706CDD970E901ULL, "follower_banter6169750.cl"},
	{0x27D733909D10A4CBULL, "w_bonepit_19_0_ll.mmh"},
	{0x27D73B0407C3AE12ULL, "hightown_male_noble_gossip6171435.cl"},
	{0x27D81C8380C369CCULL, "6117673_f.fxe"},
	{0x27D823633F7C938AULL, "bee_cyc_g.dds"},
	{0x27D827473B5418B1ULL, "art\\vfx\\models\\vfx_cylinder_2sd.mmh"},
	{0x27D841473B54456FULL, "art\\vfx\\models\\vfx_cylinder_2sd.msh"},
	{0x27D8A7787E54A3DBULL, "fex_bd_a_md_cn.mmh"},
	{0x27D8C5787E54D6C5ULL, "fex_bd_a_md_cn.msh"},
	{0x27D95BAB183E3CFFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capright.msh"},
	{0x27D961AB183E46A1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capright.mmh"},
	{0x27D96DAB183E5B42ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_capright.mao"},
	{0x27D9A56FE8DB6D9CULL, "dae000cr_grand_cleric6191347.cl"},
	{0x27DAE67C43EF3040ULL, "251605104.ogg"},
	{0x27DAF8DE348C0070ULL, "416748754.ogg"},
	{0x27DB2CB958601511ULL, "ht_0_hightown_n_53_0.rml"},
	{0x27DBF1B9D348A07EULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga104a_0d.dds"},
	{0x27DC16AD5C3FB9B6ULL, "statecnt_crafting.gda"},
	{0x27DC403F09872F14ULL, "vxw_bonepitdestroyed_a3_1h3_0_0.vlm"},
	{0x27DC612E8584011FULL, "hlac_lt_0_lowtown_a2_n_c_-196608_4x.dds"},
	{0x27DC6C6AA456F340ULL, "326492620.ogg"},
	{0x27DC7C8D22F987B0ULL, "6125653_m.fxe"},
	{0x27DC854E563AE577ULL, "339391032.ogg"},
	{0x27DCA3A6C3C9C9F3ULL, "921031845.ogg"},
	{0x27DD3E8F1197C948ULL, "dre000_varric6030013.cl"},
	{0x27DDB50CE2BDC64AULL, "6048236_f.fxe"},
	{0x27DDECB29571B899ULL, "159993258.ogg"},
	{0x27DE1E44697116D7ULL, "mag311_emile_de_launcet6043503.cl"},
	{0x27DE2B63EC34E3BBULL, "mag170_thrask6054025.cl"},
	{0x27DE4F4679C01783ULL, "isa300ip_office_door.utp"},
	{0x27DE720AC64E93D0ULL, "w_bonepit_161_0.msh"},
	{0x27DE800AC64EABEAULL, "w_bonepit_161_0.mmh"},
	{0x27DE840AC64EB221ULL, "w_bonepit_161_0.mao"},
	{0x27DEAB19403B17E6ULL, "6111195_m.fxe"},
	{0x27DEB30BD8DD2E1CULL, "llac_lt_gallowstemplar_2394_8.dds"},
	{0x27DED0D5BDF258A2ULL, "gen00fl_bethany6028822.cl"},
	{0x27DF89E2B0DD10C8ULL, "w_bonepit_28_0.mmh"},
	{0x27DF95E2B0DD256BULL, "w_bonepit_28_0.mao"},
	{0x27DFA7E2B0DD43F2ULL, "w_bonepit_28_0.msh"},
	{0x27E00D490BBF7E53ULL, "vxlt_0_lowtown_a2_n_ej6_0_0.vlm"},
	{0x27E030C89ED13BECULL, "40898146.ogg"},
	{0x27E0448613DF33B9ULL, "vxht_mansion2b_343_0_0.vlm"},
	{0x27E05785A19490EDULL, "lowtown_fem_lowlife_mob6168986.cl"},
	{0x27E07029BCE4FBDDULL, "657498532.ogg"},
	{0x27E0BA26BD0FB012ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.dura.gad"},
	{0x27E0C07AFA40285FULL, "qcr220_orsino6061335.cl"},
	{0x27E146C8894A0903ULL, "vxlt_undercityserialkiller_1im_0_0.vlm"},
	{0x27E1F46C529B3AD0ULL, "vxlt_gallowscourtyard_74k_0_0.vlm"},
	{0x27E1FBED7232C1B7ULL, "870490511.ogg"},
	{0x27E302F2240DE582ULL, "gen_im_acc_rng_act2_06.uti"},
	{0x27E3A05D608287CEULL, "151368697.ogg"},
	{0x27E3AD617D0B046BULL, "ht_0_hightown_a2_28_0_lc.mmh"},
	{0x27E4F3247E2D396AULL, "dae000_arishok6082046.cl"},
	{0x27E510E79B31E294ULL, "6155913_m.fxe"},
	{0x27E5C9A716E7A3EAULL, "follower_banter6170234.cl"},
	{0x27E5F07E9B275B0BULL, "6073805_m.fxe"},
	{0x27E60122D742ADF7ULL, "843934444.ogg"},
	{0x27E608E557CD026FULL, "vxca_caverns1_mines_cn6_0_0.vlm"},
	{0x27E641A7FC3EF947ULL, "zz_followers_banter6158181.cl"},
	{0x27E698E2DDF6952CULL, "qun221_viscount6074044.cl"},
	{0x27E77AC0AD40070AULL, "6013352_m.fxe"},
	{0x27E7A8CAD5F01C86ULL, "llac_lt_0_lowtown_a2_0_7.dds"},
	{0x27E7BD26BD15D449ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.dura.evt"},
	{0x27E7C44142C1CD3AULL, "gen_im_trs_gen_arc_sbw_r0.uti"},
	{0x27E803FF293634B1ULL, "zz_dae_debug6029072.cl"},
	{0x27E80E9953E8C8EEULL, "6149903_m.fxe"},
	{0x27E877C111FED140ULL, "vxlt_warehouse_ap_0_0.vlm"},
	{0x27E89D53D27FB636ULL, "516420372.ogg"},
	{0x27E9F46504A687EEULL, "static1.vsh.fxo"},
	{0x27EA218EC380CB15ULL, "6065933_f.fxe"},
	{0x27EA497F8993112DULL, "6148402_m.fxe"},
	{0x27EA4DC6B33C2AA4ULL, "372748874.ogg"},
	{0x27EA4EDEB10510F4ULL, "mag100st_thrask2.stg"},
	{0x27EA71E6CCEF6F75ULL, "6065084_m.fxe"},
	{0x27EA723D1BC9C9A6ULL, "vxw_blightlands_11w_0_0.vlm"},
	{0x27EA8C58742C27ECULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_hairm1.mao"},
	{0x27EA9F286A533223ULL, "isa211_isabela6202187.cl"},
	{0x27EAB21BFB7D4794ULL, "vxw_bonepitdestroyed_a3_1m0_0_0.vlm"},
	{0x27EAB7E5BF37FD37ULL, "103380682.ogg"},
	{0x27EB3006F6869856ULL, "566179879.ogg"},
	{0x27EBB10B9043799DULL, "368822916.ogg"},
	{0x27EBBB83D8FC7AA0ULL, "253598686.ogg"},
	{0x27EBDC017E23A0A1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp01.phy"},
	{0x27EC92E31607782CULL, "gen00fl_merrill6130000.cl"},
	{0x27ECA745F6FD4CF3ULL, "dae000_gamlen6145095.cl"},
	{0x27ED16295A56D1BCULL, "vxlt_backalley_g_1vc_0_0.vlm"},
	{0x27ED3B0D1B52E5FDULL, "hlac_ca_caverns2prc_5802_24y.dds"},
	{0x27EE4D4BA228A582ULL, "vxw_sundermount_36s_0_1.vlm"},
	{0x27EE7B6CFDF16400ULL, "vxlt_0_lowtown_a2_n_gw8_0_0.vlm"},
	{0x27EE7F25FDB596D3ULL, "vxlt_0_lowtown_a2_n_eo9_0_0.vlm"},
	{0x27EE825473BF6ED6ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_fire_d.dds"},
	{0x27EE8882DAA94243ULL, "hlac_ca_caverns1_24424_22z.dds"},
	{0x27EF1633715F9C33ULL, "nat380_nathaniel6068372.cl"},
	{0x27EF1AC83BABD43AULL, "lt_0_lowtown_docks_a3_n_27_0.rml"},
	{0x27EF60E204BA7FF7ULL, "34041205.ogg"},
	{0x27EF7D0E94DC26C2ULL, "vxlt_gallowscourtyard_n_4w8_0_0.vlm"},
	{0x27EF87DB3CFBA91BULL, "663952945.ogg"},
	{0x27EFC727DE20F863ULL, "vxw_bonepitmining_a2_25n_0_0.vlm"},
	{0x27F017369E6BDEBBULL, "hightown_market_guard_sta6174614.cl"},
	{0x27F0281634DD2288ULL, "437288480.ogg"},
	{0x27F0333ED88C903EULL, "hlac_ht_0_hightown_a3_-131073_0z.dds"},
	{0x27F03A0E72966A43ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd0col.msh"},
	{0x27F0480E7296819DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd0col.mmh"},
	{0x27F04C0E729688AEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterd0col.mao"},
	{0x27F0556CDCBFFB62ULL, "6114800_m.fxe"},
	{0x27F13649B49EB67DULL, "gen00fl_isabela6110218.cl"},
	{0x27F150B642DE6EBAULL, "hlac_ca_cave_2194_14x.dds"},
	{0x27F180D69D4B1BC9ULL, "835733042.ogg"},
	{0x27F1C2774D7ED951ULL, "924680916.ogg"},
	{0x27F20CA47C54057FULL, "hlac_lt_gallowstemplar_2388_3y.dds"},
	{0x27F21ADBFB9E89D3ULL, "6120976_m.fxe"},
	{0x27F2412E340DF7BCULL, "100580099.ogg"},
	{0x27F2C422A00C6A77ULL, "follower_barks6154988.cl"},
	{0x27F317DF9327216CULL, "138205016.ogg"},
	{0x27F32D2035B6052EULL, "gen00fl_aveline6139413.cl"},
	{0x27F36C8847C6B8D6ULL, "400369998.ogg"},
	{0x27F3E37BE34BA551ULL, "vxlt_0_lowtown_n_e3c_0_0.vlm"},
	{0x27F3FB291E3F073DULL, "dae020cr_ef_courtesan.utc"},
	{0x27F423F5E5225D04ULL, "6042399_f.fxe"},
	{0x27F47BE76AEB0436ULL, "lt_0_lowtown_docks_121_0.msh"},
	{0x27F485E76AEB1534ULL, "lt_0_lowtown_docks_121_0.mmh"},
	{0x27F489E76AEB1C47ULL, "lt_0_lowtown_docks_121_0.mao"},
	{0x27F50DD21931B4DFULL, "hlac_ht_mansion2b_818_59y.dds"},
	{0x27F524CA5C33853BULL, "fen210_dying_magister6108621.cl"},
	{0x27F544009B2618CCULL, "hairalpha_hair_p1.psh.fxo"},
	{0x27F5516A3CEC60DDULL, "6143072_m.fxe"},
	{0x27F5A526BD21E8B8ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.dura.ani"},
	{0x27F5BA012F29ECC5ULL, "pro000_merchant6051557.cl"},
	{0x27F5E55C4EC79055ULL, "mag101_grace6035359.cl"},
	{0x27F666F86DA4EA36ULL, "265587566.ogg"},
	{0x27F6807645DEE16EULL, "6142803_m.fxe"},
	{0x27F6BF52D71EDA92ULL, "6113961_f.fxe"},
	{0x27F6C902DB993527ULL, "vxlt_0_lowtown_a2_n_c_hsv_0_0.vlm"},
	{0x27F73FCF4CA2E70AULL, "llac_lt_0_lowtown_a2_n_131069_0.dds"},
	{0x27F747D48DED30CCULL, "268053624.ogg"},
	{0x27F7781B557A32E5ULL, "gen_im_acc_amu_act1_04.uti"},
	{0x27F7DAB58E12E54FULL, "6086218_m.fxe"},
	{0x27F7F4E5861FE6E2ULL, "6124361_f.fxe"},
	{0x27F89F68481C5EF7ULL, "655908934.ogg"},
	{0x27F8DE69E38C0B8CULL, "6152585_m.fxe"},
	{0x27F9429AE383B318ULL, "6095442_m.fxe"},
	{0x27F96CDC8957ACFFULL, "261128726.ogg"},
	{0x27F986F8DE02AC3EULL, "dae000_junar6147915.cl"},
	{0x27F99CDFA08D65B4ULL, "944705519.ogg"},
	{0x27F9BC66A2D4ADF3ULL, "vxw_encounter1_am_0_0.vlm"},
	{0x27FAA54704BBD770ULL, "vxca_caverns1_ijb_0_0.vlm"},
	{0x27FB2D8DB349E263ULL, "6016840_m.fxe"},
	{0x27FC86B68FF09C11ULL, "w_bonepitmining_a2_179_0.msh"},
	{0x27FC94B68FF0B42CULL, "w_bonepitmining_a2_179_0.mao"},
	{0x27FCA0B68FF0C8CFULL, "w_bonepitmining_a2_179_0.mmh"},
	{0x27FD1BCB2B22DC2EULL, "156430632.ogg"},
	{0x27FD74D4F3F1F133ULL, "ave200_donnic6110715.cl"},
	{0x27FD9F94804D6063ULL, "and111_lirene6169997.cl"},
	{0x27FE02CF8D64C19DULL, "lt_0_lowtown_a3_n_c_34_0_ll.mmh"},
	{0x27FE9D2FA605CA35ULL, "llac_lt_undercityencounter_-1_49.dds"},
	{0x27FEB239AA92C01DULL, "515106630.ogg"},
	{0x27FEBC7BE6FC91D4ULL, "6087500_m.fxe"},
	{0x27FEC749CEB613E2ULL, "6113900_m.fxe"},
	{0x27FFA6B2F8097AFFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_windows.phy"},
	{0x27FFE0F4250045E5ULL, "zz_vault_debug6203159.cl"},
	{0x28002DD6CEFADF59ULL, "651793941.ogg"},
	{0x280066E71B300252ULL, "zz_mag_debug6053419.cl"},
	{0x280082A91F45030FULL, "705032527.ogg"},
	{0x2800D4BA04480CC0ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_2z.dds"},
	{0x2801579868962FB7ULL, "llac_ca_caverns1_mines_24426_9.dds"},
	{0x28018BB92BDDFF33ULL, "61899660.ogg"},
	{0x28028ED3356B01B7ULL, "6155583_m.fxe"},
	{0x2802E66FA492E7ACULL, "6080112_m.fxe"},
	{0x28030310F4D982E0ULL, "dae200_bran6144577.cl"},
	{0x28032FDE68EA0AE2ULL, "vxca_caverns1_gk9_0_0.vlm"},
	{0x28036E174FC7B40AULL, "dae000_gamlen6005828.cl"},
	{0x2803F0CB3BB3BA90ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg105a.mao"},
	{0x2803F4CB3BB3C1A3ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg105a.mmh"},
	{0x2803FEAF012F9480ULL, "326353312.ogg"},
	{0x2803FFB491EA35C6ULL, "573267112.ogg"},
	{0x280402CB3BB3D9FDULL, "art\\characters\\weapons\\dagger\\w_dgr_dg105a.msh"},
	{0x28047685D4C97E60ULL, "vxht_0_hightown_aig_0_0.vlm"},
	{0x28048912C2021E1DULL, "949488979.ogg"},
	{0x2804CB8FCF3D5B31ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr90pause.gad"},
	{0x280528C967550DFAULL, "190444013.ogg"},
	{0x28057A352F7B45DBULL, "arm_dal.tnt"},
	{0x28058CE890BF7295ULL, "vxw_encounter3_bl_0_0.vlm"},
	{0x2805AAE2D523D9E0ULL, "355961270.ogg"},
	{0x2805BDECFA1DA057ULL, "hlac_ht_mansion2b_818_51y.dds"},
	{0x2805FF4B0CB4C3AAULL, "vxlt_gallowstemplar_3n5_0_0.vlm"},
	{0x2806ECB1F84DDF9BULL, "vxdr_primevaltomb_846_0_1.vlm"},
	{0x28074815FCE43DF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_force_crush.ani"},
	{0x28075F1C97E61CAFULL, "6132940_m.fxe"},
	{0x2807BC5C5A383501ULL, "650604521.ogg"},
	{0x2807D897DF3882D7ULL, "6093611_f.fxe"},
	{0x2807DE9E578E53B4ULL, "489358364.ogg"},
	{0x2807DF26C1639CF5ULL, "6033542_m.fxe"},
	{0x28082E4394CAC673ULL, "vxw_encounter1_fl_0_0.vlm"},
	{0x28088E1EB90887ACULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj03.msh"},
	{0x2808941EB908919EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj03.mmh"},
	{0x2808A53B6AA2CE85ULL, "follower_banter6136686.cl"},
	{0x2808D5BA3787ADBFULL, "6208034_m.fxe"},
	{0x2808FD92CA2FE1B0ULL, "vxlt_0_lowtown_a2_n_gmz_0_0.vlm"},
	{0x28091723F6B1EFF0ULL, "vxca_caverns1_imc_0_0.vlm"},
	{0x280946BC03F7E353ULL, "443389086.ogg"},
	{0x2809B42AF54430F5ULL, "vxlt_gallowsprison_n_6b4_0_0.vlm"},
	{0x2809F9E33E4302BCULL, "vxlt_backalley_g_1xa_0_0.vlm"},
	{0x280ADF4B8678D98DULL, "llac_lt_0_lowtown_a3_n_c_65537_6.dds"},
	{0x280B3E233848FE65ULL, "6148943_m.fxe"},
	{0x280B479CDDC77A6DULL, "6125164_m.fxe"},
	{0x280BCC8FCF437BDFULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr90pause.ani"},
	{0x280C4C5C805809F7ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic_5.mao"},
	{0x280CE653854568CFULL, "vxlt_0_lowtown_docks_n_c6q_0_0.vlm"},
	{0x280D06CB1061F323ULL, "vxw_blightlands_w1_0_0.vlm"},
	{0x280D6B4018426CBFULL, "132802883.ogg"},
	{0x280D8DF6CE6FC0B2ULL, "fen310_fenris6118708.cl"},
	{0x280DA566DA9C5077ULL, "vxca_caverns1_gez_0_0.vlm"},
	{0x280DB4CC6BB06384ULL, "483563252.ogg"},
	{0x280E245F2F838C1BULL, "6103218_m.fxe"},
	{0x280E5315FCEA6E93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_force_crush.gad"},
	{0x280ED99AE727F88BULL, "6060607_m.fxe"},
	{0x280F72B847900F08ULL, "mcr370_guardsman6087668.cl"},
	{0x280FA7BC384BE526ULL, "vxca_caverns1_don_0_0.vlm"},
	{0x280FFB5038EFF15CULL, "6050114_f.fxe"},
	{0x281004AA55C214C5ULL, "hlac_ht_0_hightown_a3_-65537_1y.dds"},
	{0x2810A7B2F817EF04ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_windows.mmh"},
	{0x2810ABB2F817F657ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_windows.mao"},
	{0x2810BDB2F81814A6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_windows.msh"},
	{0x2811305F65B7CB2EULL, "lgt260_eau_tmp_asst_harbor6143628.cl"},
	{0x28117C4BEB1966A7ULL, "vxlt_warehouse_192_0_0.vlm"},
	{0x281204B93DD0D4F7ULL, "279244252.ogg"},
	{0x28124BAAE016293FULL, "771019749.ogg"},
	{0x2812ECD05736E929ULL, "vxlt_darktown_5jd_0_0.vlm"},
	{0x2812F42DE168435CULL, "vxlt_0_lowtown_docks_n_cuk_0_0.vlm"},
	{0x28130B1A0F688CBEULL, "708649538.ogg"},
	{0x28130BC24000C197ULL, "hi_distor.mao"},
	{0x28130DDC2ADF17A6ULL, "vxlt_backalley_n_1x2_0_0.vlm"},
	{0x281320631EB67537ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zztent_0.pwk"},
	{0x28132B06878E1117ULL, "mag311_keran6063756.cl"},
	{0x2813520F8C773FC6ULL, "llac_dr_slavers_3215_9.dds"},
	{0x2813E00B7067CF5CULL, "vxht_mansion2b_1ku_0_0.vlm"},
	{0x2813E5A04E985D94ULL, "vxdr_primevaltomb_8z2_0_0.vlm"},
	{0x2814038292BC4EA9ULL, "docks_male_sailor_mob6176674.cl"},
	{0x28149675FE9F6EC6ULL, "qun110_petrice6042005.cl"},
	{0x2814F5CB3BC236C8ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg105a.phy"},
	{0x281513DD0EAF815EULL, "vxlt_0_lowtown_docks_c3h_0_0.vlm"},
	{0x281524B0ED8CD458ULL, "494398388.ogg"},
	{0x28155015FCF088F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_force_crush.evt"},
	{0x2815682FF9E8027EULL, "6173595_m.fxe"},
	{0x28159F9440CB760BULL, "vxht_bartrand_n_5po_0_0.vlm"},
	{0x2815EFB6410E5EFDULL, "6170078_m.fxe"},
	{0x281613DF44AFD914ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_dwpa_0t.dds"},
	{0x2816326F0E3AA2D0ULL, "mag320_bsc_banter6182382.cl"},
	{0x28165222EB5D50C6ULL, "vxht_0_hightown_a2_av4_0_0.vlm"},
	{0x281693C3777339F1ULL, "vxlt_0_lowtown_a2_n_c_bza_0_0.vlm"},
	{0x2816A02086C0DEF3ULL, "vxw_encounter1_ko_0_0.vlm"},
	{0x2816DA124FAB1B44ULL, "vxht_bartrand_n_6q8_0_0.vlm"},
	{0x2816EEABBBBB12E3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_rob_chaa_0s.dds"},
	{0x2817152CF1EC3602ULL, "6076295_f.fxe"},
	{0x2817159399B12D65ULL, "62100457.ogg"},
	{0x281744ABC517442BULL, "vxht_chantry_n_27n_0_0.vlm"},
	{0x2817585C5390EAAAULL, "m_fen200ar_ambush_exterior.bnk"},
	{0x2817BEA070915B27ULL, "6040700_m.fxe"},
	{0x28185441475B3654ULL, "vxlt_gallowstemplar_n_557_0_0.vlm"},
	{0x281855EC231FD393ULL, "mcr370cr_mage_messenger.utc"},
	{0x2818C9055A7F180CULL, "gen00fl_sebastian6174704.cl"},
	{0x2819B979CFBD92EDULL, "6122064_m.fxe"},
	{0x281A6D14E1E49423ULL, "zz_dog_debug6204132.cl"},
	{0x281A880CCB8DDFFBULL, "zz_approval_debug6133721.cl"},
	{0x281AAFF99F43115CULL, "gen00fl_varric6119005.cl"},
	{0x281AB3C437C3AE4BULL, "vxht_brothel_n_6f6_0_0.vlm"},
	{0x281AB3DA65DB95ADULL, "17786488.ogg"},
	{0x281AF4AE9FF76746ULL, "101045687.ogg"},
	{0x281AF5F95EF37DC5ULL, "vxdr_primevaltomb_7zv_0_2.vlm"},
	{0x281B78A802580BA3ULL, "vxw_blightlands_x0_0_0.vlm"},
	{0x281B951235F9F278ULL, "6167901_m.fxe"},
	{0x281B9B3B7D0B2F64ULL, "6148678_m.fxe"},
	{0x281BE99B2B7A1AE4ULL, "zz_followers_banter6157291.cl"},
	{0x281BEB6DB0705DEFULL, "art\\characters\\playercharacter\\humanfemale\\boo_maga.mao"},
	{0x281BF76DB070718CULL, "art\\characters\\playercharacter\\humanfemale\\boo_maga.mmh"},
	{0x281BFD6DB0707C7EULL, "art\\characters\\playercharacter\\humanfemale\\boo_maga.msh"},
	{0x281C03DE4B14B3C8ULL, "ser191_emeric6108145.cl"},
	{0x281C0C4E2FE60A8AULL, "w_bonepitmining_a2_119_0.mao"},
	{0x281C104E2FE611D9ULL, "w_bonepitmining_a2_119_0.mmh"},
	{0x281C1923D3B2D9BFULL, "character_character_p2.vsh.fxo.dx11"},
	{0x281C1A4E2FE622C7ULL, "w_bonepitmining_a2_119_0.msh"},
	{0x281C6457F4C197E1ULL, "hlac_w_sundermount_-131071_0z.dds"},
	{0x281CA980CB7B1B94ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyo_0t.dds"},
	{0x281CBA9D230CA9D7ULL, "112513097.ogg"},
	{0x281D6A4BE8A0DD0FULL, "907020245.ogg"},
	{0x281E9BD4F004EBEBULL, "hlac_lt_0_lowtown_docks_a3_n_65534_0z.dds"},
	{0x281EA546E1C2B341ULL, "dae200cr_viscount6087371.cl"},
	{0x281ECD631A9151ACULL, "6090240_f.fxe"},
	{0x281F161A395973BBULL, "vxlt_darktown_2re_0_0.vlm"},
	{0x281F3FD79F65E442ULL, "324142086.ogg"},
	{0x281FF3DFFB4BE96BULL, "6054576_m.fxe"},
	{0x282061BC26F8F82BULL, "ca_caverns1_24422_ll.mmh"},
	{0x282090E5CD8DB21CULL, "vxw_sundermount_2ep_0_0.vlm"},
	{0x28209AC07D9C9A48ULL, "870780295.ogg"},
	{0x2820DDA0C22B32F0ULL, "vxlt_warehouse_q0_0_0.vlm"},
	{0x282136A260DB5156ULL, "mrl150_merrill6085741.cl"},
	{0x28215EAD492D01A9ULL, "vxlt_darktown_5ok_0_0.vlm"},
	{0x2821EF1911A6C1B9ULL, "637327951.ogg"},
	{0x2822577D408E7614ULL, "vxdr_primevaltomb_8u3_0_0.vlm"},
	{0x28232C945F188AB7ULL, "vxht_0_hightown_c1j_0_0.vlm"},
	{0x28235BA09559594AULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_act_c.impa.gad"},
	{0x2823BEE20217403AULL, "526450194.ogg"},
	{0x2823E246C206D0F7ULL, "6106558_m.fxe"},
	{0x28241B4F86C1EB2CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v1.ani"},
	{0x28249C0B3412BB76ULL, "6149244_m.fxe"},
	{0x2825E74DC2DA56B8ULL, "685057216.ogg"},
	{0x282668F4B9480172ULL, "224412535.ogg"},
	{0x2826A41126E48382ULL, "hlac_lt_backalley_n_0_17y.dds"},
	{0x2826F15BB9912A6FULL, "789342213.ogg"},
	{0x2827648AF05B2523ULL, "6171202_m.fxe"},
	{0x2827AD2B4F54631FULL, "vxlt_0_lowtown_docks_a3_as4_0_0.vlm"},
	{0x2827BABE225D09A2ULL, "vxca_caverns1_e91_0_0.vlm"},
	{0x2827CB816C1E45F5ULL, "6141566_m.fxe"},
	{0x2827FA8CAA86277CULL, "5490336.ogg"},
	{0x282840AA44575E9FULL, "6174591_m.fxe"},
	{0x282853CFBBFA1D0FULL, "gen00fl_fenris6127012.cl"},
	{0x2828F7C83BDD1146ULL, "lt_0_lowtown_docks_a3_n_27_0.anb"},
	{0x282924D2689F17AEULL, "hlac_w_encounter3_-196608_1y.dds"},
	{0x282925A9E0E6EE69ULL, "6060155_m.fxe"},
	{0x28294F7C7BC85D0EULL, "6172527_m.fxe"},
	{0x28298CFC78381B95ULL, "vxht_0_hightown_n_bxe_0_0.vlm"},
	{0x282A07445E686505ULL, "550440205.ogg"},
	{0x282A3EA0955F46E1ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_act_c.impa.evt"},
	{0x282A9F96CE9400BDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicerchair.phy"},
	{0x282B284F86C8206EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v1.gad"},
	{0x282C3C408C7D66A5ULL, "6125465_m.fxe"},
	{0x282C5BE4C9ACBB85ULL, "hlac_w_bonepitmining_a2_4_0y.dds"},
	{0x282C674057A996F7ULL, "gen00fl_anders6119882.cl"},
	{0x282C900F74C0B809ULL, "vxdr_0_deeproads_etz_0_0.vlm"},
	{0x282CE24F9ED0BCF3ULL, "6143727_f.fxe"},
	{0x282D246D86D691E4ULL, "pro000st_opening_pc.stg"},
	{0x282DC9DB44A89C68ULL, "ser221_gascard6134513.cl"},
	{0x282E3EA7D9F30591ULL, "hlac_lt_0_lowtown_a3_n_c_131070_2z.dds"},
	{0x282EAE8F2C07EDE8ULL, "hlac_lt_undercityencounter_-1_16y.dds"},
	{0x282EF17CC92FCD27ULL, "594415102.ogg"},
	{0x282F02C2BF83CA9CULL, "vxw_sundermount_2nq_0_0.vlm"},
	{0x282F0E40594BBFF2ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_dothard.mao"},
	{0x282F4F7DB4214B70ULL, "vxlt_warehouse_z1_0_0.vlm"},
	{0x282FC731B956F95EULL, "745503671.ogg"},
	{0x28300DA68B58737CULL, "human_male_heads.gda"},
	{0x28303EE0AF08831BULL, "vxlt_gallowsprison_n_6c9_0_0.vlm"},
	{0x2830CD4B8DDB6A84ULL, "hlac_ht_0_hightown_a2_-65536_3x.dds"},
	{0x2830D5C22B3607BCULL, "da2terrain_terrain_p0.vsh.fxo.dx11"},
	{0x28314B611384CAB4ULL, "23436055.ogg"},
	{0x283171346FED43CEULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga104a_0t.dds"},
	{0x283192DC17BD4152ULL, "6172729_m.fxe"},
	{0x28319E71510EA337ULL, "vxht_0_hightown_c4e_0_0.vlm"},
	{0x2831CD720DC7CDC5ULL, "881857848.ogg"},
	{0x2832234F86CE364DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v1.evt"},
	{0x2832479B89E13793ULL, "blood_conj_03.mao"},
	{0x283273973A7B9D03ULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\prp_blightrock_sa.dds"},
	{0x2832DBB89EC46DCBULL, "975866617.ogg"},
	{0x2832FFE84950BC80ULL, "6162605_m.fxe"},
	{0x2833717C8D0C9518ULL, "94590951.ogg"},
	{0x2834D37A0AF5E7C0ULL, "6217569_m.fxe"},
	{0x28357CD814BD3D60ULL, "316327123.ogg"},
	{0x283582C0F40D0C95ULL, "998237289.ogg"},
	{0x2835DAAA63AE1591ULL, "6149575_m.fxe"},
	{0x2835E0EE36FBA13CULL, "plt_cod_bks_kirkwall_hist_4.plo"},
	{0x2835F7DEA3627258ULL, "hlac_ht_chantry_2534_29z.dds"},
	{0x28365247541580C4ULL, "ca_caverns2prc_mmap.dds"},
	{0x28365CB16C56D36CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ands.msh"},
	{0x283662B16C56DD5EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ands.mmh"},
	{0x2836D135C1AAB5ADULL, "611228221.ogg"},
	{0x28377F991DAB4ADEULL, "dae000_templar_recruit_16164433.cl"},
	{0x283836A0956B7630ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_act_c.impa.ani"},
	{0x283921664972EA34ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a_1.mao"},
	{0x28395E88F68F0835ULL, "qun230ip_to_darktown.utp"},
	{0x2839C3AE450396CCULL, "6018812_m.fxe"},
	{0x2839D804C885177DULL, "231699029.ogg"},
	{0x283A0ADD449BBE92ULL, "83363597.ogg"},
	{0x283A81F867D7CB1FULL, "6149340_m.fxe"},
	{0x283B9496CEA26129ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicerchair.mao"},
	{0x283BA096CEA275C2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicerchair.mmh"},
	{0x283BA296CEA27978ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicerchair.msh"},
	{0x283BD2BC8D64A304ULL, "6096524_m.fxe"},
	{0x283BD7C534F00D2DULL, "vxlt_gallowstemplar_5ux_0_1.vlm"},
	{0x283BF9D41D45A4BBULL, "vxlt_darktown_2xg_0_0.vlm"},
	{0x283BFF8E7D9E1DAEULL, "vxlt_gallowstemplar_n_3sm_0_0.vlm"},
	{0x283C13E5ACD5250FULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ca_basic_2.mao"},
	{0x283C5A35A694B15FULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.dura.gad"},
	{0x283C6935ADBE750DULL, "286591592.ogg"},
	{0x283D749FB179E31CULL, "vxw_sundermount_2sv_0_0.vlm"},
	{0x283D7CB242630230ULL, "hlac_w_woundedcoast_-262136_0x.dds"},
	{0x283D866F52CFADD1ULL, "hlac_lt_0_lowtown_a2_n_c_65537_5z.dds"},
	{0x283EA01C2422B5ACULL, "vxlt_0_lowtown_a3_n_c_ekl_0_1.vlm"},
	{0x283EC44FF270F38FULL, "vrc221_varric6040872.cl"},
	{0x283F15F9104D7C8FULL, "vxht_0_hightown_a3_aok_0_0.vlm"},
	{0x283F1D9C3F95091DULL, "vxlt_0_lowtown_docks_cp2_0_0.vlm"},
	{0x283F7DEFE8CC88A1ULL, "follower_banter6156158.cl"},
	{0x283FF46DD3F63DE8ULL, "vxlt_gallowscourtyard_4qk_0_0.vlm"},
	{0x2840039ABD5F7F43ULL, "mrl150_merrill6093532.cl"},
	{0x284012F58CE5DF85ULL, "6161234_m.fxe"},
	{0x284040A7990C89B3ULL, "vxw_encounter3_fw_0_1.vlm"},
	{0x2840824D28D2F28EULL, "art\\vfx\\combat\\vfx_d_single_att_c.mmh"},
	{0x2840DBA402FFB7EAULL, "6141613_m.fxe"},
	{0x2840E9766816CC7DULL, "903404461.ogg"},
	{0x2841EE72D9E0E300ULL, "749390473.ogg"},
	{0x2842497C48E0BC9BULL, "mag291_demon_marethari6027850.cl"},
	{0x2842553183223F4FULL, "llac_w_blightlands_327677_0.dds"},
	{0x284259FF18367EECULL, "gen00fl_aveline6120754.cl"},
	{0x284310655B0B99F8ULL, "6206273_m.fxe"},
	{0x28431DA365691D4FULL, "gen00fl_sebastian6177898.cl"},
	{0x28433F327C4735F6ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_tmpa.mao"},
	{0x28434F35A69ABD4CULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.dura.evt"},
	{0x2843623FEA6EE13DULL, "zz_followers_banter6159531.cl"},
	{0x28437A57825EBC69ULL, "987580817.ogg"},
	{0x2843D48FDF8FA2C9ULL, "hlac_lt_gallowstemplar_f_2388_4y.dds"},
	{0x2843E43D33AD4525ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comb_0t.dds"},
	{0x2844011BE1301ADCULL, "190900557.ogg"},
	{0x28444C8755A42E11ULL, "6142475_m.fxe"},
	{0x28445904B77C9949ULL, "6057038_m.fxe"},
	{0x2845708023801922ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_basic.msh"},
	{0x2845728023801CD8ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_basic.mmh"},
	{0x28457E802380317BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_basic.mao"},
	{0x28458993BA8C1170ULL, "459750990.ogg"},
	{0x2845A5A439A43B89ULL, "vxlt_warehouse_114_0_0.vlm"},
	{0x28468F166A8793B0ULL, "hlac_w_encounter2_camp_131071_0y.dds"},
	{0x28475C17E6CA3DEBULL, "rdr111dg_martin6055043.cl"},
	{0x2847CADE3E97E9E8ULL, "6101224_m.fxe"},
	{0x2847D94A6BB9A65EULL, "6189899_m.fxe"},
	{0x2848B600A53C8290ULL, "255347754.ogg"},
	{0x2848F9097B0BA315ULL, "6080508_m.fxe"},
	{0x2848FC4B76B8A04CULL, "vxht_mansion2b_3f0_0_0.vlm"},
	{0x2849140A899ABB3AULL, "6045243_m.fxe"},
	{0x28496A559094A4E2ULL, "llac_ca_cavern1_haunted_24426_2.dds"},
	{0x28496B1EB93FE339ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj03.phy"},
	{0x2849D21B02346B49ULL, "507801651.ogg"},
	{0x284A437017B3120BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_fnzy.ani"},
	{0x284A569B89F5E329ULL, "blood_conj_03.dds"},
	{0x284A7D952DBF21ADULL, "hlac_ht_mansion2b_818_50x.dds"},
	{0x284A8867BDFB2BB5ULL, "hlac_ht_mansion_n_3381_0z.dds"},
	{0x284ACD0B5FFE7EACULL, "6172741_m.fxe"},
	{0x284B40557B4B60B8ULL, "6125631_m.fxe"},
	{0x284BB612F2117F2FULL, "117972181.ogg"},
	{0x284BE2CDA4658692ULL, "152493370.ogg"},
	{0x284BE67CA36FFB9CULL, "vxw_sundermount_2tw_0_0.vlm"},
	{0x284BEDE337D00DD9ULL, "6115880_m.fxe"},
	{0x284BFA9EB5338EE3ULL, "686821077.ogg"},
	{0x284C0BF9297C2BF6ULL, "vxht_chantry_21b_0_0.vlm"},
	{0x284C41D1EDCE50E1ULL, "vxlt_gallowscourtyard_n_6xl_0_0.vlm"},
	{0x284C87A91AD1BDA3ULL, "549900611.ogg"},
	{0x284C8F84677886FDULL, "601159898.ogg"},
	{0x284D1AE0C8756A58ULL, "111653162.ogg"},
	{0x284D9216BFD36BFDULL, "follower_banter6163136.cl"},
	{0x284DE5017E767A16ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp01.mmh"},
	{0x284DEF017E768B14ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp01.msh"},
	{0x284E10BFCECDF2DEULL, "plt_cod_aow_crafting.plo"},
	{0x284E691F62EC629AULL, "hlac_lt_0_lowtown_a2_n_131070_2z.dds"},
	{0x284E84D27EDBF805ULL, "6166334_m.fxe"},
	{0x284EA14D28DF648DULL, "art\\vfx\\combat\\vfx_d_single_att_c.anb"},
	{0x284ED64F7CE3C96AULL, "6161259_m.fxe"},
	{0x284F3DFD40A3260FULL, "vxht_0_hightown_car_0_0.vlm"},
	{0x28503735A6A51EA5ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.dura.ani"},
	{0x285051DF357F2BB5ULL, "440576241.ogg"},
	{0x285055CF2B9A3BE6ULL, "vxht_0_hightown_a3_7g4_0_0.vlm"},
	{0x285084E76B39722FULL, "lt_0_lowtown_docks_121_0.phy"},
	{0x285099A0EBF0F8FAULL, "dae000_keep_bureaucrat_1_st6165313.cl"},
	{0x2850B3BE58793CCBULL, "art\\characters\\weapons\\axes\\w_axe_103a.mao"},
	{0x2850D8D597A061BDULL, "art\\levels\\races\\design\\prp_spindleweed.mao"},
	{0x2850DCD597A068F6ULL, "art\\levels\\races\\design\\prp_spindleweed.mmh"},
	{0x2850E6D597A079F4ULL, "art\\levels\\races\\design\\prp_spindleweed.msh"},
	{0x2850EC9DC4A9ADF7ULL, "vxlt_darktown_2rq_0_0.vlm"},
	{0x2851107DE6F2D8A9ULL, "6085275_m.fxe"},
	{0x285121924AE0B3A4ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_105a_0t.dds"},
	{0x285141E26C6DBE17ULL, "vxw_woundedcoast_1bm_0_0.vlm"},
	{0x2851B8718FF5A643ULL, "416541381.ogg"},
	{0x2851BA944D15B3BCULL, "zz_trailer_shot2.cut"},
	{0x28522C6A52C8D3CFULL, "hlac_lt_0_lowtown_65535_2z.dds"},
	{0x28522F7B653A3737ULL, "gen00fl_varric6076541.cl"},
	{0x2852CAFBC1AB7171ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_009a_0s.dds"},
	{0x2852DD44325F2F41ULL, "vxw_encounter2_4m_0_0.vlm"},
	{0x285303AFB512705EULL, "140030655.ogg"},
	{0x2853301073C3AEA9ULL, "64330924.ogg"},
	{0x285346AE566DFAF1ULL, "lgt160pt_cod_avern_lett_2.utp"},
	{0x28535A09F1716742ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb11a.phy"},
	{0x285399DBEE073AB1ULL, "6153194_m.fxe"},
	{0x2853A0A68FB05A18ULL, "vxdr_primevaltomb_3i8_0_0.vlm"},
	{0x2853F17F6E09F161ULL, "390208432.ogg"},
	{0x2854003B86E8EFA1ULL, "llac_lt_0_lowtown_docks_a3_-196610_2.dds"},
	{0x28541133454BC76FULL, "vxht_0_hightown_n_al9_0_1.vlm"},
	{0x2854BE7C0EF87C56ULL, "zz_isa_debug6133946.cl"},
	{0x28553C90E6B727DCULL, "vxlt_0_lowtown_a2_n_gwd_0_0.vlm"},
	{0x28558DFB5F6E9C2FULL, "6125557_m.fxe"},
	{0x2855A288E9DDA4B1ULL, "vxht_mansion2_27w_0_0.vlm"},
	{0x2855C05ECC53245BULL, "6053595_m.fxe"},
	{0x2855FAD450D552B1ULL, "hlac_lt_0_lowtown_n_-131072_4x.dds"},
	{0x28563238FE4AC560ULL, "vxdr_primevaltomb_77r_0_0.vlm"},
	{0x28566A5CF51E5C19ULL, "6202360_m.fxe"},
	{0x2856929B5DF6BACAULL, "1012577766.ogg"},
	{0x285704954441C484ULL, "one000_ilen6205703.cl"},
	{0x28577A4DA8EEA801ULL, "war_elementalbonus.dds"},
	{0x2857CFD4B5DBB6B4ULL, "vxlt_darktown_58b_0_1.vlm"},
	{0x28582B7017BF2682ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_fnzy.evt"},
	{0x2858B5E010DA6CCBULL, "vxw_sundermount_279_0_0.vlm"},
	{0x28595FFA7451C953ULL, "mag360_samson6062126.cl"},
	{0x28599873F7C3FFE3ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_wic.mao"},
	{0x285A58599566141CULL, "vxw_sundermount_2yt_0_0.vlm"},
	{0x285A942CCD3828C8ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_1y.dds"},
	{0x285B0B4A376755E5ULL, "vxht_0_hightown_n_bx5_0_0.vlm"},
	{0x285B19CDFB255E99ULL, "vxht_mansion2b_2wg_0_0.vlm"},
	{0x285C5A58E6B9C783ULL, "mag321ar_mansion.ncs"},
	{0x285CB5576E668DDCULL, "follower_banter6141706.cl"},
	{0x285D1B721A51ED6DULL, "dae300_bran6143523.cl"},
	{0x285DD8E4FA99F08BULL, "vxht_bartrand_n_5ob_0_0.vlm"},
	{0x285E1BBD3FF31625ULL, "one111ip_ggt_crate_5.utp"},
	{0x285E475900A33433ULL, "hlac_ht_chantry_n_2534_36x.dds"},
	{0x285E4FEC130C5D77ULL, "671642034.ogg"},
	{0x285EA5A719508EC0ULL, "6099187_m.fxe"},
	{0x285EBD9832FD4A07ULL, "vxht_0_hightown_a2_c6z_0_1.vlm"},
	{0x285EF397DA49D435ULL, "hlac_lt_0_lowtown_a3_n_c_131071_7z.dds"},
	{0x285F174B0163076EULL, "vxca_caverns1_mines_hkk_0_0.vlm"},
	{0x285F2A7017C54395ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_tal_fnzy.gad"},
	{0x285F3D1936FCD2D9ULL, "6147641_m.fxe"},
	{0x285F5E7AB69FC677ULL, "vxlt_darktown_2up_0_0.vlm"},
	{0x285FB3BF5E63D697ULL, "vxw_woundedcoast_1ml_0_0.vlm"},
	{0x285FEB9A525555BEULL, "w_sundermount_103_0.rml"},
	{0x285FFB11AB95A27DULL, "ser120_ghyslain6104188.cl"},
	{0x28601994B551D7F3ULL, "mag211_marethari6025157.cl"},
	{0x2860314EAE28FB59ULL, "hlac_ht_chantry_2534_13z.dds"},
	{0x2860B8870647FF4FULL, "art\\vfx\\combat\\vfx_m_arc_bolt_db_c.impa.gad"},
	{0x2860E386599BA8AEULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.cess.ani"},
	{0x2861F9D76EC81CD0ULL, "vxht_0_hightown_n_4kg_0_0.vlm"},
	{0x286259EBD8B67B3CULL, "6175094_m.fxe"},
	{0x28632E132EAA9688ULL, "dae000ip_crowd_mf8_drunk2.utp"},
	{0x28635A35831F0928ULL, "6059381_m.fxe"},
	{0x28635A58E6BFE6EAULL, "mag321ar_mansion.lst"},
	{0x28635E3960289086ULL, "pro000_athenril6049446.cl"},
	{0x28636DFECC1EFED6ULL, "6054490_m.fxe"},
	{0x2863915A8971A106ULL, "60392505.ogg"},
	{0x2863F71A179903D2ULL, "386421536.ogg"},
	{0x2863FB4D98CEC09BULL, "6013249_m.fxe"},
	{0x2863FFD85164B4AFULL, "6122457_m.fxe"},
	{0x28641465DBD3BD31ULL, "vxht_mansion2_20v_0_0.vlm"},
	{0x2864617D35242D9FULL, "vxlt_0_lowtown_docks_a3_bfx_0_0.vlm"},
	{0x2864BE2D00573614ULL, "gen00fl_anders6103668.cl"},
	{0x2865285A095CE392ULL, "lgt000cr_foreshadow.utc"},
	{0x2866049A5259EB82ULL, "w_sundermount_103_0.phy"},
	{0x286727BD02D0854BULL, "vxw_sundermount_288_0_0.vlm"},
	{0x2867CD87064E41FCULL, "art\\vfx\\combat\\vfx_m_arc_bolt_db_c.impa.evt"},
	{0x2867FB424221BE00ULL, "vxlt_0_lowtown_docks_n_c48_0_0.vlm"},
	{0x286831F7C5D6D631ULL, "37223619.ogg"},
	{0x28684D371DCF6EAEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_drag_flemeth_fake.ani"},
	{0x28684D492431F6D7ULL, "w_bonepitmining_a2_221_0.mmh"},
	{0x286851492431FDE4ULL, "w_bonepitmining_a2_221_0.mao"},
	{0x286854BC1A06AD96ULL, "6108786_m.fxe"},
	{0x2868634924321C49ULL, "w_bonepitmining_a2_221_0.msh"},
	{0x28688653AD34FB58ULL, "6167066_m.fxe"},
	{0x2868E378ADD785D5ULL, "llac_ht_chantry_2534_37.dds"},
	{0x286934C39C3B1081ULL, "vxlt_0_lowtown_a2_n_gz6_0_0.vlm"},
	{0x28694F4690FB4E2AULL, "6143479_m.fxe"},
	{0x28697D27295D6E65ULL, "vxht_0_hightown_n_bc4_0_0.vlm"},
	{0x2869D05226B1749CULL, "plt_dae000_mother.nss"},
	{0x286A2457558343A8ULL, "vxlt_gallowsprison_n_6ad_0_0.vlm"},
	{0x286A5F4190F3D1D1ULL, "vxlt_0_lowtown_n_h83_0_0.vlm"},
	{0x286A85547A13ED64ULL, "6113911_m.fxe"},
	{0x286AA47A8F6C11C4ULL, "hlac_lt_0_lowtown_docks_a3_65533_3x.dds"},
	{0x286AE3629BD9DE34ULL, "vxlt_house_fv_0_0.vlm"},
	{0x286BE540BA7FC9C4ULL, "6118763_m.fxe"},
	{0x286C5E295C5C44E2ULL, "470848933.ogg"},
	{0x286CFFBE907505DDULL, "partycombat2.cut"},
	{0x286D95886A116C36ULL, "6134907_m.fxe"},
	{0x286DD057A895DEF7ULL, "vxlt_darktown_2xs_0_0.vlm"},
	{0x286DE68BF97E0898ULL, "vxht_brothel_n_4b7_0_0.vlm"},
	{0x286DF8CDB8BA1387ULL, "389396555.ogg"},
	{0x286E259C5059EF17ULL, "vxw_woundedcoast_1ho_0_0.vlm"},
	{0x286E342E946C9B70ULL, "vxca_caverns1_i2j_0_0.vlm"},
	{0x286E8AFADC51AA0EULL, "776303455.ogg"},
	{0x286E9F141C3860E2ULL, "hlac_lt_backalley_g_65537_8x.dds"},
	{0x286EDB8659A7D8FFULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.cess.evt"},
	{0x286EEB1D7263D591ULL, "275464302.ogg"},
	{0x286F322E4D272623ULL, "hlac_ht_0_hightown_-65536_2z.dds"},
	{0x286FB71D4DFE25EEULL, "mrl210_merrill6116390.cl"},
	{0x286FBC52897A50E3ULL, "hlac_ht_0_hightown_a3_-65537_0x.dds"},
	{0x287041FE61F4EFB3ULL, "6080205_m.fxe"},
	{0x2870EC00C91970ECULL, "6092573_m.fxe"},
	{0x2871676FB645CF16ULL, "6084683_f.fxe"},
	{0x2871A9A83CF4FD6FULL, "6094907_m.fxe"},
	{0x287434828C4961EBULL, "801616025.ogg"},
	{0x28744C6FF3229060ULL, "zz_followers_banter6159363.cl"},
	{0x28748E858A59570FULL, "llac_lt_0_lowtown_docks_-131075_3.dds"},
	{0x2874D5870658D915ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_db_c.impa.ani"},
	{0x2874D760948521AAULL, "vxht_brothel_n_62o_0_0.vlm"},
	{0x28751F44CF8E47E9ULL, "gen00fl_dog6115100.cl"},
	{0x2875BC8659ADC2E8ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.cess.gad"},
	{0x287639B16C8C7BF9ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ands.phy"},
	{0x28769AFF7900DE96ULL, "follower_banter6170627.cl"},
	{0x28769FF9641F73D3ULL, "6021239_m.fxe"},
	{0x2876BB7968B68A10ULL, "vxht_mansion2_1kq_0_0.vlm"},
	{0x2876CFDEB21FB4A4ULL, "lt_0_lowtown_docks_a3_n_246_0.msh"},
	{0x2876E1DEB21FD30DULL, "lt_0_lowtown_docks_a3_n_246_0.mao"},
	{0x2876E5DEB21FDA46ULL, "lt_0_lowtown_docks_a3_n_246_0.mmh"},
	{0x287743F323E7A272ULL, "llac_ht_0_hightown_a2_-65536_3.dds"},
	{0x28774CA43E282299ULL, "dae000_mother6136915.cl"},
	{0x28775114797A0CBBULL, "vxlt_undercityencounter_1fx_0_0.vlm"},
	{0x28776DD8178337F2ULL, "zz_followers_banter6157565.cl"},
	{0x2878036323216020ULL, "plt_fen200pt_night_terrors.nss"},
	{0x28784109F1905895ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb11a.mmh"},
	{0x28784309F1905C0BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb11a.msh"},
	{0x2878963447795C28ULL, "vxlt_gallowsprison_n_6je_0_0.vlm"},
	{0x2878FF3A89700376ULL, "llac_lt_0_lowtown_n_131070_0.dds"},
	{0x287934410FDD62C6ULL, "vxht_brothel_n_6gl_0_0.vlm"},
	{0x28796B4DFE7ED673ULL, "471378179.ogg"},
	{0x2879B25145C2136BULL, "hlac_lt_undercityserialkiller_-1_23x.dds"},
	{0x287A07A43EFC567EULL, "6135852_m.fxe"},
	{0x287A160AE3275425ULL, "645877659.ogg"},
	{0x287A1DD55C0356A7ULL, "200297472.ogg"},
	{0x287A9019FE3A20E9ULL, "vxht_mansion2_1qz_0_0.vlm"},
	{0x287AD5374C0CA3A3ULL, "w_encounter3_100_0.mao"},
	{0x287AD9374C0CAAF0ULL, "w_encounter3_100_0.mmh"},
	{0x287AE7374C0CC20AULL, "w_encounter3_100_0.msh"},
	{0x287B17A3A9EB72D9ULL, "ser120_ghyslain6104175.cl"},
	{0x287B305CA7E45425ULL, "dog321ip_meat.utp"},
	{0x287B38B57026AEFAULL, "vxht_bartrand_n_6pa_0_0.vlm"},
	{0x287B5E3B9704697FULL, "6191873_m.fxe"},
	{0x287BCB74B1B611AEULL, "propalpha1_s1.psh.fxo"},
	{0x287BD5B80CE032A0ULL, "924459874.ogg"},
	{0x287C06651D5950A1ULL, "6185232_m.fxe"},
	{0x287C3712A3FB5EE7ULL, "hlac_lt_0_lowtown_docks_-131073_1x.dds"},
	{0x287C5868EB742118ULL, "vxht_brothel_n_4g4_0_0.vlm"},
	{0x287D26371DE188E8ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_drag_flemeth_fake.gad"},
	{0x287D70B2B40CE05AULL, "vxht_0_hightown_n_9zk_0_0.vlm"},
	{0x287D8FD7D5E04C42ULL, "402864415.ogg"},
	{0x287E5488DCC65C22ULL, "6051394_m.fxe"},
	{0x287ECA915DD97735ULL, "gen_im_npc_arm_hum_glv_slva.uti"},
	{0x287F114E303A90C6ULL, "w_bonepitmining_a2_119_0.phy"},
	{0x287F93F132F6A2B5ULL, "hlac_ca_caverns1_24424_13x.dds"},
	{0x28808F918A02A061ULL, "238116266.ogg"},
	{0x2880AD173B3A7DF7ULL, "6048782.ogg"},
	{0x2880CCD17E760C55ULL, "vxlt_0_lowtown_n_fer_0_0.vlm"},
	{0x2880D332E0A814C1ULL, "412688356.ogg"},
	{0x2880E9E96C2A0041ULL, "llac_ca_caverns2_5802_11.dds"},
	{0x2880F19E32C90CADULL, "6209871_m.fxe"},
	{0x28810C33FCE9008FULL, "336021994.ogg"},
	{0x288183208084C540ULL, "802694218.ogg"},
	{0x28821D802EDA1B20ULL, "llac_lt_darktown_-1_24.dds"},
	{0x28823EE1FF596CC7ULL, "hlac_lt_ruinedhouse_n_-1_3z.dds"},
	{0x28828FDCBDFBEF5CULL, "hlac_ca_caverns1_mines_24424_15z.dds"},
	{0x2882B7A58F699056ULL, "hlac_ht_store_187_2y.dds"},
	{0x2882D7716C69CF54ULL, "vxlt_0_lowtown_docks_n_b6s_0_0.vlm"},
	{0x28835B868FB6D162ULL, "10588219.ogg"},
	{0x2883951B0D7C0738ULL, "lt_0_lowtown_docks_a3_n_30_0_ll.mmh"},
	{0x288399DB1C1CAA59ULL, "211171653.ogg"},
	{0x28845711F6A850BAULL, "6149510_m.fxe"},
	{0x2885099698D4E53EULL, "140018763.ogg"},
	{0x288549D0F057B490ULL, "106928055.ogg"},
	{0x2885554FF279FBCEULL, "vxlt_gallowstemplar_4o5_0_0.vlm"},
	{0x288598C03CDA79E0ULL, "16455827.ogg"},
	{0x28860532ED9AEC84ULL, "sst_ss_crow_m_1.gda"},
	{0x28864333A392C39BULL, "vxlt_0_lowtown_a2_n_h1v_0_0.vlm"},
	{0x28865CA977AA7838ULL, "vxdr_primevaltomb_26p_0_0.vlm"},
	{0x28866F64D107A819ULL, "vxht_mansion2b_2fd_0_0.vlm"},
	{0x288698094067AAC3ULL, "hlac_lt_0_lowtown_docks_-131074_6x.dds"},
	{0x28872533E578FBD7ULL, "vxx_narrativeroom_3m_0_0.vlm"},
	{0x28875D6D8CEF6956ULL, "dae000ip_crowd_mf7_talk12.utp"},
	{0x28878DAD12664F15ULL, "vxw_encounter3_7s_0_0.vlm"},
	{0x2887C336A749D6B1ULL, "mcr310_party_banter6087819.cl"},
	{0x288805AC940C01A0ULL, "vxca_caverns1_mines_ii6_0_0.vlm"},
	{0x28885958E6DF014DULL, "mag321ar_mansion.are"},
	{0x2888651FD22585B1ULL, "6093532_f.fxe"},
	{0x2888897CAA02D79BULL, "181250538.ogg"},
	{0x2888BC4EA79609C7ULL, "6134982_m.fxe"},
	{0x288938730B9917C2ULL, "w_woundedcoast_69_0.msh"},
	{0x28893A730B991B78ULL, "w_woundedcoast_69_0.mmh"},
	{0x288946730B992F1BULL, "w_woundedcoast_69_0.mao"},
	{0x288970C021E0C286ULL, "pro000ip_gate.utp"},
	{0x288A99593F86CF92ULL, "vxht_0_hightown_b27_0_0.vlm"},
	{0x288B5A2B3E7C0265ULL, "vxht_0_hightown_n_9aq_0_0.vlm"},
	{0x288B5BBED776ABEBULL, "17179091.ogg"},
	{0x288BCDA3C6FEA87AULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgb.phy"},
	{0x288BDF9A527A7C36ULL, "w_sundermount_103_0.mao"},
	{0x288BEB9A527A90D5ULL, "w_sundermount_103_0.mmh"},
	{0x288BED9A527A944BULL, "w_sundermount_103_0.msh"},
	{0x288C2198FD0BB392ULL, "vxdr_0_deeproads_cuq_0_1.vlm"},
	{0x288C67BE09980B4DULL, "894870935.ogg"},
	{0x288CE5949ED0B87BULL, "20048128.ogg"},
	{0x288CF814247BC5C1ULL, "vxw_blightlands_rt_0_0.vlm"},
	{0x288D13C25768D548ULL, "gen00fl_anders6125835.cl"},
	{0x288D27061748670DULL, "hlac_dr_0_deeproads_5840_4z.dds"},
	{0x288D3A730B9D038BULL, "w_woundedcoast_69_0.rml"},
	{0x288D5CD5F4D596BBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwstraight.phy"},
	{0x288E89B894849994ULL, "vxht_brothel_n_59c_0_0.vlm"},
	{0x288ED2784E2EFF65ULL, "189739174.ogg"},
	{0x288EDE823D23D2C5ULL, "247714721.ogg"},
	{0x288EFAC835978555ULL, "hf_lgt_soph_cultist.mor"},
	{0x288F3F46C27D1834ULL, "vxdr_primevalentrance_2vz_0_2.vlm"},
	{0x288F4AE883B1ABFFULL, "llac_ca_caverns1_24423_10.dds"},
	{0x288F6BB702920396ULL, "hlac_lt_0_lowtown_a2_n_0_0z.dds"},
	{0x288F86C2D7B66A61ULL, "lt_foundry_n.arl"},
	{0x2890155D982D9F3AULL, "follower_banter6156263.cl"},
	{0x289091A821806DF1ULL, "230439121.ogg"},
	{0x2890B434B374ACCAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkleft.gad"},
	{0x28925C03FF805596ULL, "w_bonepitmining_a2_96_0.mao"},
	{0x28925C201AC6C4B3ULL, "476031761.ogg"},
	{0x28926803FF806A35ULL, "w_bonepitmining_a2_96_0.mmh"},
	{0x28926A03FF806DABULL, "w_bonepitmining_a2_96_0.msh"},
	{0x28929DCE03D73C8BULL, "825368469.ogg"},
	{0x2892ABD71C3288A6ULL, "185581899.ogg"},
	{0x2892C8EEE89E693AULL, "6142410_m.fxe"},
	{0x2892E172A406C35FULL, "6176621_m.fxe"},
	{0x2893A8DB84CCA6E9ULL, "w_sundermount_85_0.phy"},
	{0x2893A95A1FAC50B2ULL, "dae110ip_at_hightown.utp"},
	{0x2893C60B897FDAC9ULL, "hor_smo.mao"},
	{0x2893F606F1ABEC2AULL, "vxht_0_hightown_a2_av8_0_0.vlm"},
	{0x28941849D62CCD03ULL, "lla_ht_0_hightown_n_65537_0.dds"},
	{0x2894DB8AB8D78559ULL, "art\\characters\\weapons\\greataxes\\w_gax_002a.msh"},
	{0x2894E58AB8D79647ULL, "art\\characters\\weapons\\greataxes\\w_gax_002a.mmh"},
	{0x2894E98AB8D79C94ULL, "art\\characters\\weapons\\greataxes\\w_gax_002a.mao"},
	{0x28957B7989CC0C46ULL, "art\\characters\\creatures\\mubari\\animation\\com_barking.ani"},
	{0x2897813F88C4EDECULL, "6133315_m.fxe"},
	{0x28979734B37A9A61ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkleft.evt"},
	{0x289831FF43191DA8ULL, "56284391.ogg"},
	{0x28984BC2E5415D56ULL, "gen_im_npc_arm_hum_cht_cira.uti"},
	{0x28984E6069A918E8ULL, "vxlt_0_lowtown_a2_fhk_0_0.vlm"},
	{0x28993E77439301DBULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.cess.ani"},
	{0x2899B832394963DFULL, "126808608.ogg"},
	{0x289A33B4987AFA6DULL, "vxw_bonepitdestroyed_a3_bo_0_0.vlm"},
	{0x289A5E2D2FFAD945ULL, "hlac_ca_cave_2194_6x.dds"},
	{0x289A66266BC59651ULL, "art\\cinematics\\deathblows\\rockgolem\\db_golem_roskwarr.gad"},
	{0x289AB1DB84D2D4E5ULL, "w_sundermount_85_0.rml"},
	{0x289ABDCD1E5DEFD5ULL, "llac_lt_0_lowtown_a2_65535_0.dds"},
	{0x289AEE290A364E73ULL, "lla_ht_0_hightown_a3_65535_0.dds"},
	{0x289B14D423F31BDFULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medd_n.dds"},
	{0x289B69F11671DE41ULL, "vxw_blightlands_wk_0_0.vlm"},
	{0x289B79F15A3A11F1ULL, "6168321_m.fxe"},
	{0x289B7D089363D60DULL, "6055152_f.fxe"},
	{0x289C49A3E1BDE314ULL, "vxht_0_hightown_a2_c7k_0_0.vlm"},
	{0x289D1A643226E861ULL, "hlac_lt_0_lowtown_131071_7x.dds"},
	{0x289D32C94FE4DBD7ULL, "vxlt_gallowscourtyard_6xu_0_0.vlm"},
	{0x289D7B22ECD10F23ULL, "hlac_ca_caverns2prc_5802_39y.dds"},
	{0x289DB5B8BE89C101ULL, "6098844_m.fxe"},
	{0x289E267CE47498F8ULL, "810590759.ogg"},
	{0x289E7C8603E44223ULL, "6171413_m.fxe"},
	{0x289EC187A0F051D4ULL, "618953608.ogg"},
	{0x289EDBE34A309163ULL, "vxdr_0_deeproads_bp3_0_0.vlm"},
	{0x289EE9EC675E9848ULL, "vxw_encounter3_f8_0_1.vlm"},
	{0x289F031DF946E6F3ULL, "6041387_m.fxe"},
	{0x289F2C71A83D7D39ULL, "6055018_m.fxe"},
	{0x289F32F5CA18319EULL, "vxht_mansion2b_3bn_0_0.vlm"},
	{0x289FEFB38FE43B3AULL, "6015840_m.fxe"},
	{0x28A067266BCA03FFULL, "art\\cinematics\\deathblows\\rockgolem\\db_golem_roskwarr.ani"},
	{0x28A1364E7DF9FA5BULL, "vxw_bonepitmining_a2_1lf_0_0.vlm"},
	{0x28A152690CBBB3D5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_flma.mao"},
	{0x28A2F2033F1FD32AULL, "vxw_encounter2_kw_0_0.vlm"},
	{0x28A302BDCBB50B6CULL, "vxlt_0_lowtown_a2_n_c_d2n_0_0.vlm"},
	{0x28A3543BA00683A5ULL, "6125616_m.fxe"},
	{0x28A3637989D820B7ULL, "art\\characters\\creatures\\mubari\\animation\\com_barking.evt"},
	{0x28A3B2D84D7978B9ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_2z.dds"},
	{0x28A44F0C89538352ULL, "6030608_f.fxe"},
	{0x28A49C127C269C2DULL, "vxlt_0_lowtown_a2_n_c_eug_0_0.vlm"},
	{0x28A5558F4A031824ULL, "449794801.ogg"},
	{0x28A587BA479A4EF4ULL, "ht_0_hightown_a3_34_0_ll.mmh"},
	{0x28A58F34B386C9B0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkleft.ani"},
	{0x28A62330CFADECCBULL, "6112674_m.fxe"},
	{0x28A623E921A8E1E7ULL, "dae300_petrice6097856.cl"},
	{0x28A63059749B3E45ULL, "vxht_bartrand_n_6pj_0_0.vlm"},
	{0x28A65819F7DF8C76ULL, "vxdr_0_deeproads_cpk_0_0.vlm"},
	{0x28A74677439F4CD2ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.cess.evt"},
	{0x28A752C0549099C1ULL, "233815708.ogg"},
	{0x28A807209BFD46CAULL, "325145929.ogg"},
	{0x28A8180E87F06A0BULL, "935825793.ogg"},
	{0x28A856C1CD478215ULL, "281891127.ogg"},
	{0x28A8A5918A7112EDULL, "vxw_bonepitdestroyed_a3_mn_0_0.vlm"},
	{0x28A9279804CA46F7ULL, "vxlt_undercityencounter_1fl_0_0.vlm"},
	{0x28A9460A972BDFC0ULL, "vxht_keep_n_6zv_0_0.vlm"},
	{0x28A9510349A5F28DULL, "hlac_lt_0_lowtown_docks_196607_0z.dds"},
	{0x28AA25E33991F0F6ULL, "6142133_m.fxe"},
	{0x28AA55D43C06C7A0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_xenon_d.dds"},
	{0x28AA647989DE4140ULL, "art\\characters\\creatures\\mubari\\animation\\com_barking.gad"},
	{0x28AB036518991ECEULL, "653278676.ogg"},
	{0x28AB4A8D5D5F3CAEULL, "gen_im_trs_gen_shd_wdn_r0.uti"},
	{0x28AB863CC186A4BAULL, "lt_0_lowtown_a2_n_43_0.rml"},
	{0x28ABB49B6946F0EEULL, "1022008652.ogg"},
	{0x28ABCD9324807D55ULL, "563035607.ogg"},
	{0x28ABCFF37E1B657BULL, "docks_male_overseer_mob6176570.cl"},
	{0x28AC3586C25C49EAULL, "vxca_caverns1_mines_jfj_0_0.vlm"},
	{0x28AC8A88124CBF50ULL, "6033389_m.fxe"},
	{0x28ACA229CDC9CB1CULL, "llac_ca_caverns1_24422_3.dds"},
	{0x28ACE661141C651BULL, "vxht_brothel_n_5lz_0_1.vlm"},
	{0x28AD222DEEE3B204ULL, "vxlt_warehouse_at_0_0.vlm"},
	{0x28AE4B3969463DF2ULL, "6114453_m.fxe"},
	{0x28AE657743A5A065ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.cess.gad"},
	{0x28AE87BF1298F78DULL, "vxht_0_hightown_a2_byq_0_0.vlm"},
	{0x28AEDD2E6FCA120BULL, "gen00fl_anders6113080.cl"},
	{0x28AF2C6E78C0F960ULL, "6125154_m.fxe"},
	{0x28AF922A952A809FULL, "922114497.ogg"},
	{0x28AFCE4A85F4F92EULL, "vxca_cavern1_haunted_dpf_0_0.vlm"},
	{0x28B036B8B2429A6DULL, "mag291_torpor6172568.cl"},
	{0x28B0C1F2C30250DFULL, "dae000_gamlen6048433.cl"},
	{0x28B0F9510460D0D1ULL, "vxht_0_hightown_9zg_0_0.vlm"},
	{0x28B12903C736D776ULL, "vxht_chantry_2li_0_0.vlm"},
	{0x28B1444B5C823515ULL, "vxlt_0_lowtown_a2_ewt_0_0.vlm"},
	{0x28B24718C41B5873ULL, "dre180ar_ruins_crypt.ncs"},
	{0x28B252F65F11ADBAULL, "w_blightlands_76_0.mmh"},
	{0x28B256F65F11B4F1ULL, "w_blightlands_76_0.mao"},
	{0x28B264F65F11CC00ULL, "w_blightlands_76_0.msh"},
	{0x28B2ED5F54B52B50ULL, "1010068862.ogg"},
	{0x28B345083F89B939ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyo.mmh"},
	{0x28B34F083F89CA27ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyo.msh"},
	{0x28B39CE8D9AE0AB6ULL, "vxlt_gallowstemplar_660_0_0.vlm"},
	{0x28B3D25EFDC15E66ULL, "vxht_mansion2b_30l_0_0.vlm"},
	{0x28B43EF8F07F370AULL, "hlac_lt_0_lowtown_a3_n_c_65536_10y.dds"},
	{0x28B485ACD785EEBBULL, "315338974.ogg"},
	{0x28B4C9F6E9D5A4F6ULL, "vxdr_0_deeproads_cml_0_0.vlm"},
	{0x28B4FC63235502E3ULL, "plt_fen200pt_night_terrors.plo"},
	{0x28B549BF11F1BE9FULL, "522863221.ogg"},
	{0x28B56244EE21FD8FULL, "mag100cr_karras.utc"},
	{0x28B5962C2C911A7DULL, "6112988_m.fxe"},
	{0x28B5AFF8B094EEA2ULL, "503574426.ogg"},
	{0x28B5D5AAFD8CEC60ULL, "6119319_m.fxe"},
	{0x28B60A21E67C8FCFULL, "mcr370_meredith6083032.cl"},
	{0x28B628D568971F8CULL, "hlac_ca_caverns1_mines_24422_7y.dds"},
	{0x28B6414B1A35D3CFULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeenter.gad"},
	{0x28B66070804F26A4ULL, "ht_0_hightown_n_37_0_lc.mmh"},
	{0x28B673B7EF5273D4ULL, "oil_proj_01.mao"},
	{0x28B6B6398EFFEB7FULL, "722746951.ogg"},
	{0x28B704C650DE805BULL, "t3_dey_bl1.tnt"},
	{0x28B710C75D92C831ULL, "vxlt_0_lowtown_docks_n_9r9_0_0.vlm"},
	{0x28B88471531652FDULL, "vxlt_0_lowtown_gbu_0_0.vlm"},
	{0x28B8A1507EA1F742ULL, "mer110cr_miner2.utc"},
	{0x28B8D3D63063BD12ULL, "llac_lt_0_lowtown_docks_a3_-131074_2.dds"},
	{0x28B8DF13C9D692B7ULL, "6163999_m.fxe"},
	{0x28B943D5F4FAA75AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwstraight.msh"},
	{0x28B951D5F4FABE93ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwstraight.mao"},
	{0x28B955D5F4FAC5A0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwstraight.mmh"},
	{0x28B96718C421AD7AULL, "dre180ar_ruins_crypt.lst"},
	{0x28B98FD07C1A24DDULL, "mcr370_varric6086843.cl"},
	{0x28BA0ED3B807C82CULL, "123783219.ogg"},
	{0x28BAA05E8006AA50ULL, "vxlt_gallowscourtyard_7cz_0_0.vlm"},
	{0x28BAD234C214F6A6ULL, "face_face_p7.psh.fxo"},
	{0x28BAFB5D3093CD47ULL, "llac_lt_0_lowtown_a3_n_c_65537_4.dds"},
	{0x28BB1C9DCAE001FBULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srda_0t.dds"},
	{0x28BB4EC2E07FB6A9ULL, "vxlt_hangedman_n_1vs_0_1.vlm"},
	{0x28BB5A6628042140ULL, "vxlt_house_g1_0_0.vlm"},
	{0x28BBAD1D332964B0ULL, "6049829_m.fxe"},
	{0x28BBB26F15B14761ULL, "6100011_m.fxe"},
	{0x28BBBDAA9A1A08BBULL, "106279532.ogg"},
	{0x28BBE0374C444D0BULL, "w_encounter3_100_0.phy"},
	{0x28BBE719D7514607ULL, "vxht_mansion2_2nz_0_0.vlm"},
	{0x28BC0DF90125B661ULL, "hlac_w_bonepitmining_a2_131077_0z.dds"},
	{0x28BD0CF5981AFF96ULL, "hlac_lt_foundry_n_320_3z.dds"},
	{0x28BD564B1A3C167CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeenter.evt"},
	{0x28BE47A395628A33ULL, "vxlt_gallowstemplar_n_2ec_0_0.vlm"},
	{0x28BF1ADF90A1F06CULL, "6063964_m.fxe"},
	{0x28BF318C2BE741A7ULL, "vfx_f_isa_trail_shimmer_p.wbd"},
	{0x28BF494DFF1E406BULL, "syncabilityanimations.gda"},
	{0x28BF624E292A88B4ULL, "one280_iwan6101114.cl"},
	{0x28BF8C2FC052918FULL, "6141494_m.fxe"},
	{0x28BF91DB84F1BA0EULL, "w_sundermount_85_0.mmh"},
	{0x28BF9DDB84F1CEA5ULL, "w_sundermount_85_0.mao"},
	{0x28BFABDB84F1E6FCULL, "w_sundermount_85_0.msh"},
	{0x28BFC20B6BEBC71BULL, "static_motionvectors_p0.psh.fxo"},
	{0x28BFD5BD230C042AULL, "vxw_encounter2_qq_0_0.vlm"},
	{0x28BFDF33B348673EULL, "hlac_dr_dungeon_3207_3y.dds"},
	{0x28C0A0D8A7619C69ULL, "6156063_m.fxe"},
	{0x28C0A45F9302F66EULL, "mage_coldmastery.dds"},
	{0x28C13A352223D7D3ULL, "llac_ca_caverns1_24426_4.dds"},
	{0x28C14CB014C9C064ULL, "6205016_m.fxe"},
	{0x28C1CD18AB6E5B0AULL, "vxht_mansion2_1vl_0_0.vlm"},
	{0x28C1E15226FC07C7ULL, "plt_dae000_mother.plo"},
	{0x28C23DF0438EE342ULL, "gen00fl_aveline6120774.cl"},
	{0x28C2440882524B9FULL, "hlac_w_encounter2_camp_65541_0x.dds"},
	{0x28C26684556936E4ULL, "llac_w_encounter2_65540_0.dds"},
	{0x28C2856AA8F33297ULL, "528347376.ogg"},
	{0x28C2CCE4D053EA0FULL, "vxht_0_hightown_c9z_0_0.vlm"},
	{0x28C369D9DD058AAFULL, "vxw_bonepitmining_a2_pe_0_0.vlm"},
	{0x28C396064B813E1EULL, "460494557.ogg"},
	{0x28C3EB51B27EE33BULL, "283403649.ogg"},
	{0x28C44026C27BDF78ULL, "zz_lgt_1_debug6114396.cl"},
	{0x28C4751CE51CE8E2ULL, "6134507_m.fxe"},
	{0x28C4B9EA12894938ULL, "hlac_lt_0_lowtown_a2_n_-1_5y.dds"},
	{0x28C4C4FE5B0C5239ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb8.msh"},
	{0x28C4CEFE5B0C6327ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb8.mmh"},
	{0x28C5CBE8C9B48BC7ULL, "follower_banter6170488.cl"},
	{0x28C627B1EA55EDA6ULL, "dae000_arishok6067858.cl"},
	{0x28C62E1C56B395AFULL, "216970714.ogg"},
	{0x28C6300AD4498C77ULL, "hlac_ht_0_hightown_a2_65536_0y.dds"},
	{0x28C63B82C41B2998ULL, "hlac_ca_cavern1_haunted_24421_6y.dds"},
	{0x28C64B45E129E485ULL, "873239405.ogg"},
	{0x28C66B11579087F3ULL, "vxw_encounter2_camp_5p_0_0.vlm"},
	{0x28C6876B57968DDFULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firesml.dura.gad"},
	{0x28C6A170FA05A0A0ULL, "6104960_m.fxe"},
	{0x28C755CE640C2023ULL, "351307433.ogg"},
	{0x28C82645CB68ADD4ULL, "6139490_m.fxe"},
	{0x28C82E2AA01253CCULL, "414927869.ogg"},
	{0x28C8F051B6DC934DULL, "ave331_jeven6138760.cl"},
	{0x28CA05E7D2CFE304ULL, "vxlt_warehouse_oz_0_0.vlm"},
	{0x28CA5E4B1A46AD95ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeenter.ani"},
	{0x28CAEB1C72633BE1ULL, "6067864_m.fxe"},
	{0x28CAEBCC6A16A513ULL, "213835531.ogg"},
	{0x28CBB63B850235BDULL, "vxlt_hangedman_rl_0_0.vlm"},
	{0x28CBD52D4F92B023ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallc.msh"},
	{0x28CBE32D4F92C87DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallc.mmh"},
	{0x28CBE72D4F92CE8EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallc.mao"},
	{0x28CBFFCD531D2644ULL, "6125579_f.fxe"},
	{0x28CC4BFA253EA9E3ULL, "6132745_f.fxe"},
	{0x28CC99583985CD88ULL, "hla_ht_0_hightown_a3_-131072_0x.dds"},
	{0x28CCE62893343C27ULL, "vxlt_0_lowtown_a3_n_c_eow_0_1.vlm"},
	{0x28CD7C6B579C99CCULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firesml.dura.evt"},
	{0x28CE8C7E7EBAC321ULL, "w_bonepitdestroyed_a3_12_0_lc.mmh"},
	{0x28CE9047D1F54B86ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_cwk.ani"},
	{0x28CEBC662AE781DCULL, "vxlt_0_lowtown_docks_n_c4d_0_0.vlm"},
	{0x28CF016AFB952229ULL, "563350056.ogg"},
	{0x28CF2F718D698464ULL, "vxw_woundedcoast_1dt_0_0.vlm"},
	{0x28CF6D0F05238736ULL, "vxht_mansion2b_3e3_0_0.vlm"},
	{0x28CF8810591621EAULL, "hlac_lt_gallowscourtyard_n_-131072_4z.dds"},
	{0x28CFADE7889FAAF3ULL, "rog_stun.dds"},
	{0x28D0A7262DC39E89ULL, "mer320_hubert6093481.cl"},
	{0x28D0C4DDB6A62E5EULL, "zz_jp_follower_poses.cut"},
	{0x28D0F6D7F06691C8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_visa_0s.dds"},
	{0x28D10E82FCA6F3D0ULL, "vxdr_primevaltomb_85w_0_0.vlm"},
	{0x28D11C863C7C6614ULL, "vxw_bonepit_1v1_0_0.vlm"},
	{0x28D244AC2A6FC806ULL, "360103474.ogg"},
	{0x28D24E5E7A5A1695ULL, "6071257_f.fxe"},
	{0x28D29ED19C5AC977ULL, "gen00fl_aveline6139674.cl"},
	{0x28D3141F65BEC873ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.impa.ani"},
	{0x28D33E0C7A66C065ULL, "357489719.ogg"},
	{0x28D473D640F4160EULL, "vxlt_0_lowtown_a2_fij_0_0.vlm"},
	{0x28D477D17991EE04ULL, "595918863.ogg"},
	{0x28D4CB6442819BA0ULL, "6072306_m.fxe"},
	{0x28D4F1EE0B49A1C3ULL, "w_bonepitdestroyed_a3_144_0.msh"},
	{0x28D4FFEE0B49B91DULL, "w_bonepitdestroyed_a3_144_0.mmh"},
	{0x28D503EE0B49C02EULL, "w_bonepitdestroyed_a3_144_0.mao"},
	{0x28D50D5D6E710828ULL, "dae011_corff_sta6140694.cl"},
	{0x28D58B9B874A811CULL, "270953567.ogg"},
	{0x28D6490631E7267BULL, "vxlt_0_lowtown_a2_n_c_etr_0_0.vlm"},
	{0x28D697197BFDFC46ULL, "6083029_m.fxe"},
	{0x28D6F7D965F47A44ULL, "6101974_m.fxe"},
	{0x28D71B90F6185902ULL, "dae000_madam_lusine6151489.cl"},
	{0x28D7283B5462A122ULL, "6134895_m.fxe"},
	{0x28D771F65F30FEC5ULL, "w_blightlands_76_0.phy"},
	{0x28D8CE353C9C7DF5ULL, "vxlt_gallowscourtyard_74j_0_0.vlm"},
	{0x28D8D873F4E76965ULL, "llac_lt_0_lowtown_n_-65537_4.dds"},
	{0x28D9A56EE96DA2ACULL, "hlac_lt_undercityencounter_-1_27y.dds"},
	{0x28D9A8BAF7F1787CULL, "and100_anders6096021.cl"},
	{0x28D9CC172E481944ULL, "vxht_mansion2b_2kb_0_0.vlm"},
	{0x28DA04D64E545187ULL, "hlac_ht_keep_3732_14x.dds"},
	{0x28DA614780260C14ULL, "sky_lt_warehouse.mao"},
	{0x28DA646B57A6FB25ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firesml.dura.ani"},
	{0x28DADD423CCC5EA1ULL, "546707774.ogg"},
	{0x28DB8CE4C095A473ULL, "vxlt_foundry_n_100_0_0.vlm"},
	{0x28DBD9643AB01914ULL, "6139902_m.fxe"},
	{0x28DC0B80AEABF5A5ULL, "vxdr_0_deeproads_cv6_0_0.vlm"},
	{0x28DC7847D20160F7ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_cwk.evt"},
	{0x28DC811F4D9954AFULL, "6173598_m.fxe"},
	{0x28DD28E601F5D18EULL, "docks_male_laborer_mob6175406.cl"},
	{0x28DD83F3CE5FCA99ULL, "isa311_castillon6092073.cl"},
	{0x28DDA14E7F5F9CE4ULL, "vxw_woundedcoast_1ou_0_0.vlm"},
	{0x28DDA3F3AD897490ULL, "mrl150_merrill_pt7.stg"},
	{0x28DE1C69B9629EC2ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_8x.dds"},
	{0x28DE6618C440C7FDULL, "dre180ar_ruins_crypt.are"},
	{0x28DF32132E19A8FFULL, "vxht_0_hightown_c89_0_0.vlm"},
	{0x28DF367B460CADC0ULL, "6110412_f.fxe"},
	{0x28DF6F5585296CEBULL, "vxca_caverns1_mines_dq8_0_0.vlm"},
	{0x28DFDE79A63978D3ULL, "582926982.ogg"},
	{0x28E00C1F65C944AAULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.impa.evt"},
	{0x28E02C92B49E65A0ULL, "6176752_m.fxe"},
	{0x28E030A471D51097ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.anb"},
	{0x28E0DC8D56C3D9A5ULL, "rdr111dg_martin6054833.cl"},
	{0x28E166333CC45069ULL, "follower_barks6153389.cl"},
	{0x28E194C27E5067FDULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyn_0t.dds"},
	{0x28E1ABEA47584170ULL, "vxht_keep_n_6j6_0_0.vlm"},
	{0x28E1C3D8B0B75FCEULL, "vxht_0_hightown_c97_0_0.vlm"},
	{0x28E246645A267ABAULL, "vxlt_0_lowtown_docks_a3_n_cv3_0_0.vlm"},
	{0x28E27947D205CD80ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_cwk.gad"},
	{0x28E2BB1D38830060ULL, "vrc221_varric6040819.cl"},
	{0x28E30D7E5F0441C0ULL, "vxht_keep_n_6fr_0_0.vlm"},
	{0x28E3746F308C0830ULL, "6120416_m.fxe"},
	{0x28E375F7A4E1DEB1ULL, "hlac_lt_0_lowtown_docks_a3_-65540_0y.dds"},
	{0x28E3EF928B9691DDULL, "vxlt_0_lowtown_n_h2e_0_0.vlm"},
	{0x28E421ED5A66692CULL, "332871426.ogg"},
	{0x28E454DC7AC067CBULL, "485080103.ogg"},
	{0x28E50411EBB870CBULL, "dae021_torelin_mob6148350.cl"},
	{0x28E5633CC1B7E1C6ULL, "lt_0_lowtown_a2_n_43_0.anb"},
	{0x28E5843F0EB212C4ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_c.anb"},
	{0x28E58F00F8D2E4CDULL, "6093636_m.fxe"},
	{0x28E5EFF7E96C30C9ULL, "hlac_lt_0_lowtown_docks_131070_3z.dds"},
	{0x28E5F00DD9369F36ULL, "gen00fl_varric6093635.cl"},
	{0x28E6030B1AA70E76ULL, "6039812_f.fxe"},
	{0x28E62DC038520F78ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.anb"},
	{0x28E67E8B9FEF1BF3ULL, "388942148.ogg"},
	{0x28E6A4DA71E7BA60ULL, "vxdr_primevaltomb_7ex_0_0.vlm"},
	{0x28E6C6A3C74B47A3ULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgb.msh"},
	{0x28E6D4A3C74B5FFDULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgb.mmh"},
	{0x28E6EB1F65CF2B1DULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.impa.gad"},
	{0x28E730A3CFAADD56ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngbn.msh"},
	{0x28E73AA3CFAAEE54ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngbn.mmh"},
	{0x28E73EA3CFAAF567ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphngbn.mao"},
	{0x28E7A2DBB2348899ULL, "330436158.ogg"},
	{0x28E92881847761CBULL, "vxw_sundermount_2e3_0_0.vlm"},
	{0x28E93FDFDFB06D02ULL, "lt_0_lowtown_docks_a3_144_0.msh"},
	{0x28E941DFDFB070B8ULL, "lt_0_lowtown_docks_a3_144_0.mmh"},
	{0x28E94DDFDFB0855BULL, "lt_0_lowtown_docks_a3_144_0.mao"},
	{0x28E9B5388391CCA6ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.anb"},
	{0x28EA6D6B735E86A9ULL, "873133415.ogg"},
	{0x28EAD77042FB28ECULL, "mag131_keran6086208.cl"},
	{0x28EB2B10DE63A82FULL, "6163652_m.fxe"},
	{0x28EB8F5D49B76352ULL, "dae000_elegant6166925.cl"},
	{0x28EBBDD91E04910FULL, "gen00fl_fenris6110836.cl"},
	{0x28EBE18D6CF92B03ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.anb"},
	{0x28EC132B7155B564ULL, "vxw_woundedcoast_1jz_0_0.vlm"},
	{0x28EC13CAB4169C08ULL, "6160816_m.fxe"},
	{0x28EC18E18892A1F3ULL, "vxca_caverns1_j2r_0_0.vlm"},
	{0x28EC493CE69DC52FULL, "204107568.ogg"},
	{0x28ECAA38DC4415EDULL, "llac_lt_undercitydungeon_-1_48.dds"},
	{0x28ED22AF8150A48AULL, "vxht_mansion2_1go_0_0.vlm"},
	{0x28ED64044714052BULL, "35423565.ogg"},
	{0x28ED6C93BA887A15ULL, "vxht_keep_n_6iz_0_0.vlm"},
	{0x28EE5B80206C0A3EULL, "hlac_w_encounter2_camp_196606_0z.dds"},
	{0x28EEB08B6CF4398AULL, "llac_lt_0_lowtown_docks_n_262142_0.dds"},
	{0x28EEBF70B2E7D42FULL, "vxw_bonepitmining_a2_af_0_0.vlm"},
	{0x28EEEB67481DDC0DULL, "hlac_lt_0_lowtown_65536_6y.dds"},
	{0x28EF100214C23A95ULL, "vis120_commoner2.utc"},
	{0x28EF346172D34BEEULL, "narrative_flemeth6002772.cl"},
	{0x28EF5061BD39314EULL, "suc_dis.mao"},
	{0x28EF8AF76DC49E8BULL, "848605288.ogg"},
	{0x28EFFC3A64815454ULL, "6207803_m.fxe"},
	{0x28F08AC4ADA5D99FULL, "492548500.ogg"},
	{0x28F0A66C5E4BBAD5ULL, "6133556_m.fxe"},
	{0x28F0EA81DD1463B5ULL, "zz_ave_debug6056768.cl"},
	{0x28F1676F1180A5FDULL, "zz_approval_debug6003918.cl"},
	{0x28F17F5B50FA5A40ULL, "vxht_keep_n_6as_0_0.vlm"},
	{0x28F1BE34D1A77176ULL, "6100290_m.fxe"},
	{0x28F1F5EA87AA35F8ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike.phy"},
	{0x28F258D22A656F08ULL, "104564413.ogg"},
	{0x28F27B6EAD71CF9CULL, "lowtown_fem_alien_elf_2_mob6166356.cl"},
	{0x28F2A33F0EBCD1C7ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_c.mmh"},
	{0x28F2BFDCD4F2A593ULL, "6110690_m.fxe"},
	{0x28F3281393F6E411ULL, "vxlt_hangedman_21c_0_0.vlm"},
	{0x28F3E65D45EED988ULL, "6068864_m.fxe"},
	{0x28F43716739F8181ULL, "6043897_m.fxe"},
	{0x28F4A5C8BD36C08CULL, "vxlt_undercityserialkiller_1iv_0_0.vlm"},
	{0x28F5463B311B85A3ULL, "gen00fl_aveline6015635.cl"},
	{0x28F56B70518F4603ULL, "ser221_gascard6122432.cl"},
	{0x28F5862981655D93ULL, "vxca_caverns2_8le_0_0.vlm"},
	{0x28F5B0837B5BEA98ULL, "774469267.ogg"},
	{0x28F60F8CED951842ULL, "sib111_mother6098947.cl"},
	{0x28F619FDB562D0CCULL, "qun131_warrens_thug6038839.cl"},
	{0x28F684D09C8C8185ULL, "104815824.ogg"},
	{0x28F6FECD4BB6801BULL, "art\\characters\\playercharacter\\humanmale\\hlh_coma.mmh"},
	{0x28F70ACD4BB694B8ULL, "art\\characters\\playercharacter\\humanmale\\hlh_coma.mao"},
	{0x28F71CCD4BB6B305ULL, "art\\characters\\playercharacter\\humanmale\\hlh_coma.msh"},
	{0x28F79A5E766D7A4BULL, "vxw_sundermount_2n2_0_0.vlm"},
	{0x28F7B838839E0F51ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.mmh"},
	{0x28F7F2E119F4D051ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_diagonal_1.mao"},
	{0x28F88CC182973127ULL, "follower_banter6163615.cl"},
	{0x28F8B930AC2850BEULL, "538223807.ogg"},
	{0x28F8D2C870F09850ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyn.mao"},
	{0x28F8D6C870F09F63ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyn.mmh"},
	{0x28F8E4C870F0B6BDULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyn.msh"},
	{0x28F8FF24ABB34B85ULL, "vxlt_0_lowtown_n_gik_0_0.vlm"},
	{0x28F91428453B8B5AULL, "ss_danarius.bnk"},
	{0x28F932A1BE2F4B51ULL, "vxht_0_hightown_9az_0_0.vlm"},
	{0x28F9550DCA933377ULL, "clf260_harley6069951.cl"},
	{0x28F9BE8D6D052D2CULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.mmh"},
	{0x28F9C2BD2B9F0A2CULL, "6051391_f.fxe"},
	{0x28FA4BFAE10193DEULL, "hlac_lt_0_lowtown_n_-1_5y.dds"},
	{0x28FA855CE157D529ULL, "126904951.ogg"},
	{0x28FAA78A0EAC804EULL, "vxlt_0_lowtown_n_eib_0_0.vlm"},
	{0x28FAB4AD4AF26FFBULL, "21053815.ogg"},
	{0x28FB6B073623D949ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf01a.phy"},
	{0x28FB7C22E0FC3E30ULL, "251471541.ogg"},
	{0x28FC24010ACEF071ULL, "vxlt_gallowscourtyard_n_3bs_0_0.vlm"},
	{0x28FD2C9C838B9872ULL, "435061388.ogg"},
	{0x28FD314DA4DDECAFULL, "vxw_bonepitmining_a2_la_0_0.vlm"},
	{0x28FD5551E18F5AF9ULL, "vxht_keep_5jg_0_0.vlm"},
	{0x28FD7F3CA1CDBDEFULL, "6152076_m.fxe"},
	{0x28FD9167C3FF8458ULL, "mrl310_party_comments6142743.cl"},
	{0x28FDA73D02769143ULL, "6063750_m.fxe"},
	{0x28FDAE09156544B2ULL, "6121146_f.fxe"},
	{0x28FE3DBF7CCBEA7BULL, "6176513_m.fxe"},
	{0x28FEF438DC2DEBCCULL, "456570751.ogg"},
	{0x28FF96499284CC8CULL, "llac_ht_mansion2_818_17.dds"},
	{0x290038F46616492FULL, "hlac_dr_primevalentrance_1724_5x.dds"},
	{0x29003B8C2C1EE9DDULL, "vfx_f_isa_trail_shimmer_p.bnk"},
	{0x290071733A506F4FULL, "static_motionvectors_p0.vsh.fxo.dx11"},
	{0x2900CFC81254DC66ULL, "hlac_ca_cavern1_haunted_24425_5z.dds"},
	{0x2900DB965BFFA573ULL, "239238800.ogg"},
	{0x29013E08E890000CULL, "nat320_delilah6065889.cl"},
	{0x290170224B51B48CULL, "vxlt_0_lowtown_docks_n_bac_0_1.vlm"},
	{0x290219B620FD82FEULL, "vxlt_0_lowtown_a2_n_c_eja_0_0.vlm"},
	{0x29027C1577D478A8ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_001a_0n.dds"},
	{0x29029F25670F1473ULL, "ave331_jeven6137288.cl"},
	{0x2902D015E0645538ULL, "217101455.ogg"},
	{0x2902F652F116D3B5ULL, "gen00fl_aveline6139478.cl"},
	{0x2903169CA5F071E8ULL, "vxw_encounter1_en_0_0.vlm"},
	{0x290317A5AF2CD90CULL, "vxlt_undercityserialkiller_1nw_0_0.vlm"},
	{0x2903D32C7013FD89ULL, "dae000_epic_merchant_16179418.cl"},
	{0x2903F806735B7613ULL, "vxca_caverns2_8kd_0_0.vlm"},
	{0x29040B0D67C7BEBCULL, "lt_0_lowtown_docks_a3_n_126_0.mao"},
	{0x2904170D67C7D35FULL, "lt_0_lowtown_docks_a3_n_126_0.mmh"},
	{0x29041D0D67C7DD01ULL, "lt_0_lowtown_docks_a3_n_126_0.msh"},
	{0x2904A90EF2798B03ULL, "art\\effects\\cloudshadows\\w_bonepitdestroyed_a3\\cloud2.dds"},
	{0x2904AB58EDCE35A4ULL, "307758206.ogg"},
	{0x290552D4A41B9638ULL, "528419621.ogg"},
	{0x29060C3B686392CBULL, "vxw_sundermount_2s5_0_0.vlm"},
	{0x2906135FD8E4D8CDULL, "hlac_ca_caverns2prc_5802_32z.dds"},
	{0x29063D6733064EE3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefl.evt"},
	{0x2906924CBF41AB04ULL, "766691339.ogg"},
	{0x2906B920841FFAA3ULL, "48454372.ogg"},
	{0x2906D2DFDBA44BBFULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm02.mmh"},
	{0x2906D8DFDBA45661ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm02.msh"},
	{0x2906E96C8AE99446ULL, "dae000_bartrand6204162.cl"},
	{0x2907CD2B72B36B0FULL, "vxlt_undercityencounter_uk_0_1.vlm"},
	{0x2907D94B61DF7155ULL, "6110897_f.fxe"},
	{0x29082FA42A3A477CULL, "mag371_elsa6065010.cl"},
	{0x2908D7AE6BEF05A5ULL, "vxlt_0_lowtown_a2_n_byf_0_0.vlm"},
	{0x2908F6E55541E664ULL, "vxw_woundedcoast_1px_0_0.vlm"},
	{0x2909091E2424555DULL, "vxlt_gallowstemplar_n_431_0_0.vlm"},
	{0x29092DA471F81098ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.mmh"},
	{0x29093AD2B2BB08C8ULL, "hlac_w_bonepit_6_0z.dds"},
	{0x29096941C260D77AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_left_up.gad"},
	{0x29099F9C07A5A00DULL, "vxca_caverns2prc_8c3_0_0.vlm"},
	{0x2909B6F5F632583DULL, "509735822.ogg"},
	{0x2909D7C870FF1388ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyn.phy"},
	{0x290A1787946BE8A1ULL, "252710604.ogg"},
	{0x290A99D371171579ULL, "177496324.ogg"},
	{0x290B9A6563F48745ULL, "vxlt_0_lowtown_a2_n_fg0_0_0.vlm"},
	{0x290D1667330C2B64ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefl.gad"},
	{0x290DAF33C03DF415ULL, "196070136.ogg"},
	{0x290E2CC038735F13ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.mmh"},
	{0x290E6A5DB85748F7ULL, "mag291cr_wesley.utc"},
	{0x290EFC2C32C6F760ULL, "hlac_lt_0_lowtown_a2_-131072_8z.dds"},
	{0x290F56BF4BE0AAFBULL, "zz_prmpt_debug6198221.cl"},
	{0x290F6C41C2654791ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_left_up.evt"},
	{0x290FB8537E659923ULL, "vxlt_warehouse_n_qn_0_0.vlm"},
	{0x290FCF4B547AAADFULL, "6095614_m.fxe"},
	{0x290FD3C00BF97C1EULL, "30699988.ogg"},
	{0x290FDA184DB7ABD9ULL, "453066711.ogg"},
	{0x29103BAF6001B977ULL, "6098947_m.fxe"},
	{0x2910A638AEA405EEULL, "vxlt_0_lowtown_a2_n_dbb_0_0.vlm"},
	{0x2910B181453FA1F6ULL, "lt_backalley_n_mmap.dds"},
	{0x2910BA41CF5B1E83ULL, "hlac_lt_0_lowtown_a3_n_c_131073_2y.dds"},
	{0x2910E1CBE9CDCCCEULL, "art\\levels\\ztools\\helpers\\character\\_textures\\mt_har_crew_0d.dds"},
	{0x2910E787D289F652ULL, "110241603.ogg"},
	{0x2910F921D3E2B1EEULL, "dae000_flemeth6013383.cl"},
	{0x291138CC8B10B548ULL, "lt_0_lowtown_docks_a3_n_43_0.rml"},
	{0x2911887997E68A68ULL, "vxw_encounter1_jo_0_0.vlm"},
	{0x2911FA7147CA03E0ULL, "vxdr_primevaltomb_7tg_0_0.vlm"},
	{0x291289DC5EDEA66CULL, "754004299.ogg"},
	{0x29130442317C5724ULL, "dae000_gamlen6145128.cl"},
	{0x29130C976F56A9F0ULL, "6050188_m.fxe"},
	{0x29131F29B10C7C63ULL, "423766681.ogg"},
	{0x291361786B5E566CULL, "hlac_lt_0_lowtown_n_65536_12x.dds"},
	{0x291418A633257DFDULL, "vxht_chantry_n_31w_0_0.vlm"},
	{0x2914256733126342ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefl.ani"},
	{0x291439796142E590ULL, "vxlt_gallowscourtyard_3wz_0_0.vlm"},
	{0x291442AAA593609DULL, "gui_reticle_p1.psh.fxo.dx11"},
	{0x29147E185A59AB4BULL, "vxw_sundermount_2t4_0_0.vlm"},
	{0x291546083FDC8526ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyo.phy"},
	{0x291549F23CD213A5ULL, "ser270ar_alley.ncs"},
	{0x29155E2703E26645ULL, "zz_one000_barker.cnv"},
	{0x2915DA8EEE5BC8C9ULL, "gen00fl_isabela6201239.cl"},
	{0x2915E2DE8F9F7C85ULL, "vxlt_0_lowtown_n_gci_0_0.vlm"},
	{0x2915EFE2242DDC58ULL, "hla_ht_0_hightown_-196608_0y.dds"},
	{0x29167A3809446C35ULL, "6090185_m.fxe"},
	{0x2916AADF2C4F948EULL, "vxw_bonepitdestroyed_a3_2am_0_0.vlm"},
	{0x291716F142D61021ULL, "6135896_f.fxe"},
	{0x291734D7DD56A398ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masd_0n.dds"},
	{0x29175E51CA2C453BULL, "6134959_m.fxe"},
	{0x2917655B9E49223BULL, "vxlt_hangedman_1vn_0_1.vlm"},
	{0x2917BA265C5DE22EULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_005a_0s.dds"},
	{0x2917C1433B17E308ULL, "l_dae370ar_gallows.wbd"},
	{0x2917E7E5869F2BEAULL, "vxlt_gallowstemplar_n_4y0_0_0.vlm"},
	{0x2918586DF032BD29ULL, "626668634.ogg"},
	{0x2918EBE6AC417DCAULL, "271678593.ogg"},
	{0x29190258E189B87AULL, "6018810_m.fxe"},
	{0x291942273E62A5E7ULL, "288033245.ogg"},
	{0x291962239DA1CA10ULL, "vxht_keep_4nc_0_0.vlm"},
	{0x2919CFFCCE8B301AULL, "gen00fl_sebastian6174719.cl"},
	{0x291A362A744C81FCULL, "6172671_m.fxe"},
	{0x291A6B7D7DF01512ULL, "vxdr_primevalentrance_1md_0_0.vlm"},
	{0x291A8BB763B057CCULL, "vxlt_gallowstemplar_f_3r0_0_0.vlm"},
	{0x291AB813A22CE6AAULL, "zz_mer_debug6016969.cl"},
	{0x291B09924F48A02CULL, "ser221_gascard6165663.cl"},
	{0x291C02EA39CF47C0ULL, "119461521.ogg"},
	{0x291C1B08DD027B52ULL, "533867950.ogg"},
	{0x291C3878373C5AF4ULL, "plt_cod_aow_combat.plo"},
	{0x291C50C45A417E44ULL, "717288269.ogg"},
	{0x291CD32FA41A82D0ULL, "vxca_cavern1_haunted_i2s_0_0.vlm"},
	{0x291D3CCD437337C9ULL, "follower_banter6160847.cl"},
	{0x291D8441C271AE60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_left_up.ani"},
	{0x291DB2AD9A1A5F16ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutb.msh"},
	{0x291DBCAD9A1A7014ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutb.mmh"},
	{0x291DC0AD9A1A7727ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutb.mao"},
	{0x291E2A30705BB1A3ULL, "vxlt_warehouse_n_tq_0_0.vlm"},
	{0x291E480E6FA5CB3EULL, "hlac_dr_slavers_3214_4z.dds"},
	{0x291E6092CA661500ULL, "6057076_f.fxe"},
	{0x291E785DBBEFE4A3ULL, "941429855.ogg"},
	{0x291F0D05F2C7F2AAULL, "384550050.ogg"},
	{0x291F8962D030BEB8ULL, "330473509.ogg"},
	{0x291FEC1FDCD82B45ULL, "6105706_m.fxe"},
	{0x29206C4E39C01C60ULL, "vxdr_primevaltomb_7qd_0_0.vlm"},
	{0x29207F38D5B6F825ULL, "mrl150_dalish_guard6068310.cl"},
	{0x2920CBC65BE0C15EULL, "hlac_lt_0_lowtown_a3_n_c_65537_5z.dds"},
	{0x2920DB90662DD07BULL, "hlac_lt_0_lowtown_docks_65534_6z.dds"},
	{0x292247CFF2CE19B5ULL, "608839704.ogg"},
	{0x29227CB5C9596977ULL, "dae000_flemeth6000700.cl"},
	{0x2922EFF54C4FC3CBULL, "vxw_sundermount_2y7_0_0.vlm"},
	{0x29238E500B0D16E7ULL, "llac_ht_mansion2_818_43.dds"},
	{0x2923A93E331BEC4CULL, "vxlt_0_lowtown_docks_n_b8y_0_0.vlm"},
	{0x2923C9AD82BC3A26ULL, "follower_banter6170418.cl"},
	{0x2923FB492AE02BC8ULL, "471635176.ogg"},
	{0x29243B803E6B2D82ULL, "905529273.ogg"},
	{0x29245834332BB7A4ULL, "hlac_ht_brothel_n_3365_1y.dds"},
	{0x29246363D12D1397ULL, "1058660247.ogg"},
	{0x2924B6480EC4BCD4ULL, "nat380_sibling6207550.cl"},
	{0x2924DACEA8023A00ULL, "hlac_w_sundermount_327678_0x.dds"},
	{0x292527B20925D11BULL, "vxlt_0_lowtown_a2_n_hfc_0_0.vlm"},
	{0x292552B3A678DEE3ULL, "857734225.ogg"},
	{0x292577B697CCEB34ULL, "gen_im_trs_dem_lgt_hlm_r0.uti"},
	{0x2925ACFAA795498EULL, "440745116.ogg"},
	{0x2926330603A97D80ULL, "mag291_demon_marethari6027969.cl"},
	{0x29263BA069DA8ECAULL, "6182458_m.fxe"},
	{0x2926A30A039DA8B1ULL, "vxht_0_hightown_cf9_0_0.vlm"},
	{0x2926B7693B1F4A64ULL, "gen00fl_sebastian6180440.cl"},
	{0x29270E7633253CC9ULL, "435148446.ogg"},
	{0x29275BF9A323DA05ULL, "18274078.ogg"},
	{0x2927CAEF4005CDC0ULL, "hlac_lt_0_lowtown_docks_-65538_8z.dds"},
	{0x2927E81C3E6AA6D5ULL, "framebuffereffects_horizontalblur_p0.vsh.fxo"},
	{0x29285407364AA06EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf01a.mmh"},
	{0x29286B4D2DA180CDULL, "vxlt_0_lowtown_a2_n_c_cmu_0_0.vlm"},
	{0x29286E07364ACC5CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf01a.msh"},
	{0x2928A80766429A7CULL, "6175771_m.fxe"},
	{0x2928C4A48097CA08ULL, "553579622.ogg"},
	{0x2929140A0352757DULL, "574251683.ogg"},
	{0x2929DBC32B32BA50ULL, "vxw_bonepitdestroyed_a3_1cz_0_0.vlm"},
	{0x2929EDD499D92D97ULL, "vxw_woundedcoast_17b_0_0.vlm"},
	{0x292A2C2549581940ULL, "305127415.ogg"},
	{0x292A409D4286D274ULL, "vxca_caverns2prc_9ft_0_0.vlm"},
	{0x292B372DE3B8D15FULL, "w_bonepitdestroyed_a3_69_0.msh"},
	{0x292B3D2DE3B8DB01ULL, "w_bonepitdestroyed_a3_69_0.mmh"},
	{0x292B492DE3B8EFA2ULL, "w_bonepitdestroyed_a3_69_0.mao"},
	{0x292B70ABBF55A423ULL, "vxdr_0_deeproads_g2x_0_0.vlm"},
	{0x292B7447A49E8CB7ULL, "6170308_m.fxe"},
	{0x292BA0EE62EC78C8ULL, "mag291ip_desk.utp"},
	{0x292C260214F60E00ULL, "vis120_commoner2.cnv"},
	{0x292C3F287E75B851ULL, "6059699_f.fxe"},
	{0x292C6CBE30F92C23ULL, "58008352.ogg"},
	{0x292C8E11044CCD6AULL, "isa000pt_gift_ship.ncs"},
	{0x292CBC45555AD814ULL, "vxht_0_hightown_a2_caq_0_0.vlm"},
	{0x292CBEAD6C8AD04CULL, "hlac_w_blightlands_131073_0z.dds"},
	{0x292CFB13B2D8BF53ULL, "170006430.ogg"},
	{0x292D1E40FBFB4A82ULL, "6095112_m.fxe"},
	{0x292D3CCB52A1282BULL, "325503311.ogg"},
	{0x292D48909CC98BBAULL, "866093914.ogg"},
	{0x292D69EEB5CC32B9ULL, "primitives_primitive_p3.vsh.fxo"},
	{0x292DC3C4FC92AB48ULL, "785702052.ogg"},
	{0x292DE65E337055B2ULL, "sky_ht_bartrand_n.mao"},
	{0x292DE6C307A6CEBEULL, "blooddragon1.dds"},
	{0x292E1C3C878E1811ULL, "vxht_0_hightown_4c9_0_0.vlm"},
	{0x292EAC7AA741136AULL, "vxht_keep_689_0_0.vlm"},
	{0x292FA5DF027753ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v3.evt"},
	{0x292FA6708217580BULL, "907295153.ogg"},
	{0x292FA910534AA818ULL, "follower_banter6166969.cl"},
	{0x292FD575BAA4A57DULL, "vxlt_hangedman_n_lj_0_1.vlm"},
	{0x2930080214F9C438ULL, "vis120_commoner2.bnk"},
	{0x29303848F57B6D7EULL, "llac_ca_caverns2prc_5802_60.dds"},
	{0x2931AABEDBA5EF03ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheelb.msh"},
	{0x2931B8BEDBA6075DULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheelb.mmh"},
	{0x2931BCBEDBA60E6EULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheelb.mao"},
	{0x29325DE670F3E53FULL, "llac_lt_gallowstemplar_f_2383_2.dds"},
	{0x2932C441FEE4780CULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.impa.ani"},
	{0x2933E5342F8199D9ULL, "vxlt_foundry_n_x9_0_0.vlm"},
	{0x2933FE65AA84AFFEULL, "735555615.ogg"},
	{0x293453188F177863ULL, "6170898_m.fxe"},
	{0x2934CE028D33CC34ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand_1.mao"},
	{0x293525EE9B52B053ULL, "vxlt_0_lowtown_docks_n_b7o_0_0.vlm"},
	{0x293548557D796BA0ULL, "gen00fl_isabela6015833.cl"},
	{0x2935D4F83CD8C0BEULL, "362103591.ogg"},
	{0x29361A5AD2B6D1A3ULL, "6027050_f.fxe"},
	{0x29361F9266EBB1E0ULL, "6173113_m.fxe"},
	{0x29362236620EFB3AULL, "6155418_m.fxe"},
	{0x293675A3EC9AD1F4ULL, "art\\characters\\creatures\\spider\\animation\\atk_v2.ani"},
	{0x293685717C6898FDULL, "zz_followers_banter6158271.cl"},
	{0x2936BEDF027D9D6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v3.gad"},
	{0x2937CF8CE64D5A9BULL, "w_woundedcoast_7_0.rml"},
	{0x2937DECB64BED66BULL, "vxht_0_hightown_a3_bz3_0_0.vlm"},
	{0x29384E230701274CULL, "222110606.ogg"},
	{0x293871F63C8CD3ACULL, "182477399.ogg"},
	{0x29388E27C44DDA01ULL, "t3_glk_b01.tnt"},
	{0x2938B27A347CEAF4ULL, "vxca_caverns2prc_9mu_0_0.vlm"},
	{0x2938CA7A8A9AB79CULL, "vxlt_foundry_n_vn_0_0.vlm"},
	{0x293904E6C98B9B0EULL, "hlac_lt_0_lowtown_a2_n_c_1_10z.dds"},
	{0x2939C8C31C47D344ULL, "836037647.ogg"},
	{0x2939D141FEEAAD4EULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.impa.gad"},
	{0x293A70487EB27754ULL, "385298364.ogg"},
	{0x293AA43353304C25ULL, "215651881.ogg"},
	{0x293B8D17F8A605E2ULL, "6134316_m.fxe"},
	{0x293BF84885E3341DULL, "6086168_m.fxe"},
	{0x293C044970B80CE2ULL, "6113658_m.fxe"},
	{0x293C1A16B43DF34AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampstake.mao"},
	{0x293C1E16B43DF999ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampstake.mmh"},
	{0x293C227028CC192FULL, "follower_barks6153475.cl"},
	{0x293C228436B46B1CULL, "vxlt_gallowstemplar_f_5rf_0_0.vlm"},
	{0x293C2816B43E0A87ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampstake.msh"},
	{0x293C72A3EC9F3856ULL, "art\\characters\\creatures\\spider\\animation\\atk_v2.gad"},
	{0x293C8FFE171E8326ULL, "llac_lt_0_lowtown_docks_a3_-2_2.dds"},
	{0x293D4317C1635AC7ULL, "hlac_dr_0_deeproads_5840_4x.dds"},
	{0x293D879829C7D3A5ULL, "vxht_chantry_2p2_0_0.vlm"},
	{0x293DBDDF0283BA6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v3.ani"},
	{0x293DE5F6E8F00C1DULL, "dre000_sibling_has_blight6036448.cl"},
	{0x293E0EC57CA6CCBBULL, "mag330_walter6046284.cl"},
	{0x293E1FCD4BF2BB50ULL, "art\\characters\\playercharacter\\humanmale\\hlh_coma.phy"},
	{0x293EA6A45CA74ED2ULL, "vxlt_0_lowtown_docks_n_b7j_0_0.vlm"},
	{0x293EF884EC4AC68CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefish.msh"},
	{0x293EFE84EC4AD17EULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefish.mmh"},
	{0x293F0A84EC4AE515ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefish.mao"},
	{0x293FEF0A06415D58ULL, "vfx_smo_thick_threshold.mao"},
	{0x294033472A1AC460ULL, "291010272.ogg"},
	{0x2940CC41FEF0C32DULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.impa.evt"},
	{0x2941ABE49492AC63ULL, "6075251_f.fxe"},
	{0x2942FB4C339FD4E0ULL, "llac_lt_0_lowtown_n_131070_6.dds"},
	{0x294321E9D4897812ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.impa.gad"},
	{0x294360F68C4296FAULL, "gen00fl_aveline6120749.cl"},
	{0x29438DA3ECA584B5ULL, "art\\characters\\creatures\\spider\\animation\\atk_v2.evt"},
	{0x2943BAF3E9968EB2ULL, "gen_im_cft_reg_d.uti"},
	{0x2943D6B33BD61B16ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_hed_flma_0s.dds"},
	{0x2945491D1B9B499AULL, "421085385.ogg"},
	{0x29455D8277FA9503ULL, "6001350_m.fxe"},
	{0x2946BF7D0F1EEB50ULL, "vxw_bonepitdestroyed_a3_1ix_0_0.vlm"},
	{0x2946F705138FC134ULL, "vxht_0_hightown_a71_0_0.vlm"},
	{0x2947245726730374ULL, "vxca_caverns2prc_9hz_0_0.vlm"},
	{0x29477BB34FB504EEULL, "384155282.ogg"},
	{0x2947A8FE5C911646ULL, "vxw_sundermount_2jg_0_0.vlm"},
	{0x2947C813ADB821EDULL, "dre180cs_arrivalprimeval.cut"},
	{0x2947D7B0A4F22434ULL, "llac_lt_undercityserialkiller_-1_8.dds"},
	{0x294851D02704BF67ULL, "6172728_m.fxe"},
	{0x2948B6D9042A209DULL, "358292101.ogg"},
	{0x29495DADE2BD98A8ULL, "m_dre180ar_deep_roads.bnk"},
	{0x294966B458A6B377ULL, "plt_vrc000pt_n_rival_mid.plo"},
	{0x2949DD98A30E9305ULL, "542379674.ogg"},
	{0x294A24E9D48F9C49ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.impa.evt"},
	{0x294A864DB15E5128ULL, "6070429_m.fxe"},
	{0x294ACC5AE6279EDCULL, "one100dg_vanard6048256.cl"},
	{0x294B0907C0EA691AULL, "114038151.ogg"},
	{0x294B58766FE6A4A2ULL, "42651477.ogg"},
	{0x294B6B4A2642026EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_wood_flat_d.dds"},
	{0x294BDA121FD549E6ULL, "vxlt_0_lowtown_a2_n_eha_0_0.vlm"},
	{0x294CC9CB71F88C37ULL, "qun210cr_guard6083840.cl"},
	{0x294CF4EA87F6DF13ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike.mmh"},
	{0x294D4586B89D42BBULL, "584466778.ogg"},
	{0x294D5E2BCC122ADBULL, "vxdr_0_deeproads_bhc_0_0.vlm"},
	{0x294D8071CD1D2817ULL, "hlac_lt_darktown_-1_29z.dds"},
	{0x294F2D86F4392B71ULL, "6110814_m.fxe"},
	{0x294F485DF05372CAULL, "53317997.ogg"},
	{0x2950781CB5A10F10ULL, "81079353.ogg"},
	{0x29507A46CA4B48FFULL, "vrc211st_hugin.stg"},
	{0x2951255885A558C0ULL, "hlac_w_bonepit_327686_0z.dds"},
	{0x295141389B06C85FULL, "w_bonepit_193_0.msh"},
	{0x295147389B06D201ULL, "w_bonepit_193_0.mmh"},
	{0x295153389B06E6A2ULL, "w_bonepit_193_0.mao"},
	{0x2951C7A440B5407FULL, "hlac_w_bonepitmining_a2_0_7z.dds"},
	{0x2951CE140F37F90DULL, "zz_one_debug6099730.cl"},
	{0x2951F0BF944491DDULL, "hlac_lt_0_lowtown_docks_a3_-65535_0y.dds"},
	{0x295251E22064E72EULL, "llac_lt_darktown_-1_13.dds"},
	{0x295294692C921949ULL, "101057856.ogg"},
	{0x2952A1CC00A8DE60ULL, "277450027.ogg"},
	{0x2952EFE1F7B16171ULL, "plt_lgt310pt_nt_melson.nss"},
	{0x29534E13B984A9D3ULL, "841148746.ogg"},
	{0x29538A33C6BEAC13ULL, "zz_ser_debug6078151.cl"},
	{0x2953D890810F0213ULL, "spt_branch.psh.fxo"},
	{0x2953EB480EA84E84ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2.ani"},
	{0x2954011619CC5664ULL, "66294621.ogg"},
	{0x29545003C4F71FD2ULL, "dre000_hireling16038615.cl"},
	{0x29547DD15069B46EULL, "6164216_m.fxe"},
	{0x2954A7E41793DA0DULL, "ave331_jeven6137721.cl"},
	{0x2954B5ABB2C530BBULL, "814581398.ogg"},
	{0x2954C28548AB076BULL, "vxht_0_hightown_a3_bl5_0_0.vlm"},
	{0x2954C842D82A5608ULL, "6133051_f.fxe"},
	{0x2954FDE3AC28248FULL, "art\\effects\\cloudshadows\\w_encounter1\\cloud1.dds"},
	{0x29551C8CE49AB9E7ULL, "vxht_0_hightown_a2_bca_0_0.vlm"},
	{0x29554DDA71255AB5ULL, "hlac_lt_0_lowtown_n_65538_2z.dds"},
	{0x29559F4BA8EE6DA1ULL, "gen00fl_dog6107705.cl"},
	{0x29561ADB4E872EC6ULL, "vxw_sundermount_2oh_0_0.vlm"},
	{0x29567B76D9BEA675ULL, "vxw_bonepitdestroyed_a3_19i_0_0.vlm"},
	{0x2956B9FBC701DA53ULL, "777363637.ogg"},
	{0x2956E02A2D5EB74DULL, "6205761_m.fxe"},
	{0x2957E0A8D9599C17ULL, "538228999.ogg"},
	{0x2957FED0BDFF5F99ULL, "vxlt_gallowstemplar_f_2jd_0_0.vlm"},
	{0x29580CE9D49BB0B8ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.impa.ani"},
	{0x295868CA4FCB7297ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsarch.mmh"},
	{0x29586CCA4FCB79A4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsarch.mao"},
	{0x29586E7C6A7A88BBULL, "64637887.ogg"},
	{0x29587ECA4FCB9809ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsarch.msh"},
	{0x2958D3A058C95FF5ULL, "vxht_0_hightown_n_4kf_0_0.vlm"},
	{0x2958FF9707823269ULL, "hlac_w_encounter2_camp_262148_0x.dds"},
	{0x295900EF2C8CE8A0ULL, "mcr370_meredith6083059.cl"},
	{0x2959063E1AA09C1CULL, "vxlt_gallowstemplar_f_5xd_0_0.vlm"},
	{0x29591B3DBE4E0A6FULL, "6000977_m.fxe"},
	{0x29596C3C5D0A9CC4ULL, "1080148.ogg"},
	{0x29596D14F3C0E647ULL, "hlac_lt_0_lowtown_docks_196607_0x.dds"},
	{0x295A3BED654A8330ULL, "vxlt_gallowstemplar_5cf_0_0.vlm"},
	{0x295A698D9DF0952CULL, "846876151.ogg"},
	{0x295A823F25B1E814ULL, "6155508_m.fxe"},
	{0x295A8FCEE6EB005CULL, "art\\vfx\\combat\\vfx_w_modal_act_c.impa.ani"},
	{0x295AE8480EAE68E6ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2.gad"},
	{0x295B335472C5FC41ULL, "vxht_mansion2b_3gw_0_1.vlm"},
	{0x295B5D0CEE4B65FBULL, "hlac_ht_mansion2_818_29x.dds"},
	{0x295B99F2C2927E5AULL, "hlac_lt_0_lowtown_a2_n_-65535_7z.dds"},
	{0x295BA17CA1E4E6ABULL, "mag110_vincento6017670.cl"},
	{0x295BACBCBD653386ULL, "478302493.ogg"},
	{0x295BC2AF236328ACULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_conj_c.impa.ani"},
	{0x295BD008BE08435BULL, "vxdr_0_deeproads_bcb_0_0.vlm"},
	{0x295C0C10164E23B6ULL, "vxdr_dungeon_21u_0_0.vlm"},
	{0x295D43FE1E597A18ULL, "681533468.ogg"},
	{0x295D5F462E4E8529ULL, "432505532.ogg"},
	{0x295DC967C4355324ULL, "ser221_party_comments6111558.cl"},
	{0x295E10A10B2710B0ULL, "qun131tr_party_ambient.ncs"},
	{0x295E6CBA928C4AF4ULL, "hlac_w_woundedcoast_65542_0y.dds"},
	{0x295E75B458B928ECULL, "plt_vrc000pt_n_rival_mid.nss"},
	{0x295EE7433B5426CAULL, "l_dae370ar_gallows.bnk"},
	{0x295EF21194869B77ULL, "qun110_petrice6018603.cl"},
	{0x295F079002ED0A91ULL, "ht_0_hightownexplosion_n_17_0_lc.mmh"},
	{0x296032FF72453CD5ULL, "6076901_m.fxe"},
	{0x29604271EDD1C3AAULL, "hlac_lt_0_lowtown_docks_n_-65539_3y.dds"},
	{0x29609CCEE6EF819EULL, "art\\vfx\\combat\\vfx_w_modal_act_c.impa.gad"},
	{0x2960E3480EB2CBC5ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2.evt"},
	{0x2960F6AEA3C6764DULL, "6093730_f.fxe"},
	{0x29616506ED7BFEDFULL, "6098262_m.fxe"},
	{0x29618A9CCA9A3E7CULL, "hm_slaver02.mor"},
	{0x2961FC23D6E3063FULL, "6015011_f.fxe"},
	{0x296228CB83E0243BULL, "hlac_lt_gallowstemplar_n_2391_4y.dds"},
	{0x29629B2BE6BA9EE2ULL, "246979391.ogg"},
	{0x2962CFAF23695DEEULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_conj_c.impa.gad"},
	{0x296362791AABB382ULL, "lt_gallowstemplar_n_mmap.dds"},
	{0x2963E604A5B497E9ULL, "860396041.ogg"},
	{0x29642F3AA9D23486ULL, "vxlt_0_lowtown_a2_n_c_eux_0_0.vlm"},
	{0x29643E4F82A04CB1ULL, "follower_banter6143116.cl"},
	{0x29646B9049EC8522ULL, "vxdr_0_deeproads_erw_0_0.vlm"},
	{0x29648CB8407D4746ULL, "vxw_sundermount_2pi_0_0.vlm"},
	{0x2964A1057047990BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_threat_decay.ani"},
	{0x2964A828C1F23B84ULL, "6206879_m.fxe"},
	{0x2964B6F8095065B3ULL, "96194279.ogg"},
	{0x2964B91C5F794314ULL, "622139108.ogg"},
	{0x2964D79602E8A3C0ULL, "978032207.ogg"},
	{0x2964E0BEA4857E00ULL, "vxlt_gallowsprison_n_6mx_0_0.vlm"},
	{0x2964FF66230F9F58ULL, "44167374.ogg"},
	{0x29653A43E5F9872FULL, "follower_banter6151290.cl"},
	{0x296552071F54CFCDULL, "6208061_m.fxe"},
	{0x29658F3063B60297ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj05.phy"},
	{0x29661E767504666EULL, "6144395_m.fxe"},
	{0x2966BD58C8F03BD3ULL, "gen_im_npc_dwf_arm_glv_wara.uti"},
	{0x2967228F6C68CF05ULL, "6082421_m.fxe"},
	{0x2967293CB2412D10ULL, "gen00fl_isabela6088097.cl"},
	{0x29676F85079D189EULL, "6051835_m.fxe"},
	{0x296777CEE6F561BDULL, "art\\vfx\\combat\\vfx_w_modal_act_c.impa.evt"},
	{0x2967781B0C96B49CULL, "vxlt_gallowstemplar_f_5ce_0_0.vlm"},
	{0x2968636785BC32ADULL, "hlac_w_woundedcoast_196615_0x.dds"},
	{0x2968DBDFDBF72504ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm02.phy"},
	{0x2968E51051E81FE5ULL, "6075532_f.fxe"},
	{0x296963D95B6500F5ULL, "llac_ca_cavern1_haunted_24422_5.dds"},
	{0x2969CAAF236F73CDULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_conj_c.impa.evt"},
	{0x296B0A69BF5C13F2ULL, "191320424.ogg"},
	{0x296B4BD9AC8FA2E7ULL, "dae300_alistair_king6134901.cl"},
	{0x296B7347E82F71F0ULL, "6201359_m.fxe"},
	{0x296C0AAED9ADB435ULL, "6153921_m.fxe"},
	{0x296C43FD60CAB616ULL, "6077945_m.fxe"},
	{0x296E8343E7B0D2DFULL, "lgt220_shp_mag_guard6149871.cl"},
	{0x296E9A189B0002FEULL, "vxlt_gamlenshouse_n_6p_0_0.vlm"},
	{0x296EF82D9F27C67BULL, "vxlt_hangedman_n_245_0_0.vlm"},
	{0x296F4FADE2DE5052ULL, "m_dre180ar_deep_roads.wbd"},
	{0x296FFF67D9CEA9E1ULL, "6043876_f.fxe"},
	{0x2970CDE5EDD802E4ULL, "mer220_hubert6041178.cl"},
	{0x2970D56A1EFEFA6EULL, "hlac_w_bonepitdestroyed_a3_458753_0x.dds"},
	{0x2970EC531D6A7009ULL, "777402752.ogg"},
	{0x2971429A114621AFULL, "111108829.ogg"},
	{0x2971C11E66D5925FULL, "468443097.ogg"},
	{0x2971F4B5C6D69942ULL, "hlac_w_bonepitmining_a2_65538_0y.dds"},
	{0x29720046C886EAE7ULL, "vxht_0_hightown_a2_bic_0_0.vlm"},
	{0x29720978B508C580ULL, "719358720.ogg"},
	{0x2972511BD7450F08ULL, "vxht_0_hightown_a3_c87_0_0.vlm"},
	{0x297289057053AD82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_threat_decay.evt"},
	{0x2972A11F4B22933CULL, "646890607.ogg"},
	{0x2972A9EBBC1048F5ULL, "290598020.ogg"},
	{0x2973012707CD94FAULL, "hlac_lt_0_lowtown_a2_-2_4z.dds"},
	{0x29732B11E5B6C6B6ULL, "787625373.ogg"},
	{0x2973946CCEDEFAF9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_brazier.msh"},
	{0x29739AA3E36C7100ULL, "6055042_m.fxe"},
	{0x29739E6CCEDF0BE7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_brazier.mmh"},
	{0x2973A26CCEDF1234ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_brazier.mao"},
	{0x2973CC46BC9CADCCULL, "602934053.ogg"},
	{0x2973F2171D994444ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_cth_homa_0d.dds"},
	{0x29742A24EBE76CDDULL, "vxlt_0_lowtown_e1z_0_0.vlm"},
	{0x29742C4671093F71ULL, "gen00fl_isabela6088872.cl"},
	{0x2974833F7F7C7143ULL, "lgt260ip_at_to_rescuecave.utp"},
	{0x2974B7EF77888A83ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3co_basic.msh"},
	{0x2974C5EF7788A2DDULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3co_basic.mmh"},
	{0x2974C87A2888B875ULL, "vxca_cavern1_haunted_ikm_0_0.vlm"},
	{0x2974C9EF7788A9EEULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3co_basic.mao"},
	{0x2974E50E43766A1FULL, "lt_0_lowtown_docks_n_5_0.rml"},
	{0x2975394653E7D819ULL, "85417367.ogg"},
	{0x297553877BCCA95DULL, "hlac_lt_0_lowtown_docks_a3_-196610_4z.dds"},
	{0x29756777F5078410ULL, "zz_followers_banter6158733.cl"},
	{0x29756B4BE0BC3DD5ULL, "57104750.ogg"},
	{0x2975B0C8296DC5A6ULL, "6064578_m.fxe"},
	{0x2975EA4625E0F1A6ULL, "gencr_raider_ws_f.utc"},
	{0x2976218CF06E8954ULL, "503679079.ogg"},
	{0x297645533055477BULL, "638172518.ogg"},
	{0x2976C82F445C20B3ULL, "zz_followers_banter6159447.cl"},
	{0x2977D5AF2E159A1CULL, "vxca_caverns2prc_7o8_0_1.vlm"},
	{0x2978380E5135FBFEULL, "vxht_0_hightown_a2_c3t_0_1.vlm"},
	{0x29784067ED136F84ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\blendtree\\c_dsogre_cbt.anb"},
	{0x2978A05497C102DEULL, "184650718.ogg"},
	{0x2978B67262B8D6CFULL, "340595041.ogg"},
	{0x2978E99AD40A3281ULL, "llac_lt_0_lowtown_a2_1_5.dds"},
	{0x29790F6B75567AD6ULL, "vxlt_gallowstemplar_f_5a9_0_0.vlm"},
	{0x297947D40052FF5BULL, "545737123.ogg"},
	{0x297977A45EE9F4BEULL, "follower_banter6161027.cl"},
	{0x297988057059CA95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_threat_decay.gad"},
	{0x2979CE02AF67AE18ULL, "character_rendertodepth_p0.psh.fxo.dx11"},
	{0x297A49F7A06FC131ULL, "vxht_keep_5gx_0_0.vlm"},
	{0x297A953FADD6BA81ULL, "508606826.ogg"},
	{0x297AB96BBF69DB11ULL, "295340251.ogg"},
	{0x297AFAE026CEDBF8ULL, "6041241_m.fxe"},
	{0x297BA3EB36CB3A51ULL, "357435608.ogg"},
	{0x297BACF55D9E9709ULL, "hlac_lt_darktown_-1_61x.dds"},
	{0x297BBCA7A8489333ULL, "6096004_m.fxe"},
	{0x297BF15196CF772AULL, "vxw_blightlands_bs_0_0.vlm"},
	{0x297D09C957A8DF4CULL, "6097628_m.fxe"},
	{0x297D1B2E53E25627ULL, "qcr220_orsino6060721.cl"},
	{0x297D48D2867E332CULL, "w_bonepitmining_a2_196_0.msh"},
	{0x297D4ED2867E3D1EULL, "w_bonepitmining_a2_196_0.mmh"},
	{0x297D5AD2867E51B5ULL, "w_bonepitmining_a2_196_0.mao"},
	{0x297D6A0A911DDEFBULL, "vxlt_hangedman_n_234_0_0.vlm"},
	{0x297DADC70A440DCCULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.cess.ani"},
	{0x297DF290856A7739ULL, "6137310_m.fxe"},
	{0x297E8EA93F74B59DULL, "16126395.ogg"},
	{0x297EDD52E70F3C25ULL, "vxlt_0_lowtown_a2_goj_0_0.vlm"},
	{0x297EEE77CB4CF62BULL, "hlac_ht_0_hightown_a2_65537_2x.dds"},
	{0x297F1C6F7C24203DULL, "hlac_lt_0_lowtown_n_-65537_0x.dds"},
	{0x297F32EB6D018E94ULL, "hlac_lt_gallowscourtyard_0_1y.dds"},
	{0x297F9230E71D8CA6ULL, "vxht_chantry_20e_0_0.vlm"},
	{0x297FD584EC822219ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefish.phy"},
	{0x298004D304E90204ULL, "vxlt_house_g5_0_0.vlm"},
	{0x2980602F304A8817ULL, "gen00fl_anders6096516.cl"},
	{0x298084D43E303C5EULL, "isa111_isabela6067119.cl"},
	{0x298134B2B5A09700ULL, "hlac_lt_0_lowtown_a2_n_65535_3y.dds"},
	{0x2981373AA26FA535ULL, "vxlt_undercityserialkiller_175_0_0.vlm"},
	{0x2982495836A5F11FULL, "lt_0_lowtown_a3_n_c_54_0.rml"},
	{0x29827B5B1BC8CBC7ULL, "gen00fl_fenris6108037.cl"},
	{0x29840EAA1170E60BULL, "vxlt_gallowstemplar_56f_0_0.vlm"},
	{0x298418C10E2F0DB6ULL, "vxlt_gallowstemplar_f_3o1_0_0.vlm"},
	{0x298426059205C839ULL, "llac_w_encounter2_262146_0.dds"},
	{0x2984BAC70A4A420EULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.cess.gad"},
	{0x298591843B2CCCB0ULL, "vxlt_gallowstemplar_5re_0_0.vlm"},
	{0x2985DF553B91DC62ULL, "6103220_m.fxe"},
	{0x2986200227B3C2A6ULL, "vxdr_0_deeproads_f60_0_0.vlm"},
	{0x2986D28CF4AA27F2ULL, "6044648_m.fxe"},
	{0x29875144CD22F327ULL, "vxlt_0_lowtown_docks_a3_n_d25_0_0.vlm"},
	{0x298781F169C3B591ULL, "6041068_m.fxe"},
	{0x298786AE01DE1277ULL, "spellfire_blue.mao"},
	{0x29879D6EAB1A962BULL, "llac_lt_0_lowtown_a3_n_c_2_0.dds"},
	{0x2987AFEF93E6EF22ULL, "gen00fl_anders6119243.cl"},
	{0x2987EDF8A23E12BCULL, "w_bonepit_106_0.mao"},
	{0x2987EE169E3776ECULL, "727128906.ogg"},
	{0x2987F9F8A23E275FULL, "w_bonepit_106_0.mmh"},
	{0x2987FFF8A23E3101ULL, "w_bonepit_106_0.msh"},
	{0x298802B2058208C0ULL, "gen00fl_varric6103464.cl"},
	{0x29885D893A6B30B9ULL, "vxlt_gallowstemplar_n_390_0_0.vlm"},
	{0x2988B85A6EF3D1E7ULL, "qcr000_aveline6041083.cl"},
	{0x2989432C7E1A73AFULL, "vxlt_0_lowtown_n_h4y_0_0.vlm"},
	{0x2989A959DD2C0EB4ULL, "6006077_m.fxe"},
	{0x298A2D4FB38AA2F0ULL, "6140195_m.fxe"},
	{0x298A39C6BC00588EULL, "vxw_bonepitdestroyed_a3_29e_0_0.vlm"},
	{0x298A466B27292448ULL, "vxdr_primevalentrance_395_0_1.vlm"},
	{0x298A4909F8FAF0BAULL, "vxht_mansion2b_3c5_0_0.vlm"},
	{0x298AB5C70A4EA5EDULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.cess.evt"},
	{0x298B5AE3477B9135ULL, "6149463_m.fxe"},
	{0x298C55D77E3A28ECULL, "1054323426.ogg"},
	{0x298C98EC6B96DE96ULL, "6101011_m.fxe"},
	{0x298CA9A10337CC09ULL, "vxlt_gallowscourtyard_n_74y_0_0.vlm"},
	{0x298D4F6654271FACULL, "follower_banter6162103.cl"},
	{0x298D7C170E5E4B8BULL, "370855667.ogg"},
	{0x298D91E4E3A38266ULL, "vxlt_gallowstemplar_64z_0_0.vlm"},
	{0x298DB78A5B0F4A2BULL, "gen00fl_aveline6053332.cl"},
	{0x298DC963F1860983ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.open.evt"},
	{0x298DCAB536B35D40ULL, "ht_0_hightown_n_29_0_ll.mmh"},
	{0x298DDA4303D2956DULL, "224885301.ogg"},
	{0x298DF159028C15DFULL, "371750196.ogg"},
	{0x298E1A8080065274ULL, "764505478.ogg"},
	{0x298E4563FA4D2F74ULL, "6071945_m.fxe"},
	{0x299006E1F7E5365AULL, "plt_lgt310pt_nt_melson.plo"},
	{0x2990C2969A26B685ULL, "framebuffereffects_vsmblur_p0.psh.fxo.dx11"},
	{0x2990C58EDD0BD815ULL, "hlac_lt_0_lowtown_a2_0_4x.dds"},
	{0x2990C8D83DBB50C0ULL, "hlac_lt_0_lowtown_docks_-3_0x.dds"},
	{0x2990FBCDCFA2A932ULL, "466935162.ogg"},
	{0x2991102CCA820CADULL, "vxlt_undercityserialkiller_1n2_0_0.vlm"},
	{0x29912C3469E06C90ULL, "drg_clo.mao"},
	{0x29912C8F782E3975ULL, "zz_approval_debug6003756.cl"},
	{0x2991316A9C69C876ULL, "6033592_m.fxe"},
	{0x29913604BF9CEB60ULL, "6088256_f.fxe"},
	{0x299285A9EA3D2CB3ULL, "376517422.ogg"},
	{0x2992A5D2A470120AULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v2.gad"},
	{0x2992CDB1E278FE1CULL, "vxlt_gallowstemplar_f_5lx_0_0.vlm"},
	{0x2992F8BD6F6E5198ULL, "gen_im_trs_dem_lgt_cht_r2.uti"},
	{0x2993EEEDC788FEF9ULL, "6096882_m.fxe"},
	{0x299436CC11EAAB78ULL, "vxlt_0_lowtown_a3_n_c_gco_0_0.vlm"},
	{0x29943E22E07AD1A8ULL, "hlac_lt_0_lowtown_a2_n_-2_3z.dds"},
	{0x299451322D87F4E2ULL, "6104320_m.fxe"},
	{0x29946D72A1784BF6ULL, "mrl310_merrill6119947.cl"},
	{0x29948CFBDDC3197BULL, "6098895_m.fxe"},
	{0x2994A263F18BE604ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.open.gad"},
	{0x2994B79A8A2B1074ULL, "494128231.ogg"},
	{0x2994BC099384E973ULL, "hlac_lt_0_lowtown_docks_131070_3x.dds"},
	{0x2994C3E33BBBABF9ULL, "vxht_brothel_n_5dy_0_0.vlm"},
	{0x2994D7EB17F1EED9ULL, "628669958.ogg"},
	{0x2994E3E4C03C12D7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_large.phy"},
	{0x299527D0E3C9399BULL, "783903122.ogg"},
	{0x2996F45921CCDF0FULL, "zz_vault_debug6177803.cl"},
	{0x299705E32989989EULL, "gen00fl_fenris6116569.cl"},
	{0x29976E838485881BULL, "970795817.ogg"},
	{0x2998ABA3ADF6710EULL, "vxw_bonepitdestroyed_a3_24f_0_0.vlm"},
	{0x2998BAE6EAF1093AULL, "vxht_mansion2b_3f6_0_0.vlm"},
	{0x2998EB94A20B3561ULL, "dae120cr_hf_mage_1.utc"},
	{0x299924912616187DULL, "260699377.ogg"},
	{0x29993A2D42FBEB83ULL, "vxht_keep_62r_0_0.vlm"},
	{0x299988D2A475FFA1ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v2.evt"},
	{0x299A58211C889CE7ULL, "vxlt_0_lowtown_a2_hdb_0_0.vlm"},
	{0x299B23D795F09F34ULL, "6171393_m.fxe"},
	{0x299B53D3DBAC223DULL, "506786706.ogg"},
	{0x299BB163F1921EE2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.open.ani"},
	{0x299C7366F91F7383ULL, "508120586.ogg"},
	{0x299C7A21E7A85F15ULL, "vxlt_0_lowtown_byu_0_0.vlm"},
	{0x299C7DFA70C2BD3DULL, "gen_im_wep_war_one_act3_09.uti"},
	{0x299D5B9F9262D5B8ULL, "6113035_f.fxe"},
	{0x299DA6B7EC8D366EULL, "6050387_f.fxe"},
	{0x299DD74EDA4D5BCFULL, "hlac_w_blightlands_327677_0x.dds"},
	{0x299E1B7984758441ULL, "vxlt_warehouse_n_q4_0_0.vlm"},
	{0x299E1F16B490C586ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampstake.phy"},
	{0x299E327340510C84ULL, "12074101.ogg"},
	{0x299EBCFBB517308DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09b.mmh"},
	{0x299ECEFBB5174F13ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09b.msh"},
	{0x299F170384FBF773ULL, "vxw_encounter2_camp_ja_0_0.vlm"},
	{0x299F56BF027D9FF4ULL, "vxht_keep_n_3i6_0_0.vlm"},
	{0x299F690C1F1C4361ULL, "fx_m2c2.dds"},
	{0x299F8209BC78252DULL, "vxlt_undercityserialkiller_1c1_0_0.vlm"},
	{0x299F876CCF041F1CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_brazier.phy"},
	{0x299FA7E1B19303E0ULL, "6083356_f.fxe"},
	{0x299FBC49EEAAED33ULL, "13919295.ogg"},
	{0x29A0CF8CE6A64468ULL, "w_woundedcoast_7_0.mmh"},
	{0x29A0DB8CE6A6580BULL, "w_woundedcoast_7_0.mao"},
	{0x29A0ED8CE6A67692ULL, "w_woundedcoast_7_0.msh"},
	{0x29A1DC4DC91A0A00ULL, "6105817_m.fxe"},
	{0x29A247AE0AB1377FULL, "6101083_m.fxe"},
	{0x29A2753E1F18FDB0ULL, "vxlt_gallowstemplar_5xc_0_0.vlm"},
	{0x29A27B257603EFE2ULL, "6072982_m.fxe"},
	{0x29A318B4D7AFA483ULL, "plt_cod_aow_spellcombo10.plo"},
	{0x29A31D0B2339D678ULL, "hlac_lt_0_lowtown_docks_-262144_1y.dds"},
	{0x29A329A0FE011881ULL, "6167629_m.fxe"},
	{0x29A3597F376544ACULL, "dae000ip_crowd_mf7_talk10.utp"},
	{0x29A37A5DE03C4F36ULL, "6117971_m.fxe"},
	{0x29A388A6DA7565D3ULL, "6127989_f.fxe"},
	{0x29A3B778F7DDACE2ULL, "llac_ht_0_hightown_a3_-131072_1.dds"},
	{0x29A4A31FCAED7039ULL, "6210309_m.fxe"},
	{0x29A4FA75E23DBC7EULL, "6155314_f.fxe"},
	{0x29A5009D81AB2442ULL, "vxw_encounter1_c2_0_0.vlm"},
	{0x29A53151781430A9ULL, "6129472_m.fxe"},
	{0x29A53EF2B8743E1CULL, "gen00fl_anders6119520.cl"},
	{0x29A54E6A35FD59B3ULL, "mag170_thrask6054058.cl"},
	{0x29A57FBAA3E7AEFEULL, "6051592_m.fxe"},
	{0x29A680D2A4807CF0ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v2.ani"},
	{0x29A6F240E22A0FF2ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava_d.dds"},
	{0x29A71757D02ED81FULL, "6033676_m.fxe"},
	{0x29A790A220AD4549ULL, "985692563.ogg"},
	{0x29A7B7EB3BE92284ULL, "vxht_keep_n_6fv_0_0.vlm"},
	{0x29A7DFE2E5044D72ULL, "llac_lt_0_lowtown_docks_a3_-65539_0.dds"},
	{0x29A7E25A881F7F6CULL, "hlac_lt_0_lowtown_docks_-65538_6x.dds"},
	{0x29A7E4CE8A88C53DULL, "vxlt_0_lowtown_n_ert_0_0.vlm"},
	{0x29A85833D7697661ULL, "556908316.ogg"},
	{0x29A8A3CE85448531ULL, "vxht_0_hightown_c50_0_0.vlm"},
	{0x29A8BCA35457B474ULL, "6067843_m.fxe"},
	{0x29A992432FF56848ULL, "6076944_m.fxe"},
	{0x29A995B487E6B7B4ULL, "6174093_m.fxe"},
	{0x29A9D01910BDA000ULL, "6076396_f.fxe"},
	{0x29AA301C0BB06251ULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.anb"},
	{0x29AA5ADE1D2F389AULL, "vxlt_gallowstemplar_4n6_0_0.vlm"},
	{0x29AA82AF2C7A47C6ULL, "6165376_m.fxe"},
	{0x29AB9E54C5D8FFD5ULL, "vxlt_warehouse_n_15g_0_0.vlm"},
	{0x29ABA789B8FBBD20ULL, "nat380_temmerin6067395.cl"},
	{0x29AC347B53ECA657ULL, "isa111_isabela6164598.cl"},
	{0x29AC512DEFE5B61CULL, "hlac_ca_caverns1_mines_24426_0x.dds"},
	{0x29AC703063F1D08CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj05.mmh"},
	{0x29AC763063F1DB7EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj05.msh"},
	{0x29AD309AA25800A9ULL, "hlac_lt_hangedman_n_-1_23z.dds"},
	{0x29AD5C7E33FD9885ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic_2.mao"},
	{0x29AE83699F746F0CULL, "gen00fl_dog6165702.cl"},
	{0x29AE92BA337BFDF9ULL, "dae300_cullen6140643.cl"},
	{0x29AEA5911B9223F2ULL, "971573954.ogg"},
	{0x29AFB5FBB525984AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09b.phy"},
	{0x29B01531E0550E32ULL, "ser221_gascard6175523.cl"},
	{0x29B0346B9E795413ULL, "hlac_lt_gallowstemplar_n_2394_12y.dds"},
	{0x29B08050FD2BA5E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_staff.ani"},
	{0x29B0831413EAD1E3ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_03oldgrave.mao"},
	{0x29B0871413EAD830ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_03oldgrave.mmh"},
	{0x29B0951413EAF04AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_03oldgrave.msh"},
	{0x29B1227527DDB6FAULL, "gen00fl_fenris6116594.cl"},
	{0x29B17296806BD883ULL, "mag291_demon_marethari6027861.cl"},
	{0x29B1AD1C6E7B686AULL, "dae000_arishok6072889.cl"},
	{0x29B1FE02FD1D208AULL, "llac_w_encounter2_camp_65539_0.dds"},
	{0x29B27795322E7BEEULL, "vxlt_0_lowtown_a2_n_gj5_0_0.vlm"},
	{0x29B2A26E962C7129ULL, "vxlt_0_lowtown_docks_n_aps_0_0.vlm"},
	{0x29B2C706E621C2F4ULL, "gen_im_trs_chn_junk2_04.uti"},
	{0x29B3170C97402807ULL, "gen00fl_anders6125294.cl"},
	{0x29B32BA28460CC44ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_up_c.anb"},
	{0x29B3727A73A13CC2ULL, "vxw_encounter1_h3_0_0.vlm"},
	{0x29B3851E600D1058ULL, "hf_cgn_009.mop"},
	{0x29B3EED0DBF4426AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_exit.gad"},
	{0x29B3FC3106033E4EULL, "339405002.ogg"},
	{0x29B40D33CFDF8AE3ULL, "6060144_m.fxe"},
	{0x29B41C809C5C4485ULL, "vxlt_0_lowtown_a3_n_c_glw_0_0.vlm"},
	{0x29B43F5FB0D49C4DULL, "vxlt_backalley_g_1zz_0_0.vlm"},
	{0x29B4529EE330BD35ULL, "364230070.ogg"},
	{0x29B45C0B2A5F4D79ULL, "sum_conj_04.mao"},
	{0x29B48B4D10ED42F8ULL, "306189463.ogg"},
	{0x29B4A5BCDB290B98ULL, "hlac_ht_mansion_n_3369_0y.dds"},
	{0x29B4E39481FF8210ULL, "vxca_caverns1_mines_e7d_0_0.vlm"},
	{0x29B4E5C1423FDD11ULL, "synclargecreatures2.gda"},
	{0x29B545206F95BA6FULL, "6050125_m.fxe"},
	{0x29B571E8421C901BULL, "rog_paralyzebrittle.dds"},
	{0x29B5E792B802EE3FULL, "dae000ip_crowd_em3_talk5.utp"},
	{0x29B629C82DDF3B04ULL, "vxht_keep_n_6aw_0_0.vlm"},
	{0x29B64D40ACCB0C64ULL, "vxw_woundedcoast_14d_0_0.vlm"},
	{0x29B656AB7C7EDDBDULL, "vxlt_0_lowtown_n_eos_0_0.vlm"},
	{0x29B689088EE8E976ULL, "gen00fl_fenris6105859.cl"},
	{0x29B6DB0570E9992AULL, "vxw_sundermount_2aj_0_0.vlm"},
	{0x29B73477BB35CEA6ULL, "6118592_m.fxe"},
	{0x29B74AB984C5CEDBULL, "hlac_lt_0_lowtown_a3_n_c_131070_2x.dds"},
	{0x29B76E0524E65254ULL, "m_qcr270ar_gallows.bnk"},
	{0x29B791DEEAFFA7F4ULL, "6161442_m.fxe"},
	{0x29B7D3994CF4E068ULL, "hlac_lt_0_lowtown_a2_n_-3_0x.dds"},
	{0x29B7EBAED961AF00ULL, "vxht_mansion_n_4fd_0_0.vlm"},
	{0x29B8511C0BBCD7CAULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.mmh"},
	{0x29B88E00B15469E6ULL, "hf_vis_lowtown3.mor"},
	{0x29B93185C8D248F2ULL, "gen00fl_sebastian6185539.cl"},
	{0x29B9532AC9EA97C9ULL, "6036329_m.fxe"},
	{0x29B9AAA9AB7791EFULL, "pro000_cut2d_voyage6004223.cl"},
	{0x29B9B8E29B665C06ULL, "dae000_keep_noble_sta6161518.cl"},
	{0x29B9C92F191715A2ULL, "422515854.ogg"},
	{0x29BA37F66FFFB90FULL, "534225516.ogg"},
	{0x29BA6EF8570BCA64ULL, "gen00fl_anders6107509.cl"},
	{0x29BAB5C617821AA9ULL, "456763385.ogg"},
	{0x29BAC2FA58399714ULL, "docks_scene_dockworkers6176477.cl"},
	{0x29BAD1D0DBFA2F01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_exit.evt"},
	{0x29BAFAACF5EDCD94ULL, "follower_banter6154018.cl"},
	{0x29BB8E0DB9A70277ULL, "llac_w_sundermount_327684_0.dds"},
	{0x29BBA109153C81BCULL, "967133306.ogg"},
	{0x29BC19260749F9B2ULL, "6148232_m.fxe"},
	{0x29BC63456BF2F9ECULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_beth.msh"},
	{0x29BC69456BF303DEULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_beth.mmh"},
	{0x29BD0391C52E2EDAULL, "6171792_f.fxe"},
	{0x29BD576F4EE6AA91ULL, "follower_banter6167187.cl"},
	{0x29BD9850FD3658DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_staff.evt"},
	{0x29BDEFBF574710EBULL, "6180383_m.fxe"},
	{0x29BDF7B6925E1814ULL, "llac_w_bonepit_393216_0.dds"},
	{0x29BE7A0C1F3710EBULL, "fx_m2c2.mao"},
	{0x29BF1E10B9F1E173ULL, "ser270_templar6127400.cl"},
	{0x29BF2ECA2BF66766ULL, "fx_ground_exp.dds"},
	{0x29BF8C62E7CCF4F8ULL, "vxlt_0_lowtown_a3_n_c_gpj_0_0.vlm"},
	{0x29BFA9ACCC1A400EULL, "vxlt_0_lowtown_csi_0_0.vlm"},
	{0x29C0479064C66434ULL, "vxlt_gallowstemplar_3qz_0_0.vlm"},
	{0x29C04AA2846B8B47ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_up_c.mmh"},
	{0x29C06BCD7C708589ULL, "594111961.ogg"},
	{0x29C06C60BE6196D3ULL, "6129789_f.fxe"},
	{0x29C0D67828356812ULL, "6142490_m.fxe"},
	{0x29C0EE7ABFD72CEDULL, "vxlt_0_lowtown_a2_n_gnv_0_0.vlm"},
	{0x29C0F2EFE1F347FFULL, "6052772_m.fxe"},
	{0x29C11C3CC7D04791ULL, "gen_im_npc_dwf_arm_boo_lgta.uti"},
	{0x29C20416664C7DBEULL, "atl_fonttest_dxt5_00.dds"},
	{0x29C214A369E0770AULL, "574422169.ogg"},
	{0x29C24ED53626F95BULL, "vxlt_blackemporium_vr_0_0.vlm"},
	{0x29C2669ECEB5C781ULL, "hm_hunter.mor"},
	{0x29C28E5D8E525D05ULL, "vxlt_0_lowtown_a3_n_c_gcv_0_0.vlm"},
	{0x29C31DCFBC422D44ULL, "vxht_0_hightown_a3_a89_0_0.vlm"},
	{0x29C338399976A9FEULL, "vxlt_undercityserialkiller_143_0_0.vlm"},
	{0x29C4188543692FBCULL, "6106774_f.fxe"},
	{0x29C49750FD3C75CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_staff.gad"},
	{0x29C4AA33F67B9163ULL, "614248978.ogg"},
	{0x29C4C8886E74F63DULL, "vxlt_0_lowtown_n_ehr_0_0.vlm"},
	{0x29C5236877A613D8ULL, "vxlt_hangedman_w2_0_0.vlm"},
	{0x29C54CE262DFB1AAULL, "vxw_sundermount_2jk_0_0.vlm"},
	{0x29C59B0AC2A2BA4DULL, "sto_cl_dark.mao"},
	{0x29C63502BEE4D46BULL, "vxca_caverns2prc_9wp_0_0.vlm"},
	{0x29C6B4B21C2AE502ULL, "vxht_0_hightown_9zh_0_0.vlm"},
	{0x29C75EC6C21C0137ULL, "6089729_m.fxe"},
	{0x29C7E84BE6A836F5ULL, "vxdr_primevaltomb_85v_0_0.vlm"},
	{0x29C8B191BA3CA1D4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_cth_homa_0t.dds"},
	{0x29C8C9D0DC065F50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_exit.ani"},
	{0x29C9D4ACD944A66BULL, "945655061.ogg"},
	{0x29CAB57B2DA8B6C2ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_ext_c.mmh"},
	{0x29CAC7DC507AC0D9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bigwheel_sa.dds"},
	{0x29CB7215F3FC31B3ULL, "vxlt_0_lowtown_n_e2j_0_0.vlm"},
	{0x29CB9B12CFD273D3ULL, "reticle.psh.fxo"},
	{0x29CBCBA339E61404ULL, "904312641.ogg"},
	{0x29CC1A6D4F5AC770ULL, "vxlt_undercitydungeon_11o_0_0.vlm"},
	{0x29CC29E3C839DBBFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_die_lp.gad"},
	{0x29CCABC0BBF792B2ULL, "ch2_s1y.psh.fxo"},
	{0x29CCDC8799056C9AULL, "6051487_f.fxe"},
	{0x29CD0741C49405C5ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_3y.dds"},
	{0x29CD170D9A3239FCULL, "w_bonepitdestroyed_a3_102_0.mao"},
	{0x29CD230D9A324D9FULL, "w_bonepitdestroyed_a3_102_0.mmh"},
	{0x29CD290D9A325841ULL, "w_bonepitdestroyed_a3_102_0.msh"},
	{0x29CD7B2637453D6AULL, "art\\characters\\creatures\\hurlock\\animation\\blendtree\\hrlk_combat_movement_cb.anb"},
	{0x29CD85E7D5251689ULL, "abi_m_horror.bnk"},
	{0x29CDE0FB9E03E83FULL, "6018510_m.fxe"},
	{0x29CDFE3FD9C30D78ULL, "vxlt_0_lowtown_a3_n_c_gwk_0_0.vlm"},
	{0x29CEF086FFC578B6ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masb.phy"},
	{0x29CF9A8CBBEF3E39ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12ma.msh"},
	{0x29CFA48CBBEF4F27ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12ma.mmh"},
	{0x29CFA88CBBEF5674ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12ma.mao"},
	{0x29CFC9CC02B14357ULL, "hlac_lt_0_lowtown_a2_n_1_7y.dds"},
	{0x29D10A0447B8BFE4ULL, "110728823.ogg"},
	{0x29D12C525295EE88ULL, "l_qun131ar_undercity_warrens.bnk"},
	{0x29D19D702D5990CCULL, "vxlt_0_lowtown_fep_0_0.vlm"},
	{0x29D1BE5C9DE4DD74ULL, "6156167_m.fxe"},
	{0x29D1C9C1B1840A6FULL, "vxlt_0_lowtown_a3_n_c_bxt_0_0.vlm"},
	{0x29D208C305C1E3BDULL, "6071508_m.fxe"},
	{0x29D21573755FB197ULL, "hlac_ca_caverns2_5802_11z.dds"},
	{0x29D259289D4FD6FEULL, "6087475_f.fxe"},
	{0x29D2E5FE8F548420ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1b.phy"},
	{0x29D32509C67B7A76ULL, "6089482_f.fxe"},
	{0x29D39803B70539FEULL, "screenspace.mao"},
	{0x29D3BEBF54D5CA2AULL, "vxw_sundermount_2od_0_0.vlm"},
	{0x29D41CD972CE99F2ULL, "19042714.ogg"},
	{0x29D4A10437EB58DAULL, "vxlt_undercityencounter_we_0_0.vlm"},
	{0x29D5267D3EE221ECULL, "418922505.ogg"},
	{0x29D52810C3D6CDDAULL, "399133675.ogg"},
	{0x29D54C7C0C27889BULL, "vxlt_0_lowtown_docks_brw_0_0.vlm"},
	{0x29D5949C2E472F02ULL, "hlac_ht_chantry_n_2534_47y.dds"},
	{0x29D5B7D56F164B9FULL, "gen00fl_sebastian6178338.cl"},
	{0x29D6A071E42FD725ULL, "dae000_male_commoner1_mob6174466.cl"},
	{0x29D6CAE182C64A0DULL, "6144293_m.fxe"},
	{0x29D6E70C2372490BULL, "clouds.gda"},
	{0x29D73FD5E801CA51ULL, "hlac_w_encounter3_-327683_0z.dds"},
	{0x29D74DFC0FBDBFE8ULL, "vxca_caverns1_mines_j2e_0_1.vlm"},
	{0x29D80717B78F24A9ULL, "vxlt_0_lowtown_a2_n_enx_0_0.vlm"},
	{0x29D83F0E37FBD832ULL, "vxlt_undercityencounter_1bp_0_1.vlm"},
	{0x29D8D986FFCDA657ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masb.msh"},
	{0x29D8EB86FFCDC4BAULL, "art\\characters\\playercharacter\\humanfemale\\arm_masb.mao"},
	{0x29D8EF86FFCDCBC9ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masb.mmh"},
	{0x29D96BD779051DC7ULL, "63873756.ogg"},
	{0x29D9BE8A2A249A8BULL, "follower_banter6141769.cl"},
	{0x29D9C28769DF0A8AULL, "hlac_w_bonepitdestroyed_a3_131073_0y.dds"},
	{0x29D9DE51B580ED26ULL, "vxdr_0_deeproads_cwn_0_0.vlm"},
	{0x29DA17CA0DAD0D91ULL, "vxht_mansion_n_599_0_0.vlm"},
	{0x29DA725D6E518AB3ULL, "hlac_ht_0_hightown_n_-131071_1z.dds"},
	{0x29DA804A2A9E193EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_2_exit.evt"},
	{0x29DAE0E09029C1A4ULL, "gen00fl_varric6075769.cl"},
	{0x29DB05755BA75331ULL, "6163188_m.fxe"},
	{0x29DBC151AB7D0051ULL, "hlac_lt_0_lowtown_docks_-196610_3x.dds"},
	{0x29DBE2B10A262EBCULL, "dae000_pc_home_ambients6208589.cl"},
	{0x29DBF5E9BAA7BD90ULL, "vxdr_dungeon_1ls_0_0.vlm"},
	{0x29DC1A9A99295400ULL, "6145196_m.fxe"},
	{0x29DC8D66B006710EULL, "vxlt_0_lowtown_cmk_0_0.vlm"},
	{0x29DC9C7CA0E2C17BULL, "6048960_m.fxe"},
	{0x29DCB8E4C0797F2FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_large.mao"},
	{0x29DCC4E4C07993CCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_large.mmh"},
	{0x29DCCAE4C0799DBEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_large.msh"},
	{0x29DD6FA8F767E1B4ULL, "932584603.ogg"},
	{0x29DD8377374C03D9ULL, "478631811.ogg"},
	{0x29DD9DAD58189F6CULL, "vxca_caverns2_62o_0_1.vlm"},
	{0x29DDB9FE4288D564ULL, "6164315_m.fxe"},
	{0x29DDD145F0ABA6C0ULL, "art\\vfx\\materials\\vfx_smoke_puffy_add.mao"},
	{0x29DDF905F2E7907FULL, "and111_lirene6084515.cl"},
	{0x29DEDEB86FFA94A8ULL, "mag371_elsa6041261.cl"},
	{0x29DEFA3514B0EBA0ULL, "gen00fl_varric6063854.cl"},
	{0x29E0289460CDC33BULL, "285486301.ogg"},
	{0x29E049A4C2A41EE9ULL, "6084430_m.fxe"},
	{0x29E0936693360897ULL, "vxlt_foundry_n_y0_0_0.vlm"},
	{0x29E106E3C84BFC05ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_die_lp.ani"},
	{0x29E1974A2AA45F69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_2_exit.gad"},
	{0x29E1CCBDFAE91899ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgtb.mmh"},
	{0x29E1D6BDFAE92987ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgtb.msh"},
	{0x29E1FA455D236DF6ULL, "6147803_f.fxe"},
	{0x29E2309C46CBE2AAULL, "vxw_sundermount_2pe_0_0.vlm"},
	{0x29E237B5D3292583ULL, "6094763_m.fxe"},
	{0x29E24CB41DF42490ULL, "6090207_m.fxe"},
	{0x29E25E0ADCE28B95ULL, "588786407.ogg"},
	{0x29E2812B33C35979ULL, "6164736_m.fxe"},
	{0x29E298FE052420B0ULL, "hlac_w_blightlands_-65537_0z.dds"},
	{0x29E312E129E1715AULL, "vxlt_undercityencounter_rf_0_0.vlm"},
	{0x29E416BD30ED1946ULL, "hlac_lt_0_lowtown_docks_a3_-4_1y.dds"},
	{0x29E431E633EFFA12ULL, "63319495.ogg"},
	{0x29E52ACF87002049ULL, "rog_threatdecay.dds"},
	{0x29E5E988F27584BDULL, "hlac_lt_0_lowtown_a2_n_65534_2x.dds"},
	{0x29E60AF2DF4F841FULL, "729429284.ogg"},
	{0x29E63F22CF820917ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfb.phy"},
	{0x29E676BA15C2581BULL, "vxlt_gallowscourtyard_7es_0_0.vlm"},
	{0x29E708DEBAD6FBFDULL, "hlac_lt_gallowstemplar_n_2391_9x.dds"},
	{0x29E70A23540AAF02ULL, "6151069_m.fxe"},
	{0x29E70C72E2E5D64AULL, "6111353_m.fxe"},
	{0x29E8502EA77705A6ULL, "vxdr_0_deeproads_clo_0_0.vlm"},
	{0x29E8684A2AAA2DB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_2_exit.ani"},
	{0x29E8A177DE25930DULL, "llac_ht_chantry_n_2534_3.dds"},
	{0x29E8ECC0E51F1C04ULL, "s_stand_01.dds"},
	{0x29E90401126138E6ULL, "vxca_caverns1_mines_gka_0_1.vlm"},
	{0x29E9430D01980F63ULL, "vxw_bonepitmining_a2_7x_0_0.vlm"},
	{0x29E95C9F50E0591DULL, "681111520.ogg"},
	{0x29E9744176C08931ULL, "62517615.ogg"},
	{0x29E97C88936891FAULL, "pssm.psh.fxo"},
	{0x29E9AFB01E01663EULL, "dae000_meredith6092552.cl"},
	{0x29EA02F8A29100A4ULL, "w_bonepit_106_0.phy"},
	{0x29EA1EC55BDD0865ULL, "ave_act3_guardf.mor"},
	{0x29EA3176837421EEULL, "672040151.ogg"},
	{0x29EA6BB4BDDCE101ULL, "vxht_mansion_n_4xw_0_0.vlm"},
	{0x29EA9F7FA6057791ULL, "vxw_sundermount_31i_0_1.vlm"},
	{0x29EAC2EC5B73668CULL, "isa211_isabela6070192.cl"},
	{0x29EAEE4737CBC85FULL, "pro000_aveline_intro6109374.cl"},
	{0x29EB45CE8505B454ULL, "6087728_m.fxe"},
	{0x29EBAAB11CD7AB08ULL, "art\\characters\\weapons\\staff\\w_stf_011a.msh"},
	{0x29EBBCB11CD7C999ULL, "art\\characters\\weapons\\staff\\w_stf_011a.mao"},
	{0x29EBC8B11CD7DE32ULL, "art\\characters\\weapons\\staff\\w_stf_011a.mmh"},
	{0x29EC666691BC95ECULL, "fade.mao"},
	{0x29EC91303173988FULL, "204949391.ogg"},
	{0x29ED5107DF839C70ULL, "fx_energylink01.mao"},
	{0x29ED5AC595A820EDULL, "47353074.ogg"},
	{0x29EDB646F9885CE1ULL, "lt_hangedman_n_mmi.gff"},
	{0x29EE0BF234C54301ULL, "vxlt_gallowsprison_n_6mu_0_0.vlm"},
	{0x29EEACC509874B06ULL, "508288074.ogg"},
	{0x29EF0543852C2117ULL, "vxlt_foundry_n_t3_0_0.vlm"},
	{0x29EF503B3CED2063ULL, "6126493_m.fxe"},
	{0x29EF51465527C06FULL, "gen00fl_aveline6121929.cl"},
	{0x29EF520FD64D54D8ULL, "89871906.ogg"},
	{0x29EF677676295C95ULL, "dae000_gamlen6088138.cl"},
	{0x29EFF4440DA2323BULL, "214118426.ogg"},
	{0x29EFFD3CD3EAAF47ULL, "vxw_bonepitdestroyed_a3_lc_0_0.vlm"},
	{0x29F0122782A5EEEEULL, "237454793.ogg"},
	{0x29F03391B2AA9D05ULL, "gen00fl_fenris6126376.cl"},
	{0x29F0470555D28B55ULL, "lt_0_lowtown_a2_n_27_0_lc.mmh"},
	{0x29F04C29792C0596ULL, "plt_cod_aow_level_up.plo"},
	{0x29F07FDF5B55965FULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbench_0.pwk"},
	{0x29F08E141420C84BULL, "art\\levels\\races\\natural\\highlands\\props\\prp_03oldgrave.phy"},
	{0x29F0939DD9C65405ULL, "zz_personality_debug6064392.cl"},
	{0x29F1E7761879A859ULL, "7352552.ogg"},
	{0x29F2506BFCB31E02ULL, "89991822.ogg"},
	{0x29F2869E1651C0D0ULL, "6128777_m.fxe"},
	{0x29F2B23D9732BC59ULL, "llac_lt_gallowscourtyard_-65538_3.dds"},
	{0x29F2C96403C4C6E6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedflurrish.evt"},
	{0x29F33A04C287F241ULL, "mag110_arianni6016808.cl"},
	{0x29F34B5A90882DABULL, "vxlt_0_lowtown_a2_n_c_htw_0_0.vlm"},
	{0x29F35AFC0D14520EULL, "512012791.ogg"},
	{0x29F392F9BA9E16B1ULL, "dae011_conspiracy_mob6140816.cl"},
	{0x29F3B47B2DCBBA69ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_ext_c.anb"},
	{0x29F3E0C1F9E504B7ULL, "hlac_lt_undercityencounter_-1_36x.dds"},
	{0x29F635A3831EE41BULL, "1022538778.ogg"},
	{0x29F65DBCA8403573ULL, "vxht_0_hightown_n_as5_0_0.vlm"},
	{0x29F71E5252B6A632ULL, "l_qun131ar_undercity_warrens.wbd"},
	{0x29F74B9B66212B2AULL, "183992228.ogg"},
	{0x29F7848D8C1AC6D5ULL, "vxht_bartrand_n_5si_0_1.vlm"},
	{0x29F78805251C6206ULL, "m_qcr270ar_gallows.wbd"},
	{0x29F79445D9AA0A36ULL, "hlac_lt_undercitydungeon_-1_52x.dds"},
	{0x29F7B039BED252AFULL, "219889691.ogg"},
	{0x29F802FE8F73D1B5ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1b.msh"},
	{0x29F804FE8F73D52BULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1b.mmh"},
	{0x29F810FE8F73E9C8ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1b.mao"},
	{0x29F90BCD4E7BD0B3ULL, "vxca_cave_164_0_1.vlm"},
	{0x29F92EB5D0E6AAB9ULL, "vxlt_gallowstemplar_5ut_0_1.vlm"},
	{0x29F9E06403CB0C31ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedflurrish.gad"},
	{0x29F9ED18011D1A1BULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_4x.dds"},
	{0x29FB68DB2D653DB5ULL, "167146754.ogg"},
	{0x29FBFE326630FE50ULL, "llac_lt_0_lowtown_docks_a3_-1_3.dds"},
	{0x29FC3F94D470C1EDULL, "gen00fl_aveline6116482.cl"},
	{0x29FC40456C28A279ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_beth.phy"},
	{0x29FC4E8D2A5E3E3AULL, "qun200_party_comment6095263.cl"},
	{0x29FCA8F6B9C1388AULL, "vxht_brothel_n_57n_0_1.vlm"},
	{0x29FDD64F3C3CA66CULL, "6096339_m.fxe"},
	{0x29FDD70941144104ULL, "llac_dr_primevaltomb_659_4.dds"},
	{0x29FE4060E6AD1144ULL, "6128360_m.fxe"},
	{0x29FFA66DBCE0D944ULL, "66731133.ogg"},
	{0x29FFFB4F0F4D0933ULL, "lt_0_lowtown_a3_n_c_43_0.anb"},
	{0x2A00CDA87DE79B04ULL, "lt_0_lowtown_docks_n_249_0.msh"},
	{0x2A00DFA87DE7BA6DULL, "lt_0_lowtown_docks_n_249_0.mao"},
	{0x2A00E16403D12CDFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedflurrish.ani"},
	{0x2A00E3A87DE7C0A6ULL, "lt_0_lowtown_docks_n_249_0.mmh"},
	{0x2A010DA8B8CA383DULL, "248919392.ogg"},
	{0x2A0124951E1100D7ULL, "15741232.ogg"},
	{0x2A0179AA3C1F5EF6ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_3webdetail.mao"},
	{0x2A017AE777179127ULL, "578743520.ogg"},
	{0x2A0185AA3C1F7295ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_3webdetail.mmh"},
	{0x2A0187AA3C1F760BULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_3webdetail.msh"},
	{0x2A0192F568C46926ULL, "vxht_chantry_2en_0_0.vlm"},
	{0x2A01C515C1CF6918ULL, "6174912_m.fxe"},
	{0x2A01F2215BDE56A6ULL, "6077013_m.fxe"},
	{0x2A01F6954BB2D264ULL, "6119658_m.fxe"},
	{0x2A02EB9668343BB4ULL, "gen_im_trs_gen_med_boo_r2.uti"},
	{0x2A0329C348FB7A16ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_gise.phy"},
	{0x2A0368D743A8297AULL, "gen00fl_aveline6137701.cl"},
	{0x2A0413401C47BD6DULL, "w_encounter3_22_0.mmh"},
	{0x2A0417401C47C3BEULL, "w_encounter3_22_0.mao"},
	{0x2A0425401C47DBF3ULL, "w_encounter3_22_0.msh"},
	{0x2A042C7025BDF659ULL, "6125703_f.fxe"},
	{0x2A043A3E7E52981CULL, "vxlt_gallowstemplar_n_5pd_0_0.vlm"},
	{0x2A04AA44D41AC9CDULL, "6134715_m.fxe"},
	{0x2A04CF999A364DF3ULL, "vxht_0_hightown_n_at4_0_0.vlm"},
	{0x2A051C4176967033ULL, "vxht_keep_n_7a9_0_0.vlm"},
	{0x2A0533E88B6336A6ULL, "vxdr_0_deeproads_cfq_0_0.vlm"},
	{0x2A058D8F881E6E65ULL, "6036951_f.fxe"},
	{0x2A059733FB5D205BULL, "45290537.ogg"},
	{0x2A0626C6E5844063ULL, "vxw_bonepitmining_a2_9z_0_0.vlm"},
	{0x2A06C966DB6C9C9AULL, "vxht_0_hightown_c80_0_0.vlm"},
	{0x2A06DC17590407AFULL, "450881551.ogg"},
	{0x2A078451D680F08CULL, "vxlt_gallowstemplar_f_4qf_0_0.vlm"},
	{0x2A079D96F08B83BAULL, "hlac_w_bonepitdestroyed_a3_1_0y.dds"},
	{0x2A07E688079F287CULL, "lgt360_mag_orlanna6155395.cl"},
	{0x2A0827D274601BCBULL, "175973520.ogg"},
	{0x2A083095399EFA6BULL, "ts_profane_02.gda"},
	{0x2A08386F1457986BULL, "6082147_f.fxe"},
	{0x2A089A0839208926ULL, "288398394.ogg"},
	{0x2A08CD18977CE408ULL, "vxlt_0_lowtown_docks_b6m_0_1.vlm"},
	{0x2A08E374ABC0F942ULL, "vxlt_hangedman_21l_0_0.vlm"},
	{0x2A09FD541B710A31ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockb.mmh"},
	{0x2A0A09541B711ED2ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockb.mao"},
	{0x2A0A0E0A09B5CA65ULL, "abi_m_chain_light.wbd"},
	{0x2A0A17541B7136EFULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockb.msh"},
	{0x2A0A49B2CA830210ULL, "6025311_m.fxe"},
	{0x2A0AB98D8159165DULL, "140343725.ogg"},
	{0x2A0AE2663DD619A9ULL, "qun120_arishok6051329.cl"},
	{0x2A0AF76502BA5E76ULL, "6118264_m.fxe"},
	{0x2A0C1F744CE096F1ULL, "338453502.ogg"},
	{0x2A0C4FEDD938418AULL, "zz_isa_debug6175188.cl"},
	{0x2A0C7AAB19E115DCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev4_b.ani"},
	{0x2A0D12C34903A7B7ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_gise.msh"},
	{0x2A0D28C34903CD29ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_gise.mmh"},
	{0x2A0D8CE77B52FC58ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_basic.phy"},
	{0x2A0E12C1C8E38BBEULL, "llac_lt_0_lowtown_docks_a3_-262144_3.dds"},
	{0x2A0E13401C5011E6ULL, "w_encounter3_22_0.rml"},
	{0x2A0E1BF16F671834ULL, "one300_zevran6076627.cl"},
	{0x2A0E23C9D324EEFDULL, "6118797_m.fxe"},
	{0x2A0E7099533F672EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_market_stall_0_sa.dds"},
	{0x2A0E77C319270800ULL, "llac_w_sundermount_131066_0.dds"},
	{0x2A0EA0D55D49FDD8ULL, "sst_ss_qunari_mage.gda"},
	{0x2A0EA6A6B6F27A56ULL, "6076602_f.fxe"},
	{0x2A0F13EFD3FFEA9BULL, "vxlt_0_lowtown_docks_bfk_0_0.vlm"},
	{0x2A0F59C611E66B6DULL, "hlac_lt_gamlenshouse_n_-1_5x.dds"},
	{0x2A0FAB27CEB211EAULL, "vxlt_warehouse_rc_0_0.vlm"},
	{0x2A110EE9C2789D5AULL, "6137270_m.fxe"},
	{0x2A1118535965D6C5ULL, "vxht_mansion2_2id_0_1.vlm"},
	{0x2A112D26E9AF1DF4ULL, "6177279_f.fxe"},
	{0x2A11605954BD7F1CULL, "469628577.ogg"},
	{0x2A1162F109980468ULL, "vxlt_hangedman_16x_0_0.vlm"},
	{0x2A118EF665C4E599ULL, "286894361.ogg"},
	{0x2A11D3387C19C068ULL, "6075173_f.fxe"},
	{0x2A1287AB19E5971EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev4_b.gad"},
	{0x2A13823F5E6895F3ULL, "dre000_varric6028417.cl"},
	{0x2A139430E23250A5ULL, "mag110_arianni6016916.cl"},
	{0x2A13A5C57D594F26ULL, "vxdr_0_deeproads_ccr_0_0.vlm"},
	{0x2A13ACF88E1BC5F5ULL, "vxca_cavern1_haunted_i2r_0_0.vlm"},
	{0x2A144F79E09B473DULL, "vxw_bonepitmining_a2_d3_0_0.vlm"},
	{0x2A148725BA98D742ULL, "mrl150_merrill6078565.cl"},
	{0x2A1489743652C4ADULL, "hlac_lt_0_lowtown_a2_n_1_6x.dds"},
	{0x2A14EC5132F7C31DULL, "zz_approval_debug6003565.cl"},
	{0x2A150047F1C9E38DULL, "vxlt_0_lowtown_docks_c4v_0_0.vlm"},
	{0x2A1563E7D56292DBULL, "abi_m_horror.wbd"},
	{0x2A157B34061E3E1AULL, "zz_followers_banter6157617.cl"},
	{0x2A15880C5CAE2824ULL, "mer340_bonepit_comments6083159.cl"},
	{0x2A159516CEB39501ULL, "6109711_m.fxe"},
	{0x2A15D18EE7125E56ULL, "vxlt_gallowstemplar_n_5v8_0_0.vlm"},
	{0x2A15F62EC877090CULL, "vxlt_gallowstemplar_f_4vg_0_0.vlm"},
	{0x2A16AD6C6BA6E4A8ULL, "dae000_gamlen6102705.cl"},
	{0x2A16C1C8DB011392ULL, "vxlt_backalley_n_1zl_0_0.vlm"},
	{0x2A1973850B25802EULL, "379438477.ogg"},
	{0x2A19A4030615C9B7ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_7z.dds"},
	{0x2A19AFB19758627DULL, "882901911.ogg"},
	{0x2A1A9B505618739DULL, "6179860_m.fxe"},
	{0x2A1ACD3B2DA75C70ULL, "vxca_caverns1_ck7_0_0.vlm"},
	{0x2A1AEF856BD98DE8ULL, "698597451.ogg"},
	{0x2A1B00F5699D18CAULL, "vxw_bonepitmining_a2_ao_0_0.vlm"},
	{0x2A1BAC0C428DDF3EULL, "849090033.ogg"},
	{0x2A1C1B5EE10287C3ULL, "fx_lava_add.mao"},
	{0x2A1C2ECF45535CC0ULL, "vxca_caverns1_cos_0_0.vlm"},
	{0x2A1C36043255B891ULL, "6118955_m.fxe"},
	{0x2A1CFE5019AC5BDFULL, "mcr370_assignments6081731.cl"},
	{0x2A1D53F463F3E71BULL, "vxca_cavern1_haunted_i9y_0_0.vlm"},
	{0x2A1D6544EE9BF371ULL, "llac_ca_caverns1_mines_24424_1.dds"},
	{0x2A1E2D146EE9E290ULL, "6082235_m.fxe"},
	{0x2A1EDA4F2F9F0364ULL, "6113058_m.fxe"},
	{0x2A1F4287F0F2EBE9ULL, "vxlt_hangedman_26b_0_0.vlm"},
	{0x2A1FB1C481EA4B3EULL, "300905474.ogg"},
	{0x2A1FED3B0135F686ULL, "344944716.ogg"},
	{0x2A202341BB2A992AULL, "zz_vault_debug6177790.cl"},
	{0x2A2047406530ACA7ULL, "hlac_lt_hangedman_n_-1_18z.dds"},
	{0x2A20C33B11580902ULL, "vxht_chantry_n_26p_0_0.vlm"},
	{0x2A20DD52367F6E1FULL, "6118320_m.fxe"},
	{0x2A2144C5ECAE8872ULL, "hlac_lt_0_lowtown_a2_n_65535_8z.dds"},
	{0x2A21D1EA4026EEAEULL, "vxlt_darktown_471_0_0.vlm"},
	{0x2A2292A70FF9C727ULL, "mcr310_argument6075061.cl"},
	{0x2A22D8F87C8EB276ULL, "hlac_dr_slavers_3217_12x.dds"},
	{0x2A22D9685BBA670BULL, "pro000_captain_wright6004333.cl"},
	{0x2A22E8B573DB4334ULL, "vxht_0_hightown_a2_ajr_0_0.vlm"},
	{0x2A22F3287703F38BULL, "vxlt_gallowstemplar_5as_0_0.vlm"},
	{0x2A237224E3BFFC0DULL, "vxlt_0_lowtown_docks_c1u_0_0.vlm"},
	{0x2A23D40673ADB77FULL, "229995317.ogg"},
	{0x2A24CE2BDA0E42E1ULL, "ave321cr_guardsman4.utc"},
	{0x2A24DDFDA3B75E63ULL, "llac_w_bonepit_65538_0.dds"},
	{0x2A24EE72B6BFC9DBULL, "219232800.ogg"},
	{0x2A24EF11F5F78AFBULL, "lt_0_lowtown_a2_19_0_lc.mmh"},
	{0x2A251636BE29756CULL, "160459254.ogg"},
	{0x2A252250923C990EULL, "gen_im_npc_wep_sta_orsa.uti"},
	{0x2A2535C57FFB1809ULL, "vxlt_gallowstemplar_n_4qg_0_0.vlm"},
	{0x2A25EA5EEEC56D1FULL, "735908186.ogg"},
	{0x2A2680F13B9DDE8AULL, "zz_lgt_1_debug6136989.cl"},
	{0x2A268DE4A0D79515ULL, "vxca_caverns1_cjy_0_0.vlm"},
	{0x2A26B6C8CDA573D4ULL, "hlac_lt_0_lowtown_a3_n_c_-3_2y.dds"},
	{0x2A27582073D24209ULL, "vxlt_gallowstemplar_f_45s_0_0.vlm"},
	{0x2A27F5AC7E23578BULL, "llac_lt_0_lowtown_a2_1_7.dds"},
	{0x2A28422ECAEC653EULL, "vxw_sundermount_380_0_0.vlm"},
	{0x2A28AC9BE6F26014ULL, "art\\characters\\playercharacter\\humanmale\\rob_and.phy"},
	{0x2A28B5D039BE71CFULL, "one121_trap6112696.cl"},
	{0x2A28DFAE27868224ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower1_.gad"},
	{0x2A28FE5C4474E1C8ULL, "311400849.ogg"},
	{0x2A293F181F9D74F0ULL, "vxca_caverns1_cn4_0_0.vlm"},
	{0x2A2972D25B93314AULL, "vxw_bonepitmining_a2_lh_0_0.vlm"},
	{0x2A298AB2A2808232ULL, "vxht_mansion2b_2kd_0_0.vlm"},
	{0x2A29B15AAE08BC90ULL, "ser221_party_comments6166192.cl"},
	{0x2A29D7ACF21ABE27ULL, "776477409.ogg"},
	{0x2A29F9D0D7FBC82AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_prostitute_2.gad"},
	{0x2A2B9C08C8B3185EULL, "38946202.ogg"},
	{0x2A2BD165896DF5F5ULL, "6092051_m.fxe"},
	{0x2A2BFA7FE7204F5EULL, "dae000ip_crowd_lm2_talk15.utp"},
	{0x2A2C184F0F7275EFULL, "lt_0_lowtown_a3_n_c_43_0.rml"},
	{0x2A2C93B0F862CF17ULL, "zz_followers_banter6158787.cl"},
	{0x2A2C9EF160DFFB10ULL, "6085335_m.fxe"},
	{0x2A2CBEE9F84C41E3ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_2x.dds"},
	{0x2A2D1422CFBDC36FULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfb.mao"},
	{0x2A2D2022CFBDD70CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfb.mmh"},
	{0x2A2D2622CFBDE1FEULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfb.msh"},
	{0x2A2D6E728089FA82ULL, "6121314_m.fxe"},
	{0x2A2DAB01EAA5F0D3ULL, "hlac_ca_caverns2_5802_66x.dds"},
	{0x2A2E31194AFD5109ULL, "hlac_lt_blackemporium_1055_15y.dds"},
	{0x2A2E53985E91851BULL, "hlac_lt_0_lowtown_65536_1z.dds"},
	{0x2A2E56561DEC6013ULL, "vxdr_primevaltomb_6mj_0_0.vlm"},
	{0x2A2E5E46CCDE4E05ULL, "llac_ca_caverns1_24426_2.dds"},
	{0x2A2E64A5812C7C21ULL, "aip_follower_sebastian.gda"},
	{0x2A2E6CA4CE4C6313ULL, "293891623.ogg"},
	{0x2A2E8D95F9F8CB35ULL, "vxlt_undercityserialkiller_1s9_0_0.vlm"},
	{0x2A2E987742EF1D7BULL, "hlac_lt_0_lowtown_docks_-131074_8z.dds"},
	{0x2A2EFDC675CDFD8AULL, "1002104257.ogg"},
	{0x2A2F1AD3DFFAD500ULL, "vxlt_gallowsprison_n_67j_0_0.vlm"},
	{0x2A2F8B88E8A1F9D3ULL, "vxht_0_hightown_avq_0_0.vlm"},
	{0x2A2FEEAE278CBA02ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower1_.ani"},
	{0x2A2FEFC90AD00A2BULL, "6076782_f.fxe"},
	{0x2A302C0D9A86DAE4ULL, "w_bonepitdestroyed_a3_102_0.phy"},
	{0x2A3052DC551ECB62ULL, "follower_banter6170313.cl"},
	{0x2A307693C3D4DAA5ULL, "vxht_mansion_n_4vu_0_0.vlm"},
	{0x2A30DCD0D801B5C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_prostitute_2.evt"},
	{0x2A311B3CB1E5E893ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\blendtrees\\crowds.anb"},
	{0x2A3129D55DF25DE1ULL, "6156870_m.fxe"},
	{0x2A313333C487783DULL, "vxw_bonepitmining_a2_n1_0_0.vlm"},
	{0x2A31650568FA0C0BULL, "vxlt_gallowstemplar_5jr_0_0.vlm"},
	{0x2A31A38ACA8CCF61ULL, "6027389_m.fxe"},
	{0x2A3202D21C185AB6ULL, "art\\vfx\\cinematic\\vfx_mother_tears.impa.ani"},
	{0x2A32899BE6FA7911ULL, "art\\characters\\playercharacter\\humanmale\\rob_and.msh"},
	{0x2A32979BE6FA912CULL, "art\\characters\\playercharacter\\humanmale\\rob_and.mao"},
	{0x2A32A39BE6FAA5CFULL, "art\\characters\\playercharacter\\humanmale\\rob_and.mmh"},
	{0x2A32F29FA68131F0ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultmain.msh"},
	{0x2A33009FA681490AULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultmain.mmh"},
	{0x2A33049FA6815041ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultmain.mao"},
	{0x2A33469975DB83B0ULL, "131764194.ogg"},
	{0x2A334B0FC73DD6B8ULL, "mer120_hubert6059402.cl"},
	{0x2A33A7A271F13089ULL, "vxlt_gallowstemplar_n_4vf_0_0.vlm"},
	{0x2A340D6747C9B203ULL, "6174696_m.fxe"},
	{0x2A348388AB94524CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamphanging.phy"},
	{0x2A353BD14C026B5CULL, "6099231_m.fxe"},
	{0x2A35854E06691E03ULL, "w_encounter3_139_0.mao"},
	{0x2A35894E06692550ULL, "w_encounter3_139_0.mmh"},
	{0x2A35974E06693D6AULL, "w_encounter3_139_0.msh"},
	{0x2A36150CA18F4B94ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.impa.ani"},
	{0x2A362146C08EEF88ULL, "6129194_m.fxe"},
	{0x2A3649B985C5AB1FULL, "vxdr_0_deeproads_fpc_0_0.vlm"},
	{0x2A36CC123886F226ULL, "762147069.ogg"},
	{0x2A3755EA20D2EDFBULL, "fx_ice_splat2.mao"},
	{0x2A3757FC18ECFEE5ULL, "hlac_w_encounter2_camp_65535_0y.dds"},
	{0x2A376647343E8BF4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudbasic.phy"},
	{0x2A378A2E243237CFULL, "6069382_f.fxe"},
	{0x2A37B257804DB0B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_injured.anb"},
	{0x2A37DE04B7D6A0C4ULL, "6175155_m.fxe"},
	{0x2A37E4AF4D8949CAULL, "vxw_bonepitmining_a2_ki_0_0.vlm"},
	{0x2A37EF0F4334A4C4ULL, "6201478_m.fxe"},
	{0x2A385968050B88BAULL, "w_encounter3_18_0.mmh"},
	{0x2A385D68050B8FF1ULL, "w_encounter3_18_0.mao"},
	{0x2A386B68050BA700ULL, "w_encounter3_18_0.msh"},
	{0x2A386F81C19592F1ULL, "167211687.ogg"},
	{0x2A38DD9CA1BCBE9DULL, "zz_setphysics.ncs"},
	{0x2A3A0C9C568A79EDULL, "63092966.ogg"},
	{0x2A3A41B9212766D7ULL, "vxca_cavern1_haunted_eh2_0_0.vlm"},
	{0x2A3B3574ECC57961ULL, "follower_banter6155418.cl"},
	{0x2A3CB2A579FDD8F4ULL, "266900283.ogg"},
	{0x2A3CF8F707BE9703ULL, "mag101_grace6053069.cl"},
	{0x2A3D120CA19565F6ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.impa.gad"},
	{0x2A3D1685A4642063ULL, "6153376_m.fxe"},
	{0x2A3D1C5046BD169FULL, "lt_0_lowtown_docks_n_lc.mmh"},
	{0x2A3D3528C8EBEC26ULL, "6153540_m.fxe"},
	{0x2A3D3AAF20B70F7CULL, "vxlt_0_lowtown_a2_n_epc_0_0.vlm"},
	{0x2A3D61C5D0D7A873ULL, "375947767.ogg"},
	{0x2A3D8EC25A18F9E1ULL, "297969306.ogg"},
	{0x2A3D9945F9202E75ULL, "vxw_bonepitdestroyed_a3_1iy_0_0.vlm"},
	{0x2A3DB21F66168A6AULL, "dae000_sandal6147819.cl"},
	{0x2A3DC5A1785C17ACULL, "6013325_m.fxe"},
	{0x2A3DD4D0D80C3210ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_prostitute_2.ani"},
	{0x2A3DEBC3825E37D9ULL, "271730139.ogg"},
	{0x2A3F2D78F856C9A9ULL, "269909888.ogg"},
	{0x2A3F4C835BA9BB66ULL, "6139566_m.fxe"},
	{0x2A3FD6E25AF0248BULL, "vxlt_gallowstemplar_5wu_0_0.vlm"},
	{0x2A400AD21C24A5A7ULL, "art\\vfx\\cinematic\\vfx_mother_tears.impa.evt"},
	{0x2A401D9769DF1A4DULL, "llac_dr_slavers_3217_0.dds"},
	{0x2A40E74317E28222ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_pc.gad"},
	{0x2A40F0F07392FCA1ULL, "llac_w_encounter2_131077_0.dds"},
	{0x2A40F5209A6DB2A5ULL, "gen00fl_isabela6088933.cl"},
	{0x2A4116E269A17401ULL, "vxht_mansion_n_4fi_0_0.vlm"},
	{0x2A4143473446A4F1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudbasic.msh"},
	{0x2A4151473446BC0CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudbasic.mao"},
	{0x2A415D473446D0AFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudbasic.mmh"},
	{0x2A41684CEE8802F9ULL, "vxht_0_hightown_a2_c3n_0_0.vlm"},
	{0x2A41B250EF0D016AULL, "llac_lt_0_lowtown_docks_a3_65535_1.dds"},
	{0x2A420476B97DBA38ULL, "vxlt_undercityserialkiller_1ib_0_0.vlm"},
	{0x2A427A2C1E7F0454ULL, "6061791_m.fxe"},
	{0x2A427B12142B1B33ULL, "6018745_m.fxe"},
	{0x2A42A08D6CBA8C19ULL, "vxlt_hangedman_4z_0_0.vlm"},
	{0x2A432093296D6421ULL, "hlac_lt_0_lowtown_docks_a3_n_0_3z.dds"},
	{0x2A4349FBA3536E31ULL, "hlac_dr_primevalentrance_1726_10x.dds"},
	{0x2A435EB5D8F5C266ULL, "359378070.ogg"},
	{0x2A43CB76D8EECCD1ULL, "gen_im_cns_cft_psn_05.uti"},
	{0x2A43CDBDFB3BE486ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgtb.phy"},
	{0x2A43E6335F34E9F3ULL, "llac_w_sundermount_327679_0.dds"},
	{0x2A44016978181284ULL, "qun221_petrice6075978.cl"},
	{0x2A440239EF3C37C5ULL, "vxht_0_hightown_a3_a2j_0_0.vlm"},
	{0x2A442D0CA19BB255ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.impa.evt"},
	{0x2A445C20E59B78C3ULL, "6072820_m.fxe"},
	{0x2A45D382CAC1294DULL, "vxlt_0_lowtown_docks_a3_n_d2f_0_1.vlm"},
	{0x2A4627C2240F71BEULL, "165148362.ogg"},
	{0x2A462BD21C2949F0ULL, "art\\vfx\\cinematic\\vfx_mother_tears.impa.gad"},
	{0x2A4709D2A85B2E45ULL, "180032669.ogg"},
	{0x2A471EBAADFB376EULL, "darktown_male_elf_sta6174546.cl"},
	{0x2A47275E7B96CA69ULL, "164698360.ogg"},
	{0x2A47C7B11D263B2DULL, "art\\characters\\weapons\\staff\\w_stf_011a.phy"},
	{0x2A480A4317E8DC19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_pc.evt"},
	{0x2A484D84F684A800ULL, "hlac_ht_0_hightown_-131071_1x.dds"},
	{0x2A4879C7E672EE59ULL, "vxdr_0_deeproads_d3b_0_0.vlm"},
	{0x2A48899F1FCD5AD5ULL, "6091832_m.fxe"},
	{0x2A48B428B391B6F4ULL, "ibeam_pressed.cur"},
	{0x2A4985E6575B75CFULL, "llac_lt_0_lowtown_-65534_0.dds"},
	{0x2A49E80A09EB6D0FULL, "abi_m_chain_light.bnk"},
	{0x2A4A7E0FE8EA5B30ULL, "461720615.ogg"},
	{0x2A4AED0068804D07ULL, "vxht_0_hightown_n_c4r_0_0.vlm"},
	{0x2A4B242986BF0134ULL, "933437324.ogg"},
	{0x2A4B2897A15BB771ULL, "lt_0_lowtown_docks_a3_172_0.msh"},
	{0x2A4B3697A15BCE8CULL, "lt_0_lowtown_docks_a3_172_0.mao"},
	{0x2A4B3A1001D89113ULL, "vxdr_primevaltomb_6wl_0_0.vlm"},
	{0x2A4B4297A15BE32FULL, "lt_0_lowtown_docks_a3_172_0.mmh"},
	{0x2A4C0871067E01C6ULL, "6083432_m.fxe"},
	{0x2A4CA731C605A4E9ULL, "ch1armtnt.psh.fxo"},
	{0x2A4CFC1885CE2E08ULL, "dre000_cut_arrivalprimeval6033789.cl"},
	{0x2A4D1AE57D666E2DULL, "118236310.ogg"},
	{0x2A4D3F039C5C5793ULL, "dae200_saemus6091684.cl"},
	{0x2A4D7B7ADAE83B19ULL, "ht_0_hightown_a3_20_0_ll.mmh"},
	{0x2A4DC7957FBFA4D1ULL, "mag370_meredith6039463.cl"},
	{0x2A4E6B000D72A7ECULL, "isa211_isabela6071681.cl"},
	{0x2A4F3B656B1C2379ULL, "hlac_lt_0_lowtown_docks_a3_-65538_4z.dds"},
	{0x2A4F6F805C5B6150ULL, "l_env_wal_lowtown_wal_cry.bnk"},
	{0x2A4FDF1CC858C1CCULL, "static_static_p16.psh.fxo.dx11"},
	{0x2A5076A9CAB33CB4ULL, "205267796.ogg"},
	{0x2A509E6DC3C21ECFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swiss_d.dds"},
	{0x2A50E965A4551ECCULL, "vxlt_0_lowtown_n_df0_0_0.vlm"},
	{0x2A50EF265187692AULL, "llac_lt_backalley_g_0_19.dds"},
	{0x2A512B969C719B6CULL, "hlac_ca_caverns1_mines_24424_14y.dds"},
	{0x2A51808F1B9B6A35ULL, "15666249.ogg"},
	{0x2A51D03C02DB8EBEULL, "rdr110dg_guard_warehouse6016695.cl"},
	{0x2A5239CF5035AF0DULL, "plt_mrl000pt_n_midstage.nss"},
	{0x2A52DD2569960144ULL, "obliskbasen_normals.dds"},
	{0x2A52DF976C1B581CULL, "w_bonepitdestroyed_a3_19_0_lc.mmh"},
	{0x2A52E64A5A7D364FULL, "vxdr_primevalentrance_1m5_0_0.vlm"},
	{0x2A52F356E1254395ULL, "vxw_encounter2_gx_0_0.vlm"},
	{0x2A5310151C155DC1ULL, "6129032_m.fxe"},
	{0x2A540E23CDC146B7ULL, "603980746.ogg"},
	{0x2A541B3629C2AC10ULL, "dae000_arishok6082026.cl"},
	{0x2A5488816A0AF77BULL, "164558020.ogg"},
	{0x2A5491D2DDA6D97CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_1.mao"},
	{0x2A54D11F0FA3E3F6ULL, "vxlt_shop_cg_0_0.vlm"},
	{0x2A55124317F37408ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_pc.ani"},
	{0x2A556D6E8D0C8511ULL, "fadeclouds.psh.fxo"},
	{0x2A55A48A5FEB2370ULL, "vxw_encounter2_c5_0_0.vlm"},
	{0x2A55B21EC0EE681BULL, "vxw_bonepitdestroyed_a3_1es_0_0.vlm"},
	{0x2A5606B87DF779BCULL, "vxw_bonepitmining_a2_8b_0_0.vlm"},
	{0x2A563CF605DF522DULL, "6154784_m.fxe"},
	{0x2A567BFF1C195540ULL, "vxht_0_hightown_a2_au7_0_0.vlm"},
	{0x2A57061E779723C0ULL, "vxw_encounter2_oq_0_0.vlm"},
	{0x2A57889E4DEB03A7ULL, "zz_approval_debug6003772.cl"},
	{0x2A578BC6E6EBD4F5ULL, "vxlt_undercityencounter_ui_0_1.vlm"},
	{0x2A5850D007C0E610ULL, "lt_0_lowtown_docks_n_164_0.msh"},
	{0x2A585ED007C0FE2AULL, "lt_0_lowtown_docks_n_164_0.mmh"},
	{0x2A5862D007C10561ULL, "lt_0_lowtown_docks_n_164_0.mao"},
	{0x2A587620C5C41C57ULL, "63264493.ogg"},
	{0x2A5885CC5A5AA307ULL, "563211924.ogg"},
	{0x2A58C4A076613CBFULL, "6039000_f.fxe"},
	{0x2A595EDD5A766587ULL, "vxht_0_hightown_n_c1m_0_0.vlm"},
	{0x2A5B042E4DDF0C87ULL, "gen_im_cft_rec_psn_03.uti"},
	{0x2A5B31FA373491C6ULL, "dae000cr_grand_cleric6174436.cl"},
	{0x2A5BCC2A99B72425ULL, "vxht_mansion_n_4gr_0_0.vlm"},
	{0x2A5CAAE2309118A4ULL, "270116054.ogg"},
	{0x2A5CAE885EB09BC7ULL, "gen_im_trs_slv_one_axe_r2.uti"},
	{0x2A5CE2394FC6E3A3ULL, "914142299.ogg"},
	{0x2A5D36C209434DEEULL, "6201564_m.fxe"},
	{0x2A5D97C2A63927BCULL, "fx_ripple01.dds"},
	{0x2A5E6339BE583AD1ULL, "gen00fl_aveline6139311.cl"},
	{0x2A5E742E6E4E84BDULL, "dae310ip_store4.utp"},
	{0x2A5F405D61418DCAULL, "art\\vfx\\combat\\vfx_str_stomp_act_c.anb"},
	{0x2A5FAF05940C5C90ULL, "gen_im_cft_reg_f.uti"},
	{0x2A5FF3D6C13C8487ULL, "hlac_w_encounter2_camp_-65538_0y.dds"},
	{0x2A604D78AFA6E55FULL, "llac_lt_0_lowtown_a2_n_c_131072_3.dds"},
	{0x2A607E8F109E5B44ULL, "isa310ip_to_warehouse.utp"},
	{0x2A608831FBF3C185ULL, "826492820.ogg"},
	{0x2A60A406FB781BD0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxb.mao"},
	{0x2A60A806FB7822E3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxb.mmh"},
	{0x2A60B606FB783A3DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxb.msh"},
	{0x2A60CEFCC537F08AULL, "vxdr_dungeon_1mn_0_0.vlm"},
	{0x2A610D35DFD8BF3FULL, "vxlt_darktown_4k0_0_0.vlm"},
	{0x2A613D452FFA0EBDULL, "83235033.ogg"},
	{0x2A6185CEB0801077ULL, "6082092_m.fxe"},
	{0x2A6258CC727199C3ULL, "vxlt_0_lowtown_a2_n_gxr_0_0.vlm"},
	{0x2A62AB00712510E4ULL, "719648616.ogg"},
	{0x2A62C7405B0E29FCULL, "zz_prmpt_debug6200945.cl"},
	{0x2A6306E828C1DD19ULL, "vxlt_0_lowtown_n_ewg_0_0.vlm"},
	{0x2A630E4089020AE2ULL, "422809786.ogg"},
	{0x2A6371052834C656ULL, "vxlt_0_lowtown_docks_n_bcr_0_0.vlm"},
	{0x2A638D3C8FB163F0ULL, "135815582.ogg"},
	{0x2A63DC0BE272CCB0ULL, "smoke_addmore.mao"},
	{0x2A64166751E13BF0ULL, "vxw_encounter2_n2_0_0.vlm"},
	{0x2A6423FBB2E4809BULL, "vxw_bonepitdestroyed_a3_1nr_0_0.vlm"},
	{0x2A643D8C1C890F7DULL, "467311519.ogg"},
	{0x2A6457EFC89206A6ULL, "6087010_m.fxe"},
	{0x2A6459680530C3D1ULL, "w_encounter3_18_0.rml"},
	{0x2A64CCF3A7983FEBULL, "816562786.ogg"},
	{0x2A64EDDC0E0F6DC0ULL, "vxht_0_hightown_a2_ap4_0_0.vlm"},
	{0x2A650924FDA85FCDULL, "243798884.ogg"},
	{0x2A650D01248450B0ULL, "526989823.ogg"},
	{0x2A65254535028162ULL, "1013924860.ogg"},
	{0x2A6538356DA7C049ULL, "vxlt_0_lowtown_n_gff_0_0.vlm"},
	{0x2A656628807FBBE4ULL, "1034005230.ogg"},
	{0x2A656EDBDB284F72ULL, "6151335_m.fxe"},
	{0x2A66731C1AC90732ULL, "73780454.ogg"},
	{0x2A671C5273706AA7ULL, "868597846.ogg"},
	{0x2A671FAF17F86627ULL, "hlac_lt_darktown_-1_9z.dds"},
	{0x2A675863543129D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipa.msh"},
	{0x2A67626354313AD4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipa.mmh"},
	{0x2A676663543141E7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipa.mao"},
	{0x2A67CFB48CC1916DULL, "art\\characters\\weapons\\staff\\textures\\w_stf_002a_0s.dds"},
	{0x2A67E80959BDD3EFULL, "6097831_f.fxe"},
	{0x2A680DA0B577F3C9ULL, "6187775_m.fxe"},
	{0x2A683B983A5778E4ULL, "307207304.ogg"},
	{0x2A68D57790251F46ULL, "gen_im_trs_qun_blt_r1.uti"},
	{0x2A68E98848465756ULL, "vxca_cavern1_haunted_ics_0_1.vlm"},
	{0x2A69143539B96BB6ULL, "vxlt_gallowstemplar_n_3s5_0_0.vlm"},
	{0x2A692F4F2BEE66D4ULL, "6001211_f.fxe"},
	{0x2A694809F39F90C3ULL, "dae000_arishok6078178.cl"},
	{0x2A6952FCB07A5BD3ULL, "vxht_0_hightown_aju_0_0.vlm"},
	{0x2A6979E77BA13AEDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_basic.msh"},
	{0x2A6987E77BA15220ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_basic.mao"},
	{0x2A698BE77BA15973ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_basic.mmh"},
	{0x2A69921665E84FFEULL, "1058019666.ogg"},
	{0x2A69BCF31F32C7AEULL, "762935114.ogg"},
	{0x2A6A020154BD4D8DULL, "765816217.ogg"},
	{0x2A6A42ECFDA400DCULL, "6137593_m.fxe"},
	{0x2A6A4A52D7E8C0BAULL, "vxlt_gallowstemplar_f_3r6_0_0.vlm"},
	{0x2A6A5D3F164373E6ULL, "vxlt_gallowstemplar_n_38p_0_0.vlm"},
	{0x2A6B66DA540BA0FCULL, "gen00fl_anders6142542.cl"},
	{0x2A6B991BA703326FULL, "hlac_lt_gallowstemplar_n_2391_0y.dds"},
	{0x2A6C1C8250A65EF1ULL, "vxca_caverns2_951_0_0.vlm"},
	{0x2A6C4ACE40A42295ULL, "follower_banter6172725.cl"},
	{0x2A6CA896A1EE6AFDULL, "326563309.ogg"},
	{0x2A6CC169508DA90DULL, "6072545_m.fxe"},
	{0x2A6CC50E8179F62DULL, "gen00fl_bethany6101241.cl"},
	{0x2A6CDEB5E0646223ULL, "vxht_mansion2_1rm_0_0.vlm"},
	{0x2A6D700BE9A5E699ULL, "hlac_lt_darktown_-1_75y.dds"},
	{0x2A6D7EECE2A8666EULL, "6147620_m.fxe"},
	{0x2A6E321857B25EC6ULL, "6056889_m.fxe"},
	{0x2A6E4341B18F88A9ULL, "319087175.ogg"},
	{0x2A6F9A53AFA54BEAULL, "79464033.ogg"},
	{0x2A6FD94549F5EA95ULL, "vxw_woundedcoast_1ht_0_1.vlm"},
	{0x2A70517AE6336F2BULL, "follower_barks6173402.cl"},
	{0x2A70ABCCCB7EC090ULL, "vxca_caverns1_mines_ehi_0_0.vlm"},
	{0x2A70B0E6F6150AAEULL, "6095996_m.fxe"},
	{0x2A71A906FB869708ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxb.phy"},
	{0x2A7235EF4F5CF2E6ULL, "fires_sa.dds"},
	{0x2A72D3B271CEFA52ULL, "vxht_0_hightown_ajp_0_0.vlm"},
	{0x2A72EC3A2976938BULL, "6134762_m.fxe"},
	{0x2A73264CBC45EB9DULL, "154927007.ogg"},
	{0x2A7342B9F6D99E8BULL, "dae000_female_tranquil6165122.cl"},
	{0x2A735D44E789F2B9ULL, "dae000_fem_commoner2_mob.bnk"},
	{0x2A73B3EAD0B37F92ULL, "vxlt_0_lowtown_docks_a3_bgn_0_0.vlm"},
	{0x2A73C5C2E8D47FBEULL, "114010258.ogg"},
	{0x2A7451805C7A4A1AULL, "l_env_wal_lowtown_wal_cry.wbd"},
	{0x2A7457F139FEBC12ULL, "6098710_f.fxe"},
	{0x2A745A88ABC9F017ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamphanging.mmh"},
	{0x2A745E88ABC9F724ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamphanging.mao"},
	{0x2A747088ABCA1589ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamphanging.msh"},
	{0x2A755B3AE9E5F87DULL, "zz_one_debug6099929.cl"},
	{0x2A757EFB373B09EDULL, "vxw_sundermount_2z9_0_1.vlm"},
	{0x2A75944030939FB5ULL, "ambient_activity_patterns.gda"},
	{0x2A75C38553E51153ULL, "gen00fl_fenris6110894.cl"},
	{0x2A767F44E78C6219ULL, "dae000_fem_commoner2_mob.cnv"},
	{0x2A769F1E219931FAULL, "338414474.ogg"},
	{0x2A76A0947E74C7F7ULL, "lt_shop_mmap.dds"},
	{0x2A778F368FBB5C65ULL, "gen00fl_isabela6100057.cl"},
	{0x2A77C4D9A2707453ULL, "vxht_0_hightown_amt_0_0.vlm"},
	{0x2A78259586B18B39ULL, "lt_0_lowtown_docks_n_47_0.rml"},
	{0x2A783330DD1C1522ULL, "hlac_lt_0_lowtown_a2_-196608_0z.dds"},
	{0x2A784857170599E6ULL, "134092531.ogg"},
	{0x2A78DF2B64373683ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.cess.evt"},
	{0x2A790C679DB30DEAULL, "vxw_encounter3_4m_0_0.vlm"},
	{0x2A7970C8A6E3ED70ULL, "vxlt_undercitydungeon_1ms_0_0.vlm"},
	{0x2A797CF8DCD310B4ULL, "vxw_encounter3_g7_0_1.vlm"},
	{0x2A798FEDBC01202CULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_mag_0d.dds"},
	{0x2A7A4BD90BF86638ULL, "6137650_m.fxe"},
	{0x2A7A85C35B3D9F34ULL, "6068730_m.fxe"},
	{0x2A7A900798E17020ULL, "follower_banter6143322.cl"},
	{0x2A7AC07A5EE6D200ULL, "vxlt_warehouse_n_1d5_0_0.vlm"},
	{0x2A7B6B2A61651660ULL, "6151377_m.fxe"},
	{0x2A7B8945F54B8156ULL, "one111_gamlen6101413.cl"},
	{0x2A7BC6693232A5AAULL, "fen110_fenris6118063.cl"},
	{0x2A7D5523588AF964ULL, "gen_im_trs_cot_junk3_02.uti"},
	{0x2A7E17D81C2901EEULL, "857035815.ogg"},
	{0x2A7E4B223BEC0315ULL, "vxw_woundedcoast_1su_0_1.vlm"},
	{0x2A7EC134000DA80BULL, "vxdr_primevalentrance_186_0_0.vlm"},
	{0x2A7ECCD7C49BF11BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderready.evt"},
	{0x2A7F0E4FCEB2C48CULL, "qcr000_arishok6166988.cl"},
	{0x2A7F14D8D959C839ULL, "mag291_demon_marethari6028875.cl"},
	{0x2A7F2F70138EA44DULL, "hlac_lt_backalley_n_65537_7y.dds"},
	{0x2A7F35AEE812EDD9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad_0.pwk"},
	{0x2A7FB82B643D1304ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.cess.gad"},
	{0x2A800B961D3AC3CBULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyo_0d.dds"},
	{0x2A80164548723D94ULL, "art\\vfx\\combat\\vfx_gib_undead_bld_p.impa.ani"},
	{0x2A80327C178FC262ULL, "one200_gold_followers6101135.cl"},
	{0x2A8054BF0C20F756ULL, "vxlt_0_lowtown_docks_n_but_0_0.vlm"},
	{0x2A807297E620D76CULL, "follower_banter6163643.cl"},
	{0x2A80F027E398A8D2ULL, "hlac_lt_0_lowtown_docks_-65540_2y.dds"},
	{0x2A81379FE1466629ULL, "vxlt_house_gd_0_0.vlm"},
	{0x2A8161ADDE0C1F10ULL, "6025100_m.fxe"},
	{0x2A81D9B6994A6DBAULL, "vxlt_0_lowtown_docks_n_b5x_0_0.vlm"},
	{0x2A81F01FBA22DCE7ULL, "166622509.ogg"},
	{0x2A823C08C53D1C26ULL, "6019219_m.fxe"},
	{0x2A82BEA33409F7C1ULL, "6162107_m.fxe"},
	{0x2A834151B378D8D6ULL, "vxht_keep_7a7_0_0.vlm"},
	{0x2A840BDBDB8875C0ULL, "6176477_m.fxe"},
	{0x2A8439695AA8FB60ULL, "49794616.ogg"},
	{0x2A844B118F0F621FULL, "llac_ht_chantry_n_2534_32.dds"},
	{0x2A849BCA3A409B0CULL, "6027678_m.fxe"},
	{0x2A84E6776881C36FULL, "rdr110dg_party_comments6094687.cl"},
	{0x2A854474A6A87256ULL, "6136307_m.fxe"},
	{0x2A854B950FB266C0ULL, "zz_keithytest.cut"},
	{0x2A858E4CF9F87590ULL, "vxca_caverns2_8e9_0_0.vlm"},
	{0x2A859643B91052FFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_nud_0s.dds"},
	{0x2A85C5D7C4A2045CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderready.gad"},
	{0x2A86541C014AAB67ULL, "hlac_lt_0_lowtown_docks_n_-4_1z.dds"},
	{0x2A869A6DE25CD115ULL, "w_encounter1_27_0.phy"},
	{0x2A86C72B64434BE2ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.cess.ani"},
	{0x2A871345487857F6ULL, "art\\vfx\\combat\\vfx_gib_undead_bld_p.impa.gad"},
	{0x2A8728B90E0E3593ULL, "hlac_lt_0_lowtown_a2_n_c_131071_6z.dds"},
	{0x2A87900C9672DD5DULL, "one300_nuncio6070811.cl"},
	{0x2A88435D616497BDULL, "art\\vfx\\combat\\vfx_str_stomp_act_c.mmh"},
	{0x2A88789D2546EC43ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\mt_eye_nrma_0d.dds"},
	{0x2A8975A74CBF8E48ULL, "6144353_m.fxe"},
	{0x2A89B547189A7D5DULL, "vxht_mansion_n_6t5_0_0.vlm"},
	{0x2A89C26FC4509323ULL, "vxht_mansion2_1xo_0_0.vlm"},
	{0x2A89D04B690DACFDULL, "mag121_idunna6084240.cl"},
	{0x2A8A2F784F02FCD4ULL, "vis110_commoner2.utc"},
	{0x2A8A874D8CCD69E8ULL, "ht_0_hightownexplosion_n_1_0.msh"},
	{0x2A8A994D8CCD8879ULL, "ht_0_hightownexplosion_n_1_0.mao"},
	{0x2A8AA54D8CCD9C12ULL, "ht_0_hightownexplosion_n_1_0.mmh"},
	{0x2A8B1711C181ADDEULL, "hlac_lt_0_lowtown_n_-65536_4x.dds"},
	{0x2A8B28FE904C7911ULL, "hlac_w_bonepit_262144_0x.dds"},
	{0x2A8B2C25E406B17CULL, "hlac_w_woundedcoast_65543_0z.dds"},
	{0x2A8B403678CE9E68ULL, "304030394.ogg"},
	{0x2A8B7727445076CCULL, "hlac_w_encounter3_0_8y.dds"},
	{0x2A8BC4D7C4A66E5AULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderready.ani"},
	{0x2A8BCA579D080543ULL, "6027398_m.fxe"},
	{0x2A8C249C47BDDB51ULL, "6034636_f.fxe"},
	{0x2A8CBB6DE2617541ULL, "w_encounter1_27_0.rml"},
	{0x2A8D14B6F6C935AAULL, "vxlt_warehouse_n_a3_0_0.vlm"},
	{0x2A8D8F86AF6AF190ULL, "vxca_caverns1_mines_ebw_0_0.vlm"},
	{0x2A8D969F86302F15ULL, "plt_isa000pt_gift_talisman.plo"},
	{0x2A8D97AB23D166C9ULL, "257265624.ogg"},
	{0x2A8DC07F659A7DB9ULL, "lt_warehouse_water_766.msh"},
	{0x2A8DCA7F659A8EA7ULL, "lt_warehouse_water_766.mmh"},
	{0x2A8DCE7F659A95F4ULL, "lt_warehouse_water_766.mao"},
	{0x2A8E2E45487EA455ULL, "art\\vfx\\combat\\vfx_gib_undead_bld_p.impa.evt"},
	{0x2A8E359F0F61E90DULL, "615933897.ogg"},
	{0x2A8E42644E68A8B4ULL, "vxdr_primevaltomb_765_0_0.vlm"},
	{0x2A8EA2E4FF6CC833ULL, "dae000_bodhan_letter6154784.cl"},
	{0x2A8EAD3FD0E03220ULL, "vxlt_0_lowtown_docks_n_b7h_0_0.vlm"},
	{0x2A8EB5A249BBBB21ULL, "gen_im_cft_rec_pot_03.uti"},
	{0x2A8EB5EB1BBFA730ULL, "vxdr_primevalentrance_1q2_0_0.vlm"},
	{0x2A8F9BE25E8C9EE6ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_fallingleaf.mmh"},
	{0x2A8FE283C2454165ULL, "6081095_m.fxe"},
	{0x2A902A7844B1B60EULL, "art\\vfx\\combat\\vfx_w_bsk_haste_activ_c.impa.ani"},
	{0x2A9097A4B49FB092ULL, "vxlt_0_lowtown_docks_a3_bqh_0_0.vlm"},
	{0x2A90CEE50527244EULL, "6018826_f.fxe"},
	{0x2A90D6B19E7C1294ULL, "851492845.ogg"},
	{0x2A90F701F7CD647EULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_maj_c.impa.ani"},
	{0x2A90FDA09E7615B7ULL, "vxht_chantry_n_2v9_0_0.vlm"},
	{0x2A90FF9F51A4A84FULL, "605482518.ogg"},
	{0x2A9114B455A6C71BULL, "vxlt_0_lowtown_docks_b5b_0_0.vlm"},
	{0x2A91788E1F51CC2DULL, "235910610.ogg"},
	{0x2A9180E807023305ULL, "6129847_m.fxe"},
	{0x2A91EF35F00E0A44ULL, "w_bonepitmining_a2_253_0.mmh"},
	{0x2A91F335F00E1097ULL, "w_bonepitmining_a2_253_0.mao"},
	{0x2A920535F00E2FE6ULL, "w_bonepitmining_a2_253_0.msh"},
	{0x2A92498A41E33383ULL, "plt_cod_mgc_chant_maker.plo"},
	{0x2A9274222A1DF2A9ULL, "230420642.ogg"},
	{0x2A934390E1B00939ULL, "264489271.ogg"},
	{0x2A935C54EEA20628ULL, "600726748.ogg"},
	{0x2A940DE622EB4723ULL, "llac_lt_0_lowtown_docks_a3_65534_5.dds"},
	{0x2A940E7969E2953DULL, "mag320_orsino6071130.cl"},
	{0x2A946501B40CCAD4ULL, "6129833_m.fxe"},
	{0x2A94DF0B1E2706CEULL, "685002340.ogg"},
	{0x2A956BA6134586BFULL, "310330081.ogg"},
	{0x2A9577F05CD304C2ULL, "zz_interactive.ncs"},
	{0x2A962447D2140CC5ULL, "6067868_m.fxe"},
	{0x2A9727CD382F5D5EULL, "281391099.ogg"},
	{0x2A976FF26AFEFA41ULL, "zz_followers_banter6158433.cl"},
	{0x2A97BBFB5E49DBD5ULL, "dae000_arishok6072848.cl"},
	{0x2A97F5CE910AA300ULL, "hlac_lt_0_lowtown_a2_-1_8y.dds"},
	{0x2A9823F160CEED6CULL, "vrc321_housekeeper6095703.cl"},
	{0x2A98344CB646ABA3ULL, "vxht_mansion2_1cn_0_0.vlm"},
	{0x2A988D32681A9050ULL, "112029144.ogg"},
	{0x2A98BD89884086C7ULL, "llac_ht_chantry_n_2534_1.dds"},
	{0x2A98C14D2662B820ULL, "co_02_f.mao"},
	{0x2A98EF0DC4E58400ULL, "gen00fl_sebastian6175017.cl"},
	{0x2A99932EBF035750ULL, "fxe_snow_tiny_add.mao"},
	{0x2A99EA9620F08CF7ULL, "300943143.ogg"},
	{0x2A9A18A025FDAD72ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.dura.gad"},
	{0x2A9A4D8F3CD8146EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopea.phy"},
	{0x2A9A7BEF5E7A1B43ULL, "304337483.ogg"},
	{0x2A9A8D75F9F67B1CULL, "681010437.ogg"},
	{0x2A9A91D7879788ABULL, "226262021.ogg"},
	{0x2A9B0BB61A87FFAFULL, "vxlt_0_lowtown_a3_n_c_f5c_0_0.vlm"},
	{0x2A9B4A95212576D5ULL, "fen110_anso6094968.cl"},
	{0x2A9BC2F281BBF2FEULL, "6139705_f.fxe"},
	{0x2A9BC95CF7F5AF1BULL, "vxdr_primevaltomb_8xs_0_0.vlm"},
	{0x2A9C0163A1610A10ULL, "vxca_caverns1_mines_egt_0_0.vlm"},
	{0x2A9C5E1EC00B525DULL, "222065749.ogg"},
	{0x2A9C8418F327956BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comd_0s.dds"},
	{0x2A9C9D8B1CAB011CULL, "plt_cod_ltr_forbethany.nss"},
	{0x2A9CDE13086EB478ULL, "6113233_f.fxe"},
	{0x2A9CF83F2CC95D74ULL, "709979306.ogg"},
	{0x2A9CFA258A3F0DD3ULL, "art\\vfx\\materials\\vfx_crack_tile.mao"},
	{0x2A9D227844BC335FULL, "art\\vfx\\combat\\vfx_w_bsk_haste_activ_c.impa.evt"},
	{0x2A9D2CEA82A5E877ULL, "1015476805.ogg"},
	{0x2A9E187C4793773FULL, "character_character_p136.psh.fxo"},
	{0x2A9F0F01F7D9CA4FULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_maj_c.impa.evt"},
	{0x2A9F3B7911B09F54ULL, "zz_followers_banter6159378.cl"},
	{0x2A9FDCA5EAA85E3BULL, "lt_0_lowtown_docks_39_0.rml"},
	{0x2A9FE9E8ECB7BD69ULL, "vxlt_0_lowtown_docks_b6p_0_1.vlm"},
	{0x2AA08422D969A544ULL, "6172244_m.fxe"},
	{0x2AA0B7C92DDAEC5BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basicnb.msh"},
	{0x2AA0C9C92DDB0AA6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basicnb.mao"},
	{0x2AA0D5C92DDB1F45ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basicnb.mmh"},
	{0x2AA10039D7FB5446ULL, "vxht_0_hightown_a3_3ub_0_0.vlm"},
	{0x2AA11BA02603D0A9ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.dura.evt"},
	{0x2AA1A6191C6CA160ULL, "plt_lgt000pt_civ_breech.nss"},
	{0x2AA1F1E6E1B5B32CULL, "mcr310_sibling6076738.cl"},
	{0x2AA28F7D0AD1382DULL, "hlac_lt_0_lowtown_a2_65538_2y.dds"},
	{0x2AA296F02B4BB636ULL, "6148184_m.fxe"},
	{0x2AA3AC0ED7F7BB0EULL, "gen00fl_fenris6105846.cl"},
	{0x2AA4037844C21D48ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_activ_c.impa.gad"},
	{0x2AA50AB75808B361ULL, "hlac_lt_0_lowtown_n_131071_0z.dds"},
	{0x2AA5795677A056E7ULL, "vxht_keep_n_5gm_0_0.vlm"},
	{0x2AA58E71404F3383ULL, "vxlt_warehouse_w1_0_0.vlm"},
	{0x2AA5CCF4E6CE5962ULL, "393616880.ogg"},
	{0x2AA5E0CFE48F0875ULL, "hlac_lt_hangedman_n_-1_27z.dds"},
	{0x2AA61001F7DFEAF8ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_maj_c.impa.gad"},
	{0x2AA668DA6BFC72F5ULL, "6126367_m.fxe"},
	{0x2AA705EB6D0B6C6AULL, "vxht_mansion_n_615_0_0.vlm"},
	{0x2AA7AE12798E18B3ULL, "496260558.ogg"},
	{0x2AA7B5DD064EF231ULL, "vxdr_primevaltomb_131_0_0.vlm"},
	{0x2AA859D899C4FE11ULL, "626219146.ogg"},
	{0x2AA871398817BB68ULL, "vxw_encounter2_camp_fz_0_0.vlm"},
	{0x2AA8E17DD78079F6ULL, "art\\levels\\ztools\\default\\generic\\9ff_barrel.msh"},
	{0x2AA8EB7DD7808AF4ULL, "art\\levels\\ztools\\default\\generic\\9ff_barrel.mmh"},
	{0x2AA8EF7DD7809107ULL, "art\\levels\\ztools\\default\\generic\\9ff_barrel.mao"},
	{0x2AA901A645DC110EULL, "gen00fl_merrill6129471.cl"},
	{0x2AA97A52BB036D1BULL, "6036706_f.fxe"},
	{0x2AAA4BA34CA5079EULL, "fen110_party_comment6095251.cl"},
	{0x2AAA568FEFF5A0F3ULL, "hlac_lt_0_lowtown_-2_0y.dds"},
	{0x2AAAA00A82AF0B0CULL, "lla_ht_0_hightown_a2_-1_0.dds"},
	{0x2AAB009C07A7FD13ULL, "sib131ar_family_estate.ncs"},
	{0x2AAB28CF5081BB1EULL, "plt_mrl000pt_n_midstage.plo"},
	{0x2AAB3404CC82A192ULL, "182193172.ogg"},
	{0x2AAB660007E637E7ULL, "vxdr_slavers_1am_0_0.vlm"},
	{0x2AABE3082B9943E7ULL, "vxdr_primevalentrance_1j4_0_0.vlm"},
	{0x2AAC1DE35A31F1FDULL, "67043964.ogg"},
	{0x2AAC96941F7C4B53ULL, "hlac_lt_0_lowtown_a2_n_131071_5y.dds"},
	{0x2AACB5D77A3B3053ULL, "380777797.ogg"},
	{0x2AACCF924A0AA977ULL, "88362695.ogg"},
	{0x2AAD1E59B7CD56B5ULL, "349420232.ogg"},
	{0x2AAD2338D1D58D67ULL, "vxlt_0_lowtown_docks_c5k_0_0.vlm"},
	{0x2AAD75BDC0D7583AULL, "79636898.ogg"},
	{0x2AADD61859DCB3D0ULL, "vxw_bonepitmining_a2_2a5_0_0.vlm"},
	{0x2AAEF5FFCB5FBDC4ULL, "6175344_m.fxe"},
	{0x2AAF03A0260FE518ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.dura.ani"},
	{0x2AAF0CC12EE4C535ULL, "6115509_m.fxe"},
	{0x2AAF33BFD97080F0ULL, "genip_rubble.utp"},
	{0x2AAF50B1C3ABE7D4ULL, "vxw_encounter2_camp_l4_0_0.vlm"},
	{0x2AAFCEFA27DD0FCFULL, "6059320_m.fxe"},
	{0x2AB0001B3E859EEAULL, "356763778.ogg"},
	{0x2AB00D26B3E92A61ULL, "hlac_ca_caverns1_24421_1x.dds"},
	{0x2AB01E2D15A771CFULL, "vxlt_warehouse_n_14v_0_0.vlm"},
	{0x2AB089DBB3733D7CULL, "vxlt_gallowstemplar_f_38r_0_0.vlm"},
	{0x2AB164A322C8B2E1ULL, "dae000_hate_enemies6167625.cl"},
	{0x2AB2209C07AE521AULL, "sib131ar_family_estate.lst"},
	{0x2AB25F599315B3CAULL, "fake_im_arm_cht_hvy_champ.uti"},
	{0x2AB25FF8A0E08E96ULL, "zz_waldo1.cut"},
	{0x2AB289ADEF129CC9ULL, "6143410_m.fxe"},
	{0x2AB28ED24A7064C8ULL, "dre000_mother6037128.cl"},
	{0x2AB2B9E22A45C281ULL, "163518032.ogg"},
	{0x2AB32088D29E5532ULL, "vxlt_0_lowtown_a2_n_f6p_0_0.vlm"},
	{0x2AB37BBF6FDB2386ULL, "380957473.ogg"},
	{0x2AB3859CA737E0A0ULL, "182264689.ogg"},
	{0x2AB3AF0DF1D269B4ULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_fin_c.mmh"},
	{0x2AB3EB3369966F67ULL, "vxht_keep_n_5hl_0_0.vlm"},
	{0x2AB4004E32454C03ULL, "vxlt_warehouse_x0_0_0.vlm"},
	{0x2AB473AE3014AF89ULL, "t1_stb_lbr.tnt"},
	{0x2AB487A1DD017F25ULL, "mcr310_argument6072668.cl"},
	{0x2AB49B2F108EE35EULL, "936712352.ogg"},
	{0x2AB4C537DAA6B778ULL, "lt_gallowscourtyard_n_33_0.rml"},
	{0x2AB4FE0A3285EAABULL, "ave200_donnic6110636.cl"},
	{0x2AB518069A32DCA3ULL, "vxht_mansion2_1ip_0_0.vlm"},
	{0x2AB5208D16F03EF6ULL, "675412744.ogg"},
	{0x2AB639D0034A1F42ULL, "vxlt_hangedman_2up_0_0.vlm"},
	{0x2AB69749FBAA82CFULL, "vxht_0_hightown_a2_9ah_0_0.vlm"},
	{0x2AB6A5C9BDF56BB3ULL, "6120598_m.fxe"},
	{0x2AB77B0A84857DA5ULL, "dre000_varric6036971.cl"},
	{0x2AB81ADD10169676ULL, "6087322_m.fxe"},
	{0x2AB84607703A9A01ULL, "vxlt_gallowsprison_n_67g_0_0.vlm"},
	{0x2AB890B85F83C956ULL, "llac_lt_gallowstemplar_2392_0.dds"},
	{0x2AB8946DAD1B8687ULL, "l_lgt110ar_gang_hideout_lt.wbd"},
	{0x2AB8D27A7CA22032ULL, "148428823.ogg"},
	{0x2AB9142B9924E94AULL, "genip_node_l1.utp"},
	{0x2AB95AB96A083D54ULL, "799306301.ogg"},
	{0x2ABB004179FFB5B9ULL, "plt_prm_cod_duru_amulet.plo"},
	{0x2ABB26072456D04FULL, "cth_nb6.tnt"},
	{0x2ABB5BE860AEA820ULL, "93389552.ogg"},
	{0x2ABC08B44486A52EULL, "vxlt_warehouse_n_p4_0_0.vlm"},
	{0x2ABC3D027D283AEFULL, "6053546_m.fxe"},
	{0x2ABC76853F9603C0ULL, "vxw_blightlands_6u_0_0.vlm"},
	{0x2ABCA37B38AC70FAULL, "follower_barks6153560.cl"},
	{0x2ABD8BFD89D6B499ULL, "gen00fl_isabela6096804.cl"},
	{0x2ABDAEF90CAF1165ULL, "vxht_0_hightown_a2_aij_0_0.vlm"},
	{0x2ABDBDB668D053C7ULL, "6148101_m.fxe"},
	{0x2ABDC28EB5A20054ULL, "vxw_encounter2_camp_g5_0_0.vlm"},
	{0x2ABE5CF2AD686931ULL, "vxht_0_hightown_a2_c4w_0_0.vlm"},
	{0x2ABE5F01F8D05640ULL, "dre000_sibling_has_blight6036515.cl"},
	{0x2ABE9F39B79A905FULL, "one110_tree_minder6081610.cl"},
	{0x2ABEB35F291C5351ULL, "dr_0_deeproads_17036.rml"},
	{0x2ABED1BCA0BC795FULL, "w_woundedcoast_212_0.phy"},
	{0x2ABF1564D283B50DULL, "1008116880.ogg"},
	{0x2ABF38A684780CBFULL, "6040708_m.fxe"},
	{0x2ABF6309EDF420B0ULL, "6074968_m.fxe"},
	{0x2ABF9976351B683AULL, "mrl310_merrill6154548.cl"},
	{0x2ABFE03C9A45D80AULL, "ser221_gascard6149160.cl"},
	{0x2AC00BEEA1AE01F1ULL, "vxht_0_hightown_7lo_0_0.vlm"},
	{0x2AC080429AA56FE0ULL, "212949876.ogg"},
	{0x2AC0B32886740852ULL, "6196828_m.fxe"},
	{0x2AC0D18FF89AF93AULL, "210499636.ogg"},
	{0x2AC0F2B9D0F7F845ULL, "vxht_mansion2_2eq_0_0.vlm"},
	{0x2AC19265C4946DB2ULL, "vxlt_0_lowtown_a2_n_f1q_0_0.vlm"},
	{0x2AC1F41FEEACF7AFULL, "vxlt_hangedman_n_1lj_0_0.vlm"},
	{0x2AC204D52E73FB52ULL, "636706163.ogg"},
	{0x2AC21A9D598BA7A8ULL, "vxlt_0_lowtown_gpn_0_0.vlm"},
	{0x2AC25AC228A50186ULL, "6153122_m.fxe"},
	{0x2AC263C4405D9A60ULL, "t3_wpn_com.tnt"},
	{0x2AC31C729D0A5122ULL, "vxca_caverns1_mines_gh9_0_0.vlm"},
	{0x2AC337784F32CFFDULL, "vis110_commoner2.bnk"},
	{0x2AC37CBAE072DF5DULL, "vxht_mansion_n_6h1_0_0.vlm"},
	{0x2AC39F0C87918D63ULL, "lt_0_lowtown_docks_a3_n_216_0.msh"},
	{0x2AC3AD0C8791A4BDULL, "lt_0_lowtown_docks_a3_n_216_0.mmh"},
	{0x2AC3B10C8791ABCEULL, "lt_0_lowtown_docks_a3_n_216_0.mao"},
	{0x2AC3F2EA883EB63BULL, "123875119.ogg"},
	{0x2AC4A13A48029A81ULL, "957608082.ogg"},
	{0x2AC4B651999CF53BULL, "vxht_0_hightown_bc9_0_0.vlm"},
	{0x2AC4C78D58B4A3A1ULL, "6106615_m.fxe"},
	{0x2AC4E4F9D0D276F2ULL, "dae000_female_tranquil.cnv"},
	{0x2AC51FEB7A8FF2F1ULL, "zz_fen_debug6165674.cl"},
	{0x2AC533413E42B955ULL, "vxca_cavern1_haunted_gkg_0_0.vlm"},
	{0x2AC550064A84D508ULL, "follower_banter6171907.cl"},
	{0x2AC574F18EB0F855ULL, "6093544_m.fxe"},
	{0x2AC68A9276467149ULL, "vxw_sundermount_302_0_1.vlm"},
	{0x2AC719784F368505ULL, "vis110_commoner2.cnv"},
	{0x2AC792A87A2D3584ULL, "6024575_f.fxe"},
	{0x2AC79A0EBA099176ULL, "vfx_dot_blur.mao"},
	{0x2AC7A4579B6B9E8BULL, "vxca_caverns2_83t_0_0.vlm"},
	{0x2AC806F9D0D4E64AULL, "dae000_female_tranquil.bnk"},
	{0x2AC8193F74D4EA7CULL, "vxlt_0_lowtown_docks_n_bs8_0_0.vlm"},
	{0x2AC81FD611FB91D6ULL, "mage_inferno.dds"},
	{0x2AC86F3108138FB2ULL, "vxht_mansion2b_22o_0_0.vlm"},
	{0x2AC8938327B83CAAULL, "llac_ht_0_hightown_a3_65537_2.dds"},
	{0x2AC91B0A47B26CEEULL, "812043684.ogg"},
	{0x2AC94AA983720259ULL, "vxdr_primevalentrance_2zl_0_1.vlm"},
	{0x2AC986CE19E01BF0ULL, "vxw_blightlands_76_0_0.vlm"},
	{0x2AC9A571518F06F8ULL, "mag200_feynriel6030436.cl"},
	{0x2AC9B5A558026C98ULL, "6152196_m.fxe"},
	{0x2ACA33E93261B1ADULL, "12246154.ogg"},
	{0x2ACA44337CAF4579ULL, "vxca_cavern1_haunted_cln_0_0.vlm"},
	{0x2ACA77DF640FB179ULL, "vxht_keep_n_708_0_0.vlm"},
	{0x2ACA7A91367CBDAEULL, "vxlt_warehouse_n_w5_0_0.vlm"},
	{0x2ACA893A1C1EF596ULL, "mag360_thrask6043672.cl"},
	{0x2ACAB73A2DE59342ULL, "6147431_m.fxe"},
	{0x2ACAE6EA13F62CCBULL, "6128727_m.fxe"},
	{0x2ACB58977517ECB4ULL, "llac_ht_mansion_n_3368_0.dds"},
	{0x2ACB899F86657AF2ULL, "plt_isa000pt_gift_talisman.nss"},
	{0x2ACBB0545E58828BULL, "6104714_m.fxe"},
	{0x2ACC0D73A4D59363ULL, "art\\characters\\weapons\\staff\\w_stf_105a.mao"},
	{0x2ACCC3E72227B23EULL, "vxlt_undercityencounter_1bm_0_0.vlm"},
	{0x2ACCCECF9F5E81B1ULL, "vxht_0_hightown_a2_c3v_0_0.vlm"},
	{0x2ACCDDB19F9DEA43ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy_crnf.ani"},
	{0x2ACD257D7FF14B43ULL, "nat380_sibling6103572.cl"},
	{0x2ACD3AB987A8E830ULL, "256334016.ogg"},
	{0x2ACD43629AF7A969ULL, "hlac_lt_0_lowtown_a3_n_c_131071_3z.dds"},
	{0x2ACDAD6DE298F430ULL, "w_encounter1_27_0.msh"},
	{0x2ACDBB6DE2990C4AULL, "w_encounter1_27_0.mmh"},
	{0x2ACDBF6DE2991281ULL, "w_encounter1_27_0.mao"},
	{0x2ACDE1869DBBF770ULL, "6093840_m.fxe"},
	{0x2ACF3AEE33B4D82BULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.mmh"},
	{0x2ACF7176D1687CA7ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_2y.dds"},
	{0x2ACF8C49B8F0A59DULL, "ave200_donnic6111215.cl"},
	{0x2ACFD2BCA0CAEE64ULL, "w_woundedcoast_212_0.mmh"},
	{0x2ACFD6BCA0CAF4B7ULL, "w_woundedcoast_212_0.mao"},
	{0x2ACFE8BCA0CB1306ULL, "w_woundedcoast_212_0.msh"},
	{0x2AD08CA01D1D9DF9ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.br_atk_v2.evt"},
	{0x2AD09F9DF2AFE38DULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun_2.mao"},
	{0x2AD1258D3A17D076ULL, "state_image_suffixes.gda"},
	{0x2AD14CD7AA3ECFB4ULL, "mag360_samson6054683.cl"},
	{0x2AD14D534EE4CFECULL, "6182012_m.fxe"},
	{0x2AD16AE099D37F04ULL, "gen_im_trs_dal_arc_sbw_r1.uti"},
	{0x2AD1A01AEB21EC88ULL, "hlac_ca_caverns1_24425_2x.dds"},
	{0x2AD1C469E2F86BEFULL, "vxlt_warehouse_n_de_0_0.vlm"},
	{0x2AD1F1C8CBCCF137ULL, "hlac_lt_0_lowtown_docks_a3_-131075_5y.dds"},
	{0x2AD2043E6B8888B9ULL, "vxlt_0_lowtown_f0j_0_0.vlm"},
	{0x2AD22F0F8E13B815ULL, "vxht_brothel_n_6bt_0_0.vlm"},
	{0x2AD253C4EC0AFE37ULL, "and330st_base_1st_noromance.stg"},
	{0x2AD32FDE5DED89D9ULL, "vxht_0_hightown_a3_c2d_0_0.vlm"},
	{0x2AD34F29F89F13B7ULL, "hlac_lt_undercityencounter_-1_2y.dds"},
	{0x2AD369895782191DULL, "6088401_m.fxe"},
	{0x2AD3CCA69CA58CC9ULL, "lt_house_3567_ll.mmh"},
	{0x2AD3F2A5D553CBDBULL, "127135567.ogg"},
	{0x2AD475372A474501ULL, "257722763.ogg"},
	{0x2AD4FE7A91453E2DULL, "vxca_caverns1_mines_jfy_0_0.vlm"},
	{0x2AD51DB86802AC11ULL, "ss_meeran.bnk"},
	{0x2AD57644B6CBC6B4ULL, "721491204.ogg"},
	{0x2AD5852F0C98FA32ULL, "6111398_f.fxe"},
	{0x2AD5FCB24965E653ULL, "498306717.ogg"},
	{0x2AD71F9C07CD6C9DULL, "sib131ar_family_estate.are"},
	{0x2AD764F3B3BF7089ULL, "vxlt_0_lowtown_a2_e1z_0_0.vlm"},
	{0x2AD7D7AB010C48C9ULL, "vxdr_primevaltomb_7t2_0_0.vlm"},
	{0x2AD7F8AB0BD63470ULL, "vxw_blightlands_87_0_0.vlm"},
	{0x2AD82E3E37E6AAF9ULL, "gen00fl_anders6119699.cl"},
	{0x2AD88408019F2496ULL, "vxht_bartrand_n_67l_0_0.vlm"},
	{0x2AD88E4AD0E30B13ULL, "319996161.ogg"},
	{0x2AD8A137B692F698ULL, "6026989_m.fxe"},
	{0x2AD8EC6E2872D62EULL, "vxlt_warehouse_n_z6_0_0.vlm"},
	{0x2AD9CC712FE33440ULL, "6201500_m.fxe"},
	{0x2ADA39732EA6B140ULL, "6135230_m.fxe"},
	{0x2ADAEAFC5ED4ABFCULL, "1007376994.ogg"},
	{0x2ADAF5B19FAA503AULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy_crnf.evt"},
	{0x2ADB3658B258F7A0ULL, "qcr221_diplomatic_sol_st.stg"},
	{0x2ADB4C8F3D0FA931ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopea.mmh"},
	{0x2ADB588F3D0FBDD2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopea.mao"},
	{0x2ADB668F3D0FD5EFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopea.msh"},
	{0x2ADBA48B58B6FD77ULL, "174993417.ogg"},
	{0x2ADC5DBFEF9E2932ULL, "6151920_m.fxe"},
	{0x2ADC8A79AFA755EAULL, "vxlt_0_lowtown_a2_ejh_0_0.vlm"},
	{0x2ADCB20DF1F5746BULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_fin_c.anb"},
	{0x2ADCCF69FC2F2B13ULL, "one300_zevran6089714.cl"},
	{0x2ADD1BEE33C0E090ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.anb"},
	{0x2ADD30D9505FD3F5ULL, "dae300_cullen6135885.cl"},
	{0x2ADD948ED9501DFAULL, "custic00c_a.dds"},
	{0x2ADDAEC56DA09DE2ULL, "vxca_caverns2_8oe_0_0.vlm"},
	{0x2ADDB4DCF21025E1ULL, "59706455.ogg"},
	{0x2ADDD95C0380B864ULL, "gen00fl_anders6104023.cl"},
	{0x2ADDDC6C99EA8350ULL, "551223028.ogg"},
	{0x2ADDEC53AEB50851ULL, "vxlt_hangedman_n_27r_0_0.vlm"},
	{0x2ADE6FFC970A3CBFULL, "ser270_templar6114154.cl"},
	{0x2ADE9F191CA04323ULL, "plt_lgt000pt_civ_breech.plo"},
	{0x2ADEE4F893F2256CULL, "6080570_m.fxe"},
	{0x2ADEE8D73C0F5CD9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_amllcrstold_d.dds"},
	{0x2ADEEB4E45D538EAULL, "vxht_keep_n_6hf_0_0.vlm"},
	{0x2ADF3B4E2C94D905ULL, "170990032.ogg"},
	{0x2ADF58A1748C7A3EULL, "6120992_f.fxe"},
	{0x2ADF95E0348B4C1DULL, "vxlt_0_lowtown_a2_n_c_gwk_0_0.vlm"},
	{0x2ADFB8D18B95F82BULL, "vxlt_hangedman_n_2ct_0_0.vlm"},
	{0x2ADFDFA5EADE4707ULL, "lt_0_lowtown_docks_39_0.anb"},
	{0x2AE017155F66D9E3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryright.mao"},
	{0x2AE01B155F66E030ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryright.mmh"},
	{0x2AE029155F66F84AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryright.msh"},
	{0x2AE0761B5D7EA139ULL, "vxlt_0_lowtown_f5i_0_0.vlm"},
	{0x2AE0F4B19FAEBA2DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy_crnf.gad"},
	{0x2AE12065807DB7B9ULL, "vxdr_primevaltomb_6qm_0_1.vlm"},
	{0x2AE1A1BB4FE3A259ULL, "vxht_0_hightown_a3_c7k_0_0.vlm"},
	{0x2AE1ADF141553DC4ULL, "869413601.ogg"},
	{0x2AE20C5232318F69ULL, "6177589_m.fxe"},
	{0x2AE227E4E247EECAULL, "vxht_0_hightown_61z_0_0.vlm"},
	{0x2AE23EC4EE2A0C9DULL, "6144082_m.fxe"},
	{0x2AE2D8ADF539C857ULL, "character_character_p211.psh.fxo.dx11.assembly"},
	{0x2AE2E4F84E09A6E8ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunaritent_sa.dds"},
	{0x2AE30A8C5C201B04ULL, "hlac_w_bonepit_327684_0x.dds"},
	{0x2AE32A8EFB8FE3BAULL, "llac_w_sundermount_-6_0.dds"},
	{0x2AE3389E98EDFDE6ULL, "hlac_ht_0_hightown_n_-196608_2y.dds"},
	{0x2AE352CCD6FD5EFCULL, "hlac_ht_0_hightown_a2_65533_1z.dds"},
	{0x2AE368F7E727110EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_exit.ani"},
	{0x2AE376C73E98EEABULL, "822353864.ogg"},
	{0x2AE399388868A07DULL, "vxw_encounter2_camp_qk_0_1.vlm"},
	{0x2AE3B9547F066D15ULL, "vxw_blightlands_9y_0_0.vlm"},
	{0x2AE452AFC8BC2E9CULL, "6144192_f.fxe"},
	{0x2AE463D04B310129ULL, "vis121_saemus6086695.cl"},
	{0x2AE5131E607E4661ULL, "vxht_0_hightown_a3_bgj_0_0.vlm"},
	{0x2AE5323AD6E2638EULL, "vxht_0_hightown_n_ajw_0_0.vlm"},
	{0x2AE56C4A421C245EULL, "vxdr_0_deeproads_cqp_0_0.vlm"},
	{0x2AE5C9CEF9DF4F36ULL, "art\\levels\\ztools\\default\\generic\\3wi_basic.msh"},
	{0x2AE5D3CEF9DF6034ULL, "art\\levels\\ztools\\default\\generic\\3wi_basic.mmh"},
	{0x2AE5D7CEF9DF6747ULL, "art\\levels\\ztools\\default\\generic\\3wi_basic.mao"},
	{0x2AE5EDAF44A4B38AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_boo_cas_0s.dds"},
	{0x2AE62DE23CC23A4EULL, "vxht_0_hightown_c3x_0_1.vlm"},
	{0x2AE6A44D8D1BF90DULL, "ht_0_hightownexplosion_n_1_0.phy"},
	{0x2AE734B5E9BEDCC7ULL, "dae000_mother6136939.cl"},
	{0x2AE7B22E528C4372ULL, "63387034.ogg"},
	{0x2AE7C53414F83CC5ULL, "336055124.ogg"},
	{0x2AE7D1EF321E629CULL, "97570859.ogg"},
	{0x2AE85719FC1D816AULL, "vxlt_warehouse_9r_0_0.vlm"},
	{0x2AE8B249CA8CFE97ULL, "396112724.ogg"},
	{0x2AE9DBD98CAE4471ULL, "vxlt_0_lowtown_fhz_0_0.vlm"},
	{0x2AEA2A3E7BE9C59CULL, "643431373.ogg"},
	{0x2AEA3C3E65C9CFE8ULL, "988966374.ogg"},
	{0x2AEA48C1096674A6ULL, "6092302_m.fxe"},
	{0x2AEA4A7EA8980C43ULL, "vxlt_0_lowtown_docks_n_c6u_0_0.vlm"},
	{0x2AEAA0594FF10FF7ULL, "llac_lt_0_lowtown_n_0_3.dds"},
	{0x2AEAE1A54C20FE06ULL, "185143633.ogg"},
	{0x2AEB5826F99DC409ULL, "6101038_m.fxe"},
	{0x2AEBDDD2697234EEULL, "312785451.ogg"},
	{0x2AEC4975EE1C6838ULL, "mer320_hubert6090234.cl"},
	{0x2AEC5E30A0AB20D1ULL, "vxlt_hangedman_n_22q_0_0.vlm"},
	{0x2AEC6FFE18AD1FF1ULL, "vxlt_gallowstemplar_n_3r9_0_0.vlm"},
	{0x2AEC9D34E260628DULL, "6140608_m.fxe"},
	{0x2AEDA67C60B555AFULL, "and100st_anders_intro.stg"},
	{0x2AEDBAB6DCF3B0C1ULL, "gen00fl_merrill6145164.cl"},
	{0x2AEDC3CC325CA989ULL, "458711539.ogg"},
	{0x2AEE07BD2681649DULL, "vxlt_0_lowtown_a2_n_c_gpj_0_0.vlm"},
	{0x2AEE9D8F214DF6A8ULL, "hlac_lt_gallowscourtyard_n_-131071_0y.dds"},
	{0x2AEEEB8D0DAB533BULL, "vxht_mansion_n_4yf_0_0.vlm"},
	{0x2AEF12C971FFE915ULL, "vxht_brothel_n_6hr_0_0.vlm"},
	{0x2AEF14849A026873ULL, "qcr000_varric6208216.cl"},
	{0x2AEFAB870DC55FDCULL, "dae000_bodhan6146857.cl"},
	{0x2AEFAC5C380EA1E9ULL, "6122464_m.fxe"},
	{0x2AEFD32B257722C2ULL, "vxht_mansion_n_596_0_0.vlm"},
	{0x2AF01A14360DD0CFULL, "hlac_dr_0_deeproads_19793_0z.dds"},
	{0x2AF026382398F019ULL, "hlac_lt_0_lowtown_-2_1x.dds"},
	{0x2AF060F7E7318E5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_exit.evt"},
	{0x2AF092A11960DA7CULL, "vxw_sundermount_37b_0_1.vlm"},
	{0x2AF0A11920B5BB75ULL, "dae000ip_crowd_ff3_talk18.utp"},
	{0x2AF16728B9275FF4ULL, "6084512_m.fxe"},
	{0x2AF2593B63B88928ULL, "vxht_chantry_n_30u_0_0.vlm"},
	{0x2AF2BB0D3C75AD0CULL, "mcr310_sibling6095294.cl"},
	{0x2AF377040FBC73D3ULL, "conversation_markup.gda"},
	{0x2AF462390CCC20ECULL, "hlac_lt_0_lowtown_a2_n_-262144_4z.dds"},
	{0x2AF48C82F6040B09ULL, "hm_one_crow_1.mor"},
	{0x2AF4AE8B1CF59447ULL, "plt_cod_ltr_forbethany.plo"},
	{0x2AF4C951E48558AEULL, "aip_mature_dragon_base.gda"},
	{0x2AF525CE3CB5A97EULL, "mag120_macha6080632.cl"},
	{0x2AF53C2CE785765BULL, "6178348_m.fxe"},
	{0x2AF5E4C91ED5671CULL, "vxlt_hangedman_n_5e_0_0.vlm"},
	{0x2AF6DA8DB69B35CDULL, "character_character_p19.psh.fxo"},
	{0x2AF6ED289AF28F09ULL, "124398806.ogg"},
	{0x2AF712347605F0DFULL, "art\\vfx\\models\\vfx_lightning_ball.mmh"},
	{0x2AF718347605FA81ULL, "art\\vfx\\models\\vfx_lightning_ball.msh"},
	{0x2AF739A6209A38D4ULL, "gen00fl_fenris6127066.cl"},
	{0x2AF741F7E7377848ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_exit.gad"},
	{0x2AF79084286E211EULL, "840150945.ogg"},
	{0x2AF7A4FD1596E0F4ULL, "dae000_epic_merchant_46179528.cl"},
	{0x2AF7C292EA1C3E09ULL, "vxht_keep_6ai_0_0.vlm"},
	{0x2AF7DB417500F6B2ULL, "qaesnf2.arl"},
	{0x2AF85BFC941391F1ULL, "217415719.ogg"},
	{0x2AF885DBCFD1C90DULL, "377662640.ogg"},
	{0x2AF8BA6A4EFC863BULL, "vxlt_gallowstemplar_5k1_0_0.vlm"},
	{0x2AF8BA9DFB5C8D26ULL, "6097002_m.fxe"},
	{0x2AF99E6DAD532EBDULL, "l_lgt110ar_gang_hideout_lt.bnk"},
	{0x2AF9DAC18702B100ULL, "hlac_lt_0_lowtown_a2_n_c_-262143_0y.dds"},
	{0x2AFA6AC9848B579BULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.impa.ani"},
	{0x2AFAB2047B6693BBULL, "mag291_vincento6166239.cl"},
	{0x2AFACFE089E068E6ULL, "vxlt_gallowstemplar_n_3jz_0_0.vlm"},
	{0x2AFAE01C139F0C7EULL, "6179908_m.fxe"},
	{0x2AFB1017D9087F93ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_medj_0e.dds"},
	{0x2AFBEF79C5764535ULL, "744695811.ogg"},
	{0x2AFCC234B85A2412ULL, "gen00fl_dog6146967.cl"},
	{0x2AFD0A7503E639B2ULL, "gen00fl_aveline6116739.cl"},
	{0x2AFD59D8668A2D49ULL, "vxca_cave_23o_0_0.vlm"},
	{0x2AFD83B226AB89A9ULL, "vxw_woundedcoast_1hp_0_1.vlm"},
	{0x2AFD84A663F60195ULL, "vxht_brothel_n_6uq_0_0.vlm"},
	{0x2AFE3D2B13E5468CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearready.ani"},
	{0x2AFE94745FE05854ULL, "vxlt_0_lowtown_a3_n_c_gbl_0_0.vlm"},
	{0x2AFEBCA177CB7E82ULL, "gen_im_trs_bst_drg_eye_01.uti"},
	{0x2AFEECBAB8B55D22ULL, "575356052.ogg"},
	{0x2AFF047E0B56F2FCULL, "vxw_sundermount_32a_0_1.vlm"},
	{0x2AFF64DAB853ED83ULL, "vxlt_blackemporium_141_0_0.vlm"},
	{0x2B000BCC8A02FF3EULL, "6136021_f.fxe"},
	{0x2B004D9B0B3BE6D1ULL, "darktown_male_sickly_sta6175320.cl"},
	{0x2B00D7E12002BDF9ULL, "125242014.ogg"},
	{0x2B012C9F19D5AD4CULL, "vxlt_0_lowtown_n_gcd_0_0.vlm"},
	{0x2B01BF4E75345E15ULL, "vxw_sundermount_26d_0_0.vlm"},
	{0x2B01ECDAC2DEF200ULL, "182009013.ogg"},
	{0x2B02BF0E05381360ULL, "vxlt_darktown_586_0_1.vlm"},
	{0x2B038D8928E68D9AULL, "hm_lgt_soph_hunter.mor"},
	{0x2B03959D91CEF661ULL, "vxht_keep_58e_0_0.vlm"},
	{0x2B03EA7DD7CD34EFULL, "art\\levels\\ztools\\default\\generic\\9ff_barrel.phy"},
	{0x2B04947FB5D45441ULL, "688033799.ogg"},
	{0x2B054A2B13EB7BCEULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearready.gad"},
	{0x2B057549263352FAULL, "mag170_thrask6054017.cl"},
	{0x2B06346FDC125689ULL, "vxht_keep_6jh_0_0.vlm"},
	{0x2B06482BD3CC1A04ULL, "gen00fl_aveline6122099.cl"},
	{0x2B066057105473E9ULL, "vxlt_hangedman_2fr_0_0.vlm"},
	{0x2B072C4740F29EBBULL, "vxlt_gallowstemplar_5t0_0_0.vlm"},
	{0x2B0772C98495EF92ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.impa.evt"},
	{0x2B07D396F1F9DF82ULL, "qun221_viscount6070482.cl"},
	{0x2B07F559B3C0239BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_exit.evt"},
	{0x2B08A2A4A6FB5A7BULL, "vxlt_0_lowtown_docks_a3_n_d3b_0_0.vlm"},
	{0x2B09399C62FEC322ULL, "120901768.ogg"},
	{0x2B093AA7A7770E00ULL, "212846614.ogg"},
	{0x2B09AA8E10643EE5ULL, "635753613.ogg"},
	{0x2B09AAC86A8AC0A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hm_paperdoll.anb"},
	{0x2B09B3B8D86A65C1ULL, "vxht_mansion2_1u6_0_0.vlm"},
	{0x2B0ABCC4B6CEBEE7ULL, "488724589.ogg"},
	{0x2B0ADEFA11A64AB5ULL, "dae221ar_chantry.are"},
	{0x2B0B4703B522393CULL, "189346843.ogg"},
	{0x2B0B75DD18F49589ULL, "vxht_0_hightown_n_byv_0_0.vlm"},
	{0x2B0B8BDDAAD0CDEFULL, "vxlt_warehouse_n_xa_0_0.vlm"},
	{0x2B0BF58F18A1A229ULL, "vxw_woundedcoast_1sq_0_1.vlm"},
	{0x2B0BFF738468974BULL, "vxlt_0_lowtown_docks_n_bdn_0_0.vlm"},
	{0x2B0C18675D070B0CULL, "6149512_m.fxe"},
	{0x2B0C452B13F191ADULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearready.evt"},
	{0x2B0C5B3986B0AC5EULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.dura.evt"},
	{0x2B0C6EDDA535CF56ULL, "405867413.ogg"},
	{0x2B0C907E2808B078ULL, "471560874.ogg"},
	{0x2B0D20EDF55FA05EULL, "vxht_chantry_2cp_0_0.vlm"},
	{0x2B0D248173CFEDC9ULL, "ser100_party_comments6120603.cl"},
	{0x2B0D3EAE6285EC04ULL, "vxw_encounter2_dt_0_0.vlm"},
	{0x2B0D45E1291930FDULL, "123945782.ogg"},
	{0x2B0D82492C2F9009ULL, "vxht_0_hightown_n_6pm_0_0.vlm"},
	{0x2B0DA032577B97E4ULL, "6095367_m.fxe"},
	{0x2B0DB034CFBC7447ULL, "llac_lt_0_lowtown_a3_n_c_-196609_4.dds"},
	{0x2B0DF37EDA140304ULL, "6076791_m.fxe"},
	{0x2B0E531F0FDFE850ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_magb_0t.dds"},
	{0x2B0E91C9849C4325ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.impa.gad"},
	{0x2B0E94972EC536DBULL, "p_cyc_out.mao"},
	{0x2B0EEDE900CC2F0BULL, "ts_dalish_wpn_props.gda"},
	{0x2B0EEE59B3C636DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_exit.gad"},
	{0x2B0F4EAB761645FDULL, "hlac_lt_0_lowtown_a2_-196609_4z.dds"},
	{0x2B0FCA103309A577ULL, "678511698.ogg"},
	{0x2B0FD04B1B46B088ULL, "gen00fl_anders6119137.cl"},
	{0x2B1014C535F67DA8ULL, "vxca_cavern1_haunted_j5a_0_0.vlm"},
	{0x2B1081441E2D7EABULL, "984628510.ogg"},
	{0x2B1086333D0A172CULL, "735607620.ogg"},
	{0x2B10AC89640810AAULL, "fx_ab_01.dds"},
	{0x2B10DE2B736FDB05ULL, "genip_bookshelf.utp"},
	{0x2B113099DAE7E5B3ULL, "74778756.ogg"},
	{0x2B11C3D0523B82EEULL, "227328235.ogg"},
	{0x2B123393911555F8ULL, "519987328.ogg"},
	{0x2B1246F552A665A3ULL, "48907182.ogg"},
	{0x2B129B70EB279B42ULL, "6043829_m.fxe"},
	{0x2B12A81842B5D699ULL, "6111289_f.fxe"},
	{0x2B12BBDD0F13843FULL, "art\\vfx\\combat\\vfx_r_arc_burst_shot_act_c.mmh"},
	{0x2B12CE33CA1B9601ULL, "dae000_meredith6187901.cl"},
	{0x2B135D4D4DA5B05DULL, "1005992001.ogg"},
	{0x2B13723986B6F189ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.dura.gad"},
	{0x2B138FB7476BBBBEULL, "vxlt_0_lowtown_docks_a3_n_apw_0_0.vlm"},
	{0x2B144B0023224BD9ULL, "dae021_keep_noble2.utc"},
	{0x2B14CB0B0FA47520ULL, "one280_emrys6137196.cl"},
	{0x2B14D3A630C71546ULL, "lgt110_gxa_coa_meeting6117494.cl"},
	{0x2B14ED59B3CAA0DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_exit.ani"},
	{0x2B14FFFA11AED71BULL, "dae221ar_chantry.ncs"},
	{0x2B15028031D868E1ULL, "pro000_colonel_ewald6047385.cl"},
	{0x2B1512CA1C9F1E9AULL, "141031495.ogg"},
	{0x2B1516BD7DD40452ULL, "hlac_lt_gallowscourtyard_131070_0z.dds"},
	{0x2B1519CF9D9C3CF7ULL, "zz_fen_debug6215241.cl"},
	{0x2B157521EFB5D2CFULL, "vxlt_backalley_n_1n1_0_0.vlm"},
	{0x2B1646C9372C83B0ULL, "lt_0_lowtown_docks_a3_n_148_0.mao"},
	{0x2B164AC9372C8AC3ULL, "lt_0_lowtown_docks_a3_n_148_0.mmh"},
	{0x2B1658C9372CA21DULL, "lt_0_lowtown_docks_a3_n_148_0.msh"},
	{0x2B16B170FF6E728DULL, "hlac_lt_0_lowtown_docks_a3_-131075_4x.dds"},
	{0x2B17031E5953D8D5ULL, "6185385_f.fxe"},
	{0x2B1772DB00EF2C3AULL, "gen00fl_sebastian6189401.cl"},
	{0x2B180ED22C40950DULL, "hlac_lt_undercityencounter_-1_3x.dds"},
	{0x2B182595CA607E41ULL, "vxht_mansion2_1x5_0_0.vlm"},
	{0x2B189D48C821C9A2ULL, "gen00fl_isabela6078988.cl"},
	{0x2B18E0EEF906BFB6ULL, "follower_banter6163771.cl"},
	{0x2B19E7BA0AEAAE09ULL, "vxht_0_hightown_n_bdu_0_0.vlm"},
	{0x2B1A433986BCC0D7ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.dura.ani"},
	{0x2B1AA97586C94356ULL, "6143022_m.fxe"},
	{0x2B1ACB67511E6728ULL, "gen_im_cft_rec_arn_02.uti"},
	{0x2B1ACE4795698BE7ULL, "mcr371cr_anders.utc"},
	{0x2B1B3FBDC8CF455AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_diagonal_1.mao"},
	{0x2B1B4C22206FCA3CULL, "one300_nuncio6083436.cl"},
	{0x2B1B77ED2471D38AULL, "follower_banter6163182.cl"},
	{0x2B1B94A644A5449BULL, "vxlt_0_lowtown_daq_0_0.vlm"},
	{0x2B1BB08B547C0484ULL, "vxw_encounter2_ou_0_0.vlm"},
	{0x2B1BDE95837C3BC2ULL, "hlac_ht_mansion2_818_18y.dds"},
	{0x2B1BDFFA11B4BFC2ULL, "dae221ar_chantry.lst"},
	{0x2B1C1BA3495F501BULL, "vxw_woundedcoast_hj_0_0.vlm"},
	{0x2B1CBDD3AEB5ACF9ULL, "lfx_noise2.dds"},
	{0x2B1CF58D3C579501ULL, "6108886_m.fxe"},
	{0x2B1D109DF6EF4D04ULL, "lgt260_eau_tmp_angelique6143957.cl"},
	{0x2B1D12348B4D3B0AULL, "hlac_lt_0_lowtown_a2_n_-2_3x.dds"},
	{0x2B1D2FE6BA741933ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_low.phy"},
	{0x2B1D551A023BA38BULL, "vxlt_undercityserialkiller_yf_0_0.vlm"},
	{0x2B1D9FC378763D89ULL, "hlac_lt_0_lowtown_-1_0z.dds"},
	{0x2B1DE5440A0BBF39ULL, "6043567_m.fxe"},
	{0x2B1E19DE145542FFULL, "vxlt_0_lowtown_a2_n_c_e11_0_0.vlm"},
	{0x2B1E2476D705D9D3ULL, "6106948_f.fxe"},
	{0x2B1E48049A814366ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl90runf.ani"},
	{0x2B1E9445E485AB2BULL, "vxlt_darktown_5gi_0_0.vlm"},
	{0x2B1F094607F5E472ULL, "vxlt_0_lowtown_a2_epc_0_1.vlm"},
	{0x2B1F0D6C1E253672ULL, "ht_0_hightown_n_58_0.rml"},
	{0x2B1F31556FD87405ULL, "llac_ht_0_hightown_a3_-196608_0.dds"},
	{0x2B1F70C4D4D9D5F9ULL, "vxlt_0_lowtown_a2_n_c_fit_0_0.vlm"},
	{0x2B1FD997D2C842F9ULL, "vxlt_0_lowtown_a2_n_c_ep5_0_1.vlm"},
	{0x2B1FFE6163AD67F6ULL, "vxw_bonepit_n4_0_0.vlm"},
	{0x2B201631DB72E688ULL, "mcr370_negotiation6080200.cl"},
	{0x2B201B65689A7AE2ULL, "vxlt_gallowstemplar_5mj_0_0.vlm"},
	{0x2B2022155F9CD04BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryright.phy"},
	{0x2B2033A4643307A7ULL, "w_woundedcoast_32_0.rml"},
	{0x2B2046921DA31D73ULL, "6126448_m.fxe"},
	{0x2B2085B02750979FULL, "water_pan.dds"},
	{0x2B20C21098FB8D77ULL, "vxht_mansion_n_4yr_0_0.vlm"},
	{0x2B21277E613A330CULL, "zz_followers_banter6158157.cl"},
	{0x2B213EB66BC60BABULL, "vxlt_0_lowtown_a2_n_eop_0_1.vlm"},
	{0x2B220DA19429D0B6ULL, "mag311_emile_de_launcet6043698.cl"},
	{0x2B226D80692D1719ULL, "6015069_m.fxe"},
	{0x2B22DBBB0360A93EULL, "6077916_m.fxe"},
	{0x2B22E824D47ECAC5ULL, "440521461.ogg"},
	{0x2B234B5EA80010E9ULL, "6067693_m.fxe"},
	{0x2B2352B31D90DAEBULL, "z_cloud_w_smok.dds"},
	{0x2B23C7B0273B1549ULL, "447699763.ogg"},
	{0x2B23E6FEE1ABEB4FULL, "vxlt_backalley_n_1i0_0_0.vlm"},
	{0x2B2403B03A1F4CF5ULL, "vxw_bonepit_1rx_0_0.vlm"},
	{0x2B240B9D6589B6B5ULL, "mcr310_party_banter6087792.cl"},
	{0x2B24100124DECFBBULL, "vxlt_gallowstemplar_5z2_0_0.vlm"},
	{0x2B2444A0A8BCB712ULL, "177884971.ogg"},
	{0x2B24C6BD39634BB8ULL, "384184369.ogg"},
	{0x2B2500DE9390C108ULL, "197996566.ogg"},
	{0x2B2541B1FD8CCC6DULL, "vxht_mansion_n_4z9_0_0.vlm"},
	{0x2B25E9E5D2D321A9ULL, "w_woundedcoast_243_0.mao"},
	{0x2B25F5E5D2D33642ULL, "w_woundedcoast_243_0.mmh"},
	{0x2B25F7E5D2D339F8ULL, "w_woundedcoast_243_0.msh"},
	{0x2B2672A56897D137ULL, "mer220_hubert6081409.cl"},
	{0x2B268B5F6221AAFAULL, "686505547.ogg"},
	{0x2B269C3E058DB514ULL, "72417175.ogg"},
	{0x2B26E9B154FDDF00ULL, "6161198_m.fxe"},
	{0x2B2714494AAA9D8DULL, "6107542_m.fxe"},
	{0x2B273B5E5165D94FULL, "dae000_quintus6156656.cl"},
	{0x2B274BC9373AFFE8ULL, "lt_0_lowtown_docks_a3_n_148_0.phy"},
	{0x2B277E02CADC5BFFULL, "851276161.ogg"},
	{0x2B279F618E5ED187ULL, "335362423.ogg"},
	{0x2B27A40BF1D2D37CULL, "primitives_primitive_p0.vsh.fxo.dx11.assembly"},
	{0x2B27B108F642F58CULL, "6015842_m.fxe"},
	{0x2B281B5DC6207B3EULL, "89392573.ogg"},
	{0x2B2843554C9292F9ULL, "594265196.ogg"},
	{0x2B285996FCE0C689ULL, "vxht_0_hightown_n_bot_0_0.vlm"},
	{0x2B28785496501D8EULL, "hlac_ht_playershouse_n_855_14z.dds"},
	{0x2B288675CAF9A42DULL, "hlac_ht_keep_n_3731_2y.dds"},
	{0x2B28B86B0F582588ULL, "gen00fl_sebastian6179879.cl"},
	{0x2B28D33167FDA9E3ULL, "6127228_f.fxe"},
	{0x2B28D95729619CF5ULL, "lt_0_lowtown_docks_n_176_0.msh"},
	{0x2B28DB572961A06BULL, "lt_0_lowtown_docks_n_176_0.mmh"},
	{0x2B28E1EF74752E6DULL, "vxlt_gallowsprison_n_6k7_0_0.vlm"},
	{0x2B28E227CFCDCA5BULL, "dae000ip_crowd_mf5_merchant.utp"},
	{0x2B28E7572961B408ULL, "lt_0_lowtown_docks_n_176_0.mao"},
	{0x2B293FAA9584AE63ULL, "llac_w_sundermount_327680_0.dds"},
	{0x2B2969748CF6D230ULL, "zz_personality_debug6004627.cl"},
	{0x2B299848EAF44E84ULL, "vxht_0_hightown_a2_ap0_0_0.vlm"},
	{0x2B2A04A7D94BD15EULL, "vxht_chantry_2yr_0_0.vlm"},
	{0x2B2A1165C5215601ULL, "2p_200_laying_injured.stg"},
	{0x2B2A226846721D04ULL, "vxw_encounter2_jz_0_0.vlm"},
	{0x2B2ADC3A574CF87CULL, "hlac_ca_caverns1_24425_5y.dds"},
	{0x2B2B1C1F70D9DAF2ULL, "322411982.ogg"},
	{0x2B2B365E55406AFFULL, "vxlt_0_lowtown_h8j_0_0.vlm"},
	{0x2B2C8221AC54DA74ULL, "63585357.ogg"},
	{0x2B2CB4A53C6F3160ULL, "hlac_ca_caverns1_24422_3z.dds"},
	{0x2B2D9ABC430ED4C4ULL, "dae000_bartrand6027459.cl"},
	{0x2B2DAD7750B0B94CULL, "6017761_m.fxe"},
	{0x2B2E10E9F204DA2DULL, "355562018.ogg"},
	{0x2B2E2BA43CFF9CF1ULL, "hlac_dr_0_deeproads_5849_1z.dds"},
	{0x2B2E50EF08DA66B3ULL, "lowtown_male_alienage2_mob6166688.cl"},
	{0x2B2E703E55A38076ULL, "vxw_bonepit_i5_0_0.vlm"},
	{0x2B2E7757ECB27583ULL, "vxht_mansion_n_61x_0_0.vlm"},
	{0x2B2EE5CBE6FF0A1FULL, "519293222.ogg"},
	{0x2B2F0F0C01C76621ULL, "6149690_m.fxe"},
	{0x2B2FBCD05ED1DD05ULL, "286412284.ogg"},
	{0x2B30BDDF56DC8BFEULL, "zz_and_debug6134112.cl"},
	{0x2B30C252FC99642CULL, "ss_t_four_v4.cut"},
	{0x2B30CA0366F8AF3EULL, "349821522.ogg"},
	{0x2B314D97F556C1BEULL, "6078816_m.fxe"},
	{0x2B3150DBE16F7155ULL, "llac_dr_primevaltomb_657_3.dds"},
	{0x2B31FE2E6CF43399ULL, "253519143.ogg"},
	{0x2B3287A96B7C6599ULL, "dae000_hate_enemies6167618.cl"},
	{0x2B32FF77E084C693ULL, "200812020.ogg"},
	{0x2B3331049A937860ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl90runf.gad"},
	{0x2B333D9C34639865ULL, "vxw_sundermount_264_0_0.vlm"},
	{0x2B3352BC031411EBULL, "hlac_ht_mansion2b_818_16x.dds"},
	{0x2B33E9BA42C8A3E7ULL, "llac_lt_0_lowtown_docks_a3_-3_6.dds"},
	{0x2B33F20193A05FB6ULL, "hlac_lt_gallowstemplar_n_2383_9x.dds"},
	{0x2B341DEA5050BB9AULL, "vxdr_primevaltomb_4hj_0_0.vlm"},
	{0x2B342F6A35F5B1A6ULL, "948434941.ogg"},
	{0x2B34440A29B6DAC4ULL, "pro000cr_guard_f.utc"},
	{0x2B34624657B50CF9ULL, "zz_followers_banter6159315.cl"},
	{0x2B34BFD83D4E3185ULL, "lt_0_lowtown_docks_a3_24_0_lc.mmh"},
	{0x2B35665B4684AFE0ULL, "zz_t_trailer_cut_femhawkdra.cut"},
	{0x2B359A9E8B4B7DC6ULL, "532968953.ogg"},
	{0x2B35C3469D054363ULL, "266532029.ogg"},
	{0x2B360C8ECFA4F844ULL, "6114451_m.fxe"},
	{0x2B368BEB897915B3ULL, "vxlt_0_lowtown_docks_a3_cv1_0_0.vlm"},
	{0x2B36CC94197D18D9ULL, "dae000_maren6148428.cl"},
	{0x2B36F35B5992CD34ULL, "qun111_ketojan6059231.cl"},
	{0x2B370E176F7AC36FULL, "lt_0_lowtown_docks_16_0_lc.mmh"},
	{0x2B377ECBF5F842A1ULL, "fen200cr_tevinter_ambusher.utc"},
	{0x2B378B71F81CB7D4ULL, "follower_banter6173174.cl"},
	{0x2B3805E1C8189472ULL, "hlac_lt_0_lowtown_a2_-1_3z.dds"},
	{0x2B3833A4644799C4ULL, "w_woundedcoast_32_0.mmh"},
	{0x2B3837A46447A017ULL, "w_woundedcoast_32_0.mao"},
	{0x2B3849A46447BF66ULL, "w_woundedcoast_32_0.msh"},
	{0x2B3888348FA86ACDULL, "fo_01.mao"},
	{0x2B3890CDB13AB071ULL, "static_rendertodepth_p0.vsh.fxo"},
	{0x2B38F21003112B37ULL, "6072297_m.fxe"},
	{0x2B38F95F60EF380AULL, "vxht_0_hightown_a2_bxm_0_0.vlm"},
	{0x2B38FF5D2D4B811BULL, "vxw_woundedcoast_bl_0_0.vlm"},
	{0x2B3904F1483CE5DFULL, "6142477_m.fxe"},
	{0x2B3A135A856B4195ULL, "vxw_woundedcoast_12f_0_1.vlm"},
	{0x2B3A3BB10D4B23D0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\prp_secretwheeldoor_s.dds"},
	{0x2B3AB7CB2E68B2BDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masc_0n.dds"},
	{0x2B3B3B263FC0A684ULL, "hlac_lt_0_lowtown_docks_0_2z.dds"},
	{0x2B3B760290C6E13BULL, "mer120_hubert6070239.cl"},
	{0x2B3B77C1BF246AF7ULL, "6094815_m.fxe"},
	{0x2B3BA94C341826B7ULL, "6025990_m.fxe"},
	{0x2B3BBCDD0F368B7CULL, "art\\vfx\\combat\\vfx_r_arc_burst_shot_act_c.anb"},
	{0x2B3BDC51D97CE292ULL, "6125708_f.fxe"},
	{0x2B3C04AA642BC684ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.impa.ani"},
	{0x2B3C2DC166F25B8CULL, "333646504.ogg"},
	{0x2B3CD8745C2249CFULL, "647589006.ogg"},
	{0x2B3CEB6E8E6BB7F1ULL, "gen00fl_fenris6126462.cl"},
	{0x2B3D0F745BA27F53ULL, "vxdr_primevaltomb_7kq_0_0.vlm"},
	{0x2B3D24879D5901BCULL, "vxw_bonepitmining_a2_hr_0_0.vlm"},
	{0x2B3D2A4C018F4E73ULL, "6128648_m.fxe"},
	{0x2B3D6D29569A9D51ULL, "470226231.ogg"},
	{0x2B3D71D6281F13F0ULL, "mag360_samson6063771.cl"},
	{0x2B3DC50333FC7A23ULL, "117412166.ogg"},
	{0x2B3E7749C13CE040ULL, "vxw_blightlands_cz_0_0.vlm"},
	{0x2B3E9AFAC694963DULL, "vxlt_0_lowtown_a3_n_c_fe9_0_0.vlm"},
	{0x2B3ED51F86DAC0DEULL, "hlac_w_sundermount_262142_0z.dds"},
	{0x2B3F3E0F90300E33ULL, "plt_lgt200pt_tomes.plo"},
	{0x2B3F48B8347209E9ULL, "154967098.ogg"},
	{0x2B3F8743F4856A05ULL, "qcr000_arishok6060120.cl"},
	{0x2B3FE83D6953AD07ULL, "6127057_f.fxe"},
	{0x2B40A7AAF7C749CAULL, "hlac_lt_0_lowtown_a2_n_-3_0z.dds"},
	{0x2B40D874F830CB80ULL, "hlac_ht_mansion2b_818_26x.dds"},
	{0x2B413813215594FDULL, "6092507_m.fxe"},
	{0x2B41D2CEFA2DBD2FULL, "art\\levels\\ztools\\default\\generic\\3wi_basic.phy"},
	{0x2B4206AF99021E3CULL, "llac_w_woundedcoast_-131070_0.dds"},
	{0x2B427135CE033766ULL, "vxht_chantry_n_1yn_0_0.vlm"},
	{0x2B42D1F2058E013AULL, "zz_approval_debug6133713.cl"},
	{0x2B4301AA6431E0E6ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.impa.gad"},
	{0x2B431E031204DD37ULL, "one100dg_vanard6054829.cl"},
	{0x2B4353D171F5F85AULL, "43700125.ogg"},
	{0x2B440C9790E70742ULL, "115990528.ogg"},
	{0x2B44179FCDF6BBF3ULL, "gen00fl_merrill6129937.cl"},
	{0x2B444E84632EB18BULL, "and230_mage_girl6135819.cl"},
	{0x2B452A219B313DFEULL, "lgt110_gxa_vgk_silver_order6117018.cl"},
	{0x2B4552BCB1F01CC0ULL, "109763330.ogg"},
	{0x2B46017CFA07CFCFULL, "6119198_m.fxe"},
	{0x2B466B6DB6CCD8DDULL, "face2_s2y.psh.fxo"},
	{0x2B468DA3470E5DE8ULL, "dae000_gamlen6149421.cl"},
	{0x2B46902A1CF71DD2ULL, "vxdr_primevaltomb_7kt_0_0.vlm"},
	{0x2B4706222A5E4E04ULL, "vxw_encounter2_px_0_0.vlm"},
	{0x2B477C1AF330D879ULL, "6138792_f.fxe"},
	{0x2B4802029F4101E3ULL, "vxlt_gallowscourtyard_6x9_0_0.vlm"},
	{0x2B481CF862F356A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_front_arc_clear.ani"},
	{0x2B482B574FAB3476ULL, "mag360_samson6042084.cl"},
	{0x2B48803C7EF9DE5DULL, "dae000cr_grand_cleric6147729.cl"},
	{0x2B48D096FC767212ULL, "lt_0_lowtown_a3_n_c_24_0.rml"},
	{0x2B48FCAA643643C5ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.impa.evt"},
	{0x2B4971278E56E123ULL, "783025644.ogg"},
	{0x2B4AC859388618ABULL, "llac_lt_undercitydungeon_-1_2.dds"},
	{0x2B4AF3C429EFE184ULL, "156763701.ogg"},
	{0x2B4B9FF43CFFF260ULL, "6115580_m.fxe"},
	{0x2B4BC3624C1974F9ULL, "6122068_m.fxe"},
	{0x2B4BDE47ECC4C09DULL, "905780031.ogg"},
	{0x2B4C5587A4CD3A70ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medp_0s.dds"},
	{0x2B4CA2F5EF87186CULL, "llac_lt_undercityencounter_-1_2.dds"},
	{0x2B4CB0C08E5574D2ULL, "707663199.ogg"},
	{0x2B4D7F5344989A31ULL, "docks_male_human_worker_sta6175990.cl"},
	{0x2B4D8747FC7ACE1FULL, "63675600.ogg"},
	{0x2B4F27F862F98843ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_front_arc_clear.gad"},
	{0x2B50FD9BF2C473A1ULL, "50098913.ogg"},
	{0x2B511442317BC520ULL, "720577615.ogg"},
	{0x2B51D41301888607ULL, "hlac_lt_0_lowtown_131072_2x.dds"},
	{0x2B51D8B6F361D6A3ULL, "hlac_lt_0_lowtown_a2_n_-131073_2y.dds"},
	{0x2B53EBCB808889A5ULL, "vfx_base.gda"},
	{0x2B54D28F74737F7BULL, "art\\vfx\\combat\\vfx_r_dwp_combo_breaker_c.mmh"},
	{0x2B54E2FFD2C7EA38ULL, "vxlt_gallowstemplar_f_4qr_0_0.vlm"},
	{0x2B54FC993E20DFFEULL, "dae000cr_grand_cleric6185384.cl"},
	{0x2B55435303F3670DULL, "6147119_m.fxe"},
	{0x2B557DC1AB43A433ULL, "vxca_caverns2_7jx_0_0.vlm"},
	{0x2B558A2DC20BAE7DULL, "6104188_f.fxe"},
	{0x2B55D457F7CA0CE1ULL, "hlac_dr_primevaltomb_659_9x.dds"},
	{0x2B55DD1944DB690AULL, "vxht_0_hightown_a2_bfo_0_0.vlm"},
	{0x2B55F759765583ECULL, "6142389_f.fxe"},
	{0x2B5604F862FF6B28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_front_arc_clear.evt"},
	{0x2B5662E38A650248ULL, "6208675_m.fxe"},
	{0x2B56A53F286FE4F0ULL, "m_mag331ar_evelinas_hideout.bnk"},
	{0x2B56C4CEBF0B9A20ULL, "194515894.ogg"},
	{0x2B57259E17C5D078ULL, "follower_banter6173026.cl"},
	{0x2B57B24AC485477BULL, "zz_fen_debug6134331.cl"},
	{0x2B57B571F0B0886FULL, "qun111_ketojan6039243.cl"},
	{0x2B5826345A4D6CD3ULL, "apostates_courage.dds"},
	{0x2B58279CDDCF4DF0ULL, "hlac_dr_dungeon_3215_3x.dds"},
	{0x2B58527A59A1A669ULL, "6125735_m.fxe"},
	{0x2B592E7EDC44DCFEULL, "976691105.ogg"},
	{0x2B593B8C96ACBFCFULL, "97205578.ogg"},
	{0x2B59E2D93072DCE2ULL, "vxlt_gallowstemplar_5yv_0_0.vlm"},
	{0x2B5A340AE3407380ULL, "vxlt_0_lowtown_a3_n_c_eom_0_0.vlm"},
	{0x2B5A58349E3E0DCBULL, "lt_0_lowtown_a2_14_0.rml"},
	{0x2B5B5B03A5291140ULL, "vxw_blightlands_ix_0_0.vlm"},
	{0x2B5B7C813E8813DDULL, "dae000_madam_lusine6156786.cl"},
	{0x2B5D5F335DEFCF23ULL, "6056606_m.fxe"},
	{0x2B5D6AC11C5F0871ULL, "vxlt_0_lowtown_f0r_0_0.vlm"},
	{0x2B5E507E890B4518ULL, "w_sundermount_116_0.phy"},
	{0x2B5EA955E19A2B7CULL, "427513253.ogg"},
	{0x2B5EAE5C4998A91AULL, "ht_0_hightown_59_0.rml"},
	{0x2B5EAEF95B8002BBULL, "fen210_hadriana6110015.cl"},
	{0x2B5F1A9F2569562EULL, "44302968.ogg"},
	{0x2B5FCF18B211DB19ULL, "qun221_viscount6095608.cl"},
	{0x2B60375E9C077A94ULL, "gen00fl_isabela6015828.cl"},
	{0x2B606502B71CEBF5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_x2keep_1.mao"},
	{0x2B612DA0F51E2A83ULL, "lt_0_lowtown_docks_n_225_0.mao"},
	{0x2B6131A0F51E31D0ULL, "lt_0_lowtown_docks_n_225_0.mmh"},
	{0x2B613FA0F51E49EAULL, "lt_0_lowtown_docks_n_225_0.msh"},
	{0x2B625E3EEEF2559DULL, "hlac_ht_0_hightown_n_65537_2y.dds"},
	{0x2B629DCA61725003ULL, "art\\vfx\\combat\\vfx_glb_ele_s_additive_c.anb"},
	{0x2B62F38F747FF4E0ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_breaker_c.anb"},
	{0x2B638262C953725CULL, "vxlt_warehouse_n_11f_0_0.vlm"},
	{0x2B63D299AC848BA0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_magb_0d.dds"},
	{0x2B63D900DF767AC3ULL, "llac_lt_undercityserialkiller_-1_23.dds"},
	{0x2B63F3230B2B4086ULL, "87243185.ogg"},
	{0x2B6427E56400AE6CULL, "zz_vault_debug6177763.cl"},
	{0x2B644EF636D1818AULL, "vxht_0_hightown_a2_bkh_0_0.vlm"},
	{0x2B6454F4032DCA9BULL, "vxw_woundedcoast_qa_0_0.vlm"},
	{0x2B6489808CEFC9E6ULL, "689713255.ogg"},
	{0x2B649A67385BE17EULL, "6165447_f.fxe"},
	{0x2B64A38CCB014CFDULL, "vxht_bartrand_n_659_0_0.vlm"},
	{0x2B64E2416994F247ULL, "vxlt_0_lowtown_docks_n_c6i_0_0.vlm"},
	{0x2B654F7E89116230ULL, "w_sundermount_116_0.rml"},
	{0x2B65F6C165C72072ULL, "6116070_f.fxe"},
	{0x2B662864F4A7DA32ULL, "gen00fl_merrill6150545.cl"},
	{0x2B6669B82CC02335ULL, "6034367_m.fxe"},
	{0x2B66E4BFB7C8B7EFULL, "vxlt_darktown_4jw_0_0.vlm"},
	{0x2B6731E755A2A1B0ULL, "art\\levels\\ztools\\default\\generic\\1wo_angle.mao"},
	{0x2B6735E755A2A8C3ULL, "art\\levels\\ztools\\default\\generic\\1wo_angle.mmh"},
	{0x2B673925CE13389CULL, "126019065.ogg"},
	{0x2B6743E755A2C01DULL, "art\\levels\\ztools\\default\\generic\\1wo_angle.msh"},
	{0x2B675118566E8CFCULL, "160007507.ogg"},
	{0x2B681A3CA6BD66D7ULL, "vxlt_0_lowtown_a2_n_c_hr5_0_0.vlm"},
	{0x2B684A991144A51AULL, "mag121_viveka6082600.cl"},
	{0x2B68650B3184C8D3ULL, "vxdr_primevaltomb_7zr_0_0.vlm"},
	{0x2B687F28F73E4148ULL, "361458539.ogg"},
	{0x2B68856629054FBFULL, "hlac_lt_0_lowtown_docks_a3_n_-2_2x.dds"},
	{0x2B6887BE0E81B476ULL, "6202187_m.fxe"},
	{0x2B68A5E7D5368C00ULL, "vxlt_0_lowtown_a3_n_c_erj_0_0.vlm"},
	{0x2B68E60AF0E3B401ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_4x.dds"},
	{0x2B68EF90D8C7087CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tall.mao"},
	{0x2B68FB90D8C71C1FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tall.mmh"},
	{0x2B690190D8C726C1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tall.msh"},
	{0x2B6902C666994637ULL, "vxlt_0_lowtown_a2_n_c_hm4_0_0.vlm"},
	{0x2B697193C6514F16ULL, "art\\vfx\\creatures\\fxc_ogre_db_a_c.impa.evt"},
	{0x2B69B5EB103ED4FFULL, "213502277.ogg"},
	{0x2B69B7B9E36EB68AULL, "vxlt_gallowstemplar_5j7_0_0.vlm"},
	{0x2B69F525EFDFFB63ULL, "vxdr_primevaltomb_4xg_0_0.vlm"},
	{0x2B6A6FEC97C56782ULL, "247008223.ogg"},
	{0x2B6AA7AD85975523ULL, "63721769.ogg"},
	{0x2B6AF2A0FE8F047EULL, "298426847.ogg"},
	{0x2B6B030720FB85F6ULL, "hlac_lt_0_lowtown_docks_n_-2_4z.dds"},
	{0x2B6B326FB2A2CA7CULL, "follower_banter6167870.cl"},
	{0x2B6B8C10B8A1FA58ULL, "vxlt_0_lowtown_n_e1l_0_0.vlm"},
	{0x2B6C371D72803964ULL, "character_character_p128.psh.fxo.dx11.assembly"},
	{0x2B6CF9DA97299055ULL, "mag311_emile_de_launcet6043602.cl"},
	{0x2B6D794DC53F2AE9ULL, "82240050.ogg"},
	{0x2B6DA52970A08CB6ULL, "160218789.ogg"},
	{0x2B6E23AED06EEAF5ULL, "6218925_m.fxe"},
	{0x2B6E2B24E4A59969ULL, "34909150.ogg"},
	{0x2B6E2C4229AB52FAULL, "6147196_m.fxe"},
	{0x2B6F250F905907D0ULL, "plt_lgt200pt_tomes.nss"},
	{0x2B6F67815AFAFC96ULL, "vxht_0_hightown_a3_b22_0_0.vlm"},
	{0x2B6F7060A2BEDBD8ULL, "llac_dr_dungeon_3217_10.dds"},
	{0x2B6FB599E98DF3E3ULL, "88908590.ogg"},
	{0x2B6FFED1397FCCFFULL, "6208878_m.fxe"},
	{0x2B7005C3DB192DF9ULL, "26053313.ogg"},
	{0x2B7011885CE4A851ULL, "llac_lt_gallowstemplar_2391_4.dds"},
	{0x2B706893C6575EA1ULL, "art\\vfx\\creatures\\fxc_ogre_db_a_c.impa.gad"},
	{0x2B707ACA617E522CULL, "art\\vfx\\combat\\vfx_glb_ele_s_additive_c.mmh"},
	{0x2B713E82BF6C4229ULL, "752250707.ogg"},
	{0x2B71B6662CCB9360ULL, "vxlt_0_lowtown_docks_a3_n_bqq_0_0.vlm"},
	{0x2B7268719E1881A5ULL, "167617151.ogg"},
	{0x2B73110381560C4AULL, "vxht_chantry_n_286_0_0.vlm"},
	{0x2B73541E5B8B0AC7ULL, "vxlt_0_lowtown_docks_n_c1h_0_0.vlm"},
	{0x2B7372E895EFA128ULL, "563972738.ogg"},
	{0x2B737CE34584B257ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5closed.mmh"},
	{0x2B7380E34584B964ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5closed.mao"},
	{0x2B7392E34584D7C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5closed.msh"},
	{0x2B73F67F474949D8ULL, "6079952_m.fxe"},
	{0x2B74A6C463B3343CULL, "69443466.ogg"},
	{0x2B74F316B4B2A2A2ULL, "plt_one200pt_herbalist.plo"},
	{0x2B75078E19D18FFDULL, "mag110_arianni6032473.cl"},
	{0x2B7543974640480FULL, "vxlt_0_lowtown_a2_n_c_ehz_0_0.vlm"},
	{0x2B7591B223F3155AULL, "t1_skn_e10.tnt"},
	{0x2B75E76AB721B51DULL, "char_interior.arl"},
	{0x2B76022C823DFA07ULL, "vxlt_0_lowtown_docks_a3_czi_0_0.vlm"},
	{0x2B76D6E8237AE153ULL, "vxdr_primevaltomb_7wu_0_0.vlm"},
	{0x2B77368354219F2CULL, "133688078.ogg"},
	{0x2B776993C65D7F4FULL, "art\\vfx\\creatures\\fxc_ogre_db_a_c.impa.ani"},
	{0x2B777D85059959A7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.pusheddown.gad"},
	{0x2B7836E755B11DE8ULL, "art\\levels\\ztools\\default\\generic\\1wo_angle.phy"},
	{0x2B783EBD89154240ULL, "vxw_blightlands_wf_0_0.vlm"},
	{0x2B7855D276047BDEULL, "120971147.ogg"},
	{0x2B7969D6BD7EE000ULL, "279124155.ogg"},
	{0x2B79805780ADD1EBULL, "vxw_woundedcoast_y9_0_0.vlm"},
	{0x2B79BF156E7004C2ULL, "qun110_petrice6205994.cl"},
	{0x2B7A007C4D20EE79ULL, "zz_followers_banter6158688.cl"},
	{0x2B7A2D433BCD18F3ULL, "737711618.ogg"},
	{0x2B7A3A6F4DF49FFEULL, "6068673_m.fxe"},
	{0x2B7A58FB18D2214DULL, "mind_conj_distor.mao"},
	{0x2B7A6220370A72D5ULL, "gen00fl_anders6136014.cl"},
	{0x2B7AC3BACDF1726FULL, "mrl210_merrill6102121.cl"},
	{0x2B7B366177D9343DULL, "28176397.ogg"},
	{0x2B7B873F288ECDBAULL, "m_mag331ar_evelinas_hideout.wbd"},
	{0x2B7BDFDC2A5921FEULL, "hf_evelina.mor"},
	{0x2B7C73F48791D273ULL, "120776376.ogg"},
	{0x2B7C8C67CDF6B37EULL, "art\\levels\\ztools\\helpers\\texels\\_textures\\2048x1024m.dds"},
	{0x2B7CA07875D58D61ULL, "lt_0_lowtown_docks_a3_n_42_0.rml"},
	{0x2B7CDFEF94DE400CULL, "dae000_bartrand6031131.cl"},
	{0x2B7CE3BCB2C7F8E7ULL, "vxlt_backalley_n_1l3_0_0.vlm"},
	{0x2B7D3FD08C89F316ULL, "77431608.ogg"},
	{0x2B7DAC9DF727E520ULL, "136665100.ogg"},
	{0x2B7DCFCCAB68466AULL, "vxht_keep_n_61y_0_0.vlm"},
	{0x2B7E11408BC452E2ULL, "348864991.ogg"},
	{0x2B7E4E4CFBCA639BULL, "llac_lt_hangedman_-1_15.dds"},
	{0x2B7E7285059F65B4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.pusheddown.evt"},
	{0x2B7F6318AC42ACA4ULL, "vxht_0_hightown_a3_b8r_0_0.vlm"},
	{0x2B7F80F9F2591260ULL, "vxw_bonepitdestroyed_a3_n2_0_0.vlm"},
	{0x2B7FDCF82F7C3927ULL, "6201279_m.fxe"},
	{0x2B8028E6BAC88AA8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_low.mmh"},
	{0x2B8034E6BAC89F4BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_low.mao"},
	{0x2B8046E6BAC8BDD2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_low.msh"},
	{0x2B8048E3DD03383EULL, "art\\vfx\\combat\\vfx_w_off_penetra_activ_c.impa.evt"},
	{0x2B809DAFC177A544ULL, "vxlt_0_lowtown_n_e2i_0_0.vlm"},
	{0x2B80F514F6E7FBFDULL, "6064832_m.fxe"},
	{0x2B80F87DE14B82C8ULL, "qcr220_grey_wardens6135963.cl"},
	{0x2B8120F21C7AE484ULL, "vxw_blightlands_6q_0_0.vlm"},
	{0x2B822910DAACD610ULL, "6174044_m.fxe"},
	{0x2B823C42BFC4883BULL, "6149933_m.fxe"},
	{0x2B8253465623E374ULL, "vxlt_backalley_g_1en_0_0.vlm"},
	{0x2B827FC96C2DA9ECULL, "vxht_0_hightown_a3_c0k_0_0.vlm"},
	{0x2B82D59EA156D162ULL, "fen121ip_locked_door.utp"},
	{0x2B830BE2C3D6E10AULL, "mrl000_marethari6129281.cl"},
	{0x2B834E6B447E4AD4ULL, "gen_im_npc_elf_arm_cht_encb.uti"},
	{0x2B837425C5A3FCFEULL, "art\\vfx\\cinematic\\vfx_ogre_footsteps_p.mmh"},
	{0x2B83945A24186232ULL, "988850266.ogg"},
	{0x2B83C012F6DB3A95ULL, "vxw_sundermount_2ck_0_0.vlm"},
	{0x2B83D39C8211EBABULL, "vxlt_0_lowtown_docks_a3_n_d32_0_0.vlm"},
	{0x2B841E52C00BB7B4ULL, "182170933.ogg"},
	{0x2B848829DABE45E9ULL, "429070562.ogg"},
	{0x2B8589A1B922BD00ULL, "vxlt_0_lowtown_a3_n_c_exh_0_0.vlm"},
	{0x2B85BE8B26B1A845ULL, "vxht_0_hightown_n_caw_0_0.vlm"},
	{0x2B85D39DACAF2B24ULL, "mag291_ntbanter6056603.cl"},
	{0x2B85D6F2A8FC93DEULL, "cod_bks_lite_vials_1.utp"},
	{0x2B8731F6732FCC6AULL, "6093672_f.fxe"},
	{0x2B874D4D2CA738D3ULL, "one280_merin6105450.cl"},
	{0x2B875FE3DD097E69ULL, "art\\vfx\\combat\\vfx_w_off_penetra_activ_c.impa.gad"},
	{0x2B889B7FEBC7508DULL, "6135368_f.fxe"},
	{0x2B890D02B1B3DC61ULL, "lowtown_fem_alienage1_sta6162541.cl"},
	{0x2B8938CFACE81021ULL, "dae200_cullen6140310.cl"},
	{0x2B89D3EED7D5BE92ULL, "vxdr_0_deeproads_f0b_0_0.vlm"},
	{0x2B8A063CD066EA4FULL, "vxdr_0_deeproads_crr_0_2.vlm"},
	{0x2B8A98571B9C7914ULL, "vxlt_undercityserialkiller_1qf_0_0.vlm"},
	{0x2B8B5599A4BE1167ULL, "vxlt_backalley_n_1w2_0_0.vlm"},
	{0x2B8B90EB1D561922ULL, "vxlt_gallowscourtyard_59t_0_0.vlm"},
	{0x2B8B99075C7724BCULL, "llac_lt_0_lowtown_a2_n_c_-196609_2.dds"},
	{0x2B8C4D5D508EF95BULL, "hlac_ht_0_hightown_a3_-131074_1y.dds"},
	{0x2B8C6A8505AB95EDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.pusheddown.ani"},
	{0x2B8C98BE54AA032FULL, "52870779.ogg"},
	{0x2B8CE0610F9297A5ULL, "hlac_lt_0_lowtown_docks_-131074_7y.dds"},
	{0x2B8D6EFF0252F61AULL, "515595260.ogg"},
	{0x2B8D98700642BA78ULL, "rdr110cr_guard_sleeping.ncs"},
	{0x2B8DEAE8608D40A8ULL, "676299997.ogg"},
	{0x2B8DF2D6E44F2AE0ULL, "vxw_bonepitdestroyed_a3_i3_0_0.vlm"},
	{0x2B8E30E3DD0F4CB7ULL, "art\\vfx\\combat\\vfx_w_off_penetra_activ_c.impa.ani"},
	{0x2B8E9EBAF6074611ULL, "6060340_m.fxe"},
	{0x2B8EBB122F682FD6ULL, "615282943.ogg"},
	{0x2B8EC30C440DBAE0ULL, "hlac_lt_0_lowtown_-65536_0y.dds"},
	{0x2B8EE8F928682711ULL, "vxw_bonepit_27l_0_0.vlm"},
	{0x2B9007C350495BA3ULL, "mag170_recruit026080924.cl"},
	{0x2B904ED4C33EBD6BULL, "gen00fl_varric6076035.cl"},
	{0x2B90794F5AA798E3ULL, "mcr310_argument6072309.cl"},
	{0x2B90B6093967139AULL, "gen00fl_merrill6126315.cl"},
	{0x2B90C5234819FBF4ULL, "vxlt_backalley_g_1bo_0_0.vlm"},
	{0x2B90E691DA58CB77ULL, "346564164.ogg"},
	{0x2B90F1A65E23C26CULL, "vxht_0_hightown_a3_c5h_0_0.vlm"},
	{0x2B915325C5B001BDULL, "art\\vfx\\cinematic\\vfx_ogre_footsteps_p.anb"},
	{0x2B91A8C5A40606FFULL, "vxlt_0_lowtown_a2_n_c_ei9_0_0.vlm"},
	{0x2B91ED29FB31DF89ULL, "37221891.ogg"},
	{0x2B91F12D015F4E8EULL, "6099336_m.fxe"},
	{0x2B91FDD6CA1F14E6ULL, "151854493.ogg"},
	{0x2B9231EFE8D15315ULL, "vxw_sundermount_2dj_0_0.vlm"},
	{0x2B93D97540C5F33DULL, "504819046.ogg"},
	{0x2B9449581A0F6948ULL, "gen00fl_fenris6113865.cl"},
	{0x2B945E4F62C38C01ULL, "vxlt_warehouse_n_166_0_0.vlm"},
	{0x2B9491A2BEF0D686ULL, "vxca_caverns1_mines_jff_0_0.vlm"},
	{0x2B9512506D0F31E0ULL, "llac_lt_gallowscourtyard_n_-65536_1.dds"},
	{0x2B954ABCC5C244E3ULL, "vxdr_primevaltomb_4ih_0_0.vlm"},
	{0x2B957D33CCB3C229ULL, "6067979_f.fxe"},
	{0x2B95D003B8F09A77ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appa_0s.dds"},
	{0x2B963A58BC12DE2DULL, "6151048_m.fxe"},
	{0x2B96DDC7C8BC1D33ULL, "772289334.ogg"},
	{0x2B96F36B2A0E1913ULL, "dae110ar_lowtown.are"},
	{0x2B97F480DD1B835EULL, "llac_lt_0_lowtown_docks_n_-131075_5.dds"},
	{0x2B98A3B3D36FA0BFULL, "6115420_m.fxe"},
	{0x2B9956BA4E778E36ULL, "character_character_p267.psh.fxo.dx11"},
	{0x2B999C8FC5C41F9FULL, "vxlt_shop_e0_0_0.vlm"},
	{0x2B99A269E3A0C7F0ULL, "hlac_w_encounter3_-196606_0z.dds"},
	{0x2B9A17CF18A6A3BBULL, "art\\vfx\\combat\\vfx_g_appearance_c.impa.evt"},
	{0x2B9A29B471A2D7B0ULL, "w_bonepit_168_0.mao"},
	{0x2B9A2DB471A2DEC3ULL, "w_bonepit_168_0.mmh"},
	{0x2B9A3BB471A2F61DULL, "w_bonepit_168_0.msh"},
	{0x2B9AA1BCCA3543C5ULL, "vxlt_gallowstemplar_436_0_0.vlm"},
	{0x2B9ADAA77B7965E2ULL, "follower_barks6153825.cl"},
	{0x2B9B18AD7CC94C0AULL, "vxlt_0_lowtown_a2_hgj_0_0.vlm"},
	{0x2B9B37C8CBAEF4BBULL, "gen00fl_anders6111169.cl"},
	{0x2B9B608D78B7515DULL, "309834952.ogg"},
	{0x2B9B6DE77E67EA6CULL, "936193940.ogg"},
	{0x2B9B70A3288799DAULL, "fen310_fenris6118191.cl"},
	{0x2B9B7935856B469BULL, "hlac_w_bonepit_458757_0y.dds"},
	{0x2B9C09B7904FBA0CULL, "vxlt_warehouse_i4_0_0.vlm"},
	{0x2B9CA77EAA533582ULL, "zz_vault_debug6200388.cl"},
	{0x2B9CF08A6F520BE3ULL, "vxlt_0_lowtown_docks_n_b5t_0_1.vlm"},
	{0x2B9D01D3724C82F2ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firebeam2.mmh"},
	{0x2B9D0F73992FFEDDULL, "6089428_m.fxe"},
	{0x2B9D34746736D4E5ULL, "179031677.ogg"},
	{0x2B9D6243C70864BBULL, "vxht_0_hightown_b66_0_0.vlm"},
	{0x2B9D63AC6F1FB889ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.dura.ani"},
	{0x2B9DC61A547EA574ULL, "gen00fl_isabela6085612.cl"},
	{0x2B9DCD09A6CB271EULL, "vxlt_0_lowtown_a2_n_c_d06_0_0.vlm"},
	{0x2B9F597F35502B7EULL, "6032992_f.fxe"},
	{0x2B9F9A914391ADD7ULL, "ave310_gen_guardsman.utc"},
	{0x2BA010CF18AB03FCULL, "art\\vfx\\combat\\vfx_g_appearance_c.impa.gad"},
	{0x2BA01E39EA94DDC5ULL, "vxht_0_hightown_n_c3l_0_1.vlm"},
	{0x2BA091BFA2E55835ULL, "vxw_bonepitmining_a2_gn_0_0.vlm"},
	{0x2BA0A3CCDAC76B95ULL, "vxw_sundermount_2ii_0_0.vlm"},
	{0x2BA0A71482787DEAULL, "tv_mockup1.cut"},
	{0x2BA0D2193EC6123BULL, "llac_ht_mansion2_818_32.dds"},
	{0x2BA1282C8A5F19ACULL, "vxlt_gallowstemplar_4t2_0_0.vlm"},
	{0x2BA153BF95BC811BULL, "mer240_jansen6081415.cl"},
	{0x2BA1762F240E46A8ULL, "docks_male_human_worker_sta6176075.cl"},
	{0x2BA1A011E441495DULL, "589119097.ogg"},
	{0x2BA249E28C99BD80ULL, "421607884.ogg"},
	{0x2BA2703A531D20B9ULL, "gen00fl_aveline6114831.cl"},
	{0x2BA29B2B67AB7322ULL, "93620840.ogg"},
	{0x2BA4358DF8BF3A09ULL, "qcr000_cut2d_thechampion6025121.cl"},
	{0x2BA45DA62405D961ULL, "6070473_m.fxe"},
	{0x2BA47EAC6F26050BULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.dura.gad"},
	{0x2BA4962EB8075766ULL, "126891621.ogg"},
	{0x2BA4AFE143DDF6F5ULL, "vxw_bonepitmining_a2_2a4_0_0.vlm"},
	{0x2BA4F65E18A3991AULL, "6142331_m.fxe"},
	{0x2BA54F41303BA286ULL, "art\\vfx\\creatures\\fxc_lotf_c.anb"},
	{0x2BA5AF7006577362ULL, "rdr110cr_guard_sleeping.utc"},
	{0x2BA5C4F630C49BCDULL, "6018421_f.fxe"},
	{0x2BA6D094BCF03BF3ULL, "lt_0_lowtown_a2_28_0_lc.mmh"},
	{0x2BA70FCF18B120FAULL, "art\\vfx\\combat\\vfx_g_appearance_c.impa.ani"},
	{0x2BA71E3236A8CE7FULL, "478941334.ogg"},
	{0x2BA77E99166CB2EAULL, "w_sundermount_240_0.mao"},
	{0x2BA78299166CB939ULL, "w_sundermount_240_0.mmh"},
	{0x2BA78C99166CCA27ULL, "w_sundermount_240_0.msh"},
	{0x2BA7DBB378DF82EEULL, "hlac_w_sundermount_262139_0x.dds"},
	{0x2BA8640B8C196649ULL, "hlac_lt_gallowstemplar_f_2392_1y.dds"},
	{0x2BA87CCACA903E7BULL, "vxca_caverns1_mines_im5_0_0.vlm"},
	{0x2BA8890A10FC052CULL, "mcr310_argument6216182.cl"},
	{0x2BA8F734852484D0ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_lava.dds"},
	{0x2BA906E4E67CB8E8ULL, "llac_w_bonepitmining_a2_131074_0.dds"},
	{0x2BA9A62E4158F093ULL, "lt_0_lowtown_docks_a3_30_0_ll.mmh"},
	{0x2BA9EA470F70E6CBULL, "zz_fs.ncs"},
	{0x2BAA76E04488855AULL, "llac_lt_0_lowtown_a2_n_0_8.dds"},
	{0x2BAA7964077C02B1ULL, "grey.dds"},
	{0x2BAA7B948245D28CULL, "vxlt_warehouse_r5_0_0.vlm"},
	{0x2BAA8974D188FC02ULL, "6016694_m.fxe"},
	{0x2BAADE8E3B62C3E9ULL, "vxlt_0_lowtown_docks_ctj_0_0.vlm"},
	{0x2BAB2EB471B153E8ULL, "w_bonepit_168_0.phy"},
	{0x2BAB7BAC6F2C1F70ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.dura.evt"},
	{0x2BABA9AD89EB336DULL, "513064451.ogg"},
	{0x2BABEC5A70DEA8EAULL, "435778673.ogg"},
	{0x2BAD6BCCB02BEEDAULL, "6141136_m.fxe"},
	{0x2BAECB10BA97C49BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyo_0t.dds"},
	{0x2BAED4E6E54B7F8EULL, "6093136_m.fxe"},
	{0x2BAF15A9CCBD8415ULL, "vxw_sundermount_2rh_0_0.vlm"},
	{0x2BAF3FBBC55456CCULL, "zz_one_debug6113369.cl"},
	{0x2BB011BE33EC540CULL, "6145142_f.fxe"},
	{0x2BB1687B8EAF06EEULL, "hlac_lt_0_lowtown_a3_n_c_-65537_6x.dds"},
	{0x2BB29B4F35252C4EULL, "985619882.ogg"},
	{0x2BB2D0EE07E50C4EULL, "318935023.ogg"},
	{0x2BB313BC816B4D73ULL, "6019096_m.fxe"},
	{0x2BB315FA80C04612ULL, "hlac_ca_caverns1_24423_10y.dds"},
	{0x2BB3322FC7F1BD9BULL, "war_staminaregen.dds"},
	{0x2BB352413047E531ULL, "art\\vfx\\creatures\\fxc_lotf_c.mmh"},
	{0x2BB36F3BCD2FE66BULL, "778003390.ogg"},
	{0x2BB401F4224EFB86ULL, "gen00fl_bethany6101011.cl"},
	{0x2BB4B98CEF87CE49ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_sa.dds"},
	{0x2BB523A964742176ULL, "6092063_m.fxe"},
	{0x2BB55D9C3F20ACCAULL, "807112070.ogg"},
	{0x2BB5E77438301F21ULL, "gen00fl_isabela6135695.cl"},
	{0x2BB5E916B539C038ULL, "vrc221_bartrand6043392.cl"},
	{0x2BB613A223AC73D4ULL, "monster1.cut"},
	{0x2BB64282AB302EB2ULL, "6137717_m.fxe"},
	{0x2BB65AB79B0C0D62ULL, "vxca_caverns2_80t_0_0.vlm"},
	{0x2BB66082C1AA6DDBULL, "vxht_keep_675_0_0.vlm"},
	{0x2BB6652AC60A7DBFULL, "dae000_tomwise6167277.cl"},
	{0x2BB6AB1CEEBA68CCULL, "mer320_hubert6065843.cl"},
	{0x2BB6EEA7BC8656FBULL, "vxca_caverns1_mines_ij4_0_0.vlm"},
	{0x2BB70B474713A999ULL, "6129426_f.fxe"},
	{0x2BB7288B0302546BULL, "hlac_dr_slavers_3216_7x.dds"},
	{0x2BB77D70E5B33B37ULL, "w_encounter3_215_0.msh"},
	{0x2BB78B71F8D1DA3BULL, "gen00fl_merrill6105676.cl"},
	{0x2BB78F70E5B3599AULL, "w_encounter3_215_0.mao"},
	{0x2BB79370E5B360A9ULL, "w_encounter3_215_0.mmh"},
	{0x2BB7D24D0A57DF1EULL, "302687062.ogg"},
	{0x2BB7FC6760B57D0AULL, "vxlt_0_lowtown_a2_hid_0_0.vlm"},
	{0x2BB85EE5F0CA0B62ULL, "6135892_m.fxe"},
	{0x2BB883DDAA6336A8ULL, "art\\levels\\races\\design\\prp_latch.msh"},
	{0x2BB895DDAA635539ULL, "art\\levels\\races\\design\\prp_latch.mao"},
	{0x2BB89FE98FDEC443ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.impa.evt"},
	{0x2BB8A1DDAA6369D2ULL, "art\\levels\\races\\design\\prp_latch.mmh"},
	{0x2BB8A2B47798AA86ULL, "hlac_lt_0_lowtown_-65536_7z.dds"},
	{0x2BB8C77F95902B36ULL, "6028534_m.fxe"},
	{0x2BB8CA5BCF1C06A6ULL, "vxlt_gallowscourtyard_71f_0_0.vlm"},
	{0x2BB9506B2D58DC69ULL, "vxlt_0_lowtown_docks_cqi_0_0.vlm"},
	{0x2BB95E671DE25954ULL, "309022763.ogg"},
	{0x2BB9FEF668D7E4B0ULL, "llac_ht_0_hightown_a2_65535_3.dds"},
	{0x2BBA35DC38946903ULL, "6026933_m.fxe"},
	{0x2BBA3D7E895983ADULL, "w_sundermount_116_0.msh"},
	{0x2BBA44F1AD8293A1ULL, "6078976_m.fxe"},
	{0x2BBA4B7E89599BE0ULL, "w_sundermount_116_0.mao"},
	{0x2BBA4F7E8959A233ULL, "w_sundermount_116_0.mmh"},
	{0x2BBA8B3A05B2C320ULL, "6107217_m.fxe"},
	{0x2BBAB9B3718CD8FEULL, "41816147.ogg"},
	{0x2BBAF29B3624D7DCULL, "sse_ss_ambient_m.gda"},
	{0x2BBB559B78EC15DFULL, "hlac_lt_gallowscourtyard_-65535_0x.dds"},
	{0x2BBB89EB0C2E8BBAULL, "hlac_lt_0_lowtown_docks_a3_n_196607_3z.dds"},
	{0x2BBB8F5CC4FEA0E6ULL, "hlac_ht_0_hightown_65536_0z.dds"},
	{0x2BBB9DB7D5A4C91FULL, "vxlt_gallowscourtyard_775_0_0.vlm"},
	{0x2BBBA94A6297E4BFULL, "vxht_0_hightown_a3_c5g_0_0.vlm"},
	{0x2BBC11A6420E5AD4ULL, "6206982_f.fxe"},
	{0x2BBC80A2A1AB4160ULL, "llac_w_encounter2_camp_262149_0.dds"},
	{0x2BBCF7D0F62E1B8AULL, "830248122.ogg"},
	{0x2BBDA3E21C1CDDE6ULL, "hlac_ht_0_hightown_a3_131071_2y.dds"},
	{0x2BBDA8D88F606429ULL, "art\\characters\\creatures\\desiredemon\\animation\\dmg_v1.ani"},
	{0x2BBDEA3F4C582072ULL, "vxlt_0_lowtown_a2_n_c_hl2_0_0.vlm"},
	{0x2BBE1D9E1E6C3D0FULL, "zz_vault_debug6203269.cl"},
	{0x2BBE8ABC0168E09BULL, "hlac_lt_0_lowtown_131071_6y.dds"},
	{0x2BBE9215A3CB063DULL, "6071760_m.fxe"},
	{0x2BBEAEFE7D88F2FAULL, "6040907_m.fxe"},
	{0x2BBED5DFA554F20EULL, "vxlt_0_lowtown_a2_h25_0_0.vlm"},
	{0x2BBEDF7FE61ACF10ULL, "fx_hand_rev.mao"},
	{0x2BBF12C9F2324051ULL, "881194172.ogg"},
	{0x2BBF3C600230F021ULL, "vxdr_primevaltomb_7ue_0_0.vlm"},
	{0x2BBF5591C17F75D4ULL, "hlac_lt_0_lowtown_n_65536_11y.dds"},
	{0x2BBF78E98FE4A0C4ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.impa.gad"},
	{0x2BBFA5DC24F98098ULL, "dae000cr_grand_cleric6183960.cl"},
	{0x2BBFD075681ED3C2ULL, "hlac_lt_gallowstemplar_f_2394_5x.dds"},
	{0x2BBFECE139E61C57ULL, "vxht_0_hightown_a3_9uy_0_0.vlm"},
	{0x2BC001B8EEE49232ULL, "333652679.ogg"},
	{0x2BC012853560920DULL, "vis121_saemus6167895.cl"},
	{0x2BC164B22DDBDF26ULL, "art\\characters\\creatures\\mubari\\animation\\kd_bk.evt"},
	{0x2BC170071C8C96E8ULL, "925075415.ogg"},
	{0x2BC18B34595795BBULL, "hlac_w_encounter3_-131068_0z.dds"},
	{0x2BC1992F917DC2B8ULL, "hlac_w_bonepitmining_a2_0_11x.dds"},
	{0x2BC201A797D0865BULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_1.mao"},
	{0x2BC28C22E1C0AB9BULL, "vxca_caverns1_edo_0_0.vlm"},
	{0x2BC2DA4DE607DDA0ULL, "vxht_0_hightown_ajr_0_0.vlm"},
	{0x2BC2FC6B2A3363D8ULL, "dae110ar_lowtown.lst"},
	{0x2BC3BBAA4B09464EULL, "one120tr_athe_ambush.ncs"},
	{0x2BC41216A94C56DAULL, "6118201_f.fxe"},
	{0x2BC4A73594EA70C4ULL, "mag211_marethari6166433.cl"},
	{0x2BC4C3D88F66B0ABULL, "art\\characters\\creatures\\desiredemon\\animation\\dmg_v1.gad"},
	{0x2BC4D25FB3A0865BULL, "vxht_keep_604_0_0.vlm"},
	{0x2BC512192B351FC3ULL, "6155976_m.fxe"},
	{0x2BC57ACF8021B2F2ULL, "6082170_m.fxe"},
	{0x2BC587E98FE925A2ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.impa.ani"},
	{0x2BC5972CED5629ABULL, "722183415.ogg"},
	{0x2BC5BEF06FB97A1CULL, "55784427.ogg"},
	{0x2BC62FEF8C59968FULL, "fxp_pillar_sym02.dds"},
	{0x2BC675B9F0612399ULL, "6032897_m.fxe"},
	{0x2BC6CD82A9896647ULL, "hlac_lt_0_lowtown_docks_a3_-131075_4z.dds"},
	{0x2BC6D1A6B190C1EBULL, "llac_lt_0_lowtown_a2_-65536_8.dds"},
	{0x2BC6DFB3AEC412E5ULL, "pro000_aveline_intro.bnk"},
	{0x2BC6F854F62F02C7ULL, "6102929_m.fxe"},
	{0x2BC74D3AD13D444AULL, "vxlt_0_lowtown_a3_n_c_f5j_0_0.vlm"},
	{0x2BC77BB22DE07271ULL, "art\\characters\\creatures\\mubari\\animation\\kd_bk.gad"},
	{0x2BC79E352807C6C0ULL, "hlac_w_blightlands_131071_0x.dds"},
	{0x2BC7ACF6E51131A8ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_02a_0t.dds"},
	{0x2BC7EA9A92DAE3A7ULL, "dre000_sibling_has_blight6036499.cl"},
	{0x2BC82AE3D65B88C7ULL, "hlac_lt_undercityencounter_-1_3z.dds"},
	{0x2BC844096D48ED96ULL, "w_woundedcoast_3_0.msh"},
	{0x2BC84E096D48FE94ULL, "w_woundedcoast_3_0.mmh"},
	{0x2BC852096D4905A7ULL, "w_woundedcoast_3_0.mao"},
	{0x2BC8B997DA75437DULL, "character_character_p53.psh.fxo.dx11.assembly"},
	{0x2BCA0F94C79AE19FULL, "vxlt_gallowscourtyard_784_0_0.vlm"},
	{0x2BCA1C6B2A39B8C9ULL, "dae110ar_lowtown.ncs"},
	{0x2BCAC0D88F6B1710ULL, "art\\characters\\creatures\\desiredemon\\animation\\dmg_v1.evt"},
	{0x2BCAC1B3AEC7C81DULL, "pro000_aveline_intro.cnv"},
	{0x2BCAE87FBD2D0B3BULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb101a_0n.dds"},
	{0x2BCBBCDEA3EFC2D3ULL, "gen00fl_aveline6122146.cl"},
	{0x2BCBD087953E35E2ULL, "6095380_f.fxe"},
	{0x2BCBDFA797D8A098ULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_1.pwk"},
	{0x2BCBF963B0A9B515ULL, "vxw_sundermount_2xv_0_0.vlm"},
	{0x2BCC0490D91BA964ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tall.phy"},
	{0x2BCC069E24F9176EULL, "hlac_lt_backalley_g_0_19z.dds"},
	{0x2BCC2D8DC25338D6ULL, "6018534_f.fxe"},
	{0x2BCCAF70D385F1FBULL, "dr_0_deeproads_17038.rml"},
	{0x2BCCDED5550EF5D2ULL, "12154146.ogg"},
	{0x2BCD18A67DC00CA3ULL, "6079291_f.fxe"},
	{0x2BCDDCFF0F8602E3ULL, "lgt160_ntb_elf_conflict6163388.cl"},
	{0x2BCE7CB22DE6921FULL, "art\\characters\\creatures\\mubari\\animation\\kd_bk.ani"},
	{0x2BCF07AE4FD8588AULL, "6117250_m.fxe"},
	{0x2BCF670BBDFC4FDCULL, "vxlt_0_lowtown_a3_n_c_er6_0_0.vlm"},
	{0x2BCFD3E8720EB2A2ULL, "6112226_m.fxe"},
	{0x2BD054C9570E8EE7ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_wng_0s.dds"},
	{0x2BD0920EF8F4E2FEULL, "6086927_m.fxe"},
	{0x2BD0ABBA7F6B52DBULL, "isa311_isabela6095339.cl"},
	{0x2BD0C5AEA22A9722ULL, "gen_im_trs_apo_lgt_glv_r2.uti"},
	{0x2BD0C89EA198DC3FULL, "fen121ip_locked_door.ncs"},
	{0x2BD0E673CEFCF5C9ULL, "727685965.ogg"},
	{0x2BD0FC26B891174FULL, "6128886_m.fxe"},
	{0x2BD10BCF39EDB2A4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkf.ani"},
	{0x2BD1CA36C8BE9458ULL, "mag101_aombanter6052763.cl"},
	{0x2BD1F72500BEA876ULL, "llac_lt_gamlenshouse_n_-1_6.dds"},
	{0x2BD2221A99F6A5E5ULL, "vxw_sundermount_2i9_0_0.vlm"},
	{0x2BD22DCDC6B5124FULL, "6140986_f.fxe"},
	{0x2BD25B6AE2803EECULL, "465606003.ogg"},
	{0x2BD2C0C0E47851CDULL, "gen00fl_carver6097225.cl"},
	{0x2BD3351BD29C71E2ULL, "vxht_0_hightown_b5r_0_0.vlm"},
	{0x2BD42EEE88ACB41EULL, "6188122_m.fxe"},
	{0x2BD5B6EABF5B0556ULL, "576719025.ogg"},
	{0x2BD6232843217A78ULL, "vxw_bonepitdestroyed_a3_lp_0_0.vlm"},
	{0x2BD653A05D6AD53AULL, "gen00fl_sebastian6195853.cl"},
	{0x2BD708CF39F21906ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkf.gad"},
	{0x2BD730CCA78840B4ULL, "matdef_df_specular.dds"},
	{0x2BD7721C95BB8667ULL, "6196829_m.fxe"},
	{0x2BD7B14533F228AAULL, "180628142.ogg"},
	{0x2BD7EDF9B4FE573EULL, "6166975_m.fxe"},
	{0x2BD8477715EEEBA3ULL, "hlac_lt_0_lowtown_docks_a3_-65538_4x.dds"},
	{0x2BD867E32A99AA18ULL, "6128449_m.fxe"},
	{0x2BD86C95E9DD7B0EULL, "lt_0_lowtown_docks_a3_n_169_0.msh"},
	{0x2BD87616CA2C62BEULL, "vxht_mansion2b_1ws_0_0.vlm"},
	{0x2BD87A95E9DD935FULL, "lt_0_lowtown_docks_a3_n_169_0.mao"},
	{0x2BD88695E9DDA7FCULL, "lt_0_lowtown_docks_a3_n_169_0.mmh"},
	{0x2BD8A8B264DF2246ULL, "vxlt_gallowstemplar_n_53g_0_0.vlm"},
	{0x2BD8C7484B052F50ULL, "hlac_ca_caverns1_24424_3x.dds"},
	{0x2BD909F6A9F84A7BULL, "141813010.ogg"},
	{0x2BD9983EB030BA09ULL, "607936950.ogg"},
	{0x2BDA7A57037DC48DULL, "llac_lt_0_lowtown_a3_n_c_-65536_6.dds"},
	{0x2BDAA3ED5C16FDAFULL, "hlac_ca_caverns2prc_5802_12z.dds"},
	{0x2BDAB60FDCF2A73DULL, "6031994_f.fxe"},
	{0x2BDAEC16B5094A69ULL, "plt_one200pt_herbalist.nss"},
	{0x2BDBB4A98CA15060ULL, "vxht_bartrand_n_5rz_0_0.vlm"},
	{0x2BDBE80C708B9DB9ULL, "573476744.ogg"},
	{0x2BDC7A8887798484ULL, "follower_banter6143160.cl"},
	{0x2BDC7C26A451F8A7ULL, "vxlt_gallowstemplar_f_3gd_0_0.vlm"},
	{0x2BDC847B92712A0EULL, "isa100im_letter.uti"},
	{0x2BDCA3FD94BA8200ULL, "38787313.ogg"},
	{0x2BDCA40867677C41ULL, "lowtown_fem_prostitute_sta6174630.cl"},
	{0x2BDCCED908EC2169ULL, "85503743.ogg"},
	{0x2BDD037B671C8F08ULL, "598851646.ogg"},
	{0x2BDE03CF39F82FE5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkf.evt"},
	{0x2BDE62C4FC7001FEULL, "826651907.ogg"},
	{0x2BDEEB9847932950ULL, "vxdr_0_deeproads_boy_0_0.vlm"},
	{0x2BDF46D599595DBDULL, "vxca_caverns2_7vz_0_0.vlm"},
	{0x2BDFB71F4B9B7571ULL, "6147649_m.fxe"},
	{0x2BE04A707A84254AULL, "vxca_caverns2prc_500_0_0.vlm"},
	{0x2BE05B43EB0B7A06ULL, "81352155.ogg"},
	{0x2BE065FE05E6F502ULL, "6017819_m.fxe"},
	{0x2BE093F78BECBE65ULL, "vxw_sundermount_2r8_0_0.vlm"},
	{0x2BE0F1ED08825713ULL, "t3_ogr_b03.tnt"},
	{0x2BE1186469FDEB58ULL, "vxdr_primevaltomb_4hx_0_0.vlm"},
	{0x2BE19F38C83E4493ULL, "570832015.ogg"},
	{0x2BE1AC908CC95BE5ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_c.anb"},
	{0x2BE25D35CB84D557ULL, "793525644.ogg"},
	{0x2BE2678DFBBE786BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantryb_sa.dds"},
	{0x2BE26ACFBA76D606ULL, "1034764799.ogg"},
	{0x2BE2723330741B88ULL, "m_fen200ar_slavers_exterior.wbd"},
	{0x2BE2B3EBFCE0EC0EULL, "109524902.ogg"},
	{0x2BE3291187151B73ULL, "lowtown_fem_alienage1_sta6162561.cl"},
	{0x2BE3E805A634DECBULL, "6103131_m.fxe"},
	{0x2BE40C93C7C1CCA7ULL, "mcr371_hostile_aveline6082204.cl"},
	{0x2BE481BF99322B7BULL, "character_character_p251.psh.fxo.dx11.assembly"},
	{0x2BE496D5536DF108ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_inkwell_sa.dds"},
	{0x2BE4F1DAC5970359ULL, "character_character_p161.psh.fxo.dx11.assembly"},
	{0x2BE52D7B119E26C4ULL, "316049548.ogg"},
	{0x2BE540EF8C746401ULL, "fxp_pillar_sym02.mao"},
	{0x2BE5FB27BCF25DA5ULL, "vxw_bonepitdestroyed_a3_1q1_0_0.vlm"},
	{0x2BE61FD6EFBFAA20ULL, "llac_ca_caverns2prc_5802_22.dds"},
	{0x2BE64FBAB8785E0AULL, "vxht_0_hightown_a2_b4a_0_0.vlm"},
	{0x2BE66702328613CDULL, "6095075_f.fxe"},
	{0x2BE6C57C47010C91ULL, "6128413_m.fxe"},
	{0x2BE7115E4304B950ULL, "672989052.ogg"},
	{0x2BE7C920F553BCAAULL, "art\\characters\\playercharacter\\children\\textures\\kf_cth_grla_0s.dds"},
	{0x2BE7ED224B3722F6ULL, "vxdr_primevaltomb_8oo_0_0.vlm"},
	{0x2BE84052A043C0C8ULL, "421631467.ogg"},
	{0x2BE865C70565FDC5ULL, "641994307.ogg"},
	{0x2BE88236F46D347BULL, "346949631.ogg"},
	{0x2BE92A0CE83CC925ULL, "6054206_m.fxe"},
	{0x2BE94B0CC4F6595FULL, "vxht_0_hightown_ary_0_0.vlm"},
	{0x2BE97AF9130209E4ULL, "195312193.ogg"},
	{0x2BEA0B792DDE1BADULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_meat.mmh"},
	{0x2BEA0F792DDE22FEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_meat.mao"},
	{0x2BEA1D792DDE3A33ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_meat.msh"},
	{0x2BEAA51749943E58ULL, "348883349.ogg"},
	{0x2BEB3F824674B039ULL, "6095731_m.fxe"},
	{0x2BEBCEF3E1931517ULL, "vxca_caverns1_mines_cjr_0_0.vlm"},
	{0x2BECA97E083AA114ULL, "gen_im_cns_tom_war_berserkr.uti"},
	{0x2BED1293E0094F99ULL, "vxca_caverns2_77q_0_0.vlm"},
	{0x2BEDB8B28B4F763DULL, "vxca_caverns2_7ky_0_0.vlm"},
	{0x2BEDDB529573EB2EULL, "atl_demosplash_dxt5_dat.xml"},
	{0x2BEE172684A23941ULL, "6065260_m.fxe"},
	{0x2BEE56315050F008ULL, "testvoltex.mao"},
	{0x2BEE5E4421933CA5ULL, "vxht_keep_n_4n9_0_0.vlm"},
	{0x2BEEEED4A0D051C6ULL, "233863841.ogg"},
	{0x2BEF5F08C2C0F8C7ULL, "182209153.ogg"},
	{0x2BEFB6A03D0EE388ULL, "242391216.ogg"},
	{0x2BEFF19C843CAE74ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl180runf.ani"},
	{0x2BF034755A139D25ULL, "dre180ip_locked_tomb_door.ncs"},
	{0x2BF054F1F29B7A46ULL, "313124589.ogg"},
	{0x2BF0FFBBC68DF391ULL, "mag100_cullen6093742.cl"},
	{0x2BF14022990739DAULL, "108574142.ogg"},
	{0x2BF1C43002750A6BULL, "gst_standing_drunk.gda"},
	{0x2BF1D11B65765101ULL, "vxlt_0_lowtown_a3_n_c_erg_0_0.vlm"},
	{0x2BF1D40202D98F85ULL, "fen310_fenris6118244.cl"},
	{0x2BF1F83D049255F8ULL, "727342463.ogg"},
	{0x2BF25A0A8C192F7AULL, "727631617.ogg"},
	{0x2BF2E2031D74FF95ULL, "331148359.ogg"},
	{0x2BF3407D06218FCAULL, "vxht_mansion_n_3zq_0_0.vlm"},
	{0x2BF40992EFD7C4CEULL, "art\\characters\\heraldry\\hld_dwarven_carta.dds"},
	{0x2BF4372132A1BD43ULL, "6122149_m.fxe"},
	{0x2BF4602CAB083D6BULL, "follower_banter6149303.cl"},
	{0x2BF46D04AEE87625ULL, "vxw_bonepitdestroyed_a3_1z0_0_0.vlm"},
	{0x2BF4CB9839A81A19ULL, "909265851.ogg"},
	{0x2BF4E1664AD2A67EULL, "mag110_arianni6020488.cl"},
	{0x2BF51E39C610552CULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyo.msh"},
	{0x2BF52439C6105F1EULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyo.mmh"},
	{0x2BF53039C61073B5ULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyo.mao"},
	{0x2BF5D0A716C26762ULL, "mag100_cullen6086370.cl"},
	{0x2BF5EE9C844114D6ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl180runf.gad"},
	{0x2BF65E2BE20D9B39ULL, "gen00fl_isabela6015845.cl"},
	{0x2BF77CEDD2C3E2D4ULL, "follower_banter6168039.cl"},
	{0x2BF784AE0ED1B314ULL, "gen00fl_anders6114453.cl"},
	{0x2BF792C9C74C77F0ULL, "736993429.ogg"},
	{0x2BF7BCE9B6EC71DFULL, "vxht_0_hightown_aux_0_0.vlm"},
	{0x2BF7D3BDDC7E6732ULL, "6126684_f.fxe"},
	{0x2BF7F4307A40A3B9ULL, "267648563.ogg"},
	{0x2BF82178BA358626ULL, "hlac_ca_caverns2prc_5802_61y.dds"},
	{0x2BF89863708D8160ULL, "vxht_bartrand_n_5xx_0_0.vlm"},
	{0x2BF8EEC011F1DA5BULL, "324210012.ogg"},
	{0x2BF9471368AA88C4ULL, "688173555.ogg"},
	{0x2BF94AF60797A6B9ULL, "zz_lgt000ar_debug.lst"},
	{0x2BF9983926AD08F8ULL, "6064680_m.fxe"},
	{0x2BF9B52F678B1FFFULL, "155459680.ogg"},
	{0x2BF9E20D14DEA2CEULL, "570679163.ogg"},
	{0x2BF9E48CC6C2ECCEULL, "gen00fl_fenris6113926.cl"},
	{0x2BFA1681D888F6F3ULL, "art\\vfx\\combat\\vfx_gib_qunc_4.mao"},
	{0x2BFA24DCBFE96F06ULL, "vxht_keep_623_0_0.vlm"},
	{0x2BFA4BA5F5DE9D62ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls106a_0s.dds"},
	{0x2BFA4BC21C9109BAULL, "281747470.ogg"},
	{0x2BFA5B0B5F038314ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.anb"},
	{0x2BFB4E08C4CCFC5EULL, "mag200_feynriel6024080.cl"},
	{0x2BFB68BF99621482ULL, "nat380_sibling6207659.cl"},
	{0x2BFB7DBF17A57734ULL, "625556441.ogg"},
	{0x2BFBEFBDFD4E0E4CULL, "syncplaceableanimations.gda"},
	{0x2BFC98A309E4AE74ULL, "325287195.ogg"},
	{0x2BFCB1886357B671ULL, "hlac_ca_cave_2194_9y.dds"},
	{0x2BFDAC873D043E09ULL, "em_leech.mor"},
	{0x2BFE0762807E7E5CULL, "hlac_lt_0_lowtown_a2_n_-3_7y.dds"},
	{0x2BFE1486A395E8A5ULL, "vxlt_0_lowtown_a3_n_c_ehy_0_0.vlm"},
	{0x2BFE309B620C00A4ULL, "llac_w_bonepitdestroyed_a3_458757_0.dds"},
	{0x2BFEC14826CE401CULL, "68180182.ogg"},
	{0x2BFEC4021D876293ULL, "hlac_lt_0_lowtown_a2_-65535_1z.dds"},
	{0x2BFF15599DB9AD49ULL, "hlac_lt_0_lowtown_a2_0_8x.dds"},
	{0x2BFF5B2A022D7F37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v1.gad"},
	{0x2BFF694CDA0FFEBEULL, "6207463_m.fxe"},
	{0x2BFF801AF13D2493ULL, "hlac_ca_caverns1_24426_4x.dds"},
	{0x2BFFC573558C91AEULL, "art\\characters\\weapons\\dagger\\w_dag_dsdga.mao"},
	{0x2BFFF70C15E0490AULL, "vxca_caverns2_831_0_0.vlm"},
	{0x2C0042F8576C6981ULL, "vxlt_0_lowtown_a3_n_c_euf_0_0.vlm"},
	{0x2C00941A24A84B8DULL, "vrc321_varric6086958.cl"},
	{0x2C0099B882FCD349ULL, "art\\levels\\4cinematics\\ferelden\\prp_ballistabolt.mao"},
	{0x2C00A5B882FCE7E2ULL, "art\\levels\\4cinematics\\ferelden\\prp_ballistabolt.mmh"},
	{0x2C00A7B882FCEA98ULL, "art\\levels\\4cinematics\\ferelden\\prp_ballistabolt.msh"},
	{0x2C011F1B0EBB9A49ULL, "587421833.ogg"},
	{0x2C0125755A21F77CULL, "dre180ip_locked_tomb_door.utp"},
	{0x2C012FB50703CBF0ULL, "isa321_velasco6091885.cl"},
	{0x2C015FFADCCE198AULL, "799353323.ogg"},
	{0x2C01AD03F461137CULL, "311151026.ogg"},
	{0x2C01E1F1294194E1ULL, "gen00fl_isabela6099665.cl"},
	{0x2C02C3E6232FD1F4ULL, "vxlt_undercityserialkiller_24r_0_0.vlm"},
	{0x2C034BF80343B14FULL, "mcr370_cut_player_rules6098296.cl"},
	{0x2C03B9146027CD5FULL, "6094781_m.fxe"},
	{0x2C03E59727064351ULL, "vxlt_darktown_5cg_0_0.vlm"},
	{0x2C04545774CCB2ADULL, "gen00fl_varric6097191.cl"},
	{0x2C047AB9BF892C05ULL, "vxw_bonepitdestroyed_a3_li_0_0.vlm"},
	{0x2C0541392B0F4F83ULL, "vxht_keep_n_614_0_0.vlm"},
	{0x2C05702A02320EC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v1.evt"},
	{0x2C05D32A747BE7A8ULL, "6077327_m.fxe"},
	{0x2C062235FA3AB732ULL, "vxlt_0_lowtown_n_exy_0_0.vlm"},
	{0x2C0655B89CE75257ULL, "qun221_petrice6076808.cl"},
	{0x2C06A53DC042B3ADULL, "hlac_lt_shop_-1_6y.dds"},
	{0x2C06C4B356EDC3BBULL, "swashbklr_foresight.dds"},
	{0x2C06E24761BFAC7FULL, "6164424_m.fxe"},
	{0x2C070A40628399E0ULL, "vxht_bartrand_n_5cy_0_0.vlm"},
	{0x2C075405E876B70BULL, "mag170_thrask6020992.cl"},
	{0x2C076E836AD51E1AULL, "6118628_m.fxe"},
	{0x2C07711005BF82AFULL, "vxca_cavern1_haunted_e4n_0_0.vlm"},
	{0x2C0799B57D7FC930ULL, "505553054.ogg"},
	{0x2C07D2CA4BC32EB4ULL, "zz_testspiders.ncs"},
	{0x2C07F3187F2F8D5DULL, "6108965_m.fxe"},
	{0x2C083A0B5F0F88D7ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.mmh"},
	{0x2C0884FEEB494FE3ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v2.evt"},
	{0x2C08AF96ABCC8EAAULL, "abi_global.wbd"},
	{0x2C0986CE79ABD765ULL, "229826332.ogg"},
	{0x2C0A43F607A60E7EULL, "zz_lgt000ar_debug.are"},
	{0x2C0A4A1C7AAC5805ULL, "1048610933.ogg"},
	{0x2C0A602D73A58F91ULL, "ser221_gascard6149266.cl"},
	{0x2C0AAD908CEC6226ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_c.mmh"},
	{0x2C0AE4B347851C30ULL, "hlac_lt_0_lowtown_docks_n_0_1z.dds"},
	{0x2C0B1F0B7F17BD5EULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rockblend_n.dds"},
	{0x2C0BF421B1684CB5ULL, "shock_proj_light02.mao"},
	{0x2C0C269E19C90AB5ULL, "483790200.ogg"},
	{0x2C0C3BF68D064C4AULL, "follower_banter6170322.cl"},
	{0x2C0C3D4BAC805F45ULL, "llac_w_encounter2_196610_0.dds"},
	{0x2C0C4DB22123B379ULL, "vxlt_undercityserialkiller_1in_0_1.vlm"},
	{0x2C0CB68BC1767E03ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_overlay_ent.ani"},
	{0x2C0CE7BDB92F10F5ULL, "ts_generic_prog_coinbag.gda"},
	{0x2C0D1E84C7B8BCADULL, "vxlt_warehouse_cs_0_0.vlm"},
	{0x2C0D8B4660281F69ULL, "fx_f2c5.mao"},
	{0x2C0DB270D3BD8DC7ULL, "dr_0_deeproads_17038.anb"},
	{0x2C0E055C20AD262AULL, "328931927.ogg"},
	{0x2C0E835429378486ULL, "6027661_m.fxe"},
	{0x2C0EABAA3A39A574ULL, "vxw_bonepitmining_a2_21a_0_0.vlm"},
	{0x2C0EB4D549628201ULL, "vxlt_0_lowtown_a3_n_c_exe_0_0.vlm"},
	{0x2C0F553DBD199429ULL, "lt_0_lowtown_docks_228_0.mao"},
	{0x2C0F5DFEEB4F2C64ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v2.gad"},
	{0x2C0F613DBD19A8C2ULL, "lt_0_lowtown_docks_228_0.mmh"},
	{0x2C0F633DBD19AC78ULL, "lt_0_lowtown_docks_228_0.msh"},
	{0x2C0F67FBA82BB582ULL, "190863586.ogg"},
	{0x2C0FA4C0512CAB49ULL, "6074489_f.fxe"},
	{0x2C10AA6CFB8BAE09ULL, "gen00fl_anders6096604.cl"},
	{0x2C10DF6D14B1A122ULL, "vxw_encounter3_fs_0_0.vlm"},
	{0x2C1172F20F643A48ULL, "169453028.ogg"},
	{0x2C120D53B9BB0746ULL, "vxw_bonepitdestroyed_a3_1x0_0_1.vlm"},
	{0x2C1246E97429107DULL, "vxlt_0_lowtown_a2_n_f1j_0_0.vlm"},
	{0x2C12AB9DA362F5A4ULL, "tbe100ip_shop_codex_10.utp"},
	{0x2C12BA59DB5BDA22ULL, "vxht_keep_6yf_0_0.vlm"},
	{0x2C132695BF0BC705ULL, "6068362_m.fxe"},
	{0x2C137C0D0AB4F8E7ULL, "art\\vfx\\textures\\rockarmor\\vfx_rockarmor_0n.dds"},
	{0x2C13882A023E74BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v1.ani"},
	{0x2C13A594DBDF9790ULL, "hlac_lt_undercityserialkiller_-1_47y.dds"},
	{0x2C13B4361C1C0B82ULL, "vxht_keep_n_8b2_0_0.vlm"},
	{0x2C14213F785330A7ULL, "vxlt_gallowstemplar_n_37l_0_0.vlm"},
	{0x2C1435103B1F425BULL, "hlac_w_bonepit_262144_0z.dds"},
	{0x2C14450766720028ULL, "6078227_m.fxe"},
	{0x2C149412EC30CFB2ULL, "vxlt_0_lowtown_n_euz_0_0.vlm"},
	{0x2C14BA54C543D84BULL, "w_sundermount_28_0_lc.mmh"},
	{0x2C14C87B2F58F61BULL, "vxw_bonepit_15c_0_0.vlm"},
	{0x2C158FBACBB872A3ULL, "vxw_bonepitdestroyed_a3_265_0_0.vlm"},
	{0x2C15C77323CBEEEAULL, "llac_w_encounter1_0_1.dds"},
	{0x2C16090FE83BCF8CULL, "6218825_m.fxe"},
	{0x2C166CFEEB556442ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v2.ani"},
	{0x2C1679B2DD19196BULL, "hlac_lt_gamlenshouse_n_-1_8y.dds"},
	{0x2C169059A5D7FC95ULL, "6141441_m.fxe"},
	{0x2C169C420C67769FULL, "vxht_mansion2_2oa_0_0.vlm"},
	{0x2C16AECC9CD5656BULL, "mag170cr_recruit01.utc"},
	{0x2C170896A3D5A006ULL, "vxht_keep_685_0_0.vlm"},
	{0x2C17B199432615AFULL, "163354450.ogg"},
	{0x2C17ECEF58CC7E50ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_ornate.mao"},
	{0x2C17F0EF58CC8563ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_ornate.mmh"},
	{0x2C17F21C9AFCD90BULL, "hlac_lt_0_lowtown_a2_n_c_-65537_4z.dds"},
	{0x2C17FEEF58CC9CBDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_ornate.msh"},
	{0x2C188D028C661939ULL, "color_of_water.mao"},
	{0x2C189B5FDDB30554ULL, "317431850.ogg"},
	{0x2C18C02A706EB020ULL, "6211394_m.fxe"},
	{0x2C18C9F840E239A6ULL, "vxlt_shop_2t_0_0.vlm"},
	{0x2C19A29A62E21A17ULL, "120722409.ogg"},
	{0x2C19A681E20F216AULL, "zz_mag_debug6051787.cl"},
	{0x2C19CE8BC18131FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_overlay_ent.evt"},
	{0x2C1A14CDE9CF95D0ULL, "vxlt_blackemporium_kj_0_0.vlm"},
	{0x2C1A7A6F2C88AEACULL, "vxht_0_hightown_c86_0_0.vlm"},
	{0x2C1A98D5358FF110ULL, "vxlt_0_lowtown_docks_c6i_0_0.vlm"},
	{0x2C1AB329FB2F3C5FULL, "743162206.ogg"},
	{0x2C1B92029A2B6EA3ULL, "vxlt_0_lowtown_a2_f57_0_0.vlm"},
	{0x2C1BAFA73B8CEC33ULL, "6110935_m.fxe"},
	{0x2C1BB4D5AE6879C6ULL, "vxlt_0_lowtown_a3_n_c_eny_0_1.vlm"},
	{0x2C1C0C487ADC44E4ULL, "149743398.ogg"},
	{0x2C1C10647C3C53D6ULL, "166613456.ogg"},
	{0x2C1CCA544BAED6E8ULL, "98848724.ogg"},
	{0x2C1CF68E8AFCCFEEULL, "6110088_m.fxe"},
	{0x2C1CFABA1A661738ULL, "vxlt_gallowstemplar_n_4sp_0_0.vlm"},
	{0x2C1D04F36DA223E1ULL, "vxht_brothel_n_5ef_0_0.vlm"},
	{0x2C1D4E096D913E97ULL, "w_woundedcoast_3_0.rml"},
	{0x2C1D763714422210ULL, "6135762_m.fxe"},
	{0x2C1D80F5CB0EFC87ULL, "vxdr_primevaltomb_2xm_0_0.vlm"},
	{0x2C1D930967CEE138ULL, "hlac_lt_0_lowtown_a2_n_-196608_4y.dds"},
	{0x2C1E446FB20D7E49ULL, "qun230cr_guild_contact6068846.cl"},
	{0x2C1EE90E831C5BE9ULL, "6098960_f.fxe"},
	{0x2C1F3226D46B2496ULL, "vxlt_0_lowtown_a2_n_hb6_0_0.vlm"},
	{0x2C1F7420A746C75AULL, "701644279.ogg"},
	{0x2C1FAE6F66B32CB9ULL, "620271911.ogg"},
	{0x2C1FC6AA4B57B7EFULL, "one120tr_athe_ambush.utt"},
	{0x2C2001582F3ACE2AULL, "119580121.ogg"},
	{0x2C20AD77C6D81A22ULL, "6143821_f.fxe"},
	{0x2C20BAE898E7EE4FULL, "6037594_f.fxe"},
	{0x2C20CD8BC1874EEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_overlay_ent.gad"},
	{0x2C210C20600FC39CULL, "vxlt_hangedman_1qy_0_0.vlm"},
	{0x2C212C36CD51F2A2ULL, "vxht_keep_6bg_0_0.vlm"},
	{0x2C214A4B9E0BDB14ULL, "character_character_p126.psh.fxo"},
	{0x2C21546D511368F2ULL, "347657361.ogg"},
	{0x2C218C7FEA82C132ULL, "fxe_shadow.dds"},
	{0x2C21CDBEB8465CC3ULL, "one110_meeran6113250.cl"},
	{0x2C224EB1433CE70CULL, "vxw_encounter3_18_0_0.vlm"},
	{0x2C22989E5646623BULL, "vxw_bonepit_zq_0_0.vlm"},
	{0x2C23124F1B8953D9ULL, "vxw_bonepitdestroyed_a3_25r_0_0.vlm"},
	{0x2C240197BDAE8B23ULL, "vxw_bonepitdestroyed_a3_214_0_0.vlm"},
	{0x2C249A46603B1717ULL, "fx_f2c5.dds"},
	{0x2C24F0CEFC1DB8E2ULL, "vxdr_primevaltomb_6sm_0_0.vlm"},
	{0x2C254BF4DF34CF24ULL, "6089854_m.fxe"},
	{0x2C260BA5145C8B91ULL, "6116485_m.fxe"},
	{0x2C2716A7162B9494ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_setprops_sa.dds"},
	{0x2C27A0F3987B86AEULL, "vxht_mansion2_1t2_0_0.vlm"},
	{0x2C280030DD8AA4B7ULL, "6061778_m.fxe"},
	{0x2C2855251455A54DULL, "hlac_w_encounter2_327681_0x.dds"},
	{0x2C28C0B68807326FULL, "244520754.ogg"},
	{0x2C28DA83E9E9953DULL, "follower_banter6163257.cl"},
	{0x2C28F1EF58DAF988ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wi_ornate.phy"},
	{0x2C290E137FE2A244ULL, "mag321pt_bestservedcold.ncs"},
	{0x2C2944B9AB2FAC23ULL, "gen00fl_isabela6090866.cl"},
	{0x2C297EC4B570FE24ULL, "6156884_m.fxe"},
	{0x2C29983330B05F4AULL, "m_fen200ar_slavers_exterior.bnk"},
	{0x2C299DB4F6B22F8EULL, "gen00fl_fenris6127042.cl"},
	{0x2C299DBE823AD791ULL, "zz_areatranstest.ncc"},
	{0x2C2A0EFA2D98251DULL, "274388362.ogg"},
	{0x2C2A5C6E9EDBD1C2ULL, "mcr310_argument6075093.cl"},
	{0x2C2AF1EC8C9438BCULL, "173769601.ogg"},
	{0x2C2B76D05F983C61ULL, "vxht_brothel_n_5he_0_0.vlm"},
	{0x2C2B815DA770324CULL, "6036856_m.fxe"},
	{0x2C2B9DDDA1B8FC88ULL, "vxht_0_hightown_n_ai8_0_0.vlm"},
	{0x2C2C01AF0D5CC602ULL, "vxht_bartrand_n_4nq_0_0.vlm"},
	{0x2C2C6C7828FDDFBCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_3_ian.gad"},
	{0x2C2CA402FEDC8A25ULL, "1054662355.ogg"},
	{0x2C2CBAD81311EB36ULL, "lt_0_lowtown_docks_a3_221_0.msh"},
	{0x2C2CC4D81311FC34ULL, "lt_0_lowtown_docks_a3_221_0.mmh"},
	{0x2C2CC8D813120347ULL, "lt_0_lowtown_docks_a3_221_0.mao"},
	{0x2C2CE93F4D3B97E9ULL, "168790921.ogg"},
	{0x2C2D1962F628D21CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlighta.mao"},
	{0x2C2D2562F628E6BFULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlighta.mmh"},
	{0x2C2D264041C45FDAULL, "dae100_viscount6118513.cl"},
	{0x2C2D2B62F628F161ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlighta.msh"},
	{0x2C2E074D38ACB37AULL, "dae000tr_exit.ncs"},
	{0x2C2E27E93DD60877ULL, "6171390_m.fxe"},
	{0x2C2E3529741087C2ULL, "vfx_blt_fire_wall_p.bnk"},
	{0x2C2EBFB1B8A152B3ULL, "qun221_viscount6084693.cl"},
	{0x2C2F0407F35568D6ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastsand_n.dds"},
	{0x2C2F0C2249DB9DEBULL, "6034673_m.fxe"},
	{0x2C2F129B7DCD3BEFULL, "ht_0_hightown_a3_44_0.rml"},
	{0x2C2F3B2DFCE88CD1ULL, "vxlt_darktown_5rd_0_0.vlm"},
	{0x2C2F3F82ACB0DF89ULL, "160280134.ogg"},
	{0x2C2F43D768A61624ULL, "6110594_m.fxe"},
	{0x2C2FBB1E47029E76ULL, "6126395_m.fxe"},
	{0x2C2FD0C4675997FBULL, "6121001_m.fxe"},
	{0x2C30C68514106EFDULL, "vxca_cave_2kz_0_1.vlm"},
	{0x2C313C442F694D8EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2winbars.msh"},
	{0x2C314A442F6965DFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2winbars.mao"},
	{0x2C3156442F697A7CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2winbars.mmh"},
	{0x2C31845D7EC4D3DFULL, "vxht_0_hightown_aid_0_0.vlm"},
	{0x2C31DDAE32A8FC24ULL, "dae000_hate_enemies6167598.cl"},
	{0x2C321F9F979CE82BULL, "vxca_caverns1_mines_ijd_0_0.vlm"},
	{0x2C324E0E65B2267DULL, "one100dg_vanard6047412.cl"},
	{0x2C325833E6719989ULL, "zz_mer_debug6062703.cl"},
	{0x2C32839E034B0823ULL, "6154575_m.fxe"},
	{0x2C33201A4A97EFF4ULL, "follower_banter6151224.cl"},
	{0x2C332A9EF252A5DCULL, "hm_ave3_gen_guardm.mor"},
	{0x2C336B782903FCBAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_3_ian.ani"},
	{0x2C3372D94FA8887AULL, "vxht_0_hightown_n_c4s_0_0.vlm"},
	{0x2C337A2227907CFFULL, "vxlt_0_lowtown_a2_n_c_hoj_0_0.vlm"},
	{0x2C337FFBF053A79FULL, "vxht_mansion2_2uc_0_0.vlm"},
	{0x2C338AA1272B4265ULL, "dae000_gamlen6121554.cl"},
	{0x2C3463FCAA7E3D98ULL, "vxlt_0_lowtown_a2_ghr_0_0.vlm"},
	{0x2C355B91209A6028ULL, "6161077_m.fxe"},
	{0x2C356BC91E6A3D14ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxb.phy"},
	{0x2C35FB39C647B0B9ULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyo.phy"},
	{0x2C3603337F8CF27EULL, "6113280_f.fxe"},
	{0x2C3628D5E8D5F468ULL, "6139303_f.fxe"},
	{0x2C363D4E599B4729ULL, "zz_followers_banter6158745.cl"},
	{0x2C369B8FC1E6849AULL, "27151927.ogg"},
	{0x2C375E291074DFACULL, "vxht_0_hightown_c24_0_0.vlm"},
	{0x2C37BED281E8A9A9ULL, "127615104.ogg"},
	{0x2C37DC1159EB2A70ULL, "14250095.ogg"},
	{0x2C3815146027C66CULL, "415034283.ogg"},
	{0x2C3816BA7D7C6DFDULL, "hlac_lt_0_lowtown_docks_a3_131071_4y.dds"},
	{0x2C38268CDC4CB195ULL, "vxdr_0_deeproads_a29_0_1.vlm"},
	{0x2C38741B9D9B062DULL, "vxlt_warehouse_rp_0_0.vlm"},
	{0x2C39422985932D8CULL, "6108090_m.fxe"},
	{0x2C39799BC661E6D9ULL, "6139275_m.fxe"},
	{0x2C3ABBA54939257DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5_tool.phy"},
	{0x2C3B1FAF3EE9F7AEULL, "dae000_bartrand6030682.cl"},
	{0x2C3B2957294C12F0ULL, "dr_0_deeproads_probe_fyr.dds"},
	{0x2C3B38BB62138B35ULL, "284694996.ogg"},
	{0x2C3B4928B0025533ULL, "dae000cr_grand_cleric6181912.cl"},
	{0x2C3C649C2159172FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangbar_n.dds"},
	{0x2C3C6EE85CD31B68ULL, "atl_armycontro_dxt5_dat.xml"},
	{0x2C3C7D4F28B16CD5ULL, "gen_im_arm_cht_med_act0_01.uti"},
	{0x2C3CD9D757280600ULL, "hd_glow.mao"},
	{0x2C3CE92A65FA2304ULL, "vxw_blightlands_wb_0_0.vlm"},
	{0x2C3D5A6FFAB043ADULL, "193430832.ogg"},
	{0x2C3DA07AA2FEFA75ULL, "6070486_m.fxe"},
	{0x2C3DBA92F992FD72ULL, "vxlt_warehouse_13e_0_0.vlm"},
	{0x2C3DD144C31B53A7ULL, "6108317_f.fxe"},
	{0x2C3E0FF0B13E23A2ULL, "vxht_keep_6ha_0_0.vlm"},
	{0x2C3F48C91E725611ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxb.msh"},
	{0x2C3F56C91E726E2CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxb.mao"},
	{0x2C3F6146D6B49480ULL, "6116628_m.fxe"},
	{0x2C3F62C91E7282CFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxb.mmh"},
	{0x2C3FDB5A5185D2F1ULL, "zz_approval_debug6003554.cl"},
	{0x2C400E352D29060EULL, "vrc221_varric6040822.cl"},
	{0x2C40DFFC1A4A0F7AULL, "955951163.ogg"},
	{0x2C4192BD7D7D61FBULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_noba.phy"},
	{0x2C41D658209F2D0CULL, "vxht_mansion_n_61c_0_0.vlm"},
	{0x2C41E4B6419EA0FAULL, "vxht_0_hightown_n_c1l_0_0.vlm"},
	{0x2C425F70E6FA3DE8ULL, "mcr310_sibling6076392.cl"},
	{0x2C4283063AB1E125ULL, "lt_0_lowtown_a2_n_52_0_lc.mmh"},
	{0x2C42B003CB36A203ULL, "34574383.ogg"},
	{0x2C42D5D99C745618ULL, "vxlt_0_lowtown_a2_gos_0_0.vlm"},
	{0x2C43300C29199C0CULL, "59260196.ogg"},
	{0x2C434C7876643778ULL, "817691714.ogg"},
	{0x2C43515DEE2062A2ULL, "vxht_0_hightown_n_bgk_0_0.vlm"},
	{0x2C435AB44F1BD091ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv3.ani"},
	{0x2C436B5548C1D5FEULL, "vxlt_gallowscourtyard_n_7aq_0_0.vlm"},
	{0x2C43933385DEF1F7ULL, "461323617.ogg"},
	{0x2C43BA6CBC2328EBULL, "vxw_woundedcoast_77_0_0.vlm"},
	{0x2C43CD6E129078A8ULL, "6166177_m.fxe"},
	{0x2C43DA0115CA5EA6ULL, "art\\characters\\creatures\\gallowsslavestatue\\textures\\cn_bdy_css_0s.dds"},
	{0x2C440B8C91C3BE94ULL, "t3_shd_klr.tnt"},
	{0x2C4417569866501BULL, "mag131ip_loot_corpse_1.utp"},
	{0x2C441C56411BC7DFULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluebase.msh"},
	{0x2C442256411BD181ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluebase.mmh"},
	{0x2C442E56411BE622ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluebase.mao"},
	{0x2C4484AD7C67B7AEULL, "vxht_mansion2_1b4_0_0.vlm"},
	{0x2C448CE76367881BULL, "vxca_caverns1_e7f_0_0.vlm"},
	{0x2C4516AB337F2D9EULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack1_c.mmh"},
	{0x2C456A64BFB1DF50ULL, "vxlt_blackemporium_zi_0_0.vlm"},
	{0x2C45908C026D51F6ULL, "hlac_ht_brothel_n_3365_0x.dds"},
	{0x2C45A8954119B789ULL, "dae000_orsino6081375.cl"},
	{0x2C45B131C1A8369AULL, "vxlt_0_lowtown_a2_n_c_e18_0_0.vlm"},
	{0x2C45C0E4ADE5CB70ULL, "6104537_m.fxe"},
	{0x2C46361A4C2C0113ULL, "qun211_braggart6075718.cl"},
	{0x2C465E3539869B5DULL, "zz_fol_debug6096554.cl"},
	{0x2C46819DFAE04018ULL, "gen00fl_merrill6122212.cl"},
	{0x2C46A91371DCA490ULL, "dae000ip_crowd_lm1_talk1.utp"},
	{0x2C46E3463DD25EF1ULL, "hlac_lt_0_lowtown_a2_65538_5x.dds"},
	{0x2C46E5F88F911EADULL, "vxlt_warehouse_wo_0_0.vlm"},
	{0x2C475BA6AD45A3B0ULL, "llac_ca_caverns2_5802_7.dds"},
	{0x2C476691F5137340ULL, "mcr310_argument6074481.cl"},
	{0x2C47DC96B003F92BULL, "321334377.ogg"},
	{0x2C484D00904C6929ULL, "mcr370_cut_player_rules6098287.cl"},
	{0x2C489DD0E57AD942ULL, "vxht_keep_n_5js_0_0.vlm"},
	{0x2C48A86D8770BBCAULL, "397280483.ogg"},
	{0x2C48BFC7990CBFB2ULL, "6126669_m.fxe"},
	{0x2C48E15920396B1FULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.dura.gad"},
	{0x2C49951CECA08CEBULL, "268946868.ogg"},
	{0x2C4A32C71C94E4D8ULL, "hlac_lt_backalley_g_65537_0z.dds"},
	{0x2C4A65B44F220233ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv3.gad"},
	{0x2C4A6E024EC9952AULL, "837210546.ogg"},
	{0x2C4A834FA10D5BC1ULL, "vxw_encounter1_dx_0_0.vlm"},
	{0x2C4B89CABF682B76ULL, "351064878.ogg"},
	{0x2C4BA6DA22C9FB2BULL, "vxht_mansion2b_31v_0_0.vlm"},
	{0x2C4BB0A5494785E9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5_tool.mao"},
	{0x2C4BBCA549479982ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5_tool.mmh"},
	{0x2C4BBEA549479D38ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5_tool.msh"},
	{0x2C4C1EB6FAEFD88AULL, "vxht_brothel_n_4fa_0_0.vlm"},
	{0x2C4C3DB079D5C6D6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brrckbrd_d.dds"},
	{0x2C4C3ED639C60452ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_act_c.mmh"},
	{0x2C4C559BC983AF91ULL, "42551884.ogg"},
	{0x2C4D06EE9FC9AC3CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge02dirt_d.dds"},
	{0x2C4D0C97D52E8AB8ULL, "hlac_ht_0_hightown_-65537_4x.dds"},
	{0x2C4DAA90814DE245ULL, "192147951.ogg"},
	{0x2C4DEE352C28ABBBULL, "vxw_bonepit_kr_0_0.vlm"},
	{0x2C4E0824139A5276ULL, "6175481_m.fxe"},
	{0x2C4E0F21485FC6CDULL, "897556720.ogg"},
	{0x2C4E2B8FB2CE67ECULL, "315896253.ogg"},
	{0x2C4E374D1E146A55ULL, "772759969.ogg"},
	{0x2C4E8050F700108BULL, "mcr310_confused_guard6077058.cl"},
	{0x2C4E8C97A27F2987ULL, "vxlt_warehouse_1d5_0_0.vlm"},
	{0x2C4EE19C7D20121DULL, "t3_abm_b03.tnt"},
	{0x2C4EE690FC423CB6ULL, "vxlt_darktown_47i_0_0.vlm"},
	{0x2C4F147363061594ULL, "one110_gustav6089815.cl"},
	{0x2C4F2E33ED9885D6ULL, "vxlt_0_lowtown_f2o_0_0.vlm"},
	{0x2C4F7C439717CAE6ULL, "6018681_m.fxe"},
	{0x2C4F880ED37F95A8ULL, "6156570_m.fxe"},
	{0x2C4F8D96AC0857C0ULL, "abi_global.bnk"},
	{0x2C4F990C8DB942CAULL, "6119490_m.fxe"},
	{0x2C4FD659203F770CULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.dura.evt"},
	{0x2C500598C857BACCULL, "hm_cgn_002.mop"},
	{0x2C500598C857BACEULL, "hm_cgn_002.mor"},
	{0x2C5029AB04412E96ULL, "6141749_m.fxe"},
	{0x2C50CB3DCD7EC803ULL, "vxlt_undercitydungeon_17b_0_0.vlm"},
	{0x2C50DA29768877F0ULL, "6173401_m.fxe"},
	{0x2C50EB3A8E0E163AULL, "vxlt_0_lowtown_n_gcb_0_0.vlm"},
	{0x2C513488E404E2FDULL, "6175392_m.fxe"},
	{0x2C5147B68E6A6E98ULL, "vxlt_0_lowtown_a2_gbp_0_0.vlm"},
	{0x2C5162B44F281B98ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv3.evt"},
	{0x2C51C33AE0167B22ULL, "vxht_0_hightown_n_bbt_0_0.vlm"},
	{0x2C51DD323AB7EE7EULL, "vxlt_gallowscourtyard_n_7dr_0_0.vlm"},
	{0x2C526594DED8B551ULL, "lt_gallowstemplar_f_2389.rml"},
	{0x2C52E3EE5E681AEAULL, "vxw_encounter2_aa_0_0.vlm"},
	{0x2C52F5AB338B335DULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack1_c.anb"},
	{0x2C53D673DE409ED4ULL, "hlac_lt_0_lowtown_docks_-65538_5y.dds"},
	{0x2C53D9DAB488D6B7ULL, "dae000_leonato.bnk"},
	{0x2C5415CA1477D208ULL, "vxdr_primevalentrance_2y9_0_0.vlm"},
	{0x2C541F29743131E0ULL, "vfx_blt_fire_wall_p.wbd"},
	{0x2C5438A3248C1BF6ULL, "6027360_m.fxe"},
	{0x2C546F21E1969C77ULL, "vxw_bonepit_ze_0_0.vlm"},
	{0x2C54C4A0BB084126ULL, "vxca_cavern1_haunted_ihc_0_0.vlm"},
	{0x2C54E2871974ACAFULL, "vxca_caverns2prc_a8e_0_0.vlm"},
	{0x2C55328D5F1DFF3EULL, "6124691_f.fxe"},
	{0x2C5557D58187372DULL, "vxlt_warehouse_xn_0_0.vlm"},
	{0x2C557C345569499FULL, "6172607_m.fxe"},
	{0x2C558A6C1DC5B2F9ULL, "ave310_donnic6062090.cl"},
	{0x2C55991F32780447ULL, "961501517.ogg"},
	{0x2C5625E3697F5E8CULL, "6106290_m.fxe"},
	{0x2C564C966B2DBDAAULL, "vxlt_gallowstemplar_n_53k_0_0.vlm"},
	{0x2C567C11A9DE8114ULL, "core_rules.gda"},
	{0x2C56A2630DF5DBA8ULL, "llac_lt_0_lowtown_docks_65537_0.dds"},
	{0x2C56AD141695CEB7ULL, "plt_lgt312pt_shw_mer_base_2.plo"},
	{0x2C56DFBF9D28BC96ULL, "dae011_corff_sta6142382.cl"},
	{0x2C572163D8381EFBULL, "457204931.ogg"},
	{0x2C57501278ED158CULL, "6169424_m.fxe"},
	{0x2C57BBDAB48C8CCFULL, "dae000_leonato.cnv"},
	{0x2C57CF9E6DB82235ULL, "gen00fl_aveline6128546.cl"},
	{0x2C580538FD99D572ULL, "gen00fl_anders6116138.cl"},
	{0x2C593D7E4D699C27ULL, "452756251.ogg"},
	{0x2C5993CEC63E7AC7ULL, "45323594.ogg"},
	{0x2C5A40631AD97E6FULL, "vxht_mansion2_1ve_0_0.vlm"},
	{0x2C5A436C3CF28C6BULL, "6206270_m.fxe"},
	{0x2C5AB6DD1D0B1534ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.impa.ani"},
	{0x2C5AD3B413F50EE4ULL, "vxlt_gallowstemplar_356_0_0.vlm"},
	{0x2C5B035A28488287ULL, "vxht_playershouse_n_1vo_0_0.vlm"},
	{0x2C5B2C4ACD36C82BULL, "vxlt_gallowstemplar_6sv_0_0.vlm"},
	{0x2C5BACE868B7CF24ULL, "410817065.ogg"},
	{0x2C5BAF32616577FBULL, "vxlt_warehouse_hx_0_0.vlm"},
	{0x2C5BF50B3236A37DULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb105a_0n.dds"},
	{0x2C5C4F812453CF0EULL, "w_bonepit_149_0.msh"},
	{0x2C5C5D812453E75FULL, "w_bonepit_149_0.mao"},
	{0x2C5C69812453FBFCULL, "w_bonepit_149_0.mmh"},
	{0x2C5C72C58AFB0DB6ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_mag_0d.dds"},
	{0x2C5D156A542459D1ULL, "6095234_m.fxe"},
	{0x2C5D2077E277CA90ULL, "467388772.ogg"},
	{0x2C5D295C9F24E221ULL, "6037388_m.fxe"},
	{0x2C5D2BFDB2EA77FFULL, "lit_distor.mao"},
	{0x2C5D7981E6DDF3A1ULL, "dae000_bartrand6064053.cl"},
	{0x2C5D7AB8589C32CFULL, "6001022_m.fxe"},
	{0x2C5DBE59204B8C65ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.dura.ani"},
	{0x2C5E6A7D541D17C5ULL, "vxht_0_hightown_n_c4x_0_0.vlm"},
	{0x2C5E8E01192176ABULL, "6171516_m.fxe"},
	{0x2C5ED3811194740DULL, "gen_im_trs_drk_one_lsw_r2.uti"},
	{0x2C5ED4A9F594B9CEULL, "118740049.ogg"},
	{0x2C5EFCE42AACCBE9ULL, "vxht_keep_n_5gs_0_0.vlm"},
	{0x2C5F1FFF2DDC210EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodmetal_d.dds"},
	{0x2C5F9F525FAA06F9ULL, "qun221_petrice6077849.cl"},
	{0x2C5FD61BDE3BA923ULL, "art\\vfx\\combat\\material__30.mao"},
	{0x2C60015ED53D189CULL, "hlac_lt_undercityencounter_-1_26z.dds"},
	{0x2C6059D41559BEE8ULL, "vxw_sundermount_35a_0_0.vlm"},
	{0x2C609E09FD73BE9AULL, "6078049_m.fxe"},
	{0x2C60EB2A318929E0ULL, "149099028.ogg"},
	{0x2C6107B284CF49AAULL, "lt_0_lowtown_n_14_0.rml"},
	{0x2C6150A433B6CF18ULL, "6187936_m.fxe"},
	{0x2C6155CB505E336AULL, "vxw_encounter2_lb_0_0.vlm"},
	{0x2C6168676053E8AEULL, "vxht_mansion2_1h6_0_0.vlm"},
	{0x2C61A707B3AF85AAULL, "433829841.ogg"},
	{0x2C61B3DD1D112E96ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.impa.gad"},
	{0x2C61D33E0F4C277EULL, "vxca_caverns1_mines_j4d_0_0.vlm"},
	{0x2C623D9C400C7E06ULL, "314695402.ogg"},
	{0x2C6305C78D124C5AULL, "ogre_default.utc"},
	{0x2C63E21A539D2937ULL, "vxlt_hangedman_n_2w4_0_0.vlm"},
	{0x2C6463B2F04332B2ULL, "t3_crd_lt1.tnt"},
	{0x2C648D47E5CFE330ULL, "6173993_m.fxe"},
	{0x2C64BE735D23D62AULL, "vxlt_gallowstemplar_n_56d_0_0.vlm"},
	{0x2C64E98DBAF2ACE9ULL, "vxdr_slavers_1as_0_0.vlm"},
	{0x2C6521BAD4910AB0ULL, "hlac_w_sundermount_262141_0y.dds"},
	{0x2C65A1544DEAD600ULL, "hlac_lt_0_lowtown_n_65534_0y.dds"},
	{0x2C66551D11BD2C4EULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.impa.evt"},
	{0x2C667DCF1C827BD9ULL, "lgt210_eau_keran_father.bnk"},
	{0x2C673635FE132002ULL, "l_mag371ar_gallows_templar.wbd"},
	{0x2C674B1069B520DDULL, "6141755_m.fxe"},
	{0x2C678626835646DAULL, "171970403.ogg"},
	{0x2C6877AB3C106475ULL, "projdecal.mat.dx11"},
	{0x2C68CEDD1D177BF5ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.impa.evt"},
	{0x2C69317D6A8DDC57ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock4a.phy"},
	{0x2C69457119DE3E1CULL, "vxlt_hangedman_1hr_0_0.vlm"},
	{0x2C69B2749D225DD1ULL, "mag291_ntbanter6166154.cl"},
	{0x2C69BECF27A084E8ULL, "554050634.ogg"},
	{0x2C69D852F01AFAB8ULL, "gen_im_trs_gen_arc_sbw_r2.uti"},
	{0x2C6A7EB87882344EULL, "6018520_m.fxe"},
	{0x2C6A9FCF1C869E39ULL, "lgt210_eau_keran_father.cnv"},
	{0x2C6ABC1416A6912CULL, "plt_lgt312pt_shw_mer_base_2.nss"},
	{0x2C6AC05F0DD692C8ULL, "gen00fl_aveline6137552.cl"},
	{0x2C6AE315D2823C82ULL, "6062055_m.fxe"},
	{0x2C6B49DE0772E10CULL, "vxht_keep_n_6m3_0_0.vlm"},
	{0x2C6B7B23FF08B6C4ULL, "6147885_f.fxe"},
	{0x2C6BCA4AE02E6DB6ULL, "vxlt_darktown_49k_0_0.vlm"},
	{0x2C6BD817F9877226ULL, "6121857_m.fxe"},
	{0x2C6BF5258C39FAF2ULL, "vxlt_0_lowtown_n_ekf_0_2.vlm"},
	{0x2C6C4F5CCDB6D9CDULL, "gen00fl_aveline6101433.cl"},
	{0x2C6CD022125C080CULL, "hlac_lt_0_lowtown_docks_a3_0_2y.dds"},
	{0x2C6CDC5A46133045ULL, "vxht_0_hightown_n_c1g_0_0.vlm"},
	{0x2C6CDFF78EDCCD3EULL, "599268302.ogg"},
	{0x2C6D4C1D11C33BF9ULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.impa.gad"},
	{0x2C6D607AAE04CFF3ULL, "566201999.ogg"},
	{0x2C6D79BD7DA2719AULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_noba.msh"},
	{0x2C6D87BD7DA289D3ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_noba.mao"},
	{0x2C6D8BBD7DA290E0ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_noba.mmh"},
	{0x2C6D9D2A6FD653B9ULL, "255636161.ogg"},
	{0x2C6DA7B8F631966EULL, "283315307.ogg"},
	{0x2C6E3150B1EFED99ULL, "llac_lt_0_lowtown_a2_n_-4_0.dds"},
	{0x2C6E3D0D7CF3129FULL, "633300019.ogg"},
	{0x2C6F3AE2CBD5A21BULL, "dae300_alistair_king6134978.cl"},
	{0x2C6F4743D693AADAULL, "574761855.ogg"},
	{0x2C6F5863B289A5DEULL, "ave200_aveline_bar6177330.cl"},
	{0x2C6F81C69260F824ULL, "gen_im_acc_amu_act2_08.uti"},
	{0x2C6FD5BF7AD040AFULL, "mcr371_scared_mage_m6086905.cl"},
	{0x2C70235641410EFEULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluebase.phy"},
	{0x2C70363AB789514BULL, "m_rai361ar_cave.wbd"},
	{0x2C7043CB2BB74C4EULL, "6052830_m.fxe"},
	{0x2C708F562E568234ULL, "120892145.ogg"},
	{0x2C70BF5234CFD4AFULL, "820996856.ogg"},
	{0x2C70BFFB959428D0ULL, "vxlt_blackemporium_ut_0_0.vlm"},
	{0x2C70DEE05BA6FDEDULL, "vxlt_0_lowtown_a3_n_c_e16_0_0.vlm"},
	{0x2C715D9DC8209C49ULL, "566157960.ogg"},
	{0x2C71846D037582FAULL, "zz_mrl_debug6184209.cl"},
	{0x2C718AE7B2A4E7FDULL, "gen_im_trs_qun_lgt_boo_r0.uti"},
	{0x2C71AD413B98A81AULL, "hlac_w_bonepitmining_a2_0_11z.dds"},
	{0x2C7230CCD373D42EULL, "art\\characters\\weapons\\staff\\textures\\w_stf_005a_0s.dds"},
	{0x2C72F27F3A59FBFFULL, "350506774.ogg"},
	{0x2C732C8E1632F79CULL, "6118639_m.fxe"},
	{0x2C735D1D11C7C447ULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.impa.ani"},
	{0x2C73D72E5B877708ULL, "vxht_0_hightown_n_av5_0_0.vlm"},
	{0x2C740E1F1DAFF6AEULL, "6072249_m.fxe"},
	{0x2C742EB23D638CF2ULL, "bmm.80.dds"},
	{0x2C744A623776A2D0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicerchair_n.dds"},
	{0x2C74646231EF810CULL, "one000_ilen6037072.cl"},
	{0x2C74F2C4BD9A4D24ULL, "vxht_0_hightown_a3_bgc_0_0.vlm"},
	{0x2C750EE04DCC16D1ULL, "mag110_vincento6064379.cl"},
	{0x2C753DD639E907F9ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_act_c.anb"},
	{0x2C75CABF722CD7DFULL, "6169485_m.fxe"},
	{0x2C762371D90196C4ULL, "69319442.ogg"},
	{0x2C762EF4BE8A0C7BULL, "124115023.ogg"},
	{0x2C7630226A2A1D8FULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchhang_dp.dds"},
	{0x2C763C91450E8381ULL, "zz_ave_debug6056786.cl"},
	{0x2C764B2D49B473ECULL, "vxca_caverns1_mines_e8e_0_0.vlm"},
	{0x2C7655E87C401E16ULL, "vxht_0_hightown_a3_aul_0_0.vlm"},
	{0x2C766B0C167C62C0ULL, "6095099_m.fxe"},
	{0x2C769C04C4D3C06BULL, "vxca_caverns1_e50_0_0.vlm"},
	{0x2C76B6E6CB033E1EULL, "zz_vrc_debug6200190.cl"},
	{0x2C7728E8DBB3F096ULL, "418594915.ogg"},
	{0x2C77894227D92EAAULL, "character_character_p153.psh.fxo.dx11.assembly"},
	{0x2C77CF0043ED1745ULL, "547702163.ogg"},
	{0x2C7892EC4551A8FBULL, "vxlt_warehouse_vz_0_0.vlm"},
	{0x2C793633F605FF22ULL, "6053482_m.fxe"},
	{0x2C7940C11966A73FULL, "qcr221cr_qunari_mage.utc"},
	{0x2C79BBBAF968F98CULL, "vxht_keep_n_6h0_0_0.vlm"},
	{0x2C7A9837C76A625CULL, "579087660.ogg"},
	{0x2C7AF045F718E840ULL, "vxlt_warehouse_n_q9_0_0.vlm"},
	{0x2C7B5B9156B6143FULL, "168239266.ogg"},
	{0x2C7B8588D51EB582ULL, "6139248_f.fxe"},
	{0x2C7BF6237C22AA6AULL, "ave310_brennan6136881.cl"},
	{0x2C7C9164685B0CC4ULL, "507621940.ogg"},
	{0x2C7CBBF0ADF1F03AULL, "vxdr_primevaltomb_4zr_0_2.vlm"},
	{0x2C7D4A69929254DCULL, "297753722.ogg"},
	{0x2C7E52F516B7189CULL, "vxw_encounter2_camp_9g_0_0.vlm"},
	{0x2C7E7879AA36BC37ULL, "combat_ar_r1_a2.gda"},
	{0x2C7E8DE0A1DCAF89ULL, "vxht_mansion_n_6cl_0_0.vlm"},
	{0x2C7EB0864EB4304CULL, "mag320_orsino6071094.cl"},
	{0x2C7F50379B46D558ULL, "564041097.ogg"},
	{0x2C7F581C7065C226ULL, "abi_m_petrify.bnk"},
	{0x2C7FF21D32A3A140ULL, "vrc321_varric6087055.cl"},
	{0x2C8070FCD4AD45E4ULL, "vxca_caverns2_91t_0_0.vlm"},
	{0x2C80B1C5C8EDDDF2ULL, "gen_im_trs_cot_med_boo_r0.uti"},
	{0x2C80BEBE0509D645ULL, "712604805.ogg"},
	{0x2C80D17E7B09BADEULL, "vxca_caverns1_mines_dpf_0_0.vlm"},
	{0x2C80EFB98846F449ULL, "459724417.ogg"},
	{0x2C81031ABD0B39FFULL, "vxlt_hangedman_n_21f_0_0.vlm"},
	{0x2C810814AE344DACULL, "6042774_m.fxe"},
	{0x2C81E6FA82024C32ULL, "360948916.ogg"},
	{0x2C81ED9C8EDCDFAFULL, "203435213.ogg"},
	{0x2C8257B1FCF5C5D2ULL, "vxlt_darktown_4v2_0_0.vlm"},
	{0x2C826A379386DA80ULL, "218899249.ogg"},
	{0x2C827D76821E093BULL, "vxdr_primevaltomb_5am_0_0.vlm"},
	{0x2C834010A8461ADFULL, "vxdr_primevaltomb_77q_0_0.vlm"},
	{0x2C836E9815F9893BULL, "786014708.ogg"},
	{0x2C8380A33B9FE12EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_enter.evt"},
	{0x2C8469BDF842DFFAULL, "937380134.ogg"},
	{0x2C84A1693C856AAEULL, "llac_lt_undercityencounter_-1_46.dds"},
	{0x2C84A7D01ECA24E0ULL, "gen_im_trs_krk_hgh_junk3_02.uti"},
	{0x2C84C7C56E363696ULL, "vxht_0_hightown_a3_arm_0_0.vlm"},
	{0x2C850366FCF4E4A4ULL, "vxht_keep_55o_0_0.vlm"},
	{0x2C8572D4E10A5423ULL, "6049462_m.fxe"},
	{0x2C8595F9F0BBC7EFULL, "vxht_mansion2_1gf_0_0.vlm"},
	{0x2C8700F715E0CCF7ULL, "vxlt_0_lowtown_n_f5h_0_0.vlm"},
	{0x2C87368845A4FDA3ULL, "vxca_caverns2_5xl_0_0.vlm"},
	{0x2C87AD95E3C5D5D5ULL, "vxlt_0_lowtown_a2_n_fij_0_0.vlm"},
	{0x2C883C472D2046C5ULL, "757459596.ogg"},
	{0x2C8848625910F381ULL, "gen00fl_aveline6139496.cl"},
	{0x2C88509EABE90AC4ULL, "seb201_lady_harimann6177390.cl"},
	{0x2C8894FAEC969FA3ULL, "hightown_male_noble_gossip6171403.cl"},
	{0x2C88C0BA82DA6683ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_combo_breaker.evt"},
	{0x2C8918B180787E51ULL, "127211896.ogg"},
	{0x2C8A77A33BA5F0D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_enter.gad"},
	{0x2C8A796B0BF09BD0ULL, "774170011.ogg"},
	{0x2C8A866C3FDCAE2CULL, "gen_im_arm_cht_nobb_t4.uti"},
	{0x2C8A9668AD98B847ULL, "llac_lt_0_lowtown_a3_n_c_-65536_4.dds"},
	{0x2C8AC25F97636A4DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_posterope.mmh"},
	{0x2C8AC65F9763709EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_posterope.mao"},
	{0x2C8AD45F976388D3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_posterope.msh"},
	{0x2C8ADB14A051C5BAULL, "hla_ht_0_hightown_a2_65538_1z.dds"},
	{0x2C8AE56C3FB43F4AULL, "llac_lt_0_lowtown_docks_-196607_0.dds"},
	{0x2C8B02FC6F18E777ULL, "209351605.ogg"},
	{0x2C8C418FA418300FULL, "vxdr_0_deeproads_ese_0_0.vlm"},
	{0x2C8C73582A39A182ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedenter.gad"},
	{0x2C8C75701F5B511EULL, "lgt160ip_wk_soph_fake_soph.utp"},
	{0x2C8C7E54FE5960F2ULL, "800649429.ogg"},
	{0x2C8C926A8DB86732ULL, "isa211_isabela6202104.cl"},
	{0x2C8D68178F33D264ULL, "mag311_emile_de_launcet6061696.cl"},
	{0x2C8DA131C7DCC0ADULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgtb.mmh"},
	{0x2C8DA531C7DCC7FEULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgtb.mao"},
	{0x2C8DB331C7DCDF33ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgtb.msh"},
	{0x2C8E13F1D3CA6A4CULL, "lt_0_lowtown_docks_n_119_0.msh"},
	{0x2C8E19F1D3CA743EULL, "lt_0_lowtown_docks_n_119_0.mmh"},
	{0x2C8E25F1D3CA88D5ULL, "lt_0_lowtown_docks_n_119_0.mao"},
	{0x2C8E502E532B50EDULL, "vxht_keep_n_36b_0_0.vlm"},
	{0x2C8E89EE2802425FULL, "6034022_m.fxe"},
	{0x2C8F2E62F67BC004ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlighta.phy"},
	{0x2C8F37B1297F72B7ULL, "vxlt_hangedman_n_2d1_0_0.vlm"},
	{0x2C8F3F3CA39397AFULL, "6152233_f.fxe"},
	{0x2C8F3FF48F4C7EF8ULL, "vxlt_0_lowtown_docks_n_cwi_0_0.vlm"},
	{0x2C8F99BA82E04304ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_combo_breaker.gad"},
	{0x2C8FEA457F76DED6ULL, "vxw_encounter2_camp_77_0_0.vlm"},
	{0x2C901EF0CC8AE339ULL, "mag321_comtesse_de_launcet6064902.cl"},
	{0x2C90A109F7F06A74ULL, "6075750_m.fxe"},
	{0x2C90D86D0FD5B4BAULL, "vxht_brothel_n_2bs_0_0.vlm"},
	{0x2C9111573A6C4538ULL, "hightown_fem_elfservant_mob6169047.cl"},
	{0x2C9147CA41B5EE55ULL, "688301711.ogg"},
	{0x2C915687A422CB38ULL, "6043698_m.fxe"},
	{0x2C9188A33BAC2C27ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_enter.ani"},
	{0x2C91A11F27A46151ULL, "lgt110_urn_barker.bnk"},
	{0x2C929FA15C000085ULL, "21400848.ogg"},
	{0x2C92AC576270FD2CULL, "1063878191.ogg"},
	{0x2C9301128591CADBULL, "hightown_male_noble2_mob6170890.cl"},
	{0x2C9396582A3FFC79ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedenter.evt"},
	{0x2C93E948E677163CULL, "6172903_m.fxe"},
	{0x2C947C2D02EA0CFBULL, "gen_im_acc_amu_act1_02.uti"},
	{0x2C94831F27A66461ULL, "lgt110_urn_barker.cnv"},
	{0x2C948B7B07B4FC3BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_6craggyrock.msh"},
	{0x2C949D7B07B51A86ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_6craggyrock.mao"},
	{0x2C94A97B07B52F25ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_6craggyrock.mmh"},
	{0x2C94BB63886E07DDULL, "dae100_viscount6103310.cl"},
	{0x2C95E07AE3B6D488ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_basic.msh"},
	{0x2C95E4695216C6DDULL, "default_black.dds"},
	{0x2C95F27AE3B6F319ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_basic.mao"},
	{0x2C95FE7AE3B707B2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_basic.mmh"},
	{0x2C95FFF14BFDF486ULL, "6149507_m.fxe"},
	{0x2C960E06CAF48AABULL, "sib131_slave_a6165266.cl"},
	{0x2C96183AB7A9ED19ULL, "m_rai361ar_cave.bnk"},
	{0x2C9697EB33F14DD0ULL, "w_bonepit_45_0.rml"},
	{0x2C96A8BA82E67BE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_combo_breaker.ani"},
	{0x2C96DDB3AE69BE17ULL, "llac_w_encounter2_camp_131073_0.dds"},
	{0x2C96F5FB1928014BULL, "llac_ca_caverns2_5802_13.dds"},
	{0x2C97272AF61B8EE7ULL, "435654040.ogg"},
	{0x2C98897621C80B13ULL, "6118427_m.fxe"},
	{0x2C98A2BB2877D8BFULL, "hi_dust_02.mao"},
	{0x2C98E449BBD3D8A8ULL, "290742455.ogg"},
	{0x2C99D5F7512730ACULL, "gen00fl_anders6119195.cl"},
	{0x2C9A40225DA2268CULL, "hlac_ca_caverns1_mines_24423_5x.dds"},
	{0x2C9A68802AEE50B0ULL, "6106697_f.fxe"},
	{0x2C9A693527058316ULL, "gen00fl_isabela_init6110265.cl"},
	{0x2C9AA04B6C1A528EULL, "one300it_fire.uti"},
	{0x2C9ABCDFFAD52E87ULL, "lgt120_orz_bln_official6170516.cl"},
	{0x2C9AE4521EBABA20ULL, "hlac_ht_playershouse_n_855_1z.dds"},
	{0x2C9B523885618AF0ULL, "gen00fl_merrill6106816.cl"},
	{0x2C9B719A85C8E089ULL, "vxht_mansion_n_6ir_0_0.vlm"},
	{0x2C9B79F6838AE289ULL, "isa211_isabela6202206.cl"},
	{0x2C9B8BFFFCEB624AULL, "vxlt_hangedman_n_10w_0_0.vlm"},
	{0x2C9BBB5F9771D10AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_posterope.phy"},
	{0x2C9BF336AAEE79E0ULL, "llac_lt_gamlenshouse_n_-1_0.dds"},
	{0x2C9C795B1FFCB27FULL, "vxht_bartrand_n_4fx_0_0.vlm"},
	{0x2C9C7F49ECA4BC0EULL, "vxw_sundermount_34d_0_0.vlm"},
	{0x2C9C9019F26437F3ULL, "6080708_m.fxe"},
	{0x2C9D373453A8E3D3ULL, "hlac_w_encounter1_-65536_0z.dds"},
	{0x2C9D401492D12D28ULL, "6101422_m.fxe"},
	{0x2C9D53A894701D4AULL, "6110829_f.fxe"},
	{0x2C9D58B00D9C04EDULL, "character_character_p226.psh.fxo.dx11"},
	{0x2C9E4B9913FCFD4FULL, "plt_lgt310pt_fixer_3.plo"},
	{0x2C9E5AFEF3211A42ULL, "w_sundermount_168_0.msh"},
	{0x2C9E5C22716CF756ULL, "vxw_encounter2_camp_28_0_0.vlm"},
	{0x2C9E5CFEF3211DF8ULL, "w_sundermount_168_0.mmh"},
	{0x2C9E68FEF321319BULL, "w_sundermount_168_0.mao"},
	{0x2C9E6B29FD5DE8F4ULL, "351736467.ogg"},
	{0x2C9E988AB5797C5BULL, "zz_otl_debug6060953.cl"},
	{0x2C9E9A31C7EB286AULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgtb.phy"},
	{0x2C9EF0F4DCE29AA7ULL, "zz_approval_debug6003596.cl"},
	{0x2C9FE82AEE7D8EA6ULL, "vxlt_gallowscourtyard_7av_0_0.vlm"},
	{0x2CA022F8D2FCC56FULL, "mag311_keran6041477.cl"},
	{0x2CA0546194345D8FULL, "873127489.ogg"},
	{0x2CA102D14E28963FULL, "vxca_caverns1_mines_e8j_0_0.vlm"},
	{0x2CA111749E07C3CBULL, "vxlt_0_lowtown_fiw_0_0.vlm"},
	{0x2CA114F181410FFFULL, "fen110_party_comment6095072.cl"},
	{0x2CA1306E8907DBDDULL, "hlac_lt_0_lowtown_docks_1_0z.dds"},
	{0x2CA19E582A4C4768ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedselectedenter.ani"},
	{0x2CA2CE0ACA9D5801ULL, "6156759_m.fxe"},
	{0x2CA323B6F1DA1456ULL, "488902685.ogg"},
	{0x2CA3599A625F4FCCULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb04a.phy"},
	{0x2CA3D9AD30DE956EULL, "vxlt_warehouse_n_121_0_0.vlm"},
	{0x2CA3E6892B0DB89BULL, "vxw_bonepitmining_a2_256_0_0.vlm"},
	{0x2CA3F2E0D1C85DC1ULL, "vxw_bonepitdestroyed_a3_26o_0_0.vlm"},
	{0x2CA400C547B9329CULL, "vxca_caverns1_j6e_0_0.vlm"},
	{0x2CA42EB7F9FD386EULL, "855169082.ogg"},
	{0x2CA4585B20911C88ULL, "363466284.ogg"},
	{0x2CA4A45A40323C42ULL, "vxw_bonepit_27c_0_0.vlm"},
	{0x2CA55446B65733AEULL, "vxht_bartrand_n_5sw_0_0.vlm"},
	{0x2CA58FBF555EE4ABULL, "w_bonepit_20_0_ll.mmh"},
	{0x2CA5ACC932193500ULL, "dae000_madam_lusine6154909.cl"},
	{0x2CA5F162545A29DDULL, "424357811.ogg"},
	{0x2CA5FB065E5177ABULL, "6034044_m.fxe"},
	{0x2CA673508D447C68ULL, "864969750.ogg"},
	{0x2CA6FB5313BE2393ULL, "6117527_m.fxe"},
	{0x2CA6FC8C33C26E03ULL, "dae000ip_crowd_mf6_talk5.utp"},
	{0x2CA746ABD4DBAE0BULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge2.ani"},
	{0x2CA749C90ECC7DD4ULL, "126463837.ogg"},
	{0x2CA79D536DDFCAACULL, "6149116_m.fxe"},
	{0x2CA83126458A10C4ULL, "isa311_isabela6092515.cl"},
	{0x2CA84D1B9AF7C64BULL, "hlac_lt_gallowscourtyard_n_131071_2y.dds"},
	{0x2CA863FEF3297D83ULL, "w_sundermount_168_0.phy"},
	{0x2CA8E59F5F67A19AULL, "6039150_m.fxe"},
	{0x2CA913A8743A5FE6ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_goon2.ani"},
	{0x2CA932CD93F11071ULL, "vxlt_0_lowtown_a3_n_c_gjm_0_0.vlm"},
	{0x2CA95677E94ABDB4ULL, "character_character_p74.psh.fxo.dx11.assembly"},
	{0x2CA963A24EE0779AULL, "lt_warehouse_4_0.anb"},
	{0x2CA977474F694B01ULL, "6100114_m.fxe"},
	{0x2CA99E7C9CC8A549ULL, "499587009.ogg"},
	{0x2CA9B6730C6300EEULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_up_c.impa.ani"},
	{0x2CA9E37777BEF909ULL, "vxht_mansion_n_6lq_0_0.vlm"},
	{0x2CAA10CDB12D1C12ULL, "vxht_bartrand_n_627_0_0.vlm"},
	{0x2CAAB7CE7FB0821FULL, "vxdr_primevaltomb_861_0_0.vlm"},
	{0x2CAAE0DA4155D53DULL, "6100104_f.fxe"},
	{0x2CAAF126DE9AD48EULL, "vxw_sundermount_33e_0_0.vlm"},
	{0x2CAB50ACB78D5F2EULL, "hlac_w_encounter3_-65535_0x.dds"},
	{0x2CABF0F77520EE58ULL, "vxca_cavern1_haunted_fm5_0_0.vlm"},
	{0x2CABFE6BE1CA46C8ULL, "vxht_keep_n_42t_0_0.vlm"},
	{0x2CAC0EE975542A96ULL, "6139262_f.fxe"},
	{0x2CAC1EDC57866544ULL, "and111_lirene6084543.cl"},
	{0x2CAC4659F1E6797AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v1.gad"},
	{0x2CAD18FAA52C7491ULL, "gen00fl_anders6119318.cl"},
	{0x2CAD436D3FE2620EULL, "hlac_lt_gallowstemplar_f_2382_0x.dds"},
	{0x2CADE16B47D265F3ULL, "hlac_lt_0_lowtown_a2_0_8z.dds"},
	{0x2CAE61E6E2A2899AULL, "vxht_chantry_n_2ia_0_0.vlm"},
	{0x2CAF1E2B13F95189ULL, "gen_im_trs_dem_junk2_04.uti"},
	{0x2CAF4435FE50EDB8ULL, "l_mag371ar_gallows_templar.bnk"},
	{0x2CAFE0038FADD21FULL, "vis121_saemus6113520.cl"},
	{0x2CB06EA6ED722D8EULL, "vxw_encounter2_ps_0_1.vlm"},
	{0x2CB0C0DE000140E3ULL, "zz_mag_debug6053427.cl"},
	{0x2CB0F3A194551E9BULL, "follower_banter6160742.cl"},
	{0x2CB1067D6ACB48AFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock4a.mao"},
	{0x2CB107A2C47F1F33ULL, "plt_vrc000pt_main.nss"},
	{0x2CB1127D6ACB5D4CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock4a.mmh"},
	{0x2CB1187D6ACB673EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock4a.msh"},
	{0x2CB181414F2828B1ULL, "6149381_f.fxe"},
	{0x2CB19E2DE5E2A976ULL, "mag101_grace6034375.cl"},
	{0x2CB201B7AEE53C71ULL, "652676715.ogg"},
	{0x2CB24959F1EAE991ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v1.evt"},
	{0x2CB24E68A16ED1ECULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinswall.msh"},
	{0x2CB25468A16EDBDEULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinswall.mmh"},
	{0x2CB26068A16EF075ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinswall.mao"},
	{0x2CB264BDC3BE7641ULL, "vxw_bonepitdestroyed_a3_21n_0_0.vlm"},
	{0x2CB272A239AF4B1CULL, "vxca_caverns1_j3z_0_0.vlm"},
	{0x2CB27B0F062B7D21ULL, "6169641_m.fxe"},
	{0x2CB31102ADEAD2C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_left.ani"},
	{0x2CB3999D757290FDULL, "982197126.ogg"},
	{0x2CB3A0889FFF261AULL, "269781706.ogg"},
	{0x2CB3A49940044A9EULL, "vxdr_primevaltomb_7u7_0_1.vlm"},
	{0x2CB3AF31135EC70EULL, "397313637.ogg"},
	{0x2CB454361C44C138ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_basic.phy"},
	{0x2CB4779BF79DD61BULL, "art\\characters\\playercharacter\\humanfemale\\hlm_medj.msh"},
	{0x2CB4899BF79DF566ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_medj.mao"},
	{0x2CB48A8549A392BEULL, "6165681_f.fxe"},
	{0x2CB4959BF79E0905ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_medj.mmh"},
	{0x2CB502D9483C3617ULL, "vxdr_primevaltomb_6j2_0_1.vlm"},
	{0x2CB5154096350CCBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\cbt_grenade_toss.evt"},
	{0x2CB5741281A4C772ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_107a_0t.dds"},
	{0x2CB59815B078F8D6ULL, "vxca_caverns1_j49_0_0.vlm"},
	{0x2CB5FF1B70780885ULL, "hlac_lt_0_lowtown_docks_-4_3y.dds"},
	{0x2CB604A8237CB7CCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medp_0d.dds"},
	{0x2CB63B2203C64AFDULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml106a_0n.dds"},
	{0x2CB65438E14A762EULL, "llac_ca_caverns2prc_5802_55.dds"},
	{0x2CB658EC5FE5141CULL, "lt_0_lowtown_a2_22_0.rml"},
	{0x2CB65AF7F4167985ULL, "hlac_dr_primevaltomb_663_1y.dds"},
	{0x2CB66DFA718C26A3ULL, "zz_rdr_debug6066635.cl"},
	{0x2CB67757C9478D34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_loop.gad"},
	{0x2CB691BE008B97ABULL, "944985914.ogg"},
	{0x2CB69638FC0EBD0DULL, "w_woundedcoast_200_0.mmh"},
	{0x2CB69A38FC0EC45EULL, "w_woundedcoast_200_0.mao"},
	{0x2CB6A3A495F495F5ULL, "vrc321_party_barks6095773.cl"},
	{0x2CB6A838FC0EDB93ULL, "w_woundedcoast_200_0.msh"},
	{0x2CB6AE730C6D7D3FULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_up_c.impa.evt"},
	{0x2CB6B67B0BE4607CULL, "gen_im_wep_war_two_act2_05.uti"},
	{0x2CB6D6390460A3A4ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_player_arishokaddress.ani"},
	{0x2CB7173E7E1A986CULL, "hlac_lt_0_lowtown_docks_131071_5z.dds"},
	{0x2CB74CA91B7EA699ULL, "vxca_caverns2_7mc_0_0.vlm"},
	{0x2CB7A71B8D5D1E37ULL, "w_woundedcoast_164_0.phy"},
	{0x2CB8A93C09F82D4BULL, "vis121_saemus6086666.cl"},
	{0x2CB922E24ED0D43AULL, "hlac_w_encounter2_camp_327684_0z.dds"},
	{0x2CB940195158FA9AULL, "vxlt_0_lowtown_a2_n_c_ei0_0_0.vlm"},
	{0x2CB94C643BA42C4DULL, "fen210_dying_magister6108595.cl"},
	{0x2CB95500F0F04B7DULL, "904489735.ogg"},
	{0x2CB9B01CB9BF22B8ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_0z.dds"},
	{0x2CBABEF328C01621ULL, "6133589_m.fxe"},
	{0x2CBB3F9118A912D6ULL, "1022787771.ogg"},
	{0x2CBB9F3401D7EE8CULL, "rdr111dg_raider_ambusher.cnv"},
	{0x2CBC146D3A0851CDULL, "vxw_bonepitmining_a2_ep_0_0.vlm"},
	{0x2CBC2DABD4EDDF95ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge2.gad"},
	{0x2CBC2E40963B558CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\cbt_grenade_toss.gad"},
	{0x2CBC42CE4BEC1FF7ULL, "vxlt_warehouse_n_102_0_0.vlm"},
	{0x2CBC8657C94C1212ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_loop.ani"},
	{0x2CBCC281DD9F7105ULL, "6035148_m.fxe"},
	{0x2CBCD33904650A06ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_player_arishokaddress.gad"},
	{0x2CBD001E137D1C8CULL, "6182449_m.fxe"},
	{0x2CBD2E5D85F3668BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\vulnerableloop.ani"},
	{0x2CBD5C0B2352C543ULL, "vxht_0_hightown_b8d_0_0.vlm"},
	{0x2CBD7B0F10FDBA2EULL, "6118310_m.fxe"},
	{0x2CBD88D74D10AB46ULL, "6035944_f.fxe"},
	{0x2CBD8F730C736728ULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_up_c.impa.gad"},
	{0x2CBDB86120AAFC4AULL, "872747610.ogg"},
	{0x2CBDC7C42ABFA9B6ULL, "vxlt_warehouse_n_ci_0_0.vlm"},
	{0x2CBDEFDF7EB50248ULL, "vfx_blt_forcefield_p.bnk"},
	{0x2CBDFCA8744C94E0ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_goon2.gad"},
	{0x2CBE6D78FF9C3DCBULL, "art\\vfx\\combat\\vfx_r_shd_decoy_act_c.mmh"},
	{0x2CBF5A1C709BA9CCULL, "abi_m_petrify.wbd"},
	{0x2CBF813401DBA4BCULL, "rdr111dg_raider_ambusher.bnk"},
	{0x2CC00D8792A74C91ULL, "6125149_f.fxe"},
	{0x2CC06159F1F75060ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v1.ani"},
	{0x2CC070D4BD4E1CCFULL, "llac_w_encounter3_-262148_0.dds"},
	{0x2CC0C26FB07D5EEDULL, "vxlt_warehouse_n_13y_0_0.vlm"},
	{0x2CC0C6231388A2F8ULL, "gen00fl_carver6104856.cl"},
	{0x2CC0EB12AD8D80FBULL, "6174146_m.fxe"},
	{0x2CC12902ADF738BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_left.evt"},
	{0x2CC1EC5C67124AA0ULL, "vxdr_primevaltomb_7zu_0_0.vlm"},
	{0x2CC215E38AD77FC6ULL, "framebuffereffects_radialblur_p0.vsh.fxo.dx11"},
	{0x2CC22D40963FBF8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\cbt_grenade_toss.ani"},
	{0x2CC244554B21DFD7ULL, "645025351.ogg"},
	{0x2CC24E3520113629ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02a.phy"},
	{0x2CC3564D5F090789ULL, "6175868_m.fxe"},
	{0x2CC374B63A324E97ULL, "vxdr_primevaltomb_6q3_0_1.vlm"},
	{0x2CC3ADCA196076EAULL, "vxht_brothel_n_6cl_0_0.vlm"},
	{0x2CC4719900F2F73BULL, "art\\vfx\\combat\\vfx_gib_glv_lgta.mao"},
	{0x2CC540807441C86AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\pose\\pose_grovel_floor.gad"},
	{0x2CC584E865C50F33ULL, "vrc321ar_mansion_haunted.ncs"},
	{0x2CC62B88F2CDF075ULL, "w_bonepitdestroyed_a3_136_0.phy"},
	{0x2CC6F48795194D12ULL, "vxht_bartrand_n_681_0_0.vlm"},
	{0x2CC72802ADFBA2ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_left.gad"},
	{0x2CC73BD8ABE0A2B9ULL, "6098836_f.fxe"},
	{0x2CC74F9E706B8022ULL, "fen110_fenris6118122.cl"},
	{0x2CC773B008CCAC54ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.mmh"},
	{0x2CC774013CF5B0E1ULL, "33364802.ogg"},
	{0x2CC7848C51A8BDB2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.inactivetoactive.gad"},
	{0x2CC78F38FC1D24CAULL, "w_woundedcoast_200_0.phy"},
	{0x2CC7B1F6434F131AULL, "vxlt_0_lowtown_a2_n_c_en1_0_0.vlm"},
	{0x2CC7B4C93348CF1FULL, "vxht_keep_5wv_0_0.vlm"},
	{0x2CC7CC9293E1CADAULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_twitch_hands_on_hips.gad"},
	{0x2CC85F1708DEEDC9ULL, "vxlt_undercityencounter_13q_0_0.vlm"},
	{0x2CC9754A775A4769ULL, "vxw_bonepit_25b_0_0.vlm"},
	{0x2CC998941FC53F22ULL, "6076075_f.fxe"},
	{0x2CC9F9C8C1CACFAAULL, "zz_and_debug6178606.cl"},
	{0x2CC9FBBC663C80EFULL, "795832384.ogg"},
	{0x2CCA47C076C99C4DULL, "m_dae320ar_hightown.wbd"},
	{0x2CCA619B6D0068AFULL, "vxw_woundedcoast_1fi_0_0.vlm"},
	{0x2CCA727E828AAEAAULL, "dae000_bodhan6146868.cl"},
	{0x2CCAB6CCF6073CEDULL, "586056232.ogg"},
	{0x2CCB790D88FE79E3ULL, "vxca_caverns2prc_78v_0_1.vlm"},
	{0x2CCBE0FC2EF8FFC0ULL, "187908863.ogg"},
	{0x2CCBEC13E0E74423ULL, "vxca_caverns1_mines_edj_0_0.vlm"},
	{0x2CCBF59453668BB6ULL, "vxdr_primevaltomb_2wp_0_1.vlm"},
	{0x2CCC4E78FFA84630ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_act_c.anb"},
	{0x2CCC8D13210CFC3CULL, "vxlt_0_lowtown_a3_n_c_f4s_0_0.vlm"},
	{0x2CCCA2E1BB60CE84ULL, "6113238_m.fxe"},
	{0x2CCCA4E865CB643AULL, "vrc321ar_mansion_haunted.lst"},
	{0x2CCCB1ACF5645506ULL, "vxlt_warehouse_cl_0_0.vlm"},
	{0x2CCCB9C11A400D18ULL, "vxlt_gallowsprison_n_6lo_0_0.vlm"},
	{0x2CCD61FE6AF6CE8AULL, "vxlt_undercityserialkiller_1ln_0_1.vlm"},
	{0x2CCDFC3D28F6D0D8ULL, "gen00fl_carver6102568.cl"},
	{0x2CCE583F32184C7DULL, "224957969.ogg"},
	{0x2CCE878C51AEE1E9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.inactivetoactive.evt"},
	{0x2CCE9C6B7F7C5E65ULL, "gen_im_flw_two_carver_03.uti"},
	{0x2CCEC6061DE3FC23ULL, "vxdr_primevalentrance_238_0_0.vlm"},
	{0x2CCEEB22BB3ADC10ULL, "hlac_w_encounter3_0_7x.dds"},
	{0x2CCEF0F1D401C5D9ULL, "lt_0_lowtown_docks_n_119_0.phy"},
	{0x2CCF3A2BA7567A7AULL, "6199750_f.fxe"},
	{0x2CCF675DA4C9FA8BULL, "6119412_f.fxe"},
	{0x2CCF94467F1F2FF5ULL, "74631307.ogg"},
	{0x2CCFF176733D0972ULL, "245583645.ogg"},
	{0x2CD0087D720D3FDEULL, "follower_banter6163515.cl"},
	{0x2CD0097CD0A8AEE1ULL, "vxw_woundedcoast_fz_0_0.vlm"},
	{0x2CD01569D4656E3FULL, "mag101tr_grace_cower.utt"},
	{0x2CD04EF8CF8C5F61ULL, "vfx_temp2.mao"},
	{0x2CD05C4B35B00E32ULL, "vxlt_0_lowtown_n_e2o_0_0.vlm"},
	{0x2CD05E3959086320ULL, "vxdr_primevaltomb_7wr_0_0.vlm"},
	{0x2CD08BDDA0DB5C2FULL, "hlac_ht_bartrand_n_3367_1z.dds"},
	{0x2CD0BD4FF6B5378AULL, "vxdr_primevaltomb_34y_0_0.vlm"},
	{0x2CD0DE3E5BF118E3ULL, "vxlt_0_lowtown_h1u_0_0.vlm"},
	{0x2CD15B67507BA96BULL, "6206021_m.fxe"},
	{0x2CD18CBCCA4CC8B1ULL, "534594267.ogg"},
	{0x2CD201542472BF75ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_smoke02.mao"},
	{0x2CD20797BAD7DD7CULL, "hlac_lt_0_lowtown_a3_n_c_65535_1z.dds"},
	{0x2CD2155D86059815ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\vulnerableloop.gad"},
	{0x2CD23D1AD2A908BBULL, "lt_0_lowtown_docks_249_0.msh"},
	{0x2CD2429EF010DCACULL, "vxlt_0_lowtown_n_e1x_0_0.vlm"},
	{0x2CD24F1AD2A92706ULL, "lt_0_lowtown_docks_249_0.mao"},
	{0x2CD25B1AD2A93BA5ULL, "lt_0_lowtown_docks_249_0.mmh"},
	{0x2CD26955BD5662FBULL, "vxlt_blackemporium_15z_0_0.vlm"},
	{0x2CD2A0F388D32D76ULL, "zz_mcr_debug6088429.cl"},
	{0x2CD2C0E682C333FBULL, "ser120cr_ghyslain.utc"},
	{0x2CD2DED85A32130EULL, "mrl150_merrill6093975.cl"},
	{0x2CD312A29BA6BCFDULL, "vxlt_gallowstemplar_4y0_0_0.vlm"},
	{0x2CD3CC5046D310E6ULL, "hlac_ht_keep_n_3732_12z.dds"},
	{0x2CD4A5E217740B10ULL, "gen00fl_varric6076122.cl"},
	{0x2CD52FC119B4E955ULL, "qcr221cr_qunari_mage.ncs"},
	{0x2CD5697D66EE91F7ULL, "sib111_mother6098795.cl"},
	{0x2CD59815A15E605CULL, "67353309.ogg"},
	{0x2CD5F7D328951F6DULL, "hlac_lt_gallowstemplar_n_2394_4x.dds"},
	{0x2CD626A6253EE79FULL, "vxht_keep_5zq_0_0.vlm"},
	{0x2CD66400E28FC25DULL, "gen_im_npc_wep_shd_krka.uti"},
	{0x2CD6AF894337E23EULL, "one300dg_solivitus6048294.cl"},
	{0x2CD6B9DC4B21992FULL, "dae000_gamlen6149455.cl"},
	{0x2CD711560A8C7077ULL, "6205972_m.fxe"},
	{0x2CD719544834EF7CULL, "126445583.ogg"},
	{0x2CD73CE9774C7266ULL, "6153979_m.fxe"},
	{0x2CD76599AC4155B5ULL, "hlac_lt_backalley_n_0_4z.dds"},
	{0x2CD7E72769505FE9ULL, "vxw_bonepit_20a_0_0.vlm"},
	{0x2CD7E91698A2A59BULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.cess.evt"},
	{0x2CD7F6290256BC68ULL, "vxlt_darktown_2tt_0_0.vlm"},
	{0x2CD82F77D859FF1FULL, "vxw_bonepitdestroyed_a3_1s9_0_0.vlm"},
	{0x2CD86CEED5C2E036ULL, "dae000_pc_home_ambients6208473.cl"},
	{0x2CD8C519F75809C6ULL, "vxlt_hangedman_n_wi_0_0.vlm"},
	{0x2CD8D3785EF6812FULL, "vxw_woundedcoast_1ah_0_0.vlm"},
	{0x2CD8E284AA71596BULL, "vxw_bonepitmining_a2_1v8_0_0.vlm"},
	{0x2CD8E77BBE72D95AULL, "vxlt_darktown_4l8_0_0.vlm"},
	{0x2CD8F8271DF482CDULL, "vxw_bonepitmining_a2_on_0_0.vlm"},
	{0x2CDA1B807453E550ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\pose\\pose_grovel_floor.ani"},
	{0x2CDA20799639181AULL, "379237908.ogg"},
	{0x2CDA5EC8F4E9EB2EULL, "6112110_m.fxe"},
	{0x2CDAC4B069A3A0FCULL, "face_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0x2CDB6F8C51B94358ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.inactivetoactive.ani"},
	{0x2CDBBB6852F67203ULL, "23316024.ogg"},
	{0x2CDBE79293F2A1C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_twitch_hands_on_hips.ani"},
	{0x2CDBF76C4BBECD2EULL, "6042088_m.fxe"},
	{0x2CDC489A6DB4DD59ULL, "vxdr_primevalentrance_22u_0_0.vlm"},
	{0x2CDC66A24F0BD6C6ULL, "lt_warehouse_4_0.rml"},
	{0x2CDC9889F32AAA17ULL, "6062053_f.fxe"},
	{0x2CDC99801A7D8422ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_1y.dds"},
	{0x2CDCCD5A6D2898A9ULL, "6024568_m.fxe"},
	{0x2CDDC3593E1CDBA7ULL, "vxht_chantry_n_1ys_0_0.vlm"},
	{0x2CDDF3FE331C90FDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wo_basic.phy"},
	{0x2CDE60D2ABC1F919ULL, "vxlt_0_lowtown_h22_0_0.vlm"},
	{0x2CDE7B59C29EC761ULL, "vxw_woundedcoast_cy_0_0.vlm"},
	{0x2CDEE21698A8B8DCULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.cess.gad"},
	{0x2CDEEFC13234805BULL, "hm_ave1_raiderm.mor"},
	{0x2CDF501B4DE73163ULL, "vxlt_0_lowtown_h6t_0_0.vlm"},
	{0x2CDF55C116CAE6C3ULL, "lt_0_lowtown_a3_n_c_28_0_lc.mmh"},
	{0x2CDF5E4A8A398C07ULL, "783085358.ogg"},
	{0x2CDFBD8E7F047E0EULL, "dre000_bodahn6031297.cl"},
	{0x2CE01882EF2E5DC7ULL, "344188447.ogg"},
	{0x2CE0880624D8BE7EULL, "vxht_mansion2_34p_0_0.vlm"},
	{0x2CE09183FD4CA7EAULL, "vxht_brothel_n_6in_0_0.vlm"},
	{0x2CE12A69D474055EULL, "mag101tr_grace_cower.ncs"},
	{0x2CE2E1DF7ED406F2ULL, "vfx_blt_forcefield_p.wbd"},
	{0x2CE31CA232D9D741ULL, "tools_start_pt.phy"},
	{0x2CE3309A6294ED97ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb04a.mmh"},
	{0x2CE33A04489A42E3ULL, "6125724_m.fxe"},
	{0x2CE3469A62951309ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb04a.msh"},
	{0x2CE3BBFF4F1A859AULL, "332102353.ogg"},
	{0x2CE439228231C963ULL, "vxht_chantry_n_2yl_0_0.vlm"},
	{0x2CE45C089A713A13ULL, "369282678.ogg"},
	{0x2CE4CC19251374E0ULL, "986629365.ogg"},
	{0x2CE4D4F33B8BDACFULL, "llac_ht_brothel_n_3373_2.dds"},
	{0x2CE4D63A6AAC803AULL, "6032991_m.fxe"},
	{0x2CE4E11698AD22DAULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.cess.ani"},
	{0x2CE56CF1745C4566ULL, "38035276.ogg"},
	{0x2CE5F7BDE7B5BCD1ULL, "772843004.ogg"},
	{0x2CE629FA3949F0A9ULL, "6123389_f.fxe"},
	{0x2CE72E3FE2F4BB46ULL, "vxw_encounter1_cv_0_0.vlm"},
	{0x2CE736F6E94E2246ULL, "vxlt_hangedman_n_rj_0_0.vlm"},
	{0x2CE7455550EC99AFULL, "vxw_woundedcoast_1lk_0_0.vlm"},
	{0x2CE75958B068F1DAULL, "vxlt_darktown_4k9_0_0.vlm"},
	{0x2CE76A040FEA9B4DULL, "vxw_bonepitmining_a2_jm_0_0.vlm"},
	{0x2CE86A99143B6A34ULL, "plt_lgt310pt_fixer_3.nss"},
	{0x2CE91D5B00A1F336ULL, "vxlt_warehouse_n_pl_0_0.vlm"},
	{0x2CE93E68A6AC62DAULL, "gen00fl_isabela6079068.cl"},
	{0x2CE99D7AFE2C3E18ULL, "vxlt_gallowsprison_n_6rm_0_0.vlm"},
	{0x2CEA1206A8A40519ULL, "40565302.ogg"},
	{0x2CEA3D1600889CA3ULL, "vxlt_hangedman_2hb_0_0.vlm"},
	{0x2CEABA775FAAF5D9ULL, "vxdr_primevalentrance_29t_0_0.vlm"},
	{0x2CEB3EE6418E56C0ULL, "xx_plaster.dds"},
	{0x2CEB73D91C77A2E6ULL, "6110369_m.fxe"},
	{0x2CEB85EB34396F4DULL, "w_bonepit_45_0.msh"},
	{0x2CEB8FB3CB036039ULL, "6121977_m.fxe"},
	{0x2CEB93EB34398680ULL, "w_bonepit_45_0.mao"},
	{0x2CEB97EB34398DD3ULL, "w_bonepit_45_0.mmh"},
	{0x2CEBC2F14671A3CCULL, "vxw_bonepitdestroyed_a3_6v_0_0.vlm"},
	{0x2CEBD29892B3606FULL, "bur_hole_dust.mao"},
	{0x2CEBF25D01E8DD06ULL, "vxht_mansion2_33a_0_0.vlm"},
	{0x2CEC9CECEE367FFEULL, "vxlt_warehouse_qn_0_0.vlm"},
	{0x2CECD2AF9DB81199ULL, "vxlt_0_lowtown_h71_0_0.vlm"},
	{0x2CECED36B494DFE1ULL, "vxw_woundedcoast_xx_0_0.vlm"},
	{0x2CECEE689EDB9D45ULL, "221095240.ogg"},
	{0x2CED30389E98CF99ULL, "6118119_m.fxe"},
	{0x2CED3C193BCB52B2ULL, "gen00fl_aveline6140798.cl"},
	{0x2CED66AFEF88E703ULL, "lgt310_mis_burn_body6180590.cl"},
	{0x2CED68FF4D124B1CULL, "971129818.ogg"},
	{0x2CEE118207D17A77ULL, "vxht_0_hightown_a2_9v1_0_0.vlm"},
	{0x2CEE373520364A4EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02a.mmh"},
	{0x2CEE51352036763CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02a.msh"},
	{0x2CEE943F25AF21C1ULL, "148639121.ogg"},
	{0x2CEEA8A89498D7D8ULL, "396449210.ogg"},
	{0x2CEEE8FE332AF169ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wo_basic.mao"},
	{0x2CEEF4FE332B0502ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wo_basic.mmh"},
	{0x2CEEF6FE332B08B8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wo_basic.msh"},
	{0x2CEFA25C1BA30CA6ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.cess.ani"},
	{0x2CEFD85D7CB1D6CEULL, "633117452.ogg"},
	{0x2CF001552B414AA9ULL, "dae000_arishok6068450.cl"},
	{0x2CF076B008EFB60BULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.anb"},
	{0x2CF0AAAD104A5717ULL, "6081273_m.fxe"},
	{0x2CF0B92DD05D465CULL, "6089829_m.fxe"},
	{0x2CF12980242B6D52ULL, "6129605_f.fxe"},
	{0x2CF15C405CC48CB2ULL, "678878958.ogg"},
	{0x2CF1A3E865EA7EBDULL, "vrc321ar_mansion_haunted.are"},
	{0x2CF1D03E19C3EFF9ULL, "6198708_m.fxe"},
	{0x2CF1DAC3882D25DBULL, "vxlt_0_lowtown_e8c_0_0.vlm"},
	{0x2CF21BB83270755AULL, "lt_gallowstemplar_n_2395.rml"},
	{0x2CF22B68A1A47A79ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinswall.phy"},
	{0x2CF24857F76765E3ULL, "538073593.ogg"},
	{0x2CF27CD987DB57A3ULL, "vxht_0_hightown_a2_byo_0_0.vlm"},
	{0x2CF2AAB43C57EEEEULL, "353546765.ogg"},
	{0x2CF2AAFF7427E1E3ULL, "vxht_chantry_n_2to_0_0.vlm"},
	{0x2CF2D960706621EBULL, "vxw_bonepitmining_a2_29b_0_0.vlm"},
	{0x2CF38F80E38E3D3EULL, "hlac_dr_0_deeproads_5845_10y.dds"},
	{0x2CF449275CADBA10ULL, "117306859.ogg"},
	{0x2CF573873DED631DULL, "hla_ht_0_hightown_a2_-2_0z.dds"},
	{0x2CF58F4A5E44D829ULL, "vxw_encounter2_45_0_0.vlm"},
	{0x2CF5A01CD4EAD3C6ULL, "vxw_encounter1_hw_0_0.vlm"},
	{0x2CF5F46B3728C51AULL, "vxlt_0_lowtown_a2_n_c_h2d_0_0.vlm"},
	{0x2CF5FCDF0BC68695ULL, "vxlt_gallowstemplar_f_5jc_0_0.vlm"},
	{0x2CF62C71826DC638ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_02a_0d.dds"},
	{0x2CF63425C91BCD33ULL, "vxlt_warehouse_10u_0_0.vlm"},
	{0x2CF645DB1EE83D5AULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_hed_snga_0n.dds"},
	{0x2CF6C3799D0DA657ULL, "vxw_woundedcoast_13n_0_0.vlm"},
	{0x2CF6CDBEBC9337DDULL, "500015604.ogg"},
	{0x2CF6EE7BA507CC55ULL, "6110003_m.fxe"},
	{0x2CF71DF1217EF0E3ULL, "hlac_dr_0_deeproads_5843_8y.dds"},
	{0x2CF775F7442715ABULL, "dae000st_grandcleric_anders.stg"},
	{0x2CF78F37F2980BB6ULL, "vxlt_warehouse_n_wm_0_0.vlm"},
	{0x2CF791B30B099D8BULL, "qun100_mage_party_comment6019110.cl"},
	{0x2CF7ACFED2C8EE14ULL, "6156206_m.fxe"},
	{0x2CF80743CB469E86ULL, "vxlt_warehouse_ro_0_0.vlm"},
	{0x2CF91B4735A250BFULL, "vxlt_hangedman_20n_0_0.vlm"},
	{0x2CF98CF26ABDF955ULL, "6208418_m.fxe"},
	{0x2CF9EC79167587FEULL, "84208738.ogg"},
	{0x2CFA296248EAA1C4ULL, "vxlt_0_lowtown_docks_a3_cy3_0_0.vlm"},
	{0x2CFA510E2ED59A12ULL, "673566544.ogg"},
	{0x2CFAB8A44B6E213DULL, "isa111_isabela6069022.cl"},
	{0x2CFADDA6A5448B50ULL, "mcr310_argument6072708.cl"},
	{0x2CFB0EC9E02C987EULL, "vxlt_warehouse_zo_0_0.vlm"},
	{0x2CFB6E9BF7D9C9B2ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_medj.phy"},
	{0x2CFB85EE0785806FULL, "6062788_m.fxe"},
	{0x2CFBD5B451242272ULL, "1009751849.ogg"},
	{0x2CFC8A5C1BAD6E17ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.cess.evt"},
	{0x2CFD0FB6C55D2043ULL, "vxlt_warehouse_n_1bh_0_0.vlm"},
	{0x2CFD20A97F49701AULL, "hlac_ht_0_hightown_-65537_4z.dds"},
	{0x2CFE0579F7E1F734ULL, "940157121.ogg"},
	{0x2CFE7C1B8D98D78FULL, "w_woundedcoast_164_0.mao"},
	{0x2CFE881B8D98EC2CULL, "w_woundedcoast_164_0.mmh"},
	{0x2CFE8E1B8D98F61EULL, "w_woundedcoast_164_0.msh"},
	{0x2CFE95B8BDF75861ULL, "111352383.ogg"},
	{0x2CFFD60E5DFC05B6ULL, "vrc321_varric6087035.cl"},
	{0x2CFFE5B7A22F3A1FULL, "167690879.ogg"},
	{0x2CFFF5D2636C4DC8ULL, "30595482.ogg"},
	{0x2D0049EB60B322D3ULL, "hlac_w_bonepitdestroyed_a3_458758_0x.dds"},
	{0x2D004CA07A233E5BULL, "vxlt_0_lowtown_e3b_0_0.vlm"},
	{0x2D007FC331311B29ULL, "lt_0_lowtown_a2_n_35_0_ll.mmh"},
	{0x2D023FA54A511CF6ULL, "character_character_p221.psh.fxo.dx11.assembly"},
	{0x2D029A8BE1D3BD84ULL, "6164764_m.fxe"},
	{0x2D02DB71DED61583ULL, "vxlt_0_lowtown_a3_n_c_hjd_0_0.vlm"},
	{0x2D03594BB8C64FF8ULL, "dae011_conspiracy_mob6149557.cl"},
	{0x2D038B5C1BB38EA0ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.cess.gad"},
	{0x2D03978D2F7A3572ULL, "ser270_templar6164082.cl"},
	{0x2D03B40946A56031ULL, "qun111_ketojan6041127.cl"},
	{0x2D03F6185C215632ULL, "6104613_f.fxe"},
	{0x2D04062883016D9DULL, "zz_followers_banter6158622.cl"},
	{0x2D040C2904E0D110ULL, "333227798.ogg"},
	{0x2D04290F34D8CAAFULL, "vxw_woundedcoast_1ru_0_0.vlm"},
	{0x2D042B948401D77DULL, "vxht_0_hightown_cb1_0_0.vlm"},
	{0x2D04693E4137D38BULL, "1037296948.ogg"},
	{0x2D04974505EAC458ULL, "mrl150_hunter6071297.cl"},
	{0x2D04E7C967DFA448ULL, "vxw_woundedcoast_uz_0_0.vlm"},
	{0x2D0509C220323748ULL, "dae300_alistair_king6133545.cl"},
	{0x2D0517CDA5BA8FC6ULL, "zz_mer_debug6073700.cl"},
	{0x2D0567C814D3D86CULL, "239706922.ogg"},
	{0x2D05AAF1B3F16D83ULL, "vxw_encounter2_r1_0_0.vlm"},
	{0x2D060114E48E2436ULL, "vxlt_warehouse_n_zn_0_0.vlm"},
	{0x2D067920BD3CB706ULL, "vxlt_warehouse_wp_0_0.vlm"},
	{0x2D06DD9B7CEF988AULL, "gen00fl_aveline6106899.cl"},
	{0x2D07B44214BE2D2DULL, "vxht_keep_n_6hu_0_0.vlm"},
	{0x2D0863E6062D05BCULL, "6144278_m.fxe"},
	{0x2D086AD195922A84ULL, "mag360_alain6044653.cl"},
	{0x2D08966B55EFD98FULL, "vxlt_darktown_575_0_0.vlm"},
	{0x2D0940CDB78BB79AULL, "6123412_f.fxe"},
	{0x2D09904142C77979ULL, "6151329_m.fxe"},
	{0x2D09F70AC13B8910ULL, "vxw_blightlands_13w_0_0.vlm"},
	{0x2D09F7CAF97B98EFULL, "6069688_m.fxe"},
	{0x2D0A1D36CC9A5706ULL, "zz_isa_debug6133227.cl"},
	{0x2D0A9013865F6CA3ULL, "llac_w_encounter1_-65535_1.dds"},
	{0x2D0B51C077014477ULL, "m_dae320ar_hightown.bnk"},
	{0x2D0B52BA12DE7A9AULL, "t3_tst_rgb.tnt"},
	{0x2D0B5BC315D8B99AULL, "gen_im_trs_qun_lgt_hlm_r0.uti"},
	{0x2D0B9C71ADC42A8FULL, "555982287.ogg"},
	{0x2D0BC81280DDFB53ULL, "llac_ca_caverns1_24425_3.dds"},
	{0x2D0C23F414EF8990ULL, "abi_m_deathsiphon.bnk"},
	{0x2D0C2BD099DE0EDDULL, "vxw_woundedcoast_10e_0_0.vlm"},
	{0x2D0C81EB54684AD8ULL, "6208688_m.fxe"},
	{0x2D0D3E88F30A1290ULL, "w_bonepitdestroyed_a3_136_0.msh"},
	{0x2D0D4C88F30A2AAAULL, "w_bonepitdestroyed_a3_136_0.mmh"},
	{0x2D0D5088F30A31E1ULL, "w_bonepitdestroyed_a3_136_0.mao"},
	{0x2D0D89D503230877ULL, "6068733_m.fxe"},
	{0x2D0DB7C8B3A10B59ULL, "6148428_m.fxe"},
	{0x2D0E4088B75B621BULL, "852758778.ogg"},
	{0x2D0E6054609AC97CULL, "6142377_m.fxe"},
	{0x2D0E6414326729D4ULL, "zz_followers_banter6157853.cl"},
	{0x2D0F238F99A99EA6ULL, "vxw_bonepitdestroyed_a3_1av_0_0.vlm"},
	{0x2D0F33BA5A6AF8E2ULL, "hlac_lt_0_lowtown_a2_n_65538_5z.dds"},
	{0x2D0FEA0A3C57AA9AULL, "6117015_f.fxe"},
	{0x2D103860C27B169BULL, "follower_banter6171267.cl"},
	{0x2D103F46020BCD85ULL, "static1_punch_dy.psh.fxo"},
	{0x2D1041361C92FFCDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_basic.msh"},
	{0x2D104F361C931700ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_basic.mao"},
	{0x2D1053361C931E53ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_basic.mmh"},
	{0x2D107CD9CD479D03ULL, "mer240_jansen6081446.cl"},
	{0x2D10C47A47017309ULL, "6023877_m.fxe"},
	{0x2D10C8ECE65E9E8CULL, "100581009.ogg"},
	{0x2D10EE6E24EE705FULL, "6118957_m.fxe"},
	{0x2D10EE96D3D0D8F5ULL, "vxlt_blackemporium_kk_0_0.vlm"},
	{0x2D10F282830B9596ULL, "358739120.ogg"},
	{0x2D116F256454B66BULL, "gen00fl_isabela6085745.cl"},
	{0x2D120EFA12670C55ULL, "art\\characters\\playercharacter\\humanmale\\rob_maga.phy"},
	{0x2D12C61337F12DB2ULL, "745235653.ogg"},
	{0x2D12D81B76457D2AULL, "llac_w_bonepitmining_a2_327680_0.dds"},
	{0x2D12F54C1B772F74ULL, "vxlt_0_lowtown_a2_n_f53_0_0.vlm"},
	{0x2D13895F67FB4D17ULL, "vxca_cavern1_haunted_don_0_0.vlm"},
	{0x2D13A73380F9D757ULL, "vxw_woundedcoast_19h_0_0.vlm"},
	{0x2D13C54A708662ADULL, "vxlt_0_lowtown_a2_emz_0_0.vlm"},
	{0x2D1413C62D8177DFULL, "6031173_m.fxe"},
	{0x2D14829742086A03ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple.mao"},
	{0x2D14869742087150ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple.mmh"},
	{0x2D1494974208896AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple.msh"},
	{0x2D14EAFDAF32CF86ULL, "vxlt_warehouse_xq_0_0.vlm"},
	{0x2D1528D7CD75F9D5ULL, "hlac_lt_undercityencounter_-1_45x.dds"},
	{0x2D15BB27E4BF30C0ULL, "t3_wdg_gb3.tnt"},
	{0x2D1632697E029B9EULL, "hlac_lt_gallowscourtyard_65539_0x.dds"},
	{0x2D166494A89D9321ULL, "263791827.ogg"},
	{0x2D1702A2C4D5C99CULL, "plt_vrc000pt_main.plo"},
	{0x2D17084847E5F20FULL, "vxlt_darktown_584_0_0.vlm"},
	{0x2D17A27EFCE30EEEULL, "vxlt_backalley_n_1ks_0_0.vlm"},
	{0x2D18105CBA04056DULL, "6126778_m.fxe"},
	{0x2D18562A3332219AULL, "lowtown_male_drunk_mob6151162.cl"},
	{0x2D186E9CC12A5D23ULL, "vxht_keep_62c_0_1.vlm"},
	{0x2D1880C56203B602ULL, "336069999.ogg"},
	{0x2D18DCAF003FDCE4ULL, "hlac_w_bonepitmining_a2_196608_0z.dds"},
	{0x2D18E0CD3109629DULL, "hlac_w_bonepit_0_10z.dds"},
	{0x2D19080521068C2EULL, "118338396.ogg"},
	{0x2D1A1BABEB246918ULL, "dae000_arishok6068668.cl"},
	{0x2D1A7C9A100BC343ULL, "hlac_dr_0_deeproads_5842_15x.dds"},
	{0x2D1AC74FE1E9B675ULL, "llac_lt_0_lowtown_docks_a3_n_65535_1.dds"},
	{0x2D1AD4E34BBDD5F4ULL, "6142539_m.fxe"},
	{0x2D1AE45EC442FFC7ULL, "587025938.ogg"},
	{0x2D1B2F7185E14187ULL, "hlac_ca_caverns2prc_5802_32x.dds"},
	{0x2D1B6810FBFD4EC6ULL, "965620909.ogg"},
	{0x2D1BBA2E2B3F3D25ULL, "zz_qcr_debug6074957.cl"},
	{0x2D1C68E9E0A080C0ULL, "one220_yevhen6102908.cl"},
	{0x2D1E00964A0A2B63ULL, "vxht_chantry_n_2ep_0_0.vlm"},
	{0x2D1E46B60F61461FULL, "vxdr_primevaltomb_8n9_0_0.vlm"},
	{0x2D1F39C5CA851CB6ULL, "w_woundedcoast_227_0.msh"},
	{0x2D1F424B50F16677ULL, "llac_lt_0_lowtown_n_-131071_1.dds"},
	{0x2D1F43C5CA852DB4ULL, "w_woundedcoast_227_0.mmh"},
	{0x2D1F47C5CA8534C7ULL, "w_woundedcoast_227_0.mao"},
	{0x2D1F6B281894AE4EULL, "6048250_f.fxe"},
	{0x2D204D1954F2B930ULL, "lgt210_shp_kill_qunari16156856.cl"},
	{0x2D20626CBE1E343FULL, "hlac_lt_0_lowtown_a2_n_c_-65537_0z.dds"},
	{0x2D20C7BAE1913A9EULL, "18078567.ogg"},
	{0x2D20D914701E238CULL, "6079263_f.fxe"},
	{0x2D21434649B60E78ULL, "6099594_m.fxe"},
	{0x2D21E820C74983DDULL, "918565816.ogg"},
	{0x2D221D2D481D06CCULL, "dae000_fenarel6147431.cl"},
	{0x2D22269955CF612BULL, "121847660.ogg"},
	{0x2D22561A39B71B0CULL, "254235047.ogg"},
	{0x2D2264A64AB90816ULL, "lt_0_lowtown_docks_a3_n_12_0.anb"},
	{0x2D239DFBC74F6EA6ULL, "6041984_m.fxe"},
	{0x2D242424DDF5AEA5ULL, "500753676.ogg"},
	{0x2D242FBFC12C070CULL, "408175883.ogg"},
	{0x2D2465690954FF7BULL, "hlac_dr_primevaltomb_661_0z.dds"},
	{0x2D246C322C2F6D60ULL, "vxw_encounter3_gc_0_1.vlm"},
	{0x2D24BB924913264FULL, "hlac_lt_0_lowtown_65536_5z.dds"},
	{0x2D253BC0C72E2D40ULL, "311826834.ogg"},
	{0x2D25A686D16FB3DCULL, "713891654.ogg"},
	{0x2D262680A181727FULL, "zz_seb_debug6185071.cl"},
	{0x2D2639133DAFEBFFULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstrb.phy"},
	{0x2D26400B4F8075D1ULL, "83472975.ogg"},
	{0x2D265A9390E4755CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_loop.ani"},
	{0x2D268A391E01FDB4ULL, "6127230_f.fxe"},
	{0x2D26FF3C7406DCBDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlb_1.mao"},
	{0x2D277B2CCAF5C0E5ULL, "vxlt_gallowstemplar_f_5j3_0_0.vlm"},
	{0x2D278D5F36534F4EULL, "vxlt_darktown_57x_0_0.vlm"},
	{0x2D287718A41FA9D8ULL, "hlac_w_bonepit_458752_0y.dds"},
	{0x2D2879CE3AACA1AAULL, "6137475_f.fxe"},
	{0x2D2AE4BD11907CBBULL, "85368634.ogg"},
	{0x2D2B029C1EF33460ULL, "6061809_m.fxe"},
	{0x2D2B8FE08BD38A6AULL, "vxw_encounter2_6p_0_0.vlm"},
	{0x2D2BE49592D4BAF5ULL, "zz_vault_debug6203274.cl"},
	{0x2D2C07497D95CFA6ULL, "vxw_bonepitdestroyed_a3_1ox_0_0.vlm"},
	{0x2D2C2F2B14189200ULL, "626961996.ogg"},
	{0x2D2C679390E8F69EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_run_loop.gad"},
	{0x2D2C928100859F1BULL, "hlac_lt_0_lowtown_a2_n_c_65537_5x.dds"},
	{0x2D2CB89301575E9FULL, "vxdr_primevaltomb_8i8_0_0.vlm"},
	{0x2D2D08513CE3F913ULL, "gen00fl_fenris6152643.cl"},
	{0x2D2D939242B1910FULL, "montage_subs_demohack6217117.cl"},
	{0x2D2D9E56A4AFE83BULL, "vxlt_0_lowtown_gkl_0_0.vlm"},
	{0x2D2E31E800315F75ULL, "163709502.ogg"},
	{0x2D2E3D63CC6F1002ULL, "6118493_m.fxe"},
	{0x2D2E597614945FFFULL, "vxlt_hangedman_n_2uj_0_0.vlm"},
	{0x2D2EC2EBBE6DB142ULL, "989486167.ogg"},
	{0x2D2F300CE65A6ABCULL, "vxlt_0_lowtown_h7h_0_0.vlm"},
	{0x2D2F7604765BD9F9ULL, "231680449.ogg"},
	{0x2D2F7E188D2696ADULL, "mrl300im_ring_gift.uti"},
	{0x2D2FA32594CE600EULL, "625313598.ogg"},
	{0x2D2FA5EBF52C0165ULL, "75149010.ogg"},
	{0x2D30CBC575C254FAULL, "853482488.ogg"},
	{0x2D319786BE6A7C19ULL, "hlac_w_bonepitdestroyed_a3_196609_0x.dds"},
	{0x2D319B7E6C304EF0ULL, "268866956.ogg"},
	{0x2D3205F41510265AULL, "abi_m_deathsiphon.wbd"},
	{0x2D32166F320FE0FDULL, "97219061.ogg"},
	{0x2D322329C155F578ULL, "6103352_m.fxe"},
	{0x2D32A1356AB135E5ULL, "972991954.ogg"},
	{0x2D32F0329A43900DULL, "125604250.ogg"},
	{0x2D347D14BDA462B5ULL, "6117024_m.fxe"},
	{0x2D35016309AEC640ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wax_sa.dds"},
	{0x2D358537B14D91EAULL, "mag210_arianni6024535.cl"},
	{0x2D35981984E2175DULL, "vxca_caverns1_fm5_0_0.vlm"},
	{0x2D35BBDE2694AAFDULL, "hlac_lt_0_lowtown_docks_a3_-65539_1z.dds"},
	{0x2D35C6F64C22CC21ULL, "zz_t_trailer_shot2.cut"},
	{0x2D35E68ED16E087EULL, "art\\characters\\playercharacter\\children\\mt_bas_km_eyesm1.mao"},
	{0x2D3601568806098EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv1.evt"},
	{0x2D362477B86ABAC5ULL, "6155967_m.fxe"},
	{0x2D36672730E9B3BAULL, "zz_qun_debug6064339.cl"},
	{0x2D3684B222BC8882ULL, "vxlt_darktown_5ij_0_0.vlm"},
	{0x2D369F2BDECBC0B8ULL, "dre000_bodahn6031259.cl"},
	{0x2D36DADAB0BF2E8AULL, "mag320_noblewoman_26060824.cl"},
	{0x2D370FA22D5E0586ULL, "838603761.ogg"},
	{0x2D373A133DBE6004ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstrb.mmh"},
	{0x2D373E133DBE6757ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstrb.mao"},
	{0x2D3750133DBE85A6ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstrb.msh"},
	{0x2D375A9B2F5DE97EULL, "ht_0_hightown_a2_26_0_ll.mmh"},
	{0x2D375E2D549958BAULL, "vxlt_0_lowtown_a3_n_c_gwy_0_0.vlm"},
	{0x2D3760D9931EF016ULL, "vxlt_0_lowtown_docks_ary_0_0.vlm"},
	{0x2D378777388F8FEBULL, "ser200_aveline6201465.cl"},
	{0x2D37939DAD79BF40ULL, "qun111ar_safehouse.lst"},
	{0x2D37A786F735085EULL, "417478063.ogg"},
	{0x2D37AFE436E843E8ULL, "6108498_m.fxe"},
	{0x2D37F3CDE44FAB47ULL, "vxht_0_hightown_b8x_0_0.vlm"},
	{0x2D3847D2D80CD437ULL, "gen00pt_combos.ncs"},
	{0x2D3929D178AC2458ULL, "6044057_m.fxe"},
	{0x2D393B91E79B3546ULL, "6163338_m.fxe"},
	{0x2D39788BE0E95F07ULL, "vxdr_0_deeproads_ega_0_0.vlm"},
	{0x2D39F37AC077E87EULL, "vxca_caverns1_i24_0_0.vlm"},
	{0x2D3A01BD7DC9A2EAULL, "vxw_encounter2_1q_0_0.vlm"},
	{0x2D3A0C0663B0606CULL, "vxht_mansion2_1pm_0_1.vlm"},
	{0x2D3A4865DC1592F6ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_lrgchair.msh"},
	{0x2D3A5265DC15A3F4ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_lrgchair.mmh"},
	{0x2D3A5665DC15AA07ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_lrgchair.mao"},
	{0x2D3AB039AAB22433ULL, "767043325.ogg"},
	{0x2D3AB29199015BFAULL, "607251641.ogg"},
	{0x2D3ADA5B7D0667EFULL, "6070387_m.fxe"},
	{0x2D3AEDC87D91A324ULL, "6090319_m.fxe"},
	{0x2D3B3D126608722EULL, "w_woundedcoast_173_0.msh"},
	{0x2D3B4B1266088A7FULL, "w_woundedcoast_173_0.mao"},
	{0x2D3B571266089E1CULL, "w_woundedcoast_173_0.mmh"},
	{0x2D3B648B545BE839ULL, "combat_ar_r1_a4.gda"},
	{0x2D3BE0D4FFE7CB7AULL, "hlac_ca_caverns1_mines_24427_5z.dds"},
	{0x2D3C0A747C9C123FULL, "6041484_m.fxe"},
	{0x2D3C442DA9B32275ULL, "vxlt_blackemporium_zh_0_0.vlm"},
	{0x2D3C714564DFA9C0ULL, "gen00fl_fenris6126542.cl"},
	{0x2D3C92AF993DD89BULL, "fx_cone_radius.mao"},
	{0x2D3CA91DDDC67372ULL, "one100dg_nabil6033135.cl"},
	{0x2D3CBCA25048ACEFULL, "gen_im_trs_cot_med_cht_r0.uti"},
	{0x2D3CDE90C2A81A3AULL, "hlac_lt_undercityencounter_-1_17x.dds"},
	{0x2D3CF856880C1939ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv1.gad"},
	{0x2D3D939DAD7E2A91ULL, "qun111ar_safehouse.ncs"},
	{0x2D3DA6E806B980A3ULL, "mrl310_merrill6144591.cl"},
	{0x2D3DE2FC43D4DFEFULL, "6108523_m.fxe"},
	{0x2D3E1BF742578729ULL, "vxw_blightlands_w3_0_0.vlm"},
	{0x2D3E4527743535CCULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_boat_.gad"},
	{0x2D3E58360C6220A9ULL, "art\\vfx\\combat\\vfx_a_stealth_aoe_p.mmh"},
	{0x2D3EA82DB84BDF0AULL, "372549518.ogg"},
	{0x2D3F6A82B73CBEB1ULL, "art\\cinematics\\_crit path\\followerintros\\cs_merill_merillintro.ani"},
	{0x2D409506B34C0DF8ULL, "6104252_m.fxe"},
	{0x2D41273919BC3EC3ULL, "vxht_chantry_n_2uq_0_1.vlm"},
	{0x2D4136AAFEE239F5ULL, "mag291_demon_marethari6029036.cl"},
	{0x2D415C99C8224AEEULL, "hlac_lt_0_lowtown_docks_-3_4z.dds"},
	{0x2D418A9DAD82040BULL, "qun111ar_safehouse.are"},
	{0x2D41C45911F17EB7ULL, "vxlt_0_lowtown_docks_a3_n_d3v_0_0.vlm"},
	{0x2D420595F8D9665EULL, "vxw_bonepitdestroyed_a3_18y_0_0.vlm"},
	{0x2D42893AB7571C5CULL, "ser100_party_comments6108242.cl"},
	{0x2D42CD0180C6FFFBULL, "vxlt_warehouse_8h_0_0.vlm"},
	{0x2D44095688125487ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv1.ani"},
	{0x2D441F2B502BF6D9ULL, "558865855.ogg"},
	{0x2D44299C89986473ULL, "ht_store_187_ll.mmh"},
	{0x2D44C3C2A4BFBF8AULL, "hlac_lt_warehouse_0_6y.dds"},
	{0x2D44F68F14B2A102ULL, "vxlt_darktown_5rk_0_0.vlm"},
	{0x2D4522443A3D8F78ULL, "400092825.ogg"},
	{0x2D4525A2332CB0B6ULL, "tools_start_pt.mmh"},
	{0x2D452FA2332CC1B4ULL, "tools_start_pt.msh"},
	{0x2D4542F7614E144DULL, "gen00fl_fenris6148974.cl"},
	{0x2D454427743B52CAULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_boat_.ani"},
	{0x2D455E33AB9B5749ULL, "vxw_bonepitdestroyed_a3_ng_0_0.vlm"},
	{0x2D45B7A628CF666CULL, "74658251.ogg"},
	{0x2D45E8BBE0E42AD3ULL, "51257913.ogg"},
	{0x2D45F57D367D1FCAULL, "vxlt_hangedman_n_253_0_0.vlm"},
	{0x2D4653593426B019ULL, "953669708.ogg"},
	{0x2D467582B742F053ULL, "art\\cinematics\\_crit path\\followerintros\\cs_merill_merillintro.gad"},
	{0x2D46D4CCA052648DULL, "vxht_chantry_2pz_0_0.vlm"},
	{0x2D47931F1E9781E9ULL, "vxw_blightlands_11t_0_0.vlm"},
	{0x2D47C8AEECF0465AULL, "vxlt_warehouse_n_88_0_0.vlm"},
	{0x2D47CEF2A16CCDACULL, "vxdr_primevaltomb_4hl_0_0.vlm"},
	{0x2D47EC2E7AF5DE69ULL, "gen00fl_anders6125199.cl"},
	{0x2D481CC79FE101E0ULL, "plt_cod_prc_vessel_of_tears.plo"},
	{0x2D4861EBA7B45B1CULL, "339005426.ogg"},
	{0x2D48D974D4A44E90ULL, "mag121_eau_banter.cnv"},
	{0x2D48EBCADA137BE8ULL, "art\\vfx\\materials\\chunky\\vfx_mer_sword_glow.mao"},
	{0x2D4973420F25055AULL, "693351054.ogg"},
	{0x2D497D7191D602E0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curver.phy"},
	{0x2D49E6F0499803CBULL, "615445088.ogg"},
	{0x2D4B91ED55AF050EULL, "12442212.ogg"},
	{0x2D4BB3E1A196E19CULL, "6048422_m.fxe"},
	{0x2D4BC8045310CC49ULL, "6163456_m.fxe"},
	{0x2D4BCAE36AD13812ULL, "ht_0_hightown_a3_27_0.rml"},
	{0x2D4BCAEE95C234E0ULL, "zz_one_debug6077391.cl"},
	{0x2D4C12256883D386ULL, "mrl000_marethari6069875.cl"},
	{0x2D4C3FC93CE1F22AULL, "w_blightlands_26_0.mao"},
	{0x2D4C43C93CE1F979ULL, "w_blightlands_26_0.mmh"},
	{0x2D4C496C057E4EB7ULL, "hlac_lt_0_lowtown_a2_n_c_65534_0z.dds"},
	{0x2D4C4DC93CE20A67ULL, "w_blightlands_26_0.msh"},
	{0x2D4C8C5910329575ULL, "6136731_m.fxe"},
	{0x2D4C8DD4344D9FA9ULL, "vxw_blightlands_x2_0_0.vlm"},
	{0x2D4CE9EF69C3204FULL, "gencr_street_thug_2w.utc"},
	{0x2D4CFB74D4A871E8ULL, "mag121_eau_banter.bnk"},
	{0x2D4D7A25E3BB71DFULL, "vxdr_primevaltomb_7qg_0_0.vlm"},
	{0x2D4DD53348D034BDULL, "32719502.ogg"},
	{0x2D4EC5BD2544B700ULL, "gen00fl_anders6125343.cl"},
	{0x2D4F90EDA7C28089ULL, "6108950_m.fxe"},
	{0x2D4FD72A9C71C95DULL, "57506990.ogg"},
	{0x2D4FF4906F46C4F9ULL, "6121076_m.fxe"},
	{0x2D507133C28250FCULL, "6042644_m.fxe"},
	{0x2D5111598E06A034ULL, "llac_lt_gallowstemplar_2394_11.dds"},
	{0x2D5121C493B92449ULL, "6050444_m.fxe"},
	{0x2D51220F3BE1E523ULL, "6206133_m.fxe"},
	{0x2D516A00A94D9933ULL, "mcr370_meredith6083586.cl"},
	{0x2D51DF4B4AD40701ULL, "hlac_lt_gallowscourtyard_n_-2_3z.dds"},
	{0x2D52915FAF463EE6ULL, "469894703.ogg"},
	{0x2D5295B6184A873AULL, "mag101_grace6034025.cl"},
	{0x2D52EE4EBC3FCBDFULL, "hlac_lt_0_lowtown_a2_n_c_-196609_3z.dds"},
	{0x2D5317397D194949ULL, "542057635.ogg"},
	{0x2D536C4D89BB8A74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.gad"},
	{0x2D5372CE029912F5ULL, "6074033_m.fxe"},
	{0x2D53C01188FE6BFEULL, "follower_banter6161166.cl"},
	{0x2D53DD4716B0B539ULL, "llac_ht_bartrand_n_3373_2.dds"},
	{0x2D540AA22A0F68F7ULL, "vxht_keep_n_3yi_0_0.vlm"},
	{0x2D548CAF0ADCF3FCULL, "6089505_f.fxe"},
	{0x2D548D97423E616BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple.phy"},
	{0x2D5528CDA55418C6ULL, "vxca_caverns1_mines_fmk_0_0.vlm"},
	{0x2D5528DD42B499C5ULL, "dae300_vashothlead6062789.cl"},
	{0x2D5547A64AE430AAULL, "lt_0_lowtown_docks_a3_n_12_0.rml"},
	{0x2D55CD97FD7DB4B6ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_u_c.impa.evt"},
	{0x2D55FE2CBD377BFEULL, "vxw_encounter3_3p_0_0.vlm"},
	{0x2D5668FDFDA6A52AULL, "vxlt_hangedman_n_we_0_0.vlm"},
	{0x2D56E2B4FABE4AB5ULL, "vxlt_hangedman_n_1xt_0_0.vlm"},
	{0x2D56FEE74422881CULL, "vxw_encounter2_camp_nv_0_0.vlm"},
	{0x2D5725055C914D74ULL, "mag131ar_bloodmages.are"},
	{0x2D574E394CEAD4EEULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01loop.ani"},
	{0x2D57BB990918173DULL, "398870653.ogg"},
	{0x2D57BE31CB4E5646ULL, "lgt160_ntb_elf_conflict6163365.cl"},
	{0x2D586E15AC74C819ULL, "hlac_w_woundedcoast_262145_0z.dds"},
	{0x2D59139955BCE8E4ULL, "vxht_0_hightown_aqk_0_0.vlm"},
	{0x2D5921FA12A32F70ULL, "art\\characters\\playercharacter\\humanmale\\rob_maga.msh"},
	{0x2D592FFA12A3468AULL, "art\\characters\\playercharacter\\humanmale\\rob_maga.mmh"},
	{0x2D5933FA12A34DC1ULL, "art\\characters\\playercharacter\\humanmale\\rob_maga.mao"},
	{0x2D597A530C9DAEE0ULL, "932753807.ogg"},
	{0x2D59F3BF22D26CC4ULL, "art\\characters\\weapons\\unique\\textures\\w_vael_0d.dds"},
	{0x2D5A033ACC7A5ECCULL, "art\\levels\\ztools\\default\\generic\\0pp_basic.phy"},
	{0x2D5A7B4D89C1C252ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.ani"},
	{0x2D5A880F45D8B200ULL, "ave310_brennan6137228.cl"},
	{0x2D5A8BF8FFB01466ULL, "gen00fl_merrill6107031.cl"},
	{0x2D5AC6B627C110EFULL, "6106723_m.fxe"},
	{0x2D5ADA4583628C7BULL, "221821663.ogg"},
	{0x2D5B1175D4E3709AULL, "vxlt_0_lowtown_a2_n_c_hoc_0_0.vlm"},
	{0x2D5B2D3EEF8093EAULL, "39840695.ogg"},
	{0x2D5B5C60FC4FB958ULL, "vxht_chantry_n_2is_0_0.vlm"},
	{0x2D5B86EE64D6DAD8ULL, "81442838.ogg"},
	{0x2D5BE674141F0BB8ULL, "793005022.ogg"},
	{0x2D5C90F5DC05472FULL, "vxlt_darktown_5k1_0_0.vlm"},
	{0x2D5CB853C68D69F1ULL, "6017675_f.fxe"},
	{0x2D5CC497FD83C441ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_u_c.impa.gad"},
	{0x2D5CD38CF059111DULL, "vxlt_shop_j3_0_0.vlm"},
	{0x2D5D292C23F3E741ULL, "vxdr_primevalentrance_23b_0_0.vlm"},
	{0x2D5D2D99007E5813ULL, "vxht_bartrand_n_6pl_0_0.vlm"},
	{0x2D5D3B6FF4F47B33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_browsing_store.evt"},
	{0x2D5DC50244FB538EULL, "vxw_encounter2_4o_0_1.vlm"},
	{0x2D5E88056966357AULL, "vxlt_0_lowtown_docks_a3_czh_0_0.vlm"},
	{0x2D5F2A7D95066FE1ULL, "llac_dr_0_deeproads_5840_2.dds"},
	{0x2D5F74DA30002277ULL, "vxlt_0_lowtown_gkx_0_0.vlm"},
	{0x2D5FB62E0D92B98DULL, "hlac_dr_0_deeproads_5843_3z.dds"},
	{0x2D604C8B4C6E3476ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8man.msh"},
	{0x2D604F8A35417106ULL, "vxlt_blackemporium_14p_0_0.vlm"},
	{0x2D60568B4C6E4574ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8man.mmh"},
	{0x2D605A8B4C6E4B87ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8man.mao"},
	{0x2D6061A6894253FBULL, "lt_gallowstemplar_f_2387.rml"},
	{0x2D60B2B3C337B675ULL, "hlac_w_woundedcoast_131075_0y.dds"},
	{0x2D60D6EE2A3FB7D4ULL, "6110257_m.fxe"},
	{0x2D6100D1C61B7A4CULL, "t3_eye_blk.tnt"},
	{0x2D6152A47D89A6D4ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_f_c.mmh"},
	{0x2D615EE63E191A20ULL, "6031660_m.fxe"},
	{0x2D6168839A479A86ULL, "vxw_encounter3_4a_0_0.vlm"},
	{0x2D61DA48F89ED202ULL, "vxlt_darktown_5xu_0_0.vlm"},
	{0x2D621DAF50EE763BULL, "563788828.ogg"},
	{0x2D627F84617A1CEFULL, "sib131_slaver_guard6165284.cl"},
	{0x2D63125EB7078CB6ULL, "hlac_ht_mansion2_818_11z.dds"},
	{0x2D6361F610988C8BULL, "6115482_m.fxe"},
	{0x2D63658CD337FF90ULL, "6043760_m.fxe"},
	{0x2D63861B406CA686ULL, "hlac_dr_slavers_3214_14y.dds"},
	{0x2D639EE3FD49A3BDULL, "793566728.ogg"},
	{0x2D63C597FD89E4EFULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_u_c.impa.ani"},
	{0x2D642A3E016B7F8CULL, "vxlt_undercitydungeon_17y_0_0.vlm"},
	{0x2D64346FF4FA8E74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_browsing_store.gad"},
	{0x2D6446394CF5513FULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01loop.evt"},
	{0x2D648A7657433DDFULL, "vxca_cave_239_0_0.vlm"},
	{0x2D64B60A87ABC55FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow1base.phy"},
	{0x2D64C78BBB1E4DD0ULL, "6095903_m.fxe"},
	{0x2D64D223E94356AAULL, "vxw_encounter1_cz_0_0.vlm"},
	{0x2D6502746B659F63ULL, "405156200.ogg"},
	{0x2D6548969BC5727DULL, "vxlt_0_lowtown_n_eua_0_0.vlm"},
	{0x2D6570C43618A09CULL, "vxw_encounter2_camp_iw_0_0.vlm"},
	{0x2D65AFD07099F0BBULL, "244357497.ogg"},
	{0x2D65F1BB2581AC34ULL, "gen00fl_varric6103495.cl"},
	{0x2D6609E00D3EB5D7ULL, "gen00fl_anders6119419.cl"},
	{0x2D670534261D8EF1ULL, "hlac_lt_undercityserialkiller_-1_34z.dds"},
	{0x2D6755360C8520DEULL, "art\\vfx\\combat\\vfx_a_stealth_aoe_p.anb"},
	{0x2D676A0A1F5CC6F2ULL, "llac_w_sundermount_196603_0.dds"},
	{0x2D6799C47F956BF5ULL, "vxlt_blackemporium_uu_0_0.vlm"},
	{0x2D67A3950F3F0572ULL, "376820071.ogg"},
	{0x2D67A5D13506A736ULL, "hlac_ht_brothel_n_3366_0y.dds"},
	{0x2D67BA7D2EB32B57ULL, "538747634.ogg"},
	{0x2D67D1AADF9DF343ULL, "framebuffereffects_diffusiondof_p3.psh.fxo.dx11"},
	{0x2D686419660B4614ULL, "6053682_m.fxe"},
	{0x2D68C84630FCF197ULL, "gen00fl_isabela6118379.cl"},
	{0x2D68D5D25ED67F70ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig_sa.dds"},
	{0x2D6945EA2CC004ECULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_0.pwk"},
	{0x2D69540B80739857ULL, "461231248.ogg"},
	{0x2D69C42E63DA081AULL, "hlac_lt_0_lowtown_a3_n_c_-131072_0x.dds"},
	{0x2D69CAD687C76FCEULL, "hlac_lt_0_lowtown_docks_a3_-4_4z.dds"},
	{0x2D6A5CE46E18766EULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\femalemage1_.ani"},
	{0x2D6AB96FB64D210CULL, "dae000_cerimon.cnv"},
	{0x2D6B27394CFB3B28ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01loop.gad"},
	{0x2D6B436FF500C652ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_browsing_store.ani"},
	{0x2D6B9B0915E9FFC1ULL, "vxdr_primevalentrance_26a_0_0.vlm"},
	{0x2D6C4ED1A0D2E23BULL, "vxht_chantry_n_36h_0_0.vlm"},
	{0x2D6C8619F893F607ULL, "20048494.ogg"},
	{0x2D6D60C4E7D46E98ULL, "lowtown_male_beggar_sta6151041.cl"},
	{0x2D6D7A2007F79519ULL, "6172313_m.fxe"},
	{0x2D6E8510F99F9F90ULL, "8472030.ogg"},
	{0x2D6E9B6FB650D73CULL, "dae000_cerimon.bnk"},
	{0x2D6EAEB2C4C02A96ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldc_sa.dds"},
	{0x2D6F9A7191F70375ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curver.msh"},
	{0x2D6F9C7191F706EBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curver.mmh"},
	{0x2D6FA87191F71A88ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_curver.mao"},
	{0x2D6FF300FAE567EBULL, "468705124.ogg"},
	{0x2D702EBE2DACCFE9ULL, "vxlt_backalley_n_1x1_0_0.vlm"},
	{0x2D712BCC9D8309E8ULL, "255663423.ogg"},
	{0x2D71373D06932164ULL, "nhg.cut"},
	{0x2D71C31F64A588C1ULL, "6188572_m.fxe"},
	{0x2D71E49E97147200ULL, "follower_banter6161256.cl"},
	{0x2D7228592DAF0070ULL, "6116134_m.fxe"},
	{0x2D725B4AD7E026D3ULL, "dae000_surprises.bnk"},
	{0x2D7290750862704AULL, "6205609_m.fxe"},
	{0x2D72D8E9C11B3929ULL, "586177634.ogg"},
	{0x2D736F459D87ED31ULL, "dae000cr_grand_cleric6147710.cl"},
	{0x2D73B2D6648D8BA8ULL, "lowtown_fem_urchin_mob.cnv"},
	{0x2D73E2A1280EB91CULL, "vxw_encounter2_camp_dt_0_0.vlm"},
	{0x2D7413ECE9E2DB1AULL, "hlac_ht_0_hightown_a2_0_0y.dds"},
	{0x2D741C8D6FE6396AULL, "vxw_blightlands_qz_0_0.vlm"},
	{0x2D74367471C372C3ULL, "lt_0_lowtown_docks_a3_n_118_0.mao"},
	{0x2D743A7471C37910ULL, "lt_0_lowtown_docks_a3_n_118_0.mmh"},
	{0x2D74487471C3912AULL, "lt_0_lowtown_docks_a3_n_118_0.msh"},
	{0x2D74508058B17FBBULL, "hlac_w_bonepitmining_a2_393217_0z.dds"},
	{0x2D745B9F885767B4ULL, "6027595_m.fxe"},
	{0x2D746E93C26943BCULL, "gen00fl_aveline6132739.cl"},
	{0x2D756E0DC677ED53ULL, "6040627_m.fxe"},
	{0x2D757D4AD7E29623ULL, "dae000_surprises.cnv"},
	{0x2D759121AB2AA03FULL, "lt_0_lowtown_docks_a3_n_227_0.msh"},
	{0x2D759721AB2AAAE1ULL, "lt_0_lowtown_docks_a3_n_227_0.mmh"},
	{0x2D75A321AB2ABE82ULL, "lt_0_lowtown_docks_a3_n_227_0.mao"},
	{0x2D75B70A87BA3A64ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow1base.mmh"},
	{0x2D75BB0A87BA40B7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow1base.mao"},
	{0x2D75BDC888D3AAEBULL, "391189396.ogg"},
	{0x2D75CD0A87BA5F06ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow1base.msh"},
	{0x2D75D2AD7F140605ULL, "590303065.ogg"},
	{0x2D75DB4899963915ULL, "vxlt_gallowstemplar_n_5qf_0_0.vlm"},
	{0x2D7606CDF8D94B66ULL, "qcr000_arishok6096436.cl"},
	{0x2D760C055CABD356ULL, "mag131ar_bloodmages.ncs"},
	{0x2D76208C9E07E030ULL, "234578670.ogg"},
	{0x2D76348515C98C97ULL, "6084700_m.fxe"},
	{0x2D76D4D6648FFB10ULL, "lowtown_fem_urchin_mob.bnk"},
	{0x2D770672822DCDFCULL, "6149026_m.fxe"},
	{0x2D789605244C8EDAULL, "6080618_f.fxe"},
	{0x2D78E2E774389864ULL, "737805987.ogg"},
	{0x2D7905D770594F72ULL, "fx_glow02_a.dds"},
	{0x2D7974AFBFF1782FULL, "vxlt_darktown_5q3_0_0.vlm"},
	{0x2D79893F7A3B04C4ULL, "563522740.ogg"},
	{0x2D79CB50749F4667ULL, "mrl000_marethari6129314.cl"},
	{0x2D79E029DE116EC2ULL, "vxlt_darktown_2v4_0_0.vlm"},
	{0x2D7A647934E2CC20ULL, "6064856_f.fxe"},
	{0x2D7A817DAF656D82ULL, "ser100_party_comments6120596.cl"},
	{0x2D7AC0AE92C8FABBULL, "vxht_chantry_n_33k_0_0.vlm"},
	{0x2D7B42C5CAD38AAFULL, "w_woundedcoast_227_0.phy"},
	{0x2D7C6EA47B8F646CULL, "hlac_lt_gallowstemplar_2384_3z.dds"},
	{0x2D7CDFDAF555239AULL, "6118211_m.fxe"},
	{0x2D7D0B2B1D201357ULL, "6189003_m.fxe"},
	{0x2D7D0B3889832CCCULL, "art\\levels\\ztools\\default\\generic\\0ch_basic.msh"},
	{0x2D7D1138898336BEULL, "art\\levels\\ztools\\default\\generic\\0ch_basic.mmh"},
	{0x2D7D1D3889834B55ULL, "art\\levels\\ztools\\default\\generic\\0ch_basic.mao"},
	{0x2D7D24055CB21AEFULL, "mag131ar_bloodmages.lst"},
	{0x2D7D3D41F51D7655ULL, "vxlt_0_lowtown_a2_n_f4y_0_0.vlm"},
	{0x2D7D7BA605423067ULL, "vxdr_0_deeproads_cws_0_0.vlm"},
	{0x2D7E35E46E28DCA8ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\femalemage1_.gad"},
	{0x2D7E4E5928E13D03ULL, "992079090.ogg"},
	{0x2D7E4F777D5AC94AULL, "vxht_bartrand_n_4l7_0_0.vlm"},
	{0x2D7F8A37BB1C2003ULL, "rdr111dg_harbormaster6016139.cl"},
	{0x2D7FB77DC2C0485AULL, "mag320_orsino6071089.cl"},
	{0x2D801640494684B4ULL, "6075196_m.fxe"},
	{0x2D80706C8192881BULL, "dae000_meredith6080974.cl"},
	{0x2D80C05D8B7B3A70ULL, "6082213_m.fxe"},
	{0x2D80DFFF8DEB42C1ULL, "vxht_0_hightown_a2_by5_0_0.vlm"},
	{0x2D80FB62432A99F2ULL, "default_blackcubemap.dds"},
	{0x2D8162A2213BAA85ULL, "art\\vfx\\combat\\vfx_w_off_accuracy_activ_c.anb"},
	{0x2D817A6DBE05F793ULL, "6118697_f.fxe"},
	{0x2D81A362691D5615ULL, "vxht_bartrand_n_3yl_0_0.vlm"},
	{0x2D81F8AB1AAEA1B1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeexit.ani"},
	{0x2D81FE6F1CE4CD7CULL, "hlac_dr_slavers_3216_10y.dds"},
	{0x2D8233ED9E258742ULL, "nat380_party_banter6090056.cl"},
	{0x2D825A2C8D975C9DULL, "652103549.ogg"},
	{0x2D82A5294F7226AEULL, "193467202.ogg"},
	{0x2D837FC315A3F272ULL, "6064699_m.fxe"},
	{0x2D838E8CB8C457E9ULL, "6129007_m.fxe"},
	{0x2D846AEF07E462FFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masf_0n.dds"},
	{0x2D8502D517143F0BULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_dt_smoke_sml.mmh"},
	{0x2D858E899970452BULL, "m_ser221ar_murdrer_estat_nt.wbd"},
	{0x2D85ACEF023D6967ULL, "6054099_m.fxe"},
	{0x2D85B3E1C92953DEULL, "gen_im_trs_krk_hgh_junk3_04.uti"},
	{0x2D85C30C1828E135ULL, "qun110_petrice6042206.cl"},
	{0x2D866F762323FC0CULL, "6112905_m.fxe"},
	{0x2D86CF27B305D99DULL, "6154753_m.fxe"},
	{0x2D870A314263BB12ULL, "vxht_mansion2_25b_0_0.vlm"},
	{0x2D872C73DEE853B9ULL, "dae000_arishok6078338.cl"},
	{0x2D87D19B47298785ULL, "vxlt_0_lowtown_docks_n_c1b_0_0.vlm"},
	{0x2D885206D0078742ULL, "vxlt_darktown_2y5_0_0.vlm"},
	{0x2D8903AB1AB4D353ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeexit.gad"},
	{0x2D897FA96DDAC4C5ULL, "vxlt_0_lowtown_docks_a3_czc_0_0.vlm"},
	{0x2D89CA192FEA5595ULL, "6089571_f.fxe"},
	{0x2D8A4338D269EEE0ULL, "gen00fl_mrl_act3_gift2.stg"},
	{0x2D8A55A47DACB08BULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_f_c.anb"},
	{0x2D8A9B6C55EAFCE8ULL, "743419918.ogg"},
	{0x2D8ABC97EFA9AFFBULL, "6117981_m.fxe"},
	{0x2D8B06B28A6261F3ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_004a_0n.dds"},
	{0x2D8B653242BF488DULL, "hlac_lt_0_lowtown_docks_-196610_4y.dds"},
	{0x2D8BBBA51E4E981CULL, "6039750_m.fxe"},
	{0x2D8C1B00076EC625ULL, "mrl310_party_comments6143066.cl"},
	{0x2D8CA7993006FCDBULL, "465695857.ogg"},
	{0x2D8D28546AF5DC8EULL, "536333083.ogg"},
	{0x2D8D825838FA1543ULL, "vxlt_warehouse_n_14r_0_0.vlm"},
	{0x2D8DB1EF8EB72778ULL, "vxht_0_hightown_a2_cf9_0_1.vlm"},
	{0x2D8DC05F71FED109ULL, "terrain1.psh.fxo"},
	{0x2D8E1398CAE8E24AULL, "58212468.ogg"},
	{0x2D8E87E45C194D93ULL, "91150100.ogg"},
	{0x2D8E8895A6D9E303ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v1.ani"},
	{0x2D8E95A6DAE41129ULL, "6106567_m.fxe"},
	{0x2D8EC2BEE2CCD5D9ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he1.msh"},
	{0x2D8ECCBEE2CCE6C7ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he1.mmh"},
	{0x2D8F00AB1AB939B8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeexit.evt"},
	{0x2D8F35061CE626B8ULL, "vxht_mansion_n_61w_0_0.vlm"},
	{0x2D8F6207699FF906ULL, "vxht_mansion_n_619_0_0.vlm"},
	{0x2D8F96336DEF242DULL, "6017838_m.fxe"},
	{0x2D8FA8CD4B1E5983ULL, "hlac_w_bonepitdestroyed_a3_0_2y.dds"},
	{0x2D907A949206504AULL, "718271776.ogg"},
	{0x2D909C04357B96B6ULL, "33660477.ogg"},
	{0x2D90FCA24CC20E38ULL, "6059253_m.fxe"},
	{0x2D9160E0C5B7A5D3ULL, "6156894_m.fxe"},
	{0x2D9179070A9C2BB3ULL, "hlac_lt_darktown_-1_61z.dds"},
	{0x2D91BD7CDF2B9610ULL, "mcr371_hostile_merrill6173438.cl"},
	{0x2D93915B46FD4425ULL, "472732723.ogg"},
	{0x2D940F33A215F989ULL, "586284435.ogg"},
	{0x2D947C45F68C590DULL, "645175755.ogg"},
	{0x2D94AABCF3F097EBULL, "vxw_bonepitmining_a2_177_0_0.vlm"},
	{0x2D94BA755C6E0CFFULL, "368363035.ogg"},
	{0x2D94E78490A5F46CULL, "vxlt_gallowstemplar_2p8_0_0.vlm"},
	{0x2D95239D838A4F5DULL, "1023216299.ogg"},
	{0x2D955165DC624DEFULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_lrgchair.phy"},
	{0x2D956BC01EDEFD0EULL, "vxlt_0_lowtown_a2_n_c_gi2_0_0.vlm"},
	{0x2D95CF7764AFB525ULL, "vxca_cavern1_haunted_jfa_0_0.vlm"},
	{0x2D95EBE8DD05A126ULL, "gen00fl_carver6104414.cl"},
	{0x2D961CB46D2459F5ULL, "plt_lgt210pt_mistress_selby.nss"},
	{0x2D962372E77EA093ULL, "322140533.ogg"},
	{0x2D9638C9EC20D2F3ULL, "hlac_w_bonepitmining_a2_65537_0y.dds"},
	{0x2D96CE7A307FDC53ULL, "vxlt_hangedman_n_15t_0_1.vlm"},
	{0x2D9716B33596C7C5ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_104a_d.dds"},
	{0x2D97D0CCEE60E01CULL, "vxlt_0_lowtown_a3_n_c_gm9_0_0.vlm"},
	{0x2D992A790391543AULL, "6093990_m.fxe"},
	{0x2D9978F671C5267DULL, "506857235.ogg"},
	{0x2D99910612E23316ULL, "vxlt_0_lowtown_docks_b48_0_0.vlm"},
	{0x2D99BF6D47C393D5ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls102a_0t.dds"},
	{0x2D99DA3ACCAFFC97ULL, "art\\levels\\ztools\\default\\generic\\0pp_basic.mmh"},
	{0x2D99DE3ACCB003A4ULL, "art\\levels\\ztools\\default\\generic\\0pp_basic.mao"},
	{0x2D99E8102ABB0598ULL, "5394496.ogg"},
	{0x2D99F03ACCB02209ULL, "art\\levels\\ztools\\default\\generic\\0pp_basic.msh"},
	{0x2D9AC7F1911C3976ULL, "942727388.ogg"},
	{0x2D9B084FF20331BAULL, "fxe_snow_tiny.dds"},
	{0x2D9B2B49451349A1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_floor_r.dds"},
	{0x2D9B3D48E238572EULL, "d15_blood_link.dds"},
	{0x2D9B50F2608B40ABULL, "vxlt_house_eh_0_0.vlm"},
	{0x2D9BA095A6E496FAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v1.evt"},
	{0x2D9BE9542E88A1E3ULL, "dae000_elegant6166913.cl"},
	{0x2D9C67F7EB2A2B19ULL, "vxca_caverns1_mines_i1i_0_2.vlm"},
	{0x2D9CA5D77363ABD0ULL, "gen_im_trs_cot_med_boo_r2.uti"},
	{0x2D9CEA3726490597ULL, "6060231_f.fxe"},
	{0x2D9CEF10A324842FULL, "llac_ca_caverns1_mines_24421_2.dds"},
	{0x2D9CFB4F6DDB04EEULL, "748675326.ogg"},
	{0x2D9D04468922AD5EULL, "vxlt_gallowscourtyard_7bw_0_0.vlm"},
	{0x2D9D1AB7C2ADBAACULL, "6033639_m.fxe"},
	{0x2D9D7A55B787736CULL, "hlac_ht_0_hightown_n_-131070_0y.dds"},
	{0x2D9D880E772D72F8ULL, "ch2em.psh.fxo"},
	{0x2D9DE15D743CB77AULL, "325627894.ogg"},
	{0x2D9E5448AD307E66ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_ext_c.impa.evt"},
	{0x2D9EC525B2DF8DDCULL, "896893856.ogg"},
	{0x2D9F54CD6BB41D09ULL, "lt_0_lowtown_docks_20_0.rml"},
	{0x2D9F5BFCF92E225BULL, "vxht_mansion_n_3yi_0_0.vlm"},
	{0x2DA032F38202FB8EULL, "153096581.ogg"},
	{0x2DA074BA33141909ULL, "barr_distor.mao"},
	{0x2DA13315221055E0ULL, "hlac_dr_primevaltomb_9621_2z.dds"},
	{0x2DA14E44AF13FF22ULL, "hlac_ca_cavern1_haunted_24423_3z.dds"},
	{0x2DA1A3D620C44EBBULL, "6176427_m.fxe"},
	{0x2DA1DE8F479FA78BULL, "isa311_isabela6117918.cl"},
	{0x2DA1F21402690405ULL, "320486147.ogg"},
	{0x2DA20FECD1097EA4ULL, "6167480_m.fxe"},
	{0x2DA29D061B51EE21ULL, "dae000_mother6177013.cl"},
	{0x2DA29F95A6EAB3EDULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v1.gad"},
	{0x2DA2B859BD1C5D54ULL, "hlac_ht_0_hightown_a2_1_2z.dds"},
	{0x2DA342E407F91A7FULL, "870525633.ogg"},
	{0x2DA35F33FC177F56ULL, "442910293.ogg"},
	{0x2DA3EF9C5BD2428CULL, "hlac_lt_0_lowtown_a2_-3_1x.dds"},
	{0x2DA40E5BD07B4DC2ULL, "aip_rockwraith_base.gda"},
	{0x2DA4B58172EF8677ULL, "6082165_f.fxe"},
	{0x2DA501F779AEEFD5ULL, "follower_banter6169707.cl"},
	{0x2DA56B48AD36C3B1ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_ext_c.impa.gad"},
	{0x2DA5A2289D967D9FULL, "vxht_0_hightown_n_c5i_0_0.vlm"},
	{0x2DA671DEBD30E364ULL, "llac_lt_0_lowtown_a2_-262143_0.dds"},
	{0x2DA673859E57B357ULL, "art\\levels\\ztools\\default\\generic\\0do_basic.phy"},
	{0x2DA696B5ABFB76A9ULL, "vxlt_darktown_44o_0_0.vlm"},
	{0x2DA698C07A513AADULL, "vxht_keep_n_61j_0_0.vlm"},
	{0x2DA69B9DC7DFF0E0ULL, "vxca_caverns2_8ae_0_0.vlm"},
	{0x2DA6FC5E355C74FFULL, "vxlt_darktown_5as_0_0.vlm"},
	{0x2DA78C05DB93693AULL, "304086524.ogg"},
	{0x2DA7D6E59ED33FBEULL, "vxdr_0_deeproads_9gp_0_0.vlm"},
	{0x2DA7F206BDA13546ULL, "zz_suhwan_test6202891.cl"},
	{0x2DA843C93D3057D2ULL, "w_blightlands_26_0.rml"},
	{0x2DA89A6E6A5E2D9AULL, "vxlt_hangedman_n_21o_0_0.vlm"},
	{0x2DA8C3018A64E75BULL, "vxlt_hangedman_n_1hz_0_0.vlm"},
	{0x2DA8DE9A4C7C454BULL, "lt_0_lowtown_a2_n_24_0.rml"},
	{0x2DA91AF6CE6DA6D3ULL, "285576723.ogg"},
	{0x2DA963A2215CFEC6ULL, "art\\vfx\\combat\\vfx_w_off_accuracy_activ_c.mmh"},
	{0x2DAA2A4BD61C041DULL, "vxlt_gallowscourtyard_755_0_0.vlm"},
	{0x2DAA4EFA0FCF912DULL, "zz_dae_debug6184892.cl"},
	{0x2DAA575698AB580FULL, "mag160_feynriel.bnk"},
	{0x2DAA7089998F2EF9ULL, "m_ser221ar_murdrer_estat_nt.bnk"},
	{0x2DAABA164DEED46EULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dssa_0n.dds"},
	{0x2DAAD90222643559ULL, "490844891.ogg"},
	{0x2DAAF906F343DA71ULL, "hlac_lt_shop_-1_1x.dds"},
	{0x2DAB76237B18C5DEULL, "vxlt_gallowscourtyard_7gh_0_0.vlm"},
	{0x2DABA32889C19C2AULL, "vxht_mansion_n_5cl_0_0.vlm"},
	{0x2DABDF6F4C653F04ULL, "gen00fl_merrill6124023.cl"},
	{0x2DAC6C48AD3CE45FULL, "art\\vfx\\combat\\vfx_glb_spi_imp_ext_c.impa.ani"},
	{0x2DACCE0061A583FAULL, "gen00fl_aveline6132812.cl"},
	{0x2DAD6B3673638E7EULL, "vxht_keep_n_6jh_0_0.vlm"},
	{0x2DADA1126AB4A270ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgtb_t.dds"},
	{0x2DADE037F6745D91ULL, "15388233.ogg"},
	{0x2DAE795698AF7B77ULL, "mag160_feynriel.cnv"},
	{0x2DAE8AFC0EC957F6ULL, "nat380_temmerin6120939.cl"},
	{0x2DAF30516E600DD8ULL, "vxlt_warehouse_1e5_0_0.vlm"},
	{0x2DAFDA9E3B811CE4ULL, "mcr371_hostile_fenris6082235.cl"},
	{0x2DB00A3D85C75717ULL, "6040599_m.fxe"},
	{0x2DB03CF8E1621B23ULL, "ht_bartrand_n_3370_lc.mmh"},
	{0x2DB0942CF45515A5ULL, "6107160_m.fxe"},
	{0x2DB0A10EE30DD21DULL, "vxht_0_hightown_a3_bxo_0_0.vlm"},
	{0x2DB102A87EF2DB3CULL, "691853747.ogg"},
	{0x2DB123942D3CA6AFULL, "hlac_w_bonepitmining_a2_0_6y.dds"},
	{0x2DB14058A73B8F1BULL, "art\\vfx\\combat\\vfx_w_off_stagger_activ_c.impa.ani"},
	{0x2DB17F6A8C61F0AFULL, "vxw_woundedcoast_16y_0_0.vlm"},
	{0x2DB19D48AAE74089ULL, "401021449.ogg"},
	{0x2DB1D0DEFA21A78AULL, "6093468_m.fxe"},
	{0x2DB1EB72423328EEULL, "6041324_m.fxe"},
	{0x2DB1F1EF7457A9CCULL, "character_character_p49.psh.fxo.dx11.assembly"},
	{0x2DB24DBE1B804159ULL, "6088781_f.fxe"},
	{0x2DB32BA54C58A114ULL, "400792815.ogg"},
	{0x2DB330B582674298ULL, "vxlt_0_lowtown_a2_n_hf4_0_0.vlm"},
	{0x2DB33F4FD5EE1707ULL, "6121752-mrl_romance_wander2.cut"},
	{0x2DB3546E686CB305ULL, "vxdr_primevaltomb_527_0_1.vlm"},
	{0x2DB38BABD56F12DEULL, "vxht_keep_5i9_0_0.vlm"},
	{0x2DB41B5B5C5F67BDULL, "434490997.ogg"},
	{0x2DB45F85A144D15DULL, "vxlt_foundry_n_w8_0_0.vlm"},
	{0x2DB4F1B17DDD623EULL, "509259532.ogg"},
	{0x2DB50C8535DDF1F3ULL, "6023890_f.fxe"},
	{0x2DB50D7AB9D60960ULL, "vxca_caverns2_8lb_0_0.vlm"},
	{0x2DB50D9A21E697BEULL, "vxlt_0_lowtown_n_ffp_0_0.vlm"},
	{0x2DB5417471FB1C2BULL, "lt_0_lowtown_docks_a3_n_118_0.phy"},
	{0x2DB54BCC1087FDDAULL, "6197207_m.fxe"},
	{0x2DB55B50940ACFC8ULL, "vxw_bonepit_11x_0_0.vlm"},
	{0x2DB56E3B27528D7FULL, "vxlt_darktown_5jr_0_0.vlm"},
	{0x2DB587119CFCA70CULL, "isa111_isabela6075424.cl"},
	{0x2DB5B64A80943514ULL, "gen00fl_merrill6150697.cl"},
	{0x2DB5CB734ABB8BE5ULL, "vxlt_gallowstemplar_n_5t5_0_0.vlm"},
	{0x2DB6502980691FBFULL, "vxdr_primevalentrance_24u_0_0.vlm"},
	{0x2DB7274BF00A36E1ULL, "vxw_woundedcoast_6j_0_0.vlm"},
	{0x2DB74319CD1A9267ULL, "vxdr_0_deeproads_cco_0_0.vlm"},
	{0x2DB78AA2EE2D509CULL, "static_static_p12.psh.fxo"},
	{0x2DB81BB46D413E56ULL, "plt_lgt210pt_mistress_selby.plo"},
	{0x2DB915B5486B6AE3ULL, "mag311_keran6041709.cl"},
	{0x2DB937D1A80F3CECULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutan.msh"},
	{0x2DB93DD1A80F46DEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutan.mmh"},
	{0x2DB949D1A80F5B75ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutan.mao"},
	{0x2DB97C3ECF2BC3FEULL, "vxlt_0_lowtown_a2_eos_0_0.vlm"},
	{0x2DB9F6C3658B2F68ULL, "538122223.ogg"},
	{0x2DBA6339B1B0EE63ULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.dura.ani"},
	{0x2DBB0B9B0D27147FULL, "vxlt_foundry_n_12c_0_1.vlm"},
	{0x2DBB142CCE66F001ULL, "556155651.ogg"},
	{0x2DBB5676BCB44324ULL, "vxht_brothel_n_5hl_0_0.vlm"},
	{0x2DBB580B57F6BD19ULL, "360182034.ogg"},
	{0x2DBB5B35701C2949ULL, "6071907_f.fxe"},
	{0x2DBB8406EDE6F147ULL, "6049642_m.fxe"},
	{0x2DBBB5BEE2F3ADFCULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he1.phy"},
	{0x2DBBC1F22498119AULL, "6095703_m.fxe"},
	{0x2DBC042D7273643AULL, "zz_qun_debug6028238.cl"},
	{0x2DBC3FB6DD1A535BULL, "vxht_mansion_n_3sk_0_0.vlm"},
	{0x2DBC88B413528372ULL, "322319169.ogg"},
	{0x2DBCC49E0E8FD6AAULL, "6197140_m.fxe"},
	{0x2DBDE83889BA8859ULL, "art\\levels\\ztools\\default\\generic\\0ch_basic.phy"},
	{0x2DBDFDF95BDD912BULL, "vxlt_warehouse_88_0_0.vlm"},
	{0x2DBE2CB390226932ULL, "dae011_corff_sta6149642.cl"},
	{0x2DBE4858A7462712ULL, "art\\vfx\\combat\\vfx_w_off_stagger_activ_c.impa.evt"},
	{0x2DBED44B5A5619F5ULL, "pro000cr_bethany_fake.utc"},
	{0x2DBEF3989CA17541ULL, "1043100897.ogg"},
	{0x2DBF59E6FB5B2B7EULL, "vxdr_primevalentrance_31z_0_0.vlm"},
	{0x2DBFB8BF9A9C3BD0ULL, "carpetnew_sa.dds"},
	{0x2DBFCDEE403FE695ULL, "15235595.ogg"},
	{0x2DBFF1477E58092FULL, "vxw_woundedcoast_11x_0_0.vlm"},
	{0x2DC00948005AF095ULL, "6048157_m.fxe"},
	{0x2DC0608062F24FEDULL, "vxca_cavern1_haunted_j43_0_0.vlm"},
	{0x2DC075BC38FF3457ULL, "vxht_mansion2b_334_0_0.vlm"},
	{0x2DC0A8963DC0B1F3ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_c.impa.evt"},
	{0x2DC0B001B15EE2C8ULL, "vxw_woundedcoast_6o_0_0.vlm"},
	{0x2DC0B4211944E3BFULL, "481508504.ogg"},
	{0x2DC0C41AD4BFDD6BULL, "dae300_cullen6135344.cl"},
	{0x2DC18AAED492B9C5ULL, "130792959.ogg"},
	{0x2DC1DA3F269E4666ULL, "6206971_f.fxe"},
	{0x2DC23B65CBEADECBULL, "691362495.ogg"},
	{0x2DC2926421E62A56ULL, "1057646555.ogg"},
	{0x2DC3471D35B326C5ULL, "vxlt_0_lowtown_docks_a3_cno_0_0.vlm"},
	{0x2DC37F57ABCC21E0ULL, "vxca_caverns2_8kc_0_0.vlm"},
	{0x2DC410AECD4FA61AULL, "vxht_bartrand_n_4fq_0_0.vlm"},
	{0x2DC45EA39F6F180FULL, "vxlt_darktown_5t8_0_0.vlm"},
	{0x2DC49F260E655350ULL, "240219271.ogg"},
	{0x2DC538C8907EDB70ULL, "w_sundermount_17_0.mao"},
	{0x2DC53CC8907EE183ULL, "w_sundermount_17_0.mmh"},
	{0x2DC547FFD23CCDD9ULL, "6114810_m.fxe"},
	{0x2DC54AC8907EF9DDULL, "w_sundermount_17_0.msh"},
	{0x2DC56758A74C7AA5ULL, "art\\vfx\\combat\\vfx_w_off_stagger_activ_c.impa.gad"},
	{0x2DC59285A7089642ULL, "758038489.ogg"},
	{0x2DC59928E2004F61ULL, "vxw_woundedcoast_3i_0_0.vlm"},
	{0x2DC5A17C47603FDEULL, "vxlt_0_lowtown_docks_bao_0_0.vlm"},
	{0x2DC5C81CF561C32AULL, "vxlt_hangedman_1pl_0_0.vlm"},
	{0x2DC5FC7FE023691CULL, "static_static_p41.psh.fxo"},
	{0x2DC634EDE32D339FULL, "84616523.ogg"},
	{0x2DC656901366CC96ULL, "mag210_arianni6024100.cl"},
	{0x2DC65BF53BE8D635ULL, "zz_hawkereveal.cut"},
	{0x2DC6705748161424ULL, "vxlt_0_lowtown_docks_b4z_0_0.vlm"},
	{0x2DC6786FBD474DEEULL, "147632872.ogg"},
	{0x2DC6E12197B37B4DULL, "qapath2.arl"},
	{0x2DC73C7453158333ULL, "ser120_ghyslain6104832.cl"},
	{0x2DC7476A4269C586ULL, "vxht_keep_n_6hj_0_0.vlm"},
	{0x2DC77B39B1BBA15AULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.dura.evt"},
	{0x2DC7A1963DC6C434ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_c.impa.gad"},
	{0x2DC7E6906C360C08ULL, "mrl351_pridedemon6165549.cl"},
	{0x2DC82A2C8B0D9F5FULL, "zz_lgt_2_debug6211857.cl"},
	{0x2DC886E26DADCD2AULL, "vxht_mansion_n_5in_0_0.vlm"},
	{0x2DC8F0C0AA5AC246ULL, "hlac_w_bonepitmining_a2_196608_0x.dds"},
	{0x2DC8F497488F3E33ULL, "one130_thug6131629.cl"},
	{0x2DC914AC804827E9ULL, "hlac_lt_0_lowtown_docks_a3_n_0_6y.dds"},
	{0x2DC975CD54D070C7ULL, "1002821904.ogg"},
	{0x2DC9F5E3DFDD09C7ULL, "6044542_m.fxe"},
	{0x2DCA102BF576C64AULL, "follower_banter6151248.cl"},
	{0x2DCAB0253C4B1263ULL, "vxlt_0_lowtown_a2_n_c_hkr_0_0.vlm"},
	{0x2DCAB92CB519B303ULL, "w_encounter3_233_0.msh"},
	{0x2DCAC67AF738CEF8ULL, "gen00fl_aveline6053311.cl"},
	{0x2DCAC72CB519CB5DULL, "w_encounter3_233_0.mmh"},
	{0x2DCACB2CB519D26EULL, "w_encounter3_233_0.mao"},
	{0x2DCB8268138ECC9DULL, "mag291_demon_marethari6166180.cl"},
	{0x2DCBE0769B3E676EULL, "vxca_cavern1_haunted_gi2_0_0.vlm"},
	{0x2DCC5BB7654E4554ULL, "w_encounter2_camp_34_0.phy"},
	{0x2DCC85376CFF2C20ULL, "vxdr_0_deeproads_g30_0_0.vlm"},
	{0x2DCCE3691B380D6EULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.impa.evt"},
	{0x2DCD29F83E9CF3A1ULL, "hf_gen_act2.mor"},
	{0x2DCD374831CB25E9ULL, "6177861_f.fxe"},
	{0x2DCD40A866502785ULL, "261416205.ogg"},
	{0x2DCD9371E323E074ULL, "hlac_lt_0_lowtown_a2_n_65538_2y.dds"},
	{0x2DCDB0963DCB4912ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_c.impa.ani"},
	{0x2DCE012158F4528CULL, "vxlt_0_lowtown_a3_n_c_hqz_0_0.vlm"},
	{0x2DCE7A39B1C1BE4DULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.dura.gad"},
	{0x2DCEA9AE509D1C76ULL, "gen00fl_anders6125367.cl"},
	{0x2DCF9841E3230FB8ULL, "6107312_m.fxe"},
	{0x2DD0386319AAF903ULL, "vxw_blightlands_f2_0_0.vlm"},
	{0x2DD0848F4BA532D0ULL, "6120449_m.fxe"},
	{0x2DD085E08E67F5E7ULL, "vxht_chantry_2tl_0_0.vlm"},
	{0x2DD08634ABF83BB8ULL, "vxlt_warehouse_90_0_0.vlm"},
	{0x2DD099BA50437D21ULL, "361443787.ogg"},
	{0x2DD0C6A27691FFE4ULL, "zz_vault_debug6203350.cl"},
	{0x2DD105495B1568BBULL, "6088220_f.fxe"},
	{0x2DD106C5354255BBULL, "hlac_ht_0_hightown_a3_-65535_0z.dds"},
	{0x2DD1295DC53A6A00ULL, "follower_banter6170755.cl"},
	{0x2DD17D7E832B229BULL, "hlac_ca_caverns1_24421_0y.dds"},
	{0x2DD19855D38C0939ULL, "6107058_m.fxe"},
	{0x2DD1E7DEF1666DE0ULL, "vxht_0_hightown_bal_0_0.vlm"},
	{0x2DD253CD6BDF7529ULL, "lt_0_lowtown_docks_20_0.anb"},
	{0x2DD25E4D712EA920ULL, "vxlt_darktown_4v0_0_0.vlm"},
	{0x2DD2828BBF45BE9AULL, "vxht_bartrand_n_4kr_0_0.vlm"},
	{0x2DD2B2421D25E41FULL, "vxht_brothel_n_59t_0_0.vlm"},
	{0x2DD2DA691B3C6919ULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.impa.gad"},
	{0x2DD2E69F21169F17ULL, "art\\toolset\\tools_gizmo_uhm.msh"},
	{0x2DD2FC9F2116C489ULL, "art\\toolset\\tools_gizmo_uhm.mmh"},
	{0x2DD329D62D470EFCULL, "50798603.ogg"},
	{0x2DD34D1B5AF4ADA2ULL, "vxht_mansion_n_6oo_0_0.vlm"},
	{0x2DD352B7655454FCULL, "w_encounter2_camp_34_0.rml"},
	{0x2DD35EF0089F2EE6ULL, "hlac_ca_cavern1_haunted_24426_2z.dds"},
	{0x2DD3660B43E68A2FULL, "336978142.ogg"},
	{0x2DD4B88AB486AE91ULL, "613714322.ogg"},
	{0x2DD4E3BA8DDC754EULL, "vxlt_darktown_5st_0_0.vlm"},
	{0x2DD51DEED716FC63ULL, "darktown_male_fer_urch_sta6174951.cl"},
	{0x2DD5A437B1908917ULL, "581824328.ogg"},
	{0x2DD5B0D2F63D82DEULL, "vxht_0_hightown_n_cb5_0_0.vlm"},
	{0x2DD5B947345FDE06ULL, "vxht_keep_n_6ck_0_0.vlm"},
	{0x2DD5ED5659AF2A48ULL, "mag130_danzig6016448.cl"},
	{0x2DD5F5DE4D79E9C7ULL, "6067850_f.fxe"},
	{0x2DD6835657E2AB25ULL, "416520474.ogg"},
	{0x2DD738B765581151ULL, "w_encounter2_camp_34_0.msh"},
	{0x2DD746B76558296CULL, "w_encounter2_camp_34_0.mao"},
	{0x2DD752B765583D0FULL, "w_encounter2_camp_34_0.mmh"},
	{0x2DD81C0FFA0258F9ULL, "1006255297.ogg"},
	{0x2DD845B43011C4FBULL, "vxht_keep_n_62q_0_0.vlm"},
	{0x2DD8625C7E709DDEULL, "dr_dungeon_3216.rml"},
	{0x2DD8B1249B642A4CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_5craggyrock.msh"},
	{0x2DD8B7249B64343EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_5craggyrock.mmh"},
	{0x2DD8BFA43741136AULL, "isa311_castillon6094904.cl"},
	{0x2DD8C3249B6448D5ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_5craggyrock.mao"},
	{0x2DD8DB5DA5441658ULL, "6131967_f.fxe"},
	{0x2DD98E05533146ADULL, "gen00fl_isabela6110290.cl"},
	{0x2DD9A857F27C07AAULL, "6199340_m.fxe"},
	{0x2DD9EB691B42A567ULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.impa.ani"},
	{0x2DDA3B54B52646D1ULL, "lt_0_lowtown_docks_a3_n_234_0.mmh"},
	{0x2DDA4754B5265B72ULL, "lt_0_lowtown_docks_a3_n_234_0.mao"},
	{0x2DDA5554B526728FULL, "lt_0_lowtown_docks_a3_n_234_0.msh"},
	{0x2DDADB53D1173A5EULL, "vxht_keep_n_4ak_0_0.vlm"},
	{0x2DDB2CEE5BCB9D08ULL, "zz_followers_banter6157952.cl"},
	{0x2DDC15ACA7C638AAULL, "da2terrain_terrain_p0.psh.fxo"},
	{0x2DDC3CC89091C0A0ULL, "w_sundermount_17_0.rml"},
	{0x2DDCD1F9E81351D6ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_mer_tw.ani"},
	{0x2DDCEA6EC6183123ULL, "vxdr_primevaltomb_3iw_0_0.vlm"},
	{0x2DDDC71014AAC975ULL, "w_bonepitdestroyed_a3_128_0.msh"},
	{0x2DDDC91014AACCEBULL, "w_bonepitdestroyed_a3_128_0.mmh"},
	{0x2DDDD51014AAE088ULL, "w_bonepitdestroyed_a3_128_0.mao"},
	{0x2DDE107422E45365ULL, "lt_0_lowtown_45_0_ll.mmh"},
	{0x2DDE38825244B9DEULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.mmh"},
	{0x2DDE6D09D56219D7ULL, "vxlt_gallowstemplar_6t7_0_0.vlm"},
	{0x2DDE7008EA80C6EAULL, "202527802.ogg"},
	{0x2DDED47F7B1D5428ULL, "6028525_m.fxe"},
	{0x2DDF427E2005C30CULL, "vxw_encounter2_f6_0_0.vlm"},
	{0x2DE02B2EE19E99CFULL, "gen_im_trs_crc_amu_r1.uti"},
	{0x2DE0976CC8519735ULL, "vxca_cavern1_haunted_ege_0_0.vlm"},
	{0x2DE1379C67E3FF06ULL, "6169435_m.fxe"},
	{0x2DE16F004F179864ULL, "mag170_hugh6080761.cl"},
	{0x2DE1735141F3E72AULL, "6118519_f.fxe"},
	{0x2DE1E6D044C10842ULL, "6055116_m.fxe"},
	{0x2DE2BAC0CDDE80FCULL, "vxlt_0_lowtown_a3_n_c_gn0_0_0.vlm"},
	{0x2DE35CCCE0DB0486ULL, "genip_door_black.utp"},
	{0x2DE3A82F15F0D809ULL, "vxw_sundermount_21u_0_0.vlm"},
	{0x2DE42E04974635E2ULL, "art\\vfx\\models\\vfx_lightning_single.msh"},
	{0x2DE4300497463898ULL, "art\\vfx\\models\\vfx_lightning_single.mmh"},
	{0x2DE431B92932F1DAULL, "art\\vfx\\combat\\vfx_r_gn3_embu_act_c.anb"},
	{0x2DE44920C105E2CDULL, "gen_im_trs_qun_two_2hm_r0.uti"},
	{0x2DE454D0052C348AULL, "6028813_m.fxe"},
	{0x2DE4626082F80636ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masb_0n.dds"},
	{0x2DE4836AB96EDD3FULL, "vxht_0_hightown_n_a4p_0_0.vlm"},
	{0x2DE4C14AC7B05F4DULL, "474762874.ogg"},
	{0x2DE55B6D1A02550EULL, "art\\characters\\creatures\\dragon\\animation\\rdy.ani"},
	{0x2DE58EA76548E14EULL, "art\\characters\\creatures\\spider\\animation\\p_tw_v1.evt"},
	{0x2DE7EFDC3EFBB14DULL, "849138448.ogg"},
	{0x2DE809B5F4BF8EB0ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_3y.dds"},
	{0x2DE8D33947171D32ULL, "888218162.ogg"},
	{0x2DE91088717D3125ULL, "6059230_m.fxe"},
	{0x2DE9486FD70EA3D3ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_orsino_chant_ex2.ani"},
	{0x2DE9D9F9E81DE9C7ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_mer_tw.evt"},
	{0x2DE9E33328170666ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_close_c.impa.ani"},
	{0x2DE9F7354A00E598ULL, "hlac_dr_dungeon_3207_2x.dds"},
	{0x2DEA049A074B6CE9ULL, "vxlt_0_lowtown_docks_a3_bs5_0_0.vlm"},
	{0x2DEA5646BB45B52EULL, "dae311ar_tavern.ncs"},
	{0x2DEB1782524F0B9DULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.anb"},
	{0x2DEB735ACB771502ULL, "156241991.ogg"},
	{0x2DEB79C9BCD4B1D5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medb_n.dds"},
	{0x2DEB982B00471AF0ULL, "ht_0_hightown_n_30_0.rml"},
	{0x2DEBAD453A47AACDULL, "300640813.ogg"},
	{0x2DEBCA2B6807E292ULL, "llac_dr_slavers_3214_11.dds"},
	{0x2DEC14C0CF2E749EULL, "837600697.ogg"},
	{0x2DEC85A7654EF0F9ULL, "art\\characters\\creatures\\spider\\animation\\p_tw_v1.gad"},
	{0x2DED0DB8C518A89AULL, "893285627.ogg"},
	{0x2DED1C1CFD972A03ULL, "vxw_blightlands_l4_0_0.vlm"},
	{0x2DEDA448EE18C0E4ULL, "ai_movetypes.gda"},
	{0x2DEDB45B11FBDB8CULL, "vxw_encounter2_a7_0_0.vlm"},
	{0x2DEE1124D67FF6FAULL, "6091708_f.fxe"},
	{0x2DEE48859E951FAFULL, "art\\levels\\ztools\\default\\generic\\0do_basic.mao"},
	{0x2DEE54859E95344CULL, "art\\levels\\ztools\\default\\generic\\0do_basic.mmh"},
	{0x2DEE5A859E953E3EULL, "art\\levels\\ztools\\default\\generic\\0do_basic.msh"},
	{0x2DEECB98D5529EE0ULL, "vxht_0_hightown_bwj_0_0.vlm"},
	{0x2DEF6E2A2811B639ULL, "6175406_m.fxe"},
	{0x2DEFC47A49BF5CCCULL, "vxw_bonepitdestroyed_a3_ld_0_0.vlm"},
	{0x2DEFE52E33E9FFAAULL, "6117419_f.fxe"},
	{0x2DF0107EFCCA66A5ULL, "vxlt_backalley_g_1fn_0_0.vlm"},
	{0x2DF027EA8851DD93ULL, "ave100_ambush_16106192.cl"},
	{0x2DF06784E71925C9ULL, "lt_0_lowtown_docks_a3_n_192_0.mao"},
	{0x2DF07384E7193A62ULL, "lt_0_lowtown_docks_a3_n_192_0.mmh"},
	{0x2DF07584E7193D18ULL, "lt_0_lowtown_docks_a3_n_192_0.msh"},
	{0x2DF0C0F67CE09535ULL, "wttypes_base.gda"},
	{0x2DF0D7833A722071ULL, "ser221_party_comments.bnk"},
	{0x2DF0FAF9E8244010ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_mer_tw.gad"},
	{0x2DF14E46BB4BC607ULL, "dae311ar_tavern.lst"},
	{0x2DF1A16E10EF20D6ULL, "vxlt_0_lowtown_a3_n_c_gvb_0_0.vlm"},
	{0x2DF240A468918527ULL, "zz_vault_debug6203214.cl"},
	{0x2DF2536D1A0CD25FULL, "art\\characters\\creatures\\dragon\\animation\\rdy.evt"},
	{0x2DF277950E3D3A69ULL, "prty_merril_armor_1.dds"},
	{0x2DF27B687DACD889ULL, "372101294.ogg"},
	{0x2DF296A765537947ULL, "art\\characters\\creatures\\spider\\animation\\p_tw_v1.ani"},
	{0x2DF2D30D6030E00CULL, "vxca_caverns1_gk6_0_1.vlm"},
	{0x2DF2F8CCECE0C780ULL, "and_templar_3.utc"},
	{0x2DF3179D6C56B976ULL, "lt_0_lowtown_docks_a3_n_119_0.phy"},
	{0x2DF3B9833A742281ULL, "ser221_party_comments.cnv"},
	{0x2DF41EA3731FFDDCULL, "gen00fl_fenris6128052.cl"},
	{0x2DF45ADC84AFF379ULL, "gen00fl_anders6125740.cl"},
	{0x2DF4816A59DE24E4ULL, "6046486_m.fxe"},
	{0x2DF4988317B8A520ULL, "fen210_hadriana6110384.cl"},
	{0x2DF4FA623D03E240ULL, "vxht_mansion2_1x8_0_0.vlm"},
	{0x2DF57651D0951B75ULL, "6018902_f.fxe"},
	{0x2DF65BD1A61AED46ULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.cess.evt"},
	{0x2DF6AC8E8527FD87ULL, "276079709.ogg"},
	{0x2DF72082AB9B7DBBULL, "531510310.ogg"},
	{0x2DF7646B751F0D3BULL, "6201532_m.fxe"},
	{0x2DF7CB3328231AD7ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_close_c.impa.evt"},
	{0x2DF8195FAC39C6C3ULL, "w_bonepit_138_0.mao"},
	{0x2DF81D5FAC39CD10ULL, "w_bonepit_138_0.mmh"},
	{0x2DF82B5FAC39E52AULL, "w_bonepit_138_0.msh"},
	{0x2DF8D9D5AA619519ULL, "6096226_m.fxe"},
	{0x2DF9346D1A12BC48ULL, "art\\characters\\creatures\\dragon\\animation\\rdy.gad"},
	{0x2DF9424734A313EFULL, "vxdr_dungeon_1mg_0_0.vlm"},
	{0x2DF99E1E77989D99ULL, "llac_lt_backalley_n_0_11.dds"},
	{0x2DFA2D9F529F56E8ULL, "hlac_ht_chantry_2534_43z.dds"},
	{0x2DFA74D335F9E081ULL, "mag101_alain6052777.cl"},
	{0x2DFAFF2EA81F7CCEULL, "6084184_f.fxe"},
	{0x2DFB8DF9EF8D4283ULL, "vxw_blightlands_q7_0_0.vlm"},
	{0x2DFBDAA6BAB3DC5EULL, "llac_ht_playershouse_n_855_32.dds"},
	{0x2DFC263803F1F40CULL, "vxw_encounter2_l4_0_0.vlm"},
	{0x2DFC9AC91DE27FDDULL, "153589295.ogg"},
	{0x2DFCBE9BFDA57EEBULL, "888109151.ogg"},
	{0x2DFD009D6C5EE617ULL, "lt_0_lowtown_docks_a3_n_119_0.msh"},
	{0x2DFD129D6C5F057AULL, "lt_0_lowtown_docks_a3_n_119_0.mao"},
	{0x2DFD169D6C5F0B89ULL, "lt_0_lowtown_docks_a3_n_119_0.mmh"},
	{0x2DFD1F6FD71F077DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_orsino_chant_ex2.gad"},
	{0x2DFD692F8AA9DAF5ULL, "dae300_bran6143479.cl"},
	{0x2DFD72D1A6213291ULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.cess.gad"},
	{0x2DFDCBDBC56975F2ULL, "gen00fl_merrill6122368.cl"},
	{0x2DFDDA0FBA77D47FULL, "hlac_lt_0_lowtown_a2_-131073_0z.dds"},
	{0x2DFE094429078A67ULL, "123573370.ogg"},
	{0x2DFE157AABFD41C1ULL, "vfx_distortionmask_p3.vsh.fxo.dx11.assembly"},
	{0x2DFE1659D376BF82ULL, "art\\characters\\weapons\\shortbows\\wsbwsb104a.msh"},
	{0x2DFE1859D376C338ULL, "art\\characters\\weapons\\shortbows\\wsbwsb104a.mmh"},
	{0x2DFE53F41ECD9A14ULL, "vxht_keep_6kz_0_0.vlm"},
	{0x2DFE8D5EDF6303F7ULL, "vxlt_darktown_6c6_0_0.vlm"},
	{0x2DFECC3328293B60ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_close_c.impa.gad"},
	{0x2DFF3717AD3685DEULL, "vxca_cavern1_haunted_i1p_0_0.vlm"},
	{0x2E00C13842675F51ULL, "6149342_m.fxe"},
	{0x2E00EB5CF4ED2C0BULL, "hlac_lt_gallowscourtyard_n_-2_3x.dds"},
	{0x2E01117924A80FB8ULL, "qun220cr_javaris6071516.cl"},
	{0x2E0112F7DA012D97ULL, "526071262.ogg"},
	{0x2E015C7504BF355CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_gong_0.pwk"},
	{0x2E01861D7DB8E9C1ULL, "dre000_dougal6142196.cl"},
	{0x2E01877F33A323E6ULL, "vxlt_darktown_4zb_0_0.vlm"},
	{0x2E02208D075792C0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgtb_d.dds"},
	{0x2E024F46BB5A3B3CULL, "dae311ar_tavern.are"},
	{0x2E0292046D86729DULL, "6125426_f.fxe"},
	{0x2E02CC3EE658DC6CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow6base.phy"},
	{0x2E0312B2F49C76E9ULL, "gen00fl_anders6125850.cl"},
	{0x2E036C3F2EF9FAC0ULL, "vxht_mansion2_1c9_0_0.vlm"},
	{0x2E03FA4365E49F65ULL, "hlac_lt_0_lowtown_a3_n_c_131073_1z.dds"},
	{0x2E0428900CDE4413ULL, "632435617.ogg"},
	{0x2E0473D1A627533FULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.cess.ani"},
	{0x2E04779904237463ULL, "vxlt_0_lowtown_a2_n_c_hwv_0_0.vlm"},
	{0x2E047810F7A662BAULL, "350327145.ogg"},
	{0x2E047BE247A23E8BULL, "794347313.ogg"},
	{0x2E048AF66CBDDF25ULL, "6068344_m.fxe"},
	{0x2E0514FD15ABC5CAULL, "vxw_woundedcoast_1aa_0_0.vlm"},
	{0x2E05563BAFA7870DULL, "vxdr_primevaltomb_530_0_1.vlm"},
	{0x2E05A6FFAD82A01FULL, "lowtown_male_urchin_sta6164227.cl"},
	{0x2E0606CEB23737FAULL, "vxdr_primevaltomb_2xl_0_0.vlm"},
	{0x2E063812E784A1F6ULL, "64783572.ogg"},
	{0x2E06476D4FBDDA43ULL, "dae200_bran6144583.cl"},
	{0x2E06901D6C4A4743ULL, "282507265.ogg"},
	{0x2E0727871E6B9CB2ULL, "follower_banter6136702.cl"},
	{0x2E075D01C1607E16ULL, "zz_eric_docks.cut"},
	{0x2E077618467942B9ULL, "dae100cr_qunari_warrior2.utc"},
	{0x2E079131FABFE740ULL, "415682950.ogg"},
	{0x2E07A1CFDB467D21ULL, "6195831_m.fxe"},
	{0x2E081A1AF9F6FB47ULL, "vxlt_warehouse_n_14n_0_0.vlm"},
	{0x2E08F4721EAAF1B9ULL, "215496657.ogg"},
	{0x2E09128A5F9D91E4ULL, "6071843_f.fxe"},
	{0x2E099796B2A98E37ULL, "vxht_mansion2b_312_0_0.vlm"},
	{0x2E09C740DFCFB6E3ULL, "vxlt_hangedman_n_26g_0_0.vlm"},
	{0x2E09E5370507017BULL, "6032868_m.fxe"},
	{0x2E0BBFACDE594E3AULL, "6080301_m.fxe"},
	{0x2E0BE4223957C7AAULL, "394660363.ogg"},
	{0x2E0BF6DBEF3D89F3ULL, "30165378.ogg"},
	{0x2E0C19B1862F27C9ULL, "6161634_m.fxe"},
	{0x2E0C34B92954488DULL, "art\\vfx\\combat\\vfx_r_gn3_embu_act_c.mmh"},
	{0x2E0D1CA8BE3D75BDULL, "262813634.ogg"},
	{0x2E0D88F36E512E8CULL, "hlac_ht_keep_n_4519_16z.dds"},
	{0x2E0DC31E7C046563ULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.impa.ani"},
	{0x2E0E04806B8D2E48ULL, "and100_anders6095868.cl"},
	{0x2E0E6E93D5486276ULL, "fex_ruin_wl_b.msh"},
	{0x2E0E7893D5487374ULL, "fex_ruin_wl_b.mmh"},
	{0x2E0EB15234B322F4ULL, "523675943.ogg"},
	{0x2E0FD23FB7ADAD47ULL, "hlac_dr_0_deeproads_5843_3x.dds"},
	{0x2E1101569FCE5643ULL, "6129117_m.fxe"},
	{0x2E1124A609161F4BULL, "vxw_encounter2_camp_s2_0_0.vlm"},
	{0x2E116026B6675C0FULL, "6155697_m.fxe"},
	{0x2E11E4A610F048FDULL, "dae000_madam_lusine6152157.cl"},
	{0x2E12219ABBE4EBCFULL, "llac_ca_caverns2_5802_53.dds"},
	{0x2E1271B3BE94EE26ULL, "vxw_bonepit_1py_0_0.vlm"},
	{0x2E13586A7177948CULL, "6116678_f.fxe"},
	{0x2E1386DA07A1DE4AULL, "vxw_woundedcoast_1lb_0_0.vlm"},
	{0x2E1460E22DEE10EFULL, "6116199_m.fxe"},
	{0x2E1478ABA42D507AULL, "vxdr_primevaltomb_2wm_0_0.vlm"},
	{0x2E14E96EBC378DA7ULL, "320100269.ogg"},
	{0x2E1529248A4080CEULL, "87093757.ogg"},
	{0x2E153864CF50ECA4ULL, "zz_qun_debug6045394.cl"},
	{0x2E163636E5A30BF3ULL, "art\\vfx\\combat\\material__23.mao"},
	{0x2E168BF7EBED13C7ULL, "vxlt_warehouse_n_19i_0_0.vlm"},
	{0x2E16E0B228FE50ACULL, "6118367_m.fxe"},
	{0x2E1785182163D949ULL, "6087195_f.fxe"},
	{0x2E17992B006C5758ULL, "ht_0_hightown_n_30_0.anb"},
	{0x2E17AE573D573AFAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow4base.phy"},
	{0x2E17DF19904B32B2ULL, "hlac_lt_gallowstemplar_2383_12y.dds"},
	{0x2E180EAAA2577D0BULL, "239749115.ogg"},
	{0x2E1822CDC310B2E8ULL, "vxlt_0_lowtown_a2_n_gz3_0_0.vlm"},
	{0x2E191D075D165B37ULL, "vxlt_gallowstemplar_6s2_0_0.vlm"},
	{0x2E1950B3CF5964E0ULL, "dae000_pc_home_ambients6208303.cl"},
	{0x2E19A8D49EA31679ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowarch.msh"},
	{0x2E19B2D49EA32767ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowarch.mmh"},
	{0x2E19B6D49EA32DB4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowarch.mao"},
	{0x2E19F81CA3B80E8BULL, "6174911_m.fxe"},
	{0x2E1A3EFBCC838366ULL, "6115551_m.fxe"},
	{0x2E1A9016D4BE4BBCULL, "vxlt_0_lowtown_a2_n_c_hjj_0_0.vlm"},
	{0x2E1AAC242E02C1DFULL, "942855877.ogg"},
	{0x2E1ABD69A96B41C7ULL, "t3_brt_b03.tnt"},
	{0x2E1ADB1E7C0F185AULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.impa.evt"},
	{0x2E1AE8287BCFD678ULL, "vxca_caverns2_7ca_0_1.vlm"},
	{0x2E1B3EFE8558E292ULL, "vxht_keep_57w_0_0.vlm"},
	{0x2E1BAE4BCAA3539BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf02.mmh"},
	{0x2E1BCC4BCAA38685ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf02.msh"},
	{0x2E1BDBF056DF70B4ULL, "hf_leliana.mor"},
	{0x2E1BEFB50B317E24ULL, "hlac_w_woundedcoast_196616_0x.dds"},
	{0x2E1BF70BC5AC2269ULL, "6085442_m.fxe"},
	{0x2E1C0B5A52DA99EAULL, "w_woundedcoast_114_0.mao"},
	{0x2E1C0F5A52DAA039ULL, "w_woundedcoast_114_0.mmh"},
	{0x2E1C195A52DAB127ULL, "w_woundedcoast_114_0.msh"},
	{0x2E1C97062F83FE86ULL, "vxw_sundermount_30q_0_1.vlm"},
	{0x2E1CACF6A1D29A16ULL, "rdr111dg_martin6054263.cl"},
	{0x2E1D2FE07256A793ULL, "6148729_m.fxe"},
	{0x2E1D4DB70D5B9562ULL, "813651709.ogg"},
	{0x2E1D6526F43A480DULL, "vxht_0_hightown_a3_b9l_0_0.vlm"},
	{0x2E1E05D23B4B6623ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_appear.ani"},
	{0x2E1E58BC249631B6ULL, "gen00fl_isabela6101473.cl"},
	{0x2E1EF4F599197389ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_right.ani"},
	{0x2E1F7881CEE1D105ULL, "6179561_m.fxe"},
	{0x2E1F8AE015A57A52ULL, "372961471.ogg"},
	{0x2E1F908BC77EC3B5ULL, "218603025.ogg"},
	{0x2E1F9682FB0C37CBULL, "vxw_encounter2_camp_n5_0_0.vlm"},
	{0x2E201CBBCBE03A83ULL, "follower_banter6160733.cl"},
	{0x2E208D2BB594D4ECULL, "gen00fl_isabela6089586.cl"},
	{0x2E20E390B08B06A6ULL, "vxw_bonepit_1uz_0_0.vlm"},
	{0x2E20F9B3D7906A08ULL, "6097211_m.fxe"},
	{0x2E2110CEE544B0FDULL, "vxlt_0_lowtown_n_e2t_0_0.vlm"},
	{0x2E213249D8A9480AULL, "2684187.ogg"},
	{0x2E21DA1E7C15354DULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.impa.gad"},
	{0x2E222EE8D6BA0D98ULL, "vxca_caverns2prc_9uq_0_0.vlm"},
	{0x2E22646BEDE9F479ULL, "dae000ip_crowd_em1_talk5.utp"},
	{0x2E22E0337C4DFD9DULL, "mag170_hugh6080696.cl"},
	{0x2E23596314004376ULL, "det_smo_wispy.mao"},
	{0x2E237604387F015FULL, "155308360.ogg"},
	{0x2E23E92C040F241AULL, "dae000_merchant_generic_hm6172647.cl"},
	{0x2E2403F399371F2EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsblock.phy"},
	{0x2E2445796A9B376AULL, "hlac_ht_mansion_n_3370_0y.dds"},
	{0x2E244DEBE31F31CBULL, "vxca_caverns1_gke_0_1.vlm"},
	{0x2E260FF5991FC00BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_right.gad"},
	{0x2E2634A88E54A99AULL, "78507528.ogg"},
	{0x2E26AAFAC3BBE7E3ULL, "vxlt_hangedman_n_28i_0_0.vlm"},
	{0x2E26EAF1D15D74B7ULL, "vxca_caverns1_mines_j7d_0_0.vlm"},
	{0x2E26F8B8D5C8AE78ULL, "6201508_m.fxe"},
	{0x2E2701529573BADEULL, "vxdr_primevaltomb_8nt_0_0.vlm"},
	{0x2E274D580C070DF1ULL, "670067078.ogg"},
	{0x2E2756FE4362EABFULL, "6201271_m.fxe"},
	{0x2E2765BAD48C2B78ULL, "6135238_m.fxe"},
	{0x2E286427D36DD245ULL, "vxlt_0_lowtown_docks_a3_c3h_0_0.vlm"},
	{0x2E28FC9765373DA6ULL, "hlac_ht_0_hightown_131072_0y.dds"},
	{0x2E295C49C6AAD03DULL, "gen00fl_merrill6129603.cl"},
	{0x2E29BB8A782DEADEULL, "and230_mage_girl6109107.cl"},
	{0x2E2A7425276612EDULL, "303106615.ogg"},
	{0x2E2A81385EE8B959ULL, "6205017_m.fxe"},
	{0x2E2AE7C9A90D5AC0ULL, "hlac_ca_cavern1_haunted_24425_2y.dds"},
	{0x2E2AFED0AEE37629ULL, "hlac_lt_0_lowtown_n_131071_5y.dds"},
	{0x2E2AFEF447E1E492ULL, "vxht_0_hightown_a3_an5_0_0.vlm"},
	{0x2E2B20294B5EABFFULL, "gen00fl_varric6081257.cl"},
	{0x2E2B2FAE9EA21230ULL, "hlac_lt_0_lowtown_a3_n_c_65537_1x.dds"},
	{0x2E2BE08A3EC6A306ULL, "one120_athenril6088889.cl"},
	{0x2E2BE94F690866E2ULL, "6118444_m.fxe"},
	{0x2E2C0335403F2492ULL, "ave121_brennan6133088.cl"},
	{0x2E2C1DD23B57CC1AULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_appear.evt"},
	{0x2E2C5798D40F095CULL, "839498892.ogg"},
	{0x2E2CDD1609856D66ULL, "vxlt_darktown_4ke_0_0.vlm"},
	{0x2E2D26528CB9C04BULL, "hlac_lt_0_lowtown_docks_n_-65538_9y.dds"},
	{0x2E2DCB817F3EBA9FULL, "hlac_lt_0_lowtown_docks_a3_-65537_3x.dds"},
	{0x2E2DDCB13FAF14AFULL, "6068456_m.fxe"},
	{0x2E2E23B1974C40ABULL, "hf_bethany_h1s3.mor"},
	{0x2E2ED64A81C46D20ULL, "765301820.ogg"},
	{0x2E2F0E4F749A4BEEULL, "mag210_arianni6030466.cl"},
	{0x2E3043DA24EBBA86ULL, "210973846.ogg"},
	{0x2E3055139526DD14ULL, "follower_banter6148752.cl"},
	{0x2E30D7AE175A5EDCULL, "vxw_bonepitmining_a2_246_0_0.vlm"},
	{0x2E30E34661902297ULL, "vxdr_primevaltomb_7su_0_0.vlm"},
	{0x2E31ACDA9454090AULL, "vxlt_backalley_g_1u2_0_2.vlm"},
	{0x2E31D9CD15D78514ULL, "837004792.ogg"},
	{0x2E3200CD86C0B9C6ULL, "vxht_mansion_n_59r_0_0.vlm"},
	{0x2E320D7824DC14E7ULL, "474709325.ogg"},
	{0x2E331CD23B5DE90DULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_appear.gad"},
	{0x2E334210D763CA58ULL, "qun160_vashothlead6055571.cl"},
	{0x2E3383904131FEDAULL, "vxht_0_hightown_a3_c0m_0_0.vlm"},
	{0x2E3399A708B3C9B5ULL, "6148899_m.fxe"},
	{0x2E33C46C0CEA81ACULL, "6112167_m.fxe"},
	{0x2E34C35D6A46DFCAULL, "6093746_f.fxe"},
	{0x2E3517E5B93B11D3ULL, "6218952_m.fxe"},
	{0x2E351C512E6F9744ULL, "hlac_ht_chantry_n_2534_17z.dds"},
	{0x2E352EDB5DD24407ULL, "mag320_orsino6071204.cl"},
	{0x2E35E0E986FC30D2ULL, "6126434_m.fxe"},
	{0x2E364EB3BAD433E2ULL, "lt_gallowscourtyard_37_0_lc.mmh"},
	{0x2E36D7B1EC20EABDULL, "w_bonepitdestroyed_a3_16_0_lc.mmh"},
	{0x2E371E0411E8DAD0ULL, "976950075.ogg"},
	{0x2E37404C08BABA46ULL, "6155649_m.fxe"},
	{0x2E376483807ACA3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_af_transition.ani"},
	{0x2E37C904ACD5C17DULL, "zz_fol_debug6066153.cl"},
	{0x2E38AB1D54ACEE69ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robc_d.dds"},
	{0x2E39245FAC71702BULL, "w_bonepit_138_0.phy"},
	{0x2E39276131773448ULL, "gen00fl_varric6076571.cl"},
	{0x2E3975C61733F7DBULL, "21881795.ogg"},
	{0x2E3A14868311FA75ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magb_0n.dds"},
	{0x2E3A84D592292F08ULL, "6027663_m.fxe"},
	{0x2E3B3EF532E8DF30ULL, "vxht_bartrand_n_64o_0_0.vlm"},
	{0x2E3B581527404FEFULL, "r_ripple_01.mao"},
	{0x2E3B7620CB2CF232ULL, "vxlt_0_lowtown_docks_a3_cv4_0_0.vlm"},
	{0x2E3B91C190F01722ULL, "art\\vfx\\models\\vfx_sphere_2sd.msh"},
	{0x2E3B93C190F01AD8ULL, "art\\vfx\\models\\vfx_sphere_2sd.mmh"},
	{0x2E3BF2C9E0958347ULL, "889846365.ogg"},
	{0x2E3C209B79DD0DD3ULL, "6121232_m.fxe"},
	{0x2E3C39B6A326AF2BULL, "hlac_w_bonepitmining_a2_6_0y.dds"},
	{0x2E3C7A3CDEF868CBULL, "vxw_encounter2_camp_d7_0_0.vlm"},
	{0x2E3DDD86EC624EE1ULL, "vxw_bonepitmining_a2_1xx_0_0.vlm"},
	{0x2E3E71838080FF7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_af_transition.gad"},
	{0x2E3F07BB6A7D4353ULL, "dae000_paivel6111949.cl"},
	{0x2E3F5E9936DCA8B6ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.impa.evt"},
	{0x2E3FB00F17361B21ULL, "hlac_dr_0_deeproads_5845_3x.dds"},
	{0x2E3FC295D72AD52CULL, "hlac_lt_0_lowtown_a2_n_c_65535_5y.dds"},
	{0x2E404F15539F6CECULL, "pro000st_aveline_gallows.stg"},
	{0x2E40520737E8C0D0ULL, "817284741.ogg"},
	{0x2E4080A5F8CA075BULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.dura.evt"},
	{0x2E40C823087F208EULL, "zz_followers_banter6157837.cl"},
	{0x2E40CD7FB05B2D29ULL, "vxlt_warehouse_n_ql_0_0.vlm"},
	{0x2E40D592FB0D482DULL, "vxca_caverns1_cmt_0_0.vlm"},
	{0x2E410E5445DDFF41ULL, "qcr220_grey_wardens6135915.cl"},
	{0x2E417BAEEA86ED16ULL, "vxw_woundedcoast_c2_0_0.vlm"},
	{0x2E41A86FF2B52EFDULL, "6124048_m.fxe"},
	{0x2E41F56D3328175AULL, "vxht_0_hightown_a3_c5n_0_0.vlm"},
	{0x2E4326AC0354E7CDULL, "6111321_m.fxe"},
	{0x2E4389330F70BDFAULL, "vxht_playershouse_n_1vn_0_0.vlm"},
	{0x2E438DFF9E2037ABULL, "6137709_m.fxe"},
	{0x2E4392B04E20E4FDULL, "gen_im_npc_arm_hum_cht_rdrd.uti"},
	{0x2E43A33EE6902D37ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow6base.mmh"},
	{0x2E43A73EE6903444ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow6base.mao"},
	{0x2E43B93EE69052A9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow6base.msh"},
	{0x2E43F9F92D4B7B5DULL, "30604435.ogg"},
	{0x2E443E88D3F8B194ULL, "vxlt_0_lowtown_e17_0_0.vlm"},
	{0x2E4452C678F24952ULL, "6123734_m.fxe"},
	{0x2E44DC2651290C80ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medh_0s.dds"},
	{0x2E454C838086DE9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_af_transition.evt"},
	{0x2E459BD49EC83A9CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowarch.phy"},
	{0x2E462490C2083E9DULL, "vxca_cavern1_haunted_id8_0_0.vlm"},
	{0x2E46402994D50F2BULL, "783521621.ogg"},
	{0x2E46559936E2B841ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.impa.gad"},
	{0x2E468C0FC4EB21D0ULL, "llac_lt_0_lowtown_docks_a3_-2_4.dds"},
	{0x2E46C5DC37AD0CFAULL, "548320000.ogg"},
	{0x2E472A08E5C4DF34ULL, "vxht_keep_4bk_0_0.vlm"},
	{0x2E47395535F93B30ULL, "556964658.ogg"},
	{0x2E4776530C503B14ULL, "lgt200_eau_tome6171794.cl"},
	{0x2E4779A5F8D0199CULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.dura.gad"},
	{0x2E478A78B9D4293DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_vael_0d.dds"},
	{0x2E47E2AE2BCE1592ULL, "vxht_0_hightown_a3_ap7_0_0.vlm"},
	{0x2E486010065E4998ULL, "hlac_dr_slavers_3214_3y.dds"},
	{0x2E487D86DD2C8CD2ULL, "dae000_bodhan6129122.cl"},
	{0x2E487EAF1E9FAF21ULL, "6092240_m.fxe"},
	{0x2E49192ABA773D12ULL, "lgt110_gxa_vgk_silver_order6117021.cl"},
	{0x2E49802EE766084EULL, "hf_macha.mor"},
	{0x2E49DAD4C0C6CEDAULL, "978225607.ogg"},
	{0x2E49DBEE849AE160ULL, "6075164_f.fxe"},
	{0x2E49E7FDBD230AB2ULL, "vxlt_0_lowtown_docks_a3_cy5_0_0.vlm"},
	{0x2E4A3D1A09C4F79BULL, "dae000cr_grand_cleric6148146.cl"},
	{0x2E4A52673017D0BAULL, "hlac_w_encounter3_-65539_0x.dds"},
	{0x2E4AAAB21F685D11ULL, "hf_lgt_angelique.mor"},
	{0x2E4AE7421B9CDFE6ULL, "hlac_w_encounter3_-262144_0y.dds"},
	{0x2E4AF8EDFEE89679ULL, "gen_im_npc_arm_hum_cht_cotb.uti"},
	{0x2E4AFC5289CA6A38ULL, "zz_vrc_debug6200163.cl"},
	{0x2E4B108F1A85DC44ULL, "811683566.ogg"},
	{0x2E4B14E8032D6764ULL, "hlac_w_woundedcoast_0_10z.dds"},
	{0x2E4C4FB4D8F69BFEULL, "vxht_keep_n_63c_0_0.vlm"},
	{0x2E4C7730CBADDDBFULL, "hla_ht_0_hightown_65534_0x.dds"},
	{0x2E4C8CA07C8E3558ULL, "mrl150_merrill6071920.cl"},
	{0x2E4D569936E8D8EFULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.impa.ani"},
	{0x2E4D7455BF53A165ULL, "zz_followers_banter6159333.cl"},
	{0x2E4D89D6F70E8786ULL, "vxlt_undercityencounter_124_0_1.vlm"},
	{0x2E4DEBEBB4505FFBULL, "hlac_w_woundedcoast_-131064_0x.dds"},
	{0x2E4E0FA96490799EULL, "495979564.ogg"},
	{0x2E4E489E2B8B9E02ULL, "vxlt_0_lowtown_a2_n_e1r_0_0.vlm"},
	{0x2E4E78A5F8D6369AULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.dura.ani"},
	{0x2E4E8E57F4928F5EULL, "6048416_f.fxe"},
	{0x2E4EC16F3F72FE06ULL, "442801115.ogg"},
	{0x2E4ED00E8A53F40EULL, "mag121_viveka6156585.cl"},
	{0x2E4ED199548C0228ULL, "819717102.ogg"},
	{0x2E4F02FD5A484653ULL, "vxht_bartrand_n_5vq_0_0.vlm"},
	{0x2E4F5D5AD283B060ULL, "mag170_thrask6054037.cl"},
	{0x2E4FBE1DF1C2FDF2ULL, "isa311_isabela6095637.cl"},
	{0x2E4FBEEE6ECEB085ULL, "zz_fen_debug6134386.cl"},
	{0x2E4FDC7BE4C6EEF6ULL, "qaisnf2.arl"},
	{0x2E4FE932FDFD3F34ULL, "art\\vfx\\materials\\vfx_entropic.mao"},
	{0x2E4FEFA8C858C1E0ULL, "darktown_fem_elf_sta6174534.cl"},
	{0x2E502D0D88A0D412ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\dieloop.gad"},
	{0x2E505615EE5F8D15ULL, "hlac_w_bonepit_65540_0y.dds"},
	{0x2E5154C0905F688BULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.impa.evt"},
	{0x2E515577809C259BULL, "555961712.ogg"},
	{0x2E516D5342AB2C24ULL, "166491942.ogg"},
	{0x2E51B7D44FFAA8BFULL, "vxht_keep_n_55o_0_0.vlm"},
	{0x2E51D460151F6135ULL, "qun120cr_javaris6043230.cl"},
	{0x2E524C6A3C715205ULL, "6000768_m.fxe"},
	{0x2E5357D28D107955ULL, "682010721.ogg"},
	{0x2E543A0D87DC271BULL, "6173873_m.fxe"},
	{0x2E54B59A5496D4B6ULL, "vxht_mansion2_1tj_0_0.vlm"},
	{0x2E5539A916DB8307ULL, "llac_dr_slavers_3217_2.dds"},
	{0x2E5559A1A0E3398EULL, "mer221_sabin6047688.cl"},
	{0x2E558870B40F7EEEULL, "one110_tree_minder6140859.cl"},
	{0x2E55D8199FAD1348ULL, "vxw_bonepitmining_a2_1uz_0_0.vlm"},
	{0x2E55FBEE0087E602ULL, "316393283.ogg"},
	{0x2E560F8F44D32EEAULL, "hlac_lt_backalley_g_1_3x.dds"},
	{0x2E56120B328A7AC9ULL, "vxht_mansion2b_1qr_0_0.vlm"},
	{0x2E5617AABFE621FEULL, "dae150_radha6148899.cl"},
	{0x2E56F123B882A196ULL, "6047404_f.fxe"},
	{0x2E56FBEF1E0590A4ULL, "hlac_lt_0_lowtown_docks_n_-131073_0y.dds"},
	{0x2E5711E99D016C88ULL, "cd_one110_pryce.wbd"},
	{0x2E571F9B30FB3638ULL, "369495598.ogg"},
	{0x2E574CA80FC2F655ULL, "hlac_ca_caverns2_5802_1y.dds"},
	{0x2E5776ECF3188707ULL, "hc_cloud.mao"},
	{0x2E57A48731F1793DULL, "316264951.ogg"},
	{0x2E57DD6468699703ULL, "6084739_m.fxe"},
	{0x2E57F4719720E514ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_master_sa.dds"},
	{0x2E58030D3956CF99ULL, "vxlt_warehouse_15x_0_0.vlm"},
	{0x2E58533A216D0E5CULL, "gen00fl_carver6108931.cl"},
	{0x2E586DC09065B24CULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.impa.gad"},
	{0x2E5883B31B9CE4D2ULL, "vxht_bartrand_n_5vt_0_0.vlm"},
	{0x2E58F562670C43A3ULL, "vxlt_0_lowtown_docks_c4h_0_0.vlm"},
	{0x2E59863A1431FFCFULL, "6143679_f.fxe"},
	{0x2E59D6A52CD05165ULL, "975472791.ogg"},
	{0x2E5A03FD39C0ED0DULL, "vxlt_0_lowtown_n_f66_0_1.vlm"},
	{0x2E5A720F222DF43BULL, "6085804_m.fxe"},
	{0x2E5B27B71B07D6BEULL, "vxw_bonepit_pb_0_0.vlm"},
	{0x2E5B566C3D5F71E0ULL, "6106680_f.fxe"},
	{0x2E5C4AE820FB4093ULL, "916635467.ogg"},
	{0x2E5C4DF3CE3185D8ULL, "llac_lt_darktown_-1_15.dds"},
	{0x2E5C85D7F14B3829ULL, "vxdr_0_deeproads_g2z_0_0.vlm"},
	{0x2E5CC9102FF934A6ULL, "412071854.ogg"},
	{0x2E5D74DA4C3E5ED3ULL, "vxht_bartrand_n_5yp_0_0.vlm"},
	{0x2E5DC1097E9DC743ULL, "811924642.ogg"},
	{0x2E5E5AF2CE422A25ULL, "6202287_m.fxe"},
	{0x2E5E5F68CE731E16ULL, "vxw_woundedcoast_u4_0_0.vlm"},
	{0x2E5E65A3E58C3E9CULL, "lt_0_lowtown_docks_a3_n_179_0.mao"},
	{0x2E5E71A3E58C533FULL, "lt_0_lowtown_docks_a3_n_179_0.mmh"},
	{0x2E5E77A3E58C5DE1ULL, "lt_0_lowtown_docks_a3_n_179_0.msh"},
	{0x2E5EADFC37CB33B4ULL, "hlac_ca_cavern1_haunted_24423_4y.dds"},
	{0x2E5EC5100590582DULL, "6167298_m.fxe"},
	{0x2E5F6CC0906BCF4AULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.impa.ani"},
	{0x2E5F733A711D2089ULL, "6042249_m.fxe"},
	{0x2E5FE9232DC4B5A4ULL, "vxw_woundedcoast_fs_0_0.vlm"},
	{0x2E600EFA21EF34C5ULL, "vxht_playershouse_n_1sf_0_0.vlm"},
	{0x2E605B1406245C27ULL, "6096350_m.fxe"},
	{0x2E60665B6357949EULL, "vxht_0_hightown_a3_c81_0_0.vlm"},
	{0x2E606868A54A8FCBULL, "vxlt_0_lowtown_a3_n_c_gmz_0_0.vlm"},
	{0x2E6138296507D46BULL, "6085048_m.fxe"},
	{0x2E618166FEEC22A5ULL, "fen000pt_gift_sword.ncs"},
	{0x2E618F2193F80DA3ULL, "6120752_m.fxe"},
	{0x2E62617FB95112F5ULL, "gen00fl_merrill6109551.cl"},
	{0x2E62994BF506BD58ULL, "vxht_bartrand_n_4ab_0_0.vlm"},
	{0x2E62CF1EBB50DA0CULL, "ht_0_hightown_n_43_0_ll.mmh"},
	{0x2E62CF4BCADF8ED0ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf02.phy"},
	{0x2E635387A33F7C36ULL, "w_blightlands_206_0.mao"},
	{0x2E635F87A33F90D5ULL, "w_blightlands_206_0.mmh"},
	{0x2E636187A33F944BULL, "w_blightlands_206_0.msh"},
	{0x2E63AB7ABC667CEBULL, "royalarcher_woundarrow.dds"},
	{0x2E63B2C406E880A6ULL, "156937829.ogg"},
	{0x2E6402F3996D01F1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsblock.mmh"},
	{0x2E640EF3996D1592ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsblock.mao"},
	{0x2E641CF3996D2DAFULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsblock.msh"},
	{0x2E64228876C27972ULL, "fadesky.psh.fxo"},
	{0x2E642F9929824917ULL, "hlac_lt_0_lowtown_docks_a3_-196610_4x.dds"},
	{0x2E644A9776552835ULL, "vxlt_undercitydungeon_196_0_0.vlm"},
	{0x2E6479941D641AE4ULL, "lt_0_lowtown_docks_n_212_0.msh"},
	{0x2E6483E824809349ULL, "vxht_mansion2b_1tq_0_0.vlm"},
	{0x2E648B941D64394DULL, "lt_0_lowtown_docks_n_212_0.mao"},
	{0x2E648F941D643F86ULL, "lt_0_lowtown_docks_n_212_0.mmh"},
	{0x2E64D595C98136CBULL, "vxlt_0_lowtown_docks_a3_d0a_0_1.vlm"},
	{0x2E65180D88B30CB8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\dieloop.ani"},
	{0x2E66310478E1336DULL, "follower_banter6165354.cl"},
	{0x2E66368D3F8B6A06ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rockblend_sa.dds"},
	{0x2E667959040B869BULL, "lt_0_lowtown_docks_a3_n_254_0.mmh"},
	{0x2E668559040B9B38ULL, "lt_0_lowtown_docks_a3_n_254_0.mao"},
	{0x2E669759040BB985ULL, "lt_0_lowtown_docks_a3_n_254_0.msh"},
	{0x2E669784C8BD6E64ULL, "vxdr_primevaltomb_7lk_0_0.vlm"},
	{0x2E669B101E76E1D6ULL, "489792311.ogg"},
	{0x2E66F5900D92FD52ULL, "vxht_bartrand_n_5yu_0_0.vlm"},
	{0x2E672C52A292E321ULL, "hlac_ht_mansion2_818_49x.dds"},
	{0x2E6730038A5DD129ULL, "hlac_lt_undercitydungeon_-1_33z.dds"},
	{0x2E675F18564812A0ULL, "lowtown_male_alienage1_sta6162745.cl"},
	{0x2E67673F59025C23ULL, "vxlt_0_lowtown_docks_c1k_0_0.vlm"},
	{0x2E689955E4CF22EAULL, "928014167.ogg"},
	{0x2E6901031D77020CULL, "6139704_m.fxe"},
	{0x2E695A9CF45D3D66ULL, "hlac_dr_primevaltomb_658_9x.dds"},
	{0x2E69C9554BDEFE11ULL, "6134614_m.fxe"},
	{0x2E6A96F59F7F96E0ULL, "hlac_ca_caverns2_5802_54z.dds"},
	{0x2E6AAFC117598D38ULL, "hlac_w_encounter2_camp_131070_0z.dds"},
	{0x2E6AE76C5A6827FAULL, "aip_pridedemon_poison.gda"},
	{0x2E6BC7FD728B9220ULL, "6148393_m.fxe"},
	{0x2E6C4A629CE3312BULL, "861177250.ogg"},
	{0x2E6C83D6D6FBBDFCULL, "6033569_m.fxe"},
	{0x2E6C8B4B69133A9DULL, "100544933.ogg"},
	{0x2E6CD145C0693696ULL, "vxw_woundedcoast_r5_0_0.vlm"},
	{0x2E6CDA812C6BA343ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupb.msh"},
	{0x2E6CE8812C6BBA9DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupb.mmh"},
	{0x2E6CEC812C6BC1AEULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupb.mao"},
	{0x2E6D3B053B5AF6ADULL, "vxw_encounter2_fs_0_0.vlm"},
	{0x2E6D41F3B2123089ULL, "follower_banter6156024.cl"},
	{0x2E6D55AFFE10D25CULL, "qun211_braggart6182761.cl"},
	{0x2E6D5E36FB5A9576ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_c.mmh"},
	{0x2E6D6EECBD9DEBB1ULL, "llac_lt_0_lowtown_docks_a3_n_-4_6.dds"},
	{0x2E6DF705EEEA1025ULL, "vxw_bonepitmining_a2_25x_0_0.vlm"},
	{0x2E6DFEC2B3FA7CE6ULL, "433413978.ogg"},
	{0x2E6E3C86997E2987ULL, "vxlt_gallowscourtyard_n_74g_0_0.vlm"},
	{0x2E6E4B75BC8A698EULL, "hlac_lt_undercitydungeon_-1_49x.dds"},
	{0x2E6E5B001FBACE24ULL, "vxw_woundedcoast_cp_0_0.vlm"},
	{0x2E6E80D713E54D45ULL, "vxht_playershouse_n_1ve_0_0.vlm"},
	{0x2E6F1C89D0AF58DDULL, "mag100_cullen6082289.cl"},
	{0x2E6F8472A8C6CEF4ULL, "6195283_m.fxe"},
	{0x2E6FA522E12FD3C6ULL, "vxw_encounter2_camp_rh_0_0.vlm"},
	{0x2E70446BE8D381F0ULL, "603075517.ogg"},
	{0x2E70927A23BBEF4CULL, "qun160_vashothlead6053915.cl"},
	{0x2E70B08A26982A06ULL, "6170086_m.fxe"},
	{0x2E70B2A5D406A9C6ULL, "6070171_m.fxe"},
	{0x2E70E4CEDDFC229FULL, "vxht_0_hightown_a2_caf_0_0.vlm"},
	{0x2E70FB2886AF2FC7ULL, "711122647.ogg"},
	{0x2E719954388305B6ULL, "vxht_mansion2_1bl_0_0.vlm"},
	{0x2E72187C704EED85ULL, "lgt000cr_hate_qunari.utc"},
	{0x2E72594EC0945D2CULL, "474324240.ogg"},
	{0x2E72819F97B31697ULL, "gen00fl_merrill6128978.cl"},
	{0x2E728ABAC46DA1FEULL, "lt_0_lowtown_a3_n_c_33_0.rml"},
	{0x2E72A7573DA3DA23ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow4base.msh"},
	{0x2E72B5573DA3F27DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow4base.mmh"},
	{0x2E72B9573DA3F88EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow4base.mao"},
	{0x2E72E9C6247F5319ULL, "gen00fl_anders6103731.cl"},
	{0x2E738ACFDEA91C3DULL, "6161573_m.fxe"},
	{0x2E73A144A328787EULL, "zz_one_debug6025817.cl"},
	{0x2E7401E20F706206ULL, "clf260_harley6069794.cl"},
	{0x2E740547AB45820BULL, "993335061.ogg"},
	{0x2E741BE381091F23ULL, "vxdr_slavers_1jd_0_0.vlm"},
	{0x2E7468D63509A9F4ULL, "vxlt_0_lowtown_n_er9_0_0.vlm"},
	{0x2E749E0E2F2D2A2AULL, "vxca_caverns1_mines_ec2_0_0.vlm"},
	{0x2E7568018CCB5D42ULL, "gen_im_wep_mag_sta_act3_03.uti"},
	{0x2E75715D38EAFA95ULL, "6084693_f.fxe"},
	{0x2E760B566AE251AAULL, "vxlt_hangedman_2u8_0_0.vlm"},
	{0x2E7644ADFAD55D9EULL, "clf260_harley6088759.cl"},
	{0x2E7676ABFD94D412ULL, "vxdr_0_deeproads_boz_0_1.vlm"},
	{0x2E7694A527D70961ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_bed_to_knife.ani"},
	{0x2E76BA33D45DE343ULL, "w_bonepitmining_a2_30_0_lc.mmh"},
	{0x2E76DF0D213F2EC4ULL, "hlac_lt_hangedman_n_-1_17x.dds"},
	{0x2E76FA88A6CD0D76ULL, "w_bonepit_139_0.phy"},
	{0x2E7744AFC70C051BULL, "hlac_dr_primevaltomb_659_8y.dds"},
	{0x2E7753EBCD1CDA4CULL, "mag360_alain6044622.cl"},
	{0x2E783C74A708502EULL, "vxca_caverns1_mines_ee0_0_0.vlm"},
	{0x2E7841003BA7A308ULL, "mage_fireball.dds"},
	{0x2E784711B0CFD910ULL, "w_encounter3_11_0.rml"},
	{0x2E78EB7B4FF388BFULL, "vxht_0_hightown_a2_b75_0_0.vlm"},
	{0x2E78EC242B987185ULL, "llac_ca_caverns1_24425_5.dds"},
	{0x2E78EC24C6B55315ULL, "185895937.ogg"},
	{0x2E79959CD569E031ULL, "primitives_primitive_p3.psh.fxo.dx11.assembly"},
	{0x2E7A2E0D8190C56FULL, "vxht_0_hightown_a2_7iz_0_0.vlm"},
	{0x2E7A31348FFB3CA4ULL, "zz_combat_demo_start.utp"},
	{0x2E7A39DA6481AAA0ULL, "6145093_m.fxe"},
	{0x2E7A6FFC5ADD5FC2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killbag_sa.dds"},
	{0x2E7A9F5BD3E67CB1ULL, "follower_banter6170693.cl"},
	{0x2E7B347C937982E5ULL, "vxht_0_hightown_a3_c3s_0_0.vlm"},
	{0x2E7B854F48A0019AULL, "vfx_chunkyparticle_p5.psh.fxo.dx11"},
	{0x2E7BA19885A7700BULL, "mag311_huon6042703.cl"},
	{0x2E7C79E96BE7A81DULL, "vxht_keep_5av_0_0.vlm"},
	{0x2E7CCCDD11B0E6A4ULL, "vxw_woundedcoast_xq_0_0.vlm"},
	{0x2E7D5A1D9B14776AULL, "fen210_orana6133074.cl"},
	{0x2E7D9FA527DD3A03ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_bed_to_knife.gad"},
	{0x2E7E105A532D6C26ULL, "w_woundedcoast_114_0.phy"},
	{0x2E7E1206DAD71D02ULL, "hlac_ht_0_hightown_a2_-1_0y.dds"},
	{0x2E7E16FFD325EC46ULL, "vxw_encounter2_camp_mi_0_0.vlm"},
	{0x2E7E1BB408AF12C0ULL, "6120122_m.fxe"},
	{0x2E7E4B60D42D9644ULL, "916038911.ogg"},
	{0x2E7E4F98E8BFDAD7ULL, "hla_ht_0_hightown_a2_-2_0x.dds"},
	{0x2E7E71D6B224DE54ULL, "183139654.ogg"},
	{0x2E7F6C84250F7B3EULL, "zz_merpt_debug.ncs"},
	{0x2E7F7278B207574DULL, "6075980_m.fxe"},
	{0x2E7FBAB468C2E852ULL, "131187261.ogg"},
	{0x2E8012F19ECF3D15ULL, "6110784_m.fxe"},
	{0x2E8029C8CE07F487ULL, "vxlt_0_lowtown_docks_a3_n_d43_0_0.vlm"},
	{0x2E80698AFD851B32ULL, "plt_cod_cha_sebastian3.plo"},
	{0x2E80B5D5F20419E9ULL, "6147898_m.fxe"},
	{0x2E80E2A555EB5BA3ULL, "242207611.ogg"},
	{0x2E80E388A6D53A17ULL, "w_bonepit_139_0.msh"},
	{0x2E80F588A6D5597AULL, "w_bonepit_139_0.mao"},
	{0x2E80F988A6D55F89ULL, "w_bonepit_139_0.mmh"},
	{0x2E8163868AC72181ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extpatternfloor.dds"},
	{0x2E81E9806A1032AEULL, "vxht_0_hightown_a2_bp1_0_0.vlm"},
	{0x2E81F344DCC876AEULL, "457881658.ogg"},
	{0x2E827E1E032FA203ULL, "w_encounter2_46_0.mao"},
	{0x2E82821E032FA950ULL, "w_encounter2_46_0.mmh"},
	{0x2E82901E032FC16AULL, "w_encounter2_46_0.msh"},
	{0x2E829DCFD0DDFF8BULL, "vxdr_0_deeproads_fih_0_0.vlm"},
	{0x2E82AF4F3A6E7C12ULL, "63797710.ogg"},
	{0x2E82B92D6C103397ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03tree_1.mao"},
	{0x2E833775ED12321BULL, "mrl150_hunter6071007.cl"},
	{0x2E835724E89FFDDFULL, "6060342_m.fxe"},
	{0x2E836BEC5371046BULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.impa.ani"},
	{0x2E840B0B443F9B49ULL, "6195826_m.fxe"},
	{0x2E840EC2075722DAULL, "hlac_w_encounter2_196609_0y.dds"},
	{0x2E84143F4F1D5A9CULL, "one280_iwan6096863.cl"},
	{0x2E843E25D40291A7ULL, "gen00fl_dog6166427.cl"},
	{0x2E844C697B8EDAB9ULL, "856144487.ogg"},
	{0x2E847CA527E31DE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_bed_to_knife.evt"},
	{0x2E8541BC85464F93ULL, "6080151_m.fxe"},
	{0x2E85CDCFCF363A9EULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.impa.ani"},
	{0x2E85DAF60D0C710EULL, "6098249_m.fxe"},
	{0x2E866C053A566D29ULL, "1017223661.ogg"},
	{0x2E86E116203BD546ULL, "vxca_caverns1_jti_0_0.vlm"},
	{0x2E870276A60B6805ULL, "736932410.ogg"},
	{0x2E8702D105204F68ULL, "1073631203.ogg"},
	{0x2E879F90277196F8ULL, "sw_p_dis.mao"},
	{0x2E8851C756D6351BULL, "gen00fl_dog6170604.cl"},
	{0x2E88F69847083CA1ULL, "vxht_bartrand_n_5qb_0_0.vlm"},
	{0x2E896790BAE7EEFCULL, "6039769_m.fxe"},
	{0x2E89A659856F9B65ULL, "vxht_0_hightown_a3_c4r_0_0.vlm"},
	{0x2E89D92A3AB916FBULL, "6148152_m.fxe"},
	{0x2E89F67E13614526ULL, "397651320.ogg"},
	{0x2E89FC09303FB556ULL, "6025163_f.fxe"},
	{0x2E8A1EBF1F4727ADULL, "vxw_encounter2_lq_0_0.vlm"},
	{0x2E8A4D4F19D42DABULL, "505990711.ogg"},
	{0x2E8AE72F4CD1803BULL, "llac_lt_gallowstemplar_7198_6.dds"},
	{0x2E8AF6F931D431C1ULL, "6130976_f.fxe"},
	{0x2E8B3EBA03A6FF24ULL, "vxw_woundedcoast_uv_0_0.vlm"},
	{0x2E8B54501BA60C02ULL, "llac_lt_0_lowtown_a2_n_-65537_5.dds"},
	{0x2E8B7616DA8A3BFBULL, "w_sundermount_94_0.rml"},
	{0x2E8BC40D8DFC0048ULL, "mcr370_varric6208273.cl"},
	{0x2E8C88DCC51C04C6ULL, "vxw_encounter2_camp_hj_0_0.vlm"},
	{0x2E8D0F8294990114ULL, "6217969_m.fxe"},
	{0x2E8D4C6C460CC1E6ULL, "vxlt_warehouse_n_8d_0_0.vlm"},
	{0x2E8D8AF6864783CFULL, "6153611_m.fxe"},
	{0x2E8DEEE2CAE906D8ULL, "vxht_bartrand_n_4pa_0_0.vlm"},
	{0x2E8E23AD7BA2F110ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft2a.mao"},
	{0x2E8E27AD7BA2F823ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft2a.mmh"},
	{0x2E8E35AD7BA3107DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft2a.msh"},
	{0x2E8E590244F2B4F9ULL, "387938881.ogg"},
	{0x2E8E6DC601948D94ULL, "6084444_m.fxe"},
	{0x2E8E747036D13E3BULL, "6043914_m.fxe"},
	{0x2E8EABC3582B670EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railic.phy"},
	{0x2E8F8F95C5366B14ULL, "vxht_mansion_n_62n_0_0.vlm"},
	{0x2E8F9A0E7F42B66CULL, "vxdr_primevalentrance_2nh_0_0.vlm"},
	{0x2E8FAE4B30F6A76EULL, "528296022.ogg"},
	{0x2E8FB87EC621E122ULL, "hlac_ht_mansion_n_3367_1z.dds"},
	{0x2E8FFD21526979AFULL, "art\\vfx\\materials\\vfx_blood_dec2.mao"},
	{0x2E90004DA8993D6AULL, "qun110_petrice6042283.cl"},
	{0x2E9001C97CB2E738ULL, "278078567.ogg"},
	{0x2E904232B74E7F4FULL, "6166112_m.fxe"},
	{0x2E904882EC7AFB0DULL, "ser221_gascard6165642.cl"},
	{0x2E9053EC537B65E2ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.impa.evt"},
	{0x2E905B5D5C064B2EULL, "vxht_0_hightown_a2_bu2_0_0.vlm"},
	{0x2E90FADE9D9B551DULL, "197268707.ogg"},
	{0x2E90FF373653E879ULL, "vxht_keep_64j_0_0.vlm"},
	{0x2E9122021C3DB379ULL, "6034312_m.fxe"},
	{0x2E9212B073952AC7ULL, "llac_w_encounter2_camp_-65533_0.dds"},
	{0x2E9321DFFE715669ULL, "art\\characters\\weapons\\promo\\longswords\\textures\\w_lsw_025a_0d.dds"},
	{0x2E933DEBF16B3F80ULL, "hlac_ht_0_hightown_a3_-65538_1y.dds"},
	{0x2E934DF33393AE5DULL, "6027752_m.fxe"},
	{0x2E93DDE7E29FF1C8ULL, "framebuffereffects_hslmatrix_p0.psh.fxo.dx11.assembly"},
	{0x2E93E5CFCF42A16FULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.impa.evt"},
	{0x2E942FEBA3800B07ULL, "6053347_f.fxe"},
	{0x2E94E479A23D7FCCULL, "mag101_grace6053095.cl"},
	{0x2E94EF2AE2E038BEULL, "vxw_bonepit_lf_0_0.vlm"},
	{0x2E9752EC538182F5ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.impa.gad"},
	{0x2E97687538FE5521ULL, "vxht_bartrand_n_5ta_0_0.vlm"},
	{0x2E97835F983954B2ULL, "zz_mcr_debug6080865.cl"},
	{0x2E99E6CFCF470E18ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.impa.gad"},
	{0x2E9A1C8BDCC0A065ULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.anb"},
	{0x2E9A33A5CAEBD0CCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_quentin.ani"},
	{0x2E9A55067B8313DBULL, "307381544.ogg"},
	{0x2E9A9F7C89B3FB67ULL, "6033650_m.fxe"},
	{0x2E9AFAB9B7121D46ULL, "vxw_encounter2_camp_ck_0_0.vlm"},
	{0x2E9B8BB5960EFD3CULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tn180r.evt"},
	{0x2E9BB1D10266F59FULL, "w_bonepitdestroyed_a3_19_0_ll.mmh"},
	{0x2E9BCEEAD2C08DB5ULL, "vxlt_ruinedhouse_n_5z_0_0.vlm"},
	{0x2E9C5FA511D0D5B5ULL, "vxca_cavern1_haunted_e8h_0_0.vlm"},
	{0x2E9C75E08C02BE45ULL, "qun220cr_javaris6071490.cl"},
	{0x2E9D05802384F1B2ULL, "dae000_paivel6111937.cl"},
	{0x2E9D0E2E5B610260ULL, "lt_0_lowtown_n_41_0.rml"},
	{0x2E9D3A58E52A51BDULL, "dae200cr_captain_wright6061684.cl"},
	{0x2E9D5D161DFEB7BBULL, "chu_06.dds"},
	{0x2E9D901C965DEB94ULL, "6153821_m.fxe"},
	{0x2E9E37E99D3DB04AULL, "cd_one110_pryce.bnk"},
	{0x2E9F28AD7BB16D48ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft2a.phy"},
	{0x2E9F7114284A00F9ULL, "vxht_keep_61i_0_0.vlm"},
	{0x2E9F96178630AFCBULL, "44598533.ogg"},
	{0x2E9FAC25F37D35C0ULL, "vxlt_0_lowtown_a2_fep_0_0.vlm"},
	{0x2EA01788FE2F140CULL, "hlac_dr_0_deeproads_17639_0x.dds"},
	{0x2EA0625E75E0D858ULL, "mer341_jansen6083005.cl"},
	{0x2EA0A4579BC9DA20ULL, "hf_sgt_joanna.mor"},
	{0x2EA140A5CAF2050EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_quentin.gad"},
	{0x2EA15DDA4A77B55FULL, "858426501.ogg"},
	{0x2EA1BCDEDBDBDA57ULL, "hlac_w_bonepit_0_10x.dds"},
	{0x2EA2F38D64862B4EULL, "hlac_lt_warehouse_0_2y.dds"},
	{0x2EA3D950C9A4BAACULL, "gen00fl_merrill6150688.cl"},
	{0x2EA3DE63C1E0C42FULL, "mrl350ip_grave.ncs"},
	{0x2EA44E83335D9EC4ULL, "6068438_f.fxe"},
	{0x2EA4858DF7474576ULL, "vxht_mansion2b_2q2_0_0.vlm"},
	{0x2EA52CB676A2708EULL, "6107383_m.fxe"},
	{0x2EA54476F3A88A3DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\blendtree\\c_dsogre_charge.anb"},
	{0x2EA606880EBAFD00ULL, "6107002_m.fxe"},
	{0x2EA61C55138CCFD7ULL, "plt_prm000pt_lothering_tome.plo"},
	{0x2EA64B966D5C2CFEULL, "vxlt_undercitydungeon_142_0_0.vlm"},
	{0x2EA65A7081CD4919ULL, "mag100_karras6052957.cl"},
	{0x2EA6CBC8936C8AEAULL, "art\\vfx\\combat\\vfx_sg_app_c.impa.gad"},
	{0x2EA70279033358ADULL, "vxw_encounter2_ro_0_0.vlm"},
	{0x2EA73BA5CAF668EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_quentin.evt"},
	{0x2EA751A746940456ULL, "vxca_cave_2k4_0_0.vlm"},
	{0x2EA75D1BE14B63E7ULL, "hlac_w_bonepitdestroyed_a3_0_9x.dds"},
	{0x2EA784C4B8A3C66AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic_3.mao"},
	{0x2EA7D83E714D514BULL, "621516767.ogg"},
	{0x2EA7F44CB005A7D9ULL, "123687253.ogg"},
	{0x2EA946F500B8B20BULL, "terrain_toolset.psh.fxo"},
	{0x2EA9B81738F58194ULL, "160482571.ogg"},
	{0x2EAA0B3B4F016E15ULL, "vxlt_0_lowtown_a2_ffx_0_0.vlm"},
	{0x2EAADB4DE7D57796ULL, "qcr000_arishok6165086.cl"},
	{0x2EAADC6720739EACULL, "6112736_m.fxe"},
	{0x2EAAE4B98BCFC702ULL, "vxlt_0_lowtown_docks_n_alp_0_0.vlm"},
	{0x2EAB0F9117F09843ULL, "945498536.ogg"},
	{0x2EAB5FBF969740C4ULL, "vxlt_warehouse_10v_0_0.vlm"},
	{0x2EAB8862FFD11596ULL, "6208704_f.fxe"},
	{0x2EABEA3582489D03ULL, "915497145.ogg"},
	{0x2EAC15A560CA1BA4ULL, "6111210_m.fxe"},
	{0x2EAC5E6E80A0791AULL, "vxdr_dungeon_1lx_0_1.vlm"},
	{0x2EAC6DDF48E8C9D2ULL, "hlac_lt_0_lowtown_a3_n_c_196607_1y.dds"},
	{0x2EAC87BAC49F14A2ULL, "lt_0_lowtown_a3_n_c_33_0.anb"},
	{0x2EAD0539BD08B2EEULL, "vxlt_gallowscourtyard_n_6yl_0_0.vlm"},
	{0x2EAD3F173FF27C2EULL, "vxht_0_hightown_a2_bc4_0_0.vlm"},
	{0x2EAD606B825CF3F8ULL, "follower_barks6173421.cl"},
	{0x2EAD69A4BFCF49E2ULL, "214880115.ogg"},
	{0x2EADAEC893727781ULL, "art\\vfx\\combat\\vfx_sg_app_c.impa.evt"},
	{0x2EAED0D5828BE8E4ULL, "vxdr_primevaltomb_7ul_0_0.vlm"},
	{0x2EAFEEF0E039C85BULL, "ts_coterie_02.gda"},
	{0x2EB00807E00A958DULL, "ht_0_hightown_a2_62_0_lc.mmh"},
	{0x2EB05DAFC278DEB5ULL, "6076237_m.fxe"},
	{0x2EB120E92E6C1CAFULL, "ht_0_hightown_43_0.rml"},
	{0x2EB12663244E0499ULL, "6013427_m.fxe"},
	{0x2EB14089B048F1E5ULL, "ser221_quentin6123391.cl"},
	{0x2EB2482D9471D281ULL, "6040892_m.fxe"},
	{0x2EB37C1295CA15B3ULL, "vxht_0_hightown_a3_7fv_0_0.vlm"},
	{0x2EB40857FAA9BD68ULL, "hlac_lt_0_lowtown_a2_n_c_65536_7z.dds"},
	{0x2EB4896089F3800AULL, "98444436.ogg"},
	{0x2EB48EB70C369D93ULL, "6092267_f.fxe"},
	{0x2EB4B69895303F83ULL, "357791573.ogg"},
	{0x2EB4F2B5EBEA81CFULL, "hlac_lt_0_lowtown_131071_2y.dds"},
	{0x2EB4F423F68CA0EAULL, "hlac_lt_gallowscourtyard_65535_0x.dds"},
	{0x2EB5A12AB7A9EA98ULL, "399866566.ogg"},
	{0x2EB5E6F56BE74696ULL, "33535812.ogg"},
	{0x2EB5F7EB50ECF97EULL, "vxca_cavern1_haunted_hi5_0_0.vlm"},
	{0x2EB6339DE7D33937ULL, "47642709.ogg"},
	{0x2EB695366E049B0FULL, "vxlt_gallowscourtyard_n_7cj_0_0.vlm"},
	{0x2EB6A342ABC47E4AULL, "art\\vfx\\combat\\vfx_r_gn3_repulsion_aoe_p.anb"},
	{0x2EB6C91333C4FBAAULL, "6160748_m.fxe"},
	{0x2EB75D2C86E0B3FDULL, "743595387.ogg"},
	{0x2EB77C0DE4C6BB27ULL, "629033888.ogg"},
	{0x2EB7995A12A90ADAULL, "6121216_f.fxe"},
	{0x2EB80668D94EB157ULL, "vxlt_0_lowtown_docks_n_bey_0_0.vlm"},
	{0x2EB83D2CDF558BD7ULL, "446607932.ogg"},
	{0x2EB896B768961288ULL, "gen00fl_aveline6139622.cl"},
	{0x2EBA2B55139D924CULL, "plt_prm000pt_lothering_tome.nss"},
	{0x2EBA8249BABA0225ULL, "gen00fl_aveline6114817.cl"},
	{0x2EBB3A88238C52DCULL, "644944153.ogg"},
	{0x2EBB42F6AE5C5D16ULL, "character_character_p54.psh.fxo.dx11.assembly"},
	{0x2EBB6076D943FA2FULL, "359842794.ogg"},
	{0x2EBB821E035F6E13ULL, "w_encounter2_46_0.rml"},
	{0x2EBBA6C8937EA7D0ULL, "art\\vfx\\combat\\vfx_sg_app_c.impa.ani"},
	{0x2EBBCBB2D507B1FBULL, "vxw_encounter2_mx_0_0.vlm"},
	{0x2EBC060BC2F9EBC1ULL, "615583735.ogg"},
	{0x2EBC223C0597A0A0ULL, "zz_mcr_debug6080786.cl"},
	{0x2EBC35EEFDF37F22ULL, "hm_martin.mor"},
	{0x2EBC36161E19267DULL, "chu_06.mao"},
	{0x2EBC3C6001BA8708ULL, "6183815_m.fxe"},
	{0x2EBCEE9C7C03EB50ULL, "6110535_f.fxe"},
	{0x2EBD5B9FCD09DFFEULL, "6110119_f.fxe"},
	{0x2EBDB77BD406E844ULL, "hlac_lt_darktown_-1_6x.dds"},
	{0x2EBDCB8B6EC1FE11ULL, "6125160_m.fxe"},
	{0x2EBDE8A7EE1C4A57ULL, "gen00fl_anders6128363.cl"},
	{0x2EBEDF267857A05DULL, "321123346.ogg"},
	{0x2EBF08C398AFB5B6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sack_n.dds"},
	{0x2EBF3822341D38A5ULL, "mrl000_marethari6129411.cl"},
	{0x2EBF746A0906A1E2ULL, "346574511.ogg"},
	{0x2EBF984016441D19ULL, "6018527_m.fxe"},
	{0x2EBFF35959D718BDULL, "6057078_m.fxe"},
	{0x2EBFF98ACABA6302ULL, "gen00fl_anders6113032.cl"},
	{0x2EC06BAFCE006599ULL, "837948093.ogg"},
	{0x2EC087B0B01C0DD2ULL, "vxlt_gallowscourtyard_3x1_0_0.vlm"},
	{0x2EC186D1086ADE9CULL, "lt_0_lowtown_21_0_lc.mmh"},
	{0x2EC1C97221E9799FULL, "hlac_lt_0_lowtown_n_65536_6z.dds"},
	{0x2EC1CDC6FEF68033ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drga_0d.dds"},
	{0x2EC1DECFB8AE1E77ULL, "6068038_m.fxe"},
	{0x2EC20172B4EF9F08ULL, "6149505_m.fxe"},
	{0x2EC232B2739A6AA8ULL, "6152758_m.fxe"},
	{0x2EC26AE1EFE1860DULL, "aip_hurlockshield_base.gda"},
	{0x2EC2891E0365996BULL, "w_encounter2_46_0.phy"},
	{0x2EC29BC397C5192DULL, "dae100_bran6144429.cl"},
	{0x2EC2FACCDE6442EFULL, "vxht_0_hightown_a2_bwe_0_0.vlm"},
	{0x2EC31D8BDCE3A6A6ULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.mmh"},
	{0x2EC3AC8DC732A2B7ULL, "6154074_m.fxe"},
	{0x2EC3CFBD565615ECULL, "764664023.ogg"},
	{0x2EC3D54DE41A330AULL, "generate_treasure.ncs"},
	{0x2EC42A403D60D82DULL, "705309593.ogg"},
	{0x2EC43861911286BFULL, "one300_nuncio6083280.cl"},
	{0x2EC43C08D33294C4ULL, "1069229342.ogg"},
	{0x2EC475831B417F56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_2.evt"},
	{0x2EC47C18C1901BEEULL, "vxdr_0_deeproads_f5u_0_0.vlm"},
	{0x2EC4890BC91CA718ULL, "6114411_m.fxe"},
	{0x2EC4BE185FC705F8ULL, "llac_lt_0_lowtown_a2_65535_9.dds"},
	{0x2EC507135FFAB38FULL, "vxlt_gallowscourtyard_n_7fu_0_0.vlm"},
	{0x2EC50C4596CE92D7ULL, "mcr371_orsino6083314.cl"},
	{0x2EC5132FAA51B9B3ULL, "zz_seb_debug6214546.cl"},
	{0x2EC5189A774C744EULL, "6082286_m.fxe"},
	{0x2EC524E753FC31D9ULL, "vxca_caverns2_2uj_0_0.vlm"},
	{0x2EC543E15367BB05ULL, "488045576.ogg"},
	{0x2EC56496D9DB0BFDULL, "hlac_dr_dungeon_3216_11y.dds"},
	{0x2EC5986AD0300E00ULL, "6098816_m.fxe"},
	{0x2EC5A92A83D28304ULL, "6033690_f.fxe"},
	{0x2EC5BCE6933C2CFEULL, "110872401.ogg"},
	{0x2EC5D2FED179B7E8ULL, "6144850_m.fxe"},
	{0x2EC5E06957BD0C1AULL, "w_bonepitmining_a2_86_0.mmh"},
	{0x2EC5E46957BD1351ULL, "w_bonepitmining_a2_86_0.mao"},
	{0x2EC5ED166E56D008ULL, "955846423.ogg"},
	{0x2EC5F26957BD2B60ULL, "w_bonepitmining_a2_86_0.msh"},
	{0x2EC5F8EADDE3552EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_table_01_1_sa.dds"},
	{0x2EC60C647F4EEDC9ULL, "zz_hawke_docksnt.cut"},
	{0x2EC644E41AB096E7ULL, "hlac_lt_0_lowtown_docks_131073_0x.dds"},
	{0x2EC64BE795B7BB9BULL, "hlac_w_encounter3_-327683_0x.dds"},
	{0x2EC67845CB44C9D7ULL, "vxlt_0_lowtown_docks_n_bbx_0_0.vlm"},
	{0x2EC691F81A03D5CAULL, "gen_im_cns_cft_arn_08.uti"},
	{0x2EC6D24862246A74ULL, "gen00fl_aveline6101698.cl"},
	{0x2EC72D4AA92F48F8ULL, "gen00fl_anders6116114.cl"},
	{0x2EC7B53BE7B31F35ULL, "vxca_cavern1_haunted_e7u_0_0.vlm"},
	{0x2EC7C158D351DBFCULL, "6036701_f.fxe"},
	{0x2EC82023E69A168DULL, "gen00fl_anders6103670.cl"},
	{0x2EC8704EE8D5ECBBULL, "vxht_brothel_n_3sd_0_0.vlm"},
	{0x2EC87E435E6A4107ULL, "gen00fl_aveline6139920.cl"},
	{0x2EC87FB2C89E39F0ULL, "arch_blood3.mao"},
	{0x2EC8BEB39AD6A92FULL, "vxlt_0_lowtown_docks_a3_n_ctl_0_0.vlm"},
	{0x2EC8C0A587FDE7C4ULL, "952255646.ogg"},
	{0x2EC982F24A089D77ULL, "99119730.ogg"},
	{0x2EC989C2942A14DBULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twra_0n.dds"},
	{0x2ECA22D123DEAD2EULL, "vxht_0_hightown_a2_bi6_0_0.vlm"},
	{0x2ECA4615646BB9C7ULL, "6178344_m.fxe"},
	{0x2ECA519825D46B0FULL, "9071019.ogg"},
	{0x2ECA9A6623F2FE2CULL, "mcr310_argument6075819.cl"},
	{0x2ECACD635932F19BULL, "hlac_lt_0_lowtown_docks_-196610_3z.dds"},
	{0x2ECB08749EFC1796ULL, "vxca_caverns1_ec3_0_0.vlm"},
	{0x2ECB115F09874B9FULL, "6120936_m.fxe"},
	{0x2ECB2751E55554CFULL, "hlac_lt_0_lowtown_a2_n_c_-131072_8z.dds"},
	{0x2ECB6C831B478EE1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_2.gad"},
	{0x2ECB888E3CE59F51ULL, "and111_refugee_barks6111538.cl"},
	{0x2ECCEA0DA4E9DED9ULL, "ter_grassa_s.dds"},
	{0x2ECCF121A2750CACULL, "vxlt_0_lowtown_a2_n_c_ei6_0_0.vlm"},
	{0x2ECD23BE07869094ULL, "6124723_f.fxe"},
	{0x2ECD3511B117FA8DULL, "w_encounter3_11_0.msh"},
	{0x2ECD3E5D0F6B40E6ULL, "211400091.ogg"},
	{0x2ECD4311B11812C0ULL, "w_encounter3_11_0.mao"},
	{0x2ECD4711B1181913ULL, "w_encounter3_11_0.mmh"},
	{0x2ECDA85704CA469BULL, "453492541.ogg"},
	{0x2ECDCA3B0969FE66ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_enter.ani"},
	{0x2ECE26EB9D7A48EFULL, "vxht_0_hightown_bb6_0_0.vlm"},
	{0x2ECE4135F7F52C69ULL, "vxht_chantry_27g_0_0.vlm"},
	{0x2ECE93B1E2AF689AULL, "vxlt_darktown_5az_0_0.vlm"},
	{0x2ECEC38F4A11EB71ULL, "6111615_m.fxe"},
	{0x2ECED3072FBEBB48ULL, "6169503_m.fxe"},
	{0x2ECF223A05035477ULL, "llac_w_encounter3_0_2.dds"},
	{0x2ECF45556E09A5E5ULL, "1030273845.ogg"},
	{0x2ECF4EC32270B25CULL, "348238217.ogg"},
	{0x2ECFA7F7458273DBULL, "6072770_m.fxe"},
	{0x2ECFAAC35862FCD1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railic.mmh"},
	{0x2ECFB6C358631172ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railic.mao"},
	{0x2ECFC4C35863288FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railic.msh"},
	{0x2ECFF8F383325E4BULL, "mcr310_argument6074347.cl"},
	{0x2ED051F84A3F7E8EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2closed.msh"},
	{0x2ED05FF84A3F96DFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2closed.mao"},
	{0x2ED06BF84A3FAB7CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2closed.mmh"},
	{0x2ED084CC25B4B42CULL, "vxht_chantry_n_2dd_0_0.vlm"},
	{0x2ED0F49B5C74D6B0ULL, "655120816.ogg"},
	{0x2ED16CA9D05A5B6FULL, "vxht_0_hightown_a2_bxd_0_0.vlm"},
	{0x2ED16D831B4BFB8FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_2.ani"},
	{0x2ED1CC131D4605BDULL, "6048777_m.fxe"},
	{0x2ED1E2D99C426368ULL, "313647149.ogg"},
	{0x2ED20EA1507D9C10ULL, "qun221tr_follower_amb_chant.utt"},
	{0x2ED217EC37DCDB48ULL, "llac_lt_shop_-1_8.dds"},
	{0x2ED2539DC091017BULL, "458175810.ogg"},
	{0x2ED270687DC7B126ULL, "6120754_f.fxe"},
	{0x2ED2BC8274D81909ULL, "6132871_f.fxe"},
	{0x2ED2C72F1B2EFE15ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelogs_0.pwk"},
	{0x2ED2FAE8BB12BF98ULL, "6113511_m.fxe"},
	{0x2ED335122969DA08ULL, "zz_vault_debug6178920.cl"},
	{0x2ED378F051F0CC0FULL, "vxlt_gallowscourtyard_n_7it_0_0.vlm"},
	{0x2ED3F1055AF91C97ULL, "one100dg_nabil6033161.cl"},
	{0x2ED3FEC14E4676FDULL, "6053359_m.fxe"},
	{0x2ED4E0E9B5D6BA6CULL, "art\\characters\\weapons\\shields\\textures\\w_shd_004a_d.dds"},
	{0x2ED5529610466184ULL, "486723817.ogg"},
	{0x2ED591390F6C51D5ULL, "zz_syncvfxtest.ncs"},
	{0x2ED5956A45857299ULL, "vxlt_undercitydungeon_yw_0_0.vlm"},
	{0x2ED5DE9CEB38C0F6ULL, "6136703_m.fxe"},
	{0x2ED60292EB316EC9ULL, "277245651.ogg"},
	{0x2ED60638F1F480FEULL, "narrative_banter6165524.cl"},
	{0x2ED6E5F2B79976B3ULL, "6042727_f.fxe"},
	{0x2ED6F8B850C70FB8ULL, "6106089_m.fxe"},
	{0x2ED72E4665C7D54EULL, "lt_0_lowtown_docks_n_184_0.msh"},
	{0x2ED73C4665C7EC9FULL, "lt_0_lowtown_docks_n_184_0.mao"},
	{0x2ED7484665C8013CULL, "lt_0_lowtown_docks_n_184_0.mmh"},
	{0x2ED749AD77530F23ULL, "6016149_m.fxe"},
	{0x2ED76B78333035EFULL, "787064122.ogg"},
	{0x2ED7FC2B5C7CB5E2ULL, "701927448.ogg"},
	{0x2ED81428CC63472EULL, "zz_one_debug6077964.cl"},
	{0x2ED84D58529527D6ULL, "isa321_velasco6094484.cl"},
	{0x2ED8A239CC7D3FA8ULL, "897096515.ogg"},
	{0x2ED8D6E45DB80510ULL, "6043399_m.fxe"},
	{0x2ED8E1D1475313F2ULL, "mag210_arianni6021226.cl"},
	{0x2ED91521A8BF888CULL, "46820121.ogg"},
	{0x2ED98BE6F191A882ULL, "hlac_w_sundermount_-262147_0x.dds"},
	{0x2ED99CA6798A1F95ULL, "zz_tbe100_barber_flyer6190989.cl"},
	{0x2EDA21FB01CED93CULL, "243026755.ogg"},
	{0x2EDA30E949FC7C13ULL, "6128716_m.fxe"},
	{0x2EDA5F4451B84F1CULL, "93207643.ogg"},
	{0x2EDAD0E0E25D1A76ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwaywall.phy"},
	{0x2EDB58FCAAFCF5B5ULL, "252241692.ogg"},
	{0x2EDB7DE63999C45FULL, "6128411_m.fxe"},
	{0x2EDC4866C6D4778EULL, "hlac_ca_caverns1_24422_7x.dds"},
	{0x2EDCA21E3BCC2DEFULL, "vxca_caverns2_8p5_0_0.vlm"},
	{0x2EDCB312E9EB44E9ULL, "vxht_chantry_20f_0_0.vlm"},
	{0x2EDCE958C1837E63ULL, "llac_ht_bartrand_n_3373_0.dds"},
	{0x2EDD17BFF3C7CD6BULL, "hlac_lt_gallowstemplar_n_2394_9y.dds"},
	{0x2EDDD4086673D7D0ULL, "279828048.ogg"},
	{0x2EDDDBB634A9A478ULL, "hlac_dr_dungeon_3215_0y.dds"},
	{0x2EDF4C1E2CE1A161ULL, "vxlt_undercityserialkiller_19g_0_0.vlm"},
	{0x2EDFA642ABE7883DULL, "art\\vfx\\combat\\vfx_r_gn3_repulsion_aoe_p.mmh"},
	{0x2EDFEF1E0C3F2C4FULL, "319620536.ogg"},
	{0x2EE022421FBCFFA1ULL, "gen00fl_merrill6108673.cl"},
	{0x2EE0BC147618AD7BULL, "hlac_w_woundedcoast_196610_0y.dds"},
	{0x2EE11B0CAEB5BDDEULL, "vxw_woundedcoast_na_0_0.vlm"},
	{0x2EE13D39BBAECB74ULL, "lowtown_male_cityguard_mob6150514.cl"},
	{0x2EE1AAC78DFCD2E9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_anda_0s.dds"},
	{0x2EE1FAA1F786DB00ULL, "dae000_gen_turn_in_46178439.cl"},
	{0x2EE2488F2002929CULL, "w_bonepit_159_0.mao"},
	{0x2EE2548F2002A73FULL, "w_bonepit_159_0.mmh"},
	{0x2EE25A5D07AFB846ULL, "llac_lt_gallowscourtyard_n_-131072_4.dds"},
	{0x2EE25A8F2002B1E1ULL, "w_bonepit_159_0.msh"},
	{0x2EE2B33B097C3360ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_enter.gad"},
	{0x2EE340E42A817AE9ULL, "vxdr_0_deeproads_f63_0_0.vlm"},
	{0x2EE35BFFAF30FAD7ULL, "vxlt_0_lowtown_docks_n_btz_0_0.vlm"},
	{0x2EE3AED445475F80ULL, "vxht_mansion2b_24w_0_0.vlm"},
	{0x2EE3F4BE32A398F6ULL, "vxlt_undercityencounter_13s_0_1.vlm"},
	{0x2EE42D7D0E400E9CULL, "vxw_bonepitdestroyed_a3_75_0_0.vlm"},
	{0x2EE435FE81CA0A61ULL, "837478480.ogg"},
	{0x2EE45079DD2ED976ULL, "6131603_m.fxe"},
	{0x2EE4A84F104DCFDCULL, "alphaclouds02.mao"},
	{0x2EE4B9E0E2654717ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwaywall.msh"},
	{0x2EE4BC59F73E2DF0ULL, "6174022_m.fxe"},
	{0x2EE4CBE0E265667AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwaywall.mao"},
	{0x2EE4CFE0E2656C89ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwaywall.mmh"},
	{0x2EE56CAE751C4FFEULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen.evt"},
	{0x2EE5A26D7EC2DA2FULL, "vxlt_0_lowtown_docks_a3_n_czn_0_0.vlm"},
	{0x2EE5B4FDA9A49318ULL, "lgt220_mag_apost_apost36156747.cl"},
	{0x2EE641CC4848ECF4ULL, "6124715_m.fxe"},
	{0x2EE68A37154CDBA8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medf_0s.dds"},
	{0x2EE68C8759B2A97EULL, "vxca_caverns1_cki_0_0.vlm"},
	{0x2EE6EBF4AC00C6B8ULL, "359327645.ogg"},
	{0x2EE6EFC99477E4DAULL, "follower_banter6171057.cl"},
	{0x2EE6F8733F183047ULL, "hlac_lt_0_lowtown_-65535_6y.dds"},
	{0x2EE758886D1C2EC1ULL, "vxlt_0_lowtown_docks_c42_0_0.vlm"},
	{0x2EE76055A744FABDULL, "vxlt_0_lowtown_docks_a3_n_d2g_0_0.vlm"},
	{0x2EE7C60B8A8DCB3AULL, "zz_lgt_2_debug6206582.cl"},
	{0x2EE7EC2E82E84896ULL, "vxca_caverns1_ei5_0_0.vlm"},
	{0x2EE8141A1E6C8E3DULL, "zz_seb_debug6183093.cl"},
	{0x2EE83CCB7D41DD4CULL, "vxlt_0_lowtown_a2_n_d46_0_0.vlm"},
	{0x2EE841D5051B679CULL, "gen00fl_fenris6127112.cl"},
	{0x2EE8999E44ACB43AULL, "6176722_f.fxe"},
	{0x2EE92CC9F071A25EULL, "vxht_0_hightown_a2_c71_0_0.vlm"},
	{0x2EE98BAC0BD9C9D5ULL, "plt_lgt000pt_tombs_second.plo"},
	{0x2EE9B32647197734ULL, "ser221_party_comments6209851.cl"},
	{0x2EE9B4D9378F33CFULL, "6101487_m.fxe"},
	{0x2EE9B7A3B3004D89ULL, "789098684.ogg"},
	{0x2EEA13E5EB50768AULL, "6065818_m.fxe"},
	{0x2EEA15E1140DD3FFULL, "vxlt_0_lowtown_a2_n_e8d_0_0.vlm"},
	{0x2EEACF8C638B5F45ULL, "one100dg_lia.bnk"},
	{0x2EEB14604350C647ULL, "mrl310_party_comments.bnk"},
	{0x2EEB2BC3EC08110EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_flared1a.phy"},
	{0x2EEB372A8CF70908ULL, "zz_crowd_2.utc"},
	{0x2EEBB6C113AAEC48ULL, "6163703_m.fxe"},
	{0x2EEBC88BC32FE37DULL, "696396307.ogg"},
	{0x2EEBE8DE0DCEC33CULL, "vxca_cavern1_haunted_ei7_0_1.vlm"},
	{0x2EEC21ADEA70BBDAULL, "550220181.ogg"},
	{0x2EEC32CC350D4DCAULL, "vxw_woundedcoast_11q_0_0.vlm"},
	{0x2EEC511EA69867C4ULL, "6051257_m.fxe"},
	{0x2EEC83AE75229529ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen.gad"},
	{0x2EEC9A6BD4CB41D3ULL, "follower_barks6153505.cl"},
	{0x2EED4906E77E6F2AULL, "vxw_encounter2_camp_rd_0_0.vlm"},
	{0x2EEDB9E08CEEFC4BULL, "static_static_p11.psh.fxo"},
	{0x2EEE0F58A22F53C5ULL, "8122806.ogg"},
	{0x2EEE3660435335BFULL, "mrl310_party_comments.cnv"},
	{0x2EEE91130BB39B55ULL, "6149315_m.fxe"},
	{0x2EEEB18C638F157DULL, "one100dg_lia.cnv"},
	{0x2EEF2DD47CF7363CULL, "vxlt_darktown_536_0_0.vlm"},
	{0x2EF0391114EBD67CULL, "146516263.ogg"},
	{0x2EF0B5202977C6ABULL, "vxlt_gallowstemplar_3rl_0_0.vlm"},
	{0x2EF16A0EF5EB5DF1ULL, "293769780.ogg"},
	{0x2EF16A13AE4C084CULL, "789409302.ogg"},
	{0x2EF1BB114F107193ULL, "lgt200cr_cor_blimey.utc"},
	{0x2EF220B1373D7800ULL, "vxht_mansion2b_21t_0_0.vlm"},
	{0x2EF22224843870A8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkb_sa.dds"},
	{0x2EF253A7C07A5C5CULL, "6149453_m.fxe"},
	{0x2EF27AAB77E05F2EULL, "vxlt_0_lowtown_a2_hd7_0_0.vlm"},
	{0x2EF2AAFF11644E1FULL, "6074347_m.fxe"},
	{0x2EF2F0549C4E95E5ULL, "nat320_delilah6065907.cl"},
	{0x2EF2F8892D002AAEULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ff_basic_1.mao"},
	{0x2EF34451CDDCA717ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_throw.gad"},
	{0x2EF354AE75286477ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closedtoopen.ani"},
	{0x2EF45563162FC6E8ULL, "97496892.ogg"},
	{0x2EF4999DFEE3EDBDULL, "spt_fronds.vsh.fxo"},
	{0x2EF4FE644BA8C1FEULL, "vxca_caverns1_cnj_0_0.vlm"},
	{0x2EF5083DD906AF2BULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax101a_0n.dds"},
	{0x2EF5259FEEF5E585ULL, "vxdr_primevalentrance_175_0_0.vlm"},
	{0x2EF57616DAE4D7C8ULL, "w_sundermount_94_0.mmh"},
	{0x2EF58216DAE4EC6BULL, "w_sundermount_94_0.mao"},
	{0x2EF59416DAE50AF2ULL, "w_sundermount_94_0.msh"},
	{0x2EF61104C7B4968BULL, "gencr_coterie_member.utc"},
	{0x2EF6A8CDF22F50FAULL, "vxdr_0_deeproads_ete_0_0.vlm"},
	{0x2EF72123D627D241ULL, "character_character_p13.psh.fxo.dx11"},
	{0x2EF7FDA1509E4ECDULL, "qun221tr_follower_amb_chant.ncs"},
	{0x2EF85A3B61D39B89ULL, "6040984_m.fxe"},
	{0x2EF88563438CB98DULL, "vxlt_0_lowtown_a2_n_hfe_0_0.vlm"},
	{0x2EF8859916078747ULL, "100750389.ogg"},
	{0x2EF8F52D8D98A37EULL, "mcr310_party_banter6087807.cl"},
	{0x2EF93241010D973EULL, "681432101.ogg"},
	{0x2EF9560CDD722291ULL, "hlac_ht_mansion2b_818_7z.dds"},
	{0x2EF9631F5AF2AC90ULL, "vxdr_0_deeproads_cqn_0_0.vlm"},
	{0x2EF97C82735C926FULL, "vxht_0_hightown_bu3_0_0.vlm"},
	{0x2EF9E948B891B21AULL, "vxlt_darktown_5pu_0_0.vlm"},
	{0x2EFA46D2742626F7ULL, "vxht_brothel_n_3sp_0_0.vlm"},
	{0x2EFA5951CDE2E9A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_throw.evt"},
	{0x2EFAFF797B1F5193ULL, "vxht_0_hightown_a3_a2l_0_0.vlm"},
	{0x2EFB15144F9115B0ULL, "gen_im_trs_apo_sta_03.uti"},
	{0x2EFB1F91E4D6CEF6ULL, "229259865.ogg"},
	{0x2EFB28F9AD484CC6ULL, "hlac_w_woundedcoast_0_10x.dds"},
	{0x2EFB4E835763E85CULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.dura.ani"},
	{0x2EFB802BEF5A0DD3ULL, "6162798_m.fxe"},
	{0x2EFB9611AAE687BCULL, "71993980.ogg"},
	{0x2EFBE2F87385653CULL, "vxw_woundedcoast_1gu_0_0.vlm"},
	{0x2EFC2BBD7EE514CBULL, "static_static_p42.psh.fxo"},
	{0x2EFC6D9CD06F8FDAULL, "weapon_weapon_p3.psh.fxo.dx11.assembly"},
	{0x2EFC7EC24BFA8258ULL, "116916527.ogg"},
	{0x2EFCC240A63CA4EFULL, "vxht_0_hightown_a2_bka_0_0.vlm"},
	{0x2EFDCDE178C022E2ULL, "mrl150_merrill6093966.cl"},
	{0x2EFECCFE56801B79ULL, "fen310_fenris_praescia.stg"},
	{0x2EFF56ECF1275763ULL, "vxlt_hangedman_n_2mv_0_0.vlm"},
	{0x2EFF890D15819F2DULL, "w_bonepitmining_a2_131_0.msh"},
	{0x2EFF948E78933255ULL, "vxht_0_hightown_c3f_0_0.vlm"},
	{0x2EFF970D1581B760ULL, "w_bonepitmining_a2_131_0.mao"},
	{0x2EFF9B0D1581BDB3ULL, "w_bonepitmining_a2_131_0.mmh"},
	{0x2EFFC2A86E266A54ULL, "vxht_0_hightown_8pu_0_0.vlm"},
	{0x2F0068BB28B8E086ULL, "vxca_caverns1_cmk_0_0.vlm"},
	{0x2F009012AC1BD5EAULL, "6139232_m.fxe"},
	{0x2F009CC2FC155221ULL, "and111_lirene6044966.cl"},
	{0x2F00C584B27074DCULL, "6142553_m.fxe"},
	{0x2F00D55C0DAF8441ULL, "zz_seb_debug6215778.cl"},
	{0x2F00E58FC032CAF2ULL, "hlac_ca_caverns1_mines_24426_4z.dds"},
	{0x2F01303D1A6578F4ULL, "6117959_m.fxe"},
	{0x2F014372C09558B7ULL, "6128987_m.fxe"},
	{0x2F0159F9267B01FCULL, "vxlt_darktown_4ia_0_0.vlm"},
	{0x2F015B835768699EULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.dura.gad"},
	{0x2F016BE2BD853447ULL, "hlac_w_sundermount_-4_0y.dds"},
	{0x2F0175BAC1DE6824ULL, "6041740_m.fxe"},
	{0x2F0193DD6E21266DULL, "593773007.ogg"},
	{0x2F022E0CE1F1361FULL, "llac_lt_gallowscourtyard_-1_1.dds"},
	{0x2F029F4A44CBFB2AULL, "6089812_m.fxe"},
	{0x2F034D6FD3EED8B1ULL, "6067993_f.fxe"},
	{0x2F03D429AF913881ULL, "511196608.ogg"},
	{0x2F0449D3B2B12A4DULL, "6180587_m.fxe"},
	{0x2F0477D37A8E623DULL, "llac_lt_gallowstemplar_f_2382_1.dds"},
	{0x2F04DF520E234B9CULL, "lt_0_lowtown_a2_n_12_0.rml"},
	{0x2F04E589CF3B066CULL, "vrc321_party_barks6095745.cl"},
	{0x2F0522B778CBD7E0ULL, "hlac_dr_primevaltomb_665_0y.dds"},
	{0x2F0665BF158823E3ULL, "369267688.ogg"},
	{0x2F06B9C8FD6DF826ULL, "vxht_0_hightown_a2_c3l_0_1.vlm"},
	{0x2F06FC89629C7E97ULL, "6070242_m.fxe"},
	{0x2F0726347FDC6A42ULL, "723131797.ogg"},
	{0x2F075A2DCAD79C4FULL, "6166303_m.fxe"},
	{0x2F077151CDED9C9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_throw.ani"},
	{0x2F07A599C7545196ULL, "1023549657.ogg"},
	{0x2F07BB5E6834941EULL, "hlac_dr_dungeon_3215_7z.dds"},
	{0x2F08074A33A31870ULL, "185738613.ogg"},
	{0x2F083683576E49BDULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.dura.evt"},
	{0x2F08AC1BB19BAC3BULL, "w_encounter3_226_0.mmh"},
	{0x2F08B81BB19BC0D8ULL, "w_encounter3_226_0.mao"},
	{0x2F08CA1BB19BDF25ULL, "w_encounter3_226_0.msh"},
	{0x2F08D3E794351D9CULL, "gen00fl_anders6125695.cl"},
	{0x2F08EB060A189F87ULL, "vxlt_gallowstemplar_f_45m_0_0.vlm"},
	{0x2F08EE35A2E8E3E1ULL, "487643603.ogg"},
	{0x2F09439E8D5277D2ULL, "hlac_ca_caverns1_24424_9x.dds"},
	{0x2F099A571C52F000ULL, "451800257.ogg"},
	{0x2F09A5738844D72CULL, "vxlt_0_lowtown_a2_n_c_h2b_0_0.vlm"},
	{0x2F0A00A2D7E6258CULL, "lgt320cr_shw_ors_black_op.utc"},
	{0x2F0A1964FA098E98ULL, "6108465_m.fxe"},
	{0x2F0A2093FBC38F0EULL, "mer340_cowan6083544.cl"},
	{0x2F0A243F7C35497BULL, "ca_caverns1_mines.arl"},
	{0x2F0A2CC0CB6AA02AULL, "vxw_encounter2_camp_hf_0_0.vlm"},
	{0x2F0A4E04F1916E43ULL, "mag320_orsino6060228.cl"},
	{0x2F0AAC979D928DE2ULL, "dre000_bodahn6031288.cl"},
	{0x2F0AC060DDB427D6ULL, "669030718.ogg"},
	{0x2F0AC1613A493CE0ULL, "vxdr_primevaltomb_6wk_0_0.vlm"},
	{0x2F0B1FC5D7E27605ULL, "6174687_m.fxe"},
	{0x2F0B2C36E4E7FF6AULL, "506680241.ogg"},
	{0x2F0B5F76191C1F0AULL, "gen00fl_aveline6139945.cl"},
	{0x2F0B8B0A284C8737ULL, "vxlt_blackemporium_15n_0_0.vlm"},
	{0x2F0BA78A5911F6E9ULL, "vxca_caverns1_mines_gkb_0_1.vlm"},
	{0x2F0C5131DCE42EA3ULL, "gen_im_wep_war_shd_act0_01.uti"},
	{0x2F0C8F264A1FF1BDULL, "hlac_ht_0_hightown_n_-1_2y.dds"},
	{0x2F0C9B0944159E86ULL, "6089665_m.fxe"},
	{0x2F0D0160C4ABDBA0ULL, "377480489.ogg"},
	{0x2F0D11D2F9268CA2ULL, "vxw_sundermount_22k_0_0.vlm"},
	{0x2F0D581EFD441BABULL, "513967428.ogg"},
	{0x2F0D97EAC5722771ULL, "209220976.ogg"},
	{0x2F0DC8C9E31D6FE3ULL, "vxlt_hangedman_n_2hy_0_0.vlm"},
	{0x2F0DE6ACB2960338ULL, "725338064.ogg"},
	{0x2F0E066B6A894AD5ULL, "vxht_0_hightown_c6e_0_0.vlm"},
	{0x2F0E3C1AECDB6F1DULL, "20445608.ogg"},
	{0x2F0E7616A7BCF463ULL, "6129012_m.fxe"},
	{0x2F0F0B77DF994656ULL, "and100_anders6095875.cl"},
	{0x2F0FBC795FDE2EADULL, "475066834.ogg"},
	{0x2F0FCD165CDB361DULL, "vxdr_primevalentrance_fl_0_1.vlm"},
	{0x2F105700121AE586ULL, "6112976_m.fxe"},
	{0x2F10F576F384BBBDULL, "gen00fl_dog6117779.cl"},
	{0x2F10F6DC6634DF59ULL, "6113648_m.fxe"},
	{0x2F118438A0806443ULL, "493720387.ogg"},
	{0x2F119D27D8F470E5ULL, "llac_lt_0_lowtown_docks_0_3.dds"},
	{0x2F11A051E92C51C8ULL, "vxw_bonepitmining_a2_16o_0_0.vlm"},
	{0x2F11A4BD1FE5C530ULL, "t3_shd_spb.tnt"},
	{0x2F120BCDDCE42483ULL, "425577171.ogg"},
	{0x2F121D7E7F26A15BULL, "vxht_keep_n_8av_0_0.vlm"},
	{0x2F1262193D932B00ULL, "657325200.ogg"},
	{0x2F12999C9350CD98ULL, "141302788.ogg"},
	{0x2F12A5E61C401010ULL, "vxht_chantry_2fa_0_0.vlm"},
	{0x2F12AFDEEF5DBDFBULL, "ach_exp_wilderness.dds"},
	{0x2F136A7C9E6884F2ULL, "vxlt_0_lowtown_a2_n_ev7_0_0.vlm"},
	{0x2F14183ABFF1A8CBULL, "ht_0_hightown_36_0_ll.mmh"},
	{0x2F14448D5E863DC8ULL, "346074154.ogg"},
	{0x2F151B66B4B474E6ULL, "6037036_f.fxe"},
	{0x2F1621E431EE3390ULL, "6113573_m.fxe"},
	{0x2F162B24DB0A49A5ULL, "fen110_fenris6101961.cl"},
	{0x2F166D46567071E2ULL, "471740142.ogg"},
	{0x2F173A5624ACF1BAULL, "6096302_m.fxe"},
	{0x2F1746D38BBCD392ULL, "854989702.ogg"},
	{0x2F179D79A79C32EDULL, "vxht_mansion_n_503_0_1.vlm"},
	{0x2F17A34900A45BE4ULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.mmh"},
	{0x2F17A467CF1D5C92ULL, "6128339_m.fxe"},
	{0x2F17B3E3F08A7132ULL, "925599870.ogg"},
	{0x2F17C1E0BDC8C4ADULL, "vxlt_0_lowtown_a3_n_c_h22_0_0.vlm"},
	{0x2F17F32240EB356DULL, "gen00fl_aveline6120843.cl"},
	{0x2F17F6C7774AB362ULL, "gen00fl_fenris6110817.cl"},
	{0x2F18095A0F434E92ULL, "dae000_sandal6148227.cl"},
	{0x2F18590A8D2B504EULL, "dae000ip_crowd_lm3_talk12.utp"},
	{0x2F1881BBF00CE9E7ULL, "vxht_mansion2_262_0_0.vlm"},
	{0x2F189E9DBD60B8AAULL, "vxw_encounter2_camp_cg_0_0.vlm"},
	{0x2F18CE5049159CBBULL, "hlac_w_woundedcoast_-131067_0y.dds"},
	{0x2F19424670E8B975ULL, "6101698_m.fxe"},
	{0x2F1A0389E62AFAB6ULL, "268333261.ogg"},
	{0x2F1A171461306258ULL, "524100480.ogg"},
	{0x2F1A9E8F9E2A1DD8ULL, "367221328.ogg"},
	{0x2F1AC3D2C174729AULL, "hlac_w_encounter2_camp_131070_0x.dds"},
	{0x2F1AC88983EFEECCULL, "gen00fl_merrill6124052.cl"},
	{0x2F1B47EEA3B6705BULL, "vxlt_undercitydungeon_1n1_0_0.vlm"},
	{0x2F1B83AFEB1CA522ULL, "vxw_sundermount_27t_0_0.vlm"},
	{0x2F1BB32078D60E6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fb_transition.gad"},
	{0x2F1BFAE01366CEDCULL, "hlac_lt_0_lowtown_docks_n_-196610_3x.dds"},
	{0x2F1C652D38811509ULL, "442813959.ogg"},
	{0x2F1CC51D298FD470ULL, "1069785725.ogg"},
	{0x2F1D1B51DB65D2AFULL, "6084249_m.fxe"},
	{0x2F1D3394881E0507ULL, "6039717_m.fxe"},
	{0x2F1D82CE2B106F67ULL, "vxlt_shop_it_0_0.vlm"},
	{0x2F1DE0DAD10F672FULL, "dae000_arishok6081660.cl"},
	{0x2F1DFFAC4B30255EULL, "vxw_bonepit_1vk_0_0.vlm"},
	{0x2F1E168412A6CD29ULL, "vxht_0_hightown_a3_cf8_0_0.vlm"},
	{0x2F1E16D02585C737ULL, "llac_lt_0_lowtown_n_65536_8.dds"},
	{0x2F1E19B953867678ULL, "vxht_0_hightown_b8k_0_0.vlm"},
	{0x2F1EB5C2587F2D3CULL, "6197249_m.fxe"},
	{0x2F1EE5E105333E92ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\templar\\blendtree\\templar_override.anb"},
	{0x2F1EFADC1084A629ULL, "gen_im_wep_war_one_act2_02.uti"},
	{0x2F1F15DD5F4B05DDULL, "210738341.ogg"},
	{0x2F1F1DB7CEE83898ULL, "vxlt_0_lowtown_docks_b55_0_0.vlm"},
	{0x2F1F59B29A3C21A2ULL, "6041210_m.fxe"},
	{0x2F20655E305461DBULL, "961339826.ogg"},
	{0x2F206E2F88101F3EULL, "vxlt_0_lowtown_a3_n_c_ef8_0_0.vlm"},
	{0x2F20A020731CB2C2ULL, "vxdr_0_deeproads_crq_0_0.vlm"},
	{0x2F2117C30E362890ULL, "vxht_chantry_2cn_0_0.vlm"},
	{0x2F21624D8C2A2F6AULL, "mcr371_hostile_merrill6082170.cl"},
	{0x2F2163D68BF6707EULL, "vxw_encounter2_ck_0_0.vlm"},
	{0x2F2198B5CE0D337CULL, "vfx_chunkyparticle_p5.psh.fxo.dx11.assembly"},
	{0x2F21AF9FAF5F9A4FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masa_0t.dds"},
	{0x2F21D2CEBE23716AULL, "dae000_gen_turn_in_36178454.cl"},
	{0x2F22C82078DC501CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fb_transition.evt"},
	{0x2F22F5EB5588C06BULL, "6037104_m.fxe"},
	{0x2F23707D98492347ULL, "dae000ip_crowd_em1_talk3.utp"},
	{0x2F238ECB5CF356D8ULL, "934747897.ogg"},
	{0x2F23988FED5E61B5ULL, "vxca_caverns2prc_a0o_0_0.vlm"},
	{0x2F24BB19FD67147FULL, "6040546_m.fxe"},
	{0x2F25649D248E7CC2ULL, "hlac_lt_backalley_g_0_8z.dds"},
	{0x2F25E2A08895C9F9ULL, "6134590_f.fxe"},
	{0x2F267EAC0C0D61B2ULL, "plt_lgt000pt_tombs_second.nss"},
	{0x2F26BB1533F937BAULL, "mag291_torpor6025984.cl"},
	{0x2F26BF4356827A57ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isa.phy"},
	{0x2F27197E7206E427ULL, "6018631_m.fxe"},
	{0x2F274EF41EB8C8DAULL, "vxca_caverns1_mines_e8c_0_0.vlm"},
	{0x2F27581C13B0BFC0ULL, "vxw_encounter1_du_0_0.vlm"},
	{0x2F279C24D129399DULL, "6169472_m.fxe"},
	{0x2F27A621546A7A66ULL, "6077318_m.fxe"},
	{0x2F27D02E54849B1CULL, "6121932_m.fxe"},
	{0x2F27D53DFC5CD623ULL, "vxw_sundermount_312_0_0.vlm"},
	{0x2F284D33E66EB0CEULL, "6025151_m.fxe"},
	{0x2F28727B08CD6762ULL, "vxht_bartrand_n_6kl_0_0.vlm"},
	{0x2F28862F5D3A3D4AULL, "806475282.ogg"},
	{0x2F2948ED52217C99ULL, "vxlt_0_lowtown_a2_n_eny_0_1.vlm"},
	{0x2F29545D6884385EULL, "vxw_woundedcoast_wf_0_0.vlm"},
	{0x2F29F58CDD12BDA2ULL, "vxw_sundermount_28u_0_0.vlm"},
	{0x2F2A59971D6D1B97ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_ex.gad"},
	{0x2F2ABD5732C8647FULL, "follower_banter6173148.cl"},
	{0x2F2AD63572690F41ULL, "gen00fl_aveline6137582.cl"},
	{0x2F2AD796D9207FB1ULL, "vxht_keep_61q_0_0.vlm"},
	{0x2F2B190EBDF3B41DULL, "hlac_lt_0_lowtown_n_1_7y.dds"},
	{0x2F2B2DC80FE8A039ULL, "268603107.ogg"},
	{0x2F2B7D18BAD9F570ULL, "ico_gloves_heavy_1.dds"},
	{0x2F2B9B1588E1EA43ULL, "6054059_m.fxe"},
	{0x2F2BE006F70E4524ULL, "genip_crafting_station.ncs"},
	{0x2F2C000848E35626ULL, "gen00fl_sebastian6189034.cl"},
	{0x2F2C2AC3EC3FA6D1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_flared1a.mmh"},
	{0x2F2C36C3EC3FBB72ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_flared1a.mao"},
	{0x2F2C44C3EC3FD28FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_flared1a.msh"},
	{0x2F2CCE2D69068F06ULL, "vxw_encounter2_fl_0_0.vlm"},
	{0x2F2CD7576F586A50ULL, "290939037.ogg"},
	{0x2F2CEA70E6934DA4ULL, "vxlt_0_lowtown_docks_a3_cp4_0_0.vlm"},
	{0x2F2D80F1A981AE47ULL, "vxlt_darktown_6ip_0_0.vlm"},
	{0x2F2DFC2DD31DCCFEULL, "6183967_m.fxe"},
	{0x2F2E46E53CF5E524ULL, "6148375_m.fxe"},
	{0x2F2E586E9E5DD5ECULL, "6013844_m.fxe"},
	{0x2F2E6943D7667323ULL, "one220_yevhen6098783.cl"},
	{0x2F2E6CB1288594D7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowalcove.phy"},
	{0x2F2EB2833951406EULL, "dae221ip_home_amb_merr_flw2.utp"},
	{0x2F2EC159412E3FEBULL, "89951382.ogg"},
	{0x2F2EE00C7A0637BEULL, "vxlt_0_lowtown_a3_n_c_ei9_0_0.vlm"},
	{0x2F2EFA6B23451413ULL, "vxlt_hangedman_t8_0_0.vlm"},
	{0x2F2F6AE4A70BA7AFULL, "6071201_m.fxe"},
	{0x2F2FD02078E6E835ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fb_transition.ani"},
	{0x2F2FD5B37DEC88FEULL, "vxw_encounter2_nl_0_0.vlm"},
	{0x2F30176E29758D71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reaction.ani"},
	{0x2F3099DEA8895F24ULL, "vxw_bonepitdestroyed_a3_1xb_0_1.vlm"},
	{0x2F30A8E841255B7AULL, "150828697.ogg"},
	{0x2F30C577DF45B6BAULL, "6024130_m.fxe"},
	{0x2F311B7B6414A425ULL, "137932682.ogg"},
	{0x2F312AA0FF8A740CULL, "hlac_lt_0_lowtown_a3_n_c_131072_2y.dds"},
	{0x2F316E971D735E24ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_ex.evt"},
	{0x2F3186F626D8A2FCULL, "299431817.ogg"},
	{0x2F31FA5E627136BDULL, "sf_xy16icf32_t2.vsh.fxo"},
	{0x2F3235000DE35CDBULL, "art\\vfx\\creatures\\fxc_ash_proj_p.dura.ani"},
	{0x2F3235273D4478ABULL, "6116421_m.fxe"},
	{0x2F323A2FAB7BD4E3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_9.pwk"},
	{0x2F32B0EC15FF78B8ULL, "mrl150ip_pile_b.utp"},
	{0x2F32EA18FD45ED7CULL, "535717772.ogg"},
	{0x2F336BD52573D233ULL, "hlac_lt_0_lowtown_-1_0x.dds"},
	{0x2F346864EDFAD7F0ULL, "vxw_encounter1_e6_0_0.vlm"},
	{0x2F352088A6812657ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_edge.phy"},
	{0x2F3557D24C2DB759ULL, "vxw_sundermount_30s_0_0.vlm"},
	{0x2F35815651A5942DULL, "83353122.ogg"},
	{0x2F35C9F905A6D840ULL, "vxw_encounter1_ir_0_0.vlm"},
	{0x2F36FF33B84EA569ULL, "vxdr_0_deeproads_cwm_0_0.vlm"},
	{0x2F37226E297BBE13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reaction.gad"},
	{0x2F37524D103E7E99ULL, "6122368_f.fxe"},
	{0x2F37820C5F60DE2DULL, "6065289_f.fxe"},
	{0x2F37C3656771C145ULL, "ob_glow.mao"},
	{0x2F38A80DAF7A4D64ULL, "750177052.ogg"},
	{0x2F38B56C6EDD4C0FULL, "vxw_encounter3_ga_0_0.vlm"},
	{0x2F38CD5C65FD4EF5ULL, "ca_caverns2prc_mmap_ph.dds"},
	{0x2F393E218F734ACCULL, "hlac_w_encounter2_camp_262149_0z.dds"},
	{0x2F3A2CFB0DF2D113ULL, "lt_0_lowtown_n_20_0.anb"},
	{0x2F3A5CBAE91E9561ULL, "lt_0_lowtown_45_0.anb"},
	{0x2F3AD04671A1A4FAULL, "6124289_m.fxe"},
	{0x2F3AD38E1BB65ACBULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\tre_stump01_n.dds"},
	{0x2F3B383036E52A01ULL, "dre000_bodahn6031362.cl"},
	{0x2F3BB2412FD22AA4ULL, "vxw_bonepit_11t_0_0.vlm"},
	{0x2F3BC174F6778491ULL, "30376648.ogg"},
	{0x2F3C0542D30C39C8ULL, "650306778.ogg"},
	{0x2F3C2D2280B0A14EULL, "ser120_ghyslain6108839.cl"},
	{0x2F3C55304BC1273FULL, "w_sundermount_155_0.phy"},
	{0x2F3CC813E12D0ED9ULL, "llac_w_blightlands_262147_0.dds"},
	{0x2F3CE195ADB2AB68ULL, "zz_dre_debug6168281.cl"},
	{0x2F3DC2FED6F18001ULL, "55796937.ogg"},
	{0x2F3DCFAEA2BF91F8ULL, "234305812.ogg"},
	{0x2F3DDF7D7F94723EULL, "vxht_keep_n_3il_0_0.vlm"},
	{0x2F3DFB7CF2225990ULL, "vxht_chantry_2yl_0_0.vlm"},
	{0x2F3E1116E10D552BULL, "mag160_feynriel6020433.cl"},
	{0x2F3E1F6E2981D878ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_reaction.evt"},
	{0x2F3E47906FE2A17EULL, "vxw_encounter2_im_0_0.vlm"},
	{0x2F3E86971D7E111DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_ex.ani"},
	{0x2F3F8315BDF7C5BAULL, "vxw_bonepitdestroyed_a3_lb_0_0.vlm"},
	{0x2F3FD6F4358DC4ECULL, "306151901.ogg"},
	{0x2F3FDF236FFC3FB0ULL, "vxlt_0_lowtown_docks_n_9fh_0_0.vlm"},
	{0x2F3FF4A4205A4DB6ULL, "hlac_ht_mansion2b_818_38y.dds"},
	{0x2F40139AA50F636AULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack2_c.anb"},
	{0x2F40290E612B1095ULL, "vxw_encounter1_fz_0_0.vlm"},
	{0x2F403D000DEFA7D2ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.dura.evt"},
	{0x2F40A64900C765DBULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.anb"},
	{0x2F40D0A9C88099EAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robc_0s.dds"},
	{0x2F412FC1BD807894ULL, "6132987_m.fxe"},
	{0x2F4197ABE4379C41ULL, "6016288_m.fxe"},
	{0x2F421B9E8909020FULL, "llac_lt_0_lowtown_docks_a3_-196607_0.dds"},
	{0x2F42D4E7F2680577ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhanggreen.msh"},
	{0x2F42DA41DFF0F070ULL, "vxw_encounter1_j7_0_0.vlm"},
	{0x2F42E6E7F26823DAULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhanggreen.mao"},
	{0x2F42EAE7F2682AE9ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhanggreen.mmh"},
	{0x2F434DFFDCFF925DULL, "375214870.ogg"},
	{0x2F439D7004367AE1ULL, "vxdr_0_deeproads_ew7_0_0.vlm"},
	{0x2F43BC21D32822D4ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear3.ani"},
	{0x2F43F591D0B94EC8ULL, "vxlt_0_lowtown_docks_n_b4x_0_1.vlm"},
	{0x2F447FDEB5156A9DULL, "6163672_m.fxe"},
	{0x2F45213911385501ULL, "353037079.ogg"},
	{0x2F455634ECB99862ULL, "vxht_bartrand_n_6qn_0_0.vlm"},
	{0x2F457110AA44BDE9ULL, "vxdr_0_deeproads_cll_0_0.vlm"},
	{0x2F45B5646CEF2AE9ULL, "hlac_lt_backalley_n_0_0z.dds"},
	{0x2F45D43F6A939F62ULL, "202196873.ogg"},
	{0x2F465CF2E94B5837ULL, "6110272_m.fxe"},
	{0x2F4669CB9D0C4A13ULL, "gen00fl_sebastian6189424.cl"},
	{0x2F467087BA5C3B9DULL, "112706125.ogg"},
	{0x2F46A1760D52AD13ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.anb"},
	{0x2F46EE868CA7CE59ULL, "6164314_m.fxe"},
	{0x2F4716E25F5F3A8FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_pointing.gad"},
	{0x2F472092246EAFA3ULL, "6125939_m.fxe"},
	{0x2F475C000DF5FB65ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.dura.gad"},
	{0x2F48056DC5F3FAA5ULL, "6109409_m.fxe"},
	{0x2F48247437310EE1ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower3_meredithsintro_varric.ani"},
	{0x2F4832EE53AF1A9EULL, "lt_gallowscourtyard_59_0_lc.mmh"},
	{0x2F4949DF1FBBF060ULL, "lowtown_male_beggar_sta6151070.cl"},
	{0x2F49B1E74CF2C006ULL, "vxw_encounter2_ln_0_0.vlm"},
	{0x2F49B37D8002A231ULL, "368280868.ogg"},
	{0x2F49D34492CAE94FULL, "plt_lgt000pt_lite_vials.plo"},
	{0x2F4A0533FF86ED30ULL, "6179506_m.fxe"},
	{0x2F4A5AE009690138ULL, "one300dg_solivitus6060359.cl"},
	{0x2F4AB921D32E3C36ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear3.gad"},
	{0x2F4B62DFA02EA3BDULL, "6076398_m.fxe"},
	{0x2F4B89599B405EAFULL, "spt_billboards0.vsh.fxo"},
	{0x2F4C17C2A781B036ULL, "538569589.ogg"},
	{0x2F4C2022D6119F04ULL, "948645893.ogg"},
	{0x2F4C714ACFEE2805ULL, "vxht_0_hightown_b8r_0_0.vlm"},
	{0x2F4CC0F228F2E532ULL, "187711256.ogg"},
	{0x2F4CF88F6C68D185ULL, "114370131.ogg"},
	{0x2F4D2BE25F63CA3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_pointing.evt"},
	{0x2F4D56304BCF9C44ULL, "w_sundermount_155_0.mmh"},
	{0x2F4D5A304BCFA297ULL, "w_sundermount_155_0.mao"},
	{0x2F4D6C304BCFC1E6ULL, "w_sundermount_155_0.msh"},
	{0x2F4E38B7D2C76273ULL, "54573681.ogg"},
	{0x2F4EC88322794E45ULL, "818822172.ogg"},
	{0x2F4F2F7437373F83ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower3_meredithsintro_varric.gad"},
	{0x2F50016507FF854BULL, "dae000_pickpocket_scene6175094.cl"},
	{0x2F50418F2133828CULL, "character_character_p238.psh.fxo.dx11.assembly"},
	{0x2F506FB1D020A06FULL, "vxlt_0_lowtown_a2_him_0_0.vlm"},
	{0x2F50C81991A95AE1ULL, "306977522.ogg"},
	{0x2F50D6659B5CBA7FULL, "vxlt_warehouse_a3_0_0.vlm"},
	{0x2F515AB526B2376BULL, "llac_lt_0_lowtown_a2_n_1_4.dds"},
	{0x2F517016CD20CF9BULL, "585763481.ogg"},
	{0x2F51920E45184112ULL, "vxht_brothel_n_5jp_0_0.vlm"},
	{0x2F521921DB15B561ULL, "34855031.ogg"},
	{0x2F5262BFCDA57C5EULL, "character_character_p102.psh.fxo.dx11"},
	{0x2F528F8105ED14C9ULL, "art\\characters\\weapons\\longswords\\w_lsw_023a.phy"},
	{0x2F529B430FB5B5D4ULL, "zz_isa_debug6133935.cl"},
	{0x2F52F37C508F850EULL, "6202024_m.fxe"},
	{0x2F532B4655DB46D4ULL, "atl_graphicsde_dxt5_dat.xml"},
	{0x2F53AC9539835A08ULL, "117336625.ogg"},
	{0x2F53B1E74A1B838FULL, "vxht_mansion2_32k_0_0.vlm"},
	{0x2F53BE760D5D697CULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.mmh"},
	{0x2F53D048BBA7B3B0ULL, "hlac_lt_0_lowtown_n_65533_3y.dds"},
	{0x2F53E2ED9C3AD669ULL, "vxdr_0_deeproads_cis_0_0.vlm"},
	{0x2F5458D45ABC7EADULL, "vxw_encounter2_6c_0_0.vlm"},
	{0x2F54B1389BCEB0D4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_quick_modal_activate.ani"},
	{0x2F550E6A00F567D7ULL, "lt_0_lowtown_docks_a3_169_0.mmh"},
	{0x2F55126A00F56EE4ULL, "lt_0_lowtown_docks_a3_169_0.mao"},
	{0x2F551CE6B5F4EEC3ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetalb_d.dds"},
	{0x2F55246A00F58D49ULL, "lt_0_lowtown_docks_a3_169_0.msh"},
	{0x2F55A9CED74C5968ULL, "additivestreaks01.mao"},
	{0x2F5650BC6C06BE46ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_e.dds"},
	{0x2F56F55081054BF4ULL, "zz_vault_debug6203189.cl"},
	{0x2F5721D5F7B85AC2ULL, "lgt110_arl_bann_teagan6115623.cl"},
	{0x2F582751E33D6492ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl180runf.gad"},
	{0x2F588A8F704ED545ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\baseswordshield\\blendtree\\commander_movement.anb"},
	{0x2F58E18B20A3EFAAULL, "6028927_m.fxe"},
	{0x2F58F9A68982F4DBULL, "114059410.ogg"},
	{0x2F5906F5D4B5FA81ULL, "fx_lag.mao"},
	{0x2F5949E78970EB37ULL, "498300823.ogg"},
	{0x2F598419405A799CULL, "878321926.ogg"},
	{0x2F598C33BBFAFD93ULL, "llac_ht_keep_4519_16.dds"},
	{0x2F59B18D7DE2EF48ULL, "6117974_f.fxe"},
	{0x2F59FAD09A3D3ED0ULL, "6104890_m.fxe"},
	{0x2F5B33E25F701555ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_pointing.ani"},
	{0x2F5B4E1E3BD958D3ULL, "hlac_ht_0_hightown_-65536_3y.dds"},
	{0x2F5BABB8DF67D865ULL, "gen_im_trs_crc_junk3_03.uti"},
	{0x2F5BAE389BD4CA36ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_quick_modal_activate.gad"},
	{0x2F5CC23920980FB3ULL, "gen00fl_merrill6129555.cl"},
	{0x2F5CC39471C442B3ULL, "6048267_m.fxe"},
	{0x2F5CFE4B241BA740ULL, "lt_0_lowtown_docks_n_150_0.phy"},
	{0x2F5D6BE60158FBF7ULL, "379008.ogg"},
	{0x2F5DF16C5D52A214ULL, "lgt310cr_mis_burn_murder1.utc"},
	{0x2F5E1521C2F6CA8EULL, "lt_0_lowtown_docks_a3_n_196_0.mmh"},
	{0x2F5E2121C2F6DF25ULL, "lt_0_lowtown_docks_a3_n_196_0.mao"},
	{0x2F5E2F21C2F6F77CULL, "lt_0_lowtown_docks_a3_n_196_0.msh"},
	{0x2F5E3D86E89C202BULL, "6051402_f.fxe"},
	{0x2F5E7B61A1DF85D8ULL, "hlac_ca_caverns1_24424_0y.dds"},
	{0x2F5EB3D76B922169ULL, "vxht_chantry_2em_0_0.vlm"},
	{0x2F5EBD128C5D7075ULL, "llac_lt_undercityserialkiller_-1_25.dds"},
	{0x2F5EE61B1864AB9CULL, "mag311_keran6061827.cl"},
	{0x2F5F2A51E34388C9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl180runf.evt"},
	{0x2F5FC6CE1379450FULL, "vxlt_warehouse_sy_0_0.vlm"},
	{0x2F5FE038CC94D8F5ULL, "hlac_lt_undercitydungeon_-1_37z.dds"},
	{0x2F5FE587E4DA8CB7ULL, "hla_ht_0_hightown_a2_65534_0x.dds"},
	{0x2F604870E8467190ULL, "fex_bd_a_rf_cn.msh"},
	{0x2F605670E84689AAULL, "fex_bd_a_rf_cn.mmh"},
	{0x2F60D9978EF5510DULL, "6105786_m.fxe"},
	{0x2F60E4B84234530EULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_add_scl.mao"},
	{0x2F60FEE3D0434763ULL, "art\\characters\\weapons\\longswords\\w_lsw_hlka.mao"},
	{0x2F6102E3D0434DB0ULL, "art\\characters\\weapons\\longswords\\w_lsw_hlka.mmh"},
	{0x2F6110E3D04365CAULL, "art\\characters\\weapons\\longswords\\w_lsw_hlka.msh"},
	{0x2F61223F0B772C02ULL, "gen00fl_anders6119391.cl"},
	{0x2F614A5E9DC3752FULL, "780529477.ogg"},
	{0x2F614D29116D160EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedidle.evt"},
	{0x2F61543B343B12A2ULL, "6171939_m.fxe"},
	{0x2F61A27889E6DAB1ULL, "art\\characters\\creatures\\ragedemon\\animation\\rdy.gad"},
	{0x2F61C89CF6113FEAULL, "69005989.ogg"},
	{0x2F61C9389BD96395ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_quick_modal_activate.evt"},
	{0x2F61DA080FF089CEULL, "vxlt_warehouse_e2_0_0.vlm"},
	{0x2F620DA70D22442BULL, "docks_male_elven_worker_sta6176101.cl"},
	{0x2F6226FF581C4CE2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medd_0n.dds"},
	{0x2F6254CA8E30EEE9ULL, "vxdr_0_deeproads_cfr_0_0.vlm"},
	{0x2F62C2062994BB2DULL, "lgt260_eau_tmp_conrad6143792.cl"},
	{0x2F62CAB14CB2972DULL, "vxw_encounter2_1b_0_0.vlm"},
	{0x2F62CEBC27340948ULL, "6163512_m.fxe"},
	{0x2F630C9ADFD0DAD2ULL, "187273967.ogg"},
	{0x2F6399200F09C6B8ULL, "vxlt_0_lowtown_a3_n_c_hjk_0_0.vlm"},
	{0x2F6437E4E3FAFE94ULL, "lt_gallowstemplar_n_2384.rml"},
	{0x2F645F73593C182FULL, "237972311.ogg"},
	{0x2F64A97EDF699FCCULL, "vxlt_0_lowtown_docks_n_c6n_0_0.vlm"},
	{0x2F64C9F8F39F84E0ULL, "760713440.ogg"},
	{0x2F64F273734C4996ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_trqa_0n.dds"},
	{0x2F64FC34B2692278ULL, "177221592.ogg"},
	{0x2F6548A76C7995B8ULL, "158165570.ogg"},
	{0x2F65628A44C0717CULL, "227987423.ogg"},
	{0x2F65A90E1F3CB855ULL, "6149126_m.fxe"},
	{0x2F6649FB0E183DCFULL, "lt_0_lowtown_n_20_0.rml"},
	{0x2F6695A130DEF106ULL, "vxw_encounter2_rp_0_0.vlm"},
	{0x2F66E8EDE3731E30ULL, "6177706_m.fxe"},
	{0x2F676C815E5C9537ULL, "qun230cr_guild_contact6068866.cl"},
	{0x2F684429117325B9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedidle.gad"},
	{0x2F68A37889ECFB5FULL, "art\\characters\\creatures\\ragedemon\\animation\\rdy.ani"},
	{0x2F68A9448441BC79ULL, "vxca_caverns1_igp_0_0.vlm"},
	{0x2F68CAC89550F52DULL, "6111808_m.fxe"},
	{0x2F68EB4B90150184ULL, "llac_lt_0_lowtown_n_-262144_2.dds"},
	{0x2F69169AA5326D5DULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack2_c.mmh"},
	{0x2F69E6985818439BULL, "241309309.ogg"},
	{0x2F6A64E0F3997A03ULL, "gen00fl_carver6097339.cl"},
	{0x2F6A8339C6B86ADCULL, "234898565.ogg"},
	{0x2F6AA4E485653F65ULL, "6051902_m.fxe"},
	{0x2F6B357163BA5B33ULL, "6043367_m.fxe"},
	{0x2F6B4216FDD1AAA5ULL, "222426740.ogg"},
	{0x2F6B5247F0B90594ULL, "vxht_0_hightown_n_c6z_0_1.vlm"},
	{0x2F6B885226DD85CBULL, "6125796_f.fxe"},
	{0x2F6BDC12CCD4A117ULL, "hlac_ht_keep_n_3732_7y.dds"},
	{0x2F6C1FD1919E4D44ULL, "hlac_w_bonepit_327683_0y.dds"},
	{0x2F6C283FECC1E172ULL, "llac_w_blightlands_0_5.dds"},
	{0x2F6CDA07D5872481ULL, "vxht_mansion2b_24z_0_0.vlm"},
	{0x2F6D1251E34F9D38ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl180runf.ani"},
	{0x2F6D1B19F38E56F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_exit.ani"},
	{0x2F6D708123E5C5B4ULL, "dae000cr_lady_elegant.utc"},
	{0x2F6D979D7ACF22B6ULL, "537831984.ogg"},
	{0x2F6DA120A7D065F2ULL, "vxht_0_hightown_n_9zc_0_0.vlm"},
	{0x2F6DB25D0DB2B96AULL, "vxlt_gallowscourtyard_n_4vc_0_0.vlm"},
	{0x2F6DC17AF4F36A0FULL, "1029518700.ogg"},
	{0x2F6DD6FB34B54529ULL, "fen121ar_base.are"},
	{0x2F6E38AB056F5D8FULL, "vxlt_warehouse_tx_0_0.vlm"},
	{0x2F6E944356BFE6AFULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isa.mao"},
	{0x2F6E94FF965486A7ULL, "llac_w_encounter1_65535_0.dds"},
	{0x2F6EA04356BFFB4CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isa.mmh"},
	{0x2F6EA64356C0053EULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isa.msh"},
	{0x2F6EC27E25FF82ACULL, "vxlt_0_lowtown_a2_n_c_hoe_0_0.vlm"},
	{0x2F6EDC20351DF306ULL, "436590991.ogg"},
	{0x2F6EF306DA18C461ULL, "vxdr_0_deeproads_ef4_0_0.vlm"},
	{0x2F6F005381846850ULL, "205006197.ogg"},
	{0x2F6F552911796107ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandedidle.ani"},
	{0x2F70047F55F020C1ULL, "l_fen321ar_base.wbd"},
	{0x2F7008964F47B8B5ULL, "791302407.ogg"},
	{0x2F707D61A4BE53CFULL, "qun200_party_comment6090344.cl"},
	{0x2F70C15939AB1F2FULL, "vxht_mansion2b_3ef_0_0.vlm"},
	{0x2F7117AB24BFB927ULL, "485572358.ogg"},
	{0x2F718FEA68ECE42DULL, "vxw_blightlands_fl_0_0.vlm"},
	{0x2F71EFD3F3453A88ULL, "ser270_templar6122514.cl"},
	{0x2F723651CB3DE4D4ULL, "6206884_m.fxe"},
	{0x2F72A6D385CFA19AULL, "qun221_viscount6085785.cl"},
	{0x2F7307B5FEAE41C0ULL, "495455862.ogg"},
	{0x2F73689096EAD700ULL, "llac_w_encounter2_131072_0.dds"},
	{0x2F73F619F394363BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_exit.gad"},
	{0x2F743DBAE94FD8F1ULL, "lt_0_lowtown_45_0.rml"},
	{0x2F7523C1A061C824ULL, "vxlt_0_lowtown_docks_a3_co9_0_0.vlm"},
	{0x2F7588453E682B16ULL, "vxw_bonepitmining_a2_1i0_0_0.vlm"},
	{0x2F7641B128C3012FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowalcove.mao"},
	{0x2F764DB128C315CCULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowalcove.mmh"},
	{0x2F7653B128C31FBEULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowalcove.msh"},
	{0x2F768C1F4FE6D747ULL, "552611775.ogg"},
	{0x2F769F705D355147ULL, "dae000_epic_merchant_26179387.cl"},
	{0x2F77194DD643082EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xb.phy"},
	{0x2F771B217637D4F9ULL, "vxca_caverns1_ijw_0_0.vlm"},
	{0x2F77A05C4A6043B1ULL, "762263047.ogg"},
	{0x2F77EFF95CDB1213ULL, "vxht_mansion2_2ew_0_0.vlm"},
	{0x2F77F67ED0906053ULL, "zz_dog_debug6127794.cl"},
	{0x2F78019049B779B9ULL, "vxlt_0_lowtown_a2_n_gph_0_0.vlm"},
	{0x2F78B0FF5F39639EULL, "6071943_m.fxe"},
	{0x2F78C97A8EA98AE5ULL, "768650267.ogg"},
	{0x2F798B5EFE64486BULL, "810810643.ogg"},
	{0x2F7A76A270D7B8D0ULL, "gen00fl_anders6114505.cl"},
	{0x2F7ACE3BF84BA39AULL, "follower_banter6161294.cl"},
	{0x2F7B130C62F9C8A5ULL, "zz_qun_debug6028279.cl"},
	{0x2F7B1319F39A8620ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_exit.evt"},
	{0x2F7B2B7619DB7DA1ULL, "humanfemale_flemeth.fxa"},
	{0x2F7BBA2439B0ADD8ULL, "6114845_f.fxe"},
	{0x2F7BDFADE75BEBBDULL, "dre000_varric6035141.cl"},
	{0x2F7BFAF401300FA3ULL, "vxca_caverns1_dpb_0_0.vlm"},
	{0x2F7C1EF461EA3052ULL, "6119645_m.fxe"},
	{0x2F7C338C5467B391ULL, "vxht_mansion_n_4d3_0_0.vlm"},
	{0x2F7C4928885AEAA9ULL, "6169753_m.fxe"},
	{0x2F7CA72A3F3E6C6AULL, "6042408_m.fxe"},
	{0x2F7CE1244348E361ULL, "dae300_alistair_king6133784.cl"},
	{0x2F7CF30AD38E8CF8ULL, "gen00fl_isabela6075260.cl"},
	{0x2F7CF588A6BE92AFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_edge.mao"},
	{0x2F7D0188A6BEA74CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_edge.mmh"},
	{0x2F7D0788A6BEB13EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_edge.msh"},
	{0x2F7D1F6BFDA2DCB6ULL, "vxlt_0_lowtown_a2_n_d66_0_0.vlm"},
	{0x2F7D438A539299AFULL, "hlac_lt_0_lowtown_docks_65534_2z.dds"},
	{0x2F7D4E0D26D3FC53ULL, "6034516_m.fxe"},
	{0x2F7DD592110FDCAEULL, "hlac_ca_caverns2prc_5802_64z.dds"},
	{0x2F7EA4299965ACF0ULL, "845031841.ogg"},
	{0x2F7EB62262ACC728ULL, "6074062_f.fxe"},
	{0x2F7F39A9C3947C51ULL, "6148648_m.fxe"},
	{0x2F7F44BE71DE378DULL, "mag170_thrask.bnk"},
	{0x2F7F676A0632835AULL, "vxlt_hangedman_20g_0_0.vlm"},
	{0x2F7F78810613DBEEULL, "art\\characters\\weapons\\longswords\\w_lsw_023a.mmh"},
	{0x2F7F84810613EF85ULL, "art\\characters\\weapons\\longswords\\w_lsw_023a.mao"},
	{0x2F7F8B54BCEA7050ULL, "6149880_m.fxe"},
	{0x2F7F90B1178B207EULL, "912421168.ogg"},
	{0x2F7F9281061407DCULL, "art\\characters\\weapons\\longswords\\w_lsw_023a.msh"},
	{0x2F7FACC7421201A8ULL, "character_character_p54.psh.fxo.dx11"},
	{0x2F7FC76E7915684AULL, "35395102.ogg"},
	{0x2F80118E99CE9BE7ULL, "lit_bt_dust.mao"},
	{0x2F806BB3731C5406ULL, "6129744_f.fxe"},
	{0x2F8076FD1BD75123ULL, "dae000_bodhan_letter6144316.cl"},
	{0x2F80C7AD446BC1C0ULL, "vxw_bonepitdestroyed_a3_26b_0_0.vlm"},
	{0x2F80CC0BB2751DA1ULL, "785855940.ogg"},
	{0x2F80F351D3BAB9D2ULL, "dae000_oldmangossip6174764.cl"},
	{0x2F811EBFE12DDC18ULL, "6018658_m.fxe"},
	{0x2F817C31C81AA6B3ULL, "hlac_lt_gallowscourtyard_n_-196608_0y.dds"},
	{0x2F81C74BC9390433ULL, "700742726.ogg"},
	{0x2F8213A246BAC7CFULL, "vxht_0_hightown_cen_0_0.vlm"},
	{0x2F8213B52D03F145ULL, "6129160_m.fxe"},
	{0x2F821B4B243AF4D5ULL, "lt_0_lowtown_docks_n_150_0.msh"},
	{0x2F821D4B243AF84BULL, "lt_0_lowtown_docks_n_150_0.mmh"},
	{0x2F82294B243B0CE8ULL, "lt_0_lowtown_docks_n_150_0.mao"},
	{0x2F826AA8DBC9CD90ULL, "125139922.ogg"},
	{0x2F829321E0AA5D1AULL, "6096631_f.fxe"},
	{0x2F82CCB778535679ULL, "vxht_mansion2_314_0_0.vlm"},
	{0x2F8301B0FD71C04AULL, "hlac_ht_chantry_2534_43x.dds"},
	{0x2F8366BE71E25AF5ULL, "mag170_thrask.cnv"},
	{0x2F83984D94406D19ULL, "mcr310_sibling6076359.cl"},
	{0x2F83C449BAB6309CULL, "vxlt_0_lowtown_docks_a3_d0s_0_0.vlm"},
	{0x2F83E38FDEACB5EEULL, "llac_ca_caverns2prc_5802_4.dds"},
	{0x2F83EF915C91F41EULL, "6100078_m.fxe"},
	{0x2F8402CC812EFC76ULL, "6052085_m.fxe"},
	{0x2F8422860F1E4ABFULL, "550721889.ogg"},
	{0x2F84404D29F389CDULL, "6135756_m.fxe"},
	{0x2F8475D89E2E6A54ULL, "vxht_0_hightown_bcb_0_0.vlm"},
	{0x2F84E06634201D08ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basicxend.msh"},
	{0x2F84F26634203B99ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basicxend.mao"},
	{0x2F84FE6634205032ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basicxend.mmh"},
	{0x2F853BF12D7F0FDBULL, "142326816.ogg"},
	{0x2F858CFE682DED79ULL, "vxca_caverns1_imv_0_0.vlm"},
	{0x2F858E11BB461163ULL, "6129203_m.fxe"},
	{0x2F85904EF219ACDDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masf.phy"},
	{0x2F860C775E7DB3FEULL, "832124839.ogg"},
	{0x2F8661D64ED12A93ULL, "vxht_mansion2_2nv_0_0.vlm"},
	{0x2F866B1A9CBF3120ULL, "isa211_party_reaction6083425.cl"},
	{0x2F86736D3BAD9239ULL, "vxlt_0_lowtown_a2_n_gmo_0_0.vlm"},
	{0x2F87792A8C4492E6ULL, "vxht_0_hightown_a3_c52_0_0.vlm"},
	{0x2F87CC5B652CDBA7ULL, "326345903.ogg"},
	{0x2F87FE8D0A2377CCULL, "6110859_m.fxe"},
	{0x2F880ED779BDFC59ULL, "6113059_m.fxe"},
	{0x2F885C9544CDE4EAULL, "6177894_m.fxe"},
	{0x2F88833949714011ULL, "hlac_lt_backalley_n_65537_0x.dds"},
	{0x2F89162209DE8DB6ULL, "vxlt_0_lowtown_a2_n_f66_0_2.vlm"},
	{0x2F893EF7474575E9ULL, "gen_im_trs_dem_lgt_boo_r2.uti"},
	{0x2F897FF8CA43E277ULL, "art\\vfx\\combat\\vfx_c_debili_aoe_imp_p.impa.gad"},
	{0x2F89D096226E3CA2ULL, "gen_im_delete_me_08.uti"},
	{0x2F8A0008EE52B0D8ULL, "vxlt_darktown_5ji_0_0.vlm"},
	{0x2F8AA5C74EF40535ULL, "vxw_woundedcoast_1ct_0_0.vlm"},
	{0x2F8ADB5F2B791296ULL, "llac_ht_0_hightown_-65537_4.dds"},
	{0x2F8B09BCA0925176ULL, "386227309.ogg"},
	{0x2F8BA3636B3653D0ULL, "499051347.ogg"},
	{0x2F8BCD45AE0838E9ULL, "6089398.ogg"},
	{0x2F8CD443CA11AF7EULL, "art\\vfx\\combat\\material__35.mao"},
	{0x2F8CE4AC6C4AF2B7ULL, "isa311_castillon6092130.cl"},
	{0x2F8D1A74FFF38D72ULL, "280979586.ogg"},
	{0x2F8D34E219D999EFULL, "vxht_0_hightown_a2_b9k_0_0.vlm"},
	{0x2F8D42170242AB0BULL, "88015322.ogg"},
	{0x2F8D73CE7E1A4D3AULL, "dae021_sister_colinda_sta6150198.cl"},
	{0x2F8D9D2ACE258703ULL, "dae000ip_crowd_ff3_talk16.utp"},
	{0x2F8DB9485B67947BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_isabella.anb"},
	{0x2F8DBC57BB668B01ULL, "gen00fl_merrill6151006.cl"},
	{0x2F8DBF5F3E38BBC9ULL, "6020937_m.fxe"},
	{0x2F8DCD20B229E451ULL, "vxlt_gallowstemplar_3mg_0_0.vlm"},
	{0x2F8E0F5A220C94E3ULL, "760536178.ogg"},
	{0x2F8E73A44CD9152DULL, "vxw_blightlands_lj_0_0.vlm"},
	{0x2F8EB87477677811ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar1_.gad"},
	{0x2F8EBD4562C3219BULL, "6083346_f.fxe"},
	{0x2F8F398A3661DA40ULL, "vxw_bonepitdestroyed_a3_21c_0_0.vlm"},
	{0x2F8F7C58FD953E91ULL, "6018664_m.fxe"},
	{0x2F8FA7899254E6A3ULL, "mag170_thrask6020552.cl"},
	{0x2F906FBBC9A389DBULL, "tbe100ip_shop_codex_8.utp"},
	{0x2F9094F8CA4A2404ULL, "art\\vfx\\combat\\vfx_c_debili_aoe_imp_p.impa.evt"},
	{0x2F909FCB05324344ULL, "vxlt_darktown_5bm_0_0.vlm"},
	{0x2F90B3666D219D22ULL, "324874862.ogg"},
	{0x2F9111F56E88683DULL, "zz_dae_debug6029894.cl"},
	{0x2F9118A03C813EF3ULL, "780413273.ogg"},
	{0x2F92913F2EE7B27EULL, "one280_iwan6097156.cl"},
	{0x2F92A2C1BD3D9DC5ULL, "l_isa211ar_foundry.wbd"},
	{0x2F92DBD8AB912303ULL, "zz_jumpcam.ncs"},
	{0x2F93119772D5BBEEULL, "6148618_m.fxe"},
	{0x2F93AF05CAF62F82ULL, "6044002_m.fxe"},
	{0x2F93C7FB34D5FB7FULL, "fen121ar_base.ncs"},
	{0x2F93F24493095634ULL, "plt_lgt000pt_lite_vials.nss"},
	{0x2F9471626078C071ULL, "6136836_m.fxe"},
	{0x2F949AC234FF960AULL, "seb221ip_door_block_barrel.utp"},
	{0x2F95267F560F76C3ULL, "l_fen321ar_base.bnk"},
	{0x2F952A85DEF36580ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut3flat.phy"},
	{0x2F952B1F248C5501ULL, "6108752_m.fxe"},
	{0x2F954427D907E2FAULL, "pssmcanopy.psh.fxo"},
	{0x2F95461A45EB68FEULL, "vxw_blightlands_7h_0_0.vlm"},
	{0x2F95475CA9A4A03AULL, "808208149.ogg"},
	{0x2F957AFA0442FEB0ULL, "hlac_lt_0_lowtown_n_-65536_0z.dds"},
	{0x2F95B974776D98BFULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar1_.ani"},
	{0x2F95D87762F836F9ULL, "463976006.ogg"},
	{0x2F95FD5A2C955EC8ULL, "6077337_f.fxe"},
	{0x2F95FD69390D31DEULL, "vxca_caverns1_eae_0_0.vlm"},
	{0x2F96854EF2280D49ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masf.mao"},
	{0x2F96914EF22821E2ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masf.mmh"},
	{0x2F96934EF2282498ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masf.msh"},
	{0x2F96E7DC3FD53144ULL, "dae000_gamlen6149933.cl"},
	{0x2F96F04821DF4276ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_20ring.mao"},
	{0x2F96FC4821DF5615ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_20ring.mmh"},
	{0x2F96FE4821DF598BULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_20ring.msh"},
	{0x2F970F46E2B5F0B3ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pulpit.phy"},
	{0x2F97520B132407A4ULL, "vxlt_0_lowtown_docks_b7v_0_1.vlm"},
	{0x2F9769633A274BA4ULL, "966665356.ogg"},
	{0x2F97C518AAA8B2C9ULL, "vxht_0_hightown_ba9_0_0.vlm"},
	{0x2F97DAA676BC4F8FULL, "70143635.ogg"},
	{0x2F983C38A097E178ULL, "889077775.ogg"},
	{0x2F986A56854B6C6FULL, "vxca_caverns2_838_0_0.vlm"},
	{0x2F98FFFBBA1099A7ULL, "479779697.ogg"},
	{0x2F9917A440EA1DB5ULL, "vxw_woundedcoast_1ns_0_0.vlm"},
	{0x2F9924CA212FD70CULL, "art\\characters\\playercharacter\\humanmale\\glv_ric.msh"},
	{0x2F992ACA212FE1FEULL, "art\\characters\\playercharacter\\humanmale\\glv_ric.mmh"},
	{0x2F9936CA212FF595ULL, "art\\characters\\playercharacter\\humanmale\\glv_ric.mao"},
	{0x2F9959BEE4B28368ULL, "vxht_mansion_n_4xr_0_0.vlm"},
	{0x2F99C184C7AEA140ULL, "hlac_lt_undercityserialkiller_-1_46z.dds"},
	{0x2F99C2F0D5A66241ULL, "609974598.ogg"},
	{0x2F9AE7FB34DC506EULL, "fen121ar_base.lst"},
	{0x2F9BA17BD7C8EBCEULL, "vxlt_warehouse_y6_0_0.vlm"},
	{0x2F9C11C365C2163BULL, "dae000_pc_home_ambients6208698.cl"},
	{0x2F9C371ED608AFB0ULL, "243962657.ogg"},
	{0x2F9CEC0E16B97C54ULL, "z_cloud02.dds"},
	{0x2F9CF9FC5B9AE568ULL, "vxlt_gallowsprison_n_6mp_0_0.vlm"},
	{0x2F9D48FF4BFF6EE9ULL, "atl_startmenu2_dxt5_01.dds"},
	{0x2F9DACF8CA54D7FDULL, "art\\vfx\\combat\\vfx_c_debili_aoe_imp_p.impa.ani"},
	{0x2F9DF04F059356B1ULL, "vxca_caverns1_mines_j47_0_0.vlm"},
	{0x2F9E1E4BE4AA9E89ULL, "vxlt_gallowstemplar_f_5wp_0_0.vlm"},
	{0x2F9EB5C04CBDEB24ULL, "534562353.ogg"},
	{0x2F9F1A010FAECE5DULL, "6132890_m.fxe"},
	{0x2F9F605844B8B1F4ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabv2.evt"},
	{0x2F9F7208C5A2D098ULL, "210129959.ogg"},
	{0x2F9FE06C736FF69DULL, "hlac_w_sundermount_-131077_0x.dds"},
	{0x2FA0DDDC144A7496ULL, "vxw_bonepitmining_a2_1x3_0_0.vlm"},
	{0x2FA0DE79E79A8C09ULL, "mcr310_argument6182273.cl"},
	{0x2FA109E3D0793DCBULL, "art\\characters\\weapons\\longswords\\w_lsw_hlka.phy"},
	{0x2FA1504EC50C4AB5ULL, "6175006_m.fxe"},
	{0x2FA18A51AA04B6BBULL, "6171835_m.fxe"},
	{0x2FA190B4CF4F781FULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw103a.mao"},
	{0x2FA1A394399D639AULL, "6126342_m.fxe"},
	{0x2FA1C00DCD027D09ULL, "hla_ht_0_hightown_n_131073_0y.dds"},
	{0x2FA1C0A6556BFD96ULL, "414644407.ogg"},
	{0x2FA1F777322E09B4ULL, "vxca_caverns2_5yx_0_0.vlm"},
	{0x2FA2586F28640843ULL, "6083547_m.fxe"},
	{0x2FA2AA79824F9021ULL, "933626912.ogg"},
	{0x2FA2ADA1914A38EDULL, "855763038.ogg"},
	{0x2FA2E5E0637DC824ULL, "mag311_emile_de_launcet6043498.cl"},
	{0x2FA3B7F737E1817EULL, "vxw_blightlands_8i_0_0.vlm"},
	{0x2FA430E88E5D678DULL, "hightown_market_guard_sta6169661.cl"},
	{0x2FA4416CA0D09C6BULL, "944945594.ogg"},
	{0x2FA4B29FE10B3AB6ULL, "vxht_keep_49d_0_0.vlm"},
	{0x2FA5C5F9DBD1C866ULL, "zz_one_debug6099762.cl"},
	{0x2FA5DB7066AD60AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.gad"},
	{0x2FA6325CF6A77FA9ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.dura.ani"},
	{0x2FA682DFBA94C4AFULL, "6071076_m.fxe"},
	{0x2FA6DE95B4A4BEE3ULL, "6085479_m.fxe"},
	{0x2FA6E3C2D23EE1D8ULL, "vxlt_darktown_5pg_0_0.vlm"},
	{0x2FA7158FA03BABC7ULL, "584028853.ogg"},
	{0x2FA7262C61F71535ULL, "6096600_m.fxe"},
	{0x2FA746A56B689A9FULL, "vxht_playershouse_n_1rk_0_0.vlm"},
	{0x2FA76B7D9166E8F3ULL, "art\\characters\\weapons\\spears\\textures\\w_qui_qun_0n.dds"},
	{0x2FA7E1B61F83195BULL, "6201468_m.fxe"},
	{0x2FA8AA2D468E710DULL, "vxlt_0_lowtown_docks_brq_0_0.vlm"},
	{0x2FA906F3FA69DFEDULL, "dae000_epic_merchant_46179530.cl"},
	{0x2FA95D2EB229DB7AULL, "dae000_merchant_hm_mob6174219.cl"},
	{0x2FA9716AB1D0C310ULL, "6204972_m.fxe"},
	{0x2FA9B6CA79A9792DULL, "946216029.ogg"},
	{0x2FAA1C97A5A45C86ULL, "1024672554.ogg"},
	{0x2FAA55EFBD8FC17EULL, "228376965.ogg"},
	{0x2FAA71E6E55FEEFDULL, "w_woundedcoast_56_0.rml"},
	{0x2FAA79BF800E226AULL, "138301372.ogg"},
	{0x2FAB31B8C54997F5ULL, "vrc321_varric6096942.cl"},
	{0x2FAB3BDF428C65F8ULL, "294547115.ogg"},
	{0x2FAB438DC0604D6DULL, "vxlt_0_lowtown_a2_n_ffk_0_0.vlm"},
	{0x2FAB44D98F919316ULL, "vxht_playershouse_n_1lp_0_0.vlm"},
	{0x2FAB4932614B0280ULL, "hlac_ht_0_hightown_1_1y.dds"},
	{0x2FAC124CDB47DDFCULL, "867646070.ogg"},
	{0x2FAC25B53FA5575FULL, "1029756036.ogg"},
	{0x2FAC2BBBBC399B90ULL, "577880379.ogg"},
	{0x2FAC65BF04500EA9ULL, "vxht_mansion2_1un_0_0.vlm"},
	{0x2FAC8ED19C5467DFULL, "vxw_bonepitdestroyed_a3_n1_0_0.vlm"},
	{0x2FACBE7066B34E41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.evt"},
	{0x2FACC700162FE76EULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumd.mmh"},
	{0x2FACD300162FFB05ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumd.mao"},
	{0x2FACE1001630135CULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumd.msh"},
	{0x2FAD4D5CF6ADCC2BULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.dura.gad"},
	{0x2FADB81B10DAF91EULL, "6153261_m.fxe"},
	{0x2FADCFF8E219EAE3ULL, "dae000_junar6147912.cl"},
	{0x2FADDFEDD4F10537ULL, "68590012.ogg"},
	{0x2FAE057D8FF93AF8ULL, "hlac_lt_hangedman_-1_7x.dds"},
	{0x2FAE717BBD894445ULL, "genip_book_1.utp"},
	{0x2FAE9EDACAA1D9BCULL, "385074993.ogg"},
	{0x2FAEC845FB20842CULL, "dae000_gamlen6145094.cl"},
	{0x2FAF116C2ADAFD7EULL, "vxht_brothel_n_6kp_0_0.vlm"},
	{0x2FAF4FB906408D16ULL, "vxw_bonepitmining_a2_1u4_0_0.vlm"},
	{0x2FAF7A51E27D01EDULL, "gen00fl_carver6103403.cl"},
	{0x2FAF90E45F3F9595ULL, "vxlt_hangedman_1xo_0_0.vlm"},
	{0x2FAF963F078479FBULL, "606976221.ogg"},
	{0x2FAFA83D0FDA3E0FULL, "hlac_lt_0_lowtown_docks_-196611_0x.dds"},
	{0x2FAFAD45B67F64C5ULL, "zz_car_debug6122478.cl"},
	{0x2FAFB181F3703A43ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blackdust_d.dds"},
	{0x2FB04551B6ED9ECBULL, "6127186_m.fxe"},
	{0x2FB06F5741E1F49EULL, "6166544_m.fxe"},
	{0x2FB0F44C37742B19ULL, "vxlt_0_lowtown_docks_a3_n_d0q_0_0.vlm"},
	{0x2FB10358E4F5B2A7ULL, "zz_followers_banter6159432.cl"},
	{0x2FB125E30D7ABA0EULL, "hf_aveline.mor"},
	{0x2FB1307D08B8585BULL, "vxw_woundedcoast_1by_0_0.vlm"},
	{0x2FB153926ADC2923ULL, "vxw_bonepit_2ad_0_0.vlm"},
	{0x2FB15B5EE65F527BULL, "vxw_encounter2_8k_0_0.vlm"},
	{0x2FB34A5CF6B23290ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.dura.evt"},
	{0x2FB421D7DEA568ACULL, "lt_0_lowtown_a2_n_c_29_0.rml"},
	{0x2FB45B17EBB1D858ULL, "6198640_m.fxe"},
	{0x2FB477A8DC538D5DULL, "art\\characters\\heraldry\\hld_viscount_2.dds"},
	{0x2FB5559FC434FA58ULL, "vxlt_darktown_5ud_0_0.vlm"},
	{0x2FB56267137D7E41ULL, "mer210_seldrek_st.stg"},
	{0x2FB573CCB641EB71ULL, "static1_punch_s1y.psh.fxo"},
	{0x2FB5992CD5AED7C0ULL, "6055420_m.fxe"},
	{0x2FB63864026CC141ULL, "vxw_sundermount_31h_0_0.vlm"},
	{0x2FB689FF5A9D519BULL, "6114629_m.fxe"},
	{0x2FB6A246F9EE357DULL, "vxlt_0_lowtown_a2_n_c_e31_0_0.vlm"},
	{0x2FB6D6F87361EE06ULL, "6071694_f.fxe"},
	{0x2FB6FC2B81CBE0A0ULL, "vxht_bartrand_n_5yw_0_0.vlm"},
	{0x2FB7184DD678EAF1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xb.mmh"},
	{0x2FB7244DD678FE92ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xb.mao"},
	{0x2FB7324DD67916AFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xb.msh"},
	{0x2FB7C506DBB02F2FULL, "hlac_ht_mansion2_818_25x.dds"},
	{0x2FB86ED3349C5C79ULL, "6145203_m.fxe"},
	{0x2FB8DEA1CF952A69ULL, "231674868.ogg"},
	{0x2FB8F7EC39C55807ULL, "6095733_m.fxe"},
	{0x2FB9A05D672C024CULL, "hlac_lt_hangedman_-1_27z.dds"},
	{0x2FB9B67066BDCA90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.ani"},
	{0x2FB9BD5C87A8F927ULL, "vxlt_darktown_4ly_0_0.vlm"},
	{0x2FB9D46B5029408DULL, "vxlt_gallowscourtyard_7eu_0_0.vlm"},
	{0x2FBA262E7014F999ULL, "6151224_m.fxe"},
	{0x2FBA32DC58D215D0ULL, "dae300_alistair_king6133574.cl"},
	{0x2FBA4785DF12B315ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut3flat.msh"},
	{0x2FBA4985DF12B68BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut3flat.mmh"},
	{0x2FBA5585DF12CB28ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut3flat.mao"},
	{0x2FBA8EEF9CD45283ULL, "314258202.ogg"},
	{0x2FBAB630F084288FULL, "vxw_encounter3_8h_0_0.vlm"},
	{0x2FBAD79BF6462729ULL, "vxht_mansion2_1xm_0_0.vlm"},
	{0x2FBB5E2E84375A23ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.dura.ani"},
	{0x2FBBD9DCF645DC4EULL, "mrl310_merrill6120114.cl"},
	{0x2FBBDA3F0E9734EFULL, "vxdr_primevaltomb_3hl_0_0.vlm"},
	{0x2FBC045EA86C2733ULL, "vxht_0_hightown_n_bx3_0_0.vlm"},
	{0x2FBC29F802D1119EULL, "6156161_m.fxe"},
	{0x2FBC4B76BB7A3FACULL, "vxlt_hangedman_n_21i_0_0.vlm"},
	{0x2FBC4DC7BBE736FFULL, "hlac_ca_caverns1_mines_24424_9z.dds"},
	{0x2FBCCABEDDFA6478ULL, "6094961_f.fxe"},
	{0x2FBCD9B838AEB13BULL, "vxht_bartrand_n_666_0_0.vlm"},
	{0x2FBCF6E49015AF57ULL, "lgt000ip_ultra_treatise.utp"},
	{0x2FBD1A7902DA7ECDULL, "vxca_caverns2prc_acz_0_0.vlm"},
	{0x2FBD4DD1AA3F6C3AULL, "hlac_w_sundermount_327675_0y.dds"},
	{0x2FBD83491CD115FEULL, "vxht_brothel_n_6tq_0_0.vlm"},
	{0x2FBDAD6D78C0DD0CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnr90runf.evt"},
	{0x2FBE8436B208ED39ULL, "6037039_f.fxe"},
	{0x2FBEA8E8A7D5E9E6ULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_f_c.impa.evt"},
	{0x2FBED13962547929ULL, "6042708_m.fxe"},
	{0x2FBEF212E582F41AULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluelyrium.mmh"},
	{0x2FBEF612E582FB51ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluelyrium.mao"},
	{0x2FBF0412E5831360ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluelyrium.msh"},
	{0x2FBF106345472161ULL, "vxht_chantry_36q_0_0.vlm"},
	{0x2FBF826D4FCBA7A6ULL, "6165642_m.fxe"},
	{0x2FBFA259FAAE70DBULL, "vxw_woundedcoast_1mx_0_0.vlm"},
	{0x2FBFC998F1696F91ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_slider.ani"},
	{0x2FC02094727ED641ULL, "6179254_f.fxe"},
	{0x2FC0209802999F7BULL, "vxw_blightlands_oq_0_0.vlm"},
	{0x2FC04369C11ABA4CULL, "follower_barks6153258.cl"},
	{0x2FC05D3E9462B116ULL, "608541390.ogg"},
	{0x2FC0C186C981B815ULL, "vxht_mansion_n_5zl_0_0.vlm"},
	{0x2FC10703E607313AULL, "hlac_w_bonepitmining_a2_458757_0x.dds"},
	{0x2FC11FAFF53F14F1ULL, "vxlt_0_lowtown_docks_n_cte_0_0.vlm"},
	{0x2FC15BA54B30099FULL, "hlac_lt_0_lowtown_a2_65535_5y.dds"},
	{0x2FC170371C2EB871ULL, "w_bonepitmining_a2_94_0.msh"},
	{0x2FC17E371C2ECF8CULL, "w_bonepitmining_a2_94_0.mao"},
	{0x2FC18A371C2EE42FULL, "w_bonepitmining_a2_94_0.mmh"},
	{0x2FC301BC97A731B6ULL, "832262641.ogg"},
	{0x2FC36BE7D36462AAULL, "w_blightlands_109_0.mao"},
	{0x2FC36FE7D36469F9ULL, "w_blightlands_109_0.mmh"},
	{0x2FC379E7D3647AE7ULL, "w_blightlands_109_0.msh"},
	{0x2FC4389135BEA97DULL, "hlac_dr_primevaltomb_659_14x.dds"},
	{0x2FC43A75D655F485ULL, "art\\vfx\\materials\\vfx_wispy_cs.mao"},
	{0x2FC455715BD0490FULL, "hlac_ht_mansion_n_3372_0z.dds"},
	{0x2FC458FE14D6949DULL, "one110cr_meeran.utc"},
	{0x2FC4B09C778737B0ULL, "character_character_p61.psh.fxo"},
	{0x2FC4B2E4F39CFB1AULL, "one100dg_kelder6055420.cl"},
	{0x2FC4BE01B6A74C69ULL, "447858735.ogg"},
	{0x2FC51BE855A94932ULL, "hlac_ca_caverns1_mines_24424_11x.dds"},
	{0x2FC5348689AC0F7AULL, "845686995.ogg"},
	{0x2FC58579EB1879B5ULL, "6136353_m.fxe"},
	{0x2FC5888539A4AEA2ULL, "621778359.ogg"},
	{0x2FC5BFE8A7DC2F31ULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_f_c.impa.gad"},
	{0x2FC63D6EE0DFE678ULL, "6139995_m.fxe"},
	{0x2FC6CF4D6AFED602ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.gad"},
	{0x2FC6D498F16FA133ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_slider.gad"},
	{0x2FC71544025ECE7CULL, "362724224.ogg"},
	{0x2FC71856587A0765ULL, "vxlt_hangedman_27y_0_0.vlm"},
	{0x2FC78990243EDD79ULL, "llac_w_sundermount_-65541_0.dds"},
	{0x2FC7CFA1650D0D91ULL, "vxlt_shop_dy_0_0.vlm"},
	{0x2FC80A9AA3881885ULL, "622272523.ogg"},
	{0x2FC82B2724A1B71FULL, "dae000_bartrand6133012.cl"},
	{0x2FC82F39799F11A7ULL, "vxlt_darktown_4kx_0_0.vlm"},
	{0x2FC89919069BCD48ULL, "vxht_chantry_36t_0_0.vlm"},
	{0x2FC8DA0B42A5CB22ULL, "vxw_sundermount_2sx_0_0.vlm"},
	{0x2FC8F0928A3030C8ULL, "llac_lt_0_lowtown_docks_n_-131075_3.dds"},
	{0x2FC908AEC7C24A07ULL, "vxw_encounter3_ec_0_0.vlm"},
	{0x2FC90C5BB8F5B998ULL, "696697838.ogg"},
	{0x2FC9280DE27A410FULL, "vxw_encounter3_5k_0_0.vlm"},
	{0x2FC9762E8443C01AULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.dura.evt"},
	{0x2FC9FAE5DDD94407ULL, "follower_banter6163747.cl"},
	{0x2FCA763B9A623FB3ULL, "vxht_0_hightown_n_bc2_0_0.vlm"},
	{0x2FCA7C70E1FCAEF0ULL, "gen_im_trs_cot_junk1_04.uti"},
	{0x2FCADBC56614C42FULL, "vxlt_house_f1_0_0.vlm"},
	{0x2FCAF9D456286701ULL, "follower_banter6169813.cl"},
	{0x2FCB196A5D92F798ULL, "vxdr_primevalentrance_2oi_0_2.vlm"},
	{0x2FCB5043B5897747ULL, "385417074.ogg"},
	{0x2FCBB9A1A6278988ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_bld_flma_0d.dds"},
	{0x2FCC0CA30D5ABAD7ULL, "6155910_m.fxe"},
	{0x2FCC6AF597AE0DEEULL, "vxlt_0_lowtown_a2_deh_0_0.vlm"},
	{0x2FCCAF2957FCE4D4ULL, "vxht_0_hightown_bxg_0_0.vlm"},
	{0x2FCCC0E8A7E24FDFULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_f_c.impa.ani"},
	{0x2FCD138FCB3375EAULL, "6175801_m.fxe"},
	{0x2FCD79A2CD73D7B1ULL, "310662854.ogg"},
	{0x2FCD9DA9D3CFCE50ULL, "vxht_chantry_n_262_0_0.vlm"},
	{0x2FCDA76764A46D57ULL, "art\\vfx\\combat\\vfx_m_bld_wound_act_c.anb"},
	{0x2FCDF24D6B0530F9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.evt"},
	{0x2FCE045302930B32ULL, "art\\levels\\races\\design\\prp_chestb_0.pwk"},
	{0x2FCECCDB02CE7170ULL, "vxlt_0_lowtown_f9h_0_0.vlm"},
	{0x2FCF918CE7352D71ULL, "vxlt_0_lowtown_docks_n_cod_0_0.vlm"},
	{0x2FD01CC1FBD26E01ULL, "6163346_m.fxe"},
	{0x2FD0752E8449DD0DULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.dura.gad"},
	{0x2FD09C2E5A24950EULL, "dae000ip_crowd_mf4_talk17.utp"},
	{0x2FD0A7E5C4DB2E35ULL, "hlac_ht_0_hightown_a3_65534_1y.dds"},
	{0x2FD101A0A3705E5FULL, "load_movieoverlay.gfx"},
	{0x2FD11E9F0AAE57F8ULL, "6171869_m.fxe"},
	{0x2FD1243530567F1CULL, "6046947_m.fxe"},
	{0x2FD1B2A2D902BBDAULL, "6075356_m.fxe"},
	{0x2FD1F13F50040371ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_behind_both_ways.ani"},
	{0x2FD27561BFC2893BULL, "vxht_mansion2b_2lp_0_0.vlm"},
	{0x2FD2C229CF94FDDFULL, "lgt210pt_frame.ncs"},
	{0x2FD30890AB7750CCULL, "dae000_pc_home_ambients6208637.cl"},
	{0x2FD37CC1BD74F46FULL, "l_isa211ar_foundry.bnk"},
	{0x2FD389D37DE8616EULL, "6114831_m.fxe"},
	{0x2FD3BDA75151AE6FULL, "750796126.ogg"},
	{0x2FD3F756DD0E9235ULL, "6131253_m.fxe"},
	{0x2FD3FFC41C70BA8DULL, "vxlt_0_lowtown_docks_bcn_0_0.vlm"},
	{0x2FD483D2B3061447ULL, "w_bonepitdestroyed_a3_20_0.rml"},
	{0x2FD503A1D97098B0ULL, "vxlt_gallowstemplar_4ws_0_0.vlm"},
	{0x2FD55DAB377A2C54ULL, "vxlt_gallowstemplar_5qh_0_0.vlm"},
	{0x2FD5767670280C53ULL, "266260261.ogg"},
	{0x2FD5ACB63BC4F738ULL, "6177724_m.fxe"},
	{0x2FD5D19447D11EA3ULL, "6059385_f.fxe"},
	{0x2FD604D32881BA86ULL, "6133115_m.fxe"},
	{0x2FD619D61C7C4EEDULL, "m_mcr310ar_lowtown_nt.wbd"},
	{0x2FD65A62C5E17C4CULL, "dae000ip_crowd_mf1_talk9.utp"},
	{0x2FD724BC24617760ULL, "6147697_m.fxe"},
	{0x2FD728079EBA8AB8ULL, "vxlt_0_lowtown_a3_n_c_h2c_0_0.vlm"},
	{0x2FD74B41B16D17CEULL, "286396655.ogg"},
	{0x2FD74BE8349BE3A2ULL, "vxw_sundermount_2ty_0_0.vlm"},
	{0x2FD7581EB363A20BULL, "gen00fl_anders6125881.cl"},
	{0x2FD775FBDD880429ULL, "lt_0_lowtown_docks_19_0_ll.mmh"},
	{0x2FD7B983469F892DULL, "402693074.ogg"},
	{0x2FD7BB55DA325829ULL, "vxht_mansion2_1fs_0_0.vlm"},
	{0x2FD7C1B71E6BB82CULL, "vxht_bartrand_n_4fw_0_0.vlm"},
	{0x2FD7E525C3842A9EULL, "hlac_lt_hangedman_-1_6y.dds"},
	{0x2FD80E5F5B898EB3ULL, "6199024_m.fxe"},
	{0x2FD81C7E6D9FBCCAULL, "vxlt_0_lowtown_docks_a3_n_cnk_0_0.vlm"},
	{0x2FD82A8A139F93C4ULL, "hlac_lt_0_lowtown_docks_a3_-131074_8z.dds"},
	{0x2FD8FC3F500A3413ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_behind_both_ways.gad"},
	{0x2FD910D5AD89D8B0ULL, "vxht_0_hightown_a3_avo_0_0.vlm"},
	{0x2FD927828DFB01F8ULL, "hlac_ht_mansion2_818_51x.dds"},
	{0x2FD930FECA1FAC31ULL, "mer221_sabin6047515.cl"},
	{0x2FD9557880629B87ULL, "vxht_keep_6aw_0_0.vlm"},
	{0x2FD95BDEEA374258ULL, "qun120cr_javaris.stg"},
	{0x2FD9692E7B1813C5ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\blendtree\\c_dsogre_cbtmvt.anb"},
	{0x2FD9D4238745BB51ULL, "qun221_viscount6092397.cl"},
	{0x2FDA01CA21673299ULL, "art\\characters\\playercharacter\\humanmale\\glv_ric.phy"},
	{0x2FDA55257C16D586ULL, "290286300.ogg"},
	{0x2FDA670300BD46FEULL, "vxht_brothel_n_6zs_0_0.vlm"},
	{0x2FDAE67B3521DF15ULL, "vxlt_hangedman_1il_0_0.vlm"},
	{0x2FDB037A00472054ULL, "6166616_m.fxe"},
	{0x2FDB5EEC82F79E91ULL, "vxlt_gallowscourtyard_73g_0_0.vlm"},
	{0x2FDBFA4D6B117BE8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.ani"},
	{0x2FDC8613DE9AA1DBULL, "vxw_woundedcoast_1sz_0_0.vlm"},
	{0x2FDD15DC037CC85EULL, "6071024_m.fxe"},
	{0x2FDD413822A257A4ULL, "fen210_orana6109705.cl"},
	{0x2FDD70E6E58B4615ULL, "w_woundedcoast_56_0.anb"},
	{0x2FDD7C4D52A6B88CULL, "hlac_lt_gallowscourtyard_n_-65537_2x.dds"},
	{0x2FDDEA37AC5C0AE5ULL, "881055965.ogg"},
	{0x2FDEB76AC9156C38ULL, "6067119_m.fxe"},
	{0x2FDF0934721B1FF9ULL, "612323167.ogg"},
	{0x2FDF4F701F5D488EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_tw.ani"},
	{0x2FDF69A7FD85D539ULL, "llac_lt_0_lowtown_docks_n_-131074_7.dds"},
	{0x2FDF9612224C979CULL, "6041847_m.fxe"},
	{0x2FDFF93F50104E78ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_look_behind_both_ways.evt"},
	{0x2FE0375ACB5C0510ULL, "6044664_m.fxe"},
	{0x2FE05FDEEA3D67DAULL, "qun120cr_javaris.utc"},
	{0x2FE097E429D3B652ULL, "vxlt_0_lowtown_docks_a3_n_cyw_0_1.vlm"},
	{0x2FE0C56E5A60E9F7ULL, "dae000_bodhan6129114.cl"},
	{0x2FE0FD66346EAD2DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basicxend.phy"},
	{0x2FE0FDDA213C8A6FULL, "6064715_m.fxe"},
	{0x2FE10E7A5C258475ULL, "qcr000_varric6065815.cl"},
	{0x2FE1447E94E8BEDEULL, "qcr000_cut2d_thechampion6025142.cl"},
	{0x2FE1CB244113C4E5ULL, "6019164_m.fxe"},
	{0x2FE208C645EC2632ULL, "vxdr_0_deeproads_eu2_0_0.vlm"},
	{0x2FE27DF734EAE46DULL, "137693721.ogg"},
	{0x2FE28CBA5A999A8FULL, "dre000_mother6037110.cl"},
	{0x2FE2EF67B766EBC7ULL, "dae200_cullen6140293.cl"},
	{0x2FE3EC1A8A792398ULL, "vxw_woundedcoast_t9_0_0.vlm"},
	{0x2FE41112E5A24525ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2bluelyrium.phy"},
	{0x2FE411A518788C4DULL, "hlac_w_woundedcoast_131078_0z.dds"},
	{0x2FE413F14A408A51ULL, "6060186_m.fxe"},
	{0x2FE4784E5DEE7751ULL, "vxlt_gallowstemplar_3sy_0_0.vlm"},
	{0x2FE492016A5AFBEFULL, "vxht_brothel_n_4fh_0_0.vlm"},
	{0x2FE50C4D5769F516ULL, "vxht_playershouse_n_1pt_0_0.vlm"},
	{0x2FE5BDC52691FC22ULL, "vxw_sundermount_2yz_0_0.vlm"},
	{0x2FE62D32CC2870A9ULL, "vxht_mansion2_1ir_0_0.vlm"},
	{0x2FE633941061D0ACULL, "vxht_bartrand_n_4kt_0_0.vlm"},
	{0x2FE637F5C859E467ULL, "ht_0_hightown_43_0_ll.mmh"},
	{0x2FE69CA68A93837BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.resetup.evt"},
	{0x2FE7C21055BA2379ULL, "zz_isa_debug6133990.cl"},
	{0x2FE8B6502CDFF07AULL, "hlac_lt_gallowstemplar_n_2383_2y.dds"},
	{0x2FE8EF8AC5E47ED5ULL, "vxht_0_hightown_n_c71_0_0.vlm"},
	{0x2FE8F756EF34B7BBULL, "6034743_m.fxe"},
	{0x2FE8FE7912AD633CULL, "6115944_m.fxe"},
	{0x2FE9670C0A365114ULL, "vxw_bonepit_1x2_0_1.vlm"},
	{0x2FE9D0C974EDB711ULL, "vxlt_gallowscourtyard_74f_0_0.vlm"},
	{0x2FEA0E9431A70973ULL, "vxlt_0_lowtown_a3_n_c_e2m_0_0.vlm"},
	{0x2FEAA0E72D6A86A4ULL, "vxw_bonepitmining_a2_1rw_0_0.vlm"},
	{0x2FEAF62B55598736ULL, "vxlt_hangedman_26q_0_0.vlm"},
	{0x2FEAFEDA7078902AULL, "222485524.ogg"},
	{0x2FEB3914A4FACE48ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.anb"},
	{0x2FEB40936BC3CE04ULL, "668709969.ogg"},
	{0x2FEC0288F095A084ULL, "vxw_encounter1_dq_0_0.vlm"},
	{0x2FEC03491475CE59ULL, "cod_bks_chantry_hist_2.utp"},
	{0x2FEC33636C8CC79FULL, "6140348_f.fxe"},
	{0x2FEC47701F67C5DFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_tw.evt"},
	{0x2FEC533FAF80D3B3ULL, "vxht_0_hightown_n_9aw_0_0.vlm"},
	{0x2FEC716DC62E891BULL, "vxlt_0_lowtown_a2_cw0_0_0.vlm"},
	{0x2FEC83D2B31AA664ULL, "w_bonepitdestroyed_a3_20_0.mmh"},
	{0x2FEC87D2B31AACB7ULL, "w_bonepitdestroyed_a3_20_0.mao"},
	{0x2FEC95AE698BDA75ULL, "seb201_lady_harimann6197706.cl"},
	{0x2FEC99D2B31ACB06ULL, "w_bonepitdestroyed_a3_20_0.msh"},
	{0x2FECCDE55BA566FBULL, "lgt260_eau_tmp_ending6171809.cl"},
	{0x2FED6C73C601CD48ULL, "hlac_lt_undercityserialkiller_-1_11z.dds"},
	{0x2FED88397F758B6DULL, "110678419.ogg"},
	{0x2FED95A68A9995BCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.resetup.gad"},
	{0x2FEDDF6E145AE9ECULL, "6096453_f.fxe"},
	{0x2FEE9330F90E5991ULL, "228170625.ogg"},
	{0x2FEE99745C4C4787ULL, "lt_gallowstemplar_2386.rml"},
	{0x2FEF0231001417F3ULL, "vxw_bonepit_nu_0_0.vlm"},
	{0x2FEF095BE47769CAULL, "one220_yevhen6121660.cl"},
	{0x2FEF591BA3AEBA3BULL, "vxht_mansion2b_2br_0_0.vlm"},
	{0x2FEFC49B05EE7F92ULL, "vxht_0_hightown_a2_c6b_0_0.vlm"},
	{0x2FF004EC90771668ULL, "vxht_mansion_n_4fl_0_0.vlm"},
	{0x2FF008603A833F85ULL, "638187699.ogg"},
	{0x2FF01E6865AC46F5ULL, "6093665_m.fxe"},
	{0x2FF029FD1B09183EULL, "dae000_arishok6095679.cl"},
	{0x2FF0B5AD1C724BB6ULL, "pro000_opening.cnv"},
	{0x2FF11CD2E84799D0ULL, "qun211it_barrel_latch.uti"},
	{0x2FF181658DE7D3B9ULL, "vxlt_0_lowtown_docks_n_c3h_0_0.vlm"},
	{0x2FF25685C431A146ULL, "831284347.ogg"},
	{0x2FF2EEBB23B0BBFDULL, "6066715_m.fxe"},
	{0x2FF328701F6DAFC8ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_tw.gad"},
	{0x2FF329F0F950ACDDULL, "vxht_0_hightown_a6o_0_0.vlm"},
	{0x2FF395FFD1A28972ULL, "635686259.ogg"},
	{0x2FF3B7AD1C748496ULL, "pro000_opening.bnk"},
	{0x2FF3DD5CC4D85D35ULL, "ser200_aveline6201513.cl"},
	{0x2FF47F4D820A60B2ULL, "hlac_lt_darktown_-1_49y.dds"},
	{0x2FF494A68A9FB2BAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.resetup.ani"},
	{0x2FF50038518BEDCAULL, "vxlt_0_lowtown_docks_a3_n_cte_0_0.vlm"},
	{0x2FF5A46764C5BA58ULL, "art\\vfx\\combat\\vfx_m_bld_wound_act_c.mmh"},
	{0x2FF68F056B35AF5EULL, "136420110.ogg"},
	{0x2FF77E8ED3AF0E3EULL, "163597610.ogg"},
	{0x2FF7AB78981DCD8DULL, "vxht_keep_n_4nq_0_0.vlm"},
	{0x2FF81A6B1A1BCA95ULL, "vxca_caverns2prc_7o5_0_1.vlm"},
	{0x2FF842A666E3CF91ULL, "vxlt_gallowscourtyard_79e_0_0.vlm"},
	{0x2FF842DE10B213F9ULL, "vxw_encounter3_64_0_0.vlm"},
	{0x2FF88EC21EC4D396ULL, "ca_caverns1_24425_ll.mmh"},
	{0x2FF899A27A4EB39DULL, "6101491_f.fxe"},
	{0x2FF8D19D5420A5BFULL, "615932203.ogg"},
	{0x2FF8EDD43E46B877ULL, "6020496_m.fxe"},
	{0x2FF91F55D96FB9E1ULL, "6047196_m.fxe"},
	{0x2FF93C928E6B2FFFULL, "597429353.ogg"},
	{0x2FF96DE8946E37CDULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_candle_0.pwk"},
	{0x2FF970F50ACB8696ULL, "534446668.ogg"},
	{0x2FFA0846E30A6228ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pulpit.mmh"},
	{0x2FFA1446E30A76CBULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pulpit.mao"},
	{0x2FFA1BFEB627932BULL, "zz_cave_accolades_v1.cut"},
	{0x2FFA2646E30A9552ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pulpit.msh"},
	{0x2FFA7465E28BB904ULL, "vxw_encounter1_iv_0_0.vlm"},
	{0x2FFAE8DC5BF7E987ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_loop.gad"},
	{0x2FFB25732D9D7210ULL, "6034685_m.fxe"},
	{0x2FFB8D4E4BB165FAULL, "184937121.ogg"},
	{0x2FFBB08593631921ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_017a_0n.dds"},
	{0x2FFC15E7274ECA0EULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_dwfa_0n.dds"},
	{0x2FFC2163EC2DC21BULL, "vxlt_gallowscourtyard_6xa_0_0.vlm"},
	{0x2FFC4DAF11DA283CULL, "6147587_m.fxe"},
	{0x2FFD95A4E7D8DDCFULL, "aip_slaverwarrior_base.gda"},
	{0x2FFDBA392573070DULL, "vxw_bonepitdestroyed_a3_1zx_0_0.vlm"},
	{0x2FFE393184E15EF7ULL, "6115099_m.fxe"},
	{0x2FFE480506CA8F1BULL, "6202151_m.fxe"},
	{0x2FFE48DEEA563E40ULL, "qun120cr_javaris.ncs"},
	{0x2FFE65DC23F0D215ULL, "vxw_encounter2_camp_5f_0_0.vlm"},
	{0x2FFEC7FE593800D6ULL, "vxlt_undercityencounter_1c2_0_0.vlm"},
	{0x2FFF14B9054056C2ULL, "seb221_flora_harimann6177691.cl"},
	{0x2FFF5C9A58467A26ULL, "vxlt_blackemporium_if_0_0.vlm"},
	{0x2FFFA7FAD4E71B4DULL, "6135610_f.fxe"},
	{0x30002DF48E45E5F5ULL, "1015643467.ogg"},
	{0x30010F31693455D2ULL, "45299496.ogg"},
	{0x300144C99E54BE04ULL, "6077012_f.fxe"},
	{0x300150891E8C3F8FULL, "vxlt_0_lowtown_a2_exk_0_0.vlm"},
	{0x300197688F642400ULL, "6135937_m.fxe"},
	{0x3001E0BD90BA5FE1ULL, "vxlt_0_lowtown_docks_a3_n_9zr_0_0.vlm"},
	{0x3002807B480BE9F0ULL, "and230_mage_girl6135803.cl"},
	{0x300281A4270E53B1ULL, "vxca_caverns1_ijo_0_0.vlm"},
	{0x3002D159A8B202BBULL, "hlac_w_woundedcoast_196613_0x.dds"},
	{0x3002F58BA911D0B7ULL, "vxht_keep_n_67j_0_0.vlm"},
	{0x3003072AC185F19EULL, "hlac_ht_mansion2_818_50y.dds"},
	{0x300403E31945D000ULL, "6129671_m.fxe"},
	{0x3004059286F1062EULL, "vxdr_primevalentrance_1ja_0_0.vlm"},
	{0x30041466C861A44FULL, "vxht_0_hightown_c0q_0_0.vlm"},
	{0x30044BE54B12C377ULL, "vxht_mansion2b_2ld_0_0.vlm"},
	{0x300451B163633CF0ULL, "135733882.ogg"},
	{0x30048CD3C7920830ULL, "llac_ht_0_hightown_a2_65536_4.dds"},
	{0x300611DD3DFE8A08ULL, "579963852.ogg"},
	{0x30064AEEBA44A6F0ULL, "529529691.ogg"},
	{0x30066DE6E5AE461DULL, "w_woundedcoast_56_0.mao"},
	{0x300671E6E5AE4D56ULL, "w_woundedcoast_56_0.mmh"},
	{0x30067BE6E5AE5E54ULL, "w_woundedcoast_56_0.msh"},
	{0x3006A37630DEBC33ULL, "43257512.ogg"},
	{0x3006B7E20E3D97ACULL, "hlac_w_bonepit_65542_0z.dds"},
	{0x3006FEF77D861BB4ULL, "llac_w_sundermount_196609_0.dds"},
	{0x30074D7479746A37ULL, "hlac_ht_mansion2_818_31y.dds"},
	{0x3007A717C9271D3FULL, "9198952.ogg"},
	{0x30084839CB569662ULL, "vxht_mansion2b_2k4_0_0.vlm"},
	{0x3008C281369CA5DFULL, "vxht_bartrand_n_5ry_0_0.vlm"},
	{0x30091448BB4A3A81ULL, "z_cloud03.2.dds"},
	{0x3009389B55B50F97ULL, "art\\vfx\\combat\\vfx_str_spit2_aoe_p.anb"},
	{0x300999B50356E4B7ULL, "754967418.ogg"},
	{0x300A5E1A073FFAC1ULL, "vxca_caverns1_dp8_0_0.vlm"},
	{0x300A890174970FFFULL, "llac_dr_0_deeproads_17038_6.dds"},
	{0x300AD89F61058437ULL, "rdr110dg_guard_sleeping6016740.cl"},
	{0x300ADE90DC26B72FULL, "58603445.ogg"},
	{0x300B0F329F4E96A1ULL, "character_character_p198.psh.fxo.dx11.assembly"},
	{0x300B1D8C393CAD12ULL, "ht_0_hightown_n_61_0.rml"},
	{0x300B41B5C043AACDULL, "6187924_m.fxe"},
	{0x300BA9BA967F6952ULL, "6059915_f.fxe"},
	{0x300BBE62C32DF785ULL, "6023873_m.fxe"},
	{0x300BC6650F969A3EULL, "553547522.ogg"},
	{0x300BD43B8B3DDB4BULL, "vxca_caverns1_gk4_0_0.vlm"},
	{0x300C005718E1FFC3ULL, "fen210_tevinter_ambusher26108648.cl"},
	{0x300C35FA9932A460ULL, "hlac_w_sundermount_-327686_0x.dds"},
	{0x300C53DEEA628E2BULL, "qun120cr_javaris.bnk"},
	{0x300C7F038FFD495EULL, "740112650.ogg"},
	{0x300CA54FF02636C6ULL, "gen00fl_varric6077822.cl"},
	{0x300CA854E9DAB092ULL, "vxht_0_hightown_a2_c8l_0_0.vlm"},
	{0x300D28AD249AC4C1ULL, "follower_banter6151366.cl"},
	{0x300D2B3C5A685D5EULL, "hlac_ca_cavern1_haunted_24424_26x.dds"},
	{0x300D2DCAF77DB8B1ULL, "6095559_m.fxe"},
	{0x300DA8D417B550FEULL, "555311363.ogg"},
	{0x300DD41751F24A9FULL, "504705390.ogg"},
	{0x300DE73F6D6A883AULL, "587892911.ogg"},
	{0x300E293EE6C5FDBCULL, "93925303.ogg"},
	{0x300E31310B296EB6ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.dura.evt"},
	{0x300EA27C70388F54ULL, "6023887_m.fxe"},
	{0x300EDB9389D73399ULL, "gen00fl_anders6125335.cl"},
	{0x300EE4BDA8C8A528ULL, "6136316_m.fxe"},
	{0x300F195DA4493234ULL, "mag131_keran6044527.cl"},
	{0x300F55DEEA64C70BULL, "qun120cr_javaris.cnv"},
	{0x300FA89CDD89B2ABULL, "vxw_encounter2_29_0_0.vlm"},
	{0x300FBAC20B282EB7ULL, "llac_lt_0_lowtown_a2_n_c_0_3.dds"},
	{0x300FD5DC5C0A25CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_loop.ani"},
	{0x30101939A748422AULL, "vxlt_0_lowtown_docks_c01_0_0.vlm"},
	{0x301028EFA05AD93AULL, "6090195_m.fxe"},
	{0x30102A914FBFFBD1ULL, "pro000_colonel_ewald6040627.cl"},
	{0x3010F38119046C31ULL, "vxca_caverns1_imn_0_0.vlm"},
	{0x30112BEC1A472FCEULL, "446816807.ogg"},
	{0x3011572F4CB0E4CAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\attackfromkncokdown.gad"},
	{0x30116352D5DEFF20ULL, "dre000_sibling_has_blight6064109.cl"},
	{0x301167689B07E937ULL, "vxht_keep_n_62e_0_0.vlm"},
	{0x3011D9EFEC8411F1ULL, "vxlt_0_lowtown_a2_n_gmw_0_0.vlm"},
	{0x3011E3F235781ECAULL, "vxlt_0_lowtown_docks_a3_n_czg_0_0.vlm"},
	{0x3011EB9E8C9DD624ULL, "vxht_playershouse_n_1p6_0_0.vlm"},
	{0x30122561656956F5ULL, "hlac_ca_caverns1_mines_24424_2y.dds"},
	{0x30128643BA57BCCFULL, "vxht_0_hightown_c7p_0_0.vlm"},
	{0x3012AF42198AA09CULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev7.ani"},
	{0x3012D71D55FAE076ULL, "vxlt_gallowstemplar_n_2p1_0_0.vlm"},
	{0x3012F8D5776CAF17ULL, "vxlt_undercityencounter_115_0_0.vlm"},
	{0x30130B9F3F412E30ULL, "6126937_m.fxe"},
	{0x30139151D40D52BEULL, "161353368.ogg"},
	{0x3013BEFB3BAE44B8ULL, "vxw_encounter2_62_0_0.vlm"},
	{0x3013CF46560E093FULL, "6175134_m.fxe"},
	{0x30143814A51DD1E3ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.mmh"},
	{0x3014B2ACA25100ABULL, "6171270_f.fxe"},
	{0x3014FC75A2D2A68FULL, "dae000_flemeth6013326.cl"},
	{0x3015134FFD6CA0B3ULL, "635613357.ogg"},
	{0x301528310B2F7E41ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.dura.gad"},
	{0x30156632F5CE9614ULL, "57972035.ogg"},
	{0x3015724AD853306FULL, "vxdr_primevaltomb_6y2_0_0.vlm"},
	{0x301578FB2E986D9AULL, "6069677_f.fxe"},
	{0x3015E15CF1DECE2DULL, "dae000_tomwise6167293.cl"},
	{0x301623D61CB24317ULL, "m_mcr310ar_lowtown_nt.bnk"},
	{0x30162F068B73508CULL, "973959671.ogg"},
	{0x3016C713447FA9FBULL, "6099830_m.fxe"},
	{0x30171790BDCC1B5CULL, "sb_day_base.mao"},
	{0x3017345E2892BE5FULL, "vxht_bartrand_n_5ux_0_0.vlm"},
	{0x301756A01E35116FULL, "298765064.ogg"},
	{0x3017AEB1AA29D88BULL, "dae000_arishok6078426.cl"},
	{0x3017DD397B600A25ULL, "gen_im_trs_crc_lgt_glv_r1.uti"},
	{0x3017FD9001AC085CULL, "lgt360tr_orlanna.utt"},
	{0x30180F671DF21F78ULL, "596719384.ogg"},
	{0x30183A2F4CB6D261ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\attackfromkncokdown.evt"},
	{0x30187E1E6440FA51ULL, "vxht_0_hightown_aq2_0_0.vlm"},
	{0x3019B15C467661B2ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_03a_0s.dds"},
	{0x3019BC421990D5DEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev7.gad"},
	{0x3019D8F00FC852ECULL, "hlac_lt_undercityserialkiller_-1_48x.dds"},
	{0x3019E9F19586E6D3ULL, "vxca_caverns2_8mv_0_0.vlm"},
	{0x301A0AE58BEE3292ULL, "6151164_m.fxe"},
	{0x301A1B97887581D2ULL, "6056815_f.fxe"},
	{0x301A8C6D774FFD92ULL, "587790174.ogg"},
	{0x301ADA7D8B39081FULL, "dae000_gamlen6131092.cl"},
	{0x301AEE77CCC1F2D7ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_012a_0n.dds"},
	{0x301B01E4869BFBAEULL, "750065681.ogg"},
	{0x301B84A770078D01ULL, "6098937_f.fxe"},
	{0x301B8C7BD51834A7ULL, "493782156.ogg"},
	{0x301C29310B359EEFULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.dura.ani"},
	{0x301C83AEE9CB9130ULL, "hlac_ht_0_hightown_a3_2_0z.dds"},
	{0x301C9797E5957805ULL, "vxlt_warehouse_n_14d_0_0.vlm"},
	{0x301C9E4A0F84D6C3ULL, "hlac_w_bonepit_262145_0y.dds"},
	{0x301CF06403BAFFBFULL, "hurlock_lsword_shield.utc"},
	{0x301CFFB0D18D77B9ULL, "6016770_m.fxe"},
	{0x301D5499021B8333ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.cess.evt"},
	{0x301DD8C110BA7E68ULL, "gen00fl_merrill6129482.cl"},
	{0x301E5CA3DC30AE0DULL, "follower_banter6154000.cl"},
	{0x301F45C32981FBC2ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_2y.dds"},
	{0x301F463A6D2C0F81ULL, "hlac_lt_warehouse_65537_2x.dds"},
	{0x301F56F8B74E1FD1ULL, "6126459_m.fxe"},
	{0x301F878C3646B414ULL, "fen210_hadriana6110375.cl"},
	{0x301F95B6F91F6638ULL, "6112098_f.fxe"},
	{0x301FA23FAAB0F953ULL, "vxlt_gallowstemplar_2r9_0_0.vlm"},
	{0x302019A179DD83A9ULL, "6155291_m.fxe"},
	{0x30202E27BB3F628CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_diepose_aveline.ani"},
	{0x30205D7B7E93EEA4ULL, "vxht_playershouse_n_1w7_0_0.vlm"},
	{0x3020951798A6538FULL, "vxht_keep_n_6ls_0_0.vlm"},
	{0x3020C4F1DF36B025ULL, "hlac_dr_0_deeproads_5842_14y.dds"},
	{0x30213BCB5956D79FULL, "98927583.ogg"},
	{0x3021F97C5ACDF9F9ULL, "vxlt_0_lowtown_a2_ep5_0_0.vlm"},
	{0x302203B5E5353556ULL, "809446535.ogg"},
	{0x30222D625C56D6BFULL, "hlac_lt_darktown_-1_18z.dds"},
	{0x3022778C82B90850ULL, "vxht_mansion_n_3er_0_0.vlm"},
	{0x302310B4CB988905ULL, "vxht_0_hightown_a2_9az_0_0.vlm"},
	{0x30235BAA8469B851ULL, "923538838.ogg"},
	{0x30236AA756DB8552ULL, "vxca_caverns2_8ms_0_0.vlm"},
	{0x3023C60AB3F59E91ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\additivedamage.ani"},
	{0x302413AC54F2EC15ULL, "vxca_caverns1_j2d_0_0.vlm"},
	{0x30244D9902219674ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.cess.gad"},
	{0x30245FF07D18E376ULL, "265015500.ogg"},
	{0x3024F8FEDA6DCC34ULL, "385859457.ogg"},
	{0x3024FBB603FE0414ULL, "6163615_m.fxe"},
	{0x30252BF3AF42C762ULL, "vxht_mansion2b_2a6_0_0.vlm"},
	{0x302547FDBDC08DF8ULL, "ch1armtnt_d.psh.fxo"},
	{0x3025B8BBB9885DFEULL, "vxw_blightlands_eb_0_0.vlm"},
	{0x3025FF60CE1D1A85ULL, "209891884.ogg"},
	{0x3026322F4CC301B0ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\attackfromkncokdown.ani"},
	{0x302670E7D3B8E8E6ULL, "w_blightlands_109_0.phy"},
	{0x3026B13A950EC098ULL, "6127094_f.fxe"},
	{0x30273B27BB4597CEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_diepose_aveline.gad"},
	{0x30279CABBC907D07ULL, "6140224_m.fxe"},
	{0x3028467239D3829BULL, "813747371.ogg"},
	{0x302876DF49349E9AULL, "6206680_m.fxe"},
	{0x30288818D1B013CAULL, "vxlt_0_lowtown_docks_n_as1_0_0.vlm"},
	{0x30289E8FB6266627ULL, "vxlt_warehouse_n_8y_0_0.vlm"},
	{0x3028F1F5BBC102E6ULL, "6144760_m.fxe"},
	{0x30290FCFFB55508DULL, "vxw_bonepitdestroyed_a3_1eu_0_0.vlm"},
	{0x30292272B24F7499ULL, "341702923.ogg"},
	{0x30297CE829BCDEABULL, "follower_banter6141893.cl"},
	{0x3029D9AF731E6DA7ULL, "dre000_cut_arrivaldeeproads.bnk"},
	{0x3029E9FF327B5865ULL, "x360_controllerinteraction.gda"},
	{0x302A1CD7B4C1DA88ULL, "6163637_m.fxe"},
	{0x302A1D952F1A4A56ULL, "vxlt_undercityencounter_1r5_0_0.vlm"},
	{0x302A87AB72EE4EE4ULL, "hlac_ca_cavern1_haunted_24421_9x.dds"},
	{0x302AD10AB3FBD033ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\additivedamage.gad"},
	{0x302B0974D78B9085ULL, "vxlt_warehouse_n_19c_0_0.vlm"},
	{0x302B5C990227CE52ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.cess.ani"},
	{0x302B84C7FDA75C5CULL, "137595280.ogg"},
	{0x302BDE9113EFCD50ULL, "vxlt_hangedman_n_1ta_0_0.vlm"},
	{0x302C065CC9591BEDULL, "art\\vfx\\materials\\vfx_dot_add.mao"},
	{0x302C8475E2D35B2EULL, "750560465.ogg"},
	{0x302CA61FF46E890FULL, "vxlt_0_lowtown_a2_eit_0_0.vlm"},
	{0x302CFE2A51057423ULL, "pro000cr_carver_fake.utc"},
	{0x302D75D62B0F0F9BULL, "vxlt_0_lowtown_n_h1z_0_0.vlm"},
	{0x302DFBAF7322901FULL, "dre000_cut_arrivaldeeproads.cnv"},
	{0x302EC54D171B8366ULL, "6142847_m.fxe"},
	{0x302ED3CE5A8ACE86ULL, "258790784.ogg"},
	{0x302F06F48A9C6C0FULL, "vxht_keep_n_6gr_0_0.vlm"},
	{0x302F1C8F0DBFAC95ULL, "6027446_m.fxe"},
	{0x302F85DE7FBEE9E9ULL, "abi_m_arcane_shield.wbd"},
	{0x302F9B02BD5C24EEULL, "55090788.ogg"},
	{0x302F9FD2299FB22BULL, "llac_lt_0_lowtown_a2_n_c_-65539_0.dds"},
	{0x302FE429E3200C65ULL, "vxw_encounter2_camp_56_0_0.vlm"},
	{0x3030380EFC475831ULL, "w_bonepit_74_0.msh"},
	{0x3030460EFC47704CULL, "w_bonepit_74_0.mao"},
	{0x3030520EFC4784EFULL, "w_bonepit_74_0.mmh"},
	{0x30307220455F7F42ULL, "vxlt_gallowsprison_n_6b7_0_0.vlm"},
	{0x3030736C39BD591EULL, "mnd_cycle_f3.dds"},
	{0x3030892F61D70FA1ULL, "vxca_caverns1_mines_eh4_0_0.vlm"},
	{0x3031231296987C86ULL, "vxw_blightlands_fs_0_0.vlm"},
	{0x30312972736BBD85ULL, "6105441_m.fxe"},
	{0x30314C7D338C8072ULL, "one101dg_smuggler6044990.cl"},
	{0x3031CE0AB401E998ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\additivedamage.evt"},
	{0x30320B5CE6832CDEULL, "vxw_bonepitmining_a2_1na_0_0.vlm"},
	{0x3032359B55D80F98ULL, "art\\vfx\\combat\\vfx_str_spit2_aoe_p.mmh"},
	{0x3032DF89A67DAE70ULL, "art\\characters\\weapons\\quivers\\w_qui_dq01a.mao"},
	{0x3032E31DF40E93F5ULL, "146980917.ogg"},
	{0x3032E389A67DB483ULL, "art\\characters\\weapons\\quivers\\w_qui_dq01a.mmh"},
	{0x3032E4CB1559907FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\brother_pcopeningmage.gad"},
	{0x3032EF88AC8FB9F3ULL, "llac_lt_0_lowtown_docks_262142_0.dds"},
	{0x3032F189A67DCCDDULL, "art\\characters\\weapons\\quivers\\w_qui_dq01a.msh"},
	{0x3033FC7E15BC441CULL, "vxlt_0_lowtown_docks_a3_n_aoh_0_0.vlm"},
	{0x30342A98AB7E767EULL, "vxw_blightlands_nc_0_0.vlm"},
	{0x30346A4C1030B75CULL, "678824584.ogg"},
	{0x30348A90B4EC5606ULL, "567637248.ogg"},
	{0x3034D28D531B24BBULL, "art\\vfx\\cinematic\\vfx_orsino_swirly_c.mmh"},
	{0x3035EB411A4DDFDBULL, "vxht_bartrand_n_5rd_0_1.vlm"},
	{0x3036CDAB797317D3ULL, "vxca_caverns2_8wx_0_0.vlm"},
	{0x3036FFD11B5B0835ULL, "6064390_f.fxe"},
	{0x30371123250FB292ULL, "hlac_ht_chantry_n_2534_50y.dds"},
	{0x303781ACED4B690DULL, "vxw_bonepitdestroyed_a3_1nt_0_0.vlm"},
	{0x3037D44946C70DDCULL, "93986634.ogg"},
	{0x3037D6A217CCD3DEULL, "mrl351_pridedemon6165565.cl"},
	{0x3038B70D7DE896F3ULL, "conversation_categories.gda"},
	{0x30399800D49CF449ULL, "gen00fl_merrill6127511.cl"},
	{0x3039BC1C3CC60F61ULL, "209047384.ogg"},
	{0x3039FDF896974B40ULL, "vxdr_primevalentrance_23o_0_0.vlm"},
	{0x303AAE1F7D0682BEULL, "6208589_m.fxe"},
	{0x303AB8311946ECB3ULL, "6149871_f.fxe"},
	{0x303AD716CCFDE750ULL, "vxca_caverns2prc_520_0_0.vlm"},
	{0x303B89FB21C577DAULL, "dre000_varric6036992.cl"},
	{0x303BF76A26A79E9FULL, "vxlt_blackemporium_s1_0_0.vlm"},
	{0x303C2680EF2F96CAULL, "mag360_samson6042079.cl"},
	{0x303C3AB29B3A50D1ULL, "6128659_m.fxe"},
	{0x303C6F5E39AD4377ULL, "gen00fl_isabela6135709.cl"},
	{0x303C8DC5E07554DAULL, "ht_0_hightownexplosion_n_16_0_lc.mmh"},
	{0x303CBBCAB294C213ULL, "vxht_0_hightown_n_caq_0_0.vlm"},
	{0x303CD47FC23B2A0BULL, "6102815_m.fxe"},
	{0x303CDC69A57C26CAULL, "hlac_lt_0_lowtown_a2_n_c_65536_7x.dds"},
	{0x303D2B36E2DBE64EULL, "vxlt_0_lowtown_a2_ef8_0_0.vlm"},
	{0x303D2FB64CAFA190ULL, "llac_w_woundedcoast_-196608_0.dds"},
	{0x303D7136C1520E07ULL, "hlac_w_encounter2_327681_0z.dds"},
	{0x303F10855F243D91ULL, "lla_ht_0_hightown_n_131070_0.dds"},
	{0x303F44B82EC81561ULL, "6205596_m.fxe"},
	{0x303FC5542752B74BULL, "mag320_orsino6071073.cl"},
	{0x303FE5047FEA6B86ULL, "art\\vfx\\materials\\vfx_web_debris_blend.mao"},
	{0x303FF46EAF84BA05ULL, "vxht_0_hightown_a2_9wx_0_0.vlm"},
	{0x3040596ABB1BC7EEULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_bdy_hdra_0s.dds"},
	{0x30406A82AAA3446CULL, "450516098.ogg"},
	{0x3040ACDE487A581AULL, "hlac_lt_0_lowtown_docks_n_0_6y.dds"},
	{0x3040E2DFACA47088ULL, "6149036_f.fxe"},
	{0x3041441A400EA48DULL, "mrl310_party_comments6143057.cl"},
	{0x30419F3B1BDDC1CBULL, "6209859_m.fxe"},
	{0x3041F38D5325E720ULL, "art\\vfx\\cinematic\\vfx_orsino_swirly_c.anb"},
	{0x30429AFBE522D9A9ULL, "w_bonepitdestroyed_a3_75_0.mao"},
	{0x30429C759D748EFEULL, "vxw_blightlands_sd_0_0.vlm"},
	{0x3042A6FBE522EE42ULL, "w_bonepitdestroyed_a3_75_0.mmh"},
	{0x3042A8FBE522F1F8ULL, "w_bonepitdestroyed_a3_75_0.msh"},
	{0x3042B7C8AA96088CULL, "771415122.ogg"},
	{0x3043302CAB4D2401ULL, "hlac_lt_0_lowtown_65535_0z.dds"},
	{0x3043309707D14E14ULL, "llac_lt_gallowstemplar_2384_7.dds"},
	{0x30434BDE0B382EE3ULL, "vxca_caverns1_mines_iog_0_0.vlm"},
	{0x3043DA2A4913A9FCULL, "6202192_m.fxe"},
	{0x3043E6A2830363E2ULL, "341979278.ogg"},
	{0x304420512E3669F9ULL, "vrc221_bartrand6043027.cl"},
	{0x30443FBCEDE773D4ULL, "dae000_pc_home_ambients6208318.cl"},
	{0x304471C7CEC6E45CULL, "6005407_m.fxe"},
	{0x304549FC1FF6DA7AULL, "follower_banter6171541.cl"},
	{0x3045721A2150A284ULL, "vxw_bonepitdestroyed_a3_26f_0_0.vlm"},
	{0x304597D97AACBF2FULL, "6004115_m.fxe"},
	{0x30459BCF6C0C3262ULL, "gen00fl_anders6107688.cl"},
	{0x3045DD680545F11EULL, "mag130_danzig6016424.cl"},
	{0x3046C1CB1569FDC5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\brother_pcopeningmage.ani"},
	{0x3046F97EBC2046C5ULL, "wm_ico_sewerentrance_roll.dds"},
	{0x30470BA6104848C1ULL, "6163993_m.fxe"},
	{0x30475503D00BC963ULL, "979667397.ogg"},
	{0x30476AC691C5486EULL, "dae200_cullen6140328.cl"},
	{0x3047942E65407B29ULL, "lt_0_lowtown_docks_a3_128_0.mao"},
	{0x304795EB1214F4A6ULL, "vxlt_blackemporium_vk_0_0.vlm"},
	{0x3047A02E65408FC2ULL, "lt_0_lowtown_docks_a3_128_0.mmh"},
	{0x3047A22E65409378ULL, "lt_0_lowtown_docks_a3_128_0.msh"},
	{0x3047CEECB5A47B8EULL, "vxca_caverns1_mines_dq9_0_0.vlm"},
	{0x3047D87609388805ULL, "312132456.ogg"},
	{0x30482742F68E304BULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.dura.evt"},
	{0x3048A196D977DF35ULL, "hlac_ca_caverns1_mines_24424_27z.dds"},
	{0x3048DD78ECEF0990ULL, "6049717_m.fxe"},
	{0x304949B1F8B9206CULL, "vxdr_primevalentrance_2q4_0_1.vlm"},
	{0x30496A1E12AF051BULL, "vxdr_primevaltomb_4i0_0_0.vlm"},
	{0x304A5FF4184E62CAULL, "plt_prm_cod_hindsight_belt.plo"},
	{0x304A66DCBA16A161ULL, "dae000_meredith6085272.cl"},
	{0x304A8AD623EF0505ULL, "6108954_m.fxe"},
	{0x304AF315ADCE1AFDULL, "6054371_f.fxe"},
	{0x304B385BD5A825B9ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masf.msh"},
	{0x304B425BD5A836A7ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masf.mmh"},
	{0x304B465BD5A83DF4ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masf.mao"},
	{0x304B6C22E8622AFFULL, "vxht_keep_n_6f1_0_0.vlm"},
	{0x304B7F9447E82C2CULL, "6172648_m.fxe"},
	{0x304B8BB9FE3B3A1CULL, "256928594.ogg"},
	{0x304B9D13D4D1FECEULL, "vxlt_0_lowtown_a2_ei9_0_0.vlm"},
	{0x304BE63ED242E3ADULL, "6035904_m.fxe"},
	{0x304C0D1CAD15EB8DULL, "hlac_ht_mansion2_818_30x.dds"},
	{0x304CC1758F240713ULL, "zz_vault_debug6200390.cl"},
	{0x304CE9B549D27FBCULL, "vxlt_0_lowtown_docks_a3_n_c3z_0_0.vlm"},
	{0x304D7A3BC1906773ULL, "284652303.ogg"},
	{0x304D7BDC96E11115ULL, "6169191_m.fxe"},
	{0x304DAD4346283C19ULL, "hlac_w_bonepit_1_0z.dds"},
	{0x304E06CC7A84AD86ULL, "vxw_blightlands_lu_0_0.vlm"},
	{0x304EB982E3A097BEULL, "231467685.ogg"},
	{0x304EC793BD1EB9A6ULL, "hlac_w_woundedcoast_327682_0y.dds"},
	{0x304F4042F694790CULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.dura.gad"},
	{0x305083BDA0EA54A0ULL, "t3_hey_gn1.tnt"},
	{0x3050C7FA0B95CB14ULL, "vxlt_undercitydungeon_1ic_0_0.vlm"},
	{0x3050E497435E9D5BULL, "lt_0_lowtown_docks_a3_163_0.msh"},
	{0x3050F697435EBBA6ULL, "lt_0_lowtown_docks_a3_163_0.mao"},
	{0x30510297435ED045ULL, "lt_0_lowtown_docks_a3_163_0.mmh"},
	{0x305105AE97C167F0ULL, "vxlt_0_lowtown_n_gkp_0_0.vlm"},
	{0x30510E528F6AA77EULL, "vxw_blightlands_te_0_0.vlm"},
	{0x30519EFC56185D93ULL, "vxlt_gallowstemplar_430_0_0.vlm"},
	{0x3051F26E64029007ULL, "390297030.ogg"},
	{0x30523092A1AC677EULL, "vxw_sundermount_2gx_0_0.vlm"},
	{0x3052B22286B38E28ULL, "seb321_leliana6180941.cl"},
	{0x3052D5629BA692FFULL, "mag291_ntbanter6056831.cl"},
	{0x3052F20C21AE9F19ULL, "6178360_m.fxe"},
	{0x30532A5D828E98CAULL, "w_bonepit_36_0.mao"},
	{0x30532E5D828E9F19ULL, "w_bonepit_36_0.mmh"},
	{0x30533347021AD180ULL, "ser221_gamlen6133115.cl"},
	{0x3053385D828EB007ULL, "w_bonepit_36_0.msh"},
	{0x305381827450A874ULL, "ht_0_hightown_a3_17_0_lc.mmh"},
	{0x3053B9C04F854100ULL, "gen_im_trs_apo_lgt_glv_r0.uti"},
	{0x3053E3F71346BB04ULL, "vxw_bonepitdestroyed_a3_21g_0_0.vlm"},
	{0x3053FA3640850731ULL, "6123363_f.fxe"},
	{0x305478EF3996AEF7ULL, "hlac_lt_backalley_g_65536_6x.dds"},
	{0x3054DADE28A622B4ULL, "428233437.ogg"},
	{0x305515BA6533A3C5ULL, "hlac_lt_0_lowtown_docks_131072_1z.dds"},
	{0x30553F42F698E30AULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.dura.ani"},
	{0x305557DE7FDF6ACBULL, "abi_m_arcane_shield.bnk"},
	{0x305591FA14222665ULL, "vxca_caverns1_j24_0_0.vlm"},
	{0x3055A95929F40161ULL, "137014289.ogg"},
	{0x3055C38B2CE236CAULL, "aip_kirkwallguard2h_base.gda"},
	{0x3056C25F80A664FAULL, "vxlt_gallowscourtyard_n_74f_0_0.vlm"},
	{0x30572FB91EE0322CULL, "llac_w_sundermount_-65540_0.dds"},
	{0x3057CBFEF35DDFBAULL, "vxlt_0_lowtown_a2_n_goj_0_0.vlm"},
	{0x3057F32020B82BC5ULL, "hlac_lt_0_lowtown_docks_n_-3_4z.dds"},
	{0x3058559BEED282A9ULL, "vxlt_gallowscourtyard_3bp_0_0.vlm"},
	{0x30585B6639EE6AF6ULL, "818819573.ogg"},
	{0x3058B0DEBFCB60E5ULL, "271571093.ogg"},
	{0x3059C7FE1A14CF18ULL, "vxca_cavern1_haunted_jfx_0_0.vlm"},
	{0x3059DDFFDA58437FULL, "vxht_keep_n_6a0_0_0.vlm"},
	{0x305A1581C65CF784ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb8.msh"},
	{0x305A2B81C65D1D26ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb8.mmh"},
	{0x305ACDC26395F8CBULL, "vxlt_0_lowtown_docks_a3_d53_0_0.vlm"},
	{0x305AFD7E2FDD1269ULL, "dae000_flemeth6013337.cl"},
	{0x305B2672740AA855ULL, "6033627_m.fxe"},
	{0x305B409B4904CC1FULL, "15791431.ogg"},
	{0x305B5BC47681BC04ULL, "6111276_m.fxe"},
	{0x305C05B5AAC6BCC8ULL, "qun110_petrice6018574.cl"},
	{0x305C4BA81FC0F3B8ULL, "vxw_blightlands_q8_0_0.vlm"},
	{0x305C6FC55CF5FA4EULL, "vxht_keep_n_6b0_0_0.vlm"},
	{0x305C78A96C7AC606ULL, "vxw_blightlands_qv_0_0.vlm"},
	{0x305CA38ABCA578ADULL, "lt_0_lowtown_water_21475.msh"},
	{0x305CB18ABCA590E0ULL, "lt_0_lowtown_water_21475.mao"},
	{0x305CB58ABCA59733ULL, "lt_0_lowtown_water_21475.mmh"},
	{0x305CDF3AFCB69016ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.impa.ani"},
	{0x305D59B3031C197DULL, "hlac_lt_backalley_n_0_9y.dds"},
	{0x305D7AE067CD580EULL, "6142494_m.fxe"},
	{0x305DB4CC008EA6C0ULL, "vxht_0_hightown_a2_by8_0_0.vlm"},
	{0x305F1D9A0A9A999BULL, "llac_lt_gallowstemplar_2391_6.dds"},
	{0x305F27915BEE6FB6ULL, "588377831.ogg"},
	{0x305F3C91CCE607C8ULL, "llac_lt_0_lowtown_a2_n_c_-131071_2.dds"},
	{0x305F974DA15D940BULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.cess.evt"},
	{0x305FA5D5EBA2D554ULL, "vxht_0_hightown_bnt_0_1.vlm"},
	{0x305FAF2B97FEA1D4ULL, "vxca_caverns2_8qp_0_1.vlm"},
	{0x305FE4FD71E4D5EFULL, "250689436.ogg"},
	{0x305FEFABE07CDAB7ULL, "tbe100_xenon6190972.cl"},
	{0x305FF1BF8D8576C0ULL, "489439337.ogg"},
	{0x30600EF9053E901CULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.cess.ani"},
	{0x30602F97EF245FE3ULL, "vxca_caverns1_mines_iai_0_0.vlm"},
	{0x30607B6629CDCC01ULL, "272228839.ogg"},
	{0x3060A26F93A27FFEULL, "vxw_sundermount_2hy_0_0.vlm"},
	{0x3060CC0A2A08BB85ULL, "hlac_lt_0_lowtown_a2_n_c_0_5x.dds"},
	{0x30612322CD1CD3DCULL, "gen00fl_anders6202356.cl"},
	{0x3061BF518FC9C862ULL, "213459451.ogg"},
	{0x3061DFBB78F76120ULL, "ser120_ghyslain6105074.cl"},
	{0x306243F464E1DFDCULL, "775044472.ogg"},
	{0x30631DB13FA6DAE2ULL, "743927420.ogg"},
	{0x30638BAC534FAC47ULL, "120846547.ogg"},
	{0x30643C9EEED8EC3AULL, "character_character_p3.vsh.fxo.dx11.assembly"},
	{0x30646134941F54DAULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_player_isabellasreturn.gad"},
	{0x30647265900DAF8DULL, "vxlt_0_lowtown_docks_b5d_0_0.vlm"},
	{0x3064854CBE835808ULL, "zz_whatact.ncs"},
	{0x3064AD8ECE51D040ULL, "hlac_w_blightlands_327680_0y.dds"},
	{0x30652E6B2B5F00B0ULL, "6093595_f.fxe"},
	{0x30662E3E45B84732ULL, "plt_cod_mgc_tranquil.plo"},
	{0x306636062C48CB89ULL, "vxlt_gallowstemplar_n_5uv_0_0.vlm"},
	{0x30669956964F7F42ULL, "vxht_mansion_n_4w1_0_0.vlm"},
	{0x3066B04DA163DDCCULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.cess.gad"},
	{0x3066F01C49B972DCULL, "6068012_f.fxe"},
	{0x30671BF90544C55EULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.cess.gad"},
	{0x3067255E908229F1ULL, "6028388_m.fxe"},
	{0x3067684EC1405044ULL, "zz_isa_debug6133218.cl"},
	{0x30677E0B120FDF91ULL, "vxw_bonepitdestroyed_a3_19e_0_0.vlm"},
	{0x306861AA909E34CEULL, "vxht_keep_639_0_1.vlm"},
	{0x3068AFA1B5302FFAULL, "vxlt_0_lowtown_docks_a3_n_d42_0_0.vlm"},
	{0x3068DAC3E2CC53C1ULL, "dae000_sandal6147762.cl"},
	{0x306911901B7DF273ULL, "em_silsam.mor"},
	{0x3069309A66205AEFULL, "vxdr_primevaltomb_34p_0_0.vlm"},
	{0x306956EF6C7182F7ULL, "one140cr_charade6137074.cl"},
	{0x306984C9F4EBB80EULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he2a.phy"},
	{0x306A4347C21B132AULL, "lgt320_mis_resp_swordhunter6150274.cl"},
	{0x306A5F967ACD4215ULL, "6163391_m.fxe"},
	{0x306AC2D10835345DULL, "hlac_w_sundermount_-131078_0y.dds"},
	{0x306AE73AFCC2DB07ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.impa.evt"},
	{0x306AFB9121DC8C7BULL, "vxdr_0_deeproads_gb0_0_0.vlm"},
	{0x306B159500A234CBULL, "vrc321tr_housekeeper.ncs"},
	{0x306B44F5885B5A63ULL, "vxht_mansion2b_33x_0_0.vlm"},
	{0x306B4A0585670385ULL, "vxht_0_hightown_a2_9xe_0_0.vlm"},
	{0x306B4F8F16684B95ULL, "dre000_varric6030010.cl"},
	{0x306BB72C270D3062ULL, "zz_followers_banter6157828.cl"},
	{0x306BC97C3148C5CEULL, "gen00fl_merrill6142094.cl"},
	{0x306CA5CC01D84F4EULL, "hlac_ht_keep_3732_2x.dds"},
	{0x306D486AF7DBBCB3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up.ani"},
	{0x306D90C2913CD7E2ULL, "6128735_m.fxe"},
	{0x306DAF4DA169FACAULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.cess.ani"},
	{0x306DF6F9054AA57DULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.cess.evt"},
	{0x306E39ADA52518ACULL, "244988870.ogg"},
	{0x306E477BD45BB301ULL, "6206391_m.fxe"},
	{0x306F144C8598987EULL, "vxw_sundermount_2mz_0_0.vlm"},
	{0x306F270E325C978DULL, "vxdr_primevaltomb_8xu_0_0.vlm"},
	{0x306F2BC489FDA824ULL, "vxlt_undercityserialkiller_19n_0_0.vlm"},
	{0x306FE34B91470A1DULL, "6087422_m.fxe"},
	{0x30700DF11D572904ULL, "6121024_m.fxe"},
	{0x3070107540CBDBF9ULL, "vxdr_primevalentrance_311_0_0.vlm"},
	{0x307034840960B9FBULL, "6119572_m.fxe"},
	{0x30703EC19A37938BULL, "210994721.ogg"},
	{0x307079741BB735C7ULL, "161911691.ogg"},
	{0x3070E36B6890CEC6ULL, "236769186.ogg"},
	{0x30712084A1D23BB2ULL, "326694903.ogg"},
	{0x307139CFE3CC6467ULL, "6149075_f.fxe"},
	{0x307159E0255596BBULL, "vxht_mansion2b_27m_0_0.vlm"},
	{0x3071735FF4FA3EA3ULL, "784372750.ogg"},
	{0x307196CE9CC780BFULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_csand01_da.dds"},
	{0x3071A4DCE195CCE7ULL, "vxht_mansion2b_a9_0_0.vlm"},
	{0x3071E1F67EB04E39ULL, "hlac_lt_0_lowtown_a2_65536_4y.dds"},
	{0x3072083AFCC93250ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.impa.gad"},
	{0x3072D8A600767CDFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_gallowsalcov.msh"},
	{0x3072DEA600768681ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_gallowsalcov.mmh"},
	{0x3072EAA600769B22ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_gallowsalcov.mao"},
	{0x3072F1C4CFED4428ULL, "hlac_lt_0_lowtown_-131072_0x.dds"},
	{0x307301CC3EA40CAAULL, "985420589.ogg"},
	{0x3073C5128419158FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_left.gad"},
	{0x3073F49001FA57F9ULL, "lgt360tr_orlanna.ncs"},
	{0x3074AFF7DB0AFCEDULL, "dae000_gen_turn_in_1.bnk"},
	{0x30752FE08FE1C40EULL, "art\\characters\\weapons\\mauls\\w_mal_ml103.mao"},
	{0x30756B478588A712ULL, "285588815.ogg"},
	{0x3075850732C867FEULL, "ht_0_hightown_a3_16_0.rml"},
	{0x3075B5A7403BAC39ULL, "llac_lt_0_lowtown_a2_n_c_-131072_8.dds"},
	{0x3076108BB6BF17B1ULL, "vxca_caverns1_i2g_0_0.vlm"},
	{0x30763825B1E94FC3ULL, "713726683.ogg"},
	{0x3076E6B209D08A9FULL, "73076842.ogg"},
	{0x3076EAC5399BEB36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_small.evt"},
	{0x3077150D41F0A5B0ULL, "dae010pt_amb_lowtown.ncs"},
	{0x30772B5BD5CD4ADCULL, "art\\characters\\playercharacter\\humanmale\\bdy_masf.phy"},
	{0x30773952342290D1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgtb_0s.dds"},
	{0x30774E0C8784403BULL, "hlac_lt_0_lowtown_a2_n_c_-65537_5y.dds"},
	{0x307759316EE3D911ULL, "vxca_cavern1_haunted_inb_0_0.vlm"},
	{0x3077D1F7DB0D6C55ULL, "dae000_gen_turn_in_1.cnv"},
	{0x307804FC05D8F6B5ULL, "hlac_lt_darktown_-1_13y.dds"},
	{0x30787C3494302BC0ULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_player_isabellasreturn.ani"},
	{0x3078C08067BAB7F9ULL, "6147021_m.fxe"},
	{0x3078C789D82C3A99ULL, "vxht_mansion2b_325_0_0.vlm"},
	{0x307902AC6E29C872ULL, "6323921.ogg"},
	{0x3079144F2A4CEA32ULL, "hlac_ca_caverns2_5802_71y.dds"},
	{0x30799A65E014945AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_plain.msh"},
	{0x30799CC29D671876ULL, "art\\toolset\\tools_start_pt.mao"},
	{0x3079A865E014AB93ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_plain.mao"},
	{0x3079A8C29D672C15ULL, "art\\toolset\\tools_start_pt.mmh"},
	{0x3079AAC29D672F8BULL, "art\\toolset\\tools_start_pt.msh"},
	{0x3079AC65E014B2A0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_plain.mmh"},
	{0x3079B5B307A431D6ULL, "260194822.ogg"},
	{0x3079DA12841DA53CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_left.evt"},
	{0x307A2FC8575AB1A1ULL, "vxlt_hangedman_2k1_0_0.vlm"},
	{0x307A3F446E4DF270ULL, "mag311_emile_de_launcet6043502.cl"},
	{0x307A4C63F111BF54ULL, "mag170_thrask6054024.cl"},
	{0x307B406AF7E7ECEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up.evt"},
	{0x307BD7E40D9C4E87ULL, "833212956.ogg"},
	{0x307C407D749149DAULL, "vxca_cavern1_haunted_eia_0_1.vlm"},
	{0x307C5C191CFB13A8ULL, "hlac_ca_caverns2prc_5802_48z.dds"},
	{0x307D217AFF1D7038ULL, "qcr220_orsino6061334.cl"},
	{0x307D7B6422D7E807ULL, "vxht_0_hightown_n_b7b_0_0.vlm"},
	{0x307DE1C539A1FAC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_small.gad"},
	{0x307DE4862FECACBCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9dd_doorframe.mao"},
	{0x307DE8DEF8C2A6B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_back.ani"},
	{0x307DF0862FECC15FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9dd_doorframe.mmh"},
	{0x307DF6862FECCB01ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9dd_doorframe.msh"},
	{0x307E41696AEFD622ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_basic.phy"},
	{0x307EAF33E0C9DC50ULL, "mer141_jansen6031855.cl"},
	{0x307EFDA686E5C5D3ULL, "vxht_0_hightown_byc_0_0.vlm"},
	{0x307F1A861C79E1E1ULL, "mag111ip_backroom_door2.utp"},
	{0x307F7EE445A1D63EULL, "hlac_w_encounter3_0_3z.dds"},
	{0x307FB4A1525D4B65ULL, "character_character_p62.psh.fxo.dx11"},
	{0x307FCBBD174BAF3BULL, "vxht_mansion2b_20l_0_0.vlm"},
	{0x308000A520BBB035ULL, "418093187.ogg"},
	{0x3080BF38202F8EE4ULL, "155844586.ogg"},
	{0x308103CE8F3A4FA8ULL, "montage_subs_demohack6217093.cl"},
	{0x308121F9B3616511ULL, "335485135.ogg"},
	{0x3081234C1739058FULL, "custic00b.dds"},
	{0x308137CD6C815234ULL, "27060940.ogg"},
	{0x308166E4207E7D48ULL, "358753162.ogg"},
	{0x308207504F1282F4ULL, "fen310_party_comment6116859.cl"},
	{0x30821F6AF7EDD35DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up.gad"},
	{0x3082621E8844EBDBULL, "hlac_ht_mansion2b_818_7x.dds"},
	{0x308319C01034FC89ULL, "vxlt_gallowstemplar_n_5ct_0_0.vlm"},
	{0x30832A54F22D77A1ULL, "llac_lt_0_lowtown_docks_65533_1.dds"},
	{0x30837D107A3BB042ULL, "vxht_mansion_n_4a3_0_0.vlm"},
	{0x308391F27F18788EULL, "157381131.ogg"},
	{0x3083BC8BD4838247ULL, "vxdr_primevalentrance_t5_0_0.vlm"},
	{0x30842BB02430A51FULL, "llac_lt_backalley_n_65536_3.dds"},
	{0x3084776B1B4983F2ULL, "plt_cod_aow_quick_potions.plo"},
	{0x30849578381C771DULL, "6077670_m.fxe"},
	{0x3084E2C539A81B6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_small.ani"},
	{0x3084F3DEF8C8D853ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_back.gad"},
	{0x30856E8A2FB61E9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flrourish_additive.gad"},
	{0x30857A870DB837B4ULL, "6101541_m.fxe"},
	{0x3085CB0E60D9F191ULL, "vxca_cavern1_haunted_ika_0_0.vlm"},
	{0x3086862BF0F43EDDULL, "w_bonepit_194_0.mao"},
	{0x30868A2BF0F44516ULL, "w_bonepit_194_0.mmh"},
	{0x3086942BF0F45614ULL, "w_bonepit_194_0.msh"},
	{0x3086B5E401C43256ULL, "zz_followers_banter6159066.cl"},
	{0x3086CEE93328ACE1ULL, "402910632.ogg"},
	{0x3087327348100136ULL, "dae000_fem_commoner1_mob6174480.cl"},
	{0x30876B4D13889867ULL, "vxw_bonepitdestroyed_a3_1vn_0_0.vlm"},
	{0x30877698162E60C9ULL, "llac_lt_0_lowtown_a2_n_c_65538_4.dds"},
	{0x3087E2128429F055ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_left.ani"},
	{0x30887761A17C2354ULL, "vxht_0_hightown_b9p_0_0.vlm"},
	{0x3088F425A3CC4F5EULL, "vxca_cavern1_haunted_j23_0_1.vlm"},
	{0x30891A9500BB3AB2ULL, "vrc321tr_housekeeper.utt"},
	{0x308949BAE1044F7FULL, "lgt250st_dalish_conflict.stg"},
	{0x3089886F21108222ULL, "qun131_taskmaster6039692.cl"},
	{0x3089DC6F3623C7DBULL, "vxw_woundedcoast_17v_0_0.vlm"},
	{0x308A5AF84BAFD599ULL, "6115417_m.fxe"},
	{0x308A8C2E0D36ACAEULL, "qun160_vashothlead6053878.cl"},
	{0x308AE033E754DF88ULL, "hf_one_ggt_ginnis.mor"},
	{0x308AF0DEF8CD3EB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_back.evt"},
	{0x308B286C97AEC646ULL, "vxdr_primevaltomb_2w3_0_2.vlm"},
	{0x308B3593910F43F3ULL, "85962995.ogg"},
	{0x308C0F7E6DE9D924ULL, "vxlt_undercityserialkiller_13l_0_0.vlm"},
	{0x308C2FA70FE8DCE8ULL, "6043526_m.fxe"},
	{0x308C638A2FBC2A8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flrourish_additive.evt"},
	{0x308D781E201179E6ULL, "mer141_jansen6077802.cl"},
	{0x308DA5537C59535EULL, "6177685_m.fxe"},
	{0x308EDD4D676F6A8AULL, "486660581.ogg"},
	{0x308EE38427C7528BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_meda_0s.dds"},
	{0x308F4D7AB97DDA91ULL, "w_blightlands_64_0.msh"},
	{0x308F5B7AB97DF2ACULL, "w_blightlands_64_0.mao"},
	{0x308F677AB97E074FULL, "w_blightlands_64_0.mmh"},
	{0x308F944980BF1B2CULL, "hlac_dr_slavers_3217_9z.dds"},
	{0x30901AD0DB797CF8ULL, "6051640_m.fxe"},
	{0x30909F3B4691B217ULL, "vxlt_0_lowtown_docks_a3_n_brc_0_0.vlm"},
	{0x3090FB257E6C736DULL, "166833828.ogg"},
	{0x30912B351FA9DC22ULL, "dae000_keep_servant_1_sta6161375.cl"},
	{0x30913FF701A50F33ULL, "6167982_m.fxe"},
	{0x309145CA611060D4ULL, "fen210_dying_magister6108620.cl"},
	{0x3091526B93EF5E65ULL, "vxlt_hangedman_2mk_0_0.vlm"},
	{0x30915410F9AA724DULL, "dae200_bran6144574.cl"},
	{0x3091DEEB8B72DB2AULL, "hlac_dr_primevaltomb_658_2y.dds"},
	{0x3091F632F204ACD4ULL, "446664349.ogg"},
	{0x309262958BE2A331ULL, "822002612.ogg"},
	{0x30927DF4E1137A8FULL, "vxlt_gallowstemplar_395_0_0.vlm"},
	{0x3092A0D81B7B122BULL, "art\\vfx\\cinematic\\vfx_blood_directional_l.impa.ani"},
	{0x3092B6AADA12CCACULL, "315056828.ogg"},
	{0x3092DDE87131AE67ULL, "6119644_m.fxe"},
	{0x309399F3DDB32058ULL, "follower_banter6143410.cl"},
	{0x3093A745B9A4BF45ULL, "vxlt_0_lowtown_docks_a3_n_d49_0_0.vlm"},
	{0x30944B4E6B13309DULL, "hlac_w_bonepit_0_5z.dds"},
	{0x30945FA1B6A44F18ULL, "vxw_bonepit_107_0_0.vlm"},
	{0x3094C13B5F1A1C81ULL, "787613292.ogg"},
	{0x3094C85393E04310ULL, "vxca_caverns1_mines_dpx_0_0.vlm"},
	{0x3094F96CDC65619FULL, "vxlt_undercityserialkiller_1mu_0_0.vlm"},
	{0x309548460FC94A71ULL, "cod_hst_kirk_arch_1a_11.utp"},
	{0x3095BBFBE8C987C9ULL, "6129906_m.fxe"},
	{0x3095BD4B43508C6EULL, "990853088.ogg"},
	{0x3095D3ADA42901E1ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft180.ani"},
	{0x30961A9312A01E1FULL, "dae000_merchant_weapons_dm6164405.cl"},
	{0x309629014D3C8CD3ULL, "6148307_m.fxe"},
	{0x30965CB8C6E09E09ULL, "6125817_f.fxe"},
	{0x30968FB87F471852ULL, "6037601_f.fxe"},
	{0x309751E2443C0B42ULL, "m_var321ar_mansion_haunted.wbd"},
	{0x30977FDDEFA992CCULL, "qun120_arishok6050295.cl"},
	{0x30984E4C2819E05BULL, "vxw_woundedcoast_12i_0_0.vlm"},
	{0x3098533F1039B957ULL, "dae200cr_gamlen.utc"},
	{0x309886D622AA3C78ULL, "mer240cr_jansen.utc"},
	{0x309905977A4E364AULL, "707944193.ogg"},
	{0x30994CB48AC782AAULL, "llac_lt_0_lowtown_a3_n_c_-2_0.dds"},
	{0x3099A535818753A8ULL, "sst_ss_abomination_1.gda"},
	{0x3099B85082B75ECAULL, "6115506_m.fxe"},
	{0x309A4B8A2FC83FE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flrourish_additive.ani"},
	{0x309AF90DDD4F23BBULL, "gen00fl_varric6093632.cl"},
	{0x309B3581741C7A71ULL, "zz_qun_debug6034111.cl"},
	{0x309B93BBF633473BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02deadtree.mmh"},
	{0x309B9FBBF6335BD8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02deadtree.mao"},
	{0x309BB1BBF6337A25ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02deadtree.msh"},
	{0x309CAD11D42BD24CULL, "6172820_m.fxe"},
	{0x309CB20B8F89D488ULL, "6170084_m.fxe"},
	{0x309CD44300F134E9ULL, "primitives_primitive_p5.psh.fxo"},
	{0x309CDEADA42F3283ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft180.gad"},
	{0x309D02DDF06063CAULL, "vxht_mansion2b_3eo_0_0.vlm"},
	{0x309DC463E9CD2615ULL, "232372398.ogg"},
	{0x309E0EC72B1853E9ULL, "6059317_m.fxe"},
	{0x309E1CCCB2F25642ULL, "758967082.ogg"},
	{0x309ED86349CD0E94ULL, "dae300_bran6143506.cl"},
	{0x309EDFA6009BC3FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_gallowsalcov.phy"},
	{0x309F9470E291872FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chantry_statues_d.dds"},
	{0x309FFC4A6EAEE2DEULL, "6080669_f.fxe"},
	{0x30A03E629539E50EULL, "vxht_0_hightown_a3_bz2_0_0.vlm"},
	{0x30A06D10CE9BD2A3ULL, "6054521_m.fxe"},
	{0x30A088D81B8726A2ULL, "art\\vfx\\cinematic\\vfx_blood_directional_l.impa.evt"},
	{0x30A0A5C86ADFF881ULL, "6132731_m.fxe"},
	{0x30A0AAC41D47A66BULL, "hlac_ht_chantry_2534_50y.dds"},
	{0x30A0B106ADF59DC7ULL, "180246019.ogg"},
	{0x30A0CDC71D4BAC74ULL, "zz_one_debug6075014.cl"},
	{0x30A0E8F8FD8C2709ULL, "vxdr_0_deeproads_d32_0_0.vlm"},
	{0x30A101DD55D6776DULL, "hlac_w_sundermount_2_0z.dds"},
	{0x30A20E9D34223CBFULL, "ht_keep_n_9215.rml"},
	{0x30A24A34EDE62E0EULL, "rmtst_probe_3915.mtx"},
	{0x30A3116028765567ULL, "221663858.ogg"},
	{0x30A34D46AA0924F0ULL, "619093358.ogg"},
	{0x30A359876E5B398EULL, "vxlt_gallowsprison_n_64a_0_0.vlm"},
	{0x30A3873340E787CAULL, "follower_barks6153388.cl"},
	{0x30A38881C9F3F882ULL, "vxlt_gallowstemplar_3mh_0_0.vlm"},
	{0x30A3BBADA4351668ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft180.evt"},
	{0x30A3C3501E91F79AULL, "vxlt_warehouse_p5_0_0.vlm"},
	{0x30A3F8E7D9367416ULL, "6163239_m.fxe"},
	{0x30A41C696B1066D6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_basic.mao"},
	{0x30A428696B107B75ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_basic.mmh"},
	{0x30A42A696B107EEBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_basic.msh"},
	{0x30A454D18146740BULL, "sebastian.cut"},
	{0x30A482EB8F77E073ULL, "dae000_meredith6085256.cl"},
	{0x30A4C33F5954AABDULL, "vxht_mansion2b_3g2_0_0.vlm"},
	{0x30A4F9E1FB56ECE7ULL, "lit_up_04_l2.mao"},
	{0x30A4FFB4AAC054B5ULL, "art\\levels\\races\\design\\_textures\\prp_bodybagcoins_dp.dds"},
	{0x30A5699C1EC7C994ULL, "vxht_0_hightown_a3_9p9_0_0.vlm"},
	{0x30A5C7EAABBDF78AULL, "6122077_f.fxe"},
	{0x30A62874E616FECDULL, "vxw_woundedcoast_1ef_0_0.vlm"},
	{0x30A6ECCEC1DB6D12ULL, "359213575.ogg"},
	{0x30A787D81B8D43B5ULL, "art\\vfx\\cinematic\\vfx_blood_directional_l.impa.gad"},
	{0x30A7A98386454EEEULL, "6134510_m.fxe"},
	{0x30A80504E45167D7ULL, "vxca_caverns2_9fn_0_0.vlm"},
	{0x30A811B92D270BEEULL, "307350600.ogg"},
	{0x30A90525D98A11EDULL, "690137647.ogg"},
	{0x30A906BF8A178C5CULL, "vxht_0_hightown_c8f_0_0.vlm"},
	{0x30A913B95447ED46ULL, "art\\vfx\\combat\\vfx_w_wns_front_long1_c.impa.ani"},
	{0x30A93D87349C8E52ULL, "6206752_m.fxe"},
	{0x30A9E2CE0E749885ULL, "6141764_m.fxe"},
	{0x30AA14DE4FE53615ULL, "ser191_emeric6108146.cl"},
	{0x30AA1AA39F62EA06ULL, "vxht_chantry_1xf_0_0.vlm"},
	{0x30AA83C9F5234DD1ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he2a.mmh"},
	{0x30AA9DC9F523798FULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he2a.msh"},
	{0x30AADC6249B994C6ULL, "vxlt_0_lowtown_docks_a3_d1f_0_0.vlm"},
	{0x30AAE466A1FA6C4EULL, "297662486.ogg"},
	{0x30AAE838033A399FULL, "hlac_ht_keep_4519_15y.dds"},
	{0x30AB555717875012ULL, "1013192520.ogg"},
	{0x30AB724CEF4A4CEAULL, "6047450_m.fxe"},
	{0x30AB7FCC48D8ADC8ULL, "lt_0_lowtown_docks_n_252_0.msh"},
	{0x30AB91CC48D8CC59ULL, "lt_0_lowtown_docks_n_252_0.mao"},
	{0x30AB9B0947064E38ULL, "896620940.ogg"},
	{0x30AB9DCC48D8E0F2ULL, "lt_0_lowtown_docks_n_252_0.mmh"},
	{0x30ABC34CB0F6A4E0ULL, "dae000_gen_turn_in_16198711.cl"},
	{0x30ABFDE862E1E048ULL, "zz_dre_debug6056988.cl"},
	{0x30AC0FBEDB1B0401ULL, "hlac_ht_mansion2_818_7x.dds"},
	{0x30AD677F282676C2ULL, "art\\levels\\4cinematics\\dwarf\\_7_-_default.mao"},
	{0x30AE4CE62BFB127BULL, "619685969.ogg"},
	{0x30AED385177284ADULL, "924300380.ogg"},
	{0x30AEFE27262B80B6ULL, "vxht_0_hightown_a2_bpi_0_0.vlm"},
	{0x30AF2E5D82DCFE72ULL, "w_bonepit_36_0.rml"},
	{0x30AF820732F9DAA2ULL, "ht_0_hightown_a3_16_0.anb"},
	{0x30AFB112D20BAAA9ULL, "624658073.ogg"},
	{0x30AFBDAF4FAEE49CULL, "vxlt_0_lowtown_docks_n_cs1_0_0.vlm"},
	{0x30AFC4958D2F45C0ULL, "6126487_m.fxe"},
	{0x30AFF749B652B82BULL, "hlac_w_sundermount_-1_0z.dds"},
	{0x30B010BC994EDE5CULL, "401666634.ogg"},
	{0x30B02C6980765561ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportin.ani"},
	{0x30B046D083C78D23ULL, "zz_vrc_debug6064258.cl"},
	{0x30B0E2F133711082ULL, "6176004_m.fxe"},
	{0x30B0F0DB7915862CULL, "6140300_m.fxe"},
	{0x30B1517987B2CB17ULL, "46470637.ogg"},
	{0x30B1A240A29BE977ULL, "vxht_mansion2b_20x_0_0.vlm"},
	{0x30B1A2BA90EEB9C1ULL, "hf_comt_servant_1b.mor"},
	{0x30B1C0B9FCFFE6A1ULL, "6122764_f.fxe"},
	{0x30B231BB56DFC720ULL, "mcr370_cut_meredith_dies6081067.cl"},
	{0x30B2C0E3E96AE1E7ULL, "vxw_bonepitdestroyed_a3_1ak_0_0.vlm"},
	{0x30B3BAE10F9F50C9ULL, "vxdr_primevaltomb_6m8_0_0.vlm"},
	{0x30B40905B22F8098ULL, "vxlt_backalley_g_1fw_0_0.vlm"},
	{0x30B47E36BFF7E670ULL, "vxca_caverns2prc_7nv_0_0.vlm"},
	{0x30B532060C06115BULL, "vxw_woundedcoast_18k_0_0.vlm"},
	{0x30B5470211FC74CAULL, "335742532.ogg"},
	{0x30B63431775EDB40ULL, "6084866_m.fxe"},
	{0x30B6B2831E372ECDULL, "953298296.ogg"},
	{0x30B6CEA9758D5A88ULL, "6117825_m.fxe"},
	{0x30B6FBB9545401B7ULL, "art\\vfx\\combat\\vfx_w_wns_front_long1_c.impa.evt"},
	{0x30B73769807C8603ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportin.gad"},
	{0x30B7AAC8547067E3ULL, "hlac_lt_0_lowtown_docks_n_-3_7y.dds"},
	{0x30B7EE190A161134ULL, "vxlt_0_lowtown_docks_ctk_0_0.vlm"},
	{0x30B7FEDCBBD8BE0DULL, "vxca_cavern1_haunted_j69_0_0.vlm"},
	{0x30B81995C83E4D88ULL, "ser200_aveline6201587.cl"},
	{0x30B8A8E73B1380A8ULL, "253357542.ogg"},
	{0x30B8CCA8F59B4F40ULL, "vxca_caverns2prc_8cz_0_0.vlm"},
	{0x30B932B65702FB36ULL, "cn_hed_wera_0t.dds"},
	{0x30B93C4C07A1A021ULL, "51021642.ogg"},
	{0x30B9481DBE540807ULL, "6121975_m.fxe"},
	{0x30BA3623758B6B59ULL, "vxht_mansion2b_1n3_0_0.vlm"},
	{0x30BA8D425022C277ULL, "llac_lt_0_lowtown_docks_n_131071_3.dds"},
	{0x30BAB6EEC6A3D092ULL, "335754232.ogg"},
	{0x30BB1A4D3EAA665DULL, "6163225_m.fxe"},
	{0x30BBDADB49791EB5ULL, "ser221_gascard6134510.cl"},
	{0x30BC0DCC3F5B05C6ULL, "vxlt_darktown_2vp_0_0.vlm"},
	{0x30BC6B1358AE8064ULL, "6126502_m.fxe"},
	{0x30BC7CB03A72249DULL, "vxw_woundedcoast_5k_0_0.vlm"},
	{0x30BC96C8059591C7ULL, "269192792.ogg"},
	{0x30BD11898C6915FCULL, "w_bonepitmining_a2_9_0.mao"},
	{0x30BD1D898C69299FULL, "w_bonepitmining_a2_9_0.mmh"},
	{0x30BD221C7926160EULL, "vxht_0_hightown_a3_bl4_0_0.vlm"},
	{0x30BD23898C693441ULL, "w_bonepitmining_a2_9_0.msh"},
	{0x30BDA846CF2D5142ULL, "one120_athenril6088319.cl"},
	{0x30BDD0532428FDA9ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc02.phy"},
	{0x30BDDA44F1129FC7ULL, "6075721_f.fxe"},
	{0x30BDE2E4CD4BD1ACULL, "llac_lt_0_lowtown_a2_n_-262144_0.dds"},
	{0x30BDFCB9545A2240ULL, "art\\vfx\\combat\\vfx_w_wns_front_long1_c.impa.gad"},
	{0x30BE1469808269E8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportin.evt"},
	{0x30BEB393B95DF678ULL, "6034813_f.fxe"},
	{0x30BFA8D33384C8E2ULL, "w_woundedcoast_249_0.msh"},
	{0x30BFAAD33384CB98ULL, "w_woundedcoast_249_0.mmh"},
	{0x30BFB6D33384E03BULL, "w_woundedcoast_249_0.mao"},
	{0x30BFD08E0F1A7BBDULL, "hlac_ht_0_hightown_n_-196607_0y.dds"},
	{0x30C09CCC6C21417AULL, "ave200_donnic6110709.cl"},
	{0x30C132C0DB60FA67ULL, "vxw_bonepitdestroyed_a3_1jj_0_0.vlm"},
	{0x30C141B4519A5E02ULL, "40542485.ogg"},
	{0x30C148A02F7DF092ULL, "6173120_m.fxe"},
	{0x30C150CB65385867ULL, "t3_abm_h03.tnt"},
	{0x30C1692CED494E01ULL, "hlac_w_woundedcoast_-131069_0y.dds"},
	{0x30C1B25B2C1E186EULL, "vxht_mansion2_20u_0_0.vlm"},
	{0x30C267751A2423A3ULL, "501070347.ogg"},
	{0x30C27CBBA2592691ULL, "vxlt_gallowscourtyard_7cw_0_0.vlm"},
	{0x30C2B7864DB96A18ULL, "6069724_m.fxe"},
	{0x30C2D9BA0E28BEBBULL, "lowtown_male_alienage1_sta6162810.cl"},
	{0x30C30310D4A5D2D0ULL, "6142503_f.fxe"},
	{0x30C30C2ECA032FCDULL, "vxw_woundedcoast_1kd_0_0.vlm"},
	{0x30C30F12EDA927ECULL, "285508156.ogg"},
	{0x30C32B1274359E2DULL, "dae000cr_grand_cleric6124735.cl"},
	{0x30C380F0981BC0D8ULL, "6085717_m.fxe"},
	{0x30C3A03DF05C76C8ULL, "6133063_m.fxe"},
	{0x30C3D1E90ADBA35CULL, "977911873.ogg"},
	{0x30C4D1722BAEDE72ULL, "vxlt_undercityencounter_wm_0_0.vlm"},
	{0x30C50045EDB208A3ULL, "6110121_f.fxe"},
	{0x30C5EA796E03BD5CULL, "vxht_0_hightown_c2d_0_0.vlm"},
	{0x30C5F978B91F9C62ULL, "6140261_m.fxe"},
	{0x30C6649CE3D57746ULL, "242617321.ogg"},
	{0x30C69AFD3B20043EULL, "vxca_cavern1_haunted_jg1_0_0.vlm"},
	{0x30C6A5FBF5CE82B7ULL, "follower_banter6141668.cl"},
	{0x30C709F745EB2E16ULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em_facem1.mao"},
	{0x30C782E5C6893EACULL, "117637481.ogg"},
	{0x30C787EFC7E05322ULL, "vxlt_gallowstemplar_f_5rh_0_0.vlm"},
	{0x30C7B04566B727C5ULL, "llac_ht_keep_4519_10.dds"},
	{0x30C7CDE3E5999703ULL, "art\\effects\\cloudshadows\\w_bonepitmining_a2\\cloud2.dds"},
	{0x30C7E5786DACC3CFULL, "hlac_ca_caverns1_24421_4y.dds"},
	{0x30C8627062E161DCULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_glv_cas_0n.dds"},
	{0x30C8A800678183D9ULL, "vxht_mansion2b_1a2_0_0.vlm"},
	{0x30CA2AB6A8DF01BCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2keepwindow.mao"},
	{0x30CA36B6A8DF165FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2keepwindow.mmh"},
	{0x30CA3CB6A8DF2001ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_x2keepwindow.msh"},
	{0x30CA66CEC8A8B00BULL, "388780935.ogg"},
	{0x30CA7FA931511E46ULL, "vxlt_darktown_2yq_0_0.vlm"},
	{0x30CAEE8D2C683D1DULL, "vxw_woundedcoast_2j_0_0.vlm"},
	{0x30CB5A8F8B635222ULL, "198332831.ogg"},
	{0x30CB70CF31918D81ULL, "vxht_0_hightown_c87_0_0.vlm"},
	{0x30CB89E302CA5419ULL, "dae000_meredith6085269.cl"},
	{0x30CBC137D90751D4ULL, "vxdr_primevaltomb_6xf_0_0.vlm"},
	{0x30CBE8E64094DCB8ULL, "vxlt_0_lowtown_a2_n_cmv_0_0.vlm"},
	{0x30CC463C33501BE4ULL, "6172716_m.fxe"},
	{0x30CD51F464E7730CULL, "vxca_caverns2_8oj_0_1.vlm"},
	{0x30CD66DAB735EDE6ULL, "923900078.ogg"},
	{0x30CD7538201E3908ULL, "gen00fl_isabela_init6077911.cl"},
	{0x30CDA796C0A21796ULL, "6110696_f.fxe"},
	{0x30CDB45379145B28ULL, "llac_lt_0_lowtown_docks_-65539_5.dds"},
	{0x30CE0CEF0278928BULL, "mag291_torpor6026995.cl"},
	{0x30CE524C1745348BULL, "hlac_lt_warehouse_65537_2z.dds"},
	{0x30CEA375BBA8B924ULL, "isa311_isabela6092227.cl"},
	{0x30CED8FCD2A3B49FULL, "gen00fl_sebastian6174716.cl"},
	{0x30CF40BFA9234975ULL, "plt_prm_cod_deep_blade.plo"},
	{0x30D02659B5453EFDULL, "vxlt_0_lowtown_docks_a3_9f5_0_0.vlm"},
	{0x30D032BD345AA9F3ULL, "vxca_caverns2prc_989_0_0.vlm"},
	{0x30D0A10BA668A1AFULL, "vxlt_gallowstemplar_n_3gd_0_0.vlm"},
	{0x30D0AE1744415BBEULL, "6028937_f.fxe"},
	{0x30D103DA1ACB91B2ULL, "6015776_m.fxe"},
	{0x30D118C9A19F4C89ULL, "781022674.ogg"},
	{0x30D2096A3CBBB9C5ULL, "414668635.ogg"},
	{0x30D20C7D00032C70ULL, "6034937_m.fxe"},
	{0x30D2736CC38DA68EULL, "188159950.ogg"},
	{0x30D2F4A6462CE4CDULL, "w_bonepitmining_a2_245_0.mmh"},
	{0x30D2F56B209662A6ULL, "79438551.ogg"},
	{0x30D2F8A6462CEB1EULL, "w_bonepitmining_a2_245_0.mao"},
	{0x30D3028A344FC2F8ULL, "149881191.ogg"},
	{0x30D306A6462D0353ULL, "w_bonepitmining_a2_245_0.msh"},
	{0x30D3434F1DA4F6F2ULL, "vxlt_undercityencounter_rn_0_0.vlm"},
	{0x30D3CCBE2B581D7EULL, "214184112.ogg"},
	{0x30D3DC565E072043ULL, "zz_dae_debug6029496.cl"},
	{0x30D49DF3CB2D0E57ULL, "vxdr_primevaltomb_7ub_0_1.vlm"},
	{0x30D4B63875647962ULL, "vxht_0_hightown_a3_a6m_0_0.vlm"},
	{0x30D4BCCE91C8585BULL, "598857742.ogg"},
	{0x30D4EB7107CEC361ULL, "6096280_m.fxe"},
	{0x30D4F4396AD2201AULL, "333273649.ogg"},
	{0x30D50BC4A976A769ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.cess.ani"},
	{0x30D53463DDBB6D40ULL, "dae100_viscount6101863.cl"},
	{0x30D54834532AD887ULL, "57391279.ogg"},
	{0x30D5E259E6A24B01ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_hightown_fabric_s.dds"},
	{0x30D666A19CBEF301ULL, "hlac_ht_0_hightown_n_-65537_1y.dds"},
	{0x30D7004875395572ULL, "6017671_m.fxe"},
	{0x30D711CE515C4195ULL, "vxw_encounter2_camp_jw_0_0.vlm"},
	{0x30D719DD59779C59ULL, "vxht_mansion2b_1d1_0_0.vlm"},
	{0x30D741ABC50FEFAEULL, "6044886_m.fxe"},
	{0x30D7B1C7D6C833B9ULL, "vxlt_gallowstemplar_5cs_0_0.vlm"},
	{0x30D7B9BDF6103444ULL, "vxht_mansion2_3e9_0_0.vlm"},
	{0x30D7BA7CEDB18570ULL, "cc_gib100.dds"},
	{0x30D84F29748DC610ULL, "vxlt_gallowscourtyard_77f_0_0.vlm"},
	{0x30D8EAE01EF97011ULL, "vxdr_primevaltomb_85z_0_0.vlm"},
	{0x30D904ED2E920938ULL, "gen00fl_carver6103390.cl"},
	{0x30D934EC6B78F350ULL, "vxlt_hangedman_n_18u_0_0.vlm"},
	{0x30D9B4193359D807ULL, "llac_lt_gallowscourtyard_n_-131073_0.dds"},
	{0x30D9E082044FB4C1ULL, "dae000ip_crowd_ff7_talk10.utp"},
	{0x30DA9BBD1D093446ULL, "hlac_ca_cavern1_haunted_24421_9z.dds"},
	{0x30DC26C4A97CF3EBULL, "art\\vfx\\combat\\vfx_m_bld_control_c.cess.gad"},
	{0x30DC4451398CC0E9ULL, "vxca_caverns1_i1k_0_1.vlm"},
	{0x30DC8708F5801877ULL, "dae000_worthy6166589.cl"},
	{0x30DCD37F94F911C3ULL, "vxht_bartrand_n_68d_0_0.vlm"},
	{0x30DD8B027CD722C2ULL, "vxlt_shop_dv_0_0.vlm"},
	{0x30DD93C20F2C822BULL, "vxlt_darktown_48a_0_0.vlm"},
	{0x30DD94338CA8186CULL, "vxlt_0_lowtown_a2_n_c_gc2_0_0.vlm"},
	{0x30DDE7FF890C9263ULL, "vxlt_0_lowtown_docks_n_aqb_0_0.vlm"},
	{0x30DE5E7519AFD3DAULL, "vxht_0_hightown_n_alq_0_0.vlm"},
	{0x30DE5FE2447825F8ULL, "m_var321ar_mansion_haunted.bnk"},
	{0x30DE92F71EEADFC3ULL, "art\\characters\\weapons\\axes\\w_axe_ax103a.mmh"},
	{0x30DEA0F71EEAF71DULL, "art\\characters\\weapons\\axes\\w_axe_ax103a.msh"},
	{0x30DF1D8B0DAF7BB7ULL, "art\\vfx\\creatures\\fxc_withfn_c.anb"},
	{0x30DF607586455791ULL, "vxlt_gallowscourtyard_7iu_0_0.vlm"},
	{0x30DF8E090C698528ULL, "6042074_f.fxe"},
	{0x30DFD57363481C0DULL, "461836450.ogg"},
	{0x30DFEFE8ADEF60CDULL, "vxw_woundedcoast_1qb_0_0.vlm"},
	{0x30DFF986303F9AA4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9dd_doorframe.phy"},
	{0x30E05E66796AAC32ULL, "vxlt_darktown_5bk_0_0.vlm"},
	{0x30E06F60792C164CULL, "vxdr_primevalentrance_178_0_0.vlm"},
	{0x30E083F7D448A7C8ULL, "6139663_m.fxe"},
	{0x30E0FE66AC134D09ULL, "qun210ar_fq_random_enc.lst"},
	{0x30E17CC0819E2E59ULL, "v_cage_03.dds"},
	{0x30E1CAA8CC4118FEULL, "225924491.ogg"},
	{0x30E270887548455EULL, "6041859_f.fxe"},
	{0x30E2B44A60B6F44EULL, "6018894_m.fxe"},
	{0x30E2B4BBF66F8270ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_02deadtree.phy"},
	{0x30E2F0F65C4ABDD8ULL, "mrl150_merrill6094015.cl"},
	{0x30E323C4A9830D50ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.cess.evt"},
	{0x30E35E49CC5156D5ULL, "6117275_m.fxe"},
	{0x30E378240015D3AEULL, "vxlt_backalley_n_1wg_0_0.vlm"},
	{0x30E46BA9ABCC8422ULL, "vxlt_gallowstemplar_f_5xj_0_0.vlm"},
	{0x30E470350E69C889ULL, "1058502363.ogg"},
	{0x30E51595491247D1ULL, "one000_ilen6205700.cl"},
	{0x30E536166526A3E1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_winiron.phy"},
	{0x30E57EC16DCD9BE2ULL, "6039728_f.fxe"},
	{0x30E583AB43525A15ULL, "vxw_encounter2_camp_ev_0_0.vlm"},
	{0x30E5C904DFB6ED22ULL, "691139401.ogg"},
	{0x30E6787F6728E82FULL, "vxlt_gallowstemplar_3e1_0_0.vlm"},
	{0x30E67B8A4E07BDB7ULL, "vxw_encounter2_bm_0_0.vlm"},
	{0x30E6B10B1601550BULL, "10511780.ogg"},
	{0x30E7D666AC192840ULL, "qun210ar_fq_random_enc.ncs"},
	{0x30E7F89C69A62B2AULL, "vxca_caverns1_gjf_0_0.vlm"},
	{0x30E85489157DBE81ULL, "vxht_0_hightown_c25_0_0.vlm"},
	{0x30E93A349746FD0EULL, "hlac_dr_slavers_3214_11z.dds"},
	{0x30E9454E84D5F8B3ULL, "6040559_m.fxe"},
	{0x30E96A08508CD496ULL, "6095353_f.fxe"},
	{0x30E9B953244E11CEULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc02.mmh"},
	{0x30E9BC520B3493E1ULL, "dae150_chandan6148552.cl"},
	{0x30E9D353244E3DBCULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc02.msh"},
	{0x30E9EC3FAF07244EULL, "vxca_caverns1_mines_j1z_0_1.vlm"},
	{0x30EA0B0EFB42AF9DULL, "6207932_m.fxe"},
	{0x30EA42A2C0575377ULL, "6125856_m.fxe"},
	{0x30EB203654CB6BB9ULL, "793876270.ogg"},
	{0x30EB5744DA0C983EULL, "pet_spit.mao"},
	{0x30EB704E0E9AEFCDULL, "718548020.ogg"},
	{0x30EBAEA41D0722DEULL, "gen00fl_isabela6088352.cl"},
	{0x30EBDD45A6096C38ULL, "6031394_m.fxe"},
	{0x30EC9CF64D7F0520ULL, "dae011_badpoet_sta6147007.cl"},
	{0x30EDF954737BF394ULL, "6077420_m.fxe"},
	{0x30EE0292307EA328ULL, "follower_banter6160761.cl"},
	{0x30EE27B7467F8EFBULL, "vxlt_warehouse_n_v3_0_0.vlm"},
	{0x30EE507B6A1712D0ULL, "gen00fl_varric6076540.cl"},
	{0x30EE6FDA36489837ULL, "gen00fl_merrill6128703.cl"},
	{0x30EEC9DE803E8304ULL, "vxlt_hangedman_2y2_0_0.vlm"},
	{0x30EF795483C6388DULL, "vxw_woundedcoast_hl_0_0.vlm"},
	{0x30EF8293089E269CULL, "art\\characters\\playercharacter\\humanmale\\arm_roba.mao"},
	{0x30EF93F71EF954E8ULL, "art\\characters\\weapons\\axes\\w_axe_ax103a.phy"},
	{0x30EFB5C7C47E4C21ULL, "em_tamlen.mor"},
	{0x30EFC2C0B809C5F4ULL, "vxlt_0_lowtown_h5j_0_0.vlm"},
	{0x30F01800848F3831ULL, "plt_prm_cod_ashes_amulet.plo"},
	{0x30F0220C6D259DF3ULL, "6174487_m.fxe"},
	{0x30F0CDF5AB2981AEULL, "552880482.ogg"},
	{0x30F0E5F30DA461FEULL, "mag360_alain6063779.cl"},
	{0x30F1059240FDC1AAULL, "vxw_blightlands_122_0_0.vlm"},
	{0x30F1119259088F39ULL, "fx_fireringfb_a.dds"},
	{0x30F11A4D9AC1B3C2ULL, "vxlt_gallowscourtyard_73h_0_0.vlm"},
	{0x30F13585D400C568ULL, "art\\vfx\\combat\\vfx_glb_ele_f_additive_c.anb"},
	{0x30F170DEF140DAFAULL, "vxlt_gallowstemplar_f_45l_0_0.vlm"},
	{0x30F1CA9876F96564ULL, "142435612.ogg"},
	{0x30F1D026BE476F55ULL, "6112175_m.fxe"},
	{0x30F1F766AC21B4AEULL, "qun210ar_fq_random_enc.are"},
	{0x30F21FE12D41EEBDULL, "304436159.ogg"},
	{0x30F23C3E5566490BULL, "hlac_lt_0_lowtown_65535_0x.dds"},
	{0x30F2DD869DC29CA2ULL, "vxlt_gallowstemplar_f_5ck_0_0.vlm"},
	{0x30F3CD23435C5420ULL, "70434596.ogg"},
	{0x30F3DFE1AF32344AULL, "gen00fl_fenris6110826.cl"},
	{0x30F45D1CB2AF2A2DULL, "gen00fl_varric6093656.cl"},
	{0x30F461548C4B9365ULL, "llac_dr_slavers_3216_1.dds"},
	{0x30F4A3F8B8D8E7AEULL, "vxw_blightlands_140_0_0.vlm"},
	{0x30F4E4FB25CFE612ULL, "ch1alph.psh.fxo"},
	{0x30F4ED673FFDD637ULL, "vxw_encounter2_ml_0_0.vlm"},
	{0x30F5058B5526756DULL, "gen_im_trs_tmp_junk2_03.uti"},
	{0x30F50DD5BC900270ULL, "6201661_m.fxe"},
	{0x30F53AA9FBAA3013ULL, "mcr310_sibling6096584.cl"},
	{0x30F597F89CB527D5ULL, "794176919.ogg"},
	{0x30F62934FDE290FFULL, "gen00fl_bethany6101531.cl"},
	{0x30F651EB05AF687BULL, "llac_ca_caverns2_5802_66.dds"},
	{0x30F66139FF8DF2CDULL, "dae000cr_grand_cleric6147646.cl"},
	{0x30F67E4663454C80ULL, "llac_lt_backalley_n_0_14.dds"},
	{0x30F68CF1EBCE871CULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scl_c.mao"},
	{0x30F6A1A8F0B092A6ULL, "211211271.ogg"},
	{0x30F6CCC86875091EULL, "575039898.ogg"},
	{0x30F6CF711072B4D4ULL, "348173011.ogg"},
	{0x30F873B5B2E1FC7CULL, "571025237.ogg"},
	{0x30F8C3879811ED0BULL, "196810183.ogg"},
	{0x30F8D2E8A00C1408ULL, "lgt150_nat_were_supporter6110972.cl"},
	{0x30FA0DCB64D9F680ULL, "vxlt_0_lowtown_n_e1t_0_0.vlm"},
	{0x30FA10ECCF50618CULL, "vxlt_gallowscourtyard_n_4m4_0_0.vlm"},
	{0x30FA33AF75B88B82ULL, "vxlt_gallowstemplar_3sv_0_0.vlm"},
	{0x30FA4010A15BA5BCULL, "vxlt_0_lowtown_docks_a3_n_cwv_0_0.vlm"},
	{0x30FB8EF125E83219ULL, "719308306.ogg"},
	{0x30FC292E5730DF47ULL, "hlac_ht_mansion2_818_30z.dds"},
	{0x30FC7BCFC44E3DD4ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_8x.dds"},
	{0x30FCAA311E6C2B47ULL, "6110957_m.fxe"},
	{0x30FD294A5F42AEA5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_lgta_0d.dds"},
	{0x30FD8168B84CDBD2ULL, "238381003.ogg"},
	{0x30FD828F703CE1E8ULL, "6005821_m.fxe"},
	{0x30FD8ADFF4A645C0ULL, "6153388_m.fxe"},
	{0x30FD963A4E1C53FDULL, "960103545.ogg"},
	{0x30FD99B99696B19AULL, "vxlt_blackemporium_135_0_0.vlm"},
	{0x30FDD5C7ECA99ED5ULL, "269388557.ogg"},
	{0x30FDEB3175BC510DULL, "vxw_woundedcoast_ek_0_0.vlm"},
	{0x30FE41733228D189ULL, "312338041.ogg"},
	{0x30FE8BD2FDE0CA9AULL, "t3_arm_sla.tnt"},
	{0x30FEB4125696C71FULL, "811249453.ogg"},
	{0x30FF0D10476BC9DBULL, "qun131_taskmaster6039549.cl"},
	{0x30FF2C0A5808229EULL, "6110549_m.fxe"},
	{0x30FF8C2A8CB7CC42ULL, "vxlt_gallowscourtyard_74i_0_0.vlm"},
	{0x30FFA2E208AB5E88ULL, "vxlt_0_lowtown_ekl_0_1.vlm"},
	{0x30FFD4AF56C0F17EULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_player.ani"},
	{0x3100104D78431033ULL, "zz_lineup.cut"},
	{0x310022CF1D540AB4ULL, "rob_pl3.tnt"},
	{0x310077078C8DF459ULL, "zz_approval_debug6003684.cl"},
	{0x3100D568CF1B923CULL, "art\\characters\\weapons\\axes\\w_axe_ax06a.mao"},
	{0x3100E168CF1BA6DFULL, "art\\characters\\weapons\\axes\\w_axe_ax06a.mmh"},
	{0x3100E768CF1BB081ULL, "art\\characters\\weapons\\axes\\w_axe_ax06a.msh"},
	{0x3100F8849A484F83ULL, "gen00fl_anders6123645.cl"},
	{0x3101562E65EAA9C5ULL, "vxw_encounter3_fq_0_1.vlm"},
	{0x3101C62B32260024ULL, "gen_im_arm_cht_hvy_lore_01.uti"},
	{0x3101EA6238DCF2EEULL, "6103289_f.fxe"},
	{0x31026F742F59E5D9ULL, "vxht_mansion2b_1u6_0_0.vlm"},
	{0x31029D6FF7F37FCDULL, "zz_followers_banter6159360.cl"},
	{0x310303A89AF60152ULL, "hlac_lt_backalley_g_1_4y.dds"},
	{0x3103075EACAA7D39ULL, "vxlt_gallowstemplar_5rp_0_0.vlm"},
	{0x3104F98B0505F939ULL, "ts_gems_01.gda"},
	{0x3105921AFA903132ULL, "771422885.ogg"},
	{0x3105A3D9F3FDF542ULL, "ht_keep_n_3740.rml"},
	{0x3105A430DE95B6CEULL, "vxht_0_hightown_a2_a7o_0_0.vlm"},
	{0x3105CBA5E2D43ECFULL, "267937157.ogg"},
	{0x3105DE81C494DE10ULL, "6069706_m.fxe"},
	{0x3105DEEDEBE7BF63ULL, "267240532.ogg"},
	{0x31060B0B749005F0ULL, "vxca_cave_22y_0_0.vlm"},
	{0x31063029B3D203DCULL, "hlac_ht_chantry_2534_5x.dds"},
	{0x3106628977A2ED11ULL, "vxw_bonepitdestroyed_a3_1pr_0_0.vlm"},
	{0x31067F9D273D8008ULL, "vxlt_gallowstemplar_f_6t7_0_0.vlm"},
	{0x3107CBD2E280CE11ULL, "vxca_cavern1_haunted_i8x_0_0.vlm"},
	{0x3107D51A1A0B0DBFULL, "6096358_m.fxe"},
	{0x31080E423F810769ULL, "6146901_m.fxe"},
	{0x31081A8B0DD27BB8ULL, "art\\vfx\\creatures\\fxc_withfn_c.mmh"},
	{0x310882C9C1467A0CULL, "vxlt_gallowscourtyard_n_4v5_0_0.vlm"},
	{0x310954F2B204FE0CULL, "6205636_m.fxe"},
	{0x31096A1B1A82C554ULL, "332671550.ogg"},
	{0x310988D10A7DA6DEULL, "vxdr_0_deeproads_bwo_0_0.vlm"},
	{0x3109D6CD5B228757ULL, "mcr371_arthuris6087099.cl"},
	{0x310A04487774D91BULL, "40432035.ogg"},
	{0x310A0D4F2123AC51ULL, "hlac_ca_caverns2_5802_16y.dds"},
	{0x310A3524844E9D83ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small.ani"},
	{0x310A7C0B7ACFEFBBULL, "lt_0_lowtown_docks_a3_149_0.msh"},
	{0x310A8E0B7AD00E06ULL, "lt_0_lowtown_docks_a3_149_0.mao"},
	{0x310A9A0B7AD022A5ULL, "lt_0_lowtown_docks_a3_149_0.mmh"},
	{0x310AFC8F632FEFEBULL, "lt_gallowstemplar_2391.anb"},
	{0x310B040CEAE95F89ULL, "gen00fl_anders6112962.cl"},
	{0x310B0D8A819493FCULL, "6126384_m.fxe"},
	{0x310B5C1B3ED49937ULL, "216767240.ogg"},
	{0x310B789AC43C322EULL, "597325397.ogg"},
	{0x310BADF814C0CA66ULL, "zz_vault_debug6177619.cl"},
	{0x310C0B96888CCA1AULL, "vxlt_blackemporium_166_0_0.vlm"},
	{0x310C5D0E67B2698DULL, "vxw_woundedcoast_bj_0_0.vlm"},
	{0x310C8478BBD407E2ULL, "6119893_f.fxe"},
	{0x310D1A4265724484ULL, "vxdr_primevalentrance_26h_0_0.vlm"},
	{0x310D24A7128AD30BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkc_d.dds"},
	{0x310D386F31D00C9AULL, "201954494.ogg"},
	{0x310D9A5109ADC42CULL, "vxlt_darktown_5ps_0_0.vlm"},
	{0x310DBEC5C93DE11DULL, "vxht_0_hightown_a3_al9_0_1.vlm"},
	{0x310EA7771858714FULL, "6152739_m.fxe"},
	{0x310F0F727B3B730EULL, "279725413.ogg"},
	{0x310F7047FA82BAD3ULL, "vxht_0_hightown_b7m_0_0.vlm"},
	{0x310FC14081AECDA2ULL, "vxlt_gallowstemplar_f_5iu_0_0.vlm"},
	{0x310FE964EAF8E122ULL, "738622323.ogg"},
	{0x3110C4504F2FE6CCULL, "6083496_m.fxe"},
	{0x3110C63B6FDAC946ULL, "follower_banter6136687.cl"},
	{0x311119D37573EDD8ULL, "lgt210pt_search_rescue.ncs"},
	{0x31118F4AF444095BULL, "hlac_lt_backalley_n_65537_0z.dds"},
	{0x3111C70C698D70B9ULL, "llac_lt_0_lowtown_-3_4.dds"},
	{0x3111CE163D0B31AFULL, "vxlt_gallowstemplar_3t2_0_0.vlm"},
	{0x3112230967379241ULL, "44610365.ogg"},
	{0x31127ECF32859967ULL, "926120722.ogg"},
	{0x3112F5CF05E4BBE1ULL, "489740931.ogg"},
	{0x311310600A47AE5EULL, "art\\cinematics\\deathblows\\rockgolem\\cs_db_pc_rubblemage.ani"},
	{0x31136A6B73ADF04CULL, "6103392_m.fxe"},
	{0x31139EDB7B392E96ULL, "vxht_chantry_339_0_0.vlm"},
	{0x3113A6DD18B87978ULL, "vxlt_darktown_6ic_0_0.vlm"},
	{0x3113C8CCE74B1DFCULL, "dae000_gamlen6149472.cl"},
	{0x3113D049847226B9ULL, "624946989.ogg"},
	{0x3113E62BDE76B5E9ULL, "699184420.ogg"},
	{0x3114128FA22992CDULL, "hlac_lt_0_lowtown_n_-196608_4z.dds"},
	{0x31141A055FACD9B9ULL, "gen00fl_sebastian6174707.cl"},
	{0x31141D2D93C31F25ULL, "llac_lt_undercityserialkiller_-1_50.dds"},
	{0x3114D46669990591ULL, "vxw_bonepitdestroyed_a3_1uq_0_0.vlm"},
	{0x3114EB9BEA7AE268ULL, "llac_lt_warehouse_1_4.dds"},
	{0x3114EDAF56D377F8ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_player.gad"},
	{0x3115BBDC045E59E7ULL, "vxdr_primevaltomb_8re_0_0.vlm"},
	{0x3115EB3DD1E9BEE9ULL, "vxlt_gallowscourtyard_71e_0_0.vlm"},
	{0x31174D248459517AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small.evt"},
	{0x3117C8E41BB8C56EULL, "ave331_jeven6137720.cl"},
	{0x311825E8E18835DAULL, "vxht_0_hightown_n_apm_0_0.vlm"},
	{0x3118421625B6171DULL, "6072289_f.fxe"},
	{0x31186C02F0FB18E9ULL, "dae300_cullen6135878.cl"},
	{0x311889D7BDF146BAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_galloalcovlo.msh"},
	{0x31188E9BD104EFA6ULL, "ht_0_hightown_34_0_lc.mmh"},
	{0x311897D7BDF15EF3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_galloalcovlo.mao"},
	{0x31189BD7BDF16500ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_galloalcovlo.mmh"},
	{0x3118C9E078F7D38FULL, "art\\vfx\\combat\\vfx_gib_qunc_lbrt.anb"},
	{0x31191E355262682BULL, "zz_pgoto.ncs"},
	{0x311951556CBA4B75ULL, "vxht_mansion_n_3es_0_0.vlm"},
	{0x3119646E253C2BE1ULL, "hightown_male_cityguard_mob6170058.cl"},
	{0x31197C87C44226CDULL, "hlac_w_encounter3_-3_0z.dds"},
	{0x31199AEA849FB9CCULL, "6091855_m.fxe"},
	{0x3119C250E89F769EULL, "vxdr_primevalentrance_19t_0_0.vlm"},
	{0x3119E2083FBFB1B3ULL, "6087218_f.fxe"},
	{0x311A255EDB7C2FC3ULL, "199732897.ogg"},
	{0x311A3485D423C803ULL, "art\\vfx\\combat\\vfx_glb_ele_f_additive_c.mmh"},
	{0x311A847709EB2316ULL, "vxht_chantry_2tp_0_1.vlm"},
	{0x311AB3DF1905059DULL, "vxca_caverns1_e8m_0_0.vlm"},
	{0x311ADC39C4EC6BECULL, "6147657_m.fxe"},
	{0x311AE8B35C362F7CULL, "6172598_m.fxe"},
	{0x311B146AA7B09A49ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_1z.dds"},
	{0x311B5CA237456D44ULL, "6043276_m.fxe"},
	{0x311B642FB5FAE678ULL, "6129125_m.fxe"},
	{0x311B7763E17DE054ULL, "vxca_caverns2_80c_0_1.vlm"},
	{0x311BCF03AE8DBD50ULL, "seb221_ruxton_harimann6197718.cl"},
	{0x311C0A0AE641F2E1ULL, "6152764_m.fxe"},
	{0x311C96F3825F7CC9ULL, "vxht_keep_n_30l_0_0.vlm"},
	{0x311D2D328310AA09ULL, "vxht_0_hightown_a3_c95_0_0.vlm"},
	{0x311DB6E10B40905CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat1a.mao"},
	{0x311DC2E10B40A4FFULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat1a.mmh"},
	{0x311DC8E10B40AEA1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat1a.msh"},
	{0x311DE224EC78D353ULL, "vxht_0_hightown_b0l_0_0.vlm"},
	{0x311E331D73A4E622ULL, "vxlt_gallowstemplar_f_5lv_0_0.vlm"},
	{0x311E4C24845F6E6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small.gad"},
	{0x311E592201A9956BULL, "hlac_lt_0_lowtown_n_65536_2z.dds"},
	{0x311EA08CF33853C2ULL, "886509563.ogg"},
	{0x311F8F5233FBAD36ULL, "vxlt_0_lowtown_a2_hdo_0_0.vlm"},
	{0x311FC6F5A8CCFCF5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edge_1.mao"},
	{0x311FD31874B58962ULL, "6119137_m.fxe"},
	{0x311FEC5F185BB7E1ULL, "one130_thug6131599.cl"},
	{0x312002C92F08B7CFULL, "299124279.ogg"},
	{0x312030628E5B76E3ULL, "vxlt_0_lowtown_a2_n_ewq_0_0.vlm"},
	{0x3120D92EB0886AF4ULL, "vxca_cavern1_haunted_efw_0_0.vlm"},
	{0x3120DC04D33AB5BEULL, "vxlt_backalley_g_1e4_0_0.vlm"},
	{0x312177E53A726B5AULL, "qun221_petrice6077949.cl"},
	{0x3121AB114E7AC37EULL, "ht_mansion_n_3364_ll.mmh"},
	{0x3121B6E3FD759459ULL, "944897729.ogg"},
	{0x31228A299D51640CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv2.ani"},
	{0x3122A1F979E39CFAULL, "zz_lgt_debug6109143.cl"},
	{0x3122F1F3EDB9F1C3ULL, "pro000_varric_cass6208382.cl"},
	{0x31232DB54D6230DDULL, "6102659_m.fxe"},
	{0x312335EBFA81B6EAULL, "vxw_encounter1_be_0_0.vlm"},
	{0x312387DD897B04E0ULL, "lgt150_nat_were_supporter.cnv"},
	{0x3123D999D1F64A13ULL, "vxht_0_hightown_n_c1a_0_0.vlm"},
	{0x31252BAA996354E9ULL, "1073284695.ogg"},
	{0x312555C857339F11ULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_c.anb"},
	{0x3125A21E6E9F6AF6ULL, "hlac_lt_0_lowtown_a3_n_c_-1_2y.dds"},
	{0x31260F616BF757E5ULL, "vxht_0_hightown_n_aow_0_0.vlm"},
	{0x31260F6EDFF1A022ULL, "atl_guiachieve_dxt5_01.dds"},
	{0x31261D898CC2136CULL, "w_bonepitmining_a2_9_0.rml"},
	{0x312685E1010D0A61ULL, "321117231.ogg"},
	{0x31269A1B96AD1C6BULL, "vxlt_0_lowtown_ham_0_0.vlm"},
	{0x31269FF66C4591FEULL, "600738.ogg"},
	{0x3126F1D601C73070ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_2y.dds"},
	{0x3126F7395753C918ULL, "30397921.ogg"},
	{0x312703FAB01BA676ULL, "w_sundermount_90_0.msh"},
	{0x31270DFAB01BB774ULL, "w_sundermount_90_0.mmh"},
	{0x312711FAB01BBD87ULL, "w_sundermount_90_0.mao"},
	{0x312729600A5881D8ULL, "art\\cinematics\\deathblows\\rockgolem\\cs_db_pc_rubblemage.gad"},
	{0x312769DD897EBA18ULL, "lgt150_nat_were_supporter.bnk"},
	{0x3127C0D8F4BE2122ULL, "water0nospec.psh.fxo"},
	{0x3127CA9489041183ULL, "framebuffereffects_horizontalblur_p0.vsh.fxo.dx11"},
	{0x3127DFA189962A98ULL, "llac_ca_caverns2prc_5802_2.dds"},
	{0x31283C127ABB834DULL, "6123236_m.fxe"},
	{0x31284B54C7511BAEULL, "vxca_caverns1_dq8_0_0.vlm"},
	{0x312880465008302AULL, "vxlt_0_lowtown_docks_a3_d1j_0_0.vlm"},
	{0x312897B31CD00177ULL, "283830526.ogg"},
	{0x3128A3FF07047599ULL, "6085381_m.fxe"},
	{0x3128BC7E1E426E57ULL, "hlac_w_sundermount_-131077_0z.dds"},
	{0x3128BD4713850D0FULL, "vxca_caverns2_6t7_0_0.vlm"},
	{0x312940C84B9E9A8DULL, "vxw_woundedcoast_th_0_0.vlm"},
	{0x312944C8CC3B0058ULL, "qun111_ketojan6040978.cl"},
	{0x31295E60F2CAFE3EULL, "6121256_f.fxe"},
	{0x312982F91D4A44C1ULL, "444280212.ogg"},
	{0x312997299D57994EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv2.gad"},
	{0x3129B1463512CB2AULL, "414135785.ogg"},
	{0x3129CAB292A95352ULL, "6125787_m.fxe"},
	{0x3129CFCE4488A6B6ULL, "healplayer.ncs"},
	{0x312A6F97EA3F4CEBULL, "gen00fl_aveline6043104.cl"},
	{0x312A7877F92016FDULL, "zz_followers_banter6158730.cl"},
	{0x312A8E7B3CB07239ULL, "313965978.ogg"},
	{0x312AAD2FFA409C03ULL, "dae000_arishok6070822.cl"},
	{0x312B241D1A0BB41CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_forward.ani"},
	{0x312B571D62E0E517ULL, "lgt000_chanter.bnk"},
	{0x312B9F0F7506C289ULL, "vxht_0_hightown_a3_c24_0_0.vlm"},
	{0x312BF289A9166D84ULL, "mag210_arianni6057319.cl"},
	{0x312D5E3D5355AFFDULL, "512678527.ogg"},
	{0x312E21A8BB9B1A94ULL, "6136657_m.fxe"},
	{0x312E26C064C877E5ULL, "vxht_keep_n_59y_0_0.vlm"},
	{0x312E9499EF5FE30EULL, "vxca_caverns2prc_9wq_0_0.vlm"},
	{0x312ECE469C6A515AULL, "one300_zevran6075743.cl"},
	{0x312F391D62E49B2FULL, "lgt000_chanter.cnv"},
	{0x312F6F49005F7CA3ULL, "86544167.ogg"},
	{0x3130064DE4533772ULL, "hlac_lt_0_lowtown_a2_n_-2_4y.dds"},
	{0x313092299D5DAF2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv2.evt"},
	{0x31312A36EFE728F5ULL, "vxlt_hangedman_n_1qo_0_0.vlm"},
	{0x31318B6DE77C93D2ULL, "hm_carver_h1s3.mor"},
	{0x3131A7C8EC77CF6AULL, "vxw_encounter1_gf_0_0.vlm"},
	{0x3131B0DFB6FF2A75ULL, "vxca_caverns2prc_521_0_0.vlm"},
	{0x3131B8204D853691ULL, "vxw_bonepitdestroyed_a3_1cw_0_0.vlm"},
	{0x3131EF8CA44F2A18ULL, "vxw_bonepitmining_a2_25a_0_0.vlm"},
	{0x31322BF86250F2DAULL, "vxdr_primevalentrance_2km_0_0.vlm"},
	{0x3132311D1A11E95EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_forward.gad"},
	{0x313276C8573E618AULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_c.mmh"},
	{0x3132ADC84A98CE0BULL, "15455247.ogg"},
	{0x31330865DBF65922ULL, "vrc321ip_spooky_5_owner.utp"},
	{0x31334DB3AB507390ULL, "vxht_mansion2_2o2_0_0.vlm"},
	{0x313353A0B2201668ULL, "396667656.ogg"},
	{0x31337B68FC437294ULL, "6089645_m.fxe"},
	{0x313389A5FC766C96ULL, "209059375.ogg"},
	{0x3133A553478DAFEFULL, "6152527_f.fxe"},
	{0x313465F1ED0B916EULL, "hlac_w_encounter3_-65536_0y.dds"},
	{0x3134AB17AAFBD80CULL, "128918966.ogg"},
	{0x3134BAF4F3F1F4E1ULL, "6083714_m.fxe"},
	{0x3134DA7C97A315E8ULL, "vxlt_warehouse_n_112_0_0.vlm"},
	{0x3135C10D5EE276F6ULL, "vxw_bonepitmining_a2_ku_0_0.vlm"},
	{0x3135CA3E17FE8BCFULL, "6086849_m.fxe"},
	{0x3135D0A2BEFDAE65ULL, "vxlt_0_lowtown_docks_coc_0_0.vlm"},
	{0x3135DFF9CC53D2E5ULL, "6092126_f.fxe"},
	{0x3136D229F4B71134ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_enter.ani"},
	{0x3136F98F63552537ULL, "lt_gallowstemplar_2391.rml"},
	{0x31375FEA2B264906ULL, "6066899_m.fxe"},
	{0x3137772D411DA0FEULL, "6156536_m.fxe"},
	{0x3137895213E7533DULL, "311682813.ogg"},
	{0x3137B2A53D94B30DULL, "vxw_woundedcoast_qg_0_0.vlm"},
	{0x31383F89428114D4ULL, "900150865.ogg"},
	{0x31389C8AAD074908ULL, "6089667_m.fxe"},
	{0x31390C1D1A17C97DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_forward.evt"},
	{0x31394090E8FEE334ULL, "6180817_f.fxe"},
	{0x31396BAFA195D9A4ULL, "6066348_f.fxe"},
	{0x31398B7BE2C03C16ULL, "vxlt_0_lowtown_docks_a3_n_9k6_0_0.vlm"},
	{0x3139903438F25B73ULL, "vxca_cavern1_haunted_j5n_0_0.vlm"},
	{0x3139CDA6FEB4C295ULL, "character_character_p217.psh.fxo.dx11"},
	{0x313A75503429945DULL, "vxdr_primevaltomb_77k_0_0.vlm"},
	{0x313A9A6F1307DDB2ULL, "6111274_m.fxe"},
	{0x313B1C9920F9EB3CULL, "6075318_m.fxe"},
	{0x313B805892EDF7EFULL, "character_character_p268.psh.fxo.dx11"},
	{0x313B91B2B21480F6ULL, "follower_banter6163634.cl"},
	{0x313C251B571A6A21ULL, "445418782.ogg"},
	{0x313C58817B0C9008ULL, "6112974_m.fxe"},
	{0x313CB4B21A5CF382ULL, "vxca_caverns2_9a2_0_0.vlm"},
	{0x313CB6D258A1BA62ULL, "6113737_m.fxe"},
	{0x313CE518A622D874ULL, "vxht_chantry_n_2ag_0_0.vlm"},
	{0x313D0611D88B3733ULL, "6078106_f.fxe"},
	{0x313DCF29F4BD2A96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_enter.gad"},
	{0x313DF9175EE7513FULL, "263107653.ogg"},
	{0x313F55C4B2762E7FULL, "vxlt_gallowstemplar_3dr_0_0.vlm"},
	{0x313F91AB2BE4B0C1ULL, "6161010_m.fxe"},
	{0x313FA0D9F42F683EULL, "ht_keep_n_3740.anb"},
	{0x313FF11AA0E13982ULL, "vxht_0_hightown_a2_arl_0_0.vlm"},
	{0x3141E02D7991C8AAULL, "477966564.ogg"},
	{0x31421A0F9CD1F8F2ULL, "gui_guipixelshader_p2.psh.fxo.dx11"},
	{0x31423CFD86AF7E54ULL, "vxdr_primevaltomb_5a6_0_0.vlm"},
	{0x3142658393B034CAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_enter.gad"},
	{0x3142B4610C96DB51ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverenter.ani"},
	{0x3142BD82F4CFE893ULL, "6120855_m.fxe"},
	{0x3142D6CF6035E997ULL, "vxht_bartrand_n_5nu_0_0.vlm"},
	{0x31431ED65E6DB952ULL, "380320918.ogg"},
	{0x31434174C8AD780BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed.ani"},
	{0x31434C5989992E68ULL, "vxlt_warehouse_n_163_0_0.vlm"},
	{0x314391D6E7A35F2BULL, "6132856_m.fxe"},
	{0x314392B86FB1DF36ULL, "vxht_keep_n_57p_0_0.vlm"},
	{0x31439344063A05CFULL, "528189350.ogg"},
	{0x3143F1FC2904D6BAULL, "6134925_m.fxe"},
	{0x3143F8E92D3E590BULL, "gen00fl_anders6119460.cl"},
	{0x31441A4FF2D7A53FULL, "llac_lt_hangedman_n_-1_16.dds"},
	{0x314459673F599C8DULL, "dre000_carver6037589.cl"},
	{0x3144CAB1710A0AFAULL, "properties.gda"},
	{0x3144EA29F4C377F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_enter.evt"},
	{0x31450050E5716BE4ULL, "212554368.ogg"},
	{0x31457226997FE9CAULL, "mcr370_cut_epilogue.cnv"},
	{0x31461A42C726569EULL, "6045799_m.fxe"},
	{0x3146235485AB6EA1ULL, "follower_banter6168418.cl"},
	{0x3146A4E6FDAE06A9ULL, "vxw_bonepitdestroyed_a3_267_0_0.vlm"},
	{0x3146DDDF11DF1927ULL, "403649777.ogg"},
	{0x31474D670E85E7E3ULL, "vxca_caverns1_mines_i1y_0_0.vlm"},
	{0x31478A8E1FB9D610ULL, "vxw_bonepitdestroyed_a3_17f_0_0.vlm"},
	{0x314795FC92CA6EC2ULL, "hlac_w_encounter1_65535_0y.dds"},
	{0x3147D131C3D19027ULL, "dr_slavers.arl"},
	{0x3147D34BCAA909E7ULL, "da2_prc_one_de-de.cre"},
	{0x31483B16657B291DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_winiron.mao"},
	{0x31483E4469568452ULL, "vxht_brothel_n_4c2_0_0.vlm"},
	{0x31483F16657B3056ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_winiron.mmh"},
	{0x31484916657B4154ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_winiron.msh"},
	{0x31484F1497732A6EULL, "6080733_f.fxe"},
	{0x314854269981ECF2ULL, "mcr370_cut_epilogue.bnk"},
	{0x3148868880AAFADBULL, "239680472.ogg"},
	{0x3148BF610C9B59F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverenter.gad"},
	{0x31491D683A8A44AEULL, "ach_kill_upgrade.dds"},
	{0x31493F58090690CCULL, "163835202.ogg"},
	{0x3149488393B62261ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_enter.evt"},
	{0x31495630A33CE7EBULL, "6103351_m.fxe"},
	{0x31498CBB6B52CD6BULL, "430856492.ogg"},
	{0x314A11119CD4FB0EULL, "art\\characters\\creatures\\highdragon\\cn_hed_fdr_1.mao"},
	{0x314A2ABB28478E94ULL, "gen00fl_sebastian6195860.cl"},
	{0x314A954F11B186EBULL, "6086341_f.fxe"},
	{0x314B5D818F04D28BULL, "480107881.ogg"},
	{0x314BD2DEBA644C7DULL, "vxlt_warehouse_10k_0_0.vlm"},
	{0x314BDE7E7ED35863ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_critical_shot.evt"},
	{0x314BE04A8237F4BAULL, "6107054_f.fxe"},
	{0x314BE17816999498ULL, "6115588_m.fxe"},
	{0x314BE51B413ED14EULL, "895296658.ogg"},
	{0x314C9F679A536F01ULL, "lt_0_lowtown_a2_n_c_42_0.rml"},
	{0x314CA72ECC2102A9ULL, "vxlt_0_lowtown_a2_f55_0_0.vlm"},
	{0x314D22F2C7DF98B8ULL, "dae000_gamlen6048432.cl"},
	{0x314D2F6B99F7FE53ULL, "vxht_brothel_n_4h6_0_0.vlm"},
	{0x314D2FE48E6EF1C4ULL, "vxw_encounter3_b9_0_0.vlm"},
	{0x314D32D1D66869D1ULL, "236415019.ogg"},
	{0x314D34DC88BF08B4ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l90_rf.ani"},
	{0x314D9CF0A5895724ULL, "hlac_ca_cavern1_haunted_24424_7y.dds"},
	{0x314E1735914C808CULL, "476052691.ogg"},
	{0x314E2C9F1CEE0B81ULL, "6166954_m.fxe"},
	{0x314F30833E31C624ULL, "6093606_m.fxe"},
	{0x314F4C9BB9DD4AB2ULL, "6121026_m.fxe"},
	{0x314FBC610CA17358ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverenter.evt"},
	{0x314FD833FD86676DULL, "follower_banter6169332.cl"},
	{0x315028174ECDA2FEULL, "6087506_m.fxe"},
	{0x3150403208EC1397ULL, "15222416.ogg"},
	{0x3150C6D6AC759C0EULL, "hlac_dr_primevaltomb_9621_6x.dds"},
	{0x3150C86E1475A312ULL, "vxdr_primevaltomb_7zu_0_2.vlm"},
	{0x31512974C8B98C82ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed.evt"},
	{0x3151730B7B0BE352ULL, "lt_0_lowtown_docks_a3_149_0.phy"},
	{0x3151D6FCA085EF68ULL, "94752692.ogg"},
	{0x3152B77E7ED934E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_critical_shot.gad"},
	{0x3152EC37BC81F47EULL, "hlac_w_bonepitmining_a2_458755_0z.dds"},
	{0x31537A157BD5413AULL, "6040661_f.fxe"},
	{0x3153C0D91D7C41E1ULL, "hlac_lt_0_lowtown_131073_2y.dds"},
	{0x3153CBE479122CECULL, "369549268.ogg"},
	{0x315431DC88C52216ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l90_rf.gad"},
	{0x315433E2D8D188F1ULL, "6113918_m.fxe"},
	{0x31543EA3F27EF769ULL, "fx_icecloud.mao"},
	{0x315476DE188BEF56ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright_1.mao"},
	{0x3154CFFB5E56C819ULL, "vxca_caverns1_mines_i26_0_0.vlm"},
	{0x31550AF37DCD3DC9ULL, "follower_banter6172985.cl"},
	{0x315516C3EFA41F29ULL, "vxw_bonepitdestroyed_a3_216_0_0.vlm"},
	{0x3155804490F37A08ULL, "6087467_m.fxe"},
	{0x31558B59088DB741ULL, "6089723_m.fxe"},
	{0x31558FABCE9116C2ULL, "plt_prm000pt_hindsight_belt.nss"},
	{0x3155ECB2A9BD1585ULL, "llac_ht_keep_3735_0.dds"},
	{0x3155FC6B11AFEE90ULL, "vxw_bonepitdestroyed_a3_18g_0_0.vlm"},
	{0x31569F4D693A4FADULL, "6068487_m.fxe"},
	{0x3156B0215B4C9CD2ULL, "vxht_brothel_n_4h3_0_0.vlm"},
	{0x3157408393C251B0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_enter.ani"},
	{0x31576B4255F5F7C7ULL, "vxht_bartrand_n_68x_0_0.vlm"},
	{0x3157CF17AD7E88F2ULL, "fen110_hunter.cnv"},
	{0x31582874C8BFA995ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed.gad"},
	{0x315875D5002D6FBCULL, "231915304.ogg"},
	{0x31588C6C4A1F025DULL, "726975139.ogg"},
	{0x31589F6989015A61ULL, "159539302.ogg"},
	{0x3158F02514E5C27FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masa_0d.dds"},
	{0x31592BCF53D3E4C6ULL, "ico_belt_2.dds"},
	{0x3159B5F18976B837ULL, "454043474.ogg"},
	{0x3159C67E7EDF6CC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_critical_shot.ani"},
	{0x315A0C1DD0F27C6CULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_bowl.phy"},
	{0x315A6D58D5E0BBDEULL, "gen_im_trs_krk_one_mac_r1.uti"},
	{0x315AAB98B34FAABBULL, "vxht_0_hightown_a3_at6_0_0.vlm"},
	{0x315AC21994CA7C04ULL, "zz_ser_debug6078161.cl"},
	{0x315AF117AD80F84AULL, "fen110_hunter.bnk"},
	{0x315B1BD08534290BULL, "hlac_ht_mansion2_818_7z.dds"},
	{0x315B869FFA9E6104ULL, "431120782.ogg"},
	{0x315B885307974EFCULL, "6015932_m.fxe"},
	{0x315C063F09F1A8E6ULL, "6026190_m.fxe"},
	{0x315C17C58E126455ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_galotopright.msh"},
	{0x315C19C58E1267CBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_galotopright.mmh"},
	{0x315C25C58E127C68ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_galotopright.mao"},
	{0x315CD88D09AADBE3ULL, "con_ico_bribe.dds"},
	{0x315CFDB52D66C5A1ULL, "llac_lt_0_lowtown_docks_a3_196606_0.dds"},
	{0x315D4B0535E11F67ULL, "lt_0_lowtown_n_46_0_ll.mmh"},
	{0x315D8808D9C47BC0ULL, "274102572.ogg"},
	{0x315E07668249236CULL, "vxlt_gallowscourtyard_n_6p5_0_0.vlm"},
	{0x315E6133F934AADAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_one_hand_loop.gad"},
	{0x315E96EE4B785D5EULL, "6108302_m.fxe"},
	{0x315EDED0D4A69645ULL, "blood3.mao"},
	{0x315EDFFDE1C3AB73ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.dura.evt"},
	{0x315F9467187802CAULL, "6181922_f.fxe"},
	{0x315FD6D533CFB9E5ULL, "vxht_0_hightown_n_ass_0_0.vlm"},
	{0x31604F40C9CA380EULL, "230203671.ogg"},
	{0x316056957AD32889ULL, "422722453.ogg"},
	{0x3160B3512FDDB36EULL, "fen210_hadriana6110401.cl"},
	{0x3161793D26836891ULL, "405275114.ogg"},
	{0x3161AE332B060998ULL, "6207780_m.fxe"},
	{0x3161EC6DE799FC9EULL, "vxht_0_hightown_n_av3_0_0.vlm"},
	{0x3162BF7FED312B79ULL, "hlac_w_bonepitmining_a2_0_0y.dds"},
	{0x3162D5A228AAD763ULL, "vxlt_0_lowtown_docks_n_be5_0_0.vlm"},
	{0x3162E7EF109E49B6ULL, "dae010cr_sailor_m.utc"},
	{0x3163184A879EB371ULL, "vxlt_gallowstemplar_5ck_0_0.vlm"},
	{0x316334C9AF8A8C33ULL, "dae000_pc_home_ambients6208683.cl"},
	{0x316336F4E95DEE9CULL, "6207813_f.fxe"},
	{0x31649A70EBE957BFULL, "vxlt_0_lowtown_n_ep1_0_1.vlm"},
	{0x31651D6AD247D482ULL, "vxca_caverns2_61u_0_0.vlm"},
	{0x31656433F93ACEF1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_one_hand_loop.evt"},
	{0x3165660D3065337EULL, "247087648.ogg"},
	{0x3165D8FDE1C9BDB4ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.dura.gad"},
	{0x3165E00EFE7307D4ULL, "lowtown_fem_prostitute_sta.cnv"},
	{0x31661D2672086063ULL, "dae000_tomwise6167347.cl"},
	{0x3166C67A9685D425ULL, "hlac_lt_0_lowtown_docks_a3_196607_1x.dds"},
	{0x316703DB1C71645AULL, "llac_w_sundermount_262142_0.dds"},
	{0x3167063DB7974134ULL, "1058504580.ogg"},
	{0x316710051B005745ULL, "vxht_mansion2_3bi_0_0.vlm"},
	{0x316754677D376A53ULL, "ser270_cut_motherreveal.bnk"},
	{0x31676F6A3A2091ECULL, "mag170_thrask6054059.cl"},
	{0x31677D4110BA6883ULL, "llac_ht_mansion2b_818_38.dds"},
	{0x31685B4C980E606DULL, "m_dae311ar_hanged_man.bnk"},
	{0x316875B9A8589E63ULL, "llac_lt_0_lowtown_docks_n_-131074_5.dds"},
	{0x3168C20EFE750A14ULL, "lowtown_fem_prostitute_sta.bnk"},
	{0x316913BBDCBA426EULL, "6103211_m.fxe"},
	{0x3169321854CBD2EBULL, "rog_twistknife.dds"},
	{0x316938EF6CF829CEULL, "llac_lt_0_lowtown_a2_n_c_2_0.dds"},
	{0x316939A2A5644EA1ULL, "6136160_m.fxe"},
	{0x31694F57221571AEULL, "dae000_keep_maid_sta6162422.cl"},
	{0x316958AF2196202BULL, "vxlt_warehouse_n_vc_0_0.vlm"},
	{0x316963F250D25142ULL, "static_lava_p1.psh.fxo"},
	{0x31698BE336C0A790ULL, "7608368.ogg"},
	{0x316A02F9705AB82EULL, "6096522_m.fxe"},
	{0x316A0886360EAC1FULL, "564640851.ogg"},
	{0x316A115BDD70E164ULL, "45831942.ogg"},
	{0x316A6C4BE6CB2706ULL, "489388288.ogg"},
	{0x316A76677D39D9A3ULL, "ser270_cut_motherreveal.cnv"},
	{0x316AB1F97C0143A8ULL, "vxlt_0_lowtown_a3_n_c_f58_0_0.vlm"},
	{0x316AE9E359F7B7FAULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierlrg.mmh"},
	{0x316AEDE359F7BE31ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierlrg.mao"},
	{0x316AFBE359F7D640ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierlrg.msh"},
	{0x316B05F9A825D89BULL, "character_character_p277.psh.fxo.dx11.assembly"},
	{0x316B4DA3F291EF17ULL, "fx_icecloud.dds"},
	{0x316BE7FDE1CE4292ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.dura.ani"},
	{0x316BFE0A32971D58ULL, "vxlt_0_lowtown_a2_n_ets_0_0.vlm"},
	{0x316C0C6EF9601934ULL, "vxht_keep_n_5gr_0_0.vlm"},
	{0x316C35E2D9383682ULL, "llac_dr_primevalentrance_1726_2.dds"},
	{0x316C488B478DF420ULL, "mag320_nobleman_36060272.cl"},
	{0x316C7BCB48B26297ULL, "zz_apppt_debug.ncs"},
	{0x316C860E70900F34ULL, "6076580_m.fxe"},
	{0x316CD7812872EEAEULL, "183124953.ogg"},
	{0x316D51F25882C1DFULL, "ht_0_hightown_54_0.rml"},
	{0x316E66A02A452833ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dieloop.evt"},
	{0x316E9249DD581B4AULL, "vxlt_gallowstemplar_f_3qo_0_0.vlm"},
	{0x316EB44C0BA1D22FULL, "gen00fl_aveline6106944.cl"},
	{0x316EB8A9860B34CDULL, "463137597.ogg"},
	{0x316EF074463A566CULL, "vxht_chantry_28g_0_0.vlm"},
	{0x316F42A774F14375ULL, "6020530_f.fxe"},
	{0x316F9F1864DAA4EEULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupa.msh"},
	{0x316FAD1864DABC3FULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupa.mao"},
	{0x316FB91864DAD0DCULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupa.mmh"},
	{0x316FEBA3E8CAAA5BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_diagonal_1.mao"},
	{0x31704B933DDB2B4DULL, "6175161_m.fxe"},
	{0x3170704C55826109ULL, "hlac_lt_undercityencounter_-1_49x.dds"},
	{0x3170AED85B2D8FBEULL, "dae300_bran6144941.cl"},
	{0x3170CE4C66284A94ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.dura.ani"},
	{0x3170DE8F28DA86F9ULL, "6179269_f.fxe"},
	{0x3170E24C54A08D43ULL, "335279208.ogg"},
	{0x3170E9C808AC6C69ULL, "vxht_mansion2_2a9_0_0.vlm"},
	{0x31710056CBBC5764ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_jump.ani"},
	{0x31712FF875E79AC4ULL, "vxht_0_hightown_n_bx0_0_0.vlm"},
	{0x317132D2C59BCAAFULL, "vxca_caverns2_9dx_0_0.vlm"},
	{0x3171477F1AA0EFE3ULL, "vxlt_0_lowtown_docks_n_bb4_0_0.vlm"},
	{0x3171BB6C118BE15EULL, "one140cr_charade6173705.cl"},
	{0x3171DEF6F840615AULL, "6055585_m.fxe"},
	{0x3171F91889A5FA34ULL, "vxdr_slavers_1ar_0_0.vlm"},
	{0x31727C33F94581C0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_one_hand_loop.ani"},
	{0x3172E2D2C9657DE3ULL, "vxlt_0_lowtown_a2_n_gof_0_1.vlm"},
	{0x3172EDD47B7AA29BULL, "vxca_cave_1lo_0_0.vlm"},
	{0x3173037695B51E9AULL, "vxht_keep_n_6f8_0_0.vlm"},
	{0x31731B77CD928D82ULL, "l_env_wal_gallows_calmgrp.wbd"},
	{0x3173832C78FB2914ULL, "6078753_m.fxe"},
	{0x31739AB6DAF786E2ULL, "vxht_0_hightown_a3_amp_0_0.vlm"},
	{0x31746A717ACEC08DULL, "llac_lt_0_lowtown_-65535_3.dds"},
	{0x317488808C59E877ULL, "194797501.ogg"},
	{0x31748E3AC4777F86ULL, "6096812_m.fxe"},
	{0x31749744DA0F1A8DULL, "vxw_bonepit_1mm_0_0.vlm"},
	{0x31750A0615EF9FB7ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_4y.dds"},
	{0x31751B0D457E88E8ULL, "vxht_0_hightown_a2_9p4_0_0.vlm"},
	{0x31755FA02A4B3B74ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dieloop.gad"},
	{0x3175D505A4862FFBULL, "6016325_m.fxe"},
	{0x31764E70D61EDDC7ULL, "694962175.ogg"},
	{0x317652AB464C623DULL, "vxlt_gallowstemplar_f_3s2_0_0.vlm"},
	{0x31765389DECE8BDFULL, "398999420.ogg"},
	{0x31765390CF3D5585ULL, "42990036.ogg"},
	{0x3176E5D0D4BB3383ULL, "blood3.dds"},
	{0x3177630464AF9621ULL, "mag160_feynriel6020071.cl"},
	{0x31777DC2848E74ADULL, "6140279_m.fxe"},
	{0x3177CB4C662E64F6ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.dura.gad"},
	{0x3177FD56CBC270C6ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_jump.gad"},
	{0x31783446ED472AF6ULL, "hlac_lt_0_lowtown_docks_n_-131073_1x.dds"},
	{0x31788B0DC7633564ULL, "vxht_bartrand_n_5qk_0_0.vlm"},
	{0x3178AE7367479DAFULL, "vxlt_0_lowtown_n_def_0_0.vlm"},
	{0x317934A4ABEF732CULL, "992428037.ogg"},
	{0x31796CF1417D5EEEULL, "6119638_m.fxe"},
	{0x3179FAE8DD47B2B1ULL, "vxht_keep_n_681_0_0.vlm"},
	{0x317A19B801B8409CULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyn_0s.dds"},
	{0x317AA1710D033797ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaopen.mmh"},
	{0x317AA5710D033EA4ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaopen.mao"},
	{0x317AB7710D035D09ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaopen.msh"},
	{0x317ADC09937372AAULL, "dre180ip_invisible_door.utp"},
	{0x317B49D219CC0553ULL, "sst_ss_ambient_m.gda"},
	{0x317B8CE9DA3E6609ULL, "6122095_m.fxe"},
	{0x317BF60FD0190434ULL, "ca_caverns2prc_probe_9ep.dds"},
	{0x317BF78BF51CF6A7ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation1.gad"},
	{0x317C0DFAB063F777ULL, "w_sundermount_90_0.rml"},
	{0x317C2BE0BBE558A4ULL, "378212326.ogg"},
	{0x317C6EA02A517352ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dieloop.ani"},
	{0x317C887890E51BB1ULL, "mag320_evelina6051875.cl"},
	{0x317D04801C0F3E07ULL, "vxht_mansion_n_6il_0_0.vlm"},
	{0x317D396C29D9F4C2ULL, "vxw_bonepitdestroyed_a3_19j_0_0.vlm"},
	{0x317D3D6177A3C0D3ULL, "narrative_flemeth6002775.cl"},
	{0x317E0DE75C6EC9FCULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_c.impa.ani"},
	{0x317E373B7697DF2FULL, "vxca_cavern1_haunted_dpk_0_0.vlm"},
	{0x317EE64C6634B155ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.dura.evt"},
	{0x317EF856CBC886A5ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_jump.evt"},
	{0x317F5BA4FAA284E9ULL, "vxht_mansion2_2j8_0_0.vlm"},
	{0x317F760BBEA4163FULL, "vxca_caverns1_j24_0_2.vlm"},
	{0x317F7E679A7E9091ULL, "lt_0_lowtown_a2_n_c_42_0.anb"},
	{0x317F9055073A5522ULL, "6174993_f.fxe"},
	{0x317FA1D567DDB344ULL, "vxht_0_hightown_n_bc1_0_0.vlm"},
	{0x31801563EDB389BDULL, "6070342_f.fxe"},
	{0x3180493F22DF60F7ULL, "466914795.ogg"},
	{0x3180BE5CAAA74E6FULL, "hlac_lt_0_lowtown_a2_n_c_-65537_1y.dds"},
	{0x3180E46FD9333EB6ULL, "335260254.ogg"},
	{0x318111C2CB071C48ULL, "622483177.ogg"},
	{0x31813C25B397AE67ULL, "vxht_0_hightown_a2_c52_0_0.vlm"},
	{0x318174BF91F4A64FULL, "6150162_m.fxe"},
	{0x3181CFFFDAD247D5ULL, "gen_im_trs_dal_junk2_03.uti"},
	{0x31820C93CCED9F62ULL, "vxht_0_hightown_a3_ajq_0_0.vlm"},
	{0x31820CFAB068616FULL, "w_sundermount_90_0.phy"},
	{0x31820DD2958DDFF1ULL, "6076457_m.fxe"},
	{0x31820F8232CF6AF2ULL, "docks_male_sailor_mob.cnv"},
	{0x318249180CDDC304ULL, "44000187.ogg"},
	{0x318252723EF7A650ULL, "character_character_p193.psh.fxo.dx11"},
	{0x318309177295DAA8ULL, "static_static_p9.vsh.fxo.dx11"},
	{0x318328865E302328ULL, "vxlt_hangedman_n_1w0_0_0.vlm"},
	{0x318338FEF519BB81ULL, "vxlt_0_lowtown_n_e1y_0_0.vlm"},
	{0x31833A13F7252D55ULL, "6028403_m.fxe"},
	{0x318357E168A589CCULL, "dae020cr_merch_tailor.utc"},
	{0x3183808F0A9030C1ULL, "dae000_templar_knight6165164.cl"},
	{0x3183A4854269C211ULL, "272002037.ogg"},
	{0x3183E9FEC37362C0ULL, "character_character_p119.psh.fxo.dx11.assembly"},
	{0x3184766F5C2169B4ULL, "342974140.ogg"},
	{0x31847DABA172DC75ULL, "294466109.ogg"},
	{0x31851AE75C74FE3EULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_c.impa.gad"},
	{0x3185318232D1DA4AULL, "docks_male_sailor_mob.bnk"},
	{0x31855512C210970FULL, "6165243_m.fxe"},
	{0x31857CABCEBA1A45ULL, "plt_prm000pt_hindsight_belt.plo"},
	{0x31858B8D2ED6D170ULL, "water_water_p5.psh.fxo"},
	{0x3185CA112B759264ULL, "w_encounter2_17_0.msh"},
	{0x3185DC112B75B0CDULL, "w_encounter2_17_0.mao"},
	{0x3185E0112B75B706ULL, "w_encounter2_17_0.mmh"},
	{0x31873CCC295B718BULL, "gen00fl_merrill6127232.cl"},
	{0x31879F7ABFCFB47BULL, "6049677_m.fxe"},
	{0x3187E99AE7943819ULL, "vxht_0_hightown_n_9xb_0_0.vlm"},
	{0x3187F9122282508EULL, "vxca_caverns1_hu8_0_0.vlm"},
	{0x31882B79BE91300BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\blendtree\\doubledoor.anb"},
	{0x3188599D2DDA1280ULL, "282406626.ogg"},
	{0x31886CC5CF3DCB31ULL, "vxht_keep_n_630_0_0.vlm"},
	{0x3188F1C94E4F40AAULL, "6136714_m.fxe"},
	{0x318924BEBF4F5614ULL, "lt_0_lowtown_a2_39_0.rml"},
	{0x318968201F5A1DF2ULL, "gen00fl_aveline6114841.cl"},
	{0x318986E5BA956487ULL, "dae000_pc_home_ambients6208485.cl"},
	{0x318A5D9962A80C7AULL, "6175922_m.fxe"},
	{0x318A63FC7A9D3B38ULL, "hlac_lt_0_lowtown_n_-1_4x.dds"},
	{0x318A83B12CD58263ULL, "gen00fl_merrill6116063.cl"},
	{0x318B175A1D32D16AULL, "vxw_bonepitdestroyed_a3_23w_0_0.vlm"},
	{0x318B6F08BEC6C4E6ULL, "vxlt_gallowstemplar_2pb_0_0.vlm"},
	{0x318B765D0E055687ULL, "vxht_mansion_n_6lo_0_0.vlm"},
	{0x318B9123201362E4ULL, "6119140_m.fxe"},
	{0x318BEAC0C9A26B9CULL, "character_character_p1.psh.fxo"},
	{0x318BF5E75C7ADE5DULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_c.impa.evt"},
	{0x318C1BFF511101CEULL, "w_bonepitdestroyed_a3_112_0.msh"},
	{0x318C29FF5111191FULL, "w_bonepitdestroyed_a3_112_0.mao"},
	{0x318C35FF51112DBCULL, "w_bonepitdestroyed_a3_112_0.mmh"},
	{0x318C5B331F3C0D2AULL, "hf_gen_templar.mor"},
	{0x318CF56955D1C535ULL, "hlac_lt_0_lowtown_docks_n_131073_0y.dds"},
	{0x318D00C4FB934F75ULL, "630062987.ogg"},
	{0x318D2EA43C250078ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb106a_0t.dds"},
	{0x318D79547F2763E9ULL, "gen00fl_isabela6075388.cl"},
	{0x318D94AF90F23F47ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirewall.anb"},
	{0x318DDB253FBD3F01ULL, "dae011_drunk1_sta.bnk"},
	{0x318E11945AD75832ULL, "6129414_f.fxe"},
	{0x318E6DE15D80FCF1ULL, "vxlt_gallowstemplar_5rh_0_0.vlm"},
	{0x318E819E51D79922ULL, "1072387048.ogg"},
	{0x318E9D7AEA520958ULL, "llac_lt_undercityencounter_-1_40.dds"},
	{0x318F07CB7B6F4B82ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd06a.msh"},
	{0x318F09CB7B6F4F38ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd06a.mmh"},
	{0x318F62B8616632F0ULL, "one110ip_ggt_tree.ncs"},
	{0x318F8D40456C283CULL, "169518493.ogg"},
	{0x318F8F801345A746ULL, "6028444_m.fxe"},
	{0x318F9FAA0F7E8022ULL, "vxlt_gallowstemplar_n_5pj_0_0.vlm"},
	{0x318FD7112B7DFCD1ULL, "w_encounter2_17_0.phy"},
	{0x318FFA84741C2A32ULL, "6041724_m.fxe"},
	{0x31900C8036BD7B97ULL, "hlac_lt_0_lowtown_docks_1_0x.dds"},
	{0x31905850114DE6F9ULL, "6119523_m.fxe"},
	{0x31907E70BEE3B7E2ULL, "vxht_0_hightown_a3_awr_0_0.vlm"},
	{0x3190E48BF52F32EDULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation1.ani"},
	{0x3190FD253FBFAF71ULL, "dae011_drunk1_sta.cnv"},
	{0x3190FE0FE4030D08ULL, "38559349.ogg"},
	{0x3191192724B3AA52ULL, "hlac_ca_caverns1_mines_24422_2x.dds"},
	{0x319154AC1CDE2815ULL, "324955719.ogg"},
	{0x3191AC660AA17F3EULL, "follower_banter6164020.cl"},
	{0x3191B16BDA1E83ADULL, "w_blightlands_235_0.msh"},
	{0x3191BF6BDA1E9BE0ULL, "w_blightlands_235_0.mao"},
	{0x3191C36BDA1EA233ULL, "w_blightlands_235_0.mmh"},
	{0x3191D16335376E05ULL, "345116564.ogg"},
	{0x3192182EA61E752BULL, "646617562.ogg"},
	{0x31923497FD4348E2ULL, "vxht_0_hightown_cbz_0_0.vlm"},
	{0x31924D489A55DAE8ULL, "vxlt_0_lowtown_a3_n_c_eoe_0_0.vlm"},
	{0x31927032F23DE053ULL, "6116582_m.fxe"},
	{0x319290467EF26C99ULL, "vxlt_gallowstemplar_f_398_0_0.vlm"},
	{0x3192DAC7AE508BBEULL, "vxlt_undercityserialkiller_217_0_0.vlm"},
	{0x31936973D733314CULL, "398928580.ogg"},
	{0x31943A91966F62B8ULL, "18842744.ogg"},
	{0x3194514C98338997ULL, "m_dae311ar_hanged_man.wbd"},
	{0x319462AAA1C1887EULL, "6161268_m.fxe"},
	{0x3194981864F9B4A7ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupa.phy"},
	{0x3194E642CFEB22D5ULL, "hlac_lt_gallowstemplar_7198_2z.dds"},
	{0x3195326C8FBA764BULL, "dae000_bartrand6204165.cl"},
	{0x31958FFB240CFB94ULL, "818082283.ogg"},
	{0x3195A529ED167F9AULL, "nat380_party_banter6066174.cl"},
	{0x3195E4A9110A38A5ULL, "cod_ltr_ggt_note2.utp"},
	{0x3195E9D41C27940DULL, "vxca_caverns1_edi_0_0.vlm"},
	{0x31965C5AB8CB5A1BULL, "vxlt_0_lowtown_eid_0_0.vlm"},
	{0x31965F04CEA378EFULL, "661747895.ogg"},
	{0x3196617E7EFB505BULL, "vxdr_0_deeproads_a26_0_0.vlm"},
	{0x3196E0112B842ACDULL, "w_encounter2_17_0.rml"},
	{0x3196E3CBC0E10539ULL, "6217102_m.fxe"},
	{0x3197231071B0A36EULL, "fen110_fenris6117956.cl"},
	{0x31972D58A6F9213CULL, "vxca_caverns1_mines_i2o_0_0.vlm"},
	{0x319763A64045592AULL, "6131614_m.fxe"},
	{0x31977D0B8A1950FBULL, "6132786_m.fxe"},
	{0x31979759989FACF8ULL, "zz_dog_debug6127935.cl"},
	{0x3197A981F653FAF0ULL, "zz_ser_debug.ncs"},
	{0x3197B5DB6806E5E3ULL, "6135080_m.fxe"},
	{0x3197DB0D47DA7BCBULL, "qcr000tr_ambients_run.ncs"},
	{0x31985883CF1E899BULL, "vxht_chantry_n_2y4_0_0.vlm"},
	{0x3198BF70ED5EC638ULL, "573466375.ogg"},
	{0x319910CB7B77AFC3ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd06a.phy"},
	{0x31992157A11E8361ULL, "6103408_f.fxe"},
	{0x319921ED90E88E4AULL, "w_bonepit_1_0.mao"},
	{0x319925ED90E89499ULL, "w_bonepit_1_0.mmh"},
	{0x31992FED90E8A587ULL, "w_bonepit_1_0.msh"},
	{0x31993D5B32C04339ULL, "vxht_0_hightown_b4z_0_0.vlm"},
	{0x31994656AE0C8930ULL, "art\\vfx\\models\\vfx_gore.mmh"},
	{0x31995456AE0CA14AULL, "art\\vfx\\models\\vfx_gore.msh"},
	{0x319974A19998C3DBULL, "6034659_m.fxe"},
	{0x3199A4B210AC29BEULL, "6180878_m.fxe"},
	{0x3199A8184AAE78BFULL, "vxht_chantry_28h_0_0.vlm"},
	{0x3199B7064BC4771CULL, "595037096.ogg"},
	{0x3199CB9C4C463797ULL, "6131888_m.fxe"},
	{0x3199CC4BCDA09A0CULL, "hlac_ca_cave_2194_12x.dds"},
	{0x319A030012097B64ULL, "6114640_m.fxe"},
	{0x319A2903265DE06BULL, "vxlt_0_lowtown_h9e_0_0.vlm"},
	{0x319A735C3FE89B77ULL, "one120_athenril6088288.cl"},
	{0x319A854A7703CCB8ULL, "rdr111ip_at_docks.ncs"},
	{0x319ABE9A4FC7BF2FULL, "6016585_m.fxe"},
	{0x319AE31DD129CD37ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_bowl.mmh"},
	{0x319AE71DD129D444ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_bowl.mao"},
	{0x319AED68DE5F984DULL, "follower_banter6173164.cl"},
	{0x319AF91DD129F2A9ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_bowl.msh"},
	{0x319B1AF55A84102FULL, "vxca_cavern1_haunted_dnu_0_0.vlm"},
	{0x319B365FF5CADB45ULL, "hlac_lt_0_lowtown_a2_65536_3x.dds"},
	{0x319B866F78A406C2ULL, "vxca_cavern1_haunted_ikn_0_0.vlm"},
	{0x319B8F53C45D1E9EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev5.ani"},
	{0x319BB3AE4DEAEBBBULL, "6120150_m.fxe"},
	{0x319C3535D107E4EFULL, "735847672.ogg"},
	{0x319C858F4BC9E444ULL, "vxht_0_hightown_n_bi7_0_0.vlm"},
	{0x319CE92942437FC3ULL, "vxht_mansion2_257_0_0.vlm"},
	{0x319D244A7270A3D9ULL, "vxlt_gallowstemplar_f_64z_0_0.vlm"},
	{0x319D3809A2359B26ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.anb"},
	{0x319D8B6AA159893EULL, "166129610.ogg"},
	{0x319D9236E4E5A1AEULL, "dae_basic_attacks.ncs"},
	{0x319DB4096277D65EULL, "vxlt_undercityserialkiller_1m8_0_0.vlm"},
	{0x319E402271565B65ULL, "48022951.ogg"},
	{0x319EFBAB8B0BAC0CULL, "6096300_m.fxe"},
	{0x319F129D2AC0CB42ULL, "hlac_ht_0_hightown_a2_2_1x.dds"},
	{0x319FA51782725DD5ULL, "pro000_colonel_ewald6039951.cl"},
	{0x319FC8992335D003ULL, "dae000_templar_recruit_16164434.cl"},
	{0x31A01AA0883D4395ULL, "coins_sa.dds"},
	{0x31A1A413AD39846CULL, "6043971_m.fxe"},
	{0x31A1A625C1026B71ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_darren.gad"},
	{0x31A1D6AA6738B695ULL, "6070198_m.fxe"},
	{0x31A1EF5FFFEC7AE2ULL, "vxlt_gallowstemplar_n_2e6_0_0.vlm"},
	{0x31A213AEC16FDEECULL, "560161737.ogg"},
	{0x31A225DB74DC18ACULL, "130073635.ogg"},
	{0x31A25647D33116F0ULL, "947871455.ogg"},
	{0x31A28C9F37D39DA5ULL, "mrl000_marethari6205690.cl"},
	{0x31A2E5B0CAC550D7ULL, "hlac_ht_0_hightown_n_0_2y.dds"},
	{0x31A2FA36F29AE4FFULL, "6101388_m.fxe"},
	{0x31A347C324AD796FULL, "ser270_emeric6149367.cl"},
	{0x31A35872FD45417FULL, "6067387_m.fxe"},
	{0x31A39963FC8977D5ULL, "vxlt_0_lowtown_a2_n_c_emg_0_0.vlm"},
	{0x31A4F2388ED311F6ULL, "vxca_caverns2_8k4_0_1.vlm"},
	{0x31A529DB5C39E13DULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120.phy"},
	{0x31A545DCF0256F4CULL, "6130133_f.fxe"},
	{0x31A558A24E72D5F7ULL, "vxlt_gallowstemplar_4o0_0_0.vlm"},
	{0x31A5C0541D1E11B1ULL, "6110155_m.fxe"},
	{0x31A5E69960B676B5ULL, "vxlt_0_lowtown_docks_a3_n_cyu_0_0.vlm"},
	{0x31A610B6BA41CEDEULL, "mcr370_messenger6071837.cl"},
	{0x31A6255F0BCAEBEFULL, "214747226.ogg"},
	{0x31A63C6ECEA749BCULL, "qun120_arishok6051317.cl"},
	{0x31A6CA60C114A21BULL, "vxht_chantry_n_2t7_0_0.vlm"},
	{0x31A74BC4413776F8ULL, "dae021_sister_lorena_sta6153107.cl"},
	{0x31A78171B1A9F75FULL, "llac_lt_0_lowtown_a2_n_131072_1.dds"},
	{0x31A7D7F5F5AF919FULL, "vxlt_gallowstemplar_f_4ue_0_0.vlm"},
	{0x31A7E27E1B24C442ULL, "w_woundedcoast_103_0.phy"},
	{0x31A7EB446EE6E14EULL, "6033134_m.fxe"},
	{0x31A89786919AC0CFULL, "6080292_m.fxe"},
	{0x31A8A725C1088B1FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_darren.ani"},
	{0x31A8B681F66283BBULL, "zz_ser_debug.cnv"},
	{0x31A9142B1DCB3B62ULL, "lt_0_lowtown_docks_222_0.msh"},
	{0x31A9162B1DCB3E18ULL, "lt_0_lowtown_docks_222_0.mmh"},
	{0x31A9222B1DCB52BBULL, "lt_0_lowtown_docks_222_0.mao"},
	{0x31A955188F8E7531ULL, "870686729.ogg"},
	{0x31A9673206D5DC09ULL, "6084319_m.fxe"},
	{0x31A9D0F857174E3AULL, "vxca_caverns1_mines_j24_0_2.vlm"},
	{0x31A9F335F11389D4ULL, "zz_reenableguis.ncs"},
	{0x31AA46BF189C4EA1ULL, "6120407_f.fxe"},
	{0x31AAD35858E6F806ULL, "6093987_m.fxe"},
	{0x31AAE89CD65D9732ULL, "6033985_m.fxe"},
	{0x31AB0C35B1401623ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crn_rdy_rf.evt"},
	{0x31AB3B09A241DDD1ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.mmh"},
	{0x31AB519B416D2DF1ULL, "vxlt_gallowstemplar_5xn_0_0.vlm"},
	{0x31AB64CB88F9DB54ULL, "6110258_f.fxe"},
	{0x31AC496A008305C6ULL, "6019031_m.fxe"},
	{0x31AC70291DC2E8BDULL, "699310904.ogg"},
	{0x31ACD299A363518EULL, "mag291_vincento6027368.cl"},
	{0x31ACEE3EBCFE2F33ULL, "one100dg_kelder6060647.cl"},
	{0x31AD85F0EBA43819ULL, "116027749.ogg"},
	{0x31ADDBA1B5D75A3EULL, "6143448_m.fxe"},
	{0x31AE0437C3DA585DULL, "vxdr_primevaltomb_7oc_0_0.vlm"},
	{0x31AE69897C974A9AULL, "556440129.ogg"},
	{0x31AE74A4BEF2EFBEULL, "static_static_p36.psh.fxo.dx11"},
	{0x31AE8D01C4DEFDFBULL, "vxht_0_hightown_a3_ccm_0_0.vlm"},
	{0x31AF19FB9EB5C8E4ULL, "828472407.ogg"},
	{0x31AF31027E420BE8ULL, "vxlt_0_lowtown_a3_n_c_euc_0_0.vlm"},
	{0x31AF50DB3F4757C3ULL, "vxht_mansion2b_3gm_0_1.vlm"},
	{0x31AFA853C46DF218ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev5.gad"},
	{0x31B052AD32E4A751ULL, "6076307_m.fxe"},
	{0x31B0E3C7F57C68B8ULL, "hlac_lt_0_lowtown_a3_n_c_65537_2y.dds"},
	{0x31B0FB1BA1F93590ULL, "vxlt_gallowscourtyard_7bi_0_0.vlm"},
	{0x31B18E87AEEA857AULL, "vxw_encounter3_eb_0_0.vlm"},
	{0x31B1CA4FEF3B59FFULL, "6062087_m.fxe"},
	{0x31B1E535B145F2A4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crn_rdy_rf.gad"},
	{0x31B2672B185F97F9ULL, "vxlt_0_lowtown_a3_n_c_hwu_0_0.vlm"},
	{0x31B283F5615AB5DFULL, "6125162_m.fxe"},
	{0x31B29C4A771885B1ULL, "rdr111ip_at_docks.utp"},
	{0x31B305FFDB2F0455ULL, "401773757.ogg"},
	{0x31B30B81A4270445ULL, "vxht_keep_6uq_0_0.vlm"},
	{0x31B4376710F0F736ULL, "6042440_m.fxe"},
	{0x31B473365FB701D1ULL, "vxlt_darktown_41x_0_0.vlm"},
	{0x31B495C0A58A2D72ULL, "hlac_lt_undercityencounter_-1_22x.dds"},
	{0x31B4C8693FEFCCB1ULL, "gen00fl_sebastian6180443.cl"},
	{0x31B512CA046970C2ULL, "629083583.ogg"},
	{0x31B545BEBF74C914ULL, "lt_0_lowtown_a2_39_0.anb"},
	{0x31B5E00D47F381B2ULL, "qcr000tr_ambients_run.utt"},
	{0x31B60F080E563167ULL, "vxw_bonepit_1pd_0_0.vlm"},
	{0x31B624D4FA1D5308ULL, "gen00fl_aveline6115769.cl"},
	{0x31B62ADB5C485642ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120.mmh"},
	{0x31B656885CEF874AULL, "vxca_caverns2_94a_0_0.vlm"},
	{0x31B659FC39691AF1ULL, "fx_dis_streaks.mao"},
	{0x31B67C9571F8D6F3ULL, "art\\vfx\\textures\\vfx_crack01.dds"},
	{0x31B691AF91153F48ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirewall.mmh"},
	{0x31B6A9A63F6DCDC1ULL, "6161201_m.fxe"},
	{0x31B6B90813C10AE4ULL, "hlac_lt_foundry_n_320_9z.dds"},
	{0x31B73E899A589204ULL, "vxca_caverns2prc_8wb_0_0.vlm"},
	{0x31B78DA3A020C5ACULL, "hm_gencr_raider_archer.mor"},
	{0x31B7A1E243216BCEULL, "147510433.ogg"},
	{0x31B809E9E3A894BCULL, "vxlt_gallowstemplar_f_37c_0_0.vlm"},
	{0x31B84BB4779A91A9ULL, "vxca_caverns2_5xn_0_0.vlm"},
	{0x31B84E9445C8C1F4ULL, "gen00fl_aveline6139136.cl"},
	{0x31B8A5FE7D7090D6ULL, "6027702_m.fxe"},
	{0x31B8EF688D212AD5ULL, "216782896.ogg"},
	{0x31B8F435B14C2A82ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crn_rdy_rf.ani"},
	{0x31B9082E575CD0D6ULL, "vxht_0_hightown_8pg_0_0.vlm"},
	{0x31B9471BD5265799ULL, "6051384_m.fxe"},
	{0x31B954245E118447ULL, "gen_im_npc_dwf_arm_glv_meda.uti"},
	{0x31B9C9AE4991210DULL, "hlac_lt_0_lowtown_a3_n_c_-65535_5x.dds"},
	{0x31BA621EC25EF964ULL, "vxht_keep_4i6_0_0.vlm"},
	{0x31BAA4E651DB8A8FULL, "zz_ave_debug6061041.cl"},
	{0x31BAA85105F8E58FULL, "mag170_thrask6017037.cl"},
	{0x31BAD230A260D706ULL, "gen00fl_varric6118787.cl"},
	{0x31BAD9C13A2C0374ULL, "and331ar_sewers.are"},
	{0x31BAF540E560C9A2ULL, "vxlt_gallowstemplar_n_5au_0_0.vlm"},
	{0x31BB2977CDD05B38ULL, "l_env_wal_gallows_calmgrp.bnk"},
	{0x31BBC6781B370F29ULL, "dae000ip_crowd_mf8_talk5.utp"},
	{0x31BBD4ECF9524930ULL, "6026963_m.fxe"},
	{0x31BC150958353094ULL, "6200855_m.fxe"},
	{0x31BC482DD9066F72ULL, "plt_prm_cod_apostate_staff.plo"},
	{0x31BC7D17EC5F09C4ULL, "rdr111dg_martin6055042.cl"},
	{0x31BCC77C1482955BULL, "zz_isa_debug6133941.cl"},
	{0x31BCCBE32AC6F585ULL, "646093221.ogg"},
	{0x31BD26941B625E07ULL, "6175404_m.fxe"},
	{0x31BDE36644D4F7A2ULL, "6116213_m.fxe"},
	{0x31BDEDD957D455A6ULL, "zz_dae_debug6029132.cl"},
	{0x31BEDAA3A5773E7FULL, "vxht_0_hightown_a2_aut_0_0.vlm"},
	{0x31BF09630719E5B6ULL, "hlac_lt_warehouse_-65536_1z.dds"},
	{0x31BF277C7B732D37ULL, "vxw_encounter2_7b_0_0.vlm"},
	{0x31BF4B27EDDF4C28ULL, "fx_ground_exp_snow.mao"},
	{0x31C0201F835FA5F6ULL, "qcr220_orsino6060700.cl"},
	{0x31C04514B7B2C1ABULL, "6103562_m.fxe"},
	{0x31C0465F37330E16ULL, "mag320_orsino6071162.cl"},
	{0x31C0504A8F1714D8ULL, "llac_ca_caverns2prc_5802_53.dds"},
	{0x31C08EA4DE93D0D0ULL, "vxht_keep_n_8bp_0_0.vlm"},
	{0x31C0929E9C165A54ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\blendtree\\floortrap.anb"},
	{0x31C0C45E8131AFE4ULL, "vxht_bartrand_n_5hl_0_0.vlm"},
	{0x31C12AB43FC54363ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_03tent.mao"},
	{0x31C12EB43FC549B0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_03tent.mmh"},
	{0x31C13CB43FC561CAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_03tent.msh"},
	{0x31C171BA3988CD0EULL, "6031309_m.fxe"},
	{0x31C1AADAFBACDC74ULL, "741212649.ogg"},
	{0x31C1C5B718F9C024ULL, "6080097_m.fxe"},
	{0x31C1DB51678AD6FAULL, "vxht_keep_6av_0_0.vlm"},
	{0x31C2166B3A8BFF88ULL, "dae000cr_shopper_m.utc"},
	{0x31C29486B1107C1EULL, "gen_im_wep_rog_arc_lore_01.uti"},
	{0x31C2EC870F136B45ULL, "1009063716.ogg"},
	{0x31C39955824A5B0DULL, "gen00fl_isabela6015830.cl"},
	{0x31C480E5004C49E7ULL, "vxw_bonepit_1ug_0_0.vlm"},
	{0x31C4AF70B50E35BAULL, "183147815.ogg"},
	{0x31C4B9A5EA18D908ULL, "mer341_jansen6082987.cl"},
	{0x31C4BDDEC928C889ULL, "6070240_m.fxe"},
	{0x31C4F3B836EA38D9ULL, "llac_lt_0_lowtown_a3_n_c_-262144_2.dds"},
	{0x31C51B833163E641ULL, "6166301_m.fxe"},
	{0x31C58FB3EB07FC34ULL, "712400914.ogg"},
	{0x31C601073CE50B1BULL, "69135131.ogg"},
	{0x31C61E85C2FA4B1CULL, "hlac_lt_0_lowtown_a2_n_c_65535_4z.dds"},
	{0x31C7B76A555E2667ULL, "hlac_lt_hangedman_-1_11x.dds"},
	{0x31C7D79F19A18A4EULL, "6141198_m.fxe"},
	{0x31C822872881A02FULL, "one100_party_comments6133373.cl"},
	{0x31C82E835F5AD002ULL, "vxdr_primevaltomb_745_0_0.vlm"},
	{0x31C84543C6303B7FULL, "lgt160_wk_soph_sophia.bnk"},
	{0x31C8D3FBB45511E4ULL, "vxht_keep_4n7_0_0.vlm"},
	{0x31C8D7AB6A3A3E8FULL, "vxw_sundermount_30t_0_1.vlm"},
	{0x31C9E07C51C952F3ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_1x.dds"},
	{0x31CA2F02EBAC5411ULL, "6171468_m.fxe"},
	{0x31CA6367425A320AULL, "ser270_emeric6111480.cl"},
	{0x31CA7D77CE26A1B3ULL, "327878415.ogg"},
	{0x31CAA87B58BFDC31ULL, "6094506_f.fxe"},
	{0x31CAD8B65557893BULL, "6131963_m.fxe"},
	{0x31CAE69443E224F9ULL, "6093816_m.fxe"},
	{0x31CAF54F778074C0ULL, "vxht_keep_n_5k2_0_0.vlm"},
	{0x31CB155D34506AFCULL, "plt_lgt000pt_baz_dirty.nss"},
	{0x31CB1D106D2040EDULL, "zz_ave_debug6064619.cl"},
	{0x31CC2743C633F087ULL, "lgt160_wk_soph_sophia.cnv"},
	{0x31CC441F8041DDB7ULL, "6156476_m.fxe"},
	{0x31CC680F5E0653B4ULL, "mag330_walter6046159.cl"},
	{0x31CC755635A298C5ULL, "vxlt_0_lowtown_a2_e1n_0_0.vlm"},
	{0x31CC9EC3A3283168ULL, "243481929.ogg"},
	{0x31CCBD7E1B43A1F6ULL, "w_woundedcoast_103_0.mao"},
	{0x31CCC97E1B43B595ULL, "w_woundedcoast_103_0.mmh"},
	{0x31CCCB7E1B43B90BULL, "w_woundedcoast_103_0.msh"},
	{0x31CD8A0B03273B7AULL, "904893528.ogg"},
	{0x31CDD816F750D432ULL, "gen_im_npc_arm_hum_boo_mrcb.uti"},
	{0x31CE539038A271A4ULL, "da2terrain_terrain_p2.psh.fxo"},
	{0x31CE552B7911978FULL, "vxw_encounter2_qx_0_0.vlm"},
	{0x31CE964C03989086ULL, "144731415.ogg"},
	{0x31CF0A4B5A7AB79CULL, "hlac_w_blightlands_0_5x.dds"},
	{0x31CF2A5FB1DDEEABULL, "gen00fl_merrill6107450.cl"},
	{0x31CF6B91E6817A3AULL, "397659823.ogg"},
	{0x31CFA807FAEF1F88ULL, "6207977_m.fxe"},
	{0x31CFD2EB08DB5582ULL, "6125448_m.fxe"},
	{0x31CFF882C782B9F8ULL, "151959640.ogg"},
	{0x31D036AE9033FF7CULL, "hlac_lt_gallowstemplar_2383_2x.dds"},
	{0x31D068657588DC4EULL, "vxw_encounter2_k3_0_0.vlm"},
	{0x31D09E313DA26C6CULL, "art\\characters\\creatures\\ragedemon\\animation\\dmg_v1.ani"},
	{0x31D0F5915D9E0518ULL, "gen_im_arm_cht_noba_t1.uti"},
	{0x31D10EC3B9066626ULL, "6089108_m.fxe"},
	{0x31D2A5735A85E3C3ULL, "108571045.ogg"},
	{0x31D2C85EE6070BFCULL, "6218902_m.fxe"},
	{0x31D3135A5D7AE92CULL, "147466653.ogg"},
	{0x31D395699BEF63C1ULL, "lt_undercityencounter_mmi.gff"},
	{0x31D42BEAFF36C6C6ULL, "gencr_qunari_arishock.utc"},
	{0x31D4465C795C80AAULL, "ser221_party_comments6122860.cl"},
	{0x31D458973922ADBFULL, "6052909_m.fxe"},
	{0x31D4B9BAFD1ECCA9ULL, "and100_anders6096022.cl"},
	{0x31D4EDA3C794C5BCULL, "vxlt_gallowstemplar_f_39a_0_0.vlm"},
	{0x31D5123CAB42099FULL, "6173824_m.fxe"},
	{0x31D56533D7A97C3DULL, "follower_banter6163972.cl"},
	{0x31D5AEE1E7C91274ULL, "mer320_hubert6082039.cl"},
	{0x31D684D5460D94B4ULL, "vxw_bonepit_25c_0_0.vlm"},
	{0x31D6AB313DA6EDAEULL, "art\\characters\\creatures\\ragedemon\\animation\\dmg_v1.gad"},
	{0x31D76253580F49EDULL, "qun120cr_javaris6042992.cl"},
	{0x31D7CAD1E88255C0ULL, "gen_im_trs_dwv_dua_dag_r1.uti"},
	{0x31D8041DFB11DCC8ULL, "ser221_gascard6111585.cl"},
	{0x31D8381CBA233BC2ULL, "vxlt_gallowscourtyard_7cx_0_0.vlm"},
	{0x31D84FAB5FFE155CULL, "mag160_feynriel6016583.cl"},
	{0x31D884F7BC7347A0ULL, "vxht_0_hightown_byd_0_0.vlm"},
	{0x31D8D97E2280DA6BULL, "vxht_chantry_n_2ra_0_0.vlm"},
	{0x31D998516036C810ULL, "vxlt_0_lowtown_docks_baq_0_0.vlm"},
	{0x31D9A23622AD0E59ULL, "vxlt_0_lowtown_a2_gny_0_0.vlm"},
	{0x31D9C0C13A468956ULL, "and331ar_sewers.ncs"},
	{0x31D9D036724B86ABULL, "6173330_m.fxe"},
	{0x31DA7C125A54A5DCULL, "hlac_lt_0_lowtown_-196608_4z.dds"},
	{0x31DAED2F96DFE045ULL, "art\\vfx\\combat\\vfx_r_var_bayonet_v5_c.anb"},
	{0x31DB207EBEDCF29FULL, "753588924.ogg"},
	{0x31DB22CAA66DA6F4ULL, "vxlt_0_lowtown_a2_n_c_hld_0_0.vlm"},
	{0x31DB3B4FC69E6C4DULL, "hlac_w_encounter2_camp_262143_0y.dds"},
	{0x31DB3DBE89DB60B9ULL, "zz_pro_debug6059746.cl"},
	{0x31DB464D44D4A944ULL, "vxht_0_hightown_n_9ux_0_0.vlm"},
	{0x31DB7D84C1147718ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_smallcap.phy"},
	{0x31DB820BA9C80253ULL, "328456957.ogg"},
	{0x31DBB8A1DFF81C18ULL, "6082132_m.fxe"},
	{0x31DC222AA1C08355ULL, "vxht_playershouse_n_1q0_0_0.vlm"},
	{0x31DC5E9289A392B7ULL, "vxw_blightlands_ak_0_0.vlm"},
	{0x31DC862BB35F14C5ULL, "vxw_encounter3_ei_0_0.vlm"},
	{0x31DC90C5DD6D6BA6ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_5.dds"},
	{0x31DCE49B6396F08EULL, "vxlt_0_lowtown_a3_n_c_eqa_0_0.vlm"},
	{0x31DCE6CCEDE92A35ULL, "949006551.ogg"},
	{0x31DD40C1669DB91DULL, "hlac_ca_caverns2_5802_2x.dds"},
	{0x31DD59B861A844F9ULL, "one110ip_ggt_tree.utp"},
	{0x31DD60D7C461D9D5ULL, "vxlt_0_lowtown_a2_n_c_eqc_0_0.vlm"},
	{0x31DD661BF703FAFCULL, "zz_mer_debug6062739.cl"},
	{0x31DD8AE6F1F4D511ULL, "vxca_cavern1_haunted_j21_0_0.vlm"},
	{0x31DDA6313DAD038DULL, "art\\characters\\creatures\\ragedemon\\animation\\dmg_v1.evt"},
	{0x31DE3E64105B1493ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_down.ani"},
	{0x31DE809247704EF5ULL, "vxlt_hangedman_n_15s_0_0.vlm"},
	{0x31DEC25AB2214F7DULL, "ser221_party_comments6166191.cl"},
	{0x31DF53D5F9BC6E40ULL, "vxdr_slavers_1j3_0_0.vlm"},
	{0x31DF971170EBDC0FULL, "gen00fl_varric6099403.cl"},
	{0x31DFCCF844A16C9DULL, "cotbedroll_n.dds"},
	{0x31DFD72AE2903442ULL, "6173010_m.fxe"},
	{0x31DFDB13DF031FD1ULL, "6160677_m.fxe"},
	{0x31E0A7D667132738ULL, "89408137.ogg"},
	{0x31E0D8C13A4CD0EFULL, "and331ar_sewers.lst"},
	{0x31E16F2D14D7E5DDULL, "w_blightlands_61_0_ll.mmh"},
	{0x31E1C105BE36B8E4ULL, "6202138_m.fxe"},
	{0x31E1F8A5B6C384D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_r_tochest_alt.evt"},
	{0x31E20DE11A01C6B9ULL, "6092492_f.fxe"},
	{0x31E28A8931545CE9ULL, "241710842.ogg"},
	{0x31E30A3F03FEBEAFULL, "254223267.ogg"},
	{0x31E35AB6B3A34C31ULL, "gen00fl_merrill6146936.cl"},
	{0x31E37E40F76C35FCULL, "art\\levels\\ztools\\shaders\\blank_d.dds"},
	{0x31E3C641C4FAC595ULL, "vxht_keep_n_5i9_0_0.vlm"},
	{0x31E3DE8CDF0B82A8ULL, "6142629_m.fxe"},
	{0x31E4246731B1914EULL, "6099159_f.fxe"},
	{0x31E4639EE1E9BDAFULL, "6161034_m.fxe"},
	{0x31E4643DAA5B92D3ULL, "625134345.ogg"},
	{0x31E474E7CCC80E22ULL, "157515235.ogg"},
	{0x31E4803D3D539745ULL, "vxlt_gallowstemplar_n_47e_0_0.vlm"},
	{0x31E4F5098DF1DFB7ULL, "6087626_m.fxe"},
	{0x31E58370A602EC0BULL, "6079798_f.fxe"},
	{0x31E604051804C598ULL, "6104620_f.fxe"},
	{0x31E6C0017907373AULL, "873753360.ogg"},
	{0x31E6FE0B2416CEA6ULL, "prm000im_evra_ring.uti"},
	{0x31E7124E79DE71EFULL, "6036323_m.fxe"},
	{0x31E71A42AB5687A3ULL, "627453512.ogg"},
	{0x31E774BB0DB4A7F8ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml105a_0t.dds"},
	{0x31E7A558FE66D22DULL, "hlac_ht_0_hightown_n_0_3x.dds"},
	{0x31E80B9EA0D1CDC9ULL, "vxlt_0_lowtown_a2_n_c_i0x_0_0.vlm"},
	{0x31E8232B02C8B66BULL, "vxca_cavern1_haunted_edr_0_0.vlm"},
	{0x31E8250400A91325ULL, "vxdr_primevaltomb_4xq_0_0.vlm"},
	{0x31E8A91ECC8FB31EULL, "zz_dae_debug6049787.cl"},
	{0x31E8D458DF55203BULL, "vxlt_0_lowtown_n_eis_0_0.vlm"},
	{0x31E8EFA5B6C99461ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_r_tochest_alt.gad"},
	{0x31E98A5D11A71006ULL, "284599764.ogg"},
	{0x31E9FD8D5BFD8766ULL, "rdr111dg_martin6054832.cl"},
	{0x31EAD06F7B99AB37ULL, "vxw_blightlands_jj_0_0.vlm"},
	{0x31EAF889D550E631ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v3.ani"},
	{0x31EB29A314754048ULL, "793068171.ogg"},
	{0x31EB63A90A770E1EULL, "art\\vfx\\combat\\vfx_glb_sleep_c.dura.ani"},
	{0x31EC1919E3814CF6ULL, "art\\vfx\\models\\vfx_lightning_single_fat.msh"},
	{0x31EC2319E3815DF4ULL, "art\\vfx\\models\\vfx_lightning_single_fat.mmh"},
	{0x31EC2E9EE037F9F9ULL, "vxlt_0_lowtown_a3_n_c_hcy_0_0.vlm"},
	{0x31EC3664106744CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_down.evt"},
	{0x31ED842DEBDC5B50ULL, "278617214.ogg"},
	{0x31EDB466C6BECAC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_2hands.ani"},
	{0x31EDD77645C26641ULL, "vxht_chantry_2a9_0_0.vlm"},
	{0x31EE5766DCCE4212ULL, "gen_im_flw_arc_varric_04.uti"},
	{0x31EE725443353361ULL, "vxlt_0_lowtown_a2_e27_0_0.vlm"},
	{0x31EE815C1070F2FCULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyn_0t.dds"},
	{0x31EEA64136C38542ULL, "vxdr_primevaltomb_85u_0_0.vlm"},
	{0x31EED4842DEF3B62ULL, "hlac_ca_caverns1_mines_24423_1z.dds"},
	{0x31EEF0A5B6CE010FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_r_tochest_alt.ani"},
	{0x31EF2D97C0AEDBCFULL, "lt_0_lowtown_n_mmap.dds"},
	{0x31EF756F1E247B1AULL, "t1_mue_ye3.tnt"},
	{0x31F01E843F572A69ULL, "6153994_m.fxe"},
	{0x31F08ED46AB02FACULL, "707573870.ogg"},
	{0x31F0D0E9E65F02CEULL, "art\\characters\\weapons\\shields\\w_shd_kita.mmh"},
	{0x31F0DCE9E65F1765ULL, "art\\characters\\weapons\\shields\\w_shd_kita.mao"},
	{0x31F0EAE9E65F2EBCULL, "art\\characters\\weapons\\shields\\w_shd_kita.msh"},
	{0x31F1158C00E767EDULL, "ht_0_hightown_52_0_ll.mmh"},
	{0x31F20389D55717D3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v3.gad"},
	{0x31F215436EA1859EULL, "6119871_f.fxe"},
	{0x31F2A5D36837185DULL, "hlac_lt_0_lowtown_docks_-131076_1y.dds"},
	{0x31F2A916103411FAULL, "6103390_m.fxe"},
	{0x31F2E3F873B60629ULL, "vxlt_0_lowtown_docks_9nn_0_0.vlm"},
	{0x31F30B40CB57173AULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf.gad"},
	{0x31F30DD6547CA4B0ULL, "sky_w_encounter3.mao"},
	{0x31F31564106D2B3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_down.gad"},
	{0x31F3D6B84854070AULL, "gen00fl_anders6214652.cl"},
	{0x31F4BF66C6C4FC63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_2hands.gad"},
	{0x31F4E0E97742FC10ULL, "vxw_bonepitdestroyed_a3_1sj_0_0.vlm"},
	{0x31F4F4DADB804B75ULL, "6164026_m.fxe"},
	{0x31F525ED9136F3F2ULL, "w_bonepit_1_0.rml"},
	{0x31F54AEEF38F8C5CULL, "vxca_caverns2_78u_0_0.vlm"},
	{0x31F63114D41C14BDULL, "vxlt_gallowstemplar_n_3j5_0_0.vlm"},
	{0x31F6F819ADCBB0FCULL, "llac_ht_0_hightown_a2_131073_0.dds"},
	{0x31F6FAF3F2E1DFF8ULL, "239516921.ogg"},
	{0x31F7FB0A0489DF48ULL, "vxlt_0_lowtown_a2_e22_0_0.vlm"},
	{0x31F80089D55B7E38ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v3.evt"},
	{0x31F82B6FA46C2BCEULL, "sse_ss_zevran.gda"},
	{0x31F84AD878622DD8ULL, "vxlt_house_gi_0_0.vlm"},
	{0x31F8C552A675AC81ULL, "vxdr_primevaltomb_4hm_0_0.vlm"},
	{0x31F933B267345D76ULL, "6208004_m.fxe"},
	{0x31F9424C6D8FC3B7ULL, "vxw_blightlands_oe_0_0.vlm"},
	{0x31F94B81E24CCAB6ULL, "zz_ave_debug6056769.cl"},
	{0x31F97BA90A8374EFULL, "art\\vfx\\combat\\vfx_glb_sleep_c.dura.evt"},
	{0x31F9CA7C31C4C78FULL, "vxlt_undercityencounter_121_0_1.vlm"},
	{0x31FA0E40CB5D3B51ULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf.evt"},
	{0x31FA15E3A05E2924ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hallastatue_d.dds"},
	{0x31FA2D23EF57DA2EULL, "hlac_w_bonepitmining_a2_327685_0y.dds"},
	{0x31FA420286B5EC94ULL, "combinedhud.gfx"},
	{0x31FA7ED6213207D6ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.cess.ani"},
	{0x31FAD9BC6C533FFEULL, "6144572_f.fxe"},
	{0x31FB0617B2FAEECDULL, "182751391.ogg"},
	{0x31FBC5D67ABFAD8AULL, "hlac_lt_0_lowtown_-131072_0z.dds"},
	{0x31FC495337B87EC1ULL, "vxht_chantry_2z8_0_0.vlm"},
	{0x31FC5F9EC351DBD8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_coma_0s.dds"},
	{0x31FCDD522DB85B73ULL, "art\\characters\\playercharacter\\humanmale\\glv_magc.phy"},
	{0x31FCE436ACDEA292ULL, "mcr310_party_banter6087818.cl"},
	{0x31FCF76C124527E1ULL, "zz_dre_debug6168237.cl"},
	{0x31FD014012DCD829ULL, "vxca_caverns1_mines_edh_0_0.vlm"},
	{0x31FD090CF14B46E9ULL, "vxlt_gallowscourtyard_7au_0_0.vlm"},
	{0x31FD6325FDFE8CB1ULL, "6085280_m.fxe"},
	{0x31FD950F9432A837ULL, "dre000_bodahn6064146.cl"},
	{0x31FDE451DD8B8F40ULL, "hlac_ht_keep_n_3732_15y.dds"},
	{0x31FDEDB1622CD898ULL, "20004174.ogg"},
	{0x31FE4858C824ABE7ULL, "vxw_bonepit_1ic_0_0.vlm"},
	{0x31FE9A8F95B32830ULL, "6148090_m.fxe"},
	{0x31FE9FC774A35ACCULL, "527087474.ogg"},
	{0x31FEF5B69B61C35FULL, "vxht_0_hightown_bao_0_0.vlm"},
	{0x31FF7CA90A87E198ULL, "art\\vfx\\combat\\vfx_glb_sleep_c.dura.gad"},
	{0x3200014F5198DA58ULL, "dae200_saemus6091708.cl"},
	{0x32004B850B160E70ULL, "qun131_taskmaster6059010.cl"},
	{0x32004DB63BE843FAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spike.mmh"},
	{0x320051B63BE84A31ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spike.mao"},
	{0x3200597BCAB34C0DULL, "gen00fl_aveline6137469.cl"},
	{0x32005FB63BE86240ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spike.msh"},
	{0x3200863F9AD7BDC5ULL, "vxht_0_hightown_n_byb_0_0.vlm"},
	{0x3200E0200800FE9EULL, "vxdr_primevalentrance_1id_0_0.vlm"},
	{0x32011AF3022A2A7AULL, "6104290_m.fxe"},
	{0x32011C6C4067FC86ULL, "6162288_mother_shh.cut"},
	{0x320121D41C616E39ULL, "179902302.ogg"},
	{0x32013371F40679C3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masc_0n.dds"},
	{0x32014A9C0D74DE9AULL, "6160898_m.fxe"},
	{0x3201879C61620A17ULL, "hf_amb_keep_noble.mor"},
	{0x320195E8EDF457BDULL, "art\\characters\\playercharacter\\humanmale\\boo_magb.phy"},
	{0x3201C72B0CB039F7ULL, "vxht_0_hightown_c92_0_0.vlm"},
	{0x3202B822673D5989ULL, "6163210_m.fxe"},
	{0x3202B864E14D721BULL, "6126817_m.fxe"},
	{0x320305C79CDA24D0ULL, "w_encounter3_227_0.msh"},
	{0x320313C79CDA3CEAULL, "w_encounter3_227_0.mmh"},
	{0x320317C79CDA4321ULL, "w_encounter3_227_0.mao"},
	{0x320340A42F679BA9ULL, "mag371_elsa6065013.cl"},
	{0x32036D7CC6ACBF7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_ex.ani"},
	{0x320376180952139AULL, "vxht_keep_n_642_0_0.vlm"},
	{0x320394138291740AULL, "gen00fl_varric6033695.cl"},
	{0x3203EE2F9702E686ULL, "art\\vfx\\combat\\vfx_r_var_bayonet_v5_c.mmh"},
	{0x32042F14F86323EBULL, "vxht_chantry_n_2cb_0_0.vlm"},
	{0x32044604080E9A63ULL, "llac_w_encounter2_camp_393215_0.dds"},
	{0x32045573E4825AE9ULL, "6110384_m.fxe"},
	{0x3204D0FC7A09FF8FULL, "framebuffereffects_range_dof_p0.psh.fxo"},
	{0x32050668F7FA41F1ULL, "181308800.ogg"},
	{0x32050B480134A11BULL, "vxlt_hangedman_n_16x_0_0.vlm"},
	{0x32052BC4B6539334ULL, "6140289_m.fxe"},
	{0x3205302AC7CD7D0AULL, "hlac_ca_caverns2prc_5802_48x.dds"},
	{0x320557936C31918FULL, "vxht_brothel_n_6gi_0_0.vlm"},
	{0x3205AC7D6CF35C14ULL, "vxca_caverns1_fmx_0_0.vlm"},
	{0x3205F658F750E190ULL, "framebuffereffects_hslmatrix_p0.vsh.fxo"},
	{0x32063C4487D4F1E5ULL, "942064668.ogg"},
	{0x320641C3D58EE06EULL, "vxht_0_hightown_c53_0_0.vlm"},
	{0x32069586A1E764E4ULL, "atl_startmenu_dxt5_06.dds"},
	{0x3206D2B29233E765ULL, "vxht_keep_n_5dj_0_0.vlm"},
	{0x3206D7D1869A9FA3ULL, "207545998.ogg"},
	{0x3206F16BA1898942ULL, "dae000ip_crowd_mf6_talk11.utp"},
	{0x32072640CB67EE20ULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf.ani"},
	{0x32072BE370CF5727ULL, "llac_w_blightlands_196608_0.dds"},
	{0x32072E87D8B2CF69ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_explo_c.impa.ani"},
	{0x320786D6213C9FC7ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.cess.evt"},
	{0x32079C3F4598E45CULL, "art\\characters\\creatures\\dragon\\animation\\die_lp_v1.gad"},
	{0x3207B4295F85DC37ULL, "vxw_blightlands_pd_0_0.vlm"},
	{0x3207C0C8F971428BULL, "100084161.ogg"},
	{0x3207CCF70B11B516ULL, "396955786.ogg"},
	{0x3207E1586FCBA8CCULL, "vxlt_warehouse_n_14i_0_0.vlm"},
	{0x320840C905DF0FE7ULL, "hlac_lt_gallowstemplar_f_2384_1x.dds"},
	{0x32087A832E90A8ACULL, "vxlt_0_lowtown_a2_n_e2u_0_0.vlm"},
	{0x320895BF85771A20ULL, "atl_worldmap_dxt5_dat.xml"},
	{0x3208BA7D0286BE41ULL, "18584762.ogg"},
	{0x3208E4E76972F8C8ULL, "vxca_caverns1_iki_0_0.vlm"},
	{0x320980FB6E4D8084ULL, "zz_dae_debug6045828.cl"},
	{0x320A7A7CC6B2F3BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_ex.gad"},
	{0x320A7FB105662E62ULL, "genip_node_e2.utp"},
	{0x320A837252020352ULL, "116709336.ogg"},
	{0x320ABAAE772344EAULL, "mer341_jansen6082996.cl"},
	{0x320AEB0A26CE71C5ULL, "885608608.ogg"},
	{0x320B0AB84DB9ED2EULL, "6148236_m.fxe"},
	{0x320B560E27216461ULL, "vxlt_0_lowtown_a2_e85_0_0.vlm"},
	{0x320BFD5894983C42ULL, "mag210_arianni6030497.cl"},
	{0x320C452572DCBF70ULL, "6056726_m.fxe"},
	{0x320C4DAA2E824A6BULL, "531224990.ogg"},
	{0x320C699446135459ULL, "531121947.ogg"},
	{0x320C69BFB1FDC160ULL, "gen00fl_anders6111173.cl"},
	{0x320D9B3F459D4E5AULL, "art\\characters\\creatures\\dragon\\animation\\die_lp_v1.ani"},
	{0x320DCE0E9A146561ULL, "vxca_caverns1_inc_0_0.vlm"},
	{0x320E04FD7261BD6AULL, "6164353_m.fxe"},
	{0x320E3B7623300CC2ULL, "mcr310_argument6182306.cl"},
	{0x320E4987D8B91BEBULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_explo_c.impa.gad"},
	{0x320E5A7555CEB9CCULL, "vxht_0_hightown_a2_9aw_0_0.vlm"},
	{0x320E5AD44A174DF1ULL, "1014437456.ogg"},
	{0x320EA7D62142F610ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.cess.gad"},
	{0x320F232EFDB93C99ULL, "gen00fl_bethany6097591.cl"},
	{0x320F48FA0F895FD0ULL, "6059244_m.fxe"},
	{0x320F4B6E5F000BDCULL, "zz_ave_debug6207419.cl"},
	{0x32100F96B00B71ACULL, "w_blightlands_43_0.msh"},
	{0x32101596B00B7B9EULL, "w_blightlands_43_0.mmh"},
	{0x32102196B00B9035ULL, "w_blightlands_43_0.mao"},
	{0x3210531E461DB3CDULL, "vxlt_0_lowtown_a2_n_c_bz5_0_0.vlm"},
	{0x321111740C5D094BULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.mmh"},
	{0x3211280794BF0BBEULL, "qun221_petrice6083502.cl"},
	{0x3211557CC6B8D3DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_ex.evt"},
	{0x3211B46223BF5424ULL, "llac_lt_0_lowtown_a3_n_c_1_4.dds"},
	{0x3211E1E4F08DAE5AULL, "340178359.ogg"},
	{0x3211FF9C35F2C788ULL, "vxht_0_hightown_a3_bdz_0_0.vlm"},
	{0x32123B28B7C560EDULL, "6055601_m.fxe"},
	{0x32128528603E9A13ULL, "1071617383.ogg"},
	{0x32128AE8EE02B829ULL, "art\\characters\\playercharacter\\humanmale\\boo_magb.mao"},
	{0x32129116E3BA8B0BULL, "art\\levels\\races\\proxy\\_textures\\ter_3_d.dds"},
	{0x321296E8EE02CCC2ULL, "art\\characters\\playercharacter\\humanmale\\boo_magb.mmh"},
	{0x321298E8EE02D078ULL, "art\\characters\\playercharacter\\humanmale\\boo_magb.msh"},
	{0x3212A1BFA24B5BC3ULL, "hlac_lt_gallowstemplar_f_2383_6x.dds"},
	{0x32132DECC60D0E10ULL, "vxca_cavern1_haunted_i1n_0_0.vlm"},
	{0x321430495A28B37BULL, "6103412_m.fxe"},
	{0x321450A64407C195ULL, "151441144.ogg"},
	{0x32145C3F3B32C71AULL, "155975309.ogg"},
	{0x32147487EC05BFC2ULL, "44337832.ogg"},
	{0x3214B8A96337916DULL, "vxht_0_hightown_c1x_0_0.vlm"},
	{0x3214D1F9AC877F3DULL, "888152827.ogg"},
	{0x3214DE03CB4E4867ULL, "gen00fl_sebastian6180489.cl"},
	{0x3214DEC3E8761048ULL, "vxlt_0_lowtown_a2_e80_0_0.vlm"},
	{0x3214F1ACAA467D6EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo12_basic.phy"},
	{0x32150C646AC1C3CBULL, "700150873.ogg"},
	{0x3215448F8429FFE5ULL, "vxht_keep_n_5ii_0_0.vlm"},
	{0x32154687D8BF3550ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_explo_c.impa.evt"},
	{0x32154FBB54A405F7ULL, "w_bonepitmining_a2_117_0.mmh"},
	{0x321553BB54A40C04ULL, "w_bonepitmining_a2_117_0.mao"},
	{0x321565BB54A42B69ULL, "w_bonepitmining_a2_117_0.msh"},
	{0x3216073679E09044ULL, "str_clo.mao"},
	{0x3216137368813ABAULL, "vxlt_0_lowtown_a2_n_ekj_0_2.vlm"},
	{0x321614EFC12F2905ULL, "970394302.ogg"},
	{0x32161CBD19C68CCDULL, "hlac_ca_cavern1_haunted_24422_2y.dds"},
	{0x32162606517BF4B7ULL, "vxw_blightlands_ug_0_0.vlm"},
	{0x3216BD5FA277417FULL, "gen00fl_dog6116491.cl"},
	{0x3216C4DF246336DCULL, "vxht_keep_60v_0_1.vlm"},
	{0x321714BBC6735FEFULL, "fen110_fenris.bnk"},
	{0x321756C45B691148ULL, "vxca_caverns1_inf_0_0.vlm"},
	{0x32175829D5EA9EE8ULL, "hlac_lt_0_lowtown_docks_a3_-3_1z.dds"},
	{0x321782DF13161192ULL, "259830070.ogg"},
	{0x3218092AFED072C6ULL, "ser270_templar6114629.cl"},
	{0x32184043C29D3B62ULL, "llac_lt_0_lowtown_a2_n_65538_1.dds"},
	{0x3218B91C874983F7ULL, "6048808_f.fxe"},
	{0x32191E6E395989DDULL, "lt_0_lowtown_a2_n_c_water_21475.mao"},
	{0x3219226E39599016ULL, "lt_0_lowtown_a2_n_c_water_21475.mmh"},
	{0x32192C6E3959A114ULL, "lt_0_lowtown_a2_n_c_water_21475.msh"},
	{0x321991E13339BEFEULL, "6139496_m.fxe"},
	{0x3219C64A410A174AULL, "vxlt_gallowstemplar_n_3yo_0_0.vlm"},
	{0x3219E1B5EF385DF1ULL, "6134579_m.fxe"},
	{0x3219FF9B1765638FULL, "hlac_lt_backalley_n_65537_4z.dds"},
	{0x321A2AB729E4DD0EULL, "hlac_ca_caverns2_5802_50x.dds"},
	{0x321A36BBC675CF57ULL, "fen110_fenris.cnv"},
	{0x321B284F9AA3C65AULL, "vxw_bonepit_28f_0_0.vlm"},
	{0x321BBAD01D0A805BULL, "follower_banter.bnk"},
	{0x321BC0F4FADCBA4EULL, "ht_keep_n_3735.anb"},
	{0x321BD9707F4DF45FULL, "vxht_0_hightown_bwi_0_0.vlm"},
	{0x321BF97C085A56AFULL, "hm_rdr_guard_warehouse.mor"},
	{0x321C87391EE47E77ULL, "dae011_gossips.bnk"},
	{0x321CB4EA38EF03E0ULL, "vxht_bartrand_n_6pk_0_0.vlm"},
	{0x321D276015E5A857ULL, "hlac_w_bonepit_0_5x.dds"},
	{0x321D4BD5BE911B8BULL, "vxht_0_hightown_n_c3c_0_0.vlm"},
	{0x321D63248943D1A7ULL, "lowtown_male_lowlife1_sta6164942.cl"},
	{0x321DA7827A46BB98ULL, "hlac_dr_0_deeproads_5845_11x.dds"},
	{0x321DEB552AF0256BULL, "hlac_lt_0_lowtown_a2_65535_1y.dds"},
	{0x321E80E9C0137254ULL, "art\\characters\\creatures\\ragedemon\\animation\\enter.gad"},
	{0x321E9D35FB06B929ULL, "6165256_m.fxe"},
	{0x321EF2740C6911B0ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.anb"},
	{0x321F189BE6A08E67ULL, "409468101.ogg"},
	{0x321F4DFD8EB69525ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_106a_n.dds"},
	{0x321F69391EE6808FULL, "dae011_gossips.cnv"},
	{0x321F697EAE972BBAULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_04brgdebris.msh"},
	{0x321F777EAE9743F3ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_04brgdebris.mao"},
	{0x321F77B491865106ULL, "cs_dre_18_cs_ragstoriches.bnk"},
	{0x321F7B7EAE974A00ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_04brgdebris.mmh"},
	{0x321F9466D0FE894AULL, "47072864.ogg"},
	{0x321FBCD01D0E6C1BULL, "follower_banter.cnv"},
	{0x321FE0C96D12FA8CULL, "6102697_f.fxe"},
	{0x3220006FA855573FULL, "dae000ip_scavenge_d.utp"},
	{0x32205E5C81E1D81FULL, "and100_anders6111682.cl"},
	{0x3220C85A30E91EA7ULL, "6051272_m.fxe"},
	{0x32214462B5EA48A9ULL, "hlac_dr_dungeon_3214_14x.dds"},
	{0x32224B15F64DD712ULL, "vxlt_undercityserialkiller_11p_0_0.vlm"},
	{0x3222833DA86A5861ULL, "vxht_0_hightown_n_bz3_0_0.vlm"},
	{0x322293A829A8A335ULL, "vxht_mansion2b_3cn_0_0.vlm"},
	{0x3223092736CF9E8FULL, "gen00fl_anders6121175.cl"},
	{0x3223265D349AFD27ULL, "plt_lgt000pt_baz_dirty.plo"},
	{0x32232A86552DA9EDULL, "vxht_0_hightown_c4w_0_0.vlm"},
	{0x322360811350A796ULL, "qun110_petrice6042153.cl"},
	{0x32239205B9D0EBA9ULL, "vxht_0_hightown_a2_bym_0_0.vlm"},
	{0x322399976305F936ULL, "hlac_lt_0_lowtown_a3_n_c_1_7y.dds"},
	{0x32239C566FAEA790ULL, "6016609_m.fxe"},
	{0x3223F99F926FF184ULL, "dae000_party_night_comment6201649.cl"},
	{0x3224035066F62D96ULL, "vxht_mansion2_9d_0_0.vlm"},
	{0x322482E6BC458973ULL, "846537783.ogg"},
	{0x3224868324E9B447ULL, "llac_lt_0_lowtown_-65535_1.dds"},
	{0x3224D6638A62A405ULL, "ht_0_hightown_a3_51_0.rml"},
	{0x3224F158834639A9ULL, "6078169_m.fxe"},
	{0x322508A36E2ECA17ULL, "16251970.ogg"},
	{0x32255E3D127CD9ACULL, "vxlt_0_lowtown_a2_n_e8k_0_0.vlm"},
	{0x32256CB63C079405ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spike.phy"},
	{0x32258FE9C019AA32ULL, "art\\characters\\creatures\\ragedemon\\animation\\enter.ani"},
	{0x3225E342A450716AULL, "6135712_f.fxe"},
	{0x322645005A6F555CULL, "mag360_samson6185088.cl"},
	{0x32266889C42816D9ULL, "art\\characters\\creatures\\mubari\\animation\\tn90l.ani"},
	{0x3226C846C4EAE1A6ULL, "dae200_saemus6091715.cl"},
	{0x3227494F8C71508EULL, "art\\characters\\creatures\\shriek\\textures\\cn_bdy_ems_0s.dds"},
	{0x32276EADB3C913E6ULL, "hlac_ht_0_hightown_a3_-131072_1z.dds"},
	{0x3227BD9BF590AC2EULL, "dae200cr_viscount6098514.cl"},
	{0x32281A126B4E5314ULL, "vrc321_varric6154759.cl"},
	{0x3228A3AB4224B4E6ULL, "dm_lgt_urn_conleader.mor"},
	{0x3228FDD1076C3D7DULL, "579644073.ogg"},
	{0x3229A038DAF0A5E6ULL, "mrl150_dalish_guard6068311.cl"},
	{0x322A17D85B1BD18EULL, "6034218_m.fxe"},
	{0x322A3F6F637D11A9ULL, "359129756.ogg"},
	{0x322A4B4D71440CDFULL, "vxht_0_hightown_bph_0_0.vlm"},
	{0x322A695F6C1F84D6ULL, "6066518_m.fxe"},
	{0x322A835003B50159ULL, "6088855_f.fxe"},
	{0x322ACD219D6E875CULL, "6076067_m.fxe"},
	{0x322AE02D43E7E556ULL, "gen00fl_anders6111315.cl"},
	{0x322AFF5F311AA552ULL, "601021013.ogg"},
	{0x322B33B39BABE62BULL, "rdr111dg_martin6015151.cl"},
	{0x322BBDB2B087340BULL, "vxht_0_hightown_n_c8b_0_0.vlm"},
	{0x322C95D32A3876F7ULL, "gen00fl_aveline6128984.cl"},
	{0x322CBBE623606054ULL, "gen_im_wep_rog_dua_act3_08.uti"},
	{0x322CC635717448DBULL, "hc_fire_blue.mao"},
	{0x322CD6B867C889C8ULL, "llac_ht_playershouse_n_855_34.dds"},
	{0x322D3BAE73AB6F8FULL, "mag320_nobleman_26060267.cl"},
	{0x322D4389C42DF61BULL, "art\\characters\\creatures\\mubari\\animation\\tn90l.gad"},
	{0x322DDC4DF255E16CULL, "character_rendertodepth_p0.psh.fxo"},
	{0x322FD7D8CAEA610AULL, "one121_trap6112686.cl"},
	{0x322FF3DC94FE9307ULL, "6177679_m.fxe"},
	{0x3230244ED82AD126ULL, "hlac_w_woundedcoast_-65534_0y.dds"},
	{0x323051458B575624ULL, "gen_im_wep_rog_dua_act1_02.uti"},
	{0x3230A5A16D5161A4ULL, "lowtown_fem_alien_elf_2_mob6166338.cl"},
	{0x3230F51A9A6070E1ULL, "vxht_0_hightown_n_be2_0_0.vlm"},
	{0x323103C6B82DD986ULL, "hlac_w_woundedcoast_196616_0z.dds"},
	{0x323105851B9EBBB5ULL, "vxht_mansion2b_3fm_0_0.vlm"},
	{0x323112EA28FF4EFDULL, "vxlt_darktown_5bp_0_0.vlm"},
	{0x32311A39430C5436ULL, "vxdr_primevalentrance_1j9_0_0.vlm"},
	{0x32322F63A13EB1E9ULL, "6071023_m.fxe"},
	{0x323244A2FCAC81EEULL, "fxa_cen_volt2.dds"},
	{0x32325A69E9489DB6ULL, "vxlt_0_lowtown_docks_c5f_0_0.vlm"},
	{0x3232A6335C53B6D9ULL, "cloud_w.mao"},
	{0x3232E8CD32996E60ULL, "6138765_m.fxe"},
	{0x3232EC99BECE5565ULL, "follower_banter6163491.cl"},
	{0x3232F72D4C6D6BBAULL, "vxlt_0_lowtown_a2_n_eqh_0_2.vlm"},
	{0x3232FD986B3E5DD5ULL, "6043619_m.fxe"},
	{0x323315BAD58CDB71ULL, "vxht_0_hightown_b1a_0_0.vlm"},
	{0x323356A44707DC83ULL, "6129463_m.fxe"},
	{0x323369B01C7B98BDULL, "1058956379.ogg"},
	{0x3233E87CC4EC287EULL, "142249875.ogg"},
	{0x32346089C4344600ULL, "art\\characters\\creatures\\mubari\\animation\\tn90l.evt"},
	{0x3234654B48399878ULL, "6100353_m.fxe"},
	{0x3235064472708611ULL, "vxw_bonepit_1rt_0_0.vlm"},
	{0x323563B9FAFCD6E4ULL, "dae000_female_tranquil6165123.cl"},
	{0x32356F935083CFEBULL, "vxca_cave_183_0_0.vlm"},
	{0x3235AE3F00B28661ULL, "art\\vfx\\combat\\vfx_m_bld_conj_c.anb"},
	{0x3235BC2E80A9E0BFULL, "hlac_lt_hangedman_n_-1_9z.dds"},
	{0x3235CDFD942FE995ULL, "mcr370_meredith6083122.cl"},
	{0x32367D6DE891B9DAULL, "vxlt_darktown_531_0_1.vlm"},
	{0x323699F19D064103ULL, "vxlt_0_lowtown_f1g_0_0.vlm"},
	{0x3236D2FB6E032C21ULL, "6143304_m.fxe"},
	{0x3236F01D7D27D400ULL, "character_character_p144.psh.fxo.dx11"},
	{0x323709040F842155ULL, "w_bonepitdestroyed_a3_142_0.msh"},
	{0x32370B040F8424CBULL, "w_bonepitdestroyed_a3_142_0.mmh"},
	{0x323715F8E4457C2AULL, "6052096_m.fxe"},
	{0x323717040F843968ULL, "w_bonepitdestroyed_a3_142_0.mao"},
	{0x32376A84C162B5ADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_smallcap.msh"},
	{0x32377884C162CDE0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_smallcap.mao"},
	{0x32377C84C162D433ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_smallcap.mmh"},
	{0x3237B40128B35E2FULL, "genip_sword.utp"},
	{0x3237D4BD09F14374ULL, "static_static_p14.psh.fxo.dx11.assembly"},
	{0x3237ECB0F997FF8CULL, "character_character_p76.psh.fxo.dx11"},
	{0x323889B54D11EA0EULL, "6031118_m.fxe"},
	{0x3238A1982B1E5244ULL, "6201513_m.fxe"},
	{0x3238DFE80C590AA2ULL, "one100dg_lia6049837.cl"},
	{0x3238E39D9F4222A1ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation3.gad"},
	{0x323961A9B3B19CC0ULL, "gen00fl_sebastian6177882.cl"},
	{0x3239D818E97291E5ULL, "fen110st_fenris_mansion.stg"},
	{0x323A65103B4A7AAFULL, "6078435_m.fxe"},
	{0x323A6AEDDE371442ULL, "vxca_cavern1_haunted_i2q_0_0.vlm"},
	{0x323A780E24B8209AULL, "hlac_lt_0_lowtown_n_-1_4z.dds"},
	{0x323A8035975A6FD4ULL, "vxht_chantry_n_330_0_0.vlm"},
	{0x323A96EC197A5341ULL, "gen00flst_isa_act2.stg"},
	{0x323A9B773FB41E2DULL, "539669454.ogg"},
	{0x323AF827AE4A4310ULL, "vxdr_primevaltomb_8nj_0_0.vlm"},
	{0x323B282E1D96853DULL, "hlac_w_sundermount_0_3y.dds"},
	{0x323B2A795B201EB4ULL, "364699287.ogg"},
	{0x323BEADBDCD4D4B6ULL, "w_bonepitdestroyed_a3_55_0.msh"},
	{0x323BF4DBDCD4E5B4ULL, "w_bonepitdestroyed_a3_55_0.mmh"},
	{0x323BF8DBDCD4ECC7ULL, "w_bonepitdestroyed_a3_55_0.mao"},
	{0x323C43C7961B8272ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsb_0s.dds"},
	{0x323C568FB5B3E986ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.dura.ani"},
	{0x323C5752B1D1261DULL, "6127662_f.fxe"},
	{0x323C59FE7E11C24EULL, "hlac_lt_0_lowtown_docks_a3_-1_1z.dds"},
	{0x323D1A3A31117690ULL, "vxw_bonepitdestroyed_a3_1hw_0_0.vlm"},
	{0x323D1BDE6BF558A8ULL, "plt_genpt_app_aveline.plo"},
	{0x323D33AD010520ACULL, "626999098.ogg"},
	{0x323DD99D32BC3133ULL, "829518969.ogg"},
	{0x323EA263666FC040ULL, "dae000_meredith6080984.cl"},
	{0x323FB3D4E2A34F15ULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\weapons.mao"},
	{0x323FD6C3994100EBULL, "6098600_m.fxe"},
	{0x323FE49D9F48434FULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation3.ani"},
	{0x323FE6BBED7DEFA6ULL, "hlac_lt_0_lowtown_a2_n_-131072_1z.dds"},
	{0x323FEB5EC4BF1774ULL, "55004255.ogg"},
	{0x323FF41D501F61EAULL, "vxlt_hangedman_10l_0_0.vlm"},
	{0x32407CE8F9BDDE93ULL, "vxlt_0_lowtown_docks_a3_cze_0_0.vlm"},
	{0x3240AA294B10BA1BULL, "dr_0_deeproads_mmi.gff"},
	{0x3240D520118C74E6ULL, "vxdr_primevalentrance_16u_0_0.vlm"},
	{0x3240EEDAC6950D69ULL, "vxw_bonepitdestroyed_a3_1vp_0_0.vlm"},
	{0x3240F566F2C8E529ULL, "follower_banter6149288.cl"},
	{0x324123B89C2A6A19ULL, "mag100_thrask6034815.cl"},
	{0x32414CB6AFD6BC90ULL, "hlac_lt_0_lowtown_a2_n_-65536_3y.dds"},
	{0x3241AF1B0A9420CAULL, "6162082_f.fxe"},
	{0x3241C6A2E5E01B31ULL, "933817547.ogg"},
	{0x32421F41FE1ECDACULL, "ser200_aveline6201525.cl"},
	{0x32425D3EAD15852EULL, "6117969_f.fxe"},
	{0x32427428D271E51BULL, "qcr000_arishok6059306.cl"},
	{0x32428F3F00BCDBDAULL, "art\\vfx\\combat\\vfx_m_bld_conj_c.mmh"},
	{0x32429F47CBD1D863ULL, "zz_ave_debug6064103.cl"},
	{0x3242E4408C5F7CF9ULL, "vxht_brothel_n_6e5_0_0.vlm"},
	{0x3242E7EFBE9F1938ULL, "art\\characters\\playercharacter\\elfmale\\arm_lgte.phy"},
	{0x3242FFC689A2C09FULL, "304232128.ogg"},
	{0x324309D221E8FF63ULL, "vxht_0_hightown_a2_65g_0_0.vlm"},
	{0x324434076678093CULL, "887506705.ogg"},
	{0x32449A19676B751FULL, "vxlt_gallowstemplar_n_4xz_0_0.vlm"},
	{0x324503E23C2763FBULL, "6166985_m.fxe"},
	{0x324528AE80346130ULL, "gen_im_trs_krk_grd_junk1_04.uti"},
	{0x324589ACB2D4A6E0ULL, "69196737.ogg"},
	{0x3245F5C4BAE13BEAULL, "6118753_f.fxe"},
	{0x32466E23383E8FFEULL, "vxlt_gallowstemplar_f_59y_0_0.vlm"},
	{0x3246AF9E73AEA4C1ULL, "llac_lt_0_lowtown_a2_-65537_3.dds"},
	{0x324781B812D2B8E4ULL, "sb_noise.dds"},
	{0x324783B68C0F4503ULL, "hlac_lt_0_lowtown_a2_-131072_5z.dds"},
	{0x3247A021DA831749ULL, "dae000_leonato6162489.cl"},
	{0x3248C8A229B56FBFULL, "zz_vrc_debug6200254.cl"},
	{0x3248E34E43A2DA34ULL, "lgt250pt_way_dlsh_cnflc.ncs"},
	{0x3248FBBA944A484BULL, "vxdr_primevaltomb_52i_0_1.vlm"},
	{0x324937781CEA2525ULL, "llac_lt_0_lowtown_docks_n_-196607_0.dds"},
	{0x3249393F2B528F35ULL, "water_water_p0.vsh.fxo.dx11"},
	{0x32497B71DB86CF6EULL, "gen_im_trs_dal_med_hlm_r0.uti"},
	{0x3249B4E8CA8EA549ULL, "zz_pro_debug6007864.cl"},
	{0x324A3E8FB5BFFEF7ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.dura.evt"},
	{0x324A58BE1E3BEA4CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_lp.ani"},
	{0x324A5A616127F11DULL, "vxca_caverns2_43g_0_1.vlm"},
	{0x324A6A98F1BC1CA9ULL, "289922027.ogg"},
	{0x324A841FFE8D2CF7ULL, "666735123.ogg"},
	{0x324AD7B15481AF7DULL, "518717590.ogg"},
	{0x324AFB4E5A880C59ULL, "hlac_ht_0_hightown_a3_-2_0z.dds"},
	{0x324B1BC3B36607CEULL, "1022842988.ogg"},
	{0x324B8765296B0372ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\walk_forward_1_step.gad"},
	{0x324B8C1723078F10ULL, "vxw_bonepitdestroyed_a3_1mt_0_0.vlm"},
	{0x324BAAEE4FF97DA3ULL, "mcr310_argument6074435.cl"},
	{0x324BF8A245A878E4ULL, "gen00fl_merrill6162245.cl"},
	{0x324C55678CE3B8A4ULL, "zz_serpt_debug.ncs"},
	{0x324CB3BBACA74BEAULL, "vxw_encounter3_g9_0_1.vlm"},
	{0x324D30E3A32B2DFEULL, "6092607_m.fxe"},
	{0x324D39361B91C585ULL, "117216834.ogg"},
	{0x324D5CC4BBFF4BC8ULL, "follower_banter6160857.cl"},
	{0x324DE781EE1B9BD4ULL, "llac_dr_slavers_3214_2.dds"},
	{0x324E10E81633928AULL, "follower_banter6169720.cl"},
	{0x324EC3F4FB0818B2ULL, "ht_keep_n_3735.rml"},
	{0x324F144ACC55FDA2ULL, "398195968.ogg"},
	{0x324F416095F20AF7ULL, "character_character_p196.psh.fxo.dx11"},
	{0x324F4FC939B7C1C3ULL, "vxht_0_hightown_a3_akc_0_0.vlm"},
	{0x324F9DC78417D392ULL, "one350_variel.cnv"},
	{0x32501B6B41F769EEULL, "398216964.ogg"},
	{0x32503B6405A062A1ULL, "hlac_lt_darktown_-1_34z.dds"},
	{0x32503F8FB5C46B80ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.dura.gad"},
	{0x32504AD7DB20F6B3ULL, "6031278_m.fxe"},
	{0x3250A2F574B7A26EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_pleading_up.evt"},
	{0x3250E83185D4A3ABULL, "6173161_m.fxe"},
	{0x32511239E1C09829ULL, "6127019_f.fxe"},
	{0x325124BFBA61846AULL, "vxht_mansion_n_52o_0_0.vlm"},
	{0x325149B595DE5DE4ULL, "6078707_f.fxe"},
	{0x32515FE795CE7F31ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsfloor.mmh"},
	{0x325165BE1E421E8EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_lp.gad"},
	{0x32516BE795CE93D2ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsfloor.mao"},
	{0x325179E795CEABEFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsfloor.msh"},
	{0x3252385CB1E7AF76ULL, "vxht_mansion_n_33b_0_0.vlm"},
	{0x32528A65297126A9ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\walk_forward_1_step.evt"},
	{0x3252B9D62FC06F9FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_right.gad"},
	{0x32532D386B964F66ULL, "vxht_0_hightown_a2_b8a_0_0.vlm"},
	{0x32538E1A7D70F46AULL, "gen_im_trs_cot_lgt_boo_r1.uti"},
	{0x32539E7FD9DCED1BULL, "6162707_m.fxe"},
	{0x3253ADD7E269F191ULL, "425770831.ogg"},
	{0x3253BFC7841BF6EAULL, "one350_variel.bnk"},
	{0x3253CB1B56006716ULL, "6170752_m.fxe"},
	{0x3254A2A9ED4ACF34ULL, "vxw_blightlands_11u_0_0.vlm"},
	{0x3255F0ACAA7E1231ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo12_basic.mmh"},
	{0x3255FCACAA7E26D2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo12_basic.mao"},
	{0x325600C974B46B40ULL, "ht_brothel_n_3367_ll.mmh"},
	{0x32560AACAA7E3EEFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo12_basic.msh"},
	{0x3256283AD0B20BCDULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgtb_0d.dds"},
	{0x3256722299B9429CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_right_2.gad"},
	{0x325699F574BBFE19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_pleading_up.gad"},
	{0x3256B3518A31DEF6ULL, "860198518.ogg"},
	{0x325760BE1E46826DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_lp.evt"},
	{0x3257DBE192367410ULL, "vxdr_primevaltomb_8dh_0_0.vlm"},
	{0x3257F49293909E94ULL, "6092178_f.fxe"},
	{0x3257F596B048FB55ULL, "w_blightlands_43_0.rml"},
	{0x325837D5E0440E7AULL, "hlac_ca_caverns2_5802_26y.dds"},
	{0x325854BC650CCBA4ULL, "hlac_lt_warehouse_65536_2z.dds"},
	{0x32588937B2F8B1B8ULL, "vxdr_primevaltomb_55g_0_1.vlm"},
	{0x3258B0D3644E41B8ULL, "llac_w_woundedcoast_-131067_0.dds"},
	{0x32592DED1883CE11ULL, "art\\vfx\\cinematic\\vfx_shield_break_p.impa.ani"},
	{0x32595F03DEE7774DULL, "6125142_m.fxe"},
	{0x325972465213C6DFULL, "604190735.ogg"},
	{0x3259AB5DFD1E2AF2ULL, "art\\vfx\\combat\\vfx_r_gn2_hbomb_act_c.impa.gad"},
	{0x3259AED62FC67B8CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_right.evt"},
	{0x3259EFF8D0AD0D25ULL, "vrc221_bartrand6065324.cl"},
	{0x325A468D5227AA73ULL, "6043572_m.fxe"},
	{0x325A82C7D8EC1841ULL, "vxlt_0_lowtown_a2_gof_0_0.vlm"},
	{0x325AE6A4F514938DULL, "gen00fl_anders6114469.cl"},
	{0x325AF31E53A883F4ULL, "vxlt_warehouse_133_0_0.vlm"},
	{0x325B7B99F572DAAFULL, "6161265_m.fxe"},
	{0x325B827A4D9170B9ULL, "606273131.ogg"},
	{0x325B836117F17D62ULL, "6175441_m.fxe"},
	{0x325BBB3283B74824ULL, "rog_criticalshot.dds"},
	{0x325BE9D68FFBF629ULL, "284872200.ogg"},
	{0x325CE1460202E7E8ULL, "vxca_cavern1_haunted_gh3_0_0.vlm"},
	{0x325CE6872A345932ULL, "26597685.ogg"},
	{0x325D35FFA6FF3AAFULL, "hlac_lt_0_lowtown_a2_-131073_1y.dds"},
	{0x325D712299BF5F9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_right_2.ani"},
	{0x325D80E7823D9FB1ULL, "vxw_encounter2_c8_0_0.vlm"},
	{0x325DA9A63AA4B166ULL, "vxht_0_hightown_n_ap1_0_0.vlm"},
	{0x325DAAF574C23A67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_pleading_up.ani"},
	{0x325DCDDCCE77577EULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.cess.evt"},
	{0x325E500EF217E8D5ULL, "98574354.ogg"},
	{0x325E933A2F44EF42ULL, "vxca_caverns2prc_90d_0_0.vlm"},
	{0x325ED6522E0B19E8ULL, "art\\characters\\playercharacter\\humanmale\\glv_magc.mmh"},
	{0x325EE2522E0B2D8BULL, "art\\characters\\playercharacter\\humanmale\\glv_magc.mao"},
	{0x325EF4522E0B4C12ULL, "art\\characters\\playercharacter\\humanmale\\glv_magc.msh"},
	{0x325F2CCD3B6AFD95ULL, "6139271_m.fxe"},
	{0x325F79B491BC38ACULL, "cs_dre_18_cs_ragstoriches.wbd"},
	{0x325FA1A3682587D0ULL, "vxw_woundedcoast_1oa_0_0.vlm"},
	{0x325FABF5E1180A8EULL, "hlac_w_sundermount_-262140_0y.dds"},
	{0x325FFD1A9FA1B243ULL, "6081725_m.fxe"},
	{0x326010C65728404EULL, "244259008.ogg"},
	{0x326028C097C55859ULL, "llac_w_woundedcoast_0_6.dds"},
	{0x326038ED1889FFB3ULL, "art\\vfx\\cinematic\\vfx_shield_break_p.impa.gad"},
	{0x32605A8F1873EC34ULL, "vxlt_backalley_n_1r2_0_0.vlm"},
	{0x32605D5414E36794ULL, "6018576_f.fxe"},
	{0x32607265297D3B18ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\walk_forward_1_step.ani"},
	{0x3260AE5DFD244E29ULL, "art\\vfx\\combat\\vfx_r_gn2_hbomb_act_c.impa.evt"},
	{0x3260B599FA871C46ULL, "717641130.ogg"},
	{0x3260DDA10DD715C9ULL, "lt_gallowstemplar_2393.rml"},
	{0x32615A41697CA823ULL, "vxht_0_hightown_a3_b9r_0_0.vlm"},
	{0x326184A0D3B474B2ULL, "234472938.ogg"},
	{0x32618828EE424BA1ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.cess.ani"},
	{0x326192A16AD68CF9ULL, "421168241.ogg"},
	{0x3261C991B74027D1ULL, "vxlt_darktown_4ut_0_0.vlm"},
	{0x3261C9AEA85452AFULL, "vxdr_primevaltomb_7nz_0_0.vlm"},
	{0x3262105CCBD3059BULL, "6167407_m.fxe"},
	{0x32622D28E5141762ULL, "598538509.ogg"},
	{0x3262CDC2D46649C6ULL, "vxdr_0_deeproads_cr5_0_0.vlm"},
	{0x3262D93BECEB4A65ULL, "vxw_bonepit_29z_0_0.vlm"},
	{0x3262D97E9ECD6CEAULL, "6112553_f.fxe"},
	{0x3262E2125B05A10FULL, "syncdeathanimations.gda"},
	{0x326373D23423B3E3ULL, "vxca_caverns1_mines_cn2_0_0.vlm"},
	{0x32638C20A855843FULL, "hlac_ht_mansion2_818_3z.dds"},
	{0x3263A02E3F2ED01FULL, "vxlt_gallowstemplar_f_46h_0_0.vlm"},
	{0x32640434F03F5577ULL, "hlac_lt_0_lowtown_65533_0z.dds"},
	{0x3264573B493B7CC2ULL, "vxw_bonepitdestroyed_a3_1iz_0_0.vlm"},
	{0x3264E4DCCE7D9CA9ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.cess.gad"},
	{0x326598EAAF0261FCULL, "mer210_coterie_contact.cnv"},
	{0x3266013D0A00237AULL, "vxht_0_hightown_n_b7c_0_0.vlm"},
	{0x326610290A7C1CDCULL, "6150338_m.fxe"},
	{0x3266CD19BDDBF046ULL, "hlac_lt_foundry_n_320_9x.dds"},
	{0x326735ED18901918ULL, "art\\vfx\\cinematic\\vfx_shield_break_p.impa.evt"},
	{0x326796D62FD290E5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_right.ani"},
	{0x3267B39ED3C4C521ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyn_0t.dds"},
	{0x32685025A1C95F05ULL, "849812173.ogg"},
	{0x32689328EE487D43ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.cess.gad"},
	{0x3268A8988F4CA1F1ULL, "vxdr_0_deeproads_7ty_0_1.vlm"},
	{0x3268BAEAAF04D14CULL, "mer210_coterie_contact.bnk"},
	{0x3268E2F3A265D009ULL, "vxht_mansion_n_5ay_0_0.vlm"},
	{0x3268F79930103CA0ULL, "vxht_0_hightown_b7j_0_0.vlm"},
	{0x326910F55D471A56ULL, "vxht_0_hightown_brs_0_0.vlm"},
	{0x326923449F186031ULL, "vxlt_0_lowtown_a3_n_c_hfb_0_0.vlm"},
	{0x326964FB459E9C74ULL, "vxlt_warehouse_160_0_0.vlm"},
	{0x3269E5BFF3AC14C7ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_5z.dds"},
	{0x3269E8D627BA3D1CULL, "6151892_m.fxe"},
	{0x326A137F5EA97A28ULL, "scrollblend.mao"},
	{0x326B10ED36082DECULL, "mag371_elsa6154435.cl"},
	{0x326B65C7AEB9D310ULL, "723974583.ogg"},
	{0x326B7288BE4CB5C7ULL, "6166887_f.fxe"},
	{0x326BB5DCCE836BF7ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.cess.ani"},
	{0x326C44719C7756E9ULL, "vxw_bonepitdestroyed_a3_1au_0_0.vlm"},
	{0x326C5AFF081B9753ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.cess.evt"},
	{0x326C630550C9EF04ULL, "hlac_lt_gallowstemplar_2383_10y.dds"},
	{0x326C671B142760CBULL, "412749174.ogg"},
	{0x326CAC8C95F286CFULL, "hlac_lt_0_lowtown_-65535_1x.dds"},
	{0x326CE3415CBD138AULL, "vxlt_0_lowtown_docks_bgi_0_0.vlm"},
	{0x326CE43DA2BDA519ULL, "558082283.ogg"},
	{0x326CEFB874CB17F5ULL, "mag371_elsa6041262.cl"},
	{0x326D1C5436C4FC26ULL, "vxlt_gallowstemplar_n_4u8_0_0.vlm"},
	{0x326D4B351981DB0DULL, "gen00fl_varric6063857.cl"},
	{0x326E13805A1BA050ULL, "vxw_woundedcoast_1jn_0_0.vlm"},
	{0x326E9405C5DC7B59ULL, "668591840.ogg"},
	{0x326E965DFD306298ULL, "art\\vfx\\combat\\vfx_r_gn2_hbomb_act_c.impa.ani"},
	{0x326F0133D4AFA575ULL, "577736095.ogg"},
	{0x326F0FFC1E2051C3ULL, "hlac_lt_0_lowtown_docks_a3_-131073_2z.dds"},
	{0x326F7028EE4E6028ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.cess.evt"},
	{0x32701164929B0F91ULL, "6111949_m.fxe"},
	{0x32703BB0B3B04081ULL, "vrc321cr_servant_ghost_1.utc"},
	{0x32703F78994131FEULL, "lgt220_shp_mag_guard6149911.cl"},
	{0x32704BAD7ABD7A76ULL, "6208273_m.fxe"},
	{0x32705950C8A44907ULL, "238147190.ogg"},
	{0x327059F57DE996B0ULL, "vxlt_0_lowtown_a2_n_c_e8i_0_0.vlm"},
	{0x327203BD382550DFULL, "806686113.ogg"},
	{0x32721BFD2259B57FULL, "6060261_m.fxe"},
	{0x327236D95533FD86ULL, "lgt311_shw_mer_mettin6156521.cl"},
	{0x32726494EACB5D36ULL, "hlac_lt_0_lowtown_-3_3z.dds"},
	{0x327285F5B453AB24ULL, "t3_eey_am3.tnt"},
	{0x3272A5CF082BD560ULL, "6122876_m.fxe"},
	{0x3272E08F6228083EULL, "zz_dae_debug6029424.cl"},
	{0x3273023A9CC020E4ULL, "238267983.ogg"},
	{0x327328491149599AULL, "54598320.ogg"},
	{0x327353FF0821A994ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.cess.gad"},
	{0x32742EBECA88C0A0ULL, "hlac_lt_0_lowtown_docks_a3_-4_0x.dds"},
	{0x327445795EF4F323ULL, "6181779_m.fxe"},
	{0x32745A327A51090CULL, "920293359.ogg"},
	{0x3274B51B29448478ULL, "qun210cr_commoner.utc"},
	{0x3274E130E2FC586EULL, "vxdr_primevaltomb_1a8_0_0.vlm"},
	{0x32750CDE9A39F502ULL, "and_templar_1.utc"},
	{0x3275DC7F035220BAULL, "180290031.ogg"},
	{0x3275FC30D12A4B1CULL, "89579695.ogg"},
	{0x3276046FCFDEA74FULL, "6088152_f.fxe"},
	{0x327658B91D03C852ULL, "vxlt_0_lowtown_f66_0_1.vlm"},
	{0x3276EB5423B6CE88ULL, "hlac_lt_0_lowtown_a2_n_-65537_3z.dds"},
	{0x3277DCBABE8DBC2EULL, "hm_gamlen_h2s1.mor"},
	{0x3278C480F89216CDULL, "vxht_store_c9_0_0.vlm"},
	{0x3278E3A7CC79349AULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkfront.gad"},
	{0x32791E69BC9BF036ULL, "6141132_m.fxe"},
	{0x3279D2642B1E6E7FULL, "dae000_quintus6154571.cl"},
	{0x3279F42C75AA7D6AULL, "dae000_epic_merchant_16179419.cl"},
	{0x327A62FF0827E272ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.cess.ani"},
	{0x327A80443EF715ECULL, "vxlt_0_lowtown_a2_gmy_0_0.vlm"},
	{0x327AC15DEE3ACE59ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_aoe_p.impa.ani"},
	{0x327BB8134F49D4C3ULL, "art\\vfx\\materials\\vfx_tear_dis.mao"},
	{0x327BE264B5E7FE96ULL, "6133519_m.fxe"},
	{0x327C39CC00BC1204ULL, "hlac_ht_0_hightown_131071_0z.dds"},
	{0x327C5F915FF33545ULL, "6071863_m.fxe"},
	{0x327C73CEF923BCC8ULL, "vxca_caverns1_i3a_0_0.vlm"},
	{0x327D3E48FC601D34ULL, "vxlt_backalley_n_1x0_0_0.vlm"},
	{0x327D482ABCC0668BULL, "vxca_caverns2_4ti_0_0.vlm"},
	{0x327D914E266128C3ULL, "6153888_m.fxe"},
	{0x327E0C79DBA3F0CFULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliff02_n.dds"},
	{0x327E40103A4A64A3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_basic.mao"},
	{0x327E44103A4A6BF0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_basic.mmh"},
	{0x327E52103A4A830AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_basic.msh"},
	{0x327E663E8CC4A667ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyn_0d.dds"},
	{0x327EC3A5356DEDC7ULL, "gen00fl_anders6125319.cl"},
	{0x327F13D65B5968D8ULL, "dae100_bran6118162.cl"},
	{0x327F17A4DA1E7115ULL, "517229115.ogg"},
	{0x327F2B6276C5CE07ULL, "gen00fl_merrill_postromance.stg"},
	{0x327FC159D4096679ULL, "364776325.ogg"},
	{0x327FE6A7CC7F58B1ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkfront.evt"},
	{0x32805FDC03174361ULL, "328443225.ogg"},
	{0x32808747EC0911CAULL, "vxht_0_hightown_ass_0_0.vlm"},
	{0x32811D126B0AB6E3ULL, "6177451_m.fxe"},
	{0x32812D66878D6280ULL, "vxlt_0_lowtown_a2_n_c_e12_0_0.vlm"},
	{0x32814A34B0014F22ULL, "vxlt_hangedman_n_5k_0_0.vlm"},
	{0x32816060A783C4F7ULL, "gen00fl_anders6125171.cl"},
	{0x32819C5DEE40AD9BULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_aoe_p.impa.gad"},
	{0x3281AA5C3A4E3FBBULL, "plt_ssf100pt_main.nss"},
	{0x3281FC5F5D5F57CCULL, "fen110_anso6095003.cl"},
	{0x3282B0BA44B17500ULL, "149778051.ogg"},
	{0x3282BB8146AF6E53ULL, "art\\characters\\creatures\\mubari\\animation\\bnt_dblw_lp.ani"},
	{0x3282E8CEE1A7DA65ULL, "hlac_w_bonepitdestroyed_a3_0_1z.dds"},
	{0x3282EF0DCD726974ULL, "433787583.ogg"},
	{0x32831D5AFD0ABC2EULL, "778175911.ogg"},
	{0x328391C7AB69CE17ULL, "hlac_lt_gamlenshouse_n_-1_4y.dds"},
	{0x3283EDEA5D85F85FULL, "719125831.ogg"},
	{0x3284191696BB3377ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebopen.msh"},
	{0x32842B1696BB51DAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebopen.mao"},
	{0x32842F1696BB58E9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebopen.mmh"},
	{0x3284604CD1E524C2ULL, "mag101cr_decimus.utc"},
	{0x328473E989561614ULL, "vxlt_foundry_n_wu_0_0.vlm"},
	{0x32848111261F58FBULL, "753005418.ogg"},
	{0x328493646C4E58D0ULL, "6202295_m.fxe"},
	{0x32851F713DAD18FDULL, "287770132.ogg"},
	{0x32855E8246555E29ULL, "454908416.ogg"},
	{0x3285F4AF41334B56ULL, "vxht_0_hightown_bxu_0_0.vlm"},
	{0x328625CBF9C760D8ULL, "and110_refugee6111592.cl"},
	{0x328647622C53C439ULL, "vxlt_warehouse_n_12j_0_0.vlm"},
	{0x328709B6B1B7FEE7ULL, "qun160_vashothlead6064677.cl"},
	{0x3287655DE729DA0AULL, "68302760.ogg"},
	{0x328779A6CE5CC1BAULL, "vxht_0_hightown_b8y_0_0.vlm"},
	{0x328789FBECE8DAD5ULL, "475973448.ogg"},
	{0x3287A1D330010E4DULL, "vxlt_0_lowtown_docks_a3_n_cok_0_0.vlm"},
	{0x3287C58E18859647ULL, "vxca_cave_1d2_0_1.vlm"},
	{0x328806590A2E1ECCULL, "llac_lt_0_lowtown_a2_n_-196609_4.dds"},
	{0x3288A0665954E159ULL, "follower_banter6162100.cl"},
	{0x3288B95DEE46FD80ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_aoe_p.impa.evt"},
	{0x3289426581630438ULL, "hlac_w_sundermount_262140_0z.dds"},
	{0x3289509F35FD6EDAULL, "6116280_m.fxe"},
	{0x32898BAFF0E4AD1BULL, "6206955_m.fxe"},
	{0x3289B01569BAC886ULL, "628522683.ogg"},
	{0x3289F68DA608E3E4ULL, "one100dg_vanard6048238.cl"},
	{0x328A255246422299ULL, "807574895.ogg"},
	{0x328A70BFECD50A8FULL, "vxw_blightlands_bz_0_0.vlm"},
	{0x328AF73A3E07D150ULL, "vxw_woundedcoast_1pl_0_0.vlm"},
	{0x328AF874D43082B4ULL, "projdecal_projdecal_p1.vsh.fxo.dx11.assembly"},
	{0x328B887D3639F920ULL, "lgt261ar_tome_cave.are"},
	{0x328B93C82A2C00B5ULL, "hlac_lt_hangedman_n_-1_2y.dds"},
	{0x328BD34032C2CA9CULL, "77735676.ogg"},
	{0x328CC3D74FBE0B32ULL, "hm_magnus.mor"},
	{0x328D0876D5315BB9ULL, "432015130.ogg"},
	{0x328D23631BBA84F9ULL, "lowtown_fem_prostitute_sta6152340.cl"},
	{0x328D3680BDA9E094ULL, "977714674.ogg"},
	{0x328D3DAF61D5C7B0ULL, "vxlt_0_lowtown_a2_n_c_e2w_0_0.vlm"},
	{0x328DAD73F0936FE7ULL, "577799432.ogg"},
	{0x328DED2BD7253D8DULL, "cod_ltr_servantletter_.ncs"},
	{0x328DF443C0B883E4ULL, "567867483.ogg"},
	{0x328DFEA7CC8BBE80ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearattkfront.ani"},
	{0x328E0642CC206EB1ULL, "6051483_m.fxe"},
	{0x328E49E6C8788A80ULL, "vxlt_0_lowtown_h8i_0_0.vlm"},
	{0x328E5A02F0FF031FULL, "hlac_lt_0_lowtown_a3_n_c_0_5y.dds"},
	{0x328EA33120B092D6ULL, "vxlt_gallowstemplar_5qz_0_0.vlm"},
	{0x328EB2CD3B0A01F0ULL, "mcr310_argument6127594.cl"},
	{0x328EF1B1469322ADULL, "llac_lt_undercityserialkiller_-1_14.dds"},
	{0x328F27C323681290ULL, "fen210_orana6152640.cl"},
	{0x328F5C9196A79ACBULL, "combat_style.gda"},
	{0x328F81B28724A172ULL, "6105950_m.fxe"},
	{0x328F9FBC54655584ULL, "vxht_keep_n_5k6_0_0.vlm"},
	{0x3290041A494E9436ULL, "331005213.ogg"},
	{0x32906C3564B71B2CULL, "vxlt_0_lowtown_docks_n_apj_0_0.vlm"},
	{0x3290F4DBDD1D25B7ULL, "w_bonepitdestroyed_a3_55_0.rml"},
	{0x3291403F8E893236ULL, "6164414_m.fxe"},
	{0x329173FDA512CA80ULL, "796721381.ogg"},
	{0x32919883A366B359ULL, "6149665_m.fxe"},
	{0x3291AB9C5D6424F0ULL, "vxw_sundermount_359_0_0.vlm"},
	{0x3291ACD2368090B0ULL, "167804497.ogg"},
	{0x32920E963E5ADE5EULL, "6161211_m.fxe"},
	{0x329245E6B2C3F65AULL, "cs_mcr370cs_rules_surv.bnk"},
	{0x3292748DB4B84366ULL, "6115865_m.fxe"},
	{0x329287647F4A0F61ULL, "611558061.ogg"},
	{0x3292CA828B9291DDULL, "51278022.ogg"},
	{0x32930D3075102540ULL, "vxw_sundermount_31e_0_0.vlm"},
	{0x329326D33E8F689FULL, "gen00fl_sebastian6179593.cl"},
	{0x3293CE370AA71F81ULL, "172487886.ogg"},
	{0x3293DEDA7C831BC7ULL, "165069956.ogg"},
	{0x329404BF658DACFCULL, "character_character_p263.psh.fxo.dx11.assembly"},
	{0x329436B0A56869B3ULL, "vxw_sundermount_262_0_0.vlm"},
	{0x32951D99F1634EB7ULL, "456533076.ogg"},
	{0x329581DE0A0FDB35ULL, "830308751.ogg"},
	{0x3295977D36426733ULL, "lgt261ar_tome_cave.lst"},
	{0x3295FFA518ED229DULL, "dae000_hate_enemies6167580.cl"},
	{0x32961084A0BADA5AULL, "478600715.ogg"},
	{0x3296928146BFD1FDULL, "art\\characters\\creatures\\mubari\\animation\\bnt_dblw_lp.gad"},
	{0x3296C60B51751FBFULL, "vxlt_0_lowtown_a2_gvw_0_0.vlm"},
	{0x32970CE4A3B0D4D2ULL, "llac_lt_gallowstemplar_2388_5.dds"},
	{0x32976BBFD7DEC0D5ULL, "6127486_f.fxe"},
	{0x3297B71A6122EF4EULL, "282120858.ogg"},
	{0x3297FE5FB28486ECULL, "fake_im_arm_cht_lgt_champ.uti"},
	{0x32980717BA04F4B9ULL, "art\\levels\\races\\design\\animation\\beartrap.inactive.ani"},
	{0x329828BDB3E5BFC2ULL, "lt_0_lowtown_docks_169_0.msh"},
	{0x32982ABDB3E5C378ULL, "lt_0_lowtown_docks_169_0.mmh"},
	{0x3298362F44E53C9AULL, "447042869.ogg"},
	{0x329836BDB3E5D71BULL, "lt_0_lowtown_docks_169_0.mao"},
	{0x329844DFDD8F67A0ULL, "vxht_brothel_n_4c0_0_0.vlm"},
	{0x329880814DCAAA63ULL, "24071692.ogg"},
	{0x3298A5CC2F1634A1ULL, "dae000_gamlen6124368.cl"},
	{0x3298C61E99D42F96ULL, "6139319_m.fxe"},
	{0x3298D7F8B3AB6EB0ULL, "106162462.ogg"},
	{0x329928C95F1A9AD7ULL, "gen_im_npc_arm_hum_hlm_tmpa.uti"},
	{0x32992B8D373F3418ULL, "vxht_chantry_302_0_0.vlm"},
	{0x329956415CC6560BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_walk.ani"},
	{0x3299A8413AB722DAULL, "w_bonepit_66_0.mmh"},
	{0x3299AC413AB72911ULL, "w_bonepit_66_0.mao"},
	{0x3299BA413AB74120ULL, "w_bonepit_66_0.msh"},
	{0x3299EF9A9BDCFD9CULL, "vxlt_gallowstemplar_5y8_0_0.vlm"},
	{0x329A8D8F7D685376ULL, "zz_approval_debug6003757.cl"},
	{0x329AA7A7DE93ED1CULL, "seb321_leliana6197871.cl"},
	{0x329AD31E146039E3ULL, "llac_lt_0_lowtown_-3_6.dds"},
	{0x329AF250F6546845ULL, "vxht_mansion_n_66w_0_0.vlm"},
	{0x329B5F14B3DCDCFFULL, "mag360_samson6171451.cl"},
	{0x329B60F7A536C134ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.impa.ani"},
	{0x329BE18733D20CF6ULL, "isa311_isabela6175517.cl"},
	{0x329C07D562241391ULL, "dae000_bartrand6031165.cl"},
	{0x329C0D0DB37BC1F9ULL, "507447280.ogg"},
	{0x329C49B45C173D61ULL, "674207904.ogg"},
	{0x329C4A418D310BB4ULL, "439682909.ogg"},
	{0x329C5B8659D415B4ULL, "617567859.ogg"},
	{0x329CF1C9D18D821BULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar2_.ani"},
	{0x329DB3A276752ED5ULL, "vxdr_0_deeproads_ev8_0_0.vlm"},
	{0x329DBFAD954D4BCAULL, "llac_lt_warehouse_1_6.dds"},
	{0x329DE9080D447D58ULL, "852632842.ogg"},
	{0x329E13E6605EB86AULL, "ave200_donnic6112141.cl"},
	{0x329E2287FC30E355ULL, "6076372_m.fxe"},
	{0x329E2F080047CFA1ULL, "985364433.ogg"},
	{0x329ED4EFBEED57CDULL, "art\\characters\\playercharacter\\elfmale\\arm_lgte.msh"},
	{0x329EE217BA0AD4FBULL, "art\\levels\\races\\design\\animation\\beartrap.inactive.gad"},
	{0x329EE2EFBEED6F00ULL, "art\\characters\\playercharacter\\elfmale\\arm_lgte.mao"},
	{0x329EE6EFBEED7653ULL, "art\\characters\\playercharacter\\elfmale\\arm_lgte.mmh"},
	{0x329F0FE2CD3BD306ULL, "6110538_m.fxe"},
	{0x329F147F4F00C7C4ULL, "zz_dae_debug6143036.cl"},
	{0x329FFE9A4A8AE3DBULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliff2_da.dds"},
	{0x32A017047A30B2D7ULL, "qun210cr_guard6083835.cl"},
	{0x32A04FF2BDA19349ULL, "gen00fl_anders6119523.cl"},
	{0x32A081CE0EC28361ULL, "6102793_f.fxe"},
	{0x32A109C10B290C97ULL, "664271791.ogg"},
	{0x32A133D60A1EAE5CULL, "vxca_cavern1_haunted_fmq_0_0.vlm"},
	{0x32A16715A80388BDULL, "dre000_varric6034663.cl"},
	{0x32A247059E356CD7ULL, "art\\characters\\weapons\\dagger\\w_dger_106a.phy"},
	{0x32A24A194E3F0975ULL, "6101064_f.fxe"},
	{0x32A25DF7A53CDA96ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.impa.gad"},
	{0x32A2B6FA87DABAB6ULL, "art\\levels\\ztools\\default\\generic\\9ff_table.msh"},
	{0x32A2C0FA87DACBB4ULL, "art\\levels\\ztools\\default\\generic\\9ff_table.mmh"},
	{0x32A2C4FA87DAD2C7ULL, "art\\levels\\ztools\\default\\generic\\9ff_table.mao"},
	{0x32A2CCE7588C15FDULL, "vxca_caverns1_mines_jg6_0_0.vlm"},
	{0x32A33E79A3CD535CULL, "6128531_m.fxe"},
	{0x32A347D35484D045ULL, "1056205992.ogg"},
	{0x32A35D6501F9759FULL, "811399147.ogg"},
	{0x32A39ACACDB6FB36ULL, "vxw_bonepit_20r_0_0.vlm"},
	{0x32A3BAAB4970D1F0ULL, "6039854_f.fxe"},
	{0x32A3DAE71D69818CULL, "156663728.ogg"},
	{0x32A44BDA77F4B392ULL, "vxlt_undercityserialkiller_1ny_0_0.vlm"},
	{0x32A4858D13ED3F4DULL, "vxlt_0_lowtown_docks_a3_n_cui_0_0.vlm"},
	{0x32A4F5869B199999ULL, "43378194.ogg"},
	{0x32A501A840C719FBULL, "sse_ss_varric.gda"},
	{0x32A526253B020524ULL, "mer320_hubert6083766.cl"},
	{0x32A537E8436B383FULL, "vxlt_0_lowtown_a2_gmv_0_0.vlm"},
	{0x32A5C897117A0BDAULL, "790390334.ogg"},
	{0x32A60D93BACFE04DULL, "402550654.ogg"},
	{0x32A6D5EE4A9AC97FULL, "vxw_blightlands_c9_0_0.vlm"},
	{0x32A6F67EFC4A5766ULL, "hlac_ht_0_hightown_a3_-131069_0y.dds"},
	{0x32A73650F75EA7DBULL, "6111598_m.fxe"},
	{0x32A73C689D729327ULL, "darktown_male_elf_sta.bnk"},
	{0x32A73E415CD26A82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_walk.evt"},
	{0x32A79906B0E92995ULL, "vxlt_0_lowtown_a3_n_c_i0m_0_0.vlm"},
	{0x32A7BA1DD0D8D42CULL, "155146103.ogg"},
	{0x32A7C12EDBA8098DULL, "gui_reticle_p1.psh.fxo.dx11.assembly"},
	{0x32A7DA8FB3C4F0D2ULL, "6120911_m.fxe"},
	{0x32A85142A14F5D44ULL, "585183987.ogg"},
	{0x32A8F741E6D400BEULL, "216933992.ogg"},
	{0x32A91FF1ADA51AA1ULL, "hightown_market_guard_sta6169658.cl"},
	{0x32A967B3CD38804EULL, "vxw_blightlands_b7_0_0.vlm"},
	{0x32A978F7A54327F5ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.impa.evt"},
	{0x32A97F3FE4D57068ULL, "600029498.ogg"},
	{0x32A9C73F8DEAF388ULL, "6076801_m.fxe"},
	{0x32A9D469A4A230B9ULL, "gen00fl_dog6165701.cl"},
	{0x32AA6ECE40CE9A23ULL, "mag120_macha6080635.cl"},
	{0x32AA78C02D4D23CAULL, "vxca_cavern1_haunted_dpb_0_0.vlm"},
	{0x32AA9EAC152F06CBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_wf_injured.ani"},
	{0x32AAD6BF9992BCB0ULL, "zz_followers_banter6157530.cl"},
	{0x32AB1A8EA2FAF1EEULL, "hlac_w_bonepitmining_a2_458758_0y.dds"},
	{0x32AB5E689D76B59FULL, "darktown_male_elf_sta.cnv"},
	{0x32AB69F691CCAFFFULL, "gen00fl_aveline6120746.cl"},
	{0x32ABB8D6E09CAF29ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzqasign.mao"},
	{0x32ABC4D6E09CC3C2ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzqasign.mmh"},
	{0x32ABC6D6E09CC778ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzqasign.msh"},
	{0x32ABC86E3E6AC68FULL, "610604686.ogg"},
	{0x32ABEDF5FE38D734ULL, "vxlt_0_lowtown_docks_a3_akk_0_0.vlm"},
	{0x32AC24C77C04399AULL, "vxca_caverns1_co9_0_0.vlm"},
	{0x32AD2483BAC926BCULL, "vxca_caverns2_92u_0_0.vlm"},
	{0x32AD343BB6A822F4ULL, "hlac_ca_caverns1_mines_24423_6y.dds"},
	{0x32AD5A68571D20E9ULL, "qun120_arishok6051695.cl"},
	{0x32ADC313DCE9F0F5ULL, "6095407_m.fxe"},
	{0x32AE3D415CD88795ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_walk.gad"},
	{0x32AEE1B8DC173EFAULL, "6205111_m.fxe"},
	{0x32AF15B984488C74ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkleft.ani"},
	{0x32AF2CEA9005B04BULL, "6107126_m.fxe"},
	{0x32AF907AB5BC8CEBULL, "377071617.ogg"},
	{0x32AFBF4A16CE72B5ULL, "6115602_m.fxe"},
	{0x32AFE783FF4821DDULL, "6139305_m.fxe"},
	{0x32AFF2F4F1E3E92CULL, "24959672.ogg"},
	{0x32B000E9548EA158ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_min_c.anb"},
	{0x32B076CB2240BD86ULL, "llac_lt_0_lowtown_docks_a3_n_65537_0.dds"},
	{0x32B0DEA3EAA40AEBULL, "vxlt_0_lowtown_a2_n_c_hid_0_0.vlm"},
	{0x32B112755C23914CULL, "440239961.ogg"},
	{0x32B118C9D19E6DA5ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar2_.gad"},
	{0x32B17E0A988E01BEULL, "vxlt_0_lowtown_a2_n_f5i_0_0.vlm"},
	{0x32B1B55C3A777654ULL, "plt_ssf100pt_main.plo"},
	{0x32B1E00E7AF8C0CBULL, "one100dg_nabil6033118.cl"},
	{0x32B26554F78CFF1DULL, "654566278.ogg"},
	{0x32B2793DCD570FD6ULL, "gen00fl_aveline6121935.cl"},
	{0x32B38A7BC3E6CF08ULL, "plt_dae000pt_generic.nss"},
	{0x32B43D8E09DA5355ULL, "dre000st_varric_framingc.stg"},
	{0x32B4C9F1406EC08FULL, "zz_lgt_1_debug6136986.cl"},
	{0x32B4DDA9EF97CC95ULL, "vxca_caverns1_ghw_0_0.vlm"},
	{0x32B4F4379BB34359ULL, "1039618612.ogg"},
	{0x32B512B9844CF2D6ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkleft.gad"},
	{0x32B53D7C84F749AFULL, "851030338.ogg"},
	{0x32B59929684E374AULL, "ave200_donnic6059569.cl"},
	{0x32B5A82DA70889CAULL, "555188842.ogg"},
	{0x32B5B2757C2F3011ULL, "vxw_bonepitmining_a2_2a8_0_0.vlm"},
	{0x32B5B4E6A35B9E9CULL, "6061785_f.fxe"},
	{0x32B5F262043D65DBULL, "zz_followers_banter6159429.cl"},
	{0x32B652A4481B1A1DULL, "vxht_brothel_n_57m_0_1.vlm"},
	{0x32B6A9984C024A46ULL, "atl_startmenu_dxt5_04.dds"},
	{0x32B6DF887BE6396DULL, "6110959_m.fxe"},
	{0x32B6F1619A25B469ULL, "hlac_ca_caverns1_mines_24424_23z.dds"},
	{0x32B757A7802C98B5ULL, "hlac_lt_0_lowtown_a2_n_c_65538_4x.dds"},
	{0x32B7686DC89513F5ULL, "vxht_keep_n_8bq_0_0.vlm"},
	{0x32B774586FAE09ACULL, "vxlt_warehouse_p3_0_0.vlm"},
	{0x32B7986998354477ULL, "6121640_f.fxe"},
	{0x32B80D34919ED0ABULL, "vxht_mansion2_1po_0_0.vlm"},
	{0x32B85FE6B2E4F1F8ULL, "cs_mcr370cs_rules_surv.wbd"},
	{0x32B8A8989ED0C421ULL, "vxlt_backalley_n_1js_0_0.vlm"},
	{0x32B8F506794CC61AULL, "6037093_m.fxe"},
	{0x32B963BA51D7A435ULL, "hlac_ht_mansion2_818_8y.dds"},
	{0x32B98F5AD03867F7ULL, "vxlt_0_lowtown_n_e0x_0_0.vlm"},
	{0x32BA0EBF4416844CULL, "vxdr_0_deeproads_erx_0_1.vlm"},
	{0x32BA258E0378951CULL, "hlac_lt_0_lowtown_-131072_7y.dds"},
	{0x32BA89A4AD31017BULL, "vxlt_gallowscourtyard_4ux_0_0.vlm"},
	{0x32BA937BC3ECFD1BULL, "plt_dae000pt_generic.plo"},
	{0x32BB1BC85C06100EULL, "vxht_0_hightown_a2_3w1_0_0.vlm"},
	{0x32BB8332806A472FULL, "6169953_m.fxe"},
	{0x32BC2DB984533F35ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkleft.evt"},
	{0x32BC5935CE7AAD29ULL, "lgt150cr_nat_were_dalish_gd.utc"},
	{0x32BC61D3E9DE0082ULL, "t3_dog_ob3.tnt"},
	{0x32BC7319706822B1ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyn_0d.dds"},
	{0x32BCCC9C1DFB8842ULL, "6098933_m.fxe"},
	{0x32BCD289C5963EF7ULL, "dae000ip_crowd_mf8_talk3.utp"},
	{0x32BCD7A3F060F518ULL, "464330646.ogg"},
	{0x32BD4D74A4B4FE9DULL, "vxlt_undercityserialkiller_1sq_0_0.vlm"},
	{0x32BE0658AE15778CULL, "vxw_encounter1_b3_0_0.vlm"},
	{0x32BE85AC153F8555ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_wf_injured.gad"},
	{0x32BECE836D574B24ULL, "6091374_m.fxe"},
	{0x32BEDF5B4E493890ULL, "ttx_floor_small_1.mmh"},
	{0x32BEED5B4E4950AAULL, "ttx_floor_small_1.msh"},
	{0x32BF4752B7F125FCULL, "vxht_0_hightown_by8_0_0.vlm"},
	{0x32BF4B103A820E0BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_basic.phy"},
	{0x32BF513A0B82FF8CULL, "fxe_dust_sunshine.dds"},
	{0x32BFB02C8D8F843EULL, "vxlt_gallowstemplar_f_2pa_0_0.vlm"},
	{0x32C0CAAD3F008DC7ULL, "ser221_party_comments6187718.cl"},
	{0x32C0EE93EA2003B2ULL, "vxht_0_hightown_n_bx6_0_0.vlm"},
	{0x32C16D9DE8915D96ULL, "56420245.ogg"},
	{0x32C199EB75C6ADC1ULL, "232757690.ogg"},
	{0x32C1E1D0E79FACC4ULL, "vxdr_dungeon_2hh_0_0.vlm"},
	{0x32C1F18C4A979FFDULL, "6059336_f.fxe"},
	{0x32C219FE6864174BULL, "character_character_p2.psh.fxo"},
	{0x32C26E9B0F5F1E32ULL, "darktown_male_sickly_sta6175321.cl"},
	{0x32C2715FD121A3B8ULL, "388975035.ogg"},
	{0x32C2F80CE9A720C8ULL, "hlac_lt_warehouse_n_65537_0z.dds"},
	{0x32C311C75D3BE9E9ULL, "follower_banter6151357.cl"},
	{0x32C339697EC72092ULL, "hlac_lt_0_lowtown_docks_n_-65537_1y.dds"},
	{0x32C3689C885B4EACULL, "6096515_m.fxe"},
	{0x32C384822D67D3A5ULL, "vxht_chantry_n_20u_0_0.vlm"},
	{0x32C3BCE4D1F2B2FDULL, "hlac_lt_backalley_g_0_5z.dds"},
	{0x32C53B142D4ECAD5ULL, "vxlt_warehouse_12y_0_0.vlm"},
	{0x32C55A73C6EF2C8FULL, "vxw_sundermount_2ac_0_0.vlm"},
	{0x32C58472E1A60F97ULL, "vxca_caverns2_8bw_0_0.vlm"},
	{0x32C5A62E5DC0A500ULL, "579716973.ogg"},
	{0x32C5D4ECE1D044F0ULL, "seb221_flora_harimann6177279.cl"},
	{0x32C5E63561A4222CULL, "vxlt_warehouse_u0_0_0.vlm"},
	{0x32C64BC232169ED4ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_fereldencoin_sa.dds"},
	{0x32C670886B163EFFULL, "vxlt_0_lowtown_a3_n_c_byk_0_0.vlm"},
	{0x32C70FE14E5B9A24ULL, "6104735_m.fxe"},
	{0x32C7786075666427ULL, "hlac_lt_0_lowtown_a2_n_c_65533_3z.dds"},
	{0x32C7ED7E88B21B19ULL, "98974804.ogg"},
	{0x32C9AB33CC552BCCULL, "zz_ser_debug6078150.cl"},
	{0x32C9D84201F98EA2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop_lifted.gad"},
	{0x32CA1AB8A9A938E7ULL, "vxw_bonepitmining_a2_1lb_0_0.vlm"},
	{0x32CA2E9777811602ULL, "vxlt_darktown_41w_0_0.vlm"},
	{0x32CB2168F787DB6CULL, "6099395_f.fxe"},
	{0x32CB2806A16C0268ULL, "hlac_ht_store_187_1z.dds"},
	{0x32CB79A3FDE3CB7DULL, "rdr110dg_dockworker6015936.cl"},
	{0x32CBCE382D72A032ULL, "6094968_m.fxe"},
	{0x32CBFB1F6AE76C6FULL, "dae000_sandal6147816.cl"},
	{0x32CC2412CA522A0BULL, "806460911.ogg"},
	{0x32CC7835A00B900CULL, "vxw_encounter1_g0_0_0.vlm"},
	{0x32CCCE58A8FB4FAFULL, "gen_im_arm_cht_med_act3_02.uti"},
	{0x32CD15EC01F66F12ULL, "lt_0_lowtown_a2_33_0.anb"},
	{0x32CD178497858EC7ULL, "mer340_cowan6083521.cl"},
	{0x32CDD85F932C318FULL, "101857663.ogg"},
	{0x32CEEC698EA48BCBULL, "30878094.ogg"},
	{0x32CF6070DC161C32ULL, "vxht_0_hightown_n_bc7_0_0.vlm"},
	{0x32CF87F1EDF2AD1AULL, "mag360_samson6062118.cl"},
	{0x32CFE89FD00A1B15ULL, "dae000ip_crowd_ff5_talk7.utp"},
	{0x32D07AA21459F2C2ULL, "art\\vfx\\combat\\vfx_r_var_range_v5_c.impa.gad"},
	{0x32D0E861115A11A9ULL, "vxht_0_hightown_a2_b5y_0_0.vlm"},
	{0x32D0FB4201FFE899ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop_lifted.evt"},
	{0x32D1A249604D7B45ULL, "134127935.ogg"},
	{0x32D1FFC9E8BECCAAULL, "vxdr_0_deeproads_cu8_0_0.vlm"},
	{0x32D23EEE96D160ECULL, "148280689.ogg"},
	{0x32D29D48B1C36FEAULL, "vxca_caverns1_mines_j2d_0_2.vlm"},
	{0x32D3991F45384B3BULL, "art\\vfx\\creatures\\fxc_wisp_c.dura.ani"},
	{0x32D3CC50B8E5450FULL, "vxw_sundermount_2jb_0_0.vlm"},
	{0x32D3ECFA27405B97ULL, "vxht_0_hightown_bfr_0_0.vlm"},
	{0x32D4B97A5B3EF4C8ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightd.msh"},
	{0x32D4BFBFCC266C06ULL, "78292347.ogg"},
	{0x32D4CB7A5B3F1359ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightd.mao"},
	{0x32D4D77A5B3F27F2ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightd.mmh"},
	{0x32D4F9CFCC254FB9ULL, "vxht_0_hightown_a2_9zk_0_0.vlm"},
	{0x32D5270EA1D8259FULL, "llac_lt_0_lowtown_docks_-196610_1.dds"},
	{0x32D529BC9C2D3155ULL, "llac_lt_darktown_-1_69.dds"},
	{0x32D58C5282BCF521ULL, "vxlt_backalley_n_1pq_0_0.vlm"},
	{0x32D6536CD2C3304DULL, "lt_gallowscourtyard_51_0_ll.mmh"},
	{0x32D674520AF639D4ULL, "620630250.ogg"},
	{0x32D692063A7E6E48ULL, "qun131_warrens_thug6038769.cl"},
	{0x32D70FB311A862FBULL, "llac_lt_gallowscourtyard_-131072_2.dds"},
	{0x32D79DA214604CB9ULL, "art\\vfx\\combat\\vfx_r_var_range_v5_c.impa.evt"},
	{0x32D7A861037A2599ULL, "258720986.ogg"},
	{0x32D7C0B18BBA85F7ULL, "vxht_0_hightown_90c_0_0.vlm"},
	{0x32D899000768A4E1ULL, "gen00fl_fenris6105861.cl"},
	{0x32D8F5033C7D00B6ULL, "29212789.ogg"},
	{0x32D8FFE954B1A4F3ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_min_c.mmh"},
	{0x32D958592188027AULL, "788604645.ogg"},
	{0x32D9E2F80381F5B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_loop.ani"},
	{0x32DA5E897E0C1EEEULL, "hlac_ht_0_hightown_a3_-196610_0z.dds"},
	{0x32DA6E3101F33A3AULL, "fex_bd_b_bs_cn.mmh"},
	{0x32DA803101F35880ULL, "fex_bd_b_bs_cn.msh"},
	{0x32DACA5C10A8A81BULL, "vxlt_hangedman_n_281_0_0.vlm"},
	{0x32DB33281B0A725EULL, "zz_one_debug6079328.cl"},
	{0x32DBC1BB047E04D2ULL, "550947950.ogg"},
	{0x32DBC28AE05C40A1ULL, "vxht_0_hightown_a3_c72_0_1.vlm"},
	{0x32DC51A365FA4899ULL, "llac_w_sundermount_-196603_0.dds"},
	{0x32DCBA7BE2F74A09ULL, "isa321_velasco6117649.cl"},
	{0x32DCBC12AB054C8BULL, "6114412_m.fxe"},
	{0x32DD81E4F734AC92ULL, "38287873.ogg"},
	{0x32DDA156B461DEA6ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_saintsmoke.mmh"},
	{0x32DDDD30949E6043ULL, "376097528.ogg"},
	{0x32DE0342020A8088ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_lying_female_loop_lifted.ani"},
	{0x32DE076C5705509CULL, "llac_lt_0_lowtown_n_131071_3.dds"},
	{0x32DE548809F85886ULL, "pro000_athenril6048742.cl"},
	{0x32DE5CA91C5242DCULL, "477338156.ogg"},
	{0x32DFC171E96984E6ULL, "dae000_male_commoner1_mob6174467.cl"},
	{0x32DFCE19C8480FAEULL, "hlac_lt_0_lowtown_a2_131070_2y.dds"},
	{0x32E04EBCC4AD4C09ULL, "vxw_encounter2_camp_6w_0_0.vlm"},
	{0x32E071A6DAB4E52AULL, "vxdr_0_deeproads_cr9_0_0.vlm"},
	{0x32E08D892FB16A44ULL, "w_encounter3_43_0.msh"},
	{0x32E09F892FB188ADULL, "w_encounter3_43_0.mao"},
	{0x32E0A11F4542E332ULL, "art\\vfx\\creatures\\fxc_wisp_c.dura.evt"},
	{0x32E0A3892FB18FE6ULL, "w_encounter3_43_0.mmh"},
	{0x32E0D4C5AEBACB2EULL, "vxdr_primevalentrance_2q5_0_2.vlm"},
	{0x32E159DDAB484749ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magc_0d.dds"},
	{0x32E169C51CC6AC83ULL, "dae000ip_crowd_ff4_talk17.utp"},
	{0x32E16FD2DB7A7026ULL, "282254812.ogg"},
	{0x32E1CDC3EFDC8959ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devastationaoe.ani"},
	{0x32E1DFEEB128E08EULL, "vxw_woundedcoast_y8_0_0.vlm"},
	{0x32E1F218375CE720ULL, "616547661.ogg"},
	{0x32E225C499700E02ULL, "6095930_m.fxe"},
	{0x32E23E2DAADB5D8FULL, "vxw_sundermount_2om_0_0.vlm"},
	{0x32E27C15C398EE8DULL, "hlac_ht_keep_n_3736_1y.dds"},
	{0x32E2BEDB1A1B3CE9ULL, "6177682_m.fxe"},
	{0x32E38A7E91081162ULL, "6021344_m.fxe"},
	{0x32E3B5BF19A40C43ULL, "vxht_0_hightown_n_cf8_0_0.vlm"},
	{0x32E3C7D5AF848FFDULL, "6144316_m.fxe"},
	{0x32E4285C07DB8103ULL, "vxlt_0_lowtown_n_gbp_0_0.vlm"},
	{0x32E439EBAD428C13ULL, "921001352.ogg"},
	{0x32E45D9410F70DD0ULL, "6094562_m.fxe"},
	{0x32E4A5A2146AE4A8ULL, "art\\vfx\\combat\\vfx_r_var_range_v5_c.impa.ani"},
	{0x32E5D401910B09B3ULL, "llac_w_sundermount_-2_1.dds"},
	{0x32E6407BEB58ACA4ULL, "one100dg_nabil6173817.cl"},
	{0x32E69A9564854AE4ULL, "813609867.ogg"},
	{0x32E6D9F543BCB3B1ULL, "6134962_m.fxe"},
	{0x32E6DC455A23AD2CULL, "lgt000cr_bysmor.utc"},
	{0x32E6FCFAE7EF182EULL, "6060707_m.fxe"},
	{0x32E73435A1848BE6ULL, "vxlt_0_lowtown_a3_n_c_cfv_0_0.vlm"},
	{0x32E74020F3EF6A2EULL, "vxht_0_hightown_n_7j1_0_0.vlm"},
	{0x32E76E882B5D504EULL, "zz_ave_debug6056776.cl"},
	{0x32E7796581D3A91AULL, "6055152_m.fxe"},
	{0x32E7A3E8CD001BBFULL, "fx_noise02.dds"},
	{0x32E7A9460D2DCBCBULL, "6059276_m.fxe"},
	{0x32E7B88806E5E73EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_die.ani"},
	{0x32E7C01F454936C5ULL, "art\\vfx\\creatures\\fxc_wisp_c.dura.gad"},
	{0x32E7C1832BFF8488ULL, "6165451_m.fxe"},
	{0x32E7DFA6D84DC4DAULL, "863928051.ogg"},
	{0x32E7EE8A7053265CULL, "vxlt_0_lowtown_a2_n_c_e1n_0_0.vlm"},
	{0x32E7F9492CB52DF9ULL, "896356796.ogg"},
	{0x32E86C0CC43F0B64ULL, "hf_ibethany_h2s2.mor"},
	{0x32E88C7E50F58178ULL, "116871343.ogg"},
	{0x32E8A8C3EFE2689BULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devastationaoe.gad"},
	{0x32E9098333FB9F71ULL, "651938346.ogg"},
	{0x32E948B21E2439D7ULL, "gen00fl_varric6079580.cl"},
	{0x32E95A3BE1411F28ULL, "vxlt_hangedman_n_25a_0_0.vlm"},
	{0x32E99DC74B7BF2B0ULL, "framebuffereffects_verticalblurshadows_p0.vsh.fxo.dx11.assembly"},
	{0x32E9DE1BA887B580ULL, "vxht_chantry_n_2ik_0_0.vlm"},
	{0x32EA02D8229D2CF3ULL, "l_dre180ar_ruins_crypt.bnk"},
	{0x32EA085E0F8AFC07ULL, "gen00fl_isabela6075287.cl"},
	{0x32EA28059E72EDCCULL, "art\\characters\\weapons\\dagger\\w_dger_106a.mmh"},
	{0x32EA2E059E72F7BEULL, "art\\characters\\weapons\\dagger\\w_dger_106a.msh"},
	{0x32EA335B609A6EA4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_black.msh"},
	{0x32EA3E3C7336BFA8ULL, "qcr220_meredith6041536.cl"},
	{0x32EA455B609A8D0DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_black.mao"},
	{0x32EA47655AE3F6DAULL, "565943330.ogg"},
	{0x32EA495B609A9446ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_black.mmh"},
	{0x32EA6DF6EF7C8445ULL, "13993361.ogg"},
	{0x32EAA617B27C6CEBULL, "vxlt_0_lowtown_a2_n_c_hux_0_0.vlm"},
	{0x32EB445B5DAA1941ULL, "qun111_ketojan6059232.cl"},
	{0x32EB48C3F2EDBF35ULL, "834656100.ogg"},
	{0x32EB494148EE89BFULL, "qun120_arishok6049656.cl"},
	{0x32EBA523AA706DF1ULL, "6119651_m.fxe"},
	{0x32EC0D58507B1796ULL, "vxht_mansion2_3ca_0_0.vlm"},
	{0x32EC4F23761AA0FBULL, "6040978_f.fxe"},
	{0x32ECB3BA64C1619EULL, "76892202.ogg"},
	{0x32EE5ECD6DF6631FULL, "dae150ar_sundermount.are"},
	{0x32EEB7E74D3B106DULL, "hlac_lt_0_lowtown_a2_n_c_-3_1x.dds"},
	{0x32EEB9F803940C5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_loop.gad"},
	{0x32EEF695CAC98F51ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mattock_d.dds"},
	{0x32EF0E91B8BF3BD5ULL, "6076353_f.fxe"},
	{0x32EF7193D685013CULL, "6075813_m.fxe"},
	{0x32EFC5C3EFE8B880ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devastationaoe.evt"},
	{0x32EFEA6EDDAA6043ULL, "vxht_0_hightown_c7t_0_0.vlm"},
	{0x32EFF5BF8E28AF6BULL, "vxlt_hangedman_n_20i_0_0.vlm"},
	{0x32F0033463686C01ULL, "951655042.ogg"},
	{0x32F06FB2D4458337ULL, "6097421_m.fxe"},
	{0x32F0B00A9CD1760FULL, "vxw_sundermount_2pl_0_0.vlm"},
	{0x32F13D74F134B032ULL, "vxht_0_hightown_n_9ar_0_0.vlm"},
	{0x32F1A2239BF08BFEULL, "vxlt_gallowstemplar_n_51y_0_0.vlm"},
	{0x32F1A3892FC002ADULL, "w_encounter3_43_0.rml"},
	{0x32F1EC832ED88A96ULL, "6174863_f.fxe"},
	{0x32F1F2080092277EULL, "6201383_f.fxe"},
	{0x32F229BEA6D1FE9FULL, "658671079.ogg"},
	{0x32F31C53ABDF276CULL, "gen00fl_isabela6079286.cl"},
	{0x32F4498F8E30B9C2ULL, "follower_banter6163300.cl"},
	{0x32F4D08806F09A0FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_die.evt"},
	{0x32F4DD1B5CF1282FULL, "806456802.ogg"},
	{0x32F50B0AB13E4E5CULL, "vxlt_0_lowtown_h85_0_0.vlm"},
	{0x32F56553D5F9D195ULL, "vxlt_0_lowtown_docks_a3_d3k_0_0.vlm"},
	{0x32F56DF3A03E191AULL, "vxw_encounter2_j4_0_0.vlm"},
	{0x32F58613D50B1953ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_8y.dds"},
	{0x32F59DDCC2B1D169ULL, "zz_pridedemon2.cut"},
	{0x32F5A612937AA466ULL, "vxlt_0_lowtown_a3_n_c_cmw_0_0.vlm"},
	{0x32F6096E7B4D7052ULL, "6151506_m.fxe"},
	{0x32F650F6119D861AULL, "w_blightlands_10_0.mmh"},
	{0x32F654F6119D8D51ULL, "w_blightlands_10_0.mao"},
	{0x32F662F6119DA560ULL, "w_blightlands_10_0.msh"},
	{0x32F6D5C4E1205835ULL, "328268462.ogg"},
	{0x32F70D3CF1EFC58CULL, "181872056.ogg"},
	{0x32F74EFA1E48D2B7ULL, "lt_0_lowtown_docks_n_202_0.msh"},
	{0x32F760FA1E48F11AULL, "lt_0_lowtown_docks_n_202_0.mao"},
	{0x32F764FA1E48F829ULL, "lt_0_lowtown_docks_n_202_0.mmh"},
	{0x32F786A54CEE6F2AULL, "vxw_sundermount_34q_0_1.vlm"},
	{0x32F7A184EF621AC6ULL, "genip_cin_qun_amulet.utp"},
	{0x32F883FE2C1F5762ULL, "hlac_lt_0_lowtown_a2_-3_5z.dds"},
	{0x32F88BB1EC469D68ULL, "6075570_m.fxe"},
	{0x32F96F2E7CBA5DE4ULL, "6015998_m.fxe"},
	{0x32F97A593F5FC04EULL, "gen00fl_aveline6139466.cl"},
	{0x32F9C900EBD90177ULL, "75987289.ogg"},
	{0x32F9FB198AF15D45ULL, "gen00fl_aveline6116671.cl"},
	{0x32F9FC555822F001ULL, "376298110.ogg"},
	{0x32FA17009C668671ULL, "6114151_m.fxe"},
	{0x32FA4F1D5916EFA7ULL, "630470375.ogg"},
	{0x32FAE92142E29EACULL, "qun221_viscount6095614.cl"},
	{0x32FB3FE4B6744B43ULL, "hlac_lt_0_lowtown_a2_-65536_7y.dds"},
	{0x32FBD18806F6BAB8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_die.gad"},
	{0x32FBF05D33494274ULL, "834784088.ogg"},
	{0x32FC271D198D58C3ULL, "967052754.ogg"},
	{0x32FC5A860F0B5579ULL, "6206140_m.fxe"},
	{0x32FD1CF78F4E7751ULL, "w_sundermount_66_0_lc.mmh"},
	{0x32FD3F2B69561869ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_ex.ani"},
	{0x32FD76F19F57AD40ULL, "hlac_ca_cavern1_haunted_24426_5y.dds"},
	{0x32FD862A1A2D43D6ULL, "6111537_m.fxe"},
	{0x32FD8DF8A52DC8A9ULL, "hlac_lt_0_lowtown_n_131072_2y.dds"},
	{0x32FDC8E01FF5F025ULL, "569998144.ogg"},
	{0x32FE95CDF16A3D2DULL, "6172559_m.fxe"},
	{0x32FEBFFA882928AFULL, "art\\levels\\ztools\\default\\generic\\9ff_table.phy"},
	{0x32FECFF8AABD4620ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medc.phy"},
	{0x32FEDFBD2D4BAFE2ULL, "vxlt_0_lowtown_docks_a0k_0_0.vlm"},
	{0x32FF57CD6E04CABCULL, "dae150ar_sundermount.lst"},
	{0x32FFE2F133A244ECULL, "6167893_m.fxe"},
	{0x32FFE5FD8598AC39ULL, "plt_dae00pt_secondary.nss"},
	{0x330018EC0221CE4EULL, "lt_0_lowtown_a2_33_0.rml"},
	{0x33007F01687C3580ULL, "6111997_m.fxe"},
	{0x330246F2903FEAECULL, "838112089.ogg"},
	{0x33025B03AC8C3918ULL, "6072603_m.fxe"},
	{0x33027E1496E6ACC8ULL, "705364988.ogg"},
	{0x33027FF0F073C51AULL, "genip_cot.utp"},
	{0x33028542A42D5C3FULL, "83490425.ogg"},
	{0x3302905D2F5572DDULL, "6076930_m.fxe"},
	{0x330294F108967335ULL, "vxlt_gallowstemplar_f_3rm_0_0.vlm"},
	{0x3302E720E71D17F9ULL, "vxw_sundermount_2g3_0_0.vlm"},
	{0x33036EA214E371D6ULL, "ser270_emeric6149328.cl"},
	{0x33039726CC0CD93AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokena.msh"},
	{0x3303A526CC0CF173ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokena.mao"},
	{0x3303A926CC0CF780ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokena.mmh"},
	{0x3303BDAB73B5B854ULL, "qun221_viscount6084689.cl"},
	{0x3303D730C7EFEA15ULL, "vxlt_0_lowtown_docks_a3_d4j_0_0.vlm"},
	{0x33045A2B695C64EBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_ex.gad"},
	{0x330471C9626C9C66ULL, "186609927.ogg"},
	{0x3304C9EEB1A1CC32ULL, "hlac_lt_0_lowtown_a2_n_-65538_0x.dds"},
	{0x3304D07C443297A8ULL, "vxht_bartrand_n_4bg_0_0.vlm"},
	{0x330553D256994D6EULL, "vxdr_primevaltomb_132_0_0.vlm"},
	{0x3305A7CCDF25C231ULL, "hla_ht_0_hightown_1_0z.dds"},
	{0x3305F8823EE487AAULL, "vxw_sundermount_33p_0_1.vlm"},
	{0x330618680C8488E2ULL, "6020069_f.fxe"},
	{0x330657CD6E0AE905ULL, "dae150ar_sundermount.ncs"},
	{0x3306AE55A4CCC894ULL, "436394186.ogg"},
	{0x3306CCFD859EA012ULL, "plt_dae00pt_secondary.plo"},
	{0x3306DA65CB478EF5ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_8z.dds"},
	{0x330777E7E2FA556BULL, "hlac_ht_keep_4519_11y.dds"},
	{0x3307D0EC7FF8897EULL, "gen00fl_anders6116794.cl"},
	{0x3307E8012AEAE718ULL, "740406580.ogg"},
	{0x3307FEC96F35F456ULL, "vxca_caverns2_82z_0_0.vlm"},
	{0x33083C6DCA30026CULL, "vxlt_gallowstemplar_f_6sj_0_0.vlm"},
	{0x330960181A43D134ULL, "gen_im_trs_drk_shd_met_r2.uti"},
	{0x3309EC87EA5AF525ULL, "vxht_mansion_n_3f6_0_0.vlm"},
	{0x330A590A9EB67C1FULL, "lt_0_lowtown_21_0_ll.mmh"},
	{0x330A7E0CA44424A8ULL, "hlac_w_blightlands_0_10z.dds"},
	{0x330B226255FBCE33ULL, "vxlt_0_lowtown_a2_n_c_d6y_0_0.vlm"},
	{0x330B4BB9519BFC8EULL, "591281273.ogg"},
	{0x330B572B69627E50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_ex.evt"},
	{0x330C1C08B5B1B6C1ULL, "vxht_mansion_n_4vy_0_0.vlm"},
	{0x330DB697738A4C88ULL, "6202182_f.fxe"},
	{0x330DF6EE992DE06FULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.impa.gad"},
	{0x330E0D703486342FULL, "lgt220cr_mag_templar.utc"},
	{0x330E97095283B9D9ULL, "vxht_0_hightown_56w_0_0.vlm"},
	{0x330E9ADD47D68276ULL, "200634056.ogg"},
	{0x330EE463D825F484ULL, "art\\characters\\playercharacter\\humanmale\\glv_maga.msh"},
	{0x330EF663D82613EDULL, "art\\characters\\playercharacter\\humanmale\\glv_maga.mao"},
	{0x330EFA63D8261A26ULL, "art\\characters\\playercharacter\\humanmale\\glv_maga.mmh"},
	{0x330F7566B640E2CAULL, "906905287.ogg"},
	{0x330F86BE344AD148ULL, "6093480_m.fxe"},
	{0x330F8FB6075C33E2ULL, "6067702_m.fxe"},
	{0x33100D1E7516B3B8ULL, "259818922.ogg"},
	{0x331051FDCF827B66ULL, "635928517.ogg"},
	{0x3310B76B90BE2CC1ULL, "213924378.ogg"},
	{0x3310F6A8A9D7D550ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabsnb.mao"},
	{0x3310F6AA480935BDULL, "gen00fl_varric6033540.cl"},
	{0x3310FAA8A9D7DC63ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabsnb.mmh"},
	{0x331108A8A9D7F3BDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabsnb.msh"},
	{0x33114AC1E7B2C3ACULL, "vxlt_blackemporium_133_0_0.vlm"},
	{0x33114F61F47E57DDULL, "vxca_cave_2d2_0_0.vlm"},
	{0x331158FDD9133079ULL, "vxw_sundermount_2h2_0_0.vlm"},
	{0x331159F1B55EEE9BULL, "6016721_m.fxe"},
	{0x33117C3AEEC2D31EULL, "zz_one_debug6099928.cl"},
	{0x33119F63B97AEB4BULL, "vxht_0_hightown_bka_0_0.vlm"},
	{0x3311ADE4DD2A43F1ULL, "vxlt_warehouse_n_12r_0_0.vlm"},
	{0x3311B17E6EC2E86FULL, "gen_im_npc_arm_hum_boo_rdrc.uti"},
	{0x3311BC4C650315EAULL, "vxca_cave_256_0_0.vlm"},
	{0x331251AD842A4A1AULL, "vxw_encounter2_p6_0_0.vlm"},
	{0x331254457AF32458ULL, "620608675.ogg"},
	{0x3312D31DA4AEAEF0ULL, "hlac_lt_0_lowtown_docks_n_196607_0x.dds"},
	{0x3312E184941B0EE4ULL, "w_bonepitmining_a2_163_0.msh"},
	{0x3312F384941B2D4DULL, "w_bonepitmining_a2_163_0.mao"},
	{0x3312F784941B3386ULL, "w_bonepitmining_a2_163_0.mmh"},
	{0x331301847FD655DCULL, "qun221_petrice6077367.cl"},
	{0x331318BA8B2C157CULL, "84130727.ogg"},
	{0x33131A598650C77BULL, "587493391.ogg"},
	{0x3313E792797651D8ULL, "vxlt_gallowstemplar_3di_0_0.vlm"},
	{0x33143B3D8013818EULL, "nat380_nathaniel_banter6067300.cl"},
	{0x33147B95355F6C4AULL, "hlac_lt_0_lowtown_a2_n_65536_11y.dds"},
	{0x33150702F56DA7B7ULL, "44864379.ogg"},
	{0x33150BEE9934221CULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.impa.evt"},
	{0x33151A84DDF1B10CULL, "zz_followers_banter6158694.cl"},
	{0x331569AD48A8D159ULL, "lt_0_lowtown_docks_236_0.msh"},
	{0x331573AD48A8E247ULL, "lt_0_lowtown_docks_236_0.mmh"},
	{0x331577AD48A8E894ULL, "lt_0_lowtown_docks_236_0.mao"},
	{0x3315E0BB156593C6ULL, "6172677_m.fxe"},
	{0x331650A55DAF1F0DULL, "775958358.ogg"},
	{0x33166AA5DA512B3DULL, "lt_0_lowtown_n_13_0.rml"},
	{0x3316C887112E5808ULL, "follower_banter6148473.cl"},
	{0x33177A25C9143CA4ULL, "ach_soc_romance_1.dds"},
	{0x33177D3B9E72E28CULL, "215309781.ogg"},
	{0x3317D3A8DC8896C7ULL, "vxlt_0_lowtown_docks_n_95j_0_0.vlm"},
	{0x3317E718DF8E218DULL, "vxdr_0_deeproads_bun_0_0.vlm"},
	{0x33187056E37395FDULL, "w_woundedcoast_193_0.mao"},
	{0x33187456E3739C36ULL, "w_woundedcoast_193_0.mmh"},
	{0x33187E56E373AD34ULL, "w_woundedcoast_193_0.msh"},
	{0x33189646DE72A875ULL, "vxw_sundermount_30v_0_1.vlm"},
	{0x3318A2C92BE5758EULL, "one220_yevhen6103383.cl"},
	{0x3318DAC6355D852DULL, "60946530.ogg"},
	{0x3318DC7E5DC58E4CULL, "prm000ip_curio_chest.ncs"},
	{0x3318F163D82E5FF1ULL, "art\\characters\\playercharacter\\humanmale\\glv_maga.phy"},
	{0x33191F481D17BABEULL, "6101188_f.fxe"},
	{0x331A0B76399102C8ULL, "vxlt_0_lowtown_ewy_0_0.vlm"},
	{0x331A206D99D5F09AULL, "6099325_m.fxe"},
	{0x331A5E83EA12C0DDULL, "907654862.ogg"},
	{0x331AA3BA44D2C2DDULL, "zz_area_signage8.utp"},
	{0x331B2C64F402A0D9ULL, "1026258722.ogg"},
	{0x331B4C8F69FB4F87ULL, "354086687.ogg"},
	{0x331BFAE7FDB2C46EULL, "vxht_0_hightown_a2_c4t_0_0.vlm"},
	{0x331C68540E6EDA1FULL, "vxht_0_hightown_n_bwj_0_0.vlm"},
	{0x331C8B84E38F7840ULL, "vxw_bonepit_2a3_0_0.vlm"},
	{0x331D460B94D6E49FULL, "vxlt_gallowstemplar_n_43k_0_0.vlm"},
	{0x331D61780A4C4854ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.cess.ani"},
	{0x331DC22AC3536D68ULL, "ave100st_ambush_1.stg"},
	{0x331DC3A2FA286444ULL, "6088151_f.fxe"},
	{0x331E27BBA5BE6B9EULL, "qcr000_arishok6042459.cl"},
	{0x331EA9234A3F3B17ULL, "art\\levels\\ztools\\default\\generic\\1wo_basic.phy"},
	{0x331EE12954F0883DULL, "6133043_f.fxe"},
	{0x331EF389F348715AULL, "138954732.ogg"},
	{0x331EF49D6A326F61ULL, "vxlt_0_lowtown_ers_0_0.vlm"},
	{0x331F14D967456D66ULL, "470894947.ogg"},
	{0x331F20AE3AE9E944ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.mmh"},
	{0x331F20F0ABF21A68ULL, "6150712_m.fxe"},
	{0x331F2E0A934C3B78ULL, "fen310_fenris6118258.cl"},
	{0x331F67009E11DBCEULL, "vxca_caverns2_78u_0_2.vlm"},
	{0x331FA71600E7AE7EULL, "1010166678.ogg"},
	{0x331FBC9ED9A8DC2CULL, "vxlt_blackemporium_160_0_0.vlm"},
	{0x331FCADACB0948F9ULL, "vxw_sundermount_2m1_0_0.vlm"},
	{0x33200C33D6D69EB1ULL, "24201215.ogg"},
	{0x332011E4ED207000ULL, "vxlt_0_lowtown_a2_n_c_ex9_0_0.vlm"},
	{0x3320483C02B60E2AULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_109a_d.dds"},
	{0x332054054F2A81C4ULL, "vxca_caverns2_6se_0_0.vlm"},
	{0x332099D1E5FB5C3FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_exit.gad"},
	{0x3320DCDEAC364AC3ULL, "mer141_jansen6176012.cl"},
	{0x3321043D038E0F59ULL, "vxlt_gallowstemplar_n_6sk_0_0.vlm"},
	{0x3321329B0D3119B2ULL, "94631597.ogg"},
	{0x3321B0A0A44388F1ULL, "6080742_m.fxe"},
	{0x3321B89F74F435B6ULL, "vxw_blightlands_14h_0_0.vlm"},
	{0x3321BB77940954CAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_rug_sa.dds"},
	{0x3321F3F14EE964D3ULL, "hlac_dr_slavers_3217_6x.dds"},
	{0x3321FBA8A9E65088ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabsnb.phy"},
	{0x332213EE993EBA35ULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.impa.ani"},
	{0x332250F611C2C131ULL, "w_blightlands_10_0.rml"},
	{0x332287351D104E36ULL, "follower_banter6141821.cl"},
	{0x332342371BE310ACULL, "art\\characters\\weapons\\longswords\\w_lsw_019a.phy"},
	{0x3323B10C9B4FB8FEULL, "one300_nuncio6070810.cl"},
	{0x3323ECF8AADC93B5ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medc.msh"},
	{0x3323EEF8AADC972BULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medc.mmh"},
	{0x332404C8D5DA51C6ULL, "59587255.ogg"},
	{0x33245E780A5261B6ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.cess.gad"},
	{0x3324C507FC567C8AULL, "490719370.ogg"},
	{0x3324FA7C839C9E0AULL, "6106999_m.fxe"},
	{0x33252D19835F1B79ULL, "hlac_lt_darktown_-1_67z.dds"},
	{0x33257E169FC86713ULL, "vxht_playershouse_n_1vc_0_0.vlm"},
	{0x3325F14B6DEA879EULL, "mag121_idunna6084241.cl"},
	{0x3327400DC9B6746DULL, "gen00fl_sebastian6175014.cl"},
	{0x33278ED1E601682CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_exit.evt"},
	{0x3327F47CE1E20DCAULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_lowerblank_d.dds"},
	{0x332876BB17F7F881ULL, "hlac_w_sundermount_-262139_0z.dds"},
	{0x33287D532B871B48ULL, "vxlt_0_lowtown_erv_0_0.vlm"},
	{0x3328A86AE9BFB82EULL, "vxht_chantry_2t9_0_0.vlm"},
	{0x3328BFC28AEE4B4BULL, "vxlt_gallowstemplar_5mg_0_0.vlm"},
	{0x3328E2830D6BEC87ULL, "vxw_encounter2_98_0_0.vlm"},
	{0x3329CEF96C01AA1AULL, "follower_banter6171888.cl"},
	{0x332A54A0206300A9ULL, "360231587.ogg"},
	{0x332A5ED49D6BE14EULL, "hlac_ht_mansion_n_3371_0z.dds"},
	{0x332A60E98EC119BCULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180r_rnf.ani"},
	{0x332A6CC4EFA8DCEEULL, "vxht_0_hightown_a2_c3u_0_0.vlm"},
	{0x332A79780A56FB15ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.cess.evt"},
	{0x332AA13E7E2F2284ULL, "gen_im_trs_tmp_one_lsw_r1.uti"},
	{0x332AA9F7DA8DEB5AULL, "1010276281.ogg"},
	{0x332AD6C6C6074222ULL, "734185875.ogg"},
	{0x332B441FF88CD698ULL, "6061801_m.fxe"},
	{0x332BCD555E4D0FF0ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh03.mmh"},
	{0x332BD8660F42012CULL, "325292526.ogg"},
	{0x332BDB555E4D270AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh03.msh"},
	{0x332C15F5B3205366ULL, "283504512.ogg"},
	{0x332C92D67F4CE767ULL, "plt_genpt_app_isabela.plo"},
	{0x332CDE98B9EABAB2ULL, "vxlt_0_lowtown_n_eoy_0_1.vlm"},
	{0x332CF5FA7D8FE480ULL, "6056861_m.fxe"},
	{0x332D667A5C2887E1ULL, "vxlt_0_lowtown_emr_0_0.vlm"},
	{0x332DA63C1AA0BBAEULL, "884278418.ogg"},
	{0x332E08A0C6D90123ULL, "vxlt_0_lowtown_a2_n_hcx_0_0.vlm"},
	{0x332E92408EA48936ULL, "6164703_f.fxe"},
	{0x332EB580284DFD57ULL, "123240841.ogg"},
	{0x332EEFC2F8825B05ULL, "6084706_m.fxe"},
	{0x332F2F796EBDBDDEULL, "mag320_orsino6071131.cl"},
	{0x332F382672784AE6ULL, "art\\vfx\\materials\\vfx_blood_dir02.mao"},
	{0x332FD67A5B8BD1EDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightd.phy"},
	{0x332FE9C0F8FE1477ULL, "vxht_playershouse_n_202_0_0.vlm"},
	{0x3330DE750B2335F1ULL, "gen00fl_fenris6126985.cl"},
	{0x33315A59B136326BULL, "569102314.ogg"},
	{0x333169D94820B87FULL, "828615108.ogg"},
	{0x33316DE98EC74EFEULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180r_rnf.gad"},
	{0x333220D822DB1551ULL, "l_dre180ar_ruins_crypt.wbd"},
	{0x33328B5DE7446ADAULL, "6093636_f.fxe"},
	{0x33337BE163DAAAAAULL, "vxht_chantry_2s4_0_0.vlm"},
	{0x3333BB85AD61485CULL, "vxht_chantry_n_2s1_0_0.vlm"},
	{0x3334653035D3D616ULL, "78344538.ogg"},
	{0x333466C78AA421EFULL, "dae110ip_store_armor.utp"},
	{0x33355CBFC133D6C5ULL, "hlac_lt_undercitydungeon_-1_20z.dds"},
	{0x333576D1E60D7C85ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_exit.ani"},
	{0x33363FF0396D7CECULL, "592230698.ogg"},
	{0x333664EC0AD9D872ULL, "vxw_encounter2_camp_pg_0_1.vlm"},
	{0x33368976D5D858C9ULL, "ts_kirk_lowtown_03.gda"},
	{0x3336EF301D7D33C8ULL, "vxlt_0_lowtown_emw_0_0.vlm"},
	{0x3337CF38E444925FULL, "vxdr_primevaltomb_6wh_0_0.vlm"},
	{0x33380AA64AAE39F3ULL, "gen00fl_merrill6129476.cl"},
	{0x333815C823D1CA91ULL, "llac_lt_0_lowtown_a2_n_65533_1.dds"},
	{0x33381BCCC669D592ULL, "6064088_m.fxe"},
	{0x333848E98ECD2E1DULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn180r_rnf.evt"},
	{0x3338F18A97BDC75BULL, "lt_0_lowtown_docks_a3_17_0_lc.mmh"},
	{0x333956772B7DE99AULL, "hlac_w_sundermount_262140_0x.dds"},
	{0x3339C42FF7104524ULL, "314361893.ogg"},
	{0x3339EF24E658B00BULL, "vxlt_0_lowtown_docks_a3_cxg_0_0.vlm"},
	{0x333A09E3A6328C51ULL, "hlac_ht_playershouse_n_855_24z.dds"},
	{0x333A3E85A2B6A8BBULL, "class_abilities_warrior.gda"},
	{0x333A67C9B7B4DC95ULL, "793918273.ogg"},
	{0x333AAA909D95EBDCULL, "mrl150_merrill6077755.cl"},
	{0x333AE37AF4F58764ULL, "dae000_merchant_dm_mob6174239.cl"},
	{0x333BD8574E1EA061ULL, "vxlt_0_lowtown_ehq_0_0.vlm"},
	{0x333D0AE56AFB1E5CULL, "317346655.ogg"},
	{0x333D912DC52CC2F8ULL, "vxht_bartrand_n_68k_0_0.vlm"},
	{0x333D9542EEF13394ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_loop.gad"},
	{0x333E02687BB6640FULL, "and100_anders6095850.cl"},
	{0x333E80D5E083CC38ULL, "6162512_f.fxe"},
	{0x333E90EEF271E55BULL, "6171923_m.fxe"},
	{0x333F1D3F15543E7FULL, "431408362.ogg"},
	{0x333F3025BFDE650BULL, "6106666_m.fxe"},
	{0x333F53DEF07FF64CULL, "982814280.ogg"},
	{0x333F7CA567FCCDB9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzarda_d.dds"},
	{0x333FBDF9F5916BE7ULL, "827215842.ogg"},
	{0x333FEA2BBAB5EA0EULL, "6030683_m.fxe"},
	{0x33400DC1C4B12116ULL, "550005425.ogg"},
	{0x334083148112665CULL, "follower_banter6143267.cl"},
	{0x33419652471C9103ULL, "lgt210_magu_selby6142034.cl"},
	{0x3341C87CC76B588BULL, "961984129.ogg"},
	{0x33422B856CB7E0CEULL, "6059699_m.fxe"},
	{0x33432893FCAC3A22ULL, "tbe100im_desdemona_1h.uti"},
	{0x33434CE797C1F2F4ULL, "6071127_m.fxe"},
	{0x33434DD0F82EA6D9ULL, "vxht_keep_3b9_0_0.vlm"},
	{0x3343A5D91B22BF85ULL, "follower_banter6154101.cl"},
	{0x3343B3FBB2335725ULL, "vxht_mansion_n_3r2_0_0.vlm"},
	{0x33444A49E6F932B7ULL, "mrl210_merrill6166794.cl"},
	{0x33445E9D66C6833FULL, "shock_projectile_f.dds"},
	{0x3344A442EEF76C72ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_loop.ani"},
	{0x3344D200FA345F2EULL, "818617317.ogg"},
	{0x3344E452969C6CDDULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urva_0d.dds"},
	{0x334517F7E94337A8ULL, "rdr111dg_martin6014878.cl"},
	{0x3345610D0F734C48ULL, "vxlt_0_lowtown_eht_0_0.vlm"},
	{0x33460279345D4D39ULL, "6047383_m.fxe"},
	{0x334634C02A9390EEULL, "gen_im_trs_krk_grd_junk1_02.uti"},
	{0x33464115D63AAADFULL, "vxdr_primevaltomb_6zk_0_0.vlm"},
	{0x334644886D4388E0ULL, "441398481.ogg"},
	{0x33473BD92DDE1207ULL, "vxw_sundermount_2bb_0_0.vlm"},
	{0x334754DC101C52D0ULL, "hf_barwench.mor"},
	{0x3348098F6030B275ULL, "vxw_woundedcoast_1da_0_0.vlm"},
	{0x334823AE3B0CF33BULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.anb"},
	{0x33484F2BABB27218ULL, "vxw_bonepitmining_a2_9u_0_0.vlm"},
	{0x334867AD9A817D7DULL, "145016914.ogg"},
	{0x3349140DE36BEEF0ULL, "gen_im_npc_hum_arm_cht_chnb.uti"},
	{0x33492D7A19C58214ULL, "792536703.ogg"},
	{0x3349469B247380B0ULL, "dae000_bartrand6206153.cl"},
	{0x3349780B7506F2E6ULL, "hlac_w_bonepitdestroyed_a3_196613_0x.dds"},
	{0x33498917A5FD3175ULL, "vxlt_undercityencounter_123_0_1.vlm"},
	{0x334A554652159B6EULL, "hlac_lt_undercityserialkiller_-1_42x.dds"},
	{0x334A5A05DA34AEB0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sworda_n.dds"},
	{0x334AF1BC60E3B987ULL, "zz_vrc_debug6200265.cl"},
	{0x334AF9440199BC34ULL, "366062371.ogg"},
	{0x334B775F3A02097EULL, "vxlt_gallowscourtyard_4ln_0_0.vlm"},
	{0x334BBE1A791F9C74ULL, "6117956_f.fxe"},
	{0x334BFEDFC96AAB2DULL, "vxw_encounter1_bv_0_0.vlm"},
	{0x334C6D7CC64DE73DULL, "gen_im_trs_cot_med_hlm_r0.uti"},
	{0x334C8832B2C1B54DULL, "vxht_mansion2b_3do_0_0.vlm"},
	{0x334CB001FDA145ADULL, "dre000_sibling_has_blight6036516.cl"},
	{0x334D7AB5A2103334ULL, "static_static_p1.vsh.fxo"},
	{0x334D85A327A58D82ULL, "dae000_hate_enemies6167624.cl"},
	{0x334D95C1BC147183ULL, "6083402_m.fxe"},
	{0x334DAE54AFFFF436ULL, "vxlt_0_lowtown_docks_a3_aqw_0_0.vlm"},
	{0x334E1810ADD6A75BULL, "hlac_dr_0_deeproads_5845_2y.dds"},
	{0x334EBF9A4B74EF1DULL, "694301648.ogg"},
	{0x334EC85493A5AD3AULL, "6141671_m.fxe"},
	{0x334EDD080C141A22ULL, "743174240.ogg"},
	{0x334FB9F42B8E1AFCULL, "vxht_0_hightown_b76_0_0.vlm"},
	{0x3350250BEB819569ULL, "lgt310_mag_melson6152131.cl"},
	{0x3351288C3CF212A2ULL, "w_woundedcoast_ll.mmh"},
	{0x33515F2EBC3822F1ULL, "6047677_m.fxe"},
	{0x33523D75939A6297ULL, "vxlt_0_lowtown_docks_b4t_0_1.vlm"},
	{0x3352EA92B6F54548ULL, "6172675_m.fxe"},
	{0x335308AEAAF3E1E2ULL, "6122760_m.fxe"},
	{0x3353C99CFD9FFB73ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic_4.mao"},
	{0x33543C8C88A5A847ULL, "gen_im_arm_cht_comc_t2.uti"},
	{0x335461064F570B55ULL, "follower_banter6171904.cl"},
	{0x33546D004BDBD7E3ULL, "707733505.ogg"},
	{0x33547D051EE755D7ULL, "follower_barks6153798.cl"},
	{0x3354FF79EF5ABE74ULL, "6131209_m.fxe"},
	{0x33551269B73A0E16ULL, "follower_banter6163267.cl"},
	{0x33552AD0B033A692ULL, "6174745_m.fxe"},
	{0x33552F859D8B7672ULL, "6120137_m.fxe"},
	{0x335530A6A5F8A531ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean.mmh"},
	{0x33553CA6A5F8B9D2ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean.mao"},
	{0x33554AA6A5F8D1EFULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean.msh"},
	{0x3355F38489EAB402ULL, "vxlt_warehouse_n_oq_0_0.vlm"},
	{0x335621638B5E862DULL, "55763344.ogg"},
	{0x33565EA574A33D10ULL, "vxdr_slavers_11w_0_1.vlm"},
	{0x335675A068805F8FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls103a_0s.dds"},
	{0x33579FF3E40411BAULL, "vxlt_warehouse_n_14o_0_0.vlm"},
	{0x3357B79D51F50604ULL, "vxw_sundermount_31a_0_0.vlm"},
	{0x3357D82CA92C89BEULL, "621965809.ogg"},
	{0x3357FB5BDA2B5096ULL, "gen00fl_isabela6101064.cl"},
	{0x33582086D1E6F423ULL, "lt_0_lowtown_docks_n_17_0_lc.mmh"},
	{0x33582CA157981932ULL, "hlac_lt_gallowstemplar_2394_11x.dds"},
	{0x3358634E350CBD88ULL, "vxht_brothel_n_5io_0_0.vlm"},
	{0x33587FBF2672A786ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_beyondpossession_e.dds"},
	{0x3358BCF3AEC2958EULL, "dae011_corff_sta6164629.cl"},
	{0x3358BF93B0631FA6ULL, "mag370_meredith6039273.cl"},
	{0x335A10E92E40C7A9ULL, "hlac_lt_warehouse_n_1_0z.dds"},
	{0x335A2C52A29AE83FULL, "hlac_ht_0_hightown_-131074_1x.dds"},
	{0x335A70BCBB60C3ADULL, "vxw_encounter1_gu_0_0.vlm"},
	{0x335A7755388A1789ULL, "850708348.ogg"},
	{0x335A7A386313BE6FULL, "6130995_m.fxe"},
	{0x335ADB818D498AABULL, "681768043.ogg"},
	{0x335B1B5EF58D742BULL, "vxca_caverns2_8h0_0_1.vlm"},
	{0x335B57ACE0E0798CULL, "vxlt_0_lowtown_ci5_0_0.vlm"},
	{0x335C1054ACB12DC7ULL, "vxht_0_hightown_c0i_0_0.vlm"},
	{0x335C1F9078A6A525ULL, "6039392_f.fxe"},
	{0x335CF7184FDCB916ULL, "741284525.ogg"},
	{0x335CFAD773A7F28EULL, "vxca_caverns1_eb8_0_0.vlm"},
	{0x335DCAB597A793EFULL, "829402392.ogg"},
	{0x335E1910CA0722BAULL, "vxht_0_hightown_a2_9aq_0_0.vlm"},
	{0x335F9435C49FDA0FULL, "gen_im_arm_boo_med_set_high.uti"},
	{0x335FA79FF101B4B5ULL, "hlac_lt_gallowstemplar_n_2384_4y.dds"},
	{0x335FBAA641183F8DULL, "dae000_keep_servant_1_sta6156888.cl"},
	{0x335FD10BAE2E3B71ULL, "6042977_m.fxe"},
	{0x335FD3124FB31132ULL, "vxht_0_hightown_n_b19_0_0.vlm"},
	{0x335FE1EB03DC619CULL, "mag360_samson6154312.cl"},
	{0x336060527CFBCF71ULL, "vxht_0_hightown_a2_9zs_0_0.vlm"},
	{0x336067C961319498ULL, "6142892_m.fxe"},
	{0x3360DDF6B0A0E2ECULL, "sst_ss_taraneh.gda"},
	{0x33611319F8C8A058ULL, "6068576_m.fxe"},
	{0x3361837EE8D9D516ULL, "vxw_encounter3_da_0_0.vlm"},
	{0x3361D5CCAFCF8D31ULL, "vxw_blightlands_e1_0_0.vlm"},
	{0x33621D553E9E4362ULL, "hlac_ht_keep_n_4519_12x.dds"},
	{0x33621EB570ABF063ULL, "6092278_f.fxe"},
	{0x33623AC808E04223ULL, "335157924.ogg"},
	{0x3362A8C18508E922ULL, "w_bonepitmining_a2_9_0_ll.mmh"},
	{0x3362C9C9F87A6D75ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_023a_0s.dds"},
	{0x336340D0814CA4A5ULL, "vxca_cavern1_haunted_ehh_0_0.vlm"},
	{0x33634450A2E470B8ULL, "6036489_m.fxe"},
	{0x336358ACA7767A17ULL, "182252269.ogg"},
	{0x3363A061A320B083ULL, "llac_w_encounter2_camp_131077_0.dds"},
	{0x3363C53928E72775ULL, "mag370_meredith6185100.cl"},
	{0x3363CD90BAD2FFFAULL, "hlac_lt_gallowscourtyard_n_-131071_1x.dds"},
	{0x3363D07788B526FFULL, "hlac_lt_gallowstemplar_2394_9z.dds"},
	{0x336419371C1A6277ULL, "art\\characters\\weapons\\longswords\\w_lsw_019a.mmh"},
	{0x33641D371C1A6884ULL, "art\\characters\\weapons\\longswords\\w_lsw_019a.mao"},
	{0x33642F371C1A87E9ULL, "art\\characters\\weapons\\longswords\\w_lsw_019a.msh"},
	{0x33644052977FEE5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_knife_to_throat_loop.gad"},
	{0x336465617BE0CC82ULL, "vxlt_warehouse_n_rr_0_0.vlm"},
	{0x3364B0A111A98486ULL, "follower_banter6155727.cl"},
	{0x3364D38875A82D02ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_3y.dds"},
	{0x3364E28B2A52F458ULL, "6098067_m.fxe"},
	{0x33657E234A7AF56FULL, "art\\levels\\ztools\\default\\generic\\1wo_basic.mao"},
	{0x33658A234A7B090CULL, "art\\levels\\ztools\\default\\generic\\1wo_basic.mmh"},
	{0x33658A590337797AULL, "vxht_mansion_n_6im_0_0.vlm"},
	{0x336590234A7B13FEULL, "art\\levels\\ztools\\default\\generic\\1wo_basic.msh"},
	{0x336611D0D5FA2A3AULL, "vxlt_warehouse_n_19h_0_0.vlm"},
	{0x33661CD3117030D7ULL, "hlac_ca_caverns2_5802_2z.dds"},
	{0x3366B49BF5FFBE87ULL, "qun110_petrice6041274.cl"},
	{0x3366FB7E5E07E44DULL, "prm000ip_curio_chest.utp"},
	{0x336702AE20FB495FULL, "6198702_m.fxe"},
	{0x33676EBA8A36008AULL, "6141569_m.fxe"},
	{0x33677060A88DFCCFULL, "art\\characters\\creatures\\desiredemon\\animation\\blendtree\\desiredemon_overlay.anb"},
	{0x3367CDE70B25111EULL, "76572610.ogg"},
	{0x3367E2A784B425B3ULL, "6067376_m.fxe"},
	{0x3368447AF36A4E82ULL, "393012411.ogg"},
	{0x336866C7B5D4DF33ULL, "combat_deathblow_8.gda"},
	{0x33686BC3AAE27DE0ULL, "gen_im_npc_arm_hum_cht_gwdc.uti"},
	{0x3368805B0D733641ULL, "595426745.ogg"},
	{0x336892FCC5428C71ULL, "vxlt_backalley_n_1e2_0_0.vlm"},
	{0x3368DAAD443C21A7ULL, "6118104_m.fxe"},
	{0x336917C1B4E51856ULL, "character_character_p160.psh.fxo.dx11"},
	{0x336933643234A280ULL, "6152079_m.fxe"},
	{0x336960F29BA4B3F9ULL, "hlac_ht_mansion2_818_27x.dds"},
	{0x336970D33977059BULL, "6166390_m.fxe"},
	{0x33697D9072CE5796ULL, "vxlt_gallowstemplar_n_46u_0_0.vlm"},
	{0x3369DD7259190EF8ULL, "hlac_lt_gallowscourtyard_-1_4x.dds"},
	{0x336A3F529784585AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_knife_to_throat_loop.ani"},
	{0x336A82319EA74647ULL, "vxht_0_hightown_c7h_0_0.vlm"},
	{0x336AA7E272DE07AFULL, "216702924.ogg"},
	{0x336AF42B7F506E93ULL, "fv.cut"},
	{0x336B3A0AC75388F0ULL, "6102179_m.fxe"},
	{0x336BB05678834267ULL, "w_encounter2_camp_46_0.rml"},
	{0x336BE8BF41947485ULL, "vxht_bartrand_n_68r_0_0.vlm"},
	{0x336BEA5C08513AB1ULL, "gen00fl_anders6104020.cl"},
	{0x336C3A6F5CE26FA6ULL, "6095574_m.fxe"},
	{0x336C45E84047C321ULL, "vxlt_0_lowtown_docks_a3_buk_0_0.vlm"},
	{0x336C663B6FF2C2E0ULL, "6170488_m.fxe"},
	{0x336CD4555E84B20BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh03.phy"},
	{0x336D869C686A5809ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickenter.ani"},
	{0x336DB77CE9C420BBULL, "6114841_m.fxe"},
	{0x336DEB94BFA27CB0ULL, "hlac_lt_undercitydungeon_-1_13y.dds"},
	{0x336DEBC35AFF59A0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lowdoor_d.dds"},
	{0x336DEDAC4544DD51ULL, "48778935.ogg"},
	{0x336EC3FAE350BAD9ULL, "lowtown_argument6169449.cl"},
	{0x336F04327DEA7639ULL, "6176667_m.fxe"},
	{0x336F79CEB3CD46CEULL, "157751268.ogg"},
	{0x336F9295C162A669ULL, "vxw_bonepit_1pz_0_0.vlm"},
	{0x336FF3776EE0BD73ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicechair_0.pwk"},
	{0x336FF55BDACFED96ULL, "vxw_encounter3_ab_0_0.vlm"},
	{0x33700C9875BF1917ULL, "vxdr_0_deeproads_9c9_0_0.vlm"},
	{0x33700CC365B44044ULL, "hlac_lt_gallowstemplar_f_2391_4z.dds"},
	{0x33703DE3E1AAD8A7ULL, "qun110_varnell.bnk"},
	{0x337047A9A1C5A5B1ULL, "vxw_blightlands_n0_0_0.vlm"},
	{0x3371C60B18AA0A4DULL, "zz_followers_banter6157869.cl"},
	{0x3372AF5678895F7FULL, "w_encounter2_camp_46_0.phy"},
	{0x3372D73E6DD6E502ULL, "vxlt_warehouse_n_ys_0_0.vlm"},
	{0x3372EA64A6D6C9D0ULL, "6167516_m.fxe"},
	{0x337314CDD8ECB165ULL, "gen_im_wep_mag_sta_act0_01.uti"},
	{0x33731635D11191BBULL, "29230573.ogg"},
	{0x33733D445A8FA3BAULL, "al_mage_s03.gda"},
	{0x33736FBA2DC9FCE7ULL, "hlac_lt_backalley_n_1_5x.dds"},
	{0x337372714DEE8C00ULL, "6088329_m.fxe"},
	{0x3373FC35F52D91FAULL, "vxht_mansion_n_6ln_0_0.vlm"},
	{0x337430F29F670D3EULL, "dae000_pc_home_ambients6208507.cl"},
	{0x33744F3E3CC3859AULL, "gen00fl_anders6119698.cl"},
	{0x33745FE3E1AEFB1FULL, "qun110_varnell.cnv"},
	{0x337484656CA7B05AULL, "163578860.ogg"},
	{0x3374A19C6870A48BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickenter.gad"},
	{0x3374B6DC29121365ULL, "dae000cr_grand_cleric6183963.cl"},
	{0x3374E6002CF42CFEULL, "qun111_ketojan6041135.cl"},
	{0x3374FF43B2555E40ULL, "hlac_w_encounter3_-262144_1x.dds"},
	{0x3375945439336F5BULL, "hlac_ht_mansion2_818_48y.dds"},
	{0x3375B772F1D7F0A1ULL, "dae000ip_crowd_lm5_talk8.utp"},
	{0x3375C3CC950A6A0AULL, "hlac_lt_foundry_n_320_11y.dds"},
	{0x3375C4360D12CCADULL, "131995074.ogg"},
	{0x3375E35923EBB183ULL, "vxca_caverns2prc_8q7_0_0.vlm"},
	{0x337614B729FEA29BULL, "224086939.ogg"},
	{0x33764295F7C16967ULL, "vxdr_0_deeproads_bhg_0_0.vlm"},
	{0x3376548476AA3E33ULL, "6062076_m.fxe"},
	{0x3377DB6A527E2A0AULL, "6072380_m.fxe"},
	{0x33781E4678BB8C67ULL, "target_wps.gda"},
	{0x337891BF4DEDC80FULL, "219433333.ogg"},
	{0x3378F2A1EDAC58C8ULL, "vxca_cavern1_haunted_j2f_0_2.vlm"},
	{0x33790381B456C130ULL, "377249655.ogg"},
	{0x3379CB91D84A3DEAULL, "6085993_m.fxe"},
	{0x3379EAF4D21C790AULL, "vxw_woundedcoast_oj_0_0.vlm"},
	{0x337A44CBF54C977AULL, "3387108.ogg"},
	{0x337AF85D5806287FULL, "6175978_m.fxe"},
	{0x337B02B0E2BA7726ULL, "69437113.ogg"},
	{0x337B20077D36D9F8ULL, "mag160_feynriel6020450.cl"},
	{0x337B2ABDCA7546E2ULL, "6208472_f.fxe"},
	{0x337B9E9C6876BEF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickenter.evt"},
	{0x337BC397EB4E9819ULL, "follower_banter6163640.cl"},
	{0x337C130A85B94CF2ULL, "gen00fl_aveline6132931.cl"},
	{0x337C2959DBBA393BULL, "6113941_m.fxe"},
	{0x337C6CA13B7FFD53ULL, "hlac_lt_0_lowtown_a3_n_c_65534_1z.dds"},
	{0x337C7FF9C4180C96ULL, "vxht_mansion2_35k_0_0.vlm"},
	{0x337CB7B2EDAB539AULL, "277625773.ogg"},
	{0x337D6C587BE106D8ULL, "63865723.ogg"},
	{0x337DE781A77DEB81ULL, "vxlt_0_lowtown_a2_n_c_ei7_0_0.vlm"},
	{0x337E0FC07E6CC16DULL, "zz_dre_debug6092002.cl"},
	{0x337E7E7567B53197ULL, "vxdr_0_deeproads_9d8_0_0.vlm"},
	{0x337EB98693BBBE31ULL, "vxw_blightlands_s7_0_0.vlm"},
	{0x337EC2B149FED2DFULL, "6151323_m.fxe"},
	{0x337F4CBA76E71225ULL, "vxht_chantry_n_2ox_0_0.vlm"},
	{0x337F53C7DD4D1795ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_0z.dds"},
	{0x3380507533A39DB8ULL, "lgt210_qun_blk_qunari6142300.cl"},
	{0x3380CD62F6B9C8AFULL, "hlac_dr_primevaltomb_661_4z.dds"},
	{0x3381A50F532BC87CULL, "vxdr_primevaltomb_3do_0_0.vlm"},
	{0x3381BC28C836C1C8ULL, "gen00fl_anders6136008.cl"},
	{0x3381D10328091475ULL, "vxw_woundedcoast_1pm_0_0.vlm"},
	{0x3382A24DBE6D9D37ULL, "vxlt_darktown_4h5_0_0.vlm"},
	{0x3382A3404F818B13ULL, "196995015.ogg"},
	{0x3383338539857D6EULL, "vis121_saemus6167894.cl"},
	{0x33833C8DDA28AECFULL, "llac_lt_0_lowtown_a2_131072_0.dds"},
	{0x338342701FD9AFE8ULL, "946010626.ogg"},
	{0x3383B0567897D384ULL, "w_encounter2_camp_46_0.mmh"},
	{0x3383B4567897DAD7ULL, "w_encounter2_camp_46_0.mao"},
	{0x3383C545BBD855E3ULL, "dae000_epic_merchant_36179502.cl"},
	{0x3383C6567897F926ULL, "w_encounter2_camp_46_0.msh"},
	{0x3383F330DA3B357FULL, "llac_lt_0_lowtown_docks_n_-65537_1.dds"},
	{0x33844998D38AFC67ULL, "gen00fl_fenris6115343.cl"},
	{0x338490FA9CC0817FULL, "600092176.ogg"},
	{0x338506BDC23DB856ULL, "vxdr_primevaltomb_6xt_0_0.vlm"},
	{0x338576B6A92EBD71ULL, "vxlt_backalley_n_1k0_0_0.vlm"},
	{0x33858392D0527E16ULL, "vxht_0_hightown_n_bzx_0_0.vlm"},
	{0x3385AD6039660F72ULL, "lt_0_lowtown_a2_19_0.anb"},
	{0x3385BE946F891026ULL, "vxht_mansion_n_32e_0_0.vlm"},
	{0x3386D308D5E633DCULL, "vxlt_0_lowtown_a2_n_c_exe_0_0.vlm"},
	{0x338711B0885B28EEULL, "66677376.ogg"},
	{0x338715E7F834A359ULL, "6033726_m.fxe"},
	{0x33876CE9B9E4F1E5ULL, "895785337.ogg"},
	{0x3387EB04B60C2024ULL, "vxlt_gallowstemplar_n_416_0_0.vlm"},
	{0x3388436687D390ECULL, "hlac_lt_undercityserialkiller_-1_4z.dds"},
	{0x33885CD1C412918AULL, "vxw_woundedcoast_dk_0_0.vlm"},
	{0x3388A0092DB6DDE4ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc106a_0n.dds"},
	{0x338A0A8E9901D7A9ULL, "vxlt_0_lowtown_docks_c4j_0_0.vlm"},
	{0x338ADED9B7EF31D6ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdazed.evt"},
	{0x338B0B9A96FFCE00ULL, "dre000_sibling_has_blight6036498.cl"},
	{0x338BB529EAC88997ULL, "vxlt_0_lowtown_n_fe8_0_0.vlm"},
	{0x338BC3049A353EECULL, "hlac_ca_caverns2prc_5802_2z.dds"},
	{0x338BEB64D1A1D1C3ULL, "vxca_caverns2_485_0_0.vlm"},
	{0x338CB0566079F842ULL, "232511068.ogg"},
	{0x338CFCBD8AAF3B14ULL, "6091936_m.fxe"},
	{0x338D0102233EB71AULL, "334096973.ogg"},
	{0x338D042976FD8001ULL, "6076480_m.fxe"},
	{0x338D2B6385B1D6B1ULL, "vxw_blightlands_t6_0_0.vlm"},
	{0x338D96407CE51F90ULL, "vxht_0_hightown_a2_ca5_0_0.vlm"},
	{0x338DABC650F8B28AULL, "art\\levels\\races\\design\\animation\\beartrap.active.gad"},
	{0x338DD5DA99EFD956ULL, "596822675.ogg"},
	{0x338DF434DD761F2AULL, "art\\vfx\\materials\\chunky\\vfx_c_crater_weapon.mao"},
	{0x338E2AEA680F95B0ULL, "mcr310_argument6074359.cl"},
	{0x338E57761A22F390ULL, "6113887_m.fxe"},
	{0x338E6BB659823A2FULL, "537945282.ogg"},
	{0x338E97FA61AC6D5AULL, "vxlt_0_lowtown_a2_n_c_gc4_0_0.vlm"},
	{0x338EF88889BDD870ULL, "hlac_lt_0_lowtown_a3_n_c_-2_6z.dds"},
	{0x338F1E4684DECF90ULL, "6099190_m.fxe"},
	{0x338F3BD8053A53A4ULL, "vxca_caverns1_ike_0_0.vlm"},
	{0x338F3D276F3910ADULL, "662512498.ogg"},
	{0x338F7E503BF1BCD1ULL, "gen_im_flw_two_fenris_01.uti"},
	{0x338F837C8A954746ULL, "vxlt_gallowsprison_n_6hu_0_0.vlm"},
	{0x339045A73CCAF4B3ULL, "ser121_jethann6104607.cl"},
	{0x339144F99A6B95D5ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_005a_0s.dds"},
	{0x33914BB82C8D0733ULL, "dre000st_anders.stg"},
	{0x33918F2E5D6B73BCULL, "6170790_m.fxe"},
	{0x3391D5D9B7F54161ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdazed.gad"},
	{0x3391D824150E9621ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180_alt.ani"},
	{0x33920A3A1D8D819DULL, "184049667.ogg"},
	{0x3392164028D9FFDEULL, "hlac_w_sundermount_393214_0y.dds"},
	{0x33924C97859B84A2ULL, "vxht_mansion_n_5bw_0_0.vlm"},
	{0x33925DE5F1ED4A37ULL, "116283127.ogg"},
	{0x3392C6FA095EE3F4ULL, "gen00fl_varric6079798.cl"},
	{0x33936C716C7761D8ULL, "vxht_0_hightown_a2_auo_0_0.vlm"},
	{0x3393824C6DE02549ULL, "6068843_m.fxe"},
	{0x33939052C778B733ULL, "gen00fl_aveline6139902.cl"},
	{0x3393E5C78650157EULL, "6061797_m.fxe"},
	{0x3394C04C86A66349ULL, "gen00fl_varric6033621.cl"},
	{0x3394D2D6203AA99BULL, "803603674.ogg"},
	{0x3394E1C0E89B892EULL, "gen00fl_carver6097224.cl"},
	{0x3394F083F5BECF1FULL, "279422636.ogg"},
	{0x33951467F4EC12E3ULL, "hlac_lt_undercitydungeon_-1_23y.dds"},
	{0x33955B28BBC9A18AULL, "vxw_sundermount_1yq_0_0.vlm"},
	{0x3395F14A5EE80315ULL, "207113551.ogg"},
	{0x3396458B115BC073ULL, "6140215_m.fxe"},
	{0x33968DAA9AC89C80ULL, "vxlt_0_lowtown_a2_n_c_hoi_0_0.vlm"},
	{0x3396A3FA27DB27A5ULL, "911275691.ogg"},
	{0x3396B8A1BBC31A8CULL, "vxlt_undercitydungeon_22i_0_0.vlm"},
	{0x3396CEAEB608AA0AULL, "vxw_woundedcoast_ad_0_0.vlm"},
	{0x3396D30480436F54ULL, "mag291_vincento6166238.cl"},
	{0x3397003D8EFEA7C8ULL, "609542073.ogg"},
	{0x3397D1470A1C5AC4ULL, "vxht_brothel_n_618_0_0.vlm"},
	{0x3397D6D9B7F9AE0FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdazed.ani"},
	{0x3397E61FD6569906ULL, "dae000_gamlen6005778.cl"},
	{0x3398197ABC4419BDULL, "hlac_ca_caverns1_mines_24424_1x.dds"},
	{0x339831C3E5D3B4DCULL, "vxlt_0_lowtown_a2_n_c_h22_0_0.vlm"},
	{0x3398E3241514C7C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180_alt.gad"},
	{0x3398E382271FA3AFULL, "vxlt_warehouse_n_w8_0_0.vlm"},
	{0x3398F2F2E8925059ULL, "vxlt_backalley_g_1pa_0_1.vlm"},
	{0x3399685E0E448F76ULL, "vxlt_hangedman_n_10k_0_0.vlm"},
	{0x33999C2696C6E6ABULL, "hlac_w_sundermount_65538_0y.dds"},
	{0x3399C9477B72941EULL, "gen00fl_anders6111364.cl"},
	{0x3399CC7E1BD27A3EULL, "346534537.ogg"},
	{0x3399D5141987D548ULL, "31095314.ogg"},
	{0x339A07DECE79ACA8ULL, "vxw_woundedcoast_1f2_0_0.vlm"},
	{0x339A86D36E38993DULL, "mag360_samson6172232.cl"},
	{0x339AC5E75B320E79ULL, "682954160.ogg"},
	{0x339AE773B4FD4597ULL, "ave331cr_insurgent_guard.utc"},
	{0x339B1B28518F80BAULL, "211039.ogg"},
	{0x339B5C0FCCC7FD05ULL, "mer120_hubert6059401.cl"},
	{0x339B6E9A7CA55394ULL, "6096836_m.fxe"},
	{0x339B77444051BF05ULL, "805051914.ogg"},
	{0x339B82615FEDA190ULL, "vxca_caverns1_efz_0_0.vlm"},
	{0x339BB15483251CF4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_reaction_left_shoulder.ani"},
	{0x339BB171CC7F7167ULL, "hlac_lt_0_lowtown_docks_n_-1_2x.dds"},
	{0x339BCFD3EDF017B1ULL, "gen_im_npc_arm_hum_cht_thga.uti"},
	{0x339BDAAE3F8F3C40ULL, "207961410.ogg"},
	{0x339C7FA17D5D552FULL, "hlac_w_encounter2_65541_0z.dds"},
	{0x339CB0CE81AFEF18ULL, "6107225_f.fxe"},
	{0x339CBF8F52DCA061ULL, "gen_im_wep_rog_dua_iconic.uti"},
	{0x339CFA2DFE07DFEFULL, "plt_lgt220pt_shp_mag_letter.plo"},
	{0x339D13036E146ED5ULL, "pro000_friedrich6050189.cl"},
	{0x339D39EF5178E13CULL, "6118411_m.fxe"},
	{0x339DC6B3B04AE4CCULL, "cd_seb321_rebel_mage.bnk"},
	{0x339E53FCDE64F7AAULL, "isa111_isabela.cnv"},
	{0x339E67F9F0A08A1EULL, "261361634.ogg"},
	{0x339E941BA1752916ULL, "hlac_lt_0_lowtown_n_-65535_4y.dds"},
	{0x339EF3D9F9A22145ULL, "vxht_mansion_n_6le_0_0.vlm"},
	{0x339F2E24FC109205ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg105a_0d.dds"},
	{0x339F4FD0C2310D0DULL, "vxlt_0_lowtown_docks_a3_n_97m_0_0.vlm"},
	{0x339FC024151AAAA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180_alt.evt"},
	{0x33A02C3B80BD084AULL, "hlac_lt_0_lowtown_docks_a3_-3_1x.dds"},
	{0x33A074742E9DD350ULL, "mag110_samson6018587.cl"},
	{0x33A0C3064057E2F3ULL, "art\\characters\\playercharacter\\humanmale\\arm_masf.phy"},
	{0x33A13B324B0A0FA6ULL, "hlac_w_encounter3_-131070_0z.dds"},
	{0x33A14070B7CF93C6ULL, "6096764_m.fxe"},
	{0x33A186C651091D70ULL, "art\\levels\\races\\design\\animation\\beartrap.active.ani"},
	{0x33A1AE5483298356ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_reaction_left_shoulder.gad"},
	{0x33A235FCDE68ADD2ULL, "isa111_isabela.bnk"},
	{0x33A28627467BEF3BULL, "6101325_f.fxe"},
	{0x33A31B5CB936143EULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.cess.ani"},
	{0x33A371C8A77CB3C3ULL, "gen00fl_merrill6130131.cl"},
	{0x33A3DCD2BDEBD838ULL, "gen00fl_player6207708.cl"},
	{0x33A4B242A67D8F96ULL, "vxdr_0_deeproads_cuf_0_2.vlm"},
	{0x33A5310FB2FEB329ULL, "vxdr_slavers_1jf_0_0.vlm"},
	{0x33A54A118B3852ACULL, "lowtown_fem_alienage1_sta6162560.cl"},
	{0x33A56E70FA25AD2DULL, "vxw_bonepitdestroyed_a3_28e_0_0.vlm"},
	{0x33A5F66E038696F1ULL, "6208577_m.fxe"},
	{0x33A61883B83DBC13ULL, "dae121tr_alema.utt"},
	{0x33A6EBDD400C9FB8ULL, "zz_followers_banter6158283.cl"},
	{0x33A740AA4F3B999BULL, "hla_ht_0_hightown_a2_-131071_0y.dds"},
	{0x33A77660BDEDF343ULL, "820342815.ogg"},
	{0x33A790AFDF175E0EULL, "hlac_w_encounter1_-65537_0x.dds"},
	{0x33A838F2FDE637C1ULL, "631352113.ogg"},
	{0x33A86D99FDA027E2ULL, "static_static_p36.psh.fxo"},
	{0x33A879BBC06FC528ULL, "vxw_woundedcoast_1a3_0_0.vlm"},
	{0x33A8B2081EA53260ULL, "6016829_m.fxe"},
	{0x33A8B8869F48CB12ULL, "vxlt_gallowstemplar_f_4qh_0_0.vlm"},
	{0x33A8C954832FCFB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_reaction_left_shoulder.evt"},
	{0x33AAE4E54F3E71CFULL, "348718552.ogg"},
	{0x33AB18B88CCE7F85ULL, "hlac_lt_0_lowtown_a2_n_-131073_1z.dds"},
	{0x33AB96BA977107FFULL, "185356526.ogg"},
	{0x33ABD16D80259DCCULL, "zz_ser_debug6085770.cl"},
	{0x33ABFB8A6813FBE9ULL, "6142797_f.fxe"},
	{0x33AC1A03EC9E378EULL, "vxw_woundedcoast_76_0_0.vlm"},
	{0x33AC718BF13BABA2ULL, "vxlt_hangedman_1qw_0_0.vlm"},
	{0x33AC9BE1BE36363FULL, "412669492.ogg"},
	{0x33AD6CB71BDDAB65ULL, "achievements.gda"},
	{0x33ADEFC4251C7EB1ULL, "t_table_01.dds"},
	{0x33ADF30B9237FC5AULL, "6093730_m.fxe"},
	{0x33AF152AE8FADB96ULL, "6048437_m.fxe"},
	{0x33AF30516987B5A2ULL, "vxht_mansion_n_5hq_0_0.vlm"},
	{0x33B009D1AC7EBEE9ULL, "605559083.ogg"},
	{0x33B0335CB940C70FULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.cess.evt"},
	{0x33B06B3772D93A36ULL, "w_woundedcoast_208_0.mao"},
	{0x33B0773772D94ED5ULL, "w_woundedcoast_208_0.mmh"},
	{0x33B0793772D9524BULL, "w_woundedcoast_208_0.msh"},
	{0x33B08DD69389539CULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relax.ani"},
	{0x33B0C16DC22231E5ULL, "llac_ht_mansion2b_818_29.dds"},
	{0x33B0CE1939C8905DULL, "zz_vault_debug6200341.cl"},
	{0x33B0DCC019D1AA68ULL, "347032812.ogg"},
	{0x33B0F959B379FB75ULL, "63632184.ogg"},
	{0x33B1791B6400C2D3ULL, "character_character_p264.psh.fxo.dx11"},
	{0x33B18EE3ADBE1152ULL, "6095840_m.fxe"},
	{0x33B2289FABC87D5CULL, "vxlt_0_lowtown_a2_n_c_ewx_0_0.vlm"},
	{0x33B23C7C1231B7BAULL, "qcr000_varric6208228.cl"},
	{0x33B2724A31AFD27AULL, "377781140.ogg"},
	{0x33B29B481D911ADBULL, "6089821_f.fxe"},
	{0x33B36EF37447E3EEULL, "397032620.ogg"},
	{0x33B38918A23ED6D2ULL, "1046708619.ogg"},
	{0x33B39CEC39045856ULL, "721953474.ogg"},
	{0x33B3C97816C0B453ULL, "dae011_badpoet_sta6145178.cl"},
	{0x33B3E04DEC1BC5ADULL, "vxw_bonepitdestroyed_a3_23d_0_0.vlm"},
	{0x33B48A1CB899599EULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_corruptlyrium_e.dds"},
	{0x33B4B95C0A0ECAB4ULL, "599961940.ogg"},
	{0x33B516233E951F7CULL, "dre000_varric6150294.cl"},
	{0x33B6147963464F9CULL, "zz_prmpt_debug6200930.cl"},
	{0x33B6159393B24D1DULL, "vxca_caverns1_mines_gie_0_0.vlm"},
	{0x33B6162F8597A94BULL, "6125799_m.fxe"},
	{0x33B62CC323FC5AC1ULL, "6061772_m.fxe"},
	{0x33B63BDFDDA77603ULL, "6088313_m.fxe"},
	{0x33B65B130CADEBF0ULL, "6188119_m.fxe"},
	{0x33B667081D647A52ULL, "94892611.ogg"},
	{0x33B6CF946E96E08AULL, "vxw_bonepit_1wl_0_0.vlm"},
	{0x33B6E3B5B4D83EF4ULL, "hlac_lt_0_lowtown_a2_-3_2y.dds"},
	{0x33B6E9B40F58C1D3ULL, "cd_qcr220_orsino.bnk"},
	{0x33B6EB98B265DDA8ULL, "vxw_woundedcoast_1l0_0_0.vlm"},
	{0x33B70619E2F0A214ULL, "sound_bodyfall_creat_size.gda"},
	{0x33B72A3BFB026F36ULL, "109723174.ogg"},
	{0x33B7345CB946E7B8ULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.cess.gad"},
	{0x33B742BEA57A8FDFULL, "art\\vfx\\environment\\vfx_dark_fog.cess.gad"},
	{0x33B743AE3595621BULL, "170146855.ogg"},
	{0x33B79AD6938F88DEULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relax.gad"},
	{0x33B7BA6AE1FED215ULL, "6133088_m.fxe"},
	{0x33B7F9C91D04D947ULL, "1045715066.ogg"},
	{0x33B802186B1DDAFBULL, "394870080.ogg"},
	{0x33B8177E183133FCULL, "6074058_m.fxe"},
	{0x33B85E8F1F274B45ULL, "6154722_m.fxe"},
	{0x33B89957A390E874ULL, "6053093_m.fxe"},
	{0x33B8B06039916DAEULL, "lt_0_lowtown_a2_19_0.rml"},
	{0x33B8D0662CB7DC90ULL, "hightown_fem_cityguard_mob6168484.cl"},
	{0x33B92E1FA9E65F96ULL, "113390982.ogg"},
	{0x33B9A811323746F5ULL, "hlac_lt_gallowstemplar_2394_2y.dds"},
	{0x33B9C05240AC8FAFULL, "560010139.ogg"},
	{0x33B9ECB95562FDCBULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.impa.evt"},
	{0x33BA86CAB38FFD60ULL, "vxht_0_hightown_a3_a2k_0_0.vlm"},
	{0x33BA9BD38D4DB601ULL, "vxlt_0_lowtown_a2_n_c_h2c_0_0.vlm"},
	{0x33BAA22DA4D2CE7BULL, "hlac_lt_gallowstemplar_f_2383_8z.dds"},
	{0x33BB734568E2339DULL, "mrl000_marethari6129248.cl"},
	{0x33BB8BA536468529ULL, "ca_caverns1_mines_26038.rml"},
	{0x33BBA89E4B14BEB9ULL, "6070584_f.fxe"},
	{0x33BC25FE4498ED62ULL, "6106199_m.fxe"},
	{0x33BC5E6514987B3CULL, "mer221_sabin6080108.cl"},
	{0x33BDB75619848460ULL, "6024526_f.fxe"},
	{0x33BDC82117EF0053ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\tripped.ani"},
	{0x33BE37BEA5809BCCULL, "art\\vfx\\environment\\vfx_dark_fog.cess.evt"},
	{0x33BE8851D0253329ULL, "dae110ip_at_mrl_home.utp"},
	{0x33BEC6B95651BB64ULL, "6122771_m.fxe"},
	{0x33BEFF14BD5FDDCBULL, "vxca_caverns2_80y_0_0.vlm"},
	{0x33BF221F541D9EF9ULL, "854503682.ogg"},
	{0x33BF292E60A08A8FULL, "64506666.ogg"},
	{0x33BF8D376529E569ULL, "w_bonepitmining_a2_19_0.rml"},
	{0x33BFBDA3866255D6ULL, "vxlt_0_lowtown_n_f24_0_0.vlm"},
	{0x33C071312976B485ULL, "323607923.ogg"},
	{0x33C0D70537B5A248ULL, "vxca_cave_14p_0_0.vlm"},
	{0x33C0EFAD0E96D2E6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robb_0d.dds"},
	{0x33C105B95569468CULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.impa.gad"},
	{0x33C1481B9A897A96ULL, "qaperc.arl"},
	{0x33C1752A6CAEEDBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180.gad"},
	{0x33C223858F97605CULL, "gen_im_trs_drk_one_mac_r2.uti"},
	{0x33C224458BEAF38AULL, "vxw_woundedcoast_pg_0_0.vlm"},
	{0x33C318770A28DD0DULL, "lt_0_lowtown_probe_f65.dds"},
	{0x33C3198F3CF5965AULL, "hlac_lt_hangedman_-1_7z.dds"},
	{0x33C373CBFAB38E23ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.cess.ani"},
	{0x33C390DD6C3E59FCULL, "vxlt_0_lowtown_a2_h1z_0_0.vlm"},
	{0x33C3CDCEA326D99CULL, "vxw_bonepit_hx_0_0.vlm"},
	{0x33C3EE04579CC08EULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredlong.mmh"},
	{0x33C3F41D22B6C0DFULL, "mcr310_argument6074332.cl"},
	{0x33C3FA04579CD525ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredlong.mao"},
	{0x33C40804579CED7CULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredlong.msh"},
	{0x33C4AF29A1AB94E8ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb106a_0d.dds"},
	{0x33C503AF89887068ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_3x.dds"},
	{0x33C5FF534DD22E41ULL, "865540148.ogg"},
	{0x33C627FE9B60ED6BULL, "6102455_m.fxe"},
	{0x33C64324F933DCC7ULL, "92399627.ogg"},
	{0x33C6473CDBE130AFULL, "crt_spblm_fl.mao"},
	{0x33C66853650A2119ULL, "hlac_ht_bartrand_n_3370_2y.dds"},
	{0x33C6D60BE3D3D570ULL, "6054878_m.fxe"},
	{0x33C704B9556DB08AULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.impa.ani"},
	{0x33C8576DCDD2B551ULL, "74178262.ogg"},
	{0x33C86A2A6CB4F9ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180.evt"},
	{0x33C8AD2F8625B520ULL, "979425390.ogg"},
	{0x33C8CCA314375556ULL, "108946608.ogg"},
	{0x33C8CD8D251E20EAULL, "230860119.ogg"},
	{0x33C924460227E7C9ULL, "hlac_lt_0_lowtown_65535_5y.dds"},
	{0x33C9813765322549ULL, "w_bonepitmining_a2_19_0.mao"},
	{0x33C98D37653239E2ULL, "w_bonepitmining_a2_19_0.mmh"},
	{0x33C98F3765323C98ULL, "w_bonepitmining_a2_19_0.msh"},
	{0x33C996A0BFC145B7ULL, "153679677.ogg"},
	{0x33CA0B44BCBB2843ULL, "6108008_m.fxe"},
	{0x33CAC02117F97C8AULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\tripped.evt"},
	{0x33CB1FBEA58AFD25ULL, "art\\vfx\\environment\\vfx_dark_fog.cess.ani"},
	{0x33CB432505383A9AULL, "hlac_lt_0_lowtown_n_-65536_7y.dds"},
	{0x33CB6DD693F94542ULL, "vxw_bonepitmining_a2_2a7_0_0.vlm"},
	{0x33CC0DD15A967520ULL, "6151290_m.fxe"},
	{0x33CC8B575F9CB4EBULL, "character_character_p233.psh.fxo.dx11.assembly"},
	{0x33CC9D4DF4A52F91ULL, "669419279.ogg"},
	{0x33CD812D77CA7AF2ULL, "ser221_gascard6149267.cl"},
	{0x33CE2F8078586E56ULL, "vxlt_0_lowtown_n_f55_0_0.vlm"},
	{0x33CE6D41F7EDBD1AULL, "vxw_blightlands_c0_0_0.vlm"},
	{0x33CE6ED38E20A7B1ULL, "hlac_ht_0_hightown_a2_-65535_1z.dds"},
	{0x33CE8D21EF39DA35ULL, "plt_cod_cha_isabela3.plo"},
	{0x33CEA1AFEED9DC87ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.gad"},
	{0x33CEA4410F1E42CCULL, "w_blightlands_186_0.msh"},
	{0x33CEA8A97059413AULL, "dae000_hate_enemies6167619.cl"},
	{0x33CEAA410F1E4CBEULL, "w_blightlands_186_0.mmh"},
	{0x33CEB6410F1E6155ULL, "w_blightlands_186_0.mao"},
	{0x33CEC50D8B9D70A0ULL, "6164997_m.fxe"},
	{0x33CEDE8BAF97B48CULL, "dae000_arishok6068581.cl"},
	{0x33CEE711792B5353ULL, "zz_jp_poses.cut"},
	{0x33CEF0C80E2ECD6DULL, "349596946.ogg"},
	{0x33CF42E3E5B70054ULL, "vxht_bartrand_n_6ra_0_0.vlm"},
	{0x33CF4696E2219CD9ULL, "ht_brothel_n_3364.anb"},
	{0x33D0542811CE2F6CULL, "vxlt_gallowstemplar_n_6qh_0_0.vlm"},
	{0x33D09FB66E11DF47ULL, "vxlt_undercitydungeon_1im_0_1.vlm"},
	{0x33D0ACFD53CA3C03ULL, "gen_im_trs_krk_hvy_hlm_r1.uti"},
	{0x33D0D157F2BD3CE1ULL, "vxlt_undercitydungeon_12g_0_0.vlm"},
	{0x33D13BF5AFC9FF7AULL, "vxca_caverns2_78r_0_1.vlm"},
	{0x33D18BCBFABFF41AULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.cess.evt"},
	{0x33D195A87DB30536ULL, "llac_lt_hangedman_-1_21.dds"},
	{0x33D19F2117FF63FDULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\tripped.gad"},
	{0x33D1F4C471F9DFC4ULL, "gen00fl_aveline6122114.cl"},
	{0x33D21868D91A80D5ULL, "825480554.ogg"},
	{0x33D263F68DF246FFULL, "6099504_m.fxe"},
	{0x33D2A2B984F10B60ULL, "304514184.ogg"},
	{0x33D30D6C183E273CULL, "lgt000ip_grimoire_sack.utp"},
	{0x33D3184977F6E497ULL, "vxht_keep_3y6_0_0.vlm"},
	{0x33D32906A22507CCULL, "art\\characters\\playercharacter\\humanmale\\hm_arm_lgte.mao"},
	{0x33D38C543EF0F081ULL, "t3_hey_te2.tnt"},
	{0x33D3B81F4DD7B59FULL, "vxht_mansion_n_62y_0_0.vlm"},
	{0x33D3F6C5F6372551ULL, "qunari_male_heads.gda"},
	{0x33D49E24C5EB0315ULL, "6139688_m.fxe"},
	{0x33D4D652C975AFF0ULL, "hlac_ht_0_hightown_65533_2y.dds"},
	{0x33D4E9F7AC0B3688ULL, "6126390_f.fxe"},
	{0x33D547E8D5C9EDF0ULL, "6051778_m.fxe"},
	{0x33D5696FEE78CC12ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masb_0d.dds"},
	{0x33D596AFEEDFE894ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.evt"},
	{0x33D5EE8ABA9A14E0ULL, "6080649_f.fxe"},
	{0x33D6522A6CC10E05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_180.ani"},
	{0x33D67ACEE2A04B6DULL, "hlac_w_bonepit_131072_0y.dds"},
	{0x33D6969DD7661DCEULL, "530505330.ogg"},
	{0x33D71B64AFE6377BULL, "vxw_bonepitdestroyed_a3_29t_0_0.vlm"},
	{0x33D782CCC2111D8BULL, "hlac_w_sundermount_-262139_0x.dds"},
	{0x33D789C6A42C7772ULL, "vxlt_darktown_4k1_0_0.vlm"},
	{0x33D7C1A2C1F16162ULL, "mag291_vincento6027351.cl"},
	{0x33D82565EABDD69EULL, "138424149.ogg"},
	{0x33D850E4E70E836AULL, "6128372_f.fxe"},
	{0x33D8670E39A1F654ULL, "gen00fl_merrill6150712.cl"},
	{0x33D88ACBFAC6110DULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.cess.gad"},
	{0x33D8911A33ABF3F4ULL, "character_character_p14.psh.fxo.dx11"},
	{0x33D92A3B8A54A083ULL, "863781408.ogg"},
	{0x33D92AC57C516FF1ULL, "dae010_lowlife_m2.utc"},
	{0x33D93635FA080B25ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_megatiki_r.dds"},
	{0x33D93BF691498686ULL, "lt_gallowstemplar_n_2382.rml"},
	{0x33D959076FAD5E39ULL, "739534622.ogg"},
	{0x33D95F9F0798606CULL, "hlac_ht_keep_n_3740_0z.dds"},
	{0x33D9A431CC3730E9ULL, "vxht_0_hightown_a2_cb9_0_0.vlm"},
	{0x33DA008786766942ULL, "6126975_m.fxe"},
	{0x33DA2C38F80B8C56ULL, "hlac_lt_gallowscourtyard_-131074_0y.dds"},
	{0x33DA48337A60AB13ULL, "qun131_taskmaster6039390.cl"},
	{0x33DA87F0483F47CFULL, "735968809.ogg"},
	{0x33DAA3A257DC078CULL, "vxlt_0_lowtown_n_ghu_0_0.vlm"},
	{0x33DAECF18EFF947DULL, "vxlt_warehouse_b0_0_0.vlm"},
	{0x33DB50C0C6A879B4ULL, "vxht_chantry_27f_0_0.vlm"},
	{0x33DB5D5EB550DD84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_wave.ani"},
	{0x33DB794F304D2DDFULL, "dae000_bartrand6064033.cl"},
	{0x33DB8A71A3DD51E6ULL, "mrl250cr_thief_lt.utc"},
	{0x33DC3DC24321B334ULL, "397715772.ogg"},
	{0x33DCC911C342EBAEULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumb.msh"},
	{0x33DCD711C34303FFULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumb.mao"},
	{0x33DCDF1EE9E3D59AULL, "vxw_blightlands_d1_0_0.vlm"},
	{0x33DCE311C343179CULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumb.mmh"},
	{0x33DCE490EC55CF90ULL, "6064675_m.fxe"},
	{0x33DCF75D05544787ULL, "404326280.ogg"},
	{0x33DD2219A156E130ULL, "6214928_m.fxe"},
	{0x33DD2A3CFA3A3BC1ULL, "42567522.ogg"},
	{0x33DD9CCAAE48D894ULL, "633886745.ogg"},
	{0x33DDB2303427D0D8ULL, "vxca_caverns2_6h9_0_1.vlm"},
	{0x33DDF7B9B6B73D2DULL, "6069578_m.fxe"},
	{0x33DE803AF864AB0DULL, "vxht_chantry_30c_0_0.vlm"},
	{0x33DED0B3B0828CFEULL, "cd_seb321_rebel_mage.wbd"},
	{0x33DEFB9BF54ECF0EULL, "vxca_caverns1_e51_0_0.vlm"},
	{0x33DEFD670CBDB6D2ULL, "6082592_m.fxe"},
	{0x33DF11936007F7C7ULL, "vxlt_undercitydungeon_1ln_0_1.vlm"},
	{0x33DF3C059C8F620DULL, "vxlt_0_lowtown_n_eo2_0_1.vlm"},
	{0x33DF56A039450BA6ULL, "vxht_mansion_n_673_0_0.vlm"},
	{0x33DFE4E7C2998EB1ULL, "vxlt_0_lowtown_a2_n_cfv_0_0.vlm"},
	{0x33E00875458D2E34ULL, "580198595.ogg"},
	{0x33E0427C026A7313ULL, "6137608_m.fxe"},
	{0x33E0783E567EE248ULL, "hlac_ht_chantry_n_2534_7y.dds"},
	{0x33E0BBF1D6A955EDULL, "llac_ht_store_187_4.dds"},
	{0x33E0C56D062CBC7AULL, "572320199.ogg"},
	{0x33E135F1C0745904ULL, "vxw_bonepit_2a7_0_0.vlm"},
	{0x33E1B5E55C2EC2A5ULL, "402948343.ogg"},
	{0x33E25A5EB556F7E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_wave.gad"},
	{0x33E27C3493C71B13ULL, "795104492.ogg"},
	{0x33E2C77E6003FBC4ULL, "6171810_m.fxe"},
	{0x33E3101B753E058EULL, "6126636_m.fxe"},
	{0x33E3475F5C76767DULL, "llac_w_blightlands_393214_0.dds"},
	{0x33E34EEA8076D17FULL, "vxlt_warehouse_n_qz_0_0.vlm"},
	{0x33E384436A80B196ULL, "vxw_encounter3_9j_0_0.vlm"},
	{0x33E38EAFEEEC18CDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.ani"},
	{0x33E4004BB24142F0ULL, "6141128_f.fxe"},
	{0x33E40CCD6BCE72E4ULL, "hlac_w_encounter2_camp_-65535_0y.dds"},
	{0x33E5C08BA5693939ULL, "567204425.ogg"},
	{0x33E638FFB9140DA2ULL, "vxlt_hangedman_1ms_0_0.vlm"},
	{0x33E6727AEBC9EE04ULL, "follower_barks6173403.cl"},
	{0x33E679C09A2DBFEBULL, "6128683_f.fxe"},
	{0x33E691D95599EDF6ULL, "dae300_cullen6135884.cl"},
	{0x33E6C7F569E5A591ULL, "vxdr_0_deeproads_bot_0_0.vlm"},
	{0x33E6C9FC9852F02EULL, "108005484.ogg"},
	{0x33E6CDD6CF012849ULL, "zz_and_debug6134108.cl"},
	{0x33E70F8E747D1293ULL, "vxdr_0_deeproads_eqk_0_0.vlm"},
	{0x33E8192DFE47FFD4ULL, "plt_lgt220pt_shp_mag_letter.nss"},
	{0x33E8555EB55B5AC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_wave.evt"},
	{0x33E8711A4EB0EE01ULL, "follower_banter6151227.cl"},
	{0x33E8CBEAA6080148ULL, "vxlt_undercitydungeon_15c_0_0.vlm"},
	{0x33E91B34EF7BD1E5ULL, "wos_small.mao"},
	{0x33EA081B2527A3A8ULL, "hlac_dr_0_deeproads_5842_4y.dds"},
	{0x33EAB83852C15FCAULL, "376527136.ogg"},
	{0x33EABC8712F2B409ULL, "dae000_bodhan6146854.cl"},
	{0x33EADDAFA867B6F0ULL, "126441987.ogg"},
	{0x33EAFD2D66FF9317ULL, "nat380_temmerin6067343.cl"},
	{0x33EB333F61D637E7ULL, "vxlt_blackemporium_vv_0_0.vlm"},
	{0x33EB33DED59DF041ULL, "6127614_m.fxe"},
	{0x33EB50FBDBD9EE1AULL, "vxw_blightlands_i2_0_0.vlm"},
	{0x33EC9209610475E6ULL, "clf260_harley6089565.cl"},
	{0x33ECABB3C7E40690ULL, "619705.ogg"},
	{0x33ECDCDD5891962FULL, "gen00fl_varric6074976.cl"},
	{0x33ED3D92BCD32F78ULL, "164947737.ogg"},
	{0x33ED4D856B8EF9FDULL, "182547798.ogg"},
	{0x33ED725B801CD387ULL, "qun120_arishok6049627.cl"},
	{0x33EDB511D6A96DE1ULL, "vxlt_undercitydungeon_18e_0_0.vlm"},
	{0x33EDCAA908183959ULL, "rog_followthrough.dds"},
	{0x33EE0C0D41A36EB9ULL, "mcr310_sibling6095297.cl"},
	{0x33EE3B943AF75D5AULL, "hlac_ht_mansion2_818_51z.dds"},
	{0x33EE46783514260DULL, "6176021_m.fxe"},
	{0x33EE484D727C9E9FULL, "231438972.ogg"},
	{0x33EF55968256DE76ULL, "307123374.ogg"},
	{0x33EF7545F59564E3ULL, "784618344.ogg"},
	{0x33EFF2EAC4169FCFULL, "lt_0_lowtown_docks_a3_n_239_0.mao"},
	{0x33EFFEEAC416B46CULL, "lt_0_lowtown_docks_a3_n_239_0.mmh"},
	{0x33F004EAC416BE5EULL, "lt_0_lowtown_docks_a3_n_239_0.msh"},
	{0x33F0506EF934C834ULL, "vxdr_0_deeproads_f62_0_0.vlm"},
	{0x33F0905839EE2B1CULL, "114236074.ogg"},
	{0x33F09708366F428AULL, "vxw_bonepit_1kh_0_0.vlm"},
	{0x33F0BDC6FA17FB78ULL, "store_lgt_epic1_2.utm"},
	{0x33F0E0B1BCC48AECULL, "qun221_viscount6084692.cl"},
	{0x33F0EAC808D28E36ULL, "blackemp_module.dds"},
	{0x33F1C0C7726CE9FFULL, "vxlt_warehouse_n_te_0_0.vlm"},
	{0x33F1F6205C76CA16ULL, "vxw_encounter3_2k_0_0.vlm"},
	{0x33F20AF64260A0FEULL, "13804615.ogg"},
	{0x33F246B5D4940B79ULL, "llac_w_encounter2_393215_0.dds"},
	{0x33F261D34CBBD36BULL, "hlac_w_sundermount_-2_1x.dds"},
	{0x33F268A13D34BE11ULL, "fex_bd_tower.mmh"},
	{0x33F278999B8A75E5ULL, "dae000_pc_home_ambients6208353.cl"},
	{0x33F282A13D34EACFULL, "fex_bd_tower.msh"},
	{0x33F2D81C0B5C43B7ULL, "vxlt_gallowscourtyard_n_4ll_0_0.vlm"},
	{0x33F3234E37CBF24AULL, "qcr000_arishok6097552.cl"},
	{0x33F36CE3A971F30EULL, "zz_t_trailer_varric.cut"},
	{0x33F38503D7D3A527ULL, "195928184.ogg"},
	{0x33F3D34AC9622214ULL, "zz_fen_debug6134330.cl"},
	{0x33F4A0D0F76D36B5ULL, "equipmentlayout.gda"},
	{0x33F4B8F2B999DE7AULL, "hlac_ht_mansion2b_818_31z.dds"},
	{0x33F52A83AEAEDDC1ULL, "556241963.ogg"},
	{0x33F5849A18F9E58AULL, "261104282.ogg"},
	{0x33F634D35F87F596ULL, "360919419.ogg"},
	{0x33F70D177510F6D8ULL, "disarm_pressed.cur"},
	{0x33F73F3E8B6D7103ULL, "hlac_lt_undercitydungeon_-1_40z.dds"},
	{0x33F7AE25D05B92EFULL, "vxca_caverns1_mines_gkd_0_1.vlm"},
	{0x33F7AFF79CDAE64AULL, "6043230_m.fxe"},
	{0x33F7BE8A5C3165A4ULL, "vxdr_primevalentrance_3b9_0_0.vlm"},
	{0x33F7CDF92695F82AULL, "lt_0_lowtown_docks_a3_20_0_ll.mmh"},
	{0x33F8BBCBBD491E87ULL, "hlac_ht_chantry_n_2534_35y.dds"},
	{0x33F8D56767D066BCULL, "llac_w_encounter3_-327684_0.dds"},
	{0x33F8F6BDAEB81381ULL, "hlac_lt_ruinedhouse_n_-1_4y.dds"},
	{0x33F9152979C122D6ULL, "739249053.ogg"},
	{0x33F92DB6C574F507ULL, "hlac_w_woundedcoast_131078_0x.dds"},
	{0x33F93A867C595991ULL, "6147029_m.fxe"},
	{0x33F953E2CFD366BCULL, "ser221_gamlen6135269.cl"},
	{0x33F9A2E1E9294FB9ULL, "6163374_f.fxe"},
	{0x33F9C2D8CDD0069AULL, "vxw_blightlands_r3_0_0.vlm"},
	{0x33F9F05DB4EC9F14ULL, "763540508.ogg"},
	{0x33F9FF6111945F08ULL, "219089241.ogg"},
	{0x33FA4754E13E4C84ULL, "dae000_bartrand6027554.cl"},
	{0x33FA50431ED3F2E1ULL, "dae000cr_grand_cleric6147659.cl"},
	{0x33FB0087B2B5014AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_bnt_dblw.gad"},
	{0x33FB0BCC164251BFULL, "vxlt_gallowstemplar_n_6qg_0_0.vlm"},
	{0x33FB230CEE036350ULL, "vxlt_0_lowtown_a2_n_e10_0_0.vlm"},
	{0x33FB447D7AFCB2C6ULL, "zz_approval_debug6003821.cl"},
	{0x33FBCF3428D38077ULL, "hlac_w_woundedcoast_0_0x.dds"},
	{0x33FBFA94253AF2A5ULL, "art\\levels\\races\\proxy\\props\\_textures\\fires_d.dds"},
	{0x33FC91369A44EA6CULL, "one200dg_solivitus6048206.cl"},
	{0x33FCA9D727653067ULL, "165740259.ogg"},
	{0x33FCC2CC0D70722DULL, "6094611_m.fxe"},
	{0x33FCFBDDC8A4D231ULL, "vxlt_0_lowtown_docks_a3_czn_0_1.vlm"},
	{0x33FD0C2DA236AB71ULL, "hlac_ht_brothel_n_3364_0x.dds"},
	{0x33FD83A144A473E7ULL, "gen00fl_isabela6096770.cl"},
	{0x33FD8B1C80F946E8ULL, "6118342_f.fxe"},
	{0x33FD929D8DAD0332ULL, "llac_ca_caverns1_mines_24423_5.dds"},
	{0x33FDA6D7727730F8ULL, "vxht_mansion2_258_0_0.vlm"},
	{0x33FDACA4E0FFDB8DULL, "hlac_lt_undercitydungeon_-1_28z.dds"},
	{0x33FDC985C56BA0A3ULL, "vxlt_gallowscourtyard_7ek_0_0.vlm"},
	{0x33FDD255C8AD9BF4ULL, "442796387.ogg"},
	{0x33FDF9155CB3BA7DULL, "hlac_lt_darktown_-1_10x.dds"},
	{0x33FDFB8496EE3DBBULL, "dae300cr_cullen.utc"},
	{0x33FE07B40F94F731ULL, "cd_qcr220_orsino.wbd"},
	{0x33FEC8C35BF8C013ULL, "hlac_w_sundermount_196604_0z.dds"},
	{0x33FECEB4618C2472ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_enco_0n.dds"},
	{0x33FF78C4EF96B2E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_exit.gad"},
	{0x33FF989B9AB3E5BBULL, "6049215_m.fxe"},
	{0x33FF9ECCB31FD6B3ULL, "284308651.ogg"},
	{0x33FFD007077D13BDULL, "404477854.ogg"},
	{0x34000BE6FBBC29CAULL, "gen_im_arm_hlm_med_champion.uti"},
	{0x340027B11A0624FEULL, "vxw_encounter1_eh_0_0.vlm"},
	{0x34002D36C17D9E13ULL, "llac_w_encounter2_3_0.dds"},
	{0x34005132DE60A541ULL, "hlac_lt_gallowsprison_n_-1_9y.dds"},
	{0x3400B3D6F8E5B262ULL, "326023820.ogg"},
	{0x3400C667DA01DA47ULL, "qun221_petrice6083703.cl"},
	{0x34011335E91F2AE8ULL, "399082418.ogg"},
	{0x34011C178366BA02ULL, "llac_lt_0_lowtown_docks_n_-196611_2.dds"},
	{0x3401E387B2BAEEE1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_bnt_dblw.evt"},
	{0x3402530467D79C97ULL, "vxlt_undercityserialkiller_xq_0_0.vlm"},
	{0x3402B715E2A04A21ULL, "6106866_m.fxe"},
	{0x3402BC0640AAA168ULL, "art\\characters\\playercharacter\\humanmale\\arm_masf.mmh"},
	{0x3402C1F7E5DE41C3ULL, "dae000ip_crowd_lm5_talk11.utp"},
	{0x3402C80640AAB50BULL, "art\\characters\\playercharacter\\humanmale\\arm_masf.mao"},
	{0x3402DA0640AAD392ULL, "art\\characters\\playercharacter\\humanmale\\arm_masf.msh"},
	{0x3402FC9D4C11F5EBULL, "gen00fl_dog6117478.cl"},
	{0x34037157AE63C156ULL, "vxlt_warehouse_n_10o_0_0.vlm"},
	{0x34037D9A5261A070ULL, "71469968.ogg"},
	{0x3403F34858694393ULL, "vxdr_0_deeproads_egm_0_0.vlm"},
	{0x34042D11DD8665FCULL, "zz_dae_debug6049943.cl"},
	{0x34043C2AAFFE1A3FULL, "mag121im_idunna_clue01.uti"},
	{0x3404AC9D15810A3FULL, "vxht_0_hightown_a3_bu2_0_0.vlm"},
	{0x3404D6DF46C016D6ULL, "art\\characters\\weapons\\staff\\w_stf_008a.msh"},
	{0x3404E0DF46C027D4ULL, "art\\characters\\weapons\\staff\\w_stf_008a.mmh"},
	{0x3404E4DF46C02EE7ULL, "art\\characters\\weapons\\staff\\w_stf_008a.mao"},
	{0x340514DAF9897FC1ULL, "hlac_lt_backalley_g_65536_4x.dds"},
	{0x3405675D22A0F8D1ULL, "hlac_ht_keep_n_3735_1x.dds"},
	{0x34059BC4EF9B59D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_exit.evt"},
	{0x3405A39305CE4443ULL, "vxht_0_hightown_a2_c67_0_0.vlm"},
	{0x34072E2FCDC71213ULL, "llac_lt_0_lowtown_docks_n_65533_0.dds"},
	{0x3407397F6D1BE503ULL, "m_dae211ar_tavern_nt.wbd"},
	{0x3407A3B8ABA73A38ULL, "3842492.ogg"},
	{0x3407F0F6F5E4B8D2ULL, "970756206.ogg"},
	{0x340836DCF42D1861ULL, "148508713.ogg"},
	{0x3408486D05AF751EULL, "383996789.ogg"},
	{0x340873A9B323AC03ULL, "hlac_w_blightlands_262143_0z.dds"},
	{0x3408D1EE1687EBCAULL, "varric_old.cut"},
	{0x3408D6729DE1D34BULL, "6128914_m.fxe"},
	{0x3408F72219D73B0EULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrggl.mmh"},
	{0x3409032219D74FA5ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrggl.mao"},
	{0x3409112219D767FCULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrggl.msh"},
	{0x34094796E25316D9ULL, "ht_brothel_n_3364.rml"},
	{0x340965E377679074ULL, "1029715562.ogg"},
	{0x340993A64BCFDB7BULL, "vxca_caverns2prc_4x5_0_1.vlm"},
	{0x3409B860A7E300A2ULL, "vxw_encounter2_camp_9i_0_0.vlm"},
	{0x3409C92E2E60B3F4ULL, "vxht_0_hightown_a3_7j1_0_0.vlm"},
	{0x3409E8A9C362E1ECULL, "6111603_m.fxe"},
	{0x340A0F077E997495ULL, "6047413_m.fxe"},
	{0x340A2A03FD8994A7ULL, "120031916.ogg"},
	{0x340ABE89625386BEULL, "714159487.ogg"},
	{0x340B90D58202A826ULL, "gen00fl_aveline6120789.cl"},
	{0x340B9207F7164386ULL, "vxw_encounter1_bi_0_0.vlm"},
	{0x340CB11E8383C60BULL, "em_chandan.mor"},
	{0x340DBBA7F6B000E2ULL, "hlac_lt_undercitydungeon_-1_18z.dds"},
	{0x340DFEC0852AEA80ULL, "vxht_bartrand_n_4kx_0_0.vlm"},
	{0x340E2F7829524F32ULL, "lgt331ip_codex_note.utp"},
	{0x340E95B2035894B5ULL, "art\\vfx\\models\\vfx_rockarmor_ls.mmh"},
	{0x340E97B20358982BULL, "art\\vfx\\models\\vfx_rockarmor_ls.msh"},
	{0x340E998E0BFC3D7EULL, "vxw_encounter1_ji_0_0.vlm"},
	{0x340EDBC4BD7EC81DULL, "t1_skn_t01.tnt"},
	{0x340F097DE61C0515ULL, "qcr220_grey_wardens6135960.cl"},
	{0x340F3AE3254F443DULL, "207664589.ogg"},
	{0x340F81410F559E59ULL, "w_blightlands_186_0.phy"},
	{0x340F88FB81C05DB9ULL, "art\\vfx\\materials\\vfx_d_forcevert.mao"},
	{0x340FB4DBD0E04A79ULL, "6018725_m.fxe"},
	{0x340FDB87B2C71E30ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_bnt_dblw.ani"},
	{0x341018E47C32288EULL, "118876269.ogg"},
	{0x3410749EA99BC37BULL, "w_woundedcoast_64_0.msh"},
	{0x34107E759A6D200BULL, "vxht_playershouse_n_1ub_0_0.vlm"},
	{0x3410869EA99BE1C6ULL, "w_woundedcoast_64_0.mao"},
	{0x3410929EA99BF665ULL, "w_woundedcoast_64_0.mmh"},
	{0x3410A150FB2348E4ULL, "mcr310_confused_guard6077059.cl"},
	{0x3411150D4CBE125CULL, "w_bonepitdestroyed_a3_124_0.mao"},
	{0x3411210D4CBE26FFULL, "w_bonepitdestroyed_a3_124_0.mmh"},
	{0x3411270D4CBE30A1ULL, "w_bonepitdestroyed_a3_124_0.msh"},
	{0x3411F717CC17D7BDULL, "mcr370_meredith6083113.cl"},
	{0x34125DF19DC39BCEULL, "821066580.ogg"},
	{0x341332C5058024BDULL, "character_rendertodepth_p1.vsh.fxo.dx11.assembly"},
	{0x34133346A87BCE0FULL, "mcr370_guardsman.bnk"},
	{0x3413A3C4EFA7A4C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_exit.ani"},
	{0x3413AB3674B0CE80ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofextrude.phy"},
	{0x341437A18FC7112DULL, "6167614_m.fxe"},
	{0x341451EA6501FE3CULL, "6118600_m.fxe"},
	{0x341592C4295E24BFULL, "vxlt_0_lowtown_a2_n_hd5_0_0.vlm"},
	{0x34159E0FF7C24CF0ULL, "vxlt_0_lowtown_docks_n_cug_0_0.vlm"},
	{0x3416058C1FF50679ULL, "529178458.ogg"},
	{0x341697FE802F1226ULL, "hlac_lt_backalley_n_0_14x.dds"},
	{0x341727205A6E25ACULL, "vxht_keep_n_644_0_0.vlm"},
	{0x34175546A87FF177ULL, "mcr370_guardsman.cnv"},
	{0x34179C19D639BC9CULL, "266559450.ogg"},
	{0x3417A767083CC816ULL, "661654956.ogg"},
	{0x3417AB7513C9059AULL, "darktown_male_fer_urch_sta.cnv"},
	{0x3417BCD9F7E78AAAULL, "art\\characters\\playercharacter\\humanmale\\glv_wmda.mao"},
	{0x3417C0D9F7E791F9ULL, "art\\characters\\playercharacter\\humanmale\\glv_wmda.mmh"},
	{0x3417CAD9F7E7A2E7ULL, "art\\characters\\playercharacter\\humanmale\\glv_wmda.msh"},
	{0x3417F4DE91B31AA3ULL, "vxlt_0_lowtown_docks_bcp_0_0.vlm"},
	{0x341828FBB89257C0ULL, "6135803_m.fxe"},
	{0x34182A3D99D91922ULL, "vxw_encounter2_camp_4j_0_0.vlm"},
	{0x34183B0B2056CC74ULL, "vxht_0_hightown_a3_7g6_0_0.vlm"},
	{0x34186CC5BE71746FULL, "6084471_f.fxe"},
	{0x34195D171F21B3A6ULL, "mag160_feynriel6016556.cl"},
	{0x34199FD416280584ULL, "hlac_ca_caverns1_mines_24424_17x.dds"},
	{0x341A03E4E90C5C06ULL, "vxw_encounter1_gj_0_0.vlm"},
	{0x341A16E731415D01ULL, "hightown_market_guard_sta6174748.cl"},
	{0x341A6ED664E577E8ULL, "6112142_f.fxe"},
	{0x341A956226EB9BBFULL, "6116032_m.fxe"},
	{0x341AC194C1B6EECBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt_death.ani"},
	{0x341ACD7513CB75E2ULL, "darktown_male_fer_urch_sta.bnk"},
	{0x341B06EC55844DADULL, "88530515.ogg"},
	{0x341BA9B7A86E9D2FULL, "hlac_ca_caverns1_mines_24424_8y.dds"},
	{0x341BB0AF24914FCCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_loop.ani"},
	{0x341BB620291D73BBULL, "468582156.ogg"},
	{0x341BF6908242F4CCULL, "6095240_m.fxe"},
	{0x341C4FAD41F35929ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge04_d.dds"},
	{0x341C787E6667F4B9ULL, "zz_followers_banter6158154.cl"},
	{0x341C9AB041752EACULL, "6208683_f.fxe"},
	{0x341CC4A8A5D9307EULL, "721342661.ogg"},
	{0x341D60BA50630C64ULL, "149336743.ogg"},
	{0x341DEA3B66FEDFF7ULL, "hlac_lt_0_lowtown_docks_a3_n_1_0z.dds"},
	{0x341E2C532B6C2655ULL, "vxlt_0_lowtown_docks_a3_n_d1r_0_0.vlm"},
	{0x341E85919675D7E6ULL, "6098952_f.fxe"},
	{0x341EBBB5E8D0134AULL, "841885534.ogg"},
	{0x341EE3FA96D716A5ULL, "lgt260_eau_tmp_angelique.bnk"},
	{0x341EE50F1BB98D0EULL, "6060538_m.fxe"},
	{0x341EEC168759AD1FULL, "6079268_f.fxe"},
	{0x341EEF86AA5EBA32ULL, "hlac_ca_cavern1_haunted_24425_9z.dds"},
	{0x341EF540733B3C68ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml105a_0d.dds"},
	{0x341FB8A4DE20B9C6ULL, "vxlt_shop_d2_0_0.vlm"},
	{0x341FB8DE2B086181ULL, "vxlt_0_lowtown_a2_n_c_hod_0_0.vlm"},
	{0x341FC5844659BC11ULL, "art\\vfx\\cinematic\\vfx_mth_arv_qunari_c.anb"},
	{0x341FF43B1B3EAB5AULL, "vxht_chantry_28a_0_0.vlm"},
	{0x34202DCFA28EA650ULL, "220196646.ogg"},
	{0x342084136CE0D6D9ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar1_.gad"},
	{0x3420B603202549B3ULL, "6113831_m.fxe"},
	{0x3420F3B6FE369DADULL, "w_sundermount_91_0.mmh"},
	{0x3420F7B6FE36A4FEULL, "w_sundermount_91_0.mao"},
	{0x342105B6FE36BC33ULL, "w_sundermount_91_0.msh"},
	{0x34216371A41694BEULL, "vxlt_0_lowtown_n_eod_0_0.vlm"},
	{0x3421C412A9828A7FULL, "vxlt_darktown_4yn_0_0.vlm"},
	{0x3421C5FA96D919DDULL, "lgt260_eau_tmp_angelique.cnv"},
	{0x3421D91E2F729369ULL, "6132673_f.fxe"},
	{0x3422044EF5FEBCCBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_black_d.dds"},
	{0x342234CED212C35EULL, "260030575.ogg"},
	{0x3422874CE9BA7543ULL, "vxht_0_hightown_a2_c89_0_0.vlm"},
	{0x3422BDAF2497840EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_loop.gad"},
	{0x34236250F6BE42B0ULL, "11822516.ogg"},
	{0x34238016AAF917C5ULL, "mag110cr_sailor_2.utc"},
	{0x3423A8F03010A678ULL, "hlac_lt_0_lowtown_a2_n_65535_0x.dds"},
	{0x34244187077BFEAFULL, "llac_w_sundermount_262143_0.dds"},
	{0x34252E02B690B702ULL, "lowtown_fem_alienage1_sta6162540.cl"},
	{0x342583A3DE81238DULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em01.mmh"},
	{0x342595A3DE814213ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em01.msh"},
	{0x34263D1454A934EDULL, "lgt100cr_gang_3_docks.utc"},
	{0x342666BB83A93323ULL, "vxlt_0_lowtown_docks_bfs_0_0.vlm"},
	{0x34266ADCFDE7B1AAULL, "vxht_mansion_n_4wy_0_0.vlm"},
	{0x3426C964FF4C484CULL, "gen00fl_anders6125668.cl"},
	{0x3426ED6C68AA0AC1ULL, "92255293.ogg"},
	{0x3426FF6E3C7FAD14ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_7z.dds"},
	{0x3427300CBE5DAF41ULL, "350405323.ogg"},
	{0x34278D8777A64D5EULL, "6037090_f.fxe"},
	{0x342795136CE71227ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar1_.ani"},
	{0x34281FED3A40903CULL, "6136925_m.fxe"},
	{0x342845D4D7DE496EULL, "vxlt_0_lowtown_a2_n_c_e29_0_0.vlm"},
	{0x3428A584E5F91EE2ULL, "vxlt_0_lowtown_docks_a3_cye_0_0.vlm"},
	{0x3428A994C1C30342ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt_death.evt"},
	{0x3428C3C5C33CA283ULL, "gen00fl_fenris6126886.cl"},
	{0x342945475F3DB0AFULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v5.gad"},
	{0x34296BBD6E5AD424ULL, "hlac_lt_0_lowtown_docks_-65540_0y.dds"},
	{0x3429968D5323168AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_020a_s.dds"},
	{0x34299B43035E032AULL, "62737731.ogg"},
	{0x342A314F6C7017ADULL, "6083927_m.fxe"},
	{0x342A4F1A59EE0E1FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1ww_basic_1.mao"},
	{0x342A5FE3501637B2ULL, "719717073.ogg"},
	{0x342A7086F9AAE667ULL, "mag210_arianni6024118.cl"},
	{0x342AEE3CD97531CAULL, "vxlt_darktown_426_0_0.vlm"},
	{0x342AF3B6FE3EF226ULL, "w_sundermount_91_0.rml"},
	{0x342B08308C97FB5AULL, "626526010.ogg"},
	{0x342BA566E440515FULL, "88738339.ogg"},
	{0x342BD374A1459532ULL, "mag291_ntbanter6166155.cl"},
	{0x342BF7A4B8CA9AADULL, "6126815_m.fxe"},
	{0x342C2B7F6D3AE9B1ULL, "m_dae211ar_tavern_nt.bnk"},
	{0x342C3A3860905B9FULL, "495545777.ogg"},
	{0x342CE68446647E8AULL, "art\\vfx\\cinematic\\vfx_mth_arv_qunari_c.mmh"},
	{0x342D392B7F10177DULL, "6045972_m.fxe"},
	{0x342D5DD67FAB9C99ULL, "6072758_m.fxe"},
	{0x342E0B22F043C88AULL, "471332730.ogg"},
	{0x342EA894C1C76D55ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt_death.gad"},
	{0x342EBD82611E1DB1ULL, "746210450.ogg"},
	{0x342F2F9D761C7F54ULL, "73042635.ogg"},
	{0x342F6630D8E51AA2ULL, "vxca_caverns1_j6k_0_0.vlm"},
	{0x342F7C6E0710E2FFULL, "262304408.ogg"},
	{0x342F86991DBC07AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_action.ani"},
	{0x342F872F28DEC9F5ULL, "docks_male_human_worker_sta6176076.cl"},
	{0x342FAA4D8E1C6D8CULL, "vxca_caverns2prc_7vr_0_0.vlm"},
	{0x342FFF8F819B14E0ULL, "700337798.ogg"},
	{0x3430281313EE4F57ULL, "vxca_caverns1_mines_hi0_0_0.vlm"},
	{0x34303F79D40AEFE8ULL, "775856358.ogg"},
	{0x34305A475F43F35CULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v5.evt"},
	{0x3430820A3C68F1CBULL, "vxdr_primevaltomb_528_0_0.vlm"},
	{0x3430B48C157E4B45ULL, "6071104_m.fxe"},
	{0x3430E638A536D88AULL, "164692729.ogg"},
	{0x34318778AF5B9485ULL, "hlac_lt_0_lowtown_docks_a3_-65538_5y.dds"},
	{0x343188828F3F06AEULL, "gen_im_trs_cot_junk1_02.uti"},
	{0x3431AF0EE8522193ULL, "mrl000_marethari6129398.cl"},
	{0x3431B67157192D45ULL, "mag200_feynriel6030435.cl"},
	{0x3431DDE3B9649B82ULL, "gen00fl_sebastian6187757.cl"},
	{0x3431ECB6FE45056AULL, "w_sundermount_91_0.phy"},
	{0x34320C927D22BA76ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadb_1.mao"},
	{0x343314EFC7123626ULL, "vxht_mansion_n_3vq_0_0.vlm"},
	{0x343322612099940EULL, "6118955_f.fxe"},
	{0x34333D3791EDFCE4ULL, "zz_dae_debug6029813.cl"},
	{0x34334BF54496F7BEULL, "6141925_m.fxe"},
	{0x3433578A0078EA20ULL, "1009540919.ogg"},
	{0x34337659CA643666ULL, "241332373.ogg"},
	{0x3433CA7981417395ULL, "hlac_lt_0_lowtown_a2_n_c_-1_4y.dds"},
	{0x3433FD29913DB35BULL, "plt_cod_prc_hedge_witch.plo"},
	{0x343433DEF288591DULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_bench.anb"},
	{0x3434A6C0139D2FA3ULL, "vxca_cavern1_haunted_i9a_0_0.vlm"},
	{0x3434ABFF48F81B09ULL, "character_character_p235.psh.fxo.dx11"},
	{0x34357A01F9B076B0ULL, "vxlt_undercityserialkiller_1l5_0_0.vlm"},
	{0x34359C88A2F6B5C2ULL, "6034074_m.fxe"},
	{0x3435CE0ED4B1F5B8ULL, "6120419_f.fxe"},
	{0x34367CA3DE8F8B4AULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em01.phy"},
	{0x3436B976F7439421ULL, "6144976_m.fxe"},
	{0x3436C5ECCF47ACE7ULL, "528036001.ogg"},
	{0x343714BB8CB4F6F1ULL, "vxlt_darktown_6kx_0_0.vlm"},
	{0x34376EC335EE7C3AULL, "6201380_m.fxe"},
	{0x34377CC994D110A3ULL, "6109227_m.fxe"},
	{0x34377D278F77CAA7ULL, "209956440.ogg"},
	{0x343789EFCF27C886ULL, "vxdr_primevaltomb_9cr_0_0.vlm"},
	{0x3437F592D5E746D1ULL, "vxht_bartrand_n_5q2_0_0.vlm"},
	{0x3437FDED080AAFBEULL, "w_blightlands_35_0.rml"},
	{0x343822495F0EE598ULL, "mag160_varian_ilithis6018891.cl"},
	{0x3438529023E4FD5EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnl180runf.evt"},
	{0x3438687E5F253CD6ULL, "one200_gold_followers6117107.cl"},
	{0x34387EA3A09B9F76ULL, "6040161_m.fxe"},
	{0x3438C83674D01C15ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofextrude.msh"},
	{0x3438CA3674D01F8BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofextrude.mmh"},
	{0x3438D63674D03428ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofextrude.mao"},
	{0x3438F5DADF8EB9F1ULL, "6098713_f.fxe"},
	{0x3439B4AE478A15F1ULL, "vxca_cavern1_haunted_giq_0_0.vlm"},
	{0x343A5D5589701B5EULL, "6087717_m.fxe"},
	{0x343A88E03ABDE6E0ULL, "hlac_w_encounter3_-65531_0z.dds"},
	{0x343AA9162F74A3A8ULL, "vxlt_darktown_5lk_0_0.vlm"},
	{0x343AC1FBF219FFCEULL, "matt_rogue.cut"},
	{0x343AD57D2FDFB741ULL, "vxlt_0_lowtown_a3_n_c_e7x_0_0.vlm"},
	{0x343B7EEA0335DCFAULL, "gen00fl_aveline6121997.cl"},
	{0x343BA0930D602DA3ULL, "vxht_keep_n_4no_0_0.vlm"},
	{0x343BA4023513C7A2ULL, "6033124_f.fxe"},
	{0x343C07091067A1CBULL, "vxdr_primevaltomb_6vc_0_0.vlm"},
	{0x343D21A5578A64EAULL, "mag291_torpor6027010.cl"},
	{0x343D3A3BED5EACCCULL, "hlac_lt_gallowstemplar_n_2383_0y.dds"},
	{0x343D7E991DC837FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_action.evt"},
	{0x343DD80DCADB3322ULL, "vxca_caverns1_j3t_0_0.vlm"},
	{0x343E62475F503E75ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v5.ani"},
	{0x343E913A57F9FC5AULL, "gen00fl_aveline6114830.cl"},
	{0x343ED2E762BE907BULL, "mcr371_cut_orsino_defiance.bnk"},
	{0x343FEF638721330DULL, "gen_im_arm_hlm_lgt_act1_03.uti"},
	{0x34405FC599085017ULL, "gen_im_arm_cht_lgt_act1_03.uti"},
	{0x344086FE227923A5ULL, "lowtown_male_alienage1_sta6162711.cl"},
	{0x3440E471B75C43E5ULL, "w_woundedcoast_26_0_lc.mmh"},
	{0x34412D198FDE9043ULL, "l_one101ar_ggt_cavern.wbd"},
	{0x3441421BDAEC680EULL, "vxw_bonepitmining_a2_1v9_0_0.vlm"},
	{0x3441838A996F48C2ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_pota.msh"},
	{0x3441858A996F4C78ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_pota.mmh"},
	{0x3441918A996F601BULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_pota.mao"},
	{0x34421668289B2E01ULL, "6165241_m.fxe"},
	{0x34424AF4271FDD8BULL, "gen00fl_bethany6101016.cl"},
	{0x34425B2A14782F86ULL, "6071262_m.fxe"},
	{0x3442D4E762C27C3BULL, "mcr371_cut_orsino_defiance.cnv"},
	{0x3443EBDEEBA68F30ULL, "vxlt_undercityserialkiller_1a2_0_0.vlm"},
	{0x34440EBE8701F2B4ULL, "vxdr_0_deeproads_cwl_0_0.vlm"},
	{0x344418DF5311324DULL, "vxlt_0_lowtown_docks_n_b04_0_0.vlm"},
	{0x34445F991DCE21E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_action.gad"},
	{0x3444942D06A6C8BDULL, "6019013_f.fxe"},
	{0x34459E002B7811EBULL, "mer221_sabin6048286.cl"},
	{0x3445E9278741D660ULL, "vxht_mansion2_2nq_0_0.vlm"},
	{0x3445EDC5C44B433EULL, "300289426.ogg"},
	{0x3446676FC7DD5F51ULL, "vxht_bartrand_n_5t1_0_0.vlm"},
	{0x344673A811C120A9ULL, "mcr370_cut_meredith_idol6085999.cl"},
	{0x3446740D7A812363ULL, "pro000_colonel_ewald6040540.cl"},
	{0x34469639CC367DFBULL, "24599833.ogg"},
	{0x3446E538C3753421ULL, "hlac_lt_gallowscourtyard_n_65534_0y.dds"},
	{0x34474C83495D2155ULL, "110822164.ogg"},
	{0x344750ADDD1C066AULL, "hlac_ht_0_hightown_a3_-65536_0z.dds"},
	{0x344758DEE6EE6081ULL, "w_sundermount_89_0.anb"},
	{0x344761CA4EFE4D05ULL, "6155862_m.fxe"},
	{0x3447A843F9BD64C6ULL, "qcr000_arishok6060121.cl"},
	{0x3448087C1A4FD6E4ULL, "1596939.ogg"},
	{0x3448279905929079ULL, "43924643.ogg"},
	{0x34484D51BB58ED06ULL, "dae021_sister_samea_sta6142136.cl"},
	{0x344933843C2BB03EULL, "439866301.ogg"},
	{0x344934A1FB134901ULL, "dae021_keep_noblewoman.utc"},
	{0x344A3DC554D1E022ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft.msh"},
	{0x344A3FC554D1E3D8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft.mmh"},
	{0x344A4BC554D1F87BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft.mao"},
	{0x344AB7279A8BBFF8ULL, "6153380_m.fxe"},
	{0x344AC11EA8B9136EULL, "hlac_w_blightlands_131074_1z.dds"},
	{0x344AD694C7F1971DULL, "gen_im_trs_qun_two_2hs_r2.uti"},
	{0x344B1C4A44851DAFULL, "gen00fl_fenris6107182.cl"},
	{0x344B8C8A9977AC03ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_pota.phy"},
	{0x344BC23A3547C763ULL, "228762037.ogg"},
	{0x344BE1FF2336235DULL, "vxlt_undercityencounter_1n6_0_0.vlm"},
	{0x344C75E1E7E8121FULL, "fxe_day_tree_beam.mao"},
	{0x344D46C1E2045B91ULL, "vxlt_blackemporium_zd_0_0.vlm"},
	{0x344D5CAD83FF33C5ULL, "510879425.ogg"},
	{0x344D8898639B961EULL, "hlac_lt_0_lowtown_a2_n_65535_1y.dds"},
	{0x344DB5FECC346AF4ULL, "6108076_m.fxe"},
	{0x344DCD623FB07D92ULL, "vxht_keep_62g_0_0.vlm"},
	{0x344DEC166417D0FBULL, "772965005.ogg"},
	{0x344E26085142879EULL, "lt_0_lowtown_docks_34_0.rml"},
	{0x344E37C20D3B4B64ULL, "lgt210tr_frame_asst.utt"},
	{0x344E772ADD9B28C3ULL, "168628479.ogg"},
	{0x344E80CF21A263ECULL, "678254592.ogg"},
	{0x344EB558895E45ACULL, "lgt150_nat_were_supporter6111016.cl"},
	{0x344EEEFC7F6C7E9FULL, "art\\vfx\\textures\\vfx_ring_falloff.dds"},
	{0x344F34AE8DDABBE3ULL, "vxlt_0_lowtown_a3_n_c_bxp_0_0.vlm"},
	{0x344F59A7331E9D73ULL, "714772731.ogg"},
	{0x344FAE12CCF8515BULL, "lgt220_shp_mag_guard.bnk"},
	{0x344FCF0F1663F513ULL, "6072257_m.fxe"},
	{0x344FDD1FBE054E6CULL, "95137311.ogg"},
	{0x344FEF9BA7FEABEEULL, "313434208.ogg"},
	{0x3450226DF049552AULL, "480534604.ogg"},
	{0x34503EA6C99F28F6ULL, "6089900_m.fxe"},
	{0x34504E28AAD6B127ULL, "vxlt_0_lowtown_a2_n_enw_0_1.vlm"},
	{0x345072E7720C80CFULL, "isa211_isabela6071381.cl"},
	{0x3450AD8D5B37830BULL, "mag360_thrask6044278.cl"},
	{0x3451277B60546305ULL, "lt_0_lowtown_docks_24_0.rml"},
	{0x34514D972E7802E3ULL, "298803281.ogg"},
	{0x34518A79F78960A3ULL, "vxca_cavern1_haunted_i3c_0_0.vlm"},
	{0x3451929EA9D38D6EULL, "w_woundedcoast_64_0.rml"},
	{0x3451C62ACAE61298ULL, "dae000_tomwise6167276.cl"},
	{0x3451DCD889BC11BCULL, "follower_banter6161115.cl"},
	{0x3452033089BA3342ULL, "354570617.ogg"},
	{0x34526C3279A72BAEULL, "250112634.ogg"},
	{0x3452809B78F80B34ULL, "vxdr_0_deeproads_clm_0_0.vlm"},
	{0x345294C10C1B236CULL, "vxca_cavern1_haunted_imf_0_0.vlm"},
	{0x3452C11E7077CF75ULL, "hlac_lt_warehouse_n_1_4z.dds"},
	{0x3452C38533E19E2EULL, "w_woundedcoast_187_0.phy"},
	{0x34530A9392432615ULL, "vxht_chantry_215_0_0.vlm"},
	{0x345322021211470FULL, "6180813_m.fxe"},
	{0x34532F6483374DCBULL, "vxlt_0_lowtown_docks_3m7_0_0.vlm"},
	{0x3453446D05EB01B9ULL, "6109428_f.fxe"},
	{0x3453B012CCFC3D1BULL, "lgt220_shp_mag_guard.cnv"},
	{0x3453DD06E21F7CD9ULL, "plt_qun000pt_dialogs.plo"},
	{0x3453FC184C3E6BCAULL, "hlac_ht_store_187_1x.dds"},
	{0x34545A75F3A68E85ULL, "mer320_hubert6090237.cl"},
	{0x345539DEE6FA69FAULL, "w_sundermount_89_0.mmh"},
	{0x34553DDEE6FA7031ULL, "w_sundermount_89_0.mao"},
	{0x345546C554DBF763ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft.phy"},
	{0x34554BDEE6FA8840ULL, "w_sundermount_89_0.msh"},
	{0x3455F97D62AA625DULL, "vxlt_0_lowtown_docks_a3_n_bqh_0_0.vlm"},
	{0x34561AFB83643035ULL, "dae000_mother6135749.cl"},
	{0x3456747937457B8BULL, "598607468.ogg"},
	{0x34568860D0F8018FULL, "hlac_dr_0_deeproads_5845_6y.dds"},
	{0x34570F18AB68662BULL, "vxdr_dungeon_1oa_0_0.vlm"},
	{0x345710557187B8A1ULL, "vxdr_0_deeproads_f1r_0_0.vlm"},
	{0x3457121ECB2A7965ULL, "6111348_f.fxe"},
	{0x34577F0B7E457503ULL, "vxlt_backalley_g_1ps_0_1.vlm"},
	{0x3457AF9F82FC692AULL, "6047335_m.fxe"},
	{0x3457E672FD113170ULL, "vxlt_0_lowtown_a2_n_gwh_0_0.vlm"},
	{0x3457EA54D17837BDULL, "6143960_m.fxe"},
	{0x3458095C1E8D46E6ULL, "llac_lt_0_lowtown_n_-1_4.dds"},
	{0x3458980EAEB77CA6ULL, "gen_im_wep_mag_sta_act1_01.uti"},
	{0x3458EA6A2E526A29ULL, "vxw_sundermount_310_0_0.vlm"},
	{0x345A77A3E4476E1CULL, "1044525223.ogg"},
	{0x345ACFCB9AF561AEULL, "832063622.ogg"},
	{0x345B2C663EA205FEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vault_d.dds"},
	{0x345B38F7A0E1308EULL, "vxw_bonepitmining_a2_29c_0_0.vlm"},
	{0x345BE2EA16084544ULL, "dre180im_crypt_key.uti"},
	{0x345C331503E6224BULL, "hlac_w_blightlands_262142_0y.dds"},
	{0x345C51073632A962ULL, "202620330.ogg"},
	{0x345C55B5EF53A8A0ULL, "dae000_mother6136938.cl"},
	{0x345C5ABAF399B6C9ULL, "hlac_ht_0_hightown_n_-65537_2x.dds"},
	{0x345D0C741C07D918ULL, "framebuffereffects_ssao_p0.psh.fxo.dx11"},
	{0x345D2DCB95E7DCEFULL, "vxht_0_hightown_a3_b0m_0_0.vlm"},
	{0x345D5019612C7647ULL, "mag100_karras6034717.cl"},
	{0x345D5185484F5634ULL, "vxht_chantry_2ho_0_0.vlm"},
	{0x345D882A244746FAULL, "art\\characters\\creatures\\dog\\textures\\c_doga_0s.dds"},
	{0x345D9B55927CE8D9ULL, "6123356_gascard_dies.cut"},
	{0x345E6B64038E2D96ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.cess.ani"},
	{0x345EB3EC4105E0B8ULL, "genip_node_j1.utp"},
	{0x345F6A9A795C7A19ULL, "6043736_m.fxe"},
	{0x345FB0A461C61165ULL, "6112921_f.fxe"},
	{0x34605615BE6DEACAULL, "6048107_m.fxe"},
	{0x3460B029BF66DF0AULL, "hlac_lt_0_lowtown_a2_n_0_1y.dds"},
	{0x3460CFF4D0AFEE93ULL, "211712531.ogg"},
	{0x3460F3D60B82C175ULL, "784625993.ogg"},
	{0x3461069DFE113BECULL, "vxca_cavern1_haunted_ijg_0_0.vlm"},
	{0x34612425F842320DULL, "196628603.ogg"},
	{0x346302E2656B4B6CULL, "dae100_viscount6119017.cl"},
	{0x346314A647B8B828ULL, "art\\characters\\playercharacter\\humanmale\\glv_nuda.msh"},
	{0x346326A647B8D6B9ULL, "art\\characters\\playercharacter\\humanmale\\glv_nuda.mao"},
	{0x346332A647B8EB52ULL, "art\\characters\\playercharacter\\humanmale\\glv_nuda.mmh"},
	{0x3463374E469D8C9FULL, "6128118_m.fxe"},
	{0x34637CBAD32E36A5ULL, "615813283.ogg"},
	{0x3463C667CB89C0EFULL, "mag100_karras.bnk"},
	{0x3463F1ED082FD636ULL, "w_blightlands_35_0.mao"},
	{0x3463FDED082FEAD5ULL, "w_blightlands_35_0.mmh"},
	{0x3463FFED082FEE4BULL, "w_blightlands_35_0.msh"},
	{0x346414696D8B19D3ULL, "327053654.ogg"},
	{0x34646B5A54A07ADDULL, "vxlt_0_lowtown_docks_a3_n_btg_0_0.vlm"},
	{0x34648E8B397766EDULL, "661654766.ogg"},
	{0x34654B578A81C340ULL, "6042148_f.fxe"},
	{0x34657F33BC85A079ULL, "296656414.ogg"},
	{0x34658232637DD121ULL, "vxdr_0_deeproads_f4q_0_0.vlm"},
	{0x3465DBDD1F50D8CDULL, "vxca_caverns2_95e_0_0.vlm"},
	{0x3465FD2E649F24A9ULL, "6027777_f.fxe"},
	{0x34663AD7AA4E3374ULL, "126984248.ogg"},
	{0x34663AE93F99BC1DULL, "381481967.ogg"},
	{0x3466436FB7D01EA9ULL, "follower_banter6167873.cl"},
	{0x34665819F4E0C62CULL, "143034733.ogg"},
	{0x34668A0B0C0D601AULL, "test.mao"},
	{0x3466E48071CA92F5ULL, "lgt210_mag_escp_rescued6156650.cl"},
	{0x3466E867CB8C3057ULL, "mag100_karras.cnv"},
	{0x34671F198FFF47F1ULL, "l_one101ar_ggt_cavern.bnk"},
	{0x346733107B371F24ULL, "421927906.ogg"},
	{0x34674AA6D6C6D5A4ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopening_rog.ani"},
	{0x34674E6F494D390DULL, "follower_banter6156248.cl"},
	{0x34686E2672FC8417ULL, "216349820.ogg"},
	{0x3468B82E079E4BDDULL, "192189939.ogg"},
	{0x3468BE88CF37895CULL, "hm_thug03.mor"},
	{0x3469CECCC9A0941CULL, "llac_lt_0_lowtown_docks_-262144_0.dds"},
	{0x346A314D0FB76181ULL, "6187916_f.fxe"},
	{0x346A4A58D302D9B7ULL, "598129083.ogg"},
	{0x346A8B888C4A07D1ULL, "follower_banter6143163.cl"},
	{0x346AE3659EAEDC62ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownloop.gad"},
	{0x346AF4CD8FCB4A3BULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgta.msh"},
	{0x346B06CD8FCB6886ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgta.mao"},
	{0x346B12CD8FCB7D25ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgta.mmh"},
	{0x346B857293F2DE14ULL, "781959605.ogg"},
	{0x346BC3623A456EB4ULL, "vxht_chantry_2el_0_0.vlm"},
	{0x346BD95537AD5029ULL, "gen_im_wep_mag_sta_act2_07.uti"},
	{0x346C709887F1E0D0ULL, "vxlt_0_lowtown_docks_a3_n_cxc_0_0.vlm"},
	{0x346C7364039A7887ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.cess.evt"},
	{0x346CA282F5609FD8ULL, "vxw_encounter2_48_0_0.vlm"},
	{0x346D2DC7A7641924ULL, "6206178_m.fxe"},
	{0x346D41D6C0831E81ULL, "vxlt_hangedman_n_21h_0_0.vlm"},
	{0x346D47A6D6CB3C06ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopening_rog.gad"},
	{0x346D7E492BBD6BFFULL, "mag170_thrask6054018.cl"},
	{0x346DDEAB3607D736ULL, "66069078.ogg"},
	{0x346DDEE03E524E7BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6_0.pwk"},
	{0x346DE74ADCE13007ULL, "vxlt_foundry_n_xw_0_0.vlm"},
	{0x346DEDFBB6662FB5ULL, "6164440_m.fxe"},
	{0x346EA9AB831416F9ULL, "ser191_emeric6104713.cl"},
	{0x346EC7F2B064034AULL, "6043007_m.fxe"},
	{0x346ED54874CAF176ULL, "hlac_ca_cavern1_haunted_24424_6x.dds"},
	{0x346EEB7946F3379BULL, "953252753.ogg"},
	{0x346F47F8A48975A0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldbase.phy"},
	{0x346F5859AD392519ULL, "vxht_0_hightown_a3_cf9_0_1.vlm"},
	{0x346F64555CE43C34ULL, "vxdr_0_deeproads_cfs_0_0.vlm"},
	{0x347039FAA9515FF2ULL, "gen00fl_anders6119319.cl"},
	{0x3470D040E6E6F812ULL, "vxlt_gallowstemplar_n_4sj_0_0.vlm"},
	{0x347106659EB38359ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownloop.evt"},
	{0x347151828738BC21ULL, "hlac_ht_mansion2b_818_57y.dds"},
	{0x3471F5C49F25789CULL, "vxlt_hangedman_m5_0_1.vlm"},
	{0x3472C1A4983078CCULL, "gen00fl_aveline6053342.cl"},
	{0x34734674CD0D8BB8ULL, "vxw_bonepitdestroyed_a3_235_0_0.vlm"},
	{0x34734EC20D5A8E11ULL, "lgt210tr_frame_asst.ncs"},
	{0x34735D1C6ACF7111ULL, "32454117.ogg"},
	{0x3473946403A0CFD0ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.cess.gad"},
	{0x3473C6A40C2E92CEULL, "hlac_lt_0_lowtown_docks_n_65535_5x.dds"},
	{0x3473F77A03D16014ULL, "6047247_m.fxe"},
	{0x34742A0D4D12B344ULL, "w_bonepitdestroyed_a3_124_0.phy"},
	{0x3474684BB2C7D7F3ULL, "gen_im_trs_gen_brokenbow.uti"},
	{0x347485A554B61289ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_chantry.phy"},
	{0x3474BDA49CC9D256ULL, "219152518.ogg"},
	{0x3474BF6AEDC8DEB0ULL, "6115114_f.fxe"},
	{0x3474F9CF940C2C1FULL, "6147405_m.fxe"},
	{0x3475734E7A75A5FAULL, "6081399_m.fxe"},
	{0x3475D666C1DF477CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar5_.gad"},
	{0x3475D806E23C5AEEULL, "plt_qun000pt_dialogs.nss"},
	{0x3475FF45398CDC1AULL, "nat320_delilah6065920.cl"},
	{0x3476A4CD4D1985E3ULL, "194715572.ogg"},
	{0x3476AC6C65452323ULL, "vxlt_0_lowtown_a3_n_c_ey0_0_0.vlm"},
	{0x34773B6E1CD03104ULL, "seb221_random_servants.cnv"},
	{0x3477B082BD88002DULL, "504223391.ogg"},
	{0x3477F3DD720E5659ULL, "6118645_m.fxe"},
	{0x3477F6457EEAD1AAULL, "6150361_m.fxe"},
	{0x34781F4A2D17DE11ULL, "522634126.ogg"},
	{0x347833295D428EE7ULL, "hlac_lt_0_lowtown_65534_3z.dds"},
	{0x34789C58B7E6A511ULL, "vxlt_blackemporium_uy_0_0.vlm"},
	{0x3478C6F0EF9FB11BULL, "627364591.ogg"},
	{0x3478D86777ACFB65ULL, "6087490_f.fxe"},
	{0x34795A5BB595131DULL, "zz_qun_debug6045384.cl"},
	{0x347A01B200657A61ULL, "6004232_m.fxe"},
	{0x347A3BB4B538C058ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockexxl.phy"},
	{0x347A58DEE719BA05ULL, "w_sundermount_89_0.phy"},
	{0x347A5D6E1CD2A164ULL, "seb221_random_servants.bnk"},
	{0x347AA0A4CF7143EAULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgte_0n.dds"},
	{0x347AB3C15EB473EBULL, "hlac_ca_caverns2prc_5802_10x.dds"},
	{0x347AC1D9F83C10E6ULL, "art\\characters\\playercharacter\\humanmale\\glv_wmda.phy"},
	{0x347ACA3C0E1B64A4ULL, "vis121_saemus6086667.cl"},
	{0x347AF083EF7A03E7ULL, "516324417.ogg"},
	{0x347B026DC1E0BC41ULL, "vxlt_0_lowtown_docks_br5_0_0.vlm"},
	{0x347C3BD808CC802FULL, "isa211_isabela6083199.cl"},
	{0x347C4C1005293F2EULL, "6137502_m.fxe"},
	{0x347C5767B797611AULL, "w_encounter3_32_0.mmh"},
	{0x347C5B67B7976851ULL, "w_encounter3_32_0.mao"},
	{0x347C6967B7978060ULL, "w_encounter3_32_0.msh"},
	{0x347C993F3D402654ULL, "vxht_bartrand_n_66m_0_0.vlm"},
	{0x347CD566C1E5647AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar5_.ani"},
	{0x347D4C65108F45BFULL, "vxca_cavern1_haunted_imi_0_0.vlm"},
	{0x347D6749A7E88C9CULL, "llac_w_sundermount_-327680_0.dds"},
	{0x347DB1508579FA57ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_smelt_d.dds"},
	{0x347DC11CE5406606ULL, "389639226.ogg"},
	{0x347DE8EA17939726ULL, "961493480.ogg"},
	{0x347DEC7DD56A320AULL, "vxw_woundedcoast_1t_0_0.vlm"},
	{0x347DFCB5DF6C5CE7ULL, "vxlt_undercityserialkiller_1t6_0_0.vlm"},
	{0x347E143CF7C10547ULL, "gen_im_trs_gem_onyx_01.uti"},
	{0x347E4DB93C042CB3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_cliffwall_s.dds"},
	{0x347EB91D5B7FCE5CULL, "476484036.ogg"},
	{0x347ED6AF0B60C036ULL, "mer200pt_stolen_goods.ncs"},
	{0x347F0E659EBFCE48ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownloop.ani"},
	{0x347F4BE556F608F4ULL, "6154024_m.fxe"},
	{0x347FA79BF41E0D78ULL, "vxht_mansion2_2f1_0_0.vlm"},
	{0x347FFD918A95B5A0ULL, "suc_wav_cl.mao"},
	{0x34807780B1E42C4EULL, "dae000_arishok6067779.cl"},
	{0x348139DEE71FA411ULL, "w_sundermount_89_0.rml"},
	{0x3481ECDECF35602CULL, "llac_w_encounter2_393214_0.dds"},
	{0x3481F4F2921A9516ULL, "vxlt_0_lowtown_er8_0_0.vlm"},
	{0x34827B10361DB483ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fl.ani"},
	{0x3482895248DFF0EFULL, "hlac_lt_darktown_-1_19y.dds"},
	{0x3482E18C04148A61ULL, "vxlt_0_lowtown_a3_n_c_hoe_0_0.vlm"},
	{0x34833C09D2F37AF0ULL, "vxlt_0_lowtown_a2_n_gfw_0_0.vlm"},
	{0x34834BE77710C701ULL, "ave100pt_start.ncs"},
	{0x34834DF58ADA375BULL, "vxht_0_hightown_ana_0_0.vlm"},
	{0x348370497D02852EULL, "one140cr_charade6128495.cl"},
	{0x34838AA5BAC9CAC8ULL, "lgt110_orz_harrowmont6109010.cl"},
	{0x348415FE0E761333ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga101a_0d.dds"},
	{0x3484267B607FBB2DULL, "lt_0_lowtown_docks_24_0.anb"},
	{0x34842DD3AFEAEC0BULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.dura.evt"},
	{0x348439958E561496ULL, "6126584_m.fxe"},
	{0x3484AAE9FA97DA74ULL, "mag360cr_alain.utc"},
	{0x34852D2594B6E8D5ULL, "6020788_m.fxe"},
	{0x34852EEA4F1F16ECULL, "vxca_caverns2prc_9xr_0_0.vlm"},
	{0x348579AEA01E5E2BULL, "vxdr_primevaltomb_6j6_0_1.vlm"},
	{0x3485970B7DB06E62ULL, "682908175.ogg"},
	{0x3485D86E0AB9E3E3ULL, "sky_dr_primevaltomb.mao"},
	{0x34869A3637C1E51CULL, "vxlt_0_lowtown_gbp_0_0.vlm"},
	{0x3486FB29BF0FA0CFULL, "hlac_lt_0_lowtown_a2_n_c_-65536_8y.dds"},
	{0x34873BB5468DE22CULL, "w_blightlands_168_0.msh"},
	{0x348741B5468DEC1EULL, "w_blightlands_168_0.mmh"},
	{0x34874DB5468E00B5ULL, "w_blightlands_168_0.mao"},
	{0x3487ED8CCB9361B3ULL, "gen00fl_fenris6113921.cl"},
	{0x3487F71EF098CEA5ULL, "vxlt_hangedman_n_26q_0_0.vlm"},
	{0x3488397A53D12D80ULL, "hlac_ca_caverns2prc_5802_60x.dds"},
	{0x34883E80C2D92351ULL, "vxht_keep_4nn_0_0.vlm"},
	{0x348874413BFA9233ULL, "vxlt_gallowstemplar_f_5j5_0_0.vlm"},
	{0x348890C1A09DA5CAULL, "703268709.ogg"},
	{0x3488B81723749701ULL, "vxht_bartrand_n_4fv_0_0.vlm"},
	{0x3488B8EE982F77DEULL, "6099326_f.fxe"},
	{0x3488D7676BAB034CULL, "qcr221_cut_arishoks_address6042350.cl"},
	{0x34890846F8C1CABFULL, "hlac_lt_0_lowtown_n_65535_7x.dds"},
	{0x34898D9724E81BD2ULL, "hlac_lt_darktown_-1_88z.dds"},
	{0x3489B6E21B8C9DFDULL, "gen00fl_varric6076121.cl"},
	{0x3489C89A0E57EFDBULL, "6191271_m.fxe"},
	{0x3489D34519FFC75EULL, "w_bonepit_25_0.rml"},
	{0x348A1025F675437FULL, "hlac_lt_0_lowtown_1_10x.dds"},
	{0x348A2A67656F3398ULL, "vxlt_gallowscourtyard_6x6_0_0.vlm"},
	{0x348A5588034000B5ULL, "601022811.ogg"},
	{0x348AD8EF13CEC330ULL, "1005911557.ogg"},
	{0x348ADC7106C43E1CULL, "vxlt_0_lowtown_a2_n_c_i1b_0_0.vlm"},
	{0x348B0B1C2F363ED4ULL, "vxht_bartrand_n_63r_0_0.vlm"},
	{0x348B1B5FCB8D819FULL, "hightown_fem_elfservant_mob6169059.cl"},
	{0x348B46D3AFF135CCULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.dura.gad"},
	{0x348BBE4202855E3FULL, "vxca_cavern1_haunted_ijh_0_0.vlm"},
	{0x348C07B4EB3C593BULL, "523495391.ogg"},
	{0x348C61CE122D750FULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1a.mao"},
	{0x348C6DCE122D89ACULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1a.mmh"},
	{0x348C73CE122D939EULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1a.msh"},
	{0x348C843103D080B4ULL, "6143110_m.fxe"},
	{0x348C8A367AF631B3ULL, "6165180_m.fxe"},
	{0x348CF8193DF713EEULL, "llac_lt_0_lowtown_a2_n_65537_6.dds"},
	{0x348D6EEF6BFB3B96ULL, "hlac_ht_mansion2_818_58y.dds"},
	{0x348E3B82B9282F90ULL, "vxw_bonepitmining_a2_1wx_0_0.vlm"},
	{0x348E516854EFDB0AULL, "hlac_lt_foundry_n_320_2y.dds"},
	{0x348E52F8DD45A56CULL, "hlac_ca_caverns2prc_5802_53z.dds"},
	{0x348F93103628687AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fl.evt"},
	{0x34906DE31C918311ULL, "mag291ip_barrel_white.utp"},
	{0x34908F7B85A40370ULL, "288355386.ogg"},
	{0x3490F836039448A7ULL, "vxlt_0_lowtown_a3_n_c_cmj_0_0.vlm"},
	{0x349121AEE2AC70DEULL, "vxlt_0_lowtown_e1m_0_0.vlm"},
	{0x34914788559B2198ULL, "668703769.ogg"},
	{0x34916CFE87CC7F80ULL, "vxlt_hangedman_n_2ui_0_0.vlm"},
	{0x34917A749280F89DULL, "zz_personality_debug6004624.cl"},
	{0x349203EA8B5CE3DFULL, "hlac_lt_gallowscourtyard_-131072_1y.dds"},
	{0x349245D3AFF752CAULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.dura.ani"},
	{0x3492C285341780F1ULL, "w_woundedcoast_187_0.mmh"},
	{0x3492CE8534179492ULL, "w_woundedcoast_187_0.mao"},
	{0x3492DC853417ACAFULL, "w_woundedcoast_187_0.msh"},
	{0x3493521E4F168E0AULL, "hlac_w_blightlands_0_10x.dds"},
	{0x34937903CCE02242ULL, "524049107.ogg"},
	{0x34939F0286AD0155ULL, "6027688_m.fxe"},
	{0x349464F8A4A8C335ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldbase.msh"},
	{0x349466F8A4A8C6ABULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldbase.mmh"},
	{0x349472F8A4A8DB48ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldbase.mao"},
	{0x349485C2132C35C0ULL, "vxdr_0_deeproads_cv7_0_0.vlm"},
	{0x3494D8D716AFABE0ULL, "gen00fl_anders6119212.cl"},
	{0x349535901A90A677ULL, "dae000_fem_priest_mob6174195.cl"},
	{0x34963D1EB97F7263ULL, "157183237.ogg"},
	{0x34969210362E856DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_fl.gad"},
	{0x34970798C137F09CULL, "static_static_p23.psh.fxo.dx11"},
	{0x349729F4156AAF81ULL, "vxht_bartrand_n_4ku_0_0.vlm"},
	{0x34984DE01CA23F96ULL, "vxca_cave_1a5_0_0.vlm"},
	{0x34987BB37586B8D5ULL, "vxlt_gallowstemplar_f_431_0_0.vlm"},
	{0x3498DD98A390ED1BULL, "437370704.ogg"},
	{0x34996CAAB8B4140DULL, "gen_im_trs_apo_junk1_01.uti"},
	{0x349995DD46DF2409ULL, "6007718_m.fxe"},
	{0x349A301EF47B76BFULL, "vxca_cavern1_haunted_igk_0_0.vlm"},
	{0x349AD453CC272007ULL, "6041718_m.fxe"},
	{0x349B025632166258ULL, "vxlt_0_lowtown_a3_n_c_byp_0_0.vlm"},
	{0x349B093F164B784AULL, "6076238_m.fxe"},
	{0x349BB6674994BD81ULL, "6068739_m.fxe"},
	{0x349BF23794BFD548ULL, "dae251ip_at_sundermount_a.utp"},
	{0x349BF9DD2146F5AEULL, "vxht_mansion2_29z_0_0.vlm"},
	{0x349C1C8F9984D07FULL, "633419614.ogg"},
	{0x349C22C8DB5AA128ULL, "6039383_m.fxe"},
	{0x349C62704AE2DC8FULL, "434444811.ogg"},
	{0x349CBDAE2B835AFEULL, "854382640.ogg"},
	{0x349D27675CE1B10FULL, "88792296.ogg"},
	{0x349D5CDD4DD734EAULL, "vxw_sundermount_32v_0_0.vlm"},
	{0x349E1ADC26A0BE0CULL, "vxca_cave_20c_0_0.vlm"},
	{0x349E41E141A3B47BULL, "hightown_male_elfservnt_mob6170086.cl"},
	{0x349E92E8332F1CD1ULL, "art\\vfx\\materials\\vfx_fire_sin.mao"},
	{0x349EDC71FDA59861ULL, "follower_banter6173177.cl"},
	{0x349F2B69A2291928ULL, "6141127_m.fxe"},
	{0x349F9815348C17B9ULL, "art\\vfx\\combat\\vfx_sg_hurl_act_c.impa.ani"},
	{0x349FA006ED29B293ULL, "278359902.ogg"},
	{0x34A00453B1FBF57DULL, "la_sh.mao"},
	{0x34A06266D5585DA4ULL, "vxlt_0_lowtown_ewu_0_0.vlm"},
	{0x34A06EA554DB26AEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_chantry.mmh"},
	{0x34A07AA554DB3B45ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_chantry.mao"},
	{0x34A088A554DB529CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_chantry.msh"},
	{0x34A0A695CE429594ULL, "90799459.ogg"},
	{0x34A0D6101E7FE3C6ULL, "349946315.ogg"},
	{0x34A0FF7AE1E25E3DULL, "fen110_fenris6118048.cl"},
	{0x34A17667B7B6B225ULL, "w_encounter3_32_0.phy"},
	{0x34A1EE691FE4ACF7ULL, "follower_banter6141729.cl"},
	{0x34A1F1FDE1F2B9F0ULL, "vxlt_undercityencounter_13e_0_1.vlm"},
	{0x34A202AA64FE58F3ULL, "417950489.ogg"},
	{0x34A204C858A38E04ULL, "hlac_w_woundedcoast_327681_0z.dds"},
	{0x34A210F9360003CEULL, "6034636_m.fxe"},
	{0x34A276B8A1C42DF0ULL, "qun221_petrice6076809.cl"},
	{0x34A27F34BAA5C635ULL, "dre000_vampire6034367.cl"},
	{0x34A2D0AD869AE8D0ULL, "gen00fl_anders6119902.cl"},
	{0x34A2FAFFFD74DDABULL, "hlac_lt_0_lowtown_docks_-131073_2y.dds"},
	{0x34A353565C9F116DULL, "396099139.ogg"},
	{0x34A36D5F538A4725ULL, "vxht_bartrand_n_4ew_0_0.vlm"},
	{0x34A3B1ACFF5B8FF0ULL, "dae200_cullen6140515.cl"},
	{0x34A3F673B78FC7D7ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\2048texel.dds"},
	{0x34A475D19334B999ULL, "6072549_m.fxe"},
	{0x34A5C16E877EA9F2ULL, "528968787.ogg"},
	{0x34A673153491F7FBULL, "art\\vfx\\combat\\vfx_sg_hurl_act_c.impa.gad"},
	{0x34A67B3399838D46ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\prp_barkburn_d.dds"},
	{0x34A69891B844CDCEULL, "dae320ip_at_keep.utp"},
	{0x34A6AF86C7DFA427ULL, "6017476_m.fxe"},
	{0x34A6B0823EFC0F0EULL, "506264078.ogg"},
	{0x34A6ED90677CD155ULL, "vxlt_gallowstemplar_f_480_0_0.vlm"},
	{0x34A6F41DBE5BE60DULL, "llac_lt_0_lowtown_a2_n_c_65534_4.dds"},
	{0x34A74DD7BD606F6BULL, "6051643_m.fxe"},
	{0x34A76E7DD99C00B0ULL, "367221359.ogg"},
	{0x34A7781A1940EA77ULL, "w_blightlands_213_0.mmh"},
	{0x34A77C1A1940F084ULL, "w_blightlands_213_0.mao"},
	{0x34A78A4C66EA77D0ULL, "477996499.ogg"},
	{0x34A78E1A19410FE9ULL, "w_blightlands_213_0.msh"},
	{0x34A7EED613226FD4ULL, "vxht_bartrand_n_69p_0_0.vlm"},
	{0x34A7EF81E6E0036FULL, "zz_mag_debug6051788.cl"},
	{0x34A7F34A52EA95D7ULL, "6082143_f.fxe"},
	{0x34A83093F717E8AFULL, "qun231_varnell6098597.cl"},
	{0x34A85767B7BC9C31ULL, "w_encounter3_32_0.rml"},
	{0x34A866B64E51E036ULL, "vxht_chantry_20u_0_0.vlm"},
	{0x34A8678005500FA3ULL, "vxlt_0_lowtown_docks_b5z_0_0.vlm"},
	{0x34A8ADBF4456FD31ULL, "zz_dae_debug6051282.cl"},
	{0x34A8EDB752B10D3CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightownbooks_d.dds"},
	{0x34A920840D8A4863ULL, "184159155.ogg"},
	{0x34A92C0AB5D12C71ULL, "6173563_m.fxe"},
	{0x34A98E17A2DCDC90ULL, "lt_0_lowtown_docks_47_0.rml"},
	{0x34A9BB32A7EFDCC9ULL, "qcr220cr_qunari_cut_a1.utc"},
	{0x34AA62E53D9DF7C9ULL, "art\\levels\\races\\proxy\\props\\door\\_textures\\door_e.dds"},
	{0x34AB2462DC461453ULL, "zz_vault_debug6203298.cl"},
	{0x34AB722E2E4DEA00ULL, "hlac_w_bonepitdestroyed_a3_65542_0z.dds"},
	{0x34AB79A929EFCD59ULL, "6116752_f.fxe"},
	{0x34AC09855353CC6BULL, "86910423.ogg"},
	{0x34AC1A13489242FBULL, "vxdr_primevaltomb_5bi_0_1.vlm"},
	{0x34AC1D8AB60E2D18ULL, "vxdr_primevaltomb_4xh_0_0.vlm"},
	{0x34AC761A7BA03573ULL, "6020085_m.fxe"},
	{0x34AC9015349694E0ULL, "art\\vfx\\combat\\vfx_sg_hurl_act_c.impa.evt"},
	{0x34AD42CD67133619ULL, "lgt000cr_hybris.utc"},
	{0x34ADC0ED935F70FAULL, "vxw_sundermount_21j_0_0.vlm"},
	{0x34ADC8B68B1ACF47ULL, "hlac_lt_undercitydungeon_-1_28x.dds"},
	{0x34ADF336D90E8BA3ULL, "840660502.ogg"},
	{0x34AE392B2E31E3A3ULL, "hlac_lt_darktown_-1_67x.dds"},
	{0x34AE7EF2FA972B35ULL, "vxdr_primevaltomb_75y_0_0.vlm"},
	{0x34AED443C74E7624ULL, "vxlt_0_lowtown_erz_0_0.vlm"},
	{0x34AF0EAB04F40BF1ULL, "vxdr_primevaltomb_7e4_0_1.vlm"},
	{0x34AF60FD89C50CDDULL, "6208667_m.fxe"},
	{0x34AF882BE9DB6F0EULL, "6032915_m.fxe"},
	{0x34AFD672EC35E51EULL, "297902975.ogg"},
	{0x34B00D4A5F318F99ULL, "317544579.ogg"},
	{0x34B02F6BB1A6379AULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_dirt_sa.dds"},
	{0x34B072AAE6AA4559ULL, "zz_dre_debug6204366.cl"},
	{0x34B0EB610DFE0A5BULL, "6041501_m.fxe"},
	{0x34B0FC1CF3E67779ULL, "mer320_hubert6065840.cl"},
	{0x34B15E4044B7141CULL, "hlac_lt_0_lowtown_a2_n_131073_1x.dds"},
	{0x34B1697BF71866C0ULL, "vxdr_0_deeproads_ch5_0_0.vlm"},
	{0x34B1C431EF66E534ULL, "seb321_leliana6180963.cl"},
	{0x34B1EBCD90073DD2ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgta.phy"},
	{0x34B1F3B0BA966AC4ULL, "ht_0_hightown_a3_27_0_ll.mmh"},
	{0x34B222552F64820AULL, "dae000_arishok6068451.cl"},
	{0x34B2538F848B224EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_both_hands.evt"},
	{0x34B2CA17EAC57C04ULL, "art\\characters\\playercharacter\\humanmale\\arm_masd.msh"},
	{0x34B2DC17EAC59B6DULL, "art\\characters\\playercharacter\\humanmale\\arm_masd.mao"},
	{0x34B2E017EAC5A1A6ULL, "art\\characters\\playercharacter\\humanmale\\arm_masd.mmh"},
	{0x34B31C28A79EF7A3ULL, "vxdr_primevaltomb_8xk_0_0.vlm"},
	{0x34B3236CFDBB10B4ULL, "dae100_viscount6101878.cl"},
	{0x34B3C504FABD3840ULL, "mag210_arianni6024551.cl"},
	{0x34B4A731E333EDC4ULL, "488657622.ogg"},
	{0x34B4ADD76BD3D7DCULL, "hlac_lt_0_lowtown_a2_n_-1_2x.dds"},
	{0x34B50EB984123677ULL, "vxw_encounter2_camp_os_0_1.vlm"},
	{0x34B536EA9CF9AEF2ULL, "382456255.ogg"},
	{0x34B5ADD379EA5530ULL, "dae000ip_crowd_em4_talk1.utp"},
	{0x34B61659485E7E66ULL, "mcr310_argument6182467.cl"},
	{0x34B62F701D65B83CULL, "hlac_ca_cavern1_haunted_24421_1x.dds"},
	{0x34B63529DD371202ULL, "228497433.ogg"},
	{0x34B635754260BF06ULL, "111760701.ogg"},
	{0x34B6458EBB35A6F5ULL, "6174957_m.fxe"},
	{0x34B6B329EBDECFCAULL, "vxlt_warehouse_n_p0_0_0.vlm"},
	{0x34B6C5D75FA3D86FULL, "vxht_0_hightown_a3_a73_0_0.vlm"},
	{0x34B6C7451A26A0D6ULL, "w_bonepit_25_0.mao"},
	{0x34B6D3451A26B575ULL, "w_bonepit_25_0.mmh"},
	{0x34B6D5451A26B8EBULL, "w_bonepit_25_0.msh"},
	{0x34B74D3F74165B28ULL, "6164405_m.fxe"},
	{0x34B8362F355C8874ULL, "follower_banter6172326.cl"},
	{0x34B8730F27C06D01ULL, "mag291_demon_marethari6027975.cl"},
	{0x34B8A6B4FB845873ULL, "gen00fl_fenris6127045.cl"},
	{0x34B8A8450A01A325ULL, "mrl150_hunter6071294.cl"},
	{0x34B8C30A0862723CULL, "6168526_m.fxe"},
	{0x34B8DF4616E3546FULL, "670099653.ogg"},
	{0x34B90146B7A5CF57ULL, "vxht_bartrand_n_5pk_0_0.vlm"},
	{0x34B94A8F849131F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_both_hands.gad"},
	{0x34B9A2540F75D6D2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_exit.gad"},
	{0x34B9B2B30F2CD5E4ULL, "qun100_mage_party_comment6019111.cl"},
	{0x34B9DC55C8A06A95ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge04_sa.dds"},
	{0x34BA522CA413549AULL, "mag291_vincento6027108.cl"},
	{0x34BA5E12FFD743E2ULL, "zz_lgt_1_debug6188683.cl"},
	{0x34BAC94A300ECA28ULL, "mrl000_marethari6129368.cl"},
	{0x34BAD5EE6D393D9FULL, "vxht_mansion_n_6bi_0_0.vlm"},
	{0x34BC434EFCC2C9A1ULL, "6141827_m.fxe"},
	{0x34BCD717EACDE771ULL, "art\\characters\\playercharacter\\humanmale\\arm_masd.phy"},
	{0x34BCDCF5091C4AAAULL, "hlac_ca_cavern1_haunted_24423_10z.dds"},
	{0x34BCF372299C6661ULL, "235946994.ogg"},
	{0x34BD8F1E95C8E4A6ULL, "dr_0_deeproads_20354_ll.mmh"},
	{0x34BDA0C0D29F388CULL, "zz_mcr_debug6088445.cl"},
	{0x34BDE2C0EE4C7F9FULL, "6107243_m.fxe"},
	{0x34BE447B0B6B2618ULL, "338656472.ogg"},
	{0x34BE7F545BE18B78ULL, "vxht_0_hightown_a3_apm_0_0.vlm"},
	{0x34BE84F13B177E7BULL, "vxht_0_hightown_n_a2n_0_0.vlm"},
	{0x34BF31A64807484DULL, "art\\characters\\playercharacter\\humanmale\\glv_nuda.phy"},
	{0x34BF41CAF3441674ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.anb"},
	{0x34BF5B8F8495BA47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_both_hands.ani"},
	{0x34BFA5540F7A4709ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_exit.evt"},
	{0x34BFD873C4BE1D6CULL, "622283777.ogg"},
	{0x34BFF7C0C2801F5DULL, "vxht_bartrand_n_5rc_0_0.vlm"},
	{0x34C043268681757FULL, "art\\characters\\playercharacter\\elffemale\\cth_comf.phy"},
	{0x34C0DA810908FA70ULL, "6041270_m.fxe"},
	{0x34C0DFB5D13F9AF4ULL, "vxw_woundedcoast_1o4_0_1.vlm"},
	{0x34C108E307C89D89ULL, "vxw_sundermount_32h_0_1.vlm"},
	{0x34C120221508E58EULL, "vxw_sundermount_31j_0_1.vlm"},
	{0x34C121D9CEA493DBULL, "llac_lt_0_lowtown_a2_n_65533_3.dds"},
	{0x34C1301FD5C80F9BULL, "785622609.ogg"},
	{0x34C151EFD06F45C0ULL, "vxlt_0_lowtown_a2_n_byg_0_0.vlm"},
	{0x34C1ADFFCC483C84ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoor.mmh"},
	{0x34C1B1FFCC4843D7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoor.mao"},
	{0x34C1C3FFCC486226ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoor.msh"},
	{0x34C242202BA4A6D9ULL, "552493485.ogg"},
	{0x34C29F02E224E32CULL, "plt_cod_mgc_veil.plo"},
	{0x34C2A8EAD870E386ULL, "1058949101.ogg"},
	{0x34C2B8DA8D9453A7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\knife_throw_action.gad"},
	{0x34C3E7318D4ED348ULL, "docks_scene_dockworkers6176213.cl"},
	{0x34C3F8170A519790ULL, "vxht_playershouse_n_1r8_0_0.vlm"},
	{0x34C45359CCD4F6D1ULL, "vxca_caverns1_j29_0_1.vlm"},
	{0x34C473C997EAF1E8ULL, "vxw_sundermount_34c_0_1.vlm"},
	{0x34C4D9938A99DD99ULL, "21558752.ogg"},
	{0x34C4F7110FD79404ULL, "108019036.ogg"},
	{0x34C52506DDD4E84AULL, "vxlt_warehouse_n_w1_0_0.vlm"},
	{0x34C57E3C17E79B75ULL, "6031162_m.fxe"},
	{0x34C5DC0F413079C5ULL, "733194907.ogg"},
	{0x34C6119421F467E9ULL, "6078446_m.fxe"},
	{0x34C657E4F37D8ACAULL, "159910766.ogg"},
	{0x34C672A23EAE0D61ULL, "hlac_lt_0_lowtown_n_-131072_2z.dds"},
	{0x34C69023E10CF1D8ULL, "qun231cr_varnell_minion2.utc"},
	{0x34C6EBD8D51F4127ULL, "6043020_m.fxe"},
	{0x34C7594B48F322EBULL, "llac_ht_mansion2b_818_7.dds"},
	{0x34C75C1BDDE9FAD1ULL, "6079045_m.fxe"},
	{0x34C7ABF5031C6CD8ULL, "gen00fl_isabela6096836.cl"},
	{0x34C7B25EE3411421ULL, "vxdr_0_deeproads_cn2_0_0.vlm"},
	{0x34C7CA75AACF853BULL, "sky_lt_hangedman_n.mao"},
	{0x34C7F7D1BE23ED19ULL, "805802222.ogg"},
	{0x34C818B546C53DB9ULL, "w_blightlands_168_0.phy"},
	{0x34C81F45FBCF8872ULL, "674315129.ogg"},
	{0x34C81FE662F2F2A3ULL, "6179062_m.fxe"},
	{0x34C94AFF5A94CD68ULL, "follower_banter6160620.cl"},
	{0x34C97F23237C6244ULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.mmh"},
	{0x34C9ADDA8D9A5FB4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\knife_throw_action.evt"},
	{0x34C9B4712EF455DEULL, "406556479.ogg"},
	{0x34CA03B9137018C1ULL, "vxht_keep_n_4n5_0_0.vlm"},
	{0x34CA161248C8D776ULL, "6027455_f.fxe"},
	{0x34CAAD28BA477323ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_3.pwk"},
	{0x34CB56097BC868D7ULL, "w_bonepitmining_a2_83_0.mmh"},
	{0x34CB5A097BC86FE4ULL, "w_bonepitmining_a2_83_0.mao"},
	{0x34CB682D4447AC24ULL, "116073200.ogg"},
	{0x34CB6C097BC88E49ULL, "w_bonepitmining_a2_83_0.msh"},
	{0x34CBB7FDAB3AA724ULL, "vxlt_0_lowtown_ehx_0_0.vlm"},
	{0x34CBC8571AE7F0ADULL, "84757994.ogg"},
	{0x34CCCB762EB4005EULL, "hlac_lt_darktown_-1_42x.dds"},
	{0x34CD20CAF3501B37ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.mmh"},
	{0x34CD258C05349EDBULL, "816537652.ogg"},
	{0x34CD7A05F1F25ADFULL, "art\\vfx\\combat\\vfx_r_var_modal_act_c.impa.gad"},
	{0x34CD8D540F865C78ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_exit.ani"},
	{0x34CD983AEACBA684ULL, "hlac_ca_caverns2_5802_73y.dds"},
	{0x34CE699DB47637DDULL, "vxht_bartrand_n_5ub_0_0.vlm"},
	{0x34CE6F0E6A8F011EULL, "one100dg_vanard6047413.cl"},
	{0x34CE8282D4E1A92FULL, "657478752.ogg"},
	{0x34CEAB58848FB0A9ULL, "llac_lt_0_lowtown_docks_a3_65537_0.dds"},
	{0x34CEC07A6601ADF7ULL, "920347479.ogg"},
	{0x34CF2EC5CEAC21A5ULL, "6147170_f.fxe"},
	{0x34CF97113BF5AC1EULL, "950306515.ogg"},
	{0x34CFF8EC20C57129ULL, "qun131_warrens_thug.bnk"},
	{0x34D01E1086B5E250ULL, "867668926.ogg"},
	{0x34D0994934EF0F68ULL, "839386946.ogg"},
	{0x34D14426868FE984ULL, "art\\characters\\playercharacter\\elffemale\\cth_comf.mmh"},
	{0x34D14692B76E4981ULL, "967587199.ogg"},
	{0x34D14826868FF0D7ULL, "art\\characters\\playercharacter\\elffemale\\cth_comf.mao"},
	{0x34D15A2686900F26ULL, "art\\characters\\playercharacter\\elffemale\\cth_comf.msh"},
	{0x34D1664841418B70ULL, "vxlt_0_lowtown_docks_n_c4h_0_0.vlm"},
	{0x34D1859D0BE69CB5ULL, "mrl250cr_corpse_archer.utc"},
	{0x34D283CF4E436022ULL, "t3_hey_gn3.tnt"},
	{0x34D396E3CFCB00CAULL, "vxlt_warehouse_n_z2_0_0.vlm"},
	{0x34D3DAEC20C92769ULL, "qun131_warrens_thug.cnv"},
	{0x34D3F39754D0D5CEULL, "hlac_lt_foundry_n_320_15y.dds"},
	{0x34D41BB7BAC8E8E6ULL, "vxlt_0_lowtown_a2_n_c_gch_0_0.vlm"},
	{0x34D46F05F1F866CCULL, "art\\vfx\\combat\\vfx_r_var_modal_act_c.impa.evt"},
	{0x34D514EB1DBB12A0ULL, "498824302.ogg"},
	{0x34D5E0A88F0710F0ULL, "dae000_arishok6078416.cl"},
	{0x34D5FB00447A1C73ULL, "39914458.ogg"},
	{0x34D628B4B586FEEDULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockexxl.msh"},
	{0x34D62FCB3EABCA2AULL, "1038654513.ogg"},
	{0x34D636B4B5871620ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockexxl.mao"},
	{0x34D63AB4B5871D73ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockexxl.mmh"},
	{0x34D64AA28CF4A485ULL, "dae000_fenarel6133318.cl"},
	{0x34D64DF9D3ECEED2ULL, "415598015.ogg"},
	{0x34D6AC0F9AC81FDCULL, "6217958_m.fxe"},
	{0x34D6B7D547F60F75ULL, "271139758.ogg"},
	{0x34D72593B4B27195ULL, "lgt110_gxa_coa_spy6164335.cl"},
	{0x34D75909788EFF02ULL, "6104881_m.fxe"},
	{0x34D76DE2A8E5449FULL, "6177671_m.fxe"},
	{0x34D791F5064EA38BULL, "6037315_m.fxe"},
	{0x34D7A5DA8DA68FEDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\knife_throw_action.ani"},
	{0x34D7E33EA0D56CBEULL, "6077134_m.fxe"},
	{0x34D8AC4B3688D015ULL, "vxdr_0_deeproads_cg9_0_0.vlm"},
	{0x34D8B89197D40045ULL, "vxw_sundermount_21a_0_0.vlm"},
	{0x34D8F6B1D293E37BULL, "gen00fl_merrill6115798.cl"},
	{0x34D99A9FB95A921EULL, "gen00fl_aveline6139992.cl"},
	{0x34D9B4F2745704A9ULL, "hlac_lt_0_lowtown_131073_1x.dds"},
	{0x34D9B94873AE5058ULL, "6071196_m.fxe"},
	{0x34D9B9FB9572C4C1ULL, "vxlt_0_lowtown_a3_n_c_ehu_0_0.vlm"},
	{0x34DA58018722FB3BULL, "one140cr_charade6137621.cl"},
	{0x34DABE538D28832EULL, "43982073.ogg"},
	{0x34DABF74B6A26A99ULL, "zz_approval_debug6003705.cl"},
	{0x34DAD264A7DE4089ULL, "vxca_cave_2bl_0_0.vlm"},
	{0x34DAFC195E04EA14ULL, "vxw_encounter3_fd_0_1.vlm"},
	{0x34DC3F11F92E3A02ULL, "follower_banter6141555.cl"},
	{0x34DC72BBD3EA2A5EULL, "vxht_mansion_n_624_0_0.vlm"},
	{0x34DC91B0A0E30210ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown4.msh"},
	{0x34DC9FB0A0E31A2AULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown4.mmh"},
	{0x34DCAF9731FB8A38ULL, "950771366.ogg"},
	{0x34DCC34431A17EEAULL, "1011850817.ogg"},
	{0x34DCDB7AA66C505DULL, "vxht_bartrand_n_5xa_0_0.vlm"},
	{0x34DCECFE27494F6AULL, "6075129_m.fxe"},
	{0x34DD316F2363E0DAULL, "gen00fl_isabela6079789.cl"},
	{0x34DD4CD59961FA24ULL, "542914517.ogg"},
	{0x34DD8FEC92B8A9AAULL, "700661059.ogg"},
	{0x34DDF1B5F8A7C5C7ULL, "zz_qun_debug6219183.cl"},
	{0x34DECBF70CCFFB91ULL, "6148824_m.fxe"},
	{0x34DECD48FC3F43B0ULL, "6042755_f.fxe"},
	{0x34DED5021FD5C0E5ULL, "vxdr_0_deeproads_cpl_0_0.vlm"},
	{0x34DF9A8B4B3BC6FEULL, "lgt210_magu_selby6142046.cl"},
	{0x34DF9D79A4F16DB1ULL, "zz_jp_rogue_persuade.cut"},
	{0x34DFAD51A312AFFBULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.impa.ani"},
	{0x34DFED9FFD994BC1ULL, "vxca_cavern1_haunted_i39_0_0.vlm"},
	{0x34E02A4E9F1E6797ULL, "441317717.ogg"},
	{0x34E0F525A48E77A1ULL, "vxht_chantry_2c3_0_0.vlm"},
	{0x34E15705F202C825ULL, "art\\vfx\\combat\\vfx_r_var_modal_act_c.impa.ani"},
	{0x34E1C0CD714788D7ULL, "zz_lgt_1_debug6135727.cl"},
	{0x34E1EFE978F2E0CCULL, "gen00fl_anders6125981.cl"},
	{0x34E22E6B17D1101CULL, "vxht_chantry_n_31r_0_0.vlm"},
	{0x34E26452D54E7022ULL, "vxw_encounter2_camp_nx_0_0.vlm"},
	{0x34E268BE9CD1BD29ULL, "vxw_bonepit_2af_0_0.vlm"},
	{0x34E413C9C57EB808ULL, "vxw_woundedcoast_5z_0_0.vlm"},
	{0x34E456DD8D8818D7ULL, "vxht_bartrand_n_5ad_0_0.vlm"},
	{0x34E491CB14CA9705ULL, "property_data.gda"},
	{0x34E521C0A4ED6E69ULL, "mag200pt_night_terrors.ncs"},
	{0x34E53B8EC8768CCBULL, "gen_im_flw_arc_varric_01.uti"},
	{0x34E5736B6D262606ULL, "6119185_f.fxe"},
	{0x34E5F5D3D8BC4061ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.anb"},
	{0x34E62F58E75799CDULL, "vxlt_gallowstemplar_n_3rm_0_0.vlm"},
	{0x34E6AB8570DC933BULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_c.mmh"},
	{0x34E7334914CBC8C4ULL, "367209487.ogg"},
	{0x34E73A626BF1C6F7ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_3x.dds"},
	{0x34E79DEADC9B8AECULL, "hm_bartend_friend.mor"},
	{0x34E79E4F0B64CDEAULL, "hlac_lt_0_lowtown_a2_131070_6y.dds"},
	{0x34E7CA5A3E80AB76ULL, "vxlt_darktown_509_0_0.vlm"},
	{0x34E7D0EB83C2C222ULL, "ser210_witness6168143.cl"},
	{0x34E8E2E828433B54ULL, "6144431_f.fxe"},
	{0x34E92922AC506F56ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.cess.ani"},
	{0x34E92DB643673AC7ULL, "6076665_f.fxe"},
	{0x34E935D8BEBBC8B5ULL, "vxlt_0_lowtown_n_gc9_0_0.vlm"},
	{0x34EA1BF061730481ULL, "254813022.ogg"},
	{0x34EA50C2F91F0472ULL, "tbe100ip_xenon.utp"},
	{0x34EA8E86DCC0BF9BULL, "6116511_m.fxe"},
	{0x34EA943DB987C0E3ULL, "6094684_m.fxe"},
	{0x34EA9CCB84E6281BULL, "6156766_m.fxe"},
	{0x34EAE9B2D0BF1EEBULL, "hlac_lt_0_lowtown_a3_n_c_0_1y.dds"},
	{0x34EB26C7B336F234ULL, "gen_im_npc_arm_hum_cht_rdra.uti"},
	{0x34EB62757648455BULL, "6086247_m.fxe"},
	{0x34EC11D12FDC2C5FULL, "vxlt_0_lowtown_docks_a3_btv_0_0.vlm"},
	{0x34EC4CA7A5C8DF4BULL, "401103707.ogg"},
	{0x34EC565838D51069ULL, "277398889.ogg"},
	{0x34ECD6E5D8493D05ULL, "vxht_0_hightown_a3_apt_0_0.vlm"},
	{0x34ECDC8069152DD6ULL, "299036555.ogg"},
	{0x34EDB551A31EFAF2ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.impa.evt"},
	{0x34EDC87BF49BA4AAULL, "qun221_petrice6077358.cl"},
	{0x34EDE2C186827D19ULL, "6118032_m.fxe"},
	{0x34EEDC5309EDB328ULL, "6092559_m.fxe"},
	{0x34EF3ECB6DFFC332ULL, "6187839_m.fxe"},
	{0x34F06F27ABF51A76ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robb_0t.dds"},
	{0x34F083EE18E7B31BULL, "336713491.ogg"},
	{0x34F0D62FC74488A2ULL, "vxw_encounter2_camp_iy_0_0.vlm"},
	{0x34F141A6C32B5954ULL, "964659960.ogg"},
	{0x34F171A352AAA535ULL, "703372683.ogg"},
	{0x34F1E8C895B7D900ULL, "zz_followers_banter6157376.cl"},
	{0x34F1FCD5D804A675ULL, "vxlt_0_lowtown_a2_n_e11_0_0.vlm"},
	{0x34F25FA6FDFF082DULL, "360034230.ogg"},
	{0x34F26D0446777C23ULL, "hlac_ht_mansion2_818_27z.dds"},
	{0x34F28223239F6C3BULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.anb"},
	{0x34F29E7DD95E246EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1a.msh"},
	{0x34F2AC7DD95E3BBFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1a.mao"},
	{0x34F2B87DD95E505CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1a.mmh"},
	{0x34F2D13E8012937BULL, "654175791.ogg"},
	{0x34F2D6D3D8C695DAULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.mmh"},
	{0x34F2DCD49493EADBULL, "vxlt_0_lowtown_n_gj0_0_0.vlm"},
	{0x34F3CC8570E755A0ULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_c.anb"},
	{0x34F3D451A3239A85ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.impa.gad"},
	{0x34F511511489CFDDULL, "842501832.ogg"},
	{0x34F599FACE12E069ULL, "6166316_m.fxe"},
	{0x34F5B2C4914DC23EULL, "hlac_lt_gallowscourtyard_n_-131073_1z.dds"},
	{0x34F5B717F627D6A9ULL, "6080988_m.fxe"},
	{0x34F5E24858B9E449ULL, "375786096.ogg"},
	{0x34F601D98E4813C0ULL, "vxht_chantry_2p3_0_0.vlm"},
	{0x34F63122AC5B0747ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.cess.evt"},
	{0x34F64412B1229711ULL, "pro000_merchant6051269.cl"},
	{0x34F74DF17D81B064ULL, "vxca_caverns2_865_0_1.vlm"},
	{0x34F79471E8588046ULL, "dae100_bran6118339.cl"},
	{0x34F7D031B3BD5018ULL, "691132918.ogg"},
	{0x34F7EA62C5184A43ULL, "hlac_ht_mansion2b_818_47z.dds"},
	{0x34F817C8E1232465ULL, "vxht_chantry_2im_0_0.vlm"},
	{0x34F888BC12653948ULL, "hf_qmq_visguard3.mor"},
	{0x34F8F5C0A81AF6A6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic_1.mao"},
	{0x34F92B2A662E6CE2ULL, "gen00flst_sebacttwoa.stg"},
	{0x34F97C6C639BCEF9ULL, "lgt110_gxa_vgk_conspirator1.bnk"},
	{0x34F988CF104F3AEDULL, "dae000ip_crowd_mf7_drunk2.utp"},
	{0x34FA14450A59C103ULL, "vxw_bonepitdestroyed_a3_8c_0_0.vlm"},
	{0x34FA223EEAE8CBE5ULL, "166581475.ogg"},
	{0x34FA7CBAABF2B523ULL, "mag370_meredith6037401.cl"},
	{0x34FAB58FA5669AE1ULL, "vxw_sundermount_202_0_0.vlm"},
	{0x34FAB960255D48ADULL, "dae_monster_basic_attacks.ncs"},
	{0x34FAC7509F46C9BEULL, "llac_ht_keep_n_3732_5.dds"},
	{0x34FB065C12A9D5C3ULL, "788531982.ogg"},
	{0x34FB3DAA5D458A94ULL, "rog_armistice.dds"},
	{0x34FB63049D5A602CULL, "lt_0_lowtown_32_0_lc.mmh"},
	{0x34FBE27C85F38410ULL, "mag100_cullen6148669.cl"},
	{0x34FBF410C0AAC362ULL, "lt_0_lowtown_ll.mmh"},
	{0x34FC1249FB7AEA6FULL, "vxht_0_hightown_a3_byv_0_0.vlm"},
	{0x34FC9E6C639E3E59ULL, "lgt110_gxa_vgk_conspirator1.cnv"},
	{0x34FCE98D9B3C3082ULL, "6149891_m.fxe"},
	{0x34FCF81D1931436AULL, "art\\characters\\creatures\\hurlock\\textures\\cn_hed_hlk_0s.dds"},
	{0x34FD0A54246BA68EULL, "vxw_bonepitmining_a2_176_0_0.vlm"},
	{0x34FD5222AC615D90ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.cess.gad"},
	{0x34FE44AB2ECFF1D3ULL, "lowtown_fem_cityguard_mob.bnk"},
	{0x34FE9E27BE467785ULL, "zz_slate4_v1.cut"},
	{0x34FEE4B90B568D6AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar3_.gad"},
	{0x34FF131C877223E3ULL, "character_character_p206.psh.fxo.dx11.assembly"},
	{0x34FFB3FFC9D00B3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_action.gad"},
	{0x3500C8C3A7CDD14BULL, "hlac_ht_keep_3732_17y.dds"},
	{0x3500D270236CBD32ULL, "follower_banter6163443.cl"},
	{0x3500F6CFA0FC516DULL, "vxlt_0_lowtown_a2_n_c_ett_0_0.vlm"},
	{0x350128365B24BC4AULL, "222181013.ogg"},
	{0x35013A97717449D7ULL, "vxht_bartrand_n_5of_0_0.vlm"},
	{0x35015A380DC5A9E7ULL, "543616080.ogg"},
	{0x350166AB2ED26123ULL, "lowtown_fem_cityguard_mob.cnv"},
	{0x35018C1FE297DB90ULL, "6164283_m.fxe"},
	{0x35024CFD3CFDE9B1ULL, "hlac_w_encounter2_camp_393215_0z.dds"},
	{0x3502944001CB1A9AULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_cst_end.gad"},
	{0x3502CC08BADC4460ULL, "6072702_m.fxe"},
	{0x35034DB592962A1BULL, "6087796_m.fxe"},
	{0x3503AE1DF1CA94E1ULL, "dae000_male_commoner1_mob.bnk"},
	{0x35042E380DFB4D67ULL, "6085493_m.fxe"},
	{0x35043E4566BB46C8ULL, "vxw_sundermount_207_0_0.vlm"},
	{0x350657C9FAE5B67CULL, "6124786_m.fxe"},
	{0x35067F27B8AAEDB4ULL, "static_static_p1.psh.fxo.dx11.assembly"},
	{0x3506A8FFC9D6172CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_action.evt"},
	{0x3506AE88DB7B8E32ULL, "vxht_0_hightown_a2_aso_0_0.vlm"},
	{0x3506D01DF1CD0451ULL, "dae000_male_commoner1_mob.cnv"},
	{0x3507C041FD8BD017ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.impa.gad"},
	{0x3508FAF49BE307B9ULL, "plt_cod_aow_spellcombo7.plo"},
	{0x3508FE8662E89E91ULL, "6214693_m.fxe"},
	{0x35090E4E2F257F1EULL, "6082682_m.fxe"},
	{0x35091EFBF15A2B2CULL, "vxw_encounter2_j2_0_0.vlm"},
	{0x35092694A7BCD9E2ULL, "363577306.ogg"},
	{0x3509276C975CB361ULL, "vxw_sundermount_251_0_0.vlm"},
	{0x3509974001D13EB1ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_cst_end.evt"},
	{0x3509D147CB6B20C9ULL, "ser270_templar6150307.cl"},
	{0x3509F4EC20F6E2DAULL, "qun131_warrens_thug.stg"},
	{0x350A0FE6A8949480ULL, "vxlt_darktown_5ap_0_0.vlm"},
	{0x350A495986607762ULL, "hlac_lt_0_lowtown_a2_n_c_1_11y.dds"},
	{0x350AE63E2DC9C51AULL, "vxlt_warehouse_n_qs_0_0.vlm"},
	{0x350B26CC9DB26F73ULL, "hlac_ht_0_hightown_n_-65537_2z.dds"},
	{0x350B5B6A8D324902ULL, "6142991_m.fxe"},
	{0x350B5BE2D0B27DB4ULL, "dae300_alistair_king6134979.cl"},
	{0x350B9F44B792E54DULL, "6207796_m.fxe"},
	{0x350C3E70E5136230ULL, "6132864_m.fxe"},
	{0x350C512594FC0B5CULL, "llac_ca_cave_2194_7.dds"},
	{0x350C590F6675A08AULL, "art\\vfx\\combat\\vfx_c_mer_attk1_c.anb"},
	{0x350C8AF71AB1F210ULL, "fen321ar_base.lst"},
	{0x350CABE7249E258CULL, "296167326.ogg"},
	{0x350CAF1D24B017C2ULL, "6039816_f.fxe"},
	{0x350D043F89508A8AULL, "character_character_p151.psh.fxo.dx11"},
	{0x350D28FFBBE237B2ULL, "mag311_nyssa6041770.cl"},
	{0x350D590DAC7A0CB6ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_3craggyrock.msh"},
	{0x350D630DAC7A1DB4ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_3craggyrock.mmh"},
	{0x350D670DAC7A24C7ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_3craggyrock.mao"},
	{0x350D6BC51A43C597ULL, "vxlt_backalley_n_1bm_0_0.vlm"},
	{0x350DB8A161CE6CF9ULL, "807026622.ogg"},
	{0x350E0FC013DD8AA9ULL, "252066648.ogg"},
	{0x350E431D377490ADULL, "vrc321_varric6087056.cl"},
	{0x350E67A544773D43ULL, "gen_im_trs_bst_claws_01.uti"},
	{0x350ED541FD9212A4ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.impa.evt"},
	{0x350F5D8D38FDFF1FULL, "59995204.ogg"},
	{0x350F73F872543C32ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.cess.gad"},
	{0x350F88C4B0F7B9AEULL, "vxht_mansion2_2ar_0_0.vlm"},
	{0x350FA583EF932436ULL, "6150039_m.fxe"},
	{0x35107513A8E4FDF0ULL, "6120012_m.fxe"},
	{0x3510851E30704008ULL, "zz_ave_debug6064151.cl"},
	{0x3510B261D4DF41F6ULL, "363123611.ogg"},
	{0x35110094EA1990F7ULL, "254295730.ogg"},
	{0x351100B60B2E533AULL, "hlac_lt_foundry_n_320_21z.dds"},
	{0x3511264D8E890960ULL, "vxlt_gallowstemplar_437_0_0.vlm"},
	{0x3511685BF49427FAULL, "vxw_encounter2_99_0_0.vlm"},
	{0x3511F60970F823A4ULL, "isa211_isabela6069983.cl"},
	{0x3512ED9E6D0A9DAFULL, "vxht_mansion2_1b9_0_0.vlm"},
	{0x35131495C3B8628AULL, "vxw_bonepitmining_a2_1pg_0_0.vlm"},
	{0x35132372E47C531DULL, "vxca_caverns1_iod_0_0.vlm"},
	{0x35136AF71AB7DAC1ULL, "fen321ar_base.ncs"},
	{0x3513B35B0D0481D4ULL, "lt_0_lowtown_40_0.rml"},
	{0x3513BF4E1491DA74ULL, "vxlt_0_lowtown_a2_n_ehc_0_0.vlm"},
	{0x3513BFB90B68AA50ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar3_.ani"},
	{0x3513BFD56396F39BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medb_0s.dds"},
	{0x351444619452FEF0ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scroll_add_vt.mao"},
	{0x351490FFC9E22B85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_action.ani"},
	{0x3514BEDB11EE108BULL, "art\\vfx\\combat\\vfx_c_ogr_prone_p.impa.evt"},
	{0x3514E4A9B91134D8ULL, "dae000_gamlen6121548.cl"},
	{0x351515BC74CA4BF6ULL, "llac_lt_ruinedhouse_n_-1_8.dds"},
	{0x35152065CD71A6B2ULL, "vxht_0_hightown_a2_anh_0_0.vlm"},
	{0x35157F503D26C916ULL, "vxht_mansion_n_6jt_0_0.vlm"},
	{0x3515BF54A1977681ULL, "hlac_lt_0_lowtown_65536_7z.dds"},
	{0x3515C45C16C0C628ULL, "vxw_encounter2_e6_0_1.vlm"},
	{0x3515DFC033115A99ULL, "6147704_m.fxe"},
	{0x351627D187E3307BULL, "vxlt_0_lowtown_a2_he5_0_0.vlm"},
	{0x35165C6CDB882583ULL, "vxw_sundermount_378_0_0.vlm"},
	{0x3516619EB0B98D11ULL, "seb201_lady_harimann6177393.cl"},
	{0x351676F8725A6069ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.cess.evt"},
	{0x35169E7FBE65D548ULL, "ser221_gascard6122455.cl"},
	{0x3516C487BBC05D33ULL, "vxlt_gallowstemplar_n_5t3_0_0.vlm"},
	{0x3516FFB1245CF367ULL, "hlac_ca_caverns1_mines_24424_18z.dds"},
	{0x3517354F7C5E8043ULL, "llac_lt_0_lowtown_docks_-65538_3.dds"},
	{0x35174307ED835BC4ULL, "6105952_m.fxe"},
	{0x35177B7772EEA658ULL, "6175788_m.fxe"},
	{0x351781F71ABBEAFBULL, "fen321ar_base.are"},
	{0x35178F249D29AE30ULL, "hlac_ca_cavern1_haunted_24424_22z.dds"},
	{0x3517AF4001DDA480ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_cst_end.ani"},
	{0x351881C39A8AAD00ULL, "vxlt_darktown_5jq_0_0.vlm"},
	{0x3518A2D9E9939745ULL, "hlac_w_woundedcoast_262149_0z.dds"},
	{0x3518CFBB5AB47A30ULL, "dae011_corff_sta6164586.cl"},
	{0x3518EA17717648A5ULL, "hlac_lt_gallowstemplar_f_2383_7y.dds"},
	{0x351927A5E763B246ULL, "vxlt_0_lowtown_a2_byn_0_0.vlm"},
	{0x3519F884C8A99DC3ULL, "hlac_lt_backalley_n_65536_0y.dds"},
	{0x351A3C918C6B2881ULL, "824595558.ogg"},
	{0x351A51077937FB4CULL, "art\\characters\\heraldry\\hld_starkhaven.dds"},
	{0x351A5FEB17A2764CULL, "404526182.ogg"},
	{0x351A97614999275CULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_brazier_n.dds"},
	{0x351B2237B98685C7ULL, "gen00fl_fenris6128080.cl"},
	{0x351B4A0BF5D7B52AULL, "follower_banter6143258.cl"},
	{0x351B58A1A7050A1FULL, "271884070.ogg"},
	{0x351B6A4E014E1D6FULL, "vxlt_0_lowtown_a2_n_hgh_0_0.vlm"},
	{0x351B876C33A78374ULL, "6142650_m.fxe"},
	{0x351BD7DB11F45A4CULL, "art\\vfx\\combat\\vfx_c_ogr_prone_p.impa.gad"},
	{0x351BED41FD9CC59DULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.impa.ani"},
	{0x351C7DDBC6C1E687ULL, "125390702.ogg"},
	{0x351D8F4D92550DC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_loop.gad"},
	{0x351F7C46151C8645ULL, "vxw_encounter2_34_0_0.vlm"},
	{0x351F9834B856808AULL, "vxlt_0_lowtown_docks_a3_c46_0_0.vlm"},
	{0x35200D5AF91C16D2ULL, "nat380_nathaniel6090039.cl"},
	{0x35205423A6347325ULL, "6084728_m.fxe"},
	{0x35207038546D63A4ULL, "gen00fl_anders6096504.cl"},
	{0x3520B4F51B45064FULL, "vxca_caverns1_mines_jfc_0_0.vlm"},
	{0x3522D6DB11FA774AULL, "art\\vfx\\combat\\vfx_c_ogr_prone_p.impa.ani"},
	{0x3522DED90CA99728ULL, "vxca_caverns1_mines_ila_0_0.vlm"},
	{0x352332B9F2996D9BULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.dura.ani"},
	{0x35235ED9DDD3FA2AULL, "6208264_m.fxe"},
	{0x35235EF87264C1D8ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.cess.ani"},
	{0x3523BFF1A736DB51ULL, "vxca_caverns1_mines_e7i_0_0.vlm"},
	{0x3523F12D2F1CE196ULL, "vxht_mansion_n_6mu_0_0.vlm"},
	{0x3524B24D925B67B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_loop.evt"},
	{0x3524BA947CE2EAE2ULL, "vxlt_gamlenshouse_n_5q_0_0.vlm"},
	{0x3524CE3B41CC72F6ULL, "art\\vfx\\models\\vfx_rock.mmh"},
	{0x3524D83B41CC83F4ULL, "art\\vfx\\models\\vfx_rock.msh"},
	{0x35251D41D86E2DA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_enter.ani"},
	{0x35253D186A18E3CAULL, "6147611_m.fxe"},
	{0x3525A087FE3F2379ULL, "584902258.ogg"},
	{0x3525D3C86304DDE4ULL, "mag360_thrask6043520.cl"},
	{0x352602B5D5465C2CULL, "vxw_encounter2_p0_0_0.vlm"},
	{0x352653B5AD022D44ULL, "hlac_dr_primevaltomb_11985_2z.dds"},
	{0x352660A7FE22256BULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_dwn_c.mmh"},
	{0x352767E0D15D33EFULL, "vxht_0_hightown_a3_bns_0_0.vlm"},
	{0x3527CF4C1083B937ULL, "6036802_m.fxe"},
	{0x3527F12A0D8F1F80ULL, "6053549_m.fxe"},
	{0x352833AB58719BD5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scroll.msh"},
	{0x352835AB58719F4BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scroll.mmh"},
	{0x352841AB5871B3E8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scroll.mao"},
	{0x35292EA14F25FB87ULL, "hlac_lt_0_lowtown_n_-196608_4x.dds"},
	{0x352980901C019911ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_att_c.anb"},
	{0x352A3B90E89D7254ULL, "189480888.ogg"},
	{0x352A916D6989BDA5ULL, "364169547.ogg"},
	{0x352B1A41D8729406ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_enter.gad"},
	{0x352B54DBF761347DULL, "26498208.ogg"},
	{0x352BA2AFE6CABB0DULL, "718107102.ogg"},
	{0x352C19C4FF8FD8A0ULL, "6082257_m.fxe"},
	{0x352C1BCB350D56D8ULL, "vxlt_darktown_43c_0_0.vlm"},
	{0x352C36A1FC8E10CCULL, "art\\vfx\\textures\\vfx_slash_fb.dds"},
	{0x352C77AC72E94D1AULL, "vxlt_darktown_4we_0_0.vlm"},
	{0x352C7E3E641744DCULL, "6136363_m.fxe"},
	{0x352C7FC5CDEAAB07ULL, "6060409_m.fxe"},
	{0x352CA8E30439EDEFULL, "558432800.ogg"},
	{0x352D485E9FCA36BAULL, "6148881_m.fxe"},
	{0x352D4BC00B3D377EULL, "dre000cr_dougal.utc"},
	{0x352D6CA4B23B6E3CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww31_slopec.mao"},
	{0x352D78A4B23B82DFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww31_slopec.mmh"},
	{0x352D7EA4B23B8C81ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww31_slopec.msh"},
	{0x352DAE3C03CC9BBBULL, "hlac_lt_warehouse_65537_3y.dds"},
	{0x352E086DF2978A36ULL, "vxdr_0_deeproads_cfa_0_0.vlm"},
	{0x352E1FFD22F424A6ULL, "6131198_f.fxe"},
	{0x352E4A7F1DB3B535ULL, "eye_eye_p0.psh.fxo"},
	{0x352E9899713E8F87ULL, "hlac_w_encounter3_-3_0x.dds"},
	{0x352EA8E7F0690EE0ULL, "gen00fl_fenris6112053.cl"},
	{0x352EE7BC6C290753ULL, "vxht_bartrand_n_67m_0_0.vlm"},
	{0x352F60195BB2C80DULL, "6196856_m.fxe"},
	{0x352FEA50E0113D6CULL, "173932456.ogg"},
	{0x35303AB9F2A40592ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.dura.evt"},
	{0x353056D735FA4A11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3closed.mmh"},
	{0x353062D735FA5EB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3closed.mao"},
	{0x35306ADD2E902155ULL, "107208184.ogg"},
	{0x353070D735FA76CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3closed.msh"},
	{0x3530A307F87E0B74ULL, "vxlt_0_lowtown_a2_n_eva_0_0.vlm"},
	{0x3530F13B970DB0BFULL, "6144589_m.fxe"},
	{0x3531BA4D9265FFA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_loop.ani"},
	{0x35321541D878AAE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_enter.evt"},
	{0x353225DB26EB9CF2ULL, "vxht_0_hightown_a3_c5f_0_0.vlm"},
	{0x353250506DDBCA71ULL, "vxdr_0_deeproads_f56_0_0.vlm"},
	{0x35325AB63E24219BULL, "772225949.ogg"},
	{0x3532973697E1D8EBULL, "6129126_m.fxe"},
	{0x35331F44204C8852ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.cess.gad"},
	{0x353338259A0C148CULL, "gen00fl_bethany6101516.cl"},
	{0x353341A7FE2C7AD0ULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_dwn_c.anb"},
	{0x35336E6EF90898A3ULL, "vxw_woundedcoast_hr_0_0.vlm"},
	{0x353381AAB42F07FFULL, "lt_0_lowtown_a2_n_c_7_0.msh"},
	{0x353387AAB42F11A1ULL, "lt_0_lowtown_a2_n_c_7_0.mmh"},
	{0x353393AAB42F2642ULL, "lt_0_lowtown_a2_n_c_7_0.mao"},
	{0x3533B5705D3668D1ULL, "dae000_cora6162455.cl"},
	{0x35340C54836E284EULL, "90557475.ogg"},
	{0x35343E8D6DEFFCD8ULL, "hla_ht_0_hightown_a2_-131074_0y.dds"},
	{0x35345C0F6696F87DULL, "art\\vfx\\combat\\vfx_c_mer_attk1_c.mmh"},
	{0x35347B71EAB97F2DULL, "mag120_macha6080682.cl"},
	{0x353495FE68274061ULL, "hlac_w_bonepitmining_a2_458752_0x.dds"},
	{0x3535132F4FDA30FEULL, "hlac_ht_playershouse_n_855_34x.dds"},
	{0x35359C423029304BULL, "and231tr_smugglers2.utt"},
	{0x3535D9BDC3534C6FULL, "vxht_0_hightown_a3_ber_0_0.vlm"},
	{0x353604BF8BE39B6BULL, "qcr000_arishok6059997.cl"},
	{0x353622EF0C85CCBFULL, "lt_0_lowtown_docks_132_0.phy"},
	{0x3536A1901C0C5B8AULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_att_c.mmh"},
	{0x3536D55C86B4E02FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\dw_tal_crp_atk.gad"},
	{0x3536EEEA76FA561DULL, "zz_dae_debug6069797.cl"},
	{0x3537167E1215D454ULL, "6066707_m.fxe"},
	{0x35374A0B4FE48CD6ULL, "vxlt_darktown_6hj_0_0.vlm"},
	{0x353759B9F2AA5925ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.dura.gad"},
	{0x35379D7585940C86ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_min_c.anb"},
	{0x3537D563C21EA13CULL, "art\\vfx\\combat\\vfx_glb_spi_imp_maj_c.impa.ani"},
	{0x3537ED3DE0D1105EULL, "vxca_cave_1nd_0_0.vlm"},
	{0x353868722D7DA5D2ULL, "vxht_bartrand_n_67h_0_0.vlm"},
	{0x353922442050F889ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.cess.evt"},
	{0x3539570ED0991087ULL, "vxlt_gallowstemplar_n_5ri_0_0.vlm"},
	{0x3539576D2CC2D603ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampnail.msh"},
	{0x35395C86475D8093ULL, "llac_lt_0_lowtown_0_8.dds"},
	{0x3539656D2CC2EE5DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampnail.mmh"},
	{0x3539696D2CC2F56EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampnail.mao"},
	{0x3539B1F573A218DFULL, "704989079.ogg"},
	{0x353A9CF5930A9535ULL, "hlac_dr_0_deeproads_17038_3y.dds"},
	{0x353ADE5B86DA032EULL, "vxht_mansion2_2pu_0_0.vlm"},
	{0x353B04328E0A83EDULL, "ser221_party_comments6163787.cl"},
	{0x353B064E6A2706F7ULL, "hf_amb_commoner2.mor"},
	{0x353B7523256076F1ULL, "hlac_lt_0_lowtown_n_0_2y.dds"},
	{0x353B8E37217AF9C8ULL, "ser221_gascard6148837.cl"},
	{0x353BABA12C633C26ULL, "dae000_gamlen6121555.cl"},
	{0x353C28878D63DEDDULL, "ico_ring_4.dds"},
	{0x353C5FFFF908B745ULL, "vxw_encounter2_92_0_0.vlm"},
	{0x353D75F1861E57D8ULL, "fen110_party_comment6095073.cl"},
	{0x353DA0372142D156ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning2sail.msh"},
	{0x353DA2CECB58FDFAULL, "zz_qcr_debug6215494.cl"},
	{0x353DA7B6DD56793FULL, "t3_cth_nb2.tnt"},
	{0x353DAA372142E254ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning2sail.mmh"},
	{0x353DAE372142E967ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning2sail.mao"},
	{0x353DE95A07112CE2ULL, "639359465.ogg"},
	{0x353DEA5C86BB22DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\dw_tal_crp_atk.evt"},
	{0x353E729486C1BCB5ULL, "6201493_m.fxe"},
	{0x353E97FFD832E363ULL, "6108314_m.fxe"},
	{0x353EB3D09BB7C0F0ULL, "351808126.ogg"},
	{0x353EE263C224D67EULL, "art\\vfx\\combat\\vfx_glb_spi_imp_maj_c.impa.gad"},
	{0x353EE792DD145FA1ULL, "nat380_temmerin6067389.cl"},
	{0x353F0BDB5C80DB2FULL, "art\\vfx\\creatures\\fxc_shade_c.mmh"},
	{0x353F8291DF750CB3ULL, "6175963_m.fxe"},
	{0x353F9AE281AFEB92ULL, "vxht_mansion2_335_0_0.vlm"},
	{0x354038A6870EDD5BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1plank.mmh"},
	{0x354044A6870EF1F8ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1plank.mao"},
	{0x354056A6870F1045ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1plank.msh"},
	{0x3540754CE2759A7FULL, "vxht_mansion2_1ry_0_0.vlm"},
	{0x35415D9C49F23C21ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallred.phy"},
	{0x3541E115B339EA60ULL, "mcr371_scared_mage_m.utc"},
	{0x354236401BE7F8F8ULL, "vxht_0_hightown_c0z_0_0.vlm"},
	{0x35425EA1725AAA24ULL, "vxht_mansion_n_6j6_0_0.vlm"},
	{0x3542B7AFECD4E1F7ULL, "494638179.ogg"},
	{0x354307126513514EULL, "vxht_mansion2_1sw_0_0.vlm"},
	{0x3543808B12636124ULL, "dae000cr_grand_cleric6141056.cl"},
	{0x3543FB25464DA0A8ULL, "vxlt_0_lowtown_a2_n_e3b_0_0.vlm"},
	{0x3543FFB83DB2C522ULL, "923193931.ogg"},
	{0x354407E653F54B39ULL, "gen_im_trs_drk_two_2hs_r1.uti"},
	{0x354457F654B5012EULL, "hlac_lt_0_lowtown_docks_a3_n_131071_3x.dds"},
	{0x3544603E3CA00F4EULL, "6206064_m.fxe"},
	{0x3544736D49D7BD78ULL, "vxht_0_hightown_n_cf7_0_0.vlm"},
	{0x354558C0898E027DULL, "505876963.ogg"},
	{0x3545A07585A04F31ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_min_c.mmh"},
	{0x3545BD63C22AB59DULL, "art\\vfx\\combat\\vfx_glb_spi_imp_maj_c.impa.evt"},
	{0x3545C52EF031DEC1ULL, "330662063.ogg"},
	{0x35460C17EB702D44ULL, "t1_mul_pi1.tnt"},
	{0x3546A9423037B9FAULL, "and231tr_smugglers2.ncs"},
	{0x35470A44205D0DF8ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.cess.ani"},
	{0x354723EF0C9441C4ULL, "lt_0_lowtown_docks_132_0.mmh"},
	{0x354727EF0C944817ULL, "lt_0_lowtown_docks_132_0.mao"},
	{0x354739EF0C946766ULL, "lt_0_lowtown_docks_132_0.msh"},
	{0x35475CD86C8F53BFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09b.phy"},
	{0x35476B8C24722BFDULL, "vxlt_gallowstemplar_34x_0_0.vlm"},
	{0x35476E5B44E2B6DBULL, "727737684.ogg"},
	{0x354809A6B6BDC516ULL, "hlac_lt_0_lowtown_a2_2_0y.dds"},
	{0x354827C60824282BULL, "vxht_0_hightown_n_a71_0_0.vlm"},
	{0x35482AAD7BC3BDE5ULL, "402530708.ogg"},
	{0x35488886B573E755ULL, "vxlt_0_lowtown_a2_e38_0_0.vlm"},
	{0x3548D33B28F74531ULL, "hlac_lt_0_lowtown_a2_-131071_2x.dds"},
	{0x354967402D11108DULL, "267687539.ogg"},
	{0x354A2FD4C2188217ULL, "hlac_lt_gallowstemplar_n_2394_5y.dds"},
	{0x354A38449F0543D9ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.impa.ani"},
	{0x354A38A6B3A56CBFULL, "mcr310cr_bloodmage.utc"},
	{0x354ACCA40CFBB67EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3.ani"},
	{0x354B070D6413C173ULL, "gen_im_trs_prf_junk2_04.uti"},
	{0x354B4103948967F8ULL, "vis121_saemus6113521.cl"},
	{0x354B471FD7263536ULL, "6115986_m.fxe"},
	{0x354B54A315DF045AULL, "vxlt_0_lowtown_a2_df1_0_0.vlm"},
	{0x354B62C1BFFCF90FULL, "ser221_gamlen6135228.cl"},
	{0x354BD4DF455FBC08ULL, "narrative_banter6149652.cl"},
	{0x354BF25C86C76DF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\dw_tal_crp_atk.ani"},
	{0x354C0D40B6A7C1B7ULL, "812221409.ogg"},
	{0x354CA3C5E58747A7ULL, "486061592.ogg"},
	{0x354CBD543B10B253ULL, "197143642.ogg"},
	{0x354DB629C0C9114CULL, "402548243.ogg"},
	{0x354DC778A25BB612ULL, "6118156_m.fxe"},
	{0x354DD7C1345AD9CAULL, "hlac_lt_0_lowtown_a3_n_c_-131073_3z.dds"},
	{0x354E8BCF556F11A2ULL, "vxlt_0_lowtown_a3_n_c_gz2_0_0.vlm"},
	{0x354EE239C6E9581FULL, "tools_sun.msh"},
	{0x354EE729D46BB2FFULL, "vxht_mansion2_1ux_0_0.vlm"},
	{0x354EE839C6E962C1ULL, "tools_sun.mmh"},
	{0x354F31D1EEAB94CBULL, "vxlt_0_lowtown_gb3_0_0.vlm"},
	{0x354F50C836D2133AULL, "zz_mcr_debug6091900.cl"},
	{0x354FC39532306994ULL, "6096331_m.fxe"},
	{0x35505228DCF4C9A3ULL, "vxw_woundedcoast_bt_0_0.vlm"},
	{0x3550664E418528F9ULL, "279448696.ogg"},
	{0x3550D07E6450C2A4ULL, "vxht_mansion_n_6m7_0_0.vlm"},
	{0x3550FCB79AFF9B42ULL, "llac_w_woundedcoast_-131072_0.dds"},
	{0x355113449F0B231BULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.impa.gad"},
	{0x355143C96CB46DA9ULL, "632883184.ogg"},
	{0x35518EFA76674F5CULL, "zz_rdr_debug6066634.cl"},
	{0x35519D94FD965979ULL, "824432046.ogg"},
	{0x35522316BFC1AB82ULL, "dae000_mother6035949.cl"},
	{0x3553394DFF95B362ULL, "416296886.ogg"},
	{0x3553DADA2534ECDDULL, "314959013.ogg"},
	{0x3553FA37F6020554ULL, "6068907_m.fxe"},
	{0x35558D36E3D66020ULL, "6047518_f.fxe"},
	{0x3555A515D4124017ULL, "vxlt_backalley_n_1yh_0_0.vlm"},
	{0x3555BD11A5DB0636ULL, "vxht_chantry_2ta_0_0.vlm"},
	{0x3557102F6876D9C7ULL, "llac_lt_0_lowtown_a2_n_c_65534_6.dds"},
	{0x3557491FCF4BADEFULL, "6015920-intro.cut"},
	{0x3557C00C78A87C74ULL, "vxca_caverns1_mines_i0i_0_0.vlm"},
	{0x355830449F117300ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.impa.evt"},
	{0x355842C8C69BB1AFULL, "zz_and_debug6178609.cl"},
	{0x35585DD86C9DC8C4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09b.mmh"},
	{0x355871E319452B22ULL, "gen00fl_varric6075644.cl"},
	{0x355873D86C9DEE66ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09b.msh"},
	{0x35588F78B02C81E6ULL, "165226626.ogg"},
	{0x3558B3E4815D7374ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian_1.mao"},
	{0x3558BB7E875B90AFULL, "dae000_bodhan6146867.cl"},
	{0x3558D592334660DCULL, "935914339.ogg"},
	{0x3558E4A40D081C4FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3.evt"},
	{0x3559302EF0111684ULL, "vxdr_0_deeproads_cv3_0_0.vlm"},
	{0x355950AEF1B982D7ULL, "w_woundedcoast_47_0.anb"},
	{0x3559803A3797D883ULL, "595460608.ogg"},
	{0x3559C6CFC4B4A90EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_en.ani"},
	{0x3559E7BCBF8FA1C8ULL, "w_bonepitmining_a2_123_0.msh"},
	{0x3559F9BCBF8FC059ULL, "w_bonepitmining_a2_123_0.mao"},
	{0x355A05BCBF8FD4F2ULL, "w_bonepitmining_a2_123_0.mmh"},
	{0x355A4C3E7600302BULL, "161698536.ogg"},
	{0x355A8AF7A7352851ULL, "6217010_m.fxe"},
	{0x355AB956197ADCC5ULL, "vxw_sundermount_2bh_0_0.vlm"},
	{0x355C0924D15201A8ULL, "mer120_hubert6173908.cl"},
	{0x355C762D033FACCCULL, "qun131_taskmaster6039610.cl"},
	{0x355C7DBFD048FD7BULL, "6201239_m.fxe"},
	{0x355CB12E47303ADCULL, "6164962_m.fxe"},
	{0x355CF33FE008346EULL, "vxlt_undercityserialkiller_14c_0_0.vlm"},
	{0x355D08E883720740ULL, "llac_w_woundedcoast_0_3.dds"},
	{0x355D4937B69B5D0EULL, "dae011_gossips6149748.cl"},
	{0x355D4E0D96EFE85AULL, "6065797_m.fxe"},
	{0x355DA3AEE0A1AD4BULL, "vxlt_0_lowtown_gi2_0_0.vlm"},
	{0x355DDC0AA553AC5BULL, "81304513.ogg"},
	{0x355DE922CA329F47ULL, "71379602.ogg"},
	{0x355DF1ECBC20A08EULL, "730429917.ogg"},
	{0x355E4B8CF0B69946ULL, "vxw_bonepitdestroyed_a3_1p9_0_0.vlm"},
	{0x355E96568925A608ULL, "328464270.ogg"},
	{0x355F88375903C2F9ULL, "gen00fl_anders6197770.cl"},
	{0x355FE5A40D0E3CF8ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3.gad"},
	{0x355FF915B3531031ULL, "mcr371_scared_mage_m.bnk"},
	{0x35601D85C762F2ACULL, "vxw_blightlands_69_0_0.vlm"},
	{0x356058573B8382A1ULL, "301284302.ogg"},
	{0x356088AAB456011EULL, "lt_0_lowtown_a2_n_c_7_0.phy"},
	{0x3560DA9859013EA1ULL, "6077058_m.fxe"},
	{0x3561C421B3A2C01DULL, "173719139.ogg"},
	{0x3561E7D85F043BF3ULL, "mrl150_merrill6093972.cl"},
	{0x3562C2C71E31C248ULL, "388739882.ogg"},
	{0x3562C7697DBE2C34ULL, "art\\characters\\playercharacter\\elfmale\\arm_nuda.phy"},
	{0x3562FA85B81E299FULL, "504927073.ogg"},
	{0x35630222F9CE70C2ULL, "vxlt_blackemporium_zk_0_0.vlm"},
	{0x3563480536F56DBEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sis_flemeth_real.ani"},
	{0x35634A6171F323F5ULL, "vxlt_0_lowtown_docks_a3_n_cxb_0_0.vlm"},
	{0x35634CAD428BE9F6ULL, "173997864.ogg"},
	{0x35636B69C4F29BF7ULL, "242468874.ogg"},
	{0x3563DB15B356C641ULL, "mcr371_scared_mage_m.cnv"},
	{0x3563DC6A82C0430EULL, "6093914_m.fxe"},
	{0x3564ACA5A67B5A07ULL, "vxlt_gallowstemplar_n_5cj_0_0.vlm"},
	{0x356522F763213806ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraitc_d.dds"},
	{0x3565820C217B3534ULL, "vxca_caverns1_djf_0_1.vlm"},
	{0x3565BF47A9C0979CULL, "dre000_stroud6036701.cl"},
	{0x3566187B5CAA7FE3ULL, "gen00flst_bethany_carver3.stg"},
	{0x35662A1BB20252BFULL, "819623611.ogg"},
	{0x35668FB224C4B84BULL, "179445346.ogg"},
	{0x3566BECFC4BF265FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_en.evt"},
	{0x3566F2C1332FB209ULL, "llac_lt_gallowstemplar_2394_1.dds"},
	{0x3567066270E4B48EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_destroyed.msh"},
	{0x3567146270E4CCDFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_destroyed.mao"},
	{0x35671AB8B786B410ULL, "vxlt_blackemporium_tc_0_0.vlm"},
	{0x3567206270E4E17CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_destroyed.mmh"},
	{0x3567A20BE2072F04ULL, "vxdr_0_deeproads_cs0_0_0.vlm"},
	{0x3567C87BDF517178ULL, "vxlt_gallowstemplar_45u_0_0.vlm"},
	{0x3567D384BB3F0C61ULL, "mud_f_w.dds"},
	{0x3567FF37C509B0F0ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden2.msh"},
	{0x35680CDB5CA3E26CULL, "art\\vfx\\creatures\\fxc_shade_c.anb"},
	{0x35680D37C509C80AULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden2.mmh"},
	{0x3568124151A7560BULL, "art\\vfx\\creatures\\fxc_gol_sla_c.impa.evt"},
	{0x35686EDE10EFBE94ULL, "mag210_arianni6056864.cl"},
	{0x3568E625EBD0B4D5ULL, "vxht_keep_6c7_0_0.vlm"},
	{0x356924239161AF82ULL, "hlac_ht_0_hightown_a2_65534_1y.dds"},
	{0x3569620DACC87AAFULL, "art\\levels\\races\\natural\\highlands\\props\\prp_3craggyrock.phy"},
	{0x356AB28EA6F5FD34ULL, "83644858.ogg"},
	{0x356B0B49BE32C52DULL, "850913531.ogg"},
	{0x356B135AE9B0A0AAULL, "371751785.ogg"},
	{0x356B662F5ED5147AULL, "588729326.ogg"},
	{0x356CA1B801B0E515ULL, "6164310_m.fxe"},
	{0x356CA4697DC64531ULL, "art\\characters\\playercharacter\\elfmale\\arm_nuda.msh"},
	{0x356CB2697DC65D4CULL, "art\\characters\\playercharacter\\elfmale\\arm_nuda.mao"},
	{0x356CBE697DC671EFULL, "art\\characters\\playercharacter\\elfmale\\arm_nuda.mmh"},
	{0x356D35E2C0E0FAA3ULL, "vxw_woundedcoast_tv_0_0.vlm"},
	{0x356D9FCFC4C51048ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_en.gad"},
	{0x356DC4A9F22FB072ULL, "dae000_bartrand6031388.cl"},
	{0x356E7E8A167F0747ULL, "6034655_m.fxe"},
	{0x356EBCF1C49B1E6DULL, "spit_clo.mao"},
	{0x356F015272397FDEULL, "hlac_ht_keep_3741_0x.dds"},
	{0x356F22DB41CF5C24ULL, "vxlt_undercitydungeon_15o_0_0.vlm"},
	{0x356F2B4151AD9FCCULL, "art\\vfx\\creatures\\fxc_gol_sla_c.impa.gad"},
	{0x356FBFEFF95935A1ULL, "water_water_p2.vsh.fxo"},
	{0x356FCCF84D368490ULL, "mcr370_meredith6085938.cl"},
	{0x356FFC56A7063B8FULL, "llac_lt_0_lowtown_a3_n_c_65536_10.dds"},
	{0x35708C46A0AE030EULL, "gen_im_trs_dal_two_2hs_r0.uti"},
	{0x357097A16EE9AF26ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_channel_sa.dds"},
	{0x3570C388BCDC4FB5ULL, "363486126.ogg"},
	{0x35723A80E4844E47ULL, "6167964_m.fxe"},
	{0x3572BFC0D6D58948ULL, "66138073.ogg"},
	{0x3572DF4B5A21B2F8ULL, "437797836.ogg"},
	{0x35735F2405051A8EULL, "7216295.ogg"},
	{0x357362C90C4DE793ULL, "921866063.ogg"},
	{0x357378A6D9787D9EULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingd_d.dds"},
	{0x3573918FED8D1481ULL, "249755580.ogg"},
	{0x3573BD37A6FA19AFULL, "vxht_bartrand_n_5sz_0_0.vlm"},
	{0x3573CAA2234A74A5ULL, "lt_0_lowtown_docks_a3_n_mmap.dds"},
	{0x3574701A9308FB60ULL, "hlac_ht_chantry_2534_21z.dds"},
	{0x35749AE7C326C643ULL, "6161306_m.fxe"},
	{0x3574B74AC5B617D3ULL, "6095430_m.fxe"},
	{0x3574B8F73B9EDB85ULL, "232295343.ogg"},
	{0x357501357C07031CULL, "w_encounter3_28_0.rml"},
	{0x357506DB76B7D963ULL, "6170322_m.fxe"},
	{0x3575927BDC90946BULL, "y_crate_03.dds"},
	{0x3575DEF34F200375ULL, "6076403_f.fxe"},
	{0x357613E8D3FD4784ULL, "vxdr_0_deeproads_ch1_0_0.vlm"},
	{0x35762A4151B3BCCAULL, "art\\vfx\\creatures\\fxc_gol_sla_c.impa.ani"},
	{0x3576A9EEA0DF50D1ULL, "gen_im_npc_arm_hum_glv_mrcb.uti"},
	{0x3576BE549AAC4346ULL, "105460970.ogg"},
	{0x3576CF98816E0852ULL, "zz_dae_debug6029415.cl"},
	{0x357709B44499A020ULL, "hlac_ht_keep_3732_6z.dds"},
	{0x35772323B3600D5EULL, "6064905_m.fxe"},
	{0x3577610537064138ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sis_flemeth_real.gad"},
	{0x35779BF46BFBE877ULL, "391237778.ogg"},
	{0x3577BA4009409946ULL, "gen00fl_dog6165753.cl"},
	{0x3577E7357C090D71ULL, "w_encounter3_28_0.msh"},
	{0x3577F5357C09248CULL, "w_encounter3_28_0.mao"},
	{0x357801357C09392FULL, "w_encounter3_28_0.mmh"},
	{0x35780902E3731B7CULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masf_0t.dds"},
	{0x3578EFA4475D6179ULL, "547588440.ogg"},
	{0x35791EF0645A4975ULL, "plt_cod_ltr_pryce_followup.plo"},
	{0x357935EC2B80D58DULL, "430953933.ogg"},
	{0x357965F58154CD34ULL, "977080898.ogg"},
	{0x3579A9183D450B1BULL, "dae000_flemeth6000968.cl"},
	{0x3579BD1C13D797C7ULL, "6147540_m.fxe"},
	{0x357A20E588D32838ULL, "vxlt_backalley_g_1x4_0_1.vlm"},
	{0x357A62228352F811ULL, "419515266.ogg"},
	{0x357A6DEC234F9036ULL, "504570195.ogg"},
	{0x357A84ED953166E3ULL, "vxlt_warehouse_n_ub_0_0.vlm"},
	{0x357AC9DF23C2C901ULL, "zz_dae_debug6029348.cl"},
	{0x357ACD93A7E3CDB1ULL, "721880395.ogg"},
	{0x357AE939C70E9F3EULL, "tools_sun.phy"},
	{0x357B635A0F23E44FULL, "tv_door_03.dds"},
	{0x357B7677C14A73F3ULL, "vxlt_0_lowtown_n_eqs_0_0.vlm"},
	{0x357BD178889C4A97ULL, "215189807.ogg"},
	{0x357BE0229498DC0FULL, "425997609.ogg"},
	{0x357BE4BECBD2ECF6ULL, "dae000_mother6121046.cl"},
	{0x357C98903F80AB46ULL, "vxca_cavern1_haunted_i25_0_0.vlm"},
	{0x357CBD5573CB2640ULL, "627015658.ogg"},
	{0x357D94B833C574A4ULL, "vxlt_undercitydungeon_18l_0_0.vlm"},
	{0x357E1FEE6793D9D0ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelier.msh"},
	{0x357E2DEE6793F1EAULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelier.mmh"},
	{0x357E31EE6793F821ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelier.mao"},
	{0x357EC4EB877918A3ULL, "vxw_bonepit_1yo_0_0.vlm"},
	{0x357EFFAE8A45C305ULL, "vxht_0_hightown_c7b_0_0.vlm"},
	{0x357F08249B318121ULL, "298671653.ogg"},
	{0x357F31024D38B20BULL, "hlac_lt_0_lowtown_docks_196607_3y.dds"},
	{0x357FA3BB0EA05027ULL, "6102099_m.fxe"},
	{0x35804A3830A904A8ULL, "art\\characters\\playercharacter\\elffemale\\cth_comd.msh"},
	{0x35804EBA52B35715ULL, "dre00st_returnsurface.stg"},
	{0x35805C3830A92339ULL, "art\\characters\\playercharacter\\elffemale\\cth_comd.mao"},
	{0x3580683830A937D2ULL, "art\\characters\\playercharacter\\elffemale\\cth_comd.mmh"},
	{0x35808519A617C595ULL, "6103234_m.fxe"},
	{0x358096A761FF95C7ULL, "zz_one_debug6051516.cl"},
	{0x35814BAEF1DACC22ULL, "w_woundedcoast_47_0.msh"},
	{0x35814DAEF1DACFD8ULL, "w_woundedcoast_47_0.mmh"},
	{0x358159AEF1DAE47BULL, "w_woundedcoast_47_0.mao"},
	{0x358186B86483C9DBULL, "lt_0_lowtown_docks_a3_9_0_ll.mmh"},
	{0x3581E878E29B1F6FULL, "static_static_p0.vsh.fxo.dx11.assembly"},
	{0x358242E0CAAD4CD6ULL, "rob_pl1.tnt"},
	{0x3582B2AE54B3EDFBULL, "gen00fl_anders6125360.cl"},
	{0x3582F67DA8ECE156ULL, "6113874_m.fxe"},
	{0x35830CC4B51CDEC3ULL, "6166206_m.fxe"},
	{0x3583761AEECEB1DFULL, "isa111cr_lucky.utc"},
	{0x3583ABFCCD50A562ULL, "dae011_corff_sta6164632.cl"},
	{0x35841C50138BA2E5ULL, "6078183_m.fxe"},
	{0x3584360E1A26F3A8ULL, "1033514553.ogg"},
	{0x3584365440E0AC12ULL, "6140144_m.fxe"},
	{0x35846CE9F6800952ULL, "6164595_m.fxe"},
	{0x358492AA5416188BULL, "437018710.ogg"},
	{0x358533C39378210EULL, "hightown_male_noble_gossip6171318.cl"},
	{0x35854DAEF1DEB8EBULL, "w_woundedcoast_47_0.rml"},
	{0x3585A9B36C84F7CDULL, "6082120_m.fxe"},
	{0x3585C9DFCFBCE5D5ULL, "vxht_keep_6i5_0_0.vlm"},
	{0x35862FDBC6908E88ULL, "6065904_m.fxe"},
	{0x35863C5E027FF5B3ULL, "hlac_lt_undercityencounter_-1_49z.dds"},
	{0x3586482F1C3D9ECFULL, "vxht_keep_626_0_0.vlm"},
	{0x3586FE385EAE9226ULL, "6149455_m.fxe"},
	{0x358775F004BE3135ULL, "579332004.ogg"},
	{0x3588283A1CC38378ULL, "521981082.ogg"},
	{0x3588CA3AB9FBC227ULL, "6038596_m.fxe"},
	{0x3588F5D11D65AAB5ULL, "hlac_ht_mansion2b_818_58x.dds"},
	{0x3588F6CA87277F63ULL, "vxlt_warehouse_n_xe_0_0.vlm"},
	{0x358967465B884E4BULL, "mrl310_merrill6120360.cl"},
	{0x35896E9FB9123C2AULL, "hlac_lt_0_lowtown_a2_-196609_1z.dds"},
	{0x35898000639DDEF9ULL, "lt_undercityserialkiller_mmap.dds"},
	{0x358A009430795A21ULL, "one300_zevran6089741.cl"},
	{0x358A1FBDC5C6A489ULL, "one220_yevhen6102815.cl"},
	{0x358A3DAD211675DAULL, "666875233.ogg"},
	{0x358AE0C94BFB217AULL, "lt_0_lowtown_docks_a3_n_248_0.msh"},
	{0x358AEEC94BFB38B3ULL, "lt_0_lowtown_docks_a3_n_248_0.mao"},
	{0x358AF2C94BFB3FC0ULL, "lt_0_lowtown_docks_a3_n_248_0.mmh"},
	{0x358AF8BFD0FDFE1BULL, "qun160_vashothlead6064646.cl"},
	{0x358B15974CB2766CULL, "one130_thug6131628.cl"},
	{0x358B31779BE89516ULL, "6121927_m.fxe"},
	{0x358B59543B3F8BE4ULL, "vxlt_backalley_n_1dt_0_0.vlm"},
	{0x358BB4E0F3D24C7CULL, "vxlt_0_lowtown_a2_n_c_gof_0_0.vlm"},
	{0x358BB8E6D27A35D3ULL, "follower_banter6156167.cl"},
	{0x358BDDFB040F3FBAULL, "vxca_cavern1_haunted_gkm_0_1.vlm"},
	{0x358BE001EAC1614FULL, "one_special_modal.ncs"},
	{0x358BEDD3EF1352EEULL, "createitem.ncs"},
	{0x358C38DCF687817EULL, "6143013_m.fxe"},
	{0x358C6753DF87E84FULL, "vxlt_gallowstemplar_n_4y9_0_0.vlm"},
	{0x358CFF6271057847ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_destroyed.phy"},
	{0x358E57B9CFB0BA42ULL, "vxlt_blackemporium_uv_0_0.vlm"},
	{0x358E5BA0A1EE76CDULL, "w_sundermount_147_0.mmh"},
	{0x358E5FA0A1EE7D1EULL, "w_sundermount_147_0.mao"},
	{0x358E6DA0A1EE9553ULL, "w_sundermount_147_0.msh"},
	{0x358E960EFB52E7EDULL, "vxdr_primevaltomb_6um_0_1.vlm"},
	{0x358EF6F6980DDE15ULL, "6104334_m.fxe"},
	{0x358F5EED33F0E672ULL, "zz_qcr_debug6045071.cl"},
	{0x358F67219E86AF9BULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.cess.evt"},
	{0x358F81A4B28E5C24ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww31_slopec.phy"},
	{0x358F937F2531A0CFULL, "mrl210_merrill6166875.cl"},
	{0x358FF1A367D5333FULL, "vxlt_0_lowtown_docks_b67_0_0.vlm"},
	{0x359062FF051FB721ULL, "6161528_m.fxe"},
	{0x3590752CDB49E5EEULL, "6167293_m.fxe"},
	{0x3590AE456DF8D25EULL, "6036713_m.fxe"},
	{0x3591443EDD6F15B8ULL, "644787037.ogg"},
	{0x359158C962F9C1ABULL, "vxlt_0_lowtown_a2_hee_0_0.vlm"},
	{0x35916223BFA31E61ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p2_c.impa.ani"},
	{0x359209D71677F1C5ULL, "6075972_f.fxe"},
	{0x359274076D200707ULL, "105196384.ogg"},
	{0x35928865D11C1F2BULL, "one220_yevhen6103128.cl"},
	{0x3592A1D65A5001E4ULL, "mcr310_argument6127565.cl"},
	{0x359360CDAA8B71CBULL, "zz_mer_debug6073707.cl"},
	{0x3593656FB7A24A2AULL, "qun230cr_guild_contact6068847.cl"},
	{0x359388BDF30295AEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edgebase_1.mao"},
	{0x3593F95D18E9E848ULL, "lt_0_lowtown_docks_a3_n_165_0.mmh"},
	{0x3594055D18E9FCEBULL, "lt_0_lowtown_docks_a3_n_165_0.mao"},
	{0x3594175D18EA1B72ULL, "lt_0_lowtown_docks_a3_n_165_0.msh"},
	{0x359461066D1648F6ULL, "563771326.ogg"},
	{0x3594F0839437159FULL, "vxlt_undercitydungeon_1it_0_0.vlm"},
	{0x359526F75654F259ULL, "gen00fl_anders6119196.cl"},
	{0x359555553310CB2DULL, "6111295_m.fxe"},
	{0x3595701550A4AAA6ULL, "6142555_m.fxe"},
	{0x3595CBD6160A7CA2ULL, "888131342.ogg"},
	{0x3595D9A4504949DEULL, "isa111_isabela6069023.cl"},
	{0x3596200D5BB92305ULL, "vxlt_gallowstemplar_45l_0_0.vlm"},
	{0x35965733FABD2B62ULL, "hlac_lt_gallowstemplar_2393_1z.dds"},
	{0x359660219E8CC2DCULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.cess.gad"},
	{0x35966922906DD6D1ULL, "hlac_ht_mansion2b_818_65x.dds"},
	{0x35968EDD862884CDULL, "art\\levels\\races\\design\\_textures\\prp_chesta_sa.dds"},
	{0x3596E63F34B6FAD7ULL, "one280_emrys6104249.cl"},
	{0x35971358300D7101ULL, "530565450.ogg"},
	{0x359734240B38228EULL, "lgt000pt_baz_sketchy.ncs"},
	{0x359738B835227259ULL, "966076067.ogg"},
	{0x359764945B149CA6ULL, "vxht_keep_n_8ao_0_0.vlm"},
	{0x359802266261355EULL, "657912916.ogg"},
	{0x359809A321BB6E2AULL, "vxlt_gallowscourtyard_7c0_0_0.vlm"},
	{0x35981DD7D60F2665ULL, "mcr370_cut_marriage6098260.cl"},
	{0x35986D23BFA94F03ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p2_c.impa.gad"},
	{0x3598BE4D61D0E642ULL, "6041405_m.fxe"},
	{0x3598F609C624A6FEULL, "hlac_lt_gallowstemplar_n_2383_12x.dds"},
	{0x35991B050CF63F37ULL, "one110_pryce6088791.cl"},
	{0x359920E5E71C4632ULL, "vxw_sundermount_267_0_0.vlm"},
	{0x359953E0CBB2ED5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wi_sit_grnd_lp.gad"},
	{0x35996552552445BBULL, "vxdr_0_deeproads_es9_0_0.vlm"},
	{0x3599745A0F3CFEC1ULL, "tv_door_03.mao"},
	{0x3599A5F70E3CF16DULL, "vxlt_hangedman_n_243_0_0.vlm"},
	{0x3599D4B78868B8E8ULL, "party_med_targeted.mao"},
	{0x359A7E9B3A5D0A2AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwalk.mao"},
	{0x359A829B3A5D1179ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwalk.mmh"},
	{0x359A8C9B3A5D2267ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwalk.msh"},
	{0x359ADE34E46E6E61ULL, "momnt_01.mao"},
	{0x359C5F219E912CDAULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.cess.ani"},
	{0x359CAE318B2B97B7ULL, "vxw_bonepitdestroyed_a3_2ah_0_0.vlm"},
	{0x359CAF9640978ED8ULL, "mcr371_hostile_anders6082112.cl"},
	{0x359CB64FFE047522ULL, "qun200_party_comment6090329.cl"},
	{0x359CD2D01575989CULL, "6173922_m.fxe"},
	{0x359D0E8127FF54DDULL, "821684804.ogg"},
	{0x359D530B0D59F5A6ULL, "gen00fl_dog6167698.cl"},
	{0x359DB360B01EC240ULL, "vxdr_primevalentrance_fu_0_1.vlm"},
	{0x359DB8BCD8222497ULL, "vxht_playershouse_n_1p8_0_1.vlm"},
	{0x359DC88C8740C384ULL, "vxht_chantry_2z1_0_0.vlm"},
	{0x359DF9E1DAA33782ULL, "vxht_keep_4na_0_0.vlm"},
	{0x359E7618AA45F5FBULL, "vxdr_primevaltomb_3j6_0_0.vlm"},
	{0x359EAA3B8F7177B1ULL, "638778772.ogg"},
	{0x359EB73635DFCA8CULL, "841170791.ogg"},
	{0x359EC517ED9BC90BULL, "vxlt_0_lowtown_docks_a3_n_9r8_0_0.vlm"},
	{0x359F4071D4CF30C6ULL, "vxlt_0_lowtown_gff_0_0.vlm"},
	{0x359F4A23BFAF32E8ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p2_c.impa.evt"},
	{0x359F4DAA7DF5C94DULL, "llac_lt_0_lowtown_a2_n_c_131073_0.dds"},
	{0x359F52E0CBB7575AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wi_sit_grnd_lp.ani"},
	{0x359F54A0A1FCDD8AULL, "w_sundermount_147_0.phy"},
	{0x359FCAA654EFDA2BULL, "vxlt_0_lowtown_a2_hbd_0_0.vlm"},
	{0x359FD2E1E72312C5ULL, "vista3.vsh.fxo"},
	{0x35A0B4C5CE5D59FAULL, "str_slimebone_s.dds"},
	{0x35A10900E7B46DECULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_altar_n.dds"},
	{0x35A2455925D2121AULL, "324130094.ogg"},
	{0x35A27CD06FBB89B0ULL, "6173342_m.fxe"},
	{0x35A27D7E751A8F9FULL, "gen_im_acc_rng_act1_06.uti"},
	{0x35A28F7909F45E30ULL, "mag110_vincento6016333.cl"},
	{0x35A2908B84FE77B8ULL, "1034289351.ogg"},
	{0x35A32BE90029CFCFULL, "vxht_keep_680_0_0.vlm"},
	{0x35A3431D55BA9CE5ULL, "hlac_lt_hangedman_-1_19z.dds"},
	{0x35A3629C4A450F5DULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallred.mao"},
	{0x35A3669C4A451596ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallred.mmh"},
	{0x35A3709C4A452694ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallred.msh"},
	{0x35A3A8E336AD9863ULL, "409815541.ogg"},
	{0x35A3BD91F3ECE12EULL, "vxht_keep_5ko_0_0.vlm"},
	{0x35A3E1F83B0F75ECULL, "dae000_arishok6067676.cl"},
	{0x35A4580008E53602ULL, "follower_banter6165597.cl"},
	{0x35A4B6CD14E2ACE9ULL, "vxlt_blackemporium_vh_0_0.vlm"},
	{0x35A4B73CD3077F3AULL, "hlac_lt_0_lowtown_-131071_2y.dds"},
	{0x35A5504ED5CA0C2AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_ric_0n.dds"},
	{0x35A559E81BECF629ULL, "6053309_f.fxe"},
	{0x35A587CD41C6D2DBULL, "art\\vfx\\combat\\vfx_c_mer_attk2_c.impa.ani"},
	{0x35A58B17E44B218EULL, "6060340_f.fxe"},
	{0x35A5AECA83282EF0ULL, "6069874_m.fxe"},
	{0x35A67B8013B186AAULL, "vxlt_gallowscourtyard_7d1_0_0.vlm"},
	{0x35A6BA19B5401E55ULL, "6074370_m.fxe"},
	{0x35A6C522AF03726FULL, "556746026.ogg"},
	{0x35A6D6D3EE4605A8ULL, "vxlt_0_lowtown_docks_a3_n_cq4_0_0.vlm"},
	{0x35A7C134BBDC0E75ULL, "hlac_lt_warehouse_n_65536_0x.dds"},
	{0x35A815EDF0B0DC18ULL, "6066185_m.fxe"},
	{0x35A83D0E1F2BBCE4ULL, "vxlt_backalley_n_1jz_0_0.vlm"},
	{0x35A8688354A6A9F7ULL, "pro000_opening_pc6001200.cl"},
	{0x35A87933EC08196AULL, "zz_mer_debug6062702.cl"},
	{0x35A8BD4C20A188F3ULL, "6096400_m.fxe"},
	{0x35A93DFD73306B2FULL, "w_bonepitmining_a2_66_0.mao"},
	{0x35A949FD73307FCCULL, "w_bonepitmining_a2_66_0.mmh"},
	{0x35A94FFD733089BEULL, "w_bonepitmining_a2_66_0.msh"},
	{0x35A996E5DC5A5498ULL, "hlac_w_encounter3_0_4y.dds"},
	{0x35AA09C35E6C0260ULL, "6054746_m.fxe"},
	{0x35AB9047AB43971BULL, "zz_dre_debug6069198.cl"},
	{0x35ABC33CA7CB56FCULL, "gen00fl_aveline6106923.cl"},
	{0x35ABC3E5BB96C7FEULL, "lgt260_way_blood_mage6148381.cl"},
	{0x35AC5960C757F234ULL, "follower_banter6171266.cl"},
	{0x35AC5985702932BDULL, "dae000_gen_player_turn_in6177849.cl"},
	{0x35AC94153FD09DB2ULL, "6113351_m.fxe"},
	{0x35AD61104D058DCCULL, "6174581_m.fxe"},
	{0x35AD8581E63B2211ULL, "plt_vrc200pt_n_fm_complete.plo"},
	{0x35ADA6B21913D54DULL, "6051072_m.fxe"},
	{0x35ADC7D38D5965D9ULL, "hlac_w_woundedcoast_-131072_0y.dds"},
	{0x35ADDC8E05F14E3FULL, "6142837_m.fxe"},
	{0x35ADE6692B75C2AFULL, "mag200_feynriel6056921.cl"},
	{0x35ADFDDB96B2CAEFULL, "650454850.ogg"},
	{0x35AE05A900277298ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_corruptlyrium_s.dds"},
	{0x35AE095C82CA7E69ULL, "ser210cr_witness.utc"},
	{0x35AE4581B4CFB955ULL, "6131176_m.fxe"},
	{0x35AEC5C8B8D8BCA4ULL, "zz_followers_banter6157549.cl"},
	{0x35AF424CFEDA3AEFULL, "325996924.ogg"},
	{0x35AF486D110F4990ULL, "gen00fl_merrill6128683.cl"},
	{0x35AFD4226EB1608EULL, "mag100_cullen6093804.cl"},
	{0x35B03B8BB734CE92ULL, "6102902_m.fxe"},
	{0x35B07AC642D12A0EULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.impa.ani"},
	{0x35B0B5B3404FE14DULL, "gen00fl_sebastian6179986.cl"},
	{0x35B38FCD41D31DD2ULL, "art\\vfx\\combat\\vfx_c_mer_attk2_c.impa.evt"},
	{0x35B45EF614F804C9ULL, "vxlt_hangedman_tj_0_0.vlm"},
	{0x35B47AC30EBEF4E8ULL, "plt_gen00pt_skills.plo"},
	{0x35B4BF7CEFEF25CFULL, "767612765.ogg"},
	{0x35B4DFB8578DE75AULL, "186868601.ogg"},
	{0x35B4ED5D05A79F2AULL, "vxlt_gallowscourtyard_7i2_0_0.vlm"},
	{0x35B504BCBFDC7EEDULL, "w_bonepitmining_a2_123_0.phy"},
	{0x35B511091A0EA912ULL, "vxw_encounter3_45_0_0.vlm"},
	{0x35B52BF6A73636D5ULL, "6079070_m.fxe"},
	{0x35B548B0E03C1E28ULL, "vxlt_0_lowtown_docks_a3_n_cz5_0_0.vlm"},
	{0x35B611F0648DE152ULL, "plt_cod_ltr_pryce_followup.nss"},
	{0x35B632637E717614ULL, "6146997_m.fxe"},
	{0x35B63EED81E98540ULL, "w_encounter2_43_0_ll.mmh"},
	{0x35B6730358B2D817ULL, "art\\characters\\creatures\\ragedemon\\animation\\tn90r.gad"},
	{0x35B76C09F75C061BULL, "vxlt_0_lowtown_a2_n_ei7_0_0.vlm"},
	{0x35B7E0B84776F4D8ULL, "vxlt_warehouse_n_qa_0_0.vlm"},
	{0x35B7F9D8F6DFCAD8ULL, "746659306.ogg"},
	{0x35B88C5F4F517AA3ULL, "vxw_bonepit_1ms_0_0.vlm"},
	{0x35B8B4CFA4668FB9ULL, "vxht_bartrand_n_64z_0_0.vlm"},
	{0x35B928EFD4B982FEULL, "hlac_dr_slavers_3217_10z.dds"},
	{0x35B93AB1696FA1CAULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.impa.gad"},
	{0x35B940349F34575AULL, "vxlt_gallowstemplar_f_6sl_0_0.vlm"},
	{0x35B98B33711A2102ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.dura.gad"},
	{0x35B99D1F85CC756CULL, "hlac_lt_0_lowtown_a2_-4_0z.dds"},
	{0x35B9CDD3591C3BCDULL, "6137308_f.fxe"},
	{0x35B9EA4EBCED0406ULL, "vxca_caverns2prc_8kt_0_0.vlm"},
	{0x35BA1F25FEF2CCEEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sledge_sa.dds"},
	{0x35BA1FFD3C845CEEULL, "221934317.ogg"},
	{0x35BA2116EAD546D4ULL, "follower_banter6171532.cl"},
	{0x35BA77C071BB903BULL, "hlac_ht_mansion2_818_6y.dds"},
	{0x35BA8420DBCE8F5AULL, "286278168.ogg"},
	{0x35BAAC466B2CF484ULL, "vxht_chantry_2p7_0_0.vlm"},
	{0x35BAAECD41D97165ULL, "art\\vfx\\combat\\vfx_c_mer_attk2_c.impa.gad"},
	{0x35BC1555034D2ABBULL, "dae000ip_crowd_lm4_talk12.utp"},
	{0x35BC242BB8BB61C6ULL, "vxlt_0_lowtown_ghh_0_0.vlm"},
	{0x35BC8B8C417AAC73ULL, "vxw_bonepitmining_a2_p1_0_0.vlm"},
	{0x35BD5E00C7638C50ULL, "603497485.ogg"},
	{0x35BD72C642DBA75FULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.impa.evt"},
	{0x35BDE31DCCC00AC0ULL, "hlac_lt_0_lowtown_a2_-262144_2y.dds"},
	{0x35BE02F9FEAC629FULL, "vxlt_ruinedhouse_n_b9_0_0.vlm"},
	{0x35BE5265E2BCD775ULL, "248465613.ogg"},
	{0x35BE78D374BF34BEULL, "qcr000_aveline6054356.cl"},
	{0x35BE90B10EB6A301ULL, "vxlt_darktown_5pr_0_0.vlm"},
	{0x35BEB9D42AD17909ULL, "vxlt_undercityencounter_12f_0_0.vlm"},
	{0x35BEDFF47F2D4ACEULL, "hlac_lt_0_lowtown_a2_n_0_5y.dds"},
	{0x35BF2F8C6331F502ULL, "hlac_ca_caverns2_5802_32x.dds"},
	{0x35BFE1571E5995FEULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.dura.evt"},
	{0x35C004A49B2EAFF6ULL, "vrc321_party_barks6095772.cl"},
	{0x35C01DB169758F61ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.impa.evt"},
	{0x35C0A4519ED83AD9ULL, "681177933.ogg"},
	{0x35C0AE3371207BF9ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.dura.evt"},
	{0x35C0DA0E94E36552ULL, "vxht_0_hightown_a3_b75_0_0.vlm"},
	{0x35C115E37D17191BULL, "431765577.ogg"},
	{0x35C13B7BC95B3340ULL, "vxlt_gallowsprison_n_6lw_0_0.vlm"},
	{0x35C1E33F443815DCULL, "6143086_m.fxe"},
	{0x35C1EA10BDD8947DULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallgreen_0.pwk"},
	{0x35C382E60C04C192ULL, "vxw_encounter3_16_0_0.vlm"},
	{0x35C38E146324826DULL, "vxht_bartrand_n_4h8_0_0.vlm"},
	{0x35C3ECAFEC3FF087ULL, "6116542_m.fxe"},
	{0x35C44E5A45CF1803ULL, "vxw_bonepitdestroyed_a3_bq_0_0.vlm"},
	{0x35C453C642E19148ULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.impa.gad"},
	{0x35C4CB664BB09B8BULL, "hlac_lt_0_lowtown_65536_7x.dds"},
	{0x35C519895F96C0D8ULL, "44521970.ogg"},
	{0x35C51C8C5016849EULL, "6115372_f.fxe"},
	{0x35C533B2F8BFAE4AULL, "gen00fl_anders6125851.cl"},
	{0x35C57ABABA21E5F1ULL, "rog_markofdeath.dds"},
	{0x35C588898791BDA6ULL, "hlac_ht_chantry_n_2534_44x.dds"},
	{0x35C5D99720DF2885ULL, "6117445_f.fxe"},
	{0x35C5DDE6E9521E9BULL, "vxlt_0_lowtown_a2_n_er6_0_0.vlm"},
	{0x35C6C7FEC75BD87EULL, "gen00fl_fenris6126834.cl"},
	{0x35C6D95BEE2BF31AULL, "hm_gamlen_h1s2.mor"},
	{0x35C6D9DE802CB970ULL, "702409561.ogg"},
	{0x35C6F8571E5FDB29ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.dura.gad"},
	{0x35C72C36BE319E5BULL, "w_bonepitmining_a2_89_0.msh"},
	{0x35C73E36BE31BCA6ULL, "w_bonepitmining_a2_89_0.mao"},
	{0x35C74A36BE31D145ULL, "w_bonepitmining_a2_89_0.mmh"},
	{0x35C74AE026A790ABULL, "vxdr_primevalentrance_2q6_0_0.vlm"},
	{0x35C768A69BA5C640ULL, "hlac_lt_0_lowtown_a2_n_-65536_2z.dds"},
	{0x35C7A70308110D25ULL, "lt_gallowscourtyard_ll.mmh"},
	{0x35C7EF4DCD2E489DULL, "ave000im_gft_shield.uti"},
	{0x35C81F74A74BDB27ULL, "hlac_ca_caverns2_5802_65y.dds"},
	{0x35C846554B90B257ULL, "gen_im_arm_cht_lgt_act2_02.uti"},
	{0x35C86117B1E767B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent.gad"},
	{0x35C9558C07A4674DULL, "hlac_lt_blackemporium_1055_14z.dds"},
	{0x35C98C10F1BAB006ULL, "445192447.ogg"},
	{0x35C9A17AEA19B8A6ULL, "vxlt_ruinedhouse_n_gs_0_0.vlm"},
	{0x35C9FC576003E4D4ULL, "6073785_m.fxe"},
	{0x35CA1813693CDB12ULL, "532950572.ogg"},
	{0x35CAA00358C3CD9DULL, "art\\characters\\creatures\\ragedemon\\animation\\tn90r.ani"},
	{0x35CAB4CBDFCA1058ULL, "gen_im_npc_wep_two_quna.uti"},
	{0x35CB87CC62595CD8ULL, "6067649_m.fxe"},
	{0x35CCA6E845967527ULL, "6144476_f.fxe"},
	{0x35CD73E145452CD6ULL, "gen00fl_mrl_act3_pt2.stg"},
	{0x35CDB6B8297907C3ULL, "vxlt_0_lowtown_a2_n_c_euy_0_0.vlm"},
	{0x35CDC9571E65AA77ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.dura.ani"},
	{0x35CE15B16981BEB0ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.impa.ani"},
	{0x35CE457EC8965FFBULL, "llac_dr_primevalentrance_1725_0.dds"},
	{0x35CE66C12B7F8F18ULL, "mag320_orsino6071252.cl"},
	{0x35CEB633712CC6E8ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.dura.ani"},
	{0x35CF4BEB86D97DD2ULL, "vxht_0_hightown_a3_b26_0_0.vlm"},
	{0x35CF6417B1ED8BE9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent.evt"},
	{0x35CF681DEE10829FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sidel.phy"},
	{0x35CF7B418F242BD7ULL, "52369057.ogg"},
	{0x35CF7F152694AA8DULL, "vxlt_gallowscourtyard_6xg_0_0.vlm"},
	{0x35CFAD58BB514BC0ULL, "vxlt_gallowsprison_n_6it_0_0.vlm"},
	{0x35CFE400E7561FD9ULL, "6086888_f.fxe"},
	{0x35CFE7534C5A1375ULL, "6148033_m.fxe"},
	{0x35CFF5B62F92DD65ULL, "251213119.ogg"},
	{0x35D00E188981BDBFULL, "232824570.ogg"},
	{0x35D0152BF445D7ACULL, "character_character_p16.psh.fxo.dx11.assembly"},
	{0x35D017AD3B84A4E3ULL, "6080479_f.fxe"},
	{0x35D12842814FEA2DULL, "lgt101_mag_escp_terrie.bnk"},
	{0x35D12D53C749B4D1ULL, "6180449_f.fxe"},
	{0x35D18EF3C6E18202ULL, "hlac_lt_0_lowtown_docks_n_-196608_1y.dds"},
	{0x35D249F34F90A7C9ULL, "mag370_meredith6171558.cl"},
	{0x35D2BEA8958ED515ULL, "130010509.ogg"},
	{0x35D2C03737C53083ULL, "vxw_bonepitdestroyed_a3_mp_0_0.vlm"},
	{0x35D34264BE9098BAULL, "vxdr_primevalentrance_t4_0_0.vlm"},
	{0x35D40B0985D8B4C8ULL, "lla_ht_0_hightown_n_65534_0.dds"},
	{0x35D447ABD51732A3ULL, "dre120pt_reqs_maps.ncs"},
	{0x35D44A4281525995ULL, "lgt101_mag_escp_terrie.cnv"},
	{0x35D45A6471338776ULL, "976695817.ogg"},
	{0x35D497DCDA37239AULL, "6135368_m.fxe"},
	{0x35D4E28BB61E48DFULL, "llac_w_encounter2_camp_327681_0.dds"},
	{0x35D4EF0E1870DD60ULL, "vxht_0_hightown_n_c4y_0_0.vlm"},
	{0x35D5043977F886B6ULL, "hlac_dr_primevaltomb_658_10z.dds"},
	{0x35D5A74B61484DF5ULL, "llac_ht_mansion_n_3365_0.dds"},
	{0x35D5DD7CA5191580ULL, "mag100_cullen6140971.cl"},
	{0x35D62F7D5C7B2669ULL, "hlac_lt_darktown_-1_31y.dds"},
	{0x35D6CD2331049E5CULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_masa_0n.dds"},
	{0x35D6D395F0A39FC4ULL, "116213943.ogg"},
	{0x35D72A2DBC0D5120ULL, "197891568.ogg"},
	{0x35D72DB436D23917ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr.phy"},
	{0x35D7D5335F5AE05BULL, "dre000st_varric_wayarounda.stg"},
	{0x35D8C7E80B94A95CULL, "mrl310_merrill6144590.cl"},
	{0x35D8F22C0827A89FULL, "dae000_merchant_generic_hm6172648.cl"},
	{0x35D98D6D08FF9BFFULL, "zz_mrl_debug6184206.cl"},
	{0x35D9DEEE9169F423ULL, "vxca_caverns1_edw_0_0.vlm"},
	{0x35DA0085BB026D5DULL, "gen000tr_lothering_burn.utt"},
	{0x35DA0611BC0C1398ULL, "zz_dae_debug6029912.cl"},
	{0x35DA1BEE9A93244DULL, "zz_one_debug6077392.cl"},
	{0x35DA5B256D54B58BULL, "mrl000_marethari6069872.cl"},
	{0x35DA7650C750113BULL, "575359984.ogg"},
	{0x35DA7FE7978D8033ULL, "473072703.ogg"},
	{0x35DB9D4D54706276ULL, "vxw_woundedcoast_16l_0_0.vlm"},
	{0x35DBBD2258569F31ULL, "zz_vault_debug6200378.cl"},
	{0x35DBC45E17FFF681ULL, "324927992.ogg"},
	{0x35DBEF70F70534AAULL, "vxw_bonepitdestroyed_a3_1p5_0_0.vlm"},
	{0x35DC2D91047E5433ULL, "vxht_0_hightown_a3_c3u_0_0.vlm"},
	{0x35DC4C17B1F7ED58ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ent.ani"},
	{0x35DC673830F794CDULL, "art\\characters\\playercharacter\\elffemale\\cth_comd.phy"},
	{0x35DC8B2F3DCBE7C1ULL, "character_character_p118.psh.fxo.dx11"},
	{0x35DD1C9C12AADDA5ULL, "hlac_lt_darktown_-1_21x.dds"},
	{0x35DD829E62874FE0ULL, "6044178_m.fxe"},
	{0x35DD9C9EAAB29734ULL, "and121_karl6093899.cl"},
	{0x35DDE3F37367FDA7ULL, "follower_banter6163178.cl"},
	{0x35DEC7A7642A8F12ULL, "vxlt_0_lowtown_a2_ev4_0_0.vlm"},
	{0x35DF174D423BF1B2ULL, "6049460_f.fxe"},
	{0x35DF234440954509ULL, "hlac_w_encounter1_65539_0z.dds"},
	{0x35DF44ECED59AC67ULL, "vxht_keep_671_0_0.vlm"},
	{0x35DF6D6CA490E5D5ULL, "697004191.ogg"},
	{0x35DF8829F0B619D1ULL, "vxca_caverns1_mines_ehd_0_0.vlm"},
	{0x35DF9527D0DA292BULL, "art\\vfx\\textures\\vfx_distortion_ball_alp.dds"},
	{0x35E04F59085B7915ULL, "540366957.ogg"},
	{0x35E0691DEE1EF7A4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sidel.mmh"},
	{0x35E06D1DEE1EFEF7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sidel.mao"},
	{0x35E07F1DEE1F1D46ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sidel.msh"},
	{0x35E0BFF6405FF586ULL, "6106380_m.fxe"},
	{0x35E0D6A7DD278A1CULL, "lgt000ip_seal_crate.utp"},
	{0x35E1531BC5F22262ULL, "lt_0_lowtown_docks_a3_122_0.msh"},
	{0x35E1551BC5F22518ULL, "lt_0_lowtown_docks_a3_122_0.mmh"},
	{0x35E1611BC5F239BBULL, "lt_0_lowtown_docks_a3_122_0.mao"},
	{0x35E177F23CCA2783ULL, "6084021_m.fxe"},
	{0x35E19AD574CBBE37ULL, "lt_0_lowtown_docks_n_9_0.anb"},
	{0x35E19EAD7179E130ULL, "242244323.ogg"},
	{0x35E1C27F22F9DC8CULL, "293927792.ogg"},
	{0x35E309B9CC2D3C1AULL, "6130068_m.fxe"},
	{0x35E311C6507CA623ULL, "vxca_cavern1_haunted_jf7_0_0.vlm"},
	{0x35E34D47EC50169EULL, "551438873.ogg"},
	{0x35E360EB0A66F5E0ULL, "vxht_0_hightown_n_c1f_0_0.vlm"},
	{0x35E397538206ABA6ULL, "27023019.ogg"},
	{0x35E415E6ACD031B4ULL, "vxlt_shop_ik_0_0.vlm"},
	{0x35E496FBEC3F0F0FULL, "6132658_m.fxe"},
	{0x35E4A67D7CAB6634ULL, "139257212.ogg"},
	{0x35E4B2213D03FAF7ULL, "711464358.ogg"},
	{0x35E4ED789A272F7CULL, "one100_solivitus6048123.cl"},
	{0x35E509838BD5E72EULL, "715243303.ogg"},
	{0x35E5A644C8DD3735ULL, "follower_banter6172880.cl"},
	{0x35E5F8C049A8C08BULL, "814449502.ogg"},
	{0x35E60731D01F384BULL, "lgt160_ntb_elf_conflict6163362.cl"},
	{0x35E68FC06CFD44CCULL, "lt_gallowscourtyard_35_0_lc.mmh"},
	{0x35E69E77033D76EBULL, "cod_bks_lite_tombs_last_3.utp"},
	{0x35E75310DB7AA925ULL, "qun160_vashothlead6055572.cl"},
	{0x35E7CBAD58F09FB6ULL, "vxlt_gallowstemplar_n_5ut_0_1.vlm"},
	{0x35E7E4037E8C9558ULL, "vxlt_darktown_4rt_0_0.vlm"},
	{0x35E7F0F3D38139AEULL, "vxht_keep_n_4i3_0_0.vlm"},
	{0x35E81468FF9D5005ULL, "vxht_brothel_n_5g4_0_0.vlm"},
	{0x35E8A5BB9C8841ABULL, "hlac_lt_0_lowtown_docks_-131075_1z.dds"},
	{0x35E8E970F8421B79ULL, "6110955_m.fxe"},
	{0x35E8F0D528551AC0ULL, "vxht_mansion_n_4vt_0_0.vlm"},
	{0x35E96ADB741C3CCFULL, "vxlt_gallowscourtyard_n_7gf_0_0.vlm"},
	{0x35EA0F2A46667AF6ULL, "vxw_woundedcoast_11m_0_0.vlm"},
	{0x35EA614DE8FB4D2AULL, "vxw_bonepitdestroyed_a3_1u6_0_0.vlm"},
	{0x35EA9F6DF6746CB3ULL, "vxht_0_hightown_a3_c4t_0_0.vlm"},
	{0x35EB424FEB8E71A8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wi_basic.msh"},
	{0x35EB46317F3C376BULL, "gen00fl_aveline6116628.cl"},
	{0x35EB544FEB8E9039ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wi_basic.mao"},
	{0x35EB5C1BC5FA85A3ULL, "lt_0_lowtown_docks_a3_122_0.phy"},
	{0x35EB604FEB8EA4D2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wi_basic.mmh"},
	{0x35EB642B723D05C8ULL, "hlac_lt_0_lowtown_n_65537_7y.dds"},
	{0x35EC32B415BF0FE3ULL, "ave100_donnic6106001.cl"},
	{0x35EC813314B646CEULL, "hlac_lt_foundry_n_320_6y.dds"},
	{0x35ECAFE9E08BCCDDULL, "mcr310_argument6156719.cl"},
	{0x35EDA3929DFAECC9ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_4z.dds"},
	{0x35EDB6C9DF4FC4E7ULL, "vxht_keep_600_0_0.vlm"},
	{0x35EE7018A9C70B39ULL, "6139294_f.fxe"},
	{0x35EE98ACC9FA2BA4ULL, "6082231_m.fxe"},
	{0x35EECAB103BDC956ULL, "6165145_m.fxe"},
	{0x35EED9FD162E4B23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_rf.evt"},
	{0x35EF88F83D378B1EULL, "100571989.ogg"},
	{0x35EFF45A11B0CCFDULL, "gen_im_trs_rdr_junk2_02.uti"},
	{0x35F0195D72F049C8ULL, "6074335_m.fxe"},
	{0x35F02B1D74571D64ULL, "134665933.ogg"},
	{0x35F02DB757F78A09ULL, "hlac_lt_gallowstemplar_7198_6z.dds"},
	{0x35F04457A99CD2C1ULL, "vxw_bonepit_1s7_0_0.vlm"},
	{0x35F06A976C3F8CA6ULL, "vxca_caverns1_icm_0_2.vlm"},
	{0x35F07066ABACB4F4ULL, "hlac_lt_gallowscourtyard_n_-65537_1y.dds"},
	{0x35F07E953AFFE79FULL, "225095439.ogg"},
	{0x35F090273AFFE7ACULL, "vxw_blightlands_d7_0_0.vlm"},
	{0x35F0FD3B483FE3AEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headhigh.evt"},
	{0x35F10910DEBB0265ULL, "6088409_m.fxe"},
	{0x35F155656FEF5BA5ULL, "pro000_colonel_ewald6047378.cl"},
	{0x35F19FA4BE30BA83ULL, "600751338.ogg"},
	{0x35F2C1531F3621DAULL, "6018421_m.fxe"},
	{0x35F2FB08C298B6EEULL, "261697953.ogg"},
	{0x35F45C8CE410E104ULL, "vxca_cavern1_haunted_gkc_0_1.vlm"},
	{0x35F494DCB350059EULL, "vxlt_gallowstemplar_f_6t1_0_0.vlm"},
	{0x35F4B43B5756DDCAULL, "140238561.ogg"},
	{0x35F4F06B9ABF9F0CULL, "6050176_f.fxe"},
	{0x35F50C5DF3A83B1EULL, "dae000cr_grand_cleric6173598.cl"},
	{0x35F533D4371DB37AULL, "follower_banter6163726.cl"},
	{0x35F53926455E99BCULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_wng_0d.dds"},
	{0x35F54337F4267D19ULL, "6119387_m.fxe"},
	{0x35F597D49454379CULL, "784167991.ogg"},
	{0x35F5B2FD163427A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_rf.gad"},
	{0x35F5DD787D446024ULL, "vxlt_0_lowtown_a3_n_c_hen_0_0.vlm"},
	{0x35F61697A605A01CULL, "vxht_brothel_n_3eu_0_0.vlm"},
	{0x35F61FDDB41AE6FEULL, "vxw_encounter1_j8_0_1.vlm"},
	{0x35F6530009530E73ULL, "vxw_bonepitmining_a2_l5_0_0.vlm"},
	{0x35F6669771953A32ULL, "pro000_athenril6048720.cl"},
	{0x35F68645F1936885ULL, "vxht_brothel_n_5j3_0_0.vlm"},
	{0x35F6F027FAC1B515ULL, "vxlt_gallowsprison_n_6pz_0_0.vlm"},
	{0x35F798A7720DDF74ULL, "zz_approval_debug6003768.cl"},
	{0x35F7CA2B502A2DFFULL, "6176074_m.fxe"},
	{0x35F7F43B4845F359ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headhigh.gad"},
	{0x35F87A2381929761ULL, "6005778_m.fxe"},
	{0x35F8A1F7FDA82B1DULL, "373902117.ogg"},
	{0x35F90FC75C2A8218ULL, "vxlt_0_lowtown_ex4_0_0.vlm"},
	{0x35F922573FF66B85ULL, "hightown_fem_elfservant_mob6169044.cl"},
	{0x35FA1D661F56D9BCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robb_0n.dds"},
	{0x35FA3C7445CF46AAULL, "vxca_cavern1_haunted_i29_0_0.vlm"},
	{0x35FA3E704C3C8EBDULL, "466878226.ogg"},
	{0x35FB4AB68E9BB368ULL, "6198486_m.fxe"},
	{0x35FB4EBAAD54D953ULL, "dae000_merchant_tailor_hm6164346.cl"},
	{0x35FB553181263D38ULL, "423938441.ogg"},
	{0x35FB7BBE05968602ULL, "weapontrail_normal.dds"},
	{0x35FBC5301F4215CEULL, "959591660.ogg"},
	{0x35FC056DC976BB90ULL, "llac_lt_0_lowtown_n_-1_2.dds"},
	{0x35FC2B85FA5D75A8ULL, "gen_im_npc_wep_shd_crta.uti"},
	{0x35FC6BE3D4A24AD1ULL, "vxca_caverns1_mines_ebz_0_0.vlm"},
	{0x35FC8AC593078534ULL, "6149196_m.fxe"},
	{0x35FCC1FD163A5F82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_rf.ani"},
	{0x35FD839B3AB19066ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwalk.phy"},
	{0x35FDA4061960B5C4ULL, "6136882_m.fxe"},
	{0x35FF02042CF6002CULL, "vxw_blightlands_i4_0_0.vlm"},
	{0x35FF053B484C2EA7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headhigh.ani"},
	{0x35FFCEDE83B18E1BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masf_0s.dds"},
	{0x3600831D63172F16ULL, "6080644_m.fxe"},
	{0x3600C2E67CC40954ULL, "character_character_p197.psh.fxo.dx11"},
	{0x3601883C859F3825ULL, "hlac_lt_0_lowtown_a2_-65536_4z.dds"},
	{0x36019ABAA5B99232ULL, "vxlt_0_lowtown_a2_n_c_d3s_0_0.vlm"},
	{0x3601B8E4698544ACULL, "vxlt_0_lowtown_a3_n_c_byd_0_0.vlm"},
	{0x360211E7A5913066ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.impa.evt"},
	{0x360235C97158B78CULL, "856938889.ogg"},
	{0x360263F3FD2277E2ULL, "w_bonepit_181_0.msh"},
	{0x360265F3FD227A98ULL, "w_bonepit_181_0.mmh"},
	{0x360271F3FD228F3BULL, "w_bonepit_181_0.mao"},
	{0x360272352C8F9C1BULL, "gen00fl_isabela_init6110262.cl"},
	{0x36029D5BAC1B40C5ULL, "vxca_caverns1_flz_0_0.vlm"},
	{0x3603C23D20C55B73ULL, "576931990.ogg"},
	{0x3603CDD85577CFC9ULL, "lgt110_gxa_coa_spy6117456.cl"},
	{0x360428D877213A27ULL, "6043918_m.fxe"},
	{0x3604C4DCFB4926F3ULL, "vxw_bonepitmining_a2_k4_0_0.vlm"},
	{0x3605340BA8403AFEULL, "vxdr_primevaltomb_9a1_0_1.vlm"},
	{0x3605C1EF2039F5A7ULL, "hlac_dr_slavers_3215_4x.dds"},
	{0x36065EE13CB46F6BULL, "vxlt_0_lowtown_a2_n_euc_0_0.vlm"},
	{0x36069081E6875E76ULL, "plt_vrc200pt_n_fm_complete.nss"},
	{0x360690A1CE826BABULL, "27699790.ogg"},
	{0x3606C74FFC1F4BF3ULL, "546457398.ogg"},
	{0x3606E40100385446ULL, "art\\levels\\races\\natural\\highlands\\vista\\_textures\\prp_occluder_d.dds"},
	{0x3607119A915F1DF2ULL, "mag131_taraneh6088734.cl"},
	{0x360736C5E7CFA52EULL, "lt_0_lowtown_a2_n_29_0.rml"},
	{0x360781A44E209A98ULL, "vxlt_0_lowtown_es5_0_0.vlm"},
	{0x3608BA5D19A66297ULL, "mrl000_marethari6205612.cl"},
	{0x360928E7A59775B1ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.impa.gad"},
	{0x36093DFBEE198B59ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_lavaspillxa.msh"},
	{0x360947FBEE199C47ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_lavaspillxa.mmh"},
	{0x36094BFBEE19A294ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_lavaspillxa.mao"},
	{0x36099939F6219810ULL, "follower_banter6154201.cl"},
	{0x3609AAB9941BA520ULL, "gen00fl_anders6110057.cl"},
	{0x360A73DE1A6A7584ULL, "vxlt_0_lowtown_a2_gbl_0_0.vlm"},
	{0x360ADDC0C6986351ULL, "vxca_caverns1_mines_egy_0_0.vlm"},
	{0x360BBD764BED6C31ULL, "hlac_lt_0_lowtown_a2_n_65534_5y.dds"},
	{0x360BF942C5EB0754ULL, "vxht_0_hightown_a3_aol_0_0.vlm"},
	{0x360C2F06D08B0984ULL, "sib131_slave_a6165267.cl"},
	{0x360C389C02562288ULL, "vxlt_undercityserialkiller_18q_0_0.vlm"},
	{0x360C7235C642FBBAULL, "6080208_m.fxe"},
	{0x360C905751D1852BULL, "llac_w_blightlands_0_8.dds"},
	{0x360CB5B30425A86CULL, "vfxchunky0_voltex.vsh.fxo"},
	{0x360D0ADC01E9CC40ULL, "6167273_m.fxe"},
	{0x360D59F63A9C7521ULL, "811594607.ogg"},
	{0x360D70081419456FULL, "mag101_grace6034233.cl"},
	{0x360D73E11EEC18ACULL, "vxw_blightlands_r5_0_0.vlm"},
	{0x360D97D574F0F44BULL, "lt_0_lowtown_docks_n_9_0.rml"},
	{0x360DE075401DCD4CULL, "hlac_lt_0_lowtown_docks_a3_-196609_0y.dds"},
	{0x360F30F338FB4D1BULL, "6093681_m.fxe"},
	{0x360FB03677F397E5ULL, "vxw_bonepitdestroyed_a3_1xg_0_1.vlm"},
	{0x360FCF3AD3C83576ULL, "art\\vfx\\combat\\vfx_w_wns_left_close1_c.impa.ani"},
	{0x361029E7A59D965FULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.impa.ani"},
	{0x361058457B7F4FFBULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb06a.msh"},
	{0x361076457B7F82E5ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb06a.mmh"},
	{0x36109740C187B085ULL, "6078126_f.fxe"},
	{0x3610982BD8F2AF6BULL, "vxca_caverns1_mines_i9y_0_0.vlm"},
	{0x3610DFD1F9712CBCULL, "640841513.ogg"},
	{0x361111C5C72BE5EFULL, "548861599.ogg"},
	{0x3611126453660917ULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyn.phy"},
	{0x36114BFC2D72DF03ULL, "vxht_chantry_353_0_1.vlm"},
	{0x361187E18106C47DULL, "hlac_lt_hangedman_n_-1_1x.dds"},
	{0x3611D3C40A28BA81ULL, "150773202.ogg"},
	{0x3612C13261309124ULL, "vxlt_0_lowtown_a3_n_c_hol_0_0.vlm"},
	{0x361306A3221F4A98ULL, "vxlt_0_lowtown_h1z_0_0.vlm"},
	{0x36133BE2B465A3AAULL, "475199227.ogg"},
	{0x36139E5CD1247640ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_big.phy"},
	{0x36140F914479899DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_diagonal_1.mao"},
	{0x3614466BFE376440ULL, "vxht_mansion_n_4gs_0_0.vlm"},
	{0x361481F70FF219B0ULL, "hlac_lt_0_lowtown_a2_n_65536_4y.dds"},
	{0x3614AA8F40A47DD6ULL, "293266262.ogg"},
	{0x3614F8AEE7928C5FULL, "6051613_f.fxe"},
	{0x3615117C9A160F98ULL, "336886723.ogg"},
	{0x3615384E3CE36E5DULL, "fen210_fenris6109399.cl"},
	{0x361560C19960BBC5ULL, "ca_caverns1_mines_mmi.gff"},
	{0x36159F622E649561ULL, "6118793_m.fxe"},
	{0x3615B6E4BEDD96AAULL, "vxw_bonepitdestroyed_a3_1d1_0_0.vlm"},
	{0x3615BF673D605D7FULL, "vxht_0_hightown_n_8pu_0_0.vlm"},
	{0x3615C0F40CD721D2ULL, "vxca_caverns1_mines_dr0_0_0.vlm"},
	{0x3615F3814016B318ULL, "vxlt_0_lowtown_en2_0_0.vlm"},
	{0x3616210F02906E94ULL, "art\\characters\\playercharacter\\humanmale\\hoo_noba.phy"},
	{0x36165DEBE3C7DFD4ULL, "vxlt_gallowstemplar_4uy_0_0.vlm"},
	{0x3616B0BFAE9825A0ULL, "999918481.ogg"},
	{0x36177F62127C4E72ULL, "vxca_caverns1_mines_e8k_0_0.vlm"},
	{0x3617DC4853603C48ULL, "w_bonepit_145_0.mmh"},
	{0x3617E848536050EBULL, "w_bonepit_145_0.mao"},
	{0x3617FA4853606F72ULL, "w_bonepit_145_0.msh"},
	{0x361875E96FF431DFULL, "llac_lt_undercitydungeon_-1_37.dds"},
	{0x3618A756DF21FEE1ULL, "2272251.ogg"},
	{0x3618FD4F2DF5505DULL, "889887881.ogg"},
	{0x3619A710855C872BULL, "vxdr_primevaltomb_3jf_0_0.vlm"},
	{0x3619CCAAE0DF224EULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathv1.evt"},
	{0x361A3FF6E87C4D24ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.impa.ani"},
	{0x361A6B1FB7E11FD4ULL, "vxht_0_hightown_a3_atm_0_0.vlm"},
	{0x361B21AA68C322B2ULL, "vxw_sundermount_2c8_0_0.vlm"},
	{0x361B7A6EEDDA739CULL, "hlac_lt_0_lowtown_a2_n_-131072_9x.dds"},
	{0x361B8713228E0BACULL, "135680942.ogg"},
	{0x361BAB20CFAC199FULL, "vxlt_0_lowtown_a3_n_c_h1u_0_0.vlm"},
	{0x361C60E777511ED1ULL, "zz_mag_debug6054608.cl"},
	{0x361CD73AD3D2CD67ULL, "art\\vfx\\combat\\vfx_w_wns_left_close1_c.impa.evt"},
	{0x361CF8E41493A447ULL, "vxlt_0_lowtown_a2_n_c_efh_0_0.vlm"},
	{0x361CFDA5097CD688ULL, "650274996.ogg"},
	{0x361D1D80AB7D0116ULL, "dae320ip_chant_board_ht_nt.utp"},
	{0x361D9B405BABD180ULL, "vxca_caverns1_mines_ihe_0_0.vlm"},
	{0x361DA969ED0C2623ULL, "vxw_bonepit_10x_0_0.vlm"},
	{0x361DE398D5143AB7ULL, "389996319.ogg"},
	{0x361E02B4370DF36FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr.mao"},
	{0x361E0D44A6E37113ULL, "vxht_mansion2_3bo_0_0.vlm"},
	{0x361E0EB4370E070CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr.mmh"},
	{0x361E14B4370E11FEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr.msh"},
	{0x361EAEF60CD27437ULL, "vxw_bonepitdestroyed_a3_24c_0_0.vlm"},
	{0x361F4BB5F5039550ULL, "vxht_0_hightown_a2j_0_0.vlm"},
	{0x361FC290FD32E425ULL, "vxw_woundedcoast_134_0_0.vlm"},
	{0x361FFE0F02988791ULL, "art\\characters\\playercharacter\\humanmale\\hoo_noba.msh"},
	{0x36200C0F02989FACULL, "art\\characters\\playercharacter\\humanmale\\hoo_noba.mao"},
	{0x3620180F0298B44FULL, "art\\characters\\playercharacter\\humanmale\\hoo_noba.mmh"},
	{0x36203CF6E880B386ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.impa.gad"},
	{0x36209CF30D21B088ULL, "ht_mansion2b_818_ll.mmh"},
	{0x3620C3AAE0E531F9ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathv1.gad"},
	{0x3620ED4C26F2013CULL, "seb321_leliana6180912.cl"},
	{0x36216B2A8C1EDA3CULL, "mag100_thrask6035907.cl"},
	{0x36219F7112C6F70BULL, "6074497_m.fxe"},
	{0x3621AB774664F758ULL, "vxlt_darktown_4np_0_0.vlm"},
	{0x3621AE17C6E2BF23ULL, "217993136.ogg"},
	{0x3621DC570F1C9317ULL, "6143148_m.fxe"},
	{0x362206AF2E0C7BAAULL, "vxdr_primevaltomb_8m5_0_0.vlm"},
	{0x362231507E7DA5F4ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medd.phy"},
	{0x36224B693FD3B14AULL, "vxht_mansion2_1t6_0_0.vlm"},
	{0x3622B197A01CBDDAULL, "vxlt_0_lowtown_a2_eki_0_2.vlm"},
	{0x3622E463368BE2FFULL, "hlac_ht_mansion_n_3382_1z.dds"},
	{0x3622FAC8A09D0C34ULL, "vxlt_gallowscourtyard_71d_0_0.vlm"},
	{0x3623776583D926AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\blendtree\\trooper_overlay.anb"},
	{0x3623D6DEC88C2489ULL, "vxlt_undercityencounter_1mo_0_0.vlm"},
	{0x3623F83AD3D923B0ULL, "art\\vfx\\combat\\vfx_w_wns_left_close1_c.impa.gad"},
	{0x3624113F205AADE1ULL, "6117693_m.fxe"},
	{0x36241228965E03C2ULL, "character_character_p243.psh.fxo.dx11.assembly"},
	{0x3624655E320CCB98ULL, "vxlt_0_lowtown_ei3_0_0.vlm"},
	{0x3624C0BDCD3A240EULL, "ave121_jeven6105641.cl"},
	{0x3624C94827A373D5ULL, "llac_w_woundedcoast_327688_0.dds"},
	{0x3624D7064658AD43ULL, "plt_cod_aow_spellcombo5.plo"},
	{0x362566E0678DB01BULL, "vxlt_0_lowtown_docks_n_bbl_0_0.vlm"},
	{0x36259F9BA2CD7959ULL, "6126503_m.fxe"},
	{0x3625D4D6EA737F72ULL, "254672167.ogg"},
	{0x3626333F0A084FD8ULL, "6088916_m.fxe"},
	{0x36266187485FFB18ULL, "vxht_chantry_n_2yc_0_0.vlm"},
	{0x36266718D4155633ULL, "6080186_m.fxe"},
	{0x36267624956913B4ULL, "hlac_w_encounter2_camp_65536_0y.dds"},
	{0x3626C1A8A9F0C94AULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.dura.gad"},
	{0x3626C873D0BBAF86ULL, "llac_lt_0_lowtown_a3_n_c_1_6.dds"},
	{0x3626D35B93492862ULL, "6168149_f.fxe"},
	{0x3626D4AAE0E9BA47ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathv1.ani"},
	{0x362737F6E886CA65ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.impa.evt"},
	{0x36274454B762F57AULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.cess.gad"},
	{0x3627465BCD82A70AULL, "vxdr_0_deeproads_ex9_0_0.vlm"},
	{0x362757CD936A6E96ULL, "vxht_0_hightown_a3_9v1_0_0.vlm"},
	{0x36279AC7D78228BBULL, "plt_fen300pt_opening.nss"},
	{0x3627C72828AED52EULL, "vxw_bonepitdestroyed_a3_1o0_0_0.vlm"},
	{0x36283D4CDC5B8311ULL, "6049648_m.fxe"},
	{0x36285A85BCE560A1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_slabs.phy"},
	{0x3628853DE6D5E97DULL, "lt_warehouse_n_probe_1gp.dds"},
	{0x3628BB9C6832FED2ULL, "125947895.ogg"},
	{0x3628D7231912C945ULL, "gen00fl_carver6104855.cl"},
	{0x3629500145E94ACAULL, "739635771.ogg"},
	{0x362993875AB93B32ULL, "vxw_sundermount_2d9_0_0.vlm"},
	{0x362A53B6EEECFDD9ULL, "gen_im_trs_bst_claws_03.uti"},
	{0x362A69BF63947142ULL, "6060314_m.fxe"},
	{0x362A9323E836C954ULL, "vxht_0_hightown_cdz_0_0.vlm"},
	{0x362B6315913D37BDULL, "110860773.ogg"},
	{0x362C0E507E85BEF1ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medd.msh"},
	{0x362C0F888701F3EEULL, "hlac_w_bonepitmining_a2_327686_0x.dds"},
	{0x362C1C507E85D60CULL, "art\\characters\\playercharacter\\humanmale\\hlf_medd.mao"},
	{0x362C28507E85EAAFULL, "art\\characters\\playercharacter\\humanmale\\hlf_medd.mmh"},
	{0x362C815D4F098019ULL, "399101996.ogg"},
	{0x362CF964A54DC960ULL, "vxdr_slavers_2hg_0_1.vlm"},
	{0x362D4754B7676591ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.cess.evt"},
	{0x362DA1433BD3B5DFULL, "6145209_m.fxe"},
	{0x362DA4A8A9F6B6E1ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.dura.evt"},
	{0x362DAB494C2B61A9ULL, "dae000_friend_turn_in6179298.cl"},
	{0x362E25A5E62E1A96ULL, "vxht_brothel_n_6di_0_0.vlm"},
	{0x362EC94605F39652ULL, "6090854_m.fxe"},
	{0x362F1F84B2C7F698ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyn_0s.dds"},
	{0x36301BF2115898D9ULL, "800492118.ogg"},
	{0x363045AFA359C280ULL, "vxlt_0_lowtown_a2_n_e2y_0_0.vlm"},
	{0x3630793F8E5C5759ULL, "hlac_lt_gallowscourtyard_65536_1y.dds"},
	{0x3630D23E89E5E894ULL, "6080578_m.fxe"},
	{0x3630DDCD6133075EULL, "hlac_lt_0_lowtown_a2_1_2z.dds"},
	{0x3630EDF2D998BDC4ULL, "798824943.ogg"},
	{0x3630F42AE8E08898ULL, "llac_lt_0_lowtown_a2_n_65537_0.dds"},
	{0x36313C24BE592B5AULL, "gen00fl_aveline6132940.cl"},
	{0x3631663DC8619517ULL, "217512041.ogg"},
	{0x36316BA774CB4C56ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tray.msh"},
	{0x363175A774CB5D54ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tray.mmh"},
	{0x363179A774CB6467ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tray.mao"},
	{0x3631CCE51544FDD7ULL, "t3_rob_pl4.tnt"},
	{0x36323BA2564DA690ULL, "395893575.ogg"},
	{0x3632427B0B000474ULL, "vxlt_gallowstemplar_n_38r_0_0.vlm"},
	{0x3633635F20F799DAULL, "6095075_m.fxe"},
	{0x363388320C2905FFULL, "m_dae380ar_deep_roads.bnk"},
	{0x363422DD09816F82ULL, "6093557_f.fxe"},
	{0x36346AA0C94CC7A8ULL, "6108093_f.fxe"},
	{0x3634848E6AE71557ULL, "qun231_varnell6078111.cl"},
	{0x363494C573C73BCAULL, "251373852.ogg"},
	{0x3634EA51B21DEA88ULL, "6154158_m.fxe"},
	{0x3635B15865565D5AULL, "mrl310_party_comments6142763.cl"},
	{0x3636587C3AE3E9B0ULL, "war_multiwepattack.dds"},
	{0x36366BD86447BAB0ULL, "hlac_lt_0_lowtown_a3_n_c_-262143_0z.dds"},
	{0x36367845B57710BCULL, "6134457_m.fxe"},
	{0x363715419BABB77AULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.cess.gad"},
	{0x3637685A86F8611CULL, "seb221_flora_harimann6187801.cl"},
	{0x3637EF4FBDE94266ULL, "497360197.ogg"},
	{0x3637F7051704F1E3ULL, "581163780.ogg"},
	{0x3637F97A3CB7F13CULL, "200055291.ogg"},
	{0x363825FBF6226FE1ULL, "hc_cloud2.mao"},
	{0x3638517D7797C503ULL, "follower_banter6163512.cl"},
	{0x3638BB5CD143C3D5ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_big.msh"},
	{0x3638BD5CD143C74BULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_big.mmh"},
	{0x3638C62AFCB06525ULL, "6155844_m.fxe"},
	{0x3639129D5BD8D37DULL, "rune_weapon_spirit.dds"},
	{0x36397B52BF00EF82ULL, "vxca_caverns1_mines_e7f_0_0.vlm"},
	{0x3639A65AF1B38670ULL, "mag110cr_guardsman.utc"},
	{0x3639C80FBF09FD2AULL, "vxlt_0_lowtown_ghd_0_0.vlm"},
	{0x363A0BF220CC718EULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_barge_.ani"},
	{0x363A87B5ADC491D6ULL, "vxlt_0_lowtown_docks_a3_n_czj_0_1.vlm"},
	{0x363AF4CE7297B2A7ULL, "dae000_denier6162499.cl"},
	{0x363B07A145287225ULL, "ser120_ghyslain6105046.cl"},
	{0x363B097202A1CE7DULL, "vxw_encounter2_g0_0_0.vlm"},
	{0x363B543FD34B3BFDULL, "400456954.ogg"},
	{0x363B5D618FCC5F21ULL, "hlac_lt_0_lowtown_a2_n_1_1y.dds"},
	{0x363B5F54B773CC60ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.cess.ani"},
	{0x363B87E04AD51E54ULL, "82525802.ogg"},
	{0x363B9CA8AA02E630ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.dura.ani"},
	{0x363C1747A75CA7B4ULL, "6168139_m.fxe"},
	{0x363C70C0C54FF51EULL, "hlac_dr_slavers_3207_8z.dds"},
	{0x363C7AF97E78FC87ULL, "and000ip_amulet_container.utp"},
	{0x363C9782D8243316ULL, "vxht_brothel_n_6aj_0_0.vlm"},
	{0x363CA64AE11F1525ULL, "vxw_woundedcoast_192_0_0.vlm"},
	{0x363CEE05309E5742ULL, "1013029015.ogg"},
	{0x363D18419BB02791ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.cess.evt"},
	{0x363D3B22F7E9AED2ULL, "6095754_m.fxe"},
	{0x363D52CDE89757A3ULL, "vxlt_backalley_g_1f8_0_0.vlm"},
	{0x363D576204C8AD44ULL, "hlac_w_bonepitdestroyed_a3_65540_0x.dds"},
	{0x363D5D0DC4BADB66ULL, "meleering1.gda"},
	{0x363D9A3ABC847089ULL, "ser100_party_comments6108241.cl"},
	{0x363D9FDB097605C4ULL, "vxlt_gallowstemplar_f_5j6_0_0.vlm"},
	{0x363DDA60A0AC674BULL, "6095878_m.fxe"},
	{0x363E3958142A0A78ULL, "gen00fl_anders6113052.cl"},
	{0x363F1EE85B2119EAULL, "vxht_mansion2b_1z0_0_0.vlm"},
	{0x363F2F2323BFE24AULL, "vxht_mansion2_1b0_0_0.vlm"},
	{0x3640BA98AC785589ULL, "vxlt_undercityencounter_1gm_0_0.vlm"},
	{0x3640FD825B17E5B5ULL, "gen00fl_merrill6108523.cl"},
	{0x364115F366E06CD9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_020a_0t.dds"},
	{0x36413563FC18C8E2ULL, "dae000_flemeth6013307.cl"},
	{0x364137DC0A4F2F12ULL, "dae000cr_grand_cleric6173675.cl"},
	{0x36424A9A4B79E11BULL, "vxlt_0_lowtown_docks_n_btn_0_0.vlm"},
	{0x3642544B0E879322ULL, "6114490_f.fxe"},
	{0x3642835586B9AA59ULL, "6128703_m.fxe"},
	{0x36431D7BDF482AE8ULL, "241789360.ogg"},
	{0x36441B52553D0F71ULL, "vxht_bartrand_n_64b_0_0.vlm"},
	{0x36447255EFCFB9CAULL, "vxlt_0_lowtown_docks_a3_n_buw_0_0.vlm"},
	{0x3644E96FD69C6C26ULL, "vxw_woundedcoast_1ma_0_0.vlm"},
	{0x36453A7D9CAA0446ULL, "zz_dae_debug6049996.cl"},
	{0x364577E912CD0B27ULL, "mrl000_marethari6129299.cl"},
	{0x3645A54184AD6F72ULL, "6103057_m.fxe"},
	{0x3645BAB3779BE8DEULL, "hlac_ht_0_hightown_2_1z.dds"},
	{0x3645CE2A5921FB47ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comb_0n.dds"},
	{0x3645DD5AE3640A0DULL, "284263824.ogg"},
	{0x364678AFDC70EC77ULL, "w_sundermount_65_0_ll.mmh"},
	{0x3646BBDFC5450408ULL, "fex_bd_a_md_icn.msh"},
	{0x3646D9DFC5453732ULL, "fex_bd_a_md_icn.mmh"},
	{0x36474ED5E4C0CFCEULL, "sse_ss_fleshgolem_1.gda"},
	{0x36475F4FEBDD01CDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wi_basic.phy"},
	{0x3647AA81041961E8ULL, "t3_dey_gr1.tnt"},
	{0x3647BAACC83DEA2FULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.mmh"},
	{0x3647DE30589F77C4ULL, "995530977.ogg"},
	{0x3648AEBBB90F4BCFULL, "939742377.ogg"},
	{0x3648BB1161BE2867ULL, "gen00fl_merrill6162020.cl"},
	{0x3648C644BF2A0450ULL, "6105479_m.fxe"},
	{0x3648FBBD3CAE743AULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.dura.gad"},
	{0x3649165F0AF0F549ULL, "gen00fl_aveline6139214.cl"},
	{0x36498FE875A018F9ULL, "vxht_keep_707_0_0.vlm"},
	{0x3649C2C1E2EA595FULL, "vxht_bartrand_n_6ph_0_0.vlm"},
	{0x3649CFA93EE3F99EULL, "hlac_lt_0_lowtown_a2_196607_0z.dds"},
	{0x364B095FCA1A4B96ULL, "vxht_brothel_n_6jk_0_0.vlm"},
	{0x364B30419BBC8E60ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.cess.ani"},
	{0x364C1DD9B5FB8DC3ULL, "lt_0_lowtown_a3_n_c_7_0.mao"},
	{0x364C21D9B5FB9410ULL, "lt_0_lowtown_a3_n_c_7_0.mmh"},
	{0x364C2FD9B5FBAC2AULL, "lt_0_lowtown_a3_n_c_7_0.msh"},
	{0x364C81B47DC6FDE1ULL, "gen00fl_aveline6152543.cl"},
	{0x364CCCC3822ED8CAULL, "gen00fl_aveline6115271.cl"},
	{0x364D29698745F380ULL, "vxlt_0_lowtown_a2_n_e8g_0_0.vlm"},
	{0x364D35A443B980A4ULL, "dae000_bartrand6206148.cl"},
	{0x364DE4F220DCD8C8ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_barge_.gad"},
	{0x364E0534E97C07F7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchhang_sa.dds"},
	{0x364E6C3B7360B890ULL, "171820704.ogg"},
	{0x364E87357120B860ULL, "vxlt_0_lowtown_docks_cs7_0_1.vlm"},
	{0x364EC7C4778B159CULL, "6130079_m.fxe"},
	{0x364F53BD2394B5F2ULL, "209845998.ogg"},
	{0x364FE6ACF3051B3CULL, "art\\characters\\creatures\\hurlock\\animation\\blendtree\\hrlk_sw.anb"},
	{0x364FFEBD3CB49851ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.dura.evt"},
	{0x36506409207E9901ULL, "634617174.ogg"},
	{0x3650727C524C900BULL, "889500553.ogg"},
	{0x3650785176FA3196ULL, "6150703_m.fxe"},
	{0x365090130B40F730ULL, "vxlt_gallowscourtyard_n_7dl_0_0.vlm"},
	{0x3650F91E10ED0463ULL, "6078475_m.fxe"},
	{0x3650FB4A9BFCA0C0ULL, "hlac_ht_brothel_n_2995_6x.dds"},
	{0x36515B3DE1939733ULL, "vxht_keep_4hf_0_0.vlm"},
	{0x36515FD17A8A7C8DULL, "1049778678.ogg"},
	{0x365174F4F10C5DBDULL, "6068026_f.fxe"},
	{0x36518ABD5EA8C035ULL, "256168177.ogg"},
	{0x365199ED13A62236ULL, "one220_yevhen6104144.cl"},
	{0x3652B8F7F18F804BULL, "dae000_flemeth6000784.cl"},
	{0x3653506FAA186C8CULL, "440845317.ogg"},
	{0x36535B4CC89284A6ULL, "vxw_woundedcoast_1hb_0_0.vlm"},
	{0x3653AB9CAF61BF5BULL, "835163432.ogg"},
	{0x36544697DC15B521ULL, "vxht_mansion2_24r_0_0.vlm"},
	{0x365458C1B9676928ULL, "102482737.ogg"},
	{0x365475D6058CCB00ULL, "vxca_caverns1_mines_j59_0_0.vlm"},
	{0x3654BF88CCCD2B62ULL, "687243556.ogg"},
	{0x365563F2EF8034C5ULL, "75498305.ogg"},
	{0x36566D23C4096B7AULL, "vxlt_foundry_n_xv_0_0.vlm"},
	{0x3656820B719C0027ULL, "ele_f2.dds"},
	{0x36569FA9AC8A4304ULL, "gen00fl_fenris6126348.cl"},
	{0x3657A314F87F7521ULL, "6097049_m.fxe"},
	{0x3657A5C7D7AB5F54ULL, "plt_fen300pt_opening.plo"},
	{0x3657E76453A1C36FULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyn.mao"},
	{0x3657F36453A1D70CULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyn.mmh"},
	{0x3657F96453A1E1FEULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyn.msh"},
	{0x36584F457BBCF592ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb06a.phy"},
	{0x36587B3AC27A070AULL, "hlac_w_sundermount_65537_0y.dds"},
	{0x3658C578B2E72D3EULL, "vxca_caverns2_7kn_0_0.vlm"},
	{0x3658D4FF51B884CFULL, "vxlt_warehouse_ci_0_0.vlm"},
	{0x365946C5395FC296ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm04.msh"},
	{0x365950C5395FD394ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm04.mmh"},
	{0x3659CEAAEE67F8E7ULL, "6139933_f.fxe"},
	{0x3659EB2F59844CBEULL, "dae000_arishok6068550.cl"},
	{0x3659F9FD81230E51ULL, "zz_followers_banter6158529.cl"},
	{0x365A876F4869BE81ULL, "mag360_samson6042056.cl"},
	{0x365B16F1F208B9FEULL, "hlac_lt_0_lowtown_-65536_2y.dds"},
	{0x365B1DD5F307F41DULL, "llac_lt_darktown_-1_58.dds"},
	{0x365B512B96C7F763ULL, "art\\vfx\\combat\\vfx_glb_ele_f_extreme_c.impa.evt"},
	{0x365BCE22F5E4DDAAULL, "vxdr_primevaltomb_8y1_0_0.vlm"},
	{0x365BFAB05DB4D32CULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsmb.msh"},
	{0x365C00B05DB4DD1EULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsmb.mmh"},
	{0x365C0D18B1C40508ULL, "6084228_m.fxe"},
	{0x365CBBB020ACA5CBULL, "llac_lt_0_lowtown_a2_-65537_1.dds"},
	{0x365CC41D436EC051ULL, "vxht_0_hightown_a2_bos_0_0.vlm"},
	{0x365D16BD3CBF4B20ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.dura.ani"},
	{0x365D9B4015AF518AULL, "6107291_m.fxe"},
	{0x365DB5E4CD4EC096ULL, "vxw_sundermount_20y_0_0.vlm"},
	{0x365E01F8392E9540ULL, "561041579.ogg"},
	{0x365E49DD47EE4686ULL, "dae300_vashothlead6062788.cl"},
	{0x365ED113C5603E30ULL, "mag360_thrask6046479.cl"},
	{0x365F44DC0EDF989CULL, "438506361.ogg"},
	{0x365F745787002081ULL, "hlac_lt_undercityencounter_-1_34x.dds"},
	{0x365F971CF3710547ULL, "319807674.ogg"},
	{0x365F9992220A75ACULL, "586478174.ogg"},
	{0x365FCD1AD389AFB3ULL, "vxht_keep_4mi_0_0.vlm"},
	{0x36606E320C4FC735ULL, "m_dae380ar_deep_roads.wbd"},
	{0x3660861EDFD53030ULL, "vxca_caverns1_mines_j4z_0_0.vlm"},
	{0x3660F439A7B67845ULL, "307134889.ogg"},
	{0x36611F415B42D096ULL, "vxht_0_hightown_a3_9b5_0_0.vlm"},
	{0x366195E96BFDC14CULL, "follower_banter6148705.cl"},
	{0x3661B19DBAED6725ULL, "hlac_ca_caverns2_5802_19x.dds"},
	{0x3661CD29BA889D26ULL, "vxw_woundedcoast_1sc_0_0.vlm"},
	{0x36622A2B96CDD3E4ULL, "art\\vfx\\combat\\vfx_glb_ele_f_extreme_c.impa.gad"},
	{0x366290B90EE27437ULL, "vxlt_0_lowtown_cmn_0_0.vlm"},
	{0x3662B874CE0BCDA1ULL, "vxht_mansion2_21q_0_0.vlm"},
	{0x3662E622EB453856ULL, "gen00fl_varric6033673.cl"},
	{0x366343961E526B14ULL, "gen_im_trs_rdr_two_2ha_r1.uti"},
	{0x366447272D2A9C74ULL, "art\\characters\\heraldry\\hld_ferelden.dds"},
	{0x3664B66D26B2D69FULL, "6031656_m.fxe"},
	{0x3664C38EC11E9932ULL, "hlac_w_bonepitdestroyed_a3_0_11z.dds"},
	{0x3664E6105504CB63ULL, "mcr310_cut_chantry_blows_up6116286.cl"},
	{0x3664ED2496DB5F79ULL, "hlac_w_encounter2_-65533_0x.dds"},
	{0x366611064F83FDB7ULL, "6150988_m.fxe"},
	{0x366661413601F578ULL, "vxht_0_hightown_a2_c68_0_0.vlm"},
	{0x36669BA837201F13ULL, "mag320_bsc_banter6100763.cl"},
	{0x366703A1509CC0F7ULL, "wm_ico_darktown.dds"},
	{0x366746DC43AE9D4FULL, "vxlt_warehouse_dh_0_0.vlm"},
	{0x36679F3C2FCC83FAULL, "499335672.ogg"},
	{0x36680CA08FC88E1AULL, "vxht_mansion2_1rp_0_0.vlm"},
	{0x3668271E3666A3C6ULL, "hlac_lt_0_lowtown_a2_-196608_1y.dds"},
	{0x366834377D8BE38FULL, "gen00fl_varric6093665.cl"},
	{0x3668370A6219377AULL, "vxlt_gallowstemplar_f_5wo_0_0.vlm"},
	{0x3668B800857FC713ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v2.ani"},
	{0x3668E2A1C1A60CEEULL, "vxdr_primevaltomb_8px_0_2.vlm"},
	{0x3669392B96D40BC2ULL, "art\\vfx\\combat\\vfx_glb_ele_f_extreme_c.impa.ani"},
	{0x36695C80B77FBD87ULL, "642490293.ogg"},
	{0x366ACAE0546B49BEULL, "15060699.ogg"},
	{0x366B3C30DD7A84EBULL, "vxlt_gallowstemplar_f_38q_0_0.vlm"},
	{0x366B6A4BBE507645ULL, "dae011_conspiracy_mob6149554.cl"},
	{0x366B77553DFC81BFULL, "hlac_dr_0_deeproads_5840_3y.dds"},
	{0x366B83D2CABE65C3ULL, "zz_followers_banter6159696.cl"},
	{0x366BD228D6E904D7ULL, "1060363653.ogg"},
	{0x366BFEF5098C6401ULL, "mrl310_merrill6198097.cl"},
	{0x366C14E220FD3BEAULL, "hlac_w_encounter2_camp_65538_0y.dds"},
	{0x366C27C1BF44D916ULL, "vxw_sundermount_25z_0_0.vlm"},
	{0x366C6DC6F35EB7ADULL, "6059783_f.fxe"},
	{0x366CA35DF58F2724ULL, "6208676_f.fxe"},
	{0x366D197F2C507E87ULL, "771465508.ogg"},
	{0x366D3C6315BA280EULL, "hlac_lt_0_lowtown_docks_a3_-2_3y.dds"},
	{0x366D68CE12092706ULL, "hlac_lt_warehouse_65536_2x.dds"},
	{0x366D9C90B9716151ULL, "772122796.ogg"},
	{0x366DCF6093CFECD0ULL, "62534662.ogg"},
	{0x366DF897CABF0984ULL, "6173405_m.fxe"},
	{0x366E6F08D8570BE8ULL, "hlac_lt_0_lowtown_docks_n_-131075_6y.dds"},
	{0x366E72BBBF4E6798ULL, "869724747.ogg"},
	{0x366EC752E368CC5CULL, "zz_one_debug6099971.cl"},
	{0x366F6969487B7B8CULL, "plt_cod_hst_orlais.plo"},
	{0x366F97D7F9F608F2ULL, "vxlt_hangedman_20o_0_0.vlm"},
	{0x366FC90E1D58E761ULL, "vxlt_0_lowtown_a2_n_c_bxk_0_0.vlm"},
	{0x36707C5428CD5ACFULL, "122734673.ogg"},
	{0x367083575EAB1156ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_rblt.anb"},
	{0x3670BBACC860F16CULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.anb"},
	{0x3671150263A64115ULL, "art\\characters\\playercharacter\\qunarimale\\eye_mag.mao"},
	{0x36712ED85440E95FULL, "dae350ip_at_scaverns_a.utp"},
	{0x367174EC568C9141ULL, "vxca_cavern1_haunted_jfm_0_0.vlm"},
	{0x36718A5E358C34AAULL, "root_mage_primal1.dds"},
	{0x36720C68BD96A210ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_9y.dds"},
	{0x36721B50D8A712B9ULL, "vxca_caverns2_7gc_0_2.vlm"},
	{0x367351923F48EA0FULL, "vxw_encounter1_cs_0_0.vlm"},
	{0x36735A4945A2510FULL, "vxlt_hangedman_n_ro_0_0.vlm"},
	{0x3673F69737ADA4B3ULL, "345584006.ogg"},
	{0x367458EF005E380CULL, "gen00fl_varric6076067.cl"},
	{0x367471433F85B472ULL, "qun221_petrice6083825.cl"},
	{0x3675765E00EE5757ULL, "vxca_caverns2_8da_0_0.vlm"},
	{0x367595DB3EF50753ULL, "vxlt_0_lowtown_a2_n_hb7_0_0.vlm"},
	{0x3675B51437F00A61ULL, "zz_followers_banter6157850.cl"},
	{0x36762AF90E4A8259ULL, "vxlt_backalley_g_1vz_0_0.vlm"},
	{0x36763DE78A4744F7ULL, "dre000_sibling_has_blight6036464.cl"},
	{0x36766862E28E4065ULL, "vxlt_0_lowtown_docks_b7s_0_1.vlm"},
	{0x36767E7D81BEA69AULL, "vxht_mansion2_1uq_0_0.vlm"},
	{0x3676B000858BF74AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v2.evt"},
	{0x3676CD94C162DD30ULL, "hlac_w_bonepitmining_a2_393221_0x.dds"},
	{0x3676E8C0A4877C9FULL, "ser270_emeric6149280.cl"},
	{0x36776769B9F9D5DBULL, "em_ath_entouragem.mor"},
	{0x3677CE844C94F1B8ULL, "hlac_w_encounter2_262149_0x.dds"},
	{0x3677DEE78226EC1FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05b.msh"},
	{0x3677E4E78226F6C1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05b.mmh"},
	{0x3677F8543058DA6FULL, "6077415_f.fxe"},
	{0x367878C3D9F69B9DULL, "mer220_hubert6081393.cl"},
	{0x367890647552127AULL, "540230705.ogg"},
	{0x367951C3DEAD9348ULL, "vxlt_0_lowtown_a2_n_c_bxn_0_0.vlm"},
	{0x3679C2AB2B38C578ULL, "6125503_m.fxe"},
	{0x3679C84150E0EC9CULL, "6098261_m.fxe"},
	{0x367A1B4FBB739487ULL, "6101004_f.fxe"},
	{0x367A2EF1624CF7E8ULL, "958792769.ogg"},
	{0x367A31D61C98A368ULL, "gencr_buzz_glide.utc"},
	{0x367A5ED958672FE8ULL, "vxht_0_hightown_c82_0_0.vlm"},
	{0x367A6D603A2318AEULL, "628809126.ogg"},
	{0x367A820C7450B789ULL, "vxlt_undercityencounter_1sq_0_0.vlm"},
	{0x367A9642DA8C8CF2ULL, "785881025.ogg"},
	{0x367B34A9D14EF3B8ULL, "gen00fl_anders6121189.cl"},
	{0x367C34FAA1016CACULL, "6208356_m.fxe"},
	{0x367C40397444EB47ULL, "6111594_m.fxe"},
	{0x367C87684310C3EAULL, "qcr220_grey_wardens6060403.cl"},
	{0x367CD48C23A9F711ULL, "6206148_m.fxe"},
	{0x367D0C3449AADE5CULL, "hm_slavermage.mor"},
	{0x367D51C3321ED5BEULL, "305219138.ogg"},
	{0x367D8EDAC5E5D17DULL, "922218094.ogg"},
	{0x367D8F008591DDBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v2.gad"},
	{0x367DB143F208EC8CULL, "panning1_h.cur"},
	{0x367DF952D75FB9B4ULL, "vxht_mansion2_2a8_0_0.vlm"},
	{0x367F0DDB2BFF7F9EULL, "llac_lt_darktown_-1_71.dds"},
	{0x367F12FD35F1096EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsarch.msh"},
	{0x367F20FD35F120BFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsarch.mao"},
	{0x367F2CFD35F1355CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsarch.mmh"},
	{0x367FCEC10670D974ULL, "gen_im_trs_tmp_one_mac_r0.uti"},
	{0x36800046434A1C66ULL, "6139250_m.fxe"},
	{0x368049E57802BA8DULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\blendtree\\armoire.anb"},
	{0x368063394F58D82AULL, "llac_ca_cavern1_haunted_24426_13.dds"},
	{0x368063FB6E1F813EULL, "134642543.ogg"},
	{0x3680ED9CD586E324ULL, "ser221_party_comments6166116.cl"},
	{0x36811319BD08B688ULL, "rog_defensebuff.dds"},
	{0x3681658C513C22AEULL, "llac_ht_mansion2b_818_31.dds"},
	{0x3681AD6E9167A6BCULL, "383779647.ogg"},
	{0x3681C36F313F028FULL, "vxw_encounter1_hr_0_0.vlm"},
	{0x3681F3EB42A7D688ULL, "6074044_f.fxe"},
	{0x36822C5F5062E760ULL, "qcr000_arishok6042488.cl"},
	{0x36825E11DBCE8C55ULL, "299693528.ogg"},
	{0x3682C53F7456418FULL, "lowtown_male_lowlife1_sta6165004.cl"},
	{0x3682E9831C89948DULL, "fen210_hadriana6110387.cl"},
	{0x3682FA6512633F34ULL, "vxlt_shop_2v_0_0.vlm"},
	{0x36833CD48C9EF96DULL, "6130962_m.fxe"},
	{0x368344FB19EE2678ULL, "vxht_0_hightown_a2_c86_0_0.vlm"},
	{0x36837295C185B293ULL, "vxlt_0_lowtown_a2_e1h_0_0.vlm"},
	{0x3683D78638487B73ULL, "mag131_taraneh6094527.cl"},
	{0x36840C72443B5F68ULL, "lgt261ip_at_coast.utp"},
	{0x368446CB792D91E8ULL, "gr_ele.mao"},
	{0x36845B89E0C38E59ULL, "hlac_ca_caverns1_24424_19z.dds"},
	{0x36845C644471955CULL, "vxca_caverns1_mines_ih9_0_0.vlm"},
	{0x36848BD80886C3D7ULL, "6153822_m.fxe"},
	{0x3684E81171DC7F04ULL, "903016690.ogg"},
	{0x3684F05A73B4BF1AULL, "vxht_mansion2_1xr_0_0.vlm"},
	{0x36851BF797707251ULL, "vxlt_0_lowtown_a2_hfr_0_0.vlm"},
	{0x368550B6C92DB1D3ULL, "zz_followers_banter6158970.cl"},
	{0x3685D6455E974EB6ULL, "vxlt_shop_sw_0_0.vlm"},
	{0x3685D792C5DEF44CULL, "zz_vrc_debug6200233.cl"},
	{0x3685FB3C28F83706ULL, "702910874.ogg"},
	{0x368667C0EE2BAE24ULL, "dae000_arishok6068480.cl"},
	{0x3686EA67AD230FAAULL, "vxht_chantry_n_32i_0_0.vlm"},
	{0x3686F8E765305D85ULL, "vxht_brothel_n_589_0_0.vlm"},
	{0x36870C042D1F0371ULL, "dre000_varric6027788.cl"},
	{0x368721CCF5A17531ULL, "503840807.ogg"},
	{0x36872F2989B18AA9ULL, "vxlt_0_lowtown_a2_n_by2_0_0.vlm"},
	{0x368744BF0DE1B794ULL, "dae000_mother6136941.cl"},
	{0x36884AEFE78191C5ULL, "6112980_m.fxe"},
	{0x36885E195A7CDF9DULL, "lgt210_shp_kill_qunari16156855.cl"},
	{0x36886F0DA1B68920ULL, "vxht_bartrand_n_67j_0_0.vlm"},
	{0x36888869CEED66D6ULL, "vxht_bartrand_n_6rs_0_0.vlm"},
	{0x3688D4CDA75C3C55ULL, "6095890_m.fxe"},
	{0x36896ABA6A9CEFA7ULL, "559487720.ogg"},
	{0x36897F5342377188ULL, "6131606_f.fxe"},
	{0x368A5F85BD3833DDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_slabs.mao"},
	{0x368A6385BD383A16ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_slabs.mmh"},
	{0x368A6D85BD384B14ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_slabs.msh"},
	{0x368B4123EABF01EEULL, "gen00fl_anders6103671.cl"},
	{0x368B48A896F34595ULL, "hlac_ca_caverns1_24424_25y.dds"},
	{0x368C16166B456674ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_teleport.ani"},
	{0x368C6B2FC955D234ULL, "vxht_mansion2_2j9_0_0.vlm"},
	{0x368C9FBAE98BAF7AULL, "6174583_m.fxe"},
	{0x368CCE7F222251BDULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_5y.dds"},
	{0x368D28D9B633372BULL, "lt_0_lowtown_a3_n_c_7_0.phy"},
	{0x368DA98E4108D6B2ULL, "and111_refugee_barks6111539.cl"},
	{0x368E4722E78297C4ULL, "6168687_m.fxe"},
	{0x368EAFF4D7B62A64ULL, "115536878.ogg"},
	{0x368EEB5CC69DB3F4ULL, "llac_lt_gallowstemplar_f_2391_8.dds"},
	{0x368F495A994BEED9ULL, "1039591508.ogg"},
	{0x368F761C976DE4C1ULL, "68579183.ogg"},
	{0x368F76DD273204E9ULL, "6087710_m.fxe"},
	{0x368F90C716925195ULL, "73334296.ogg"},
	{0x368FB6C09D0EA8FFULL, "vxw_encounter1_d6_0_0.vlm"},
	{0x368FDECF3B24592CULL, "316998321.ogg"},
	{0x36905499C23545D4ULL, "6115786_m.fxe"},
	{0x369166A03867C1B1ULL, "971608266.ogg"},
	{0x3691A1C2D7D70BD9ULL, "6036744_f.fxe"},
	{0x3691D93D1E90611DULL, "rune_weapon_damage.dds"},
	{0x3691DD527226A6F6ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm.phy"},
	{0x369213166B49CCD6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_teleport.gad"},
	{0x369233ECAAFF3131ULL, "183075528.ogg"},
	{0x36925C0B3C0EE12FULL, "ser270_emeric6113571.cl"},
	{0x36929C731990E6CFULL, "vxlt_warehouse_wu_0_0.vlm"},
	{0x3692E5E50A7BC20CULL, "vxlt_darktown_2w9_0_0.vlm"},
	{0x36930CBCB1B63EC3ULL, "water_pan2_blend.mao"},
	{0x36930EB2F236B359ULL, "vxlt_backalley_g_1xx_0_0.vlm"},
	{0x36933486696B303EULL, "vxht_chantry_n_2z8_0_0.vlm"},
	{0x3693836B3F422E93ULL, "vxw_encounter3_eo_0_0.vlm"},
	{0x36938DD489668AD1ULL, "vxlt_0_lowtown_a2_hcq_0_0.vlm"},
	{0x36946605B876A256ULL, "138358257.ogg"},
	{0x3694B360175F9A65ULL, "6170834_m.fxe"},
	{0x3694E5FC8F88D57FULL, "6099420_f.fxe"},
	{0x36952603E76D4846ULL, "100903207.ogg"},
	{0x3696181E6D32FAEFULL, "hlac_lt_hangedman_n_-1_8y.dds"},
	{0x36968E39BFC7C981ULL, "lowtown_male_cityguard_mob6150517.cl"},
	{0x3696ACBB8969FCC9ULL, "ca_cave_2194.rml"},
	{0x3696D53F1BDF88FDULL, "135970880.ogg"},
	{0x36972D1ABE6DC91BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleftcrn_1.mao"},
	{0x369795961A8050C4ULL, "zz_followers_banter6158481.cl"},
	{0x36981A911D3A8EF4ULL, "t3_hey_hz2.tnt"},
	{0x3698631B4B352F8FULL, "pro000_wesley6007719.cl"},
	{0x36988736FB890578ULL, "hlac_lt_0_lowtown_docks_n_196607_3y.dds"},
	{0x3698C2C7ADB1F457ULL, "6110412_m.fxe"},
	{0x36992E166B501935ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_teleport.evt"},
	{0x36994BD96218025DULL, "6110072_m.fxe"},
	{0x36998F1E4567D911ULL, "zz_demo_attract.cut"},
	{0x369B0EA86649F51BULL, "6118631_f.fxe"},
	{0x369B1F4199B63E34ULL, "599864294.ogg"},
	{0x369B54C3605BF2A9ULL, "art\\vfx\\materials\\chunky\\vfx_c_gore.mao"},
	{0x369B7D7CC936B113ULL, "vxlt_gallowstemplar_n_47c_0_0.vlm"},
	{0x369BBFE4299030E5ULL, "vxlt_shop_fo_0_0.vlm"},
	{0x369BC652722ED397ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm.msh"},
	{0x369BDC52722EF909ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm.mmh"},
	{0x369BF10CDAC0D475ULL, "449951009.ogg"},
	{0x369C4936E8DF1174ULL, "vxht_0_hightown_a2_a88_0_0.vlm"},
	{0x369C6FA44813A573ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_4x.dds"},
	{0x369CD7B05DEC2EB9ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpsmb.phy"},
	{0x369D02D8D173ED21ULL, "mag321_comtesse_de_launcet6042807.cl"},
	{0x369D54E202DB19D6ULL, "dae000_meredith6085332.cl"},
	{0x369DA71D8295C562ULL, "dre000_dougal6142197.cl"},
	{0x369DE6BEA8A7E95FULL, "zz_spy130_darktown_res_36055343.cl"},
	{0x369E04264B30C341ULL, "ser221_party_comments6209852.cl"},
	{0x369E289D8F04C17FULL, "vxw_encounter1_i1_0_0.vlm"},
	{0x369E70273673CCBFULL, "zz_qun_debug6064336.cl"},
	{0x369EA9C8D45D0559ULL, "ico_poison_vunerable.dds"},
	{0x369FFEBC5DAF5E4AULL, "vxht_bartrand_n_5ss_0_0.vlm"},
	{0x36A047E846305766ULL, "mag291_demon_marethari6028852.cl"},
	{0x36A0C8701C9C86D0ULL, "485246816.ogg"},
	{0x36A10E500B86FF4FULL, "vxlt_warehouse_xt_0_0.vlm"},
	{0x36A119BEBED5A184ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallgreen.msh"},
	{0x36A12BBEBED5C0EDULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallgreen.mao"},
	{0x36A12FBEBED5C726ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallgreen.mmh"},
	{0x36A175FAABC9582CULL, "6054972_f.fxe"},
	{0x36A1D41457A0F01AULL, "vxht_mansion2_1fl_0_0.vlm"},
	{0x36A1FD3EB290631DULL, "lt_0_lowtown_a3_n_c_45_0_ll.mmh"},
	{0x36A2B9FF42837FB6ULL, "vxlt_shop_iy_0_0.vlm"},
	{0x36A2C07AB94E9E29ULL, "llac_w_encounter1_2_0.dds"},
	{0x36A2CA55800C46CCULL, "dae321ip_at_cht_htown.utp"},
	{0x36A31A02D0F8399FULL, "mrl310_merrill6120012.cl"},
	{0x36A3521E38A56FFAULL, "rog_trip.dds"},
	{0x36A37DA594DE1AAFULL, "vxdr_dungeon_1t8_0_0.vlm"},
	{0x36A3E5E7824C333EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05b.phy"},
	{0x36A40BFD36101927ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsarch.phy"},
	{0x36A4E1AFB937A696ULL, "ht_0_hightown_n_30_0_ll.mmh"},
	{0x36A5035D6D068BC1ULL, "hlac_lt_backalley_n_65537_6z.dds"},
	{0x36A50FD9CD544F73ULL, "388836904.ogg"},
	{0x36A60FE9F1B2ADA4ULL, "ave200_aveline_bar6116373.cl"},
	{0x36A66E482EB6DF10ULL, "hlac_lt_0_lowtown_docks_65535_0x.dds"},
	{0x36A6F11B40030E3AULL, "vxht_bartrand_n_68y_0_0.vlm"},
	{0x36A70192FE1F20ADULL, "fx_rock_chip.mao"},
	{0x36A70E1E383050D0ULL, "vxlt_gallowstemplar_3s4_0_0.vlm"},
	{0x36A7B919354F245DULL, "vxht_0_hightown_bdv_0_0.vlm"},
	{0x36A8194F97E20F3FULL, "art\\levels\\ztools\\helpers\\texels\\_textures\\512m.dds"},
	{0x36A861C185B50A7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\persuade_die_lp.gad"},
	{0x36A8BDDBB883E44CULL, "vxlt_gallowstemplar_4xv_0_0.vlm"},
	{0x36A989F91D9F83BCULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.impa.ani"},
	{0x36A98C351CF39F1AULL, "zz_lgt_2_debug6211865.cl"},
	{0x36AAE6DB6D772308ULL, "tree_01.dds"},
	{0x36AB26BEBEDE0CF1ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallgreen.phy"},
	{0x36AB3D6512C788EBULL, "6152546_m.fxe"},
	{0x36AB6689D0723B22ULL, "dae000_arishok6067762.cl"},
	{0x36AB89E3714686F0ULL, "306033814.ogg"},
	{0x36AB90FCCCEFB442ULL, "6212087_m.fxe"},
	{0x36AB928A8CABB7D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_getup_v3.gad"},
	{0x36ABD511A49802D8ULL, "narrative_varric_demo6170837.cl"},
	{0x36ABD995E68F3E75ULL, "6060542_m.fxe"},
	{0x36AC5E370EB76179ULL, "vxlt_0_lowtown_ffp_0_0.vlm"},
	{0x36AC89BDE54D9088ULL, "gen00fl_aveline6139143.cl"},
	{0x36ACBBC1A0C2CACCULL, "hlac_ca_caverns2_5802_24y.dds"},
	{0x36ACC642354FC7A6ULL, "vxlt_undercitydungeon_1je_0_1.vlm"},
	{0x36AD627AFF18E19EULL, "6054080_f.fxe"},
	{0x36ADFF87AE4DEB51ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_combat_enter.gad"},
	{0x36AE1E88BB4B0B8DULL, "hlac_ht_brothel_n_3371_1y.dds"},
	{0x36AFEBE8A31B8CC2ULL, "hlac_dr_0_deeproads_5849_8z.dds"},
	{0x36B07382AD0F56B0ULL, "309854748.ogg"},
	{0x36B096F91DA5B8FEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.impa.gad"},
	{0x36B1958A8CB02809ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_getup_v3.evt"},
	{0x36B1CB7A38A24001ULL, "856929099.ogg"},
	{0x36B242899210C8FEULL, "6110251_m.fxe"},
	{0x36B373EBAA9CC74BULL, "6060742_m.fxe"},
	{0x36B3B3181B5C21F5ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal_2.mao"},
	{0x36B40087AE5258FFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_combat_enter.ani"},
	{0x36B42B220D7F6E29ULL, "ser121_jethann.bnk"},
	{0x36B49EAC6C8EFCBBULL, "vxlt_darktown_5nb_0_0.vlm"},
	{0x36B4D3D08EFCC6CAULL, "6204875_m.fxe"},
	{0x36B4F4D0C591AB08ULL, "mer210cr_coterie_thug_3.utc"},
	{0x36B50EB79645B3F5ULL, "vxdr_0_deeproads_9e4_0_1.vlm"},
	{0x36B51AE99088F1DAULL, "6151623_m.fxe"},
	{0x36B54FC539AE308FULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm04.phy"},
	{0x36B5ED152AB5F2E5ULL, "gui_lightblend_p0.psh.fxo.dx11"},
	{0x36B62AF627453CDDULL, "vxht_0_hightown_bau_0_0.vlm"},
	{0x36B64E56BBF3865AULL, "vxht_0_hightown_a3_bj6_0_0.vlm"},
	{0x36B682EE4DAC5817ULL, "one120_athenril6088910.cl"},
	{0x36B6A4DE9C42343FULL, "hlac_ht_0_hightown_a3_-65537_4x.dds"},
	{0x36B771F91DAB981DULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.impa.evt"},
	{0x36B7837F26BAD793ULL, "vxht_0_hightown_n_byd_0_0.vlm"},
	{0x36B796CB6FA23990ULL, "mag291_ambient_16206770.cl"},
	{0x36B7B3F42ACA10ADULL, "vxw_sundermount_37f_0_0.vlm"},
	{0x36B7DF75EE8C726CULL, "187476818.ogg"},
	{0x36B7E55C9C35A065ULL, "6154973_m.fxe"},
	{0x36B80D220D832469ULL, "ser121_jethann.cnv"},
	{0x36B81E77E8D20A83ULL, "vxht_mansion2b_1gz_0_0.vlm"},
	{0x36B820152D2AEC5EULL, "76752733.ogg"},
	{0x36B8B4CE022B16A9ULL, "dae000ip_party_picker.utp"},
	{0x36B8E25D77918916ULL, "dae000_bodhan6146828.cl"},
	{0x36B8F469193B5916ULL, "434353994.ogg"},
	{0x36B90548D65BD8F3ULL, "hm_apostate01.mor"},
	{0x36B93EE5DA915AF4ULL, "962327267.ogg"},
	{0x36BA3256CC50DAD0ULL, "601191976.ogg"},
	{0x36BB36F9EE528EDCULL, "vxca_caverns1_mines_j5e_0_0.vlm"},
	{0x36BC3EC185C577C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\persuade_die_lp.ani"},
	{0x36BC697DB10F0871ULL, "zz_personality_debug6004637.cl"},
	{0x36BD0E085866D254ULL, "art\\vfx\\textures\\vfx_gui_ring.dds"},
	{0x36BD84FF52F08658ULL, "dae200_bran6144559.cl"},
	{0x36BDA6590E610557ULL, "vxht_0_hightown_bhx_0_0.vlm"},
	{0x36BDFC08D86600B1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_left_hand.ani"},
	{0x36BE8DD5587A5982ULL, "mag311_emile_de_launcet6043724.cl"},
	{0x36BEED3D06C0FDE7ULL, "zz_vault_debug6203399.cl"},
	{0x36BF06BC6D17EDA5ULL, "llac_lt_0_lowtown_docks_-4_4.dds"},
	{0x36BF7D8A8CBC3D78ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_getup_v3.ani"},
	{0x36C0E4342E959CDDULL, "609762087.ogg"},
	{0x36C177E2FDCF909BULL, "lt_ruinedhouse_n_mmap.dds"},
	{0x36C1FBB9064F7DC4ULL, "follower_banter6155718.cl"},
	{0x36C202FADE370B3DULL, "vxlt_0_lowtown_a2_n_df9_0_0.vlm"},
	{0x36C25556580760DCULL, "110382805.ogg"},
	{0x36C2B51C7E9CF974ULL, "vxlt_gallowstemplar_n_3jx_0_0.vlm"},
	{0x36C2BA05BBA8900AULL, "ser210_witness6168172.cl"},
	{0x36C2EB0123698A7AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxb.mmh"},
	{0x36C2EF01236990B1ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxb.mao"},
	{0x36C2FD012369A8C0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxb.msh"},
	{0x36C3E81094DE9E6FULL, "hlac_ht_mansion2_818_2y.dds"},
	{0x36C3EE4F20778AEBULL, "6083576_m.fxe"},
	{0x36C4099F025E1274ULL, "zz_tbe_debug6189389.cl"},
	{0x36C4CA5AFC928A22ULL, "zz_fol_debug6062215.cl"},
	{0x36C50708D86C3253ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_left_hand.gad"},
	{0x36C5653862EF499CULL, "gen00fl_merrill6121125.cl"},
	{0x36C5CE88CE600559ULL, "tbe100st_magic_a3.utm"},
	{0x36C625D11CC0292DULL, "vxw_sundermount_32e_0_0.vlm"},
	{0x36C646DDFF86E936ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_basic_2.mao"},
	{0x36C651EC65BA1794ULL, "vxdr_primevaltomb_7t3_0_0.vlm"},
	{0x36C654DAEDDE4CD7ULL, "vxlt_gallowstemplar_5aw_0_0.vlm"},
	{0x36C68C94A13F8922ULL, "abi_m_inferno.wbd"},
	{0x36C6D04FA8A82E7DULL, "6087515_f.fxe"},
	{0x36C6FD6BDCE5AD2BULL, "vxdr_primevaltomb_36r_0_0.vlm"},
	{0x36C78129503BF3D8ULL, "gen00fl_varric6081256.cl"},
	{0x36C78EC99A82184BULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_exit.evt"},
	{0x36C7B9D1B926BCD0ULL, "442355939.ogg"},
	{0x36C7D72B3CDDCAA4ULL, "zz_fol_debug6103804.cl"},
	{0x36C816081EF705A2ULL, "hla_ht_0_hightown_n_131072_0y.dds"},
	{0x36C88270CDB983FBULL, "lt_0_lowtown_a2_36_0_ll.mmh"},
	{0x36C89FEEABA74950ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_6y.dds"},
	{0x36C95CA973F5524AULL, "vxlt_0_lowtown_docks_b81_0_1.vlm"},
	{0x36C9636A644AD2A1ULL, "hlac_lt_0_lowtown_a2_65538_3z.dds"},
	{0x36C9ABBB899554E9ULL, "ca_cave_2194.anb"},
	{0x36CA15BF485CDD7DULL, "379388429.ogg"},
	{0x36CA16DBAAF4441FULL, "llac_lt_backalley_g_0_10.dds"},
	{0x36CA296A345FEF48ULL, "6144869_m.fxe"},
	{0x36CA48FBBF698589ULL, "6059013_f.fxe"},
	{0x36CA876D44F49F32ULL, "zz_mag_debug6024439.cl"},
	{0x36CAB0C99307F317ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tal_bfft.evt"},
	{0x36CAEBADE2A8FA71ULL, "vxlt_undercityencounter_13h_0_1.vlm"},
	{0x36CAFF02B08ED476ULL, "lt_0_lowtown_a3_n_c_6_0.phy"},
	{0x36CB0408D87098B8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_left_hand.evt"},
	{0x36CB18D75743288AULL, "757477274.ogg"},
	{0x36CBAB0FE4A07B67ULL, "6072312_m.fxe"},
	{0x36CBCF6FA082F7FDULL, "vxw_blightlands_91_0_0.vlm"},
	{0x36CBD9026D7578C4ULL, "mrl150_merrill6078523.cl"},
	{0x36CBF021893BD0C4ULL, "vxlt_gallowstemplar_n_5t0_0_0.vlm"},
	{0x36CC8BA365E1F248ULL, "01_fire_f2_blue.dds"},
	{0x36CCA796373E7D41ULL, "face_face_p22.psh.fxo.dx11"},
	{0x36CCB99CE6CA00D3ULL, "vxht_0_hightown_a3_aum_0_0.vlm"},
	{0x36CCCF82375432EDULL, "6181903_f.fxe"},
	{0x36CCD7FE6E33558BULL, "vxlt_gallowstemplar_f_47c_0_0.vlm"},
	{0x36CD2FE2E925D83EULL, "627244216.ogg"},
	{0x36CDB2312CCAB42DULL, "zz_approval_debug6003653.cl"},
	{0x36CE2E30022B42A8ULL, "6043144_m.fxe"},
	{0x36CE35CA52A512B7ULL, "mrl310_merrill_pt2.stg"},
	{0x36CE80955331411FULL, "hlac_lt_0_lowtown_docks_-131074_2x.dds"},
	{0x36CEA7C99A88610CULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_exit.gad"},
	{0x36CEF95CBC611B09ULL, "vxht_0_hightown_a3_8pq_0_0.vlm"},
	{0x36CF568603D5BD17ULL, "830621971.ogg"},
	{0x36D05397F0D97553ULL, "6137037_m.fxe"},
	{0x36D0DCBEC3FC5E73ULL, "6110727_f.fxe"},
	{0x36D11FCE9D137548ULL, "840502782.ogg"},
	{0x36D12181B8B9DCBBULL, "6173514_m.fxe"},
	{0x36D14CF2CD338C2EULL, "vxw_woundedcoast_l7_0_0.vlm"},
	{0x36D16086288A727DULL, "305939163.ogg"},
	{0x36D18266507B2DBBULL, "vxlt_darktown_5td_0_0.vlm"},
	{0x36D1EF175AC902FBULL, "6201600_m.fxe"},
	{0x36D29D77C0B183E3ULL, "403206174.ogg"},
	{0x36D3095FBAF675FFULL, "502502319.ogg"},
	{0x36D31E599E8CF8EBULL, "338794852.ogg"},
	{0x36D33D771AB3F468ULL, "l_ser130ar_emeric_encounter.bnk"},
	{0x36D371689E8D8240ULL, "vxlt_shop_31_0_0.vlm"},
	{0x36D3AAE6B1C9DF0BULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.mmh"},
	{0x36D3BF90E07C996DULL, "6095151_m.fxe"},
	{0x36D4861EA1C652FCULL, "6208037_m.fxe"},
	{0x36D4A6C99A8CCB0AULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_exit.ani"},
	{0x36D4BF44EC23039EULL, "vxht_0_hightown_a3_bwy_0_0.vlm"},
	{0x36D4E802B0970117ULL, "lt_0_lowtown_a3_n_c_6_0.msh"},
	{0x36D4FA02B097207AULL, "lt_0_lowtown_a3_n_c_6_0.mao"},
	{0x36D4FE02B0972689ULL, "lt_0_lowtown_a3_n_c_6_0.mmh"},
	{0x36D53B51D07020B5ULL, "6176308_m.fxe"},
	{0x36D59B87C01B6885ULL, "6133534_m.fxe"},
	{0x36D62CB1341C9B39ULL, "hlac_lt_gallowstemplar_n_2391_6y.dds"},
	{0x36D64BAA63870FF4ULL, "6005757_m.fxe"},
	{0x36D64D329DB4F7C0ULL, "982221596.ogg"},
	{0x36D6549D1785B881ULL, "vxlt_0_lowtown_a2_n_e8j_0_0.vlm"},
	{0x36D6BDCD73E6762BULL, "6126798_m.fxe"},
	{0x36D6F9E480EB280EULL, "6125149_m.fxe"},
	{0x36D7038747CB4E49ULL, "vxlt_gallowstemplar_43b_0_0.vlm"},
	{0x36D70C3E56BD4567ULL, "6176746_m.fxe"},
	{0x36D72F52B691EB0BULL, "zz_followers_banter6158673.cl"},
	{0x36D79DCE0245A020ULL, "dae000ip_party_picker.ncs"},
	{0x36D7AB4FF7302FE3ULL, "vxht_mansion2_1va_0_0.vlm"},
	{0x36D820A6627F6DCCULL, "vxht_0_hightown_a3_apy_0_0.vlm"},
	{0x36D8F0D8DB1C4504ULL, "vxht_mansion_n_4gw_0_0.vlm"},
	{0x36D9539100844FCAULL, "14110777.ogg"},
	{0x36D956E3AFE0AE8CULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hc1a.msh"},
	{0x36D95CE3AFE0B97EULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hc1a.mmh"},
	{0x36D95EB00EEF70AEULL, "vxht_0_hightown_a2_c5g_0_0.vlm"},
	{0x36D96320A86FD13DULL, "267908908.ogg"},
	{0x36D9E925F6989C8AULL, "vfx_vfx_p5.vsh.fxo.dx11"},
	{0x36DA17515CFF2283ULL, "860265648.ogg"},
	{0x36DAE09F41E4431BULL, "1066243535.ogg"},
	{0x36DB0487F6A2963BULL, "156495755.ogg"},
	{0x36DB2B79D8C01953ULL, "vxht_0_hightown_a3_arl_0_0.vlm"},
	{0x36DB87284083FACFULL, "dae000_mother6127191.cl"},
	{0x36DC83F5513B001DULL, "ser270_emeric6149383.cl"},
	{0x36DCF7F6478FEA27ULL, "w_bonepitmining_a2_31_0.rml"},
	{0x36DD0BD755EBAAA0ULL, "835390534.ogg"},
	{0x36DD3BF35C3D2AC2ULL, "6107732_m.fxe"},
	{0x36DD876BBAF034A2ULL, "art\\levels\\ztools\\default\\generic\\0wo_basic.phy"},
	{0x36DDA10995CC9001ULL, "vxca_caverns1_mines_j54_0_0.vlm"},
	{0x36DDBC2DD80B333DULL, "hlac_ht_0_hightown_n_-131073_0y.dds"},
	{0x36DDCD4CF88C77B5ULL, "441271587.ogg"},
	{0x36DDD90E8F261CF3ULL, "mag121_viveka6156582.cl"},
	{0x36DDE2318E5F1788ULL, "587674865.ogg"},
	{0x36DE00FF9F9F884FULL, "vxlt_blackemporium_11j_0_0.vlm"},
	{0x36DE7077E5650FBCULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbed_exit.ani"},
	{0x36DEC4067A872833ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_flourish_additive.evt"},
	{0x36DF3AE6F561C730ULL, "vxht_mansion2_34n_0_0.vlm"},
	{0x36E0007093AA62A6ULL, "hlac_lt_0_lowtown_a2_n_c_65536_12z.dds"},
	{0x36E0A4FDF6776B19ULL, "97500004.ogg"},
	{0x36E0E4AA5904CC64ULL, "w_encounter3_7_0.msh"},
	{0x36E0F6AA5904EACDULL, "w_encounter3_7_0.mao"},
	{0x36E0FAAA5904F106ULL, "w_encounter3_7_0.mmh"},
	{0x36E119D2AEAF24A4ULL, "mag370_meredith6046541.cl"},
	{0x36E12AB5305025F8ULL, "450071304.ogg"},
	{0x36E14C2001A8478FULL, "vxca_caverns1_mines_fmn_0_0.vlm"},
	{0x36E1702E69742D8FULL, "gen00fl_aveline6121910.cl"},
	{0x36E1808CFD27865DULL, "vxht_0_hightown_bpr_0_0.vlm"},
	{0x36E18BE6B1D5E870ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.anb"},
	{0x36E19A145BAE25D5ULL, "123588033.ogg"},
	{0x36E2B38FD37EE74CULL, "gen00fl_anders6094087.cl"},
	{0x36E33894D1CA7DD7ULL, "vxlt_gallowstemplar_5wq_0_0.vlm"},
	{0x36E372E21F19C9E8ULL, "vrc221_varric6043876.cl"},
	{0x36E45EE1B019526AULL, "6133559_m.fxe"},
	{0x36E473F05794AA4EULL, "384880489.ogg"},
	{0x36E55A264D6A458CULL, "hlac_lt_undercityencounter_-1_11x.dds"},
	{0x36E57D77E56B44FEULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbed_exit.gad"},
	{0x36E5946B6C04B1D4ULL, "54907472.ogg"},
	{0x36E59BB947DC787DULL, "hlac_lt_gallowstemplar_n_2384_7x.dds"},
	{0x36E5BD067A8D3B74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_flourish_additive.gad"},
	{0x36E68CD78B3A0A50ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_3.dds"},
	{0x36E710D9771FC73CULL, "181835390.ogg"},
	{0x36E76B138C778EA9ULL, "6109770_f.fxe"},
	{0x36E7D4FD9A65E9BDULL, "267618908.ogg"},
	{0x36E7FD5592B8FB95ULL, "vxdr_primevaltomb_56b_0_0.vlm"},
	{0x36E80A012388DA85ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxb.phy"},
	{0x36E8386EEE2B62FAULL, "ts_books_02.gda"},
	{0x36E8A20EE9FDDAD0ULL, "6030627_m.fxe"},
	{0x36E8D2ADEF383D93ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_wall_0.pwk"},
	{0x36E960E7E08A9AFBULL, "vxw_sundermount_34u_0_0.vlm"},
	{0x36E9ADE132D62BF7ULL, "art\\levels\\races\\design\\_textures\\plc_coinpile01_s.dds"},
	{0x36E9C37FD243A4CCULL, "6070810_m.fxe"},
	{0x36EA6946169BC180ULL, "149251765.ogg"},
	{0x36EABB409DDC7BDAULL, "gencr_slaver_2w.utc"},
	{0x36EB09206D2B4A83ULL, "vxlt_0_lowtown_a2_n_gnh_0_0.vlm"},
	{0x36EB0D06F1A5413BULL, "vxlt_gallowscourtyard_n_769_0_0.vlm"},
	{0x36EB268B15D48B25ULL, "hlac_w_blightlands_262140_0y.dds"},
	{0x36EB3CA796A94613ULL, "6130014_m.fxe"},
	{0x36EB48F0FF88AD5BULL, "6081695_m.fxe"},
	{0x36EB4B32A241F0F0ULL, "557010844.ogg"},
	{0x36EB4FCDBEBF88B0ULL, "vxlt_warehouse_qp_0_0.vlm"},
	{0x36EBA02E46E71FCFULL, "6141142_m.fxe"},
	{0x36EBD419CB37FC14ULL, "vxca_cave_23n_0_0.vlm"},
	{0x36EC2C7018B77A18ULL, "qun221cr_qunari_assassin.utc"},
	{0x36EC5877E571241DULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbed_exit.evt"},
	{0x36EC5F9FDFDDF7ADULL, "llac_lt_0_lowtown_a2_n_-65536_7.dds"},
	{0x36EC8E12C76117F3ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_rnf.evt"},
	{0x36ECA00321587567ULL, "gen_im_npc_elf_arm_hlm_dalb.uti"},
	{0x36ECBF6DA254DFF2ULL, "genip_station_runes.utp"},
	{0x36ECCC067A937352ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_flourish_additive.ani"},
	{0x36ED11B382F88D09ULL, "art\\levels\\races\\design\\trap_floorlever_12.pwk"},
	{0x36ED4E099B21FC58ULL, "dae000_denier.cnv"},
	{0x36ED8BD5F69BCA4FULL, "vxw_encounter3_4d_0_0.vlm"},
	{0x36EDC8FA7C9283C8ULL, "hlac_lt_undercitydungeon_-1_0y.dds"},
	{0x36EEA7C4A66B6352ULL, "ico_2h_sword_1.dds"},
	{0x36EEAABC10FEE078ULL, "static_static_p23.psh.fxo"},
	{0x36EEC77A852A202CULL, "6103741_m.fxe"},
	{0x36EEC82DD5679174ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock2_sa.dds"},
	{0x36EF2FA3784D5209ULL, "gen00fl_fenris6128051.cl"},
	{0x36EFA5976F386780ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk_d.dds"},
	{0x36F01A616EFD86A1ULL, "404102970.ogg"},
	{0x36F081780C52B4E2ULL, "6103920_f.fxe"},
	{0x36F102AFEFE039ABULL, "llac_ht_0_hightown_n_-65537_4.dds"},
	{0x36F130099B25B260ULL, "dae000_denier.bnk"},
	{0x36F1A6A3E8C1917DULL, "gen00fl_fenris6110917.cl"},
	{0x36F1FAAA591364CDULL, "w_encounter3_7_0.rml"},
	{0x36F2307678BD2451ULL, "dae000_pc_home_ambients6208667.cl"},
	{0x36F23500FCA7FE64ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_4z.dds"},
	{0x36F270FE52F42794ULL, "811310121.ogg"},
	{0x36F29C07C8D8A130ULL, "781146397.ogg"},
	{0x36F2D251A2D55EC4ULL, "follower_barks6153535.cl"},
	{0x36F33C55834AA0EFULL, "vxht_keep_n_40u_0_0.vlm"},
	{0x36F38712C7672A34ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_rnf.gad"},
	{0x36F3B39EFFFE6EE9ULL, "vxlt_gallowstemplar_f_4n6_0_0.vlm"},
	{0x36F41F23425279C5ULL, "6166217_m.fxe"},
	{0x36F4CC7B9B824612ULL, "zz_followers_banter6159273.cl"},
	{0x36F4F7F647A47C44ULL, "w_bonepitmining_a2_31_0.mmh"},
	{0x36F4FBF647A48297ULL, "w_bonepitmining_a2_31_0.mao"},
	{0x36F50DF647A4A1E6ULL, "w_bonepitmining_a2_31_0.msh"},
	{0x36F531FA87082BD1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_over_heart.ani"},
	{0x36F5438B08802E02ULL, "vxca_caverns1_mines_ehk_0_0.vlm"},
	{0x36F548D34266687DULL, "ave200_aveline_bar6116198.cl"},
	{0x36F55A32B598CEE1ULL, "262397231.ogg"},
	{0x36F5DCF07E6DA729ULL, "6102670_f.fxe"},
	{0x36F70E6228B7499BULL, "art\\characters\\playercharacter\\humanmale\\hlf_medb.msh"},
	{0x36F71C71A9525239ULL, "vxlt_0_lowtown_docks_n_9ra_0_0.vlm"},
	{0x36F7206228B768E6ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medb.mao"},
	{0x36F72C6228B77C85ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medb.mmh"},
	{0x36F7D2C4D280B37BULL, "vxw_sundermount_33t_0_0.vlm"},
	{0x36F7E27E0BEC043CULL, "6103194_m.fxe"},
	{0x36F82F771AD2F812ULL, "l_ser130ar_emeric_encounter.wbd"},
	{0x36F865E8472AAE6FULL, "698981042.ogg"},
	{0x36F8807DA0F6582FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wic_0d.dds"},
	{0x36F928AFA968D908ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_dummy_n.dds"},
	{0x36F96DF06A88116AULL, "vxlt_0_lowtown_a2_gys_0_0.vlm"},
	{0x36F99612C76BAF12ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_rnf.ani"},
	{0x36F9BCC75324F9E2ULL, "hlac_ht_playershouse_n_855_15y.dds"},
	{0x36F9F9E16FB3B3A0ULL, "6054508_m.fxe"},
	{0x36FA0B452965526CULL, "and100_anders6095899.cl"},
	{0x36FA0F18F57B0EC4ULL, "88922337.ogg"},
	{0x36FB3CFA870CAA73ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_over_heart.gad"},
	{0x36FB7E23256DB015ULL, "vxw_bonepitdestroyed_a3_164_0_0.vlm"},
	{0x36FBBA1A636D448AULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graved.mao"},
	{0x36FBBE1A636D4BD9ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graved.mmh"},
	{0x36FBC81A636D5CC7ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graved.msh"},
	{0x36FBFFC50AC008C6ULL, "vxca_caverns1_mines_ics_0_1.vlm"},
	{0x36FC2C68FED95C22ULL, "hlac_lt_hangedman_-1_23x.dds"},
	{0x36FC65F950DAA22FULL, "hlac_w_encounter2_327680_0y.dds"},
	{0x36FC9A6332E1CE16ULL, "vxw_sundermount_2kd_0_0.vlm"},
	{0x36FCC8155ECF29C4ULL, "657793883.ogg"},
	{0x36FCE10CB251043BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm03.msh"},
	{0x36FCEB085EBA6427ULL, "gen00fl_carver6104365.cl"},
	{0x36FCFF0CB2513725ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm03.mmh"},
	{0x36FD0B01339FAEB1ULL, "dre000_dougal6131136.cl"},
	{0x36FD1C9902F4F8F8ULL, "static_static_p52.psh.fxo"},
	{0x36FD2F7BBC1CD31BULL, "mag291_torpor6027044.cl"},
	{0x36FD601C5F76FA74ULL, "912934075.ogg"},
	{0x36FD822C35D80FEFULL, "6078704_f.fxe"},
	{0x36FE24B15F57B948ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_0x.dds"},
	{0x36FE3C45FD7229EFULL, "vxca_cave_17s_0_1.vlm"},
	{0x36FF3C2DF2282692ULL, "6127654_f.fxe"},
	{0x36FF7FA08B529D48ULL, "123433536.ogg"},
	{0x3700043A322E8A60ULL, "vxlt_0_lowtown_docks_a3_czb_0_0.vlm"},
	{0x3700106895343ED6ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangc.mao"},
	{0x37001C6895345375ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangc.mmh"},
	{0x37001E68953456EBULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangc.msh"},
	{0x37006AC72C0B4CECULL, "vxw_bonepitdestroyed_a3_1wk_0_1.vlm"},
	{0x3700A61FB74F1D40ULL, "qcr000_arishok6059336.cl"},
	{0x3701A0A72C4DE5B8ULL, "741182586.ogg"},
	{0x3701B9748A30F438ULL, "hlac_w_encounter2_camp_-65531_0y.dds"},
	{0x370239FA8712C3D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_over_heart.evt"},
	{0x37026C0C51129354ULL, "173213250.ogg"},
	{0x37026ECD17263020ULL, "vxht_0_hightown_a3_b86_0_0.vlm"},
	{0x37029288ADB344C9ULL, "cod_bks_lite_tombs_last_1.utp"},
	{0x3702964D37D968C2ULL, "190317034.ogg"},
	{0x3702A5B01D7D0D24ULL, "957785425.ogg"},
	{0x370300E6CD127963ULL, "vxht_mansion2_1gb_0_0.vlm"},
	{0x370311BA268EA8EEULL, "vxca_caverns1_mines_gjk_0_0.vlm"},
	{0x37034EA18CDC69CCULL, "6002802_m.fxe"},
	{0x3703626BBB10C556ULL, "art\\levels\\ztools\\default\\generic\\0wo_basic.mao"},
	{0x37036E6BBB10D9F5ULL, "art\\levels\\ztools\\default\\generic\\0wo_basic.mmh"},
	{0x3703706BBB10DD6BULL, "art\\levels\\ztools\\default\\generic\\0wo_basic.msh"},
	{0x370382A7A09A05B4ULL, "6144915_m.fxe"},
	{0x37051DF61D2FFC9DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexd.phy"},
	{0x37052556D2A351D6ULL, "497988247.ogg"},
	{0x37055E72CA335A12ULL, "vxlt_undercitydungeon_17w_0_0.vlm"},
	{0x3705633AA121B0C3ULL, "431251696.ogg"},
	{0x3705C805333221BAULL, "gen00fl_fenris6126610.cl"},
	{0x370618829803BB37ULL, "vxht_0_hightown_n_a7u_0_0.vlm"},
	{0x3706278E0F5A2EC9ULL, "hlac_w_bonepitdestroyed_a3_5_0z.dds"},
	{0x37065E0F29D9FA36ULL, "str_wispy.mao"},
	{0x37075A59183833DAULL, "402347754.ogg"},
	{0x37080CB6B2C149CFULL, "6166008_m.fxe"},
	{0x37081591C85AE578ULL, "hlac_lt_0_lowtown_n_65534_3x.dds"},
	{0x3708B9BDEB6028D9ULL, "6036533_f.fxe"},
	{0x37092784D61482B9ULL, "575189703.ogg"},
	{0x37099DCCAC94D824ULL, "vxca_cave_18y_0_1.vlm"},
	{0x370A0E8915C8A9D6ULL, "6103394_m.fxe"},
	{0x370A1897BFFE9434ULL, "vxlt_gallowscourtyard_7at_0_0.vlm"},
	{0x370A1941E8883115ULL, "vxw_bonepit_xe_0_0.vlm"},
	{0x370A38566D348321ULL, "6128920_f.fxe"},
	{0x370A5671C2BFC7F4ULL, "vxht_chantry_n_1yl_0_0.vlm"},
	{0x370A932BF1D2CDAAULL, "vxw_bonepit_1r3_0_0.vlm"},
	{0x370B0C4024D7E696ULL, "vxw_sundermount_2le_0_0.vlm"},
	{0x370C248C0C952F74ULL, "308182231.ogg"},
	{0x370C7E05B4BDA3C0ULL, "store_dae310_access.utm"},
	{0x370C8B35ADC69558ULL, "336166061.ogg"},
	{0x370C8E16472BDE08ULL, "hlac_ht_brothel_n_3367_0x.dds"},
	{0x370C952D877D5755ULL, "vxht_mansion_n_6k0_0_0.vlm"},
	{0x370D20A470A9C918ULL, "6167163_m.fxe"},
	{0x370D92A1CA13C890ULL, "gen00fl_merrill6116000.cl"},
	{0x370D9A94A17BA3D8ULL, "abi_m_inferno.bnk"},
	{0x370D9ACB7AEAC692ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v1.gad"},
	{0x370DE4002FD433CBULL, "141650912.ogg"},
	{0x370E2CB5254AD8F0ULL, "6097866_m.fxe"},
	{0x370E319269ADF3AEULL, "vxw_bonepit_1p1_0_0.vlm"},
	{0x370E80693119458BULL, "hlac_lt_undercityencounter_-1_34z.dds"},
	{0x370EEE7D0BAAF654ULL, "6165110_m.fxe"},
	{0x370F10ACE565A8B9ULL, "6126943_m.fxe"},
	{0x370F23851211C326ULL, "vxw_woundedcoast_17o_0_0.vlm"},
	{0x370F586411CA8C7BULL, "6128309_m.fxe"},
	{0x370F7B7CD56B4889ULL, "335649498.ogg"},
	{0x370FDC6B81F842B5ULL, "verticalblur7.vsh.fxo"},
	{0x37100C42C52C8CB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_mw.anb"},
	{0x37100ED02594F3A1ULL, "vxht_mansion2_2um_0_0.vlm"},
	{0x37117C1892BD2D46ULL, "90263460.ogg"},
	{0x3711D8252B427417ULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf_p_rf.gad"},
	{0x3711D998B6E33CC2ULL, "m_and121ar_chantry.wbd"},
	{0x3711FF2B0A859E86ULL, "gen00fl_fenris6126711.cl"},
	{0x37120A64D31590DAULL, "vxca_cavern1_haunted_ija_0_0.vlm"},
	{0x37125E3F41879C6FULL, "vxw_woundedcoast_oc_0_0.vlm"},
	{0x3712E08E5B5EC761ULL, "456072782.ogg"},
	{0x3713DC0B5C6350ADULL, "6156764_m.fxe"},
	{0x3713DE9317C154DAULL, "lt_0_lowtown_docks_n_247_0.msh"},
	{0x3713EC9317C16C13ULL, "lt_0_lowtown_docks_n_247_0.mao"},
	{0x3713F09317C17320ULL, "lt_0_lowtown_docks_n_247_0.mmh"},
	{0x371416A3C26B0027ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_spear_movement.anb"},
	{0x37146E53D0E9F5C4ULL, "6055433_m.fxe"},
	{0x37149DCB7AF0EAC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v1.evt"},
	{0x3714C005AB4432D4ULL, "llac_ht_0_hightown_a3_-65538_0.dds"},
	{0x371528C71EAC517AULL, "dae011_corff_sta6142579.cl"},
	{0x3715B18F2447E847ULL, "mcr310_argument6075058.cl"},
	{0x371612F61D3E5D09ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexd.mao"},
	{0x371618D611F7408EULL, "417802174.ogg"},
	{0x37161EF61D3E71A2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexd.mmh"},
	{0x371620F61D3E7558ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexd.msh"},
	{0x3716257EDF04D875ULL, "vxht_0_hightown_a2k_0_0.vlm"},
	{0x37172758C7E12679ULL, "hlac_w_woundedcoast_4_0z.dds"},
	{0x37173AD91A05D9C9ULL, "6180495_m.fxe"},
	{0x37177F80937E428BULL, "vxht_mansion2_3dq_0_0.vlm"},
	{0x37178FC0C9F99903ULL, "955041803.ogg"},
	{0x371793338110741BULL, "vxca_caverns1_mines_cnj_0_0.vlm"},
	{0x3717ECDF0C969F3AULL, "6096009_m.fxe"},
	{0x3717F202CA5FF083ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv2.ani"},
	{0x37184C40C48E38CFULL, "hlac_ht_keep_n_3736_2z.dds"},
	{0x3718A73879092E4CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenc.msh"},
	{0x3718AD387909383EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenc.mmh"},
	{0x3718B93879094CD5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenc.mao"},
	{0x3718CC1766AA059CULL, "6093576-deadman.cut"},
	{0x3718ED252B48B6A4ULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf_p_rf.evt"},
	{0x371923BDE15E1EE8ULL, "353971369.ogg"},
	{0x3719418197732AE0ULL, "art\\vfx\\materials\\vfx_glyph_damage.mao"},
	{0x37194EF740373ADAULL, "vxca_caverns2prc_9vz_0_0.vlm"},
	{0x3719D50D92CC8295ULL, "mcr370_varric6208270.cl"},
	{0x3719F4CF58D266F5ULL, "6101971_m.fxe"},
	{0x371A3291335711D1ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_spi_c.impa.ani"},
	{0x371A33E3B0180A19ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hc1a.phy"},
	{0x371AB2E793475FF2ULL, "hlac_lt_0_lowtown_a2_n_c_65535_0x.dds"},
	{0x371ABF2DB3DC6467ULL, "vxlt_0_lowtown_e8g_0_0.vlm"},
	{0x371BDC1BABC95051ULL, "vxdr_primevaltomb_3lv_0_0.vlm"},
	{0x371C1FE7D6145792ULL, "6129854_f.fxe"},
	{0x371C5A086A60BACBULL, "35132883.ogg"},
	{0x371CA36F5BA40C2EULL, "vxw_bonepit_1u2_0_0.vlm"},
	{0x371D8B019C4D1280ULL, "303708359.ogg"},
	{0x371D95620407DBA6ULL, "vxw_woundedcoast_12p_0_0.vlm"},
	{0x371E074A861EC8A1ULL, "gen00fl_merrill6150694.cl"},
	{0x371E97467EE5D72CULL, "vxlt_warehouse_n_qu_0_0.vlm"},
	{0x371EB0E518B08BD6ULL, "6096953_m.fxe"},
	{0x371EB256251130ACULL, "mcr310_sibling6076367.cl"},
	{0x371ED8CD21D632F9ULL, "vxca_caverns1_mines_fo6_0_0.vlm"},
	{0x371EE03DCCE3A7F3ULL, "dae200cr_viscount6087368.cl"},
	{0x371F14C3640A048FULL, "vxlt_0_lowtown_cox_0_0.vlm"},
	{0x371F1B7EE4748127ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_001a_0s.dds"},
	{0x371F61D3671521B4ULL, "dre000_varric6029459.cl"},
	{0x371F6FFA15645D0EULL, "zz_dae_debug6184893.cl"},
	{0x371F70864FA92891ULL, "hlac_ca_cavern1_haunted_24422_5x.dds"},
	{0x371F8842ABB5E9F7ULL, "6143113_m.fxe"},
	{0x37203D91335B9073ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_spi_c.impa.gad"},
	{0x3720927F7C24E8BDULL, "141154331.ogg"},
	{0x3720D01C337DB4EFULL, "vxw_woundedcoast_db_0_0.vlm"},
	{0x3721080005B1261DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_fakedoor.phy"},
	{0x372117317E9ABC89ULL, "vxht_mansion2b_1ab_0_0.vlm"},
	{0x372118BE90CD2599ULL, "vxht_0_hightown_a3_c8z_0_0.vlm"},
	{0x372139ACA2FE976CULL, "73413538.ogg"},
	{0x372146C4ACBA24B4ULL, "vxdr_0_deeproads_cug_0_1.vlm"},
	{0x3721A9A4A00B50D3ULL, "785377107.ogg"},
	{0x3721DCE7B7C14BFAULL, "vxlt_gallowstemplar_n_5rh_0_0.vlm"},
	{0x372243AC53F3B7BBULL, "453321278.ogg"},
	{0x37226373342382C9ULL, "vxca_caverns2_7bc_0_1.vlm"},
	{0x372285CB7AFCFF38ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v1.ani"},
	{0x372290D03A053C57ULL, "lgt311_shw_mer_mettin6156517.cl"},
	{0x3722AB51631981CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_talk.evt"},
	{0x3722C75D320D6396ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_basic.msh"},
	{0x3722D15D320D7494ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_basic.mmh"},
	{0x3722D55D320D7BA7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_basic.mao"},
	{0x372378F79B6BD658ULL, "vxht_mansion2_1uc_0_0.vlm"},
	{0x37239EAD2AC7CC72ULL, "hlac_w_blightlands_0_1z.dds"},
	{0x3724385E8C36DBA8ULL, "6084181_m.fxe"},
	{0x3724E6C9B9950484ULL, "dae000_arishok6074056.cl"},
	{0x37250A02CA6AA47AULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv2.evt"},
	{0x37257156BBFE9C0EULL, "sd.cut"},
	{0x37259BD7AB5AC20EULL, "seb000im_gft_bow.uti"},
	{0x372605252B53699DULL, "art\\characters\\creatures\\stonegolem\\animation\\rnf_p_rf.ani"},
	{0x37265ED219232565ULL, "6153165_m.fxe"},
	{0x3726FCFBCC746215ULL, "vxw_bonepit_nc_0_0.vlm"},
	{0x372724479BDD7ACBULL, "124302362.ogg"},
	{0x37273A913361A9D8ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_spi_c.impa.evt"},
	{0x3727A4D547DF7319ULL, "792254945.ogg"},
	{0x3727BBCB94A07582ULL, "126659340.ogg"},
	{0x3727D4C8EA897E39ULL, "214196505.ogg"},
	{0x3727E295F6AFD71AULL, "hlac_w_encounter2_262149_0z.dds"},
	{0x3727EFFA08C41796ULL, "vxw_sundermount_2zg_0_0.vlm"},
	{0x37282B7D2FA950D7ULL, "6132984_m.fxe"},
	{0x3728398C4B50EBC6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_heal.evt"},
	{0x37283D03662DDA95ULL, "vxca_cavern1_haunted_ii7_0_0.vlm"},
	{0x3728ED4886E100C6ULL, "mcr370st_anders_magecrisis.stg"},
	{0x372920F84173168AULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_hugetreasure_1.mao"},
	{0x37294D7853F99530ULL, "1069881351.ogg"},
	{0x3729A251631F9179ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_talk.gad"},
	{0x3729C6708FB9E286ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundrysmoke_lrg.mmh"},
	{0x372A5D14328C3716ULL, "6136112_m.fxe"},
	{0x372A8A299283A30AULL, "6017037_m.fxe"},
	{0x372B041E8BCD35C7ULL, "fen110_fenris6118099.cl"},
	{0x372B104E7419B09AULL, "gen00fl_fenris6127983.cl"},
	{0x372B1942386C73D3ULL, "356897467.ogg"},
	{0x372B5B7B1F593406ULL, "6028138_m.fxe"},
	{0x372B6359B806076AULL, "767369233.ogg"},
	{0x372B7DC26AC83623ULL, "6023901_m.fxe"},
	{0x372BE3F1A47C6991ULL, "gen00fl_sebastian6174079.cl"},
	{0x372C0902CA70C16DULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv2.gad"},
	{0x372C37BDD8DD074CULL, "6033600_m.fxe"},
	{0x372C9B2E965CAC12ULL, "one300_zevran6076953.cl"},
	{0x372CA5180A3B570DULL, "201676823.ogg"},
	{0x372CCCBA44BCD7F2ULL, "vxdr_primevalentrance_24t_0_0.vlm"},
	{0x372CFC70E49CEA65ULL, "vxw_bonepitdestroyed_a3_16d_0_0.vlm"},
	{0x372D66F56289BA79ULL, "follower_banter6172691.cl"},
	{0x372DCE732B34C4EFULL, "vxw_sundermount_1yx_0_0.vlm"},
	{0x372ECCB2FE1A36DDULL, "vxht_0_hightown_a2_cbd_0_1.vlm"},
	{0x372F41F92573CD6FULL, "vxw_woundedcoast_am_0_0.vlm"},
	{0x372F45058E1C7EB9ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf06.msh"},
	{0x372F4F058E1C8FA7ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf06.mmh"},
	{0x372F508C4B573111ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_heal.gad"},
	{0x372FB351632419C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_lean_forward_talk.ani"},
	{0x372FDB7BA77BC386ULL, "6088889_f.fxe"},
	{0x37302D1D098671FAULL, "aip_mercenarywarrior_base.gda"},
	{0x373037E21E5FAF5BULL, "w_bonepitdestroyed_a3_151_0.mmh"},
	{0x373043E21E5FC3F8ULL, "w_bonepitdestroyed_a3_151_0.mao"},
	{0x373055E21E5FE245ULL, "w_bonepitdestroyed_a3_151_0.msh"},
	{0x373074D1EE02D59DULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg103a_0n.dds"},
	{0x3730F0FCA248E1D5ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.anb"},
	{0x3731491CC84AB5D1ULL, "1072533811.ogg"},
	{0x373153613CC7F704ULL, "dre000_cut2d_ragstoriches6025109.cl"},
	{0x3731B36CD9BFD203ULL, "831903593.ogg"},
	{0x3731F539FBE5D51EULL, "hlac_lt_0_lowtown_n_65534_2y.dds"},
	{0x3731FD0005BF8689ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_fakedoor.mao"},
	{0x3732090005BF9B22ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_fakedoor.mmh"},
	{0x37320B0005BF9ED8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_fakedoor.msh"},
	{0x373333DDD669F258ULL, "681692914.ogg"},
	{0x3733506E1610856DULL, "vxlt_0_lowtown_docks_a3_n_d3d_0_0.vlm"},
	{0x37356ED8BE6A7A95ULL, "vxw_bonepit_ib_0_0.vlm"},
	{0x37361BB9596DCEF2ULL, "vxlt_gallowstemplar_2p6_0_0.vlm"},
	{0x3736518C4B5D51BFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_heal.ani"},
	{0x373691AC30202BD2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chac_d.dds"},
	{0x3736EB831E63010BULL, "vxw_encounter2_pr_0_1.vlm"},
	{0x37378ECBDB46CEDCULL, "vxlt_warehouse_ob_0_0.vlm"},
	{0x373832A19F50F41FULL, "411681431.ogg"},
	{0x3738DBF19FA0783AULL, "static_rendertodepth_p1.psh.fxo"},
	{0x3738E3CC76C8CFFFULL, "6196821_m.fxe"},
	{0x373926A37F9BD9BEULL, "288548708.ogg"},
	{0x3739462F607B3702ULL, "381410030.ogg"},
	{0x373991A2D78862C5ULL, "vxht_0_hightown_a3_cb0_0_0.vlm"},
	{0x3739D1529A9FF6ABULL, "553819178.ogg"},
	{0x373A791BE7F40CA6ULL, "vxw_woundedcoast_18r_0_0.vlm"},
	{0x373B4411FE2FF7B5ULL, "6147537_m.fxe"},
	{0x373B982472CEF0F8ULL, "469422797.ogg"},
	{0x373BBCC2B1DF76E6ULL, "937080665.ogg"},
	{0x373BE414F4018AF8ULL, "hlac_w_encounter2_camp_262145_0z.dds"},
	{0x373BFC0FC86B8422ULL, "29930421.ogg"},
	{0x373CBF4287210DFAULL, "vxdr_0_deeproads_aaz_0_0.vlm"},
	{0x373CCFC750D7FE4FULL, "llac_lt_0_lowtown_docks_65534_4.dds"},
	{0x373D61D13EC6BC30ULL, "w_blightlands_24_0.rml"},
	{0x373D855E58E80D01ULL, "hlac_lt_0_lowtown_docks_a3_-131075_3y.dds"},
	{0x373DF355E140976CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appa_0d.dds"},
	{0x373DFC7874B95699ULL, "vxht_0_hightown_a3_c6x_0_0.vlm"},
	{0x373E056228F33D32ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medb.phy"},
	{0x373E5598FA3CB21BULL, "6101523_f.fxe"},
	{0x373E64751D4FF90DULL, "vxlt_0_lowtown_exe_0_0.vlm"},
	{0x373E6FA12D990029ULL, "414594354.ogg"},
	{0x373E8C4A14231E9EULL, "hlac_lt_0_lowtown_a3_n_c_-2_2x.dds"},
	{0x373EC0A19BAD7CFAULL, "vxlt_gallowstemplar_n_5hj_0_0.vlm"},
	{0x373EE2BF85BA2F81ULL, "hlac_lt_undercityencounter_-1_4y.dds"},
	{0x373F3423B5E97925ULL, "vxlt_0_lowtown_docks_n_aov_0_0.vlm"},
	{0x373F9C2A891209BDULL, "hlac_lt_gallowstemplar_2394_1x.dds"},
	{0x373FF1312F15E973ULL, "vxht_chantry_n_28k_0_0.vlm"},
	{0x374028B4C4055F2CULL, "vxlt_darktown_4wc_0_0.vlm"},
	{0x374043B06EFCCEB9ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firemed.mmh"},
	{0x37405CB17F580758ULL, "vxht_mansion2_1ca_0_0.vlm"},
	{0x37408509BE16ECC5ULL, "vxlt_gallowstemplar_f_53h_0_0.vlm"},
	{0x3740FDE07D91C411ULL, "lgt110cr_gxa_vgk_cnspr_4.utc"},
	{0x374105F8C1D9FE8DULL, "gen_im_arm_cht_coma_t6.uti"},
	{0x374145571677E13FULL, "dae000_arishok6067703.cl"},
	{0x3741629569BFFA01ULL, "vxlt_0_lowtown_docks_n_apk_0_0.vlm"},
	{0x3741C0AB2F08AC3FULL, "vxlt_backalley_g_1kw_0_0.vlm"},
	{0x37426460DF736D20ULL, "follower_banter6166406.cl"},
	{0x37427569AED188CAULL, "6016180_m.fxe"},
	{0x37427F1067C77DE2ULL, "l_dae311ar_tavern.wbd"},
	{0x3742904A4EE755A7ULL, "mer221_sabin6080134.cl"},
	{0x3742976735D9F2D3ULL, "6025367_m.fxe"},
	{0x374326E0CD691EB9ULL, "6082018_m.fxe"},
	{0x374334722AC776C7ULL, "vxlt_undercityserialkiller_1c3_0_0.vlm"},
	{0x3743D80CB28CF7D2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm03.phy"},
	{0x37440660157CB5EEULL, "hlac_dr_primevaltomb_658_6y.dds"},
	{0x37445D2EE3A8AC5CULL, "vxlt_0_lowtown_a2_n_en1_0_0.vlm"},
	{0x3744B678A83D54BEULL, "6019112_m.fxe"},
	{0x374600A8CD3CE75CULL, "vxlt_warehouse_pc_0_0.vlm"},
	{0x37463F309BCA6C30ULL, "vxw_encounter3_6a_0_0.vlm"},
	{0x3746A2E50192020FULL, "6118694_m.fxe"},
	{0x3746BAEFFC3D2866ULL, "380928237.ogg"},
	{0x3746F8ADF402A9CEULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.impa.ani"},
	{0x3746F916EA138554ULL, "6136958_m.fxe"},
	{0x3747EFA507F32134ULL, "vxht_keep_n_4bx_0_0.vlm"},
	{0x374819A3EDAF7C3BULL, "6207066_m.fxe"},
	{0x3748586D90E6195CULL, "6097326_f.fxe"},
	{0x3748B7F6ADDDA825ULL, "6109702_m.fxe"},
	{0x3748D18719AA942FULL, "883913029.ogg"},
	{0x374909ECD62F5108ULL, "llac_lt_darktown_-1_77.dds"},
	{0x37491B77BCC9384CULL, "6039400_m.fxe"},
	{0x37497472EF541C34ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_ice_c.anb"},
	{0x3749D506F594A69FULL, "vxca_caverns1_fmo_0_0.vlm"},
	{0x374A0F07BAEC692FULL, "146165428.ogg"},
	{0x374A3024D673815BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\fireball.ani"},
	{0x374A75D938AC2D87ULL, "vxht_mansion_n_5ag_0_0.vlm"},
	{0x374AA4153D9C6E76ULL, "hlac_lt_0_lowtown_docks_-65540_1x.dds"},
	{0x374AA786623924C7ULL, "zz_approval_debug6003729.cl"},
	{0x374AF1C7B0489DACULL, "282895284.ogg"},
	{0x374B82EDDDDC1701ULL, "33471445.ogg"},
	{0x374BEFA5A5FDE898ULL, "vxlt_hangedman_n_26h_0_0.vlm"},
	{0x374C0AF3CAE591F3ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appear.evt"},
	{0x374C51D2E28FD52AULL, "art\\characters\\playercharacter\\humanmale\\eye_viscount.mao"},
	{0x374C6CC8547D0609ULL, "vxht_mansion2b_1rg_0_0.vlm"},
	{0x374CDB1F0EE46A9BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfallr.msh"},
	{0x374CED1F0EE489E6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfallr.mao"},
	{0x374CF91F0EE49D85ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfallr.mmh"},
	{0x374CFAE7B283594AULL, "vxht_0_hightown_n_c3w_0_1.vlm"},
	{0x374D327E8DA3957AULL, "vxlt_gallowstemplar_n_5ck_0_0.vlm"},
	{0x374D374267452B32ULL, "6173407_m.fxe"},
	{0x374D7C3149BE5DEAULL, "6117702_m.fxe"},
	{0x374DA3B10BDCC040ULL, "187784173.ogg"},
	{0x374DA89DF0CFB859ULL, "lt_0_lowtown_docks_a3_136_0.msh"},
	{0x374DB29DF0CFC947ULL, "lt_0_lowtown_docks_a3_136_0.mmh"},
	{0x374DB69DF0CFCF94ULL, "lt_0_lowtown_docks_a3_136_0.mao"},
	{0x374DE9B6FADD3FF0ULL, "6063753_m.fxe"},
	{0x374DFA920F0D68CBULL, "vxht_bartrand_n_6qc_0_0.vlm"},
	{0x374DFE6595485AB4ULL, "vxw_bonepitdestroyed_a3_1vq_0_0.vlm"},
	{0x374E2F915E37E0B7ULL, "6102657_m.fxe"},
	{0x374EE87FE5AE8A2FULL, "6066407_m.fxe"},
	{0x374F28AD67DF9C3BULL, "isa111_isabela6164604.cl"},
	{0x374F5733ED74DE9DULL, "hlac_lt_0_lowtown_a3_n_c_131070_6z.dds"},
	{0x374F5A6324733EA1ULL, "hlac_ht_keep_n_3731_3z.dds"},
	{0x374F98A9CEF63DB8ULL, "6085252_m.fxe"},
	{0x3750105A9A0CA5B6ULL, "vxlt_shop_5e_0_0.vlm"},
	{0x37503E49C86A73CDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic_4.mao"},
	{0x375044B203E02451ULL, "mcr310_argument6127698.cl"},
	{0x375084F7377860D0ULL, "6060389_m.fxe"},
	{0x3750D15B1DD93187ULL, "273945898.ogg"},
	{0x375109E80890124FULL, "vxlt_0_lowtown_a2_eja_0_0.vlm"},
	{0x3751376D9AEC0B6CULL, "hlac_w_bonepitmining_a2_262148_0y.dds"},
	{0x3751C853B330DFE3ULL, "scaleform_textcolor.mao"},
	{0x3751CFAC8E9EB340ULL, "6088794_m.fxe"},
	{0x3752265370D1C87FULL, "1016715822.ogg"},
	{0x37522A1748E82DE5ULL, "gencr_darkspawn_ogre_deepr.utc"},
	{0x375294A935BD1F24ULL, "6174040_m.fxe"},
	{0x3752D8572D06D76AULL, "6136927_m.fxe"},
	{0x375303F3CAEBA434ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appear.gad"},
	{0x37536ABC9EBDF54FULL, "vxht_keep_n_6ho_0_0.vlm"},
	{0x37539702B5D60CB2ULL, "vxlt_gallowstemplar_f_5v4_0_0.vlm"},
	{0x3753D2BE6EB2A048ULL, "345801580.ogg"},
	{0x3753DF983529E327ULL, "ht_keep_3735.rml"},
	{0x3753FA1707312CE7ULL, "6095755_m.fxe"},
	{0x375414542C2E6BAFULL, "statecnt_door_cage.gda"},
	{0x3754464C687A244FULL, "6144453_m.fxe"},
	{0x375453E2AF8ADFD3ULL, "fen110_intro.stg"},
	{0x37545EC555D04729ULL, "hlac_lt_warehouse_1_3z.dds"},
	{0x3754B10D8DC084B0ULL, "vxw_encounter3_3n_0_0.vlm"},
	{0x3754F0ADF40ED91FULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.impa.evt"},
	{0x375511DABC47BE98ULL, "vxlt_0_lowtown_a2_n_c_hrv_0_0.vlm"},
	{0x375516A7E3D8C9E4ULL, "vxdr_0_deeproads_f4x_0_0.vlm"},
	{0x375585AED24FC1DDULL, "gen00fl_aveline6129071.cl"},
	{0x3755A9682CB91EE4ULL, "209422689.ogg"},
	{0x375725886E8B53E1ULL, "6097603_m.fxe"},
	{0x37575372EF6021F7ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_ice_c.mmh"},
	{0x3757D73A34842040ULL, "vxlt_0_lowtown_docks_br8_0_0.vlm"},
	{0x37583824D67FCC52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\fireball.evt"},
	{0x37583CD75D671A39ULL, "vxht_0_hightown_ccb_0_0.vlm"},
	{0x37587B498BA39C65ULL, "vxw_bonepit_n3_0_0.vlm"},
	{0x37587CD35E9AE178ULL, "character_character_p92.psh.fxo.dx11.assembly"},
	{0x3758E798B71F5778ULL, "m_and121ar_chantry.bnk"},
	{0x3758F1FCA26A3516ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.mmh"},
	{0x375912F3CAF02912ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appear.ani"},
	{0x3759AC6BAFB8692BULL, "hightown_male_noble1_sta6170789.cl"},
	{0x3759B01370665694ULL, "gen00fl_aveline6120866.cl"},
	{0x3759BB51255D14E5ULL, "vxca_cavern1_haunted_iig_0_0.vlm"},
	{0x375A3433D35B6DDFULL, "1047602269.ogg"},
	{0x375A978FFB5616EFULL, "vxw_woundedcoast_pn_0_0.vlm"},
	{0x375ADEA546731E89ULL, "vxht_mansion2b_1uf_0_0.vlm"},
	{0x375AFFFB7DC29572ULL, "vxht_0_hightown_n_a4q_0_0.vlm"},
	{0x375B38058E41A3DCULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf06.phy"},
	{0x375B5CDA67AC07AEULL, "734147338.ogg"},
	{0x375BB0A93C87C502ULL, "hlac_ht_mansion2b_818_23y.dds"},
	{0x375BD1ADF414C308ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.impa.gad"},
	{0x375BD7922285DB4CULL, "vxht_0_hightown_c0n_0_0.vlm"},
	{0x375BDC3354D081A3ULL, "vxdr_0_deeproads_bup_0_0.vlm"},
	{0x375C8FB1861E9A88ULL, "732311450.ogg"},
	{0x375D3C72424F3E9FULL, "6033993_m.fxe"},
	{0x375D6AE8896F71FEULL, "vxw_sundermount_35g_0_0.vlm"},
	{0x375DB4BBCFF81C6FULL, "vxw_bonepit_1xd_0_0.vlm"},
	{0x375DEFAC9DE43D91ULL, "llac_ca_caverns1_mines_24424_10.dds"},
	{0x375E717058256C5AULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.cess.gad"},
	{0x375E888811429DA9ULL, "gen00fl_sebastian6180847.cl"},
	{0x375EBF1A63C1CAC6ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graved.phy"},
	{0x375F299C026DEDD8ULL, "6077693_m.fxe"},
	{0x375F5724D6861FE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\fireball.gad"},
	{0x375F8DF0B987FF64ULL, "204771335.ogg"},
	{0x375FE808C053A26DULL, "hlac_w_encounter3_-262140_0z.dds"},
	{0x376041898094CBC0ULL, "6083694_m.fxe"},
	{0x3760690578FA6692ULL, "dae000_gamlen6048508.cl"},
	{0x376126022FC83D57ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_stars.dds"},
	{0x3761345BFC58D656ULL, "vxht_bartrand_n_69b_0_0.vlm"},
	{0x37613F228072767BULL, "mag100_karras6034726.cl"},
	{0x37616D5E0FA747E7ULL, "455332162.ogg"},
	{0x37617C26396D531DULL, "dae300_alistair_king6133623.cl"},
	{0x3761BA56E982C7D7ULL, "gen00fl_sebastian6179953.cl"},
	{0x37629C551A868310ULL, "vxht_chantry_2tf_0_1.vlm"},
	{0x3762A9CB173B3C2AULL, "6040709_m.fxe"},
	{0x3762D57F5DCEDAE8ULL, "640579913.ogg"},
	{0x376344D38ECE529DULL, "643567729.ogg"},
	{0x376344FA4D2177F4ULL, "art\\vfx\\combat\\vfx_m_arc_modal_act_c.mmh"},
	{0x37649ACDA44C5487ULL, "vxlt_hangedman_1rg_0_0.vlm"},
	{0x3764F667B74C8EAFULL, "vxlt_hangedman_2i5_0_0.vlm"},
	{0x3764F88D10FDA1B9ULL, "vxht_0_hightown_a3_asy_0_0.vlm"},
	{0x376538B51AACE81CULL, "vxht_mansion2b_2ax_0_0.vlm"},
	{0x37655234EA856A75ULL, "ser120_ghyslain6118390.cl"},
	{0x37657470582B9071ULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.cess.evt"},
	{0x37658F014C4FADDBULL, "6038624_m.fxe"},
	{0x37662C44DFAD7BF0ULL, "vxlt_warehouse_n_11j_0_0.vlm"},
	{0x376639C2A1468C8FULL, "vxca_caverns1_mines_ei9_0_0.vlm"},
	{0x376648EA8DE85D4CULL, "ave100_ambush_16106193.cl"},
	{0x37678698019C28FEULL, "llac_w_encounter2_camp_262143_0.dds"},
	{0x37679BBA3679F897ULL, "vxdr_0_deeproads_fhc_0_0.vlm"},
	{0x376833E5C210F03DULL, "vxca_caverns1_mines_e8p_0_0.vlm"},
	{0x376840B06F1E1BAEULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firemed.anb"},
	{0x37686AB874B6E881ULL, "vxlt_warehouse_p2_0_0.vlm"},
	{0x3768D35F89EA1998ULL, "vxlt_hangedman_n_28f_0_0.vlm"},
	{0x37690E7DB2625FC1ULL, "hlac_lt_0_lowtown_docks_a3_131070_1y.dds"},
	{0x376913843CB4420AULL, "vxca_caverns1_mines_ghq_0_0.vlm"},
	{0x37691691AC965FCAULL, "w_sundermount_82_0.mao"},
	{0x37691A91AC966619ULL, "w_sundermount_82_0.mmh"},
	{0x376922792A323605ULL, "qun220cr_javaris6071515.cl"},
	{0x37692491AC967707ULL, "w_sundermount_82_0.msh"},
	{0x3769B114251528BDULL, "ss_issac_hawke_2.cut"},
	{0x3769B93A39BB47A5ULL, "vxdr_primevaltomb_3sa_0_2.vlm"},
	{0x3769BA0BF332428DULL, "vxlt_0_lowtown_eib_0_0.vlm"},
	{0x3769C32EB3BA3CE1ULL, "hlac_lt_0_lowtown_docks_a3_-196610_6x.dds"},
	{0x3769E86299270948ULL, "vxlt_gallowstemplar_n_5x8_0_0.vlm"},
	{0x376A28F35124218BULL, "6149991_m.fxe"},
	{0x376A496F147BF3CCULL, "vxht_0_hightown_c7o_0_0.vlm"},
	{0x376B358D35BE66B7ULL, "vxht_0_hightown_n_ajr_0_0.vlm"},
	{0x376B5C13389E92B7ULL, "art\\vfx\\combat\\vfx_gib_hm_meda_rblt.anb"},
	{0x376B6811BF98C08BULL, "6044901_f.fxe"},
	{0x376C917E4437B01FULL, "cr_spe_crack.dds"},
	{0x376CD1003FEFF54DULL, "6118159_m.fxe"},
	{0x376D2EE19B13EC6EULL, "6043591_m.fxe"},
	{0x376D4D66D144FF0CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chant2frame.msh"},
	{0x376D5366D14509FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chant2frame.mmh"},
	{0x376D59922D563526ULL, "6118835_m.fxe"},
	{0x376D5F66D1451D95ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chant2frame.mao"},
	{0x376F31689D34677CULL, "vxht_bartrand_n_676_0_0.vlm"},
	{0x376F49BFE47949EBULL, "hlac_w_encounter2_-65532_0y.dds"},
	{0x37701AD53BED1979ULL, "614132958.ogg"},
	{0x37701E30B3DA806FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_har_flma_0d.dds"},
	{0x3770D87E385184A1ULL, "vxlt_gallowscourtyard_7b4_0_0.vlm"},
	{0x3771513768BCF9C2ULL, "hlac_lt_undercitydungeon_-1_9x.dds"},
	{0x3771A4EBD248E8F9ULL, "mag360_alain6044621.cl"},
	{0x37728C7058364340ULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.cess.ani"},
	{0x3772B1032346DFC6ULL, "6149020_m.fxe"},
	{0x3772B1C1CFB88574ULL, "hlac_lt_0_lowtown_a2_n_65536_8y.dds"},
	{0x37730CAA96426D07ULL, "vxlt_hangedman_1yf_0_0.vlm"},
	{0x37732E6119A19694ULL, "590144913.ogg"},
	{0x377375C353FB7BA6ULL, "hlac_w_bonepitmining_a2_65539_0z.dds"},
	{0x37749E21D1A39470ULL, "vxlt_warehouse_n_16k_0_0.vlm"},
	{0x3774EB47864B0C4AULL, "t3_hey_bl4.tnt"},
	{0x3774F02A2D29AB0FULL, "dae000ip_crowd_ff6_talk6.utp"},
	{0x37755FE6746A501AULL, "gen00fl_varric6076058.cl"},
	{0x377667E4C9ABAE1FULL, "845193067.ogg"},
	{0x3776DC9566AD0101ULL, "vxlt_warehouse_u1_0_0.vlm"},
	{0x3776EF0907EF0968ULL, "ser270_cut_aftermath6162310.cl"},
	{0x37771A1B0BBBD6B8ULL, "vxw_sundermount_377_0_0.vlm"},
	{0x377736650C474ABAULL, "dae000_merchant_generic_hm6172632.cl"},
	{0x377773C3642369A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right.ani"},
	{0x377781FFDDD7775EULL, "59232438.ogg"},
	{0x3777CAA8B0DB98E4ULL, "gen00fl_merrill6107802.cl"},
	{0x3778090315B1EEAEULL, "dae000ip_crowd_em4_talk14.utp"},
	{0x377858E21E9D9C90ULL, "w_bonepitdestroyed_a3_151_0.phy"},
	{0x3778806D23AA4E28ULL, "zz_add_all_mage.ncs"},
	{0x377898005B0205CBULL, "hlac_lt_0_lowtown_131072_1y.dds"},
	{0x3779045FA5FD2BF6ULL, "493901622.ogg"},
	{0x37791E8AF3DC0DB9ULL, "6031364_m.fxe"},
	{0x37796BAD8576D907ULL, "vxlt_gallowsprison_n_5sb_0_0.vlm"},
	{0x3779E6B78CE47120ULL, "6216508_m.fxe"},
	{0x3779F2B73A98E1C2ULL, "6180148_m.fxe"},
	{0x377A4F7AC05C8FC8ULL, "233074098.ogg"},
	{0x377A9B9DF0F6907CULL, "lt_0_lowtown_docks_a3_136_0.phy"},
	{0x377BA8083095622FULL, "vxht_0_hightown_a2_bc9_0_0.vlm"},
	{0x377BCB6F1F4C4DA6ULL, "6117935_m.fxe"},
	{0x377CE4D1AE20840BULL, "vxlt_0_lowtown_a2_e7z_0_0.vlm"},
	{0x377D4366BBBE8E61ULL, "vxlt_gallowstemplar_n_5v3_0_0.vlm"},
	{0x377D43B3C116B50AULL, "6070489_m.fxe"},
	{0x377D9926223C94B6ULL, "vxca_cavern1_haunted_ihs_0_0.vlm"},
	{0x377DA4AE82A6DB27ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.anb"},
	{0x377E6FE51423750FULL, "hlac_ht_chantry_n_2534_32x.dds"},
	{0x377E8EC36429B62BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right.gad"},
	{0x377ED05D325BD18FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_basic.phy"},
	{0x377EEAD70592D749ULL, "hlac_lt_ruinedhouse_n_-1_7x.dds"},
	{0x377EF18406156A5AULL, "hlac_lt_gallowscourtyard_-1_4z.dds"},
	{0x377FFE7A0EDE86BEULL, "art\\vfx\\creatures\\fxc_shockwave_p.impa.evt"},
	{0x378021C98C537B56ULL, "6047515_m.fxe"},
	{0x3780498054514B28ULL, "183301822.ogg"},
	{0x3780576C78358EB6ULL, "510111067.ogg"},
	{0x378076B2E8CD9D92ULL, "vxw_encounter2_f8_0_0.vlm"},
	{0x3780E4DC8966F0D6ULL, "gen00fl_aveline6122754.cl"},
	{0x37817E8788388587ULL, "vxlt_hangedman_1da_0_0.vlm"},
	{0x37819C5CE74C2508ULL, "6134916_m.fxe"},
	{0x3781F4F484F5CB78ULL, "6170035_m.fxe"},
	{0x37822FEDBD23AFCDULL, "6103844_f.fxe"},
	{0x378271F9F98500A8ULL, "zz_seb_debug6186215.cl"},
	{0x3782ED741DD64EB2ULL, "hlac_ht_chantry_2534_44y.dds"},
	{0x37838935391A26C3ULL, "llac_lt_gallowstemplar_f_7198_7.dds"},
	{0x378393ACEFCBEE19ULL, "6103269_carver_sit_around.cut"},
	{0x37848BC3642E1C90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right.evt"},
	{0x378492AB02816BD4ULL, "zz_followers_banter6159015.cl"},
	{0x3784E62C356CA1EEULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appb_0s.dds"},
	{0x37851039B1FC31ABULL, "vxw_blightlands_10m_0_0.vlm"},
	{0x37853E1CC4672FACULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodburn_sa.dds"},
	{0x37858BF7FDB1EF38ULL, "vxw_sundermount_324_0_0.vlm"},
	{0x3785B8F94A6BC186ULL, "vxw_sundermount_32z_0_0.vlm"},
	{0x378643B251ABD0BBULL, "6090042_m.fxe"},
	{0x3786B7EC44FB746DULL, "6085950_m.fxe"},
	{0x3786CF3120BC55A4ULL, "928675117.ogg"},
	{0x3787157A0EE4CCE9ULL, "art\\vfx\\creatures\\fxc_shockwave_p.impa.gad"},
	{0x3787706BD6732593ULL, "follower_banter6155427.cl"},
	{0x378783E142855B71ULL, "6149294_m.fxe"},
	{0x37878A1CD2F58B6CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallows.msh"},
	{0x3787901CD2F5955EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallows.mmh"},
	{0x37879C1CD2F5A9F5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallows.mao"},
	{0x3787F56C59671348ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1sml.msh"},
	{0x3788076C596731D9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1sml.mao"},
	{0x3788136C59674672ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1sml.mmh"},
	{0x378822C0A65E70F3ULL, "art\\vfx\\combat\\vfx_a_disengage_aoe_p.impa.evt"},
	{0x37884CDC810C8436ULL, "102667822.ogg"},
	{0x378862057DD3D9ACULL, "vxlt_darktown_4nd_0_0.vlm"},
	{0x3788758C296ADAEEULL, "vxht_keep_61r_0_0.vlm"},
	{0x3788840947D4F32CULL, "hlac_ht_0_hightown_n_-65536_2z.dds"},
	{0x37890A52A5DA65EFULL, "vxw_bonepit_1ka_0_0.vlm"},
	{0x37890C7FA6193B34ULL, "6141804_m.fxe"},
	{0x37898D1068039798ULL, "l_dae311ar_tavern.bnk"},
	{0x3789D1C26C4AA4D6ULL, "6029137_m.fxe"},
	{0x378AB2EE77DF9306ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_twitch.ani"},
	{0x378B32AE9A91B55EULL, "dae000_mother6103060.cl"},
	{0x378B3E1712C04A9AULL, "378681265.ogg"},
	{0x378B47FA4D42CEABULL, "art\\vfx\\combat\\vfx_m_arc_modal_act_c.anb"},
	{0x378C2965AAE7A11FULL, "6055108_m.fxe"},
	{0x378C3BF72A46B15FULL, "6053591_m.fxe"},
	{0x378CF5A72BDD000BULL, "vxw_encounter3_cq_0_0.vlm"},
	{0x378D5E767DAE6EB2ULL, "vxlt_gallowstemplar_f_5j0_0_0.vlm"},
	{0x378D9204223EE398ULL, "mag311_nyssa6167209.cl"},
	{0x378DC48DFDA8D2D3ULL, "vxlt_0_lowtown_docks_arx_0_0.vlm"},
	{0x378DCEAF13099D8BULL, "vxca_caverns1_ih7_0_0.vlm"},
	{0x378DDAC4F67C1674ULL, "6071128_f.fxe"},
	{0x378DE67A0EEA9B37ULL, "art\\vfx\\creatures\\fxc_shockwave_p.impa.ani"},
	{0x378DF832B987F7C9ULL, "6153107_m.fxe"},
	{0x378E5B0C593BFC64ULL, "867873032.ogg"},
	{0x378ECAA06D8214E1ULL, "6048288_f.fxe"},
	{0x378ED093F287390BULL, "vxht_keep_6bj_0_0.vlm"},
	{0x378EE88FDAC3B612ULL, "vxw_encounter2_a9_0_0.vlm"},
	{0x378F1BC0A6648334ULL, "art\\vfx\\combat\\vfx_a_disengage_aoe_p.impa.gad"},
	{0x378F3579A7698E79ULL, "mag101_grace6053096.cl"},
	{0x378F42E3F8794A40ULL, "mcr370_meredith6122905.cl"},
	{0x378F46C781E59204ULL, "6116130_m.fxe"},
	{0x378F58BA208038DFULL, "352243587.ogg"},
	{0x378F7B13C7823A53ULL, "6149793_m.fxe"},
	{0x378FEE63AD2E7625ULL, "vxlt_0_lowtown_a2_n_erx_0_0.vlm"},
	{0x37903837BD2EF151ULL, "6104954_m.fxe"},
	{0x379039FB30AC36CBULL, "statecnt_door_secret.gda"},
	{0x3790620369164C12ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gouda_n.dds"},
	{0x3790DD3E5424DDB3ULL, "vxw_encounter2_camp_50_0_0.vlm"},
	{0x37910B541AA744F6ULL, "172353481.ogg"},
	{0x3791504F86954972ULL, "zz_dae_debug6029868.cl"},
	{0x3791592AD1F9C7C4ULL, "vxht_0_hightown_a3_c3v_0_0.vlm"},
	{0x379174D341C6280EULL, "909602143.ogg"},
	{0x37924FD13F0EDDADULL, "w_blightlands_24_0.msh"},
	{0x37925DD13F0EF5E0ULL, "w_blightlands_24_0.mao"},
	{0x379261D13F0EFC33ULL, "w_blightlands_24_0.mmh"},
	{0x37926F6D9B723C29ULL, "vxht_0_hightown_a3_b9p_0_0.vlm"},
	{0x3792B7596FC02348ULL, "hlac_dr_dungeon_3207_8z.dds"},
	{0x37932DC3E5E85FA5ULL, "hlac_w_blightlands_65533_0x.dds"},
	{0x37933950EA531D6BULL, "gen00fl_isabela6135653.cl"},
	{0x3793EA00A8FA0760ULL, "721834199.ogg"},
	{0x3793FDD4EFA807B8ULL, "vxw_sundermount_395_0_0.vlm"},
	{0x37942D687EFD2173ULL, "vxlt_0_lowtown_a3_n_c_f57_0_0.vlm"},
	{0x3794E29835617F3BULL, "ht_keep_3735.anb"},
	{0x3795001C218CE36FULL, "gen_im_trs_bks_bestiary.uti"},
	{0x379518CEC3465842ULL, "art\\characters\\creatures\\strider\\animation\\attkfr_leg.gad"},
	{0x37952AC0A6690812ULL, "art\\vfx\\combat\\vfx_a_disengage_aoe_p.impa.ani"},
	{0x379540F1AF6A0A24ULL, "hlac_ht_0_hightown_65538_0z.dds"},
	{0x379637B64F16D534ULL, "vxw_bonepitdestroyed_a3_1oz_0_0.vlm"},
	{0x3796466B6DE26899ULL, "vxht_keep_4ao_0_0.vlm"},
	{0x3796D3D350C1254BULL, "1000288570.ogg"},
	{0x3796D5FB1FFD18B5ULL, "hlac_ca_cave_2194_8z.dds"},
	{0x3797ECEA72C7BB17ULL, "vis121cr_saemus.utc"},
	{0x3797EFE1D0EA129BULL, "lt_0_lowtown_docks_17_0.rml"},
	{0x3798640AD92CE53CULL, "205045181.ogg"},
	{0x37989AEE77EBA877ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_twitch.evt"},
	{0x3798A645A5E3BFF9ULL, "6164493_m.fxe"},
	{0x3798A8AD70FDDE0EULL, "zz_hawkeshield1.cut"},
	{0x3798FFC1A04D321EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_barter.evt"},
	{0x379912F1C6FC6E2FULL, "420382020.ogg"},
	{0x3799312F9BE9F77BULL, "6113869_f.fxe"},
	{0x379964D8307303A4ULL, "970908735.ogg"},
	{0x379965D894EAE469ULL, "art\\vfx\\materials\\chunky\\vfx_lightbeam_1.mao"},
	{0x379989E99E5D5582ULL, "zz_qun_debug6085734.cl"},
	{0x37998A886CD4F96CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_drink.ani"},
	{0x3799AF080F514013ULL, "one200dg_solivitus6048179.cl"},
	{0x379A05B7E6B1F0CBULL, "425899248.ogg"},
	{0x379A13731D45418FULL, "793004493.ogg"},
	{0x379ABFB07CC48028ULL, "6087477_f.fxe"},
	{0x379AD6C242037F8AULL, "mer340_bonepit_comments.cnv"},
	{0x379ADADD59440727ULL, "hlac_lt_undercityserialkiller_-1_27x.dds"},
	{0x379ADF3C01B990F6ULL, "vxht_mansion2b_3es_0_0.vlm"},
	{0x379B7C67D38F1265ULL, "vxdr_primevaltomb_5c9_0_0.vlm"},
	{0x379B8E22B1280CE4ULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.mmh"},
	{0x379BE665C23FA59CULL, "hla_ht_0_hightown_a2_-65536_0x.dds"},
	{0x379C16D1775C3439ULL, "271428755.ogg"},
	{0x379C2C539743BA6CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgtp_0t.dds"},
	{0x379C2CD51463958AULL, "313134643.ogg"},
	{0x379C2D74BCD6ED81ULL, "store_dae170cr_mage.utm"},
	{0x379C3BCEC34CB239ULL, "art\\characters\\creatures\\strider\\animation\\attkfr_leg.evt"},
	{0x379C4B10E1CC3FC2ULL, "hlac_lt_backalley_g_0_18y.dds"},
	{0x379D049625F48A11ULL, "290084581.ogg"},
	{0x379D3C7D5F782D61ULL, "6043388_f.fxe"},
	{0x379D49CD4D476362ULL, "205609857.ogg"},
	{0x379D76CCCA8BC51EULL, "ave200_aveline_bar6116418.cl"},
	{0x379D76CD2CAC4265ULL, "6153316_m.fxe"},
	{0x379DE7E7223193F5ULL, "vxlt_gallowstemplar_3s5_0_0.vlm"},
	{0x379E1A47CC510ED7ULL, "645013901.ogg"},
	{0x379E5290CAD72FA2ULL, "6015291_m.fxe"},
	{0x379E60409F248EA5ULL, "vxlt_0_lowtown_a2_n_ewg_0_0.vlm"},
	{0x379E756921729DC0ULL, "vxw_bonepitdestroyed_a3_1q0_0_0.vlm"},
	{0x379E964750699502ULL, "mcr310_sibling6076343.cl"},
	{0x379E9BEE77F01500ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_twitch.gad"},
	{0x379EB8C2420735B2ULL, "mer340_bonepit_comments.bnk"},
	{0x379EE8B897C197E0ULL, "vxlt_0_lowtown_docks_a3_c3i_0_0.vlm"},
	{0x379F16C1A051C549ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_barter.gad"},
	{0x379F1CD5C60AEEC6ULL, "vxlt_0_lowtown_a2_n_eoy_0_1.vlm"},
	{0x379F8D574E14E1F4ULL, "6131206_f.fxe"},
	{0x379F97886CD97AAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_drink.gad"},
	{0x379FCB07C3EFE044ULL, "vxht_0_hightown_a3_c4w_0_0.vlm"},
	{0x379FD7124541A15CULL, "vxlt_blackemporium_13c_0_0.vlm"},
	{0x379FFAB827C64B27ULL, "283094407.ogg"},
	{0x37A01624C2AC8D6AULL, "gencr_coterie_thug_2w_f.utc"},
	{0x37A0585012A1F197ULL, "dae000_tomwise6167332.cl"},
	{0x37A059D70D5BCBF4ULL, "t3_dog_on1.tnt"},
	{0x37A0981F74FCD99EULL, "vxht_keep_5db_0_0.vlm"},
	{0x37A14065FF28B78FULL, "vxht_mansion_n_53u_0_0.vlm"},
	{0x37A2176443252C4BULL, "6126832_f.fxe"},
	{0x37A24912A6C2E3C6ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_4x.dds"},
	{0x37A26288BD91C781ULL, "dae000cr_grand_cleric6124685.cl"},
	{0x37A31E7FF8C7E483ULL, "hlac_ca_caverns2_5802_59z.dds"},
	{0x37A3DC3F865E8B9CULL, "75554307.ogg"},
	{0x37A406B84B1B7730ULL, "vxht_0_hightown_n_b8d_0_0.vlm"},
	{0x37A409A050995389ULL, "art\\levels\\races\\proxy\\props\\_textures\\bedsingle_n.dds"},
	{0x37A47A89FD9777B3ULL, "587231803.ogg"},
	{0x37A4A2BEE1FB5B4FULL, "follower_banter6172741.cl"},
	{0x37A5B6EA94FFC6B7ULL, "w_bonepitmining_a2_191_0.msh"},
	{0x37A5C8EA94FFE51AULL, "w_bonepitmining_a2_191_0.mao"},
	{0x37A5CCEA94FFEC29ULL, "w_bonepitmining_a2_191_0.mmh"},
	{0x37A5E7C1A0579397ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_barter.ani"},
	{0x37A5FA8313F73365ULL, "dae210_gate_guard6084334.cl"},
	{0x37A660A2E9A900B8ULL, "250526671.ogg"},
	{0x37A68E830D6D4C80ULL, "6070388_m.fxe"},
	{0x37A692886CDF908DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_drink.evt"},
	{0x37A6A1AE82C9DB28ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.mmh"},
	{0x37A6D88586137CC0ULL, "vxht_keep_n_4n8_0_0.vlm"},
	{0x37A6F7B070D0F2C1ULL, "primitive_w.mao"},
	{0x37A7287D620DC799ULL, "mag110_vincento6016288.cl"},
	{0x37A7398F86D952D5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_broken_1.mao"},
	{0x37A783037AFF8CE0ULL, "vxdr_primevaltomb_2xf_0_0.vlm"},
	{0x37A7CB21EC8A0F2CULL, "313148066.ogg"},
	{0x37A85BA81B4E3943ULL, "follower_banter6167053.cl"},
	{0x37A8608E5137BA09ULL, "6162232_f.fxe"},
	{0x37A86D093320AF3DULL, "766728777.ogg"},
	{0x37A87F2B66B89D78ULL, "129968937.ogg"},
	{0x37A8F1FD407B0174ULL, "1064366446.ogg"},
	{0x37A8FDCDAA8AE19AULL, "6095422_m.fxe"},
	{0x37A90E525CEF9DCBULL, "vrc321ip_spooky7_obj_7.utp"},
	{0x37A93E24C3A93AC9ULL, "vxht_0_hightown_n_c13_0_0.vlm"},
	{0x37A943CEC3574A28ULL, "art\\characters\\creatures\\strider\\animation\\attkfr_leg.ani"},
	{0x37A9850AA56C5A21ULL, "vxht_0_hightown_n_cbv_0_0.vlm"},
	{0x37AB0FA1F3C82CB2ULL, "llac_w_woundedcoast_196616_0.dds"},
	{0x37ABB44DD6736A0BULL, "vxht_keep_6hl_0_0.vlm"},
	{0x37ABC17C289E15CEULL, "vxht_0_hightown_a2_brw_0_0.vlm"},
	{0x37AC01C5394FDBFCULL, "follower_banter6173192.cl"},
	{0x37AC0AA773C6BEC9ULL, "mag370_meredith6064968.cl"},
	{0x37AC3837FADB1878ULL, "810968613.ogg"},
	{0x37AC6B1FE8C65141ULL, "hlac_w_woundedcoast_0_2x.dds"},
	{0x37ACC06526A69674ULL, "hlac_ca_cave_2194_11y.dds"},
	{0x37ACCCC5D4B88009ULL, "art\\vfx\\combat\\vfx_h_cb_tal_threaten.mmh"},
	{0x37ACD21D911AA725ULL, "vxlt_0_lowtown_a2_n_exf_0_0.vlm"},
	{0x37ACE7461368B640ULL, "vxw_bonepitdestroyed_a3_1z1_0_0.vlm"},
	{0x37AD65085A13DE29ULL, "ht_0_hightown_52_0.rml"},
	{0x37AD72B2E705D6DCULL, "llac_lt_0_lowtown_docks_a3_n_-131075_6.dds"},
	{0x37ADCCB501896557ULL, "6218871_m.fxe"},
	{0x37AE1F9E4F1263B2ULL, "gui_guivertexshader_p0.vsh.fxo.dx11.assembly"},
	{0x37AE2A66D17C5A99ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chant2frame.phy"},
	{0x37AE7FAB5B9A4A80ULL, "vxht_keep_n_6f2_0_0.vlm"},
	{0x37AE99BBF8EA0BFBULL, "hla_ht_0_hightown_n_2_0x.dds"},
	{0x37AFC8AE1623AFA6ULL, "618838031.ogg"},
	{0x37AFCA5ECB401CD8ULL, "ht_0_hightown_a2_36_0_lc.mmh"},
	{0x37AFDBD26990A12AULL, "hlac_lt_0_lowtown_docks_a3_-65539_4z.dds"},
	{0x37B01FE8F2960D60ULL, "cod_bks_kirkwall_guards.utp"},
	{0x37B0473D3236C588ULL, "art\\characters\\creatures\\dragonling\\animation\\blendtree\\dragonling_movement.anb"},
	{0x37B09D71B734833AULL, "dae000_bartrand6205972.cl"},
	{0x37B0F5BB1355A90BULL, "vxht_0_hightown_n_bgf_0_0.vlm"},
	{0x37B18F37D6A858C1ULL, "429646338.ogg"},
	{0x37B1C6330C1072C9ULL, "31034934.ogg"},
	{0x37B1E25B56007668ULL, "character_character_p81.psh.fxo.dx11.assembly"},
	{0x37B1EE91F2D6E25CULL, "6134844_m.fxe"},
	{0x37B26BE3AC704C0DULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyo.mmh"},
	{0x37B26FE3AC70535EULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyo.mao"},
	{0x37B27DE3AC706A93ULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyo.msh"},
	{0x37B28C8390219366ULL, "6144218_m.fxe"},
	{0x37B2BE2B5DA15672ULL, "6125515_f.fxe"},
	{0x37B30D528F549085ULL, "zz_vrc_debug6200160.cl"},
	{0x37B3100873305575ULL, "65435116.ogg"},
	{0x37B3124621BECF08ULL, "6810793.ogg"},
	{0x37B3573B794CD20FULL, "652323529.ogg"},
	{0x37B372FCEB5CDB3EULL, "vxlt_0_lowtown_a2_n_c_hx2_0_0.vlm"},
	{0x37B374FB7E33B8D4ULL, "6172811_m.fxe"},
	{0x37B3B46AEC42A1A2ULL, "21382956.ogg"},
	{0x37B3D989A42F981FULL, "578074534.ogg"},
	{0x37B3F8DCF082E64DULL, "682763245.ogg"},
	{0x37B41403DF35DAB2ULL, "ser221_gascard6165445.cl"},
	{0x37B44C71DF3ED843ULL, "one100dg_vanard6017503.cl"},
	{0x37B46AF6D0571E90ULL, "6171872_m.fxe"},
	{0x37B481264CF23A70ULL, "8433811.ogg"},
	{0x37B483C05FA7A9F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkback.ani"},
	{0x37B4B98C22943954ULL, "hf_lgt_urn_shill.mor"},
	{0x37B4BE4C0D9B4B21ULL, "vxw_bonepitdestroyed_a3_1s7_0_0.vlm"},
	{0x37B4F39FB972E773ULL, "hlac_w_bonepitdestroyed_a3_5_0x.dds"},
	{0x37B5BEF8144737E5ULL, "686146479.ogg"},
	{0x37B5C98EB9E428E4ULL, "gen_im_acc_blt_act3_06.uti"},
	{0x37B5DDF299EDF16CULL, "610062287.ogg"},
	{0x37B5FB8A5C0C9B9AULL, "follower_banter6163472.cl"},
	{0x37B672D303141C25ULL, "dae000_arishok6067815.cl"},
	{0x37B686BCAB9962D2ULL, "354925123.ogg"},
	{0x37B68B0E8526F7B3ULL, "vxca_caverns1_j22_0_0.vlm"},
	{0x37B68EC8081628C0ULL, "vxlt_0_lowtown_a3_n_c_ehx_0_0.vlm"},
	{0x37B6B66A64E637ABULL, "216795995.ogg"},
	{0x37B7070694824B70ULL, "6087188_m.fxe"},
	{0x37B82F1F9FB4CCB0ULL, "vxlt_0_lowtown_a2_eom_0_0.vlm"},
	{0x37B869CF4DF8E766ULL, "vxlt_gallowstemplar_n_6t1_0_0.vlm"},
	{0x37B91B8F9C2E4640ULL, "644070052.ogg"},
	{0x37B9BE92D81C375DULL, "hlac_lt_0_lowtown_a2_n_c_-1_7x.dds"},
	{0x37BA8FF435E2AFB0ULL, "vxht_keep_n_6ew_0_0.vlm"},
	{0x37BC744BCCB38881ULL, "qun231cr_varnell_minion7.utc"},
	{0x37BC79FB1ADDB4E1ULL, "llac_lt_undercitydungeon_-1_39.dds"},
	{0x37BCBDEFA883B744ULL, "hm_qmq_patronm2.mor"},
	{0x37BCC26C703CAFC0ULL, "vxca_cavern1_haunted_i34_0_0.vlm"},
	{0x37BCF1884D906300ULL, "vxht_keep_n_6a3_0_0.vlm"},
	{0x37BD840B54A1DA5AULL, "6115371_m.fxe"},
	{0x37BDAC5F14766E50ULL, "sst_ss_coterie_mage_f.gda"},
	{0x37BDB6ACED04807AULL, "llac_ht_keep_n_4519_14.dds"},
	{0x37BDF99154321A99ULL, "316940754.ogg"},
	{0x37BE21C3F80620D9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottlewine_sa.dds"},
	{0x37BE241FE314E88FULL, "vxht_mansion_n_59w_0_0.vlm"},
	{0x37BE2CB7B6CCA9C8ULL, "6142122_m.fxe"},
	{0x37BE30999978237CULL, "6103671_m.fxe"},
	{0x37BEE9C284FE2341ULL, "vxdr_0_deeproads_bd5_0_0.vlm"},
	{0x37BEFB2D41DAA008ULL, "6111089_m.fxe"},
	{0x37BF343CE8A94A3CULL, "652481596.ogg"},
	{0x37BF543BC1270AEAULL, "6172808_m.fxe"},
	{0x37C04C5BA98AAABDULL, "vxlt_hangedman_20t_0_0.vlm"},
	{0x37C06F416B0F1E93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\enter.evt"},
	{0x37C0CE3C32C24DFBULL, "hlac_lt_darktown_-1_56x.dds"},
	{0x37C0E8F54E59028FULL, "6161607_m.fxe"},
	{0x37C1025AE7A246CBULL, "6030677_m.fxe"},
	{0x37C106711682EE82ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.active.gad"},
	{0x37C107D974F2D7D3ULL, "dae000cr_grand_cleric6124786.cl"},
	{0x37C1946FA79D651BULL, "43107369.ogg"},
	{0x37C1CA11636C9CFBULL, "6098327_m.fxe"},
	{0x37C1E1A25AF3A987ULL, "vxw_encounter2_camp_6i_0_0.vlm"},
	{0x37C1F859A3ECAC8FULL, "hlac_w_woundedcoast_196614_0y.dds"},
	{0x37C21AB902E8DE15ULL, "hlac_ht_keep_3739_0x.dds"},
	{0x37C2396D9A5FAD59ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm03.mmh"},
	{0x37C24121ECBBA281ULL, "vxlt_blackemporium_132_0_0.vlm"},
	{0x37C2436D9A5FBE47ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm03.msh"},
	{0x37C284DEE88ADBB3ULL, "6024584_m.fxe"},
	{0x37C28CFDC9C549BCULL, "6208415_m.fxe"},
	{0x37C32C67753C8D2FULL, "hlac_lt_gallowstemplar_2394_8y.dds"},
	{0x37C364E3AC7EB3CAULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyo.phy"},
	{0x37C379B31C9060CFULL, "572477201.ogg"},
	{0x37C3A349BFF3AFE6ULL, "gen00fl_aveline6114816.cl"},
	{0x37C3B34EE1436845ULL, "6005774_m.fxe"},
	{0x37C3CA6872DD03A3ULL, "qcr220_meredith6054737.cl"},
	{0x37C4453515F64DB0ULL, "6048726_f.fxe"},
	{0x37C49122B14B16DBULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.anb"},
	{0x37C4EDE941A9756EULL, "6090813_f.fxe"},
	{0x37C50978EBF1D7F1ULL, "vxht_0_hightown_a3_all_0_0.vlm"},
	{0x37C51A91ACE4C572ULL, "w_sundermount_82_0.rml"},
	{0x37C52D9A6311B503ULL, "610053666.ogg"},
	{0x37C5699C241031FAULL, "455820994.ogg"},
	{0x37C57D451F4EFFA5ULL, "vxca_caverns1_mines_i1o_0_0.vlm"},
	{0x37C58740EE20B6F6ULL, "lt_0_lowtown_docks_a3_n_161_0.msh"},
	{0x37C59140EE20C7F4ULL, "lt_0_lowtown_docks_a3_n_161_0.mmh"},
	{0x37C59540EE20CE07ULL, "lt_0_lowtown_docks_a3_n_161_0.mao"},
	{0x37C5B8B155922564ULL, "vxdr_0_deeproads_ck8_0_0.vlm"},
	{0x37C5BFF13C57B51AULL, "vxw_encounter1_i8_0_0.vlm"},
	{0x37C5D6AFEE563019ULL, "face_rendertodepth_p3.psh.fxo.dx11"},
	{0x37C6044F3E0D2B36ULL, "vxlt_0_lowtown_docks_buk_0_0.vlm"},
	{0x37C668416B137ED4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\enter.gad"},
	{0x37C7054B0A7DF866ULL, "dae210_gate_guard6085134.cl"},
	{0x37C71E5737A5AC58ULL, "follower_banter6173149.cl"},
	{0x37C722750E834F0BULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg101a_0n.dds"},
	{0x37C7671CD32B33F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallows.phy"},
	{0x37C7EAEB2B39E0EAULL, "vxht_mansion_n_4zj_0_0.vlm"},
	{0x37C7EE159A706467ULL, "dae000_madam_lusine6152278.cl"},
	{0x37C83465144E60F2ULL, "vxht_keep_n_55n_0_0.vlm"},
	{0x37C85AC05FB80C9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkback.gad"},
	{0x37C888D277FC752DULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_ballistabolt01_0_n.dds"},
	{0x37C89C7311F7609FULL, "vxlt_foundry_n_wb_0_0.vlm"},
	{0x37C9145EEE270DECULL, "1011681990.ogg"},
	{0x37C98F7E0C4B8063ULL, "and100_anders6206346.cl"},
	{0x37C9C576D74837CCULL, "plt_cod_mgc_chant_laws.plo"},
	{0x37CA1CC86079A160ULL, "vxlt_warehouse_1a1_0_0.vlm"},
	{0x37CA929A46FF9CF9ULL, "438055411.ogg"},
	{0x37CAE2DCFCE0F661ULL, "vxlt_0_lowtown_docks_n_bcy_0_0.vlm"},
	{0x37CB1B91ACE93206ULL, "w_sundermount_82_0.phy"},
	{0x37CBA8249A213AE0ULL, "6109013_m.fxe"},
	{0x37CC3CAB5121F6E9ULL, "6064902_m.fxe"},
	{0x37CC4589F48FED7FULL, "vxlt_hangedman_2eq_0_0.vlm"},
	{0x37CC67E320AED9AAULL, "pro000_cut2d_real_story6001259.cl"},
	{0x37CCB3DCC933FDE9ULL, "6087848_m.fxe"},
	{0x37CD77416B19B6B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\enter.ani"},
	{0x37CD78FF5B4CF27BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic_3.mao"},
	{0x37CD7C7DF8A6009DULL, "llac_lt_0_lowtown_65536_10.dds"},
	{0x37CDE7CC9202BA54ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_c.impa.ani"},
	{0x37CE2F9E213801F2ULL, "mer240_jansen6050447.cl"},
	{0x37CE8DEA4B49AEF4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_rhand.ani"},
	{0x37CEF7739B9D400FULL, "6069886_m.fxe"},
	{0x37CF22F91AA72373ULL, "50624099.ogg"},
	{0x37CF2CD70BAA0C7FULL, "vxlt_0_lowtown_docks_9rd_0_0.vlm"},
	{0x37CFBF2A4B935F4AULL, "vxdr_dungeon_1t1_0_0.vlm"},
	{0x37D009121B6F7F49ULL, "6071896_m.fxe"},
	{0x37D031716CBE6179ULL, "798952799.ogg"},
	{0x37D0537F4CE9C207ULL, "vxw_encounter2_camp_1h_0_0.vlm"},
	{0x37D08836E3BE5FF3ULL, "238633592.ogg"},
	{0x37D08DDD83566579ULL, "8165454.ogg"},
	{0x37D0B2FEDEB1BB01ULL, "vxlt_blackemporium_161_0_0.vlm"},
	{0x37D0D238B0082F36ULL, "vxht_keep_5k7_0_0.vlm"},
	{0x37D12851B3074D91ULL, "183214299.ogg"},
	{0x37D149A18C8A6203ULL, "hlac_dr_0_deeproads_5841_1x.dds"},
	{0x37D15864695522CFULL, "426347546.ogg"},
	{0x37D1A98978F92ED8ULL, "zz_dae_debug6047602.cl"},
	{0x37D1CF3D22CFE71EULL, "gen00fl_varric6033642.cl"},
	{0x37D1CF42A4313215ULL, "vxca_cavern1_haunted_gk6_0_1.vlm"},
	{0x37D238EEEFFFA953ULL, "hlac_lt_0_lowtown_1_1x.dds"},
	{0x37D25017469D8761ULL, "1001843651.ogg"},
	{0x37D2BD821DEBD035ULL, "zz_followers_banter6158811.cl"},
	{0x37D2D2F59C60057CULL, "6096230_m.fxe"},
	{0x37D30B4F5C655D21ULL, "vxca_cavern1_haunted_i13_0_0.vlm"},
	{0x37D31958209E466FULL, "244317181.ogg"},
	{0x37D32A9439A5696DULL, "vxca_caverns1_mines_im3_0_0.vlm"},
	{0x37D3480C6E696113ULL, "vxca_caverns2_3yi_0_0.vlm"},
	{0x37D35E2989D380CCULL, "vxht_brothel_n_5g9_0_0.vlm"},
	{0x37D37B55DDE7F071ULL, "vxht_0_hightown_a3_aic_0_0.vlm"},
	{0x37D3F9CBE1D15C9FULL, "6060401_m.fxe"},
	{0x37D4736E8EF1B293ULL, "lgt220pt_shp_mag_invst.ncs"},
	{0x37D48AEA4B4E1556ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_rhand.gad"},
	{0x37D4B81A0B2ACF85ULL, "6168589_m.fxe"},
	{0x37D4C4719B43CAABULL, "hlac_w_encounter2_196612_0x.dds"},
	{0x37D4E4CC9208D3B6ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_c.impa.gad"},
	{0x37D5073D766C6F7EULL, "vxlt_darktown_2tr_0_0.vlm"},
	{0x37D596C8B18B8DA4ULL, "340186060.ogg"},
	{0x37D5C9C5D4DB803EULL, "art\\vfx\\combat\\vfx_h_cb_tal_threaten.anb"},
	{0x37D61F85B54E91A7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_ands_0s.dds"},
	{0x37D6317116959468ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.active.ani"},
	{0x37D6938AE642FA60ULL, "vxht_playershouse_n_1sg_0_0.vlm"},
	{0x37D6E9FB583BD62FULL, "491837960.ogg"},
	{0x37D7B8CA4A39634FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal_2.mao"},
	{0x37D7CDB3EEA57835ULL, "vxw_bonepitdestroyed_a3_l9_0_0.vlm"},
	{0x37D8B3EE1C0DFDB4ULL, "qun200pt_following_qun.ncs"},
	{0x37D8C8A9BF99AD2FULL, "gen_im_cft_reg_e2.uti"},
	{0x37D8F2E1D121AF67ULL, "lt_0_lowtown_docks_17_0.anb"},
	{0x37D9508B09410845ULL, "478258670.ogg"},
	{0x37D986B1D03D1948ULL, "hlac_lt_0_lowtown_a2_n_65537_5y.dds"},
	{0x37D9B4BBDC6E3798ULL, "355283842.ogg"},
	{0x37D9EFA1096F791FULL, "gen00fl_aveline6114411.cl"},
	{0x37DA52CC3C261065ULL, "vxw_bonepitdestroyed_a3_1rh_0_0.vlm"},
	{0x37DA557ADC6FD775ULL, "hlac_lt_0_lowtown_n_131073_2z.dds"},
	{0x37DAFFCC920D6D15ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_c.impa.evt"},
	{0x37DBA5EA4B5461B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_rhand.evt"},
	{0x37DBA6DCF96AB63BULL, "lava0volume.psh.fxo"},
	{0x37DBB4D1E4B6B61BULL, "126900904.ogg"},
	{0x37DCCC9A5E026BA3ULL, "512161412.ogg"},
	{0x37DD315E1C9F3ABEULL, "55682114.ogg"},
	{0x37DDEE04A1CB60E7ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_homa_0t.dds"},
	{0x37DE381EE56CBACCULL, "6126447_m.fxe"},
	{0x37DEAB497879273AULL, "6018748_m.fxe"},
	{0x37DECAAAD7D7C9F6ULL, "hlac_lt_gallowscourtyard_131070_1y.dds"},
	{0x37DEEE5A3123D0E6ULL, "vxht_0_hightown_a3_bph_0_0.vlm"},
	{0x37DFFB5E1650203DULL, "pro000_refugee_angry.bnk"},
	{0x37E013E2E37D94A1ULL, "vxw_bonepitdestroyed_a3_1b4_0_0.vlm"},
	{0x37E07194537C8E06ULL, "vxlt_darktown_2wc_0_0.vlm"},
	{0x37E0E4597D2CF1BFULL, "256020073.ogg"},
	{0x37E115284921634DULL, "407305637.ogg"},
	{0x37E1597841D4F543ULL, "6127054_m.fxe"},
	{0x37E174AFC47D9A5BULL, "vxw_bonepitdestroyed_a3_e0_0_0.vlm"},
	{0x37E190A865ABC250ULL, "6036492_m.fxe"},
	{0x37E19C712B9B81EDULL, "vxca_caverns1_mines_ij2_0_0.vlm"},
	{0x37E1C4A9156F1239ULL, "6034235_m.fxe"},
	{0x37E1C7DE683E0231ULL, "1031940216.ogg"},
	{0x37E1DDD4EC3C2C5DULL, "823239107.ogg"},
	{0x37E21DEB5C9D0073ULL, "299844094.ogg"},
	{0x37E2279481383EEFULL, "6087374_m.fxe"},
	{0x37E2DD5E16522345ULL, "pro000_refugee_angry.cnv"},
	{0x37E2F72A6DD8C0FFULL, "vxht_0_hightown_a2_bge_0_0.vlm"},
	{0x37E37CE4948888F3ULL, "follower_banter6160614.cl"},
	{0x37E387987DA9C16DULL, "gen00fl_fenris6112102.cl"},
	{0x37E3DD5E125AEA4FULL, "vxht_keep_n_66q_0_0.vlm"},
	{0x37E46216A4450664ULL, "gen00fl_anders6201681.cl"},
	{0x37E47F8817CE2443ULL, "900210544.ogg"},
	{0x37E4A8C5381604D5ULL, "850888460.ogg"},
	{0x37E50567D83912E0ULL, "vxht_playershouse_n_1vd_0_0.vlm"},
	{0x37E55A5B52FF6302ULL, "183779746.ogg"},
	{0x37E591A57E2ED265ULL, "gen00fl_anders6111123.cl"},
	{0x37E659D65BFE7325ULL, "vxlt_gallowstemplar_f_34k_0_0.vlm"},
	{0x37E739799C53C49CULL, "vxht_mansion2b_22q_0_0.vlm"},
	{0x37E7AEABDAF5A240ULL, "hlac_lt_gallowscourtyard_65534_0y.dds"},
	{0x37E80DFE654DBBD7ULL, "vxdr_primevaltomb_6yz_0_0.vlm"},
	{0x37E83D58AB8E03ADULL, "static_static_p6.vsh.fxo.dx11.assembly"},
	{0x37E84DEC1765EFE1ULL, "fx_tendrils01.dds"},
	{0x37E84FA4AA51D569ULL, "vxht_0_hightown_n_627_0_0.vlm"},
	{0x37E86EE2860E73EAULL, "vxlt_gallowsprison_n_6qf_0_0.vlm"},
	{0x37E8AF78D1040F39ULL, "vxht_0_hightown_c5h_0_0.vlm"},
	{0x37E8D8EE48E3073BULL, "art\\characters\\weapons\\axes\\textures\\w_axe_03a_0n.dds"},
	{0x37E8FB8130892AE4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mb.mmh"},
	{0x37E8FF8130893137ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mb.mao"},
	{0x37E9118130894F86ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mb.msh"},
	{0x37E9312812A40E7CULL, "vxht_0_hightown_a3_b8j_0_0.vlm"},
	{0x37E98BAB08A8B506ULL, "665913721.ogg"},
	{0x37EA5135046080D8ULL, "vxht_0_hightown_n_8pn_0_0.vlm"},
	{0x37EA585083F05CB0ULL, "1049266675.ogg"},
	{0x37EAC3E529E50114ULL, "dae000_epic_merchant_46179511.cl"},
	{0x37EB1FE16FCB56F5ULL, "6095394_f.fxe"},
	{0x37EBDCDDAAA37E0FULL, "385308398.ogg"},
	{0x37EC101B842C8648ULL, "w_blightlands_227_0.msh"},
	{0x37EC221B842CA4D9ULL, "w_blightlands_227_0.mao"},
	{0x37EC2E1B842CB972ULL, "w_blightlands_227_0.mmh"},
	{0x37EC566BC3785ED6ULL, "971482676.ogg"},
	{0x37EC5D009E6C07D1ULL, "369100600.ogg"},
	{0x37EC609263F4AE83ULL, "6108574_m.fxe"},
	{0x37EDC42757C0B015ULL, "6121545_f.fxe"},
	{0x37EDF93641AE27A3ULL, "hlac_w_encounter2_-65533_0z.dds"},
	{0x37EE07120E9D11D5ULL, "gen00fl_sebastian6179797.cl"},
	{0x37EE8B259A0056E7ULL, "vxlt_hangedman_n_1bx_0_0.vlm"},
	{0x37EEAAEE532BBB2DULL, "198934788.ogg"},
	{0x37EEB548F2AB3B83ULL, "6167199_m.fxe"},
	{0x37EEEBBCB3FE1D15ULL, "vxdr_primevaltomb_8o8_0_0.vlm"},
	{0x37EF047870280BDDULL, "vxdr_primevaltomb_6wr_0_0.vlm"},
	{0x37EF2A63061DDC8BULL, "732089314.ogg"},
	{0x37EF64CD9730B493ULL, "hlac_w_sundermount_-196607_0z.dds"},
	{0x37EF680C144740F2ULL, "vxht_0_hightown_a2_b74_0_0.vlm"},
	{0x37EF7F893412C1FEULL, "6110989_m.fxe"},
	{0x37EF9439C5CED23BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic_1.mao"},
	{0x37EFAD80D442994BULL, "753363868.ogg"},
	{0x37EFB6E7742882A1ULL, "182596903.ogg"},
	{0x37EFE20F2C00245FULL, "zz_followers_banter6158505.cl"},
	{0x37EFF4BA7D6C15D3ULL, "vxlt_0_lowtown_docks_b49_0_0.vlm"},
	{0x37F03F4F0639E890ULL, "aip_hurlockcrossbow_base.gda"},
	{0x37F08DD5E2DC4C5BULL, "vxlt_0_lowtown_docks_a3_n_brw_0_0.vlm"},
	{0x37F0B7E95D21164FULL, "6111665_f.fxe"},
	{0x37F159E613EF74B6ULL, "vxlt_0_lowtown_docks_bbn_0_0.vlm"},
	{0x37F19A0C26BCA3D6ULL, "weapon_weapon_p5.psh.fxo.dx11"},
	{0x37F24B8004F42FB3ULL, "vxlt_hangedman_2mm_0_0.vlm"},
	{0x37F24CA30FD6881DULL, "rdr111dg_martin6014697.cl"},
	{0x37F24F3B045102CFULL, "vxht_keep_n_61p_0_0.vlm"},
	{0x37F299D3A7F991EAULL, "862661953.ogg"},
	{0x37F350575EAF20A7ULL, "379806226.ogg"},
	{0x37F360F74992C501ULL, "461640170.ogg"},
	{0x37F3ED35D8F2C68EULL, "dae300_bran6143445.cl"},
	{0x37F4084A14730276ULL, "zz_followers_banter6159417.cl"},
	{0x37F41BEDC9A02BE0ULL, "zz_approval_debug6003085.cl"},
	{0x37F43BDF4979B4C8ULL, "vxht_keep_7a9_0_0.vlm"},
	{0x37F44E1249A84D2BULL, "llac_ht_mansion2_818_50.dds"},
	{0x37F4CBB34DF48BA5ULL, "vxlt_gallowstemplar_f_33j_0_0.vlm"},
	{0x37F6C43BCE8DBBFAULL, "6043410_m.fxe"},
	{0x37F7459BDE137FE9ULL, "hlac_ht_mansion2b_818_54x.dds"},
	{0x37F7A305049A26FCULL, "vxht_0_hightown_a3_b7k_0_0.vlm"},
	{0x37F7C4B0D3A150C3ULL, "dre000_dougal6129791.cl"},
	{0x37F83B4053058615ULL, "vxlt_0_lowtown_gcj_0_0.vlm"},
	{0x37F8DE526C004E11ULL, "246086696.ogg"},
	{0x37F912A8D86A6C62ULL, "6135104_m.fxe"},
	{0x37F97BDCD0BF563EULL, "vxlt_warehouse_13i_0_0.vlm"},
	{0x37F97C2EB317C442ULL, "6201649_m.fxe"},
	{0x37F98C9C30A8C651ULL, "15816600.ogg"},
	{0x37F98E19EA4EAD76ULL, "fxa_cen_volt.dds"},
	{0x37F98EA0B019C13AULL, "mer241cr_queen_spider.utc"},
	{0x37FA51B030A0CE03ULL, "qcr270_fake_meredith6066339.cl"},
	{0x37FA54EE27AA7BCBULL, "gen00fl_anders6135765.cl"},
	{0x37FAD58A44A29221ULL, "vxdr_primevaltomb_8n7_0_0.vlm"},
	{0x37FAF9CB1F3E979EULL, "hla_ht_0_hightown_a3_65536_0x.dds"},
	{0x37FB17FB5345267FULL, "art\\vfx\\environment\\vfx_trap_glow_p.impa.gad"},
	{0x37FBC8753D84028FULL, "vxw_encounter2_camp_rm_0_0.vlm"},
	{0x37FBDF31FACBB9E4ULL, "vxht_chantry_2r9_0_0.vlm"},
	{0x37FC5BA41B100EEAULL, "630829949.ogg"},
	{0x37FC6A8709FC07E7ULL, "801872784.ogg"},
	{0x37FC993A358B29EFULL, "vxw_bonepit_13i_0_0.vlm"},
	{0x37FD31697EDF8E83ULL, "33732577.ogg"},
	{0x37FD7655621E245DULL, "vxdr_primevaltomb_6zq_0_0.vlm"},
	{0x37FDD9E9063D5972ULL, "vxht_0_hightown_a2_b85_0_0.vlm"},
	{0x37FDDA093378B04BULL, "gen_im_cns_tom_mag_bldmage.uti"},
	{0x37FE19203E97A9C3ULL, "vxca_caverns1_mines_jfg_0_0.vlm"},
	{0x37FE5869A869CB5BULL, "vxw_bonepitdestroyed_a3_k2_0_0.vlm"},
	{0x37FE6820A118020CULL, "vxlt_0_lowtown_a2_n_gns_0_0.vlm"},
	{0x37FEE013262961F5ULL, "hlac_lt_gallowstemplar_n_7198_6y.dds"},
	{0x37FF811AAB372845ULL, "hlac_lt_gallowstemplar_n_2391_1x.dds"},
	{0x37FF9397A003A197ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_006a_0d.dds"},
	{0x37FFB26B95BC1212ULL, "vxht_brothel_n_6c4_0_0.vlm"},
	{0x38003689D468C719ULL, "vrc321_party_barks6095746.cl"},
	{0x38003B8C41E0DE83ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barreltiny_0.pwk"},
	{0x3800B6A99DAFF227ULL, "darktown_male_fereldan_sta6175148.cl"},
	{0x3800E67F2084A775ULL, "24592742.ogg"},
	{0x380100D3F81770D5ULL, "ser270_templar6122517.cl"},
	{0x38010CFB53497F6CULL, "art\\vfx\\environment\\vfx_trap_glow_p.impa.evt"},
	{0x38012E4240200239ULL, "896822884.ogg"},
	{0x38018998F8B998BEULL, "36440168.ogg"},
	{0x3801CDB6A4B17D6FULL, "30503327.ogg"},
	{0x38020CD962DB4DF9ULL, "w_bonepitmining_a2_247_0.msh"},
	{0x380216D962DB5EE7ULL, "w_bonepitmining_a2_247_0.mmh"},
	{0x38021AD962DB6534ULL, "w_bonepitmining_a2_247_0.mao"},
	{0x38026D033B61EFE4ULL, "vxdr_0_deeproads_fpd_0_0.vlm"},
	{0x380291BA2AC7B2A0ULL, "llac_lt_hangedman_-1_27.dds"},
	{0x38040E76A28146C9ULL, "6179786_f.fxe"},
	{0x38042F2CA5C57317ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_upperblank_d.dds"},
	{0x38052BCB2FC20B67ULL, "vrc221_hugin6065260.cl"},
	{0x38054DEAEAFEF5B6ULL, "vxht_mansion2b_f5_0_0.vlm"},
	{0x38058AC01885405FULL, "6119529_m.fxe"},
	{0x3805A56B08E86141ULL, "37945346.ogg"},
	{0x380635E0931EB162ULL, "vxht_mansion2_3eo_0_0.vlm"},
	{0x38066C14A93C0590ULL, "vxlt_0_lowtown_n_h7o_0_0.vlm"},
	{0x380671EEC0281BEDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightc_sa.dds"},
	{0x3806852E2C04DEB1ULL, "vxlt_undercitydungeon_24r_0_0.vlm"},
	{0x380686D7AF0F0AC4ULL, "vxht_keep_4he_0_0.vlm"},
	{0x3806A8101AC9CB1EULL, "6065317_m.fxe"},
	{0x3806F39F809A4E1BULL, "vxca_caverns2_9fz_0_0.vlm"},
	{0x38075D337A48B12DULL, "hlac_w_bonepit_131077_0z.dds"},
	{0x38075EEC1780BD6BULL, "fx_tendrils01.mao"},
	{0x3807778D14E4208AULL, "6149482_m.fxe"},
	{0x380795810C50CA3CULL, "dae200_saemus6091678.cl"},
	{0x3807BCAEFF224C2BULL, "follower_banter6163981.cl"},
	{0x3807E43A5A47DE77ULL, "ave200_donnic6111271.cl"},
	{0x3807EDB9C2B56EBEULL, "vxlt_warehouse_16j_0_0.vlm"},
	{0x3808765B057981ADULL, "221832786.ogg"},
	{0x380894018708323FULL, "llac_ht_bartrand_n_2995_1.dds"},
	{0x3808A5E04A955DF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_force_crush.gad"},
	{0x380941912550EA9BULL, "6173051_m.fxe"},
	{0x380942CE2F8FC4C8ULL, "768848290.ogg"},
	{0x3809444C47E0DF83ULL, "6106774_m.fxe"},
	{0x380947673698AAA1ULL, "vxdr_primevaltomb_8i6_0_0.vlm"},
	{0x380A02AFAFAC6009ULL, "cd_ave121_jeven.bnk"},
	{0x380A3A522F7A1B0FULL, "vxw_encounter2_camp_ml_0_0.vlm"},
	{0x380B728A31C2561CULL, "vxlt_gallowstemplar_f_2be_0_0.vlm"},
	{0x380BF7363E7D3385ULL, "6152243_m.fxe"},
	{0x380C163BC8185E9FULL, "6217108_m.fxe"},
	{0x380C6E8BED6E2FA6ULL, "6147912_m.fxe"},
	{0x380CA27E2AFBA8B6ULL, "vxw_bonepitdestroyed_a3_1vc_0_0.vlm"},
	{0x380CD9FD930E1A8CULL, "vxlt_0_lowtown_a2_n_gkp_0_0.vlm"},
	{0x380D3AE1106BCA14ULL, "580761768.ogg"},
	{0x380D86F6C1196FDCULL, "hlac_lt_0_lowtown_a2_n_-196608_3x.dds"},
	{0x380DE89D115FA49EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_action.evt"},
	{0x380E298D7CB6F60EULL, "861989661.ogg"},
	{0x380EE5BD1C4ADADFULL, "zz_vault_debug6178666.cl"},
	{0x380EF3307E02CAF9ULL, "mrl350ip_marethari_corpse.utp"},
	{0x380EF4FB535593C5ULL, "art\\vfx\\environment\\vfx_trap_glow_p.impa.ani"},
	{0x380FA8E04A9B8129ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_force_crush.evt"},
	{0x38108830E6566934ULL, "one110_gustav6089789.cl"},
	{0x38108FEB7379BA35ULL, "6117962_f.fxe"},
	{0x3810DC98D7B5EEAEULL, "180264043.ogg"},
	{0x3811BBEF5435BBABULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmld_1.mao"},
	{0x3812155CB977D9E5ULL, "hlac_lt_0_lowtown_a2_-65537_3x.dds"},
	{0x38121C120C25C772ULL, "6122438_m.fxe"},
	{0x38127C2874A6D241ULL, "634094933.ogg"},
	{0x3812AA338A6C64CAULL, "6052755_m.fxe"},
	{0x381304687C90040FULL, "vxca_caverns1_jtl_0_0.vlm"},
	{0x381347391C9D593EULL, "dae000_gamlen6145131.cl"},
	{0x38136575F2468C0DULL, "zz_personality_debug.cnv"},
	{0x3813D8EEAB9648C9ULL, "art\\levels\\races\\vfx_models\\prp_brazlogwall.mao"},
	{0x3813E4EEAB965D62ULL, "art\\levels\\races\\vfx_models\\prp_brazlogwall.mmh"},
	{0x3813E6EEAB966018ULL, "art\\levels\\races\\vfx_models\\prp_brazlogwall.msh"},
	{0x3813FC9D0A7B58DFULL, "6114229_m.fxe"},
	{0x3813FF9D116437C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_action.gad"},
	{0x3814BC1AA190208DULL, "isa211_party_reaction6083426.cl"},
	{0x3814F8B4A1052344ULL, "vxht_keep_4mj_0_0.vlm"},
	{0x381509587299F944ULL, "6084871_m.fxe"},
	{0x3815375076692C73ULL, "art\\characters\\creatures\\ragedemon\\animation\\abilities.evt"},
	{0x3815716B87E71A45ULL, "follower_barks6173422.cl"},
	{0x381664712ECD6E3AULL, "hf_ibethany_h3s1.mor"},
	{0x3816F0B9543AA0BFULL, "hlac_lt_undercityencounter_-1_30z.dds"},
	{0x38178AF77049B966ULL, "gen00fl_isabela6109792.cl"},
	{0x3817908E486549C2ULL, "vxht_chantry_2td_0_1.vlm"},
	{0x3817BBB23C973468ULL, "253551999.ogg"},
	{0x38184B6F18825D7BULL, "6043392_m.fxe"},
	{0x3818B4ED691D53B9ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11a.msh"},
	{0x3818BD74EB1C5FAAULL, "812294277.ogg"},
	{0x3818BEED691D64A7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11a.mmh"},
	{0x3818FBF3E7E38884ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_2z.dds"},
	{0x3819F4C5759260FEULL, "momnt_dis.mao"},
	{0x381AB17D9EACFA5EULL, "vxw_woundedcoast_15q_0_0.vlm"},
	{0x381AD09D116A0617ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_action.ani"},
	{0x381ADDABEE828296ULL, "402823861.ogg"},
	{0x381B20ECA68780CBULL, "hlac_lt_backalley_g_65536_4z.dds"},
	{0x381B5D0F81DC2C67ULL, "hlac_ca_caverns1_24423_2y.dds"},
	{0x381B97FD20B279DCULL, "dae000_bodhan_letter6144317.cl"},
	{0x381B9F9F27359EDFULL, "mind_dis.mao"},
	{0x381BAEBCFD7439B2ULL, "vxlt_gallowstemplar_n_5t6_0_0.vlm"},
	{0x381BB66BB2DB4EEAULL, "vxlt_0_lowtown_n_fi5_0_0.vlm"},
	{0x381BDC7C732BEC23ULL, "950531451.ogg"},
	{0x381BFF78C3E1A12DULL, "hlac_lt_0_lowtown_a2_n_c_-65535_5y.dds"},
	{0x381C3050766F3EB4ULL, "art\\characters\\creatures\\ragedemon\\animation\\abilities.gad"},
	{0x381C5B587A1F61FCULL, "476115679.ogg"},
	{0x381C795BC9AD2D64ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_2z.dds"},
	{0x381CA5F23BDC3BF0ULL, "1073611582.ogg"},
	{0x381CC81BDBFDFD65ULL, "vxca_caverns2prc_96q_0_0.vlm"},
	{0x381CCADB45D0DD33ULL, "153192851.ogg"},
	{0x381CFABAFBF05631ULL, "hla_ht_0_hightown_a3_0_0z.dds"},
	{0x381D20BB03597F1FULL, "qcr000_arishok6059687.cl"},
	{0x381D535E56300FEBULL, "hlac_lt_0_lowtown_docks_65534_0x.dds"},
	{0x381D90E04AA79598ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_force_crush.ani"},
	{0x381DAF3B51A14580ULL, "6120795_m.fxe"},
	{0x381DB8A6ED1F0BA7ULL, "vxlt_0_lowtown_a2_daq_0_0.vlm"},
	{0x381E07C9BC2C853FULL, "w_encounter3_109_0.msh"},
	{0x381E0DC9BC2C8FE1ULL, "w_encounter3_109_0.mmh"},
	{0x381E19C9BC2CA382ULL, "w_encounter3_109_0.mao"},
	{0x381E2BE32C5D1759ULL, "plt_isa310pt_isabelas_trail.nss"},
	{0x381E395467B21F31ULL, "lt_0_lowtown_docks_a3_n_212_0.mmh"},
	{0x381E455467B233D2ULL, "lt_0_lowtown_docks_a3_n_212_0.mao"},
	{0x381E535467B24BEFULL, "lt_0_lowtown_docks_a3_n_212_0.msh"},
	{0x381F4C24E043F766ULL, "fen110_fenris6101960.cl"},
	{0x381FB94D991D48BAULL, "mcr310_sibling6076358.cl"},
	{0x38201DCE2BA39B77ULL, "art\\vfx\\models\\vfx_lyrium_c.mmh"},
	{0x382033CE2BA3C0E9ULL, "art\\vfx\\models\\vfx_lyrium_c.msh"},
	{0x382040DC97D48AA5ULL, "hlac_lt_backalley_n_65536_3z.dds"},
	{0x38206A0A732D5765ULL, "vxdr_primevaltomb_8oh_0_0.vlm"},
	{0x3821C7D2E48B900AULL, "one110_pryce6088778.cl"},
	{0x3821EC9ADBC08CD8ULL, "zz_ave_debug6056688.cl"},
	{0x38223D430500FB2EULL, "vxw_bonepitmining_a2_1l7_0_0.vlm"},
	{0x38223F507673C392ULL, "art\\characters\\creatures\\ragedemon\\animation\\abilities.ani"},
	{0x38224CDDEAA2E9C8ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.anb"},
	{0x38227219D7D3CEC9ULL, "and110tr_refugees.ncs"},
	{0x3822E1848132234BULL, "vxlt_0_lowtown_n_enw_0_1.vlm"},
	{0x38230671957D1E16ULL, "61842359.ogg"},
	{0x38237E3DE6BC4EE6ULL, "gen_im_trs_qun_horn_04.uti"},
	{0x3823B2AC365626DCULL, "vxht_keep_n_6ao_0_0.vlm"},
	{0x3823E07FB233EDBAULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magb.mmh"},
	{0x3823F27FB2340C00ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magb.msh"},
	{0x382405A5A11EEB49ULL, "353416845.ogg"},
	{0x3824378352325D14ULL, "art\\cinematics\\act1\\cs_fp_pained.ani"},
	{0x38244B7A9356D0ABULL, "ca_cave_mmap_ph.dds"},
	{0x3824E5AF40B1854FULL, "screenspace.vsh.fxo"},
	{0x382512E32C630B32ULL, "plt_isa310pt_isabelas_trail.plo"},
	{0x382529EB81C3565CULL, "6110910_m.fxe"},
	{0x38253A6D9AB42C46ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm03.phy"},
	{0x3825972A9123C5FBULL, "w_sundermount_165_0.msh"},
	{0x3825A92A9123E446ULL, "w_sundermount_165_0.mao"},
	{0x3825B52A9123F8E5ULL, "w_sundermount_165_0.mmh"},
	{0x3825FA60263E997EULL, "gen00fl_sebastian6180459.cl"},
	{0x3826DF4F6F7937C8ULL, "vxht_0_hightown_n_5h3_0_0.vlm"},
	{0x3826DF9615252465ULL, "and111_donation_box6200721.cl"},
	{0x38271E0C13664C0FULL, "vxw_encounter2_camp_cn_0_0.vlm"},
	{0x382758EB3E3FCE44ULL, "6071839_m.fxe"},
	{0x382766FB14800219ULL, "mrl150_dalish_guard6068268.cl"},
	{0x38279A38702E0007ULL, "vxht_0_hightown_n_a84_0_0.vlm"},
	{0x382915F553E87D9BULL, "hlac_ht_playershouse_n_855_24x.dds"},
	{0x382918DBB37BC530ULL, "360265442.ogg"},
	{0x3829235A90A312DEULL, "vxw_woundedcoast_10r_0_0.vlm"},
	{0x38297127CD9570A3ULL, "328860417.ogg"},
	{0x382994D3DD33121EULL, "vxht_brothel_n_5cw_0_0.vlm"},
	{0x3829B32BC0793217ULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_p.anb"},
	{0x382A205D54439F19ULL, "341756791.ogg"},
	{0x382A5549AAB4AE5BULL, "vxlt_0_lowtown_docks_a3_n_bfk_0_0.vlm"},
	{0x382A9FCBF4F92980ULL, "hlac_w_blightlands_196610_0z.dds"},
	{0x382B2BFCBF59B14CULL, "6103041_f.fxe"},
	{0x382B348352387776ULL, "art\\cinematics\\act1\\cs_fp_pained.gad"},
	{0x382BBC4D740CB9C7ULL, "403516084.ogg"},
	{0x382C614DEB7B1DE4ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden1.msh"},
	{0x382C70FADDE33994ULL, "6076150_f.fxe"},
	{0x382C774DEB7B4286ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden1.mmh"},
	{0x382C9ABA285A4DB2ULL, "6109684_m.fxe"},
	{0x382CE312CAD11CEBULL, "6071825_m.fxe"},
	{0x382D63E865DF7F7EULL, "hlac_ht_keep_3734_0x.dds"},
	{0x382D792B19BC4E21ULL, "hlac_lt_shop_-1_7z.dds"},
	{0x382DABE51B03D084ULL, "vxlt_gallowscourtyard_7e8_0_0.vlm"},
	{0x382DB0096BE01082ULL, "isa120_hayder6068789.cl"},
	{0x382DB66B2A0EC736ULL, "6067395_m.fxe"},
	{0x382DB6E8AFAB8FF3ULL, "hlac_lt_ruinedhouse_n_-1_7z.dds"},
	{0x382E66961EBAEA5AULL, "lt_0_lowtown_docks_a3_n_3_0.anb"},
	{0x382F1A02C3D7D0FCULL, "gen00fl_fenris6110868.cl"},
	{0x382FCBA7277D249FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv3.gad"},
	{0x38314B92B1FBF882ULL, "6114505_m.fxe"},
	{0x38319405C69DDFCBULL, "424021512.ogg"},
	{0x3831CDBCBFEE060CULL, "6076229_m.fxe"},
	{0x38320C43E5A691E1ULL, "dre000_stroud6036676.cl"},
	{0x38321CC8951B416FULL, "6153215_m.fxe"},
	{0x38327F762FF78ECFULL, "956206199.ogg"},
	{0x3832BE9CA5A0BEE0ULL, "vis121_saemus6099443.cl"},
	{0x3832CB764C07534EULL, "78132410.ogg"},
	{0x3832DD7253CFA61EULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_l_p.mmh"},
	{0x38335B19D7E21AACULL, "and110tr_refugees.utt"},
	{0x3833AD5BB8D1BE79ULL, "vxca_caverns1_jl2_0_0.vlm"},
	{0x3834475EFFC3BE24ULL, "185186845.ogg"},
	{0x3834CA02A176656EULL, "vxca_caverns1_iyi_0_0.vlm"},
	{0x38357D39DB2C9899ULL, "672701679.ogg"},
	{0x3835A3A1BBA33C28ULL, "vxht_0_hightown_a2_atp_0_0.vlm"},
	{0x3835B4410DBE8CC1ULL, "89971324.ogg"},
	{0x3835B602A8979FE0ULL, "6054852_m.fxe"},
	{0x3835CD711B40F9D2ULL, "mag100_karras6034509.cl"},
	{0x3835E9E75C9FFBA3ULL, "vxlt_blackemporium_8h_0_0.vlm"},
	{0x383645B96322CB4FULL, "68364840.ogg"},
	{0x3836A1E38ECB72D2ULL, "dae000cr_grand_cleric6148093.cl"},
	{0x3836A9479BB7D1E5ULL, "vxlt_backalley_g_1w0_0_1.vlm"},
	{0x3836C0A72783308CULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv3.evt"},
	{0x3836C82107A49F9CULL, "vxlt_gallowstemplar_f_2mx_0_0.vlm"},
	{0x3837597FECC03F9CULL, "vxlt_0_lowtown_docks_n_9a1_0_0.vlm"},
	{0x38377A1CC870743BULL, "hlac_lt_0_lowtown_a2_n_-262144_3x.dds"},
	{0x3837AE9D0CDA22A9ULL, "549888816.ogg"},
	{0x383832C36A9EF1D4ULL, "dae000_pc_home_ambients6208699.cl"},
	{0x38384DB0273F475EULL, "6101019_m.fxe"},
	{0x38387033107875ADULL, "6093683_m.fxe"},
	{0x38389276E1606AB2ULL, "vxlt_gallowstemplar_n_5b0_0_0.vlm"},
	{0x3838B2B508CC5AFBULL, "6119378_m.fxe"},
	{0x3838BD01DD0D01DAULL, "hlac_lt_0_lowtown_a2_n_65535_0z.dds"},
	{0x383912622611919EULL, "6060199_f.fxe"},
	{0x38394238880CF9B3ULL, "6095012_m.fxe"},
	{0x383955225DB1EDF9ULL, "vxw_encounter2_camp_p1_0_0.vlm"},
	{0x383986423E5E8316ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv5.ani"},
	{0x38398D79C315D833ULL, "hf_lgt_margitte.mor"},
	{0x3839E89F054B8207ULL, "vxw_sundermount_375_0_1.vlm"},
	{0x3839F77D33BFE190ULL, "vxlt_gallowstemplar_n_438_0_0.vlm"},
	{0x383AC895F3D0D50EULL, "vxlt_0_lowtown_a3_n_c_fgt_0_0.vlm"},
	{0x383ADF00579493CDULL, "845626060.ogg"},
	{0x383B49D4FDF7A766ULL, "mcr370_meredith6083084.cl"},
	{0x383B75EAECD5F669ULL, "6019156_f.fxe"},
	{0x383C15A5F44DC8B4ULL, "6078177_m.fxe"},
	{0x383C3B734C2885BBULL, "dae000_fem_commoner1_mob6174487.cl"},
	{0x383CC82CC5E0B532ULL, "hlac_lt_0_lowtown_a2_n_c_65536_0y.dds"},
	{0x383D081A4B586E7FULL, "sst_ss_zevran.gda"},
	{0x383D0CFB90FEBA0AULL, "vxlt_gallowscourtyard_6we_0_0.vlm"},
	{0x383D6C659D6336FCULL, "art\\characters\\creatures\\darkspawnogre\\cn_bdy_ogra.mao"},
	{0x383D78659D634A9FULL, "art\\characters\\creatures\\darkspawnogre\\cn_bdy_ogra.mmh"},
	{0x383D7E659D635541ULL, "art\\characters\\creatures\\darkspawnogre\\cn_bdy_ogra.msh"},
	{0x383D919404B062B6ULL, "6055581_m.fxe"},
	{0x383DB86CB89B801DULL, "6161058_m.fxe"},
	{0x383DC72F5E9D4B35ULL, "6111563_m.fxe"},
	{0x383DE1646852B439ULL, "mag170_hugh6080807.cl"},
	{0x383E26EA486E04CBULL, "6173859_m.fxe"},
	{0x383E3A029D35CC3DULL, "vxlt_foundry_n_11a_0_0.vlm"},
	{0x383EF24CCF373F80ULL, "vxht_keep_n_648_0_0.vlm"},
	{0x383EF96335EAD5AEULL, "6135957_m.fxe"},
	{0x383F13CA2CF29E9BULL, "mag291_torpor6056813.cl"},
	{0x383F782FD00A7A11ULL, "gen00fl_varric6033719.cl"},
	{0x383F942E4305DC38ULL, "hlac_ht_chantry_n_2534_6z.dds"},
	{0x3840BC7253DBABDDULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_l_p.anb"},
	{0x3841228D6024B40EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xc.phy"},
	{0x3841A12711B8D7EEULL, "6087387_f.fxe"},
	{0x384248CFCEC7E2E1ULL, "63683470.ogg"},
	{0x3842B99FA19C73BDULL, "follower_banter6155898.cl"},
	{0x3843819871EAD096ULL, "vxlt_gallowstemplar_f_5ay_0_0.vlm"},
	{0x3843B27235D3152AULL, "llac_dr_slavers_3215_5.dds"},
	{0x3843BCDD13C55F31ULL, "6096635_m.fxe"},
	{0x3843F1F0173B683BULL, "vxht_mansion2_385_0_0.vlm"},
	{0x384427DF9A8DB860ULL, "6051752_m.fxe"},
	{0x38448CE25283BFD0ULL, "weapon2_s2y.psh.fxo"},
	{0x3844A7ED694278DCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye11a.phy"},
	{0x3844A8A7278F45E5ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv3.ani"},
	{0x3844BB3CCF92821DULL, "vxw_woundedcoast_1t9_0_0.vlm"},
	{0x3844D9320430F537ULL, "vxlt_0_lowtown_a2_h20_0_0.vlm"},
	{0x3845670898840E7DULL, "zz_followers_banter6157906.cl"},
	{0x3845A7AB9B8CDE3EULL, "6175489_m.fxe"},
	{0x3845CEE097AD1D54ULL, "gen_im_trs_dwv_med_boo_r1.uti"},
	{0x38461BA8585A5E78ULL, "ch1_s2y.psh.fxo"},
	{0x38463017093CB9F6ULL, "hlac_lt_gallowscourtyard_n_-131072_0z.dds"},
	{0x384669F1F2D40AB6ULL, "vxw_bonepitdestroyed_a3_1jg_0_0.vlm"},
	{0x3846C2CB4AE4E518ULL, "zz_and_debug6178561.cl"},
	{0x3846D406B2708EAFULL, "297403875.ogg"},
	{0x38472491FAC2737BULL, "6114478_m.fxe"},
	{0x384780FFF552D414ULL, "554297831.ogg"},
	{0x38478E423E6ACE07ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv5.evt"},
	{0x3847AAF3E1C9FF25ULL, "follower_banter6143413.cl"},
	{0x3847C6FF4FA80679ULL, "vxw_encounter2_camp_k0_0_0.vlm"},
	{0x3847D83512D32F62ULL, "w_bonepitdestroyed_a3_22_0_lc.mmh"},
	{0x3847DAC5A25B769CULL, "vxdr_primevaltomb_2ul_0_1.vlm"},
	{0x38480114F87C02F3ULL, "233850774.ogg"},
	{0x38483D61E617D353ULL, "vxht_0_hightown_n_c0r_0_0.vlm"},
	{0x3848BC354B7F5DB2ULL, "hlac_ht_chantry_2534_1z.dds"},
	{0x3848D94B9110C608ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_setprops_n.dds"},
	{0x3848FF7FB2533EC5ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magb.phy"},
	{0x38496A2C28970AF6ULL, "w_bonepit_141_0.msh"},
	{0x3849742C28971BF4ULL, "w_bonepit_141_0.mmh"},
	{0x3849782C28972207ULL, "w_bonepit_141_0.mao"},
	{0x38499EF655585F13ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.cess.ani"},
	{0x3849D0FF22EAFCB9ULL, "llac_lt_hangedman_-1_4.dds"},
	{0x3849E2DCFB165133ULL, "mrl310_merrill6120113.cl"},
	{0x384AA5C841B7B234ULL, "hlac_ca_caverns1_mines_24425_7x.dds"},
	{0x384ABE11B060E199ULL, "6179052_f.fxe"},
	{0x384AC12783A0A246ULL, "6065065_m.fxe"},
	{0x384ACEB6F7328825ULL, "193627342.ogg"},
	{0x384B0B565F8C6711ULL, "vrc221_hawke.bnk"},
	{0x384B1444D9AE9FD6ULL, "isa311_isabela6092496.cl"},
	{0x384B4BDDEAC5ED63ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.mmh"},
	{0x384BD82AAC7EA7BEULL, "6027368_m.fxe"},
	{0x384C66BB9588D029ULL, "hlac_lt_0_lowtown_n_-131072_7y.dds"},
	{0x384CBF600E415E25ULL, "hlac_w_bonepit_65536_0x.dds"},
	{0x384CDB969EB64F79ULL, "321637827.ogg"},
	{0x384CDE7A3F2ECE4CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v3.ani"},
	{0x384E0EFC7A04D7C9ULL, "398084360.ogg"},
	{0x384EAF423E712550ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv5.gad"},
	{0x384EC182D84C3F26ULL, "circle.mao"},
	{0x384ED3B6679A057CULL, "6173092_m.fxe"},
	{0x384EED565F901D21ULL, "vrc221_hawke.cnv"},
	{0x384F8AEE0AC51776ULL, "vxlt_gallowstemplar_f_3oq_0_0.vlm"},
	{0x384FD2350DF2ED0AULL, "vxht_mansion2_3d4_0_0.vlm"},
	{0x38502C84BDB3FCE2ULL, "hindsight_belt.dds"},
	{0x38504FB40AB3A18CULL, "hlac_lt_0_lowtown_a2_n_196607_1x.dds"},
	{0x3850BC125461C910ULL, "vxht_keep_6n1_0_0.vlm"},
	{0x3850BE08669E9D4DULL, "and341tr_spider_ambush3.utt"},
	{0x38517C39298C4E90ULL, "357199129.ogg"},
	{0x3851968AA583E093ULL, "418464275.ogg"},
	{0x3851ADBC8562966EULL, "vxca_caverns1_igk_0_0.vlm"},
	{0x3851BE17A76C71D2ULL, "vxht_0_hightown_n_c0w_0_0.vlm"},
	{0x3851C3EB138D3D47ULL, "t3_dps_b04.tnt"},
	{0x3851E0AFAFE9DC5BULL, "cd_ave121_jeven.wbd"},
	{0x3852795C0572B835ULL, "6121511_m.fxe"},
	{0x3852875B9F8F6D28ULL, "vxht_0_hightown_a2_ajn_0_0.vlm"},
	{0x3852B02BC09C3218ULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_p.mmh"},
	{0x3852D601224377BAULL, "6120595_m.fxe"},
	{0x3852EA212882BFD4ULL, "392345615.ogg"},
	{0x385382EB78E51338ULL, "art\\characters\\playercharacter\\elfmale\\cth_hat_comb.phy"},
	{0x3853EB7A3F35028EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v3.gad"},
	{0x3853F617FAC8FCD2ULL, "267846735.ogg"},
	{0x3853FB5698FC7043ULL, "6139667_f.fxe"},
	{0x3855F11985853B13ULL, "vxht_mansion_n_6lc_0_0.vlm"},
	{0x3855F6EE16A2307FULL, "169527755.ogg"},
	{0x385796F655648F4AULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.cess.evt"},
	{0x385836227CC9BF8AULL, "mcr370_negotiation6080224.cl"},
	{0x38589F56B7236F15ULL, "hlac_lt_0_lowtown_docks_a3_65534_1z.dds"},
	{0x3858A784935458A3ULL, "6060283_m.fxe"},
	{0x38593F8F94C15BEFULL, "6087125_m.fxe"},
	{0x3859998AF1186C7CULL, "6164005_m.fxe"},
	{0x3859E67A3F39666DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v3.evt"},
	{0x3859F0B574EAEB0AULL, "vxlt_gallowscourtyard_6yg_0_0.vlm"},
	{0x3859F71B1D9200C9ULL, "mag311_keran6061824.cl"},
	{0x385A0BFCB8DD400DULL, "883640668.ogg"},
	{0x385A0D212695EBBAULL, "hla_ht_0_hightown_-131073_0z.dds"},
	{0x385A38C033BA4D3AULL, "ts_dwarven_03.gda"},
	{0x385AD37E343F23E7ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_act_c.impa.gad"},
	{0x385B3AFAF20A3CEDULL, "6162352_m.fxe"},
	{0x385B5E8CBBF8F7CDULL, "lowtown_fem_prostitute_sta6152317.cl"},
	{0x385B893A7A6BEFCDULL, "hm_one_mo_vanard.mor"},
	{0x385BB77030A7D7CBULL, "6085238_f.fxe"},
	{0x385BD1DA2228BE7FULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundrysmoke_sml.mmh"},
	{0x385C55435FA365D9ULL, "lgt110ar_gang_hideout_d.are"},
	{0x385C60088EB7562CULL, "6098881_m.fxe"},
	{0x385C990514B9D4DAULL, "6142256_m.fxe"},
	{0x385CCBF7D6146A85ULL, "vxlt_0_lowtown_a2_n_c_euo_0_0.vlm"},
	{0x385CD6319FF09119ULL, "follower_banter6155865.cl"},
	{0x385CDFC81F43D8E9ULL, "6036505_m.fxe"},
	{0x385D9B803F2C9325ULL, "574409086.ogg"},
	{0x385DFCCAFCBB2FF6ULL, "vxlt_gallowstemplar_f_3jr_0_0.vlm"},
	{0x385E75F6556A75BDULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.cess.gad"},
	{0x385F1DCCF8CE9529ULL, "vxlt_0_lowtown_a2_n_hcz_0_0.vlm"},
	{0x385FDC1E5D02E080ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm01.phy"},
	{0x38601F997758AEEEULL, "vxca_caverns1_ijl_0_0.vlm"},
	{0x386040885DCFD512ULL, "1065662341.ogg"},
	{0x3860C87E34437CF4ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_act_c.impa.evt"},
	{0x3860D5A9FB27993BULL, "vxht_mansion2_327_0_0.vlm"},
	{0x386169961EE64886ULL, "lt_0_lowtown_docks_a3_n_3_0.rml"},
	{0x386169B6CC27179FULL, "art\\vfx\\combat\\vfx_r_var_activation1_c.mmh"},
	{0x3861BB6BDA600D8CULL, "follower_barks6153504.cl"},
	{0x3861C84686929192ULL, "6069634_m.fxe"},
	{0x38627A1C0B4664FCULL, "372574435.ogg"},
	{0x3863032B988F2420ULL, "follower_banter6171860.cl"},
	{0x38630FB0DC05CAE1ULL, "gen00fl_varric6093595.cl"},
	{0x38631FD5FE577E84ULL, "vxw_bonepitdestroyed_a3_1q4_0_0.vlm"},
	{0x38634E8F79550B6FULL, "vxw_bonepitmining_a2_1oc_0_0.vlm"},
	{0x3863AC5A3F904243ULL, "vxht_keep_622_0_0.vlm"},
	{0x38641C6195C8E68BULL, "470858443.ogg"},
	{0x3864297A61309C28ULL, "6139707_f.fxe"},
	{0x386459897D476A20ULL, "hlac_lt_0_lowtown_-131072_8x.dds"},
	{0x3864814848538DF4ULL, "gen00fl_anders6125790.cl"},
	{0x38649B92389F636BULL, "dae300_vashothlead6062800.cl"},
	{0x3864AAB933943779ULL, "vxw_encounter2_camp_a6_0_0.vlm"},
	{0x3864F7EEEF7F9D7FULL, "vxht_0_hightown_a2_b0l_0_0.vlm"},
	{0x3865007607938824ULL, "qun110_petrice6042003.cl"},
	{0x38668C50689F97F9ULL, "zz_lgt_1_debug6114368.cl"},
	{0x3867047D31DF2903ULL, "91225181.ogg"},
	{0x38670D851A42DFCEULL, "52159956.ogg"},
	{0x386790796666214BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ext.ani"},
	{0x38679643F3774505ULL, "833091993.ogg"},
	{0x3867E4156BFD9784ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall_1.mao"},
	{0x3867E6C3E9285B2DULL, "hlac_lt_gallowstemplar_2382_2z.dds"},
	{0x38684058EB756954ULL, "vxht_keep_n_73p_0_0.vlm"},
	{0x3868870B6CC27512ULL, "hlac_lt_0_lowtown_a2_n_c_1_7z.dds"},
	{0x3868DBD9E520DD02ULL, "one351ip_area_transitions.ncs"},
	{0x38695B371EE6C914ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingc_n.dds"},
	{0x386990370D9CA007ULL, "vxca_caverns1_jfb_0_0.vlm"},
	{0x3869CC47C41C92FFULL, "449037686.ogg"},
	{0x3869FE2B022161F4ULL, "6139542_m.fxe"},
	{0x386AB27E0A54348AULL, "zz_mcr_debug6071542.cl"},
	{0x386AB555C03C4729ULL, "hlac_lt_0_lowtown_-196609_2z.dds"},
	{0x386BAF00E9A6C935ULL, "152442278.ogg"},
	{0x386BF032DFD185ABULL, "vxlt_0_lowtown_a2_n_c_gbr_0_0.vlm"},
	{0x386C6EA7EEB14876ULL, "vxlt_gallowstemplar_f_3qs_0_0.vlm"},
	{0x386C88AE4388C989ULL, "453997428.ogg"},
	{0x386D8E2A91616B92ULL, "w_sundermount_165_0.phy"},
	{0x386E85721E6BB3B2ULL, "6116132_m.fxe"},
	{0x386E9176694EC76EULL, "vxca_caverns1_imm_0_0.vlm"},
	{0x386EA70866B773C0ULL, "and341tr_spider_ambush3.ncs"},
	{0x386EC07E344FAC2DULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_act_c.impa.ani"},
	{0x386F296FBE5A369BULL, "art\\characters\\creatures\\desiredemon\\animation\\p.ani"},
	{0x38700991DE7D9961ULL, "592230336.ogg"},
	{0x3870261A1C010B24ULL, "342750500.ogg"},
	{0x387105494934EACCULL, "6019099_m.fxe"},
	{0x387112498AF9CD19ULL, "1001519912.ogg"},
	{0x3871487CD5097BEBULL, "164329048.ogg"},
	{0x38716FA8570C94E1ULL, "hlac_ht_keep_n_3732_1y.dds"},
	{0x387191B2F04D9704ULL, "vxw_bonepitdestroyed_a3_1z5_0_0.vlm"},
	{0x38720EDACD312BF7ULL, "w_encounter3_88_0.msh"},
	{0x387220DACD314A5AULL, "w_encounter3_88_0.mao"},
	{0x387224DACD315169ULL, "w_encounter3_88_0.mmh"},
	{0x3872A858E3B594E7ULL, "6202124_m.fxe"},
	{0x3872D30F0C36C191ULL, "6125982_m.fxe"},
	{0x3872D4040CDE9D41ULL, "llac_lt_foundry_n_320_4.dds"},
	{0x3872F6FF313D2BDFULL, "vxht_mansion2_2nr_0_0.vlm"},
	{0x3873A7760B14253FULL, "hlac_lt_0_lowtown_a2_n_c_-131072_6x.dds"},
	{0x3873CC5D57C957B4ULL, "74005383.ogg"},
	{0x3873CC71BC9A5008ULL, "vxw_bonepitmining_a2_i5_0_0.vlm"},
	{0x3873D3A7F6AE2D73ULL, "character_character_p130.psh.fxo.dx11"},
	{0x387415FB81DA8EF1ULL, "vxht_0_hightown_c5p_0_0.vlm"},
	{0x387515833107A2A4ULL, "llac_dr_dungeon_3216_6.dds"},
	{0x387532E2E1E8E8F5ULL, "dae300_petrice6097869.cl"},
	{0x38754B1B8DC19C45ULL, "6156449_m.fxe"},
	{0x387571BB438B6680ULL, "706024274.ogg"},
	{0x38757879667235C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ext.evt"},
	{0x3875C97BE694758BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_down.ani"},
	{0x3876426F9A12A45FULL, "vrc321ip_static_book_1.utp"},
	{0x3876AD3410F9C563ULL, "vxdr_dungeon_1mc_0_0.vlm"},
	{0x3876F7EBC976B861ULL, "959175760.ogg"},
	{0x3877172F43A6C54AULL, "1024188356.ogg"},
	{0x38785225EBA78C0BULL, "vxdr_primevalentrance_32r_0_0.vlm"},
	{0x3878D367E672E251ULL, "vxw_blightlands_13z_0_0.vlm"},
	{0x3878F7F470BC1B76ULL, "fex_ruin_arch_a.mmh"},
	{0x387901F470BC2C74ULL, "fex_ruin_arch_a.msh"},
	{0x38797C208DAF48BBULL, "hla_ht_0_hightown_n_1_0y.dds"},
	{0x3879B72E4FC2A80FULL, "6092264_m.fxe"},
	{0x387A098161307FA1ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls108a_0n.dds"},
	{0x387A5879F03296F3ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv1.evt"},
	{0x387AC2F476D66505ULL, "119399579.ogg"},
	{0x387B6E7547C3C299ULL, "vxlt_hangedman_25g_0_0.vlm"},
	{0x387B777966769FD5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_ext.gad"},
	{0x387B83AF661452C0ULL, "6034729_f.fxe"},
	{0x387BAC072B7D6EE2ULL, "vxw_encounter3_bo_0_0.vlm"},
	{0x387BBDA8E0E4A02CULL, "vxht_mansion2_1rv_0_0.vlm"},
	{0x387C16D6DA42C121ULL, "6169326_m.fxe"},
	{0x387C6D1108930343ULL, "90349866.ogg"},
	{0x387C93F611685D4AULL, "vxht_0_hightown_a2_bp5_0_0.vlm"},
	{0x387C9940609D517DULL, "gen00fl_anders6119880.cl"},
	{0x387D234F901AF63EULL, "6036955_m.fxe"},
	{0x387D54DD90DB7D99ULL, "qun131_taskmaster6039525.cl"},
	{0x387D813DF45179BDULL, "vxdr_primevalentrance_24o_0_0.vlm"},
	{0x387E0812867B2D55ULL, "6104907_m.fxe"},
	{0x387E705305DC71E8ULL, "6048729_m.fxe"},
	{0x387E9C2369FB0D49ULL, "6214583_m.fxe"},
	{0x387EF58177726FD5ULL, "hlac_lt_gallowsprison_n_-1_6x.dds"},
	{0x387F43B58C2A4D5EULL, "446531748.ogg"},
	{0x387FB3680B9F9A4AULL, "452364471.ogg"},
	{0x38801593E1F7C5C0ULL, "dre000_varric6208001.cl"},
	{0x3880865B7F4C1E99ULL, "hla_ht_0_hightown_65538_0z.dds"},
	{0x38813BAA8915AB40ULL, "vxlt_0_lowtown_do0_0_0.vlm"},
	{0x38815179F038A934ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv1.gad"},
	{0x38816EC6D1CBB31CULL, "one300_zevran6093499.cl"},
	{0x3881DD5A6BE963F7ULL, "vxlt_warehouse_e7_0_0.vlm"},
	{0x3882218D605C49D1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xc.mmh"},
	{0x38822D8D605C5E72ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xc.mao"},
	{0x388230FCD2D7C9A1ULL, "vxht_0_hightown_n_c5g_0_0.vlm"},
	{0x38823B8D605C758FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xc.msh"},
	{0x3882AF57D5A4F4F5ULL, "art\\characters\\playercharacter\\dwarfmale\\cth_comx.phy"},
	{0x3882C886CE204481ULL, "ca_caverns2_5802.anb"},
	{0x3882CB94F55FDC14ULL, "1036481773.ogg"},
	{0x3882D3CFC84658C3ULL, "368105932.ogg"},
	{0x3883506FBE6B2225ULL, "art\\characters\\creatures\\desiredemon\\animation\\p.gad"},
	{0x388352CE27BCDE22ULL, "gen00fl_fenris6106700.cl"},
	{0x3883B17BE6A08A02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_down.evt"},
	{0x3883F55C93B306FFULL, "989103445.ogg"},
	{0x38846748A808AD96ULL, "484959366.ogg"},
	{0x3884754987E9150DULL, "6135671_m.fxe"},
	{0x38847C5F32793D52ULL, "vxw_woundedcoast_i4_0_0.vlm"},
	{0x3884F91E5D222E15ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm01.msh"},
	{0x3884FB1E5D22318BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm01.mmh"},
	{0x388527DF31EDACF6ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_wkf.ani"},
	{0x38854559D7DC42C6ULL, "pro000_meeran6050067.cl"},
	{0x3885A39E53F11BCCULL, "hlac_lt_warehouse_0_5z.dds"},
	{0x3885C29B9AB88DB9ULL, "art\\vfx\\combat\\vfx_ha_ar_armrestor_act_c.mmh"},
	{0x3886337A77C76287ULL, "hlac_lt_0_lowtown_a2_n_-131071_0y.dds"},
	{0x3886474B2F6E786DULL, "mag330_walter6185091.cl"},
	{0x388729B1A69DA45FULL, "sib100_father_contact6109123.cl"},
	{0x3887604FF3218797ULL, "6033683_m.fxe"},
	{0x38876079F03D2E12ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv1.ani"},
	{0x38877C7E2BF25939ULL, "fx_dazz_circle.mao"},
	{0x38879B12744B268FULL, "24950377.ogg"},
	{0x3887D5D52BEFFA6FULL, "gen00fl_fenris6126860.cl"},
	{0x38881888239A3A5FULL, "dae000_pc_home_ambients6208607.cl"},
	{0x388829F587F1E5A5ULL, "470470230.ogg"},
	{0x388853192F0FB058ULL, "6095828_m.fxe"},
	{0x388866435FC8BDFEULL, "lgt110ar_gang_hideout_d.lst"},
	{0x3888EEE560824194ULL, "lgt260_eau_tmp_ending6171808.cl"},
	{0x388A2F85D2DAB8ACULL, "vxht_mansion2_1uw_0_0.vlm"},
	{0x388A6AB6CC4A1EDCULL, "art\\vfx\\combat\\vfx_r_var_activation1_c.anb"},
	{0x388AB07BE6A6A715ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_down.gad"},
	{0x388B05D3035E75CAULL, "vxht_0_hightown_a2_bu6_0_0.vlm"},
	{0x388B0C0F6DF7B4CBULL, "541275782.ogg"},
	{0x388B2A1FA9C52A35ULL, "426146758.ogg"},
	{0x388B8A8A6ED7A810ULL, "6179317_m.fxe"},
	{0x388BAE33A3DF98BAULL, "vxht_keep_5e1_0_0.vlm"},
	{0x388BCB3E0CF5DD35ULL, "962315609.ogg"},
	{0x388BF941CB31F23AULL, "122919550.ogg"},
	{0x388C60A7BC6D35ADULL, "vxlt_0_lowtown_a2_n_gn6_0_0.vlm"},
	{0x388C876677557984ULL, "gen_im_acc_rng_act2_08.uti"},
	{0x388C8C0D850D104FULL, "vxca_caverns1_cmn_0_0.vlm"},
	{0x388CB7F038DBAB01ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stand_d.dds"},
	{0x388D44A38B0AE2A4ULL, "vxlt_gallowstemplar_f_5o9_0_0.vlm"},
	{0x388D9595BD270B2FULL, "888002856.ogg"},
	{0x388DAA9931D0F90CULL, "970687191.ogg"},
	{0x388E1B238A906372ULL, "zz_charge_forest.cut"},
	{0x388E330E4A0D4DF0ULL, "vxca_cavern1_haunted_ghk_0_0.vlm"},
	{0x388E3C6B871C809EULL, "weapon_distortionmask_p0.vsh.fxo"},
	{0x388EF9B2C7282B0EULL, "vxlt_0_lowtown_hal_0_0.vlm"},
	{0x388F3F67945E562CULL, "gen00fl_merrill6108515.cl"},
	{0x388F56B3D6DD7C66ULL, "mag131_keran6086174.cl"},
	{0x38902308F1C66DDAULL, "6111321_f.fxe"},
	{0x3890655F145DC357ULL, "6149708_m.fxe"},
	{0x3890813923269541ULL, "6095945_f.fxe"},
	{0x389097EE53660670ULL, "128559983.ogg"},
	{0x3890EE70C8C3D15DULL, "6149324_m.fxe"},
	{0x3891043BD79FAF8DULL, "dae000ip_crowd_ff6_talk4.utp"},
	{0x389108F37ECEAABCULL, "vxca_caverns1_mines_e41_0_0.vlm"},
	{0x3891445367235F8DULL, "vxlt_hangedman_n_269_0_0.vlm"},
	{0x389149F712A28E13ULL, "22671520.ogg"},
	{0x389177F4BF823F7FULL, "mag330_walter6064738.cl"},
	{0x389217C8918C9BF4ULL, "128422216.ogg"},
	{0x38922FDF31F844E7ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_wkf.evt"},
	{0x38924F1A7FFFB9B1ULL, "6164949_m.fxe"},
	{0x3892FF81F72F9FF5ULL, "fen110_fenris6100359.cl"},
	{0x389328751229F650ULL, "6125341_m.fxe"},
	{0x38932B95CC830232ULL, "llac_lt_0_lowtown_a2_-1_7.dds"},
	{0x38935B2DA99D780AULL, "62378918.ogg"},
	{0x38939D4BE43D387FULL, "vxdr_primevaltomb_3sa_0_0.vlm"},
	{0x38939F20791210C2ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_table.msh"},
	{0x3893A12079121478ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_table.mmh"},
	{0x3893AD207912281BULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_table.mao"},
	{0x3893B9DE11FC3D1DULL, "6078419_m.fxe"},
	{0x3894DA6D036F7E9AULL, "art\\levels\\races\\design\\animation\\floortrappulled.gad"},
	{0x38955380530FF915ULL, "824523612.ogg"},
	{0x389598CE53757BE9ULL, "456954059.ogg"},
	{0x38966954C1615A76ULL, "one280_merin6116754.cl"},
	{0x3896C9B730027FB3ULL, "vxlt_0_lowtown_docks_coe_0_0.vlm"},
	{0x38976B02D099D614ULL, "fx_explosion02_flip.dds"},
	{0x389774715F6403E2ULL, "823730412.ogg"},
	{0x3897D64E2BE497C1ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_1x.dds"},
	{0x3897DF635510CFD3ULL, "vxw_woundedcoast_c3_0_0.vlm"},
	{0x38981151E313F90EULL, "hlac_ht_0_hightown_a2_65535_3x.dds"},
	{0x3898A162C4D0D12CULL, "vxht_mansion2_1xt_0_0.vlm"},
	{0x3898FA90BE2EF221ULL, "6163126_m.fxe"},
	{0x389950DF31FE9B30ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_wkf.gad"},
	{0x38997A32D667E076ULL, "llac_ca_cavern1_haunted_24426_6.dds"},
	{0x3899B08C0CA54A86ULL, "dae000_pc_home_ambients6208183.cl"},
	{0x389AAEE3C024622EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_basic.mmh"},
	{0x389ABAE3C02476C5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_basic.mao"},
	{0x389AC8E3C0248E1CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_basic.msh"},
	{0x389AEE4FF4F718CBULL, "gen00fl_varric6077825.cl"},
	{0x389AF67CEBC9D07BULL, "vxht_mansion2b_24i_0_1.vlm"},
	{0x389B0612FB5B0C30ULL, "631811149.ogg"},
	{0x389B8F7694EDD4D5ULL, "vxw_encounter2_5n_0_0.vlm"},
	{0x389BB36DF8723A49ULL, "hlac_lt_0_lowtown_65536_2y.dds"},
	{0x389BD89FC322E3A8ULL, "6121554_m.fxe"},
	{0x389BF08F57581DA9ULL, "195124083.ogg"},
	{0x389C48067450AC3BULL, "6013361_f.fxe"},
	{0x389C7E0912DDD9B0ULL, "gencr_city_guard_2h_f.utc"},
	{0x389C85870BABA7D0ULL, "vxdr_0_deeproads_gc0_0_0.vlm"},
	{0x389D31019911D33FULL, "6048810_m.fxe"},
	{0x389D3500FDECF3B9ULL, "177087241.ogg"},
	{0x389DA820791A7403ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_table.phy"},
	{0x389DA85E48FA00EDULL, "gen00fl_dog6167633.cl"},
	{0x389DC2D0821DDE88ULL, "6034380_m.fxe"},
	{0x389DD0A435B300CDULL, "llac_w_bonepit_0_12.dds"},
	{0x389EA5A7F9AAF948ULL, "vxlt_0_lowtown_docks_n_b9n_0_0.vlm"},
	{0x389F1DE15EC27E97ULL, "w_bonepit_63_0.mmh"},
	{0x389F21E15EC285A4ULL, "w_bonepit_63_0.mao"},
	{0x389F23EC0F584126ULL, "dae000_party_night_comment.cnv"},
	{0x389F309372915E82ULL, "vxlt_0_lowtown_docks_c6k_0_0.vlm"},
	{0x389F33E15EC2A409ULL, "w_bonepit_63_0.msh"},
	{0x38A024F8BD77EA89ULL, "6105631_f.fxe"},
	{0x38A0336A72B3EEA3ULL, "hlac_w_woundedcoast_4_0x.dds"},
	{0x38A081659DB7D7E4ULL, "art\\characters\\creatures\\darkspawnogre\\cn_bdy_ogra.phy"},
	{0x38A1284D8503CBBEULL, "w_bonepitdestroyed_a3_36_0.rml"},
	{0x38A1601916656E52ULL, "vxw_woundedcoast_c6_0_0.vlm"},
	{0x38A1906B77C5C2F2ULL, "63998803.ogg"},
	{0x38A19A5204200ED0ULL, "6122241_m.fxe"},
	{0x38A225EC0F5A7AE6ULL, "dae000_party_night_comment.bnk"},
	{0x38A30802DFD3B452ULL, "dae300_alistair_king6134865.cl"},
	{0x38A34E49AE91FF87ULL, "gen_im_trs_crc_junk1_03.uti"},
	{0x38A37CD0EC67D8D4ULL, "251067212.ogg"},
	{0x38A38ECF2A4C65E1ULL, "vxlt_0_lowtown_docks_n_b6j_0_0.vlm"},
	{0x38A45111A5A012ADULL, "818346405.ogg"},
	{0x38A4732C28E3C5EFULL, "w_bonepit_141_0.phy"},
	{0x38A514C9230F366BULL, "vxdr_primevaltomb_7uf_0_1.vlm"},
	{0x38A51883FB82F910ULL, "vxlt_0_lowtown_e1s_0_0.vlm"},
	{0x38A5B370B7FD035CULL, "vxht_keep_n_64d_0_0.vlm"},
	{0x38A5C9FA03A4375AULL, "6054083_m.fxe"},
	{0x38A60596C91FB2DEULL, "dre000_cut_return_surface6030202.cl"},
	{0x38A67378044322FDULL, "6101250_m.fxe"},
	{0x38A6F99F65E25FD0ULL, "rdr110dg_guard_sleeping6016741.cl"},
	{0x38A7133FB6C6E9ACULL, "vxht_mansion2_1cu_0_0.vlm"},
	{0x38A7BF2954012468ULL, "6120990_f.fxe"},
	{0x38A7E98CE74AA6CAULL, "vxht_0_hightown_a2_bc0_0_0.vlm"},
	{0x38A85C6AB3F18999ULL, "186768539.ogg"},
	{0x38A87C97FA7BF1DBULL, "hlac_ca_cavern1_haunted_24422_5z.dds"},
	{0x38A9F56D03820880ULL, "art\\levels\\races\\design\\animation\\floortrappulled.ani"},
	{0x38AA1946004E45D9ULL, "dae000_gamlen6145097.cl"},
	{0x38AA94196A54A440ULL, "plt_cod_ltr_nt_feynriel.plo"},
	{0x38AA971150ED3C6EULL, "llac_w_encounter2_camp_65535_0.dds"},
	{0x38AB0FEF318D16D9ULL, "6120866_m.fxe"},
	{0x38AB4220626F4C70ULL, "20138808.ogg"},
	{0x38AB47717B054651ULL, "hlac_ht_0_hightown_n_65534_1y.dds"},
	{0x38AB5A211A47E756ULL, "vxlt_gallowstemplar_n_2r7_0_0.vlm"},
	{0x38AB87E0332385EFULL, "vxw_bonepitmining_a2_1pn_0_0.vlm"},
	{0x38ABA2DE8B07EBBFULL, "6043910_m.fxe"},
	{0x38AC4D961373F093ULL, "art\\cinematics\\act3\\romances\\anders\\cs_pcmale_andersrom.ani"},
	{0x38ACA9BAB63758A3ULL, "6071937_m.fxe"},
	{0x38ACAA87992161B6ULL, "hlac_lt_0_lowtown_n_-3_6z.dds"},
	{0x38ACB492D98ADB32ULL, "llac_lt_0_lowtown_docks_131071_7.dds"},
	{0x38AD1784EBA111C8ULL, "vxlt_0_lowtown_docks_n_b6o_0_0.vlm"},
	{0x38AD32F141CBAD77ULL, "vxlt_warehouse_t0_0_0.vlm"},
	{0x38ADBF9B9AD9DAAEULL, "art\\vfx\\combat\\vfx_ha_ar_armrestor_act_c.anb"},
	{0x38AE280D4B0F908DULL, "vxlt_hangedman_n_287_0_0.vlm"},
	{0x38AEEE91D50F031FULL, "vxdr_0_deeproads_aff_0_0.vlm"},
	{0x38AF3C0592969067ULL, "vrc321_spooky_scenes6215547.cl"},
	{0x38AF5D43EEF4FEEDULL, "follower_barks6153198.cl"},
	{0x38AF6FEB793351CDULL, "art\\characters\\playercharacter\\elfmale\\cth_hat_comb.msh"},
	{0x38AF81EB79337053ULL, "art\\characters\\playercharacter\\elfmale\\cth_hat_comb.mmh"},
	{0x38B01B8868FC1529ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig2_e.dds"},
	{0x38B07A258562DB44ULL, "vxca_caverns1_mines_e6g_0_0.vlm"},
	{0x38B165D9AE8B147AULL, "777190412.ogg"},
	{0x38B39FD5ED6AE71BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01gravecairn.phy"},
	{0x38B3AD7113EEB0B3ULL, "vxlt_0_lowtown_docks_cug_0_0.vlm"},
	{0x38B3C839691CAAE4ULL, "pro000_athenril6049440.cl"},
	{0x38B3F3C89912845CULL, "hlac_ht_0_hightown_a2_0_3z.dds"},
	{0x38B4A4BC5F95E161ULL, "gen00fl_sebastian6179999.cl"},
	{0x38B4B539FE2E1823ULL, "vxw_bonepitmining_a2_78_0_0.vlm"},
	{0x38B4C31D38FD00D3ULL, "vxw_woundedcoast_u5_0_0.vlm"},
	{0x38B4D067CC249D1DULL, "6161249_m.fxe"},
	{0x38B53BD29A1231D9ULL, "hlac_lt_0_lowtown_docks_-196611_2x.dds"},
	{0x38B5851CA8BD022CULL, "vxht_mansion2_1fj_0_0.vlm"},
	{0x38B5BE4B0CBF60E0ULL, "mrl150_merrill6085831.cl"},
	{0x38B627D7452AD704ULL, "mcr371_hostile_fenris6082240.cl"},
	{0x38B63ABEF20A2367ULL, "6143819_f.fxe"},
	{0x38B65B69D940BF4AULL, "vxht_0_hightown_a2_bd1_0_0.vlm"},
	{0x38B6735B46867502ULL, "gen00fl_aveline6139189.cl"},
	{0x38B6A68A9574521CULL, "vxlt_gallowstemplar_n_2je_0_0.vlm"},
	{0x38B6D9586D8CBC94ULL, "vxlt_undercityencounter_13p_0_0.vlm"},
	{0x38B7206C27E50F9BULL, "937044916.ogg"},
	{0x38B734FAED2C42A0ULL, "30502554.ogg"},
	{0x38B74BB5186FCA15ULL, "6149661_m.fxe"},
	{0x38B77A7DAFC26F68ULL, "449354162.ogg"},
	{0x38B7887EFCE7BFC3ULL, "159375496.ogg"},
	{0x38B8BEB402DCF956ULL, "zz_onept_debug.ncs"},
	{0x38B8C68CA31E3A55ULL, "vxw_blightlands_gg_0_0.vlm"},
	{0x38B8F17FC55ABECFULL, "vxw_encounter2_fi_0_0.vlm"},
	{0x38B9186E82A25BCAULL, "qcr220_grey_wardens6135943.cl"},
	{0x38B94112E7EA0EDBULL, "1021041288.ogg"},
	{0x38B968970D053FCFULL, "32446361.ogg"},
	{0x38BA080A88B45CBAULL, "vxht_0_hightown_c7i_0_0.vlm"},
	{0x38BA76C54603C6D3ULL, "6174058_m.fxe"},
	{0x38BA8C878EDE6649ULL, "6018578_m.fxe"},
	{0x38BAC7573034AAE3ULL, "886676828.ogg"},
	{0x38BAD9A150D70C12ULL, "gen00fl_sebastian6174863.cl"},
	{0x38BAE41207CA890AULL, "6191050_m.fxe"},
	{0x38BB04DEA3E3B6D6ULL, "167604013.ogg"},
	{0x38BB0662B5700FF3ULL, "6137460_m.fxe"},
	{0x38BBA2FB86552CC2ULL, "271670346.ogg"},
	{0x38BBA4CE33C1C5F7ULL, "vxlt_warehouse_y3_0_0.vlm"},
	{0x38BC1376B8C7FC5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lean_on_table_loop.ani"},
	{0x38BC79E8B7D1E416ULL, "t3_dog_on3.tnt"},
	{0x38BC7ECBE6546AF7ULL, "lt_0_lowtown_docks_a3_1_0.anb"},
	{0x38BC8ACBA2A2C9CCULL, "gen00fl_sebastian6189425.cl"},
	{0x38BC964CDF47AF09ULL, "art\\characters\\creatures\\mubari\\animation\\sleep.gad"},
	{0x38BCA2912B281A45ULL, "fen110_fenris6098898.cl"},
	{0x38BCA986CE518811ULL, "ca_caverns2_5802.rml"},
	{0x38BCD2F9A64F69EBULL, "qcr220_orsino6061784.cl"},
	{0x38BD01019FA079A1ULL, "vxht_0_hightown_a2_c7r_0_0.vlm"},
	{0x38BD186314202C49ULL, "191064192.ogg"},
	{0x38BDB52EAB284502ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztable.msh"},
	{0x38BDB72EAB2848B8ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztable.mmh"},
	{0x38BDC17900DD0880ULL, "gen00fl_aveline6132762.cl"},
	{0x38BDC32EAB285D5BULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztable.mao"},
	{0x38BE3FCD6B14AE9AULL, "mag170_recruit02.cnv"},
	{0x38BE43D2FA519F52ULL, "vxw_woundedcoast_u0_0_0.vlm"},
	{0x38BE8557628B8394ULL, "6047248_f.fxe"},
	{0x38BEF533FDB9B529ULL, "825694650.ogg"},
	{0x38BF6DF106F1A92BULL, "75890619.ogg"},
	{0x38BF77B23BA2BD26ULL, "6076061_m.fxe"},
	{0x38BF8B0C2F5D9A61ULL, "6034694_f.fxe"},
	{0x38BFB7E542476476ULL, "vxca_caverns2_6h6_0_0.vlm"},
	{0x38C00B92166E8903ULL, "vxlt_0_lowtown_n_fgz_0_0.vlm"},
	{0x38C08FDA303C4569ULL, "645897161.ogg"},
	{0x38C0910EE9E558FDULL, "hlac_dr_0_deeproads_17038_0x.dds"},
	{0x38C095A2FF81B5C9ULL, "6097276_m.fxe"},
	{0x38C0B09CDBA11848ULL, "branch2.dds"},
	{0x38C0CFE9A4D48B33ULL, "vxht_0_hightown_n_asu_0_0.vlm"},
	{0x38C124961386073DULL, "art\\cinematics\\act3\\romances\\anders\\cs_pcmale_andersrom.gad"},
	{0x38C1431AE61BAA66ULL, "gen00fl_varric6063867.cl"},
	{0x38C15A3778C851F6ULL, "follower_banter6147975.cl"},
	{0x38C161CD6B171EE2ULL, "mag170_recruit02.bnk"},
	{0x38C162BCF8165AE6ULL, "mrl251ip_dead_radha.utp"},
	{0x38C1794644F55737ULL, "plt_zz_lgtpt_1_debug.plo"},
	{0x38C191900E659B64ULL, "hlac_ca_caverns1_24423_12y.dds"},
	{0x38C19E9706DB0DF7ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.fly_long.evt"},
	{0x38C1C69635776A14ULL, "w_blightlands_15_0.rml"},
	{0x38C22DCBFB5D8D6DULL, "mcr370_cullen6087428.cl"},
	{0x38C237CE4DFEF959ULL, "vxw_bonepitmining_a2_6u_0_0.vlm"},
	{0x38C295844C00C5BEULL, "6096496_m.fxe"},
	{0x38C2F16FCE734BFFULL, "hlac_ht_0_hightown_a2_-65537_3x.dds"},
	{0x38C334FA2AF31953ULL, "vxw_woundedcoast_r4_0_0.vlm"},
	{0x38C34244BE1AB59DULL, "6166349_m.fxe"},
	{0x38C355CB60146191ULL, "6062171_m.fxe"},
	{0x38C3674CDF4D7E57ULL, "art\\characters\\creatures\\mubari\\animation\\sleep.ani"},
	{0x38C3723FCAFFE628ULL, "6075347_m.fxe"},
	{0x38C37A9990D9EF60ULL, "vxdr_0_deeproads_egj_0_0.vlm"},
	{0x38C3BA67633CCA58ULL, "vxlt_0_lowtown_a2_f58_0_0.vlm"},
	{0x38C3F6F99AB31AACULL, "vxht_mansion2_1ik_0_0.vlm"},
	{0x38C4516461F67FFCULL, "vxlt_0_lowtown_a2_n_him_0_0.vlm"},
	{0x38C4BF1748BFE491ULL, "6055605_m.fxe"},
	{0x38C51867876A6A9CULL, "vxlt_gallowstemplar_n_2oz_0_0.vlm"},
	{0x38C5BCF41C626A56ULL, "6114018_f.fxe"},
	{0x38C5E173EB70CF49ULL, "vxlt_0_lowtown_docks_a3_cz7_0_0.vlm"},
	{0x38C5ED7E3B74E797ULL, "vxht_mansion2_22n_0_0.vlm"},
	{0x38C663247F8DF1E8ULL, "vxlt_0_lowtown_h8a_0_0.vlm"},
	{0x38C6979851C1AE80ULL, "453343156.ogg"},
	{0x38C6FCAD792E34D2ULL, "704233827.ogg"},
	{0x38C719FE904B701AULL, "vxca_caverns1_mines_edw_0_0.vlm"},
	{0x38C72117D2E889ACULL, "789880411.ogg"},
	{0x38C73869951452D5ULL, "vxw_blightlands_hf_0_0.vlm"},
	{0x38C751AD073AB785ULL, "lt_gamlenshouse_n_3567_ll.mmh"},
	{0x38C7BE2EAB30A943ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztable.phy"},
	{0x38C7D9F3A1903E3BULL, "vxw_encounter3_84_0_0.vlm"},
	{0x38C81D805F770481ULL, "vxht_keep_n_645_0_0.vlm"},
	{0x38C830D96FBD5731ULL, "6062773_m.fxe"},
	{0x38C88F02815AEA86ULL, "art\\vfx\\combat\\vfx_m_bld_sac_act_c.impa.evt"},
	{0x38C8D610B89ACF13ULL, "6175523_m.fxe"},
	{0x38C92B76B8D2AF2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lean_on_table_loop.evt"},
	{0x38C93EECD3E40E00ULL, "gen00fl_bethany6097831.cl"},
	{0x38C95CCC4E92D9EAULL, "6076577_m.fxe"},
	{0x38C9C257D5E11710ULL, "art\\characters\\playercharacter\\dwarfmale\\cth_comx.msh"},
	{0x38C9C7913EB84B13ULL, "6027678_f.fxe"},
	{0x38C9D057D5E12F2AULL, "art\\characters\\playercharacter\\dwarfmale\\cth_comx.mmh"},
	{0x38C9D457D5E13661ULL, "art\\characters\\playercharacter\\dwarfmale\\cth_comx.mao"},
	{0x38CA01BF51920037ULL, "953671303.ogg"},
	{0x38CA567580572B09ULL, "and100_anders6206379.cl"},
	{0x38CA6F80F40078CFULL, "mag360_samson6042076.cl"},
	{0x38CA884A49501284ULL, "character_character_p61.psh.fxo.dx11"},
	{0x38CA8E6EAC1C3368ULL, "llac_lt_0_lowtown_a2_-3_3.dds"},
	{0x38CB01B1E8D86162ULL, "vxht_0_hightown_n_bia_0_0.vlm"},
	{0x38CB72DE91969221ULL, "vxht_0_hightown_a2_c2q_0_0.vlm"},
	{0x38CC8D6D73C812C6ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_2x.dds"},
	{0x38CCB5AFEC47B7D2ULL, "vxw_woundedcoast_r1_0_0.vlm"},
	{0x38CD1C4D8528F236ULL, "w_bonepitdestroyed_a3_36_0.mao"},
	{0x38CD284D852906D5ULL, "w_bonepitdestroyed_a3_36_0.mmh"},
	{0x38CD2A4D85290A4BULL, "w_bonepitdestroyed_a3_36_0.msh"},
	{0x38CD64BFAC159A7FULL, "vxdr_primevaltomb_3gm_0_0.vlm"},
	{0x38CD7997233FDFD0ULL, "gen_im_cft_rec_bmb_04.uti"},
	{0x38CE57B36AA9B395ULL, "20314232.ogg"},
	{0x38CEB14E88F69941ULL, "gen00fl_fenris6126578.cl"},
	{0x38CEFD68B8A57DD0ULL, "6061607_m.fxe"},
	{0x38CF55F1569D1359ULL, "qun110_petrice6042255.cl"},
	{0x38CF9D5F38A9FB1DULL, "llac_lt_0_lowtown_docks_a3_n_-65538_3.dds"},
	{0x38CFA371CCFE4656ULL, "vxlt_gallowstemplar_4uk_0_0.vlm"},
	{0x38CFA602816130D1ULL, "art\\vfx\\combat\\vfx_m_bld_sac_act_c.impa.gad"},
	{0x38CFDC3DD530AB01ULL, "85170073.ogg"},
	{0x38CFE5590277808CULL, "weapon2_d.psh.fxo"},
	{0x38CFFA47352AD4B2ULL, "ts_beasts_spdr_prog_glands.gda"},
	{0x38D02C76B8D8CFD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lean_on_table_loop.gad"},
	{0x38D069F00036B216ULL, "6160872_m.fxe"},
	{0x38D06C8D551D97AEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4.phy"},
	{0x38D0855C30FA5146ULL, "vxlt_blackemporium_u2_0_0.vlm"},
	{0x38D1B690BC8F1DB2ULL, "dm_dougal_thug.mor"},
	{0x38D1BA9C559E1094ULL, "309938976.ogg"},
	{0x38D1CF71735FB214ULL, "781344133.ogg"},
	{0x38D2415968AD699CULL, "vxht_0_hightown_cb4_0_0.vlm"},
	{0x38D2E7E8B7D4D1FFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_higharmoire_sa.dds"},
	{0x38D2F0BE779D786BULL, "6103049_m.fxe"},
	{0x38D2F2981CD03094ULL, "follower_barks6153809.cl"},
	{0x38D300DEA5D58548ULL, "art\\vfx\\combat\\vfx_glb_ele_s_minor_c.anb"},
	{0x38D3AE81070972EEULL, "vxca_caverns1_i2d_0_0.vlm"},
	{0x38D3AFF7B82E3DEEULL, "6121919_f.fxe"},
	{0x38D3CB77040614CCULL, "666070728.ogg"},
	{0x38D48648C1DDF630ULL, "w_encounter1_17_0.mao"},
	{0x38D48A48C1DDFD43ULL, "w_encounter1_17_0.mmh"},
	{0x38D49848C1DE149DULL, "w_encounter1_17_0.msh"},
	{0x38D4B7E85D05D98EULL, "6125160_f.fxe"},
	{0x38D566CCDF771756ULL, "vxht_0_hightown_cag_0_0.vlm"},
	{0x38D58846450619ACULL, "plt_zz_lgtpt_1_debug.nss"},
	{0x38D5A70281659E7FULL, "art\\vfx\\combat\\vfx_m_bld_sac_act_c.impa.ani"},
	{0x38D5AA46870A6B55ULL, "vxw_blightlands_me_0_0.vlm"},
	{0x38D5D539A946EFCFULL, "vxw_encounter2_lk_0_0.vlm"},
	{0x38D5F635A6E9D123ULL, "6161456_m.fxe"},
	{0x38D64BD0938656BBULL, "vxw_encounter3_57_0_0.vlm"},
	{0x38D6A1A90CA998E7ULL, "seb321_leliana6181756.cl"},
	{0x38D6A2B661B36FB1ULL, "6069673_m.fxe"},
	{0x38D6C663F925980EULL, "vxht_0_hightown_a2_dw_0_0.vlm"},
	{0x38D74974487BDC76ULL, "6147853_m.fxe"},
	{0x38D795D289F986A8ULL, "docks_male_elven_worker_sta6175804.cl"},
	{0x38D7F05D0E10E016ULL, "hlac_lt_0_lowtown_a2_n_65535_9y.dds"},
	{0x38D817B8FC293ECBULL, "577214520.ogg"},
	{0x38D8334598475657ULL, "hlac_lt_0_lowtown_a3_n_c_131070_6x.dds"},
	{0x38D851CE28AD2E40ULL, "cth_nb1.tnt"},
	{0x38DA5A8365DF0D7AULL, "6128645_m.fxe"},
	{0x38DABAD8FEB76B14ULL, "714885933.ogg"},
	{0x38DAEE3BE38CA49DULL, "hlac_w_bonepitdestroyed_a3_262146_0z.dds"},
	{0x38DBDAF0413A57F5ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgta_d.dds"},
	{0x38DD5BF18388D131ULL, "6098807_f.fxe"},
	{0x38DDE33DCEC419C3ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_cast_c.anb"},
	{0x38DE24CDBAEF9793ULL, "vxht_brothel_n_60l_0_0.vlm"},
	{0x38DF47D4CBB80C6CULL, "vxht_chantry_359_0_0.vlm"},
	{0x38DF546C7444A6BDULL, "hlac_ht_mansion_n_2995_6z.dds"},
	{0x38E0316DAD1FECAFULL, "art\\vfx\\combat\\vfx_glb_restore_stamina_c.mmh"},
	{0x38E07A64A242A2A2ULL, "391746329.ogg"},
	{0x38E086D5ED91AABAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01gravecairn.msh"},
	{0x38E094D5ED91C2F3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01gravecairn.mao"},
	{0x38E098D5ED91C900ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01gravecairn.mmh"},
	{0x38E0C3CE08A56E96ULL, "gen00fl_dog6166815.cl"},
	{0x38E0D47D5B31A079ULL, "mcr310ar_lowtown_nt.lst"},
	{0x38E117C6463230B2ULL, "vxht_0_hightown_a3_c3p_0_0.vlm"},
	{0x38E152431749956FULL, "trigger_trap.cur"},
	{0x38E1D6C093327182ULL, "6162603_m.fxe"},
	{0x38E23B44AD5A4937ULL, "291033993.ogg"},
	{0x38E25DCB1BB4DAA6ULL, "abi_patch.gda"},
	{0x38E2D1381F611897ULL, "vxht_mansion2_28h_0_0.vlm"},
	{0x38E351A41B1D3203ULL, "dae300_cullen6135982.cl"},
	{0x38E366ECD79AE457ULL, "6078523_m.fxe"},
	{0x38E38118CC4791F5ULL, "6172241_m.fxe"},
	{0x38E3D7F6821DDBFAULL, "34884023.ogg"},
	{0x38E3E6D41F1E8926ULL, "6139463_m.fxe"},
	{0x38E4053748E6231FULL, "mag131_keran6086270.cl"},
	{0x38E448BE51562E38ULL, "6125250_f.fxe"},
	{0x38E4681298DFE9A3ULL, "6166556_m.fxe"},
	{0x38E468D4A51AB961ULL, "604490561.ogg"},
	{0x38E50DE7DD0824F0ULL, "1022400593.ogg"},
	{0x38E58B48C1EC7268ULL, "w_encounter1_17_0.phy"},
	{0x38E59E61AA531FA8ULL, "qun200_party_comment6090345.cl"},
	{0x38E5BB513A71F4F6ULL, "6140953_m.fxe"},
	{0x38E5CDC60E779498ULL, "llac_ht_0_hightown_131071_0.dds"},
	{0x38E5E7ECECC0D829ULL, "lgt110_arl_bann_teagan6115569.cl"},
	{0x38E6B3BC9AEA7E34ULL, "vxht_0_hightown_a2_cb8_0_0.vlm"},
	{0x38E6FBBE48B706DBULL, "llac_ca_caverns1_mines_24424_12.dds"},
	{0x38E70F1DFC6337FEULL, "zz_arishock_army.cut"},
	{0x38E779F0C7F641DEULL, "and100_anders6111670.cl"},
	{0x38E79CC6A0332101ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\blendtree\\trooper_movement.anb"},
	{0x38E7CC7D5B37B150ULL, "mcr310ar_lowtown_nt.ncs"},
	{0x38E8A08FEE3AE606ULL, "202220338.ogg"},
	{0x38E8C4262F18E346ULL, "234799338.ogg"},
	{0x38E8DD5E3C4AF5E9ULL, "hm_amb_tailor.mor"},
	{0x38E97BCBE67B530BULL, "lt_0_lowtown_docks_a3_1_0.rml"},
	{0x38EA8EB23BFC9551ULL, "6091682_f.fxe"},
	{0x38EB0462994651C8ULL, "668960337.ogg"},
	{0x38EB35680DEC5FC2ULL, "vxht_chantry_n_260_0_0.vlm"},
	{0x38EB4BBC07ED9BDBULL, "vxlt_0_lowtown_a2_n_gy9_0_0.vlm"},
	{0x38EBC03DCED01BECULL, "art\\vfx\\combat\\vfx_d_qun_aoe_cast_c.mmh"},
	{0x38EC8A48C1F28F60ULL, "w_encounter1_17_0.rml"},
	{0x38ED25211F3889CAULL, "vxlt_0_lowtown_a2_hd3_0_0.vlm"},
	{0x38ED6EEC0E3A099CULL, "vxht_0_hightown_a3_ay2_0_0.vlm"},
	{0x38ED7396F84CB7A7ULL, "hlac_w_encounter2_camp_327682_0y.dds"},
	{0x38ED854B811BC83FULL, "375160896.ogg"},
	{0x38EDAAC4373F839AULL, "vxht_0_hightown_a2_byr_0_0.vlm"},
	{0x38EE4799532B3215ULL, "llac_dr_slavers_3214_7.dds"},
	{0x38EE723FE94E6430ULL, "vxca_cavern1_haunted_hu7_0_0.vlm"},
	{0x38EEB598D59CCC96ULL, "vxlt_gallowstemplar_n_5iy_0_0.vlm"},
	{0x38EF3662A083DAD8ULL, "mag291_ntbanter6056830.cl"},
	{0x38EF89A338284932ULL, "vxht_0_hightown_a3_c4q_0_0.vlm"},
	{0x38EFA83FED20C19AULL, "hlac_ht_chantry_n_2534_6x.dds"},
	{0x38EFCAE069C92DC3ULL, "smoke.mao"},
	{0x38F0169D240B9235ULL, "vxlt_0_lowtown_a2_n_go1_0_0.vlm"},
	{0x38F022B359CF06A3ULL, "6156155_m.fxe"},
	{0x38F030BB7DC8508DULL, "ser120_ghyslain6105077.cl"},
	{0x38F07F15DDEE16A7ULL, "lgt311ar_gang_hideout_lt.lst"},
	{0x38F0E50676D8FC16ULL, "6149169_m.fxe"},
	{0x38F14A285E1606ADULL, "697375363.ogg"},
	{0x38F175EA0013DC99ULL, "ht_0_hightown_a3_51_0_ll.mmh"},
	{0x38F19091EBAC243BULL, "vxlt_0_lowtown_a2_er5_0_0.vlm"},
	{0x38F1CD7D5B40073EULL, "mcr310ar_lowtown_nt.are"},
	{0x38F1F77477581720ULL, "591699421.ogg"},
	{0x38F2B6A2144C05CDULL, "805798453.ogg"},
	{0x38F2B8F38D3320CFULL, "vxw_encounter2_ru_0_0.vlm"},
	{0x38F2BEDC3C14D302ULL, "vxht_mansion2b_3gh_0_1.vlm"},
	{0x38F428BA31987D15ULL, "vxht_brothel_n_32m_0_0.vlm"},
	{0x38F4645AAFCF91F7ULL, "dae000_orana.stg"},
	{0x38F4736ED056EA1BULL, "hlac_ht_keep_n_3735_1z.dds"},
	{0x38F47481CAB696B7ULL, "6165139_m.fxe"},
	{0x38F4B4DCBCC539D1ULL, "dae000ip_crowd_em1_talk18.utp"},
	{0x38F4E43D34B9B6BEULL, "hlac_lt_undercitydungeon_-1_50z.dds"},
	{0x38F53C47731DA91AULL, "357339462.ogg"},
	{0x38F5CD72E250DB24ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_act_c.impa.ani"},
	{0x38F5D56FD33D559FULL, "one300lia_aveline_st.stg"},
	{0x38F6260E23583DA7ULL, "follower_banter6141444.cl"},
	{0x38F6E551023DFCF4ULL, "6036676_m.fxe"},
	{0x38F7288EA258BC83ULL, "710785984.ogg"},
	{0x38F73BAF570859DEULL, "hlac_ca_cavern1_haunted_24425_0y.dds"},
	{0x38F78E856ADCCCD7ULL, "324086567.ogg"},
	{0x38F7C68DC90822F6ULL, "mer340_cowan.cnv"},
	{0x38F7EFDEDCFACA7CULL, "da2terrain_terrain_tools_p0.vsh.fxo"},
	{0x38F8023ADE8EA297ULL, "qcr000_arishok6060117.cl"},
	{0x38F89CB4058B7D4EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon_sa.dds"},
	{0x38F8AD610947B489ULL, "491295199.ogg"},
	{0x38F8D176D65CCADFULL, "qun131_warrens_thug6019156.cl"},
	{0x38F8D533DA7DF88AULL, "vxlt_gallowscourtyard_66z_0_0.vlm"},
	{0x38F8E02A97F93FD5ULL, "dae000_elegant6166942.cl"},
	{0x38F95C1C3B6EE419ULL, "vxlt_foundry_n_12t_0_0.vlm"},
	{0x38F98C803905FBFFULL, "vxht_keep_n_5wu_0_0.vlm"},
	{0x38F98EFA58AE970FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffa_n.dds"},
	{0x38F9A744FFE27842ULL, "vxht_chantry_n_211_0_0.vlm"},
	{0x38F9BD98F9E3B45BULL, "vxlt_0_lowtown_a2_n_gv8_0_0.vlm"},
	{0x38F9D27C36193AB3ULL, "gen00fl_merrill6142093.cl"},
	{0x38FA02B5FB04E9D9ULL, "752603567.ogg"},
	{0x38FAB1D57068E9C1ULL, "120207913.ogg"},
	{0x38FABEEE6E771376ULL, "vxlt_gallowstemplar_n_3gq_0_0.vlm"},
	{0x38FAF41F35FD6FBBULL, "988247438.ogg"},
	{0x38FB40261DD2E9D3ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v3.ani"},
	{0x38FB82C009723CA9ULL, "6141500_m.fxe"},
	{0x38FBC69635A8E2D7ULL, "w_blightlands_15_0.mmh"},
	{0x38FBC88DC90C0ED6ULL, "mer340_cowan.bnk"},
	{0x38FBCA72E2554186ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_act_c.impa.gad"},
	{0x38FBCA9635A8E9E4ULL, "w_blightlands_15_0.mao"},
	{0x38FBCD6384D2B8A5ULL, "6129476_m.fxe"},
	{0x38FBDC9635A90849ULL, "w_blightlands_15_0.msh"},
	{0x38FBDFE0911DB6EBULL, "follower_banter6164841.cl"},
	{0x38FBFFDEA5F888E3ULL, "art\\vfx\\combat\\vfx_glb_ele_s_minor_c.mmh"},
	{0x38FC79F54F233B36ULL, "6060277_m.fxe"},
	{0x38FC9E959EAD508FULL, "llac_lt_0_lowtown_a2_n_0_1.dds"},
	{0x38FD09890F57CDC8ULL, "hlac_ht_0_hightown_-65534_1z.dds"},
	{0x38FDDA9C8494CEB6ULL, "vxlt_0_lowtown_ded_0_0.vlm"},
	{0x38FE43433B1DD837ULL, "103784695.ogg"},
	{0x38FF2A07DD964778ULL, "llac_lt_darktown_-1_60.dds"},
	{0x3900026EDDA23CBBULL, "vxlt_0_lowtown_a2_eu4_0_0.vlm"},
	{0x39007977BC3439A8ULL, "hlac_dr_primevalentrance_1725_1x.dds"},
	{0x3900A8CD39D3BBC6ULL, "723792349.ogg"},
	{0x39015E9B7D24BBD7ULL, "zz_gen_im_cut_invis_staff.ncs"},
	{0x39016F1D902AEB54ULL, "6020412_f.fxe"},
	{0x39018015DDFC8BDCULL, "lgt311ar_gang_hideout_lt.are"},
	{0x3902889A56D8EF0EULL, "vxlt_0_lowtown_h9d_0_0.vlm"},
	{0x3902C572E25B5865ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_act_c.impa.evt"},
	{0x390381196AA0ACCFULL, "plt_cod_ltr_nt_feynriel.nss"},
	{0x390440ED94539084ULL, "6137283_m.fxe"},
	{0x390477EF714CAA90ULL, "one140cr_charade6137075.cl"},
	{0x3904955B7020B39BULL, "456229135.ogg"},
	{0x39056B07AB914DBAULL, "plt_qun200pt_blackform2.nss"},
	{0x390578E416745432ULL, "530952183.ogg"},
	{0x39058792CA7A4C57ULL, "mag321_comte_de_launcet6064923.cl"},
	{0x3905F82337BC4D4CULL, "34550364.ogg"},
	{0x3905FB0414088EB9ULL, "zz_dae_debug6142951.cl"},
	{0x39064E654E3F46C0ULL, "6053174_m.fxe"},
	{0x3906E9758239C6E4ULL, "748134494.ogg"},
	{0x3907095AB9D06777ULL, "vxlt_blackemporium_170_0_0.vlm"},
	{0x39076FE450581221ULL, "173128227.ogg"},
	{0x3907A5965D8D2E60ULL, "6129734_m.fxe"},
	{0x3907FE5D2AFC147FULL, "vxht_keep_n_5xt_0_0.vlm"},
	{0x3908066BDC4CEE5DULL, "6149175_m.fxe"},
	{0x39082568DCAABA6BULL, "ico_mm_plotgiver2.dds"},
	{0x390838261DDD660AULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v3.evt"},
	{0x3908715A79B5FBCFULL, "409119348.ogg"},
	{0x3908F8D2F9D58AADULL, "330499005.ogg"},
	{0x3909326DAD42F3ECULL, "art\\vfx\\combat\\vfx_glb_restore_stamina_c.anb"},
	{0x3909FF78D02C2EBFULL, "vxht_chantry_356_0_0.vlm"},
	{0x390A3F2DC88F9D3CULL, "hlac_ht_0_hightown_a3_-131073_0x.dds"},
	{0x390A8E7E1B2BB49AULL, "vxht_0_hightown_a2_bgl_0_0.vlm"},
	{0x390A9B3423667ED0ULL, "358407773.ogg"},
	{0x390AE746CB52E469ULL, "ef_progenitor_2.mor"},
	{0x390B58BD8E674BF2ULL, "vxw_bonepit_28n_0_0.vlm"},
	{0x390C1F94C4445DBBULL, "173118785.ogg"},
	{0x390C4AB8C96298B0ULL, "mag210_arianni6021211.cl"},
	{0x390C767029323FBEULL, "seb201_lady_harimann6197134.cl"},
	{0x390C8C85D703F7C6ULL, "625509447.ogg"},
	{0x390CEEECE50329F6ULL, "411048101.ogg"},
	{0x390D034917D560FCULL, "vxht_0_hightown_a3_au6_0_0.vlm"},
	{0x390D73F7577CE6E5ULL, "pro000_colonel_ewald6005413.cl"},
	{0x390D91A97EF9D460ULL, "6167022_m.fxe"},
	{0x390E744BCF98553BULL, "vxlt_0_lowtown_a2_ex7_0_0.vlm"},
	{0x390E84C93586F80DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masa_0s.dds"},
	{0x390F14CE940AD2F5ULL, "montage_subs_demohack6217090.cl"},
	{0x390F17261DE34D7DULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v3.gad"},
	{0x390FF134A3F9D8AAULL, "lgt000cr_baz_sketch.utc"},
	{0x390FF59414F0B708ULL, "6083598_f.fxe"},
	{0x39105B6E2086CD7EULL, "986411018.ogg"},
	{0x39106B8D55537A71ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4.mmh"},
	{0x39108800ADEA2C72ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower1_fenrintro.gad"},
	{0x3910D189CB11E2BDULL, "vxht_0_hightown_n_akc_0_0.vlm"},
	{0x39113FD03E5891D0ULL, "vxlt_warehouse_n_r0_0_0.vlm"},
	{0x391162137722A23AULL, "hlac_lt_house_-1_6x.dds"},
	{0x3911A10C84F57342ULL, "hlac_lt_0_lowtown_docks_-3_5y.dds"},
	{0x3911D48F21BE6DC4ULL, "lt_0_lowtown_docks_99_0.msh"},
	{0x3911E68F21BE8C2DULL, "lt_0_lowtown_docks_99_0.mao"},
	{0x3911E7B954E006A1ULL, "411336667.ogg"},
	{0x3911EA8F21BE9366ULL, "lt_0_lowtown_docks_99_0.mmh"},
	{0x39121961294575F5ULL, "llac_lt_0_lowtown_docks_-65538_5.dds"},
	{0x391257775C783A83ULL, "gen00fl_isabela6099317.cl"},
	{0x3912D5AAE43CFCF3ULL, "vxlt_hangedman_n_25n_0_0.vlm"},
	{0x39131DFA18D1C3A4ULL, "6053857_m.fxe"},
	{0x39134E4DB77A80FFULL, "dre000_sibling_has_blight6037041.cl"},
	{0x3913ED932DA25896ULL, "316993384.ogg"},
	{0x3914066E82371400ULL, "vxw_encounter2_d0_0_0.vlm"},
	{0x391427D3949492A7ULL, "cli_armies.gda"},
	{0x39145CBB71C8EAA8ULL, "251243872.ogg"},
	{0x3914779F04A8613BULL, "vxlt_house_hc_0_0.vlm"},
	{0x3914974CAD7A893BULL, "6041578_m.fxe"},
	{0x3914A596F802FFA5ULL, "rock_ref.dds"},
	{0x3914AD58867EDBF9ULL, "follower_banter6170011.cl"},
	{0x3915192107FA4F6DULL, "208299061.ogg"},
	{0x391533E2372665FEULL, "vxlt_0_lowtown_a2_n_gz5_0_0.vlm"},
	{0x39163F458DF8DAE1ULL, "zz_qcr_debug6058854.cl"},
	{0x39166EA2F1BB4F82ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_weightshift.gad"},
	{0x39173C356F8FCA56ULL, "gen00fl_aveline6139268.cl"},
	{0x3918079AE8A2C863ULL, "197831460.ogg"},
	{0x3918556BCEDE84E9ULL, "513195148.ogg"},
	{0x3918DC261905BBB8ULL, "vxht_mansion2b_1gu_0_0.vlm"},
	{0x3918E7CF18EF418BULL, "6017498_m.fxe"},
	{0x3919098D5D82F0DBULL, "6119971_f.fxe"},
	{0x39193916635CFE75ULL, "vxdr_primevaltomb_7xz_0_0.vlm"},
	{0x3919CF1050206112ULL, "891768537.ogg"},
	{0x391A77246E6C4AD7ULL, "6089646_m.fxe"},
	{0x391BA9FED63359CAULL, "6116370_m.fxe"},
	{0x391BF253D524BBB4ULL, "w_woundedcoast_233_0.phy"},
	{0x391C7896CE66A032ULL, "gen00fl_isabela6088068.cl"},
	{0x391C9C7E10A47B9BULL, "lgt110_gxa_vgk_conspirator3.bnk"},
	{0x391CB820F38FE33EULL, "488949741.ogg"},
	{0x391CEFCF69541454ULL, "llac_lt_gallowscourtyard_n_-131071_1.dds"},
	{0x391D081DF201FFB2ULL, "hlac_lt_0_lowtown_-196608_0x.dds"},
	{0x391D73F3BBD4F2DDULL, "character_character_p226.psh.fxo.dx11.assembly"},
	{0x391D91A2F1C1AA79ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_weightshift.evt"},
	{0x391DA4721770AEE8ULL, "208822468.ogg"},
	{0x391DE87E7FD93593ULL, "gen_im_trs_tmp_insignia_03.uti"},
	{0x391E5ABAA7713A05ULL, "501983230.ogg"},
	{0x391E611C6C2113BAULL, "ht_chantry_mmap.dds"},
	{0x391E91F353425851ULL, "467037338.ogg"},
	{0x391E98A7FD4DD58FULL, "vxlt_0_lowtown_docks_c08_0_0.vlm"},
	{0x391ED5925428E3BDULL, "vxdr_slavers_13u_0_0.vlm"},
	{0x391F715532D30396ULL, "vxlt_0_lowtown_docks_n_bf1_0_0.vlm"},
	{0x391F9E7E10A6B55BULL, "lgt110_gxa_vgk_conspirator3.cnv"},
	{0x391FB1AD304EAA50ULL, "vxlt_warehouse_n_y1_0_0.vlm"},
	{0x391FBE6D0D15BD92ULL, "vxlt_0_lowtown_a2_n_cml_0_0.vlm"},
	{0x392016B75C7F7930ULL, "vxw_encounter2_cu_0_0.vlm"},
	{0x3920243433133EF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_down_loop.gad"},
	{0x3920BCFBA129992CULL, "430844683.ogg"},
	{0x392110FDBCDF9129ULL, "testgenlock.dds"},
	{0x39216C290985FA53ULL, "vxca_caverns1_ec2_0_0.vlm"},
	{0x3921EE2C18AF5E00ULL, "vxht_0_hightown_a2_auw_0_0.vlm"},
	{0x3921F7595F9D1CE8ULL, "vxht_chantry_n_2ic_0_0.vlm"},
	{0x3922829C4B80BF37ULL, "vxw_sundermount_34a_0_0.vlm"},
	{0x3922C86EEBBB7632ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink.gad"},
	{0x3922D0D1A54E83FDULL, "gen00fl_aveline6139676.cl"},
	{0x39235697E502AD4DULL, "607637808.ogg"},
	{0x3923943E31988F83ULL, "220586468.ogg"},
	{0x3923A5BF291C7E7EULL, "vxlt_0_lowtown_a2_n_gw6_0_0.vlm"},
	{0x3923D7AD8B9868D1ULL, "hlac_lt_gallowscourtyard_n_65539_0z.dds"},
	{0x392403DC1D6E418EULL, "hlac_ht_chantry_2534_25x.dds"},
	{0x39245F6191820150ULL, "vxlt_darktown_4u9_0_0.vlm"},
	{0x39247629C05ABA4CULL, "6119694_m.fxe"},
	{0x39248973EC601A4AULL, "112432515.ogg"},
	{0x3924EC7E8E7AEC1CULL, "vxlt_0_lowtown_n_exg_0_0.vlm"},
	{0x39257300ADFC6418ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower1_fenrintro.ani"},
	{0x3925CF53D52CD4B1ULL, "w_woundedcoast_233_0.msh"},
	{0x3925DD53D52CECCCULL, "w_woundedcoast_233_0.mao"},
	{0x3925E953D52D016FULL, "w_woundedcoast_233_0.mmh"},
	{0x3926090794080663ULL, "594624049.ogg"},
	{0x39263F5B20E30454ULL, "ser121_jethann6106925.cl"},
	{0x392694F13F282F83ULL, "6070888_m.fxe"},
	{0x39269E59AA41EA82ULL, "face2_dy.psh.fxo"},
	{0x3926A8D7FCD74970ULL, "862610098.ogg"},
	{0x3926B8A073683493ULL, "588165260.ogg"},
	{0x3926DC4D2AFC156BULL, "6167277_f.fxe"},
	{0x3926EE019C9909BCULL, "6115337_m.fxe"},
	{0x3926FF3BBF8C1230ULL, "vxlt_undercitydungeon_1g5_0_0.vlm"},
	{0x3927033772B90AFAULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammer.mmh"},
	{0x3927073772B91131ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammer.mao"},
	{0x3927153772B92940ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammer.msh"},
	{0x39275D0EC3467174ULL, "follower_banter6160642.cl"},
	{0x39276369F7403162ULL, "vxlt_0_lowtown_a2_eww_0_0.vlm"},
	{0x3927863A77CD1AA5ULL, "358564686.ogg"},
	{0x392863BC009263B1ULL, "6178368_m.fxe"},
	{0x3928A7779621D4BAULL, "vxdr_primevaltomb_1cv_0_0.vlm"},
	{0x3929044BAC373B09ULL, "vxlt_0_lowtown_n_fab_0_0.vlm"},
	{0x3929319EAF3F1675ULL, "145894949.ogg"},
	{0x3929BCAA7A07199DULL, "gen00fl_merrill6090271.cl"},
	{0x3929CB6EEBC19A69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink.evt"},
	{0x3929F19706132C0AULL, "vxlt_0_lowtown_e0n_0_0.vlm"},
	{0x3929F4193A0A9A5BULL, "gen00fl_anders6119297.cl"},
	{0x392A0AACB0C63881ULL, "152055407.ogg"},
	{0x392A7E5AAFFD4D58ULL, "dae000_orana.cnv"},
	{0x392AECDECADA98D2ULL, "vxca_caverns1_ec7_0_0.vlm"},
	{0x392B7C022F97ED20ULL, "wm_tbe_city_night.map"},
	{0x392B99A2F1CDF568ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_weightshift.ani"},
	{0x392B99D29ADC0C68ULL, "hlac_dr_dungeon_3217_6y.dds"},
	{0x392BB746003BF773ULL, "239426131.ogg"},
	{0x392C9BE204B3013FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_maga_0s.dds"},
	{0x392CD58BCE538DE8ULL, "vxlt_0_lowtown_a2_n_c_ei2_0_0.vlm"},
	{0x392D00EF17D08AA8ULL, "6165445_f.fxe"},
	{0x392D03F20AA98D9FULL, "71599123.ogg"},
	{0x392D613AE25C1EB6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_04fern.msh"},
	{0x392D6B3AE25C2FB4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_04fern.mmh"},
	{0x392D6F3AE25C36C7ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_04fern.mao"},
	{0x392DAB0D83EBD103ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtc_0t.dds"},
	{0x392DB7BF7495A2D5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightlighta_n.dds"},
	{0x392DB882FCAE271EULL, "6071121_m.fxe"},
	{0x392E2BF7388962F8ULL, "pro000_varric6170658.cl"},
	{0x392E4BEB472E7573ULL, "vxht_bartrand_n_4bh_0_0.vlm"},
	{0x392E605AB0010360ULL, "dae000_orana.bnk"},
	{0x392E88944E7591B0ULL, "vxw_encounter2_nr_0_0.vlm"},
	{0x392F209B9EB3942AULL, "6025142_m.fxe"},
	{0x392F671461B5F8C6ULL, "6177817_f.fxe"},
	{0x392FDA3B89398F14ULL, "character_rendertodepth_p2.vsh.fxo.dx11"},
	{0x39301F0B6CE97CD1ULL, "6067636_f.fxe"},
	{0x393060090AA57680ULL, "vxht_0_hightown_a2_apt_0_0.vlm"},
	{0x39310C09859D3248ULL, "hlac_ht_0_hightown_0_0y.dds"},
	{0x39317B93177BB3F8ULL, "dae000_merchant_weapons_dm6164404.cl"},
	{0x3931CC936D4E2349ULL, "6217968_m.fxe"},
	{0x3931D03345B869CFULL, "follower_barks6153387.cl"},
	{0x3931D9FE95038971ULL, "298073460.ogg"},
	{0x3931F0AC6678740DULL, "hlac_lt_gallowstemplar_f_2388_5z.dds"},
	{0x3932DCC8CBAAD195ULL, "dae000ip_crowd_lm1_talk8.utp"},
	{0x3932FBB21FD468FAULL, "vxht_mansion_n_5af_0_0.vlm"},
	{0x39331F4D0643CD1AULL, "gen00fl_isabela6096953.cl"},
	{0x393353F400DE5DFFULL, "vxht_keep_n_5kq_0_0.vlm"},
	{0x39335E5B8071049CULL, "vxlt_0_lowtown_n_eud_0_0.vlm"},
	{0x39337E96F81D6E4FULL, "rock_ref.mao"},
	{0x39338A09F87DADFAULL, "6028426_m.fxe"},
	{0x393456D173DC682AULL, "hlac_ht_mansion2_818_10y.dds"},
	{0x3934FC7B1D2DAB3BULL, "plt_vrc000pt_n_gift.nss"},
	{0x393501FC4AED78ACULL, "art\\vfx\\materials\\vfx_smoke_soft.mao"},
	{0x39350F3433257698ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_down_loop.ani"},
	{0x39357118B1822AB0ULL, "vxlt_undercitydungeon_1j2_0_0.vlm"},
	{0x3936392A750B2843ULL, "vxlt_warehouse_cm_0_0.vlm"},
	{0x393683CEF73AB256ULL, "vxlt_0_lowtown_n_ev7_0_0.vlm"},
	{0x3936B36EEBCBFBD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink.ani"},
	{0x3936D12ECA39A694ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_noba_0s.dds"},
	{0x3936F0A6544823C2ULL, "hlac_w_bonepit_65539_0z.dds"},
	{0x3936F789598AAFF6ULL, "vxlt_undercityencounter_z2_0_0.vlm"},
	{0x393700C1F9A1A358ULL, "6087665_m.fxe"},
	{0x39372682EED80909ULL, "qun200pt_missing_qunari.ncs"},
	{0x3937719DC8436D2FULL, "gen00fl_isabela_init.bnk"},
	{0x3937823C70E2026EULL, "mrl000_marethari6129447.cl"},
	{0x3937E2104F841BD9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\dwoverlay.anb"},
	{0x39382D387249334FULL, "vxht_brothel_n_6ce_0_0.vlm"},
	{0x3938A79327951945ULL, "lla_ht_0_hightown_a3_2_0.dds"},
	{0x39392CDF821F51AFULL, "326275150.ogg"},
	{0x393A40B5525E5A55ULL, "6208366_m.fxe"},
	{0x393A939DC845DC97ULL, "gen00fl_isabela_init.cnv"},
	{0x393A9F0F95E20D46ULL, "531528855.ogg"},
	{0x393B01E8AAD5DFE7ULL, "vxht_keep_n_8ar_0_0.vlm"},
	{0x393B0A2EC1037F20ULL, "vxlt_0_lowtown_docks_b11_0_0.vlm"},
	{0x393B463925C941B6ULL, "vxw_bonepit_1pi_0_0.vlm"},
	{0x393B4768C049A668ULL, "vxlt_0_lowtown_a2_n_c_en3_0_0.vlm"},
	{0x393BE1F363E14848ULL, "6112053_m.fxe"},
	{0x393C54813C36FF77ULL, "gen_im_trs_tmp_med_glv_r1.uti"},
	{0x393CFA71406BAA30ULL, "vxw_encounter2_is_0_0.vlm"},
	{0x393CFD4F3252CA87ULL, "286181054.ogg"},
	{0x393D28E10B1E4BBAULL, "440319197.ogg"},
	{0x393D43EA85795025ULL, "461541253.ogg"},
	{0x393D8D958D603651ULL, "hlac_lt_0_lowtown_a2_0_0x.dds"},
	{0x393DD83FB28767ECULL, "gen00fl_isabela6078991.cl"},
	{0x393E342538845EDFULL, "vxlt_gallowstemplar_434_0_0.vlm"},
	{0x393E4FE2ED722B53ULL, "vxca_caverns1_ei4_0_0.vlm"},
	{0x393F3D805EAD0E9CULL, "6111719_m.fxe"},
	{0x393FA78AD1AA063BULL, "vxdr_primevalentrance_331_0_0.vlm"},
	{0x393FC3CC1D21509BULL, "vxdr_primevaltomb_7ys_0_0.vlm"},
	{0x39401B3A7693003BULL, "llac_lt_0_lowtown_a3_n_c_65537_0.dds"},
	{0x39405BFFE669DB37ULL, "396012688.ogg"},
	{0x394065B9BABAF463ULL, "vxht_0_hightown_a2_bwa_0_0.vlm"},
	{0x394069CF15548544ULL, "662547996.ogg"},
	{0x3940F12833D112ADULL, "vxw_sundermount_2zy_0_1.vlm"},
	{0x3940F1788029F40BULL, "vxht_mansion_n_6ob_0_0.vlm"},
	{0x3941F35EDF5AC14BULL, "hlac_ca_caverns2prc_5802_31y.dds"},
	{0x394235C61429F7E6ULL, "vxht_brothel_n_5kw_0_0.vlm"},
	{0x39424D2D6F130DC9ULL, "6153820_m.fxe"},
	{0x39426D200D146922ULL, "tal_04.dds"},
	{0x39431393726AA5FFULL, "998148147.ogg"},
	{0x39431860043E0341ULL, "vxw_bonepitmining_a2_7b_0_0.vlm"},
	{0x39434BAD23700561ULL, "27345369.ogg"},
	{0x394431BCEEE80538ULL, "vxht_mansion2b_1pp_0_0.vlm"},
	{0x39447E91FE9546C4ULL, "6103244_m.fxe"},
	{0x3944C0C1E0B47443ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_screen_sa.dds"},
	{0x3944F138318E1B7EULL, "vxw_sundermount_356_0_1.vlm"},
	{0x3944F5ABE930CAD6ULL, "vxlt_0_lowtown_n_es8_0_0.vlm"},
	{0x3944F678801C7BE0ULL, "vxca_cave_uk_0_1.vlm"},
	{0x39455F7404C005D1ULL, "6202108_m.fxe"},
	{0x394608D821A05144ULL, "vxw_encounter2_camp_53_0_0.vlm"},
	{0x39460D0327B01F0DULL, "hlac_ht_mansion2b_818_62y.dds"},
	{0x394638E2BECF1B02ULL, "w_woundedcoast_168_0.msh"},
	{0x3946398F616843F1ULL, "618104591.ogg"},
	{0x39463AE2BECF1EB8ULL, "w_woundedcoast_168_0.mmh"},
	{0x39463B3CD412B127ULL, "41960593.ogg"},
	{0x394646E2BECF335BULL, "w_woundedcoast_168_0.mao"},
	{0x39464E692666D1E0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain_2.mao"},
	{0x394657CA0815D021ULL, "prop1_ds2.psh.fxo"},
	{0x3947263CB9F518EEULL, "vxht_keep_n_6j8_0_0.vlm"},
	{0x39472A95CD1083D5ULL, "ser200_aveline6201584.cl"},
	{0x394780F9811FD926ULL, "primitives_primitive_p4.vsh.fxo.dx11"},
	{0x3948244F2B98EA08ULL, "6147929_m.fxe"},
	{0x394849DD63C6D32CULL, "hlac_ht_0_hightown_a2_-131073_1x.dds"},
	{0x39490A0AC3A8AEA9ULL, "vxlt_0_lowtown_docks_bcr_0_0.vlm"},
	{0x3949764F4E0C5EC2ULL, "vxlt_hangedman_n_1tc_0_0.vlm"},
	{0x3949A0FAD1FF6F1AULL, "6031330_m.fxe"},
	{0x3949B81617BF5A36ULL, "vxw_bonepit_1uj_0_0.vlm"},
	{0x394AA2B770A2B711ULL, "641279350.ogg"},
	{0x394AE5BE104FE9D5ULL, "dre000_varric6208055.cl"},
	{0x394BB792ADBDF7E4ULL, "146106213.ogg"},
	{0x394BC6CE9D5EFBB8ULL, "vxlt_0_lowtown_a2_n_gng_0_0.vlm"},
	{0x394C223772D85B05ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammer.phy"},
	{0x394C50A66806E4A4ULL, "vxlt_0_lowtown_docks_n_bfs_0_0.vlm"},
	{0x394C8C3A8ACCF749ULL, "6089950_m.fxe"},
	{0x394D20A68B748FFAULL, "vxlt_hangedman_1rf_0_0.vlm"},
	{0x394D371F79B6F671ULL, "754573220.ogg"},
	{0x394D574F9E1EEA1AULL, "hlac_ca_cavern1_haunted_24424_25y.dds"},
	{0x394D73A3133D6EA1ULL, "vxht_0_hightown_a2_cah_0_0.vlm"},
	{0x394DCCB8FA9C33F1ULL, "6135033_m.fxe"},
	{0x394DD0E05257EF2EULL, "llac_ca_caverns1_mines_24427_5.dds"},
	{0x394E82AF05D33D2CULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\dazed.ani"},
	{0x394F81793252DFC5ULL, "vxht_mansion_n_5dn_0_0.vlm"},
	{0x394FB567FB963330ULL, "142624335.ogg"},
	{0x395041E2BED77F43ULL, "w_woundedcoast_168_0.phy"},
	{0x39513060F4EC9567ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_cartrock_1.mao"},
	{0x39514630A122B05AULL, "6180587_f.fxe"},
	{0x39526D1BCEDCB1BEULL, "vxlt_0_lowtown_docks_a3_n_d2p_0_0.vlm"},
	{0x3952910C232E9E47ULL, "vxht_0_hightown_n_ds_0_0.vlm"},
	{0x3952EC61E7C99123ULL, "w_blightlands_1_0.rml"},
	{0x39530F1A889A4556ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magc.phy"},
	{0x395367449A8A0AB1ULL, "hlac_w_bonepitmining_a2_0_8y.dds"},
	{0x39536788DB26E356ULL, "vxlt_0_lowtown_n_el9_0_0.vlm"},
	{0x3953C0746C52D710ULL, "hlac_lt_gallowstemplar_f_2394_4y.dds"},
	{0x39546DE985A430F1ULL, "123852598.ogg"},
	{0x39550D1218013685ULL, "llac_lt_0_lowtown_a3_n_c_65535_8.dds"},
	{0x395510F25635644FULL, "vxht_brothel_n_6ig_0_0.vlm"},
	{0x395524F36289C073ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.up.ani"},
	{0x39552F51FEF0ADEAULL, "vxlt_0_lowtown_a2_bxm_0_0.vlm"},
	{0x395556E980D595D3ULL, "6201222_m.fxe"},
	{0x39558FAF05D9726EULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\dazed.gad"},
	{0x3955B7C0D701E051ULL, "plt_isa300pt_wicked.plo"},
	{0x3955D36AC374CB1FULL, "llac_lt_gallowstemplar_n_2394_11.dds"},
	{0x395612F53653B5B2ULL, "753186844.ogg"},
	{0x39564573234773B2ULL, "vxht_keep_4hc_0_0.vlm"},
	{0x3956588FA3787857ULL, "llac_lt_0_lowtown_65536_12.dds"},
	{0x39568EB0641F31EAULL, "gen_im_trs_crc_lgt_cht_r0.uti"},
	{0x3956BD2951ED2F68ULL, "gen00fl_bethany6101215.cl"},
	{0x3956F48707F626A5ULL, "154015944.ogg"},
	{0x395706556B0B0DE6ULL, "6134323_m.fxe"},
	{0x39573FA6973D87A5ULL, "6064058_m.fxe"},
	{0x39573FD90D8A4864ULL, "268853642.ogg"},
	{0x3957536C93DDADA4ULL, "dae000_bartrand6204164.cl"},
	{0x3957B1EF1665F385ULL, "vxlt_undercityserialkiller_1c9_0_0.vlm"},
	{0x3957E82C40027742ULL, "vxlt_hangedman_n_1ql_0_0.vlm"},
	{0x39581067BCFBB7A0ULL, "dae200_cullen6140292.cl"},
	{0x3958BEF53F7C8EF4ULL, "hlac_lt_gallowstemplar_2383_11z.dds"},
	{0x395905DBE10DA1AEULL, "32902370.ogg"},
	{0x395938C8FAAB6596ULL, "vxlt_0_lowtown_docks_n_br5_0_0.vlm"},
	{0x3959D7040F6E9A5AULL, "ht_0_hightown_a2_20_0.anb"},
	{0x395A0D0B1822DF63ULL, "vxca_caverns2_8p1_0_0.vlm"},
	{0x395A1C1C534378DBULL, "qcr220_orsino6041849.cl"},
	{0x395A4A761A15AE0EULL, "llac_lt_0_lowtown_docks_a3_-131075_7.dds"},
	{0x395A7EE7EB240157ULL, "6078750_m.fxe"},
	{0x395AC28359FCFD24ULL, "vxlt_0_lowtown_docks_n_bcp_0_0.vlm"},
	{0x395B92837D6AA87AULL, "vxlt_hangedman_1yg_0_0.vlm"},
	{0x395C290D844A7088ULL, "isa211_walleyedsam6070007.cl"},
	{0x395C3422D24A2809ULL, "gen00fl_anders6202355.cl"},
	{0x395C50E774CD0909ULL, "145998010.ogg"},
	{0x395C5ED57B88B390ULL, "hlac_lt_0_lowtown_docks_a3_131073_0z.dds"},
	{0x395C76166FDFE6F3ULL, "vxca_caverns2prc_a70_0_0.vlm"},
	{0x395C8AAF05DF884DULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\dazed.evt"},
	{0x395D0E68E282CFAEULL, "follower_banter6173165.cl"},
	{0x395D31B81A5CECBEULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.impa.evt"},
	{0x395DDA3878CA4CD2ULL, "dae000_arishok6068569.cl"},
	{0x395DF3562448F845ULL, "vxht_mansion_n_5am_0_0.vlm"},
	{0x395DF81A88A425F7ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magc.msh"},
	{0x395E0A1A88A4445AULL, "art\\characters\\playercharacter\\humanfemale\\rob_magc.mao"},
	{0x395E0E1A88A44B69ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magc.mmh"},
	{0x395E53EBCAF663FDULL, "131623180.ogg"},
	{0x395EA4077F20EB1AULL, "vxw_bonepitmining_a2_1ws_0_1.vlm"},
	{0x395EB6CBDEB11224ULL, "mcr310_party_banter.cnv"},
	{0x395EE62C4AC1B11CULL, "162580921.ogg"},
	{0x395F6074197780BFULL, "lgt210_magu_selby.bnk"},
	{0x395FCEB21EF617C7ULL, "vxca_caverns1_mines_j6k_0_0.vlm"},
	{0x395FD562933E5169ULL, "vxlt_hangedman_n_1yc_0_0.vlm"},
	{0x39604C82CF3096D5ULL, "163698172.ogg"},
	{0x39604ED79C2F0D65ULL, "455701496.ogg"},
	{0x39606F929D988112ULL, "hightown_fem_courtesan_mob6168590.cl"},
	{0x39607FF24AF2987BULL, "6048185_f.fxe"},
	{0x39608F31B3D3DB7AULL, "plt_zz_andpt_debug.nss"},
	{0x3960B40FBC162980ULL, "mcr310_loyalist_mages6076854.cl"},
	{0x3961872DF28EDD79ULL, "116753215.ogg"},
	{0x39618EC14AED71C3ULL, "vxlt_warehouse_rn_0_0.vlm"},
	{0x3961D8CBDEB38184ULL, "mcr310_party_banter.bnk"},
	{0x3961F1866C9F147AULL, "vxlt_gallowsprison_n_5sc_0_0.vlm"},
	{0x396210FAD3CD72FAULL, "hlac_ht_chantry_2534_35y.dds"},
	{0x396216482C05362DULL, "dae000_hate_follower_react6187869.cl"},
	{0x39624274197983C7ULL, "lgt210_magu_selby.cnv"},
	{0x396285CFFD680223ULL, "qun120cr_javaris6172020.cl"},
	{0x396366E673A0E424ULL, "38768570.ogg"},
	{0x39636966E9515981ULL, "follower_banter6170749.cl"},
	{0x396376448097B1E0ULL, "llac_ca_cavern1_haunted_24426_0.dds"},
	{0x39638BB255467662ULL, "vxht_keep_n_6iy_0_0.vlm"},
	{0x39639420F2B4BD22ULL, "366010790.ogg"},
	{0x3963D1F1FB2F49F3ULL, "gen_im_trs_gem_coral_02.uti"},
	{0x39641C01CA259CF6ULL, "889085493.ogg"},
	{0x39643743EC623FF8ULL, "6207674_m.fxe"},
	{0x396448B81A6332E9ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.impa.gad"},
	{0x39647775750031F1ULL, "6162296_player_carver.cut"},
	{0x396487387022BA39ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingb.msh"},
	{0x396491387022CB27ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingb.mmh"},
	{0x396495387022D274ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingb.mao"},
	{0x3965077B1D56E1D4ULL, "plt_vrc000pt_n_gift.plo"},
	{0x396568C328D0B0C8ULL, "ser270_emeric6149366.cl"},
	{0x396596123210ABE3ULL, "6063839_m.fxe"},
	{0x3965BFD97B606464ULL, "vxlt_shop_e7_0_0.vlm"},
	{0x39660CA839747022ULL, "6060105_m.fxe"},
	{0x39665B9555B6038CULL, "vxlt_gallowsprison_n_6k2_0_0.vlm"},
	{0x3966ED01E0B36B66ULL, "186316744.ogg"},
	{0x39670FE3F3B7AED8ULL, "vxlt_hangedman_2yn_0_0.vlm"},
	{0x39677BBFD9A72492ULL, "lowtown_male_alienage2_mob6166513.cl"},
	{0x39678E5252480CDAULL, "vxlt_0_lowtown_a2_n_c_cmv_0_0.vlm"},
	{0x3967904D23B7562DULL, "352734027.ogg"},
	{0x396799C54E47EF35ULL, "79105982.ogg"},
	{0x396806A43F0FEF71ULL, "hlac_ca_caverns2prc_5802_25z.dds"},
	{0x39684055F914F429ULL, "60563719.ogg"},
	{0x396852FAFEF2CDBEULL, "719202384.ogg"},
	{0x3968FBF3629A231DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.up.gad"},
	{0x3969092F45C19203ULL, "6033698_m.fxe"},
	{0x396923077370FFE9ULL, "6108789_m.fxe"},
	{0x396989DDB4235868ULL, "vxlt_0_lowtown_a2_n_c_h2f_0_0.vlm"},
	{0x3969FF880220E208ULL, "one100_solivitus6048145.cl"},
	{0x396A39FCD77F4A78ULL, "gen00fl_sebastian6174717.cl"},
	{0x396B19B81A690137ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.impa.ani"},
	{0x396B8807ABE8322DULL, "plt_qun200pt_blackform2.plo"},
	{0x396BA291C8820D1FULL, "dae000_gamlen6121570.cl"},
	{0x396BB6A852A26B44ULL, "vxca_caverns1_j21_0_0.vlm"},
	{0x396BC27444B03609ULL, "999535773.ogg"},
	{0x396BDC69B3B1EC1CULL, "1055454617.ogg"},
	{0x396C8EAC28DE328EULL, "vxht_chantry_n_2cc_0_0.vlm"},
	{0x396CB5BF3F55C1A9ULL, "vxht_keep_n_4nm_0_0.vlm"},
	{0x396D7ACAEFDC2AEDULL, "vxlt_warehouse_8n_0_0.vlm"},
	{0x396D8B4EF1FB72ECULL, "art\\vfx\\combat\\vfx_c_mer_attk3_c.impa.ani"},
	{0x396DB870937E99B3ULL, "vxdr_slavers_1bl_0_0.vlm"},
	{0x396DBD1B4C896AFDULL, "594330287.ogg"},
	{0x396DF9BB7E0E1B79ULL, "6198699_f.fxe"},
	{0x396E473F853469E9ULL, "vxlt_hangedman_n_1vb_0_0.vlm"},
	{0x396EB75A267CCAD2ULL, "vxw_bonepit_112_0_0.vlm"},
	{0x3970009E3CE38A43ULL, "vxlt_warehouse_wq_0_0.vlm"},
	{0x39702A027FF20DC7ULL, "6116201_m.fxe"},
	{0x39702E1880F1F7E6ULL, "6065652_m.fxe"},
	{0x397091E2B485B4C5ULL, "vxw_bonepitdestroyed_a3_256_0_0.vlm"},
	{0x3971477DFA023535ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala1a.msh"},
	{0x3971497DFA0238ABULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala1a.mmh"},
	{0x3971557DFA024D48ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala1a.mao"},
	{0x3971AD29A08128D3ULL, "vxca_caverns2_5c8_0_1.vlm"},
	{0x3971C6EEF99066C5ULL, "64971895.ogg"},
	{0x397219C83385E026ULL, "gen00fl_isabela6106173.cl"},
	{0x3972D5CF199F1FFBULL, "711506054.ogg"},
	{0x39733D6E3E140E5BULL, "vxlt_0_lowtown_docks_n_c58_0_0.vlm"},
	{0x39734C6650DFCC76ULL, "gen_im_trs_tmp_hvy_boo_r0.uti"},
	{0x3973984EF1FFF42EULL, "art\\vfx\\combat\\vfx_c_mer_attk3_c.impa.gad"},
	{0x3973B3EA239FD30DULL, "6119008_m.fxe"},
	{0x39741478A9F89FD8ULL, "vxca_caverns1_mines_ede_0_0.vlm"},
	{0x3974291F8776777BULL, "qcr220_orsino6060707.cl"},
	{0x39745F1B0D282A82ULL, "456322600.ogg"},
	{0x3974D46367E309DCULL, "pro000_meeran6050120.cl"},
	{0x3974D574633F52D3ULL, "793346788.ogg"},
	{0x39750656E848800DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexa.mmh"},
	{0x39750A56E848875EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexa.mao"},
	{0x39751856E8489E93ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexa.msh"},
	{0x39751E7267C0D51DULL, "vxlt_0_lowtown_a2_n_c_byl_0_0.vlm"},
	{0x397547C27410716CULL, "6151012_m.fxe"},
	{0x39756E04263543F0ULL, "vrc221_varric6044561.cl"},
	{0x39763EA951379455ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_06a_0t.dds"},
	{0x3976978C12D30D90ULL, "140292088.ogg"},
	{0x3976E76F4FB343E3ULL, "vxht_0_hightown_bu7_0_0.vlm"},
	{0x3976EE6AE6F557D8ULL, "vxdr_primevalentrance_237_0_0.vlm"},
	{0x3976EF37BE187445ULL, "25147280.ogg"},
	{0x3977C757215F959DULL, "6118182_m.fxe"},
	{0x3977FC68260CC06EULL, "hlac_lt_0_lowtown_a2_n_-65536_6x.dds"},
	{0x39783E456CFB3A97ULL, "6118562_m.fxe"},
	{0x3978684B70A13B9DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic_3.mao"},
	{0x3978DFB6AF330C94ULL, "6144520_m.fxe"},
	{0x3978F2EF697381A3ULL, "3218106.ogg"},
	{0x397922FA8B1ADF2BULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_0t.dds"},
	{0x3979AF645B5C05F4ULL, "follower_banter6155808.cl"},
	{0x397A407DA13D7527ULL, "vxht_0_hightown_a3_bpu_0_0.vlm"},
	{0x397A92CFE5410FA4ULL, "vxht_keep_7a5_0_0.vlm"},
	{0x397A934EF2060A0DULL, "art\\vfx\\combat\\vfx_c_mer_attk3_c.impa.evt"},
	{0x397AD88BE50C33F5ULL, "ht_0_hightownexplosion_n_9_0.phy"},
	{0x397AEDF6524FF48DULL, "dae011_badpoet_sta6147004.cl"},
	{0x397B97823871F7BBULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa.phy"},
	{0x397C5D5A11B1ACACULL, "vxht_0_hightown_n_8ay_0_0.vlm"},
	{0x397C7317B8EED477ULL, "vxht_keep_n_6b5_0_0.vlm"},
	{0x397CBED2E02D46B3ULL, "llac_lt_gallowstemplar_2394_3.dds"},
	{0x397D635487A9D4EBULL, "847406077.ogg"},
	{0x397E2A8C4EEFE3A0ULL, "hlac_lt_0_lowtown_docks_n_65535_1z.dds"},
	{0x397E63A5FD623EF6ULL, "vxht_0_hightown_aso_0_0.vlm"},
	{0x397E727B2ED9A2C3ULL, "vxlt_warehouse_xp_0_0.vlm"},
	{0x397EB08C2BEC8F55ULL, "vxca_caverns1_fn4_0_0.vlm"},
	{0x397EF13C312ABDDDULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_0z.dds"},
	{0x397F03BFA67BCD45ULL, "vxw_bonepitdestroyed_a3_205_0_0.vlm"},
	{0x397F396EC513D97BULL, "vxlt_0_lowtown_a2_gwl_0_0.vlm"},
	{0x397F62DFB4BB28C4ULL, "lt_0_lowtown_docks_a3_232_0.mmh"},
	{0x397F66DFB4BB2F17ULL, "lt_0_lowtown_docks_a3_232_0.mao"},
	{0x397F78DFB4BB4E66ULL, "lt_0_lowtown_docks_a3_232_0.msh"},
	{0x397F9C0CD6B72E2BULL, "906594240.ogg"},
	{0x397FA21344544583ULL, "vxlt_0_lowtown_ep2_0_1.vlm"},
	{0x398004086956D41FULL, "1020717621.ogg"},
	{0x398074B7F0DDD0B1ULL, "gen00fl_aveline6129048.cl"},
	{0x3980E38A61DA7692ULL, "6120515_m.fxe"},
	{0x3981C3F4D7E9D5E9ULL, "vxht_mansion_n_3ys_0_0.vlm"},
	{0x398228E1B403164FULL, "gen00fl_fenris6110829.cl"},
	{0x39827CCD7DE6ABECULL, "hlac_ca_caverns1_mines_24421_1z.dds"},
	{0x39829FA6297EF4A9ULL, "vxca_cavern1_haunted_i3a_0_0.vlm"},
	{0x39837F89DF97BC36ULL, "vxw_bonepit_1in_0_0.vlm"},
	{0x3983E82572693352ULL, "hla_ht_0_hightown_65537_0x.dds"},
	{0x3984484A881FF5D3ULL, "6117673_m.fxe"},
	{0x3985AB6C5CD3305FULL, "278135911.ogg"},
	{0x3985DA61E7F4CC3AULL, "w_blightlands_1_0.msh"},
	{0x3985E861E7F4E473ULL, "w_blightlands_1_0.mao"},
	{0x3985EC61E7F4EA80ULL, "w_blightlands_1_0.mmh"},
	{0x398605E183359D1FULL, "fpstest.arl"},
	{0x3986181A2FDF46A4ULL, "vxlt_0_lowtown_docks_n_brw_0_0.vlm"},
	{0x39868A2781DF37C5ULL, "vxlt_hangedman_1yl_0_0.vlm"},
	{0x3986B0225EF15317ULL, "6117462_m.fxe"},
	{0x3986D83847B55ABDULL, "788210979.ogg"},
	{0x39879D1E269FE382ULL, "hm_plinth.mor"},
	{0x3987E3E8A4DE4A55ULL, "lgt150_nat_were_supporter6110971.cl"},
	{0x3987E8B23C2C0C80ULL, "vxw_blightlands_65_0_0.vlm"},
	{0x3988AB030C3B6DE0ULL, "hlac_lt_backalley_g_65537_8z.dds"},
	{0x3988AFA99DFEF652ULL, "100324084.ogg"},
	{0x3988B25A93338DA7ULL, "vxht_0_hightown_a3_bot_0_0.vlm"},
	{0x3988BE78456C033CULL, "360001903.ogg"},
	{0x3988CA612E7D6242ULL, "1026109955.ogg"},
	{0x39890CBDB57C54BDULL, "688752113.ogg"},
	{0x398933BFD1810422ULL, "gen00fl_anders6125419.cl"},
	{0x3989396F5CF2F05CULL, "vxw_encounter2_om_0_0.vlm"},
	{0x3989CB9B8CD4D278ULL, "llac_ht_0_hightown_a2_65533_1.dds"},
	{0x3989CF1FA9FE72F1ULL, "vxht_0_hightown_a2_c47_0_0.vlm"},
	{0x398A529A96B069F6ULL, "6076102_m.fxe"},
	{0x398AD5F28745391CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_tochest.ani"},
	{0x398B025CF7739A0EULL, "dae000_tomwise6167292.cl"},
	{0x398B39E92A7EBD08ULL, "6101408_m.fxe"},
	{0x398BA36BAB85A161ULL, "vxlt_hangedman_1t7_0_0.vlm"},
	{0x398BC75AFB5D3608ULL, "832906002.ogg"},
	{0x398BF57631A5C42AULL, "vxht_mansion2b_23d_0_0.vlm"},
	{0x398C5908DDFF0B93ULL, "follower_banter6141563.cl"},
	{0x398CD6663BA2FAE4ULL, "9681831.ogg"},
	{0x398CDA040F99F886ULL, "ht_0_hightown_a2_20_0.rml"},
	{0x398D240139347B8BULL, "pro000_merchant6051551.cl"},
	{0x398D3E11E647335AULL, "character_character_p258.psh.fxo.dx11.assembly"},
	{0x398D547031C186D5ULL, "follower_barks6153477.cl"},
	{0x398E827A4558F44FULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p_injured.gad"},
	{0x398EABAF491CA32FULL, "rdr111dg_martin6015069.cl"},
	{0x39907A387047DF5CULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingb.phy"},
	{0x39914232085E8DC6ULL, "hlac_ca_cavern1_haunted_24423_2y.dds"},
	{0x39915217CDFF8806ULL, "vxca_caverns2_720_0_1.vlm"},
	{0x3991E2F2874B6E5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_tochest.gad"},
	{0x39929BE3012BCBBCULL, "112952237.ogg"},
	{0x399391D7210DEAB7ULL, "533205777.ogg"},
	{0x3993E61FC35C7918ULL, "vxca_caverns1_mines_idi_0_0.vlm"},
	{0x39942FAECF1C3316ULL, "6041118_f.fxe"},
	{0x399436F9656D2BDEULL, "290573051.ogg"},
	{0x39949415A1C7E200ULL, "6018536_f.fxe"},
	{0x3994C18C04628461ULL, "ts_qunari_03.gda"},
	{0x39951DF13E5A945FULL, "6196699_f.fxe"},
	{0x3995977A455F36FCULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p_injured.evt"},
	{0x3995E97B4C013D4DULL, "art\\vfx\\materials\\vfx_fire_sin_blend.mao"},
	{0x39961B9831410591ULL, "vxca_cavern1_haunted_e6n_0_0.vlm"},
	{0x39966E3AD962CEA2ULL, "w_sundermount_186_0.msh"},
	{0x3996703AD962D258ULL, "w_sundermount_186_0.mmh"},
	{0x39967C3AD962E6FBULL, "w_sundermount_186_0.mao"},
	{0x3996C8E9D7C238BEULL, "dre000_dougal6129709.cl"},
	{0x39978102B9097091ULL, "1044740724.ogg"},
	{0x399799CB0C073D19ULL, "mag320cr_noblewoman_1.utc"},
	{0x3997AB4C4EE908DCULL, "vxw_encounter2_jb_0_0.vlm"},
	{0x39980017054BC9A1ULL, "457790728.ogg"},
	{0x399804AD20BA046CULL, "zz_dre_debug6056979.cl"},
	{0x399840FC9BF48B71ULL, "vxht_0_hightown_a2_c36_0_0.vlm"},
	{0x39985F1797DD0A82ULL, "57696148.ogg"},
	{0x39987B4A728C7B14ULL, "hlac_ht_bartrand_n_3382_1x.dds"},
	{0x3998BDF287514E7DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_tochest.evt"},
	{0x3998E7E0B74E20C6ULL, "occlusionquery.vsh.fxo"},
	{0x39996D943A836DDAULL, "475241205.ogg"},
	{0x3999789ABE4B9232ULL, "410238259.ogg"},
	{0x3999B45A49BC8522ULL, "explo_smoke_blend.dds"},
	{0x399A19702E98E848ULL, "follower_banter6155556.cl"},
	{0x399A3C801CE1579DULL, "148643300.ogg"},
	{0x399ABCEC2F914639ULL, "6126657_m.fxe"},
	{0x399AD8F51F97194BULL, "vxht_0_hightown_cbw_0_0.vlm"},
	{0x399B011F8EC8A99AULL, "vxht_0_hightown_a2_b5f_0_0.vlm"},
	{0x399B50B58DAA3C76ULL, "vxca_caverns2_9dm_0_0.vlm"},
	{0x399B5C7E350EA09FULL, "dae000_player_generic_find6179044.cl"},
	{0x399BCB03F728DE40ULL, "ss_taraneh.bnk"},
	{0x399C6EA6DB124B1EULL, "hlac_w_woundedcoast_-262141_0x.dds"},
	{0x399D0361432C19E6ULL, "mag121_idunna6084483.cl"},
	{0x399D25946D4CD87EULL, "vxlt_0_lowtown_docks_n_c46_0_0.vlm"},
	{0x399D6C3A47D2133EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keep_1.mao"},
	{0x399DCF2A80FEDE8EULL, "vxca_cave_182_0_0.vlm"},
	{0x399E0571DB37FADDULL, "dae000_merchant_dm_mob6174243.cl"},
	{0x399E4BB2B1EBCCEBULL, "105332968.ogg"},
	{0x399E86E221909630ULL, "474197301.ogg"},
	{0x399EA824D5B7C9D3ULL, "vxlt_0_lowtown_docks_a3_coz_0_0.vlm"},
	{0x399EBC3A0D21BACCULL, "6160665_m.fxe"},
	{0x399F0F3B74ABAB4BULL, "follower_banter6136680.cl"},
	{0x399FA1B7921F903BULL, "vxlt_gallowstemplar_39i_0_0.vlm"},
	{0x39A00F4EE4BC46CFULL, "vxw_encounter2_6y_0_0.vlm"},
	{0x39A0875E92884DECULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mugpewter01_0_n.dds"},
	{0x39A0D757BB9010D8ULL, "vxlt_hangedman_2ej_0_0.vlm"},
	{0x39A139601F5E6EE4ULL, "182466788.ogg"},
	{0x39A159D8FE02C6E4ULL, "167106768.ogg"},
	{0x39A1773AD96CE5E3ULL, "w_sundermount_186_0.phy"},
	{0x39A17F158921A315ULL, "vxht_brothel_n_3vq_0_0.vlm"},
	{0x39A18AA34431C348ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhangceltic.msh"},
	{0x39A19CA34431E1D9ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhangceltic.mao"},
	{0x39A19ED09D15A89CULL, "352781130.ogg"},
	{0x39A1A8A34431F672ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhangceltic.mmh"},
	{0x39A232FD74261E0AULL, "vxdr_primevalentrance_396_0_0.vlm"},
	{0x39A29F7A4569CE15ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p_injured.ani"},
	{0x39A2DD8034F716A1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_basic.phy"},
	{0x39A304F3087D1C9EULL, "6053095_m.fxe"},
	{0x39A31CEC3668288CULL, "528892255.ogg"},
	{0x39A36F37FEC2A12FULL, "vxlt_backalley_n_1lk_0_0.vlm"},
	{0x39A39E3276A7B488ULL, "mag360cr_templar_archer.utc"},
	{0x39A45C5D79E21329ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent_overlay.ani"},
	{0x39A468C9870B2504ULL, "killtarget.ncs"},
	{0x39A547B0C77428B8ULL, "6152206_m.fxe"},
	{0x39A59614771FBEA7ULL, "vxht_0_hightown_a3_bav_0_0.vlm"},
	{0x39A5A3BC353ED1D2ULL, "588484623.ogg"},
	{0x39A5F7CD5FFF62F0ULL, "mcr371_arthuris6087098.cl"},
	{0x39A6234E07997B26ULL, "6208219_m.fxe"},
	{0x39A668B5EBD91A0EULL, "6018664_f.fxe"},
	{0x39A77719D26FA344ULL, "vxlt_hangedman_2mn_0_0.vlm"},
	{0x39A77E823897085AULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa.msh"},
	{0x39A78C8238971F93ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa.mao"},
	{0x39A79082389726A0ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa.mmh"},
	{0x39A79E8E683B4119ULL, "558414464.ogg"},
	{0x39A9559D6D608159ULL, "vxw_bonepitmining_a2_fe_0_0.vlm"},
	{0x39A96183AD7369C1ULL, "6141140_m.fxe"},
	{0x39A98A9AAC86AC01ULL, "6141904_m.fxe"},
	{0x39A9B8DE625B5AD5ULL, "art\\characters\\weapons\\unique\\w_vael_quiver.msh"},
	{0x39A9BADE625B5E4BULL, "art\\characters\\weapons\\unique\\w_vael_quiver.mmh"},
	{0x39A9C7AF1094C709ULL, "hlac_lt_0_lowtown_docks_a3_n_131068_0x.dds"},
	{0x39A9FBC814590706ULL, "hm_bancroft.mor"},
	{0x39AA2C3FC8E4C219ULL, "vxw_sundermount_311_0_1.vlm"},
	{0x39AA2EDF013DF809ULL, "vxht_keep_3ih_0_0.vlm"},
	{0x39AAE2A1917B57C9ULL, "vxht_playershouse_n_1v1_0_0.vlm"},
	{0x39AB195FA8EF2301ULL, "vxht_0_hightown_a2_auz_0_0.vlm"},
	{0x39AB775D79E85FABULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent_overlay.gad"},
	{0x39AB804B1B287E62ULL, "6128924_m.fxe"},
	{0x39AB8ED79140A3F5ULL, "459594492.ogg"},
	{0x39AC04FC692B6997ULL, "follower_banter6173377.cl"},
	{0x39AC8BF394835D29ULL, "vxw_bonepitdestroyed_a3_1nh_0_0.vlm"},
	{0x39ACB4074E358054ULL, "vxca_caverns2_7o9_0_1.vlm"},
	{0x39ACBDCA79C3F9F3ULL, "ser200_aveline6201483.cl"},
	{0x39ACC152406FEC71ULL, "6095762_m.fxe"},
	{0x39ACC9196D3DBA15ULL, "6033722_m.fxe"},
	{0x39AD6ED0987340DCULL, "172697707.ogg"},
	{0x39AD8E4C9A05FB01ULL, "hlac_ht_0_hightown_a2_-196608_2y.dds"},
	{0x39ADB3F56AA36D6FULL, "403433873.ogg"},
	{0x39ADB6E44E5A8A23ULL, "llac_w_bonepitdestroyed_a3_458758_0.dds"},
	{0x39ADC2C0D74C68B6ULL, "plt_isa300pt_wicked.nss"},
	{0x39ADC44CFB640630ULL, "vxht_brothel_n_6kn_0_0.vlm"},
	{0x39AE06C7F3615996ULL, "art\\vfx\\combat\\vfx_v_fir_single_att_c.impa.evt"},
	{0x39AE40FCD560B9F6ULL, "w_sundermount_81_0_lc.mmh"},
	{0x39AE7B53E9207C42ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv3.gad"},
	{0x39AE812BD6B25F4FULL, "vxw_encounter2_1x_0_0.vlm"},
	{0x39AEFCD2E2D276D2ULL, "gen00fl_sebastian6177923.cl"},
	{0x39AF294849FFFF3EULL, "qcr220_orsino6061358.cl"},
	{0x39AF4934AD862958ULL, "vxlt_hangedman_2bk_0_0.vlm"},
	{0x39B03B056489B55AULL, "gen00fl_sebastian6174706.cl"},
	{0x39B0726C2F21D4B0ULL, "145561796.ogg"},
	{0x39B08E71EB2AEFAEULL, "stoney_s.dds"},
	{0x39B0A0402F98590CULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gsw_002a_d.dds"},
	{0x39B0D0ED36080E47ULL, "vxlt_warehouse_cq_0_0.vlm"},
	{0x39B0ECE81D9D4C0FULL, "hlac_ca_cavern1_haunted_24422_1z.dds"},
	{0x39B12A8DA6965A92ULL, "hlac_lt_gallowstemplar_n_2383_5x.dds"},
	{0x39B14252B9A32337ULL, "ht_0_hightownexplosion_n_50_0.rml"},
	{0x39B1745D79ECC610ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent_overlay.evt"},
	{0x39B182A1BC542157ULL, "6207956_m.fxe"},
	{0x39B1E114F0B8B9AFULL, "vxlt_backalley_n_1wj_0_0.vlm"},
	{0x39B2391F9C38E301ULL, "hlac_lt_0_lowtown_docks_-196610_5z.dds"},
	{0x39B2424A73A7D38AULL, "dre000_bodahn6031352.cl"},
	{0x39B3A25987210F20ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cenotaph.phy"},
	{0x39B3B8C9E8A356ACULL, "w_woundedcoast_211_0.phy"},
	{0x39B3E16489D03C3EULL, "6115435_f.fxe"},
	{0x39B432BBAF5432B4ULL, "603082065.ogg"},
	{0x39B469BC2AF23207ULL, "llac_lt_0_lowtown_a2_n_c_131073_2.dds"},
	{0x39B49B0CB785B52FULL, "vxw_blightlands_112_0_0.vlm"},
	{0x39B4A2DD52BFE878ULL, "715991394.ogg"},
	{0x39B4FDC7F3676921ULL, "art\\vfx\\combat\\vfx_v_fir_single_att_c.impa.gad"},
	{0x39B55F057D5D489CULL, "6031321_m.fxe"},
	{0x39B5856E2A190682ULL, "hightown_male_cityguard_mob6170059.cl"},
	{0x39B59E53E926D639ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv3.evt"},
	{0x39B600223121735DULL, "vxlt_0_lowtown_a3_n_c_ekj_0_2.vlm"},
	{0x39B615C501EE48D9ULL, "220630637.ogg"},
	{0x39B6A67347662848ULL, "dae021_chantry_initiate_mob6147620.cl"},
	{0x39B74FBD670F760BULL, "art\\vfx\\combat\\vfx_w_2nd_atk_lng_c.impa.ani"},
	{0x39B7911E6ABAEBBCULL, "6149318_m.fxe"},
	{0x39B7C77A5F5699D9ULL, "vxw_bonepitmining_a2_md_0_0.vlm"},
	{0x39B7E4C634CA4F98ULL, "239264895.ogg"},
	{0x39B7F8625B43525FULL, "6104549_m.fxe"},
	{0x39B82E57615439E1ULL, "vxw_encounter3_z_0_0.vlm"},
	{0x39B87DC3B6F4CB13ULL, "llac_lt_gallowstemplar_f_2383_6.dds"},
	{0x39B88961ACBDF317ULL, "lt_0_lowtown_docks_a3_n_26_0.anb"},
	{0x39B8D9583411FD7AULL, "vxw_encounter2_camp_1i_0_0.vlm"},
	{0x39B8E08F578B1D24ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave2statue.mmh"},
	{0x39B8E48F578B2477ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave2statue.mao"},
	{0x39B8F68F578B42C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave2statue.msh"},
	{0x39B914E6506E1399ULL, "art\\characters\\creatures\\dragon\\animation\\dblow_blunt.ani"},
	{0x39B92C609D7E27E7ULL, "follower_banter6163219.cl"},
	{0x39B98B3C9AE53B81ULL, "vxht_0_hightown_a2_apy_0_0.vlm"},
	{0x39B9C37C2C5726ECULL, "hlac_ca_caverns1_24426_10z.dds"},
	{0x39BA155BF6630A01ULL, "vxw_encounter2_j3_0_0.vlm"},
	{0x39BA9E6215F11E87ULL, "lowtown_male_lowlife2_mob6166772.cl"},
	{0x39BAA8DBC33B480BULL, "vxlt_0_lowtown_a3_n_c_ht6_0_0.vlm"},
	{0x39BABC878B1AC95AULL, "6033664_m.fxe"},
	{0x39BB28FC6AAF978FULL, "floorcolor.dds"},
	{0x39BB7459417F5D09ULL, "vxdr_primevaltomb_50t_0_0.vlm"},
	{0x39BBBDD9B583C9ADULL, "dae300_alistair_king6134903.cl"},
	{0x39BBFEC7F36D89CFULL, "art\\vfx\\combat\\vfx_v_fir_single_att_c.impa.ani"},
	{0x39BC0D5F1D389282ULL, "one130_thug6131598.cl"},
	{0x39BC3629ED5A1EB0ULL, "vxht_brothel_n_6to_0_0.vlm"},
	{0x39BC6992C428D551ULL, "achievementbrowser.gfx"},
	{0x39BC7EDAE5A9809AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_mera_0s.dds"},
	{0x39BC8571760BC13BULL, "vxlt_gallowstemplar_33k_0_0.vlm"},
	{0x39BCD6EBC544181AULL, "pro000_soldier_leader.cnv"},
	{0x39BD5579DCEC2716ULL, "gen00fl_fenris6128003.cl"},
	{0x39BD589D81C53E01ULL, "6163427_m.fxe"},
	{0x39BE84B37428C50CULL, "6111412_m.fxe"},
	{0x39BEF6D35E5D0F90ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_flat.mao"},
	{0x39BEFAD35E5D16A3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_flat.mmh"},
	{0x39BF08D35E5D2EFDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_flat.msh"},
	{0x39BF202CB3ED831DULL, "llac_ca_caverns2prc_5802_16.dds"},
	{0x39BF4A9A8E699CBFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_meda_0d.dds"},
	{0x39BFC96FE1A27FD9ULL, "dae000_dalish_apprentices6147574.cl"},
	{0x39BFE11E924318A8ULL, "vxlt_darktown_47g_0_0.vlm"},
	{0x39BFEFE65073F3DBULL, "art\\characters\\creatures\\dragon\\animation\\dblow_blunt.gad"},
	{0x39BFF8EBC5468862ULL, "pro000_soldier_leader.bnk"},
	{0x39C05341F6E2C7F1ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_2x.dds"},
	{0x39C07FFA7F41E9EDULL, "zz_lgt_2_debug6140441.cl"},
	{0x39C085F145C81E77ULL, "hlac_lt_0_lowtown_docks_a3_65535_0x.dds"},
	{0x39C0DE74E0DD386AULL, "vxlt_0_lowtown_a3_n_c_emy_0_0.vlm"},
	{0x39C0FCD659A04AC9ULL, "424888384.ogg"},
	{0x39C1517AE6A722E9ULL, "hlac_lt_0_lowtown_a2_n_1_2x.dds"},
	{0x39C19C21FB757FACULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_tal_aoe.ani"},
	{0x39C1EB8BE5485610ULL, "ht_0_hightownexplosion_n_9_0.msh"},
	{0x39C1F98BE5486E2AULL, "ht_0_hightownexplosion_n_9_0.mmh"},
	{0x39C1FD8BE5487561ULL, "ht_0_hightownexplosion_n_9_0.mao"},
	{0x39C20489E90A80EEULL, "6162966_m.fxe"},
	{0x39C22085DC346B79ULL, "127953976.ogg"},
	{0x39C288E9A624C0A5ULL, "vxht_0_hightown_ap1_0_0.vlm"},
	{0x39C2A653E9316E28ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv3.ani"},
	{0x39C3074DA99B720BULL, "ht_mansion_n_3374.anb"},
	{0x39C362E4BBB73A33ULL, "6115396_m.fxe"},
	{0x39C3DE245E1B1728ULL, "22061087.ogg"},
	{0x39C42673BC91466EULL, "lt_0_lowtown_a2_n_c_37_0_lc.mmh"},
	{0x39C42A445158C710ULL, "fen210_hadriana6110582.cl"},
	{0x39C45AD3B65BD444ULL, "vxlt_hangedman_2gl_0_0.vlm"},
	{0x39C537BD671B8A82ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_lng_c.impa.evt"},
	{0x39C545C6C33AEE9AULL, "arc_blur.mao"},
	{0x39C5614FAA460E10ULL, "hlac_lt_0_lowtown_a3_n_c_65538_0z.dds"},
	{0x39C5B68DA7251A37ULL, "dae000_arishok6081596.cl"},
	{0x39C5F85315463344ULL, "hlac_ht_0_hightown_a3_-1_1y.dds"},
	{0x39C61AB3B63067D2ULL, "dae000_male_commoner2_mob.cnv"},
	{0x39C623436B8A8C6DULL, "hlac_lt_0_lowtown_n_131071_4z.dds"},
	{0x39C63957514CB259ULL, "vxw_bonepitmining_a2_hk_0_0.vlm"},
	{0x39C63F483E5FB8BAULL, "vxht_0_hightown_c3l_0_1.vlm"},
	{0x39C667505CFD64AAULL, "6075936_m.fxe"},
	{0x39C69B0E14B71BD6ULL, "598051514.ogg"},
	{0x39C6AC31B42ABFEDULL, "plt_zz_andpt_debug.plo"},
	{0x39C6BA365C71D94BULL, "gen00flst_varric_actone.stg"},
	{0x39C70CE6507A43C0ULL, "art\\characters\\creatures\\dragon\\animation\\dblow_blunt.evt"},
	{0x39C7321A190E50E4ULL, "loadhints_ambush.gda"},
	{0x39C7DCF918D503BFULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgta_0t.dds"},
	{0x39C7F61291FADD6FULL, "mag311_keran6061817.cl"},
	{0x39C83FBB4BAA6825ULL, "6110875_f.fxe"},
	{0x39C84E2C89B9CD8DULL, "444557802.ogg"},
	{0x39C8790877E0EE09ULL, "6108565_m.fxe"},
	{0x39C8A921FB7BB4EEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_tal_aoe.gad"},
	{0x39C8C649E57B2F88ULL, "214231498.ogg"},
	{0x39C8EC2998CAB526ULL, "t1_mul_pi3.tnt"},
	{0x39C93CB3B632D72AULL, "dae000_male_commoner2_mob.bnk"},
	{0x39C9A67C8386449AULL, "vxdr_primevaltomb_3ti_0_0.vlm"},
	{0x39C9ACD35129EB15ULL, "6039522_m.fxe"},
	{0x39C9D707B4E84443ULL, "vxlt_blackemporium_14q_0_0.vlm"},
	{0x39C9F38781CA32F0ULL, "dae300_cullen6140622.cl"},
	{0x39CABDFE2EB8E935ULL, "449259215.ogg"},
	{0x39CAD727301AB687ULL, "vxca_cave_2gm_0_0.vlm"},
	{0x39CAFF1C9B7DAEACULL, "vxw_encounter1_d9_0_0.vlm"},
	{0x39CB28871D6CE5B1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn5plank.msh"},
	{0x39CB36871D6CFDCCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn5plank.mao"},
	{0x39CB42871D6D126FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn5plank.mmh"},
	{0x39CB91F8008AC4D2ULL, "vxlt_gamlenshouse_n_nd_0_0.vlm"},
	{0x39CB95675DBCF2BEULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_012a_0d.dds"},
	{0x39CBB841E4E2C4CFULL, "vxw_blightlands_kq_0_0.vlm"},
	{0x39CBEBF9A8F20DD3ULL, "vxw_bonepitmining_a2_1i1_0_0.vlm"},
	{0x39CC36BD6721A795ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_lng_c.impa.gad"},
	{0x39CCCC5782974078ULL, "6174448_m.fxe"},
	{0x39CD7DF07A1DD165ULL, "877132625.ogg"},
	{0x39CD8FEEEF3ECCC4ULL, "453772642.ogg"},
	{0x39CDBBFF84FBB9FDULL, "6128664_m.fxe"},
	{0x39CE00846B7C513FULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_0x.dds"},
	{0x39CE94E00DF474D1ULL, "224700736.ogg"},
	{0x39CE951F4D8A6E06ULL, "mrl210_merrill6117472.cl"},
	{0x39CFA421FB81CACDULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_tal_aoe.evt"},
	{0x39CFFBD35E6B8BC8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_flat.phy"},
	{0x39D0154B731DFEFAULL, "vxca_caverns1_j5j_0_0.vlm"},
	{0x39D0AA052B2C70EAULL, "220997188.ogg"},
	{0x39D0F84192CF3F60ULL, "816512441.ogg"},
	{0x39D1041C4476A5A5ULL, "vxca_caverns2_9f4_0_0.vlm"},
	{0x39D14F470897692BULL, "6094464_m.fxe"},
	{0x39D1769D26C6D6FEULL, "hlac_dr_primevaltomb_665_2y.dds"},
	{0x39D292142BCC63F5ULL, "hlac_lt_0_lowtown_docks_a3_-65540_1z.dds"},
	{0x39D2DD10CDBDC5E3ULL, "llac_lt_hangedman_-1_6.dds"},
	{0x39D36A924A5B408CULL, "vxlt_0_lowtown_n_fga_0_0.vlm"},
	{0x39D379D9EC8A1B96ULL, "hlac_ca_caverns1_mines_24425_7z.dds"},
	{0x39D4796E8AFCD013ULL, "gen00fl_isabela6123221.cl"},
	{0x39D482FAA6B220A0ULL, "vxw_woundedcoast_i6_0_0.vlm"},
	{0x39D48564C156AC9EULL, "qun160_vashothlead.stg"},
	{0x39D4B6034458871BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleft.msh"},
	{0x39D4C8034458A666ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleft.mao"},
	{0x39D4D4034458BA05ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleft.mmh"},
	{0x39D504BB7FB43679ULL, "dae000_mother6103124.cl"},
	{0x39D5589FE6148ED3ULL, "gen_im_cns_flu_bethany_01.uti"},
	{0x39D55F1F46907445ULL, "vxw_encounter2_camp_6c_0_0.vlm"},
	{0x39D5B442CD4A33AAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww13_gallowsalcov.mao"},
	{0x39D5B842CD4A3AF9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww13_gallowsalcov.mmh"},
	{0x39D5C242CD4A4BE7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww13_gallowsalcov.msh"},
	{0x39D5DF483D2C0A6CULL, "6178471_m.fxe"},
	{0x39D604C875CD6A46ULL, "dae000_bodhan_letter6144272.cl"},
	{0x39D630F7859E6337ULL, "6137052_f.fxe"},
	{0x39D67985A0FFA013ULL, "vxw_encounter2_32_0_0.vlm"},
	{0x39D6F915DA4F3B01ULL, "vxw_encounter2_p1_0_0.vlm"},
	{0x39D72F5E71755130ULL, "hlac_lt_darktown_-1_46z.dds"},
	{0x39D7359AC769AF29ULL, "mrl150_merrill6093530.cl"},
	{0x39D745F5E684F065ULL, "ss_tanner_deep_roads.cut"},
	{0x39D798DE2D920DEEULL, "vxht_keep_n_682_0_0.vlm"},
	{0x39D7C603633319C9ULL, "643809234.ogg"},
	{0x39D8BF5987405CB5ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cenotaph.msh"},
	{0x39D8C1598740602BULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cenotaph.mmh"},
	{0x39D8C31A40125FEAULL, "118636029.ogg"},
	{0x39D8CD59874074C8ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cenotaph.mao"},
	{0x39D975079A3E292EULL, "894790548.ogg"},
	{0x39D97CAC3745A3A9ULL, "vxlt_0_lowtown_docks_b5x_0_0.vlm"},
	{0x39D9CBFF22411B3AULL, "gen00fl_aveline6120756.cl"},
	{0x39DA2A1ED6D8DD4FULL, "vxw_blightlands_lp_0_0.vlm"},
	{0x39DA82BF5A35342BULL, "6173419_m.fxe"},
	{0x39DA871F313E71F1ULL, "art\\characters\\weapons\\quivers\\w_qui_dq03a.mmh"},
	{0x39DA931F313E8592ULL, "art\\characters\\weapons\\quivers\\w_qui_dq03a.mao"},
	{0x39DA9C9A5E1778CCULL, "6172613_m.fxe"},
	{0x39DAA11F313E9DAFULL, "art\\characters\\weapons\\quivers\\w_qui_dq03a.msh"},
	{0x39DB9F1C0299F545ULL, "581918313.ogg"},
	{0x39DBD60279566FA8ULL, "6124298_m.fxe"},
	{0x39DC26840BEA57C7ULL, "vxlt_warehouse_rr_0_0.vlm"},
	{0x39DC3FE7637DE514ULL, "6110868_m.fxe"},
	{0x39DC4F6F839A153CULL, "follower_banter6149294.cl"},
	{0x39DCC66876D24E3BULL, "vxlt_0_lowtown_bz2_0_0.vlm"},
	{0x39DD36ABC69B032FULL, "vxlt_backalley_n_1xw_0_0.vlm"},
	{0x39DDD9B7E13BE4B8ULL, "997930728.ogg"},
	{0x39DE719DB4A0D007ULL, "hlac_lt_blackemporium_1055_14x.dds"},
	{0x39DE99F7F54F6DBCULL, "544910375.ogg"},
	{0x39DEC82C89A3AD25ULL, "mcr310_sibling6076319.cl"},
	{0x39DEF388E26316D4ULL, "art\\vfx\\combat\\vfx_gib_undead_bld_p.anb"},
	{0x39DF220917DF4AC5ULL, "427868820.ogg"},
	{0x39E144548A869742ULL, "follower_banter6168419.cl"},
	{0x39E1504473D665DDULL, "mag311_emile_de_launcet6043501.cl"},
	{0x39E186017B156AA8ULL, "6119482_f.fxe"},
	{0x39E20CBE109367C7ULL, "hlac_lt_gallowstemplar_f_2388_5x.dds"},
	{0x39E2A45E3AB82519ULL, "57031601.ogg"},
	{0x39E2BF8BC203BB64ULL, "vxht_mansion2_2of_0_0.vlm"},
	{0x39E35E477D924F66ULL, "40934996.ogg"},
	{0x39E37EDA155790B6ULL, "hlac_lt_backalley_g_0_11z.dds"},
	{0x39E3D0FC38868CC5ULL, "vxw_encounter2_camp_1b_0_0.vlm"},
	{0x39E3D14C7A3E63D1ULL, "37472318.ogg"},
	{0x39E3DB95A8EEE171ULL, "371341923.ogg"},
	{0x39E41EC6C3555D4CULL, "arc_blur.dds"},
	{0x39E42B1D03D4ECEBULL, "1003680995.ogg"},
	{0x39E43EA27F39EBEBULL, "793367587.ogg"},
	{0x39E447AC305178E7ULL, "one111ip_ggt_crate_7.utp"},
	{0x39E48661ACE3292BULL, "lt_0_lowtown_docks_a3_n_26_0.rml"},
	{0x39E4AD2BEF1E01AEULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v2.evt"},
	{0x39E4E22C49AFA1A1ULL, "one300_nuncio6083512.cl"},
	{0x39E53D921CACC843ULL, "and100_anders6095803.cl"},
	{0x39E5B9CF4F926A8AULL, "vxlt_0_lowtown_eoc_0_0.vlm"},
	{0x39E60ABB1F88266EULL, "vxht_keep_n_633_0_0.vlm"},
	{0x39E67A5D566D6733ULL, "gen00fl_sebastian6179337.cl"},
	{0x39E69640D752F9DBULL, "art\\characters\\playercharacter\\humanmale\\hlm_masd.mmh"},
	{0x39E6A240D7530E78ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masd.mao"},
	{0x39E6A96639A606CBULL, "mrl000_marethari6205609.cl"},
	{0x39E6B440D7532CC5ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masd.msh"},
	{0x39E7BAE11406E787ULL, "vxca_cave_2mo_0_0.vlm"},
	{0x39E89BFBC8CEF5CFULL, "vxw_blightlands_qs_0_0.vlm"},
	{0x39E90E77500D914CULL, "6175713_m.fxe"},
	{0x39E91BCF9C1207DEULL, "981889089.ogg"},
	{0x39E930F054DBF0CCULL, "6125508_f.fxe"},
	{0x39E93FC18C97C416ULL, "6099307_m.fxe"},
	{0x39E9DC5F4BD15160ULL, "plt_cod_mgc_annulment.plo"},
	{0x39EA21C6E2C37691ULL, "149595343.ogg"},
	{0x39EA5B6B1D9E53F6ULL, "139851801.ogg"},
	{0x39EA82D53F5CA4B0ULL, "lgt210_eau_symp_templar6144472.cl"},
	{0x39EA9081EF7B89D4ULL, "hlac_ht_chantry_n_2534_22z.dds"},
	{0x39EA9860FDE07047ULL, "vxlt_warehouse_wm_0_0.vlm"},
	{0x39EACBE53728F00DULL, "6119619_m.fxe"},
	{0x39EADD155DD17C83ULL, "vxlt_warehouse_n_13g_0_0.vlm"},
	{0x39EB241C18709ABCULL, "gen_im_trs_dem_blt_r1.uti"},
	{0x39EBA42BEF241159ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v2.gad"},
	{0x39EBD288E26D6897ULL, "art\\vfx\\combat\\vfx_gib_undead_bld_p.mmh"},
	{0x39EBE6850FA76DA4ULL, "224566775.ogg"},
	{0x39EBF4F5D3674C88ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_003a_0t.dds"},
	{0x39EC02228944AB09ULL, "vxw_bonepit_yv_0_0.vlm"},
	{0x39EC1A876F1A4890ULL, "6016843_m.fxe"},
	{0x39ECBA91DCBD1464ULL, "6173038_m.fxe"},
	{0x39ED054C32EEB536ULL, "507067744.ogg"},
	{0x39ED3F683CDE22B4ULL, "gen_im_cft_reg_j.uti"},
	{0x39ED8FB9698E4896ULL, "682303976.ogg"},
	{0x39EDB38799B9C1ADULL, "llac_ht_brothel_n_3382_0.dds"},
	{0x39EDD7B97B691581ULL, "zz_sexyhawke.cut"},
	{0x39EDDB378D40FA4EULL, "hlac_lt_0_lowtown_a2_-196608_4z.dds"},
	{0x39EE5BE56B4C70E6ULL, "mag100_cullen6093790.cl"},
	{0x39EE78B88A23B166ULL, "lt_0_lowtown_a2_53_0_ll.mmh"},
	{0x39EEA68900B84FD2ULL, "6187766_f.fxe"},
	{0x39EEA9D624F1D05CULL, "vxw_blightlands_6i_0_0.vlm"},
	{0x39EEB178E97AEC16ULL, "hlac_lt_0_lowtown_a2_-1_2y.dds"},
	{0x39EEEDBCB350A772ULL, "gen_im_wep_war_one_lore_03.uti"},
	{0x39EF044DA9C0A857ULL, "ht_mansion_n_3374.rml"},
	{0x39EF86CE6DB7A06FULL, "dae000_merchant_mage_hf6164461.cl"},
	{0x39EF98FB3FDC2FBCULL, "black.vsh.fxo"},
	{0x39EFA3D1F3F221B1ULL, "6082289_m.fxe"},
	{0x39F0704830850434ULL, "159628155.ogg"},
	{0x39F07B780C1EC702ULL, "445741799.ogg"},
	{0x39F0925DA394EE61ULL, "vxht_mansion_n_29v_0_0.vlm"},
	{0x39F0B3DEFA22D52DULL, "vxht_mansion_n_4zy_0_0.vlm"},
	{0x39F13FBE9F38A000ULL, "6129330_m.fxe"},
	{0x39F166B48A9E51A0ULL, "vxw_woundedcoast_c4_0_0.vlm"},
	{0x39F1A44FB09A7621ULL, "static_static_p24.psh.fxo.dx11.assembly"},
	{0x39F29AC8A0DDC662ULL, "377884227.ogg"},
	{0x39F2B52BEF2A4CA7ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v2.ani"},
	{0x39F2FC578CC4B989ULL, "vxlt_0_lowtown_a3_n_c_epd_0_1.vlm"},
	{0x39F35D3F84EBD113ULL, "vxw_encounter2_94_0_0.vlm"},
	{0x39F3661F9D3FDEEFULL, "hlac_ht_0_hightown_a3_65535_3z.dds"},
	{0x39F3A128B1D93CB5ULL, "qun101ar_talvashoth_cave.are"},
	{0x39F423BA014A81C0ULL, "6084291_m.fxe"},
	{0x39F481A0AD30D76DULL, "vxht_keep_n_67x_0_0.vlm"},
	{0x39F48FC9E8DAA877ULL, "w_woundedcoast_211_0.mmh"},
	{0x39F493C9E8DAAE84ULL, "w_woundedcoast_211_0.mao"},
	{0x39F4A5C9E8DACDE9ULL, "w_woundedcoast_211_0.msh"},
	{0x39F4A69F640A53C3ULL, "vxlt_0_lowtown_a2_eva_0_0.vlm"},
	{0x39F4A87AE5335125ULL, "709668689.ogg"},
	{0x39F4DC0C61F6926BULL, "299302897.ogg"},
	{0x39F4EF55ED92D9B3ULL, "hlac_w_encounter1_65539_0x.dds"},
	{0x39F500F8E776912FULL, "106588410.ogg"},
	{0x39F53D8538640413ULL, "6145142_m.fxe"},
	{0x39F5BB6193C4A076ULL, "6172004_m.fxe"},
	{0x39F62AFB1DE02DA6ULL, "vxca_caverns2_98a_0_0.vlm"},
	{0x39F67C372FA14928ULL, "hlac_dr_dungeon_3214_2x.dds"},
	{0x39F6B445B7F401AFULL, "vxca_caverns1_dq5_0_0.vlm"},
	{0x39F727669845A9E8ULL, "wm_wilderness.map"},
	{0x39F741907ED45753ULL, "vxw_bonepitmining_a2_1x2_0_0.vlm"},
	{0x39F7585726E5E693ULL, "dae000_keep_maid_sta6162425.cl"},
	{0x39F82858A965ABC0ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_8z.dds"},
	{0x39F82C7BD468F3A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_assassinate.gad"},
	{0x39F896CA66994161ULL, "fen210_dying_magister6108623.cl"},
	{0x39F8D72233037753ULL, "mcr371_scared_mage_m6087692.cl"},
	{0x39F8E984E946CC43ULL, "vxht_mansion_n_618_0_0.vlm"},
	{0x39F90A3DEFD688C7ULL, "vxlt_warehouse_xl_0_0.vlm"},
	{0x39F94EF24FC79503ULL, "vxlt_warehouse_n_18f_0_0.vlm"},
	{0x39F9899657FD6C69ULL, "lgt110_orz_harrowmont6109035.cl"},
	{0x39F9EB18DC5EE15AULL, "30938672.ogg"},
	{0x39FA147CD8B2E930ULL, "190124780.ogg"},
	{0x39FA2F2239BEEBBBULL, "hlac_lt_0_lowtown_docks_131071_2x.dds"},
	{0x39FA998B4C5EE38DULL, "mag320_nobleman_36060271.cl"},
	{0x39FB0CEF77928E45ULL, "vxca_caverns1_j5a_0_0.vlm"},
	{0x39FB990F5CBC8238ULL, "314730622.ogg"},
	{0x39FC112C6AC144FFULL, "6033658_m.fxe"},
	{0x39FC3A8C6B4A6948ULL, "qcr000_varric6066277.cl"},
	{0x39FCBA47CC9B72A3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stonesedge.msh"},
	{0x39FCC847CC9B8AFDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stonesedge.mmh"},
	{0x39FCCC47CC9B910EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stonesedge.mao"},
	{0x39FD77F8484F4955ULL, "vxlt_foundry_n_z7_0_0.vlm"},
	{0x39FD96DFA7169DD7ULL, "890628066.ogg"},
	{0x39FDC228B1E1C91BULL, "qun101ar_talvashoth_cave.ncs"},
	{0x39FE32832350417EULL, "6141738_m.fxe"},
	{0x39FE698EB4B24214ULL, "6141453_m.fxe"},
	{0x39FEB0082E4930CAULL, "vxdr_primevalentrance_1je_0_0.vlm"},
	{0x39FEF19F9D67FC3FULL, "63612553.ogg"},
	{0x39FF48AA2AE622B6ULL, "6102938_f.fxe"},
	{0x39FF4F7BD46F4D99ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_assassinate.evt"},
	{0x39FFD78F8451C64DULL, "673538386.ogg"},
	{0x39FFDF72CDAF061EULL, "hlac_lt_0_lowtown_a2_131072_3x.dds"},
	{0x3A00914706ECF8CCULL, "hlac_lt_blackemporium_1055_8x.dds"},
	{0x3A00A70ED7A51098ULL, "6069917_m.fxe"},
	{0x3A00C2464028F5D2ULL, "vxw_bonepitmining_a2_1x7_0_0.vlm"},
	{0x3A00C4A836C497D7ULL, "6139736_m.fxe"},
	{0x3A00F248EAA59B6EULL, "one300st_nuncio.stg"},
	{0x3A0100B78623CDE6ULL, "398650030.ogg"},
	{0x3A0186AF0437F51AULL, "217707215.ogg"},
	{0x3A018F0C26639AEEULL, "hlac_ht_brothel_n_2995_2z.dds"},
	{0x3A01A88DFA1B75D3ULL, "vxht_playershouse_n_1lq_0_0.vlm"},
	{0x3A01D9F99AB4725EULL, "6172949_m.fxe"},
	{0x3A01FCE5C26C32DFULL, "vxht_0_hightown_n_c4z_0_0.vlm"},
	{0x3A02411025A79143ULL, "w_encounter2_20_0.rml"},
	{0x3A0242DCA1346F3BULL, "vxht_0_hightown_a2_a2q_0_0.vlm"},
	{0x3A02A3C2F0CF6CEAULL, "273904409.ogg"},
	{0x3A02D122A37D2984ULL, "457746670.ogg"},
	{0x3A02DD6CCDFFE899ULL, "vxdr_primevalentrance_2yt_0_0.vlm"},
	{0x3A02F37D9F26EFEDULL, "vxht_keep_n_62w_0_0.vlm"},
	{0x3A032CAA6B4E9318ULL, "vxlt_0_lowtown_docks_n_aoo_0_0.vlm"},
	{0x3A03447A1B277CDBULL, "6054969_m.fxe"},
	{0x3A03706432A4F1F2ULL, "zz_followers_banter6158106.cl"},
	{0x3A03F95CF3B09639ULL, "575306591.ogg"},
	{0x3A03FAE617CAF141ULL, "hlac_lt_0_lowtown_a3_n_c_131074_0x.dds"},
	{0x3A042D3E85BAB8F6ULL, "6175104_m.fxe"},
	{0x3A04A228B1E7B1C2ULL, "qun101ar_talvashoth_cave.lst"},
	{0x3A04AFB9E98080CCULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rtl.ani"},
	{0x3A04D91B0EDD2668ULL, "vxlt_0_lowtown_a3_n_c_eqf_0_1.vlm"},
	{0x3A04E2803549E9DDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_basic.mao"},
	{0x3A04E6803549F016ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_basic.mmh"},
	{0x3A04F080354A0114ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp12_basic.msh"},
	{0x3A05565584CFAE36ULL, "6175940_m.fxe"},
	{0x3A056E4328633664ULL, "vxw_bonepitdestroyed_a3_183_0_0.vlm"},
	{0x3A057FCA02CDB95DULL, "339208974.ogg"},
	{0x3A05B36D70CA6FD3ULL, "vxw_bonepitmining_a2_1u5_0_0.vlm"},
	{0x3A05CDFA46192761ULL, "vxw_encounter2_camp_23_0_0.vlm"},
	{0x3A05F9C904F51EB3ULL, "hlac_lt_gallowstemplar_7198_6x.dds"},
	{0x3A0616C8E9CB7730ULL, "gen00fl_aveline6137677.cl"},
	{0x3A06B37A10978896ULL, "444277.ogg"},
	{0x3A070399E0C11885ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.anb"},
	{0x3A074A03B0D04EDEULL, "vxlt_0_lowtown_docks_n_bze_0_0.vlm"},
	{0x3A07504DCAEF8D1FULL, "vxlt_0_lowtown_a2_eoo_0_1.vlm"},
	{0x3A079964C18228E5ULL, "qun160_vashothlead.bnk"},
	{0x3A07AF9F1B15210DULL, "6113019_m.fxe"},
	{0x3A07B096E7B12FB7ULL, "optionsmenu.gfx"},
	{0x3A07D8175A2F15E8ULL, "dae000_gamlen6005826.cl"},
	{0x3A07DF56C568F8F4ULL, "gen00fl_carver6102335.cl"},
	{0x3A0819992859D4F5ULL, "vxlt_warehouse_n_r1_0_0.vlm"},
	{0x3A0855AC4FAEA2F8ULL, "hlac_dr_primevaltomb_657_4x.dds"},
	{0x3A08676F106F173CULL, "hlac_w_encounter2_camp_196606_0x.dds"},
	{0x3A089181F62DC396ULL, "6104841_m.fxe"},
	{0x3A08C32FAE65B988ULL, "46023120.ogg"},
	{0x3A091EDA07F3EC99ULL, "dae000ip_crowd_mf3_drunk2.utp"},
	{0x3A09313B048B8EB3ULL, "art\\characters\\creatures\\qunari\\textures\\qm_har_stn_0s.dds"},
	{0x3A09481025ADBBBBULL, "w_encounter2_20_0.phy"},
	{0x3A098F31ED748E57ULL, "133292841.ogg"},
	{0x3A09907C949195F7ULL, "6098517_f.fxe"},
	{0x3A09E72619F15D3DULL, "lt_0_lowtown_docks_a3_n_220_0.mao"},
	{0x3A09EB2619F16476ULL, "lt_0_lowtown_docks_a3_n_220_0.mmh"},
	{0x3A09F52619F17574ULL, "lt_0_lowtown_docks_a3_n_220_0.msh"},
	{0x3A0A2BD67DF884E0ULL, "vxlt_0_lowtown_docks_a3_n_d48_0_0.vlm"},
	{0x3A0A642751EBC4B5ULL, "6067367_m.fxe"},
	{0x3A0A9FCEF6304B63ULL, "vxht_0_hightown_a2_b9o_0_0.vlm"},
	{0x3A0AD94525A86876ULL, "vxca_caverns2_79m_0_1.vlm"},
	{0x3A0B2B439B642E8FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.impa.gad"},
	{0x3A0B6B59988827FBULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsb_0n.dds"},
	{0x3A0B7B64C185DE1DULL, "qun160_vashothlead.cnv"},
	{0x3A0BBCB9E986B50EULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rtl.gad"},
	{0x3A0BD1DB895720EFULL, "dae221ip_at_cht_htown.utp"},
	{0x3A0BD8EAFC023BF9ULL, "194336850.ogg"},
	{0x3A0C0356E52E7AE7ULL, "pri_water.mao"},
	{0x3A0C577BD479E588ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_assassinate.ani"},
	{0x3A0D0A05B552F6D1ULL, "6107050_m.fxe"},
	{0x3A0D759D5CF548B7ULL, "6144343_m.fxe"},
	{0x3A0DE1F0F0F3565FULL, "6108592_m.fxe"},
	{0x3A0E188C3AE21072ULL, "634501751.ogg"},
	{0x3A0E69319C90284EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_lft_180.evt"},
	{0x3A0F56B0076DD348ULL, "vxw_encounter2_camp_26_0_0.vlm"},
	{0x3A0FCB7F75118923ULL, "6084506_m.fxe"},
	{0x3A0FD702A81C276FULL, "vxht_chantry_34u_0_0.vlm"},
	{0x3A0FEAA34390C2BAULL, "hlac_w_encounter2_2_0z.dds"},
	{0x3A10201157563B7AULL, "vxht_0_hightown_n_a85_0_0.vlm"},
	{0x3A10221ACB98C309ULL, "vxlt_0_lowtown_n_f1r_0_0.vlm"},
	{0x3A106EC2B4624B5FULL, "vxht_0_hightown_n_c1e_0_0.vlm"},
	{0x3A1113E5CC6BD181ULL, "vxw_blightlands_68_0_0.vlm"},
	{0x3A113CC1EDCE095DULL, "hlac_ca_caverns1_24424_26x.dds"},
	{0x3A1140439B68BE3CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.impa.evt"},
	{0x3A1198EB4165CA4BULL, "91205529.ogg"},
	{0x3A11B695A55DF693ULL, "vxw_sundermount_2bn_0_0.vlm"},
	{0x3A11D6175A12B181ULL, "6117796_m.fxe"},
	{0x3A12AB063825B571ULL, "vxca_cavern1_haunted_ck7_0_0.vlm"},
	{0x3A12F425A614D604ULL, "vxca_cavern1_haunted_jfd_0_0.vlm"},
	{0x3A132351C6C93721ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.dura.ani"},
	{0x3A13F563C16FBD3BULL, "6105380_m.fxe"},
	{0x3A14D5D557CE4235ULL, "vxlt_0_lowtown_n_cms_0_0.vlm"},
	{0x3A14FEF178C6080FULL, "6077327_f.fxe"},
	{0x3A155E318DFAEE1DULL, "lt_0_lowtown_docks_156_0.mao"},
	{0x3A1560319C9637F9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_lft_180.gad"},
	{0x3A1562318DFAF556ULL, "lt_0_lowtown_docks_156_0.mmh"},
	{0x3A156C318DFB0654ULL, "lt_0_lowtown_docks_156_0.msh"},
	{0x3A158DEBF18681A6ULL, "gen00fl_anders6125727.cl"},
	{0x3A15AF214B5ABD27ULL, "6172997_m.fxe"},
	{0x3A160C80F9DF4DA5ULL, "6083475_m.fxe"},
	{0x3A166FC1CA08466AULL, "153820791.ogg"},
	{0x3A168B761A4FED75ULL, "vxlt_warehouse_n_y0_0_0.vlm"},
	{0x3A178D972CA878B1ULL, "w_bonepit_229_0.mmh"},
	{0x3A1799972CA88D52ULL, "w_bonepit_229_0.mao"},
	{0x3A17A7972CA8A56FULL, "w_bonepit_229_0.msh"},
	{0x3A17E5431ADBB3ADULL, "415670543.ogg"},
	{0x3A183525B92D5142ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_rf.gad"},
	{0x3A1848B3DD3FF40CULL, "vxlt_0_lowtown_docks_a3_cyj_0_1.vlm"},
	{0x3A1869DEA45B5A21ULL, "vxdr_primevaltomb_4zu_0_2.vlm"},
	{0x3A18A9DE21F0EF01ULL, "character_character_p125.psh.fxo.dx11"},
	{0x3A1A2E51C6CF68C3ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.dura.gad"},
	{0x3A1A47490D1E39B7ULL, "6102791_f.fxe"},
	{0x3A1A675FC9603DA4ULL, "pro000_friedrich6050159.cl"},
	{0x3A1AA5E9A6655A88ULL, "hf_amb_fereldan.mor"},
	{0x3A1AD73390727B65ULL, "m_qun210ar_fq_random_enc.wbd"},
	{0x3A1B6611C33B1171ULL, "6140257_m.fxe"},
	{0x3A1B71319C9AC047ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_lft_180.ani"},
	{0x3A1BB6B69E6568DDULL, "dae030cr_male_fer_urchin.utc"},
	{0x3A1C10914DFEA477ULL, "6141807_m.fxe"},
	{0x3A1C2FE0E758C0E8ULL, "vxlt_hangedman_n_21m_0_0.vlm"},
	{0x3A1C9DAA7FA637EDULL, "6037128_m.fxe"},
	{0x3A1CBC4B60809B20ULL, "vxw_woundedcoast_r3_0_0.vlm"},
	{0x3A1CFD226613CF35ULL, "228317765.ogg"},
	{0x3A1CFFE4BD29FAAEULL, "follower_banter6168042.cl"},
	{0x3A1D09ED852528E1ULL, "vxca_caverns1_j42_0_0.vlm"},
	{0x3A1D908C63AFD3F2ULL, "705858373.ogg"},
	{0x3A1DB84628B40FC8ULL, "6197146_m.fxe"},
	{0x3A1DF0956F84664DULL, "249261940.ogg"},
	{0x3A1E7BF47766CACEULL, "vxw_blightlands_12v_0_0.vlm"},
	{0x3A1E93F7BD8EDB89ULL, "vxlt_0_lowtown_n_f4q_0_0.vlm"},
	{0x3A1EFA0CD5F54E21ULL, "zz_approval_debug6133723.cl"},
	{0x3A1F48439B750955ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.impa.ani"},
	{0x3A1F5825B933AB39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_rf.evt"},
	{0x3A1FA18F0F6D0C62ULL, "dae000_templar_knight6165165.cl"},
	{0x3A2003310413D610ULL, "dae000_orana6149008.cl"},
	{0x3A2010644F3AC418ULL, "vxlt_0_lowtown_docks_n_aqm_0_0.vlm"},
	{0x3A20253150284911ULL, "hlac_lt_darktown_-1_69x.dds"},
	{0x3A20B126085C204EULL, "gen00fl_varric6072986.cl"},
	{0x3A20E69F548B0EAEULL, "6171269_m.fxe"},
	{0x3A210B51C6D54BA8ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.dura.evt"},
	{0x3A21260508521B9EULL, "character_character_p229.psh.fxo.dx11.assembly"},
	{0x3A21A41EAF599287ULL, "llac_w_encounter2_camp_1_0.dds"},
	{0x3A22153421351F57ULL, "gen00fl_varric6079948.cl"},
	{0x3A229AE7FEB24B6AULL, "6103196_m.fxe"},
	{0x3A229ED1150F8D99ULL, "vxht_0_hightown_n_avx_0_0.vlm"},
	{0x3A247990227394C5ULL, "gen_im_acc_rng_act1_04.uti"},
	{0x3A2497FC8D47F49EULL, "92090804.ogg"},
	{0x3A24FC9895725424ULL, "vxlt_0_lowtown_docks_n_b9b_0_0.vlm"},
	{0x3A2524862315A4FCULL, "vxht_brothel_n_5ht_0_0.vlm"},
	{0x3A254C284C668C75ULL, "llac_lt_0_lowtown_a2_n_-196608_0.dds"},
	{0x3A25730D7DF0A9DBULL, "vxca_caverns2_8bc_0_0.vlm"},
	{0x3A2588789D5A22A3ULL, "lgt220_shp_mag_guard6149916.cl"},
	{0x3A25A7833A8E90F7ULL, "vxht_0_hightown_a2_a7j_0_0.vlm"},
	{0x3A25ADC1DB157037ULL, "hlac_lt_0_lowtown_n_131070_2y.dds"},
	{0x3A267147536A5040ULL, "vxw_bonepitmining_a2_25y_0_0.vlm"},
	{0x3A2675656A96E6D8ULL, "art\\vfx\\materials\\vfx_particle_fade.mmh"},
	{0x3A2692A34679D4C8ULL, "vxca_caverns1_j47_0_0.vlm"},
	{0x3A269C4B7FD62DB0ULL, "vxw_encounter2_qj_0_1.vlm"},
	{0x3A26CD30A1BF1A00ULL, "vxw_blightlands_i8_0_0.vlm"},
	{0x3A26D613087C0061ULL, "m_dae370ar_gallows.bnk"},
	{0x3A27074DE376C723ULL, "489302809.ogg"},
	{0x3A27155C4D7AFC4EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxb.msh"},
	{0x3A27235C4D7B139FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxb.mao"},
	{0x3A272F5C4D7B283CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxb.mmh"},
	{0x3A275B57F6A7A375ULL, "215923113.ogg"},
	{0x3A27F6865FFDC188ULL, "345880785.ogg"},
	{0x3A280F40A0033DE0ULL, "zz_followers_banter6158027.cl"},
	{0x3A2860DD28C018C0ULL, "756663502.ogg"},
	{0x3A28A5CC71AB5A7FULL, "ave200_donnic6110706.cl"},
	{0x3A2900D6489C8963ULL, "143749582.ogg"},
	{0x3A2951A888535E1CULL, "6118954_m.fxe"},
	{0x3A29788FAF0E207BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand.mmh"},
	{0x3A298410DE4E5E80ULL, "vxlt_0_lowtown_a3_n_c_byh_0_0.vlm"},
	{0x3A29848FAF0E3418ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand.mao"},
	{0x3A29968FAF0E5365ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand.msh"},
	{0x3A29C82E06B13E45ULL, "vxdr_primevalentrance_221_0_0.vlm"},
	{0x3A2A34D96118AEC7ULL, "gen_im_trs_crc_junk2_02.uti"},
	{0x3A2AAD2719772D3AULL, "6116363_m.fxe"},
	{0x3A2B14C112889CB9ULL, "qun111_ketojan6019016.cl"},
	{0x3A2B6233DA9D4309ULL, "mag370_meredith_st.stg"},
	{0x3A2B9322E62B7A39ULL, "6084749_f.fxe"},
	{0x3A2C2306E09A17C0ULL, "589073792.ogg"},
	{0x3A2C6025B93E4328ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_rf.ani"},
	{0x3A2C6FCDB5B73D29ULL, "zz_mer_debug6045638.cl"},
	{0x3A2C770090E6659CULL, "6031170_m.fxe"},
	{0x3A2C7D8AFDB6E333ULL, "416790147.ogg"},
	{0x3A2C85CFC6611F35ULL, "gen00fl_fenris6127010.cl"},
	{0x3A2C99E4AB14831BULL, "842879401.ogg"},
	{0x3A2CFBA00B32B532ULL, "vxlt_hangedman_27z_0_0.vlm"},
	{0x3A2D3DF0F9A234A7ULL, "gen_im_cns_flu_aveline_04.uti"},
	{0x3A2D4AD3CE6143C3ULL, "llac_ht_chantry_n_2534_43.dds"},
	{0x3A2E220618DAF516ULL, "6149851_m.fxe"},
	{0x3A2E6E132BE2B5C3ULL, "vxlt_0_lowtown_a2_eje_0_0.vlm"},
	{0x3A2EB740D790E710ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masd.phy"},
	{0x3A2EC251C0C085A8ULL, "vxlt_warehouse_n_cg_0_0.vlm"},
	{0x3A2F0499E0E26CC6ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.mmh"},
	{0x3A2F587C4681272EULL, "6174169_m.fxe"},
	{0x3A2F5919253A421DULL, "vxlt_gallowscourtyard_n_3ta_0_0.vlm"},
	{0x3A2FE52071EC2352ULL, "62372332.ogg"},
	{0x3A30603830241E9EULL, "vxca_caverns1_e8z_0_0.vlm"},
	{0x3A307BAFB3106947ULL, "6114023_m.fxe"},
	{0x3A30CEE7C21098C3ULL, "vxht_keep_n_6hk_0_0.vlm"},
	{0x3A30D46B65B1177AULL, "hlac_w_sundermount_-196609_0y.dds"},
	{0x3A31CC49F5C6D37DULL, "vxht_0_hightown_a3_c3k_0_0.vlm"},
	{0x3A31D08FEE4F6247ULL, "vis121_saemus6100106.cl"},
	{0x3A3235548339929EULL, "hlac_dr_primevaltomb_657_5y.dds"},
	{0x3A32ADD11F5516FFULL, "vxca_caverns1_dnt_0_0.vlm"},
	{0x3A32BA58123BF074ULL, "mer120_hubert6072904.cl"},
	{0x3A32C3E9FBF139D2ULL, "vxdr_primevaltomb_2sq_0_0.vlm"},
	{0x3A32CC22B19E8093ULL, "hlac_lt_0_lowtown_docks_n_-4_5z.dds"},
	{0x3A336E7587686CA4ULL, "vxlt_0_lowtown_docks_n_b6c_0_0.vlm"},
	{0x3A33AE7D764CF764ULL, "sib111_mother6121436.cl"},
	{0x3A33E4EA6FE6C25BULL, "vxca_caverns2_8ib_0_0.vlm"},
	{0x3A342D33C8FFF48DULL, "6148834_f.fxe"},
	{0x3A347017D16B4D56ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.cess.ani"},
	{0x3A348281F16B5F5AULL, "6118336_m.fxe"},
	{0x3A352F1025D2CC5AULL, "w_encounter2_20_0.msh"},
	{0x3A353D1025D2E393ULL, "w_encounter2_20_0.mao"},
	{0x3A353F0D93B53280ULL, "vxw_blightlands_r9_0_0.vlm"},
	{0x3A35411025D2EAA0ULL, "w_encounter2_20_0.mmh"},
	{0x3A354355A590484EULL, "6126459_f.fxe"},
	{0x3A35962D1125790BULL, "hlac_dr_0_deeproads_5843_0y.dds"},
	{0x3A364375BBFC2B97ULL, "art\\vfx\\materials\\vfx_crater_half.mao"},
	{0x3A369A39AD82EF3AULL, "hlac_ht_0_hightown_n_65535_1y.dds"},
	{0x3A36FFB4A60DC38FULL, "vxlt_0_lowtown_docks_a3_d1c_0_0.vlm"},
	{0x3A370030FCFDBBECULL, "467176623.ogg"},
	{0x3A371AF9CCA6B298ULL, "strider_heads.gda"},
	{0x3A37DB0F22224340ULL, "dae000_flemeth6000910.cl"},
	{0x3A380E4937203752ULL, "gen_im_trs_prf_amu_r2.uti"},
	{0x3A3853FD5FA8D83CULL, "hlac_dr_slavers_3214_4x.dds"},
	{0x3A3859CC0F0B3A2AULL, "vxlt_backalley_g_1jh_0_1.vlm"},
	{0x3A38B942CD9EB9E6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww13_gallowsalcov.phy"},
	{0x3A390705705F3F7CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3_lp.ani"},
	{0x3A3937DE89E1962EULL, "gen00fl_fenris6107170.cl"},
	{0x3A395D65DB4C0FECULL, "6026955_f.fxe"},
	{0x3A39B364CC59E695ULL, "6139144_f.fxe"},
	{0x3A39B5E2E806EBA9ULL, "hlac_dr_primevalentrance_1726_13y.dds"},
	{0x3A39C05156B3F6FAULL, "6080705_m.fxe"},
	{0x3A3A19B026F57F9CULL, "dae000_meredith6092554.cl"},
	{0x3A3A8FF64D642A26ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\w_var\\bianca_d.dds"},
	{0x3A3A9C000CA040C8ULL, "6199346_m.fxe"},
	{0x3A3AA9DAD5F58978ULL, "mag320_bsc_banter6182373.cl"},
	{0x3A3B2EE994E5E815ULL, "6142398_m.fxe"},
	{0x3A3B54EEDC4A3F8EULL, "qun211cr_elf_warrior.cnv"},
	{0x3A3B7001C1F3D7D3ULL, "vxht_playershouse_n_1pu_0_0.vlm"},
	{0x3A3BC6520D2D04A5ULL, "hlac_lt_0_lowtown_a3_n_c_65536_3x.dds"},
	{0x3A3BCB0CB8EC81CCULL, "244990549.ogg"},
	{0x3A3BE9992812ABBCULL, "dae000_templar_recruit_16164435.cl"},
	{0x3A3D70D10E1243C9ULL, "264497081.ogg"},
	{0x3A3DE8672E648888ULL, "p_imp_dust.mao"},
	{0x3A3E29AD9423F2DCULL, "187709761.ogg"},
	{0x3A3E52B4557C5413ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext.evt"},
	{0x3A3E76EEDC4CAFFEULL, "qun211cr_elf_warrior.bnk"},
	{0x3A3E9129A3331BAEULL, "6031976_m.fxe"},
	{0x3A3EECDF34F03797ULL, "6057076_m.fxe"},
	{0x3A3F05CB77FA09E8ULL, "hla_ht_0_hightown_131073_1x.dds"},
	{0x3A3F0966B0A7D499ULL, "dae000ip_crowd_lm4_talk10.utp"},
	{0x3A3F40C4B406B143ULL, "vxht_keep_n_6cj_0_0.vlm"},
	{0x3A3F7220D8E6FA02ULL, "vxlt_0_lowtown_docks_a3_bz5_0_0.vlm"},
	{0x3A3F7717929B81D0ULL, "hlac_ht_playershouse_n_855_30z.dds"},
	{0x3A3F84BAEA2CBBC6ULL, "6202194_m.fxe"},
	{0x3A3F889F96B3D449ULL, "hlac_lt_0_lowtown_a2_1_10x.dds"},
	{0x3A3FC9A531969DFBULL, "vxht_0_hightown_a3_b67_0_0.vlm"},
	{0x3A401405706573BEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3_lp.gad"},
	{0x3A401737C5264389ULL, "mer110_jansen6032824.cl"},
	{0x3A402AA0E85E5BEDULL, "6041503_m.fxe"},
	{0x3A40807A6A1D4171ULL, "bad_large.mao"},
	{0x3A40F607DB6B6C0CULL, "w_encounter3_37_0.phy"},
	{0x3A414F566DFD97B0ULL, "vxw_blightlands_sz_0_0.vlm"},
	{0x3A417817D175E547ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.cess.evt"},
	{0x3A429023C204361DULL, "6026180_m.fxe"},
	{0x3A42F64FFCD7F5B5ULL, "vrc221_varric6040870.cl"},
	{0x3A4300ADCEF70B59ULL, "hro_anvil_lava01a.dds"},
	{0x3A43A3900FC61FA3ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medi.mao"},
	{0x3A444BB45580B454ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext.gad"},
	{0x3A44F0B783487652ULL, "vxht_playershouse_n_1pp_0_0.vlm"},
	{0x3A458869DDDA1116ULL, "848280343.ogg"},
	{0x3A45F37F57A9A60DULL, "vxw_woundedcoast_19z_0_0.vlm"},
	{0x3A4617FE3C4051E8ULL, "vxht_bartrand_n_4kp_0_0.vlm"},
	{0x3A463066FDA3462EULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgta_0n.dds"},
	{0x3A4656CFC6A03EB0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1a.mao"},
	{0x3A465ACFC6A045C3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1a.mmh"},
	{0x3A4668CFC6A05D1DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1a.msh"},
	{0x3A46CC3FF4D48F03ULL, "zz_followers_banter6159537.cl"},
	{0x3A47CD07DB714514ULL, "w_encounter3_37_0.rml"},
	{0x3A47FB99A875D958ULL, "vxht_0_hightown_a2_b5t_0_0.vlm"},
	{0x3A4815B8604A9B4CULL, "vxlt_0_lowtown_a2_n_c_eub_0_0.vlm"},
	{0x3A48311E9BAF348FULL, "vxlt_darktown_2vu_0_0.vlm"},
	{0x3A4848219B488494ULL, "hlac_ca_caverns1_24425_6x.dds"},
	{0x3A489917D17C3B90ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.cess.gad"},
	{0x3A48F0CD3DCF3A9AULL, "6111331_m.fxe"},
	{0x3A492AB6B5DB3640ULL, "qun160_vashothlead6064676.cl"},
	{0x3A49573DCE068A81ULL, "llac_w_blightlands_0_11.dds"},
	{0x3A4990AE6FC40E55ULL, "llac_lt_gallowstemplar_2384_2.dds"},
	{0x3A49B6B14BC3033FULL, "560076706.ogg"},
	{0x3A49C392602E45CBULL, "vxlt_0_lowtown_docks_a3_cy9_0_1.vlm"},
	{0x3A4A0E05D8D184A1ULL, "vxht_playershouse_n_1na_0_0.vlm"},
	{0x3A4A1A703123049DULL, "24418596.ogg"},
	{0x3A4A33AB89AA88ECULL, "vxw_bonepitmining_a2_eu_0_0.vlm"},
	{0x3A4A801B5EC3FFB8ULL, "hlac_lt_0_lowtown_n_65537_6z.dds"},
	{0x3A4B25F1373B2D0BULL, "vxht_brothel_n_6fv_0_0.vlm"},
	{0x3A4B5AB45586EC32ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext.ani"},
	{0x3A4B905D228C8FF5ULL, "dae000_meredith6080998.cl"},
	{0x3A4BF550FD7DF344ULL, "vxlt_0_lowtown_docks_cof_0_0.vlm"},
	{0x3A4CA457C358A15AULL, "6135610_m.fxe"},
	{0x3A4D702DDD336D4BULL, "451559233.ogg"},
	{0x3A4DBD6FDBBB11F8ULL, "hlac_ht_0_hightown_1_2x.dds"},
	{0x3A4DD8D2CC97FAD1ULL, "llac_ht_keep_3736_3.dds"},
	{0x3A4E4430507FC2FDULL, "hlac_dr_0_deeproads_5849_5z.dds"},
	{0x3A4E92D97C6E1568ULL, "655111140.ogg"},
	{0x3A4E968F817F24FBULL, "zz_approval_debug6003750.cl"},
	{0x3A4EDAEFB4C53CFCULL, "6002005_f.fxe"},
	{0x3A4EE1DEC4BCBDE8ULL, "6189398_m.fxe"},
	{0x3A4F51E9C11C4936ULL, "vxw_bonepitmining_a2_1lo_0_0.vlm"},
	{0x3A4F918B034147FFULL, "vxca_caverns1_dpv_0_0.vlm"},
	{0x3A4FB572B522A8A4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_exit.ani"},
	{0x3A50B32612FBBF02ULL, "6105811_m.fxe"},
	{0x3A50E37372357030ULL, "vxw_sundermount_2gf_0_0.vlm"},
	{0x3A5140DAD809424BULL, "19658822.ogg"},
	{0x3A51C078E974882AULL, "art\\characters\\creatures\\dragon\\animation\\crnf.gad"},
	{0x3A520D1ADABA7940ULL, "232447772.ogg"},
	{0x3A52160FF4C4DB7AULL, "vxca_caverns1_jfc_0_0.vlm"},
	{0x3A525D44D2B50111ULL, "632019648.ogg"},
	{0x3A5299F9536D47E5ULL, "warrior_3.cut"},
	{0x3A5305AEE2F77B8EULL, "vxlt_0_lowtown_a2_n_c_huy_0_0.vlm"},
	{0x3A534CF2DB08686FULL, "255722240.ogg"},
	{0x3A53628056EE9CCAULL, "llac_lt_0_lowtown_a2_-3_1.dds"},
	{0x3A539781975642B9ULL, "194212785.ogg"},
	{0x3A53CAA5EEEB0F55ULL, "mer341_jansen6082984.cl"},
	{0x3A53E6EA62939018ULL, "vxlt_0_lowtown_a2_n_era_0_0.vlm"},
	{0x3A54234D16706F9AULL, "vxlt_0_lowtown_docks_c4u_0_0.vlm"},
	{0x3A5458C556FF6285ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_aoe_imp_p.mmh"},
	{0x3A5470A468A1DB07ULL, "character_character_p185.psh.fxo.dx11"},
	{0x3A547B48169605A7ULL, "hlac_ht_mansion2b_818_12x.dds"},
	{0x3A54CFD79F7E211FULL, "llac_lt_0_lowtown_docks_a3_n_-131073_3.dds"},
	{0x3A550554150CD01EULL, "6106270_m.fxe"},
	{0x3A5531A83AC95274ULL, "vxlt_warehouse_n_8f_0_0.vlm"},
	{0x3A55B272B5270F06ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_exit.gad"},
	{0x3A55B2A8407A68F3ULL, "vxlt_darktown_5ke_0_0.vlm"},
	{0x3A55CC118AC2D905ULL, "mag370_meredith6046887.cl"},
	{0x3A5628C5643FE06BULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb107a_0n.dds"},
	{0x3A567580300FDB3BULL, "vxlt_undercityencounter_xq_0_0.vlm"},
	{0x3A56A2FB8DA54D0FULL, "vxlt_darktown_2yt_0_0.vlm"},
	{0x3A56A63FCC164164ULL, "6144516_f.fxe"},
	{0x3A570209B91DD93EULL, "hlac_lt_0_lowtown_docks_65535_4z.dds"},
	{0x3A570C87A039A02FULL, "vxlt_0_lowtown_docks_a3_bss_0_0.vlm"},
	{0x3A573541AFA5028DULL, "zz_fen_debug6134328.cl"},
	{0x3A575BCFC6AEBAE8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1a.phy"},
	{0x3A57D1702CE419AEULL, "mcr371ip_sandal_store.utp"},
	{0x3A57D9B5CCCC0C3CULL, "vxlt_0_lowtown_docks_cp7_0_0.vlm"},
	{0x3A57E041076E7DC0ULL, "490123256.ogg"},
	{0x3A585556BEA3BE0EULL, "vxlt_hangedman_n_20h_0_0.vlm"},
	{0x3A587162CD8E138DULL, "mer220_hubert6065081.cl"},
	{0x3A587970E37C72D7ULL, "llac_lt_0_lowtown_docks_a3_n_-65538_1.dds"},
	{0x3A589641922D2344ULL, "1023167530.ogg"},
	{0x3A58A378E97A75C1ULL, "art\\characters\\creatures\\dragon\\animation\\crnf.evt"},
	{0x3A58E57F8F00740AULL, "6101697_m.fxe"},
	{0x3A58ECF818299A7FULL, "qcr221_cut_qunari_withdraw6042771.cl"},
	{0x3A590C239881EB91ULL, "vxw_woundedcoast_1dm_0_0.vlm"},
	{0x3A59CED9248A4593ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_avea_0d.dds"},
	{0x3A5A8FB39DA9AF5BULL, "6096201_f.fxe"},
	{0x3A5A9E2226289F0AULL, "lgt220_mag_apost_apost36170143.cl"},
	{0x3A5AB13390A81E0FULL, "m_qun210ar_fq_random_enc.bnk"},
	{0x3A5ABDC9F054F779ULL, "6034657_m.fxe"},
	{0x3A5ACB63FDCE0F44ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.cess.ani"},
	{0x3A5B2C432EA2C4E0ULL, "208827935.ogg"},
	{0x3A5B4C6646CA330FULL, "qun120_arishok6051327.cl"},
	{0x3A5B6D95D8F8F936ULL, "345673314.ogg"},
	{0x3A5C0692391436E6ULL, "546103111.ogg"},
	{0x3A5C2D194794567CULL, "cd_mag311_keran.wbd"},
	{0x3A5CAD72B52D25E5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_exit.evt"},
	{0x3A5D650ECC372440ULL, "6131909_f.fxe"},
	{0x3A5DD6E8DB30906FULL, "ser121_jethann6107105.cl"},
	{0x3A5DE086FECA8583ULL, "hlac_lt_0_lowtown_a2_n_c_131071_0x.dds"},
	{0x3A5E161DF6D5FA09ULL, "6142040_m.fxe"},
	{0x3A5ED6CE6ECB5AF8ULL, "vxca_caverns1_mines_jfh_0_0.vlm"},
	{0x3A5F3775EA553D17ULL, "6043679_m.fxe"},
	{0x3A5F466C46430E46ULL, "74350926.ogg"},
	{0x3A5F680FB787E200ULL, "788772278.ogg"},
	{0x3A5F973C9B900B2FULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.dura.gad"},
	{0x3A5FA7E5091978AEULL, "llac_lt_0_lowtown_a2_65534_6.dds"},
	{0x3A606815B8C12176ULL, "lowtown_male_lowlife1_sta6164967.cl"},
	{0x3A60DD3D391DE22CULL, "379487859.ogg"},
	{0x3A613247EBA7C9D7ULL, "vxlt_gallowscourtyard_n_6q1_0_0.vlm"},
	{0x3A6145314652080BULL, "hlac_lt_0_lowtown_docks_-196610_5x.dds"},
	{0x3A6155C55709E852ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_aoe_imp_p.anb"},
	{0x3A61778BD4ED940EULL, "vxlt_0_lowtown_a2_n_c_hpz_0_0.vlm"},
	{0x3A619A9DFF4535C9ULL, "llac_w_encounter3_-131073_0.dds"},
	{0x3A61C863FDD428A6ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.cess.gad"},
	{0x3A61FCEE46B1AD63ULL, "vxht_brothel_n_4fl_0_0.vlm"},
	{0x3A62952A0866881AULL, "vxlt_0_lowtown_docks_c1v_0_0.vlm"},
	{0x3A62A87E3CE2C5A3ULL, "vxht_mansion2b_1pn_0_1.vlm"},
	{0x3A630B8F26CD3610ULL, "qcr221cr_qunari_archer.utc"},
	{0x3A63CA4D8E5F1C57ULL, "hlac_w_bonepitdestroyed_a3_262146_0x.dds"},
	{0x3A643BCD3BD2A2EEULL, "character_rendertodepth_p2.psh.fxo.dx11"},
	{0x3A645F66D04BC0EEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_rf.evt"},
	{0x3A64839B2AFE2CA9ULL, "vxw_woundedcoast_hp_0_0.vlm"},
	{0x3A64B5D8700EF044ULL, "6101227_f.fxe"},
	{0x3A651A93478E43B1ULL, "88685770.ogg"},
	{0x3A6556513BEBDD63ULL, "zz_approval_debug6003563.cl"},
	{0x3A659B78E984F210ULL, "art\\characters\\creatures\\dragon\\animation\\crnf.ani"},
	{0x3A65E0CBED01FFFEULL, "708884874.ogg"},
	{0x3A65FE42C690333BULL, "mag321_comtesse_de_launcet6043367.cl"},
	{0x3A66707CD075B7D8ULL, "273228912.ogg"},
	{0x3A66AC3C9B964DDCULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.dura.evt"},
	{0x3A66D2DDDA6C3FB2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_enter.gad"},
	{0x3A67008985A18EC3ULL, "gen00fl_fenris6106624.cl"},
	{0x3A6717656D96B9ECULL, "vxw_bonepitmining_a2_ok_0_0.vlm"},
	{0x3A6728127C12AA4EULL, "vxlt_darktown_2v8_0_0.vlm"},
	{0x3A677E008A780411ULL, "vxw_woundedcoast_1ol_0_0.vlm"},
	{0x3A67C2395AE3EF94ULL, "llac_lt_0_lowtown_-65537_0.dds"},
	{0x3A6815A41FD09189ULL, "vxca_caverns2prc_9j1_0_0.vlm"},
	{0x3A6828613D34EF72ULL, "character_character_p25.psh.fxo.dx11.assembly"},
	{0x3A683636F8F3B32FULL, "lowtown_male_lowlife2_mob6169193.cl"},
	{0x3A6889C58898E7A8ULL, "vxlt_warehouse_n_wc_0_0.vlm"},
	{0x3A68C363FDDA3E85ULL, "art\\vfx\\combat\\vfx_meredith_powerup_c.cess.evt"},
	{0x3A6A5666D0501C99ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_rf.gad"},
	{0x3A6A6D6A0EB6F18DULL, "hlac_lt_0_lowtown_n_131070_3x.dds"},
	{0x3A6C220B420D4926ULL, "32023516.ogg"},
	{0x3A6C3EC58BE72163ULL, "235834635.ogg"},
	{0x3A6C741966300349ULL, "vxdr_0_deeproads_eq9_0_0.vlm"},
	{0x3A6C883D62DC2B1BULL, "6102767_m.fxe"},
	{0x3A6CEB7D63F675CAULL, "llac_lt_0_lowtown_docks_131073_1.dds"},
	{0x3A6D352D3E2F8784ULL, "vxlt_backalley_g_1fk_0_0.vlm"},
	{0x3A6DC72D5621A130ULL, "vxw_sundermount_2md_0_0.vlm"},
	{0x3A6DD5DDDA7263E9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_enter.evt"},
	{0x3A6DDD2CF0359F1CULL, "165219477.ogg"},
	{0x3A6DE41308B81A13ULL, "m_dae370ar_gallows.wbd"},
	{0x3A6E04111DA56F7CULL, "6137001_m.fxe"},
	{0x3A6E281F01270B5BULL, "6173662_m.fxe"},
	{0x3A6E460AE978527DULL, "ht_keep_n_4023.rml"},
	{0x3A6E6632A0575AD0ULL, "vxht_mansion2_1w5_0_0.vlm"},
	{0x3A6E835BE67F508CULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.impa.ani"},
	{0x3A6EFB5183027652ULL, "6016755_m.fxe"},
	{0x3A6F2D60D7DCD0C7ULL, "gen00fl_merrill6162132.cl"},
	{0x3A6F316AB4FB5B2EULL, "82477501.ogg"},
	{0x3A6F3B34967B40CAULL, "6095803_f.fxe"},
	{0x3A6F87561EFF28F8ULL, "fxe_watervol.mao"},
	{0x3A701E9B7FFF8147ULL, "vrc321_party_barks6095766.cl"},
	{0x3A70998FAF4A5AB0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand.phy"},
	{0x3A70AA9AA698ACDCULL, "vxw_blightlands_cf_0_0.vlm"},
	{0x3A70C1C6B04A2A77ULL, "647057521.ogg"},
	{0x3A715EEEDC77E001ULL, "qun211cr_elf_warrior.stg"},
	{0x3A716766D05658E7ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_rf.ani"},
	{0x3A7170AC7202A200ULL, "seb100im_flint_chest.uti"},
	{0x3A721C1C8A40F4B4ULL, "ash_leave.mao"},
	{0x3A7277CEF4C0687AULL, "633349167.ogg"},
	{0x3A72A3A5790E4654ULL, "fx_dirt_spec.dds"},
	{0x3A72D96D5FB916FBULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal.mmh"},
	{0x3A72E11D5863668CULL, "mrl210_merrill6116396.cl"},
	{0x3A72E56D5FB92A98ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal.mao"},
	{0x3A72F76D5FB949E5ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal.msh"},
	{0x3A73D2FD96B3CEEBULL, "vxlt_darktown_4vo_0_0.vlm"},
	{0x3A73E90EA75C427DULL, "1007202454.ogg"},
	{0x3A748C61C63BAEF9ULL, "vxlt_0_lowtown_docks_a3_d73_0_0.vlm"},
	{0x3A74B43C9BA298F5ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.dura.ani"},
	{0x3A74F4D63298C397ULL, "lt_0_lowtown_a2_n_42_0.anb"},
	{0x3A751923F435CA88ULL, "6201395_m.fxe"},
	{0x3A7589425F8CD26CULL, "vxw_bonepitmining_a2_jh_0_0.vlm"},
	{0x3A75905BE68585CEULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.impa.gad"},
	{0x3A75F13815B886DEULL, "6083353_m.fxe"},
	{0x3A75FB83724FFEC4ULL, "vxht_0_hightown_n_asv_0_0.vlm"},
	{0x3A7607FF119BFC71ULL, "341084656.ogg"},
	{0x3A7689767F29E9F5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas2_sa.dds"},
	{0x3A769F6D65B80AA7ULL, "zz_lgtpt_1_debug.ncs"},
	{0x3A76DB18354CC6E4ULL, "vxdr_primevaltomb_866_0_0.vlm"},
	{0x3A7717D1391D6E49ULL, "vxlt_warehouse_11t_0_0.vlm"},
	{0x3A7718B8BE3877E7ULL, "6126822_m.fxe"},
	{0x3A77E07BF54FF361ULL, "6131323_m.fxe"},
	{0x3A77E0BC609EA97AULL, "6042938_m.fxe"},
	{0x3A77E2A99EBA7CE9ULL, "hlac_lt_gallowscourtyard_-65537_2z.dds"},
	{0x3A780B2CFA77D013ULL, "hlac_lt_0_lowtown_n_196607_0z.dds"},
	{0x3A7810CAE57729A0ULL, "6106621_m.fxe"},
	{0x3A78352349A68EEAULL, "root_leave.mao"},
	{0x3A783FC3F17F5D6CULL, "hlac_lt_0_lowtown_a3_n_c_-131072_6x.dds"},
	{0x3A786AEEDC7E131BULL, "qun211cr_elf_warrior.utc"},
	{0x3A793652A03DC6F1ULL, "vxlt_0_lowtown_a2_n_c_e8d_0_0.vlm"},
	{0x3A79E16D137F7D51ULL, "vxht_0_hightown_n_7lx_0_0.vlm"},
	{0x3A79EE50B6E5A786ULL, "354038593.ogg"},
	{0x3A7A12FAA9A714EBULL, "6083690_f.fxe"},
	{0x3A7A322DB47B3BC1ULL, "293525863.ogg"},
	{0x3A7A3811DF108C6AULL, "vxca_caverns1_mines_eia_0_1.vlm"},
	{0x3A7ABDDDDA7CC558ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_enter.ani"},
	{0x3A7AE35AB6FE2A1EULL, "ser221_party_comments6166190.cl"},
	{0x3A7AF37A3F137C20ULL, "vxca_caverns1_ec5_0_0.vlm"},
	{0x3A7B28B4C8945EFAULL, "art\\levels\\races\\vfx_models\\prp_lightbeamrec.msh"},
	{0x3A7B36B4C8947633ULL, "art\\levels\\races\\vfx_models\\prp_lightbeamrec.mao"},
	{0x3A7B3AB4C8947D40ULL, "art\\levels\\races\\vfx_models\\prp_lightbeamrec.mmh"},
	{0x3A7B62B2C2B88182ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsfloor.msh"},
	{0x3A7B64B2C2B88538ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsfloor.mmh"},
	{0x3A7B70B2C2B899DBULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsfloor.mao"},
	{0x3A7C390A4817B9B0ULL, "vxw_sundermount_2ve_0_0.vlm"},
	{0x3A7C8B5BE68B9BADULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.impa.evt"},
	{0x3A7CC45C78C261A9ULL, "245323358.ogg"},
	{0x3A7CD80F924D7350ULL, "vxht_mansion2_1z2_0_0.vlm"},
	{0x3A7CF10D29183CE8ULL, "6165683_f.fxe"},
	{0x3A7D0DB3F9396AC5ULL, "vxca_caverns1_jfh_0_0.vlm"},
	{0x3A7D2756EBC4677EULL, "vxlt_gallowsprison_n_6gt_0_0.vlm"},
	{0x3A7DD89182A22629ULL, "hf_taranah.mor"},
	{0x3A7EAA042E73578EULL, "llac_lt_0_lowtown_a2_n_c_-131072_3.dds"},
	{0x3A7EBD841FA29F3CULL, "leftarrow.dds"},
	{0x3A7F1C77988EC55CULL, "vxw_blightlands_dg_0_0.vlm"},
	{0x3A7FAC6C1F6C6FC6ULL, "hlac_ht_mansion_n_3367_0y.dds"},
	{0x3A7FE4755EE8BE26ULL, "zz_followers_banter6157392.cl"},
	{0x3A80478DA04433EEULL, "vxca_caverns1_ck9_0_0.vlm"},
	{0x3A80B0120F4D24CAULL, "781284525.ogg"},
	{0x3A8167550EEA5DA9ULL, "vxw_woundedcoast_bv_0_0.vlm"},
	{0x3A81C5D59070FFC6ULL, "blood_conj_07_l2.mao"},
	{0x3A81CCE497A21622ULL, "vxca_caverns2_804_0_0.vlm"},
	{0x3A81CD07DBA2BDD7ULL, "w_encounter3_37_0.mmh"},
	{0x3A81D107DBA2C4E4ULL, "w_encounter3_37_0.mao"},
	{0x3A81E307DBA2E349ULL, "w_encounter3_37_0.msh"},
	{0x3A821757163CFDFCULL, "hlac_lt_0_lowtown_docks_n_-262144_1x.dds"},
	{0x3A8232E47C8BE3E4ULL, "ts_slavers_acc_table.gda"},
	{0x3A8240532313492FULL, "hlac_ht_mansion_n_3382_0y.dds"},
	{0x3A828ED7746D0805ULL, "llac_ca_caverns1_24423_12.dds"},
	{0x3A828F827788306CULL, "hlac_ht_0_hightown_a2_0_2y.dds"},
	{0x3A82CD753899BFAEULL, "400361886.ogg"},
	{0x3A83D579A0A9E6A1ULL, "vxht_playershouse_n_1re_0_0.vlm"},
	{0x3A83E488F5361DEAULL, "448506344.ogg"},
	{0x3A8428261A6FF5C7ULL, "hlac_lt_0_lowtown_docks_131070_0y.dds"},
	{0x3A84EA296CED617FULL, "vxw_bonepit_2ap_0_0.vlm"},
	{0x3A8512D38F298CD8ULL, "vxdr_primevaltomb_3ix_0_0.vlm"},
	{0x3A852336641636D1ULL, "vxdr_0_deeproads_cka_0_0.vlm"},
	{0x3A8532C7615F214AULL, "follower_banter6151356.cl"},
	{0x3A8575343E23C81BULL, "hlac_ht_mansion2b_818_65z.dds"},
	{0x3A85A65563A207A1ULL, "546485926.ogg"},
	{0x3A86C61CC6EE778EULL, "hlac_lt_0_lowtown_a3_n_c_65533_3x.dds"},
	{0x3A86F1BD320AE3D5ULL, "gen_im_trs_rdr_med_glv_r2.uti"},
	{0x3A876EEE64F55F2EULL, "hlac_lt_0_lowtown_docks_n_-2_1y.dds"},
	{0x3A889904C39D06AEULL, "357570632.ogg"},
	{0x3A88A665AE4B6F39ULL, "1064085636.ogg"},
	{0x3A89280D1051EB8BULL, "vxw_sundermount_27y_0_0.vlm"},
	{0x3A892B1947BB40F6ULL, "cd_mag311_keran.bnk"},
	{0x3A894D896706A30AULL, "hlac_dr_primevalentrance_1725_1z.dds"},
	{0x3A8965F654A0520CULL, "18114128.ogg"},
	{0x3A897A655F56E5EEULL, "vxca_cavern1_haunted_huy_0_0.vlm"},
	{0x3A89C3EB1A814697ULL, "mrl150_merrill6070984.cl"},
	{0x3A8A17370CBBCC6AULL, "vxlt_hangedman_2xn_0_1.vlm"},
	{0x3A8A8A2B043939B0ULL, "537209433.ogg"},
	{0x3A8A902CF83C25C2ULL, "vxlt_gallowscourtyard_7ac_0_1.vlm"},
	{0x3A8AAAA57922E382ULL, "fx_dirt_spec.mao"},
	{0x3A8B11F29003D60BULL, "hm_earl.mor"},
	{0x3A8BFB8E70CB33F2ULL, "303135056.ogg"},
	{0x3A8C20A6401F28C6ULL, "6092070_m.fxe"},
	{0x3A8CB44E54FA54B9ULL, "vxlt_0_lowtown_a2_n_c_eoi_0_0.vlm"},
	{0x3A8D2C0640FEC3AFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curver_1.mao"},
	{0x3A8D2E5FEB330C85ULL, "vxca_caverns1_mines_jfq_0_0.vlm"},
	{0x3A8D6CC364DF51E8ULL, "1186934.ogg"},
	{0x3A8D8E548A84DDDCULL, "vxw_blightlands_id_0_0.vlm"},
	{0x3A8E1C1F6F0AA3C8ULL, "dae000_sandal6147817.cl"},
	{0x3A8E7AD44249DDB0ULL, "follower_banter6155655.cl"},
	{0x3A8EE97565DEE11DULL, "72308405.ogg"},
	{0x3A8EF91277493B49ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scraplight_n.dds"},
	{0x3A8F25F63925F7F2ULL, "ave100_donnic6106082.cl"},
	{0x3A8F5C02004D30C7ULL, "489852760.ogg"},
	{0x3A8F84F953A0D263ULL, "warrior_3.utc"},
	{0x3A8FBEDC65A0A249ULL, "835616574.ogg"},
	{0x3A8FF420336D81D1ULL, "vxht_chantry_2cc_0_0.vlm"},
	{0x3A904534C714225CULL, "vxlt_0_lowtown_a3_n_c_by4_0_0.vlm"},
	{0x3A906604359E39D9ULL, "hlac_ht_mansion2b_818_9x.dds"},
	{0x3A90CAB8C63205E0ULL, "vxlt_gallowstemplar_n_4qr_0_0.vlm"},
	{0x3A91C7F7CB00ABB6ULL, "997354461.ogg"},
	{0x3A91CF09720ADC3FULL, "vxw_bonepitmining_a2_od_0_0.vlm"},
	{0x3A91E416A935208BULL, "static_static_p4.psh.fxo"},
	{0x3A9222001EC9AE6EULL, "hlac_ca_cavern1_haunted_24427_3z.dds"},
	{0x3A9222D4D9198CB1ULL, "348246464.ogg"},
	{0x3A92C468B171D4A2ULL, "127806192.ogg"},
	{0x3A92D397605A4D91ULL, "vxw_woundedcoast_1pa_0_0.vlm"},
	{0x3A93E21AB52D8AEAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\set_lowtowndoor_d.dds"},
	{0x3A942D847B57A55DULL, "6112144_m.fxe"},
	{0x3A947C4B90AE9190ULL, "1025863496.ogg"},
	{0x3A9489B09BC3E635ULL, "131806972.ogg"},
	{0x3A95647F0AD49AD4ULL, "dm_denier.mor"},
	{0x3A95738A94C2BDBEULL, "6165743_m.fxe"},
	{0x3A95C77395D8BA32ULL, "vxw_encounter2_camp_55_0_0.vlm"},
	{0x3A95CF892E83C539ULL, "hlac_w_encounter2_-65534_0y.dds"},
	{0x3A965DE866B72922ULL, "274667321.ogg"},
	{0x3A965F2E7830705DULL, "zz_dae_debug6029829.cl"},
	{0x3A966AD358C8367EULL, "vxht_mansion_n_4xt_0_0.vlm"},
	{0x3A974ACA8581457DULL, "lgt260_eau_tmp_lyrium_mess.bnk"},
	{0x3A975EE26A515C26ULL, "182514911.ogg"},
	{0x3A97638CE04B1CEFULL, "follower_banter6163574.cl"},
	{0x3A9799EA0248040BULL, "vxw_sundermount_28x_0_0.vlm"},
	{0x3A97D73422FFAD20ULL, "vxca_caverns1_ei3_0_0.vlm"},
	{0x3A98108F32B33B69ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopeb.mao"},
	{0x3A981C8F32B34F02ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopeb.mmh"},
	{0x3A981E8F32B352B8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopeb.msh"},
	{0x3A985BF8EA1FEB18ULL, "ser120_ghyslain6108842.cl"},
	{0x3A98A0D8B59527BBULL, "hlac_ca_caverns1_24423_9x.dds"},
	{0x3A98AE28753D1CEAULL, "6047717_m.fxe"},
	{0x3A9903AE7BF426EFULL, "mer341_jansen6082999.cl"},
	{0x3A9987F97FB6A131ULL, "vxw_sundermount_354_0_0.vlm"},
	{0x3A99B9DBF74EF74AULL, "6113922_m.fxe"},
	{0x3A9A0B10CFB0987EULL, "vxlt_gallowsprison_n_6mv_0_0.vlm"},
	{0x3A9A511725038BB2ULL, "396491965.ogg"},
	{0x3A9A871A0069EA4FULL, "6132896_m.fxe"},
	{0x3A9ABABFB6CEB0CDULL, "gen00fl_anders6111170.cl"},
	{0x3A9AF46D644CE658ULL, "6067747_f.fxe"},
	{0x3A9B17F684992FE2ULL, "dae000_arishok6187907.cl"},
	{0x3A9B2CCA8584FA85ULL, "lgt260_eau_tmp_lyrium_mess.cnv"},
	{0x3A9B7ABF0A187A24ULL, "hlac_ht_mansion_n_3364_0z.dds"},
	{0x3A9B804AAFA19368ULL, "699784997.ogg"},
	{0x3A9BB9B882826163ULL, "861005641.ogg"},
	{0x3A9C00317C7AF65CULL, "vxw_blightlands_re_0_0.vlm"},
	{0x3A9CAD033B959162ULL, "vxca_caverns1_cjz_0_0.vlm"},
	{0x3A9D517918E16C09ULL, "50844066.ogg"},
	{0x3A9D6C239FBA86E8ULL, "6142118_f.fxe"},
	{0x3A9DAD228374D8D1ULL, "6177673_m.fxe"},
	{0x3A9DC87C6FB2B61CULL, "1000569821.ogg"},
	{0x3A9E4B0EF2D68EA9ULL, "vxw_woundedcoast_tt_0_0.vlm"},
	{0x3A9EF27A48963492ULL, "vrc321ip_spooky7_obj_2.utp"},
	{0x3A9F58B8F8F73B58ULL, "vxht_brothel_n_4bw_0_0.vlm"},
	{0x3A9F6082FB35F8EDULL, "dae000cr_grand_cleric6164472.cl"},
	{0x3A9F7DE2E6431A02ULL, "hlac_lt_0_lowtown_docks_n_65534_2z.dds"},
	{0x3A9F893AE1C85B8DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medp_0s.dds"},
	{0x3A9FBD6DA9985C21ULL, "lt_0_lowtown_docks_28_0.anb"},
	{0x3A9FC758B485D30EULL, "mag110_arianni6016795.cl"},
	{0x3A9FE2DC423DF320ULL, "6101270_f.fxe"},
	{0x3AA03C0A7031F2C4ULL, "324228900.ogg"},
	{0x3AA040E66400F4BFULL, "vxw_bonepitmining_a2_jg_0_0.vlm"},
	{0x3AA0F1D632BDF9ABULL, "lt_0_lowtown_a2_n_42_0.rml"},
	{0x3AA186874008C681ULL, "vxw_blightlands_d6_0_0.vlm"},
	{0x3AA23A3C0BCFA81CULL, "rdr110dg_guard_warehouse6016693.cl"},
	{0x3AA247BF76797ABEULL, "zz_otl_debug6060594.cl"},
	{0x3AA249EFDADB2C64ULL, "vxht_keep_5zv_0_0.vlm"},
	{0x3AA2BC5932BCD21EULL, "ser270_emeric6113499.cl"},
	{0x3AA43622F724E181ULL, "llac_ht_0_hightown_n_-196608_0.dds"},
	{0x3AA462D77A653DBDULL, "95580521.ogg"},
	{0x3AA51D5AB93AB9D0ULL, "671466330.ogg"},
	{0x3AA57567ACC0B706ULL, "vxlt_gallowsprison_n_6ng_0_0.vlm"},
	{0x3AA579E7C84DDB6AULL, "543860626.ogg"},
	{0x3AA5D3488BD077BCULL, "transparent.dds"},
	{0x3AA6A8ECB555E4B7ULL, "dot_blur_a.dds"},
	{0x3AA6D44D1D6B1DFFULL, "one120_athenril6088301.cl"},
	{0x3AA6F9F6D2DC167DULL, "vxht_keep_4hx_0_0.vlm"},
	{0x3AA888B5ABC67330ULL, "947256582.ogg"},
	{0x3AA8FA9E56E0DE6DULL, "zz_approval_debug6003770.cl"},
	{0x3AA905848397118AULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgte_0s.dds"},
	{0x3AAA9548D5E91BFCULL, "vxht_keep_n_49a_0_0.vlm"},
	{0x3AAAC852219F0375ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_qu101a_0t.dds"},
	{0x3AAAD18239D9F99DULL, "428087415.ogg"},
	{0x3AAB3AD3DDC0A274ULL, "llac_lt_hangedman_n_-1_0.dds"},
	{0x3AAB657BC6066173ULL, "6176672_m.fxe"},
	{0x3AAB66AA830D7EC7ULL, "vxht_keep_n_6hw_0_0.vlm"},
	{0x3AAB7F3E25B82CA4ULL, "hlac_lt_backalley_g_1_6y.dds"},
	{0x3AAB8428600E3F95ULL, "vxht_mansion2b_2qe_0_0.vlm"},
	{0x3AABD6A3E48ACE6DULL, "448868854.ogg"},
	{0x3AACD50CF8C6F549ULL, "6169414_m.fxe"},
	{0x3AACD7DA1759B2D1ULL, "vxht_chantry_2ya_0_0.vlm"},
	{0x3AADA5734F220BF8ULL, "vxlt_gallowstemplar_n_4o4_0_0.vlm"},
	{0x3AADBC09AB461776ULL, "gen00fl_merrill6106960.cl"},
	{0x3AADCA95EAED53D8ULL, "vxht_brothel_n_4gt_0_0.vlm"},
	{0x3AAE33860F3AD09CULL, "629939521.ogg"},
	{0x3AAE5F9FA636C93AULL, "6137071_f.fxe"},
	{0x3AAF02DC99882139ULL, "6180158_m.fxe"},
	{0x3AAF7E7D44879E2CULL, "plt_dae000pt_amb_actions.plo"},
	{0x3AAF9937F8BD953BULL, "vxht_0_hightown_a2_avm_0_0.vlm"},
	{0x3AAFDA17B96EACA9ULL, "vxw_bonepit_1ym_0_0.vlm"},
	{0x3AAFF86431FEDF01ULL, "vxw_blightlands_i5_0_0.vlm"},
	{0x3AB0046E7818CB22ULL, "art\\characters\\creatures\\spider\\textures\\c_giaspidra_0s.dds"},
	{0x3AB01AE2F872ADC5ULL, "mag210_arianni6021245.cl"},
	{0x3AB0470727CE6D86ULL, "vxht_mansion_n_4zf_0_0.vlm"},
	{0x3AB063F9E513D031ULL, "llac_lt_0_lowtown_docks_a3_-131071_0.dds"},
	{0x3AB06EDC35CF87EBULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml104a_0n.dds"},
	{0x3AB0F19E552EF373ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.cess.evt"},
	{0x3AB1C657E395FA34ULL, "vxlt_blackemporium_vi_0_0.vlm"},
	{0x3AB27BBB2DD22221ULL, "gen00fl_sebastian6195863.cl"},
	{0x3AB298719490255AULL, "fire.mao"},
	{0x3AB2AF446E8E2381ULL, "vxlt_0_lowtown_a3_n_c_bye_0_0.vlm"},
	{0x3AB2B9F11373D24EULL, "lowtown_male_urchin_sta6161586.cl"},
	{0x3AB2BF1F305925A1ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsa.phy"},
	{0x3AB2C0D961D30667ULL, "vxw_woundedcoast_17r_0_0.vlm"},
	{0x3AB2D22496070E12ULL, "593157440.ogg"},
	{0x3AB2EB40014FF8CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmenter.evt"},
	{0x3AB327CADD86D756ULL, "narrative_varric6172518.cl"},
	{0x3AB353FCFE604052ULL, "6130988_f.fxe"},
	{0x3AB3B74316BCF231ULL, "245055394.ogg"},
	{0x3AB3F57273CBD5A1ULL, "442408542.ogg"},
	{0x3AB40BCE1CA3984AULL, "6029025_m.fxe"},
	{0x3AB4204C80E19309ULL, "6105683_m.fxe"},
	{0x3AB49BE2490CED81ULL, "vfx_dust_01.dds"},
	{0x3AB51272629CF304ULL, "107833677.ogg"},
	{0x3AB533F2CD69BF05ULL, "dae000_gamlen6048431.cl"},
	{0x3AB6637D448D8F63ULL, "plt_dae000pt_amb_actions.nss"},
	{0x3AB6D995850B7B2FULL, "6144435_m.fxe"},
	{0x3AB6E266FB1FCB35ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chainlong.msh"},
	{0x3AB6E466FB1FCEABULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chainlong.mmh"},
	{0x3AB6F066FB1FE348ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chainlong.mao"},
	{0x3AB711B5FD36A4CDULL, "6167157_m.fxe"},
	{0x3AB777B2DED5B447ULL, "vxlt_0_lowtown_a2_emx_0_0.vlm"},
	{0x3AB792BC96A5CEA4ULL, "vxht_0_hightown_7iy_0_0.vlm"},
	{0x3AB7BE2F80D9F2CBULL, "vxlt_0_lowtown_n_e22_0_0.vlm"},
	{0x3AB7BF004D7AA7B6ULL, "dae000_flemeth6000930.cl"},
	{0x3AB7EA9E553505B4ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.cess.gad"},
	{0x3AB84A0381B98E2BULL, "6044937_m.fxe"},
	{0x3AB8B4C3EA6F79CBULL, "w_sundermount_77_0_lc.mmh"},
	{0x3AB8E11A0CFCA907ULL, "vxlt_0_lowtown_docks_a3_d2b_0_0.vlm"},
	{0x3AB955AE7E9863FCULL, "6118636_f.fxe"},
	{0x3AB9F60552045815ULL, "vxht_mansion2b_2jd_0_0.vlm"},
	{0x3ABA04400156418CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmenter.gad"},
	{0x3ABAF45DCAD6E1F9ULL, "96805438.ogg"},
	{0x3ABB0F6F1A048CCBULL, "hlac_lt_backalley_n_65537_6x.dds"},
	{0x3ABB1398A1256FB1ULL, "follower_banter6160946.cl"},
	{0x3ABB7543C6DAD432ULL, "vxca_caverns1_j27_0_0.vlm"},
	{0x3ABC0C4E5082496BULL, "vxlt_darktown_4ib_0_0.vlm"},
	{0x3ABC3A8A27C7AEB6ULL, "158212492.ogg"},
	{0x3ABC893686CA5D6CULL, "ico_potion_restore.dds"},
	{0x3ABD168583E4DFBCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_3stackxa.mao"},
	{0x3ABD228583E4F45FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_3stackxa.mmh"},
	{0x3ABD288583E4FE01ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_3stackxa.msh"},
	{0x3ABD992D41E5C356ULL, "6077989_m.fxe"},
	{0x3ABDF99E55398A92ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.cess.ani"},
	{0x3ABE07E5909EF924ULL, "zz_followers_banter6158553.cl"},
	{0x3ABE3FFF5821F24BULL, "135850234.ogg"},
	{0x3ABE6A4123F4F781ULL, "vxw_blightlands_r4_0_0.vlm"},
	{0x3ABEBDF59AB5AE20ULL, "vxw_bonepit_110_0_0.vlm"},
	{0x3ABEE5357BD9A3A4ULL, "vxdr_0_deeproads_458_0_1.vlm"},
	{0x3ABEF3A70F6A9800ULL, "56129737.ogg"},
	{0x3ABF76FD3C3BAE2DULL, "lt_0_lowtown_docks_a3_n_160_0.mmh"},
	{0x3ABF7AFD3C3BB57EULL, "lt_0_lowtown_docks_a3_n_160_0.mao"},
	{0x3ABF86883733B7BBULL, "hlac_lt_gallowstemplar_2384_4x.dds"},
	{0x3ABF88FD3C3BCCB3ULL, "lt_0_lowtown_docks_a3_n_160_0.msh"},
	{0x3ABFC2D443B35D87ULL, "541856318.ogg"},
	{0x3AC00340015AAB8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmenter.ani"},
	{0x3AC00A26B20CB3F2ULL, "jobboardplots.gda"},
	{0x3AC0535E294C2D13ULL, "vxca_caverns1_i1j_0_2.vlm"},
	{0x3AC054B670325DC3ULL, "6071010_m.fxe"},
	{0x3AC06CF5C9B7C2E9ULL, "190396236.ogg"},
	{0x3AC06F5504B267DEULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r90_rf.ani"},
	{0x3AC0853E95EB9204ULL, "mer210_smith6051401.cl"},
	{0x3AC0B5714F6DEECAULL, "rogue_cu.cut"},
	{0x3AC1A6CD5CEE69D6ULL, "zz_area_signage14.utp"},
	{0x3AC1EBC1964C47FFULL, "face0_s2y.psh.fxo"},
	{0x3AC2BDEAE0A31693ULL, "6060715_m.fxe"},
	{0x3AC2E0ABE3AA0DECULL, "art\\levels\\races\\proxy\\_textures\\default_white.dds"},
	{0x3AC32F16B3827A9FULL, "vxht_0_hightown_au8_0_0.vlm"},
	{0x3AC33126EEAF77BCULL, "vxlt_0_lowtown_a3_n_c_csg_0_0.vlm"},
	{0x3AC358B302CA6B9BULL, "si_expo.mao"},
	{0x3AC37977AFC2D0C9ULL, "hlac_lt_0_lowtown_docks_a3_-131075_0x.dds"},
	{0x3AC3E7096BA89B78ULL, "817377201.ogg"},
	{0x3AC416B0494FC9D5ULL, "vxht_0_hightown_a3_cai_0_0.vlm"},
	{0x3AC489372D0B3317ULL, "vxca_cave_19q_0_1.vlm"},
	{0x3AC4ADCA3A6879DEULL, "llac_ca_caverns2_5802_38.dds"},
	{0x3AC4D6D8DC94F349ULL, "hlac_lt_undercityencounter_-1_7x.dds"},
	{0x3AC547928EFA8028ULL, "gen00fl_fenris6115359.cl"},
	{0x3AC54B72770193AFULL, "6141735_m.fxe"},
	{0x3AC55442E463F720ULL, "341598651.ogg"},
	{0x3AC61C9CD23079EAULL, "gen00fl_fenris_base_act2_1.stg"},
	{0x3AC6C139EF62BB6EULL, "vxlt_0_lowtown_a3_n_c_hfa_0_0.vlm"},
	{0x3AC6EF2177D14449ULL, "hlac_lt_0_lowtown_docks_a3_65534_5z.dds"},
	{0x3AC7BF857F284F34ULL, "vxht_chantry_n_357_0_1.vlm"},
	{0x3AC8399A49DBA225ULL, "vxlt_0_lowtown_a2_n_c_goj_0_1.vlm"},
	{0x3AC84D21EC93B2FAULL, "6175711_m.fxe"},
	{0x3AC95FD1278E44ADULL, "vxht_mansion_n_47j_0_0.vlm"},
	{0x3AC99C62F2010E46ULL, "440228301.ogg"},
	{0x3AC9B7329426177CULL, "vxw_woundedcoast_fk_0_0.vlm"},
	{0x3AC9BD739D5A853FULL, "829497822.ogg"},
	{0x3AC9D1FA7F1456B9ULL, "mag360_samson6062124.cl"},
	{0x3ACA2BF3836209AAULL, "follower_banter6172984.cl"},
	{0x3ACA52B3F960CD96ULL, "gen_im_trs_qun_biscuit_02.uti"},
	{0x3ACA5C657B47C297ULL, "6076396_m.fxe"},
	{0x3ACBF0FCA8A4A437ULL, "6205169_m.fxe"},
	{0x3ACC7CF1DCA9C63BULL, "vxht_0_hightown_a2_alo_0_0.vlm"},
	{0x3ACD09CC97547132ULL, "7298842.ogg"},
	{0x3ACD875504BD1AAFULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r90_rf.evt"},
	{0x3ACDE047E219BC87ULL, "hlac_lt_0_lowtown_-1_3y.dds"},
	{0x3ACE1B128264C09FULL, "clf260tr_start_attack.utt"},
	{0x3ACF19B79C3331F7ULL, "6078174_m.fxe"},
	{0x3ACF61DA4D9463D5ULL, "dae010_arguing_female.utc"},
	{0x3ACF72E79801511BULL, "139751104.ogg"},
	{0x3ACFB36627A3A643ULL, "vxht_keep_n_61t_0_0.vlm"},
	{0x3ACFFEBD87199700ULL, "hlac_lt_0_lowtown_a2_n_-131072_6y.dds"},
	{0x3AD00CD519254625ULL, "vxdr_primevaltomb_7m1_0_0.vlm"},
	{0x3AD069AE74CE7374ULL, "gen00fl_merrill6130101.cl"},
	{0x3AD0717194A8E00CULL, "fire.dds"},
	{0x3AD0B2B617A4688BULL, "6072706_f.fxe"},
	{0x3AD1805F662F71B1ULL, "60592948.ogg"},
	{0x3AD23A5AE57FD420ULL, "582221611.ogg"},
	{0x3AD35050D7AE621FULL, "417435263.ogg"},
	{0x3AD360AB652B6989ULL, "mag160_feynriel6016580.cl"},
	{0x3AD387FDED958577ULL, "vxw_encounter2_e4_0_0.vlm"},
	{0x3AD396BEA0A8FADEULL, "237815893.ogg"},
	{0x3AD3AAFE6911FC58ULL, "gen00fl_varric6076081.cl"},
	{0x3AD3D9E9CE44E9B6ULL, "6099703_m.fxe"},
	{0x3AD3E226EC6BDEB8ULL, "llac_lt_0_lowtown_a3_n_c_-1_5.dds"},
	{0x3AD4178207C230CEULL, "534260778.ogg"},
	{0x3AD428622DE60C80ULL, "plt_cod_cha_merrill.plo"},
	{0x3AD434CB64BFFD7FULL, "art\\characters\\creatures\\hurlock\\textures\\cn_bdy_hlkb_0n.dds"},
	{0x3AD4885504C33B58ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r90_rf.gad"},
	{0x3AD4AD31FAC77E61ULL, "6116060_f.fxe"},
	{0x3AD69F2005EC4E98ULL, "272575877.ogg"},
	{0x3AD70CFFB3C3CA9FULL, "mag121_idunna6084871.cl"},
	{0x3AD774914CBC1E7FULL, "vxlt_0_lowtown_a2_n_c_gnj_0_0.vlm"},
	{0x3AD8511716844827ULL, "hlac_dr_slavers_3216_3x.dds"},
	{0x3AD881E36F1562C4ULL, "hlac_lt_0_lowtown_a2_n_-2_6y.dds"},
	{0x3AD894B6648D1D06ULL, "850274066.ogg"},
	{0x3AD93ACD8481F00DULL, "vxca_caverns1_mines_id8_0_0.vlm"},
	{0x3AD9499227BEF51AULL, "718605384.ogg"},
	{0x3AD9793A2557D774ULL, "llac_lt_gallowstemplar_f_2394_5.dds"},
	{0x3AD99E6DA9C99FB1ULL, "lt_0_lowtown_docks_28_0.rml"},
	{0x3AD9B21EE2575327ULL, "596540080.ogg"},
	{0x3ADA046C17166683ULL, "one140cr_charade6173702.cl"},
	{0x3ADA975688A2320CULL, "6120597_m.fxe"},
	{0x3ADACB06E167822CULL, "vxca_caverns1_mines_edq_0_0.vlm"},
	{0x3ADBB476C5D54B5CULL, "451234151.ogg"},
	{0x3ADBDB55FC47B1A6ULL, "vxca_cavern1_haunted_dqa_0_0.vlm"},
	{0x3ADBE54D3AE742C2ULL, "275489483.ogg"},
	{0x3ADC1005564F8ABCULL, "atl_fonttest_dxt5_02.dds"},
	{0x3ADC30523DCEB0B5ULL, "6118469_m.fxe"},
	{0x3ADC902AF1FC37D8ULL, "vxht_keep_n_5xo_0_0.vlm"},
	{0x3ADD87E984439C27ULL, "vxht_brothel_n_5kj_0_0.vlm"},
	{0x3ADE2299A76DFE4AULL, "vxlt_backalley_n_1wc_0_0.vlm"},
	{0x3ADE266318B06242ULL, "vxht_keep_n_8br_0_0.vlm"},
	{0x3ADE4747E04E882EULL, "429142505.ogg"},
	{0x3ADE6B968AA3AE79ULL, "135217042.ogg"},
	{0x3ADEC047D0AEB31CULL, "zz_ave_debug6064102.cl"},
	{0x3ADEC57F8F8EF6D5ULL, "gen_im_cft_rec_pot_07.uti"},
	{0x3ADEC7B0B3E211D4ULL, "6110015_m.fxe"},
	{0x3ADEF1E278A2DF9CULL, "llac_w_encounter3_-196606_0.dds"},
	{0x3ADF1FC8C6EB92C3ULL, "6066339_m.fxe"},
	{0x3ADF2920D2831FD5ULL, "6097356_f.fxe"},
	{0x3ADF3012827357BEULL, "clf260tr_start_attack.ncs"},
	{0x3ADF6B03492D6472ULL, "vxlt_gallowscourtyard_n_5gq_0_0.vlm"},
	{0x3ADFED56DD21D403ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_beta_s.dds"},
	{0x3AE00985686764E8ULL, "254122132.ogg"},
	{0x3AE02FF324336830ULL, "480026934.ogg"},
	{0x3AE047E062A5ECFCULL, "6025351_m.fxe"},
	{0x3AE04DB75CEC7802ULL, "character.mat.dx11"},
	{0x3AE100161AEA4CC6ULL, "dre000_dougal6129806.cl"},
	{0x3AE1771B15357A76ULL, "vxlt_0_lowtown_docks_a3_cz5_0_1.vlm"},
	{0x3AE19306EA4EF856ULL, "hlac_lt_gallowstemplar_2383_11x.dds"},
	{0x3AE1FB11FACA8340ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_018a_0s.dds"},
	{0x3AE2FD809DE7CDD7ULL, "6168983_m.fxe"},
	{0x3AE364A13B7E9B4DULL, "mer340_cowan_st.stg"},
	{0x3AE38E712C4C12C4ULL, "fen210_orana6109770.cl"},
	{0x3AE3EB91FD837CD4ULL, "vxlt_darktown_583_0_0.vlm"},
	{0x3AE3EEB3BC4055CFULL, "hlac_ht_brothel_n_3371_2z.dds"},
	{0x3AE40AFA63657886ULL, "824350930.ogg"},
	{0x3AE47812A46F64ACULL, "vxht_0_hightown_a2_bsv_0_0.vlm"},
	{0x3AE4C6861F8B28FAULL, "726240667.ogg"},
	{0x3AE52A31DC15818EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_craggyrock.phy"},
	{0x3AE606684547EFB8ULL, "llac_w_encounter2_-65532_0.dds"},
	{0x3AE6291C9EFA35F6ULL, "6208542_f.fxe"},
	{0x3AE67591EAE1720AULL, "159284663.ogg"},
	{0x3AE69AEC7812487CULL, "vxw_woundedcoast_xi_0_0.vlm"},
	{0x3AE69B8742770C63ULL, "vxdr_primevaltomb_34t_0_0.vlm"},
	{0x3AE6D94C4BCF2AA6ULL, "251401489.ogg"},
	{0x3AE6FD62581C9880ULL, "gen_im_trs_dal_shd_wdn_r1.uti"},
	{0x3AE708368AD2BB6FULL, "gen00fl_sebastian6180423.cl"},
	{0x3AE7B47F38CCCCEEULL, "dae000_flemeth6013264.cl"},
	{0x3AE7C7DD37BC1A28ULL, "6139177_m.fxe"},
	{0x3AE7F5BF2DDC7145ULL, "6152621_m.fxe"},
	{0x3AE8D7790749E3A2ULL, "734892678.ogg"},
	{0x3AE91C80B44EA773ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_surprised_back.ani"},
	{0x3AEA2A2923E563F5ULL, "1040987716.ogg"},
	{0x3AEA63A0956B936FULL, "10501019.ogg"},
	{0x3AEAA373022BBB1FULL, "w_sundermount_65_0.rml"},
	{0x3AEABF18581492B9ULL, "252207302.ogg"},
	{0x3AEADB425E252F36ULL, "6114230_f.fxe"},
	{0x3AEB0251518B5101ULL, "gen_im_trs_gen_med_cht_r2.uti"},
	{0x3AEB1139507B7EA0ULL, "468578671.ogg"},
	{0x3AEB47A6EB7A61AAULL, "6118747_m.fxe"},
	{0x3AEB9DA2D1BD7789ULL, "vxht_keep_n_3zr_0_0.vlm"},
	{0x3AEC0520B795DFB4ULL, "68799069.ogg"},
	{0x3AEC3ADC796FF7B6ULL, "6080165_m.fxe"},
	{0x3AEC533EB15F3B3CULL, "lt_0_lowtown_docks_n_231_0.mao"},
	{0x3AEC5F3EB15F4FDFULL, "lt_0_lowtown_docks_n_231_0.mmh"},
	{0x3AEC653EB15F5981ULL, "lt_0_lowtown_docks_n_231_0.msh"},
	{0x3AEC6BBC371D5493ULL, "ser120_ghyslain6165796.cl"},
	{0x3AEC7FEDFE5F75A2ULL, "cr_spe_glow.dds"},
	{0x3AECB9321A91467DULL, "hlac_w_sundermount_-196614_0x.dds"},
	{0x3AED09332941A3DEULL, "6083142_m.fxe"},
	{0x3AED1601048CC97EULL, "vxht_mansion_n_4fj_0_0.vlm"},
	{0x3AEDE2093339506BULL, "235977857.ogg"},
	{0x3AEE0A152BC69D8FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_meda_0t.dds"},
	{0x3AEE54B140CC8CF3ULL, "hlac_lt_0_lowtown_a2_1_10z.dds"},
	{0x3AEE617BB64C22DDULL, "835014822.ogg"},
	{0x3AEF200FC86838C1ULL, "vxdr_primevaltomb_4ir_0_0.vlm"},
	{0x3AEF284E0521C359ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masd.mmh"},
	{0x3AEF324E0521D447ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masd.msh"},
	{0x3AEF4E6E6031124AULL, "vxw_blightlands_144_0_0.vlm"},
	{0x3AEF51CD2479F611ULL, "hlac_lt_0_lowtown_docks_a3_-65536_0x.dds"},
	{0x3AEF89D17B869CEAULL, "6013334_m.fxe"},
	{0x3AEFA46105F3FA69ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd04a.phy"},
	{0x3AEFEFD9286B9335ULL, "gen00fl_fenris6110834.cl"},
	{0x3AF03B256241187BULL, "vxlt_gallowsprison_n_64p_0_0.vlm"},
	{0x3AF06BB7D181B677ULL, "vxw_encounter2_k6_0_0.vlm"},
	{0x3AF088889AE37502ULL, "one220_yevhen6119964.cl"},
	{0x3AF1685B0374314BULL, "vxht_keep_n_7aa_0_0.vlm"},
	{0x3AF1A9789679E792ULL, "mag320_evelina6051874.cl"},
	{0x3AF1AB72B4DA2A2EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_tarpile.msh"},
	{0x3AF1B972B4DA427FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_tarpile.mao"},
	{0x3AF1C572B4DA561CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_tarpile.mmh"},
	{0x3AF2BDE61234C939ULL, "167193738.ogg"},
	{0x3AF2C41CA88499EEULL, "6067773_m.fxe"},
	{0x3AF37BAD2368D67BULL, "46376447.ogg"},
	{0x3AF3A8373D9268CCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03a.msh"},
	{0x3AF3AAEC53B6FE05ULL, "seb201_lady_harimann6177410.cl"},
	{0x3AF3AE373D9272BEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03a.mmh"},
	{0x3AF3B2EE3A974CA3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_bnt_dblw.ani"},
	{0x3AF3DF29AFCEF3F2ULL, "plt_cod_cha_flemeth.plo"},
	{0x3AF3FE0066B62425ULL, "follower_banter6172808.cl"},
	{0x3AF405C0A6EC993BULL, "vfx0_volumetexture.psh.fxo"},
	{0x3AF41E1B9247EC99ULL, "vxdr_primevaltomb_351_0_0.vlm"},
	{0x3AF485765DE254E9ULL, "vxht_keep_n_8al_0_0.vlm"},
	{0x3AF4F4F3B246DF73ULL, "art\\vfx\\textures\\vfx_smoke_whispy_cs.dds"},
	{0x3AF50CC96A0860FCULL, "vxw_woundedcoast_un_0_0.vlm"},
	{0x3AF54FE88A8A7B02ULL, "lt_0_lowtown_docks_n_44_0.rml"},
	{0x3AF59BA472F40D01ULL, "6195857_m.fxe"},
	{0x3AF5F29C82F7CE72ULL, "vxca_caverns1_j25_0_2.vlm"},
	{0x3AF61480B45923AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_surprised_back.evt"},
	{0x3AF62FF2DB22234CULL, "arm_mer.tnt"},
	{0x3AF69C73B632048FULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgta_0d.dds"},
	{0x3AF6ED4417D516FDULL, "ht_0_hightown_a3_36_0_lc.mmh"},
	{0x3AF6F81BC143E6C1ULL, "lit_aoe_02_l2.mao"},
	{0x3AF70E8E33E11CCBULL, "zz_lgt_1_debug6114216.cl"},
	{0x3AF720D20086D990ULL, "vxht_brothel_n_501_0_1.vlm"},
	{0x3AF73195EBB10094ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.impa.ani"},
	{0x3AF7C4633B0D27DCULL, "6045854_m.fxe"},
	{0x3AF81D23F48BCF92ULL, "gen_im_trs_bst_pelt_01.uti"},
	{0x3AF82CF25777D26FULL, "dm_gen_1.mor"},
	{0x3AF831C30E2A3BD0ULL, "388594261.ogg"},
	{0x3AF8DCFCA72ED0F3ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobc.phy"},
	{0x3AF90CA5A789DA38ULL, "gen00fl_anders6119713.cl"},
	{0x3AF91C090108DAACULL, "vxlt_hangedman_2hy_0_0.vlm"},
	{0x3AF976A37A94EB1FULL, "6103607_m.fxe"},
	{0x3AF98D9122454984ULL, "vxw_bonepitmining_a2_20b_0_0.vlm"},
	{0x3AF9AC4B22A7A967ULL, "gen00fl_dog6115119.cl"},
	{0x3AF9DB0018F51A2BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanl.evt"},
	{0x3AFA4DDB8E68A81EULL, "vxlt_undercityserialkiller_18w_0_0.vlm"},
	{0x3AFA942D08DEE51AULL, "hlac_lt_0_lowtown_n_65537_6x.dds"},
	{0x3AFB4CF2CCC063EDULL, "lt_0_lowtown_docks_n_18_0_ll.mmh"},
	{0x3AFB584A64265AACULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.cess.ani"},
	{0x3AFBDE6EBFE9975EULL, "6141741_m.fxe"},
	{0x3AFC10C4B64ED7B0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave5statue.mao"},
	{0x3AFC14C4B64EDEC3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave5statue.mmh"},
	{0x3AFC22C4B64EF61DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave5statue.msh"},
	{0x3AFCC23B4D078B11ULL, "dre000_bodahn6031278.cl"},
	{0x3AFCC2B2B8E2AD22ULL, "106712173.ogg"},
	{0x3AFCF380B45F0A1DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_surprised_back.gad"},
	{0x3AFE2E95EBB71AF6ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.impa.gad"},
	{0x3AFE4818CFC7B039ULL, "vxw_sundermount_2zu_0_1.vlm"},
	{0x3AFF7152E11CE6EFULL, "6075700_m.fxe"},
	{0x3AFF88A426C1ECA9ULL, "6092033_m.fxe"},
	{0x3AFFC3A756DB81B7ULL, "lowtown_male_alienage2_mob6166543.cl"},
	{0x3B00D83A31C0B52DULL, "38835042.ogg"},
	{0x3B00F40018FB63ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanl.gad"},
	{0x3B00F8016ECDD2DDULL, "vxht_bartrand_n_6pr_0_0.vlm"},
	{0x3B0123CBC45083A1ULL, "6064388_m.fxe"},
	{0x3B015BCC885B95ACULL, "vxht_0_hightown_a2_byt_0_0.vlm"},
	{0x3B0175277DC7D44FULL, "6070022_m.fxe"},
	{0x3B017F1CB1E9F0F3ULL, "ch2armtnt_nolit.psh.fxo"},
	{0x3B0186CEAC1F656FULL, "vxca_caverns1_mines_ghx_0_0.vlm"},
	{0x3B01CAEE3AA3B29AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_bnt_dblw.evt"},
	{0x3B024FC9E83DD362ULL, "vxht_0_hightown_a2_ai9_0_0.vlm"},
	{0x3B02654A642C8FEEULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.cess.gad"},
	{0x3B02681839D6BDE1ULL, "vxlt_0_lowtown_a3_n_c_ghw_0_0.vlm"},
	{0x3B02AD6753146F09ULL, "6070891_m.fxe"},
	{0x3B02F4EB40EB6602ULL, "qun111_ketojan6019040.cl"},
	{0x3B03269C5A8C149FULL, "6125221_m.fxe"},
	{0x3B039FBF1FD9077BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_extreme_v1.ani"},
	{0x3B03E58662B60938ULL, "vxlt_0_lowtown_n_fiw_0_0.vlm"},
	{0x3B0440DC788ECFB8ULL, "704475256.ogg"},
	{0x3B04C148AAACB52EULL, "160292862.ogg"},
	{0x3B04E092722C9445ULL, "277907634.ogg"},
	{0x3B04F70B4F2F0236ULL, "221956418.ogg"},
	{0x3B054995EBBD6755ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.impa.evt"},
	{0x3B0569C45ED28207ULL, "vxlt_undercityencounter_1mq_0_0.vlm"},
	{0x3B05AC9D8186EA25ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pickaxe_n.dds"},
	{0x3B05CC5D5EECD54EULL, "6082930_m.fxe"},
	{0x3B06577BCE3475FAULL, "428012278.ogg"},
	{0x3B0699894FE97D68ULL, "hlac_lt_0_lowtown_docks_n_196609_0z.dds"},
	{0x3B069D83F4C9D7C2ULL, "707541910.ogg"},
	{0x3B0745F39BAA4E45ULL, "6146963_m.fxe"},
	{0x3B07811A6856FF61ULL, "6069869_f.fxe"},
	{0x3B07F300190180EAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanl.ani"},
	{0x3B08318DC25724D3ULL, "6128312_m.fxe"},
	{0x3B08C9EE3AA9CF8DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_bnt_dblw.gad"},
	{0x3B095E025DE25FB3ULL, "docks_male_overseer_mob6176556.cl"},
	{0x3B09604A6432A5CDULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.cess.evt"},
	{0x3B0A037DB9E826FBULL, "629409531.ogg"},
	{0x3B0A5A6F168323DEULL, "190907415.ogg"},
	{0x3B0AE1CE5DC67550ULL, "gen_im_wep_war_one_lore_01.uti"},
	{0x3B0AF41AC35F8829ULL, "vxca_caverns1_edu_0_0.vlm"},
	{0x3B0B2CE8EBFB9250ULL, "gen00fl_fenris6126793.cl"},
	{0x3B0B861EDA2DCC1FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01bonetrack.msh"},
	{0x3B0B8C1EDA2DD6C1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01bonetrack.mmh"},
	{0x3B0B981EDA2DEB62ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01bonetrack.mao"},
	{0x3B0BB0C89DC6B680ULL, "203794197.ogg"},
	{0x3B0BF0CDFB2B69C8ULL, "vxlt_0_lowtown_a3_n_c_ghr_0_0.vlm"},
	{0x3B0C4D92BFD4D0EFULL, "dae000_oldmangossip6174921.cl"},
	{0x3B0CE0899261532FULL, "233926219.ogg"},
	{0x3B0D0692F6F61B4DULL, "657084227.ogg"},
	{0x3B0D26B682CD8FA0ULL, "vxlt_gamlenshouse_n_ky_0_0.vlm"},
	{0x3B0D4BDD3F53BA16ULL, "6087278_m.fxe"},
	{0x3B0D4F71B56DE777ULL, "vxw_encounter2_q0_0_0.vlm"},
	{0x3B0D507468F6F621ULL, "hlac_w_bonepit_0_12x.dds"},
	{0x3B0D55C203C6E782ULL, "llac_lt_gallowstemplar_n_2383_10.dds"},
	{0x3B0D600CF2E531E8ULL, "llac_lt_gallowstemplar_n_2394_1.dds"},
	{0x3B0DD79EAAFD07CFULL, "101139400.ogg"},
	{0x3B0E4445EAE10033ULL, "414471196.ogg"},
	{0x3B0E7714EE254DD2ULL, "zz_followers_banter6159615.cl"},
	{0x3B0EF9F64C62347CULL, "vxlt_gallowstemplar_6sd_0_1.vlm"},
	{0x3B0F8FA42213B7D8ULL, "vxlt_0_lowtown_docks_c1d_0_0.vlm"},
	{0x3B0F9D383F3DB6FFULL, "lla_ht_0_hightown_n_131072_0.dds"},
	{0x3B0FB1B7E1B6FBD3ULL, "gen00fl_merrill6121685.cl"},
	{0x3B1024A90F0230F2ULL, "vxht_chantry_28i_0_0.vlm"},
	{0x3B105AC947AEA212ULL, "llac_lt_0_lowtown_65538_0.dds"},
	{0x3B1098A5B33B57C9ULL, "and230cr_templar3.utc"},
	{0x3B10BB9D16232D17ULL, "vxht_0_hightown_n_c9u_0_0.vlm"},
	{0x3B114DAB69B67CF5ULL, "llac_w_sundermount_196604_0.dds"},
	{0x3B1199045D4E6800ULL, "6127408_m.fxe"},
	{0x3B11A7BF1FE55272ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_extreme_v1.evt"},
	{0x3B1215FFA94F2E6FULL, "6169250_m.fxe"},
	{0x3B127B561D3DA73BULL, "vxht_brothel_n_6g5_0_0.vlm"},
	{0x3B1319CC3FD9899EULL, "6015686_m.fxe"},
	{0x3B14273F3C53B0B1ULL, "lava0.psh.fxo"},
	{0x3B143E3A50E2EDCEULL, "6085830_m.fxe"},
	{0x3B1491955BCA77BBULL, "6142052_f.fxe"},
	{0x3B14C81F30ABFF16ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsa.mmh"},
	{0x3B14D21F30AC1014ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsa.msh"},
	{0x3B14F40C49F33A52ULL, "vxdr_0_deeproads_epy_0_0.vlm"},
	{0x3B15588A7DD90B66ULL, "hlac_lt_backalley_g_0_9y.dds"},
	{0x3B15FF9D3856E869ULL, "llac_lt_0_lowtown_docks_a3_-131074_5.dds"},
	{0x3B163EAC58F3A746ULL, "vxlt_0_lowtown_n_e3h_0_0.vlm"},
	{0x3B1681FC79D71421ULL, "art\\characters\\playercharacter\\qunarimale\\eye_nud.mao"},
	{0x3B171FD2B0838300ULL, "89153145.ogg"},
	{0x3B17255C326EC23EULL, "vxlt_gallowstemplar_f_357_0_0.vlm"},
	{0x3B17260A38511156ULL, "810353736.ogg"},
	{0x3B1731957254D91EULL, "vxlt_undercityserialkiller_12i_0_0.vlm"},
	{0x3B173273BA3AB4F5ULL, "hlac_lt_0_lowtown_a2_n_131071_3y.dds"},
	{0x3B1765BB57E9681DULL, "250921594.ogg"},
	{0x3B1769562F57E646ULL, "6014718_f.fxe"},
	{0x3B177DF3715AD5E8ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_0x.dds"},
	{0x3B17C6BF1FE9F205ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_extreme_v1.gad"},
	{0x3B17F0BEB16A97E7ULL, "483122388.ogg"},
	{0x3B181AD105D0D471ULL, "vxlt_0_lowtown_a2_n_c_eoq_0_0.vlm"},
	{0x3B1887A873F62419ULL, "541682926.ogg"},
	{0x3B18BD6C98152243ULL, "6059144_m.fxe"},
	{0x3B18D8FAD7A668E2ULL, "364681147.ogg"},
	{0x3B1988742BE4CD83ULL, "w_blightlands_180_0.mao"},
	{0x3B198C742BE4D4D0ULL, "w_blightlands_180_0.mmh"},
	{0x3B199A742BE4ECEAULL, "w_blightlands_180_0.msh"},
	{0x3B199AAE83EEE08BULL, "vxw_sundermount_2es_0_0.vlm"},
	{0x3B1A33AEA14008D5ULL, "6089851_m.fxe"},
	{0x3B1A62AAED218248ULL, "vxlt_0_lowtown_a3_n_c_gos_0_0.vlm"},
	{0x3B1A897BB8D1E547ULL, "hlac_lt_0_lowtown_n_131070_3z.dds"},
	{0x3B1AA18FAC1DEF08ULL, "vxlt_undercitydungeon_19o_0_0.vlm"},
	{0x3B1B250C9406ADA1ULL, "vxlt_gallowstemplar_n_6s2_0_1.vlm"},
	{0x3B1B501403E35625ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_4x.dds"},
	{0x3B1B747070C59418ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_003a_0d.dds"},
	{0x3B1B8D6106190D8EULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd04a.mmh"},
	{0x3B1B932D7E3C17ECULL, "hlac_lt_gallowstemplar_f_2383_10x.dds"},
	{0x3B1B989374C3A820ULL, "vxlt_gamlenshouse_n_nf_0_0.vlm"},
	{0x3B1BA76106193A7CULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd04a.msh"},
	{0x3B1BADB258B26B3FULL, "6208688_f.fxe"},
	{0x3B1BB1C594E2C891ULL, "221709532.ogg"},
	{0x3B1C2503EB6884B1ULL, "hla_ht_0_hightown_a3_-65535_0y.dds"},
	{0x3B1C44484A1A7AD8ULL, "vxdr_slavers_1jk_0_0.vlm"},
	{0x3B1C8160AC5EEC90ULL, "gen00fl_anders6125170.cl"},
	{0x3B1CB2CDF32DE7C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v1.gad"},
	{0x3B1CCA1476D7DC7DULL, "hlac_ca_cave_2194_5y.dds"},
	{0x3B1D29A3755740DAULL, "hlac_lt_0_lowtown_n_65534_3z.dds"},
	{0x3B1D9030FD3E300CULL, "250903832.ogg"},
	{0x3B1DDD66CE3FB9BBULL, "llac_ht_chantry_n_2534_5.dds"},
	{0x3B1DE770162A70C4ULL, "6113985_f.fxe"},
	{0x3B1E23D386331F3DULL, "hlac_lt_gallowscourtyard_n_-65536_2z.dds"},
	{0x3B1E3F866C47C6ACULL, "vxht_0_hightown_a2_bgj_0_0.vlm"},
	{0x3B1F1C031320471CULL, "dm_varric.mor"},
	{0x3B1F2B858437CDA4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_3stackxa.phy"},
	{0x3B1F7DEF3E6A1ECFULL, "hlac_ht_playershouse_n_855_28x.dds"},
	{0x3B1FAC9B4A64352FULL, "938821833.ogg"},
	{0x3B1FB45989B3BAD3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_treeleafs_dp.dds"},
	{0x3B1FF649C1E4BFE3ULL, "ave200_donnic6111213.cl"},
	{0x3B1FF940645EB58AULL, "zz_prmpt_debug6200947.cl"},
	{0x3B20204638D80BA7ULL, "6072554_m.fxe"},
	{0x3B20585A98847C96ULL, "vxlt_gallowsprison_n_5wc_0_0.vlm"},
	{0x3B20864B79059D99ULL, "hlac_lt_0_lowtown_131071_4y.dds"},
	{0x3B20B623CEE2478FULL, "art\\characters\\playercharacter\\children\\mt_bas_kf_eyesm1.mao"},
	{0x3B20F43FFB2AAE26ULL, "898207727.ogg"},
	{0x3B21432905B3E933ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgtb_0n.dds"},
	{0x3B2185DED4ABCFD6ULL, "vxht_keep_n_73b_0_0.vlm"},
	{0x3B21B9EF034E9663ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_paintings_single_da.dds"},
	{0x3B21F8448497B040ULL, "144960765.ogg"},
	{0x3B22F1B14DEA0FD0ULL, "hlac_ht_mansion_n_3374_1x.dds"},
	{0x3B23BF4C41E3B1B2ULL, "457013998.ogg"},
	{0x3B23D5CDF33441B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v1.evt"},
	{0x3B242B912AB14FD5ULL, "6110591_m.fxe"},
	{0x3B2465592B6CA59CULL, "605951194.ogg"},
	{0x3B24E1406FE7BB10ULL, "248723747.ogg"},
	{0x3B24EEB5C9A29283ULL, "6019210_m.fxe"},
	{0x3B25A137CE2B1E19ULL, "vxw_bonepit_26g_0_0.vlm"},
	{0x3B25A372644AF19EULL, "vxlt_undercityserialkiller_17j_0_0.vlm"},
	{0x3B25AC4D0B474219ULL, "km_urchin_1.mor"},
	{0x3B25FF8247B858D7ULL, "vxdr_0_deeproads_fil_0_0.vlm"},
	{0x3B262931DC4D1751ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_craggyrock.mmh"},
	{0x3B263531DC4D2BF2ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_craggyrock.mao"},
	{0x3B2642B4DA84C637ULL, "704051730.ogg"},
	{0x3B264331DC4D430FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_craggyrock.msh"},
	{0x3B2669B52101626EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_fa_transition.ani"},
	{0x3B26A1D561B0F545ULL, "llac_lt_gallowstemplar_f_2383_0.dds"},
	{0x3B26D703F755137AULL, "llac_w_sundermount_-196608_0.dds"},
	{0x3B26F6580290B0DBULL, "6082603_m.fxe"},
	{0x3B270EA97EE89C09ULL, "llac_lt_0_lowtown_65534_3.dds"},
	{0x3B271470484896E2ULL, "art\\vfx\\models\\vfx_potion.msh"},
	{0x3B27167048489998ULL, "art\\vfx\\models\\vfx_potion.mmh"},
	{0x3B278CB4C7A29178ULL, "vxlt_gallowscourtyard_n_7gm_0_0.vlm"},
	{0x3B27DFAC7A0BDB2AULL, "hlac_lt_0_lowtown_a2_65534_4x.dds"},
	{0x3B280C8B75E4F90BULL, "vxw_sundermount_2nr_0_0.vlm"},
	{0x3B281EB0B8E698BCULL, "llac_lt_0_lowtown_docks_a3_-262144_1.dds"},
	{0x3B282848192A7802ULL, "nat380_sibling6207552.cl"},
	{0x3B283F2982A79B7DULL, "vxw_encounter1_f2_0_0.vlm"},
	{0x3B2879F2A37E4BE3ULL, "dae000_pc_home_ambients6208500.cl"},
	{0x3B28D487DF179AC8ULL, "vxlt_0_lowtown_a3_n_c_gbp_0_0.vlm"},
	{0x3B29446320975F9AULL, "lowtown_fem_prostitute_sta6152341.cl"},
	{0x3B2948A49E27F9A3ULL, "ts_qunari_arm_props.gda"},
	{0x3B29626A7EFCBEBEULL, "vxht_0_hightown_n_9zo_0_0.vlm"},
	{0x3B299012687AC9E0ULL, "vxht_keep_6up_0_0.vlm"},
	{0x3B2990DEB07F142FULL, "hlac_ht_0_hightown_65537_0x.dds"},
	{0x3B2995BE61EC4306ULL, "6096030_m.fxe"},
	{0x3B29BEA347C78F43ULL, "hlac_ht_mansion2b_818_19y.dds"},
	{0x3B29C0B38D948A5AULL, "hlac_lt_0_lowtown_a3_n_c_-2_1y.dds"},
	{0x3B29C6661008A1A0ULL, "gen00fl_isabela6097167.cl"},
	{0x3B2A22EC2D7FBC7CULL, "6031149_m.fxe"},
	{0x3B2A2F00310D1DA3ULL, "qun111_ketojan6041132.cl"},
	{0x3B2A9FC9F906580DULL, "rock_h_f.dds"},
	{0x3B2AC00245920FF8ULL, "combat_smcs_r4_a1.gda"},
	{0x3B2AC6730261DA83ULL, "w_sundermount_65_0.anb"},
	{0x3B2AD968E6094CCFULL, "vis121_saemus6118482.cl"},
	{0x3B2B00985C747E55ULL, "genip_chest_fancy.utp"},
	{0x3B2B39BDB29F667CULL, "vxca_caverns1_mines_j21_0_1.vlm"},
	{0x3B2B6862D23B5377ULL, "188343183.ogg"},
	{0x3B2C1CCEEB659621ULL, "gen_im_trs_rdr_shd_wdn_r1.uti"},
	{0x3B2C3975CD806525ULL, "6050101_m.fxe"},
	{0x3B2C520B774FAE8DULL, "6096789_m.fxe"},
	{0x3B2D1211DC29DFDFULL, "vxlt_0_lowtown_docks_a3_cza_0_0.vlm"},
	{0x3B2D96E786E6240EULL, "one220_yevhen6098775.cl"},
	{0x3B2DBDAF0FB90761ULL, "vxlt_0_lowtown_a3_n_c_gyt_0_0.vlm"},
	{0x3B2E11F95A2D14CCULL, "gui_guivertexshader_p0.vsh.fxo.dx11"},
	{0x3B2E84F7D7C3E5D1ULL, "572064078.ogg"},
	{0x3B2EA7E7FEFFAC17ULL, "vxht_0_hightown_a2_c1n_0_0.vlm"},
	{0x3B2F01208C7CF36EULL, "vxw_blightlands_78_0_0.vlm"},
	{0x3B2F7E7EE0FBEF43ULL, "art\\levels\\races\\vfx_models\\prp_firecoalsml.mao"},
	{0x3B2F827EE0FBF590ULL, "art\\levels\\races\\vfx_models\\prp_firecoalsml.mmh"},
	{0x3B2F907EE0FC0DAAULL, "art\\levels\\races\\vfx_models\\prp_firecoalsml.msh"},
	{0x3B2FA8E1987D9765ULL, "vxlt_blackemporium_ya_0_0.vlm"},
	{0x3B302D273464F675ULL, "6151332_m.fxe"},
	{0x3B3085D25D38D952ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglehead_n.dds"},
	{0x3B30BE992E1150C9ULL, "1057979713.ogg"},
	{0x3B30D3672F9339FAULL, "6206763_m.fxe"},
	{0x3B30DDCDF33ED9A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v1.ani"},
	{0x3B30E640325DC079ULL, "936284953.ogg"},
	{0x3B310A6CFA7460FEULL, "597367557.ogg"},
	{0x3B311A4AAF66F6EFULL, "764556514.ogg"},
	{0x3B3164F9D1368759ULL, "docks_male_overseer_mob6176569.cl"},
	{0x3B31770EFABBB923ULL, "docks_male_laborer_mob6175386.cl"},
	{0x3B319A1BE0DDA372ULL, "1062417195.ogg"},
	{0x3B31C38C705376EDULL, "305810723.ogg"},
	{0x3B31E825CA21F840ULL, "116531223.ogg"},
	{0x3B3269483676235FULL, "62696343.ogg"},
	{0x3B335978CB468EF3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\pose_a.ani"},
	{0x3B3361B5210BDEBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_fa_transition.evt"},
	{0x3B341314C0213699ULL, "vxw_bonepit_21f_0_0.vlm"},
	{0x3B3420DFCEBDD89AULL, "6139799_m.fxe"},
	{0x3B3485373DC9C459ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03a.phy"},
	{0x3B3489D2AEA60BCCULL, "vxlt_undercityserialkiller_1n7_0_0.vlm"},
	{0x3B35563CAA8259A0ULL, "var_trigger_ambient.gda"},
	{0x3B357226CD3FB286ULL, "zz_lgt_1_debug6114394.cl"},
	{0x3B35A1267C015343ULL, "6053344_m.fxe"},
	{0x3B35E50FB309056DULL, "llac_lt_0_lowtown_a3_n_c_65538_5.dds"},
	{0x3B35F94B883B4FACULL, "sbl_cl_still.mao"},
	{0x3B3617A42B9414EEULL, "350857778.ogg"},
	{0x3B362802FA8D96D9ULL, "ts_raiders_03.gda"},
	{0x3B363B5ABB66EB34ULL, "6075475_m.fxe"},
	{0x3B3650B0DF98144DULL, "199893873.ogg"},
	{0x3B367E6867DB118BULL, "vxw_sundermount_2su_0_0.vlm"},
	{0x3B36A09F9180138FULL, "6036976_m.fxe"},
	{0x3B36B106749DB3FDULL, "vxw_encounter1_k1_0_0.vlm"},
	{0x3B36C8ED5E80E553ULL, "vxdr_0_deeproads_eca_0_0.vlm"},
	{0x3B374881BCB477A6ULL, "lt_0_lowtown_docks_a3_n_19_0_lc.mmh"},
	{0x3B37677748B16EAEULL, "vxht_keep_67d_0_0.vlm"},
	{0x3B37C99ADC50AC86ULL, "dr_slavers_3218.rml"},
	{0x3B37E3B80BA2B7BCULL, "6103903_m.fxe"},
	{0x3B384529048C85F1ULL, "gen_im_arm_glv_lgt_set_med.uti"},
	{0x3B386517F5369D07ULL, "6034066_m.fxe"},
	{0x3B389773026DC80FULL, "w_sundermount_65_0.mao"},
	{0x3B38A373026DDCACULL, "w_sundermount_65_0.mmh"},
	{0x3B38A973026DE69EULL, "w_sundermount_65_0.msh"},
	{0x3B396C6152189458ULL, "gen00fl_merrill6151108.cl"},
	{0x3B3994BCAF75D5A4ULL, "hlac_ca_caverns1_mines_24422_4x.dds"},
	{0x3B39E28894FC392BULL, "402253964.ogg"},
	{0x3B3A19FBA95B8EB6ULL, "270989661.ogg"},
	{0x3B3A42B52111C8A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_fa_transition.gad"},
	{0x3B3A43A52977EB82ULL, "310957306.ogg"},
	{0x3B3B41B5A3F45830ULL, "hlac_lt_0_lowtown_a2_1_6x.dds"},
	{0x3B3B4A016292FF98ULL, "vrc221_bartrand6065338.cl"},
	{0x3B3B69178D83FAA5ULL, "10470726.ogg"},
	{0x3B3C3ED19C94A774ULL, "6018815_f.fxe"},
	{0x3B3C510D583D06ADULL, "vxlt_0_lowtown_a3_n_c_e1v_0_0.vlm"},
	{0x3B3CE48089A9934DULL, "hla_ht_0_hightown_n_131073_1z.dds"},
	{0x3B3D4D62CFD26750ULL, "w_bonepitdestroyed_a3_159_0.mao"},
	{0x3B3D5162CFD26E63ULL, "w_bonepitdestroyed_a3_159_0.mmh"},
	{0x3B3D5F62CFD285BDULL, "w_bonepitdestroyed_a3_159_0.msh"},
	{0x3B3D72FD7E730BEEULL, "vxw_blightlands_89_0_0.vlm"},
	{0x3B3D76084881B0B2ULL, "vxca_caverns1_jg8_0_0.vlm"},
	{0x3B3DD39CDFCF180AULL, "137631111.ogg"},
	{0x3B3E067D3446D43FULL, "hlac_dr_0_deeproads_5843_4y.dds"},
	{0x3B3E6612EE8848C4ULL, "967078999.ogg"},
	{0x3B3E9960348D528FULL, "gen00fl_fenris6126514.cl"},
	{0x3B3EADE22AB1968EULL, "dae000_bartrand6033979.cl"},
	{0x3B3EF85CAD610301ULL, "qcr220_orsino6041779.cl"},
	{0x3B3F59AC84D83330ULL, "571392911.ogg"},
	{0x3B3FE4920CEE50BDULL, "mag311_emile_de_launcet6062690.cl"},
	{0x3B4049A31FD583D2ULL, "vxdr_0_deeproads_ecd_0_0.vlm"},
	{0x3B4073A4ECA3B3E4ULL, "gen_im_trs_krk_hgh_junk1_04.uti"},
	{0x3B40C15D425CA922ULL, "vxw_bonepit_mu_0_0.vlm"},
	{0x3B40C4E3D7B3D844ULL, "lt_0_lowtown_docks_a3_n_32_0.anb"},
	{0x3B41CF319380B453ULL, "6175141_m.fxe"},
	{0x3B4229245FA9FBE8ULL, "vxlt_darktown_5ax_0_0.vlm"},
	{0x3B4253262C14825AULL, "996224374.ogg"},
	{0x3B42C4B369F79AD0ULL, "vxht_bartrand_n_5hx_0_0.vlm"},
	{0x3B42FBAFA09C244CULL, "vxlt_undercityserialkiller_1c4_0_0.vlm"},
	{0x3B4359E876B2022DULL, "w_bonepit_140_0.mmh"},
	{0x3B435DE876B2097EULL, "w_bonepit_140_0.mao"},
	{0x3B436BE876B220B3ULL, "w_bonepit_140_0.msh"},
	{0x3B43B2B3D18CA5AEULL, "1017120244.ogg"},
	{0x3B4440F1B2804342ULL, "hightown_market_guard_sta6169659.cl"},
	{0x3B44C0F9F7E48AB7ULL, "221153856.ogg"},
	{0x3B44C2EA84377C5AULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.dura.gad"},
	{0x3B44F04559D12A0BULL, "vxw_sundermount_2tt_0_0.vlm"},
	{0x3B450B9331709C85ULL, "hlac_w_bonepitmining_a2_0_7x.dds"},
	{0x3B4521047A07AE96ULL, "599545913.ogg"},
	{0x3B4521A559ADD189ULL, "6033681_m.fxe"},
	{0x3B4549ED81C474F8ULL, "14612431.ogg"},
	{0x3B460C58BAF28D55ULL, "561831883.ogg"},
	{0x3B465594FD98D03CULL, "6104803_m.fxe"},
	{0x3B466272F3AEF100ULL, "vxlt_warehouse_n_qy_0_0.vlm"},
	{0x3B466BC6974C426FULL, "659737413.ogg"},
	{0x3B46D3E446572D13ULL, "vxlt_foundry_n_sz_0_0.vlm"},
	{0x3B473078CB56F19DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\pose_a.gad"},
	{0x3B47CAF696A9F7D8ULL, "gen00fl_aveline6120747.cl"},
	{0x3B47FC3E5EBFFAD7ULL, "llac_ca_caverns2prc_5802_14.dds"},
	{0x3B480E23C66044D0ULL, "140948284.ogg"},
	{0x3B4810B198DDB052ULL, "233743165.ogg"},
	{0x3B483735164DE139ULL, "hlac_w_sundermount_65539_0x.dds"},
	{0x3B484E2E28D1B462ULL, "vxht_brothel_n_6bw_0_0.vlm"},
	{0x3B486C3D631FAE86ULL, "6111407_m.fxe"},
	{0x3B4894E83F27A159ULL, "6137611_m.fxe"},
	{0x3B48961210BEE086ULL, "986082793.ogg"},
	{0x3B48A2EF7EFE2433ULL, "ser270_templar6149978.cl"},
	{0x3B48A4968A760969ULL, "mag291_demon_marethari6027863.cl"},
	{0x3B48E603730D2B3EULL, "659359728.ogg"},
	{0x3B498093FCEAB027ULL, "6084898_m.fxe"},
	{0x3B4A4B28E8A08C47ULL, "vxht_keep_n_61h_0_0.vlm"},
	{0x3B4A68A6C5A14D15ULL, "vxht_mansion2b_28j_0_0.vlm"},
	{0x3B4ADD778C8FE6D5ULL, "6015973_m.fxe"},
	{0x3B4AEEE2A0E21BBBULL, "primitives_primitive_p5.psh.fxo.dx11.assembly"},
	{0x3B4B43A13CA2ACE5ULL, "dae000tr_location_qunari.utt"},
	{0x3B4BA15CE74C0D84ULL, "hlac_lt_0_lowtown_a2_n_c_65534_5z.dds"},
	{0x3B4BC5EA843DA071ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.dura.evt"},
	{0x3B4BD2EF35ECAAC8ULL, "6040615_m.fxe"},
	{0x3B4C1E0C7C75B525ULL, "fxe_fswater_ripple.dds"},
	{0x3B4C53C3849F4992ULL, "901602925.ogg"},
	{0x3B4C914D9F017B23ULL, "6061334_m.fxe"},
	{0x3B4CE4EA69197DB8ULL, "hlac_lt_undercitydungeon_-1_1z.dds"},
	{0x3B4D4C28D2AB6B74ULL, "one120_athenril6113291.cl"},
	{0x3B4D61B773B0A12CULL, "dae150_chandan6148793.cl"},
	{0x3B4DB16C5182602BULL, "6138757_m.fxe"},
	{0x3B4DB554CC51AB5BULL, "6174528_m.fxe"},
	{0x3B4DE3CAD264ABC3ULL, "art\\vfx\\models\\vfx_cylinder.mmh"},
	{0x3B4DF1CAD264C31DULL, "art\\vfx\\models\\vfx_cylinder.msh"},
	{0x3B4E6F4EFB2A147CULL, "hlac_ca_caverns1_24424_7x.dds"},
	{0x3B4E8A90EBCE152DULL, "ht_0_hightown_53_0_lc.mmh"},
	{0x3B4EBE961F01BA29ULL, "vxw_bonepit_10z_0_0.vlm"},
	{0x3B4EE802C477C844ULL, "40614405.ogg"},
	{0x3B4EF860206748C1ULL, "mag210cr_arianni.utc"},
	{0x3B4EFD4A72D02E88ULL, "weapon1_dy.psh.fxo"},
	{0x3B4F2AEC2E2C2F66ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg.dura.evt"},
	{0x3B4F2EC57A86AE12ULL, "dae000_oldmangossip6174984.cl"},
	{0x3B4F55959D496070ULL, "hlac_lt_blackemporium_1055_7y.dds"},
	{0x3B4F5E6681012374ULL, "vxht_0_hightown_a3_c50_0_0.vlm"},
	{0x3B509B0151A01468ULL, "vxlt_darktown_5jy_0_0.vlm"},
	{0x3B50C8E1B461D920ULL, "vxw_bonepitmining_a2_1x5_0_0.vlm"},
	{0x3B50D88948CD183EULL, "6169898_f.fxe"},
	{0x3B50ED44A92FE6FDULL, "vxca_caverns1_mines_e6z_0_0.vlm"},
	{0x3B50F826A0FDE65AULL, "hlac_w_encounter2_camp_262145_0x.dds"},
	{0x3B518F8103AEE0D4ULL, "220169622.ogg"},
	{0x3B52294E05764246ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masd.phy"},
	{0x3B52B69D47999C88ULL, "vxca_cavern1_haunted_e81_0_0.vlm"},
	{0x3B52DE70954E8693ULL, "dae000_gamlen6149482.cl"},
	{0x3B530ABC77B2DA60ULL, "vxw_encounter3_eh_0_0.vlm"},
	{0x3B534234A13E068BULL, "llac_ht_0_hightown_n_-196608_2.dds"},
	{0x3B5362224BC7428BULL, "vxw_sundermount_2yw_0_0.vlm"},
	{0x3B5386B6955D1736ULL, "vxlt_blackemporium_xy_0_0.vlm"},
	{0x3B538C0A92664544ULL, "dae000ip_crowd_mf4_merchant.utp"},
	{0x3B53F3C5598A2B67ULL, "454403333.ogg"},
	{0x3B5442CE49BC3193ULL, "llac_lt_0_lowtown_-196608_3.dds"},
	{0x3B5451F5E0476DFCULL, "art\\vfx\\creatures\\fxc_rage_burst_p.impa.ani"},
	{0x3B5452E876C069EAULL, "w_bonepit_140_0.phy"},
	{0x3B547F884614CB3CULL, "6072261_m.fxe"},
	{0x3B54BE9B969D53A9ULL, "gen00fl_isabela6088348.cl"},
	{0x3B54D44FE5A50980ULL, "vxlt_warehouse_n_tf_0_0.vlm"},
	{0x3B55229991C4070AULL, "411634786.ogg"},
	{0x3B5546E90752408DULL, "122615930.ogg"},
	{0x3B55857432B6653DULL, "mag110_samson6018584.cl"},
	{0x3B5593D641325079ULL, "112361669.ogg"},
	{0x3B55E446440A4305ULL, "vxlt_gallowscourtyard_n_7gt_0_0.vlm"},
	{0x3B55F94252DD0AE2ULL, "additiveclouds02.mao"},
	{0x3B5641EC2E3274B1ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg.dura.gad"},
	{0x3B565A3F96EF122AULL, "6101954_m.fxe"},
	{0x3B56757E2B571EA1ULL, "6136641_m.fxe"},
	{0x3B56C1164E5DAF97ULL, "vxca_caverns2_719_0_0.vlm"},
	{0x3B56D876D671AEF0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_tmpa_s.dds"},
	{0x3B5732E3A8B060C6ULL, "6015934_m.fxe"},
	{0x3B57B780A6702FDAULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsmglw.mmh"},
	{0x3B57BB80A6703611ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsmglw.mao"},
	{0x3B57C980A6704E20ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsmglw.msh"},
	{0x3B58DA83B7976595ULL, "vxht_mansion2b_25i_0_0.vlm"},
	{0x3B58DDEA84485340ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.dura.ani"},
	{0x3B58DEDF923E9089ULL, "6103968_m.fxe"},
	{0x3B58FDDD7F314787ULL, "sortedemitter.mao"},
	{0x3B591345E59B50E1ULL, "vxlt_0_lowtown_a3_n_c_gvi_0_0.vlm"},
	{0x3B5961B4C6A2507BULL, "art\\characters\\playercharacter\\elfmale\\em_har_zeva.msh"},
	{0x3B597FB4C6A28365ULL, "art\\characters\\playercharacter\\elfmale\\em_har_zeva.mmh"},
	{0x3B598ACB99F65C57ULL, "hla_ht_0_hightown_a3_-131074_0y.dds"},
	{0x3B59D296928AC3B3ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\hf_hed_casa_0n.dds"},
	{0x3B5A059513D143C4ULL, "819525989.ogg"},
	{0x3B5A099F850184ACULL, "6111016_m.fxe"},
	{0x3B5A0FD6D04BED7BULL, "character_character_p282.psh.fxo.dx11"},
	{0x3B5AC7D7BA92D531ULL, "6110977_m.fxe"},
	{0x3B5AD5FCA7818F68ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobc.mmh"},
	{0x3B5AF3FCA781C192ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobc.msh"},
	{0x3B5B5EF5E04DA23EULL, "art\\vfx\\creatures\\fxc_rage_burst_p.impa.gad"},
	{0x3B5BA8974931BA98ULL, "dae000_madam_lusine6151490.cl"},
	{0x3B5C0C9DC41AFDBAULL, "6034239_f.fxe"},
	{0x3B5C57A61262042BULL, "lt_0_lowtown_n_28_0.anb"},
	{0x3B5CC29F8265134DULL, "419603578.ogg"},
	{0x3B5D1156304C58C4ULL, "hlac_lt_undercityencounter_-1_24x.dds"},
	{0x3B5D307310F7D2A9ULL, "vxw_bonepit_15y_0_0.vlm"},
	{0x3B5D30BD95329B7FULL, "hlac_ht_keep_3732_13y.dds"},
	{0x3B5D42EC2E38955FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg.dura.ani"},
	{0x3B5D477E319F437DULL, "926337551.ogg"},
	{0x3B5D4F2A44E8607DULL, "305085336.ogg"},
	{0x3B5D792A2FD2CA06ULL, "dae000ip_ambient_liecudlm_s.utp"},
	{0x3B5D9F4FB6F9DBFFULL, "hlac_lt_0_lowtown_docks_-4_6x.dds"},
	{0x3B5E13445939E3EAULL, "6048139_f.fxe"},
	{0x3B5E3D2AA7B6ED47ULL, "llac_lt_0_lowtown_a2_131073_1.dds"},
	{0x3B5F0CDE43962CE8ULL, "vxlt_darktown_5ov_0_0.vlm"},
	{0x3B5F2D548A9AC62FULL, "qun230ip_to_smugglers_cut.utp"},
	{0x3B5F3ABEA657F1A0ULL, "vxw_bonepitmining_a2_1u2_0_0.vlm"},
	{0x3B5F56CDD7339A80ULL, "6115406_f.fxe"},
	{0x3B5FDCFF69CC98EBULL, "103761445.ogg"},
	{0x3B603358C56A9BA6ULL, "gen00fl_merrill6151115.cl"},
	{0x3B611E61185B2136ULL, "hm_qmq_orwald.mor"},
	{0x3B61AEFE389A798CULL, "552476614.ogg"},
	{0x3B61E01D9C0661D9ULL, "78338761.ogg"},
	{0x3B620408FED01D8CULL, "240929350.ogg"},
	{0x3B6239F5E053825DULL, "art\\vfx\\creatures\\fxc_rage_burst_p.impa.evt"},
	{0x3B62652622D5599FULL, "hlac_dr_0_deeproads_5842_11x.dds"},
	{0x3B62EFE8E6BE7317ULL, "hlac_lt_0_lowtown_a2_n_c_-3_0y.dds"},
	{0x3B638CCE6F199574ULL, "vxdr_0_deeproads_dq5_0_0.vlm"},
	{0x3B638E452ABC9EB8ULL, "6032478_m.fxe"},
	{0x3B63A7F3F326128BULL, "vfx_dust_03.dds"},
	{0x3B64C2BDDA8C037AULL, "6123650_f.fxe"},
	{0x3B64C62EF821D6D5ULL, "follower_banter6179315.cl"},
	{0x3B6522E44DA4312CULL, "6148308_m.fxe"},
	{0x3B6531E80CBDE562ULL, "vxht_brothel_n_6hq_0_0.vlm"},
	{0x3B653FFB8A589DF5ULL, "vxht_mansion2_1w4_0_0.vlm"},
	{0x3B65C3C1678563B4ULL, "6163145_m.fxe"},
	{0x3B65D06BAABF0C85ULL, "329976366.ogg"},
	{0x3B65F26A5D623982ULL, "art\\levels\\races\\natural\\highlands\\foliage\\_textures\\plc_flowerwild_01_0_d.dds"},
	{0x3B6604DB7495D393ULL, "6079263_m.fxe"},
	{0x3B660887D6E30065ULL, "6088797_runoff.cut"},
	{0x3B66BE2CAF9F4AFDULL, "814560106.ogg"},
	{0x3B66D985A40241DFULL, "vxlt_0_lowtown_docks_a3_cnm_0_0.vlm"},
	{0x3B672A0F394C9EE3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_45ang.mao"},
	{0x3B672E0F394CA530ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_45ang.mmh"},
	{0x3B673638C8BCD17DULL, "w_bonepitmining_a2_51_0.rml"},
	{0x3B673C0F394CBD4AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_45ang.msh"},
	{0x3B6743F08B0A7811ULL, "gen00fl_dog6122241.cl"},
	{0x3B675C85109E81DDULL, "qun131_taskmaster6059013.cl"},
	{0x3B67B7799BE72371ULL, "vxlt_gamlenshouse_n_3l_0_0.vlm"},
	{0x3B67BD28338639ECULL, "6197960_m.fxe"},
	{0x3B68B5297EDBB84BULL, "6037613_f.fxe"},
	{0x3B68CE4F6E014D16ULL, "vxw_encounter2_camp_2x_0_0.vlm"},
	{0x3B68DC32E1DE01AEULL, "vxlt_gallowstemplar_n_4up_0_0.vlm"},
	{0x3B69E5ED1A1E5000ULL, "6152310_m.fxe"},
	{0x3B6A028FE5C31153ULL, "vxht_chantry_338_0_0.vlm"},
	{0x3B6A0AEF01FC1FD4ULL, "follower_banter6163777.cl"},
	{0x3B6A1242ED982CEBULL, "dae011_drunk1_sta6144135.cl"},
	{0x3B6A63CF86ECAF5DULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_coma_0n.dds"},
	{0x3B6A7E7AB611163AULL, "6031139_f.fxe"},
	{0x3B6A90B75C0C150AULL, "70971316.ogg"},
	{0x3B6B56C6A19F5506ULL, "llac_ht_mansion_n_3367_1.dds"},
	{0x3B6C7EBCFEA0BABAULL, "vxlt_0_lowtown_a2_n_c_efi_0_0.vlm"},
	{0x3B6D33381F2505E9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_1.ani"},
	{0x3B6D7EBB358C4568ULL, "vxlt_darktown_5pw_0_0.vlm"},
	{0x3B6DAC9B984E0A20ULL, "vxw_bonepitmining_a2_1r3_0_0.vlm"},
	{0x3B6DDC591DA0ACBBULL, "hlac_lt_undercityencounter_-1_35y.dds"},
	{0x3B6DE16630D06F7DULL, "hightown_fem_cityguard_mob6168487.cl"},
	{0x3B6EC784B04C00C2ULL, "vxw_woundedcoast_1db_0_0.vlm"},
	{0x3B6EEDF680ADA233ULL, "gen_im_trs_krk_drk_junk2_02.uti"},
	{0x3B6EF0684BE9D6D0ULL, "6109736_m.fxe"},
	{0x3B6F2B2D5F7B196CULL, "vxlt_gallowscourtyard_620_0_0.vlm"},
	{0x3B6F62DF0304E656ULL, "80498764.ogg"},
	{0x3B702F3B2816901DULL, "897098426.ogg"},
	{0x3B708D069AE4C3FBULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziercap.mmh"},
	{0x3B7099069AE4D798ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziercap.mao"},
	{0x3B70AB069AE4F6E5ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziercap.msh"},
	{0x3B70CE194CC0C7C5ULL, "6062688_f.fxe"},
	{0x3B71659D2B81C7D2ULL, "6179059_f.fxe"},
	{0x3B71C5725332C264ULL, "vxht_0_hightown_n_c5n_0_0.vlm"},
	{0x3B71D0A091012A36ULL, "ht_0_hightownexplosion_n_2_0.phy"},
	{0x3B71F1DB92284FD5ULL, "6107274_m.fxe"},
	{0x3B72163C4E7B3D00ULL, "qun110_petrice6172065.cl"},
	{0x3B722E3D53CDE5CEULL, "hlac_w_encounter2_camp_65539_0z.dds"},
	{0x3B72458959DB8973ULL, "hlac_lt_0_lowtown_docks_a3_-131075_0z.dds"},
	{0x3B7258825BA56C28ULL, "hlac_ht_bartrand_n_2995_7y.dds"},
	{0x3B72C1FEB56D319AULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\takeoffflylandscream.gad"},
	{0x3B72C74E57820103ULL, "art\\levels\\races\\design\\trap_dart.mao"},
	{0x3B72CB4E57820850ULL, "art\\levels\\races\\design\\trap_dart.mmh"},
	{0x3B72D94E5782206AULL, "art\\levels\\races\\design\\trap_dart.msh"},
	{0x3B72F543F2184105ULL, "lgt220_shp_mag_guard6149873.cl"},
	{0x3B73A2EA86ADABF3ULL, "hlac_lt_undercityencounter_-1_7z.dds"},
	{0x3B73A3C4FEB3FDE2ULL, "vxht_brothel_n_6ur_0_0.vlm"},
	{0x3B73B1D87C4EB675ULL, "vxht_mansion2_1z3_0_0.vlm"},
	{0x3B744C2F0A949485ULL, "vxca_caverns1_mines_j6a_0_0.vlm"},
	{0x3B744E381F2B526BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_1.gad"},
	{0x3B745E33712071D4ULL, "vxlt_darktown_5n9_0_0.vlm"},
	{0x3B74D90A704432C4ULL, "vxw_encounter2_camp_r_0_0.vlm"},
	{0x3B75386F692D095AULL, "6123236_f.fxe"},
	{0x3B75517BF02AE2F1ULL, "one100dg_nabil6173814.cl"},
	{0x3B755869DADD62AEULL, "w_sundermount_10_0.msh"},
	{0x3B756669DADD7AFFULL, "w_sundermount_10_0.mao"},
	{0x3B757269DADD8E9CULL, "w_sundermount_10_0.mmh"},
	{0x3B757788302DC533ULL, "zz_ave_debug6056771.cl"},
	{0x3B75A96DB46DFA8EULL, "ts_prog_opal.gda"},
	{0x3B75BB3321E9FCF3ULL, "hlac_lt_0_lowtown_docks_a3_65534_5x.dds"},
	{0x3B75FD55F0C591DFULL, "6122880_m.fxe"},
	{0x3B76069611B0A988ULL, "vxlt_0_lowtown_a3_n_c_hrs_0_0.vlm"},
	{0x3B768F2FF67096B6ULL, "hlac_ca_caverns1_mines_24421_10z.dds"},
	{0x3B7696562AF53A67ULL, "192370957.ogg"},
	{0x3B76D7841C32F233ULL, "713797724.ogg"},
	{0x3B770AD05F2B8360ULL, "vxht_0_hightown_n_byc_0_0.vlm"},
	{0x3B778B2B59F8D507ULL, "vxlt_0_lowtown_docks_a3_n_art_0_0.vlm"},
	{0x3B77DE6D0FF25A62ULL, "vxw_sundermount_26g_0_0.vlm"},
	{0x3B785B0230AD96C9ULL, "character_character_p190.psh.fxo.dx11.assembly"},
	{0x3B786A7903B5C464ULL, "6121143_m.fxe"},
	{0x3B791940FE0A2377ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.gad"},
	{0x3B79311C88FF6940ULL, "390651325.ogg"},
	{0x3B793EECC5CE7186ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkf.evt"},
	{0x3B797BE73270E9A6ULL, "llac_lt_0_lowtown_-1_1.dds"},
	{0x3B798CA24E0F0DECULL, "lyriumghost_spiritpulse.dds"},
	{0x3B79C4FEB57355B1ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\takeoffflylandscream.evt"},
	{0x3B79DE6A2603A0CEULL, "art\\characters\\creatures\\mubari\\animation\\sleep_p.ani"},
	{0x3B79F181F1E858D7ULL, "747318788.ogg"},
	{0x3B7A4ABD458C04CDULL, "gen00fl_bethany6028816.cl"},
	{0x3B7ABDD2A7610D1DULL, "251488287.ogg"},
	{0x3B7B4B381F316BD0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_1.evt"},
	{0x3B7B6C18D0C7A8A0ULL, "atl_quickbar_dxt5_01.dds"},
	{0x3B7B9C9340B02CD8ULL, "sib111_mother6098952.cl"},
	{0x3B7BB9A0910957D7ULL, "ht_0_hightownexplosion_n_2_0.msh"},
	{0x3B7BCBA09109763AULL, "ht_0_hightownexplosion_n_2_0.mao"},
	{0x3B7BCFA091097D49ULL, "ht_0_hightownexplosion_n_2_0.mmh"},
	{0x3B7C4D5F628B6679ULL, "fen110_anso6095000.cl"},
	{0x3B7C57B2F8016356ULL, "vxca_caverns2_7w2_0_0.vlm"},
	{0x3B7C884A405EC46BULL, "6135808_m.fxe"},
	{0x3B7C9F0A6EF7383AULL, "hlac_lt_0_lowtown_docks_n_-131075_7x.dds"},
	{0x3B7D2247AAF5B37EULL, "6146800.cut"},
	{0x3B7D3961A2421942ULL, "vxw_woundedcoast_1oc_0_0.vlm"},
	{0x3B7D4F77B7DCC7D6ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.cess.ani"},
	{0x3B7D9D0A517131ECULL, "vxlt_gallowscourtyard_691_0_0.vlm"},
	{0x3B7DB023BAC757FDULL, "vxlt_hangedman_27a_0_0.vlm"},
	{0x3B7DFADD97F41A02ULL, "vxdr_0_deeproads_cql_0_0.vlm"},
	{0x3B7E6D46E6011C75ULL, "llac_lt_gallowstemplar_f_7198_1.dds"},
	{0x3B7E727D3E27AC95ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_lgtc_0n.dds"},
	{0x3B7E89634EA5DCE7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_broken_1.mao"},
	{0x3B7EE821F4622C0EULL, "210571941.ogg"},
	{0x3B7EFC91351EAEABULL, "6125922_m.fxe"},
	{0x3B7F1E39793094B8ULL, "dae000ip_crowd_mf3_talk3.utp"},
	{0x3B7F5048DA73B28AULL, "hlac_dr_dungeon_3214_2z.dds"},
	{0x3B7F7F81150FF964ULL, "441040809.ogg"},
	{0x3B7F9E6586B6F038ULL, "808427883.ogg"},
	{0x3B800A42AD0D4E96ULL, "vxca_caverns1_j4y_0_0.vlm"},
	{0x3B801A536F28CB3FULL, "893402719.ogg"},
	{0x3B802E40FE106504ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.evt"},
	{0x3B80374F4528DAE4ULL, "vxht_0_hightown_n_c2o_0_0.vlm"},
	{0x3B8050B561168160ULL, "follower_banter6169784.cl"},
	{0x3B80557092E17BEBULL, "246031753.ogg"},
	{0x3B8055ECC5D4B7D1ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkf.gad"},
	{0x3B808CB9F7AA015CULL, "hlac_lt_0_lowtown_docks_n_0_5z.dds"},
	{0x3B80C7C992543A29ULL, "6095568_m.fxe"},
	{0x3B80E3E3D7E9F0A4ULL, "lt_0_lowtown_docks_a3_n_32_0.rml"},
	{0x3B81094EE3BDA33FULL, "32310345.ogg"},
	{0x3B8127DF5FD0A55CULL, "zz_and_debug6134114.cl"},
	{0x3B812DBF45A96AB6ULL, "31474160.ogg"},
	{0x3B81EE5A86090ABDULL, "hlac_w_sundermount_-327676_0z.dds"},
	{0x3B8228211FA0A598ULL, "6116871_m.fxe"},
	{0x3B82846DAC0A2FA3ULL, "vxht_mansion_n_3ue_0_0.vlm"},
	{0x3B834956DF065147ULL, "486757732.ogg"},
	{0x3B834E0C56EEE4CAULL, "vxlt_0_lowtown_docks_a3_bsz_0_0.vlm"},
	{0x3B8378A5B1AFD78AULL, "vxht_brothel_n_6f3_0_0.vlm"},
	{0x3B83C4D60E8A8CEEULL, "684968827.ogg"},
	{0x3B83F6389686C41AULL, "llac_lt_0_lowtown_a3_n_c_-1_7.dds"},
	{0x3B843B4C20460A29ULL, "760727266.ogg"},
	{0x3B847E395F7CACEFULL, "vignettegui.mao"},
	{0x3B84BF25F5252E3AULL, "llac_lt_0_lowtown_65535_5.dds"},
	{0x3B84D42C7D0424BDULL, "hlac_lt_gallowstemplar_n_7198_5x.dds"},
	{0x3B84D79B1CBAAA00ULL, "vxlt_darktown_4ym_0_0.vlm"},
	{0x3B85CA45053B4EA4ULL, "1051398076.ogg"},
	{0x3B85D742E6D842A3ULL, "374292509.ogg"},
	{0x3B85ED5B262E870DULL, "gen00fl_fenris6108035.cl"},
	{0x3B861901E310E962ULL, "967533120.ogg"},
	{0x3B8656ECC5D9257FULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkf.ani"},
	{0x3B86FD9DE3CF30CBULL, "zz_personality_debug6064394.cl"},
	{0x3B86FEFF834709CFULL, "6075173_m.fxe"},
	{0x3B87655B6286F4E2ULL, "qun111_ketojan6059233.cl"},
	{0x3B87D66A260FD01FULL, "art\\characters\\creatures\\mubari\\animation\\sleep_p.evt"},
	{0x3B87DCFEB57FBB80ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\takeoffflylandscream.ani"},
	{0x3B88295F62B31C4DULL, "lt_0_lowtown_docks_106_0.mmh"},
	{0x3B882D5F62B3229EULL, "lt_0_lowtown_docks_106_0.mao"},
	{0x3B883B5F62B33AD3ULL, "lt_0_lowtown_docks_106_0.msh"},
	{0x3B884C62C542B3E1ULL, "43560562.ogg"},
	{0x3B88655ED385E40CULL, "gen00fl_aveline6105386.cl"},
	{0x3B88D2CCFC7A711BULL, "lt_gallowscourtyard_n_56_0.rml"},
	{0x3B88F2F95539DD06ULL, "vxht_mansion_n_47u_0_0.vlm"},
	{0x3B88F5BF4947515DULL, "mag360_thrask6043538.cl"},
	{0x3B89255D35A1C1E1ULL, "hlac_lt_gamlenshouse_n_-1_2y.dds"},
	{0x3B8954A61288ED77ULL, "lt_0_lowtown_n_28_0.rml"},
	{0x3B898CDD798B4EB6ULL, "6080272_f.fxe"},
	{0x3B8A40A02B23EB23ULL, "vxca_cave_1lw_0_0.vlm"},
	{0x3B8A5777B7E75FC7ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.cess.evt"},
	{0x3B8A58D176C69FE5ULL, "vxht_mansion2b_259_0_0.vlm"},
	{0x3B8AC07D02059D1EULL, "vxlt_undercityserialkiller_1ja_0_0.vlm"},
	{0x3B8B1CC8AD4E4EECULL, "dae000_arishok6072825.cl"},
	{0x3B8B619DFF08C158ULL, "llac_ht_mansion2b_818_37.dds"},
	{0x3B8BAB3E943831C2ULL, "vxw_woundedcoast_1jl_0_0.vlm"},
	{0x3B8BCAB1441CF01BULL, "vxht_mansion2_1v9_0_0.vlm"},
	{0x3B8BDA905BEB20F8ULL, "771394766.ogg"},
	{0x3B8C08FDDC58F6DFULL, "6024975_m.fxe"},
	{0x3B8C17592400D582ULL, "49178926.ogg"},
	{0x3B8C349CAD1367ADULL, "vxlt_warehouse_n_139_0_0.vlm"},
	{0x3B8C4DA86BF7959DULL, "6147674_m.fxe"},
	{0x3B8CBA72069BD3C6ULL, "vxlt_0_lowtown_n_h88_0_0.vlm"},
	{0x3B8CE140A6E1DA89ULL, "6149878_m.fxe"},
	{0x3B8D4640FE1B18FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.ani"},
	{0x3B8D50EDF57F77C3ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgta_0s.dds"},
	{0x3B8D5C132CCD09EEULL, "follower_banter6149223.cl"},
	{0x3B8DB6B25EE72D81ULL, "6076589_m.fxe"},
	{0x3B8DC766469FA4A7ULL, "dae200_saemus6091644.cl"},
	{0x3B8E1E58D23001B3ULL, "102083939.ogg"},
	{0x3B8E2619D6BB33A4ULL, "dae000_ineria6136979.cl"},
	{0x3B8E7C1F9F036716ULL, "vxca_caverns1_j1z_0_0.vlm"},
	{0x3B8EB76A2615BA08ULL, "art\\characters\\creatures\\mubari\\animation\\sleep_p.gad"},
	{0x3B8F09E0B8433D57ULL, "vxlt_0_lowtown_a2_db4_0_0.vlm"},
	{0x3B8F43424239B513ULL, "mag291_torpor6026973.cl"},
	{0x3B8F8F60FCD9BC06ULL, "255958264.ogg"},
	{0x3B8FB7D4FEE2A86EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_out.evt"},
	{0x3B90794A8313DD2EULL, "26620044.ogg"},
	{0x3B90800CEF0FACFAULL, "vxw_encounter2_pg_0_1.vlm"},
	{0x3B9098D8CF438868ULL, "6015820_m.fxe"},
	{0x3B90B81A2D4E9741ULL, "vxht_chantry_n_25z_0_0.vlm"},
	{0x3B90CFD0AC68D419ULL, "113291585.ogg"},
	{0x3B910C5EF5F9C189ULL, "vxlt_hangedman_n_30j_0_0.vlm"},
	{0x3B914C755C25D39AULL, "410094196.ogg"},
	{0x3B91645AD58FF5B7ULL, "162445971.ogg"},
	{0x3B917877B7EDB610ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.cess.gad"},
	{0x3B9213EA693E4A79ULL, "vxlt_0_lowtown_docks_n_cur_0_0.vlm"},
	{0x3B92D01215BBA73BULL, "llac_lt_0_lowtown_a3_n_c_-196609_0.dds"},
	{0x3B92EA4FF59CDA88ULL, "vxlt_0_lowtown_a3_n_c_hdq_0_0.vlm"},
	{0x3B939874E7740BE9ULL, "vxw_woundedcoast_1bc_0_0.vlm"},
	{0x3B93E2905F4242AAULL, "vxlt_0_lowtown_n_e3d_0_0.vlm"},
	{0x3B943CBB4A401248ULL, "6015936_m.fxe"},
	{0x3B947269DAF83FEFULL, "w_sundermount_10_0.rml"},
	{0x3B94F752F78159A0ULL, "vxht_playershouse_n_1pr_0_0.vlm"},
	{0x3B95267AA3DD80F1ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_nervous.ani"},
	{0x3B95AED4FEE70419ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_out.gad"},
	{0x3B96172496DEE835ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_close_c.mmh"},
	{0x3B961A0EA7C0A2EDULL, "hlac_w_encounter3_-393216_0y.dds"},
	{0x3B964631FE052923ULL, "6122163_m.fxe"},
	{0x3B9655AAE8AD6243ULL, "vxw_encounter2_fm_0_0.vlm"},
	{0x3B965F838E53383BULL, "w_bonepitmining_a2_154_0.phy"},
	{0x3B9767C14FFC9576ULL, "vxdr_primevaltomb_2w0_0_1.vlm"},
	{0x3B97D453D483043AULL, "vxlt_0_lowtown_a2_n_c_eud_0_0.vlm"},
	{0x3B989A544C472550ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallceltic.mao"},
	{0x3B989E544C472C63ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallceltic.mmh"},
	{0x3B98AC544C4743BDULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallceltic.msh"},
	{0x3B98CE7D9D8874AAULL, "6140944_m.fxe"},
	{0x3B98F2097D18870BULL, "gen00fl_anders6125597.cl"},
	{0x3B99225F62C1830AULL, "lt_0_lowtown_docks_106_0.phy"},
	{0x3B993259F3FBB59EULL, "vxlt_undercityserialkiller_1ob_0_0.vlm"},
	{0x3B99B8FD969372FCULL, "llac_lt_0_lowtown_65537_5.dds"},
	{0x3B99D0A9459F7265ULL, "t3_wer_b03.tnt"},
	{0x3B9AA6799F09802DULL, "vxlt_warehouse_n_188_0_0.vlm"},
	{0x3B9AB84BAFD79B17ULL, "zz_followers_banter6158130.cl"},
	{0x3B9ADD8A3A4A25ABULL, "static_static_p0.vsh.fxo.dx11"},
	{0x3B9B26592C33E135ULL, "zz_vault_debug6177801.cl"},
	{0x3B9BC3FDAC2AF370ULL, "hlac_lt_0_lowtown_docks_-65537_2y.dds"},
	{0x3B9BCC84321AE452ULL, "gen_im_arm_glv_med_set_med.uti"},
	{0x3B9C317AA3E3B193ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_nervous.gad"},
	{0x3B9C82C47AEC15EBULL, "606246195.ogg"},
	{0x3B9CBFD4FEED4067ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_out.ani"},
	{0x3B9CD59588535B3FULL, "seb221_ruxton_harimann6177209.cl"},
	{0x3B9D121186B32173ULL, "1014284296.ogg"},
	{0x3B9D786D94DFEC58ULL, "gen_im_trs_cot_dua_dag_r2.uti"},
	{0x3B9DCD9B2864F528ULL, "art\\characters\\weapons\\dagger\\w_dger_104a.msh"},
	{0x3B9DEB9B28652852ULL, "art\\characters\\weapons\\dagger\\w_dger_104a.mmh"},
	{0x3B9ED58BD2C228E4ULL, "vxht_brothel_n_59s_0_0.vlm"},
	{0x3B9EE919A4F89D99ULL, "scaleform_sf_main_p1.psh.fxo.dx11"},
	{0x3B9F29F71F44AFC1ULL, "vxht_chantry_n_20y_0_0.vlm"},
	{0x3B9F5B7FA27FC1DEULL, "qun221_petrice6076890.cl"},
	{0x3B9FBF7944078E5EULL, "w_sundermount_99_0.rml"},
	{0x3BA00CA4D47E5044ULL, "art\\levels\\races\\proxy\\props\\_textures\\cotbedroll_d.dds"},
	{0x3BA017FEA12196B2ULL, "hlac_lt_0_lowtown_docks_a3_-3_6y.dds"},
	{0x3BA04CA13CEAFBA8ULL, "dae000tr_location_qunari.ncs"},
	{0x3BA04E6A0FB77FDCULL, "hlac_w_encounter2_camp_131072_0x.dds"},
	{0x3BA158B4C6DFF612ULL, "art\\characters\\playercharacter\\elfmale\\em_har_zeva.phy"},
	{0x3BA15F8ACA3DB67CULL, "vxlt_0_lowtown_a2_n_c_ffp_0_0.vlm"},
	{0x3BA1EBF8736E609DULL, "165246432.ogg"},
	{0x3BA20A51D96A2469ULL, "vxw_woundedcoast_1mb_0_0.vlm"},
	{0x3BA21128D8B835EFULL, "hlac_ca_caverns1_24423_5x.dds"},
	{0x3BA2186C2C0EB72BULL, "vxht_0_hightown_a3_bfw_0_0.vlm"},
	{0x3BA26324C19186FCULL, "vxw_woundedcoast_9z_0_0.vlm"},
	{0x3BA27E1D3016D9F1ULL, "6172577_m.fxe"},
	{0x3BA327A1300271EEULL, "lgt000_chanter6147179.cl"},
	{0x3BA32B4F9A8CFC35ULL, "6080213_f.fxe"},
	{0x3BA35CAE7941EF68ULL, "mag320_nobleman_26060266.cl"},
	{0x3BA39F54FD2F0F83ULL, "ts_templars_prog_junk2.gda"},
	{0x3BA3B4B2ADF26378ULL, "6188116_f.fxe"},
	{0x3BA4052DEF204756ULL, "6174587_m.fxe"},
	{0x3BA4142496EB20C2ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_close_c.anb"},
	{0x3BA46890481B9DEAULL, "vxlt_backalley_g_1ds_0_0.vlm"},
	{0x3BA49834E5008699ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windown.msh"},
	{0x3BA4A234E5009787ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windown.mmh"},
	{0x3BA4A634E5009ED4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windown.mao"},
	{0x3BA4F72D09FAAE6BULL, "tm_hwp_h02.tnt"},
	{0x3BA566994714F95CULL, "dae000cr_grand_cleric6185382.cl"},
	{0x3BA568B59734F1F8ULL, "vxlt_warehouse_dc_0_0.vlm"},
	{0x3BA5A848BDA84D64ULL, "w_blightlands_58_0.msh"},
	{0x3BA5AF814B379602ULL, "vxht_chantry_2cl_0_0.vlm"},
	{0x3BA5BA48BDA86BCDULL, "w_blightlands_58_0.mao"},
	{0x3BA5BE48BDA87206ULL, "w_blightlands_58_0.mmh"},
	{0x3BA5D98716008E55ULL, "follower_banter6148470.cl"},
	{0x3BA5ECA41CB36DCEULL, "follower_banter6172773.cl"},
	{0x3BA63B53B14535C8ULL, "gen00fl_anders6119128.cl"},
	{0x3BA6435E59C5AC03ULL, "vxca_cavern1_haunted_jw8_0_0.vlm"},
	{0x3BA6483B6E14069CULL, "lt_0_lowtown_docks_36_0.rml"},
	{0x3BA654810C4BE73CULL, "350423601.ogg"},
	{0x3BA65A9E82F75FF9ULL, "808213450.ogg"},
	{0x3BA6D879440DD722ULL, "w_sundermount_99_0.phy"},
	{0x3BA7245A0F101C23ULL, "vxca_cave_1by_0_0.vlm"},
	{0x3BA7350F3982954BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wi_45ang.phy"},
	{0x3BA7898184FB59D9ULL, "6093180_m.fxe"},
	{0x3BA7E27D0F353614ULL, "6048141_m.fxe"},
	{0x3BA7E771F9A44215ULL, "qun111_ketojan6039241.cl"},
	{0x3BA7EA52A19EDAC3ULL, "347760897.ogg"},
	{0x3BA8029B8FA74BAEULL, "one280_emrys6104299.cl"},
	{0x3BA88EF8782462C2ULL, "vxw_woundedcoast_1pn_0_0.vlm"},
	{0x3BA8E43374B4539DULL, "vxlt_0_lowtown_docks_n_b5k_0_0.vlm"},
	{0x3BA93142FAFB125BULL, "hlac_lt_darktown_-1_69z.dds"},
	{0x3BA96EB1340EF28CULL, "6052762_m.fxe"},
	{0x3BA99F544C55A088ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallceltic.phy"},
	{0x3BA9D50231360C57ULL, "6092578_m.fxe"},
	{0x3BAA381917E47C06ULL, "dae300_cullen6140596.cl"},
	{0x3BAA468BBC4157E9ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobe.phy"},
	{0x3BAA55035C666586ULL, "hlac_ht_0_hightown_65538_0x.dds"},
	{0x3BAA68445421B3F0ULL, "character_character_p182.psh.fxo.dx11.assembly"},
	{0x3BAAFE30EC3CE06BULL, "mag110_arianni6016910.cl"},
	{0x3BAB6AFB69242A88ULL, "6114293_m.fxe"},
	{0x3BAB7EB80E535909ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blend_1.mao"},
	{0x3BABE20A3A517D7CULL, "fen110_fenris6096213.cl"},
	{0x3BAC31E339F69296ULL, "6144063_m.fxe"},
	{0x3BAC5180BE3A81D8ULL, "6103781_m.fxe"},
	{0x3BAC5E552B58A73CULL, "6117278_m.fxe"},
	{0x3BACAD96F6A635DAULL, "follower_banter6164297.cl"},
	{0x3BAD45B09971A823ULL, "160063201.ogg"},
	{0x3BAD947247B5D1C3ULL, "hlac_dr_slavers_3216_8y.dds"},
	{0x3BADF19DF418F89BULL, "vxlt_0_lowtown_docks_a3_n_bsx_0_0.vlm"},
	{0x3BAE2A9F1A3BBCAAULL, "19007902.ogg"},
	{0x3BAE5598C359156FULL, "6139909_m.fxe"},
	{0x3BAE5C99CB39C849ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_1.dds"},
	{0x3BAE7F358926E780ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_008a_0s.dds"},
	{0x3BAEFDDD5CB4CD88ULL, "gen00fl_varric6074977.cl"},
	{0x3BAF69DA0402823DULL, "6118821_m.fxe"},
	{0x3BAF7D1308BFBA07ULL, "6034237_m.fxe"},
	{0x3BAFCE09D9890B88ULL, "vxlt_0_lowtown_a3_n_c_hno_0_0.vlm"},
	{0x3BB07C2ECB603CE9ULL, "vxw_woundedcoast_1ha_0_0.vlm"},
	{0x3BB0A2F3266C4E24ULL, "6044660_m.fxe"},
	{0x3BB0A6DCB14B78E5ULL, "6078778_m.fxe"},
	{0x3BB0A8AC80A68897ULL, "vxht_0_hightown_a2_cda_0_0.vlm"},
	{0x3BB0B3917F9233EDULL, "vxw_encounter2_7p_0_0.vlm"},
	{0x3BB0EFFA220CF271ULL, "6175477_m.fxe"},
	{0x3BB213E71AC460AAULL, "vxht_0_hightown_91m_0_0.vlm"},
	{0x3BB316DA4418B69FULL, "453009646.ogg"},
	{0x3BB322B84CE6AE36ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_paintings_crazy_da.dds"},
	{0x3BB33964CC999343ULL, "vxw_encounter2_lo_0_0.vlm"},
	{0x3BB36D44709711E9ULL, "hlac_lt_shop_-1_2y.dds"},
	{0x3BB3F25696B55AB2ULL, "6000877_m.fxe"},
	{0x3BB4218590D7500EULL, "obliskbasespec.dds"},
	{0x3BB523CF6D25A889ULL, "qun231cr_varnell.utc"},
	{0x3BB53412EC197BBBULL, "w_bonepitdestroyed_a3_4_0_ll.mmh"},
	{0x3BB545E2335A85A0ULL, "417961327.ogg"},
	{0x3BB5AF6F1159F33CULL, "46984146.ogg"},
	{0x3BB5BB41D182D30AULL, "6116311_f.fxe"},
	{0x3BB5CB27ACCD39EEULL, "gen00fl_merrill6129004.cl"},
	{0x3BB5E29D945B281CULL, "llac_w_sundermount_393213_0.dds"},
	{0x3BB601799E31AB58ULL, "6125367_m.fxe"},
	{0x3BB61613D7E7E69EULL, "vxlt_undercityserialkiller_1yd_0_0.vlm"},
	{0x3BB62FD9BF4D0998ULL, "202531238.ogg"},
	{0x3BB63788B1C36178ULL, "vxht_mansion2_33o_0_0.vlm"},
	{0x3BB64BE7FE285E67ULL, "gen00fl_fenris6110854.cl"},
	{0x3BB6BE48BDB6E5CDULL, "w_blightlands_58_0.rml"},
	{0x3BB6FC19E3BF87A6ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5_b.evt"},
	{0x3BB7198E7ADEA34FULL, "plt_lgt210pt_eau_debt.plo"},
	{0x3BB7B731AADB0C12ULL, "465194627.ogg"},
	{0x3BB7E54D5067F9A5ULL, "97553097.ogg"},
	{0x3BB844DC9A4E1B59ULL, "vxlt_0_lowtown_h7q_0_0.vlm"},
	{0x3BB8CA63B89107E2ULL, "dae000ip_crowd_em2_talk9.utp"},
	{0x3BB8E21CA221993FULL, "6120154_m.fxe"},
	{0x3BB977813B589C5FULL, "6055607_m.fxe"},
	{0x3BB9F42AD97E4674ULL, "hlac_ca_caverns2_5802_72z.dds"},
	{0x3BB9FEBBD16865C6ULL, "1031823448.ogg"},
	{0x3BBAD0321D4EBFBCULL, "6112178_m.fxe"},
	{0x3BBB075411F18AE4ULL, "vxht_0_hightown_a2_c2x_0_0.vlm"},
	{0x3BBB6811D66FD55FULL, "vxht_playershouse_n_1hy_0_0.vlm"},
	{0x3BBB978CDF52AC6FULL, "lt_0_lowtown_a2_n_c_probe_gpy.dds"},
	{0x3BBB9C9B872F2816ULL, "vxht_mansion_n_57l_0_0.vlm"},
	{0x3BBBE77F028FB0EAULL, "mag161st_varian_ilithis.stg"},
	{0x3BBC3A3235E3B085ULL, "119224828.ogg"},
	{0x3BBC5CEC6F6A832DULL, "258206040.ogg"},
	{0x3BBC60C51FAF7727ULL, "305086894.ogg"},
	{0x3BBC9CCE91F4AE7BULL, "vxlt_0_lowtown_docks_bt1_0_0.vlm"},
	{0x3BBCB869A19DE98AULL, "338550297.ogg"},
	{0x3BBCF1A9E8A4703EULL, "prp_str_mirrorstatue_01d.dds"},
	{0x3BBD1319E3C41AF1ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5_b.gad"},
	{0x3BBD59117C86D695ULL, "hf_qfq_mob4.mor"},
	{0x3BBD62C63298F9FFULL, "zz_followers_banter6158952.cl"},
	{0x3BBDA5F33A931E22ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha09a.msh"},
	{0x3BBDA7F33A9321D8ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha09a.mmh"},
	{0x3BBDDBB6F5F89ABDULL, "6117921_m.fxe"},
	{0x3BBDEEEFFAB5D287ULL, "6031137_m.fxe"},
	{0x3BBE1BCE02B0FEC9ULL, "vxlt_gallowscourtyard_n_749_0_0.vlm"},
	{0x3BBE1BFA5DF5708DULL, "6088904_m.fxe"},
	{0x3BBE2471A2C8B18AULL, "follower_banter6155697.cl"},
	{0x3BBE3FE6CB7F2408ULL, "vxlt_0_lowtown_a3_n_c_hkl_0_0.vlm"},
	{0x3BBECF5270E97774ULL, "6111539_m.fxe"},
	{0x3BBF1C87530D7634ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railpost.phy"},
	{0x3BBF3D768513030FULL, "81560276.ogg"},
	{0x3BBF46DEA57DB7FCULL, "vxw_woundedcoast_3x_0_0.vlm"},
	{0x3BC015FF0687EB89ULL, "hm_raider_6.mor"},
	{0x3BC0499D8C491E21ULL, "378714595.ogg"},
	{0x3BC072436CC98FD7ULL, "dae000_keep_male_servant_mo6160898.cl"},
	{0x3BC11AF368D75FE5ULL, "6034355_f.fxe"},
	{0x3BC128CF024B7AA6ULL, "6136365_m.fxe"},
	{0x3BC1CB24205C82F7ULL, "403833950.ogg"},
	{0x3BC24125348CE79FULL, "vxdr_primevaltomb_7o9_0_0.vlm"},
	{0x3BC246838E7848DAULL, "w_bonepitmining_a2_154_0.msh"},
	{0x3BC24C6F7B2122F8ULL, "vxlt_warehouse_ra_0_0.vlm"},
	{0x3BC24E17EF5EB60AULL, "311531255.ogg"},
	{0x3BC254838E786013ULL, "w_bonepitmining_a2_154_0.mao"},
	{0x3BC258838E786720ULL, "w_bonepitmining_a2_154_0.mmh"},
	{0x3BC2773195C4524BULL, "hlac_w_woundedcoast_0_2z.dds"},
	{0x3BC33238C90B289DULL, "w_bonepitmining_a2_51_0.mao"},
	{0x3BC33638C90B2FD6ULL, "w_bonepitmining_a2_51_0.mmh"},
	{0x3BC34038C90B40D4ULL, "w_bonepitmining_a2_51_0.msh"},
	{0x3BC3FC33D1813A79ULL, "zz_ser_debug6078153.cl"},
	{0x3BC41419E3CA3A9FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv5_b.ani"},
	{0x3BC54E0EB7567D0BULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.dura.ani"},
	{0x3BC581EE369E9C80ULL, "6000893_m.fxe"},
	{0x3BC5B027EA97B0A8ULL, "vxht_mansion2_1td_0_0.vlm"},
	{0x3BC5BA1EE68867B2ULL, "vxht_0_hightown_n_asp_0_0.vlm"},
	{0x3BC61CD51AA38C81ULL, "mcr370_cullen6087433.cl"},
	{0x3BC63F0A2126CFCEULL, "6076307_f.fxe"},
	{0x3BC693880C3CBD0CULL, "558880327.ogg"},
	{0x3BC6B817040046CEULL, "563978925.ogg"},
	{0x3BC7D9DD22CC734AULL, "hla_ht_0_hightown_131073_1z.dds"},
	{0x3BC872174F15D4CBULL, "597390103.ogg"},
	{0x3BC8AEF33A9D3563ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha09a.phy"},
	{0x3BC8F2B2017AEAD9ULL, "dae000_hate_enemies6167607.cl"},
	{0x3BC8F98753158F31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railpost.msh"},
	{0x3BC907875315A74CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railpost.mao"},
	{0x3BC913875315BBEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railpost.mmh"},
	{0x3BC96B0629315A77ULL, "vxw_blightlands_b2_0_0.vlm"},
	{0x3BC9793103E7A364ULL, "vxht_0_hightown_a2_c9y_0_0.vlm"},
	{0x3BC9D9EEC865EDDFULL, "vxht_playershouse_n_1ox_0_0.vlm"},
	{0x3BCAEB41FAE892FBULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.dura.ani"},
	{0x3BCAF336E90C5FD3ULL, "vxlt_backalley_n_1og_0_1.vlm"},
	{0x3BCB7DAB31311008ULL, "vxca_caverns2prc_a59_0_0.vlm"},
	{0x3BCB910B88ADBB24ULL, "6034681_m.fxe"},
	{0x3BCB9B3D279BAC13ULL, "6112905_f.fxe"},
	{0x3BCBAA165C5FE6F6ULL, "6182065_m.fxe"},
	{0x3BCBE9B33994EF5EULL, "vxht_0_hightown_auu_0_0.vlm"},
	{0x3BCBF69041BCED5FULL, "vxlt_0_lowtown_docks_a3_c3j_0_0.vlm"},
	{0x3BCCB379442E67D6ULL, "w_sundermount_99_0.mao"},
	{0x3BCCBF79442E7C75ULL, "w_sundermount_99_0.mmh"},
	{0x3BCCC179442E7FEBULL, "w_sundermount_99_0.msh"},
	{0x3BCCD105EE7A3563ULL, "seb321_rebel_mage6180875.cl"},
	{0x3BCDD4CC9739BCE3ULL, "conversation_abilities.gda"},
	{0x3BCDE50191BD8F50ULL, "6085267_m.fxe"},
	{0x3BCDE59EF21000EEULL, "vxw_blightlands_n3_0_0.vlm"},
	{0x3BCDEAA78DC2996DULL, "vxw_blightlands_ay_0_0.vlm"},
	{0x3BCDF64663F8C9FBULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_enco_0s.dds"},
	{0x3BCE1E27BAAD53F2ULL, "368947225.ogg"},
	{0x3BCE23EFF007C975ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxc.phy"},
	{0x3BCE6995F0BFDFB2ULL, "6034931_m.fxe"},
	{0x3BCF5CB8271F222FULL, "vxdr_0_deeproads_cf4_0_0.vlm"},
	{0x3BCF8DA683EEB601ULL, "vxlt_warehouse_n_qt_0_0.vlm"},
	{0x3BCF9AABF4419326ULL, "6131265_m.fxe"},
	{0x3BCFCE0DEA9BD434ULL, "494606582.ogg"},
	{0x3BCFDB5039908D64ULL, "zz_mcr_debug6080849.cl"},
	{0x3BD00910373AC9C9ULL, "vxlt_0_lowtown_docks_a3_n_d4m_0_0.vlm"},
	{0x3BD01D1EB085C443ULL, "vxw_encounter2_rq_0_0.vlm"},
	{0x3BD03EFB7AD8DDE9ULL, "51772870.ogg"},
	{0x3BD08B34E525ABBCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45windown.phy"},
	{0x3BD08E0F7DD689CEULL, "152408886.ogg"},
	{0x3BD13E66DFFD5084ULL, "mag110_samson6018631.cl"},
	{0x3BD142CA8C658615ULL, "zz_mage_forcecrush.cut"},
	{0x3BD1BD666B73D8E5ULL, "6161348_m.fxe"},
	{0x3BD23BC555E4516FULL, "zz_fol_debug6098103.cl"},
	{0x3BD2DEB994503CC9ULL, "6136656_m.fxe"},
	{0x3BD2E5042487F623ULL, "1025953275.ogg"},
	{0x3BD2F9CDBBD4179EULL, "vxlt_undercityserialkiller_1sf_0_0.vlm"},
	{0x3BD3360EB7629182ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.dura.evt"},
	{0x3BD354817A08D2C6ULL, "vxht_0_hightown_9zl_0_0.vlm"},
	{0x3BD3B854A8A06A8BULL, "252833820.ogg"},
	{0x3BD3C047139CA385ULL, "vxlt_warehouse_dz_0_0.vlm"},
	{0x3BD490DB24FAE25FULL, "vxdr_primevaltomb_2xe_0_0.vlm"},
	{0x3BD4A2151BCD0478ULL, "dae100_alistair_drunk6135124.cl"},
	{0x3BD4A3BA4B305AF7ULL, "46396070.ogg"},
	{0x3BD5019F39FB1E24ULL, "pro000_refugee_female6051851.cl"},
	{0x3BD5057884890E63ULL, "146269380.ogg"},
	{0x3BD5067CBC2684A0ULL, "character_character_p162.psh.fxo.dx11"},
	{0x3BD52750C1287E7DULL, "lt_0_lowtown_a2_n_c_33_0.rml"},
	{0x3BD57BD00E43B283ULL, "hlac_lt_0_lowtown_a2_-65535_7x.dds"},
	{0x3BD62941C0F6C3AFULL, "vxht_0_hightown_bns_0_1.vlm"},
	{0x3BD62F8BBC666B0EULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobe.mmh"},
	{0x3BD63B8BBC667FA5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobe.mao"},
	{0x3BD6498BBC6697FCULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobe.msh"},
	{0x3BD6B9A33F12EA90ULL, "463285611.ogg"},
	{0x3BD70112C8C1293AULL, "585027900.ogg"},
	{0x3BD7715FDE550B04ULL, "hlac_ht_chantry_2534_46y.dds"},
	{0x3BD78E9756A9C812ULL, "vxlt_0_lowtown_a2_fg8_0_0.vlm"},
	{0x3BD7C1AB7606E034ULL, "llac_ht_mansion2_818_66.dds"},
	{0x3BD83879D4F894C9ULL, "6089644_m.fxe"},
	{0x3BD8771B68C98BC0ULL, "6083948_m.fxe"},
	{0x3BD8F341FAF4DDF2ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.dura.evt"},
	{0x3BD96513DB027853ULL, "vxlt_backalley_n_1jh_0_1.vlm"},
	{0x3BD99C7B7B19F6E8ULL, "w_bonepit_57_0.msh"},
	{0x3BD9AE7B7B1A1579ULL, "w_bonepit_57_0.mao"},
	{0x3BD9BA7B7B1A2912ULL, "w_bonepit_57_0.mmh"},
	{0x3BDA350EB768AE95ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.dura.gad"},
	{0x3BDAD04770697464ULL, "6111764_m.fxe"},
	{0x3BDAFFB441E1A18DULL, "6086704_m.fxe"},
	{0x3BDB3722BF07DB77ULL, "llac_lt_0_lowtown_a2_n_-131073_1.dds"},
	{0x3BDBC6927157419BULL, "one200dg_solivitus6060340.cl"},
	{0x3BDBFFD87A8D492DULL, "hlac_ca_caverns1_24424_10y.dds"},
	{0x3BDC02DD9AB92F3FULL, "513176985.ogg"},
	{0x3BDC577BE406196EULL, "vxw_blightlands_s4_0_0.vlm"},
	{0x3BDC845B7459165CULL, "994832115.ogg"},
	{0x3BDC9AB7B9EC682BULL, "l_ser270ar_alley.wbd"},
	{0x3BDCE828A6D8E303ULL, "829757580.ogg"},
	{0x3BDD7BAE9EDFF954ULL, "6143198_m.fxe"},
	{0x3BDD9246F3515577ULL, "634930668.ogg"},
	{0x3BDDA2DF0764452CULL, "zz_area_signage16.utp"},
	{0x3BDE2CF8CDA5721BULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.cess.evt"},
	{0x3BDE54A41BB78145ULL, "vxca_cave_2gg_0_0.vlm"},
	{0x3BDE88F43529AED3ULL, "hlac_ht_mansion2_818_26y.dds"},
	{0x3BDF1241FAF97D85ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.dura.gad"},
	{0x3BDF2A4A97FEF58AULL, "18481479.ogg"},
	{0x3BDFAE6C6882FB9DULL, "llac_lt_gallowscourtyard_-131071_3.dds"},
	{0x3BE0015457B1B579ULL, "hm_gxa_vgk_cnspr_3.mor"},
	{0x3BE03D34F8BD8690ULL, "6126490_m.fxe"},
	{0x3BE06A50F77E2FD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_tippy_toes.evt"},
	{0x3BE08EC353B699E8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wic_0d.dds"},
	{0x3BE0A5ED524F0934ULL, "342474796.ogg"},
	{0x3BE0F504EA542093ULL, "6137701_m.fxe"},
	{0x3BE1D198ABB73C64ULL, "787690704.ogg"},
	{0x3BE1E31DD6B1D611ULL, "vxca_caverns1_mines_in2_0_0.vlm"},
	{0x3BE22ECC9B5550ABULL, "llac_lt_0_lowtown_docks_n_65532_1.dds"},
	{0x3BE2AE5934AEC6C2ULL, "hlac_lt_0_lowtown_docks_n_-65539_4x.dds"},
	{0x3BE3351FDF5BE117ULL, "bad_arrow.mao"},
	{0x3BE381A31B0F3CB8ULL, "llac_lt_0_lowtown_docks_-131073_2.dds"},
	{0x3BE3D11727710076ULL, "6110178_m.fxe"},
	{0x3BE487FB695C99C7ULL, "vxca_caverns1_cmv_0_0.vlm"},
	{0x3BE489DC044F038EULL, "828301604.ogg"},
	{0x3BE48F1A2E2B1305ULL, "vxht_mansion2_33v_0_0.vlm"},
	{0x3BE4BC51569435CBULL, "fort_water_cyc.mao"},
	{0x3BE51309FE496993ULL, "hlac_lt_0_lowtown_docks_0_5x.dds"},
	{0x3BE525F8CDAB855CULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.cess.gad"},
	{0x3BE55C3298FE4168ULL, "14016323.ogg"},
	{0x3BE6D604511D0033ULL, "hlac_w_bonepitdestroyed_a3_6_0y.dds"},
	{0x3BE758BB918C31A3ULL, "35979275.ogg"},
	{0x3BE76350F7844214ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_tippy_toes.gad"},
	{0x3BE7FD4DF5D020ADULL, "hlac_w_encounter3_-262145_0y.dds"},
	{0x3BE80724A98E36BFULL, "6131967_m.fxe"},
	{0x3BE8269EC6C0960FULL, "art\\vfx\\combat\\vfx_m_bld_sac_imp_c.impa.gad"},
	{0x3BE914E7172BC9D1ULL, "180752895.ogg"},
	{0x3BE91726AF5EB711ULL, "882260616.ogg"},
	{0x3BE92A67647319F8ULL, "art\\characters\\playercharacter\\humanmale\\hlh_ric.phy"},
	{0x3BE95E2DD05CF299ULL, "786757250.ogg"},
	{0x3BE96D00A08D1EB7ULL, "hlac_w_encounter3_-196607_0z.dds"},
	{0x3BE9A7C430327C86ULL, "qun221_petrice6077989.cl"},
	{0x3BEA0128AF626C0CULL, "art\\characters\\weapons\\shields\\w_shd_008a.msh"},
	{0x3BEA0728AF6276FEULL, "art\\characters\\weapons\\shields\\w_shd_008a.mmh"},
	{0x3BEA1328AF628A95ULL, "art\\characters\\weapons\\shields\\w_shd_008a.mao"},
	{0x3BEA1C73FCEE26CFULL, "688839768.ogg"},
	{0x3BEA7DF721F6E1D2ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_4z.dds"},
	{0x3BEAC958D5FC31EEULL, "vxw_blightlands_t5_0_0.vlm"},
	{0x3BEAFB1AA5FDA691ULL, "6102124_f.fxe"},
	{0x3BEB24F8CDAFEF5AULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.cess.ani"},
	{0x3BEB5C30664BB727ULL, "hlac_ht_mansion_n_3373_1z.dds"},
	{0x3BEB9F6FF5CA98B1ULL, "dae000_porfiria6162614.cl"},
	{0x3BEC80B63711CE53ULL, "hf_gencr_raider_archer.mor"},
	{0x3BECEA58399FD051ULL, "6095835_f.fxe"},
	{0x3BED31F997C98388ULL, "907479486.ogg"},
	{0x3BED334963277865ULL, "mag160_varian_ilithis6018892.cl"},
	{0x3BED3518CE10DA15ULL, "art\\vfx\\materials\\chunky\\vfx_c_grenade_weapon.mao"},
	{0x3BEDF592FE9DBFDBULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_bug01.mao"},
	{0x3BEE00467F67748FULL, "6028525_f.fxe"},
	{0x3BEE3B9EC6C525BCULL, "art\\vfx\\combat\\vfx_m_bld_sac_imp_c.impa.evt"},
	{0x3BEE6D53B10CE819ULL, "gen00fl_isabela6079285.cl"},
	{0x3BEE7250F78A7AF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_tippy_toes.ani"},
	{0x3BEEF11530BA788FULL, "vxdr_0_deeproads_cr0_0_0.vlm"},
	{0x3BEEFDDFDFD2EEC0ULL, "6094775_railing.cut"},
	{0x3BEF3C02AE98A8C4ULL, "vxlt_0_lowtown_a3_n_c_i08_0_0.vlm"},
	{0x3BEF61F05BE7733BULL, "vxw_sundermount_2o1_0_0.vlm"},
	{0x3BEFD1B6BB3D1C54ULL, "581282624.ogg"},
	{0x3BEFD48E0C01DBC4ULL, "z_dust5.dds"},
	{0x3BEFEA9214B1C435ULL, "vxlt_warehouse_n_144_0_0.vlm"},
	{0x3BF054FAC8A7EE91ULL, "vxca_caverns1_mines_ik1_0_0.vlm"},
	{0x3BF07EB05CF7B062ULL, "755855830.ogg"},
	{0x3BF08EE2636B7C93ULL, "vxht_0_hightown_a3_cb6_0_0.vlm"},
	{0x3BF0A400F788D485ULL, "vxlt_warehouse_rx_0_0.vlm"},
	{0x3BF0D6FFFAEDF887ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\_8_-_default.mao"},
	{0x3BF15BBEE56470A6ULL, "6164964_m.fxe"},
	{0x3BF198F7F0E8B631ULL, "gencr_mature_dragon.utc"},
	{0x3BF1DA8AA9D0951CULL, "one200dg_solivitus6048185.cl"},
	{0x3BF20E96A0B317B3ULL, "6209851_m.fxe"},
	{0x3BF21253CCAC34F9ULL, "gen00fl_fenris6113956.cl"},
	{0x3BF263FFE4560E6FULL, "vxlt_gallowscourtyard_6yn_0_0.vlm"},
	{0x3BF2BC6C5497847DULL, "hlac_lt_undercityencounter_-1_38x.dds"},
	{0x3BF2F465A05BBEFFULL, "856973469.ogg"},
	{0x3BF32D6CF9BBF4FAULL, "213683595.ogg"},
	{0x3BF331BED5C7FDE8ULL, "follower_banter6151347.cl"},
	{0x3BF3E1209CFEB21CULL, "6128412_m.fxe"},
	{0x3BF423DB7A945D66ULL, "gen00fl_aveline6140080.cl"},
	{0x3BF44B2D9E9F4AB3ULL, "combat_dw_nmv_default.gda"},
	{0x3BF49A2B086783AFULL, "vxlt_gallowscourtyard_n_69a_0_0.vlm"},
	{0x3BF49B6325B6F677ULL, "127185407.ogg"},
	{0x3BF4ACC837FB6FE4ULL, "vxlt_0_lowtown_docks_b4k_0_1.vlm"},
	{0x3BF4FCF7DAC22EABULL, "follower_banter6172833.cl"},
	{0x3BF5432E7BB8AE9AULL, "6073934_m.fxe"},
	{0x3BF5C3CD4092755DULL, "mcr310_argument6127597.cl"},
	{0x3BF61266AC4912B3ULL, "6188560-kirkwallneedsvcount.cut"},
	{0x3BF6799F58471D7EULL, "499838054.ogg"},
	{0x3BF6A0F58C16494FULL, "809464237.ogg"},
	{0x3BF6A72FBE5C532BULL, "vxca_caverns2prc_9w0_0_0.vlm"},
	{0x3BF6D4FF29E356C3ULL, "6162585_m.fxe"},
	{0x3BF7115BDA629E2FULL, "vxht_chantry_2t4_0_0.vlm"},
	{0x3BF71764EF031642ULL, "ave200cr_donnic.utc"},
	{0x3BF72AA00A74F6DFULL, "vxlt_warehouse_1a2_0_0.vlm"},
	{0x3BF739E3038131B2ULL, "hlac_lt_0_lowtown_a2_n_-1_6z.dds"},
	{0x3BF763C1ADE757A9ULL, "hlac_ht_keep_n_3732_2x.dds"},
	{0x3BF84693D3ACBD2EULL, "6148175_m.fxe"},
	{0x3BF893A374D2431BULL, "hm_noble_gen_4.mor"},
	{0x3BF89B86077BFFFFULL, "llac_lt_gallowscourtyard_n_0_0.dds"},
	{0x3BF8B33A9A9801E7ULL, "23662899.ogg"},
	{0x3BF8DDBDD3E9BE8EULL, "6117052_m.fxe"},
	{0x3BF92E6E7D0A6655ULL, "539329541.ogg"},
	{0x3BF9403E63A4E2EDULL, "vxw_blightlands_pv_0_0.vlm"},
	{0x3BF9DF31919936E2ULL, "vxw_sundermount_2ch_0_0.vlm"},
	{0x3BF9E39D2B8B8D80ULL, "125933942.ogg"},
	{0x3BF9EA9B28B3854DULL, "art\\characters\\weapons\\dagger\\w_dger_104a.phy"},
	{0x3BFAC05DFEA2ADA9ULL, "vxw_bonepitmining_a2_203_0_0.vlm"},
	{0x3BFB385DFFA3B245ULL, "vxca_cave_2me_0_0.vlm"},
	{0x3BFB6C66E5143956ULL, "vxlt_0_lowtown_a2_n_c_eho_0_0.vlm"},
	{0x3BFBE43585E89F06ULL, "vxlt_0_lowtown_a3_n_c_e1i_0_0.vlm"},
	{0x3BFC439EC6D170D5ULL, "art\\vfx\\combat\\vfx_m_bld_sac_imp_c.impa.ani"},
	{0x3BFCF8FC1334631AULL, "hlac_lt_undercitydungeon_-1_1x.dds"},
	{0x3BFD912D88BC5619ULL, "6164277_m.fxe"},
	{0x3BFDD3CD4DDD8BBBULL, "vxw_sundermount_2p0_0_0.vlm"},
	{0x3BFE5C6F06A7DCB5ULL, "vxlt_warehouse_n_193_0_0.vlm"},
	{0x3BFE9C7775FFEB51ULL, "439115758.ogg"},
	{0x3BFEACF17493D5A0ULL, "539571910.ogg"},
	{0x3BFEB08E1BA5E1CEULL, "lt_0_lowtown_docks_n_152_0.phy"},
	{0x3BFEC436599FFB60ULL, "6153292_m.fxe"},
	{0x3BFF0BE4FA0F121AULL, "vxlt_0_lowtown_a2_n_c_gnc_0_0.vlm"},
	{0x3BFF15DDE97EED05ULL, "vxlt_warehouse_wg_0_0.vlm"},
	{0x3BFF3320EB006FA0ULL, "444713081.ogg"},
	{0x3BFF82204B510B90ULL, "988778595.ogg"},
	{0x3BFF90F1BB045681ULL, "vxlt_darktown_4ra_0_0.vlm"},
	{0x3C00C009DD08A5FFULL, "hm_arthuris.mor"},
	{0x3C00C329CEB1BC34ULL, "vxlt_gallowstemplar_f_4n7_0_0.vlm"},
	{0x3C01029D5C439E80ULL, "322376322.ogg"},
	{0x3C01388E7B1D1034ULL, "plt_lgt210pt_eau_debt.nss"},
	{0x3C0147C2158F604CULL, "hlac_lt_0_lowtown_a2_-65534_0y.dds"},
	{0x3C017B728AA6F6E0ULL, "6166591_m.fxe"},
	{0x3C017CB7BA0B51F9ULL, "l_ser270ar_alley.bnk"},
	{0x3C01FABAF6B0FFC0ULL, "mag121_idunna6084184.cl"},
	{0x3C02095568087817ULL, "vxlt_0_lowtown_docks_a3_cqt_0_0.vlm"},
	{0x3C020F96A6D2E022ULL, "6179759_m.fxe"},
	{0x3C02218706C0F3F6ULL, "pro000_cut2d_voyage6024969.cl"},
	{0x3C02BFED87C64DFDULL, "6095388_m.fxe"},
	{0x3C033AFD7EE32BFBULL, "vxw_bonepitdestroyed_a3_1w9_0_0.vlm"},
	{0x3C033B68DEE7B6BDULL, "w_bonepit_239_0.mao"},
	{0x3C033F68DEE7BDF6ULL, "w_bonepit_239_0.mmh"},
	{0x3C034968DEE7CEF4ULL, "w_bonepit_239_0.msh"},
	{0x3C03A162903E4FDFULL, "vxht_playershouse_n_1sd_0_0.vlm"},
	{0x3C03A4B0B9179C51ULL, "vxca_caverns1_mines_dpu_0_0.vlm"},
	{0x3C03D1CB4818F513ULL, "vxlt_gallowstemplar_n_5re_0_0.vlm"},
	{0x3C041755A9466ED2ULL, "face.mat.dx11"},
	{0x3C0421E05E5F83D7ULL, "6150203_m.fxe"},
	{0x3C04A76A7EDDB0DCULL, "424695301.ogg"},
	{0x3C05190CB0526BABULL, "vxca_caverns2prc_9r3_0_0.vlm"},
	{0x3C0604C289C276BEULL, "gen00fl_fenris6106656.cl"},
	{0x3C0607881BD97DD3ULL, "640636266.ogg"},
	{0x3C067579F4389551ULL, "6025803_f.fxe"},
	{0x3C06E461E91B520EULL, "dae000cr_grand_cleric6185498.cl"},
	{0x3C0726A897F09725ULL, "vxlt_warehouse_n_xs_0_0.vlm"},
	{0x3C073F9EE6553A58ULL, "vxca_caverns1_dno_0_0.vlm"},
	{0x3C075B8931B0D906ULL, "creeper_earthmove.dds"},
	{0x3C077D71D2C87860ULL, "vxht_0_hightown_n_b7i_0_0.vlm"},
	{0x3C07BAD8E6DAB77CULL, "6041487_m.fxe"},
	{0x3C07CFFFA5FA0DA2ULL, "vxlt_0_lowtown_a2_n_c_i4c_0_0.vlm"},
	{0x3C082650C153D595ULL, "lt_0_lowtown_a2_n_c_33_0.anb"},
	{0x3C08510E838F4F62ULL, "vxw_sundermount_2di_0_0.vlm"},
	{0x3C097C35784C7919ULL, "vxlt_0_lowtown_docks_a3_aon_0_0.vlm"},
	{0x3C09ABED680338E4ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredidol.gad"},
	{0x3C09E4FAE82BE37AULL, "lowtown_argument6169448.cl"},
	{0x3C0A3B03A8C00620ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_stump01_sa.dds"},
	{0x3C0A5E560CEA6F2AULL, "vxlt_0_lowtown_n_h84_0_0.vlm"},
	{0x3C0AA00F37483C75ULL, "355384834.ogg"},
	{0x3C0AE40E519C6A04ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_noba.msh"},
	{0x3C0AFA0E519C8FA6ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_noba.mmh"},
	{0x3C0B7053F89884CBULL, "6135912_m.fxe"},
	{0x3C0B7CC728AFCF48ULL, "6218906_m.fxe"},
	{0x3C0C439770DDB842ULL, "zz_qcr_debug6056624.cl"},
	{0x3C0CAF2F34B618AFULL, "1011279419.ogg"},
	{0x3C0D429C15C8D7F2ULL, "vxlt_0_lowtown_a2_gvv_0_0.vlm"},
	{0x3C0D87BADB750585ULL, "vxlt_warehouse_xf_0_0.vlm"},
	{0x3C0DDCCE827A30CBULL, "6129610_m.fxe"},
	{0x3C0E22DE84161C77ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobb_0d.dds"},
	{0x3C0E23C1366590F9ULL, "vxlt_0_lowtown_a2_n_gca_0_0.vlm"},
	{0x3C0E86572FDF967FULL, "vxht_0_hightown_n_7iy_0_0.vlm"},
	{0x3C0E962B4D5E4671ULL, "554947386.ogg"},
	{0x3C0F293984852E20ULL, "mnd_con_disc.mao"},
	{0x3C0F39B082CC8A66ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_1x.dds"},
	{0x3C0F4EEEDEDC55B6ULL, "vxlt_blackemporium_9f_0_0.vlm"},
	{0x3C102297F7BD3FDBULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.cess.evt"},
	{0x3C106B84E31F72B9ULL, "zz_followers_banter6158697.cl"},
	{0x3C10BAED680970C2ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredidol.ani"},
	{0x3C10C9AEFFEE90ACULL, "293010151.ogg"},
	{0x3C10ED6DA9AA4847ULL, "vxw_encounter2_fq_0_0.vlm"},
	{0x3C113C853E55F253ULL, "vis121_saemus6167893.cl"},
	{0x3C11761434D9E833ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_concrete.phy"},
	{0x3C1176E1BAA82DE0ULL, "vxht_mansion_n_66v_0_0.vlm"},
	{0x3C11ACDA70D9447BULL, "vxw_bonepitdestroyed_a3_1x8_0_0.vlm"},
	{0x3C11DA53A9ACA85CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curvednb.mao"},
	{0x3C11E653A9ACBCFFULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curvednb.mmh"},
	{0x3C11EC53A9ACC6A1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curvednb.msh"},
	{0x3C12133F8234685FULL, "vxht_playershouse_n_1vg_0_0.vlm"},
	{0x3C127E439215E0F6ULL, "74781388.ogg"},
	{0x3C12B21C2B7FC46AULL, "6201523_f.fxe"},
	{0x3C12B6D9191AABC9ULL, "lgt000cr_hate_gen_challenge.utc"},
	{0x3C12E7BF9F1CE31DULL, "zz_followers_banter6157533.cl"},
	{0x3C13669B1CF0B987ULL, "panning_v.cur"},
	{0x3C145D33ADA5A645ULL, "6124040_m.fxe"},
	{0x3C14F10E51A4D571ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_noba.phy"},
	{0x3C1536EFF043EB90ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxc.msh"},
	{0x3C1544EFF04403AAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxc.mmh"},
	{0x3C1548EFF0440AE1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxc.mao"},
	{0x3C1573ABEE19D815ULL, "w_encounter3_244_0.msh"},
	{0x3C1575ABEE19DB8BULL, "w_encounter3_244_0.mmh"},
	{0x3C1581ABEE19F028ULL, "w_encounter3_244_0.mao"},
	{0x3C158A3A3EF324E0ULL, "1006521441.ogg"},
	{0x3C15C9913A1BA7ABULL, "gen_im_cft_rec_pot_01.uti"},
	{0x3C1618F6D9C8B826ULL, "vxdr_primevaltomb_7sx_0_0.vlm"},
	{0x3C161AD809B40EB6ULL, "551528695.ogg"},
	{0x3C167547CE76175BULL, "vxlt_warehouse_n_159_0_0.vlm"},
	{0x3C16941F28362CC9ULL, "164383738.ogg"},
	{0x3C16C2EB758567E2ULL, "vxw_sundermount_2ij_0_0.vlm"},
	{0x3C171B97F7C3521CULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.cess.gad"},
	{0x3C1723D05F2390D6ULL, "follower_banter6172298.cl"},
	{0x3C173C2CB4F296BAULL, "89350027.ogg"},
	{0x3C174576AE5E99F7ULL, "20413858.ogg"},
	{0x3C174E0D62DC9361ULL, "6068911_m.fxe"},
	{0x3C177BA11DB642B3ULL, "746585914.ogg"},
	{0x3C17965B1F9A3C08ULL, "467906602.ogg"},
	{0x3C17E4ABB5BE1693ULL, "6076785_m.fxe"},
	{0x3C19491147FF52A4ULL, "6049713_m.fxe"},
	{0x3C1955CB9C5A6971ULL, "dae000_ineria.bnk"},
	{0x3C196C291D664ABAULL, "194972666.ogg"},
	{0x3C19DAB8173E04B4ULL, "186944431.ogg"},
	{0x3C19DD3C8D1D8CE1ULL, "dae000_mother6148961.cl"},
	{0x3C1A4639EB1A06A5ULL, "vxlt_darktown_4sz_0_0.vlm"},
	{0x3C1A823DD2E128DBULL, "gen00fl_aveline6121932.cl"},
	{0x3C1AF1B6D51D3884ULL, "6059622_m.fxe"},
	{0x3C1B8DD233D74D46ULL, "vxht_0_hightown_9aq_0_0.vlm"},
	{0x3C1BB47907BEF072ULL, "vxlt_0_lowtown_a2_gmw_0_0.vlm"},
	{0x3C1BC96DCB0B2029ULL, "mer110_jansen6070219.cl"},
	{0x3C1C315BDA693493ULL, "972366729.ogg"},
	{0x3C1C37CB9C5C6B81ULL, "dae000_ineria.cnv"},
	{0x3C1D685AC56FAC6CULL, "vxlt_hangedman_n_1pa_0_0.vlm"},
	{0x3C1E1A97F7C96F1AULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.cess.ani"},
	{0x3C1E664AA9145531ULL, "vxht_mansion2b_1kt_0_0.vlm"},
	{0x3C1E85466812F6AEULL, "vxht_keep_6gt_0_0.vlm"},
	{0x3C1F2486AB596784ULL, "6188612_m.fxe"},
	{0x3C1F5F4A9BA060C7ULL, "vxw_encounter2_ap_0_0.vlm"},
	{0x3C1F67FEAA36A977ULL, "hlac_ca_caverns1_24425_9z.dds"},
	{0x3C1FE645C0B5309CULL, "dae000_epic_merchant_36179503.cl"},
	{0x3C1FED2C76E16740ULL, "vxw_bonepitmining_a2_7o_0_0.vlm"},
	{0x3C200A6FF88752CAULL, "922927313.ogg"},
	{0x3C2039BFEAAE9AE9ULL, "732577897.ogg"},
	{0x3C20A4309577A0E7ULL, "hlac_lt_0_lowtown_131070_2z.dds"},
	{0x3C20AC1359FA7892ULL, "hlac_ca_caverns2_5802_21x.dds"},
	{0x3C20D1E7642ABFCAULL, "6103202_m.fxe"},
	{0x3C20DABC80789C0EULL, "6140140_m.fxe"},
	{0x3C22686C0E68751CULL, "65074495.ogg"},
	{0x3C2269FF1FCFEABFULL, "lt_0_lowtown_docks_a3_16_0.rml"},
	{0x3C22A8F1491C247CULL, "6033824_m.fxe"},
	{0x3C22D8051656D6CBULL, "vxlt_0_lowtown_docks_a3_cvi_0_0.vlm"},
	{0x3C23454658A9E770ULL, "llac_ca_caverns1_24421_0.dds"},
	{0x3C242358CA416B58ULL, "vxca_caverns1_dpm_0_0.vlm"},
	{0x3C242F3171A35CFBULL, "dae011_gossips6149754.cl"},
	{0x3C24990A4FCD9AFFULL, "vxht_chantry_2dt_0_0.vlm"},
	{0x3C24B7D18A6A1672ULL, "6127078_f.fxe"},
	{0x3C24D3CE141DC37BULL, "6055571_m.fxe"},
	{0x3C252A5E89E475D1ULL, "vxlt_shop_en_0_0.vlm"},
	{0x3C2534C8677B8062ULL, "vxw_sundermount_2rk_0_0.vlm"},
	{0x3C2683B73E67251EULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.cess.ani"},
	{0x3C26CB5BA16E4868ULL, "6155478_m.fxe"},
	{0x3C26F5FBD24F0BE3ULL, "100879073.ogg"},
	{0x3C272A474613A383ULL, "qun221_viscount6069379.cl"},
	{0x3C276CF3636C2D77ULL, "590736904.ogg"},
	{0x3C2826CFA5A061F3ULL, "lt_0_lowtown_n_46_0.anb"},
	{0x3C2839FD726C687FULL, "vxdr_0_deeproads_ceq_0_0.vlm"},
	{0x3C28646FBBF3560AULL, "follower_banter6167872.cl"},
	{0x3C295BE97B15272BULL, "vxlt_0_lowtown_docks_bqn_0_0.vlm"},
	{0x3C29A75E27690AE7ULL, "6083099_m.fxe"},
	{0x3C29C639577CCF7BULL, "vxlt_0_lowtown_a3_n_c_eqp_0_0.vlm"},
	{0x3C2BADAF4A221EF8ULL, "vxw_encounter3_4o_0_0.vlm"},
	{0x3C2BE97FF3292534ULL, "llac_lt_0_lowtown_docks_n_65535_1.dds"},
	{0x3C2CEC50188E1728ULL, "581850456.ogg"},
	{0x3C2D2FADBD437AC6ULL, "plt_dae000pt_epilogue.plo"},
	{0x3C2D4103BF5F7CDAULL, "6160649_m.fxe"},
	{0x3C2D47FBBDA8BC85ULL, "mag160_varian_ilithis.bnk"},
	{0x3C2DA955AA9F7CC3ULL, "hlac_w_encounter2_65538_0z.dds"},
	{0x3C2DD1278D967947ULL, "vxw_encounter2_ls_0_0.vlm"},
	{0x3C2DFFADEAC36387ULL, "vxht_mansion2b_1ry_0_0.vlm"},
	{0x3C2E037F60348690ULL, "6153375_m.fxe"},
	{0x3C2E83C6F414ECFBULL, "6037113_m.fxe"},
	{0x3C2E9A6704037FBEULL, "hlac_lt_0_lowtown_docks_a3_n_196609_0y.dds"},
	{0x3C2ECB3B11F96434ULL, "dae300_petrice.cnv"},
	{0x3C2F22779A367906ULL, "906008179.ogg"},
	{0x3C2F27621DFB3E93ULL, "vxlt_gallowstemplar_n_5cf_0_0.vlm"},
	{0x3C2F43C45620D84AULL, "6104302_m.fxe"},
	{0x3C2F98DBEFBC447CULL, "mag170_thrask6016008.cl"},
	{0x3C2FFCA353D9469FULL, "vxca_cave_23y_0_0.vlm"},
	{0x3C3029FBBDAABFBDULL, "mag160_varian_ilithis.cnv"},
	{0x3C30C0F42C315FEAULL, "vxw_bonepitmining_a2_26z_0_0.vlm"},
	{0x3C30F9C02E5CD269ULL, "581284450.ogg"},
	{0x3C3149E2084CEF4BULL, "vxlt_0_lowtown_docks_a3_cyh_0_0.vlm"},
	{0x3C317CD9B899DAD0ULL, "58686517.ogg"},
	{0x3C31D44340FBE60BULL, "vxlt_foundry_n_vm_0_0.vlm"},
	{0x3C31EB910BB8C8FEULL, "vxlt_warehouse_n_114_0_0.vlm"},
	{0x3C32AD3B11FD1A74ULL, "dae300_petrice.bnk"},
	{0x3C330864A976A416ULL, "vxw_encounter2_camp_lf_0_0.vlm"},
	{0x3C33154B5065B749ULL, "llac_ca_caverns2_5802_42.dds"},
	{0x3C333157DB12E5A5ULL, "6053566_m.fxe"},
	{0x3C337156AC8822A7ULL, "361836019.ogg"},
	{0x3C340EE588930BD6ULL, "llac_lt_hangedman_n_-1_2.dds"},
	{0x3C345FBE05CBB6C9ULL, "vxht_0_hightown_a3_7iy_0_0.vlm"},
	{0x3C349BB73E738BEFULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.cess.evt"},
	{0x3C352C4CED15998BULL, "vxlt_gallowstemplar_f_5re_0_0.vlm"},
	{0x3C357E93D0B7A76DULL, "mcr371_hostile_aveline6082206.cl"},
	{0x3C364F53B6E40B71ULL, "zz_mag_debug6051368.cl"},
	{0x3C3677E50FA9816FULL, "166417176.ogg"},
	{0x3C36AF2716D18030ULL, "vxlt_gallowstemplar_f_56f_0_0.vlm"},
	{0x3C373C0FDE4BE259ULL, "6116579_m.fxe"},
	{0x3C378620AF74B670ULL, "llac_w_sundermount_-196611_0.dds"},
	{0x3C378B3CAFDFB77CULL, "vxw_bonepitmining_a2_1xi_0_0.vlm"},
	{0x3C37B76C8252111AULL, "6032924_m.fxe"},
	{0x3C37C1E1CB4AD46BULL, "200997370.ogg"},
	{0x3C37CDC66D0B3FABULL, "vxlt_0_lowtown_docks_bta_0_0.vlm"},
	{0x3C3838164972E7FBULL, "vxlt_0_lowtown_a3_n_c_ets_0_0.vlm"},
	{0x3C38A2A1BCC41338ULL, "hlac_lt_0_lowtown_a2_n_65537_4z.dds"},
	{0x3C3940F9E4200295ULL, "hlac_lt_0_lowtown_docks_a3_-131076_0y.dds"},
	{0x3C3956C70BD8AB49ULL, "6161362_m.fxe"},
	{0x3C39586582DCB881ULL, "vxlt_darktown_4ne_0_0.vlm"},
	{0x3C39EE881DB75A62ULL, "hlac_lt_undercityencounter_-1_15z.dds"},
	{0x3C3A1F8C3C183778ULL, "vxw_encounter3_1l_0_0.vlm"},
	{0x3C3A862784AFF6ABULL, "vxlt_gallowscourtyard_7fj_0_0.vlm"},
	{0x3C3A9CB73E77F898ULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.cess.gad"},
	{0x3C3A9CEBD30A0746ULL, "dae310ip_at_bsc_warehouse.utp"},
	{0x3C3B851109A289F2ULL, "vxlt_0_lowtown_a2_df9_0_0.vlm"},
	{0x3C3B9414863FBA89ULL, "follower_banter6143264.cl"},
	{0x3C3B9D19984F3480ULL, "480664196.ogg"},
	{0x3C3C718ADCB97C07ULL, "vxht_mansion2b_1ux_0_0.vlm"},
	{0x3C3CB6A918A1A630ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2b_1.mao"},
	{0x3C3CC8916E32AB2AULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.anb"},
	{0x3C3CCA16DB5F0669ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_01stump.phy"},
	{0x3C3CD0E65ACD9840ULL, "vxw_bonepitmining_a2_9m_0_0.vlm"},
	{0x3C3CF5A3E60B057BULL, "6030013_m.fxe"},
	{0x3C3D081893986764ULL, "vxht_0_hightown_a2_caa_0_0.vlm"},
	{0x3C3D1A2FA9799B10ULL, "vxw_woundedcoast_10l_0_0.vlm"},
	{0x3C3D46E20E939384ULL, "507651468.ogg"},
	{0x3C3D6DC2336EC727ULL, "560534541.ogg"},
	{0x3C3DB000632C9A1CULL, "hlac_dr_dungeon_3214_5y.dds"},
	{0x3C3DF4AFB126724AULL, "6035917_f.fxe"},
	{0x3C3E067B825BF5E6ULL, "hlac_ht_0_hightown_-196608_0z.dds"},
	{0x3C3E0A8498E4FBC1ULL, "6020656_m.fxe"},
	{0x3C3E0E7584D7BD41ULL, "vxht_chantry_n_2yv_0_0.vlm"},
	{0x3C3E71BD63C0B44DULL, "hlac_ca_caverns1_24425_10x.dds"},
	{0x3C3EAF8E1BDBC391ULL, "lt_0_lowtown_docks_n_152_0.mmh"},
	{0x3C3EBB8E1BDBD832ULL, "lt_0_lowtown_docks_n_152_0.mao"},
	{0x3C3EC98E1BDBF04FULL, "lt_0_lowtown_docks_n_152_0.msh"},
	{0x3C3EE080782ACDABULL, "vxca_caverns2prc_9n7_0_0.vlm"},
	{0x3C3EE645732980C1ULL, "vxca_cavern1_haunted_ehd_0_0.vlm"},
	{0x3C3FCE1690DA9233ULL, "vxdr_0_deeproads_cpj_0_0.vlm"},
	{0x3C3FD6BAF83E7FACULL, "vxht_brothel_n_4vg_0_0.vlm"},
	{0x3C3FE6FD2D097E96ULL, "dev_res_01.mao"},
	{0x3C405D6DFDAEE17EULL, "vxlt_warehouse_n_165_0_0.vlm"},
	{0x3C412C36ADD8D7ACULL, "qcr220_orsino6061811.cl"},
	{0x3C41350D697F3FAEULL, "gen_im_trs_apo_sta_ele_r0.uti"},
	{0x3C417A419B6CBC96ULL, "vxw_encounter2_camp_gg_0_0.vlm"},
	{0x3C41894BF7A76D8FULL, "zz_approval_debug6003602.cl"},
	{0x3C41E020939EC048ULL, "docks_male_human_worker_sta6175975.cl"},
	{0x3C4218824B67B162ULL, "vxw_sundermount_2xu_0_0.vlm"},
	{0x3C42BDCE7EDFA961ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timber45post.phy"},
	{0x3C432477EDADE1BBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoleft.mmh"},
	{0x3C433077EDADF658ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoleft.mao"},
	{0x3C434277EDAE14A5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoleft.msh"},
	{0x3C43C0D41A078464ULL, "qun120_arishok6051667.cl"},
	{0x3C43F906ACBE97E3ULL, "175576246.ogg"},
	{0x3C44176764BFA48DULL, "art\\characters\\playercharacter\\humanmale\\hlh_ric.msh"},
	{0x3C44296764BFC313ULL, "art\\characters\\playercharacter\\humanmale\\hlh_ric.mmh"},
	{0x3C450DB589ED0BDFULL, "hlac_lt_0_lowtown_docks_-131075_5z.dds"},
	{0x3C456CAC12CE4DE9ULL, "l_env_wal_lowtowndocks_military_day.bnk"},
	{0x3C465CED0384A40CULL, "945155922.ogg"},
	{0x3C46B7ECE23625BEULL, "art\\vfx\\combat\\vfx_glb_ele_s_deathblow_c.mmh"},
	{0x3C46ED50C2A5ED08ULL, "w_sundermount_236_0.msh"},
	{0x3C46FF50C2A60B99ULL, "w_sundermount_236_0.mao"},
	{0x3C470B50C2A62032ULL, "w_sundermount_236_0.mmh"},
	{0x3C479D76680E3D17ULL, "w_blightlands_248_0.mmh"},
	{0x3C47A176680E4424ULL, "w_blightlands_248_0.mao"},
	{0x3C47B376680E6289ULL, "w_blightlands_248_0.msh"},
	{0x3C47EA38F63CAEFDULL, "vxlt_hangedman_2ms_0_0.vlm"},
	{0x3C481FFEC9E3CEBFULL, "vxlt_hangedman_n_1pn_0_0.vlm"},
	{0x3C484F25954EFC83ULL, "358711163.ogg"},
	{0x3C4866887CFBDF9DULL, "vxlt_0_lowtown_fee_0_0.vlm"},
	{0x3C48BAA49648DDDEULL, "242222152.ogg"},
	{0x3C4913ECFB1AC91BULL, "376704914.ogg"},
	{0x3C4A9F006031DD33ULL, "vxht_chantry_2lh_0_0.vlm"},
	{0x3C4AB4E17182AA47ULL, "vxw_encounter2_rm_0_0.vlm"},
	{0x3C4B4501719933DCULL, "298777363.ogg"},
	{0x3C4B54057DF114E5ULL, "vxlt_gallowstemplar_n_2p6_0_0.vlm"},
	{0x3C4B8F5720E443C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2_dead_loop.gad"},
	{0x3C4B9AC38E052DB8ULL, "vxlt_warehouse_n_13h_0_0.vlm"},
	{0x3C4C27DD176D0ACAULL, "276232790.ogg"},
	{0x3C4C4028EB3E646AULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.cess.gad"},
	{0x3C4C48DCE7F01CF0ULL, "hlac_ca_caverns1_mines_24426_11x.dds"},
	{0x3C4C4B25C2A38A25ULL, "571700364.ogg"},
	{0x3C4C6C6BBBC05585ULL, "vxca_caverns1_i01_0_0.vlm"},
	{0x3C4CCDC620A290C1ULL, "6108675_m.fxe"},
	{0x3C4DFC2DA85837C2ULL, "crt_monster_spider.bnk"},
	{0x3C4E3FF382D0AAB3ULL, "vxdr_0_deeproads_cmm_0_0.vlm"},
	{0x3C4EA121E92221B0ULL, "438902651.ogg"},
	{0x3C4ECCFEBDC1F3FFULL, "6026989_f.fxe"},
	{0x3C4F49C923C6B56FULL, "vxw_encounter3_f2_0_0.vlm"},
	{0x3C4FB0967387E64DULL, "6105900_m.fxe"},
	{0x3C500BA7F17EC0ABULL, "vxca_caverns1_huy_0_0.vlm"},
	{0x3C504244CD74A6CAULL, "573290583.ogg"},
	{0x3C50E4F1086F4D96ULL, "666687641.ogg"},
	{0x3C50F2E0FF366021ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_rght.ani"},
	{0x3C5150F32D6D9531ULL, "vxlt_0_lowtown_a2_n_gzf_0_0.vlm"},
	{0x3C515D32ADB0F31FULL, "455428887.ogg"},
	{0x3C51C5CA5F3166EBULL, "zz_vis_debug6050342.cl"},
	{0x3C51ED62FE638B0AULL, "vxw_sundermount_2e6_0_0.vlm"},
	{0x3C521006D101CA8BULL, "vxlt_gallowstemplar_f_5xg_0_0.vlm"},
	{0x3C5229BA1F23508EULL, "ave310_donnic6134734.cl"},
	{0x3C525802910565CDULL, "target_pressed.cur"},
	{0x3C5262910FB45E80ULL, "6025161_f.fxe"},
	{0x3C527F6AC5468B52ULL, "gen00fl_anders6113128.cl"},
	{0x3C52FE49BE914CD4ULL, "zz_followers_banter6158034.cl"},
	{0x3C532328EB445101ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.cess.evt"},
	{0x3C539167A0624265ULL, "llac_lt_0_lowtown_a2_n_c_131072_1.dds"},
	{0x3C53D48AF9FAF48FULL, "6180938_m.fxe"},
	{0x3C5443CFA5C5CDAFULL, "lt_0_lowtown_n_46_0.rml"},
	{0x3C547015336349AEULL, "mag370_meredith6037662.cl"},
	{0x3C548CC0FEA227B9ULL, "art\\levels\\races\\proxy\\props\\_textures\\foodstorage_sa.dds"},
	{0x3C5496ECE2422B7DULL, "art\\vfx\\combat\\vfx_glb_ele_s_deathblow_c.anb"},
	{0x3C5563D6058D2F16ULL, "357154506.ogg"},
	{0x3C565D756E4D863FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be05a.mmh"},
	{0x3C5663756E4D90E1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be05a.msh"},
	{0x3C56A97FD68E69D5ULL, "zz_mag_debug6051236.cl"},
	{0x3C56D6C03C54A7AFULL, "dre000_dougal6129773.cl"},
	{0x3C56DED957139832ULL, "gen00fl_aveline6115473.cl"},
	{0x3C572DF54B0AF26CULL, "follower_banter6163768.cl"},
	{0x3C573F4360FE8BBAULL, "1025481999.ogg"},
	{0x3C575C68400C8D6DULL, "298047092.ogg"},
	{0x3C5794456DBF0F3EULL, "mrl000_marethari6129249.cl"},
	{0x3C57FDE0FF3C91C3ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_rght.gad"},
	{0x3C5834BB344C409BULL, "ht_0_hightown_20_0_lc.mmh"},
	{0x3C583B04D1826465ULL, "697258686.ogg"},
	{0x3C58843175EF4493ULL, "6116678_m.fxe"},
	{0x3C58CA4D734B74C9ULL, "vxlt_gallowstemplar_f_453_0_0.vlm"},
	{0x3C58D40AE25C2404ULL, "611283858.ogg"},
	{0x3C5910DD5227F5B3ULL, "vxht_chantry_2ik_0_0.vlm"},
	{0x3C597BF733ED84D8ULL, "vxw_bonepit_2ak_0_0.vlm"},
	{0x3C59A733197A7D78ULL, "vxht_mansion_n_617_0_0.vlm"},
	{0x3C59C2EE2B6BADA7ULL, "vxlt_0_lowtown_a2_n_c_fen_0_0.vlm"},
	{0x3C5A0540C689D085ULL, "vxw_encounter3_4v_0_0.vlm"},
	{0x3C5A0CA07FFB4638ULL, "vxlt_warehouse_n_18i_0_0.vlm"},
	{0x3C5AF22F68C3EE41ULL, "vxht_chantry_n_2ot_0_0.vlm"},
	{0x3C5AF33575FA3857ULL, "6047244_m.fxe"},
	{0x3C5B2DF5F2782B95ULL, "6172713_m.fxe"},
	{0x3C5C1BA9C5DEE5C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyemid.ani"},
	{0x3C5D00947D83CE35ULL, "2p_200_r.stg"},
	{0x3C5D28AB755280F3ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_3.dds"},
	{0x3C5D36A54BAC04E1ULL, "vxw_bonepitdestroyed_a3_5b_0_0.vlm"},
	{0x3C5DB34F9D60B442ULL, "480390165.ogg"},
	{0x3C5E607E17A5FCFFULL, "vxht_chantry_2pp_0_0.vlm"},
	{0x3C5E6CD4A0B7259BULL, "483066839.ogg"},
	{0x3C5E8C07E712B4C6ULL, "577547847.ogg"},
	{0x3C5EDAE0FF4274A8ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_rght.evt"},
	{0x3C5EFB611293EDBBULL, "157365019.ogg"},
	{0x3C5F381CBEACACFAULL, "hlac_dr_0_deeproads_5842_5x.dds"},
	{0x3C5F81AAF55C1CCDULL, "mag100_cullen6095102.cl"},
	{0x3C5F9C8FA66FDA9EULL, "6163824_m.fxe"},
	{0x3C5FBA5720F535A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2_dead_loop.ani"},
	{0x3C5FC2D01F63ADB1ULL, "vxlt_0_lowtown_a2_n_gwe_0_0.vlm"},
	{0x3C605F3FF059A38AULL, "vxw_sundermount_2n7_0_0.vlm"},
	{0x3C607E123917E682ULL, "6112100_f.fxe"},
	{0x3C6081E3C2F7E30BULL, "vxlt_gallowstemplar_f_5cf_0_0.vlm"},
	{0x3C60B952D154CC72ULL, "122898983.ogg"},
	{0x3C60D92CBE96D2BDULL, "vxht_chantry_28r_0_0.vlm"},
	{0x3C611B28EB508150ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.cess.ani"},
	{0x3C6143D0983504F0ULL, "6129114_m.fxe"},
	{0x3C61C83E813DB2D7ULL, "gen00fl_isabela6079030.cl"},
	{0x3C62059B56D8A8AEULL, "zz_arishock_forestcu.cut"},
	{0x3C624D4BD02A40D6ULL, "llac_lt_gallowstemplar_f_2394_7.dds"},
	{0x3C624E96751F671FULL, "815275937.ogg"},
	{0x3C6267E37FF86B91ULL, "6098318_m.fxe"},
	{0x3C627E62465491ECULL, "6119464_f.fxe"},
	{0x3C628CFF20060A23ULL, "lt_0_lowtown_docks_a3_16_0.anb"},
	{0x3C6314330F17CDBEULL, "6118757_f.fxe"},
	{0x3C6333265594C31AULL, "1017217715.ogg"},
	{0x3C64249BE5E0C674ULL, "6191948_m.fxe"},
	{0x3C64577D89FB7B29ULL, "nat380_sibling6103570.cl"},
	{0x3C64734C66D3ECF3ULL, "436017398.ogg"},
	{0x3C64FC4F1177A8F1ULL, "6067716_m.fxe"},
	{0x3C6583C2FE6CFF36ULL, "art\\characters\\creatures\\spider\\animation\\tal_threaten.ani"},
	{0x3C65B949CB598A6FULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_table_n.dds"},
	{0x3C65CB916E55B51DULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.mmh"},
	{0x3C65D1C24E7BE259ULL, "llac_w_encounter1_65537_0.dds"},
	{0x3C661531F37E3141ULL, "seb321_leliana6180960.cl"},
	{0x3C669BC0E67435D2ULL, "vxlt_gallowscourtyard_n_75y_0_0.vlm"},
	{0x3C6782BA441E0E33ULL, "vxht_chantry_2bj_0_0.vlm"},
	{0x3C67C6346FDB66F0ULL, "llac_w_bonepit_458757_0.dds"},
	{0x3C67D455678C81ACULL, "vxlt_0_lowtown_docks_c4s_0_0.vlm"},
	{0x3C6835EF792BE608ULL, "l_env_wal_lowtown_crowd_bed.wbd"},
	{0x3C683686FBC2C115ULL, "6128707_m.fxe"},
	{0x3C68771DB87FE905ULL, "vxw_encounter3_1u_0_0.vlm"},
	{0x3C689D01C25AEEB6ULL, "6104056_m.fxe"},
	{0x3C68B316DB84198EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_01stump.mmh"},
	{0x3C68BF16DB842E25ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_01stump.mao"},
	{0x3C68CD16DB84467CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_01stump.msh"},
	{0x3C69E30B9E2040F3ULL, "vxht_mansion2b_3e2_0_0.vlm"},
	{0x3C69E3B2F5177CBFULL, "vxlt_0_lowtown_eu4_0_0.vlm"},
	{0x3C6AD9D631C3695EULL, "108937455.ogg"},
	{0x3C6AF373DE0664BDULL, "157903046.ogg"},
	{0x3C6C0D40C0E1584BULL, "170977143.ogg"},
	{0x3C6C1E3CE25E0814ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic_1.mao"},
	{0x3C6C446714DF7B69ULL, "vxw_woundedcoast_17l_0_0.vlm"},
	{0x3C6C96B061FB0927ULL, "hlac_lt_0_lowtown_a2_n_c_65535_9z.dds"},
	{0x3C6CFBA3F23F672BULL, "6129368_m.fxe"},
	{0x3C6D2C5A82297011ULL, "6179261_f.fxe"},
	{0x3C6D4B105793DD2BULL, "hlac_w_encounter3_-65540_0y.dds"},
	{0x3C6DDC1F64439F2AULL, "vxht_0_hightown_9pb_0_0.vlm"},
	{0x3C6E681B8D1CB777ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_quick_modal_activate.gad"},
	{0x3C6EAA946176B3E3ULL, "vxlt_0_lowtown_a2_n_c_gjr_0_0.vlm"},
	{0x3C6ED11CE24FBC0AULL, "vxw_sundermount_2s0_0_0.vlm"},
	{0x3C6FC6D73EC5E25BULL, "gen00fl_merrill6127544.cl"},
	{0x3C6FCFDBFF2416AAULL, "t3_arm_cot.tnt"},
	{0x3C702CD9BF3EEAEDULL, "vxlt_darktown_469_0_0.vlm"},
	{0x3C7104A9C5F11AC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyemid.gad"},
	{0x3C72986CBE162CCAULL, "6072350_f.fxe"},
	{0x3C729875DE3A6578ULL, "zz_followers_banter6158166.cl"},
	{0x3C72A2A5D5327632ULL, "vxdr_slavers_1bi_0_0.vlm"},
	{0x3C72C10A66CAA2E4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robd_0d.dds"},
	{0x3C72E10A068F8601ULL, "hlac_ht_mansion2_818_37y.dds"},
	{0x3C72E1B80F3426AEULL, "vxlt_0_lowtown_e82_0_0.vlm"},
	{0x3C738BC2FE794A27ULL, "art\\characters\\creatures\\spider\\animation\\tal_threaten.evt"},
	{0x3C73E62DA878E1E0ULL, "crt_monster_spider.wbd"},
	{0x3C744CF5AA445D11ULL, "vxlt_0_lowtown_docks_a3_n_cxn_0_0.vlm"},
	{0x3C746F14352E59A8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_concrete.mmh"},
	{0x3C747B14352E6E4BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_concrete.mao"},
	{0x3C74837AF09A7151ULL, "follower_barks6173400.cl"},
	{0x3C748D14352E8CD2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_concrete.msh"},
	{0x3C74EF53AA014944ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curvednb.phy"},
	{0x3C75145F0E0A36C0ULL, "seb201_lady_harimann6179352.cl"},
	{0x3C75598BC3768E75ULL, "lt_0_lowtown_docks_a3_n_243_0.msh"},
	{0x3C755B8BC37691EBULL, "lt_0_lowtown_docks_a3_n_243_0.mmh"},
	{0x3C75678BC376A588ULL, "lt_0_lowtown_docks_a3_n_243_0.mao"},
	{0x3C757D1B8D22F904ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_quick_modal_activate.evt"},
	{0x3C7659DAE9AD9D7AULL, "gen00fl_sebastian6188572.cl"},
	{0x3C774CA940C3BAEFULL, "hlac_lt_undercityencounter_-1_31y.dds"},
	{0x3C7771C1670D1E7BULL, "qun221_viscount6074193.cl"},
	{0x3C7778A983685DFAULL, "12976766.ogg"},
	{0x3C77CB651A127D4CULL, "6074353_m.fxe"},
	{0x3C7810ADBD833105ULL, "plt_dae000pt_epilogue.nss"},
	{0x3C78343FA0FB23D4ULL, "6127591_f.fxe"},
	{0x3C785DDECF4CBF5BULL, "hlac_lt_0_lowtown_docks_a3_-65536_0z.dds"},
	{0x3C78AADDFDE311DAULL, "lt_0_lowtown_docks_220_0.mmh"},
	{0x3C78AEDDFDE31811ULL, "lt_0_lowtown_docks_220_0.mao"},
	{0x3C78BCDDFDE33020ULL, "lt_0_lowtown_docks_220_0.msh"},
	{0x3C78E5909A705A15ULL, "6076551_f.fxe"},
	{0x3C78F97210C257A0ULL, "ts_raiders_wpn_props.gda"},
	{0x3C791415495C8C61ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower1_meredithsintro_varric.gad"},
	{0x3C79874BCE3D008EULL, "art\\characters\\creatures\\arh_abm_rev\\c_abm_head.mao"},
	{0x3C79ACC2FE7DEE70ULL, "art\\characters\\creatures\\spider\\animation\\tal_threaten.gad"},
	{0x3C79B50BA8D52229ULL, "dae000_light_tombs_boss6179474.cl"},
	{0x3C7A7CDC930EE7FBULL, "hlac_lt_backalley_g_65536_5y.dds"},
	{0x3C7AB64406D593E9ULL, "vxw_woundedcoast_12s_0_0.vlm"},
	{0x3C7ACC63283E0BC4ULL, "art\\vfx\\combat\\vfx_gib_human_bld_p.anb"},
	{0x3C7ADB0965D557EBULL, "clf260_harley6089562.cl"},
	{0x3C7C4883E22266AEULL, "gen00fl_varric6118729.cl"},
	{0x3C7C7C4301DDC8ACULL, "289128811.ogg"},
	{0x3C7D42F9D445D48AULL, "vxw_sundermount_2t1_0_0.vlm"},
	{0x3C7D659DA6E4140BULL, "vxlt_gallowstemplar_f_5ix_0_0.vlm"},
	{0x3C7D8330D5CA18E6ULL, "llac_ht_chantry_n_2534_19.dds"},
	{0x3C7D9975698468F8ULL, "6162307_f.fxe"},
	{0x3C7E5AC72F862611ULL, "268557169.ogg"},
	{0x3C7E73B6D6914F3BULL, "424483426.ogg"},
	{0x3C7F0A1115B485F5ULL, "468829590.ogg"},
	{0x3C7F15154960F90FULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower1_meredithsintro_varric.ani"},
	{0x3C7F16B1FF8AF564ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation.ani"},
	{0x3C7FC223094F8611ULL, "character_character_p2.psh.fxo.dx11.assembly"},
	{0x3C7FF33091353746ULL, "mag291_vincento6027683.cl"},
	{0x3C811306E95BF844ULL, "flip_cover.cur"},
	{0x3C815395012A3F2EULL, "vxlt_0_lowtown_e33_0_0.vlm"},
	{0x3C81CD8C66F9EA09ULL, "82209620.ogg"},
	{0x3C8228237CF3B3B6ULL, "6176044_m.fxe"},
	{0x3C82951B8D2DACFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_quick_modal_activate.ani"},
	{0x3C8295522BF9C3AEULL, "gen_im_npc_dwf_arm_boo_lega.uti"},
	{0x3C82A80676DA3420ULL, "6163475_m.fxe"},
	{0x3C82AC69E28E74D5ULL, "lt_0_lowtown_docks_n_36_0.rml"},
	{0x3C82CBEA7AEEE925ULL, "6154300_m.fxe"},
	{0x3C82CEEEA048F5A8ULL, "6099525_m.fxe"},
	{0x3C82D765E52A1845ULL, "vxw_sundermount_307_0_0.vlm"},
	{0x3C82D8D1F64C2051ULL, "trap_activation.ncs"},
	{0x3C8344863842FAE8ULL, "hlac_lt_0_lowtown_n_65533_0x.dds"},
	{0x3C84000238C679F1ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg_b.anb"},
	{0x3C843440909CEED4ULL, "750396385.ogg"},
	{0x3C84486E22DE74BAULL, "vxlt_undercityserialkiller_1n1_0_0.vlm"},
	{0x3C8466CA2AF0DDA1ULL, "vxlt_gallowstemplar_f_4uk_0_0.vlm"},
	{0x3C84D5CDA290F0D2ULL, "lt_0_lowtown_docks_11_0_lc.mmh"},
	{0x3C85069E23428AB7ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipslice.msh"},
	{0x3C85189E2342A91AULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipslice.mao"},
	{0x3C851C9E2342B029ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipslice.mmh"},
	{0x3C854750E33E2FFEULL, "59020710.ogg"},
	{0x3C85926492C76B87ULL, "6026995_m.fxe"},
	{0x3C85E13F69ECA9C2ULL, "6082116_m.fxe"},
	{0x3C8613B1FF910EC6ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation.gad"},
	{0x3C863D420C0895CCULL, "6079053_m.fxe"},
	{0x3C86A3835307EC34ULL, "6078178_f.fxe"},
	{0x3C86CF11BE9A9C2BULL, "vxht_keep_n_4ha_0_0.vlm"},
	{0x3C871DC0399DB4FDULL, "hlac_lt_undercityserialkiller_-1_30z.dds"},
	{0x3C8784186510D317ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_avea_0d.dds"},
	{0x3C87B19F6DC5A0B2ULL, "character_character_p230.psh.fxo.dx11"},
	{0x3C87E98D104BE6E5ULL, "1066488985.ogg"},
	{0x3C87EB632848CAC7ULL, "art\\vfx\\combat\\vfx_gib_human_bld_p.mmh"},
	{0x3C885854E60ECFD1ULL, "dae000_bartrand6027557.cl"},
	{0x3C887725511EE4F3ULL, "672519274.ogg"},
	{0x3C8885C9A11E9FD5ULL, "6162785_m.fxe"},
	{0x3C88E015B9DC9E4BULL, "llac_lt_foundry_n_320_6.dds"},
	{0x3C8944241601F3EFULL, "plt_fen000pt_gift_sword.plo"},
	{0x3C8945CE99A313AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_draw.ani"},
	{0x3C898D7D7FCD94CBULL, "zz_approval_debug6003826.cl"},
	{0x3C898F1F4A31090EULL, "6187866_m.fxe"},
	{0x3C8993B083E3B714ULL, "425687343.ogg"},
	{0x3C899D6D8C3620BDULL, "108982526.ogg"},
	{0x3C89A0D3580E413EULL, "581122827.ogg"},
	{0x3C89CFCB254ABD45ULL, "6059151_m.fxe"},
	{0x3C89D3F314C98BADULL, "vxlt_hangedman_n_lz_0_1.vlm"},
	{0x3C8A2994DE03FE06ULL, "llac_dr_dungeon_3216_4.dds"},
	{0x3C8A4577EDEA1CF0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoleft.phy"},
	{0x3C8A57BC54BB6A8FULL, "ave100cr_raider_assassin.utc"},
	{0x3C8A71D4AFBB4F97ULL, "gen00fl_varric6107300.cl"},
	{0x3C8B108938D7FF4AULL, "lgt220_shp_mag_guard6149524.cl"},
	{0x3C8B422F7EED2629ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_act_c.impa.ani"},
	{0x3C8B91DD1D7CFBB0ULL, "fake_im_flw_sta_bethany_03.uti"},
	{0x3C8C174810360E62ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medg_0t.dds"},
	{0x3C8C43E3A968D16DULL, "llac_lt_0_lowtown_docks_n_-1_0.dds"},
	{0x3C8C4AAC130A163BULL, "l_env_wal_lowtowndocks_military_day.wbd"},
	{0x3C8C8F13BCA23268ULL, "vxca_caverns1_mines_j51_0_0.vlm"},
	{0x3C8CBB488F916353ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bottlewine_sa.dds"},
	{0x3C8CD063540B8F74ULL, "vxlt_hangedman_209_0_0.vlm"},
	{0x3C8CDDD24BEFD5ABULL, "vxw_bonepitdestroyed_a3_1xh_0_0.vlm"},
	{0x3C8CF15F86C615A9ULL, "vxdr_0_deeproads_bur_0_0.vlm"},
	{0x3C8D5879B732755BULL, "6137697_m.fxe"},
	{0x3C8D7E730F9415ABULL, "lowtown_fem_alien_elf_2_mob6166443.cl"},
	{0x3C8E202604C7AA6EULL, "337923951.ogg"},
	{0x3C8EA6B53B3E68B9ULL, "6125124_m.fxe"},
	{0x3C8F3F23A8121D6AULL, "hlac_lt_0_lowtown_docks_n_-2_5y.dds"},
	{0x3C8F48BF6CC8015AULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc104a_0n.dds"},
	{0x3C8F5E5D0E6B8DFEULL, "vxlt_0_lowtown_a3_n_c_eoc_0_0.vlm"},
	{0x3C8F613B01E36428ULL, "1066800629.ogg"},
	{0x3C8F6D9AFABBE6CAULL, "hlac_lt_0_lowtown_docks_n_196609_0x.dds"},
	{0x3C8F8DA803BDBB2CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wo_galowint.msh"},
	{0x3C8F93569858693FULL, "dae021_sister_lorena_sta.bnk"},
	{0x3C8F93A803BDC51EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wo_galowint.mmh"},
	{0x3C8F99606110470BULL, "39349292.ogg"},
	{0x3C8F9FA803BDD9B5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wo_galowint.mao"},
	{0x3C901D95060CCEECULL, "916838403.ogg"},
	{0x3C90503E940E6720ULL, "vxdr_0_deeproads_ecf_0_0.vlm"},
	{0x3C913835361E662CULL, "vrc221_varric6040824.cl"},
	{0x3C91487E49E91BEFULL, "vxlt_gallowscourtyard_66s_0_0.vlm"},
	{0x3C914C18865F7184ULL, "hlac_lt_0_lowtown_a2_n_c_65536_2y.dds"},
	{0x3C918C95F24449F8ULL, "vxht_keep_n_6hd_0_0.vlm"},
	{0x3C91D9F3EA0C360EULL, "vxca_caverns1_mines_iih_0_0.vlm"},
	{0x3C91EA5D15983E54ULL, "vxlt_gallowscourtyard_n_7fr_0_0.vlm"},
	{0x3C921A03C71D31C9ULL, "6127060_f.fxe"},
	{0x3C92210238D2EF6AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg_b.mmh"},
	{0x3C925D2F7EF372ABULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_act_c.impa.gad"},
	{0x3C927556985A6C47ULL, "dae021_sister_lorena_sta.cnv"},
	{0x3C92BA4B14D48D3AULL, "vxlt_undercityserialkiller_1c2_0_0.vlm"},
	{0x3C92CCA96456032EULL, "1015238732.ogg"},
	{0x3C92DD1BA779DA35ULL, "6119826_m.fxe"},
	{0x3C935F7E97906ACEULL, "51012791.ogg"},
	{0x3C9395B4C52A221AULL, "llac_lt_0_lowtown_docks_-131073_0.dds"},
	{0x3C93B833DE91B451ULL, "vxht_mansion2_2b8_0_0.vlm"},
	{0x3C93D073CBD8D587ULL, "986214968.ogg"},
	{0x3C944EE9B340C832ULL, "ht_0_hightownexplosion_n_42_0.rml"},
	{0x3C94E38B4F54B7A3ULL, "lgt210_magu_selby6142041.cl"},
	{0x3C94EC41DF503E29ULL, "terrain_blend1.psh.fxo"},
	{0x3C957679C3DC6374ULL, "ch1_nolit.psh.fxo"},
	{0x3C959506DD11628BULL, "476277233.ogg"},
	{0x3C960ABE5B3559ADULL, "960326779.ogg"},
	{0x3C96341E9DB79B4AULL, "llac_lt_gallowstemplar_n_2394_3.dds"},
	{0x3C963D2300F7D084ULL, "835576983.ogg"},
	{0x3C96714323B0CD82ULL, "dae000cr_grand_cleric6147658.cl"},
	{0x3C9680281302378DULL, "hlac_lt_gallowstemplar_f_2392_0z.dds"},
	{0x3C97275E6803E798ULL, "6082034_f.fxe"},
	{0x3C973DCE99AF43FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_draw.evt"},
	{0x3C97787974484303ULL, "mag320_orsino6071136.cl"},
	{0x3C979E454DC4558BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st02a_0n.dds"},
	{0x3C979F1F34D9B5BBULL, "gen00fl_aveline_qun_cris.stg"},
	{0x3C97C8B118A14D95ULL, "vxdr_0_deeproads_8wf_0_0.vlm"},
	{0x3C97CFF76E6765F2ULL, "isa111_isabela6075417.cl"},
	{0x3C98229CC98FB1FAULL, "mag110_samson6020788.cl"},
	{0x3C985A2F7EF7D910ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_act_c.impa.evt"},
	{0x3C9882CCFB611ABAULL, "6174016_f.fxe"},
	{0x3C98DD1CED65B40BULL, "art\\characters\\creatures\\stonegolem\\animation\\enter.ani"},
	{0x3C9A036AC3193E71ULL, "6125678_m.fxe"},
	{0x3C9A08EFCD9345DFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rug.phy"},
	{0x3C9A1584B1F60273ULL, "art\\vfx\\combat\\vfx_ha_hd_leap_act_c.impa.evt"},
	{0x3C9A448A517E712DULL, "vxca_cavern1_haunted_j3r_0_0.vlm"},
	{0x3C9AAA90163F4FC3ULL, "vxw_blightlands_fr_0_0.vlm"},
	{0x3C9AB7F2F70838D2ULL, "dre000_dougal6129712.cl"},
	{0x3C9AE3B86117A457ULL, "15559200.ogg"},
	{0x3C9B633C78BC2E29ULL, "vxdr_0_deeproads_bpq_0_0.vlm"},
	{0x3C9B8E44F0E4864EULL, "946366853.ogg"},
	{0x3C9B91876FB89114ULL, "static1_s1y.psh.fxo"},
	{0x3C9BE9CC443C95D9ULL, "vxca_caverns2prc_aax_0_0.vlm"},
	{0x3C9CDC26B3415366ULL, "6045113_m.fxe"},
	{0x3C9CEFAC2FECB079ULL, "zz_dae_debug6045449.cl"},
	{0x3C9CF39D91860EE0ULL, "hlac_ca_caverns1_24424_8y.dds"},
	{0x3C9D9D9E4E0C7F78ULL, "vxlt_0_lowtown_a2_emk_0_0.vlm"},
	{0x3C9DB913255CA979ULL, "75968161.ogg"},
	{0x3C9DC2527D7CEF63ULL, "character_character_p192.psh.fxo.dx11"},
	{0x3C9E1ECE99B52DE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_draw.gad"},
	{0x3C9E78BBEE0151A7ULL, "473234581.ogg"},
	{0x3C9E9F73E121A7EFULL, "vxht_0_hightown_n_bm4_0_0.vlm"},
	{0x3C9EC9E1E96ECC5DULL, "vxht_0_hightown_a3_a2r_0_0.vlm"},
	{0x3C9ECD04FE1AD6EEULL, "art\\levels\\races\\elf\\props\\_textures\\prp_post_sa.dds"},
	{0x3C9ED37EB195F62EULL, "6084983_m.fxe"},
	{0x3C9EDF4E9D3D3E34ULL, "2181330.ogg"},
	{0x3C9EFCA18AF4480CULL, "234720710.ogg"},
	{0x3C9F4EDBC2F739FDULL, "art\\characters\\playercharacter\\humanmale\\glv_lgte.phy"},
	{0x3C9F84260A3C37AEULL, "vxca_caverns1_i6h_0_0.vlm"},
	{0x3C9FFE72E43A6278ULL, "vxht_keep_n_6ce_0_0.vlm"},
	{0x3CA052051C2D3F4AULL, "hlac_lt_0_lowtown_a2_n_c_-262144_0z.dds"},
	{0x3CA05C3A078E56D4ULL, "vxlt_gallowscourtyard_n_7is_0_0.vlm"},
	{0x3CA0F2CC3574DB27ULL, "6176113_m.fxe"},
	{0x3CA10E84B1FC14B4ULL, "art\\vfx\\combat\\vfx_ha_hd_leap_act_c.impa.gad"},
	{0x3CA18C73C9FD151FULL, "character_character_p116.psh.fxo.dx11"},
	{0x3CA23522744D1F94ULL, "vxdr_primevaltomb_6m9_0_0.vlm"},
	{0x3CA23BB25E06810EULL, "615517544.ogg"},
	{0x3CA2B52CEA5C5427ULL, "art\\vfx\\environment\\vfx_prp_partycall_c.dura.gad"},
	{0x3CA2CCC6370A94A0ULL, "183201450.ogg"},
	{0x3CA2CF40A250630FULL, "dirty_bb.dds"},
	{0x3CA34A9B7B14B13CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyn_0s.dds"},
	{0x3CA388D555ADBA00ULL, "vxht_mansion2_1rz_0_0.vlm"},
	{0x3CA3A4E688B1FC59ULL, "vxlt_0_lowtown_a2_n_c_hjs_0_0.vlm"},
	{0x3CA3ACCD16398E7AULL, "dre100cr_blighted_carver.utc"},
	{0x3CA3FA85E1B1D636ULL, "sib100_father_contact6108453.cl"},
	{0x3CA4B94FA01BA7ABULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_cth_comx_0s.dds"},
	{0x3CA53A1B989E91E3ULL, "403640464.ogg"},
	{0x3CA5439A9CF07D91ULL, "253748998.ogg"},
	{0x3CA57403734CD5D6ULL, "pro000_friedrich6050188.cl"},
	{0x3CA5C2CE7F342E9DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timber45post.mao"},
	{0x3CA5C6CE7F3435D6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timber45post.mmh"},
	{0x3CA5D0CE7F3446D4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timber45post.msh"},
	{0x3CA6C51CED71C882ULL, "art\\characters\\creatures\\stonegolem\\animation\\enter.evt"},
	{0x3CA71D84B2009992ULL, "art\\vfx\\combat\\vfx_ha_hd_leap_act_c.impa.ani"},
	{0x3CA73E3A9B4FE97AULL, "art\\levels\\4cinematics\\ferelden\\prp_bookinquisbinside.mao"},
	{0x3CA7424B25514DD6ULL, "gen00fl_anders6119135.cl"},
	{0x3CA782BF8729DD99ULL, "vxlt_0_lowtown_a2_n_c_gz0_0_0.vlm"},
	{0x3CA7BE22108D3573ULL, "627896278.ogg"},
	{0x3CA8512E1AD0BEA8ULL, "6096942_m.fxe"},
	{0x3CA89AA485D1D717ULL, "hlac_lt_0_lowtown_a2_n_c_-1_7z.dds"},
	{0x3CA8AB7724BF3BBEULL, "character_rendertodepth_p6.vsh.fxo.dx11.assembly"},
	{0x3CA8B7E7FDDADEBBULL, "vxlt_warehouse_n_af_0_0.vlm"},
	{0x3CA8C37C454287FBULL, "832737829.ogg"},
	{0x3CA8F1DE488A6BA2ULL, "216695009.ogg"},
	{0x3CA93A90DD71C506ULL, "vxlt_0_lowtown_a3_n_c_emu_0_0.vlm"},
	{0x3CA970A292F7DCF2ULL, "vxlt_gallowstemplar_f_6sd_0_0.vlm"},
	{0x3CA97986A2EF3E36ULL, "mage_arcanesource.dds"},
	{0x3CA9AA2CEA626034ULL, "art\\vfx\\environment\\vfx_prp_partycall_c.dura.evt"},
	{0x3CA9B6CF5EBDDCC1ULL, "dae000_porfiria6156617.cl"},
	{0x3CA9F18E1B13525CULL, "297030625.ogg"},
	{0x3CAA7E1C55575F67ULL, "hlac_ca_caverns1_24421_12x.dds"},
	{0x3CAA8368F3D5EF52ULL, "6056838_m.fxe"},
	{0x3CAB06EE7584922AULL, "vxht_chantry_n_26x_0_0.vlm"},
	{0x3CAB09EFCDA1BAE4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rug.mmh"},
	{0x3CAB0DEFCDA1C137ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rug.mao"},
	{0x3CAB1FEFCDA1DF86ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rug.msh"},
	{0x3CAB3CB201F71AE9ULL, "142536318.ogg"},
	{0x3CAB75B7DE0518FEULL, "165185534.ogg"},
	{0x3CABB0A1A07195F0ULL, "vxht_0_hightown_c8j_0_0.vlm"},
	{0x3CAC420083553F41ULL, "183936038.ogg"},
	{0x3CACDE30978D93C9ULL, "46195093.ogg"},
	{0x3CACF5C45473A963ULL, "ser191_emeric6165790.cl"},
	{0x3CAD3E10BA4984C3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\cuddle_lying_female_loop.ani"},
	{0x3CADA6620BDA878BULL, "gui_guipixelshader_p5.psh.fxo.dx11"},
	{0x3CADC41CED77E595ULL, "art\\characters\\creatures\\stonegolem\\animation\\enter.gad"},
	{0x3CADE25FD8E298CBULL, "hlac_lt_0_lowtown_docks_a3_n_-3_1z.dds"},
	{0x3CAE32DC2DD9254AULL, "vxw_encounter3_fj_0_1.vlm"},
	{0x3CAF5BEF796829CAULL, "l_env_wal_lowtown_crowd_bed.bnk"},
	{0x3CAF60ED7309C5D9ULL, "dae000_varric_deeproads.stg"},
	{0x3CB043DBC3059A69ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgte.mao"},
	{0x3CB04FDBC305AE02ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgte.mmh"},
	{0x3CB051DBC305B1B8ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgte.msh"},
	{0x3CB0D310FB0E30DCULL, "215659286.ogg"},
	{0x3CB196A5309342DDULL, "hlac_lt_0_lowtown_a2_n_c_65537_1z.dds"},
	{0x3CB1ABBBC987C29BULL, "793103756.ogg"},
	{0x3CB20085B767394DULL, "rdr110dg_guard_warehouse6016719.cl"},
	{0x3CB21CFD30824313ULL, "77639470.ogg"},
	{0x3CB2290DBEEF0F96ULL, "6130082_m.fxe"},
	{0x3CB25F725AD9A046ULL, "1819441.ogg"},
	{0x3CB2ECB5E2AF6987ULL, "llac_w_bonepit_0_10.dds"},
	{0x3CB318516945FF52ULL, "qcr220_orsino6061361.cl"},
	{0x3CB371428DAABE15ULL, "vxlt_0_lowtown_n_eqe_0_0.vlm"},
	{0x3CB38AC761194613ULL, "one300_zevran6089093.cl"},
	{0x3CB3CA36ACB4A145ULL, "vxlt_undercityencounter_12r_0_0.vlm"},
	{0x3CB42F5AF5BDB6D6ULL, "645150595.ogg"},
	{0x3CB528AE5F773A9BULL, "art\\vfx\\models\\vfx_mer_sword.msh"},
	{0x3CB546AE5F776D85ULL, "art\\vfx\\models\\vfx_mer_sword.mmh"},
	{0x3CB666EB44F12F51ULL, "vxw_bonepitdestroyed_a3_17k_0_0.vlm"},
	{0x3CB6698769FA72FAULL, "366674174.ogg"},
	{0x3CB6D133CD568BE5ULL, "mag210_arianni6167143.cl"},
	{0x3CB6EF4179421253ULL, "character_character_p88.psh.fxo"},
	{0x3CB70333993E9C98ULL, "115451715.ogg"},
	{0x3CB7406AE8D138AAULL, "6175135_m.fxe"},
	{0x3CB78831E57D83D6ULL, "mcr370_negotiation6080202.cl"},
	{0x3CB78E49FA2B80C3ULL, "vxw_blightlands_lt_0_0.vlm"},
	{0x3CB796DEF8B7BC3DULL, "follower_banter6160825.cl"},
	{0x3CB7A22CEA6E906DULL, "art\\vfx\\environment\\vfx_prp_partycall_c.dura.ani"},
	{0x3CB83260ABD86903ULL, "vxw_bonepitmining_a2_267_0_0.vlm"},
	{0x3CB8336921D21F2BULL, "vxw_bonepitdestroyed_a3_1km_0_0.vlm"},
	{0x3CB866756EA05F84ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be05a.phy"},
	{0x3CB886E50C87B330ULL, "gen00fl_anders6135777.cl"},
	{0x3CB99B3F392DD5AAULL, "hlac_lt_gallowstemplar_n_2390_1x.dds"},
	{0x3CBA7EA7541BD962ULL, "hlac_dr_slavers_3214_10y.dds"},
	{0x3CBA935DFF3B8F39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_left.ani"},
	{0x3CBAB3F3E44DD77EULL, "vxlt_0_lowtown_a3_n_c_exf_0_0.vlm"},
	{0x3CBABD28399D458CULL, "dre000_sandal6033753.cl"},
	{0x3CBADDCFA83C85A2ULL, "6088918_m.fxe"},
	{0x3CBB2BEA546C3889ULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.mmh"},
	{0x3CBB316E3115F313ULL, "355538947.ogg"},
	{0x3CBB65CC5A26E709ULL, "qun220cr_javaris6071898.cl"},
	{0x3CBB74492C181647ULL, "vxlt_0_lowtown_a3_n_c_h20_0_0.vlm"},
	{0x3CBB776BEC9BBC68ULL, "llac_dr_primevaltomb_659_8.dds"},
	{0x3CBBCD53EA4F23BDULL, "195640307.ogg"},
	{0x3CBC4D11090284DDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic_1.mao"},
	{0x3CBC96CD5B23C063ULL, "117254690.ogg"},
	{0x3CBC9B0AF5C7B357ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_d_ink_d.dds"},
	{0x3CBD1B69BCC4271BULL, "follower_banter6163260.cl"},
	{0x3CBD73D3BE9CDFF9ULL, "w_sundermount_145_0.msh"},
	{0x3CBD7DD3BE9CF0E7ULL, "w_sundermount_145_0.mmh"},
	{0x3CBD81D3BE9CF734ULL, "w_sundermount_145_0.mao"},
	{0x3CBD911B73891A2FULL, "338137105.ogg"},
	{0x3CBDC47BE6E8312DULL, "6177857_m.fxe"},
	{0x3CBDECE32A448F75ULL, "6074459_m.fxe"},
	{0x3CBE8D6B35371EBBULL, "lowtown_male_lowlife2_mob6166769.cl"},
	{0x3CBE96CE50E3DDFDULL, "custic_light_f2.dds"},
	{0x3CBEAF28A5BAB047ULL, "121391565.ogg"},
	{0x3CBECD3D123153E0ULL, "vxht_mansion_n_6rz_0_0.vlm"},
	{0x3CBF01B3AEE42450ULL, "vxca_caverns1_mines_eb7_0_0.vlm"},
	{0x3CBF53BCC21A8837ULL, "6108794_m.fxe"},
	{0x3CBFD8E005C3E097ULL, "6027628_m.fxe"},
	{0x3CC02FB04E584E4CULL, "gen00fl_anders6103897.cl"},
	{0x3CC06C8F3999EB00ULL, "vxht_mansion2_1xx_0_0.vlm"},
	{0x3CC107BF83C5AE55ULL, "mcr371_scared_mage_m6086907.cl"},
	{0x3CC15510BA5A54ADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\cuddle_lying_female_loop.gad"},
	{0x3CC16E5DFF416F7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_left.gad"},
	{0x3CC192E5C999F52AULL, "mcr310_argument6074429.cl"},
	{0x3CC22CF2DC01FE41ULL, "dae000_arishok6072871.cl"},
	{0x3CC2D63C52B6927CULL, "one100dg_vanard6020962.cl"},
	{0x3CC477188878CCE7ULL, "374229420.ogg"},
	{0x3CC4874F217BFD65ULL, "277268456.ogg"},
	{0x3CC489A32313E2E5ULL, "hlac_dr_0_deeproads_5841_0y.dds"},
	{0x3CC497A5E47C9BB2ULL, "vxw_encounter2_camp_t_0_0.vlm"},
	{0x3CC4D8C836E747D1ULL, "vxw_bonepitdestroyed_a3_18j_0_0.vlm"},
	{0x3CC53CB3708B1ED1ULL, "539672388.ogg"},
	{0x3CC5611E6B382AD3ULL, "character_character_p79.psh.fxo"},
	{0x3CC5BCBEC12D84CAULL, "6095105_m.fxe"},
	{0x3CC60026EC219943ULL, "vxw_blightlands_qw_0_0.vlm"},
	{0x3CC76CAD8C0D9463ULL, "674553450.ogg"},
	{0x3CC781F09B567FE7ULL, "6060423_m.fxe"},
	{0x3CC78B5DFF460C60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_left.evt"},
	{0x3CC84DFF03B66825ULL, "6108763_m.fxe"},
	{0x3CC8BC859FA3F832ULL, "vxw_bonepit_29y_0_0.vlm"},
	{0x3CC8BF00CA7EA4BFULL, "cd_mcr310_sibling.bnk"},
	{0x3CC8C2166197D6DCULL, "dre000_vampire6064156.cl"},
	{0x3CC8E0CB16C7A94DULL, "6185504_m.fxe"},
	{0x3CC8E898A14CF535ULL, "6080197_m.fxe"},
	{0x3CC8FA72ADD23306ULL, "ht_0_hightown_a2_16_0_lc.mmh"},
	{0x3CC8FC5DA39554DBULL, "270812933.ogg"},
	{0x3CC94FAC9A329E22ULL, "6083818_m.fxe"},
	{0x3CC9F27DFB0EFE42ULL, "gen00fl_fenris6113900.cl"},
	{0x3CCA38378B868E5EULL, "hlac_ht_keep_n_3732_17y.dds"},
	{0x3CCACE5FB991E534ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep.phy"},
	{0x3CCBC1ECAAD1661EULL, "vxca_caverns1_mines_gir_0_0.vlm"},
	{0x3CCBEEF7749BEACCULL, "329447913.ogg"},
	{0x3CCBF52FCA743105ULL, "vxlt_0_lowtown_a2_emr_0_0.vlm"},
	{0x3CCC5E96F89973C5ULL, "vxlt_0_lowtown_docks_a3_d2h_0_0.vlm"},
	{0x3CCC9A68F94F1282ULL, "llac_ht_playershouse_n_855_21.dds"},
	{0x3CCCF5A715C5A042ULL, "fx_tent_64frm.dds"},
	{0x3CCCF9F34CE32121ULL, "6076871_m.fxe"},
	{0x3CCDCC655876E2A8ULL, "232182589.ogg"},
	{0x3CCE99742CDAB4B3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffold.phy"},
	{0x3CCF5F6EE46BA9A4ULL, "qcr000_arishok6087520.cl"},
	{0x3CCFC267393346A6ULL, "857845593.ogg"},
	{0x3CD06AA803F516B9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wo_galowint.phy"},
	{0x3CD0867464EC71F6ULL, "391922699.ogg"},
	{0x3CD115CF0A5C5563ULL, "hlac_w_woundedcoast_7_0y.dds"},
	{0x3CD14346C120AA63ULL, "hlac_w_sundermount_65539_0z.dds"},
	{0x3CD178E255464622ULL, "llac_ca_caverns1_mines_24425_7.dds"},
	{0x3CD185543FC5FD7EULL, "163483883.ogg"},
	{0x3CD22EAEE3A752FAULL, "156840042.ogg"},
	{0x3CD2672EACA0C828ULL, "aip_coteriethug2w_base.gda"},
	{0x3CD2B1E676DE4156ULL, "894091814.ogg"},
	{0x3CD438459EB14827ULL, "vxlt_backalley_g_1z8_0_0.vlm"},
	{0x3CD46324164213D4ULL, "plt_fen000pt_gift_sword.nss"},
	{0x3CD4AB5FB999FE31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep.msh"},
	{0x3CD4B95FB99A164CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep.mao"},
	{0x3CD4C55FB99A2AEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_leftkeep.mmh"},
	{0x3CD4D6957F23C3B6ULL, "186284797.ogg"},
	{0x3CD51D9CEFCCCB7EULL, "vxlt_0_lowtown_gjz_0_0.vlm"},
	{0x3CD55E336DB46B03ULL, "vxlt_0_lowtown_a2_n_ffu_0_0.vlm"},
	{0x3CD5F143FE8E46CBULL, "qcr000_arishok6060126.cl"},
	{0x3CD6B08818340C99ULL, "hlac_w_blightlands_262144_0y.dds"},
	{0x3CD6BDC03B25C152ULL, "478548530.ogg"},
	{0x3CD6FF2F115F81FFULL, "dae000ip_crowd_ff7_talk3.utp"},
	{0x3CD769CBB23E5B5CULL, "339949472.ogg"},
	{0x3CD8D44B619F220FULL, "651153866.ogg"},
	{0x3CD8F71107C312E8ULL, "6034025_m.fxe"},
	{0x3CD942A5639407B1ULL, "vxlt_0_lowtown_docks_n_c4e_0_0.vlm"},
	{0x3CD9DC645821E0C0ULL, "hlac_lt_gallowstemplar_f_2394_5z.dds"},
	{0x3CD9E14BB31D4597ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_homa.msh"},
	{0x3CD9F34BB31D64FAULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_homa.mao"},
	{0x3CD9F74BB31D6B09ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_homa.mmh"},
	{0x3CDA1E0D8055FD09ULL, "888605686.ogg"},
	{0x3CDA8E6B892B18F7ULL, "vxlt_warehouse_n_ar_0_0.vlm"},
	{0x3CDA93922E06EFB4ULL, "75393627.ogg"},
	{0x3CDABD461144D64DULL, "76959001.ogg"},
	{0x3CDB92F95CB4ADECULL, "523137221.ogg"},
	{0x3CDC1F0EEAFC7C19ULL, "6208692_m.fxe"},
	{0x3CDC938D70533771ULL, "llac_ht_0_hightown_a2_65537_0.dds"},
	{0x3CDC9BC26CEF5433ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep.phy"},
	{0x3CDCEE8FDFEEED95ULL, "710454707.ogg"},
	{0x3CDD50491D861C00ULL, "vxht_mansion2_1ff_0_0.vlm"},
	{0x3CDD6C5A508A5E59ULL, "vxlt_0_lowtown_a2_n_c_hvw_0_0.vlm"},
	{0x3CDD81654E0FBE87ULL, "6171318_m.fxe"},
	{0x3CDE4A4C481AF39DULL, "61494049.ogg"},
	{0x3CDF0E0CEDBC57EDULL, "vxlt_warehouse_n_b9_0_0.vlm"},
	{0x3CDF271945CA60B4ULL, "6108098_m.fxe"},
	{0x3CDF7FC859807576ULL, "701976864.ogg"},
	{0x3CDF97B225CF9042ULL, "w_bonepitdestroyed_a3_95_0.msh"},
	{0x3CDF99B225CF93F8ULL, "w_bonepitdestroyed_a3_95_0.mmh"},
	{0x3CDFA5B225CFA79BULL, "w_bonepitdestroyed_a3_95_0.mao"},
	{0x3CDFD0B534ADB0AFULL, "mag370_meredith6046834.cl"},
	{0x3CE0679D745E5E56ULL, "6070816_m.fxe"},
	{0x3CE0E577AAAFE007ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy.gad"},
	{0x3CE10B5C0DE60692ULL, "gen00fl_anders6104021.cl"},
	{0x3CE1BF1F1D27B3D3ULL, "6204962_m.fxe"},
	{0x3CE1C9CE15776CE5ULL, "6076844_m.fxe"},
	{0x3CE1E15CFA7E6FF3ULL, "llac_ca_caverns2_5802_40.dds"},
	{0x3CE248FF42BA00A1ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_104a_sa.dds"},
	{0x3CE28658FA1E0BD5ULL, "6070481_f.fxe"},
	{0x3CE298CD1F0CB507ULL, "dae000_loyal_mage6165239.cl"},
	{0x3CE3310782C10045ULL, "mag160_feynriel6020453.cl"},
	{0x3CE378739C67D8A5ULL, "nat380_sibling6067988.cl"},
	{0x3CE3EC4D40E11C47ULL, "543522868.ogg"},
	{0x3CE428EA548F38BEULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.anb"},
	{0x3CE42B2F8838E61FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_minehammer_1.mao"},
	{0x3CE4979953C64804ULL, "vxw_bonepitmining_a2_7k_0_0.vlm"},
	{0x3CE4BAFAE624AB50ULL, "plt_prm_cod_frozen_boots.plo"},
	{0x3CE4C5F644B32080ULL, "344290725.ogg"},
	{0x3CE4E981F7B35B20ULL, "llac_ht_0_hightown_a3_-65535_1.dds"},
	{0x3CE53B1B3C456262ULL, "864885638.ogg"},
	{0x3CE54DE4B0E05D90ULL, "vxlt_foundry_n_w1_0_0.vlm"},
	{0x3CE56F59E655F65EULL, "540740943.ogg"},
	{0x3CE580193DB97A60ULL, "zz_approval_debug6003669.cl"},
	{0x3CE5A985DA5BDC80ULL, "556061188.ogg"},
	{0x3CE5EFE98370C3FDULL, "dae000_junar6147932.cl"},
	{0x3CE60F9FBFC82F64ULL, "325006543.ogg"},
	{0x3CE635855EA79EB9ULL, "gen00fl_fenris6110896.cl"},
	{0x3CE6A083C3DD2C24ULL, "close_pressed.cur"},
	{0x3CE6A44288B24615ULL, "413199530.ogg"},
	{0x3CE70A830E130097ULL, "vxht_playershouse_n_1qz_0_0.vlm"},
	{0x3CE7568F32BB2D08ULL, "art\\levels\\ztools\\default\\generic\\2cp_basic.mmh"},
	{0x3CE7628F32BB41ABULL, "art\\levels\\ztools\\default\\generic\\2cp_basic.mao"},
	{0x3CE7748F32BB6032ULL, "art\\levels\\ztools\\default\\generic\\2cp_basic.msh"},
	{0x3CE77A8796462055ULL, "gen00fl_merrill6116032.cl"},
	{0x3CE7DA77AAB5EC14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy.evt"},
	{0x3CE86175392DC405ULL, "lgt210_qun_blk_qunari6142303.cl"},
	{0x3CE8B6B366DEF89AULL, "hlac_lt_0_lowtown_a2_n_65537_4x.dds"},
	{0x3CE9278B46E0FA69ULL, "rdr111dg_martin6053595.cl"},
	{0x3CE9A7EFBD4D0751ULL, "gen00fl_merrill6146943.cl"},
	{0x3CE9E84BA61A8F9BULL, "689842680.ogg"},
	{0x3CE9EC740241EF8CULL, "hlac_lt_0_lowtown_docks_a3_-2_4z.dds"},
	{0x3CEA89B7D46FD49DULL, "dae000_arishok6068498.cl"},
	{0x3CEAF5B032857E63ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_trapdoor_sa.dds"},
	{0x3CEB2D2699BAD312ULL, "vxlt_0_lowtown_f1r_0_0.vlm"},
	{0x3CEB654281DB788BULL, "qun100cr_qunari_warrior2.utc"},
	{0x3CEB91B76C61A6D7ULL, "vxht_0_hightown_a3_bs8_0_0.vlm"},
	{0x3CEB94B60BC5C551ULL, "llac_ht_keep_3739_0.dds"},
	{0x3CEC10D63EF7EFDBULL, "6115470_f.fxe"},
	{0x3CECE0C94B10294DULL, "531086716.ogg"},
	{0x3CED317624A27E8DULL, "hlac_lt_0_lowtown_a2_0_7y.dds"},
	{0x3CED3EBD3BF7EEA2ULL, "203649882.ogg"},
	{0x3CEDB9516E0F2A67ULL, "lgt320_shw_ors_marlein6154743.cl"},
	{0x3CEDEC3778267EE7ULL, "hlac_ca_caverns1_24426_7y.dds"},
	{0x3CEDEDEEE4802E99ULL, "gen00fl_merrill6106994.cl"},
	{0x3CEDEF9D45FABD7AULL, "vxlt_undercityencounter_1mp_0_0.vlm"},
	{0x3CEECEABC9998301ULL, "art\\characters\\creatures\\pridedemon\\animation\\dblw_lp.gad"},
	{0x3CEEED7C7CBC0016ULL, "91636063.ogg"},
	{0x3CEF0A8ABE6DA301ULL, "zz_otl_debug6060951.cl"},
	{0x3CF0509F90959C68ULL, "vxlt_0_lowtown_docks_n_apn_0_0.vlm"},
	{0x3CF054F8DBF07F15ULL, "mag311_keran6041475.cl"},
	{0x3CF088713AD544DFULL, "vxdr_0_deeproads_egi_0_0.vlm"},
	{0x3CF0917CEE31BE0BULL, "426405642.ogg"},
	{0x3CF17AFAE15FAD1CULL, "9612930.ogg"},
	{0x3CF186F18A3536A5ULL, "fen110_party_comment6095070.cl"},
	{0x3CF1E8E296E52EAAULL, "44444505.ogg"},
	{0x3CF23E74E05DA501ULL, "325325210.ogg"},
	{0x3CF2A75B02C9FF55ULL, "vxlt_0_lowtown_a2_f4n_0_0.vlm"},
	{0x3CF2DA2228F1B107ULL, "fen210_orana6109696.cl"},
	{0x3CF31329834A2CF1ULL, "ht_chantry_probe_2a4.dds"},
	{0x3CF330C084018D4EULL, "zz_dre_debug6092003.cl"},
	{0x3CF3602851E4FE8FULL, "985432671.ogg"},
	{0x3CF398728FE6A8DBULL, "6171101_m.fxe"},
	{0x3CF3D8B6A8742AC4ULL, "qcr270st_fake_meredith.stg"},
	{0x3CF3E88A4556725BULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em05.phy"},
	{0x3CF430FD2254C5C8ULL, "gen_im_trs_gem_diamond_04.uti"},
	{0x3CF43BFC25DB7230ULL, "969760351.ogg"},
	{0x3CF44EB0ECEBD1FFULL, "515782911.ogg"},
	{0x3CF4A500CAA3B3F5ULL, "cd_mcr310_sibling.wbd"},
	{0x3CF4C05A5ED4A23FULL, "61674779.ogg"},
	{0x3CF4F3E76AB0F1B5ULL, "weapon_weapon_p0.vsh.fxo"},
	{0x3CF55F2F790522B4ULL, "vxht_0_hightown_n_626_0_0.vlm"},
	{0x3CF5CFABC99FA3AFULL, "art\\characters\\creatures\\pridedemon\\animation\\dblw_lp.ani"},
	{0x3CF5D277AAC21C4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_rdy.ani"},
	{0x3CF5E1F1816453ABULL, "lgt310_mag_melson_guards6153016.cl"},
	{0x3CF6820CAF4ADD73ULL, "vxdr_dungeon_21t_0_0.vlm"},
	{0x3CF6C7D6FECDB6B3ULL, "gen_im_trs_gen_sta_fir_r0.uti"},
	{0x3CF740440A5DDA27ULL, "gen00fl_anders6201147.cl"},
	{0x3CF7FA501EF9258EULL, "6148805_m.fxe"},
	{0x3CF80B3CC90BD047ULL, "vxw_encounter2_6a_0_0.vlm"},
	{0x3CF84E3FA67041F4ULL, "hlac_lt_undercityencounter_-1_12y.dds"},
	{0x3CF8844B335271ACULL, "6069442_f.fxe"},
	{0x3CF894B0DA09B5E0ULL, "vxht_mansion_n_6ff_0_0.vlm"},
	{0x3CF89CCCE8C0BDDAULL, "305359498.ogg"},
	{0x3CF947B945B26F0CULL, "hlac_w_blightlands_196609_0x.dds"},
	{0x3CF9BFCF7B344788ULL, "static_static_p12.psh.fxo.dx11"},
	{0x3CF9E1CC92926885ULL, "6170153_m.fxe"},
	{0x3CF9F5B8CD9AC215ULL, "181342851.ogg"},
	{0x3CF9FD09C188E1C6ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1_lp.ani"},
	{0x3CFA03445911868CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_rng_v2.ani"},
	{0x3CFA9999A0C26FA6ULL, "dae000_pc_home_ambients6208352.cl"},
	{0x3CFAEF868B8942D5ULL, "dae000_meredith6092588.cl"},
	{0x3CFB25DC2C72C3ECULL, "331607130.ogg"},
	{0x3CFB37725EAEDDDAULL, "vxw_bonepitmining_a2_es_0_0.vlm"},
	{0x3CFB618BA478245EULL, "6081210_m.fxe"},
	{0x3CFB99E56E0B4ABAULL, "zz_vault_debug6177761.cl"},
	{0x3CFC6EF635741E58ULL, "69559170.ogg"},
	{0x3CFCAE349B7C8189ULL, "6057074_m.fxe"},
	{0x3CFCE7753D2E39F0ULL, "231366814.ogg"},
	{0x3CFD440734B3C68BULL, "vxlt_gallowstemplar_n_5pg_0_0.vlm"},
	{0x3CFE4AA758D9662BULL, "vxdr_primevaltomb_8vm_0_0.vlm"},
	{0x3CFEA69E6C97E326ULL, "vxw_encounter2_camp_p3_0_1.vlm"},
	{0x3CFEC7EB8B5246E7ULL, "mcr310_confused_guard6077090.cl"},
	{0x3CFF15FC2D679B36ULL, "6095071_m.fxe"},
	{0x3CFF30EE5A647E75ULL, "gen_im_wep_war_two_lore_02.uti"},
	{0x3CFF3E11E2B3BA29ULL, "zz_dae_debug6049940.cl"},
	{0x3CFF68833C593012ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.mmh"},
	{0x3CFFB7D1E22C3B56ULL, "6048514_m.fxe"},
	{0x3CFFB9F0243336F8ULL, "one220_yevhen6115733.cl"},
	{0x3CFFD55B48CFB8ADULL, "6051146_m.fxe"},
	{0x3D00CECE11FAAD7FULL, "86339411.ogg"},
	{0x3D010715D5EF76ABULL, "crate_03_a.dds"},
	{0x3D0110445917BBCEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_rng_v2.gad"},
	{0x3D01349F35ECFC12ULL, "gen00fl_merrill6150985.cl"},
	{0x3D017B5337B27904ULL, "vxw_bonepitmining_a2_9i_0_0.vlm"},
	{0x3D017E7871BEF1FCULL, "121418708.ogg"},
	{0x3D02065ECCB75B70ULL, "974135335.ogg"},
	{0x3D027656398A8BBAULL, "81079012.ogg"},
	{0x3D03A410861CBCCDULL, "682462816.ogg"},
	{0x3D0486CB5BF3A212ULL, "mag330_walter6044818.cl"},
	{0x3D04FB632BC1063CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_chantryfloor_sa.dds"},
	{0x3D05649056455089ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtd.phy"},
	{0x3D0598B6247AD6D4ULL, "vxw_encounter3_gf_0_0.vlm"},
	{0x3D068006C7A2262BULL, "764823081.ogg"},
	{0x3D06C3D6B3B867BCULL, "vxlt_0_lowtown_a2_n_c_gpo_0_0.vlm"},
	{0x3D07068DCBFFCE60ULL, "vxht_mansion_n_6ig_0_0.vlm"},
	{0x3D073AA40889FB33ULL, "656329417.ogg"},
	{0x3D07792CC79E8C85ULL, "hlac_lt_darktown_-1_66y.dds"},
	{0x3D07EF759974073AULL, "zz_mrl_debug6184216.cl"},
	{0x3D080B44591DD1ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_rng_v2.evt"},
	{0x3D088D99BA9D65D1ULL, "6041712_m.fxe"},
	{0x3D08A23E639972B2ULL, "6056720_m.fxe"},
	{0x3D08E6E5A9967268ULL, "vxlt_warehouse_o6_0_0.vlm"},
	{0x3D0A7E6DBAA6A737ULL, "404540596.ogg"},
	{0x3D0AA05402D683BDULL, "vxdr_primevalentrance_b1_0_0.vlm"},
	{0x3D0AC9AC696BD9BBULL, "982232429.ogg"},
	{0x3D0ACAAB87F0F19AULL, "ser191_emeric6104712.cl"},
	{0x3D0AF436EAF7B221ULL, "6165421_m.fxe"},
	{0x3D0B3B89C3F060AAULL, "6050196_f.fxe"},
	{0x3D0B94BF1229774DULL, "hlac_lt_undercityencounter_-1_46y.dds"},
	{0x3D0C2D10362DBD13ULL, "one100dg_kelder6060539.cl"},
	{0x3D0C752D77052AF8ULL, "76016679.ogg"},
	{0x3D0C8A236A300184ULL, "vxlt_0_lowtown_a2_n_c_eqg_0_1.vlm"},
	{0x3D0D5F696D20E58FULL, "qun211_party_comment6072171.cl"},
	{0x3D0D6EDF23FB67CAULL, "911828111.ogg"},
	{0x3D0E185A7422F375ULL, "139852649.ogg"},
	{0x3D0E32D4E9FAED52ULL, "406800431.ogg"},
	{0x3D0E843E4571FB7EULL, "hlac_lt_0_lowtown_docks_n_-131073_3z.dds"},
	{0x3D0EE609C19B16C0ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1_lp.gad"},
	{0x3D0EECE763EEE2E0ULL, "6131134_f.fxe"},
	{0x3D0EF343650C2B8CULL, "w_bonepitmining_a2_202_0.msh"},
	{0x3D0EF943650C367EULL, "w_bonepitmining_a2_202_0.mmh"},
	{0x3D0EFCDD4596C605ULL, "zz_followers_banter6158280.cl"},
	{0x3D0F0543650C4A15ULL, "w_bonepitmining_a2_202_0.mao"},
	{0x3D0FB645F397EB3BULL, "llac_lt_0_lowtown_a3_n_c_-65536_0.dds"},
	{0x3D0FF0B16589A426ULL, "583786201.ogg"},
	{0x3D0FF2C8E7C8CB00ULL, "6188670_m.fxe"},
	{0x3D1009643336E94FULL, "vxlt_0_lowtown_a2_fgi_0_0.vlm"},
	{0x3D1014E59ECA1F81ULL, "170702378.ogg"},
	{0x3D1036746546DFDFULL, "272553805.ogg"},
	{0x3D1060474149E82EULL, "1059254064.ogg"},
	{0x3D10939C4F4ECCE3ULL, "159836805.ogg"},
	{0x3D10E8102BBA8150ULL, "vxht_0_hightown_a2_bj1_0_0.vlm"},
	{0x3D111C4DD2025DEDULL, "vxht_chantry_13x_0_0.vlm"},
	{0x3D113320BFA4BB13ULL, "qun110_petrice6059147.cl"},
	{0x3D11794981D2FA13ULL, "one140cr_charade6128492.cl"},
	{0x3D119BA5BF9A4D15ULL, "lgt110_orz_harrowmont6109013.cl"},
	{0x3D11E41627E5E0B6ULL, "100430152.ogg"},
	{0x3D126A5F42EC2156ULL, "vxdr_0_deeproads_esp_0_0.vlm"},
	{0x3D133512746DD6DEULL, "vxdr_primevaltomb_7ti_0_0.vlm"},
	{0x3D13573341D4CD30ULL, "66515744.ogg"},
	{0x3D137EE1CDF27074ULL, "6124283_m.fxe"},
	{0x3D13A44CD00CA487ULL, "vxlt_0_lowtown_a2_gn7_0_0.vlm"},
	{0x3D140F6AAF665783ULL, "llac_lt_0_lowtown_a2_-2_5.dds"},
	{0x3D143E198A92A2DAULL, "llac_lt_darktown_-1_62.dds"},
	{0x3D1487198F078E50ULL, "follower_banter6169201.cl"},
	{0x3D14AAF94BC3CB86ULL, "6101086_m.fxe"},
	{0x3D1523B3B7355265ULL, "hairalpha_rendertodepth_p0.psh.fxo.dx11.assembly"},
	{0x3D15685715DDBC6EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowera.msh"},
	{0x3D15765715DDD3BFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowera.mao"},
	{0x3D15786ABDF5E6E0ULL, "vxht_mansion_n_6ld_0_0.vlm"},
	{0x3D15825715DDE85CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowera.mmh"},
	{0x3D15ACD0B289B65EULL, "6060524_f.fxe"},
	{0x3D16586AEB212259ULL, "lgt351cr_shw_mer_sympathize.utc"},
	{0x3D166EA2961D0A7DULL, "vxht_0_hightown_n_ask_0_0.vlm"},
	{0x3D168C040956B663ULL, "11073016.ogg"},
	{0x3D16CEBA3145F657ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_min_c.impa.gad"},
	{0x3D16D23341564CA0ULL, "312570385.ogg"},
	{0x3D16F64B5F38E44EULL, "hlac_ca_cavern1_haunted_24423_7z.dds"},
	{0x3D175E9E2192554FULL, "gen_im_trs_apo_rng_r1.uti"},
	{0x3D181B2C429B0EDAULL, "vxw_bonepitmining_a2_om_0_0.vlm"},
	{0x3D18AD3FDB53E8DEULL, "art\\characters\\creatures\\dragon\\animation\\crn_rdy_rf.evt"},
	{0x3D18BACB281FF4C2ULL, "vxlt_0_lowtown_a2_n_e12_0_0.vlm"},
	{0x3D18D580B594B9EDULL, "vxlt_warehouse_n_v5_0_0.vlm"},
	{0x3D1908AE3A3F08B9ULL, "gen00fl_anders6103748.cl"},
	{0x3D19B4FABC3BEDB3ULL, "6051491_f.fxe"},
	{0x3D19E8A39F0FDE9CULL, "vxlt_warehouse_10n_0_0.vlm"},
	{0x3D19EA0EC7C7B7D8ULL, "6098841_m.fxe"},
	{0x3D1A31D60382B480ULL, "6056740_m.fxe"},
	{0x3D1A4B0FD2610FBCULL, "lowtown_fem_urchin_mob6165907.cl"},
	{0x3D1AC0CABDE3A9E5ULL, "em_dead5.mor"},
	{0x3D1AD11A0BAED8DEULL, "mer221_sabin6047459.cl"},
	{0x3D1B78EE227D67D8ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_swordb.phy"},
	{0x3D1BA6CEE6E9760FULL, "vxlt_gallowsprison_n_6hp_0_0.vlm"},
	{0x3D1BBA61CB2A8CDDULL, "vxht_mansion2_2ei_0_0.vlm"},
	{0x3D1BF0251AF9E5F4ULL, "89234734.ogg"},
	{0x3D1C635A242CA68FULL, "286593884.ogg"},
	{0x3D1DE3BA314C38E4ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_min_c.impa.evt"},
	{0x3D1E71B6715F71C1ULL, "dae000_keep_male_servant_mo6161460.cl"},
	{0x3D1F1662E03B8320ULL, "560675621.ogg"},
	{0x3D1F87F4FF7F55ACULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he4a.phy"},
	{0x3D1F9A0A065CEC9DULL, "6058888_m.fxe"},
	{0x3D1FC43FDB5A2E09ULL, "art\\characters\\creatures\\dragon\\animation\\crn_rdy_rf.gad"},
	{0x3D1FCF8A457B82FAULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em05.msh"},
	{0x3D1FD778C7308ED1ULL, "vxlt_gallowscourtyard_7bd_0_0.vlm"},
	{0x3D1FE18A457BA140ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em05.mmh"},
	{0x3D2098DF8DF9F9EFULL, "994164471.ogg"},
	{0x3D21085DF828FD5EULL, "hlac_ht_chantry_2534_8y.dds"},
	{0x3D21953C1D980351ULL, "301678653.ogg"},
	{0x3D21FDB427447634ULL, "383855507.ogg"},
	{0x3D220F8EB58360DCULL, "6155799_m.fxe"},
	{0x3D223937379786B6ULL, "ip_restrict_follower_carver.ncs"},
	{0x3D22DBDB7AF7C0D6ULL, "6033606_m.fxe"},
	{0x3D232799DF097E83ULL, "4599266.ogg"},
	{0x3D2335C4B2465557ULL, "vxht_mansion2_2ll_0_0.vlm"},
	{0x3D233D97A5F6F23EULL, "vxca_caverns1_mines_j3t_0_1.vlm"},
	{0x3D233DE4DFB13239ULL, "atl_deathscree_dxt1_00.dds"},
	{0x3D23753A35ECD959ULL, "vxlt_warehouse_158_0_0.vlm"},
	{0x3D2377DE7565AF25ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.mmh"},
	{0x3D237BB21CDED2D1ULL, "6033991_m.fxe"},
	{0x3D23CC293ED5374CULL, "521296670.ogg"},
	{0x3D2433DD47A07CD4ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_5y.dds"},
	{0x3D24799B5D9AB53EULL, "787967110.ogg"},
	{0x3D2485783EF10641ULL, "vxlt_0_lowtown_a2_n_c_hkh_0_0.vlm"},
	{0x3D25B91E6D199860ULL, "92920042.ogg"},
	{0x3D268D093491275AULL, "vxw_bonepitmining_a2_jn_0_0.vlm"},
	{0x3D26953FDB5FFD57ULL, "art\\characters\\creatures\\dragon\\animation\\crn_rdy_rf.ani"},
	{0x3D27591E3C656545ULL, "vxlt_undercityencounter_1jz_0_0.vlm"},
	{0x3D2799425487BE8AULL, "365965934.ogg"},
	{0x3D27A4754496A00CULL, "vxlt_darktown_60h_0_0.vlm"},
	{0x3D27BC89073964E5ULL, "gencr_coterie_thug_ws_f.utc"},
	{0x3D2867833C7C33B9ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.anb"},
	{0x3D28999E0A96100BULL, "vxlt_gallowstemplar_n_5ax_0_0.vlm"},
	{0x3D28B0E37A2D5A8BULL, "6180615_m.fxe"},
	{0x3D2939BF2C2F6794ULL, "646503073.ogg"},
	{0x3D293A01C3F6F279ULL, "gen_im_cns_flu_fenris_03.uti"},
	{0x3D295CD9CD373CB7ULL, "vxlt_gallowsprison_n_64d_0_0.vlm"},
	{0x3D297B845A20053AULL, "vis160cr_winter_rogue.utc"},
	{0x3D29AA84EFBAE4E8ULL, "292152744.ogg"},
	{0x3D29FF7D4AB49BECULL, "hlac_lt_gallowstemplar_2391_1z.dds"},
	{0x3D2A2C3EBD20A55DULL, "vxht_mansion2_2nh_0_0.vlm"},
	{0x3D2A6A96642121B9ULL, "qcr221_cut_diplomatic_sol6201291.cl"},
	{0x3D2B3D523A20575CULL, "259929411.ogg"},
	{0x3D2BFBBA31589EDDULL, "art\\vfx\\combat\\vfx_glb_ice_imp_min_c.impa.ani"},
	{0x3D2CB408E5ED7F01ULL, "vxht_mansion2_1rw_0_0.vlm"},
	{0x3D2CD3862DACE11CULL, "llac_ht_chantry_2534_29.dds"},
	{0x3D2D324EB5D45887ULL, "6060306_m.fxe"},
	{0x3D2DF901BBDF9316ULL, "tools_camera.msh"},
	{0x3D2E0301BBDFA414ULL, "tools_camera.mmh"},
	{0x3D2ECE0A6FF8D42FULL, "6110650_m.fxe"},
	{0x3D2F56A8CF690544ULL, "38079586.ogg"},
	{0x3D2F706FAEBB9C7AULL, "6120942_m.fxe"},
	{0x3D2FEDBECFE9BE7BULL, "dae000_mother6121041.cl"},
	{0x3D300A3446DFD0DBULL, "vxca_cavern1_haunted_gk9_0_0.vlm"},
	{0x3D3056901F6BCE10ULL, "dae000_fem_priest_mob6174194.cl"},
	{0x3D3074DE757034F2ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.anb"},
	{0x3D3124969D59BD1EULL, "404281708.ogg"},
	{0x3D313ABB0F4528F8ULL, "1021951511.ogg"},
	{0x3D314D90566A64AEULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtd.mmh"},
	{0x3D315990566A7945ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtd.mao"},
	{0x3D316790566A909CULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtd.msh"},
	{0x3D3192742D2F2628ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffold.mmh"},
	{0x3D319E742D2F3ACBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffold.mao"},
	{0x3D31B0742D2F5952ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffold.msh"},
	{0x3D31B2EE8FB86CFAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin.mmh"},
	{0x3D31B6EE8FB87331ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin.mao"},
	{0x3D31C4EE8FB88B40ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin.msh"},
	{0x3D32260CBB2866C7ULL, "vxw_blightlands_lh_0_0.vlm"},
	{0x3D32A90130C9EE5FULL, "em_fenarel.mor"},
	{0x3D32BEE55A6AE625ULL, "vrc221_varric6044166.cl"},
	{0x3D33027E9B07F47AULL, "rock_large_normals.dds"},
	{0x3D33563B2305DC06ULL, "6052777_m.fxe"},
	{0x3D33790DFAC7DF99ULL, "hlac_ca_caverns1_24421_2y.dds"},
	{0x3D338265F0E0BEC9ULL, "hlac_ht_0_hightown_a2_-196608_1x.dds"},
	{0x3D3383024DE02101ULL, "lt_0_lowtown_27_0.rml"},
	{0x3D33D82BFD4428B5ULL, "llac_lt_0_lowtown_a2_-131071_2.dds"},
	{0x3D349813F998060AULL, "6164029_m.fxe"},
	{0x3D34F0D86B8DE07FULL, "vxht_0_hightown_a3_b9f_0_0.vlm"},
	{0x3D350C9AF46F9F8DULL, "6166297_m.fxe"},
	{0x3D35A51D79228725ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_4z.dds"},
	{0x3D35C8E82521D30CULL, "391712951.ogg"},
	{0x3D35D69B773C7F1CULL, "6070224_f.fxe"},
	{0x3D35FD669DC356CDULL, "w_bonepitmining_a2_23_0.mmh"},
	{0x3D3601669DC35D1EULL, "w_bonepitmining_a2_23_0.mao"},
	{0x3D360F669DC37553ULL, "w_bonepitmining_a2_23_0.msh"},
	{0x3D362E4785D84B23ULL, "vxht_0_hightown_a3_9uu_0_0.vlm"},
	{0x3D364AC426B645CEULL, "53954499.ogg"},
	{0x3D3670A3B589E1BCULL, "qun111_ketojan.cnv"},
	{0x3D36DBA467601A84ULL, "646341136.ogg"},
	{0x3D37486258BF4FDDULL, "6096752_m.fxe"},
	{0x3D37BC342C3191EEULL, "6037017_m.fxe"},
	{0x3D382D38F313A6C9ULL, "hlac_lt_0_lowtown_docks_-196610_0y.dds"},
	{0x3D3874ADC28C3B04ULL, "441262472.ogg"},
	{0x3D38FA7D5857A817ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_harrowing_d.dds"},
	{0x3D39A868AFA07C38ULL, "gen00fl_isabela6079066.cl"},
	{0x3D3A0D3DDCD3A56FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_maincavern.mao"},
	{0x3D3A17320ECF5377ULL, "621205245.ogg"},
	{0x3D3A193DDCD3B90CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_maincavern.mmh"},
	{0x3D3A1F3DDCD3C3FEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_maincavern.msh"},
	{0x3D3A8279AAEE18FDULL, "ico_shortbow_4.dds"},
	{0x3D3A92A3B58E040CULL, "qun111_ketojan.bnk"},
	{0x3D3AA508386921CFULL, "w_bonepitdestroyed_a3_220_0.mao"},
	{0x3D3AB1083869366CULL, "w_bonepitdestroyed_a3_220_0.mmh"},
	{0x3D3AB7083869405EULL, "w_bonepitdestroyed_a3_220_0.msh"},
	{0x3D3AC48C68C81209ULL, "character_character_p86.psh.fxo"},
	{0x3D3AE5BFA1AE8A02ULL, "vxlt_shop_ea_0_0.vlm"},
	{0x3D3B25E5D7E39781ULL, "vxht_mansion2_1uv_0_0.vlm"},
	{0x3D3B9839793E5BD7ULL, "dae200cr_viscount6087250.cl"},
	{0x3D3C457651DAD0FBULL, "6116252_m.fxe"},
	{0x3D3CBB32AB1CBFD1ULL, "vxlt_gallowscourtyard_7hz_0_0.vlm"},
	{0x3D3CC6D2D895F6BFULL, "961184631.ogg"},
	{0x3D3CE80A186DC849ULL, "vxht_0_hightown_n_by6_0_0.vlm"},
	{0x3D3CEDC3C2B9F9D5ULL, "ser120_ghyslain6104203.cl"},
	{0x3D3D769676A40247ULL, "951248950.ogg"},
	{0x3D3D7882DE836895ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_sand01_d.dds"},
	{0x3D3D7F79F734087CULL, "273332149.ogg"},
	{0x3D3DBFF618CE6AD6ULL, "vxdr_0_deeproads_ebs_0_0.vlm"},
	{0x3D3DC50E2E65107DULL, "398147317.ogg"},
	{0x3D3DDCD914EEA92AULL, "hlac_lt_backalley_g_0_0z.dds"},
	{0x3D3E7FF609A271C3ULL, "fen210_orana6109784.cl"},
	{0x3D3E96ECE2D98E9CULL, "796878438.ogg"},
	{0x3D3ED9695A1832A7ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgtp_0d.dds"},
	{0x3D3EE68F08DBDFFDULL, "vxdr_primevalentrance_2oj_0_0.vlm"},
	{0x3D3F197BA05D6832ULL, "166364105.ogg"},
	{0x3D3F2AD266A02E37ULL, "6033150_f.fxe"},
	{0x3D3F94C26D43C5A8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep.mmh"},
	{0x3D3FA0C26D43DA4BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep.mao"},
	{0x3D3FB2C26D43F8D2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep.msh"},
	{0x3D3FBF955EE3F1D6ULL, "6039406_m.fxe"},
	{0x3D3FFD669DCBAB46ULL, "w_bonepitmining_a2_23_0.rml"},
	{0x3D40215007A01F6CULL, "zz_followers_banter6158049.cl"},
	{0x3D402F0274701C21ULL, "vxlt_gallowstemplar_f_46f_0_0.vlm"},
	{0x3D405458B19039C1ULL, "llac_ca_caverns1_mines_24423_11.dds"},
	{0x3D4097E9AD1E7F47ULL, "vxw_blightlands_qk_0_0.vlm"},
	{0x3D40B217D72165E2ULL, "718089801.ogg"},
	{0x3D40E3D3365D3E0AULL, "mcr370_cut_player_survives6171835.cl"},
	{0x3D4141D53312CE83ULL, "300681636.ogg"},
	{0x3D4222BAC7038D69ULL, "6083145_m.fxe"},
	{0x3D422AA314DF76DAULL, "gen_im_wep_war_one_act2_05.uti"},
	{0x3D424EB40128E6F6ULL, "vxlt_0_lowtown_docks_a3_n_buk_0_0.vlm"},
	{0x3D42C83C8450AFD6ULL, "hlac_ca_caverns2_5802_72x.dds"},
	{0x3D430800918660EFULL, "144517566.ogg"},
	{0x3D43256ED8021937ULL, "vxht_0_hightown_a3_bl1_0_0.vlm"},
	{0x3D4362B55D83F8FFULL, "vxht_0_hightown_a3_b4a_0_0.vlm"},
	{0x3D458CB8B2B507E4ULL, "hlac_ht_chantry_n_2534_52y.dds"},
	{0x3D463AA808011F13ULL, "da2terrain_terrain_p1.psh.fxo"},
	{0x3D46505B7E1D13F9ULL, "atl_login_dxt5_dat.xml"},
	{0x3D469B715F1DD5EFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bigwheel_n.dds"},
	{0x3D469C39BD1D2B47ULL, "hlac_lt_gallowstemplar_f_2392_0x.dds"},
	{0x3D46FDF1A0CC5D0CULL, "rob_rev.tnt"},
	{0x3D47321833246FA4ULL, "927068125.ogg"},
	{0x3D473D2149C9582CULL, "mag211_marethari6021295.cl"},
	{0x3D47C3B313FD5831ULL, "qun100_mage_party_comment6019112.cl"},
	{0x3D47DFD39A1782C4ULL, "dae000_orsino6085479.cl"},
	{0x3D484798CFDE4381ULL, "6070613_m.fxe"},
	{0x3D4936695ABE2A89ULL, "character_character_p77.psh.fxo"},
	{0x3D493A518434CDE4ULL, "art\\vfx\\combat\\vfx_gib_corm_rtl.ani"},
	{0x3D493A518434CDEFULL, "art\\vfx\\combat\\vfx_gib_corm_rtl.anb"},
	{0x3D4997C2C9D9B001ULL, "vxht_mansion2_1xu_0_0.vlm"},
	{0x3D4A06588E8C0759ULL, "lgt150_nat_were_supporter6111015.cl"},
	{0x3D4B54430113F634ULL, "hlac_ht_0_hightown_a3_-1_0z.dds"},
	{0x3D4B65D58D33F106ULL, "vxca_cavern1_haunted_j4l_0_0.vlm"},
	{0x3D4BC43EDB4626BCULL, "framebuffereffects_dofblend_p0.psh.fxo.dx11.assembly"},
	{0x3D4BCE158E7215F0ULL, "move.ncs"},
	{0x3D4BDE63B32F8A61ULL, "701945999.ogg"},
	{0x3D4BE41233031006ULL, "329888499.ogg"},
	{0x3D4BF1C27C5998D2ULL, "vxht_mansion2_3ce_0_0.vlm"},
	{0x3D4C31D30AC48356ULL, "vxdr_0_deeproads_eot_0_0.vlm"},
	{0x3D4C96817372897BULL, "ht_chantry_n_mmap.dds"},
	{0x3D4D1036B3BA5F39ULL, "265000765.ogg"},
	{0x3D4D94920E1C151BULL, "6056923_m.fxe"},
	{0x3D4DEE10EA36C8B4ULL, "art\\characters\\creatures\\dog\\textures\\c_eye_doga_0s.dds"},
	{0x3D4E0766D10A1CF3ULL, "gen_im_trs_krk_drk_junk1_03.uti"},
	{0x3D4E1EAA169EAC6DULL, "plt_bet000pt_main.nss"},
	{0x3D4ECFE204D28455ULL, "64037447.ogg"},
	{0x3D4F141635699923ULL, "darktown_fem_lowlife_mob6176172.cl"},
	{0x3D4F2873D61FAD6CULL, "hlac_lt_0_lowtown_docks_n_0_4y.dds"},
	{0x3D4FD04365438719ULL, "w_bonepitmining_a2_202_0.phy"},
	{0x3D4FEFEF8909C6DBULL, "rock_small_specular.dds"},
	{0x3D503751843AE746ULL, "art\\vfx\\combat\\vfx_gib_corm_rtl.gad"},
	{0x3D509B03A8DCA1E7ULL, "175403959.ogg"},
	{0x3D50A2E8EBCADF2EULL, "334178310.ogg"},
	{0x3D50D0A94ABCE537ULL, "vxht_0_hightown_a3_car_0_1.vlm"},
	{0x3D50D9D484B8B93AULL, "6060519_f.fxe"},
	{0x3D51000E66F165D4ULL, "vrc321_varric6087033.cl"},
	{0x3D514F79B1B303FFULL, "vxht_mansion_n_4v7_0_0.vlm"},
	{0x3D515A3BA2EC7A65ULL, "vxw_bonepitmining_a2_at_0_0.vlm"},
	{0x3D517AEF4B2E40CBULL, "132995668.ogg"},
	{0x3D51F831921F5595ULL, "docks_scene_dockworkers6176210.cl"},
	{0x3D5218720A9EEFC0ULL, "vxlt_0_lowtown_a2_cmq_0_0.vlm"},
	{0x3D5220BF93A5D5DEULL, "fen310_fenris6117680.cl"},
	{0x3D5255C0F33DCB95ULL, "vxlt_0_lowtown_n_e8j_0_0.vlm"},
	{0x3D529C3EF108AE49ULL, "49945495.ogg"},
	{0x3D554EB632DD1E5EULL, "dae200cr_viscount6087158.cl"},
	{0x3D5581CDAEAEA924ULL, "zz_mer_debug6073706.cl"},
	{0x3D5581D59302EEDDULL, "vxht_mansion2_2qe_0_0.vlm"},
	{0x3D55DB2E07F947A5ULL, "vxht_mansion2_1vw_0_0.vlm"},
	{0x3D56B592015E34D1ULL, "art\\vfx\\textures\\vfx_streak01.dds"},
	{0x3D56BD062ADE7E3CULL, "175035540.ogg"},
	{0x3D56D1EE8FD7BD05ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin.phy"},
	{0x3D56DD1ABCAA97FFULL, "506391832.ogg"},
	{0x3D56E68B776439D1ULL, "vxlt_0_lowtown_docks_bds_0_0.vlm"},
	{0x3D575ADCDBEB69BAULL, "6197229_m.fxe"},
	{0x3D57A8464CB44309ULL, "character_character_p24.psh.fxo"},
	{0x3D589CD19E85E532ULL, "mag360_alain6044651.cl"},
	{0x3D58E157F52E545FULL, "fireflipbookadd.mao"},
	{0x3D5976161B762F90ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.mmh"},
	{0x3D5A4260B4548547ULL, "779566818.ogg"},
	{0x3D5A639F6E4FB152ULL, "vxht_mansion2_3ff_0_0.vlm"},
	{0x3D5A8736D58E7164ULL, "zz_isa_debug6133221.cl"},
	{0x3D5AAF7799D846E0ULL, "322224348.ogg"},
	{0x3D5B41071864AA35ULL, "850793047.ogg"},
	{0x3D5B8AE6030E246CULL, "mag131_keran6086194.cl"},
	{0x3D5BB5B4D5E96789ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2webdetail.mao"},
	{0x3D5BB9CC80090C23ULL, "ene_star.mao"},
	{0x3D5BC1B4D5E97C22ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2webdetail.mmh"},
	{0x3D5BC3B4D5E97FD8ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2webdetail.msh"},
	{0x3D5BE172D377293CULL, "vxlt_gallowscourtyard_n_6rc_0_0.vlm"},
	{0x3D5C1BF65B2FF559ULL, "hlac_dr_0_deeproads_5845_0y.dds"},
	{0x3D5C244ACEECB5A1ULL, "zz_fen_debug6134333.cl"},
	{0x3D5C64C4284D9938ULL, "llac_lt_warehouse_n_65537_4.dds"},
	{0x3D5CC95519C25FC5ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_2z.dds"},
	{0x3D5CE6E7B18F3382ULL, "vxlt_gallowscourtyard_77d_0_0.vlm"},
	{0x3D5D6128433F6D59ULL, "196048742.ogg"},
	{0x3D5DEA8F32264680ULL, "vxlt_gallowscourtyard_3be_0_0.vlm"},
	{0x3D5DF71FC55D79BDULL, "vxlt_0_lowtown_a2_gvm_0_0.vlm"},
	{0x3D5DFC70CCD6F7E1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robc_0t.dds"},
	{0x3D5E0A6AAF6B9EB3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright_1.mao"},
	{0x3D5E5951D5C22228ULL, "6208600_m.fxe"},
	{0x3D5E7BFB889C9736ULL, "dae000_mother6135748.cl"},
	{0x3D5EC8A5D43F0905ULL, "vxht_keep_n_61b_0_0.vlm"},
	{0x3D5EC9E112F718A3ULL, "570399138.ogg"},
	{0x3D5EE187FB969560ULL, "mrl150_merrill6078602.cl"},
	{0x3D5EEEDAB9ABC567ULL, "6116564_m.fxe"},
	{0x3D5F00CDEE5AB056ULL, "256773297.ogg"},
	{0x3D5F454776078F18ULL, "vxlt_0_lowtown_eop_0_1.vlm"},
	{0x3D5F5FB2853E1915ULL, "495972203.ogg"},
	{0x3D5F6F1747C0C013ULL, "vxlt_undercityserialkiller_1il_0_1.vlm"},
	{0x3D5F8CBB7D356235ULL, "hlac_w_encounter1_-131071_0x.dds"},
	{0x3D5FA72D3075DC38ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_basic.phy"},
	{0x3D5FCC1894E292E5ULL, "vxw_bonepitmining_a2_ls_0_0.vlm"},
	{0x3D5FE69522F6F9F2ULL, "dae000ip_crowd_lm1_talk18.utp"},
	{0x3D5FE9123C23EE84ULL, "6044606_m.fxe"},
	{0x3D5FFEEF7685658DULL, "gen00fl_dog6167242.cl"},
	{0x3D6011044121107AULL, "vxlt_0_lowtown_docks_a3_n_aru_0_0.vlm"},
	{0x3D601E5E058F101EULL, "llac_lt_0_lowtown_a2_n_65535_0.dds"},
	{0x3D604C9B7B39658EULL, "vxlt_0_lowtown_a2_n_epa_0_0.vlm"},
	{0x3D605EF4FFB6A777ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he4a.mmh"},
	{0x3D6064416D99FE16ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtbzz.phy"},
	{0x3D6065788496F1C3ULL, "lgt261cr_thug.utc"},
	{0x3D6074F4FFB6CCE9ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he4a.msh"},
	{0x3D6095BD20D94996ULL, "llac_ht_mansion2_818_64.dds"},
	{0x3D60A28A9C114799ULL, "326219882.ogg"},
	{0x3D60AA60CB6F3E41ULL, "follower_banter6171265.cl"},
	{0x3D6148CCEFF31F64ULL, "art\\characters\\weapons\\unique\\aveline\\textures\\w_shd_009a_0d.dds"},
	{0x3D61E1B148ACD59EULL, "vxw_bonepitmining_a2_i3_0_0.vlm"},
	{0x3D6218B64CFE7E12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\blendtree\\templarseeker2_override.anb"},
	{0x3D6223F322A22EDAULL, "zz_modify_item.ncs"},
	{0x3D62525793ED064DULL, "vxlt_backalley_n_1ef_0_0.vlm"},
	{0x3D63A5AD17607C06ULL, "vxlt_0_lowtown_docks_a3_n_d03_0_1.vlm"},
	{0x3D63BD469C7A5551ULL, "vxw_bonepitdestroyed_a3_1sg_0_0.vlm"},
	{0x3D656106B8E0D7DEULL, "6110606_m.fxe"},
	{0x3D6599A9367B1CB6ULL, "165173171.ogg"},
	{0x3D659F63DF2B9257ULL, "671021992.ogg"},
	{0x3D665B1612A49F33ULL, "llac_lt_0_lowtown_docks_a3_131070_4.dds"},
	{0x3D6662024E0B4291ULL, "lt_0_lowtown_27_0.anb"},
	{0x3D667B7CADC5E101ULL, "vxht_mansion2_1fk_0_0.vlm"},
	{0x3D66AA0A0A78FB87ULL, "hla_ht_0_hightown_n_65538_0y.dds"},
	{0x3D66EC02E3CE80F8ULL, "6100264_f.fxe"},
	{0x3D66FF138E958FA9ULL, "vxlt_blackemporium_8j_0_0.vlm"},
	{0x3D674E622C5B6FDAULL, "6094533_m.fxe"},
	{0x3D678DB04AD8CAD7ULL, "gen00fl_anders6125243.cl"},
	{0x3D6799161B82A8AFULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.anb"},
	{0x3D683DDBA2065BBEULL, "6053885_f.fxe"},
	{0x3D688A7E13557357ULL, "llac_lt_gallowscourtyard_-131071_1.dds"},
	{0x3D68AE29083DDB6BULL, "mrl150_merrill6089650.cl"},
	{0x3D69341B26CA4DFAULL, "120672352.ogg"},
	{0x3D693451309E4F3FULL, "ts_darkspawn_02.gda"},
	{0x3D6957C1EEA4D699ULL, "6085837_m.fxe"},
	{0x3D6A4D416DA22BB7ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtbzz.msh"},
	{0x3D6A63416DA25129ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtbzz.mmh"},
	{0x3D6AC7DCD7CB67CDULL, "56291226.ogg"},
	{0x3D6AF4E7FC8D43A1ULL, "art\\characters\\playercharacter\\children\\textures\\kf_cth_grla_0t.dds"},
	{0x3D6B0C6B31A0F436ULL, "644840383.ogg"},
	{0x3D6B1E52F0133E80ULL, "vxlt_0_lowtown_a2_fe0_0_0.vlm"},
	{0x3D6B2D1991431BA2ULL, "zz_ave_debug6064621.cl"},
	{0x3D6B58C4A3854C02ULL, "vxlt_gallowscourtyard_78e_0_0.vlm"},
	{0x3D6B590856D93BA7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_hammer_n.dds"},
	{0x3D6BE002FD24EA44ULL, "6167242_m.fxe"},
	{0x3D6C68FCB753923DULL, "vxlt_0_lowtown_a2_gml_0_0.vlm"},
	{0x3D6D028D256324E7ULL, "789327021.ogg"},
	{0x3D6D8A984DCA82DCULL, "320132010.ogg"},
	{0x3D6E2EE3C410E00CULL, "hairalpha_hair_p1.vsh.fxo.dx11.assembly"},
	{0x3D6E3DF586D8AB65ULL, "vxw_bonepitmining_a2_kr_0_0.vlm"},
	{0x3D6EBE786D2F7E0EULL, "vxlt_0_lowtown_a2_n_eub_0_0.vlm"},
	{0x3D6EC342B8509748ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggera.mmh"},
	{0x3D6ECF42B850ABEBULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggera.mao"},
	{0x3D6EE142B850CA72ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggera.msh"},
	{0x3D6F7EEB6A58EDA2ULL, "w_encounter2_camp_14_0.rml"},
	{0x3D6F926EF633DFC5ULL, "vxlt_undercityencounter_1se_0_0.vlm"},
	{0x3D6FF5BB4DB0AD23ULL, "vxht_0_hightown_a3_9ay_0_0.vlm"},
	{0x3D7121DD05BAEF41ULL, "hlac_lt_0_lowtown_docks_a3_65535_2x.dds"},
	{0x3D712C96554D447FULL, "vxlt_0_lowtown_a2_n_h1z_0_0.vlm"},
	{0x3D71885CDA9609F2ULL, "vxlt_gallowstemplar_n_6qf_0_0.vlm"},
	{0x3D7223FB5824516CULL, "hlac_lt_gallowstemplar_f_7198_0z.dds"},
	{0x3D72383093086963ULL, "hf_ibethany_h1s2.mor"},
	{0x3D72BB503F998317ULL, "323350452.ogg"},
	{0x3D731C69F722D0FBULL, "gen00fl_merrill6162129.cl"},
	{0x3D7369EF2104DF64ULL, "vxht_playershouse_n_1rl_0_0.vlm"},
	{0x3D73E369DD76C5A2ULL, "105289753.ogg"},
	{0x3D73F10CDD89F91DULL, "417291056.ogg"},
	{0x3D740DA49F45817BULL, "vrc321_party_barks6095775.cl"},
	{0x3D746F59817DD937ULL, "658230582.ogg"},
	{0x3D74ED599FBBF981ULL, "vxht_mansion2_1ij_0_0.vlm"},
	{0x3D756A754563770EULL, "lla_ht_0_hightown_131072_0.dds"},
	{0x3D757FEB6A5D5B56ULL, "w_encounter2_camp_14_0.phy"},
	{0x3D75D6C9B8634D4DULL, "vxlt_backalley_g_1pm_0_1.vlm"},
	{0x3D75F51134AD083EULL, "hlac_lt_0_lowtown_a3_n_c_65538_4x.dds"},
	{0x3D7636438DD3AF26ULL, "6034298_m.fxe"},
	{0x3D76634E8E2C023DULL, "439351693.ogg"},
	{0x3D7690E23F9FCBF6ULL, "434807176.ogg"},
	{0x3D769880AA759925ULL, "zz_seb_debug6185073.cl"},
	{0x3D76E76DC712757EULL, "bad_med.mao"},
	{0x3D771C6A53161F01ULL, "dae000ip_crowd_mf8_talk18.utp"},
	{0x3D77233209FA6204ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_sleeping.anb"},
	{0x3D7765EE22CBA66DULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_swordb.msh"},
	{0x3D7773EE22CBBDA0ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_swordb.mao"},
	{0x3D7777EE22CBC4F3ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_swordb.mmh"},
	{0x3D7778DEB1CE3665ULL, "vxht_keep_1ab_0_0.vlm"},
	{0x3D77BF5BD3ED59B3ULL, "6053946_m.fxe"},
	{0x3D77D845E780E895ULL, "w_encounter3_83_0.phy"},
	{0x3D78C908519B46F1ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonc.mmh"},
	{0x3D78D508519B5A92ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonc.mao"},
	{0x3D78E308519B72AFULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonc.msh"},
	{0x3D78F7C3096DBE0EULL, "vxca_caverns1_mines_i9x_0_0.vlm"},
	{0x3D79D4BCC3A8EB49ULL, "llac_lt_0_lowtown_a2_n_c_-65537_0.dds"},
	{0x3D7A44EC2A73B6F0ULL, "6100322_m.fxe"},
	{0x3D7AFE0B857EE5E9ULL, "452093080.ogg"},
	{0x3D7B292DD9D4C272ULL, "vxht_0_hightown_a3_bu3_0_0.vlm"},
	{0x3D7BDD2E28A5694FULL, "hlac_ht_mansion2b_818_61z.dds"},
	{0x3D7C20FF5255B18FULL, "448481318.ogg"},
	{0x3D7C2AEF8D6E6033ULL, "rand_climax.gda"},
	{0x3D7C426941AEBFA6ULL, "follower_banter6169052.cl"},
	{0x3D7CA51087954D7FULL, "vxht_mansion_n_4g0_0_0.vlm"},
	{0x3D7DE1CD76246470ULL, "zz_lgt_1_debug6135726.cl"},
	{0x3D7E1648F8BB237EULL, "dae000_orana6149051.cl"},
	{0x3D7E1A267C33A089ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_blightfire.dura.ani"},
	{0x3D7E4D62E763EF24ULL, "6120070_m.fxe"},
	{0x3D7EA1C59DBFA820ULL, "1058099084.ogg"},
	{0x3D7F361177D9374DULL, "vxlt_backalley_n_1kd_0_0.vlm"},
	{0x3D7F36D60A33044EULL, "vxht_mansion_n_4d0_0_0.vlm"},
	{0x3D7F94A457D3A093ULL, "6104602_m.fxe"},
	{0x3D7FA10905875D68ULL, "vxlt_blackemporium_164_0_0.vlm"},
	{0x3D80043A2D39E3ADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12man.msh"},
	{0x3D80123A2D39FBE0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12man.mao"},
	{0x3D80163A2D3A0233ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12man.mmh"},
	{0x3D8053A1AD8E6FF1ULL, "hlac_ca_caverns1_24424_17x.dds"},
	{0x3D805769318AC408ULL, "219535627.ogg"},
	{0x3D8068EB6A673BF7ULL, "w_encounter2_camp_14_0.msh"},
	{0x3D807AEB6A675A5AULL, "w_encounter2_camp_14_0.mao"},
	{0x3D807EEB6A676169ULL, "w_encounter2_camp_14_0.mmh"},
	{0x3D811A667FCDA5F5ULL, "625101541.ogg"},
	{0x3D811FFF14CA2575ULL, "hlac_lt_gallowscourtyard_-65536_1x.dds"},
	{0x3D8130C4DDDB9125ULL, "vxht_mansion2_1gt_0_0.vlm"},
	{0x3D816CC7D95C087CULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower3_fenrintro.gad"},
	{0x3D819B806D7DD64FULL, "hlac_lt_0_lowtown_docks_a3_-196610_3y.dds"},
	{0x3D81F41B57464D0FULL, "515295619.ogg"},
	{0x3D8244ACE2C7D871ULL, "vxca_caverns1_mines_jdg_0_0.vlm"},
	{0x3D82B1322310EE89ULL, "518716467.ogg"},
	{0x3D839E0AEB4EC0DBULL, "store_dae221_bodahn.utm"},
	{0x3D849A6DE2F229CDULL, "31264472.ogg"},
	{0x3D84F7C8277915D9ULL, "vxw_woundedcoast_1gl_0_0.vlm"},
	{0x3D850C0505C9A984ULL, "17856070.ogg"},
	{0x3D850FE061F4A821ULL, "286809091.ogg"},
	{0x3D8535267C39ED0BULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_blightfire.dura.gad"},
	{0x3D85F7D1C4F8F09AULL, "vxht_chantry_2py_0_0.vlm"},
	{0x3D86437F71771D11ULL, "6095155_m.fxe"},
	{0x3D8673F7A9EB68A5ULL, "vxlt_0_lowtown_docks_a3_n_bv5_0_0.vlm"},
	{0x3D875D7253F0D049ULL, "6165436_m.fxe"},
	{0x3D8761F85580B3C2ULL, "mcr370_meredith6083044.cl"},
	{0x3D87B9B650BE39FBULL, "vxca_cavern1_haunted_jst_0_0.vlm"},
	{0x3D886BC7D962257AULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower3_fenrintro.ani"},
	{0x3D88AB580A6A91C3ULL, "gen00fl_aveline6121943.cl"},
	{0x3D897A44683EDA07ULL, "402485891.ogg"},
	{0x3D89CDE6BF75251BULL, "7950257.ogg"},
	{0x3D8A0201BC2E010FULL, "tools_camera.phy"},
	{0x3D8A0856C20E7242ULL, "vxlt_0_lowtown_docks_a3_n_cxa_0_0.vlm"},
	{0x3D8A246F13D64D27ULL, "llac_ht_0_hightown_n_131071_2.dds"},
	{0x3D8A85589CBC7495ULL, "hlac_ht_mansion2_818_38x.dds"},
	{0x3D8A8B247227EE44ULL, "835754941.ogg"},
	{0x3D8AD9E701B04DACULL, "vxca_caverns1_dpy_0_0.vlm"},
	{0x3D8AE5EB3DFACE3FULL, "6114845_m.fxe"},
	{0x3D8B0F1FDE8BAAB7ULL, "13075834.ogg"},
	{0x3D8BAB82F6D60D9BULL, "m_mcr371ar_galows_prison_nt.bnk"},
	{0x3D8BC6E9633DB4D2ULL, "hlac_lt_0_lowtown_docks_131068_0y.dds"},
	{0x3D8BC985E5FF842EULL, "268177443.ogg"},
	{0x3D8BDC1D0BBF6DB4ULL, "6124676_m.fxe"},
	{0x3D8C0F54EDB1DCF2ULL, "vxlt_0_lowtown_a2_n_hd4_0_0.vlm"},
	{0x3D8C32267C400770ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_blightfire.dura.evt"},
	{0x3D8C546FF2D5282EULL, "clf260cr_fell_orden.utc"},
	{0x3D8C740EE92C995CULL, "llac_ht_bartrand_n_3374_0.dds"},
	{0x3D8CC6EB47E1B9B2ULL, "vxlt_0_lowtown_docks_a3_d38_0_0.vlm"},
	{0x3D8D6D0900CAB989ULL, "vxht_mansion2b_2pv_0_0.vlm"},
	{0x3D8D847D97BFFAA2ULL, "gen00fl_varric6021120.cl"},
	{0x3D8DA7EE69CF4FCDULL, "vxlt_backalley_n_1vc_0_0.vlm"},
	{0x3D8DC314877001ABULL, "8327896.ogg"},
	{0x3D8DE1E966F6E78FULL, "6074062_m.fxe"},
	{0x3D8E4C24FB4C08FAULL, "llac_lt_undercityserialkiller_-1_6.dds"},
	{0x3D8E8D0BDB13CD5DULL, "vxlt_hangedman_59_0_0.vlm"},
	{0x3D8F0BB5B9478612ULL, "161392860.ogg"},
	{0x3D8F12DD725C9ED1ULL, "vxw_bonepitdestroyed_a3_1bd_0_0.vlm"},
	{0x3D8F1711C024A665ULL, "zz_dae_debug6029911.cl"},
	{0x3D8F1ADD7C69DE9BULL, "vxdr_primevalentrance_1m8_0_1.vlm"},
	{0x3D8FFE659C4EB277ULL, "art\\levels\\races\\proxy\\props\\_textures\\tent_n.dds"},
	{0x3D9027F972242C38ULL, "hlac_ht_0_hightown_0_1z.dds"},
	{0x3D902BD29ED16D8FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_htwood_n.dds"},
	{0x3D9046E6F9E822C5ULL, "llac_w_woundedcoast_131076_0.dds"},
	{0x3D904DD5C3EEC12AULL, "hlac_lt_warehouse_-65536_0y.dds"},
	{0x3D90B7724670775AULL, "vxlt_0_lowtown_docks_cw1_0_1.vlm"},
	{0x3D91AAEAC330AC8CULL, "986263389.ogg"},
	{0x3D92123FD6FF652BULL, "540471626.ogg"},
	{0x3D9323C751285F22ULL, "282874500.ogg"},
	{0x3D932D0889832E37ULL, "6000804_f.fxe"},
	{0x3D9380D4CDA40B91ULL, "gen00fl_varric6076037.cl"},
	{0x3D941EE5695744EFULL, "1072945180.ogg"},
	{0x3D94818DC1A73AC2ULL, "gen00fl_carver6097390.cl"},
	{0x3D9488EE2D67E181ULL, "937782229.ogg"},
	{0x3D94D0D4168A3E1BULL, "vxlt_gallowstemplar_53i_0_0.vlm"},
	{0x3D94FD6DF78AECDEULL, "99622707.ogg"},
	{0x3D95B84F0C90BFF0ULL, "gen00fl_aveline6152588.cl"},
	{0x3D960B579817CBD4ULL, "vxw_encounter3_5l_0_0.vlm"},
	{0x3D966727DA90DD2BULL, "660850091.ogg"},
	{0x3D966F44E2CBBA4FULL, "itemupgrade.gfx"},
	{0x3D97117C73C224F5ULL, "mrl150_merrill6072328.cl"},
	{0x3D978C3B00B87130ULL, "vxht_keep_70b_0_0.vlm"},
	{0x3D97A16BB67609BCULL, "plt_lgt000pt_ultra_2_02.nss"},
	{0x3D97B55A536A734FULL, "vxw_encounter1_bl_0_0.vlm"},
	{0x3D983F70E853EEA9ULL, "mag320_orsino6071109.cl"},
	{0x3D984FB5CF7225C8ULL, "6139968_m.fxe"},
	{0x3D99121B39CFB3FAULL, "vxht_0_hightown_a3_8pn_0_0.vlm"},
	{0x3D99C4544EC4D438ULL, "dae011_conspiracy_mob6149548.cl"},
	{0x3D99DA5A8C4C2542ULL, "zz_mcr_debug6091896.cl"},
	{0x3D9A390BBC60F0D1ULL, "6060403_m.fxe"},
	{0x3D9AA0FDB8AC5B04ULL, "81906183.ogg"},
	{0x3D9AA84FE8958983ULL, "gen00fl_aveline6131223.cl"},
	{0x3D9B0BE442CCE85FULL, "281018795.ogg"},
	{0x3D9B38C839D7D232ULL, "vxlt_0_lowtown_docks_a3_d49_0_0.vlm"},
	{0x3D9B5E7A7EA9F786ULL, "art\\levels\\races\\proxy\\props\\_textures\\shelf_d.dds"},
	{0x3D9BDEE5F2C0D209ULL, "vxht_mansion2b_2mu_0_0.vlm"},
	{0x3D9C40E8E7159786ULL, "542314412.ogg"},
	{0x3D9C53C658ED000FULL, "6079362_m.fxe"},
	{0x3D9C96F2B6316563ULL, "177355991.ogg"},
	{0x3D9D48A412EE1DCAULL, "gen00fl_aveline6106886.cl"},
	{0x3D9DDB1A6F088B76ULL, "vxht_bartrand_n_5so_0_0.vlm"},
	{0x3D9DDEA1A380F998ULL, "lla_ht_0_hightown_n_1_0.dds"},
	{0x3D9E3A1011E802F7ULL, "6100094_f.fxe"},
	{0x3D9EA325785C3B24ULL, "vxca_cave_2k6_0_0.vlm"},
	{0x3D9EDFAD39D12DDAULL, "llac_ht_0_hightown_a2_65533_3.dds"},
	{0x3D9EE0ADBF3AC903ULL, "77966501.ogg"},
	{0x3DA08C9155230027ULL, "vxht_brothel_n_4bu_0_1.vlm"},
	{0x3DA09F4565295340ULL, "vxdr_primevaltomb_4xp_0_0.vlm"},
	{0x3DA0D7B54EC6F75CULL, "hlac_lt_0_lowtown_a3_n_c_65537_5x.dds"},
	{0x3DA0D8766BD7A895ULL, "wm_ico_caravan.dds"},
	{0x3DA0FD5F22D55555ULL, "6111215_m.fxe"},
	{0x3DA16A5BB385C625ULL, "6206382_m.fxe"},
	{0x3DA1B10E652748D2ULL, "51296195.ogg"},
	{0x3DA2C7D0C13E75C6ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_butterflys.dds"},
	{0x3DA336B8E1F76915ULL, "6121949_f.fxe"},
	{0x3DA3EC83FA991DDFULL, "qun221_petrice6076707.cl"},
	{0x3DA417F7F47176A7ULL, "lt_0_lowtown_docks_a3_41_0.rml"},
	{0x3DA42A1191F2855CULL, "follower_banter6161160.cl"},
	{0x3DA4B4BEDC93590CULL, "9569959.ogg"},
	{0x3DA4EAAF2D9925EFULL, "dae000ip_crowd_mf1_talk6.utp"},
	{0x3DA585F7AC90B12CULL, "223300929.ogg"},
	{0x3DA60DAA16E9057EULL, "plt_bet000pt_main.plo"},
	{0x3DA6273745608BCFULL, "vxw_encounter1_go_0_0.vlm"},
	{0x3DA6E51265AFC56EULL, "mag320_mage_conspirator6040767.cl"},
	{0x3DA6FAFF28945503ULL, "hlac_w_blightlands_196606_0z.dds"},
	{0x3DA7F37C59687135ULL, "llac_lt_0_lowtown_a2_-2_3.dds"},
	{0x3DA82831D4426FA4ULL, "lgt160_ntb_elf_conflict6163363.cl"},
	{0x3DA8884D25748239ULL, "503946208.ogg"},
	{0x3DA89BFAE2C7AC17ULL, "vxdr_0_deeproads_evr_0_0.vlm"},
	{0x3DA8A3070C01EB7FULL, "6094521_m.fxe"},
	{0x3DA8FB6DDA7ED165ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb103a_0d.dds"},
	{0x3DA93F170090EFC9ULL, "follower_banter6151465.cl"},
	{0x3DA94A1AFEC2371FULL, "6137077_m.fxe"},
	{0x3DA97394F65BC882ULL, "vxht_mansion2_2b7_0_0.vlm"},
	{0x3DA9D484D3819C51ULL, "vxdr_primevaltomb_8ng_0_0.vlm"},
	{0x3DA9F284802550ABULL, "vxw_woundedcoast_yy_0_0.vlm"},
	{0x3DA9F9BB948E49CBULL, "6094011_m.fxe"},
	{0x3DAA52AA76AE8185ULL, "vxlt_undercitydungeon_174_0_0.vlm"},
	{0x3DAA7D94D88590D6ULL, "156610302.ogg"},
	{0x3DAA99B165CDFB97ULL, "3514774.ogg"},
	{0x3DAB3FBE7FA49054ULL, "follower_banter6169775.cl"},
	{0x3DAB549CF0033DCFULL, "6139303_m.fxe"},
	{0x3DABB8F7630A1A00ULL, "dae000_fenarel6127244.cl"},
	{0x3DABF6975648CFD1ULL, "vxw_bonepitdestroyed_a3_1hz_0_0.vlm"},
	{0x3DAC11BBFD77A59EULL, "follower_banter6164697.cl"},
	{0x3DACBF45EC69BA0CULL, "68051437.ogg"},
	{0x3DACCEC3E792D9D4ULL, "plt_cod_cha_varric2.plo"},
	{0x3DACDB99C6E3C846ULL, "6136710_m.fxe"},
	{0x3DACEE97D6EB2E7BULL, "20422582.ogg"},
	{0x3DAD02931B9A6D79ULL, "153805780.ogg"},
	{0x3DAD0FB054C7562BULL, "vxw_bonepit_1qo_0_0.vlm"},
	{0x3DAD15026A5253A4ULL, "vxca_cave_2l7_0_0.vlm"},
	{0x3DAD179502324AD4ULL, "llac_lt_0_lowtown_a2_n_c_131070_2.dds"},
	{0x3DAE139934EF3B4CULL, "6174488_m.fxe"},
	{0x3DAE1D62CA10994EULL, "hlac_lt_0_lowtown_docks_-1_0x.dds"},
	{0x3DAE998A69A772E5ULL, "hlac_lt_gallowscourtyard_-65535_0z.dds"},
	{0x3DAEC4D39D3565DAULL, "553098345.ogg"},
	{0x3DAEF8A3028B1F7DULL, "vxht_keep_n_5cs_0_0.vlm"},
	{0x3DAF37653E5DCA38ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_glow01.mao"},
	{0x3DAF6B1642EED0CCULL, "vxlt_hangedman_15t_0_1.vlm"},
	{0x3DAFA9B92872F36DULL, "lt_0_lowtown_docks_214_0.msh"},
	{0x3DAFB7B928730AA0ULL, "lt_0_lowtown_docks_214_0.mao"},
	{0x3DAFBBB9287311F3ULL, "lt_0_lowtown_docks_214_0.mmh"},
	{0x3DAFE77D8E66ADB3ULL, "6111049_f.fxe"},
	{0x3DB04437E861A37CULL, "llac_ca_caverns2prc_5802_62.dds"},
	{0x3DB0982AABDCBBCCULL, "vxlt_warehouse_cv_0_0.vlm"},
	{0x3DB1438ED4FF312EULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkback.ani"},
	{0x3DB156AA95553129ULL, "138150546.ogg"},
	{0x3DB2464EBE7B0DE0ULL, "818781166.ogg"},
	{0x3DB29BE0830DD1FCULL, "6147846_m.fxe"},
	{0x3DB2E251A1397771ULL, "hlac_lt_0_lowtown_docks_a3_131071_5z.dds"},
	{0x3DB323F3142D6388ULL, "mag100_cullen6082372.cl"},
	{0x3DB3BBEB697C0871ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robc_0d.dds"},
	{0x3DB3D2F8F67632DBULL, "ca_caverns1_mines_24427_ll.mmh"},
	{0x3DB3E4C867D56031ULL, "mag360_thrask6043523.cl"},
	{0x3DB44D3B7878A555ULL, "vxdr_primevalentrance_27j_0_0.vlm"},
	{0x3DB451E352D17840ULL, "6037077_m.fxe"},
	{0x3DB46B098F0E0AC0ULL, "205074026.ogg"},
	{0x3DB497EC4E3BE637ULL, "vxca_caverns2prc_9wt_0_0.vlm"},
	{0x3DB4BD919DFB8E96ULL, "llac_lt_0_lowtown_docks_n_65535_3.dds"},
	{0x3DB5196E1B3093DFULL, "240807426.ogg"},
	{0x3DB603F36556D37FULL, "480391219.ogg"},
	{0x3DB694E6944C2F4CULL, "hlac_ht_chantry_2534_37y.dds"},
	{0x3DB77DD34212A6DBULL, "230013075.ogg"},
	{0x3DB79126E1952296ULL, "hlac_ca_caverns1_mines_24426_5y.dds"},
	{0x3DB86461721B692BULL, "vxw_woundedcoast_vx_0_0.vlm"},
	{0x3DB8D34DCE8C5981ULL, "vxlt_gallowscourtyard_4r3_0_0.vlm"},
	{0x3DB911750D05404AULL, "197919424.ogg"},
	{0x3DB9529B16E0F759ULL, "ht_0_hightown_a3_37_0.anb"},
	{0x3DB95A10C4FE3F61ULL, "448771196.ogg"},
	{0x3DB9F4D6256FD968ULL, "hm_ave2_bandit.mor"},
	{0x3DBA44B6D55690FAULL, "287211300.ogg"},
	{0x3DBA5773AB42722DULL, "60571316.ogg"},
	{0x3DBAEB0D68090320ULL, "6195805_m.fxe"},
	{0x3DBAFD7203379661ULL, "gen00fl_merrill6105674.cl"},
	{0x3DBB818D46BD6EABULL, "vxw_bonepit_1zn_0_0.vlm"},
	{0x3DBB8F5315652752ULL, "288929451.ogg"},
	{0x3DBB942D30C41ACDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_basic.msh"},
	{0x3DBBA22D30C43200ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_basic.mao"},
	{0x3DBBA62D30C43953ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_basic.mmh"},
	{0x3DBBB67F3E350D5AULL, "rdr110dg_guard_warehouse6174119.cl"},
	{0x3DBC80FD4D7A9BE3ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_1y.dds"},
	{0x3DBC89C04B309AE8ULL, "336512901.ogg"},
	{0x3DBCB90AF1CF9D79ULL, "ser270_emeric6111470.cl"},
	{0x3DBCF9E7F539FE4DULL, "gen00fl_fenris6112050.cl"},
	{0x3DBD3ACE0D09E5E4ULL, "ttx_roof_crn_1.msh"},
	{0x3DBD4F7BD911BF96ULL, "gencr_circle_mage.utc"},
	{0x3DBD50CE0D0A0A86ULL, "ttx_roof_crn_1.mmh"},
	{0x3DBD6A7FF48137FDULL, "vxht_keep_n_5dr_0_0.vlm"},
	{0x3DBD732DAD66AD74ULL, "lowtown_male_lowlife1_sta6164958.cl"},
	{0x3DBD84CBA25D3777ULL, "6027673_m.fxe"},
	{0x3DBDD3392826B911ULL, "224043457.ogg"},
	{0x3DBEEB45E7BD0BB0ULL, "w_encounter3_83_0.msh"},
	{0x3DBEF945E7BD23CAULL, "w_encounter3_83_0.mmh"},
	{0x3DBEFD45E7BD2A01ULL, "w_encounter3_83_0.mao"},
	{0x3DBF3B8ED50B617FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkback.evt"},
	{0x3DBF76B9B5382909ULL, "gen00fl_isabela6090864.cl"},
	{0x3DBF93C0BD925BB3ULL, "hlac_lt_0_lowtown_docks_a3_n_131068_0z.dds"},
	{0x3DBFB4629930A131ULL, "art\\characters\\playercharacter\\children\\textures\\kf_cth_grla_0d.dds"},
	{0x3DBFC1F5E233B1D4ULL, "vxlt_0_lowtown_a2_euo_0_0.vlm"},
	{0x3DBFE144DD583A53ULL, "vxlt_gallowstemplar_n_46t_0_0.vlm"},
	{0x3DC095C317D44996ULL, "art\\characters\\creatures\\mubari\\animation\\choke.ani"},
	{0x3DC12692DA277A0BULL, "6016822_m.fxe"},
	{0x3DC1425D031D6D67ULL, "6056839_m.fxe"},
	{0x3DC167E914B85652ULL, "vxw_encounter3_de_0_0.vlm"},
	{0x3DC1AB76E0AC833CULL, "6028837_m.fxe"},
	{0x3DC1C9450F3B50E6ULL, "mrl150_hunter6071295.cl"},
	{0x3DC1D4EC717A7DE1ULL, "ht_0_hightown_9_0.phy"},
	{0x3DC214C09B6DF223ULL, "gen00fl_carver6102657.cl"},
	{0x3DC21E4A95622046ULL, "vxlt_ruinedhouse_n_60_0_0.vlm"},
	{0x3DC3387D20BA5CDBULL, "100588939.ogg"},
	{0x3DC34CC2DAA491BCULL, "302397044.ogg"},
	{0x3DC387B0AA79A370ULL, "mage_earthmastery.dds"},
	{0x3DC3AD422B7BF84FULL, "vxdr_primevaltomb_9cw_0_0.vlm"},
	{0x3DC3B74580866F68ULL, "zz_approval_debug6003962.cl"},
	{0x3DC3F42E6D6A0516ULL, "mag101_alain6053003.cl"},
	{0x3DC41736203E6F79ULL, "167017980.ogg"},
	{0x3DC45F567D16436DULL, "6162312_strong_girl.cut"},
	{0x3DC4C4CE1AC1017AULL, "137096587.ogg"},
	{0x3DC4F4700DA19C57ULL, "follower_banter6143154.cl"},
	{0x3DC53C8C720ED646ULL, "6101172_m.fxe"},
	{0x3DC559CA348820D3ULL, "316392713.ogg"},
	{0x3DC56CC3C8293CA7ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.dura.gad"},
	{0x3DC5A3CCF45C9719ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunarimage_movement.anb"},
	{0x3DC5D6DA240036A7ULL, "mcr370_meredith6136141.cl"},
	{0x3DC61C8ED5114B68ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkback.gad"},
	{0x3DC649773F9454A1ULL, "vxlt_0_lowtown_n_h72_0_0.vlm"},
	{0x3DC659104559D053ULL, "vxw_encounter3_ac_0_0.vlm"},
	{0x3DC69121B14DB36CULL, "qcr220_meredith6041543.cl"},
	{0x3DC6A1CEC07939ADULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_cgrass01_da.dds"},
	{0x3DC6B83EB76DCD51ULL, "vxdr_primevaltomb_8de_0_0.vlm"},
	{0x3DC7233AD4969803ULL, "244761140.ogg"},
	{0x3DC73ED82428E891ULL, "103877742.ogg"},
	{0x3DC77F24FDDC1AE6ULL, "vxca_caverns1_mines_gkp_0_0.vlm"},
	{0x3DC828A4FEEAF851ULL, "vrc321tr_servant_corpses.utt"},
	{0x3DC836BBE9A1A63AULL, "vxca_caverns1_mines_jfz_0_0.vlm"},
	{0x3DC863A947FD9C9EULL, "lt_0_lowtown_34_0_lc.mmh"},
	{0x3DC87D4871868D54ULL, "vxht_0_hightown_n_b4d_0_0.vlm"},
	{0x3DC8DDEC7180ABFDULL, "ht_0_hightown_9_0.rml"},
	{0x3DC931389867E2CDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_diningtable_n.dds"},
	{0x3DC96AAA91160ACBULL, "vxdr_0_deeproads_euo_0_0.vlm"},
	{0x3DC97C522CD861A5ULL, "lgt160cr_wk_avern_archer_f.utc"},
	{0x3DC99F37264B7C15ULL, "ser221_gascard6148834.cl"},
	{0x3DC9CAD74D52AC5EULL, "hlac_ca_cavern1_haunted_24426_7y.dds"},
	{0x3DCA3F690D3268D1ULL, "569027383.ogg"},
	{0x3DCA9CF0ADAA53CBULL, "gen_im_flw_shd_aveline_01.uti"},
	{0x3DCACE66086E2BB7ULL, "6132777_m.fxe"},
	{0x3DCAD2EB1AD85E4EULL, "vxlt_shop_dz_0_0.vlm"},
	{0x3DCB081CE4876324ULL, "mer120cr_hubert.utc"},
	{0x3DCB805638A61077ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_beyondpossession_sa.dds"},
	{0x3DCB8181C737EA1FULL, "gen00fl_aveline6152523.cl"},
	{0x3DCB8575C3723837ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_blda_0d.dds"},
	{0x3DCB9693F654DB1DULL, "6136472_m.fxe"},
	{0x3DCBB23E8CB350EBULL, "6085462_m.fxe"},
	{0x3DCBD61E1CAF518DULL, "vxht_0_hightown_apy_0_0.vlm"},
	{0x3DCC2886DD0FD3DBULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.dura.ani"},
	{0x3DCC61C3C82F48B4ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.dura.evt"},
	{0x3DCCB2D0BE9A2D15ULL, "art\\levels\\races\\design\\prp_elfroot.msh"},
	{0x3DCCB4D0BE9A308BULL, "art\\levels\\races\\design\\prp_elfroot.mmh"},
	{0x3DCCB795F4F9C6F6ULL, "hlac_lt_backalley_g_1_7x.dds"},
	{0x3DCCC0D0BE9A4528ULL, "art\\levels\\races\\design\\prp_elfroot.mao"},
	{0x3DCD3A7BD423926CULL, "vxca_caverns2prc_a6t_0_0.vlm"},
	{0x3DCDE6925C801BFFULL, "955220178.ogg"},
	{0x3DCE053BCDE73ED9ULL, "6135114_m.fxe"},
	{0x3DCE33D2D429CA54ULL, "vxlt_0_lowtown_a2_exl_0_0.vlm"},
	{0x3DCE9DC317E09487ULL, "art\\characters\\creatures\\mubari\\animation\\choke.evt"},
	{0x3DCEA2933F7819A6ULL, "ts_chantry_arm_table.gda"},
	{0x3DCF3437E363DC8CULL, "prp_str_slime_d.dds"},
	{0x3DCF6A3E1046CB64ULL, "6030652_m.fxe"},
	{0x3DCF9584AE0D8BC3ULL, "6083059_m.fxe"},
	{0x3DCFD9C606AE6ED2ULL, "vxw_encounter3_af_0_0.vlm"},
	{0x3DCFDB6478E9F855ULL, "gen00fl_anders6096635.cl"},
	{0x3DD0236C99A14817ULL, "vxlt_warehouse_123_0_0.vlm"},
	{0x3DD0C191A5077AB6ULL, "hlac_ca_caverns1_24423_13x.dds"},
	{0x3DD0D9004A679DA7ULL, "6085901_m.fxe"},
	{0x3DD0E26C8A86AFC1ULL, "dae000_meredith6080976.cl"},
	{0x3DD0E9C12B6C5283ULL, "245856065.ogg"},
	{0x3DD1604524897843ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_lowtownjimmy_n.dds"},
	{0x3DD17C54FD90870EULL, "tow_02_f.mao"},
	{0x3DD1C0C53B007AC9ULL, "dae000_flemeth6000720.cl"},
	{0x3DD29982F711F199ULL, "m_mcr371ar_galows_prison_nt.wbd"},
	{0x3DD2E3AE2EA1EF53ULL, "vxht_mansion2_35j_0_0.vlm"},
	{0x3DD3267D6FDB918AULL, "638555220.ogg"},
	{0x3DD39BC80DF5938EULL, "vxlt_gallowstemplar_f_38p_0_0.vlm"},
	{0x3DD3C8A84A0A0BA5ULL, "6132722_m.fxe"},
	{0x3DD3D3C5D02CFA92ULL, "hlac_lt_0_lowtown_a2_65534_3y.dds"},
	{0x3DD49DC9B23C232DULL, "vxht_mansion_n_57r_0_1.vlm"},
	{0x3DD4BB54318A6D21ULL, "vxlt_0_lowtown_n_h21_0_0.vlm"},
	{0x3DD4BFE70471583AULL, "6095130_m.fxe"},
	{0x3DD4FFFA4866CFFFULL, "hlac_lt_0_lowtown_131072_5y.dds"},
	{0x3DD548A69CDCDD03ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_dssba_0s.dds"},
	{0x3DD555093607E043ULL, "6169975_m.fxe"},
	{0x3DD5BEC317E6EBD0ULL, "art\\characters\\creatures\\mubari\\animation\\choke.gad"},
	{0x3DD6CB2E6C3AD9C1ULL, "105879359.ogg"},
	{0x3DD738FDF3B0AF37ULL, "mer221_sabin6080581.cl"},
	{0x3DD7AD24AECCBE53ULL, "6036856_f.fxe"},
	{0x3DD80CB9414DF37EULL, "art\\vfx\\combat\\vfx_r_ibl_fli_app_imp_p.impa.ani"},
	{0x3DD84E62980DC047ULL, "6054965_m.fxe"},
	{0x3DD88F1DF5EB5140ULL, "lla_ht_0_hightown_a2_2_0.dds"},
	{0x3DD8924AB4F5A365ULL, "6096855_m.fxe"},
	{0x3DD8D170F1032736ULL, "mcr310_sibling6076390.cl"},
	{0x3DD8DD1C44A2FD5FULL, "vxlt_gallowstemplar_f_4eu_0_0.vlm"},
	{0x3DD96EA01D41F7B4ULL, "rdr111dg_martin6054489.cl"},
	{0x3DD99BCE93B8A5A5ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_up_c.mmh"},
	{0x3DD9E9FF131AB4F0ULL, "6152574_m.fxe"},
	{0x3DDA3086DD1C1ED2ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.dura.evt"},
	{0x3DDA59C3C83B78EDULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.dura.ani"},
	{0x3DDA94294EB15C02ULL, "vxw_bonepitdestroyed_a3_18e_0_0.vlm"},
	{0x3DDAE76CB1E614C2ULL, "hlac_ht_0_hightown_n_-2_0y.dds"},
	{0x3DDB073587D2B1A0ULL, "213983457.ogg"},
	{0x3DDB0A02055D313BULL, "6016220_m.fxe"},
	{0x3DDBC1347755887AULL, "6120813_f.fxe"},
	{0x3DDBEDC181BF054CULL, "vxlt_warehouse_ru_0_0.vlm"},
	{0x3DDC40E97E1EEF79ULL, "gen00fl_anders6125982.cl"},
	{0x3DDC6463EFF68DD2ULL, "vxht_mansion2_35o_0_0.vlm"},
	{0x3DDCE74319382439ULL, "vxlt_0_lowtown_docks_n_b63_0_1.vlm"},
	{0x3DDCEBBEFE106421ULL, "6088777_m.fxe"},
	{0x3DDD162CC90FFDCAULL, "vxlt_0_lowtown_eo2_0_1.vlm"},
	{0x3DDD1AD9BA66A16BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_enter.evt"},
	{0x3DDD63DAA1EC105BULL, "mag311_emile_de_launcet6043604.cl"},
	{0x3DDD78531F281EB8ULL, "plt_cod_hst_seekers.plo"},
	{0x3DDE8180CB88B3F1ULL, "6016741_m.fxe"},
	{0x3DDE81E77B745632ULL, "zz_mag_debug6054609.cl"},
	{0x3DDE9ABCC6E0D972ULL, "4609375.ogg"},
	{0x3DDF9338697C3BC9ULL, "zz_ser_debug6078566.cl"},
	{0x3DE0863320E43CDDULL, "vxht_0_hightown_n_cao_0_0.vlm"},
	{0x3DE09D9BA30ABB81ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_2y.dds"},
	{0x3DE0E5618929C075ULL, "63991119.ogg"},
	{0x3DE1002F27102868ULL, "ave121_jeven6105496.cl"},
	{0x3DE14F86DD227265ULL, "art\\vfx\\combat\\vfx_m_bld_mode_c.dura.gad"},
	{0x3DE1667D6662B751ULL, "vxlt_0_lowtown_d2n_0_0.vlm"},
	{0x3DE1F54C98F4DA6BULL, "prm000im_frozen_boots.uti"},
	{0x3DE2005E17CB0D25ULL, "vxht_bartrand_n_5p1_0_0.vlm"},
	{0x3DE20D4BB29106B2ULL, "6109208_m.fxe"},
	{0x3DE24E779AF97773ULL, "hlac_ht_0_hightown_a2_-196608_1z.dds"},
	{0x3DE2D6D58F2D561BULL, "vxca_caverns2_8ct_0_0.vlm"},
	{0x3DE327C5F6948592ULL, "248318660.ogg"},
	{0x3DE333D9BA6B372CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_enter.gad"},
	{0x3DE3FC640C7DF855ULL, "w_bonepitmining_a2_10_0_lc.mmh"},
	{0x3DE40B82BDCF7BAAULL, "617642282.ogg"},
	{0x3DE4377303526DEBULL, "6165485_m.fxe"},
	{0x3DE5EE093B92F407ULL, "llac_w_bonepit_131077_0.dds"},
	{0x3DE624B9415A594FULL, "art\\vfx\\combat\\vfx_r_ibl_fli_app_imp_p.impa.evt"},
	{0x3DE6736AFC8F5FB9ULL, "6016975_f.fxe"},
	{0x3DE698CE93C32B72ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_up_c.anb"},
	{0x3DE7042B2236A24CULL, "414916865.ogg"},
	{0x3DE7047CB3779B28ULL, "hlac_lt_0_lowtown_n_-65535_7z.dds"},
	{0x3DE76AB28108FAA4ULL, "6100249_m.fxe"},
	{0x3DE7ACA825EA416DULL, "vxht_0_hightown_a3_cdj_0_0.vlm"},
	{0x3DE7D7E6E3C070F1ULL, "vxht_keep_642_0_0.vlm"},
	{0x3DE826101EE7D180ULL, "gen00fl_anders6119289.cl"},
	{0x3DE862A383F4B0E8ULL, "dae000ip_crowd_mf2_talk17.utp"},
	{0x3DE9A2E0C9D0CB62ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_mage.gad"},
	{0x3DE9DD3BD5AF9D42ULL, "6174305_m.fxe"},
	{0x3DEA32D9BA71542AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_enter.ani"},
	{0x3DEA5F9E73B51DCCULL, "vxlt_warehouse_wj_0_0.vlm"},
	{0x3DEB13DC472071DAULL, "hlac_lt_0_lowtown_a2_131072_0y.dds"},
	{0x3DEB6FC9C630AB6EULL, "6085544_m.fxe"},
	{0x3DEB755B6A7F2BD1ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_flare01.mao"},
	{0x3DEBA4DE0341D1FAULL, "435235596.ogg"},
	{0x3DEBD79E10064E81ULL, "gen_im_wep_war_shd_act2_01.uti"},
	{0x3DEBE07DA1E9CB04ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_ex.gad"},
	{0x3DEBFA6734543562ULL, "vxlt_gallowstemplar_n_5qe_0_0.vlm"},
	{0x3DEC002B6FF732FCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutd.mao"},
	{0x3DEC0C2B6FF7469FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutd.mmh"},
	{0x3DEC122B6FF75141ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutd.msh"},
	{0x3DEC26BDD0C21D07ULL, "6142715_f.fxe"},
	{0x3DECC468A6908349ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrysmelt.mmh"},
	{0x3DECD4132AB907CEULL, "vxlt_gallowscourtyard_74e_0_0.vlm"},
	{0x3DED25B9416079F8ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_app_imp_p.impa.gad"},
	{0x3DED78614E43DE54ULL, "vxht_keep_n_6lt_0_0.vlm"},
	{0x3DEE33DCFD19F17AULL, "one110_meeran6089464.cl"},
	{0x3DEE471BA89EE9B3ULL, "vxlt_shop_cf_0_0.vlm"},
	{0x3DEE7B57AED7A52BULL, "hlac_lt_hangedman_-1_12y.dds"},
	{0x3DEF014BE0341A4BULL, "gen_im_trs_bst_pelt_04.uti"},
	{0x3DEF26C0D447F557ULL, "gen00fl_aveline6115423.cl"},
	{0x3DEF464F4C88C8D7ULL, "vxht_0_hightown_a2_c9w_0_1.vlm"},
	{0x3DEF63038618B2A1ULL, "373592242.ogg"},
	{0x3DEFA2C603FE0D2CULL, "hlac_lt_0_lowtown_docks_a3_-131074_0z.dds"},
	{0x3DEFB26BB6C09CE7ULL, "plt_lgt000pt_ultra_2_02.plo"},
	{0x3DEFC5E0C9D57259ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_mage.evt"},
	{0x3DF04790613ABC51ULL, "6072348_f.fxe"},
	{0x3DF0BE66DB4BF1A3ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_dsgaa_0s.dds"},
	{0x3DF0C4E70C9FAB96ULL, "739711020.ogg"},
	{0x3DF0EA6EA470F404ULL, "762083370.ogg"},
	{0x3DF10E9E7FE109AFULL, "587381757.ogg"},
	{0x3DF187BE21403ACAULL, "dae231ip_at_to_darktown.utp"},
	{0x3DF1BB99C6A4AFE5ULL, "6088702_m.fxe"},
	{0x3DF1C985DF66F72CULL, "6126753_m.fxe"},
	{0x3DF1D78A24177CBEULL, "294881343.ogg"},
	{0x3DF29AE920CAB6F0ULL, "6083544_m.fxe"},
	{0x3DF2CA491046C880ULL, "vxw_encounter1_d5_0_0.vlm"},
	{0x3DF2EF7DA1F003E2ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_ex.ani"},
	{0x3DF2F413367699CCULL, "fen110_fenris6096419.cl"},
	{0x3DF3293DA06AC13FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowedge.phy"},
	{0x3DF3539B17127159ULL, "ht_0_hightown_a3_37_0.rml"},
	{0x3DF3B5BB06CDE721ULL, "vxht_keep_6xn_0_0.vlm"},
	{0x3DF3C8F5D4E2FDF2ULL, "vxca_cavern1_haunted_imh_0_0.vlm"},
	{0x3DF43D9895F5B3C1ULL, "6150982_m.fxe"},
	{0x3DF4DE0A1FDAD4AAULL, "hlac_lt_0_lowtown_a3_n_c_1_4x.dds"},
	{0x3DF548296A955A5FULL, "524890522.ogg"},
	{0x3DF549010E95D0ABULL, "vxw_bonepit_1nb_0_0.vlm"},
	{0x3DF5655DFAED0C24ULL, "6120603_m.fxe"},
	{0x3DF5A98AF8FBB49AULL, "lgt000cr_kirk_turn_in_2.utc"},
	{0x3DF5E2F19D9B7115ULL, "one130_thug.bnk"},
	{0x3DF649C3D5B68971ULL, "vxht_keep_611_0_0.vlm"},
	{0x3DF6992CCAD50D67ULL, "dae220ar_hightown.ncs"},
	{0x3DF7135918E11378ULL, "vxw_encounter2_fb_0_0.vlm"},
	{0x3DF81AAC7229E87EULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\mw.anb"},
	{0x3DF86ECEBF8AABA6ULL, "400602770.ogg"},
	{0x3DF8D17B65AB364CULL, "vxlt_warehouse_xk_0_0.vlm"},
	{0x3DF8F72843B02621ULL, "vxht_0_hightown_n_bwd_0_0.vlm"},
	{0x3DF98969AA0C13D4ULL, "vxlt_0_lowtown_a2_eis_0_0.vlm"},
	{0x3DFA04F19D9F946DULL, "one130_thug.cnv"},
	{0x3DFABA6A7956CB4BULL, "hlac_w_blightlands_196607_0y.dds"},
	{0x3DFAF6AE19376F42ULL, "gen00fl_anders6114451.cl"},
	{0x3DFB01E1F62E071FULL, "vxlt_undercityencounter_13g_0_0.vlm"},
	{0x3DFBEA3E4039F6D4ULL, "vxht_keep_n_6gu_0_0.vlm"},
	{0x3DFC116F5558F9B2ULL, "gen00fl_merrill6124021.cl"},
	{0x3DFC4FA2009BF3EEULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtb.msh"},
	{0x3DFC5DA2009C0B3FULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtb.mao"},
	{0x3DFC69A2009C1FDCULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtb.mmh"},
	{0x3DFCEC84DAC43FB8ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stonetree_n.dds"},
	{0x3DFD295A50E8F13EULL, "art\\vfx\\combat\\vfx_c_mer_attk1_c.impa.ani"},
	{0x3DFD792CCADAF516ULL, "dae220ar_hightown.lst"},
	{0x3DFD7DBB103BC9D7ULL, "6018693_m.fxe"},
	{0x3DFD886452B455BEULL, "214499588.ogg"},
	{0x3DFDCDE0C9E1BD48ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_mage.ani"},
	{0x3DFDFD7620108ECFULL, "6076936_m.fxe"},
	{0x3DFE0E1DFA848C1AULL, "lt_0_lowtown_a2_n_c_28_0_lc.mmh"},
	{0x3DFE7654FC1BDC56ULL, "592434319.ogg"},
	{0x3DFEDA1A5B2741FAULL, "follower_banter6141876.cl"},
	{0x3DFEDA91EA8F2DB0ULL, "vxw_encounter1_ev_0_0.vlm"},
	{0x3DFF7A36710F11C0ULL, "gen00fl_sebastian6188646.cl"},
	{0x3DFF85CFDBB17B6CULL, "lt_0_lowtown_docks_a3_26_0_ll.mmh"},
	{0x3DFFCA9A2DAD50C4ULL, "vxht_chantry_2lk_0_0.vlm"},
	{0x3E002D76B89AC865ULL, "6083602_m.fxe"},
	{0x3E004784245D0A46ULL, "vxlt_hangedman_n_27y_0_0.vlm"},
	{0x3E005E48DCF7CC5BULL, "685967471.ogg"},
	{0x3E007FB27F462F51ULL, "vxdr_primevaltomb_8py_0_0.vlm"},
	{0x3E00EB3D115CBAA8ULL, "gen00fl_aveline6120813.cl"},
	{0x3E011052F8D078F3ULL, "zz_seb_debug6215747.cl"},
	{0x3E01325F17DF7C16ULL, "gen00fl_aveline6137550.cl"},
	{0x3E013C26023CE100ULL, "vxw_encounter1_i2_0_0.vlm"},
	{0x3E018264AA38F83DULL, "family_rules_headskintone.gda"},
	{0x3E018424DF801F68ULL, "61503670.ogg"},
	{0x3E021B20B36D3235ULL, "vxdr_0_deeproads_f0q_0_0.vlm"},
	{0x3E023AD2C6D91672ULL, "vxca_cavern1_haunted_iji_0_0.vlm"},
	{0x3E02A7DD03378386ULL, "gen00fl_isabela_init6077382.cl"},
	{0x3E02AE20140FDD29ULL, "vxw_bonepitmining_a2_98_0_0.vlm"},
	{0x3E02F6F7F99DCD1BULL, "vxht_0_hightown_a2_bx8_0_0.vlm"},
	{0x3E036B8D504C1E44ULL, "vxdr_0_deeproads_cmn_0_0.vlm"},
	{0x3E038572A6A839F6ULL, "403163654.ogg"},
	{0x3E038A31D2B62573ULL, "vxlt_gallowstemplar_n_3s4_0_0.vlm"},
	{0x3E03BD2E7BE724EAULL, "213874928.ogg"},
	{0x3E03FAB4F59FD0EBULL, "llac_ca_caverns1_mines_24423_8.dds"},
	{0x3E042A3DA0793644ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowedge.mmh"},
	{0x3E042E3DA0793C97ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowedge.mao"},
	{0x3E0432B31A8546DFULL, "vxlt_gallowstemplar_f_4tv_0_0.vlm"},
	{0x3E04403DA0795BE6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowedge.msh"},
	{0x3E047D2530AECBADULL, "39300891.ogg"},
	{0x3E04E4AFB6604690ULL, "hlac_lt_undercityserialkiller_-1_19z.dds"},
	{0x3E06238FA37600DFULL, "6072774_m.fxe"},
	{0x3E06A7966152B9C9ULL, "931824925.ogg"},
	{0x3E07010C89F95D38ULL, "vxca_cavern1_haunted_jw7_0_0.vlm"},
	{0x3E072943F413D4A7ULL, "460621547.ogg"},
	{0x3E073CDE10BEAA53ULL, "vxht_0_hightown_n_bkz_0_0.vlm"},
	{0x3E07882CCAE36301ULL, "dae220ar_hightown.are"},
	{0x3E0799D7EA9B704AULL, "lt_0_lowtown_docks_a3_n_191_0.mao"},
	{0x3E079DD7EA9B7699ULL, "lt_0_lowtown_docks_a3_n_191_0.mmh"},
	{0x3E07A7D7EA9B8787ULL, "lt_0_lowtown_docks_a3_n_191_0.msh"},
	{0x3E07DE42957693E9ULL, "gen00fl_sebastian6189398.cl"},
	{0x3E07FE5BA243E9B5ULL, "139252328.ogg"},
	{0x3E087E577F33A773ULL, "gen00fl_varric6097197.cl"},
	{0x3E08BE0122C2163BULL, "30064868.ogg"},
	{0x3E09474C777CE92FULL, "vxht_mansion2_2px_0_0.vlm"},
	{0x3E098CD7B4874124ULL, "364693995.ogg"},
	{0x3E09B68A18289DB3ULL, "168276906.ogg"},
	{0x3E0A3CD6AE6C4602ULL, "330592673.ogg"},
	{0x3E0A415A50F3A40FULL, "art\\vfx\\combat\\vfx_c_mer_attk1_c.impa.evt"},
	{0x3E0AA3F38F7FF752ULL, "6195817_f.fxe"},
	{0x3E0AE35D089EBFB0ULL, "6106706_m.fxe"},
	{0x3E0AFD5921EA6BD5ULL, "6168164_f.fxe"},
	{0x3E0C1897E315644AULL, "hlac_lt_0_lowtown_n_65533_0z.dds"},
	{0x3E0C6AA9BF3EDD6AULL, "6070838_m.fxe"},
	{0x3E0CA0976E5CC774ULL, "vxht_0_hightown_a2_b7b_0_0.vlm"},
	{0x3E0CF789726A8D1AULL, "hlac_lt_darktown_-1_41y.dds"},
	{0x3E0D39F76D1995D6ULL, "120230808.ogg"},
	{0x3E0D4C6EDC854630ULL, "vxw_encounter1_jw_0_0.vlm"},
	{0x3E0D9181F85C00AFULL, "isa321_velasco6094476.cl"},
	{0x3E0DD8B5AD1B195FULL, "6164499_m.fxe"},
	{0x3E0DE1BC798A0115ULL, "363477737.ogg"},
	{0x3E0E2C6C650F9F9BULL, "vxca_caverns2_8pi_0_0.vlm"},
	{0x3E0E342F5D9B8B63ULL, "dae000_arishok6068557.cl"},
	{0x3E0E4A7021D005F0ULL, "257253596.ogg"},
	{0x3E0ED83FB35493AFULL, "ser221_gascard6148829.cl"},
	{0x3E0FBECC3C5F7D1EULL, "art\\vfx\\combat\\vfx_r_var_kback_act_c.impa.ani"},
	{0x3E10EDF9DAD75F2CULL, "zz_mcr_debug6088430.cl"},
	{0x3E10FAFDFE27FE5AULL, "vxdr_slavers_1t2_0_1.vlm"},
	{0x3E11425A50F9C4B8ULL, "art\\vfx\\combat\\vfx_c_mer_attk1_c.impa.gad"},
	{0x3E117F7913D6BA4FULL, "gen00fl_dog6167662.cl"},
	{0x3E12E3BF394E5A1BULL, "hlac_lt_gallowscourtyard_n_65539_0x.dds"},
	{0x3E1305D4E268F9FEULL, "265337909.ogg"},
	{0x3E137863A7A12538ULL, "dae200cr_viscount6087206.cl"},
	{0x3E1398ED42A32347ULL, "art\\vfx\\combat\\vfx_gib_hm_meda_lbrt.anb"},
	{0x3E143A5989153248ULL, "118935210.ogg"},
	{0x3E145C0D3E35910FULL, "265385211.ogg"},
	{0x3E1460A79F69C2D0ULL, "aip_gateguardian_s_plot.gda"},
	{0x3E14818CCF2BE05AULL, "dae000ip_crowd_ff8_talk16.utp"},
	{0x3E14BF9BA77C64B1ULL, "443603452.ogg"},
	{0x3E14F095B02C1213ULL, "vxht_mansion_n_5ak_0_0.vlm"},
	{0x3E152ACD4B5B23C6ULL, "272082548.ogg"},
	{0x3E153943F45322B8ULL, "gen00fl_anders6125628.cl"},
	{0x3E15C168A6B3837EULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrysmelt.anb"},
	{0x3E160E91FAB9EF58ULL, "vxht_chantry_2zm_0_0.vlm"},
	{0x3E162794007DE4C5ULL, "vxht_mansion2b_1av_0_0.vlm"},
	{0x3E162F597634C41DULL, "vxlt_0_lowtown_a3_n_c_glo_0_0.vlm"},
	{0x3E168586D1EB9EFAULL, "vxht_mansion2b_1rx_0_0.vlm"},
	{0x3E168C90CE39FBE4ULL, "vxlt_blackemporium_t7_0_0.vlm"},
	{0x3E169B9050AE4A2BULL, "vxca_caverns1_egf_0_0.vlm"},
	{0x3E16C2DEE783D084ULL, "vxlt_0_lowtown_a2_cmu_0_0.vlm"},
	{0x3E17BB330FD75E61ULL, "gen00fl_varric6201271.cl"},
	{0x3E184E133820B000ULL, "qcr220_orsino6041853.cl"},
	{0x3E18729CE1FE1C47ULL, "genip_store_sacks.utp"},
	{0x3E18779FBAE91AC2ULL, "aip_raiderreaver_base.gda"},
	{0x3E18B08F55E4657CULL, "6053893_m.fxe"},
	{0x3E18DA15B98DB60BULL, "6069914_m.fxe"},
	{0x3E18E4EC58F4902DULL, "865466752.ogg"},
	{0x3E1971094F389AFDULL, "vxw_bonepit_29b_0_0.vlm"},
	{0x3E1979C0C1B0B2D1ULL, "hlac_lt_0_lowtown_docks_-131076_0z.dds"},
	{0x3E198B5EB4648576ULL, "6082185_m.fxe"},
	{0x3E19A1FE3DF7B551ULL, "6135070_f.fxe"},
	{0x3E19C2642C075F2EULL, "vxca_caverns2_737_0_1.vlm"},
	{0x3E1A06F7C4C8F24BULL, "hlac_lt_0_lowtown_a3_n_c_131074_0z.dds"},
	{0x3E1A6E98F0F3FB9DULL, "hlac_ht_mansion2b_818_3z.dds"},
	{0x3E1AFFE87EE51322ULL, "llac_lt_0_lowtown_docks_a3_-65540_0.dds"},
	{0x3E1B12746052DFF4ULL, "vxht_0_hightown_a2_b8c_0_0.vlm"},
	{0x3E1B6AC03ECB0116ULL, "zz_and_debug6178615.cl"},
	{0x3E1B87F775F0D1A5ULL, "hlac_w_encounter2_camp_65541_0z.dds"},
	{0x3E1BCAFC5DB25E6AULL, "vxw_bonepitmining_a2_5s_0_0.vlm"},
	{0x3E1BEF4FFCD08327ULL, "qun111_ketojan6150159.cl"},
	{0x3E1CD467542ABDE4ULL, "w_blightlands_155_0.msh"},
	{0x3E1CE667542ADC4DULL, "w_blightlands_155_0.mao"},
	{0x3E1CEA67542AE286ULL, "w_blightlands_155_0.mmh"},
	{0x3E1CFB794E714480ULL, "6117624_f.fxe"},
	{0x3E1D0500A5A1DC86ULL, "gen00fl_aveline6002896.cl"},
	{0x3E1D2DF08745395DULL, "art\\characters\\playercharacter\\humanmale\\bld_viscount.mao"},
	{0x3E1D69BDFCAAFE5AULL, "hlac_dr_primevaltomb_657_4z.dds"},
	{0x3E1D7A845C88782DULL, "69474289.ogg"},
	{0x3E1DD6CC3C6BE3EFULL, "art\\vfx\\combat\\vfx_r_var_kback_act_c.impa.evt"},
	{0x3E1EBA20AA75ABB3ULL, "vxlt_gallowscourtyard_76b_0_0.vlm"},
	{0x3E1F64E951E86B7BULL, "lt_0_lowtown_docks_164_0.msh"},
	{0x3E1F6DE3D010E4F4ULL, "5687886.ogg"},
	{0x3E1F76E951E889C6ULL, "lt_0_lowtown_docks_164_0.mao"},
	{0x3E1F82E951E89E65ULL, "lt_0_lowtown_docks_164_0.mmh"},
	{0x3E20A9CA511458D1ULL, "character_character_p22.psh.fxo.dx11.assembly"},
	{0x3E2136FBB8572E61ULL, "ht_bartrand_n_3373.anb"},
	{0x3E213DF8C5418595ULL, "hightown_male_noble1_sta.bnk"},
	{0x3E2148A200BB03A7ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtb.phy"},
	{0x3E21C45EC4249032ULL, "6179060_f.fxe"},
	{0x3E224C6DBD699922ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45nw.msh"},
	{0x3E224E6DBD699CD8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45nw.mmh"},
	{0x3E225A6DBD69B17BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45nw.mao"},
	{0x3E2279594C270E7EULL, "6075118_m.fxe"},
	{0x3E23C5551947D7BBULL, "and100_anders6095797.cl"},
	{0x3E23D38606EEEDCDULL, "875754597.ogg"},
	{0x3E23D7CC3C705098ULL, "art\\vfx\\combat\\vfx_r_var_kback_act_c.impa.gad"},
	{0x3E249019EE34FCDBULL, "6065396_m.fxe"},
	{0x3E249C2B305A4775ULL, "mcr310_argument6136209.cl"},
	{0x3E24A136682ADC9DULL, "vxlt_0_lowtown_a3_n_c_gcn_0_0.vlm"},
	{0x3E24B740FA1BC7D7ULL, "mnd_con_dis.mao"},
	{0x3E24D9EC71CF031DULL, "ht_0_hightown_9_0.mao"},
	{0x3E24DDEC71CF0A56ULL, "ht_0_hightown_9_0.mmh"},
	{0x3E24E7EC71CF1B54ULL, "ht_0_hightown_9_0.msh"},
	{0x3E24F763C3E1B77AULL, "vxht_mansion2b_1uy_0_0.vlm"},
	{0x3E250D6D42A462ABULL, "vxca_caverns1_eby_0_0.vlm"},
	{0x3E255FF8C545A8EDULL, "hightown_male_noble1_sta.cnv"},
	{0x3E258EE22F6141F4ULL, "vxht_0_hightown_n_ap3_0_0.vlm"},
	{0x3E261B16BFFBE146ULL, "vrc221_bartrand6043390.cl"},
	{0x3E2629028F725FF3ULL, "15525500.ogg"},
	{0x3E2679026FD68FD9ULL, "396403701.ogg"},
	{0x3E26AC4153D30DAEULL, "86559256.ogg"},
	{0x3E26C7CBB32F0E5EULL, "664185693.ogg"},
	{0x3E26E16754332851ULL, "w_blightlands_155_0.phy"},
	{0x3E278F224068CE93ULL, "vxw_bonepitdestroyed_a3_250_0_0.vlm"},
	{0x3E282EE57B483975ULL, "llac_ht_chantry_n_2534_45.dds"},
	{0x3E28A0CE6DC1A3F3ULL, "llac_lt_0_lowtown_a2_n_c_-65537_2.dds"},
	{0x3E28AC6E4B60F84DULL, "821423463.ogg"},
	{0x3E28E99F9480D3B8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masc_0d.dds"},
	{0x3E293311FAD3F8B1ULL, "vxht_mansion_n_4f9_0_0.vlm"},
	{0x3E293FA94A6E05E0ULL, "hlac_lt_0_lowtown_a2_n_65538_5x.dds"},
	{0x3E29A75F0C386E34ULL, "122896635.ogg"},
	{0x3E29C3476875F3F9ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrc_0n.dds"},
	{0x3E2A1A6FAFBEE188ULL, "llac_lt_0_lowtown_a2_n_65535_6.dds"},
	{0x3E2A340EA2A765FBULL, "vxlt_backalley_g_1jl_0_0.vlm"},
	{0x3E2AA57E751BFEDDULL, "hlac_ht_0_hightown_n_-65535_0x.dds"},
	{0x3E2AD1043342821FULL, "776106785.ogg"},
	{0x3E2ADE41B98E9598ULL, "6085968_m.fxe"},
	{0x3E2BAC4C314A2F92ULL, "950631693.ogg"},
	{0x3E2C3EFA6C6F5324ULL, "art\\vfx\\zztest\\zzeventstest.impa.ani"},
	{0x3E2C4483F44F4303ULL, "follower_banter6163251.cl"},
	{0x3E2C99A73B16279BULL, "hlac_lt_0_lowtown_a2_0_0z.dds"},
	{0x3E2CE1940AB85DD6ULL, "6089096_m.fxe"},
	{0x3E2D27B89BC8420DULL, "art\\characters\\creatures\\desiredemon\\c_desiredemon_eye.mao"},
	{0x3E2D556DBD73B063ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45nw.phy"},
	{0x3E2E6E22CF68CE81ULL, "812549488.ogg"},
	{0x3E2EB2E14DA9D1A6ULL, "984640373.ogg"},
	{0x3E2EB54CA2E5F912ULL, "193222373.ogg"},
	{0x3E2F04B65F588EABULL, "hlac_lt_gallowstemplar_f_2384_2y.dds"},
	{0x3E2F158A1BCC3913ULL, "vxlt_hangedman_1rk_0_0.vlm"},
	{0x3E2F4370244151F2ULL, "hlac_ht_keep_3736_0z.dds"},
	{0x3E2F714BE829951FULL, "6082203_m.fxe"},
	{0x3E303E586E05A89CULL, "hlac_lt_gallowscourtyard_65539_0z.dds"},
	{0x3E30E7FAC09CF408ULL, "vxlt_0_lowtown_docks_a3_n_cv1_0_0.vlm"},
	{0x3E30F2D94162ED0AULL, "lgt150_nat_were_supporter6110992.cl"},
	{0x3E30FEBA5F709B2DULL, "916918454.ogg"},
	{0x3E312FF761B6D3D1ULL, "vxht_0_hightown_n_c57_0_0.vlm"},
	{0x3E3133E135BBF13BULL, "vxht_mansion2_1sf_0_0.vlm"},
	{0x3E317672DC1B5ABCULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge5.ani"},
	{0x3E31905252FA73FFULL, "200332922.ogg"},
	{0x3E31A5F5DAB63651ULL, "6116548_m.fxe"},
	{0x3E32014C9537B68BULL, "vxlt_0_lowtown_docks_n_b11_0_0.vlm"},
	{0x3E323BFA6C73B986ULL, "art\\vfx\\zztest\\zzeventstest.impa.gad"},
	{0x3E32833EB4785501ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgtp_0n.dds"},
	{0x3E32EABE437D6567ULL, "isa211_isabela6070580.cl"},
	{0x3E32FFDA743C95F7ULL, "dre000_dougal6129724.cl"},
	{0x3E33B5527C1D4B9BULL, "vxlt_gallowstemplar_5jb_0_0.vlm"},
	{0x3E3447B335B935ECULL, "vxw_woundedcoast_1ec_0_0.vlm"},
	{0x3E34B8141D5A8D5AULL, "252317130.ogg"},
	{0x3E351E06145955FEULL, "6169048_m.fxe"},
	{0x3E355511A5BB5694ULL, "6169183_m.fxe"},
	{0x3E3600FF325EE713ULL, "vxw_bonepitdestroyed_a3_203_0_0.vlm"},
	{0x3E37B7C433ACDC7DULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyo_0t.dds"},
	{0x3E383A626063E826ULL, "lt_0_lowtown_n_29_0.rml"},
	{0x3E388372DC218FFEULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge5.gad"},
	{0x3E38A17E6B3AB3C6ULL, "vxht_bartrand_n_4nu_0_0.vlm"},
	{0x3E38A48A58B19A29ULL, "6101131_m.fxe"},
	{0x3E38A5EB949D7E7BULL, "vxlt_backalley_g_1wo_0_0.vlm"},
	{0x3E38ADE29815DDEBULL, "41375456.ogg"},
	{0x3E39074E0BB671ECULL, "274436433.ogg"},
	{0x3E3936FA6C79D065ULL, "art\\vfx\\zztest\\zzeventstest.impa.evt"},
	{0x3E39C354CF490B6AULL, "6177354_f.fxe"},
	{0x3E39C9D4317C38BAULL, "hlac_dr_dungeon_3217_7x.dds"},
	{0x3E39E7F7D50A9453ULL, "vxw_encounter3_9k_0_0.vlm"},
	{0x3E3A60EDB4C67E48ULL, "hlac_lt_0_lowtown_n_-2_1z.dds"},
	{0x3E3AA39C32DF8B02ULL, "gen_im_npc_arm_hum_cht_bmga.uti"},
	{0x3E3AB7FA3FFB6C7CULL, "202334987.ogg"},
	{0x3E3B367042AEA387ULL, "dae160ar_wounded_coast.ncs"},
	{0x3E3BC1C88A143A50ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_basic_2.mao"},
	{0x3E3BF0AA733E7441ULL, "hlac_ca_cavern1_haunted_24422_3z.dds"},
	{0x3E3C8CB496E7B7CEULL, "843750597.ogg"},
	{0x3E3D070986237423ULL, "6082694_m.fxe"},
	{0x3E3DE328DA1FAD9FULL, "6085303_m.fxe"},
	{0x3E3E142F5F1F815BULL, "tau_star_g.dds"},
	{0x3E3E2043C0AE2BAEULL, "753969568.ogg"},
	{0x3E3E2C362FEBCED3ULL, "6069716_m.fxe"},
	{0x3E3E475C8DEC5EF7ULL, "gen00fl_fenris6126972.cl"},
	{0x3E3E9292E3C82FEEULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbl.evt"},
	{0x3E3F3C5611EE8F77ULL, "6183809_m.fxe"},
	{0x3E3FE4BFEF19F603ULL, "sky_sky2_p0.vsh.fxo.dx11.assembly"},
	{0x3E3FE566DF2CB047ULL, "rock_jagged_normals.dds"},
	{0x3E3FEBD302FA64F4ULL, "6110851_m.fxe"},
	{0x3E4003FF922DE378ULL, "6053314_m.fxe"},
	{0x3E403C0B1C3F119AULL, "hlac_ht_0_hightown_0_1x.dds"},
	{0x3E40E39FBEE36543ULL, "gen00fl_aveline6139995.cl"},
	{0x3E41112BE8E15BF6ULL, "6034359_m.fxe"},
	{0x3E41216D3D036EECULL, "6148726_m.fxe"},
	{0x3E412F33204E2AE7ULL, "6214839_m.fxe"},
	{0x3E4157172B738F46ULL, "57942320.ogg"},
	{0x3E416BA5EE84AD0EULL, "6049661_m.fxe"},
	{0x3E417D2AD6602E45ULL, "vxht_mansion2b_1rs_0_0.vlm"},
	{0x3E41E45206D0D78CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03b.msh"},
	{0x3E41EA5206D0E27EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03b.mmh"},
	{0x3E420E605AACB8F2ULL, "zz_testhelper.ncc"},
	{0x3E42167042B48C36ULL, "dae160ar_wounded_coast.lst"},
	{0x3E424EA47934F605ULL, "vxw_encounter2_ly_0_0.vlm"},
	{0x3E428D4E72CDDDB9ULL, "vxca_caverns1_ij7_0_0.vlm"},
	{0x3E430E880ABBEB28ULL, "vxlt_hangedman_2fo_0_0.vlm"},
	{0x3E431C9DB248C0FBULL, "llac_lt_0_lowtown_1_9.dds"},
	{0x3E4342787A504446ULL, "477306605.ogg"},
	{0x3E436196C017C08FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comc_0n.dds"},
	{0x3E4368AD965F32D2ULL, "vxw_encounter3_9n_0_0.vlm"},
	{0x3E4389835981D190ULL, "pro000_opening_pc6001201.cl"},
	{0x3E439607FDF61E4BULL, "782685753.ogg"},
	{0x3E43A907A7A038AAULL, "hlac_lt_0_lowtown_-3_2y.dds"},
	{0x3E43FF31AFDEDC6EULL, "152847249.ogg"},
	{0x3E4404B6D0F0BB82ULL, "vxlt_gallowscourtyard_7gt_0_0.vlm"},
	{0x3E444FA185601F2FULL, "and100_anders6095825.cl"},
	{0x3E448992E3CC8B99ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbl.gad"},
	{0x3E451D36E1705BD9ULL, "6135305_m.fxe"},
	{0x3E4544379E5DDBE4ULL, "251481685.ogg"},
	{0x3E454ABDEA37EC35ULL, "6104006_m.fxe"},
	{0x3E460433DE86C2B7ULL, "vxlt_0_lowtown_a3_n_c_ekj_0_0.vlm"},
	{0x3E46DE2D7093922EULL, "vxca_cavern1_haunted_hvl_0_0.vlm"},
	{0x3E47628CDB6D6EE6ULL, "6118431_m.fxe"},
	{0x3E47B147B02072B4ULL, "zz_dre_debug6069199.cl"},
	{0x3E47F2DE0A0B70C9ULL, "zz_mag_debug6053425.cl"},
	{0x3E4859D4C700ACD3ULL, "vxw_encounter3_2j_0_0.vlm"},
	{0x3E493309DC2B5915ULL, "vxht_0_hightown_9ur_0_0.vlm"},
	{0x3E493C77E29A4621ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldb_d.dds"},
	{0x3E49870D0AC61134ULL, "vxht_mansion_n_671_0_0.vlm"},
	{0x3E4A93A22694F2F1ULL, "881534659.ogg"},
	{0x3E4AE0E6B278D1BBULL, "6125470_m.fxe"},
	{0x3E4B4E4B920EA628ULL, "vxht_0_hightown_a2_bcb_0_0.vlm"},
	{0x3E4B78E678198CA3ULL, "6085794_m.fxe"},
	{0x3E4B9A92E3D2C7E7ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbl.ani"},
	{0x3E4C257042BCFA21ULL, "dae160ar_wounded_coast.are"},
	{0x3E4C3AB0D8084DDAULL, "566938740.ogg"},
	{0x3E4C841F0EA3C79FULL, "lt_undercityencounter.arl"},
	{0x3E4CB5157DCE618DULL, "llac_lt_0_lowtown_-131072_7.dds"},
	{0x3E4CC2F5F78C8994ULL, "ht_0_hightown_a3_58_0.rml"},
	{0x3E4D29B92B2D302AULL, "697198396.ogg"},
	{0x3E4D8898E3082D89ULL, "m_ave100ar_patrol.bnk"},
	{0x3E4DAC9535622DA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ent.evt"},
	{0x3E4DD8096327D9BEULL, "6140622_f.fxe"},
	{0x3E4DF101960FB685ULL, "vxlt_0_lowtown_a3_n_c_hm3_0_0.vlm"},
	{0x3E4DFBF2F709D02EULL, "6040544_f.fxe"},
	{0x3E4E17810FBEE8E7ULL, "hlac_lt_undercityserialkiller_-1_8y.dds"},
	{0x3E4E8E8CCBB74807ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_diagonal_1.mao"},
	{0x3E4ED90477F04104ULL, "358906614.ogg"},
	{0x3E4EDEE7C5D83FEAULL, "6033628_m.fxe"},
	{0x3E4EF55BECACDE22ULL, "885634697.ogg"},
	{0x3E4FC36A4C5075E1ULL, "hlac_lt_gallowstemplar_n_2394_3y.dds"},
	{0x3E4FDD82E07AC58EULL, "6044561_m.fxe"},
	{0x3E4FEF07C85646C5ULL, "vxht_mansion2b_1ur_0_0.vlm"},
	{0x3E4FF6CD3E0D261DULL, "vxlt_0_lowtown_a3_n_c_gpk_0_0.vlm"},
	{0x3E509492D4DB1877ULL, "w_encounter3_113_0.phy"},
	{0x3E50990C60097C9BULL, "vxlt_gallowstemplar_5pd_0_0.vlm"},
	{0x3E50FF2B64C3F639ULL, "vxca_caverns1_im6_0_0.vlm"},
	{0x3E510435E18088B2ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_act_c.mmh"},
	{0x3E512495A236001BULL, "1019959315.ogg"},
	{0x3E512ABF61597335ULL, "hlac_w_sundermount_196607_0y.dds"},
	{0x3E512B6D19A566ECULL, "vxw_woundedcoast_1ka_0_0.vlm"},
	{0x3E518CD0E60D97C4ULL, "415426107.ogg"},
	{0x3E51DA8A88554B52ULL, "vxw_encounter3_2o_0_0.vlm"},
	{0x3E5201BFF84586FEULL, "6163248_m.fxe"},
	{0x3E522C61293D3ABFULL, "ave331_jeven6138780.cl"},
	{0x3E527693C2E6D402ULL, "vxlt_gallowscourtyard_7hu_0_0.vlm"},
	{0x3E52DA730C30CE5BULL, "gen_im_arm_cht_nobe_t2.uti"},
	{0x3E531F517DF8F586ULL, "hlac_lt_0_lowtown_a2_n_65538_4y.dds"},
	{0x3E5337781F87841DULL, "218803232.ogg"},
	{0x3E537345B719BFF7ULL, "6142419_m.fxe"},
	{0x3E53946138F4130FULL, "hlac_lt_0_lowtown_-1_4x.dds"},
	{0x3E545C4606F9E4FBULL, "vxw_bonepitmining_a2_1j4_0_0.vlm"},
	{0x3E54859535680A24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ent.gad"},
	{0x3E54CB83A254D4B3ULL, "93870785.ogg"},
	{0x3E54D54A9A62039AULL, "mag291_vincento6056856.cl"},
	{0x3E54EB3BB364A04CULL, "vxw_encounter2_camp_du_0_1.vlm"},
	{0x3E553041D42B3BBDULL, "323573561.ogg"},
	{0x3E55CE542B6E6955ULL, "dae000_bodhan6129141.cl"},
	{0x3E55F249A2730CE3ULL, "6083327_m.fxe"},
	{0x3E561EC28F3B66EEULL, "384356757.ogg"},
	{0x3E564DFB083F0660ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallows.phy"},
	{0x3E56A8C9B310B595ULL, "6042831_f.fxe"},
	{0x3E57CCF200248DD8ULL, "llac_ca_caverns1_mines_24427_3.dds"},
	{0x3E5816C7FA948287ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\twohanded\\ledgedropland.gad"},
	{0x3E587659D435A2A9ULL, "25929136.ogg"},
	{0x3E58FB184371E659ULL, "995653567.ogg"},
	{0x3E592C3FF3CD4B41ULL, "vxlt_hangedman_n_2ml_0_0.vlm"},
	{0x3E593CAE80C2F2D0ULL, "vxdr_primevalentrance_1le_0_0.vlm"},
	{0x3E596C702D952771ULL, "vxlt_gallowscourtyard_n_7da_0_0.vlm"},
	{0x3E5989E7FB1A542DULL, "vxw_bonepitmining_a2_26i_0_0.vlm"},
	{0x3E598B6CF90C8C5CULL, "vxw_encounter1_di_0_0.vlm"},
	{0x3E5A2ECDC737C4B8ULL, "6102877_f.fxe"},
	{0x3E5AC35FDCC9F9CAULL, "724047240.ogg"},
	{0x3E5B17FBB88871F1ULL, "ht_bartrand_n_3373.rml"},
	{0x3E5B26027F7436F4ULL, "849853302.ogg"},
	{0x3E5B45B9988309B5ULL, "gen00fl_fenris6116511.cl"},
	{0x3E5B9495356E4202ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ent.ani"},
	{0x3E5C694259C7FDF9ULL, "6116203_m.fxe"},
	{0x3E5C882C18E0D41AULL, "vxht_0_hightown_a3_b9o_0_0.vlm"},
	{0x3E5C91DD771A3C7AULL, "mag131_keran6086189.cl"},
	{0x3E5C9954EB62BF90ULL, "isa311_isabela6095550.cl"},
	{0x3E5CC36B7D0467B8ULL, "6108524_m.fxe"},
	{0x3E5D15EDFE0C3C6BULL, "vxht_0_hightown_bs9_0_0.vlm"},
	{0x3E5D9EE99BB78A66ULL, "vxw_bonepitmining_a2_o1_0_0.vlm"},
	{0x3E5DE6C099F0EFBFULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargebrown.msh"},
	{0x3E5DECC099F0FA61ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargebrown.mmh"},
	{0x3E5DF8C099F10E02ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargebrown.mao"},
	{0x3E5E0428C1525948ULL, "192929051.ogg"},
	{0x3E5E29CBD66108CDULL, "gen_im_cns_pot_technique_02.uti"},
	{0x3E5E2ADA08B72BE5ULL, "vrc221_varric6043938.cl"},
	{0x3E5E3A19F8078FEEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleoff.mmh"},
	{0x3E5E4619F807A385ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleoff.mao"},
	{0x3E5E5419F807BBDCULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candleoff.msh"},
	{0x3E5E68AA30033E9DULL, "vxlt_0_lowtown_a3_n_c_gwj_0_0.vlm"},
	{0x3E5EFF573A2D583FULL, "vxw_woundedcoast_1el_0_0.vlm"},
	{0x3E5F0BC7FA9A8E94ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\twohanded\\ledgedropland.evt"},
	{0x3E5F29251961CA1AULL, "6125170_m.fxe"},
	{0x3E5F325E5D212705ULL, "vxw_encounter2_rg_0_0.vlm"},
	{0x3E5F6B673F6EF8BDULL, "554821415.ogg"},
	{0x3E5F7B348B73978AULL, "follower_banter6151275.cl"},
	{0x3E5FE611A8AEE1A5ULL, "narrative_varric_demo6170834.cl"},
	{0x3E600BA74D5AE181ULL, "llac_lt_0_lowtown_a2_n_c_-65535_6.dds"},
	{0x3E603489F15262FEULL, "one110_tree_minder6144073.cl"},
	{0x3E606AD9D5F64DF4ULL, "dae000ip_crowd_lm5_drunk1.utp"},
	{0x3E606C2DC1A20C07ULL, "vxht_keep_3xu_0_0.vlm"},
	{0x3E6087107FC76818ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg03.phy"},
	{0x3E60F1555FD43CCFULL, "dae000_bodhan6166375.cl"},
	{0x3E60F3917BD3B420ULL, "follower_barks6169912.cl"},
	{0x3E60FC1B770E6CB1ULL, "hlac_w_sundermount_0_2z.dds"},
	{0x3E611CB69201A664ULL, "vxlt_undercityserialkiller_1mr_0_0.vlm"},
	{0x3E61F26061967089ULL, "541544947.ogg"},
	{0x3E6215951024C066ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_meredith.evt"},
	{0x3E62D18188B76D5AULL, "hlac_ht_0_hightown_1_2z.dds"},
	{0x3E62E7EDC272F3B7ULL, "vxlt_0_lowtown_a3_n_c_eqd_0_0.vlm"},
	{0x3E62FC05D0B8BD62ULL, "6116406_m.fxe"},
	{0x3E639E961F267A53ULL, "vxlt_0_lowtown_a2_eqh_0_0.vlm"},
	{0x3E63EE3EECC39F51ULL, "vxw_woundedcoast_nr_0_0.vlm"},
	{0x3E64018ED7F42D12ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrb_0n.dds"},
	{0x3E64A20F82D83D8EULL, "hlac_ht_0_hightown_-131073_1x.dds"},
	{0x3E66F5681EC06F20ULL, "montage_subs_demohack6217141.cl"},
	{0x3E676FB9FAB4A90AULL, "mrl000_marethari6197796.cl"},
	{0x3E67AE8B72B90B50ULL, "vxdr_primevalentrance_1ab_0_0.vlm"},
	{0x3E67C10D491F1098ULL, "gencr_mercenary_warrior_2wf.utc"},
	{0x3E67FD49EB02A4DCULL, "vxw_encounter1_in_0_0.vlm"},
	{0x3E680578ECD091C0ULL, "6085520_m.fxe"},
	{0x3E688164279F4A84ULL, "886145947.ogg"},
	{0x3E68B5DFE56C8203ULL, "dae000_hate_follower_react6167476.cl"},
	{0x3E68C5A8303D3939ULL, "vxlt_warehouse_qe_0_0.vlm"},
	{0x3E68D7E89E4694ABULL, "vxw_bonepit_16z_0_0.vlm"},
	{0x3E68EB7A4C9A9236ULL, "6051397_f.fxe"},
	{0x3E692C95102B05B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_meredith.gad"},
	{0x3E69627C02296071ULL, "vxca_caverns1_mines_j4w_0_0.vlm"},
	{0x3E698F21A6F4581FULL, "vxw_bonepitdestroyed_a3_cv_0_0.vlm"},
	{0x3E69A81AB4A78861ULL, "zz_followers_banter6157713.cl"},
	{0x3E69BC1FEBCEDE39ULL, "vxlt_gallowscourtyard_5xp_0_0.vlm"},
	{0x3E69C3BD07254099ULL, "w_sundermount_192_0.msh"},
	{0x3E69CDBD07255187ULL, "w_sundermount_192_0.mmh"},
	{0x3E69D1BD072558D4ULL, "w_sundermount_192_0.mao"},
	{0x3E69F3787B7FDD49ULL, "gen_im_trs_krk_drk_junk1_01.uti"},
	{0x3E6A8154C7DB304EULL, "vxw_bonepitdestroyed_a3_19f_0_0.vlm"},
	{0x3E6A888DB2DA3274ULL, "mag131_taraneh6088691.cl"},
	{0x3E6B4245A6C6C1A3ULL, "790994180.ogg"},
	{0x3E6BD6C6567382ACULL, "gen00fl_anders6107691.cl"},
	{0x3E6BF0011044542EULL, "lt_0_lowtown_docks_n_4_0.rml"},
	{0x3E6BFEB0A4877B3DULL, "dae000_paivel6111590.cl"},
	{0x3E6C43CEEA298B5EULL, "475414595.ogg"},
	{0x3E6C78E66B524A38ULL, "482955995.ogg"},
	{0x3E6CD0B68E0CD844ULL, "hlac_w_encounter1_-131069_0x.dds"},
	{0x3E6CD51FACEE8B30ULL, "vxca_caverns1_mines_foc_0_0.vlm"},
	{0x3E6CE4ED61F19EB4ULL, "vxlt_hangedman_n_1yb_0_0.vlm"},
	{0x3E6D03C7FAA6BECDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\twohanded\\ledgedropland.ani"},
	{0x3E6D04DFE8FFD523ULL, "vxw_woundedcoast_13b_0_0.vlm"},
	{0x3E6D1F4BE07B18D2ULL, "vxlt_0_lowtown_a2_eqm_0_0.vlm"},
	{0x3E6D5995A595D4C3ULL, "1029824032.ogg"},
	{0x3E6D8CE69FF1FB40ULL, "vxht_chantry_n_25w_0_0.vlm"},
	{0x3E6DA109A25C74EEULL, "6054241_m.fxe"},
	{0x3E6DCD4DDEE05D97ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_0x.dds"},
	{0x3E6DF49C42141C59ULL, "vxlt_0_lowtown_docks_a3_n_d52_0_0.vlm"},
	{0x3E6E00ECDB484973ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_de_transition.evt"},
	{0x3E6E0F26FD9197ECULL, "vxw_woundedcoast_1qg_0_0.vlm"},
	{0x3E6E514983BE5517ULL, "lt_gallowscourtyard_n_52_0.rml"},
	{0x3E6F444C5FA0F693ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90runf.evt"},
	{0x3E6FC07DDF0421D7ULL, "6084866_f.fxe"},
	{0x3E702D951031265FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncfail_meredith.ani"},
	{0x3E705C8CD28E07EFULL, "hlac_w_bonepit_196608_0y.dds"},
	{0x3E706D56B387C0A4ULL, "718882424.ogg"},
	{0x3E70E2AC3D37A81CULL, "vxca_caverns1_mines_jg3_0_0.vlm"},
	{0x3E714D19479245EEULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.impa.evt"},
	{0x3E7171B182DE6E07ULL, "gen00fl_merrill6129919.cl"},
	{0x3E71BE9BE52B65CFULL, "194823231.ogg"},
	{0x3E72902DCACEE6E4ULL, "6126681_m.fxe"},
	{0x3E72F72A53BA417CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.open2.gad"},
	{0x3E7448069BED88D8ULL, "vxlt_0_lowtown_a2_go3_0_0.vlm"},
	{0x3E74532C0D033E78ULL, "dae000_merchant_generic_hm6172649.cl"},
	{0x3E74F9ECDB4E5BB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_de_transition.gad"},
	{0x3E7526AA64C21BA8ULL, "143761820.ogg"},
	{0x3E753D4C5FA556D4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90runf.gad"},
	{0x3E75D4EBBECA47A6ULL, "vxlt_0_lowtown_docks_a3_n_brn_0_0.vlm"},
	{0x3E75EE6D0DDCE3D8ULL, "zz_mrl_debug6184207.cl"},
	{0x3E764F11C5CF87F8ULL, "vxht_brothel_n_6cn_0_0.vlm"},
	{0x3E766DA1DF06852DULL, "vxw_bonepitmining_a2_28g_0_0.vlm"},
	{0x3E76A9019A708ACAULL, "634005451.ogg"},
	{0x3E76D08D37A596E2ULL, "art\\vfx\\textures\\vfx_gradient.dds"},
	{0x3E770B8A864CA6E8ULL, "vxht_keep_n_640_0_0.vlm"},
	{0x3E773785223351B9ULL, "vxlt_warehouse_zd_0_0.vlm"},
	{0x3E7744194796A199ULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.impa.gad"},
	{0x3E7788E693EC4AFFULL, "hlac_lt_backalley_g_65536_0z.dds"},
	{0x3E77C9BBCBE5DD2AULL, "6052908_m.fxe"},
	{0x3E77EFC1D2EA8ED6ULL, "gen00fl_fenris6107228.cl"},
	{0x3E7800FE98EA709FULL, "vxw_bonepitdestroyed_a3_nq_0_0.vlm"},
	{0x3E78AD5719CC9C5DULL, "sg5_cl.mao"},
	{0x3E7978A7B4446982ULL, "vxw_bonepitdestroyed_a3_1sh_0_0.vlm"},
	{0x3E79A72560F098B8ULL, "6106724_m.fxe"},
	{0x3E79F62A53C05E7AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.open2.ani"},
	{0x3E7A0335E1A38C59ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_act_c.anb"},
	{0x3E7A20E6D568ECFCULL, "zz_vrc_debug6200196.cl"},
	{0x3E7A251894526823ULL, "429788719.ogg"},
	{0x3E7B08ECDB52E092ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_de_transition.ani"},
	{0x3E7B352CA26781EAULL, "dae000_flemeth6000686.cl"},
	{0x3E7B498ABFEC1631ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrpglw.msh"},
	{0x3E7B56CA53E7B734ULL, "vxlt_hangedman_n_1vc_0_0.vlm"},
	{0x3E7B578ABFEC2E4CULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrpglw.mao"},
	{0x3E7B638ABFEC42EFULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrpglw.mmh"},
	{0x3E7B92295452D2A5ULL, "gen00fl_varric6081255.cl"},
	{0x3E7BE3111E19893FULL, "vxw_woundedcoast_1kn_0_0.vlm"},
	{0x3E7BF57CA0868D81ULL, "vxw_encounter1_d8_0_0.vlm"},
	{0x3E7C4C4C5FAB8EB2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90runf.ani"},
	{0x3E7C6679340A34D9ULL, "vxlt_0_lowtown_docks_a3_n_d01_0_0.vlm"},
	{0x3E7C6AFB086006F5ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallows.msh"},
	{0x3E7C6CFB08600A6BULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallows.mmh"},
	{0x3E7C78FB08601E08ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallows.mao"},
	{0x3E7D6A0C1CD80200ULL, "366593365.ogg"},
	{0x3E7D7ACBA4B972F6ULL, "character_character_p278.psh.fxo.dx11"},
	{0x3E7DCEA12E6D0F5FULL, "w_bonepitdestroyed_a3_109_0.phy"},
	{0x3E7DEB682AABA5AAULL, "vxlt_hangedman_n_27u_0_0.vlm"},
	{0x3E7E5519479CDDE7ULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.impa.ani"},
	{0x3E7E78AE737CFE68ULL, "hlac_ca_caverns1_mines_24423_9x.dds"},
	{0x3E7ED83CAB484834ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\qm_injured_movement.anb"},
	{0x3E7EE65A10B1DFA8ULL, "wm_ico_mansion.dds"},
	{0x3E7F05CA012FBE4EULL, "art\\vfx\\combat\\vfx_m_blood_change_c.impa.evt"},
	{0x3E7F70E90E539BF4ULL, "28191870.ogg"},
	{0x3E7F710E07A1D616ULL, "vxlt_darktown_2xn_0_0.vlm"},
	{0x3E7FB8BFF4E8C72DULL, "407836560.ogg"},
	{0x3E80EFA8A79F2E27ULL, "lgt120_orz_bln_official6140212.cl"},
	{0x3E810E3D0AE43540ULL, "zz_vault_debug6203398.cl"},
	{0x3E8171430FD7B8B8ULL, "mag211_marethari6025354.cl"},
	{0x3E817349C75D3474ULL, "hlac_lt_0_lowtown_a2_n_-3_4x.dds"},
	{0x3E81C8A7EA540479ULL, "vxlt_0_lowtown_a3_n_c_ep4_0_1.vlm"},
	{0x3E81D823D492E60BULL, "art\\vfx\\combat\\vfx_c_pride_cast_s_c.impa.evt"},
	{0x3E82A4ED4BBD297DULL, "hlac_w_sundermount_-393215_0x.dds"},
	{0x3E82C15207083319ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03b.phy"},
	{0x3E844BB206D60821ULL, "17508646.ogg"},
	{0x3E846CE610E1916CULL, "168534094.ogg"},
	{0x3E8541CB7D602682ULL, "6128057_f.fxe"},
	{0x3E85FCCA0135CDF9ULL, "art\\vfx\\combat\\vfx_m_blood_change_c.impa.gad"},
	{0x3E862D53491B51A2ULL, "137139006.ogg"},
	{0x3E865EBF5A9EA817ULL, "dae000_bodhan_letter6144266.cl"},
	{0x3E8672DB8AE0891FULL, "vxw_bonepitdestroyed_a3_ip_0_0.vlm"},
	{0x3E86C6001F0EE52AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_noba_0t.dds"},
	{0x3E8839A30AC0790FULL, "fen110_fenris6100317.cl"},
	{0x3E88BAF5A300C421ULL, "hlac_w_bonepit_0_7x.dds"},
	{0x3E88BBEFF6FD3DABULL, "dae011_order_drink6205118.cl"},
	{0x3E88F123D4992FCCULL, "art\\vfx\\combat\\vfx_c_pride_cast_s_c.impa.gad"},
	{0x3E891A78D9F511C7ULL, "6084189_f.fxe"},
	{0x3E892BE338B50514ULL, "vxlt_hangedman_n_31t_0_0.vlm"},
	{0x3E89A57940581FF6ULL, "vxca_caverns1_mines_e58_0_0.vlm"},
	{0x3E89E899CCEC0623ULL, "vxw_woundedcoast_19d_0_0.vlm"},
	{0x3E89EDC09A1636DEULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargebrown.phy"},
	{0x3E89EE47E7995FBBULL, "6198669_m.fxe"},
	{0x3E8AB544D9DDE3E5ULL, "vxht_mansion_n_33e_0_0.vlm"},
	{0x3E8B6EDAAD1B15EFULL, "vxlt_0_lowtown_a3_n_c_htz_0_0.vlm"},
	{0x3E8B868C4FDF196EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_basic.msh"},
	{0x3E8B948C4FDF30BFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_basic.mao"},
	{0x3E8BA08C4FDF455CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_basic.mmh"},
	{0x3E8BFBE73C58C079ULL, "lt_0_lowtown_docks_a3_n_188_0.msh"},
	{0x3E8C05E73C58D167ULL, "lt_0_lowtown_docks_a3_n_188_0.mmh"},
	{0x3E8C09E73C58D7B4ULL, "lt_0_lowtown_docks_a3_n_188_0.mao"},
	{0x3E8C0DCA013A5647ULL, "art\\vfx\\combat\\vfx_m_blood_change_c.impa.ani"},
	{0x3E8C3B254E9D1985ULL, "137540607.ogg"},
	{0x3E8C5C5ED2738E13ULL, "vxw_encounter2_camp_6e_0_0.vlm"},
	{0x3E8CF1C61460593BULL, "6067743_m.fxe"},
	{0x3E8D6E633529E775ULL, "860460041.ogg"},
	{0x3E8DF58555A1BBB3ULL, "vxw_bonepitdestroyed_a3_16b_0_0.vlm"},
	{0x3E8E3E91E49ADCAFULL, "100838935.ogg"},
	{0x3E8E62C7E1A21362ULL, "mag371_elsa6041241.cl"},
	{0x3E8E821A9EA64EA5ULL, "980693572.ogg"},
	{0x3E8E8AA23E081059ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd4a.msh"},
	{0x3E8E94A23E082147ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd4a.mmh"},
	{0x3E8ECFA12E7B8464ULL, "w_bonepitdestroyed_a3_109_0.mmh"},
	{0x3E8ED37FA69D2334ULL, "vxht_mansion_n_3yr_0_0.vlm"},
	{0x3E8ED3A12E7B8AB7ULL, "w_bonepitdestroyed_a3_109_0.mao"},
	{0x3E8EE5A12E7BA906ULL, "w_bonepitdestroyed_a3_109_0.msh"},
	{0x3E8F06F30EDFBA22ULL, "hlac_lt_0_lowtown_docks_a3_-4_5y.dds"},
	{0x3E8F4F8F7517E04DULL, "113795810.ogg"},
	{0x3E8F571D8A691E8EULL, "119813265.ogg"},
	{0x3E8FF023D49F4CCAULL, "art\\vfx\\combat\\vfx_c_pride_cast_s_c.impa.ani"},
	{0x3E9000A01977750EULL, "mag101_alain6032143.cl"},
	{0x3E904370EECB48D2ULL, "604583662.ogg"},
	{0x3E9089DFC9FC8928ULL, "vxlt_0_lowtown_a2_hda_0_0.vlm"},
	{0x3E90C085AE135D70ULL, "hlac_lt_warehouse_n_0_2z.dds"},
	{0x3E90EBD997ED86D6ULL, "vxca_caverns1_mines_jld_0_0.vlm"},
	{0x3E915B4021881BE4ULL, "6135788_m.fxe"},
	{0x3E91A84CB63099BBULL, "dae000pt_generic.ncs"},
	{0x3E91F223117480ACULL, "zz_followers_banter6157831.cl"},
	{0x3E9254DE836F62BBULL, "247613398.ogg"},
	{0x3E926CEF4FC4A7E8ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrowgl.mmh"},
	{0x3E927079EF6A5373ULL, "hlac_lt_0_lowtown_-65535_2y.dds"},
	{0x3E9278EF4FC4BB8BULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrowgl.mao"},
	{0x3E928AEF4FC4DA12ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrowgl.msh"},
	{0x3E92DF9CFB7FB3EDULL, "vxht_mansion2_1a6_0_0.vlm"},
	{0x3E9305E727EED7A1ULL, "mag100st_cullen_wilmod.stg"},
	{0x3E930BAADD0513DFULL, "hlac_ht_bartrand_n_3373_2z.dds"},
	{0x3E9324C9DA0A2250ULL, "331893847.ogg"},
	{0x3E9332471B895DB8ULL, "6170116_m.fxe"},
	{0x3E9332CBA9BBB8F8ULL, "vxht_brothel_n_6il_0_0.vlm"},
	{0x3E941B677F52A05FULL, "392873551.ogg"},
	{0x3E94883B4EDB8D73ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_basic.phy"},
	{0x3E94D4550D5B40F0ULL, "hlac_ca_caverns2prc_5802_40z.dds"},
	{0x3E951FB2D7E25BE8ULL, "plt_seb321pt_faith.plo"},
	{0x3E956698E345A9DBULL, "m_ave100ar_patrol.wbd"},
	{0x3E958E6CEC0CFFBCULL, "356812113.ogg"},
	{0x3E9600D6D9704BE0ULL, "vxw_encounter2_35_0_0.vlm"},
	{0x3E9625F316F45E7AULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.cess.gad"},
	{0x3E9687C4F482687DULL, "qun211_party_comment6072249.cl"},
	{0x3E96C99782FFE57DULL, "hla_ht_0_hightown_n_-196608_0y.dds"},
	{0x3E97117C22BE4DC8ULL, "hlac_lt_undercityserialkiller_-1_44x.dds"},
	{0x3E97605B10EA117BULL, "fxa_pois_dirtys.mao"},
	{0x3E9764058AE6394CULL, "2513351.ogg"},
	{0x3E9774338FCB4529ULL, "abi_r_rainofarrows.wbd"},
	{0x3E97AF7F845B9F53ULL, "vxht_0_hightown_n_b9d_0_0.vlm"},
	{0x3E97FBB998EC948DULL, "gen00fl_anders6110054.cl"},
	{0x3E9809A0FE69B548ULL, "6052090_m.fxe"},
	{0x3E98255D4359220FULL, "t1_skn_d03.tnt"},
	{0x3E986992D51884CFULL, "w_encounter3_113_0.mao"},
	{0x3E987592D518996CULL, "w_encounter3_113_0.mmh"},
	{0x3E987B92D518A35EULL, "w_encounter3_113_0.msh"},
	{0x3E98818726E0668EULL, "vxlt_0_lowtown_docks_n_b76_0_0.vlm"},
	{0x3E98B26B61D4DEEDULL, "hlac_lt_0_lowtown_n_131072_3z.dds"},
	{0x3E98C6CB0205BA3FULL, "vxw_woundedcoast_1qh_0_0.vlm"},
	{0x3E9A2B506E610AF1ULL, "hlac_w_encounter2_camp_-65537_0z.dds"},
	{0x3E9ACE3BC469A693ULL, "vxw_encounter2_camp_1d_0_0.vlm"},
	{0x3E9B7BC22A3B2096ULL, "dae300_alistair_king6133547.cl"},
	{0x3E9BF85DF0927C20ULL, "vxht_mansion2_3cg_0_0.vlm"},
	{0x3E9BFC7CBA03AAADULL, "188108825.ogg"},
	{0x3E9C28F316F8CE91ULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.cess.evt"},
	{0x3E9D268D111577BDULL, "hlac_lt_0_lowtown_a2_n_131071_0x.dds"},
	{0x3E9D39EA454E1E07ULL, "w_bonepitmining_a2_17_0.rml"},
	{0x3E9DB5B2B49C0151ULL, "vxw_woundedcoast_zv_0_0.vlm"},
	{0x3E9DC02B7A0EA120ULL, "vxlt_gallowscourtyard_n_7ek_0_0.vlm"},
	{0x3E9DC082C4935D67ULL, "vxht_0_hightown_akj_0_0.vlm"},
	{0x3E9E2EE0053AEACDULL, "lgt120_orz_bln_official6170514.cl"},
	{0x3E9E84388FC6D9FEULL, "gen00fl_merrill6106814.cl"},
	{0x3E9FCF5ADB77D72EULL, "mag170_thrask6054035.cl"},
	{0x3E9FF65ECB4F74B6ULL, "vxlt_0_lowtown_e8j_0_0.vlm"},
	{0x3EA03ED7DB472026ULL, "mcr370_cut_marriage6098261.cl"},
	{0x3EA04C93066DFA00ULL, "atl_chargen_st_dxt5_01.dds"},
	{0x3EA061A8D14CE8AEULL, "darktown_fem_elf_sta6174536.cl"},
	{0x3EA0B2980931B74BULL, "435357091.ogg"},
	{0x3EA0DF153434A9C0ULL, "dae000_meredith6085280.cl"},
	{0x3EA1303545B03DD2ULL, "vxht_0_hightown_n_b9a_0_0.vlm"},
	{0x3EA1374EF9867B94ULL, "follower_banter6170733.cl"},
	{0x3EA1CB37C869A0AFULL, "hlac_ca_caverns2_5802_62x.dds"},
	{0x3EA1DEF85CC2F244ULL, "92807763.ogg"},
	{0x3EA2770E33F9FC93ULL, "943477241.ogg"},
	{0x3EA27DE3FD44F5F8ULL, "llac_lt_0_lowtown_131073_0.dds"},
	{0x3EA2965A600E6EE0ULL, "368851821.ogg"},
	{0x3EA2A764AC6EFB5FULL, "lgt260cr_eau_tmp_dockworker.utc"},
	{0x3EA314DE5473F8B7ULL, "vxlt_0_lowtown_docks_b09_0_0.vlm"},
	{0x3EA3A896192AF0A1ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar2_.gad"},
	{0x3EA3E8FC0EA28C6FULL, "590294802.ogg"},
	{0x3EA414C49680CAF2ULL, "6154192_m.fxe"},
	{0x3EA430034724E5F8ULL, "834192081.ogg"},
	{0x3EA4A6A342373985ULL, "vxht_brothel_n_6cw_0_0.vlm"},
	{0x3EA4CE3E8A26B302ULL, "vxw_bonepitdestroyed_a3_1bk_0_0.vlm"},
	{0x3EA5350B72CBCE34ULL, "llac_lt_0_lowtown_docks_a3_n_-2_5.dds"},
	{0x3EA54AF843B7E10FULL, "vxlt_0_lowtown_a2_byk_0_0.vlm"},
	{0x3EA5E206EF231EFCULL, "6128342_m.fxe"},
	{0x3EA6891177C1018FULL, "vxlt_gallowscourtyard_7c9_0_0.vlm"},
	{0x3EA7258BE3CC8DDBULL, "dae000_madam_lusine6152160.cl"},
	{0x3EA7D0B989850476ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medb.msh"},
	{0x3EA7DAB989851574ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medb.mmh"},
	{0x3EA7DEB989851B87ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medb.mao"},
	{0x3EA81A9594C70F7BULL, "vxw_bonepitmining_a2_29r_0_0.vlm"},
	{0x3EA84A89F3CA64F6ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_humanmale.msh"},
	{0x3EA85489F3CA75F4ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_humanmale.mmh"},
	{0x3EA85889F3CA7C07ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_humanmale.mao"},
	{0x3EA88334988B653DULL, "vxlt_gallowscourtyard_72p_0_0.vlm"},
	{0x3EA8BA2B3E7FFF25ULL, "6162288_m.fxe"},
	{0x3EA8F110C1D53B42ULL, "follower_banter6143319.cl"},
	{0x3EA957D88943923FULL, "6103034_m.fxe"},
	{0x3EA97575C060E2BEULL, "387525394.ogg"},
	{0x3EA9A36BE7B198B7ULL, "follower_banter6155922.cl"},
	{0x3EA9AC042623C376ULL, "47662364.ogg"},
	{0x3EAA2C5EDD21B2E5ULL, "365774331.ogg"},
	{0x3EAA40F317053560ULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.cess.ani"},
	{0x3EAA4DA1E2C04431ULL, "6043325_m.fxe"},
	{0x3EAA5B36D650FECFULL, "vxdr_primevalentrance_21z_0_0.vlm"},
	{0x3EAA65AF4A65CE5BULL, "6033815_f.fxe"},
	{0x3EAA8F721651093AULL, "var_trigger_talk.gda"},
	{0x3EAA9C990F5A9E1CULL, "6151108_m.fxe"},
	{0x3EAAA9961931114FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar2_.ani"},
	{0x3EAB629F2E0173A6ULL, "6115616_m.fxe"},
	{0x3EAB791BC9BBD307ULL, "vxdr_primevaltomb_55e_0_0.vlm"},
	{0x3EABB7398A895434ULL, "vxht_mansion_n_3sp_0_0.vlm"},
	{0x3EABD7E6D8F4F4D3ULL, "vxlt_0_lowtown_a2_eho_0_0.vlm"},
	{0x3EAC278FA69219D1ULL, "vxw_woundedcoast_wu_0_0.vlm"},
	{0x3EAC32086C04B9A0ULL, "vxlt_gallowscourtyard_n_7hh_0_0.vlm"},
	{0x3EAC3ED4109557FCULL, "vxht_mansion2_49_0_0.vlm"},
	{0x3EAC9E9E0C7A6362ULL, "6187949_m.fxe"},
	{0x3EAD7AC9E8EB3804ULL, "look_at.gda"},
	{0x3EAD8302FBDCAE84ULL, "vxdr_primevaltomb_4gy_0_0.vlm"},
	{0x3EADB4D4F2279F01ULL, "6143119_m.fxe"},
	{0x3EADF6C85F8340AFULL, "vxw_bonepitdestroyed_a3_1v6_0_0.vlm"},
	{0x3EAE567288911326ULL, "lowtown_fem_prostitute_sta6152327.cl"},
	{0x3EAE9319D6BD63B6ULL, "vxht_mansion_n_32u_0_0.vlm"},
	{0x3EAEB360FDCFED87ULL, "6119424_m.fxe"},
	{0x3EAF548B09032E3AULL, "vxca_caverns1_mines_j6j_0_0.vlm"},
	{0x3EB046F3831901A1ULL, "vxht_mansion_n_62w_0_0.vlm"},
	{0x3EB04D535126F31BULL, "vxht_0_hightown_a2_b44_0_0.vlm"},
	{0x3EB079BD415D3CCEULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_chanterboard.mmh"},
	{0x3EB07F8C4FFE2927ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_basic.phy"},
	{0x3EB08ED0B714D586ULL, "hlac_ht_mansion_n_3364_0x.dds"},
	{0x3EB0BC75958EB236ULL, "6169589_m.fxe"},
	{0x3EB143F8EEB601ABULL, "374714830.ogg"},
	{0x3EB186BB466A1137ULL, "vxlt_0_lowtown_docks_b78_0_0.vlm"},
	{0x3EB20A2F0375A813ULL, "vxca_caverns1_j5g_0_0.vlm"},
	{0x3EB2AE3FACE30602ULL, "qun211cr_crazy_woman.cnv"},
	{0x3EB2E490BD5C7CE0ULL, "vxw_encounter2_93_0_0.vlm"},
	{0x3EB31C7652450B3FULL, "573618296.ogg"},
	{0x3EB31E5280B97BEAULL, "plt_genpt_party_triggers.plo"},
	{0x3EB32123A4F06ECBULL, "gen_im_trs_gen_one_lsw_r1.uti"},
	{0x3EB3261DCEABC7BFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_basic.phy"},
	{0x3EB3400ECA341F32ULL, "808114372.ogg"},
	{0x3EB39E4706EF0E4EULL, "w_bonepitdestroyed_a3_140_0.phy"},
	{0x3EB3F4EC5CE51CC0ULL, "vxht_mansion_n_3i6_0_0.vlm"},
	{0x3EB43725585D5C50ULL, "6180014_m.fxe"},
	{0x3EB439EA4560FD24ULL, "w_bonepitmining_a2_17_0.mmh"},
	{0x3EB43DB896186322ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_carcass_d.dds"},
	{0x3EB43DEA45610477ULL, "w_bonepitmining_a2_17_0.mao"},
	{0x3EB44FEA456122C6ULL, "w_bonepitmining_a2_17_0.msh"},
	{0x3EB451767CE05BB2ULL, "dae000_pc_home_ambients6208666.cl"},
	{0x3EB4595AA84C27E3ULL, "6152164_m.fxe"},
	{0x3EB4732F55D8C68BULL, "narrative_flemeth6002793.cl"},
	{0x3EB4A78A6C824180ULL, "fen310_varania6118600.cl"},
	{0x3EB4AEBDA0DBD7BAULL, "126735239.ogg"},
	{0x3EB5D24BAD3D97CCULL, "6033378_m.fxe"},
	{0x3EB6456271894F2AULL, "vxht_bartrand_n_4ny_0_0.vlm"},
	{0x3EB6685851EF90F6ULL, "vxlt_hangedman_261_0_0.vlm"},
	{0x3EB6721EFD3DD65CULL, "vxca_cavern1_haunted_j22_0_2.vlm"},
	{0x3EB6903FACE6BC3AULL, "qun211cr_crazy_woman.bnk"},
	{0x3EB69BB81D674B67ULL, "152762723.ogg"},
	{0x3EB6A3A8443BC64CULL, "hlac_lt_foundry_n_320_12z.dds"},
	{0x3EB6AC00144FC327ULL, "vxlt_backalley_n_1vc_0_2.vlm"},
	{0x3EB6D6ACE4006E22ULL, "vxht_mansion2b_2uj_0_0.vlm"},
	{0x3EB71CB2D7FF3CA7ULL, "plt_seb321pt_faith.nss"},
	{0x3EB734F933F64AF7ULL, "163929265.ogg"},
	{0x3EB7AE3625A48C24ULL, "fen210_tevinter_ambusher6108308.cl"},
	{0x3EB81362486F945AULL, "dre000pt_extra_funds2.ncs"},
	{0x3EB8F65A809CA00EULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewedge.msh"},
	{0x3EB9045A809CB85FULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewedge.mao"},
	{0x3EB9105A809CCCFCULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewedge.mmh"},
	{0x3EB928A710ABEC8BULL, "vxlt_hangedman_n_st_0_0.vlm"},
	{0x3EB9745DFCA86DB3ULL, "vxw_bonepit_n5_0_0.vlm"},
	{0x3EB9D34E2ADAD910ULL, "219998608.ogg"},
	{0x3EB9D47C1FF78C26ULL, "6110916_m.fxe"},
	{0x3EBA80A20D88DD96ULL, "893130406.ogg"},
	{0x3EBAB4659661E633ULL, "vxca_cavern1_haunted_iia_0_0.vlm"},
	{0x3EBAC84E63D63E16ULL, "vxlt_undercityencounter_1fq_0_0.vlm"},
	{0x3EBB351564C3499AULL, "142914876.ogg"},
	{0x3EBB7DA23E2EE87CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd4a.phy"},
	{0x3EBBE4F726EC02F6ULL, "1011573505.ogg"},
	{0x3EBC46338FEA130BULL, "abi_r_rainofarrows.bnk"},
	{0x3EBC68640B8F4F27ULL, "hlac_dr_slavers_3217_5y.dds"},
	{0x3EBC74108015A6ADULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg03.msh"},
	{0x3EBC86108015C533ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg03.mmh"},
	{0x3EBCAD525B5F9898ULL, "vxlt_warehouse_n_um_0_0.vlm"},
	{0x3EBD6CF705A8258BULL, "rdr111dg_martin6053566.cl"},
	{0x3EBD7435E15FE9FAULL, "159542750.ogg"},
	{0x3EBD871CE53A00CDULL, "870279684.ogg"},
	{0x3EBD8B4E883734F4ULL, "vxlt_0_lowtown_a3_n_c_cmu_0_0.vlm"},
	{0x3EBD8CD53A5F9929ULL, "798094611.ogg"},
	{0x3EBD98BD4167FBCDULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_chanterboard.anb"},
	{0x3EBDD890702D5B5FULL, "vxlt_gallowstemplar_n_4qq_0_0.vlm"},
	{0x3EBDDBDA19AC5E64ULL, "plt_cod_bks_lite_foreshadow.plo"},
	{0x3EBE0FCCF8835A10ULL, "107169682.ogg"},
	{0x3EBE1627AA1508E2ULL, "art\\characters\\weapons\\shortbows\\w_sbw_dssbb.mao"},
	{0x3EBE1EBD40D8C722ULL, "vxht_bartrand_n_652_0_0.vlm"},
	{0x3EBE71EA934BFAC9ULL, "vxht_mansion_n_6f3_0_0.vlm"},
	{0x3EBE7A87EEF68BE6ULL, "6118660_m.fxe"},
	{0x3EBEBF30431D0B9BULL, "vxht_0_hightown_a2_b97_0_0.vlm"},
	{0x3EBF8FE5EB4BB082ULL, "vxdr_primevaltomb_8nh_0_0.vlm"},
	{0x3EBFA8D49F48AC01ULL, "6094450_f.fxe"},
	{0x3EC03E6B191873B4ULL, "mrl150_merrill6202029.cl"},
	{0x3EC0408DCCE72B25ULL, "6075067_m.fxe"},
	{0x3EC0420A430F4178ULL, "6138808_m.fxe"},
	{0x3EC053847759BEDFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basicwood.phy"},
	{0x3EC05951DDB36A3AULL, "391831973.ogg"},
	{0x3EC0C010C0AFA2A4ULL, "gen00fl_aveline6114865.cl"},
	{0x3EC0F5CEFACEAEB8ULL, "hlac_lt_0_lowtown_a2_1_5y.dds"},
	{0x3EC11B60F01A5DABULL, "vxw_bonepitmining_a2_1me_0_0.vlm"},
	{0x3EC11E31E8094DA3ULL, "1014866654.ogg"},
	{0x3EC15DF18B03AFE3ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.impa.evt"},
	{0x3EC18A5D26236A85ULL, "vxht_brothel_n_6iu_0_0.vlm"},
	{0x3EC1B1F86E12E402ULL, "vxw_bonepitdestroyed_a3_1hu_0_0.vlm"},
	{0x3EC1E335C4EBCDC9ULL, "6150201_m.fxe"},
	{0x3EC1F8E1C601EEBBULL, "6125621_m.fxe"},
	{0x3EC266C94EDB3540ULL, "vxht_mansion_n_3f7_0_0.vlm"},
	{0x3EC2A3A1CE2C5B7DULL, "gen00fl_merrill6116003.cl"},
	{0x3EC369127BBE44D1ULL, "hlac_lt_0_lowtown_n_65536_9y.dds"},
	{0x3EC37B86AE9DECD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_modal_activate.gad"},
	{0x3EC39DFB4E5D4353ULL, "mag160_feynriel6020068.cl"},
	{0x3EC3B50B0D481CB9ULL, "vxw_encounter3_6t_0_0.vlm"},
	{0x3EC40B44AC1C91DCULL, "zz_one_debug6025811.cl"},
	{0x3EC4271DCEBA3CC4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_basic.mmh"},
	{0x3EC42B1DCEBA4317ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_basic.mao"},
	{0x3EC43D1DCEBA6266ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_basic.msh"},
	{0x3EC48C95C5DBE0A8ULL, "w_encounter2_2_0.msh"},
	{0x3EC49E95C5DBFF39ULL, "w_encounter2_2_0.mao"},
	{0x3EC4A38858CABD42ULL, "dae000_flemeth6013255.cl"},
	{0x3EC4A7E7C6974FDBULL, "sst_ss_fell_orden.gda"},
	{0x3EC4AA95C5DC13D2ULL, "w_encounter2_2_0.mmh"},
	{0x3EC509BE248D9FEBULL, "6014947_m.fxe"},
	{0x3EC54889D5F686A2ULL, "vxht_mansion2b_2nk_0_0.vlm"},
	{0x3EC54A581132982BULL, "6016634_f.fxe"},
	{0x3EC580055824774EULL, "vxlt_gallowscourtyard_7ct_0_0.vlm"},
	{0x3EC586E8746BD142ULL, "515377608.ogg"},
	{0x3EC6218EC7A36E60ULL, "557252233.ogg"},
	{0x3EC6FA34BFD1710AULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.bl_atk_v3.evt"},
	{0x3EC7140FC668CF8FULL, "hightown_fem_noble1_sta6169245.cl"},
	{0x3EC75016EDD94F22ULL, "gen00fl_sebastian6175006.cl"},
	{0x3EC78A54D2C492DBULL, "vxdr_0_deeproads_fhw_0_0.vlm"},
	{0x3EC7A2732FE2CA47ULL, "art\\vfx\\combat\\vfx_str_disapp_c.impa.gad"},
	{0x3EC7BDEA852C77A7ULL, "dae000_madam_lusine6151477.cl"},
	{0x3EC7C998E2D3F89CULL, "lt_0_lowtown_a2_n_c_44_0_lc.mmh"},
	{0x3EC7E63AEE9E8633ULL, "vxw_bonepit_i4_0_0.vlm"},
	{0x3EC836F18B098C64ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.impa.gad"},
	{0x3EC88D752A34B7D0ULL, "llac_ht_keep_n_3732_18.dds"},
	{0x3EC90EEE53C643F1ULL, "6137619_m.fxe"},
	{0x3EC9EAF8537D92B7ULL, "6169720_m.fxe"},
	{0x3ECA8158D559FE31ULL, "lt_gallowscourtyard_n_47_0.rml"},
	{0x3ECA9086AEA42F64ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_modal_activate.evt"},
	{0x3ECB1F2F4D55B118ULL, "vxlt_warehouse_n_xj_0_0.vlm"},
	{0x3ECC26DB804875E3ULL, "vxht_0_hightown_a3_b5f_0_0.vlm"},
	{0x3ECC8366D234FB58ULL, "dre000_varric6027743.cl"},
	{0x3ECCAF5B7BDE4E63ULL, "lt_0_lowtown_a2_n_21_0_lc.mmh"},
	{0x3ECCD3EC7BB9FF32ULL, "gen00fl_isabela6078979.cl"},
	{0x3ECCE3C785421349ULL, "vxht_mansion_n_6i2_0_0.vlm"},
	{0x3ECD677B1AD6F6A3ULL, "matdef_dcc_diffuse.dds"},
	{0x3ECD6DA35AAE04E3ULL, "6144475_m.fxe"},
	{0x3ECD7A56240B46C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.closedtoopen.ani"},
	{0x3ECD95081F8B0887ULL, "vxw_bonepit_yh_0_0.vlm"},
	{0x3ECE38D630D70D96ULL, "lt_darktown_water_8752.mao"},
	{0x3ECE44D630D72235ULL, "lt_darktown_water_8752.mmh"},
	{0x3ECE46D630D725ABULL, "lt_darktown_water_8752.msh"},
	{0x3ECE97732FE8D654ULL, "art\\vfx\\combat\\vfx_str_disapp_c.impa.evt"},
	{0x3ECF1F6527618D57ULL, "vxht_0_hightown_n_c8f_0_0.vlm"},
	{0x3ECF2C2DD5F93B0DULL, "6120111_f.fxe"},
	{0x3ECF45F18B0FC442ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.impa.ani"},
	{0x3ECF8D3DE210762BULL, "vxw_bonepitmining_a2_1jd_0_0.vlm"},
	{0x3ECFCCE3AE890818ULL, "vxlt_darktown_4tb_0_0.vlm"},
	{0x3ED0BA9D53A040ABULL, "vxlt_backalley_g_1y1_0_0.vlm"},
	{0x3ED0D16D45E41037ULL, "6005026_m.fxe"},
	{0x3ED15484776833E4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basicwood.mmh"},
	{0x3ED1588477683A37ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basicwood.mao"},
	{0x3ED16A8477685886ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basicwood.msh"},
	{0x3ED183A19D7F428BULL, "vxlt_0_lowtown_fi7_0_0.vlm"},
	{0x3ED200A1982C8F00ULL, "6033122_f.fxe"},
	{0x3ED2B6059129C8CCULL, "6039178_m.fxe"},
	{0x3ED3837DA17AE64BULL, "zz_dae_debug6049991.cl"},
	{0x3ED426FE046AA910ULL, "6125727_m.fxe"},
	{0x3ED42854ABE65F96ULL, "hlac_ht_0_hightown_a3_-1_0x.dds"},
	{0x3ED434AE3332AE9FULL, "vxht_bartrand_n_5r9_0_0.vlm"},
	{0x3ED465C871EFB3F9ULL, "vxlt_0_lowtown_a2_bz4_0_0.vlm"},
	{0x3ED4901BDCB64259ULL, "6202221_m.fxe"},
	{0x3ED4943FACFFC0CFULL, "qun211cr_crazy_woman.utc"},
	{0x3ED499C5B6D0BE4CULL, "llac_dr_dungeon_3215_9.dds"},
	{0x3ED527A6D67026C5ULL, "pro000cr_ogre_fake.ncs"},
	{0x3ED560C336FA5724ULL, "944602562.ogg"},
	{0x3ED58C43BA55528BULL, "6129730_f.fxe"},
	{0x3ED629230F1CEA99ULL, "vxht_0_hightown_a3_2xc_0_0.vlm"},
	{0x3ED67155C63E4638ULL, "273437464.ogg"},
	{0x3ED73C12092C84ACULL, "299119140.ogg"},
	{0x3ED740DFFAF8E7B4ULL, "w_bonepitdestroyed_a3_23_0.rml"},
	{0x3ED76BE196E4CE8DULL, "662711332.ogg"},
	{0x3ED7D24685731130ULL, "6016160_m.fxe"},
	{0x3ED7D63E242A3C67ULL, "vxlt_undercitydungeon_200_0_0.vlm"},
	{0x3ED7D8D7F4194F86ULL, "war_accuracy.dds"},
	{0x3ED7E74F7D3015B8ULL, "zz_hawke_cu_ve2_night.cut"},
	{0x3ED84022230DA299ULL, "hlac_lt_0_lowtown_-65535_3x.dds"},
	{0x3ED858A6FCA477D3ULL, "vxlt_0_lowtown_er9_0_0.vlm"},
	{0x3ED873F01056FF60ULL, "259510671.ogg"},
	{0x3ED87C2684EB0283ULL, "266095403.ogg"},
	{0x3ED8A886AEB0955DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_modal_activate.ani"},
	{0x3ED8EAA05F716A08ULL, "vxht_mansion2b_3c7_0_0.vlm"},
	{0x3ED9590CB6B5BC1CULL, "6122061_m.fxe"},
	{0x3ED9A56725BFDEF1ULL, "vxht_mansion2_31l_0_0.vlm"},
	{0x3EDA68DAB7B5FBD4ULL, "art\\vfx\\combat\\vfx_d_qun_direct_cast_c.mmh"},
	{0x3EDB206BADE2482FULL, "6070215_m.fxe"},
	{0x3EDB2CCA707398F9ULL, "hlac_ht_0_hightown_65537_2x.dds"},
	{0x3EDB59865CDC69FBULL, "llac_dr_slavers_3217_6.dds"},
	{0x3EDB625624175B37ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.closedtoopen.evt"},
	{0x3EDB857ABBB242BAULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_noba_0d.dds"},
	{0x3EDB8F732FF3528DULL, "art\\vfx\\combat\\vfx_str_disapp_c.impa.ani"},
	{0x3EDBAA8D7FE18EFDULL, "framebuffereffects_ps3zcullreload_p0.psh.fxo.dx11.assembly"},
	{0x3EDBC3B8554498B5ULL, "6095003_m.fxe"},
	{0x3EDBCEF566A2043EULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_003a_0n.dds"},
	{0x3EDBDF804CA41D63ULL, "6141775_m.fxe"},
	{0x3EDBEE29D4C4C0CEULL, "vxdr_primevaltomb_85y_0_0.vlm"},
	{0x3EDC06E511812107ULL, "vxw_bonepit_dk_0_0.vlm"},
	{0x3EDC385AE9A27F16ULL, "qun221_petrice6077318.cl"},
	{0x3EDC739FCF37E182ULL, "vxdr_primevaltomb_8dj_0_0.vlm"},
	{0x3EDC83AE96690E22ULL, "6217093_m.fxe"},
	{0x3EDD4878CC2C234EULL, "6079045_f.fxe"},
	{0x3EDD52AEA8DE77EEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v1.evt"},
	{0x3EDDF8EF5D3F8C04ULL, "6041274_m.fxe"},
	{0x3EDE3042FB8DFA95ULL, "6021350_m.fxe"},
	{0x3EDE312838429B78ULL, "vxw_encounter2_6r_0_0.vlm"},
	{0x3EDEF434ECF85A53ULL, "hlac_ht_chantry_n_2534_10x.dds"},
	{0x3EDF40FB060FDD4EULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v3.ani"},
	{0x3EDFBF88DD60F364ULL, "t3_wdg_gn3.tnt"},
	{0x3EDFD32F1EA004A6ULL, "vxht_bartrand_n_5ws_0_0.vlm"},
	{0x3EDFEFCD04C9B9D2ULL, "6125202_m.fxe"},
	{0x3EE007CE0DA1E33DULL, "248867549.ogg"},
	{0x3EE031A355326C81ULL, "850251103.ogg"},
	{0x3EE0492B9A8E843AULL, "177720110.ogg"},
	{0x3EE0C56D92945813ULL, "vxdr_primevalentrance_227_0_0.vlm"},
	{0x3EE109AEEC16D5BBULL, "480146046.ogg"},
	{0x3EE1DDEDE53DD74CULL, "ser221_gamlen6135373.cl"},
	{0x3EE2275280E0FBE1ULL, "plt_genpt_party_triggers.nss"},
	{0x3EE26356241D7BC0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.closedtoopen.gad"},
	{0x3EE270B1C8E5BFA1ULL, "6205651_m.fxe"},
	{0x3EE349AEA8E2D399ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v1.gad"},
	{0x3EE3879CD275BD06ULL, "genip_node_i2.utp"},
	{0x3EE3BD9162028575ULL, "vxdr_primevaltomb_3a9_0_0.vlm"},
	{0x3EE46390779771E8ULL, "21963373.ogg"},
	{0x3EE617796FB70C81ULL, "cd_qcr220_meredith_viscount.bnk"},
	{0x3EE618A6D67E800FULL, "pro000cr_ogre_fake.utc"},
	{0x3EE64C32A006867EULL, "76666184.ogg"},
	{0x3EE68E40CAF6F0F0ULL, "vxlt_0_lowtown_n_bys_0_0.vlm"},
	{0x3EE75C7D51678288ULL, "vxht_mansion2b_3f4_0_0.vlm"},
	{0x3EE86C36F3F285DCULL, "6074407_m.fxe"},
	{0x3EE895B5C23155EAULL, "t3_dog_d02.tnt"},
	{0x3EE90C2F430C8151ULL, "vxw_bonepit_1os_0_0.vlm"},
	{0x3EE95FF13F4538DAULL, "hf_ave3_insurgent_guard3.mor"},
	{0x3EE992489FD860AFULL, "6075115_m.fxe"},
	{0x3EEA5AAEA8E90FE7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v1.ani"},
	{0x3EEA6C2B042B0E32ULL, "vxht_chantry_21f_0_0.vlm"},
	{0x3EEA78C203773987ULL, "vxw_bonepit_oj_0_0.vlm"},
	{0x3EEAE7BE4AA091AFULL, "281638232.ogg"},
	{0x3EEB2EF38BA46158ULL, "hlac_lt_0_lowtown_-65536_3x.dds"},
	{0x3EEB8B7593C452FFULL, "vxlt_0_lowtown_docks_bcd_0_0.vlm"},
	{0x3EEBE71187DB862EULL, "gen_im_npc_arm_hum_glv_cotc.uti"},
	{0x3EEBEAFE201061BFULL, "custic00b_a.dds"},
	{0x3EEC1A972E1DA3A0ULL, "lgt220_shp_mag_guard6149896.cl"},
	{0x3EEC50DC46C5220DULL, "lgt260_eau_tmp_ending6171810.cl"},
	{0x3EECA3052A38B3F8ULL, "vxw_encounter2_1s_0_0.vlm"},
	{0x3EECC00B21C42162ULL, "dre000_sibling_has_blight6036508.cl"},
	{0x3EED18373535CD46ULL, "464407342.ogg"},
	{0x3EED2180A544A074ULL, "5558241.ogg"},
	{0x3EED5A55F3D10D45ULL, "vxlt_0_lowtown_a2_gz1_0_0.vlm"},
	{0x3EEDC583C5982F77ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_lp.gad"},
	{0x3EEDCE9E7D7EB9B0ULL, "dark_fog_b.mao"},
	{0x3EEDDD9631EF1095ULL, "art\\vfx\\textures\\level_fx\\vfx_bird_02.dds"},
	{0x3EEE0744CE159E36ULL, "follower_banter6172881.cl"},
	{0x3EEE3EE58D55C89BULL, "481818014.ogg"},
	{0x3EEE3F4881654ECBULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.dura.evt"},
	{0x3EEE5A39516BFAA4ULL, "qun231_varnell6152904.cl"},
	{0x3EEEB194872C2253ULL, "vxca_cave_1a4_0_0.vlm"},
	{0x3EEEB9D010B3A58FULL, "6042074_m.fxe"},
	{0x3EEF5EA08B3974CDULL, "t1_mub_wh2.tnt"},
	{0x3EEF7A036D13CDACULL, "6103648_m.fxe"},
	{0x3EEFD753A72C0619ULL, "gui_guipixelshader_p1.psh.fxo"},
	{0x3EF03C09593ABE24ULL, "hlac_lt_0_lowtown_docks_a3_0_1x.dds"},
	{0x3EF07410E0B456E6ULL, "qun160_vashothlead6055573.cl"},
	{0x3EF074D515AF3A0FULL, "gen00fl_merrill6119624.cl"},
	{0x3EF09E20ABD8D022ULL, "vxht_mansion2b_2av_0_0.vlm"},
	{0x3EF0ED135531E734ULL, "66521509.ogg"},
	{0x3EF145E7D61779C8ULL, "6137489_m.fxe"},
	{0x3EF1622852CB6891ULL, "vxht_0_hightown_atq_0_0.vlm"},
	{0x3EF1764BBC0CB295ULL, "6136191_m.fxe"},
	{0x3EF178732CB308E7ULL, "6078740_f.fxe"},
	{0x3EF1C4E8D23FA868ULL, "6121254_f.fxe"},
	{0x3EF1E0F494D41BFDULL, "cod_hst_qunari.utp"},
	{0x3EF214F4E7ADE493ULL, "vxht_0_hightown_n_a88_0_0.vlm"},
	{0x3EF28240E60CAB8FULL, "mage_attractionwave.dds"},
	{0x3EF28A2CCA29C9F9ULL, "seb221_flora_harimann.bnk"},
	{0x3EF2AAA2BD9D916CULL, "437937025.ogg"},
	{0x3EF2FD5C0A525C07ULL, "vxca_caverns2prc_9t1_0_0.vlm"},
	{0x3EF31BA9AF873AFDULL, "624694811.ogg"},
	{0x3EF38D28E05D4BC2ULL, "1046877775.ogg"},
	{0x3EF39D470724F011ULL, "w_bonepitdestroyed_a3_140_0.mmh"},
	{0x3EF3A24766F351F4ULL, "dae000_ineria6140549.cl"},
	{0x3EF3A947072504B2ULL, "w_bonepitdestroyed_a3_140_0.mao"},
	{0x3EF3B74707251CCFULL, "w_bonepitdestroyed_a3_140_0.msh"},
	{0x3EF419FB0621F688ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v3.gad"},
	{0x3EF47D877C8E5FEAULL, "vxdr_0_deeproads_cui_0_1.vlm"},
	{0x3EF48AC2C4896A24ULL, "6096765_f.fxe"},
	{0x3EF48FC22BAEA016ULL, "vxlt_undercityencounter_1ru_0_0.vlm"},
	{0x3EF498D41DE7BBD2ULL, "primitives_primitive_p4.psh.fxo.dx11.assembly"},
	{0x3EF4DA83C59E7104ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_lp.evt"},
	{0x3EF5012E11EA0942ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_twitch.gad"},
	{0x3EF521092126BC93ULL, "face_face_p0.psh.fxo.dx11"},
	{0x3EF55848816B978CULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.dura.gad"},
	{0x3EF5AC2CCA2C3959ULL, "seb221_flora_harimann.cnv"},
	{0x3EF5B0C48A354132ULL, "hlac_w_woundedcoast_-262136_0z.dds"},
	{0x3EF5FE9C27C053A8ULL, "vxlt_0_lowtown_docks_cqt_0_0.vlm"},
	{0x3EF6813B4F2E4BE8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_basic.mmh"},
	{0x3EF68D3B4F2E5F8BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_basic.mao"},
	{0x3EF69F3B4F2E7E12ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_basic.msh"},
	{0x3EF7A7F8162047E0ULL, "6128889_m.fxe"},
	{0x3EF7B4ECC79857F0ULL, "vxlt_0_lowtown_h89_0_0.vlm"},
	{0x3EF7EBE576C5A39BULL, "326171843.ogg"},
	{0x3EF8324A4880C0D2ULL, "vxca_cave_1a1_0_0.vlm"},
	{0x3EF83F1FD0024B20ULL, "hlac_lt_0_lowtown_a2_-196608_0x.dds"},
	{0x3EF862625040FE14ULL, "art\\vfx\\combat\\vfx_glb_potion_drink_c.anb"},
	{0x3EF8AFF7ADDCFE8CULL, "ht_mansion_n_9156_lc.mmh"},
	{0x3EF92D34A6B03904ULL, "gen00fl_carver6102573.cl"},
	{0x3EF9FD5285BA6B7FULL, "vxlt_0_lowtown_docks_bfg_0_0.vlm"},
	{0x3EFA24396DD96D22ULL, "6200743_m.fxe"},
	{0x3EFA37DDDB1B1379ULL, "hlac_ht_0_hightown_a2_65536_3z.dds"},
	{0x3EFA43AF7F48A797ULL, "vxlt_gallowscourtyard_n_73r_0_0.vlm"},
	{0x3EFA97C2FF9F68E9ULL, "570344682.ogg"},
	{0x3EFADD8812FF1622ULL, "hlac_lt_gallowscourtyard_n_-131071_3z.dds"},
	{0x3EFAF0DA19E02FABULL, "plt_cod_bks_lite_foreshadow.nss"},
	{0x3EFB227A846B5198ULL, "vxlt_darktown_4ea_0_0.vlm"},
	{0x3EFB57488170018AULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.dura.ani"},
	{0x3EFB6805EF0FC041ULL, "vxw_woundedcoast_138_0_0.vlm"},
	{0x3EFB92369DC27272ULL, "hlac_w_blightlands_393212_0y.dds"},
	{0x3EFBD66A29D0BF1BULL, "vxdr_primevaltomb_3m3_0_0.vlm"},
	{0x3EFC242E11F06339ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_twitch.evt"},
	{0x3EFC2A25B734DFFAULL, "vxlt_0_lowtown_a2_gn6_0_0.vlm"},
	{0x3EFC4ACBDA8C71D1ULL, "lt_0_lowtown_docks_207_0.msh"},
	{0x3EFC58CBDA8C89ECULL, "lt_0_lowtown_docks_207_0.mao"},
	{0x3EFC64CBDA8C9D8FULL, "lt_0_lowtown_docks_207_0.mmh"},
	{0x3EFC729DA21C78D1ULL, "vxht_0_hightown_a2_ca8_0_0.vlm"},
	{0x3EFCD12727E95547ULL, "llac_lt_0_lowtown_-131072_5.dds"},
	{0x3EFD25ED8C4819ACULL, "vrc321ip_spooky7_obj_11.utp"},
	{0x3EFD4EFC783F4B84ULL, "dae000_gen_turn_in_56198669.cl"},
	{0x3EFD66AB2414FF9FULL, "gen_im_npc_dwf_arm_hlm_wara.uti"},
	{0x3EFDC731DEB51CC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_loop.ani"},
	{0x3EFDCB5A01FD1BCBULL, "903589153.ogg"},
	{0x3EFE1F63BF2A3798ULL, "603133291.ogg"},
	{0x3EFE227614A9436CULL, "6128719_m.fxe"},
	{0x3EFEA7A32E3745D4ULL, "928818266.ogg"},
	{0x3EFFE1D5334F4B58ULL, "vxlt_gallowstemplar_n_650_0_0.vlm"},
	{0x3EFFF0F4AAD0CB0BULL, "vxht_mansion_n_5bz_0_0.vlm"},
	{0x3F0018D39B83A917ULL, "hlac_ca_caverns1_24424_26z.dds"},
	{0x3F0026DD53D53907ULL, "vxlt_0_lowtown_n_f4o_0_0.vlm"},
	{0x3F0056C6201524BAULL, "vxlt_warehouse_cp_0_0.vlm"},
	{0x3F0084235512157BULL, "vxdr_0_deeproads_bb1_0_0.vlm"},
	{0x3F00B87612397FC3ULL, "6089505_m.fxe"},
	{0x3F00C5B563BC35D2ULL, "138868893.ogg"},
	{0x3F0124328E5BEA6EULL, "lt_0_lowtown_docks_a3_n_178_0.mmh"},
	{0x3F0130328E5BFE05ULL, "lt_0_lowtown_docks_a3_n_178_0.mao"},
	{0x3F013E328E5C165CULL, "lt_0_lowtown_docks_a3_n_178_0.msh"},
	{0x3F01827A11787C15ULL, "vxlt_hangedman_n_1i7_0_0.vlm"},
	{0x3F01F283C5A924FDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_lp.ani"},
	{0x3F02025A9BFEC846ULL, "6124780_m.fxe"},
	{0x3F027FE38880A02FULL, "vxdr_primevaltomb_8f1_0_0.vlm"},
	{0x3F02D9B989D1BF6FULL, "art\\characters\\playercharacter\\humanfemale\\arm_medb.phy"},
	{0x3F0321DF4A0C12C6ULL, "374345778.ogg"},
	{0x3F036BDAB7D9058BULL, "art\\vfx\\combat\\vfx_d_qun_direct_cast_c.anb"},
	{0x3F03E231DEB9B64BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_loop.gad"},
	{0x3F057A9FAD452498ULL, "6135872_m.fxe"},
	{0x3F064162504D03D7ULL, "art\\vfx\\combat\\vfx_glb_potion_drink_c.mmh"},
	{0x3F0646979F712657ULL, "605621502.ogg"},
	{0x3F066E342C28EB72ULL, "vxlt_0_lowtown_docks_b66_0_0.vlm"},
	{0x3F068D527237CF95ULL, "vxht_mansion_n_3w9_0_0.vlm"},
	{0x3F06D4CE4A4E2D5CULL, "568930450.ogg"},
	{0x3F0799E9A9B7CC6FULL, "plt_cod_ltr_otl_comte.plo"},
	{0x3F07A94A6AB9EFE6ULL, "vxca_caverns1_ico_0_0.vlm"},
	{0x3F07C936A51A2B02ULL, "vxdr_primevaltomb_8uu_0_0.vlm"},
	{0x3F081FA2CE1E8027ULL, "301941566.ogg"},
	{0x3F08CD78B5316B6DULL, "6079082_f.fxe"},
	{0x3F08CE8E59B74D50ULL, "mag160_varian_ilithis6016488.cl"},
	{0x3F08ED2D4F8A6B39ULL, "944782634.ogg"},
	{0x3F09074F356E63E8ULL, "6126616_m.fxe"},
	{0x3F092C2E11FAFB28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_twitch.ani"},
	{0x3F0950031D128901ULL, "hlac_ht_keep_n_3736_0z.dds"},
	{0x3F0953697511B424ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy.ani"},
	{0x3F0954B1A9E8D82BULL, "vxw_bonepitmining_a2_1vx_0_0.vlm"},
	{0x3F096170AB305798ULL, "6200107_m.fxe"},
	{0x3F0965D32D6E3DCEULL, "vxw_bonepitdestroyed_a3_1pq_0_0.vlm"},
	{0x3F09B9EAEADE8F96ULL, "follower_banter6143218.cl"},
	{0x3F0A333A584FE44AULL, "vxht_mansion2_2av_0_0.vlm"},
	{0x3F0A56DD59C6A868ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf05a.msh"},
	{0x3F0A5B292066BCD0ULL, "483010165.ogg"},
	{0x3F0A5EBE8524FAD1ULL, "vxca_caverns1_efw_0_0.vlm"},
	{0x3F0A6C495595626BULL, "vxht_0_hightown_b75_0_0.vlm"},
	{0x3F0A74DD59C6DA92ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf05a.mmh"},
	{0x3F0AA26D894F725FULL, "vxht_mansion2_3bk_0_0.vlm"},
	{0x3F0B2300B467E3DCULL, "hlac_lt_0_lowtown_n_-65536_4z.dds"},
	{0x3F0B269900DC6B35ULL, "791177434.ogg"},
	{0x3F0B3261EBF79F56ULL, "mcr310_cut_chantry_blows_up6087554.cl"},
	{0x3F0B6E6D512C1747ULL, "6780581.ogg"},
	{0x3F0C7DEF8CEE6551ULL, "mag320cr_nobleman_3.utc"},
	{0x3F0C88B9B4AA4D05ULL, "vxw_encounter2_6k_0_0.vlm"},
	{0x3F0CA75C00491425ULL, "254126810.ogg"},
	{0x3F0CEF8AB7F1F330ULL, "dre000_bodahn6033402.cl"},
	{0x3F0D28F8F1088A37ULL, "mnd_con_dust.mao"},
	{0x3F0D8000726E2524ULL, "w_bonepitmining_a2_220_0.msh"},
	{0x3F0D9200726E438DULL, "w_bonepitmining_a2_220_0.mao"},
	{0x3F0D943C0E8D7B6EULL, "zz_mcr_debug6080784.cl"},
	{0x3F0D9600726E4AC6ULL, "w_bonepitmining_a2_220_0.mmh"},
	{0x3F0DEEE51A2B5B15ULL, "gen00fl_aveline6122768.cl"},
	{0x3F0DFD44A3788953ULL, "7743765.ogg"},
	{0x3F0E7F1A0A38B19BULL, "2224041.ogg"},
	{0x3F0F17B8F774068BULL, "llac_lt_0_lowtown_a2_n_c_-65535_4.dds"},
	{0x3F0F19768C7C84A9ULL, "hlac_lt_gamlenshouse_n_-1_1x.dds"},
	{0x3F0F3E601E704CBBULL, "qun120cr_javaris6043236.cl"},
	{0x3F0F506975161A86ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy.gad"},
	{0x3F0F72CFAFCDFA19ULL, "dae011_corff_sta6142567.cl"},
	{0x3F0F9632B74DDC9BULL, "354070112.ogg"},
	{0x3F0FA9B53205FF78ULL, "6167677_m.fxe"},
	{0x3F0FB79984451DA2ULL, "ave200_donnic6111719.cl"},
	{0x3F0FC79619F859C1ULL, "w_woundedcoast_1_0_ll.mmh"},
	{0x3F0FCAF51967C68BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_exit.evt"},
	{0x3F0FFBB0EFFA3517ULL, "370724679.ogg"},
	{0x3F0FFE9CDA591971ULL, "ser221_party_comments6166115.cl"},
	{0x3F106F225C478954ULL, "zz_approval_debug6003676.cl"},
	{0x3F10BB125A75370AULL, "267904099.ogg"},
	{0x3F1140DFFB2A6077ULL, "w_bonepitdestroyed_a3_23_0.mmh"},
	{0x3F1144DFFB2A6684ULL, "w_bonepitdestroyed_a3_23_0.mao"},
	{0x3F1156DFFB2A85E9ULL, "w_bonepitdestroyed_a3_23_0.msh"},
	{0x3F116E8488F139A0ULL, "vxw_encounter3_dg_0_0.vlm"},
	{0x3F11C276DFCD7B94ULL, "6151496_m.fxe"},
	{0x3F1249C0A9631278ULL, "ser270_emeric6149281.cl"},
	{0x3F1265B208035BB2ULL, "mcr310_argument6127699.cl"},
	{0x3F129FCF3303CE74ULL, "6042323_m.fxe"},
	{0x3F13217D5788390DULL, "vxdr_primevaltomb_7yu_0_0.vlm"},
	{0x3F13893E97899C16ULL, "82773661.ogg"},
	{0x3F1416F83637D91BULL, "98943233.ogg"},
	{0x3F146AA427D1475EULL, "6027544_m.fxe"},
	{0x3F1470D30D0ED3F9ULL, "vxht_0_hightown_a2_9uz_0_0.vlm"},
	{0x3F14A65BAEDB4445ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_0x.dds"},
	{0x3F151C8317C0442EULL, "l_one121ar_warehouse.wbd"},
	{0x3F1545EA9CC69F7FULL, "vxw_bonepitdestroyed_a3_1zr_0_0.vlm"},
	{0x3F1578591003A862ULL, "lt_0_lowtown_docks_n_20_0_ll.mmh"},
	{0x3F1578C0F2FDE471ULL, "dae000_arishok6068483.cl"},
	{0x3F15C47622ED118FULL, "vxw_bonepitdestroyed_a3_1c9_0_0.vlm"},
	{0x3F161887A6BC5FA4ULL, "dae000_party_night_comment6201616.cl"},
	{0x3F164B69751C3165ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy.evt"},
	{0x3F16790A822A70A8ULL, "6081388_m.fxe"},
	{0x3F16E3F5196E104CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_exit.gad"},
	{0x3F16EA304F5F1E5DULL, "pro000_athenril6049458.cl"},
	{0x3F16F9402A46D163ULL, "store_dae110_weapons.utm"},
	{0x3F1731861CB8ED78ULL, "6088840_m.fxe"},
	{0x3F17581055DD6C1DULL, "mag211_marethari6025339.cl"},
	{0x3F17A21BD2517E2FULL, "mrl000_marethari6129233.cl"},
	{0x3F17F48324841C58ULL, "vxht_0_hightown_a3_b4z_0_0.vlm"},
	{0x3F1839FCE664832CULL, "static_static_p30.psh.fxo.dx11"},
	{0x3F18A5174A45FCCAULL, "vxht_mansion2_2jw_0_0.vlm"},
	{0x3F18EDA053BB35CCULL, "93816786.ogg"},
	{0x3F18F00EDC0C1A38ULL, "vxw_bonepitmining_a2_268_0_0.vlm"},
	{0x3F191AAF4BF6D92FULL, "6163245_m.fxe"},
	{0x3F191D1028C5EC86ULL, "vxw_bonepitmining_a2_26v_0_0.vlm"},
	{0x3F194A23EF8F76D3ULL, "gen00fl_anders6103676.cl"},
	{0x3F1963F51F4EFE67ULL, "6108574_fenris_get_up.cut"},
	{0x3F19689612E5BC20ULL, "vxlt_gallowstemplar_n_46s_0_0.vlm"},
	{0x3F19DE80863F7608ULL, "177301863.ogg"},
	{0x3F1A15E91670F20BULL, "vxlt_hangedman_1qz_0_0.vlm"},
	{0x3F1AFA96A6A06585ULL, "vxw_encounter2_1j_0_0.vlm"},
	{0x3F1B125F2D97064BULL, "vxlt_gallowscourtyard_n_70o_0_0.vlm"},
	{0x3F1B856C9B89F8DAULL, "vxlt_0_lowtown_ef8_0_0.vlm"},
	{0x3F1BBB11E5CCE4C0ULL, "vxca_cavern1_haunted_ehi_0_0.vlm"},
	{0x3F1BC13EC0432967ULL, "vxlt_0_lowtown_n_gj4_0_0.vlm"},
	{0x3F1C04C76D8EFE61ULL, "w_encounter1_1_0.phy"},
	{0x3F1C0BE38317238FULL, "110871405.ogg"},
	{0x3F1C30C674FAD911ULL, "6167056_m.fxe"},
	{0x3F1C5322AAAAE2DEULL, "lgt323pt_shw_ors_blckop.ncs"},
	{0x3F1C6124FC5E4FB6ULL, "mer120_hubert6036656.cl"},
	{0x3F1C7C2A914C2F69ULL, "mag100_thrask6035904.cl"},
	{0x3F1CD0EB0398F8BCULL, "6128971_m.fxe"},
	{0x3F1CFDA9CD43E622ULL, "6095826_m.fxe"},
	{0x3F1D40182FACFA71ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_act_c.anb"},
	{0x3F1DA6298549824AULL, "vxlt_shop_59_0_0.vlm"},
	{0x3F1DB4A164D6C485ULL, "581836721.ogg"},
	{0x3F1DE01E7275195FULL, "tbe100ip_shop_codex_4.utp"},
	{0x3F1DE2F519742D4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_exit.ani"},
	{0x3F1E25F25D8AFD73ULL, "6034224_m.fxe"},
	{0x3F1E8368A373A981ULL, "dae011_drunk1_sta6149682.cl"},
	{0x3F1EC1792DAD4ED4ULL, "558471231.ogg"},
	{0x3F1ED4DA9398EE82ULL, "hlac_w_bonepit_327680_0z.dds"},
	{0x3F1F3500F0D5F7CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_second_wind.gad"},
	{0x3F1F75D0A4027255ULL, "6116593_m.fxe"},
	{0x3F1FA567155EE31BULL, "vxlt_undercityserialkiller_yv_0_0.vlm"},
	{0x3F1FBF1B4DECAD58ULL, "dae000_orsino6187913.cl"},
	{0x3F1FCB7FC2A1FE7BULL, "gen00fl_merrill6109557.cl"},
	{0x3F1FE0617AE75220ULL, "vxw_encounter3_ad_0_0.vlm"},
	{0x3F200B33A419B87FULL, "6120918_m.fxe"},
	{0x3F20252AAE2B8CF3ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.cess.evt"},
	{0x3F20547821B15FCBULL, "398023329.ogg"},
	{0x3F206FD39CCCAF3FULL, "6103265_m.fxe"},
	{0x3F20A9881565D50AULL, "gen00fl_sebastian6180846.cl"},
	{0x3F20A995C62A70CDULL, "w_encounter2_2_0.phy"},
	{0x3F2100D9D8D18167ULL, "6166270_f.fxe"},
	{0x3F2173BC8484F12DULL, "6102871_m.fxe"},
	{0x3F217C92BC9927B9ULL, "vxht_0_hightown_n_b8q_0_0.vlm"},
	{0x3F21D690DF0E6096ULL, "qun221_petrice6076685.cl"},
	{0x3F21E3E72156042AULL, "vxlt_darktown_56o_0_0.vlm"},
	{0x3F230DC76D952C7DULL, "w_encounter1_1_0.rml"},
	{0x3F238500117955EBULL, "mrl310_party_comments6143060.cl"},
	{0x3F239F6A638227F2ULL, "6107098_f.fxe"},
	{0x3F24185008EC3910ULL, "vxht_0_hightown_a6f_0_0.vlm"},
	{0x3F24365314E32A0FULL, "vxw_bonepitdestroyed_a3_1d8_0_0.vlm"},
	{0x3F24C92628E74160ULL, "hlac_ca_caverns1_mines_24422_3y.dds"},
	{0x3F24FCEB94A31B36ULL, "mag131_taraneh6086039.cl"},
	{0x3F2504ABE0639AA9ULL, "6039979_m.fxe"},
	{0x3F2567C1E711A9F5ULL, "6020093_f.fxe"},
	{0x3F261157DE1B84DDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small_1.mao"},
	{0x3F2640C79CAE4546ULL, "vxlt_darktown_5cl_0_0.vlm"},
	{0x3F264A00F0DC3A7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_second_wind.evt"},
	{0x3F26662B526D01F8ULL, "fen310_fenris6117673.cl"},
	{0x3F26AA95C62EDE79ULL, "w_encounter2_2_0.rml"},
	{0x3F26BDD51C554A78ULL, "vxw_blightlands_qx_0_0.vlm"},
	{0x3F2716F43C3C154AULL, "vxht_mansion2_2wp_0_0.vlm"},
	{0x3F271E2AAE319F34ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.cess.gad"},
	{0x3F27F85F67D948A6ULL, "llac_lt_0_lowtown_a2_n_c_-131073_2.dds"},
	{0x3F280C7FD86E7F26ULL, "vxht_bartrand_n_5nx_0_0.vlm"},
	{0x3F281BC4034D64E7ULL, "water_water_p4.psh.fxo.dx11"},
	{0x3F282BB0E79DEF75ULL, "face_face_p0.vsh.fxo.dx11"},
	{0x3F2914AB97BEA397ULL, "vxlt_0_lowtown_docks_a3_n_d18_0_0.vlm"},
	{0x3F2926F8FD24AF80ULL, "vxht_0_hightown_a2_byx_0_0.vlm"},
	{0x3F29551EE3F09BBAULL, "ser111ip_bone_sack.utp"},
	{0x3F29727746EE0E73ULL, "6118980_m.fxe"},
	{0x3F2993F1B56A47B9ULL, "vxht_keep_n_6jc_0_0.vlm"},
	{0x3F29BFCFC2DAB285ULL, "vxw_blightlands_fd_0_0.vlm"},
	{0x3F2A3D6307F1A7D0ULL, "vxlt_backalley_n_1jn_0_0.vlm"},
	{0x3F2B61182FB96FEAULL, "art\\vfx\\combat\\vfx_m_bld_mgc_act_c.mmh"},
	{0x3F2BAC5CF5F76E3AULL, "vxlt_warehouse_rs_0_0.vlm"},
	{0x3F2C0FAB3C86444DULL, "vxdr_0_deeproads_f5b_0_0.vlm"},
	{0x3F2C6AFF642F7120ULL, "vxht_mansion2_35m_0_0.vlm"},
	{0x3F2C97CF4F8115F3ULL, "6031124_m.fxe"},
	{0x3F2D25796FF32733ULL, "cd_qcr220_meredith_viscount.wbd"},
	{0x3F2D2D2AAE362412ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.cess.ani"},
	{0x3F2D75B129A105AFULL, "gen00fl_isabela6090857.cl"},
	{0x3F2E3D02A8815E4AULL, "aip_templararcher_base.gda"},
	{0x3F2E7BC66360BEA9ULL, "vxx_narrativeroom_3c_0_0.vlm"},
	{0x3F2E90E84B01396BULL, "mag291_demon_marethari6028855.cl"},
	{0x3F2EE1B08EC2C7BFULL, "6104365_m.fxe"},
	{0x3F2EFF9EB34A78AFULL, "w_bonepitdestroyed_a3_219_0.mao"},
	{0x3F2F0B9EB34A8D4CULL, "w_bonepitdestroyed_a3_219_0.mmh"},
	{0x3F2F119EB34A973EULL, "w_bonepitdestroyed_a3_219_0.msh"},
	{0x3F2F25B952A4854BULL, "lt_0_lowtown_n_42_0_ll.mmh"},
	{0x3F2F591267C80D00ULL, "vxlt_hangedman_2er_0_0.vlm"},
	{0x3F2F83EBB52C0B39ULL, "gen00fl_dog6107356.cl"},
	{0x3F2FB73419BF6677ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.dura.gad"},
	{0x3F30674F89F7A084ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hgrass2_darelief.dds"},
	{0x3F306D9E79A1A0C0ULL, "dae000ip_crowd_ff8_talk14.utp"},
	{0x3F30B489F5322440ULL, "mcr370_assignments6177466.cl"},
	{0x3F31F272E9C5BF55ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbreadoven_0.pwk"},
	{0x3F32795A8C25B649ULL, "seb221_flora_harimann6187802.cl"},
	{0x3F32987B03614706ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.evt"},
	{0x3F33029C2EF4CE90ULL, "hlac_ht_0_hightown_a2_2_0y.dds"},
	{0x3F331B543892D1C9ULL, "dae000_elegant6166911.cl"},
	{0x3F3346DE25E6D082ULL, "569436962.ogg"},
	{0x3F335200F0E6D195ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_second_wind.ani"},
	{0x3F33899BBBBB1499ULL, "679398843.ogg"},
	{0x3F33EF2EE21483FCULL, "672211948.ogg"},
	{0x3F340AF3851C8493ULL, "vxca_caverns1_jfn_0_0.vlm"},
	{0x3F349A8F637F1528ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_sword_1.mao"},
	{0x3F34BB6A0350874EULL, "vxw_bonepitdestroyed_a3_1ct_0_0.vlm"},
	{0x3F34CA33DF748B39ULL, "6137196_m.fxe"},
	{0x3F34E71A758C247DULL, "449813310.ogg"},
	{0x3F351484F382C37CULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_pc_arishokresponse.ani"},
	{0x3F353CB324807E93ULL, "6043811_m.fxe"},
	{0x3F359CD502FFDC91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_reaction.ani"},
	{0x3F35A4FA11327D39ULL, "vxlt_0_lowtown_a2_eox_0_0.vlm"},
	{0x3F35F5A86B5AF652ULL, "llac_ht_0_hightown_a3_0_0.dds"},
	{0x3F360C85959FE628ULL, "ser120cr_shade_cut.utc"},
	{0x3F3626DC3138E5BDULL, "6018538_m.fxe"},
	{0x3F364DB517B30542ULL, "6201574_f.fxe"},
	{0x3F36548F0313110DULL, "follower_barks6153873.cl"},
	{0x3F3696AB30533479ULL, "vxw_bonepit_81_0_0.vlm"},
	{0x3F369F055B6A229AULL, "6052944_m.fxe"},
	{0x3F36CC3419C5A804ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.dura.evt"},
	{0x3F37842160D8D802ULL, "6119295_f.fxe"},
	{0x3F37A8686833459AULL, "vxlt_0_lowtown_a2_n_hff_0_0.vlm"},
	{0x3F37C6FBDB675240ULL, "vxht_chantry_n_2oy_0_0.vlm"},
	{0x3F37F21021DD2BBAULL, "gen00fl_isabela6075301.cl"},
	{0x3F37FE014661A4D6ULL, "6113092_m.fxe"},
	{0x3F380790BA54F6ACULL, "6119500_m.fxe"},
	{0x3F3831ACB4D0CB05ULL, "vxw_blightlands_kc_0_0.vlm"},
	{0x3F397A47F42157CBULL, "vxca_caverns1_mines_jgp_0_0.vlm"},
	{0x3F39817A0C6BFD16ULL, "6131139_m.fxe"},
	{0x3F39AF7B03678D51ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.gad"},
	{0x3F39D1273B4F6198ULL, "zz_qun_debug6064337.cl"},
	{0x3F3A0BCA7B1E167CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_roundtable_n.dds"},
	{0x3F3A1E39E7ED86BAULL, "vxlt_warehouse_wl_0_0.vlm"},
	{0x3F3A2526500D9EE2ULL, "ser221_party_comments6209853.cl"},
	{0x3F3A2D90D4613B51ULL, "mcr371_post_harvester6083685.cl"},
	{0x3F3B15A63BC4AA4EULL, "qun110_petrice6150078.cl"},
	{0x3F3B4F1D0DFF783FULL, "dae000_mother6035951.cl"},
	{0x3F3B9ABDEA1FC6D5ULL, "gen00fl_aveline6139140.cl"},
	{0x3F3C2184F388F7BEULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_pc_arishokresponse.gad"},
	{0x3F3C57647D330FEBULL, "vxdr_primevaltomb_3dl_0_0.vlm"},
	{0x3F3CA7D503060E33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_reaction.gad"},
	{0x3F3CDABB2BE630B3ULL, "llac_lt_0_lowtown_65534_1.dds"},
	{0x3F3CE4E47A4DEC1BULL, "llac_ht_keep_3736_1.dds"},
	{0x3F3D165A46B61B9FULL, "gen00fl_varric6118753.cl"},
	{0x3F3DC111427B3CCBULL, "772661181.ogg"},
	{0x3F3EB5992769BC6CULL, "hf_vis_hightown1.mor"},
	{0x3F3ED34F108BAF03ULL, "vxht_0_hightown_c1f_0_0.vlm"},
	{0x3F3FA90A69AC1DEBULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.dura.ani"},
	{0x3F3FB07B036BFAFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.ani"},
	{0x3F4080340859C55BULL, "mer120_hubert6017838.cl"},
	{0x3F40A9306ECCF484ULL, "vxht_chantry_n_20p_0_0.vlm"},
	{0x3F40E6F613E9D6B9ULL, "6207770_m.fxe"},
	{0x3F415A9747ADADE7ULL, "581572711.ogg"},
	{0x3F41B886E3E17142ULL, "6110797_f.fxe"},
	{0x3F42128317E7215CULL, "l_one121ar_warehouse.bnk"},
	{0x3F424285F2DE8B2CULL, "mag110_vincento6016294.cl"},
	{0x3F42AC6A84E6C77BULL, "ca_caverns1_mmap_ph.dds"},
	{0x3F4300E2C7F77F1FULL, "6041439_m.fxe"},
	{0x3F432481809A7646ULL, "vxlt_darktown_5in_0_0.vlm"},
	{0x3F43A4D5030C2798ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_reaction.evt"},
	{0x3F43E43419D05BFDULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.dura.ani"},
	{0x3F44677FF3C77423ULL, "6083521_m.fxe"},
	{0x3F447D7984779FBDULL, "vxca_cavern1_haunted_ims_0_0.vlm"},
	{0x3F44BF389A85B1ABULL, "6216712_m.fxe"},
	{0x3F44DC42DBE9EB68ULL, "6060554_m.fxe"},
	{0x3F44F3F628AB6B70ULL, "hlac_w_encounter2_camp_327683_0y.dds"},
	{0x3F455734B30257CAULL, "6174824_m.fxe"},
	{0x3F457EE64D80B821ULL, "vxca_caverns1_mines_j23_0_2.vlm"},
	{0x3F460AB2E110E080ULL, "vxht_0_hightown_a2_bgf_0_0.vlm"},
	{0x3F460DA1194133CFULL, "vxca_caverns2prc_8kq_0_0.vlm"},
	{0x3F46323DACFB28E5ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_3x.dds"},
	{0x3F463C932F5E3023ULL, "vxlt_gallowstemplar_6q8_0_0.vlm"},
	{0x3F4664B004963BDCULL, "mag370_meredith6064974.cl"},
	{0x3F46A389A6C6E385ULL, "vxw_blightlands_lb_0_0.vlm"},
	{0x3F46EB587980E802ULL, "vxht_0_hightown_n_c58_0_0.vlm"},
	{0x3F46FC990196A21BULL, "vxdr_primevaltomb_6yn_0_0.vlm"},
	{0x3F47211CEBDDD8D0ULL, "vxlt_backalley_n_1pl_0_0.vlm"},
	{0x3F479CFE45F9B90BULL, "vxw_sundermount_31k_0_1.vlm"},
	{0x3F47A92B328EB1B0ULL, "6125910_m.fxe"},
	{0x3F4825C8E1DFB2FCULL, "6172266_m.fxe"},
	{0x3F489016D9E39F3AULL, "vxlt_warehouse_xm_0_0.vlm"},
	{0x3F48954D645E46D3ULL, "6116007_m.fxe"},
	{0x3F49170A1FEF5126ULL, "hlac_w_encounter1_65534_0z.dds"},
	{0x3F495C835BFE54BCULL, "dae000_oldmangossip6174906.cl"},
	{0x3F49748C35B308EBULL, "ht_0_hightown_61_0.rml"},
	{0x3F497EAFD682B36DULL, "916428753.ogg"},
	{0x3F49BF2EBC0B8A87ULL, "870968256.ogg"},
	{0x3F4A2C6CE617F6B0ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_infa_0n.dds"},
	{0x3F4A68737EE9A823ULL, "qun110cr_throne_guard.utc"},
	{0x3F4AD8DA4929EC3AULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.dura.gad"},
	{0x3F4B8CF4D9757943ULL, "mag311_emile_de_launcet6043635.cl"},
	{0x3F4BC8D62FCAD3A8ULL, "qun221_petrice.cnv"},
	{0x3F4BE3A03A60CF94ULL, "map_night_1280x1024.dds"},
	{0x3F4C1FAC394C7B68ULL, "vxw_encounter2_d8_0_0.vlm"},
	{0x3F4C28139C89253AULL, "6121509_m.fxe"},
	{0x3F4C910A69B67F62ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.dura.evt"},
	{0x3F4CB8369619EBE8ULL, "mag160_feynriel6020098.cl"},
	{0x3F4CE2DFC7D13AFDULL, "container.gfx"},
	{0x3F4CEB4BE1D7CDFFULL, "vxdr_primevaltomb_8ls_0_0.vlm"},
	{0x3F4D32D25DC9AA8CULL, "hlac_lt_0_lowtown_-131071_0y.dds"},
	{0x3F4D3C960245D26EULL, "6099488_m.fxe"},
	{0x3F4D6043DDA4435BULL, "character_character_p138.psh.fxo.dx11"},
	{0x3F4DAE731579CB15ULL, "6122774_m.fxe"},
	{0x3F4DC782983A5DC2ULL, "743128528.ogg"},
	{0x3F4EA8CE5C723106ULL, "hlac_ca_caverns1_mines_24422_4z.dds"},
	{0x3F4EB9E5D2F03C20ULL, "243918090.ogg"},
	{0x3F4EEAD62FCD4310ULL, "qun221_petrice.bnk"},
	{0x3F4EED625801321FULL, "vxht_mansion2b_1mx_0_0.vlm"},
	{0x3F4EF9D612125062ULL, "vxht_keep_73u_0_0.vlm"},
	{0x3F4F195DEB4D8C4CULL, "6104920_m.fxe"},
	{0x3F4F8935A1E5B9B2ULL, "vxht_chantry_2lm_0_0.vlm"},
	{0x3F4FBDB0A8352E2CULL, "hf_apostate1.mor"},
	{0x3F4FF5C105EE235FULL, "llac_lt_0_lowtown_docks_a3_65533_2.dds"},
	{0x3F503A18A0283FD4ULL, "vxht_brothel_n_62a_0_0.vlm"},
	{0x3F50727B583A423FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp3a.phy"},
	{0x3F50A2DF71DFEF5DULL, "6131125_m.fxe"},
	{0x3F50DE225DEB6B12ULL, "zz_vault_debug6200379.cl"},
	{0x3F50F130A331741EULL, "104451878.ogg"},
	{0x3F51208895540665ULL, "hlac_lt_0_lowtown_a2_n_c_131071_1y.dds"},
	{0x3F514441FA07D150ULL, "525048756.ogg"},
	{0x3F51965E72908EC6ULL, "vxlt_darktown_5ro_0_0.vlm"},
	{0x3F51DBDA49301051ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.dura.evt"},
	{0x3F52009236A5FC07ULL, "hla_ht_0_hightown_n_131073_1x.dds"},
	{0x3F5237BD2FAA734EULL, "gen00fl_anders6125341.cl"},
	{0x3F5271E76B8C93D0ULL, "6050117_m.fxe"},
	{0x3F52A3EE528933B0ULL, "one120_athenril6088911.cl"},
	{0x3F52B8E9A9F7EC54ULL, "plt_cod_ltr_otl_comte.nss"},
	{0x3F52EF56766DB83DULL, "vxca_cavern1_haunted_ijr_0_0.vlm"},
	{0x3F534BDD20ADE5E9ULL, "vxca_caverns2_98b_0_0.vlm"},
	{0x3F53538F703E6AF9ULL, "276791111.ogg"},
	{0x3F53567573C21542ULL, "130686125.ogg"},
	{0x3F53900A69BC9C75ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.dura.gad"},
	{0x3F54A1692EBD211EULL, "50150180.ogg"},
	{0x3F54C775CEF2BBA6ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_0x.dds"},
	{0x3F54DD60589CB202ULL, "6107004_m.fxe"},
	{0x3F54ECA003ED8E36ULL, "dae011_corff_sta.cnv"},
	{0x3F54F6B08AEECE76ULL, "dae000ip_crowd_lm2_talk4.utp"},
	{0x3F55156698BCFC05ULL, "vxw_blightlands_qa_0_0.vlm"},
	{0x3F555D356B770082ULL, "vxht_0_hightown_n_c29_0_0.vlm"},
	{0x3F55761D99283B2DULL, "6148961_m.fxe"},
	{0x3F55DA427DDC3D30ULL, "distor_mask.dds"},
	{0x3F5637BEE022BD2FULL, "sky_motionvectorssky_p0.vsh.fxo"},
	{0x3F56DE9CCD873636ULL, "vxlt_hangedman_n_1hs_0_0.vlm"},
	{0x3F56E82B41B000F1ULL, "zz_fol_debug6103807.cl"},
	{0x3F57895464466CDDULL, "6079027_m.fxe"},
	{0x3F5794581FAB09D7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunkhigh_s.dds"},
	{0x3F57D9F267489856ULL, "vxht_mansion2_3gu_0_0.vlm"},
	{0x3F57DF2447BFDACCULL, "165489711.ogg"},
	{0x3F57EEA003EFC716ULL, "dae011_corff_sta.bnk"},
	{0x3F5812F48079DE9EULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rtlb.ani"},
	{0x3F5839FA2D568BDCULL, "mer210_coterie_contact6079365.cl"},
	{0x3F5877D199A2ECE5ULL, "w_encounter2_13_0_ll.mmh"},
	{0x3F589CAC878AA40BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_ex.ani"},
	{0x3F59EE0BFA31B96FULL, "vxw_sundermount_275_0_0.vlm"},
	{0x3F5A07099D6A8EDCULL, "291048615.ogg"},
	{0x3F5A8BE3436E8826ULL, "vxht_mansion2b_1b4_0_0.vlm"},
	{0x3F5A91892B4293E8ULL, "vxw_encounter2_o9_0_0.vlm"},
	{0x3F5B8FACCB2448F0ULL, "6071871_m.fxe"},
	{0x3F5BE653C6424ACAULL, "prm000im_blood_gloves.uti"},
	{0x3F5BE9752E4689B6ULL, "vxht_mansion_n_3va_0_0.vlm"},
	{0x3F5C9C87679AC186ULL, "21280113.ogg"},
	{0x3F5C9F0DE5B451BCULL, "6027685_m.fxe"},
	{0x3F5D33C0430D8A3BULL, "6095908_m.fxe"},
	{0x3F5D3CD43CA7CC7FULL, "follower_banter6163729.cl"},
	{0x3F5D8D6AB014B4F1ULL, "vxw_encounter3_3c_0_0.vlm"},
	{0x3F5DD6009016E8C7ULL, "gen00fl_merrill6106978.cl"},
	{0x3F5DFB1293DBD232ULL, "vxht_chantry_2in_0_0.vlm"},
	{0x3F5E07A99327FD87ULL, "vxw_bonepit_7b_0_0.vlm"},
	{0x3F5E22E7D7E93BADULL, "6187951_m.fxe"},
	{0x3F5E88545CBA4E75ULL, "vxdr_primevalentrance_1ac_0_0.vlm"},
	{0x3F5ECA68BFC9A111ULL, "hlac_w_blightlands_-1_0z.dds"},
	{0x3F5EF3DA493AC320ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.dura.ani"},
	{0x3F5FE3EA6257A2F7ULL, "489149828.ogg"},
	{0x3F6132D1269B34B2ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_basket_sa.dds"},
	{0x3F6161336863D0BDULL, "vxca_cavern1_haunted_igq_0_0.vlm"},
	{0x3F61737B5848B744ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp3a.mmh"},
	{0x3F61777B5848BD97ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp3a.mao"},
	{0x3F61897B5848DCE6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp3a.msh"},
	{0x3F6259A06CB2433BULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.cess.ani"},
	{0x3F62D1486DF5BF27ULL, "vxca_caverns1_mines_gkm_0_0.vlm"},
	{0x3F62D9C254012970ULL, "998149.ogg"},
	{0x3F63C042EB395765ULL, "299619368.ogg"},
	{0x3F64256A470E8BC2ULL, "36641895.ogg"},
	{0x3F659F2EB4209953ULL, "6085040_m.fxe"},
	{0x3F65A19B9BC9BFA4ULL, "ave200_party_comment6116359.cl"},
	{0x3F65B2B18D9BE9E6ULL, "vxlt_gallowscourtyard_62j_0_0.vlm"},
	{0x3F65F79DFC505177ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_exp_p.impa.gad"},
	{0x3F661E431A633B47ULL, "ts_dwarven_prog_junk1.gda"},
	{0x3F6636AC88C9E606ULL, "147017905.ogg"},
	{0x3F6673DD5A15378DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf05a.phy"},
	{0x3F6677C7A5050DD1ULL, "6076084_m.fxe"},
	{0x3F6684AC8796B882ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_ex.evt"},
	{0x3F669AA5258CDFD6ULL, "hf_eustice.mor"},
	{0x3F66C839B0D5E772ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\small_n.dds"},
	{0x3F67050049872405ULL, "6087215_m.fxe"},
	{0x3F67275878DE7832ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_bench"},
	{0x3F679DBA223FEBCDULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh02a.mmh"},
	{0x3F67AFBA22400A53ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh02a.msh"},
	{0x3F6825985DC3A47FULL, "672767071.ogg"},
	{0x3F685FE8EC27D1EFULL, "vxw_sundermount_284_0_0.vlm"},
	{0x3F6903661D38AC68ULL, "vxw_encounter2_j6_0_0.vlm"},
	{0x3F6954DDBF6173CCULL, "hm_templar02.mor"},
	{0x3F69582847F2F747ULL, "6155724_f.fxe"},
	{0x3F69A3F6B6E61758ULL, "llac_lt_0_lowtown_a2_65534_0.dds"},
	{0x3F69B8AE7FFAAB1AULL, "gen00fl_merrill6108218.cl"},
	{0x3F69EA942A105FA0ULL, "49959503.ogg"},
	{0x3F6A4D914BEA710FULL, "vxdr_primevaltomb_8y8_0_0.vlm"},
	{0x3F6A64189DBC379EULL, "hair_tints.gda"},
	{0x3F6B57D1657F67D3ULL, "428750034.ogg"},
	{0x3F6B5CE385DB5035ULL, "dre000_varric6034685.cl"},
	{0x3F6B6D0FB165EA27ULL, "gen_im_trs_qun_two_2hm_r2.uti"},
	{0x3F6B84BF0B3D33A9ULL, "gen00fl_anders6111286.cl"},
	{0x3F6BC131D83184D2ULL, "vxca_cave_199_0_0.vlm"},
	{0x3F6BE304A7B0ABD3ULL, "vxht_mansion_n_6ju_0_0.vlm"},
	{0x3F6C2BF4808AB218ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rtlb.gad"},
	{0x3F6C31497BF317DCULL, "vxdr_primevalentrance_23s_0_0.vlm"},
	{0x3F6C4FEEE1212C49ULL, "darktown_male_fer_urch_sta6174953.cl"},
	{0x3F6CD911CF901D98ULL, "follower_banter6141867.cl"},
	{0x3F6CDB1C5D8293DCULL, "vxca_caverns2_5a8_0_0.vlm"},
	{0x3F6D0C9DFC569304ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_exp_p.impa.evt"},
	{0x3F6D2D3B9370CFB3ULL, "6094518_m.fxe"},
	{0x3F6D5DD4140EB5B3ULL, "default_white.dds"},
	{0x3F6D6AC7B7F4566BULL, "41545268.ogg"},
	{0x3F6D813FE4AFB1EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_over_left_shoulder.gad"},
	{0x3F6D83AC879CD595ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_ex.gad"},
	{0x3F6DB2A00E92761CULL, "177007503.ogg"},
	{0x3F6DFF499212A633ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi02.phy"},
	{0x3F6E08A201FFF831ULL, "vxlt_warehouse_n_11g_0_0.vlm"},
	{0x3F6E15CFBB76720CULL, "6214710_f.fxe"},
	{0x3F6E8F16B279CF31ULL, "537629203.ogg"},
	{0x3F6E93105537477BULL, "394366334.ogg"},
	{0x3F6ED570C8132985ULL, "571771405.ogg"},
	{0x3F6F06667C88DD30ULL, "art\\levels\\races\\natural\\proxy\\foliage\\bsh_zzbushmed_uhm.msh"},
	{0x3F6F14667C88F54AULL, "art\\levels\\races\\natural\\proxy\\foliage\\bsh_zzbushmed_uhm.mmh"},
	{0x3F6F18667C88FB81ULL, "art\\levels\\races\\natural\\proxy\\foliage\\bsh_zzbushmed_uhm.mao"},
	{0x3F6F406FE3EB08A5ULL, "6115379_f.fxe"},
	{0x3F6F58BEFD4F0A54ULL, "vxht_0_hightown_a2_av6_0_0.vlm"},
	{0x3F6F61A06CBCDB32ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.cess.evt"},
	{0x3F6F6C364746BE19ULL, "6051601_m.fxe"},
	{0x3F6FA9EF058BF9B9ULL, "gen00fl_varric6076064.cl"},
	{0x3F6FD88E5E4A048AULL, "hlac_lt_0_lowtown_n_-65535_7x.dds"},
	{0x3F7007BEA3F1B3DFULL, "vxlt_0_lowtown_a2_e1l_0_0.vlm"},
	{0x3F702AD2B3815AF1ULL, "mag370_meredith6046542.cl"},
	{0x3F7040AB587BBD64ULL, "6103716_m.fxe"},
	{0x3F706FBAB2E9A50CULL, "dae000_merchant_tailor_hm6164347.cl"},
	{0x3F70981B5856DD16ULL, "lgt310_mag_melson6152116.cl"},
	{0x3F70ADB212AF5A6EULL, "gen00fl_sebastian6195879.cl"},
	{0x3F70F351C3C4069AULL, "mag371_elsa6041418.cl"},
	{0x3F71AFF9C0924DBDULL, "ts_balanceinfo.ncs"},
	{0x3F720C2F04B84E3BULL, "vxlt_hangedman_n_w0_0_0.vlm"},
	{0x3F7232C0EB976AD0ULL, "llac_lt_gallowstemplar_n_2382_0.dds"},
	{0x3F7293C9EF34AE5BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railtop.phy"},
	{0x3F73C20C8A9BC778ULL, "77189318.ogg"},
	{0x3F73FB3FAB4BD664ULL, "vxlt_gallowstemplar_f_4ub_0_0.vlm"},
	{0x3F74643FE4B59E81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_over_left_shoulder.evt"},
	{0x3F746A40550C65A2ULL, "character_character_p227.psh.fxo.dx11"},
	{0x3F746F431A11C302ULL, "6098810_m.fxe"},
	{0x3F74AF71FE6BF2ABULL, "hlac_lt_gallowstemplar_f_2382_1z.dds"},
	{0x3F74B7C03DA2231EULL, "gen00fl_merrill6121817.cl"},
	{0x3F75031318476953ULL, "vxht_mansion2_24i_0_1.vlm"},
	{0x3F7563BA69054A52ULL, "vxht_mansion_n_6jp_0_0.vlm"},
	{0x3F758FA67D90143CULL, "vxlt_shop_j6_0_0.vlm"},
	{0x3F76063419A2F920ULL, "700657562.ogg"},
	{0x3F763B0F3A4037B2ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkleft.gad"},
	{0x3F7680A06CC32EC5ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.cess.gad"},
	{0x3F7692E37E698027ULL, "w_bonepitdestroyed_a3_46_0.rml"},
	{0x3F76B5AB1EEAF8D1ULL, "1004451169.ogg"},
	{0x3F76B5B3869C0BA5ULL, "6155295_m.fxe"},
	{0x3F76D3D1E5993DADULL, "6056921_m.fxe"},
	{0x3F7717A2C77B933DULL, "vxca_caverns2prc_9xg_0_0.vlm"},
	{0x3F775CBDCD18470BULL, "vxw_encounter2_camp_9d_0_0.vlm"},
	{0x3F776F9D275AB926ULL, "vxht_mansion2b_1x6_0_0.vlm"},
	{0x3F78117311140C2DULL, "6095141_f.fxe"},
	{0x3F7896BA224E528AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh02a.phy"},
	{0x3F789D2D0560874EULL, "vxlt_0_lowtown_d6y_0_0.vlm"},
	{0x3F78E0C592B79EA1ULL, "mag120_macha6091378.cl"},
	{0x3F790980CCE67184ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.dura.ani"},
	{0x3F797D36D7D19BE3ULL, "vxht_0_hightown_a3_byr_0_0.vlm"},
	{0x3F7980B28DC4DD28ULL, "910988070.ogg"},
	{0x3F79839F00201B5FULL, "gen00fl_bethany6109225.cl"},
	{0x3F79C16F61FF37EFULL, "art\\characters\\creatures\\shade\\animation\\blendtree\\shade_overlay.anb"},
	{0x3F79F75C305320CDULL, "vxca_caverns1_icr_0_2.vlm"},
	{0x3F7A249DFC6146FDULL, "art\\vfx\\cinematic\\vfx_meredith_swd_exp_p.impa.ani"},
	{0x3F7A54E199A6C453ULL, "vxht_mansion_n_6mt_0_0.vlm"},
	{0x3F7AA3266DE9305CULL, "vxdr_primevalentrance_26p_0_0.vlm"},
	{0x3F7AFF957E9087CAULL, "pro000_merchant6051581.cl"},
	{0x3F7B2E90D3B24FE9ULL, "hlac_ht_0_hightown_n_65538_0z.dds"},
	{0x3F7B621579D8DA13ULL, "6209411_m.fxe"},
	{0x3F7BFE17076147FFULL, "vxlt_0_lowtown_docks_b5n_0_0.vlm"},
	{0x3F7C1F7A73E259DEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile_12.pwk"},
	{0x3F7C5AF24365BC53ULL, "t3_gsp_b01.tnt"},
	{0x3F7CB7600104FB87ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_stairs_d.dds"},
	{0x3F7D15118D76FAF1ULL, "vxht_mansion_n_634_0_0.vlm"},
	{0x3F7D34CB4812972EULL, "zz_qun_debug6085815.cl"},
	{0x3F7D3E0F3A465BE9ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkleft.evt"},
	{0x3F7DCA9BEF4522D4ULL, "vxht_0_hightown_a2_aq7_0_0.vlm"},
	{0x3F7DF40E851C571EULL, "498525542.ogg"},
	{0x3F7E43329645A16BULL, "vxlt_0_lowtown_a2_h1y_0_0.vlm"},
	{0x3F7E54BB4AC2F154ULL, "vxw_encounter2_ky_0_0.vlm"},
	{0x3F7E9996883821A8ULL, "vxdr_0_deeproads_coy_0_1.vlm"},
	{0x3F7EB2884A71D5E4ULL, "6104616_m.fxe"},
	{0x3F7EEF208E24BD89ULL, "hlac_lt_backalley_n_65537_3y.dds"},
	{0x3F7F09C76DE3839DULL, "w_encounter1_1_0.mao"},
	{0x3F7F0DC76DE38AD6ULL, "w_encounter1_1_0.mmh"},
	{0x3F7F17C76DE39BD4ULL, "w_encounter1_1_0.msh"},
	{0x3F7F3989077E9E53ULL, "1009396355.ogg"},
	{0x3F800680CCEC8BE6ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.dura.gad"},
	{0x3F800998C4E556E9ULL, "6153135_f.fxe"},
	{0x3F80189421AE2FDFULL, "vxw_encounter3_ek_0_0.vlm"},
	{0x3F8089081AB0AE19ULL, "vxlt_gallowscourtyard_75i_0_0.vlm"},
	{0x3F8093E50CDD2CAAULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_spear1_high_arishokresponse.gad"},
	{0x3F8097CA411D7F8CULL, "828245199.ogg"},
	{0x3F80E351A7A5E111ULL, "follower_barks6153536.cl"},
	{0x3F811A0B4FD65798ULL, "hlac_lt_0_lowtown_docks_65535_3y.dds"},
	{0x3F8162E44B657EDEULL, "vxlt_0_lowtown_docks_a3_n_btp_0_0.vlm"},
	{0x3F81631EE43228B3ULL, "vxdr_primevaltomb_8on_0_0.vlm"},
	{0x3F81DE32EA4999BAULL, "mcr370_negotiation6080138.cl"},
	{0x3F820DE7C16AF550ULL, "vxdr_primevalentrance_2z9_0_1.vlm"},
	{0x3F825C3FE4C1CED0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_over_left_shoulder.ani"},
	{0x3F82809842F9014DULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_0t.dds"},
	{0x3F82852FB3A2B299ULL, "zz_seb_debug6214544.cl"},
	{0x3F829DB2F31C59AEULL, "6103435_m.fxe"},
	{0x3F82BA61CB8DA51BULL, "6043625_f.fxe"},
	{0x3F836ABB85EAC469ULL, "hm_dead_1.mor"},
	{0x3F83A95BE0AF5B77ULL, "6128760_f.fxe"},
	{0x3F83D5975AFB62D2ULL, "vxht_mansion_n_6mq_0_0.vlm"},
	{0x3F848286758C7F7CULL, "vxlt_gallowscourtyard_n_7e7_0_0.vlm"},
	{0x3F85071DC8D23E6EULL, "w_bonepit_158_0.mmh"},
	{0x3F85131DC8D25205ULL, "w_bonepit_158_0.mao"},
	{0x3F85211DC8D26A5CULL, "w_bonepit_158_0.msh"},
	{0x3F85E7200124DD68ULL, "vxw_encounter2_p4_0_0.vlm"},
	{0x3F860180CCF0EEC5ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.dura.evt"},
	{0x3F86346D393439F0ULL, "hlac_ht_0_hightown_a3_65536_4y.dds"},
	{0x3F866FD38D1EEF97ULL, "dae000_bartrand6206974.cl"},
	{0x3F8688032D5F406FULL, "481404024.ogg"},
	{0x3F86E8463C711F89ULL, "vxlt_0_lowtown_n_eww_0_0.vlm"},
	{0x3F86F7F911328A8FULL, "vxht_chantry_n_35a_0_0.vlm"},
	{0x3F86FC2BADFBB122ULL, "917437456.ogg"},
	{0x3F86FFCB27A27C19ULL, "vxht_0_hightown_a3_bx3_0_0.vlm"},
	{0x3F875174E8CA4630ULL, "vxw_encounter2_camp_pd_0_0.vlm"},
	{0x3F87B533C8ADCDCDULL, "6044166_m.fxe"},
	{0x3F87E9364B331AF8ULL, "vxlt_0_lowtown_a2_n_ch9_0_0.vlm"},
	{0x3F8891D9625A30D9ULL, "gen_im_trs_bks_champ_kirk.uti"},
	{0x3F88BBD6877C4CA0ULL, "lt_gallowscourtyard_n_probe_5fy.dds"},
	{0x3F892DCCB4958B52ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_008a_n.dds"},
	{0x3F89346D8F6ABC2EULL, "art\\vfx\\combat\\vfx_ha_hd_overwhelm_act_c.mmh"},
	{0x3F896D8AE1D595E8ULL, "338942708.ogg"},
	{0x3F897CEAEC403D70ULL, "vxlt_0_lowtown_a2_n_c_exi_0_0.vlm"},
	{0x3F89A4DD905E7DF1ULL, "vxht_keep_n_6yf_0_0.vlm"},
	{0x3F89F509262A55E1ULL, "6050394_m.fxe"},
	{0x3F8A21BC4EBB29A4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed2.ani"},
	{0x3F8A260F3A50BD58ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkleft.ani"},
	{0x3F8A8C8DA687CDA2ULL, "6179984_m.fxe"},
	{0x3F8ABEDBA4B609F0ULL, "ico_longbow_3.dds"},
	{0x3F8AEB8BD1AC326FULL, "208373055.ogg"},
	{0x3F8B95D8C062F4AFULL, "vxht_brothel_n_50c_0_0.vlm"},
	{0x3F8BC166E6706101ULL, "vxca_caverns1_mines_edp_0_0.vlm"},
	{0x3F8CB0AF38C1AA17ULL, "vxht_keep_63m_0_0.vlm"},
	{0x3F8CB0CA1050DEB3ULL, "cnvprev_human_male.utc"},
	{0x3F8D03F3827F4ED2ULL, "vxht_0_hightown_c70_0_1.vlm"},
	{0x3F8D2EAF22258881ULL, "vxlt_hangedman_2vz_0_0.vlm"},
	{0x3F8D9652687CF088ULL, "vxlt_gallowstemplar_f_3m9_0_0.vlm"},
	{0x3F8DC7A3ED9E6C1EULL, "gen00fl_fenris6110916.cl"},
	{0x3F8E1B4B2E752574ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_cutoff_d.dds"},
	{0x3F8E92D9F0990647ULL, "936315248.ogg"},
	{0x3F8E92E37E7E1244ULL, "w_bonepitdestroyed_a3_46_0.mmh"},
	{0x3F8E96E37E7E1897ULL, "w_bonepitdestroyed_a3_46_0.mao"},
	{0x3F8EA8E37E7E37E6ULL, "w_bonepitdestroyed_a3_46_0.msh"},
	{0x3F8EC577185988FBULL, "hlac_ca_caverns1_24427_3z.dds"},
	{0x3F8F6AC24EE4963FULL, "art\\vfx\\combat\\vfx_gib_qunc_rblt.anb"},
	{0x3F90043D1AE0B516ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutb.mao"},
	{0x3F90103D1AE0C9B5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutb.mmh"},
	{0x3F90123D1AE0CD2BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutb.msh"},
	{0x3F901EBC4EBF9006ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed2.gad"},
	{0x3F903BF99A277422ULL, "lt_0_lowtown_docks_147_0.msh"},
	{0x3F903DF99A2777D8ULL, "lt_0_lowtown_docks_147_0.mmh"},
	{0x3F9049F99A278C7BULL, "lt_0_lowtown_docks_147_0.mao"},
	{0x3F90A265229BAF1FULL, "403497570.ogg"},
	{0x3F91176FC64570EFULL, "zz_vault_debug6203333.cl"},
	{0x3F9169D34743431EULL, "ave200_aveline_bar6116199.cl"},
	{0x3F9182B718E9D7EEULL, "seb201_lady_harimann6177572.cl"},
	{0x3F91B22F9E06BC1BULL, "894071196.ogg"},
	{0x3F92F463678297FCULL, "vxlt_gallowscourtyard_n_7h4_0_0.vlm"},
	{0x3F936F490A981620ULL, "vxw_encounter3_9l_0_0.vlm"},
	{0x3F939A1E294A07CAULL, "vxlt_0_lowtown_docks_bu7_0_0.vlm"},
	{0x3F941C7B92513F73ULL, "6105487_m.fxe"},
	{0x3F944A0F41E2516DULL, "6197859_m.fxe"},
	{0x3F9453570B46EA26ULL, "vxht_mansion2b_1v8_0_0.vlm"},
	{0x3F9454FAD618FCA4ULL, "742912728.ogg"},
	{0x3F946EE50CED9690ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_spear1_high_arishokresponse.ani"},
	{0x3F950DBBA0793B8DULL, "166840347.ogg"},
	{0x3F9562AF0FD83B6EULL, "vxlt_0_lowtown_ffk_0_0.vlm"},
	{0x3F956E72A9784381ULL, "vxht_0_hightown_a2_bsw_0_0.vlm"},
	{0x3F959C6B00E4B24CULL, "vxlt_undercitydungeon_179_0_0.vlm"},
	{0x3F95C351DAC05EB0ULL, "vxw_encounter2_camp_ke_0_0.vlm"},
	{0x3F963319CFE09D2AULL, "217317557.ogg"},
	{0x3F9660F0BBBDCCE3ULL, "vxht_0_hightown_a3_bst_0_0.vlm"},
	{0x3F96E579AD750D30ULL, "lgt110_urn_barker6132984.cl"},
	{0x3F9719BC4EC5A6E5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed2.evt"},
	{0x3F9726B311CFAC9FULL, "vxht_mansion2b_1tg_0_0.vlm"},
	{0x3F97536D8F772E2DULL, "art\\vfx\\combat\\vfx_ha_hd_overwhelm_act_c.anb"},
	{0x3F976FE81F0C3FADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9mw_pulleya_2.mao"},
	{0x3F981BEBDE56E4E6ULL, "vxca_caverns1_i1y_0_0.vlm"},
	{0x3F98A7C81B06B42FULL, "365763922.ogg"},
	{0x3F98AF4E22B62C13ULL, "one110_gustav.bnk"},
	{0x3F98BEB7AED35C94ULL, "vxca_caverns1_gkn_0_1.vlm"},
	{0x3F98D7239C48DD50ULL, "199355620.ogg"},
	{0x3F99E2B76097AF6EULL, "vxw_encounter1_e8_0_0.vlm"},
	{0x3F9A33A787437BABULL, "hlac_lt_backalley_n_1_6y.dds"},
	{0x3F9A5379DEE73581ULL, "llac_ht_mansion_n_3369_0.dds"},
	{0x3F9ABB2F28162F3CULL, "w_bonepitmining_a2_180_0.mao"},
	{0x3F9AC72F281643DFULL, "w_bonepitmining_a2_180_0.mmh"},
	{0x3F9ACD2F28164D81ULL, "w_bonepitmining_a2_180_0.msh"},
	{0x3F9ACE05B9209631ULL, "372090888.ogg"},
	{0x3F9BAB1525144328ULL, "vxht_keep_n_5gn_0_0.vlm"},
	{0x3F9BD14E22B89C63ULL, "one110_gustav.cnv"},
	{0x3F9BDBFAB3D28E49ULL, "dre000_hireling16038621.cl"},
	{0x3F9C28D4ADB51EAEULL, "6173403_m.fxe"},
	{0x3F9C3AF7267DE128ULL, "6095231_m.fxe"},
	{0x3F9C8C918B11A035ULL, "890510318.ogg"},
	{0x3F9D15D5610E20FAULL, "llac_dr_dungeon_3217_1.dds"},
	{0x3F9D1FBCB42936FFULL, "w_bonepit_9_0_lc.mmh"},
	{0x3F9D911A6C6F55FCULL, "95086712.ogg"},
	{0x3F9E5B58CCBF8B7FULL, "hlac_ca_caverns2prc_5802_35y.dds"},
	{0x3F9E68079B450BCAULL, "vxht_keep_5kk_0_0.vlm"},
	{0x3F9E725C95BFEAECULL, "vxlt_0_lowtown_a2_n_c_bz0_0_0.vlm"},
	{0x3F9E7AC9EF59BEFAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railtop.msh"},
	{0x3F9E88C9EF59D633ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railtop.mao"},
	{0x3F9E8CC9EF59DD40ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railtop.mmh"},
	{0x3FA114A2F32034DBULL, "609692562.ogg"},
	{0x3FA190056E983B4CULL, "hlac_lt_0_lowtown_a2_n_0_2z.dds"},
	{0x3FA197BEB55A2428ULL, "vxdr_slavers_1an_0_0.vlm"},
	{0x3FA214C6D90D3028ULL, "vxdr_primevalentrance_1j7_0_0.vlm"},
	{0x3FA2AC8AF12E37E7ULL, "6134757_m.fxe"},
	{0x3FA2CBFF010B195FULL, "ser191_emeric6120627.cl"},
	{0x3FA34AAA9BC67357ULL, "hlac_ht_mansion2b_818_3x.dds"},
	{0x3FA354F77F4979A8ULL, "vxlt_0_lowtown_docks_c5h_0_0.vlm"},
	{0x3FA372B5761A5D44ULL, "vxlt_shop_ce_0_0.vlm"},
	{0x3FA3CC00205D5089ULL, "vxlt_0_lowtown_n_emu_0_0.vlm"},
	{0x3FA4352ECCB67730ULL, "vxw_encounter2_camp_fb_0_0.vlm"},
	{0x3FA45253D8CF3E6CULL, "10772210.ogg"},
	{0x3FA4D2CDADB3E563ULL, "vxht_0_hightown_a3_bnw_0_0.vlm"},
	{0x3FA660A4D02C6E70ULL, "vxlt_0_lowtown_a2_n_c_erw_0_0.vlm"},
	{0x3FA6D3C61952F7ABULL, "art\\characters\\creatures\\ragedemon\\textures\\c_ragedmon_0n.dds"},
	{0x3FA6FD5AF5C93386ULL, "125006316.ogg"},
	{0x3FA7096A7B519494ULL, "157879711.ogg"},
	{0x3FA7C7A0FF9256AFULL, "vxht_0_hightown_a2_c5j_0_0.vlm"},
	{0x3FA7D65009F2F462ULL, "vxw_encounter1_ax_0_0.vlm"},
	{0x3FA80EA6D8A7ED96ULL, "mag100_cullen6140924.cl"},
	{0x3FA85494528DC7EEULL, "vxw_encounter1_j9_0_0.vlm"},
	{0x3FA9212F5673055EULL, "hlac_lt_0_lowtown_a2_n_131070_4x.dds"},
	{0x3FA9237016DD4B3BULL, "vxca_caverns1_mines_e3w_0_0.vlm"},
	{0x3FA998E90F005FEFULL, "mcr370_negotiation6095522.cl"},
	{0x3FAA12690611B981ULL, "vxlt_hangedman_2hx_0_0.vlm"},
	{0x3FAA1CF2170A5BA8ULL, "vxht_keep_n_5ho_0_0.vlm"},
	{0x3FAA63B8E188B8D0ULL, "hlac_w_sundermount_393215_0x.dds"},
	{0x3FAAC669303B5C14ULL, "6169598_m.fxe"},
	{0x3FAB423CD26C0290ULL, "6085612_m.fxe"},
	{0x3FAB7A3FCD0B57EEULL, "43001668.ogg"},
	{0x3FAB7C6AF010798BULL, "vxca_caverns1_j3y_0_0.vlm"},
	{0x3FAB85870C4CCE09ULL, "207150420.ogg"},
	{0x3FABC4591A75387AULL, "llac_lt_0_lowtown_a3_n_c_-131071_1.dds"},
	{0x3FABDA24F7F48448ULL, "art\\vfx\\models\\vfx_cylinder_2sd01.mmh"},
	{0x3FABF824F7F4B772ULL, "art\\vfx\\models\\vfx_cylinder_2sd01.msh"},
	{0x3FAC45758C1DC267ULL, "llac_w_sundermount_131076_0.dds"},
	{0x3FACCE9B9BADF978ULL, "6042565_m.fxe"},
	{0x3FACD08C53947E66ULL, "dm_arintal.mor"},
	{0x3FACE3156D6FA771ULL, "vxht_0_hightown_n_b8i_0_0.vlm"},
	{0x3FAD0770D527525FULL, "101378917.ogg"},
	{0x3FAD5029476618A4ULL, "6064671_m.fxe"},
	{0x3FAD66E80FA8A2BDULL, "6217928_m.fxe"},
	{0x3FAD75B9D4505ABEULL, "6092122_m.fxe"},
	{0x3FAE341488391313ULL, "vxca_cave_27q_0_0.vlm"},
	{0x3FAE3D036AA0C3F8ULL, "6206930_m.fxe"},
	{0x3FAE4932AFDB6B31ULL, "vxht_0_hightown_a2_9pa_0_0.vlm"},
	{0x3FAE51D32B13F6F8ULL, "195929373.ogg"},
	{0x3FAEA823D53E9B0DULL, "hlac_w_encounter1_-1_0z.dds"},
	{0x3FAEF2EA0DFF9A7FULL, "479624388.ogg"},
	{0x3FB03DBFD35A970EULL, "art\\levels\\races\\proxy\\props\\door\\_textures\\prp_blackdoor_d.dds"},
	{0x3FB06632CC5D2498ULL, "ico_mm_allyarrow.dds"},
	{0x3FB1009E21DC1057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_right.anb"},
	{0x3FB1D12FF19B4A90ULL, "mcr310_confused_guard6077019.cl"},
	{0x3FB1D223654BA2D0ULL, "417343923.ogg"},
	{0x3FB2522C8D647481ULL, "vxht_0_hightown_a2_byu_0_0.vlm"},
	{0x3FB279F6991BB15AULL, "and111_lirene6084530.cl"},
	{0x3FB2851BD3865D50ULL, "vxlt_0_lowtown_docks_a3_bfk_0_0.vlm"},
	{0x3FB2D341C16DF47EULL, "6133739_m.fxe"},
	{0x3FB3076A4D809177ULL, "vxht_keep_42t_0_0.vlm"},
	{0x3FB30C9E45E66B12ULL, "mcr371_hostile_fenris6082237.cl"},
	{0x3FB37113CC661582ULL, "vxw_woundedcoast_zy_0_0.vlm"},
	{0x3FB3BB3AC21AF06AULL, "ser100_party_comments6108240.cl"},
	{0x3FB4DE8F24910E51ULL, "vxw_bonepitmining_a2_1nr_0_0.vlm"},
	{0x3FB56658BA9C0140ULL, "isa321_velasco6091852.cl"},
	{0x3FB5C5CBA8873212ULL, "6085971_m.fxe"},
	{0x3FB6482CFBE90CE2ULL, "vxw_encounter1_fy_0_0.vlm"},
	{0x3FB72CF450BB8466ULL, "gen00fl_carver6103680.cl"},
	{0x3FB7580997EB2DC3ULL, "mag321ar_store.ncs"},
	{0x3FB7A566AC4F8EB9ULL, "6072689_f.fxe"},
	{0x3FB8432DA4478709ULL, "m_dae120ar_hightown.bnk"},
	{0x3FB85C14C72BAE0BULL, "hlac_ht_keep_n_3736_0x.dds"},
	{0x3FB89E81675202BDULL, "qun230cr_guild_contact6068864.cl"},
	{0x3FB8A67DEA36E1A4ULL, "vxlt_0_lowtown_docks_n_cqg_0_0.vlm"},
	{0x3FB8EBE93E5F3A08ULL, "vxlt_gallowstemplar_f_3r4_0_0.vlm"},
	{0x3FB90B57C530C94FULL, "qcr000pt_main.ncs"},
	{0x3FB95007C7434D98ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_90_c.anb"},
	{0x3FBA5850475EE875ULL, "106853132.ogg"},
	{0x3FBAF164C8DBE939ULL, "6082102_m.fxe"},
	{0x3FBB64BBA39B3B6BULL, "vxlt_backalley_n_1pr_0_1.vlm"},
	{0x3FBB7550A1E18090ULL, "841223732.ogg"},
	{0x3FBBC206395C3124ULL, "6063771_m.fxe"},
	{0x3FBC15C46BE62646ULL, "vxca_caverns2_8f7_0_1.vlm"},
	{0x3FBC4C28B60E6466ULL, "mrl000_marethari6129191.cl"},
	{0x3FBC890AF6309126ULL, "ht_0_hightown_45_0_lc.mmh"},
	{0x3FBD25E754B3FC20ULL, "vxlt_0_lowtown_a2_eqo_0_0.vlm"},
	{0x3FBDA8727AC2A9D6ULL, "335569973.ogg"},
	{0x3FBDAE5982F55CE8ULL, "838793765.ogg"},
	{0x3FBDED6815BC0131ULL, "hlac_ca_caverns1_mines_24424_6y.dds"},
	{0x3FBE9597CB840C25ULL, "vxht_0_hightown_a2_bww_0_0.vlm"},
	{0x3FBEE10E79C0F6DCULL, "hlac_w_sundermount_262142_0x.dds"},
	{0x3FBEE3228161D0FCULL, "6103598_m.fxe"},
	{0x3FBF226FECD62CCCULL, "986661387.ogg"},
	{0x3FBF3B93D2ECBB17ULL, "vxlt_0_lowtown_docks_n_b6m_0_1.vlm"},
	{0x3FBF6E2D7CD91218ULL, "zz_qun_debug6028236.cl"},
	{0x3FBF8E630E422EEEULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.impa.ani"},
	{0x3FC016750277797BULL, "6053289_m.fxe"},
	{0x3FC068500ADA034DULL, "lt_0_lowtown_docks_a3_206_0.mmh"},
	{0x3FC06C500ADA099EULL, "lt_0_lowtown_docks_a3_206_0.mao"},
	{0x3FC07A500ADA21D3ULL, "lt_0_lowtown_docks_a3_206_0.msh"},
	{0x3FC0C4097F5A8D01ULL, "vxht_0_hightown_a2_bbt_0_0.vlm"},
	{0x3FC0D7244E39A6CDULL, "vxdr_primevalentrance_fm_0_0.vlm"},
	{0x3FC0DF107129DE92ULL, "6054847_m.fxe"},
	{0x3FC1BE8104C3FCF5ULL, "plt_lgt000pt_baz_delyrium.nss"},
	{0x3FC210F03FFD31E1ULL, "vxht_0_hightown_biy_0_0.vlm"},
	{0x3FC21A7826CCDE69ULL, "6071886_f.fxe"},
	{0x3FC22C6C4CB3119BULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crn_rdy_rf.evt"},
	{0x3FC22DB062A62F5FULL, "6164396_m.fxe"},
	{0x3FC24A32C70A0038ULL, "6128385_f.fxe"},
	{0x3FC2755F25F8F02BULL, "hlac_lt_0_lowtown_docks_n_-1_1y.dds"},
	{0x3FC2E6656353073CULL, "vxlt_gallowscourtyard_750_0_0.vlm"},
	{0x3FC340444BC7E928ULL, "6144570_f.fxe"},
	{0x3FC35A3A5E73BCBCULL, "274996454.ogg"},
	{0x3FC36C515836DFF1ULL, "vxht_keep_n_6ez_0_0.vlm"},
	{0x3FC3B7233A5F01CEULL, "ht_0_hightown_a3_52_0.rml"},
	{0x3FC3F0BFB7D65184ULL, "421985663.ogg"},
	{0x3FC3F61ADF252B91ULL, "dae300_cullen6135346.cl"},
	{0x3FC4869862112082ULL, "and120st_chantry.stg"},
	{0x3FC4A5DC724CE242ULL, "vxlt_gallowstemplar_3s6_0_0.vlm"},
	{0x3FC56D6C155A3F54ULL, "llac_ca_caverns1_24424_9.dds"},
	{0x3FC5736DF1B5AE78ULL, "756677486.ogg"},
	{0x3FC5BB70369F381FULL, "530178308.ogg"},
	{0x3FC5E61A25CC2807ULL, "90554923.ogg"},
	{0x3FC63BE741233991ULL, "dae000_hate_enemies6167504.cl"},
	{0x3FC681F74848F585ULL, "757346320.ogg"},
	{0x3FC692D479451F97ULL, "6107002_f.fxe"},
	{0x3FC694866EE1BD90ULL, "gst_wall_lean_right.gda"},
	{0x3FC6F1CEDDA001BDULL, "gen00fl_anders6103915.cl"},
	{0x3FC6FC733F6908ADULL, "hlac_w_encounter3_-393213_0z.dds"},
	{0x3FC7AD3BD77210E3ULL, "759646961.ogg"},
	{0x3FC7BBBD8FC652CEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.impa.evt"},
	{0x3FC7BDA8408FFBFBULL, "6076401_f.fxe"},
	{0x3FC7C2F68EFEA409ULL, "dae011st_tavern_ambients.stg"},
	{0x3FC82FA9EB9E75A2ULL, "hlac_lt_warehouse_0_1x.dds"},
	{0x3FC86F5F1C52D89DULL, "llac_lt_0_lowtown_n_-65535_2.dds"},
	{0x3FC8B4FF3768DCBDULL, "319730226.ogg"},
	{0x3FC9256C4CB924DCULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crn_rdy_rf.gad"},
	{0x3FC9A353526125A6ULL, "254439177.ogg"},
	{0x3FC9E66A01B7FCE1ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v5.ani"},
	{0x3FCA2EDB74CC52CAULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfl.gad"},
	{0x3FCA80901DCDC134ULL, "mag210_arianni6024106.cl"},
	{0x3FCA86B63B7EB0D9ULL, "lt_warehouse_n_mmi.gff"},
	{0x3FCAE21857880C3FULL, "fen110_fenris6067902.cl"},
	{0x3FCB80C1F5720470ULL, "140109085.ogg"},
	{0x3FCBA86F4479976CULL, "lt_0_lowtown_docks_a3_173_0.msh"},
	{0x3FCBAE1128C05CB3ULL, "151355663.ogg"},
	{0x3FCBAE6F4479A15EULL, "lt_0_lowtown_docks_a3_173_0.mmh"},
	{0x3FCBBA6F4479B5F5ULL, "lt_0_lowtown_docks_a3_173_0.mao"},
	{0x3FCBFEAEF0F9D401ULL, "872528461.ogg"},
	{0x3FCC86630E4CAB3FULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.impa.evt"},
	{0x3FCC9C2C95750D85ULL, "zz_lgt_2_debug6211855.cl"},
	{0x3FCC9C30B2A5B131ULL, "6034639_m.fxe"},
	{0x3FCD783878BA939DULL, "6079654_m.fxe"},
	{0x3FCEB2BD8FCC6279ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.impa.gad"},
	{0x3FCF246C4CBD8EDAULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crn_rdy_rf.ani"},
	{0x3FCF31440421374CULL, "hlac_lt_foundry_n_320_1z.dds"},
	{0x3FCF35E67150A581ULL, "vxht_0_hightown_a2_bgk_0_0.vlm"},
	{0x3FCF391BE170F699ULL, "vxht_0_hightown_a3_bc6_0_0.vlm"},
	{0x3FCF3A4278DF2A2AULL, "vxlt_darktown_5rc_0_0.vlm"},
	{0x3FCF89C1D100E9CDULL, "206203767.ogg"},
	{0x3FCFEAABB83122F2ULL, "hlac_w_blightlands_-65534_0y.dds"},
	{0x3FCFF81DE59B82E8ULL, "gen00fl_anders6111331.cl"},
	{0x3FD010CD25A292AAULL, "w_woundedcoast_74_0_lc.mmh"},
	{0x3FD0213FB1B3DBB7ULL, "6101201_m.fxe"},
	{0x3FD07B4CFFC0E708ULL, "vxlt_0_lowtown_docks_n_b5z_0_0.vlm"},
	{0x3FD082CD31F34A61ULL, "vxht_0_hightown_brx_0_0.vlm"},
	{0x3FD0F16A01BE2D83ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v5.gad"},
	{0x3FD0F849926717A8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi02.mmh"},
	{0x3FD111B3229159ECULL, "x_partypicker_probe_2q.dds"},
	{0x3FD111DB74D24061ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfl.evt"},
	{0x3FD1164992674AD2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi02.msh"},
	{0x3FD118DED35A8AD8ULL, "181767190.ogg"},
	{0x3FD1B4FF0C6C75E4ULL, "hlac_lt_0_lowtown_docks_n_-65537_3y.dds"},
	{0x3FD1E546363AC3E5ULL, "vxlt_hangedman_n_10l_0_0.vlm"},
	{0x3FD1FCF63ECA7378ULL, "hlac_ca_caverns1_mines_24426_12y.dds"},
	{0x3FD23ADD21CF1C54ULL, "vxht_brothel_n_6gn_0_0.vlm"},
	{0x3FD25EBD88BD5150ULL, "6212117_m.fxe"},
	{0x3FD2A6105AF67DDAULL, "6082120_f.fxe"},
	{0x3FD316B82DCF9B81ULL, "785552821.ogg"},
	{0x3FD367630E529528ULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.impa.gad"},
	{0x3FD3C90224152E30ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flowerpot.mao"},
	{0x3FD3CD0224153543ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flowerpot.mmh"},
	{0x3FD3DB0224154C9DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flowerpot.msh"},
	{0x3FD3E5BA77F11F44ULL, "vxlt_gallowscourtyard_76a_0_0.vlm"},
	{0x3FD40548E172BE6CULL, "mag310cr_controlled_3.utc"},
	{0x3FD4298C88E34BCAULL, "vxht_keep_n_4b6_0_0.vlm"},
	{0x3FD49FB9356E1B41ULL, "vxlt_gallowstemplar_6qb_0_0.vlm"},
	{0x3FD4C3BD8FD0EAC7ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.impa.ani"},
	{0x3FD67D5454ADD580ULL, "vxht_0_hightown_a2_b5l_0_0.vlm"},
	{0x3FD6F8FE204691FDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masa_0t.dds"},
	{0x3FD7866AE6721154ULL, "7531923.ogg"},
	{0x3FD7A4FD65ABF7DEULL, "vxlt_darktown_4u6_0_1.vlm"},
	{0x3FD7CE6A01C41168ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v5.evt"},
	{0x3FD7CF05623AEDBCULL, "vxht_0_hightown_a3_bwk_0_0.vlm"},
	{0x3FD868B3C9B10B71ULL, "85923648.ogg"},
	{0x3FD8A17FE28A2CB9ULL, "vxlt_0_lowtown_bxv_0_0.vlm"},
	{0x3FD8E5EF26921B93ULL, "4585335.ogg"},
	{0x3FD91C4E423EF501ULL, "64973866.ogg"},
	{0x3FD9466195356194ULL, "vxlt_0_lowtown_docks_a3_d0j_0_1.vlm"},
	{0x3FD9568570EAC003ULL, "1931632.ogg"},
	{0x3FD9ED4AF6A8D5B1ULL, "vxlt_0_lowtown_a3_n_c_er7_0_0.vlm"},
	{0x3FDA00E74E9F1F9EULL, "fen210_hadriana6110603.cl"},
	{0x3FDADCB6997405BEULL, "gr_03_flow.mao"},
	{0x3FDB458BFC6563A7ULL, "820149188.ogg"},
	{0x3FDB50D4AD4AA23DULL, "220131785.ogg"},
	{0x3FDBD50DB621A0F2ULL, "512753278.ogg"},
	{0x3FDBE74477F287C0ULL, "101769527.ogg"},
	{0x3FDBFA12D210BF02ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf05.msh"},
	{0x3FDBFC12D210C2B8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf05.mmh"},
	{0x3FDC1269A7E20B02ULL, "86847986.ogg"},
	{0x3FDC5EE62F92EBF5ULL, "450711016.ogg"},
	{0x3FDC764FB55770EDULL, "6095862_f.fxe"},
	{0x3FDCA6571B537618ULL, "dae000_arishok6067702.cl"},
	{0x3FDCBCA83CB69ECCULL, "mag320_bsc_banter6100762.cl"},
	{0x3FDCDFE6117D5A74ULL, "ch2_ds1y.psh.fxo"},
	{0x3FDCEF4AB945235EULL, "vxht_bartrand_n_5rt_0_0.vlm"},
	{0x3FDD426FA0DEDF34ULL, "166649358.ogg"},
	{0x3FDD50D75D330957ULL, "vxlt_0_lowtown_docks_a3_cxc_0_0.vlm"},
	{0x3FDD798A3543AF45ULL, "6043065_f.fxe"},
	{0x3FDD8F0082ADE825ULL, "llac_ht_chantry_n_2534_30.dds"},
	{0x3FDDABC3A8B36446ULL, "gen00fl_fenris6126694.cl"},
	{0x3FDE1CFB5D104105ULL, "6087026_m.fxe"},
	{0x3FDE5AD08FF50A34ULL, "plt_cod_cha_bethany2.plo"},
	{0x3FDEC6DB69CF7A01ULL, "68730652.ogg"},
	{0x3FDF09DB74DE6FB0ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkfl.ani"},
	{0x3FDF3EFB0315C2DFULL, "795907941.ogg"},
	{0x3FDF5C0F56AE37E1ULL, "llac_w_woundedcoast_327684_0.dds"},
	{0x3FDF9363BA5C2B0CULL, "6176121_m.fxe"},
	{0x3FDFB294F2284AD6ULL, "6175715_m.fxe"},
	{0x3FE0657EC2B1C584ULL, "vxca_cavern1_haunted_ehm_0_0.vlm"},
	{0x3FE08E30A7C025CEULL, "vxca_cavern1_haunted_j22_0_0.vlm"},
	{0x3FE09C37CAA56C9CULL, "505055327.ogg"},
	{0x3FE11172B6BF2CF0ULL, "llac_lt_warehouse_0_5.dds"},
	{0x3FE1313FBE2EB3C2ULL, "223744668.ogg"},
	{0x3FE13CB59D69BBECULL, "gen00fl_aveline6127430.cl"},
	{0x3FE16B4DFC736C3BULL, "vxlt_hangedman_1p9_0_0.vlm"},
	{0x3FE1C0352A0656FEULL, "6115780_m.fxe"},
	{0x3FE1C77D6CFCF4FBULL, "scaleform.mat"},
	{0x3FE240C7C6EDC249ULL, "417407109.ogg"},
	{0x3FE24F07C7665133ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_90_c.mmh"},
	{0x3FE285B0BB1ECF72ULL, "llac_lt_0_lowtown_docks_a3_n_131070_0.dds"},
	{0x3FE29B697AD9644AULL, "vxht_keep_n_4i7_0_0.vlm"},
	{0x3FE2BC9E9FAFCAE9ULL, "narrative_flemeth6002805.cl"},
	{0x3FE3275A0BBC4D1AULL, "vxca_caverns1_hih_0_0.vlm"},
	{0x3FE378601588EB05ULL, "433259317.ogg"},
	{0x3FE3BD8104E0E156ULL, "plt_lgt000pt_baz_delyrium.plo"},
	{0x3FE4582C1A6AD4BBULL, "vxlt_hangedman_21q_0_0.vlm"},
	{0x3FE49DCB2AC2E671ULL, "11185243.ogg"},
	{0x3FE4D9C94FD6633DULL, "6096626_m.fxe"},
	{0x3FE4F2BFB45D0F4EULL, "hlac_w_blightlands_131073_0x.dds"},
	{0x3FE53B4F936B8B00ULL, "6163453_m.fxe"},
	{0x3FE5EF9DB24D3656ULL, "vxlt_warehouse_sl_0_0.vlm"},
	{0x3FE60312D2192343ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf05.phy"},
	{0x3FE67E45FDE642EEULL, "hlac_ht_mansion2_818_14y.dds"},
	{0x3FE68156219A5F40ULL, "mag320_orsino6071177.cl"},
	{0x3FE687D63E52B53BULL, "dae000_loyal_mage6165240.cl"},
	{0x3FE6C44EBEEDACA7ULL, "764190743.ogg"},
	{0x3FE719B0B58BDD4EULL, "6180449_m.fxe"},
	{0x3FE7BE2AB83355D0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upperwin.mao"},
	{0x3FE7C22AB8335CE3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upperwin.mmh"},
	{0x3FE7D02AB833743DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upperwin.msh"},
	{0x3FE8369E3E4F92ECULL, "6121954_m.fxe"},
	{0x3FE8A867ACB64506ULL, "6072595_m.fxe"},
	{0x3FE8B5C76C8EF3BCULL, "t3_dog_d04.tnt"},
	{0x3FE8C9FB6E44C186ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action_loop.ani"},
	{0x3FE8CC6EFAA7F07FULL, "mrl111ar_house.are"},
	{0x3FE8FB7BD1CFE0D7ULL, "vxw_woundedcoast_15_0_0.vlm"},
	{0x3FE9EB2EA16655A5ULL, "vxht_0_hightown_a2_bft_0_0.vlm"},
	{0x3FEA950FD413E300ULL, "6050441_m.fxe"},
	{0x3FEAACEF0AB4D8B0ULL, "1032914512.ogg"},
	{0x3FEBD3155444439EULL, "vxlt_undercitydungeon_13w_0_0.vlm"},
	{0x3FEBD951A1D817ABULL, "lt_gallowstemplar_2382.rml"},
	{0x3FEC1219F919F859ULL, "dre000_vampire6034352.cl"},
	{0x3FEC7B0F3F6F02BAULL, "6031275_m.fxe"},
	{0x3FECA50F9A1D7117ULL, "vxlt_gallowscourtyard_57s_0_0.vlm"},
	{0x3FED27DC5AD4D807ULL, "414064114.ogg"},
	{0x3FED51995C0A5030ULL, "gen00fl_aveline6140002.cl"},
	{0x3FEDCBA1358AF25AULL, "vxca_caverns2prc_adx_0_0.vlm"},
	{0x3FEE1F28AA44D374ULL, "gen00fl_varric6080361.cl"},
	{0x3FEE6EDFAE81EAA3ULL, "992105054.ogg"},
	{0x3FEEA602EC697051ULL, "vxw_bonepitmining_a2_1zv_0_0.vlm"},
	{0x3FEEDBE0DF29705CULL, "vxca_caverns2_521_0_0.vlm"},
	{0x3FEEF54F9DA8028DULL, "6035922_m.fxe"},
	{0x3FEF4A89C9365D0CULL, "follower_banter6165421.cl"},
	{0x3FF028366675B56DULL, "crt_sw_dis.mao"},
	{0x3FF0284D580A3983ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv2.evt"},
	{0x3FF032121BFC6F7EULL, "6110880_m.fxe"},
	{0x3FF07436D3D5D591ULL, "vxdr_primevaltomb_7lr_0_0.vlm"},
	{0x3FF11DFEAB333085ULL, "hlac_lt_0_lowtown_a2_-131073_0x.dds"},
	{0x3FF127CCA8ECAC35ULL, "vxlt_0_lowtown_a2_d2n_0_0.vlm"},
	{0x3FF136D0B9E92120ULL, "vxht_0_hightown_n_b9c_0_0.vlm"},
	{0x3FF17B5B8B20970AULL, "vxlt_hangedman_n_s1_0_0.vlm"},
	{0x3FF1822A1E47A985ULL, "gen00fl_merrill6124406.cl"},
	{0x3FF18F431CF5BF96ULL, "6169493_m.fxe"},
	{0x3FF222D2A516E235ULL, "zz_dog_debug6127734.cl"},
	{0x3FF2270E4219AD01ULL, "6139954_m.fxe"},
	{0x3FF235D4C1D4CEFBULL, "6155994_m.fxe"},
	{0x3FF2C083A768483EULL, "140776858.ogg"},
	{0x3FF35EFD99E5F624ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_put_up_your_dukes.ani"},
	{0x3FF3618855706C4BULL, "257748934.ogg"},
	{0x3FF3FE875E1B7C93ULL, "ai_leadercommands.gda"},
	{0x3FF43B43E792AEF3ULL, "6026982_m.fxe"},
	{0x3FF4617AA4434ED6ULL, "vxlt_warehouse_tm_0_0.vlm"},
	{0x3FF48C46506FE967ULL, "495356906.ogg"},
	{0x3FF49BEA61A05304ULL, "hlac_lt_0_lowtown_docks_a3_-3_4y.dds"},
	{0x3FF4CB91C1042C73ULL, "127652631.ogg"},
	{0x3FF4F9D3068B7EE0ULL, "ser210_witness6168112.cl"},
	{0x3FF56147C3143C27ULL, "106288752.ogg"},
	{0x3FF5A2450B293F7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_clench_fist.ani"},
	{0x3FF5E8C7D94D618AULL, "6174958_m.fxe"},
	{0x3FF5F300D95FA6ADULL, "953829474.ogg"},
	{0x3FF61508F55CC2E1ULL, "vxlt_gallowstemplar_5qq_0_0.vlm"},
	{0x3FF6B1FB6E50D6F7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action_loop.evt"},
	{0x3FF6CF99EC020A84ULL, "6023960_m.fxe"},
	{0x3FF6DD2F84D0C109ULL, "579600648.ogg"},
	{0x3FF7014D58101604ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv2.gad"},
	{0x3FF72CC55E2A128CULL, "428977836.ogg"},
	{0x3FF76F124DD7BBB6ULL, "802851954.ogg"},
	{0x3FF7CE6A87350EF3ULL, "vxlt_darktown_42k_0_0.vlm"},
	{0x3FF7D43EEE395ACCULL, "fen120ip_to_mansion.ncs"},
	{0x3FF81941C4FC404CULL, "w_bonepitdestroyed_a3_20_0_lc.mmh"},
	{0x3FF821A122D55E67ULL, "montage_subs_demohack6217135.cl"},
	{0x3FF87A0D59E7EBD1ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_005a_n.dds"},
	{0x3FF8950E83C50814ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.dura.ani"},
	{0x3FF8C32AB841D108ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upperwin.phy"},
	{0x3FF8D5235D6A49C9ULL, "hlac_ht_mansion2_818_34x.dds"},
	{0x3FF8E63F79ECC168ULL, "lowtown_male_lowlife1_sta6165005.cl"},
	{0x3FF8E99E82757421ULL, "810475309.ogg"},
	{0x3FF8F77074CFCC9AULL, "vxlt_gallowscourtyard_7ev_0_0.vlm"},
	{0x3FF90ECDEC788B9CULL, "392987512.ogg"},
	{0x3FF95BFD99EA5C86ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_put_up_your_dukes.gad"},
	{0x3FF9C56EFAB6571CULL, "mrl111ar_house.lst"},
	{0x3FF9D3049D31545EULL, "vxht_bartrand_n_5xv_0_0.vlm"},
	{0x3FFA08E20B369490ULL, "hlac_ht_chantry_2534_38x.dds"},
	{0x3FFA146C7CE79823ULL, "vxlt_0_lowtown_a2_n_ex0_0_0.vlm"},
	{0x3FFA1F8AF661C28FULL, "llac_ht_chantry_2534_35.dds"},
	{0x3FFAEC836D1A953AULL, "vxht_keep_n_6hv_0_0.vlm"},
	{0x3FFAEEA67E842439ULL, "vxw_encounter2_camp_qw_0_1.vlm"},
	{0x3FFAF28D7F25BFE7ULL, "llac_lt_0_lowtown_docks_a3_-131075_2.dds"},
	{0x3FFB38E456466FA3ULL, "555875839.ogg"},
	{0x3FFBE50B0EEACDACULL, "w_bonepit_10_0_ll.mmh"},
	{0x3FFC02346D3D4B52ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_017_0s.dds"},
	{0x3FFC086B01606805ULL, "1370055.ogg"},
	{0x3FFC3D7E27810ADAULL, "vxca_caverns2prc_acy_0_0.vlm"},
	{0x3FFCB2FB6E554380ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action_loop.gad"},
	{0x3FFD104B5577C672ULL, "693571612.ogg"},
	{0x3FFD17DFDE5F88D1ULL, "vxw_bonepitmining_a2_1wu_0_0.vlm"},
	{0x3FFE104D58164EE2ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv2.ani"},
	{0x3FFF28A4E266B9E9ULL, "528389181.ogg"},
	{0x3FFF920E83CB2276ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.dura.gad"},
	{0x3FFFB67CE552FFABULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb107a_0d.dds"},
	{0x3FFFF8B099E869F0ULL, "vxlt_0_lowtown_a2_n_c_ho9_0_0.vlm"},
	{0x3FFFFEBA5A1DC7F0ULL, "972349703.ogg"},
	{0x4000212DA485035BULL, "m_dae120ar_hightown.wbd"},
	{0x400056FD99F07365ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_put_up_your_dukes.evt"},
	{0x40010ADC0C637A9DULL, "756852963.ogg"},
	{0x4001432CD6270B25ULL, "character_character_p109.psh.fxo"},
	{0x400195012C95A234ULL, "vxht_keep_n_8am_0_0.vlm"},
	{0x4001C4CE8479BA09ULL, "vxw_bonepitmining_a2_f5_0_0.vlm"},
	{0x4001C92E1EFB626DULL, "vxw_bonepitdestroyed_a3_29r_0_0.vlm"},
	{0x40021377DFCF7656ULL, "6044367_m.fxe"},
	{0x400213AC4ED164D9ULL, "367854381.ogg"},
	{0x4002A7B10514344DULL, "qun221_viscount6070168.cl"},
	{0x4002D35796396756ULL, "vxlt_warehouse_yn_0_0.vlm"},
	{0x40033916D8F7C7BDULL, "6136371_m.fxe"},
	{0x4003BA450B35A54FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_clench_fist.evt"},
	{0x4003C72C02EC8E5BULL, "character_character_p93.psh.fxo"},
	{0x40046896665D5941ULL, "vxlt_0_lowtown_eki_0_1.vlm"},
	{0x4004C3618581B63CULL, "vxht_keep_53e_0_0.vlm"},
	{0x4004D56A6FDE2678ULL, "llac_lt_0_lowtown_docks_-3_6.dds"},
	{0x4004F95803801A27ULL, "438106616.ogg"},
	{0x4005077C48885164ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.anb"},
	{0x4005253FCCF5C1FEULL, "hlac_ca_cavern1_haunted_24424_4z.dds"},
	{0x40055DE2086532DBULL, "dae000_meredith6085335.cl"},
	{0x4005B09E4827231FULL, "685798196.ogg"},
	{0x4005DBA8B5AC1B31ULL, "gen00fl_merrill6107801.cl"},
	{0x4005E7E8F65CB1D4ULL, "hlac_lt_gallowstemplar_f_2394_8y.dds"},
	{0x40064047792B2773ULL, "vxlt_darktown_49j_0_0.vlm"},
	{0x4006A2BC7665307CULL, "vxw_encounter3_a8_0_0.vlm"},
	{0x4006AD0E83D16ED5ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.dura.evt"},
	{0x4006CA30478DC1DAULL, "6137308_m.fxe"},
	{0x4006D604BA33E9EAULL, "221643817.ogg"},
	{0x4006FD8BC8E2CABAULL, "vxlt_0_lowtown_a2_emy_0_0.vlm"},
	{0x40070505FB19D0FFULL, "vxw_woundedcoast_hf_0_0.vlm"},
	{0x400726DF16F14A42ULL, "vxca_caverns2_9ar_0_0.vlm"},
	{0x4007446751BADF6FULL, "vxw_sundermount_2s9_0_0.vlm"},
	{0x40074CC01E4F67CAULL, "hlac_ca_caverns1_mines_24423_9z.dds"},
	{0x40078FE1C9934419ULL, "con_ico_combat.dds"},
	{0x4007CB75ED6E5E6DULL, "vxlt_0_lowtown_a2_f6p_0_0.vlm"},
	{0x40082EDF4AAFAFE2ULL, "176207779.ogg"},
	{0x400844E18F276CDEULL, "vxht_bartrand_n_5cw_0_0.vlm"},
	{0x4008B6CF3830749EULL, "vxlt_undercitydungeon_19i_0_0.vlm"},
	{0x4008CC49161160E2ULL, "vxlt_hangedman_1vb_0_0.vlm"},
	{0x4009F14729CB289EULL, "vxca_cavern1_haunted_hkk_0_0.vlm"},
	{0x400A1B2544B11553ULL, "gen00fl_aveline6137683.cl"},
	{0x400A475B722F9DD6ULL, "hlac_lt_0_lowtown_a2_n_-3_4z.dds"},
	{0x400ABB450B3BC5F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_clench_fist.gad"},
	{0x400B4E98133DF60FULL, "ave300pt_favour_fault.ncs"},
	{0x400B8E68AEB1B5F4ULL, "6048196_m.fxe"},
	{0x400C03DCFF39896CULL, "mrl310_merrill6120112.cl"},
	{0x400C6FBD6FB37CEEULL, "240601452.ogg"},
	{0x400CD22C99A1B132ULL, "6120014_m.fxe"},
	{0x400CD3604A48A65CULL, "6071164_f.fxe"},
	{0x400D57F0B7C20691ULL, "vxdr_primevaltomb_7bp_0_0.vlm"},
	{0x400DC5EBD725C39AULL, "mag360_alain6044620.cl"},
	{0x400E117014A6D019ULL, "vxca_cavern1_haunted_ioe_0_0.vlm"},
	{0x400E276BEDCE3FADULL, "m_fen121ar_mansion_nt.wbd"},
	{0x400E6550E6282586ULL, "root_rog_shadow.dds"},
	{0x400E801722DF9319ULL, "llac_lt_0_lowtown_a2_n_-65537_2.dds"},
	{0x400E841B50CBAF68ULL, "pro000_wesley6007718.cl"},
	{0x400ED90E66F1C2F3ULL, "vxdr_0_deeproads_gam_0_0.vlm"},
	{0x400EDA3DB57281CBULL, "vxht_mansion2_3eb_0_0.vlm"},
	{0x400F70FD862D63B3ULL, "6046874_m.fxe"},
	{0x400FEB6AEF5D5FF1ULL, "855113699.ogg"},
	{0x40103B0B10F17AEDULL, "vxw_bonepitdestroyed_a3_24q_0_0.vlm"},
	{0x40109B35DB7F7C86ULL, "dae000ip_crowd_em3_talk8.utp"},
	{0x4010B4AB1F7EF5CCULL, "vxw_encounter2_fv_0_0.vlm"},
	{0x4012135F446B3177ULL, "71991431.ogg"},
	{0x4012267C48931067ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.mmh"},
	{0x401226AA8C3806A9ULL, "6089789_m.fxe"},
	{0x40122DFEB9E68D02ULL, "vxht_0_hightown_a2_ca7_0_0.vlm"},
	{0x40123908F4E2A6DBULL, "character_character_p44.psh.fxo"},
	{0x4012BF940B1BF0A3ULL, "vxca_caverns1_mines_i19_0_0.vlm"},
	{0x4012E0DE95B966FEULL, "hlac_ca_cavern1_haunted_24421_10x.dds"},
	{0x401322C94117469AULL, "vxlt_0_lowtown_docks_bcm_0_0.vlm"},
	{0x40138911F7EED202ULL, "379474077.ogg"},
	{0x40143ECE00ACDB62ULL, "180918122.ogg"},
	{0x4014F7F92B261C1EULL, "llac_ht_0_hightown_131073_0.dds"},
	{0x40151A5C156A0AC9ULL, "vxw_sundermount_215_0_0.vlm"},
	{0x40154EA61B55F6C3ULL, "dae000_madam_lusine6152151.cl"},
	{0x40156866A927D510ULL, "869144112.ogg"},
	{0x4015C1117FA3456CULL, "vxlt_0_lowtown_docks_a3_n_con_0_0.vlm"},
	{0x4015D9661D6C02DAULL, "6082585_m.fxe"},
	{0x4016D5A77800C205ULL, "44369663.ogg"},
	{0x40178EF7597AE9F4ULL, "294892994.ogg"},
	{0x4017988F905CB8C0ULL, "6064929_m.fxe"},
	{0x4018006FD001B95CULL, "llac_lt_0_lowtown_docks_n_-131075_7.dds"},
	{0x401823D615344C34ULL, "6015845_f.fxe"},
	{0x4018D926707279FEULL, "vxht_mansion2b_24y_0_0.vlm"},
	{0x401916AD5E29062EULL, "6090410_m.fxe"},
	{0x40192A0189B288CBULL, "147493117.ogg"},
	{0x40194A0D1BD492FDULL, "487676715.ogg"},
	{0x401955229C386404ULL, "vxht_chantry_n_2ec_0_0.vlm"},
	{0x40195540AF20E12EULL, "vxw_woundedcoast_128_0_0.vlm"},
	{0x401955FF881E450BULL, "art\\vfx\\combat\\vfx_r_asn_regen_imp_c.mmh"},
	{0x40196A64D9B7EED2ULL, "zz_qun_debug6045396.cl"},
	{0x4019B06B8D22A9F5ULL, "551335753.ogg"},
	{0x4019D2484C6C8C01ULL, "gen00fl_anders6125793.cl"},
	{0x4019D3B4BD5E6198ULL, "vxht_mansion2_1vn_0_0.vlm"},
	{0x401A0045A0A7CE74ULL, "6043096_m.fxe"},
	{0x401A343929E90899ULL, "gen00fl_merrill6129557.cl"},
	{0x401AA4FC4A833BC3ULL, "6092404_m.fxe"},
	{0x401AD998A7C65541ULL, "vxlt_0_lowtown_docks_n_aoz_0_0.vlm"},
	{0x401B0591658C3BC2ULL, "hlac_w_blightlands_131074_0y.dds"},
	{0x401BB721C9E436BFULL, "ser191_emeric6108830.cl"},
	{0x401BB875FF97825BULL, "hlac_w_bonepit_0_13y.dds"},
	{0x401C7590869E399FULL, "554683482.ogg"},
	{0x401C85D0507ED2BBULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.impa.ani"},
	{0x401CC853D969B546ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.impa.ani"},
	{0x401D76C25E3E0176ULL, "vxht_keep_n_581_0_0.vlm"},
	{0x401D7C9CAFC77069ULL, "vxht_0_hightown_awp_0_0.vlm"},
	{0x401EBA52A41632BFULL, "w_encounter3_202_0.msh"},
	{0x401EC052A4163D61ULL, "w_encounter3_202_0.mmh"},
	{0x401ECC52A4165102ULL, "w_encounter3_202_0.mao"},
	{0x401F268811750E4CULL, "vxw_encounter2_aw_0_0.vlm"},
	{0x401F593046F2A3A3ULL, "821776686.ogg"},
	{0x401F5B15FFBDDCF5ULL, "6120894_f.fxe"},
	{0x401FE62B95DD957AULL, "6103822_m.fxe"},
	{0x40201A1F9CEF0F02ULL, "vxca_caverns1_efx_0_0.vlm"},
	{0x4020AAE5E6D8BF5BULL, "character_character_p35.psh.fxo"},
	{0x4020B891E6F6DC93ULL, "85100133.ogg"},
	{0x4020E7AE883FFC7BULL, "243588225.ogg"},
	{0x4021172BF1F2EAF5ULL, "vxlt_backalley_n_1jo_0_0.vlm"},
	{0x40219996C69FDA46ULL, "qun231tr_varnell_initiates.utt"},
	{0x4021C74D5837A11CULL, "hlac_dr_dungeon_3215_7x.dds"},
	{0x4021D321824BB97BULL, "hlac_w_woundedcoast_0_3y.dds"},
	{0x40225A704B688FA3ULL, "hlac_w_encounter3_-327675_0y.dds"},
	{0x4022719DD006FFCAULL, "vxw_bonepitdestroyed_a3_1o4_0_0.vlm"},
	{0x40238AC42C65F6A2ULL, "6105993_f.fxe"},
	{0x4023E367EEEA4E54ULL, "924406834.ogg"},
	{0x40240E03ED42CA18ULL, "mag311_keran6041560.cl"},
	{0x40244B430F621B5AULL, "6085485_m.fxe"},
	{0x4024791EF16A15E6ULL, "462138231.ogg"},
	{0x4024D3F6F1165113ULL, "gen00fl_fenris6105878.cl"},
	{0x4025B002F9FD91E2ULL, "vxlt_hangedman_1xl_0_0.vlm"},
	{0x4025CBCAEE81DA49ULL, "6065923_m.fxe"},
	{0x4025D5508375306EULL, "vxlt_0_lowtown_f4e_0_0.vlm"},
	{0x40265C7BA2E11762ULL, "739298257.ogg"},
	{0x40269814F9DE5A3AULL, "w_encounter1_42_0.phy"},
	{0x4026B60DBD0BF56AULL, "6043616_m.fxe"},
	{0x4026EDA4FB9DC502ULL, "follower_banter6167991.cl"},
	{0x4026F97540B89BFFULL, "282136089.ogg"},
	{0x402718E03AC69B61ULL, "gen00fl_anders6119228.cl"},
	{0x402734831C50CED5ULL, "6099403_m.fxe"},
	{0x402736FF882A4E70ULL, "art\\vfx\\combat\\vfx_r_asn_regen_imp_c.anb"},
	{0x402793AEBE7EB7E2ULL, "w_blightlands_44_0.rml"},
	{0x4027A05495E07845ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl_1.mao"},
	{0x40282F78AD9DC00EULL, "725003825.ogg"},
	{0x40284591AF547A18ULL, "vxht_mansion2_1yo_0_0.vlm"},
	{0x4028C0C4E9A26A2EULL, "860563220.ogg"},
	{0x402987DB268E0473ULL, "l_env_wal_gallows_angry_crowd.bnk"},
	{0x40298DD050896AB2ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.impa.evt"},
	{0x4029984ECCD41D6EULL, "703112068.ogg"},
	{0x4029EB3CD4F7117EULL, "6133568_m.fxe"},
	{0x4029EB6A7DDFAB6FULL, "412542825.ogg"},
	{0x402A5875DA38D408ULL, "vxlt_gallowstemplar_n_3n8_0_0.vlm"},
	{0x402A642339BCC199ULL, "6162091_m.fxe"},
	{0x402AB053D975C9B7ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.impa.evt"},
	{0x402B67E10205E71AULL, "6151357_m.fxe"},
	{0x402B9C8C232B01A2ULL, "qun110_petrice6037089.cl"},
	{0x402BBDB7B89B65FCULL, "vxlt_gallowscourtyard_65e_0_0.vlm"},
	{0x402BDA11AAE7F8DFULL, "seb221_brett_harimann6177264.cl"},
	{0x402BF9DF66E82C8FULL, "llac_lt_0_lowtown_a2_65535_2.dds"},
	{0x402C1B526018A48DULL, "fx_vol_light.dds"},
	{0x402C72EB2F707AD5ULL, "follower_banter6156015.cl"},
	{0x402CC649D1107E03ULL, "gen00fl_merrill6129605.cl"},
	{0x402D984D055A6942ULL, "hlac_ht_keep_n_4519_3z.dds"},
	{0x402D9865036B26CCULL, "vxw_encounter2_lt_0_0.vlm"},
	{0x402E258A82954BBCULL, "and230_mage_girl6109101.cl"},
	{0x402E47CF2540E023ULL, "vxlt_0_lowtown_docks_n_beu_0_0.vlm"},
	{0x402EC1683F35E4CAULL, "vxdr_primevaltomb_8f8_0_0.vlm"},
	{0x402F0405E32F40A9ULL, "455482337.ogg"},
	{0x402F26C2D95438A5ULL, "vxlt_0_lowtown_a2_hgq_0_0.vlm"},
	{0x402F7BE5D8F12C73ULL, "314814374.ogg"},
	{0x402FB2E5D95B81C4ULL, "gen00fl_anders6112891.cl"},
	{0x402FF6911A859012ULL, "zz_hawke_cu_nt_v1.cut"},
	{0x403010EAE02AC3D6ULL, "vxht_0_hightown_n_9h1_0_0.vlm"},
	{0x403078481F4DBA00ULL, "vxdr_primevaltomb_3gn_0_0.vlm"},
	{0x4030ACD0508FBE45ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.impa.gad"},
	{0x40319B634CA8703FULL, "vxht_store_c3_0_0.vlm"},
	{0x4031B153D97BEA40ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.impa.gad"},
	{0x40321D564B7A3AB7ULL, "vxw_woundedcoast_73_0_0.vlm"},
	{0x40324300AAB5FEF5ULL, "vxw_blightlands_136_0_0.vlm"},
	{0x40325E647EAEA38AULL, "6025124_m.fxe"},
	{0x403272686B03203AULL, "473399326.ogg"},
	{0x40328684E98C7A07ULL, "74408901.ogg"},
	{0x4032B19B2D72C3E3ULL, "336916836.ogg"},
	{0x4032D4742F88F0D1ULL, "vxht_playershouse_n_1r5_0_0.vlm"},
	{0x4032F5E0B59BA395ULL, "gen00fl_sebastian6188708.cl"},
	{0x40330605A76DFF5DULL, "vxlt_gamlenshouse_n_js_0_0.vlm"},
	{0x4033DFE0960D4D0BULL, "qun220cr_javaris6071496.cl"},
	{0x40341A19D3439C92ULL, "vxlt_darktown_2w7_0_0.vlm"},
	{0x403421DFEBF3AA62ULL, "vxlt_hangedman_1gm_0_0.vlm"},
	{0x40345EA8DEC42CE8ULL, "738554174.ogg"},
	{0x40348B497CC3CDDFULL, "vxht_0_hightown_n_b7j_0_0.vlm"},
	{0x4035ABD741877F70ULL, "rog_duelistattackbonus.dds"},
	{0x4035B1B623721231ULL, "dae200_cullen6140509.cl"},
	{0x4035B3931EC42AE6ULL, "vxlt_0_lowtown_eua_0_0.vlm"},
	{0x4035C5A8988BBC56ULL, "vxdr_primevalentrance_1ok_0_1.vlm"},
	{0x403617D9F6E8D229ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.active.gad"},
	{0x4036767146386094ULL, "vxlt_0_lowtown_a2_n_fge_0_0.vlm"},
	{0x4036CE7DBCC4599AULL, "vxht_keep_n_4nr_0_0.vlm"},
	{0x4036E0A5643FB462ULL, "vxht_mansion_n_5zo_0_0.vlm"},
	{0x4037A368C0AA4957ULL, "75782907.ogg"},
	{0x4037BD528BB28641ULL, "vxlt_0_lowtown_docks_n_aqx_0_0.vlm"},
	{0x4038098E283BD6CEULL, "vxw_bonepit_1rw_0_0.vlm"},
	{0x40387DAEBE8D0537ULL, "w_blightlands_44_0.msh"},
	{0x40388FAEBE8D239AULL, "w_blightlands_44_0.mao"},
	{0x403893AEBE8D2AA9ULL, "w_blightlands_44_0.mmh"},
	{0x4038A4D4553EA9EEULL, "hlac_w_bonepitdestroyed_a3_131072_0z.dds"},
	{0x4038C0CDBE6CA957ULL, "vxca_caverns1_e98_0_0.vlm"},
	{0x4038D04ECDF7ECA8ULL, "follower_banter6169300.cl"},
	{0x4038F9A4C1997AB0ULL, "cod_ltr_sk3_scribbles3.utp"},
	{0x4038FB94038BB420ULL, "ht_0_hightownexplosion_n_mmi.gff"},
	{0x4038FC37FF1569E9ULL, "vxca_cavern1_haunted_dqb_0_0.vlm"},
	{0x4039612B5C0A28B6ULL, "fxe_ripple_blend.mao"},
	{0x4039A80A6AFC0EAEULL, "511957213.ogg"},
	{0x4039CC214FC58246ULL, "vxca_caverns1_mines_iny_0_0.vlm"},
	{0x4039DE17478FA221ULL, "ser270_emeric.bnk"},
	{0x4039EB87ED2634C5ULL, "152692628.ogg"},
	{0x403A5C2814FEC14BULL, "mag110_ws_banter6079648.cl"},
	{0x403AED91315337BAULL, "lgt000cr_hate_slaver.utc"},
	{0x403B1FDB350E868EULL, "222444527.ogg"},
	{0x403B3304B73B6F43ULL, "zz_fol_debug6066155.cl"},
	{0x403B9D041714E7F2ULL, "hlac_lt_0_lowtown_docks_-262144_2z.dds"},
	{0x403C7950177ECD30ULL, "dae000_tomwise6167333.cl"},
	{0x403CB9AC1736F8A3ULL, "vxlt_0_lowtown_docks_n_bbt_0_0.vlm"},
	{0x403CE8D9F6EEA177ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.active.ani"},
	{0x403CFBBACF303CE9ULL, "1034496491.ogg"},
	{0x403D96FA5E65A3BDULL, "vxht_brothel_n_5k8_0_0.vlm"},
	{0x403DA812557A6CACULL, "6016331_m.fxe"},
	{0x403DDAAD23E7894DULL, "vxht_bartrand_n_646_0_0.vlm"},
	{0x403DFAE5D5DF1BF5ULL, "vxlt_backalley_n_1pm_0_0.vlm"},
	{0x403E00174793C491ULL, "ser270_emeric.cnv"},
	{0x403E02F65150A480ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgta_0n.dds"},
	{0x403E84B964A3835FULL, "vxht_mansion_n_66u_0_0.vlm"},
	{0x403F221BFEBE957AULL, "hm_lgt_roderick.mor"},
	{0x403F321C6DBE14C2ULL, "dae000ip_crowd_lm3_drunk1.utp"},
	{0x403F708C149189C0ULL, "hlac_lt_backalley_g_0_8x.dds"},
	{0x403FA163D33F85CEULL, "573039805.ogg"},
	{0x403FC56CE2DD65CEULL, "28111775.ogg"},
	{0x403FE537D69F8055ULL, "654982210.ogg"},
	{0x404019891DC7C738ULL, "6202154-anygoodseatsleft.cut"},
	{0x4040302DFD981571ULL, "hlac_dr_dungeon_3216_10z.dds"},
	{0x4040716055190EC0ULL, "vxlt_hangedman_n_26p_0_0.vlm"},
	{0x404078B5841767BFULL, "vxlt_0_lowtown_docks_a3_n_coa_0_0.vlm"},
	{0x4040AB9957AAC399ULL, "vxca_caverns2prc_9zq_0_0.vlm"},
	{0x4040BB1FBC6AEDC5ULL, "hla_ht_0_hightown_65534_0z.dds"},
	{0x4040CBDCD3AC221EULL, "vxca_cavern1_haunted_e87_0_0.vlm"},
	{0x4040CC79C2F232FFULL, "vxw_woundedcoast_tb_0_0.vlm"},
	{0x4040DA1048754454ULL, "passive_abilities_seb.gda"},
	{0x404139BE3AA16E36ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_templarbadge_dp.dds"},
	{0x404177E2996417DDULL, "vxlt_gamlenshouse_n_mr_0_0.vlm"},
	{0x404202B474E85068ULL, "vxw_blightlands_c2_0_0.vlm"},
	{0x4042531AABEF5A8CULL, "hlac_lt_0_lowtown_a2_n_65536_12z.dds"},
	{0x4042AB90A40BAD69ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.dura.ani"},
	{0x4042FA40D3809189ULL, "hlac_lt_0_lowtown_docks_a3_131070_2x.dds"},
	{0x4043D477F34D3E40ULL, "611604046.ogg"},
	{0x4043D8622E5FC652ULL, "6098920_m.fxe"},
	{0x4044183D003CFFD5ULL, "6093603_m.fxe"},
	{0x4044257010BA4366ULL, "vxlt_0_lowtown_epb_0_0.vlm"},
	{0x40443DAC3684E710ULL, "art\\levels\\races\\vfx_models\\prp_lichenglow.mao"},
	{0x404441AC3684EE23ULL, "art\\levels\\races\\vfx_models\\prp_lichenglow.mmh"},
	{0x40444FAC3685067DULL, "art\\levels\\races\\vfx_models\\prp_lichenglow.msh"},
	{0x404464D37F0174DBULL, "884253566.ogg"},
	{0x4044EF294FBC6F57ULL, "w_encounter3_148_0.phy"},
	{0x4045294B9340AB18ULL, "vxht_mansion2_1gm_0_0.vlm"},
	{0x40459222022F78E1ULL, "955021483.ogg"},
	{0x4045ADDDFCBFC8C7ULL, "727032400.ogg"},
	{0x4045AEFB34FD808CULL, "632192166.ogg"},
	{0x4045F33EEE7BB0CDULL, "fen120ip_to_mansion.utp"},
	{0x4045FEFE26205846ULL, "plt_mrl000pt_gift_halla.plo"},
	{0x4046431292FF3A6FULL, "938968033.ogg"},
	{0x4046910E79EC36FBULL, "200964602.ogg"},
	{0x4046A1B12A02874EULL, "pro000_colonel_ewald6040590.cl"},
	{0x4046CB7A9F2200CAULL, "30959417.ogg"},
	{0x4046DB888C489193ULL, "follower_banter6163664.cl"},
	{0x40470375256C7072ULL, "llac_lt_0_lowtown_docks_a3_-131073_1.dds"},
	{0x4048EE1006C5FD34ULL, "383258656.ogg"},
	{0x40495F8B89005A8BULL, "llac_ht_mansion_n_3369_2.dds"},
	{0x404970191281E85AULL, "vxca_caverns2prc_6t7_0_0.vlm"},
	{0x4049C690A411F9EBULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.dura.gad"},
	{0x404A299432062AF2ULL, "6138811_m.fxe"},
	{0x404A3572364C1CEDULL, "hlac_lt_warehouse_n_65536_3y.dds"},
	{0x404A44D324C0C85BULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.impa.ani"},
	{0x404A746298BF8CBCULL, "6096506_m.fxe"},
	{0x404A7C1EE75757CCULL, "vxw_encounter2_rj_0_0.vlm"},
	{0x404AA33ECD817E81ULL, "6074060_m.fxe"},
	{0x404AD075B09D4715ULL, "vxlt_hangedman_n_25x_0_0.vlm"},
	{0x404B0C319BBBC179ULL, "ss_erictest.cut"},
	{0x404BEBE9A9707E62ULL, "787929457.ogg"},
	{0x404C323872A356C1ULL, "6133501_m.fxe"},
	{0x404CF4A4C416F4D6ULL, "vxht_0_hightown_n_9v3_0_0.vlm"},
	{0x404D4688CF12F714ULL, "mag120_macha6095531.cl"},
	{0x404D81A92F6326F0ULL, "vxlt_hangedman_n_215_0_0.vlm"},
	{0x404DBF2E0A2603F6ULL, "mag160_feynriel6020085.cl"},
	{0x404DEBE300AD5C7DULL, "hlac_w_sundermount_327681_0y.dds"},
	{0x404E020FC720FCF3ULL, "67598341.ogg"},
	{0x404E3E1C8E934CE2ULL, "gen_im_cns_cft_wrn_06.uti"},
	{0x404F316BEE05E7D7ULL, "m_fen121ar_mansion_nt.bnk"},
	{0x404F3E56B4E84B7FULL, "vxw_woundedcoast_qm_0_0.vlm"},
	{0x404F498ADD053F7CULL, "460839514.ogg"},
	{0x404F74EBD7410A51ULL, "972521966.ogg"},
	{0x404FBB02863547BEULL, "ht_0_hightown_a2_28_0.rml"},
	{0x404FF785BC516C39ULL, "hm_lucky.mor"},
	{0x405008AFFA83B68BULL, "vxw_encounter2_camp_nu_0_0.vlm"},
	{0x4050749166DE68E8ULL, "vxw_blightlands_d3_0_0.vlm"},
	{0x40507DD3189652A2ULL, "vxlt_gallowstemplar_n_2ot_0_0.vlm"},
	{0x4050C390A4181350ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.dura.evt"},
	{0x405153987DAED4C6ULL, "602315368.ogg"},
	{0x40519FEF7316168BULL, "774530276.ogg"},
	{0x4051D610DB02FBBEULL, "6110448_m.fxe"},
	{0x4051FFCE96B53DBDULL, "583406096.ogg"},
	{0x40528AB94539BDC5ULL, "67347877.ogg"},
	{0x405380A6BEC7E8BDULL, "687764472.ogg"},
	{0x40538DCF10BD1D07ULL, "hlac_ca_caverns1_24425_10z.dds"},
	{0x4053DEF2CBD3D4F1ULL, "vxht_keep_n_63p_0_0.vlm"},
	{0x4054344FF853BA0FULL, "112635892.ogg"},
	{0x40543EBBC9738E95ULL, "71191103.ogg"},
	{0x4054CBDF69703D35ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb101a.mmh"},
	{0x4054CDDF697040ABULL, "art\\characters\\weapons\\longbows\\w_lbw_lb101a.msh"},
	{0x4054F8B957F2C374ULL, "308513017.ogg"},
	{0x405527630931A68AULL, "vxht_mansion_n_5h9_0_0.vlm"},
	{0x405566D2C52FB91DULL, "444568031.ogg"},
	{0x4055AAA1372FCA39ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be03a.mmh"},
	{0x4055B4A1372FDB27ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be03a.msh"},
	{0x405610620C3D9B04ULL, "6141589_m.fxe"},
	{0x4056A2223E27C86BULL, "mrl000_marethari6129417.cl"},
	{0x4057240061F5BF70ULL, "gen00fl_anders6125589.cl"},
	{0x405725C1E5B94C0CULL, "59780478.ogg"},
	{0x40572A42F070903EULL, "6183826_m.fxe"},
	{0x40574410779FD269ULL, "vxht_0_hightown_akt_0_0.vlm"},
	{0x405837A409E9DFBAULL, "dae000_mother6121073.cl"},
	{0x40583DC0C3BC6373ULL, "vxlt_warehouse_n_ch_0_0.vlm"},
	{0x40584CD324CD1352ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.impa.evt"},
	{0x405893026F798DFAULL, "6127197_f.fxe"},
	{0x40594C88B5E685BCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_alchem_sa.dds"},
	{0x40599D65FB2329A3ULL, "vxlt_0_lowtown_docks_n_btv_0_0.vlm"},
	{0x405A378D02220A02ULL, "gen00fl_varric6118732.cl"},
	{0x405A56C4D67AAE06ULL, "zz_followers_banter6157240.cl"},
	{0x405A8A4BB514F21FULL, "hlac_lt_gallowstemplar_f_2383_2x.dds"},
	{0x405A93CE35ECAA80ULL, "764856154.ogg"},
	{0x405AE86D89698BC6ULL, "vxlt_0_lowtown_a2_n_e1v_0_0.vlm"},
	{0x405AEF7457269849ULL, "6146996_m.fxe"},
	{0x405AF2EE7E48ED07ULL, "389780390.ogg"},
	{0x405BE7A0B80A8740ULL, "vxht_bartrand_n_4ev_0_0.vlm"},
	{0x405C0A9F69FAF1BDULL, "rdr110dg_guard_sleeping6016742.cl"},
	{0x405C367D3E588BDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ae_transition.evt"},
	{0x405C60B6FCA993B2ULL, "vxht_0_hightown_a3_2es_0_0.vlm"},
	{0x405C813C6AD0128CULL, "6095132_m.fxe"},
	{0x405CA008AC3E1B80ULL, "vxht_keep_n_5wv_0_0.vlm"},
	{0x405D0F4FF91A5024ULL, "gen00fl_varric6077824.cl"},
	{0x405D21FBAD1539C1ULL, "vxlt_0_lowtown_a2_n_eiu_0_0.vlm"},
	{0x405D42AF781D64AFULL, "gen_im_wep_war_two_iconic.uti"},
	{0x405D46C64510338BULL, "vxca_cavern1_haunted_k6j_0_0.vlm"},
	{0x405D47EB95B3C601ULL, "6168130_m.fxe"},
	{0x405D60314CD7134EULL, "6117298_m.fxe"},
	{0x405D62448AC25D5EULL, "6075859_m.fxe"},
	{0x405D63D65A44AE9DULL, "dae221cr_torelin.utc"},
	{0x405D98F6AB4C2C2FULL, "llac_lt_gallowstemplar_2383_3.dds"},
	{0x405DAE3B357D7160ULL, "llac_lt_0_lowtown_a2_n_c_196606_0.dds"},
	{0x405DEA02B84CAE71ULL, "716961049.ogg"},
	{0x405E72B50D5CF142ULL, "art\\vfx\\combat\\vfx_f_and_fury_c.mmh"},
	{0x405E7A8CEC79CF0BULL, "vxw_encounter2_camp_it_0_0.vlm"},
	{0x405EC4357F598EC7ULL, "hlac_w_encounter1_-1_0x.dds"},
	{0x405EE66E58D48168ULL, "vxw_blightlands_i0_0_0.vlm"},
	{0x405F6BD324D366E5ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.impa.gad"},
	{0x405FE0F6890AE517ULL, "ts_beasts_spiders_02.gda"},
	{0x405FE5CBCB46000AULL, "mer221_sabin6080568.cl"},
	{0x405FEE344DD63F2AULL, "6069611_f.fxe"},
	{0x40602E1AD82DC8C4ULL, "vxw_bonepit_c1_0_0.vlm"},
	{0x40605827A15F9370ULL, "cod_tomes_tome_0.utp"},
	{0x40605A18737C077EULL, "6165116_m.fxe"},
	{0x406073A7F0CBFA76ULL, "580690387.ogg"},
	{0x4060A42C366E7F0CULL, "vxht_0_hightown_c4r_0_0.vlm"},
	{0x4060FE8CA7525026ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_small.ani"},
	{0x4061101CDD7287C9ULL, "104225906.ogg"},
	{0x40617BCE0E004184ULL, "mag100_thrask6047537.cl"},
	{0x4061EE3D09F178EDULL, "gen_im_trs_drk_amu_r2.uti"},
	{0x40620D05772CDC18ULL, "vxht_mansion2_1mk_0_0.vlm"},
	{0x406261F9F2318014ULL, "gen00fl_varric6018423.cl"},
	{0x4062B5429BB962B9ULL, "vxca_caverns1_ckb_0_0.vlm"},
	{0x4062C985F8052C4DULL, "fen110_fenris6118116.cl"},
	{0x40634D7D3E5ED109ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ae_transition.gad"},
	{0x406355C30D9D8A37ULL, "vxlt_backalley_g_1jx_0_0.vlm"},
	{0x406408029360AC71ULL, "vxlt_0_lowtown_bxn_0_0.vlm"},
	{0x40641BD59DC32F1EULL, "vxlt_0_lowtown_a3_n_c_hru_0_0.vlm"},
	{0x406438DC1B466123ULL, "hlac_ht_0_hightown_65537_2z.dds"},
	{0x40648B67FF29E025ULL, "mag100_karras6052944.cl"},
	{0x4064FEEE542AD237ULL, "vxca_caverns1_e54_0_0.vlm"},
	{0x40650C90718AAA93ULL, "6102697_m.fxe"},
	{0x40655CA1662D1997ULL, "vxht_keep_6fb_0_0.vlm"},
	{0x4065BCFBA6A95117ULL, "439846467.ogg"},
	{0x4065E46488EED935ULL, "6115387_m.fxe"},
	{0x40664FDAB3B37B5EULL, "dae000_pc_home_ambients6208558.cl"},
	{0x40668D929961AB9DULL, "vxw_bonepitdestroyed_a3_lq_0_0.vlm"},
	{0x406696E6A7A73EA3ULL, "dae200_bran6094718.cl"},
	{0x40671BDFC286D564ULL, "6029364_m.fxe"},
	{0x4067702B62AA521CULL, "vxlt_0_lowtown_a2_ew9_0_0.vlm"},
	{0x4067E0B1B31D9331ULL, "dae000_arishok6078424.cl"},
	{0x4067FF656434CAFFULL, "441432605.ogg"},
	{0x4068275296796916ULL, "follower_banter6156191.cl"},
	{0x406832D51FFBDC9AULL, "vxw_bonepitdestroyed_a3_1ev_0_0.vlm"},
	{0x40684915385FA07BULL, "6035357_m.fxe"},
	{0x40684D0D4CCD4210ULL, "mag100_cullen6082325.cl"},
	{0x406888DC070E46CAULL, "vxdr_primevaltomb_8r4_0_0.vlm"},
	{0x40689BB8CE9A837BULL, "vrc321_varric6096944.cl"},
	{0x4068B5BC9286DB5BULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztablelong_0.pwk"},
	{0x4069272465439E25ULL, "267716534.ogg"},
	{0x4069524CC2CF3E48ULL, "art\\vfx\\combat\\vfx_a_stealth_act_c.anb"},
	{0x40696AB686C2F26AULL, "vxht_chantry_n_27k_0_0.vlm"},
	{0x4069A263CDBDEC73ULL, "115479031.ogg"},
	{0x4069A9BF0DC5D109ULL, "art\\vfx\\combat\\vfx_c_pride_cast_s_c.mmh"},
	{0x406A1E7D3E64A057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ae_transition.ani"},
	{0x406A9865DEE5855FULL, "llac_lt_0_lowtown_n_65538_2.dds"},
	{0x406A997724594073ULL, "one100dg_lia6014137.cl"},
	{0x406AA598E6970797ULL, "6065046_m.fxe"},
	{0x406AA71233747756ULL, "zz_vault_debug6178922.cl"},
	{0x406AB9CC75824D8AULL, "art\\vfx\\materials\\vfx_glyph_repulsion.mao"},
	{0x406AF2148D7932AEULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.cess.ani"},
	{0x406B11E59E343400ULL, "vxht_keep_n_5xw_0_0.vlm"},
	{0x406B6132959BC0BEULL, "6142436_m.fxe"},
	{0x406B63AF43C52090ULL, "vxdr_primevaltomb_77b_0_0.vlm"},
	{0x406B7BC629776317ULL, "vxlt_gallowstemplar_n_45u_0_0.vlm"},
	{0x406B93D89F0B5241ULL, "vxlt_0_lowtown_a2_n_ert_0_0.vlm"},
	{0x406C668F1D437CC8ULL, "6119873_f.fxe"},
	{0x406D584B4ACA99E8ULL, "vxw_blightlands_r1_0_0.vlm"},
	{0x406DE68CA75CB197ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_small.evt"},
	{0x406DEE89F46D2812ULL, "568980280.ogg"},
	{0x406E9FF7CA23E144ULL, "vxw_bonepit_n6_0_0.vlm"},
	{0x406EB334EFBF8476ULL, "ser120_ghyslain6118391.cl"},
	{0x406EC0A43E9B2CC5ULL, "art\\vfx\\combat\\vfx_w_wns_front_close_c.anb"},
	{0x406F58C8A5810FABULL, "6109230_m.fxe"},
	{0x406F61D6CB4E866CULL, "6162951_m.fxe"},
	{0x406F86A3E5260E53ULL, "follower_banter6154006.cl"},
	{0x406FC649AC3451F7ULL, "6160965_m.fxe"},
	{0x406FDFFF63DD59C4ULL, "vxca_cavern1_haunted_iib_0_0.vlm"},
	{0x4070393469FCE62CULL, "vxht_0_hightown_a3_b9i_0_0.vlm"},
	{0x40708C96C6E3981FULL, "qun231tr_varnell_initiates.ncs"},
	{0x4071271F8DAF7B39ULL, "vxca_caverns1_cna_0_0.vlm"},
	{0x4071A5DB26CBECD1ULL, "l_env_wal_gallows_angry_crowd.wbd"},
	{0x4071CAD49F6F1AC5ULL, "6089532_m.fxe"},
	{0x4073157B603BE528ULL, "vxlt_backalley_n_1l0_0_0.vlm"},
	{0x4073ED1FDAAE67C3ULL, "6036701_m.fxe"},
	{0x4074519EB239D436ULL, "1051677508.ogg"},
	{0x4074621296247148ULL, "780460355.ogg"},
	{0x407484FABEAECE62ULL, "vxw_encounter2_camp_5e_0_0.vlm"},
	{0x4074A050CF02886FULL, "hlac_ca_caverns1_24426_0x.dds"},
	{0x4074E78CA762D220ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_small.gad"},
	{0x4074EFD2AA66E390ULL, "649903391.ogg"},
	{0x4075198576BD16C3ULL, "vxw_encounter1_bh_0_0.vlm"},
	{0x40752AD13088BBC6ULL, "6149258_f.fxe"},
	{0x40755242E8E86E40ULL, "649383308.ogg"},
	{0x40755FBD15C5BD42ULL, "6109110_m.fxe"},
	{0x407563377CDF237BULL, "follower_banter6147972.cl"},
	{0x4075953586C2E672ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wlkf.gad"},
	{0x4075ACB9B8887864ULL, "77438296.ogg"},
	{0x4076A4B211F1F51AULL, "vxw_bonepitdestroyed_a3_1nw_0_0.vlm"},
	{0x4077BFA29DC473BDULL, "6069664_m.fxe"},
	{0x4077F5984F00BF7FULL, "679373402.ogg"},
	{0x4078448929922ECAULL, "clf221_jalen6069477.cl"},
	{0x4078C1CA769EC1C9ULL, "vxw_encounter2_9f_0_0.vlm"},
	{0x4078CC93CC670A48ULL, "24238258.ogg"},
	{0x4078EA148D8562FFULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.cess.evt"},
	{0x4078F9570536D048ULL, "vxca_caverns2_82t_0_0.vlm"},
	{0x40790FEB95057117ULL, "vxht_0_hightown_n_bpw_0_0.vlm"},
	{0x407913B11C5D0061ULL, "601455710.ogg"},
	{0x407917D6C4B56B3BULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.mmh"},
	{0x40792B3B8C10FEE3ULL, "6112916_f.fxe"},
	{0x40792C6B7B34731CULL, "503042501.ogg"},
	{0x407983A81DD7AD58ULL, "11722643.ogg"},
	{0x4079BBF826CEBB39ULL, "139918363.ogg"},
	{0x407A05B591016AC1ULL, "vxlt_0_lowtown_a2_n_ewk_0_0.vlm"},
	{0x407A2B64DFA5FA99ULL, "19661046.ogg"},
	{0x407A31A40615F47CULL, "vxw_encounter3_90_0_0.vlm"},
	{0x407A5DD015034AA5ULL, "vxht_0_hightown_a2_b4z_0_0.vlm"},
	{0x407B74B915A27972ULL, "6125606_m.fxe"},
	{0x407B9E8E1DFC3EF5ULL, "6201516_m.fxe"},
	{0x407BAA9FD7F8D315ULL, "243254829.ogg"},
	{0x407C50F1CA053FE1ULL, "pro000_merchant6051816.cl"},
	{0x407C983586C909A9ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wlkf.evt"},
	{0x407D11D4BC19F9C4ULL, "vxw_bonepit_i7_0_0.vlm"},
	{0x407D28A787594171ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_diagonal.mmh"},
	{0x407D34A787595512ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_diagonal.mao"},
	{0x407D42A787596D2FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_diagonal.msh"},
	{0x407D590568FCCF9DULL, "hlac_lt_0_lowtown_docks_n_131072_1y.dds"},
	{0x407D8C5780E410C8ULL, "gencr_gate_slave_r1.utc"},
	{0x407DBB81FDF45153ULL, "vxht_0_hightown_a3_9v0_0_0.vlm"},
	{0x407DC576D6F99B26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_exit.evt"},
	{0x407DE27E35D83CE1ULL, "vxca_caverns2_89p_0_0.vlm"},
	{0x407E3E42A927E75AULL, "vxca_caverns2prc_a6r_0_0.vlm"},
	{0x407ECD8F2D22D7C8ULL, "6083588_m.fxe"},
	{0x407EE5635C4B5791ULL, "vxlt_darktown_418_0_0.vlm"},
	{0x407F462A38470944ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rl_basic.msh"},
	{0x407F582A384727ADULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rl_basic.mao"},
	{0x407F5C2A38472EE6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rl_basic.mmh"},
	{0x407FCB148D8B4CE8ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.cess.gad"},
	{0x4080452C01388BFFULL, "6108465_f.fxe"},
	{0x4080C718A3B253C3ULL, "hlac_lt_0_lowtown_docks_a3_-65537_0y.dds"},
	{0x408121AA20CE0838ULL, "gen00fl_merrill6130079.cl"},
	{0x408187585231FDA8ULL, "vxlt_backalley_n_1w1_0_0.vlm"},
	{0x408203548CEBD819ULL, "isa211_isabela6198640.cl"},
	{0x40826CEDCE711B4DULL, "zz_approval_debug6003086.cl"},
	{0x40829114FA2CAD63ULL, "w_encounter1_42_0.msh"},
	{0x40829BAD321B3631ULL, "154048173.ogg"},
	{0x40829F14FA2CC4BDULL, "w_encounter1_42_0.mmh"},
	{0x4082A11411284CC1ULL, "6165133_m.fxe"},
	{0x4082A314FA2CCBCEULL, "w_encounter1_42_0.mao"},
	{0x4082C178A87EE863ULL, "vxlt_0_lowtown_docks_bgd_0_0.vlm"},
	{0x4082F635DDC4EF73ULL, "dae300_bran6143442.cl"},
	{0x408317A35E009105ULL, "6076390_m.fxe"},
	{0x40833261DA7FAE5DULL, "hlac_lt_0_lowtown_docks_a3_196607_2y.dds"},
	{0x408343EF85EDDBA3ULL, "hlac_ht_0_hightown_a2_65536_3x.dds"},
	{0x40836485AE11D951ULL, "w_encounter3_237_0.mmh"},
	{0x40837085AE11EDF2ULL, "w_encounter3_237_0.mao"},
	{0x408377F8E03D8850ULL, "llac_lt_0_lowtown_-1_7.dds"},
	{0x40837E85AE12050FULL, "w_encounter3_237_0.msh"},
	{0x40838B6268B32F43ULL, "vxw_encounter1_gk_0_0.vlm"},
	{0x40839357999EACF3ULL, "vxlt_warehouse_n_pm_0_0.vlm"},
	{0x4083DC76D6FE2E71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_exit.gad"},
	{0x4084037A9E5B68D9ULL, "hlac_ht_chantry_n_2534_30x.dds"},
	{0x408413871456F445ULL, "6135243_f.fxe"},
	{0x4084171582AF2F76ULL, "70638952.ogg"},
	{0x408450644DD13F4BULL, "vxw_encounter3_bb_0_0.vlm"},
	{0x4084895A50B7FD53ULL, "vxht_brothel_n_6dh_0_0.vlm"},
	{0x4084D9A5921D7768ULL, "6065298_m.fxe"},
	{0x408521FFFF8C20BFULL, "art\\characters\\playercharacter\\humanmale\\rob_appb.msh"},
	{0x408527FFFF8C2B61ULL, "art\\characters\\playercharacter\\humanmale\\rob_appb.mmh"},
	{0x408533FFFF8C3F02ULL, "art\\characters\\playercharacter\\humanmale\\rob_appb.mao"},
	{0x4085490BCC3765C5ULL, "91447765.ogg"},
	{0x40858D17BEEDF5FAULL, "vxdr_primevaltomb_50k_0_0.vlm"},
	{0x4085BC5A0B26B969ULL, "463155764.ogg"},
	{0x4085C7EF3A233086ULL, "vxlt_hangedman_o4_0_0.vlm"},
	{0x408638D6C4C02DA0ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_frd_loop_c.anb"},
	{0x4086FC5BEA4E651AULL, "633610971.ogg"},
	{0x4087536C3479C292ULL, "vxw_bonepitdestroyed_a3_me_0_0.vlm"},
	{0x408771B50D7FF4E9ULL, "art\\vfx\\combat\\vfx_f_and_fury_c.anb"},
	{0x40884403081D7A29ULL, "vxht_mansion_n_3fb_0_0.vlm"},
	{0x4088779282F78341ULL, "vxlt_0_lowtown_a2_n_exj_0_0.vlm"},
	{0x4088B3DBEC8FA2E9ULL, "577982036.ogg"},
	{0x4088B80286650862ULL, "ht_0_hightown_a2_28_0.anb"},
	{0x4089532A384F73B1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rl_basic.phy"},
	{0x4089ADE08B453425ULL, "hlac_w_blightlands_196605_0y.dds"},
	{0x4089D902199A5C02ULL, "6090393_m.fxe"},
	{0x408A00AAAD6DC069ULL, "qun211_braggart6075532.cl"},
	{0x408A11698C6EB805ULL, "298476162.ogg"},
	{0x408A37037ABB09DEULL, "hlac_lt_0_lowtown_a2_-262144_0y.dds"},
	{0x408A7025E288DAEEULL, "hlac_lt_gallowstemplar_f_2394_1x.dds"},
	{0x408A7D6D0EB5D475ULL, "6130126_m.fxe"},
	{0x408A803586D51E18ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wlkf.ani"},
	{0x408ADD76D7044E1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_exit.ani"},
	{0x408B085D13C89FE3ULL, "gen00fl_fenris6103344.cl"},
	{0x408B14F73F2A49B7ULL, "ico_boots_med_2.dds"},
	{0x408B2E3231201D6AULL, "314807973.ogg"},
	{0x408B410B299D4BB8ULL, "w_woundedcoast_88_0.rml"},
	{0x408B6CBE515DB74BULL, "vxht_0_hightown_n_bri_0_0.vlm"},
	{0x408BB8F6F0AC900FULL, "follower_banter6171920.cl"},
	{0x408C4F5205E4C730ULL, "6219448_m.fxe"},
	{0x408C97D5B02209ACULL, "vxht_mansion_n_4v8_0_0.vlm"},
	{0x408CA51E528A169FULL, "gen_im_trs_drk_two_2hm_r1.uti"},
	{0x408CB2F1CFB75816ULL, "6118945_m.fxe"},
	{0x408CC3DB533E3761ULL, "258856906.ogg"},
	{0x408CC4294FF9DBAFULL, "w_encounter3_148_0.mao"},
	{0x408CD0294FF9F04CULL, "w_encounter3_148_0.mmh"},
	{0x408CD6294FF9FA3EULL, "w_encounter3_148_0.msh"},
	{0x408E33C2CFE5848EULL, "mag121ip_idunna_clue01.ncs"},
	{0x408E3834A719CA07ULL, "isa311_isabela6092378.cl"},
	{0x408E46E9D21F4F43ULL, "260083073.ogg"},
	{0x408E6311AD6492A1ULL, "vxlt_undercitydungeon_1wx_0_0.vlm"},
	{0x408EAE8E835B4BC5ULL, "6087732_m.fxe"},
	{0x408FF063CC6D6A4CULL, "vxht_brothel_n_6cz_0_0.vlm"},
	{0x409015335BDB02C7ULL, "gen_im_arm_cht_hvy_champion.uti"},
	{0x40901B23CB6ECB22ULL, "6127033_m.fxe"},
	{0x4090A8DEBA97F5D3ULL, "llac_dr_0_deeproads_17038_2.dds"},
	{0x4090DFFE26600E85ULL, "plt_mrl000pt_gift_halla.nss"},
	{0x4090F754304824F1ULL, "mag320_orsino6071071.cl"},
	{0x40918B7F2B22CEF1ULL, "642795356.ogg"},
	{0x4091D24CA95EB315ULL, "6149374_f.fxe"},
	{0x409205348B94C573ULL, "vxlt_warehouse_n_wl_0_0.vlm"},
	{0x4092514CC2F241E3ULL, "art\\vfx\\combat\\vfx_a_stealth_act_c.mmh"},
	{0x40926E1A490404D3ULL, "mrl310_party_comments6143051.cl"},
	{0x4092A0779D463966ULL, "llac_ca_cavern1_haunted_24424_0.dds"},
	{0x4092A6BF0DE8D13EULL, "art\\vfx\\combat\\vfx_c_pride_cast_s_c.anb"},
	{0x4093FEF4B0E40E7AULL, "vxdr_primevaltomb_55d_0_0.vlm"},
	{0x4094015BE02524F8ULL, "mag170_hugh6080817.cl"},
	{0x409421B84F752E90ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avel.mmh"},
	{0x40942FB84F7546AAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avel.msh"},
	{0x40951BB2B3A12DEAULL, "6135153_m.fxe"},
	{0x40953602F91EAB3CULL, "vxht_keep_5as_0_0.vlm"},
	{0x409551E3A79A9B13ULL, "vxca_cave_2ga_0_0.vlm"},
	{0x4095DDA6C45BFC7BULL, "374757835.ogg"},
	{0x4095F3A578F1A217ULL, "vxht_0_hightown_n_bfq_0_0.vlm"},
	{0x40960F6B6C5FCA31ULL, "vxca_cavern1_haunted_ghf_0_0.vlm"},
	{0x409610F1B997CFC7ULL, "zz_slate5_hcu.cut"},
	{0x40964D5B2675674EULL, "vxdr_primevalentrance_23s_0_2.vlm"},
	{0x40967E40B4CBDBEAULL, "840080723.ogg"},
	{0x4096EE15B01BA449ULL, "365168361.ogg"},
	{0x40971A575E547B3FULL, "static_static_p103.psh.fxo.dx11"},
	{0x409725EBAD63C09CULL, "6151320_m.fxe"},
	{0x4097C1A43EBE3306ULL, "art\\vfx\\combat\\vfx_w_wns_front_close_c.mmh"},
	{0x4097D9C5BAAFE8DEULL, "hlac_lt_gallowsprison_n_-1_12z.dds"},
	{0x40981A836673585BULL, "6088920_m.fxe"},
	{0x4098794098BA58EBULL, "6152667_m.fxe"},
	{0x4098926F300F977CULL, "86608459.ogg"},
	{0x4098A42A2CCDC8CEULL, "516721165.ogg"},
	{0x4098BAB7478C1312ULL, "6167037_m.fxe"},
	{0x4098BBF5D0463599ULL, "95276675.ogg"},
	{0x4098C2A910CCD040ULL, "seb321_leliana6181757.cl"},
	{0x409924D297181CBDULL, "fen311cr_praescia.utc"},
	{0x40992970ABCD699EULL, "vxlt_undercitydungeon_1oc_0_0.vlm"},
	{0x4099D101D2ADA2BAULL, "vxht_keep_n_61i_0_0.vlm"},
	{0x409A9F3BE1E08253ULL, "vxht_0_hightown_a3_9h2_0_0.vlm"},
	{0x409ACD104CF65FB2ULL, "hlac_w_blightlands_-65537_0x.dds"},
	{0x409B04CEAB152E04ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comf_0n.dds"},
	{0x409B513C529E0650ULL, "509167574.ogg"},
	{0x409B65EA9D6A2C00ULL, "6048238_f.fxe"},
	{0x409C00F3AF857974ULL, "vxlt_0_lowtown_a2_n_c_gcj_0_0.vlm"},
	{0x409C7BED0402B452ULL, "vxht_brothel_n_6an_0_0.vlm"},
	{0x409CF9571484BB99ULL, "6171600_m.fxe"},
	{0x409D5FD23C3DC3E8ULL, "hlac_lt_undercitydungeon_-1_10x.dds"},
	{0x409DCAF91D6AB932ULL, "35130538.ogg"},
	{0x409E186ABAD02D60ULL, "298168211.ogg"},
	{0x409EED58A51DC2ECULL, "180768118.ogg"},
	{0x409F39A0C3545C07ULL, "gen00fl_aveline6139887.cl"},
	{0x409F53980F49B2C1ULL, "rdr111dg_harbormaster6107806.cl"},
	{0x409F7273A68E91DEULL, "6018677_f.fxe"},
	{0x409F76F1C56A4991ULL, "6114040_f.fxe"},
	{0x40A077117D8ADDF3ULL, "vxlt_warehouse_n_zo_0_0.vlm"},
	{0x40A0A7FFB6275934ULL, "vxw_woundedcoast_1bb_0_0.vlm"},
	{0x40A0AF9A6E82CFD4ULL, "vxw_sundermount_2je_0_0.vlm"},
	{0x40A0D22375FBF412ULL, "hlac_lt_0_lowtown_a3_n_c_1_3y.dds"},
	{0x40A1116D4ECE9BC3ULL, "vxdr_primevaltomb_9cs_0_0.vlm"},
	{0x40A11D1C1EF5AF6AULL, "vxca_cavern1_haunted_fo9_0_0.vlm"},
	{0x40A14C307680DBE0ULL, "6173650_m.fxe"},
	{0x40A16D1434A42E53ULL, "vxht_brothel_n_6jj_0_0.vlm"},
	{0x40A1B9A1E01D587BULL, "168950877.ogg"},
	{0x40A1D2A675FDD084ULL, "773088917.ogg"},
	{0x40A20BE373E4E232ULL, "6069872_m.fxe"},
	{0x40A29B1E1DB0A335ULL, "307992597.ogg"},
	{0x40A2FB6DDAD39427ULL, "vxca_caverns1_icr_0_0.vlm"},
	{0x40A30D747FEB13DCULL, "vxht_brothel_n_4g0_0_0.vlm"},
	{0x40A311AE98282702ULL, "hlac_ca_caverns2prc_5802_65y.dds"},
	{0x40A352DE7B76FBB6ULL, "6039460_m.fxe"},
	{0x40A3932152CC9848ULL, "llac_w_encounter2_camp_131074_0.dds"},
	{0x40A3956AB4B43B9AULL, "vxlt_0_lowtown_docks_b5g_0_0.vlm"},
	{0x40A39873C0B0971CULL, "vxht_keep_n_6i7_0_0.vlm"},
	{0x40A3DEAC741929FDULL, "vxdr_0_deeproads_cms_0_0.vlm"},
	{0x40A3EED3EC0EFDF0ULL, "hlac_w_encounter2_131076_0z.dds"},
	{0x40A4D960A745644AULL, "6206962_f.fxe"},
	{0x40A4E4A5E1AF0A91ULL, "hlac_lt_0_lowtown_n_1_6z.dds"},
	{0x40A52F890EDB39ECULL, "739718843.ogg"},
	{0x40A564A046E033C5ULL, "640489189.ogg"},
	{0x40A56A6D20D41010ULL, "qun131_taskmaster6059029.cl"},
	{0x40A5A9732913A32BULL, "172181210.ogg"},
	{0x40A5C4A095B80D48ULL, "6164629_m.fxe"},
	{0x40A5D3F7751A9B6BULL, "gen00fl_isabela6109795.cl"},
	{0x40A65F3A04F65365ULL, "6071225_m.fxe"},
	{0x40A68C60586970DBULL, "6083820_m.fxe"},
	{0x40A7A13507830273ULL, "hlac_ht_mansion2_818_34z.dds"},
	{0x40A850783549E84BULL, "vxht_0_hightown_n_bxk_0_0.vlm"},
	{0x40A86BC72D21B684ULL, "hlac_ht_store_187_4y.dds"},
	{0x40A92EBF5F5EB9E1ULL, "vxht_0_hightown_b9i_0_0.vlm"},
	{0x40A93F0B29B64595ULL, "w_woundedcoast_88_0.msh"},
	{0x40A9410B29B6490BULL, "w_woundedcoast_88_0.mmh"},
	{0x40A94D0B29B65DA8ULL, "w_woundedcoast_88_0.mao"},
	{0x40A985BE0391DFBAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\runleanl.evt"},
	{0x40A9C4C616EC4CC3ULL, "541819438.ogg"},
	{0x40A9CB5F01664BF2ULL, "vxlt_backalley_g_1nu_0_0.vlm"},
	{0x40A9D7C78F604480ULL, "6043338_f.fxe"},
	{0x40AA3ABACD7BC956ULL, "vxdr_primevalentrance_36w_0_0.vlm"},
	{0x40AA72335AE36B9BULL, "llac_ca_caverns1_24423_9.dds"},
	{0x40AA84708C343DBFULL, "mag100_karras6052959.cl"},
	{0x40AAEDC9F5F8CCD2ULL, "vxht_brothel_n_6jo_0_0.vlm"},
	{0x40AAF6D33B0ECB70ULL, "261155585.ogg"},
	{0x40AB915B9ECEC84EULL, "392116101.ogg"},
	{0x40ABFE6EBFF34239ULL, "vxw_encounter2_ng_0_0.vlm"},
	{0x40AC2E88ED6DD6D2ULL, "6151281_m.fxe"},
	{0x40AC55265791A99EULL, "vxlt_0_lowtown_a3_n_c_hkj_0_0.vlm"},
	{0x40AC74F253BC4A2CULL, "vxlt_0_lowtown_a2_n_h1u_0_0.vlm"},
	{0x40ACD41DB0599B4CULL, "vxht_brothel_n_6ix_0_0.vlm"},
	{0x40AD7585937B2C7FULL, "vxca_caverns1_edc_0_0.vlm"},
	{0x40AD8B64F9D3B4F1ULL, "art\\characters\\creatures\\mubari\\animation\\prst.ani"},
	{0x40AD9524E514D96BULL, "fen110_fenris6101967.cl"},
	{0x40ADA96D190CE4DCULL, "645910210.ogg"},
	{0x40AE4A135703239AULL, "vxdr_primevaltomb_8xv_0_0.vlm"},
	{0x40AEAF96437074FDULL, "vxht_chantry_2iu_0_0.vlm"},
	{0x40AEBCB6A6137F1BULL, "61386878.ogg"},
	{0x40AF21776078E854ULL, "vxw_sundermount_2oj_0_0.vlm"},
	{0x40AF8AECFDA5EF72ULL, "6083959_m.fxe"},
	{0x40AFA97C1C78CC9CULL, "vxlt_0_lowtown_a2_en2_0_0.vlm"},
	{0x40B010D2E95C720FULL, "one110_pryce6088777.cl"},
	{0x40B016339BDC2860ULL, "vxht_0_hightown_a3_cb1_0_0.vlm"},
	{0x40B051EA1FA6D422ULL, "vxlt_0_lowtown_n_exi_0_0.vlm"},
	{0x40B0AF594891A46DULL, "and121_karl6093904.cl"},
	{0x40B0E9CEA609F8DAULL, "lt_0_lowtown_docks_a3_120_0.mmh"},
	{0x40B0EDCEA609FF11ULL, "lt_0_lowtown_docks_a3_120_0.mao"},
	{0x40B0FBCEA60A1720ULL, "lt_0_lowtown_docks_a3_120_0.msh"},
	{0x40B106DBACD4E230ULL, "6116390_m.fxe"},
	{0x40B128FFFFB167DEULL, "art\\characters\\playercharacter\\humanmale\\rob_appb.phy"},
	{0x40B13A898DF8D81AULL, "gen00fl_merrill6124050.cl"},
	{0x40B18CD21613C262ULL, "6093904_f.fxe"},
	{0x40B25089660F427DULL, "vxdr_0_deeproads_cjr_0_0.vlm"},
	{0x40B2B77520B365C8ULL, "vxht_0_hightown_b9l_0_0.vlm"},
	{0x40B2CEEF93427A56ULL, "vxlt_0_lowtown_a2_eia_0_0.vlm"},
	{0x40B36EBBC487ED29ULL, "vxlt_0_lowtown_exy_0_0.vlm"},
	{0x40B37A67336ECA79ULL, "1052125638.ogg"},
	{0x40B38741960AE0C7ULL, "6180426_m.fxe"},
	{0x40B3A23E7091E369ULL, "hlac_lt_gallowstemplar_n_2393_0y.dds"},
	{0x40B3BC087F2B694CULL, "hlac_dr_0_deeproads_5842_3x.dds"},
	{0x40B3E02811477957ULL, "vxht_playershouse_n_1uf_0_0.vlm"},
	{0x40B4199937D8A353ULL, "vxw_sundermount_20x_0_0.vlm"},
	{0x40B4546887E7C31FULL, "hlac_ht_playershouse_n_855_0z.dds"},
	{0x40B45D629242BDDEULL, "408296425.ogg"},
	{0x40B4647BAF4A5053ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.cess.ani"},
	{0x40B468CE6E08DED0ULL, "eye_eye_p1.psh.fxo.dx11"},
	{0x40B49664F9D9E593ULL, "art\\characters\\creatures\\mubari\\animation\\prst.gad"},
	{0x40B52FC429705D56ULL, "vxht_keep_n_6gg_0_0.vlm"},
	{0x40B53C3C6B63F4FFULL, "372847738.ogg"},
	{0x40B57A6C5A9A3479ULL, "156306460.ogg"},
	{0x40B595FBB1CC1E2CULL, "mag110_arianni6016836.cl"},
	{0x40B5B42D6F64348BULL, "375183253.ogg"},
	{0x40B60D2A8FB99A9EULL, "vxlt_undercitydungeon_1ue_0_0.vlm"},
	{0x40B61AE106B343FAULL, "vxw_bonepit_yi_0_0.vlm"},
	{0x40B6AD70DB04F9D5ULL, "6080092_m.fxe"},
	{0x40B6B9570009FE9DULL, "hlac_w_bonepit_393220_0z.dds"},
	{0x40B7099A826AB260ULL, "vxlt_gallowstemplar_f_53i_0_0.vlm"},
	{0x40B74D0D37F0DE9BULL, "301244641.ogg"},
	{0x40B75E58F7D9690AULL, "339350122.ogg"},
	{0x40B7700556141DAAULL, "vxca_caverns1_mines_inu_0_0.vlm"},
	{0x40B7706AEFC27CDBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_enter.ani"},
	{0x40B782F5C5CCB353ULL, "vxht_0_hightown_a3_9b4_0_0.vlm"},
	{0x40B7ABA137829626ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be03a.phy"},
	{0x40B7C975068835EBULL, "6146925_m.fxe"},
	{0x40B7D0A98E103617ULL, "vxht_0_hightown_n_9b2_0_0.vlm"},
	{0x40B7F0BF4C108880ULL, "507926690.ogg"},
	{0x40B81362EC425D86ULL, "qun221_petrice6076748.cl"},
	{0x40B83D3BF35C6472ULL, "vxlt_backalley_g_1kv_0_0.vlm"},
	{0x40B8815B97A5C6BCULL, "134230287.ogg"},
	{0x40B8BB0642B1344DULL, "89821227.ogg"},
	{0x40B8EC6EAAA1C5E7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masb_0s.dds"},
	{0x40B99AA523488804ULL, "hlac_w_sundermount_-262142_0y.dds"},
	{0x40B9B2DFBD3FA89DULL, "6208673_f.fxe"},
	{0x40B9F8C26EA76E5CULL, "247834640.ogg"},
	{0x40BA174A5D4B4609ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic_3.mao"},
	{0x40BA322F32C0CB53ULL, "vxlt_0_lowtown_docks_ba7_0_1.vlm"},
	{0x40BA704BB1E95AB9ULL, "vxw_encounter2_if_0_0.vlm"},
	{0x40BB1E487486E125ULL, "vxca_caverns1_eic_0_1.vlm"},
	{0x40BB2C5B70A8896BULL, "dae210_gate_guard6209975.cl"},
	{0x40BB2E52948C8B46ULL, "zz_vrc_debug6200161.cl"},
	{0x40BB3B86A1ECC3CCULL, "vxlt_0_lowtown_a3_n_c_h27_0_0.vlm"},
	{0x40BB55CC5D197F29ULL, "vxca_cave_1lu_0_0.vlm"},
	{0x40BB9364F9DFFFF8ULL, "art\\characters\\creatures\\mubari\\animation\\prst.evt"},
	{0x40BBBECB869CD8C1ULL, "6076934_m.fxe"},
	{0x40BC3C14ECB1E698ULL, "6129037_m.fxe"},
	{0x40BCE7BB8703E006ULL, "113451276.ogg"},
	{0x40BD041E47D6443BULL, "vxlt_hangedman_2dn_0_0.vlm"},
	{0x40BD1093C273F574ULL, "w_encounter2_camp_37_0_lc.mmh"},
	{0x40BD162F75D0207DULL, "6070910_m.fxe"},
	{0x40BD770464DBA72EULL, "6116136_m.fxe"},
	{0x40BD9354526F00D4ULL, "vxw_sundermount_2pk_0_0.vlm"},
	{0x40BD94745D6EAD1DULL, "443211851.ogg"},
	{0x40BE4FBDCD3F1933ULL, "54995739.ogg"},
	{0x40BFAFB19805903CULL, "6105893_m.fxe"},
	{0x40C03DF5C111D89BULL, "6125805_m.fxe"},
	{0x40C042E233A3C20CULL, "dre000_dougal6130965.cl"},
	{0x40C05008A022E4D5ULL, "vxlt_undercityencounter_1b2_0_0.vlm"},
	{0x40C0AF595E1AA560ULL, "vxht_bartrand_n_3sk_0_0.vlm"},
	{0x40C0B3382A1839D6ULL, "188462241.ogg"},
	{0x40C0C26658055AFDULL, "vxdr_0_deeproads_cgq_0_0.vlm"},
	{0x40C0C6B28B5AAA33ULL, "al_warrior_attributes.gda"},
	{0x40C15C7BAF54CC8AULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.cess.evt"},
	{0x40C188F5B687A3AEULL, "vxlt_foundry_n_yu_0_0.vlm"},
	{0x40C198BC7A7493D1ULL, "zz_end_lineup.cut"},
	{0x40C1FF4B4DDB4F41ULL, "vxlt_gallowstemplar_f_34g_0_0.vlm"},
	{0x40C20B7D017F482DULL, "6032273_m.fxe"},
	{0x40C24E07B3C2744CULL, "6199324_m.fxe"},
	{0x40C2CCED936DEF8BULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cratequest_0.pwk"},
	{0x40C32E7D1E098B44ULL, "hlac_lt_0_lowtown_a2_n_c_65535_6x.dds"},
	{0x40C3591C8089A0E0ULL, "vxlt_gallowstemplar_n_5rb_0_0.vlm"},
	{0x40C399E682564B3BULL, "725180071.ogg"},
	{0x40C3A1A11B6675D6ULL, "vxht_keep_n_6bh_0_0.vlm"},
	{0x40C3F600A91FB9D4ULL, "6165932_m.fxe"},
	{0x40C44FC9B1E565E4ULL, "vxlt_0_lowtown_n_h28_0_0.vlm"},
	{0x40C458901BD102B7ULL, "vxw_bonepitmining_a2_27d_0_0.vlm"},
	{0x40C48CBDF8A95C7AULL, "vxw_bonepit_dj_0_0.vlm"},
	{0x40C4C7080F23E9E8ULL, "llac_w_sundermount_65530_0.dds"},
	{0x40C56A55DD3E2DA0ULL, "vxht_mansion_n_6jr_0_0.vlm"},
	{0x40C579DA09903F2FULL, "atl_partypicke_dxt5_02.dds"},
	{0x40C582624B205E8AULL, "mrl000_marethari6142413.cl"},
	{0x40C5E27969A35D42ULL, "6082981_m.fxe"},
	{0x40C60B5A34917821ULL, "seb221_flora_harimann6177728.cl"},
	{0x40C6513AA0345FD5ULL, "lt_0_lowtown_a2_water_21475.msh"},
	{0x40C6533AA034634BULL, "lt_0_lowtown_a2_water_21475.mmh"},
	{0x40C65F3AA03477E8ULL, "lt_0_lowtown_a2_water_21475.mao"},
	{0x40C6D739A132EFE0ULL, "84315830.ogg"},
	{0x40C6F7C47C70E338ULL, "co_dot_01.mao"},
	{0x40C76BFAF035C053ULL, "6099184_m.fxe"},
	{0x40C7B78D7499A02FULL, "6002525_f.fxe"},
	{0x40C804C7AF39AE1EULL, "vxca_caverns2prc_a00_0_0.vlm"},
	{0x40C83B7BAF5AB3FDULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.cess.gad"},
	{0x40C85551ABA6B03FULL, "mag291_torpor6026955.cl"},
	{0x40C8573094C04289ULL, "dae000_gamlen6145109.cl"},
	{0x40C85F8EA2E47867ULL, "6018587_m.fxe"},
	{0x40C89018C7484762ULL, "107592457.ogg"},
	{0x40C8A51FDAA89C50ULL, "vxdr_0_deeproads_f1o_0_0.vlm"},
	{0x40C8BBD1A4CD3365ULL, "hlac_lt_0_lowtown_a2_-65535_6y.dds"},
	{0x40C907093359AFAAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candleon_d.dds"},
	{0x40C95652F33BD17EULL, "dae000_bartrand6030633.cl"},
	{0x40C96389E6D8BF36ULL, "vxlt_undercitydungeon_1n8_0_0.vlm"},
	{0x40C992D502C8896BULL, "mcr370_meredith6083083.cl"},
	{0x40CA1BA89B263919ULL, "gen00fl_merrill6141209.cl"},
	{0x40CA3DC6C1C54508ULL, "901654502.ogg"},
	{0x40CA3F0D7FAA4200ULL, "6119396_f.fxe"},
	{0x40CAD9CAC4AFE752ULL, "6000792_m.fxe"},
	{0x40CB5B4A4377E43BULL, "hlac_lt_undercityserialkiller_-1_3x.dds"},
	{0x40CB9528F66327F7ULL, "projdecal_projdecal_p0.psh.fxo.dx11.assembly"},
	{0x40CBF768748BB6CAULL, "vxw_woundedcoast_l3_0_0.vlm"},
	{0x40CBFD968C09A2B4ULL, "vxw_woundedcoast_1sa_0_0.vlm"},
	{0x40CC08539A9B9B71ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutd.mmh"},
	{0x40CC14539A9BAF12ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutd.mao"},
	{0x40CC22539A9BC72FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutd.msh"},
	{0x40CC322AC577D9E9ULL, "6027654_f.fxe"},
	{0x40CC66216AD38E6DULL, "nat380_sibling6068026.cl"},
	{0x40CC976AEFD51B65ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_enter.gad"},
	{0x40CC994BF6E7FD2EULL, "6067385_f.fxe"},
	{0x40CC9E8B177B052AULL, "6082196_m.fxe"},
	{0x40CCB8FA74B015D5ULL, "6128549_m.fxe"},
	{0x40CCD6D7A31879FAULL, "w_bonepitmining_a2_11_0.mmh"},
	{0x40CCDAD7A3188031ULL, "w_bonepitmining_a2_11_0.mao"},
	{0x40CCE8D7A3189840ULL, "w_bonepitmining_a2_11_0.msh"},
	{0x40CD1E2FAA10AA23ULL, "mer210cr_smith.utc"},
	{0x40CD51C9E57490DCULL, "514400019.ogg"},
	{0x40CDDA3F0533539EULL, "6036858_m.fxe"},
	{0x40CE932476CFC3B8ULL, "6097901_m.fxe"},
	{0x40CEAFD2B307F11BULL, "6122905_m.fxe"},
	{0x40CEDAE855D9923BULL, "hlac_ca_caverns2prc_5802_5x.dds"},
	{0x40CF2E96FA0DB4B6ULL, "335618411.ogg"},
	{0x40D030AF42DF64C6ULL, "6208031_m.fxe"},
	{0x40D12A51601C0306ULL, "6147149_m.fxe"},
	{0x40D155D8DBD89BA6ULL, "qcr000_arishok6057112.cl"},
	{0x40D1A5085887E559ULL, "dae000_hate_follower_react6103205.cl"},
	{0x40D1B5DA39ED19FBULL, "6102460_m.fxe"},
	{0x40D1FAA6DD3A4E5AULL, "w_bonepit_183_0.mmh"},
	{0x40D1FEA6DD3A5491ULL, "w_bonepit_183_0.mao"},
	{0x40D20CA6DD3A6CA0ULL, "w_bonepit_183_0.msh"},
	{0x40D20F56C69F023DULL, "art\\levels\\4cinematics\\ferelden\\prp_crown.mao"},
	{0x40D21356C69F0976ULL, "art\\levels\\4cinematics\\ferelden\\prp_crown.mmh"},
	{0x40D21D56C69F1A74ULL, "art\\levels\\4cinematics\\ferelden\\prp_crown.msh"},
	{0x40D27677260F3EBFULL, "zz_rdr_debug6206279.cl"},
	{0x40D2808E6F16C300ULL, "pro000cr_soldier6.utc"},
	{0x40D2FE9AEA9F74FAULL, "vxw_bonepit_ok_0_0.vlm"},
	{0x40D329396F159040ULL, "vxca_caverns1_cn5_0_1.vlm"},
	{0x40D35395B7BB94B6ULL, "zz_dre_debug6168283.cl"},
	{0x40D373B2CA24B023ULL, "mcr371cr_alain.utc"},
	{0x40D3DC32CF344620ULL, "vxht_mansion_n_6ms_0_0.vlm"},
	{0x40D40925B6479B4AULL, "vxht_0_hightown_a2_c5c_0_0.vlm"},
	{0x40D44D38C67ADE74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\flurry.ani"},
	{0x40D4B47B9369E9F5ULL, "vxlt_0_lowtown_docks_a3_buw_0_0.vlm"},
	{0x40D4C8C523416839ULL, "294340904.ogg"},
	{0x40D4F821333F887DULL, "vxw_sundermount_2a8_0_1.vlm"},
	{0x40D5055CB8B59AB4ULL, "242962832.ogg"},
	{0x40D528B84FACD1ABULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avel.phy"},
	{0x40D592D7C30EC4A7ULL, "6072158_m.fxe"},
	{0x40D59995E205D935ULL, "hlac_lt_0_lowtown_n_0_3z.dds"},
	{0x40D5DEB42878ADE1ULL, "dae000_madam_lusine6151398.cl"},
	{0x40D5F715DCBE0D51ULL, "vxlt_blackemporium_tn_0_0.vlm"},
	{0x40D614113782E3B9ULL, "art\\vfx\\combat\\vfx_gib_qunc_rtl.anb"},
	{0x40D708CEA62AFCE5ULL, "lt_0_lowtown_docks_a3_120_0.phy"},
	{0x40D716FCCC9EB4D0ULL, "vxdr_0_deeproads_f4l_0_0.vlm"},
	{0x40D71E3577AD3173ULL, "6126642_m.fxe"},
	{0x40D772138705247EULL, "6165325_m.fxe"},
	{0x40D7D6713903DA5CULL, "hlac_ht_0_hightown_a3_0_3y.dds"},
	{0x40D7EAE336D997C7ULL, "playvfx.ncs"},
	{0x40D7EB424189CBD5ULL, "vrc321st_spooky_scene_8.stg"},
	{0x40D8230565601E9DULL, "one100dg_nabil6033163.cl"},
	{0x40D85C7351056154ULL, "dae000_fem_commoner1_mob6174486.cl"},
	{0x40D863E99CB32C99ULL, "qcr220_orsino6041825.cl"},
	{0x40D88C518FB8272AULL, "vxlt_0_lowtown_a2_n_e1z_0_0.vlm"},
	{0x40D8D65077201B42ULL, "467876028.ogg"},
	{0x40D926A9C2445645ULL, "144981817.ogg"},
	{0x40D986F970D93ADAULL, "fake_im_arm_hlm_lgt_champ.uti"},
	{0x40D98E9D8CAF32C8ULL, "mcr310_argument6075121.cl"},
	{0x40D9BEBA92DD8FECULL, "hlac_lt_0_lowtown_n_-65536_5y.dds"},
	{0x40DA09D4484D8B26ULL, "260262787.ogg"},
	{0x40DA4A38C67F44D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\flurry.gad"},
	{0x40DA4FD4321A5322ULL, "150872935.ogg"},
	{0x40DAA7710263EC6CULL, "vxht_mansion2b_3dj_0_0.vlm"},
	{0x40DABC7674C57AB2ULL, "dae000_friend_turn_in6179075.cl"},
	{0x40DAEF18CF895363ULL, "648264931.ogg"},
	{0x40DB629D3132AA63ULL, "6081202_m.fxe"},
	{0x40DB8334F17A977AULL, "vxca_caverns2prc_9t0_0_0.vlm"},
	{0x40DBD4FC7EA6AAF1ULL, "art\\characters\\creatures\\stonegolem\\animation\\blendtree\\stonegolem_awakened.anb"},
	{0x40DC17A4C9F23607ULL, "776683581.ogg"},
	{0x40DC3E28D6CA3BCCULL, "zz_one_debug6077962.cl"},
	{0x40DC49C6052ABB8AULL, "6169888_m.fxe"},
	{0x40DC8424723D8B8EULL, "6062118_m.fxe"},
	{0x40DCFAF3996E3143ULL, "vxw_bonepitdestroyed_a3_23z_0_0.vlm"},
	{0x40DD33C2840F15D5ULL, "vxlt_undercityencounter_1t0_0_0.vlm"},
	{0x40DD36521BBAB33FULL, "ef_shadow.mor"},
	{0x40DD392B379409E0ULL, "zz_isa_debug6133153.cl"},
	{0x40DD6F53B2ED25BCULL, "vxht_keep_5xt_0_0.vlm"},
	{0x40DE06B8A4F3CFE1ULL, "vxdr_primevaltomb_6ux_0_0.vlm"},
	{0x40DE454556499EBCULL, "hlac_lt_undercityserialkiller_-1_43y.dds"},
	{0x40DE83E9A7A0EC74ULL, "6087805_m.fxe"},
	{0x40DF43549688D43BULL, "fen121_cut_fenris_finds6098872.cl"},
	{0x40DF75480C21DD3EULL, "vxht_0_hightown_a3_c0i_0_0.vlm"},
	{0x40DFB7E24C347078ULL, "6169987_m.fxe"},
	{0x40E02AD65FCD9A2DULL, "vxht_0_hightown_c18_0_0.vlm"},
	{0x40E02CCC577AC525ULL, "vxht_brothel_n_4fz_0_0.vlm"},
	{0x40E031156ABC5BEEULL, "vxca_caverns2prc_7zd_0_0.vlm"},
	{0x40E06A35EBFC4264ULL, "llac_ht_keep_3731_1.dds"},
	{0x40E0DA346B9E1825ULL, "6064560_m.fxe"},
	{0x40E112850B27D345ULL, "vxw_bonepit_yb_0_0.vlm"},
	{0x40E13C49FFBD33B7ULL, "vxw_bonepitmining_a2_29f_0_0.vlm"},
	{0x40E16538C6859135ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\flurry.evt"},
	{0x40E17E11029E1E93ULL, "dae200_bran6144572.cl"},
	{0x40E186E68E0F42B0ULL, "6153515_m.fxe"},
	{0x40E1964EF473336DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masc_0d.dds"},
	{0x40E21F22AD58815CULL, "art\\characters\\playercharacter\\qunarimale\\eye_ars.mao"},
	{0x40E2FA59787E51DEULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_conj_c.impa.ani"},
	{0x40E31E5445A1077FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_loop.gad"},
	{0x40E3472B2ABFCE7AULL, "fen110_fenris6096422.cl"},
	{0x40E384D66BEF73C2ULL, "462817279.ogg"},
	{0x40E3AB8B30C88515ULL, "art\\vfx\\combat\\vfx_w_wns_front_long1_c.mmh"},
	{0x40E3B7527B4C3439ULL, "296063733.ogg"},
	{0x40E443DC48594955ULL, "hlac_lt_0_lowtown_a2_-131072_2y.dds"},
	{0x40E469D59B03EE1EULL, "mrl250cr_thief_d.utc"},
	{0x40E4962B39A35F25ULL, "6139077_m.fxe"},
	{0x40E4F35D0C3253EAULL, "599318078.ogg"},
	{0x40E4FDD0E74C1204ULL, "6217554_m.fxe"},
	{0x40E54B7A0ABD8ADAULL, "vxw_woundedcoast_1ee_0_0.vlm"},
	{0x40E5DB19E20F0524ULL, "6103363_m.fxe"},
	{0x40E5E1C1737A227EULL, "61930911.ogg"},
	{0x40E6153CBE7DFBECULL, "gen_im_acc_blt_act2_05.uti"},
	{0x40E658363CD4FBFBULL, "6095081_m.fxe"},
	{0x40E66586543BD685ULL, "seb221_flora_harimann6177676.cl"},
	{0x40E670C3B90A2CABULL, "6109061_m.fxe"},
	{0x40E6A56B0506D504ULL, "docks_male_elven_worker_sta6175798.cl"},
	{0x40E6DEAFB18B1E79ULL, "vxht_brothel_n_3f5_0_0.vlm"},
	{0x40E7D67A6A9C6A5BULL, "hlac_w_blightlands_-1_0x.dds"},
	{0x40E7D8A9102D3ABAULL, "822889298.ogg"},
	{0x40E8770A0B48D6FDULL, "vxht_chantry_2uq_0_0.vlm"},
	{0x40E8A37404627FD6ULL, "hlac_ca_cavern1_haunted_24425_8y.dds"},
	{0x40E8ACB63AFD747AULL, "vxlt_0_lowtown_n_f4n_0_0.vlm"},
	{0x40E8D9583F4AE1ADULL, "vxlt_gallowstemplar_483_0_0.vlm"},
	{0x40E97D05DAD42E2EULL, "6069185_f.fxe"},
	{0x40E98EAFCAA30DB5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowflat.phy"},
	{0x40E98EEEDB62923FULL, "vxlt_gallowstemplar_65x_0_0.vlm"},
	{0x40E9F58E1A329BE9ULL, "246436347.ogg"},
	{0x40EA114A62B686DAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg2a.mmh"},
	{0x40EA234A62B6A520ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg2a.msh"},
	{0x40EA3EC2D033F62BULL, "mag121ip_idunna_clue01.utp"},
	{0x40EA7C4C843FE448ULL, "vxdr_primevaltomb_545_0_0.vlm"},
	{0x40EA898C08A6E3DBULL, "24088708.ogg"},
	{0x40EB119D75AB530AULL, "803842555.ogg"},
	{0x40EB4101CEA4A3D8ULL, "vxca_cavern1_haunted_i6o_0_0.vlm"},
	{0x40EB7C21761EC1BCULL, "6148212_m.fxe"},
	{0x40EB7E1A28D23C03ULL, "zz_seb_debug6183095.cl"},
	{0x40EBB724E7EF17A2ULL, "6035932_m.fxe"},
	{0x40EBDB14BC2CA95FULL, "vxht_mansion_n_6ry_0_0.vlm"},
	{0x40EC1F417AB0E9F5ULL, "300872670.ogg"},
	{0x40EC42CC4411D8C6ULL, "6149049_f.fxe"},
	{0x40ED677571B5C57BULL, "art\\characters\\playercharacter\\humanmale\\boo_lgte.msh"},
	{0x40ED797571B5E3C6ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgte.mao"},
	{0x40ED857571B5F865ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgte.mmh"},
	{0x40EDC6968AC8C491ULL, "vxlt_warehouse_n_e8_0_0.vlm"},
	{0x40EDD530F5A99236ULL, "mrl150_merrill6070495.cl"},
	{0x40EDE724FE17F5BEULL, "vxht_0_hightown_a3_c5j_0_0.vlm"},
	{0x40EE9CB351C3B2ADULL, "vxht_0_hightown_c47_0_0.vlm"},
	{0x40EEAEB3566BEA60ULL, "vxlt_gallowstemplar_n_5ca_0_0.vlm"},
	{0x40EEC8F4A747A595ULL, "vxlt_darktown_50n_0_0.vlm"},
	{0x40EFB14837B9FCC7ULL, "vxw_encounter1_bt_0_0.vlm"},
	{0x40F01259788904AFULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_conj_c.impa.evt"},
	{0x40F0CD1AFD475E81ULL, "vxht_brothel_n_4vf_0_0.vlm"},
	{0x40F0F08D101488A5ULL, "131232641.ogg"},
	{0x40F0F472736FEB0AULL, "ave310tr_donnic_fight.ncs"},
	{0x40F1042C28A863AFULL, "6133565_m.fxe"},
	{0x40F1542B9D60138DULL, "follower_banter6171863.cl"},
	{0x40F196957ABBBC37ULL, "qun221_viscount6084712.cl"},
	{0x40F1A88B30D4BDA2ULL, "art\\vfx\\combat\\vfx_w_wns_front_long1_c.anb"},
	{0x40F23FC726418622ULL, "zz_one_debug6075016.cl"},
	{0x40F2765056800F69ULL, "6083745_cleric_youstand.cut"},
	{0x40F33150EA52C632ULL, "vxlt_shop_cc_0_0.vlm"},
	{0x40F33CC300BA5FFEULL, "546678053.ogg"},
	{0x40F3858AD531C4C1ULL, "vxlt_0_lowtown_docks_n_a0k_0_0.vlm"},
	{0x40F389ECD7FE64BEULL, "dwarf_female_heads.gda"},
	{0x40F3C564D9992440ULL, "6051579_f.fxe"},
	{0x40F3C705DF2FE1CAULL, "one280_merin6105668.cl"},
	{0x40F3DF818B2176FBULL, "6171879_m.fxe"},
	{0x40F3E6EFF5794B04ULL, "hlac_ht_0_hightown_n_131071_1z.dds"},
	{0x40F41D9CD558FF07ULL, "252819455.ogg"},
	{0x40F42C2AA5B4B5FAULL, "hlac_lt_0_lowtown_docks_n_-131074_4z.dds"},
	{0x40F4347E4903CBADULL, "app_wispy_f.mao"},
	{0x40F43521E7DC8829ULL, "w_bonepitmining_a2_28_0_ll.mmh"},
	{0x40F43D192E73F901ULL, "vxht_keep_n_5xz_0_0.vlm"},
	{0x40F6014B583E9448ULL, "vxdr_primevaltomb_6xz_0_0.vlm"},
	{0x40F621BF1974E42BULL, "6087171_f.fxe"},
	{0x40F6681213D578D6ULL, "gen00fl_sebastian6179796.cl"},
	{0x40F6694678914455ULL, "gen00fl_anders6125142.cl"},
	{0x40F6AFBB3F8A6A5DULL, "6165685_m.fxe"},
	{0x40F6D0B8DBF34BBEULL, "6121196_m.fxe"},
	{0x40F6FB5445B174C5ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_loop.ani"},
	{0x40F71359788F2558ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_conj_c.impa.gad"},
	{0x40F77A9394A162FDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appp_0n.dds"},
	{0x40F786795E4ADA4CULL, "art\\characters\\weapons\\dagger\\w_dgr_005a.phy"},
	{0x40F7A4979C22C1EDULL, "vxht_0_hightown_n_a7g_0_0.vlm"},
	{0x40F7C258FE6AA0A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_right.ani"},
	{0x40F7D27A3EE07DCCULL, "vxw_encounter2_6f_0_0.vlm"},
	{0x40F819A6DD5B5265ULL, "w_bonepit_183_0.phy"},
	{0x40F8742E549053A9ULL, "6139419_m.fxe"},
	{0x40F8902CE2CD91CFULL, "98155050.ogg"},
	{0x40F8D6D7A33DB411ULL, "w_bonepitmining_a2_11_0.rml"},
	{0x40F9EDFE6814DA3CULL, "6145112_m.fxe"},
	{0x40FAD7D9493F64FDULL, "hlac_lt_gallowstemplar_f_2394_11z.dds"},
	{0x40FAD80486EC4CF0ULL, "vxlt_hangedman_n_2u9_0_0.vlm"},
	{0x40FB20D3EB9C303CULL, "art\\characters\\creatures\\mubari\\animation\\rnf.ani"},
	{0x40FBD650B7621590ULL, "vxdr_primevaltomb_7eh_0_0.vlm"},
	{0x40FC160150517C1AULL, "262565778.ogg"},
	{0x40FC2D70652135DEULL, "dae000_worthy6166627.cl"},
	{0x40FC46E1F2060E3AULL, "514699145.ogg"},
	{0x40FC8DA23E85FAF1ULL, "vxca_caverns1_cny_0_0.vlm"},
	{0x40FD276D4CAE9AF3ULL, "747733444.ogg"},
	{0x40FD37657C5C0B66ULL, "gen00fl_aveline6139839.cl"},
	{0x40FDFCC3CE6147BEULL, "28495134.ogg"},
	{0x40FE232529B01547ULL, "vxw_encounter1_gw_0_0.vlm"},
	{0x40FE931C8DFC5855ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.mmh"},
	{0x40FEC5C86596B600ULL, "gr_sp_fl02.mao"},
	{0x40FEDD58FE70ED2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_right.gad"},
	{0x40FF5F85ADE368C5ULL, "dae000_bartrand6030658.cl"},
	{0x40FF7DD03D1D1EE2ULL, "6104441_m.fxe"},
	{0x40FFC8095125FA82ULL, "923264734.ogg"},
	{0x40FFFCA718525C6FULL, "1028340131.ogg"},
	{0x410058B57F969665ULL, "fxa_caltrop.mao"},
	{0x4100E7DD434424B6ULL, "vxlt_ruinedhouse_n_jd_0_0.vlm"},
	{0x4100E99466A49E3DULL, "6115408_m.fxe"},
	{0x4100EE5B0486D44EULL, "939075146.ogg"},
	{0x4102158CEC3067BEULL, "vxdr_0_deeproads_dqo_0_0.vlm"},
	{0x41022DD3EBA2657EULL, "art\\characters\\creatures\\mubari\\animation\\rnf.gad"},
	{0x41022F33EEA9BBDAULL, "vxw_woundedcoast_1kg_0_0.vlm"},
	{0x410276EA1C8118A4ULL, "6072801_m.fxe"},
	{0x4102D961AA17C9E3ULL, "gen00fl_anders6113130.cl"},
	{0x41030F8318BA6DE1ULL, "vxw_bonepit_z3_0_0.vlm"},
	{0x410331E8FA6B6648ULL, "504224913.ogg"},
	{0x4104616404338482ULL, "vxw_bonepitmining_a2_1zy_0_0.vlm"},
	{0x41047FAF5A28654FULL, "one100_ht_banter6060907.cl"},
	{0x4104E096E604C9FEULL, "6172817_m.fxe"},
	{0x4105327D4D7BEB45ULL, "vxlt_0_lowtown_n_f1f_0_0.vlm"},
	{0x41055B886684168AULL, "lt_0_lowtown_docks_n_196_0.mao"},
	{0x41055F1506D80EBFULL, "vxht_mansion2b_3de_0_0.vlm"},
	{0x41055F8866841DD9ULL, "lt_0_lowtown_docks_n_196_0.mmh"},
	{0x4105698866842EC7ULL, "lt_0_lowtown_docks_n_196_0.msh"},
	{0x4106445730D6964CULL, "vxw_encounter2_1g_0_0.vlm"},
	{0x4106C65D1FBA7156ULL, "vxlt_backalley_n_1wn_0_1.vlm"},
	{0x41082D8B83F92696ULL, "vxht_0_hightown_bc0_0_0.vlm"},
	{0x410833C7AC4A2763ULL, "6108651_m.fxe"},
	{0x410856B7703CBFC2ULL, "6043991_m.fxe"},
	{0x41085A50EFE7E944ULL, "gen00fl_isabela6135652.cl"},
	{0x4108A9AD8DB55D30ULL, "6087045_m.fxe"},
	{0x4108D6D5B60B9483ULL, "vxlt_0_lowtown_hb0_0_0.vlm"},
	{0x4108DC0811061E4AULL, "vxw_bonepit_1p5_0_0.vlm"},
	{0x4108EC37EE90183BULL, "vxw_encounter2_camp_r1_0_0.vlm"},
	{0x410908D3EBA8449DULL, "art\\characters\\creatures\\mubari\\animation\\rnf.evt"},
	{0x41091F180FC4B241ULL, "248372274.ogg"},
	{0x41093673DF3A7259ULL, "vxlt_gallowstemplar_n_31x_0_0.vlm"},
	{0x41095B00BAA99A38ULL, "vxlt_gallowstemplar_4so_0_0.vlm"},
	{0x410967137F07C540ULL, "6098290_m.fxe"},
	{0x410A0BE3B10090EDULL, "vxht_0_hightown_a2_be5_0_0.vlm"},
	{0x410A25030B2BD2AEULL, "520266308.ogg"},
	{0x410A30763BB9BF7CULL, "vxlt_0_lowtown_ewm_0_0.vlm"},
	{0x410AFE4331987676ULL, "vxdr_primevalentrance_1oz_0_0.vlm"},
	{0x410B573A1677A78FULL, "707863497.ogg"},
	{0x410B67355C32CB62ULL, "follower_banter6171088.cl"},
	{0x410B901C8E06DDE2ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.anb"},
	{0x410BD021DA6A5515ULL, "hlac_lt_0_lowtown_docks_a3_-131073_5y.dds"},
	{0x410C33ADC72A4AEAULL, "6135342_m.fxe"},
	{0x410C9838DA0F19C8ULL, "vxw_bonepit_z6_0_0.vlm"},
	{0x410CD6EEED1BA0ACULL, "6166991_m.fxe"},
	{0x410D113169C86FFCULL, "6016061_m.fxe"},
	{0x410D122A4AF8621FULL, "386534086.ogg"},
	{0x410D36943301D39EULL, "art\\characters\\creatures\\dragon\\animation\\cwbl.evt"},
	{0x410D36DB4F82BF0CULL, "6040556_m.fxe"},
	{0x410D3E4FC569FFFBULL, "hlac_lt_0_lowtown_docks_a3_n_-3_0y.dds"},
	{0x410D5844EA860972ULL, "vxlt_undercityserialkiller_25w_0_0.vlm"},
	{0x410EC47DC9706744ULL, "vxca_cavern1_haunted_i0i_0_0.vlm"},
	{0x410EE5963D3C15F7ULL, "301972191.ogg"},
	{0x410F018DF7893598ULL, "pow_in_cyc.mao"},
	{0x410F53A1120F0F09ULL, "rdr110dg_party_comments6094678.cl"},
	{0x410FA289FCF7191BULL, "w_encounter3_253_0.mmh"},
	{0x410FAE89FCF72DB8ULL, "w_encounter3_253_0.mao"},
	{0x410FC089FCF74C05ULL, "w_encounter3_253_0.msh"},
	{0x410FD00814E7BFCCULL, "one200dg_solivitus6048178.cl"},
	{0x410FD99363482C14ULL, "mag291_vincento6027375.cl"},
	{0x4110304A62D78AE5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg2a.phy"},
	{0x411072891630C966ULL, "hf_desp_refugee.mor"},
	{0x4110AEC36A233993ULL, "570721630.ogg"},
	{0x4110FB5DD147601EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.evt"},
	{0x4111531158F68486ULL, "586835342.ogg"},
	{0x41116C653E8B209DULL, "6047462_m.fxe"},
	{0x41116ED587F22239ULL, "vxw_blightlands_7c_0_0.vlm"},
	{0x4111D80AAA01806AULL, "vxlt_0_lowtown_h4w_0_0.vlm"},
	{0x41123BD943BB08EFULL, "hlac_w_encounter3_-327684_0y.dds"},
	{0x41126DA256AC0B1BULL, "xx_floor.dds"},
	{0x4112B48751C57F0DULL, "218529503.ogg"},
	{0x4112D340F6299D02ULL, "vxw_bonepitmining_a2_1wz_0_0.vlm"},
	{0x4112D7349A2F201EULL, "102530049.ogg"},
	{0x41134D94330666C9ULL, "art\\characters\\creatures\\dragon\\animation\\cwbl.gad"},
	{0x41134F853254AF7DULL, "zz_rogue5.cut"},
	{0x4113A45A3F7203C5ULL, "vxlt_0_lowtown_n_f4e_0_0.vlm"},
	{0x4113C6D668296783ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkf.ani"},
	{0x4114205C6ECBF5E6ULL, "vxca_cavern1_haunted_hhz_0_0.vlm"},
	{0x4114C46489E70C81ULL, "art\\levels\\races\\proxy\\props\\door\\door_2.mao"},
	{0x4114E939265F0331ULL, "207605676.ogg"},
	{0x411521EBF1741648ULL, "narrative_varric6172556.cl"},
	{0x4115680530109582ULL, "6074363_f.fxe"},
	{0x41169F6875EF3F16ULL, "vxht_0_hightown_bl1_0_0.vlm"},
	{0x4116A8FD25DFCE09ULL, "dae000_bodhan_letter6144314.cl"},
	{0x4116FB0A2A8476ABULL, "dae000_merchant_access_hf6162948.cl"},
	{0x4117125DD14BF349ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.gad"},
	{0x41174DE502FC36CAULL, "vxw_bonepit_1u6_0_0.vlm"},
	{0x41175E14E08630BBULL, "vxw_encounter2_camp_m0_0_0.vlm"},
	{0x41181DBA9379E671ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultdoor.mmh"},
	{0x411829BA9379FA12ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultdoor.mao"},
	{0x411837BA937A122FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_vaultdoor.msh"},
	{0x41189A4F69DCCEEDULL, "l_qun111ar_safehouse.bnk"},
	{0x4118CAB56C956081ULL, "vxlt_0_lowtown_docks_c4r_0_0.vlm"},
	{0x4119A6FE4152F093ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_tarp_new_d.dds"},
	{0x411A1E94330C3517ULL, "art\\characters\\creatures\\dragon\\animation\\cwbl.ani"},
	{0x411A87EF7895C829ULL, "vxht_chantry_n_2e2_0_0.vlm"},
	{0x411AC33FE2EC2ECFULL, "dae111ar_tavern_nt.are"},
	{0x411AFADB6499E3C6ULL, "495000707.ogg"},
	{0x411B0A15CC053248ULL, "vxw_bonepit_e7_0_0.vlm"},
	{0x411B2648AA5087BFULL, "vxlt_gallowstemplar_f_31v_0_0.vlm"},
	{0x411B564E97E72669ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_basic_1.mao"},
	{0x411BA9300FCB81C7ULL, "vxdr_primevaltomb_9co_0_0.vlm"},
	{0x411D18D8E1C791B6ULL, "6112132_m.fxe"},
	{0x411D7E56849A5A00ULL, "6195823_m.fxe"},
	{0x411D8E2DC541786BULL, "48139327.ogg"},
	{0x411D9FB3DBAE5E6BULL, "mag131_keran6086173.cl"},
	{0x411DE35DD151C197ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.ani"},
	{0x411E54DEAFAA7D0BULL, "vrc321ip_projectile_target.ncs"},
	{0x411EBC30A5EE3B19ULL, "hlac_dr_dungeon_3214_12z.dds"},
	{0x411EC5F1962FF551ULL, "6035193_m.fxe"},
	{0x411F12EDD295ECDAULL, "vxw_woundedcoast_1qa_0_0.vlm"},
	{0x411F521B93836967ULL, "llac_lt_0_lowtown_a2_n_c_-131072_6.dds"},
	{0x411FA24272247B18ULL, "6185479_m.fxe"},
	{0x411FB4319F5AB8B9ULL, "6034806_m.fxe"},
	{0x411FE0B279E83AB9ULL, "vxw_blightlands_8b_0_0.vlm"},
	{0x412053A4BA21CA78ULL, "vxlt_gallowscourtyard_n_4va_0_0.vlm"},
	{0x412060691F202B59ULL, "335694290.ogg"},
	{0x4120920D94EF6804ULL, "vxht_bartrand_n_4er_0_0.vlm"},
	{0x4120DED668341B7AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkf.evt"},
	{0x4120E39E1F6819D9ULL, "6094710_m.fxe"},
	{0x4121EDD9E2783417ULL, "vxht_mansion_n_60k_0_0.vlm"},
	{0x41226115CACAF8A6ULL, "6118013_m.fxe"},
	{0x412301A12597C8CFULL, "hlac_lt_0_lowtown_a2_0_4z.dds"},
	{0x41231231D0757843ULL, "vxdr_primevaltomb_90z_0_0.vlm"},
	{0x412325519414B3F9ULL, "723037571.ogg"},
	{0x412351A4A511E859ULL, "6105856_m.fxe"},
	{0x41239F2C2B9F24D8ULL, "vxw_bonepitdestroyed_a3_1ne_0_0.vlm"},
	{0x41243C4792ABA79AULL, "740452437.ogg"},
	{0x4124437CD3663AC1ULL, "art\\cinematics\\additional animation\\cs_malenobleiv_cinematicanim.gad"},
	{0x4124798828778238ULL, "dae000_pc_home_ambients6208606.cl"},
	{0x4125CFF1D27C493BULL, "vxw_encounter2_camp_h3_0_0.vlm"},
	{0x4125D9899C5D6389ULL, "mag170_thrask6020550.cl"},
	{0x412602D76E29EA78ULL, "hlac_w_blightlands_0_9x.dds"},
	{0x412662C846CD492FULL, "vxht_playershouse_n_1iz_0_0.vlm"},
	{0x41268A88B3BCF61EULL, "6155739_m.fxe"},
	{0x41272BE78D445F10ULL, "vxdr_primevaltomb_7tw_0_0.vlm"},
	{0x41273C925E8B7901ULL, "vxlt_0_lowtown_docks_c1q_0_0.vlm"},
	{0x4127DDD6683A386DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkf.gad"},
	{0x4127E5C36F328F4BULL, "93029879.ogg"},
	{0x4128F3147ED3A971ULL, "vxw_encounter2_cx_0_0.vlm"},
	{0x412935631D937ED0ULL, "zz_followers_banter6159222.cl"},
	{0x412944CA38E0E27BULL, "dae000_fem_commoner2_mob6174590.cl"},
	{0x4129B5C2FCEAB3D1ULL, "character_character_p141.psh.fxo"},
	{0x4129C78D7082C285ULL, "gen00fl_merrill6108259.cl"},
	{0x412A121B0759561FULL, "character_character_p143.psh.fxo"},
	{0x412A4CF1393ADCF3ULL, "qcr000_arishok6060096.cl"},
	{0x412A8DD41FBB395DULL, "lgt310ip_drop_zone_2.ncs"},
	{0x412B447CD36C5B6FULL, "art\\cinematics\\additional animation\\cs_malenobleiv_cinematicanim.ani"},
	{0x412B4B2379816A84ULL, "6097862_m.fxe"},
	{0x412BBC3FE2FA958CULL, "dae111ar_tavern_nt.lst"},
	{0x412BCFB046690B08ULL, "839912552.ogg"},
	{0x412C112A58E86367ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.impa.gad"},
	{0x412CA9426C3C6E05ULL, "llac_lt_0_lowtown_docks_n_-65538_8.dds"},
	{0x412D25A3E1A31A91ULL, "6086241_m.fxe"},
	{0x412DF0EDF56606D3ULL, "gen_im_fla_aveline_01.uti"},
	{0x412E26339DA31825ULL, "one300_nuncio6073775.cl"},
	{0x412E26FD24954C9DULL, "365237351.ogg"},
	{0x412E574B2379F865ULL, "372328208.ogg"},
	{0x412E75F782B0017BULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb107a_0t.dds"},
	{0x412E9C59A86CC49AULL, "vxw_woundedcoast_ho_0_0.vlm"},
	{0x412F00961A5D1E26ULL, "and111_donation_box6200720.cl"},
	{0x412F764A152208FFULL, "vxca_caverns1_e7j_0_0.vlm"},
	{0x412F869EE062E479ULL, "37554061.ogg"},
	{0x412FBDF5717FE1B1ULL, "6041208_m.fxe"},
	{0x412FDB98E1E04271ULL, "vxlt_undercitydungeon_1g8_0_0.vlm"},
	{0x413007C7199401CBULL, "vxlt_0_lowtown_a2_ewz_0_0.vlm"},
	{0x41304D2776067517ULL, "mag291_vincento6027675.cl"},
	{0x4130CB519A471A3BULL, "502083575.ogg"},
	{0x4130E2927B9F9934ULL, "one280ip_cache.utp"},
	{0x4130F0D291D096A3ULL, "6083709_m.fxe"},
	{0x41310F80DEB2FECDULL, "llac_lt_0_lowtown_a2_n_131070_5.dds"},
	{0x41313C6D130AA01AULL, "lt_gallowstemplar_8165.anb"},
	{0x4131A1AFCAE0E3D0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowflat.msh"},
	{0x4131ABC14A3A879EULL, "6136206_m.fxe"},
	{0x4131AFAFCAE0FBEAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowflat.mmh"},
	{0x4131B3AFCAE10221ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0cc_galowflat.mao"},
	{0x4131ED4A30FD144CULL, "vxw_blightlands_lo_0_0.vlm"},
	{0x413201381633C97AULL, "131400241.ogg"},
	{0x413262C6F8078496ULL, "vxht_mansion_n_4yo_0_0.vlm"},
	{0x41326AAEC68363B5ULL, "vxht_keep_5hi_0_0.vlm"},
	{0x4132958FF3F41B0AULL, "81630996.ogg"},
	{0x41329C3FE3007E35ULL, "dae111ar_tavern_nt.ncs"},
	{0x413375C275AB7F7AULL, "6140218_m.fxe"},
	{0x41340DD45369B03AULL, "vxca_caverns1_cmw_0_0.vlm"},
	{0x41342A54118003D3ULL, "6110859_f.fxe"},
	{0x4134829A3D54D5F4ULL, "llac_dr_slavers_3207_6.dds"},
	{0x4134D37C8C9F3FC8ULL, "lt_gallowstemplar_mmap_ph.dds"},
	{0x41355C15A35956DAULL, "dae000_fem_commoner1_mob.cnv"},
	{0x41355E7571F36B12ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgte.phy"},
	{0x41359DC47F3A7790ULL, "vxdr_primevaltomb_7qt_0_0.vlm"},
	{0x4135B77A7B13FA2BULL, "hlac_lt_undercitydungeon_-1_48z.dds"},
	{0x4135FC1526CD264CULL, "plt_pro000pt_cavril_store.plo"},
	{0x41361A6512631199ULL, "da2terrain_terrain_p1.hsh.fxo.dx11"},
	{0x41368EB8B1AD9C44ULL, "vxdr_primevaltomb_8om_0_0.vlm"},
	{0x41369387A4EC0D9CULL, "450844858.ogg"},
	{0x41372A24F198CE2EULL, "vfx_lavajet_lrg.wbd"},
	{0x41375D795E807817ULL, "art\\characters\\weapons\\dagger\\w_dgr_005a.mmh"},
	{0x413761795E807F24ULL, "art\\characters\\weapons\\dagger\\w_dgr_005a.mao"},
	{0x413773795E809D89ULL, "art\\characters\\weapons\\dagger\\w_dgr_005a.msh"},
	{0x4137821DCEF39587ULL, "dae220ip_at_keep.utp"},
	{0x4137901CF1EE71A3ULL, "6146837_m.fxe"},
	{0x41385F4AB3F31C06ULL, "gen00fl_anders6116116.cl"},
	{0x41387E15A35BC622ULL, "dae000_fem_commoner1_mob.bnk"},
	{0x41388ABD4D5C4889ULL, "hla_ht_0_hightown_a3_131073_1x.dds"},
	{0x4138FA89474560E8ULL, "sse_ss_wilmod.gda"},
	{0x4139AC045ED7CF39ULL, "vxht_brothel_n_6tz_0_0.vlm"},
	{0x4139D4A5710D3213ULL, "vxw_sundermount_301_0_0.vlm"},
	{0x413A3A9337EF73EAULL, "961318826.ogg"},
	{0x413AB6D5969FBA2BULL, "750527249.ogg"},
	{0x413B1671454114BCULL, "vxht_brothel_n_5fz_0_0.vlm"},
	{0x413B391E09159857ULL, "hlac_w_bonepitdestroyed_a3_0_6y.dds"},
	{0x413C1C131377F3D6ULL, "hf_shw_ors_sympathize.mor"},
	{0x413C2FB808DEE483ULL, "195413242.ogg"},
	{0x413C31A43958B796ULL, "vxlt_gallowstemplar_5k8_0_0.vlm"},
	{0x413C7A692CD3B354ULL, "aip_follower_fake_carver.gda"},
	{0x413CA0857DB054F8ULL, "545973720.ogg"},
	{0x413CE11526D31683ULL, "plt_pro000pt_cavril_store.nss"},
	{0x413D0E369A62DD1AULL, "vxw_woundedcoast_eh_0_0.vlm"},
	{0x413D97DE2C21CD5CULL, "gen00fl_sebastian6180017.cl"},
	{0x413E827B730ACD5EULL, "gen00fl_varric6076542.cl"},
	{0x413F0428597BB2F1ULL, "hlac_dr_primevaltomb_659_13y.dds"},
	{0x413FE0D5C627AF20ULL, "pro000cr_athenril_entourage.utc"},
	{0x413FF70BB3784CC2ULL, "vxdr_primevaltomb_7xy_0_0.vlm"},
	{0x413FFE2A58F8EBADULL, "art\\vfx\\combat\\vfx_gib_hlk_p.impa.ani"},
	{0x41405B56F08BE2CDULL, "vxca_cave_17z_0_2.vlm"},
	{0x41405F2722F32CCCULL, "vxw_blightlands_ql_0_0.vlm"},
	{0x41409F647C98CE89ULL, "hlac_lt_0_lowtown_docks_a3_-65538_9y.dds"},
	{0x4140B409DCFC7AE8ULL, "6111073_f.fxe"},
	{0x4140D59AF6F85B6DULL, "vxlt_0_lowtown_a2_he3_0_0.vlm"},
	{0x41411DB5FAFAA5CBULL, "dae000_merchant_generic_hm.bnk"},
	{0x41417EA8D12E9979ULL, "6122110_m.fxe"},
	{0x4141D90FA170479FULL, "581582648.ogg"},
	{0x4141D9396DED2D31ULL, "pro000_athenril6049443.cl"},
	{0x41424316EB745751ULL, "mag160_feynriel6020431.cl"},
	{0x41425E74D8014F0AULL, "6031855_m.fxe"},
	{0x414281F0CA2AD0F0ULL, "hlac_lt_0_lowtown_docks_n_131071_5y.dds"},
	{0x4142864267F13C5FULL, "vxht_mansion_n_6lg_0_0.vlm"},
	{0x4142BEBCDAB84277ULL, "966138437.ogg"},
	{0x4142FFDBD2C2D9FCULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_001a_0d.dds"},
	{0x41431C5D5EEE37D1ULL, "725854800.ogg"},
	{0x41441FB5FAFCDEABULL, "dae000_merchant_generic_hm.cnv"},
	{0x414438D749FB5A51ULL, "mcr371_hostile_fenris6082243.cl"},
	{0x41447AE2557C985EULL, "6139311_m.fxe"},
	{0x4144904F6A01F817ULL, "l_qun111ar_safehouse.wbd"},
	{0x41462EEE775D31E8ULL, "470959488.ogg"},
	{0x41468B3A08819F13ULL, "dae000cr_grand_cleric6147640.cl"},
	{0x41468D827A5038FAULL, "vxw_bonepit_7c_0_0.vlm"},
	{0x41469B2C28ED8E0FULL, "182214007.ogg"},
	{0x4146F542922EB278ULL, "885439728.ogg"},
	{0x4147687753CD3C99ULL, "plt_rumor_enemies_among_us.plo"},
	{0x4147A1AD90CD5F01ULL, "zz_vault_debug6203209.cl"},
	{0x41487FBAD12163CEULL, "781257217.ogg"},
	{0x41489306FD6B8A50ULL, "dae000_bodhan6129098.cl"},
	{0x414920D96874FEC7ULL, "vxw_bonepitdestroyed_a3_28g_0_0.vlm"},
	{0x41494E1832F5BBA8ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_021a_0d.dds"},
	{0x4149E02040F13FE5ULL, "vxw_woundedcoast_1ly_0_0.vlm"},
	{0x414A46DD80D85F5FULL, "6118288_f.fxe"},
	{0x414A4DB20402BE68ULL, "vxlt_0_lowtown_el9_0_1.vlm"},
	{0x414AA74ABD8FB92FULL, "931777009.ogg"},
	{0x414B15CC95072C94ULL, "14182686.ogg"},
	{0x414B6F2A5DA2F7C8ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1a.msh"},
	{0x414B80138C58F59AULL, "vxw_woundedcoast_bi_0_0.vlm"},
	{0x414B812A5DA31659ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1a.mao"},
	{0x414B8D2A5DA32AF2ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1a.mmh"},
	{0x414BF77004D7A567ULL, "gen_im_trs_drk_one_lsw_r0.uti"},
	{0x414C0324C459B4D2ULL, "hlac_lt_gallowscourtyard_-2_3z.dds"},
	{0x414D077273BE6A1BULL, "ave310tr_donnic_fight.utt"},
	{0x414D1C942C3EE9C3ULL, "wm_ico_lowtown_roll.dds"},
	{0x414D30ECEC1A3DE2ULL, "vxw_encounter2_camp_jt_0_0.vlm"},
	{0x414E734922A33966ULL, "668714165.ogg"},
	{0x414F5784F1F5C45CULL, "137442621.ogg"},
	{0x414F8C1AEAEC8C6BULL, "gen00fl_varric6063860.cl"},
	{0x414FE27FA1B4DA1EULL, "hlac_w_blightlands_0_8y.dds"},
	{0x41500D37C464DB87ULL, "184529259.ogg"},
	{0x41502D078B134E4AULL, "463882247.ogg"},
	{0x415051FA26BC89F6ULL, "6048110_m.fxe"},
	{0x415066DC3DC7A116ULL, "vxht_0_hightown_bx5_0_0.vlm"},
	{0x415072B6581D0766ULL, "485898961.ogg"},
	{0x41508FE4CE616B58ULL, "vxlt_0_lowtown_docks_b5u_0_0.vlm"},
	{0x4150A46EB4E65B04ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefr.ani"},
	{0x41511558CAD498CAULL, "vxw_bonepit_1i2_0_0.vlm"},
	{0x41514B70C7255057ULL, "llac_lt_0_lowtown_n_-65535_0.dds"},
	{0x4151533FB6FF263BULL, "vxdr_slavers_19q_0_0.vlm"},
	{0x415183C859D22DCFULL, "776433481.ogg"},
	{0x4151E18CEB78CD05ULL, "1054577235.ogg"},
	{0x41521EE3D62097F8ULL, "57301633.ogg"},
	{0x415252A8B9620F35ULL, "vxht_0_hightown_c72_0_0.vlm"},
	{0x41528520D9774E4BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_ric_0n.dds"},
	{0x415288AE5FDA5D2CULL, "plt_zz_lgtpt_2_debug.plo"},
	{0x4152A601EDD07C1CULL, "vxlt_0_lowtown_by8_0_0.vlm"},
	{0x4152C715E29A46F2ULL, "456074220.ogg"},
	{0x4152FE179D6BB0D3ULL, "vxw_sundermount_2ke_0_0.vlm"},
	{0x415341847EEBE176ULL, "hlac_lt_darktown_-1_89y.dds"},
	{0x4153626BD1823CBBULL, "art\\characters\\playercharacter\\humanfemale\\hn_bdy_tmpa.mao"},
	{0x4153BF6878264311ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingc_d.dds"},
	{0x4153F69937012D1FULL, "6053825_m.fxe"},
	{0x41547FF3D2BA56D7ULL, "static_static_p15.psh.fxo"},
	{0x4154D9130FCFAC6BULL, "vxlt_0_lowtown_a2_n_c_ghd_0_0.vlm"},
	{0x41554248EDF7AEA6ULL, "vxdr_0_deeproads_9c4_0_0.vlm"},
	{0x41560183ADC53230ULL, "6127422_m.fxe"},
	{0x41560D431A1B7102ULL, "zz_isa_debug6133937.cl"},
	{0x41562579FD77DE43ULL, "mcr371_hostile_anders6082092.cl"},
	{0x4156606A5E8E3ACBULL, "llac_ca_caverns1_mines_24423_13.dds"},
	{0x41568A4ECBB48B04ULL, "220591427.ogg"},
	{0x4156D1927BC04BEDULL, "one280ip_cache.ncs"},
	{0x415792B65A6B1747ULL, "vxw_bonepitdestroyed_a3_23f_0_0.vlm"},
	{0x4157A16EB4EC7566ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefr.gad"},
	{0x4157AD6BFFFFE57BULL, "1010084269.ogg"},
	{0x4157E73D7E37FB70ULL, "panning.cur"},
	{0x415896FC7D6B174FULL, "llac_w_encounter3_-3_0.dds"},
	{0x4158A2764B204ACBULL, "231710342.ogg"},
	{0x4158CFF1694FEEA6ULL, "6140661_m.fxe"},
	{0x41596DAE5FE04E63ULL, "plt_zz_lgtpt_2_debug.nss"},
	{0x415A02DC4BBE718BULL, "mag170_thrask6016975.cl"},
	{0x415B1C4B8B8045C8ULL, "113032329.ogg"},
	{0x415BA2C9DE105662ULL, "vxw_encounter2_camp_eu_0_0.vlm"},
	{0x415D0C6BDF8DCF39ULL, "follower_barks6153507.cl"},
	{0x415D930A359A1AD9ULL, "mer120_hubert6036662.cl"},
	{0x415D9C6EB4F0D845ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runpausefr.evt"},
	{0x415E25FE131F5B1EULL, "323523949.ogg"},
	{0x415EC3D718B2A276ULL, "6043010_m.fxe"},
	{0x415EE15AC0525937ULL, "6014723_m.fxe"},
	{0x415F4C184BDAA736ULL, "vxw_woundedcoast_1bp_0_0.vlm"},
	{0x415F7D9D4D2FD320ULL, "453650951.ogg"},
	{0x415FD2F0D4EE4F28ULL, "zz_dae_debug6029325.cl"},
	{0x415FD86432E0F5E7ULL, "6076144_f.fxe"},
	{0x415FDB72215BE4D7ULL, "6144523_m.fxe"},
	{0x4160069E04022430ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mb.mao"},
	{0x41600A9E04022B43ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mb.mmh"},
	{0x4160189E0402429DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mb.msh"},
	{0x41605D8E08A415BEULL, "dr_primevaltomb_11985.rml"},
	{0x4160734693B75EBAULL, "vxw_encounter2_fp_0_0.vlm"},
	{0x4160824607C1B8EDULL, "264218200.ogg"},
	{0x41609AD1FFB461EDULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_dotsoft.dds"},
	{0x41609B1AD856EC76ULL, "vxw_bonepit_251_0_0.vlm"},
	{0x416101CCF054AF2CULL, "docks_scene_dockworkers6176510.cl"},
	{0x41616FF48F61C953ULL, "vxw_sundermount_2ld_0_0.vlm"},
	{0x416178B2FAD5167DULL, "6638077.ogg"},
	{0x4161B053C1B03A43ULL, "vxlt_0_lowtown_docks_b6r_0_1.vlm"},
	{0x4161F5E365E94314ULL, "primitives_primitive_p6.vsh.fxo"},
	{0x4162F1D0C4B06F57ULL, "static_static_p46.psh.fxo"},
	{0x41636E5168EEC7D4ULL, "204968366.ogg"},
	{0x4163801D560A7322ULL, "art\\vfx\\combat\\vfx_w_rev_incense_act_c.impa.gad"},
	{0x4163CA6C7299C216ULL, "72560495.ogg"},
	{0x4163EF097667A732ULL, "278324781.ogg"},
	{0x41642024F1BFAB5CULL, "vfx_lavajet_lrg.bnk"},
	{0x41643F6D1335FF46ULL, "lt_gallowstemplar_8165.rml"},
	{0x4164AFB0723FA538ULL, "vis100pt_rescue.ncs"},
	{0x4164BA4000DB4C9EULL, "mcr310_argument6072324.cl"},
	{0x4164F4A4F2936338ULL, "llac_lt_gallowsprison_n_-1_3.dds"},
	{0x416590A59F0A34ACULL, "llac_lt_0_lowtown_a3_n_c_0_5.dds"},
	{0x4165A6D28ECA09F5ULL, "docks_male_elven_worker_sta6175807.cl"},
	{0x4165D2AF22A2F1C1ULL, "lt_0_lowtown_a2_n_c_0_0.phy"},
	{0x416623672F3ED462ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdga_0d.dds"},
	{0x416691B4F9E57F49ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_1x.dds"},
	{0x4166C3DA24DD70E5ULL, "vxw_woundedcoast_1rg_0_0.vlm"},
	{0x41675B53C4CFFC8CULL, "hlac_w_encounter3_0_10x.dds"},
	{0x4167750E59EE9DBBULL, "w_woundedcoast_165_0.msh"},
	{0x41677DDD215F1D7BULL, "hlac_w_woundedcoast_6_0z.dds"},
	{0x4167870E59EEBC06ULL, "w_woundedcoast_165_0.mao"},
	{0x4167930E59EED0A5ULL, "w_woundedcoast_165_0.mmh"},
	{0x416798CAB6F1557AULL, "590421009.ogg"},
	{0x4167BEA2D189FDB8ULL, "mag311_emile_de_launcet6188171.cl"},
	{0x4167DEABFC992066ULL, "vxca_cavern1_haunted_e58_0_0.vlm"},
	{0x4167F065E1A8C34DULL, "one300_zevran6075579.cl"},
	{0x4168178F606BB69CULL, "6208632_f.fxe"},
	{0x416837478805696EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_up.evt"},
	{0x416863CD7045269AULL, "vxw_woundedcoast_tk_0_0.vlm"},
	{0x41689FECA2C6177CULL, "194946973.ogg"},
	{0x4168D2AA2DE0BF66ULL, "15272838.ogg"},
	{0x41690535BF4F9792ULL, "hf_viveka.mor"},
	{0x4169637753EA1AAEULL, "plt_rumor_enemies_among_us.nss"},
	{0x41697A90B5803A1AULL, "dae000_pc_home_ambients6208635.cl"},
	{0x41698665D0E5A310ULL, "825876341.ogg"},
	{0x4169A9C50513352EULL, "hlac_ht_0_hightown_-65538_1x.dds"},
	{0x4169FFE22C27A291ULL, "gen00fl_player6207769.cl"},
	{0x416A322E39091AC0ULL, "gen_im_cns_cft_wrn_04.uti"},
	{0x416A5D4D686F6D07ULL, "mag200_feynriel6023895.cl"},
	{0x416A6B81813049DBULL, "vxht_0_hightown_c38_0_0.vlm"},
	{0x416AA31D5610CD19ULL, "art\\vfx\\combat\\vfx_w_rev_incense_act_c.impa.evt"},
	{0x416AF0AA50B667D2ULL, "vxw_sundermount_2la_0_0.vlm"},
	{0x416AFBF6D6E0F4C5ULL, "vfx_env_dust_02_alpha.mao"},
	{0x416AFE20B64BBE1EULL, "788177259.ogg"},
	{0x416B4D64008A89CDULL, "483175676.ogg"},
	{0x416B855DE72CDD6EULL, "vxca_cave_22w_0_0.vlm"},
	{0x416C3F05C6296137ULL, "6165207_m.fxe"},
	{0x416C4089FE2BAB37ULL, "zz_vault_debug6203364.cl"},
	{0x416C94E6AFD0BDCFULL, "vxlt_darktown_2wf_0_0.vlm"},
	{0x416CB1ACE5869C92ULL, "vxht_mansion2b_1go_0_0.vlm"},
	{0x416CDA096DA0F730ULL, "23533297.ogg"},
	{0x416D3B3D376B721CULL, "6129482_m.fxe"},
	{0x416DBDF53DD0BFB6ULL, "vxw_woundedcoast_1mq_0_0.vlm"},
	{0x416E2E478809C519ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_up.gad"},
	{0x416EB0C092C9443CULL, "6039803_f.fxe"},
	{0x416EE52385AD773AULL, "vxw_encounter2_aq_0_0.vlm"},
	{0x416F0CF7CA4D04F6ULL, "vxw_bonepit_202_0_0.vlm"},
	{0x416FE1D18157E1D3ULL, "vxw_sundermount_2qg_0_0.vlm"},
	{0x4170249E74F04BF1ULL, "382570574.ogg"},
	{0x4170734839277810ULL, "plt_cod_mgc_spirits.plo"},
	{0x4170ED1012AACB32ULL, "vxdr_primevaltomb_510_0_0.vlm"},
	{0x4171E1E5436684B8ULL, "qun221_petrice6077947.cl"},
	{0x4172242B358B3544ULL, "hlac_lt_gallowscourtyard_n_-65538_3z.dds"},
	{0x4172538290B47D02ULL, "6021211_f.fxe"},
	{0x41727ED41FF8D524ULL, "lgt310ip_drop_zone_2.utp"},
	{0x4172BF14A9113090ULL, "6072193_m.fxe"},
	{0x4172DD9FE2E04526ULL, "601029800.ogg"},
	{0x4172EDA41BDAA8CAULL, "vxht_0_hightown_a2_cbv_0_0.vlm"},
	{0x41736CB37A9DEA4DULL, "6098721_m.fxe"},
	{0x41738F2D78FBE9B0ULL, "vxca_caverns1_gie_0_0.vlm"},
	{0x4173B795D398F275ULL, "811214031.ogg"},
	{0x41746BE544649D48ULL, "vxlt_warehouse_n_10a_0_0.vlm"},
	{0x41752A9B27698BF7ULL, "zz_followers_banter6158202.cl"},
	{0x41753F4788100167ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_up.ani"},
	{0x4176551FB11DA4E4ULL, "6173320_m.fxe"},
	{0x41767B0B77167790ULL, "6129200_m.fxe"},
	{0x41776FEDCC7CE2FEULL, "vxht_0_hightown_c84_0_0.vlm"},
	{0x417793FF949F1B67ULL, "583901900.ogg"},
	{0x4177AB1D561B6508ULL, "art\\vfx\\combat\\vfx_w_rev_incense_act_c.impa.ani"},
	{0x4177C0BF4BADCF8BULL, "zz_mag_debug6025383.cl"},
	{0x41782E162ECCB8C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_hands_up_whoa.ani"},
	{0x4178B3E8AC26A009ULL, "gen_im_trs_gen_sta_fir_r2.uti"},
	{0x4178E109ABF2F42FULL, "249710325.ogg"},
	{0x417925BAEEF528D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_blueceiling_d.dds"},
	{0x417953F1E392C8B4ULL, "vxw_woundedcoast_17m_0_0.vlm"},
	{0x41795B3405BAB97EULL, "6144320_m.fxe"},
	{0x417B0C74DFE182D3ULL, "gen00fl_fenris6113918.cl"},
	{0x417B33972D362ABAULL, "6133788_m.fxe"},
	{0x417B592BFE65BBABULL, "49233750.ogg"},
	{0x417C240B4CD0BFF1ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03skeleton.msh"},
	{0x417C2FD22FC6D836ULL, "vxw_woundedcoast_1hr_0_0.vlm"},
	{0x417C320B4CD0D70CULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03skeleton.mao"},
	{0x417C3E0B4CD0EBAFULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03skeleton.mmh"},
	{0x417CE8E7D6B143B7ULL, "mag170_hugh6080733.cl"},
	{0x417CE96FF2E67363ULL, "680471845.ogg"},
	{0x417D570077A38FBAULL, "vxw_encounter2_lr_0_0.vlm"},
	{0x417DA87FF9E089B0ULL, "mag370_meredith6046952.cl"},
	{0x417DF3FC978AD974ULL, "follower_banter6163166.cl"},
	{0x417E53AE734DFA53ULL, "vxw_sundermount_2zf_0_0.vlm"},
	{0x417E7BB0AAC45868ULL, "vxlt_warehouse_n_qq_0_0.vlm"},
	{0x417F2B162ED2D226ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_hands_up_whoa.gad"},
	{0x417F6BFDBC9C80F5ULL, "631063400.ogg"},
	{0x417FDA1425467E7BULL, "1058446821.ogg"},
	{0x417FF8C63CB719ADULL, "hlac_w_bonepitdestroyed_a3_0_7x.dds"},
	{0x41804A2070AB0986ULL, "dae021_torelin_mob6150222.cl"},
	{0x418082B7FBF2980AULL, "dm_jucktugust.mor"},
	{0x418109264497A5D3ULL, "147604959.ogg"},
	{0x4181297376B36A03ULL, "llac_w_sundermount_196611_0.dds"},
	{0x41813358D420F07DULL, "445153771.ogg"},
	{0x41815D70BF4A3412ULL, "hla_ht_0_hightown_a3_65534_0z.dds"},
	{0x4181E5F31183BFD7ULL, "zz_mag_debug6044337.cl"},
	{0x4181FF5391325F49ULL, "ch1multnt.psh.fxo"},
	{0x4182AA1B2D986E16ULL, "319237915.ogg"},
	{0x41835BD6847C8D93ULL, "vxlt_0_lowtown_docks_a3_d2n_0_0.vlm"},
	{0x4183F007BA026E49ULL, "mcr370_meredith6085959.cl"},
	{0x41847689C795D055ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_buttressroof.phy"},
	{0x418511C5E8FBB65FULL, "84486388.ogg"},
	{0x41858A46185089ACULL, "vxlt_shop_6r_0_0.vlm"},
	{0x41859D73320C6ABBULL, "6085512_m.fxe"},
	{0x4185DEB1C6E6E15BULL, "lgt120_orz_bln_official6140209.cl"},
	{0x4185F4AF5278C9A4ULL, "vxlt_gallowstemplar_5qx_0_0.vlm"},
	{0x418626162ED8E805ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_hands_up_whoa.evt"},
	{0x41872DEEB2B8F04BULL, "hlac_lt_0_lowtown_docks_a3_65535_2z.dds"},
	{0x418753B0F0A40138ULL, "hlac_ca_cavern1_haunted_24425_1x.dds"},
	{0x418784EDB633B2E2ULL, "6090344_m.fxe"},
	{0x4187C5CED588E134ULL, "vxw_woundedcoast_12r_0_0.vlm"},
	{0x4187D46434A298D2ULL, "vxw_sundermount_2zc_0_0.vlm"},
	{0x4187F04853586946ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_105a_d.dds"},
	{0x4187F48DE6613212ULL, "hf_slaver01.mor"},
	{0x418827F1B454B1FCULL, "69342815.ogg"},
	{0x41885345F9CEABFCULL, "hlac_w_bonepitdestroyed_a3_393222_0y.dds"},
	{0x41885B6474F8610AULL, "vxw_encounter2_camp_n0_0_0.vlm"},
	{0x418976CB80C1A34DULL, "dae300_bran6144841.cl"},
	{0x418AA1AF21BCF0B6ULL, "vxw_woundedcoast_1ss_0_0.vlm"},
	{0x418B73D4A6FF4A9AULL, "6082152_m.fxe"},
	{0x418B9C73B4A87460ULL, "6065280_f.fxe"},
	{0x418CB209E8B0C1CEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1ww_basic.msh"},
	{0x418CC009E8B0D91FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1ww_basic.mao"},
	{0x418CCC09E8B0EDBCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1ww_basic.mmh"},
	{0x418CED8D9CBA70E8ULL, "vxlt_warehouse_n_tn_0_0.vlm"},
	{0x418D38ABA8BF2329ULL, "llac_lt_gallowstemplar_2388_2.dds"},
	{0x418D93827A23AE02ULL, "215411611.ogg"},
	{0x418DED3DD0EA9D3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_crouching_throat_slit.gad"},
	{0x418E5E4BE82BE0E1ULL, "6020528_m.fxe"},
	{0x418E804DAE39E64DULL, "hlac_lt_gallowstemplar_n_2388_3x.dds"},
	{0x418ED23FCB289287ULL, "323247222.ogg"},
	{0x418F1BB42FF6822CULL, "263418367.ogg"},
	{0x418F674C61A10D47ULL, "one100dg_nabil6055644.cl"},
	{0x418FB9259BEC2BAFULL, "462327041.ogg"},
	{0x418FC92C1DDE4DE3ULL, "vxw_woundedcoast_og_0_0.vlm"},
	{0x4190059F537C1489ULL, "gen_im_arm_cht_nobc_t2.uti"},
	{0x4190404C886EC9FEULL, "6096904_m.fxe"},
	{0x41907E47961727C6ULL, "llac_ht_keep_3731_3.dds"},
	{0x419085A11243D670ULL, "dre000_hireling2.cnv"},
	{0x4190AC9B2934D60BULL, "lightingglitchtest.arl"},
	{0x4190C6D43CB55193ULL, "gen00fl_anders6202533.cl"},
	{0x4190D521DE47AF45ULL, "art\\vfx\\materials\\vfx_small_dot_blend.mao"},
	{0x4190EC7C6C373380ULL, "gen00fl_aveline6139954.cl"},
	{0x4191067C8BCB6D22ULL, "827942828.ogg"},
	{0x419130799B7D5257ULL, "121268430.ogg"},
	{0x4191BF4E263D4BD1ULL, "vxlt_blackemporium_5y_0_0.vlm"},
	{0x4191BF97EF22CD00ULL, "587888631.ogg"},
	{0x4191D7B6CC5AC1C3ULL, "vxlt_hangedman_10a_0_0.vlm"},
	{0x4191EF3537D1F588ULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari.anb"},
	{0x41920BCDCCD81125ULL, "392778584.ogg"},
	{0x419272D1917BB204ULL, "6067818_m.fxe"},
	{0x419283B5ADD74E71ULL, "art\\vfx\\combat\\vfx_m_arc_conj_c.impa.ani"},
	{0x4192DFBA97559832ULL, "vxw_bonepitdestroyed_a3_16g_0_0.vlm"},
	{0x4192F1F1F11B421BULL, "6037373_m.fxe"},
	{0x4193A7A1124645C8ULL, "dre000_hireling2.bnk"},
	{0x41942AE9363213B1ULL, "gen00fl_anders6119462.cl"},
	{0x41945E9F56A88E28ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highwardrobe_sa.dds"},
	{0x41946D5ADE4CB441ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.dura.ani"},
	{0x41947BF578EE1603ULL, "zz_dae_debug6029892.cl"},
	{0x4194881143F248DEULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs103a_0s.dds"},
	{0x4194B8F252D31E4FULL, "6110074_m.fxe"},
	{0x4194E514A8948F4BULL, "6064529_f.fxe"},
	{0x4194F03DD0F0C151ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_crouching_throat_slit.evt"},
	{0x4195504EFFD9A9D5ULL, "817789177.ogg"},
	{0x419577A948DF863FULL, "824670365.ogg"},
	{0x41963421C6A649FDULL, "6112016_m.fxe"},
	{0x419697D56FDCE3C8ULL, "vxlt_darktown_519_0_1.vlm"},
	{0x4197EC3B44242AD3ULL, "art\\characters\\weapons\\dagger\\w_dgr_101a.mao"},
	{0x419810208CD12154ULL, "6149111_f.fxe"},
	{0x419854B863D41B20ULL, "391427176.ogg"},
	{0x41989833DDCB4F6DULL, "hlac_lt_0_lowtown_a2_-65536_2x.dds"},
	{0x41998AD335A4E33CULL, "hlac_w_encounter3_0_3x.dds"},
	{0x41998EB5ADDD7F13ULL, "art\\vfx\\combat\\vfx_m_arc_conj_c.impa.gad"},
	{0x4199BCF039CC0571ULL, "vxht_brothel_n_6em_0_0.vlm"},
	{0x419A17E2C692B9D6ULL, "ser221_party_comments6209901.cl"},
	{0x419A3ABA5B8FC0BAULL, "vxw_encounter2_rl_0_0.vlm"},
	{0x419AA1659B681A49ULL, "vxht_0_hightown_a3_c6y_0_1.vlm"},
	{0x419AE7B2C17EB915ULL, "140764450.ogg"},
	{0x419B0EF367D3DEADULL, "vxca_cave_vw_0_1.vlm"},
	{0x419B1F125C3176D7ULL, "233415939.ogg"},
	{0x419B785ADE52E5E3ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.dura.gad"},
	{0x419B88FE6B6A790EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyb.msh"},
	{0x419B96FE6B6A915FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyb.mao"},
	{0x419BA2FE6B6AA5FCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyb.mmh"},
	{0x419C1F4FFED0CB95ULL, "vxlt_darktown_5lr_0_0.vlm"},
	{0x419C67032C7A94FAULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_dothalf.dds"},
	{0x419CDD019C3DBDFFULL, "6127094_m.fxe"},
	{0x419DEBA11A8AA27BULL, "gen00fl_isabela6097290.cl"},
	{0x419E1FFE206C3E0CULL, "493822318.ogg"},
	{0x419E3B090FD46663ULL, "vxw_woundedcoast_df_0_0.vlm"},
	{0x419E62CEDB69D180ULL, "411276971.ogg"},
	{0x419E9E1EBF88F82CULL, "hlac_w_encounter1_-131072_0x.dds"},
	{0x419F329B21F93E3DULL, "6085392_m.fxe"},
	{0x419F4D7AF14119ACULL, "6137038_m.fxe"},
	{0x419F90FD40BF6038ULL, "vxht_0_hightown_c1i_0_0.vlm"},
	{0x41A021698AC89F6CULL, "6067722_m.fxe"},
	{0x41A028AC1279BF43ULL, "vxlt_0_lowtown_n_gm1_0_0.vlm"},
	{0x41A045A7C0083D22ULL, "follower_banter6161265.cl"},
	{0x41A072B6DE48E297ULL, "hlac_lt_0_lowtown_a2_n_c_65537_1x.dds"},
	{0x41A08BB5ADE39978ULL, "art\\vfx\\combat\\vfx_m_arc_conj_c.impa.evt"},
	{0x41A0A4309B82B797ULL, "10180445.ogg"},
	{0x41A152990BB20C71ULL, "456040426.ogg"},
	{0x41A1E176FB8F1739ULL, "vxw_blightlands_ei_0_0.vlm"},
	{0x41A1E188D3B88235ULL, "dae000ip_crowd_em4_talk8.utp"},
	{0x41A2083DD0FB7420ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_crouching_throat_slit.ani"},
	{0x41A219743EB3C464ULL, "hlac_lt_gallowstemplar_2390_0z.dds"},
	{0x41A2555ADE58C8C8ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.dura.evt"},
	{0x41A2B82A9E8B2017ULL, "plt_lgt210pt_homewtown.nss"},
	{0x41A2D93779A5D394ULL, "vxlt_hangedman_tk_0_0.vlm"},
	{0x41A305A8D159A931ULL, "324073965.ogg"},
	{0x41A313790C3AB9FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\combo_finish.ani"},
	{0x41A316862D023B8FULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.impa.gad"},
	{0x41A37089214295FDULL, "mag161tr_finalfight.ncs"},
	{0x41A37220A5E27E3FULL, "6090046_m.fxe"},
	{0x41A3A633602D2968ULL, "486850595.ogg"},
	{0x41A3B46B855E082BULL, "hlac_ca_caverns1_24422_6z.dds"},
	{0x41A3E269FF279E3EULL, "gen_im_wep_war_two_act2_07.uti"},
	{0x41A3EF8BA7B010DDULL, "vxlt_warehouse_147_0_0.vlm"},
	{0x41A3F90C0945612DULL, "lgt000ip_map_crate.utp"},
	{0x41A3FA177452ADD8ULL, "688131140.ogg"},
	{0x41A4C0D4C1BF01A8ULL, "973983140.ogg"},
	{0x41A53F1E58E4920AULL, "vxw_encounter2_camp_d2_0_0.vlm"},
	{0x41A5566FE42F5F2CULL, "hlac_ht_0_hightown_a2_-196609_0y.dds"},
	{0x41A62494AB7119CEULL, "6104954_f.fxe"},
	{0x41A642C50A22E39CULL, "art\\cinematics\\act3\\romances\\isabella\\cs_isabella_isarom.ani"},
	{0x41A648AC34CD622AULL, "hlac_dr_primevaltomb_658_11y.dds"},
	{0x41A6E6918BC83DC5ULL, "vxht_keep_n_3fz_0_0.vlm"},
	{0x41A72939B743E1FBULL, "472346292.ogg"},
	{0x41A74DF8543D66ADULL, "723001138.ogg"},
	{0x41A7AF8DB280908BULL, "90575346.ogg"},
	{0x41A7BDB0C6B8BE5DULL, "vxdr_0_deeproads_egs_0_0.vlm"},
	{0x41A7D63EF483C3DEULL, "vxca_caverns2_8ak_0_0.vlm"},
	{0x41A81D1FFC3E6786ULL, "gen_im_trs_gem_jade_02.uti"},
	{0x41A85E950B1C3EA5ULL, "6153558_m.fxe"},
	{0x41A89B5D5A909F6CULL, "gen00fl_sebastian6179336.cl"},
	{0x41A92B862D06CB3CULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.impa.evt"},
	{0x41AA20790C40EE3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\combo_finish.gad"},
	{0x41AADD887E52E384ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_002a_0t.dds"},
	{0x41AB47352ABEBA71ULL, "vxw_blightlands_8z_0_0.vlm"},
	{0x41AB6CA40A642285ULL, "610598507.ogg"},
	{0x41AB7B25773E3C81ULL, "6042076_m.fxe"},
	{0x41ABBA141A6E5AD7ULL, "71233777.ogg"},
	{0x41AC3BA80286CCDEULL, "gen00fl_isabela6101558.cl"},
	{0x41ACACE601CA7EE3ULL, "vxw_woundedcoast_ai_0_0.vlm"},
	{0x41AD0975BFE5AF1FULL, "dots_02.dds"},
	{0x41AD2520B0B4A047ULL, "hc_fire_shot.mao"},
	{0x41AD4FC50A2918DEULL, "art\\cinematics\\act3\\romances\\isabella\\cs_isabella_isarom.gad"},
	{0x41AE20DAE5D1C2E7ULL, "71140002.ogg"},
	{0x41AE4F22788F9045ULL, "407310572.ogg"},
	{0x41AE6A927B2FAFD0ULL, "70972702.ogg"},
	{0x41AE6C0E5A2A9152ULL, "w_woundedcoast_165_0.phy"},
	{0x41AECBEB31EBA5AAULL, "llac_lt_0_lowtown_a2_65536_11.dds"},
	{0x41AEF1024B6A2A73ULL, "plt_mer100pt_missing_miners.nss"},
	{0x41AF91B6ACA29305ULL, "hlac_ca_caverns1_mines_24424_9x.dds"},
	{0x41AFD60ADA59BE6CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_basic.msh"},
	{0x41AFDC0ADA59C85EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_basic.mmh"},
	{0x41AFE80ADA59DCF5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_basic.mao"},
	{0x41B004FCF25A6AC7ULL, "vxca_cavern1_haunted_j2e_0_2.vlm"},
	{0x41B05353ED852FB9ULL, "vxw_blightlands_nh_0_0.vlm"},
	{0x41B06DC0F6D0E59EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_1_darren.ani"},
	{0x41B09CC8AF938DDEULL, "llac_w_bonepitdestroyed_a3_196611_0.dds"},
	{0x41B0FB790C46CE5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\combo_finish.evt"},
	{0x41B1021A088BF72DULL, "6104459_m.fxe"},
	{0x41B11CDA261FD6A6ULL, "6175319_m.fxe"},
	{0x41B16A2A16B15307ULL, "173600150.ogg"},
	{0x41B16F510AA9F346ULL, "vxlt_house_dn_0_0.vlm"},
	{0x41B1A0E94BC4471FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprailsb.phy"},
	{0x41B1A7CE71DAD7C8ULL, "dae000_merchant_mage_hf6164460.cl"},
	{0x41B21B075ED10F0BULL, "6067832_f.fxe"},
	{0x41B296D339EB85D6ULL, "vxlt_undercityencounter_v6_0_0.vlm"},
	{0x41B2A51376796780ULL, "vxca_caverns1_dpu_0_0.vlm"},
	{0x41B2E3CFA15B9A3AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic_1.mao"},
	{0x41B31564F45BF8A3ULL, "t1_mue_re1.tnt"},
	{0x41B38DB618ACDA83ULL, "dae000_sketch6179018.cl"},
	{0x41B3B0AC465BF67DULL, "vxlt_gallowsprison_n_6a5_0_0.vlm"},
	{0x41B4219AD340B41BULL, "722347039.ogg"},
	{0x41B44231207D452BULL, "6143247_m.fxe"},
	{0x41B447FCAB00C8E3ULL, "mcr370_negotiation6080304.cl"},
	{0x41B44AC11AC367B6ULL, "gen00fl_merrill6129480.cl"},
	{0x41B4968846692A24ULL, "hlac_lt_0_lowtown_a3_n_c_131072_1x.dds"},
	{0x41B501EC8CD5D562ULL, "gen00pt_party.ncs"},
	{0x41B545A4D63D2703ULL, "docks_fem_sailor_mob6176288.cl"},
	{0x41B5B6614F2C107AULL, "vxht_keep_n_3zm_0_0.vlm"},
	{0x41B5F63C6F42608BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgte_0t.dds"},
	{0x41B5FE89458CAC09ULL, "6182479_m.fxe"},
	{0x41B63886129BF09BULL, "pro000_aveline_intro6109132.cl"},
	{0x41B67DE4378364F2ULL, "414805785.ogg"},
	{0x41B6A0AA1DB83671ULL, "vxht_brothel_n_6kc_0_0.vlm"},
	{0x41B6DBCBA7CED879ULL, "gen00fl_sebastian6189426.cl"},
	{0x41B6DE06680F7DFBULL, "vxlt_0_lowtown_a2_cox_0_0.vlm"},
	{0x41B733862D131655ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.impa.ani"},
	{0x41B74A31908BE12FULL, "6038981_m.fxe"},
	{0x41B762E4C7085428ULL, "776071893.ogg"},
	{0x41B7D339F4D5487AULL, "propalpha1_nolit.psh.fxo"},
	{0x41B806AB2F4D0812ULL, "68843704.ogg"},
	{0x41B88F71533AF7E0ULL, "dae000_arishok6067755.cl"},
	{0x41B8A3F62E9943FBULL, "dae000_sabina6161464.cl"},
	{0x41B8AC94C37DF97AULL, "mag211_marethari6025154.cl"},
	{0x41B8D598D727F718ULL, "vxdr_dungeon_1ml_0_0.vlm"},
	{0x41B95EDCCDF11B2EULL, "hm_amb_keep_guard1.mor"},
	{0x41B9DB2A3A913546ULL, "6103677_m.fxe"},
	{0x41BA2EE6AF6BAD12ULL, "433308791.ogg"},
	{0x41BA7CC5EA8B3BF8ULL, "prp_str_tentical_d.dds"},
	{0x41BB67632EB99C35ULL, "vxht_brothel_n_5gd_0_0.vlm"},
	{0x41BBAA965C20A3CAULL, "lgt110_orz_harrowmont6109034.cl"},
	{0x41BBAE1B39FC273EULL, "278913778.ogg"},
	{0x41BBF79B1066D3D5ULL, "6104730_m.fxe"},
	{0x41BBFA647FD71387ULL, "art\\characters\\playercharacter\\humanfemale\\hat_isa.mao"},
	{0x41BCF0D8D3D01168ULL, "vxlt_darktown_4ye_0_0.vlm"},
	{0x41BD053B10256192ULL, "lt_0_lowtown_docks_27_0.anb"},
	{0x41BE293BE432309BULL, "6207570_m.fxe"},
	{0x41BE44E8FF4CC3E1ULL, "hlac_lt_0_lowtown_-4_0y.dds"},
	{0x41BE5E933E5C4A32ULL, "vxw_bonepit_n0_0_0.vlm"},
	{0x41BEC530DF7B4839ULL, "vxw_blightlands_so_0_0.vlm"},
	{0x41BEFD2EF42B883CULL, "l_ser221ar_murderer_estate_nt.wbd"},
	{0x41BF1F871F3B024BULL, "372157879.ogg"},
	{0x41BF342370BF1095ULL, "6137638_f.fxe"},
	{0x41BF395CDAA8C1F8ULL, "follower_barks6153896.cl"},
	{0x41BF4C4805A777E9ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.cess.ani"},
	{0x41BF742469259901ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.cess.ani"},
	{0x41BF9E9AD815C2B2ULL, "vxca_cavern1_haunted_iid_0_0.vlm"},
	{0x41C118D86421A91CULL, "dae300_bran6144947.cl"},
	{0x41C181FE6B8B3CC7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyb.phy"},
	{0x41C2503D91AC2CC6ULL, "649528422.ogg"},
	{0x41C2A1E94BD2BC24ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprailsb.mmh"},
	{0x41C2A5E94BD2C377ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprailsb.mao"},
	{0x41C2B7E94BD2E1C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprailsb.msh"},
	{0x41C3375B88654F33ULL, "plt_lgt160pt_wk_ave_reward.plo"},
	{0x41C368EBCDA408E1ULL, "691009482.ogg"},
	{0x41C486C0F6E1B918ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_1_darren.gad"},
	{0x41C4A32654D2EADFULL, "143169519.ogg"},
	{0x41C4C54C06170B38ULL, "6127400_m.fxe"},
	{0x41C50CF8979F4099ULL, "6148464_m.fxe"},
	{0x41C540815A6BF1DCULL, "vxw_sundermount_31y_0_0.vlm"},
	{0x41C546091AE33EADULL, "vxdr_primevaltomb_4zs_0_0.vlm"},
	{0x41C56E518868BEB6ULL, "194388892.ogg"},
	{0x41C57F24692A17A3ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.cess.gad"},
	{0x41C5F40CA5D4D1B6ULL, "art\\characters\\weapons\\dagger\\w_dgr_107.mao"},
	{0x41C6674805ADC46BULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.cess.gad"},
	{0x41C6E6BC9F6702CBULL, "6108725_m.fxe"},
	{0x41C767B896EBE77BULL, "6198687_m.fxe"},
	{0x41C7C7AF57CBB0BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_varian.evt"},
	{0x41C7D7AF22F5C4FDULL, "lt_0_lowtown_a2_n_c_0_0.mao"},
	{0x41C7DBAF22F5CB36ULL, "lt_0_lowtown_a2_n_c_0_0.mmh"},
	{0x41C7E5AF22F5DC34ULL, "lt_0_lowtown_a2_n_c_0_0.msh"},
	{0x41C80673CC3DA211ULL, "996197899.ogg"},
	{0x41C823A8199E5B2FULL, "vxdr_0_deeproads_by8_0_0.vlm"},
	{0x41C88C29EA26A01CULL, "6134458_f.fxe"},
	{0x41C923E2195B20EDULL, "vxdr_primevaltomb_3j0_0_0.vlm"},
	{0x41C9A9BA9E2C294AULL, "lt_0_lowtown_a2_n_c_41_0.rml"},
	{0x41C9D94020AFB4B5ULL, "vxht_brothel_n_5jc_0_0.vlm"},
	{0x41CA3A2D38D88B13ULL, "llac_lt_hangedman_n_-1_12.dds"},
	{0x41CA5A3DC0CDBD8CULL, "w_sundermount_100_0.msh"},
	{0x41CA603DC0CDC87EULL, "w_sundermount_100_0.mmh"},
	{0x41CA63454FCF90EFULL, "hlac_ht_0_hightown_a2_-131070_0y.dds"},
	{0x41CA6C3DC0CDDC15ULL, "w_sundermount_100_0.mao"},
	{0x41CACD44F8B1F324ULL, "crt_spell.dds"},
	{0x41CADB8038CCA898ULL, "hlac_ca_caverns2_5802_28y.dds"},
	{0x41CB0975E5602537ULL, "42445040.ogg"},
	{0x41CB8989C7D1F370ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_buttressroof.msh"},
	{0x41CB9789C7D20A8AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_buttressroof.mmh"},
	{0x41CB9B89C7D211C1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_buttressroof.mao"},
	{0x41CC095EA4085022ULL, "llac_ht_mansion2b_818_13.dds"},
	{0x41CC0FBD0F611486ULL, "1034038658.ogg"},
	{0x41CC5C24692FFA88ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.cess.evt"},
	{0x41CD10C99F964987ULL, "vxlt_backalley_g_1x6_0_0.vlm"},
	{0x41CD370DD17160B9ULL, "vxw_blightlands_tn_0_0.vlm"},
	{0x41CD3AB0AC926A4DULL, "vxht_0_hightown_a3_ai7_0_0.vlm"},
	{0x41CD644805B3DDD0ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.cess.evt"},
	{0x41CDC0AF57D010FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_varian.gad"},
	{0x41CDF0C5B69FF7F5ULL, "vxdr_0_deeproads_f4m_0_0.vlm"},
	{0x41CE046127325E81ULL, "644978894.ogg"},
	{0x41CE23F68D0C74DFULL, "gen00fl_sebastian6178374.cl"},
	{0x41CE594DE3B320B9ULL, "vxw_sundermount_2gs_0_0.vlm"},
	{0x41CEC7A4BC236FECULL, "6046011_m.fxe"},
	{0x41CED221E01B9BF9ULL, "hlac_lt_darktown_-1_58x.dds"},
	{0x41CF2C4AAB9535DAULL, "6103844_m.fxe"},
	{0x41CF815C0AC25925ULL, "6087008_m.fxe"},
	{0x41CFEDFB1FFBFAA2ULL, "497015123.ogg"},
	{0x41D0253F24137616ULL, "6145144_f.fxe"},
	{0x41D0E1D3F630BF56ULL, "436318870.ogg"},
	{0x41D0FACCF09BA98AULL, "dae000_gamlen6149470.cl"},
	{0x41D114FD948DFFA9ULL, "62752294.ogg"},
	{0x41D1263039C57E91ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratequest_n.dds"},
	{0x41D1432138074578ULL, "llac_ht_mansion2_818_6.dds"},
	{0x41D16ADC98E9F8BBULL, "hlac_ca_caverns2prc_5802_54x.dds"},
	{0x41D18E3254CDB6FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_enter.gad"},
	{0x41D1F68B4A4DFBA3ULL, "vxht_mansion_n_2lw_0_0.vlm"},
	{0x41D2381DA2B3696CULL, "vxlt_0_lowtown_docks_n_b01_0_0.vlm"},
	{0x41D322801E0A40DEULL, "vxht_0_hightown_bab_0_0.vlm"},
	{0x41D340ABE9A8A73EULL, "vxca_caverns1_mines_e8g_0_0.vlm"},
	{0x41D3BA1E46F57830ULL, "vxlt_darktown_2tl_0_0.vlm"},
	{0x41D435871E3B2CE8ULL, "vxca_cave_134_0_1.vlm"},
	{0x41D4BFAF57D62DFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_varian.ani"},
	{0x41D4CC25FFF89C3FULL, "vxlt_gallowstemplar_n_34q_0_0.vlm"},
	{0x41D50CD940CE201AULL, "vxca_caverns1_edj_0_0.vlm"},
	{0x41D528853C6A34ECULL, "6125684_m.fxe"},
	{0x41D5B783A32B38A9ULL, "hlac_lt_gallowstemplar_n_2394_0x.dds"},
	{0x41D5DF09D4E51DBAULL, "6118471_m.fxe"},
	{0x41D669FA6A799134ULL, "6124679_f.fxe"},
	{0x41D76ECC324F2C31ULL, "gen00fl_merrill6127230.cl"},
	{0x41D7833254D20FECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_enter.evt"},
	{0x41D7A2C914D78ED9ULL, "890419804.ogg"},
	{0x41D7AFC246D8C488ULL, "vxca_caverns2prc_7uk_0_0.vlm"},
	{0x41D7F8375EE11D78ULL, "54053975.ogg"},
	{0x41D8027CD7ACC863ULL, "vxw_woundedcoast_pj_0_0.vlm"},
	{0x41D8254A26D922F1ULL, "hlac_lt_backalley_n_0_8z.dds"},
	{0x41D83CD261B32440ULL, "vxw_woundedcoast_135_0_0.vlm"},
	{0x41D84B3878FCA06FULL, "hlac_ca_caverns2prc_5802_9x.dds"},
	{0x41D8756A9B7D4F7FULL, "lt_0_lowtown_docks_n_8_0_lc.mmh"},
	{0x41D8A3186462A7B3ULL, "6076210_f.fxe"},
	{0x41D8D36AD4D34C0EULL, "hlac_lt_0_lowtown_a2_n_65538_1z.dds"},
	{0x41D953EAC29FABA2ULL, "448506847.ogg"},
	{0x41D9BC923E342F44ULL, "dae300_vashothlead6062801.cl"},
	{0x41D9DD863C8742B9ULL, "451887132.ogg"},
	{0x41D9E540F2C85666ULL, "gen00fl_fenris6108045.cl"},
	{0x41DA04DE0723C1BCULL, "6117615_f.fxe"},
	{0x41DAEE477209BEEEULL, "6102546_m.fxe"},
	{0x41DB1A8607B2E86CULL, "123302605.ogg"},
	{0x41DB2E072F2DE909ULL, "vxca_caverns2_7ms_0_0.vlm"},
	{0x41DBAC8D9E8882CDULL, "vxht_0_hightown_a3_av6_0_0.vlm"},
	{0x41DBE6AD40CC95AAULL, "pro000_cut2d_voyage.cnv"},
	{0x41DC51743F2F65E9ULL, "6025798_m.fxe"},
	{0x41DCAEDC4AD36F58ULL, "lgt110_arl_bann_teagan6115636.cl"},
	{0x41DCCB2AD5A93939ULL, "vxw_sundermount_2hr_0_0.vlm"},
	{0x41DD3745C4A13D0BULL, "vxw_bonepitdestroyed_a3_76_0_0.vlm"},
	{0x41DD8C5E4314C4CCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.inactive.gad"},
	{0x41DD96204260139FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_lgta_0t.dds"},
	{0x41DDD4ACC82A3A15ULL, "6072220_m.fxe"},
	{0x41DE10AB86457F20ULL, "vxht_brothel_n_6do_0_0.vlm"},
	{0x41DE1FE88479EC86ULL, "gen00fl_sebastian6179886.cl"},
	{0x41DEF2D81F4D9AE3ULL, "116644489.ogg"},
	{0x41DF7A03764C2E08ULL, "vxht_brothel_n_2bq_0_0.vlm"},
	{0x41DFC8AD40D04BD2ULL, "pro000_cut2d_voyage.bnk"},
	{0x41DFE8980EBE1DCDULL, "6054848_m.fxe"},
	{0x41DFF8D7E039D837ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic_2.mao"},
	{0x41E017845E9275D6ULL, "890943263.ogg"},
	{0x41E1945D1000595EULL, "vxht_0_hightown_bjc_0_0.vlm"},
	{0x41E1AE311B940707ULL, "364476234.ogg"},
	{0x41E1CBA06E852908ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend02_n.dds"},
	{0x41E216661395989CULL, "follower_banter6164026.cl"},
	{0x41E24EF099BF4B91ULL, "nat380_nathaniel6066707.cl"},
	{0x41E29D04CDC290C4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45ne.msh"},
	{0x41E2AF04CDC2AF2DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45ne.mao"},
	{0x41E2B304CDC2B666ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45ne.mmh"},
	{0x41E2D8249481F483ULL, "vxlt_0_lowtown_a2_n_c_ekg_0_0.vlm"},
	{0x41E2F9EF7EDB0415ULL, "6174283_m.fxe"},
	{0x41E3388A167EDCCBULL, "131765082.ogg"},
	{0x41E3CA76835333CBULL, "6106525_m.fxe"},
	{0x41E48B5E431AE1CAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.inactive.ani"},
	{0x41E48E1A97A717AEULL, "735289109.ogg"},
	{0x41E5654A26DB1A57ULL, "6141450_m.fxe"},
	{0x41E56B3254DE2445ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_enter.ani"},
	{0x41E5893E7CFCF76CULL, "vxca_caverns2_90c_0_0.vlm"},
	{0x41E5BC8433A0121CULL, "6174769_m.fxe"},
	{0x41E5C01E00931AB6ULL, "84696250.ogg"},
	{0x41E5CFC0C0C830B4ULL, "art\\characters\\playercharacter\\humanfemale\\glv_urva.phy"},
	{0x41E5D4455CB53F9AULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.anb"},
	{0x41E62DA8836AE213ULL, "595742490.ogg"},
	{0x41E643B676A3BA54ULL, "6065827_m.fxe"},
	{0x41E668C659C6FC06ULL, "6095127_m.fxe"},
	{0x41E68372D679AEBEULL, "one100dg_nabil6173824.cl"},
	{0x41E72F4B6D496029ULL, "hlac_w_blightlands_65538_0y.dds"},
	{0x41E739799D133F7FULL, "vxdr_0_deeproads_bny_0_0.vlm"},
	{0x41E786EB6EFAA68BULL, "art\\vfx\\combat\\vfx_glb_restore_mana_c.impa.evt"},
	{0x41E7E8A9D099DA6DULL, "lt_0_lowtown_docks_a3_114_0.msh"},
	{0x41E7F6A9D099F1A0ULL, "lt_0_lowtown_docks_a3_114_0.mao"},
	{0x41E7FAA9D099F8F3ULL, "lt_0_lowtown_docks_a3_114_0.mmh"},
	{0x41E80B581654EE81ULL, "mer120_hubert6072907.cl"},
	{0x41E857B44F4BBBCCULL, "6128126_m.fxe"},
	{0x41E8687EDB52EDB9ULL, "zz_dog_debug6127796.cl"},
	{0x41E92348CD130266ULL, "456575660.ogg"},
	{0x41E9388C4F06186FULL, "cd_gen00fl_aveline.wbd"},
	{0x41E9CB3F1FB0F64EULL, "vxdr_0_deeproads_bow_0_0.vlm"},
	{0x41E9D7E2974BF67DULL, "vxht_mansion_n_4h3_0_0.vlm"},
	{0x41E9FA08B485444BULL, "652641580.ogg"},
	{0x41EA1C38FAC5B220ULL, "6143356_m.fxe"},
	{0x41EA1E6A907E9B4DULL, "vxht_0_hightown_a3_as5_0_0.vlm"},
	{0x41EA45E49011F8D9ULL, "790799613.ogg"},
	{0x41EA5F88FCE26E9BULL, "878832344.ogg"},
	{0x41EAA9C445684B4CULL, "225298972.ogg"},
	{0x41EAFB2EF450BFB6ULL, "l_ser221ar_murderer_estate_nt.bnk"},
	{0x41EB3D07C79F51B9ULL, "vxw_sundermount_2mq_0_0.vlm"},
	{0x41EB6189218031C0ULL, "mag161tr_finalfight.utt"},
	{0x41EBB276F4882182ULL, "vxlt_blackemporium_ta_0_0.vlm"},
	{0x41EBC648DD54F222ULL, "fx_icecrystal.dds"},
	{0x41EC45119CDDE7A9ULL, "zz_seb_debug6183088.cl"},
	{0x41EC8288783B97A0ULL, "vxht_brothel_n_6al_0_0.vlm"},
	{0x41ECAA04CDCAFB31ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45ne.phy"},
	{0x41ECDFDC2AEF3EAFULL, "95691326.ogg"},
	{0x41ED49D265E8B12DULL, "gen_im_wep_war_shd_start.uti"},
	{0x41ED9B2A9ECADAF8ULL, "plt_lgt210pt_homewtown.plo"},
	{0x41EE250ADE526006ULL, "gen00fl_isabela6075262.cl"},
	{0x41EE48BA6D5793F8ULL, "452413878.ogg"},
	{0x41EE49BC2581BD33ULL, "6095870_m.fxe"},
	{0x41EE4FCD24C9E518ULL, "vxlt_0_lowtown_docks_n_bdx_0_1.vlm"},
	{0x41EE9FEB6F00F04CULL, "art\\vfx\\combat\\vfx_glb_restore_mana_c.impa.gad"},
	{0x41EF19DC72FB6B6AULL, "6084881_f.fxe"},
	{0x41EF1A494075FC84ULL, "seagull_f.dds"},
	{0x41EF2B96ED938E73ULL, "vxht_mansion2_1tk_0_0.vlm"},
	{0x41EF866043469CF1ULL, "qun131cr_warrens_thug.utc"},
	{0x41EF99D8EB83810FULL, "llac_ht_bartrand_n_3382_0.dds"},
	{0x41EFACC0C0D049B1ULL, "art\\characters\\playercharacter\\humanfemale\\glv_urva.msh"},
	{0x41EFB30ADA8F66F9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_basic.phy"},
	{0x41EFC6C0C0D0766FULL, "art\\characters\\playercharacter\\humanfemale\\glv_urva.mmh"},
	{0x41F005F1507035C3ULL, "6202192_f.fxe"},
	{0x41F0063A01F671DEULL, "vxht_0_hightown_bwd_0_0.vlm"},
	{0x41F0083B1050C0CEULL, "lt_0_lowtown_docks_27_0.rml"},
	{0x41F1DC32279F6FD3ULL, "mag130_danzig6018847.cl"},
	{0x41F277B5B4CFA616ULL, "qun110_petrice6018576.cl"},
	{0x41F28E8C47999B0BULL, "6102118_m.fxe"},
	{0x41F2E9C3011BE47EULL, "qun231_varnell6077148.cl"},
	{0x41F31E5B888E48D0ULL, "plt_lgt160pt_wk_ave_reward.nss"},
	{0x41F34B19724A8098ULL, "6161499_m.fxe"},
	{0x41F379C32DA13315ULL, "ser270_emeric6149365.cl"},
	{0x41F4E999368DC0B9ULL, "sst_ss_coterie_assassin_f.gda"},
	{0x41F5208C459F5540ULL, "vxw_woundedcoast_193_0_0.vlm"},
	{0x41F535758EBE8457ULL, "dae100_gamlen_letter6146980.cl"},
	{0x41F53D1C581EA174ULL, "qcr220_orsino6041848.cl"},
	{0x41F55B9F512613FBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\.mmh"},
	{0x41F5799F512646E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\.msh"},
	{0x41F58F6517E00EF6ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_f_c.anb"},
	{0x41F59EEB6F070D4AULL, "art\\vfx\\combat\\vfx_glb_restore_mana_c.impa.ani"},
	{0x41F5E96FEDC9D654ULL, "344696208.ogg"},
	{0x41F6134B6BB1E455ULL, "6119578_m.fxe"},
	{0x41F64E58947692B2ULL, "dae110ar_docks_nt.are"},
	{0x41F6D8625CD85F2CULL, "vxca_caverns1_hin_0_0.vlm"},
	{0x41F6E501DC20E6C8ULL, "6103675_m.fxe"},
	{0x41F7095E3E379088ULL, "6047427_m.fxe"},
	{0x41F73448AC74E3B6ULL, "hlac_lt_0_lowtown_a2_n_c_196606_0x.dds"},
	{0x41F73D5940976446ULL, "357207622.ogg"},
	{0x41F73E11D1A806A5ULL, "839814853.ogg"},
	{0x41F75B5FA8F5AC84ULL, "269115673.ogg"},
	{0x41F762FB89D17561ULL, "448423657.ogg"},
	{0x41F7CBA6453E62B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v1.ani"},
	{0x41F8296BA80C7D95ULL, "hlac_ca_caverns2_5802_17z.dds"},
	{0x41F87A6C17AB4DDCULL, "293033400.ogg"},
	{0x41F8A4E5F69BA272ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_slaterock_d.dds"},
	{0x41F9253FD55F4812ULL, "6105745_m.fxe"},
	{0x41F9BBD10E4ABC70ULL, "one100dg_vanard6054640.cl"},
	{0x41FA0119E8163A50ULL, "6132716_m.fxe"},
	{0x41FA03E2A39CEA7CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_ands_0d.dds"},
	{0x41FA2566BCEE5ADBULL, "16022299.ogg"},
	{0x41FAF4656A31B020ULL, "vxht_brothel_n_6jm_0_0.vlm"},
	{0x41FB7E3ABC510F2CULL, "vxlt_0_lowtown_docks_brt_0_0.vlm"},
	{0x41FBC76C6F8A2B4BULL, "wm_ico_house.dds"},
	{0x41FBCE5F8C2548C2ULL, "vxlt_hangedman_n_1vt_0_1.vlm"},
	{0x41FC75AF2077B04FULL, "vxca_caverns2_5cl_0_1.vlm"},
	{0x41FCB6409823DEE2ULL, "gen_im_npc_wep_two_2hmb.uti"},
	{0x41FDA278349B0052ULL, "6116169_m.fxe"},
	{0x41FDC82290C07BD6ULL, "335792296.ogg"},
	{0x41FDCCF065526C05ULL, "llac_dr_0_deeproads_17038_4.dds"},
	{0x41FE903CF4F1B684ULL, "dae000_keep_maid_sta6162415.cl"},
	{0x41FEA6A6454442FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v1.gad"},
	{0x41FEBCCD18BCCE7DULL, "zz_dre_debug6091999.cl"},
	{0x41FF0996C15AEDF1ULL, "6098215_m.fxe"},
	{0x41FF5047A0FE2B24ULL, "vxca_caverns2_82x_0_0.vlm"},
	{0x42004558947ED865ULL, "dae110ar_docks_nt.lst"},
	{0x4200DC420CAC2678ULL, "6153407_m.fxe"},
	{0x420108D1ECE5D830ULL, "303641056.ogg"},
	{0x420120B5A2E6ECF5ULL, "704359940.ogg"},
	{0x42019CB71DB68AFBULL, "w_bonepit_250_0.msh"},
	{0x4201AEB71DB6A946ULL, "w_bonepit_250_0.mao"},
	{0x4201BAB71DB6BDE5ULL, "w_bonepit_250_0.mmh"},
	{0x4201BCF664409918ULL, "6166599_m.fxe"},
	{0x4201DC930B7552C6ULL, "ser270_emeric6113345.cl"},
	{0x42020B9583373D64ULL, "vxht_mansion_n_3p9_0_0.vlm"},
	{0x42027AC12B9DEE4FULL, "w_bonepitmining_a2_2_0_ll.mmh"},
	{0x4202D311BAED1353ULL, "71956052.ogg"},
	{0x420309AE75CFB34EULL, "6084499_m.fxe"},
	{0x420381F7A292D9C2ULL, "vxlt_warehouse_n_e7_0_0.vlm"},
	{0x4203926517EC5161ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_f_c.mmh"},
	{0x42047B7C8DE89BC0ULL, "71878052.ogg"},
	{0x4204C3A64548DFE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v1.evt"},
	{0x420530CD6072368EULL, "487508860.ogg"},
	{0x420556DECE16FF48ULL, "6104775_m.fxe"},
	{0x42055DD3F7C5E776ULL, "6135276_m.fxe"},
	{0x4205702139C0BCB5ULL, "214358167.ogg"},
	{0x4205EA4A889BA8F1ULL, "385377.ogg"},
	{0x4205FA9F0C0BCFE7ULL, "llac_w_sundermount_-131075_0.dds"},
	{0x42072149408A9AD2ULL, "seagull_f.mao"},
	{0x42075D5894851FF4ULL, "dae110ar_docks_nt.ncs"},
	{0x4207AAAF46E94B76ULL, "510865122.ogg"},
	{0x420857312CC9691DULL, "552227893.ogg"},
	{0x42088C128EB073CCULL, "6140515_m.fxe"},
	{0x4208F1BFA2F1063DULL, "vxlt_backalley_g_1km_0_0.vlm"},
	{0x42090B8414F78680ULL, "mrl150_merrill6085844.cl"},
	{0x42091D56354BA884ULL, "vxlt_hangedman_n_2hj_0_0.vlm"},
	{0x42091F36E74D609CULL, "vxw_encounter2_g5_0_0.vlm"},
	{0x420957ECABFC3349ULL, "vxlt_gallowstemplar_f_5k4_0_0.vlm"},
	{0x42097F9E895EA656ULL, "vxht_brothel_n_62s_0_0.vlm"},
	{0x4209F1DA22B81656ULL, "420910147.ogg"},
	{0x420A09966C9E4358ULL, "vxw_bonepit_1sb_0_0.vlm"},
	{0x420A52B8A1D657FAULL, "369955762.ogg"},
	{0x420A5919F02CD8E7ULL, "vxw_bonepitdestroyed_a3_e4_0_0.vlm"},
	{0x420A89CE16078C32ULL, "6097860_m.fxe"},
	{0x420AD6B08645A61FULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga103a_0n.dds"},
	{0x420AFA2216252CBAULL, "vxlt_0_lowtown_a3_n_c_h21_0_0.vlm"},
	{0x420B00B6E62615D8ULL, "135604645.ogg"},
	{0x420B373DC1051919ULL, "w_sundermount_100_0.phy"},
	{0x420B62F533119EE8ULL, "6122065_m.fxe"},
	{0x420B6929AF61A20AULL, "6112569_m.fxe"},
	{0x420BFA979E462DEDULL, "and230_mage_girl6215861.cl"},
	{0x420C0F50D17FBF73ULL, "vxht_mansion2_1bm_0_0.vlm"},
	{0x420C0FBF86F30D25ULL, "1024295107.ogg"},
	{0x420C53D7AF735D9FULL, "6172089_m.fxe"},
	{0x420CA878C5EBDEACULL, "vxlt_gallowscourtyard_7ep_0_0.vlm"},
	{0x420CC8B6A7B521BAULL, "501074225.ogg"},
	{0x420CEEFA75EF05C6ULL, "mrl000_marethari6117705.cl"},
	{0x420D397C1D76BC01ULL, "zz_isa_debug6133943.cl"},
	{0x420D89673E1D9AD9ULL, "gen00fl_anders6096473.cl"},
	{0x420DBC18F83A1726ULL, "17637912.ogg"},
	{0x420DD70C5A1BB9BEULL, "6182461_m.fxe"},
	{0x420E561EDA5ABA98ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothb_1.mao"},
	{0x420E6CB5BB451FF1ULL, "6107170_f.fxe"},
	{0x420EB3EEB999FAF1ULL, "vxca_caverns2_73t_0_1.vlm"},
	{0x420ECFB5D729FE9EULL, "dae000_flemeth6000703.cl"},
	{0x420ED7455CD84A4DULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.mmh"},
	{0x420F3985E50271D0ULL, "6041360-aftermath-pt1.cut"},
	{0x420F95F26D6F1B87ULL, "dae000_mother6135750.cl"},
	{0x4210D06D978F98F6ULL, "gen00fl_aveline6139974.cl"},
	{0x421118AB992E0E11ULL, "one100dg_nabil6032533.cl"},
	{0x42115986A56977E6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_male.ani"},
	{0x421159C5F3F0B277ULL, "sw_blood_sh.mao"},
	{0x4211683CFFDBDB8EULL, "128576681.ogg"},
	{0x4211CA2F6FD5E83EULL, "mf.cut"},
	{0x42122BC8516B6D90ULL, "123914617.ogg"},
	{0x4212403DC10B4735ULL, "w_sundermount_100_0.rml"},
	{0x4212BEE353E7BCCFULL, "222007189.ogg"},
	{0x4212C95E4E8ECCCEULL, "gen00fl_dog6167632.cl"},
	{0x4212F7054F9E1555ULL, "6113378_m.fxe"},
	{0x4212FEA1A31617B9ULL, "6107667_m.fxe"},
	{0x42147877A9B047D2ULL, "279333065.ogg"},
	{0x4214C3558B3FBB53ULL, "gen00fl_isabela6015836.cl"},
	{0x4214E888480C1311ULL, "hlac_lt_undercityencounter_-1_41x.dds"},
	{0x4214EC024BC0D4DCULL, "plt_mer100pt_missing_miners.plo"},
	{0x4214FB11CB174BC3ULL, "w_encounter3_185_0.msh"},
	{0x421508B69CAE489AULL, "llac_lt_gallowsprison_n_-1_1.dds"},
	{0x42150911CB17631DULL, "w_encounter3_185_0.mmh"},
	{0x42150D11CB176A2EULL, "w_encounter3_185_0.mao"},
	{0x42153480ED92719EULL, "plt_cod_ltr_ggt_note1.nss"},
	{0x42155DC6A3FE37F3ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_1z.dds"},
	{0x42155E4789F25658ULL, "vxw_woundedcoast_ty_0_0.vlm"},
	{0x421671514D13E6C5ULL, "hlac_lt_darktown_-1_18x.dds"},
	{0x421736CEBDF76A62ULL, "485796575.ogg"},
	{0x42175C5F19175DAAULL, "w_woundedcoast_158_0.mao"},
	{0x4217605F191764F9ULL, "w_woundedcoast_158_0.mmh"},
	{0x4217639C94E71EBDULL, "vxlt_backalley_g_1pl_0_0.vlm"},
	{0x42176A5F191775E7ULL, "w_woundedcoast_158_0.msh"},
	{0x4217BA52B7F80D67ULL, "llac_lt_0_lowtown_a2_n_196607_0.dds"},
	{0x42181A7F72C8100EULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_nath.phy"},
	{0x42188193B63585B0ULL, "vxw_bonepitdestroyed_a3_o1_0_0.vlm"},
	{0x421934916080BF45ULL, "385596498.ogg"},
	{0x421961FF28A0335AULL, "gen_im_fla_bethany_01.uti"},
	{0x4219FDC091D15A90ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishok_movement.anb"},
	{0x421A0B5CFC45C2F3ULL, "dae000_tomwise6167295.cl"},
	{0x421A1D5716942313ULL, "rog_speedmode.dds"},
	{0x421BA97C5FB5FC6FULL, "vxlt_0_lowtown_docks_n_bam_0_0.vlm"},
	{0x421C565972CD22FDULL, "417623291.ogg"},
	{0x421C581D5029D34EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_loop.ani"},
	{0x421C97B99192C6EDULL, "6137695_m.fxe"},
	{0x421CC8CA1752257DULL, "943174050.ogg"},
	{0x421CE7B8AC5925CAULL, "vxw_bonepitmining_a2_1l3_0_0.vlm"},
	{0x421E0076737543E6ULL, "825152096.ogg"},
	{0x421E466B304396A3ULL, "vxca_caverns2_9fb_0_0.vlm"},
	{0x421E9E68E8E908E8ULL, "hlac_ca_caverns1_mines_24425_13y.dds"},
	{0x421ED7A3F22709D7ULL, "6042148_m.fxe"},
	{0x421EDF2B83291C7FULL, "hlac_w_woundedcoast_0_6z.dds"},
	{0x421EEC8D22CCFB81ULL, "vxca_caverns1_dnz_0_0.vlm"},
	{0x421F2EDDFEEC019CULL, "6051051_m.fxe"},
	{0x421F4186A5758C57ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_male.evt"},
	{0x421FC8C572C14EDBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_simple.phy"},
	{0x421FEEB2EAA52B25ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaricommander_movement.anb"},
	{0x42200A885F7469D8ULL, "898904884.ogg"},
	{0x42215A33A1FA3520ULL, "hlac_ca_cavern1_haunted_24423_3x.dds"},
	{0x4221F35EE9BFD97EULL, "llac_lt_0_lowtown_docks_n_-3_1.dds"},
	{0x422228AA9147E02AULL, "6171085_m.fxe"},
	{0x42226E357C76202BULL, "6129764_m.fxe"},
	{0x42228B3988B6B579ULL, "975766900.ogg"},
	{0x422299BED15D33DFULL, "123023967.ogg"},
	{0x4222E55E783D3E58ULL, "dm_kanky_hammertoe.mor"},
	{0x42234B9BC95CC31EULL, "51801225.ogg"},
	{0x422371CC3C93D743ULL, "vxca_caverns2prc_8ku_0_0.vlm"},
	{0x4223CF6282282C2DULL, "6106606_m.fxe"},
	{0x4223D0247BE86ED8ULL, "vxw_woundedcoast_qv_0_0.vlm"},
	{0x4223FC17A625D363ULL, "ca_caverns1_mines_24422_lc.mmh"},
	{0x42246BE580FD580EULL, "vxlt_darktown_4ic_0_0.vlm"},
	{0x42247E43F489CACEULL, "follower_barks6153199.cl"},
	{0x4224F4768BF89461ULL, "art\\characters\\creatures\\dog\\textures\\c_doga_0d.dds"},
	{0x4225550391BBAE1BULL, "6129438_m.fxe"},
	{0x422590D18D85D41BULL, "gen00fl_fenris6116548.cl"},
	{0x4225A98C3E087A05ULL, "vxca_caverns1_j2e_0_1.vlm"},
	{0x4225EBE98F90D9B5ULL, "hlac_lt_gallowscourtyard_-131072_0x.dds"},
	{0x4225EF62E5C5ED17ULL, "vxht_mansion_n_6nu_0_0.vlm"},
	{0x422633E3E7102D4AULL, "hlac_lt_undercitydungeon_-1_10z.dds"},
	{0x42264286A57BACE0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_male.gad"},
	{0x422665D22087DB78ULL, "619693312.ogg"},
	{0x422673DFF01650FDULL, "6087390_m.fxe"},
	{0x4226D3D1923358ACULL, "vxlt_0_lowtown_docks_bck_0_0.vlm"},
	{0x4227D1C208A45AA3ULL, "vxlt_gallowstemplar_4sq_0_1.vlm"},
	{0x4227E2E1BD0F2A22ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangbn.msh"},
	{0x4227E4E1BD0F2DD8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangbn.mmh"},
	{0x4227F0E1BD0F427BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangbn.mao"},
	{0x42282627FB7959E1ULL, "99061329.ogg"},
	{0x42283C88FF2B4142ULL, "ico_boots_light_1.dds"},
	{0x4228AE107DF1884DULL, "qcr221cr_qunari_warrior.utc"},
	{0x4228E84D8E198CE4ULL, "271636926.ogg"},
	{0x4229528C4F3C2845ULL, "cd_gen00fl_aveline.bnk"},
	{0x4229A3E61018F359ULL, "docks_male_laborer_mob6175403.cl"},
	{0x4229A945C0BB967CULL, "vxca_caverns2prc_6ng_0_0.vlm"},
	{0x422A5736FD16EA88ULL, "lowtown_male_lowlife2_mob6169192.cl"},
	{0x422ACF3763DA0666ULL, "vxw_woundedcoast_1e9_0_0.vlm"},
	{0x422AE40BFB102BE6ULL, "6041812_m.fxe"},
	{0x422B10571D6044D9ULL, "6060726_m.fxe"},
	{0x422B4C6147FCFBEBULL, "mag121_idunna6084484.cl"},
	{0x422B67BE92CF0DDFULL, "zz_pro_debug6059748.cl"},
	{0x422BBC6AAE59EB3BULL, "vxlt_0_lowtown_docks_c0d_0_0.vlm"},
	{0x422BE9F7F2A3A907ULL, "50677200.ogg"},
	{0x422C49BFF6A18DBCULL, "169922830.ogg"},
	{0x422D1B870BDD69A3ULL, "363486612.ogg"},
	{0x422D79C9D7B2C40BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkf.evt"},
	{0x422DAC38F84E3DB9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portrait_n.dds"},
	{0x422DF0B78C81D3DBULL, "hlac_lt_0_lowtown_n_1_6x.dds"},
	{0x422E563A759E5DFDULL, "lt_0_lowtown_docks_a3_21_0.rml"},
	{0x422E56A126F69D01ULL, "6087083_f.fxe"},
	{0x422EA1B945BF51AEULL, "6129354_m.fxe"},
	{0x422ED78E78752342ULL, "vxlt_warehouse_n_r2_0_0.vlm"},
	{0x422F381D900510FEULL, "vxlt_0_lowtown_n_e1z_0_0.vlm"},
	{0x422F536A1667AB25ULL, "82085842.ogg"},
	{0x422FAD4ED0056F4CULL, "llac_w_encounter2_4_0.dds"},
	{0x422FBD21C367D121ULL, "lgt310cr_mag_ors_informant.utc"},
	{0x422FD3239077F38FULL, "706280743.ogg"},
	{0x4231311D503BEC88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_loop.gad"},
	{0x4231BC4B943305E5ULL, "ring_deep_green.dds"},
	{0x4231F3F9ABE435C4ULL, "qcr220_orsino6061785.cl"},
	{0x423208DB428BD1A5ULL, "992486986.ogg"},
	{0x4232712A7749A5AEULL, "6137285_m.fxe"},
	{0x4232A3A1D68427C4ULL, "llac_lt_house_-1_6.dds"},
	{0x4234451E815D5B7FULL, "189576045.ogg"},
	{0x4234475678D34FBDULL, "vxlt_backalley_g_1zj_0_0.vlm"},
	{0x423492C9D7B90DCCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkf.gad"},
	{0x423539846D9AC48DULL, "w_woundedcoast_147_0.mmh"},
	{0x42353A2B016C5D88ULL, "hlac_lt_0_lowtown_docks_n_1_0z.dds"},
	{0x42353D846D9ACBDEULL, "w_woundedcoast_147_0.mao"},
	{0x423545AE8429712CULL, "vxlt_0_lowtown_docks_bfh_0_0.vlm"},
	{0x42354B846D9AE313ULL, "w_woundedcoast_147_0.msh"},
	{0x4235A48770899300ULL, "6027832_m.fxe"},
	{0x423697E096B03567ULL, "200092965.ogg"},
	{0x42369C89482FAE10ULL, "llac_ca_cavern1_haunted_24424_6.dds"},
	{0x4236A81CCE147EE8ULL, "dre000_sibling_has_blight6036524.cl"},
	{0x4236AE47DC4EE34CULL, "llac_lt_0_lowtown_docks_n_65534_0.dds"},
	{0x4236BD7E39EA3678ULL, "dae000_player_generic_find6179045.cl"},
	{0x42371E722CBDFB6CULL, "697184009.ogg"},
	{0x42374ECC00F2594EULL, "mcr370_cullen6087429.cl"},
	{0x423767C29ABEE69AULL, "hlac_ca_cavern1_haunted_24425_1z.dds"},
	{0x42379A4133F957D6ULL, "vxw_encounter2_kk_0_0.vlm"},
	{0x4238ABE02A18A7CFULL, "art\\vfx\\materials\\vfx_d_heatnoise_nodist.mao"},
	{0x42391CCB7FC9225EULL, "qun210cr_guard6083843.cl"},
	{0x42392E5BD8B2E750ULL, "hf_amb_keep_maid.mor"},
	{0x423998DDB1A524F2ULL, "422232456.ogg"},
	{0x4239DF634349578EULL, "391330324.ogg"},
	{0x423A2671E014D67EULL, "dae000_merchant_dm_mob6174242.cl"},
	{0x423A540FD89AA5E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_action.evt"},
	{0x423A5EB000767B16ULL, "6126607_m.fxe"},
	{0x423A75809976F8BCULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_bladestorm.ani"},
	{0x423A7B4AB0AF879EULL, "llac_lt_0_lowtown_docks_a3_196609_0.dds"},
	{0x423A7D491AB28DE2ULL, "6082325_m.fxe"},
	{0x423A8819BFF43E00ULL, "vxlt_0_lowtown_a2_n_c_gni_0_0.vlm"},
	{0x423ADF349C6618E9ULL, "gen00fl_varric6033658.cl"},
	{0x423B218E8BDA8011ULL, "llac_lt_undercityserialkiller_-1_10.dds"},
	{0x423B341364B65229ULL, "6093495_f.fxe"},
	{0x423B5DED5006E4B1ULL, "vxw_bonepitmining_a2_247_0_0.vlm"},
	{0x423B91C9D7BF2ACAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkf.ani"},
	{0x423BB9D69E5BAF71ULL, "vxw_blightlands_np_0_0.vlm"},
	{0x423BD04706B92C81ULL, "vxca_caverns1_dpx_0_0.vlm"},
	{0x423BD34E3897DAADULL, "qun210tr_follower_amb.utt"},
	{0x423BF5704C2EE5B3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_catacombs.phy"},
	{0x423C40A2223434BCULL, "mrl351_pridedemon6165563.cl"},
	{0x423C996998212246ULL, "vxlt_gallowstemplar_35d_0_0.vlm"},
	{0x423CB0A1FD32F984ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyn.msh"},
	{0x423CC2A1FD3318EDULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyn.mao"},
	{0x423CC6A1FD331F26ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyn.mmh"},
	{0x423CD4503AF6BD84ULL, "6114091_m.fxe"},
	{0x423D496B6A6B3BC2ULL, "vxlt_warehouse_n_y3_0_0.vlm"},
	{0x423D767491ED0E51ULL, "dae200_saemus6141194.cl"},
	{0x423D8FC56A2A6246ULL, "gen00fl_varric6076018.cl"},
	{0x423DFB8A90424B33ULL, "vxht_mansion2b_2q3_0_0.vlm"},
	{0x423E80D5D28F0462ULL, "hlac_lt_0_lowtown_docks_a3_-2_0x.dds"},
	{0x423EBCA3C2F9F23AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dazed.gad"},
	{0x423ECAA190914A7FULL, "art\\characters\\weapons\\greataxes\\w_gax_004a.msh"},
	{0x423ED0A190915421ULL, "art\\characters\\weapons\\greataxes\\w_gax_004a.mmh"},
	{0x423EDCA1909168C2ULL, "art\\characters\\weapons\\greataxes\\w_gax_004a.mao"},
	{0x423F7FC7284112DEULL, "llac_lt_0_lowtown_65533_3.dds"},
	{0x423F88F7AFBB02EDULL, "character_character_p97.psh.fxo.dx11.assembly"},
	{0x423F9568AADC7657ULL, "hlac_w_bonepit_393220_0x.dds"},
	{0x423FA8A0D50E900FULL, "621026666.ogg"},
	{0x423FAA8ECDA742CCULL, "6179461_m.fxe"},
	{0x423FFD155A3ABCDAULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_ferelden_jimmy_da.dds"},
	{0x424054473878A39EULL, "hlac_lt_warehouse_n_0_6x.dds"},
	{0x4240B8AAD9B90021ULL, "rdr111dg_assistant6016638.cl"},
	{0x42416B0FD8A0EB31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_action.gad"},
	{0x42418280997D2DFEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_bladestorm.gad"},
	{0x4241BEF46925482BULL, "vxlt_gallowscourtyard_6py_0_0.vlm"},
	{0x4241C6C99C9167B5ULL, "hlac_w_sundermount_-196604_0y.dds"},
	{0x4241F74A709AFE25ULL, "347579464.ogg"},
	{0x424205E7D178B993ULL, "vxlt_0_lowtown_docks_a3_n_arx_0_0.vlm"},
	{0x42421B65C742C7BFULL, "796193842.ogg"},
	{0x42421FBE64A0D92CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint_12.pwk"},
	{0x4242C7CDFFB990FCULL, "vxlt_0_lowtown_a3_n_c_hsp_0_0.vlm"},
	{0x4243055FA87796B4ULL, "gen00fl_aveline6200890.cl"},
	{0x42432613F4F63DBDULL, "wm_ico_clinic_roll.dds"},
	{0x4243463157EB53FBULL, "fxc_smoke_add.mao"},
	{0x4243A4EEA8D0CA45ULL, "921788877.ogg"},
	{0x4243AC2767BAD14DULL, "6143281_m.fxe"},
	{0x4243B6BD25A4A84CULL, "218123302.ogg"},
	{0x4243E5D499E768FDULL, "vxlt_shop_cx_0_0.vlm"},
	{0x42442529E2E082F1ULL, "537763729.ogg"},
	{0x4244681697C0871EULL, "hlac_ca_caverns2prc_5802_44x.dds"},
	{0x42446B173B18713EULL, "6082168_f.fxe"},
	{0x4244846CC1DC0EEAULL, "vxdr_primevaltomb_8md_0_1.vlm"},
	{0x424486AB6DFEEF92ULL, "336832645.ogg"},
	{0x4244D7E790A6C7DCULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1lrgtileweb.mao"},
	{0x4244E3E790A6DC7FULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1lrgtileweb.mmh"},
	{0x4244E9E790A6E621ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1lrgtileweb.msh"},
	{0x4245BFA3C3001651ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dazed.evt"},
	{0x424632846DA92C4AULL, "w_woundedcoast_147_0.phy"},
	{0x4246487043602D7EULL, "6208646_m.fxe"},
	{0x424655A09DED21AEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_long_1.mao"},
	{0x4246BDA1FD3B64F1ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyn.phy"},
	{0x4246D5E2B4503456ULL, "6120946_m.fxe"},
	{0x4246F5A096B14B3BULL, "62477256.ogg"},
	{0x424703F32C12C1DDULL, "llac_w_woundedcoast_327680_0.dds"},
	{0x42478954F65F3FEAULL, "6128973_m.fxe"},
	{0x42479A1BCFD675D2ULL, "352316561.ogg"},
	{0x4247DC6133D95DDEULL, "plt_mcr371pt_harvester.nss"},
	{0x4247DC8EC194603CULL, "gen_im_trs_dal_hvy_boo_r2.uti"},
	{0x4247EB85C65ECC42ULL, "6044925_m.fxe"},
	{0x4247FBF2586C6B46ULL, "vxht_0_hightown_a2_61z_0_0.vlm"},
	{0x42485D8099830D1DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_bladestorm.evt"},
	{0x42486C0FD8A70BDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_action.ani"},
	{0x4248C837C769EC86ULL, "177469372.ogg"},
	{0x4248D812829956CCULL, "6069590_f.fxe"},
	{0x42498D9B59EA1779ULL, "follower_banter6154039.cl"},
	{0x4249A651DFEFF47AULL, "6020088_m.fxe"},
	{0x4249DC056B79ADA8ULL, "hlac_ca_cavern1_haunted_24424_16y.dds"},
	{0x4249DE3AF5B6026AULL, "hlac_lt_0_lowtown_a3_n_c_-1_1x.dds"},
	{0x4249EB3822B28B72ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_begging.gad"},
	{0x4249EDDFE1C46518ULL, "919447961.ogg"},
	{0x424A2BB39051C7F1ULL, "vxw_blightlands_sw_0_0.vlm"},
	{0x424A2C1ED344BDABULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_nud_0n.dds"},
	{0x424A33260A9AC547ULL, "164411507.ogg"},
	{0x424AAC0A3C65E122ULL, "6125369_m.fxe"},
	{0x424ADD8773A5CF8CULL, "263860632.ogg"},
	{0x424AF345D504A384ULL, "art\\vfx\\combat\\vfx_w_gn2_taunt_c.anb"},
	{0x424B2A65DD366A13ULL, "vxlt_gallowscourtyard_n_5z7_0_0.vlm"},
	{0x424BAFC572E65F7AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_simple.msh"},
	{0x424BBDC572E676B3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_simple.mao"},
	{0x424BC1C572E67DC0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_simple.mmh"},
	{0x424C6A2896F5511FULL, "vxlt_0_lowtown_n_f0k_0_0.vlm"},
	{0x424C6D67823863B3ULL, "vxht_mansion2b_2j2_0_0.vlm"},
	{0x424D78AEC616B397ULL, "hlac_lt_0_lowtown_docks_a3_n_196606_0y.dds"},
	{0x424DBD35F255ACD3ULL, "6103359_m.fxe"},
	{0x424DED5C5DA65D4DULL, "vxht_keep_5jk_0_0.vlm"},
	{0x424E21E9D6C67174ULL, "6081393_m.fxe"},
	{0x424E7C36981D4FEEULL, "121965903.ogg"},
	{0x424EE02BACD1A5A3ULL, "shock_projectile_f_e.dds"},
	{0x424F78EABED20E16ULL, "vxdr_primevaltomb_6x4_0_0.vlm"},
	{0x424FC35773713DA8ULL, "139456103.ogg"},
	{0x42504A0B97CFAC4EULL, "vxlt_blackemporium_zg_0_0.vlm"},
	{0x42506B0618411B26ULL, "zz_seb_debug6214538.cl"},
	{0x4250AF9DF60FFC90ULL, "vxca_cave_2i8_0_0.vlm"},
	{0x4250EE3822B8AEA9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_begging.evt"},
	{0x42510D82A2BD6A24ULL, "6053286_m.fxe"},
	{0x4251233C19BA0CDEULL, "vxlt_0_lowtown_n_fj4_0_0.vlm"},
	{0x4251602CFE088A62ULL, "vxdr_slavers_1b9_0_0.vlm"},
	{0x4251FCBC203C754BULL, "hlac_ca_cavern1_haunted_24422_3x.dds"},
	{0x42522D85E8CEA9C6ULL, "hlac_lt_gallowstemplar_2390_0x.dds"},
	{0x425246BCD3910B04ULL, "qun120_arishok_guard6060610.cl"},
	{0x425257B18BDD817DULL, "vxlt_shop_fw_0_0.vlm"},
	{0x425263658E0F8B15ULL, "static_static_p95.psh.fxo.dx11"},
	{0x4252D7A3C30AC920ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dazed.ani"},
	{0x4253305862A5677FULL, "llac_lt_0_lowtown_a2_n_65534_4.dds"},
	{0x42536723885E2579ULL, "hlac_ca_caverns2_5802_60x.dds"},
	{0x4253937FF69D470BULL, "coins_n.dds"},
	{0x4253C67DCE81965BULL, "vxht_0_hightown_n_c7o_0_0.vlm"},
	{0x425436BC231248A3ULL, "zz_approval_debug6001819.cl"},
	{0x42547DFB17E588D6ULL, "vxw_encounter2_qm_0_0.vlm"},
	{0x4254890150052A64ULL, "hlac_ht_brothel_n_3382_1z.dds"},
	{0x4254BCBA4956E736ULL, "hlac_ht_chantry_n_2534_53x.dds"},
	{0x4254C5507EA0F85CULL, "m_dae251ar_mount_caverns.wbd"},
	{0x42552E08EF485D43ULL, "vxca_cave_208_0_0.vlm"},
	{0x4255B23AD9253E92ULL, "store_dae170cr_weapons.utm"},
	{0x4255D8E16615FC48ULL, "w_bonepitdestroyed_a3_17_0.msh"},
	{0x4255E90BDF74EEF7ULL, "vxht_keep_n_5jz_0_0.vlm"},
	{0x4255EAE166161AD9ULL, "w_bonepitdestroyed_a3_17_0.mao"},
	{0x4255F6E166162F72ULL, "w_bonepitdestroyed_a3_17_0.mmh"},
	{0x4256647B818E669AULL, "828946232.ogg"},
	{0x425691DA5725B00CULL, "6171942_m.fxe"},
	{0x4256EBF983E8874DULL, "art\\characters\\weapons\\greatswords\\w_gsw_008a.mmh"},
	{0x4256EFF983E88D9EULL, "art\\characters\\weapons\\greatswords\\w_gsw_008a.mao"},
	{0x4256F1279E1B8012ULL, "lt_0_lowtown_a2_n_5_0.rml"},
	{0x4256FDF983E8A5D3ULL, "art\\characters\\weapons\\greatswords\\w_gsw_008a.msh"},
	{0x425737067414D5EBULL, "6053317_m.fxe"},
	{0x4258369D70C7D4A4ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.dura.ani"},
	{0x42583B369150A7CCULL, "vxht_chantry_12n_0_0.vlm"},
	{0x42589D908247E071ULL, "vxw_blightlands_tv_0_0.vlm"},
	{0x42591245D5111587ULL, "art\\vfx\\combat\\vfx_w_gn2_taunt_c.mmh"},
	{0x4259197F72FFA5D1ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_nath.mmh"},
	{0x4259337F72FFD18FULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_nath.msh"},
	{0x4259763253CBB681ULL, "al_mage_t03.gda"},
	{0x42597DB16979AE03ULL, "6118600_f.fxe"},
	{0x4259BFD09489A071ULL, "vxw_sundermount_2gk_0_0.vlm"},
	{0x425A039F62CADFD2ULL, "fen310_fenris6144518.cl"},
	{0x425ABC4E38B26420ULL, "qun210tr_follower_amb.ncs"},
	{0x425B58A22CF8BCCEULL, "802189174.ogg"},
	{0x425BB188E9FA9A80ULL, "docks_male_sailor_mob6176664.cl"},
	{0x425BB51740D53C19ULL, "6209893_m.fxe"},
	{0x425BB5EAF5F8FBF0ULL, "hlac_lt_0_lowtown_a2_n_-65535_5z.dds"},
	{0x425C12B9674A1258ULL, "12767360.ogg"},
	{0x425C2F8A7924FADDULL, "614781396.ogg"},
	{0x425C5E5FDC57DDD2ULL, "vxlt_0_lowtown_a2_n_c_ep2_0_1.vlm"},
	{0x425C8E48B19373CCULL, "6177661_m.fxe"},
	{0x425CA8AE8D4E800FULL, "vxlt_blackemporium_u7_0_0.vlm"},
	{0x425CD3AD141D9CBCULL, "383366221.ogg"},
	{0x425CDE1F525B500BULL, "mrl210_merrill6117475.cl"},
	{0x425CE90B05AB2CD2ULL, "hlac_lt_0_lowtown_a2_65537_1z.dds"},
	{0x425CF55D25C1FF23ULL, "hlac_ht_0_hightown_a2_-65537_0y.dds"},
	{0x425D80C16D501906ULL, "6095376_m.fxe"},
	{0x425DA19213B5B80DULL, "art\\levels\\races\\proxy\\props\\_textures\\corpses_d.dds"},
	{0x425DBE952CF77027ULL, "64711361.ogg"},
	{0x425E06E8EEEB5234ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masc_0s.dds"},
	{0x425E339D70CC3B06ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.dura.gad"},
	{0x425EA16FC3251214ULL, "vxlt_hangedman_5t_0_0.vlm"},
	{0x425ED63822C4C318ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_begging.ani"},
	{0x425F0193C4D1F0C2ULL, "llac_lt_undercityserialkiller_-1_39.dds"},
	{0x425FAB87E3A5C1FCULL, "vxlt_0_lowtown_a3_n_c_hev_0_0.vlm"},
	{0x42601980EDD22FE9ULL, "plt_cod_ltr_ggt_note1.plo"},
	{0x42609AD628DB4E55ULL, "1056909856.ogg"},
	{0x4260D31C41574773ULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.impa.ani"},
	{0x4261553A75C9B515ULL, "lt_0_lowtown_docks_a3_21_0.anb"},
	{0x4261553D17C0F2C7ULL, "vxca_caverns2prc_7v5_0_0.vlm"},
	{0x42615D5F0E6540E3ULL, "fa_cl_a.mao"},
	{0x42617764A670B37FULL, "gen00fl_isabela6099577.cl"},
	{0x4261B13E53AC1C23ULL, "6071115_m.fxe"},
	{0x42627091C7EEE122ULL, "gen00fl_merrill6108540.cl"},
	{0x4262AB539CEA2234ULL, "6155312_f.fxe"},
	{0x42633539EBB7C052ULL, "fxe_rippleout.dds"},
	{0x42634DA14F286036ULL, "vxw_woundedcoast_18b_0_0.vlm"},
	{0x42636E6ED7F7D64AULL, "qun120_arishok6051315.cl"},
	{0x42639DA0FE91141CULL, "45093362.ogg"},
	{0x4263B57F3E7668EEULL, "art\\vfx\\combat\\vfx_d_teleport_full_c.impa.ani"},
	{0x42644DC87A9E4C4BULL, "dae000_bodhan_letter6144275.cl"},
	{0x426469C41C4BECDCULL, "6112960_f.fxe"},
	{0x42647B6A2DD6A8EAULL, "923600527.ogg"},
	{0x4264C82BC1494C3AULL, "vxw_blightlands_fo_0_0.vlm"},
	{0x4264F90F9009E9CAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_2_ian.gad"},
	{0x42652E9D70D251E5ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.dura.evt"},
	{0x42654528D040DE72ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\hf_hed_mera_0d.dds"},
	{0x42656CB095A27FA8ULL, "520320251.ogg"},
	{0x4265B7977770577FULL, "hm_guard01.mor"},
	{0x4265CA87B3D7A48CULL, "6034177_m.fxe"},
	{0x4265FFDAD13D171FULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf3a.phy"},
	{0x42660FD840A29396ULL, "vxw_bonepit_zx_0_0.vlm"},
	{0x4266374D90D196EAULL, "vxht_bartrand_n_4hk_0_0.vlm"},
	{0x42664B447262FEC8ULL, "687817719.ogg"},
	{0x4266A6C5E127ABB7ULL, "llac_lt_0_lowtown_a2_n_-3_5.dds"},
	{0x42674373B7CDDF29ULL, "vxht_0_hightown_a3_9uw_0_0.vlm"},
	{0x4267E4F983F6EE0AULL, "art\\characters\\weapons\\greatswords\\w_gsw_008a.phy"},
	{0x42680BCCE185F335ULL, "vxht_0_hightown_a2_c6q_0_0.vlm"},
	{0x426831AD867FB8F1ULL, "vxw_sundermount_2hj_0_0.vlm"},
	{0x42685021087076F6ULL, "llac_lt_0_lowtown_a2_n_65538_5.dds"},
	{0x42695F689B0C5627ULL, "vxca_cave_17z_0_0.vlm"},
	{0x42696D8673A25BD5ULL, "vxlt_0_lowtown_a2_gxk_0_0.vlm"},
	{0x426AB020A24FDB10ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_30ring.msh"},
	{0x426ABE20A24FF32AULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_30ring.mmh"},
	{0x426AC220A24FFA61ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_30ring.mao"},
	{0x426AD5D2149FA837ULL, "hlac_lt_0_lowtown_-262144_2x.dds"},
	{0x426B05A185C057D4ULL, "6106882_f.fxe"},
	{0x426B10890EC9D31AULL, "lgt150ip_rand_3.utp"},
	{0x426B36BF82337856ULL, "lt_0_lowtown_docks_a3_n_168_0.mao"},
	{0x426B42BF82338CF5ULL, "lt_0_lowtown_docks_a3_n_168_0.mmh"},
	{0x426B44BF8233906BULL, "lt_0_lowtown_docks_a3_n_168_0.msh"},
	{0x426BBA50D36D0329ULL, "6360714.ogg"},
	{0x426BD1A190B8439EULL, "art\\characters\\weapons\\greataxes\\w_gax_004a.phy"},
	{0x426C7987F9E7590FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackleft.gad"},
	{0x426CA0C42D189367ULL, "hlac_w_sundermount_131069_0x.dds"},
	{0x426CB06C5223B7F3ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_010a_0s.dds"},
	{0x426CCBBBA098762EULL, "art\\characters\\weapons\\spears\\w_spr_001a_p.msh"},
	{0x426CE5BBA098A21CULL, "art\\characters\\weapons\\spears\\w_spr_001a_p.mmh"},
	{0x426D10FB5B040019ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage3_.gad"},
	{0x426D1F87414B42CEULL, "910957106.ogg"},
	{0x426D4BB43EB84003ULL, "6136925_f.fxe"},
	{0x426D4F5F6592ACA2ULL, "515918421.ogg"},
	{0x426D835789378ABDULL, "513185215.ogg"},
	{0x426DC142348DA1C2ULL, "llac_lt_0_lowtown_docks_a3_-65538_9.dds"},
	{0x426DCB1C4161C3AAULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.impa.evt"},
	{0x426DD4B019090A2FULL, "vxht_0_hightown_cbj_0_1.vlm"},
	{0x426DE65F020FD6A4ULL, "vxdr_primevaltomb_6uq_0_0.vlm"},
	{0x426E301E4F6F22D3ULL, "vxht_0_hightown_a2_cas_0_1.vlm"},
	{0x426E6F82B75C3AEBULL, "597687351.ogg"},
	{0x426EFA19EBD6BBF8ULL, "hlac_w_woundedcoast_-196605_0y.dds"},
	{0x426EFFEE97111707ULL, "vxht_mansion2b_2ph_0_0.vlm"},
	{0x426FFD2B6E13E9A6ULL, "319484568.ogg"},
	{0x42700B7973839AD9ULL, "942131012.ogg"},
	{0x42703553E3CE5682ULL, "120357462.ogg"},
	{0x4270AA37B26DC75BULL, "vxht_0_hightown_n_c9a_0_0.vlm"},
	{0x4270AD7F3E80E53FULL, "art\\vfx\\combat\\vfx_d_teleport_full_c.impa.evt"},
	{0x4270CAB623668A00ULL, "6131394_f.fxe"},
	{0x4270D33D20073F24ULL, "vxdr_primevaltomb_549_0_0.vlm"},
	{0x4270D6574006CF93ULL, "6043319-dragonling_reveal.cut"},
	{0x42723308B0AA23B2ULL, "llac_ht_0_hightown_a2_65535_1.dds"},
	{0x42728E87F9EBE8BCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackleft.evt"},
	{0x4272E49ADC2E03AEULL, "gen00fl_isabela6079082.cl"},
	{0x4273AD64B2B61CA8ULL, "vxdr_0_deeproads_cwp_0_0.vlm"},
	{0x4273FF15C1A2911DULL, "6164347_m.fxe"},
	{0x427421FB5B0A3C67ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_mage3_.ani"},
	{0x4274302E7E529892ULL, "gen00fl_anders6113083.cl"},
	{0x42744E9736F9D0ECULL, "vxlt_gallowstemplar_n_3rh_0_0.vlm"},
	{0x4274AA1C4167AA1DULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.impa.gad"},
	{0x4274CE126D14F530ULL, "6181750_m.fxe"},
	{0x4275DFD198B99A13ULL, "887265847.ogg"},
	{0x4275F36D551E6D49ULL, "vxht_0_hightown_a3_cbd_0_0.vlm"},
	{0x42760D292BFF0E9EULL, "sib131_letter_react6127301.cl"},
	{0x4276428FC2A9D711ULL, "vxht_mansion2_1w8_0_0.vlm"},
	{0x4276461198625D12ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v2_c.impa.gad"},
	{0x4276A38A7875D171ULL, "vxw_sundermount_2mi_0_0.vlm"},
	{0x427700DAD14B8C24ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf3a.mmh"},
	{0x4277062692BE89E8ULL, "713895355.ogg"},
	{0x427716DAD14BB1C6ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf3a.msh"},
	{0x427717AB8FACF97EULL, "cd_mag131_keran.bnk"},
	{0x427773B7BC773BF6ULL, "vxw_encounter2_3b_0_1.vlm"},
	{0x42778E7F3E86CF28ULL, "art\\vfx\\combat\\vfx_d_teleport_full_c.impa.gad"},
	{0x4277E78BBF49973DULL, "47989909.ogg"},
	{0x42789B2EF2974971ULL, "360569833.ogg"},
	{0x4278C34A1A8DD1A9ULL, "820023192.ogg"},
	{0x427963691119E5A7ULL, "184218078.ogg"},
	{0x4279A242617C4301ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.cess.ani"},
	{0x4279D40F901C06B0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_2_ian.ani"},
	{0x427A042711E1DB63ULL, "vxw_bonepit_13m_0_0.vlm"},
	{0x427A7E01DF42E7C8ULL, "269416006.ogg"},
	{0x427A84922F4660D4ULL, "lgt260_eau_tmp_asst_harbor6143640.cl"},
	{0x427B9FA26DB1F5CEULL, "vxlt_blackemporium_uz_0_0.vlm"},
	{0x427BA416AFAE5271ULL, "mcr370_assignments6182065.cl"},
	{0x427BE3DD7117EEACULL, "vxw_bonepitdestroyed_a3_1ep_0_0.vlm"},
	{0x427BE695AD9A125CULL, "llac_ht_playershouse_n_855_30.dds"},
	{0x427BFC8759A7D9E5ULL, "719814346.ogg"},
	{0x427C4AFA06231B5FULL, "vxw_sundermount_21c_0_0.vlm"},
	{0x427C583BF405EF24ULL, "vxdr_primevaltomb_6xv_0_0.vlm"},
	{0x427C5CA02FFC6A11ULL, "lgt200_eau_tome.bnk"},
	{0x427CCE1271175F0BULL, "vxlt_0_lowtown_docks_a3_d0p_0_0.vlm"},
	{0x427D491198688149ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v2_c.impa.evt"},
	{0x427D88AAFB4C73CAULL, "6119829_m.fxe"},
	{0x427D8F2E2A346457ULL, "281647195.ogg"},
	{0x427E207E21EE6D97ULL, "lt_0_lowtown_docks_225_0.mmh"},
	{0x427E247E21EE74A4ULL, "lt_0_lowtown_docks_225_0.mao"},
	{0x427E367E21EE9309ULL, "lt_0_lowtown_docks_225_0.msh"},
	{0x427E50E215A08CFCULL, "ico_armor_medium_2.dds"},
	{0x427E525AA1710CA1ULL, "137611020.ogg"},
	{0x427E71B28D6E0B6DULL, "gen00fl_dog6115096.cl"},
	{0x427F0C99FB59131CULL, "90438598.ogg"},
	{0x427FAD426180C1A3ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.cess.gad"},
	{0x427FD781A565F704ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debilitationcast.ani"},
	{0x42803EA030002021ULL, "lgt200_eau_tome.cnv"},
	{0x42807C09947FE6F2ULL, "vxht_chantry_357_0_0.vlm"},
	{0x42809687F9F833D5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackleft.ani"},
	{0x428099441B8E4E4AULL, "540224076.ogg"},
	{0x4280C3507EC62FD6ULL, "m_dae251ar_mount_caverns.bnk"},
	{0x4280E87C78DC22EFULL, "274372048.ogg"},
	{0x4281ABE5A5357D3AULL, "vxw_blightlands_li_0_0.vlm"},
	{0x4281CDA01E184DF6ULL, "6096618_m.fxe"},
	{0x4281F8B5146B2867ULL, "vxlt_undercitydungeon_1c3_0_1.vlm"},
	{0x428207535A7B4FD8ULL, "vxca_caverns1_edx_0_0.vlm"},
	{0x42821F41A4AC3528ULL, "vxdr_0_deeproads_clq_0_0.vlm"},
	{0x4282316752BE4293ULL, "6174704_m.fxe"},
	{0x4284EE1CE4013F1BULL, "vxdr_0_deeproads_fix_0_0.vlm"},
	{0x42850B5F4CD10D90ULL, "mrl351_phantomelves6170360.cl"},
	{0x428515676A6BE9F1ULL, "vxw_sundermount_2vh_0_0.vlm"},
	{0x428558D82F35EE63ULL, "vxw_woundedcoast_4v_0_0.vlm"},
	{0x4285FDDCA8B0A4BAULL, "w_sundermount_46_0.rml"},
	{0x428647D270654C32ULL, "93629344.ogg"},
	{0x42864D5425E60532ULL, "vxdr_primevaltomb_8ok_0_0.vlm"},
	{0x428686A7B7561B4AULL, "32347663.ogg"},
	{0x42868A426186A488ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.cess.evt"},
	{0x4286CA13B76586C6ULL, "6013415_m.fxe"},
	{0x4286D481A56C1166ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debilitationcast.gad"},
	{0x4286F93822D58264ULL, "vxht_mansion_n_6bn_0_0.vlm"},
	{0x428706465ACB9EF0ULL, "vxca_caverns1_mines_ih5_0_0.vlm"},
	{0x4287535C73BC4591ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf_scalpm1.mao"},
	{0x4287AA33E7B6B3FBULL, "vxlt_0_lowtown_a3_n_c_fge_0_0.vlm"},
	{0x4287C84055ADA3D9ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_encn_0n.dds"},
	{0x428852D0870C7224ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrainenter.ani"},
	{0x42885AEF943BD51CULL, "vxht_brothel_n_6b9_0_0.vlm"},
	{0x4289078D966C895CULL, "t1_mul_te1.tnt"},
	{0x428911E219A0E939ULL, "6043286_m.fxe"},
	{0x42892E63206F4BF6ULL, "494808467.ogg"},
	{0x4289365639CBE5C9ULL, "vxlt_gallowstemplar_n_5r7_0_0.vlm"},
	{0x42895ADBF6E2ED57ULL, "sib131im_key_vault.uti"},
	{0x4289DFE40C2973F4ULL, "zz_followers_banter6159060.cl"},
	{0x4289E409164B0F83ULL, "657650381.ogg"},
	{0x428A47AEA3757607ULL, "zz_one_debug6075020.cl"},
	{0x428A695B4B78F83BULL, "vxca_caverns2_2ry_0_0.vlm"},
	{0x428AC9FA22051A6AULL, "hla_ht_0_hightown_a2_-1_0x.dds"},
	{0x428AF7040F36A5F3ULL, "601132815.ogg"},
	{0x428B0D1B6B43B459ULL, "vxlt_0_lowtown_a2_n_c_do0_0_0.vlm"},
	{0x428B3111987495B8ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v2_c.impa.ani"},
	{0x428B9F7D2790FEFEULL, "603972513.ogg"},
	{0x428BA48F66688237ULL, "vxw_sundermount_36z_0_1.vlm"},
	{0x428BF352FA9C75A9ULL, "mag121_viveka6156508.cl"},
	{0x428C030795B53F0FULL, "vxht_mansion2b_26b_0_0.vlm"},
	{0x428C0C98FF2B9CA8ULL, "lt_0_lowtown_docks_a3_n_143_0.mmh"},
	{0x428C1898FF2BB14BULL, "lt_0_lowtown_docks_a3_n_143_0.mao"},
	{0x428C2A98FF2BCFD2ULL, "lt_0_lowtown_docks_a3_n_143_0.msh"},
	{0x428CCF81A5707445ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debilitationcast.evt"},
	{0x428CF02B74038028ULL, "vxht_chantry_2hs_0_0.vlm"},
	{0x428E19C0ACEADE69ULL, "6101337_m.fxe"},
	{0x428E4FD08710D886ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrainenter.gad"},
	{0x428EE416DE28BDBAULL, "24946296.ogg"},
	{0x428EF4430B5C2A2FULL, "isa311_isabela6092240.cl"},
	{0x428F433B12F26F44ULL, "gen00fl_carver6104885.cl"},
	{0x428F4CEC669E35EBULL, "98911692.ogg"},
	{0x428F95A625652113ULL, "mag121_viveka6095654.cl"},
	{0x428F9CD7C4ED6E32ULL, "vxlt_0_lowtown_gc9_0_0.vlm"},
	{0x428FAA36ADAE7582ULL, "hightown_market_guard_sta6174617.cl"},
	{0x428FC091568CD78CULL, "590750936.ogg"},
	{0x42900A0657FC3CB4ULL, "6087473_f.fxe"},
	{0x42901DC2972B95BAULL, "vxw_blightlands_qj_0_0.vlm"},
	{0x4290281C7C400415ULL, "dae000_arishok6072884.cl"},
	{0x42903D7943F394C2ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzfence.phy"},
	{0x4290A9BBE41A1D49ULL, "6058949_m.fxe"},
	{0x4290FFFB72DF3E98ULL, "vxht_0_hightown_a2_bxo_0_0.vlm"},
	{0x42911F9F063E538DULL, "6066386_m.fxe"},
	{0x4291A83AF509656BULL, "6125409_m.fxe"},
	{0x4291D1335B2628ECULL, "art\\levels\\races\\design\\_textures\\plc_coinpile01_d.dds"},
	{0x4291D77999933E26ULL, "gen_im_cft_rec_arn_04.uti"},
	{0x42926B6456C80F6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v1.ani"},
	{0x42927DEEB1797A08ULL, "173746633.ogg"},
	{0x42928619C7C0C9C3ULL, "mag311cr_huon.ncs"},
	{0x4292A3EABFFE99D4ULL, "llac_ht_0_hightown_65533_3.dds"},
	{0x4292C16134191B29ULL, "plt_mcr371pt_harvester.plo"},
	{0x4292DB6C7F06CE99ULL, "vxw_woundedcoast_57_0_0.vlm"},
	{0x4292E628DB660CC1ULL, "qcr000_arishok6059304.cl"},
	{0x429321EF29B134EDULL, "682091506.ogg"},
	{0x429375F39FDD542BULL, "hlac_ht_0_hightown_a3_65535_1x.dds"},
	{0x42939CEEADE5CDF9ULL, "vrc321cr_guard_ghost_1.utc"},
	{0x4293CAB5212C06E3ULL, "vxw_woundedcoast_1y_0_0.vlm"},
	{0x4293FF6447B77555ULL, "hlac_lt_undercitydungeon_-1_47y.dds"},
	{0x42944E9EEFBB2356ULL, "325837714.ogg"},
	{0x4294B8877FB04A3AULL, "410995088.ogg"},
	{0x42954AD08716EF65ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrainenter.evt"},
	{0x4295B24ED4AD7DF1ULL, "773933571.ogg"},
	{0x4295B57036F9EDD6ULL, "follower_barks6153476.cl"},
	{0x4295C0217F3BDDC2ULL, "lowtown_male_alienage1_sta6162758.cl"},
	{0x4295CC8C2750F363ULL, "hlac_lt_0_lowtown_docks_a3_65532_0y.dds"},
	{0x4296743E3F46E215ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\arcanehorror_overlay.anb"},
	{0x4296B5FADB489A7FULL, "304849397.ogg"},
	{0x42971E653E79A886ULL, "6040543_m.fxe"},
	{0x42973C2D7F3A68EDULL, "329857690.ogg"},
	{0x429795700DFF0AEBULL, "gen00fl_anders6113008.cl"},
	{0x42998C931D05DA45ULL, "dae000_merchant_weapons_dm6164407.cl"},
	{0x42999A84CD28B7F1ULL, "468410334.ogg"},
	{0x4299E54CDC74B353ULL, "vxlt_gallowstemplar_f_5ax_0_0.vlm"},
	{0x4299EF2532CBBC2EULL, "6054546_m.fxe"},
	{0x429AB931D2244221ULL, "40134050.ogg"},
	{0x429ABFE82429465EULL, "character_character_p270.psh.fxo.dx11"},
	{0x429AD482CCBB0D24ULL, "lgt110_urn_barker6132975.cl"},
	{0x429AF6484EFC8A6AULL, "vxht_chantry_n_359_0_1.vlm"},
	{0x429B620865F998A8ULL, "vxht_chantry_2ep_0_0.vlm"},
	{0x429CE73F3E980504ULL, "vxw_woundedcoast_131_0_0.vlm"},
	{0x429D977E0942D69FULL, "859793941.ogg"},
	{0x429DCB562DB71B5DULL, "hlac_w_bonepitdestroyed_a3_327684_0y.dds"},
	{0x429DF57F3450A34DULL, "lgt310_mag_grigor6153031.cl"},
	{0x429E731EA19894F2ULL, "stone_flare.mao"},
	{0x429EEE704C835728ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_catacombs.mmh"},
	{0x429EFA704C836BCBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_catacombs.mao"},
	{0x429F063B3B6DF33FULL, "vxlt_gallowstemplar_n_3rg_0_0.vlm"},
	{0x429F0C704C838A52ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_catacombs.msh"},
	{0x429F1D7C132F04D0ULL, "vxlt_0_lowtown_a2_n_h20_0_0.vlm"},
	{0x429F636456D28BBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v1.evt"},
	{0x429FCCB3979B2E09ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180r.ani"},
	{0x429FDFD5FC661B66ULL, "hlac_dr_0_deeproads_5849_9y.dds"},
	{0x42A01A17E6FF7DEBULL, "6122509_m.fxe"},
	{0x42A0441E1474E6F3ULL, "vxht_keep_4bx_0_0.vlm"},
	{0x42A08FBED3FE0170ULL, "108029556.ogg"},
	{0x42A0D28CC7177351ULL, "hlac_lt_0_lowtown_a2_n_-262143_0z.dds"},
	{0x42A0F929D2576E32ULL, "370877954.ogg"},
	{0x42A105D31546953AULL, "hlac_lt_darktown_-1_80z.dds"},
	{0x42A12B742D49D8B7ULL, "825040299.ogg"},
	{0x42A183E23FDFCA60ULL, "431631161.ogg"},
	{0x42A1B221B40912B9ULL, "hlac_lt_warehouse_65537_1y.dds"},
	{0x42A1EF14FDA4B240ULL, "vxht_mansion2b_33o_0_0.vlm"},
	{0x42A334FA3B7ABDD5ULL, "vxlt_0_lowtown_a2_glw_0_0.vlm"},
	{0x42A371D0C3C3C82BULL, "120379841.ogg"},
	{0x42A376DCEA9835B6ULL, "zz_mcr_debug6071696.cl"},
	{0x42A451C4879EE77AULL, "6045571_m.fxe"},
	{0x42A457F1256C919BULL, "vxlt_0_lowtown_docks_a3_8ug_0_0.vlm"},
	{0x42A4C471416101FEULL, "vxdr_primevaltomb_4hn_0_0.vlm"},
	{0x42A500BB112D107BULL, "vxlt_0_lowtown_e21_0_0.vlm"},
	{0x42A52B474738ACA1ULL, "vxlt_foundry_n_wl_0_0.vlm"},
	{0x42A551605DE961CBULL, "vxca_caverns1_cjw_0_0.vlm"},
	{0x42A6446456D875A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v1.gad"},
	{0x42A6E7B397A17A8BULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180r.gad"},
	{0x42A6E93961D0D6C5ULL, "dae000cr_commoner_m2.utc"},
	{0x42A6F1334B4D35A8ULL, "follower_barks6153386.cl"},
	{0x42A787F09A244D50ULL, "6166933_m.fxe"},
	{0x42A7A3159EA002A8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mb.msh"},
	{0x42A7B5159EA02139ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mb.mao"},
	{0x42A7C1159EA035D2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mb.mmh"},
	{0x42A84498E5E568C8ULL, "vxlt_darktown_6hd_0_0.vlm"},
	{0x42A89493A76BD6EFULL, "220364655.ogg"},
	{0x42A8C1C1FB9B9B9BULL, "dae000_merchant_armor_hm6164281.cl"},
	{0x42A933219743456FULL, "one220_yevhen6173606.cl"},
	{0x42A97C16309E529EULL, "dae000_fenarel6133409.cl"},
	{0x42A98BC23E0E6A9FULL, "llac_ca_caverns2prc_5802_32.dds"},
	{0x42A9A6606AE8E73BULL, "llac_lt_0_lowtown_-65535_5.dds"},
	{0x42AA2CD67F43A01BULL, "6179599_f.fxe"},
	{0x42AAB2A507E5AC7AULL, "6126324_m.fxe"},
	{0x42AB08E4C1FA6F93ULL, "follower_banter6168045.cl"},
	{0x42AB737097D2DA36ULL, "365005629.ogg"},
	{0x42AC8568D2F932A0ULL, "vxw_sundermount_2kb_0_0.vlm"},
	{0x42ACA991365381C1ULL, "vxca_caverns2_9f8_0_0.vlm"},
	{0x42AD64AD48CF71E4ULL, "305838135.ogg"},
	{0x42AD74D87A8E7EA8ULL, "vxdr_0_deeproads_ccl_0_0.vlm"},
	{0x42AD85AFED6E99B3ULL, "vxlt_blackemporium_wi_0_0.vlm"},
	{0x42ADE4642EEBD020ULL, "t3_wpn_stl.tnt"},
	{0x42ADE4B397A794F0ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180r.evt"},
	{0x42AEBA260D2C9533ULL, "gen00fl_varric6072981.cl"},
	{0x42AEEE82477EFC19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_chest_out.ani"},
	{0x42AF6BB139D2004DULL, "6135860_m.fxe"},
	{0x42AF6FE611CB3327ULL, "406006331.ogg"},
	{0x42AF9399BC8F3CE9ULL, "dae000cr_merchant_generic.utc"},
	{0x42B1A6D72D70D655ULL, "vxlt_0_lowtown_a2_gcv_0_0.vlm"},
	{0x42B1C22A44A2C4EDULL, "86789083.ogg"},
	{0x42B1DB5CEEA6B3B0ULL, "20569011.ogg"},
	{0x42B1EF953A2C51F7ULL, "vxw_bonepit_27j_0_0.vlm"},
	{0x42B235DFF4831259ULL, "344445607.ogg"},
	{0x42B237991A9FC5EEULL, "6177704_m.fxe"},
	{0x42B251EA69875D24ULL, "6122128_m.fxe"},
	{0x42B295873EA74DC1ULL, "w_encounter1_13_0_lc.mmh"},
	{0x42B2A26C7B9A3931ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_4z.dds"},
	{0x42B3393D79924A18ULL, "dae000_gamlen6145087.cl"},
	{0x42B36F73817EDE50ULL, "combat_dw_r3_a1.gda"},
	{0x42B4029188F0A15EULL, "hlac_lt_gallowscourtyard_65534_2y.dds"},
	{0x42B432774B134ABDULL, "219179899.ogg"},
	{0x42B48BED0FAE2F49ULL, "vxlt_gallowstemplar_n_5c4_0_0.vlm"},
	{0x42B5187944127276ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzfence.mao"},
	{0x42B5247944128615ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzfence.mmh"},
	{0x42B52607E3D108FEULL, "6126521_f.fxe"},
	{0x42B526794412898BULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzfence.msh"},
	{0x42B52D58A2D991F4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsmb.phy"},
	{0x42B5C32802A6EE58ULL, "6084912_m.fxe"},
	{0x42B5C9824784DC5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_chest_out.gad"},
	{0x42B66040A4D42D4DULL, "zz_followers_banter6158024.cl"},
	{0x42B6673F8A1B2D1AULL, "vxw_bonepitmining_a2_253_0_0.vlm"},
	{0x42B6946F8665C0F0ULL, "vxlt_0_lowtown_a2_n_e87_0_0.vlm"},
	{0x42B6B4D98A16ADA6ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney2_sa.dds"},
	{0x42B6C906C060E453ULL, "vxlt_gallowstemplar_f_5oz_0_0.vlm"},
	{0x42B6E711D8DD9660ULL, "418688294.ogg"},
	{0x42B6F6B3C91161E4ULL, "art\\levels\\races\\dwarf\\props\\dead\\prp_sarcophagus.msh"},
	{0x42B708B3C911804DULL, "art\\levels\\races\\dwarf\\props\\dead\\prp_sarcophagus.mao"},
	{0x42B70CB3C9118686ULL, "art\\levels\\races\\dwarf\\props\\dead\\prp_sarcophagus.mmh"},
	{0x42B73262B10CBF63ULL, "659640513.ogg"},
	{0x42B741927F618579ULL, "lt_0_lowtown_docks_104_0.msh"},
	{0x42B7467305D04DACULL, "vxlt_0_lowtown_docks_b5a_0_0.vlm"},
	{0x42B74B927F619667ULL, "lt_0_lowtown_docks_104_0.mmh"},
	{0x42B74F927F619CB4ULL, "lt_0_lowtown_docks_104_0.mao"},
	{0x42B791F5AA41515AULL, "llac_lt_0_lowtown_131073_2.dds"},
	{0x42B7E9AB8FE44184ULL, "cd_mag131_keran.wbd"},
	{0x42B7F6E16668F919ULL, "w_bonepitdestroyed_a3_17_0.rml"},
	{0x42B840B011C8290DULL, "247521262.ogg"},
	{0x42B8BE70DD8FA39FULL, "574479681.ogg"},
	{0x42B90FD415E17598ULL, "vxht_0_hightown_bu8_0_0.vlm"},
	{0x42B91E8C1AC3D840ULL, "hlac_ht_0_hightown_a2_2_1z.dds"},
	{0x42B94C98D8B45F47ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_2y.dds"},
	{0x42B9CAF922843604ULL, "vxw_woundedcoast_197_0_0.vlm"},
	{0x42B9DDA901E4CD8CULL, "hlac_lt_house_-1_0x.dds"},
	{0x42BA2B7DBCB6766FULL, "cd_dae000_light_tombs_boss.wbd"},
	{0x42BA3D4D9E6FBDAAULL, "vxlt_gallowstemplar_35h_0_0.vlm"},
	{0x42BAB3E3847826E7ULL, "179198008.ogg"},
	{0x42BAF745C4EF4B20ULL, "vxw_sundermount_2lc_0_0.vlm"},
	{0x42BB2B94142AE684ULL, "1072289717.ogg"},
	{0x42BBC37E43160264ULL, "46563354.ogg"},
	{0x42BC52BA20909C36ULL, "6123334_m.fxe"},
	{0x42BC9F9C601AF19DULL, "58408570.ogg"},
	{0x42BCE682478B2C40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_chest_out.evt"},
	{0x42BD2BB54183650AULL, "1008663590.ogg"},
	{0x42BD56ABE8273F56ULL, "llac_dr_slavers_3207_4.dds"},
	{0x42BDBE1DADDB1C27ULL, "gen00fl_isabela_init6127677.cl"},
	{0x42BDDF3BE6DC3B49ULL, "397234003.ogg"},
	{0x42BDE0DC04AC9870ULL, "vxca_caverns1_mines_j5i_0_0.vlm"},
	{0x42BE43406D32ADB2ULL, "vxdr_primevalentrance_2zk_0_1.vlm"},
	{0x42BE753522773B53ULL, "gen00fl_varric6063851.cl"},
	{0x42BE7BE07C785191ULL, "vxht_mansion2_1n3_0_0.vlm"},
	{0x42BE8AA7AF95F1B5ULL, "dae200_alistair_drunk.bnk"},
	{0x42BE8ACC3E106CC3ULL, "dm_bartrand.mor"},
	{0x42BEC44A115A2137ULL, "gen_im_flw_arc_sebastian_02.uti"},
	{0x42BECD43F4275BD6ULL, "follower_banter6151293.cl"},
	{0x42BF0A58A2E1AAF1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsmb.msh"},
	{0x42BF2458A2E1D6AFULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpsmb.mmh"},
	{0x42BF5D94887A039FULL, "llac_lt_0_lowtown_docks_n_-131073_2.dds"},
	{0x42BF96327B0A7C53ULL, "6103392_f.fxe"},
	{0x42C0039C9955BF80ULL, "988751859.ogg"},
	{0x42C0355F24903A1DULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2win.phy"},
	{0x42C103B3C919CC51ULL, "art\\levels\\races\\dwarf\\props\\dead\\prp_sarcophagus.phy"},
	{0x42C10C6B80F99C7DULL, "llac_lt_undercitydungeon_-1_4.dds"},
	{0x42C1328D6583BB83ULL, "hlac_lt_backalley_g_65537_5z.dds"},
	{0x42C1ACA7AF98610DULL, "dae200_alistair_drunk.cnv"},
	{0x42C1BE55F55EBD18ULL, "vxlt_gallowstemplar_5j5_0_0.vlm"},
	{0x42C20D3F1DC891FFULL, "6115588_f.fxe"},
	{0x42C217692CE2C1D7ULL, "lgt260_eau_tmp_roderick6143707.cl"},
	{0x42C28CEE29F9AE0FULL, "llac_dr_primevaltomb_657_1.dds"},
	{0x42C2A0D95E2817E4ULL, "lgt311_shw_mer_mettin6156527.cl"},
	{0x42C30EAE836BA15FULL, "vxw_encounter2_36_0_0.vlm"},
	{0x42C318C58F575B32ULL, "hlac_lt_0_lowtown_docks_n_0_1x.dds"},
	{0x42C34A8F6B1C219CULL, "zz_dae_debug6029422.cl"},
	{0x42C3915557C4F9E8ULL, "mer120_hubert6083922.cl"},
	{0x42C3B34D50340301ULL, "vxlt_0_lowtown_a2_n_c_gnd_0_0.vlm"},
	{0x42C3CD4FCD07AAF4ULL, "hlac_ht_chantry_2534_47z.dds"},
	{0x42C42EAD0ED8C2ACULL, "gen_im_acc_blt_act1_04.uti"},
	{0x42C4D30681C35D7CULL, "vxht_brothel_n_6d3_0_0.vlm"},
	{0x42C4F2A46D0B6349ULL, "follower_banter6161022.cl"},
	{0x42C51E2A290C192EULL, "485423691.ogg"},
	{0x42C57DC8AC791D14ULL, "vxlt_gallowstemplar_43c_0_0.vlm"},
	{0x42C580FEAA20A524ULL, "vxlt_0_lowtown_n_f57_0_0.vlm"},
	{0x42C5C6B18BFC45D3ULL, "6091855_f.fxe"},
	{0x42C5D1CEB61D81D8ULL, "vxw_bonepit_10w_0_0.vlm"},
	{0x42C5F2C5FD94335EULL, "character_character_p106.psh.fxo"},
	{0x42C5FB553000B624ULL, "6149021_f.fxe"},
	{0x42C63A80935975D0ULL, "vxca_caverns1_mines_dp8_0_0.vlm"},
	{0x42C66E6128F0A686ULL, "6155748_m.fxe"},
	{0x42C6881BC1E0D087ULL, "genip_party_storage.ncs"},
	{0x42C6A92DED0D6374ULL, "vxlt_gallowstemplar_f_6s2_0_0.vlm"},
	{0x42C6B79F3BDBE228ULL, "vxht_chantry_2to_0_0.vlm"},
	{0x42C6C4D854A79A5AULL, "vxlt_0_lowtown_docks_a3_n_coh_0_0.vlm"},
	{0x42C7095814FCACDAULL, "mag360st_samson_cullen.stg"},
	{0x42C7C7EF4E1FD9EFULL, "dae000_hate_follower_react6167490.cl"},
	{0x42C7C92DF6E01F9FULL, "atl_prccontent_dxt5_02.dds"},
	{0x42C8096600CEA955ULL, "character_character_p248.psh.fxo.dx11.assembly"},
	{0x42C8C24CC0D5C73BULL, "vxht_mansion2_1bt_0_1.vlm"},
	{0x42C93E98896BE132ULL, "455310606.ogg"},
	{0x42C9C9AF354AD207ULL, "1064077558.ogg"},
	{0x42C9D9CF985ED8DDULL, "vxlt_gallowstemplar_f_4yk_0_0.vlm"},
	{0x42C9E7B52374B9CAULL, "6032050_m.fxe"},
	{0x42CA172BBC625262ULL, "character_character_p74.psh.fxo.dx11"},
	{0x42CABDD67C6387C9ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_6.dds"},
	{0x42CB2532B811136CULL, "art\\characters\\weapons\\axes\\w_axe_02a.msh"},
	{0x42CB2B32B8111D5EULL, "art\\characters\\weapons\\axes\\w_axe_02a.mmh"},
	{0x42CB3732B81131F5ULL, "art\\characters\\weapons\\axes\\w_axe_02a.mao"},
	{0x42CB7F02620597ECULL, "docks_male_overseer_mob6176557.cl"},
	{0x42CC8C2300A2CF33ULL, "mag311_emile_de_launcet6064728.cl"},
	{0x42CCD2431DBB9EC0ULL, "mrl150cr_dalish_guard1.utc"},
	{0x42CCDDDCA8EC7171ULL, "w_sundermount_46_0.mmh"},
	{0x42CCE9DCA8EC8512ULL, "w_sundermount_46_0.mao"},
	{0x42CCF7DCA8EC9D2FULL, "w_sundermount_46_0.msh"},
	{0x42CCFAB450AC06BFULL, "172188411.ogg"},
	{0x42CDB8DBE5E84967ULL, "6096326_f.fxe"},
	{0x42CE1D19337928B9ULL, "6072839_m.fxe"},
	{0x42CEF1C6C97EA84BULL, "vxht_brothel_n_61c_0_0.vlm"},
	{0x42CFB3E30D3148BFULL, "dae000_meredith6085267.cl"},
	{0x42D05B2F5289F53AULL, "zz_followers_banter6159444.cl"},
	{0x42D071A67658E81AULL, "6135671_f.fxe"},
	{0x42D0980662CB0029ULL, "t3_har_dbl.tnt"},
	{0x42D12A5F249E9A89ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2win.mao"},
	{0x42D1365F249EAF22ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2win.mmh"},
	{0x42D1385F249EB2D8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2win.msh"},
	{0x42D177AB524EF51DULL, "6071521_f.fxe"},
	{0x42D19260B0777F7DULL, "gen00fl_anders6125173.cl"},
	{0x42D19C0B8A479D22ULL, "prm000im_duru_amulet.uti"},
	{0x42D1B6D29E37104BULL, "vxw_encounter2_gv_0_0.vlm"},
	{0x42D1B7A97F6D5A51ULL, "lt_0_lowtown_docks_a3_n_180_0.msh"},
	{0x42D1C5A97F6D726CULL, "lt_0_lowtown_docks_a3_n_180_0.mao"},
	{0x42D1D1A97F6D860FULL, "lt_0_lowtown_docks_a3_n_180_0.mmh"},
	{0x42D1DF776ED2BF09ULL, "557086316.ogg"},
	{0x42D201ED055D544DULL, "t3_wdg_w26.tnt"},
	{0x42D245C84AEB50C8ULL, "6018517_m.fxe"},
	{0x42D24AFCDD0970C5ULL, "gen00fl_sebastian6174714.cl"},
	{0x42D25522D7E0A7EAULL, "gen00fl_anders6202354.cl"},
	{0x42D25B25FD0559C9ULL, "vxdr_0_deeproads_d3r_0_0.vlm"},
	{0x42D26334FE46AFB9ULL, "dae000_flemeth6000812.cl"},
	{0x42D2FF7F63D9C8F9ULL, "6051667_m.fxe"},
	{0x42D30B73EBAE5743ULL, "750078509.ogg"},
	{0x42D3B91A56F0ECEBULL, "6067798_m.fxe"},
	{0x42D3E3D67B19E1F4ULL, "6095248_m.fxe"},
	{0x42D443ABA8139A58ULL, "vxw_bonepit_15t_0_0.vlm"},
	{0x42D466F88323DEE8ULL, "6052951_m.fxe"},
	{0x42D46D3221F6CEA4ULL, "6142121_f.fxe"},
	{0x42D477E6E89CBDE0ULL, "vxht_mansion_n_5al_0_0.vlm"},
	{0x42D4F27614C4A601ULL, "hlac_ht_brothel_n_3371_0z.dds"},
	{0x42D51259DF5BC7C5ULL, "6121685_m.fxe"},
	{0x42D538C9C7A927BCULL, "vxw_woundedcoast_5n_0_0.vlm"},
	{0x42D5D733A6F0C7F8ULL, "vxw_encounter1_bg_0_0.vlm"},
	{0x42D6750DA4CD1F17ULL, "vrc221_bartrand6043388.cl"},
	{0x42D67DD43CA40B12ULL, "66764861.ogg"},
	{0x42D6E4D68073B12FULL, "hlac_lt_backalley_g_65536_1y.dds"},
	{0x42D71BBA42948579ULL, "hlac_w_woundedcoast_262148_0y.dds"},
	{0x42D73748319A020EULL, "dae000_hate_follower_react6187868.cl"},
	{0x42D74728D9CCFD0EULL, "seb321_leliana6180958.cl"},
	{0x42D7606F88C76A69ULL, "follower_banter6149297.cl"},
	{0x42D77BB6BAAC25ADULL, "qun160_vashothlead6064675.cl"},
	{0x42D7C899B4E3508BULL, "vxlt_undercityencounter_1dn_0_0.vlm"},
	{0x42D7D9FA323506FFULL, "6207780_f.fxe"},
	{0x42D7DAFFA8DB7C20ULL, "vxw_sundermount_2za_0_0.vlm"},
	{0x42D80191A7D942F9ULL, "vxht_mansion2b_211_0_0.vlm"},
	{0x42D8DD9A7E4525C3ULL, "dae000ip_crowd_em4_talk6.utp"},
	{0x42D92071A4AD1B1EULL, "6107234_f.fxe"},
	{0x42D927ACF5DD67FFULL, "hlac_lt_0_lowtown_docks_a3_-65538_0x.dds"},
	{0x42D958E039407CFAULL, "6147417_m.fxe"},
	{0x42D96A46326BB682ULL, "6170594_m.fxe"},
	{0x42D9AF1A9A89DD81ULL, "qun221_viscount6070788.cl"},
	{0x42DA034732EBDAE4ULL, "122069299.ogg"},
	{0x42DAA24F5AB47E24ULL, "vxca_caverns2_8fe_0_1.vlm"},
	{0x42DAB116169C2352ULL, "976032896.ogg"},
	{0x42DAB3E61C6F35A1ULL, "6117976_m.fxe"},
	{0x42DADB2CBC0D06EFULL, "hlac_ca_caverns2prc_5802_58x.dds"},
	{0x42DB0FC33FCBB950ULL, "6108559_m.fxe"},
	{0x42DB299E37FD300FULL, "6045242_m.fxe"},
	{0x42DB7072D2AC29E0ULL, "hlac_lt_0_lowtown_docks_a3_0_6y.dds"},
	{0x42DB96B07E571713ULL, "vxht_mansion2b_1ds_0_0.vlm"},
	{0x42DBA2130CC32168ULL, "vxht_mansion2_1rr_0_0.vlm"},
	{0x42DBAB743703C00EULL, "6055605_f.fxe"},
	{0x42DBB650D7FD9528ULL, "209524833.ogg"},
	{0x42DC14293C0D6DE3ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_aravell_sa.dds"},
	{0x42DC81AB7F4D88DDULL, "vxlt_gallowscourtyard_n_74u_0_0.vlm"},
	{0x42DC9E130F6E905DULL, "6208335_m.fxe"},
	{0x42DCB208DC7BEAEDULL, "llac_lt_ruinedhouse_n_-1_7.dds"},
	{0x42DCC159569E3F56ULL, "6125740_f.fxe"},
	{0x42DCEB4C4BD7E4AAULL, "6174038_f.fxe"},
	{0x42DD135EF0B72AA7ULL, "vxca_caverns2prc_ac8_0_0.vlm"},
	{0x42DD3F923FE80914ULL, "vxlt_0_lowtown_docks_as4_0_0.vlm"},
	{0x42DD52EE96279E42ULL, "6138745_m.fxe"},
	{0x42DDA6B07360BF49ULL, "framebuffereffects_composedof_p0.psh.fxo.dx11.assembly"},
	{0x42DE7ED8DFC10580ULL, "6057354_m.fxe"},
	{0x42DE9E3D919832FDULL, "ser270_alley.lst"},
	{0x42DFF2686757D25FULL, "vxw_encounter2_90_0_0.vlm"},
	{0x42E0174DEA2094DCULL, "darktown_fem_lowlife_mob6174157.cl"},
	{0x42E0B97C55ABBCE3ULL, "vxw_bonepitmining_a2_1og_0_0.vlm"},
	{0x42E15FCEDA1F1DDDULL, "rog_blendedmodes.dds"},
	{0x42E1B05476DEAC61ULL, "6111673_m.fxe"},
	{0x42E1B6C065AF8E7CULL, "vxht_brothel_n_6j1_0_0.vlm"},
	{0x42E1C06222CC81BFULL, "235785432.ogg"},
	{0x42E285DA7663C457ULL, "6101814_m.fxe"},
	{0x42E290D6122E7FCCULL, "67989339.ogg"},
	{0x42E334927F86A99CULL, "lt_0_lowtown_docks_104_0.phy"},
	{0x42E344ADCCF6A56AULL, "vxlt_0_lowtown_a3_n_c_eki_0_2.vlm"},
	{0x42E3668121E098E5ULL, "vxlt_0_lowtown_docks_n_b7x_0_1.vlm"},
	{0x42E3939B6FC761DCULL, "hlac_lt_undercitydungeon_-1_17y.dds"},
	{0x42E3AAA6B99F403CULL, "vxw_woundedcoast_2o_0_0.vlm"},
	{0x42E437F170E771BDULL, "hlac_lt_0_lowtown_a2_n_c_0_6y.dds"},
	{0x42E4491098E6E078ULL, "vxw_encounter1_gd_0_0.vlm"},
	{0x42E4B5B70C9F625BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgte_0d.dds"},
	{0x42E55BEDB7D2AE4DULL, "vxht_brothel_n_5du_0_0.vlm"},
	{0x42E55C65B9211837ULL, "6175769_m.fxe"},
	{0x42E5675B5DA1BA4FULL, "463270111.ogg"},
	{0x42E5D76BC94A60C6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasea_n.dds"},
	{0x42E5F9D8C295B4E3ULL, "6020456_m.fxe"},
	{0x42E678FF6AF359D0ULL, "6075075_m.fxe"},
	{0x42E7167378D97A60ULL, "vxw_bonepitdestroyed_a3_257_0_0.vlm"},
	{0x42E773CF38674AC5ULL, "ser270_templar6127486.cl"},
	{0x42E78B19C809124DULL, "mag311cr_huon.utc"},
	{0x42E7BE43DF7A899CULL, "genpt_dog.ncs"},
	{0x42E7D6303E483F6CULL, "hlac_ht_mansion2_818_62z.dds"},
	{0x42E8E50F77F870BFULL, "6071196_f.fxe"},
	{0x42EA13EFFEB939E8ULL, "vxht_mansion2_1us_0_0.vlm"},
	{0x42EB4AC72A82DE2BULL, "hairalpha_hair_p0.psh.fxo.dx11"},
	{0x42EB87E611613DA2ULL, "llac_dr_0_deeproads_5842_10.dds"},
	{0x42EB9F9AD5E98A4AULL, "6126557_m.fxe"},
	{0x42EC10B4FFE9FB94ULL, "129467405.ogg"},
	{0x42EC9750F6817D42ULL, "145605429.ogg"},
	{0x42ECC4F7A9DF450BULL, "543718948.ogg"},
	{0x42ECF53F980D1934ULL, "713373439.ogg"},
	{0x42EE106D26A572CBULL, "vxht_0_hightown_a2_as2_0_0.vlm"},
	{0x42EE37CECCF2F55EULL, "646985693.ogg"},
	{0x42EEDAB11B20B1FBULL, "6174551_m.fxe"},
	{0x42EF04EEA7C71A61ULL, "175998987.ogg"},
	{0x42EF19AABCA9CC56ULL, "w_bonepit_148_0.mao"},
	{0x42EF25AABCA9E0F5ULL, "w_bonepit_148_0.mmh"},
	{0x42EF27AABCA9E46BULL, "w_bonepit_148_0.msh"},
	{0x42EF61A9E071BD99ULL, "llac_dr_primevalentrance_1726_5.dds"},
	{0x42F018A4276E83BCULL, "gen00fl_isabela6088354.cl"},
	{0x42F0467F57F63572ULL, "zz_dae_debug6143034.cl"},
	{0x42F0E98EAF0F32C4ULL, "hla_ht_0_hightown_a3_-2_0z.dds"},
	{0x42F1129B5D99AF99ULL, "vxca_caverns1_e81_0_0.vlm"},
	{0x42F12E34E0ED7CD1ULL, "dr_primevaltomb_658_ll.mmh"},
	{0x42F26CC6F8825E25ULL, "hlac_lt_0_lowtown_a3_n_c_65534_6y.dds"},
	{0x42F2B66621ED86DCULL, "515035924.ogg"},
	{0x42F31F498EB95298ULL, "vxlt_0_lowtown_a2_hgh_0_0.vlm"},
	{0x42F3EA773ACA6CE6ULL, "6096463_m.fxe"},
	{0x42F4DDCAEEF67AFCULL, "6082194_m.fxe"},
	{0x42F53450F98F14B1ULL, "mag331ip_hideout_exit.utp"},
	{0x42F561236E96416BULL, "54072685.ogg"},
	{0x42F57EFFFF296D14ULL, "llac_lt_0_lowtown_-131073_3.dds"},
	{0x42F588506ACF92E0ULL, "vxw_bonepitdestroyed_a3_204_0_0.vlm"},
	{0x42F6179B0D0ABD91ULL, "432413334.ogg"},
	{0x42F680BA6C969213ULL, "6094958_m.fxe"},
	{0x42F68F303146C4ECULL, "vxca_cavern1_haunted_djf_0_0.vlm"},
	{0x42F69BBBE3D21BFAULL, "528277337.ogg"},
	{0x42F6E12472E864D5ULL, "dae200_bran6094802.cl"},
	{0x42F725EED000FC6EULL, "vxw_sundermount_357_0_0.vlm"},
	{0x42F7A9B56A0DCD2FULL, "2893834.ogg"},
	{0x42F7AE651A69B940ULL, "vxlt_0_lowtown_docks_n_aow_0_0.vlm"},
	{0x42F83EF3DB11EC6AULL, "6175517_m.fxe"},
	{0x42F864E0D5EBBFBCULL, "hla_ht_0_hightown_-131070_0y.dds"},
	{0x42F885CCF0AF5268ULL, "vxht_mansion2_1xp_0_0.vlm"},
	{0x42F8871CBD161EF3ULL, "gen00fl_varric6093650.cl"},
	{0x42F8A05D7408AA19ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint_6.pwk"},
	{0x42F8F4AABC8608F5ULL, "mer221_sabin6080529.cl"},
	{0x42F99E2E0F09A5A7ULL, "740619552.ogg"},
	{0x42FA053347B81B8BULL, "dbg_settag.ncs"},
	{0x42FA457DBCEC8645ULL, "cd_dae000_light_tombs_boss.bnk"},
	{0x42FA6C265F00CF03ULL, "1028428680.ogg"},
	{0x42FAC2D5E02BF824ULL, "vxlt_warehouse_n_10m_0_0.vlm"},
	{0x42FB138B87EA8322ULL, "art\\vfx\\combat\\vfx_r_var_grenade_toss_c.mmh"},
	{0x42FBA4CF879502AEULL, "6099506_f.fxe"},
	{0x42FC3CD8B26F6BC6ULL, "vxdr_primevaltomb_505_0_2.vlm"},
	{0x42FC824A189B8B4BULL, "vxht_0_hightown_a2_an5_0_0.vlm"},
	{0x42FCBBBFBEAB608FULL, "w_encounter2_camp_20_0_ll.mmh"},
	{0x42FD872CB2A486ABULL, "vxlt_gallowscourtyard_63t_0_0.vlm"},
	{0x42FDC94DC390956FULL, "290333751.ogg"},
	{0x42FDE1ACBD7CD511ULL, "6208001-sisterwasincircle.cut"},
	{0x42FE081A0E37D38FULL, "6092559_f.fxe"},
	{0x42FE4BBE87C9F7DFULL, "vxw_sundermount_2bj_0_0.vlm"},
	{0x42FF8E412DCE7BE5ULL, "6042519_m.fxe"},
	{0x42FF8FB8F517E2B2ULL, "llac_ca_caverns2_5802_5.dds"},
	{0x42FFBC2434529C96ULL, "vxdr_primevaltomb_5uh_0_0.vlm"},
	{0x42FFE489DFF450E8ULL, "327633091.ogg"},
	{0x43002656757AEA63ULL, "6075155_m.fxe"},
	{0x43007B0F2A3EBDCEULL, "6091682_m.fxe"},
	{0x4300E7F3D7066B9FULL, "gen00fl_player6207745.cl"},
	{0x43010891B9F23BB2ULL, "33274278.ogg"},
	{0x43011B4D398A7EF0ULL, "llac_ca_caverns1_mines_24426_2.dds"},
	{0x4301EDAE392D3314ULL, "1011174036.ogg"},
	{0x4301F7BD5D116D1EULL, "439412416.ogg"},
	{0x430267BFD27B9C91ULL, "vxht_bartrand_n_5qr_0_0.vlm"},
	{0x4302D081F15539CFULL, "858287793.ogg"},
	{0x4302DB2BC2CD452BULL, "vxlt_0_lowtown_a3_n_c_fg5_0_0.vlm"},
	{0x4303901B31831AB1ULL, "hlac_ht_0_hightown_n_-131073_1z.dds"},
	{0x43039BCD70B7D43BULL, "vxca_caverns1_gj3_0_0.vlm"},
	{0x43040817DC222C58ULL, "vxca_caverns1_e7q_0_0.vlm"},
	{0x43041CE809A95BE8ULL, "w_bonepitmining_a2_219_0.msh"},
	{0x43042EE809A97A79ULL, "w_bonepitmining_a2_219_0.mao"},
	{0x43043AE809A98E12ULL, "w_bonepitmining_a2_219_0.mmh"},
	{0x43044DCC331991B9ULL, "527552654.ogg"},
	{0x43049D12FA200FC6ULL, "hlac_ht_brothel_n_3382_1x.dds"},
	{0x43050D7E96ED9AC3ULL, "vxlt_backalley_g_1n1_0_1.vlm"},
	{0x4305501B06371986ULL, "hlac_lt_0_lowtown_docks_a3_0_1z.dds"},
	{0x43058E05CF2334B3ULL, "hlac_ht_mansion2b_818_8y.dds"},
	{0x4305964036A62F55ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\prp_zzhumanbody_d.dds"},
	{0x4305FDF8A67AEFE9ULL, "hlac_lt_0_lowtown_a2_65536_9z.dds"},
	{0x43065169A129569AULL, "761236759.ogg"},
	{0x43067AEF98B85C0AULL, "211711122.ogg"},
	{0x4306BEF69B666CE1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_roba_0n.dds"},
	{0x4306EC4754396093ULL, "vxht_mansion2b_1ut_0_0.vlm"},
	{0x4306F7A9E2A56AE8ULL, "vxht_mansion2_1cq_0_0.vlm"},
	{0x4307019492B6B56EULL, "299780124.ogg"},
	{0x430705B88F366FFAULL, "llac_ht_0_hightown_n_65533_2.dds"},
	{0x43071F5089B96C7BULL, "859862571.ogg"},
	{0x43072F5D5EA59752ULL, "6139914_f.fxe"},
	{0x4307B41A0EF3DAC3ULL, "6015932_f.fxe"},
	{0x43082834562218D4ULL, "zz_isa_debug6133148.cl"},
	{0x43088354B9D5BB4AULL, "6128757_m.fxe"},
	{0x430885D26F66A41BULL, "hlac_lt_0_lowtown_docks_-131076_0x.dds"},
	{0x4308BF4FD018A875ULL, "qun131_warrens_thug6057192.cl"},
	{0x4308E0731DF24860ULL, "hlac_ca_caverns1_mines_24423_1x.dds"},
	{0x43091FCA248F8A26ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_pc_.ani"},
	{0x43092B8594B91C5CULL, "6085858_m.fxe"},
	{0x4309725541795110ULL, "6040700_f.fxe"},
	{0x4309ACB828D67133ULL, "dae010_alienage_elf_m2.utc"},
	{0x4309BF5B062DDEA9ULL, "hlac_w_bonepitmining_a2_262150_0y.dds"},
	{0x4309F2948FEC5C2AULL, "and111_lirene6169994.cl"},
	{0x430A1F7A1F83DE7DULL, "vxlt_0_lowtown_a3_n_c_e23_0_0.vlm"},
	{0x430A4A59C64762FFULL, "hlac_lt_0_lowtown_docks_a3_n_-3_5z.dds"},
	{0x430A8261C741B51FULL, "6119305_m.fxe"},
	{0x430AA73D91BD7CDAULL, "ser270_alley.are"},
	{0x430B0232B846BBF9ULL, "art\\characters\\weapons\\axes\\w_axe_02a.phy"},
	{0x430BFC0266ED0FDEULL, "vxdr_primevaltomb_6we_0_0.vlm"},
	{0x430C3D674EFB6837ULL, "art\\characters\\creatures\\dragon\\animation\\wkf.gad"},
	{0x430C559ADFBD156FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_lgta_0d.dds"},
	{0x430C59E5BCEDF20DULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpb.mmh"},
	{0x430C6BE5BCEE1093ULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpb.msh"},
	{0x430D58407AB6C497ULL, "fx_woodgrain.dds"},
	{0x430DD7A067085C40ULL, "plt_car100pt_gift.nss"},
	{0x430DEDFE6F3BB0D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_two_handed_ready.anb"},
	{0x430E2D7563B9A02BULL, "zz_followers_banter6157395.cl"},
	{0x430E41991D0D7177ULL, "mcr371_post_harvester.bnk"},
	{0x430E9F60073A7E87ULL, "6207772_m.fxe"},
	{0x430F46CF66C9AF82ULL, "6177919_m.fxe"},
	{0x430FC46DB9081720ULL, "hlac_ht_mansion_n_3367_1x.dds"},
	{0x430FE447E64297FEULL, "mcr310_confused_guard6077022.cl"},
	{0x430FEF8439A1F0A8ULL, "w_bonepit_163_0.mmh"},
	{0x430FFB8439A2054BULL, "w_bonepit_163_0.mao"},
	{0x43100D8439A223D2ULL, "w_bonepit_163_0.msh"},
	{0x4310485D9322633EULL, "llac_ht_0_hightown_n_1_1.dds"},
	{0x43108A6395BBE55DULL, "366503949.ogg"},
	{0x4310998F8BA8FBD6ULL, "vxw_blightlands_bo_0_0.vlm"},
	{0x43109BD2791F961FULL, "lt_0_lowtown_a2_n_c_44_0_ll.mmh"},
	{0x4310D0C9878AFCD8ULL, "vxlt_backalley_g_1u2_0_0.vlm"},
	{0x4310D99CC471B511ULL, "vxht_bartrand_n_5tq_0_0.vlm"},
	{0x4310E41867A7DEEBULL, "hlac_ht_playershouse_n_855_4z.dds"},
	{0x431123991D0F738FULL, "mcr371_post_harvester.cnv"},
	{0x43119469C2634A87ULL, "con_ico_diplomacy.dds"},
	{0x4311E98E5D537B40ULL, "6179560_m.fxe"},
	{0x431236B0E2E42C88ULL, "254087633.ogg"},
	{0x43123DA9B9A3B296ULL, "pro000_cut2d_voyage6004220.cl"},
	{0x431252674EFFF7C4ULL, "art\\characters\\creatures\\dragon\\animation\\wkf.evt"},
	{0x431274FD9FA9679FULL, "qun221_viscount6069230.cl"},
	{0x43127506086726C3ULL, "art\\vfx\\combat\\vfx_d_teleport_start_p.impa.evt"},
	{0x431307AD4629C3BEULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.dura.evt"},
	{0x43140EB14F9FC5EDULL, "vxw_sundermount_34s_0_0.vlm"},
	{0x43144AD2D24FDE42ULL, "hlac_dr_primevaltomb_658_5x.dds"},
	{0x4314E2AA46211E8EULL, "620602389.ogg"},
	{0x4314E39254698C1DULL, "559465568.ogg"},
	{0x43156986D49B8368ULL, "vxht_mansion2_1fn_0_0.vlm"},
	{0x43167BB7AF242C4FULL, "161618895.ogg"},
	{0x4316C63B0F33EE4BULL, "954283561.ogg"},
	{0x4316FFB2CB21A521ULL, "6077335_m.fxe"},
	{0x43177269297188DBULL, "vxca_cavern1_haunted_cnz_0_0.vlm"},
	{0x4317B110C81E02FAULL, "ser270_templar6127403.cl"},
	{0x4318229BB98AFAEEULL, "912137951.ogg"},
	{0x4318D833F29C0242ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right_2.gad"},
	{0x4318F43EB937CD9FULL, "6114405_m.fxe"},
	{0x43190FB3DA2B507BULL, "ave310_donnic6134728.cl"},
	{0x43194E06086D0344ULL, "art\\vfx\\combat\\vfx_d_teleport_start_p.impa.gad"},
	{0x4319819022D19E97ULL, "hlac_ht_0_hightown_n_-65535_0z.dds"},
	{0x431A1EAD463009E9ULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.dura.gad"},
	{0x431A2AF9AED7C76FULL, "gen00fl_varric6119008.cl"},
	{0x431A6D89A1B3E844ULL, "6005822_f.fxe"},
	{0x431AAE698E20E71FULL, "vxlt_0_lowtown_a2_n_c_i0n_0_0.vlm"},
	{0x431B12D2A8082A7CULL, "ht_0_hightown_a2.anb"},
	{0x431B1ED2A8083EEEULL, "ht_0_hightown_a2.arl"},
	{0x431B751BC228F0DAULL, "genip_party_storage.utp"},
	{0x431BF4B688F57479ULL, "nat380_sibling6207665.cl"},
	{0x431C7011DEE9C4A1ULL, "385452848.ogg"},
	{0x431CC1DF0191523DULL, "vxlt_undercityserialkiller_157_0_0.vlm"},
	{0x431D08CA24A00C20ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_pc_.gad"},
	{0x431D52E5BCFC59CAULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpb.phy"},
	{0x431D7288A36B4E0CULL, "llac_dr_0_deeproads_5838_2.dds"},
	{0x431DB8254BA725E2ULL, "vxlt_0_lowtown_a3_n_c_i0n_0_0.vlm"},
	{0x431DCA5BDD3A348EULL, "llac_ht_mansion2_818_15.dds"},
	{0x431DF5D402EC2CC3ULL, "100479565.ogg"},
	{0x431F5D0608718822ULL, "art\\vfx\\combat\\vfx_d_teleport_start_p.impa.ani"},
	{0x431FFB33F2A25C39ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right_2.evt"},
	{0x43206A674F0C5DBDULL, "art\\characters\\creatures\\dragon\\animation\\wkf.ani"},
	{0x4320A2A73F5C565EULL, "gen00fl_bethany6083991.cl"},
	{0x4320EFAD4635D837ULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.dura.ani"},
	{0x432163A3D61245B9ULL, "hlac_w_encounter3_-327676_0z.dds"},
	{0x4321675C168538AAULL, "lt_undercitydungeon_3567_ll.mmh"},
	{0x43216D49FC1DF3E6ULL, "745534314.ogg"},
	{0x43218E2726CB717AULL, "vxca_cavern1_haunted_jl8_0_0.vlm"},
	{0x4321A4D795BA9AF8ULL, "186104880.ogg"},
	{0x432201FFA5FFC58BULL, "851922716.ogg"},
	{0x4322808E4195DE6DULL, "vxw_sundermount_33r_0_0.vlm"},
	{0x43228C8DA1C4B5F6ULL, "mag121_idunna6083961.cl"},
	{0x4322D24C7B1DE792ULL, "hlac_w_woundedcoast_-262143_0z.dds"},
	{0x4322DBA01482F8A4ULL, "vxca_caverns2_8yh_0_1.vlm"},
	{0x43230E06CADDB2EEULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardazed.evt"},
	{0x432317ADEE30D27AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_avea_0n.dds"},
	{0x4323397E1C9A11A8ULL, "6060183_m.fxe"},
	{0x4323461A739B03B8ULL, "gen00fl_isabela6090247.cl"},
	{0x4323C3528F103009ULL, "art\\characters\\playercharacter\\qunarimale\\eye_stn.mao"},
	{0x4323C672A787E71BULL, "40610671.ogg"},
	{0x4323DB63C6919BE8ULL, "vxht_mansion2_1io_0_0.vlm"},
	{0x4323F41DA5255536ULL, "6028875_m.fxe"},
	{0x4324128B880D86C9ULL, "art\\vfx\\combat\\vfx_r_var_grenade_toss_c.anb"},
	{0x432442005620D9D6ULL, "vxdr_0_deeproads_cf0_0_1.vlm"},
	{0x432541C9748687AEULL, "character_character_p193.psh.fxo.dx11.assembly"},
	{0x43258EB9BF9B86D1ULL, "746006024.ogg"},
	{0x4325E3962F274855ULL, "lgt110cr_arl_connor.utc"},
	{0x4325E4EB1F5E2230ULL, "mrl150_merrill6070985.cl"},
	{0x432606343A7C1C0BULL, "6167029_m.fxe"},
	{0x43262C6111CB94BDULL, "lowtown_fem_alienage1_sta6161575.cl"},
	{0x4326421EE36889A8ULL, "6118327_m.fxe"},
	{0x4326CCE6E2FD02AFULL, "284470666.ogg"},
	{0x4326DCBC13ECF1EBULL, "6167674_m.fxe"},
	{0x4326DF150BB6F0C6ULL, "490332579.ogg"},
	{0x4327033403700F7DULL, "vxlt_0_lowtown_a3_n_c_e81_0_0.vlm"},
	{0x432740BCE41D0288ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_seaway.msh"},
	{0x432752BCE41D2119ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_seaway.mao"},
	{0x43275EBCE41D35B2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_seaway.mmh"},
	{0x4327661BAB2DE61FULL, "6113505_m.fxe"},
	{0x43278DAFA5C919CDULL, "gen_im_trs_dal_med_cht_r1.uti"},
	{0x4327B65D7BEC7E3CULL, "6099244_m.fxe"},
	{0x4327DA119984BFAEULL, "zz_followers_banter6157709.cl"},
	{0x4327ED22FA805658ULL, "mcr310_confused_guard.cnv"},
	{0x43282C2542CD6545ULL, "gen00fl_carver6102552.cl"},
	{0x43283AD90517EE56ULL, "hlac_lt_0_lowtown_docks_n_-196610_6y.dds"},
	{0x4328F2CD0F7A3763ULL, "vxw_bonepitmining_a2_1pj_0_0.vlm"},
	{0x43290506CAE20E99ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardazed.gad"},
	{0x43291CF244AFB95BULL, "127677021.ogg"},
	{0x4329603A9D37B547ULL, "695012651.ogg"},
	{0x4329ACBC82A98CD4ULL, "lt_0_lowtown_docks_a3_107_0.phy"},
	{0x432A143DE685AB3CULL, "ser221_party_comment_2.stg"},
	{0x432A59B87E1C037BULL, "mag371_elsa6041264.cl"},
	{0x432A6D0768E741A0ULL, "6209976_f.fxe"},
	{0x432A7975B9885CEEULL, "art\\vfx\\combat\\vfx_c_debili_cast_aoe_c.mmh"},
	{0x432A85384D83DD12ULL, "hlac_ht_chantry_2534_32x.dds"},
	{0x432A997EC306730EULL, "6072713_m.fxe"},
	{0x432B08F379C13BE6ULL, "6067737_m.fxe"},
	{0x432BCF22FA840C60ULL, "mcr310_confused_guard.bnk"},
	{0x432C0A17E1CA0F3CULL, "71819497.ogg"},
	{0x432C6E46D9AA4376ULL, "6121660_m.fxe"},
	{0x432CFCED22635DECULL, "character_character_p50.psh.fxo.dx11"},
	{0x432D0333F2ACF428ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_right_2.ani"},
	{0x432D96938CC17C08ULL, "art\\vfx\\textures\\vfx_crack_decal.dds"},
	{0x432DBFD43C7E7A66ULL, "vfx_rendertoshadowmap_p0.vsh.fxo.dx11"},
	{0x432DE24D3BFD1A52ULL, "316086473.ogg"},
	{0x432E18B0B8C5F3FEULL, "dae000_bodhan6146882.cl"},
	{0x432E1A31A38AEDB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v2.ani"},
	{0x432E30AEC679C3A2ULL, "6196690_f.fxe"},
	{0x432E72ECC5E4070FULL, "772544857.ogg"},
	{0x432E777C1B225ED6ULL, "hlac_w_bonepitdestroyed_a3_0_10y.dds"},
	{0x432E7BA621F80739ULL, "hlac_ht_mansion2_818_4y.dds"},
	{0x432E9FF563119F18ULL, "118465458.ogg"},
	{0x432EBE8812EC72E4ULL, "pro000_athenril6048744.cl"},
	{0x432F037872A76896ULL, "784635847.ogg"},
	{0x432F2571CB13AD6EULL, "6094731_m.fxe"},
	{0x432F63967C30F24BULL, "508814296.ogg"},
	{0x432F8ADF913F5000ULL, "vxht_0_hightown_n_c3m_0_1.vlm"},
	{0x433006547F0F158BULL, "6207004_m.fxe"},
	{0x43301606CAE84AE7ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardazed.ani"},
	{0x4330A0CDEDC36361ULL, "6169542_m.fxe"},
	{0x4330F090276A455FULL, "201886370.ogg"},
	{0x433104878752A65DULL, "dae300_cullen6140621.cl"},
	{0x433170FE3E279399ULL, "144996978.ogg"},
	{0x4331E583B609E45BULL, "505845390.ogg"},
	{0x43324C32B2A04860ULL, "llac_lt_0_lowtown_a2_n_65538_3.dds"},
	{0x4332C01B517D3806ULL, "hlac_ca_caverns1_mines_24423_0y.dds"},
	{0x4332E97EB8597F74ULL, "65499102.ogg"},
	{0x43336FF8A63C08BCULL, "vxca_caverns1_e8h_0_0.vlm"},
	{0x43337D5974E13966ULL, "547792301.ogg"},
	{0x433417B2704BCF8DULL, "446677595.ogg"},
	{0x433489BC82B358D1ULL, "lt_0_lowtown_docks_a3_107_0.msh"},
	{0x433497BC82B370ECULL, "lt_0_lowtown_docks_a3_107_0.mao"},
	{0x4334A3BC82B3848FULL, "lt_0_lowtown_docks_a3_107_0.mmh"},
	{0x4334B40A9F83EFA0ULL, "6176719_f.fxe"},
	{0x4334F2A528BFEA27ULL, "vxlt_0_lowtown_a3_n_c_hlj_0_0.vlm"},
	{0x43352531A3911F53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v2.gad"},
	{0x43352AB02C22D4C9ULL, "dae000_meredith6092557.cl"},
	{0x43354B841E0AC7FAULL, "6032138_f.fxe"},
	{0x4335A7FB7ADE1A9BULL, "dre000_varric6034654.cl"},
	{0x4335B81E39BEBE09ULL, "w_bonepitdestroyed_a3_226_0.mao"},
	{0x4335C41E39BED2A2ULL, "w_bonepitdestroyed_a3_226_0.mmh"},
	{0x4335C61E39BED658ULL, "w_bonepitdestroyed_a3_226_0.msh"},
	{0x4336A7BAE67E1897ULL, "35601596.ogg"},
	{0x4336DA06A26B5A09ULL, "hlac_w_sundermount_131067_0x.dds"},
	{0x4336FA992D4000E9ULL, "dae000_templar_recruit_16164436.cl"},
	{0x43378C06011564E6ULL, "and111_lirene6084516.cl"},
	{0x43379875B9931BEDULL, "art\\vfx\\combat\\vfx_c_debili_cast_aoe_c.anb"},
	{0x4338374CD4C537CDULL, "946499321.ogg"},
	{0x43384D8F3D3C2D07ULL, "follower_banter6161231.cl"},
	{0x43385E3EE394B545ULL, "llac_lt_hangedman_n_-1_14.dds"},
	{0x4339252511EE53D6ULL, "719261282.ogg"},
	{0x4339F6C81FB89DC0ULL, "vxlt_0_lowtown_a2_n_eix_0_0.vlm"},
	{0x433AD72C06E614FDULL, "736087246.ogg"},
	{0x433ADAC64C4AD5A6ULL, "vxlt_backalley_n_1ye_0_0.vlm"},
	{0x433B05C6DBAB698AULL, "6180391_m.fxe"},
	{0x433B2231A39585B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v2.evt"},
	{0x433B47C6BF913B63ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03a.mmh"},
	{0x433B55C6BF9152BDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03a.msh"},
	{0x433BBB4151E2B165ULL, "qun120_arishok6049654.cl"},
	{0x433C3E1220169A2AULL, "vxlt_gallowstemplar_3ja_0_0.vlm"},
	{0x433C8FE4877E0061ULL, "mag320_orsino6071219.cl"},
	{0x433CD669F095F86BULL, "6083943_m.fxe"},
	{0x433DB8A1C9A1E278ULL, "vxw_bonepitdestroyed_a3_23u_0_0.vlm"},
	{0x433DE4862892BF0FULL, "art\\characters\\weapons\\bolts\\w_blt_hlk01.mao"},
	{0x433DF0862892D3ACULL, "art\\characters\\weapons\\bolts\\w_blt_hlk01.mmh"},
	{0x433DF6862892DD9EULL, "art\\characters\\weapons\\bolts\\w_blt_hlk01.msh"},
	{0x433EEF552C1C28A7ULL, "6038673_m.fxe"},
	{0x433EFC15DF05F6C6ULL, "970077822.ogg"},
	{0x433F1182B6E5F10EULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.impa.ani"},
	{0x433F289C7A07F473ULL, "vxht_keep_49e_0_0.vlm"},
	{0x433FD8BB2A307D3DULL, "704415729.ogg"},
	{0x434042FF53BCA68FULL, "357171313.ogg"},
	{0x4340DA2CF3CCB7B5ULL, "597341204.ogg"},
	{0x43410C86B30AFC04ULL, "123008721.ogg"},
	{0x43412A9F406F2D66ULL, "43316485.ogg"},
	{0x434167406473C23BULL, "dae000_mother6035919.cl"},
	{0x43418D1A041E9E87ULL, "dae000_hate_follower_react6103229.cl"},
	{0x4341D4FF27CB343FULL, "gen00fl_aveline6120759.cl"},
	{0x4341DC92260FD1B6ULL, "6128003_m.fxe"},
	{0x43424D61F988D6ACULL, "vxw_woundedcoast_hi_0_0.vlm"},
	{0x43426ED58166922BULL, "289916930.ogg"},
	{0x4342CDE12CE440A8ULL, "vxca_cavern1_haunted_j25_0_1.vlm"},
	{0x4343137FCB24295FULL, "gen00fl_varric6076435.cl"},
	{0x434318D2D8297C7EULL, "art\\vfx\\combat\\vfx_r_gn3_repulsion_aoe_p.impa.ani"},
	{0x43433BEC9B60CBE3ULL, "dr_primevaltomb_657_lc.mmh"},
	{0x4343E960A2B3AE69ULL, "hlac_lt_0_lowtown_n_0_7z.dds"},
	{0x4343F6348C8D60E8ULL, "hlac_lt_0_lowtown_a2_n_65536_7x.dds"},
	{0x4345194D4026AF53ULL, "vxlt_gallowstemplar_n_5ix_0_0.vlm"},
	{0x4345AE494FD1761BULL, "6127560_m.fxe"},
	{0x4345FB5C14036EE1ULL, "6088459_f.fxe"},
	{0x43460B685B2BA54EULL, "6067636_m.fxe"},
	{0x434617F94410C6A0ULL, "674674114.ogg"},
	{0x4346E3045AC531BBULL, "vxlt_darktown_2s6_0_0.vlm"},
	{0x43476EC061422D99ULL, "lowtown_male_alienage1_sta6162809.cl"},
	{0x43483B2B90D5F5E0ULL, "hlac_w_woundedcoast_2_0y.dds"},
	{0x4348AA8AE396F0F0ULL, "vxlt_warehouse_po_0_0.vlm"},
	{0x4348D27A2AB34E52ULL, "vxlt_undercityserialkiller_1ii_0_1.vlm"},
	{0x4348D914D577874AULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench2_sebastianintro.gad"},
	{0x4348F045F0F7122EULL, "llac_w_woundedcoast_65538_0.dds"},
	{0x434936ED39542A46ULL, "gen00fl_carver6103392.cl"},
	{0x434977A3CDA1820AULL, "6185091_m.fxe"},
	{0x434982B92A6233B0ULL, "hlac_ht_keep_4519_5y.dds"},
	{0x43498BCB35DA3718ULL, "vxca_caverns2_842_0_0.vlm"},
	{0x4349A45948536D6CULL, "gen00fl_aveline6139460.cl"},
	{0x4349AFE31B595D50ULL, "194253668.ogg"},
	{0x4349F24FCEB4D813ULL, "533742564.ogg"},
	{0x4349F3502070143FULL, "vxht_mansion_n_272_0_0.vlm"},
	{0x434A7C42DF35F0C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_up.gad"},
	{0x434AA1108C4A1711ULL, "vxht_bartrand_n_5hu_0_0.vlm"},
	{0x434AAFEF120CB2AAULL, "vxlt_gallowstemplar_3gb_0_0.vlm"},
	{0x434AF8913A093393ULL, "22879315.ogg"},
	{0x434B47E3157F1663ULL, "478295428.ogg"},
	{0x434B575002105CB6ULL, "vrc221_varric6040871.cl"},
	{0x434B62F38D5DFC42ULL, "docks_male_overseer_mob6176573.cl"},
	{0x434B834365C31DD6ULL, "vxw_sundermount_2av_0_0.vlm"},
	{0x434BA49695D253A6ULL, "hlac_ht_mansion2b_818_20x.dds"},
	{0x434BCC4A0F51AFFBULL, "vxca_cave_1o6_0_0.vlm"},
	{0x434BD0A0673DB103ULL, "plt_car100pt_gift.plo"},
	{0x434BDAFFDDEC002EULL, "421298222.ogg"},
	{0x434BE3B00AE439E0ULL, "vxw_encounter2_camp_6b_0_0.vlm"},
	{0x434C072AE4A81779ULL, "llac_lt_0_lowtown_docks_a3_-65537_5.dds"},
	{0x434C0982B6F06E5FULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.impa.evt"},
	{0x434C48C6BF9FAF88ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw03a.phy"},
	{0x434CAF4479BE3B12ULL, "art\\vfx\\materials\\vfx_markofdeath.mao"},
	{0x434D0BEC23C62233ULL, "pro000_cut2d_thefirstyear.bnk"},
	{0x434D18E595712F71ULL, "zz_followers_banter6158550.cl"},
	{0x434D40B0143818DFULL, "fen210_hadriana6110572.cl"},
	{0x434D50D323E14ED4ULL, "dae150ip_barrier.ncs"},
	{0x434D5B7CF246A9CDULL, "625493787.ogg"},
	{0x434DAE8F1CEB4D88ULL, "6040938_m.fxe"},
	{0x434DE24FA8189C3AULL, "art\\characters\\playercharacter\\humanmale\\glv_masa.mmh"},
	{0x434DE64FA818A371ULL, "art\\characters\\playercharacter\\humanmale\\glv_masa.mao"},
	{0x434DF44FA818BA80ULL, "art\\characters\\playercharacter\\humanmale\\glv_masa.msh"},
	{0x434E4B5A7D7CC48EULL, "qcr000_aveline6041080.cl"},
	{0x434E571E1C2B6ECEULL, "instancing_depthwrite.vsh.fxo"},
	{0x434E963E9ABC1551ULL, "mer210_smith6051402.cl"},
	{0x434E9F5F8289DCC9ULL, "vxht_0_hightown_a3_c9u_0_0.vlm"},
	{0x434F27DABD313A50ULL, "6120786_m.fxe"},
	{0x434F6A384ABECDFBULL, "vxlt_0_lowtown_h9u_0_0.vlm"},
	{0x43502DEC23C89183ULL, "pro000_cut2d_thefirstyear.cnv"},
	{0x43502FC62257B29DULL, "plt_lgt000pt_ultra_4_06.plo"},
	{0x435037691B6611F5ULL, "559095523.ogg"},
	{0x4350F28B4B2584CBULL, "vxht_brothel_n_6bz_0_0.vlm"},
	{0x435130D2D835E24FULL, "art\\vfx\\combat\\vfx_r_gn3_repulsion_aoe_p.impa.evt"},
	{0x435135722DC2C9ACULL, "593460256.ogg"},
	{0x43517142DF3BFCD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_up.evt"},
	{0x435171A3F781FB53ULL, "6130133_m.fxe"},
	{0x4351C6D283577F06ULL, "vxht_chantry_n_27g_0_0.vlm"},
	{0x4351DFF9E7CD751AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_medium.gad"},
	{0x43522EC12C59591EULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_mother_andersintro.ani"},
	{0x4352A8F234090255ULL, "6095400_f.fxe"},
	{0x4352EA82B6F65848ULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.impa.gad"},
	{0x43535FE3177FA15DULL, "61362489.ogg"},
	{0x43540A77681A8123ULL, "lt_0_lowtown_docks_n_38_0_ll.mmh"},
	{0x43540A8A699983B2ULL, "gen00fl_aveline6053331.cl"},
	{0x4355083CEC2AE8DCULL, "qun220cr_javaris.cnv"},
	{0x43551F56E0E1B12DULL, "356804762.ogg"},
	{0x4355638DA2709E90ULL, "mer221ar_player_home.lst"},
	{0x43564E03B8D38DD0ULL, "vxht_mansion2_2bu_0_0.vlm"},
	{0x43566143017B296EULL, "lt_0_lowtown_docks_a3_19_0_ll.mmh"},
	{0x435676AA22C016E9ULL, "6133810_f.fxe"},
	{0x43570C588A590F91ULL, "hlac_ca_caverns2_5802_5y.dds"},
	{0x43571697CCCA9923ULL, "vxdr_primevaltomb_7yk_0_0.vlm"},
	{0x43571C67D58D0970ULL, "vxlt_warehouse_ul_0_0.vlm"},
	{0x435732C6225DD6CAULL, "plt_lgt000pt_ultra_4_06.nss"},
	{0x4357BB0D2F36EA12ULL, "vxlt_0_lowtown_n_fim_0_0.vlm"},
	{0x4357DE338AEE6423ULL, "hlac_lt_darktown_-1_58z.dds"},
	{0x4357FDA827D04F98ULL, "vxca_caverns2_833_0_0.vlm"},
	{0x43582A3CEC2D582CULL, "qun220cr_javaris.bnk"},
	{0x435831D2D83C02F8ULL, "art\\vfx\\combat\\vfx_r_gn3_repulsion_aoe_p.impa.gad"},
	{0x43587460B73EEAEAULL, "hlac_ht_mansion_n_2995_3z.dds"},
	{0x4358A948DCBECFF3ULL, "vxca_caverns1_mines_e5y_0_0.vlm"},
	{0x4358BED11F81997BULL, "6054037_m.fxe"},
	{0x4358D288E39BF639ULL, "hlac_ht_0_hightown_n_-65537_4z.dds"},
	{0x4358E2F9E7D39931ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_medium.evt"},
	{0x43594F7B036BE368ULL, "vxlt_gallowscourtyard_4w9_0_0.vlm"},
	{0x43597BEFF9411811ULL, "918896800.ogg"},
	{0x435997D5E477DE80ULL, "hlac_w_sundermount_-262147_0z.dds"},
	{0x4359D4CFD20CAAF7ULL, "6093999_m.fxe"},
	{0x4359F52057B93656ULL, "vxw_sundermount_2jw_0_0.vlm"},
	{0x435A07AF81A9B19AULL, "40436096.ogg"},
	{0x435A3A60CD0860A9ULL, "445924135.ogg"},
	{0x435A558CFCDA5260ULL, "vxw_encounter2_camp_1c_0_0.vlm"},
	{0x435AAA3E4D94460CULL, "character_character_p148.psh.fxo.dx11"},
	{0x435AD37DE1ACD730ULL, "t3_cth_nb5.tnt"},
	{0x435B11D2A83E0C0CULL, "ht_0_hightown_a2.rml"},
	{0x435B4BD0F3022E4EULL, "6202108_f.fxe"},
	{0x435B5C91ADCF4B86ULL, "167891022.ogg"},
	{0x435C2013C17DE930ULL, "vxlt_0_lowtown_a2_n_gc4_0_0.vlm"},
	{0x435C2FCC1F33C2FFULL, "6104620_m.fxe"},
	{0x435C438DA2768741ULL, "mer221ar_player_home.ncs"},
	{0x435CCD899B419CF1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockback.ani"},
	{0x435D113C747FF549ULL, "vxht_0_hightown_a3_c2t_0_0.vlm"},
	{0x435D42AA1497BC45ULL, "548028389.ogg"},
	{0x435D9E5654137B85ULL, "vxw_bonepitdestroyed_a3_28m_0_0.vlm"},
	{0x435DB414D589A430ULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench2_sebastianintro.ani"},
	{0x435E3C202BAD054AULL, "400855758.ogg"},
	{0x435E588FE5B87934ULL, "6071291_m.fxe"},
	{0x435E6942DF46790DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_up.ani"},
	{0x435EDAA77792DF28ULL, "ts_qunari_wpn_props.gda"},
	{0x435F311BDD7507ACULL, "vxw_woundedcoast_bo_0_0.vlm"},
	{0x435F8C53E7721A17ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_1_cailin.gad"},
	{0x435FB41A83F7CC17ULL, "vxw_sundermount_23q_0_0.vlm"},
	{0x435FDA4F69EA631FULL, "mcr371cr_emile_de_launcet.utc"},
	{0x435FF7EF0FC23E80ULL, "vxdr_0_deeproads_cvw_0_0.vlm"},
	{0x4360100F2B62D166ULL, "6125695_m.fxe"},
	{0x4360218A2FF5BC7FULL, "star_b.dds"},
	{0x43604AB6C5C0274EULL, "378517458.ogg"},
	{0x43605A8DA27A977BULL, "mer221ar_player_home.are"},
	{0x4360A73ADE90F927ULL, "6128770_m.fxe"},
	{0x4360D1B11C66801FULL, "324137511.ogg"},
	{0x4360D5696D9D5692ULL, "gen00fl_isabela6099479.cl"},
	{0x436113CD49D35A64ULL, "vxlt_undercitydungeon_1is_0_0.vlm"},
	{0x43613FA288B04BEBULL, "31161959.ogg"},
	{0x436169791EAE4A87ULL, "850610230.ogg"},
	{0x4361961B052A0CA2ULL, "hlac_lt_0_lowtown_docks_a3_-1_0y.dds"},
	{0x4361E9FE587D19DBULL, "vxlt_0_lowtown_docks_8rm_0_0.vlm"},
	{0x43627CD03C1839CBULL, "vxw_blightlands_9w_0_0.vlm"},
	{0x43629135AD67387FULL, "ts_prog_snow_obsidian.gda"},
	{0x4362EA97BC41D0B6ULL, "302177543.ogg"},
	{0x4363A9A4E4BD9419ULL, "follower_banter6163882.cl"},
	{0x4363B4A4A8F4F417ULL, "6140596_m.fxe"},
	{0x4363D8899B47CD93ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockback.gad"},
	{0x4363DD513A42F40CULL, "fen210_hadriana6110407.cl"},
	{0x43648BE0B2A26F40ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bed_d.dds"},
	{0x4364BFE0AAC9A650ULL, "vxht_mansion2_2or_0_0.vlm"},
	{0x4364F2C712BDE481ULL, "tbe100_mirror6200108.cl"},
	{0x436515E1A8E8E208ULL, "qun210tr_random_thug.utt"},
	{0x43652FD323F5A805ULL, "dae150ip_barrier.utp"},
	{0x43654C5EF59AE740ULL, "vxlt_0_lowtown_a2_n_exg_0_0.vlm"},
	{0x4365653D8908E1ACULL, "nat380_nathaniel_banter6067306.cl"},
	{0x43657017E289F131ULL, "hlac_lt_gallowstemplar_2394_6y.dds"},
	{0x4365C088EC522809ULL, "6141526_m.fxe"},
	{0x4365D961F95AA1D1ULL, "6115497_m.fxe"},
	{0x4365F27F7FD094B0ULL, "6035224_m.fxe"},
	{0x4365F87D1BE61ED4ULL, "381159241.ogg"},
	{0x436647C12C6A2C98ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_mother_andersintro.gad"},
	{0x4366FAF9E7DFFF00ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_weightshift_medium.ani"},
	{0x43675CB239798ACDULL, "vxht_brothel_n_57l_0_0.vlm"},
	{0x43682112360EA761ULL, "str_slimebone_d.dds"},
	{0x436821962D769788ULL, "lgt310pt_shw_mer_main.ncs"},
	{0x43686329B85210D5ULL, "dre000_varric6207986.cl"},
	{0x436866FD49AF4ED6ULL, "vxw_sundermount_2ox_0_0.vlm"},
	{0x43687A65E031303BULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v2.ani"},
	{0x436882576EAEF8B9ULL, "dre000_varric6027762.cl"},
	{0x4368BFA0F5CF1965ULL, "48470408.ogg"},
	{0x436A0F810059A486ULL, "227535815.ogg"},
	{0x436AD5899B4DE7F8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockback.evt"},
	{0x436ADF9C8317562DULL, "vxht_chantry_n_2li_0_0.vlm"},
	{0x436B847270B7CDE2ULL, "vxlt_0_lowtown_docks_a3_d3h_0_0.vlm"},
	{0x436B8DC942D643FCULL, "224416850.ogg"},
	{0x436C1A55881E9423ULL, "6108062_m.fxe"},
	{0x436C24FB01DF5C13ULL, "475347003.ogg"},
	{0x436C2CC068C366C7ULL, "art\\characters\\playercharacter\\humanfemale\\eye_hlm.mao"},
	{0x436CAD61ACA593A7ULL, "hlac_lt_0_lowtown_1_2y.dds"},
	{0x436CD7BEE44FA885ULL, "523809345.ogg"},
	{0x436D126654B395E9ULL, "vxht_0_hightown_a3_5io_0_0.vlm"},
	{0x436DA38E826AA031ULL, "1037515689.ogg"},
	{0x436DD5FA20CC692FULL, "6134461_m.fxe"},
	{0x436DD84A85A18DBCULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm_lashesm1.mao"},
	{0x436DDD59E678DE81ULL, "6170867_m.fxe"},
	{0x436DED5EC370CDCFULL, "vxlt_0_lowtown_a2_gyz_0_0.vlm"},
	{0x436DF8223898430CULL, "mcr371_scared_mage_m6087693.cl"},
	{0x436E25F775EDE497ULL, "vxw_sundermount_24p_0_0.vlm"},
	{0x436E27C8EF559D9DULL, "gen00fl_aveline6137674.cl"},
	{0x436E321D5D912839ULL, "mrl210_merrill6116395.cl"},
	{0x436E72D4F902CB42ULL, "mag371_elsa.cnv"},
	{0x436E9BF2A7D6E4BEULL, "386062090.ogg"},
	{0x436ECDDC3398F2B6ULL, "w_encounter3_40_0.rml"},
	{0x436EEAFAFA99C201ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_water_sa.dds"},
	{0x436F2D698AE2BAAFULL, "vxlt_gallowstemplar_f_4ss_0_0.vlm"},
	{0x436F85AA3BC972E4ULL, "vxlt_undercitydungeon_1lp_0_0.vlm"},
	{0x436F895AD951CB51ULL, "6137075_m.fxe"},
	{0x436FB72B59E67120ULL, "6060096_f.fxe"},
	{0x436FDD12928EA3EEULL, "w_encounter3_229_0.msh"},
	{0x436FE35EBA2CB1F1ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_hed_hlk_0d.dds"},
	{0x436FEB12928EBB3FULL, "w_encounter3_229_0.mao"},
	{0x436FF712928ECFDCULL, "w_encounter3_229_0.mmh"},
	{0x43705C82E049D70EULL, "501257400.ogg"},
	{0x437069CCF6492ADBULL, "vxlt_0_lowtown_docks_a3_ctc_0_0.vlm"},
	{0x4370C2744534B943ULL, "1000908867.ogg"},
	{0x4371194CB28217B7ULL, "936072264.ogg"},
	{0x437151A21C7D6FE5ULL, "6089763_m.fxe"},
	{0x437154D4F904CE7AULL, "mag371_elsa.bnk"},
	{0x43715772C80E7567ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc107a_0t.dds"},
	{0x437199DFA317B600ULL, "vxht_0_hightown_n_7iz_0_0.vlm"},
	{0x4371CE973758FCE3ULL, "cart_sa.dds"},
	{0x437289464DBE216EULL, "6171526_m.fxe"},
	{0x4373014FA837ED45ULL, "art\\characters\\playercharacter\\humanmale\\glv_masa.phy"},
	{0x437331BD9CBFBED0ULL, "vxht_mansion2_2hs_0_0.vlm"},
	{0x437336B428384B78ULL, "6063821_m.fxe"},
	{0x4373B1352605AE54ULL, "follower_banter6141827.cl"},
	{0x4373B953E7830F9DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_1_cailin.ani"},
	{0x4373FE2F33CD517DULL, "llac_lt_0_lowtown_docks_a3_-65536_0.dds"},
	{0x43741B0CA443D25CULL, "one300_nuncio6070816.cl"},
	{0x43743CE57E50E966ULL, "791033221.ogg"},
	{0x437443BBDDCDFF60ULL, "follower_banter6169825.cl"},
	{0x437488C553637195ULL, "848524813.ogg"},
	{0x4374991D052A048AULL, "vxca_caverns1_gk1_0_0.vlm"},
	{0x43752CCC8B85D3EAULL, "6096504_m.fxe"},
	{0x43754871CE735ED1ULL, "vxlt_0_lowtown_n_h7b_0_0.vlm"},
	{0x437574A28830EDC8ULL, "atl_logininfos_dxt5_dat.xml"},
	{0x43758265E03BC832ULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v2.evt"},
	{0x437594359B15A8ECULL, "art\\characters\\weapons\\dagger\\w_dgr_dg101a.phy"},
	{0x4375D5099D5CE653ULL, "vxlt_0_lowtown_docks_n_bf0_0_0.vlm"},
	{0x43761D82BC3B0313ULL, "dre180_db_rock_warr.cut"},
	{0x43761EF284F66004ULL, "vxlt_gallowstemplar_6qk_0_0.vlm"},
	{0x4376465B67ACA509ULL, "892553722.ogg"},
	{0x4376635B7480E2B8ULL, "lgt160_ntb_elf_conflict6163334.cl"},
	{0x43766769B137B324ULL, "mer141ar_caverns.lst"},
	{0x4376D8DA3BA56756ULL, "vxw_sundermount_2py_0_0.vlm"},
	{0x43786CE8024B95C0ULL, "gen00fl_fenris6110855.cl"},
	{0x4378AA0CA5C91BAEULL, "gen00fl_anders6125297.cl"},
	{0x4379860E755489DFULL, "art\\vfx\\combat\\vfx_m_spi_frc_cast_c.mmh"},
	{0x437989E8267C4073ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_4.dds"},
	{0x4379C289DD0A2352ULL, "320908453.ogg"},
	{0x4379C8DFDC873021ULL, "6140859_m.fxe"},
	{0x4379DCBF3DEFFB1EULL, "hlac_w_sundermount_131071_0x.dds"},
	{0x4379F64F62ADE662ULL, "vxlt_0_lowtown_docks_a3_d4i_0_0.vlm"},
	{0x4379FE17C2CB51CEULL, "540829064.ogg"},
	{0x437A45A618AD7825ULL, "static_static_p8.psh.fxo.dx11"},
	{0x437AE8FCB4F67E26ULL, "hlac_dr_slavers_3207_0z.dds"},
	{0x437B6851CE10E61FULL, "6106671_m.fxe"},
	{0x437B7CC90895A1EFULL, "vxlt_hangedman_n_t9_0_0.vlm"},
	{0x437B8474D19885E7ULL, "qcr220_grey_wardens6135957.cl"},
	{0x437BB709BBA7168FULL, "art\\characters\\creatures\\desiredemon\\animation\\a_lp.gad"},
	{0x437BDE4FF2A999BEULL, "vxht_0_hightown_n_akt_0_0.vlm"},
	{0x437BE46B4D061209ULL, "vxlt_warehouse_15h_0_0.vlm"},
	{0x437C08EE5D98CC92ULL, "hlac_lt_0_lowtown_docks_65533_0x.dds"},
	{0x437C3592984841F3ULL, "plt_cod_ltr_ggt_note3.plo"},
	{0x437C6769B13C1E8DULL, "mer141ar_caverns.ncs"},
	{0x437C7BCFF578BDCEULL, "6178420_m.fxe"},
	{0x437CA165E0421BC5ULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v2.gad"},
	{0x437CB0B00F5250E9ULL, "6133516_m.fxe"},
	{0x437CB5B9EBF93870ULL, "562006779.ogg"},
	{0x437CDFA5A88D4EA1ULL, "vxca_caverns2_7mz_0_1.vlm"},
	{0x437D038F603E5F03ULL, "one110_pryce6088409.cl"},
	{0x437D0CDDD6F5763FULL, "6072150_m.fxe"},
	{0x437D9F467CD8D32FULL, "vxlt_gallowstemplar_f_4xr_0_0.vlm"},
	{0x437DA80875121174ULL, "hlac_lt_0_lowtown_a2_n_c_65536_3z.dds"},
	{0x437DF78683CF3FC0ULL, "vxca_caverns1_mines_i1n_0_0.vlm"},
	{0x437DF7872DBF8B64ULL, "vxlt_undercitydungeon_1wq_0_0.vlm"},
	{0x437E005CBE1A6136ULL, "108226867.ogg"},
	{0x437E4FF8F0DAF7F8ULL, "6036773_m.fxe"},
	{0x437EE25598A1F6F4ULL, "6077987_m.fxe"},
	{0x437F6BE61C1879CBULL, "458540407.ogg"},
	{0x438082CA32462ACBULL, "vxw_sundermount_26j_0_0.vlm"},
	{0x4380ABB7F010E0BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v3.ani"},
	{0x43810CFE6C60609AULL, "6117049_m.fxe"},
	{0x4381253CEC5054AFULL, "qun220cr_javaris.ncs"},
	{0x43814FCA4F6840CEULL, "gui_reticle_p0.psh.fxo"},
	{0x43815075A5BB04B2ULL, "189685784.ogg"},
	{0x438177E94BB7EF03ULL, "mcr371_hostile_aveline.bnk"},
	{0x4381C53F5666915BULL, "6153725_m.fxe"},
	{0x4381CC09BBABA63CULL, "art\\characters\\creatures\\desiredemon\\animation\\a_lp.evt"},
	{0x43825546DDB9DAABULL, "art\\characters\\playercharacter\\elfmale\\animation\\blendtree\\em_exp_idle.anb"},
	{0x43829E4BFB7BCACEULL, "1066013054.ogg"},
	{0x43829FEB6E6955D1ULL, "oc_dust.mao"},
	{0x43830AF9F7201D0AULL, "vxca_caverns1_gf2_0_0.vlm"},
	{0x43835AD66EC82CCBULL, "279966279.ogg"},
	{0x43838E7A78A8ECCDULL, "vxht_0_hightown_a2_c4c_0_0.vlm"},
	{0x438395D17DEF4CABULL, "mrl150_dalish_guard6068293.cl"},
	{0x4383B0278E8B04A4ULL, "ttx_pillar_base_a_1.msh"},
	{0x4383BA4EC0697751ULL, "vxlt_0_lowtown_n_h2a_0_0.vlm"},
	{0x4383C6278E8B2A46ULL, "ttx_pillar_base_a_1.mmh"},
	{0x4383FE87BEDDCB0BULL, "hlac_ht_brothel_n_3371_0x.dds"},
	{0x438446E68F52FED3ULL, "vxlt_0_lowtown_docks_n_bc3_0_0.vlm"},
	{0x438459E94BB9F233ULL, "mcr371_hostile_aveline.cnv"},
	{0x43849AC7B15C77A1ULL, "6201362_f.fxe"},
	{0x4384A5346CF0039CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magb_0s.dds"},
	{0x4384C8C024A94E85ULL, "6136014_m.fxe"},
	{0x4384D68F2B6767A4ULL, "6171876_m.fxe"},
	{0x43856C27EE5640F3ULL, "vxlt_hangedman_26p_0_0.vlm"},
	{0x438596AFAF4C6F2EULL, "qun221_petrice6076201.cl"},
	{0x43865BF6EDB71B61ULL, "zz_approval_debug6003099.cl"},
	{0x4386FC9AB93F9380ULL, "lgt150ip_rand_1.utp"},
	{0x43870F7072DD40A5ULL, "hlac_lt_0_lowtown_docks_a3_-65537_3z.dds"},
	{0x438786E55A9BE643ULL, "6000753_m.fxe"},
	{0x4387AC92C305BA80ULL, "vxht_chantry_2dw_0_0.vlm"},
	{0x4387C3566703872DULL, "vxht_chantry_n_2bg_0_0.vlm"},
	{0x4387C87F8AE3EDC8ULL, "6142449_m.fxe"},
	{0x438815D961C0E695ULL, "ts_chantry_wpn_table.gda"},
	{0x4388A2D523D74C10ULL, "6135083_m.fxe"},
	{0x4388A9CEBC2C9107ULL, "6082100_m.fxe"},
	{0x4388DBB37DD8EC70ULL, "203383276.ogg"},
	{0x43890F15A4725510ULL, "vxw_woundedcoast_1hu_0_1.vlm"},
	{0x438970A8466B6B2CULL, "6174057_m.fxe"},
	{0x43899448767B9F51ULL, "zz_o_11_gallows_nt_cinetest.cut"},
	{0x4389EEA5FA8BBA6FULL, "vxlt_hangedman_n_s8_0_0.vlm"},
	{0x438A6E75D3BE5211ULL, "6032022_m.fxe"},
	{0x438B0E3CEC58822BULL, "qun220cr_javaris.stg"},
	{0x438B38B516A9077EULL, "430039818.ogg"},
	{0x438B3FAA89275F67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_wesley.gad"},
	{0x438B4D85E5A48800ULL, "vxdr_0_deeproads_cer_0_0.vlm"},
	{0x438BAB37D768415AULL, "vxht_mansion2b_3dl_0_0.vlm"},
	{0x438BD7F14E89C0D7ULL, "arlathanfocusingcrystal.dds"},
	{0x438BE078B80044C8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medc_n.dds"},
	{0x438BEE2886D63B14ULL, "181620567.ogg"},
	{0x438BFDF0DA73EC42ULL, "vxht_mansion2_1w7_0_0.vlm"},
	{0x438C044C72207ADBULL, "6089475_m.fxe"},
	{0x438CEBCF1BF0B351ULL, "llac_ht_store_187_0.dds"},
	{0x438D0CCB873390C9ULL, "136553387.ogg"},
	{0x438D4AF4C12184B2ULL, "zz_qun_debug6036927.cl"},
	{0x438D56A5EE9F1E9EULL, "mag311_keran6041724.cl"},
	{0x438D70F0389B7E64ULL, "framebuffereffects_hslmatrix_p0.vsh.fxo.dx11"},
	{0x438D8B787B88A959ULL, "396789711.ogg"},
	{0x438DB3B7F01B78B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v3.evt"},
	{0x438DBD69CB329659ULL, "hlac_ht_mansion2b_818_52z.dds"},
	{0x438DC79C50A79D52ULL, "vxlt_0_lowtown_docks_n_bc6_0_0.vlm"},
	{0x438E355C0069A0E5ULL, "6087563_m.fxe"},
	{0x438E36D230491F5CULL, "dae000_templar_recruit_16164440.cl"},
	{0x438EB7B7C54F1150ULL, "mcr310_argument6075110.cl"},
	{0x438EB819BF2B4E9BULL, "vxht_0_hightown_a3_b01_0_0.vlm"},
	{0x438EC2CC65416F43ULL, "bind_dot.mao"},
	{0x438ED54F0A5C6C67ULL, "gen00fl_fenris6113819.cl"},
	{0x438F796CB94A54ADULL, "art\\levels\\ztools\\default\\generic\\4so_basic.mmh"},
	{0x438F7D6CB94A5BFEULL, "art\\levels\\ztools\\default\\generic\\4so_basic.mao"},
	{0x438F8B6CB94A7333ULL, "art\\levels\\ztools\\default\\generic\\4so_basic.msh"},
	{0x438FD409BBB7F155ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_lp.ani"},
	{0x4390FBB3CEBEDAB2ULL, "6072825_m.fxe"},
	{0x439134AA892BB874ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_wesley.evt"},
	{0x4391C649F879C08DULL, "489027472.ogg"},
	{0x4391ECAAA4C4C1F9ULL, "follower_banter6167118.cl"},
	{0x439200576A9F054DULL, "vxht_0_hightown_a2_c3b_0_0.vlm"},
	{0x43925DDC4E6B4C9AULL, "599470954.ogg"},
	{0x43928AA9BE4F4124ULL, "vxht_keep_n_3y6_0_0.vlm"},
	{0x4392B8C381491753ULL, "vxlt_0_lowtown_docks_n_bx2_0_0.vlm"},
	{0x4392C563E51C62DAULL, "6202118_m.fxe"},
	{0x4392D7FB3F1F7726ULL, "mag360_alain6044606.cl"},
	{0x4393F8A688F250C1ULL, "abi_r_trip.bnk"},
	{0x4394AA8881624D95ULL, "hlac_ht_mansion2b_818_40y.dds"},
	{0x4394B0C00CA7B880ULL, "gen00fl_anders6119542.cl"},
	{0x4394B4BCF46D2C74ULL, "vxht_mansion2b_9n_0_0.vlm"},
	{0x4394D2B7F021CC45ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v3.gad"},
	{0x439559B9643C0710ULL, "vxlt_0_lowtown_docks_a3_n_btn_0_0.vlm"},
	{0x4395B3761AE9C698ULL, "vxht_brothel_n_4ab_0_0.vlm"},
	{0x4395B384E628E140ULL, "gui_reticle_p0.vsh.fxo.dx11"},
	{0x4395F744FD3047F5ULL, "vxlt_0_lowtown_a2_n_h21_0_0.vlm"},
	{0x439722F739A1505BULL, "hlac_w_bonepit_0_6y.dds"},
	{0x43973ABA8FF0F9B2ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.cess.gad"},
	{0x439775C5E449F47AULL, "225492703.ogg"},
	{0x439780F296686D90ULL, "vxw_woundedcoast_1st_0_1.vlm"},
	{0x4397B86628570994ULL, "vxht_0_hightown_n_c12_0_0.vlm"},
	{0x43992FF381DFD630ULL, "blood_imp_04.mao"},
	{0x43999469F33E81E9ULL, "6139316_m.fxe"},
	{0x439A4D107C15E40CULL, "fen110_fenris6117950.cl"},
	{0x439A550A789AACA4ULL, "gen_im_delete_me_06.uti"},
	{0x439A6FCDCC6A04C2ULL, "vxht_mansion2_1z0_0_0.vlm"},
	{0x439A89DF8BA5CAD5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wrba_0d.dds"},
	{0x439AC560A2655465ULL, "llac_lt_0_lowtown_a2_n_131072_3.dds"},
	{0x439ADE9940A93318ULL, "258993099.ogg"},
	{0x439B0F45F7D86D37ULL, "al_rogue_attributes.gda"},
	{0x439B6E69B156DB17ULL, "mer141ar_caverns.are"},
	{0x439BDC4D723667A7ULL, "hlac_ca_caverns2prc_5802_14x.dds"},
	{0x439CB57A1A4FEF08ULL, "396375483.ogg"},
	{0x439D64602729A3FBULL, "vxca_cavern1_haunted_edb_0_0.vlm"},
	{0x439DF499F2DEDC5BULL, "hlac_lt_undercityencounter_-1_41z.dds"},
	{0x439E3DBA8FF71DE9ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.cess.evt"},
	{0x439EC0C712A60DB4ULL, "333558784.ogg"},
	{0x439F10983BEAE103ULL, "mag101_grace6053111.cl"},
	{0x439F2CAA8937E7ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_wesley.ani"},
	{0x439F5D73C28601ACULL, "lt_0_lowtown_43_0_lc.mmh"},
	{0x439FAA232ACF3FC4ULL, "llac_ht_0_hightown_a3_-65537_3.dds"},
	{0x43A0123CEC6AE541ULL, "qun220cr_javaris.utc"},
	{0x43A02D6101BCE561ULL, "6002115_m.fxe"},
	{0x43A04CBECB0503E0ULL, "vxdr_primevalentrance_220_0_0.vlm"},
	{0x43A0726CB958BC6AULL, "art\\levels\\ztools\\default\\generic\\4so_basic.phy"},
	{0x43A0B29ADAD821C6ULL, "gen00fl_merrill6199887.cl"},
	{0x43A0C6B2C1D2CDF4ULL, "darktown_male_fereldan_sta6175156.cl"},
	{0x43A103B073EFCEBCULL, "gen_im_npc_arm_hum_cht_tmpd.uti"},
	{0x43A12AA0733F2FD3ULL, "vxlt_0_lowtown_docks_n_bu5_0_0.vlm"},
	{0x43A137FB8C4CAB54ULL, "6168030_m.fxe"},
	{0x43A178E0DDB12894ULL, "llac_lt_0_lowtown_a2_131071_0.dds"},
	{0x43A1870E7575DD1CULL, "art\\vfx\\combat\\vfx_m_spi_frc_cast_c.anb"},
	{0x43A1BB612008C914ULL, "932720544.ogg"},
	{0x43A1FDF7069D50EAULL, "6131179_m.fxe"},
	{0x43A280F45B9BAAF0ULL, "vxlt_blackemporium_13o_0_0.vlm"},
	{0x43A29B85C1B949ABULL, "67875505.ogg"},
	{0x43A2E28549E9C42AULL, "llac_lt_0_lowtown_a2_n_c_131071_5.dds"},
	{0x43A33790A728A5E8ULL, "vxlt_0_lowtown_a2_fe8_0_0.vlm"},
	{0x43A38FC9C11DA965ULL, "309442764.ogg"},
	{0x43A3C35CEBFAF998ULL, "hlac_lt_warehouse_0_9z.dds"},
	{0x43A3C6D8016514F0ULL, "6078735_m.fxe"},
	{0x43A3E37D3CAB55EBULL, "vxht_0_hightown_a3_b8i_0_0.vlm"},
	{0x43A457957FB42713ULL, "vxlt_0_lowtown_a2_gz7_0_0.vlm"},
	{0x43A49B84155E3B29ULL, "6110436_m.fxe"},
	{0x43A51CBA625A9D69ULL, "6101727_f.fxe"},
	{0x43A62FC41A6A5BDFULL, "vxca_caverns1_flx_0_0.vlm"},
	{0x43A6B0787E93E713ULL, "zz_mer_debug6180867.cl"},
	{0x43A6C6D2C35AB245ULL, "vxlt_backalley_g_1l0_0_0.vlm"},
	{0x43A6CF78C8338CF0ULL, "6100999_m.fxe"},
	{0x43A73C36D9E80836ULL, "201940998.ogg"},
	{0x43A7497D15C5C276ULL, "70311910.ogg"},
	{0x43A83A3F285B479BULL, "6088348_m.fxe"},
	{0x43A86A4541AC53E6ULL, "m_one300ar_vanard_fight.bnk"},
	{0x43A8B1A63BE35290ULL, "and121_karl6093990.cl"},
	{0x43A8C7780610130BULL, "vxlt_0_lowtown_docks_n_cs2_0_0.vlm"},
	{0x43A8D3DA58F73AD1ULL, "vxlt_gallowstemplar_n_435_0_0.vlm"},
	{0x43A947E047C76374ULL, "vxlt_0_lowtown_a2_n_hdb_0_0.vlm"},
	{0x43A95DBD48A432A8ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_1x.dds"},
	{0x43A961CEE62F8E9EULL, "zz_followers_banter6158379.cl"},
	{0x43A9E82DB69358D7ULL, "128181057.ogg"},
	{0x43AA27744B8497D6ULL, "6016428_f.fxe"},
	{0x43AA365D8D38543AULL, "6099729_m.fxe"},
	{0x43AA784CD48FBE9BULL, "372696455.ogg"},
	{0x43AA7AB6C4A92FBCULL, "mcr370_messenger6071831.cl"},
	{0x43AA823D35E28CA5ULL, "mcr370_assignments6177322.cl"},
	{0x43AAC4D4319C60F0ULL, "dae300_vashothlead6062791.cl"},
	{0x43AAD6DF9411C6CAULL, "6127698_f.fxe"},
	{0x43AAFE6D2ACDF8FBULL, "vxht_chantry_n_2hs_0_0.vlm"},
	{0x43AB1C92986F8790ULL, "plt_cod_ltr_ggt_note3.nss"},
	{0x43AB25BA90017F58ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.cess.ani"},
	{0x43AB2955EB7399BEULL, "vxht_0_hightown_a2_b78_0_0.vlm"},
	{0x43ACA404F73913A4ULL, "185832104.ogg"},
	{0x43ACE219E98D002AULL, "vxlt_0_lowtown_a2_n_eou_0_1.vlm"},
	{0x43AD45F4B274CBC0ULL, "vxlt_gallowstemplar_f_33k_0_0.vlm"},
	{0x43AE8423A9C6A79DULL, "vxht_0_hightown_b8j_0_0.vlm"},
	{0x43AE9F3DF3B2FDE1ULL, "6005015_m.fxe"},
	{0x43AEE9BF5B2341C2ULL, "zz_prmpt_debug6198222.cl"},
	{0x43AF40656EFD50EEULL, "6162405_m.fxe"},
	{0x43AF7501A28A32A5ULL, "6076331_f.fxe"},
	{0x43AFD4CAE0938BEAULL, "gen00fl_isabela6100081.cl"},
	{0x43B023D9F92BF17BULL, "519522490.ogg"},
	{0x43B0FFE12E024D6BULL, "51283549.ogg"},
	{0x43B1169A202E9850ULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_1z.dds"},
	{0x43B156F46839B67AULL, "hlac_dr_dungeon_3216_4z.dds"},
	{0x43B16FD100BC19E8ULL, "6198705_m.fxe"},
	{0x43B26A31C686E99DULL, "228448324.ogg"},
	{0x43B29DD801D115D1ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody1_.gad"},
	{0x43B2AFE00527EBB4ULL, "6149475_m.fxe"},
	{0x43B383FB349B3E0CULL, "lt_0_lowtown_a2_n_20_0_lc.mmh"},
	{0x43B3E8B7C8107A6CULL, "docks_male_elven_worker_sta6175831.cl"},
	{0x43B51AA11A9D9EE4ULL, "follower_banter6155721.cl"},
	{0x43B520C6EB8C098AULL, "vxlt_0_lowtown_docks_a3_d05_0_0.vlm"},
	{0x43B542D1426E9EEEULL, "hlac_w_bonepit_131078_0x.dds"},
	{0x43B602F0CBA0673AULL, "mag120cr_brothelguard.utc"},
	{0x43B6145C710AFC58ULL, "gen00fl_merrill6109596.cl"},
	{0x43B66B359B4CF9B7ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg101a.mmh"},
	{0x43B681359B4D1F29ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg101a.msh"},
	{0x43B69F02E6DEE6B9ULL, "6147685_m.fxe"},
	{0x43B750D3F000A1EDULL, "hlac_lt_backalley_n_65536_5x.dds"},
	{0x43B789BF3C64194FULL, "zz_vault_debug6203225.cl"},
	{0x43B89ED801D5837FULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody1_.ani"},
	{0x43B8EF9EB30411D3ULL, "6053292_m.fxe"},
	{0x43B9396F189F27A7ULL, "ht_0_hightown_a3_34_0_lc.mmh"},
	{0x43B9704A1CC4117BULL, "vxht_chantry_n_2cr_0_0.vlm"},
	{0x43B9F90F01E03EB4ULL, "vxdr_primevalentrance_16u_0_2.vlm"},
	{0x43BA161D5D7D40B9ULL, "ts_dalish_prog_junk1.gda"},
	{0x43BA323140C51A74ULL, "mag121_idunna6084214.cl"},
	{0x43BA494151CD430AULL, "w_bonepitdestroyed_a3_104_0.mao"},
	{0x43BA4D4151CD4A59ULL, "w_bonepitdestroyed_a3_104_0.mmh"},
	{0x43BA4DC28DD9B563ULL, "gen00fl_fenris6106651.cl"},
	{0x43BA574151CD5B47ULL, "w_bonepitdestroyed_a3_104_0.msh"},
	{0x43BAE1330A589BD7ULL, "vxht_mansion2_3du_0_0.vlm"},
	{0x43BAFDC5FEE19A29ULL, "vxw_woundedcoast_19f_0_0.vlm"},
	{0x43BB0242C6A2AAF8ULL, "hlac_w_bonepitdestroyed_a3_196611_0x.dds"},
	{0x43BB2B214BF7A87BULL, "392141319.ogg"},
	{0x43BB38337BF6B2F5ULL, "gen00fl_merrill6150979.cl"},
	{0x43BB4DCE27DA4BF6ULL, "gen_im_arm_cht_elf_coma_t3.uti"},
	{0x43BC1749C6C19A9CULL, "ave200_donnic6111212.cl"},
	{0x43BC3131C8986AADULL, "6037371_m.fxe"},
	{0x43BC58D1F74E9A04ULL, "vxlt_0_lowtown_docks_n_aos_0_0.vlm"},
	{0x43BC91A3B8C9BBF4ULL, "541489638.ogg"},
	{0x43BD14497D5AB8F5ULL, "vxca_caverns1_mines_dp9_0_0.vlm"},
	{0x43BD34DEBDDE76B6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ci_galow.phy"},
	{0x43BDF8B48D429E76ULL, "hlac_ht_0_hightown_a3_-65536_4z.dds"},
	{0x43BE0EB1B014CE57ULL, "6215155_m.fxe"},
	{0x43BF93D21E90B590ULL, "313675720.ogg"},
	{0x43C020133E60D3C3ULL, "llac_ht_0_hightown_n_-131074_2.dds"},
	{0x43C044326C4B7CDFULL, "6191275_m.fxe"},
	{0x43C0F4339747CF4AULL, "563451905.ogg"},
	{0x43C104E1A9372435ULL, "qun210tr_random_thug.ncs"},
	{0x43C15DFE0DF3EAEAULL, "art\\levels\\races\\design\\_textures\\trap_floorlever_d.dds"},
	{0x43C18A94E2F93AECULL, "gen00fl_aveline6116485.cl"},
	{0x43C19E4396A6ABA6ULL, "gen00fl_isabela6099538.cl"},
	{0x43C1D7AD1532EBDDULL, "vxht_mansion2_3bq_0_0.vlm"},
	{0x43C29A899BB3C5ABULL, "6141595_m.fxe"},
	{0x43C37438F3A3A4FEULL, "vxlt_0_lowtown_h8g_0_0.vlm"},
	{0x43C392A3DD82220AULL, "vxlt_0_lowtown_docks_a3_d56_0_0.vlm"},
	{0x43C3A21BDD5D622BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_5.pwk"},
	{0x43C3BB26D210948BULL, "zz_lgt_1_debug6114393.cl"},
	{0x43C417A0AC78143EULL, "vxht_0_hightown_n_api_0_0.vlm"},
	{0x43C42DA036AD495BULL, "llac_lt_undercityserialkiller_-1_12.dds"},
	{0x43C45E7E7533F834ULL, "188651618.ogg"},
	{0x43C4705F419A02B4ULL, "mag320_orsino6071164.cl"},
	{0x43C4CC9D868E0CC4ULL, "al_merril_talents.gda"},
	{0x43C5700A4C87B66BULL, "6206933_m.fxe"},
	{0x43C5CEAEE3E8DF6DULL, "448254132.ogg"},
	{0x43C63D8295382954ULL, "vxht_0_hightown_n_8q1_0_0.vlm"},
	{0x43C67B0393872221ULL, "717132181.ogg"},
	{0x43C6EF0FD22DB02AULL, "vxht_0_hightown_at6_0_0.vlm"},
	{0x43C6FFDFEB83EF9EULL, "vxht_mansion2b_3c1_0_0.vlm"},
	{0x43C71DDEBDE6A457ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ci_galow.msh"},
	{0x43C72FDEBDE6C2BAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ci_galow.mao"},
	{0x43C733DEBDE6C9C9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ci_galow.mmh"},
	{0x43C87AEA424E5B22ULL, "lt_0_lowtown_docks_a3_247_0.msh"},
	{0x43C87CEA424E5ED8ULL, "lt_0_lowtown_docks_a3_247_0.mmh"},
	{0x43C888EA424E737BULL, "lt_0_lowtown_docks_a3_247_0.mao"},
	{0x43C89603EEEE5341ULL, "vxdr_primevaltomb_7c3_0_0.vlm"},
	{0x43C8A17E0BF87342ULL, "6141873_m.fxe"},
	{0x43C8E9675AE68418ULL, "dae000_dalish_apprentices6147546.cl"},
	{0x43C95699C8D7312BULL, "follower_banter6163497.cl"},
	{0x43CA0A07BFD566CFULL, "191964980.ogg"},
	{0x43CA485D9103F1B9ULL, "215518551.ogg"},
	{0x43CA5EF6A061BAFFULL, "1020716936.ogg"},
	{0x43CA894FDCD4335FULL, "qcr000_arishok6166985.cl"},
	{0x43CAF446699DBEF5ULL, "839168101.ogg"},
	{0x43CB07FE835A53F3ULL, "6125157_m.fxe"},
	{0x43CB2708212DF475ULL, "vxht_mansion2_21m_0_0.vlm"},
	{0x43CBDACD23B242FFULL, "qun221_viscount6073875.cl"},
	{0x43CBEB9932BAC100ULL, "6084549_m.fxe"},
	{0x43CC44ACCD177D4AULL, "105801272.ogg"},
	{0x43CC84F3E262BEB4ULL, "698607200.ogg"},
	{0x43CC85F9D611AFFAULL, "docks_male_overseer_mob6176568.cl"},
	{0x43CC980EFF96E1DCULL, "docks_male_laborer_mob6175387.cl"},
	{0x43CCEDA56F8DBB2CULL, "llac_lt_undercityserialkiller_-1_37.dds"},
	{0x43CD00096C70B6B9ULL, "hlac_dr_0_deeproads_5849_11y.dds"},
	{0x43CD74CBDEBCC8EBULL, "vxca_caverns1_mines_dni_0_0.vlm"},
	{0x43CD9077C49357B1ULL, "follower_banter6136729.cl"},
	{0x43CE080530B7CF4CULL, "6078466_f.fxe"},
	{0x43CE3B7FB763E1DCULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_rune.mao"},
	{0x43CE477FB763F67FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_rune.mmh"},
	{0x43CE4D7FB7640021ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_rune.msh"},
	{0x43CE941E1F49C9F4ULL, "428871895.ogg"},
	{0x43CED4DBDC9E19B9ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax107a_0n.dds"},
	{0x43CEF6D1A1762B0CULL, "207341414.ogg"},
	{0x43CFC9E2955942F3ULL, "6104435_m.fxe"},
	{0x43D008CCF8153C6EULL, "hlac_dr_0_deeproads_5839_0y.dds"},
	{0x43D0552A8A198D81ULL, "nat380_sibling6068039.cl"},
	{0x43D0597B36803A69ULL, "1048249287.ogg"},
	{0x43D08D427274D428ULL, "hightown_male_cityguard_mob.cnv"},
	{0x43D0B6F394B098EFULL, "vxht_mansion_n_5b6_0_0.vlm"},
	{0x43D16D664C60B2E8ULL, "qun120_arishok6051326.cl"},
	{0x43D1BFDC33ED5803ULL, "w_encounter3_40_0.msh"},
	{0x43D1CDDC33ED705DULL, "w_encounter3_40_0.mmh"},
	{0x43D1D1DC33ED776EULL, "w_encounter3_40_0.mao"},
	{0x43D1D288C842C10CULL, "s_ballista_01_l3.mao"},
	{0x43D1D75088A3C7CDULL, "vxlt_undercitydungeon_1c1_0_1.vlm"},
	{0x43D2292CFBC2AAF4ULL, "vxht_0_hightown_a3_bai_0_0.vlm"},
	{0x43D2E90A745ED51FULL, "gen_im_trs_dwv_junk3_02.uti"},
	{0x43D34F3EAD838D70ULL, "6166205_m.fxe"},
	{0x43D35C2429AFC565ULL, "vxht_mansion2b_3et_0_0.vlm"},
	{0x43D35C5FBC44F0D1ULL, "gen00fl_merrill6107452.cl"},
	{0x43D3AF4272774390ULL, "hightown_male_cityguard_mob.bnk"},
	{0x43D3FC8B1C5C6F8DULL, "t1_mue_wh1.tnt"},
	{0x43D4CE52BD38BDC6ULL, "vxdr_primevaltomb_741_0_0.vlm"},
	{0x43D4D7071B5F2AB6ULL, "916165497.ogg"},
	{0x43D4DDCF40D8B127ULL, "6177419_m.fxe"},
	{0x43D570AF99BC6530ULL, "6150688_m.fxe"},
	{0x43D571BCDD7A081EULL, "vxht_mansion2b_3f2_0_0.vlm"},
	{0x43D5DC254C8C3A6CULL, "art\\vfx\\combat\\vfx_r_arc_burst_shot_act_c.impa.ani"},
	{0x43D5F9A743BD99EAULL, "vxlt_0_lowtown_docks_a3_n_d37_0_0.vlm"},
	{0x43D624E708921ED8ULL, "hlac_lt_0_lowtown_docks_n_65533_3z.dds"},
	{0x43D6A86695C4A812ULL, "56586237.ogg"},
	{0x43D6FC79628D87D2ULL, "865155938.ogg"},
	{0x43D701D7D57AA7FDULL, "vxdr_primevaltomb_8op_0_0.vlm"},
	{0x43D716B9681B65A3ULL, "art\\vfx\\models\\vfx_grenade.msh"},
	{0x43D71F9CEF7DC68CULL, "vxlt_hangedman_n_246_0_0.vlm"},
	{0x43D724B9681B7DFDULL, "art\\vfx\\models\\vfx_grenade.mmh"},
	{0x43D754A0BD3DCEB1ULL, "hlac_lt_0_lowtown_a2_n_1_11x.dds"},
	{0x43D7CB6C1B793E7FULL, "fo_distor.mao"},
	{0x43D8059774C6BF4FULL, "mag100_cullen6140941.cl"},
	{0x43D8A6AAF703DB2AULL, "vxht_0_hightown_a2_ame_0_0.vlm"},
	{0x43D8EAE86BDDFCA9ULL, "zz_qcr_debug6053499.cl"},
	{0x43D8EAEB4FFC121EULL, "dre000_hireling16038642.cl"},
	{0x43D949461F857D0AULL, "6140636_m.fxe"},
	{0x43D9872E9619B84FULL, "art\\vfx\\combat\\vfx_r_arc_scloud_act_c.impa.gad"},
	{0x43D9DAF87242B74DULL, "one140cr_charade6137638.cl"},
	{0x43D9F0E67280785DULL, "365532201.ogg"},
	{0x43DA44A69B05BF1DULL, "299782796.ogg"},
	{0x43DAB8302B1595E4ULL, "665672338.ogg"},
	{0x43DAC527767C1636ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanl.evt"},
	{0x43DB06A6892E6B73ULL, "abi_r_trip.wbd"},
	{0x43DB64F63D6AED4CULL, "6034520_m.fxe"},
	{0x43DBAD53DF496C2CULL, "llac_lt_0_lowtown_docks_a3_-65538_7.dds"},
	{0x43DBADE885EBCA09ULL, "198999918.ogg"},
	{0x43DBE9254C90BBAEULL, "art\\vfx\\combat\\vfx_r_arc_burst_shot_act_c.impa.gad"},
	{0x43DC73353330EDA3ULL, "hlac_ca_caverns2_5802_60z.dds"},
	{0x43DD797776ADE416ULL, "rdr110dg_party_comments6094684.cl"},
	{0x43DDE12037AA6D3CULL, "follower_banter6170391.cl"},
	{0x43DE971718882B99ULL, "9085331.ogg"},
	{0x43DEC209E141387DULL, "6044287_m.fxe"},
	{0x43DEC31F46AA2F37ULL, "hlac_lt_0_lowtown_a3_n_c_-3_6z.dds"},
	{0x43DF15CF38451A8CULL, "hlac_ht_0_hightown_n_65535_3y.dds"},
	{0x43DF1C481B4555D0ULL, "551815939.ogg"},
	{0x43DF436AED81A757ULL, "271875489.ogg"},
	{0x43DFAE52AE67E388ULL, "qun211_braggart6075496.cl"},
	{0x43E067ABA3CF9D5FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_n.dds"},
	{0x43E07618DB49E5A0ULL, "605134911.ogg"},
	{0x43E092803195D1D1ULL, "hlac_lt_0_lowtown_a2_n_c_-1_0y.dds"},
	{0x43E09C2E961FFAFCULL, "art\\vfx\\combat\\vfx_r_arc_scloud_act_c.impa.evt"},
	{0x43E118D0AAD99735ULL, "lt_0_lowtown_a2_n_c_37_0_ll.mmh"},
	{0x43E128F8AABA1302ULL, "6037404_m.fxe"},
	{0x43E15F6B4B95EE8CULL, "llac_lt_0_lowtown_n_-3_0.dds"},
	{0x43E1BB5D1EC747DCULL, "art\\characters\\creatures\\pridedemon\\animation\\cast.ani"},
	{0x43E1BC27768225C1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanl.gad"},
	{0x43E1E9955E44B7E8ULL, "fx_icepoof.mao"},
	{0x43E22C5ABC86FD43ULL, "ser221_party_comments6166197.cl"},
	{0x43E238C822E9FB51ULL, "939726306.ogg"},
	{0x43E292EEC7EE666CULL, "vxlt_0_lowtown_docks_n_b7u_0_1.vlm"},
	{0x43E2E4254C96D18DULL, "art\\vfx\\combat\\vfx_r_arc_burst_shot_act_c.impa.evt"},
	{0x43E3419202DEFE2BULL, "amb_scene_ravens.wbd"},
	{0x43E365F8ECBA1584ULL, "sound_vegetation_type.gda"},
	{0x43E59179E173DF0CULL, "vxlt_hangedman_n_237_0_0.vlm"},
	{0x43E5C8FE118BD807ULL, "263047823.ogg"},
	{0x43E766F96BF1975FULL, "6041505_m.fxe"},
	{0x43E78883F52A757AULL, "hlac_ht_bartrand_n_2995_6x.dds"},
	{0x43E78B2E522FE66DULL, "hlac_lt_0_lowtown_n_-131072_6z.dds"},
	{0x43E7C85D1ECBC91EULL, "art\\characters\\creatures\\pridedemon\\animation\\cast.gad"},
	{0x43E8502588E74A1BULL, "vxht_0_hightown_a3_a7o_0_0.vlm"},
	{0x43E8839EE833F2C2ULL, "109847784.ogg"},
	{0x43E8BD277688466FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanl.ani"},
	{0x43E8BDE191EA322CULL, "vxca_caverns1_edl_0_0.vlm"},
	{0x43E928B84713E146ULL, "158973100.ogg"},
	{0x43E96C4541E3ED8CULL, "m_one300ar_vanard_fight.wbd"},
	{0x43E9F9822614A900ULL, "6027335_m.fxe"},
	{0x43EA5270528FDE3FULL, "mag131_keran6086205.cl"},
	{0x43EBE1ECD888AAF0ULL, "746343320.ogg"},
	{0x43EC67FEA0E41C4AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bedroll_d.dds"},
	{0x43EC8F8D1F36E9BEULL, "vxdr_primevalentrance_2nj_0_0.vlm"},
	{0x43ECC2E64A3BFB0AULL, "35014553.ogg"},
	{0x43ECCD9A9A305D9EULL, "mage_entropiccloud.dds"},
	{0x43ED2299A900915BULL, "6161157_m.fxe"},
	{0x43ED2CB23D55B73DULL, "65370819.ogg"},
	{0x43ED79E020D8FA47ULL, "672461689.ogg"},
	{0x43EDA42E962A9215ULL, "art\\vfx\\combat\\vfx_r_arc_scloud_act_c.impa.ani"},
	{0x43EE1CF4952781B7ULL, "dae000ip_crowd_lm1_drunk2.utp"},
	{0x43EE21FEAFD36D6FULL, "t3_wdg_w24.tnt"},
	{0x43EE3E9F8577735DULL, "dae011cr_drunk_1.utc"},
	{0x43EE66647029D5D6ULL, "gen00fl_isabela6101059.cl"},
	{0x43EE9645CD435AB3ULL, "6173851_m.fxe"},
	{0x43EEA35D1ED1A93DULL, "art\\characters\\creatures\\pridedemon\\animation\\cast.evt"},
	{0x43EF530F9099D372ULL, "vis121_saemus6084919.cl"},
	{0x43EFAA43FEA94D9BULL, "al_mage_t01.gda"},
	{0x43EFFA41A39996E1ULL, "vxht_bartrand_n_6rx_0_0.vlm"},
	{0x43F0E61FD21BAB90ULL, "vxht_bartrand_n_5rj_0_0.vlm"},
	{0x43F0E9DE0DECE0ABULL, "966007418.ogg"},
	{0x43F159F8DAB79CEBULL, "vrc221_bartrand6065322.cl"},
	{0x43F17B92449BB79FULL, "6044630_m.fxe"},
	{0x43F29564FC8E72F8ULL, "vxlt_hangedman_10n_0_0.vlm"},
	{0x43F2977CE37C0ACAULL, "6121992_m.fxe"},
	{0x43F2E2F4A39E7ADCULL, "189775957.ogg"},
	{0x43F2EF56CEE984B5ULL, "151993940.ogg"},
	{0x43F327245DBDBEC4ULL, "vxht_mansion2b_2q0_0_0.vlm"},
	{0x43F35831E59EB689ULL, "art\\vfx\\combat\\vfx_glb_restore_mana_c.mmh"},
	{0x43F37A0DEC3E9113ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\mt_hed_dwfa_0n.dds"},
	{0x43F39244BD8D3060ULL, "vxca_caverns1_jfi_0_0.vlm"},
	{0x43F3B86D42F05595ULL, "6135110_m.fxe"},
	{0x43F3EB9AC4D09CC1ULL, "vxdr_primevaltomb_7r0_0_0.vlm"},
	{0x43F4460B1D6AC4D3ULL, "one280_emrys6137193.cl"},
	{0x43F44EA63E8D64F1ULL, "lgt110_gxa_coa_meeting6117491.cl"},
	{0x43F4CBA0E8FD2203ULL, "860628777.ogg"},
	{0x43F4EB528ED7A033ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pot_sa.dds"},
	{0x43F567F4046414F8ULL, "6036522_m.fxe"},
	{0x43F683BD731814F6ULL, "6034755_m.fxe"},
	{0x43F720D292070B98ULL, "vxlt_0_lowtown_a2_h9v_0_0.vlm"},
	{0x43F7372BF9AB351EULL, "47237252.ogg"},
	{0x43F75E0C67B19D6AULL, "6021209_m.fxe"},
	{0x43F81BF39CF9CA4DULL, "dae000_bartrand6206868.cl"},
	{0x43F837B560225507ULL, "darktown_fem_fer_urch_sta6177742.cl"},
	{0x43F8A945784571ECULL, "59821171.ogg"},
	{0x43F95C372D346D49ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02a.phy"},
	{0x43F976C85934B509ULL, "vxlt_undercitydungeon_yg_0_0.vlm"},
	{0x43F997332921AA9BULL, "6000959_m.fxe"},
	{0x43FA700738DCD6C0ULL, "1070503208.ogg"},
	{0x43FA8FC86BA714D9ULL, "6135789_m.fxe"},
	{0x43FAC51333577B01ULL, "vxht_0_hightown_n_7iw_0_0.vlm"},
	{0x43FC35340A7D6571ULL, "401734349.ogg"},
	{0x43FC8D633F84CF6DULL, "753674933.ogg"},
	{0x43FD24F2F98CC3C5ULL, "92093190.ogg"},
	{0x43FD5D9C0480923CULL, "vxlt_undercityserialkiller_18e_0_0.vlm"},
	{0x43FDA97D46D91FFBULL, "w_blightlands_97_0.msh"},
	{0x43FDBB7D46D93E46ULL, "w_blightlands_97_0.mao"},
	{0x43FDC77D46D952E5ULL, "w_blightlands_97_0.mmh"},
	{0x43FDD6ABD053E9FAULL, "66088835.ogg"},
	{0x43FEFE2291F33564ULL, "vxdr_0_deeproads_cud_0_2.vlm"},
	{0x43FF57FCC411C410ULL, "vxht_bartrand_n_5uk_0_0.vlm"},
	{0x44004FB06DE58DCAULL, "182271726.ogg"},
	{0x4400BE1EB8DF730CULL, "6039545_m.fxe"},
	{0x440110FE8EDD1FF0ULL, "lt_0_lowtown_a2_44_0.rml"},
	{0x44016CCFAB6CD21EULL, "zz_fen_debug6215242.cl"},
	{0x440173DC57370013ULL, "character_character_p228.psh.fxo.dx11"},
	{0x440199014FB3D744ULL, "vxht_mansion2b_2j1_0_0.vlm"},
	{0x4401FFB6D9458924ULL, "llac_lt_0_lowtown_docks_a3_n_-65537_5.dds"},
	{0x440224966FC98255ULL, "vxht_mansion_n_43f_0_0.vlm"},
	{0x44024D6AF781E745ULL, "743581207.ogg"},
	{0x4402838EB3ED074BULL, "vxw_sundermount_2ce_0_0.vlm"},
	{0x4403652689D309AAULL, "6048129_m.fxe"},
	{0x44036B0F4397FF7BULL, "6071493_f.fxe"},
	{0x4403C60766D09578ULL, "vxht_mansion_n_52m_0_0.vlm"},
	{0x4403E7290259798BULL, "ht_0_hightownexplosion_n_18_0.rml"},
	{0x4404185C72D14E0DULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_015a_0t.dds"},
	{0x440432DE6F9744DDULL, "fxe_water_smoke2_add.dds"},
	{0x4404D2EA1C388636ULL, "6085929_f.fxe"},
	{0x44050226A288E24AULL, "266871242.ogg"},
	{0x44050A364F605A02ULL, "hlac_lt_0_lowtown_a2_131070_1x.dds"},
	{0x44051AB4D96CB777ULL, "208139055.ogg"},
	{0x4405474E848C1C31ULL, "6072094_f.fxe"},
	{0x4405D068395454AFULL, "vxht_keep_67i_0_0.vlm"},
	{0x4405E6CA5270FE14ULL, "209941003.ogg"},
	{0x440616DE5054E8A7ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.dura.gad"},
	{0x44064A5CBEE77B5BULL, "6030444_m.fxe"},
	{0x44071303B684340DULL, "vxlt_gallowstemplar_5jd_0_0.vlm"},
	{0x4407F5E87B7DBD1FULL, "gen00fl_aveline6132824.cl"},
	{0x4408239202FDE7F9ULL, "amb_scene_ravens.bnk"},
	{0x44084B9A365E9472ULL, "vxw_bonepitmining_a2_oe_0_0.vlm"},
	{0x4408C959A3638006ULL, "zz_dog_debug6127937.cl"},
	{0x4408CE044927D7E4ULL, "6129411_m.fxe"},
	{0x4409B019F6A80644ULL, "vxw_sundermount_31p_0_1.vlm"},
	{0x4409B50F69C7B5B2ULL, "6036982_m.fxe"},
	{0x440A130E203C1B53ULL, "llac_lt_0_lowtown_docks_n_-65537_5.dds"},
	{0x440A4D16DCE029B0ULL, "dae110ar_lowtown_nt.are"},
	{0x440A554B25F1B6D4ULL, "6148209_m.fxe"},
	{0x440A6CC8D97B054DULL, "365002609.ogg"},
	{0x440AE4DA680979A1ULL, "gen00fl_sebastian6180388.cl"},
	{0x440B16D6DBDB0ED6ULL, "follower_banter6172719.cl"},
	{0x440B1EF452FCE8ECULL, "llac_dr_primevaltomb_661_5.dds"},
	{0x440C4A28BA86961DULL, "lgt160_ntb_elf_conflict6163359.cl"},
	{0x440C6A6106D02F4DULL, "vxca_cavern1_haunted_ckc_0_0.vlm"},
	{0x440C96363840ECC3ULL, "dae000_arishok6082023.cl"},
	{0x440CA0726CB9D9C6ULL, "237401980.ogg"},
	{0x440CE02DE8E4B62EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_loop.ani"},
	{0x440D0BDE505AF4B4ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.dura.evt"},
	{0x440D6416DCE286B2ULL, "dae110ar_lowtown_nt.ncs"},
	{0x440D7ABC38C8C1CCULL, "258691210.ogg"},
	{0x440DC9D9B607DC90ULL, "vxht_bartrand_n_5xh_0_0.vlm"},
	{0x440DDE47C5062568ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_act_c.anb"},
	{0x440E0ACB0E4C438AULL, "vxht_0_hightown_c8q_0_1.vlm"},
	{0x440E93D0685BC2FDULL, "vxlt_0_lowtown_d1w_0_0.vlm"},
	{0x440EA1B565E770CDULL, "follower_banter6169787.cl"},
	{0x440ED09BB125C494ULL, "art\\vfx\\combat\\vfx_r_brd_psong_act_c.impa.ani"},
	{0x440EFE7D230DDEF9ULL, "6174526_m.fxe"},
	{0x440F0981F452F37CULL, "em_comt_servant_dist.mor"},
	{0x440F1628D9E1E1ECULL, "plt_prm000pt_evra_ring.plo"},
	{0x440F7421E375B439ULL, "dae000_flemeth6013386.cl"},
	{0x440FE0EFFB3B32A7ULL, "6061396_m.fxe"},
	{0x44100E36B503D86CULL, "llac_w_bonepit_0_2.dds"},
	{0x44100F178D34DDDBULL, "pro000_colonel_ewald6039957.cl"},
	{0x441062CF01BADA28ULL, "98418008.ogg"},
	{0x44106D9A620BB847ULL, "344490052.ogg"},
	{0x4410AB6C2E5B7DF4ULL, "645848168.ogg"},
	{0x4410D7C653457F81ULL, "vxht_chantry_2dz_0_0.vlm"},
	{0x4410F56BA5E31FCBULL, "vxw_sundermount_2dd_0_0.vlm"},
	{0x44115F752DCA4BD8ULL, "6108757_m.fxe"},
	{0x44116158AA2F82FBULL, "219413551.ogg"},
	{0x44126E96C7786B18ULL, "hlac_lt_0_lowtown_a2_n_65533_0z.dds"},
	{0x441278F93BAAA8C7ULL, "6067655_m.fxe"},
	{0x44128FED1C226460ULL, "hlac_lt_0_lowtown_a2_-3_5x.dds"},
	{0x44138953C58EE38AULL, "gen00fl_fenris6148980.cl"},
	{0x4413AAEFA9EDF1EAULL, "320558720.ogg"},
	{0x441442452B4A6D2FULL, "vxht_keep_60h_0_0.vlm"},
	{0x44145C16DCE897A3ULL, "dae110ar_lowtown_nt.lst"},
	{0x4415A776E1225ACCULL, "384661097.ogg"},
	{0x4415CD9BB12BDEF6ULL, "art\\vfx\\combat\\vfx_r_brd_psong_act_c.impa.gad"},
	{0x4415FB28D9E7D223ULL, "plt_prm000pt_evra_ring.nss"},
	{0x4416B05E7217D716ULL, "gen00fl_dog6169326.cl"},
	{0x4416BD772854ACF2ULL, "vxw_bonepitmining_a2_jf_0_0.vlm"},
	{0x4416C6FAA31519A5ULL, "438874281.ogg"},
	{0x441776DBA0AB760AULL, "540618444.ogg"},
	{0x4417EAB15ABA17F0ULL, "vxht_keep_n_6fb_0_0.vlm"},
	{0x44181C24B6568707ULL, "art\\levels\\races\\ferelden\\props\\_textures\\ossuary_n.dds"},
	{0x441832DB6AFF2CABULL, "hlac_dr_primevaltomb_9621_7z.dds"},
	{0x4418C72817E7CF54ULL, "6145109_m.fxe"},
	{0x44191D9EDF4AC087ULL, "vxw_bonepitmining_a2_85_0_0.vlm"},
	{0x44195E54F9405595ULL, "t3_rob_cir.tnt"},
	{0x4419C6C852339EADULL, "6133592_m.fxe"},
	{0x4419E1BFBED5A2A5ULL, "zz_followers_banter6158451.cl"},
	{0x4419E4A7243D0353ULL, "hlac_lt_gallowstemplar_f_2391_3x.dds"},
	{0x4419E97C1CDA81DAULL, "llac_lt_0_lowtown_docks_-3_4.dds"},
	{0x441A71281D24D805ULL, "108187507.ogg"},
	{0x441A7DB214B9ECF3ULL, "gen00fl_varric6103461.cl"},
	{0x441AB4A316C5349EULL, "307470171.ogg"},
	{0x441B03DE506724EDULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.dura.ani"},
	{0x441B0D611587FFA5ULL, "hlac_lt_0_lowtown_n_65536_6x.dds"},
	{0x441B0FA4EC71AE37ULL, "6064666_f.fxe"},
	{0x441B0FE9E026B3CDULL, "vxht_keep_n_704_0_0.vlm"},
	{0x441B19D50E8D3DD0ULL, "vxht_0_hightown_n_cbk_0_0.vlm"},
	{0x441B40EECD0B71A2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphanga.msh"},
	{0x441B42EECD0B7558ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphanga.mmh"},
	{0x441B4EEECD0B89FBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphanga.mao"},
	{0x441B712098182652ULL, "88601203.ogg"},
	{0x441B84EEE089AF84ULL, "vxlt_0_lowtown_a2_n_ewb_0_0.vlm"},
	{0x441BC4DDF1CA37A1ULL, "hlac_dr_dungeon_3217_11x.dds"},
	{0x441BE59175F0809EULL, "6115757_m.fxe"},
	{0x441C3BB6A7FDF510ULL, "vxht_bartrand_n_5ci_0_0.vlm"},
	{0x441C5531E5C1B6BEULL, "art\\vfx\\combat\\vfx_glb_restore_mana_c.anb"},
	{0x441CCF78BE704369ULL, "art\\vfx\\materials\\chunky\\vfx_c_flat_add.mao"},
	{0x441CE89BB1322B55ULL, "art\\vfx\\combat\\vfx_r_brd_psong_act_c.impa.evt"},
	{0x441D4E415221C946ULL, "w_bonepitdestroyed_a3_104_0.phy"},
	{0x441DA2473CA7DE7DULL, "hlac_lt_0_lowtown_a2_65536_0y.dds"},
	{0x441DB8CE03958616ULL, "ser191_emeric6164743.cl"},
	{0x441DCACB8945EB9DULL, "hlac_lt_0_lowtown_131074_0z.dds"},
	{0x441F086DD3B2CFFEULL, "vxht_chantry_n_2ie_0_0.vlm"},
	{0x441F674897D9384BULL, "vxw_sundermount_2ig_0_0.vlm"},
	{0x441F7A69194E6F5DULL, "946242594.ogg"},
	{0x442023E28E7145EEULL, "zz_getinfo.ncs"},
	{0x442029436FDA8F1BULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_002a_0n.dds"},
	{0x44202E993B17FD6BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wo_basic_1.mao"},
	{0x44207F4F9262A36EULL, "475007771.ogg"},
	{0x4420B32F3F8E65BEULL, "gen00fl_anders6096515.cl"},
	{0x4420ECF678F62485ULL, "vxlt_hangedman_10w_0_0.vlm"},
	{0x44210B4DD3065817ULL, "6111997_f.fxe"},
	{0x44211F94EE64EAC2ULL, "vxw_woundedcoast_1bu_0_1.vlm"},
	{0x4421A2AAF97F542EULL, "mag100_cullen6095103.cl"},
	{0x4421AD15133A9E35ULL, "vxlt_0_lowtown_a2_n_c_ep0_0_0.vlm"},
	{0x4421B92DE8F6D068ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_loop.gad"},
	{0x4421E57CB0C140D7ULL, "vxw_encounter3_fz_0_0.vlm"},
	{0x442326C47AEF4D72ULL, "gen00fl_aveline6122116.cl"},
	{0x44241819F6D1C5F8ULL, "931881118.ogg"},
	{0x44242AE5EB5DD9D2ULL, "6219194_m.fxe"},
	{0x4424576BBD351974ULL, "6104295_f.fxe"},
	{0x4424A6685ADB216AULL, "6145110_m.fxe"},
	{0x4425767566D08503ULL, "6175652_f.fxe"},
	{0x442645372D5B346EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02a.mmh"},
	{0x44265A73311045DBULL, "vxdr_primevaltomb_2un_0_0.vlm"},
	{0x44265C8E4CB03070ULL, "vxht_keep_n_6ac_0_0.vlm"},
	{0x44265F372D5B605CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02a.msh"},
	{0x442733FC1B0B04B7ULL, "static_static_p33.psh.fxo"},
	{0x44273CFE2FD52D5DULL, "vxlt_0_lowtown_a2_e1v_0_0.vlm"},
	{0x44278826BAD5A64BULL, "6120774_m.fxe"},
	{0x44284F6CFB282D21ULL, "298783848.ogg"},
	{0x442873E17E673FFCULL, "ts_qunari_acc_table.gda"},
	{0x4428944CC9B69B24ULL, "hlac_ht_playershouse_n_855_16z.dds"},
	{0x4428B332DA62DBBCULL, "vxlt_undercityserialkiller_17x_0_0.vlm"},
	{0x44292B751E16A27EULL, "185996770.ogg"},
	{0x442996070B64C472ULL, "6088684_m.fxe"},
	{0x4429BEBCC1EF0F3AULL, "45705726.ogg"},
	{0x4429E6A0426940FEULL, "vxlt_0_lowtown_a2_n_c_ei4_0_0.vlm"},
	{0x442A001D7166F0C5ULL, "hlac_lt_hangedman_n_-1_9x.dds"},
	{0x442A3EF71B41F033ULL, "gen_im_trs_cot_lgt_cht_r1.uti"},
	{0x442A514280932240ULL, "one100dg_nabil6054981.cl"},
	{0x442A73973EAE4458ULL, "289515330.ogg"},
	{0x442A7ECA5F7A1205ULL, "63178819.ogg"},
	{0x442AA6B42EF1BD27ULL, "gui_guipixelshader_p3.psh.fxo"},
	{0x442B3246CF524284ULL, "dae200_saemus6091713.cl"},
	{0x442B3F0687B3D805ULL, "gen00fl_sebastian6180296.cl"},
	{0x442B79E5FD9B9423ULL, "6149748_m.fxe"},
	{0x442C8229C4A56535ULL, "6048786_m.fxe"},
	{0x442C864AD74514C1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_varric_.ani"},
	{0x442C885CED8C642CULL, "vxw_bonepit_10c_0_0.vlm"},
	{0x442D11FE8F025C58ULL, "lt_0_lowtown_a2_44_0.anb"},
	{0x442D1B61FB0C7965ULL, "6144778_m.fxe"},
	{0x442D3213E7394E35ULL, "vxlt_0_lowtown_a2_n_c_f4w_0_0.vlm"},
	{0x442D7A4AC5A8E87EULL, "vxht_chantry_n_2df_0_0.vlm"},
	{0x442DAF1F67540B3AULL, "627331912.ogg"},
	{0x442DBB803731B081ULL, "vxht_chantry_2zx_0_0.vlm"},
	{0x442DF1F9384EAE63ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v4.evt"},
	{0x442E5EAD6EB54D0DULL, "6167463_m.fxe"},
	{0x442ED3FFEF3A5C77ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeballopen.gad"},
	{0x442F47966A8DB577ULL, "336197718.ogg"},
	{0x442FDAC1FD6741E6ULL, "dae000_meredith6085228.cl"},
	{0x442FF64D3E0A85CAULL, "6104009_m.fxe"},
	{0x44305759A2B75957ULL, "vxw_encounter3_cu_0_0.vlm"},
	{0x44305B5ADA4ED642ULL, "hlac_w_woundedcoast_131073_0x.dds"},
	{0x443123D215A66ABFULL, "6105384_m.fxe"},
	{0x4431306189E3F6D7ULL, "vxca_caverns1_ih3_0_0.vlm"},
	{0x443198BA0914E6CFULL, "vxlt_gallowsprison_n_6nb_0_0.vlm"},
	{0x44321D98E3384705ULL, "vxht_mansion_n_52t_0_0.vlm"},
	{0x4432324669619257ULL, "vxht_keep_6bn_0_0.vlm"},
	{0x443337313D8D1932ULL, "37117229.ogg"},
	{0x4433914AD74B4663ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_varric_.gad"},
	{0x4433EF5120A6B22BULL, "6102416_f.fxe"},
	{0x443444E206D15389ULL, "712516890.ogg"},
	{0x4434CAF938548AE4ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v4.gad"},
	{0x44351BA2DEA5C483ULL, "1863526.ogg"},
	{0x4435534875E7A1E0ULL, "6152736_m.fxe"},
	{0x44355F060E52703FULL, "6088916_f.fxe"},
	{0x4435E8FFEF409E04ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeballopen.evt"},
	{0x443602057B39130CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_thinking.anb"},
	{0x44361EA4D6B5FC90ULL, "644028097.ogg"},
	{0x4436A47AF4BDA8B2ULL, "follower_barks6173401.cl"},
	{0x4436C8422598C31AULL, "6119619_f.fxe"},
	{0x4436DD47C5292803ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_act_c.mmh"},
	{0x443721D9FB7D5E28ULL, "hlac_lt_0_lowtown_a2_-1_1z.dds"},
	{0x44374DD3AD91A95DULL, "vxw_encounter3_eq_0_0.vlm"},
	{0x4437EE98C6358FD9ULL, "6116412_m.fxe"},
	{0x443873F74FDEC98DULL, "vxlt_0_lowtown_a2_hg9_0_0.vlm"},
	{0x44388740B867825BULL, "vxht_playershouse_n_1zo_0_0.vlm"},
	{0x443909FB69D70FE0ULL, "qcr221_cut_qunari_withdraw.cnv"},
	{0x443A1C329F374F13ULL, "6050176_m.fxe"},
	{0x443B035B4C9211ECULL, "422580670.ogg"},
	{0x443B09C93529EC55ULL, "vxdr_primevaltomb_539_0_0.vlm"},
	{0x443BD9F9385AC2C2ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_attack_v4.ani"},
	{0x443C28BBDDF20409ULL, "isa311_castillon6092110.cl"},
	{0x443C6A5984229D4FULL, "vxht_mansion_n_4zc_0_0.vlm"},
	{0x443C6F954FAD5406ULL, "hlac_lt_0_lowtown_a2_-3_4y.dds"},
	{0x443C7DAFE2C9F43CULL, "6014253_f.fxe"},
	{0x443CD86CDFCA35F9ULL, "6015285_m.fxe"},
	{0x443CEBFB69DAC518ULL, "qcr221_cut_qunari_withdraw.bnk"},
	{0x443CF411216DBCE9ULL, "vxht_bartrand_n_5wp_0_0.vlm"},
	{0x443D7FDDFDD47F74ULL, "prx_black.dds"},
	{0x443E034ED2511BC2ULL, "vxw_woundedcoast_1hw_0_1.vlm"},
	{0x443E220C46640FF2ULL, "6127653_m.fxe"},
	{0x443ED8FC3C8D6256ULL, "828915917.ogg"},
	{0x443F6DBB2EDCB233ULL, "22644999.ogg"},
	{0x443FD3D529400A65ULL, "mag100_karras6052845.cl"},
	{0x443FE59AB7A9CFE7ULL, "vxca_caverns2_8gd_0_0.vlm"},
	{0x44400A96FB0AFF4FULL, "vxlt_gallowsprison_n_6km_0_0.vlm"},
	{0x444034DBF4E674D4ULL, "vxw_encounter1_ct_0_0.vlm"},
	{0x44403D92FB3FDBD4ULL, "vxlt_hangedman_n_rh_0_0.vlm"},
	{0x44405C80F3BCA92EULL, "6201663_m.fxe"},
	{0x444074100C38BDE0ULL, "gen00flst_varric_acttwoc.stg"},
	{0x4441851CD0841ACDULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_light_d.dds"},
	{0x4441952CE42A6E4DULL, "41140045.ogg"},
	{0x4441B9A5889A3C76ULL, "art\\effects\\meshparams.dat"},
	{0x44421345A34DC5B8ULL, "hlac_ht_0_hightown_65536_4x.dds"},
	{0x4442146370D57B8EULL, "dae000_meredith6080986.cl"},
	{0x4442A1F360B42084ULL, "vxca_caverns1_mines_i1j_0_0.vlm"},
	{0x4442BCFB07FAD1C9ULL, "488613186.ogg"},
	{0x4442D3661D5B3BA1ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_4z.dds"},
	{0x444300FFEF4B51FDULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeballopen.ani"},
	{0x44439667CD96E80FULL, "6108093_m.fxe"},
	{0x4443CBD2D261CFB0ULL, "302978412.ogg"},
	{0x4443CEA7A2E1D0C9ULL, "6108842_m.fxe"},
	{0x444470133A5D1579ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_chop.ani"},
	{0x4444931285677EF8ULL, "vxlt_backalley_g_1xt_0_1.vlm"},
	{0x444553E0D63CE31CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_bnt_dblw_lp.gad"},
	{0x4445FF6101A4256FULL, "6078758_m.fxe"},
	{0x44461F0CA77E9D79ULL, "564239938.ogg"},
	{0x444679C0EF50F572ULL, "6086884_m.fxe"},
	{0x4446EE42818D3884ULL, "ht_mansion2_mmi.gff"},
	{0x444715ACC1BACA37ULL, "hlac_lt_0_lowtown_n_-131073_0y.dds"},
	{0x4447567E94AF1F86ULL, "vxht_chantry_n_2bx_0_0.vlm"},
	{0x44479A9D6631FADDULL, "vxht_keep_6ae_0_0.vlm"},
	{0x444814D116B02C07ULL, "vxlt_hangedman_4t_0_0.vlm"},
	{0x44486F58FE351C97ULL, "dae000_arishok6068522.cl"},
	{0x4448F438168A5A5BULL, "vxlt_warehouse_1b7_0_0.vlm"},
	{0x444913EA8C4339FAULL, "vxht_mansion2b_2wh_0_0.vlm"},
	{0x444966D119B4409CULL, "479522069.ogg"},
	{0x44497423BF283A73ULL, "vxht_0_hightown_a2_bph_0_0.vlm"},
	{0x444ABDA3163455A0ULL, "6167347_m.fxe"},
	{0x444AE3519C3C6F5FULL, "hlac_lt_hangedman_-1_16y.dds"},
	{0x444B03BFF13E9A84ULL, "hlac_ht_0_hightown_a2_0_1x.dds"},
	{0x444B23B8D9129778ULL, "vxdr_0_deeproads_cu8_0_2.vlm"},
	{0x444B4B133A62F4BBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_chop.gad"},
	{0x444B8F510CEA4D82ULL, "256852965.ogg"},
	{0x444BD20E639CF964ULL, "ht_0_hightown_13_0.rml"},
	{0x444C09C4CC3A3097ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_knockback.gad"},
	{0x444C52E0D643001AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_bnt_dblw_lp.ani"},
	{0x444C752BC4473442ULL, "vxw_woundedcoast_1sv_0_1.vlm"},
	{0x444CA16177DA1456ULL, "hlac_ht_chantry_2534_47x.dds"},
	{0x444D0B0987A93DF7ULL, "plt_bet100pt_gift.nss"},
	{0x444D518E60A88179ULL, "6116680_f.fxe"},
	{0x444D715CF17CF6FBULL, "802395428.ogg"},
	{0x444E42E4BB5541B7ULL, "6143467_m.fxe"},
	{0x444E5777A99FE867ULL, "vxca_caverns2_8bg_0_0.vlm"},
	{0x444E9C9931EA2158ULL, "follower_banter6165403.cl"},
	{0x444EA6B8E6DC8D54ULL, "vxw_encounter1_hu_0_0.vlm"},
	{0x444EFE99A7DD96A5ULL, "vxdr_0_deeproads_fi6_0_0.vlm"},
	{0x444F0348D4936B4BULL, "vxht_chantry_1yp_0_0.vlm"},
	{0x444F16004D4DC357ULL, "vxht_keep_6hh_0_0.vlm"},
	{0x444F91E3755AFEB3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_roll_shoulders.evt"},
	{0x444F9AF124842BA2ULL, "hlac_ht_mansion_n_3371_1y.dds"},
	{0x444FA3FBB676BD41ULL, "526746586.ogg"},
	{0x444FF2E1D899797BULL, "vxca_caverns2prc_aen_0_0.vlm"},
	{0x4450333A19EB3C2DULL, "ave200pt_long_road.ncs"},
	{0x44517B69C6F3C24EULL, "hlac_w_woundedcoast_131074_0y.dds"},
	{0x445193D9A3E28D7BULL, "w_bonepit.rml"},
	{0x4451C3BD2A8D3027ULL, "vxlt_0_lowtown_docks_b5w_0_1.vlm"},
	{0x4451D927226F45CEULL, "lt_0_lowtown_docks_n_mmi.gff"},
	{0x4451FB2271850C73ULL, "6173110_m.fxe"},
	{0x445253EBBC686F2EULL, "gencr_dalish_archer.utc"},
	{0x445268133A6944A0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_chop.evt"},
	{0x44529F2B901AA005ULL, "279184264.ogg"},
	{0x445308505A073A65ULL, "780523389.ogg"},
	{0x44531C2982336242ULL, "6016801_m.fxe"},
	{0x44531EC4CC407324ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_knockback.evt"},
	{0x4453254548D47859ULL, "and230st_base_first_time.stg"},
	{0x44538B82712C4896ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliff2_n.dds"},
	{0x4453ABFC0984F41AULL, "6113019_f.fxe"},
	{0x4453D299CE7C8934ULL, "534716349.ogg"},
	{0x4454576D8A300257ULL, "vxht_0_hightown_n_bgb_0_0.vlm"},
	{0x4454C036BAE01377ULL, "sky_lt_0_lowtown_docks_a3_n.mao"},
	{0x4454D5DC9F886356ULL, "192684646.ogg"},
	{0x445557B133CAFA8DULL, "vxlt_0_lowtown_a2_hi7_0_0.vlm"},
	{0x44556AFA9C53B35BULL, "vxht_playershouse_n_1ta_0_0.vlm"},
	{0x44560484F5310507ULL, "lgt000cr_dwf_turn_in_1.utc"},
	{0x445667B4476A143AULL, "ca_caverns1_24421_ll.mmh"},
	{0x44568AE3756111F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_roll_shoulders.gad"},
	{0x4456BF8C59343E36ULL, "701946913.ogg"},
	{0x4457000F37AEBAAEULL, "vxlt_0_lowtown_n_gpo_0_0.vlm"},
	{0x44571E760486AE31ULL, "vxlt_gallowstemplar_4na_0_0.vlm"},
	{0x445785C77E39527AULL, "vxht_mansion2b_2pi_0_0.vlm"},
	{0x4457ADFC1D0EDFFBULL, "vxlt_0_lowtown_a2_n_c_hnu_0_0.vlm"},
	{0x4457D00F99129445ULL, "hlac_ht_mansion2_818_3x.dds"},
	{0x4457D8B972D012DEULL, "hlac_ht_0_hightown_-65537_2x.dds"},
	{0x4458C1F7B7A7112DULL, "lt_0_lowtown_docks_n_19_0_lc.mmh"},
	{0x4458C652807E2633ULL, "hlac_lt_0_lowtown_docks_a3_131070_2z.dds"},
	{0x4458DA77F9BD6AFAULL, "llac_lt_backalley_n_65537_3.dds"},
	{0x44596C614DF64CB3ULL, "579983471.ogg"},
	{0x44599595CB08AFF8ULL, "vxdr_0_deeproads_cr9_0_2.vlm"},
	{0x445A076F1FBBD2D2ULL, "6216519_m.fxe"},
	{0x445A2CD2A39D1794ULL, "6115997_m.fxe"},
	{0x445A44E4C25895E2ULL, "243648360.ogg"},
	{0x445A93CCAD7EE8FAULL, "380587641.ogg"},
	{0x445AE07970AE07C0ULL, "one300dg_solivitus6065648.cl"},
	{0x445B1848D088CEDAULL, "vxlt_gallowstemplar_n_34x_0_0.vlm"},
	{0x445B19083EF5305CULL, "vxlt_0_lowtown_a2_ghv_0_0.vlm"},
	{0x445B35F1C3594E75ULL, "one350_variel6076798.cl"},
	{0x445B4DE7950A525DULL, "vxht_0_hightown_n_byz_0_0.vlm"},
	{0x445B665C99196247ULL, "vxw_sundermount_37d_0_0.vlm"},
	{0x445B71FBD58FBC25ULL, "llac_lt_gallowscourtyard_-1_3.dds"},
	{0x445B99A994CAE975ULL, "vxht_mansion2_2os_0_0.vlm"},
	{0x445BE13FDAE7A584ULL, "6143396_m.fxe"},
	{0x445C54553B8744E2ULL, "6134997_f.fxe"},
	{0x445C8635AA6C37C1ULL, "dae00st_rockwraith_reveal.stg"},
	{0x445CFD4AD071805BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grainsack_n.dds"},
	{0x445D3A166BDEF937ULL, "character_character_p99.psh.fxo.dx11.assembly"},
	{0x445D5935A29E0755ULL, "6137040_m.fxe"},
	{0x445D6B13C364A303ULL, "vxlt_0_lowtown_docks_a3_n_d3z_0_0.vlm"},
	{0x445D99E3756749D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_roll_shoulders.ani"},
	{0x445DA58492A98A1FULL, "627891154.ogg"},
	{0x445DA60AA046F2EDULL, "war_hasteafterkills.dds"},
	{0x445E62B4947DF3DDULL, "112414286.ogg"},
	{0x445E7A21F433B8F0ULL, "gen00fl_fenris6126726.cl"},
	{0x445F4AC3BB41073EULL, "art\\characters\\creatures\\spider\\animation\\p.evt"},
	{0x445F9D56D6B71962ULL, "hlac_w_sundermount_-327683_0y.dds"},
	{0x445FD519585EA350ULL, "80646374.ogg"},
	{0x445FE63C843D150FULL, "839840858.ogg"},
	{0x445FFB253ADA61E5ULL, "vxdr_dungeon_212_0_0.vlm"},
	{0x445FFD8D0082AB66ULL, "seb221_ruxton_harimann6177215.cl"},
	{0x44602E4390E4778BULL, "47234533.ogg"},
	{0x446036C4CC4B261DULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_knockback.ani"},
	{0x44609B7131346740ULL, "6126512_f.fxe"},
	{0x4460B35C0B6C99FCULL, "539720907.ogg"},
	{0x4460D43B6F1480ECULL, "follower_banter6141838.cl"},
	{0x44611DA5AD140085ULL, "gen00fl_anders6119710.cl"},
	{0x44611F83A23AFE22ULL, "6119606_f.fxe"},
	{0x4461531A95866B3EULL, "768507768.ogg"},
	{0x446173111349EFEBULL, "llac_lt_0_lowtown_a3_n_c_131069_0.dds"},
	{0x446231265EFC8C2FULL, "mrl310_merrill6143969.cl"},
	{0x44623DD51F806822ULL, "mcr370_cullen6087432.cl"},
	{0x446247AB450C9C1FULL, "hlac_w_woundedcoast_8_0x.dds"},
	{0x4462B149BAEA0D44ULL, "vxlt_blackemporium_wj_0_0.vlm"},
	{0x44637E849D639404ULL, "vxlt_gallowstemplar_f_34n_0_0.vlm"},
	{0x4464129F3ECD5471ULL, "pro000_refugee_female6051852.cl"},
	{0x44648CB8358ED44BULL, "209119702.ogg"},
	{0x4464B49D05B236FBULL, "6170890_m.fxe"},
	{0x44651599BA71BFF9ULL, "w_bonepitmining_a2_21_0.msh"},
	{0x44651F99BA71D0E7ULL, "w_bonepitmining_a2_21_0.mmh"},
	{0x44652399BA71D734ULL, "w_bonepitmining_a2_21_0.mao"},
	{0x44654E5645DC7AF4ULL, "6174589_m.fxe"},
	{0x446661C3BB474D69ULL, "art\\characters\\creatures\\spider\\animation\\p.gad"},
	{0x4466793668BA52C9ULL, "hlac_ca_caverns2_5802_13z.dds"},
	{0x4466803D0D89890CULL, "character_character_p195.psh.fxo.dx11.assembly"},
	{0x44683A436A586532ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_dq101a_0n.dds"},
	{0x44687B2BD2B77D07ULL, "vxlt_0_lowtown_n_ewi_0_0.vlm"},
	{0x4468837E70F666E1ULL, "hlac_lt_0_lowtown_a2_n_c_-3_6y.dds"},
	{0x44689EAF95933014ULL, "6110697_m.fxe"},
	{0x4468CDEA303FCA37ULL, "550823626.ogg"},
	{0x446976C3E0D64FFCULL, "vxlt_0_lowtown_a2_n_c_bxz_0_0.vlm"},
	{0x4469AEA1BEBE8146ULL, "6137007_m.fxe"},
	{0x4469BFC487006ADDULL, "vxht_0_hightown_n_bdy_0_0.vlm"},
	{0x4469D8398B0F7AC7ULL, "vxw_sundermount_32g_0_0.vlm"},
	{0x4469DFF3955D60B7ULL, "423250694.ogg"},
	{0x4469F6F8F90EA714ULL, "lt_gallowscourtyard_n_42_0.rml"},
	{0x446A37F59F340C1DULL, "hlac_lt_0_lowtown_a2_-131072_1x.dds"},
	{0x446A7285D25F59BFULL, "w_encounter2_camp_0_0.rml"},
	{0x446A9C71B9433CF9ULL, "vxlt_gallowstemplar_47f_0_0.vlm"},
	{0x446ABEA613910EACULL, "vxca_caverns1_e7e_0_0.vlm"},
	{0x446B211BBB8849B9ULL, "vxht_bartrand_n_32y_0_0.vlm"},
	{0x446B3B318D8C1967ULL, "vxca_caverns2_8ta_0_0.vlm"},
	{0x446C04295CB9A999ULL, "95080702.ogg"},
	{0x446C515B420B3E62ULL, "416496970.ogg"},
	{0x446C7F2942F0D0EDULL, "vxw_encounter2_qs_0_1.vlm"},
	{0x446CC1137CEF400AULL, "hm_one_nuncio.mor"},
	{0x446D32C3BB4D1BB7ULL, "art\\characters\\creatures\\spider\\animation\\p.ani"},
	{0x446DB778F373E6CFULL, "6075570_f.fxe"},
	{0x446E2250C2D6AFD5ULL, "llac_lt_0_lowtown_-65537_5.dds"},
	{0x446E3A59211CD02DULL, "vxlt_0_lowtown_a2_n_c_hta_0_0.vlm"},
	{0x446E655F69C632B2ULL, "331582552.ogg"},
	{0x446E820044CB3187ULL, "242707010.ogg"},
	{0x446F71C1B793F35EULL, "gen00fl_isabela6079279.cl"},
	{0x446F87EFF8F1B54EULL, "isa211_party_reaction.cnv"},
	{0x446F8C8C89854FA0ULL, "6141853_m.fxe"},
	{0x446FA268ACE4067DULL, "mag320_orsino6038230.cl"},
	{0x446FD2A1EE81F120ULL, "mag291_demon_marethari6029048.cl"},
	{0x446FFB5DEB7FCBE7ULL, "6131348_f.fxe"},
	{0x44702B98E6A591E6ULL, "6142278_m.fxe"},
	{0x44709185D263FA57ULL, "w_encounter2_camp_0_0.phy"},
	{0x4471DC2EEFEFD88BULL, "follower_banter6170044.cl"},
	{0x4471EFCAD22BC83FULL, "dae000_gamlen6121505.cl"},
	{0x44728B1B4A3CEEC9ULL, "6115769_m.fxe"},
	{0x4472A9EFF8F424BEULL, "isa211_party_reaction.bnk"},
	{0x4472F5AD6EE840CDULL, "vxlt_0_lowtown_a2_n_c_e80_0_0.vlm"},
	{0x4472FE3F31CFF993ULL, "hlac_ht_bartrand_n_3371_0z.dds"},
	{0x4473990CCBB44EABULL, "575002509.ogg"},
	{0x4473C09BE110BC6CULL, "6072258_m.fxe"},
	{0x4473E3C91B9AEBAEULL, "vxlt_0_lowtown_n_gja_0_0.vlm"},
	{0x44752B64E9E06323ULL, "framebuffereffects_composedof_p0.vsh.fxo.dx11.assembly"},
	{0x44754DA0D812FAA1ULL, "vxdr_0_deeproads_7tx_0_0.vlm"},
	{0x44759EFED6753122ULL, "32763593.ogg"},
	{0x44765FB6D434DCEEULL, "art\\characters\\creatures\\buzzard\\animation\\blendtree\\buzzardgliding.anb"},
	{0x4476BF77EC8ADC0AULL, "6114058_f.fxe"},
	{0x4476C51102578444ULL, "6121935_f.fxe"},
	{0x44770F4D62224FF9ULL, "gen00fl_merrill.utc"},
	{0x44777CA387EB845DULL, "814387378.ogg"},
	{0x4477FCC222E1615CULL, "vxlt_0_lowtown_a2_gbt_0_0.vlm"},
	{0x447831A178F6835DULL, "vxht_0_hightown_n_box_0_0.vlm"},
	{0x44789C7D5B14118BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose2.evt"},
	{0x44794245486F8B15ULL, "6130001_m.fxe"},
	{0x44795AA79C434638ULL, "6130992_m.fxe"},
	{0x44798CF2702FF37CULL, "6142834_m.fxe"},
	{0x447A573835146804ULL, "6075769_m.fxe"},
	{0x447A84191F30090CULL, "6112673_m.fxe"},
	{0x447B18626A5F246EULL, "vxlt_0_lowtown_docks_n_cui_0_0.vlm"},
	{0x447C043FF95EB06EULL, "vxht_mansion2b_1kw_0_0.vlm"},
	{0x447C29442EA15060ULL, "hlac_ht_keep_n_4519_12z.dds"},
	{0x447C3449B892A7F3ULL, "unbroken_circle.dds"},
	{0x447C841420158A6FULL, "139252407.ogg"},
	{0x447D7C6A8E05849CULL, "6187842_m.fxe"},
	{0x447DA45B66E3840BULL, "485484895.ogg"},
	{0x447DEBC44C217FDFULL, "274000581.ogg"},
	{0x447DEEFA55F63D2AULL, "hlac_lt_0_lowtown_n_-3_7y.dds"},
	{0x447E097F06C97DEBULL, "893286292.ogg"},
	{0x447E3BCBD3FB2380ULL, "vxlt_gallowsprison_n_6c6_0_0.vlm"},
	{0x447E9965EF60BED0ULL, "vxlt_0_lowtown_docks_b99_0_0.vlm"},
	{0x447F49ECD86C0409ULL, "vrc321ip_spooky_2_owner.utp"},
	{0x447FB57D5B1A5B4CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose2.gad"},
	{0x447FCE479857753DULL, "hlac_w_sundermount_327684_0z.dds"},
	{0x447FFBF13C1F2E61ULL, "522485731.ogg"},
	{0x44802990B76FC79AULL, "6148834_m.fxe"},
	{0x4480378D86CD1F4FULL, "6077755_m.fxe"},
	{0x44804674E22FAFDCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_archopen.mao"},
	{0x44804E18A4A8DDCBULL, "art\\vfx\\combat\\vfx_w_rev_incense_act_c.mmh"},
	{0x44805274E22FC47FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_archopen.mmh"},
	{0x44805874E22FCE21ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_archopen.msh"},
	{0x4480843169EA5C2CULL, "6043899_f.fxe"},
	{0x44808E646DC472ADULL, "vxht_bartrand_n_4cy_0_0.vlm"},
	{0x448142BEDA988035ULL, "follower_banter6151344.cl"},
	{0x44814F4B455C77E4ULL, "6174714_m.fxe"},
	{0x4481EEAEC8B56C55ULL, "dae300_bran6144939.cl"},
	{0x448203BFB1F6DA93ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medd.phy"},
	{0x44821CF853E5B310ULL, "114464435.ogg"},
	{0x448255A60D91042EULL, "vxlt_0_lowtown_n_gob_0_0.vlm"},
	{0x44827D6B82ADE1C5ULL, "vxht_mansion2b_2pb_0_0.vlm"},
	{0x448287A8D65FC09DULL, "w_encounter3_115_0.phy"},
	{0x4482950AB4961292ULL, "727055507.ogg"},
	{0x4482E24746436178ULL, "gen00fl_aveline6053269.cl"},
	{0x4483474B242299C8ULL, "6218818_m.fxe"},
	{0x44838B5B36460651ULL, "6116532_f.fxe"},
	{0x4483A4D83C04E370ULL, "227522765.ogg"},
	{0x4483B65ED8B3A5B9ULL, "gen00fl_aveline6105385.cl"},
	{0x4483D8ACB7BDB3C1ULL, "llac_ca_caverns1_mines_24422_3.dds"},
	{0x448405A0650AC5E3ULL, "vxht_0_hightown_a3_du_0_0.vlm"},
	{0x448445AA5948966AULL, "vxca_caverns1_mines_ija_0_0.vlm"},
	{0x4484C4D7901C8F3AULL, "6061813_f.fxe"},
	{0x44855EE5B6A3AE07ULL, "vxlt_0_lowtown_n_emk_0_0.vlm"},
	{0x4485ADDAEA422441ULL, "oip_rock_02.mao"},
	{0x4485C39C5349DCA9ULL, "vxlt_gallowstemplar_6t9_0_0.vlm"},
	{0x44864890F4F55AAAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rail8m.mao"},
	{0x44864C90F4F561F9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rail8m.mmh"},
	{0x44865690F4F572E7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rail8m.msh"},
	{0x4486B47D5B20784AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose2.ani"},
	{0x448705E09564BA54ULL, "vxht_mansion_n_5cb_0_0.vlm"},
	{0x44876FF6410F1EC1ULL, "730332652.ogg"},
	{0x4487ECC0E8AB779AULL, "lla_ht_0_hightown_65535_0.dds"},
	{0x44887B4C60DB6B55ULL, "vxw_bonepit_l9_0_0.vlm"},
	{0x448B014C5C0CDDEBULL, "vxht_0_hightown_a3_bhy_0_0.vlm"},
	{0x448B1A34C2236C2DULL, "801348283.ogg"},
	{0x448B247B501C439FULL, "211912083.ogg"},
	{0x448B6CB04CBD58DFULL, "hlac_lt_gallowstemplar_f_2384_6y.dds"},
	{0x448B9E0F6E46FCFFULL, "gen00flst_isa_multi_act.stg"},
	{0x448BC8F94C7E52BFULL, "w_encounter2_51_0.rml"},
	{0x448BD554F55C4B8DULL, "hlac_lt_0_lowtown_n_-131073_1x.dds"},
	{0x448CC838ED370EA2ULL, "scaleform_sf_main_p0.psh.fxo"},
	{0x448D1E275E4D728EULL, "549226332.ogg"},
	{0x448D2C281A4D6AE6ULL, "atl_controller_dxt5_dat.xml"},
	{0x448D372B4D4E935CULL, "hlac_ca_cavern1_haunted_24424_26z.dds"},
	{0x448DD03D781D873CULL, "vxlt_undercityserialkiller_1js_0_0.vlm"},
	{0x448DE1FC87DB9130ULL, "67980637.ogg"},
	{0x448E2F18A4B4E630ULL, "art\\vfx\\combat\\vfx_w_rev_incense_act_c.anb"},
	{0x448ECF3BBD62B810ULL, "435087926.ogg"},
	{0x448ED19B16DE7285ULL, "122416106.ogg"},
	{0x448EE56152ADA053ULL, "art\\characters\\playercharacter\\humanmale\\glv_maso.phy"},
	{0x448F173C87C7F470ULL, "hlac_lt_0_lowtown_a2_n_c_131070_3y.dds"},
	{0x448F201B8842F3DDULL, "hlac_ht_playershouse_n_855_8x.dds"},
	{0x449130E2D8B8CDDFULL, "llac_lt_0_lowtown_a2_0_2.dds"},
	{0x4491C3A5A9A0EBC3ULL, "695112928.ogg"},
	{0x4491E7F94C82F357ULL, "w_encounter2_51_0.phy"},
	{0x449208A3552FCBBBULL, "dae000_gamlen6149424.cl"},
	{0x4492AAD9A41A4B4DULL, "w_bonepit.arl"},
	{0x44931B0F2B837FC9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas_d.dds"},
	{0x449348621461A7C8ULL, "556084343.ogg"},
	{0x44937CA8D66E2109ULL, "w_encounter3_115_0.mao"},
	{0x449388A8D66E35A2ULL, "w_encounter3_115_0.mmh"},
	{0x44938AA8D66E3958ULL, "w_encounter3_115_0.msh"},
	{0x44938D214AC5A076ULL, "422798257.ogg"},
	{0x4493FEBDB878AE97ULL, "6088329_f.fxe"},
	{0x44942CAD65B27020ULL, "521456100.ogg"},
	{0x44951DB429513F30ULL, "vxht_mansion_n_4xj_0_0.vlm"},
	{0x4495293D2E826735ULL, "609232391.ogg"},
	{0x44953BCB0BCCEB19ULL, "223800086.ogg"},
	{0x44957CAF3F937A36ULL, "hlac_w_encounter2_camp_5_0y.dds"},
	{0x4495DBB491335C4DULL, "6162110_m.fxe"},
	{0x4495EF1F2AD5E75FULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst.gad"},
	{0x449617F8F9341A14ULL, "lt_gallowscourtyard_n_42_0.anb"},
	{0x449679B324B16465ULL, "vxlt_gallowstemplar_f_3ju_0_0.vlm"},
	{0x4497EE0987E8F7D8ULL, "plt_bet100pt_gift.plo"},
	{0x44981F99BA9D2A44ULL, "w_bonepitmining_a2_21_0.rml"},
	{0x4498BDF1A039A130ULL, "vxlt_gallowsprison_n_6mh_0_0.vlm"},
	{0x4498F54D623F09A4ULL, "gen00fl_merrill.cnv"},
	{0x44990076486D99B5ULL, "6123360_m.fxe"},
	{0x44996595676D3578ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyo_0d.dds"},
	{0x4499EB75DC60F470ULL, "vxht_keep_n_69k_0_0.vlm"},
	{0x449A10E87EB9C549ULL, "6169915_m.fxe"},
	{0x449A5520A1DF02C9ULL, "vxht_mansion_n_5a9_0_0.vlm"},
	{0x449AD5BA1001E61AULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.cess.gad"},
	{0x449AD9077942BE95ULL, "llac_lt_0_lowtown_196606_0.dds"},
	{0x449ADBE0C5079B74ULL, "plt_cod_cha_player.plo"},
	{0x449BC49A7555CF7FULL, "6107074_m.fxe"},
	{0x449C174D62417904ULL, "gen00fl_merrill.bnk"},
	{0x449C17CEF9E450E9ULL, "1047816440.ogg"},
	{0x449C41F86D5A48E1ULL, "1066760498.ogg"},
	{0x449CE41F2ADBF34CULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst.evt"},
	{0x449DA3B3E61DAC97ULL, "6218637_m.fxe"},
	{0x449E9837ADFC9BACULL, "m_dae111ar_gamlens_house_nt.bnk"},
	{0x449EC0BD1988B11BULL, "962987123.ogg"},
	{0x449ECAEDE87025B5ULL, "zz_isa_debug6175185.cl"},
	{0x449ED2164ABF1F80ULL, "hlac_w_sundermount_-65534_0y.dds"},
	{0x449F18EF711AA8DDULL, "206378656.ogg"},
	{0x449F51331F171DD9ULL, "ser221_gascard6133865.cl"},
	{0x449F704A5A860E92ULL, "vxlt_warehouse_n_13r_0_0.vlm"},
	{0x449F93475CC2AA8FULL, "hlac_w_bonepit_0_2y.dds"},
	{0x44A01494C6F3E42DULL, "755536240.ogg"},
	{0x44A0687A9C205E52ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.dura.gad"},
	{0x44A0A44D110B22D9ULL, "423400660.ogg"},
	{0x44A0DA8EAF851D07ULL, "hightown_male_cityguard_mob6170012.cl"},
	{0x44A0FDC19EF0A6A6ULL, "llac_ht_0_hightown_-131071_1.dds"},
	{0x44A1194F9D3168D6ULL, "llac_w_encounter3_-196608_0.dds"},
	{0x44A15DA43406BC71ULL, "6094718_m.fxe"},
	{0x44A16FE6FFBBD3FFULL, "6061801_f.fxe"},
	{0x44A1D8BA10080A31ULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.cess.evt"},
	{0x44A1FA7C7359296CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxined.msh"},
	{0x44A2007C7359335EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxined.mmh"},
	{0x44A20C7C735947F5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxined.mao"},
	{0x44A22B8794FB3147ULL, "6108242_m.fxe"},
	{0x44A240C7CA8EAA37ULL, "ctplt_glow.mao"},
	{0x44A2AB904DB3476BULL, "dae020cr_chantry3.utc"},
	{0x44A2EDDAD3825A5BULL, "6075398_f.fxe"},
	{0x44A3129A6D374317ULL, "370029057.ogg"},
	{0x44A38759316C4836ULL, "zz_vault_debug6177800.cl"},
	{0x44A3B17DFD075C0CULL, "625473106.ogg"},
	{0x44A3E17321FB3333ULL, "vxw_bonepitmining_a2_zs_0_0.vlm"},
	{0x44A43A148CEE7161ULL, "lt_0_lowtown_docks_a3_n_24_0.anb"},
	{0x44A465BB159E16CEULL, "w_bonepitdestroyed_a3_222_0.mmh"},
	{0x44A471BB159E2B65ULL, "w_bonepitdestroyed_a3_222_0.mao"},
	{0x44A47FBB159E42BCULL, "w_bonepitdestroyed_a3_222_0.msh"},
	{0x44A49F318DAB9F1CULL, "6053592_m.fxe"},
	{0x44A4EB9016A77CE5ULL, "vxlt_gallowstemplar_f_3qt_0_0.vlm"},
	{0x44A4F2EEEB47C3D6ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdgc_0n.dds"},
	{0x44A4FEE243F764DAULL, "vxca_caverns2_95f_0_0.vlm"},
	{0x44A510B280843223ULL, "6155381_m.fxe"},
	{0x44A536F4B2899CFBULL, "vxlt_hangedman_n_20y_0_0.vlm"},
	{0x44A5451C97B7A8A4ULL, "hlac_w_woundedcoast_-196601_0y.dds"},
	{0x44A596AAB37D2CA2ULL, "6162522_m.fxe"},
	{0x44A59FEE9309AA45ULL, "496686664.ogg"},
	{0x44A5E9D85E288594ULL, "vxlt_0_lowtown_docks_a3_cw1_0_0.vlm"},
	{0x44A664EA1600368FULL, "hlac_lt_undercityencounter_-1_45z.dds"},
	{0x44A66B7A9C24CE89ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.dura.evt"},
	{0x44A6C92AFB42EBB8ULL, "6109260_m.fxe"},
	{0x44A6CD5C3BFDF3B4ULL, "vxlt_0_lowtown_a3_n_c_do0_0_0.vlm"},
	{0x44A6D952BC7768C8ULL, "vxht_bartrand_n_66q_0_0.vlm"},
	{0x44A6E72DE48E0C52ULL, "llac_ht_keep_n_3732_9.dds"},
	{0x44A6F30A3F7ED1A9ULL, "fen110_fenris6096210.cl"},
	{0x44A70128CA71BF9EULL, "vxlt_gallowstemplar_f_3r2_0_0.vlm"},
	{0x44A704CFDB3D816EULL, "6105424_m.fxe"},
	{0x44A72F2473539CDAULL, "436592997.ogg"},
	{0x44A7AEE654E4DA68ULL, "281183867.ogg"},
	{0x44A7B283EA289F74ULL, "hlac_ht_keep_4519_9y.dds"},
	{0x44A85D52CE570CF0ULL, "vxht_keep_n_64h_0_0.vlm"},
	{0x44A93AFE76227E9DULL, "6169714_f.fxe"},
	{0x44A9566C0A428199ULL, "6071188_m.fxe"},
	{0x44A9582545690752ULL, "mer320_hubert6083764.cl"},
	{0x44A959ABAD30B635ULL, "6135129_m.fxe"},
	{0x44A9896829492F75ULL, "907018327.ogg"},
	{0x44A99C549E825652ULL, "292411533.ogg"},
	{0x44A9CC1F2AE654A5ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst.ani"},
	{0x44AA4969E695576EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_brokewood.phy"},
	{0x44AA6FB580E50EE3ULL, "hlac_w_encounter3_-327676_0x.dds"},
	{0x44AA97DDB326285DULL, "dae000cr_grand_cleric6147691.cl"},
	{0x44AB35F1DBBF9CD2ULL, "6056917_m.fxe"},
	{0x44AB3A6CDE138C7EULL, "vxca_caverns2_78v_0_1.vlm"},
	{0x44AB488B4B0431CAULL, "rdr111dg_martin6053594.cl"},
	{0x44ABA198A4457DBDULL, "6164466_m.fxe"},
	{0x44ABC279ED18D561ULL, "vxht_bartrand_n_63k_0_0.vlm"},
	{0x44AC95F32AAB32DBULL, "gen00fl_anders6214493.cl"},
	{0x44ACCA8914308899ULL, "501440566.ogg"},
	{0x44ACD6AFF49439EAULL, "317115282.ogg"},
	{0x44ACE3FC62A880DCULL, "dae000_gamlen6048511.cl"},
	{0x44ACF3AD321BBF49ULL, "vxw_bonepitdestroyed_a3_25b_0_0.vlm"},
	{0x44AD060C508E0137ULL, "vxlt_gallowscourtyard_641_0_0.vlm"},
	{0x44ADE2274C7C2712ULL, "vxlt_warehouse_n_18s_0_0.vlm"},
	{0x44ADEF105F82C961ULL, "6101254_m.fxe"},
	{0x44AE2F69B745D254ULL, "vxca_caverns1_mines_fmi_0_0.vlm"},
	{0x44AE4262F2E696BCULL, "hlac_w_encounter2_1_0y.dds"},
	{0x44AEB674E98CEC0FULL, "combat_in_mv_default.gda"},
	{0x44AEEEE87DB7F06EULL, "vxlt_0_lowtown_a2_n_gze_0_0.vlm"},
	{0x44AEF7486DD8125BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha02a.phy"},
	{0x44AF47201B3F5A12ULL, "6083400_f.fxe"},
	{0x44AF583B1487E84CULL, "dae000_arishok6070935.cl"},
	{0x44AF7A4A8E451948ULL, "lgt110_gxa_vgk_silver_order6116811.cl"},
	{0x44AF8CE804B16414ULL, "6045627_m.fxe"},
	{0x44AFF0BA10147000ULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.cess.ani"},
	{0x44B13A27F1A80ED8ULL, "ser221_gamlen6131654.cl"},
	{0x44B2035342A7AB1FULL, "6034928_f.fxe"},
	{0x44B2063CF157FC78ULL, "380385382.ogg"},
	{0x44B2870BE92F3A05ULL, "vxca_caverns2prc_8kc_0_0.vlm"},
	{0x44B305E55065ACA3ULL, "53684797.ogg"},
	{0x44B3851799345630ULL, "hlac_lt_0_lowtown_a2_n_131070_0z.dds"},
	{0x44B38CFEC2C41231ULL, "vxht_0_hightown_c8g_0_0.vlm"},
	{0x44B3A9BEFD45CAF7ULL, "hlac_ca_caverns1_24425_11y.dds"},
	{0x44B3D824B0B8A990ULL, "vxht_mansion2b_1t4_0_0.vlm"},
	{0x44B40887A5734522ULL, "6167622_m.fxe"},
	{0x44B4291A4D72EBFBULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb08a.mmh"},
	{0x44B4471A4D731EE5ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb08a.msh"},
	{0x44B4537A9C30E3F8ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.dura.ani"},
	{0x44B4593194521BD3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic_2.mao"},
	{0x44B45BB5501E9E14ULL, "vxlt_0_lowtown_docks_a3_cz6_0_0.vlm"},
	{0x44B4630224EB2380ULL, "vxht_mansion_n_4v4_0_0.vlm"},
	{0x44B4729164069B62ULL, "fen110_fenris6100376.cl"},
	{0x44B47A150D7F29C9ULL, "vxlt_hangedman_1r9_0_0.vlm"},
	{0x44B541C0C64BDF1CULL, "vxca_cavern1_haunted_k2t_0_1.vlm"},
	{0x44B54B2FAE6D8148ULL, "vxht_bartrand_n_63n_0_0.vlm"},
	{0x44B54D78BC1787A9ULL, "vxht_mansion_n_6nj_0_1.vlm"},
	{0x44B5D6B19EC6FDA6ULL, "6182722_m.fxe"},
	{0x44B5E24D1DFE2AD6ULL, "class_abilities_carver.gda"},
	{0x44B5EBB71F7C0E78ULL, "vxca_cave_207_0_0.vlm"},
	{0x44B5FE3C91FA6782ULL, "dae000_mother6148960.cl"},
	{0x44B666C9D0573F5CULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.impa.ani"},
	{0x44B6951923BBE838ULL, "sb03_front.dds"},
	{0x44B6E07FB12C0B9CULL, "6048343_m.fxe"},
	{0x44B6F5AD9544EBF4ULL, "500183713.ogg"},
	{0x44B70C24406A2154ULL, "hlac_lt_undercityserialkiller_-1_15z.dds"},
	{0x44B72B7B207CA81BULL, "948071392.ogg"},
	{0x44B740628910B90DULL, "hlac_lt_0_lowtown_65535_4z.dds"},
	{0x44B7ACC5463C4069ULL, "vxht_mansion2b_1b7_0_0.vlm"},
	{0x44B803FE5C18EA2AULL, "6201537_m.fxe"},
	{0x44B82DD6594B6898ULL, "mer221_sabin6042708.cl"},
	{0x44B86685D2A166AFULL, "w_encounter2_camp_0_0.mao"},
	{0x44B871005ADC672AULL, "6135207_m.fxe"},
	{0x44B87285D2A17B4CULL, "w_encounter2_camp_0_0.mmh"},
	{0x44B87885D2A1853EULL, "w_encounter2_camp_0_0.msh"},
	{0x44B8DB761E86105BULL, "486919853.ogg"},
	{0x44B8FBF14AD5C2B5ULL, "zz_lgt_1_debug6136984.cl"},
	{0x44B909DEBDACCADDULL, "pro000_captain_wright6047482.cl"},
	{0x44B90F0CFF6C2852ULL, "water_water_p1.vsh.fxo.dx11"},
	{0x44B92724626DF384ULL, "6110917_f.fxe"},
	{0x44B9376814F5AB6CULL, "74164143.ogg"},
	{0x44B9FF42CA596A3CULL, "923429144.ogg"},
	{0x44BA054ECCF866F3ULL, "vxlt_0_lowtown_a2_hdn_0_0.vlm"},
	{0x44BA61440E82C480ULL, "gen00fl_anders6201146.cl"},
	{0x44BA78975FE5392EULL, "vxca_caverns2prc_7qi_0_0.vlm"},
	{0x44BA8400DBCF07D7ULL, "6175044_m.fxe"},
	{0x44BAC0701EDEF1D1ULL, "vxca_cave_1nw_0_0.vlm"},
	{0x44BAF8AC19F67A08ULL, "t3_rob_pl3.tnt"},
	{0x44BAFF4D7FE96E2BULL, "629764593.ogg"},
	{0x44BB3E987A1E7C32ULL, "llac_w_encounter2_camp_-65532_0.dds"},
	{0x44BC557D3D98D755ULL, "vxca_caverns2prc_9fq_0_0.vlm"},
	{0x44BC73C9D05BC09EULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.impa.gad"},
	{0x44BCA1529DA982CAULL, "740427958.ogg"},
	{0x44BDED069FBF8F83ULL, "gen_im_trs_cot_arc_sbw_r1.uti"},
	{0x44BE01EAC9BF8B67ULL, "zz_mcr_debug6080880.cl"},
	{0x44BE5553345B4669ULL, "6043538_m.fxe"},
	{0x44BEC7A6CCE4DE7CULL, "ss_fenris.bnk"},
	{0x44C00F8DD574D402ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml103a_0t.dds"},
	{0x44C04401D6D056B6ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.cess.ani"},
	{0x44C05694735CFE33ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.dura.evt"},
	{0x44C082F476E78FB8ULL, "100897203.ogg"},
	{0x44C0C52D05E76433ULL, "vxw_bonepitmining_a2_lu_0_0.vlm"},
	{0x44C0F8E8DB255285ULL, "vxca_caverns2prc_8lb_0_0.vlm"},
	{0x44C15F11E6D6F18AULL, "zz_dae_debug6049941.cl"},
	{0x44C19587425325EFULL, "95152072.ogg"},
	{0x44C21B7FF53D54F9ULL, "vxdr_primevaltomb_1mw_0_0.vlm"},
	{0x44C2252AEE72A93FULL, "6033386_m.fxe"},
	{0x44C2EBF1FF754249ULL, "vxlt_hangedman_1y8_0_0.vlm"},
	{0x44C2EC9083BC9D00ULL, "fxc_puke_dust.mao"},
	{0x44C31582598808D0ULL, "vxdr_primevaltomb_8p4_0_0.vlm"},
	{0x44C34EC9D061A0BDULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.impa.evt"},
	{0x44C38FD8ED426A00ULL, "dae000_meredith6085327.cl"},
	{0x44C39F7AD291E6F7ULL, "892389769.ogg"},
	{0x44C4393ACB012D7AULL, "862042251.ogg"},
	{0x44C45D94117226F8ULL, "vxca_cave_254_0_0.vlm"},
	{0x44C4B89DF6B572C8ULL, "hlac_lt_0_lowtown_a2_n_-65536_0x.dds"},
	{0x44C55EB317F9BE0EULL, "w_sundermount.arl"},
	{0x44C58A58C0CE866FULL, "6101057_f.fxe"},
	{0x44C5FFD7C82BA10FULL, "144232210.ogg"},
	{0x44C6400D54DDF301ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1_up.ani"},
	{0x44C6B67FAB0E5E95ULL, "vxca_caverns2_6ha_0_0.vlm"},
	{0x44C6BBF255EA454BULL, "llac_w_woundedcoast_262152_0.dds"},
	{0x44C6CE3FD2AEBA26ULL, "6016486_f.fxe"},
	{0x44C74F9473631174ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.dura.gad"},
	{0x44C7B59E41DCA86BULL, "vxht_0_hightown_a3_au5_0_0.vlm"},
	{0x44C7BAA15F8C8A75ULL, "lgt311_shw_mer_mettin6156461.cl"},
	{0x44C86424937BB92AULL, "hlac_dr_0_deeproads_5849_0z.dds"},
	{0x44C8752ABDF08E2DULL, "gencr_dalish_archer_f.utc"},
	{0x44C8A633D1050661ULL, "vxht_bartrand_n_69i_0_0.vlm"},
	{0x44C8EA7451DB51AEULL, "vxca_caverns2prc_7tj_0_0.vlm"},
	{0x44C919670051ABC8ULL, "vxlt_0_lowtown_a2_enw_0_1.vlm"},
	{0x44C929B22D012262ULL, "vxht_chantry_216_0_0.vlm"},
	{0x44C93D2D370A937BULL, "502226868.ogg"},
	{0x44C96EE0CF6AB2E7ULL, "6050095_m.fxe"},
	{0x44C9C45BBE892D63ULL, "zz_qun_debug6045382.cl"},
	{0x44CA1847DB373F2CULL, "vxw_bonepitmining_a2_255_0_0.vlm"},
	{0x44CA85A3B4F1D3D1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1ladder.mmh"},
	{0x44CA91A3B4F1E872ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1ladder.mao"},
	{0x44CA9DBA36B866D8ULL, "vxht_chantry_30r_0_0.vlm"},
	{0x44CA9FA3B4F1FF8FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1ladder.msh"},
	{0x44CBB7FCC5F3B6FEULL, "vxlt_0_lowtown_a2_n_c_hog_0_0.vlm"},
	{0x44CBC22499A34386ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga102a_0n.dds"},
	{0x44CC3AEDC3C8E7D3ULL, "vfx_bird_raven02.dds"},
	{0x44CC4B0D54E271A3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1_up.gad"},
	{0x44CCCA46375FCA4AULL, "hlac_lt_0_lowtown_a2_n_65536_7z.dds"},
	{0x44CCE6B107A6ACBEULL, "zz_followers_banter6158784.cl"},
	{0x44CE1EF78344A1CFULL, "vxlt_backalley_n_1pm_0_2.vlm"},
	{0x44CE4C01D6DCA1A7ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.cess.evt"},
	{0x44CE4DB24075CB04ULL, "qcr221_cut_heros_reward6043148.cl"},
	{0x44CE5E9473694952ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.dura.ani"},
	{0x44CEE5163F2108AAULL, "mag211_marethari6021341.cl"},
	{0x44CF268658696A68ULL, "zz_mer_debug6017191.cl"},
	{0x44CF2B29B17F0B95ULL, "lt_0_lowtown_a2_n_13_0.rml"},
	{0x44CF3709F7DD7CB3ULL, "vxw_bonepitmining_a2_kt_0_0.vlm"},
	{0x44CF5C5AD2EA6820ULL, "vxlt_0_lowtown_docks_n_bf7_0_0.vlm"},
	{0x44CF9D04E79E84D5ULL, "6025310_m.fxe"},
	{0x44CF9E95D4EA9842ULL, "6087087_m.fxe"},
	{0x44CFA78F53F295E4ULL, "character_character_p135.psh.fxo.dx11"},
	{0x44D003E1D5F103B1ULL, "638519551.ogg"},
	{0x44D08984A1EB4A0DULL, "mer340_cowan6083523.cl"},
	{0x44D0DADC8880710DULL, "262781531.ogg"},
	{0x44D12EF53084BCE6ULL, "vxht_mansion2b_3d0_0_0.vlm"},
	{0x44D15BEF07850061ULL, "follower_banter6163774.cl"},
	{0x44D17DB9583215FFULL, "255194513.ogg"},
	{0x44D1C531761A5BB5ULL, "1006080235.ogg"},
	{0x44D22EE99259B248ULL, "vxht_bartrand_n_69l_0_0.vlm"},
	{0x44D23C491856B5ACULL, "hlac_lt_0_lowtown_a3_n_c_-2_3y.dds"},
	{0x44D242D232D8851EULL, "507042110.ogg"},
	{0x44D287A185B97061ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_loop.ani"},
	{0x44D3280D54E85488ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1_up.evt"},
	{0x44D3DF8228708DC4ULL, "6024539_f.fxe"},
	{0x44D4401EC8856400ULL, "vxlt_0_lowtown_a2_n_h1y_0_0.vlm"},
	{0x44D46D01D6E145F0ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.cess.gad"},
	{0x44D4907F2A287169ULL, "vxht_mansion2b_1x5_0_0.vlm"},
	{0x44D4D5AA4B236D89ULL, "vxlt_0_lowtown_docks_a3_n_d0a_0_0.vlm"},
	{0x44D5BD04EFD7192DULL, "730949821.ogg"},
	{0x44D5F9D9E7556481ULL, "llac_w_sundermount_196608_0.dds"},
	{0x44D6098E31EC01BCULL, "vxlt_undercityserialkiller_1st_0_0.vlm"},
	{0x44D64B9BF176C480ULL, "hlac_dr_dungeon_3215_8y.dds"},
	{0x44D6B22D793E9CE8ULL, "vxca_caverns1_mines_icm_0_1.vlm"},
	{0x44D72738F61E5B3CULL, "dae000_madam_lusine6152231.cl"},
	{0x44D7B1303DBD2BF3ULL, "167177851.ogg"},
	{0x44D7D182C23D6DB2ULL, "864973755.ogg"},
	{0x44D80BE50EA01315ULL, "vxht_0_hightown_a74_0_0.vlm"},
	{0x44D80DD79EAAB7CAULL, "lt_0_lowtown_docks_a3_29_0_lc.mmh"},
	{0x44D93094E699336EULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkright.ani"},
	{0x44D992A185BFA103ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_loop.gad"},
	{0x44D9BCF94CC05FAFULL, "w_encounter2_51_0.mao"},
	{0x44D9C8F94CC0744CULL, "w_encounter2_51_0.mmh"},
	{0x44D9CEF94CC07E3EULL, "w_encounter2_51_0.msh"},
	{0x44DA0FD9252B9C4BULL, "follower_banter6154107.cl"},
	{0x44DADE486DFD22FAULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha02a.msh"},
	{0x44DAF0486DFD4140ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha02a.mmh"},
	{0x44DB0823DC0F3D8DULL, "vxht_brothel_n_4a3_0_0.vlm"},
	{0x44DB2D216BD752E6ULL, "65453326.ogg"},
	{0x44DB5F9B6F476571ULL, "966383448.ogg"},
	{0x44DD240B45D77DDFULL, "vxht_0_hightown_a3_cb2_0_0.vlm"},
	{0x44DD272FD7766947ULL, "705936919.ogg"},
	{0x44DD2A8D94D24B52ULL, "cutscene_wrapper6000073.cl"},
	{0x44DDC278D2EB68A9ULL, "llac_w_woundedcoast_-65532_0.dds"},
	{0x44DDCE37C4E080A0ULL, "vxlt_0_lowtown_docks_n_bc4_0_0.vlm"},
	{0x44DE08B508AE98F3ULL, "vxht_keep_n_57q_0_0.vlm"},
	{0x44DE1273B54DFE1FULL, "dae210_gate_guard6084310.cl"},
	{0x44DE1B148D1FB4F1ULL, "lt_0_lowtown_docks_a3_n_24_0.rml"},
	{0x44DEB10639C3190DULL, "plt_cod_bks_kirkwall_hist_1.plo"},
	{0x44DED012FB0453F4ULL, "art\\characters\\playercharacter\\children\\km_cth_boya.mao"},
	{0x44DEEAB676CEF297ULL, "6169294_m.fxe"},
	{0x44DEF44DF7E5E724ULL, "hlac_w_encounter2_camp_196609_0y.dds"},
	{0x44DF1D5461F8DA40ULL, "dae000_bodhan6146831.cl"},
	{0x44DF59133362AB75ULL, "261796363.ogg"},
	{0x44DF5CD7464968F7ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit_3.mao"},
	{0x44DFA237AE3443DEULL, "m_dae111ar_gamlens_house_nt.wbd"},
	{0x44DFB898FACD6D00ULL, "vxht_mansion_n_4g3_0_0.vlm"},
	{0x44DFEEABD575C179ULL, "vxlt_0_lowtown_n_e11_0_0.vlm"},
	{0x44DFEEAEB78BCDB5ULL, "6032130_m.fxe"},
	{0x44E04E2F4B4CFB84ULL, "w_bonepitdestroyed_a3_28_0_lc.mmh"},
	{0x44E0B9EF7B433AC6ULL, "949975878.ogg"},
	{0x44E0E500A30E5536ULL, "hlac_lt_0_lowtown_docks_n_-65537_2x.dds"},
	{0x44E12BE977459167ULL, "ser221_gascard6111755.cl"},
	{0x44E188D853E3FB9EULL, "49225107.ogg"},
	{0x44E188E51CB4BF06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_up.ani"},
	{0x44E196BBE86E6DABULL, "hlac_ca_caverns2_5802_51z.dds"},
	{0x44E1D77C738ED1F9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxined.phy"},
	{0x44E1DE9A8E6EBF63ULL, "hlac_ht_0_hightown_n_-65537_4x.dds"},
	{0x44E20B395A1B5DB6ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masb.msh"},
	{0x44E215395A1B6EB4ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masb.mmh"},
	{0x44E219395A1B75C7ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masb.mao"},
	{0x44E30288910BC322ULL, "static1_dy.psh.fxo"},
	{0x44E358A667E09FBCULL, "319245893.ogg"},
	{0x44E363C0E6745132ULL, "hlac_lt_0_lowtown_a3_n_c_65537_1z.dds"},
	{0x44E380668C056555ULL, "6082234_m.fxe"},
	{0x44E3B922D2442629ULL, "ser100_party_comments6108238.cl"},
	{0x44E3FCBFB2499908ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medd.mmh"},
	{0x44E4013D2AB022EFULL, "6163876_m.fxe"},
	{0x44E417722C6607DFULL, "vxlt_gallowstemplar_f_53j_0_0.vlm"},
	{0x44E41ABFB249CC32ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medd.msh"},
	{0x44E45AADEF6C41F0ULL, "lla_ht_0_hightown_-131071_0.dds"},
	{0x44E485B9C6E757ABULL, "6071920_m.fxe"},
	{0x44E4CA0EFDCCF757ULL, "vxht_0_hightown_n_b5t_0_0.vlm"},
	{0x44E4FCEFBCC0E75CULL, "vxlt_gallowsprison_n_6cj_0_0.vlm"},
	{0x44E5672F48401BB8ULL, "vxlt_0_lowtown_a2_n_c_hm7_0_0.vlm"},
	{0x44E58EADF46DA6D4ULL, "6142306_m.fxe"},
	{0x44E5EEDA08D0B0E1ULL, "w_bonepitdestroyed_a3_200_0.phy"},
	{0x44E62894E6A3AFBFULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkright.evt"},
	{0x44E7007752831670ULL, "167342524.ogg"},
	{0x44E70AC69DE842FFULL, "vxlt_gallowscourtyard_6xm_0_0.vlm"},
	{0x44E7189B0982D799ULL, "6175792_m.fxe"},
	{0x44E779D1EB00CA41ULL, "vxlt_darktown_4tw_0_0.vlm"},
	{0x44E81D32EDB4E95CULL, "6181912_m.fxe"},
	{0x44E81DB4B81D7A81ULL, "385013757.ogg"},
	{0x44E8A21064DCFBADULL, "floorspec.dds"},
	{0x44E94D90F549E0E6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rail8m.phy"},
	{0x44EACA74172EAE0AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic.mao"},
	{0x44EACE74172EB559ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic.mmh"},
	{0x44EAD874172EC647ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic.msh"},
	{0x44EB4869E6CCEC31ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_brokewood.mmh"},
	{0x44EB5469E6CD00D2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_brokewood.mao"},
	{0x44EB6269E6CD18EFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_brokewood.msh"},
	{0x44EB78D5432489AFULL, "6205363_m.fxe"},
	{0x44EB7D09BAF2A1C4ULL, "hlac_w_sundermount_-262145_0x.dds"},
	{0x44EBC08908A7475DULL, "vxht_0_hightown_n_b7p_0_0.vlm"},
	{0x44EBC8E1D515D230ULL, "vxht_mansion_n_4ft_0_0.vlm"},
	{0x44EC2C3EBDFFC98CULL, "character_character_p18.psh.fxo"},
	{0x44EC8C07C02CE2E4ULL, "vxw_woundedcoast_108_0_0.vlm"},
	{0x44ECBBCD53D3CD33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_rotateback.evt"},
	{0x44ECEE3758B5DCAFULL, "vxht_keep_6gy_0_0.vlm"},
	{0x44ED0994E6A999A8ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkright.gad"},
	{0x44ED3E22AF111109ULL, "vxca_caverns1_mines_j28_0_1.vlm"},
	{0x44EDA2FAEFD075BBULL, "vxdr_0_deeproads_cum_0_0.vlm"},
	{0x44EE86FA0A66387DULL, "297680562.ogg"},
	{0x44EEA88A531CC16AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_basic.mao"},
	{0x44EEAC8A531CC7B9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_basic.mmh"},
	{0x44EEB68A531CD8A7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_basic.msh"},
	{0x44EF7885AD8EAA4CULL, "vxw_encounter1_bs_0_0.vlm"},
	{0x44EFC44BC43FE82BULL, "6100017_m.fxe"},
	{0x44EFCE76D7C010F5ULL, "vxht_keep_5if_0_0.vlm"},
	{0x44F0005798A174CCULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetold_sa.dds"},
	{0x44F01C9CC6D288AFULL, "t1_mue_wh3.tnt"},
	{0x44F02C494A43704DULL, "llac_w_encounter2_196607_0.dds"},
	{0x44F066F42A84F65FULL, "vxlt_gallowstemplar_n_5ra_0_0.vlm"},
	{0x44F1D9E59B61249BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinl.msh"},
	{0x44F1DE61530211C8ULL, "art\\characters\\playercharacter\\humanmale\\glv_maso.mmh"},
	{0x44F1EA615302266BULL, "art\\characters\\playercharacter\\humanmale\\glv_maso.mao"},
	{0x44F1EBE59B6143E6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinl.mao"},
	{0x44F1F7E59B615785ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinl.mmh"},
	{0x44F1FC61530244F2ULL, "art\\characters\\playercharacter\\humanmale\\glv_maso.msh"},
	{0x44F20658AEF53012ULL, "494256549.ogg"},
	{0x44F227574D68AB1AULL, "hlac_ht_0_hightown_65536_4z.dds"},
	{0x44F29D87008099CDULL, "117557757.ogg"},
	{0x44F2B4534ABDB5F8ULL, "421407561.ogg"},
	{0x44F2C2E84E82FD4DULL, "388621463.ogg"},
	{0x44F302AD8F8EA35EULL, "gen00fl_anders6119900.cl"},
	{0x44F33127FF27062CULL, "47049262.ogg"},
	{0x44F347F3CCA057BAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave4statue.mmh"},
	{0x44F34BF3CCA05EF1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave4statue.mao"},
	{0x44F359F3CCA07600ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave4statue.msh"},
	{0x44F366C338310CE5ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs08a_0n.dds"},
	{0x44F370610A9247D7ULL, "1055409605.ogg"},
	{0x44F3B4113A395834ULL, "6113286_f.fxe"},
	{0x44F3B4CD53D9E074ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_rotateback.gad"},
	{0x44F3C5C73BF8AD50ULL, "vxht_0_hightown_n_c5z_0_0.vlm"},
	{0x44F497D8FA202788ULL, "lt_0_lowtown_docks_a3_n_226_0.mmh"},
	{0x44F4A3D8FA203C2BULL, "lt_0_lowtown_docks_a3_n_226_0.mao"},
	{0x44F4B5D8FA205AB2ULL, "lt_0_lowtown_docks_a3_n_226_0.msh"},
	{0x44F571E51CC54100ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_up.gad"},
	{0x44F59E9E5F34C764ULL, "6094940_m.fxe"},
	{0x44F5A2549F6D0F4CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerpot01_0_d.dds"},
	{0x44F5D07C31AA136EULL, "hlac_lt_0_lowtown_-3_6y.dds"},
	{0x44F5D3C00FC8C7AFULL, "vxlt_0_lowtown_docks_a3_cvt_0_1.vlm"},
	{0x44F5E75E03C3CF17ULL, "lowtown_male_urchin_sta6164177.cl"},
	{0x44F60189DF66D06FULL, "151019198.ogg"},
	{0x44F6879E6C5EC971ULL, "vxht_bartrand_n_32a_0_0.vlm"},
	{0x44F6BFB1A3B03CA8ULL, "677116000.ogg"},
	{0x44F6D2544BD8D4D2ULL, "hlac_lt_0_lowtown_a3_n_c_-1_6y.dds"},
	{0x44F6EAE92E1B62EBULL, "6169984_m.fxe"},
	{0x44F75CE068AC2619ULL, "602447725.ogg"},
	{0x44F78147476B3146ULL, "pro000_aveline_intro6109377.cl"},
	{0x44F7A33D0E488F23ULL, "hlac_lt_0_lowtown_docks_-4_5y.dds"},
	{0x44F7ED753097E1CEULL, "gen_im_trs_krk_grd_junk2_03.uti"},
	{0x44F86629B92EA3E1ULL, "86161371.ogg"},
	{0x44F8FEA3FCA658FEULL, "gui_passthrough_p0.vsh.fxo"},
	{0x44F931618ECDC4F1ULL, "vxlt_gallowstemplar_n_51j_0_0.vlm"},
	{0x44F9888356A4B7DDULL, "hlac_lt_darktown_-1_22y.dds"},
	{0x44F9D929C0FFFE36ULL, "vxca_caverns1_mines_cjo_0_0.vlm"},
	{0x44FA3F2857E20EB0ULL, "354807135.ogg"},
	{0x44FABABBD375C49FULL, "zz_one_debug6113364.cl"},
	{0x44FAC3CD53E01852ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_rotateback.ani"},
	{0x44FAD9CE04EEC1AAULL, "456639838.ogg"},
	{0x44FB4A1A4DAF2630ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb08a.phy"},
	{0x44FB4A2E6AD86E44ULL, "6020074_f.fxe"},
	{0x44FC2F1BF415AE2FULL, "6013326_m.fxe"},
	{0x44FC5FD611F69578ULL, "6035141_m.fxe"},
	{0x44FD19662D93469BULL, "vxlt_0_lowtown_docks_bs8_0_0.vlm"},
	{0x44FD297AC459F71BULL, "w_bonepitmining_a2_236_0.mmh"},
	{0x44FD357AC45A0BB8ULL, "w_bonepitmining_a2_236_0.mao"},
	{0x44FD477AC45A2A05ULL, "w_bonepitmining_a2_236_0.msh"},
	{0x44FDEA629F84C2CCULL, "vxw_encounter1_gp_0_0.vlm"},
	{0x44FE2F6F519FA1C0ULL, "6023872_m.fxe"},
	{0x44FE8B7466C68B46ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\blendtree\\floortrap.anb"},
	{0x44FE8B7DDF0B39C8ULL, "154190637.ogg"},
	{0x44FF2E902FAAB39FULL, "hlac_dr_0_deeproads_17039_0y.dds"},
	{0x44FF398CE92FDAF4ULL, "gen00fl_isabela6079949.cl"},
	{0x44FF686BE6F2E02DULL, "hlac_lt_0_lowtown_docks_n_-65536_0z.dds"},
	{0x44FF98EB5BDF50D8ULL, "407836666.ogg"},
	{0x44FFA0A7ABEE2C02ULL, "192978226.ogg"},
	{0x44FFB57C7602A8FAULL, "936465988.ogg"},
	{0x44FFD76A2243FA02ULL, "321442927.ogg"},
	{0x4500409275243DD5ULL, "353226178.ogg"},
	{0x4500A6575E297421ULL, "mag360_samson6042081.cl"},
	{0x450194F9DE4001E6ULL, "hlac_lt_0_lowtown_docks_-3_6x.dds"},
	{0x45020873450B7021ULL, "546485100.ogg"},
	{0x4502B634510F4E52ULL, "9804586.ogg"},
	{0x4503095A9B81ECE8ULL, "199451203.ogg"},
	{0x450383E6D1241CECULL, "6049471_f.fxe"},
	{0x45039A337B975D43ULL, "gen_im_trs_bst_hunt_dag.uti"},
	{0x4503A5CDA4CB133FULL, "506741910.ogg"},
	{0x4503B7994F644A78ULL, "842933227.ogg"},
	{0x45041128BA88510CULL, "103507149.ogg"},
	{0x4504344D732059DFULL, "253510964.ogg"},
	{0x4504C3A7514671FBULL, "6116818_m.fxe"},
	{0x450541366CF0FD10ULL, "hlac_lt_undercitydungeon_-1_8y.dds"},
	{0x45054A009E8CD584ULL, "llac_lt_0_lowtown_a2_n_-2_3.dds"},
	{0x4505B9E8D678675AULL, "gen00fl_isabela6086453.cl"},
	{0x450626F53F100B0DULL, "art\\characters\\creatures\\shriek\\textures\\cn_hed_ems_0d.dds"},
	{0x450651B318313F2CULL, "w_sundermount.rml"},
	{0x45066456BE455107ULL, "vxlt_0_lowtown_a2_n_c_ekj_0_1.vlm"},
	{0x45067C1A1FE47AD6ULL, "hlac_lt_0_lowtown_a2_n_c_65536_3x.dds"},
	{0x45069908DB3ECEBEULL, "82098164.ogg"},
	{0x4506B7CB2CB17337ULL, "476744180.ogg"},
	{0x45078D38C6FD078FULL, "lgt120ip_at_gang_hideout_ht.utp"},
	{0x45079839FE4AED63ULL, "6133525_m.fxe"},
	{0x4507A0B31A86A7E0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_basic.phy"},
	{0x4508B52A43B57195ULL, "6209325_m.fxe"},
	{0x45091915C81F4F05ULL, "942806624.ogg"},
	{0x4509CC1BA4E43D88ULL, "docks_fem_sailor_mob6176304.cl"},
	{0x450A1B0208E68874ULL, "vxht_brothel_n_5ku_0_0.vlm"},
	{0x450A55D2DDCA386EULL, "73133317.ogg"},
	{0x450A5D0897658A2AULL, "6045984_m.fxe"},
	{0x450ADBEA06BB9997ULL, "lgt260_eau_tmp_ending6173894.cl"},
	{0x450B60495FC48758ULL, "lt_0_lowtown_docks_a3_1_0_lc.mmh"},
	{0x450BE86D4A07720DULL, "vxdr_primevaltomb_6ri_0_0.vlm"},
	{0x450CBCD8F40EE803ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_l_fist_clench.ani"},
	{0x450D3ECA28600108ULL, "586525258.ogg"},
	{0x450DD8B84B4EC272ULL, "gen00fl_sebastian6191251.cl"},
	{0x450DFDE256BE9816ULL, "6151111_m.fxe"},
	{0x450E212F61960EB5ULL, "vxlt_undercityserialkiller_1ik_0_0.vlm"},
	{0x450E3A6F0D0697AAULL, "vxdr_primevaltomb_7u2_0_0.vlm"},
	{0x450F1581D12C495CULL, "6076147_f.fxe"},
	{0x450F6D6CB8354924ULL, "qun120_arishok6051902.cl"},
	{0x450F83F124C8A3FDULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_1z.dds"},
	{0x450F89280FB712D0ULL, "hlac_ca_cavern1_haunted_24424_0x.dds"},
	{0x450FA5CC4CF5D90CULL, "w_bonepitmining_a2_109_0.msh"},
	{0x450FABCC4CF5E3FEULL, "w_bonepitmining_a2_109_0.mmh"},
	{0x450FB7CC4CF5F795ULL, "w_bonepitmining_a2_109_0.mao"},
	{0x4510528692A330DCULL, "vxlt_gallowsprison_n_6ri_0_0.vlm"},
	{0x45108DA3931E68B6ULL, "vxca_caverns2_97p_0_0.vlm"},
	{0x45113C31ED2EA73CULL, "6164746_m.fxe"},
	{0x451149CBB395D04CULL, "6055136_m.fxe"},
	{0x4511B1AC145881B2ULL, "6155973_m.fxe"},
	{0x4511C0E9C4267EC9ULL, "vxw_encounter2_camp_67_0_0.vlm"},
	{0x4511D3AF7A061240ULL, "llac_w_woundedcoast_327681_0.dds"},
	{0x4511D5D1459FA427ULL, "vxlt_0_lowtown_docks_a3_n_8uh_0_0.vlm"},
	{0x4511F39DF88E80F5ULL, "vxht_0_hightown_n_cbj_0_0.vlm"},
	{0x45127022A5832356ULL, "hlac_ht_chantry_2534_0y.dds"},
	{0x4512D34D3A28335DULL, "6028933_m.fxe"},
	{0x4512DC9588DEEC7FULL, "6107225_m.fxe"},
	{0x4513FA51060B51D5ULL, "vxlt_0_lowtown_a3_n_c_efg_0_0.vlm"},
	{0x4513FF0588812B6EULL, "isa311_castillon6092098.cl"},
	{0x451460AA526EE383ULL, "gen00fl_varric6033546.cl"},
	{0x4514FD0526042E37ULL, "vxlt_0_lowtown_hai_0_0.vlm"},
	{0x4515454812D30B73ULL, "hlac_ca_caverns2_5802_13x.dds"},
	{0x4515494C94339E98ULL, "hlac_w_sundermount_262149_0y.dds"},
	{0x4515F850AB30EB64ULL, "vxca_caverns1_fmh_0_0.vlm"},
	{0x45160C18EBB83CB4ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtd.phy"},
	{0x45161A2AB8E6441CULL, "6214692_m.fxe"},
	{0x45168B18DA5390A9ULL, "hlac_lt_0_lowtown_a2_65535_8x.dds"},
	{0x4516D3CE7D115042ULL, "dae120st_merch_weapon.utm"},
	{0x45171989996FED9CULL, "dae000_fenarel6128115.cl"},
	{0x4517EF1AD33FC533ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masa.phy"},
	{0x4518152D0D9F5EACULL, "cd_dae000_mother.bnk"},
	{0x4518297940D41F6DULL, "character_character_p0.vsh.fxo.dx11.assembly"},
	{0x45185E7B02382B4AULL, "hlac_lt_0_lowtown_docks_-65539_6y.dds"},
	{0x45185F38DCC9907FULL, "vxlt_0_lowtown_n_h1v_0_0.vlm"},
	{0x451940F55B6E73FEULL, "vxlt_hangedman_n_21k_0_0.vlm"},
	{0x45197E51B58EE305ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pile_rags_0_d.dds"},
	{0x4519D4D8F4199BFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_l_fist_clench.evt"},
	{0x451A9836E95E9124ULL, "705497367.ogg"},
	{0x451AA0479C8105DDULL, "ser221_quentin6123311.cl"},
	{0x451B1844B1F3EBABULL, "vxlt_0_lowtown_a2_n_c_hu8_0_0.vlm"},
	{0x451BB3D86281E557ULL, "vxdr_primevalentrance_32v_0_0.vlm"},
	{0x451BBC8B00673FDFULL, "vxlt_gallowstemplar_n_5cb_0_0.vlm"},
	{0x451C930C538C2735ULL, "vxlt_undercityserialkiller_1nj_0_0.vlm"},
	{0x451CAEA5C70649A2ULL, "6162751_m.fxe"},
	{0x451CDA1E741DEF58ULL, "6132734_m.fxe"},
	{0x451CEB7C5BFB56B6ULL, "m_qcr210ar_lowtown_nt.wbd"},
	{0x451D27F669EE2902ULL, "6196705_m.fxe"},
	{0x451D4536D6E002EBULL, "mag101_aombanter6052766.cl"},
	{0x451E91DFBF1BF8D9ULL, "wm_ico_darktown_roll.dds"},
	{0x451EDA130A8071ADULL, "6081683_f.fxe"},
	{0x451FE918EBC055B1ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtd.msh"},
	{0x451FF535346997E2ULL, "6088370_m.fxe"},
	{0x451FF718EBC06DCCULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtd.mao"},
	{0x45200318EBC0826FULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtd.mmh"},
	{0x452032C6B61C9749ULL, "vxw_encounter2_camp_16_0_0.vlm"},
	{0x4520627B1E0BED79ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_011a_0s.dds"},
	{0x4520CEAEE463846FULL, "vxw_encounter2_camp_n9_0_0.vlm"},
	{0x4520D3D8F41FB8EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_l_fist_clench.gad"},
	{0x4520D5E0BB8813D0ULL, "vxht_brothel_n_59g_0_0.vlm"},
	{0x4520F0AE46A9AA60ULL, "17467290.ogg"},
	{0x45212F946F05C501ULL, "vxht_0_hightown_a3_b9h_0_0.vlm"},
	{0x45217CD600E0DBD1ULL, "275128518.ogg"},
	{0x452229A418A87B98ULL, "61762439.ogg"},
	{0x4522468F2C02D8D0ULL, "fen310_fenris6175084.cl"},
	{0x452295B2D7B9DD20ULL, "6078182_m.fxe"},
	{0x4522EDD66435159FULL, "hlac_dr_slavers_3215_3y.dds"},
	{0x45231FEF7AAA50EEULL, "gen00fl_dog6167243.cl"},
	{0x45239A7F01DD67C0ULL, "6105440_m.fxe"},
	{0x4523C924A3FCE01BULL, "vxca_caverns1_hw3_0_0.vlm"},
	{0x45246CDF4AC3DBE7ULL, "qun221_viscount6167451.cl"},
	{0x45247BF3E2371E63ULL, "6139325_m.fxe"},
	{0x45251A9FF9B9638CULL, "hlac_lt_0_lowtown_docks_n_-131075_1x.dds"},
	{0x4525525E0BFE25DAULL, "vxlt_gallowstemplar_n_3rn_0_0.vlm"},
	{0x4525E006729A767EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_drunk.evt"},
	{0x4525FCCD2B1A8E68ULL, "gen00fl_aveline6122730.cl"},
	{0x452692D4B4982B30ULL, "gen00fl_varric6107301.cl"},
	{0x4527635BBB10685CULL, "6179384_m.fxe"},
	{0x4527C796E6B30DE0ULL, "hlac_lt_undercitydungeon_-1_18x.dds"},
	{0x4527C9EC1E83CD68ULL, "3208733.ogg"},
	{0x452809D6FAA5261BULL, "141550834.ogg"},
	{0x4528860337365651ULL, "6127680_f.fxe"},
	{0x4528E011AE2F71D7ULL, "character_character_p6.psh.fxo"},
	{0x452918399512F986ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_up.ani"},
	{0x4529B04238846682ULL, "vxlt_0_lowtown_e1q_0_0.vlm"},
	{0x4529E768AB7A2652ULL, "hlac_lt_0_lowtown_a3_n_c_2_1z.dds"},
	{0x452A6A0DD41BB533ULL, "gen00fl_sebastian6175012.cl"},
	{0x452A8C277E7ABDB4ULL, "vxht_0_hightown_awq_0_0.vlm"},
	{0x452AE6EA3921F093ULL, "vxht_keep_3xy_0_0.vlm"},
	{0x452AED9831A19311ULL, "llac_lt_darktown_-1_29.dds"},
	{0x452B2082B8FC6222ULL, "6167451_m.fxe"},
	{0x452B2998AEDD2C29ULL, "vxlt_0_lowtown_a2_n_ex2_0_0.vlm"},
	{0x452B7FBB6B7E0922ULL, "hlac_w_sundermount_-327680_0x.dds"},
	{0x452B8E4E39AACC85ULL, "370542712.ogg"},
	{0x452C2E7DFBE13171ULL, "6034055_m.fxe"},
	{0x452C697CC6E22AA0ULL, "vxht_bartrand_n_34j_0_0.vlm"},
	{0x452CD59A8A47785FULL, "172061967.ogg"},
	{0x452CF70672A0BBA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_drunk.gad"},
	{0x452D6C52B6524656ULL, "art\\toolset\\_textures\\iconrgby_d.dds"},
	{0x452D6CF800BE0FFFULL, "hightown_market_guard_sta6169641.cl"},
	{0x452DA080020E9638ULL, "vxlt_0_lowtown_a2_n_c_ht0_0_0.vlm"},
	{0x452DBDB31AA7A875ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_basic.msh"},
	{0x452DBFB31AA7ABEBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_basic.mmh"},
	{0x452DCBB31AA7BF88ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_basic.mao"},
	{0x452DEB5B6D16C877ULL, "vxdr_0_deeproads_crx_0_0.vlm"},
	{0x452E05F60A3F941DULL, "6115035_m.fxe"},
	{0x452E637436CCA5F2ULL, "6036516_f.fxe"},
	{0x452E95CC08C15D50ULL, "vxht_0_hightown_a2_c7o_0_0.vlm"},
	{0x452F408BD6599CEFULL, "vxw_encounter2_camp_i8_0_0.vlm"},
	{0x452F84C03783871BULL, "w_bonepit_37_0_ll.mmh"},
	{0x452FE314C23A5D68ULL, "436552704.ogg"},
	{0x452FEDB0CAC65D26ULL, "vxdr_0_deeproads_fbl_0_0.vlm"},
	{0x4530048B542FE05CULL, "lgt210_magu_selby6142040.cl"},
	{0x45302539093CDBD6ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgta.msh"},
	{0x45302F39093CECD4ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgta.mmh"},
	{0x45303339093CF3E7ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgta.mao"},
	{0x453077AB48348495ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tool.phy"},
	{0x453087988F5840BAULL, "6103134_m.fxe"},
	{0x453109B82DE44CBAULL, "dae121st_viscount_1.stg"},
	{0x453140F399C6664CULL, "zz_dae_debug6029167.cl"},
	{0x453231CEF3769C0AULL, "hlac_lt_0_lowtown_a3_n_c_-196609_1z.dds"},
	{0x4532C2183CFDB1DEULL, "6013374_m.fxe"},
	{0x4532E9354757905CULL, "zz_fol_debug6096553.cl"},
	{0x453356EC27BE3E6CULL, "hlac_w_encounter2_camp_327679_0x.dds"},
	{0x45335F48FCD41423ULL, "dae000_orana6149056.cl"},
	{0x45338BD1887BC94BULL, "dev_res_up1.mao"},
	{0x4533C80672A68AF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_drunk.ani"},
	{0x45340C33921E8D87ULL, "vxca_caverns2_a4h_0_0.vlm"},
	{0x4534753984940A9DULL, "6070239_m.fxe"},
	{0x4534B735BE5FEC7EULL, "vxht_bartrand_n_4fu_0_0.vlm"},
	{0x4534D7931356DA2EULL, "hlac_ca_caverns2_5802_36x.dds"},
	{0x453534CC04D49517ULL, "zz_ave_debug6207692.cl"},
	{0x453563216B4EBAA4ULL, "488865368.ogg"},
	{0x4535A8B8DDD73BCBULL, "841309787.ogg"},
	{0x4536E09AB3DE62EEULL, "vxlt_0_lowtown_docks_n_c4f_0_0.vlm"},
	{0x45370039951F0EF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_up.evt"},
	{0x45376FA692636901ULL, "979951805.ogg"},
	{0x4537D4B84353784EULL, "vxlt_0_lowtown_n_h8p_0_0.vlm"},
	{0x4538020BFC6170E5ULL, "6072560_f.fxe"},
	{0x4538D0E59B9D1832ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinl.phy"},
	{0x453AA3C8B186E3A9ULL, "gen00fl_merrill6130133.cl"},
	{0x453AAF710D460368ULL, "vxca_caverns1_mines_edu_0_0.vlm"},
	{0x453ADB59B8D84320ULL, "vxht_bartrand_n_33k_0_0.vlm"},
	{0x453AE9D9FAA557E7ULL, "6219195_m.fxe"},
	{0x453AEDAD0FEBC8CFULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirecamp_0.pwk"},
	{0x453B151078A893A0ULL, "follower_banter6156113.cl"},
	{0x453B16E91B2645DAULL, "hlac_w_blightlands_0_9z.dds"},
	{0x453B1B65D4DF5078ULL, "884286475.ogg"},
	{0x453BF1669F773F47ULL, "hlac_lt_0_lowtown_n_-131073_1z.dds"},
	{0x453C28F9516B29B4ULL, "6039967_m.fxe"},
	{0x453C91754A559834ULL, "hlac_lt_0_lowtown_-131072_4x.dds"},
	{0x453D013995237B80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_up.gad"},
	{0x453D8E35B52AE881ULL, "vxht_mansion_n_4v9_0_0.vlm"},
	{0x453DAB0E5FE5E4BEULL, "439453583.ogg"},
	{0x453E14395A69CBAFULL, "art\\characters\\playercharacter\\humanfemale\\glv_masb.phy"},
	{0x453E566DCBD6A4FEULL, "437349648.ogg"},
	{0x453E906F60D930BBULL, "dae221ip_home_amb_orana_lut.utp"},
	{0x453F3F226DAAC2A9ULL, "hlac_lt_0_lowtown_a2_-131071_1y.dds"},
	{0x453F5F3BCE0F9815ULL, "21425607.ogg"},
	{0x453F9EA197B16B0FULL, "1062425968.ogg"},
	{0x4540598CDF50B3E2ULL, "vxdr_0_deeproads_cj9_0_0.vlm"},
	{0x454083C9E1FFB1CAULL, "vxht_keep_60a_0_0.vlm"},
	{0x4540A2262BD502AAULL, "6048811_m.fxe"},
	{0x4541899A3613201DULL, "hlac_lt_0_lowtown_a2_-65537_0y.dds"},
	{0x4541CCC646090B28ULL, "vxht_keep_5k9_0_0.vlm"},
	{0x4541F00EBF4C986CULL, "w_bonepitmining_a2_95_0.msh"},
	{0x4541F60EBF4CA25EULL, "w_bonepitmining_a2_95_0.mmh"},
	{0x4542020EBF4CB6F5ULL, "w_bonepitmining_a2_95_0.mao"},
	{0x45423B4C3D6D2C10ULL, "one100dg_kelder6056889.cl"},
	{0x4542436D8A2E871AULL, "zz_ser_debug6085772.cl"},
	{0x454251BE924D4FE1ULL, "6139622_m.fxe"},
	{0x45429BB17FAA69F4ULL, "349126805.ogg"},
	{0x4542B9F27E47C13FULL, "art\\cinematics\\deathblows\\ogre\\cs_db_carver_ogremage.gad"},
	{0x4542E716768A231DULL, "6075139_m.fxe"},
	{0x45432912B05604FEULL, "vxht_bartrand_n_4kv_0_0.vlm"},
	{0x45438FC76CA1B17DULL, "258860864.ogg"},
	{0x4543C6DB25627221ULL, "hlac_lt_0_lowtown_a3_n_c_131070_4x.dds"},
	{0x454447159EFC995CULL, "6177184_m.fxe"},
	{0x45445C7CD04DE5E8ULL, "vxht_0_hightown_a2_bsr_0_0.vlm"},
	{0x4544C7E3F6FB6529ULL, "mag321_comtesse_de_launcet6064883.cl"},
	{0x45455F8CD4E45A99ULL, "gen00fl_fenris6113923.cl"},
	{0x45457FCE2893DC17ULL, "vxlt_gallowstemplar_f_5bq_0_0.vlm"},
	{0x4545EF719AA090B9ULL, "519120810.ogg"},
	{0x45460BC2CDC8B734ULL, "vxca_cavern1_haunted_dqc_0_0.vlm"},
	{0x45461DD6BFABA356ULL, "6167287_f.fxe"},
	{0x454663BD315993D1ULL, "vxht_mansion_n_6b7_0_0.vlm"},
	{0x45466538B27664E4ULL, "6110643_f.fxe"},
	{0x45466757EE152E60ULL, "vxca_cave_2by_0_0.vlm"},
	{0x454673D5D9BC36EBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_reaction.evt"},
	{0x45483F8017B484AEULL, "556646701.ogg"},
	{0x45484F747FF5C21EULL, "gen_im_trs_drk_one_mac_r0.uti"},
	{0x454872789A7CDFA8ULL, "follower_banter6170764.cl"},
	{0x4548F3DA0925361DULL, "w_bonepitdestroyed_a3_200_0.mao"},
	{0x4548F7DA09253D56ULL, "w_bonepitdestroyed_a3_200_0.mmh"},
	{0x454901DA09254E54ULL, "w_bonepitdestroyed_a3_200_0.msh"},
	{0x4549117C5C20D224ULL, "m_qcr210ar_lowtown_nt.bnk"},
	{0x45493CD8E8E63EF8ULL, "dae100_viscount6118490.cl"},
	{0x4549D70B5D13EF92ULL, "481997875.ogg"},
	{0x454A1878A4226D8DULL, "876634894.ogg"},
	{0x454A58893E39DEDCULL, "6150272_f.fxe"},
	{0x454A5F021105A3A4ULL, "fen310_fenris6118243.cl"},
	{0x454A705B30EC09C6ULL, "874242779.ogg"},
	{0x454A7BDC6D9E4B3DULL, "690138750.ogg"},
	{0x454ABE35907803C3ULL, "6088241_m.fxe"},
	{0x454B3AAB3C860451ULL, "6171312_m.fxe"},
	{0x454B58575F5892D1ULL, "615822294.ogg"},
	{0x454C113DDD667736ULL, "zz_followers_banter6157806.cl"},
	{0x454C6F722ECD3280ULL, "479455086.ogg"},
	{0x454C8CD5D9C0CCACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_reaction.gad"},
	{0x454CCADD66A9353DULL, "139651336.ogg"},
	{0x454CD16AAEB28E26ULL, "vxdr_0_deeproads_fhn_0_0.vlm"},
	{0x454D1C923E24653CULL, "6176514_m.fxe"},
	{0x454D789B2EDA82BEULL, "dae000_bartrand6206151.cl"},
	{0x454DCF7417833446ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic.phy"},
	{0x454DF7249C69C9D2ULL, "644063872.ogg"},
	{0x454DFC79AD5137CEULL, "825318179.ogg"},
	{0x454E1F6397891AD7ULL, "996647484.ogg"},
	{0x454E56CEFA59DD33ULL, "hla_ht_0_hightown_a3_131073_1z.dds"},
	{0x454E5B83A5794437ULL, "dae000cr_grand_cleric6187954.cl"},
	{0x454E63BC6B4975CDULL, "zz_vrc_debug6200267.cl"},
	{0x454E70419357F845ULL, "hlac_ht_0_hightown_-131074_1z.dds"},
	{0x454E882AFF4C4852ULL, "vxlt_hangedman_1tx_0_0.vlm"},
	{0x454ECD9D6A9903C2ULL, "llac_lt_darktown_-1_40.dds"},
	{0x454ED41558319317ULL, "529967931.ogg"},
	{0x454FDBE5CE6AD72FULL, "mcr310_argument6074426.cl"},
	{0x454FEBC657352C92ULL, "dae000_quintus6154778.cl"},
	{0x45501277B23D173AULL, "428302887.ogg"},
	{0x455082CC4D2D3499ULL, "w_bonepitmining_a2_109_0.phy"},
	{0x4550AD8A536F93A6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_basic.phy"},
	{0x4550C32202F7DDB7ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinstower.phy"},
	{0x4550D6C0560AF256ULL, "6173814_m.fxe"},
	{0x4551038974C6B183ULL, "vxca_caverns1_mines_j2k_0_0.vlm"},
	{0x455123E4F8D58CB6ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_stub_0t.dds"},
	{0x455148725B2D5876ULL, "vxlt_shop_s7_0_0.vlm"},
	{0x4551579CC3EC4032ULL, "vxht_mansion2b_2j7_0_0.vlm"},
	{0x45515C5D36C91C10ULL, "vxdr_0_deeproads_a28_0_1.vlm"},
	{0x45517F353FB78E69ULL, "6093584_m.fxe"},
	{0x4551A73F1A3DFD86ULL, "hlac_lt_undercitydungeon_-1_19y.dds"},
	{0x455266988FE523A2ULL, "937452349.ogg"},
	{0x45528DBE08CF926FULL, "dae000_mother6103041.cl"},
	{0x4552FA96CB81BC96ULL, "vxht_bartrand_n_6r3_0_0.vlm"},
	{0x4553385419FB5C7BULL, "513811831.ogg"},
	{0x45538BD5D9C6E9AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_reaction.ani"},
	{0x4553A8CB93CF7BCDULL, "526997973.ogg"},
	{0x4554448BBAF9107AULL, "hlac_w_bonepit_131074_0x.dds"},
	{0x45548AF258F24559ULL, "weapon2_s1y.psh.fxo"},
	{0x455494A01C6FF082ULL, "pro000cr_soldier4.utc"},
	{0x4554F2299851585DULL, "1014436217.ogg"},
	{0x4554F3433E21E2B5ULL, "llac_w_woundedcoast_196611_0.dds"},
	{0x45556CC57F2E1D99ULL, "261556930.ogg"},
	{0x4555959285DDF598ULL, "dae000_fenarel6134323.cl"},
	{0x455664F59AA26EB1ULL, "927350665.ogg"},
	{0x45567C398282D70BULL, "mcr371_hostile_aveline6082198.cl"},
	{0x45569E9777B18D66ULL, "mag101tr_aom_way_out.utt"},
	{0x4556F3D30F542A58ULL, "6060386_f.fxe"},
	{0x4556F727E8B63986ULL, "415232890.ogg"},
	{0x4557146679708305ULL, "hlac_lt_gallowstemplar_2394_9x.dds"},
	{0x4557269E7052604DULL, "vxlt_0_lowtown_a2_n_c_f9h_0_0.vlm"},
	{0x455759F3D03FC4CCULL, "vxw_bonepitdestroyed_a3_23a_0_0.vlm"},
	{0x45576B49C3C7E275ULL, "6055436_m.fxe"},
	{0x455793D0AA09662EULL, "6056724_m.fxe"},
	{0x455796F27E59E185ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_carver_ogremage.ani"},
	{0x4557C948D3C63959ULL, "hlac_lt_0_lowtown_a2_n_1_4z.dds"},
	{0x4557EB7F8960A2B7ULL, "935746364.ogg"},
	{0x45587162D4B9C162ULL, "qcr000_varric6066280.cl"},
	{0x4558A54CA8C2017BULL, "6059405_m.fxe"},
	{0x45591A31EA7F8BCCULL, "6060411_m.fxe"},
	{0x45591F2D0DD706DEULL, "cd_dae000_mother.wbd"},
	{0x45593C69C1A102B4ULL, "follower_banter6163261.cl"},
	{0x4559809DC07AFD89ULL, "vxlt_0_lowtown_a2_n_ekf_0_1.vlm"},
	{0x4559D203ABFBC09AULL, "dae200_saemus6091687.cl"},
	{0x455A5039252235B8ULL, "hlac_ht_keep_n_3732_16x.dds"},
	{0x455A85FCCF841E94ULL, "759800478.ogg"},
	{0x455A9386BFDEA24CULL, "758523651.ogg"},
	{0x455B80B6C79048D7ULL, "gen00fl_anders6125922.cl"},
	{0x455BF390A5E0B531ULL, "dae300_cullen6140618.cl"},
	{0x455BFE90B06BB539ULL, "hlac_ht_0_hightown_a2_-65534_0z.dds"},
	{0x455C255BE4924534ULL, "gen00fl_isabela6101062.cl"},
	{0x455CB8F7E7FBDB1BULL, "6053528_m.fxe"},
	{0x455CDD9EC73F73D8ULL, "vxw_bonepitmining_a2_77_0_0.vlm"},
	{0x455D2993BACA8084ULL, "mag370_meredith6039275.cl"},
	{0x455D6B5258C52901ULL, "vxlt_0_lowtown_a2_n_h1t_0_0.vlm"},
	{0x455D7720089804A6ULL, "qun111_ketojan6039019.cl"},
	{0x455DBAB3AF549CA8ULL, "171053467.ogg"},
	{0x455E000185FAE091ULL, "vxlt_0_lowtown_a2_hc1_0_0.vlm"},
	{0x455E12E1941AF89EULL, "vxlt_0_lowtown_fer_0_0.vlm"},
	{0x455F38D5CBD5B082ULL, "6032476_m.fxe"},
	{0x455F5CFB57367429ULL, "vxlt_0_lowtown_docks_n_bew_0_0.vlm"},
	{0x455F77D3D65D018FULL, "vxht_0_hightown_9zi_0_0.vlm"},
	{0x455F91EBC3C5B617ULL, "dre200ar_dougal_encounter.lst"},
	{0x455FE934C3F6B1BBULL, "dre000_vampire6034361.cl"},
	{0x45601985673FE33FULL, "character_character_p220.psh.fxo.dx11.assembly"},
	{0x45602CD182A031F5ULL, "gen00fl_anders6125476.cl"},
	{0x45614036B43A16E8ULL, "vxht_0_hightown_a2_byp_0_0.vlm"},
	{0x4561644571D25E2BULL, "hlac_ht_0_hightown_n_131073_0x.dds"},
	{0x4561711A6C9BF45AULL, "545975931.ogg"},
	{0x456178B3BC7159C4ULL, "vxlt_gallowstemplar_n_2p3_0_0.vlm"},
	{0x4561AE9AADCBD869ULL, "6076530_m.fxe"},
	{0x4561B6E0C75D89A9ULL, "999382557.ogg"},
	{0x4561E2A2031CA5FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\runforward.evt"},
	{0x456276F46A16063AULL, "aip_spidergiant_base.gda"},
	{0x45629353E7C8A1C6ULL, "vxlt_warehouse_n_ou_0_0.vlm"},
	{0x4562F806DBAEF830ULL, "vxlt_gallowsprison_n_67z_0_0.vlm"},
	{0x456368AE637FE050ULL, "vxw_bonepit_1v5_0_0.vlm"},
	{0x4564A65C9F0ACA19ULL, "llac_lt_0_lowtown_a2_n_c_-2_0.dds"},
	{0x4564B2C03C225900ULL, "vxlt_0_lowtown_a2_n_hco_0_0.vlm"},
	{0x4564B432BDF09D8AULL, "hlac_w_bonepit_5_0y.dds"},
	{0x456566301A8BEC75ULL, "vxw_bonepitdestroyed_a3_1h6_0_0.vlm"},
	{0x4565755A3D5EBAC6ULL, "hightown_male_noble1_sta6170870.cl"},
	{0x4565AA5142E08326ULL, "456756153.ogg"},
	{0x45669B497E86F760ULL, "168043099.ogg"},
	{0x4566CA1D209DCE47ULL, "w_bonepitmining_a2_2_0.rml"},
	{0x4566DFD256F0B779ULL, "151555137.ogg"},
	{0x45675960FA0C537CULL, "496217476.ogg"},
	{0x45675C74332BACC7ULL, "hlac_lt_0_lowtown_65535_4x.dds"},
	{0x4567B425AC6A84D5ULL, "199946609.ogg"},
	{0x45685AF48386FDBAULL, "seb000im_leliana_chest.uti"},
	{0x45687168CE0D8FCAULL, "6125453_m.fxe"},
	{0x4568F9A20322EB29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\runforward.gad"},
	{0x45690722D5586163ULL, "7399645.ogg"},
	{0x4569569458A08339ULL, "534301182.ogg"},
	{0x45699D30F682E4D9ULL, "dre000_varric6030292.cl"},
	{0x456A3506AC4E87F6ULL, "vxht_0_hightown_a3_c9w_0_1.vlm"},
	{0x456A4C5B8AE5B914ULL, "vxdr_primevaltomb_2w3_0_0.vlm"},
	{0x456B8849795DC917ULL, "zz_vrc_debug6200157.cl"},
	{0x456BE54864137ABAULL, "hlac_lt_0_lowtown_a2_n_c_-2_6y.dds"},
	{0x456C71DE77F0F911ULL, "vxlt_0_lowtown_a2_hh0_0_0.vlm"},
	{0x456CDEC6CD4C2437ULL, "167835609.ogg"},
	{0x456CE0F08129B981ULL, "hlac_ht_mansion2b_818_63z.dds"},
	{0x456D8CD112DA6F51ULL, "mrl310_merrill6113535.cl"},
	{0x456DC71294B73410ULL, "vxlt_gallowstemplar_55p_0_0.vlm"},
	{0x456DCED8492C8CA9ULL, "vxlt_0_lowtown_docks_n_bbv_0_0.vlm"},
	{0x456E23AFB4C64254ULL, "vxht_mansion_n_53r_0_0.vlm"},
	{0x456E5574F2126334ULL, "llac_lt_warehouse_n_0_2.dds"},
	{0x456EBD5F80D3F350ULL, "vxw_woundedcoast_w8_0_0.vlm"},
	{0x456F8852227E2A16ULL, "player_core.ncs"},
	{0x456FB213A6302F68ULL, "vxht_0_hightown_a2_bbq_0_0.vlm"},
	{0x456FBF7F424BDEDFULL, "rdr110dg_guard_warehouse6174116.cl"},
	{0x456FCAA20328BA77ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\runforward.ani"},
	{0x456FDE8247FD72C8ULL, "mag371tr_close_orsino.utt"},
	{0x45705E59F1F87326ULL, "97639190.ogg"},
	{0x457092EBC3D42B2CULL, "dre200ar_dougal_encounter.are"},
	{0x4570CA13F7515A8CULL, "vxlt_0_lowtown_docks_a3_n_d3a_0_0.vlm"},
	{0x45710530D9BEBA46ULL, "vxlt_warehouse_n_rv_0_0.vlm"},
	{0x45715229352FFAE2ULL, "6000947_m.fxe"},
	{0x45715B4865487615ULL, "6202056_m.fxe"},
	{0x4571946F8C82ABBEULL, "308930399.ogg"},
	{0x4571EDFAEDB5FC7FULL, "lowtown_argument6169447.cl"},
	{0x457272FB0189A678ULL, "hlac_lt_0_lowtown_a2_n_-131072_5x.dds"},
	{0x4572B1547242392BULL, "vxlt_ruinedhouse_n_ih_0_0.vlm"},
	{0x4572B6E38826904FULL, "6118342_m.fxe"},
	{0x4572E77759567509ULL, "hlac_w_encounter2_262145_0y.dds"},
	{0x4572EAE9451CA12BULL, "vxht_0_hightown_a3_9zd_0_0.vlm"},
	{0x4573000A9A893E17ULL, "654620774.ogg"},
	{0x4573B809C9262861ULL, "genpt_amb_notif.ncs"},
	{0x4574195383260437ULL, "6074329_m.fxe"},
	{0x4574584AC9FE46F3ULL, "gen00fl_aveline6112299.cl"},
	{0x4574CFA1954BEBAFULL, "llac_lt_0_lowtown_131072_2.dds"},
	{0x4575732ED96201F5ULL, "vxlt_0_lowtown_docks_b98_0_0.vlm"},
	{0x4575B82ECFB41DB7ULL, "586455976.ogg"},
	{0x4575D7700D96E87DULL, "vxca_caverns1_mines_eib_0_1.vlm"},
	{0x4575EA0B67759409ULL, "follower_banter6141648.cl"},
	{0x4575F3D8413BB793ULL, "vxdr_primevaltomb_55i_0_0.vlm"},
	{0x4576390B5C3918BCULL, "622057161.ogg"},
	{0x45765BAF2884B5F2ULL, "hlac_dr_dungeon_3215_3z.dds"},
	{0x4576620CF83CADBCULL, "483559055.ogg"},
	{0x45766457A4672E89ULL, "vxlt_0_lowtown_a2_n_eqd_0_1.vlm"},
	{0x45767AC537A2A7D5ULL, "hlac_dr_primevaltomb_9621_8y.dds"},
	{0x4576BC0F030E5F53ULL, "vxlt_gallowsprison_n_5wb_0_0.vlm"},
	{0x4576D36CE62E5FFEULL, "follower_banter6163906.cl"},
	{0x4577371FCAF69185ULL, "llac_lt_0_lowtown_docks_n_-65537_3.dds"},
	{0x457776FF8577F8CEULL, "6173708_m.fxe"},
	{0x45778AAB4870A7B0ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tool.msh"},
	{0x457798AB4870BFCAULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tool.mmh"},
	{0x45779CAB4870C601ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tool.mao"},
	{0x45787AC8D658FE62ULL, "6034301_f.fxe"},
	{0x4578C4AC8070C3F1ULL, "hlac_lt_darktown_-1_17y.dds"},
	{0x4578D0DBA801C6F0ULL, "440546809.ogg"},
	{0x4578D5CE50437E34ULL, "follower_banter6172722.cl"},
	{0x4578E94B5A2938EDULL, "gen_im_trs_cot_lgt_hlm_r1.uti"},
	{0x4579800DF256BFD8ULL, "mag101_aombanter6053184.cl"},
	{0x4579C382D20E87A0ULL, "simplecreature.gda"},
	{0x4579CDBC0787227BULL, "lt_backalley_g_3_0_lc.mmh"},
	{0x457AE81AD39436A8ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masa.mmh"},
	{0x457AF41AD3944B4BULL, "art\\characters\\playercharacter\\humanfemale\\boo_masa.mao"},
	{0x457B032B0AA500C0ULL, "vxht_0_hightown_ap0_0_0.vlm"},
	{0x457B061AD39469D2ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masa.msh"},
	{0x457B20BEC673C40AULL, "one110_meeran6113253.cl"},
	{0x457C506C5C0CDE39ULL, "vxw_encounter1_ec_0_0.vlm"},
	{0x457C8CC4C5FC860FULL, "6155316_m.fxe"},
	{0x457D16C79226BBA6ULL, "6074401_m.fxe"},
	{0x457D9D0BC0FC84A8ULL, "vxw_bonepitdestroyed_a3_1vm_0_0.vlm"},
	{0x457DD94AC8B076E1ULL, "730175835.ogg"},
	{0x457E23F0982647E8ULL, "vxht_0_hightown_a2_bgn_0_0.vlm"},
	{0x457E2AA11D3363D0ULL, "6064727_m.fxe"},
	{0x457ECA1D20B26064ULL, "w_bonepitmining_a2_2_0.mmh"},
	{0x457ECE1D20B266B7ULL, "w_bonepitmining_a2_2_0.mao"},
	{0x457EE01D20B28506ULL, "w_bonepitmining_a2_2_0.msh"},
	{0x457F1566C94A0DD6ULL, "6129263_f.fxe"},
	{0x457F770DCBB4D2C6ULL, "vxlt_warehouse_n_yw_0_0.vlm"},
	{0x457FCD25573E8E95ULL, "6209156_m.fxe"},
	{0x457FD9E800B59DA4ULL, "vxht_bartrand_n_6rq_0_0.vlm"},
	{0x45803CC4C462FDD2ULL, "vxlt_gallowsprison_n_5wg_0_0.vlm"},
	{0x4580720B13D7BD02ULL, "llac_lt_0_lowtown_docks_a3_n_-65540_1.dds"},
	{0x45812D7F0E894D03ULL, "764700159.ogg"},
	{0x45815CC63712B9ABULL, "vxht_0_hightown_a3_9wg_0_0.vlm"},
	{0x4581926EE292A2E2ULL, "gen_im_trs_drk_junk2_02.uti"},
	{0x4581B299A6648D4CULL, "648213909.ogg"},
	{0x4582568563847A5AULL, "gen00fl_fenris6110897.cl"},
	{0x4582CCA7B75000F2ULL, "hlac_lt_0_lowtown_n_-65535_0y.dds"},
	{0x45830BBC2DDD5464ULL, "vxw_sundermount_2z4_0_1.vlm"},
	{0x4583C81C3CC7EFA4ULL, "ca_caverns2prc.arl"},
	{0x45842AE8AD4DECFBULL, "plt_zz_vispt_debug.nss"},
	{0x4584680D8FC50598ULL, "6001025_m.fxe"},
	{0x45846D8C9C8E1907ULL, "zz_dre_debug6168275.cl"},
	{0x45858A84FD169186ULL, "static_static_p27.psh.fxo.dx11.assembly"},
	{0x45863F40DD65DB17ULL, "gen00fl_fenris6105607.cl"},
	{0x45865D8543EABE0CULL, "vis121_saemus6167892.cl"},
	{0x45880787E23CCECFULL, "vxlt_0_lowtown_a3_n_c_e1l_0_0.vlm"},
	{0x45888BECB5B4F237ULL, "vxlt_0_lowtown_h9a_0_0.vlm"},
	{0x4588DB22A01048EEULL, "6180517_m.fxe"},
	{0x4588E299A59351ABULL, "dae000_pc_home_ambients6208355.cl"},
	{0x4588E368331AE7E7ULL, "268625196.ogg"},
	{0x45897E5DA8F6E5C0ULL, "vxca_caverns2_9f5_0_0.vlm"},
	{0x4589E2F5F2D6965BULL, "vxht_bartrand_n_738_0_0.vlm"},
	{0x458A29FDC016FEDCULL, "hlac_lt_0_lowtown_a2_n_65536_0y.dds"},
	{0x458A37B40BFAAC5EULL, "lt_gallowstemplar_f_2391.rml"},
	{0x458AB2922D18BDA9ULL, "vxlt_0_lowtown_docks_n_btt_0_0.vlm"},
	{0x458AC2494E02F6B9ULL, "vxw_encounter1_jb_0_0.vlm"},
	{0x458B076998B27354ULL, "vxht_mansion_n_59p_0_0.vlm"},
	{0x458B6E0D139883D2ULL, "hlac_lt_0_lowtown_docks_n_-65538_7z.dds"},
	{0x458B78CAD3AC5AB7ULL, "vxlt_0_lowtown_ev9_0_0.vlm"},
	{0x458BAF2E7624B989ULL, "lt_gallowscourtyard_n_51_0_lc.mmh"},
	{0x458BE351734CF9F0ULL, "vxw_encounter2_oa_0_0.vlm"},
	{0x458C29A360BF6AFAULL, "hlac_lt_0_lowtown_a3_n_c_-262144_4y.dds"},
	{0x458C2E39098B49CFULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgta.phy"},
	{0x458C387907F6BB87ULL, "zz_followers_banter6158808.cl"},
	{0x458C744FD846F719ULL, "vxca_caverns1_mines_gi9_0_0.vlm"},
	{0x458C86AC8D80707EULL, "follower_banter6141595.cl"},
	{0x458D78321DD07E28ULL, "vxca_cavern1_haunted_i7p_0_0.vlm"},
	{0x458E05ED08E97426ULL, "28667375.ogg"},
	{0x458E78E45CC4E9A6ULL, "one100dg_elren6032527.cl"},
	{0x458E9350B3F1AD0EULL, "art\\characters\\playercharacter\\humanmale\\arm_medb.phy"},
	{0x458F236BF400BA4FULL, "6133608_m.fxe"},
	{0x458F9A9D6C12C14FULL, "6112142_m.fxe"},
	{0x458FD575E62B4149ULL, "w_bonepit_26_0_ll.mmh"},
	{0x45904769DA5796ACULL, "170057127.ogg"},
	{0x459059F9400E17FDULL, "character_character_p0.vsh.fxo.dx11"},
	{0x45908FFC4D74D71AULL, "302014266.ogg"},
	{0x4590AD0F340DD6ADULL, "isa211_qunari_ambush6081816.cl"},
	{0x45918BE10AE4DFC6ULL, "213150305.ogg"},
	{0x459198C15A4E3275ULL, "hlac_dr_primevalentrance_1724_0y.dds"},
	{0x45935448493D6099ULL, "6095497_f.fxe"},
	{0x45945D39BD4D0649ULL, "vxht_0_hightown_n_a6h_0_0.vlm"},
	{0x4594BFBCC793A564ULL, "vxlt_foundry_n_we_0_0.vlm"},
	{0x45954E8B8E0D07A9ULL, "gen00fl_fenris6128028.cl"},
	{0x45957AE8E33B67D8ULL, "dae000cr_grand_cleric6124761.cl"},
	{0x45958C2B4B364532ULL, "vxlt_blackemporium_yb_0_0.vlm"},
	{0x4595C930AAE70B33ULL, "lowtown_male_cityguard_mob6150502.cl"},
	{0x45960CCB3183B4D8ULL, "hm_amb_commoner2.mor"},
	{0x4596E8CDAC3D5B05ULL, "268618495.ogg"},
	{0x459798220333970FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinstower.mao"},
	{0x4597A4220333ABACULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinstower.mmh"},
	{0x4597AA220333B59EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinstower.msh"},
	{0x4597B34CB754024DULL, "character_character_p112.psh.fxo"},
	{0x45980FC4970AED13ULL, "vxw_bonepit_yd_0_0.vlm"},
	{0x459876D48934A285ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\xx_emissive2.mao"},
	{0x4599592742E684FBULL, "vxht_bartrand_n_42p_0_0.vlm"},
	{0x45995CE2F6292694ULL, "vxht_playershouse_n_1v0_0_0.vlm"},
	{0x4599C2240BBD3FACULL, "6201632_m.fxe"},
	{0x4599EAA7C5A27337ULL, "vxlt_0_lowtown_eq8_0_0.vlm"},
	{0x459A552E65431270ULL, "vxw_encounter2_jn_0_0.vlm"},
	{0x459A868912FA8571ULL, "lt_0_lowtown_a2_n_40_0.rml"},
	{0x459B5C441A4F8529ULL, "hlac_lt_0_lowtown_n_-65537_2z.dds"},
	{0x459B71B9A5D16FD0ULL, "6165266_m.fxe"},
	{0x459B755B4232EBEDULL, "vxlt_0_lowtown_a2_gpk_0_0.vlm"},
	{0x459C52A33514961EULL, "hlac_lt_0_lowtown_a2_n_-131072_4y.dds"},
	{0x459CFA1502100D6DULL, "274477358.ogg"},
	{0x459D30222B584250ULL, "vxw_bonepit_1j1_0_0.vlm"},
	{0x459E9C3FDF0DA2D8ULL, "hlac_lt_0_lowtown_docks_n_-131073_4y.dds"},
	{0x459EEC040EE10C5FULL, "vxht_0_hightown_n_5io_0_0.vlm"},
	{0x459F0E5B0CFAFF92ULL, "vxca_cavern1_haunted_j48_0_0.vlm"},
	{0x459F17278BE532E8ULL, "gen00fl_merrill6130084.cl"},
	{0x459F79E112EA22D6ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgtp_0s.dds"},
	{0x459FBDEAE8C3C6A1ULL, "vxht_0_hightown_au6_0_0.vlm"},
	{0x459FCAF1BFDA95E0ULL, "hlac_lt_0_lowtown_a3_n_c_131072_6y.dds"},
	{0x45A06CE6F148B67FULL, "mcr310_sibling6076735.cl"},
	{0x45A06E3BC19B0907ULL, "hlac_lt_gallowstemplar_7198_5y.dds"},
	{0x45A097319E961867ULL, "6148956_m.fxe"},
	{0x45A0DF9CF14020ECULL, "662863236.ogg"},
	{0x45A1864F4C6339E5ULL, "ach_exp_exodus.dds"},
	{0x45A1ABFC8956F403ULL, "qun120_arishok6050311.cl"},
	{0x45A1E522ECA0B3A8ULL, "295843922.ogg"},
	{0x45A1E9B40729B894ULL, "791100451.ogg"},
	{0x45A1F0CB38361D61ULL, "vxlt_0_lowtown_docks_n_9do_0_0.vlm"},
	{0x45A2A5A9E261BCA2ULL, "6096796_f.fxe"},
	{0x45A31054481022DDULL, "vxht_mansion_n_6rc_0_0.vlm"},
	{0x45A323E07CC6BAE0ULL, "qun221_petrice6185479.cl"},
	{0x45A340F6B40D01C3ULL, "vxca_cave_me_0_0.vlm"},
	{0x45A36DE93893869BULL, "w_woundedcoast_97_0.mmh"},
	{0x45A379E938939B38ULL, "w_woundedcoast_97_0.mao"},
	{0x45A37EF43116433FULL, "hlac_w_sundermount_131075_0y.dds"},
	{0x45A38BE93893B985ULL, "w_woundedcoast_97_0.msh"},
	{0x45A39C5F5B364F07ULL, "hlac_lt_gallowstemplar_n_2388_3z.dds"},
	{0x45A4919777F3983FULL, "mag101tr_aom_way_out.ncs"},
	{0x45A4BE4CA95EF70AULL, "hlac_ca_caverns1_mines_24425_6y.dds"},
	{0x45A53D9600C616DFULL, "vxlt_warehouse_un_0_1.vlm"},
	{0x45A5DB42466CA42EULL, "199108525.ogg"},
	{0x45A5E5CFB84DD8ECULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\8gr_boosterpack.msh"},
	{0x45A5EBCFB84DE2DEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\8gr_boosterpack.mmh"},
	{0x45A5F7CFB84DF775ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\8gr_boosterpack.mao"},
	{0x45A610922CE9701DULL, "vxlt_0_lowtown_docks_a3_bv2_0_0.vlm"},
	{0x45A62602AC08A772ULL, "vxlt_0_lowtown_a3_n_c_f2g_0_0.vlm"},
	{0x45A6260F85AE6795ULL, "6128930_m.fxe"},
	{0x45A6A6DD34186357ULL, "hlac_lt_0_lowtown_131074_0x.dds"},
	{0x45A730B3803C99EEULL, "39106896.ogg"},
	{0x45A74910B2080B4CULL, "6073802_m.fxe"},
	{0x45A77D16C5C5A734ULL, "6117915_m.fxe"},
	{0x45A7812BEAB7A1F3ULL, "shriek_normal.utc"},
	{0x45A8E9CA1EB77471ULL, "6001889_f.fxe"},
	{0x45A9E7383429046DULL, "vxlt_0_lowtown_a2_gwj_0_0.vlm"},
	{0x45A9EB77810EA08FULL, "6087477_m.fxe"},
	{0x45AA2F95601E9C90ULL, "hlac_dr_dungeon_3207_0z.dds"},
	{0x45AA3A5523D2A5DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_browsing_store.gad"},
	{0x45AA692BD0E33E36ULL, "6120920_m.fxe"},
	{0x45AACBAC7925EE3AULL, "63179514.ogg"},
	{0x45AB2CD4E1AD49C3ULL, "hlac_lt_0_lowtown_n_-196609_1x.dds"},
	{0x45ABB7A9871ABA51ULL, "480793963.ogg"},
	{0x45AC13E720879EFEULL, "6075200_m.fxe"},
	{0x45AC2229CE3FF68AULL, "6034661_m.fxe"},
	{0x45AD0624A62C845EULL, "lt_0_lowtown_n_21_0.rml"},
	{0x45AD1A90E84AE463ULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.cess.ani"},
	{0x45AE27E9C7103463ULL, "hlac_ht_0_hightown_-131074_2y.dds"},
	{0x45AE5188B15728ECULL, "hlac_ca_cavern1_haunted_24426_11z.dds"},
	{0x45AE9FC5BF3FF3AFULL, "mag120_macha6080626.cl"},
	{0x45AEAF353A32606CULL, "ts_coterie_arm_table.gda"},
	{0x45AF163B63F0136DULL, "40369230.ogg"},
	{0x45AF19414BC07C1BULL, "gen00fl_bethany6101229.cl"},
	{0x45AF73B0E2D6FE42ULL, "6154912_m.fxe"},
	{0x45AFFDE99FB6636EULL, "6050047_m.fxe"},
	{0x45B049A774981084ULL, "246276115.ogg"},
	{0x45B10800B002A584ULL, "6177306_m.fxe"},
	{0x45B1205F91C4A4B8ULL, "hlac_ht_chantry_2534_9x.dds"},
	{0x45B13D5523D8C9F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_browsing_store.evt"},
	{0x45B15418AEE11528ULL, "terrain_blend.psh.fxo"},
	{0x45B25A9031763240ULL, "154132475.ogg"},
	{0x45B2D6E96B9E6160ULL, "794555791.ogg"},
	{0x45B312068671DDCEULL, "nat380_party_banter.cnv"},
	{0x45B335E8AD756F94ULL, "plt_zz_vispt_debug.plo"},
	{0x45B34A66F2FA845DULL, "gen00fl_bethany6112223.cl"},
	{0x45B3FD956EEB9515ULL, "vxht_0_hightown_a2_ajw_0_0.vlm"},
	{0x45B44D9A87188627ULL, "6076382_f.fxe"},
	{0x45B45C2AFF780413ULL, "1004081629.ogg"},
	{0x45B48C6B4DC5627CULL, "6048318_m.fxe"},
	{0x45B4C34CD22C6C2AULL, "hlac_lt_backalley_g_0_10y.dds"},
	{0x45B53EC244C554C3ULL, "qcr210ar_lowtown_nt.ncs"},
	{0x45B61EB5CE68467CULL, "vxlt_0_lowtown_docks_n_bfk_0_0.vlm"},
	{0x45B6340686744D3EULL, "nat380_party_banter.bnk"},
	{0x45B68B08B25C15BEULL, "1053241868.ogg"},
	{0x45B6B77080FBAFF8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appc_0n.dds"},
	{0x45B738E8492F4370ULL, "vxw_encounter2_pl_0_0.vlm"},
	{0x45B75599253B69C5ULL, "projdecal_projdecal_p0.vsh.fxo.dx11"},
	{0x45B768A8B1BC91CDULL, "6156188_m.fxe"},
	{0x45B7D67C2E562FF2ULL, "hlac_w_encounter3_-196606_0x.dds"},
	{0x45B7E5569C6E2C08ULL, "723136512.ogg"},
	{0x45B7FF1FDF1CA677ULL, "352010620.ogg"},
	{0x45B80F38FFE66217ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.gad"},
	{0x45B85915261F1CEDULL, "vxlt_0_lowtown_a2_gji_0_0.vlm"},
	{0x45B86062454225B8ULL, "plt_cod_cha_elthina.plo"},
	{0x45B863CD025F83AAULL, "1001475077.ogg"},
	{0x45B86D50A03A26DAULL, "6129633_m.fxe"},
	{0x45B89F183E45E5E8ULL, "lt_0_lowtown_docks_251_0.msh"},
	{0x45B8B1183E460479ULL, "lt_0_lowtown_docks_251_0.mao"},
	{0x45B8BD183E461812ULL, "lt_0_lowtown_docks_251_0.mmh"},
	{0x45B9865CCB8A258FULL, "vxht_chantry_n_26q_0_0.vlm"},
	{0x45B9D9E434EABFB5ULL, "hlac_w_bonepit_393218_0y.dds"},
	{0x45B9EF4B43894BF5ULL, "vxdr_primevaltomb_8p5_0_0.vlm"},
	{0x45B9FBBCE9E7D079ULL, "hlac_ht_0_hightown_a3_65537_1x.dds"},
	{0x45BA3290E855975AULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.cess.evt"},
	{0x45BA40DCA7E9B5D2ULL, "vxw_encounter2_camp_7s_0_0.vlm"},
	{0x45BA4F49C41A2D61ULL, "zz_followers_banter6158037.cl"},
	{0x45BA9176783DFB44ULL, "460743385.ogg"},
	{0x45BA92B6763C4D62ULL, "dae000_keep_male_servant_mo6161461.cl"},
	{0x45BAB32CB99276F2ULL, "follower_banter6149300.cl"},
	{0x45BB385C5D11B8AAULL, "llac_ca_caverns1_24427_0.dds"},
	{0x45BC3EC244CB732AULL, "qcr210ar_lowtown_nt.lst"},
	{0x45BC95CB61AFEF97ULL, "zz_slate2_tracking2.cut"},
	{0x45BCB1153A754303ULL, "zz_tr_testlevel.ncs"},
	{0x45BCD50C435CB298ULL, "68949916.ogg"},
	{0x45BD6C992D0B15ECULL, "gen_im_cft_reg_i1.uti"},
	{0x45BE1CD3B9816901ULL, "6089961_m.fxe"},
	{0x45BE555523E37CC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_browsing_store.ani"},
	{0x45BE8EF20705FFDEULL, "583886914.ogg"},
	{0x45BE9C8F8E3786F2ULL, "vxlt_hangedman_n_1po_0_0.vlm"},
	{0x45BEFCD71FB4E54AULL, "368753792.ogg"},
	{0x45BF214AC31B8DFEULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die.evt"},
	{0x45BF2438FFECA4A4ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.evt"},
	{0x45BF3203D88B2FD3ULL, "vxw_encounter2_camp_2y_0_0.vlm"},
	{0x45BF4E4CEBEA47B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\em_injured_movement.anb"},
	{0x45BF502F6683CBD4ULL, "439780795.ogg"},
	{0x45BF711C3143DCDBULL, "llac_w_encounter2_262148_0.dds"},
	{0x45BFBD69132C23E9ULL, "286881525.ogg"},
	{0x45BFF5C5CD46D369ULL, "gen00fl_fenris6126884.cl"},
	{0x45BFF5EBA64FC78AULL, "hlac_lt_0_lowtown_a2_-1_1x.dds"},
	{0x45C004EFE01E1B89ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masd_0s.dds"},
	{0x45C13190E85BB44DULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.cess.gad"},
	{0x45C16A8DE2B6F34EULL, "884727790.ogg"},
	{0x45C1CE677A553DE2ULL, "173687860.ogg"},
	{0x45C1F4E5EB071AE7ULL, "6056916_m.fxe"},
	{0x45C3770309E4D626ULL, "gen00fl_aveline6013249.cl"},
	{0x45C39B88052A5132ULL, "22814997.ogg"},
	{0x45C3C48512086F3BULL, "6044818_m.fxe"},
	{0x45C3DC913254EA23ULL, "124718017.ogg"},
	{0x45C4060A9424D944ULL, "dre000_varric6036976.cl"},
	{0x45C434B40C2C1F02ULL, "lt_gallowstemplar_f_2391.anb"},
	{0x45C4B90E644F9CEDULL, "225607300.ogg"},
	{0x45C575852E874B0EULL, "vxht_0_hightown_bs8_0_0.vlm"},
	{0x45C5D65C7ACAFF28ULL, "vxw_bonepitdestroyed_a3_1of_0_0.vlm"},
	{0x45C5E38F88DEB850ULL, "vxca_cavern1_haunted_cml_0_0.vlm"},
	{0x45C60B8071432E10ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basic.mao"},
	{0x45C60F8071433523ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basic.mmh"},
	{0x45C61D8071434D7DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basic.msh"},
	{0x45C6384AC321D329ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die.gad"},
	{0x45C69750B2F799FEULL, "vxlt_hangedman_n_2ug_0_0.vlm"},
	{0x45C6A06DB90375ABULL, "6092080_m.fxe"},
	{0x45C6E6CA64C632C4ULL, "zz_vis_debug6050343.cl"},
	{0x45C760535ED218FCULL, "art\\vfx\\cinematic\\vfx_meredith_smokey_c.impa.ani"},
	{0x45C814B642F2B001ULL, "6110421_m.fxe"},
	{0x45C82257899F80D3ULL, "6095240_f.fxe"},
	{0x45C824401234B2A4ULL, "gen00fl_dog6165755.cl"},
	{0x45C82CD2F61DD1DEULL, "6043255_f.fxe"},
	{0x45C898294A2E9E65ULL, "vxht_0_hightown_a2_at9_0_0.vlm"},
	{0x45C8ABBEF3218D54ULL, "vxw_encounter2_camp_rj_0_0.vlm"},
	{0x45C8BFA1D0D439E6ULL, "vxlt_0_lowtown_docks_n_b0k_0_0.vlm"},
	{0x45C933D4896465A7ULL, "6120837_m.fxe"},
	{0x45C968BF88FE1556ULL, "mcr371_scared_mage_m6086906.cl"},
	{0x45C9AA99F3658586ULL, "hlac_lt_0_lowtown_a3_n_c_131072_1z.dds"},
	{0x45C9FF429999CB4AULL, "gen00fl_sebastian6189399.cl"},
	{0x45CB8BF9E870B902ULL, "485905344.ogg"},
	{0x45CB9035DC090DF3ULL, "vxdr_primevalentrance_1j8_0_0.vlm"},
	{0x45CBB5212845A28EULL, "z_cloud.dds"},
	{0x45CBCD82484BB4F5ULL, "mag371tr_close_orsino.ncs"},
	{0x45CC3C38FFF7579DULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.ani"},
	{0x45CCD06682455773ULL, "vxlt_0_lowtown_docks_c5g_0_0.vlm"},
	{0x45CCEF28DE2575F5ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_crown_sa.dds"},
	{0x45CD094AC327A277ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_die.ani"},
	{0x45CD358CD151328CULL, "703589149.ogg"},
	{0x45CD5768EF411FF7ULL, "plt_cod_aow_spellcombo9.plo"},
	{0x45CD5B5B61BAF183ULL, "vxlt_warehouse_vb_0_0.vlm"},
	{0x45CE5AE2BE33E441ULL, "6075269_m.fxe"},
	{0x45CE6D535ED84D3EULL, "art\\vfx\\cinematic\\vfx_meredith_smokey_c.impa.gad"},
	{0x45CF18715E58066EULL, "6116600_m.fxe"},
	{0x45CF3F42B890F241ULL, "146230335.ogg"},
	{0x45CF50A8576884EEULL, "plt_one200pt_fools_gold.plo"},
	{0x45CF9250B42942D1ULL, "art\\characters\\playercharacter\\humanmale\\arm_medb.mmh"},
	{0x45CF9C924F703B3AULL, "6176219_m.fxe"},
	{0x45CF9E50B4295772ULL, "art\\characters\\playercharacter\\humanmale\\arm_medb.mao"},
	{0x45CFAC50B4296E8FULL, "art\\characters\\playercharacter\\humanmale\\arm_medb.msh"},
	{0x45CFB281FC7F3808ULL, "isa321_velasco6094477.cl"},
	{0x45D02C0BAA8F2399ULL, "6125753_m.fxe"},
	{0x45D1C55AFC2124AAULL, "vxlt_warehouse_n_hx_0_0.vlm"},
	{0x45D1D37EE336C342ULL, "hlac_ht_brothel_n_2995_3y.dds"},
	{0x45D29C1478E2A42EULL, "dae150_harshal6148843.cl"},
	{0x45D2E448DA8E1634ULL, "6036697_f.fxe"},
	{0x45D2F826AF682A7CULL, "hlac_w_bonepitmining_a2_458755_0x.dds"},
	{0x45D3212A737ADAF2ULL, "ht_0_hightownexplosion_n.arl"},
	{0x45D362F63B55F94FULL, "152440652.ogg"},
	{0x45D36FED7DAC1C18ULL, "624693886.ogg"},
	{0x45D3CD9DB5A070E5ULL, "vxca_caverns2_9dj_0_0.vlm"},
	{0x45D4205ADC2C43E5ULL, "264805187.ogg"},
	{0x45D47457CFDC464AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pulpit_n.dds"},
	{0x45D4932DEB26FD7EULL, "304575459.ogg"},
	{0x45D4DA55A4AE01EDULL, "hlac_dr_primevalentrance_1726_12z.dds"},
	{0x45D4DD4C70008B7BULL, "mag110_samson6018700.cl"},
	{0x45D51496A92825C6ULL, "418623628.ogg"},
	{0x45D548535EDE2D5DULL, "art\\vfx\\cinematic\\vfx_meredith_smokey_c.impa.evt"},
	{0x45D66DF717973153ULL, "vxca_caverns1_j4x_0_0.vlm"},
	{0x45D68614921FC837ULL, "vxca_cavern1_haunted_edv_0_0.vlm"},
	{0x45D710807151AA48ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basic.phy"},
	{0x45D71D9BE517A5D4ULL, "vxw_encounter2_camp_mk_0_0.vlm"},
	{0x45D7946BFE92B8D2ULL, "440561202.ogg"},
	{0x45D7F0FE2C83B36BULL, "lowtown_male_alienage1_sta6162717.cl"},
	{0x45D85196E557E626ULL, "gen_im_trs_crc_lgt_hlm_r2.uti"},
	{0x45D865BEED341FACULL, "hlac_lt_0_lowtown_a3_n_c_-65534_0y.dds"},
	{0x45D8781779E2C709ULL, "189622353.ogg"},
	{0x45D90D71F49659C2ULL, "187368520.ogg"},
	{0x45D9186BDF51C68DULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ichm.mmh"},
	{0x45D92A6BDF51E513ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ichm.msh"},
	{0x45D93C63AEDA03ACULL, "vxw_bonepitmining_a2_1xh_0_1.vlm"},
	{0x45D9F0EC741FEB73ULL, "220238275.ogg"},
	{0x45DA4A444C2A2416ULL, "hlac_w_blightlands_0_0y.dds"},
	{0x45DAD8000967952BULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v1.evt"},
	{0x45DAE273F15ADE71ULL, "llac_ht_keep_n_4519_13.dds"},
	{0x45DC10C2AB1DC7E6ULL, "vxca_caverns2_909_0_0.vlm"},
	{0x45DC300CAE760CBBULL, "llac_lt_0_lowtown_a2_n_c_65534_2.dds"},
	{0x45DC8975E99BABE4ULL, "640696775.ogg"},
	{0x45DCAB73E259D4F2ULL, "6068350_m.fxe"},
	{0x45DD1033E08DD464ULL, "6000930_m.fxe"},
	{0x45DD5ED70B16869EULL, "vxlt_0_lowtown_n_df2_0_0.vlm"},
	{0x45DD6F9F17CF2B25ULL, "llac_lt_backalley_n_65536_1.dds"},
	{0x45DD85D9348C6526ULL, "ash_02_l2.mao"},
	{0x45DE343A905B4A48ULL, "ss_raider_m_1.wbd"},
	{0x45DF8963E196ACCAULL, "6125642_m.fxe"},
	{0x45DFD6DB4D1CEC6BULL, "vxht_brothel_n_5hw_0_0.vlm"},
	{0x45DFF85BBC7FC78EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wo_basic.msh"},
	{0x45E0065BBC7FDFDFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wo_basic.mao"},
	{0x45E0125BBC7FF47CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wo_basic.mmh"},
	{0x45E03737EE173D2AULL, "vxlt_warehouse_n_oy_0_0.vlm"},
	{0x45E053EA5F64EE4EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_nud_0n.dds"},
	{0x45E057CF20C17426ULL, "plt_gen00pt_combat.plo"},
	{0x45E0BBF8BE1AC9D3ULL, "6177129_m.fxe"},
	{0x45E0CDCC4D5112A9ULL, "6077447_m.fxe"},
	{0x45E0D206669C8565ULL, "vxca_cavern1_haunted_dpy_0_0.vlm"},
	{0x45E0EB263BE11F4EULL, "lt_gallowscourtyard_n_48_0.rml"},
	{0x45E0FF8F86662BC1ULL, "6169519_m.fxe"},
	{0x45E1297A4867F2C7ULL, "gen_im_trs_tmp_junk2_01.uti"},
	{0x45E1744CA44A8FFCULL, "vxlt_0_lowtown_docks_n_bun_0_0.vlm"},
	{0x45E19FD7FD8B7501ULL, "6142413_m.fxe"},
	{0x45E1CA3E37893532ULL, "vxht_mansion2b_289_0_0.vlm"},
	{0x45E1EE474FC4C014ULL, "427385577.ogg"},
	{0x45E1EEDA569D5EFFULL, "hlac_lt_gallowscourtyard_-65538_3z.dds"},
	{0x45E1F100096DDEECULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v1.gad"},
	{0x45E23DC244EC408DULL, "qcr210ar_lowtown_nt.are"},
	{0x45E241F19A4FBF97ULL, "vxlt_gallowstemplar_n_5wv_0_0.vlm"},
	{0x45E2580F3ED795C9ULL, "6087765_m.fxe"},
	{0x45E263EC691A985FULL, "hlac_lt_0_lowtown_a2_-131072_9z.dds"},
	{0x45E26546B354DFA7ULL, "zz_prmpt_debug6200953.cl"},
	{0x45E2C13C67666457ULL, "dae000_bartrand6027560.cl"},
	{0x45E341A85779147DULL, "plt_one200pt_fools_gold.nss"},
	{0x45E34C837558ED1AULL, "vxlt_gallowscourtyard_6ki_0_0.vlm"},
	{0x45E38B4A9189B07BULL, "6037931_m.fxe"},
	{0x45E4413795450A8EULL, "gen_im_trs_cot_junk2_03.uti"},
	{0x45E45D681B45FCFFULL, "vxht_chantry_n_207_0_0.vlm"},
	{0x45E4A3F52AD2ABD2ULL, "204355249.ogg"},
	{0x45E4EC720B0C553CULL, "812242540.ogg"},
	{0x45E50BD5EFA12288ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_5x.dds"},
	{0x45E552D972D40B51ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3achar.msh"},
	{0x45E560D972D4236CULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3achar.mao"},
	{0x45E56CD972D4370FULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3achar.mmh"},
	{0x45E58F78D70DBE54ULL, "vxw_encounter2_camp_hh_0_0.vlm"},
	{0x45E5D0AC5273FAD9ULL, "398464829.ogg"},
	{0x45E65732E503A0DAULL, "llac_ht_mansion2_818_4.dds"},
	{0x45E689DE9F40577BULL, "hlac_lt_0_lowtown_docks_a3_65535_3y.dds"},
	{0x45E70324A65DF702ULL, "lt_0_lowtown_n_21_0.anb"},
	{0x45E8201BC9044AF4ULL, "lgt210ar_gang_hideout_d.are"},
	{0x45E85247D8F138DCULL, "gen_im_arm_cht_comb_t6.uti"},
	{0x45E8F0000973FBEAULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v1.ani"},
	{0x45E8FE78C00EA05CULL, "vxlt_gallowsprison_n_69x_0_0.vlm"},
	{0x45E9640E20E3AB30ULL, "llac_lt_0_lowtown_n_0_8.dds"},
	{0x45E9F12AE05E8FFBULL, "hlac_lt_0_lowtown_1_9x.dds"},
	{0x45E9F2266419D072ULL, "vxlt_hangedman_n_1gj_0_0.vlm"},
	{0x45EA116BDF602E4AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ichm.phy"},
	{0x45EA35163A44C1DCULL, "darktown_fem_lowlife_mob6176173.cl"},
	{0x45EAAB53A9172C58ULL, "qun211cr_final1.utc"},
	{0x45EAC5B27C8069A3ULL, "mag320_orsino6071183.cl"},
	{0x45EAF0BBD87EBD6BULL, "vxlt_0_lowtown_docks_b4a_0_0.vlm"},
	{0x45EB1DEFFAD1FA1BULL, "230273440.ogg"},
	{0x45EB3D870E8B79BFULL, "6126578_m.fxe"},
	{0x45EB461976F8B413ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v1.impa.evt"},
	{0x45EB9D0E2A6D43E3ULL, "mag110_samson6018133.cl"},
	{0x45EC055AF38357BBULL, "204383258.ogg"},
	{0x45EC08032EB54BC0ULL, "art\\characters\\weapons\\maces\\w_mce_mc102.mao"},
	{0x45EC294D923DFBD4ULL, "324263750.ogg"},
	{0x45EC41D1CC1E6C80ULL, "llac_dr_primevalentrance_1726_0.dds"},
	{0x45EC6B371A85C8A6ULL, "one300dg_solivitus6048324.cl"},
	{0x45ECC4632F690BC0ULL, "vxw_encounter1_j6_0_1.vlm"},
	{0x45ED149E594EEECCULL, "27252082.ogg"},
	{0x45ED28AFC0E8C641ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockbxxl.phy"},
	{0x45ED2D1EE8AA4C5CULL, "ht_0_hightown_a2_13_0.rml"},
	{0x45ED8376248E22F9ULL, "252425212.ogg"},
	{0x45EDF8E03C56CFCBULL, "combat_in_range.gda"},
	{0x45EE580699BA47C4ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf5a.msh"},
	{0x45EE6E0699BA6D66ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf5a.mmh"},
	{0x45EE8B4BFBB43225ULL, "68216039.ogg"},
	{0x45EEC211B5C94EDEULL, "6148837_f.fxe"},
	{0x45EED4D7D9AC20A7ULL, "182843816.ogg"},
	{0x45EED733E144D866ULL, "909832609.ogg"},
	{0x45EF3FA93F472B29ULL, "6078547_m.fxe"},
	{0x45EF46DCF6CFEB2DULL, "977752211.ogg"},
	{0x45F07C3290DCDB96ULL, "360335983.ogg"},
	{0x45F0912890D2629CULL, "zz_followers_banter6158625.cl"},
	{0x45F0B3CE8C45D817ULL, "vxlt_gallowstemplar_n_5jq_0_0.vlm"},
	{0x45F13F1976FD1454ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v1.impa.gad"},
	{0x45F153B83B4BD9F0ULL, "vxw_blightlands_6e_0_0.vlm"},
	{0x45F15CA18D9DE277ULL, "6071292_m.fxe"},
	{0x45F15CEF9840A7CDULL, "gen00fl_varric6075155.cl"},
	{0x45F17233D74589CEULL, "hlac_w_encounter2_camp_262149_0x.dds"},
	{0x45F18C74DCF1B808ULL, "6182719_m.fxe"},
	{0x45F25A1795BC9D00ULL, "vxlt_hangedman_34h_0_0.vlm"},
	{0x45F291FE87BA35B7ULL, "vxht_chantry_n_2cf_0_0.vlm"},
	{0x45F314074A0683D7ULL, "hlac_lt_0_lowtown_a2_-131072_1z.dds"},
	{0x45F3147DE40E3C08ULL, "244209511.ogg"},
	{0x45F37F8E1AD87163ULL, "784535494.ogg"},
	{0x45F39425E158541FULL, "6035215_f.fxe"},
	{0x45F396CED9479E4CULL, "902913660.ogg"},
	{0x45F3B6AE76D0C936ULL, "6055004_m.fxe"},
	{0x45F49EC456A805CBULL, "vxlt_gallowstemplar_n_5qh_0_0.vlm"},
	{0x45F4A0856EB019BEULL, "981685829.ogg"},
	{0x45F4CDBB595AC712ULL, "6061340_m.fxe"},
	{0x45F56E0DCF2F345BULL, "pose_transitions.gda"},
	{0x45F5D39666C629D5ULL, "hlac_lt_0_lowtown_a3_n_c_65535_4x.dds"},
	{0x45F684283323F5C1ULL, "lt_warehouse_n_4_0.anb"},
	{0x45F69BA174A845DFULL, "hlac_lt_backalley_n_1_2y.dds"},
	{0x45F720FA2EBF89F7ULL, "gen00fl_fenris6118717.cl"},
	{0x45F7557AF5FEF67FULL, "6135963_m.fxe"},
	{0x45F7598CDC8D5A3AULL, "43377259.ogg"},
	{0x45F7AA070038BFBDULL, "cst_lightning.mao"},
	{0x45F7AD14E89AB3A5ULL, "vxht_0_hightown_a3_byd_0_0.vlm"},
	{0x45F840BB8DA5C5CFULL, "6117453_m.fxe"},
	{0x45F84E1977034C32ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v1.impa.ani"},
	{0x45F8650699C2B231ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf5a.phy"},
	{0x45F89A835D9A647DULL, "pro000_opening_pc6001202.cl"},
	{0x45F9116D17F008F5ULL, "vxht_0_hightown_n_c2z_0_0.vlm"},
	{0x45F92FB3DBE92022ULL, "hlac_lt_foundry_n_320_16x.dds"},
	{0x45F95BF6A97DDDE0ULL, "vxw_sundermount_306_0_0.vlm"},
	{0x45F969286AB644E9ULL, "vxca_caverns2_8gz_0_0.vlm"},
	{0x45F9D32A7BDAA897ULL, "6155547_m.fxe"},
	{0x45FA41214F319624ULL, "lt_0_lowtown_docks_a3_168_0.msh"},
	{0x45FA53214F31B48DULL, "lt_0_lowtown_docks_a3_168_0.mao"},
	{0x45FA57214F31BBC6ULL, "lt_0_lowtown_docks_a3_168_0.mmh"},
	{0x45FA66B8C007CF5EULL, "vxlt_backalley_n_1pc_0_1.vlm"},
	{0x45FAA7F18F6EE466ULL, "fen110_party_comment6095071.cl"},
	{0x45FAB3FE93608774ULL, "cod_mgc_circle_history.utp"},
	{0x45FABA1668BA7DA1ULL, "6093896_m.fxe"},
	{0x45FB18AA086D00CBULL, "6129311_m.fxe"},
	{0x45FC2584973F387DULL, "168405045.ogg"},
	{0x45FC2DE5C48AE15CULL, "lgt260_way_blood_mage6148387.cl"},
	{0x45FC4FD0BE583681ULL, "vxlt_0_lowtown_a2_n_hfa_0_0.vlm"},
	{0x45FC558175E1D946ULL, "hlac_w_woundedcoast_262152_0y.dds"},
	{0x45FCF5CD9DC433C6ULL, "6169862_m.fxe"},
	{0x45FD3DBF62CE21AFULL, "vxht_mansion_n_3v4_0_0.vlm"},
	{0x45FD83FF70EF733CULL, "73522717.ogg"},
	{0x45FE1BC84EB5CA8EULL, "395745067.ogg"},
	{0x45FE23D57AC616ECULL, "hlac_ht_chantry_n_2534_0x.dds"},
	{0x45FEA825D00958E6ULL, "6072241_m.fxe"},
	{0x45FECDA4FC1DCEDBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.evt"},
	{0x45FEEB95CE43BBFCULL, "6070790_f.fxe"},
	{0x45FF1869346B3055ULL, "mag200_feynriel6056923.cl"},
	{0x45FF25AB7E3BF097ULL, "vxlt_gallowstemplar_n_5mp_0_0.vlm"},
	{0x45FF31580EC34E07ULL, "vxlt_0_lowtown_a3_n_c_ff2_0_0.vlm"},
	{0x45FF39E889B148EBULL, "lt_0_lowtown_docks_31_0_ll.mmh"},
	{0x45FFC465EE8EBB28ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_lgtc_0n.dds"},
	{0x45FFCB2BD8B25FCFULL, "25233679.ogg"},
	{0x45FFD12E2C3C1231ULL, "vxca_caverns2_78t_0_0.vlm"},
	{0x45FFDC773AE0585BULL, "vxht_0_hightown_a3_c70_0_1.vlm"},
	{0x45FFE20EDA72659CULL, "vxht_0_hightown_a3_c3n_0_0.vlm"},
	{0x460017AB925DC0A9ULL, "hlac_ht_keep_4519_18x.dds"},
	{0x46003FE4A58CB809ULL, "6142537_f.fxe"},
	{0x4600637C84976AF5ULL, "6113961_m.fxe"},
	{0x4600B463ED95F5CCULL, "w_encounter2_camp_35_0.msh"},
	{0x4600BA63ED95FFBEULL, "w_encounter2_camp_35_0.mmh"},
	{0x4600C663ED961455ULL, "w_encounter2_camp_35_0.mao"},
	{0x4600DFB349438D93ULL, "gen00fl_sebastian6179980.cl"},
	{0x46010BD181A6A5DEULL, "vxht_bartrand_n_6jc_0_0.vlm"},
	{0x4601118659D62A43ULL, "857322873.ogg"},
	{0x4601990F33987645ULL, "6124361_m.fxe"},
	{0x4601A377C672FBFAULL, "vxw_bonepitmining_a2_84_0_0.vlm"},
	{0x4601C19EF7C6F655ULL, "570761441.ogg"},
	{0x4601C8A63336C52AULL, "vxlt_darktown_4us_0_0.vlm"},
	{0x4601E7779441820EULL, "6102124_m.fxe"},
	{0x4602288B68C6E1B6ULL, "6052104_m.fxe"},
	{0x4602295D90A0E9A9ULL, "hlac_lt_0_lowtown_a2_n_65534_0z.dds"},
	{0x46022A45B24D79CEULL, "w_encounter3_111_0.mmh"},
	{0x46023645B24D8E65ULL, "w_encounter3_111_0.mao"},
	{0x46024445B24DA5BCULL, "w_encounter3_111_0.msh"},
	{0x4602D6B527E1F8CEULL, "6095835_m.fxe"},
	{0x4602FE35A85BC785ULL, "137590839.ogg"},
	{0x460309B0BA188208ULL, "vxlt_0_lowtown_docks_n_c6j_0_0.vlm"},
	{0x4603508690C1A9D6ULL, "dae000_meredith6092589.cl"},
	{0x4603FD0F1ACCFA10ULL, "vxht_0_hightown_bwz_0_0.vlm"},
	{0x4604423F92929999ULL, "6083808_m.fxe"},
	{0x46049C705592FFF6ULL, "dae000cr_grand_cleric6173688.cl"},
	{0x4604A3EEE2CF7CEDULL, "gen_im_cns_wrn_sandal_act1.uti"},
	{0x4604EBEF61BDF549ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedleft_enter.ani"},
	{0x4604FF2EDF714171ULL, "art\\levels\\4cinematics\\dwarf\\_textures\\dwarfchair_n.dds"},
	{0x460532CB86B8CA31ULL, "fex_ruin_circ_a.msh"},
	{0x4605358B0E1A4FCBULL, "hlac_lt_blackemporium_1055_17y.dds"},
	{0x46054CCB86B8F6EFULL, "fex_ruin_circ_a.mmh"},
	{0x46058C5B1242536EULL, "fx_lit_half.dds"},
	{0x4605C6A4FC23E11CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.gad"},
	{0x4605F15BBCA08B47ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wo_basic.phy"},
	{0x46068C34990A0A49ULL, "qun220cr_javaris6068908.cl"},
	{0x4606B2987F9BDE4FULL, "6112753_m.fxe"},
	{0x4606EAD8305F797EULL, "art\\characters\\creatures\\pridedemon\\animation\\cast2.ani"},
	{0x4606F4BE0A622F16ULL, "chargen_dae.gda"},
	{0x4607A6D26086DC92ULL, "834918724.ogg"},
	{0x4607D0AB6A0B338BULL, "hlac_lt_gallowstemplar_f_2388_6y.dds"},
	{0x46080F9869A91176ULL, "vxht_mansion2_2aj_0_0.vlm"},
	{0x4608239A598A6702ULL, "dae270pt_gallows.ncs"},
	{0x4608D0B8FE3297F7ULL, "6088691_f.fxe"},
	{0x4608DBCB537D8676ULL, "mcr370_meredith6136160.cl"},
	{0x4608DC318EE5745FULL, "gen00fl_fenris6108060.cl"},
	{0x46090E25E76177E1ULL, "33032123.ogg"},
	{0x460917A6A9EFCF6DULL, "w_bonepitmining_a2_18_0_lc.mmh"},
	{0x46096EA5714F6059ULL, "vxht_0_hightown_a3_c68_0_0.vlm"},
	{0x4609A3F8DC1CB37EULL, "art\\vfx\\creatures\\vfxroot.mmh"},
	{0x4609D55C9CD89372ULL, "hlac_ht_0_hightown_a3_1_0y.dds"},
	{0x460A644ACF3D1B1AULL, "hlac_ht_keep_n_3732_16z.dds"},
	{0x460A8B69467FA1ABULL, "follower_banter6169055.cl"},
	{0x460ABDDF1D7508CAULL, "vxht_keep_6zw_0_0.vlm"},
	{0x460AC28A7C61B719ULL, "413868348.ogg"},
	{0x460B06EF61C28ECBULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedleft_enter.gad"},
	{0x460B29342438A9A4ULL, "6939218.ogg"},
	{0x460B49AC805E421EULL, "hlac_lt_0_lowtown_a2_n_-65535_1x.dds"},
	{0x460BAF9C54C43A2FULL, "vxht_mansion_n_3s7_0_0.vlm"},
	{0x460BE2CA57455A37ULL, "hlac_lt_0_lowtown_docks_-65538_1z.dds"},
	{0x460C36B7032C554CULL, "vxlt_hangedman_10z_0_0.vlm"},
	{0x460C39468A31215EULL, "401873715.ogg"},
	{0x460CC5A4FC29FE1AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.ani"},
	{0x460CFA41AD104254ULL, "mrl000_marethari6129373.cl"},
	{0x460D1C3649A6582CULL, "mag170_thrask6017043.cl"},
	{0x460D6B0D8D1C35E0ULL, "vxlt_undercityencounter_18u_0_0.vlm"},
	{0x460E1F1BC925186FULL, "lgt210ar_gang_hideout_d.lst"},
	{0x460EA21A4B3B369AULL, "vxlt_gallowscourtyard_6xd_0_0.vlm"},
	{0x460EDD74B45566C7ULL, "6055081_m.fxe"},
	{0x460F3B4E1C8A9875ULL, "hlac_lt_0_lowtown_a2_65535_4x.dds"},
	{0x460F78F0132A9CB9ULL, "6051689_m.fxe"},
	{0x460FB0A72D143B08ULL, "57711368.ogg"},
	{0x460FBEB38150CF3CULL, "hlac_lt_gallowscourtyard_n_-131073_1x.dds"},
	{0x46102D8206FEFD97ULL, "framebuffereffects_horizontalblur_p0.vsh.fxo.dx11.assembly"},
	{0x46102EA4A4225C14ULL, "vrc321_party_barks6095774.cl"},
	{0x4610A43A1AE63E8CULL, "vxca_caverns1_mines_im6_0_0.vlm"},
	{0x4610BF09BB090650ULL, "vxw_bonepit_129_0_0.vlm"},
	{0x46112A45DFBE429BULL, "vxht_0_hightown_n_c6p_0_0.vlm"},
	{0x461139562FC2FE24ULL, "gen00fl_varric6076463.cl"},
	{0x46117B8DAC0E9A88ULL, "vxlt_0_lowtown_docks_n_c1k_0_0.vlm"},
	{0x4611B6153ABD8B9AULL, "zz_tr_testlevel.utt"},
	{0x4612004FF1B90AD3ULL, "vxht_mansion_n_57m_0_0.vlm"},
	{0x461203EF61C8A830ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedleft_enter.evt"},
	{0x46120678FBAAC13DULL, "6188697_m.fxe"},
	{0x461211E47D373161ULL, "6081737_f.fxe"},
	{0x46122FFB107FB24BULL, "953639628.ogg"},
	{0x4612655BF743BEBEULL, "dae000_pickpocket_scene6175066.cl"},
	{0x4612705BDA638360ULL, "6179793_m.fxe"},
	{0x46127CE9568EC1ECULL, "gen00fl_anders6103822.cl"},
	{0x4612B41C8488B219ULL, "6086708_m.fxe"},
	{0x46130C2A73B09A88ULL, "ht_0_hightownexplosion_n.rml"},
	{0x46135F96D9C61656ULL, "61353640.ogg"},
	{0x46139E2A03160266ULL, "art\\vfx\\combat\\vfx_str_att_back_r_leg_c.impa.evt"},
	{0x46140037BF122D95ULL, "mag210_arianni6024538.cl"},
	{0x46142237842C4E19ULL, "vxw_bonepitmining_a2_by_0_0.vlm"},
	{0x4614928E2914197CULL, "mag110_arianni6032474.cl"},
	{0x4614D33F8F49D6BCULL, "llac_ht_keep_n_3732_7.dds"},
	{0x461502D8306BDF4FULL, "art\\characters\\creatures\\pridedemon\\animation\\cast2.evt"},
	{0x4615A87F669601C5ULL, "152375413.ogg"},
	{0x4615DC41C7561627ULL, "vxlt_gallowscourtyard_n_3be_0_0.vlm"},
	{0x461628CBFED97671ULL, "vxw_encounter1_jz_0_0.vlm"},
	{0x461681755B9F29F6ULL, "vxht_mansion2_2jk_0_0.vlm"},
	{0x461684D0515DDD7DULL, "380009908.ogg"},
	{0x4616D77877F3A6B6ULL, "gen00fl_isabela6088675.cl"},
	{0x46170518EE6DE6A5ULL, "vxlt_0_lowtown_a2_n_hgz_0_0.vlm"},
	{0x461731FED04FF1DCULL, "gen00fl_fenris6126832.cl"},
	{0x46175BFC938B4BDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch02.ani"},
	{0x461782F8DC28B83DULL, "art\\vfx\\creatures\\vfxroot.anb"},
	{0x4617AFA9A58956F5ULL, "vxht_brothel_n_59f_0_0.vlm"},
	{0x4617F35E27231C83ULL, "37189014.ogg"},
	{0x461823D4608DE3A4ULL, "hlac_w_woundedcoast_262151_0z.dds"},
	{0x4618F5C7E864FC5CULL, "zz_ext.arl"},
	{0x46190B7B981B323CULL, "w_bonepitdestroyed_a3_199_0.mao"},
	{0x4619177B981B46DFULL, "w_bonepitdestroyed_a3_199_0.mmh"},
	{0x46191D7B981B5081ULL, "w_bonepitdestroyed_a3_199_0.msh"},
	{0x46192FBC0F6B214AULL, "vxht_keep_6gp_0_0.vlm"},
	{0x46194340FED89FA0ULL, "zz_followers_banter6159468.cl"},
	{0x4619C5F7F187FE7BULL, "w_encounter1_53_0.mmh"},
	{0x4619D1F7F1881218ULL, "w_encounter1_53_0.mao"},
	{0x4619E3F7F1883165ULL, "w_encounter1_53_0.msh"},
	{0x4619F37D9E92381DULL, "hlac_lt_undercitydungeon_-1_44x.dds"},
	{0x461AB52A031C47B1ULL, "art\\vfx\\combat\\vfx_str_att_back_r_leg_c.impa.gad"},
	{0x461AE893E2478547ULL, "302565835.ogg"},
	{0x461BED022B42DE8BULL, "hlac_ht_mansion2b_818_63x.dds"},
	{0x461C03D83071FFF8ULL, "art\\characters\\creatures\\pridedemon\\animation\\cast2.gad"},
	{0x461C24563D066428ULL, "primitives_primitive_p0.psh.fxo.dx11"},
	{0x461C50CD17E42ED7ULL, "6169180_m.fxe"},
	{0x461C54AEB048C348ULL, "vxlt_gallowstemplar_n_2r9_0_0.vlm"},
	{0x461C61D6E27ABDCEULL, "6025803_m.fxe"},
	{0x461C86C507B8E0F7ULL, "dae000ip_wall_lean_right_d.utp"},
	{0x461C948DE7D5D95DULL, "vxlt_undercityencounter_u0_0_0.vlm"},
	{0x461CD8CB4B30E42BULL, "dre180cs_ragstoriches_war.cut"},
	{0x461D66917F79094EULL, "6129251_m.fxe"},
	{0x461D68FC938FCD1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch02.gad"},
	{0x461DEF8B6592D6DEULL, "vxht_bartrand_n_6pe_0_0.vlm"},
	{0x461EA38ED1A17900ULL, "vxca_caverns1_hib_0_0.vlm"},
	{0x461ECC75C57D39DBULL, "zz_qun_debug6206594.cl"},
	{0x461ED50908464408ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_008a_0d.dds"},
	{0x461F16170CDC570CULL, "vxca_caverns1_mines_ij7_0_0.vlm"},
	{0x461F57444978FA95ULL, "vxdr_primevaltomb_760_0_0.vlm"},
	{0x461F7BD5E073F88FULL, "store_dae310_armor.utm"},
	{0x461F7FE9F9F1B164ULL, "z_cloud031.dds"},
	{0x461F85C1B2AF70EFULL, "mcr370_templar26087011.cl"},
	{0x461F9131EB079CDBULL, "mcr370_negotiation6080205.cl"},
	{0x461F9C22D1B45B1BULL, "vxht_0_hightown_n_c3s_0_0.vlm"},
	{0x461FB9569F631365ULL, "6076295_m.fxe"},
	{0x4620539291F4B579ULL, "vxca_cavern1_haunted_fmh_0_0.vlm"},
	{0x4620892C84258264ULL, "6185152_m.fxe"},
	{0x46217ADD666213F1ULL, "follower_banter6155649.cl"},
	{0x4621AB89F1EDD4F4ULL, "517484706.ogg"},
	{0x4621B62A0322685FULL, "art\\vfx\\combat\\vfx_str_att_back_r_leg_c.impa.ani"},
	{0x4622621AFD080C30ULL, "vxlt_0_lowtown_a2_bza_0_0.vlm"},
	{0x46231C93344E7B7EULL, "lgt320_shw_ors_secret_meet6156366.cl"},
	{0x46234868356FB34EULL, "762718830.ogg"},
	{0x4623A8C7811051CCULL, "6127645_f.fxe"},
	{0x4623D27CDFDAE1B7ULL, "vxca_cave_187_0_0.vlm"},
	{0x462443FC9395AD3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch02.evt"},
	{0x46245E789C14D316ULL, "vxht_mansion2b_2iz_0_0.vlm"},
	{0x4624F3524D954276ULL, "vxht_mansion2_2wl_0_0.vlm"},
	{0x462523B6BB716170ULL, "100029439.ogg"},
	{0x46252742DCECFB2AULL, "6048283_f.fxe"},
	{0x462542BCAC6D6130ULL, "qun111_ketojan6041025.cl"},
	{0x46255A3A90978D0AULL, "ss_raider_m_1.bnk"},
	{0x46256F94F2C2A075ULL, "vxht_0_hightown_a2_c7n_0_0.vlm"},
	{0x4625E832FB666AA4ULL, "vxht_mansion2b_2v5_0_0.vlm"},
	{0x4625EF63AA216CF8ULL, "6101035_m.fxe"},
	{0x462646796E0870AAULL, "zz_prmpt_debug6200932.cl"},
	{0x4626835E9ECFFE2CULL, "vxdr_0_deeproads_b4d_0_0.vlm"},
	{0x4626CD91C1856170ULL, "follower_banter6161339.cl"},
	{0x4626F3BFBC1AB1B8ULL, "2099109.ogg"},
	{0x462731C63F728A70ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.mmh"},
	{0x46277B9A09F469B1ULL, "dae000_fenarel6133321.cl"},
	{0x4627D5E3825EF1FEULL, "6149309_m.fxe"},
	{0x4627E6561A2C8ED6ULL, "vxlt_gallowscourtyard_4xj_0_0.vlm"},
	{0x4627E8BDFA9CB3DBULL, "6125373_m.fxe"},
	{0x46286CBE8DE467E9ULL, "one100dg_lia6055136.cl"},
	{0x4628B00335C40893ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_ogre_ogrewarr.ani"},
	{0x46294AF7FD3F95B4ULL, "6089797_m.fxe"},
	{0x4629D2496DD35EDEULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax104a_0n.dds"},
	{0x4629FF6D11F3C2A5ULL, "zz_mrl_debug6184204.cl"},
	{0x462A45EEA386D093ULL, "zz_one_debug6077394.cl"},
	{0x462A59F9D444B6F5ULL, "369662270.ogg"},
	{0x462A72C14317EC4DULL, "580568070.ogg"},
	{0x462A966EA939DEC5ULL, "vxlt_gallowstemplar_n_2go_0_0.vlm"},
	{0x462B88B0B1EA0F5AULL, "hlac_lt_warehouse_n_0_5y.dds"},
	{0x462C5F0351BA9BC2ULL, "6156248_m.fxe"},
	{0x462CCD8FA69729AAULL, "487113392.ogg"},
	{0x462CF597AE3AE400ULL, "6068846_m.fxe"},
	{0x462D4C265F1BCF44ULL, "zz_tbe100_barber6191005.cl"},
	{0x462D654F766A0830ULL, "6133809_m.fxe"},
	{0x462DA8CE1812CA62ULL, "dae000tr_preload_creature.utt"},
	{0x462E0DFFC3AA739BULL, "vxht_0_hightown_n_c8r_0_0.vlm"},
	{0x462E17E6607C677DULL, "qun231cr_qunari2.utc"},
	{0x462EADD9CBF3D447ULL, "dae300_petrice6097870.cl"},
	{0x462EB5727F992BCDULL, "422577839.ogg"},
	{0x462F42AD0397FE23ULL, "vxht_bartrand_n_5pg_0_0.vlm"},
	{0x462F437EF996DEF4ULL, "523913106.ogg"},
	{0x462F9C801131F2DAULL, "hlac_ca_caverns2prc_5802_47y.dds"},
	{0x462FB3B9696C21D4ULL, "vxlt_0_lowtown_docks_cs2_0_0.vlm"},
	{0x463013102ABC8747ULL, "vxlt_0_lowtown_a2_n_gn4_0_0.vlm"},
	{0x4630652833553951ULL, "lt_warehouse_n_4_0.rml"},
	{0x46309AA9FDD8677AULL, "vxlt_hangedman_4u_0_0.vlm"},
	{0x4630FA0D9C1F110FULL, "365293221.ogg"},
	{0x46310BE815DEAD14ULL, "578270926.ogg"},
	{0x463140716B84DC87ULL, "370818677.ogg"},
	{0x46316990CDFC8636ULL, "vxht_0_hightown_anh_0_0.vlm"},
	{0x46319BF23A3BA3F0ULL, "6127057_m.fxe"},
	{0x4633652F3F8B5AF6ULL, "vxht_mansion2_2pm_0_0.vlm"},
	{0x4633C57AE62F0F5DULL, "mer140cr_jansen.utc"},
	{0x4634F53822864755ULL, "llac_lt_warehouse_65536_3.dds"},
	{0x46350FFBCDB99957ULL, "6034729_m.fxe"},
	{0x463554C63F7F028FULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.anb"},
	{0x46355DE339C5F4CFULL, "vxlt_0_lowtown_a3_n_c_emp_0_0.vlm"},
	{0x463580B699C780D0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_maga_0d.dds"},
	{0x4635BF7D732D284EULL, "658358384.ogg"},
	{0x46360D058BFECA3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back.gad"},
	{0x4636288EE303CDCAULL, "ca_caverns1_mines_26035_ll.mmh"},
	{0x46363F8E56242807ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_102a_sa.dds"},
	{0x46366F379C534B12ULL, "zz_dae_debug6029811.cl"},
	{0x4636A89A1906E375ULL, "6116834_m.fxe"},
	{0x4636EC746F297F3CULL, "lgt110_urn_barker6133002.cl"},
	{0x4637BA64EEEB361CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_45walkway.mao"},
	{0x4637C664EEEB4ABFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_45walkway.mmh"},
	{0x4637CC64EEEB5561ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_45walkway.msh"},
	{0x463810AF4747A3D1ULL, "hlac_ca_caverns1_24424_21y.dds"},
	{0x463863EF89045F8EULL, "vxdr_primevalentrance_14s_0_0.vlm"},
	{0x4638CBF29BBEE057ULL, "6167898_m.fxe"},
	{0x46393486FBD05FD7ULL, "6163380_m.fxe"},
	{0x46393A1691D58464ULL, "6050162_m.fxe"},
	{0x46395EFF6E835E05ULL, "vfx_bird_raven04.dds"},
	{0x463A60FD0F7EC51BULL, "art\\characters\\creatures\\ragedemon\\animation\\tn90l.ani"},
	{0x463A67656D5EA602ULL, "282202763.ogg"},
	{0x463AF1E6F12FA03CULL, "6107279_m.fxe"},
	{0x463BA2204A34DC24ULL, "vxht_bartrand_n_63b_0_0.vlm"},
	{0x463C040902734E44ULL, "llac_w_bonepitmining_a2_0_13.dds"},
	{0x463C384F5A774624ULL, "445425035.ogg"},
	{0x463CD8A4B4234761ULL, "vxht_0_hightown_cal_0_0.vlm"},
	{0x463CF8B12CDB2FBDULL, "6048236_m.fxe"},
	{0x463D02058C04D62CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back.evt"},
	{0x463D3BEF0080D7E8ULL, "vxdr_primevaltomb_3tk_0_0.vlm"},
	{0x463D751CD55A3824ULL, "dae000ip_crowd_lm3_talk10.utp"},
	{0x463D870335D61F3DULL, "art\\cinematics\\deathblows\\ogre\\cs_db_ogre_ogrewarr.gad"},
	{0x463E10C766B3B4F3ULL, "vxw_bonepit_20s_0_0.vlm"},
	{0x463E47F55BCC2E7CULL, "hlac_lt_0_lowtown_docks_a3_n_131071_2y.dds"},
	{0x463E4C1EA510D062ULL, "w_woundedcoast_186_0.msh"},
	{0x463E4E1EA510D318ULL, "w_woundedcoast_186_0.mmh"},
	{0x463E5A1EA510E7BBULL, "w_woundedcoast_186_0.mao"},
	{0x463E829FB85E7512ULL, "6122135_m.fxe"},
	{0x463F1E47B7B4AB59ULL, "gen_im_cns_flu_carver_01.uti"},
	{0x463F372121C7133AULL, "vxw_encounter1_bu_0_0.vlm"},
	{0x463FAD97E789E184ULL, "vxht_0_hightown_ap4_0_0.vlm"},
	{0x463FBE4678013E27ULL, "6041178_m.fxe"},
	{0x463FCD3687B06B7EULL, "159489352.ogg"},
	{0x463FF27FEBCC3C41ULL, "vxht_mansion2_1yz_0_0.vlm"},
	{0x46403D3C35B047CFULL, "nat380_sibling6068015.cl"},
	{0x46408C8164FC82CCULL, "6129845_f.fxe"},
	{0x46409DFF62FF2248ULL, "vxlt_gallowstemplar_4ue_0_0.vlm"},
	{0x4640A57A4FEEC92DULL, "200210314.ogg"},
	{0x46411D953B3FF081ULL, "466565076.ogg"},
	{0x46413EFFB5A8E065ULL, "84224515.ogg"},
	{0x46419163EDCD5159ULL, "w_encounter2_camp_35_0.phy"},
	{0x4641AB652CD9B224ULL, "459083872.ogg"},
	{0x464256C5926B6F29ULL, "male_version.cut"},
	{0x4642E327BCED3960ULL, "vxlt_0_lowtown_docks_a3_d2i_0_0.vlm"},
	{0x46443B57ED0CA627ULL, "hlac_w_encounter2_4_0y.dds"},
	{0x4644AB3337186248ULL, "vxht_0_hightown_a3_atq_0_0.vlm"},
	{0x4644C8DA5D2039ACULL, "vxlt_blackemporium_8s_0_0.vlm"},
	{0x46454EC3C7F260D6ULL, "ser120_ghyslain6104202.cl"},
	{0x46457C2323649EBBULL, "dae000_hate_follower_react6103250.cl"},
	{0x4645B0C424018726ULL, "ft_l_dust.mao"},
	{0x4645F608B0FB53AAULL, "qun221_viscount6069185.cl"},
	{0x46462EB1E12BC6B4ULL, "6087597_m.fxe"},
	{0x4646615A7577F348ULL, "vxht_0_hightown_cai_0_0.vlm"},
	{0x46468796DCA5394DULL, "vxlt_undercityserialkiller_1ak_0_0.vlm"},
	{0x46477C62F7C858D5ULL, "6174045_m.fxe"},
	{0x464783CEC9AF8A09ULL, "35124392.ogg"},
	{0x46479A63EDD1CC75ULL, "w_encounter2_camp_35_0.rml"},
	{0x4648551EA51933A3ULL, "w_woundedcoast_186_0.phy"},
	{0x4648FEA65CF4EFCDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basic.mmh"},
	{0x464902A65CF4F61EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basic.mao"},
	{0x46490B531ED29E76ULL, "vxca_cavern1_haunted_ih3_0_0.vlm"},
	{0x464910A65CF50E53ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basic.msh"},
	{0x4649326B3BE6C906ULL, "hlac_lt_0_lowtown_a3_n_c_131069_0y.dds"},
	{0x46497218B06F793AULL, "272202568.ogg"},
	{0x4649945A67B9CEE1ULL, "vxht_0_hightown_a3_aqk_0_0.vlm"},
	{0x464A58A59EC21C8CULL, "6080189_m.fxe"},
	{0x464ABA3385233979ULL, "qun131_taskmaster6039392.cl"},
	{0x464AEA058C10EA85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back.ani"},
	{0x464BC1F55E29A4BCULL, "6151153_m.fxe"},
	{0x464C34DC126B9325ULL, "vxht_mansion2_2l6_0_0.vlm"},
	{0x464C7A0E59959B9EULL, "877110693.ogg"},
	{0x464CA4407161D5DCULL, "846360298.ogg"},
	{0x464D2F70280C000CULL, "263734016.ogg"},
	{0x464D47ACB2D7BC90ULL, "w_encounter3_195_0.msh"},
	{0x464D55ACB2D7D4AAULL, "w_encounter3_195_0.mmh"},
	{0x464D59ACB2D7DBE1ULL, "w_encounter3_195_0.mao"},
	{0x464D8807D9F7561BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2walk.msh"},
	{0x464D9A07D9F77566ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2walk.mao"},
	{0x464DA607D9F78905ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2walk.mmh"},
	{0x464DA8FE13BD2BBAULL, "vxw_encounter1_gv_0_0.vlm"},
	{0x464DAF0C4C153DB2ULL, "hlac_lt_0_lowtown_n_-65536_0x.dds"},
	{0x464DBDC89CF357D1ULL, "6107726_m.fxe"},
	{0x464E60D5471C569BULL, "1047992456.ogg"},
	{0x464E87FD0F8FB0A5ULL, "art\\characters\\creatures\\ragedemon\\animation\\tn90l.gad"},
	{0x464EFBE16D37D196ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masf.msh"},
	{0x464F05E16D37E294ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masf.mmh"},
	{0x464F2DAFC13B997DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockbxxl.mao"},
	{0x464F31AFC13B9FB6ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockbxxl.mmh"},
	{0x464F3BAFC13BB0B4ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockbxxl.msh"},
	{0x464FA96DB70E0794ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_c.impa.ani"},
	{0x464FFE24FA4645F8ULL, "vxw_sundermount_32t_0_0.vlm"},
	{0x46501245140C32EBULL, "mrl150_hunter6071292.cl"},
	{0x465018DABE0F21E3ULL, "w_blightlands_149_0.msh"},
	{0x465026DABE0F393DULL, "w_blightlands_149_0.mmh"},
	{0x46502ADABE0F404EULL, "w_blightlands_149_0.mao"},
	{0x465047E72913AE6FULL, "217845262.ogg"},
	{0x4650B0F27528E4A7ULL, "vxlt_gallowstemplar_650_0_0.vlm"},
	{0x4650EA0486D2717BULL, "692195175.ogg"},
	{0x46510104B9DFB87AULL, "vxlt_0_lowtown_n_ewh_0_0.vlm"},
	{0x46523C2D28C278FAULL, "vxdr_0_deeproads_8si_0_0.vlm"},
	{0x4652E4353583C3C1ULL, "gen00fl_isabela_init6110260.cl"},
	{0x4652EB32DCC08EE3ULL, "vxlt_gallowscourtyard_6wh_0_0.vlm"},
	{0x465326B7BA4CA86FULL, "gen00fl_aveline_generic_4.stg"},
	{0x46532C2C4D9A633FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_rightkeep_1.mao"},
	{0x465367E4E1AE3758ULL, "450984817.ogg"},
	{0x46539BB1B3897A85ULL, "vxw_bonepitdestroyed_a3_85_0_0.vlm"},
	{0x4653CD3893DFCAA3ULL, "gen00fl_merrill6106813.cl"},
	{0x4653FB8A2DFF7E56ULL, "266845775.ogg"},
	{0x46540D973A345F75ULL, "dae000ip_crowd_mf7_talk15.utp"},
	{0x46541155DAD704B5ULL, "lgt220cr_mag_tmplr_archer.utc"},
	{0x4654A4D2A7155030ULL, "gen00fl_merrill6114055.cl"},
	{0x4654AA88E1E56CF0ULL, "llac_lt_0_lowtown_n_131072_0.dds"},
	{0x4654D337676E0BC8ULL, "vxht_0_hightown_cdf_0_0.vlm"},
	{0x4654F973CE9B51CDULL, "vxlt_undercityserialkiller_1fj_0_0.vlm"},
	{0x46554068B640E19AULL, "hlac_lt_0_lowtown_a2_n_c_65536_8y.dds"},
	{0x4655F02694A3917FULL, "vxlt_0_lowtown_exo_0_0.vlm"},
	{0x4655FC4B269AAA5BULL, "vxlt_gallowstemplar_f_4ws_0_0.vlm"},
	{0x4656367714457CFBULL, "ser270_templar6165688.cl"},
	{0x4656A66DB71421F6ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_c.impa.gad"},
	{0x4656B510547A4D55ULL, "6031830_m.fxe"},
	{0x4656B664B4CDC997ULL, "zz_mageattacks.cut"},
	{0x4656CA4C0CF19107ULL, "6043284_m.fxe"},
	{0x4656E06F56170750ULL, "316876537.ogg"},
	{0x46570A151F3B153BULL, "art\\vfx\\creatures\\fxc_ashwraith_c.impa.evt"},
	{0x46571E7208147102ULL, "gen00fl_merrill6105675.cl"},
	{0x4657423A1052B228ULL, "hlac_ht_0_hightown_65533_1x.dds"},
	{0x46576D6C10887CBEULL, "vxlt_0_lowtown_a2_n_c_gy2_0_0.vlm"},
	{0x4657F849F37DCA14ULL, "one110_meeran6089848.cl"},
	{0x465803609218B462ULL, "287068669.ogg"},
	{0x4658064E6577E5F5ULL, "6084874_m.fxe"},
	{0x46588EBC12014FC1ULL, "6142220_m.fxe"},
	{0x4659CA1B778959B1ULL, "lt_0_lowtown_docks_25_0_lc.mmh"},
	{0x4659E19649E81B66ULL, "mcr371_hostile_anders6082110.cl"},
	{0x4659E2D8E6EE53B8ULL, "hlac_ca_cavern1_haunted_24426_6x.dds"},
	{0x4659F7A65D03568AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0rw_basic.phy"},
	{0x465A2BE0F8A39B3EULL, "6205612_m.fxe"},
	{0x465A42774D812375ULL, "vxw_bonepit_1v4_0_0.vlm"},
	{0x465A712262BB2689ULL, "6028946_m.fxe"},
	{0x465A9843D97A47A3ULL, "vxht_bartrand_n_5ah_0_0.vlm"},
	{0x465B0987BB2D52B2ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsa_0n.dds"},
	{0x465B3C1C21DC38A8ULL, "vxht_brothel_n_508_0_0.vlm"},
	{0x465BC74DEFEEC835ULL, "6128909_m.fxe"},
	{0x465BD2E52BC43F6CULL, "hlac_ca_cavern1_haunted_24424_27y.dds"},
	{0x465C0A23999D1E78ULL, "hlac_w_bonepitdestroyed_a3_458756_0y.dds"},
	{0x465C1F1E4923A802ULL, "vxht_mansion_n_6b8_0_0.vlm"},
	{0x465C6106D97EC432ULL, "sib131_slave_a6165265.cl"},
	{0x465C74A300D9F00AULL, "304621648.ogg"},
	{0x465CA580D7F3FF3AULL, "qcr270_fake_meredith.cnv"},
	{0x465CD639CFB86D41ULL, "vxht_mansion2_1gx_0_0.vlm"},
	{0x465CE3A92A0D24BCULL, "gen00fl_sebastian6180808.cl"},
	{0x465D03151F3F757CULL, "art\\vfx\\creatures\\fxc_ashwraith_c.impa.gad"},
	{0x465D10DAF8F28B00ULL, "vxlt_0_lowtown_docks_bfd_0_0.vlm"},
	{0x465D35D7CF7E25C7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoorcobble_sa.dds"},
	{0x465D393FFF40906CULL, "6173944_m.fxe"},
	{0x465D8378B0C1CE8EULL, "6054843_m.fxe"},
	{0x465D9C45AFA1FD28ULL, "zz_mrl_debug6185316.cl"},
	{0x465DA2081D0CFF15ULL, "mag101_grace6034231.cl"},
	{0x465DA824BE7D05BFULL, "293873959.ogg"},
	{0x465DBE43CF685F56ULL, "rog_escape.dds"},
	{0x465DC16DB71A6E55ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_c.impa.evt"},
	{0x465E0A8E3A15A157ULL, "vxlt_0_lowtown_a3_n_c_ht2_0_0.vlm"},
	{0x465E4BB607EE622FULL, "6115301_m.fxe"},
	{0x465EEC46566D6CE7ULL, "qun211cr_henchman.ncs"},
	{0x465FBEE979DF17B4ULL, "188462283.ogg"},
	{0x465FFE05C1AC0AC2ULL, "6071859_m.fxe"},
	{0x46600D634D2CA500ULL, "zz_followers_banter6157581.cl"},
	{0x46609464F8186089ULL, "250759797.ogg"},
	{0x4660AC32CD17478CULL, "6032041_m.fxe"},
	{0x4660C2196B92318DULL, "mag100_karras6034715.cl"},
	{0x4660C780D7F82182ULL, "qcr270_fake_meredith.bnk"},
	{0x4660CC474EB73F1BULL, "vxw_woundedcoast_a1_0_0.vlm"},
	{0x4661167F3E197F6DULL, "llac_ca_cavern1_haunted_24426_9.dds"},
	{0x4661190BA3156F11ULL, "vxdr_primevalentrance_1lh_0_0.vlm"},
	{0x466134713BDF8A1AULL, "hlac_ht_chantry_2534_9z.dds"},
	{0x466204F692157BC8ULL, "isa211_isabela6202209.cl"},
	{0x46621B4B8EACD9E2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglebody_sa.dds"},
	{0x4662523F828671D9ULL, "719448530.ogg"},
	{0x46626672DB9CD6C2ULL, "hlac_ca_cavern1_haunted_24427_2y.dds"},
	{0x46626E8D1AF14371ULL, "character_character_p208.psh.fxo.dx11.assembly"},
	{0x4663041B3C72EE79ULL, "vxlt_0_lowtown_n_byf_0_0.vlm"},
	{0x466383A92D3F0D21ULL, "hlac_lt_0_lowtown_docks_a3_1_0y.dds"},
	{0x46639EDF68AC765DULL, "hlac_lt_0_lowtown_a2_-65538_0x.dds"},
	{0x466402151F45927AULL, "art\\vfx\\creatures\\fxc_ashwraith_c.impa.ani"},
	{0x46646289A8B3690CULL, "vxca_caverns1_mines_j2p_0_0.vlm"},
	{0x46648F43A0B86970ULL, "vxw_encounter2_4p_0_0.vlm"},
	{0x46649D0F8695BB45ULL, "6175820_m.fxe"},
	{0x4664FD4CD1E46FACULL, "hlac_lt_gallowscourtyard_-131071_0x.dds"},
	{0x4665479DFFF0636BULL, "vxht_keep_n_49b_0_0.vlm"},
	{0x466565062C79CAA1ULL, "dae000_fem_commoner1_mob6174504.cl"},
	{0x46656FC27FBD7237ULL, "166637118.ogg"},
	{0x46658A5611BB6920ULL, "323891581.ogg"},
	{0x466592129DCB3D14ULL, "vxlt_warehouse_11u_0_0.vlm"},
	{0x466597286AD53675ULL, "vxw_woundedcoast_w9_0_0.vlm"},
	{0x4665B06C59B4C1E3ULL, "vxht_mansion2_3d9_0_0.vlm"},
	{0x4665F3B15C449EE8ULL, "vxht_brothel_n_4bg_0_0.vlm"},
	{0x466621F9FFAFABC8ULL, "32090013.ogg"},
	{0x4666971442257EF9ULL, "18105556.ogg"},
	{0x46670C996C43E3DEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2_0.pwk"},
	{0x4667760F93F9E24AULL, "753677957.ogg"},
	{0x46677B8447994CF3ULL, "6115583_m.fxe"},
	{0x4667C3E97D731766ULL, "307468512.ogg"},
	{0x4667FB222E867DE0ULL, "fen210_orana6109697.cl"},
	{0x4667FF4F61AB1E48ULL, "zz_followers_banter6159510.cl"},
	{0x46680EF27F236643ULL, "876210649.ogg"},
	{0x46682B04E8BC51CAULL, "ht_0_hightown_n_45_0_ll.mmh"},
	{0x46683BF65536CEA8ULL, "mag360_samson6044909.cl"},
	{0x4668F53CB11BE28AULL, "gen00fl_aveline6106921.cl"},
	{0x4669B374E42D44D1ULL, "6036564_f.fxe"},
	{0x4669DF4656775E74ULL, "qun211cr_henchman.cnv"},
	{0x466A192DF460E921ULL, "mag101_grace6034370.cl"},
	{0x466A8FCFA9A1F87EULL, "vxlt_darktown_5pq_0_0.vlm"},
	{0x466AD61B651AA166ULL, "vxlt_gallowstemplar_n_3r2_0_0.vlm"},
	{0x466B00C4BFB616ABULL, "6110566_m.fxe"},
	{0x466B29F6D694185BULL, "art\\characters\\weapons\\longswords\\w_lsw_007a.phy"},
	{0x466B2F08F9D8983AULL, "vxdr_primevaltomb_9cn_0_0.vlm"},
	{0x466B9D20D236586AULL, "hlac_ht_0_hightown_-131069_0x.dds"},
	{0x466BD1D40F1B45BDULL, "6119398_m.fxe"},
	{0x466BEFD7D789B468ULL, "mer120_hubert6059385.cl"},
	{0x466CC146567960B4ULL, "qun211cr_henchman.bnk"},
	{0x466D7A9DB931FD95ULL, "dr_slavers_mmap.dds"},
	{0x466D8C2DB68C514AULL, "vxlt_0_lowtown_e86_0_0.vlm"},
	{0x466E6DA2E9CE3DBAULL, "dre000st_varric_framing.stg"},
	{0x466ED52A1A95AAA8ULL, "vxlt_0_lowtown_dej_0_0.vlm"},
	{0x466F32DDD349000BULL, "6173654_f.fxe"},
	{0x466F833A69026529ULL, "1008619660.ogg"},
	{0x466F8AE8950B8791ULL, "vxdr_primevalentrance_1ao_0_0.vlm"},
	{0x466FC534EDF03D87ULL, "llac_w_woundedcoast_0_8.dds"},
	{0x466FCEECC0ACBFE3ULL, "vxlt_gallowscourtyard_6yj_0_0.vlm"},
	{0x466FD53B8B27E9B0ULL, "119208584.ogg"},
	{0x466FE3D984B7DF05ULL, "vxw_sundermount_37n_0_0.vlm"},
	{0x467065990125E04EULL, "6179993_m.fxe"},
	{0x4670C86558918A7EULL, "350856654.ogg"},
	{0x46714BFFB4E29F92ULL, "lowtown_argument6169369.cl"},
	{0x4671883FB0A9CF5BULL, "plt_and100pt_tranquility.nss"},
	{0x46718ED0CC44574CULL, "qun120_arishok6051421.cl"},
	{0x4671D46FAEFE414EULL, "hlac_lt_hangedman_-1_27x.dds"},
	{0x4672CBE08610710EULL, "vxht_0_hightown_b74_0_0.vlm"},
	{0x46731173E27756F5ULL, "743940064.ogg"},
	{0x46732A3E8C0F89BBULL, "548939188.ogg"},
	{0x4673547CBCF2B670ULL, "vxw_blightlands_cj_0_0.vlm"},
	{0x46735541657ABC8FULL, "6139707_m.fxe"},
	{0x4673E491904994F7ULL, "vxht_mansion_n_4w8_0_0.vlm"},
	{0x4673E8468E2D2339ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliff01_da.dds"},
	{0x467403EF8FC15594ULL, "vxlt_warehouse_14z_0_0.vlm"},
	{0x46740CDDBDC5FB3BULL, "2932111.ogg"},
	{0x46750D4ED820AA68ULL, "l_qun100ar_exterior.wbd"},
	{0x46756426954D22BCULL, "423580424.ogg"},
	{0x4675AF086213B3DBULL, "6150096_m.fxe"},
	{0x4675C58061BBF113ULL, "205543345.ogg"},
	{0x4675EABDD62F842CULL, "ave121_jeven6105647.cl"},
	{0x4675F0C6D362D7C4ULL, "vxw_bonepitmining_a2_lv_0_0.vlm"},
	{0x46760BA83F918124ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\emissive.mao"},
	{0x467657B0EF7430A9ULL, "art\\vfx\\combat\\vfx_str_spit1_act_c.mmh"},
	{0x467663BCF30CCB99ULL, "gen00fl_anders6114432.cl"},
	{0x46770A703D585EAEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupb_n.dds"},
	{0x46776DE4610B9AA8ULL, "vxht_0_hightown_a2_c51_0_0.vlm"},
	{0x46777BFDBD6678A3ULL, "vxht_bartrand_n_5oj_0_0.vlm"},
	{0x467859730FFEBF00ULL, "vxw_bonepitdestroyed_a3_1zq_0_0.vlm"},
	{0x467884DAA6C738F4ULL, "mag311_emile_de_launcet6043605.cl"},
	{0x467916D88445D160ULL, "959371697.ogg"},
	{0x4679283BBF339A45ULL, "6101258_m.fxe"},
	{0x46795BCA9971FFF1ULL, "6026948_m.fxe"},
	{0x467A94CE71BAB843ULL, "al_aveline_talents.gda"},
	{0x467ACB57CF7CDAC4ULL, "818066748.ogg"},
	{0x467AF7F7D1E090E6ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_stn_0d.dds"},
	{0x467B207B986E2024ULL, "w_bonepitdestroyed_a3_199_0.phy"},
	{0x467B5B24CFE8A6C7ULL, "6081380_f.fxe"},
	{0x467B62A4F59516B9ULL, "hlac_lt_ruinedhouse_n_-1_1z.dds"},
	{0x467B768F7F68D7ECULL, "255524276.ogg"},
	{0x467BC3CE1855193BULL, "dae000tr_preload_creature.ncs"},
	{0x467BF8A8BE5447C5ULL, "vxlt_0_lowtown_n_ewc_0_0.vlm"},
	{0x467BFD77DF431568ULL, "31578680.ogg"},
	{0x467BFE0AA88269CAULL, "vxlt_0_lowtown_e37_0_0.vlm"},
	{0x467C34854A7E5F18ULL, "vxht_0_hightown_a3_4c9_0_0.vlm"},
	{0x467C3DA4EE8E249AULL, "485372454.ogg"},
	{0x467C44E9D4BA3ADFULL, "zz_sneak02.arl"},
	{0x467C4F258BFE33DFULL, "face_rendertodepth_p3.vsh.fxo.dx11"},
	{0x467C50A7768D1566ULL, "vxca_caverns1_im4_0_1.vlm"},
	{0x467D4C35E97EDAC5ULL, "hlac_lt_0_lowtown_n_65535_7z.dds"},
	{0x467D5414E7309F85ULL, "hlac_lt_0_lowtown_1_10z.dds"},
	{0x467E3B1B6FB082AFULL, "hlac_w_woundedcoast_0_7y.dds"},
	{0x467E4341975B4B3DULL, "lgt250_way_dalish_conflict.bnk"},
	{0x467E55B676ADF785ULL, "vxw_sundermount_32m_0_0.vlm"},
	{0x467E605F6EB74362ULL, "172341064.ogg"},
	{0x467EEC44772500B0ULL, "890017119.ogg"},
	{0x467F8FE6E32B766EULL, "genip_node_a2.utp"},
	{0x467F96CBCD369AACULL, "ts_chantry_03.gda"},
	{0x467F97644A2E584CULL, "fen210_dying_magister6108592.cl"},
	{0x46801316A7ECFB3AULL, "6120807_m.fxe"},
	{0x46802B0FBD7F84FEULL, "6085763_f.fxe"},
	{0x46804963AB49BDB4ULL, "vxdr_primevaltomb_7sz_0_0.vlm"},
	{0x46804F0AA47D9B4DULL, "vxlt_undercityserialkiller_1ug_0_0.vlm"},
	{0x4680774CB2104B27ULL, "gen00fl_anders6121210.cl"},
	{0x4680811BCDFEC542ULL, "lt_0_lowtown_docks_a3_n_151_0.phy"},
	{0x4680973883E9BED7ULL, "dae011_corff_sta6152739.cl"},
	{0x4680A7B6CE6E953FULL, "lgt210_qun_blk_qunari6142389.cl"},
	{0x4680CD2D0745759BULL, "vxw_bonepit_1w9_0_0.vlm"},
	{0x46812541975D4E45ULL, "lgt250_way_dalish_conflict.cnv"},
	{0x4681940383E1721BULL, "combat_st_r1_a2.gda"},
	{0x4681C659AEE8CEF0ULL, "vxw_blightlands_dk_0_0.vlm"},
	{0x4681D30354E5C6B2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic_1.mao"},
	{0x468200F504AC3180ULL, "vxlt_0_lowtown_a3_n_c_eh8_0_0.vlm"},
	{0x46823DAEB1A56E8FULL, "dre000_varric6208031.cl"},
	{0x46824E103BC43CCCULL, "one100dg_kelder6060538.cl"},
	{0x468264187F8C180CULL, "6126520_m.fxe"},
	{0x4683806972B76568ULL, "qun211_party_comment6072170.cl"},
	{0x46838198A97DC11CULL, "vxht_keep_4mr_0_0.vlm"},
	{0x468409E5F1B6362CULL, "art\\characters\\weapons\\maces\\w_mce_mc07.phy"},
	{0x468418B7706D4B8EULL, "6179261_m.fxe"},
	{0x468462A3C558F044ULL, "vxw_bonepitmining_a2_kw_0_0.vlm"},
	{0x46846987BE2D5EEBULL, "vxw_woundedcoast_fh_0_0.vlm"},
	{0x4684AC9F1981D487ULL, "73103694.ogg"},
	{0x468552FA27864245ULL, "hlac_ht_0_hightown_a2_-65534_1y.dds"},
	{0x46856F95A1028114ULL, "zz_vault_debug6203273.cl"},
	{0x468599263C8595F0ULL, "231806039.ogg"},
	{0x4685A5EA0C3E316EULL, "w_encounter1_10_0.msh"},
	{0x4685B3EA0C3E48BFULL, "w_encounter1_10_0.mao"},
	{0x4685BFEA0C3E5D5CULL, "w_encounter1_10_0.mmh"},
	{0x4685DECE032413C0ULL, "vxlt_0_lowtown_a3_n_c_bxg_0_0.vlm"},
	{0x4685E6AEDE1533A1ULL, "546707525.ogg"},
	{0x46860FF1E907805BULL, "lt_0_lowtown_docks_a3_n_40_0.rml"},
	{0x46875420C53B3ACCULL, "qun110_petrice6059146.cl"},
	{0x4688781077B3C06DULL, "6162640_m.fxe"},
	{0x46887B60D7040A50ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_04tree_1.mao"},
	{0x468887167DA58640ULL, "vxdr_primevaltomb_7m0_0_0.vlm"},
	{0x4688A1A7561167B1ULL, "t3_dog_ow3.tnt"},
	{0x4689091DB80CD733ULL, "lfx_glow01.dds"},
	{0x46891496438F9DB7ULL, "vxca_caverns2prc_aez_0_0.vlm"},
	{0x46892779E152DDCDULL, "382845205.ogg"},
	{0x46893DB08F133FBCULL, "637991643.ogg"},
	{0x46896C19140086D3ULL, "vxw_encounter2_camp_lg_0_0.vlm"},
	{0x468A6A85B04A6045ULL, "vxlt_0_lowtown_n_epb_0_0.vlm"},
	{0x468A81B8555F82D2ULL, "21101403.ogg"},
	{0x468B0031366C2420ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_maga_0t.dds"},
	{0x468B8A7D9F3A1C2BULL, "204380721.ogg"},
	{0x468B98C134D01BA6ULL, "mag320_orsino6071250.cl"},
	{0x468BACA3AFE307F3ULL, "239434190.ogg"},
	{0x468BB4492CCA6D97ULL, "6002513_m.fxe"},
	{0x468C86A0189F55DBULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he6a.msh"},
	{0x468CA4A0189F88C5ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he6a.mmh"},
	{0x468CC929A3FA34B3ULL, "ch1alph_vl.psh.fxo"},
	{0x468CDABE893DA74AULL, "6109438_f.fxe"},
	{0x468CE0CF117E4A1FULL, "plt_cod_aow_locked_items.plo"},
	{0x468DA2108097151EULL, "vxlt_0_lowtown_n_gcf_0_0.vlm"},
	{0x468DCB3675E0012DULL, "gen00fl_sebastian6188645.cl"},
	{0x468E194F1F1A03AEULL, "6044600_m.fxe"},
	{0x468E8865632AACE3ULL, "pro000cs_utter_bullshit.cut"},
	{0x468EFC3D162D3DF5ULL, "gen00fl_aveline6120810.cl"},
	{0x468F3F09F93B8E1BULL, "vxw_bonepit_1x8_0_0.vlm"},
	{0x468F71D03F8DA8A6ULL, "hlac_lt_0_lowtown_docks_n_65534_3y.dds"},
	{0x46903836A0DEE770ULL, "vxw_blightlands_ih_0_0.vlm"},
	{0x469046AB4C88C732ULL, "6177304_m.fxe"},
	{0x469072D1F6A24A00ULL, "vxlt_0_lowtown_a3_n_c_es9_0_0.vlm"},
	{0x4690DAFF5E29659DULL, "vxdr_0_deeproads_f09_0_0.vlm"},
	{0x4690F0DD0808658BULL, "gen00fl_isabela_init6077385.cl"},
	{0x4691415768308C01ULL, "hlac_lt_0_lowtown_n_131070_4y.dds"},
	{0x4691754C6D61B1D0ULL, "vxlt_0_lowtown_docks_b6u_0_1.vlm"},
	{0x4692DB64B023776BULL, "vxw_woundedcoast_ck_0_0.vlm"},
	{0x4692ECCED5552552ULL, "vxw_encounter2_camp_lb_0_0.vlm"},
	{0x4692EF16DD6E2BF0ULL, "vxlt_0_lowtown_a3_n_c_by8_0_0.vlm"},
	{0x4693525F497C7204ULL, "mnd_imp_custic.mao"},
	{0x469409EB15598575ULL, "vxw_bonepit_1j0_0_0.vlm"},
	{0x46940C795C868CF5ULL, "hlac_lt_0_lowtown_n_-65537_6z.dds"},
	{0x4695A742F60175FEULL, "alphadot01.mao"},
	{0x4696E05A5AEF7DA7ULL, "ht_0_hightownexplosion_n_25_0.rml"},
	{0x4696E9ED69A9D8EDULL, "6167540_m.fxe"},
	{0x469710F6D6B928FAULL, "art\\characters\\weapons\\longswords\\w_lsw_007a.msh"},
	{0x46971EF6D6B94033ULL, "art\\characters\\weapons\\longswords\\w_lsw_007a.mao"},
	{0x469722F6D6B94740ULL, "art\\characters\\weapons\\longswords\\w_lsw_007a.mmh"},
	{0x469747FF1753458AULL, "vxht_0_hightown_bqf_0_0.vlm"},
	{0x4697766F8729ABF8ULL, "776479732.ogg"},
	{0x469795709D304883ULL, "vxlt_warehouse_8p_0_0.vlm"},
	{0x4697994701100C4CULL, "w_bonepit_186_0.phy"},
	{0x4697A9C6C33753ADULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestmetal.mmh"},
	{0x4697ADC6C3375AFEULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestmetal.mao"},
	{0x4697BBC6C3377233ULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestmetal.msh"},
	{0x4697DDF605F69F53ULL, "vxw_encounter2_camp_gf_0_0.vlm"},
	{0x46988F7D25255511ULL, "llac_lt_0_lowtown_a2_n_-65536_3.dds"},
	{0x4698A6B25282153AULL, "vxw_bonepitdestroyed_a3_28f_0_0.vlm"},
	{0x4698E8A96DFA8874ULL, "vxw_bonepit_17b_0_0.vlm"},
	{0x4699874591F955EAULL, "demo_rogue_bow.utc"},
	{0x4699E28BEC7DFBD1ULL, "dae000tr_location_alienage.utt"},
	{0x469A073A761E6214ULL, "dae000_bartrand6027604.cl"},
	{0x469A26261EC08617ULL, "6005413_m.fxe"},
	{0x469A269218537170ULL, "qcr220_meredith6054743.cl"},
	{0x469A86561A289B7DULL, "gen00fl_fenris6126431.cl"},
	{0x469AC1AB0D100014ULL, "mag291_demon_marethari6029031.cl"},
	{0x469AE8F97C0BFED3ULL, "fx_crt_cycle.mao"},
	{0x469B3DA26B4AACAFULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue2_d.dds"},
	{0x469BB95995CB4399ULL, "llac_ht_0_hightown_n_0_3.dds"},
	{0x469C04E1DF4AD705ULL, "6093820_f.fxe"},
	{0x469C831AD0EAAF30ULL, "vxlt_0_lowtown_a3_n_c_erz_0_0.vlm"},
	{0x469CB33659A9B0A5ULL, "453803116.ogg"},
	{0x469CCD211BE083A7ULL, "77104938.ogg"},
	{0x469D86255E97BCF1ULL, "730122871.ogg"},
	{0x469DB33E898A3720ULL, "isa121cr_hayder_henchman.utc"},
	{0x469E06ED075D66F4ULL, "234390591.ogg"},
	{0x469E7F01C3239ED7ULL, "gen00fl_isabela6075276.cl"},
	{0x469E9CC1363B722DULL, "vxca_caverns1_mines_ims_0_0.vlm"},
	{0x469EAA1392D4FFF0ULL, "vxw_blightlands_ri_0_0.vlm"},
	{0x469ECFF969CE33A1ULL, "vxdr_primevaltomb_7o7_0_0.vlm"},
	{0x469ED246E7EC1C47ULL, "mag311_nyssa6064989.cl"},
	{0x469F3B11971FD989ULL, "follower_banter6161163.cl"},
	{0x469F47D26378490CULL, "6128720_m.fxe"},
	{0x469F54B0EF9730DEULL, "art\\vfx\\combat\\vfx_str_spit1_act_c.anb"},
	{0x469F8E13E28AD6A5ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga105a_0n.dds"},
	{0x469FCBCC031F8B4FULL, "plt_lgt000pt_civ_witch.plo"},
	{0x469FF3DD759BD70BULL, "vxw_bonepit_by_0_0.vlm"},
	{0x46A01CAD694BB790ULL, "vxdr_0_deeproads_9p7_0_0.vlm"},
	{0x46A052664E62E73BULL, "w_blightlands_122_0.phy"},
	{0x46A08EA30F29DF19ULL, "one300_zevran6089763.cl"},
	{0x46A0E7E20ED67F40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\blendtree\\trooper_movement.anb"},
	{0x46A15EABC74B3DD2ULL, "vxw_encounter2_camp_gc_0_0.vlm"},
	{0x46A16216011BDADCULL, "gen_im_arm_boo_med_start.uti"},
	{0x46A1933FB0D305F4ULL, "plt_and100pt_tranquility.plo"},
	{0x46A1B968B52AA285ULL, "gen00fl_isabela6079065.cl"},
	{0x46A21F78A377BB0AULL, "one100_solivitus6048121.cl"},
	{0x46A2B3E14C0B22A1ULL, "hightown_male_elfservnt_mob6170084.cl"},
	{0x46A39FD12488C826ULL, "def_dirt_d.dds"},
	{0x46A3BFEA0C575AAFULL, "w_encounter1_10_0.rml"},
	{0x46A41EAD66923503ULL, "vxdr_primevalentrance_15x_0_1.vlm"},
	{0x46A432CAA9A71157ULL, "gen00fl_merrill6144958.cl"},
	{0x46A50F2BC73821BCULL, "lt_backalley_g_1_0.mao"},
	{0x46A51B2BC738365FULL, "lt_backalley_g_1_0.mmh"},
	{0x46A5212BC7384001ULL, "lt_backalley_g_1_0.msh"},
	{0x46A551F7395251B3ULL, "390367342.ogg"},
	{0x46A55C1BCE1DA2F6ULL, "lt_0_lowtown_docks_a3_n_151_0.mao"},
	{0x46A5681BCE1DB695ULL, "lt_0_lowtown_docks_a3_n_151_0.mmh"},
	{0x46A56A1BCE1DBA0BULL, "lt_0_lowtown_docks_a3_n_151_0.msh"},
	{0x46A5B9DC09495E0AULL, "vxht_0_hightown_bzg_0_0.vlm"},
	{0x46A5DC0602011124ULL, "194611635.ogg"},
	{0x46A64FD2F7ECB7D3ULL, "vxw_encounter2_camp_bi_0_0.vlm"},
	{0x46A6B864A5D4991DULL, "w_encounter3_137_0.mao"},
	{0x46A6BC64A5D4A056ULL, "w_encounter3_137_0.mmh"},
	{0x46A6C664A5D4B154ULL, "w_encounter3_137_0.msh"},
	{0x46A70BB312E22CA7ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar1_.gad"},
	{0x46A7188F44782DBAULL, "vxw_bonepitdestroyed_a3_23g_0_0.vlm"},
	{0x46A73D3EE918DAD3ULL, "lt_undercityserialkiller.arl"},
	{0x46A73F58005E14B5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_basic.msh"},
	{0x46A74158005E182BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_basic.mmh"},
	{0x46A746A168C8A171ULL, "856529222.ogg"},
	{0x46A74D58005E2CC8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_basic.mao"},
	{0x46A77E5C8D3B8E48ULL, "vxlt_0_lowtown_n_ev9_0_0.vlm"},
	{0x46A79485532D51E2ULL, "183995005.ogg"},
	{0x46A7CEC261E13E01ULL, "vxca_caverns1_hio_0_0.vlm"},
	{0x46A82748C997B2EDULL, "743794364.ogg"},
	{0x46A84E97BFEA0A34ULL, "304859801.ogg"},
	{0x46A8793F35D2570DULL, "6153442_m.fxe"},
	{0x46A8A078A38A7ACAULL, "vxca_caverns1_i9m_0_0.vlm"},
	{0x46A8A2C6C345BB6AULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestmetal.phy"},
	{0x46A8BB6147B33CB1ULL, "156237823.ogg"},
	{0x46A942A832A26FACULL, "6108888_f.fxe"},
	{0x46A96CB42669ACA5ULL, "6162381_m.fxe"},
	{0x46A996E9D62E94AAULL, "6034746_m.fxe"},
	{0x46AA0F3742D4CB68ULL, "231534231.ogg"},
	{0x46AA24DFA86C5EC3ULL, "character_character_p260.psh.fxo.dx11.assembly"},
	{0x46AA3D288524A459ULL, "fx_1014_diff.dds"},
	{0x46AA9EEA0C5D4127ULL, "w_encounter1_10_0.phy"},
	{0x46AAB5BE47CAA049ULL, "878725369.ogg"},
	{0x46AAB6C003571278ULL, "vxlt_darktown_2wm_0_0.vlm"},
	{0x46AAC172B24767B4ULL, "275919366.ogg"},
	{0x46AAFD00B77B6DBAULL, "mcr370_meredith6083585.cl"},
	{0x46AB04E16D863F8FULL, "art\\characters\\playercharacter\\humanmale\\hlf_masf.phy"},
	{0x46AB232EB3AD7D26ULL, "vxht_mansion2_22c_0_0.vlm"},
	{0x46AB47EE1FD3C797ULL, "hf_dalesdottir.mor"},
	{0x46ABBA1542CB3431ULL, "vxca_cavern1_haunted_fmp_0_0.vlm"},
	{0x46AC2A149E8EDB8DULL, "488930409.ogg"},
	{0x46AC40C6BBCA1DA8ULL, "vxw_bonepit_1pg_0_0.vlm"},
	{0x46AC4D3BCAE6CFF2ULL, "102423305.ogg"},
	{0x46AC6783BDDCFAE1ULL, "vxlt_0_lowtown_n_es3_0_0.vlm"},
	{0x46AC8B0830473E3EULL, "vxlt_gallowscourtyard_n_6k6_0_0.vlm"},
	{0x46ACEA8987469B5CULL, "hightown_fem_courtesan_mob6168589.cl"},
	{0x46AD0E9E28318AADULL, "vxca_caverns1_mines_ijr_0_0.vlm"},
	{0x46AD107FC86EBE96ULL, "ser221_gascard6122457.cl"},
	{0x46AD2271C23511F1ULL, "mag210_arianni6057345.cl"},
	{0x46AD9592104927B6ULL, "art\\characters\\creatures\\corpse\\animation\\blendtree\\corpse_movement.anb"},
	{0x46ADB36A6BF13537ULL, "vxlt_0_lowtown_a2_n_c_hdc_0_0.vlm"},
	{0x46ADFD44AE810AEFULL, "vxw_bonepitdestroyed_a3_7z_0_0.vlm"},
	{0x46AE9C1EF1319CE7ULL, "58393991.ogg"},
	{0x46AFBF1E940FA86BULL, "vxw_woundedcoast_um_0_0.vlm"},
	{0x46AFD088B9415652ULL, "vxw_encounter2_camp_bl_0_0.vlm"},
	{0x46B037A5E0E4D198ULL, "6163996_m.fxe"},
	{0x46B06322EAC03860ULL, "hlac_lt_gallowstemplar_2393_1x.dds"},
	{0x46B0737155320B1AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebb.mmh"},
	{0x46B0777155321251ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebb.mao"},
	{0x46B07BF697F7B422ULL, "gen_im_trs_tmp_med_cht_r2.uti"},
	{0x46B0857155322A60ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebb.msh"},
	{0x46B0AA7930FC1411ULL, "6127107_f.fxe"},
	{0x46B0B9EE536D75DFULL, "6172795_m.fxe"},
	{0x46B11B9963EC09C9ULL, "6084683_m.fxe"},
	{0x46B1575C71C37EC4ULL, "6088686_m.fxe"},
	{0x46B162B0025E0183ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_y_neg.ani"},
	{0x46B1663C49F7B317ULL, "w_bonepitdestroyed_a3_78_0.mmh"},
	{0x46B16A3C49F7BA24ULL, "w_bonepitdestroyed_a3_78_0.mao"},
	{0x46B17C3C49F7D889ULL, "w_bonepitdestroyed_a3_78_0.msh"},
	{0x46B242D47B3B7889ULL, "hlac_x_narrativeroom_0_0z.dds"},
	{0x46B2632AF0C566E9ULL, "vxdr_0_deeproads_9c7_0_0.vlm"},
	{0x46B30032316B09BDULL, "character_character_p253.psh.fxo.dx11.assembly"},
	{0x46B322AE58E315A7ULL, "gen00fl_aveline6139638.cl"},
	{0x46B3337D782108EEULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage1_.ani"},
	{0x46B37059A9CD5A7CULL, "560815900.ogg"},
	{0x46B398BCB1EE7319ULL, "622352711.ogg"},
	{0x46B39CAAF10F86FFULL, "zz_dre_debug6204368.cl"},
	{0x46B3BC11E27E4F38ULL, "genip_door_highin_keep.utp"},
	{0x46B3D94656B58C29ULL, "qun211cr_henchman.utc"},
	{0x46B429F39DD304F0ULL, "1027286981.ogg"},
	{0x46B483CA2A71E16CULL, "951900192.ogg"},
	{0x46B48FB039045881ULL, "114524183.ogg"},
	{0x46B4D1F30CC2D521ULL, "6108517_f.fxe"},
	{0x46B524B00CDB7B06ULL, "lt_0_lowtown_a3_n_c_mmap.dds"},
	{0x46B55D7294E59323ULL, "gen_im_trs_tmp_med_boo_r2.uti"},
	{0x46B5BA081EFC6B38ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef01.phy"},
	{0x46B5C16BC49D47A6ULL, "fen110_fenris6118140.cl"},
	{0x46B5E236665F9575ULL, "vxht_0_hightown_a2_cat_0_0.vlm"},
	{0x46B5F0397F31A6C8ULL, "vxlt_0_lowtown_n_es6_0_0.vlm"},
	{0x46B5F7423C87444DULL, "art\\effects\\cloudshadows\\w_sundermount\\cloud1.dds"},
	{0x46B61CD7D503CF17ULL, "vxht_0_hightown_a3_c13_0_0.vlm"},
	{0x46B65F6ECA155497ULL, "lt_0_lowtown_docks_a3_125_0.mmh"},
	{0x46B6636ECA155BA4ULL, "lt_0_lowtown_docks_a3_125_0.mao"},
	{0x46B6756ECA157A09ULL, "lt_0_lowtown_docks_a3_125_0.msh"},
	{0x46B6BCE5D9E122FCULL, "vxca_caverns1_mines_coj_0_0.vlm"},
	{0x46B6DCD77FA3070DULL, "6116539_m.fxe"},
	{0x46B737050522F38EULL, "mag210_arianni6024553.cl"},
	{0x46B7B3CE6210A508ULL, "prm000im_blood_helm.uti"},
	{0x46B806C49D3F271FULL, "hlac_lt_undercityserialkiller_-1_20y.dds"},
	{0x46B87434A8578E8EULL, "hlac_ca_caverns1_mines_24423_5z.dds"},
	{0x46B87846B72EB304ULL, "mag131_keran6086253.cl"},
	{0x46B8D948947485F4ULL, "zz_duel_arishok.ncs"},
	{0x46B8F22555DD2DD6ULL, "6199322_m.fxe"},
	{0x46B8FFB9EB80FE3BULL, "vxlt_shop_dk_0_0.vlm"},
	{0x46B91F8B6A492250ULL, "hlac_ca_caverns2_5802_33y.dds"},
	{0x46B966D4B4D6E030ULL, "vxlt_0_lowtown_a3_n_c_exx_0_0.vlm"},
	{0x46B9805952C42C44ULL, "mcr310_argument6182461.cl"},
	{0x46B9F0420528B748ULL, "vxht_0_hightown_c9a_0_0.vlm"},
	{0x46B9FD0AA9EFE3EEULL, "128208836.ogg"},
	{0x46BA271EB91B36D5ULL, "vxca_cavern1_haunted_jgn_0_0.vlm"},
	{0x46BA8720A8FA6A9AULL, "6108580_m.fxe"},
	{0x46BAB2A3ADC03628ULL, "vxw_bonepit_1ud_0_0.vlm"},
	{0x46BAC39C04A73F60ULL, "6154073_m.fxe"},
	{0x46BB12846FA63D96ULL, "sib131_slaver_guard6165287.cl"},
	{0x46BB78A4EC4AB96AULL, "6019887_f.fxe"},
	{0x46BB7D1F1E2B9A50ULL, "lgt110_arl_bann_teagan6115580.cl"},
	{0x46BB807B1A27A32DULL, "vxca_caverns1_mines_igq_0_0.vlm"},
	{0x46BB9C9464E86658ULL, "vxlt_gallowscourtyard_6xv_0_0.vlm"},
	{0x46BBB3B34DBA64A1ULL, "vxdr_primevaltomb_7e5_0_0.vlm"},
	{0x46BBF8B312F468EDULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar1_.ani"},
	{0x46BC2E9C0DF0EF73ULL, "vxw_blightlands_14i_0_0.vlm"},
	{0x46BC334ED85CED2AULL, "l_qun100ar_exterior.bnk"},
	{0x46BC82A9F6564904ULL, "6167307_m.fxe"},
	{0x46BCA2D5C1832EB7ULL, "6149263_m.fxe"},
	{0x46BCBD5872DFFB75ULL, "vxca_cavern1_haunted_cmm_0_0.vlm"},
	{0x46BCC0ECC5744687ULL, "vxca_caverns2_7mm_0_0.vlm"},
	{0x46BCDBD737B0F79AULL, "6167143_m.fxe"},
	{0x46BCE06283227882ULL, "lgt000ip_ultra_dwf_book.ncs"},
	{0x46BCE4B319922412ULL, "qun100_mage_party_comment6019113.cl"},
	{0x46BD9ED8CAF4BD81ULL, "vxlt_gallowscourtyard_7eq_0_0.vlm"},
	{0x46BE28C1F3985438ULL, "vxlt_0_lowtown_docks_a3_n_d3u_0_0.vlm"},
	{0x46BE2CB0E26AC73DULL, "557485476.ogg"},
	{0x46BE30FB8605C0EBULL, "vxw_woundedcoast_rl_0_0.vlm"},
	{0x46BE4DAFA29C43E1ULL, "1055369598.ogg"},
	{0x46BE61A4842C13D3ULL, "37784998.ogg"},
	{0x46BEA9611641AF46ULL, "306258354.ogg"},
	{0x46BEC45149EFE9DFULL, "452135823.ogg"},
	{0x46BEC6AB6AE0CB03ULL, "art\\vfx\\combat\\vfx_w_wns_flurry_c.impa.ani"},
	{0x46BEEBF9622C087BULL, "6110496_m.fxe"},
	{0x46BF604838EA3431ULL, "301942370.ogg"},
	{0x46BFA623343BAFE4ULL, "hlac_lt_gallowstemplar_n_2383_3x.dds"},
	{0x46BFADAFB8A16250ULL, "vxw_bonepitmining_a2_1w8_0_0.vlm"},
	{0x46BFE993D08AB24FULL, "gen00fl_aveline6132734.cl"},
	{0x46C00E5950DE8E12ULL, "llac_lt_0_lowtown_docks_a3_n_0_6.dds"},
	{0x46C0645A455B1B35ULL, "6041961_m.fxe"},
	{0x46C0D507E2BB7F69ULL, "vxdr_0_deeproads_9d6_0_0.vlm"},
	{0x46C0EFCBCC9F6193ULL, "8056729.ogg"},
	{0x46C17B1BFF1B96D2ULL, "994614270.ogg"},
	{0x46C23FF32FF1E3ABULL, "30742883.ogg"},
	{0x46C37CEB0ABBC17AULL, "meleering5.gda"},
	{0x46C3AE15FB421889ULL, "gen_im_trs_krk_junk2_04.uti"},
	{0x46C3D1FEE1B84EDCULL, "vxlt_0_lowtown_docks_bf8_0_0.vlm"},
	{0x46C3DE78CE0BA315ULL, "hlac_lt_0_lowtown_docks_n_196608_0x.dds"},
	{0x46C44B99E363F7ADULL, "hlac_lt_darktown_-1_30z.dds"},
	{0x46C4D834DD515FA0ULL, "6153164_m.fxe"},
	{0x46C4E0E5F1ED87F7ULL, "art\\characters\\weapons\\maces\\w_mce_mc07.mmh"},
	{0x46C4F6E5F1EDAD69ULL, "art\\characters\\weapons\\maces\\w_mce_mc07.msh"},
	{0x46C56586F5280196ULL, "hlac_lt_0_lowtown_-131072_4z.dds"},
	{0x46C579B0026ED26DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_y_neg.gad"},
	{0x46C594B1398AFF3BULL, "hlac_ca_caverns1_mines_24421_6x.dds"},
	{0x46C5C3F0D874608DULL, "gen00fl_varric6018419.cl"},
	{0x46C614941B47F135ULL, "vxht_keep_62t_0_0.vlm"},
	{0x46C665DE32031D24ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mc.msh"},
	{0x46C677DE32033B8DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mc.mao"},
	{0x46C67BDE320342C6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mc.mmh"},
	{0x46C703C81E1DA4F1ULL, "6204164_m.fxe"},
	{0x46C70C7D78316F28ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage1_.gad"},
	{0x46C799FC82808276ULL, "vxht_keep_5tz_0_0.vlm"},
	{0x46C7A254A705D2A6ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose2.anb"},
	{0x46C7B8BF936B02E8ULL, "vxlt_0_lowtown_docks_c4w_0_0.vlm"},
	{0x46C89EC76F276D3DULL, "llac_lt_0_lowtown_a2_n_1_2.dds"},
	{0x46C8EF850C8D209EULL, "6087803_m.fxe"},
	{0x46C98CFBE0B5D552ULL, "11973191.ogg"},
	{0x46C9DBA8D6A0B6B4ULL, "zz_isa_debug6175238.cl"},
	{0x46CA70DD05A0A039ULL, "6162019_m.fxe"},
	{0x46CAAA25A416B1DAULL, "gen00fl_bethany6101514.cl"},
	{0x46CBDEAB6AEB7EFAULL, "art\\vfx\\combat\\vfx_w_wns_flurry_c.impa.evt"},
	{0x46CBE5F2E62E89BEULL, "vxht_chantry_28e_0_0.vlm"},
	{0x46CC39664E87F7DAULL, "w_blightlands_122_0.msh"},
	{0x46CC47664E880F13ULL, "w_blightlands_122_0.mao"},
	{0x46CC4B664E881620ULL, "w_blightlands_122_0.mmh"},
	{0x46CC8F054F6B57E3ULL, "196371011.ogg"},
	{0x46CCB4B6001D66E3ULL, "vxw_encounter3_f6_0_0.vlm"},
	{0x46CD1C2BC759899CULL, "lt_backalley_g_1_0.anb"},
	{0x46CD7130EBBF0E83ULL, "896412009.ogg"},
	{0x46CDB3BF5923273BULL, "static_static_p2.vsh.fxo"},
	{0x46CDC10919B6DD27ULL, "zz_vault_debug6178936.cl"},
	{0x46CE0BBC65CED769ULL, "vxlt_0_lowtown_n_fe6_0_0.vlm"},
	{0x46CE20DA7917BD90ULL, "dre000_dougal6129725.cl"},
	{0x46CE2B933BBD8346ULL, "vxht_mansion2_1ws_0_0.vlm"},
	{0x46CE3B1655AFAFC4ULL, "6086886_m.fxe"},
	{0x46CE8B20DBC8EFCCULL, "qun211ip_barrel_latch.ncs"},
	{0x46CEC0AB55DD774DULL, "art\\characters\\playercharacter\\humanmale\\arm_ric.mmh"},
	{0x46CEC4AB55DD7D9EULL, "art\\characters\\playercharacter\\humanmale\\arm_ric.mao"},
	{0x46CED2AB55DD95D3ULL, "art\\characters\\playercharacter\\humanmale\\arm_ric.msh"},
	{0x46CF01BC441E6030ULL, "152386397.ogg"},
	{0x46CF0AD9F775810DULL, "zz_followers_banter6158406.cl"},
	{0x46D015ADA5F24DA4ULL, "6028922_f.fxe"},
	{0x46D04360389BE120ULL, "vxlt_gallowsprison_n_5we_0_0.vlm"},
	{0x46D0C458F026B180ULL, "725261098.ogg"},
	{0x46D0E61ACCF3C9E8ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyo_0t.dds"},
	{0x46D19554AAE56055ULL, "one100dg_kelder6055605.cl"},
	{0x46D1CDD451775913ULL, "6049624_m.fxe"},
	{0x46D1F75B289AE871ULL, "vxlt_undercitydungeon_242_0_0.vlm"},
	{0x46D214D9DDEBF4A6ULL, "lowtown_argument6169408.cl"},
	{0x46D2578AAE704EB2ULL, "6051695_m.fxe"},
	{0x46D2821C28AA69EFULL, "ht_0_hightownexplosion_n_25_0_ll.mmh"},
	{0x46D2D3F3631DD7C8ULL, "vxlt_0_lowtown_n_ei4_0_0.vlm"},
	{0x46D2DDAB6AF19BEDULL, "art\\vfx\\combat\\vfx_w_wns_flurry_c.impa.gad"},
	{0x46D2F79AE101F068ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hat_medc_0s.dds"},
	{0x46D31447B95E72B8ULL, "ser221_gascard6149258.cl"},
	{0x46D348E0C94A9379ULL, "624130984.ogg"},
	{0x46D3A7F71E97A022ULL, "art\\characters\\playercharacter\\humanfemale\\boo_nuda.msh"},
	{0x46D3A9F71E97A3D8ULL, "art\\characters\\playercharacter\\humanfemale\\boo_nuda.mmh"},
	{0x46D3B5F71E97B87BULL, "art\\characters\\playercharacter\\humanfemale\\boo_nuda.mao"},
	{0x46D3EC7C3AD28E97ULL, "vxlt_hangedman_n_4q_0_0.vlm"},
	{0x46D41383257EF4D3ULL, "fen210_hadriana6110381.cl"},
	{0x46D454F6699FF3A3ULL, "mer200ar_exterior.are"},
	{0x46D456C37C42D76FULL, "667951650.ogg"},
	{0x46D473EDF40A2B62ULL, "6053510_m.fxe"},
	{0x46D4772BDCA2F453ULL, "120043485.ogg"},
	{0x46D47911F227BFA8ULL, "hlac_w_sundermount_-327675_0z.dds"},
	{0x46D47DA018DCFC72ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he6a.phy"},
	{0x46D4B7C1319DD032ULL, "hlac_ht_0_hightown_a3_2_0x.dds"},
	{0x46D63CF8E2F92A17ULL, "gen00fl_aveline6114002.cl"},
	{0x46D71C3C7DD57901ULL, "866665590.ogg"},
	{0x46D766CA104DD0B3ULL, "6107214_m.fxe"},
	{0x46D767170D35369CULL, "451353177.ogg"},
	{0x46D770470145AA17ULL, "w_bonepit_186_0.mmh"},
	{0x46D774470145B124ULL, "w_bonepit_186_0.mao"},
	{0x46D786470145CF89ULL, "w_bonepit_186_0.msh"},
	{0x46D78927E98820CFULL, "vxca_cavern1_haunted_j4i_0_0.vlm"},
	{0x46D7B063FA322EFBULL, "art\\vfx\\models\\vfx_rockarmor_lback.mmh"},
	{0x46D7C100FAAAD5D7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblelarge.phy"},
	{0x46D7C197444D1B70ULL, "vxlt_undercityencounter_uh_0_1.vlm"},
	{0x46D7CE63FA3261E5ULL, "art\\vfx\\models\\vfx_rockarmor_lback.msh"},
	{0x46D81E1DD9E0C101ULL, "227284673.ogg"},
	{0x46D89F04A81BB89EULL, "hlac_w_encounter2_camp_327680_0z.dds"},
	{0x46D8A76902D0543DULL, "352099269.ogg"},
	{0x46D8E059E4DE3031ULL, "617700907.ogg"},
	{0x46D912A8996C5FFEULL, "dae000_arishok6078414.cl"},
	{0x46D94CB1C4EABFDBULL, "6115733_f.fxe"},
	{0x46D9685C92C78690ULL, "gen00fl_fenris6126973.cl"},
	{0x46D99BB03E2FB8ABULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.evt"},
	{0x46D9E192044B5773ULL, "mcr310_argument6074484.cl"},
	{0x46D9EBC160EF8532ULL, "hlac_w_encounter3_-196605_0y.dds"},
	{0x46DA42CB1E4B2806ULL, "hlac_lt_gallowstemplar_2393_0y.dds"},
	{0x46DA4E51CA759C35ULL, "wm_ico_tavern.dds"},
	{0x46DB2692F2137F63ULL, "vxw_encounter3_c9_0_0.vlm"},
	{0x46DB813775582710ULL, "6094687_m.fxe"},
	{0x46DBEFB218406D78ULL, "t3_dps_b03.tnt"},
	{0x46DC5B60EC4D6791ULL, "849827285.ogg"},
	{0x46DD03DE1D3716E3ULL, "hlac_lt_0_lowtown_n_65535_4y.dds"},
	{0x46DD1EA396CA2101ULL, "6218824_m.fxe"},
	{0x46DD2CEEB604BAE1ULL, "6189899_f.fxe"},
	{0x46DD67F1CC0BA700ULL, "941830689.ogg"},
	{0x46DD970569F61C70ULL, "658636976.ogg"},
	{0x46DDCA019188B761ULL, "one140cr_charade6137623.cl"},
	{0x46DDDFB6E0CEA847ULL, "dae000_porfiria6156620.cl"},
	{0x46DE1F4C30428DCAULL, "seb321_leliana6180910.cl"},
	{0x46DE583DC60EF425ULL, "399163831.ogg"},
	{0x46DE58A91D7FC137ULL, "vxlt_hangedman_n_20m_0_0.vlm"},
	{0x46DE647CD3C42BFAULL, "756400840.ogg"},
	{0x46DE85B4AB0F0D2EULL, "6177209_m.fxe"},
	{0x46DE9D2A956F66CAULL, "mag100_thrask6035905.cl"},
	{0x46DEAC4086A86AA2ULL, "combat_arishock_range.gda"},
	{0x46DEB0F71EA1B763ULL, "art\\characters\\playercharacter\\humanfemale\\boo_nuda.phy"},
	{0x46DF119F7A996780ULL, "vxlt_0_lowtown_docks_b5m_0_0.vlm"},
	{0x46DF124C19672695ULL, "481492177.ogg"},
	{0x46DF8352F49BCF87ULL, "mag121_idunna6084897.cl"},
	{0x46DF9554E55AF877ULL, "6121096_m.fxe"},
	{0x46DFACDABDDC743DULL, "23674520.ogg"},
	{0x46DFB9AB55EBDE0AULL, "art\\characters\\playercharacter\\humanmale\\arm_ric.phy"},
	{0x46E03147C6D9811EULL, "plt_zz_seb000pt_debug.plo"},
	{0x46E0B4B03E36026CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.gad"},
	{0x46E0B7583D88DE8EULL, "lgt110_orz_harr_carta6109058.cl"},
	{0x46E0BA5B95A7A90CULL, "vxlt_warehouse_vy_0_0.vlm"},
	{0x46E0F7DA5052F2F4ULL, "hlac_lt_0_lowtown_docks_a3_-3_5z.dds"},
	{0x46E1E3C0054A34B4ULL, "hlac_w_blightlands_131068_0z.dds"},
	{0x46E2A30E8AC82E1FULL, "394734207.ogg"},
	{0x46E2A90F56D463AEULL, "one300_zevran6075394.cl"},
	{0x46E38E0B45044ED5ULL, "ser270_emeric6113573.cl"},
	{0x46E3C2C872630C40ULL, "gen00fl_fenris6116530.cl"},
	{0x46E3D20F58B629D4ULL, "vxht_keep_n_5j5_0_0.vlm"},
	{0x46E3E933D212842DULL, "vxca_caverns1_mines_j76_0_0.vlm"},
	{0x46E41C1106459522ULL, "hlac_lt_0_lowtown_a2_n_0_6x.dds"},
	{0x46E44A2CC91BBA4FULL, "6127498_m.fxe"},
	{0x46E4843005B8AD90ULL, "6162568_m.fxe"},
	{0x46E50AA25B3E7E63ULL, "hlac_ht_0_hightown_a2_-65534_0x.dds"},
	{0x46E5C5306C71CA40ULL, "6040981_m.fxe"},
	{0x46E62495B2B84EEFULL, "mag360_thrask6043543.cl"},
	{0x46E629BF9173D2BEULL, "hlac_w_bonepit_131076_0z.dds"},
	{0x46E63C0E89325001ULL, "vxlt_0_lowtown_docks_bfi_0_0.vlm"},
	{0x46E64428F59E1EF4ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb105a_0s.dds"},
	{0x46E6A2C9C6C0B00EULL, "vxlt_0_lowtown_a2_h1x_0_0.vlm"},
	{0x46E73A4DEDC0A5ADULL, "333619758.ogg"},
	{0x46E781236D79D34DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_3windows.mmh"},
	{0x46E785236D79D99EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_3windows.mao"},
	{0x46E793236D79F1D3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_3windows.msh"},
	{0x46E7B3B03E3C1F6AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.ani"},
	{0x46E89EA347386C9CULL, "vxht_keep_40u_0_0.vlm"},
	{0x46E8A18BDD17545CULL, "161508140.ogg"},
	{0x46E8DC3B490871AAULL, "46314910.ogg"},
	{0x46E8F28A1A642896ULL, "453448576.ogg"},
	{0x46E94BB4FCFF2A0BULL, "mrl210_merrill6094194.cl"},
	{0x46E972BEED57C03BULL, "vxlt_gallowscourtyard_79w_0_0.vlm"},
	{0x46E9822F2910FD19ULL, "vxca_caverns1_io8_0_0.vlm"},
	{0x46E9EACC035DF834ULL, "plt_lgt000pt_civ_witch.nss"},
	{0x46EA590D032C699BULL, "vxca_caverns1_ee0_0_0.vlm"},
	{0x46EA748C7B4F0ED1ULL, "6180150_m.fxe"},
	{0x46EA75FF3915FDD3ULL, "181017612.ogg"},
	{0x46EAC300B50AE9E4ULL, "hla_ht_0_hightown_n_65535_0y.dds"},
	{0x46EB1BA6B7783C40ULL, "715554511.ogg"},
	{0x46EB1DF436924A62ULL, "mrl310_merrill6143980.cl"},
	{0x46EB91FBFD6B9265ULL, "235162291.ogg"},
	{0x46EC9F49D50C445BULL, "vxht_keep_63y_0_0.vlm"},
	{0x46ECDCAD65230662ULL, "537509207.ogg"},
	{0x46ECEE97A2AA2A9CULL, "645841862.ogg"},
	{0x46ED80BE8FE7B4A5ULL, "vxca_cave_14h_0_1.vlm"},
	{0x46EE02539FC30EE7ULL, "855926705.ogg"},
	{0x46EE891866AA07B3ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_aoe_p.anb"},
	{0x46EEB1F34EBAA1C6ULL, "6201274_m.fxe"},
	{0x46EEE38BECC63D24ULL, "dae000tr_location_alienage.ncs"},
	{0x46EF221D244DF57CULL, "6130105_m.fxe"},
	{0x46EF449180A4A38DULL, "follower_barks6169911.cl"},
	{0x46EF6CEB3988B3CDULL, "6135095_m.fxe"},
	{0x46EF808327886C79ULL, "lt_0_lowtown_docks_a3_204_0.msh"},
	{0x46EF8A8327887D67ULL, "lt_0_lowtown_docks_a3_204_0.mmh"},
	{0x46EF8E83278883B4ULL, "lt_0_lowtown_docks_a3_204_0.mao"},
	{0x46EFC39A5780DCA4ULL, "enemy_attack.ncs"},
	{0x46EFCB5344D11B6DULL, "343842911.ogg"},
	{0x46EFD1255265B040ULL, "mag110_arianni6016843.cl"},
	{0x46EFD5F81C692373ULL, "vxca_caverns1_dqa_0_0.vlm"},
	{0x46EFF4AAB2EF21FFULL, "vxlt_hangedman_n_16t_0_0.vlm"},
	{0x46F0397EBAC352BAULL, "llac_ht_mansion_n_3381_1.dds"},
	{0x46F04F20614855CDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basicnb.msh"},
	{0x46F05D2061486D00ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basicnb.mao"},
	{0x46F0612061487453ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basicnb.mmh"},
	{0x46F144B3EAF5E100ULL, "ser211cr_demon_ninette.utc"},
	{0x46F1BABE58AC2EE6ULL, "19505455.ogg"},
	{0x46F1E8A09FB5FAAEULL, "mag311_nyssa6041859.cl"},
	{0x46F22808C4D6B54AULL, "gen00fl_isabela6088248.cl"},
	{0x46F259FA09FEFD1CULL, "6069658_m.fxe"},
	{0x46F25B10C4089CADULL, "vxca_caverns1_mines_j25_0_0.vlm"},
	{0x46F2D7D37F8384DBULL, "6149905_f.fxe"},
	{0x46F2F60CF7AEC610ULL, "6167668_m.fxe"},
	{0x46F3BC130E74B850ULL, "vxlt_0_lowtown_a2_n_c_f5e_0_0.vlm"},
	{0x46F3EE649ABE4A2EULL, "vxlt_undercityencounter_11p_0_0.vlm"},
	{0x46F417BA2DA3DBDCULL, "gen_im_trs_dal_one_lsw_r1.uti"},
	{0x46F429A6C63AC106ULL, "6110894_f.fxe"},
	{0x46F42A38B99AC01BULL, "6059001_m.fxe"},
	{0x46F5B4AC1C2E0E61ULL, "zz_fol_debug.ncs"},
	{0x46F654EE67532D9CULL, "vxw_encounter1_f7_0_0.vlm"},
	{0x46F729869CE4CC96ULL, "llac_lt_warehouse_n_0_0.dds"},
	{0x46F74E9FD05DBBEFULL, "551049667.ogg"},
	{0x46F74EA8DB63CE8CULL, "805569793.ogg"},
	{0x46F87A236D883A0AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_3windows.phy"},
	{0x46F8C7E0D494C875ULL, "hlac_ht_keep_4519_14x.dds"},
	{0x46F8CC83D55BEB25ULL, "one100dg_vanard6032539.cl"},
	{0x46F9003F864DF490ULL, "377961245.ogg"},
	{0x46F9248D92177539ULL, "plt_cod_ltr_sk_gascard.plo"},
	{0x46F97A61DE1CDB56ULL, "vxw_encounter1_cf_0_0.vlm"},
	{0x46F996F1010613F1ULL, "217095405.ogg"},
	{0x46FA4B48029D8731ULL, "6035382_m.fxe"},
	{0x46FAD992D2D7F175ULL, "hlac_lt_0_lowtown_a2_n_65534_4z.dds"},
	{0x46FB1B4838E0F636ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtop.msh"},
	{0x46FB254838E10734ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtop.mmh"},
	{0x46FB294838E10E47ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowtop.mao"},
	{0x46FBA61866B4C31CULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_aoe_p.mmh"},
	{0x46FC0AFCBC3FCA61ULL, "dae000_bartrand6206873.cl"},
	{0x46FD000F7EACA568ULL, "plt_dre120pt_sandal.nss"},
	{0x46FD60AFC8065CAFULL, "6142204_m.fxe"},
	{0x46FD7AA7B8D247F4ULL, "gen00fstl_fenris_gift1.stg"},
	{0x46FE09C247606059ULL, "312683952.ogg"},
	{0x46FEC45CDDAFA446ULL, "vxlt_0_lowtown_docks_a3_n_bw5_0_0.vlm"},
	{0x46FF3310A943BF22ULL, "31948591.ogg"},
	{0x46FF9149E9EAD431ULL, "vxlt_0_lowtown_h84_0_0.vlm"},
	{0x46FFA26F22D1E59AULL, "dae000_bartrand6027505.cl"},
	{0x46FFD32013CE787FULL, "321273450.ogg"},
	{0x46FFD7AC1C369E66ULL, "zz_fol_debug.cnv"},
	{0x46FFFFE9DE7CB73BULL, "694367888.ogg"},
	{0x47005DF669C53D88ULL, "mer200ar_exterior.lst"},
	{0x4700E8133396115FULL, "vxlt_0_lowtown_a3_n_c_d3s_0_0.vlm"},
	{0x47013DB47C99AB2AULL, "hlac_lt_undercityserialkiller_-1_35z.dds"},
	{0x470185BE9298135FULL, "vxht_mansion_n_5ao_0_0.vlm"},
	{0x4701A7E76D20D333ULL, "402354342.ogg"},
	{0x47021FF7DED4C262ULL, "348033692.ogg"},
	{0x4702A3052B9E301CULL, "lowtown_fem_cityguard_mob6164835.cl"},
	{0x4702CE3AA702A8B6ULL, "gen_im_wep_war_two_act3_04.uti"},
	{0x4702EE17692562CEULL, "hlac_ca_caverns1_mines_24425_2y.dds"},
	{0x4703021EED7086CEULL, "364688852.ogg"},
	{0x4703090F7EB1207BULL, "plt_dre120pt_sandal.plo"},
	{0x4703DA3B8A18E314ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_rfoot.ani"},
	{0x4704640708751942ULL, "w_bonepit_171_0.phy"},
	{0x4704C6CB5949461CULL, "vxw_encounter1_k4_0_0.vlm"},
	{0x4704C74F5A47F6B9ULL, "vxw_encounter2_camp_pq_0_0.vlm"},
	{0x4704CB973713FB4AULL, "gen00fl_aveline6139989.cl"},
	{0x4704D33D6E221D0CULL, "474993103.ogg"},
	{0x4705337B87E52ABBULL, "6086513_m.fxe"},
	{0x47054186AE01D19FULL, "750181964.ogg"},
	{0x4705511AC0FBE906ULL, "plt_ssf300pt_main.plo"},
	{0x47058B941758FC8CULL, "326138454.ogg"},
	{0x4705C91AB9846302ULL, "zz_followers_banter6157712.cl"},
	{0x4705DF23C37663C1ULL, "vxht_bartrand_n_5o0_0_0.vlm"},
	{0x4706B4A95B15407BULL, "vxht_mansion2b_35j_0_0.vlm"},
	{0x4706FCCD757AA92BULL, "lt_0_lowtown_a2_n_36_0_lc.mmh"},
	{0x4707242BC78B0FA4ULL, "lt_backalley_g_1_0.phy"},
	{0x47075232C2DDD415ULL, "6121146_m.fxe"},
	{0x47075DF669CB5CD9ULL, "mer200ar_exterior.ncs"},
	{0x470798D2A50A4975ULL, "vxw_bonepit_128_0_0.vlm"},
	{0x4707EC3ED012F3D6ULL, "vxw_encounter1_hg_0_0.vlm"},
	{0x4707F3070395E546ULL, "vxht_mansion2_1kw_0_0.vlm"},
	{0x4708739D364B7307ULL, "75319491.ogg"},
	{0x4708B00CC990DAB3ULL, "901928433.ogg"},
	{0x4708DD31604D38C7ULL, "647940218.ogg"},
	{0x47095DBB8C430E3EULL, "6024116_m.fxe"},
	{0x4709F2BB7E1253BCULL, "536395643.ogg"},
	{0x4709F6EA9109391AULL, "hlac_ca_cavern1_haunted_24426_6z.dds"},
	{0x470A47781C229920ULL, "vxw_encounter2_camp_7q_0_0.vlm"},
	{0x470A53CCBE06315AULL, "41934126.ogg"},
	{0x470A90FB203A075BULL, "llac_lt_0_lowtown_n_2_1.dds"},
	{0x470AD73B8A1EFD76ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_rfoot.gad"},
	{0x470B64A1062C783AULL, "hlac_lt_gallowstemplar_2388_1z.dds"},
	{0x470BBAADABFAA91EULL, "187413787.ogg"},
	{0x470BEC672A40C3BCULL, "hlac_lt_0_lowtown_a2_n_-65536_7y.dds"},
	{0x470C270B04447A2EULL, "436553125.ogg"},
	{0x470C71C3FB175FFEULL, "hlac_lt_0_lowtown_a3_n_c_196607_0z.dds"},
	{0x470C8B77891A92A1ULL, "vxca_cave_23w_0_0.vlm"},
	{0x470D0C67749527EFULL, "vxca_caverns1_gk8_0_0.vlm"},
	{0x470D2C9106ED2567ULL, "6032204_m.fxe"},
	{0x470D7098E348E284ULL, "_test2.cut"},
	{0x470D7687B1CC94EDULL, "hlac_lt_undercityserialkiller_-1_29x.dds"},
	{0x470D772E897EC3A8ULL, "gen00fl_anders6125196.cl"},
	{0x470D93765226314BULL, "static.mat"},
	{0x470E1B2BC7911F2CULL, "lt_backalley_g_1_0.rml"},
	{0x470E726F967C3091ULL, "follower_banter6165475.cl"},
	{0x470E8DED1F3415C2ULL, "213456256.ogg"},
	{0x470F3177640EE946ULL, "6118743_m.fxe"},
	{0x470F6B338A1EF29CULL, "mag170_hugh6080691.cl"},
	{0x471040FB8D2C8EBFULL, "hlac_lt_gallowstemplar_f_2388_2y.dds"},
	{0x4711A7081F4AA9CDULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef01.msh"},
	{0x4711B9081F4AC853ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef01.mmh"},
	{0x4711BD4F9BCE9122ULL, "hlac_lt_foundry_n_320_5x.dds"},
	{0x4711E7972F30E3EEULL, "art\\levels\\ztools\\default\\generic\\1ow_basic.msh"},
	{0x4711F23B8A2549D5ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_p_rfoot.evt"},
	{0x4711F5972F30FB3FULL, "art\\levels\\ztools\\default\\generic\\1ow_basic.mao"},
	{0x471201972F310FDCULL, "art\\levels\\ztools\\default\\generic\\1ow_basic.mmh"},
	{0x4713392C4C3E0F39ULL, "vxw_encounter2_camp_kp_0_0.vlm"},
	{0x47135C460C6D0AC2ULL, "6061230_m.fxe"},
	{0x47139BEB0C105FD3ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend01_n.dds"},
	{0x4714244B22D4CFDFULL, "vxw_bonepitdestroyed_a3_254_0_0.vlm"},
	{0x4714E607AD62286FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottle_sa.dds"},
	{0x4714EB844E363E07ULL, "vxca_cavern1_haunted_ec3_0_0.vlm"},
	{0x47152988650F5ECCULL, "6093624_m.fxe"},
	{0x4715301C9FC76730ULL, "552178390.ogg"},
	{0x471536064657F311ULL, "vxlt_backalley_n_1dm_0_0.vlm"},
	{0x4715DC6D945CA609ULL, "hlac_ca_cavern1_haunted_24422_6y.dds"},
	{0x471623BA823CF04BULL, "6148949_m.fxe"},
	{0x4716280F4A893129ULL, "mcr370_meredith6083108.cl"},
	{0x471765388589EA1FULL, "character_character_p189.psh.fxo.dx11.assembly"},
	{0x47178DE720B34EA5ULL, "382573430.ogg"},
	{0x471795AB4F8BDA55ULL, "qun220cr_javaris6071858.cl"},
	{0x47184A78551BE090ULL, "427529762.ogg"},
	{0x4718CC7F97ED5E97ULL, "sky_lt_warehouse_n.mao"},
	{0x471952194555EE8AULL, "vxca_caverns2_85c_0_0.vlm"},
	{0x4719942B72AA4B0DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0cc_basic.msh"},
	{0x4719999C9C085111ULL, "zz_followers_banter6158925.cl"},
	{0x47199E0586E20F01ULL, "vxlt_0_lowtown_a3_n_c_es4_0_0.vlm"},
	{0x4719A22B72AA6340ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0cc_basic.mao"},
	{0x4719A62B72AA6993ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0cc_basic.mmh"},
	{0x4719CE84A77540C3ULL, "vxht_mansion_n_4zg_0_0.vlm"},
	{0x471A5AC669AEDF82ULL, "gen00fl_bethany6028808.cl"},
	{0x471A9EA1D83D115DULL, "hlac_ht_mansion2b_818_43x.dds"},
	{0x471B1F8D9234534EULL, "plt_cod_ltr_sk_gascard.nss"},
	{0x471BAAB2FE383A5FULL, "vxlt_hangedman_1ro_0_0.vlm"},
	{0x471C95D07EA9469FULL, "141708244.ogg"},
	{0x471CAA20DC0B45CDULL, "qun211ip_barrel_latch.utp"},
	{0x471DA354560501C6ULL, "6117643_m.fxe"},
	{0x471DB9FD52BE13BBULL, "sst_ss_hadriana.gda"},
	{0x471DC2F02FC2413DULL, "vxlt_hangedman_n_1ms_0_0.vlm"},
	{0x471E17E295362C57ULL, "vxw_sundermount_22b_0_0.vlm"},
	{0x471E65A6277B6701ULL, "6169425_m.fxe"},
	{0x471E9C465A84F503ULL, "vxht_keep_n_67f_0_0.vlm"},
	{0x471ED569CFE02EF5ULL, "6104150_m.fxe"},
	{0x471F35D0B42E0805ULL, "6208304_m.fxe"},
	{0x471F9600FAE8422FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblelarge.mao"},
	{0x471FA200FAE856CCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblelarge.mmh"},
	{0x471FA800FAE860BEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblelarge.msh"},
	{0x471FA94B17F0CA1EULL, "vxdr_primevaltomb_77m_0_1.vlm"},
	{0x471FB31D068946A7ULL, "675638703.ogg"},
	{0x47200D25A2AA67D1ULL, "6104795_f.fxe"},
	{0x47201145FF6E62A1ULL, "6136326_m.fxe"},
	{0x47211DBEB54697BEULL, "6065955_m.fxe"},
	{0x4721315C4E511AC6ULL, "19651455.ogg"},
	{0x47217A3038087B5AULL, "508491184.ogg"},
	{0x4721AB093E3427B9ULL, "vxw_encounter2_camp_fw_0_0.vlm"},
	{0x4721EA11D3A8B065ULL, "follower_banter6141864.cl"},
	{0x47228C38E344232CULL, "art\\vfx\\environment\\vfx_dark_fog.impa.ani"},
	{0x4722962814CAE85FULL, "vxw_bonepitdestroyed_a3_207_0_0.vlm"},
	{0x47234E3EBB7BD3E4ULL, "6059701_m.fxe"},
	{0x4723B56F482CFC74ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firelogs_e.dds"},
	{0x4723FC6A2C8A5A3BULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv1.ani"},
	{0x472400F01E2C92B1ULL, "vxdr_primevaltomb_50m_0_1.vlm"},
	{0x4724297850C03FBFULL, "follower_banter6173104.cl"},
	{0x47246728435BDC85ULL, "combat_ws_r3_a1.gda"},
	{0x4724A2385ED465D2ULL, "gen00fl_anders6096506.cl"},
	{0x4724EDCAE43F23D7ULL, "164303769.ogg"},
	{0x472594A17E6B0035ULL, "art\\characters\\playercharacter\\humanfemale\\arm_urva_1.mao"},
	{0x4725B734C84D4829ULL, "6039812_m.fxe"},
	{0x472602F5519014E6ULL, "unwrap.vsh.fxo"},
	{0x47266178737C82A1ULL, "vxlt_0_lowtown_a2_f66_0_2.vlm"},
	{0x472699F1089B6FF0ULL, "one100_party_comments6133457.cl"},
	{0x4726ED42952DEA8EULL, "6162333_m.fxe"},
	{0x472733E5DACD12F1ULL, "vxht_mansion_n_54b_0_0.vlm"},
	{0x472770DDFEDACCFDULL, "6103928_m.fxe"},
	{0x4727B64F9B218E31ULL, "6188665_f.fxe"},
	{0x47284116307164D0ULL, "252165005.ogg"},
	{0x47293F070893F6F6ULL, "w_bonepit_171_0.mao"},
	{0x47294B0708940A95ULL, "w_bonepit_171_0.mmh"},
	{0x47294D0708940E0BULL, "w_bonepit_171_0.msh"},
	{0x47299938E34A586EULL, "art\\vfx\\environment\\vfx_dark_fog.impa.gad"},
	{0x4729B53FD6BE02CBULL, "vxca_caverns2_78t_0_2.vlm"},
	{0x472A06DD054F6FBDULL, "6150069_m.fxe"},
	{0x472B1081C06CAFF1ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_lyrium_0.pwk"},
	{0x472B5ABF184D2917ULL, "6060683_m.fxe"},
	{0x472BBBCAF04A5348ULL, "6201276_m.fxe"},
	{0x472BFD2E72F41E1BULL, "mag101_alain6053004.cl"},
	{0x472C05000652C577ULL, "mcr310_argument6074457.cl"},
	{0x472C349A3E53BB7BULL, "hlac_lt_0_lowtown_docks_a3_65533_0z.dds"},
	{0x472C34CD21B859BDULL, "vxlt_hangedman_n_1jr_0_0.vlm"},
	{0x472C4D7F2BAB5160ULL, "hlac_w_woundedcoast_-262139_0y.dds"},
	{0x472C5CE212AD3F60ULL, "6162670_m.fxe"},
	{0x472C78330E5A7A1FULL, "vxht_mansion_n_4cu_0_1.vlm"},
	{0x472CCE6A7F89353AULL, "hlac_ca_caverns1_mines_24425_12x.dds"},
	{0x472D395C0BFA0489ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_monique.ani"},
	{0x472DA5BF0F606B0AULL, "gen00fl_anders6111287.cl"},
	{0x472DD85AE0484C13ULL, "mag170_thrask6054032.cl"},
	{0x472E3FFAE6738C26ULL, "art\\vfx\\materials\\vfx_blood02.mao"},
	{0x472E8322F1647B21ULL, "6131226_m.fxe"},
	{0x472F30153905992DULL, "dae000_meredith6085283.cl"},
	{0x472F84F994034990ULL, "6108317_m.fxe"},
	{0x472FB468838A11F9ULL, "qcr000_varric6068104.cl"},
	{0x472FE475DC44DBA2ULL, "t3_wpn_stn.tnt"},
	{0x4730005C6E121E1CULL, "6118377_f.fxe"},
	{0x473014B454431A08ULL, "3649565.ogg"},
	{0x47301CE6302A4039ULL, "vxw_encounter2_camp_av_0_0.vlm"},
	{0x47301E0FC496D818ULL, "vxht_0_hightown_n_7vk_0_0.vlm"},
	{0x4730259AFF377884ULL, "6117112_f.fxe"},
	{0x47306A4E80D567CFULL, "6215047_m.fxe"},
	{0x473070B2D92781DFULL, "105284513.ogg"},
	{0x47309438E3506E4DULL, "art\\vfx\\environment\\vfx_dark_fog.impa.evt"},
	{0x4730B5249719E895ULL, "ico_poison_deathroot.dds"},
	{0x4730E6F7958AD23AULL, "6147689_f.fxe"},
	{0x4730E714FEFEF914ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic_3.mao"},
	{0x4731046A2C94F232ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv1.evt"},
	{0x4731314B6B338683ULL, "gen00fl_sebastian6198189.cl"},
	{0x47314BE035221955ULL, "vxlt_backalley_g_1oi_0_0.vlm"},
	{0x47327C51DBA7530CULL, "gen_im_arm_hlm_med_set_high.uti"},
	{0x47329E16A6647600ULL, "vxca_cavern1_haunted_j60_0_0.vlm"},
	{0x4732A8C4F95F431EULL, "qun211_party_comment6072248.cl"},
	{0x4732DE5AA73F9533ULL, "vxlt_hangedman_n_10j_0_0.vlm"},
	{0x47334890EF288055ULL, "6201252_f.fxe"},
	{0x47338039920694DDULL, "vxw_sundermount_21y_0_0.vlm"},
	{0x4733A154504D1DDAULL, "gen00fl_aveline6101429.cl"},
	{0x4733B592EDEBFD55ULL, "6135702_f.fxe"},
	{0x4734545C0C00510BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_monique.gad"},
	{0x47349D65D11049E1ULL, "6080720_m.fxe"},
	{0x4735E6D472A0DE10ULL, "1032233051.ogg"},
	{0x473618AF696E0BD5ULL, "mag100_cullen6140918.cl"},
	{0x4736D01A1795FAE2ULL, "dae000cr_grand_cleric6148145.cl"},
	{0x4736E0972F4FF3A7ULL, "art\\levels\\ztools\\default\\generic\\1ow_basic.phy"},
	{0x47370FB28F1FA3F0ULL, "dae000_orsino6085439.cl"},
	{0x4738236A2C9B45C5ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv1.gad"},
	{0x473936BF9648E991ULL, "qcr000_arishok6059995.cl"},
	{0x4739680F0B20FB4EULL, "6033991_f.fxe"},
	{0x47399B9953FD93AEULL, "716693019.ogg"},
	{0x473A17FDF29BACAAULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobc.mao"},
	{0x473A1BFDF29BB3F9ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobc.mmh"},
	{0x473A25FDF29BC4E7ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobc.msh"},
	{0x473A26A35248A6DCULL, "928433724.ogg"},
	{0x473A2EB13AABE46CULL, "6098203_m.fxe"},
	{0x473A8BF36BB51CB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_earth_movement.ani"},
	{0x473A8E50029808E0ULL, "vxht_0_hightown_a3_8pd_0_0.vlm"},
	{0x473A90C50DAC630CULL, "vxlt_blackemporium_15y_0_0.vlm"},
	{0x473B41DAF0FC72EDULL, "zz_followers_banter6159078.cl"},
	{0x473B53F8451A123AULL, "dae000_arishok6067674.cl"},
	{0x473BF02FF0DA7CA0ULL, "6153317_m.fxe"},
	{0x473C1109E5715763ULL, "544839918.ogg"},
	{0x473C2CB61C1AAB0FULL, "6165445_m.fxe"},
	{0x473C956550C4B41BULL, "vxw_bonepit_144_0_0.vlm"},
	{0x473C9F728D61F52BULL, "lowtown_fem_prostitute_sta6152320.cl"},
	{0x473CB51316A82BA5ULL, "6149067_m.fxe"},
	{0x473CC9E5E60962EAULL, "277257871.ogg"},
	{0x473D25F82912F2A8ULL, "gencr_spider_deeproads.utc"},
	{0x473D740126BFF4E0ULL, "6110275_m.fxe"},
	{0x473D772C59F71A80ULL, "gen00fl_aveline6139271.cl"},
	{0x473DF8E2ED596941ULL, "6077757_m.fxe"},
	{0x473E2E32986FC4B3ULL, "ser221_party_comments6163781.cl"},
	{0x473E41C4E60DCE31ULL, "vxht_chantry_n_2s0_0_0.vlm"},
	{0x473E7BEA69B203A7ULL, "hlac_lt_0_lowtown_docks_65534_4x.dds"},
	{0x473E8BA66869C7A9ULL, "dae021_guardsman_garol.utc"},
	{0x473F5E508328390EULL, "mrl000_marethari6129317.cl"},
	{0x473FA2A19370A32AULL, "338118114.ogg"},
	{0x473FD069A873D401ULL, "leave_05.dds"},
	{0x47402F19C740D2CDULL, "lt_0_lowtown_docks_a3_99_0.msh"},
	{0x47403D19C740EA00ULL, "lt_0_lowtown_docks_a3_99_0.mao"},
	{0x47404119C740F153ULL, "lt_0_lowtown_docks_a3_99_0.mmh"},
	{0x47404B667E9B20D6ULL, "vxht_mansion_n_5cp_0_0.vlm"},
	{0x4740554F4442D5CCULL, "6047399_m.fxe"},
	{0x47408A3FC72DC674ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml106a_0s.dds"},
	{0x4740D2066788061CULL, "art\\vfx\\cinematic\\vfx_dust_imp_med_p.impa.ani"},
	{0x474166F36BBAFCFBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_earth_movement.gad"},
	{0x474196D5CE8AC439ULL, "287480893.ogg"},
	{0x4741BBFB3383F283ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_in_c.impa.evt"},
	{0x4741FB23520EA877ULL, "6111095_m.fxe"},
	{0x47422CADC7026437ULL, "6142110_m.fxe"},
	{0x47425C2241618CA9ULL, "llac_ht_0_hightown_a3_131071_0.dds"},
	{0x47428AC567BE18CDULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_03brgdebris.mmh"},
	{0x47428EC567BE1F1EULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_03brgdebris.mao"},
	{0x47429CC567BE3753ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_03brgdebris.msh"},
	{0x4742AB693F424924ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkf.ani"},
	{0x4742D1AB2B7F7D57ULL, "6111608_m.fxe"},
	{0x4742DB2AEB2374EAULL, "vxlt_gallowstemplar_3ri_0_0.vlm"},
	{0x474307CE94BA98A3ULL, "hlac_ht_0_hightown_a3_65537_1z.dds"},
	{0x474350E1E0A346FCULL, "6085803_m.fxe"},
	{0x47438C8358395888ULL, "vxlt_0_lowtown_a2_n_c_eou_0_1.vlm"},
	{0x4743F74862AC894EULL, "1003020385.ogg"},
	{0x47440451ABCACC72ULL, "follower_barks6153537.cl"},
	{0x47446AAAB270C42AULL, "179885893.ogg"},
	{0x47449E5A6EDC6B8AULL, "vxht_0_hightown_b5z_0_0.vlm"},
	{0x4744B7528422C753ULL, "hlac_w_bonepit_131073_0x.dds"},
	{0x4744E940D30A2BDAULL, "qun110_petrice6042302.cl"},
	{0x4744F92590A288F9ULL, "6053922_m.fxe"},
	{0x4745A8731E309C45ULL, "6113658_f.fxe"},
	{0x4745D2C3635F2B5CULL, "vxlt_0_lowtown_docks_b51_0_0.vlm"},
	{0x4746283F6E88E30AULL, "art\\characters\\playercharacter\\humanmale\\hlf_medf.mao"},
	{0x47462C3F6E88EA59ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medf.mmh"},
	{0x4746363F6E88FB47ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medf.msh"},
	{0x4746397D9A42A124ULL, "205786963.ogg"},
	{0x474696CF94713095ULL, "6094714_m.fxe"},
	{0x474783F36BBF99E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_earth_movement.evt"},
	{0x47478BADA915CF4BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\ready.ani"},
	{0x4747AABFA605EE2DULL, "vxlt_0_lowtown_docks_n_aru_0_0.vlm"},
	{0x4747D3A2B345722AULL, "hlac_ca_caverns1_mines_24421_11y.dds"},
	{0x4747DF06678E3B5EULL, "art\\vfx\\cinematic\\vfx_dust_imp_med_p.impa.gad"},
	{0x4747E765D6A6C8ACULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_abomination_sacredmage.ani"},
	{0x474894FB3389CF04ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_in_c.impa.gad"},
	{0x4748956F72876876ULL, "6164458_m.fxe"},
	{0x4748A8693F46AF86ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkf.gad"},
	{0x4748D2E5928A6630ULL, "6106931_m.fxe"},
	{0x4748E4301250108BULL, "vxht_keep_n_3co_0_0.vlm"},
	{0x4749243D943AC8F8ULL, "hlac_lt_backalley_n_0_10z.dds"},
	{0x4749B5B50F6D72AEULL, "prm000im_adder_longbow.uti"},
	{0x4749C46928CFD8DAULL, "6083083_m.fxe"},
	{0x474A32FB7CAE2543ULL, "504280264.ogg"},
	{0x474BC5E89F72EA1CULL, "vxlt_hangedman_27d_0_0.vlm"},
	{0x474BC9AC67E5FE5FULL, "vxdr_primevaltomb_995_0_0.vlm"},
	{0x474C335FCEE709BDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_hlmeyesm1.mao"},
	{0x474C6E1BA0FB2A22ULL, "494424818.ogg"},
	{0x474C88735E2749A7ULL, "cd_seb221_flora_harimann.bnk"},
	{0x474C8909C971F37FULL, "dae000ip_sit_ground_d.utp"},
	{0x474CA6184F68EEB3ULL, "hlac_w_sundermount_131067_0z.dds"},
	{0x474CA74E1A0B2D1CULL, "vxlt_undercityencounter_1bn_0_1.vlm"},
	{0x474CF8962AC3CCBFULL, "759177933.ogg"},
	{0x474D31835A8A6704ULL, "vxdr_primevaltomb_7m4_0_0.vlm"},
	{0x474DCE445733956FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomeb_sa.dds"},
	{0x474E0D7AFD7D894FULL, "mcr371_scared_mage_f6086926.cl"},
	{0x474E5483A5A48B98ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_008a_0t.dds"},
	{0x474EBA0667941B7DULL, "art\\vfx\\cinematic\\vfx_dust_imp_med_p.impa.evt"},
	{0x474EEB5C163C97D6ULL, "vxlt_hangedman_247_0_0.vlm"},
	{0x474EF465D6ACFDEEULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_abomination_sacredmage.gad"},
	{0x474EFD7A0E305511ULL, "vxlt_backalley_n_1pa_0_0.vlm"},
	{0x474FA3693F4CC665ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkf.evt"},
	{0x474FA3FB339007E2ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_in_c.impa.ani"},
	{0x474FE0433816D55DULL, "996393844.ogg"},
	{0x47502BD0638BC7A1ULL, "fx_glyph_neutral.dds"},
	{0x4750321AC13BA045ULL, "plt_ssf300pt_main.nss"},
	{0x4750461597D5205DULL, "636630773.ogg"},
	{0x475088D5B30FD307ULL, "mud_f.dds"},
	{0x475123D4AD086D4AULL, "cn_bdy_wera_0t.dds"},
	{0x475185D50025F03CULL, "6162788_m.fxe"},
	{0x4752C64CCD4362D9ULL, "879263295.ogg"},
	{0x4753C815047CC7D5ULL, "6118761_f.fxe"},
	{0x4753F83E8F491298ULL, "6014697_f.fxe"},
	{0x475408C762B1535EULL, "gen00fl_anders6125510.cl"},
	{0x47540CE37761F8F7ULL, "223608958.ogg"},
	{0x475498EDD7435D4EULL, "948332115.ogg"},
	{0x4754C17D53DC6842ULL, "dae110cr_elf_m_01.utc"},
	{0x4754E0190EA84E5EULL, "vxca_caverns1_mines_k6k_0_0.vlm"},
	{0x4754F72029F7272AULL, "6109005_f.fxe"},
	{0x475545B625482595ULL, "vxdr_0_deeproads_gfz_0_0.vlm"},
	{0x475573ADA921E3C2ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\ready.evt"},
	{0x4755B0ACC790EFB5ULL, "755948898.ogg"},
	{0x4755F3899AF9C2DAULL, "6041706_m.fxe"},
	{0x47567CAFB553021CULL, "hlac_dr_slavers_3207_8x.dds"},
	{0x475689A8D39AF599ULL, "39500435.ogg"},
	{0x47569915C5C8BF02ULL, "gen00fl_anders6125288.cl"},
	{0x4756A1A30EC5BAD9ULL, "6173152_m.fxe"},
	{0x4756CF27D34CD4C7ULL, "6153729_m.fxe"},
	{0x4757002EBF274C33ULL, "gen00fl_varric6098075.cl"},
	{0x47584DF675AAB5E7ULL, "vxlt_0_lowtown_docks_a3_brt_0_0.vlm"},
	{0x475864A11ACE5B9BULL, "6180967_m.fxe"},
	{0x4758798A0B36A326ULL, "vxht_mansion2_2vo_0_0.vlm"},
	{0x475927F6C32C06D9ULL, "hlac_lt_0_lowtown_-1_6x.dds"},
	{0x475A0F6CEA76EB96ULL, "116459134.ogg"},
	{0x475A50C34B8F48C6ULL, "vxht_bartrand_n_5q9_0_0.vlm"},
	{0x475B2280F3A4D508ULL, "zz_followers_banter6158250.cl"},
	{0x475B72ADA9264DD5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\ready.gad"},
	{0x475C45F7296A3EAFULL, "vxlt_undercitydungeon_1kb_0_0.vlm"},
	{0x475C9D04CC92172EULL, "gen00fl_varric6093608.cl"},
	{0x475CC6E628AFF608ULL, "378941918.ogg"},
	{0x475D1579EB98CE6BULL, "vxw_woundedcoast_9y_0_0.vlm"},
	{0x475D34AC3E651F20ULL, "hlac_w_sundermount_-65541_0y.dds"},
	{0x475D5D390832B056ULL, "vxlt_hangedman_218_0_0.vlm"},
	{0x475DA0E988F1B832ULL, "6129112_m.fxe"},
	{0x475DC79E1BCC7CC5ULL, "12486277.ogg"},
	{0x475DE169A88CEE8BULL, "leave_05.mao"},
	{0x475DF5DA1BB334CCULL, "6072449_m.fxe"},
	{0x475E1C1FE80615D8ULL, "296063325.ogg"},
	{0x475E26D2DDE3E00EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_reaction.ani"},
	{0x475E2FE963E70A86ULL, "vxca_caverns1_mines_iml_0_0.vlm"},
	{0x475E532DEEC9D237ULL, "gen_im_acc_rng_act2_10.uti"},
	{0x475E95B6A9E03B08ULL, "cd_mrl150_cut_amuletritual.wbd"},
	{0x475EDE812CF62A8AULL, "6120050_m.fxe"},
	{0x475F93BEA9ECFBC4ULL, "vxht_mansion_n_4va_0_1.vlm"},
	{0x475F9A6D5DEB418AULL, "dae200_bran6144580.cl"},
	{0x475FF7B1F21C08BCULL, "6167488_m.fxe"},
	{0x47609C7FBFBDE73AULL, "llac_lt_gallowscourtyard_n_-65537_2.dds"},
	{0x4760FCA9EBA645C1ULL, "279687941.ogg"},
	{0x47617314DF1EEA96ULL, "w_sundermount_18_0.msh"},
	{0x47617D14DF1EFB94ULL, "w_sundermount_18_0.mmh"},
	{0x47618114DF1F02A7ULL, "w_sundermount_18_0.mao"},
	{0x4761811D25A3189DULL, "19710571.ogg"},
	{0x4761B36FB6A1BAF6ULL, "hlac_lt_gallowscourtyard_n_-1_4y.dds"},
	{0x4762692B1379B8E4ULL, "gen00fl_fenris6126717.cl"},
	{0x4762BDC648668B7CULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgtp_0s.dds"},
	{0x47630F159462E049ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.impa.ani"},
	{0x47633B3274EFB429ULL, "hm_saemus.mor"},
	{0x47636760B628B779ULL, "56836115.ogg"},
	{0x476371405D9B34FDULL, "508024950.ogg"},
	{0x4763DB982EE7069CULL, "hlac_lt_0_lowtown_a2_196607_0x.dds"},
	{0x4763E8CB09D0E3E2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2b.msh"},
	{0x4763EACB09D0E698ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2b.mmh"},
	{0x4763F6CB09D0FB3BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2b.mao"},
	{0x476441A6CA14AC7EULL, "355042328.ogg"},
	{0x47644F7C29B28B1CULL, "vxca_caverns1_mines_fn8_0_0.vlm"},
	{0x4764931CEA825B3FULL, "430929273.ogg"},
	{0x4764E0BC3DB3865CULL, "6148407_m.fxe"},
	{0x47656A4578CAABE8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blend.msh"},
	{0x47657001CE84DCFEULL, "vxlt_0_lowtown_a2_n_e2w_0_0.vlm"},
	{0x47657C4578CACA79ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blend.mao"},
	{0x4765884578CADE12ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blend.mmh"},
	{0x4765EB90B60D72F5ULL, "cod_bks_kirkwall_hist_2.utp"},
	{0x4765F282336C3B16ULL, "6134972_m.fxe"},
	{0x4767217B634D385BULL, "gen_im_acc_amu_act2_03.uti"},
	{0x476727C65BA14459ULL, "gen00fl_anders6107692.cl"},
	{0x47674152EF0F970DULL, "dae000_pc_home_ambients6208704.cl"},
	{0x476749291A054D20ULL, "hla_ht_0_hightown_a2_65536_0y.dds"},
	{0x476791C46BCB599EULL, "mrl150_merrill6093566.cl"},
	{0x4767F2F0BC333C7CULL, "art\\vfx\\combat\\vfx_r_var_pin_shot_act_c.impa.ani"},
	{0x476888C48913C2EFULL, "118979510.ogg"},
	{0x47688949F557C9BBULL, "vxca_cavern1_haunted_inp_0_0.vlm"},
	{0x47689BA03AC04B63ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.impa.ani"},
	{0x4768A44933A5411CULL, "gen00fl_bethany6097884.cl"},
	{0x4768E1660350787CULL, "6048813_m.fxe"},
	{0x47691B9A35DE4E10ULL, "vxlt_0_lowtown_n_f08_0_0.vlm"},
	{0x47692A15946779CBULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.impa.gad"},
	{0x47693B5F1D69951BULL, "gen00fl_aveline6137557.cl"},
	{0x4769889CCDB4188EULL, "39605856.ogg"},
	{0x4769A649B2CEC92FULL, "hlac_lt_0_lowtown_docks_a3_n_-3_4y.dds"},
	{0x4769ADE2C1882CEBULL, "vxht_playershouse_n_1rf_0_1.vlm"},
	{0x476A188E6E51F080ULL, "vxw_woundedcoast_he_0_0.vlm"},
	{0x476ACC38E8B2A207ULL, "669702892.ogg"},
	{0x476B1ED2DDEE5D5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_reaction.evt"},
	{0x476B8756DD8EE6EBULL, "vxw_woundedcoast_6x_0_0.vlm"},
	{0x476B93A0D1611419ULL, "62789298.ogg"},
	{0x476BBED1A0B70DF3ULL, "674214696.ogg"},
	{0x476BD35A9CE4493EULL, "6173186_m.fxe"},
	{0x476C56DCFB67C66CULL, "art\\characters\\playercharacter\\humanmale\\arm_medp.phy"},
	{0x476C70B742E88DF4ULL, "646295444.ogg"},
	{0x476CD8E549C6CE1FULL, "dae000_orsino6085452.cl"},
	{0x476CE72970434352ULL, "376837624.ogg"},
	{0x476D0F68C768A1ADULL, "vxlt_0_lowtown_a2_n_elj_0_0.vlm"},
	{0x476D3890D219C4CEULL, "vxlt_ruinedhouse_n_6x_0_0.vlm"},
	{0x476D738B83C821C8ULL, "327426468.ogg"},
	{0x476E5FD11A12BBBAULL, "llac_lt_gallowstemplar_f_2391_6.dds"},
	{0x476E6EB09CEB6CF2ULL, "vxca_caverns2prc_a6z_0_0.vlm"},
	{0x476E7357B9336A0CULL, "llac_w_bonepitmining_a2_262145_0.dds"},
	{0x476E952B55C393D2ULL, "164981346.ogg"},
	{0x476EFFF0BC3970BEULL, "art\\vfx\\combat\\vfx_r_var_pin_shot_act_c.impa.gad"},
	{0x476F29411A341866ULL, "s_ballista_03.dds"},
	{0x476FBA34DE569546ULL, "hlac_lt_gallowstemplar_n_2383_3z.dds"},
	{0x476FE23F203AF682ULL, "character_character_p43.psh.fxo.dx11.assembly"},
	{0x47702715946D9330ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.impa.evt"},
	{0x477052A3B5308B7DULL, "hlac_ht_bartrand_n_3366_0x.dds"},
	{0x477064031B2135A0ULL, "dae000cr_grand_cleric6124750.cl"},
	{0x4770767D95C8FCCFULL, "6198486_f.fxe"},
	{0x4771980B4153BD5DULL, "6191044_m.fxe"},
	{0x4771E5892588043EULL, "535420830.ogg"},
	{0x4771FFD2DDF44748ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_friend_reaction.gad"},
	{0x4772D727B6FD1CE8ULL, "6102926_m.fxe"},
	{0x47733FF886EEF0CDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sider_1.mao"},
	{0x477360E369185161ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_2x.dds"},
	{0x4773BED88352B0CBULL, "6152331_m.fxe"},
	{0x477478AADB8F79AFULL, "6142015_m.fxe"},
	{0x4774E4C764A73862ULL, "gen00fl_dog6170607.cl"},
	{0x4775B3A03ACAFE5AULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.impa.evt"},
	{0x4775CC4E16DC3D16ULL, "vxlt_gallowstemplar_n_2rw_0_0.vlm"},
	{0x4775DAF0BC3F50DDULL, "art\\vfx\\combat\\vfx_r_var_pin_shot_act_c.impa.evt"},
	{0x47766DB68B97D6C4ULL, "6064958_m.fxe"},
	{0x4776D0210CFC536DULL, "6148382_f.fxe"},
	{0x4777E9C4A6691D13ULL, "6173411_m.fxe"},
	{0x477866EC48599ECCULL, "prm000im_irons_belt.uti"},
	{0x4778C9F389D61B21ULL, "6096024_m.fxe"},
	{0x4778F187B4DF8F5FULL, "vxw_encounter2_camp_6a_0_0.vlm"},
	{0x4778F7AEEC0F45BDULL, "6147515_m.fxe"},
	{0x47796FBBDA0DA2CAULL, "follower_banter6160730.cl"},
	{0x47797E735E4E26DDULL, "cd_seb221_flora_harimann.wbd"},
	{0x477AB1D765D8921BULL, "6065294_f.fxe"},
	{0x477B13A347D33B86ULL, "vxca_caverns1_mines_ign_0_0.vlm"},
	{0x477B43E36DA77DF6ULL, "6114853_m.fxe"},
	{0x477B45F296CFF0E6ULL, "619023131.ogg"},
	{0x477BE73C4A50A0EFULL, "6047456_f.fxe"},
	{0x477BF70252051573ULL, "1067534797.ogg"},
	{0x477C2FE3BE82FE02ULL, "vxlt_ruinedhouse_n_hj_0_0.vlm"},
	{0x477C83A995B336A0ULL, "473778967.ogg"},
	{0x477C90F16AFFB82BULL, "mer320_hubert6051980.cl"},
	{0x477C9C73FCBEB2F3ULL, "gen00fl_merrill6150379.cl"},
	{0x477CB2A03AD11B4DULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.impa.gad"},
	{0x477D74656264CFC3ULL, "gen00fl_anders6119143.cl"},
	{0x477D877A41F6851AULL, "vxlt_0_lowtown_exf_0_0.vlm"},
	{0x477DF1EED2F2EF47ULL, "6087679_m.fxe"},
	{0x477EBD10E58538EBULL, "qun160_vashothlead6055574.cl"},
	{0x477F32411E29E0DAULL, "1030779450.ogg"},
	{0x477F5F5E2F70CC2EULL, "vxw_bonepitdestroyed_a3_kc_0_0.vlm"},
	{0x477F6A4EC180C32AULL, "art\\characters\\creatures\\pridedemon\\animation\\cindes\\pose\\ready_loop.gad"},
	{0x477F7D8643F5316BULL, "hlac_dr_slavers_3215_7y.dds"},
	{0x477FA3FEEE23138BULL, "vxlt_0_lowtown_gbs_0_0.vlm"},
	{0x477FAF52E8A7FD1FULL, "6061815_m.fxe"},
	{0x4780338A24B65A16ULL, "llac_w_bonepit_65540_0.dds"},
	{0x4780386083462D62ULL, "vxw_bonepit_xf_0_0.vlm"},
	{0x4780543A33A40B82ULL, "w_sundermount_69_0.msh"},
	{0x4780563A33A40F38ULL, "w_sundermount_69_0.mmh"},
	{0x47805AD51250E705ULL, "lt_0_lowtown_a2_n_c_36_0_lc.mmh"},
	{0x4780623A33A423DBULL, "w_sundermount_69_0.mao"},
	{0x4780EEFDB4631CC8ULL, "985821269.ogg"},
	{0x4781254FCE8DAB93ULL, "364997041.ogg"},
	{0x47813999F0E11705ULL, "vxlt_gallowscourtyard_n_5yh_0_0.vlm"},
	{0x4781C506CEF5C2D6ULL, "115136047.ogg"},
	{0x478253BBB2710DFEULL, "vxlt_0_lowtown_a2_n_e8i_0_0.vlm"},
	{0x4782E8EBBFCA8435ULL, "341487547.ogg"},
	{0x47830CBF6B5C0259ULL, "6135976_m.fxe"},
	{0x4783563A33A6454BULL, "w_sundermount_69_0.rml"},
	{0x4783C667DBDCBE42ULL, "hf_nella.mor"},
	{0x47846DBBB96A666DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookgen_d.dds"},
	{0x4784B329598C8066ULL, "gen00fl_varric6081254.cl"},
	{0x4784DF1E7AD8BB26ULL, "framebuffereffects_dofblend_p0.psh.fxo"},
	{0x47852B4EF3A74B30ULL, "vxca_caverns2prc_907_0_1.vlm"},
	{0x4785F858C7F0FF65ULL, "412867216.ogg"},
	{0x4785FBFF945063B8ULL, "vxht_mansion2b_311_0_0.vlm"},
	{0x47865D5DAEDA03CEULL, "llac_w_encounter3_-131074_0.dds"},
	{0x4786CFF16FF2A175ULL, "mnd_pro_01.mao"},
	{0x4786E789BB9DD65BULL, "gen00fl_fenris6110920.cl"},
	{0x4786F8F72262C904ULL, "vxdr_primevaltomb_7y8_0_0.vlm"},
	{0x4786FC48523E2180ULL, "vxw_woundedcoast_bc_0_0.vlm"},
	{0x478730411A48B5B4ULL, "s_ballista_03.mao"},
	{0x47873E34AB80BC51ULL, "gen00fl_carver6102570.cl"},
	{0x47877B87F1733375ULL, "6103241_m.fxe"},
	{0x478839B4AE666DF4ULL, "heraldry.gda"},
	{0x478923F6D092CCEFULL, "gencr_trap_explosives_1.utc"},
	{0x47895F7C7F01ADF2ULL, "161182290.ogg"},
	{0x4789A02830B7E57EULL, "vxca_caverns1_mines_j57_0_0.vlm"},
	{0x4789B5C05F9D9676ULL, "6113953_m.fxe"},
	{0x478A5D3A33AC6FC3ULL, "w_sundermount_69_0.phy"},
	{0x478A69BB0A1F3DEBULL, "6063822_m.fxe"},
	{0x478A82A5E0B4DAB8ULL, "524938766.ogg"},
	{0x478A84A723D40A77ULL, "vxw_bonepitdestroyed_a3_19c_0_0.vlm"},
	{0x478A860B663E46F1ULL, "vxca_caverns1_jlj_0_0.vlm"},
	{0x478AC8DFF727E1F7ULL, "mag120_macha6080657.cl"},
	{0x478B38C9927E6E18ULL, "zz_qun_debug6045408.cl"},
	{0x478BD1532432C05BULL, "6098851_m.fxe"},
	{0x478BDBC0CEFEB1C7ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_7y.dds"},
	{0x478C5378A4C3CDF1ULL, "hlac_lt_hangedman_n_-1_6y.dds"},
	{0x478C97BD640F74D4ULL, "hlac_lt_warehouse_n_65537_4z.dds"},
	{0x478CAC24C4D1AA60ULL, "vxht_mansion2b_1aw_0_0.vlm"},
	{0x478E1140E807DA28ULL, "hlac_ca_caverns2_5802_42y.dds"},
	{0x478E57A1ECF3BDB5ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax105a_0t.dds"},
	{0x478F4FA77EFF61B1ULL, "plt_cod_cha_aveline2.plo"},
	{0x478F70E5F1BB19FEULL, "mag291_torpor6026985.cl"},
	{0x478FB2692ACEE32BULL, "vxlt_0_lowtown_docks_n_b6i_0_1.vlm"},
	{0x478FDE9E74CD649BULL, "hlac_lt_0_lowtown_a2_n_-262143_0x.dds"},
	{0x478FF2EBB65951FCULL, "485589827.ogg"},
	{0x478FF92D347F1F5FULL, "6089586_f.fxe"},
	{0x479013C8323A8C3BULL, "55228861.ogg"},
	{0x47902DD20B1E7671ULL, "vxw_encounter2_camp_pi_0_0.vlm"},
	{0x47904EF63EC3D9CDULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_013a_0d.dds"},
	{0x479118E05B55E792ULL, "dae000ip_crowd_em4_talk18.utp"},
	{0x4791B7571F6C08E5ULL, "dae000_arishok6067701.cl"},
	{0x4791F5306EA6B0A8ULL, "268260181.ogg"},
	{0x47921C178769567AULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinsarc.mmh"},
	{0x4792201787695CB1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinsarc.mao"},
	{0x47922E17876974C0ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinsarc.msh"},
	{0x4793454EC1912D10ULL, "art\\characters\\creatures\\pridedemon\\animation\\cindes\\pose\\ready_loop.ani"},
	{0x47938F71D1E29D2CULL, "hlac_lt_0_lowtown_a2_131072_2y.dds"},
	{0x4793F61B6FAA4194ULL, "216044520.ogg"},
	{0x479490074428C4B9ULL, "vxht_chantry_n_33p_0_1.vlm"},
	{0x4794BC0E3962DF77ULL, "878572493.ogg"},
	{0x4794E7FAFC359C6AULL, "hightown_male_noble_gossip6171400.cl"},
	{0x479581703EBD30B8ULL, "vxlt_gallowstemplar_6s1_0_0.vlm"},
	{0x4795E923512D0B30ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_keep.mao"},
	{0x4795ED23512D1243ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_keep.mmh"},
	{0x4795FB23512D299DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_keep.msh"},
	{0x479604007ADCF169ULL, "lt_0_lowtown_docks_a3_n_219_0.mao"},
	{0x479610007ADD0502ULL, "lt_0_lowtown_docks_a3_n_219_0.mmh"},
	{0x479612007ADD08B8ULL, "lt_0_lowtown_docks_a3_n_219_0.msh"},
	{0x479614272CAD14AFULL, "hlac_dr_slavers_3217_2x.dds"},
	{0x479667A7D10FCEF3ULL, "6051214_m.fxe"},
	{0x47967E5BBF4BD9BCULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatiki.mao"},
	{0x47968A5BBF4BEE5FULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatiki.mmh"},
	{0x4796905BBF4BF801ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatiki.msh"},
	{0x4796A866A47087ACULL, "ser211cr_demon_alessa.utc"},
	{0x47977ED2D54E1AD2ULL, "6204985_m.fxe"},
	{0x479850F760BA33FFULL, "gen00fl_anders6119198.cl"},
	{0x479868043A37FD34ULL, "gen00fl_varric6076218.cl"},
	{0x479879A0086191C4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_k_bk.ani"},
	{0x4798A4FD6E92EC4BULL, "lgt110_gxa_coa_spy.bnk"},
	{0x4798F3EC5C1000EDULL, "530350126.ogg"},
	{0x479919CA1C028C9CULL, "llac_lt_darktown_-1_73.dds"},
	{0x47991EE1E70C564EULL, "vxca_cavern1_haunted_e6m_0_0.vlm"},
	{0x47992EAC7018C484ULL, "llac_lt_0_lowtown_a2_n_c_-65536_8.dds"},
	{0x4799C12E221B21B3ULL, "494008794.ogg"},
	{0x4799DF2D40B8ACB3ULL, "character_character_p15.psh.fxo.dx11.assembly"},
	{0x479A5FCD80A803F7ULL, "vxca_cavern1_haunted_ind_0_0.vlm"},
	{0x479AA15080D120EFULL, "433861174.ogg"},
	{0x479AA99DAA791D3AULL, "473186746.ogg"},
	{0x479AC3BD5D3EBF94ULL, "dae100_bran6144435.cl"},
	{0x479BA6FD6E95252BULL, "lgt110_gxa_coa_spy.cnv"},
	{0x479C250244DD69C0ULL, "133672606.ogg"},
	{0x479CF12C7F8FC208ULL, "hm_refugee_5.mor"},
	{0x479D1CFDF2F032E6ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobc.phy"},
	{0x479D84AFDCBADBD1ULL, "vxw_woundedcoast_15b_0_0.vlm"},
	{0x479E0ABC0E853DC3ULL, "vxca_caverns1_i1i_0_1.vlm"},
	{0x479E5C2200EBD6E2ULL, "vxca_cavern1_haunted_ii4_0_0.vlm"},
	{0x479E9F57E5E1A95FULL, "vxca_caverns1_j5c_0_0.vlm"},
	{0x479E9FAEFD148EF1ULL, "vxw_encounter2_camp_kh_0_0.vlm"},
	{0x479EBC5C7D292BFBULL, "vxlt_0_lowtown_a2_n_eoq_0_0.vlm"},
	{0x479F2B218B29E523ULL, "follower_banter6172789.cl"},
	{0x479F695B958F7710ULL, "xenonreload.vsh.fxo"},
	{0x479F76A00867AB26ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_k_bk.gad"},
	{0x479FD90561BB72ACULL, "gen00fl_isabela6110297.cl"},
	{0x479FF00448554DA2ULL, "632424045.ogg"},
	{0x47A0325B45C3D0CEULL, "follower_banter6155811.cl"},
	{0x47A0793715C37A21ULL, "6105639_f.fxe"},
	{0x47A0C08A354BAA55ULL, "vxw_bonepitmining_a2_21d_0_0.vlm"},
	{0x47A150B7D868F69CULL, "hlac_lt_0_lowtown_65536_10y.dds"},
	{0x47A1CE2B1192BD8CULL, "w_blightlands_205_0.msh"},
	{0x47A1D42B1192C87EULL, "w_blightlands_205_0.mmh"},
	{0x47A1E02B1192DC15ULL, "w_blightlands_205_0.mao"},
	{0x47A205CE3E927105ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs08.mao"},
	{0x47A2AB9F4C101E24ULL, "vxlt_gallowstemplar_n_2r5_0_0.vlm"},
	{0x47A2B8031E803AB1ULL, "vxlt_0_lowtown_a2_n_c_h23_0_0.vlm"},
	{0x47A2C9F94BC65C9FULL, "vxlt_hangedman_5c_0_0.vlm"},
	{0x47A2CF8F4964AFD7ULL, "hlac_lt_undercitydungeon_-1_44z.dds"},
	{0x47A337BC09DD166FULL, "plottypes.gda"},
	{0x47A351528DFAB616ULL, "6111069_m.fxe"},
	{0x47A37028AE5DD181ULL, "gen00fl_varric6080362.cl"},
	{0x47A3E002362A5280ULL, "vxw_woundedcoast_ta_0_0.vlm"},
	{0x47A409A2CE72A270ULL, "453414070.ogg"},
	{0x47A4B70E6CAFA3C5ULL, "hlac_lt_ruinedhouse_n_-1_0y.dds"},
	{0x47A51A9AF8D2E560ULL, "hm_one_crow_4.mor"},
	{0x47A5BBB6AA1C7ECAULL, "cd_mrl150_cut_amuletritual.bnk"},
	{0x47A5E3361C0184F3ULL, "vrc221ip_servant_corpse_2.utp"},
	{0x47A6687B7FDB9C10ULL, "426209030.ogg"},
	{0x47A671A0086DC105ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_k_bk.evt"},
	{0x47A6C179145B603EULL, "mag110_vincento6016331.cl"},
	{0x47A6EE23513B8768ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_keep.phy"},
	{0x47A6F19B9F9B0DDCULL, "6048574_m.fxe"},
	{0x47A70B1CFAC73A82ULL, "zz_test.cut"},
	{0x47A711643C6E7E4BULL, "follower_banter6166681.cl"},
	{0x47A71857819B3B1CULL, "6118504_f.fxe"},
	{0x47A7BDD1DBD16024ULL, "470130427.ogg"},
	{0x47A849EE7CFAC4A4ULL, "zz_fen_debug6134381.cl"},
	{0x47A887128D61D243ULL, "242348363.ogg"},
	{0x47A8D004DE76F7FFULL, "vxlt_hangedman_n_263_0_0.vlm"},
	{0x47A8DD1117D8CE9AULL, "vxlt_0_lowtown_eia_0_0.vlm"},
	{0x47A92D3F6EDD6946ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medf.phy"},
	{0x47A9829C1C0C6FAEULL, "307242846.ogg"},
	{0x47A9FB101384EDC6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_armsup.ani"},
	{0x47AA99B138812546ULL, "vxht_keep_4ne_0_0.vlm"},
	{0x47AAA1A0CFEE6659ULL, "w_woundedcoast_192_0.msh"},
	{0x47AAA819D67C0C31ULL, "hlac_lt_gallowstemplar_n_7198_2y.dds"},
	{0x47AAABA0CFEE7747ULL, "w_woundedcoast_192_0.mmh"},
	{0x47AAAFA0CFEE7D94ULL, "w_woundedcoast_192_0.mao"},
	{0x47AAF470D11D000CULL, "vxlt_warehouse_8k_0_0.vlm"},
	{0x47AB61CA6114AECEULL, "vxlt_hangedman_n_271_0_0.vlm"},
	{0x47ABF68CCEB0F451ULL, "vxw_woundedcoast_10a_0_0.vlm"},
	{0x47ABFC79A83C27BFULL, "921187858.ogg"},
	{0x47AC35B2933CA670ULL, "hlac_lt_0_lowtown_-65537_2x.dds"},
	{0x47AC4E7FC31E525FULL, "hlac_lt_gallowstemplar_n_2382_3x.dds"},
	{0x47AC5A889604457FULL, "hlac_lt_0_lowtown_n_-196608_3y.dds"},
	{0x47AC7757D64E280FULL, "vxht_keep_n_8b3_0_0.vlm"},
	{0x47AC7D2465099CCAULL, "vxlt_0_lowtown_docks_a3_d0d_0_1.vlm"},
	{0x47AC960F675ECD94ULL, "vxlt_gallowscourtyard_n_748_0_0.vlm"},
	{0x47ACC26837CF1393ULL, "greenscreen.sub.xml"},
	{0x47AD118BEF0AA771ULL, "vxw_encounter2_camp_fo_0_0.vlm"},
	{0x47AD2DDCFB9F1737ULL, "art\\characters\\playercharacter\\humanmale\\arm_medp.mmh"},
	{0x47AD31DCFB9F1E44ULL, "art\\characters\\playercharacter\\humanmale\\arm_medp.mao"},
	{0x47AD43DCFB9F3CA9ULL, "art\\characters\\playercharacter\\humanmale\\arm_medp.msh"},
	{0x47ADAF9B8D192902ULL, "vrc221_varric6044228.cl"},
	{0x47ADCF3FEE77C292ULL, "gen_im_arm_cht_noba.uti"},
	{0x47AE1A1983F6215DULL, "hlac_ht_playershouse_n_855_20z.dds"},
	{0x47AE1E1CFACD7981ULL, "zz_test.arl"},
	{0x47AE3D01F27EEAB1ULL, "vxlt_0_lowtown_a2_n_c_erz_0_0.vlm"},
	{0x47AE6EB6BBE6A26CULL, "gen00fl_anders6125426.cl"},
	{0x47AF36A521F30C32ULL, "6124748_m.fxe"},
	{0x47AF78869E0210F5ULL, "6176173_m.fxe"},
	{0x47B01BEC8B8DBF1EULL, "vxdr_primevaltomb_7eg_0_1.vlm"},
	{0x47B0500603124035ULL, "435475382.ogg"},
	{0x47B07AAD2C46C4CBULL, "vxw_encounter1_au_0_0.vlm"},
	{0x47B0A8CD53CDCFA4ULL, "6126494_m.fxe"},
	{0x47B1C137A4FA62B8ULL, "vxca_caverns1_mines_j2d_0_0.vlm"},
	{0x47B214CE3332EAC4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeleft.msh"},
	{0x47B226CE3333092DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeleft.mao"},
	{0x47B22ACE33331066ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeleft.mmh"},
	{0x47B26EBAF4C3F6BBULL, "vxlt_0_lowtown_a2_by7_0_0.vlm"},
	{0x47B30B30543BF9FEULL, "pro000_athenril6049459.cl"},
	{0x47B3764998FF319DULL, "6117770_m.fxe"},
	{0x47B379105ABA47BEULL, "mag211_marethari6025338.cl"},
	{0x47B3ACD428CD173BULL, "vxlt_gallowscourtyard_7ce_0_0.vlm"},
	{0x47B3ACF1720F84A2ULL, "6156752_m.fxe"},
	{0x47B3D5AFCEAD79FDULL, "531077950.ogg"},
	{0x47B3FA1666F74B32ULL, "1062415535.ogg"},
	{0x47B3FDB8DE701099ULL, "vxw_encounter2_k_0_0.vlm"},
	{0x47B415D80D2A6A3AULL, "static_static_p65.psh.fxo.dx11"},
	{0x47B480B2D5BDB112ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_ice_p.mmh"},
	{0x47B495DE4451056DULL, "vxca_caverns2prc_9re_0_0.vlm"},
	{0x47B522AECDDED72BULL, "138463934.ogg"},
	{0x47B56F0AFCC27080ULL, "vxw_bonepit_1yx_0_0.vlm"},
	{0x47B580FE13AD715DULL, "vxw_sundermount_2bp_0_0.vlm"},
	{0x47B5FE646116744EULL, "follower_barks.cnv"},
	{0x47B639A1DEC843B9ULL, "mcr370_varric6208285.cl"},
	{0x47B64630AC66196CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobb_0s.dds"},
	{0x47B67D14DF673B97ULL, "w_sundermount_18_0.rml"},
	{0x47B68778A2A2A575ULL, "vxw_bonepitmining_a2_1w9_0_0.vlm"},
	{0x47B691043D8CD13EULL, "6018627_m.fxe"},
	{0x47B705BB3419BF7CULL, "vxlt_gallowsprison_n_5w9_0_0.vlm"},
	{0x47B70A10272E7BF1ULL, "1018370878.ogg"},
	{0x47B72F87EE775218ULL, "6166115_m.fxe"},
	{0x47B74DE7B1134B48ULL, "6101952_m.fxe"},
	{0x47B77C2C11458587ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetalb_sa.dds"},
	{0x47B796ADEADC3F38ULL, "6106917_m.fxe"},
	{0x47B7E31013910237ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_armsup.evt"},
	{0x47B801BB9AB3F3E0ULL, "vxht_mansion2b_1rr_0_0.vlm"},
	{0x47B81016C9E874AFULL, "clf260_harley6069670.cl"},
	{0x47B81757FAC465F1ULL, "hlac_lt_0_lowtown_a2_n_c_65538_0x.dds"},
	{0x47B8257D0564BC52ULL, "hlac_lt_0_lowtown_a2_131073_1y.dds"},
	{0x47B8D26F7C47AB1DULL, "6136933_m.fxe"},
	{0x47B8D3A4771C2BCEULL, "zz_vault_debug6203217.cl"},
	{0x47B920646118E3BEULL, "follower_barks.bnk"},
	{0x47B92C03D42E984FULL, "gen_im_trs_gen_lgt_cht_r1.uti"},
	{0x47B9368F89271F78ULL, "6177830_m.fxe"},
	{0x47B94B172022169AULL, "vxht_chantry_n_2t2_0_0.vlm"},
	{0x47B984D7925B2BB7ULL, "dre000_varric6035132.cl"},
	{0x47B9B5ABBBBC2453ULL, "vxca_cave_2fr_0_0.vlm"},
	{0x47B9D3A7530AC74EULL, "vxlt_hangedman_n_222_0_0.vlm"},
	{0x47BA6178F5D8D991ULL, "vxlt_0_lowtown_docks_a3_bff_0_0.vlm"},
	{0x47BA71B9AB8801D7ULL, "lgt160pt_wk_ave_reward.ncs"},
	{0x47BAD99FF74791FCULL, "165892074.ogg"},
	{0x47BB30E10140CC6CULL, "724767076.ogg"},
	{0x47BB6EFD1AB00393ULL, "404600751.ogg"},
	{0x47BB8368E100BFF1ULL, "vxw_encounter2_camp_an_0_0.vlm"},
	{0x47BC00632D240AA8ULL, "hlac_lt_blackemporium_1055_24x.dds"},
	{0x47BC23AF93F5ACFBULL, "mag110_samson6016180.cl"},
	{0x47BC5616F0C1E93EULL, "6052917_m.fxe"},
	{0x47BCAEDEE4750331ULL, "vxlt_0_lowtown_a2_n_c_ewy_0_0.vlm"},
	{0x47BCB84E161CFC46ULL, "6140940_m.fxe"},
	{0x47BCBD8633CE343EULL, "one350_vinell6087509.cl"},
	{0x47BCD982CAC3B636ULL, "614723773.ogg"},
	{0x47BCFC60FAC939D7ULL, "vxw_sundermount_2ew_0_0.vlm"},
	{0x47BD5236BA364A6BULL, "sw_blood_cl.mao"},
	{0x47BE83519BE89894ULL, "vxlt_0_lowtown_docks_a3_n_d4l_0_0.vlm"},
	{0x47BEB758D92A4018ULL, "6089726_m.fxe"},
	{0x47BEE410139722C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_armsup.gad"},
	{0x47BEEC8A1E3CDD4BULL, "vxw_encounter1_ft_0_0.vlm"},
	{0x47BF1C44B149E609ULL, "zz_one_debug6025812.cl"},
	{0x47BF393062C022C9ULL, "vxw_sundermount_30c_0_0.vlm"},
	{0x47BFA0E9FC70D740ULL, "vxlt_warehouse_n_xr_0_0.vlm"},
	{0x47C0B84084DDF731ULL, "hlac_lt_0_lowtown_a2_n_c_196607_0x.dds"},
	{0x47C1099FEEDF4DAFULL, "6001191_f.fxe"},
	{0x47C1313796B7EBD5ULL, "212552640.ogg"},
	{0x47C1874579193B0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blend.phy"},
	{0x47C21EB11AC32FBBULL, "vxlt_gallowscourtyard_7dd_0_0.vlm"},
	{0x47C223AEA687AA8EULL, "hlac_lt_0_lowtown_a2_-3_1z.dds"},
	{0x47C260C4C25F9B82ULL, "hlac_lt_hangedman_n_-1_22z.dds"},
	{0x47C355A35F9611DAULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm.mmh"},
	{0x47C367A35F963020ULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm.msh"},
	{0x47C3D10E5B3A16CFULL, "art\\vfx\\combat\\vfx_r_gn4_escape_act_c.impa.gad"},
	{0x47C49ADD7CA4557FULL, "mag131_keran6086186.cl"},
	{0x47C49F57CB7C7FEEULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnr.evt"},
	{0x47C4C4D0C78EE28FULL, "vxht_bartrand_n_4np_0_0.vlm"},
	{0x47C4E5FBADD34679ULL, "6211922_m.fxe"},
	{0x47C54B3C0277B3DBULL, "vxht_keep_6fv_0_0.vlm"},
	{0x47C5A7CAF2DBD9BAULL, "6085760_m.fxe"},
	{0x47C5AF62479B40B2ULL, "vxw_bonepitmining_a2_lp_0_0.vlm"},
	{0x47C5C9E67D5E6A78ULL, "gen00fl_varric6076056.cl"},
	{0x47C65310839FD255ULL, "6033146_m.fxe"},
	{0x47C673988CAA0C60ULL, "vxht_mansion2b_1us_0_0.vlm"},
	{0x47C6DDA2F6442A86ULL, "6042366_f.fxe"},
	{0x47C744BAB7CA115FULL, "w_sundermount_51_0.msh"},
	{0x47C74ABAB7CA1B01ULL, "w_sundermount_51_0.mmh"},
	{0x47C756BAB7CA2FA2ULL, "w_sundermount_51_0.mao"},
	{0x47C75B9FFCC5F1DBULL, "eye_eyelash_p0.psh.fxo.dx11"},
	{0x47C77180E6178813ULL, "34560085.ogg"},
	{0x47C7D51632300193ULL, "hlac_ca_caverns1_24423_6y.dds"},
	{0x47C7F6E84A9BE01FULL, "6020962_m.fxe"},
	{0x47C8BDC7F7BE0078ULL, "seb321_leliana6197963.cl"},
	{0x47C8C77F7435FF69ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_1y.dds"},
	{0x47C8D4F7AF03D1F3ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_f_c.impa.evt"},
	{0x47C91EF8A9BA5422ULL, "vxw_encounter3_70_0_0.vlm"},
	{0x47C9B983032C1860ULL, "lowtown_male_beggar_sta.cnv"},
	{0x47C9F7D7931E99C0ULL, "hlac_dr_0_deeproads_5849_8x.dds"},
	{0x47C9FFFF57F50F95ULL, "vxlt_warehouse_n_wz_0_0.vlm"},
	{0x47CA9657CB80DB99ULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnr.gad"},
	{0x47CACFDDE3E912DCULL, "vxht_0_hightown_c6y_0_1.vlm"},
	{0x47CAE60E5B40597CULL, "art\\vfx\\combat\\vfx_r_gn4_escape_act_c.impa.evt"},
	{0x47CAE73E5CB7FFAAULL, "gen_im_arm_cht_lgt_lore_01.uti"},
	{0x47CB04EF66C8FA33ULL, "vxca_cavern1_haunted_e4z_0_0.vlm"},
	{0x47CB0E111F64B81CULL, "660487210.ogg"},
	{0x47CB6E3DECBF5257ULL, "vxw_sundermount_2nv_0_0.vlm"},
	{0x47CBA697524A536EULL, "vxlt_warehouse_n_11d_0_0.vlm"},
	{0x47CC4D16193D1949ULL, "6101011_f.fxe"},
	{0x47CC5E2B25BE0531ULL, "6109368_m.fxe"},
	{0x47CC9AAF3B7C2548ULL, "485028560.ogg"},
	{0x47CCA4D289D5FC8DULL, "6092491_m.fxe"},
	{0x47CCB84B113F4E24ULL, "vxht_0_hightown_cae_0_0.vlm"},
	{0x47CD1F930EC2EBAEULL, "ice_shard.dds"},
	{0x47CD2BF8BD64B133ULL, "157490836.ogg"},
	{0x47CD5A967500595FULL, "vxdr_primevalentrance_223_0_0.vlm"},
	{0x47CD8C8737BA00A4ULL, "6128694_m.fxe"},
	{0x47CD9B83032FCD98ULL, "lowtown_male_beggar_sta.bnk"},
	{0x47CDFA6929C86608ULL, "706613263.ogg"},
	{0x47CDFB981B6CF41DULL, "vxht_bartrand_n_68j_0_0.vlm"},
	{0x47CE9C26D542CB26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_twitch_hands_on_hips.ani"},
	{0x47CEF617CF21B280ULL, "394119474.ogg"},
	{0x47CF2F6E34827150ULL, "lt_0_lowtown_docks_n_2_0.rml"},
	{0x47CF97B2DED3DEAAULL, "dre000_dougal6206178.cl"},
	{0x47CFCDF7AF09E434ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_f_c.impa.gad"},
	{0x47CFDC53F776148AULL, "6082178_m.fxe"},
	{0x47D01654739F065AULL, "hlac_w_bonepitdestroyed_a3_196611_0z.dds"},
	{0x47D0282478BC051BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flma_0n.dds"},
	{0x47D066D135EC04F4ULL, "541352433.ogg"},
	{0x47D069848D43D2FEULL, "6139610_m.fxe"},
	{0x47D08494FE91DC88ULL, "21148983.ogg"},
	{0x47D0CA74A985AF0EULL, "dae021_sister_colinda_sta.cnv"},
	{0x47D0D5E7076D9CA2ULL, "332023878.ogg"},
	{0x47D0D9B25717B45CULL, "vxw_woundedcoast_h9_0_0.vlm"},
	{0x47D167A95C4BA595ULL, "6034059_m.fxe"},
	{0x47D19030428D2A4EULL, "6035929_m.fxe"},
	{0x47D1A757CB8717E7ULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnr.ani"},
	{0x47D268D8189B3935ULL, "vxlt_warehouse_w7_0_0.vlm"},
	{0x47D2F2AE7874EA18ULL, "834591624.ogg"},
	{0x47D3BD18F46DCC5BULL, "vxht_keep_6ci_0_0.vlm"},
	{0x47D3EC74A9881F7EULL, "dae021_sister_colinda_sta.bnk"},
	{0x47D4213F39915932ULL, "vxw_bonepitmining_a2_kq_0_0.vlm"},
	{0x47D44B3DEF49B57BULL, "vxca_caverns1_mines_i6h_0_0.vlm"},
	{0x47D527F6B050DAC1ULL, "rdr111dg_martin6054266.cl"},
	{0x47D562D5B9B4FCD1ULL, "vxca_caverns1_fm1_0_0.vlm"},
	{0x47D56B61FF10FED4ULL, "747578963.ogg"},
	{0x47D57BE7FE811072ULL, "vxw_woundedcoast_1em_0_0.vlm"},
	{0x47D5CCC64703E741ULL, "6133079_m.fxe"},
	{0x47D5DCF7AF0E6912ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_f_c.impa.ani"},
	{0x47D5F147B482B8AAULL, "dre000_stroud6036703.cl"},
	{0x47D5F67C407F478AULL, "6120627_m.fxe"},
	{0x47D61255656908A8ULL, "dae000_bodhan6166374.cl"},
	{0x47D646409C7A4A52ULL, "sg_water_cyc.mao"},
	{0x47D68D715936D89DULL, "vxca_cavern1_haunted_j27_0_2.vlm"},
	{0x47D787D17A7B48FBULL, "178553606.ogg"},
	{0x47D7EE0E5B4AF095ULL, "art\\vfx\\combat\\vfx_r_gn4_escape_act_c.impa.ani"},
	{0x47D7F1F19CE47A98ULL, "75225643.ogg"},
	{0x47D7F4A751DDE42AULL, "vxht_bartrand_n_5q5_0_0.vlm"},
	{0x47D827CA9FA6E8CDULL, "339506955.ogg"},
	{0x47D88CD59763793BULL, "vxdr_primevaltomb_80k_0_0.vlm"},
	{0x47D8D4223C8C84CDULL, "6177466_m.fxe"},
	{0x47D8E3771C776B3FULL, "81081832.ogg"},
	{0x47D8E509CDFB5919ULL, "vxlt_hangedman_n_1rf_0_0.vlm"},
	{0x47D8F5ACD3BC2769ULL, "980103705.ogg"},
	{0x47D911147DF6447CULL, "155016839.ogg"},
	{0x47D91F03F15B5CE5ULL, "mag311_keran6041563.cl"},
	{0x47D942C5DD7F7281ULL, "244872842.ogg"},
	{0x47D9599593788BF1ULL, "vxca_cave_224_0_0.vlm"},
	{0x47D9901CA057AAD8ULL, "218590601.ogg"},
	{0x47D9D37B84418EAEULL, "vxlt_foundry_n_sr_0_1.vlm"},
	{0x47DB2A28033566A4ULL, "vxht_0_hightown_cdj_0_0.vlm"},
	{0x47DB54453BCDBEC7ULL, "6030448_m.fxe"},
	{0x47DB6CD50A81CE93ULL, "lt_0_lowtown_19_0_lc.mmh"},
	{0x47DB8426D54D2C97ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_twitch_hands_on_hips.evt"},
	{0x47DBB7175249697EULL, "gen_im_trs_gen_two_2ha_r1.uti"},
	{0x47DBF12090244D4DULL, "sp_module.ncs"},
	{0x47DC93B827B1F850ULL, "lt_0_lowtown_a3_n_c_53_0.rml"},
	{0x47DCAC953D1ED943ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockbc.mao"},
	{0x47DCB0953D1EDF90ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockbc.mmh"},
	{0x47DCBA6950060789ULL, "378287772.ogg"},
	{0x47DCBE953D1EF7AAULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockbc.msh"},
	{0x47DCF1E349E881BEULL, "403525201.ogg"},
	{0x47DD122E5C66F79BULL, "vxw_bonepitmining_a2_1p2_0_0.vlm"},
	{0x47DD3D2242A8D76DULL, "art\\characters\\creatures\\dragonling\\textures\\cn_eye_draa_0d.dds"},
	{0x47DD7FB2D5E0B4B9ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_ice_p.anb"},
	{0x47DE90259D471600ULL, "w_bonepitmining_a2_140_0.phy"},
	{0x47DEBD5367222AD2ULL, "777426252.ogg"},
	{0x47DF343DBEC26339ULL, "one120_athenril6088325.cl"},
	{0x47DFD9864E41BE15ULL, "6031359_m.fxe"},
	{0x47E0164BBB251B8AULL, "hlac_ht_0_hightown_65533_1z.dds"},
	{0x47E040BD347AE833ULL, "gen00fl_anders6125346.cl"},
	{0x47E041010E6B1D32ULL, "dae000_orsino6081385.cl"},
	{0x47E045ECA245D88CULL, "6176068_m.fxe"},
	{0x47E0DAB50A9151B5ULL, "vxlt_warehouse_x6_0_0.vlm"},
	{0x47E0ECB89FD2C1B7ULL, "vxht_bartrand_n_47g_0_0.vlm"},
	{0x47E1952FF0D4475BULL, "496114725.ogg"},
	{0x47E19741B26AEACDULL, "vxht_keep_64f_0_0.vlm"},
	{0x47E1E007533BEEE3ULL, "337356501.ogg"},
	{0x47E22D169F453D78ULL, "gen00fl_dog6176625.cl"},
	{0x47E24569BC3E640CULL, "hlac_dr_dungeon_3217_1x.dds"},
	{0x47E28526D5534D20ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_twitch_hands_on_hips.gad"},
	{0x47E2EAA246BDFCBFULL, "hlac_w_encounter1_3_0y.dds"},
	{0x47E2F4B7150EC36FULL, "6118099_f.fxe"},
	{0x47E34E53159A1278ULL, "vxht_mansion2b_1g5_0_0.vlm"},
	{0x47E36B6D605C1756ULL, "lgt000cr_hate_apostate.utc"},
	{0x47E37CDD18D764B2ULL, "66043056.ogg"},
	{0x47E3897680228A29ULL, "hlac_lt_0_lowtown_a3_n_c_0_8x.dds"},
	{0x47E58357B41D5177ULL, "vxlt_gallowscourtyard_7cq_0_0.vlm"},
	{0x47E5F4BF2B6166F2ULL, "6001355_m.fxe"},
	{0x47E600C6B06C0A87ULL, "gen00fl_fenris6127027.cl"},
	{0x47E640BDFD2A8FD8ULL, "static_static_p106.psh.fxo.dx11"},
	{0x47E6668443D3FCAAULL, "vxht_bartrand_n_5t6_0_0.vlm"},
	{0x47E74A980109D96DULL, "vxlt_0_lowtown_docks_bt7_0_0.vlm"},
	{0x47E79EFB9CBD4DDAULL, "6060435_m.fxe"},
	{0x47E7ADF4DE5254FCULL, "mag311_emile_de_launcet6043634.cl"},
	{0x47E7DB7597E67466ULL, "247701535.ogg"},
	{0x47E851F7D0AB8357ULL, "vxw_sundermount_2tp_0_0.vlm"},
	{0x47E85EC71C89D08DULL, "follower_banter6149169.cl"},
	{0x47E8BAD9BE2F5339ULL, "hm_amb_keep_bureaucrat1.mor"},
	{0x47E8C5706024EE0BULL, "488900361.ogg"},
	{0x47E974A35FB715E5ULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm.phy"},
	{0x47E99B3FAC89E227ULL, "dae000_mother6099203.cl"},
	{0x47E9B7586525B3B2ULL, "vxlt_0_lowtown_n_e89_0_0.vlm"},
	{0x47EB867A8390169CULL, "191358043.ogg"},
	{0x47EB86FD912DA681ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back_mirror.ani"},
	{0x47EC2ABAB7E9012AULL, "w_sundermount_51_0.rml"},
	{0x47EC3E6C97537E65ULL, "mag311tr_keran_cower.utt"},
	{0x47ECF1EA4C99D876ULL, "lt_gallowscourtyard_n_77_0_ll.mmh"},
	{0x47ED61B404AB4F7EULL, "vxlt_0_lowtown_docks_n_aph_0_0.vlm"},
	{0x47EDC7ED108FC2FDULL, "fx_streak02.mao"},
	{0x47EE15B2852E97E2ULL, "6049474_m.fxe"},
	{0x47EE281114D81869ULL, "20260162.ogg"},
	{0x47EE28C86595821FULL, "6064643_m.fxe"},
	{0x47EE7797C7D129A9ULL, "hlac_lt_0_lowtown_a2_n_c_-3_5x.dds"},
	{0x47EE8DDEC5859DE8ULL, "vrc221_bartrand6043275.cl"},
	{0x47EED6EAC5F58900ULL, "707472904.ogg"},
	{0x47EEDC7B528CF90BULL, "leave_07.dds"},
	{0x47EFBC5D885992DEULL, "vxlt_hangedman_t1_0_0.vlm"},
	{0x47EFF3C105E9462DULL, "follower_banter6156042.cl"},
	{0x47F0091EA461034DULL, "vxht_keep_61e_0_0.vlm"},
	{0x47F0B1C87F5F8B0FULL, "6119482_m.fxe"},
	{0x47F0BE533ED514D6ULL, "6147413_m.fxe"},
	{0x47F156C9FE58E2C5ULL, "zz_dae_debug6029119.cl"},
	{0x47F165473B842120ULL, "vxw_encounter2_camp_ga_0_0.vlm"},
	{0x47F18EE86CBDA9D0ULL, "vxlt_undercityserialkiller_1bg_0_0.vlm"},
	{0x47F191FD91322523ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back_mirror.gad"},
	{0x47F1C03007902AF8ULL, "vxht_mansion2b_1z2_0_0.vlm"},
	{0x47F1C5CD1225B399ULL, "130984143.ogg"},
	{0x47F1C7BCD2A6FDCDULL, "6147794_f.fxe"},
	{0x47F20996F90C6C43ULL, "861398071.ogg"},
	{0x47F25FA1E26D4172ULL, "vxw_woundedcoast_1ko_0_0.vlm"},
	{0x47F2679AB56143B1ULL, "6106630_f.fxe"},
	{0x47F2B0802AF01841ULL, "6139679_m.fxe"},
	{0x47F33F62ACE7827EULL, "hlac_ht_0_hightown_a3_131074_0x.dds"},
	{0x47F343C1FE91B581ULL, "vxw_woundedcoast_hh_0_0.vlm"},
	{0x47F34BBAB7EF587EULL, "w_sundermount_51_0.phy"},
	{0x47F4C3209196CC1FULL, "vxht_0_hightown_n_ca5_0_0.vlm"},
	{0x47F5412B165BA613ULL, "lt_0_lowtown_a2_n_c_26_0_ll.mmh"},
	{0x47F54561299DD5BDULL, "185789270.ogg"},
	{0x47F5630DF24E748DULL, "llac_lt_undercityencounter_-1_7.dds"},
	{0x47F58330F0FC22B8ULL, "763628903.ogg"},
	{0x47F5875CD0624FA1ULL, "359642787.ogg"},
	{0x47F5A7F1FA2BC139ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm05.msh"},
	{0x47F5B1F1FA2BD227ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm05.mmh"},
	{0x47F5C8C3B1E78A19ULL, "vxlt_hangedman_n_1dd_0_0.vlm"},
	{0x47F6481416D8C322ULL, "hlac_dr_0_deeproads_5842_7z.dds"},
	{0x47F6C3D4C2A19BD7ULL, "vxw_sundermount_2ys_0_0.vlm"},
	{0x47F6FD8BC674FF2CULL, "vxlt_gallowscourtyard_6ko_0_0.vlm"},
	{0x47F72F00535AD77BULL, "6108026_m.fxe"},
	{0x47F7A2627C4570CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\ready.evt"},
	{0x47F842EE686EF765ULL, "mag311cr_keran_civ.utc"},
	{0x47F868A002F295C1ULL, "41195083.ogg"},
	{0x47F86EFD91380808ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_back_mirror.evt"},
	{0x47F8756ADF8E29ABULL, "558379857.ogg"},
	{0x47F8935BBF9EC7A4ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatiki.phy"},
	{0x47F8AAA99465457DULL, "dae000ip_crowd_lm2_talk12.utp"},
	{0x47F8FA4B6A60C03AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoortrim_1.mao"},
	{0x47F95CA33C0D6811ULL, "6054008_m.fxe"},
	{0x47F97D311E1B62FFULL, "328176372.ogg"},
	{0x47F9C225B7ACADC8ULL, "6140942_m.fxe"},
	{0x47FA29A13ED69EA3ULL, "vxca_caverns2_8cl_0_0.vlm"},
	{0x47FA637F52B0A9F8ULL, "ave100_donnic.stg"},
	{0x47FAC1DB6C5B37AEULL, "mag320_orsino6071207.cl"},
	{0x47FAF00EA8B93488ULL, "322204661.ogg"},
	{0x47FB1E812752194DULL, "61593854.ogg"},
	{0x47FB306E34A7ADB8ULL, "lt_0_lowtown_docks_n_2_0.anb"},
	{0x47FB4DBB67D828CFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_cth_homa_0s.dds"},
	{0x47FB530195F36AFFULL, "vxht_0_hightown_a3_9xf_0_0.vlm"},
	{0x47FBE562E892CD48ULL, "qun131tr_warrens_thug_amb.utt"},
	{0x47FBE6B2D86DAB43ULL, "mrl350ip_at_downmountain.ncs"},
	{0x47FBECAC852EF10BULL, "6166308_f.fxe"},
	{0x47FC0D46E921291AULL, "6069230_m.fxe"},
	{0x47FC304588040634ULL, "llac_ht_keep_3732_13.dds"},
	{0x47FC74142FF8F30FULL, "dae011_badpoet_sta6146931.cl"},
	{0x47FC7D281921F8A4ULL, "mag110_ws_banter6079649.cl"},
	{0x47FCF0A4EFE9BF57ULL, "follower_banter6155997.cl"},
	{0x47FD901C485C4A0AULL, "6074456_m.fxe"},
	{0x47FDED8A0F8CB078ULL, "6175865_m.fxe"},
	{0x47FE4555FB64BD7BULL, "vxlt_0_lowtown_docks_n_bg7_0_0.vlm"},
	{0x47FE87A4B7AA8BBBULL, "133066259.ogg"},
	{0x47FE99627C4B8079ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\ready.gad"},
	{0x47FF61162DF820FCULL, "demo_weapon_swap6207148.cl"},
	{0x47FF94095519C963ULL, "946227676.ogg"},
	{0x47FFAEB54E227DB6ULL, "6134247_f.fxe"},
	{0x47FFD22109D39BC1ULL, "fx_firetwister_add.mao"},
	{0x47FFD7242D7A39A0ULL, "vxw_encounter2_camp_bn_0_0.vlm"},
	{0x47FFDAA0FB6E4A3CULL, "629344062.ogg"},
	{0x47FFDCE60193FA37ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_012a_0s.dds"},
	{0x48000187F39794A8ULL, "hlac_ca_caverns1_mines_24426_6z.dds"},
	{0x4800074F82D399B8ULL, "root_mage_bloodmage.dds"},
	{0x48001C6572C2D25FULL, "hlac_ca_caverns1_24422_2z.dds"},
	{0x48003CC6F572F63AULL, "fx_exp_cloud.mao"},
	{0x480052EDF81F063FULL, "gen00fl_fenris6107151.cl"},
	{0x4800AF6ECBDD176BULL, "vxw_bonepitmining_a2_1um_0_0.vlm"},
	{0x48010FB3C711893DULL, "vxca_cavern1_haunted_hm6_0_0.vlm"},
	{0x480115CD84317619ULL, "479042216.ogg"},
	{0x4801C82E0E3CD57BULL, "mag160_feynriel6020082.cl"},
	{0x48022244FF7AE203ULL, "vxw_encounter2_ba_0_0.vlm"},
	{0x4802671198098277ULL, "vxlt_gallowscourtyard_7is_0_0.vlm"},
	{0x48028AAEA71CDC64ULL, "490321146.ogg"},
	{0x4802D3C5D406E3A1ULL, "hlac_ca_caverns1_24424_11z.dds"},
	{0x4803233ECFA3964EULL, "6173894_m.fxe"},
	{0x480336281593D55DULL, "164780793.ogg"},
	{0x4803566C545F45A3ULL, "m_dae110ar_lowtown_nt.wbd"},
	{0x480383F9D77AA808ULL, "542293377.ogg"},
	{0x480384C22FC5399BULL, "dae300_alistair_king6133540.cl"},
	{0x48038F0056925F21ULL, "llac_ca_cavern1_haunted_24426_14.dds"},
	{0x4803AD259D666395ULL, "w_bonepitmining_a2_140_0.msh"},
	{0x4803AF259D66670BULL, "w_bonepitmining_a2_140_0.mmh"},
	{0x4803BB259D667BA8ULL, "w_bonepitmining_a2_140_0.mao"},
	{0x4803E2FE5BAB4E54ULL, "vxlt_0_lowtown_docks_a3_d1d_0_0.vlm"},
	{0x4803EAE56186EC5EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_posting_dp.dds"},
	{0x4804541D073254DBULL, "vxht_0_hightown_n_bad_0_0.vlm"},
	{0x48046EB6A067DFE3ULL, "hlac_lt_ruinedhouse_n_-1_1x.dds"},
	{0x480472E1293A0734ULL, "381666332.ogg"},
	{0x4804895FE7DDC890ULL, "395690819.ogg"},
	{0x4804AA627C5008C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\ready.ani"},
	{0x480537C763D5D183ULL, "260883477.ogg"},
	{0x48055A16746850DFULL, "vxw_encounter2_pn_0_1.vlm"},
	{0x48056B86504F5C35ULL, "6173772_m.fxe"},
	{0x4805C39BF5474157ULL, "hlac_lt_0_lowtown_-131073_3y.dds"},
	{0x4806047EF00CF63BULL, "vxw_bonepitdestroyed_a3_1uc_0_0.vlm"},
	{0x480633ED4F7CE4CEULL, "6072348_m.fxe"},
	{0x480747377E2EF86CULL, "6202066_m.fxe"},
	{0x4807D8E82320E93AULL, "vxdr_primevaltomb_91e_0_1.vlm"},
	{0x480805B3DADCFD55ULL, "6053852_m.fxe"},
	{0x4808937C1F991984ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_coma_0d.dds"},
	{0x480894B827D734B8ULL, "lt_0_lowtown_a3_n_c_53_0.anb"},
	{0x4808FE34BF4AAB61ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_ice_c.impa.ani"},
	{0x48090FC296843A6CULL, "hlac_lt_0_lowtown_n_65537_0x.dds"},
	{0x480910B3B43FEC06ULL, "dae000ip_crowd_ff5_talk11.utp"},
	{0x4809387F803B76D8ULL, "6076485_m.fxe"},
	{0x4809884EFF110F21ULL, "follower_banter6170730.cl"},
	{0x480AA37DCFA12E14ULL, "82001418.ogg"},
	{0x480B3DCB4CE30C13ULL, "zz_qun_debug6085812.cl"},
	{0x480BB9D860AF8EE5ULL, "llac_lt_undercitydungeon_-1_35.dds"},
	{0x480BE7280025C66EULL, "vxht_0_hightown_a2_9pb_0_0.vlm"},
	{0x480C73EDC1CBE847ULL, "201878717.ogg"},
	{0x480C9AFAE64B076BULL, "6036521_m.fxe"},
	{0x480CA4BE65D19EB9ULL, "hlac_ca_caverns1_24423_7x.dds"},
	{0x480CE3E0EF343EB7ULL, "zz_approval_debug6003105.cl"},
	{0x480D9582634BBC0FULL, "6113282_m.fxe"},
	{0x480DF57B52A7D4EDULL, "leave_07.mao"},
	{0x480DF90A2EA765A5ULL, "vxw_woundedcoast_oq_0_0.vlm"},
	{0x480E6848E2689594ULL, "6118420_m.fxe"},
	{0x480E720AD231E262ULL, "hlac_lt_house_-1_4z.dds"},
	{0x480E94E77243B24AULL, "6137641_m.fxe"},
	{0x480EEB593C7EFA14ULL, "hlac_dr_0_deeproads_17035_0z.dds"},
	{0x480F435BC6597272ULL, "vxw_woundedcoast_1qi_0_0.vlm"},
	{0x480FCC9291A9ABDCULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.dur.ani"},
	{0x48100934BF50DC03ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_ice_c.impa.gad"},
	{0x481009B26BEB5B36ULL, "6033318_m.fxe"},
	{0x4810223D912DB004ULL, "6014878_m.fxe"},
	{0x481075FFD04FC5C4ULL, "6107176_m.fxe"},
	{0x481080C9CC5DF408ULL, "6060266_m.fxe"},
	{0x48111E03D881EA57ULL, "6133460_m.fxe"},
	{0x481170A63865314BULL, "vxlt_hangedman_1vo_0_0.vlm"},
	{0x4812724AABB5E5C3ULL, "49680064.ogg"},
	{0x481292E0937CFF56ULL, "6015952_m.fxe"},
	{0x4812A02ED6EC22EDULL, "vxlt_0_lowtown_docks_be4_0_0.vlm"},
	{0x4812C5F9F9286D5BULL, "vxht_0_hightown_n_blg_0_0.vlm"},
	{0x4812D9337421B429ULL, "6089482_m.fxe"},
	{0x4812F22B1E838967ULL, "41133283.ogg"},
	{0x48130CD05E66FE5AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_swordflame.phy"},
	{0x4813B29AA204A939ULL, "vfx_chunkyparticle_p0.psh.fxo.dx11.assembly"},
	{0x4813C185246623EBULL, "135077996.ogg"},
	{0x4813CD3E44C6B19EULL, "character_character_p70.psh.fxo.dx11"},
	{0x4813D00A23A147CFULL, "dae000_gen_turn_in_3.bnk"},
	{0x481401C59792C742ULL, "mag120_macha6091379.cl"},
	{0x4814299AC645BDCEULL, "sky_ht_mansion2.mao"},
	{0x481469D70FA23DFEULL, "vxlt_warehouse_p1_0_0.vlm"},
	{0x48148BE64ACF7466ULL, "6135171_m.fxe"},
	{0x48151468514CBF54ULL, "vxlt_darktown_2vr_0_0.vlm"},
	{0x48151644BE6E0153ULL, "plt_cod_ltr_alain_mercy.plo"},
	{0x481549A7BE282623ULL, "6036662_m.fxe"},
	{0x4815CEE8D6DD9608ULL, "vxlt_warehouse_n_1bw_0_0.vlm"},
	{0x4815D99291AE2D1EULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.dur.gad"},
	{0x4815DFA54CE0EFBBULL, "vxca_cavern1_haunted_i2l_0_0.vlm"},
	{0x4815E49F04FD6338ULL, "gen00fl_bethany6109224.cl"},
	{0x4815FFFDDAE8411CULL, "vxlt_hangedman_2mv_0_0.vlm"},
	{0x4816285227473F05ULL, "hlac_ht_mansion_n_3382_1x.dds"},
	{0x4816410776320BEFULL, "263976547.ogg"},
	{0x4816757A27F3EDBDULL, "nat380_nathaniel6070252.cl"},
	{0x4816E634BF56BFE8ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_ice_c.impa.evt"},
	{0x4816F15B33DC6AD7ULL, "6131909_m.fxe"},
	{0x4816F20A23A3B737ULL, "dae000_gen_turn_in_3.cnv"},
	{0x481731E6B34E36ACULL, "zz_mer_debug6045138.cl"},
	{0x48176B3AF8A1C5C3ULL, "character_character_p127.psh.fxo.dx11"},
	{0x48176E70FDE9F487ULL, "vxlt_0_lowtown_a2_n_c_hv9_0_0.vlm"},
	{0x4817AFDE8D2FE3D7ULL, "77816299.ogg"},
	{0x4818182F2710801FULL, "vxlt_gallowstemplar_46w_0_0.vlm"},
	{0x481843EA2897A6FDULL, "254828330.ogg"},
	{0x4819130310E62EDCULL, "vxw_woundedcoast_q2_0_0.vlm"},
	{0x4819811588B16FD6ULL, "lt_0_lowtown_docks_a3_n_147_0.msh"},
	{0x48198B1588B180D4ULL, "lt_0_lowtown_docks_a3_n_147_0.mmh"},
	{0x48198F1588B187E7ULL, "lt_0_lowtown_docks_a3_n_147_0.mao"},
	{0x4819FE43E2A3FAC8ULL, "936136523.ogg"},
	{0x481A1E368FF4A4F1ULL, "nat380_temmerin6067378.cl"},
	{0x481AEE4064B30967ULL, "6113527_m.fxe"},
	{0x481C0521F19D0DC4ULL, "6168154_f.fxe"},
	{0x481C26CB49249D06ULL, "6172725_m.fxe"},
	{0x481C4992CFE7DD9AULL, "zz_vrc_debug6200231.cl"},
	{0x481C518FB42F2AB0ULL, "follower_banter6164053.cl"},
	{0x481C6AE7209D7E25ULL, "vxw_woundedcoast_dp_0_0.vlm"},
	{0x481D6A6C06CB255FULL, "ts_profane_acc_props.gda"},
	{0x481DB7953D5682ABULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockbc.phy"},
	{0x481DDDE2AF92A47AULL, "genip_barrel.utp"},
	{0x481DE1B9022578D9ULL, "6085213_m.fxe"},
	{0x481DEF22DC7D0B53ULL, "298902924.ogg"},
	{0x481E8AB2E22DFD0FULL, "art\\characters\\creatures\\strider\\animation\\attkfr_arms.gad"},
	{0x481EDED44138F53AULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.impa.gad"},
	{0x481F05FEE3671303ULL, "vxw_encounter2_hc_0_0.vlm"},
	{0x481F75C280D796CDULL, "81609339.ogg"},
	{0x481F8BB71DBA4CD3ULL, "seb201_lady_harimann6177575.cl"},
	{0x481FAB7554907CF0ULL, "609406430.ogg"},
	{0x48205FED8A0975AFULL, "static_static_p44.psh.fxo.dx11"},
	{0x482064CBB4D71BA6ULL, "follower_banter6163735.cl"},
	{0x4820A01C678885DFULL, "vxca_caverns1_jfj_0_0.vlm"},
	{0x4820EAE732971DCEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_cmov_rf.evt"},
	{0x482160CF86F40540ULL, "6084707_m.fxe"},
	{0x482181390015EE89ULL, "narrative_banter6165521.cl"},
	{0x4821854FA85A88CDULL, "221692889.ogg"},
	{0x48218E4AD27500DEULL, "hlac_lt_0_lowtown_docks_a3_n_-65536_0x.dds"},
	{0x48219AF1FA50E65CULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm05.phy"},
	{0x48219B8ECFF81E5BULL, "llac_lt_0_lowtown_a2_n_-65536_1.dds"},
	{0x4821AB83A1CA1CF6ULL, "306365090.ogg"},
	{0x4821AE6F0AC9A7A9ULL, "6147803_m.fxe"},
	{0x482253229C5748ACULL, "vxlt_gallowscourtyard_6xb_0_0.vlm"},
	{0x4822BC34F3D655FBULL, "ser120_ghyslain6118396.cl"},
	{0x4822D9FF6A798594ULL, "vxht_0_hightown_a3_7ix_0_0.vlm"},
	{0x4822E838D3F9273BULL, "vxw_bonepitdestroyed_a3_1ce_0_0.vlm"},
	{0x482350FCB56B25C4ULL, "meleering3.gda"},
	{0x482386454342D7D4ULL, "vxlt_darktown_2ys_0_0.vlm"},
	{0x4823943C8F3A33EDULL, "hlac_lt_gallowscourtyard_-65536_2y.dds"},
	{0x48244D5A8DA4469DULL, "hlac_lt_0_lowtown_docks_a3_131072_0x.dds"},
	{0x48249FB2E2328CBCULL, "art\\characters\\creatures\\strider\\animation\\attkfr_arms.evt"},
	{0x4824E62978760D82ULL, "6166900_m.fxe"},
	{0x48251419DFC460FEULL, "113816668.ogg"},
	{0x482560B93EA849BBULL, "plt_cod_crt_dragon.plo"},
	{0x482575E2685D1BA3ULL, "vxht_0_hightown_bou_0_0.vlm"},
	{0x48257F3814B8E823ULL, "vxca_caverns2_8pq_0_0.vlm"},
	{0x482590D50C9950BAULL, "65361726.ogg"},
	{0x4825DB296906623BULL, "693425008.ogg"},
	{0x4825E0242DC7CC3FULL, "llac_lt_0_lowtown_a2_65537_4.dds"},
	{0x4825E04DEFE00D07ULL, "vxlt_0_lowtown_a2_n_c_hq8_0_0.vlm"},
	{0x4825E1D4413F1951ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.impa.evt"},
	{0x4826529287EE8987ULL, "follower_banter6173135.cl"},
	{0x48265E621E34C0CBULL, "dae000_arishok6068539.cl"},
	{0x4827C4DF2D808C13ULL, "lowtown_male_beggar_sta6151075.cl"},
	{0x4827E1E7329D2D79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_cmov_rf.gad"},
	{0x4827FB2FAD1B9499ULL, "545466922.ogg"},
	{0x4828358B4E540AA8ULL, "6052115_m.fxe"},
	{0x48288FACA45E8DF6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wo_basic.msh"},
	{0x482899ACA45E9EF4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wo_basic.mmh"},
	{0x48289DACA45EA507ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wo_basic.mao"},
	{0x4828E6BD37700F74ULL, "467763829.ogg"},
	{0x4829486C547FFD51ULL, "m_dae110ar_lowtown_nt.bnk"},
	{0x48296925896717F7ULL, "vxw_bonepitdestroyed_a3_1pt_0_0.vlm"},
	{0x4829E632D86179D5ULL, "vxht_mansion2b_2e9_0_0.vlm"},
	{0x482A81F45BA3B4D8ULL, "vxlt_0_lowtown_ext_0_0.vlm"},
	{0x482AAAA41FEE429BULL, "728583295.ogg"},
	{0x482B14F734F2C5FEULL, "6112169_m.fxe"},
	{0x482B1A24213924B3ULL, "518982295.ogg"},
	{0x482B51EDBF06A370ULL, "606407715.ogg"},
	{0x482B9DC66BC5CCA7ULL, "aip_highdragon_base.gda"},
	{0x482BBD05B906968EULL, "hlac_ht_keep_n_4519_16x.dds"},
	{0x482BDE0ED5393756ULL, "793479140.ogg"},
	{0x482C3CB0740D6B66ULL, "lgt310tr_melson.utt"},
	{0x482C3F4A70412F75ULL, "964969075.ogg"},
	{0x482C417EFA51BC6FULL, "art\\vfx\\materials\\vfx_d_blast.mao"},
	{0x482CA96840DC449BULL, "775822354.ogg"},
	{0x482CE79867649286ULL, "935292107.ogg"},
	{0x482D27FB89BBFB4CULL, "art\\characters\\playercharacter\\humanfemale\\hf_hoo_robc.mao"},
	{0x482D5B7F52DBF5CBULL, "ave100_donnic.bnk"},
	{0x482DF2E732A1B5C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_cmov_rf.ani"},
	{0x482DF7D99C97AE39ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_dsmla_0t.dds"},
	{0x482E190991EEA2B8ULL, "vxlt_warehouse_vm_0_0.vlm"},
	{0x482E6AFD9576ABA5ULL, "llac_lt_0_lowtown_docks_-196610_3.dds"},
	{0x482E9B78BAA5C366ULL, "abi_w_global.bnk"},
	{0x482EB092D8E319BDULL, "vxca_caverns2prc_adk_0_0.vlm"},
	{0x482EFDDF51DC7F22ULL, "vxht_keep_n_6i9_0_0.vlm"},
	{0x482F1A72B32B8169ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v2.ani"},
	{0x482F4E01738C1F73ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_com_c.anb"},
	{0x482F8E5B2C39DDCEULL, "6135070_m.fxe"},
	{0x48305D7F52DE2EABULL, "ave100_donnic.cnv"},
	{0x483135751900E13BULL, "2p_200_one280_iwan.stg"},
	{0x48315A15C5EF3FBBULL, "vxw_bonepitdestroyed_a3_1dd_0_0.vlm"},
	{0x48320A0F0F550C3AULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.cess.gad"},
	{0x48326904F43546D8ULL, "vxca_cave_1lx_0_0.vlm"},
	{0x4832A7B2E23ED7D5ULL, "art\\characters\\creatures\\strider\\animation\\attkfr_arms.ani"},
	{0x4832B57B497A6BCFULL, "6096754_m.fxe"},
	{0x4832C09C5523283BULL, "6100772_m.fxe"},
	{0x4832F9D44149CC20ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.impa.ani"},
	{0x4833CA058C3A6DE1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangsign_d.dds"},
	{0x48343371B29AB247ULL, "vxlt_0_lowtown_docks_a3_cz8_0_1.vlm"},
	{0x48347728439BF1A7ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.cess.gad"},
	{0x4834B3630B85CE4BULL, "vxlt_gallowsprison_n_5re_0_0.vlm"},
	{0x4834B93B4EBF5916ULL, "zz_qun_debug6083563.cl"},
	{0x4834CC86008D5755ULL, "6034918_m.fxe"},
	{0x4834E1F293D1F4BDULL, "mer221_sabin6065652.cl"},
	{0x4834E4C01CAF1DC8ULL, "6119679_f.fxe"},
	{0x48350E4B0C73FC46ULL, "6043726_m.fxe"},
	{0x48354A857EC347BCULL, "194946601.ogg"},
	{0x483597C0A6B0EE74ULL, "mag291_vincento6152606.cl"},
	{0x4836129C2E00D7E5ULL, "6121314_f.fxe"},
	{0x48363572B331CDEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v2.gad"},
	{0x4836418DF0790FC7ULL, "vxca_caverns2_7tw_0_1.vlm"},
	{0x483760BBDAB2F385ULL, "character_character_p130.psh.fxo"},
	{0x48378B4046F330EEULL, "vxlt_0_lowtown_a3_n_c_er4_0_0.vlm"},
	{0x4838325E808EED31ULL, "6197877_m.fxe"},
	{0x48388B43ACB73F11ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\prp_secretwheeldoor_n.dds"},
	{0x4838B836B14DB1F8ULL, "323791041.ogg"},
	{0x4838B9FE2B428226ULL, "gen_im_arm_hlm_nobc.uti"},
	{0x4838BEFB53F20F0CULL, "mag160_feynriel6020069.cl"},
	{0x4838E461546AA9B5ULL, "6145184_m.fxe"},
	{0x4838FC4EE28897AFULL, "6106886_m.fxe"},
	{0x48390D0F0F5B3051ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.cess.evt"},
	{0x48391860638635DFULL, "6127699_f.fxe"},
	{0x48394E1A2C871D01ULL, "6155314_m.fxe"},
	{0x483A26925250FA36ULL, "montage_subs_demohack6217114.cl"},
	{0x483A81286DAB1077ULL, "6117916_m.fxe"},
	{0x483ABB903F188D96ULL, "127892445.ogg"},
	{0x483ABDF59ED1C38AULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgta_0n.dds"},
	{0x483B146DE70A5286ULL, "vxlt_0_lowtown_a2_n_eqv_0_0.vlm"},
	{0x483B6C2843A1FDB4ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.cess.evt"},
	{0x483B9DF6C502351AULL, "lt_0_lowtown_docks_3_0.rml"},
	{0x483C0C6FB5A931CCULL, "hlac_ht_bartrand_n_2995_0x.dds"},
	{0x483C11FE9E94E9ECULL, "6128948_m.fxe"},
	{0x483C9523976CD10AULL, "seb201_lady_harimann6177367.cl"},
	{0x483CFFF771F58A08ULL, "hlac_lt_backalley_n_0_11y.dds"},
	{0x483D226FCAD9323DULL, "vxca_caverns2prc_acj_0_0.vlm"},
	{0x483D3272B337E750ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v2.evt"},
	{0x483D5FA6C721CE2DULL, "vxht_mansion2b_2id_0_1.vlm"},
	{0x483D6B0173988DDCULL, "art\\vfx\\combat\\vfx_w_2nd_atk_com_c.mmh"},
	{0x483D9385F80C4CD9ULL, "mag110_vincento6016297.cl"},
	{0x483D949F3A9022D2ULL, "qun160_party_comment_16061227.cl"},
	{0x483DA85E76B2F686ULL, "hlac_ht_playershouse_n_855_16x.dds"},
	{0x483DAE992D5A4213ULL, "plt_cod_ltr_isabelaapology.plo"},
	{0x483DB8CC6BD690E2ULL, "232480696.ogg"},
	{0x483E56346907C69AULL, "ht_0_hightown_a3_29_0_ll.mmh"},
	{0x483FBE717DF13644ULL, "vxht_mansion_n_3zo_0_0.vlm"},
	{0x4840171C707AF7EBULL, "qun160_vashothlead6053899.cl"},
	{0x48406C760D9BCE85ULL, "lt_0_lowtown_a2_38_0.anb"},
	{0x48406D238067C50CULL, "vxht_keep_n_62r_0_0.vlm"},
	{0x48408DEABAF9F049ULL, "hlac_lt_0_lowtown_a3_n_c_65536_9z.dds"},
	{0x4840A909D7B891A0ULL, "vxlt_0_lowtown_a3_n_c_gy9_0_0.vlm"},
	{0x4840F007DBDD539BULL, "art\\characters\\playercharacter\\humanmale\\hlm_masc.mao"},
	{0x4840F498942FF692ULL, "mag311_huon6042700.cl"},
	{0x48410A58D86A2E2BULL, "vxca_caverns1_mines_i99_0_0.vlm"},
	{0x48411959B882487CULL, "vxlt_hangedman_173_0_0.vlm"},
	{0x4841476C979BCD28ULL, "mag311tr_keran_cower.ncs"},
	{0x4841572A2C9162A6ULL, "isa311_hawke_comments6117911.cl"},
	{0x484175B009C39009ULL, "mag370_meredith6064977.cl"},
	{0x4842058D6F748956ULL, "201087040.ogg"},
	{0x484261C0068BE147ULL, "6127021_m.fxe"},
	{0x4842ED745CB60085ULL, "915466921.ogg"},
	{0x48437D0922BCB135ULL, "llac_lt_0_lowtown_a3_n_c_65534_1.dds"},
	{0x48438C524FC1E57FULL, "vxht_0_hightown_a3_9am_0_0.vlm"},
	{0x4843BC96B72B2FD7ULL, "379893057.ogg"},
	{0x4843FE8B6973B9E6ULL, "gen00fl_fenris6127099.cl"},
	{0x4844324A1D66AE94ULL, "zz_followers_banter6159411.cl"},
	{0x48444CEF81E8D9D6ULL, "vxht_mansion_n_59b_0_0.vlm"},
	{0x48447CA4C44AC2E6ULL, "vxlt_0_lowtown_a3_n_c_gdh_0_0.vlm"},
	{0x4844DDE6262A82B1ULL, "6051609_m.fxe"},
	{0x4844F65F7FC06FB9ULL, "49386038.ogg"},
	{0x4844FD44BE96FAF0ULL, "plt_cod_ltr_alain_mercy.nss"},
	{0x4845B3CEE9365F6EULL, "art\\characters\\creatures\\dragon\\animation\\cwkf.ani"},
	{0x4845B607B7A27499ULL, "vxca_cavern1_haunted_egq_0_0.vlm"},
	{0x4845C94993B80599ULL, "w_bonepit_249_0.msh"},
	{0x4845D34993B81687ULL, "w_bonepit_249_0.mmh"},
	{0x4845D74993B81DD4ULL, "w_bonepit_249_0.mao"},
	{0x48460225D49BE1EDULL, "vxlt_0_lowtown_a3_n_c_enw_0_0.vlm"},
	{0x4846250F0F65E320ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.cess.ani"},
	{0x484650172D033FBDULL, "6111353_f.fxe"},
	{0x4846BB2994892E9AULL, "485265840.ogg"},
	{0x4846DDA72333B8FFULL, "77462708.ogg"},
	{0x48470E060CE3D45EULL, "857378981.ogg"},
	{0x484765AE3F8FE5D8ULL, "vxlt_0_lowtown_enr_0_0.vlm"},
	{0x4847A7E79424B2FEULL, "6072202_f.fxe"},
	{0x4847D521420E3942ULL, "vxlt_backalley_n_1jl_0_0.vlm"},
	{0x48485EC439530B3DULL, "art\\vfx\\materials\\vfx_wis_smo_blend.mao"},
	{0x4848F25B4D77E144ULL, "6059226_m.fxe"},
	{0x4849642843AE2DEDULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.cess.ani"},
	{0x4849C0B94BD90FC0ULL, "gen00fl_anders6125848.cl"},
	{0x484A0185E49CF62BULL, "818918262.ogg"},
	{0x484A834428E8C2ADULL, "hlac_lt_0_lowtown_-131073_2x.dds"},
	{0x484AD125E28D384EULL, "gen00fl_dog6166424.cl"},
	{0x484ADDA9CB21CD25ULL, "vxw_bonepit_1ww_0_0.vlm"},
	{0x484AE239E319817DULL, "809963471.ogg"},
	{0x484AEC05E9FD87C9ULL, "one000_ilen6037080.cl"},
	{0x484B081E2BC69423ULL, "vxlt_gallowstemplar_5jz_0_0.vlm"},
	{0x484B86EE309FE971ULL, "gen00fl_anders6135767.cl"},
	{0x484B9C66F81AAB14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flourish.ani"},
	{0x484CC246F791DDE6ULL, "284552022.ogg"},
	{0x484CC9A5F1596A95ULL, "gencr_darkspawn_emissary_tr.utc"},
	{0x484CDCCD8AC4AF73ULL, "6171026_m.fxe"},
	{0x484D0A1A088C9040ULL, "ts_qunari_prog_junk1.gda"},
	{0x484D42C389FF84FFULL, "636960713.ogg"},
	{0x484D8509D1B24D81ULL, "6112904_m.fxe"},
	{0x484DABE356940F0DULL, "6084900_m.fxe"},
	{0x484DCB0A03E19358ULL, "hlac_w_blightlands_196611_0y.dds"},
	{0x484E09553EAEEDA4ULL, "696899463.ogg"},
	{0x484ED75C16F31C2FULL, "6078748_f.fxe"},
	{0x484F84ABD451F73DULL, "hlac_lt_0_lowtown_a2_65533_0z.dds"},
	{0x48500C3364368655ULL, "24095128.ogg"},
	{0x48507A3A2C236CC9ULL, "lt_0_lowtown_docks_n_159_0.mao"},
	{0x4850863A2C238162ULL, "lt_0_lowtown_docks_n_159_0.mmh"},
	{0x4850883A2C238418ULL, "lt_0_lowtown_docks_n_159_0.msh"},
	{0x4850EBB2D8B5F3DEULL, "mrl350ip_at_downmountain.utp"},
	{0x4850F5DB01AB9B98ULL, "305696803.ogg"},
	{0x485164F7988D28C8ULL, "6053897_m.fxe"},
	{0x485176C0E7F220F0ULL, "4864385.ogg"},
	{0x48517BDD7012DCF5ULL, "489023519.ogg"},
	{0x4851B0EE338A3F24ULL, "903829418.ogg"},
	{0x4851BEBC208FA7B0ULL, "vxht_brothel_n_3hn_0_0.vlm"},
	{0x48520CBCA3A9D867ULL, "6108453_m.fxe"},
	{0x4852108EDB8A1402ULL, "rdr110dg_guard_warehouse6016721.cl"},
	{0x48529966F820C576ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flourish.gad"},
	{0x4852ABCEE940DBBFULL, "art\\characters\\creatures\\dragon\\animation\\cwkf.evt"},
	{0x4853326653AE2D56ULL, "6177667_m.fxe"},
	{0x485346C9134A84EDULL, "vxca_caverns1_k2q_0_0.vlm"},
	{0x48538AEE9E75B46FULL, "6065929_m.fxe"},
	{0x4853A794DB38962CULL, "vxlt_0_lowtown_n_h1y_0_0.vlm"},
	{0x4853D504DDF8E724ULL, "481170737.ogg"},
	{0x4854538962557E00ULL, "vxw_bonepit_10o_0_0.vlm"},
	{0x48547402C691FA6DULL, "vxlt_0_lowtown_a3_n_c_eqv_0_0.vlm"},
	{0x4854E5FEDD232B23ULL, "6175387_m.fxe"},
	{0x4855210EA6A493F8ULL, "6088233_m.fxe"},
	{0x48555E0651B5A331ULL, "vxlt_hangedman_n_218_0_0.vlm"},
	{0x48556C9E448CEADDULL, "dae300st_solivitus_plus.utm"},
	{0x485595A5A5DBC1DEULL, "vxht_mansion2_2kp_0_0.vlm"},
	{0x4855B98361BBAD29ULL, "llac_ht_brothel_n_3371_2.dds"},
	{0x4855D78B3185FE58ULL, "vxlt_0_lowtown_eis_0_0.vlm"},
	{0x48569382FB05270CULL, "ser221_gascard6165645.cl"},
	{0x4856CD53DB72F86BULL, "6108951_f.fxe"},
	{0x48575483724B9347ULL, "77471119.ogg"},
	{0x485798AB714BD131ULL, "711557199.ogg"},
	{0x4857D462E8E10F75ULL, "qun131tr_warrens_thug_amb.ncs"},
	{0x48584FF4384CDC07ULL, "vxlt_gallowstemplar_6s3_0_1.vlm"},
	{0x48587EA66F3E5AC3ULL, "743510438.ogg"},
	{0x4858983904262B9CULL, "hlac_lt_0_lowtown_a3_n_c_-2_2z.dds"},
	{0x4858F1BD976B55D5ULL, "6053873_f.fxe"},
	{0x48592AE3FEE532A4ULL, "vxlt_0_lowtown_n_ei8_0_0.vlm"},
	{0x4859770CCB913469ULL, "519316251.ogg"},
	{0x48598CCEE946C5A8ULL, "art\\characters\\creatures\\dragon\\animation\\cwkf.gad"},
	{0x4859B466F82711D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\flourish.evt"},
	{0x4859D3BD308C7164ULL, "6049465_m.fxe"},
	{0x485A219662529DFBULL, "205008624.ogg"},
	{0x485B2D9A0EEE0FCCULL, "vxw_sundermount_37c_0_0.vlm"},
	{0x485B53A21A23D730ULL, "370968502.ogg"},
	{0x485B8B2C04ACF775ULL, "follower_banter6151245.cl"},
	{0x485C0F6D4E4D3C4EULL, "vxw_woundedcoast_1dn_0_0.vlm"},
	{0x485C5AA832B44017ULL, "vxw_encounter2_camp_8w_0_0.vlm"},
	{0x485CAECD79D19EF0ULL, "hairalpha_rendertoshadowmap_p0.psh.fxo.dx11"},
	{0x485CF1CEA55F5F3AULL, "700537695.ogg"},
	{0x485D4D239CFA290AULL, "hlac_w_sundermount_-327675_0x.dds"},
	{0x485D52F68A85DFD6ULL, "6096409_m.fxe"},
	{0x485D5E2E79879251ULL, "653630341.ogg"},
	{0x485D7750E34C6F87ULL, "vxlt_warehouse_15v_0_0.vlm"},
	{0x485DC4124F3DCC13ULL, "vxlt_backalley_g_1l6_0_0.vlm"},
	{0x485E22B8FD04E834ULL, "mag131_taraneh6094548.cl"},
	{0x485E5184F4D722EAULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_vial_d.dds"},
	{0x485E52F6BC3E2A14ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_coma_0t.dds"},
	{0x485EF34519036555ULL, "mrl351_phantomelves6170352.cl"},
	{0x485F5EC284EB11ECULL, "186363845.ogg"},
	{0x485FC8D036D08B51ULL, "t3_wdg_bb3.tnt"},
	{0x485FE881FBCD69E2ULL, "vxlt_0_lowtown_gdh_0_0.vlm"},
	{0x486067506B9981F1ULL, "hlac_lt_gallowstemplar_n_2384_0y.dds"},
	{0x486090DAB800339DULL, "520868031.ogg"},
	{0x486097526AB62236ULL, "528625370.ogg"},
	{0x4860BBD914BA837CULL, "vxht_mansion_n_507_0_0.vlm"},
	{0x486186848AB77900ULL, "llac_lt_0_lowtown_a2_n_c_-2_5.dds"},
	{0x4861AAE6774DC864ULL, "vxlt_0_lowtown_docks_n_b6r_0_1.vlm"},
	{0x4861D168C1709596ULL, "849659384.ogg"},
	{0x486204711A7AF138ULL, "vxca_caverns1_mines_i70_0_0.vlm"},
	{0x4862762E3B377E5BULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardb.msh"},
	{0x4862942E3B37B145ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardb.mmh"},
	{0x48629FDBE5EF4A66ULL, "pro000_captain_wright6004411.cl"},
	{0x4862C566544B9680ULL, "vxw_bonepit_15l_0_0.vlm"},
	{0x48633099513F79F7ULL, "vxw_bonepitdestroyed_a3_1dp_0_0.vlm"},
	{0x486336BCB3CABC91ULL, "gen00fl_aveline6131892.cl"},
	{0x486392EB989AAC78ULL, "6083333_m.fxe"},
	{0x4863DEE6B824D2E0ULL, "vxlt_0_lowtown_a2_gz0_0_0.vlm"},
	{0x48647CD706867369ULL, "ave200_donnic6112166.cl"},
	{0x48651833D8A1F21BULL, "qun231_varnell6099648.cl"},
	{0x486531EC6A2435BEULL, "vxlt_gallowstemplar_f_6sh_0_0.vlm"},
	{0x486535B12BF6FCACULL, "gen00fl_merrill6141237.cl"},
	{0x48656BB3058F7243ULL, "vxlt_0_lowtown_a3_n_c_e1h_0_0.vlm"},
	{0x48656C972B15D136ULL, "dae300_alistair_king6134898.cl"},
	{0x4865A578BB91024CULL, "ser121_jethann6104578.cl"},
	{0x486643185C63A118ULL, "fen110_fenris6067903.cl"},
	{0x4866ACCCE7993781ULL, "6087475_m.fxe"},
	{0x4866CE6F7108212AULL, "framebuffereffects_twopassblur_p0.psh.fxo.dx11"},
	{0x4866D3DDC3D23BFBULL, "w_bonepitdestroyed_a3_98_0.msh"},
	{0x4866DBBFC2A86892ULL, "517695946.ogg"},
	{0x4866E5DDC3D25A46ULL, "w_bonepitdestroyed_a3_98_0.mao"},
	{0x4866EAE135F22582ULL, "blood_dot_l3.mao"},
	{0x4866F1DDC3D26EE5ULL, "w_bonepitdestroyed_a3_98_0.mmh"},
	{0x486749C3BC5C7D31ULL, "45466778.ogg"},
	{0x486818F7D038426DULL, "216906128.ogg"},
	{0x4868459A228289E4ULL, "6044765_m.fxe"},
	{0x48693EA8F5D21FD0ULL, "6180836_m.fxe"},
	{0x4869CBEBFB255C56ULL, "hlac_lt_0_lowtown_docks_a3_-3_5x.dds"},
	{0x4869F6F0D549DDEBULL, "6068368_m.fxe"},
	{0x486A334B1BF698B3ULL, "follower_banter6172899.cl"},
	{0x486A33B6B83CF1DDULL, "1066655714.ogg"},
	{0x486A814A404354CEULL, "vxw_woundedcoast_1oo_0_0.vlm"},
	{0x486AACEC236CB7D9ULL, "gen00fl_aveline6139522.cl"},
	{0x486AB7D1B01C9E16ULL, "hlac_w_blightlands_131068_0x.dds"},
	{0x486ACC8524AA5897ULL, "vxw_encounter2_camp_3v_0_0.vlm"},
	{0x486B01E690150902ULL, "746718490.ogg"},
	{0x486B52F1C1B5B965ULL, "mag210_arianni6030500.cl"},
	{0x486B653BBEFCCC6FULL, "560882253.ogg"},
	{0x486B9DEA45CBFAF7ULL, "905316395.ogg"},
	{0x486BCE3F28DC9F2EULL, "6031288_m.fxe"},
	{0x486C630E50793556ULL, "151837641.ogg"},
	{0x486C7454A53555A8ULL, "vxdr_0_deeproads_bhd_0_0.vlm"},
	{0x486C81A5F5E5EDB2ULL, "dae000ip_exit_gamlen_house.utp"},
	{0x486D06F38C7BA88DULL, "hlac_lt_ruinedhouse_n_-1_8y.dds"},
	{0x486D123995DD9647ULL, "vxht_mansion_n_47h_0_0.vlm"},
	{0x486D2D77558718C6ULL, "vxdr_dungeon_1lu_0_0.vlm"},
	{0x486D922DFA564B61ULL, "317197474.ogg"},
	{0x486D95992D833BB0ULL, "plt_cod_ltr_isabelaapology.nss"},
	{0x486DAEE9B772F797ULL, "dae000_gamlen6048428.cl"},
	{0x486E0B7F910D7B31ULL, "ca_caverns1_mines_24421.rml"},
	{0x486E1B5CB1CD5003ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_die.evt"},
	{0x486E404087A6F7FEULL, "vxlt_blackemporium_131_0_0.vlm"},
	{0x486E5A5EEDC38262ULL, "vxlt_0_lowtown_gci_0_0.vlm"},
	{0x486E890E242339A5ULL, "dae011_gossips6149796.cl"},
	{0x486EC4224F35AB19ULL, "6072167_m.fxe"},
	{0x486EE2B953463B12ULL, "2p_follower.stg"},
	{0x486EF07687960967ULL, "6106653_m.fxe"},
	{0x486F05D05EB55083ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_swordflame.msh"},
	{0x486F079728DDED67ULL, "llac_lt_0_lowtown_docks_a3_n_-196608_0.dds"},
	{0x486F13D05EB568DDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_swordflame.mmh"},
	{0x486F17D05EB56FEEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_swordflame.mao"},
	{0x486F9BDCC2D88BACULL, "vxdr_primevaltomb_4i3_0_0.vlm"},
	{0x486F9D78BADD5D0CULL, "abi_w_global.wbd"},
	{0x486FD7F193E5D7FBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_roll_shoulders.evt"},
	{0x4870845A1F36EE04ULL, "llac_ht_playershouse_n_855_1.dds"},
	{0x4870C6C6A7C2E1F8ULL, "6105448_m.fxe"},
	{0x4870CA022D33D5F6ULL, "vxht_0_hightown_a2_cbj_0_0.vlm"},
	{0x4871A27643359277ULL, "vxw_bonepitdestroyed_a3_1is_0_0.vlm"},
	{0x48721822AE03ADC8ULL, "6043724_m.fxe"},
	{0x48721C3D5540CC6DULL, "25245214.ogg"},
	{0x48723EE7B316E9DFULL, "llac_w_bonepitdestroyed_a3_458754_0.dds"},
	{0x487256BA0EBFDFAAULL, "124442137.ogg"},
	{0x4872D6BA6725779DULL, "6093486_f.fxe"},
	{0x48735B2F8E8EA1E0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowright.phy"},
	{0x48739ABBC9E3E575ULL, "fen200ip_to_slaver_hideout.utp"},
	{0x4873D79915BA0897ULL, "949646191.ogg"},
	{0x48747ADF3D8DE982ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_isab.msh"},
	{0x48747CDF3D8DED38ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_isab.mmh"},
	{0x48748FBB9E8A0E07ULL, "vxca_caverns1_mines_j2f_0_1.vlm"},
	{0x4874F45CB1D32C84ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_die.gad"},
	{0x487512CC2243AA81ULL, "35977886.ogg"},
	{0x487529181C90F49BULL, "t1_mue_bk3.tnt"},
	{0x48756168BC25ED4DULL, "hlac_lt_gallowsprison_n_-1_5y.dds"},
	{0x48757AF6C5337226ULL, "lt_0_lowtown_docks_3_0.anb"},
	{0x4875BE89B3302B44ULL, "follower_banter6154054.cl"},
	{0x487659290228BB98ULL, "vxht_0_hightown_a3_a78_0_0.vlm"},
	{0x487689BEDEE0CBBEULL, "544859526.ogg"},
	{0x4876B2798D15A88FULL, "6150250_m.fxe"},
	{0x4876D0F193EBEA3CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_roll_shoulders.gad"},
	{0x4876EC79C0D6DFB6ULL, "fen110_party_comment6095228.cl"},
	{0x48772301D8E62E27ULL, "vxlt_0_lowtown_f0d_0_0.vlm"},
	{0x487733BAACA8AFCFULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.cess.gad"},
	{0x4877415C1A7AD05CULL, "115912649.ogg"},
	{0x48777153FCF686BBULL, "vxdr_primevaltomb_8yd_0_0.vlm"},
	{0x487773DADF3D3BD6ULL, "dae000_arishok6081663.cl"},
	{0x487781FA04D73E5AULL, "880762195.ogg"},
	{0x4877C02261D069E1ULL, "zz_approval_debug6003675.cl"},
	{0x48783531149EBD6DULL, "294671828.ogg"},
	{0x48796D760DCB958DULL, "lt_0_lowtown_a2_38_0.rml"},
	{0x4879764DD67B4EABULL, "vxlt_0_lowtown_docks_n_bgg_0_0.vlm"},
	{0x487A2FB0744F763FULL, "lgt310tr_melson.ncs"},
	{0x487A5AC0AEED38C5ULL, "ser270_emeric6149282.cl"},
	{0x487A835EFFF42143ULL, "follower_barks6153686.cl"},
	{0x487A85CB245428ABULL, "vxht_mansion_n_5m6_0_1.vlm"},
	{0x487AC0EB3414AA10ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eagleheadsml_d.dds"},
	{0x487AC48D80EAB52EULL, "vxlt_0_lowtown_a2_n_gvi_0_0.vlm"},
	{0x487B171F75FD827FULL, "1006496560.ogg"},
	{0x487B7ACC4C7C993DULL, "6111541_m.fxe"},
	{0x487BD16D84CF262DULL, "15155501.ogg"},
	{0x487BF7E0C9A6A49BULL, "llac_ht_store_187_2.dds"},
	{0x487C035CB1D96562ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_die.ani"},
	{0x487C50ECCBC6C2DBULL, "6039277_f.fxe"},
	{0x487C63A88544FBB5ULL, "6207680_m.fxe"},
	{0x487CB21D799D107EULL, "vxlt_blackemporium_162_0_0.vlm"},
	{0x487D0C44CE211604ULL, "366868702.ogg"},
	{0x487D2957E1029ECBULL, "vxw_encounter2_camp_5h_0_0.vlm"},
	{0x487DCFF193F2073AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_roll_shoulders.ani"},
	{0x487DFC1AE7774C75ULL, "6098710_m.fxe"},
	{0x487E2F4892AD1C84ULL, "lt_0_lowtown_docks_a3_n_172_0.mmh"},
	{0x487E334892AD23D7ULL, "lt_0_lowtown_docks_a3_n_172_0.mao"},
	{0x487E3A7648B427B3ULL, "6031663_m.fxe"},
	{0x487E454892AD4226ULL, "lt_0_lowtown_docks_a3_n_172_0.msh"},
	{0x487E48BAACAEF27CULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.cess.evt"},
	{0x487E6B7303C588CEULL, "zz_isa_debug6133959.cl"},
	{0x487E83DF3D964DC3ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_isab.phy"},
	{0x487EA9F5D3459045ULL, "gen00fl_merrill6127267.cl"},
	{0x487EADE1AAB181B7ULL, "hlac_w_sundermount_65534_0y.dds"},
	{0x487F4541B1A6AFE5ULL, "5611549.ogg"},
	{0x487F5194DD4C2FA1ULL, "171664765.ogg"},
	{0x487FC552C4E2EEEEULL, "hlac_lt_0_lowtown_n_-131071_0z.dds"},
	{0x4880817FED682917ULL, "6076073_m.fxe"},
	{0x4880A2893FDEAB1CULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar1_mcr370_playerrules.gad"},
	{0x4880B4BA6232AB3AULL, "mag100_cullen6095121.cl"},
	{0x48812806764CF000ULL, "llac_lt_0_lowtown_docks_n_-196611_0.dds"},
	{0x4881F144DC57F3D6ULL, "vxw_encounter2_camp_rx_0_0.vlm"},
	{0x488243240FF29D64ULL, "6109138_m.fxe"},
	{0x48827B76D42E943BULL, "315189925.ogg"},
	{0x4882A8D6432F90D4ULL, "dae000_loyal_mage6165241.cl"},
	{0x4882B56ADACF6CACULL, "281410028.ogg"},
	{0x4882CEF2868FF772ULL, "6094766_m.fxe"},
	{0x48838743994E0ECCULL, "788141999.ogg"},
	{0x488398ACA4AB48EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0wo_basic.phy"},
	{0x48844B56D955B804ULL, "vxlt_warehouse_n_xv_0_0.vlm"},
	{0x48851F9CDFEDE552ULL, "ser221_party_comments6166114.cl"},
	{0x488594DECADC46A7ULL, "vxlt_0_lowtown_f5g_0_0.vlm"},
	{0x48873319FDF520FAULL, "dre000_vampire6034353.cl"},
	{0x48876CC5515816B3ULL, "mag101_aombanter6052902.cl"},
	{0x4887A1893FE4C81AULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar1_mcr370_playerrules.ani"},
	{0x4887E13084BABDA2ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3co_basic_1.mao"},
	{0x4887EF7309F35CB8ULL, "vxlt_blackemporium_15m_0_0.vlm"},
	{0x48881F305290157DULL, "hlac_ca_caverns2prc_5802_21z.dds"},
	{0x48882BC3FEC02D29ULL, "vxdr_primevaltomb_7yi_0_0.vlm"},
	{0x48885C95BBD4193BULL, "zz_dre_debug6168284.cl"},
	{0x4888C3CBB2A00E8EULL, "384433393.ogg"},
	{0x4888DCC7527513CDULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinscol.mmh"},
	{0x4888E0C752751A1EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinscol.mao"},
	{0x4888EEC752753253ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinscol.msh"},
	{0x4888F7B6A957FFE1ULL, "vxht_keep_n_73i_0_0.vlm"},
	{0x488969E63601485AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_3.gad"},
	{0x488996A8F22F1D8DULL, "622585241.ogg"},
	{0x4889BC88EB8B8D66ULL, "6128069_m.fxe"},
	{0x4889DF90E498799BULL, "qun221_petrice6076682.cl"},
	{0x488A26C0BE3AD1C0ULL, "plt_cod_ltr_instructions.plo"},
	{0x488A6924687A097CULL, "6171832_m.fxe"},
	{0x488ABB5B296DDCD6ULL, "vxca_caverns1_jkx_0_0.vlm"},
	{0x488B2A9C2C5BF086ULL, "261366778.ogg"},
	{0x488B4B8ABD629A60ULL, "243132087.ogg"},
	{0x488B50BAACB98995ULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.cess.ani"},
	{0x488B5D2658C006C0ULL, "hlac_lt_undercitydungeon_-1_9z.dds"},
	{0x488BB3D138D98389ULL, "vxlt_0_lowtown_n_db5_0_0.vlm"},
	{0x488CFEB4635DA117ULL, "hlac_w_encounter2_65535_0y.dds"},
	{0x488D08E6753932CCULL, "6104713_m.fxe"},
	{0x488D88D15DFFAD9AULL, "128661095.ogg"},
	{0x488E1141C8B7621DULL, "vxht_mansion2_259_0_0.vlm"},
	{0x488E11436CF334AEULL, "gen00fl_aveline6139923.cl"},
	{0x488E6B23F524428CULL, "gen00fl_anders6103677.cl"},
	{0x488E772B57F72845ULL, "fen310_fenris6117670.cl"},
	{0x488E8E6CD502439AULL, "vxca_caverns2prc_6hc_0_0.vlm"},
	{0x488EB3DECD091813ULL, "llac_ht_bartrand_n_2995_5.dds"},
	{0x488EC47DD32F2A1CULL, "w_encounter2_camp_43_0_ll.mmh"},
	{0x488EC9536F7C17B1ULL, "tattoos.gda"},
	{0x488EDB898249BB66ULL, "zz_dae_debug6047600.cl"},
	{0x488F6F4DDC398201ULL, "mag130_danzig6016450.cl"},
	{0x488F85CA35E6C641ULL, "mag291_torpor6056811.cl"},
	{0x488FCE79C2A8DABBULL, "135759302.ogg"},
	{0x48901335B65FEFAEULL, "vxht_mansion2b_1vp_0_0.vlm"},
	{0x48902782273CBBBBULL, "zz_followers_banter6158817.cl"},
	{0x48906321CE4E0C56ULL, "vxw_encounter2_camp_my_0_0.vlm"},
	{0x48906CE636076C71ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_3.evt"},
	{0x489075B5ACC66F92ULL, "follower_banter6141586.cl"},
	{0x4890C2D40C951C00ULL, "vxw_encounter3_fp_0_1.vlm"},
	{0x4890C85861136781ULL, "isa321_velasco6094481.cl"},
	{0x4890C97F8604BFD5ULL, "mag101_grace6035849.cl"},
	{0x4890DBA2328EB856ULL, "6121202_m.fxe"},
	{0x4890EA87B9BD7072ULL, "dae321ar_player_home.ncs"},
	{0x48925526457E0B24ULL, "hlac_w_bonepitdestroyed_a3_458752_0y.dds"},
	{0x4892806C6AACABC8ULL, "vxht_keep_n_73l_0_0.vlm"},
	{0x489289D2885A2880ULL, "vxca_caverns1_e7i_0_0.vlm"},
	{0x4892EA2B4D148CD4ULL, "fen210_orana6109684.cl"},
	{0x489316FA84F047A5ULL, "vxw_bonepit_1lr_0_0.vlm"},
	{0x489501B602E1F1BDULL, "vxlt_0_lowtown_a3_n_c_hl8_0_0.vlm"},
	{0x48953D4166916958ULL, "hlac_ca_cavern1_haunted_24424_3y.dds"},
	{0x48954B3CDDF3B43CULL, "129907237.ogg"},
	{0x4895D6E116259E4EULL, "vxw_woundedcoast_1pb_0_0.vlm"},
	{0x4895F098C7D5299CULL, "vxht_keep_n_53f_0_0.vlm"},
	{0x48964BF391BC97D2ULL, "mcr310_argument6074344.cl"},
	{0x4897490784D397E2ULL, "534456889.ogg"},
	{0x4897FBA84F188DE9ULL, "vxlt_backalley_n_1yf_0_0.vlm"},
	{0x48993562C9030217ULL, "669729514.ogg"},
	{0x489945AF569B275DULL, "215518717.ogg"},
	{0x4899782F8EAFA275ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowright.msh"},
	{0x48997A2F8EAFA5EBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowright.mmh"},
	{0x4899862F8EAFB988ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowright.mao"},
	{0x4899B44196F31101ULL, "vxca_caverns1_edm_0_0.vlm"},
	{0x4899C0C61F331D89ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierfix.mao"},
	{0x4899CCC61F333222ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierfix.mmh"},
	{0x4899CEC61F3335D8ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierfix.msh"},
	{0x4899D5C752837A8AULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinscol.phy"},
	{0x4899E4BE64BBB4CBULL, "llac_ca_caverns1_mines_24422_1.dds"},
	{0x4899F1D0062014E2ULL, "97127453.ogg"},
	{0x489A15BE504EA9E6ULL, "6130020_m.fxe"},
	{0x489BB8E8BAEBFA44ULL, "lt_0_lowtown_a2_n_28_0_lc.mmh"},
	{0x489BF8B90865F61AULL, "hlac_w_bonepitdestroyed_a3_196614_0y.dds"},
	{0x489C4C57AB06695DULL, "follower_banter6163411.cl"},
	{0x489CB6F24C06B4C1ULL, "vxlt_gallowstemplar_2ot_0_0.vlm"},
	{0x489CFF832847ABB4ULL, "6048178_m.fxe"},
	{0x489D6400C327C3D6ULL, "w_bonepit_167_0.msh"},
	{0x489D6E00C327D4D4ULL, "w_bonepit_167_0.mmh"},
	{0x489D7200C327DBE7ULL, "w_bonepit_167_0.mao"},
	{0x489D84E636121F40ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_3.ani"},
	{0x489DB494EE956D74ULL, "vxht_chantry_35a_0_0.vlm"},
	{0x489DE582AFB8914AULL, "combat_smaoe_r2_range.gda"},
	{0x489DFDC1230C2DC7ULL, "6035946_m.fxe"},
	{0x489F30FA2675628EULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc101a_0s.dds"},
	{0x489F9C885CDD0F14ULL, "79995281.ogg"},
	{0x48A0E6BC9A49A622ULL, "vxlt_gallowstemplar_34s_0_0.vlm"},
	{0x48A18C97903C30C6ULL, "6128344_m.fxe"},
	{0x48A242C016E78D77ULL, "34447702.ogg"},
	{0x48A2D7281204B8CBULL, "vxca_caverns1_j2j_0_0.vlm"},
	{0x48A2FBBEC7233AD3ULL, "gen_im_wep_rog_arc_act3_01.uti"},
	{0x48A2FE2F902AD1B7ULL, "6118243_m.fxe"},
	{0x48A36D607408BA12ULL, "6207113_m.fxe"},
	{0x48A43EAB30869855ULL, "dbg_setattrib.ncs"},
	{0x48A4540ED740A847ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_lightbeam2.dds"},
	{0x48A4937F17F4FF79ULL, "317368351.ogg"},
	{0x48A494A0D2396990ULL, "601689271.ogg"},
	{0x48A50A2C16893455ULL, "vxlt_0_lowtown_a2_n_f56_0_0.vlm"},
	{0x48A546C5AC9C81FAULL, "vxca_caverns2_7ml_0_0.vlm"},
	{0x48A57F1F9C696847ULL, "llac_lt_undercityencounter_-1_5.dds"},
	{0x48A59C2F2D892551ULL, "1060880802.ogg"},
	{0x48A5D7AD57D71CD9ULL, "vxht_0_hightown_n_ai5_0_0.vlm"},
	{0x48A6418A4A7A8B08ULL, "vxlt_warehouse_n_14m_0_0.vlm"},
	{0x48A6EB5BFD829D7CULL, "w_bonepitmining_a2_29_0.rml"},
	{0x48A73EE940531E5BULL, "6083969_m.fxe"},
	{0x48A75F446D0FE46BULL, "primitives_primitive_p3.psh.fxo"},
	{0x48A7A048FBBBBBF7ULL, "vxca_caverns2_941_0_1.vlm"},
	{0x48A855EE301BCEBCULL, "735844476.ogg"},
	{0x48A8ACE9AF5700C1ULL, "vxht_0_hightown_a2_bxz_0_0.vlm"},
	{0x48AA17EED0E73654ULL, "hlac_ht_chantry_n_2534_3y.dds"},
	{0x48AA2034689706BFULL, "6067747_m.fxe"},
	{0x48AA6D2E3B7524F2ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardb.phy"},
	{0x48AAD15BFD8659D1ULL, "w_bonepitmining_a2_29_0.msh"},
	{0x48AADF5BFD8671ECULL, "w_bonepitmining_a2_29_0.mao"},
	{0x48AAE8CD8CFF4445ULL, "hlac_ht_0_hightown_a3_-65537_4z.dds"},
	{0x48AAEB5BFD86858FULL, "w_bonepitmining_a2_29_0.mmh"},
	{0x48AAF453AC1E3567ULL, "743111136.ogg"},
	{0x48AAFB070B726537ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_small.gad"},
	{0x48AB888CC4ECA413ULL, "lowtown_fem_prostitute_sta6152311.cl"},
	{0x48ABE583D661D477ULL, "539006035.ogg"},
	{0x48AC625DFB728F6DULL, "657004061.ogg"},
	{0x48ACBAA7307D9C08ULL, "vxht_0_hightown_a2_9as_0_0.vlm"},
	{0x48AD55CBF746A076ULL, "gen_im_cft_rec_pot_08.uti"},
	{0x48AD5A77E258769CULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_023a_0n.dds"},
	{0x48AD6BB98135BAF9ULL, "68386644.ogg"},
	{0x48ADBF76B0396007ULL, "zz_fen_debug6165690.cl"},
	{0x48AE542B646EFCEAULL, "dae000_bartrand6107548.cl"},
	{0x48AE6C2E5F0A9A5FULL, "llac_lt_0_lowtown_docks_n_-196609_0.dds"},
	{0x48AF4F69D2CD4653ULL, "64793639.ogg"},
	{0x48AF6B581EEDDD86ULL, "gen00fl_anders6113050.cl"},
	{0x48AF86B7FE34502CULL, "zz_qun_debug6050208.cl"},
	{0x48AF952E868CA893ULL, "ico_amulet_2.dds"},
	{0x48B08FE2FE52826BULL, "vxlt_undercityserialkiller_13w_0_0.vlm"},
	{0x48B110070B76F4C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_small.evt"},
	{0x48B11F4F45A5EFA2ULL, "6129267_f.fxe"},
	{0x48B1512EC0ACBFF5ULL, "969966549.ogg"},
	{0x48B1752CB0D836A5ULL, "hlac_lt_0_lowtown_n_-196609_0y.dds"},
	{0x48B25E23D972FC23ULL, "6149343_f.fxe"},
	{0x48B26419F902740FULL, "vxlt_darktown_5ci_0_0.vlm"},
	{0x48B34010F484F002ULL, "vxw_woundedcoast_15m_0_0.vlm"},
	{0x48B3A7CB9F8FF7B4ULL, "hlac_dr_0_deeproads_5842_0y.dds"},
	{0x48B3BB9FE8B4DCEAULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.impa.gad"},
	{0x48B48C0148C3172EULL, "vxlt_0_lowtown_a2_n_gbm_0_0.vlm"},
	{0x48B49FC70F754B84ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_lar_p.mmh"},
	{0x48B4C371775E66C7ULL, "bn.cut"},
	{0x48B5C9B4B088F716ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_basic.mao"},
	{0x48B5D5B4B0890BB5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_basic.mmh"},
	{0x48B5D7B4B0890F2BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_basic.msh"},
	{0x48B60C48BBDBD25BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeloop.ani"},
	{0x48B743F30905F75CULL, "gen_im_trs_dal_med_boo_r1.uti"},
	{0x48B7B6507194F71FULL, "zz_lgt_1_debug6114366.cl"},
	{0x48B82322A8F78999ULL, "docks_male_overseer_mob.bnk"},
	{0x48B8A8A00637ECC4ULL, "6125288_m.fxe"},
	{0x48B8F1088E3F87D3ULL, "vxlt_darktown_2rm_0_0.vlm"},
	{0x48B9BC0F449628E9ULL, "lt_0_lowtown_docks.rml"},
	{0x48BA9E9FE8BAC981ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.impa.evt"},
	{0x48BB14AD4B1B41DCULL, "vxw_bonepit_103_0_0.vlm"},
	{0x48BB4522A8F9F9F9ULL, "docks_male_overseer_mob.cnv"},
	{0x48BB6FEA5C4ACD97ULL, "llac_ht_0_hightown_a2_65534_1.dds"},
	{0x48BBA5D457AD17D6ULL, "159791738.ogg"},
	{0x48BC2529CC2E867CULL, "134063692.ogg"},
	{0x48BC4F09A524FEF7ULL, "6089794_m.fxe"},
	{0x48BC5026861F24B6ULL, "446020859.ogg"},
	{0x48BC8E565299A126ULL, "llac_ht_playershouse_n_855_10.dds"},
	{0x48BD119957826EA9ULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_bda.phy"},
	{0x48BD3D27B6D010EEULL, "6155883_m.fxe"},
	{0x48BE4B9B48E47C0EULL, "hlac_dr_0_deeproads_17639_0z.dds"},
	{0x48BE82DFCDF3401DULL, "436707626.ogg"},
	{0x48BF28070B835ABDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_small.ani"},
	{0x48BF5363018953CCULL, "zz_dre_debug6168223.cl"},
	{0x48BF7EB608EBB9DCULL, "94942268.ogg"},
	{0x48BFF3CDA9C6E02DULL, "vxht_keep_n_678_0_0.vlm"},
	{0x48C0EDD05D84F756ULL, "hlac_lt_0_lowtown_a2_n_-1_7y.dds"},
	{0x48C14DAFEA9E8937ULL, "mag360_thrask6043514.cl"},
	{0x48C150F8385D8993ULL, "vxlt_undercityserialkiller_238_0_0.vlm"},
	{0x48C17550F8A8D389ULL, "gen_im_trs_bow.uti"},
	{0x48C17CC1951C5D7BULL, "680601844.ogg"},
	{0x48C188345E21B302ULL, "hlac_lt_backalley_n_0_18x.dds"},
	{0x48C1917222E20B47ULL, "332371309.ogg"},
	{0x48C1B1EDE67B0882ULL, "vxw_woundedcoast_10n_0_0.vlm"},
	{0x48C271BE4F942652ULL, "vxlt_darktown_2rh_0_0.vlm"},
	{0x48C2D3D7A66C4B93ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_5x.dds"},
	{0x48C2F7644C06290EULL, "716074911.ogg"},
	{0x48C35CA0368BA9F4ULL, "6044369_m.fxe"},
	{0x48C3B0F93D300BE6ULL, "lt_0_lowtown_a3_n_c_34_0_lc.mmh"},
	{0x48C3BF06232D78AAULL, "vxca_caverns1_mines_e7n_0_0.vlm"},
	{0x48C4581209F4F4CCULL, "vxlt_darktown_2wy_0_0.vlm"},
	{0x48C473EDE6282841ULL, "6112751_m.fxe"},
	{0x48C481442B5DF328ULL, "vxlt_blackemporium_xw_0_0.vlm"},
	{0x48C532E250A90BCDULL, "598612147.ogg"},
	{0x48C568636BF497F0ULL, "393971350.ogg"},
	{0x48C590A3934331C1ULL, "vxht_0_hightown_a2_bfx_0_0.vlm"},
	{0x48C6E14ED2C86FF5ULL, "follower_banter6169303.cl"},
	{0x48C714CD45D00F96ULL, "350976242.ogg"},
	{0x48C71ACF77F76F17ULL, "6070388_f.fxe"},
	{0x48C762E58035A053ULL, "vxlt_darktown_2ul_0_0.vlm"},
	{0x48C7674CF13512BFULL, "6166844_m.fxe"},
	{0x48C78677EE178601ULL, "and100_anders6095870.cl"},
	{0x48C7A9931E61DFB0ULL, "6067850_m.fxe"},
	{0x48C7B88D3C56C69CULL, "628298854.ogg"},
	{0x48C7BCEE95E24233ULL, "vxw_woundedcoast_1ra_0_0.vlm"},
	{0x48C7F1FD4E564E5BULL, "vxca_cave_27h_0_1.vlm"},
	{0x48C8969FE8C6F9D0ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.impa.ani"},
	{0x48C978046E9F76F6ULL, "370754017.ogg"},
	{0x48C9EE4A954B9A18ULL, "247589777.ogg"},
	{0x48C9EE93CDD2AD62ULL, "vxht_chantry_n_30o_0_0.vlm"},
	{0x48CA17788FF5E618ULL, "vxht_0_hightown_a3_btz_0_0.vlm"},
	{0x48CA417B06CD840BULL, "gen00fl_aveline6053314.cl"},
	{0x48CAA90503666F8FULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_lichenglowdots.mmh"},
	{0x48CAC2DFD7F5FAB8ULL, "hm_badpoet.mor"},
	{0x48CB24453BCCF277ULL, "6147935_m.fxe"},
	{0x48CB3348BBEE70E5ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeloop.gad"},
	{0x48CBAC8EC2FECDDCULL, "564203466.ogg"},
	{0x48CC092B4744CCD2ULL, "zz_fol_debug6103806.cl"},
	{0x48CC2FFC56FE6BB4ULL, "6215181_m.fxe"},
	{0x48CCB514D14E9E6DULL, "hlac_lt_0_lowtown_a2_n_c_-196608_2x.dds"},
	{0x48CCD1C25AEAC9F4ULL, "hlac_lt_house_-1_3y.dds"},
	{0x48CD186E68317F1EULL, "dae000_bodhan6129117.cl"},
	{0x48CD83CEA4B05D0FULL, "6047235_m.fxe"},
	{0x48CDC3AA83CD9A6AULL, "w_bonepitmining_a2_63_0.mao"},
	{0x48CDC7AA83CDA0B9ULL, "w_bonepitmining_a2_63_0.mmh"},
	{0x48CDD1AA83CDB1A7ULL, "w_bonepitmining_a2_63_0.msh"},
	{0x48CE065F0E4E6280ULL, "dae011_corff_sta6164577.cl"},
	{0x48CE410E72121500ULL, "pro000_colonel_ewald6039949.cl"},
	{0x48CE65AA9BBCF8ADULL, "vxht_keep_n_627_0_0.vlm"},
	{0x48CE860B14DA18AEULL, "77578011.ogg"},
	{0x48CE865106B184F0ULL, "6153982_m.fxe"},
	{0x48CED80194C9EAC7ULL, "dae000cr_grand_cleric6148111.cl"},
	{0x48CF1BA8228D4A91ULL, "6137691_m.fxe"},
	{0x48CF2874C236AE07ULL, "vxlt_darktown_59c_0_0.vlm"},
	{0x48CF47D3DCEEA50FULL, "vxlt_darktown_5ik_0_0.vlm"},
	{0x48CF7480DD07457BULL, "fx_black_halo.mao"},
	{0x48CF954464C81B2BULL, "llac_lt_0_lowtown_docks_-131074_6.dds"},
	{0x48D0203F042214D6ULL, "6094674_m.fxe"},
	{0x48D03E69B1111145ULL, "vxca_caverns2_7mg_0_0.vlm"},
	{0x48D14209AED0DE2BULL, "hlac_ca_caverns1_24421_11y.dds"},
	{0x48D1761BA60A935CULL, "6032539_m.fxe"},
	{0x48D1DABB9EBD8A1CULL, "6055604_m.fxe"},
	{0x48D1DC313731A8F3ULL, "zz_approval_debug6003655.cl"},
	{0x48D2062A8FE77881ULL, "201277741.ogg"},
	{0x48D2B9FEDB518FEEULL, "563762133.ogg"},
	{0x48D2EC5BFDA7D9CCULL, "w_bonepitmining_a2_29_0.anb"},
	{0x48D30C1C66A5D89AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04b.mmh"},
	{0x48D31E1C66A5F7E0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04b.msh"},
	{0x48D364C1F46A58CDULL, "vxht_mansion2_318_0_0.vlm"},
	{0x48D38B051A87A067ULL, "vxdr_0_deeproads_fbq_0_0.vlm"},
	{0x48D38DBF7D0889A8ULL, "562911606.ogg"},
	{0x48D4356FCB5B3210ULL, "vxlt_0_lowtown_docks_a3_bte_0_0.vlm"},
	{0x48D45C2C45222D86ULL, "78670303.ogg"},
	{0x48D4603A82450B1BULL, "vxdr_0_deeproads_erj_0_0.vlm"},
	{0x48D4A4A3FB3083F2ULL, "6174693_m.fxe"},
	{0x48D4BD2C9AAD0846ULL, "zz_lgt_2_debug6211854.cl"},
	{0x48D50AACCFAEF7C0ULL, "t3_fat_f01.tnt"},
	{0x48D529EEC2C55B0DULL, "dae000_porfiria6155380.cl"},
	{0x48D535F5B7D63F57ULL, "vxlt_foundry_n_vi_0_0.vlm"},
	{0x48D54A2CA4B7E72FULL, "6118364_m.fxe"},
	{0x48D575AD84A80CFEULL, "447646180.ogg"},
	{0x48D5C177C3B94D6DULL, "6077274_m.fxe"},
	{0x48D5D3191ECB7B86ULL, "vxlt_house_ea_0_0.vlm"},
	{0x48D5D4C2722BB8D3ULL, "vxlt_darktown_2xo_0_0.vlm"},
	{0x48D5F89405FC2E68ULL, "6150198_m.fxe"},
	{0x48D6C93971E09BB0ULL, "hlac_w_bonepitmining_a2_0_13z.dds"},
	{0x48D6EACC23396AF3ULL, "vxht_mansion_n_4vb_0_1.vlm"},
	{0x48D70DA841E2AD79ULL, "mag320_bsc_banter6100761.cl"},
	{0x48D741CD7E502BCEULL, "5332414.ogg"},
	{0x48D790B62B6F5321ULL, "vxlt_undercityencounter_13i_0_0.vlm"},
	{0x48D818B5F539BF1DULL, "6092280_m.fxe"},
	{0x48D83E252FF016BAULL, "s_deca_cr3.mao"},
	{0x48D85FE807544BB6ULL, "27462760.ogg"},
	{0x48D8647954F4B103ULL, "gen00fl_aveline6137500.cl"},
	{0x48D865702D6A656DULL, "cd_qun220cr_jarvis.bnk"},
	{0x48D88DFF63EFF2D7ULL, "vxlt_gallowstemplar_f_5ra_0_0.vlm"},
	{0x48D8C1A416F70B45ULL, "dae000cr_grand_cleric6199331.cl"},
	{0x48D90D93AA6F91A6ULL, "vxht_0_hightown_n_c7v_0_0.vlm"},
	{0x48D91A8EB14E2867ULL, "hlac_lt_0_lowtown_a2_n_c_-131071_1x.dds"},
	{0x48D938CE5B0DE27EULL, "6019040_m.fxe"},
	{0x48D977148351282EULL, "fx_smokenewss.dds"},
	{0x48D993A3CB762B39ULL, "vxw_sundermount_35l_0_0.vlm"},
	{0x48DA7790BDB69E16ULL, "980675056.ogg"},
	{0x48DA9C6E40B99B24ULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.mmh"},
	{0x48DACE372CE65183ULL, "vxw_encounter2_7n_0_0.vlm"},
	{0x48DB09ADEF7B1287ULL, "vxlt_undercitydungeon_yy_0_0.vlm"},
	{0x48DB25C2A0A4E879ULL, "llac_lt_0_lowtown_docks_a3_-4_3.dds"},
	{0x48DC0F1E0FDAC417ULL, "6126729_m.fxe"},
	{0x48DC2C6FC2B08F5DULL, "vxlt_foundry_n_xe_0_0.vlm"},
	{0x48DC39ABC25991F5ULL, "hlac_lt_0_lowtown_a3_n_c_0_4x.dds"},
	{0x48DC69E46DA991A1ULL, "127899124.ogg"},
	{0x48DC7AB978AC3C47ULL, "745935451.ogg"},
	{0x48DC83F11F5A8485ULL, "hlac_ht_0_hightown_a2_65536_2y.dds"},
	{0x48DCA2C70F96A37BULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_lar_p.anb"},
	{0x48DD3E224F73AE6CULL, "gen00fl_aveline6120844.cl"},
	{0x48DD7EBCF2954301ULL, "vxw_bonepit_10b_0_0.vlm"},
	{0x48DD9BB0FC0F1241ULL, "6087242_m.fxe"},
	{0x48DDB9B0CEE4BD8FULL, "vxlt_darktown_5rj_0_0.vlm"},
	{0x48DE0C3F3D092009ULL, "one280_iwan6097153.cl"},
	{0x48DE44AA90E8C29EULL, "zz_varric_arrows_tst.cut"},
	{0x48DE966F31CBE248ULL, "6128346_m.fxe"},
	{0x48DF2958BA118995ULL, "6111398_m.fxe"},
	{0x48DF3E8239A3218FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar7_.gad"},
	{0x48DF557833805752ULL, "vxlt_darktown_2xj_0_0.vlm"},
	{0x48DF5D95360BA6E4ULL, "6044534_m.fxe"},
	{0x48DF9D1185B238CDULL, "6114439_f.fxe"},
	{0x48E00375C68C5847ULL, "vxlt_0_lowtown_a3_n_c_e1t_0_0.vlm"},
	{0x48E0D4905194DD01ULL, "6207965_m.fxe"},
	{0x48E164FE0F4A2428ULL, "vxlt_blackemporium_vu_0_0.vlm"},
	{0x48E170EEE6B5F82AULL, "darktown_male_fer_urch_sta6174952.cl"},
	{0x48E1738CBAFD9142ULL, "hlac_lt_0_lowtown_a3_n_c_131071_2z.dds"},
	{0x48E186C07052C107ULL, "21589903.ogg"},
	{0x48E1E03ECDDD23B9ULL, "hlac_lt_undercityencounter_-1_37y.dds"},
	{0x48E23BE415677154ULL, "6126357_m.fxe"},
	{0x48E2711CC55876DBULL, "168266734.ogg"},
	{0x48E2BD40ABDCB570ULL, "6110392_m.fxe"},
	{0x48E3752F2B076688ULL, "74687390.ogg"},
	{0x48E37D7270527383ULL, "409727039.ogg"},
	{0x48E3BC0996ADFFAFULL, "6117264_m.fxe"},
	{0x48E3C98ACCAD9361ULL, "vxca_caverns2_7w9_0_0.vlm"},
	{0x48E484677AC89C9CULL, "vxht_0_hightown_n_asn_0_0.vlm"},
	{0x48E4F3EA7E68B684ULL, "189168045.ogg"},
	{0x48E58A8097EE86A0ULL, "643763137.ogg"},
	{0x48E5994F0883126BULL, "565185629.ogg"},
	{0x48E6334FF8AFAD2FULL, "vxlt_0_lowtown_a2_n_cow_0_0.vlm"},
	{0x48E6355350001269ULL, "llac_lt_warehouse_n_-65535_0.dds"},
	{0x48E69354CA550694ULL, "one280_merin6116752.cl"},
	{0x48E6CA153049300CULL, "gen_im_trs_bst_drg_scale_04.uti"},
	{0x48E6FB3273E21718ULL, "vxht_0_hightown_a3_bnx_0_0.vlm"},
	{0x48E74DC8F6026EC9ULL, "qcr220st_meredith_keep.stg"},
	{0x48E7BE319FC3C72FULL, "vxw_woundedcoast_125_0_0.vlm"},
	{0x48E7F35C59D28D66ULL, "plt_zz_fenpt_debug.nss"},
	{0x48E809F474BB08C4ULL, "vxlt_hangedman_n_10i_0_0.vlm"},
	{0x48E85BAD838BA9A1ULL, "118842049.ogg"},
	{0x48E8771B4953584BULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.dura.ani"},
	{0x48E8FA9957A782CEULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_bda.mmh"},
	{0x48E9149957A7AEBCULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_bda.msh"},
	{0x48E92F3D5B46E5AEULL, "778128474.ogg"},
	{0x48E940141EDC6A03ULL, "vxw_encounter2_2q_0_0.vlm"},
	{0x48E951EA8060938CULL, "461867390.ogg"},
	{0x48E957195CAEC6C0ULL, "355803062.ogg"},
	{0x48EA80FB01D0DC97ULL, "6123429_m.fxe"},
	{0x48EACBAB87EC7E43ULL, "dae000_gamlen6147206.cl"},
	{0x48EAD15B5BD5B42AULL, "vxdr_dungeon_1ly_0_0.vlm"},
	{0x48EB050C1EAA04E7ULL, "177843179.ogg"},
	{0x48EB1DAF70A3F01EULL, "zz_dae_debug6050749.cl"},
	{0x48EC204FFDEC8671ULL, "gen00fl_varric6077827.cl"},
	{0x48ECC60380C75233ULL, "795751939.ogg"},
	{0x48ECC70F44C194A3ULL, "lt_0_lowtown_docks.arl"},
	{0x48ED4B126FB61C9AULL, "vxlt_hangedman_n_284_0_0.vlm"},
	{0x48EDABF18529A8A7ULL, "6149887_m.fxe"},
	{0x48EDBBDD6CDCA573ULL, "vxlt_gallowscourtyard_n_695_0_0.vlm"},
	{0x48EE10FE11A59759ULL, "vxht_0_hightown_n_av8_0_0.vlm"},
	{0x48EE24228FC5B880ULL, "vrc221_varric6040914.cl"},
	{0x48EF4413A1EA1307ULL, "vxlt_0_lowtown_a2_n_el9_0_1.vlm"},
	{0x48EF59FC6512A8F3ULL, "hlac_lt_0_lowtown_a3_n_c_65536_9x.dds"},
	{0x48EF99D6910B9892ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack01.gad"},
	{0x48EFAE41493F610AULL, "gen00fl_merrill6129619.cl"},
	{0x48EFF3E9CB9A09DBULL, "gen00fl_anders6142561.cl"},
	{0x48F06EBEFE73D167ULL, "vxdr_0_deeproads_fhs_0_0.vlm"},
	{0x48F071A914E7817BULL, "vxca_caverns1_mines_ck5_0_0.vlm"},
	{0x48F092305065CE70ULL, "gst_standing_neutral.gda"},
	{0x48F0DF152C0E8B66ULL, "157378773.ogg"},
	{0x48F11133CD14FC7FULL, "w_bonepit_152_0.phy"},
	{0x48F187E2F874E47AULL, "6116636_f.fxe"},
	{0x48F194A2C44FD51BULL, "qun220cr_javaris6071863.cl"},
	{0x48F1B552523A1AC9ULL, "art\\levels\\races\\design\\prp_metaltrp.impact.gad"},
	{0x48F1E37450005900ULL, "vxw_bonepitmining_a2_259_0_0.vlm"},
	{0x48F232196A09D95AULL, "vis120_commoner26069043.cl"},
	{0x48F29CDA7BE03603ULL, "zz_and_debug6178663.cl"},
	{0x48F2D509D8047D49ULL, "hlac_ht_mansion2b_818_66y.dds"},
	{0x48F2EB5E61CD6EB7ULL, "423282367.ogg"},
	{0x48F32FD99AB3034CULL, "400297385.ogg"},
	{0x48F35B8239B3FC55ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar7_.ani"},
	{0x48F35EDF90591727ULL, "lgt150_nat_were_supporter6110968.cl"},
	{0x48F3AD60BCAA256EULL, "vxca_cavern1_haunted_ghe_0_0.vlm"},
	{0x48F44462C51B0F2EULL, "vxlt_0_lowtown_docks_n_c5i_0_0.vlm"},
	{0x48F4659ADC79E788ULL, "267545634.ogg"},
	{0x48F469078594E146ULL, "plt_gen00pt_specializations.plo"},
	{0x48F56AC06E74BA29ULL, "6064883_m.fxe"},
	{0x48F571B947DC23D7ULL, "vxlt_gallowstemplar_f_5xc_0_0.vlm"},
	{0x48F5F14D8E5BC2A6ULL, "vxht_0_hightown_n_c9x_0_0.vlm"},
	{0x48F617CB6AEA7849ULL, "989390917.ogg"},
	{0x48F65F1B495F6CC2ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.dura.evt"},
	{0x48F69CD69111BCC9ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack01.evt"},
	{0x48F6DA8E48B50575ULL, "vxlt_undercityserialkiller_1ge_0_0.vlm"},
	{0x48F708BCDE3C16F4ULL, "301881456.ogg"},
	{0x48F71E441C2F6109ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_wf.ani"},
	{0x48F75293415463C5ULL, "vxlt_0_lowtown_h26_0_0.vlm"},
	{0x48F78AF80B4A284FULL, "arena.arl"},
	{0x48F8054D3B16B703ULL, "vxw_blightlands_ag_0_0.vlm"},
	{0x48F814C89B9D7478ULL, "hlac_ht_chantry_n_2534_40z.dds"},
	{0x48F82B1C66C529A5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04b.phy"},
	{0x48F864DFE68F1615ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunarimage.anb"},
	{0x48F88652523FE917ULL, "art\\levels\\races\\design\\prp_metaltrp.impact.ani"},
	{0x48F92B7091A014B4ULL, "dae200cr_viscount6087383.cl"},
	{0x48F92B7C37E063B3ULL, "hlac_ht_chantry_n_2534_31y.dds"},
	{0x48F92D780389E0F3ULL, "aip_dragonling_base.gda"},
	{0x48F96D00C37631CFULL, "w_bonepit_167_0.phy"},
	{0x48FA5D4E343D321CULL, "611480660.ogg"},
	{0x48FA9D6AB2D0EE8FULL, "vxlt_darktown_5xt_0_0.vlm"},
	{0x48FB2213FA62B025ULL, "vxw_sundermount_2z1_0_1.vlm"},
	{0x48FB7270E3853842ULL, "vxlt_0_lowtown_docks_a3_but_0_0.vlm"},
	{0x48FC1DCE771FC127ULL, "6142987_m.fxe"},
	{0x48FC3CD7131C7D16ULL, "mnd_con_dis01.mao"},
	{0x48FC3E3691805E68ULL, "752620613.ogg"},
	{0x48FC5E1B4963D6D5ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.dura.gad"},
	{0x48FD085C59E50CA1ULL, "plt_zz_fenpt_debug.plo"},
	{0x48FD1B5B05A248E8ULL, "6080268_m.fxe"},
	{0x48FD5B2E3A44307DULL, "353746715.ogg"},
	{0x48FD715D355E797AULL, "vxca_cavern1_haunted_ec2_0_0.vlm"},
	{0x48FD7DBD17A940DEULL, "vxht_0_hightown_n_cax_0_0.vlm"},
	{0x48FDA56AC3FA74CFULL, "182208200.ogg"},
	{0x48FE2E8F76D97D86ULL, "6080108_f.fxe"},
	{0x48FE39441C35AD8BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_wf.gad"},
	{0x48FE62A40B2677F9ULL, "vxlt_darktown_5j4_0_0.vlm"},
	{0x48FEB839158098F9ULL, "6067657_m.fxe"},
	{0x48FEE06683A8D45CULL, "one280_golem6115978.cl"},
	{0x48FEF679B2FF339DULL, "lgt110_urn_barker6132987.cl"},
	{0x48FF1AFC2F6C054AULL, "zz_fol_debug6066149.cl"},
	{0x4900600666491E09ULL, "vxlt_0_lowtown_a2_n_c_def_0_0.vlm"},
	{0x49006D56D696043FULL, "plt_fen000pt_main.nss"},
	{0x49008C133D4D51BDULL, "25470928.ogg"},
	{0x4900918303B8FFF7ULL, "vxca_cavern1_haunted_j27_0_0.vlm"},
	{0x4900949A9179A31DULL, "701531000.ogg"},
	{0x4900ED85C2116B19ULL, "vxlt_ruinedhouse_n_5n_0_0.vlm"},
	{0x49013B2706CCB99BULL, "6122069_f.fxe"},
	{0x4901660C782B6FAFULL, "589954975.ogg"},
	{0x4901D26A6EEC5526ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrypot.dura.evt"},
	{0x4901E35A665E1BC6ULL, "vxca_caverns1_mines_eho_0_0.vlm"},
	{0x49021233CD237084ULL, "w_bonepit_152_0.mmh"},
	{0x49021633CD2377D7ULL, "w_bonepit_152_0.mao"},
	{0x49022833CD239626ULL, "w_bonepit_152_0.msh"},
	{0x4902731F5BEB2D04ULL, "art\\vfx\\combat\\vfx_w_wns_front_close2_c.mmh"},
	{0x4902B6681121E1A3ULL, "6061099_m.fxe"},
	{0x49034E73128EE4EFULL, "6080573_m.fxe"},
	{0x4903798CD80F6C6AULL, "1035701513.ogg"},
	{0x49039F6E40DCA51BULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.anb"},
	{0x4903E39639D23C57ULL, "vxlt_gallowstemplar_f_5cb_0_0.vlm"},
	{0x4903EECAD006E31BULL, "688720323.ogg"},
	{0x4903F4CB543571A6ULL, "zz_and_debug6178563.cl"},
	{0x4904181FC1B4D88EULL, "qcr000_arishok6059334.cl"},
	{0x49044674AD18ABD5ULL, "689927855.ogg"},
	{0x49045B702D8F8E97ULL, "cd_qun220cr_jarvis.wbd"},
	{0x490484D6911DD138ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack01.ani"},
	{0x4904A1EB83AFF82FULL, "vxw_woundedcoast_187_0_0.vlm"},
	{0x490623CE02C89B03ULL, "vxw_encounter2_8s_0_0.vlm"},
	{0x4906304B15B387AEULL, "mrl150_merrill6085833.cl"},
	{0x49064FEEBD584725ULL, "vxlt_0_lowtown_docks_a3_cx9_0_0.vlm"},
	{0x4906772A2D0CCF83ULL, "vxw_blightlands_jf_0_0.vlm"},
	{0x4906BA2C02A0DC00ULL, "docks_male_overseer_mob6176580.cl"},
	{0x490754DD04674A19ULL, "mrl310_merrill6120111.cl"},
	{0x4907D46C0E4BB58CULL, "6120805_m.fxe"},
	{0x4907E96A6EF0E871ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrypot.dura.gad"},
	{0x4907F2090539CAFDULL, "vxw_bonepitmining_a2_al_0_0.vlm"},
	{0x4908CD4F6E24E8E9ULL, "gen00fl_isabela_init6110211.cl"},
	{0x4909C83220D54B4AULL, "408778233.ogg"},
	{0x490A1FE2BC58FB85ULL, "sib111_mother6098807.cl"},
	{0x490B7491A0D9B86CULL, "hlac_lt_undercitydungeon_-1_7x.dds"},
	{0x490BA6615C16EE62ULL, "lgt110_orz_harr_carta6109061.cl"},
	{0x490BB918B0DE470BULL, "18370545.ogg"},
	{0x490BD9C560FB277AULL, "dae000_mother6136958.cl"},
	{0x490BEF5809B1180BULL, "266439566.ogg"},
	{0x490C132230D7FE31ULL, "506040389.ogg"},
	{0x490CF1E8ADCA93E0ULL, "hlac_lt_0_lowtown_docks_a3_65537_0y.dds"},
	{0x490CFEE10EFC7A4CULL, "llac_lt_gallowstemplar_n_2383_5.dds"},
	{0x490E08B5674C3944ULL, "6035138_m.fxe"},
	{0x490E556F6C7987C0ULL, "vxht_mansion2_1vv_0_0.vlm"},
	{0x490EEA6A6EF7081FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrypot.dura.ani"},
	{0x490F655D22CED6ADULL, "llac_lt_darktown_-1_5.dds"},
	{0x490F9CF7EC28A327ULL, "art\\vfx\\combat\\vfx_g_spider_fla_bst_c.impa.gad"},
	{0x490FC9E56EE5FBBCULL, "6174141_m.fxe"},
	{0x491017A6DE32069BULL, "mag100_cullen6140923.cl"},
	{0x4911047F30C6E2EDULL, "6153727_m.fxe"},
	{0x49115EC7438FC5ECULL, "hla_ht_0_hightown_a3_131072_0y.dds"},
	{0x4911F16628074214ULL, "hlac_lt_gallowscourtyard_-131071_3y.dds"},
	{0x4911FA9331F51EA5ULL, "vxw_woundedcoast_1o_0_0.vlm"},
	{0x4912228E2652ACDBULL, "6102662_m.fxe"},
	{0x491260EEBB2CDC5EULL, "vxlt_gallowstemplar_439_0_0.vlm"},
	{0x4912C20E5D189843ULL, "vxlt_0_lowtown_a3_n_c_emt_0_0.vlm"},
	{0x4912C329C9B837E3ULL, "6013366_m.fxe"},
	{0x49132552C51D8D10ULL, "s_ballista_05.dds"},
	{0x491359B66CFAE40DULL, "6061101_m.fxe"},
	{0x49138376DCDD657AULL, "6134767_m.fxe"},
	{0x49138CB782C0ED69ULL, "hlac_w_encounter2_327685_0x.dds"},
	{0x49138FFAC0231C3EULL, "vxlt_gallowscourtyard_72g_0_0.vlm"},
	{0x4913BBBB61CDDE6DULL, "6166644_m.fxe"},
	{0x4913BE482CA13675ULL, "vxlt_undercityserialkiller_1qc_0_0.vlm"},
	{0x4913DD98773841A6ULL, "vxht_0_hightown_a2_c1c_0_0.vlm"},
	{0x491405637E1C1901ULL, "hlac_lt_0_lowtown_a2_0_6z.dds"},
	{0x49142AF91806BCEDULL, "110218066.ogg"},
	{0x491448A35E00084AULL, "230290543.ogg"},
	{0x49144CC96DAEB9B1ULL, "394109557.ogg"},
	{0x491496747CFF42EBULL, "hlac_lt_undercityserialkiller_-1_24y.dds"},
	{0x4914E9071F02E803ULL, "vxw_blightlands_oi_0_0.vlm"},
	{0x491505BA0F41A49CULL, "hla_ht_0_hightown_a3_65536_0z.dds"},
	{0x49152A29AD7D2921ULL, "180400350.ogg"},
	{0x49156056D6A84A50ULL, "plt_fen000pt_main.plo"},
	{0x4915AE87AA5F86BFULL, "vxlt_gallowstemplar_5c5_0_0.vlm"},
	{0x491654651689BED2ULL, "dae000_pickpocket_scene6175097.cl"},
	{0x491691F7EC2EAF34ULL, "art\\vfx\\combat\\vfx_g_spider_fla_bst_c.impa.evt"},
	{0x491705DE2248B6A1ULL, "800247477.ogg"},
	{0x49173D6A627F1A66ULL, "qun221_petrice6077893.cl"},
	{0x49178CDAB128F53DULL, "hlac_dr_primevalentrance_1724_3x.dds"},
	{0x491795869BA59122ULL, "6063870_m.fxe"},
	{0x4917DD4648E83937ULL, "vxca_caverns1_mines_iim_0_0.vlm"},
	{0x49180DEB07CF003CULL, "6110834_f.fxe"},
	{0x491868FDF87014B2ULL, "mag110_arianni6206385.cl"},
	{0x49188AE964147753ULL, "6126484_m.fxe"},
	{0x49191BABAEEF0663ULL, "qun160_vashothlead6064589.cl"},
	{0x4919399F1CC934E4ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_com_0n.dds"},
	{0x49194FF7BC9BEFB4ULL, "llac_lt_0_lowtown_docks_-131071_0.dds"},
	{0x4919945E9334644EULL, "vxlt_darktown_5x9_0_0.vlm"},
	{0x4919E8A527A9FCA4ULL, "183970446.ogg"},
	{0x491A4F051456B138ULL, "hlac_lt_0_lowtown_a2_-262144_1x.dds"},
	{0x491AE5ABBAB9A5E9ULL, "3356686.ogg"},
	{0x491B16C6599E77E6ULL, "6202204_m.fxe"},
	{0x491B17AE4D407337ULL, "hlac_lt_0_lowtown_docks_131071_6x.dds"},
	{0x491B288AC4690E1AULL, "6120111_m.fxe"},
	{0x491BB5061899ED81ULL, "vxca_caverns1_e7d_0_0.vlm"},
	{0x491BB6CD37D4F8DEULL, "var_item.gda"},
	{0x491C39E98E2676DFULL, "46340554.ogg"},
	{0x491C3ED2258870E2ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robc_s.dds"},
	{0x491C8A3FD6022E3EULL, "6134943_m.fxe"},
	{0x491CC74C5E6FA040ULL, "vxht_mansion2_1yw_0_0.vlm"},
	{0x491D28699A4AFCD0ULL, "zz_followers_banter6159543.cl"},
	{0x491D69789117E8ADULL, "6059003_m.fxe"},
	{0x491E1614BC358D30ULL, "hightown_market_guard_sta6169598.cl"},
	{0x491E644D10ECB01EULL, "6125216_f.fxe"},
	{0x491EC7144A4A4CC6ULL, "vxca_caverns1_mines_ebq_0_0.vlm"},
	{0x491EFA854E440CD3ULL, "6084865_m.fxe"},
	{0x491F30EAD7550CE7ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v3.gad"},
	{0x492049D1BEABA422ULL, "zz_followers_banter6157345.cl"},
	{0x4920EBC221BB7574ULL, "vxw_bonepitmining_a2_1wm_0_1.vlm"},
	{0x4920EF9B4B4B4334ULL, "6115514_f.fxe"},
	{0x4921374D934742E3ULL, "hlac_lt_0_lowtown_65536_8y.dds"},
	{0x49217C2C77A4DA43ULL, "hlac_lt_0_lowtown_docks_-131074_1y.dds"},
	{0x4921C3CA1F999E7AULL, "vxlt_0_lowtown_docks_b56_0_1.vlm"},
	{0x4922260F92F79B43ULL, "947095964.ogg"},
	{0x492246956E957C27ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue3_e.dds"},
	{0x49225505C610D2B0ULL, "6088769_m.fxe"},
	{0x4922B40CE42AB6FFULL, "6056831_m.fxe"},
	{0x4922BF91329B5506ULL, "359703263.ogg"},
	{0x49232E4FF00D31BAULL, "681821952.ogg"},
	{0x49233DDEC9FCD6E0ULL, "6170206_m.fxe"},
	{0x49235AE410F90083ULL, "vxw_blightlands_ph_0_0.vlm"},
	{0x49235D5772AEE7C0ULL, "315150858.ogg"},
	{0x492436488D4B4E8FULL, "vxw_woundedcoast_143_0_0.vlm"},
	{0x49245D77447B3982ULL, "6121036_f.fxe"},
	{0x492489F7EC3ADF6DULL, "art\\vfx\\combat\\vfx_g_spider_fla_bst_c.impa.ani"},
	{0x4924B8315764EE67ULL, "gen00fl_merrill6141085.cl"},
	{0x4924D5C2E925FBFDULL, "vxw_bonepitmining_a2_kj_0_0.vlm"},
	{0x4924FFD6CEF81FAEULL, "6023966_m.fxe"},
	{0x492525EAD75965F4ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v3.evt"},
	{0x492615B0F243D2BEULL, "fen000ip_book_container.utp"},
	{0x4926225518755F8FULL, "hlac_lt_gallowstemplar_7198_2x.dds"},
	{0x4926306BAE15B23DULL, "vxw_woundedcoast_1ev_0_0.vlm"},
	{0x49266075F6DA9A85ULL, "945923409.ogg"},
	{0x492669F9723DA57FULL, "245092490.ogg"},
	{0x4926FD60BF9765FDULL, "729781689.ogg"},
	{0x4927295D18A57A9CULL, "gen00fl_fenris6103345.cl"},
	{0x49276B64D3366FFCULL, "vxlt_0_lowtown_a3_n_c_gye_0_0.vlm"},
	{0x4927C973FD4548CAULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc102a_0d.dds"},
	{0x4927DE281A6D7921ULL, "hlac_lt_0_lowtown_a3_n_c_131073_0y.dds"},
	{0x4927F1C7544B9038ULL, "hlac_lt_gallowsprison_n_-1_15y.dds"},
	{0x492804125C0F5A82ULL, "886848835.ogg"},
	{0x49282A462035A326ULL, "99065036.ogg"},
	{0x49284E5095CE51BAULL, "vxdr_primevalentrance_2st_0_1.vlm"},
	{0x4928690139D308C5ULL, "vxca_cavern1_haunted_ec9_0_0.vlm"},
	{0x49290D06381EE72CULL, "lgt260_eau_tmp_conrad6143795.cl"},
	{0x49291E9581AAD7D2ULL, "vxlt_undercityencounter_1cv_0_0.vlm"},
	{0x492947644C6E80E4ULL, "vxdr_primevaltomb_707_0_0.vlm"},
	{0x492A2CDF886AC3DBULL, "hlac_dr_dungeon_3217_10y.dds"},
	{0x492A761F5C0C84FBULL, "art\\vfx\\combat\\vfx_w_wns_front_close2_c.anb"},
	{0x492ABC9AE0CB6164ULL, "follower_banter6161326.cl"},
	{0x492B3452C53238AEULL, "s_ballista_05.mao"},
	{0x492B624B7D1B9714ULL, "vxht_0_hightown_n_by7_0_0.vlm"},
	{0x492B703ADC217997ULL, "lgt220_shp_mag_guard6149866.cl"},
	{0x492B8B1E5ED958A5ULL, "cd_ser270_cut_aftermath.wbd"},
	{0x492BFC506E62682EULL, "gen00fl_aveline6104170.cl"},
	{0x492C1D7DE9B88C08ULL, "dae000ip_crowd_mf6_talk13.utp"},
	{0x492C8B234BDE3E40ULL, "gen00fl_isabela6106401.cl"},
	{0x492CAF63BC9DF8B1ULL, "6095911_m.fxe"},
	{0x492D13527A646467ULL, "gen_im_arm_hlm_lgt_act1_01.uti"},
	{0x492D18F6AE88CFADULL, "6095967_f.fxe"},
	{0x492D20E2304C8DF2ULL, "gen00fl_player6207768.cl"},
	{0x492D49A6B1C262BEULL, "vxlt_gallowstemplar_n_6qj_0_0.vlm"},
	{0x492D894F842D4E39ULL, "mag210_arianni6030463.cl"},
	{0x492DBC653BAC4AC0ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh04a.phy"},
	{0x492EB2CE9C28893BULL, "823574015.ogg"},
	{0x492F27A8E79FCDE4ULL, "hlac_ca_caverns2_5802_27x.dds"},
	{0x492F48048AA7678FULL, "658203092.ogg"},
	{0x492F820435D76227ULL, "251336964.ogg"},
	{0x492FB9CCD0D8306AULL, "vxlt_darktown_48d_0_0.vlm"},
	{0x49301696CC055381ULL, "6139705_m.fxe"},
	{0x493168AC6BA15EC1ULL, "hlac_lt_undercityencounter_-1_47z.dds"},
	{0x4931C4EEB9938CEBULL, "vxdr_0_deeproads_ec9_0_0.vlm"},
	{0x4931CCC102EF1903ULL, "vxw_blightlands_uk_0_0.vlm"},
	{0x49321BA0035B5535ULL, "vxht_keep_n_612_0_0.vlm"},
	{0x493264FF1753CA8CULL, "vxw_encounter2_hx_0_0.vlm"},
	{0x4932B6AC9B6A2C54ULL, "6015965_m.fxe"},
	{0x49331DEAD765952DULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v3.ani"},
	{0x49337A9070396859ULL, "art\\vfx\\combat\\vfx_w_wns_left_b_close1_c.impa.ani"},
	{0x49343FDB513BF58FULL, "926051174.ogg"},
	{0x4934C1002CD46A37ULL, "vxca_caverns1_mines_ico_0_0.vlm"},
	{0x49355BA0FB417E11ULL, "6171107_m.fxe"},
	{0x493597386DB16AAAULL, "gen00fl_merrill6121127.cl"},
	{0x493600943AC200A5ULL, "ave121_jeven6105639.cl"},
	{0x493616BA388C25FCULL, "dr_0_deeproads_17639.rml"},
	{0x49363C4F455836EEULL, "zz_vrc_debug6200327.cl"},
	{0x4936F832C5082BA3ULL, "vxlt_undercityserialkiller_yn_0_0.vlm"},
	{0x4937A261410A369BULL, "gen00fl_varric6076574.cl"},
	{0x4937CC5E4E841D5FULL, "6196714_f.fxe"},
	{0x493822CF473C739CULL, "gen00fl_aveline6114076.cl"},
	{0x49383024AE3AACF6ULL, "245979370.ogg"},
	{0x49384D3B91C10C18ULL, "6116794_m.fxe"},
	{0x49385293BE9ECBEAULL, "vxdr_0_deeproads_gbe_0_0.vlm"},
	{0x4938CACA79083C76ULL, "6176101_m.fxe"},
	{0x4938EB54C4CC1782ULL, "68780111.ogg"},
	{0x49396FBFE78FEC0BULL, "gen00fl_anders6125939.cl"},
	{0x49397AAB0E90FB93ULL, "vxht_mansion2b_2pd_0_0.vlm"},
	{0x49397DE812E88DF9ULL, "llac_ht_brothel_n_2995_5.dds"},
	{0x4939AB06425BD140ULL, "vxht_mansion2_1gu_0_0.vlm"},
	{0x4939E1BF4FD107E4ULL, "zz_mag_debug6025382.cl"},
	{0x493A187AECD8E934ULL, "6080202_m.fxe"},
	{0x493A4FACD34F5517ULL, "sib100_father_contact6108274.cl"},
	{0x493A5590703F479BULL, "art\\vfx\\combat\\vfx_w_wns_left_b_close1_c.impa.gad"},
	{0x493A997A6A534A94ULL, "qcr000_varric6065812.cl"},
	{0x493B421381FF9B4DULL, "vxca_caverns1_i2s_0_0.vlm"},
	{0x493B8B8263E4872AULL, "w_bonepit_83_0.mao"},
	{0x493B8F8263E48E79ULL, "w_bonepit_83_0.mmh"},
	{0x493B998263E49F67ULL, "w_bonepit_83_0.msh"},
	{0x493BB9FAFB9E1495ULL, "atl_startmenu2_dxt1_01.dds"},
	{0x493BF6C6C6964526ULL, "vxca_caverns2_714_0_0.vlm"},
	{0x493C199BC809344EULL, "6188711_m.fxe"},
	{0x493C382FFB36B8A5ULL, "6089110_m.fxe"},
	{0x493C8B9C3037B96DULL, "vxw_woundedcoast_g1_0_0.vlm"},
	{0x493C8E398944BB7FULL, "151711960.ogg"},
	{0x493CADC0C929FEA1ULL, "vxlt_undercityencounter_1nb_0_0.vlm"},
	{0x493CE179F79BA41EULL, "qcr220_grey_wardens6060429.cl"},
	{0x493D187F4C3CD3FBULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.cess.ani"},
	{0x493D32706BBFF7EEULL, "dae000_epic_merchant_26179384.cl"},
	{0x493DD7ADAAA8B37AULL, "87889178.ogg"},
	{0x493E1B195FA68B52ULL, "w_sundermount_101_0_ll.mmh"},
	{0x493E3B87BC2B6817ULL, "6120795_f.fxe"},
	{0x493E3C137ACD6723ULL, "w_sundermount_234_0.mao"},
	{0x493E40137ACD6E70ULL, "w_sundermount_234_0.mmh"},
	{0x493E4E137ACD858AULL, "w_sundermount_234_0.msh"},
	{0x493F26E3D98B69B8ULL, "940193550.ogg"},
	{0x493F38BFA9FD03B0ULL, "6086969_m.fxe"},
	{0x493F3ECB80E749FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_enter.ani"},
	{0x493F5FC585F514AEULL, "genip_cin_moneypurseb.utp"},
	{0x493FB0146AA66D51ULL, "static_lava_p0.psh.fxo.dx11.assembly"},
	{0x4940448B7894C6F9ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_sqr.mmh"},
	{0x49406AFFD172E7A5ULL, "6156360_m.fxe"},
	{0x4940A02129326BD6ULL, "art\\levels\\races\\proxy\\props\\_textures\\cart_sa.dds"},
	{0x4940A2B2A46A6525ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_024a_0d.dds"},
	{0x4941041E8041D03FULL, "vxlt_gallowstemplar_5r6_0_0.vlm"},
	{0x494121C7BA4E53B8ULL, "315702796.ogg"},
	{0x4941729070459780ULL, "art\\vfx\\combat\\vfx_w_wns_left_b_close1_c.impa.evt"},
	{0x49423077E4ADE486ULL, "6103123_m.fxe"},
	{0x49423A2D95128633ULL, "ser221_party_comments6128807.cl"},
	{0x49427277E8A8629EULL, "vxw_bonepitdestroyed_a3_pb_0_0.vlm"},
	{0x494314259201E33DULL, "vxw_woundedcoast_1kt_0_0.vlm"},
	{0x49431E19781E08F2ULL, "6177829_m.fxe"},
	{0x4943E3B365228342ULL, "6143255_m.fxe"},
	{0x4944D474D7F6740AULL, "hlac_lt_blackemporium_1055_24z.dds"},
	{0x4944FAB4B2B752CFULL, "vxlt_0_lowtown_docks_bac_0_1.vlm"},
	{0x4945077A76D74BA0ULL, "6137464_f.fxe"},
	{0x494542970D5BAC38ULL, "6161411_m.fxe"},
	{0x4945807B86F572F5ULL, "6125708_m.fxe"},
	{0x4945E570910FD298ULL, "mag100_karras6052958.cl"},
	{0x4945F597472B7109ULL, "2261602.ogg"},
	{0x4945FC7B704B7F42ULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_1x.dds"},
	{0x49471DB5BB994C17ULL, "hlac_w_encounter3_-196611_0y.dds"},
	{0x49474F09F56E5B42ULL, "338800545.ogg"},
	{0x4947D09EC4DE35BEULL, "vxlt_0_lowtown_a2_n_hd8_0_0.vlm"},
	{0x49489D5B2604870AULL, "vxlt_darktown_6os_0_0.vlm"},
	{0x4948DAE88CC9164DULL, "vxw_encounter3_68_0_0.vlm"},
	{0x4948F12BBF679D3BULL, "6060382_m.fxe"},
	{0x4949784BE1E57BDAULL, "829441051.ogg"},
	{0x4949B12B058F5D72ULL, "ico_gloves_heavy_3.dds"},
	{0x4949CDD6F8FEF40EULL, "art\\vfx\\combat\\vfx_f_mer_emove_act_c.mmh"},
	{0x494A039BF0210A34ULL, "vxht_bartrand_n_5wq_0_0.vlm"},
	{0x494A21694FD0BBC3ULL, "224460430.ogg"},
	{0x494A2B930B9222BBULL, "6050077_m.fxe"},
	{0x494A4D02104B5F00ULL, "vxdr_0_deeproads_bnz_0_0.vlm"},
	{0x494A4F21A11DB79FULL, "vxca_caverns2_4y2_0_0.vlm"},
	{0x494AB415189EB387ULL, "6044911_m.fxe"},
	{0x494AC3BABB8C2E52ULL, "52802353.ogg"},
	{0x494B0D7DE418C8F7ULL, "qun211cr_elf_warrior6072330.cl"},
	{0x494B109D5F4383A6ULL, "6147628_m.fxe"},
	{0x494B207F4C491EF2ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.cess.evt"},
	{0x494B22BDB734BF09ULL, "766594714.ogg"},
	{0x494B5222A1797DC6ULL, "762681493.ogg"},
	{0x494B7E71E9490829ULL, "one100dg_vanard6017501.cl"},
	{0x494B97CB84E4DAAEULL, "dae300_bran6144840.cl"},
	{0x494C106F5C004CF9ULL, "414863782.ogg"},
	{0x494CBCD421BC495CULL, "vxlt_undercityserialkiller_20v_0_0.vlm"},
	{0x494CD765219A365CULL, "hlac_lt_darktown_-1_42z.dds"},
	{0x494CFB1FB9BF96D4ULL, "dae000_friend_turn_in6179262.cl"},
	{0x494D2E69E5CAAB58ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_shortbed.phy"},
	{0x494D40FF76CAC040ULL, "gen00fl_anders6144078.cl"},
	{0x494D56CB80F3AFCFULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_enter.evt"},
	{0x494D685CD3062454ULL, "ser200_aveline6201514.cl"},
	{0x494D6BA1A9FF1065ULL, "follower_banter6169871.cl"},
	{0x494D86384862ADE6ULL, "6183818_f.fxe"},
	{0x494E166615B950E7ULL, "165716960.ogg"},
	{0x494F5FEAEB3202F3ULL, "6107672_m.fxe"},
	{0x494FE92A98D3575BULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtb.msh"},
	{0x494FFB2A98D375A6ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtb.mao"},
	{0x4950072A98D38A45ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtb.mmh"},
	{0x49506CC8D1B5DC68ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ande.msh"},
	{0x49506E5857961258ULL, "test_ext.arl"},
	{0x49508AC8D1B60E92ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ande.mmh"},
	{0x4950A254D6A3FD06ULL, "6104023_m.fxe"},
	{0x49512FF8B87328B9ULL, "vxlt_darktown_2ty_0_0.vlm"},
	{0x49513278918BBE49ULL, "follower_banter6141749.cl"},
	{0x49513F7F4C4DBE85ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.cess.gad"},
	{0x4951B8D11154681CULL, "vxht_keep_n_5ps_0_0.vlm"},
	{0x49523E63E669547FULL, "547776834.ogg"},
	{0x49524409A07FF4C4ULL, "308958412.ogg"},
	{0x495250F2A4778331ULL, "art\\vfx\\materials\\vfx_wis_smo_add.mao"},
	{0x495256257636A586ULL, "69037698.ogg"},
	{0x4952871C564189FDULL, "cod_bks_lite_vials_6.utp"},
	{0x4952C39394B305ACULL, "74943478.ogg"},
	{0x4952D9653BCB9855ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh04a.msh"},
	{0x4952DB653BCB9BCBULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh04a.mmh"},
	{0x495367D52EF39633ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_hed_hdra_0n.dds"},
	{0x495457CB80F9D078ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_enter.gad"},
	{0x495484CE2BF6C6BDULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ichm.phy"},
	{0x495534F2E82245EAULL, "589777030.ogg"},
	{0x4955693BC4749841ULL, "120456885.ogg"},
	{0x4955A9EF7A42181EULL, "one140cr_charade6137077.cl"},
	{0x4955B5992A513D32ULL, "vxht_mansion_n_3nm_0_0.vlm"},
	{0x4955D54A0E6E0E4BULL, "vxw_encounter2_camp_jy_0_0.vlm"},
	{0x4955EDE5634BCD48ULL, "llac_dr_0_deeproads_5843_8.dds"},
	{0x4956103E9B111FF0ULL, "810882849.ogg"},
	{0x495626361442693BULL, "6121051_f.fxe"},
	{0x49568EC026480240ULL, "vxht_mansion2_1ms_0_0.vlm"},
	{0x4956BAB11AD10DE0ULL, "6119532_f.fxe"},
	{0x4956D175056CDED0ULL, "608404914.ogg"},
	{0x49574815C69ED029ULL, "424635702.ogg"},
	{0x49574CC57EBF2ECDULL, "vxw_encounter3_37_0_0.vlm"},
	{0x4957D221AD818154ULL, "894047662.ogg"},
	{0x4957ECD6F90B660DULL, "art\\vfx\\combat\\vfx_f_mer_emove_act_c.anb"},
	{0x49580CE08434B0F7ULL, "vxw_bonepit_1rr_0_0.vlm"},
	{0x49582C4581EC78AEULL, "341161680.ogg"},
	{0x495832395126DEDDULL, "hlac_lt_0_lowtown_a3_n_c_65536_0y.dds"},
	{0x49584F248BF242D4ULL, "6028914_m.fxe"},
	{0x4958A49838C61CDCULL, "vxw_bonepitmining_a2_25e_0_0.vlm"},
	{0x4958F0E20DD6E3A2ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.cess.gad"},
	{0x495968F703F829D3ULL, "t1_tat_gry.tnt"},
	{0x4959827A51399C26ULL, "6140728_m.fxe"},
	{0x495A8601F2E67C40ULL, "6125190_m.fxe"},
	{0x495AC356EA8596BEULL, "hlac_dr_dungeon_3207_4x.dds"},
	{0x495C004AAA1FAB82ULL, "17095392.ogg"},
	{0x495C8C81E8C5EFEDULL, "vxw_bonepit_1q9_0_0.vlm"},
	{0x495D3BA7FD5E74FEULL, "gen00fl_anders6128360.cl"},
	{0x495D3CFCACC4AD38ULL, "6106709_f.fxe"},
	{0x495D7F30CEFD3E7FULL, "vxw_woundedcoast_19p_0_0.vlm"},
	{0x495E4F64B13FEAA3ULL, "vxdr_primevaltomb_6yv_0_0.vlm"},
	{0x495F98E06BBFD1D5ULL, "236017354.ogg"},
	{0x495FDA8989FD4A0FULL, "mag360_samson6042069.cl"},
	{0x495FF7DF75EE143DULL, "vxw_woundedcoast_1qr_0_0.vlm"},
	{0x496013E20DDD3D99ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.cess.evt"},
	{0x49604268602B3382ULL, "llac_lt_backalley_g_0_4.dds"},
	{0x4960941CF5DCD6ABULL, "vxlt_0_lowtown_docks_n_b7w_0_0.vlm"},
	{0x49618D17E71D9AD7ULL, "plt_cod_bks_kirkwall_hist_3.plo"},
	{0x4961D82EC0A81B25ULL, "dae000_merchant_hm_mob6174214.cl"},
	{0x49623B93D52A9146ULL, "vxht_0_hightown_a2_bav_0_0.vlm"},
	{0x49625D6B9724D5D2ULL, "6016908_m.fxe"},
	{0x496292C7F2E81217ULL, "dae000ip_crowd_lm2_talk1.utp"},
	{0x4962B9496542FD34ULL, "mnd_con_smoke.mao"},
	{0x4963078D4D24977EULL, "hlac_lt_backalley_g_1_5z.dds"},
	{0x4963189468F89CD6ULL, "374018845.ogg"},
	{0x4963B651B0824518ULL, "mag291_torpor6026954.cl"},
	{0x4963FE4A10931E64ULL, "vxlt_0_lowtown_n_gfw_0_0.vlm"},
	{0x49644727006426CBULL, "vxw_encounter2_camp_ex_0_0.vlm"},
	{0x49645EB08EB53F8DULL, "mcr371_emile_de_launcet6087018.cl"},
	{0x49645F45A610BD7EULL, "mcr371_orsino6083317.cl"},
	{0x496477E3585C2C38ULL, "char1_vertexlit.vsh.fxo"},
	{0x4964A59646F79BEAULL, "6043865_m.fxe"},
	{0x4964A867DB69E4A4ULL, "zz_teststage.stg"},
	{0x4964BC127D0818E8ULL, "shapechange.gda"},
	{0x4964F9FB89FC3B59ULL, "6171204_m.fxe"},
	{0x4964FA267C4CAAF1ULL, "vxw_sundermount_35t_0_0.vlm"},
	{0x49650CFD2995EC3BULL, "129935455.ogg"},
	{0x496585CE2C053BC2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ichm.mmh"},
	{0x496587CE2C053F78ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ichm.msh"},
	{0x496589042FB36D74ULL, "crt_aoe_flare.mao"},
	{0x496617579390619CULL, "vxdr_slavers_1sp_0_0.vlm"},
	{0x49663CA8A00314BAULL, "gen00fl_merrill6141208.cl"},
	{0x4966570EAC36C635ULL, "vxdr_primevaltomb_6j8_0_1.vlm"},
	{0x4967368F0B201823ULL, "vxht_brothel_n_4z0_0_0.vlm"},
	{0x496820BDA6401206ULL, "752130260.ogg"},
	{0x49682B0158BBE28DULL, "art\\characters\\weapons\\promo\\shields\\w_shd_011a.msh"},
	{0x4968390158BBFAC0ULL, "art\\characters\\weapons\\promo\\shields\\w_shd_011a.mao"},
	{0x49683D0158BC0113ULL, "art\\characters\\weapons\\promo\\shields\\w_shd_011a.mmh"},
	{0x49684AAD4C0E2C99ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanl.ani"},
	{0x4968DD0E923E3271ULL, "34782508.ogg"},
	{0x4968F7DEE4DC94C0ULL, "6122090_m.fxe"},
	{0x4968FFE6907BD0F0ULL, "6059147_m.fxe"},
	{0x49690B96F995C648ULL, "503571104.ogg"},
	{0x49692CDF412F5902ULL, "341018485.ogg"},
	{0x49693086C756DCE5ULL, "411267248.ogg"},
	{0x49699A87B59676E3ULL, "688796757.ogg"},
	{0x4969C12D5E6CAB6AULL, "559796010.ogg"},
	{0x4969E5A2EDE1CB02ULL, "vxlt_foundry_n_w3_0_0.vlm"},
	{0x496A035348B39DDCULL, "vxw_bonepitmining_a2_1q2_0_0.vlm"},
	{0x496A2D76913CC206ULL, "6176071_m.fxe"},
	{0x496A4D59A1B648E2ULL, "mrl000_marethari6129308.cl"},
	{0x496A6C339053516AULL, "hlac_lt_gallowscourtyard_n_-131072_1y.dds"},
	{0x496A7BE89CE7938CULL, "hightown_market_guard_sta6169666.cl"},
	{0x496AFE5EDABC086DULL, "vxw_bonepit_1z8_0_0.vlm"},
	{0x496B8A2B3C64F94DULL, "zz_isa_debug6133150.cl"},
	{0x496BA2A0EFFE12ABULL, "vxca_caverns1_mines_cke_0_0.vlm"},
	{0x496BA41EBA0C692DULL, "vxlt_gallowstemplar_3rz_0_0.vlm"},
	{0x496BC35BB685B4E4ULL, "follower_banter6169285.cl"},
	{0x496BDCCD00DB5242ULL, "follower_banter6170379.cl"},
	{0x496C01B22C2FF1D4ULL, "407700529.ogg"},
	{0x496C651E5F10AF4FULL, "cd_ser270_cut_aftermath.bnk"},
	{0x496C66F47B151548ULL, "vxlt_0_lowtown_a2_n_c_eha_0_0.vlm"},
	{0x496C90C68D686503ULL, "36883308.ogg"},
	{0x496CC6085D630DFAULL, "dae000_hate_follower_react6103204.cl"},
	{0x496D1BE20DE7D588ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.cess.ani"},
	{0x496D54BD262CDFF0ULL, "w_encounter2_34_0.mao"},
	{0x496D5587E0C735E6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_s2.anb"},
	{0x496D58BD262CE603ULL, "w_encounter2_34_0.mmh"},
	{0x496D66BD262CFE5DULL, "w_encounter2_34_0.msh"},
	{0x496DAC6804638DE6ULL, "mag100_karras6052945.cl"},
	{0x496DFE54253F0477ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_paperlamp.msh"},
	{0x496E1054253F22DAULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_paperlamp.mao"},
	{0x496E1454253F29E9ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_paperlamp.mmh"},
	{0x496E69C88D3DECA4ULL, "vxht_playershouse_n_20m_0_0.vlm"},
	{0x496F25AD4C140CDBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanl.gad"},
	{0x496FB588B0CA4280ULL, "598291965.ogg"},
	{0x496FC42ED30EC81EULL, "lgt110pt_urn_miracle.ncs"},
	{0x49711F0D93C14DE4ULL, "106893025.ogg"},
	{0x4971A72F5A91259FULL, "hlac_ht_mansion2b_818_15y.dds"},
	{0x4971C1559723C02AULL, "6100257_f.fxe"},
	{0x4971FFB42D558882ULL, "dae000_madam_lusine6151399.cl"},
	{0x49724F5150E674A8ULL, "fen110_fenris6118032.cl"},
	{0x49726C9643A35DD9ULL, "weapon_weapon_p4.psh.fxo.dx11.assembly"},
	{0x49728FAFDA609124ULL, "1062443581.ogg"},
	{0x4972ADDC67505583ULL, "dae300_alistair_king6133571.cl"},
	{0x4973706E32CBD16BULL, "vxht_chantry_36c_0_0.vlm"},
	{0x4973752F981083DFULL, "vxca_cave_2bz_0_0.vlm"},
	{0x49741C9EFA6259FEULL, "vxht_keep_n_646_0_0.vlm"},
	{0x497427D0AE10B862ULL, "6169787_m.fxe"},
	{0x49742E748CE04928ULL, "vxlt_undercityserialkiller_1t5_0_0.vlm"},
	{0x497473CEF0E9A06BULL, "gen00fl_varric6105409.cl"},
	{0x4974A8587C787C7CULL, "69006061.ogg"},
	{0x4974C551F1057AECULL, "gen00fl_carver6103404.cl"},
	{0x49751349ED6B6833ULL, "vxw_woundedcoast_16m_0_0.vlm"},
	{0x4975377DD625CE39ULL, "6027661_f.fxe"},
	{0x49758B10787AE8B0ULL, "mrl310_party_comments6142831.cl"},
	{0x4975E6940E9EA212ULL, "6054025_m.fxe"},
	{0x49763F5D6850DBFCULL, "zz_area_signage5.utp"},
	{0x497642AD4C1A5CC0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanl.evt"},
	{0x49769E5838E2CA10ULL, "hlac_lt_0_lowtown_docks_n_-65539_5y.dds"},
	{0x4976A3CF9E3B66AAULL, "925576190.ogg"},
	{0x4976B1590741F4DEULL, "gen00fl_varric6206046.cl"},
	{0x4976B594CF4CD8E2ULL, "art\\characters\\weapons\\promo\\longswords\\textures\\w_lsw_025a_0s.dds"},
	{0x4977971F2404EB21ULL, "primitives_primitive_p5.vsh.fxo.dx11.assembly"},
	{0x49779CE424307F87ULL, "570281064.ogg"},
	{0x49781FFECDF0E3B8ULL, "6087122_m.fxe"},
	{0x4979A6495EB2CC60ULL, "6077244_m.fxe"},
	{0x497A661942FEC101ULL, "gen00fl_anders6119295.cl"},
	{0x497A8C32D7A8F1F8ULL, "vxht_mansion_n_4zh_0_0.vlm"},
	{0x497AFCA8BB40E712ULL, "gen00fl_merrill6107800.cl"},
	{0x497B0EA7E0401E01ULL, "vxw_bonepitmining_a2_254_0_0.vlm"},
	{0x497BE8E1E146A473ULL, "t3_wdg_bb1.tnt"},
	{0x497C39EBB080C49EULL, "vxw_bonepitdestroyed_a3_lf_0_0.vlm"},
	{0x497C88EE99E8A65AULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc102a_0t.dds"},
	{0x497D0A5F6E1D29DFULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.dura.gad"},
	{0x497D780840F268DAULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpb.mmh"},
	{0x497D8A0840F28720ULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpb.msh"},
	{0x497D8E59311CBE1AULL, "6116316_m.fxe"},
	{0x497DA5A4A5CACDB5ULL, "vxlt_undercitydungeon_17d_0_0.vlm"},
	{0x497DFC2D35F89397ULL, "hlac_ht_playershouse_n_855_8z.dds"},
	{0x497E2C704E9153F8ULL, "vxlt_gallowsprison_n_6kf_0_0.vlm"},
	{0x497E2F30F0C5DCDFULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.impa.gad"},
	{0x497E59BD263B5B28ULL, "w_encounter2_34_0.phy"},
	{0x497E5E74175A99D3ULL, "and121_karl6045431.cl"},
	{0x497E6D5420ED032CULL, "vxca_caverns2prc_559_0_0.vlm"},
	{0x497EE7F42C1F65C7ULL, "vxlt_0_lowtown_a3_n_c_eji_0_0.vlm"},
	{0x497EFEBCA3BC9143ULL, "6096230_f.fxe"},
	{0x497F1F4DB916C246ULL, "vxht_0_hightown_a2_box_0_0.vlm"},
	{0x497F4128E0F75D80ULL, "art\\levels\\ztools\\default\\generic\\0rr_basic.phy"},
	{0x497F59F48AB8A638ULL, "vxht_keep_n_67i_0_0.vlm"},
	{0x497FD17205BE1EA1ULL, "502737913.ogg"},
	{0x497FDDA2D7A7BD84ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galo2floor.msh"},
	{0x497FEFA2D7A7DCEDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galo2floor.mao"},
	{0x497FF3A2D7A7E326ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galo2floor.mmh"},
	{0x4980308934446DC4ULL, "vxca_cavern1_haunted_e4y_0_0.vlm"},
	{0x498112C50F028015ULL, "96440507.ogg"},
	{0x49822EF51FB7BE2EULL, "358188647.ogg"},
	{0x4982CCFFC43003D9ULL, "123455335.ogg"},
	{0x49830030A499BB87ULL, "llac_lt_gallowstemplar_n_2393_1.dds"},
	{0x498364385F08F7CDULL, "vxht_keep_n_6y2_0_0.vlm"},
	{0x49838526DF6180B3ULL, "vxw_woundedcoast_11l_0_0.vlm"},
	{0x4983C2E6584402CCULL, "llac_lt_0_lowtown_-196609_1.dds"},
	{0x4983FF5F6E2335CCULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.dura.evt"},
	{0x49840EDC93CC3274ULL, "gen00fl_aveline6122752.cl"},
	{0x49840F4FF7C74A3CULL, "vxlt_0_lowtown_a3_n_c_hde_0_0.vlm"},
	{0x498458BD263FC520ULL, "w_encounter2_34_0.rml"},
	{0x4985045732D66F96ULL, "llac_ht_keep_3732_11.dds"},
	{0x49852430F0CBE8CCULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.impa.evt"},
	{0x4985DDFD8210875BULL, "vxht_mansion_n_6na_0_0.vlm"},
	{0x4985E3A0746801F0ULL, "6053347_m.fxe"},
	{0x4985FA4C88ABC5CBULL, "wm_ico_sewerentrance.dds"},
	{0x49868407BBC0F158ULL, "vxlt_0_lowtown_a2_n_c_hr6_0_0.vlm"},
	{0x49869099ABF74567ULL, "gen_im_trs_apo_junk1_03.uti"},
	{0x49876A3F42A839EAULL, "32023970.ogg"},
	{0x4987870CAE8601DAULL, "hlac_lt_0_lowtown_a2_n_-131072_5z.dds"},
	{0x4988A38D7D269FB3ULL, "6148390_m.fxe"},
	{0x4988D5999E69FE0AULL, "hlac_ca_caverns1_mines_24426_6x.dds"},
	{0x498903939A65B012ULL, "6163562_f.fxe"},
	{0x4989CF8A7F5B0B8DULL, "hlac_lt_0_lowtown_65536_3z.dds"},
	{0x4989EAA2D7B028F1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galo2floor.phy"},
	{0x498A17C044C18B62ULL, "zz_amb_test6105066.cl"},
	{0x498A29B8AE89F16EULL, "gen_im_trs_qun_hvy_hlm_r2.uti"},
	{0x498A3F0D9D8F029BULL, "mcr370_varric6208276.cl"},
	{0x498A425BFF9B3FC4ULL, "vxlt_0_lowtown_docks_b58_0_1.vlm"},
	{0x498AA1434B764970ULL, "vxlt_gallowscourtyard_4wa_0_0.vlm"},
	{0x498ACC79389E798BULL, "119700373.ogg"},
	{0x498B138CC26AAFCCULL, "mag311_nyssa6042218.cl"},
	{0x498B36F63429182BULL, "6097286_f.fxe"},
	{0x498B4ABF73965F1FULL, "distortiongeomvol.psh.fxo"},
	{0x498B7D8C9D060C28ULL, "dae200cr_viscount6087185.cl"},
	{0x498B92727E3330D3ULL, "893619645.ogg"},
	{0x498C53315A9BF150ULL, "6033650_f.fxe"},
	{0x498C79BDAFC8157FULL, "58053833.ogg"},
	{0x498C9299D7AB9851ULL, "314503986.ogg"},
	{0x498CAB2894332A18ULL, "clf221_jalen6070341.cl"},
	{0x498D59D11E157E47ULL, "vxlt_0_lowtown_a3_n_c_emh_0_0.vlm"},
	{0x498DCBD17CAEBEB8ULL, "vxht_keep_n_62f_0_0.vlm"},
	{0x498DCF8DABE79B9FULL, "vxca_cavern1_haunted_j1w_0_1.vlm"},
	{0x498E15E135D276B5ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw105a.mao"},
	{0x498E6AC76ED12179ULL, "6108240_m.fxe"},
	{0x498EEC360F149410ULL, "mrl000_marethari6129265.cl"},
	{0x498F33C05DE86D01ULL, "6179908_f.fxe"},
	{0x498F58433FD34AE9ULL, "w_bonepit_222_0.mao"},
	{0x498F64433FD35E82ULL, "w_bonepit_222_0.mmh"},
	{0x498F6460F298AB74ULL, "vxca_caverns1_mines_gkr_0_0.vlm"},
	{0x498F66433FD36238ULL, "w_bonepit_222_0.msh"},
	{0x498F8DE5EC6237AEULL, "follower_banter6163744.cl"},
	{0x498F9E47C2638839ULL, "gen00fl_aveline6106836.cl"},
	{0x498FF1E6A72A89AFULL, "vxlt_foundry_n_yx_0_0.vlm"},
	{0x498FF62C80908844ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masc.msh"},
	{0x4990082C8090A6ADULL, "art\\characters\\playercharacter\\humanfemale\\boo_masc.mao"},
	{0x49900C2C8090ADE6ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masc.mmh"},
	{0x499079CAADC6A91FULL, "hlac_ht_keep_3732_7z.dds"},
	{0x4990E75F6E2D9725ULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.dura.ani"},
	{0x499173FAEDA13848ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbench.msh"},
	{0x499185FAEDA156D9ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbench.mao"},
	{0x499191FAEDA16B72ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbench.mmh"},
	{0x4991D69C125729A8ULL, "w_woundedcoast_16_0.msh"},
	{0x4991E89C12574839ULL, "w_woundedcoast_16_0.mao"},
	{0x4991F2B9275263A8ULL, "zz_approval_debug6003782.cl"},
	{0x4991F49C12575CD2ULL, "w_woundedcoast_16_0.mmh"},
	{0x49920367D06ED257ULL, "6083948_f.fxe"},
	{0x49920C30F0D64A25ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.impa.ani"},
	{0x49923CECA9EF84B4ULL, "vxht_bartrand_n_5nz_0_0.vlm"},
	{0x49927880FE2B978DULL, "art\\vfx\\models\\vfx_vol_light_narrow.mmh"},
	{0x49928A80FE2BB613ULL, "art\\vfx\\models\\vfx_vol_light_narrow.msh"},
	{0x49930F4A326FC35AULL, "397820748.ogg"},
	{0x49931A7EED93DA41ULL, "vxdr_primevaltomb_3hr_0_0.vlm"},
	{0x49938B68028A7847ULL, "character_character_p130.psh.fxo.dx11.assembly"},
	{0x4993D0BA01E48DAAULL, "6055417_m.fxe"},
	{0x4993EE500CCF9101ULL, "narrative_banter.bnk"},
	{0x4994BD605C9F4E0CULL, "6124778_f.fxe"},
	{0x4994EE6E39522098ULL, "vxca_caverns1_hmf_0_0.vlm"},
	{0x4995301E183EC35EULL, "hlac_lt_0_lowtown_n_-1_2x.dds"},
	{0x49956D8C7903B002ULL, "llac_ca_caverns2_5802_27.dds"},
	{0x4995B0D010A467E3ULL, "hlac_ca_caverns1_24423_7z.dds"},
	{0x4995C87ADAC57278ULL, "mag320st_mage_conspirator.stg"},
	{0x4996DFD472828749ULL, "775112305.ogg"},
	{0x499710500CD20171ULL, "narrative_banter.cnv"},
	{0x4997156A6F1CB833ULL, "6145090_m.fxe"},
	{0x4997E02A9910FDF2ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtb.phy"},
	{0x499841D00A08D3B0ULL, "follower_banter6154119.cl"},
	{0x4998B1B8FF67E09FULL, "vxht_0_hightown_a2_cf3_0_0.vlm"},
	{0x4998E2795EBF9284ULL, "6059420_m.fxe"},
	{0x4999638E01EE1795ULL, "hlac_ht_0_hightown_n_65534_0z.dds"},
	{0x4999699C87F3917EULL, "llac_lt_undercityencounter_-1_20.dds"},
	{0x4999FB1A2E76F671ULL, "461229078.ogg"},
	{0x499A032C8098F2B1ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masc.phy"},
	{0x499A7A9F7C4B4D40ULL, "vxht_bartrand_n_5p0_0_0.vlm"},
	{0x499ABF30755A0617ULL, "vxht_0_hightown_n_aq3_0_0.vlm"},
	{0x499ACC72DA90ED63ULL, "one100dg_nabil6173823.cl"},
	{0x499ADAB64812B8D1ULL, "540720612.ogg"},
	{0x499B530C39AF4494ULL, "dae000cr_grand_cleric6124729.cl"},
	{0x499C6DC7A175FCF5ULL, "362372158.ogg"},
	{0x499D4C61288AFB98ULL, "mrl250ip_at_elfcave.utp"},
	{0x499D567D3216CF25ULL, "6161436_m.fxe"},
	{0x499DA92D3276C84BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_ian.ani"},
	{0x499DBC706799EF8AULL, "6054685_m.fxe"},
	{0x499DDA7628540335ULL, "gen00fl_aveline6139948.cl"},
	{0x499DDDBCD7C1E5AAULL, "6069862_f.fxe"},
	{0x499E0984C0D0E03AULL, "564687987.ogg"},
	{0x499E1224D902ED05ULL, "vxlt_gallowsprison_n_6np_0_0.vlm"},
	{0x499E889F97A05D4DULL, "hlac_w_blightlands_65534_0y.dds"},
	{0x499EC39E345B36DBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basicnb.phy"},
	{0x499ECAC66DF2CD5FULL, "vxca_cave_2mg_0_0.vlm"},
	{0x499F3D3D28A24DAAULL, "753363722.ogg"},
	{0x499F3EBF67C60B84ULL, "vxdr_primevalentrance_2vy_0_1.vlm"},
	{0x499FCDF321B0ED93ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_hed_snga_0s.dds"},
	{0x49A0A5B9AB8E1FD7ULL, "vxlt_gallowstemplar_n_5pc_0_0.vlm"},
	{0x49A13C269C913BF7ULL, "zz_dae000ip_letter_desk.utp"},
	{0x49A25450D937BDFFULL, "gen00fl_merrill6150685.cl"},
	{0x49A2CBCB71E4AD4EULL, "158301091.ogg"},
	{0x49A3970841136CE5ULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpb.phy"},
	{0x49A3D876D9061C4DULL, "6155592_m.fxe"},
	{0x49A408120E960481ULL, "w_encounter2_30_0_lc.mmh"},
	{0x49A442EDFD896670ULL, "6179490_m.fxe"},
	{0x49A45039EAD536A6ULL, "vxht_0_hightown_a2_ccm_0_0.vlm"},
	{0x49A45E28E116AB15ULL, "art\\levels\\ztools\\default\\generic\\0rr_basic.msh"},
	{0x49A46028E116AE8BULL, "art\\levels\\ztools\\default\\generic\\0rr_basic.mmh"},
	{0x49A46C28E116C328ULL, "art\\levels\\ztools\\default\\generic\\0rr_basic.mao"},
	{0x49A4A3B75F1A985FULL, "6173666_m.fxe"},
	{0x49A4D9B4D7CF8595ULL, "vxht_bartrand_n_5s8_0_0.vlm"},
	{0x49A51CDAB86AC770ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingc_sa.dds"},
	{0x49A52DCA555D7A9FULL, "1042837958.ogg"},
	{0x49A572994DC06945ULL, "342546423.ogg"},
	{0x49A6C3AFC754A0D4ULL, "ave121_brennan6105915.cl"},
	{0x49A765BA9020A202ULL, "647215780.ogg"},
	{0x49A7728746CEF50AULL, "gen00fl_fenris6110783.cl"},
	{0x49A777280D187446ULL, "vxlt_0_lowtown_a2_hfy_0_0.vlm"},
	{0x49A7DB882E40EC23ULL, "6033402_f.fxe"},
	{0x49A83BE7BC27EE29ULL, "6202187_f.fxe"},
	{0x49A87009045B97F7ULL, "217721041.ogg"},
	{0x49A8870D74BF9B15ULL, "_testing.bnk"},
	{0x49A9111DF305EC25ULL, "6153231_m.fxe"},
	{0x49A91B69E618E9EDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_shortbed.msh"},
	{0x49A92969E6190120ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_shortbed.mao"},
	{0x49A92D69E6190873ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_shortbed.mmh"},
	{0x49A94FE9D30F668FULL, "6096303_m.fxe"},
	{0x49A9C24B0955879BULL, "60148534.ogg"},
	{0x49A9CA7D0212FB7AULL, "llac_lt_0_lowtown_a3_n_c_131072_1.dds"},
	{0x49A9ED76BDBF591EULL, "6111656_m.fxe"},
	{0x49AA1FAB5E0B0288ULL, "6201543_f.fxe"},
	{0x49AA5FAE74CC1970ULL, "mcr370_assignments6081680.cl"},
	{0x49AA873D3DC8B4FBULL, "6041825_f.fxe"},
	{0x49AA9D758F50D1D4ULL, "zz_qunari_sit_test.cut"},
	{0x49AAEC35832678BAULL, "vxw_sundermount_37e_0_0.vlm"},
	{0x49ABAEC4D3928AEEULL, "823045758.ogg"},
	{0x49AC81964E2A06FFULL, "character_character_p233.psh.fxo.dx11"},
	{0x49AC8640EB8DEDC7ULL, "6098194_m.fxe"},
	{0x49AC89C8D2046B8DULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ande.phy"},
	{0x49AD0CDCF14BB5EAULL, "vxlt_warehouse_n_se_0_0.vlm"},
	{0x49AD296C3876BE57ULL, "hlac_lt_0_lowtown_docks_a3_131072_0z.dds"},
	{0x49AD3C161DFF438CULL, "em_gen_act3.mor"},
	{0x49AD5DD1D2E7A584ULL, "art\\effects\\cloudshadows\\w_encounter2_camp\\cloud1.dds"},
	{0x49AD6281F461EAD4ULL, "vxdr_0_deeproads_cu6_0_0.vlm"},
	{0x49ADB008F52CD60FULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_dwpa_0s.dds"},
	{0x49AE633792AC92D1ULL, "ave331_jeven6138737.cl"},
	{0x49AE6611A1011F0AULL, "zz_seb_debug6183089.cl"},
	{0x49AEA2DAA11D1239ULL, "314550270.ogg"},
	{0x49AF3AA7EC1B7659ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha1a.mmh"},
	{0x49AF44A7EC1B8747ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha1a.msh"},
	{0x49AF64E6CDA993BCULL, "vxlt_0_lowtown_a3_n_c_hkx_0_0.vlm"},
	{0x49AF8A3FAC0F56BEULL, "vxca_cavern1_haunted_imd_0_0.vlm"},
	{0x49AFE2C58C915976ULL, "6176310_m.fxe"},
	{0x49B03E65797101FAULL, "165175380.ogg"},
	{0x49B0C3826873FAA1ULL, "vxht_bartrand_n_5r7_0_0.vlm"},
	{0x49B1902D328746D5ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_ian.gad"},
	{0x49B25562C755A99EULL, "hlac_lt_0_lowtown_docks_196606_0z.dds"},
	{0x49B2677FBFD364EDULL, "vxca_cavern1_haunted_jsr_0_0.vlm"},
	{0x49B2A265CEF9AB51ULL, "486920877.ogg"},
	{0x49B3D4E5A88A175FULL, "gen00fl_isabela6096851.cl"},
	{0x49B3DDB785616ADAULL, "6076094_m.fxe"},
	{0x49B445AA7BF8188FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyo_0s.dds"},
	{0x49B4A5DA41CC219FULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef04.msh"},
	{0x49B4ABDA41CC2C41ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef04.mmh"},
	{0x49B4B16CACDF0D0DULL, "6075687_m.fxe"},
	{0x49B4CEB3E3F7941DULL, "zz_dre_debug.ncs"},
	{0x49B4ED530CDAC443ULL, "vxca_cavern1_haunted_j4m_0_0.vlm"},
	{0x49B4EFEA178432AEULL, "dre000_sibling_has_blight6036548.cl"},
	{0x49B50C8288F4C290ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_vfx_suca_0t.dds"},
	{0x49B51F1E42C066FCULL, "hlac_ht_playershouse_n_855_34z.dds"},
	{0x49B52961F8123B2CULL, "6120067_m.fxe"},
	{0x49B57F09D465932DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01faltree.msh"},
	{0x49B58D09D465AB60ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01faltree.mao"},
	{0x49B59109D465B1B3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01faltree.mmh"},
	{0x49B5E904FF0E8CC6ULL, "vxlt_0_lowtown_a2_hcz_0_0.vlm"},
	{0x49B5F1E4DFBDFDD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_2.gad"},
	{0x49B67F87B8829DC9ULL, "6151497_m.fxe"},
	{0x49B69C5CDB194167ULL, "212526915.ogg"},
	{0x49B6AA1A619A1050ULL, "llac_lt_undercitydungeon_-1_52.dds"},
	{0x49B7A08CFCC75D60ULL, "6166674_m.fxe"},
	{0x49B813BFFBDDF06EULL, "fex_bd_a_rf.mmh"},
	{0x49B82DBFFBDE1C5CULL, "fex_bd_a_rf.msh"},
	{0x49B845CF273DD18EULL, "6201087_m.fxe"},
	{0x49B871479D860F33ULL, "28607282.ogg"},
	{0x49B88C0480F6193BULL, "6111052_f.fxe"},
	{0x49B8B7E137CBF92FULL, "gen00fl_aveline6122061.cl"},
	{0x49B910DF8B2A69EEULL, "984868715.ogg"},
	{0x49B911AE1A786177ULL, "872856366.ogg"},
	{0x49B9204C0896F9A9ULL, "53848443.ogg"},
	{0x49B92C31A6E2FABEULL, "hlac_dr_slavers_3214_9y.dds"},
	{0x49B95E12751C913AULL, "vxw_sundermount_32f_0_0.vlm"},
	{0x49BAFB7B568A29A3ULL, "and230_mage_girl6135806.cl"},
	{0x49BB66521930A988ULL, "ser270_templar6112593.cl"},
	{0x49BB7E979C6E0FACULL, "art\\vfx\\materials\\vfx_arc_cl_blend.mao"},
	{0x49BBBA0C99E9315EULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.cess.evt"},
	{0x49BBD2414957A141ULL, "vxht_brothel_n_4fv_0_0.vlm"},
	{0x49BC23DB5370FA9AULL, "179057568.ogg"},
	{0x49BC2713C2F28884ULL, "6156057_m.fxe"},
	{0x49BD06E4DFC44064ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_2.evt"},
	{0x49BD49961F2E002BULL, "and111_donation_box6200727.cl"},
	{0x49BDD5113DFEDF3FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_basic.msh"},
	{0x49BDDB113DFEE9E1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_basic.mmh"},
	{0x49BDE076C1AB2873ULL, "ser270_cut_aftermath6162283.cl"},
	{0x49BDE7113DFEFD82ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_basic.mao"},
	{0x49BE44DBA74982BDULL, "gen_im_npc_arm_hum_hlm_tmpc.uti"},
	{0x49BE4ADC7BBFFB3EULL, "zz_followers_banter6158568.cl"},
	{0x49BE52A408F94FBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_stagger_1.ani"},
	{0x49BE7273424A4667ULL, "llac_ht_mansion2_818_10.dds"},
	{0x49BE7B08581B3A51ULL, "gen00fl_sebastian6189031.cl"},
	{0x49BE8A427F48102CULL, "vxht_chantry_n_2y7_0_0.vlm"},
	{0x49BEBCBD37746136ULL, "573071092.ogg"},
	{0x49BED6008705F2C8ULL, "1014070504.ogg"},
	{0x49BF355F5A6A1321ULL, "vxht_bartrand_n_5u6_0_0.vlm"},
	{0x49BF90488EDDF659ULL, "vxdr_primevaltomb_35q_0_0.vlm"},
	{0x49BFF1459FC51039ULL, "hlac_lt_0_lowtown_docks_a3_-131075_6z.dds"},
	{0x49C08EDE0DF989C4ULL, "plt_mag110pt_ws_thrask.nss"},
	{0x49C123DF935B0E7DULL, "905239766.ogg"},
	{0x49C14A6B41F71545ULL, "gen00fl_aveline6033291.cl"},
	{0x49C14EA6CC9732B9ULL, "hlac_lt_undercityencounter_-1_1z.dds"},
	{0x49C241753C99BABFULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bottlewine_d.dds"},
	{0x49C2440B6F23B328ULL, "static_static_p43.psh.fxo.dx11.assembly"},
	{0x49C254111FE4465CULL, "6119636_m.fxe"},
	{0x49C2C58391776C78ULL, "vxht_mansion_n_4ee_0_0.vlm"},
	{0x49C2D10C99EF7689ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.cess.gad"},
	{0x49C2F055282CEBDFULL, "zz_followers_banter6157654.cl"},
	{0x49C3117528353E0BULL, "hlac_lt_0_lowtown_a2_0_6x.dds"},
	{0x49C3D2E950FB533AULL, "49942538.ogg"},
	{0x49C40DF3C7909EF2ULL, "337760563.ogg"},
	{0x49C42DC0DD044770ULL, "hlac_lt_0_lowtown_0_2y.dds"},
	{0x49C4824BC070E86FULL, "6131350_m.fxe"},
	{0x49C4BC5014EF8DB4ULL, "vxht_mansion2b_1b6_0_0.vlm"},
	{0x49C4E9DD6EAD1A18ULL, "vxlt_0_lowtown_docks_bgk_0_0.vlm"},
	{0x49C52D40A27BCBBFULL, "6125367_f.fxe"},
	{0x49C53E7C5ED85562ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_anda_0d.dds"},
	{0x49C54AE0ED076AA5ULL, "922525590.ogg"},
	{0x49C55B58CB1816C1ULL, "vxlt_undercityserialkiller_y4_0_0.vlm"},
	{0x49C55FA408FF84FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_stagger_1.gad"},
	{0x49C5AA345B91A129ULL, "6175323_m.fxe"},
	{0x49C5DBECEFB3D22CULL, "6084919_f.fxe"},
	{0x49C5E4B14759480AULL, "lt_gallowsprison_n_mmap.dds"},
	{0x49C61861E4ACE3EDULL, "vxca_caverns1_eg0_0_0.vlm"},
	{0x49C6352EA76EADB0ULL, "mag291_demon_marethari6028919.cl"},
	{0x49C69664B912A2A6ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.impa.ani"},
	{0x49C6DE393D0C9C21ULL, "llac_dr_slavers_3215_2.dds"},
	{0x49C70D49CA595DC4ULL, "gen00fl_aveline6114810.cl"},
	{0x49C756197511548DULL, "6169177_m.fxe"},
	{0x49C8403D8F785CC9ULL, "302773281.ogg"},
	{0x49C84DD47DA906D9ULL, "572916443.ogg"},
	{0x49C85FB80822C60CULL, "360035427.ogg"},
	{0x49C9055067E95DD1ULL, "dae000_ineria6140344.cl"},
	{0x49C91A91E1EF4EB3ULL, "vxht_0_hightown_a3_c4e_0_1.vlm"},
	{0x49C922C802B88C1BULL, "6170670_m.fxe"},
	{0x49C977F40788CB08ULL, "302162976.ogg"},
	{0x49C99FB3E00FA19DULL, "ach_kill_hybris.dds"},
	{0x49C9A20C99F545D7ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.cess.ani"},
	{0x49CA1546D23CDFB9ULL, "6153349_m.fxe"},
	{0x49CA34AD9B4D6C19ULL, "499317723.ogg"},
	{0x49CA571F941B399EULL, "art\\characters\\playercharacter\\shared\\textures\\default_gray50.dds"},
	{0x49CA6316BE71969AULL, "hlac_lt_0_lowtown_a2_-262144_1z.dds"},
	{0x49CAAA9E3480477AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basicnb.msh"},
	{0x49CAB89E34805EB3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basicnb.mao"},
	{0x49CABC9E348065C0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basicnb.mmh"},
	{0x49CB1EE4DFD0A65DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_2.ani"},
	{0x49CB256555A5EA4BULL, "668392698.ogg"},
	{0x49CB2B7475654730ULL, "122648964.ogg"},
	{0x49CBEF7FE1F4716EULL, "458605677.ogg"},
	{0x49CBFA3160923FD9ULL, "mrl251ar_elf_cave.are"},
	{0x49CC3AA40905641DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_stagger_1.evt"},
	{0x49CC7ECE9D5BBE7BULL, "montage_subs_demohack6217096.cl"},
	{0x49CC88F06AF9764EULL, "209987111.ogg"},
	{0x49CCB2A6C3D9D6B7ULL, "379607098.ogg"},
	{0x49CCFC1F713E28ACULL, "vxht_chantry_n_2t4_0_0.vlm"},
	{0x49CD03595D139DD5ULL, "6114472_m.fxe"},
	{0x49CDA73C4C602BA1ULL, "vxht_bartrand_n_5x5_0_0.vlm"},
	{0x49CDB2CDA1D18FDEULL, "6077022_m.fxe"},
	{0x49CDE5EED5F5EFB0ULL, "llac_w_encounter3_0_9.dds"},
	{0x49CE1B5422EA2AE6ULL, "888244642.ogg"},
	{0x49CE249E77538FF1ULL, "vxlt_darktown_2qb_0_0.vlm"},
	{0x49CE57B35F4A5FF8ULL, "vxlt_0_lowtown_n_fi7_0_0.vlm"},
	{0x49CE60601D25B6BEULL, "root_forcemage.dds"},
	{0x49CEE6CEBE0F6364ULL, "hlac_lt_0_lowtown_docks_a3_-4_6z.dds"},
	{0x49CF3A632B46FB92ULL, "wm_ico_bonepit.dds"},
	{0x49CF9742D4CD28E1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_ceiling_sa.dds"},
	{0x49CFA4574CB2BD9DULL, "art\\vfx\\materials\\vfx_icecrystals01.mao"},
	{0x49CFB953B0E93B35ULL, "877150760.ogg"},
	{0x49CFCE3322CC9551ULL, "hlac_w_blightlands_2_0x.dds"},
	{0x49D0524B6C2F1CE4ULL, "364948257.ogg"},
	{0x49D0FDBC73D3DD94ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_dummy_sa.dds"},
	{0x49D1E01DEC652FAAULL, "6023972_f.fxe"},
	{0x49D20913DF85A6D6ULL, "6034526_m.fxe"},
	{0x49D2341CAF0AC4DAULL, "mrl150_merrill6078550.cl"},
	{0x49D2F1CA8C4AB95DULL, "vxlt_gallowstemplar_n_5cx_0_0.vlm"},
	{0x49D2FFDC495E6884ULL, "vxht_mansion2_1vr_0_0.vlm"},
	{0x49D37E64B91D0417ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.impa.evt"},
	{0x49D4AD0D74E5169FULL, "_testing.wbd"},
	{0x49D4CFF4424FF0A6ULL, "gen00fl_varric6102186.cl"},
	{0x49D574A1BA666F84ULL, "art\\vfx\\combat\\vfx_gib_human_bld_p.impa.ani"},
	{0x49D574B142B6C545ULL, "6095380_m.fxe"},
	{0x49D58FE1C833320DULL, "6173079_m.fxe"},
	{0x49D594652FADE0B9ULL, "vxlt_0_lowtown_a2_n_g93_0_0.vlm"},
	{0x49D5AEB32FF23843ULL, "767555901.ogg"},
	{0x49D627509E9B5568ULL, "gen00fl_aveline6201453.cl"},
	{0x49D629AF4514B05EULL, "vxht_0_hightown_n_c4w_0_0.vlm"},
	{0x49D64BD8B578AE04ULL, "vxca_caverns1_mines_hio_0_0.vlm"},
	{0x49D6580296FEBFE5ULL, "vxht_bartrand_n_5sh_0_0.vlm"},
	{0x49D7132C0C488ECDULL, "follower_banter6141815.cl"},
	{0x49D757C3C6B63D53ULL, "6103041_m.fxe"},
	{0x49D7DE10E81F5037ULL, "vxlt_0_lowtown_a2_n_eoe_0_0.vlm"},
	{0x49D7EBB3FF53FEB0ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.mmh"},
	{0x49D83EE8E0D53115ULL, "vxht_keep_4bn_0_0.vlm"},
	{0x49D870E4EDD6FD36ULL, "vxlt_0_lowtown_docks_a3_n_bfb_0_0.vlm"},
	{0x49D99B6AB1022ABCULL, "6117986_m.fxe"},
	{0x49D9A9B3E41671BAULL, "zz_dre_debug.cnv"},
	{0x49D9FCDCED9AFEBBULL, "6036499_m.fxe"},
	{0x49DA7198A19CF169ULL, "llac_ht_brothel_n_3369_1.dds"},
	{0x49DA7F64B92324A0ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.impa.gad"},
	{0x49DAC2DA44192E5EULL, "gen00fl_merrill6128700.cl"},
	{0x49DACED78C84CDF6ULL, "vxw_sundermount_30a_0_1.vlm"},
	{0x49DC19193E564421ULL, "vxht_bartrand_n_5c4_0_0.vlm"},
	{0x49DC71A1BA6C89E6ULL, "art\\vfx\\combat\\vfx_gib_human_bld_p.impa.gad"},
	{0x49DD00D729F29466ULL, "llac_w_encounter3_-262144_1.dds"},
	{0x49DD5AAD806A211BULL, "lgt000ip_hat_sack.utp"},
	{0x49DD61A113D53545ULL, "gen00fl_aveline6114413.cl"},
	{0x49DDCF9B7C8E668CULL, "static1_s2.psh.fxo"},
	{0x49DDDFE290647290ULL, "hlac_dr_0_deeproads_5845_13z.dds"},
	{0x49DDE09127DB31ABULL, "336719369.ogg"},
	{0x49DE56AC4EE9A6D8ULL, "vxht_0_hightown_a3_9uz_0_0.vlm"},
	{0x49DE6F1C17862A29ULL, "hlac_w_encounter2_327682_0y.dds"},
	{0x49DF0898959B7D72ULL, "6088208_m.fxe"},
	{0x49DF1D5C16D7E6E7ULL, "sse_ss_abomination_1.gda"},
	{0x49DF265FF2703150ULL, "744139974.ogg"},
	{0x49DF39372CBEAF37ULL, "20096853.ogg"},
	{0x49DF3BBA91BAB346ULL, "hlac_ca_caverns2_5802_27z.dds"},
	{0x49DF760309AD8D6BULL, "1010493539.ogg"},
	{0x49DF839AC5ED8D96ULL, "6076018_m.fxe"},
	{0x49DFC8BF68158EFBULL, "vxht_0_hightown_av9_0_0.vlm"},
	{0x49E01B7DE8678587ULL, "6138771_m.fxe"},
	{0x49E027D30E8DD854ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_wesley_victim.gad"},
	{0x49E06F144AB53F37ULL, "hlac_lt_gallowstemplar_2384_0x.dds"},
	{0x49E0ACDA41F168BEULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef04.phy"},
	{0x49E0D8305C44DA09ULL, "6034349_m.fxe"},
	{0x49E11287641E429DULL, "vxht_0_hightown_c7z_0_0.vlm"},
	{0x49E1A009F8DBBEB4ULL, "vxht_mansion2b_1x4_0_0.vlm"},
	{0x49E22468EF134A3EULL, "vxca_caverns2prc_9xp_0_0.vlm"},
	{0x49E26CA1BA70ECC5ULL, "art\\vfx\\combat\\vfx_gib_human_bld_p.impa.evt"},
	{0x49E288910959CB7EULL, "6099651_m.fxe"},
	{0x49E28FDE0E1671BFULL, "plt_mag110pt_ws_thrask.plo"},
	{0x49E361EE78BCD09CULL, "vxw_encounter2_camp_j_0_0.vlm"},
	{0x49E439D358FD858DULL, "6163377_m.fxe"},
	{0x49E454A51429660CULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_carv.msh"},
	{0x49E45AA5142970FEULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_carv.mmh"},
	{0x49E4BB04E3F45940ULL, "lgt110pt_harrow_escape.ncs"},
	{0x49E4E6678C7D26CEULL, "vxca_caverns1_mines_in1_0_0.vlm"},
	{0x49E50EB3FF5EC4CFULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.anb"},
	{0x49E5CAD5C87E8363ULL, "vxw_sundermount_280_0_0.vlm"},
	{0x49E61F3227CB1CBFULL, "6120446_m.fxe"},
	{0x49E63CDE14C52DC2ULL, "vxdr_0_deeproads_f1m_0_0.vlm"},
	{0x49E64FEDDA1568B7ULL, "vxlt_0_lowtown_a2_n_epd_0_0.vlm"},
	{0x49E6759ABD4D35DDULL, "6093672_m.fxe"},
	{0x49E6B198880F0233ULL, "gen00fl_fenris6112104.cl"},
	{0x49E70BBFE5BDBA8FULL, "vxlt_0_lowtown_a2_n_e1s_0_0.vlm"},
	{0x49E736D30E941032ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_wesley_victim.ani"},
	{0x49E766AA42A5647DULL, "333665915.ogg"},
	{0x49E77C87FD53A0B5ULL, "vxlt_0_lowtown_a3_n_c_hqg_0_0.vlm"},
	{0x49E786CE33E42D1CULL, "vxw_bonepit_29g_0_0.vlm"},
	{0x49E78B4E794BBAEFULL, "vxca_caverns2_8od_0_0.vlm"},
	{0x49E7B032518FDC1DULL, "443130334.ogg"},
	{0x49E7E87D881BECBAULL, "404737139.ogg"},
	{0x49E89591EA5A2D18ULL, "vxht_keep_n_4jl_0_0.vlm"},
	{0x49E90A27B35DD265ULL, "6017819_f.fxe"},
	{0x49E94C63DC7A979CULL, "6116303_m.fxe"},
	{0x49E9562CCAAD7AA4ULL, "hlac_dr_slavers_3214_13z.dds"},
	{0x49E9DC113E24265EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_basic.phy"},
	{0x49EA83D38F7B70D9ULL, "6174715_m.fxe"},
	{0x49EADF3B3DA07EF3ULL, "ser120_ghyslain6118389.cl"},
	{0x49EB3B6D433690F8ULL, "vxlt_0_lowtown_n_fg5_0_0.vlm"},
	{0x49EBF08279583588ULL, "hlac_ht_chantry_n_2534_41y.dds"},
	{0x49EC1BC1073795C3ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.cess.evt"},
	{0x49EC7A64F94115F9ULL, "hlac_lt_0_lowtown_docks_a3_131070_4z.dds"},
	{0x49EC90FAEDEE156DULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbench.phy"},
	{0x49ED2F15FA98F219ULL, "hlac_lt_0_lowtown_docks_a3_-196610_5y.dds"},
	{0x49ED7A75879195F2ULL, "6083308_m.fxe"},
	{0x49ED7DD503D2ADB3ULL, "6060452_m.fxe"},
	{0x49EDF8B28D221485ULL, "6051394_f.fxe"},
	{0x49EE9794B1971ED2ULL, "51322612.ogg"},
	{0x49EF0D41651FB5DCULL, "hlac_ht_keep_3741_0z.dds"},
	{0x49EF828AB6FA6A6AULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.impa.gad"},
	{0x49EFBBC45B8ADCF2ULL, "gen00fl_aveline6139765.cl"},
	{0x49EFCFE838471832ULL, "follower_banter6141890.cl"},
	{0x49F01CE5AB14BBFDULL, "vxdr_0_deeproads_cuj_0_1.vlm"},
	{0x49F0C67EFBD5887FULL, "vxca_caverns1_mines_id2_0_0.vlm"},
	{0x49F0CBEDF8AD5C5DULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_roba_0s.dds"},
	{0x49F0DE08E66CCCE8ULL, "lt_0_lowtown_docks_a3_151_0.msh"},
	{0x49F0F008E66CEB79ULL, "lt_0_lowtown_docks_a3_151_0.mao"},
	{0x49F0FC08E66CFF12ULL, "lt_0_lowtown_docks_a3_151_0.mmh"},
	{0x49F10B3160B178AFULL, "mrl251ar_elf_cave.ncs"},
	{0x49F11D0BD6E2407EULL, "hlac_ca_caverns1_mines_24422_1y.dds"},
	{0x49F17BEB7C29D6A2ULL, "309673592.ogg"},
	{0x49F1B351D0434E69ULL, "6149166_m.fxe"},
	{0x49F1F9BE69CE0A7FULL, "hlac_w_sundermount_-131075_0x.dds"},
	{0x49F20BE44EB9299BULL, "6141832_m.fxe"},
	{0x49F2EA84DBAF0F7EULL, "qun211_braggart6075472.cl"},
	{0x49F2F4C1073D7244ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.cess.gad"},
	{0x49F31D50E1881140ULL, "6072253_f.fxe"},
	{0x49F32F7E14B50E8CULL, "950868691.ogg"},
	{0x49F358447E733F4EULL, "vxca_caverns1_mines_ik2_0_0.vlm"},
	{0x49F366840D404112ULL, "vxht_0_hightown_c1s_0_0.vlm"},
	{0x49F3F49C12AA2779ULL, "w_woundedcoast_16_0.rml"},
	{0x49F407A6387471E0ULL, "vxca_caverns2_6ng_0_1.vlm"},
	{0x49F463792EE4E1BFULL, "414982171.ogg"},
	{0x49F4AEBB06BB4642ULL, "vxdr_0_deeproads_f4n_0_0.vlm"},
	{0x49F4C1CACC0B8137ULL, "vxlt_0_lowtown_a2_n_eug_0_0.vlm"},
	{0x49F57CFE5A7F681FULL, "hlac_dr_slavers_3216_4y.dds"},
	{0x49F60304C2DF30C3ULL, "llac_lt_0_lowtown_docks_a3_-131076_1.dds"},
	{0x49F627F03BD77C38ULL, "478347823.ogg"},
	{0x49F6658AB7005701ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.impa.evt"},
	{0x49F672CFE0203EAFULL, "dae000ip_crowd_lm1_talk11.utp"},
	{0x49F67E15F9CFADDAULL, "hlac_lt_0_lowtown_a2_n_-65535_2y.dds"},
	{0x49F6A5002B617FDDULL, "6076026_m.fxe"},
	{0x49F6AC02A15BA542ULL, "462803083.ogg"},
	{0x49F721C249548214ULL, "dre100cr_hireling_2.utc"},
	{0x49F73248F4844DECULL, "hlac_lt_0_lowtown_a2_n_65537_2x.dds"},
	{0x49F79B40F22209A1ULL, "280006319.ogg"},
	{0x49F79DF50B3A4A41ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.anb"},
	{0x49F7FAE32A4E5943ULL, "6108493_m.fxe"},
	{0x49F80B3160B797FEULL, "mrl251ar_elf_cave.lst"},
	{0x49F8191A0469D2F7ULL, "6201389_m.fxe"},
	{0x49F81EDF467CE674ULL, "llac_lt_0_lowtown_docks_n_-2_2.dds"},
	{0x49F832E63591A29DULL, "vxlt_gallowstemplar_45t_0_0.vlm"},
	{0x49F8EE0B82A54201ULL, "6165447_m.fxe"},
	{0x49F903C10741F722ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.cess.ani"},
	{0x49F91E1E9CA3F356ULL, "vxw_bonepit_277_0_0.vlm"},
	{0x49F9B0C9B4160D6DULL, "plt_cod_aow_tome.plo"},
	{0x49F9FAF2808C1214ULL, "444031765.ogg"},
	{0x49FA3F4810425CD7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_curtains_d.dds"},
	{0x49FBB05F9D5C208EULL, "hlac_lt_gallowscourtyard_n_-65537_2z.dds"},
	{0x49FBDBFD9EBBA67AULL, "422324452.ogg"},
	{0x49FC125BE5AF4B45ULL, "mag170_hugh6080814.cl"},
	{0x49FC27C6E7FA7851ULL, "819722709.ogg"},
	{0x49FC8120E0CC029FULL, "6066191_m.fxe"},
	{0x49FC9F1452C9218CULL, "vxw_encounter2_2j_0_0.vlm"},
	{0x49FCC331BF655DAAULL, "vxht_0_hightown_a2_bot_0_0.vlm"},
	{0x49FCD7F077C54245ULL, "llac_ht_bartrand_n_2995_3.dds"},
	{0x49FD0034A089504AULL, "gen00fl_varric6033659.cl"},
	{0x49FD15EB9DA401FAULL, "dae000_meredith6085255.cl"},
	{0x49FE19F8EF0AD398ULL, "417673237.ogg"},
	{0x49FE55731F40B204ULL, "vxht_mansion2_1gq_0_0.vlm"},
	{0x49FFF449E5122FFAULL, "vxlt_0_lowtown_a2_n_c_hv8_0_0.vlm"},
	{0x4A000DA2FF19652AULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_4x.dds"},
	{0x4A000EA8FEE24764ULL, "331089790.ogg"},
	{0x4A0041B0F6AAB088ULL, "6080170_m.fxe"},
	{0x4A00EA89C833D3F4ULL, "llac_w_sundermount_327676_0.dds"},
	{0x4A012DBE7537564EULL, "680270452.ogg"},
	{0x4A0159556C51F634ULL, "685287101.ogg"},
	{0x4A0177E789F6F695ULL, "6126684_m.fxe"},
	{0x4A01C13D4724797CULL, "6137550_f.fxe"},
	{0x4A02384A68788419ULL, "dr_dungeon_3218_lc.mmh"},
	{0x4A023ECE2ED72F3CULL, "207696761.ogg"},
	{0x4A027322C86AE92CULL, "503614045.ogg"},
	{0x4A03342C60AF5B9CULL, "hlac_lt_blackemporium_1055_23y.dds"},
	{0x4A03B84F47BEFCF7ULL, "hlac_lt_gallowstemplar_n_2388_2y.dds"},
	{0x4A045D8AB70C8750ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.impa.ani"},
	{0x4A046089DECFBBBFULL, "vrc321_party_barks6095748.cl"},
	{0x4A0465CB38D6A0EDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papera.msh"},
	{0x4A0473CB38D6B820ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papera.mao"},
	{0x4A0477CB38D6BF73ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papera.mmh"},
	{0x4A0543696A361D23ULL, "gen_im_trs_krk_one_lsw_r1.uti"},
	{0x4A057EF50B4652BAULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.mmh"},
	{0x4A06185435DCF0D2ULL, "mag320_orsino6071070.cl"},
	{0x4A069A4251223769ULL, "follower_banter6155955.cl"},
	{0x4A070FA15EBDAD20ULL, "361059020.ogg"},
	{0x4A0721714E8A4FB8ULL, "character_character_p234.psh.fxo.dx11.assembly"},
	{0x4A074C80993E9D34ULL, "178817488.ogg"},
	{0x4A074EE7A3C820AFULL, "gen00fl_anders6125698.cl"},
	{0x4A076094F11512E0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scale_d.dds"},
	{0x4A078FFB8E9A0BD6ULL, "vxw_bonepit_228_0_0.vlm"},
	{0x4A08618A030886D0ULL, "zz_vault_debug6203365.cl"},
	{0x4A08A7B485D45836ULL, "vxht_bartrand_n_5wc_0_0.vlm"},
	{0x4A08FEB1720A7B3EULL, "lgt250_way_dalish_conflict6147119.cl"},
	{0x4A091340481084C4ULL, "qcr000_arishok6057099.cl"},
	{0x4A0949E8ED1A2797ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x8portcull.phy"},
	{0x4A0983F0A25AD083ULL, "hlac_lt_0_lowtown_docks_n_-4_3x.dds"},
	{0x4A09A1219A66AB1CULL, "w_blightlands_6_0.mao"},
	{0x4A09AD219A66BFBFULL, "w_blightlands_6_0.mmh"},
	{0x4A09B3219A66CA61ULL, "w_blightlands_6_0.msh"},
	{0x4A09D1C6B0FCBB9DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basicnb.mao"},
	{0x4A09D5C6B0FCC2D6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basicnb.mmh"},
	{0x4A09DFC6B0FCD3D4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basicnb.msh"},
	{0x4A0A4A5C981D9AF2ULL, "6161145_m.fxe"},
	{0x4A0A651AAE462E7FULL, "lt_house_mmap.dds"},
	{0x4A0A8BFD259966E0ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_vfx_suca_0d.dds"},
	{0x4A0AB48705008384ULL, "6086868_m.fxe"},
	{0x4A0AF2217C954F0DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_zbm_0n.dds"},
	{0x4A0AFD1A96FEF7ECULL, "748032709.ogg"},
	{0x4A0B3147B8B4A019ULL, "one130cr_mekel.utc"},
	{0x4A0B34A844F7B0C5ULL, "hlac_lt_undercityencounter_-1_30x.dds"},
	{0x4A0B4E7B02AF46A0ULL, "686730828.ogg"},
	{0x4A0B87890A9C6B37ULL, "caustic_flipbook.mao"},
	{0x4A0BAD113FE620DEULL, "6166452_m.fxe"},
	{0x4A0BB79758B1DB6CULL, "6126960_m.fxe"},
	{0x4A0BD09302E941A2ULL, "40410085.ogg"},
	{0x4A0C01D00A33198EULL, "6067827_m.fxe"},
	{0x4A0CA81C22311399ULL, "vxw_woundedcoast_1if_0_0.vlm"},
	{0x4A0CD035774EB217ULL, "hlac_lt_0_lowtown_a2_-65536_3y.dds"},
	{0x4A0DB395AF219A32ULL, "846896469.ogg"},
	{0x4A0DD83F0403D777ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtoweratall.msh"},
	{0x4A0DE54A3F0AC47AULL, "hlac_lt_0_lowtown_a2_-2_1y.dds"},
	{0x4A0DEA3F0403F5DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtoweratall.mao"},
	{0x4A0DEE3F0403FCE9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtoweratall.mmh"},
	{0x4A0E2455388B1033ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_loop.ani"},
	{0x4A0EE7495905FBF0ULL, "6039717_f.fxe"},
	{0x4A0EF02C24BCC3BCULL, "hlac_lt_undercitydungeon_-1_50x.dds"},
	{0x4A0FA7FC44CB777CULL, "dae000_meredith6085367.cl"},
	{0x4A0FC1982749C853ULL, "plt_cod_itm_vstmt_sacrifice.plo"},
	{0x4A107FD8AA40319FULL, "vxlt_0_lowtown_n_bxw_0_0.vlm"},
	{0x4A109AFF06571CE1ULL, "532333192.ogg"},
	{0x4A10CE9DD5F5B460ULL, "870680605.ogg"},
	{0x4A112353BD93C56CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft.msh"},
	{0x4A112953BD93CF5EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft.mmh"},
	{0x4A1131DE81677AA3ULL, "322600334.ogg"},
	{0x4A113553BD93E3F5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft.mao"},
	{0x4A115E3ECAD629C7ULL, "6033679_f.fxe"},
	{0x4A1166A2CC03D407ULL, "6125944_m.fxe"},
	{0x4A11AC527B995033ULL, "zz_followers_banter6159177.cl"},
	{0x4A11C3E05C862A72ULL, "hm_jakeson_hall.mor"},
	{0x4A11EB419677604FULL, "vxlt_hangedman_o1_0_0.vlm"},
	{0x4A11F2B8D019D201ULL, "art\\vfx\\materials\\chunky\\vfx_c_crater_h_weapon.mao"},
	{0x4A1210FB0BA94840ULL, "689270392.ogg"},
	{0x4A123BA7EC6FF546ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha1a.phy"},
	{0x4A1349B21A9ACFE9ULL, "6134907_f.fxe"},
	{0x4A13B30F8536C80BULL, "6001277_m.fxe"},
	{0x4A13C05B2567BD22ULL, "vxht_brothel_n_6b7_0_0.vlm"},
	{0x4A13F33639AF3BB5ULL, "6030608_m.fxe"},
	{0x4A13FB0F0B1B8521ULL, "6137273_m.fxe"},
	{0x4A14E3B5421BCC3EULL, "plt_lgt110pt_har_escape_rew.nss"},
	{0x4A14EC8EDB773986ULL, "lgt320_shw_ors_secret_nob6156449.cl"},
	{0x4A150042BC0E6C72ULL, "6037574_m.fxe"},
	{0x4A1525E99667FF2EULL, "hlac_lt_0_lowtown_docks_n_65534_6z.dds"},
	{0x4A1551CA73C82A63ULL, "character_character_p185.psh.fxo.dx11.assembly"},
	{0x4A15CF51222FF897ULL, "hightown_male_noble1_sta6170868.cl"},
	{0x4A164AFEB2F95DEFULL, "797393593.ogg"},
	{0x4A16A4AE5B8C6160ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comc_0n.dds"},
	{0x4A17199177CA70B6ULL, "vxht_bartrand_n_5zd_0_0.vlm"},
	{0x4A1731624C38F492ULL, "qun131_taskmaster6039731.cl"},
	{0x4A17F73024E5FB1EULL, "4886404.ogg"},
	{0x4A188CFDF218D211ULL, "6098075_m.fxe"},
	{0x4A18A30ACC9ADBD5ULL, "lowtown_male_urchin_sta6164193.cl"},
	{0x4A18CD3342EEDA10ULL, "6018631_f.fxe"},
	{0x4A18F8BFD0A9B992ULL, "vxlt_gallowstemplar_42z_0_0.vlm"},
	{0x4A1901F8B90ADCFCULL, "hlac_lt_gallowstemplar_n_2383_12z.dds"},
	{0x4A1909E7DB8E1F50ULL, "mag170_hugh6080732.cl"},
	{0x4A1987980415093EULL, "follower_barks6153634.cl"},
	{0x4A1A1EA7E6D40434ULL, "and100_anders6197199.cl"},
	{0x4A1B392D032CE304ULL, "vxht_mansion2_1mw_0_0.vlm"},
	{0x4A1B64501CCDCFBDULL, "11104639.ogg"},
	{0x4A1BD141EEE7D714ULL, "vxca_caverns2_3j7_0_0.vlm"},
	{0x4A1C06A6C424F2B0ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney_sa.dds"},
	{0x4A1C135D7A59BC1BULL, "zz_approval_debug6003938.cl"},
	{0x4A1C4ED468065473ULL, "6017490_m.fxe"},
	{0x4A1CEAF05B2E6DCFULL, "6120990_m.fxe"},
	{0x4A1D0A1813F6376FULL, "662321229.ogg"},
	{0x4A1DB9403148D2C5ULL, "ico_amulet_4.dds"},
	{0x4A1DDA84F67A1670ULL, "llac_lt_0_lowtown_n_65536_3.dds"},
	{0x4A1E06F316609B70ULL, "zz_mag_debug6044336.cl"},
	{0x4A1E10575FBC3F74ULL, "709124568.ogg"},
	{0x4A1E41DB12DBB6DAULL, "primitives_primitive_p2.vsh.fxo.dx11.assembly"},
	{0x4A1F140DFBDE69C6ULL, "26799371.ogg"},
	{0x4A1F3D7C710823EDULL, "gen00fl_aveline6139957.cl"},
	{0x4A1F70E15C60886CULL, "vxdr_primevalentrance_1qn_0_0.vlm"},
	{0x4A1FCBECF8767213ULL, "one000_ilen6057347.cl"},
	{0x4A20887E3E65F7A8ULL, "dae000_flemeth6013338.cl"},
	{0x4A20F20C53198290ULL, "368329405.ogg"},
	{0x4A2115791D73A16DULL, "lt_0_lowtown_docks_a3_n_223_0.mmh"},
	{0x4A2119791D73A7BEULL, "lt_0_lowtown_docks_a3_n_223_0.mao"},
	{0x4A2127791D73BFF3ULL, "lt_0_lowtown_docks_a3_n_223_0.msh"},
	{0x4A217F88BD0899DBULL, "257625165.ogg"},
	{0x4A22B624EAA9A544ULL, "fen110_fenris6101966.cl"},
	{0x4A22FB55389D26DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_loop.gad"},
	{0x4A23E399D90234CDULL, "6166811_m.fxe"},
	{0x4A247B0A0935B48BULL, "vxht_mansion2b_2ug_0_0.vlm"},
	{0x4A249E01E57F6BACULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_boardtop.phy"},
	{0x4A24CD61F22391E8ULL, "static_static_p34.psh.fxo.dx11"},
	{0x4A251B0C13670FAAULL, "vxlt_0_lowtown_a2_hfu_0_0.vlm"},
	{0x4A251C4E0EA4FF03ULL, "288339512.ogg"},
	{0x4A2531A51460C199ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_carv.phy"},
	{0x4A254BEC5D716260ULL, "550226893.ogg"},
	{0x4A25513000DB1359ULL, "452294249.ogg"},
	{0x4A255CC10AF6ABC9ULL, "llac_lt_gallowscourtyard_-131073_1.dds"},
	{0x4A25B1679E67407AULL, "gen00fl_merrill6108517.cl"},
	{0x4A25FDE432F19B30ULL, "vxht_mansion2b_21d_0_0.vlm"},
	{0x4A2631D2EEF2F1E8ULL, "one110_pryce6088776.cl"},
	{0x4A264565D468F940ULL, "aip_follower_fake_bethany.gda"},
	{0x4A2669CE5379335FULL, "vxw_sundermount_305_0_0.vlm"},
	{0x4A274D5B8B6DA5BCULL, "mag121_viveka6156514.cl"},
	{0x4A276B7DF9E2B680ULL, "6153102_m.fxe"},
	{0x4A276C834D2957E8ULL, "6111470_m.fxe"},
	{0x4A27EDD0DF74E413ULL, "vxlt_gallowstemplar_f_2jf_0_0.vlm"},
	{0x4A28480752EF9F8CULL, "vxw_blightlands_or_0_0.vlm"},
	{0x4A284FFFDD9D1864ULL, "6171898_m.fxe"},
	{0x4A28ACE97D47B7CFULL, "mag311_keran6041411.cl"},
	{0x4A28C249BA34304BULL, "vxht_keep_n_5u3_0_0.vlm"},
	{0x4A293D7ECB325E30ULL, "lla_ht_0_hightown_a3_-2_0.dds"},
	{0x4A2989C832661E49ULL, "hlac_w_blightlands_-131071_0x.dds"},
	{0x4A29BC20716717C9ULL, "vxlt_0_lowtown_a2_gze_0_0.vlm"},
	{0x4A29BC5BACF0D059ULL, "hlac_w_bonepit_0_11y.dds"},
	{0x4A2A5578181B480FULL, "61315319.ogg"},
	{0x4A2AE3EDDB6D3CACULL, "6016670_m.fxe"},
	{0x4A2AEBEDE986BF45ULL, "vxlt_0_lowtown_a2_n_c_hv3_0_0.vlm"},
	{0x4A2AF26DDFA7FF92ULL, "6127103_m.fxe"},
	{0x4A2B3746E7B80186ULL, "vxlt_gallowstemplar_3re_0_0.vlm"},
	{0x4A2B7ABAEE8C77E8ULL, "6118755_f.fxe"},
	{0x4A2C55C39532D811ULL, "27015493.ogg"},
	{0x4A2D26F2B8E62E3FULL, "art\\characters\\playercharacter\\humanmale\\hlh_mede.phy"},
	{0x4A2D37E60D8A3F06ULL, "hlac_w_woundedcoast_262151_0x.dds"},
	{0x4A2D3BEFBB376350ULL, "zz_mag_debug6053408.cl"},
	{0x4A2D7A9B4BF1D524ULL, "6125510_m.fxe"},
	{0x4A2D834B38198B85ULL, "toggle_gender.ncs"},
	{0x4A2D8DA414BBEABFULL, "lgt310_mis_burn_dump6180592.cl"},
	{0x4A2E570FD0BDFDAAULL, "6114844_m.fxe"},
	{0x4A2E587021F83F6CULL, "6174611_m.fxe"},
	{0x4A2E8F5F5FD71589ULL, "654978220.ogg"},
	{0x4A2EEE3CC132D304ULL, "vxlt_0_lowtown_docks_a3_n_bg7_0_0.vlm"},
	{0x4A2F5C13A7B93AA0ULL, "vxlt_darktown_2xh_0_0.vlm"},
	{0x4A2F983C08E1FAFEULL, "hf_gencr_merc_warrior.mor"},
	{0x4A2FA02CCAC218F3ULL, "hlac_lt_0_lowtown_docks_131070_4y.dds"},
	{0x4A2FD7072FA50779ULL, "hlac_w_woundedcoast_0_1y.dds"},
	{0x4A2FDA4DA09134CCULL, "vxlt_0_lowtown_a2_n_gn3_0_0.vlm"},
	{0x4A30338BEF146D15ULL, "aip_pridedemon_base.gda"},
	{0x4A3037003A21B2AFULL, "lt_0_lowtown_docks_n_126_0.mao"},
	{0x4A3043003A21C74CULL, "lt_0_lowtown_docks_n_126_0.mmh"},
	{0x4A3046F42B6496BEULL, "524824061.ogg"},
	{0x4A3049003A21D13EULL, "lt_0_lowtown_docks_n_126_0.msh"},
	{0x4A309B537CEE2DC0ULL, "6139832_m.fxe"},
	{0x4A30A4150953EE22ULL, "vxht_brothel_n_6h1_0_0.vlm"},
	{0x4A30CEF696C729B3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain.phy"},
	{0x4A317229A73AD45EULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.impa.evt"},
	{0x4A31B8907AEA0EAEULL, "hlac_w_encounter2_camp_-2_0y.dds"},
	{0x4A3221BF86E73BBCULL, "vxw_woundedcoast_1c9_0_0.vlm"},
	{0x4A3285588C3D7335ULL, "vxw_encounter2_r7_0_0.vlm"},
	{0x4A332C0226B0D594ULL, "6126511_m.fxe"},
	{0x4A335B58359B06AFULL, "6043600_m.fxe"},
	{0x4A338CE9055D282AULL, "vxlt_0_lowtown_a2_hcv_0_0.vlm"},
	{0x4A33AD237F998C01ULL, "6032992_m.fxe"},
	{0x4A35C76069B7F7FEULL, "242736002.ogg"},
	{0x4A3616B251788428ULL, "6033309_m.fxe"},
	{0x4A370A5721116CB6ULL, "328192744.ogg"},
	{0x4A3740F9EB07F891ULL, "zz_one_debug6099767.cl"},
	{0x4A37BC4CE5581632ULL, "6219600_m.fxe"},
	{0x4A37C88FE8F371B2ULL, "vxlt_hangedman_n_10o_0_0.vlm"},
	{0x4A37CC8F6820429DULL, "character_character_p129.psh.fxo.dx11"},
	{0x4A388929A7411989ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.impa.gad"},
	{0x4A38A8DE314F2189ULL, "gen00fl_sebastian6180014.cl"},
	{0x4A38D37267F51E13ULL, "default_board.tga"},
	{0x4A3934DEC79FB8FCULL, "hlac_dr_slavers_3217_10x.dds"},
	{0x4A395DCADB7CD7C5ULL, "vxlt_0_lowtown_a2_n_c_hq2_0_0.vlm"},
	{0x4A397D923EA011FBULL, "follower_banter6160764.cl"},
	{0x4A39EB9C2975FF47ULL, "hlac_lt_0_lowtown_65536_3x.dds"},
	{0x4A3A78E537178B5FULL, "vxlt_undercityencounter_18v_0_0.vlm"},
	{0x4A3AA9F9105B22ECULL, "665146884.ogg"},
	{0x4A3AF1AC8C17D59AULL, "6035922_f.fxe"},
	{0x4A3AFEBBAE65A3E4ULL, "dae000_orsino6085426.cl"},
	{0x4A3B5CA60DA26271ULL, "hlac_dr_0_deeproads_17038_7y.dds"},
	{0x4A3B67511F8A132CULL, "644212353.ogg"},
	{0x4A3BBD532654A3BEULL, "t1_tat_pur.tnt"},
	{0x4A3C2FE7A6FEA8F9ULL, "353416900.ogg"},
	{0x4A3C549B404CB96CULL, "vxdr_primevalentrance_1kl_0_0.vlm"},
	{0x4A3CC8ECC2B4162AULL, "6113944_m.fxe"},
	{0x4A3D376B6BE7BF8DULL, "492137171.ogg"},
	{0x4A3E109F013F5E40ULL, "and110cr_street_thug.utc"},
	{0x4A3E27F2B8F4A344ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mede.mmh"},
	{0x4A3E2BF2B8F4A997ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mede.mao"},
	{0x4A3E3DF2B8F4C8E6ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mede.msh"},
	{0x4A3E3FC5DA3F2631ULL, "966367161.ogg"},
	{0x4A3EC26EAAE2247EULL, "follower_banter6136737.cl"},
	{0x4A3F15F1FB4A06A2ULL, "vxht_brothel_n_6u2_0_0.vlm"},
	{0x4A3F30E410427201ULL, "zz_followers_banter6159063.cl"},
	{0x4A3F3B7B56182B00ULL, "hlac_ca_caverns2_5802_32z.dds"},
	{0x4A3F5A29A746E8D7ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.impa.ani"},
	{0x4A3FB8CE76AB5A15ULL, "dae000_merchant_mage_hf6164463.cl"},
	{0x4A3FEC8B1B7752E2ULL, "6016171_m.fxe"},
	{0x4A403AC31641A91BULL, "6170421_m.fxe"},
	{0x4A4070DA451D0041ULL, "vxdr_primevaltomb_34n_0_0.vlm"},
	{0x4A412E23B69B93F6ULL, "zz_pro_cutscenes.cnv"},
	{0x4A414ACA4324C52FULL, "hlac_lt_gallowscourtyard_-65538_2y.dds"},
	{0x4A419C17520740FCULL, "40993597.ogg"},
	{0x4A41AE9A44586522ULL, "6103686_m.fxe"},
	{0x4A4264F1D93F3859ULL, "6064578_f.fxe"},
	{0x4A42836121748FF5ULL, "hlac_w_bonepitmining_a2_0_9z.dds"},
	{0x4A42980D0C0588A9ULL, "6096316_m.fxe"},
	{0x4A42CB8D7EDCEFDEULL, "6197706_m.fxe"},
	{0x4A43FA2379A096D3ULL, "hlac_lt_darktown_-1_59y.dds"},
	{0x4A4490DBFC3EDBC0ULL, "one220_yevhen6120157.cl"},
	{0x4A452BC136DBD08CULL, "vxw_blightlands_up_0_0.vlm"},
	{0x4A453E537E487BF7ULL, "llac_ca_caverns1_24422_4.dds"},
	{0x4A45FD29CA74AAFAULL, "vxlt_warehouse_15w_0_0.vlm"},
	{0x4A462A62121C8E8BULL, "440464618.ogg"},
	{0x4A46788C55427B71ULL, "6103577_m.fxe"},
	{0x4A4683EE24911815ULL, "13503504.ogg"},
	{0x4A46AF2E6729B6CDULL, "m_dae170ar_gallows.wbd"},
	{0x4A4826BFE4F31B0CULL, "6217330_m.fxe"},
	{0x4A483B2FFE51CD71ULL, "244717161.ogg"},
	{0x4A485E74D278A8E8ULL, "6112555_f.fxe"},
	{0x4A487CB4F46471B6ULL, "qun221cr_qunari_guard.utc"},
	{0x4A48D5FB0BFEF183ULL, "6208210_m.fxe"},
	{0x4A498DF9C2046300ULL, "vxca_cave_163_0_1.vlm"},
	{0x4A49E65381757892ULL, "character_character_p92.psh.fxo"},
	{0x4A49F7DC13DF3D48ULL, "character_character_p72.psh.fxo.dx11"},
	{0x4A4A0E11E009ABE6ULL, "mcr370_negotiation6080178.cl"},
	{0x4A4A4081335091EFULL, "6085995_f.fxe"},
	{0x4A4B80FDA3BFA022ULL, "qun211_braggart6075663.cl"},
	{0x4A4C572701A61C5CULL, "292430552.ogg"},
	{0x4A4C7C975D772A75ULL, "ser221ip_lillies.utp"},
	{0x4A4CFB08E6BB5C0DULL, "lt_0_lowtown_docks_a3_151_0.phy"},
	{0x4A4D1D925979F504ULL, "zz_ave_debug6056693.cl"},
	{0x4A4E1452FEBFAD0AULL, "mag121_viveka6156509.cl"},
	{0x4A4E3E44E413E174ULL, "isa311_isabela6092490.cl"},
	{0x4A4EE3D95C9AC6F8ULL, "gen_im_trs_qun_sta_cld_r1.uti"},
	{0x4A4EFF6005A1F7ABULL, "6071101_f.fxe"},
	{0x4A4F25B7204C78E2ULL, "6111484_f.fxe"},
	{0x4A4F2AC98FEA35D7ULL, "vxht_0_hightown_n_5gy_0_0.vlm"},
	{0x4A4FEE6AFF94004DULL, "art\\vfx\\materials\\vfx_half_dot.mao"},
	{0x4A500FBC0A9D0694ULL, "6128711_m.fxe"},
	{0x4A501EE8ED55E1EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x8portcull.mao"},
	{0x4A502AE8ED55F58CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x8portcull.mmh"},
	{0x4A5030E8ED56007EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x8portcull.msh"},
	{0x4A50467DF21EFB07ULL, "837751815.ogg"},
	{0x4A506F70B69AC396ULL, "hlac_w_sundermount_-65532_0y.dds"},
	{0x4A509A55763E6A07ULL, "vxht_keep_n_62d_0_1.vlm"},
	{0x4A50E1053FA2D2B6ULL, "vxht_bartrand_n_5nh_0_0.vlm"},
	{0x4A50FC95444B30C0ULL, "hlac_w_bonepit_65539_0x.dds"},
	{0x4A510053BDC96DF9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft.phy"},
	{0x4A5117979AB8DADFULL, "art\\levels\\ztools\\default\\generic\\0cw_basic.msh"},
	{0x4A511B0AEED29109ULL, "hlac_lt_0_lowtown_a3_n_c_65535_8x.dds"},
	{0x4A511D979AB8E481ULL, "art\\levels\\ztools\\default\\generic\\0cw_basic.mmh"},
	{0x4A5129979AB8F922ULL, "art\\levels\\ztools\\default\\generic\\0cw_basic.mao"},
	{0x4A5132F3F187236FULL, "ss_hawkecu3.cut"},
	{0x4A5166A4DB1A02BCULL, "docks_fem_sailor_mob6176289.cl"},
	{0x4A516B98A714289BULL, "lgt220_shp_mag_guard6149505.cl"},
	{0x4A51CE688BE2FE29ULL, "24782617.ogg"},
	{0x4A522EDCBC02FE5BULL, "hlac_lt_0_lowtown_a3_n_c_131070_5y.dds"},
	{0x4A524A81F0C43162ULL, "lt_0_lowtown_docks_n_31_0.rml"},
	{0x4A524DBF7F40B72FULL, "dae021_fem_worshipper_sta6150358.cl"},
	{0x4A52A16A795C1CF4ULL, "hm_noble_gen_3.mor"},
	{0x4A52F66D550188CFULL, "6079021_m.fxe"},
	{0x4A52FA1CCFEEC6B1ULL, "fxa_deacti_dis.mao"},
	{0x4A52FCCBACABB31AULL, "gen00fl_sebastian6189427.cl"},
	{0x4A537D931E8CA4E3ULL, "zz_fol_debug6092318.cl"},
	{0x4A540A7AE5FD21B1ULL, "vxlt_backalley_g_1x1_0_1.vlm"},
	{0x4A555BA106376CBAULL, "lt_0_lowtown_docks_a3_n_237_0.mmh"},
	{0x4A555FA1063773F1ULL, "lt_0_lowtown_docks_a3_n_237_0.mao"},
	{0x4A55618F42A0A15FULL, "scaleform_sf_main_p1.vsh.fxo.dx11"},
	{0x4A556DA106378B00ULL, "lt_0_lowtown_docks_a3_n_237_0.msh"},
	{0x4A558E267E686526ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_floor_sa.dds"},
	{0x4A565C825692A6F3ULL, "245223834.ogg"},
	{0x4A566B5177CB41B6ULL, "vxlt_0_lowtown_docks_a3_n_d0c_0_1.vlm"},
	{0x4A57187E43DA787EULL, "vxht_0_hightown_a2_auy_0_0.vlm"},
	{0x4A579715CF7B98E0ULL, "vxw_bonepit_yc_0_0.vlm"},
	{0x4A57AC92C64DBB49ULL, "pro000_opening_pc6001220.cl"},
	{0x4A57AEC8C20C3B75ULL, "86704178.ogg"},
	{0x4A57C7A9CDF0A25BULL, "combat_damage_react.gda"},
	{0x4A584F13D0B426E8ULL, "6108595_m.fxe"},
	{0x4A585830736B9112ULL, "character_character_p45.psh.fxo"},
	{0x4A5938552438EA6CULL, "vxdr_primevalentrance_1eb_0_0.vlm"},
	{0x4A597B605C777FAEULL, "6097864_m.fxe"},
	{0x4A599F1A6CD196B7ULL, "vxlt_gallowstemplar_f_38u_0_0.vlm"},
	{0x4A5A048A205C5ACBULL, "vxdr_slavers_1b4_0_0.vlm"},
	{0x4A5A28B76D71AD4DULL, "6163984_m.fxe"},
	{0x4A5AC8CF1176F884ULL, "6092195_m.fxe"},
	{0x4A5AF9B7432C202BULL, "vxht_0_hightown_avi_0_0.vlm"},
	{0x4A5B39FFD87F70A5ULL, "6218033_m.fxe"},
	{0x4A5B41CD6571B8BFULL, "ave110cr_dbandit_leader.utc"},
	{0x4A5BB840A536A6E4ULL, "dae000ip_crowd_mf4_talk15.utp"},
	{0x4A5BCA4AA14A15E3ULL, "zz_mag_debug6051357.cl"},
	{0x4A5C354F0F6695F0ULL, "vxw_encounter1_de_0_0.vlm"},
	{0x4A5C417711F78A20ULL, "vxlt_undercityencounter_1iv_0_0.vlm"},
	{0x4A5C52CCF58270D9ULL, "docks_scene_dockworkers6176513.cl"},
	{0x4A5CC1EC54A76F7AULL, "176485461.ogg"},
	{0x4A5D46563185CA41ULL, "667427849.ogg"},
	{0x4A5E856BEA288BD4ULL, "6080817_m.fxe"},
	{0x4A5EC8B54259D689ULL, "plt_lgt110pt_har_escape_rew.plo"},
	{0x4A5F65F3E8711676ULL, "531545609.ogg"},
	{0x4A5F67D423E85DDDULL, "t3_wdg_b01.tnt"},
	{0x4A6002171CF2BE7DULL, "vxht_chantry_n_285_0_0.vlm"},
	{0x4A60EBF5FC50A76AULL, "6107450_f.fxe"},
	{0x4A62DFF32FAE9338ULL, "vxw_encounter2_bn_0_0.vlm"},
	{0x4A62E17D55EBE23BULL, "561028410.ogg"},
	{0x4A631EAFAC56D42AULL, "mag100_cullen6092278.cl"},
	{0x4A6353B503B2A96DULL, "vxlt_backalley_g_1wi_0_0.vlm"},
	{0x4A638394393F79C9ULL, "vxlt_0_lowtown_a2_gni_0_0.vlm"},
	{0x4A63AD99D881EA16ULL, "vxca_caverns2_7at_0_0.vlm"},
	{0x4A63AF80B0279DADULL, "800139437.ogg"},
	{0x4A63BF623C1DF9EFULL, "vxlt_gallowstemplar_5yw_0_0.vlm"},
	{0x4A6431D44B77B0A3ULL, "llac_lt_0_lowtown_docks_a3_-4_1.dds"},
	{0x4A649E9528C3AC93ULL, "96194943.ogg"},
	{0x4A657501E5B6BD77ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_boardtop.mmh"},
	{0x4A657901E5B6C384ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_boardtop.mao"},
	{0x4A658A5B35D090FEULL, "vxht_0_hightown_a2_apz_0_0.vlm"},
	{0x4A658B01E5B6E2E9ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_boardtop.msh"},
	{0x4A659C2C923452BDULL, "223022792.ogg"},
	{0x4A65A47CA1387CA3ULL, "6136218_m.fxe"},
	{0x4A66147A914E5F7EULL, "vxw_encounter2_j0_0_0.vlm"},
	{0x4A66C589B20D6850ULL, "gen00fl_aveline6132652.cl"},
	{0x4A66CA0D6561A992ULL, "character_character_p34.psh.fxo"},
	{0x4A670C3F44C0DBBFULL, "vxdr_primevalentrance_1kc_0_0.vlm"},
	{0x4A6764BEBA5AE28EULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.dura.evt"},
	{0x4A67D9DA2928CF66ULL, "ico_staff_5.dds"},
	{0x4A67EB64FBD13545ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_6z.dds"},
	{0x4A682E40F799386BULL, "gen00fl_fenris6108042.cl"},
	{0x4A687F197D104B0DULL, "vxht_mansion_n_700_0_0.vlm"},
	{0x4A6914747BEF9714ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeexit.ani"},
	{0x4A692D3E1004951EULL, "vxlt_0_lowtown_a2_n_gxk_0_0.vlm"},
	{0x4A69AAB8EE294CBEULL, "llac_lt_0_lowtown_a2_-262144_1.dds"},
	{0x4A69BADF5C19BAF5ULL, "fen110_fenris6100290.cl"},
	{0x4A6A84F691E9BCB8ULL, "gen00fl_sebastian6178375.cl"},
	{0x4A6A8BF2ACD109D0ULL, "vxht_0_hightown_n_bwy_0_0.vlm"},
	{0x4A6AA72C015CAE70ULL, "vxw_encounter1_ib_0_0.vlm"},
	{0x4A6AF196ACED4D44ULL, "206997829.ogg"},
	{0x4A6B3C8565F4E2E4ULL, "gen00fl_dog6117426.cl"},
	{0x4A6B69AA33391450ULL, "vxlt_gallowstemplar_n_43x_0_0.vlm"},
	{0x4A6B940C369EC7CEULL, "vxlt_hangedman_n_2xq_0_1.vlm"},
	{0x4A6BE6B135E8AF11ULL, "vxca_caverns1_mines_k2t_0_0.vlm"},
	{0x4A6BF7F457CA6BB9ULL, "vxlt_0_lowtown_docks_n_cp1_0_0.vlm"},
	{0x4A6D053CFF2666DDULL, "vxht_0_hightown_a3_c99_0_0.vlm"},
	{0x4A6D17610658EF3BULL, "hlac_lt_gallowstemplar_2391_6x.dds"},
	{0x4A6D817A744506E8ULL, "zz_qun_debug6028283.cl"},
	{0x4A6DA51B4114F40BULL, "pro000tr_barrier_hurlock.ncs"},
	{0x4A6DDF30344B2BF9ULL, "19137170.ogg"},
	{0x4A6E5BBEBA60F239ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.dura.gad"},
	{0x4A6E8746853BED4AULL, "529400220.ogg"},
	{0x4A6EE637AF984746ULL, "gridnorm.dds"},
	{0x4A6F6F47F4A8D829ULL, "6126395_f.fxe"},
	{0x4A6FF0FF66E88EDFULL, "vxlt_0_lowtown_docks_a3_d2j_0_0.vlm"},
	{0x4A7011747BF5B176ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeexit.gad"},
	{0x4A7078BF65845E82ULL, "hlac_ht_chantry_n_2534_48x.dds"},
	{0x4A70CADAAC9502F6ULL, "571076976.ogg"},
	{0x4A7151D021A4ABB8ULL, "vxw_encounter2_mo_0_0.vlm"},
	{0x4A716AD402D9F0DBULL, "ser270_templar6122511.cl"},
	{0x4A71DA178CAF1F6BULL, "vrc321_varric6097150.cl"},
	{0x4A725FCBDA51A1FDULL, "hlac_ht_0_hightown_a2_131072_0x.dds"},
	{0x4A72AD219ABFA88CULL, "w_blightlands_6_0.rml"},
	{0x4A731FB89CF54D0FULL, "gen00fl_aveline6151892.cl"},
	{0x4A7337882EE3B5DDULL, "dup_cutoff.dds"},
	{0x4A73590FE83A347AULL, "410723454.ogg"},
	{0x4A738DA87D82F5FBULL, "6096432_f.fxe"},
	{0x4A73A93129C99A24ULL, "vxca_caverns2_7w0_0_0.vlm"},
	{0x4A73B4EC7773F6D2ULL, "gen_im_ammo_quiver_07.uti"},
	{0x4A73B5874710F8A1ULL, "vxht_mansion2b_1ti_0_0.vlm"},
	{0x4A73C209F9D16597ULL, "vxca_cave_2jz_0_0.vlm"},
	{0x4A73D5D9D9BB76EBULL, "152174029.ogg"},
	{0x4A749AD55863F6DFULL, "hlac_dr_primevaltomb_9621_3z.dds"},
	{0x4A75112780F06C7FULL, "art\\vfx\\combat\\vfx_v_nature_act_c.impa.gad"},
	{0x4A756CBEBA672D87ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.dura.ani"},
	{0x4A766DE08004CFD2ULL, "49368473.ogg"},
	{0x4A770656A84608B8ULL, "vfx_vfx_p1.psh.fxo"},
	{0x4A772C747BFBFDD5ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeexit.evt"},
	{0x4A773CD9DD48CF12ULL, "dre000_bethany6037320.cl"},
	{0x4A77503F600FADCEULL, "vxlt_0_lowtown_docks_a3_n_cxm_0_0.vlm"},
	{0x4A7829510BB11196ULL, "6140131_m.fxe"},
	{0x4A78D141FB533EE9ULL, "772122705.ogg"},
	{0x4A791BAB5AD39CAAULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdbl.mao"},
	{0x4A791FAB5AD3A3F9ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdbl.mmh"},
	{0x4A7929AB5AD3B4E7ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdbl.msh"},
	{0x4A792DC94652F977ULL, "220668366.ogg"},
	{0x4A7968A7A99E44A7ULL, "hlac_ht_chantry_n_2534_13y.dds"},
	{0x4A79A0AF5FBD3128ULL, "one100_ht_banter6060906.cl"},
	{0x4A79AF01CA067862ULL, "hlac_lt_0_lowtown_docks_n_-131075_5z.dds"},
	{0x4A7A7FC8F3BE8313ULL, "298354532.ogg"},
	{0x4A7B062780F4C56CULL, "art\\vfx\\combat\\vfx_v_nature_act_c.impa.evt"},
	{0x4A7BB70E11CFB0E0ULL, "vxlt_0_lowtown_n_f1g_0_0.vlm"},
	{0x4A7BDBA5CB2BC6F2ULL, "vxht_mansion2b_3dd_0_0.vlm"},
	{0x4A7BF854D833A453ULL, "wm_ico_clinic.dds"},
	{0x4A7C25BE7C283CADULL, "vxw_encounter2_70_0_0.vlm"},
	{0x4A7C2B86D7062891ULL, "gen00fl_carver6102949.cl"},
	{0x4A7CA0DA5E7FAFD6ULL, "vxca_caverns1_gjz_0_0.vlm"},
	{0x4A7CC2EB06835C06ULL, "qun111_ketojan6039806.cl"},
	{0x4A7CD9CDBEAB568FULL, "zz_mer_debug6045636.cl"},
	{0x4A7CE888635DB5C4ULL, "vxw_woundedcoast_1rb_0_0.vlm"},
	{0x4A7CF0C387B3D37FULL, "vxlt_0_lowtown_docks_n_aq6_0_0.vlm"},
	{0x4A7CF18F84F73C81ULL, "gen_im_npc_wep_sta_genb.uti"},
	{0x4A7CF306D19460EFULL, "836172886.ogg"},
	{0x4A7D1E979ADE21FEULL, "art\\levels\\ztools\\default\\generic\\0cw_basic.phy"},
	{0x4A7D67102143090CULL, "zz_warriorslate_trans.cut"},
	{0x4A7D90B4B0D45DD4ULL, "lt_0_lowtown_docks_a3_n_21_0.rml"},
	{0x4A7D99B9CF6BBD79ULL, "mrl150_dalish_guard6068743.cl"},
	{0x4A7E5D280BA62F68ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb3.mmh"},
	{0x4A7E606DFD121330ULL, "171360529.ogg"},
	{0x4A7E6579E6F8B62CULL, "6023951_m.fxe"},
	{0x4A7E7B280BA66192ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb3.msh"},
	{0x4A7EFAE0682A48C6ULL, "hlac_lt_0_lowtown_docks_a3_-4_6x.dds"},
	{0x4A7F111EA22C00E4ULL, "1024519756.ogg"},
	{0x4A7FC3AD139AC438ULL, "vxw_encounter2_hl_0_0.vlm"},
	{0x4A800811FBCE118DULL, "vxlt_0_lowtown_docks_b6l_0_1.vlm"},
	{0x4A80376EE79EDA6DULL, "vxlt_backalley_g_1yg_0_0.vlm"},
	{0x4A8055CA6F0B79AAULL, "sw_p_dark.mao"},
	{0x4A80DF1DB2000680ULL, "gen00fl_isabela_init6127676.cl"},
	{0x4A811AC3D9A80BF9ULL, "486150724.ogg"},
	{0x4A815000C2CCFD5FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetop.msh"},
	{0x4A815600C2CD0701ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetop.mmh"},
	{0x4A816200C2CD1BA2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetop.mao"},
	{0x4A81654E8BF49083ULL, "vxlt_darktown_4f7_0_0.vlm"},
	{0x4A81D939215909F6ULL, "art\\characters\\creatures\\strider\\animation\\die_v1.evt"},
	{0x4A81ED01F2142B7DULL, "hlac_lt_gallowstemplar_n_2391_2y.dds"},
	{0x4A821B0E1BBFB2A4ULL, "vxca_caverns2_7l1_0_0.vlm"},
	{0x4A82F834753A907EULL, "vxw_encounter2_p2_0_0.vlm"},
	{0x4A83729772606222ULL, "gen00fl_varric6191285.cl"},
	{0x4A83E8A16B074159ULL, "6118592_f.fxe"},
	{0x4A8429B2829592E5ULL, "6139248_m.fxe"},
	{0x4A843C09F168BCA0ULL, "dae000_arishok6067658.cl"},
	{0x4A8492810C603E7CULL, "6120157_m.fxe"},
	{0x4A84A01A4533E16BULL, "690885676.ogg"},
	{0x4A84A7E2C9FBC699ULL, "6135077_m.fxe"},
	{0x4A84D902C39CE123ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.anb"},
	{0x4A850183C5D3D55DULL, "296364319.ogg"},
	{0x4A8516847C50321AULL, "vxw_woundedcoast_19y_0_0.vlm"},
	{0x4A854E6D0C7CBD98ULL, "150602824.ogg"},
	{0x4A8574A117A58EEAULL, "rdr110dg_party_comments6094679.cl"},
	{0x4A8632A29109A9B6ULL, "hlac_lt_0_lowtown_a2_n_c_1_6y.dds"},
	{0x4A866CC339A6FC76ULL, "vxlt_undercityserialkiller_1kv_0_0.vlm"},
	{0x4A86DE7C5324DC5EULL, "llac_lt_undercitydungeon_-1_25.dds"},
	{0x4A87184B3AEB7F9DULL, "157228912.ogg"},
	{0x4A87730C842DF3E1ULL, "345899492.ogg"},
	{0x4A87B92E67615EF7ULL, "m_dae170ar_gallows.bnk"},
	{0x4A87D039215D6581ULL, "art\\characters\\creatures\\strider\\animation\\die_v1.gad"},
	{0x4A88260DF3A0C824ULL, "6108029_m.fxe"},
	{0x4A88EE278100D9C5ULL, "art\\vfx\\combat\\vfx_v_nature_act_c.impa.ani"},
	{0x4A8943B09FA22295ULL, "mag110_samson6053552.cl"},
	{0x4A89761282BFCA2CULL, "dae000cr_grand_cleric6124732.cl"},
	{0x4A89A7A3A3BD3509ULL, "457867877.ogg"},
	{0x4A89C10BC2C2955EULL, "6120112_f.fxe"},
	{0x4A89E8EF52431148ULL, "dae000_hate_follower_react6167491.cl"},
	{0x4A89E8F6E31297DDULL, "vxht_0_hightown_a3_c77_0_0.vlm"},
	{0x4A8A4165677E1A59ULL, "106147318.ogg"},
	{0x4A8A55E1539755EAULL, "vxht_0_hightown_a3_cd9_0_0.vlm"},
	{0x4A8AC6854013CE2EULL, "6064651_m.fxe"},
	{0x4A8B75172C1CC512ULL, "hlac_lt_gallowscourtyard_131069_0x.dds"},
	{0x4A8C4781F0F5A45EULL, "lt_0_lowtown_docks_n_31_0.anb"},
	{0x4A8CAA1B412FADF2ULL, "pro000tr_barrier_hurlock.utt"},
	{0x4A8CC9FD2B764DEAULL, "dae000_bodhan_letter6144315.cl"},
	{0x4A8CD6D58D75F66DULL, "6072369_m.fxe"},
	{0x4A8D0E92CD941569ULL, "vxdr_0_deeproads_fbo_0_0.vlm"},
	{0x4A8D5C97D5EAEF53ULL, "art\\vfx\\creatures\\fxc_pridedemon_c.mmh"},
	{0x4A8D60A2D9A63291ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic_1.mao"},
	{0x4A8D9E4520C15D23ULL, "767630840.ogg"},
	{0x4A8DA5BA4DC35D8AULL, "633707222.ogg"},
	{0x4A8E18EB89D08B4BULL, "6179044_m.fxe"},
	{0x4A8ED1392163862FULL, "art\\characters\\creatures\\strider\\animation\\die_v1.ani"},
	{0x4A8EE2A37D1E1B31ULL, "vxca_caverns1_mines_ih8_0_0.vlm"},
	{0x4A8FCC1B6A1A61D5ULL, "vxw_bonepitmining_a2_ny_0_0.vlm"},
	{0x4A900D654919B9E7ULL, "306881243.ogg"},
	{0x4A9027CBC56F6115ULL, "59669129.ogg"},
	{0x4A902CCACBA7046AULL, "309198363.ogg"},
	{0x4A9034D09F19577EULL, "gen00fl_sebastian6174038.cl"},
	{0x4A9038F46E919130ULL, "37714354.ogg"},
	{0x4A904A6AB69B1ACFULL, "vxw_bonepitmining_a2_5z_0_0.vlm"},
	{0x4A90AC34A24697E1ULL, "w_encounter2_48_0.phy"},
	{0x4A915DA067CC93DAULL, "mrl351ip_at_sundermount.utp"},
	{0x4A91A35F763228FFULL, "hlac_w_bonepitmining_a2_131073_0x.dds"},
	{0x4A91A59501473A88ULL, "qcr000_varric6066268.cl"},
	{0x4A91A73ED0B36006ULL, "864196822.ogg"},
	{0x4A9255E2FC13ED0FULL, "141703270.ogg"},
	{0x4A925BD2515C98D9ULL, "6127548_m.fxe"},
	{0x4A928AE26849166CULL, "6206862_m.fxe"},
	{0x4A92B602C3A8E34CULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.mmh"},
	{0x4A92C0B3E1432A44ULL, "mag131_keran6086172.cl"},
	{0x4A92F4AD2E239251ULL, "vxht_0_hightown_a6k_0_0.vlm"},
	{0x4A93C7F6971B9B28ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain.mmh"},
	{0x4A93D3F6971BAFCBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain.mao"},
	{0x4A93E5F6971BCE52ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain.msh"},
	{0x4A94053B01034FCAULL, "6135126_m.fxe"},
	{0x4A9423F2DA9872F0ULL, "art\\characters\\weapons\\mauls\\w_mal_ml101.mao"},
	{0x4A946A85FA0434F2ULL, "589305310.ogg"},
	{0x4A9482D4E5E6DE9EULL, "vxlt_0_lowtown_a2_n_gon_0_0.vlm"},
	{0x4A9484BB99FEDC93ULL, "vxw_sundermount_379_0_1.vlm"},
	{0x4A94C12B0B04822BULL, "vxht_0_hightown_ajm_0_0.vlm"},
	{0x4A94D13B385A499BULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_3y.dds"},
	{0x4A9505BE10127DA3ULL, "art\\characters\\creatures\\dragon\\animation\\appr.evt"},
	{0x4A954C1BC4FBF69FULL, "vxht_playershouse_n_1o8_0_0.vlm"},
	{0x4A95618E111FCCAAULL, "gen00fl_isabela6101086.cl"},
	{0x4A95AB5DCEF1BA0FULL, "ht_0_hightown_a3_35_0.anb"},
	{0x4A96210244474A86ULL, "vxca_caverns2_62d_0_0.vlm"},
	{0x4A97B534A24CC5FDULL, "w_encounter2_48_0.rml"},
	{0x4A97C30194E87540ULL, "6125868_m.fxe"},
	{0x4A98FBE2A0513FEEULL, "vxw_bonepitmining_a2_244_0_0.vlm"},
	{0x4A9928523CEDC74DULL, "vxlt_undercitydungeon_11v_0_0.vlm"},
	{0x4A9944E2824E80FAULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_qunguard_arishokaddress.gad"},
	{0x4A995CD48A58A22DULL, "vxw_blightlands_a9_0_0.vlm"},
	{0x4A99CACE7F30A4F9ULL, "vxlt_warehouse_au_0_0.vlm"},
	{0x4A9A9CC694B6B8FDULL, "857727887.ogg"},
	{0x4A9B37FA7ABFE7CBULL, "mrl000_marethari6117702.cl"},
	{0x4A9BBCB15E1EEF37ULL, "6160774_m.fxe"},
	{0x4A9BDEBE10185A24ULL, "art\\characters\\creatures\\dragon\\animation\\appr.gad"},
	{0x4A9C7C4D4059AEE6ULL, "661818312.ogg"},
	{0x4A9CC87E5D534D1AULL, "1006781898.ogg"},
	{0x4A9CEBCA257AA7DFULL, "gen_im_npc_arm_hum_glv_gwdb.uti"},
	{0x4A9DB7CD5198E094ULL, "mer221_sabin6047336.cl"},
	{0x4A9DCB8F5023C862ULL, "one220_yevhen6135896.cl"},
	{0x4A9E28D5F6446954ULL, "l_lgt311ar_rand_sewer_1_nt.wbd"},
	{0x4A9E49086FE0C183ULL, "vxlt_darktown_4h9_0_0.vlm"},
	{0x4A9E8047C68D8C2FULL, "6076349_m.fxe"},
	{0x4A9F8C4837E170AFULL, "ht_brothel_n_9156_lc.mmh"},
	{0x4A9FA1CF616BE1CDULL, "vxca_caverns1_cp3_0_0.vlm"},
	{0x4AA01696E91A5502ULL, "6129336_m.fxe"},
	{0x4AA0803A69255D90ULL, "311210121.ogg"},
	{0x4AA0B5B5FFEBB4FCULL, "dm_amb_commoner2.mor"},
	{0x4AA0D25E535F41B3ULL, "gen00fl_dog6167635.cl"},
	{0x4AA0D72D6C56468BULL, "mer240_jansen6045107.cl"},
	{0x4AA0DC6C2336FC97ULL, "lgt210_qun_blk_qunari6142316.cl"},
	{0x4AA1EC54E80E1A16ULL, "vxlt_gamlenshouse_n_b4_0_0.vlm"},
	{0x4AA210FAE0063E09ULL, "mag100_karras6034694.cl"},
	{0x4AA232F80C93FA04ULL, "hlac_lt_0_lowtown_n_-65536_6x.dds"},
	{0x4AA23E834EBD4E0FULL, "6090194_m.fxe"},
	{0x4AA2DAEF5EC4C363ULL, "mag120_macha6080679.cl"},
	{0x4AA2EDBE101E9202ULL, "art\\characters\\creatures\\dragon\\animation\\appr.ani"},
	{0x4AA331D243E67829ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_basic.phy"},
	{0x4AA344C046D2DF35ULL, "vxlt_0_lowtown_a3_n_c_h2j_0_0.vlm"},
	{0x4AA4B367DA200D4BULL, "gen_im_arm_cht_nobd.uti"},
	{0x4AA4B6F39FBACF25ULL, "6160783_m.fxe"},
	{0x4AA4CDAA606E59AEULL, "vxca_caverns2prc_4od_0_0.vlm"},
	{0x4AA531CDED7A9AE1ULL, "vxdr_0_deeproads_ebx_0_0.vlm"},
	{0x4AA53595902960F6ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_012a_0s.dds"},
	{0x4AA6064C6880D661ULL, "hlac_lt_0_lowtown_a2_n_c_65537_3x.dds"},
	{0x4AA6523DC3780FC7ULL, "298396646.ogg"},
	{0x4AA694757F1B2700ULL, "6053686_m.fxe"},
	{0x4AA75E5D12FEA884ULL, "vxht_bartrand_n_5o9_0_0.vlm"},
	{0x4AA7CEB17C4EBAADULL, "vxw_blightlands_j8_0_0.vlm"},
	{0x4AA86CC68CE65756ULL, "6067767_f.fxe"},
	{0x4AA89D95C79242F5ULL, "vxlt_gallowstemplar_f_46z_0_0.vlm"},
	{0x4AA8A2CF2B567D78ULL, "lt_gallowsprison_n_3567.rml"},
	{0x4AA8AA6742F8C37AULL, "gen00fl_anders6096472.cl"},
	{0x4AA8ECFB8EB4D060ULL, "ht_chantry_n_2534_ll.mmh"},
	{0x4AA91CC552AF68EBULL, "6000762_m.fxe"},
	{0x4AA9364CDFCC51B3ULL, "vxlt_0_lowtown_docks_a3_n_czz_0_0.vlm"},
	{0x4AA94504B560CEE6ULL, "dae000_arishok6080479.cl"},
	{0x4AA98E86C80F0789ULL, "151018043.ogg"},
	{0x4AA9B1B4B0F9D0D4ULL, "lt_0_lowtown_docks_a3_n_21_0.anb"},
	{0x4AA9CD113DF367ECULL, "follower_banter6151233.cl"},
	{0x4AA9F24CB1804669ULL, "vxdr_0_deeproads_fhm_0_0.vlm"},
	{0x4AA9FCE108E635EDULL, "vxlt_undercityserialkiller_22i_0_0.vlm"},
	{0x4AAA71C4661110B3ULL, "ss_merrillfire.cut"},
	{0x4AAA7645D597F863ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_adjust_talk.evt"},
	{0x4AAAE6BC0EB2EA9DULL, "lowtown_male_alienage1_sta6162791.cl"},
	{0x4AAB4717B7DB6406ULL, "6196824_m.fxe"},
	{0x4AAB74253F9C74CEULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\pine_sa.dds"},
	{0x4AABD062D06F4C46ULL, "vxlt_0_lowtown_docks_c6o_0_0.vlm"},
	{0x4AAC2B711E38B55FULL, "hlac_lt_0_lowtown_docks_n_131070_2x.dds"},
	{0x4AAC50E82BC46192ULL, "mag100_thrask.cnv"},
	{0x4AAC81D710371ECFULL, "mcr371_hostile_aveline6131937.cl"},
	{0x4AACEC2BF7F9C994ULL, "vxht_mansion_n_6f2_0_0.vlm"},
	{0x4AAD5700C2F2447EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetop.phy"},
	{0x4AAD5FE2825F57E0ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_qunguard_arishokaddress.ani"},
	{0x4AAD94AF9B380223ULL, "mag110ip_vincento_store.utp"},
	{0x4AAE464E84F2E3E2ULL, "llac_lt_gallowstemplar_f_2388_4.dds"},
	{0x4AAF22D5A3908FFBULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.impa.ani"},
	{0x4AAF29140AC2E92DULL, "516027976.ogg"},
	{0x4AAF2971B197BA0DULL, "cd_one300dg_lia.bnk"},
	{0x4AAF5BF64E5F512BULL, "115387264.ogg"},
	{0x4AAF5C79A25F6526ULL, "vxht_playershouse_n_1gu_0_0.vlm"},
	{0x4AAFC06FF043006BULL, "art\\vfx\\combat\\vfx_a_disengage_act_c.mmh"},
	{0x4AB04C9822C50CE8ULL, "art\\levels\\races\\proxy\\_textures\\ter_0_d.dds"},
	{0x4AB0532AFDE68211ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_range.ani"},
	{0x4AB05638C8E1A637ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_003a_0d.dds"},
	{0x4AB072E82BC884EAULL, "mag100_thrask.bnk"},
	{0x4AB09215C1B4CD94ULL, "lowtown_male_lowlife1_sta6164961.cl"},
	{0x4AB0BF1C69C2167BULL, "w_encounter3_163_0.msh"},
	{0x4AB0D11C69C234C6ULL, "w_encounter3_163_0.mao"},
	{0x4AB0DB537A07E123ULL, "6135601_f.fxe"},
	{0x4AB0DD1C69C24965ULL, "w_encounter3_163_0.mmh"},
	{0x4AB0E9BC54193D5CULL, "hlac_lt_0_lowtown_a2_1_2x.dds"},
	{0x4AB14F45D59DD4E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_adjust_talk.gad"},
	{0x4AB1E591CDD2B812ULL, "6171126_m.fxe"},
	{0x4AB34F5CE5C50E80ULL, "981787475.ogg"},
	{0x4AB37EA7D54A96A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_closed.ani"},
	{0x4AB43E4E6CD9F67DULL, "vxlt_0_lowtown_a3_n_c_f5i_0_0.vlm"},
	{0x4AB43FFFAF7B9CA7ULL, "zz_dae_debug6029302.cl"},
	{0x4AB4AD18DE8EF084ULL, "6154712_m.fxe"},
	{0x4AB4F5F62FFD4A26ULL, "244310660.ogg"},
	{0x4AB503DBECA559D4ULL, "hlac_lt_0_lowtown_n_65537_3y.dds"},
	{0x4AB5133B87FF6509ULL, "llac_ht_chantry_2534_20.dds"},
	{0x4AB520620C520497ULL, "6018536_m.fxe"},
	{0x4AB53BDF13C8A4BBULL, "hlac_lt_gallowstemplar_f_7198_7x.dds"},
	{0x4AB608D4AE0AD3F6ULL, "6121318_m.fxe"},
	{0x4AB627B5622D41EEULL, "vxlt_0_lowtown_a2_n_fhr_0_0.vlm"},
	{0x4AB6355EBF2EA805ULL, "hlac_ht_0_hightown_a2_-65537_3z.dds"},
	{0x4AB6408E6E44D32DULL, "vxw_blightlands_o7_0_0.vlm"},
	{0x4AB6634EB3AE250EULL, "6114040_m.fxe"},
	{0x4AB67A95CB15581CULL, "hlac_w_woundedcoast_-262141_0z.dds"},
	{0x4AB67AFA0EF86522ULL, "dae021_male_worshipper_sta6154420.cl"},
	{0x4AB6BF7388F9920AULL, "vxht_bartrand_n_6qf_0_0.vlm"},
	{0x4AB701A77BB5BE09ULL, "81111515.ogg"},
	{0x4AB751DED916F471ULL, "6125185_f.fxe"},
	{0x4AB75E2AFDECB3B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_range.gad"},
	{0x4AB77CF51F1ADF0FULL, "6096942_f.fxe"},
	{0x4AB7AB7B72168A0EULL, "163207315.ogg"},
	{0x4AB82F55D202ED95ULL, "lt_0_lowtown_docks_188_0.msh"},
	{0x4AB83155D202F10BULL, "lt_0_lowtown_docks_188_0.mmh"},
	{0x4AB83D55D20305A8ULL, "lt_0_lowtown_docks_188_0.mao"},
	{0x4AB85E45D5A40CC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_adjust_talk.ani"},
	{0x4AB90D1489619994ULL, "1060431425.ogg"},
	{0x4AB9683701E920D5ULL, "lowtown_male_lowlife2_mob6169191.cl"},
	{0x4AB974F6495BB791ULL, "fen110_fenris6117962.cl"},
	{0x4ABA99A7D550E32BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_closed.gad"},
	{0x4ABAA8E6829A3C08ULL, "474509588.ogg"},
	{0x4ABAEC807C54DBFCULL, "351622304.ogg"},
	{0x4ABB3B20F1FF56C3ULL, "511496612.ogg"},
	{0x4ABBE2FCA1729CECULL, "mag320_bsc_banter.cnv"},
	{0x4ABC29D8694EFE49ULL, "dae300_bran6144944.cl"},
	{0x4ABC336B5DDFCE4AULL, "vxlt_foundry_n_yq_0_0.vlm"},
	{0x4ABC3D0F567E21D5ULL, "llac_lt_0_lowtown_docks_n_-262144_2.dds"},
	{0x4ABC7332F0D94DBFULL, "llac_dr_primevaltomb_659_3.dds"},
	{0x4ABC8D5435056096ULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_qnari_isabellasreturn.ani"},
	{0x4ABCA16FF04D55D0ULL, "art\\vfx\\combat\\vfx_a_disengage_act_c.anb"},
	{0x4ABCAC3A0E181ECCULL, "dae000cr_grand_cleric6147641.cl"},
	{0x4ABCB867E8C4BC85ULL, "1027153631.ogg"},
	{0x4ABD1CCE9331132CULL, "6103683_f.fxe"},
	{0x4ABD2AD5A39CDAF2ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.impa.evt"},
	{0x4ABD66F2BAB68B15ULL, "vxlt_hangedman_26f_0_0.vlm"},
	{0x4ABE4D2DBA6659AAULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.anb"},
	{0x4ABE5B2AFDF2CD18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\medium_range.evt"},
	{0x4ABE68DF940DF76FULL, "6058943_m.fxe"},
	{0x4ABE780907404D2DULL, "plt_cod_hst_the_coterie.plo"},
	{0x4ABEC4FCA1749F1CULL, "mag320_bsc_banter.bnk"},
	{0x4ABF0C0704FEB6FAULL, "hlac_ca_cavern1_haunted_24424_17x.dds"},
	{0x4ABF124D363F8C89ULL, "hlac_w_encounter2_camp_262142_0z.dds"},
	{0x4ABF8A666B51CFA4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\wk_p_rf.ani"},
	{0x4ABFA2B52042665BULL, "vxw_sundermount_23e_0_0.vlm"},
	{0x4AC0576EBFC8D092ULL, "6174226_m.fxe"},
	{0x4AC0DA7377337AC6ULL, "6208209_f.fxe"},
	{0x4AC1426E9182841FULL, "hlac_w_sundermount_-393214_0y.dds"},
	{0x4AC15D9ACF03D336ULL, "666607238.ogg"},
	{0x4AC1885DCF16B9B3ULL, "ht_0_hightown_a3_35_0.rml"},
	{0x4AC1ED925A82DB01ULL, "364801887.ogg"},
	{0x4AC2045CD68AE116ULL, "6172329_m.fxe"},
	{0x4AC218B5073D337CULL, "6016883_m.fxe"},
	{0x4AC24EC5F4BB2CF7ULL, "follower_banter6136330.cl"},
	{0x4AC2A2B335DD6842ULL, "ht_0_hightown_n_34_0_ll.mmh"},
	{0x4AC2C995F29505DAULL, "6134749_m.fxe"},
	{0x4AC2FA7A6A8D5F31ULL, "hlac_lt_0_lowtown_a2_n_131071_7y.dds"},
	{0x4AC320C755C12045ULL, "495927528.ogg"},
	{0x4AC349D5A3A17A85ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.impa.gad"},
	{0x4AC35E9FD765DD47ULL, "6126454_f.fxe"},
	{0x4AC38F86D6A34DFAULL, "vxlt_undercitydungeon_yx_0_0.vlm"},
	{0x4AC3D1A21551BC22ULL, "w_bonepitdestroyed_a3_215_0.msh"},
	{0x4AC3D3A21551BFD8ULL, "w_bonepitdestroyed_a3_215_0.mmh"},
	{0x4AC3DFA21551D47BULL, "w_bonepitdestroyed_a3_215_0.mao"},
	{0x4AC416BD3217B070ULL, "vxw_bonepitdestroyed_a3_21s_0_0.vlm"},
	{0x4AC4A9E1DFB99C91ULL, "lt_gallowstemplar_n_2392.rml"},
	{0x4AC4AD1AF0815844ULL, "gen00fl_varric6063861.cl"},
	{0x4AC4B26B603AEBADULL, "vxw_blightlands_p6_0_0.vlm"},
	{0x4AC520655512EE79ULL, "vxlt_warehouse_pz_0_0.vlm"},
	{0x4AC587666B563606ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\wk_p_rf.gad"},
	{0x4AC59EF9ACB8359DULL, "414254647.ogg"},
	{0x4AC5B08A1D9504CCULL, "vxlt_0_lowtown_a2_gye_0_0.vlm"},
	{0x4AC5BD3926FF14B1ULL, "vxca_caverns1_mines_j5d_0_0.vlm"},
	{0x4AC5C6A26A821EDCULL, "hlac_ht_0_hightown_2_1x.dds"},
	{0x4AC5EA2B0D5B411CULL, "mag311_keran6041494.cl"},
	{0x4AC60309592E16E3ULL, "vxw_sundermount_37i_0_1.vlm"},
	{0x4AC6A3606E47C5FCULL, "6060824_m.fxe"},
	{0x4AC7A30D2F897163ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_5y.dds"},
	{0x4AC7C1E8D47311B7ULL, "6126372_m.fxe"},
	{0x4AC80EE628390D33ULL, "hlac_x_narrativeroom_0_0x.dds"},
	{0x4AC8474C19CFA6D6ULL, "gen00fl_aveline6106947.cl"},
	{0x4AC88F2521E32053ULL, "hlac_ca_caverns2_5802_61y.dds"},
	{0x4AC8B9EA0E486EA5ULL, "6081666_m.fxe"},
	{0x4AC95B6F9F5E93A4ULL, "hlac_w_sundermount_262138_0y.dds"},
	{0x4AC98A8B008486D6ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_ltr.anb"},
	{0x4AC9BB4B31BF9E03ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_fire01.mao"},
	{0x4ACA56D7C9D65301ULL, "6113280_m.fxe"},
	{0x4ACAEB86AAB4805FULL, "vxlt_undercityencounter_1jx_0_0.vlm"},
	{0x4ACB16D5F66B38EEULL, "l_lgt311ar_rand_sewer_1_nt.bnk"},
	{0x4ACB3E806AB44B47ULL, "mcr310_argument6075825.cl"},
	{0x4ACB6A0A2805E86BULL, "93242463.ogg"},
	{0x4ACBCEFA51D8E816ULL, "92979662.ogg"},
	{0x4ACC26CAEE98EFDEULL, "plt_cod_ltr_ser_alriks_pape.nss"},
	{0x4ACC4033864B9626ULL, "vxht_playershouse_n_1qw_0_0.vlm"},
	{0x4ACC51237CA10CD9ULL, "203548977.ogg"},
	{0x4ACCB0ECDE49CA4EULL, "gen00fl_bethany6097833.cl"},
	{0x4ACD45A63B8B971BULL, "ss_crazy_f_1.bnk"},
	{0x4ACD5E02EC3EF284ULL, "6081722_m.fxe"},
	{0x4ACDA180FE65C7F5ULL, "mag360_samson6042074.cl"},
	{0x4ACE08AE72AC04AFULL, "6160639_m.fxe"},
	{0x4ACE149212387EDBULL, "vxw_sundermount_24d_0_0.vlm"},
	{0x4ACE3946042A98D5ULL, "831288249.ogg"},
	{0x4ACE3FDA22F07272ULL, "follower_banter6164561.cl"},
	{0x4ACE5F70E2B7345EULL, "vxht_0_hightown_n_cbd_0_1.vlm"},
	{0x4ACE627980863B37ULL, "492308007.ogg"},
	{0x4ACECBA61D41FB24ULL, "vxlt_gamlenshouse_n_bv_0_0.vlm"},
	{0x4ACEF60C78CE38E4ULL, "vxlt_0_lowtown_a3_n_c_gon_0_1.vlm"},
	{0x4ACF1AD2440B8C4EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_basic.mmh"},
	{0x4ACF26D2440BA0E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_basic.mao"},
	{0x4ACF34D2440BB83CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_basic.msh"},
	{0x4ACF546EB7D87468ULL, "6018728_m.fxe"},
	{0x4ACF5DC48F8878A8ULL, "6104951_m.fxe"},
	{0x4ACF8E865F768074ULL, "vxca_caverns1_icm_0_0.vlm"},
	{0x4ACFAB7DCB3C539AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocke.mmh"},
	{0x4ACFAF7DCB3C5AD1ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocke.mao"},
	{0x4ACFBD7DCB3C72E0ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocke.msh"},
	{0x4AD01D9E95DB6768ULL, "vxht_0_hightown_a3_b9m_0_0.vlm"},
	{0x4AD07639C880F996ULL, "6175229_m.fxe"},
	{0x4AD1156EA6A0B2FFULL, "one110_meeran6141082.cl"},
	{0x4AD185750AE97818ULL, "follower_banter6161201.cl"},
	{0x4AD1B654351802D0ULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_qnari_isabellasreturn.gad"},
	{0x4AD1F190E40C7F98ULL, "vxw_woundedcoast_oh_0_0.vlm"},
	{0x4AD24A67E6927850ULL, "6115298_m.fxe"},
	{0x4AD282B59A8C20E9ULL, "316414068.ogg"},
	{0x4AD2C85F03DC3A6DULL, "vxw_bonepitmining_a2_29t_0_0.vlm"},
	{0x4AD321D14326E013ULL, "vxlt_gallowstemplar_n_2bf_0_0.vlm"},
	{0x4AD324485231042DULL, "vxw_blightlands_u5_0_0.vlm"},
	{0x4AD474E64B242F63ULL, "vxw_sundermount_32j_0_1.vlm"},
	{0x4AD54CAF33F46294ULL, "hlac_lt_0_lowtown_196607_0z.dds"},
	{0x4AD576AB9ACD98A8ULL, "vxht_keep_672_0_0.vlm"},
	{0x4AD583CF2B7D3630ULL, "lt_gallowsprison_n_3567.anb"},
	{0x4AD6958DA5DB877BULL, "mag121_idunna6083966.cl"},
	{0x4AD6B862C3B2CDFAULL, "qun131ip_treasure2.utp"},
	{0x4AD6D3BE85B16630ULL, "6053189_m.fxe"},
	{0x4AD737E8630ACA06ULL, "framebuffereffects_ssao_p0.psh.fxo.dx11.assembly"},
	{0x4AD7401A2B3F7D7BULL, "6085471_m.fxe"},
	{0x4AD80EC9FA6129D5ULL, "508611044.ogg"},
	{0x4AD855D9DC7ED6F3ULL, "hlac_w_blightlands_-131071_0z.dds"},
	{0x4AD906178909B7A9ULL, "zz_isa_debug6175178.cl"},
	{0x4AD925C23BFED35DULL, "6172335_m.fxe"},
	{0x4AD9AA0B0A72C9EAULL, "707487946.ogg"},
	{0x4AD9F2E548F0AB05ULL, "qun221_petrice6077944.cl"},
	{0x4ADA431BFA77978DULL, "vxca_caverns1_hm7_0_0.vlm"},
	{0x4ADA7709582B2528ULL, "6173823_m.fxe"},
	{0x4ADB28162B294700ULL, "vfx_chunkyparticle_p3.psh.fxo.dx11"},
	{0x4ADBCFDFDE350B04ULL, "6084422_m.fxe"},
	{0x4ADBD9EAC60660B9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2basic.msh"},
	{0x4ADBE3EAC60671A7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2basic.mmh"},
	{0x4ADBE7EAC60678F4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2basic.mao"},
	{0x4ADC147A411346F3ULL, "vxlt_shop_2e_0_0.vlm"},
	{0x4ADC1F71B1BE9737ULL, "cd_one300dg_lia.wbd"},
	{0x4ADC299C597F2D91ULL, "vxht_mansion2_21a_0_0.vlm"},
	{0x4ADC866F042E975BULL, "vxw_sundermount_29g_0_0.vlm"},
	{0x4ADC8E43FA33411CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_enter.ani"},
	{0x4ADCA8B05765002DULL, "616254980.ogg"},
	{0x4ADCF3D28B0B595FULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_deeproads_plants_dp.dds"},
	{0x4ADD26E11132C1E7ULL, "vxht_0_hightown_a2_5h3_0_0.vlm"},
	{0x4ADD33547475123AULL, "dae000_tomwise6167289.cl"},
	{0x4ADDD0CB98F1D0BCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masb_0d.dds"},
	{0x4ADDE109AA2D7941ULL, "619089709.ogg"},
	{0x4ADE3BD95A4F5048ULL, "678075195.ogg"},
	{0x4ADE730078EBA5E3ULL, "zz_and_debug6134159.cl"},
	{0x4ADED1DD46034055ULL, "444485990.ogg"},
	{0x4ADF04DD66911203ULL, "mcr310_party_banter6182941.cl"},
	{0x4AE0246495A6B0F8ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_enco_0t.dds"},
	{0x4AE0636DD6029818ULL, "vxw_woundedcoast_di_0_0.vlm"},
	{0x4AE0F7B236FE5BF9ULL, "mrl150_merrill6071949.cl"},
	{0x4AE1489F6850A4C2ULL, "244216034.ogg"},
	{0x4AE14B17D0BBCE8DULL, "320447489.ogg"},
	{0x4AE1654E497AFA96ULL, "645664539.ogg"},
	{0x4AE214ECF85062D1ULL, "llac_lt_0_lowtown_a3_n_c_65538_1.dds"},
	{0x4AE24F5B5DD24FE4ULL, "hlac_lt_0_lowtown_a2_65534_1y.dds"},
	{0x4AE29C1AE856F5CDULL, "vxca_cavern1_haunted_hzo_0_0.vlm"},
	{0x4AE33DBC28D71364ULL, "vxca_cave_2cn_0_0.vlm"},
	{0x4AE361B0BCDEE4A3ULL, "dae000_bodhan6146885.cl"},
	{0x4AE36DCB17D619E5ULL, "art\\vfx\\combat\\vfx_w_tmp_silence_activ_c.anb"},
	{0x4AE3985B56CAA513ULL, "vxw_bonepitmining_a2_89_0_0.vlm"},
	{0x4AE39B43FA39765EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_enter.gad"},
	{0x4AE3E8888CC3B128ULL, "vxht_keep_603_0_0.vlm"},
	{0x4AE4326A19FEAEB9ULL, "31579996.ogg"},
	{0x4AE443EEF0FBE5D4ULL, "175356025.ogg"},
	{0x4AE4EDCE130C6334ULL, "gen00fl_dog6166813.cl"},
	{0x4AE5725A5B5EE0BDULL, "vxht_mansion2b_34o_0_0.vlm"},
	{0x4AE584DEB0374056ULL, "6041154_f.fxe"},
	{0x4AE58902EFE221A8ULL, "184355019.ogg"},
	{0x4AE5C59AB42EB0F1ULL, "6163290_m.fxe"},
	{0x4AE5DE35007082F0ULL, "hla_ht_0_hightown_-131074_0y.dds"},
	{0x4AE5E50A265EDCA2ULL, "599053535.ogg"},
	{0x4AE6401C41C825BBULL, "6108615_m.fxe"},
	{0x4AE74C364FC29291ULL, "llac_lt_gallowscourtyard_131069_1.dds"},
	{0x4AE7502DBA89639DULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.mmh"},
	{0x4AE7D914A9FED1ABULL, "177794589.ogg"},
	{0x4AE7E831DD94F158ULL, "mag360_samson6044875.cl"},
	{0x4AE7FEDBF0F40D94ULL, "185668996.ogg"},
	{0x4AE80162187E6001ULL, "6093993_m.fxe"},
	{0x4AE8C046F564B66EULL, "vxdr_primevaltomb_6pv_0_2.vlm"},
	{0x4AE945AF16DF9D45ULL, "hlac_lt_0_lowtown_docks_a3_-131075_7y.dds"},
	{0x4AEA0288EECB8AEDULL, "docks_male_sailor_mob6176667.cl"},
	{0x4AEA48CA5C6A08F9ULL, "182569628.ogg"},
	{0x4AEA7643FA3F567DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_enter.evt"},
	{0x4AEAA31043B1BFC5ULL, "hlac_lt_undercityencounter_-1_0y.dds"},
	{0x4AEAC128E3CC4283ULL, "llac_lt_0_lowtown_n_-2_2.dds"},
	{0x4AEAD1756C49A4C6ULL, "hlac_lt_gallowscourtyard_n_-131071_2y.dds"},
	{0x4AEB640F77C6A206ULL, "49586321.ogg"},
	{0x4AEBC31322FA6549ULL, "6139262_m.fxe"},
	{0x4AEC03CE0EB24045ULL, "vxlt_hangedman_n_2p7_0_0.vlm"},
	{0x4AEC72404362B174ULL, "vxca_caverns1_iis_0_0.vlm"},
	{0x4AED61C4CDB9A818ULL, "vxw_sundermount_1ys_0_0.vlm"},
	{0x4AEDBE4BDBDEE7FDULL, "gen00fl_carver6108918.cl"},
	{0x4AEDEF892010C5E8ULL, "927598338.ogg"},
	{0x4AEE33590BAD0DE0ULL, "one350_variel6070997.cl"},
	{0x4AEE872ADB17DD45ULL, "vxlt_undercitydungeon_ys_0_0.vlm"},
	{0x4AEED54AC7F8B098ULL, "vxw_woundedcoast_af_0_0.vlm"},
	{0x4AEEE54079E5C565ULL, "vxlt_hangedman_266_0_0.vlm"},
	{0x4AEF1FBA6A3903B6ULL, "dre000_mother6037113.cl"},
	{0x4AEF2ADB10FD18CBULL, "ave200_donnic_bar.stg"},
	{0x4AF07195E38F3D6CULL, "vxlt_backalley_n_1ec_0_0.vlm"},
	{0x4AF0927C01A48032ULL, "6143718_m.fxe"},
	{0x4AF0C2EE32DDA3B5ULL, "prp_str_rock_n.dds"},
	{0x4AF0C662FD7F229CULL, "69151637.ogg"},
	{0x4AF11EF0462B079EULL, "vxlt_0_lowtown_docks_n_c6l_0_0.vlm"},
	{0x4AF210207C1EA226ULL, "6066315_f.fxe"},
	{0x4AF21B6AA5150A7AULL, "qun211ip_area_exit.utp"},
	{0x4AF2B0CB1D085356ULL, "hlac_w_bonepitdestroyed_a3_327685_0y.dds"},
	{0x4AF2B1BF7F6D941CULL, "zz_otl_debug6060592.cl"},
	{0x4AF2BDE516DC7024ULL, "vxlt_0_lowtown_a2_n_c_ehm_0_0.vlm"},
	{0x4AF3004C2EFAD1F3ULL, "6147694_m.fxe"},
	{0x4AF301CBDAE6A618ULL, "vxdr_primevaltomb_6ix_0_1.vlm"},
	{0x4AF3B036AF97A980ULL, "6146878_m.fxe"},
	{0x4AF3B134A29B1D1DULL, "w_encounter2_48_0.mao"},
	{0x4AF3B534A29B2456ULL, "w_encounter2_48_0.mmh"},
	{0x4AF3BF34A29B3554ULL, "w_encounter2_48_0.msh"},
	{0x4AF4DB8F9DEB4A19ULL, "hlac_ht_keep_n_3732_4z.dds"},
	{0x4AF56B3F179C6968ULL, "552050979.ogg"},
	{0x4AF656D3A50929ACULL, "6037103_m.fxe"},
	{0x4AF65995D5B7D529ULL, "vxlt_0_lowtown_docks_a3_n_bsb_0_0.vlm"},
	{0x4AF6BA3DF6398058ULL, "ser211ip_blood_trail_2.utp"},
	{0x4AF6CBAB8C42269DULL, "vxht_0_hightown_a2_c69_0_0.vlm"},
	{0x4AF7464D265F44A5ULL, "one120_athenril6088303.cl"},
	{0x4AF841CC3726C7C4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_firestatue_sa.dds"},
	{0x4AF8670D1F26ED6CULL, "w_encounter3_29_0.msh"},
	{0x4AF86D0D1F26F75EULL, "w_encounter3_29_0.mmh"},
	{0x4AF8790D1F270BF5ULL, "w_encounter3_29_0.mao"},
	{0x4AF8B61C69FFBC12ULL, "w_encounter3_163_0.phy"},
	{0x4AF93AB4BF4AE7FCULL, "vxlt_darktown_2r6_0_0.vlm"},
	{0x4AFA1DA067BE031EULL, "414643208.ogg"},
	{0x4AFAFD0453C24503ULL, "gen_im_trs_dal_hvy_hlm_r2.uti"},
	{0x4AFB03FB4BEE17CDULL, "498148089.ogg"},
	{0x4AFBDAC332D99A16ULL, "ser270_emeric6149364.cl"},
	{0x4AFC974CDE8D86BDULL, "6110829_m.fxe"},
	{0x4AFCB0EBEC67526CULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef06.phy"},
	{0x4AFCC7E2C4EF79F2ULL, "213727600.ogg"},
	{0x4AFCFFF0A345D825ULL, "vxht_chantry_31x_0_0.vlm"},
	{0x4AFD12C9F1F6FEB5ULL, "vxw_blightlands_f4_0_0.vlm"},
	{0x4AFD32EDD9BBEB50ULL, "vxca_caverns1_mines_gkg_0_1.vlm"},
	{0x4AFE4F5E0DA5AE29ULL, "207739232.ogg"},
	{0x4AFE9EF58B4211F7ULL, "lt_0_lowtown_docks_n_198_0.mmh"},
	{0x4AFEA2F58B421804ULL, "lt_0_lowtown_docks_n_198_0.mao"},
	{0x4AFEA37A71D591F8ULL, "422550763.ogg"},
	{0x4AFEB4F58B423769ULL, "lt_0_lowtown_docks_n_198_0.msh"},
	{0x4AFF18D25607A1A2ULL, "vxw_sundermount_2w4_0_0.vlm"},
	{0x4AFF286517A3CD15ULL, "6103713_m.fxe"},
	{0x4AFF45A9A05BEDB2ULL, "229298174.ogg"},
	{0x4AFF5676DD81952BULL, "vfx_cloud_dark2.dds"},
	{0x4AFFFD43CCD3ED1BULL, "791001720.ogg"},
	{0x4B003D54ECE3BFA2ULL, "6112081_f.fxe"},
	{0x4B00C343F32CCC85ULL, "6149449_m.fxe"},
	{0x4B00FC1AC0D22EACULL, "vxlt_hangedman_n_282_0_0.vlm"},
	{0x4B013EDB5ED25D5EULL, "vxlt_0_lowtown_docks_a3_czl_0_0.vlm"},
	{0x4B014A289092C6F7ULL, "hlac_ht_0_hightown_n_65533_3x.dds"},
	{0x4B01CD95098ACFD4ULL, "vxht_mansion_n_4ww_0_0.vlm"},
	{0x4B0223900E9793A9ULL, "zz_unequip.ncs"},
	{0x4B0225731F02C3A1ULL, "white.dds"},
	{0x4B02FC13C12CA09DULL, "6046906_m.fxe"},
	{0x4B0323D625B456BBULL, "6106415_m.fxe"},
	{0x4B03D37B7E5CB4B8ULL, "w_woundedcoast_163_0.phy"},
	{0x4B047C52E823848BULL, "vxlt_gallowstemplar_f_2bf_0_0.vlm"},
	{0x4B04C91155571D54ULL, "410871991.ogg"},
	{0x4B05F5099C7A9730ULL, "blue_ground.dds"},
	{0x4B0666699D8AEC5BULL, "vxw_encounter2_50_0_0.vlm"},
	{0x4B067676EC397253ULL, "dr_slavers_3218_lc.mmh"},
	{0x4B06D405CDC45893ULL, "mcr310_argument6128103.cl"},
	{0x4B072796CF4DF0B0ULL, "llac_w_encounter3_-65540_0.dds"},
	{0x4B075685871ACD94ULL, "710776325.ogg"},
	{0x4B07A999AEBAF3B6ULL, "935307960.ogg"},
	{0x4B07AC91B141007CULL, "vxlt_darktown_2u7_0_0.vlm"},
	{0x4B07CCEAC62B85DCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2basic.phy"},
	{0x4B07D099AB5FBEF8ULL, "vxlt_0_lowtown_a2_bxo_0_0.vlm"},
	{0x4B07E3CCF1064E72ULL, "mag291_ambient_26218716.cl"},
	{0x4B081F0B0B0C67F4ULL, "6120948_m.fxe"},
	{0x4B08AE07FB8E44F4ULL, "328993349.ogg"},
	{0x4B08BBC20AE181C6ULL, "6016067_m.fxe"},
	{0x4B09039FA9B79EDBULL, "379759483.ogg"},
	{0x4B09BEBBB819A4A0ULL, "6136675_m.fxe"},
	{0x4B09E70CFC238B33ULL, "art\\characters\\playercharacter\\qunarimale\\har_nud.mao"},
	{0x4B0A602312487CD8ULL, "6026202_m.fxe"},
	{0x4B0B1EC07A9A9E14ULL, "6017865_f.fxe"},
	{0x4B0B4FCDE9DE55FAULL, "211765684.ogg"},
	{0x4B0C43146D26648EULL, "vxht_0_hightown_a3_b8h_0_0.vlm"},
	{0x4B0C6ECB17F92026ULL, "art\\vfx\\combat\\vfx_w_tmp_silence_activ_c.mmh"},
	{0x4B0D02E189A3F1D9ULL, "t3_wdg_gr2.tnt"},
	{0x4B0D554FC77B6E6CULL, "vxlt_backalley_n_1ka_0_0.vlm"},
	{0x4B0D8AAF47FDBA22ULL, "vxw_sundermount_2x5_0_0.vlm"},
	{0x4B0D9A420999E595ULL, "6104613_m.fxe"},
	{0x4B0DD2BA41E9917CULL, "w_encounter3_248_0.mao"},
	{0x4B0DDEBA41E9A51FULL, "w_encounter3_248_0.mmh"},
	{0x4B0DE4BA41E9AFC1ULL, "w_encounter3_248_0.msh"},
	{0x4B0E30CB0A0A8C55ULL, "6049716_m.fxe"},
	{0x4B0ED6308086E718ULL, "vxw_bonepit_1wn_0_0.vlm"},
	{0x4B0EFF627EF27591ULL, "304712457.ogg"},
	{0x4B0F4E45C53DD338ULL, "63855405.ogg"},
	{0x4B0F5FA4BD1674AEULL, "251472753.ogg"},
	{0x4B0F61DD80697E92ULL, "61292349.ogg"},
	{0x4B0F7E76DC1C7DF0ULL, "llac_lt_0_lowtown_a2_n_c_0_8.dds"},
	{0x4B0FD2364FF518A1ULL, "m_mag111ar_warehouse.bnk"},
	{0x4B0FF6A919AA6288ULL, "297483160.ogg"},
	{0x4B104150B19C8CB0ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs105a.mao"},
	{0x4B104550B19C93C3ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs105a.mmh"},
	{0x4B105350B19CAB1DULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs105a.msh"},
	{0x4B107CB8D3C99ABEULL, "mag210_arianni6021213.cl"},
	{0x4B10AE76BED381CAULL, "vxca_caverns2_852_0_1.vlm"},
	{0x4B10D6F649703B07ULL, "follower_banter6160636.cl"},
	{0x4B1126C9A7551191ULL, "6108261_f.fxe"},
	{0x4B118FDA25D9084FULL, "hlac_lt_0_lowtown_docks_a3_n_-4_0x.dds"},
	{0x4B11A4E832318AB1ULL, "6180446_m.fxe"},
	{0x4B11BB7A4D061793ULL, "620937266.ogg"},
	{0x4B11C74DD33E5F03ULL, "vxht_0_hightown_a3_cdt_0_0.vlm"},
	{0x4B11F7D991C4C5FDULL, "distortionwater.psh.fxo"},
	{0x4B1270511F014F2BULL, "fxa_deacti_cir.mao"},
	{0x4B128A7B3359A423ULL, "mcr310ip_to_gallows.utp"},
	{0x4B128F8D8E301093ULL, "vxlt_hangedman_4x_0_0.vlm"},
	{0x4B12B69BAC94C6E1ULL, "815065217.ogg"},
	{0x4B1334E7992ABF1FULL, "876178798.ogg"},
	{0x4B134C6BE7F487DDULL, "277003699.ogg"},
	{0x4B13AF65702E579DULL, "vxht_0_hightown_a2_c87_0_0.vlm"},
	{0x4B1433A63BC77B19ULL, "ss_crazy_f_1.wbd"},
	{0x4B14C3D3323DB3B8ULL, "315383975.ogg"},
	{0x4B152D44B21130D8ULL, "vxw_woundedcoast_13m_0_0.vlm"},
	{0x4B161E6EA33718FCULL, "vxlt_darktown_2x4_0_0.vlm"},
	{0x4B164A4018CB6439ULL, "vxlt_0_lowtown_a2_n_bxm_0_0.vlm"},
	{0x4B16CBCC157805DAULL, "6125215_m.fxe"},
	{0x4B170BCAEED8AD29ULL, "plt_cod_ltr_ser_alriks_pape.plo"},
	{0x4B178B4F4B88F2E7ULL, "zz_lock.ncs"},
	{0x4B17D08BE216CF2EULL, "213267427.ogg"},
	{0x4B18CFF284233002ULL, "6165273_m.fxe"},
	{0x4B18DFA5BC2DED45ULL, "472496913.ogg"},
	{0x4B18E8D338090C1EULL, "gen00fl_aveline6128987.cl"},
	{0x4B1918989168E6EEULL, "6207579_m.fxe"},
	{0x4B1950608D7873F0ULL, "6140224_f.fxe"},
	{0x4B1998EC25C0DA07ULL, "6170472_m.fxe"},
	{0x4B19A071B39B446CULL, "llac_w_encounter2_camp_2_0.dds"},
	{0x4B19CFCDA6A1821BULL, "vxht_mansion2_3da_0_0.vlm"},
	{0x4B19EB61BDB2131CULL, "hlac_lt_0_lowtown_a2_131072_3z.dds"},
	{0x4B19F683D5E32FB5ULL, "vxw_blightlands_l2_0_0.vlm"},
	{0x4B1A13934ACBD3CDULL, "hlac_lt_backalley_n_65537_2z.dds"},
	{0x4B1A2AE19DDAFA18ULL, "vxw_woundedcoast_pe_0_0.vlm"},
	{0x4B1A3F5414A3EE91ULL, "663928928.ogg"},
	{0x4B1A51504242F694ULL, "art\\cinematics\\_crit path\\followerintros\\cs_drunk1_isa_intro.ani"},
	{0x4B1AB4F15F1C7D0EULL, "vxht_0_hightown_a3_b7i_0_0.vlm"},
	{0x4B1B423572C34DF9ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_2y.dds"},
	{0x4B1B731F000FE26CULL, "hlac_lt_darktown_-1_43y.dds"},
	{0x4B1B8696827DBA49ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_floor_tile_sa.dds"},
	{0x4B1B8C650B389545ULL, "538658716.ogg"},
	{0x4B1BA86BC51BC158ULL, "dae000ip_crowd_mf5_talk12.utp"},
	{0x4B1BC7E23C672A55ULL, "zz_followers_banter6159672.cl"},
	{0x4B1BE10FB88BF441ULL, "vxdr_primevaltomb_911_0_2.vlm"},
	{0x4B1D326BA848B877ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masg.phy"},
	{0x4B1DBC1F22EB329BULL, "255267846.ogg"},
	{0x4B1E0F5ACFDDBF0DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_tunnel_1.mao"},
	{0x4B1E86CA06DAEE9CULL, "6072777_m.fxe"},
	{0x4B1EB39047C5A036ULL, "qun120_arishok_guard6054008.cl"},
	{0x4B1EF6CECD17A879ULL, "vxlt_blackemporium_13z_0_0.vlm"},
	{0x4B1F514042D0730BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_enter.ani"},
	{0x4B214650B1AB08E8ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs105a.phy"},
	{0x4B214E50424910F6ULL, "art\\cinematics\\_crit path\\followerintros\\cs_drunk1_isa_intro.gad"},
	{0x4B21590981365D36ULL, "vxlt_0_lowtown_docks_n_c51_0_0.vlm"},
	{0x4B21DBAE7DC44D22ULL, "gen00fl_merrill6130103.cl"},
	{0x4B21DBBDEC39E550ULL, "vxht_mansion_n_62j_0_0.vlm"},
	{0x4B21EA50A23D9B34ULL, "gen00fl_isabela6079664.cl"},
	{0x4B228B73B024F532ULL, "zz_prmpt_debug.cnv"},
	{0x4B22955DC7FAC370ULL, "hlac_dr_primevaltomb_664_2x.dds"},
	{0x4B22A5F699B59123ULL, "lowtown_argument6169357.cl"},
	{0x4B22ECAE8AF53EE6ULL, "136083144.ogg"},
	{0x4B239D7FABBB51DBULL, "vxw_blightlands_g9_0_0.vlm"},
	{0x4B23D7902B3067A8ULL, "389904742.ogg"},
	{0x4B23EE4E0D99801BULL, "vxlt_gallowscourtyard_6y2_0_0.vlm"},
	{0x4B23F83B2E421B49ULL, "300312835.ogg"},
	{0x4B246F751A8B5683ULL, "p_bl.mao"},
	{0x4B24BA8827793CCCULL, "6115485_m.fxe"},
	{0x4B25127559E999ABULL, "12089991.ogg"},
	{0x4B253151300723B1ULL, "vxlt_warehouse_am_0_0.vlm"},
	{0x4B263AD1AFB431C3ULL, "gen00fl_aveline6139670.cl"},
	{0x4B26468235232FE7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_aoe_kback.gad"},
	{0x4B26ABE7D81E2377ULL, "292909196.ogg"},
	{0x4B275B2A8C2D0F38ULL, "cnvprev_human_male_mage.utc"},
	{0x4B27609840A304BEULL, "6141697_m.fxe"},
	{0x4B27D78D134C37E7ULL, "486499627.ogg"},
	{0x4B2862C325154C37ULL, "art\\levels\\ztools\\default\\generic\\0fi_basic.phy"},
	{0x4B286860C7D94835ULL, "vxw_blightlands_q1_0_0.vlm"},
	{0x4B28E11C0525688FULL, "zz_mer_debug6062734.cl"},
	{0x4B292FFFB5031283ULL, "259664111.ogg"},
	{0x4B29C17D00DA0C56ULL, "535561261.ogg"},
	{0x4B2A0F2331B62A8DULL, "vxlt_0_lowtown_docks_app_0_0.vlm"},
	{0x4B2A3EE3BAE6DBC4ULL, "vxw_woundedcoast_16n_0_0.vlm"},
	{0x4B2A704D62C47F5DULL, "261493497.ogg"},
	{0x4B2ABE52B32F63EFULL, "gen00fl_aveline6140810.cl"},
	{0x4B2AFD39B5890275ULL, "125270550.ogg"},
	{0x4B2C077DA28980DCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appa_0t.dds"},
	{0x4B2C2EF44F0ADA02ULL, "gen00fl_anders6111386.cl"},
	{0x4B2C3B82352788F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_aoe_kback.evt"},
	{0x4B2C606DEAC1C197ULL, "6217926_m.fxe"},
	{0x4B2CE98E651A370DULL, "320928311.ogg"},
	{0x4B2D394042DC8782ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_enter.evt"},
	{0x4B2D957C115E85F8ULL, "574925572.ogg"},
	{0x4B2D97ED82587A74ULL, "vxlt_gallowscourtyard_62h_0_0.vlm"},
	{0x4B2DE7AC428838DAULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash_p.anb"},
	{0x4B2E2F516FB454E0ULL, "6111468_f.fxe"},
	{0x4B2E6AF1795832F7ULL, "6114283_m.fxe"},
	{0x4B2E6B39B5C10A04ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb1.msh"},
	{0x4B2E8139B5C12FA6ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb1.mmh"},
	{0x4B2ECD5807C8D5EFULL, "vxdr_primevaltomb_6sl_0_0.vlm"},
	{0x4B2F0CD0C9EE55FFULL, "918136890.ogg"},
	{0x4B2FCAE6732C75B6ULL, "vxlt_0_lowtown_docks_n_c02_0_0.vlm"},
	{0x4B301523ECC25462ULL, "vxht_mansion2b_2jg_0_0.vlm"},
	{0x4B30166EE00017EEULL, "hlac_lt_0_lowtown_n_-3_3y.dds"},
	{0x4B3021099D903729ULL, "vxlt_0_lowtown_docks_a3_n_bgf_0_0.vlm"},
	{0x4B30993F60219C74ULL, "hole_s.dds"},
	{0x4B309FCB4EF1F008ULL, "144064748.ogg"},
	{0x4B30A083E8BEF356ULL, "mag311_nyssa6041961.cl"},
	{0x4B315BB7B2003891ULL, "character_character_p271.psh.fxo.dx11.assembly"},
	{0x4B316A1623DE6624ULL, "dre000_dougal6129800.cl"},
	{0x4B317CD266FA70D4ULL, "hairalpha_rendertoshadowmap_p0.psh.fxo"},
	{0x4B318113BC25A018ULL, "gen00fl_anders6128393.cl"},
	{0x4B31DBFC5B3E3264ULL, "133156564.ogg"},
	{0x4B31E9CBEA16FCF8ULL, "hlac_lt_0_lowtown_-65537_1y.dds"},
	{0x4B320F5C9DB16A5BULL, "vxw_blightlands_h8_0_0.vlm"},
	{0x4B3210FE95FD61D8ULL, "vxw_woundedcoast_19k_0_0.vlm"},
	{0x4B3216E67E3B9F80ULL, "dae170ip_at_lowtown.utp"},
	{0x4B32883733C9D5C1ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_follower_andersintro.ani"},
	{0x4B32F5E84A376193ULL, "vxlt_0_lowtown_n_ewe_0_0.vlm"},
	{0x4B33700845525E00ULL, "gen_im_trs_tmp_med_cht_r0.uti"},
	{0x4B338073B03355B5ULL, "zz_prmpt_debug.ncs"},
	{0x4B338F6746E504D2ULL, "6072674_m.fxe"},
	{0x4B33986E6FA57032ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_dual_movement.anb"},
	{0x4B33E00E4DAC4019ULL, "736092675.ogg"},
	{0x4B34384042E2A495ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_enter.gad"},
	{0x4B3455389AB496B3ULL, "vxht_mansion2b_3er_0_0.vlm"},
	{0x4B34B24E84721DCEULL, "6035193_f.fxe"},
	{0x4B34C07135418072ULL, "fen210_orana6109772.cl"},
	{0x4B350233501DB8F5ULL, "follower_barks6153385.cl"},
	{0x4B3541DD37F31D31ULL, "6195814_f.fxe"},
	{0x4B3588D6292CC86AULL, "vxht_mansion2_1ae_0_0.vlm"},
	{0x4B35D4630F37D78EULL, "vxca_caverns1_mines_dnh_0_0.vlm"},
	{0x4B36282B27C77085ULL, "vxlt_0_lowtown_a2_bxv_0_0.vlm"},
	{0x4B36827B0E90CF52ULL, "414353853.ogg"},
	{0x4B370518D3F6BD08ULL, "hla_ht_0_hightown_a3_-65537_0z.dds"},
	{0x4B370BC746E0EE1FULL, "656055770.ogg"},
	{0x4B37416870F8BABCULL, "art\\characters\\weapons\\unique\\w_vael_bow.mao"},
	{0x4B374D6870F8CF5FULL, "art\\characters\\weapons\\unique\\w_vael_bow.mmh"},
	{0x4B37536870F8D901ULL, "art\\characters\\weapons\\unique\\w_vael_bow.msh"},
	{0x4B37AF51DE74CB7EULL, "6065821_m.fxe"},
	{0x4B37BF8EBFA26B5FULL, "vxlt_gallowscourtyard_n_5z3_0_0.vlm"},
	{0x4B380CF3CBEF90BFULL, "vxlt_backalley_n_1kn_0_0.vlm"},
	{0x4B380EB952BF5FF4ULL, "hlac_lt_0_lowtown_docks_n_-131075_2y.dds"},
	{0x4B38B0C0ACDCF444ULL, "vxw_woundedcoast_11o_0_0.vlm"},
	{0x4B39933733D00763ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_follower_andersintro.gad"},
	{0x4B39AC85095D3045ULL, "1063587744.ogg"},
	{0x4B39FC2936AF2E54ULL, "lt_0_lowtown_a2_n_45_0.rml"},
	{0x4B3A0779ADFD1A32ULL, "6092197_m.fxe"},
	{0x4B3A33823533B82DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_aoe_kback.ani"},
	{0x4B3A3E12D615708BULL, "ht_0_hightown_a3_30_0_lc.mmh"},
	{0x4B3A4455A01E31B2ULL, "vxca_caverns2prc_a2v_0_0.vlm"},
	{0x4B3AAB161FC88F2BULL, "460553082.ogg"},
	{0x4B3AAC3C7B47430CULL, "mrl000_marethari6129441.cl"},
	{0x4B3AE6420783D179ULL, "6056889_f.fxe"},
	{0x4B3B799AC2D63E9DULL, "hlac_ht_keep_n_3733_1z.dds"},
	{0x4B3B8BE55D1A02B8ULL, "vxw_encounter2_7a_0_0.vlm"},
	{0x4B3C332998780A5EULL, "ach_comp_armor_all.dds"},
	{0x4B3D3F34F9BEEE6FULL, "vxdr_primevaltomb_6vo_0_0.vlm"},
	{0x4B3D6D0E1617A1EBULL, "vxht_mansion2_3c8_0_0.vlm"},
	{0x4B3D87EBEC9EA337ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef06.mmh"},
	{0x4B3D92A366FE220FULL, "vxlt_house_dk_0_0.vlm"},
	{0x4B3D9DEBEC9EC8A9ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef06.msh"},
	{0x4B3DA574D5BBBD40ULL, "qcr220_grey_wardens6135956.cl"},
	{0x4B3DAE0731D4E45FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv2.gad"},
	{0x4B3DFEB5D5D337DFULL, "hlac_ca_caverns2_5802_55z.dds"},
	{0x4B3E3456A046D053ULL, "isa200pt_thief.ncs"},
	{0x4B3E431AF6CF43B0ULL, "12579336.ogg"},
	{0x4B3EF2BCDA8EE33AULL, "t3_dey_gy2.tnt"},
	{0x4B3EF8C9CC2FDE3BULL, "6084757_m.fxe"},
	{0x4B3F1E78D6FF5E90ULL, "dae000_flemeth6013406.cl"},
	{0x4B3F4D0D1F62C315ULL, "w_encounter3_29_0.rml"},
	{0x4B3F582915CA3A67ULL, "463762469.ogg"},
	{0x4B3F6B5988D45BB3ULL, "319084877.ogg"},
	{0x4B3F85CD41AC9666ULL, "vxlt_0_lowtown_docks_n_ctn_0_0.vlm"},
	{0x4B40D058F917B70AULL, "348803109.ogg"},
	{0x4B4167C53C2D7A13ULL, "vxlt_0_lowtown_n_epd_0_0.vlm"},
	{0x4B4187ADF442E744ULL, "28209988.ogg"},
	{0x4B4205C30CD932A7ULL, "hlac_dr_primevaltomb_661_2x.dds"},
	{0x4B42150B13F354B1ULL, "vxlt_warehouse_oc_0_0.vlm"},
	{0x4B42296A96F98B4AULL, "6053060_m.fxe"},
	{0x4B4230767F2136F0ULL, "zz_seb_debug6212574.cl"},
	{0x4B439D097E0A0DAFULL, "537598932.ogg"},
	{0x4B43FAB31B22E0EAULL, "vxht_mansion2_1df_0_0.vlm"},
	{0x4B441A93228C48B8ULL, "hlac_lt_gallowscourtyard_65534_3x.dds"},
	{0x4B444B493D89A8DFULL, "ave331_jeven6138757.cl"},
	{0x4B44648207013605ULL, "vxlt_warehouse_n_131_0_0.vlm"},
	{0x4B44A30731DAF04CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv2.evt"},
	{0x4B44CD5EDAA11B3EULL, "ht_0_hightownexplosion_n_ll.mmh"},
	{0x4B455FD56FDB0B9FULL, "320289524.ogg"},
	{0x4B45B988824C31FBULL, "vfx_vfx_p6.vsh.fxo.dx11.assembly"},
	{0x4B463DEE9EE0BE3FULL, "t3_wdg_w11.tnt"},
	{0x4B46EA13A07D78CDULL, "893175235.ogg"},
	{0x4B473B01492A550AULL, "934524340.ogg"},
	{0x4B476B46037EA974ULL, "lgt000ip_bones_disturb1.ncs"},
	{0x4B47B012ECA19979ULL, "hlac_dr_slavers_3217_0x.dds"},
	{0x4B481BA31960ABC3ULL, "3322816.ogg"},
	{0x4B484B3C3E0D87E7ULL, "6166769_m.fxe"},
	{0x4B488864EA3302D1ULL, "404357916.ogg"},
	{0x4B489DA4485F4918ULL, "vxw_bonepit_1kj_0_0.vlm"},
	{0x4B48CD752E694B90ULL, "gen00fl_bethany6102926.cl"},
	{0x4B48FD574A97D963ULL, "hlac_lt_0_lowtown_docks_a3_-131075_6x.dds"},
	{0x4B492E8790C5B2E9ULL, "6117971_f.fxe"},
	{0x4B49F319563A7899ULL, "mi_gla_shd.mao"},
	{0x4B4A2AC39BC9ED86ULL, "vxw_encounter2_20_0_0.vlm"},
	{0x4B4A5AB87769FBE3ULL, "hlac_lt_undercityencounter_-1_1x.dds"},
	{0x4B4A90CF4EC5EAC5ULL, "deathhints_random.gda"},
	{0x4B4ADD9EED044705ULL, "vxht_0_hightown_a2_9b7_0_0.vlm"},
	{0x4B4B48B6C44C5472ULL, "vxlt_gallowstemplar_n_34p_0_0.vlm"},
	{0x4B4B76E527ED2C6EULL, "6133599_m.fxe"},
	{0x4B4B7FDF385F61D9ULL, "art\\characters\\heraldry\\hld_tevinter_slavers.dds"},
	{0x4B4B9B794AA81CE1ULL, "274265247.ogg"},
	{0x4B4BAC7045375FE0ULL, "dae000cr_grand_cleric6124673.cl"},
	{0x4B4BB5E97D3EB0BAULL, "6062801_f.fxe"},
	{0x4B4BC7313CB5FB14ULL, "w_sundermount_29_0_lc.mmh"},
	{0x4B4CBA0E5A6D1048ULL, "hair_pass1_depth.vsh.fxo"},
	{0x4B4CCB2E3B2A7E8EULL, "208562212.ogg"},
	{0x4B4D849C15D45BDEULL, "6128083_m.fxe"},
	{0x4B4D85A5DAF41BFEULL, "vxw_blightlands_d5_0_0.vlm"},
	{0x4B4DA6B89189DF6BULL, "vxlt_gallowscourtyard_n_7h7_0_0.vlm"},
	{0x4B4DF41C6F9925C4ULL, "6171591_m.fxe"},
	{0x4B4E1C9068AE5992ULL, "vxca_caverns2_74o_0_0.vlm"},
	{0x4B4E8AA83068E95BULL, "350964370.ogg"},
	{0x4B4EE36E1C553231ULL, "vxlt_0_lowtown_a2_n_en0_0_0.vlm"},
	{0x4B4EF032750E8ED9ULL, "6019219_f.fxe"},
	{0x4B4FBE4F8C9A4556ULL, "6060417_m.fxe"},
	{0x4B4FD9A22E239293ULL, "vxlt_0_lowtown_n_emg_0_0.vlm"},
	{0x4B4FE31BDB80F15DULL, "6103667_m.fxe"},
	{0x4B50B1D962F89A31ULL, "lgt311_shw_mer_mettin6156524.cl"},
	{0x4B5138FF89CD5D4DULL, "195255118.ogg"},
	{0x4B517E87F2DA3093ULL, "one130_thug6131603.cl"},
	{0x4B518B0731E551A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv2.ani"},
	{0x4B527AF116621617ULL, "hlac_lt_0_lowtown_a2_-65538_0z.dds"},
	{0x4B528027FE8C7C26ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.impa.evt"},
	{0x4B52B13DF687CFF9ULL, "ser211ip_blood_trail_2.ncs"},
	{0x4B52C4ED012E96ABULL, "ef_gen_1.mor"},
	{0x4B533F800BCC649DULL, "6208403_m.fxe"},
	{0x4B537B04CE0BAFBAULL, "198893282.ogg"},
	{0x4B5425C24EAE60DCULL, "fx_lens_flare.mao"},
	{0x4B5483F81266D299ULL, "6107436_m.fxe"},
	{0x4B554B397FE99F1FULL, "6067861_f.fxe"},
	{0x4B555391BF430DC2ULL, "vxdr_primevaltomb_758_0_2.vlm"},
	{0x4B55CBC61461F4E7ULL, "163838552.ogg"},
	{0x4B55EAAC42A98F8DULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash_p.mmh"},
	{0x4B56132D361992E6ULL, "6133822_f.fxe"},
	{0x4B5617B2C5EBCC01ULL, "darktown_male_fereldan_sta6175155.cl"},
	{0x4B56E69423B47076ULL, "hlac_w_sundermount_327680_0y.dds"},
	{0x4B570F813A556198ULL, "vxw_bonepit_1lk_0_0.vlm"},
	{0x4B5795CF700492C2ULL, "615268547.ogg"},
	{0x4B57DC6D82EE04DCULL, "6069412_m.fxe"},
	{0x4B57E0365032E653ULL, "m_mag111ar_warehouse.wbd"},
	{0x4B589727FE910F71ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.impa.gad"},
	{0x4B58C2FB6B4A6838ULL, "vxw_blightlands_ah_0_0.vlm"},
	{0x4B5993D9CD8E071FULL, "art\\characters\\creatures\\mubari\\animation\\com_roll.gad"},
	{0x4B59BCD58E8CF1B6ULL, "308722004.ogg"},
	{0x4B5A443F084C6807ULL, "vxca_caverns1_mines_i28_0_0.vlm"},
	{0x4B5A9F2CDF24588FULL, "hlac_lt_0_lowtown_a3_n_c_131070_1y.dds"},
	{0x4B5AEA3A60975F91ULL, "vxw_bonepit_1wz_0_1.vlm"},
	{0x4B5B856C9E44AFD2ULL, "dae000_bartrand6204166.cl"},
	{0x4B5B85F66D4D7EFBULL, "art\\vfx\\combat\\vfx_gib_hm_arm_ltrb.ani"},
	{0x4B5BC9741DA1EC9EULL, "hlac_lt_0_lowtown_-65537_6z.dds"},
	{0x4B5BF782CCEA347EULL, "vxw_blightlands_i6_0_0.vlm"},
	{0x4B5C0365718A4547ULL, "929276608.ogg"},
	{0x4B5C45E319E809A0ULL, "gen00fl_isabela6087746.cl"},
	{0x4B5C89864720289BULL, "171921783.ogg"},
	{0x4B5D52FBC430E5A8ULL, "90411641.ogg"},
	{0x4B5D554B0E4B4AB1ULL, "vxlt_0_lowtown_a2_n_es7_0_0.vlm"},
	{0x4B5DC2A63FFBE57DULL, "and121_karl6093993.cl"},
	{0x4B5E2A60DB260A11ULL, "vxht_mansion2_2bx_0_0.vlm"},
	{0x4B5E384F4137245AULL, "hlac_lt_backalley_n_0_10x.dds"},
	{0x4B5E4B7F2019AB13ULL, "vxlt_0_lowtown_n_eff_0_0.vlm"},
	{0x4B5E913FEE830953ULL, "vxht_0_hightown_bc1_0_0.vlm"},
	{0x4B5EAE63097978FEULL, "vxlt_0_lowtown_a3_n_c_byf_0_0.vlm"},
	{0x4B5EF8C4F7DF85B1ULL, "vxlt_warehouse_ua_0_0.vlm"},
	{0x4B5F430203F9948FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masd_0n.dds"},
	{0x4B5F4A46039302A5ULL, "lgt000ip_bones_disturb1.utp"},
	{0x4B5F5244D7833DB6ULL, "gen00fl_aveline6132896.cl"},
	{0x4B5F7D56A19D89ABULL, "hlac_ht_0_hightown_a2_65535_2z.dds"},
	{0x4B5F9827FE972F1FULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.impa.ani"},
	{0x4B5FC07B7EAAF34DULL, "w_woundedcoast_163_0.msh"},
	{0x4B5FCE7B7EAB0A80ULL, "w_woundedcoast_163_0.mao"},
	{0x4B5FD27B7EAB11D3ULL, "w_woundedcoast_163_0.mmh"},
	{0x4B6057D9358D9229ULL, "vxht_bartrand_n_5pe_0_0.vlm"},
	{0x4B6088D9CD94130CULL, "art\\characters\\creatures\\mubari\\animation\\com_roll.evt"},
	{0x4B6165BB96D24CF5ULL, "vxht_0_hightown_n_bwx_0_0.vlm"},
	{0x4B61AE82B66F4671ULL, "270893378.ogg"},
	{0x4B61B06F420DFA31ULL, "273093412.ogg"},
	{0x4B6249B07A19230DULL, "6166925_m.fxe"},
	{0x4B628003A8140AC7ULL, "925571143.ogg"},
	{0x4B62ED84B4B45A50ULL, "6149075_m.fxe"},
	{0x4B63F629F9850A93ULL, "gen00fl_varric6201268.cl"},
	{0x4B6415CEE082B80AULL, "vxht_bartrand_n_65z_0_0.vlm"},
	{0x4B64FC70E3D21971ULL, "vxlt_0_lowtown_a2_n_gc9_0_0.vlm"},
	{0x4B65136BA886396CULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masg.mmh"},
	{0x4B65196BA886435EULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masg.msh"},
	{0x4B657FDFE56C3801ULL, "zz_dre_debug6056990.cl"},
	{0x4B65CBCAEA897F61ULL, "dae300_alistair_king6134925.cl"},
	{0x4B65DAAEC5B9576EULL, "vxlt_0_lowtown_a2_n_c_elj_0_1.vlm"},
	{0x4B661D2936D4A154ULL, "lt_0_lowtown_a2_n_45_0.anb"},
	{0x4B662662070A4937ULL, "257902989.ogg"},
	{0x4B66F1A5358D57EBULL, "plt_cod_itm_glandivalis.plo"},
	{0x4B67333745558127ULL, "ht_0_hightown_a3_17_0_ll.mmh"},
	{0x4B6734D85D4080B8ULL, "vxw_blightlands_ji_0_0.vlm"},
	{0x4B674456858330D1ULL, "648087312.ogg"},
	{0x4B679C27AC935E5FULL, "vxht_0_hightown_a3_8pg_0_0.vlm"},
	{0x4B6811F5AFD7A7D2ULL, "vxht_0_hightown_bc4_0_0.vlm"},
	{0x4B685A9F0BD95D6EULL, "6139399_m.fxe"},
	{0x4B68715D9F821672ULL, "927617149.ogg"},
	{0x4B687DA2B62CA5ABULL, "vxdr_primevaltomb_7ki_0_0.vlm"},
	{0x4B68D6EE277ED690ULL, "vxca_caverns1_mines_k6u_0_0.vlm"},
	{0x4B691CA6666A7373ULL, "ts_coterie_prog_junk1.gda"},
	{0x4B69EBB899CFC538ULL, "vxlt_0_lowtown_a3_n_c_bz9_0_0.vlm"},
	{0x4B6A092F4F8FE122ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_basic.msh"},
	{0x4B6A0B2F4F8FE4D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_basic.mmh"},
	{0x4B6A172F4F8FF97BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_basic.mao"},
	{0x4B6A695FBEE04CFEULL, "vxw_blightlands_r7_0_0.vlm"},
	{0x4B6B048D1DECD2AEULL, "6114097_m.fxe"},
	{0x4B6B312C0532DC0DULL, "ave300cr_donnic.utc"},
	{0x4B6B41A04F2EA396ULL, "one300_zevran6089124.cl"},
	{0x4B6B48354904E8F3ULL, "vxht_keep_5k6_0_0.vlm"},
	{0x4B6CAD739535F4F9ULL, "vxlt_house_g4_0_0.vlm"},
	{0x4B6D031CE07921D3ULL, "vxht_0_hightown_bl0_0_0.vlm"},
	{0x4B6D4094135CBEB6ULL, "706943012.ogg"},
	{0x4B6D8518DA0C428FULL, "6208600_f.fxe"},
	{0x4B6D9A729D6DA657ULL, "dae000_orana6149081.cl"},
	{0x4B6E70D9CDA02865ULL, "art\\characters\\creatures\\mubari\\animation\\com_roll.ani"},
	{0x4B6F1491D2E7C945ULL, "dae000_gamlen6121572.cl"},
	{0x4B6F37C32551058FULL, "art\\levels\\ztools\\default\\generic\\0fi_basic.mao"},
	{0x4B6F43C325511A2CULL, "art\\levels\\ztools\\default\\generic\\0fi_basic.mmh"},
	{0x4B6F49C32551241EULL, "art\\levels\\ztools\\default\\generic\\0fi_basic.msh"},
	{0x4B6F4C2E84F497D0ULL, "66172220.ogg"},
	{0x4B6F5A55F4CF1CAFULL, "6140303_m.fxe"},
	{0x4B6F6091AEA12B37ULL, "mag121st_idunna.stg"},
	{0x4B6F833144DE1881ULL, "mag121_idunna6084217.cl"},
	{0x4B6F9C0D1F5466D6ULL, "6137469_m.fxe"},
	{0x4B6FACF66D5E6985ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_ltrb.gad"},
	{0x4B6FD0FC8E74B817ULL, "132761665.ogg"},
	{0x4B6FE61685B76B78ULL, "aip_cartaassassin_base.gda"},
	{0x4B701BB577963307ULL, "follower_banter6171133.cl"},
	{0x4B704D5DF3C88677ULL, "vxlt_blackemporium_zb_0_0.vlm"},
	{0x4B70505FB0A515E6ULL, "gen00fl_dog6116492.cl"},
	{0x4B7088C0CA049724ULL, "w_bonepitmining_a2_82_0.msh"},
	{0x4B709AC0CA04B58DULL, "w_bonepitmining_a2_82_0.mao"},
	{0x4B709EC0CA04BCC6ULL, "w_bonepitmining_a2_82_0.mmh"},
	{0x4B70A9773A51A604ULL, "hlac_lt_0_lowtown_docks_196609_0y.dds"},
	{0x4B70FDADEDC95053ULL, "vfx_vfx_p8.psh.fxo.dx11.assembly"},
	{0x4B710DF8A920E63CULL, "hlac_lt_0_lowtown_docks_65535_4x.dds"},
	{0x4B713BFB755AEA06ULL, "6115362_m.fxe"},
	{0x4B71691B394EEC5CULL, "6178448_m.fxe"},
	{0x4B71C2D4EE242F6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_lean_forward_talk.ani"},
	{0x4B71EE7BFF9CE4DFULL, "335581149.ogg"},
	{0x4B7201C06327FE04ULL, "vxdr_dungeon_1mp_0_0.vlm"},
	{0x4B732E01C738E927ULL, "48473270.ogg"},
	{0x4B733EC24EC93BCEULL, "fx_lens_flare.dds"},
	{0x4B73A18C9E100D85ULL, "199132026.ogg"},
	{0x4B73E43525E51EA3ULL, "vxlt_0_lowtown_a2_n_c_gv6_0_0.vlm"},
	{0x4B7428CF930B9E4BULL, "vxlt_0_lowtown_docks_a3_d3e_0_0.vlm"},
	{0x4B7471D93FFC5D92ULL, "mrl310_merrill6120064.cl"},
	{0x4B74A953EF02D1DEULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_chac_n.dds"},
	{0x4B74E55414599BD7ULL, "vxlt_0_lowtown_fi3_0_0.vlm"},
	{0x4B75122F4F99F863ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_basic.phy"},
	{0x4B75499935E6BC37ULL, "dae000_arishok6078471.cl"},
	{0x4B758676A413DE23ULL, "hlac_lt_0_lowtown_docks_a3_131070_4x.dds"},
	{0x4B75A6B54F369938ULL, "vxw_blightlands_of_0_0.vlm"},
	{0x4B7603A54D7E6752ULL, "6035864_m.fxe"},
	{0x4B76282C78AFF562ULL, "hlac_w_encounter2_262147_0x.dds"},
	{0x4B762C32A663761EULL, "6101275_m.fxe"},
	{0x4B7683D2A1CDC052ULL, "vxht_0_hightown_bl5_0_0.vlm"},
	{0x4B770FD930044C40ULL, "vxht_0_hightown_a2_bwv_0_0.vlm"},
	{0x4B771227CF51CB55ULL, "w_encounter3_183_0.msh"},
	{0x4B771427CF51CECBULL, "w_encounter3_183_0.mmh"},
	{0x4B772027CF51E368ULL, "w_encounter3_183_0.mao"},
	{0x4B7790FB7477EBFDULL, "plt_cod_ltr_pro_meeran.plo"},
	{0x4B77D046877D1C4DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft_1.mao"},
	{0x4B7889ADA671E599ULL, "402310111.ogg"},
	{0x4B789B1F91DDE5A1ULL, "qcr220_orsino6060705.cl"},
	{0x4B78BF40F0127354ULL, "vxlt_shop_d0_0_0.vlm"},
	{0x4B7982C569A3CD2EULL, "one350_vinell6087343.cl"},
	{0x4B7A270897FE3914ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_basic.phy"},
	{0x4B7A9F51A96F4779ULL, "498696825.ogg"},
	{0x4B7AB9AD5F5592A5ULL, "qun110_petrice6036410.cl"},
	{0x4B7B950986931B58ULL, "qun100cr_qunari_warrior5.utc"},
	{0x4B7BADD52D7C4BA4ULL, "6085616_m.fxe"},
	{0x4B7BEA8CE6471A84ULL, "176444312.ogg"},
	{0x4B7BF0F90E95C3C4ULL, "lt_0_lowtown_docks_n_171_0.mmh"},
	{0x4B7BF4F90E95CA17ULL, "lt_0_lowtown_docks_n_171_0.mao"},
	{0x4B7C06F90E95E966ULL, "lt_0_lowtown_docks_n_171_0.msh"},
	{0x4B7C3857F1BF9220ULL, "hlac_lt_hangedman_-1_23z.dds"},
	{0x4B7C58A5F5B0C451ULL, "6207778_m.fxe"},
	{0x4B7C90D25499454CULL, "397587625.ogg"},
	{0x4B7CA3D63F95A788ULL, "6104021_m.fxe"},
	{0x4B7CB3CF331318ADULL, "cart_n.dds"},
	{0x4B7CC238E79B9B4DULL, "6207768_f.fxe"},
	{0x4B7DD62B245B5BB4ULL, "6066493_m.fxe"},
	{0x4B7E93FB747E0F2AULL, "plt_cod_ltr_pro_meeran.nss"},
	{0x4B7EA53A8DCE5C35ULL, "llac_lt_0_lowtown_n_-2_4.dds"},
	{0x4B7F3402F5BF8283ULL, "llac_lt_0_lowtown_a3_n_c_-65535_7.dds"},
	{0x4B7F98E914C99DBAULL, "vxlt_0_lowtown_a2_n_gn5_0_0.vlm"},
	{0x4B80B06C074B4EF4ULL, "hlac_ht_keep_3735_0y.dds"},
	{0x4B80B65A00FF75A8ULL, "vxlt_0_lowtown_n_h5q_0_0.vlm"},
	{0x4B80D4174D0A271BULL, "vxw_encounter3_fn_0_0.vlm"},
	{0x4B80F2F0F14F4FD6ULL, "llac_lt_0_lowtown_docks_n_-2_0.dds"},
	{0x4B814244A1E04E31ULL, "wm_ico_ambush_roll.dds"},
	{0x4B820A9DF5EB530EULL, "rdr111dg_martin6057033.cl"},
	{0x4B822B892152F412ULL, "84160364.ogg"},
	{0x4B82443697793415ULL, "6093966_m.fxe"},
	{0x4B829AAC8501B6CBULL, "vxlt_0_lowtown_docks_a3_d4d_0_0.vlm"},
	{0x4B82A48DA86CBFC3ULL, "vxw_bonepitmining_a2_26w_0_0.vlm"},
	{0x4B82C16EA32A752CULL, "23794029.ogg"},
	{0x4B82C35B09B24B98ULL, "1008485828.ogg"},
	{0x4B8308D7B9590013ULL, "6152340_m.fxe"},
	{0x4B83D97B106B51F9ULL, "llac_dr_dungeon_3215_0.dds"},
	{0x4B84040898065211ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_basic.msh"},
	{0x4B84120898066A2CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_basic.mao"},
	{0x4B841892412CB1B8ULL, "vxw_blightlands_pg_0_0.vlm"},
	{0x4B841E0898067ECFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_basic.mmh"},
	{0x4B84F5AF93C3D8D2ULL, "vxht_0_hightown_bi6_0_0.vlm"},
	{0x4B84FEA0931F98FEULL, "6119228_f.fxe"},
	{0x4B858C2449A34CCDULL, "6087534_m.fxe"},
	{0x4B859BD4EE3495A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_lean_forward_talk.gad"},
	{0x4B861EA4398522B9ULL, "character_character_p266.psh.fxo.dx11.assembly"},
	{0x4B86792DDE5BB967ULL, "6020547_f.fxe"},
	{0x4B8697A309A2F5CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_table_hands_clasped_2.anb"},
	{0x4B869CA9F77E723CULL, "161643870.ogg"},
	{0x4B86EC50A391D0F1ULL, "1059740425.ogg"},
	{0x4B8739DC2AA26E61ULL, "ser210_witness6168128.cl"},
	{0x4B8779E6900EF1A7ULL, "mag130im_danzig_map.uti"},
	{0x4B87EB68C97CC0B1ULL, "6117297_m.fxe"},
	{0x4B87F72F3548EA20ULL, "4100601.ogg"},
	{0x4B87FD9BB0A03140ULL, "vxlt_gallowstemplar_2jf_0_0.vlm"},
	{0x4B884A7364583335ULL, "zz_followers_banter6158082.cl"},
	{0x4B887D139888147DULL, "vxlt_hangedman_n_10t_0_0.vlm"},
	{0x4B88C67B62916753ULL, "vxht_mansion_n_4yn_0_0.vlm"},
	{0x4B88C7DF2AE1A769ULL, "hlac_ca_caverns1_24424_14y.dds"},
	{0x4B893E37309A9437ULL, "pro000_opening6207376.cl"},
	{0x4B8A5216369905D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_meredith.ani"},
	{0x4B8ABF56BFB2EFA4ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkleft.gad"},
	{0x4B8BAD700B6FDBA9ULL, "vxht_bartrand_n_5aj_0_0.vlm"},
	{0x4B8BCFE2ACEC2E73ULL, "vxlt_0_lowtown_docks_bbo_0_0.vlm"},
	{0x4B8BE2750A78689FULL, "llac_lt_gallowscourtyard_n_-131072_1.dds"},
	{0x4B8CA05D8225B290ULL, "vxlt_gallowstemplar_f_2r9_0_0.vlm"},
	{0x4B8D02C4ED70CE56ULL, "498068616.ogg"},
	{0x4B8DB457FE072BA4ULL, "gen_im_cft_rec_psn_04.uti"},
	{0x4B8DD617FD919FD5ULL, "zz_mer_debug6206307.cl"},
	{0x4B8DDA8BF4F7D83CULL, "vxlt_0_lowtown_a2_n_c_byi_0_0.vlm"},
	{0x4B8E10783A03CF77ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_003a_0n.dds"},
	{0x4B8F45F43F003F9BULL, "vxw_encounter3_ca_0_0.vlm"},
	{0x4B8FFE60F648A625ULL, "427204752.ogg"},
	{0x4B9017AE803C70BBULL, "701997002.ogg"},
	{0x4B905D16369D8473ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_meredith.gad"},
	{0x4B90A63F6944862BULL, "lt_0_lowtown_docks_n_30_0_ll.mmh"},
	{0x4B90CFF15E1128BEULL, "t1_tat_blk.tnt"},
	{0x4B90F376D29CB8F3ULL, "llac_w_woundedcoast_-131064_0.dds"},
	{0x4B91BE674DA64EAEULL, "6061414_m.fxe"},
	{0x4B91CE56BFB92782ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkleft.ani"},
	{0x4B92473123E605D2ULL, "vxht_mansion_n_4yk_0_0.vlm"},
	{0x4B9287FEC9D26979ULL, "llac_lt_0_lowtown_a3_n_c_-131073_0.dds"},
	{0x4B92DFB00FBDBCB2ULL, "vxw_woundedcoast_1fx_0_0.vlm"},
	{0x4B9333363F30BC06ULL, "7360019.ogg"},
	{0x4B93657117BDE285ULL, "vxlt_0_lowtown_a3_n_c_ejc_0_0.vlm"},
	{0x4B937ECD3EA90572ULL, "ter_grassa_d.dds"},
	{0x4B9389CA4D509767ULL, "hightown_male_noble_gossip.bnk"},
	{0x4B93D3398C0EEF2BULL, "vxdr_primevaltomb_7zj_0_0.vlm"},
	{0x4B93EBC37B151232ULL, "6154710_m.fxe"},
	{0x4B94488CA984121AULL, "181863038.ogg"},
	{0x4B945BF3E460F253ULL, "551180454.ogg"},
	{0x4B94608BCC361579ULL, "vxht_keep_n_6fw_0_0.vlm"},
	{0x4B955CB75C387CD3ULL, "6125508_m.fxe"},
	{0x4B95C14C20B71CADULL, "6144947_m.fxe"},
	{0x4B95F794602642DFULL, "126258328.ogg"},
	{0x4B9613408E9B0E42ULL, "6162126_m.fxe"},
	{0x4B962ED0136C0173ULL, "5556231.ogg"},
	{0x4B96A96E40B83F47ULL, "6053331_m.fxe"},
	{0x4B96ABCA4D5306DFULL, "hightown_male_noble_gossip.cnv"},
	{0x4B96D8B3FD4EFD19ULL, "8483116.ogg"},
	{0x4B975A1636A39DD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_meredith.evt"},
	{0x4B977C3C557D182EULL, "vxlt_warehouse_n_15h_0_0.vlm"},
	{0x4B97FE024A05AD5FULL, "narrative_varric_demo6170810.cl"},
	{0x4B983253638245C9ULL, "284501322.ogg"},
	{0x4B984F5885CC31BCULL, "6201565_f.fxe"},
	{0x4B9857726053B8D8ULL, "6177907_f.fxe"},
	{0x4B997F4641A70570ULL, "2342791.ogg"},
	{0x4B9A45FF938BB189ULL, "346893754.ogg"},
	{0x4B9A70A66F521C61ULL, "6149232_m.fxe"},
	{0x4B9AFA3A2B576E70ULL, "326579574.ogg"},
	{0x4B9B12355E965073ULL, "vxw_woundedcoast_e4_0_1.vlm"},
	{0x4B9BAB657E01F970ULL, "vxlt_0_lowtown_docks_9r7_0_0.vlm"},
	{0x4B9BBF66CBBAB4E0ULL, "892956113.ogg"},
	{0x4B9BD1BBD74AB64CULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tal_dissp.evt"},
	{0x4B9C026750426353ULL, "876349240.ogg"},
	{0x4B9C3E1145B490ACULL, "zz_vault_debug6201716.cl"},
	{0x4B9C48F3DBE20178ULL, "gen00fl_player6207744.cl"},
	{0x4B9C502B7CEE519CULL, "gen00fl_aveline6113631.cl"},
	{0x4B9CAAF0D966BD72ULL, "vxlt_0_lowtown_n_d4t_0_0.vlm"},
	{0x4B9CAD536697C2ECULL, "qun120cr_javaris6042995.cl"},
	{0x4B9CF3B828443B49ULL, "398536366.ogg"},
	{0x4B9D730BDED4AAF7ULL, "mer210_coterie_contact6079346.cl"},
	{0x4B9E11A230435040ULL, "820044002.ogg"},
	{0x4B9EBE28DBA14522ULL, "one120_athenril6113293.cl"},
	{0x4B9ECE7E3F745CC5ULL, "dae000_player_generic_find6179046.cl"},
	{0x4B9F35EB9B3464BDULL, "213933543.ogg"},
	{0x4B9F514A66C35015ULL, "hlac_w_blightlands_-131072_0y.dds"},
	{0x4BA1073595D8B2D3ULL, "vxlt_gallowstemplar_5t8_0_0.vlm"},
	{0x4BA109BA36A328F3ULL, "vxht_keep_n_4bk_0_0.vlm"},
	{0x4BA15BA237E19126ULL, "zz_vrc_debug6200257.cl"},
	{0x4BA16963EAC445CEULL, "6088806_m.fxe"},
	{0x4BA1961839FD0D98ULL, "dae000ip_crowd_mf2_talk6.utp"},
	{0x4BA245167E0507ABULL, "vxdr_primevaltomb_7wm_0_0.vlm"},
	{0x4BA2657005E695C0ULL, "vxht_0_hightown_a2_bfu_0_0.vlm"},
	{0x4BA2AD93223F8706ULL, "dae000_merchant_weapons_dm6164406.cl"},
	{0x4BA2AE91A57DEA99ULL, "692090164.ogg"},
	{0x4BA2B7BA29370F02ULL, "qun221_viscount6070172.cl"},
	{0x4BA2D268BE2C2DF9ULL, "vxht_keep_n_6av_0_0.vlm"},
	{0x4BA3B7FB29175024ULL, "6093645_f.fxe"},
	{0x4BA3EA8A8B8F39A6ULL, "6051624_m.fxe"},
	{0x4BA3FEE7EAC37604ULL, "vxlt_0_lowtown_a2_ekg_0_2.vlm"},
	{0x4BA43F876A35B949ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.presseddown.ani"},
	{0x4BA484B2ED229C3DULL, "art\\levels\\ztools\\default\\generic\\0fw_basic.phy"},
	{0x4BA4851D809126C2ULL, "6167226_m.fxe"},
	{0x4BA4B17626B4B928ULL, "art\\characters\\creatures\\stonegolem\\c_stonegolem.msh"},
	{0x4BA4C37626B4D7B9ULL, "art\\characters\\creatures\\stonegolem\\c_stonegolem.mao"},
	{0x4BA4CF7626B4EC52ULL, "art\\characters\\creatures\\stonegolem\\c_stonegolem.mmh"},
	{0x4BA55E2EFBDF9682ULL, "452201925.ogg"},
	{0x4BA59D90BF733732ULL, "935601024.ogg"},
	{0x4BA5ED523DB7205EULL, "6140337_m.fxe"},
	{0x4BA7187AC3F86273ULL, "vxw_bonepitdestroyed_a3_1vb_0_0.vlm"},
	{0x4BA7E0B438C51665ULL, "6088287_m.fxe"},
	{0x4BA82BC254EAF522ULL, "771135356.ogg"},
	{0x4BA8492852D7ACB3ULL, "181343043.ogg"},
	{0x4BA8594C84EBE66AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon2_d.dds"},
	{0x4BA862CF9CD749F8ULL, "6040596_m.fxe"},
	{0x4BA89129EF5C0CA9ULL, "vxht_bartrand_n_5oh_0_0.vlm"},
	{0x4BA894492CD25DEBULL, "122788601.ogg"},
	{0x4BA8BDA4A7A35FDEULL, "vxdr_primevaltomb_2xh_0_0.vlm"},
	{0x4BA8DBD6224EA2CEULL, "162927234.ogg"},
	{0x4BA968011B7859EAULL, "6121004_m.fxe"},
	{0x4BA9870F6A2B1101ULL, "6036691_m.fxe"},
	{0x4BA9BA9C266C879FULL, "vxlt_warehouse_11b_0_0.vlm"},
	{0x4BAA0AB0C9CDDC13ULL, "hlac_w_bonepit_458756_0z.dds"},
	{0x4BAA5A876A3A52CBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.presseddown.gad"},
	{0x4BAAAD78C36503E5ULL, "vxlt_hangedman_2w9_0_0.vlm"},
	{0x4BAAFBD28F3BFEF0ULL, "6121975_f.fxe"},
	{0x4BAB2B85D5365F0FULL, "zz_dae_debug6050733.cl"},
	{0x4BAB3DDAE52DD8E8ULL, "vxlt_0_lowtown_a2_n_hfd_0_0.vlm"},
	{0x4BAB45A29E29554EULL, "667972151.ogg"},
	{0x4BAB77EBFFA266F1ULL, "llac_lt_undercityserialkiller_-1_1.dds"},
	{0x4BABD33E3F61247DULL, "184319670.ogg"},
	{0x4BAC07FD6DDC692FULL, "hlac_ht_0_hightown_-196609_0x.dds"},
	{0x4BAC439D4B42B5BFULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm02.msh"},
	{0x4BAC499D4B42C061ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm02.mmh"},
	{0x4BACAB0AEC0DA982ULL, "hlac_lt_0_lowtown_docks_n_-2_2x.dds"},
	{0x4BACAE6D6D480EF1ULL, "cod_mgc_veil.utp"},
	{0x4BAD460B8CBAC108ULL, "one100dg_elren6014723.cl"},
	{0x4BAD65A6BA20C7F3ULL, "t3_arm_gld.tnt"},
	{0x4BAD9D11C6113C81ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_diagonal_1.mao"},
	{0x4BADF3C940BBE4BCULL, "407333404.ogg"},
	{0x4BAE5254C6B7FAD0ULL, "mcr310_party_banter6200734.cl"},
	{0x4BAEDCF8F002D6F5ULL, "6167247_m.fxe"},
	{0x4BAF19610E205FFFULL, "hlac_lt_darktown_-1_63x.dds"},
	{0x4BAF729CEA285587ULL, "6015828_f.fxe"},
	{0x4BAF86A65B66FD8FULL, "66818439.ogg"},
	{0x4BAF8DF9C1FA95B0ULL, "6075721_m.fxe"},
	{0x4BAFB0534BD5808BULL, "vxlt_gallowstemplar_n_2jf_0_0.vlm"},
	{0x4BAFC369F3A9EDB2ULL, "vxw_woundedcoast_1lz_0_0.vlm"},
	{0x4BB0492AFBAA1385ULL, "vxlt_0_lowtown_a3_n_c_epa_0_0.vlm"},
	{0x4BB06004EC62B04DULL, "vxht_brothel_n_6k7_0_0.vlm"},
	{0x4BB0801342F7537AULL, "618372861.ogg"},
	{0x4BB0D0640512666AULL, "follower_banter6164778.cl"},
	{0x4BB0DA2BF400D720ULL, "594693242.ogg"},
	{0x4BB157876A406C30ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.presseddown.evt"},
	{0x4BB248475EB8FCE5ULL, "252469762.ogg"},
	{0x4BB2CBE39CCD747FULL, "6108732_m.fxe"},
	{0x4BB2E334AE765F6BULL, "vxw_encounter3_d8_0_0.vlm"},
	{0x4BB2F2F5637DC48DULL, "hlac_lt_undercityencounter_-1_8y.dds"},
	{0x4BB2F69ECC0E2B95ULL, "vxlt_hangedman_2cu_0_0.vlm"},
	{0x4BB3560C4CA530F3ULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_exit.ani"},
	{0x4BB370B9C1DD316DULL, "780216036.ogg"},
	{0x4BB423E50B092917ULL, "cod_mgc_andraste.utp"},
	{0x4BB49C11A2E7DD93ULL, "263930040.ogg"},
	{0x4BB50CB14493693CULL, "610094592.ogg"},
	{0x4BB55AD0677D9009ULL, "6076602_m.fxe"},
	{0x4BB579B2ED30FCA9ULL, "art\\levels\\ztools\\default\\generic\\0fw_basic.mao"},
	{0x4BB585B2ED311142ULL, "art\\levels\\ztools\\default\\generic\\0fw_basic.mmh"},
	{0x4BB587B2ED3114F8ULL, "art\\levels\\ztools\\default\\generic\\0fw_basic.msh"},
	{0x4BB65BFA25A94CCEULL, "339430929.ogg"},
	{0x4BB6C2B8BAB96B42ULL, "hlac_ht_mansion2_818_17x.dds"},
	{0x4BB6DC2B8DC7E8B9ULL, "ci_smoke03.dds"},
	{0x4BB6FCEAB937E759ULL, "6092302_f.fxe"},
	{0x4BB70BA90C042B05ULL, "da2_prc_one_es-es.cre"},
	{0x4BB768AAF6FAD7D4ULL, "hlac_lt_undercitydungeon_-1_4y.dds"},
	{0x4BB79495D7D303DBULL, "ser200_aveline6201582.cl"},
	{0x4BB7FBB86538CD82ULL, "580775837.ogg"},
	{0x4BB8005E1856020EULL, "hlac_ca_cave_2194_12z.dds"},
	{0x4BB82C791862A01FULL, "vxlt_warehouse_14m_0_0.vlm"},
	{0x4BB8348EF4532C08ULL, "genip_bucket.utp"},
	{0x4BB88D87A21B0414ULL, "1036923262.ogg"},
	{0x4BB892C6B31A5ACDULL, "vrc321ip_spooky7_obj_9.utp"},
	{0x4BB898C4445B9530ULL, "vxlt_gallowstemplar_473_0_0.vlm"},
	{0x4BB8AC9E44D62DA7ULL, "6179776_m.fxe"},
	{0x4BB8BAFE1E7B8CDDULL, "vxht_mansion_n_5au_0_0.vlm"},
	{0x4BB8C68519944E23ULL, "qun131_taskmaster6059015.cl"},
	{0x4BB8CF2E6FE43FB1ULL, "vxlt_blackemporium_16a_0_0.vlm"},
	{0x4BB9B7A895CFC598ULL, "452113226.ogg"},
	{0x4BB9B9FC6ECCFFDFULL, "vxht_mansion2b_1at_0_0.vlm"},
	{0x4BB9E2DC5E0AB1D8ULL, "470177783.ogg"},
	{0x4BBA3CF3A11D01AEULL, "dae000_bartrand6206869.cl"},
	{0x4BBA43D82F1F0D70ULL, "art\\levels\\ztools\\default\\generic\\0dd22_basic.mao"},
	{0x4BBA47D82F1F1383ULL, "art\\levels\\ztools\\default\\generic\\0dd22_basic.mmh"},
	{0x4BBA55D82F1F2BDDULL, "art\\levels\\ztools\\default\\generic\\0dd22_basic.msh"},
	{0x4BBA933FFB44E24DULL, "6101496_m.fxe"},
	{0x4BBAD4072E5DAE71ULL, "689000587.ogg"},
	{0x4BBAFE5E0F989F6DULL, "sst_ss_city_guard_f_1.gda"},
	{0x4BBB36625360E695ULL, "vxht_0_hightown_a2_bdz_0_0.vlm"},
	{0x4BBB41C73C13F98AULL, "6201519_m.fxe"},
	{0x4BBB6B624AAB4DBFULL, "hlac_w_sundermount_-5_0z.dds"},
	{0x4BBBD123B5E08E54ULL, "353074893.ogg"},
	{0x4BBC30D04D31F773ULL, "character_character_p202.psh.fxo.dx11"},
	{0x4BBC8391A588A530ULL, "mag100_karras6034543.cl"},
	{0x4BBCB53A4D4B9642ULL, "vxlt_0_lowtown_a2_n_h22_0_0.vlm"},
	{0x4BBCF2F2BE5F5AF3ULL, "415992782.ogg"},
	{0x4BBD355063DFCB1CULL, "vxlt_0_lowtown_a3_n_c_el9_0_1.vlm"},
	{0x4BBD88FE61FC08C6ULL, "440869920.ogg"},
	{0x4BBE10622056DACDULL, "lowtown_male_lowlife2_mob6166770.cl"},
	{0x4BBE2C4528AE5600ULL, "6081740_m.fxe"},
	{0x4BBE91C9E33999F8ULL, "833991421.ogg"},
	{0x4BBED1E1DE58C8CDULL, "vxht_brothel_n_6t6_0_0.vlm"},
	{0x4BBF40153F2F8294ULL, "gen00fl_merrill6108465.cl"},
	{0x4BBF6D6605672EABULL, "vxlt_0_lowtown_a3_n_c_gkp_0_0.vlm"},
	{0x4BBF72B5BF78B434ULL, "vxdr_0_deeproads_9c6_0_0.vlm"},
	{0x4BC015965F2F78BDULL, "111647081.ogg"},
	{0x4BC0366105975557ULL, "vxht_mansion_n_5hx_0_0.vlm"},
	{0x4BC04E0C4CAFAD2AULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_exit.evt"},
	{0x4BC07C563E79E092ULL, "6152723_m.fxe"},
	{0x4BC087C5605F4962ULL, "vxht_mansion2b_28i_0_0.vlm"},
	{0x4BC0945390CCF3E6ULL, "areadata.gda"},
	{0x4BC0F0B1DE69DF0BULL, "6161016_m.fxe"},
	{0x4BC14C09689F309AULL, "6134587_m.fxe"},
	{0x4BC14D6116A6BD5EULL, "lowtown_fem_alienage1_sta6161574.cl"},
	{0x4BC14F1CC1424994ULL, "6162959_m.fxe"},
	{0x4BC22E198799156BULL, "vxca_caverns1_inq_0_0.vlm"},
	{0x4BC22E5342EE467DULL, "6034428_m.fxe"},
	{0x4BC2321411C1D467ULL, "vxlt_0_lowtown_docks_aqa_0_0.vlm"},
	{0x4BC244735C396145ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_0z.dds"},
	{0x4BC267E9C722F2F3ULL, "dae000_arishok6072868.cl"},
	{0x4BC2A05E54A1DC76ULL, "mnd_imp_distor.mao"},
	{0x4BC35C77CD8D4D0EULL, "hlac_ht_playershouse_n_855_19y.dds"},
	{0x4BC3A9FA89A72AB3ULL, "zz_lgt_2_debug6140447.cl"},
	{0x4BC3B4103A36C4F8ULL, "130119034.ogg"},
	{0x4BC408A71864FF24ULL, "vxlt_0_lowtown_docks_a3_9rd_0_0.vlm"},
	{0x4BC50701FD7B5E7FULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_act_c.impa.gad"},
	{0x4BC52D9D764BB03FULL, "6027981_m.fxe"},
	{0x4BC5581903E02C45ULL, "1011060130.ogg"},
	{0x4BC59E5157D56465ULL, "vxlt_0_lowtown_n_coy_0_0.vlm"},
	{0x4BC5F68389FE8DF3ULL, "fen110_anso6094948.cl"},
	{0x4BC6E8588A622330ULL, "144725650.ogg"},
	{0x4BC72D0C4CB5939DULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_exit.gad"},
	{0x4BC7813CF5642018ULL, "vxw_woundedcoast_4y_0_0.vlm"},
	{0x4BC7C461D9C00D54ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_4x.dds"},
	{0x4BC7CA3B2BAE39A2ULL, "gen_im_acc_amu_lore_04.uti"},
	{0x4BC80C944E8D12C0ULL, "6133535_m.fxe"},
	{0x4BC82BD960C3185FULL, "vxht_mansion2b_1dw_0_0.vlm"},
	{0x4BC842060689166CULL, "83446087.ogg"},
	{0x4BC876AEA7C1DE89ULL, "6147018_m.fxe"},
	{0x4BC8BCBBF475510FULL, "vxlt_gallowstemplar_35a_0_0.vlm"},
	{0x4BC930019A908B1AULL, "fx_f3c3.mao"},
	{0x4BC985EF8C40520EULL, "hlac_lt_0_lowtown_docks_-65539_2y.dds"},
	{0x4BC9A83F4556FF15ULL, "vxht_0_hightown_a2_biy_0_0.vlm"},
	{0x4BC9E60CCD0695C5ULL, "vxlt_0_lowtown_docks_a3_n_d0u_0_0.vlm"},
	{0x4BC9ECC056697A9AULL, "vxlt_0_lowtown_a2_n_ei2_0_0.vlm"},
	{0x4BCA20A62CCD8F0AULL, "6130129_m.fxe"},
	{0x4BCA64380A607086ULL, "240663315.ogg"},
	{0x4BCA74BD4E7FB113ULL, "gen00fl_bethany6028810.cl"},
	{0x4BCA755D340C586CULL, "6041410_m.fxe"},
	{0x4BCA9E68CA062598ULL, "vxw_bonepit_14c_0_0.vlm"},
	{0x4BCAFC01FD7FB76CULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_act_c.impa.evt"},
	{0x4BCB31D3E1B060D5ULL, "6069707_m.fxe"},
	{0x4BCB38610C756217ULL, "vxlt_undercityencounter_1bx_0_0.vlm"},
	{0x4BCB48D82F2D88A8ULL, "art\\levels\\ztools\\default\\generic\\0dd22_basic.phy"},
	{0x4BCB4B221F0E980BULL, "vxht_0_hightown_cb7_0_0.vlm"},
	{0x4BCC0CED63F6B016ULL, "6076270_m.fxe"},
	{0x4BCC5972C41CDEF0ULL, "vxht_0_hightown_a2_bm4_0_0.vlm"},
	{0x4BCCA723D7961EB2ULL, "vxw_woundedcoast_1rd_0_0.vlm"},
	{0x4BCCD4F06AFBC106ULL, "genip_crate_2.utp"},
	{0x4BCD1C3745CB7613ULL, "hlac_lt_0_lowtown_131070_6z.dds"},
	{0x4BCD6479E4FD6EF5ULL, "vxlt_0_lowtown_e12_0_0.vlm"},
	{0x4BCDE492B16ECCB4ULL, "vxdr_0_deeproads_9d7_0_0.vlm"},
	{0x4BCE36DF3A51416BULL, "hlac_ht_mansion2b_818_11y.dds"},
	{0x4BCE73C858533554ULL, "235972743.ogg"},
	{0x4BCEB7C05AA09866ULL, "warrior_4.cut"},
	{0x4BCEF525B558113BULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_106a_0t.dds"},
	{0x4BCEF9A2525561E2ULL, "vxht_mansion2b_25j_0_0.vlm"},
	{0x4BCF2E16BF03D04CULL, "dae000ip_crowd_mf3_talk7.utp"},
	{0x4BCF628ED05FF78BULL, "6166116_m.fxe"},
	{0x4BCFC336E4FBED32ULL, "hlac_ca_cavern1_haunted_24424_22x.dds"},
	{0x4BCFD8AEAC8E27E3ULL, "eye_eye_p0.psh.fxo.dx11"},
	{0x4BD008D705DB9D17ULL, "mer210_seldrek6078862.cl"},
	{0x4BD1A0CC07E8CF6BULL, "and110_refugee6111597.cl"},
	{0x4BD1C8FB7FBAF634ULL, "dre000_varric6034655.cl"},
	{0x4BD2DFF28A1BB3DDULL, "vxlt_hangedman_1ru_0_0.vlm"},
	{0x4BD2EA04FF55C3F2ULL, "hlac_lt_0_lowtown_a2_-1_6y.dds"},
	{0x4BD2FFBF04C23F06ULL, "370647708.ogg"},
	{0x4BD38EA8C112F8B6ULL, "vxca_caverns2_8bj_0_0.vlm"},
	{0x4BD4FC2F43A5A463ULL, "gen00fl_anders6096512.cl"},
	{0x4BD550838657E01DULL, "lt_0_lowtown_a2_52_0_ll.mmh"},
	{0x4BD5732E7FD46AB2ULL, "llac_lt_0_lowtown_131070_5.dds"},
	{0x4BD5A6B8B91A32DDULL, "si_cl_2.mao"},
	{0x4BD65D46FE36EB0DULL, "4667387.ogg"},
	{0x4BD67620BC19C768ULL, "6024972_m.fxe"},
	{0x4BD695D084F6474FULL, "434803616.ogg"},
	{0x4BD6DC29311E11D0ULL, "gen00fl_merrill6106870.cl"},
	{0x4BD6EA277AC5D83DULL, "mrl251ip_bosshorde_b.utp"},
	{0x4BD722F5EA28D082ULL, "775473810.ogg"},
	{0x4BD72C5BB3318D48ULL, "hlac_lt_gallowstemplar_f_2394_7x.dds"},
	{0x4BD7389FE503C75FULL, "70999391.ogg"},
	{0x4BD7CE2C12E6CD5EULL, "vxht_playershouse_n_1si_0_0.vlm"},
	{0x4BD7CF91C5A8E771ULL, "gen00fl_fenris6126372.cl"},
	{0x4BD805B920F76452ULL, "sib111_mother6098851.cl"},
	{0x4BD84A9D4B67FCDEULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm02.phy"},
	{0x4BD85E9D485F931AULL, "vxlt_0_lowtown_a2_n_er3_0_0.vlm"},
	{0x4BD8E401FD8BCBC5ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_act_c.impa.ani"},
	{0x4BD955E695C79F23ULL, "art\\vfx\\textures\\level_fx\\vfx_bug_01.dds"},
	{0x4BD9C9467F138CDEULL, "vxlt_warehouse_1d0_0_0.vlm"},
	{0x4BD9F2D0370219F6ULL, "881395783.ogg"},
	{0x4BDA03B01709078CULL, "6208298_f.fxe"},
	{0x4BDA075E07279589ULL, "374191033.ogg"},
	{0x4BDB1B47348224B4ULL, "vxlt_0_lowtown_n_fe7_0_0.vlm"},
	{0x4BDB50AEBA98BECCULL, "6174916_m.fxe"},
	{0x4BDB6A1502C79511ULL, "dae000ip_craft_potions.utp"},
	{0x4BDB6BFCE2431D86ULL, "gen00fl_sebastian6174715.cl"},
	{0x4BDB8BCE2F04794AULL, "hlac_lt_0_lowtown_a2_n_-3_5y.dds"},
	{0x4BDBC078AF6DCC80ULL, "6153825_m.fxe"},
	{0x4BDBD51AE84D0D74ULL, "vxlt_hangedman_n_1p9_0_0.vlm"},
	{0x4BDBDB8121CEE841ULL, "qun110_petrice6042156.cl"},
	{0x4BDC687A3F0AF726ULL, "hlac_ht_0_hightown_n_-3_0z.dds"},
	{0x4BDC8674CB1826AFULL, "gen00fl_anders6119165.cl"},
	{0x4BDC8DA5F65A8AF7ULL, "209278557.ogg"},
	{0x4BDCC0BA61A97584ULL, "6125576_m.fxe"},
	{0x4BDCE7CFEA66ABE5ULL, "6108921_m.fxe"},
	{0x4BDD21BD9B514A92ULL, "vxca_cavern1_haunted_foa_0_0.vlm"},
	{0x4BDE316C3D747A6AULL, "399115396.ogg"},
	{0x4BDE7A9B4C852002ULL, "6661739.ogg"},
	{0x4BDEA4D8D5D94C93ULL, "6000861_m.fxe"},
	{0x4BDF241813668A14ULL, "6047724_m.fxe"},
	{0x4BDF903A7029942DULL, "861768269.ogg"},
	{0x4BE05709372BA78BULL, "6179797_f.fxe"},
	{0x4BE0A6FBCA5AC39FULL, "203454432.ogg"},
	{0x4BE0DFEE8BD0C473ULL, "vxw_bonepitdestroyed_a3_1jf_0_0.vlm"},
	{0x4BE14F636D11514FULL, "art\\characters\\weapons\\unique\\textures\\w_vael_0s.dds"},
	{0x4BE173A964AA682DULL, "art\\characters\\weapons\\longbows\\w_lbw_002a.mmh"},
	{0x4BE177A964AA6F7EULL, "art\\characters\\weapons\\longbows\\w_lbw_002a.mao"},
	{0x4BE17A2491EDFB02ULL, "774939803.ogg"},
	{0x4BE185A964AA86B3ULL, "art\\characters\\weapons\\longbows\\w_lbw_002a.msh"},
	{0x4BE18E02DAA75A5EULL, "llac_ht_keep_n_4519_3.dds"},
	{0x4BE1C795D883BA98ULL, "dae000_orsino6085501.cl"},
	{0x4BE1D8898F0CD9B2ULL, "614747638.ogg"},
	{0x4BE1F73006AD44B6ULL, "vxht_0_hightown_bfo_0_0.vlm"},
	{0x4BE20085B3091136ULL, "vxca_caverns2_8ik_0_0.vlm"},
	{0x4BE20387281C7F1DULL, "ht_keep_3733.rml"},
	{0x4BE2F18F90B1077EULL, "700552030.ogg"},
	{0x4BE325418BCBCFB7ULL, "6060150_f.fxe"},
	{0x4BE38FC0661F093AULL, "lowtown_male_alienage1_sta6162808.cl"},
	{0x4BE428A830EA9049ULL, "dae000_bodhan6146898.cl"},
	{0x4BE433FB34BE33DAULL, "vxlt_gallowstemplar_5x9_0_0.vlm"},
	{0x4BE447ADB22AC2ADULL, "vxlt_warehouse_n_vu_0_0.vlm"},
	{0x4BE487F71A252B0DULL, "vxht_0_hightown_n_c6v_0_0.vlm"},
	{0x4BE48A96449D50D3ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chab.phy"},
	{0x4BE50F9344AF495FULL, "vxht_mansion2b_1rq_0_0.vlm"},
	{0x4BE547CFE611A264ULL, "482001306.ogg"},
	{0x4BE6400904DCE5DEULL, "vxht_playershouse_n_1vj_0_0.vlm"},
	{0x4BE65FE6C68A7F27ULL, "as.cut"},
	{0x4BE6700CE2A2AD6BULL, "vxlt_gallowscourtyard_65f_0_0.vlm"},
	{0x4BE71A74FF592E6DULL, "382893130.ogg"},
	{0x4BE7A1043DD478C0ULL, "vxlt_0_lowtown_a2_hgp_0_0.vlm"},
	{0x4BE809019AAAF9CCULL, "fx_f3c3.dds"},
	{0x4BE87C99A44B465AULL, "6018599_m.fxe"},
	{0x4BE8AB4EE997EFAEULL, "497800231.ogg"},
	{0x4BE919B98BBF38BCULL, "vxw_woundedcoast_1ch_0_1.vlm"},
	{0x4BE92409860C41B1ULL, "gen00fl_anders6125595.cl"},
	{0x4BEA2FD18B2DD69BULL, "vxdr_primevalentrance_3dh_0_0.vlm"},
	{0x4BEAB4CFB5DAE9C0ULL, "aip_coward.gda"},
	{0x4BEAC70C42AB698AULL, "99956518.ogg"},
	{0x4BEAF362C32FBC61ULL, "6042681_m.fxe"},
	{0x4BEB54C49E7688CBULL, "zz_mcr_debug6098603.cl"},
	{0x4BEB6FE0DBF7D22BULL, "hlac_ht_chantry_2534_24z.dds"},
	{0x4BEC725BFE815B1DULL, "llac_w_sundermount_-1_0.dds"},
	{0x4BEC7D2F14718108ULL, "vxlt_0_lowtown_a2_eod_0_1.vlm"},
	{0x4BECE68640F88FC9ULL, "vxdr_primevaltomb_504_0_0.vlm"},
	{0x4BED097A8D86C4EBULL, "hlac_ht_keep_3732_3z.dds"},
	{0x4BED18D09D46435FULL, "424750827.ogg"},
	{0x4BED41C098B6BD76ULL, "248434627.ogg"},
	{0x4BEDA5F701D2015DULL, "82738380.ogg"},
	{0x4BEDED258FE60E68ULL, "llac_lt_0_lowtown_docks_a3_n_65534_1.dds"},
	{0x4BEE5DE5EF542130ULL, "6110957_f.fxe"},
	{0x4BEE64F7F86F8BD8ULL, "vxlt_gallowstemplar_6q7_0_0.vlm"},
	{0x4BEE98157CF86478ULL, "pro000st_captain_wright.stg"},
	{0x4BEEA8E88A132BCBULL, "6016930_m.fxe"},
	{0x4BEF65EC49A942BBULL, "gen00fl_varric6076220.cl"},
	{0x4BEF6D09C7B62312ULL, "141304256.ogg"},
	{0x4BEF7D3EC3AF3F3BULL, "6098828_m.fxe"},
	{0x4BEFDFEDE926A7CAULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastplants_dp.dds"},
	{0x4BEFF7685498FD00ULL, "6110415_m.fxe"},
	{0x4BEFFCD2B024FA6DULL, "260553936.ogg"},
	{0x4BF1E4FD714942C2ULL, "vxht_mansion2_21n_0_0.vlm"},
	{0x4BF200199958B115ULL, "vxlt_0_lowtown_a2_hdx_0_0.vlm"},
	{0x4BF21BA13733A9E4ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_meda_0d.dds"},
	{0x4BF21CD43045864FULL, "6165683_m.fxe"},
	{0x4BF2984BCBC9551CULL, "28270395.ogg"},
	{0x4BF2F9D40C1B438DULL, "vxht_0_hightown_n_c3u_0_0.vlm"},
	{0x4BF30513E7D259CFULL, "600937541.ogg"},
	{0x4BF355CD14F91D39ULL, "gen_im_npc_wep_one_maca.uti"},
	{0x4BF360E38C87EC8EULL, "vxht_0_hightown_a3_bhx_0_0.vlm"},
	{0x4BF3817036A561DFULL, "vxht_mansion2b_1up_0_0.vlm"},
	{0x4BF42EA4CCA72E1AULL, "hlac_lt_gallowscourtyard_65534_3z.dds"},
	{0x4BF43DE6E8555D6CULL, "6030649_m.fxe"},
	{0x4BF485FA6968C842ULL, "91246969.ogg"},
	{0x4BF4CDE016D5DE96ULL, "dae000_keep_male_servant_mo6161436.cl"},
	{0x4BF51F076611F97CULL, "clf260_harley6069697.cl"},
	{0x4BF57940A5D4D093ULL, "hlac_lt_0_lowtown_a2_n_c_65534_2x.dds"},
	{0x4BF58CAE3A711E44ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal_2.mao"},
	{0x4BF5F8BA91313CF0ULL, "hlac_w_sundermount_65533_1x.dds"},
	{0x4BF69614EB021A6FULL, "91941681.ogg"},
	{0x4BF6F64079BA5F12ULL, "944864265.ogg"},
	{0x4BF84DE8AFA0CA5BULL, "lgt150_nat_were_supporter6110977.cl"},
	{0x4BF8CD4B8C1CED41ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopen_v2.gad"},
	{0x4BF9498BFA99E376ULL, "634442010.ogg"},
	{0x4BF99DE60BA8BA4FULL, "vxw_bonepitmining_a2_1wj_0_1.vlm"},
	{0x4BF9D291FFA859B8ULL, "dae021_sister_samea_sta6142260.cl"},
	{0x4BFB06B6BC21C8C3ULL, "6002005_m.fxe"},
	{0x4BFB458BFCDE7CEEULL, "615921145.ogg"},
	{0x4BFC1BB8E3473DF0ULL, "vxca_caverns2prc_8dk_0_0.vlm"},
	{0x4BFC50EB2DD742D0ULL, "718889249.ogg"},
	{0x4BFC885145F60D49ULL, "zz_approval_debug6003561.cl"},
	{0x4BFD3C506969A5C7ULL, "vxw_bonepitmining_a2_26k_0_0.vlm"},
	{0x4BFD98CF1380BB74ULL, "vxlt_0_lowtown_euc_0_0.vlm"},
	{0x4BFDB9853544A89BULL, "512013977.ogg"},
	{0x4BFDBE703C8406DBULL, "follower_barks6153471.cl"},
	{0x4BFDCDAB1158A5A2ULL, "vxw_sundermount_32o_0_1.vlm"},
	{0x4BFE97E0604D11AEULL, "gen_im_trs_tmp_hvy_cht_r1.uti"},
	{0x4BFED2F7D9C63F53ULL, "125583977.ogg"},
	{0x4BFF137047F302F6ULL, "static_static_p12.psh.fxo.dx11.assembly"},
	{0x4BFF2CC191C0F71BULL, "6171461_f.fxe"},
	{0x4BFF4E3134E6A79EULL, "161004354.ogg"},
	{0x4BFF86A2E97AE6D8ULL, "gen_im_trs_qun_biscuit_04.uti"},
	{0x4BFFBF2C865AF81DULL, "175525738.ogg"},
	{0x4BFFCE4B8C230DEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopen_v2.ani"},
	{0x4C00CCA16AEBE83CULL, "475776273.ogg"},
	{0x4C016BB0FE115C0DULL, "vxht_0_hightown_n_c8t_0_0.vlm"},
	{0x4C01C357BB9AB69CULL, "hf_mother_h2s1.mor"},
	{0x4C01CD80E37E95F5ULL, "509088658.ogg"},
	{0x4C01E6672DEE62C8ULL, "vxca_caverns1_mines_jlj_0_0.vlm"},
	{0x4C02B02B47432391ULL, "6053012_m.fxe"},
	{0x4C034EBC8DF66458ULL, "6020499_m.fxe"},
	{0x4C03AE54220C0291ULL, "257288786.ogg"},
	{0x4C03F932C3FF5E7AULL, "6115487_m.fxe"},
	{0x4C047A5C95EE048EULL, "135020922.ogg"},
	{0x4C05B0EC4719C5D7ULL, "vxlt_hangedman_1hu_0_0.vlm"},
	{0x4C05E4A487AC4E33ULL, "vxlt_gallowstemplar_f_3qr_0_0.vlm"},
	{0x4C0620AE3C95169FULL, "matdef_dh_diffuse.dds"},
	{0x4C0621ED6E409631ULL, "6068560_m.fxe"},
	{0x4C06349A1A32517AULL, "377388495.ogg"},
	{0x4C06AEB3FF7D8A6BULL, "6172966_m.fxe"},
	{0x4C06C5BF0638D9AFULL, "vxlt_0_lowtown_a3_n_c_dow_0_0.vlm"},
	{0x4C06FEE793147482ULL, "plt_isa000pt_midstage.nss"},
	{0x4C071C755869450AULL, "476296167.ogg"},
	{0x4C0797E9FDDC352EULL, "385903895.ogg"},
	{0x4C08306CDE0CA8B2ULL, "762861549.ogg"},
	{0x4C0870A3205E4C60ULL, "vxdr_primevaltomb_7u9_0_1.vlm"},
	{0x4C087238BC53865AULL, "588003859.ogg"},
	{0x4C089A19D4910AD8ULL, "6115096_m.fxe"},
	{0x4C08D50545173B48ULL, "148705920.ogg"},
	{0x4C097AB385F2B117ULL, "hlac_ht_mansion2b_818_43z.dds"},
	{0x4C0A3080EF527852ULL, "lt_0_lowtown_a3_n_c_38_0_lc.mmh"},
	{0x4C0A531D626E03DAULL, "mrl210_merrill6116394.cl"},
	{0x4C0A68CFE653008FULL, "ser221st_gascard_hostage.stg"},
	{0x4C0B00FE4E922580ULL, "hlac_dr_slavers_3207_7y.dds"},
	{0x4C0B053450075419ULL, "vxlt_0_lowtown_docks_a3_n_cya_0_0.vlm"},
	{0x4C0B8B450E2886F5ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_008a_0n.dds"},
	{0x4C0B994356FE21D7ULL, "153573744.ogg"},
	{0x4C0B9E46AB8B071AULL, "6103071_m.fxe"},
	{0x4C0BD5EBD4124410ULL, "ico_staff_3.dds"},
	{0x4C0CB6701393D6C4ULL, "gen00fl_anders6113009.cl"},
	{0x4C0D46F30DB78988ULL, "6176222_m.fxe"},
	{0x4C0D50203C7721ADULL, "6180671_m.fxe"},
	{0x4C0E094D3D61CC03ULL, "6217957_m.fxe"},
	{0x4C0E1A2898BE0C69ULL, "6087277_m.fxe"},
	{0x4C0E7D33D874A47BULL, "6127072_m.fxe"},
	{0x4C0EDE4028E18F21ULL, "hlac_ht_mansion2b_818_1x.dds"},
	{0x4C1056DB49480080ULL, "vxlt_gallowscourtyard_6lb_0_0.vlm"},
	{0x4C109EB7CCC4E036ULL, "6054974_f.fxe"},
	{0x4C10F0EC615050A8ULL, "vxlt_0_lowtown_e8d_0_0.vlm"},
	{0x4C110B53E02BB664ULL, "dae000_arishok6072397.cl"},
	{0x4C11345BE03917F7ULL, "gen_im_trs_gen_statue_02.uti"},
	{0x4C121C407A6225FCULL, "character_character_p139.psh.fxo"},
	{0x4C12822F9257CCB9ULL, "barr_wood.mao"},
	{0x4C1297EAA6E7D04FULL, "6142118_m.fxe"},
	{0x4C1318149FC19C35ULL, "llac_lt_0_lowtown_a3_n_c_-65535_1.dds"},
	{0x4C1360252E6A1DEDULL, "6157194_m.fxe"},
	{0x4C13610E70ACCE3EULL, "art\\vfx\\combat\\vfx_glb_ent_imp_maj_c.mmh"},
	{0x4C137E3DBFE49136ULL, "lt_0_lowtown_docks_n_17_0.rml"},
	{0x4C137FA40B0CA6ACULL, "llac_dr_dungeon_3214_0.dds"},
	{0x4C13800A1D37EB3EULL, "vxlt_0_lowtown_a2_n_c_cmr_0_0.vlm"},
	{0x4C1423067DAFB90EULL, "zz_dre_debug6054669.cl"},
	{0x4C14F2C9CA29735BULL, "228774492.ogg"},
	{0x4C1501E5FC0AC7EBULL, "215936287.ogg"},
	{0x4C1502872847D735ULL, "ht_keep_3733.anb"},
	{0x4C1582CBFFC1AB72ULL, "vxlt_gallowstemplar_n_3rf_0_0.vlm"},
	{0x4C15AB013D0E7578ULL, "vxca_cavern1_haunted_j4b_0_0.vlm"},
	{0x4C15C827FC66A1C2ULL, "gen00fl_aveline6107009.cl"},
	{0x4C16DF482CCF76A3ULL, "160762175.ogg"},
	{0x4C16E915FE380599ULL, "gen_im_trs_dal_hvy_hlm_r0.uti"},
	{0x4C16F22337CA9C0BULL, "6078457_m.fxe"},
	{0x4C17463EA531F68DULL, "vxw_woundedcoast_o9_0_0.vlm"},
	{0x4C17C5810D159E3FULL, "130334023.ogg"},
	{0x4C17E4D43B59E2AAULL, "6114061_f.fxe"},
	{0x4C17F172733E020FULL, "app_smo.mao"},
	{0x4C1815E8CD6CBD06ULL, "vxlt_darktown_48h_0_0.vlm"},
	{0x4C18A9AAA54960D8ULL, "231430214.ogg"},
	{0x4C18B586D63B2875ULL, "vxht_mansion_n_62k_0_0.vlm"},
	{0x4C1901F73F9D05B3ULL, "798542679.ogg"},
	{0x4C1910650A1C7C91ULL, "240947889.ogg"},
	{0x4C19F7EBFBEDE284ULL, "gen00fl_anders6125721.cl"},
	{0x4C1A1ABCE3AB95D7ULL, "770583494.ogg"},
	{0x4C1A200A4D55D6C7ULL, "vxw_bonepitmining_a2_28e_0_0.vlm"},
	{0x4C1A9E21839BE806ULL, "placeable_interactions.gda"},
	{0x4C1D0BCB9CE912EDULL, "hlac_lt_0_lowtown_n_2_0y.dds"},
	{0x4C1D1549E3C8A3D8ULL, "qun120_arishok6049648.cl"},
	{0x4C1D480B2119C49CULL, "83032156.ogg"},
	{0x4C1D744F88BDDDC9ULL, "vxw_bonepit_y6_0_0.vlm"},
	{0x4C1D8C3A5EA4F7E6ULL, "hlac_ht_keep_n_4519_2y.dds"},
	{0x4C1DB1354397E677ULL, "464316515.ogg"},
	{0x4C1F62C953466928ULL, "vxlt_0_lowtown_e3e_0_0.vlm"},
	{0x4C1FB5306EA00814ULL, "hla_ht_0_hightown_a3_65537_0y.dds"},
	{0x4C1FC70E7D0C9431ULL, "vxht_keep_n_6fo_0_0.vlm"},
	{0x4C20AEAFBB335886ULL, "vxdr_0_deeproads_gbi_0_0.vlm"},
	{0x4C21399851F2BC85ULL, "142550747.ogg"},
	{0x4C21400E70B8D3FDULL, "art\\vfx\\combat\\vfx_glb_ent_imp_maj_c.anb"},
	{0x4C2158B873F88F91ULL, "vxlt_0_lowtown_docks_bd3_0_0.vlm"},
	{0x4C216B09E9AEB1C6ULL, "522152753.ogg"},
	{0x4C217066E8F10B32ULL, "mag110_samson6018633.cl"},
	{0x4C21962CB2011F22ULL, "w_encounter3_144_0.msh"},
	{0x4C21982CB20122D8ULL, "w_encounter3_144_0.mmh"},
	{0x4C21A42CB201377BULL, "w_encounter3_144_0.mao"},
	{0x4C21C8F8B879AD36ULL, "hlac_w_sundermount_327683_0x.dds"},
	{0x4C222E27C473BE2CULL, "llac_lt_0_lowtown_a3_n_c_-131072_0.dds"},
	{0x4C22413F825D1C16ULL, "732975104.ogg"},
	{0x4C226DC55ED80B15ULL, "zz_fol_debug6098101.cl"},
	{0x4C22B807B2EB1DDCULL, "225075508.ogg"},
	{0x4C22BB1F075E6E89ULL, "248536236.ogg"},
	{0x4C22F63230DFBA11ULL, "vxht_0_hightown_n_cbf_0_0.vlm"},
	{0x4C23157489A34120ULL, "w_encounter2_37_0_lc.mmh"},
	{0x4C235ACA9E92CA25ULL, "llac_lt_backalley_g_0_12.dds"},
	{0x4C23919618E61FC7ULL, "l_one300ar_vanard_fight.bnk"},
	{0x4C23B3C3CDADD2E4ULL, "6170448_m.fxe"},
	{0x4C23B4BDFB5C5032ULL, "170914606.ogg"},
	{0x4C241CDE2F048DF8ULL, "vxca_cavern1_haunted_j3c_0_0.vlm"},
	{0x4C244255406CF3AAULL, "mrl310_merrill6144431.cl"},
	{0x4C24755CC704FB95ULL, "vxht_keep_n_58f_0_0.vlm"},
	{0x4C2515DCF94EF196ULL, "vxdr_0_deeproads_eh5_0_0.vlm"},
	{0x4C25B81B97280F0DULL, "vxw_woundedcoast_d8_0_0.vlm"},
	{0x4C25C8A136B139BFULL, "l_dae221ar_brothel_nt.bnk"},
	{0x4C25DD46EF879E50ULL, "vxht_mansion_n_6hx_0_0.vlm"},
	{0x4C2670B69184A288ULL, "868558857.ogg"},
	{0x4C268A7ECA751283ULL, "714608513.ogg"},
	{0x4C26F5B0A45FB281ULL, "mer200ip_cavern_chest.utp"},
	{0x4C276D560246CD40ULL, "gen00fl_aveline6137549.cl"},
	{0x4C2798478A25FE88ULL, "lt_0_lowtown_docks_a3_27_0_lc.mmh"},
	{0x4C27C48446CFC725ULL, "hlac_lt_0_lowtown_docks_-131074_2z.dds"},
	{0x4C27CEC1F2AE6036ULL, "359927771.ogg"},
	{0x4C27EFF2AABA7884ULL, "vxlt_0_lowtown_docks_9bs_0_0.vlm"},
	{0x4C28783B9392CB78ULL, "678845624.ogg"},
	{0x4C28EE65E96304F4ULL, "vxlt_0_lowtown_eff_0_0.vlm"},
	{0x4C29304D35E8B8D7ULL, "vxw_encounter2_camp_4c_0_0.vlm"},
	{0x4C29ADF3E087945CULL, "505008974.ogg"},
	{0x4C2AE588BF47B4F2ULL, "57873521.ogg"},
	{0x4C2AFB3DB93FFA69ULL, "152989777.ogg"},
	{0x4C2B416246450541ULL, "t3_wdg_tn1.tnt"},
	{0x4C2B57A86D06C5CAULL, "186755396.ogg"},
	{0x4C2B88D589840497ULL, "hlac_lt_backalley_n_65536_4y.dds"},
	{0x4C2BF665F178FC64ULL, "535121072.ogg"},
	{0x4C2C38927A4B6941ULL, "one200dg_solivitus6060342.cl"},
	{0x4C2C682436D2330BULL, "vxlt_0_lowtown_n_eug_0_0.vlm"},
	{0x4C2C9F2CB20B3663ULL, "w_encounter3_144_0.phy"},
	{0x4C2D9B35DD42673CULL, "93537807.ogg"},
	{0x4C2E29E7B51D2FE7ULL, "gen_im_arm_cht_coma_t4.uti"},
	{0x4C2E2D8CF46915F7ULL, "gencr_gate_slave_l4.utc"},
	{0x4C2E38EB6F02ACB1ULL, "vxht_keep_n_6an_0_0.vlm"},
	{0x4C2F8E2C467A4C33ULL, "blood_clu_l3.mao"},
	{0x4C2FB6428B48173DULL, "6114014_m.fxe"},
	{0x4C312EB2CD2C942DULL, "vxlt_darktown_4mh_0_0.vlm"},
	{0x4C31E237CFCAFFB2ULL, "283387850.ogg"},
	{0x4C3232079BEFF379ULL, "178996690.ogg"},
	{0x4C32E690D219D1FDULL, "vxlt_hangedman_n_250_0_0.vlm"},
	{0x4C3322FC0AC17C01ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\hf.anb"},
	{0x4C3323D6FF09D576ULL, "6016698_m.fxe"},
	{0x4C3334B12695CC60ULL, "vxht_brothel_n_34j_0_0.vlm"},
	{0x4C335F82FC796C7CULL, "lt_warehouse_5_0_lc.mmh"},
	{0x4C33BB575AB4FE14ULL, "6208424_m.fxe"},
	{0x4C33C7A0767961E1ULL, "747049248.ogg"},
	{0x4C3429F8891E278DULL, "vxw_woundedcoast_a7_0_0.vlm"},
	{0x4C34E3F9301D3BEBULL, "vxw_encounter3_93_0_0.vlm"},
	{0x4C3511744A96A1A4ULL, "vxdr_0_deeproads_ebq_0_0.vlm"},
	{0x4C35428FF8B51D8DULL, "vis121_saemus6100104.cl"},
	{0x4C35AE7210874477ULL, "hlac_lt_darktown_-1_52x.dds"},
	{0x4C35B93C81582E7BULL, "qcr220_meredith6041533.cl"},
	{0x4C35D04EE4086780ULL, "addparty.ncs"},
	{0x4C3615A75BEF504DULL, "591308094.ogg"},
	{0x4C364890E435BC22ULL, "6000730_m.fxe"},
	{0x4C36EBE7933D7805ULL, "plt_isa000pt_midstage.plo"},
	{0x4C373066515A559EULL, "gen00fl_aveline6106976.cl"},
	{0x4C3761D56CC50ABBULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_dq101a_0s.dds"},
	{0x4C377167EB115ABEULL, "vxw_bonepit_178_0_0.vlm"},
	{0x4C379927628694C0ULL, "6076966_m.fxe"},
	{0x4C38B49215380459ULL, "vxlt_gallowscourtyard_n_514_0_0.vlm"},
	{0x4C38C25F2E37D878ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\prp_barkburn_sa.dds"},
	{0x4C38EDB5DD287A9CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv2.ani"},
	{0x4C39FB0F3F56770AULL, "art\\characters\\playercharacter\\shared\\textures\\bdy_bnb_0s.dds"},
	{0x4C3A12747D326818ULL, "hightown_male_cityguard_mob6170040.cl"},
	{0x4C3A13A472A02F2AULL, "follower_banter6161023.cl"},
	{0x4C3AB4133B7B0FE4ULL, "vxlt_undercityencounter_1kh_0_0.vlm"},
	{0x4C3AFF9CA961A0D6ULL, "414399020.ogg"},
	{0x4C3B763B7B40BBD1ULL, "dae000_merchant_generic_hm6172668.cl"},
	{0x4C3BF49D39DB21D9ULL, "hlac_lt_0_lowtown_n_1_8x.dds"},
	{0x4C3C3FC7D4EC3F65ULL, "6141711_m.fxe"},
	{0x4C3D1C9F04BAFC26ULL, "180169535.ogg"},
	{0x4C3D5D134699C050ULL, "vxca_caverns1_icx_0_1.vlm"},
	{0x4C3D61DE94488ACCULL, "gen00fl_fenris6107176.cl"},
	{0x4C3E9D5D916858BDULL, "llac_ht_mansion2b_818_1.dds"},
	{0x4C3F2570FCCB5763ULL, "vxlt_gallowstemplar_f_518_0_0.vlm"},
	{0x4C3F3F74E08E142AULL, "follower_barks6153300.cl"},
	{0x4C3FA08FBF22ACADULL, "vxlt_darktown_4hg_0_0.vlm"},
	{0x4C3FFAB5DD2EAFDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv2.gad"},
	{0x4C4020C587CD39D0ULL, "gen_im_fla_acc_isabela_01.uti"},
	{0x4C40E9F414BEE4A2ULL, "6202166_f.fxe"},
	{0x4C4162639F878787ULL, "zz_isa_debug6133974.cl"},
	{0x4C419447CB11F5C9ULL, "vxlt_0_lowtown_n_f12_0_0.vlm"},
	{0x4C41A68E188BE4E0ULL, "vxht_brothel_n_33k_0_0.vlm"},
	{0x4C41BF5521526BE5ULL, "267333913.ogg"},
	{0x4C41DDFD7EA23FB7ULL, "one200it_heart.uti"},
	{0x4C421570B26BF1C9ULL, "alphacrack01.mao"},
	{0x4C429983B9C93A76ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_en.ani"},
	{0x4C429A80B87A6E9FULL, "6177330_m.fxe"},
	{0x4C42E4B0658AF42BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_floor_sa.dds"},
	{0x4C433C53D5A1941FULL, "gen00fl_fenris6113958.cl"},
	{0x4C43D506FBD7FBF4ULL, "6113093_f.fxe"},
	{0x4C43F6AC7AEC9FB9ULL, "dae300_vashothlead6062833.cl"},
	{0x4C4406AD33E31E2BULL, "character_character_p42.psh.fxo.dx11.assembly"},
	{0x4C44F5594D812E19ULL, "gen00fl_aveline6139463.cl"},
	{0x4C44FA1CA47C6809ULL, "vxht_0_hightown_a3_c8j_0_0.vlm"},
	{0x4C45C5A3DE29C079ULL, "6166648_m.fxe"},
	{0x4C45E16CB45D7F24ULL, "ser111_fleeing_mage.cut"},
	{0x4C461445B0AB94D3ULL, "451803138.ogg"},
	{0x4C461F7711A4730AULL, "vxht_0_hightown_a3_a6e_0_0.vlm"},
	{0x4C4655150706833BULL, "hlac_ht_0_hightown_a2_-65536_0z.dds"},
	{0x4C465924CDC3592FULL, "hlac_lt_0_lowtown_1_5x.dds"},
	{0x4C468C5980E07976ULL, "334321383.ogg"},
	{0x4C46D5B5DD348FFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv2.evt"},
	{0x4C472DCD498841A3ULL, "mcr310_argument6127591.cl"},
	{0x4C476DFD78BA20EAULL, "fxa_pois_dirty.mao"},
	{0x4C47839644F1C248ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chab.mmh"},
	{0x4C47A19644F1F572ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chab.msh"},
	{0x4C47DB1BD3D8BA74ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_meda_0t.dds"},
	{0x4C48EEEEF317ABAAULL, "vxca_caverns2_56z_0_1.vlm"},
	{0x4C4902F905266315ULL, "vxlt_0_lowtown_docks_ctn_0_0.vlm"},
	{0x4C4924D3CFE81875ULL, "hlac_ht_mansion2_818_20y.dds"},
	{0x4C492A32B6ADE260ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_post_1.mao"},
	{0x4C4A0EC5520F73D8ULL, "358020121.ogg"},
	{0x4C4A1E4F11206280ULL, "vxlt_gallowscourtyard_6xn_0_0.vlm"},
	{0x4C4A2C22FEDF4584ULL, "vxht_0_hightown_a2_bxs_0_0.vlm"},
	{0x4C4A386C501D1E9DULL, "and100_anders6095966.cl"},
	{0x4C4A5851F7605F94ULL, "vxw_encounter2_3a_0_0.vlm"},
	{0x4C4AAEFC9E8A309EULL, "6082274_f.fxe"},
	{0x4C4AEC2992F3EF4AULL, "443994623.ogg"},
	{0x4C4B140C36B10818ULL, "qcr000_arishok6202345.cl"},
	{0x4C4B9DF2D7729BCBULL, "dae000_gamlen6048437.cl"},
	{0x4C4BDCCB48445F00ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pew_n.dds"},
	{0x4C4C314C022C5BF1ULL, "art\\vfx\\materials\\chunky\\vfx_shield.mao"},
	{0x4C4C9CB6C040F18EULL, "qun160_vashothlead6064674.cl"},
	{0x4C4CA8088D851204ULL, "vxlt_gallowstemplar_2jb_0_0.vlm"},
	{0x4C4CFF3280CC6F74ULL, "hlac_w_blightlands_393213_0x.dds"},
	{0x4C4D629198A9AD3CULL, "vxlt_0_lowtown_docks_a3_n_cy8_0_0.vlm"},
	{0x4C4D7B3DC01604EAULL, "lt_0_lowtown_docks_n_17_0.anb"},
	{0x4C4D8B74750C3D95ULL, "t3_eey_bl2.tnt"},
	{0x4C4D9B840143FE7FULL, "llac_ht_chantry_2534_53.dds"},
	{0x4C4EB80551F9FA69ULL, "clf260_harley6069948.cl"},
	{0x4C4F0C8CABC9D116ULL, "6126796_f.fxe"},
	{0x4C4F4220F0B65637ULL, "pro000_athenril6049474.cl"},
	{0x4C4F8796190B49FDULL, "l_one300ar_vanard_fight.wbd"},
	{0x4C4FA183B9D3D267ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_en.evt"},
	{0x4C4FBDC1589360E1ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar2_mcr370_playerflees.ani"},
	{0x4C4FD04258F476A8ULL, "991212491.ogg"},
	{0x4C50049D0EFAE0E1ULL, "420150690.ogg"},
	{0x4C506BC7DD45FD27ULL, "6208316_f.fxe"},
	{0x4C50E40A069DDA0FULL, "dae000_arishok6078174.cl"},
	{0x4C50E54AFA3A259DULL, "06_smokeglow.mao"},
	{0x4C5147A4F5931FB9ULL, "ser120cr_market_guard_talk.utc"},
	{0x4C514E80657B49C9ULL, "vxca_caverns2prc_9sn_0_0.vlm"},
	{0x4C51AEA136D648F5ULL, "l_dae221ar_brothel_nt.wbd"},
	{0x4C520CAE3A3F83EBULL, "hlac_dr_slavers_3216_0y.dds"},
	{0x4C5213F8CF0AF48FULL, "236733315.ogg"},
	{0x4C52746FAB6C3F85ULL, "vxca_cavern1_haunted_j3z_0_0.vlm"},
	{0x4C527E30D9364627ULL, "6094565_m.fxe"},
	{0x4C53088F8BE69321ULL, "zz_approval_debug6003752.cl"},
	{0x4C53C32A6BC612D8ULL, "125773579.ogg"},
	{0x4C5526FB5DB05056ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_fireplace.phy"},
	{0x4C5536BE985729F1ULL, "847594509.ogg"},
	{0x4C5539E5808BBC29ULL, "298475394.ogg"},
	{0x4C55F3AA863435E5ULL, "vxht_keep_n_586_0_0.vlm"},
	{0x4C566EC292B6901CULL, "gen00fl_fenris6106650.cl"},
	{0x4C56C283B9DA28B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_en.gad"},
	{0x4C56C8C158999183ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar2_mcr370_playerflees.gad"},
	{0x4C57774627DE4A1AULL, "874295115.ogg"},
	{0x4C57CBAF5807E0EEULL, "ht_0_hightown_n_26_0_lc.mmh"},
	{0x4C590FFE4D5A4F07ULL, "vxlt_gallowstemplar_37c_0_0.vlm"},
	{0x4C591E7B86F0CBD0ULL, "dae000_mother6127176.cl"},
	{0x4C592DE1DB426A33ULL, "vxca_caverns2_6h7_0_0.vlm"},
	{0x4C5AB9F807A58052ULL, "6216630_m.fxe"},
	{0x4C5B19E57F7B2A84ULL, "vxlt_gallowstemplar_2ec_0_0.vlm"},
	{0x4C5B1EFDB76464B1ULL, "fxa_pois_split_b.mao"},
	{0x4C5B9B442EE54502ULL, "6140169_f.fxe"},
	{0x4C5BAD0EB5D4965CULL, "243877739.ogg"},
	{0x4C5C01FF36F247B9ULL, "29675419.ogg"},
	{0x4C5C2187D937A377ULL, "850414126.ogg"},
	{0x4C5C21A7E8BE83F9ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.impa.ani"},
	{0x4C5D37BA093B4D9EULL, "6076414_m.fxe"},
	{0x4C5D835A682A9CB6ULL, "839825384.ogg"},
	{0x4C5DF2F56C5D7B32ULL, "6049096_m.fxe"},
	{0x4C5E058F9A5A3A49ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tableholy.phy"},
	{0x4C5E840E9AF15F1AULL, "1171413.ogg"},
	{0x4C5E90299C3F735EULL, "narrative_banter6165542.cl"},
	{0x4C5F107E2EAC94D2ULL, "hlac_ht_keep_3732_19z.dds"},
	{0x4C5F2E513F70B5B9ULL, "fen210_hadriana6110404.cl"},
	{0x4C5FF7BAD70EDAE9ULL, "534541900.ogg"},
	{0x4C600FFB5DBA30F7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_fireplace.msh"},
	{0x4C6021FB5DBA4F5AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_fireplace.mao"},
	{0x4C6025FB5DBA5669ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_fireplace.mmh"},
	{0x4C6094581DAF9477ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wraptable.phy"},
	{0x4C60C698DDA80A69ULL, "6166455_m.fxe"},
	{0x4C6163069A8F5976ULL, "vxlt_gallowstemplar_n_4uh_0_0.vlm"},
	{0x4C619C5246AB9F10ULL, "ht_0_hightown_53_0.rml"},
	{0x4C619D84FB7F19FFULL, "hlac_lt_blackemporium_1055_13y.dds"},
	{0x4C61DDD688689909ULL, "vxht_0_hightown_a3_c6h_0_0.vlm"},
	{0x4C61E3D6EE6EDC76ULL, "597588592.ogg"},
	{0x4C62FCA7E8C4633BULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.impa.gad"},
	{0x4C636A9C5C2EFCF8ULL, "6189422_m.fxe"},
	{0x4C6478468F6CE502ULL, "vxht_0_hightown_aui_0_0.vlm"},
	{0x4C651DC93A089759ULL, "one220_yevhen6103386.cl"},
	{0x4C65274FA8D50947ULL, "hlac_lt_0_lowtown_docks_a3_65534_6y.dds"},
	{0x4C654AA7A4AEE8C8ULL, "dae000_meredith6092547.cl"},
	{0x4C656704521D973BULL, "6117018_f.fxe"},
	{0x4C658120309C6B17ULL, "85579586.ogg"},
	{0x4C65817585F4EFC3ULL, "6118636_m.fxe"},
	{0x4C658CA8D892A4BEULL, "vxlt_backalley_g_1ny_0_0.vlm"},
	{0x4C65FED6F2DBC71CULL, "335319981.ogg"},
	{0x4C6600DF2DC62874ULL, "6174480_m.fxe"},
	{0x4C66113F3999FD81ULL, "6096301_m.fxe"},
	{0x4C664F23E0905B47ULL, "6167656_m.fxe"},
	{0x4C66EA4FE1B17B38ULL, "qcr000_arishok6166984.cl"},
	{0x4C66EA6A49B5473FULL, "mag170_thrask6054054.cl"},
	{0x4C66EAC15C3646EEULL, "hlac_lt_backalley_g_0_14y.dds"},
	{0x4C670FDCE2CB7684ULL, "vxht_0_hightown_a2_bfq_0_0.vlm"},
	{0x4C671F8887073BC7ULL, "607685803.ogg"},
	{0x4C673C0BDB4C9094ULL, "vxw_encounter2_9g_0_0.vlm"},
	{0x4C67A344E34E3916ULL, "vxlt_darktown_58a_0_0.vlm"},
	{0x4C6808AF0980DC17ULL, "70594550.ogg"},
	{0x4C684E4EC556BC92ULL, "mcr371_orsino6083308.cl"},
	{0x4C684EEB50EC60DBULL, "vxlt_foundry_n_zi_0_0.vlm"},
	{0x4C686F6B1BDCA042ULL, "156272470.ogg"},
	{0x4C6872A0C570103CULL, "886574082.ogg"},
	{0x4C68ACB6FE83B54CULL, "vxca_caverns2_7oa_0_1.vlm"},
	{0x4C68B2AA1C7C09D0ULL, "vxca_caverns1_ile_0_1.vlm"},
	{0x4C6900603367CE76ULL, "6017815_m.fxe"},
	{0x4C693EBC469382B9ULL, "vxht_mansion_n_4fa_0_0.vlm"},
	{0x4C6949223DC604B9ULL, "mcr371_scared_mage_m6087690.cl"},
	{0x4C697042D0F5EF61ULL, "mag321_comtesse_de_launcet6043365.cl"},
	{0x4C69CB58B68BE905ULL, "vxlt_darktown_6j7_0_0.vlm"},
	{0x4C6A136217FDF906ULL, "vxlt_warehouse_n_bf_0_0.vlm"},
	{0x4C6A19A7E8CAB320ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.impa.evt"},
	{0x4C6A21E76CABE5F8ULL, "llac_ht_0_hightown_a2_1_2.dds"},
	{0x4C6A7E841D5939FEULL, "vxw_encounter2_pc_0_1.vlm"},
	{0x4C6B1B31D5A86918ULL, "vxw_bonepitmining_a2_1pe_0_0.vlm"},
	{0x4C6B408B18F77B7DULL, "vxht_bartrand_n_4l4_0_0.vlm"},
	{0x4C6BEF40008A7873ULL, "536399727.ogg"},
	{0x4C6C336DD3FF398FULL, "mer110_jansen6070217.cl"},
	{0x4C6C5E186D293E1EULL, "lt_0_lowtown_45_0_lc.mmh"},
	{0x4C6C6B4B71610805ULL, "vxca_cavern1_haunted_gf0_0_0.vlm"},
	{0x4C6C762A8EF66922ULL, "nat380_sibling6068038.cl"},
	{0x4C6C8CE2F1CA5FE8ULL, "vxlt_gallowscourtyard_7et_0_0.vlm"},
	{0x4C6C9CA45D3B91C0ULL, "441937075.ogg"},
	{0x4C6D7AF486242CB9ULL, "seb221cr_elf_servant.utc"},
	{0x4C6DFEBA2B3966B7ULL, "574745131.ogg"},
	{0x4C6FAB66FC927BD4ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualready.ani"},
	{0x4C701A354076345BULL, "6050334_m.fxe"},
	{0x4C705C38935322CDULL, "mrl000_marethari6129349.cl"},
	{0x4C71164041E98A8BULL, "gen_im_trs_gen_med_cht_r0.uti"},
	{0x4C718DCB52D5AAEFULL, "6108848_m.fxe"},
	{0x4C71AEF6B89452F4ULL, "6076312_f.fxe"},
	{0x4C71F47158C4BC1CULL, "vxdr_0_deeproads_cmv_0_0.vlm"},
	{0x4C722E41F9556BD4ULL, "vxlt_0_lowtown_a2_byl_0_0.vlm"},
	{0x4C7286FFACD4FDD1ULL, "455447059.ogg"},
	{0x4C732B9A4D2C1B23ULL, "hlac_lt_0_lowtown_docks_a3_-65538_3y.dds"},
	{0x4C735694DD56B62BULL, "6205590_m.fxe"},
	{0x4C73909B8A653C8DULL, "vrc321_party_barks6095764.cl"},
	{0x4C73E5C1F2F01F42ULL, "vxht_mansion2_2bw_0_0.vlm"},
	{0x4C73FE85CA88BD3EULL, "vxlt_backalley_g_1kz_0_0.vlm"},
	{0x4C74380D6B820934ULL, "444488400.ogg"},
	{0x4C7457731BF8C56BULL, "6125194_f.fxe"},
	{0x4C74F033E6EC05BCULL, "follower_banter6163975.cl"},
	{0x4C750DE05E3CA7FAULL, "228993287.ogg"},
	{0x4C75A67B29966F1DULL, "hlac_lt_0_lowtown_a2_n_c_-196609_0y.dds"},
	{0x4C7674BDF51AC78CULL, "6135638_m.fxe"},
	{0x4C76A866FC989536ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualready.gad"},
	{0x4C76CCA759B4CA43ULL, "zz_mrl_debug6184189.cl"},
	{0x4C76F96DC0789A7BULL, "447977140.ogg"},
	{0x4C7700FE2EB90464ULL, "57492193.ogg"},
	{0x4C77169ED1F699D7ULL, "lyriumghost_regainstamina.dds"},
	{0x4C7792287630A624ULL, "t3_wer_h00.tnt"},
	{0x4C7798C26FF5AD77ULL, "t3_dey_am1.tnt"},
	{0x4C78CF35C1207F42ULL, "gen00fl_merrill6105730.cl"},
	{0x4C78FF077382CDD0ULL, "mcr371_angry_mage_m6087032.cl"},
	{0x4C797B82E3F8F536ULL, "hlac_lt_0_lowtown_docks_a3_-65539_0z.dds"},
	{0x4C79FD7F2F1F44B2ULL, "vxw_woundedcoast_16h_0_0.vlm"},
	{0x4C7AC5F27A819EEBULL, "6083330_m.fxe"},
	{0x4C7B3FA0A22AB07CULL, "592722711.ogg"},
	{0x4C7B72C15001B000ULL, "vxlt_0_lowtown_n_h1u_0_0.vlm"},
	{0x4C7B8CC51EDA0B7EULL, "vxlt_warehouse_n_qw_0_0.vlm"},
	{0x4C7B9CB3F86C1D53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_exit.ani"},
	{0x4C7C076514BDF152ULL, "gen_im_wep_war_shd_act2_06.uti"},
	{0x4C7C602DAA656C02ULL, "hlac_lt_undercitydungeon_-1_32z.dds"},
	{0x4C7C87C472B3BCC2ULL, "660958974.ogg"},
	{0x4C7CAAE7FA30D4E7ULL, "vxht_0_hightown_n_c7k_0_0.vlm"},
	{0x4C7CC366FC9D2E95ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualready.evt"},
	{0x4C7CC55B281C071CULL, "vxht_chantry_2ip_0_0.vlm"},
	{0x4C7CFF943A815670ULL, "vxdr_primevaltomb_3i0_0_0.vlm"},
	{0x4C7D00818E2C4144ULL, "743627870.ogg"},
	{0x4C7DAAF6EB58BBDAULL, "l_env_wal_gallows_military_day.bnk"},
	{0x4C7DCBE49CE6D80EULL, "art\\characters\\creatures\\mubari\\animation\\charge.ani"},
	{0x4C7DF46FCE3A86F0ULL, "59243007.ogg"},
	{0x4C7E0F270993E4C7ULL, "6126857_m.fxe"},
	{0x4C7EA399EA1FE32AULL, "6104279_f.fxe"},
	{0x4C7EEC27CD79AA7FULL, "lowtown_male_alienage1_sta6162762.cl"},
	{0x4C7F0D1B46646947ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowa_d.dds"},
	{0x4C7FE31A41BDBD17ULL, "6115406_m.fxe"},
	{0x4C80A28D7D444E01ULL, "ss_sebastian.bnk"},
	{0x4C82579EE4E637C2ULL, "vxht_mansion2_2op_0_0.vlm"},
	{0x4C828637043CC3C7ULL, "qun160_vashothbandits6069284.cl"},
	{0x4C82998490AB2BC1ULL, "hm_clf_fell_orden.mor"},
	{0x4C82AB02B0694772ULL, "gen00fl_varric6105752.cl"},
	{0x4C82C5D8F1751BDAULL, "814954735.ogg"},
	{0x4C8399FF69AC5A69ULL, "vxlt_0_lowtown_a2_eoy_0_1.vlm"},
	{0x4C83AD8D7D469686ULL, "ss_sebastian.cut"},
	{0x4C83F396C6B7A784ULL, "vxht_0_hightown_a2_blw_0_0.vlm"},
	{0x4C844BC2EFFABC5CULL, "vxlt_0_lowtown_a2_cmm_0_0.vlm"},
	{0x4C8467CEEBC06E0CULL, "vxht_mansion_n_4bu_0_1.vlm"},
	{0x4C84DE1D52151D61ULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.anb"},
	{0x4C8513285AE1B12EULL, "vxht_0_hightown_a2_90b_0_0.vlm"},
	{0x4C852A4762F70539ULL, "514593417.ogg"},
	{0x4C85381B0CF888DAULL, "dae_mage_aoe_instant.ncs"},
	{0x4C8585BA8E4EE2D5ULL, "755268479.ogg"},
	{0x4C85C06EC84BD0BDULL, "6144389_m.fxe"},
	{0x4C860C3BFD805336ULL, "gen_im_trs_slv_sta_cld_r0.uti"},
	{0x4C8620EE04DE638BULL, "353278258.ogg"},
	{0x4C8675C311B33B11ULL, "vxlt_0_lowtown_docks_b94_0_0.vlm"},
	{0x4C86B83BBE11D9E8ULL, "vxlt_0_lowtown_docks_bco_0_0.vlm"},
	{0x4C86C0A7E76662D5ULL, "vxht_chantry_n_2zg_0_0.vlm"},
	{0x4C874D27A6CB339DULL, "643783611.ogg"},
	{0x4C87B24318F98099ULL, "vxht_mansion2b_3cz_0_0.vlm"},
	{0x4C8802F1FC70EDC5ULL, "mag360_samson6062115.cl"},
	{0x4C8894B3F876998AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_exit.evt"},
	{0x4C8947D235767389ULL, "dae000_templar_recruit_16164443.cl"},
	{0x4C8966D47EEEBFEEULL, "vxlt_gallowstemplar_4qo_0_0.vlm"},
	{0x4C89DFED59C4E8EAULL, "317543076.ogg"},
	{0x4C89ECE7313B3D40ULL, "548860256.ogg"},
	{0x4C8A14EC5DBFDACBULL, "seb201_lady_harimann6177416.cl"},
	{0x4C8AA28C4A169563ULL, "vxlt_0_lowtown_docks_n_bdf_0_0.vlm"},
	{0x4C8AC3E49CF1555FULL, "art\\characters\\creatures\\mubari\\animation\\charge.evt"},
	{0x4C8AEE8F9A81016EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tableholy.mmh"},
	{0x4C8AFA8F9A811505ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tableholy.mao"},
	{0x4C8B088F9A812D5CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tableholy.msh"},
	{0x4C8B0CC56198F15CULL, "dae000ip_crowd_ff5_talk13.utp"},
	{0x4C8B2A54BA4AFBB3ULL, "campaign_base_pl-pl.cre"},
	{0x4C8B480EEAF6057AULL, "gen00fl_fenris6105842.cl"},
	{0x4C8B50CAFA5BF5F1ULL, "6207550_m.fxe"},
	{0x4C8BD25608EC1271ULL, "vxht_keep_372_0_0.vlm"},
	{0x4C8C0B6ADC1CB2CAULL, "al_warrior_t02.gda"},
	{0x4C8C34E6BBFA4A09ULL, "vxdr_primevaltomb_9e6_0_1.vlm"},
	{0x4C8C60260552C482ULL, "character_character_p102.psh.fxo"},
	{0x4C8CDAFBAD6C87B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sweep.ani"},
	{0x4C8D0BF8D6356CE9ULL, "1053296759.ogg"},
	{0x4C8D0ED08A1D0068ULL, "380104628.ogg"},
	{0x4C8DF5EB24E8489DULL, "mrl150_merrill6070986.cl"},
	{0x4C8E9D5246D28F78ULL, "ht_0_hightown_53_0.anb"},
	{0x4C8ED82B3CB0ED1CULL, "vxdr_0_deeproads_cgt_0_0.vlm"},
	{0x4C8F0F52713AD1E0ULL, "mag100_cullen6171210.cl"},
	{0x4C8F73B3F87C80FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_exit.gad"},
	{0x4C90C97BD6DC5042ULL, "vxht_mansion2_2hq_0_0.vlm"},
	{0x4C90ECBE6220285FULL, "vxlt_0_lowtown_a2_gz3_0_0.vlm"},
	{0x4C91197025EE583EULL, "272149704.ogg"},
	{0x4C91572FB8C2EB83ULL, "6151248_m.fxe"},
	{0x4C91A4E49CF73F48ULL, "art\\characters\\creatures\\mubari\\animation\\charge.gad"},
	{0x4C91B94A6DBEA545ULL, "vxht_mansion2b_321_0_0.vlm"},
	{0x4C91BF1D521F72DAULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.mmh"},
	{0x4C928FCF6D4FB8CEULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_templarbadge_n.dds"},
	{0x4C92F6D12FEAA80FULL, "vxlt_gallowstemplar_3sw_0_0.vlm"},
	{0x4C938FB39F92C94DULL, "vrc221_cassandra6040882.cl"},
	{0x4C93E5FBAD72B953ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sweep.gad"},
	{0x4C93E8D51CDB987FULL, "plt_cod_prc_velvet_cailan.plo"},
	{0x4C940701B464493DULL, "6093746_m.fxe"},
	{0x4C94767F6A4C4291ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_pc_mcr370_playerrules.gad"},
	{0x4C94A058E8DB2A8CULL, "gen_im_trs_dal_two_2hs_r2.uti"},
	{0x4C94CAFE41A25B11ULL, "118259191.ogg"},
	{0x4C952A18B007F268ULL, "vxlt_0_lowtown_docks_bfl_0_0.vlm"},
	{0x4C9568F8EDE04286ULL, "vxlt_warehouse_n_si_0_0.vlm"},
	{0x4C959E108143A5B9ULL, "fen110_fenris6117953.cl"},
	{0x4C95F1795DEC8466ULL, "33114035.ogg"},
	{0x4C960F1391C95835ULL, "gen00fl_varric6033698.cl"},
	{0x4C9624200AEF9919ULL, "vxht_mansion2b_3fy_0_0.vlm"},
	{0x4C96A54D6628DC49ULL, "6047404_m.fxe"},
	{0x4C96D6B13C747E2AULL, "gen00fl_merrill6116060.cl"},
	{0x4C96DF42F4E0D9D5ULL, "follower_banter6164739.cl"},
	{0x4C96E4EEB9FE69D5ULL, "lgt110cr_orz_harrowmont.utc"},
	{0x4C971259A834620BULL, "zz_dog_debug6127930.cl"},
	{0x4C9795AB9A94E5F1ULL, "hlac_ca_cave_2194_4z.dds"},
	{0x4C97FE71C267B959ULL, "502155964.ogg"},
	{0x4C984D8F215A4AAEULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_barge.mmh"},
	{0x4C98598F215A5F45ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_barge.mao"},
	{0x4C98678F215A769CULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_barge.msh"},
	{0x4C98F8B80A4789F6ULL, "qcr221_cut_arishoks_address.cnv"},
	{0x4C9914693C0CADE3ULL, "vxlt_0_lowtown_docks_n_bai_0_0.vlm"},
	{0x4C993D71AB4AFDA2ULL, "hlac_w_encounter2_262144_0y.dds"},
	{0x4C995863A0C6272AULL, "follower_banter6170146.cl"},
	{0x4C995FA3E79B78DEULL, "vxdr_0_deeproads_eta_0_0.vlm"},
	{0x4C998671C4D8FD29ULL, "17432037.ogg"},
	{0x4C99E2FBAD771FB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sweep.evt"},
	{0x4C99F97CD4F6DCC9ULL, "hlac_lt_0_lowtown_a2_0_3y.dds"},
	{0x4C9A643E287A5413ULL, "hlac_ca_caverns1_24426_3y.dds"},
	{0x4C9A88D95757DC5CULL, "518345664.ogg"},
	{0x4C9AE9525549F94AULL, "lgt210_magu_selby6142037.cl"},
	{0x4C9B2A5B5E7B6186ULL, "6115318_f.fxe"},
	{0x4C9B5F72E7274779ULL, "767952069.ogg"},
	{0x4C9B777F6A52633FULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_pc_mcr370_playerrules.ani"},
	{0x4C9BBBB71D6B59F0ULL, "hlac_w_encounter2_5_0y.dds"},
	{0x4C9BF38C345FE134ULL, "232860739.ogg"},
	{0x4C9BFD3A73E0F63DULL, "vxlt_gallowstemplar_n_34k_0_0.vlm"},
	{0x4C9CC05E6DDC877BULL, "hairalpha_hair_p2.psh.fxo"},
	{0x4C9CF62B31ABC117ULL, "hlac_ht_playershouse_n_855_20x.dds"},
	{0x4C9CFAB80A4B75D6ULL, "qcr221_cut_arishoks_address.bnk"},
	{0x4C9D1E3363F9937EULL, "vxlt_darktown_4wa_0_0.vlm"},
	{0x4C9D231CECA1B097ULL, "vxht_mansion2b_1cj_0_0.vlm"},
	{0x4C9D465D6DF75741ULL, "635619546.ogg"},
	{0x4C9D59194799855CULL, "zz_vault_debug6200346.cl"},
	{0x4C9DAD5A5B9F21C3ULL, "w_blightlands_63_0.mao"},
	{0x4C9DB15A5B9F2810ULL, "w_blightlands_63_0.mmh"},
	{0x4C9DBF5A5B9F402AULL, "w_blightlands_63_0.msh"},
	{0x4C9DFA114BA56F71ULL, "vxht_mansion_n_4xg_0_0.vlm"},
	{0x4C9E2AB8BD51DF00ULL, "gen00fl_fenris6107231.cl"},
	{0x4C9ECD697882FD07ULL, "vxht_0_hightown_a2_a6i_0_0.vlm"},
	{0x4C9EFDADF0470B8AULL, "6053939_f.fxe"},
	{0x4C9F359E96A90A59ULL, "mcr310_argument6075078.cl"},
	{0x4C9F4213DF9F28B2ULL, "462771766.ogg"},
	{0x4C9FBBF39168B072ULL, "6108632_m.fxe"},
	{0x4CA008E964909ABCULL, "163747541.ogg"},
	{0x4CA067791B000590ULL, "ttx_roof_int_a_1.msh"},
	{0x4CA075791B001DAAULL, "ttx_roof_int_a_1.mmh"},
	{0x4CA0B122822090EAULL, "mag311_keran6041489.cl"},
	{0x4CA0D66161EA9557ULL, "gen_im_arm_glv_hvy_set_med.uti"},
	{0x4CA12104C2C1CA26ULL, "547130844.ogg"},
	{0x4CA1425D421076DCULL, "mag200_feynriel6034450.cl"},
	{0x4CA18960F77FE32DULL, "hlac_lt_shop_-1_3z.dds"},
	{0x4CA193D0B8FB3CB7ULL, "gen00fl_anders6119433.cl"},
	{0x4CA19A4EC28DD171ULL, "vxlt_gallowsprison_n_6me_0_0.vlm"},
	{0x4CA20DC0B8FB16FFULL, "577728612.ogg"},
	{0x4CA2EC5C4A1C150EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01a.phy"},
	{0x4CA326624AA56D54ULL, "hlac_lt_0_lowtown_a2_n_65537_1y.dds"},
	{0x4CA34CE97B68C8C0ULL, "ser221_gascard6111754.cl"},
	{0x4CA3775211EA0AB0ULL, "vxdr_primevaltomb_4hp_0_0.vlm"},
	{0x4CA3AF03A7AA13DAULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm03.phy"},
	{0x4CA3C4F6EB79B778ULL, "l_env_wal_gallows_military_day.wbd"},
	{0x4CA3DAD5DFD65B06ULL, "vxlt_warehouse_n_vj_0_0.vlm"},
	{0x4CA3F4FDFA82392CULL, "915197578.ogg"},
	{0x4CA4E1FCBB81DB65ULL, "812205687.ogg"},
	{0x4CA56F487D08F952ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_act_c.mmh"},
	{0x4CA5DA22D6675D8AULL, "ser100_party_comments6108239.cl"},
	{0x4CA5DC9F295BF0D3ULL, "vxdr_primevaltomb_6x5_0_0.vlm"},
	{0x4CA61F8CC34254EFULL, "62110025.ogg"},
	{0x4CA681F2F7E168A7ULL, "6108295_m.fxe"},
	{0x4CA6E20CF64797B4ULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.mmh"},
	{0x4CA6E425F2F260D5ULL, "llac_w_sundermount_-196607_0.dds"},
	{0x4CA7D2D7DF08AB91ULL, "524154790.ogg"},
	{0x4CA7FE8B7ACE7454ULL, "6217147_m.fxe"},
	{0x4CA869581DED00CFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wraptable.mao"},
	{0x4CA875581DED156CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wraptable.mmh"},
	{0x4CA87B581DED1F5EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wraptable.msh"},
	{0x4CA88AD028FB7577ULL, "936165124.ogg"},
	{0x4CA8F9F6BFD95AEEULL, "vxlt_undercityserialkiller_1c7_0_1.vlm"},
	{0x4CA90413345AABA7ULL, "364309488.ogg"},
	{0x4CA929817589751AULL, "qcr220_grey_wardens6207442.cl"},
	{0x4CA993D78442CCF0ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg04.mmh"},
	{0x4CA9A1D78442E40AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg04.msh"},
	{0x4CAA064F33F948D1ULL, "hlac_lt_0_lowtown_docks_a3_196607_3z.dds"},
	{0x4CAA63F039BEECE8ULL, "vxht_keep_n_4np_0_0.vlm"},
	{0x4CAA677ED8029AC9ULL, "llac_lt_shop_-1_5.dds"},
	{0x4CAA6F1765D70EBDULL, "vxlt_gallowstemplar_n_33j_0_0.vlm"},
	{0x4CAAA64A156D75EBULL, "ser270_emeric6113737.cl"},
	{0x4CAC01BC553A8492ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_3y.dds"},
	{0x4CAC234D06FBC689ULL, "6096213_m.fxe"},
	{0x4CAC249E61461F33ULL, "zz_approval_debug6003776.cl"},
	{0x4CAC4B711AB95984ULL, "vxlt_0_lowtown_a2_hgt_0_0.vlm"},
	{0x4CACD1C4A80C3D61ULL, "6039153_m.fxe"},
	{0x4CAD6FD758C5B57CULL, "hlac_ht_keep_3734_0z.dds"},
	{0x4CADAE9EAD7BB712ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirewall.dura.gad"},
	{0x4CAE447DC5D962A1ULL, "qun110cr_petrice.utc"},
	{0x4CAEC608731659F3ULL, "50158260.ogg"},
	{0x4CAF38E348F43B65ULL, "ser221_gamlen6123423.cl"},
	{0x4CAF5D54EAB08F52ULL, "vxdr_primevaltomb_6x0_0_0.vlm"},
	{0x4CAF5E0C9D78599EULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.cess.ani"},
	{0x4CAF65E82FBE474BULL, "6161650_m.fxe"},
	{0x4CAF9B5E1E13EFAEULL, "gen00fl_isabela6075284.cl"},
	{0x4CB039AFF0A98931ULL, "vxht_keep_n_64e_0_0.vlm"},
	{0x4CB072B0C7142668ULL, "vxw_bonepitdestroyed_a3_1zy_0_0.vlm"},
	{0x4CB0FECC96880526ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.anb"},
	{0x4CB11128CDE6D1AEULL, "vxca_cavern1_haunted_gkh_0_0.vlm"},
	{0x4CB112D6197A80B3ULL, "gen00fl_fenris6127009.cl"},
	{0x4CB19A4A75355FEAULL, "6174752_m.fxe"},
	{0x4CB246684569E817ULL, "614936255.ogg"},
	{0x4CB295026A39CAF3ULL, "bink_bink_p0.vsh.fxo.dx11.assembly"},
	{0x4CB2EE0EAB9A6FD5ULL, "vxca_cavern1_haunted_efr_0_0.vlm"},
	{0x4CB3105712DCEFBDULL, "w_encounter3_205_0.mao"},
	{0x4CB3145712DCF6F6ULL, "w_encounter3_205_0.mmh"},
	{0x4CB31E5712DD07F4ULL, "w_encounter3_205_0.msh"},
	{0x4CB350425EF9DE55ULL, "gen00fl_fenris6126923.cl"},
	{0x4CB3CF4EEFA90CF2ULL, "vxht_chantry_n_13x_0_0.vlm"},
	{0x4CB4356B4E06A668ULL, "mrl000_marethari6129324.cl"},
	{0x4CB459CF8FE223EFULL, "6078982_f.fxe"},
	{0x4CB4B19EAD81DB49ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirewall.dura.evt"},
	{0x4CB51E73776EE791ULL, "62285745.ogg"},
	{0x4CB5632EF0E3333FULL, "245311241.ogg"},
	{0x4CB59A6C0E045B69ULL, "vxht_mansion2_2vl_0_0.vlm"},
	{0x4CB5F640BF879751ULL, "389244470.ogg"},
	{0x4CB624268AC20EF8ULL, "6122098_m.fxe"},
	{0x4CB637C2F8BD62D8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic_1.mao"},
	{0x4CB6AE1A09B36B60ULL, "dae000_hate_follower_react6103228.cl"},
	{0x4CB6CD8F78121534ULL, "48073071.ogg"},
	{0x4CB6E5DA36E8230CULL, "70584439.ogg"},
	{0x4CB75D87ADD9A66EULL, "hlac_w_sundermount_65536_0z.dds"},
	{0x4CB770714F34230DULL, "6198195_m.fxe"},
	{0x4CB78D5388528CF0ULL, "genip_door_lowin_industrial.utp"},
	{0x4CB8189124108B20ULL, "vxht_0_hightown_bc6_0_0.vlm"},
	{0x4CB854EC3C9EBC03ULL, "vxw_bonepitdestroyed_a3_2ad_0_0.vlm"},
	{0x4CB87ACCE0DC084CULL, "win_smoke.mao"},
	{0x4CB89C55B5707314ULL, "6120852_m.fxe"},
	{0x4CB910B08DD68DDBULL, "one100dg_kelder6175893.cl"},
	{0x4CB94346D422C5D1ULL, "dae200_saemus6091710.cl"},
	{0x4CB9E4CE7F479682ULL, "w_bonepit_54_0.rml"},
	{0x4CBA01ED47E5C47EULL, "vxlt_darktown_4yc_0_0.vlm"},
	{0x4CBA2EE98C287B6CULL, "vxdr_0_deeproads_f5g_0_0.vlm"},
	{0x4CBA618B2593F7AAULL, "6098775_m.fxe"},
	{0x4CBA89F7875FD46AULL, "llac_dr_dungeon_3216_8.dds"},
	{0x4CBAC58EA70FCC16ULL, "284819600.ogg"},
	{0x4CBAE5F3379B1B4DULL, "dae000_sabina6162333.cl"},
	{0x4CBB31F1B5A0A814ULL, "lgt000_chanter6166477.cl"},
	{0x4CBBD4258C8109F8ULL, "6117682_f.fxe"},
	{0x4CBC738CAB2C7653ULL, "vxw_bonepit_zy_0_0.vlm"},
	{0x4CBC9FAD7554E172ULL, "6106432_m.fxe"},
	{0x4CBCEA308EC85D25ULL, "lgt000ip_ultra_hoop.ncs"},
	{0x4CBD20C4D813900BULL, "and111_lirene6085109.cl"},
	{0x4CBD2CA6047F25B6ULL, "1756575.ogg"},
	{0x4CBD760C9D84C06FULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.cess.evt"},
	{0x4CBDBD7766344988ULL, "6120874_m.fxe"},
	{0x4CBE1BC843C973E2ULL, "6127437_m.fxe"},
	{0x4CBE3D7477BA3820ULL, "6165415_m.fxe"},
	{0x4CBE7225D2514D28ULL, "vxht_mansion_n_3yn_0_0.vlm"},
	{0x4CBF01CC969447D1ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.mmh"},
	{0x4CBF07DC1C44F3E9ULL, "6043651_m.fxe"},
	{0x4CBF1356CAB934BCULL, "6119575_m.fxe"},
	{0x4CBF28A07747D274ULL, "gen_im_arm_boo_lgt_act3_01.uti"},
	{0x4CC04AC4F25FCB96ULL, "376866345.ogg"},
	{0x4CC072E6447D28FEULL, "lt_0_lowtown_a2_n_c_38_0.rml"},
	{0x4CC1AEC209A0E8C2ULL, "6091978_m.fxe"},
	{0x4CC200059A218069ULL, "vxht_0_hightown_a2_5y2_0_0.vlm"},
	{0x4CC2999EAD8DEFB8ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirewall.dura.ani"},
	{0x4CC2BA5F51E694C3ULL, "285999113.ogg"},
	{0x4CC30A5BEF93C660ULL, "vxca_cavern1_haunted_jl2_0_0.vlm"},
	{0x4CC33587705C13F0ULL, "hlac_lt_hangedman_-1_5z.dds"},
	{0x4CC352267A911EDCULL, "194793090.ogg"},
	{0x4CC3770C9D892D18ULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.cess.gad"},
	{0x4CC404FBD6FD48FBULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90r.evt"},
	{0x4CC4162C1BD28F8EULL, "gen00fl_isabela_init6110258.cl"},
	{0x4CC455AC6DA8B657ULL, "hlac_ht_keep_n_3733_1x.dds"},
	{0x4CC4B53ABD7DC25EULL, "vxdr_0_deeproads_egd_0_0.vlm"},
	{0x4CC4E2CB65C106EBULL, "vxlt_0_lowtown_ewn_0_0.vlm"},
	{0x4CC59F41626CBA51ULL, "77608991.ogg"},
	{0x4CC5A9879C8AE11BULL, "vrc221cr_crazed_commander.utc"},
	{0x4CC5BECDE95031B2ULL, "6108060_f.fxe"},
	{0x4CC5C0A5CE5E0223ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium1_c.anb"},
	{0x4CC5C405F49E6B54ULL, "vxw_bonepitdestroyed_a3_1u8_0_0.vlm"},
	{0x4CC5EBB734881939ULL, "gen00fl_anders6111149.cl"},
	{0x4CC6009BA25CE396ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_up_c.impa.ani"},
	{0x4CC6412E5AD2198BULL, "6128984_f.fxe"},
	{0x4CC6E284160DCE65ULL, "6219382_m.fxe"},
	{0x4CC73B77CFAB335DULL, "6110469_m.fxe"},
	{0x4CC783564FF15733ULL, "vrc221_varric6040869.cl"},
	{0x4CC7B494E9039C77ULL, "plt_cod_mgc_rules.plo"},
	{0x4CC7BA45E3EE476EULL, "follower_banter6170708.cl"},
	{0x4CC814945F6A8591ULL, "165854507.ogg"},
	{0x4CC8175FDEA1A47BULL, "6135088_m.fxe"},
	{0x4CC85B490BFA07BFULL, "49365448.ogg"},
	{0x4CC88C0FADB7B168ULL, "vxlt_0_lowtown_a3_n_c_es1_0_0.vlm"},
	{0x4CC9839FC0F7D06FULL, "atl_fonttest_dxt1_01.dds"},
	{0x4CC9C806B3D0D4FAULL, "6031129_m.fxe"},
	{0x4CCA9C4D86C59948ULL, "vxlt_gallowscourtyard_n_7io_0_0.vlm"},
	{0x4CCACECE7F55E4D7ULL, "w_bonepit_54_0.msh"},
	{0x4CCAE0CE7F56033AULL, "w_bonepit_54_0.mao"},
	{0x4CCAE4CE7F560A49ULL, "w_bonepit_54_0.mmh"},
	{0x4CCAFDFBD7035B3CULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90r.gad"},
	{0x4CCB684630F94B5CULL, "629734285.ogg"},
	{0x4CCB768F8FB2B895ULL, "hlac_ht_0_hightown_n_-131074_0x.dds"},
	{0x4CCBF6C89637C114ULL, "rdr111ip_desk.utp"},
	{0x4CCC0858B2742997ULL, "fen210_orana6109744.cl"},
	{0x4CCCA045925D26DBULL, "vxdr_0_deeproads_a6z_0_0.vlm"},
	{0x4CCCDF2BBE5C9493ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.dura.evt"},
	{0x4CCD42364450A330ULL, "330747385.ogg"},
	{0x4CCD6FAE8C1534BFULL, "vxlt_gallowscourtyard_n_6yq_0_0.vlm"},
	{0x4CCDD37900AC6E02ULL, "348384857.ogg"},
	{0x4CCDDB308ED6B77CULL, "lgt000ip_ultra_hoop.utp"},
	{0x4CCDFC5ECF4A7B33ULL, "vxlt_gallowstemplar_n_3st_0_0.vlm"},
	{0x4CCE11890017273EULL, "hlac_ca_caverns1_24423_11z.dds"},
	{0x4CCE6E487D2BFCF9ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_act_c.anb"},
	{0x4CCEF0E4E6D736FEULL, "454348097.ogg"},
	{0x4CCFC223F3E04082ULL, "lla_ht_0_hightown_a2_65535_0.dds"},
	{0x4CCFE50CF66AA26BULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.anb"},
	{0x4CD0389874CAEA4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90_alt.ani"},
	{0x4CD05942DCC0890CULL, "vxht_mansion2_1a3_0_0.vlm"},
	{0x4CD05D309A2950A4ULL, "mag291_vincento6027685.cl"},
	{0x4CD06D85F285B5B3ULL, "vxlt_undercityencounter_z3_0_0.vlm"},
	{0x4CD074A83164DD5AULL, "vxdr_primevalentrance_1qh_0_0.vlm"},
	{0x4CD0BC24977461A3ULL, "hlac_dr_slavers_3217_0z.dds"},
	{0x4CD0D9869B25F8BEULL, "6048309_m.fxe"},
	{0x4CD182F44F1A569EULL, "hlac_w_bonepitdestroyed_a3_262149_0x.dds"},
	{0x4CD18E6E732E8D57ULL, "lt_0_lowtown_docks_a3_n_44_0.rml"},
	{0x4CD1FCFBD709783AULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90r.ani"},
	{0x4CD208003A692FE8ULL, "gen00flst_sebactthreea.stg"},
	{0x4CD2D2964A851FDCULL, "vxdr_primevaltomb_4xl_0_0.vlm"},
	{0x4CD2D82BBE60F4D4ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.dura.gad"},
	{0x4CD2FADDB4B99D7DULL, "hlac_lt_0_lowtown_docks_a3_131071_1z.dds"},
	{0x4CD30770A35FB49DULL, "243395865.ogg"},
	{0x4CD34982A1602781ULL, "vxlt_gallowscourtyard_6xc_0_0.vlm"},
	{0x4CD354A857B71F6BULL, "vxlt_0_lowtown_era_0_0.vlm"},
	{0x4CD37F268532C3E7ULL, "hlac_lt_0_lowtown_docks_0_6y.dds"},
	{0x4CD39DA5CE6A044CULL, "art\\vfx\\combat\\vfx_w_wns_front_medium1_c.mmh"},
	{0x4CD3BFCE7CBEA0DCULL, "375126057.ogg"},
	{0x4CD3D9E3BA268697ULL, "6090262_m.fxe"},
	{0x4CD3FCECB43B2D04ULL, "88500728.ogg"},
	{0x4CD3FE511952A2FCULL, "983447775.ogg"},
	{0x4CD4089BA2692E87ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_up_c.impa.evt"},
	{0x4CD577D3D48F100FULL, "vxca_caverns2_9a3_0_0.vlm"},
	{0x4CD59AEB1622D8D5ULL, "6116070_m.fxe"},
	{0x4CD59F2C517264D4ULL, "vxca_cavern1_haunted_i27_0_0.vlm"},
	{0x4CD5BC35BEC5FB73ULL, "vxw_bonepit_1ph_0_0.vlm"},
	{0x4CD5F115A444289EULL, "6032599_m.fxe"},
	{0x4CD67033EE523BD8ULL, "vxlt_0_lowtown_docks_n_bez_0_0.vlm"},
	{0x4CD6EA90B47A1297ULL, "vxht_mansion2b_1wf_0_0.vlm"},
	{0x4CD6FF0F13A69762ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_basic.phy"},
	{0x4CD71446D47ACD34ULL, "6115632_m.fxe"},
	{0x4CD7459874D11E8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90_alt.gad"},
	{0x4CD7578F301B6ABDULL, "fen310_fenris6175087.cl"},
	{0x4CD796B7697AC79FULL, "w_encounter2_camp_43_0_lc.mmh"},
	{0x4CD79D5D56D320EFULL, "gen_ip_bodybag.utp"},
	{0x4CD7B15A5BD0A0D3ULL, "w_blightlands_63_0.rml"},
	{0x4CD7D55BF36BFF5DULL, "vxca_caverns1_jfp_0_0.vlm"},
	{0x4CD83840196F837DULL, "art\\levels\\races\\proxy\\props\\_textures\\storage_d.dds"},
	{0x4CD84BCB0A5D77CEULL, "ico_poison_fel.dds"},
	{0x4CD88287B5C5AE3BULL, "llac_ht_0_hightown_a2_-65535_1.dds"},
	{0x4CD8B43C5B13560FULL, "vxht_0_hightown_a2_arm_0_0.vlm"},
	{0x4CD8F7DF14B60563ULL, "ca_cavern1_haunted_24423.rml"},
	{0x4CD90DB4E2407FD8ULL, "6072693_f.fxe"},
	{0x4CD9289B16211535ULL, "hlac_dr_primevaltomb_659_12z.dds"},
	{0x4CD92D03805AB9EBULL, "6152151_m.fxe"},
	{0x4CD9956E68A7782FULL, "follower_barks6153660.cl"},
	{0x4CD9DDB1D1CE2D91ULL, "6088926_m.fxe"},
	{0x4CD9E72BBE672CB2ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.dura.ani"},
	{0x4CDA5522CAD549F1ULL, "email.ncs"},
	{0x4CDA6C21EECF58EBULL, "6075141_f.fxe"},
	{0x4CDA7DE4206BB9DBULL, "llac_lt_0_lowtown_docks_n_131073_1.dds"},
	{0x4CDA90EAC60B1D1CULL, "6164395_m.fxe"},
	{0x4CDAAC45C1BAB7C1ULL, "6120992_m.fxe"},
	{0x4CDB02352B8E8EE1ULL, "follower_banter6141824.cl"},
	{0x4CDB23601CA3B23DULL, "mag320_evelina6051880.cl"},
	{0x4CDB2707F7F6650EULL, "6122131_m.fxe"},
	{0x4CDB299BA26F85D0ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_up_c.impa.gad"},
	{0x4CDB685280041E81ULL, "gen_im_trs_slv_sta_ele_r2.uti"},
	{0x4CDBC3D7393FE4CDULL, "hlac_lt_0_lowtown_docks_a3_-65538_8z.dds"},
	{0x4CDBC8479CF66FE8ULL, "vxw_bonepitdestroyed_a3_1et_0_0.vlm"},
	{0x4CDC5E7C2C3F61A7ULL, "6109418_m.fxe"},
	{0x4CDCCAAFBB8137C9ULL, "6205004_m.fxe"},
	{0x4CDD1B1A8F389330ULL, "6155526_m.fxe"},
	{0x4CDD1BFBF4AE269AULL, "6066386_f.fxe"},
	{0x4CDD409874D5826DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90_alt.evt"},
	{0x4CDD548A09BE69C2ULL, "255243034.ogg"},
	{0x4CDDC16562C1378DULL, "6094808_m.fxe"},
	{0x4CDDF8C6F1964E9BULL, "gencr_mabari.utc"},
	{0x4CDE745B7A0B0905ULL, "lgt160_ntb_elf_conflict6163337.cl"},
	{0x4CDE9622D03D59B9ULL, "486703630.ogg"},
	{0x4CDE9FB076E57AB7ULL, "680127162.ogg"},
	{0x4CDF336BAFA72D71ULL, "dae300_petrice6098240.cl"},
	{0x4CDF33E65C1BF139ULL, "hlac_ca_caverns1_24426_2x.dds"},
	{0x4CDF60F65D5D0FCDULL, "6174229_m.fxe"},
	{0x4CDFB2A318265624ULL, "68331799.ogg"},
	{0x4CDFCAB33C6105CCULL, "vxlt_0_lowtown_a3_n_c_e1s_0_0.vlm"},
	{0x4CE01ED4C623B517ULL, "468776980.ogg"},
	{0x4CE02AEE6643C9A9ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic_4.mao"},
	{0x4CE0F002E3E6A4E9ULL, "vxht_mansion2_2yq_0_0.vlm"},
	{0x4CE0F41E72AFB8B8ULL, "6092354_f.fxe"},
	{0x4CE110A5956FAA6FULL, "vxlt_darktown_6oz_0_0.vlm"},
	{0x4CE17234235BFA4EULL, "cod_hst_kirk_arch_1a_1.utp"},
	{0x4CE1952E6BB0D95BULL, "6219234_m.fxe"},
	{0x4CE222FCBC2841B8ULL, "vxlt_darktown_4f8_0_0.vlm"},
	{0x4CE249F97A7FEAC5ULL, "follower_banter6171885.cl"},
	{0x4CE24DCC981EE920ULL, "vxht_mansion_n_4yi_0_0.vlm"},
	{0x4CE3142232994550ULL, "vxlt_undercitydungeon_1ig_0_0.vlm"},
	{0x4CE361CF9402E9FFULL, "vxht_0_hightown_aps_0_0.vlm"},
	{0x4CE392E0B4375B1BULL, "art\\characters\\weapons\\dagger\\w_dgr_dg103a.msh"},
	{0x4CE39433BF525F7DULL, "vxw_woundedcoast_1fc_0_0.vlm"},
	{0x4CE3B0E0B4378E05ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg103a.mmh"},
	{0x4CE3E1FC3EDDA3E1ULL, "6048416_m.fxe"},
	{0x4CE3EB5C4A53AAD1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01a.mmh"},
	{0x4CE3EC20527F3D3AULL, "vxht_mansion_n_4zz_0_0.vlm"},
	{0x4CE4055C4A53D68FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01a.msh"},
	{0x4CE42E12B0BC13F3ULL, "vxw_bonepit_1uk_0_0.vlm"},
	{0x4CE4E68D909C9DBFULL, "vxdr_0_deeproads_f5h_0_0.vlm"},
	{0x4CE503F57737808AULL, "vxht_0_hightown_a3_amx_0_0.vlm"},
	{0x4CE524B9256ADF2FULL, "w_bonepitdestroyed_a3_174_0.mao"},
	{0x4CE530B9256AF3CCULL, "w_bonepitdestroyed_a3_174_0.mmh"},
	{0x4CE536B9256AFDBEULL, "w_bonepitdestroyed_a3_174_0.msh"},
	{0x4CE53842AED1052BULL, "6071869_m.fxe"},
	{0x4CE557841DC80DFEULL, "vxlt_darktown_4nf_0_0.vlm"},
	{0x4CE5E3D3E74F8C3DULL, "vxca_caverns2_95u_0_0.vlm"},
	{0x4CE5EF0745A3A996ULL, "hlac_ht_keep_3736_1y.dds"},
	{0x4CE60E3AA128BF17ULL, "6000893_f.fxe"},
	{0x4CE629A1A462AEF0ULL, "w_blightlands_115_0.msh"},
	{0x4CE637A1A462C60AULL, "w_blightlands_115_0.mmh"},
	{0x4CE63BA1A462CD41ULL, "w_blightlands_115_0.mao"},
	{0x4CE71E9DC1D46F85ULL, "6118829_m.fxe"},
	{0x4CE71FDD200A9DB9ULL, "6129744_m.fxe"},
	{0x4CE7821AF0C7B430ULL, "one110_gustav6089613.cl"},
	{0x4CE78A522B7B0FFEULL, "6131148_m.fxe"},
	{0x4CE7C923810EBFD3ULL, "vxw_bonepit_kz_0_0.vlm"},
	{0x4CE8305C8051718BULL, "vxlt_0_lowtown_n_e2r_0_0.vlm"},
	{0x4CE8D18C8745618DULL, "zz_vault_debug6203249.cl"},
	{0x4CE8D2268ACB03EBULL, "317459037.ogg"},
	{0x4CE9190673CC9A34ULL, "42364577.ogg"},
	{0x4CE93BA10D4B46AAULL, "147724731.ogg"},
	{0x4CE949CDA3BC5327ULL, "vxlt_0_lowtown_docks_n_cor_0_0.vlm"},
	{0x4CE96DD629BCEB75ULL, "946998981.ogg"},
	{0x4CE980D268300A44ULL, "vxht_mansion2b_3eq_0_0.vlm"},
	{0x4CE9CC6128BAB42DULL, "plt_mrl100pt_n_update.nss"},
	{0x4CEA3A248EEC8868ULL, "vxw_bonepitdestroyed_a3_1nu_0_0.vlm"},
	{0x4CEA864C844DCCB1ULL, "6123418_gamlen_blame.cut"},
	{0x4CEA9AD7847A6F0BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg04.phy"},
	{0x4CEB3C8CAD775049ULL, "6208704_m.fxe"},
	{0x4CEB708F867FA2DFULL, "156250149.ogg"},
	{0x4CEBD64BC0A8CB63ULL, "538208896.ogg"},
	{0x4CEBEC156EFEF32DULL, "ts_kirkwall_prog_junk1.gda"},
	{0x4CEC228229D0D3A2ULL, "gen00fl_aveline6132759.cl"},
	{0x4CEC2A11808B4536ULL, "gen00fl_varric6099400.cl"},
	{0x4CECD68E2505CA3DULL, "y_crate_05.dds"},
	{0x4CED4EA03CBD2EECULL, "736371944.ogg"},
	{0x4CEE19AC38E20F9FULL, "zz_dae_debug6045447.cl"},
	{0x4CEE468D05538D6BULL, "seb221_ruxton_harimann6177212.cl"},
	{0x4CEE578C6E5DDAF0ULL, "hlac_ht_mansion2_818_53z.dds"},
	{0x4CEE9FEAAC19F9B2ULL, "hlac_lt_0_lowtown_a3_n_c_-262143_0x.dds"},
	{0x4CEEB13BEB042EB0ULL, "227937565.ogg"},
	{0x4CEEE9EC233DE756ULL, "6099422_m.fxe"},
	{0x4CEF3D8ABB32EC2FULL, "llac_dr_dungeon_3217_8.dds"},
	{0x4CEF5F1F25233320ULL, "435025236.ogg"},
	{0x4CEF82302905D45FULL, "6085177_f.fxe"},
	{0x4CEF90805DCE448CULL, "587290513.ogg"},
	{0x4CEF912EA6380FA6ULL, "10114359.ogg"},
	{0x4CEFA35F4B1EA372ULL, "follower_banter6172975.cl"},
	{0x4CEFEFDC7E2ECC2FULL, "vxlt_undercitydungeon_27i_0_0.vlm"},
	{0x4CF00EFD96948830ULL, "901162511.ogg"},
	{0x4CF038623BA3506BULL, "vxlt_0_lowtown_ehc_0_0.vlm"},
	{0x4CF082B26D95EC63ULL, "w_bonepit_73_0.mao"},
	{0x4CF086B26D95F2B0ULL, "w_bonepit_73_0.mmh"},
	{0x4CF094B26D960ACAULL, "w_bonepit_73_0.msh"},
	{0x4CF1C2720F7A0CA1ULL, "144685427.ogg"},
	{0x4CF1DD023254CEFFULL, "vxca_caverns2_9fv_0_0.vlm"},
	{0x4CF1E5C8965873CDULL, "rdr111ip_desk.ncs"},
	{0x4CF1F66987980B6FULL, "framebuffereffects_radialblur_p1.psh.fxo.dx11"},
	{0x4CF20610B14877FDULL, "vxw_woundedcoast_1ab_0_0.vlm"},
	{0x4CF2A396F41DBC7EULL, "6168036_m.fxe"},
	{0x4CF36A62715E0FD7ULL, "vxw_encounter2_camp_nq_0_0.vlm"},
	{0x4CF375D2692D990AULL, "vxht_0_hightown_a3_ajy_0_0.vlm"},
	{0x4CF45ABCD9E81B7AULL, "lgt260_way_blood_mage6148402.cl"},
	{0x4CF464877372079BULL, "986573342.ogg"},
	{0x4CF465720896B94EULL, "vxht_0_hightown_atr_0_0.vlm"},
	{0x4CF4A38B1006FF58ULL, "6125859_m.fxe"},
	{0x4CF4CDA3337141AEULL, "dre000_vampire6034270.cl"},
	{0x4CF4D23A59DC63DBULL, "6063848_m.fxe"},
	{0x4CF5419617494410ULL, "character_character_p135.psh.fxo"},
	{0x4CF54CAB79F4F937ULL, "gen_im_wep_rog_arc_default.uti"},
	{0x4CF5890E690EA920ULL, "gen_im_cns_flu_sebastian_01.uti"},
	{0x4CF5907AB3CA8805ULL, "6117929_m.fxe"},
	{0x4CF621A966F5F242ULL, "567371245.ogg"},
	{0x4CF62659459C9B1FULL, "6216983_m.fxe"},
	{0x4CF673D8B5D39269ULL, "lt_0_lowtown_n_34_0_lc.mmh"},
	{0x4CF68CDBCA998703ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.fly_short.evt"},
	{0x4CF69BC23198556FULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_grass_d.dds"},
	{0x4CF6AC70DD5740FBULL, "hlac_ht_chantry_n_2534_18x.dds"},
	{0x4CF6CBBD7ED56884ULL, "6149769_m.fxe"},
	{0x4CF7BBCAFF3C2642ULL, "6083623_m.fxe"},
	{0x4CF7CA3AFF3C7550ULL, "mcr370_meredith6083600.cl"},
	{0x4CF815E656EB2E16ULL, "vxlt_darktown_5nk_0_0.vlm"},
	{0x4CF8ADB395A759BBULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saint.msh"},
	{0x4CF8BFB395A77806ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saint.mao"},
	{0x4CF8CBB395A78CA5ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saint.mmh"},
	{0x4CF98AF6DC4C50E7ULL, "vxca_caverns2_4zs_0_0.vlm"},
	{0x4CF9995324755373ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\body_curtsey.evt"},
	{0x4CF9A19AFD4D72BCULL, "6093866_m.fxe"},
	{0x4CF9E5CAB12B3EC6ULL, "ht_0_hightownexplosion_n_37_0.rml"},
	{0x4CF9E74EF2D2DEF9ULL, "vxw_encounter2_du_0_0.vlm"},
	{0x4CFA3FD07BAF2570ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1glowworms.mao"},
	{0x4CFA43D07BAF2B83ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1glowworms.mmh"},
	{0x4CFA51D07BAF43DDULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1glowworms.msh"},
	{0x4CFA635A4BB47759ULL, "llac_lt_gallowscourtyard_n_65536_1.dds"},
	{0x4CFA6FE644AE9BA2ULL, "lt_0_lowtown_a2_n_c_38_0.anb"},
	{0x4CFA8B70C4674BABULL, "6091386_m.fxe"},
	{0x4CFB5F70D7313CC1ULL, "6136021_m.fxe"},
	{0x4CFB9315A0B5D02EULL, "hightown_male_noble2_mob.cnv"},
	{0x4CFBA2245E4B2991ULL, "vxht_0_hightown_n_c5w_0_0.vlm"},
	{0x4CFBDA0F13C57416ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_basic.mao"},
	{0x4CFBE60F13C588B5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_basic.mmh"},
	{0x4CFBE80F13C58C2BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_basic.msh"},
	{0x4CFC465CCAA6ABB8ULL, "ash_04_f.dds"},
	{0x4CFC70BF163E905FULL, "555893615.ogg"},
	{0x4CFCBA4824B1D3C0ULL, "hlac_lt_0_lowtown_docks_n_-65539_4z.dds"},
	{0x4CFD1EA95EECEC12ULL, "6097009_f.fxe"},
	{0x4CFD214EF4706B15ULL, "vxht_keep_n_5gw_0_0.vlm"},
	{0x4CFD8A8B08559AA5ULL, "334897212.ogg"},
	{0x4CFDF729A10EC920ULL, "dae000_bartrand6032013.cl"},
	{0x4CFE5C71E0DE69F7ULL, "827091611.ogg"},
	{0x4CFE93D7E27A724BULL, "dae000_hate_enemies6167528.cl"},
	{0x4CFEB515A0B83F9EULL, "hightown_male_noble2_mob.bnk"},
	{0x4CFF06B6A01472B8ULL, "vxlt_darktown_4h6_0_0.vlm"},
	{0x4CFFA803A7F86603ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm03.msh"},
	{0x4CFFB603A7F87E5DULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm03.mmh"},
	{0x4D004064DDFB9792ULL, "hlac_w_sundermount_131073_0z.dds"},
	{0x4D0051E1DDDF05A0ULL, "vxht_0_hightown_bx3_0_0.vlm"},
	{0x4D0077EDA33E907DULL, "vxw_woundedcoast_1la_0_0.vlm"},
	{0x4D009253247B65B4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\body_curtsey.gad"},
	{0x4D009CB36039D4A6ULL, "6142296_m.fxe"},
	{0x4D00A878BBEA8CD7ULL, "vxca_cavern1_haunted_k6n_0_0.vlm"},
	{0x4D010264B643E4E9ULL, "6173938_m.fxe"},
	{0x4D012B4641ECBF8DULL, "vxlt_0_lowtown_docks_a3_cur_0_0.vlm"},
	{0x4D01C19B32F1C163ULL, "dae000_bartrand6206156.cl"},
	{0x4D01DC3F63542857ULL, "vxw_encounter2_camp_ip_0_0.vlm"},
	{0x4D0286CD84D309D7ULL, "gen00fl_aveline6004167.cl"},
	{0x4D030DF884B64C15ULL, "vxdr_slavers_1aw_0_0.vlm"},
	{0x4D03769384D673CAULL, "class_abilities_merril.gda"},
	{0x4D0453D4A3C871EAULL, "6117984_m.fxe"},
	{0x4D049DF4E0417501ULL, "vxlt_0_lowtown_n_h1x_0_0.vlm"},
	{0x4D04BE6A9C2CAB5AULL, "6178521_m.fxe"},
	{0x4D04E272192B1FC7ULL, "6206915_m.fxe"},
	{0x4D0526A056842562ULL, "6097068_m.fxe"},
	{0x4D0586E8CDA4B92AULL, "6036992_f.fxe"},
	{0x4D05D2C606FF435EULL, "6163194_m.fxe"},
	{0x4D05DE261849A13DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_wallcap.phy"},
	{0x4D05F7D05D0AB06BULL, "llac_ca_caverns2_5802_8.dds"},
	{0x4D062DA7F1323EC2ULL, "6086923_m.fxe"},
	{0x4D06322869908132ULL, "dae000ip_crowd_mf3_talk1.utp"},
	{0x4D06A153247FEA92ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\body_curtsey.ani"},
	{0x4D074F05B2DCB7D4ULL, "6116585_m.fxe"},
	{0x4D07C6649E99FA3DULL, "6029141_m.fxe"},
	{0x4D07EAECFC5EB531ULL, "llac_lt_0_lowtown_65535_2.dds"},
	{0x4D07F5656E5DC1E3ULL, "260869709.ogg"},
	{0x4D0900CDDAEE1C35ULL, "6095591_m.fxe"},
	{0x4D0915914DFA4BBDULL, "428473654.ogg"},
	{0x4D0936047DB35927ULL, "art\\vfx\\creatures\\fxc_bm_vom_c.impa.gad"},
	{0x4D095D72F7BEDBBAULL, "6094515_m.fxe"},
	{0x4D095F43FC211DCBULL, "lgt220_shp_mag_guard6149875.cl"},
	{0x4D0975F6B32735DFULL, "6068544_m.fxe"},
	{0x4D09FBFD5B1CD01EULL, "1053702105.ogg"},
	{0x4D0A140150414211ULL, "vxht_0_hightown_n_c2v_0_0.vlm"},
	{0x4D0A56D2ED96034BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medc_0t.dds"},
	{0x4D0A76582E0A90DFULL, "180160348.ogg"},
	{0x4D0B896E40813675ULL, "794175561.ogg"},
	{0x4D0BCDA6072A1856ULL, "6093622_m.fxe"},
	{0x4D0BD7357B1D6A2CULL, "6099590_m.fxe"},
	{0x4D0C18B78D0D9C78ULL, "147045306.ogg"},
	{0x4D0C52173991D95FULL, "6167718_f.fxe"},
	{0x4D0CC0CD1C48B11BULL, "gen00fl_merrill6121458.cl"},
	{0x4D0CE58CBB3E1A23ULL, "llac_dr_dungeon_3215_2.dds"},
	{0x4D0D1588F9CB9B51ULL, "dae000_meredith6154282.cl"},
	{0x4D0D1F6F73BA32C1ULL, "ser221ip_note_starkhaven.utp"},
	{0x4D0D28DA4899CFDAULL, "hlac_ht_chantry_n_2534_40x.dds"},
	{0x4D0DE7CFB35F49E2ULL, "vxdr_0_deeproads_8we_0_0.vlm"},
	{0x4D0E29B03EC9675DULL, "6045419_m.fxe"},
	{0x4D0E4A5C0FF73926ULL, "vxlt_0_lowtown_docks_a3_n_d15_0_0.vlm"},
	{0x4D0E9A29E4B74E2EULL, "dae000ip_crowd_mf2_talk8.utp"},
	{0x4D0EDB03B014451FULL, "dae200_saemus6091688.cl"},
	{0x4D0EDEBDA63D76A2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01b.phy"},
	{0x4D0EE4CD836FC311ULL, "6108248_m.fxe"},
	{0x4D0EE5974B9C1ED1ULL, "353086945.ogg"},
	{0x4D0EEC8FA943F0E3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_urn_d.dds"},
	{0x4D0F9D2333E2D80DULL, "vxlt_0_lowtown_docks_a3_cxq_0_0.vlm"},
	{0x4D102B047DB96534ULL, "art\\vfx\\creatures\\fxc_bm_vom_c.impa.evt"},
	{0x4D105F8DD0F9932FULL, "hla_ht_0_hightown_a3_65538_1x.dds"},
	{0x4D10D5EFD98CDB50ULL, "one280_emrys6120031.cl"},
	{0x4D10DBF62AD7EF02ULL, "vxca_caverns1_mines_gke_0_1.vlm"},
	{0x4D112AF4130E6D6EULL, "vxht_0_hightown_c8d_0_0.vlm"},
	{0x4D114413F846C41CULL, "follower_banter6141654.cl"},
	{0x4D114A110ACA100DULL, "6127207_m.fxe"},
	{0x4D117492307FC910ULL, "vxht_0_hightown_n_caf_0_0.vlm"},
	{0x4D11AAD8C2CFBFB5ULL, "l_qun211ar_poison_encounter.bnk"},
	{0x4D11D5D43724D45DULL, "dae300_vashothlead6062792.cl"},
	{0x4D11FCE5D28E0E88ULL, "mcr310_argument6074427.cl"},
	{0x4D1232F512D2E303ULL, "744092601.ogg"},
	{0x4D126C3C1691C92AULL, "rdr110dg_guard_warehouse6016691.cl"},
	{0x4D127C0F01D11490ULL, "6051272_f.fxe"},
	{0x4D12FB020C2B71ACULL, "dre000_sibling_has_blight6036511.cl"},
	{0x4D131F7C86149354ULL, "al_warrior_t04.gda"},
	{0x4D1330478E22C3DAULL, "6173421_m.fxe"},
	{0x4D135E2AA6F80C1EULL, "6019097_m.fxe"},
	{0x4D147B820B03AAA6ULL, "894811344.ogg"},
	{0x4D158D83562C47C2ULL, "w_blightlands_62_0.rml"},
	{0x4D15F992D2F3187EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_right_hand.evt"},
	{0x4D16013F5C7801D8ULL, "6077819_m.fxe"},
	{0x4D1617031BC0F467ULL, "6075299_f.fxe"},
	{0x4D1640E97B414179ULL, "vxht_0_hightown_a2_ap1_0_0.vlm"},
	{0x4D168C6C27DB5F42ULL, "vxht_keep_5ie_0_0.vlm"},
	{0x4D16D326185801A9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_wallcap.mao"},
	{0x4D16DF2618581642ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_wallcap.mmh"},
	{0x4D16E126185819F8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_wallcap.msh"},
	{0x4D172ADD31AECEE8ULL, "vxlt_0_lowtown_docks_b5e_0_0.vlm"},
	{0x4D1761D49790F04BULL, "550074600.ogg"},
	{0x4D17724D6EC5245DULL, "vxdr_slavers_140_0_0.vlm"},
	{0x4D17A4E9FB2D4568ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_enco_0d.dds"},
	{0x4D17AF67636DC40AULL, "302997691.ogg"},
	{0x4D17D9A2965D44B3ULL, "vxht_0_hightown_asn_0_0.vlm"},
	{0x4D17EB70256A0ACAULL, "6132818_f.fxe"},
	{0x4D187EB08A1B2BF4ULL, "476807572.ogg"},
	{0x4D18E917F07F3961ULL, "6148778_m.fxe"},
	{0x4D1957244EAAD955ULL, "427071879.ogg"},
	{0x4D198FCCB47A50CFULL, "6141555_m.fxe"},
	{0x4D19DADF14ED6ADFULL, "ca_cavern1_haunted_24423.anb"},
	{0x4D1B306D46E78999ULL, "6018681_f.fxe"},
	{0x4D1B318AF64BA86EULL, "gen00fl_dog6122028.cl"},
	{0x4D1B692D32B4D0E6ULL, "816530829.ogg"},
	{0x4D1B882FE5A2A1B7ULL, "464147195.ogg"},
	{0x4D1B941074A531A3ULL, "llac_lt_0_lowtown_a3_n_c_-131073_2.dds"},
	{0x4D1B99CD07DDF9D3ULL, "ave321cr_jeven_loyalist.utc"},
	{0x4D1BC42AF1117EAAULL, "6085908_m.fxe"},
	{0x4D1C5864D7041D53ULL, "vxdr_primevaltomb_5ae_0_0.vlm"},
	{0x4D1CAAA57C8D65B3ULL, "6125272_m.fxe"},
	{0x4D1CB70FD988E175ULL, "vxht_mansion_n_6hy_0_0.vlm"},
	{0x4D1CB718F2BD512CULL, "6103356_m.fxe"},
	{0x4D1D1092D2F95DA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_right_hand.gad"},
	{0x4D1D2053B7151F27ULL, "vxlt_0_lowtown_a2_n_gpf_0_0.vlm"},
	{0x4D1D4B6387AE77E0ULL, "vxlt_backalley_g_1l1_0_0.vlm"},
	{0x4D1D8B22205037B9ULL, "6071694_m.fxe"},
	{0x4D1DCEA279B4FF06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_flemeth.evt"},
	{0x4D1DF586789475F3ULL, "vxw_bonepit_1io_0_0.vlm"},
	{0x4D1E23047DC5956DULL, "art\\vfx\\creatures\\fxc_bm_vom_c.impa.ani"},
	{0x4D1E8D0AB2A9488FULL, "382699671.ogg"},
	{0x4D1E8DAD56769ACCULL, "157529182.ogg"},
	{0x4D1EAC51B8BF1815ULL, "follower_banter6155973.cl"},
	{0x4D1EFD8425DDD8D7ULL, "ach_exp_containers.dds"},
	{0x4D1F02DC86EBDCAFULL, "hlac_ht_0_hightown_a2_196607_0y.dds"},
	{0x4D1F6B430F7678D4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_6.ani"},
	{0x4D1F858C73C29A2EULL, "llac_lt_undercityencounter_-1_15.dds"},
	{0x4D1FD75EDD90815AULL, "gen00fl_aveline6105384.cl"},
	{0x4D1FE395B2CCC410ULL, "254865682.ogg"},
	{0x4D2088BF9B17CEB6ULL, "llac_ca_caverns1_mines_24425_3.dds"},
	{0x4D2157A79B3B1FF4ULL, "hlac_w_blightlands_-65535_0x.dds"},
	{0x4D21DF85D3FDB6E8ULL, "vxdr_primevaltomb_8xt_0_0.vlm"},
	{0x4D2262318B7163B5ULL, "6098804_m.fxe"},
	{0x4D22725F7C3C1902ULL, "411624053.ogg"},
	{0x4D22B7B1E4A2082AULL, "6151076_m.fxe"},
	{0x4D22FF05FE47CD65ULL, "50284165.ogg"},
	{0x4D232861E8E70022ULL, "hlac_w_encounter2_camp_-65536_0z.dds"},
	{0x4D23D6CFF3B61F31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_standing_cheer.ani"},
	{0x4D23E192D2FF2CF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_gesture_right_hand.ani"},
	{0x4D2404A38AD51329ULL, "6131603_f.fxe"},
	{0x4D2463406EC61768ULL, "zz_prmpt_debug6200949.cl"},
	{0x4D24B2C66D3759F9ULL, "vxht_0_hightown_a2_ak0_0_0.vlm"},
	{0x4D24E5A279BB4551ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_flemeth.gad"},
	{0x4D2569002851A0AFULL, "vxlt_0_lowtown_n_gpb_0_0.vlm"},
	{0x4D25A4D18674910FULL, "6081388_f.fxe"},
	{0x4D25C0006A040B6FULL, "mag360_samson6185085.cl"},
	{0x4D2668430F7C9236ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_6.gad"},
	{0x4D267783563A9517ULL, "w_blightlands_62_0.msh"},
	{0x4D26841BBA6121D3ULL, "dae000_bodhan_letter6144293.cl"},
	{0x4D268983563AB47AULL, "w_blightlands_62_0.mao"},
	{0x4D268D83563ABA89ULL, "w_blightlands_62_0.mmh"},
	{0x4D26C851C91D89B1ULL, "dae021_sister_samea_sta6142133.cl"},
	{0x4D2726865E0F6F22ULL, "6109631_m.fxe"},
	{0x4D27B40ED41386A5ULL, "follower_banter6169956.cl"},
	{0x4D28B23DDDDA9011ULL, "vxht_brothel_n_59j_0_0.vlm"},
	{0x4D28BAEF5F099B2CULL, "6093790_m.fxe"},
	{0x4D28F8FCCC369D21ULL, "vxlt_0_lowtown_n_f0e_0_0.vlm"},
	{0x4D290D7C8A76463EULL, "lt_0_lowtown_docks_a3_24_0_ll.mmh"},
	{0x4D293BF0695ED236ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\forcefield.evt"},
	{0x4D2990612868BA62ULL, "vxlt_0_lowtown_n_eqf_0_0.vlm"},
	{0x4D29A3E442BF6B89ULL, "mag110_sailor6107407.cl"},
	{0x4D29B7B88DDAAF66ULL, "vxlt_gallowstemplar_5xe_0_0.vlm"},
	{0x4D2A1872B8FA86E4ULL, "llac_lt_0_lowtown_a3_n_c_65536_2.dds"},
	{0x4D2A89E0B4734EB2ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg103a.phy"},
	{0x4D2AE1CFF3BC50D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_standing_cheer.gad"},
	{0x4D2AE52D9BDA46D9ULL, "hlac_ht_0_hightown_n_0_1z.dds"},
	{0x4D2AE6A279BFB2FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_die_flemeth.ani"},
	{0x4D2B3ECD70CC32A9ULL, "vxca_caverns2_8cn_0_0.vlm"},
	{0x4D2BBCDB62D41A62ULL, "214187614.ogg"},
	{0x4D2BE3D3CC4F8003ULL, "vxw_bonepitdestroyed_a3_29l_0_0.vlm"},
	{0x4D2C1387CC18A5B9ULL, "vxlt_warehouse_n_1d0_0_0.vlm"},
	{0x4D2C83430F812B95ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_6.evt"},
	{0x4D2CA13C4284FF41ULL, "6078816_f.fxe"},
	{0x4D2CCA3E4B41C645ULL, "gen00fl_anders6119695.cl"},
	{0x4D2CF043A201EE47ULL, "ser121_jethann6105705.cl"},
	{0x4D2D2B62887E2565ULL, "lgt160_wk_avern_body.bnk"},
	{0x4D2D5FFA360A4013ULL, "ht_0_hightown_n_21_0_lc.mmh"},
	{0x4D2D6D0B8465564EULL, "qun110_petrice6037139.cl"},
	{0x4D2DEC18300EEE99ULL, "vxlt_gallowstemplar_f_3fu_0_0.vlm"},
	{0x4D2E29A35A0CA754ULL, "dae000_gamlen6149425.cl"},
	{0x4D2E606BE957F511ULL, "vxw_woundedcoast_n2_0_0.vlm"},
	{0x4D2EEDA351ECE15BULL, "vxht_0_hightown_a3_ca7_0_0.vlm"},
	{0x4D2F01A709351C32ULL, "hlac_w_bonepitmining_a2_393221_0z.dds"},
	{0x4D2F05498A2C892DULL, "843118.ogg"},
	{0x4D2F28A8D9F7BBA7ULL, "zz_dae_debug6036584.cl"},
	{0x4D2FF25B19B9632FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_up_loop.gad"},
	{0x4D2FF4E1D545E0C0ULL, "follower_banter6143201.cl"},
	{0x4D3032F06964E1C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\forcefield.gad"},
	{0x4D307DBBA30D3561ULL, "vxlt_warehouse_ea_0_0.vlm"},
	{0x4D30DECFF3C0B738ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drinking_standing_cheer.evt"},
	{0x4D3102B53DA31E55ULL, "mag370_meredith6046836.cl"},
	{0x4D31072FEF6416F2ULL, "hlac_lt_0_lowtown_docks_n_196607_0z.dds"},
	{0x4D310D628881DA9DULL, "lgt160_wk_avern_body.cnv"},
	{0x4D32C81184CB0724ULL, "tm_war_d02.tnt"},
	{0x4D32F8027AC366A4ULL, "character_character_p115.psh.fxo.dx11.assembly"},
	{0x4D32FC332233732DULL, "vxht_mansion2_1py_0_0.vlm"},
	{0x4D3360068CEBD2C6ULL, "gen00fl_sebastian6180297.cl"},
	{0x4D337171FD256554ULL, "6095513_f.fxe"},
	{0x4D337B1C37F60721ULL, "hlac_w_encounter3_-393215_0x.dds"},
	{0x4D3387D64405D3E1ULL, "364958415.ogg"},
	{0x4D33AEC2BAAB40F9ULL, "vxw_encounter2_py_0_0.vlm"},
	{0x4D33EFD404B07F76ULL, "vxht_keep_6mn_0_0.vlm"},
	{0x4D34428173810ABDULL, "907249165.ogg"},
	{0x4D348341102D81A8ULL, "w_encounter3_225_0.mmh"},
	{0x4D348F41102D964BULL, "w_encounter3_225_0.mao"},
	{0x4D34A141102DB4D2ULL, "w_encounter3_225_0.msh"},
	{0x4D34C5BDA65E1BF5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01b.mmh"},
	{0x4D34C7BDA65E1F6BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01b.msh"},
	{0x4D34D08C16B16E27ULL, "635287608.ogg"},
	{0x4D35279614E1F832ULL, "fen310_varania6118574.cl"},
	{0x4D355E4D834249DFULL, "605751483.ogg"},
	{0x4D355EA25596A0D4ULL, "llac_ht_mansion_n_2995_1.dds"},
	{0x4D35CEF798FFD34AULL, "1033224746.ogg"},
	{0x4D36270C9365913EULL, "6017406_m.fxe"},
	{0x4D3672292B13D991ULL, "vxht_0_hightown_a2_c7b_0_0.vlm"},
	{0x4D367940D4DD3F17ULL, "6174995_m.fxe"},
	{0x4D36B25C3D161FA7ULL, "gen00fl_dog6117782.cl"},
	{0x4D36D9CB8C8B78E2ULL, "hlac_dr_dungeon_3217_5z.dds"},
	{0x4D3733F0696B026FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\forcefield.ani"},
	{0x4D377688BDA808EFULL, "gen00fl_aveline6137522.cl"},
	{0x4D38023E1A5ED2E2ULL, "vxlt_0_lowtown_n_ejg_0_0.vlm"},
	{0x4D386F2CD026D87BULL, "hlac_lt_hangedman_n_-1_14z.dds"},
	{0x4D38B19348A9CF42ULL, "vxht_0_hightown_n_cbi_0_0.vlm"},
	{0x4D394851F0418219ULL, "propalpha1_s2.psh.fxo"},
	{0x4D3A06716461E148ULL, "vxlt_warehouse_ed_0_0.vlm"},
	{0x4D3A55B0BE459883ULL, "vxw_bonepitdestroyed_a3_24o_0_0.vlm"},
	{0x4D3A8C0D0FAE6BCBULL, "vxdr_0_deeproads_cm5_0_0.vlm"},
	{0x4D3AC82323B3966CULL, "hlac_lt_0_lowtown_docks_n_-131072_0z.dds"},
	{0x4D3AE3EBFCC38F4EULL, "921343163.ogg"},
	{0x4D3B17B38A6A0EECULL, "6102143_f.fxe"},
	{0x4D3B25FC84575256ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v1.evt"},
	{0x4D3CC6AC4F2DC11BULL, "6061271_m.fxe"},
	{0x4D3DD0D8C2F53B3FULL, "l_qun211ar_poison_encounter.wbd"},
	{0x4D3DE59AC49FBFEFULL, "vxht_brothel_n_363_0_0.vlm"},
	{0x4D3DF9B2962618EBULL, "art\\characters\\creatures\\dragon\\animation\\tal_kckbk.ani"},
	{0x4D3E99CB319DBECDULL, "106373395.ogg"},
	{0x4D3F17192533FAE1ULL, "34982740.ogg"},
	{0x4D3F2844CE0251E9ULL, "696082581.ogg"},
	{0x4D3F6A048972DD7EULL, "qun210cr_guard6083836.cl"},
	{0x4D3FA4B395E34D52ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saint.phy"},
	{0x4D4033EB5584E543ULL, "dre000_hireling16038645.cl"},
	{0x4D409A420DB1D0FFULL, "ser200_aveline6201528.cl"},
	{0x4D4145A1731F4562ULL, "114552361.ogg"},
	{0x4D419D2ACD577855ULL, "rune_armor_defense.dds"},
	{0x4D41CFAC91996123ULL, "follower_banter6141592.cl"},
	{0x4D421CFC845D61E1ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v1.gad"},
	{0x4D424CBA0C3DD1AFULL, "vxlt_0_lowtown_n_gjl_0_0.vlm"},
	{0x4D427F16B53503DAULL, "503515445.ogg"},
	{0x4D42BB612906C03EULL, "plt_mrl100pt_n_update.plo"},
	{0x4D42CA17EF74A37AULL, "vxca_caverns1_mines_i29_0_0.vlm"},
	{0x4D42DC677397229CULL, "6130967_m.fxe"},
	{0x4D42FAF6AD8011B3ULL, "465209122.ogg"},
	{0x4D4341C5934D339BULL, "vxht_mansion_n_6or_0_0.vlm"},
	{0x4D43EA82E7A18DBFULL, "singleroofa_diff.dds"},
	{0x4D43EFE6597834FFULL, "110844504.ogg"},
	{0x4D44E4061D09F211ULL, "vxht_0_hightown_a2_c2a_0_0.vlm"},
	{0x4D44F4B7A26D016CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulleylong.msh"},
	{0x4D44FAB7A26D0B5EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulleylong.mmh"},
	{0x4D44FF3672EBE83BULL, "fxe_ripple_dis.mao"},
	{0x4D4506B7A26D1FF5ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulleylong.mao"},
	{0x4D450F5B19CBF0F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_up_loop.ani"},
	{0x4D45BDB1731D4DF0ULL, "hlac_lt_0_lowtown_a2_n_c_131072_6x.dds"},
	{0x4D45EE92C3379151ULL, "6061817_m.fxe"},
	{0x4D4648583921B400ULL, "6169411_m.fxe"},
	{0x4D471E0D9332C688ULL, "vxlt_gallowstemplar_4xr_0_0.vlm"},
	{0x4D4774BE189F5FCBULL, "hlac_lt_undercityencounter_-1_47x.dds"},
	{0x4D478EE68F69B8EDULL, "528699543.ogg"},
	{0x4D47ED8E3A9679E7ULL, "vxht_0_hightown_a2_ccp_0_0.vlm"},
	{0x4D47F181AEAF15BEULL, "6048118_m.fxe"},
	{0x4D480888B1EB6BEFULL, "qun120cr_javaris6042550.cl"},
	{0x4D481DFC8461CE8FULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v1.ani"},
	{0x4D4836727D97FB0EULL, "vxht_brothel_n_5hv_0_0.vlm"},
	{0x4D48B2441353B4EDULL, "gen00fl_anders6201145.cl"},
	{0x4D48FDEA01A4844BULL, "vxdr_0_deeproads_cj4_0_0.vlm"},
	{0x4D49274563B6C5AFULL, "darktown_fem_lowlife_mob6174167.cl"},
	{0x4D496DDBFAA45C91ULL, "6167544_m.fxe"},
	{0x4D49DC98E213EF0EULL, "gen00fl_fenris6115340.cl"},
	{0x4D4A995D8E8D267FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7side.phy"},
	{0x4D4AAE5DE4D3D0DEULL, "768227091.ogg"},
	{0x4D4ACF7A7751E761ULL, "dae000_arishok6067749.cl"},
	{0x4D4AD5447EF050C8ULL, "vxht_mansion2_32p_0_0.vlm"},
	{0x4D4AE1B296307A62ULL, "art\\characters\\creatures\\dragon\\animation\\tal_kckbk.evt"},
	{0x4D4B1169602211FBULL, "vxca_caverns1_mines_hvm_0_0.vlm"},
	{0x4D4B98970A5C8083ULL, "w_encounter2_camp_33_0.mao"},
	{0x4D4B9C970A5C87D0ULL, "w_encounter2_camp_33_0.mmh"},
	{0x4D4BAA970A5C9FEAULL, "w_encounter2_camp_33_0.msh"},
	{0x4D4BC1EA1E9211D1ULL, "6063550_m.fxe"},
	{0x4D4C3D0B31A6293AULL, "mag110st_vincento.utm"},
	{0x4D4CBF80D790B7A5ULL, "vxht_keep_6y5_0_0.vlm"},
	{0x4D4D26977A5B8B28ULL, "mag100_cullen6140940.cl"},
	{0x4D4D617586F96661ULL, "vxlt_warehouse_sg_0_0.vlm"},
	{0x4D4D75129AF99173ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_meda.phy"},
	{0x4D4E0C8E3C33A263ULL, "art\\vfx\\combat\\vfx_str_disapp_c.anb"},
	{0x4D4E32218578082BULL, "hlac_lt_0_lowtown_a3_n_c_65533_2z.dds"},
	{0x4D4E334DC22A06BBULL, "6139840_m.fxe"},
	{0x4D4E583F3C23AC56ULL, "136223382.ogg"},
	{0x4D4EDA5B2D078CE5ULL, "llac_lt_0_lowtown_docks_a3_n_131068_0.dds"},
	{0x4D4EF247DB70D42AULL, "zz_ave_debug6064100.cl"},
	{0x4D4F03BAF5E9A5DEULL, "6031749_m.fxe"},
	{0x4D4F15E3DC1DC81DULL, "gen00fl_fenris6105586.cl"},
	{0x4D4F6154D1EB55ACULL, "l_dae170ar_gallows.bnk"},
	{0x4D4F8B4C73C02413ULL, "146211765.ogg"},
	{0x4D4FC5652A29916FULL, "hlac_ht_0_hightown_a2_-65536_4z.dds"},
	{0x4D4FDE8315367936ULL, "474514625.ogg"},
	{0x4D501BE794D093ABULL, "6152762_m.fxe"},
	{0x4D5026649C4870CBULL, "144891044.ogg"},
	{0x4D507098CB3FBC02ULL, "gui_guipixelshader_p0.psh.fxo.dx11.assembly"},
	{0x4D50804273A9B18BULL, "6116797_m.fxe"},
	{0x4D50A2692890725BULL, "vxlt_gallowstemplar_n_5je_0_0.vlm"},
	{0x4D50E33EA7B53805ULL, "hlac_lt_0_lowtown_docks_-4_6z.dds"},
	{0x4D512C83A987941AULL, "228842956.ogg"},
	{0x4D5173CC5B74B444ULL, "vrc200im_ring_gift.uti"},
	{0x4D51AF4417534081ULL, "6110119_m.fxe"},
	{0x4D51E0B296369775ULL, "art\\characters\\creatures\\dragon\\animation\\tal_kckbk.gad"},
	{0x4D51E0D8F2135A6DULL, "dae000_meredith6085324.cl"},
	{0x4D528428FEECA09CULL, "pro000cr_captain_wright.utc"},
	{0x4D528FB8B3629632ULL, "237526711.ogg"},
	{0x4D5303D7417C4B8BULL, "hm_drunk2.mor"},
	{0x4D53C95D8149CAA8ULL, "6040660_m.fxe"},
	{0x4D53FA770DBA6760ULL, "hlac_lt_undercityencounter_-1_15x.dds"},
	{0x4D545D9B632B7505ULL, "mag210_arianni6057330.cl"},
	{0x4D54868FD65860EEULL, "6000950_m.fxe"},
	{0x4D55B31686A15AACULL, "hlac_w_blightlands_-131070_0z.dds"},
	{0x4D568B0E9A52AFA9ULL, "vxht_0_hightown_bow_0_0.vlm"},
	{0x4D56946446AE7C29ULL, "vxca_caverns2_8ps_0_0.vlm"},
	{0x4D56EA2B484E1248ULL, "vxlt_warehouse_sb_0_0.vlm"},
	{0x4D5709D7F48AAEF8ULL, "vxlt_backalley_g_1dq_0_0.vlm"},
	{0x4D57619F1D7C5A8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_reaction.ani"},
	{0x4D576FC6F39A9CCBULL, "vxdr_0_deeproads_cg7_0_0.vlm"},
	{0x4D57BD9CB1CF320CULL, "1007742412.ogg"},
	{0x4D57E9A68A260E8EULL, "374434418.ogg"},
	{0x4D593390821B5373ULL, "llac_lt_shop_-1_7.dds"},
	{0x4D5945CE92434282ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal_2.mao"},
	{0x4D5948943D89DD51ULL, "isa211_isabela6202174.cl"},
	{0x4D59584FA3B3837FULL, "498436571.ogg"},
	{0x4D596AA0213467FEULL, "6069376_f.fxe"},
	{0x4D59A8FC9B2E8D1DULL, "464975264.ogg"},
	{0x4D59C67282404420ULL, "fen310_varania6118639.cl"},
	{0x4D59CCBAD269394CULL, "6060378_m.fxe"},
	{0x4D59E58E8CF512C2ULL, "hlac_lt_backalley_n_0_1z.dds"},
	{0x4D5A1C4EEF8EF3FDULL, "484483527.ogg"},
	{0x4D5A678D5D89B735ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_stala1a.msh"},
	{0x4D5A698D5D89BAABULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_stala1a.mmh"},
	{0x4D5A758D5D89CF48ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_stala1a.mao"},
	{0x4D5AA7C610C45A3BULL, "art\\levels\\races\\design\\prp_ballista_0.pwk"},
	{0x4D5ABCA183317A46ULL, "397122514.ogg"},
	{0x4D5AE98E3C3DF08CULL, "art\\vfx\\combat\\vfx_str_disapp_c.mmh"},
	{0x4D5B315DC986D025ULL, "vxht_keep_6b4_0_0.vlm"},
	{0x4D5B9A5D8E9B9A84ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7side.mmh"},
	{0x4D5B9E5D8E9BA1D7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7side.mao"},
	{0x4D5BB05D8E9BC026ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7side.msh"},
	{0x4D5BD35278EF7EE1ULL, "vxlt_warehouse_tf_0_0.vlm"},
	{0x4D5C9AC799A82176ULL, "6086282_m.fxe"},
	{0x4D5CF0E1439A9143ULL, "gen_im_wep_war_one_act1_05.uti"},
	{0x4D5D263D8F34ABB0ULL, "6166887_m.fxe"},
	{0x4D5D37865D39ECB5ULL, "zz_mer_debug6017192.cl"},
	{0x4D5D3C7D71EF8DEBULL, "vxca_cave_19d_0_0.vlm"},
	{0x4D5D942527609619ULL, "613655339.ogg"},
	{0x4D5DB2043017CFE9ULL, "331614247.ogg"},
	{0x4D5DD1B28A71B4C1ULL, "983063758.ogg"},
	{0x4D5E51C9F815BCADULL, "vxht_mansion2_1av_0_0.vlm"},
	{0x4D5F31632A3BA1C1ULL, "6021251_m.fxe"},
	{0x4D5F976175F47DE2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_baserocks_1.mao"},
	{0x4D5FA30E29559986ULL, "vxlt_warehouse_n_9w_0_0.vlm"},
	{0x4D60219E1A4C2140ULL, "974145206.ogg"},
	{0x4D6107067874B901ULL, "character_character_p213.psh.fxo.dx11"},
	{0x4D61D95C7D67043CULL, "459629874.ogg"},
	{0x4D61FCB20810A883ULL, "6128971_f.fxe"},
	{0x4D62286F74028188ULL, "ser221ip_note_starkhaven.ncs"},
	{0x4D6270EE01E6CA89ULL, "18625696.ogg"},
	{0x4D6312B10D68F190ULL, "ts_prog_ruby.gda"},
	{0x4D633B426782BC29ULL, "907841071.ogg"},
	{0x4D63A6A92C2416FAULL, "vxlt_gallowstemplar_f_375_0_0.vlm"},
	{0x4D6441828BF317F3ULL, "m_dae321ar_keep.bnk"},
	{0x4D645D9124447494ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_mom_flemeth_real.ani"},
	{0x4D64E90F885FDB9FULL, "6180624_m.fxe"},
	{0x4D6547DA64C17C4DULL, "dre120pt_sandal.ncs"},
	{0x4D65499F1D886F02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_reaction.evt"},
	{0x4D65AE5106303207ULL, "6051874_m.fxe"},
	{0x4D66E54D25176740ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_viscdesk_d.dds"},
	{0x4D674CC6360679EFULL, "one220_yevhen6116324.cl"},
	{0x4D67847C99B5853DULL, "575149361.ogg"},
	{0x4D679209D62B64C9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunklo.mao"},
	{0x4D6799B5411DE695ULL, "vxw_bonepit_25f_0_0.vlm"},
	{0x4D679E09D62B7962ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunklo.mmh"},
	{0x4D67A009D62B7C18ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunklo.msh"},
	{0x4D67B38BF7E75099ULL, "vxlt_gallowstemplar_f_3ry_0_0.vlm"},
	{0x4D67B90D2A98198AULL, "dre000_varric6034679.cl"},
	{0x4D68656271F324FDULL, "302417512.ogg"},
	{0x4D68755DA67BA784ULL, "dae000_meredith6092208.cl"},
	{0x4D689A52A2A690AAULL, "vxlt_0_lowtown_docks_a3_b10_0_0.vlm"},
	{0x4D68CCD1A28095ABULL, "6083712_m.fxe"},
	{0x4D68F3FF55236FD9ULL, "vxht_brothel_n_57q_0_1.vlm"},
	{0x4D693D71B894045EULL, "character_character_p29.psh.fxo.dx11.assembly"},
	{0x4D6982E0A70D9BFDULL, "gen00fl_varric6104949.cl"},
	{0x4D699914631A8D8BULL, "gen_im_ammo_quiver_02.uti"},
	{0x4D69CA2649C217B1ULL, "280129584.ogg"},
	{0x4D69CAB5648B9EB4ULL, "551464041.ogg"},
	{0x4D69D56F032B6C24ULL, "vxlt_gallowscourtyard_n_79s_0_0.vlm"},
	{0x4D6A025187C09A20ULL, "422764937.ogg"},
	{0x4D6A452F6AE59761ULL, "vxlt_warehouse_ye_0_0.vlm"},
	{0x4D6A74DEB0C4C3A4ULL, "art\\vfx\\textures\\vfx_arc_cl.dds"},
	{0x4D6AB77C20AFF865ULL, "qcr000_varric6208225.cl"},
	{0x4D6B450B85D25C91ULL, "vxw_bonepit_1v8_0_0.vlm"},
	{0x4D6B5A91244A8EF6ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_mom_flemeth_real.gad"},
	{0x4D6B9CA0CEA23D57ULL, "571891556.ogg"},
	{0x4D6BAE5A63E5A66BULL, "vxca_cave_14g_0_0.vlm"},
	{0x4D6BE3A2B122154AULL, "991885334.ogg"},
	{0x4D6C489F1D8E8C15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_reaction.gad"},
	{0x4D6CC8B507E56493ULL, "218743693.ogg"},
	{0x4D6D37AAA17FABD5ULL, "vxlt_0_lowtown_a2_n_c_hla_0_0.vlm"},
	{0x4D6D430B8B20CDD8ULL, "105373955.ogg"},
	{0x4D6DAB6D6BC60FA9ULL, "vxlt_gallowstemplar_5ay_0_0.vlm"},
	{0x4D6DC1BBF3E932C5ULL, "vxca_caverns1_mines_i22_0_0.vlm"},
	{0x4D6DC992ACC7D779ULL, "vxw_blightlands_6p_0_0.vlm"},
	{0x4D6DE56778A93437ULL, "6136720_m.fxe"},
	{0x4D6F74B7043A5598ULL, "lt_0_lowtown_docks_n_9_0_lc.mmh"},
	{0x4D70AB3A171A6068ULL, "hlac_lt_0_lowtown_docks_n_-196610_5z.dds"},
	{0x4D70C0C5C3632B08ULL, "mag120_macha6080627.cl"},
	{0x4D70D112C752E519ULL, "vxlt_0_lowtown_docks_n_b80_0_0.vlm"},
	{0x4D710851CC2A882CULL, "501134053.ogg"},
	{0x4D713C71F89FB511ULL, "dae000_male_commoner1_mob6174462.cl"},
	{0x4D714F6D4607666BULL, "980726788.ogg"},
	{0x4D7168127FD9A663ULL, "6117475_m.fxe"},
	{0x4D71E04FCBC5E8C2ULL, "494356652.ogg"},
	{0x4D72D6A5B76E3980ULL, "chu_bl_clo.mao"},
	{0x4D72E1B772DE4E89ULL, "6018534_m.fxe"},
	{0x4D72E413721F0C1FULL, "6141887_m.fxe"},
	{0x4D72EEDD7A3E8117ULL, "6025161_m.fxe"},
	{0x4D73539FEFC37A35ULL, "lt_0_lowtown_docks_a3_19_0_lc.mmh"},
	{0x4D75083CF382377CULL, "6103037_m.fxe"},
	{0x4D7529898E03FBCBULL, "art\\characters\\playercharacter\\children\\animation\\blendtree\\child_paperdoll.anb"},
	{0x4D75731212C0A9D7ULL, "art\\vfx\\combat\\vfx_gib_corm_ltr.anb"},
	{0x4D75731212C0A9DCULL, "art\\vfx\\combat\\vfx_gib_corm_ltr.ani"},
	{0x4D75F2DE679D45B9ULL, "888666652.ogg"},
	{0x4D760B923313FF15ULL, "vxw_bonepit_20e_0_0.vlm"},
	{0x4D7629EB111B45BDULL, "742844404.ogg"},
	{0x4D7685181D3937B8ULL, "zz_fen_debug6134352.cl"},
	{0x4D768FB1B43203B5ULL, "6179433_m.fxe"},
	{0x4D7699BCA3266F91ULL, "vxw_woundedcoast_w5_0_0.vlm"},
	{0x4D76C1373AB877CAULL, "llac_lt_0_lowtown_docks_a3_n_65534_3.dds"},
	{0x4D76E7DD2D24D0E8ULL, "hlac_w_bonepitdestroyed_a3_0_13z.dds"},
	{0x4D770BAD7923DDA0ULL, "6096854_m.fxe"},
	{0x4D781517AD730125ULL, "vxht_keep_6h2_0_0.vlm"},
	{0x4D7853D1B845A112ULL, "vxlt_hangedman_n_248_0_0.vlm"},
	{0x4D79B53FAB4AA025ULL, "one300dg_solivitus6048318.cl"},
	{0x4D79CB31E1F02092ULL, "hlac_dr_dungeon_3216_9y.dds"},
	{0x4D7A8C6C82C0316AULL, "vxdr_primevalentrance_288_0_0.vlm"},
	{0x4D7AC609E39E5ABFULL, "vxlt_gallowscourtyard_n_65e_0_0.vlm"},
	{0x4D7AF9AFABD50205ULL, "4303470.ogg"},
	{0x4D7B401510B90B31ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v1.ani"},
	{0x4D7B801212C52B1EULL, "art\\vfx\\combat\\vfx_gib_corm_ltr.gad"},
	{0x4D7DDA1F41455706ULL, "hlac_lt_undercityencounter_-1_14y.dds"},
	{0x4D7E0574A051ED9CULL, "zz_personality_debug6004623.cl"},
	{0x4D7EB531F2088911ULL, "hlac_lt_gallowsprison_n_-1_2x.dds"},
	{0x4D7EB927A535B930ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_0x.dds"},
	{0x4D7EC81BE9264631ULL, "6071180_m.fxe"},
	{0x4D7F0EE50B81EDE6ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.impa.ani"},
	{0x4D7F42EFB948FD99ULL, "vxlt_0_lowtown_docks_n_b57_0_0.vlm"},
	{0x4D7F479B82EE91FBULL, "hlac_lt_darktown_-1_5z.dds"},
	{0x4D7F8933D40F9781ULL, "6071298_m.fxe"},
	{0x4D8023B7EC798939ULL, "gen00fl_merrill6121687.cl"},
	{0x4D803D5928C4CEAEULL, "hlac_w_encounter2_196606_0x.dds"},
	{0x4D805789C4A9E906ULL, "6148669_m.fxe"},
	{0x4D806E0EF4638302ULL, "vxdr_primevaltomb_7ej_0_0.vlm"},
	{0x4D807CA8D9583CE5ULL, "ch1notnt_ds2y.psh.fxo"},
	{0x4D808A63101047FAULL, "vxlt_gallowstemplar_f_397_0_0.vlm"},
	{0x4D815F8608CB5BEAULL, "vxht_0_hightown_a2_bzk_0_0.vlm"},
	{0x4D81BB3E9E5A2CB4ULL, "llac_lt_0_lowtown_a2_-65539_0.dds"},
	{0x4D81BC7AB6005352ULL, "vxw_bonepit_9a_0_0.vlm"},
	{0x4D81D2170414E52FULL, "6096516_m.fxe"},
	{0x4D824B1510BF3CD3ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v1.gad"},
	{0x4D824C8127CE2C46ULL, "758008363.ogg"},
	{0x4D8253BE29F0766DULL, "hlac_w_blightlands_65538_1z.dds"},
	{0x4D827DEED8E4996CULL, "376442143.ogg"},
	{0x4D82A4AB068CA33EULL, "6076201_m.fxe"},
	{0x4D831EDC9449DF17ULL, "128356284.ogg"},
	{0x4D839D496D30542BULL, "mag160_varian_ilithis6018894.cl"},
	{0x4D83C5BEF40CD291ULL, "675800191.ogg"},
	{0x4D83F4FCDC3D1B00ULL, "325103705.ogg"},
	{0x4D849C970A8C4C93ULL, "w_encounter2_camp_33_0.rml"},
	{0x4D84AD55618E63C4ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.dura.ani"},
	{0x4D8532DE45CCFFE1ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.cess.ani"},
	{0x4D85616970F26085ULL, "vxlt_backalley_g_1dh_0_0.vlm"},
	{0x4D8575B7C793F3CEULL, "6137075_f.fxe"},
	{0x4D8632759A592F0FULL, "6166189_m.fxe"},
	{0x4D86DA983134B80AULL, "6107151_f.fxe"},
	{0x4D86DCAC9C18BC8FULL, "plt_dae020pt_amb_hightown.plo"},
	{0x4D8723A97ED781E5ULL, "dae000_hate_enemies6167614.cl"},
	{0x4D8804DD9A3B82D5ULL, "one140cr_charade6134301.cl"},
	{0x4D88113564A76BDEULL, "vxdr_primevaltomb_2vw_0_1.vlm"},
	{0x4D88481510C3A338ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v1.evt"},
	{0x4D884FAB38424100ULL, "mrl251im_amu_harshal.uti"},
	{0x4D88915F04318BDDULL, "vxdr_primevalentrance_2l7_0_0.vlm"},
	{0x4D889C942BB884AFULL, "hlac_lt_0_lowtown_docks_a3_65533_4z.dds"},
	{0x4D88E1063378924CULL, "gen00fl_anders6125710.cl"},
	{0x4D89122F370C9014ULL, "docks_male_human_worker_sta6176071.cl"},
	{0x4D8A117CBDDDBAAEULL, "6143390_m.fxe"},
	{0x4D8A3D00FCE21DF9ULL, "6035944_m.fxe"},
	{0x4D8A6A56CE279A74ULL, "522065156.ogg"},
	{0x4D8A93A6F90AF8C7ULL, "813568266.ogg"},
	{0x4D8AB54F18556707ULL, "6044182_m.fxe"},
	{0x4D8AEBC3A9482FB8ULL, "dae000_arishok6067870.cl"},
	{0x4D8B2619BC3AA8D8ULL, "811192037.ogg"},
	{0x4D8B41716546F464ULL, "mag200_feynriel6030432.cl"},
	{0x4D8BA3970A9277EBULL, "w_encounter2_camp_33_0.phy"},
	{0x4D8BAA5561947D26ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.dura.gad"},
	{0x4D8BCFFF873CDA37ULL, "vxlt_0_lowtown_a3_n_c_gn7_0_0.vlm"},
	{0x4D8C24256AA73AC2ULL, "vxw_bonepitdestroyed_a3_1h5_0_0.vlm"},
	{0x4D8C3DDE45D33083ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.cess.gad"},
	{0x4D8C959FAC4582D2ULL, "562615825.ogg"},
	{0x4D8CE5CDCC09FF4CULL, "hlac_ht_0_hightown_n_2_1y.dds"},
	{0x4D8CF6E50B8E0257ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.impa.evt"},
	{0x4D8D210E93EA2BCCULL, "vxlt_0_lowtown_a3_n_c_emo_0_0.vlm"},
	{0x4D8E5F8D8A673C56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_enter.evt"},
	{0x4D8E73A41ACB6CBEULL, "hlac_dr_0_deeproads_5845_17x.dds"},
	{0x4D8E815F12C52FF3ULL, "6122875_m.fxe"},
	{0x4D8F058DC355BDF1ULL, "vxca_caverns2prc_4x6_0_0.vlm"},
	{0x4D8F2B49CF51B8D9ULL, "vxlt_gallowstemplar_f_6q9_0_0.vlm"},
	{0x4D8F409C65702225ULL, "6101135_m.fxe"},
	{0x4D901E203E96AB5AULL, "6090436_m.fxe"},
	{0x4D9034FC8AD2EA20ULL, "6088703_m.fxe"},
	{0x4D9051A17733D285ULL, "6143821_m.fxe"},
	{0x4D906B54D222FDDEULL, "l_dae170ar_gallows.wbd"},
	{0x4D90B264A9EE01FEULL, "392110877.ogg"},
	{0x4D91A85BE6D2AD6CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar6_.gad"},
	{0x4D9204A79E3BC193ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy.ani"},
	{0x4D922368C371560CULL, "6081710_f.fxe"},
	{0x4D92697E8B308A45ULL, "vxht_0_hightown_n_9wg_0_0.vlm"},
	{0x4D929D91DEB029CFULL, "955429786.ogg"},
	{0x4D92A555619A9305ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.dura.evt"},
	{0x4D92A74812436EFDULL, "vxlt_darktown_5q9_0_0.vlm"},
	{0x4D931ADE45D91468ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.cess.evt"},
	{0x4D935F23C774C29CULL, "6095154_m.fxe"},
	{0x4D93A33FD238056BULL, "vxlt_undercitydungeon_15t_0_0.vlm"},
	{0x4D93A75360EE7D9DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsblock.phy"},
	{0x4D93AE6D0A455741ULL, "vxlt_0_lowtown_a2_dbb_0_0.vlm"},
	{0x4D93CA121B371992ULL, "vxht_bartrand_n_4c2_0_0.vlm"},
	{0x4D93D725630985CEULL, "ef_merril.mor"},
	{0x4D93F7E50B9422E0ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.impa.gad"},
	{0x4D9456C9D68D109EULL, "vxlt_warehouse_n_14k_0_0.vlm"},
	{0x4D9465B4173E347CULL, "vxlt_0_lowtown_docks_n_8rm_0_0.vlm"},
	{0x4D9474ABC5F0E4F4ULL, "524919217.ogg"},
	{0x4D94BA0BACC52A08ULL, "dae000_meredith6085381.cl"},
	{0x4D95568D8A6D4BE1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_enter.gad"},
	{0x4D96B37BDAF8C891ULL, "one300ip_st_ilen.utp"},
	{0x4D975365427E37AFULL, "vxca_caverns2prc_7tg_0_0.vlm"},
	{0x4D979AA2B81CEC9FULL, "hlac_lt_0_lowtown_docks_n_65535_0z.dds"},
	{0x4D9804CCB6869F9AULL, "hlac_ht_0_hightown_-65537_1y.dds"},
	{0x4D980AF9C2ED8F76ULL, "vxca_caverns1_i6p_0_0.vlm"},
	{0x4D981C5B2D88BE14ULL, "6115346_m.fxe"},
	{0x4D989FBBA3B0307EULL, "hlac_lt_gallowstemplar_2392_1y.dds"},
	{0x4D98A75BE6D8CA6AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar6_.ani"},
	{0x4D98D9E5503D8A80ULL, "zz_followers_banter6157949.cl"},
	{0x4D991803004D20E5ULL, "vxw_bonepit_256_0_0.vlm"},
	{0x4D991C72F6173DB1ULL, "53557329.ogg"},
	{0x4D994045762270E7ULL, "vxlt_0_lowtown_docks_8uh_0_0.vlm"},
	{0x4D994E3721F86967ULL, "vxlt_0_lowtown_docks_a3_ctg_0_0.vlm"},
	{0x4D99B9575D946622ULL, "vxht_0_hightown_a3_c4a_0_0.vlm"},
	{0x4D99FFF66E59459AULL, "256846384.ogg"},
	{0x4D9A4DD0C9CFD66AULL, "vxht_0_hightown_n_a74_0_0.vlm"},
	{0x4D9A8104659E5D7DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shortbed_d.dds"},
	{0x4D9B3D81EA8DA579ULL, "425589114.ogg"},
	{0x4D9B578D8A71B88FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\overlay_enter.ani"},
	{0x4D9BB67A248D639FULL, "fxa_evl_circle_dis.mao"},
	{0x4D9C4EE2B0E9391BULL, "691794920.ogg"},
	{0x4D9C532F786D0CE5ULL, "6044002_f.fxe"},
	{0x4D9CA41E30D77431ULL, "sse_ss_athenril.gda"},
	{0x4D9D0141C81D520CULL, "hlac_w_bonepit_327682_0z.dds"},
	{0x4D9D04A20B77456AULL, "200300746.ogg"},
	{0x4D9D412AD907812BULL, "dae000_tomwise6167273.cl"},
	{0x4D9D57D897DD804FULL, "follower_banter6161118.cl"},
	{0x4D9DF3232E28CD8DULL, "vxht_playershouse_n_1jy_0_0.vlm"},
	{0x4D9E33C023D971ACULL, "194757465.ogg"},
	{0x4D9EF653FC08551DULL, "hm_vis_lowtown1.mor"},
	{0x4D9F0CD964182DA1ULL, "dae300_cullen6135881.cl"},
	{0x4D9F4FD35B730509ULL, "vxlt_hangedman_25w_0_0.vlm"},
	{0x4D9FFCA79E47F1CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy.evt"},
	{0x4DA06878E5F19C66ULL, "122212381.ogg"},
	{0x4DA07CBD5E9828C8ULL, "dre000_varric6035160.cl"},
	{0x4DA07DDD7E0F8BFDULL, "framebuffereffects_motionblur_p0.psh.fxo"},
	{0x4DA0DB5B7D26A2C5ULL, "vxht_0_hightown_n_9xf_0_0.vlm"},
	{0x4DA104327C26961BULL, "mcr370_varric6098327.cl"},
	{0x4DA15358CE7874A3ULL, "1028121194.ogg"},
	{0x4DA1755CB893925AULL, "vxlt_backalley_n_1ee_0_0.vlm"},
	{0x4DA187E405BBE0DFULL, "lgt000pt_tombs_second.ncs"},
	{0x4DA194F3ACE574D2ULL, "w_bonepitmining_a2_20_0_lc.mmh"},
	{0x4DA1CF9E85BD6756ULL, "vxlt_gallowsprison_n_6he_0_0.vlm"},
	{0x4DA2151CC42E1DEBULL, "vxlt_undercitydungeon_18w_0_0.vlm"},
	{0x4DA2318B87958EB1ULL, "vxw_encounter2_jc_0_0.vlm"},
	{0x4DA231CC256757E8ULL, "vxw_woundedcoast_hm_0_0.vlm"},
	{0x4DA23BEF0D2D3212ULL, "vxht_bartrand_n_4h3_0_0.vlm"},
	{0x4DA26C907683178CULL, "w_blightlands_36_0.msh"},
	{0x4DA272907683227EULL, "w_blightlands_36_0.mmh"},
	{0x4DA27E9076833615ULL, "w_blightlands_36_0.mao"},
	{0x4DA2936253C8EC12ULL, "6143048_m.fxe"},
	{0x4DA32C755BBD607AULL, "t1_mue_te2.tnt"},
	{0x4DA4071C3746857BULL, "6152612_m.fxe"},
	{0x4DA41D39D701CC29ULL, "380801724.ogg"},
	{0x4DA41F6320E05C13ULL, "vxca_caverns2_7gc_0_0.vlm"},
	{0x4DA49C5360FCDE09ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsblock.mao"},
	{0x4DA4A85360FCF2A2ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsblock.mmh"},
	{0x4DA4AA5360FCF658ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsblock.msh"},
	{0x4DA4AF4A52B2F256ULL, "gen00fl_fenris6107181.cl"},
	{0x4DA50B331DCBDB34ULL, "vxlt_backalley_n_1yg_0_0.vlm"},
	{0x4DA51643DC16FC43ULL, "6119882_m.fxe"},
	{0x4DA526DCE6FA9E59ULL, "hlac_w_bonepit_0_4y.dds"},
	{0x4DA526FAAE5CE380ULL, "w_encounter2_5_0_ll.mmh"},
	{0x4DA58BB4C2F76DBCULL, "6203021_m.fxe"},
	{0x4DA6CCB39630B421ULL, "hlac_w_bonepitdestroyed_a3_0_0y.dds"},
	{0x4DA6D1BBC028C870ULL, "zz_followers_banter6157259.cl"},
	{0x4DA6DBA79E4DD83DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy.gad"},
	{0x4DA6F927F9DF2C28ULL, "w_sundermount_175_0.msh"},
	{0x4DA70B27F9DF4AB9ULL, "w_sundermount_175_0.mao"},
	{0x4DA71727F9DF5F52ULL, "w_sundermount_175_0.mmh"},
	{0x4DA71E9785E4915AULL, "art\\characters\\creatures\\shriek\\animation\\combat\\heavydamage.gad"},
	{0x4DA72D24B7A5D8C5ULL, "373731562.ogg"},
	{0x4DA737DEFEFE2061ULL, "vxlt_blackemporium_10g_0_0.vlm"},
	{0x4DA789DFF2433965ULL, "vxw_bonepit_205_0_0.vlm"},
	{0x4DA7C177C3C63D17ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04a.phy"},
	{0x4DA7C45880A728FEULL, "hlac_lt_0_lowtown_docks_a3_0_4y.dds"},
	{0x4DA7D8F845A3E8D5ULL, "llac_lt_0_lowtown_a2_131071_5.dds"},
	{0x4DA83B67233B68BFULL, "static_static_p11.vsh.fxo.dx11"},
	{0x4DA8B87B2796EBA2ULL, "llac_lt_0_lowtown_a2_n_c_65536_12.dds"},
	{0x4DAA17DD91A82862ULL, "hlac_lt_0_lowtown_a2_n_65538_0y.dds"},
	{0x4DAAC470A5D0FC76ULL, "6174406_m.fxe"},
	{0x4DAAC9C30D12BC41ULL, "llac_lt_0_lowtown_n_65537_7.dds"},
	{0x4DAADBFFF8A319A9ULL, "754268931.ogg"},
	{0x4DAB20A5D9A49A0EULL, "m_dre180ar_ruins_crypt.wbd"},
	{0x4DABA134EA7B6FBAULL, "6046879_m.fxe"},
	{0x4DABC3A5CA45CC82ULL, "vxdr_primevaltomb_7tu_0_0.vlm"},
	{0x4DAC0DD41D6BB05DULL, "357204901.ogg"},
	{0x4DAC58D6DD8A0D88ULL, "zz_and_debug6134107.cl"},
	{0x4DAC5F828C310051ULL, "m_dae321ar_keep.wbd"},
	{0x4DAD2C9AD4734072ULL, "hlac_lt_0_lowtown_a3_n_c_-2_6x.dds"},
	{0x4DADB3DF2E8C15E3ULL, "6077034_m.fxe"},
	{0x4DAE219785EAB571ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\heavydamage.evt"},
	{0x4DAE27D08B36A231ULL, "31702567.ogg"},
	{0x4DAE4FD2CD377D93ULL, "vxlt_0_lowtown_h20_0_0.vlm"},
	{0x4DAE73B6D03BDAF9ULL, "gen00fl_anders6103757.cl"},
	{0x4DAE81A19975E58DULL, "vxlt_0_lowtown_docks_a3_c1f_0_0.vlm"},
	{0x4DAEA3417F2C622FULL, "6131145_m.fxe"},
	{0x4DAF4DB903A14964ULL, "hlac_lt_0_lowtown_a3_n_c_1_1y.dds"},
	{0x4DAF6E129B4C4FE8ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_meda.mmh"},
	{0x4DAF7A129B4C638BULL, "art\\characters\\playercharacter\\humanfemale\\bdy_meda.mao"},
	{0x4DAF85C16653EBBBULL, "vxlt_0_lowtown_n_e31_0_0.vlm"},
	{0x4DAF8C129B4C8212ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_meda.msh"},
	{0x4DB068CB82E68B25ULL, "655282643.ogg"},
	{0x4DB06E104CA9F208ULL, "w_woundedcoast_70_0.msh"},
	{0x4DB06FDF96CDE00DULL, "vxw_bonepitmining_a2_1p4_0_0.vlm"},
	{0x4DB080104CAA1099ULL, "w_woundedcoast_70_0.mao"},
	{0x4DB08C104CAA2532ULL, "w_woundedcoast_70_0.mmh"},
	{0x4DB08D5551B8DB4AULL, "lt_0_lowtown_24_0.rml"},
	{0x4DB0A3A9175D7068ULL, "vxw_woundedcoast_ej_0_0.vlm"},
	{0x4DB0C094C052CC48ULL, "vxlt_blackemporium_10b_0_0.vlm"},
	{0x4DB0CE8AE99159C8ULL, "800776431.ogg"},
	{0x4DB1006DC78B46B0ULL, "vxlt_0_lowtown_n_by3_0_0.vlm"},
	{0x4DB11A1E29513D51ULL, "vxca_cave_2d6_0_0.vlm"},
	{0x4DB149D0F378B267ULL, "vxlt_0_lowtown_n_bz5_0_0.vlm"},
	{0x4DB1D98E2558F5C8ULL, "vxw_encounter3_8s_0_0.vlm"},
	{0x4DB1E882178C6D56ULL, "6073890_f.fxe"},
	{0x4DB24F50A91583C6ULL, "vxht_chantry_2ch_0_0.vlm"},
	{0x4DB3055B8EA5C72EULL, "qun120_arishok6049624.cl"},
	{0x4DB34196BFEE2213ULL, "6214710_m.fxe"},
	{0x4DB3652906A385EEULL, "hlac_lt_undercityserialkiller_-1_39z.dds"},
	{0x4DB37DEF9C63DF2EULL, "gen00fl_varric6075154.cl"},
	{0x4DB37E5C3FA0D711ULL, "vxw_bonepit_1o3_0_0.vlm"},
	{0x4DB3E69FEE374343ULL, "6041487_f.fxe"},
	{0x4DB422DC1A635395ULL, "llac_ht_keep_4519_9.dds"},
	{0x4DB4543B07B1CDC1ULL, "hlac_lt_gallowstemplar_n_2388_4y.dds"},
	{0x4DB4628E56661755ULL, "6037119_m.fxe"},
	{0x4DB480C3800C6B0AULL, "vxlt_0_lowtown_docks_aon_0_0.vlm"},
	{0x4DB560BD4748F2C3ULL, "hlac_dr_0_deeproads_5842_12y.dds"},
	{0x4DB570FB5B4E2655ULL, "vxlt_0_lowtown_a2_n_c_hsl_0_0.vlm"},
	{0x4DB5BEFD960F5093ULL, "564952497.ogg"},
	{0x4DB5CBC813D2A737ULL, "lgt110cr_gxa_vgk_slvr_or_gd.utc"},
	{0x4DB5E83D5AC00D2FULL, "671853157.ogg"},
	{0x4DB5FB182D0C2688ULL, "vxlt_0_lowtown_docks_n_95i_0_0.vlm"},
	{0x4DB6509F2DD698DCULL, "art\\characters\\creatures\\mubari\\animation\\com_submit.ani"},
	{0x4DB7E81710C1F8F8ULL, "6030493_f.fxe"},
	{0x4DB7EDA77BF7889CULL, "vxht_mansion_n_4h6_0_0.vlm"},
	{0x4DB7F12D10C58E3EULL, "fen210_tevinter_ambusher6108311.cl"},
	{0x4DB8391F8622D92CULL, "vxlt_0_lowtown_docks_n_aq9_0_0.vlm"},
	{0x4DB903AF07DBB100ULL, "vxlt_0_lowtown_exl_0_0.vlm"},
	{0x4DB9144A70E3B3E4ULL, "plt_lgt220pt_shp_mag_invst.nss"},
	{0x4DB92F08CEC34FDBULL, "zz_dae_debug6029319.cl"},
	{0x4DB965DFD8373BC7ULL, "464351929.ogg"},
	{0x4DB97A24EDE8D375ULL, "art\\levels\\races\\design\\_textures\\plc_metaltrp010_0_n.dds"},
	{0x4DB9A7D602BAE807ULL, "ser121_jethann6108786.cl"},
	{0x4DBA2B9597D9E25FULL, "6053928_m.fxe"},
	{0x4DBA3582BC3BE502ULL, "vxdr_primevaltomb_7qv_0_0.vlm"},
	{0x4DBA419C65647EF1ULL, "llac_dr_dungeon_3217_6.dds"},
	{0x4DBA5E435FED0083ULL, "28182042.ogg"},
	{0x4DBA7C7E874E0757ULL, "992653553.ogg"},
	{0x4DBB103E5527C1C4ULL, "vxlt_gallowstemplar_f_3qq_0_0.vlm"},
	{0x4DBB399785F56840ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\heavydamage.ani"},
	{0x4DBC5D9F2DDB1A1EULL, "art\\characters\\creatures\\mubari\\animation\\com_submit.gad"},
	{0x4DBDA62C9F15AADFULL, "vxca_caverns2_5a6_0_1.vlm"},
	{0x4DBDFFAC629E18A5ULL, "vxlt_0_lowtown_e0u_0_0.vlm"},
	{0x4DBE0A8C733D39A3ULL, "dae000ip_crowd_mf1_talk2.utp"},
	{0x4DBE3D4A496913FCULL, "gen00fl_aveline_desk.stg"},
	{0x4DBE59169C7FC35AULL, "vxlt_backalley_n_1kg_0_0.vlm"},
	{0x4DBF15456B81BFB1ULL, "vxw_encounter2_pa_0_0.vlm"},
	{0x4DBF1586095388E8ULL, "vxw_woundedcoast_bk_0_0.vlm"},
	{0x4DBF9439FEF92FBCULL, "clf200pt_cliff_raiders.ncs"},
	{0x4DBFAE176D3CAD61ULL, "6095700_m.fxe"},
	{0x4DC040836BA31EB3ULL, "vxdr_primevaltomb_73i_0_0.vlm"},
	{0x4DC12A0C2A73BB94ULL, "hlac_w_woundedcoast_-262140_0y.dds"},
	{0x4DC132EB5DF91AE0ULL, "6132899_m.fxe"},
	{0x4DC171AC47B92F5FULL, "mrl251ip_dead_chandan.utp"},
	{0x4DC185F4DCF2535BULL, "1062319005.ogg"},
	{0x4DC1A04D01EB56BCULL, "vxlt_0_lowtown_docks_n_b5n_0_0.vlm"},
	{0x4DC1D42CB077518CULL, "927819345.ogg"},
	{0x4DC219D6ABC0938FULL, "184646114.ogg"},
	{0x4DC22719C42CADB0ULL, "vxlt_0_lowtown_h8y_0_0.vlm"},
	{0x4DC316A1532D668EULL, "gen00fl_isabela6096773.cl"},
	{0x4DC32D167024D011ULL, "lt_0_lowtown_a2_n_c_20_0_ll.mmh"},
	{0x4DC3389F2DE0FA3DULL, "art\\characters\\creatures\\mubari\\animation\\com_submit.evt"},
	{0x4DC35070A01114F9ULL, "dae000_gamlen6149480.cl"},
	{0x4DC35750A514627BULL, "6146910_m.fxe"},
	{0x4DC40E3C6B0CE975ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsblock.phy"},
	{0x4DC431AE57EC9D32ULL, "6170002_m.fxe"},
	{0x4DC490E311986636ULL, "vxlt_warehouse_n_150_0_0.vlm"},
	{0x4DC4B3D865DEAABCULL, "6016322_m.fxe"},
	{0x4DC4ED927C23C9C2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magc_0s.dds"},
	{0x4DC5B313DE594C6DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbaseb_n.dds"},
	{0x4DC5C69661F9291EULL, "vxdr_primevaltomb_6pu_0_1.vlm"},
	{0x4DC5ED05BF2D0CDDULL, "379969930.ogg"},
	{0x4DC6697FF03E54B4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shielda_sa.dds"},
	{0x4DC68A6800663FD0ULL, "llac_ht_playershouse_n_855_16.dds"},
	{0x4DC6E4CA9EB0CE91ULL, "vxht_0_hightown_a2_cax_0_0.vlm"},
	{0x4DC75967E88C80EEULL, "qun221_petrice6083700.cl"},
	{0x4DC776A437C2CE9BULL, "hlac_lt_0_lowtown_a2_n_c_65537_2y.dds"},
	{0x4DC782580ABF3A13ULL, "one320cr_magistrate.utc"},
	{0x4DC7EA6370C5FB70ULL, "vxw_sundermount_31u_0_0.vlm"},
	{0x4DC80228DCE82657ULL, "one280_emrys6120045.cl"},
	{0x4DC874DF8DFBC7E2ULL, "vxlt_0_lowtown_n_e8i_0_0.vlm"},
	{0x4DC8777A784F52BDULL, "hlac_w_bonepitmining_a2_0_4y.dds"},
	{0x4DC8F575C32C4AF7ULL, "mcr370_negotiation6080281.cl"},
	{0x4DC91027939CC77CULL, "vxht_mansion_n_4y5_0_0.vlm"},
	{0x4DC966107D79830DULL, "follower_banter6156110.cl"},
	{0x4DC9E8C3A4101EB4ULL, "eye_eye_p1.psh.fxo"},
	{0x4DCA02E87744F2F5ULL, "lgt000cr_mage_turn_in_2.utc"},
	{0x4DCA3082F570B2FCULL, "vxht_0_hightown_bri_0_0.vlm"},
	{0x4DCA3AEC957A3105ULL, "13650013.ogg"},
	{0x4DCADFD49D2EA299ULL, "689962122.ogg"},
	{0x4DCB241AF5AC7781ULL, "488520390.ogg"},
	{0x4DCBF9EB7389C06CULL, "vxht_bartrand_n_643_0_0.vlm"},
	{0x4DCC1CB20BE02C7FULL, "dae000_hate_enemies6167609.cl"},
	{0x4DCC986BA0C6CB4DULL, "66645617.ogg"},
	{0x4DCCBCA24159B06BULL, "905700796.ogg"},
	{0x4DCCD43CAEF793AEULL, "vxw_sundermount_309_0_1.vlm"},
	{0x4DCD437C0261F455ULL, "qun221_petrice6077355.cl"},
	{0x4DCD72918F78970CULL, "1027695462.ogg"},
	{0x4DCDA735306B720EULL, "vxht_keep_n_49c_0_0.vlm"},
	{0x4DCE6B54B98F8936ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_4.ani"},
	{0x4DCEC32DF018EBDEULL, "hlac_ht_keep_3733_1z.dds"},
	{0x4DCF330A8D01B4C6ULL, "vxht_chantry_2yj_0_0.vlm"},
	{0x4DCF498F91121AD0ULL, "6153227_m.fxe"},
	{0x4DD0923E74ACB773ULL, "6062124_m.fxe"},
	{0x4DD09C241EBDE181ULL, "vxht_chantry_n_209_0_0.vlm"},
	{0x4DD0F5B7989358C6ULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_fin_c.impa.ani"},
	{0x4DD101B3913DC916ULL, "872842003.ogg"},
	{0x4DD1958B01AF9F19ULL, "vrc321_housekeeper.stg"},
	{0x4DD1C17CE42662A8ULL, "hlac_lt_blackemporium_1055_4x.dds"},
	{0x4DD1FBAC9C58DD74ULL, "plt_dae020pt_amb_hightown.nss"},
	{0x4DD2017A1B444CEAULL, "vxca_caverns1_mines_icl_0_2.vlm"},
	{0x4DD20776AE202FE1ULL, "vxlt_0_lowtown_a2_erw_0_0.vlm"},
	{0x4DD23E71F6025919ULL, "62924099.ogg"},
	{0x4DD26C0B5012098FULL, "6144570_m.fxe"},
	{0x4DD2E568E496C05DULL, "6090444_m.fxe"},
	{0x4DD2FDEC73B1F328ULL, "vxlt_ruinedhouse_n_1w_0_0.vlm"},
	{0x4DD3191B904A9396ULL, "885745148.ogg"},
	{0x4DD39C7F90C1229AULL, "636254989.ogg"},
	{0x4DD48DD653FFBD77ULL, "387838938.ogg"},
	{0x4DD4AA9816317DD1ULL, "927630246.ogg"},
	{0x4DD4D1615FE3B99CULL, "vxht_mansion_n_4b4_0_0.vlm"},
	{0x4DD5418A48875936ULL, "46000693.ogg"},
	{0x4DD5D3442B9ED8D6ULL, "hlac_w_blightlands_393213_0z.dds"},
	{0x4DD68E0125DE8C86ULL, "gen_im_trs_qun_lgt_glv_r0.uti"},
	{0x4DD6C4F83724FAFAULL, "hlac_ca_caverns1_mines_24427_0y.dds"},
	{0x4DD716A5D9C9C43CULL, "m_dre180ar_ruins_crypt.bnk"},
	{0x4DD73212B30835BDULL, "vxlt_0_lowtown_f06_0_0.vlm"},
	{0x4DD79E36B7E1C1FAULL, "qcr220_orsino6061813.cl"},
	{0x4DD8887D3F257C1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverexit.gad"},
	{0x4DD8B32780D8D47EULL, "vxht_mansion2_1rt_0_0.vlm"},
	{0x4DD904F4C79F60BCULL, "vxlt_0_lowtown_docks_a3_n_cvt_0_1.vlm"},
	{0x4DD905BE07C589AAULL, "701063443.ogg"},
	{0x4DD9201893C0EA99ULL, "vxlt_gallowstemplar_n_3nu_0_0.vlm"},
	{0x4DD923DFD0A9FFFDULL, "vxw_woundedcoast_11r_0_0.vlm"},
	{0x4DD9700215D7D9F1ULL, "fen310_fenris6118240.cl"},
	{0x4DD9EDEB1C9A8875ULL, "vxlt_undercitydungeon_1if_0_0.vlm"},
	{0x4DDA2C8463156C46ULL, "llac_lt_0_lowtown_a3_n_c_65536_0.dds"},
	{0x4DDA705D818281DAULL, "6085100_m.fxe"},
	{0x4DDAC9550BEFE6CBULL, "6081800_m.fxe"},
	{0x4DDB154A71009BDFULL, "plt_lgt220pt_shp_mag_invst.plo"},
	{0x4DDB523913D2E8E1ULL, "t1_stb_dbr.tnt"},
	{0x4DDBF93FED3FB9E8ULL, "vxw_woundedcoast_ti_0_0.vlm"},
	{0x4DDC1DBF0C83EB5DULL, "lt_0_lowtown_a2_n_c_46_0.rml"},
	{0x4DDC1E4263157660ULL, "vxht_mansion2b_2ze_0_0.vlm"},
	{0x4DDC7354B99BD427ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_4.evt"},
	{0x4DDCD2AF55CE6028ULL, "6177676_m.fxe"},
	{0x4DDD2B5CDA4D3F31ULL, "w_encounter1_46_0.mmh"},
	{0x4DDD375CDA4D53D2ULL, "w_encounter1_46_0.mao"},
	{0x4DDD3C45DAAA8E63ULL, "6161544_m.fxe"},
	{0x4DDD455CDA4D6BEFULL, "w_encounter1_46_0.msh"},
	{0x4DDD9DC8A7858CACULL, "vxlt_0_lowtown_a2_n_ei4_0_0.vlm"},
	{0x4DDDA1E17DCB786CULL, "6119445_m.fxe"},
	{0x4DDDCE7984597400ULL, "vxca_cave_1bn_0_0.vlm"},
	{0x4DDE9A8CA5EE0D6DULL, "zz_mer_debug6017189.cl"},
	{0x4DDEDDB7989F6D37ULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_fin_c.impa.evt"},
	{0x4DDF385FE20424FBULL, "276613725.ogg"},
	{0x4DDF7D7D3F2B880CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverexit.evt"},
	{0x4DE07953A0164861ULL, "vxlt_0_lowtown_a2_euv_0_0.vlm"},
	{0x4DE0A1B5E2629853ULL, "115808422.ogg"},
	{0x4DE0AD4093A49A42ULL, "vxdr_primevaltomb_8p6_0_0.vlm"},
	{0x4DE1051E123466F9ULL, "vxw_encounter2_4e_0_0.vlm"},
	{0x4DE13F340C6915DBULL, "115473327.ogg"},
	{0x4DE15E0D36E9DDDDULL, "857701861.ogg"},
	{0x4DE1808CD9C1363AULL, "gen00fl_fenris6113922.cl"},
	{0x4DE19342DF5ACEC7ULL, "83207892.ogg"},
	{0x4DE1A6C1B6D2A848ULL, "mcr370_templar26087010.cl"},
	{0x4DE1F01394A54504ULL, "vxdr_primevaltomb_34g_0_0.vlm"},
	{0x4DE29454B9A07870ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_4.gad"},
	{0x4DE3623BD8A070F8ULL, "vxw_bonepitmining_a2_26x_0_0.vlm"},
	{0x4DE3634ACE2AFD3CULL, "vxca_caverns2_969_0_0.vlm"},
	{0x4DE46653F82A9589ULL, "6066283_m.fxe"},
	{0x4DE4E79D79405E61ULL, "mcr310_party_banter6087796.cl"},
	{0x4DE4E862E60EE870ULL, "845159505.ogg"},
	{0x4DE5363DC81465DEULL, "6131131_m.fxe"},
	{0x4DE591AA9BFC8722ULL, "gen_im_acc_blt_lore_01.uti"},
	{0x4DE5A3EFA4FE4E3DULL, "vxlt_0_lowtown_f55_0_0.vlm"},
	{0x4DE5B789E85B2115ULL, "vxw_blightlands_11p_0_0.vlm"},
	{0x4DE5CA60B4AA3F33ULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_gen_c.anb"},
	{0x4DE5DEB798A58DC0ULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_fin_c.impa.gad"},
	{0x4DE62C73E7F7480DULL, "vxht_playershouse_n_1ut_0_0.vlm"},
	{0x4DE639CB773243F2ULL, "299876326.ogg"},
	{0x4DE6E39D52842ACAULL, "6094722_m.fxe"},
	{0x4DE7143CD95CE3FCULL, "vxht_0_hightown_bxo_0_0.vlm"},
	{0x4DE7250472CEECFEULL, "vxht_mansion2_1uu_0_0.vlm"},
	{0x4DE73ABE99CFE289ULL, "dae221ip_home_amb_letters.utp"},
	{0x4DE846589AEAA447ULL, "vxw_bonepitmining_a2_5b_0_0.vlm"},
	{0x4DE85FC80E90A0F5ULL, "vxlt_undercitydungeon_1le_0_0.vlm"},
	{0x4DE8770A14ADA290ULL, "aip_rubblegolemb_base.gda"},
	{0x4DE88DBF41AE7C38ULL, "follower_banter6170430.cl"},
	{0x4DE8F6E69B99AB63ULL, "6153563_m.fxe"},
	{0x4DE985F6F244396CULL, "247803475.ogg"},
	{0x4DE9973FA9BFA0F1ULL, "vxlt_0_lowtown_a3_n_c_e3d_0_0.vlm"},
	{0x4DE9A9E05C81E680ULL, "gen00fl_anders6125453.cl"},
	{0x4DE9C7E65A1902F3ULL, "vxw_bonepitmining_a2_1ln_0_0.vlm"},
	{0x4DEA529076C0A135ULL, "w_blightlands_36_0.rml"},
	{0x4DEA6867A7B137A4ULL, "6054153_m.fxe"},
	{0x4DEA6B1CDF35D268ULL, "vxw_woundedcoast_qf_0_0.vlm"},
	{0x4DEB7A4E984E3EFAULL, "hlac_w_bonepit_458758_0y.dds"},
	{0x4DEBA5E9AA10BA07ULL, "461380481.ogg"},
	{0x4DEC0FA5997BA52CULL, "vxlt_0_lowtown_a2_n_er5_0_0.vlm"},
	{0x4DEC171FF75EF0ECULL, "104859137.ogg"},
	{0x4DEC2038070B547CULL, "6103266_m.fxe"},
	{0x4DECC91EEEA86D8EULL, "vxw_woundedcoast_fi_0_0.vlm"},
	{0x4DECF249F46B7157ULL, "vxlt_undercityencounter_13n_0_1.vlm"},
	{0x4DED07FE4EBA88D5ULL, "vxlt_warehouse_136_0_0.vlm"},
	{0x4DED167FE4D3F1B7ULL, "ser120_ghyslain6118424.cl"},
	{0x4DED473BF59C27EFULL, "dr_primevaltomb_12048.rml"},
	{0x4DED657D3F379D65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightrecoverexit.ani"},
	{0x4DEDBF75B6D73658ULL, "combat_ws_r1_a2.gda"},
	{0x4DEE54DCFB9AF539ULL, "vxht_0_hightown_a2_9p9_0_0.vlm"},
	{0x4DEEA277C4020B0CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04a.mmh"},
	{0x4DEEA877C40215FEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04a.msh"},
	{0x4DEF7F347CA9F443ULL, "dae000_dalish_apprentices6147749.cl"},
	{0x4DF00AFEEEFD78DBULL, "vxdr_0_deeproads_9ej_0_1.vlm"},
	{0x4DF01A4F39B69FD9ULL, "445556471.ogg"},
	{0x4DF05826154E4129ULL, "8240630.ogg"},
	{0x4DF06DAD01BA1F04ULL, "6166201_m.fxe"},
	{0x4DF0D7A5132DF5FCULL, "327094915.ogg"},
	{0x4DF16F6F13843E15ULL, "vxlt_backalley_n_1r7_0_0.vlm"},
	{0x4DF1F0B1169B886AULL, "vxlt_0_lowtown_a2_hba_0_0.vlm"},
	{0x4DF259DF6A94796AULL, "zz_and_debug6134116.cl"},
	{0x4DF2DC74767E5C1DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grand.phy"},
	{0x4DF2E6B6031949F9ULL, "hlac_lt_gallowscourtyard_-65538_4y.dds"},
	{0x4DF2E760B4B4FA9CULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_gen_c.mmh"},
	{0x4DF2F3FB54D884BEULL, "vxca_cavern1_haunted_j1z_0_1.vlm"},
	{0x4DF306B3B5F8A675ULL, "26640805.ogg"},
	{0x4DF3778089A1264EULL, "vxdr_primevaltomb_7lq_0_0.vlm"},
	{0x4DF4861145A78DB0ULL, "6005645_m.fxe"},
	{0x4DF4C74BE6074280ULL, "vxht_chantry_n_2t8_0_0.vlm"},
	{0x4DF4E04EBF2D0F88ULL, "6082675_m.fxe"},
	{0x4DF5398315446074ULL, "hlac_dr_dungeon_3217_2y.dds"},
	{0x4DF55B3F90B95EC4ULL, "133967270.ogg"},
	{0x4DF577B69522A87BULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnr.evt"},
	{0x4DF596E164C5057EULL, "vxht_mansion2_1xv_0_0.vlm"},
	{0x4DF5994110C938E2ULL, "6054731_m.fxe"},
	{0x4DF5C6D1D3F57DFCULL, "gen00fl_aveline6122046.cl"},
	{0x4DF6DB70DF3C9D99ULL, "679682731.ogg"},
	{0x4DF7D22E10AA0214ULL, "vxlt_gallowstemplar_f_5k5_0_0.vlm"},
	{0x4DF88967E53A2B73ULL, "192378250.ogg"},
	{0x4DF892614FE15110ULL, "vxht_0_hightown_n_c1v_0_0.vlm"},
	{0x4DF8C0BCDBD5D8DFULL, "hlac_lt_0_lowtown_a2_n_c_131071_4x.dds"},
	{0x4DF901FC4701A760ULL, "vxht_mansion2b_2pc_0_0.vlm"},
	{0x4DF930155D9E5631ULL, "vxw_blightlands_6h_0_0.vlm"},
	{0x4DF9793576128CDEULL, "vxlt_gallowscourtyard_n_74b_0_0.vlm"},
	{0x4DF9E7C3193370B1ULL, "861416319.ogg"},
	{0x4DFA0D8580BB90A5ULL, "1067322718.ogg"},
	{0x4DFA37C3FD34BD88ULL, "vxht_0_hightown_c0j_0_0.vlm"},
	{0x4DFAC50C039E04D5ULL, "follower_banter6143255.cl"},
	{0x4DFAD3D1794F1082ULL, "6143413_m.fxe"},
	{0x4DFAFE876336DF10ULL, "character_character_p98.psh.fxo.dx11"},
	{0x4DFB06D9DB7909E9ULL, "6140330_m.fxe"},
	{0x4DFB3AFBE09E860EULL, "vxw_woundedcoast_cj_0_0.vlm"},
	{0x4DFB780C46270F74ULL, "vxca_cave_1dm_0_1.vlm"},
	{0x4DFB79DB40B0A155ULL, "vxlt_warehouse_165_0_0.vlm"},
	{0x4DFBBB70786576B6ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_aoe_p.impa.evt"},
	{0x4DFC50A3065873C7ULL, "gen00fl_varric6080092.cl"},
	{0x4DFC6B832629FB52ULL, "6109125_m.fxe"},
	{0x4DFC70B69528BABCULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnr.gad"},
	{0x4DFC81E941A49240ULL, "111834978.ogg"},
	{0x4DFCDA8817F0F69EULL, "6080709_m.fxe"},
	{0x4DFD7A2329C0D86BULL, "vxlt_0_lowtown_a3_n_c_hss_0_0.vlm"},
	{0x4DFE2D11D91BDA4CULL, "6102812_m.fxe"},
	{0x4DFE6A3E77A9D606ULL, "hlac_dr_slavers_3214_13x.dds"},
	{0x4DFEA19921E4301FULL, "663942597.ogg"},
	{0x4DFEA60C3ECC9ADEULL, "59157319.ogg"},
	{0x4DFEFF5FB93962CAULL, "vxht_0_hightown_an4_0_0.vlm"},
	{0x4DFF3400156BDBF9ULL, "hla_ht_0_hightown_65535_0y.dds"},
	{0x4DFF4033D9BBDB7BULL, "vxht_bartrand_n_5v9_0_0.vlm"},
	{0x4DFF5B4FD70E06ACULL, "w_bonepit_26_0.msh"},
	{0x4DFF614FD70E109EULL, "w_bonepit_26_0.mmh"},
	{0x4DFF63F05EE972A0ULL, "vxdr_primevaltomb_6x2_0_0.vlm"},
	{0x4DFF6D4FD70E2535ULL, "w_bonepit_26_0.mao"},
	{0x4DFF75892D6E88F4ULL, "dae000_mother6006251.cl"},
	{0x4DFFB5424CFC4379ULL, "mag291_torpor6026971.cl"},
	{0x4DFFF28B096E89B3ULL, "971365046.ogg"},
	{0x4E00BC8E3D31563DULL, "860058666.ogg"},
	{0x4E016E2856F0FEC6ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_014a_0n.dds"},
	{0x4E01B611BBB68E3CULL, "lt_0_lowtown_docks_3_0_ll.mmh"},
	{0x4E01C3940545BD56ULL, "qun231_varnell6098594.cl"},
	{0x4E0208D82CE93D39ULL, "vxw_bonepit_8q_0_0.vlm"},
	{0x4E02B270786B8641ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_aoe_p.impa.gad"},
	{0x4E02DBF0BF37F535ULL, "ca_cavern1_haunted_24421.rml"},
	{0x4E02DCCE9A37274EULL, "w_encounter3_238_0.mmh"},
	{0x4E02E8CE9A373BE5ULL, "w_encounter3_238_0.mao"},
	{0x4E02F6CE9A37533CULL, "w_encounter3_238_0.msh"},
	{0x4E036FB6952ED7BAULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnr.ani"},
	{0x4E037066EED7D8EAULL, "6148157_m.fxe"},
	{0x4E03D174768CBC89ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grand.mao"},
	{0x4E03DD74768CD122ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grand.mmh"},
	{0x4E03DF74768CD4D8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grand.msh"},
	{0x4E0436FC997F54A0ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_brickround01_n.dds"},
	{0x4E05000083D9EE72ULL, "60688026.ogg"},
	{0x4E056CBB0116BB0EULL, "mag121_idunna6084186.cl"},
	{0x4E057F29051D1D48ULL, "zz_isa_debug6175461.cl"},
	{0x4E05C2565DBE0ADFULL, "50462046.ogg"},
	{0x4E075D482F94A866ULL, "6116644_m.fxe"},
	{0x4E07B65B0F4AA21BULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.impa.ani"},
	{0x4E08241D828FFD09ULL, "197827270.ogg"},
	{0x4E08370DBBDD1262ULL, "1025259462.ogg"},
	{0x4E087B8B01DEB5B6ULL, "vrc321_housekeeper.cnv"},
	{0x4E0893452EDFBE92ULL, "102051959.ogg"},
	{0x4E08A9A0EF2AD608ULL, "vxht_0_hightown_c7k_0_0.vlm"},
	{0x4E09B3707871A6EFULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_aoe_p.impa.ani"},
	{0x4E09B9244AA6C0BFULL, "vxlt_0_lowtown_a2_dow_0_0.vlm"},
	{0x4E0A378D5BD880FFULL, "hlac_w_encounter2_393214_0y.dds"},
	{0x4E0A7BD167854DC9ULL, "352722785.ogg"},
	{0x4E0A8CD56EE611D6ULL, "w_bonepitdestroyed_a3_9_0.rml"},
	{0x4E0AF9234155DD9AULL, "vxht_0_hightown_apz_0_0.vlm"},
	{0x4E0B213C6B490B90ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsblock.msh"},
	{0x4E0B2F3C6B4923AAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsblock.mmh"},
	{0x4E0B333C6B492AE1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsblock.mao"},
	{0x4E0B6677AA9C57DEULL, "vxlt_0_lowtown_docks_a3_n_d46_0_0.vlm"},
	{0x4E0B7D8B01E0EE96ULL, "vrc321_housekeeper.bnk"},
	{0x4E0B9C00AFDB6ADDULL, "106568869.ogg"},
	{0x4E0BF1003A024B72ULL, "mer221_sabin6048285.cl"},
	{0x4E0C91447C807719ULL, "hlac_lt_gamlenshouse_n_-1_7z.dds"},
	{0x4E0C9EEECB11F2CCULL, "6107712_m.fxe"},
	{0x4E0D23E7FE2DAA93ULL, "gen00fl_fenris6112056.cl"},
	{0x4E0D404FD71A165DULL, "w_bonepit_26_0.anb"},
	{0x4E0D53256F32E4FFULL, "6082034_m.fxe"},
	{0x4E0D97BA1CA95737ULL, "6055162_m.fxe"},
	{0x4E0DB210CBB1F3FBULL, "vxht_bartrand_n_5y8_0_0.vlm"},
	{0x4E0E2FBF0C2674E1ULL, "vxht_keep_n_6lm_0_0.vlm"},
	{0x4E0E5328F7706F15ULL, "vxlt_backalley_n_1x5_0_0.vlm"},
	{0x4E0F1CBF0CAF4375ULL, "lt_0_lowtown_a2_n_c_46_0.anb"},
	{0x4E0F1EE94B8DCFC2ULL, "small_d.dds"},
	{0x4E0F982D3C82C7F4ULL, "vxw_woundedcoast_1k9_0_0.vlm"},
	{0x4E0FAA75B70D94D2ULL, "vxht_chantry_n_36e_0_0.vlm"},
	{0x4E0FF7784B50FA50ULL, "6057321_m.fxe"},
	{0x4E1019AD9B3FC5E3ULL, "6087033_m.fxe"},
	{0x4E101C66E8A53810ULL, "1010065357.ogg"},
	{0x4E10270B119C5153ULL, "130490965.ogg"},
	{0x4E105B3A6D8D574EULL, "vxdr_primevaltomb_7bs_0_0.vlm"},
	{0x4E10C7D3E1AE5AA3ULL, "6018907_m.fxe"},
	{0x4E112303EE2CF9C3ULL, "lla_ht_0_hightown_131073_0.dds"},
	{0x4E1139B509B079FAULL, "gen00fl_fenris6127046.cl"},
	{0x4E115D8576153EC2ULL, "vxht_0_hightown_n_c5x_0_0.vlm"},
	{0x4E11631201E83006ULL, "367307017.ogg"},
	{0x4E11B9CD55082285ULL, "vxw_bonepitmining_a2_26a_0_0.vlm"},
	{0x4E121BDBFF8F6257ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue_d.dds"},
	{0x4E122B36A42B3A51ULL, "gen00fl_isabela6100053.cl"},
	{0x4E12514B03244C76ULL, "ave260ip_signal_fire2.ncs"},
	{0x4E127A9B48B1367EULL, "vxht_mansion2_1fh_0_0.vlm"},
	{0x4E12841921251FEAULL, "vxht_brothel_n_47f_0_0.vlm"},
	{0x4E12D40ECA33CB91ULL, "6077689_m.fxe"},
	{0x4E12EC1A86E10032ULL, "6112993_f.fxe"},
	{0x4E1326F5113DDB6BULL, "gen00fl_isabela6096833.cl"},
	{0x4E13436F15C0B5E2ULL, "zz_area_signage3.utp"},
	{0x4E134D605780C703ULL, "vxht_0_hightown_n_a2v_0_0.vlm"},
	{0x4E138C104CFEA2D9ULL, "w_woundedcoast_70_0.rml"},
	{0x4E13CA3285E78F13ULL, "llac_lt_gallowscourtyard_-65537_2.dds"},
	{0x4E13E8BF1E3CE93DULL, "6043938_m.fxe"},
	{0x4E141FF7D4E953F7ULL, "gen00fl_merrill6113648.cl"},
	{0x4E147BDFE0BD5FE5ULL, "vxlt_backalley_n_1wx_0_0.vlm"},
	{0x4E14BE5B0F553A12ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.impa.evt"},
	{0x4E150C838F234237ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_black.phy"},
	{0x4E153C54C611F044ULL, "626398778.ogg"},
	{0x4E15B7121E5D7A76ULL, "hlac_lt_gallowscourtyard_-2_2y.dds"},
	{0x4E15C9C67D8AEDE7ULL, "water_wireframe.vsh.fxo"},
	{0x4E15E5B62AEDD860ULL, "vxht_mansion2b_2fa_0_0.vlm"},
	{0x4E15FB8EB519E9E0ULL, "hightown_male_cityguard_mob6170013.cl"},
	{0x4E16438720C30E5BULL, "follower_banter6148476.cl"},
	{0x4E164AC268209074ULL, "6209975_m.fxe"},
	{0x4E165B6DC266EF2EULL, "vxlt_0_lowtown_docks_a3_d2f_0_1.vlm"},
	{0x4E16EE0A5CB35376ULL, "vxdr_primevaltomb_559_0_1.vlm"},
	{0x4E179F84C1F9DE59ULL, "6206364_m.fxe"},
	{0x4E17CEA27572EB97ULL, "art\\characters\\playercharacter\\humanfemale\\glv_wic.phy"},
	{0x4E181701A81DA545ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfallr_1.mao"},
	{0x4E181AA2EF56FE29ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv3.ani"},
	{0x4E181EB5C48AB70EULL, "vxw_woundedcoast_ul_0_0.vlm"},
	{0x4E182220277D4983ULL, "6174982_m.fxe"},
	{0x4E18320B8A66E6A3ULL, "176406652.ogg"},
	{0x4E189B66DD5B8291ULL, "vxw_bonepit_124_0_0.vlm"},
	{0x4E18C19E2B7DBDE4ULL, "follower_banter6173022.cl"},
	{0x4E192E85C9075296ULL, "lgt220cr_apost1.ncs"},
	{0x4E1970EAE8533EEAULL, "vxdr_primevalentrance_2q3_0_0.vlm"},
	{0x4E1A5117759AE0F7ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_p.anb"},
	{0x4E1A624C0357920BULL, "zz_vault_debug6200329.cl"},
	{0x4E1A6C030EA85922ULL, "vxht_keep_735_0_0.vlm"},
	{0x4E1A708345E2C10CULL, "6123318_m.fxe"},
	{0x4E1B5D20F2D9CBDEULL, "llac_lt_gallowscourtyard_n_-65535_0.dds"},
	{0x4E1BAF51DB215F09ULL, "709155770.ogg"},
	{0x4E1BC5026128FD7FULL, "6085134_m.fxe"},
	{0x4E1BDD5B0F5B8DA5ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.impa.gad"},
	{0x4E1C0C230EF59C8EULL, "647616103.ogg"},
	{0x4E1C869EF2B11E26ULL, "hlac_w_blightlands_131075_0z.dds"},
	{0x4E1D0E07E4A7EFEDULL, "6177843_f.fxe"},
	{0x4E1D3D7A4F1E2689ULL, "vxlt_foundry_n_12d_0_0.vlm"},
	{0x4E1DE4F225C15506ULL, "vxht_mansion2_1ai_0_0.vlm"},
	{0x4E1E2FE4F307F5A8ULL, "hlac_lt_0_lowtown_docks_-65537_1x.dds"},
	{0x4E1E6024E6BCDE3CULL, "hlac_lt_gallowstemplar_f_2394_0y.dds"},
	{0x4E1E92FA860FC10BULL, "6068539_m.fxe"},
	{0x4E1EC544C1908999ULL, "vxlt_0_lowtown_fe9_0_0.vlm"},
	{0x4E1EE10EF76A7834ULL, "vxlt_0_lowtown_docks_a3_brk_0_0.vlm"},
	{0x4E1F35A2EF5D4AABULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv3.gad"},
	{0x4E1F6D6AA478065FULL, "fadeshear.dds"},
	{0x4E1FC4D2F0A174DCULL, "vxlt_0_lowtown_ex0_0_0.vlm"},
	{0x4E1FCF62680B5742ULL, "vxht_0_hightown_n_c2y_0_0.vlm"},
	{0x4E206D767089139CULL, "vxlt_shop_fr_0_0.vlm"},
	{0x4E2074D8B8638C86ULL, "6126438_m.fxe"},
	{0x4E208310FAC61C0AULL, "vxlt_0_lowtown_n_e27_0_0.vlm"},
	{0x4E20A9C612B0E364ULL, "38134810.ogg"},
	{0x4E2106B8BDA4A927ULL, "849004554.ogg"},
	{0x4E2159694D8F6519ULL, "vxlt_gallowstemplar_n_3qx_0_0.vlm"},
	{0x4E21BC476D9FB1E2ULL, "mer140_bonepit_comments6077839.cl"},
	{0x4E21BF3D4976DF83ULL, "vxht_0_hightown_n_a7y_0_0.vlm"},
	{0x4E2227403184D042ULL, "142531271.ogg"},
	{0x4E222BEF80CE5F47ULL, "6088102_m.fxe"},
	{0x4E224C1F63DCEE16ULL, "vxht_mansion2_3g5_0_0.vlm"},
	{0x4E22750C4E84B36BULL, "one300_zevran6089699.cl"},
	{0x4E22CC9F0B9775E3ULL, "6197718_m.fxe"},
	{0x4E22D9D5E1741A2FULL, "6148202_m.fxe"},
	{0x4E22EDBCD2B37865ULL, "vxlt_backalley_n_1rg_0_0.vlm"},
	{0x4E22F2C80D339233ULL, "vxca_caverns1_mines_j29_0_2.vlm"},
	{0x4E23CB6CE925EACDULL, "hlac_lt_0_lowtown_a3_n_c_-65536_2x.dds"},
	{0x4E23EC0BA8D3C0C4ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.dura.ani"},
	{0x4E244B28FB2F8526ULL, "good_targeted.mao"},
	{0x4E245097D0815419ULL, "964115289.ogg"},
	{0x4E2457931CE3F0E0ULL, "vxht_mansion2b_2cn_0_0.vlm"},
	{0x4E24E79C57956EC0ULL, "6104422.cut"},
	{0x4E25224AE00CD42AULL, "vxdr_0_deeproads_ckf_0_0.vlm"},
	{0x4E2532A2EF61B110ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv3.evt"},
	{0x4E25C0F4E6DFD884ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0sc_basic.msh"},
	{0x4E25D2F4E6DFF7EDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0sc_basic.mao"},
	{0x4E25D6F4E6DFFE26ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0sc_basic.mmh"},
	{0x4E25FF12839754A1ULL, "117855969.ogg"},
	{0x4E26018713FC183BULL, "vxlt_0_lowtown_n_h8a_0_0.vlm"},
	{0x4E26495579FC780DULL, "hlac_lt_0_lowtown_docks_-196610_1z.dds"},
	{0x4E2670E8B8265485ULL, "930784798.ogg"},
	{0x4E269092B680CF8EULL, "vxw_woundedcoast_rm_0_0.vlm"},
	{0x4E27C5B6047390E0ULL, "120118478.ogg"},
	{0x4E28294A2F7E6365ULL, "sky_lt_shop.mao"},
	{0x4E2861E207FEB0A8ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_001a_0n.dds"},
	{0x4E2864B12CE7041BULL, "6109249_m.fxe"},
	{0x4E286AD2345E7839ULL, "vxlt_0_lowtown_a3_n_c_eun_0_0.vlm"},
	{0x4E288D4484372C33ULL, "203915883.ogg"},
	{0x4E289B46E286615BULL, "6125298_m.fxe"},
	{0x4E2991CAF64715F9ULL, "vxw_blightlands_wc_0_0.vlm"},
	{0x4E299667299B4E37ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxed.mmh"},
	{0x4E299A67299B5544ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxed.mao"},
	{0x4E29AC67299B73A9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxed.msh"},
	{0x4E29B3652682C977ULL, "atl_optionsmen_dxt5_00.dds"},
	{0x4E29FB70F9F88754ULL, "mcr310_sibling6076396.cl"},
	{0x4E2A3119BD99F83BULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsa_0s.dds"},
	{0x4E2AA33E26DA294AULL, "vxlt_undercitydungeon_20a_0_0.vlm"},
	{0x4E2AE90BA8D9DA26ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.dura.gad"},
	{0x4E2B01DED2CD5120ULL, "655786054.ogg"},
	{0x4E2B0CF51905D2A2ULL, "396343923.ogg"},
	{0x4E2B1378F012A5E1ULL, "vxht_keep_n_6bc_0_0.vlm"},
	{0x4E2B158B595D3489ULL, "lgt210_magu_selby6142043.cl"},
	{0x4E2BE9567B5B0CE7ULL, "6079664_f.fxe"},
	{0x4E2C2D2B9631F130ULL, "vxht_chantry_n_2dx_0_0.vlm"},
	{0x4E2CBE87D0196806ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_flare02.dds"},
	{0x4E2D282790B5B535ULL, "gen00fl_merrill6130087.cl"},
	{0x4E2D954C6299860FULL, "345334968.ogg"},
	{0x4E2DFA354C84E489ULL, "zz_fol_debug6096550.cl"},
	{0x4E2E1B131B00C1D6ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookstand.mao"},
	{0x4E2E27131B00D675ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookstand.mmh"},
	{0x4E2E29131B00D9EBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookstand.msh"},
	{0x4E2E36AFE2978D5CULL, "vxlt_0_lowtown_es1_0_0.vlm"},
	{0x4E2E86FCCBB36F22ULL, "6108984_m.fxe"},
	{0x4E2EC1FFB3339283ULL, "577502364.ogg"},
	{0x4E2F23BE836CCB53ULL, "hlac_ht_0_hightown_a3_65537_0y.dds"},
	{0x4E2F4C5343AC57CCULL, "vxca_cavern1_haunted_j4v_0_0.vlm"},
	{0x4E2F894EB0996EBAULL, "vxlt_0_lowtown_a3_n_c_e1u_0_0.vlm"},
	{0x4E2FC35AA5C92B45ULL, "6078046_m.fxe"},
	{0x4E2FCDF4E6E843F1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0sc_basic.phy"},
	{0x4E2FF334AF641F2AULL, "vxht_chantry_2cd_0_0.vlm"},
	{0x4E3012F5BE352B27ULL, "tal_sha_02.mao"},
	{0x4E309155B0BD2D75ULL, "6032013_m.fxe"},
	{0x4E3092FC6882E86CULL, "458733884.ogg"},
	{0x4E309DCC72C477C7ULL, "6083002_m.fxe"},
	{0x4E30D77268077689ULL, "llac_lt_0_lowtown_docks_a3_n_262142_0.dds"},
	{0x4E3133A75849CC28ULL, "vxht_keep_n_8aq_0_0.vlm"},
	{0x4E318BFAC0182571ULL, "vxdr_primevaltomb_5bj_0_0.vlm"},
	{0x4E31E40BA8DFF005ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.dura.evt"},
	{0x4E324E54C270FCC2ULL, "6202148_f.fxe"},
	{0x4E324F041478846FULL, "vxht_bartrand_n_6km_0_0.vlm"},
	{0x4E32DC1F6E6D89E6ULL, "vxdr_primevalentrance_1k6_0_0.vlm"},
	{0x4E33BEDC188C186CULL, "llac_dr_0_deeproads_5845_13.dds"},
	{0x4E340CFB3C21079FULL, "hlac_lt_0_lowtown_a3_n_c_65534_3x.dds"},
	{0x4E349C2EB16751C8ULL, "vxht_keep_n_6bf_0_0.vlm"},
	{0x4E3531828F978F7AULL, "773819438.ogg"},
	{0x4E362E75AD3D49E9ULL, "vxht_0_hightown_n_c7u_0_0.vlm"},
	{0x4E36B0FD93A49982ULL, "165631558.ogg"},
	{0x4E36C43B73624BAAULL, "vxht_0_hightown_n_7fy_0_0.vlm"},
	{0x4E36D0AA87D9F5DEULL, "lt_0_lowtown_a2_n_52_0_ll.mmh"},
	{0x4E36DA9D85105304ULL, "plt_fen000pt_midstage.nss"},
	{0x4E36DCAF265490B9ULL, "vxlt_0_lowtown_a3_n_c_exm_0_0.vlm"},
	{0x4E36FA495C3EB7FEULL, "6142300_m.fxe"},
	{0x4E374196F1993A6BULL, "vxlt_0_lowtown_a3_n_c_how_0_0.vlm"},
	{0x4E375C4B03437A83ULL, "ave260ip_signal_fire2.utp"},
	{0x4E37621FA8BE1DB7ULL, "art\\vfx\\cinematic\\vfx_arishok_spark_p.impa.gad"},
	{0x4E380395E2791AAAULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingd.mao"},
	{0x4E380795E27921F9ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingd.mmh"},
	{0x4E381195E27932E7ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingd.msh"},
	{0x4E3851A47AA868E5ULL, "hlac_lt_0_lowtown_docks_-131075_5x.dds"},
	{0x4E387E0A7B227AC0ULL, "198846048.ogg"},
	{0x4E38A893DB1CE833ULL, "mcr371_hostile_aveline6082200.cl"},
	{0x4E39164D8AF3041BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medc_0d.dds"},
	{0x4E3918F08E4E0BF1ULL, "jobboardplots_seb.gda"},
	{0x4E392B12ABCBC67CULL, "387188195.ogg"},
	{0x4E393CBDD020F785ULL, "6076075_m.fxe"},
	{0x4E393F6DDC687332ULL, "vxht_mansion2b_3ew_0_0.vlm"},
	{0x4E3947AB34F3BA00ULL, "412583469.ogg"},
	{0x4E3968468583A3B2ULL, "vxca_caverns1_mines_gk5_0_1.vlm"},
	{0x4E39DE95C73CB7F6ULL, "heroicstats.gda"},
	{0x4E39E37E4E5F8A91ULL, "170360374.ogg"},
	{0x4E3A11D37D7B22BCULL, "mag360_samson6172235.cl"},
	{0x4E3A186BB50A9FDFULL, "6211928_m.fxe"},
	{0x4E3A3BA01D7D4327ULL, "follower_banter6173386.cl"},
	{0x4E3A5E177CD99B28ULL, "vxlt_undercitydungeon_1n6_0_0.vlm"},
	{0x4E3A63E838EEACEAULL, "215961224.ogg"},
	{0x4E3A64F451B8B108ULL, "hlac_w_encounter2_camp_4_0x.dds"},
	{0x4E3ABE226E0CD717ULL, "6056740_f.fxe"},
	{0x4E3B036C21082FC8ULL, "mcr371_orsino6083463.cl"},
	{0x4E3B3A237B87CDE0ULL, "gencr_dragonling.utc"},
	{0x4E3B5FBAE1A7AC1BULL, "mrl210_merrill6102125.cl"},
	{0x4E3BA8B82D344873ULL, "576446227.ogg"},
	{0x4E3BE1466AD5A16AULL, "6077586_f.fxe"},
	{0x4E3C3267E6512EA2ULL, "hlac_ca_cavern1_haunted_24423_6y.dds"},
	{0x4E3C87410F7A2C8AULL, "6179337_m.fxe"},
	{0x4E3CCDE6F625FE58ULL, "mcr310_sibling6076734.cl"},
	{0x4E3D771FA8C2AD44ULL, "art\\vfx\\cinematic\\vfx_arishok_spark_p.impa.evt"},
	{0x4E3DA2BCDDFADDD5ULL, "vxht_0_hightown_c3w_0_1.vlm"},
	{0x4E3DBE3035A2704CULL, "vxca_cavern1_haunted_j3w_0_0.vlm"},
	{0x4E3E6D9EF5A4A542ULL, "vxht_brothel_n_59e_0_0.vlm"},
	{0x4E3F120F9C3DC8E2ULL, "zz_dre_debug6054672.cl"},
	{0x4E3F42D4003C378CULL, "vxw_bonepitdestroyed_a3_29w_0_0.vlm"},
	{0x4E3FD176B69FA965ULL, "vxlt_backalley_n_1xe_0_0.vlm"},
	{0x4E3FE8368D85047FULL, "568471181.ogg"},
	{0x4E400BB079B8F872ULL, "6122081_m.fxe"},
	{0x4E402343D2E0C02EULL, "6102403_m.fxe"},
	{0x4E407A41913EC246ULL, "gen00fl_isabela6135632.cl"},
	{0x4E4106D7C68FF43CULL, "103703411.ogg"},
	{0x4E41298DA003DE3EULL, "sb03_cloud.dds"},
	{0x4E4195D734828A7AULL, "vxlt_gallowstemplar_37b_0_0.vlm"},
	{0x4E41F726705688FCULL, "6013299_m.fxe"},
	{0x4E422EE2981B7601ULL, "vxlt_0_lowtown_exa_0_0.vlm"},
	{0x4E434E1775BDE0F8ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_p.mmh"},
	{0x4E4394EA772E8F95ULL, "6085217_m.fxe"},
	{0x4E43B177F5E3C600ULL, "follower_banter6155682.cl"},
	{0x4E4411368B0A3444ULL, "hlac_lt_0_lowtown_docks_-196609_1y.dds"},
	{0x4E4470FE326AC691ULL, "216878698.ogg"},
	{0x4E448F9D6D1F37DCULL, "vxca_cave_1b2_0_0.vlm"},
	{0x4E448FE896E16F81ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_basic.phy"},
	{0x4E44C43CDC5D4516ULL, "296189434.ogg"},
	{0x4E44EF63D467A4A3ULL, "vxht_0_hightown_a3_by2_0_0.vlm"},
	{0x4E455A53EF41C7A2ULL, "6126868_f.fxe"},
	{0x4E45D4152690D786ULL, "dae100_alistair_drunk6135126.cl"},
	{0x4E4639C90F88CD4EULL, "w_encounter1.arl"},
	{0x4E46C808BFDADFABULL, "hightown_fem_courtesan_mob6169024.cl"},
	{0x4E46ED2239961330ULL, "48500812.ogg"},
	{0x4E473123B59A4CDCULL, "235646666.ogg"},
	{0x4E47414FD74B9055ULL, "w_bonepit_26_0.rml"},
	{0x4E47837D105963FEULL, "mag370_meredith6039457.cl"},
	{0x4E47D57E563DF87FULL, "vxlt_0_lowtown_a3_n_c_gij_0_0.vlm"},
	{0x4E48A24BBDC8B368ULL, "vxca_caverns1_edh_0_0.vlm"},
	{0x4E48E93F6BBE6FFFULL, "one300_murder_followers6166957.cl"},
	{0x4E49DACC28F9CF27ULL, "one140cr_charade6137038.cl"},
	{0x4E49EF660B3D9063ULL, "6043648_m.fxe"},
	{0x4E4A9251DE65AF26ULL, "vxht_keep_6fo_0_0.vlm"},
	{0x4E4A95688B55CD36ULL, "and100_anders6095853.cl"},
	{0x4E4B1A69C683BE5CULL, "vxlt_0_lowtown_ei7_0_0.vlm"},
	{0x4E4B8F1FA8CF133DULL, "art\\vfx\\cinematic\\vfx_arishok_spark_p.impa.ani"},
	{0x4E4B989F17B736ACULL, "vxlt_0_lowtown_docks_n_bbo_0_0.vlm"},
	{0x4E4C2D8A42DDEEC2ULL, "vxht_0_hightown_a2_c7m_0_0.vlm"},
	{0x4E4CC5832C0A367DULL, "dae011_corff_sta6149629.cl"},
	{0x4E4CD6EE9350502AULL, "vxht_chantry_2yf_0_0.vlm"},
	{0x4E4CEE3F7A079B52ULL, "6082996_m.fxe"},
	{0x4E4D08A1CF64B0EAULL, "gen00fl_fenris6126593.cl"},
	{0x4E4D0AB90125DABCULL, "9876516.ogg"},
	{0x4E4D155A1ED5688CULL, "static_static_p29.psh.fxo.dx11.assembly"},
	{0x4E4D1692DCD6EDDCULL, "vxht_chantry_n_2yg_0_0.vlm"},
	{0x4E4D3C48047BF34FULL, "205659350.ogg"},
	{0x4E4D763E4A5E6C9BULL, "vxht_brothel_n_6ff_0_0.vlm"},
	{0x4E4DB409E38CEA4AULL, "569999962.ogg"},
	{0x4E4DB4B0F232500CULL, "vxw_bonepitdestroyed_a3_24t_0_0.vlm"},
	{0x4E4DC3574A371D86ULL, "vxht_keep_5xo_0_1.vlm"},
	{0x4E4DC8F1A010FB9CULL, "3040421.ogg"},
	{0x4E4DEE62CB28B15CULL, "vxca_cavern1_haunted_ehu_0_0.vlm"},
	{0x4E4E21E071074A63ULL, "vxht_mansion_n_5b2_0_0.vlm"},
	{0x4E4E22D80EF05C4FULL, "6034025_f.fxe"},
	{0x4E4F0A2C97CEBB56ULL, "character_rendertodepth_p4.vsh.fxo.dx11"},
	{0x4E4F2B709910CF24ULL, "vxht_mansion2_3gw_0_0.vlm"},
	{0x4E4F32BDF864B56FULL, "vxht_bartrand_n_6qo_0_0.vlm"},
	{0x4E4F948D004FEA20ULL, "6047265_m.fxe"},
	{0x4E4FC63000889999ULL, "llac_ht_playershouse_n_855_26.dds"},
	{0x4E5003299FA2B608ULL, "61929443.ogg"},
	{0x4E508631AC4C3CC1ULL, "6124691_m.fxe"},
	{0x4E508BBD58E78F21ULL, "hightown_male_noble_gossip6171306.cl"},
	{0x4E50F33E228CF424ULL, "vxlt_gallowscourtyard_n_7ic_0_0.vlm"},
	{0x4E51465831D384B9ULL, "vxlt_0_lowtown_gjq_0_0.vlm"},
	{0x4E5170C3F4964A1CULL, "6054827_m.fxe"},
	{0x4E51AA84DF06E54AULL, "vxlt_0_lowtown_n_gpk_0_0.vlm"},
	{0x4E51AB3C172B2411ULL, "355711490.ogg"},
	{0x4E51CB98873E47D2ULL, "6059075_m.fxe"},
	{0x4E5223BD1BFC0F03ULL, "vxht_0_hightown_a2_c3m_0_1.vlm"},
	{0x4E526542642C844EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_exit.ani"},
	{0x4E5271F8243885D9ULL, "vxht_0_hightown_a3_bxs_0_0.vlm"},
	{0x4E527B132806624BULL, "character_character_p237.psh.fxo.dx11"},
	{0x4E532E0465F0E768ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medg.msh"},
	{0x4E533273FB24581AULL, "vxht_0_hightown_aow_0_0.vlm"},
	{0x4E53400465F105F9ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medg.mao"},
	{0x4E534C0465F11992ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medg.mmh"},
	{0x4E54172431B1A05FULL, "vxdr_primevaltomb_50b_0_0.vlm"},
	{0x4E5479D6FC7B8D79ULL, "vxdr_0_deeproads_cue_0_2.vlm"},
	{0x4E551241C7C0371FULL, "6069907_m.fxe"},
	{0x4E5582A71BF72C1EULL, "static_billboard_p0.psh.fxo"},
	{0x4E559F0735AD027BULL, "6081189_m.fxe"},
	{0x4E561728EAF7074EULL, "656398250.ogg"},
	{0x4E57D1D95BE40675ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbroke.phy"},
	{0x4E580EF9BB52DAEEULL, "cd_dae000_flemeth.wbd"},
	{0x4E5848429E6AAD4FULL, "gen00fl_sebastian6189396.cl"},
	{0x4E587B2709539E8EULL, "2p_200.stg"},
	{0x4E58DB9D852D3BFFULL, "plt_fen000pt_midstage.plo"},
	{0x4E5950F4E300CC0FULL, "216118932.ogg"},
	{0x4E59A40C0A69DCE3ULL, "255268865.ogg"},
	{0x4E5A020A89F3E98CULL, "follower_banter6148749.cl"},
	{0x4E5A2D3C7D605A8AULL, "vxht_keep_n_7ad_0_0.vlm"},
	{0x4E5A9F6734D40742ULL, "vxht_0_hightown_a2_c2n_0_0.vlm"},
	{0x4E5AD33EDF0A1C19ULL, "411025792.ogg"},
	{0x4E5B774B08D3C9CAULL, "757362398.ogg"},
	{0x4E5B7B403B6ED824ULL, "follower_banter6141939.cl"},
	{0x4E5B886FCECD065CULL, "vxht_chantry_n_2td_0_0.vlm"},
	{0x4E5BBCFA1B022657ULL, "75060487.ogg"},
	{0x4E5BE1838F5EFB8FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_black.mao"},
	{0x4E5BED838F5F102CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_black.mmh"},
	{0x4E5BF3838F5F1A1EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_black.msh"},
	{0x4E5C419B9C9E134CULL, "1005288305.ogg"},
	{0x4E5CC11D5B8616E0ULL, "211277000.ogg"},
	{0x4E5CC9CEDAF9A39FULL, "one140cr_charade6137643.cl"},
	{0x4E5D43F949AF50B4ULL, "571020322.ogg"},
	{0x4E5D62ABC7797B8FULL, "zz_followers_banter6158862.cl"},
	{0x4E5D6F237C866D6EULL, "vxca_caverns2_78w_0_0.vlm"},
	{0x4E5DAAD2FB50E4A7ULL, "dae021_sister_lorena_sta6142650.cl"},
	{0x4E5DBB7B9F1213D9ULL, "6053269_m.fxe"},
	{0x4E5DC2A6B88675C6ULL, "follower_banter6173019.cl"},
	{0x4E5E3660FDCA156FULL, "35894884.ogg"},
	{0x4E5E6A222EB38888ULL, "363150592.ogg"},
	{0x4E5EA3A275AEA5EFULL, "art\\characters\\playercharacter\\humanfemale\\glv_wic.mao"},
	{0x4E5EAB89BFB1814EULL, "6074984_m.fxe"},
	{0x4E5EAFA275AEB98CULL, "art\\characters\\playercharacter\\humanfemale\\glv_wic.mmh"},
	{0x4E5EB5A275AEC47EULL, "art\\characters\\playercharacter\\humanfemale\\glv_wic.msh"},
	{0x4E5F7C2445672A84ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medc.msh"},
	{0x4E5F87F447A1EFC8ULL, "6104945_f.fxe"},
	{0x4E5F8E24456749EDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medc.mao"},
	{0x4E5F922445675026ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medc.mmh"},
	{0x4E5FBA226926A3E9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge_sa.dds"},
	{0x4E605D426438B39FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_exit.evt"},
	{0x4E60CDA9D48725B5ULL, "6129605_m.fxe"},
	{0x4E613241959202E4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01a.msh"},
	{0x4E61484195922786ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01a.mmh"},
	{0x4E617A10181D5A00ULL, "vxlt_hangedman_n_13t_0_0.vlm"},
	{0x4E617F046C30DAD7ULL, "plt_prm000pt_lost_tome.plo"},
	{0x4E620C8AB492B62AULL, "one200dg_solivitus6048187.cl"},
	{0x4E62483F341E44EEULL, "6107691_m.fxe"},
	{0x4E626935FF8980CCULL, "hlac_lt_0_lowtown_a2_-2_3y.dds"},
	{0x4E639AF51294286EULL, "hlac_w_encounter2_camp_-1_0x.dds"},
	{0x4E646DAEB9CC3C14ULL, "vxht_0_hightown_a3_cbe_0_0.vlm"},
	{0x4E6498F859A5506BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_enter.ani"},
	{0x4E64DA658DDCF8CEULL, "lt_blackemporium_probe_12s.dds"},
	{0x4E64FA3E8BA301DDULL, "gen00fl_isabela6079032.cl"},
	{0x4E6523FE7A097D5AULL, "ave331_post_fight6138821.cl"},
	{0x4E65357FF23B6672ULL, "zz_one000_merchant.cnv"},
	{0x4E658BBAC3A9D0E7ULL, "dr_0_deeproads_17005.rml"},
	{0x4E65970F43018C85ULL, "6119967_m.fxe"},
	{0x4E65E1784DFD1188ULL, "1013584525.ogg"},
	{0x4E667A295D0EB390ULL, "176962315.ogg"},
	{0x4E66A215073495D4ULL, "vxw_bonepitmining_a2_2aq_0_0.vlm"},
	{0x4E66E71C9BD025B3ULL, "hlac_lt_0_lowtown_a3_n_c_65535_8z.dds"},
	{0x4E673E42643E9D88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_exit.gad"},
	{0x4E6746C23853BD7DULL, "llac_ht_0_hightown_n_-65537_2.dds"},
	{0x4E680D5EF6FDBFA9ULL, "vxlt_undercityserialkiller_yl_0_0.vlm"},
	{0x4E6826C1AB570B87ULL, "7396965.ogg"},
	{0x4E683FF806EEBA63ULL, "hlac_ca_caverns1_24426_2z.dds"},
	{0x4E6846D4DCA77163ULL, "vxlt_hangedman_1q2_0_0.vlm"},
	{0x4E686862553439F0ULL, "llac_lt_0_lowtown_n_65535_4.dds"},
	{0x4E69599E26A2A46CULL, "llac_ca_caverns2_5802_29.dds"},
	{0x4E698924456F95F1ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medc.phy"},
	{0x4E698DD187DA4BF6ULL, "gen00fl_anders6125477.cl"},
	{0x4E6A7CC7250303E9ULL, "zz_fen_debug6215258.cl"},
	{0x4E6A953939985E21ULL, "llac_ht_mansion_n_3373_1.dds"},
	{0x4E6ADFFA46B7DC9AULL, "plt_cod_itm_centurion.plo"},
	{0x4E6AFB4AB414C13EULL, "6091710_m.fxe"},
	{0x4E6B3F41959A6D51ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01a.phy"},
	{0x4E6B5607BA273EA5ULL, "vxlt_0_lowtown_ety_0_0.vlm"},
	{0x4E6C16495D309AB9ULL, "mag330_walter6045979.cl"},
	{0x4E6C39CE15E731B8ULL, "581252452.ogg"},
	{0x4E6C7ED56F38C423ULL, "w_bonepitdestroyed_a3_9_0.msh"},
	{0x4E6C83BD61DFD99BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_lp.evt"},
	{0x4E6C8CD56F38DC7DULL, "w_bonepitdestroyed_a3_9_0.mmh"},
	{0x4E6C8D7B38F719C5ULL, "vxlt_gallowstemplar_37i_0_0.vlm"},
	{0x4E6C90D56F38E28EULL, "w_bonepitdestroyed_a3_9_0.mao"},
	{0x4E6D65CABA21C6BBULL, "mer341_jansen6054880.cl"},
	{0x4E6D6B758FE5E850ULL, "hlac_lt_backalley_g_1_1x.dds"},
	{0x4E6D6FA5CCAF2B09ULL, "6151251_m.fxe"},
	{0x4E6E18A4F4CFC87DULL, "686717538.ogg"},
	{0x4E6E34C09FEB4713ULL, "154195277.ogg"},
	{0x4E6EAF526B2A78D2ULL, "6053275_m.fxe"},
	{0x4E6F1AFDF38328EAULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.dura.gad"},
	{0x4E6F80A28450EF01ULL, "vxht_chantry_n_2y6_0_0.vlm"},
	{0x4E6FDD87D60CD967ULL, "vxht_playershouse_n_1qj_0_0.vlm"},
	{0x4E7016DCBCB97A9CULL, "one200_gold_followers6120143.cl"},
	{0x4E7044AEC7A45BEAULL, "hlac_w_bonepitdestroyed_a3_262145_0x.dds"},
	{0x4E708E1F22442C3DULL, "lgt110_arl_bann_teagan6115583.cl"},
	{0x4E709AB2A4917D0CULL, "vxlt_0_lowtown_a2_f9h_0_0.vlm"},
	{0x4E70D2D1CCC0C430ULL, "vxlt_0_lowtown_a2_n_c_er7_0_0.vlm"},
	{0x4E711C34F8AE2FE7ULL, "vxlt_0_lowtown_a2_n_c_eq9_0_0.vlm"},
	{0x4E71563B0FB4434AULL, "6170280_m.fxe"},
	{0x4E716D767ABBDAA6ULL, "dae111ip_at_ws_lowtown.utp"},
	{0x4E7180F859AFB1E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_enter.evt"},
	{0x4E7185E3EB9727D7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed.gad"},
	{0x4E71ABC515535380ULL, "6089641_m.fxe"},
	{0x4E71F6951B439FF4ULL, "220326553.ogg"},
	{0x4E71FA62A4AEBADDULL, "6025347_m.fxe"},
	{0x4E7269804DC7DCE1ULL, "6071024_f.fxe"},
	{0x4E72D38D695863DDULL, "334991183.ogg"},
	{0x4E737CBD61E5ECDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_lp.gad"},
	{0x4E73974A09200F3EULL, "6175012_m.fxe"},
	{0x4E73EB4B78670CF8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge_1.mao"},
	{0x4E747FC53D2E8D05ULL, "6150340_m.fxe"},
	{0x4E751F4D6DE1023DULL, "227337193.ogg"},
	{0x4E7537FD88D1A2A8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masb_0d.dds"},
	{0x4E758E046C419D4CULL, "plt_prm000pt_lost_tome.nss"},
	{0x4E75B3A3B7D381CBULL, "vxlt_gallowsprison_n_6pt_0_0.vlm"},
	{0x4E75D12FD8CEFDD3ULL, "6167270_m.fxe"},
	{0x4E75D31C24EFC7FFULL, "6116160_m.fxe"},
	{0x4E75DAD9C2C5D507ULL, "lt_gallowscourtyard_n_34_0.rml"},
	{0x4E75FDFDF3891581ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.dura.evt"},
	{0x4E760C9C5F59A03CULL, "658315796.ogg"},
	{0x4E76260FB92EE466ULL, "688449201.ogg"},
	{0x4E768516C8EFFAA4ULL, "vrc221_bartrand6043396.cl"},
	{0x4E778FA4A0530B2BULL, "qcr000_aveline6185156.cl"},
	{0x4E77A26FF5A10C5FULL, "hlac_w_bonepitmining_a2_196614_0x.dds"},
	{0x4E77B0979E79B494ULL, "vxlt_gallowstemplar_n_5r6_0_0.vlm"},
	{0x4E7812CD1AF345E2ULL, "one350_terath6106356.cl"},
	{0x4E787FF859B5CEF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_enter.gad"},
	{0x4E789AE3EB9D6A64ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed.evt"},
	{0x4E78DCF2EE1051D9ULL, "gen_im_wep_war_one_act1_07.uti"},
	{0x4E78F942C6BA9B37ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.up.gad"},
	{0x4E790B28BE263F7EULL, "6042691_m.fxe"},
	{0x4E796D27A60AD9BEULL, "6096610_m.fxe"},
	{0x4E797BBD61EA56DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_bldwnd_lp.ani"},
	{0x4E79B727DF18F395ULL, "vxlt_0_lowtown_docks_a3_bn2_0_0.vlm"},
	{0x4E79C7E4AC1D5725ULL, "vxlt_0_lowtown_eox_0_0.vlm"},
	{0x4E79DAC4A71D47C5ULL, "6101531_m.fxe"},
	{0x4E7A207776D5371EULL, "694178234.ogg"},
	{0x4E7A2A4CF5921D1CULL, "llac_lt_0_lowtown_a2_n_65535_2.dds"},
	{0x4E7A43CC7D502DBEULL, "379219012.ogg"},
	{0x4E7A712BB4D26CABULL, "vxht_bartrand_n_5vi_0_0.vlm"},
	{0x4E7B30D9DF4532B1ULL, "vxw_blightlands_cg_0_0.vlm"},
	{0x4E7B76D1378FCB08ULL, "282463726.ogg"},
	{0x4E7C24EC1E595A87ULL, "49073239.ogg"},
	{0x4E7D001BB4E92ECAULL, "vxlt_0_lowtown_n_gof_0_0.vlm"},
	{0x4E7D59680A1C6849ULL, "mag100_karras6034535.cl"},
	{0x4E7DBC629D47FC14ULL, "202732956.ogg"},
	{0x4E7DE340C8FA6C42ULL, "vxw_woundedcoast_z9_0_0.vlm"},
	{0x4E7DF27F76470781ULL, "vxht_chantry_n_2t5_0_0.vlm"},
	{0x4E7E0300AB6E63D3ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_aoe_imp_p.anb"},
	{0x4E7E28A1A09462B8ULL, "aibehaviors.gda"},
	{0x4E7E4F64C802F1E7ULL, "vxht_playershouse_n_1te_0_0.vlm"},
	{0x4E7E67AB1A413018ULL, "vxht_0_hightown_3xr_0_0.vlm"},
	{0x4E7EAEBD48D46C2AULL, "296128556.ogg"},
	{0x4E7EB6D79C4006A3ULL, "vxht_0_hightown_a3_be6_0_0.vlm"},
	{0x4E7EEB200554FD56ULL, "6125237_m.fxe"},
	{0x4E7F855BABCAA4EBULL, "vxht_brothel_n_6d0_0_0.vlm"},
	{0x4E7FCD1F1D3C3560ULL, "vxht_mansion2_1kp_0_1.vlm"},
	{0x4E80404629F2CB50ULL, "vxdr_primevaltomb_6qk_0_2.vlm"},
	{0x4E80854332E1384BULL, "vxht_mansion_n_4hu_0_0.vlm"},
	{0x4E8098F83B7B9657ULL, "vxw_bonepitdestroyed_a3_72_0_0.vlm"},
	{0x4E80ADE5298EFC8BULL, "vxca_caverns2_7oc_0_0.vlm"},
	{0x4E80C6ED3994C453ULL, "qcr000_arishok6059665.cl"},
	{0x4E81006C9D9C71C2ULL, "vxw_bonepit_1v7_0_0.vlm"},
	{0x4E8124F080800571ULL, "ht_0_hightownexplosion_n_27_0_ll.mmh"},
	{0x4E8141DC38B16E04ULL, "dae000cr_grand_cleric6183964.cl"},
	{0x4E81BA59F87F15A2ULL, "vxw_encounter1_f9_0_0.vlm"},
	{0x4E81F4CB7597D594ULL, "chargenmorphcfg.xml"},
	{0x4E82D9E3C949AE77ULL, "448622571.ogg"},
	{0x4E8327F89CC5EEC4ULL, "vxlt_gallowstemplar_n_3sw_0_0.vlm"},
	{0x4E83381990181A53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\blendtree\\trooper.anb"},
	{0x4E8346B98A35545BULL, "6108743_m.fxe"},
	{0x4E83525D7E40E8A9ULL, "555532581.ogg"},
	{0x4E835CBE4C718C2DULL, "isa211_isabela6070582.cl"},
	{0x4E83838EC69C3412ULL, "6128991_m.fxe"},
	{0x4E83889E249F6FE8ULL, "static_static_p16.psh.fxo"},
	{0x4E838DD2DF87CC8CULL, "6107121_m.fxe"},
	{0x4E83C5BCBF7FAACAULL, "hm_lgt_kirkwall_rep.mor"},
	{0x4E83F5FDF39545D0ULL, "art\\vfx\\cinematic\\vfx_fire_bunch_p.dura.ani"},
	{0x4E8404F9BB78041CULL, "cd_dae000_flemeth.bnk"},
	{0x4E8406CFB86E3EACULL, "420495794.ogg"},
	{0x4E8408E802F40A78ULL, "zz_followers_banter6157772.cl"},
	{0x4E840C4B8D924016ULL, "305547796.ogg"},
	{0x4E84FAB8726FBFD5ULL, "vxlt_hangedman_n_2xo_0_0.vlm"},
	{0x4E85991FC0012944ULL, "vxlt_undercityencounter_z0_0_0.vlm"},
	{0x4E85DA492BB0CD7CULL, "war_stunningblows.dds"},
	{0x4E862CC90FBE9B6CULL, "w_encounter1.rml"},
	{0x4E8638233D80DD28ULL, "387626143.ogg"},
	{0x4E8666D00D0F85E8ULL, "88275215.ogg"},
	{0x4E868FB0BA4E5D51ULL, "vxlt_undercityencounter_139_0_0.vlm"},
	{0x4E86B2E3EBA9D05DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed.ani"},
	{0x4E86E8D29A966839ULL, "6086014_m.fxe"},
	{0x4E86EB46C8514C97ULL, "follower_banter6172921.cl"},
	{0x4E8753425FAB387AULL, "vxht_0_hightown_a2_a6h_0_0.vlm"},
	{0x4E8839C19E136FA5ULL, "vxlt_0_lowtown_ejg_0_0.vlm"},
	{0x4E8950690A459316ULL, "vxw_encounter3_g4_0_0.vlm"},
	{0x4E8967A5A534BE97ULL, "vxdr_0_deeproads_9ik_0_0.vlm"},
	{0x4E89769F6C0F4F8BULL, "6129812_m.fxe"},
	{0x4E89A2B6D13B4B31ULL, "vxw_blightlands_df_0_0.vlm"},
	{0x4E89C9F80778E87EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_rage.ani"},
	{0x4E89D8181C781ACCULL, "vxdr_0_deeproads_cwv_0_2.vlm"},
	{0x4E8A2AB30B7B99B5ULL, "91962837.ogg"},
	{0x4E8AC2D5F73A6798ULL, "892859595.ogg"},
	{0x4E8B3D45085BC86BULL, "hlac_ca_caverns1_24427_5x.dds"},
	{0x4E8B502D764370BEULL, "nat380_temmerin6067340.cl"},
	{0x4E8B6EBE5600FD12ULL, "6108265_m.fxe"},
	{0x4E8B785A23752B1BULL, "vxw_sundermount_27i_0_0.vlm"},
	{0x4E8B82F5C53A123DULL, "562823375.ogg"},
	{0x4E8C188CBF6101B7ULL, "hlac_ca_caverns1_24426_8x.dds"},
	{0x4E8C2000AB7AD23CULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_aoe_imp_p.mmh"},
	{0x4E8C31F51FFBE26FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_2.gad"},
	{0x4E8C396BEC10E7D9ULL, "vxht_0_hightown_a3_bdw_0_0.vlm"},
	{0x4E8C8BC0AD9A3347ULL, "vxlt_gallowstemplar_481_0_0.vlm"},
	{0x4E8CDDE4DE0F5E0AULL, "28258012.ogg"},
	{0x4E8D2642C6CB90BDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.up.ani"},
	{0x4E8D7A59BBBD83B1ULL, "6069439_m.fxe"},
	{0x4E8D877A3726537DULL, "vxlt_0_lowtown_a2_n_e33_0_0.vlm"},
	{0x4E8D8E68D66E0519ULL, "6206971_m.fxe"},
	{0x4E8DC299131B7C91ULL, "vxw_bonepitmining_a2_1k1_0_0.vlm"},
	{0x4E8E019B4508E67BULL, "519705279.ogg"},
	{0x4E8EA3D4BA22519DULL, "gen00fl_varric6107302.cl"},
	{0x4E8F26F904E1F78FULL, "weapon_stand_01.dds"},
	{0x4E8F35431E9D7EB3ULL, "vxlt_0_lowtown_a2_n_ekg_0_2.vlm"},
	{0x4E8F6490E3357EA9ULL, "vxdr_primevaltomb_6yt_0_0.vlm"},
	{0x4E8FD5E967F17692ULL, "vfx_rendertoshadowmap_p0.psh.fxo.dx11.assembly"},
	{0x4E90010024B0153CULL, "hlac_lt_0_lowtown_a3_n_c_65538_4z.dds"},
	{0x4E9014022997C566ULL, "vxlt_warehouse_n_12y_0_0.vlm"},
	{0x4E90336B851DB7A6ULL, "6209872_m.fxe"},
	{0x4E903C89A510EE03ULL, "llac_ca_cavern1_haunted_24421_4.dds"},
	{0x4E90625C10B03167ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic_2.mao"},
	{0x4E908BC238EE1D2CULL, "1012324546.ogg"},
	{0x4E914097EEBE81F6ULL, "ts_chantry_prog_junk2.gda"},
	{0x4E9150293C38EB1EULL, "99246323.ogg"},
	{0x4E9178178BCD4F58ULL, "321383897.ogg"},
	{0x4E917D71CC3E1524ULL, "163280886.ogg"},
	{0x4E91B011F532273FULL, "487536226.ogg"},
	{0x4E91C8099F9AA259ULL, "657283101.ogg"},
	{0x4E91FA7B16958868ULL, "static_static_p45.psh.fxo"},
	{0x4E923BB8DFE7CD01ULL, "hf_zombiemother.mor"},
	{0x4E924F1BC0130447ULL, "vxht_0_hightown_c45_0_0.vlm"},
	{0x4E926636DC74CEC4ULL, "mag101_aombanter6052767.cl"},
	{0x4E9316CA6B6D06DFULL, "vxw_bonepit_dc_0_0.vlm"},
	{0x4E9319062AE6EED4ULL, "vxht_brothel_n_3nl_0_0.vlm"},
	{0x4E9346F52002241CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_2.evt"},
	{0x4E935CC0F82E8C79ULL, "gen00fl_carver6097221.cl"},
	{0x4E9427B4E6AD02DEULL, "vxlt_gallowstemplar_f_45h_0_0.vlm"},
	{0x4E943D2498607ED1ULL, "6004406_f.fxe"},
	{0x4E9499BD8D51FC39ULL, "270508517.ogg"},
	{0x4E94B5D8CBD1C126ULL, "docks_male_overseer_mob6176528.cl"},
	{0x4E950A19EC5F1DB5ULL, "character_character_p121.psh.fxo.dx11.assembly"},
	{0x4E95C51F51A150FAULL, "vxht_0_hightown_a2_a1i_0_0.vlm"},
	{0x4E96899D6DFAC72DULL, "zz_tbe100_barber_flyer6190992.cl"},
	{0x4E969E5A28481697ULL, "vxlt_0_lowtown_n_fhk_0_0.vlm"},
	{0x4E96ED9B0148C6F1ULL, "vxlt_0_lowtown_a3_n_c_eop_0_0.vlm"},
	{0x4E96F3D8007D76D1ULL, "339086822.ogg"},
	{0x4E97E1F807854E4FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_rage.evt"},
	{0x4E97FCEDC6EF73F6ULL, "aip_shadowwarrior_base.gda"},
	{0x4E981493C33163B1ULL, "vxw_blightlands_ie_0_0.vlm"},
	{0x4E9891DF9C5A6BE2ULL, "llac_w_bonepit_393221_0.dds"},
	{0x4E98C580C563DDF9ULL, "6098196_m.fxe"},
	{0x4E9917A8C963FD69ULL, "vxca_caverns2_717_0_0.vlm"},
	{0x4E9933BB0CDFBD59ULL, "678395971.ogg"},
	{0x4E9977B824882ECEULL, "6116532_m.fxe"},
	{0x4E9985FCCCAD974EULL, "6156749_m.fxe"},
	{0x4E99BAA9C2873D7BULL, "hlac_lt_0_lowtown_docks_a3_-131074_7x.dds"},
	{0x4E9A84AE0D64B4A1ULL, "w_sundermount_56_0.phy"},
	{0x4E9AAB48DE070059ULL, "vxht_0_hightown_a3_bcv_0_0.vlm"},
	{0x4E9AEB6C706133C0ULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.mmh"},
	{0x4E9B0895E2CDA0E6ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingd.phy"},
	{0x4E9B2E60768B728BULL, "182658861.ogg"},
	{0x4E9B53693494DC22ULL, "267491313.ogg"},
	{0x4E9BD1E8131F989AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_exit.gad"},
	{0x4E9C69158FB6D5EBULL, "vxht_brothel_n_6j2_0_0.vlm"},
	{0x4E9C8C9CD3F712D8ULL, "mag110_samson6020786.cl"},
	{0x4E9CF305A040F41EULL, "519142376.ogg"},
	{0x4E9CFFA65DA6230DULL, "119603405.ogg"},
	{0x4E9D2C9BCC931198ULL, "arm_pl4.tnt"},
	{0x4E9D68FD16CD694BULL, "vxht_mansion_n_4bw_0_0.vlm"},
	{0x4E9DB1BDEE9E9617ULL, "266551080.ogg"},
	{0x4E9DE9164695E05BULL, "6180469_f.fxe"},
	{0x4E9E6E0C896F38C5ULL, "526297925.ogg"},
	{0x4E9EB07B474B9EF7ULL, "30677314.ogg"},
	{0x4E9EE2F8078B6EF8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_rage.gad"},
	{0x4E9EE4D95C202890ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbroke.msh"},
	{0x4E9EF2D95C2040AAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbroke.mmh"},
	{0x4E9EF6D95C2047E1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbroke.mao"},
	{0x4E9F4DE9D69B0024ULL, "hlac_lt_0_lowtown_a3_n_c_-1_4y.dds"},
	{0x4E9F4F444F817A46ULL, "vxca_cave_1at_0_1.vlm"},
	{0x4E9F94065B001430ULL, "hlac_lt_0_lowtown_n_-1_6z.dds"},
	{0x4E9FAE19526314E5ULL, "narrative_varric6124286.cl"},
	{0x4E9FD0B9C2C95FD1ULL, "hlac_w_encounter1_-65534_0z.dds"},
	{0x4EA00F848A94F1CFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\blendtree\\floortrap.anb"},
	{0x4EA03836EFF3FF2AULL, "package_groups.gda"},
	{0x4EA04A5634953471ULL, "gen00fl_varric6076460.cl"},
	{0x4EA04EF5200CBC35ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_2.ani"},
	{0x4EA0B7DCEC9966FDULL, "124631747.ogg"},
	{0x4EA1096D280B5420ULL, "art\\characters\\creatures\\mubari\\c_dog_body.mao"},
	{0x4EA12A356057EB43ULL, "6036293_m.fxe"},
	{0x4EA161CE00B2FE6DULL, "6102660_m.fxe"},
	{0x4EA16C4CA4DBB7C1ULL, "6160981_m.fxe"},
	{0x4EA175F6B459B5E4ULL, "hlac_lt_0_lowtown_docks_a3_-1_3z.dds"},
	{0x4EA18DAE0D6AE2BDULL, "w_sundermount_56_0.rml"},
	{0x4EA1DE72565BF0D5ULL, "vxlt_hangedman_n_2bm_0_0.vlm"},
	{0x4EA2087FD8C78EDFULL, "6107140_m.fxe"},
	{0x4EA237E309715255ULL, "6064674_m.fxe"},
	{0x4EA2D4E81325BCB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_exit.evt"},
	{0x4EA3062E745BFE14ULL, "vxlt_gallowstemplar_n_5c5_0_0.vlm"},
	{0x4EA36B6A8DC905DFULL, "6204158_f.fxe"},
	{0x4EA37CBFD29D39DAULL, "character_character_p135.psh.fxo.dx11.assembly"},
	{0x4EA38DB5E2917C6AULL, "6126746_m.fxe"},
	{0x4EA43EADB7562629ULL, "19994807.ogg"},
	{0x4EA4648094B4E039ULL, "character_character_p88.psh.fxo.dx11.assembly"},
	{0x4EA535A20E84D877ULL, "lgt250_way_dalish_conflict6147134.cl"},
	{0x4EA55F77F33EDF71ULL, "vxlt_0_lowtown_a3_n_c_erw_0_0.vlm"},
	{0x4EA566DD4F9FA2CBULL, "zz_followers_banter6158286.cl"},
	{0x4EA5C8FF56942A3EULL, "hlac_w_encounter3_-393212_0z.dds"},
	{0x4EA6113AD2929FECULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_inkwell_n.dds"},
	{0x4EA62CABEC61EA70ULL, "ser221_gascard6113796.cl"},
	{0x4EA62FB6C61B06C9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_coma_0n.dds"},
	{0x4EA64B5F8920EF97ULL, "vxdr_0_deeproads_9ge_0_0.vlm"},
	{0x4EA68670B5277C31ULL, "vxw_blightlands_rd_0_0.vlm"},
	{0x4EA715B4D59A93A7ULL, "gen00fl_aveline6139177.cl"},
	{0x4EA794E89735F5BDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_basic.mao"},
	{0x4EA798E89735FCF6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_basic.mmh"},
	{0x4EA7A2E897360DF4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_basic.msh"},
	{0x4EA7BBB0242498B2ULL, "zz_upgrade.ncs"},
	{0x4EA7E907F02E9CA3ULL, "vxht_mansion2_2ic_0_0.vlm"},
	{0x4EA849771ED519F8ULL, "hlac_dr_primevaltomb_664_1y.dds"},
	{0x4EA864DE624A8C5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium.ani"},
	{0x4EA8C4E5BBCB065FULL, "vxlt_0_lowtown_n_f1d_0_0.vlm"},
	{0x4EA8CE6C706D3FDFULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.anb"},
	{0x4EA8EF4044C63255ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_mede.mao"},
	{0x4EA98B161568C158ULL, "vxca_caverns2_9xo_0_0.vlm"},
	{0x4EAA0979C62C8339ULL, "220240119.ogg"},
	{0x4EAA2BB9460D8956ULL, "hlac_w_blightlands_-65535_0z.dds"},
	{0x4EAA6F37DFD95B37ULL, "vxlt_darktown_4if_0_0.vlm"},
	{0x4EAB007039D3B699ULL, "w_bonepitdestroyed_a3_81_0.msh"},
	{0x4EAB0A7039D3C787ULL, "w_bonepitdestroyed_a3_81_0.mmh"},
	{0x4EAB0E7039D3CED4ULL, "w_bonepitdestroyed_a3_81_0.mao"},
	{0x4EAB6CB501226264ULL, "mrl210_merrill6094195.cl"},
	{0x4EAB7EF6B9267FCFULL, "vxlt_0_lowtown_a2_n_c_hyx_0_0.vlm"},
	{0x4EABB47835993054ULL, "6208046_m.fxe"},
	{0x4EABCAFD246A4B6CULL, "green50.dds"},
	{0x4EABD8172E5997D1ULL, "mag160_feynriel6016553.cl"},
	{0x4EAD51B1413E6AD7ULL, "6051579_m.fxe"},
	{0x4EADEBB64EF6CA1EULL, "vxca_cavern1_haunted_gjk_0_0.vlm"},
	{0x4EADFD41DBC3977EULL, "vxht_0_hightown_n_8px_0_0.vlm"},
	{0x4EAE98AD7EE911E4ULL, "vxht_keep_36t_0_0.vlm"},
	{0x4EAED0262F456354ULL, "167102874.ogg"},
	{0x4EAF4B04663F768DULL, "art\\characters\\playercharacter\\humanmale\\hlh_medg.phy"},
	{0x4EAF4D5F4D6B7CC3ULL, "730989409.ogg"},
	{0x4EAFDE2996D1D041ULL, "6165681_m.fxe"},
	{0x4EB0ECE813322280ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_exit.ani"},
	{0x4EB14DD8EE706545ULL, "dae100_viscount6118493.cl"},
	{0x4EB1E614508597C1ULL, "hlac_ht_0_hightown_n_65533_1x.dds"},
	{0x4EB22DCCF7D5212AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\pose_a_weightshift_medium.gad"},
	{0x4EB24AE6641FC7C5ULL, "vxht_0_hightown_a2_a6c_0_0.vlm"},
	{0x4EB24E7E95896556ULL, "dae000_bodhan6146864.cl"},
	{0x4EB2659184C98EEEULL, "follower_barks6169910.cl"},
	{0x4EB30E848F58C5E9ULL, "vxca_caverns2_4zm_0_0.vlm"},
	{0x4EB31D9430D48534ULL, "gen_im_arm_glv_hvy_act3_02.uti"},
	{0x4EB369C8CB369EEAULL, "gen_im_cns_flu_fenris_04.uti"},
	{0x4EB4065EFF35A726ULL, "617749594.ogg"},
	{0x4EB4D2C27C05F38BULL, "83625517.ogg"},
	{0x4EB5072276C04C05ULL, "vxlt_0_lowtown_n_eqd_0_1.vlm"},
	{0x4EB550B9F2D550C9ULL, "413590494.ogg"},
	{0x4EB57CDE62553F2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium.evt"},
	{0x4EB5F46643F96343ULL, "556951666.ogg"},
	{0x4EB622FE228DE886ULL, "mcr370_cut_player_survives6098266.cl"},
	{0x4EB70147A08635F0ULL, "dae010cr_worker_m.utc"},
	{0x4EB72399746D2F63ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.dura.ani"},
	{0x4EB785EFDB3F7E12ULL, "6162661_m.fxe"},
	{0x4EB7DA8E003989C8ULL, "llac_lt_undercitydungeon_-1_23.dds"},
	{0x4EB84EF7BF753863ULL, "288632936.ogg"},
	{0x4EB8805A02C196B0ULL, "spt_fronds_low.psh.fxo"},
	{0x4EB891798960E2BDULL, "gen_im_arm_glv_lgt_act2_02.uti"},
	{0x4EB8B8E4794FE45CULL, "547197962.ogg"},
	{0x4EB910CCF7DB0EC1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\pose_a_weightshift_medium.evt"},
	{0x4EB97495542416FEULL, "811471079.ogg"},
	{0x4EB9BDA08617C39EULL, "vxlt_darktown_6iu_0_0.vlm"},
	{0x4EB9ED75CA586274ULL, "zz_qun_debug6206595.cl"},
	{0x4EBA01DA4749BAB0ULL, "one140cr_charade6134584.cl"},
	{0x4EBA08C0E30A0FADULL, "dae100_bran6175865.cl"},
	{0x4EBAE610E448C3A3ULL, "vxht_mansion_n_751_0_0.vlm"},
	{0x4EBB14DE90541E2CULL, "133468283.ogg"},
	{0x4EBBE319817EEEE8ULL, "654199615.ogg"},
	{0x4EBBFA8DBBD00C22ULL, "mag131_taraneh6088693.cl"},
	{0x4EBC53DAB2F937B0ULL, "al_trees.gda"},
	{0x4EBC58C57F4FDC34ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.anb"},
	{0x4EBC7DDE625B5FD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium.gad"},
	{0x4EBC8BAC669A44B4ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em01.phy"},
	{0x4EBC8F94C2DD6E5AULL, "276025345.ogg"},
	{0x4EBCEF7742EA4348ULL, "247460825.ogg"},
	{0x4EBDD0AA93E09695ULL, "224464968.ogg"},
	{0x4EBE0E20AC04473AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagle.mmh"},
	{0x4EBE1220AC044E71ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagle.mao"},
	{0x4EBE2020AC046580ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagle.msh"},
	{0x4EBE8253022171D9ULL, "zz_seb_debug6215745.cl"},
	{0x4EBEBE76ADEB413AULL, "art\\characters\\playercharacter\\humanmale\\bdy_meda.phy"},
	{0x4EBECEC7FFD5C3B0ULL, "6087548_m.fxe"},
	{0x4EBEDA44D082869BULL, "hlac_w_blightlands_2_0z.dds"},
	{0x4EBF522876E95562ULL, "vxht_0_hightown_9uq_0_0.vlm"},
	{0x4EBF8F2964BE96EEULL, "target_circle_pressed.cur"},
	{0x4EBF9497EB664F4CULL, "21082825.ogg"},
	{0x4EBF96FB617322E9ULL, "339297909.ogg"},
	{0x4EBFCFCC648E48EFULL, "qun220cr_javaris6071896.cl"},
	{0x4EC01BE7AA099D34ULL, "hlac_dr_dungeon_3214_6x.dds"},
	{0x4EC0716FC67DAD6AULL, "6128546_m.fxe"},
	{0x4EC0D56761B32894ULL, "vxdr_0_deeproads_d3s_0_0.vlm"},
	{0x4EC11C33265F51C8ULL, "llac_w_sundermount_196612_0.dds"},
	{0x4EC13A7AAC4A2832ULL, "6161303_m.fxe"},
	{0x4EC1451B65001078ULL, "llac_lt_gallowsprison_n_-1_10.dds"},
	{0x4EC213591878FDA3ULL, "756528552.ogg"},
	{0x4EC23FD86181D222ULL, "ico_helm_medium_1.dds"},
	{0x4EC2AA7C6EA0E72BULL, "vxht_bartrand_n_5ml_0_0.vlm"},
	{0x4EC2B62D4F68B374ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.impa.ani"},
	{0x4EC4007C37CF1ED6ULL, "1026293311.ogg"},
	{0x4EC42AE83F33B279ULL, "6152643_m.fxe"},
	{0x4EC43B997477E25AULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.dura.evt"},
	{0x4EC518F19D5B10A5ULL, "vxht_mansion_n_6p1_0_0.vlm"},
	{0x4EC5603B3730ACB5ULL, "hlac_ht_bartrand_n_3372_0y.dds"},
	{0x4EC5DE6EED930FB2ULL, "vxdr_dungeon_211_0_0.vlm"},
	{0x4EC608CCF7E58B10ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\pose_a_weightshift_medium.ani"},
	{0x4EC668AC66A25DB1ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em01.msh"},
	{0x4EC66ECFF954F258ULL, "vxca_caverns2_9fm_0_0.vlm"},
	{0x4EC682AC66A28A6FULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em01.mmh"},
	{0x4EC73E329151188BULL, "460242841.ogg"},
	{0x4EC817AA040253D7ULL, "hlac_lt_0_lowtown_n_-196607_0y.dds"},
	{0x4EC88B7F73C6A32EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_1_sylvesta.ani"},
	{0x4EC8923CF0B5E28EULL, "vxw_bonepit_11n_0_0.vlm"},
	{0x4EC8B32D4F6D19D6ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.impa.gad"},
	{0x4EC8CD6643B7A9F0ULL, "964726312.ogg"},
	{0x4EC8D30EF2380088ULL, "dre000_vampire6034283.cl"},
	{0x4EC8E99E1D0F3D3CULL, "150454151.ogg"},
	{0x4EC8FF43A0817772ULL, "ts_kirk_alienage_02.gda"},
	{0x4EC94291767B1FFCULL, "vxdr_primevaltomb_6nh_0_0.vlm"},
	{0x4EC95EB900A92154ULL, "6167490_f.fxe"},
	{0x4EC9B37D6566EB26ULL, "llac_ht_0_hightown_65535_3.dds"},
	{0x4ECA37C57F5BE1F7ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.mmh"},
	{0x4ECA6ADD967CBFB0ULL, "vxlt_0_lowtown_a2_n_c_hoy_0_0.vlm"},
	{0x4ECA8749F3343325ULL, "mrl150_merrill6070797.cl"},
	{0x4ECADF4D34FCF9B3ULL, "llac_ht_chantry_2534_22.dds"},
	{0x4ECAF770C7FDEAA5ULL, "itm_base.gda"},
	{0x4ECB2365ED8A4DB9ULL, "6162132_m.fxe"},
	{0x4ECB3A99747DFF4DULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_fire_p.dura.gad"},
	{0x4ECB7BD7E41A61FDULL, "gen_im_arm_cht_lgt_act1_01.uti"},
	{0x4ECBC5605B79F263ULL, "one280_golem.bnk"},
	{0x4ECC103C9B0CFFD6ULL, "33703033.ogg"},
	{0x4ECC44DEF7C43ACFULL, "vxht_bartrand_n_4cc_0_0.vlm"},
	{0x4ECCE079F2BF50D4ULL, "vxht_brothel_n_3zp_0_0.vlm"},
	{0x4ECDE4AF55289B3CULL, "565547996.ogg"},
	{0x4ECDFF5CCCF12867ULL, "313199298.ogg"},
	{0x4ECE3AFDF0B3EBF1ULL, "vxlt_0_lowtown_goj_0_0.vlm"},
	{0x4ECE6DCD162C7C3AULL, "vxlt_0_lowtown_a3_n_c_ejh_0_0.vlm"},
	{0x4ECEA7605B7BF493ULL, "one280_golem.cnv"},
	{0x4ECEB853170488FBULL, "383483165.ogg"},
	{0x4ECECCA28E5CA14BULL, "6087147_m.fxe"},
	{0x4ECF2AF31E9AFF60ULL, "280069757.ogg"},
	{0x4ECFAE8480423B6DULL, "vxlt_0_lowtown_e27_0_0.vlm"},
	{0x4ECFCE0CBEA31A88ULL, "plt_zz_betpt_debug.nss"},
	{0x4ECFCE2D4F736635ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.impa.evt"},
	{0x4ECFD8E55808C38BULL, "lt_0_lowtown_a2_n_33_0_ll.mmh"},
	{0x4ECFF7067A5FDF1CULL, "vxca_cavern1_haunted_ile_0_0.vlm"},
	{0x4ED0795905A22838ULL, "gen00fl_merrill6141134.cl"},
	{0x4ED12DDB76FC4FB4ULL, "896463877.ogg"},
	{0x4ED14ABB6932BAE3ULL, "dae011_corff_sta6164583.cl"},
	{0x4ED2291F5260099EULL, "hlac_dr_primevaltomb_664_6z.dds"},
	{0x4ED284D8F11E7C0FULL, "ave331_post_fight6152563.cl"},
	{0x4ED2E56264092441ULL, "hlac_w_encounter2_131071_0z.dds"},
	{0x4ED379FD5760E44CULL, "6214942_m.fxe"},
	{0x4ED3916EEED2ABD2ULL, "qcr000_arishok6087522.cl"},
	{0x4ED3AAF3093481BFULL, "vxht_mansion2_27i_0_0.vlm"},
	{0x4ED3E6EE0E09F58EULL, "vxw_woundedcoast_6y_0_0.vlm"},
	{0x4ED48E1D45FEBCACULL, "vrc321_varric6087051.cl"},
	{0x4ED4DE5EE33D2133ULL, "hlac_w_encounter2_camp_262142_0x.dds"},
	{0x4ED4F0C9DA6A1E54ULL, "2182652.ogg"},
	{0x4ED536030E5E6109ULL, "326673965.ogg"},
	{0x4ED5730DEEE8CFA0ULL, "6126952_f.fxe"},
	{0x4ED63B6BA2E905A4ULL, "llac_w_blightlands_196610_0.dds"},
	{0x4ED6656724176BC7ULL, "hlac_lt_0_lowtown_docks_-196610_1x.dds"},
	{0x4ED66967A1BAC401ULL, "hlac_lt_0_lowtown_-262144_0x.dds"},
	{0x4ED6D70CBEA9489BULL, "plt_zz_betpt_debug.plo"},
	{0x4ED72B54309ED222ULL, "vxlt_hangedman_27j_0_0.vlm"},
	{0x4ED7A85076F6E38CULL, "t1_mul_br2.tnt"},
	{0x4ED8353F042F709CULL, "vxlt_0_lowtown_a3_n_c_e26_0_0.vlm"},
	{0x4ED8366EEDCA591BULL, "196400076.ogg"},
	{0x4ED85D98615F08D4ULL, "follower_banter6161249.cl"},
	{0x4ED99542DF806639ULL, "6167232_m.fxe"},
	{0x4EDA1C6F9CB64A58ULL, "447535242.ogg"},
	{0x4EDA400846676B81ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen.ani"},
	{0x4EDA45545F36F440ULL, "zz_approval_debug6003940.cl"},
	{0x4EDB2E838FFB146AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isab_0d.dds"},
	{0x4EDB42647165C939ULL, "6096812_f.fxe"},
	{0x4EDB7E6CA31FDE8BULL, "12096195.ogg"},
	{0x4EDBAD380D8783ADULL, "vxw_bonepitdestroyed_a3_241_0_0.vlm"},
	{0x4EDC3366B6670681ULL, "vxlt_hangedman_n_16z_0_0.vlm"},
	{0x4EDC856C430E39DCULL, "300369524.ogg"},
	{0x4EDCA02BB67AE3C2ULL, "vxht_0_hightown_a2_caw_0_0.vlm"},
	{0x4EDCACDAE2AA0471ULL, "vxlt_0_lowtown_gji_0_0.vlm"},
	{0x4EDCC384A1DE63B7ULL, "dae010_cityguard_f1.utc"},
	{0x4EDD411BE4156D24ULL, "vxlt_warehouse_sn_0_0.vlm"},
	{0x4EDD647F73D8BD68ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_1_sylvesta.gad"},
	{0x4EDD6BBC75A1FBFAULL, "6102810_m.fxe"},
	{0x4EDDD09A33070E48ULL, "vxht_0_hightown_a2_a79_0_0.vlm"},
	{0x4EDDEA5C3FD4A5ECULL, "llac_w_bonepit_131074_0.dds"},
	{0x4EDDF838087EA669ULL, "796027022.ogg"},
	{0x4EDE18B72D3B7CCBULL, "6189025_m.fxe"},
	{0x4EDE25CF8A021FB6ULL, "vxca_cavern1_haunted_gk0_0_0.vlm"},
	{0x4EDE68E36C55F79CULL, "vxca_cavern1_haunted_iiz_0_0.vlm"},
	{0x4EDE9CC53CB17C34ULL, "274849970.ogg"},
	{0x4EDECF370BF7EFDEULL, "vxht_mansion2_3bf_0_0.vlm"},
	{0x4EDEDABD13DFE3EFULL, "6084689_m.fxe"},
	{0x4EDEE6FD0613FB0FULL, "character_character_p237.psh.fxo.dx11.assembly"},
	{0x4EDF383D92F8BAAEULL, "6177300_m.fxe"},
	{0x4EDFA81B018FB357ULL, "vxca_caverns1_gkp_0_0.vlm"},
	{0x4EE0262A16418C75ULL, "one100dg_vanard6059697.cl"},
	{0x4EE04B08466BEA23ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen.gad"},
	{0x4EE1807CC9E0D617ULL, "gen00fl_varric6109207.cl"},
	{0x4EE18E56E31FA556ULL, "vxca_cavern1_haunted_in9_0_0.vlm"},
	{0x4EE19D2328417A54ULL, "dae000_hate_follower_react6103251.cl"},
	{0x4EE1B537C8112C6EULL, "gen00fl_fenris6128083.cl"},
	{0x4EE1E92D7569D913ULL, "vxlt_0_lowtown_a2_n_c_hv5_0_0.vlm"},
	{0x4EE1F14BED289FFDULL, "vxlt_0_lowtown_docks_a3_n_ctf_0_0.vlm"},
	{0x4EE27D36F74C7DD6ULL, "vxht_0_hightown_n_4n2_0_0.vlm"},
	{0x4EE32D20AC239845ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eagle.phy"},
	{0x4EE344333D9B7918ULL, "hlac_dr_slavers_3214_8x.dds"},
	{0x4EE3699089840D92ULL, "818923243.ogg"},
	{0x4EE37B4DC2FB49C2ULL, "vxca_cavern1_haunted_cmn_0_0.vlm"},
	{0x4EE430325D4EC11BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroof.msh"},
	{0x4EE442325D4EE066ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroof.mao"},
	{0x4EE44E325D4EF405ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroof.mmh"},
	{0x4EE5973F13C8486DULL, "441801634.ogg"},
	{0x4EE5BD4E763CD3EDULL, "lt_0_lowtown_docks_n_24_0_lc.mmh"},
	{0x4EE621855570A2C6ULL, "6117009_m.fxe"},
	{0x4EE6264B5A6750FCULL, "vxdr_primevaltomb_6xn_0_0.vlm"},
	{0x4EE6693E0E5FB9A0ULL, "335041611.ogg"},
	{0x4EE68085B1F95BB0ULL, "gen00fl_isabela_init6175666.cl"},
	{0x4EE6C60F0303ACEEULL, "592229108.ogg"},
	{0x4EE6F91A11B00DE6ULL, "vxlt_backalley_n_1e9_0_0.vlm"},
	{0x4EE728084671CD08ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen.evt"},
	{0x4EE75086A53E102DULL, "dae020cr_ef_courtesan2.utc"},
	{0x4EE75D7F30FE5EA2ULL, "815132622.ogg"},
	{0x4EE7620885952576ULL, "17474818.ogg"},
	{0x4EE7752217E58FE0ULL, "vxlt_gallowstemplar_f_2ja_0_0.vlm"},
	{0x4EE80F5CB8C3C72AULL, "mag100_thrask6164881.cl"},
	{0x4EE82A964EB8FD6BULL, "mcr371_hostile_anders6082117.cl"},
	{0x4EE83F2BCA1B55AEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_diagonal_1.mao"},
	{0x4EE86663B396F7D3ULL, "6174443_m.fxe"},
	{0x4EE8E961653182B3ULL, "6151293_m.fxe"},
	{0x4EE905122772C34BULL, "vxlt_shop_f1_0_0.vlm"},
	{0x4EE90EA8BB361318ULL, "6071847_f.fxe"},
	{0x4EE9125DB837D7DDULL, "6025130_m.fxe"},
	{0x4EE925B9A1E1F4D3ULL, "gen00fl_anders6110052.cl"},
	{0x4EE9B5B5A2116F8CULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation6.gad"},
	{0x4EE9D3D1324A1E23ULL, "890572334.ogg"},
	{0x4EE9EC9FD66F3140ULL, "64374428.ogg"},
	{0x4EEA73BD1633A0B2ULL, "359056448.ogg"},
	{0x4EEAE3FE32D54E43ULL, "39668742.ogg"},
	{0x4EEB98F43F732404ULL, "6117689_m.fxe"},
	{0x4EEBA09D96387E33ULL, "6164994_m.fxe"},
	{0x4EEBB2F8D60B85A4ULL, "vxlt_warehouse_to_0_0.vlm"},
	{0x4EEC23AE6F085BB7ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha11a.phy"},
	{0x4EECBD7287AB811DULL, "6137475_m.fxe"},
	{0x4EED0417DA1BF7B6ULL, "art\\characters\\playercharacter\\humanfemale\\hm_cth_homa.msh"},
	{0x4EED0E17DA1C08B4ULL, "art\\characters\\playercharacter\\humanfemale\\hm_cth_homa.mmh"},
	{0x4EED7B0B46BB5C1BULL, "6125597_m.fxe"},
	{0x4EEE2517C2A3B840ULL, "gen00fl_aveline6206188.cl"},
	{0x4EEEB6A6441CD08FULL, "vxlt_warehouse_n_op_0_0.vlm"},
	{0x4EEECBFC4F33D138ULL, "dae000ip_crowd_mf1_talk16.utp"},
	{0x4EEF22EB1B462B45ULL, "6039728_m.fxe"},
	{0x4EEF4138C50B8C1DULL, "vxca_caverns1_mines_jfi_0_0.vlm"},
	{0x4EEF4873E407B12AULL, "lt_0_lowtown_docks_a3_n_205_0.mao"},
	{0x4EEF4C73E407B879ULL, "lt_0_lowtown_docks_a3_n_205_0.mmh"},
	{0x4EEF5673E407C967ULL, "lt_0_lowtown_docks_a3_n_205_0.msh"},
	{0x4EEFA6C599D6CB6CULL, "lgtev_item_acquired.ncs"},
	{0x4EEFB4B5A215D98AULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation6.ani"},
	{0x4EEFB4E7BADE7322ULL, "vxca_caverns1_mines_fn6_0_0.vlm"},
	{0x4EF05B0A675FF193ULL, "vxlt_0_lowtown_a2_n_c_hq4_0_0.vlm"},
	{0x4EF0E3F2B6E710A9ULL, "dae000cr_grand_cleric6147879.cl"},
	{0x4EF1AE24C2FB04DDULL, "vxlt_undercityencounter_18l_0_0.vlm"},
	{0x4EF1C4D68E9450A7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_loop.gad"},
	{0x4EF230CB470025D5ULL, "hlac_ca_caverns2_5802_4z.dds"},
	{0x4EF2489B38773F96ULL, "hlac_lt_0_lowtown_docks_-262144_3y.dds"},
	{0x4EF27AFD38602738ULL, "3851629.ogg"},
	{0x4EF28BD80BC04AE7ULL, "gen_im_trs_dwv_junk1_04.uti"},
	{0x4EF2A212D0549104ULL, "gen_im_trs_qun_lgt_glv_r2.uti"},
	{0x4EF2AA8DE04AC2C1ULL, "vxlt_0_lowtown_a2_n_hgv_0_0.vlm"},
	{0x4EF307DDE2F18A5DULL, "996796678.ogg"},
	{0x4EF3CB420AB205CCULL, "hlac_w_encounter2_327684_0z.dds"},
	{0x4EF3E4B26F155252ULL, "6208177_f.fxe"},
	{0x4EF48BB0ECB31666ULL, "480163378.ogg"},
	{0x4EF518F8E81BA19CULL, "vxlt_0_lowtown_a3_n_c_e84_0_0.vlm"},
	{0x4EF54CF14D6C8939ULL, "hlac_ht_mansion2_818_32z.dds"},
	{0x4EF5C439977D47D1ULL, "vxlt_0_lowtown_docks_a3_n_bql_0_0.vlm"},
	{0x4EF5D1FF6B644189ULL, "follower_banter6164447.cl"},
	{0x4EF5F907E834E93CULL, "849204371.ogg"},
	{0x4EF5FF213BDD8F33ULL, "art\\characters\\creatures\\desiredemon\\animation\\singletarget.evt"},
	{0x4EF6104F667BA095ULL, "zz_followers_banter6159513.cl"},
	{0x4EF73EE5C9B83589ULL, "lgt260_way_blood_mage6148384.cl"},
	{0x4EF7689539DBDB5FULL, "6053010_m.fxe"},
	{0x4EF79261DAE83E01ULL, "6115512_m.fxe"},
	{0x4EF7978BECC0B481ULL, "dae000_madam_lusine6152162.cl"},
	{0x4EF7EB542781EE57ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaritrooper_overlay.anb"},
	{0x4EF83EFFA26F00C6ULL, "dae000ip_crowd_lm4_talk6.utp"},
	{0x4EF8FBF4CCD4C83DULL, "6115506_f.fxe"},
	{0x4EF97F4BC1ECC9CAULL, "hlac_lt_0_lowtown_docks_n_-196610_5x.dds"},
	{0x4EF9A692860213E6ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_bird01.mao"},
	{0x4EF9AD56B646F8E3ULL, "vxca_caverns1_j31_0_0.vlm"},
	{0x4EFA1E9849FF2151ULL, "vxlt_undercityencounter_1k1_0_0.vlm"},
	{0x4EFA24D5C8019E24ULL, "vxlt_warehouse_yl_0_0.vlm"},
	{0x4EFA51FBD620B443ULL, "con_ico_lie.dds"},
	{0x4EFAC671F943AB2CULL, "mag120_macha6080685.cl"},
	{0x4EFB217AD734515BULL, "follower_banter6161099.cl"},
	{0x4EFB4C9D5042289CULL, "vxca_cavern1_haunted_icx_0_0.vlm"},
	{0x4EFBECE838B1749BULL, "6122497_m.fxe"},
	{0x4EFC4F227CD48259ULL, "mag100_cullen6093801.cl"},
	{0x4EFC57BF9A6DD4F2ULL, "qcr000_arishok6059994.cl"},
	{0x4EFC85DB9BEC3406ULL, "w_bonepitdestroyed_a3_16_0_ll.mmh"},
	{0x4EFC89AE0DB787DDULL, "w_sundermount_56_0.mao"},
	{0x4EFC8DAE0DB78E16ULL, "w_sundermount_56_0.mmh"},
	{0x4EFC97AE0DB79F14ULL, "w_sundermount_56_0.msh"},
	{0x4EFCC3F238E45165ULL, "hlac_lt_0_lowtown_docks_n_-4_2y.dds"},
	{0x4EFCF8213BE3A274ULL, "art\\characters\\creatures\\desiredemon\\animation\\singletarget.gad"},
	{0x4EFD28833612E90FULL, "vxlt_warehouse_n_rs_0_0.vlm"},
	{0x4EFDC0EAA257EF2EULL, "fci_floor_01.msh"},
	{0x4EFDDAEAA2581B1CULL, "fci_floor_01.mmh"},
	{0x4EFE00CDAE1F8539ULL, "6086205_m.fxe"},
	{0x4EFE46DE6EF05589ULL, "vxht_0_hightown_n_aie_0_0.vlm"},
	{0x4EFE97DC4DF2B1C7ULL, "423295994.ogg"},
	{0x4EFED505D114D0FDULL, "vxlt_0_lowtown_docks_a3_n_czd_0_0.vlm"},
	{0x4EFEE28AE135C495ULL, "zz_followers_banter.cnv"},
	{0x4EFEFFBD60396D9FULL, "6068899_f.fxe"},
	{0x4EFF0399AB29D084ULL, "dae000_pc_home_ambients6208354.cl"},
	{0x4EFF3164C5037BF8ULL, "hlac_w_blightlands_65532_0z.dds"},
	{0x4EFFAC3973E693D4ULL, "553861804.ogg"},
	{0x4EFFBBEED7F73A4AULL, "hlac_w_bonepitdestroyed_a3_0_13x.dds"},
	{0x4EFFC42D5C6F1FEBULL, "nat380_sibling6207568.cl"},
	{0x4EFFE5E587A37015ULL, "gen_im_trs_slv_lgt_glv_r0.uti"},
	{0x4F012D4F87B0FDABULL, "134839476.ogg"},
	{0x4F014D348397DCFFULL, "plt_cod_cha_teagan.plo"},
	{0x4F01C68221CEA91CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomedclosed.mao"},
	{0x4F01D28221CEBDBFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomedclosed.mmh"},
	{0x4F01D88221CEC861ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomedclosed.msh"},
	{0x4F0407213BE9DA52ULL, "art\\characters\\creatures\\desiredemon\\animation\\singletarget.ani"},
	{0x4F040B09F0E14FCAULL, "454744879.ogg"},
	{0x4F0474C2C3732D4AULL, "6132670_f.fxe"},
	{0x4F054EDBDF69DAF7ULL, "6134865_m.fxe"},
	{0x4F05F7C956FB7EE8ULL, "aip_evelina.gda"},
	{0x4F06085359610E42ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killervase_sa.dds"},
	{0x4F06AB94C4C0EE68ULL, "8872456.ogg"},
	{0x4F06B1D68EA68CEDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_loop.ani"},
	{0x4F06BB956D4FF0ACULL, "119893227.ogg"},
	{0x4F06C3D385B2CD10ULL, "vxlt_0_lowtown_docks_a3_cwu_0_0.vlm"},
	{0x4F071C0DBA885B9DULL, "hlac_ht_brothel_n_3374_1z.dds"},
	{0x4F0759FE1826AB60ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_woodstack1a.phy"},
	{0x4F07796939A39756ULL, "mag200_feynriel6056922.cl"},
	{0x4F07C1439CDB525BULL, "hlac_lt_gallowsprison_n_-1_2z.dds"},
	{0x4F0813D195F48D71ULL, "mer240cr_miner1.utc"},
	{0x4F081D9E0E982D84ULL, "gen00fl_merrill6122216.cl"},
	{0x4F0825775638B174ULL, "mag370_meredith6172398.cl"},
	{0x4F086111F6374A91ULL, "485925316.ogg"},
	{0x4F08D9721D192368ULL, "da2terrain_terrain_tools_p0.vsh.fxo.dx11"},
	{0x4F08F7D29E195D8AULL, "vxca_caverns2_aa2_0_0.vlm"},
	{0x4F09130E7FECAEA6ULL, "mag210_arianni6031708.cl"},
	{0x4F0925F4A9293942ULL, "371939374.ogg"},
	{0x4F09580A93158F6EULL, "vxca_cavern1_haunted_fms_0_0.vlm"},
	{0x4F099D2AC5B35CE6ULL, "plt_ave300pt_favour_fault.nss"},
	{0x4F09F4976FEB04B1ULL, "792202933.ogg"},
	{0x4F0A0BDECA5F5931ULL, "hlac_lt_0_lowtown_a2_n_c_0_7z.dds"},
	{0x4F0A7AEEBA072B14ULL, "779375378.ogg"},
	{0x4F0ACED9C9602AB2ULL, "vxlt_gallowstemplar_f_3qw_0_0.vlm"},
	{0x4F0AE06D124601F3ULL, "vxlt_blackemporium_vz_0_0.vlm"},
	{0x4F0B0A785443351DULL, "6142901_m.fxe"},
	{0x4F0B5975A3D9B518ULL, "zz_mrl_debug6184218.cl"},
	{0x4F0B9A602809018FULL, "vxlt_warehouse_n_yr_0_0.vlm"},
	{0x4F0BE38D51AF1780ULL, "vxlt_hangedman_n_260_0_0.vlm"},
	{0x4F0C245DD1FC4AF4ULL, "6140212_m.fxe"},
	{0x4F0C30AA4CFF2F18ULL, "752642738.ogg"},
	{0x4F0C421C49D42642ULL, "6050919_m.fxe"},
	{0x4F0D0ACB69C96DF0ULL, "hla_ht_0_hightown_n_-1_0y.dds"},
	{0x4F0D122539895CBBULL, "vxht_mansion_n_3hj_0_0.vlm"},
	{0x4F0D19FDA24FF1C1ULL, "vxht_0_hightown_a3_bel_0_0.vlm"},
	{0x4F0D5BC06E485249ULL, "6110696_m.fxe"},
	{0x4F0D9EF70E9BE031ULL, "rdr111dg_martin6053564.cl"},
	{0x4F0DFE6FB5764C9FULL, "419777367.ogg"},
	{0x4F0E1981E8E5B79EULL, "1035429532.ogg"},
	{0x4F0E2D27B953DCE1ULL, "6150519_m.fxe"},
	{0x4F0E37D213787952ULL, "gen00fl_aveline6137666.cl"},
	{0x4F0E51A679B8B9C4ULL, "vxlt_0_lowtown_n_dow_0_0.vlm"},
	{0x4F0F90EB769CCA40ULL, "character_character_p12.psh.fxo.dx11.assembly"},
	{0x4F103391E8D343E7ULL, "467736653.ogg"},
	{0x4F104486C52CA62AULL, "231917726.ogg"},
	{0x4F104814893C588FULL, "vxlt_darktown_4uz_0_0.vlm"},
	{0x4F11FE48C0193512ULL, "vxlt_0_lowtown_docks_a3_n_d3o_0_0.vlm"},
	{0x4F1201F91678912AULL, "1066487.ogg"},
	{0x4F1224C2D701B36FULL, "vxht_mansion2b_2k5_0_0.vlm"},
	{0x4F1242FED57D4609ULL, "gen00fl_fenris6126831.cl"},
	{0x4F124782D61E2AEFULL, "gencr_gate_slave_crouching.utc"},
	{0x4F12EAB7117F3DBBULL, "dus_clo_li.mao"},
	{0x4F138FD91627BDD0ULL, "280653710.ogg"},
	{0x4F138FF2C57D734CULL, "hlac_ca_cavern1_haunted_24424_11x.dds"},
	{0x4F139D8DE9E80899ULL, "vxdr_primevaltomb_75m_0_0.vlm"},
	{0x4F13B036D4E448D0ULL, "160617100.ogg"},
	{0x4F1460DDEED5A60CULL, "hlac_lt_0_lowtown_a2_n_c_-2_4y.dds"},
	{0x4F14803C4B7021BBULL, "661698540.ogg"},
	{0x4F15792C6D4CC5FDULL, "vxlt_0_lowtown_docks_n_as2_0_0.vlm"},
	{0x4F15D73BC604C644ULL, "196916354.ogg"},
	{0x4F1628A2694C9745ULL, "6119893_m.fxe"},
	{0x4F1652B267F7145EULL, "475513395.ogg"},
	{0x4F1675ACC8764F22ULL, "6120422_f.fxe"},
	{0x4F16C011B0599050ULL, "vxlt_foundry_n_wq_0_0.vlm"},
	{0x4F173356D22F8D48ULL, "mag311_emile_de_launcet6198480.cl"},
	{0x4F17422EE6C13444ULL, "vxdr_primevaltomb_4xe_0_1.vlm"},
	{0x4F17C9DEEE1E02E9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_vael_0s.dds"},
	{0x4F17F3D62BF77CB0ULL, "vxlt_hangedman_n_21u_0_0.vlm"},
	{0x4F1884BCEA032236ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.dura.evt"},
	{0x4F18871333EE232AULL, "dae000ip_crowd_lm2_talk8.utp"},
	{0x4F19129A6CD26D38ULL, "vxw_bonepitdestroyed_a3_2ak_0_0.vlm"},
	{0x4F19352CDA0208A4ULL, "405055577.ogg"},
	{0x4F19E1C139A0FDABULL, "mag320_orsino6071257.cl"},
	{0x4F1A1AE7C906467AULL, "6187930_m.fxe"},
	{0x4F1A480E3F529233ULL, "6114011_f.fxe"},
	{0x4F1A6BDE7825E012ULL, "art\\characters\\creatures\\spider\\animation\\rdy.gad"},
	{0x4F1A7209AACE23BBULL, "6129441_m.fxe"},
	{0x4F1A9094DA5577FFULL, "91061792.ogg"},
	{0x4F1AB776AE399463ULL, "art\\characters\\playercharacter\\humanmale\\bdy_meda.msh"},
	{0x4F1AC576AE39ABBDULL, "art\\characters\\playercharacter\\humanmale\\bdy_meda.mmh"},
	{0x4F1AC976AE39B2CEULL, "art\\characters\\playercharacter\\humanmale\\bdy_meda.mao"},
	{0x4F1ADA8DD902661CULL, "6162357_m.fxe"},
	{0x4F1AFE4976CA2F18ULL, "dae000cr_grand_cleric6147663.cl"},
	{0x4F1B8FFBB3CF8AF1ULL, "hlac_lt_gallowstemplar_n_2394_10y.dds"},
	{0x4F1BBD7FE0BE4B05ULL, "6172304_m.fxe"},
	{0x4F1BF64A22730583ULL, "vxht_brothel_n_4d2_0_0.vlm"},
	{0x4F1C14178D0E3450ULL, "art\\vfx\\models\\vfx_forearm_closed.msh"},
	{0x4F1C1AF956D4CF3CULL, "942273407.ogg"},
	{0x4F1C22178D0E4C6AULL, "art\\vfx\\models\\vfx_forearm_closed.mmh"},
	{0x4F1C26178D0E52A1ULL, "art\\vfx\\models\\vfx_forearm_closed.mao"},
	{0x4F1C6137C44A9496ULL, "mag210_arianni6024539.cl"},
	{0x4F1C891610DD9AB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_2.ani"},
	{0x4F1CF01108C91EC9ULL, "vxlt_backalley_g_1xh_0_0.vlm"},
	{0x4F1E1F67CB0957EDULL, "hlac_lt_0_lowtown_docks_a3_-65537_5x.dds"},
	{0x4F1E5661FC8D297BULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.impa.ani"},
	{0x4F1EB22AC5C5DC21ULL, "plt_ave300pt_favour_fault.plo"},
	{0x4F1F46DA55B4C4A1ULL, "dae111ip_letter_desk.utp"},
	{0x4F1F7BBCEA0931C1ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.dura.gad"},
	{0x4F1F89D1D9295000ULL, "hlac_lt_0_lowtown_docks_n_65534_2x.dds"},
	{0x4F202FC4E7F1B857ULL, "vxlt_0_lowtown_docks_a3_d0t_0_0.vlm"},
	{0x4F203337CC3D5492ULL, "384610749.ogg"},
	{0x4F208761098CF06CULL, "19550749.ogg"},
	{0x4F20D18F5A678302ULL, "30277373.ogg"},
	{0x4F216EDE782C0449ULL, "art\\characters\\creatures\\spider\\animation\\rdy.evt"},
	{0x4F218CD09AB904F2ULL, "hf_gencr_coterie_thug.mor"},
	{0x4F21F1C0A6F5833DULL, "plt_lgt000pt_gangs.plo"},
	{0x4F22190808B08D98ULL, "hlac_ht_mansion_n_3366_0x.dds"},
	{0x4F227B7A63B6A9C9ULL, "mer340cr_cavern_guard.utc"},
	{0x4F2346C8E47442BDULL, "mag110_vincento6017805.cl"},
	{0x4F23600BECCFC225ULL, "195365536.ogg"},
	{0x4F23941610E3CC53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_2.gad"},
	{0x4F23B9EAA278B2E7ULL, "fci_floor_01.phy"},
	{0x4F23F23C71E3FF6EULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs107a.phy"},
	{0x4F2409A128E140CFULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.mmh"},
	{0x4F24209B4F1307B5ULL, "llac_ca_cavern1_haunted_24421_2.dds"},
	{0x4F248DA778964D9DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_chantry_1.mao"},
	{0x4F24AC54250718EFULL, "dae000_meredith6166317.cl"},
	{0x4F252F34D46CC45BULL, "803895936.ogg"},
	{0x4F256CDACAC3DC8EULL, "vxlt_gamlenshouse_n_n8_0_0.vlm"},
	{0x4F25BE6F57542958ULL, "314861874.ogg"},
	{0x4F2635A782D5CA94ULL, "vxht_0_hightown_n_5i8_0_0.vlm"},
	{0x4F267CBCEA0F526FULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.dura.ani"},
	{0x4F269CF85FAEF276ULL, "hlac_lt_0_lowtown_a2_65537_0y.dds"},
	{0x4F26AC608B6D7E8CULL, "vxht_0_hightown_n_a2m_0_0.vlm"},
	{0x4F27263EF2CC085DULL, "vxlt_0_lowtown_docks_a3_d2p_0_0.vlm"},
	{0x4F273A7F6F2FBC18ULL, "zz_followers_banter6159381.cl"},
	{0x4F277D48F7C76AC4ULL, "6108630_m.fxe"},
	{0x4F27950B21901D01ULL, "726511595.ogg"},
	{0x4F27AD9ACCDC0009ULL, "hlac_ca_cave_2193_1x.dds"},
	{0x4F27EFE403290F4EULL, "mag100_cullen6095176.cl"},
	{0x4F27F20D5ACC5413ULL, "vxlt_warehouse_15z_0_0.vlm"},
	{0x4F27F4C0A6F9F46AULL, "plt_lgt000pt_gangs.nss"},
	{0x4F27FF8E85CEC07FULL, "vxlt_warehouse_n_x1_0_0.vlm"},
	{0x4F28CC111D9F6827ULL, "467893914.ogg"},
	{0x4F28D6947A7BB496ULL, "805404223.ogg"},
	{0x4F28E632BF9CD814ULL, "dae000_bartrand6032008.cl"},
	{0x4F291E31FE330F49ULL, "6095353_m.fxe"},
	{0x4F29FCB20715C068ULL, "vxw_bonepitmining_a2_251_0_0.vlm"},
	{0x4F2A11167A547D32ULL, "788472974.ogg"},
	{0x4F2A5CD88908389BULL, "vxw_sundermount_2nb_0_0.vlm"},
	{0x4F2ABC1FD47E7E15ULL, "19120985.ogg"},
	{0x4F2B483D42122C44ULL, "hlac_lt_0_lowtown_n_-65535_2y.dds"},
	{0x4F2B95CBD6451B1CULL, "413911207.ogg"},
	{0x4F2BD27B527AD05CULL, "vxca_caverns2prc_aft_0_0.vlm"},
	{0x4F2C5E61FC997472ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.impa.evt"},
	{0x4F2CAD42E702177FULL, "vxlt_darktown_4t9_0_0.vlm"},
	{0x4F2D08EEE91DFB82ULL, "dae000_pc_home_ambients6208477.cl"},
	{0x4F2D55F3F486C497ULL, "6018882_m.fxe"},
	{0x4F2D76FE1847ABF5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_woodstack1a.msh"},
	{0x4F2D78FE1847AF6BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_woodstack1a.mmh"},
	{0x4F2D84FE1847C308ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_woodstack1a.mao"},
	{0x4F2DBDCC8F137BCEULL, "gencr_talvashoth_mage.utc"},
	{0x4F2E0575126FD710ULL, "933724299.ogg"},
	{0x4F2E50950DA89092ULL, "501882471.ogg"},
	{0x4F2E6E13DE3A7C9CULL, "qun160cr_rebel_leader6035898.cl"},
	{0x4F2EE499C4A62E88ULL, "6113956_f.fxe"},
	{0x4F2EE5582E7F4734ULL, "gen00fl_fenris6113861.cl"},
	{0x4F2EFD896A2412EFULL, "6201590_m.fxe"},
	{0x4F2F0396E1D5C43AULL, "6204321_m.fxe"},
	{0x4F2F087CBAEDE46FULL, "vxht_mansion2b_2a7_0_0.vlm"},
	{0x4F2F2FA4F7C83C87ULL, "hlac_lt_backalley_n_65537_2x.dds"},
	{0x4F2F3F08699FCE4EULL, "vxlt_darktown_4u7_0_0.vlm"},
	{0x4F2F56DE783818B8ULL, "art\\characters\\creatures\\spider\\animation\\rdy.ani"},
	{0x4F2F76BB55BA8868ULL, "111084177.ogg"},
	{0x4F2F7CF4DF0C5BE7ULL, "140732459.ogg"},
	{0x4F2F7DA913F5C418ULL, "369674250.ogg"},
	{0x4F2FB04A8A1BFB67ULL, "942827771.ogg"},
	{0x4F3018CB6C1115E6ULL, "vxw_encounter2_camp_ps_0_1.vlm"},
	{0x4F308C5CF4925115ULL, "113779900.ogg"},
	{0x4F30AE8C6B77305BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_adjust_talk.ani"},
	{0x4F30FA1B64A820C1ULL, "53904927.ogg"},
	{0x4F312D1016CE8242ULL, "gen_im_npc_dwf_arm_hlm_lgta.uti"},
	{0x4F31A22C469A1BACULL, "follower_banter6171091.cl"},
	{0x4F327D61FC9E1405ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.impa.gad"},
	{0x4F3304AE6F4429ACULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha11a.mmh"},
	{0x4F330AAE6F44339EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha11a.msh"},
	{0x4F335CE3AEAC52BBULL, "vxlt_hangedman_n_1qq_0_0.vlm"},
	{0x4F336A168EE8F530ULL, "vxht_bartrand_n_4fk_0_0.vlm"},
	{0x4F338D9596F036DCULL, "vxlt_0_lowtown_a2_n_c_hky_0_0.vlm"},
	{0x4F344DA426775B09ULL, "329614525.ogg"},
	{0x4F34BEF6A8D3D49FULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunaricontroller_d.dds"},
	{0x4F353C04E6218416ULL, "6055615_m.fxe"},
	{0x4F354FEEB8F9F761ULL, "6080669_m.fxe"},
	{0x4F35DB58E341EADEULL, "hlac_lt_darktown_-1_39z.dds"},
	{0x4F36551DD3E69E82ULL, "345127105.ogg"},
	{0x4F365C8A4DAFF7EEULL, "mag291_vincento6027386.cl"},
	{0x4F36D4C2A9B664E0ULL, "6109750_m.fxe"},
	{0x4F36EC64F283E6E1ULL, "vxlt_undercityencounter_x3_0_0.vlm"},
	{0x4F3752E6DEB9788AULL, "zz_vrc_debug6200194.cl"},
	{0x4F38B7915A566693ULL, "character_character_p178.psh.fxo.dx11.assembly"},
	{0x4F38CEB57AFE511BULL, "vxw_sundermount_2se_0_0.vlm"},
	{0x4F39F4166C5BAAEAULL, "dre000_vampire6064154.cl"},
	{0x4F3A5653456F6E94ULL, "lgt120_orz_bln_traitor_lead.cnv"},
	{0x4F3B181E6246C4D0ULL, "sst_ss_templars_f_1.gda"},
	{0x4F3B61614172BF0BULL, "238077356.ogg"},
	{0x4F3C16487C0B893EULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.impa.evt"},
	{0x4F3D19A0E737AB94ULL, "vxht_0_hightown_a3_c9t_0_0.vlm"},
	{0x4F3D87C5B8FB6717ULL, "6117624_m.fxe"},
	{0x4F3D8C9737137E6EULL, "lgt220_shp_mag_guard6149894.cl"},
	{0x4F3D8E944BDF5772ULL, "811879835.ogg"},
	{0x4F3DD4E150E64B42ULL, "hightown_male_elfservnt_mob6170085.cl"},
	{0x4F3DD7B09BC35ACEULL, "hlac_lt_warehouse_0_5x.dds"},
	{0x4F3E0FDDA5623D1DULL, "vxlt_gallowstemplar_n_4o5_0_0.vlm"},
	{0x4F3E3853457324D4ULL, "lgt120_orz_bln_traitor_lead.bnk"},
	{0x4F3E385A1B6B8BCCULL, "zz_lgt_debug.cnv"},
	{0x4F3E52B31D78E660ULL, "320327780.ogg"},
	{0x4F3E6949950E0F22ULL, "6042273_m.fxe"},
	{0x4F3E726025787ECEULL, "6127680_m.fxe"},
	{0x4F3F45B443892B6FULL, "6147654_m.fxe"},
	{0x4F3F80EBFD5F34A0ULL, "6031313_f.fxe"},
	{0x4F3FAC12F3A35112ULL, "hlac_w_encounter1_5_0z.dds"},
	{0x4F4053CAAE83ECF0ULL, "gen00fl_merrill6144959.cl"},
	{0x4F40751AB3D892C8ULL, "vxlt_undercityencounter_x6_0_0.vlm"},
	{0x4F409DC829453017ULL, "dae000_unhappy_mage6165325.cl"},
	{0x4F40B19F62E07D9EULL, "468135076.ogg"},
	{0x4F411D56C1C00626ULL, "w_blightlands_61_0_lc.mmh"},
	{0x4F41769A3238E47EULL, "682521538.ogg"},
	{0x4F43258E96AC2318ULL, "220759168.ogg"},
	{0x4F432D487C11CF69ULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.impa.gad"},
	{0x4F434A6961C7FAA0ULL, "6068477_m.fxe"},
	{0x4F4395A976A6E734ULL, "dae310ar_docks_nt.are"},
	{0x4F43B064DA477F97ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.impa.gad"},
	{0x4F43CFAE469DE566ULL, "lgt210_qun_blk_qunari6142395.cl"},
	{0x4F43DE3016AB2CE0ULL, "6039304_f.fxe"},
	{0x4F443488269A315EULL, "vxlt_0_lowtown_a2_e1a_0_0.vlm"},
	{0x4F444622F5DF2BA4ULL, "vxht_0_hightown_n_9v1_0_0.vlm"},
	{0x4F4483665A4C4761ULL, "6020065_m.fxe"},
	{0x4F44B9E5F5D40AA3ULL, "mag291_torpor6026982.cl"},
	{0x4F4506FD4175D5EAULL, "634902026.ogg"},
	{0x4F45D2494C627B12ULL, "dae000cr_cerimon.utc"},
	{0x4F45D58C6B89CEE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_adjust_talk.gad"},
	{0x4F45F221A17C6997ULL, "vxdr_primevaltomb_53s_0_0.vlm"},
	{0x4F46802F28BED009ULL, "vxht_0_hightown_n_avh_0_0.vlm"},
	{0x4F468946BBFF3651ULL, "mag131_keran6086250.cl"},
	{0x4F4729A7708A3C6EULL, "zz_one_debug6051515.cl"},
	{0x4F473B3AD532595EULL, "docks_male_human_worker_sta6175946.cl"},
	{0x4F4740926CF4699BULL, "vxw_sundermount_2td_0_0.vlm"},
	{0x4F477865D6CEEA69ULL, "hlac_dr_primevaltomb_659_16z.dds"},
	{0x4F4785899241660CULL, "vxht_0_hightown_a2_be0_0_0.vlm"},
	{0x4F47858EFF8C77C6ULL, "vxlt_shop_ee_0_0.vlm"},
	{0x4F48DF42F7C6A8CEULL, "ser221_gascard6149159.cl"},
	{0x4F49FE487C179DB7ULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.impa.ani"},
	{0x4F4A4C7DCFEDC3A3ULL, "337978099.ogg"},
	{0x4F4A500D7AA5CCA6ULL, "487092345.ogg"},
	{0x4F4A78F9FA1758DFULL, "vxht_keep_n_3fx_0_0.vlm"},
	{0x4F4AC564DA4DC224ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.impa.evt"},
	{0x4F4AFA18D2056770ULL, "t3_shd_yew.tnt"},
	{0x4F4B15392449EE18ULL, "331369021.ogg"},
	{0x4F4B8B7DD92DC414ULL, "vxht_0_hightown_a3_c2u_0_0.vlm"},
	{0x4F4BD91198A70B27ULL, "15427480.ogg"},
	{0x4F4C6F53BB92C401ULL, "38602200.ogg"},
	{0x4F4CA10BFA67B84AULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs107a_0s.dds"},
	{0x4F4CBD467540EC50ULL, "6036913_m.fxe"},
	{0x4F4CCDFB92140127ULL, "vxht_0_hightown_b1w_0_0.vlm"},
	{0x4F4D0AA12904470CULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.anb"},
	{0x4F4D13AEF3FC42DFULL, "565018084.ogg"},
	{0x4F4D1B57B62ABC9DULL, "6042090_m.fxe"},
	{0x4F4D3C82FBDBD85DULL, "dae000_pc_home_ambients6208177.cl"},
	{0x4F4D67F41D83C0BAULL, "517651784.ogg"},
	{0x4F4E12BD71445234ULL, "6118077_m.fxe"},
	{0x4F4E655EE374E747ULL, "651210620.ogg"},
	{0x4F4EBCDFBD1A7D2AULL, "vxht_0_hightown_a3_b1w_0_0.vlm"},
	{0x4F4ECC13835EF5EEULL, "w_blightlands_mmi.gff"},
	{0x4F4FB564BB170780ULL, "6163879_m.fxe"},
	{0x4F4FFDAAF5ECCED8ULL, "zz_dre_debug6204369.cl"},
	{0x4F5012BEF94A076EULL, "vxlt_0_lowtown_a2_n_c_e1l_0_0.vlm"},
	{0x4F5039B19E77BDBFULL, "mind_dust.mao"},
	{0x4F504DD072D52630ULL, "vxht_bartrand_n_4pi_0_0.vlm"},
	{0x4F50AF893F07DC38ULL, "one100_party_comments6106772.cl"},
	{0x4F510D3787E036FEULL, "6030420_f.fxe"},
	{0x4F518A085E749B46ULL, "hlac_lt_0_lowtown_docks_a3_-1_3x.dds"},
	{0x4F51C80AAE155795ULL, "hm_amb_guard4.mor"},
	{0x4F52C70D8A201F2AULL, "pro000_colonel_ewald6040543.cl"},
	{0x4F53BBD15CA5675AULL, "vxlt_gallowscourtyard_n_6yx_0_0.vlm"},
	{0x4F54455DA456AD5EULL, "vxw_encounter3_ef_0_0.vlm"},
	{0x4F5456CB4B74CECFULL, "hlac_ht_keep_3739_0z.dds"},
	{0x4F548EEDC441B55AULL, "20506869.ogg"},
	{0x4F549DE0DA75D512ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_3x.dds"},
	{0x4F5518FB2449F3A2ULL, "vxlt_foundry_n_10e_0_0.vlm"},
	{0x4F55B26F5EEA821BULL, "vxw_sundermount_2yg_0_0.vlm"},
	{0x4F55B9007ACF51C1ULL, "llac_w_sundermount_262141_0.dds"},
	{0x4F55BDBDEA4B6783ULL, "vxht_brothel_n_4x6_0_0.vlm"},
	{0x4F56259D3D87B8DDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_02fern.mao"},
	{0x4F56299D3D87BF16ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_02fern.mmh"},
	{0x4F56339D3D87D014ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_02fern.msh"},
	{0x4F567507B1D0B3AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_enter.ani"},
	{0x4F57898E374874A4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw7sm.mmh"},
	{0x4F578D8E37487BF7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw7sm.mao"},
	{0x4F579F8E37489A46ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw7sm.msh"},
	{0x4F57D699E6F7FD4CULL, "6123224_m.fxe"},
	{0x4F57DD64DA58751DULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.impa.ani"},
	{0x4F57EED8EF049296ULL, "vxlt_house_hj_0_0.vlm"},
	{0x4F5815219D421652ULL, "mag100_cullen6206088.cl"},
	{0x4F58223A1A0E2940ULL, "150872393.ogg"},
	{0x4F5840F0CE11E17FULL, "811679007.ogg"},
	{0x4F59906D2750B5B3ULL, "274629803.ogg"},
	{0x4F599FDF8A8F8718ULL, "rdr111dg_harbormaster6015971.cl"},
	{0x4F59B3CDB915AB52ULL, "zz_mer_debug6073704.cl"},
	{0x4F59FD5ACB23DC94ULL, "vxht_0_hightown_a3_c7z_0_0.vlm"},
	{0x4F59FEE36357D863ULL, "729729461.ogg"},
	{0x4F5A29458A8F58B6ULL, "zz_approval_debug6003960.cl"},
	{0x4F5A7FC29FFD5DA6ULL, "and111_lirene.cnv"},
	{0x4F5A958E8EF8CC0AULL, "hlac_lt_blackemporium_1055_4z.dds"},
	{0x4F5AE083B7859568ULL, "hlac_ht_chantry_2534_3z.dds"},
	{0x4F5B0A7918729EE3ULL, "mag110_vincento6016336.cl"},
	{0x4F5B561A0964AFCEULL, "6129880_m.fxe"},
	{0x4F5B943B4464A1ABULL, "vxca_caverns2_82r_0_1.vlm"},
	{0x4F5BF578F2AD26C8ULL, "hlac_lt_0_lowtown_docks_a3_-262144_1z.dds"},
	{0x4F5C4A4BDE33B037ULL, "crust_webs2.dds"},
	{0x4F5C4D909F413CFEULL, "72225981.ogg"},
	{0x4F5C8CA4A98073D5ULL, "498724863.ogg"},
	{0x4F5CF1E60C0B9FC7ULL, "6074348_f.fxe"},
	{0x4F5D271DE1740508ULL, "6072569_f.fxe"},
	{0x4F5D2F3F3A352F6EULL, "vxlt_0_lowtown_h87_0_0.vlm"},
	{0x4F5D81C29FFF9766ULL, "and111_lirene.bnk"},
	{0x4F5DB936DFF3BF92ULL, "zz_isa_debug6133223.cl"},
	{0x4F5DC8F19C96C6F8ULL, "ts_kirk_darktown_02.gda"},
	{0x4F5E668A092B8A80ULL, "vxht_0_hightown_a3_9xe_0_0.vlm"},
	{0x4F5F005E831EF31CULL, "vxw_bonepitdestroyed_a3_1ru_0_0.vlm"},
	{0x4F5FB1112B467820ULL, "vxht_chantry_n_36g_0_0.vlm"},
	{0x4F5FC27CF75304D1ULL, "vxht_bartrand_n_5rg_0_0.vlm"},
	{0x4F6031B04ECAFA9FULL, "atl_startmenu_dxt5_01.dds"},
	{0x4F6062DD00F8DE0AULL, "6174532_m.fxe"},
	{0x4F60CDCE233073A6ULL, "6091829_f.fxe"},
	{0x4F6104F53AEDAF32ULL, "vxca_caverns2_99z_0_0.vlm"},
	{0x4F617CA976BFBA16ULL, "dae310ar_docks_nt.ncs"},
	{0x4F624F13CB0D0956ULL, "vxw_bonepit_4v_0_0.vlm"},
	{0x4F6265BE1554086EULL, "144554698.ogg"},
	{0x4F629A5720497376ULL, "store_lgt_epic3_2.utm"},
	{0x4F6369908CAB2D25ULL, "ave310_donnic6134762.cl"},
	{0x4F637458755246B5ULL, "vxlt_0_lowtown_f1a_0_0.vlm"},
	{0x4F63F464B7098B90ULL, "vxlt_gamlenshouse_n_jz_0_0.vlm"},
	{0x4F64309EFE03B751ULL, "lgt000pt_tombs_final.ncs"},
	{0x4F646D07B1DCE3FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_enter.evt"},
	{0x4F6490E5468D59FCULL, "vxlt_warehouse_n_10u_0_0.vlm"},
	{0x4F6492275CF6FEDEULL, "vxht_keep_6ar_0_0.vlm"},
	{0x4F64F13C721B9431ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs107a.mmh"},
	{0x4F650B3C721BC0EFULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs107a.msh"},
	{0x4F65336739FC8CF7ULL, "vxlt_hangedman_n_1qe_0_0.vlm"},
	{0x4F65BCF9D0BEAC21ULL, "378012996.ogg"},
	{0x4F65C791446E1734ULL, "6201485_m.fxe"},
	{0x4F66030822456616ULL, "mag101_grace6034230.cl"},
	{0x4F661084D94F3037ULL, "hm_qsw_warrens2.mor"},
	{0x4F664876D8EE15CCULL, "6128324_m.fxe"},
	{0x4F6660DD1F414A77ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww12_basic.phy"},
	{0x4F66842366E0D659ULL, "vxht_keep_5xm_0_0.vlm"},
	{0x4F67306BF6D314CFULL, "llac_lt_gallowsprison_n_-1_8.dds"},
	{0x4F68896B6CC414E6ULL, "6119436_m.fxe"},
	{0x4F688A8C9685C5E7ULL, "6059074_m.fxe"},
	{0x4F6894A976C601AFULL, "dae310ar_docks_nt.lst"},
	{0x4F68D41BDB46EBD5ULL, "mrl000_marethari6129231.cl"},
	{0x4F690F05FC58260EULL, "vxw_bonepitmining_a2_1ul_0_0.vlm"},
	{0x4F691B2984C18EAEULL, "vxlt_0_lowtown_a2_n_f4v_0_0.vlm"},
	{0x4F6932644091B5A4ULL, "follower_banter6166680.cl"},
	{0x4F69BBA9D35667D1ULL, "dae000_paivel6111969.cl"},
	{0x4F69CBCBEE2AB9ECULL, "vxht_0_hightown_a3_av3_0_0.vlm"},
	{0x4F69FF20B3A4EC84ULL, "vxlt_0_lowtown_a2_n_c_gja_0_0.vlm"},
	{0x4F6A13EA4D3B0EC6ULL, "qcr220_orsino6061769.cl"},
	{0x4F6A155A1B908B5FULL, "zz_lgt_debug.ncs"},
	{0x4F6A1B39D39BF228ULL, "6031651_m.fxe"},
	{0x4F6A81D83238647BULL, "ts_darkspawn_acc_props.gda"},
	{0x4F6A9224921BD0FDULL, "llac_ht_mansion2_818_56.dds"},
	{0x4F6ACE82A9057CECULL, "112217899.ogg"},
	{0x4F6B32B0612CD05AULL, "vxdr_0_deeproads_f5a_0_0.vlm"},
	{0x4F6B4E07B1E2CDE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_enter.gad"},
	{0x4F6B5818F3399CEEULL, "gen00fl_merrill6129897.cl"},
	{0x4F6B6F000C5596CFULL, "313546405.ogg"},
	{0x4F6B9F1339CABC08ULL, "gen00fl_dog6107575.cl"},
	{0x4F6BD6973BCE09BCULL, "950621571.ogg"},
	{0x4F6BE1E510AA9DC2ULL, "498427059.ogg"},
	{0x4F6C201A40399E17ULL, "dae000_templar_knight6165180.cl"},
	{0x4F6C50886E813900ULL, "t1_tat_blu.tnt"},
	{0x4F6C896033C0E1D6ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden7.msh"},
	{0x4F6C936033C0F2D4ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden7.mmh"},
	{0x4F6CA6F01CA42D73ULL, "686320332.ogg"},
	{0x4F6CDE6A54448558ULL, "mag121_idunna6083927.cl"},
	{0x4F6CFA064CC33644ULL, "6034277_f.fxe"},
	{0x4F6D1B74FB4772E6ULL, "mcr310_argument6075084.cl"},
	{0x4F6E043DA6F88AA3ULL, "vxlt_0_lowtown_docks_apn_0_0.vlm"},
	{0x4F6E3A941AFFB362ULL, "vxht_mansion_n_3z9_0_0.vlm"},
	{0x4F6ED30FEFBD71F5ULL, "6154195_m.fxe"},
	{0x4F6F3FAB15268C26ULL, "6096929_f.fxe"},
	{0x4F6F6D7C8087FB32ULL, "vxht_0_hightown_a7g_0_0.vlm"},
	{0x4F701503B54DBC03ULL, "gen_im_trs_krk_hvy_boo_r2.uti"},
	{0x4F705C05EB4F05DFULL, "hlac_lt_0_lowtown_a2_-65535_3x.dds"},
	{0x4F709D0B67CCB4E1ULL, "401432571.ogg"},
	{0x4F70AB29FB28914CULL, "vxlt_backalley_g_1de_0_0.vlm"},
	{0x4F70F13DDE14ADEEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4_b.ani"},
	{0x4F70F1765E0AD661ULL, "vxht_brothel_n_62x_0_0.vlm"},
	{0x4F713BBDDB5D45D9ULL, "ave121_jeven6105644.cl"},
	{0x4F71B20486EA6CB1ULL, "hlac_ht_0_hightown_-65536_1y.dds"},
	{0x4F71CFFA7C3203D8ULL, "6195852_m.fxe"},
	{0x4F71EB8D0064880BULL, "gen_im_wep_war_shd_act2_03.uti"},
	{0x4F729C7D1AB9F343ULL, "uh_hed_olda_0d.dds"},
	{0x4F72A4B13A74DB5CULL, "vxlt_hangedman_n_26l_0_0.vlm"},
	{0x4F72AFCE1D405C29ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a_1.mao"},
	{0x4F7408E903EBBD8BULL, "dae000cr_serendipity.utc"},
	{0x4F743DD5C9E5895CULL, "149733018.ogg"},
	{0x4F744B41B29922E1ULL, "mrl000_marethari6129370.cl"},
	{0x4F744FDA55FD1468ULL, "dae111ip_letter_desk.ncs"},
	{0x4F74CBA200A1FEE2ULL, "gen00fl_fenris6115379.cl"},
	{0x4F75C926070177B6ULL, "lt_0_lowtown_docks_a3_n_253_0.msh"},
	{0x4F75D326070188B4ULL, "lt_0_lowtown_docks_a3_n_253_0.mmh"},
	{0x4F75D72607018FC7ULL, "lt_0_lowtown_docks_a3_n_253_0.mao"},
	{0x4F75EB22ED1EA5BEULL, "dae000_flemeth6013295.cl"},
	{0x4F767A137E66DAEDULL, "vxca_cave_1o0_0_0.vlm"},
	{0x4F76CED455D94EC0ULL, "588209181.ogg"},
	{0x4F76ED402D37FA4EULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3whitelyrium.mmh"},
	{0x4F76F9402D380EE5ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3whitelyrium.mao"},
	{0x4F7707402D38263CULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3whitelyrium.msh"},
	{0x4F77A198AC367862ULL, "vxlt_hangedman_n_1i4_0_0.vlm"},
	{0x4F77AFFCA2BB71ADULL, "mag360_samson6044911.cl"},
	{0x4F77C42CBF9359E1ULL, "6041859_m.fxe"},
	{0x4F77F894016339C6ULL, "vxlt_gallowscourtyard_78a_0_0.vlm"},
	{0x4F7845718BCF565EULL, "vxht_0_hightown_n_bym_0_0.vlm"},
	{0x4F78B904109EAA01ULL, "371897954.ogg"},
	{0x4F7944B58F46F821ULL, "gen_im_trs_tmp_hvy_hlm_r0.uti"},
	{0x4F79CD61DF859623ULL, "mcr310_argument6075784.cl"},
	{0x4F7A1801B9D3F8EDULL, "vxlt_0_lowtown_h9s_0_0.vlm"},
	{0x4F7A307B15E272B3ULL, "gen_im_npc_arm_hum_glv_rdrd.uti"},
	{0x4F7A86124481C2D4ULL, "1035243077.ogg"},
	{0x4F7A9F35E89AF012ULL, "vxlt_undercitydungeon_1id_0_1.vlm"},
	{0x4F7AD77DB3AEE415ULL, "vxht_chantry_20b_0_0.vlm"},
	{0x4F7B0F1144BD0CC2ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_olda_0n.dds"},
	{0x4F7BC6E650BEB15CULL, "vxlt_0_lowtown_a2_n_c_hrr_0_0.vlm"},
	{0x4F7CA636DB3F35D1ULL, "vxht_bartrand_n_5xe_0_0.vlm"},
	{0x4F7CCD4A58232DE4ULL, "llac_lt_0_lowtown_a3_n_c_-196608_2.dds"},
	{0x4F7DE93DDE1F2A3FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4_b.evt"},
	{0x4F7E1D4B1231F80DULL, "6185221_m.fxe"},
	{0x4F7E222557369FADULL, "mag110_arianni6016840.cl"},
	{0x4F7E4F40D839308CULL, "vxlt_0_lowtown_a2_he6_0_0.vlm"},
	{0x4F7E6414FEF938F2ULL, "art\\vfx\\materials\\chunky\\vfx_c_foff_v_add.mao"},
	{0x4F7E987B4D568BA2ULL, "hlac_ht_bartrand_n_2995_4z.dds"},
	{0x4F7EF6C057E5BBE9ULL, "gen_im_trs_gen_broknshield.uti"},
	{0x4F7F017BB0B0E531ULL, "vxca_cavern1_haunted_ei6_0_1.vlm"},
	{0x4F7F0AEEB7E4C0BAULL, "vxca_cavern1_haunted_j4p_0_0.vlm"},
	{0x4F7F38AC7E195598ULL, "gen00fl_fenris6127057.cl"},
	{0x4F7F6757EE2C4082ULL, "vxlt_undercityencounter_194_0_0.vlm"},
	{0x4F7F7A513184179CULL, "vxdr_0_deeproads_el9_0_2.vlm"},
	{0x4F7FCC17BCB06882ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_c.mmh"},
	{0x4F801D85AEBDCE52ULL, "llac_ht_0_hightown_a3_65536_3.dds"},
	{0x4F8044352A565D0AULL, "6136888_m.fxe"},
	{0x4F807108C9A7974FULL, "gen00fl_isabela6088247.cl"},
	{0x4F80C629ABEFF5F3ULL, "vxlt_undercityserialkiller_1z8_0_0.vlm"},
	{0x4F80D608B4DA8330ULL, "763131782.ogg"},
	{0x4F8113CD8089B37EULL, "zz_lgt_1_debug6135724.cl"},
	{0x4F81280C15D32014ULL, "gut_sp_dot.mao"},
	{0x4F815E1ABC71CDCDULL, "499618637.ogg"},
	{0x4F81654EF4451A15ULL, "vxdr_0_deeproads_f67_0_0.vlm"},
	{0x4F81A520CA674979ULL, "qun110_petrice6059145.cl"},
	{0x4F81AE894B4FF797ULL, "hlac_lt_undercityserialkiller_-1_28y.dds"},
	{0x4F81D3ECCBB5145AULL, "follower_banter6171567.cl"},
	{0x4F81DA9C038BE93AULL, "6122446_f.fxe"},
	{0x4F81EE4430A3B945ULL, "llac_lt_gallowscourtyard_-65537_4.dds"},
	{0x4F81F1046DE8883BULL, "llac_lt_0_lowtown_-131072_1.dds"},
	{0x4F8220C63697F9DDULL, "vxlt_undercityencounter_1jb_0_0.vlm"},
	{0x4F831EC4BC063796ULL, "vxdr_primevalentrance_338_0_0.vlm"},
	{0x4F833E99336EB5B0ULL, "6127170_m.fxe"},
	{0x4F8363C8A4EE0A33ULL, "zz_followers_banter6157373.cl"},
	{0x4F83969D773DBD06ULL, "6113049_m.fxe"},
	{0x4F84CA3DDE251428ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4_b.gad"},
	{0x4F8575794BD3E90BULL, "hlac_lt_0_lowtown_-262144_0z.dds"},
	{0x4F8584E455678A26ULL, "1072730374.ogg"},
	{0x4F858B6D467617F5ULL, "zz_followers_banter6159126.cl"},
	{0x4F862AE6C408F7A3ULL, "6050389_m.fxe"},
	{0x4F862B2E487AEB4DULL, "743917006.ogg"},
	{0x4F865BCF605ED85AULL, "6028537.cut"},
	{0x4F86CD30637C0DE2ULL, "6134588_f.fxe"},
	{0x4F86FAC085046E0AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback_rockwraith.gad"},
	{0x4F872154A058397BULL, "ht_0_hightownexplosion_n_15_0.msh"},
	{0x4F8721F97EA81177ULL, "vxht_mansion_n_3qu_0_0.vlm"},
	{0x4F872B182B1656DCULL, "green25.dds"},
	{0x4F873354A05857C6ULL, "ht_0_hightownexplosion_n_15_0.mao"},
	{0x4F873F54A0586C65ULL, "ht_0_hightownexplosion_n_15_0.mmh"},
	{0x4F88714D329780D8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medd_0n.dds"},
	{0x4F8917AA5D150CE6ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_hed_0d.dds"},
	{0x4F898C7754AA0708ULL, "893691291.ogg"},
	{0x4F89E48C6FA6EEC4ULL, "qun131_taskmaster.stg"},
	{0x4F89F3F18BFD15FDULL, "vxht_mansion2b_3el_0_0.vlm"},
	{0x4F89F6C13EE12335ULL, "897675106.ogg"},
	{0x4F8A54CB10672BBBULL, "6177903_m.fxe"},
	{0x4F8A62FEB4A53B19ULL, "t3_bow_wht.tnt"},
	{0x4F8A714ECA8A2777ULL, "6132828_f.fxe"},
	{0x4F8AB44489D51CB3ULL, "llac_w_encounter2_camp_262142_0.dds"},
	{0x4F8AFED5E6C0CBA8ULL, "786275425.ogg"},
	{0x4F8B315447C07A13ULL, "vxlt_darktown_54l_0_0.vlm"},
	{0x4F8B667004A40A3FULL, "395902270.ogg"},
	{0x4F8B9DFB7FF1DC42ULL, "6084240_m.fxe"},
	{0x4F8CA181FC833138ULL, "vxw_bonepitdestroyed_a3_29c_0_0.vlm"},
	{0x4F8CC11DCA2F490CULL, "vxlt_0_lowtown_a2_hb7_0_0.vlm"},
	{0x4F8D3B892B907E16ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_filler.phy"},
	{0x4F8D61B0F6AD1D80ULL, "llac_lt_gallowstemplar_n_2383_12.dds"},
	{0x4F8D7CCBA9DAD93AULL, "vxca_cavern1_haunted_j3q_0_0.vlm"},
	{0x4F8DDDC0850A5BA1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback_rockwraith.evt"},
	{0x4F8E8E19659C0B70ULL, "l_env_wal_lt_qcr_mcr_dist_combat_bed.wbd"},
	{0x4F8EC162D088F7DBULL, "6132742_m.fxe"},
	{0x4F8F5BA19F720672ULL, "306670868.ogg"},
	{0x4F8F7ACCD3B3D4E4ULL, "hlac_ht_mansion_n_3371_2z.dds"},
	{0x4F8FCD04C5011B34ULL, "art\\vfx\\cinematic\\vfx_arishok_spittle_c.anb"},
	{0x4F8FF48615A06BF5ULL, "285120120.ogg"},
	{0x4F9034877635DB72ULL, "6077839_m.fxe"},
	{0x4F9074D0114BF1ABULL, "100092405.ogg"},
	{0x4F9092A3288E125DULL, "vxlt_undercityencounter_1ga_0_0.vlm"},
	{0x4F911EC85E1D960CULL, "6094517_m.fxe"},
	{0x4F919804C93ABB28ULL, "732002165.ogg"},
	{0x4F92099122B6552DULL, "vxca_caverns1_mines_efz_0_0.vlm"},
	{0x4F9314973BE4DD4FULL, "gen00fl_aveline6139986.cl"},
	{0x4F934750ADE936BBULL, "weapon_vfx_p0.vsh.fxo.dx11"},
	{0x4F9363A4CD2DE897ULL, "vxw_bonepit_ls_0_0.vlm"},
	{0x4F9398E1A6357501ULL, "character_character_p123.psh.fxo"},
	{0x4F93C70A97BDCEE6ULL, "dae000_elegant6166899.cl"},
	{0x4F948B0664DFE5B1ULL, "568941820.ogg"},
	{0x4F94D5B6E36EC8F6ULL, "vxlt_undercitydungeon_1nx_0_0.vlm"},
	{0x4F94EE745328D6D1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\ossuary_sa.dds"},
	{0x4F951397A2EEC098ULL, "vxlt_gallowscourtyard_6wg_0_0.vlm"},
	{0x4F9529577D709D50ULL, "dae000_porfiria6162640.cl"},
	{0x4F955EF32D4B00D4ULL, "6103640_m.fxe"},
	{0x4F95E63042E39206ULL, "6126001_m.fxe"},
	{0x4F96AE5DBEEA204FULL, "477790852.ogg"},
	{0x4F9724892B98ABB7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_filler.msh"},
	{0x4F9736892B98CA1AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_filler.mao"},
	{0x4F973A892B98D129ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_filler.mmh"},
	{0x4F97B5BE084FE795ULL, "6129414_m.fxe"},
	{0x4F982BFCC11CA502ULL, "dae000_bartrand6206872.cl"},
	{0x4F98633B98C79A10ULL, "6017476_f.fxe"},
	{0x4F98642099340BB9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_meda_0d.dds"},
	{0x4F988BC703882040ULL, "6145146_f.fxe"},
	{0x4F98A09B1CE2615AULL, "6171306_m.fxe"},
	{0x4F98CC173BD3EEF0ULL, "vxw_sundermount_30w_0_1.vlm"},
	{0x4F98D9F35ED3C53BULL, "341804180.ogg"},
	{0x4F99BF7CA080B3ECULL, "6122003_m.fxe"},
	{0x4F99C2AEC091369CULL, "vxlt_hangedman_lz_0_1.vlm"},
	{0x4F99D1BBA9992DEFULL, "6076571_f.fxe"},
	{0x4F99FE4BE3FB0758ULL, "mrl000_marethari6129463.cl"},
	{0x4F9A0EC38239188AULL, "vxlt_0_lowtown_docks_a3_cym_0_0.vlm"},
	{0x4F9A1E1D9C4762B4ULL, "vxdr_0_deeproads_fbn_0_0.vlm"},
	{0x4F9A88826945ACD4ULL, "gen00fl_merrill6108524.cl"},
	{0x4F9AA3773125592FULL, "llac_ht_0_hightown_-65537_1.dds"},
	{0x4F9AAC647FB1F6D1ULL, "lgt210ip_ingredient4.utp"},
	{0x4F9AC9F321D3D36EULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzskeleton.mmh"},
	{0x4F9AD5C08514D8F0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback_rockwraith.ani"},
	{0x4F9AD5F321D3E705ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzskeleton.mao"},
	{0x4F9AE3F321D3FF5CULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzskeleton.msh"},
	{0x4F9BC9B2F76FFA72ULL, "hlac_ca_caverns1_mines_24426_1y.dds"},
	{0x4F9BDFBA3658296EULL, "vxht_chantry_n_2s3_0_0.vlm"},
	{0x4F9D0E810B220C88ULL, "w_encounter3_108_0.mmh"},
	{0x4F9D1A810B22212BULL, "w_encounter3_108_0.mao"},
	{0x4F9D2C810B223FB2ULL, "w_encounter3_108_0.msh"},
	{0x4F9D8DA332D50A4DULL, "744098164.ogg"},
	{0x4F9DAC04C50D20F7ULL, "art\\vfx\\cinematic\\vfx_arishok_spittle_c.mmh"},
	{0x4F9DBF9435E7CA8BULL, "mag330ip_to_hideout.utp"},
	{0x4F9DC48A95DA14B0ULL, "dae000_gamlen6082922.cl"},
	{0x4F9EA7D0B1746DC2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_basic_1.mao"},
	{0x4F9EFEBADFCB6416ULL, "llac_w_bonepitdestroyed_a3_327684_0.dds"},
	{0x4F9F24922D9923D9ULL, "6120754_m.fxe"},
	{0x4F9FAE257BDD3C12ULL, "mrl000_marethari6069871.cl"},
	{0x4FA0041B14BA70C2ULL, "299145460.ogg"},
	{0x4FA01FA738F55462ULL, "6109257_m.fxe"},
	{0x4FA04D3FB4E864C2ULL, "hlac_dr_0_deeproads_5839_3x.dds"},
	{0x4FA066EEA91D504CULL, "zz_one_debug6077395.cl"},
	{0x4FA0B2A41D53CBA8ULL, "gen00fl_aveline6106888.cl"},
	{0x4FA0CCDABB19204BULL, "vxlt_0_lowtown_a3_n_c_e2u_0_0.vlm"},
	{0x4FA0F94D83E51124ULL, "w_blightlands_14_0.msh"},
	{0x4FA10B4D83E52F8DULL, "w_blightlands_14_0.mao"},
	{0x4FA10F4D83E536C6ULL, "w_blightlands_14_0.mmh"},
	{0x4FA1837015318402ULL, "llac_w_encounter2_-1_0.dds"},
	{0x4FA19065CEC38E3BULL, "w_sundermount_64_0.msh"},
	{0x4FA19E197D85FEFCULL, "w_woundedcoast_84_0.mao"},
	{0x4FA1A265CEC3AC86ULL, "w_sundermount_64_0.mao"},
	{0x4FA1AA197D86129FULL, "w_woundedcoast_84_0.mmh"},
	{0x4FA1AE65CEC3C125ULL, "w_sundermount_64_0.mmh"},
	{0x4FA1B0197D861D41ULL, "w_woundedcoast_84_0.msh"},
	{0x4FA1B0C0FD54D9DCULL, "75818223.ogg"},
	{0x4FA23D654D71D19CULL, "llac_lt_0_lowtown_a2_n_c_65535_8.dds"},
	{0x4FA2785EF3D47652ULL, "6171885_m.fxe"},
	{0x4FA279DD66998DFBULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdgliding01.mmh"},
	{0x4FA2CF57E34C7A41ULL, "vxlt_0_lowtown_docks_bsz_0_0.vlm"},
	{0x4FA2DD3681FE4FF0ULL, "w_bonepitdestroyed_a3_49_0.rml"},
	{0x4FA3564D921BC006ULL, "gen00fl_aveline6139862.cl"},
	{0x4FA3857494E4D918ULL, "vxlt_gallowscourtyard_6rd_0_0.vlm"},
	{0x4FA3A3C5350E39AAULL, "vxca_caverns1_mines_dpr_0_0.vlm"},
	{0x4FA3B778262E3738ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masb_0t.dds"},
	{0x4FA3C6947B7D4E84ULL, "vxlt_0_lowtown_a2_n_c_gve_0_0.vlm"},
	{0x4FA3E2B06E6AA1C1ULL, "llac_lt_0_lowtown_n_-196609_3.dds"},
	{0x4FA4254CCF5230ACULL, "455965754.ogg"},
	{0x4FA43FD04F4BB6BFULL, "hlac_lt_gallowscourtyard_n_-2_0y.dds"},
	{0x4FA4625B4AF710CEULL, "mag110_samson6018721.cl"},
	{0x4FA489586F2C6C8AULL, "llac_w_bonepitmining_a2_65540_0.dds"},
	{0x4FA4B7123D8125A0ULL, "hightown_market_guard_sta6169613.cl"},
	{0x4FA546A44C454553ULL, "123164639.ogg"},
	{0x4FA5662F67A5BB49ULL, "dae000_arishok6068555.cl"},
	{0x4FA5846A66FAF713ULL, "vxlt_undercitydungeon_yu_0_0.vlm"},
	{0x4FA5B6F58EF2C69DULL, "6118519_m.fxe"},
	{0x4FA5C07789862ECBULL, "vxw_bonepit_nm_0_0.vlm"},
	{0x4FA5D5C0C1B6428EULL, "762003619.ogg"},
	{0x4FA618BD824A5165ULL, "one100dg_vanard6020613.cl"},
	{0x4FA621718FE87BE7ULL, "6043987_m.fxe"},
	{0x4FA6260940C40807ULL, "6155499_m.fxe"},
	{0x4FA65E4AA62A62CFULL, "vxlt_hangedman_n_2cx_0_0.vlm"},
	{0x4FA6DFA4F3CB7E80ULL, "art\\characters\\creatures\\dragon\\c_dragon_eyes.mao"},
	{0x4FA7007F233FA74BULL, "vxca_cavern1_haunted_ii9_0_0.vlm"},
	{0x4FA703F69DDA5F0AULL, "hlac_lt_0_lowtown_docks_-65537_1z.dds"},
	{0x4FA7292A299EBC3EULL, "vxw_bonepitmining_a2_ew_0_0.vlm"},
	{0x4FA75FF9E4E0497AULL, "zz_mcr_debug6088432.cl"},
	{0x4FA79DA0B284A059ULL, "954024481.ogg"},
	{0x4FA7CB17BCD1B929ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_c.anb"},
	{0x4FA894CEC9FA33C8ULL, "vxw_encounter2_ke_0_0.vlm"},
	{0x4FA89874522B14FDULL, "6028813_f.fxe"},
	{0x4FA95E4B360B9F53ULL, "vxht_bartrand_n_6r2_0_0.vlm"},
	{0x4FA9C4EB77D90F73ULL, "vxlt_blackemporium_9g_0_0.vlm"},
	{0x4FA9F80938075C09ULL, "w_encounter1_63_0.mao"},
	{0x4FAA014F3032E2D4ULL, "6096263_m.fxe"},
	{0x4FAA0409380770A2ULL, "w_encounter1_63_0.mmh"},
	{0x4FAA060938077458ULL, "w_encounter1_63_0.msh"},
	{0x4FAA2879D7B4502DULL, "vxht_0_hightown_a2_bzx_0_0.vlm"},
	{0x4FAA4FEB1B68D155ULL, "vxht_0_hightown_a3_7g3_0_0.vlm"},
	{0x4FAB92D9C17C3BF5ULL, "429774527.ogg"},
	{0x4FABF655527D1B10ULL, "hf_ser_alessa.mor"},
	{0x4FAC052165F17A5BULL, "vxdr_primevaltomb_911_0_0.vlm"},
	{0x4FAC5DF9CB469203ULL, "w_sundermount_133_0.rml"},
	{0x4FAC74EB18E57D21ULL, "89117423.ogg"},
	{0x4FAC81425B389424ULL, "vxht_keep_n_6ge_0_0.vlm"},
	{0x4FAD10E97A1F2F1FULL, "follower_banter6148708.cl"},
	{0x4FAD120510F43CB6ULL, "6056827_m.fxe"},
	{0x4FAD57C3CD7C79DBULL, "ser120_ghyslain6104205.cl"},
	{0x4FAD98E3FE5BD456ULL, "141637055.ogg"},
	{0x4FAE13C6931CA725ULL, "llac_lt_0_lowtown_docks_a3_n_-131073_1.dds"},
	{0x4FAE35DD1F7EB6CFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww12_basic.mao"},
	{0x4FAE41DD1F7ECB6CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww12_basic.mmh"},
	{0x4FAE47DD1F7ED55EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww12_basic.msh"},
	{0x4FAE4D81889ECCADULL, "6020744_m.fxe"},
	{0x4FAE8F17C99A149EULL, "llac_ca_caverns1_24424_10.dds"},
	{0x4FAEAB65CECE46F2ULL, "w_sundermount_64_0.anb"},
	{0x4FAF3F6D7489E434ULL, "llac_ht_chantry_2534_18.dds"},
	{0x4FAF6F1B4E1854C2ULL, "56816349.ogg"},
	{0x4FAF721CEEFBC15BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flmb_0s.dds"},
	{0x4FAF7BBC07DD547CULL, "follower_banter6164691.cl"},
	{0x4FB002F175B04881ULL, "342197410.ogg"},
	{0x4FB0509D784D67DAULL, "isa120_hayder6066909.cl"},
	{0x4FB09ADD66A60360ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdgliding01.anb"},
	{0x4FB1852F02A71615ULL, "hlac_lt_0_lowtown_n_65536_8z.dds"},
	{0x4FB1D86EE3AE1912ULL, "vxca_caverns1_mines_im8_0_0.vlm"},
	{0x4FB1F27AC5DB0D81ULL, "vxlt_hangedman_n_283_0_0.vlm"},
	{0x4FB20F4D83F3A98DULL, "w_blightlands_14_0.rml"},
	{0x4FB250F2279ABEC4ULL, "hlac_ht_0_hightown_a3_1_2y.dds"},
	{0x4FB25E33E0E02E9BULL, "lt_0_lowtown_docks_197_0.mmh"},
	{0x4FB26A33E0E04338ULL, "lt_0_lowtown_docks_197_0.mao"},
	{0x4FB27C33E0E06185ULL, "lt_0_lowtown_docks_197_0.msh"},
	{0x4FB2DF00F7603DD2ULL, "vxht_bartrand_n_6r7_0_0.vlm"},
	{0x4FB364F9CB4CBD7BULL, "w_sundermount_133_0.phy"},
	{0x4FB3DC879740F403ULL, "6039803_m.fxe"},
	{0x4FB3F0970D7A94C9ULL, "125281231.ogg"},
	{0x4FB428D280E8DF1DULL, "6117419_m.fxe"},
	{0x4FB45919E0E2EDA7ULL, "hlac_lt_gallowstemplar_f_2391_0y.dds"},
	{0x4FB4594F5A6F7EA9ULL, "hlac_ca_caverns1_mines_24424_3z.dds"},
	{0x4FB45FE0EF72E3F1ULL, "512136999.ogg"},
	{0x4FB475F367D465B9ULL, "6100033_m.fxe"},
	{0x4FB4B1D9A8B30F7AULL, "352908043.ogg"},
	{0x4FB50A267D5A557EULL, "6141572_m.fxe"},
	{0x4FB57AD8D9227866ULL, "vxlt_darktown_6ks_0_0.vlm"},
	{0x4FB5E50F3859B0E4ULL, "llac_ht_mansion2b_818_60.dds"},
	{0x4FB5F741A5A2F4C3ULL, "811951145.ogg"},
	{0x4FB60FA3C872A2F3ULL, "5383019.ogg"},
	{0x4FB613645439B646ULL, "vxht_mansion2_2b3_0_0.vlm"},
	{0x4FB658CEA7724265ULL, "prx_upper.dds"},
	{0x4FB658F45D52D278ULL, "hairalpha_hair_p1.psh.fxo.dx11"},
	{0x4FB6621689C1E41BULL, "m_dae220ar_hightown_nt.bnk"},
	{0x4FB675841DBA5083ULL, "380442590.ogg"},
	{0x4FB6EB0F28918AB0ULL, "227936489.ogg"},
	{0x4FB701D7803393B4ULL, "vxdr_0_deeproads_fhl_0_0.vlm"},
	{0x4FB74CE2FD2683FEULL, "6104917_m.fxe"},
	{0x4FB752BF08F96B24ULL, "vxdr_dungeon_239_0_0.vlm"},
	{0x4FB76475013D0B28ULL, "vxw_woundedcoast_12n_0_0.vlm"},
	{0x4FB7D496DA7FE80DULL, "507961917.ogg"},
	{0x4FB7F57DDE5539A2ULL, "hlac_lt_0_lowtown_0_9z.dds"},
	{0x4FB7F71F1DE2C29DULL, "vxlt_gallowscourtyard_n_7gl_0_0.vlm"},
	{0x4FB86DB02579F486ULL, "862652793.ogg"},
	{0x4FB89F8DA6073710ULL, "6043020_f.fxe"},
	{0x4FB8C300ED5C7A78ULL, "hlac_w_bonepit_65537_0z.dds"},
	{0x4FB94B6CBDAB66B2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lowdoor_sa.dds"},
	{0x4FB9B7868BA3ABF8ULL, "722419809.ogg"},
	{0x4FB9B99112A3BEABULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyo_0s.dds"},
	{0x4FB9E3998BC358F6ULL, "6032061_m.fxe"},
	{0x4FBA49D6207A801EULL, "art\\vfx\\environment\\vfx_tiki_fire.dura.ani"},
	{0x4FBA5A8AF163D9D2ULL, "853854025.ogg"},
	{0x4FBA96006583E4D3ULL, "6088797_m.fxe"},
	{0x4FBAF31F4D2EACA4ULL, "vxht_keep_n_6bj_0_0.vlm"},
	{0x4FBB4E5002FE5243ULL, "vxht_chantry_n_27f_0_0.vlm"},
	{0x4FBB66D946905EACULL, "hf_qmq_minion2.mor"},
	{0x4FBC089FCD7DF242ULL, "324150474.ogg"},
	{0x4FBC4942CEFEDA5BULL, "vxw_encounter2_camp_8c_0_0.vlm"},
	{0x4FBC5B78A4E3A6CFULL, "42909528.ogg"},
	{0x4FBC7F40AB258C69ULL, "117191971.ogg"},
	{0x4FBC986901748B59ULL, "340247438.ogg"},
	{0x4FBCCC59B9C5B2ABULL, "587591246.ogg"},
	{0x4FBCDC8C6FD23ACFULL, "qun131_taskmaster.bnk"},
	{0x4FBCDD8CF27258ADULL, "t3_dey_te2.tnt"},
	{0x4FBDC4508750BF6CULL, "art\\characters\\creatures\\mubari\\animation\\rst_lp.ani"},
	{0x4FBE0B9E0BAAB3F9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_en_charge.ani"},
	{0x4FBE46CC69B9875DULL, "art\\characters\\weapons\\longswords\\w_lsw_009a.phy"},
	{0x4FBE596806E3111CULL, "vxca_caverns2prc_98b_0_0.vlm"},
	{0x4FBED80B62956C3DULL, "6181922_m.fxe"},
	{0x4FBEEF18C87D8C21ULL, "vxlt_foundry_n_12m_0_1.vlm"},
	{0x4FBEF6057BE2CDF9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2.mmh"},
	{0x4FBF12893AD78C7EULL, "137285941.ogg"},
	{0x4FBF2F1A60E39569ULL, "302100418.ogg"},
	{0x4FBF4668436E0596ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_swq.anb"},
	{0x4FBF69E1C113ED98ULL, "6089429_m.fxe"},
	{0x4FBFD49028A331B6ULL, "hlac_w_encounter3_-196609_0y.dds"},
	{0x4FBFE1B45E5919C7ULL, "146276054.ogg"},
	{0x4FBFFE8C6FD4AA37ULL, "qun131_taskmaster.cnv"},
	{0x4FC04A4BD5A43192ULL, "vxca_caverns1_mines_ij9_0_0.vlm"},
	{0x4FC0585F9F70DC19ULL, "gen00fl_merrill6111217.cl"},
	{0x4FC099E8A4D6A43BULL, "6172896_m.fxe"},
	{0x4FC16CB14BB08873ULL, "39978913.ogg"},
	{0x4FC19783D3B85A80ULL, "rog_smokegrenade.dds"},
	{0x4FC19B78E6CF5720ULL, "fen110_fenris6100341.cl"},
	{0x4FC1D502E25CECE5ULL, "6121577_m.fxe"},
	{0x4FC2143A049FD8C3ULL, "follower_banter6154204.cl"},
	{0x4FC262319CE1D59BULL, "docks_scene_dockworkers6176216.cl"},
	{0x4FC26814D933A1F3ULL, "65776169.ogg"},
	{0x4FC29AAF72DEF79DULL, "6112719_m.fxe"},
	{0x4FC39CB197501C98ULL, "6156479_m.fxe"},
	{0x4FC3A1A688FACB6EULL, "vxlt_0_lowtown_a2_n_c_eid_0_0.vlm"},
	{0x4FC3AEA0C4865A39ULL, "6067806_m.fxe"},
	{0x4FC3D150875540AEULL, "art\\characters\\creatures\\mubari\\animation\\rst_lp.gad"},
	{0x4FC3EB7B7A197BABULL, "550674093.ogg"},
	{0x4FC40CE40D8AED3EULL, "vxw_bonepitmining_a2_oi_0_0.vlm"},
	{0x4FC47D162FF1123AULL, "252877322.ogg"},
	{0x4FC4E69E0BB0933BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_en_charge.gad"},
	{0x4FC527ADF1F14E5CULL, "vxht_0_hightown_a3_9x9_0_0.vlm"},
	{0x4FC5297305DB132AULL, "vxlt_shop_ei_0_0.vlm"},
	{0x4FC5334519A0FEC4ULL, "mrl150_hunter6071293.cl"},
	{0x4FC54E24712FA806ULL, "propalpha1_s2y.psh.fxo"},
	{0x4FC5507F6B11AEF4ULL, "vxht_store_cd_0_0.vlm"},
	{0x4FC5809EAA6DDB6CULL, "follower_banter6161252.cl"},
	{0x4FC59FC491DC06BBULL, "126215341.ogg"},
	{0x4FC5AE7BD6091BFBULL, "hlac_ht_bartrand_n_3369_1x.dds"},
	{0x4FC5FD31F530E39DULL, "hlac_lt_0_lowtown_n_1_2z.dds"},
	{0x4FC60C08DC63CFA3ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.anb"},
	{0x4FC6B0F312F5A678ULL, "mag100_thrask6031990.cl"},
	{0x4FC6D934563A8D07ULL, "6125548_m.fxe"},
	{0x4FC6DF71FFB35E55ULL, "6164282_m.fxe"},
	{0x4FC726000B2DED10ULL, "mcr310_argument6074456.cl"},
	{0x4FC781682FD22E73ULL, "dre000st_cut_return_surface.stg"},
	{0x4FC7A0558C6C69ABULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_0.pwk"},
	{0x4FC80068D2279101ULL, "qun200_party_comment6182715.cl"},
	{0x4FC8435E08CB0726ULL, "604929932.ogg"},
	{0x4FC861D62086E6EFULL, "art\\vfx\\environment\\vfx_tiki_fire.dura.evt"},
	{0x4FC879FF3B84022DULL, "6075427_m.fxe"},
	{0x4FC8AE59AE84E38AULL, "6100360_m.fxe"},
	{0x4FC907DD5779FD53ULL, "6083697_m.fxe"},
	{0x4FC931DC1020DF72ULL, "637613426.ogg"},
	{0x4FC958D86400B508ULL, "lgt110_gxa_coa_spy6117459.cl"},
	{0x4FC988194A464DADULL, "vxlt_gallowstemplar_2bg_0_0.vlm"},
	{0x4FC9BEE8035E77F0ULL, "vxlt_undercityencounter_122_0_1.vlm"},
	{0x4FC9F551A0F3193AULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.cess.gad"},
	{0x4FCA497146296730ULL, "6097549_m.fxe"},
	{0x4FCA9F90812F40E8ULL, "vxdr_0_deeproads_buo_0_0.vlm"},
	{0x4FCAAB197DA919DCULL, "w_woundedcoast_84_0.anb"},
	{0x4FCABB1FC0F4F2DBULL, "vxw_encounter2_camp_3b_0_0.vlm"},
	{0x4FCACC50875B568DULL, "art\\characters\\creatures\\mubari\\animation\\rst_lp.evt"},
	{0x4FCB2656B2700090ULL, "vxdr_primevaltomb_2ww_0_0.vlm"},
	{0x4FCB3903E4005853ULL, "486848232.ogg"},
	{0x4FCB3F78DF8E2C81ULL, "art\\characters\\creatures\\hurlock\\textures\\w_qui_hlk_0d.dds"},
	{0x4FCC039E0BB6E320ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_en_charge.evt"},
	{0x4FCC2C43628ACC2CULL, "704355718.ogg"},
	{0x4FCC49F0FB4C34E9ULL, "ip_restrict_follower_bethan.ncs"},
	{0x4FCC4B4E4B7052FCULL, "hm_captain_qerth.mor"},
	{0x4FCC720FEF84901DULL, "111318514.ogg"},
	{0x4FCD2F0061B2DCD1ULL, "qcr220_orsino6041788.cl"},
	{0x4FCDC77ADD7B5011ULL, "6089473_m.fxe"},
	{0x4FCE4FF85CB101B3ULL, "hlac_lt_0_lowtown_-1_7y.dds"},
	{0x4FCE62D6208B5398ULL, "art\\vfx\\environment\\vfx_tiki_fire.dura.gad"},
	{0x4FCE704DB3E381D6ULL, "vxlt_0_lowtown_ghx_0_0.vlm"},
	{0x4FCEC53EDF042861ULL, "600522196.ogg"},
	{0x4FCEC58F2092AEA8ULL, "479890294.ogg"},
	{0x4FCF1854A095DF12ULL, "ht_0_hightownexplosion_n_15_0.phy"},
	{0x4FCF18ED26C2DB3DULL, "hlac_ht_mansion2_818_23x.dds"},
	{0x4FCF3BCC69C7E7C9ULL, "art\\characters\\weapons\\longswords\\w_lsw_009a.mao"},
	{0x4FCF47CC69C7FC62ULL, "art\\characters\\weapons\\longswords\\w_lsw_009a.mmh"},
	{0x4FCF49CC69C7FF18ULL, "art\\characters\\weapons\\longswords\\w_lsw_009a.msh"},
	{0x4FCF536655E486CCULL, "gen_im_arm_boo_lgt_act2_02.uti"},
	{0x4FCF94171E755F5BULL, "gen00fl_varric6118999.cl"},
	{0x4FD030D7AC15BD34ULL, "470632266.ogg"},
	{0x4FD067F942F7CBD7ULL, "vxlt_0_lowtown_docks_a3_n_aoo_0_0.vlm"},
	{0x4FD0AB5A69DE711AULL, "439952037.ogg"},
	{0x4FD0B99F49F82BC4ULL, "hlac_w_bonepitmining_a2_458753_0z.dds"},
	{0x4FD0F851A0F93D51ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.cess.evt"},
	{0x4FD120FEA606541BULL, "llac_lt_0_lowtown_a3_n_c_65538_3.dds"},
	{0x4FD131047988C26FULL, "6208639_f.fxe"},
	{0x4FD15D9F962620A6ULL, "vxht_mansion_n_6zt_0_0.vlm"},
	{0x4FD1DD59F6A1C933ULL, "6082027_m.fxe"},
	{0x4FD1EAF9920E8FC3ULL, "6041848_m.fxe"},
	{0x4FD24C47870EE72DULL, "gen00fl_isabela6079674.cl"},
	{0x4FD27EC0FF8105BEULL, "vxw_bonepitmining_a2_jj_0_0.vlm"},
	{0x4FD29D1DEB2ADDD9ULL, "vxht_brothel_n_5ib_0_0.vlm"},
	{0x4FD2E69410FE57B2ULL, "vxlt_gallowstemplar_n_3sq_0_0.vlm"},
	{0x4FD395116ECAC4EBULL, "follower_banter6171854.cl"},
	{0x4FD3E908DC6FD1CCULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.mmh"},
	{0x4FD404463CAC3AACULL, "hlac_lt_warehouse_n_0_7y.dds"},
	{0x4FD433EAA8DC2E3FULL, "ht_bartrand_n_3371.rml"},
	{0x4FD4482EE5293C28ULL, "vxw_woundedcoast_18l_0_0.vlm"},
	{0x4FD4E40D35081C95ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.anb"},
	{0x4FD4EFF876746E48ULL, "mag100_karras6035302.cl"},
	{0x4FD5239E82124495ULL, "vxdr_0_deeproads_cwi_0_0.vlm"},
	{0x4FD52A5CE3BC0930ULL, "844788157.ogg"},
	{0x4FD5A701DA8633CDULL, "40577886.ogg"},
	{0x4FD5BBAF5359D59AULL, "dae000_bartrand6030686.cl"},
	{0x4FD60AF88BB6A6AAULL, "248812032.ogg"},
	{0x4FD6641965D97932ULL, "l_env_wal_lt_qcr_mcr_dist_combat_bed.bnk"},
	{0x4FD689FF7018C4ADULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg01.mmh"},
	{0x4FD693540FBBCC1FULL, "6116865_m.fxe"},
	{0x4FD69BFF7018E333ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg01.msh"},
	{0x4FD6C911A94606E2ULL, "dre100cr_warden_m.utc"},
	{0x4FD7AFC6806ED512ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_aoe_p.mmh"},
	{0x4FD7B910F8D5C877ULL, "365541656.ogg"},
	{0x4FD7D6880B07D425ULL, "771433761.ogg"},
	{0x4FD818159BF41E2CULL, "hlac_ca_caverns2prc_5802_45y.dds"},
	{0x4FD87714B65232E5ULL, "6125448_f.fxe"},
	{0x4FD8990506F0947BULL, "vxca_caverns2_8h1_0_0.vlm"},
	{0x4FD9102FE4733894ULL, "515362402.ogg"},
	{0x4FD9F654B01DC756ULL, "one100dg_kelder6055604.cl"},
	{0x4FDA0A07868D763DULL, "terrain_blend5.psh.fxo"},
	{0x4FDAE16B144F30B0ULL, "6076665_m.fxe"},
	{0x4FDAFB7BFDCB0849ULL, "6173578_m.fxe"},
	{0x4FDB24CA335EE637ULL, "isa111_isabela6067071.cl"},
	{0x4FDB2A67465AE01FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_exit.gad"},
	{0x4FDBC2EFE4301D49ULL, "vxht_mansion2b_2f4_0_0.vlm"},
	{0x4FDBE905B27EB133ULL, "follower_banter6160658.cl"},
	{0x4FDDEC807A8C29D4ULL, "6119508_m.fxe"},
	{0x4FDDF69B7CC08B08ULL, "54817593.ogg"},
	{0x4FDE1051A103F020ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.cess.ani"},
	{0x4FDE814A5B3A29EFULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_flare03.mao"},
	{0x4FDE92C1810F966FULL, "vxht_0_hightown_a3_a6l_0_0.vlm"},
	{0x4FDEAF3311461690ULL, "combat_smaoe_r1_a5.gda"},
	{0x4FDEE58F1B1B4F1CULL, "171867165.ogg"},
	{0x4FDEF521F08949B8ULL, "vxlt_0_lowtown_cmm_0_0.vlm"},
	{0x4FDF35A235BFBBAFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanr.gad"},
	{0x4FDF4BF9CB71CD1AULL, "w_sundermount_133_0.msh"},
	{0x4FDF59F9CB71E553ULL, "w_sundermount_133_0.mao"},
	{0x4FDF5DF9CB71EC60ULL, "w_sundermount_133_0.mmh"},
	{0x4FDF5F395779A5E0ULL, "vfxchunky0_cs.vsh.fxo"},
	{0x4FDF926A8B912E9BULL, "dae000ip_crowd_mf2_talk10.utp"},
	{0x4FE037EC6C30A73EULL, "gen_im_wep_war_shd_act3_05.uti"},
	{0x4FE0CF9064843399ULL, "6037036_m.fxe"},
	{0x4FE10EFADD20F659ULL, "vxht_brothel_n_5la_0_0.vlm"},
	{0x4FE11DA3EBC329F4ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_4x.dds"},
	{0x4FE151453B743737ULL, "58304014.ogg"},
	{0x4FE1AE65CEF9A52EULL, "w_sundermount_64_0.rml"},
	{0x4FE1B97E9CF61826ULL, "ave150tr_start_ambush.utt"},
	{0x4FE21F674660EC0CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_exit.evt"},
	{0x4FE2AF759FEE20F6ULL, "6119462_f.fxe"},
	{0x4FE2C06EF3E96141ULL, "lt_gallowscourtyard_n_50_0_ll.mmh"},
	{0x4FE2D679F636FE00ULL, "6076863_m.fxe"},
	{0x4FE32CF5C6781ABCULL, "hlac_lt_0_lowtown_a2_65536_12y.dds"},
	{0x4FE38D3859A03316ULL, "vxht_0_hightown_n_c9y_0_1.vlm"},
	{0x4FE3EFED9F8CB22DULL, "vxht_0_hightown_a2_bnt_0_0.vlm"},
	{0x4FE41E138BD93C6DULL, "vxht_chantry_n_2ct_0_0.vlm"},
	{0x4FE4218F8DD78597ULL, "vxca_caverns2prc_9h5_0_0.vlm"},
	{0x4FE4578F863D427CULL, "335371437.ogg"},
	{0x4FE4D91012E59E87ULL, "564569293.ogg"},
	{0x4FE5078F6F2D68E0ULL, "vxdr_dungeon_1lr_0_1.vlm"},
	{0x4FE539BEFF1A15CFULL, "vxlt_darktown_5ge_0_0.vlm"},
	{0x4FE591AA6FE8C666ULL, "118084839.ogg"},
	{0x4FE64A5DD07A1541ULL, "hlac_lt_darktown_-1_50x.dds"},
	{0x4FE64AA235C5FE5CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanr.evt"},
	{0x4FE64B254866C14DULL, "6109118_f.fxe"},
	{0x4FE733F8AC8F49C6ULL, "vxw_bonepitdestroyed_a3_18a_0_0.vlm"},
	{0x4FE782FF70272C6AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg01.phy"},
	{0x4FE7A24EA35E4D7AULL, "45013661.ogg"},
	{0x4FE8169E373F2B54ULL, "6121012_f.fxe"},
	{0x4FE82CA9AE0A2559ULL, "static_static_p37.psh.fxo.dx11"},
	{0x4FE88FD66EA983C7ULL, "443664581.ogg"},
	{0x4FE8E8576257938EULL, "lt_0_lowtown_n_54_0.rml"},
	{0x4FE99A826344F0C8ULL, "320489158.ogg"},
	{0x4FE9C27BC9EDD662ULL, "mag291_torpor6027047.cl"},
	{0x4FEA23CAED619863ULL, "375456133.ogg"},
	{0x4FEA3380F8770B55ULL, "zz_followers_banter6158253.cl"},
	{0x4FEA3ACA73DC00A3ULL, "w_bonepitmining_a2_3_0_lc.mmh"},
	{0x4FEAA9A4863C1CD5ULL, "vxca_cavern1_haunted_cn5_0_1.vlm"},
	{0x4FEAB2E6D94C5E2EULL, "6167301_m.fxe"},
	{0x4FEAB458B111216DULL, "677110582.ogg"},
	{0x4FEADB68FFAF5FCFULL, "hlac_lt_0_lowtown_docks_a3_65534_1x.dds"},
	{0x4FEADF157028D4E6ULL, "vxlt_0_lowtown_docks_d2y_0_1.vlm"},
	{0x4FEAE8C0AAC9EC81ULL, "mag291_vincento6152605.cl"},
	{0x4FEAFDAAA2C24E0AULL, "ico_bomb_phoenix.dds"},
	{0x4FEBC5012BBD5590ULL, "vxht_playershouse_n_1sw_0_0.vlm"},
	{0x4FEC43D656F96BB7ULL, "hlac_lt_undercityserialkiller_-1_7x.dds"},
	{0x4FEC9D9EE64D3E2AULL, "llac_lt_0_lowtown_n_131074_0.dds"},
	{0x4FECCB2CC561958CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl04.phy"},
	{0x4FECCD0A04A54EC0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_bda_0n.dds"},
	{0x4FED268FA8261645ULL, "389353617.ogg"},
	{0x4FED2DC1687A66B7ULL, "468218107.ogg"},
	{0x4FED5CFD9FA8ED6BULL, "588840103.ogg"},
	{0x4FEDC095A68D14A1ULL, "zz_vault_debug6203270.cl"},
	{0x4FEEF73D5EDD14EEULL, "vxlt_0_lowtown_a2_n_c_exg_0_0.vlm"},
	{0x4FEEFCB5044D8B4EULL, "6064967_m.fxe"},
	{0x4FEF010C0E9677C4ULL, "hlac_lt_hangedman_-1_8y.dds"},
	{0x4FEF294381730D32ULL, "zz_followers_banter6157639.cl"},
	{0x4FEFA45C94CAAE8FULL, "6017421_m.fxe"},
	{0x4FEFC374695B19B7ULL, "hlac_ca_caverns2prc_5802_1x.dds"},
	{0x4FF00767466D0165ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_exit.ani"},
	{0x4FF0627F5B360F45ULL, "519789088.ogg"},
	{0x4FF07A2B184A3B31ULL, "gen00fl_fenris6126714.cl"},
	{0x4FF07D9D98DB9AFFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v3.gad"},
	{0x4FF09ED2CA4927BDULL, "lgt360_mag_orlanna6155309.cl"},
	{0x4FF0DC5D8B165447ULL, "172406087.ogg"},
	{0x4FF0FF6FC7BCABBBULL, "qun110_petrice6036856.cl"},
	{0x4FF17F5CF4742F36ULL, "hlac_lt_0_lowtown_a2_65534_0x.dds"},
	{0x4FF1B3BB691DA3FDULL, "6073933_m.fxe"},
	{0x4FF1F10E49C2708AULL, "llac_ht_0_hightown_2_1.dds"},
	{0x4FF2075865FE7595ULL, "vxdr_0_deeproads_ciw_0_0.vlm"},
	{0x4FF21E4A99198043ULL, "126917162.ogg"},
	{0x4FF26357CA9A06C2ULL, "vxdr_primevaltomb_2vz_0_0.vlm"},
	{0x4FF27D3982A80D60ULL, "vxlt_0_lowtown_n_ewb_0_0.vlm"},
	{0x4FF28CEE02B6B197ULL, "6033122_m.fxe"},
	{0x4FF2F23ADFB1B2E1ULL, "vxht_brothel_n_6gw_0_0.vlm"},
	{0x4FF35A4B043A3919ULL, "6077318_f.fxe"},
	{0x4FF3EA9295C371E1ULL, "6108302_f.fxe"},
	{0x4FF415BB48A0BF4EULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rdy.ani"},
	{0x4FF41F08FC3B55B5ULL, "vxlt_0_lowtown_n_ghh_0_0.vlm"},
	{0x4FF42A5297DD086BULL, "hlac_lt_0_lowtown_docks_n_65535_4z.dds"},
	{0x4FF452A235D24975ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanr.ani"},
	{0x4FF48DC44C99A579ULL, "hlac_lt_0_lowtown_docks_a3_65535_1y.dds"},
	{0x4FF5640454916DD7ULL, "6118423_m.fxe"},
	{0x4FF573693E80F8FBULL, "vxht_0_hightown_bm7_0_0.vlm"},
	{0x4FF59C7807B1D52AULL, "vxlt_gallowscourtyard_78m_0_0.vlm"},
	{0x4FF5CDCEBEB614B8ULL, "llac_ca_caverns1_24424_5.dds"},
	{0x4FF5D36E899B93BFULL, "925763586.ogg"},
	{0x4FF5EDF0C6A6A27AULL, "blood_wave.mao"},
	{0x4FF5FE5094C34172ULL, "vxca_caverns1_mines_hkw_0_0.vlm"},
	{0x4FF6001C528E6D10ULL, "lt_0_lowtown_docks_a3_176_0.mao"},
	{0x4FF6041C528E7423ULL, "lt_0_lowtown_docks_a3_176_0.mmh"},
	{0x4FF6121C528E8C7DULL, "lt_0_lowtown_docks_a3_176_0.msh"},
	{0x4FF6151282D19D34ULL, "60338477.ogg"},
	{0x4FF62F22CC251BF0ULL, "vxlt_0_lowtown_a2_n_byw_0_0.vlm"},
	{0x4FF6412FBB6171D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_beseech.ani"},
	{0x4FF6729D98DFF3ECULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v3.evt"},
	{0x4FF6BBF17DC33032ULL, "6085456_m.fxe"},
	{0x4FF7CB368246716DULL, "w_bonepitdestroyed_a3_49_0.msh"},
	{0x4FF7D936824688A0ULL, "w_bonepitdestroyed_a3_49_0.mao"},
	{0x4FF7DD3682468FF3ULL, "w_bonepitdestroyed_a3_49_0.mmh"},
	{0x4FF7F5108DB7DDF9ULL, "hlac_lt_0_lowtown_1_7x.dds"},
	{0x4FF80265E7A27AADULL, "6163117_m.fxe"},
	{0x4FF8108A1379C4DCULL, "6173874_m.fxe"},
	{0x4FF85261BFC69E3FULL, "vxht_keep_5gf_0_0.vlm"},
	{0x4FF86C8A961D4E76ULL, "zz_ave_debug6154140.cl"},
	{0x4FF882532DE4DA4FULL, "395127914.ogg"},
	{0x4FF93332CCBA3496ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_action.ani"},
	{0x4FF94F05AE4DE856ULL, "602623168.ogg"},
	{0x4FF956CA745B09F4ULL, "and100_anders6096007.cl"},
	{0x4FF98176C506E13FULL, "451511817.ogg"},
	{0x4FFA36DE1DB36E10ULL, "vxht_playershouse_n_1vt_0_0.vlm"},
	{0x4FFA6CD0BBB72E58ULL, "722416093.ogg"},
	{0x4FFAA9F53799C451ULL, "character_character_p66.psh.fxo.dx11"},
	{0x4FFAFFF2ACBF8699ULL, "vxlt_0_lowtown_a3_n_c_gc2_0_0.vlm"},
	{0x4FFB21389353135CULL, "6072320_m.fxe"},
	{0x4FFC7AF0A1065EC8ULL, "vxht_brothel_n_6gr_0_0.vlm"},
	{0x4FFCC3D586892105ULL, "vxw_bonepitmining_a2_5h_0_0.vlm"},
	{0x4FFD038DDF3E534CULL, "vxw_bonepitmining_a2_26l_0_0.vlm"},
	{0x4FFD41341FEB38F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_reaction.ani"},
	{0x4FFD501689FDC819ULL, "m_dae220ar_hightown_nt.wbd"},
	{0x4FFD6E91F643A3CEULL, "6063867_m.fxe"},
	{0x4FFDB40530CB8549ULL, "lowtown_fem_cityguard_mob6164836.cl"},
	{0x4FFDE50D352B23D6ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.mmh"},
	{0x4FFE344B9553456DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_07bonetrack.msh"},
	{0x4FFE424B95535CA0ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_07bonetrack.mao"},
	{0x4FFE464B955363F3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_07bonetrack.mmh"},
	{0x4FFE5AB591438D4CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basick.phy"},
	{0x4FFFB9B0E4352BE6ULL, "6024128_f.fxe"},
	{0x4FFFD23677F23CA0ULL, "hlac_lt_gallowstemplar_n_2383_4y.dds"},
	{0x4FFFE7DFEF76B2E9ULL, "dae000_hate_follower_react6167474.cl"},
	{0x500055F86ECA95EEULL, "vxlt_0_lowtown_a2_n_c_h20_0_0.vlm"},
	{0x5000731500B24492ULL, "hlac_lt_0_lowtown_docks_n_-3_1z.dds"},
	{0x5000793557F48E15ULL, "vxdr_0_deeproads_cfv_0_0.vlm"},
	{0x5000AEC68091D8B9ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_aoe_p.anb"},
	{0x5000EF16749E25E0ULL, "vxlt_0_lowtown_n_epc_0_0.vlm"},
	{0x50011035A2DA0812ULL, "skintest.tnt"},
	{0x500137921DDBE4DDULL, "qcr220_meredith6054740.cl"},
	{0x5001C6C2EBA99296ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdsbl.msh"},
	{0x5001D0C2EBA9A394ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdsbl.mmh"},
	{0x5001D4C2EBA9AAA7ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdsbl.mao"},
	{0x50020DBB48ACEE9FULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rdy.evt"},
	{0x50021D696CD2AAFDULL, "485295163.ogg"},
	{0x50022F050C745B2CULL, "6109325_m.fxe"},
	{0x50028D4C375CFDDEULL, "zz_rogue2.cut"},
	{0x50034142818CE37AULL, "hlac_ca_caverns2_5802_43x.dds"},
	{0x5003492FBB6C09C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_beseech.evt"},
	{0x50038679FC126111ULL, "411190074.ogg"},
	{0x5004214D3FDCFA5BULL, "6075256_m.fxe"},
	{0x50045A9D98EC0845ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v3.ani"},
	{0x5004ADCDD64043D4ULL, "621904118.ogg"},
	{0x50052D9A9D7BEA8BULL, "llac_lt_gallowscourtyard_-131074_0.dds"},
	{0x5006370E41F6E77FULL, "135717551.ogg"},
	{0x5006A1EC41417EE5ULL, "vxdr_0_deeproads_cw9_0_0.vlm"},
	{0x5006C1AC0AD40206ULL, "717445129.ogg"},
	{0x5006C43EB1BCB6BFULL, "vxht_keep_5ja_0_0.vlm"},
	{0x50073B32CCC67F87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_action.evt"},
	{0x50073CC13300F8ADULL, "mag170_thrask6016036.cl"},
	{0x50074FCE56F8482CULL, "404931267.ogg"},
	{0x500778DB0C5DD6C0ULL, "vxdr_0_deeproads_fi7_0_0.vlm"},
	{0x5007FA40B5D31089ULL, "454216616.ogg"},
	{0x50086149ED1F4BCEULL, "419006590.ogg"},
	{0x5008B812FF6C4B5EULL, "vxht_0_hightown_n_b7g_0_0.vlm"},
	{0x5008EEBB48B2D888ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rdy.gad"},
	{0x5008F66E45A2864DULL, "art\\levels\\races\\elf\\props\\_textures\\prp_tiki_n.dds"},
	{0x500971CF9EB59F19ULL, "vxlt_0_lowtown_a3_n_c_gf1_0_0.vlm"},
	{0x5009ED6AB55DAEF0ULL, "48741286.ogg"},
	{0x5009F1C726486C21ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_noba_0s.dds"},
	{0x500A49341FF5D0E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_reaction.evt"},
	{0x500A5A8DB74F26A2ULL, "zz_vrc_vase_trap_test.ncs"},
	{0x500A6A2FBB726010ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_beseech.gad"},
	{0x500AAA197DDEFC6CULL, "w_woundedcoast_84_0.rml"},
	{0x500B1C4D2418E38BULL, "6068311_m.fxe"},
	{0x500B25135DFE82A6ULL, "vxht_mansion_n_6nx_0_0.vlm"},
	{0x500BDDB1E4532DE4ULL, "6109220_m.fxe"},
	{0x500C47C7F38E642DULL, "vxlt_0_lowtown_a2_hes_0_0.vlm"},
	{0x500C86E3C03D7FF7ULL, "520303879.ogg"},
	{0x500CEEFEA89BC858ULL, "mcr370_meredith6083077.cl"},
	{0x500D20C8B0067BD6ULL, "vxht_chantry_n_35t_0_0.vlm"},
	{0x500D64FBB81174A7ULL, "630494302.ogg"},
	{0x500D766702869C32ULL, "221736638.ogg"},
	{0x500DB0D987D20D58ULL, "vxdr_primevaltomb_7re_0_0.vlm"},
	{0x500DDF47598EB30EULL, "414434882.ogg"},
	{0x500E5C32CCCCD6D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_action.gad"},
	{0x500EDD1AC64B9898ULL, "vxw_encounter3_f9_0_0.vlm"},
	{0x500FA463C2B5C9A5ULL, "ss_ambient_male.wbd"},
	{0x50101AB4982735C7ULL, "one350_terath6106360.cl"},
	{0x5010200378237C29ULL, "lla_ht_0_hightown_a2_-131074_0.dds"},
	{0x50103193BFFCAC8FULL, "hlac_lt_gallowsprison_n_-1_6z.dds"},
	{0x501046C7B72E9797ULL, "971614926.ogg"},
	{0x5010B3EC9EF221AEULL, "gen00fl_sebastian6188598.cl"},
	{0x501102C2E02786B5ULL, "vxlt_0_lowtown_n_gbv_0_0.vlm"},
	{0x501153D1489149C2ULL, "vxlt_0_lowtown_a2_n_c_f5g_0_0.vlm"},
	{0x50116A341FFC2730ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_reaction.gad"},
	{0x5011E79768C06B40ULL, "6206869_f.fxe"},
	{0x50128A1487BC07C8ULL, "llac_ht_keep_n_4519_5.dds"},
	{0x50128AFC4E36920EULL, "6054021_m.fxe"},
	{0x50128D837E0FF298ULL, "6141145_m.fxe"},
	{0x50129D8B4C5C7662ULL, "plt_cod_aow_traps.plo"},
	{0x5012B182E74C00DEULL, "vxht_mansion_n_66x_0_0.vlm"},
	{0x5012F26C55710D0BULL, "vxlt_warehouse_10m_0_0.vlm"},
	{0x501321759AEE6FF4ULL, "6080606_m.fxe"},
	{0x501331B62C00C1A6ULL, "mag101_grace6034029.cl"},
	{0x50138544A26C8EDBULL, "6000708_m.fxe"},
	{0x5013AC85EE837A1CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityloop.ani"},
	{0x5013BC88EAB83B43ULL, "6208445_m.fxe"},
	{0x5014082911578FEEULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_open_n.dds"},
	{0x501410CF51CB4C55ULL, "210596905.ogg"},
	{0x5014114312D7ABF8ULL, "354466831.ogg"},
	{0x5014217796EFF5A8ULL, "hlac_lt_0_lowtown_65537_4y.dds"},
	{0x50143B114CE4DC1CULL, "dae000ip_crowd_lm4_talk4.utp"},
	{0x501456EAA9124DA3ULL, "ht_bartrand_n_3371.anb"},
	{0x5015027265AE8304ULL, "character_character_p173.psh.fxo.dx11.assembly"},
	{0x501561D505F1DB56ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxa.phy"},
	{0x5015C480185BC052ULL, "fen310_fenris6118689.cl"},
	{0x501600F2A9B34743ULL, "6013340_m.fxe"},
	{0x501670A5613A2835ULL, "162020796.ogg"},
	{0x50176A1657EE8602ULL, "vxca_cave_17r_0_1.vlm"},
	{0x5017D86BAC98AEE0ULL, "6118027_f.fxe"},
	{0x5017E78056D72914ULL, "6018419_f.fxe"},
	{0x501870E83F643209ULL, "757557557.ogg"},
	{0x5018A319E8B046A2ULL, "vxw_encounter2_camp_j4_0_0.vlm"},
	{0x5018EF7A2ECD910EULL, "vxlt_undercityserialkiller_13v_0_0.vlm"},
	{0x5018F54617B1B39CULL, "hlac_lt_blackemporium_1055_3y.dds"},
	{0x50194945C378EFFBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforward.ani"},
	{0x5019E747C32A844CULL, "vxw_bonepitmining_a2_28b_0_0.vlm"},
	{0x501A4CD434BF5E6EULL, "vxlt_0_lowtown_a2_n_c_ewz_0_0.vlm"},
	{0x501A8C2B6A129775ULL, "ser211ip_mothers_ring.utp"},
	{0x501AB985EE89AF5EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityloop.gad"},
	{0x501AB9A4E5847CADULL, "vxlt_0_lowtown_a2_hbr_0_0.vlm"},
	{0x501AE9A3CDD6FD01ULL, "93319776.ogg"},
	{0x501B193976F7E8B5ULL, "llac_lt_0_lowtown_docks_n_-65539_1.dds"},
	{0x501B33D1BAD96785ULL, "dae221st_dog_player_plus.stg"},
	{0x501B4E69D1C58597ULL, "isa111_isabela6175105.cl"},
	{0x501B51244A92A900ULL, "t1_skn_d04.tnt"},
	{0x501B6122F9C37801ULL, "gen00fl_varric6033676.cl"},
	{0x501BC01342866C4EULL, "qcr220_orsino6041851.cl"},
	{0x501BC213E751DBA3ULL, "474352277.ogg"},
	{0x501C4476E1119267ULL, "vxlt_backalley_n_1vp_0_1.vlm"},
	{0x501C7881046471FCULL, "597128817.ogg"},
	{0x501CC3D7D8112132ULL, "vxht_0_hightown_ask_0_0.vlm"},
	{0x501D3EBAEE695F6DULL, "vxlt_0_lowtown_a2_n_c_h1z_0_0.vlm"},
	{0x501D4EF7B841B118ULL, "vxw_encounter3_c6_0_0.vlm"},
	{0x501D9727DFED2757ULL, "401336855.ogg"},
	{0x501DC49E005FF99BULL, "6133802_m.fxe"},
	{0x501DD2D0588A56E0ULL, "vxlt_0_lowtown_n_efa_0_0.vlm"},
	{0x501E337A9ED2A1F0ULL, "6060409_f.fxe"},
	{0x501E361FCC60C4FAULL, "llac_ht_mansion2_818_68.dds"},
	{0x501EA15AC4871645ULL, "hlac_ht_mansion2_818_33y.dds"},
	{0x501EA798C5759EF3ULL, "6072701_m.fxe"},
	{0x501EAD3D8DA3CA04ULL, "6097619_m.fxe"},
	{0x501EFF06626438AEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a.evt"},
	{0x501F01FB7D748DB6ULL, "mag100_thrask6051778.cl"},
	{0x501F080D77CE2DCFULL, "follower_banter6143198.cl"},
	{0x501F242E8115BCDBULL, "vxdr_primevalentrance_274_0_0.vlm"},
	{0x501F30A50B9B7183ULL, "6117615_m.fxe"},
	{0x501F94C04931F5B4ULL, "zz_and_debug6178613.cl"},
	{0x501FC75D6C1B505DULL, "one111_thug6144336.cl"},
	{0x501FF6AB54C44156ULL, "qun220cr_javaris6071859.cl"},
	{0x50204AD505FBBBF7ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxa.msh"},
	{0x50205CD505FBDA5AULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxa.mao"},
	{0x502060D505FBE169ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxa.mmh"},
	{0x5020701FBF0F8AAEULL, "art\\vfx\\cinematic\\vfx_flash_p.impa.ani"},
	{0x50207E0269B8A5AAULL, "6072151_m.fxe"},
	{0x502081B22FB94395ULL, "mag160cr_slaver_mage.utc"},
	{0x5020A7F1AEC70756ULL, "395770302.ogg"},
	{0x5020B8D979A006DFULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.mmh"},
	{0x50211D85BE0C79D7ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyn_0s.dds"},
	{0x50219485EE8F8F7DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityloop.evt"},
	{0x5021DF60ED39BCD3ULL, "follower_banter6166403.cl"},
	{0x5021F7057C374CE6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2.phy"},
	{0x50224371C7C9DDD2ULL, "mag210_arianni6057344.cl"},
	{0x502245AE63F684C2ULL, "gen00fl_anders6125363.cl"},
	{0x502254ED3C40BFF4ULL, "6020082_m.fxe"},
	{0x50226640C93903BDULL, "zz_mcr_debug6098528.cl"},
	{0x5022B481D5F002D1ULL, "vxht_mansion2b_1t9_0_0.vlm"},
	{0x5022BAA55E2C9628ULL, "hlac_w_blightlands_0_3z.dds"},
	{0x5023CE5372A5DA5EULL, "6135811_f.fxe"},
	{0x5023F39FBA9DF889ULL, "943044300.ogg"},
	{0x5023F6CF292F583FULL, "gen00fl_anders6111190.cl"},
	{0x502439F937BDECAAULL, "6129242_m.fxe"},
	{0x50244722A57AA984ULL, "hlac_ca_cavern1_haunted_24424_24x.dds"},
	{0x502468AEFAC2C50EULL, "vxca_cavern1_haunted_icx_0_2.vlm"},
	{0x502488725F334BA0ULL, "18691504.ogg"},
	{0x50249FCA73B1FE40ULL, "vxht_0_hightown_n_7vs_0_0.vlm"},
	{0x5024A15F3B555AC5ULL, "dr_primevalentrance_1724_lc.mmh"},
	{0x5024AF060A237BD5ULL, "zz_qun_debug6028899.cl"},
	{0x5025F606626A4859ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a.gad"},
	{0x50269DD70AA9AACAULL, "ave200_donnic6112167.cl"},
	{0x502714F6DAA65F22ULL, "vxw_encounter2_camp_e5_0_0.vlm"},
	{0x50271DAC422E0ED6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v2.ani"},
	{0x5027375523AD93E1ULL, "and100_anders6095795.cl"},
	{0x50275145C3853AF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforward.evt"},
	{0x502757BA0986F733ULL, "495247451.ogg"},
	{0x50276142FB7F0A9DULL, "vxlt_warehouse_cc_0_0.vlm"},
	{0x502797FA3A08A2CBULL, "vxw_bonepitdestroyed_a3_1r6_0_0.vlm"},
	{0x502808DAF93784C2ULL, "615163565.ogg"},
	{0x50286D288452580EULL, "gen00fl_dog6167236.cl"},
	{0x50287D118F678B23ULL, "vxht_0_hightown_n_c6h_0_0.vlm"},
	{0x50288F3E0754F3CBULL, "dre000_hireling26038679.cl"},
	{0x5028A337E16936A6ULL, "272218350.ogg"},
	{0x5028C6E38F2C3BBBULL, "dre000_varric6034683.cl"},
	{0x5028CCFB85120898ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_0z.dds"},
	{0x5029611589194E69ULL, "vxht_playershouse_n_1qt_0_0.vlm"},
	{0x5029C3B5AE8778B6ULL, "ht_0_hightown_52_0_lc.mmh"},
	{0x5029C496A5EDC7B7ULL, "6164635_m.fxe"},
	{0x502ABBF0CB1AC57EULL, "vxlt_hangedman_2yx_0_0.vlm"},
	{0x502ACB32FC19DAD8ULL, "50956016.ogg"},
	{0x502ACBE64B303217ULL, "vxca_caverns2_7f9_0_1.vlm"},
	{0x502AEE13F9EC92BBULL, "hlac_ht_mansion2_818_65x.dds"},
	{0x502B662599B169A7ULL, "vxlt_warehouse_n_12d_0_0.vlm"},
	{0x502BC7801CDC737EULL, "6112020_m.fxe"},
	{0x502C5EBC6C60933BULL, "288506813.ogg"},
	{0x502C7E49FCD84D99ULL, "596679072.ogg"},
	{0x502CE15C023E1346ULL, "llac_lt_0_lowtown_a3_n_c_-196608_0.dds"},
	{0x502D0706627083A7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a.ani"},
	{0x502D255E30C3D4ACULL, "vxca_caverns2_9fy_0_0.vlm"},
	{0x502D424DA87E86C0ULL, "hf_refugee_girl.mor"},
	{0x502D7045C389DA85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforward.gad"},
	{0x502DA13673B70790ULL, "hlac_lt_0_lowtown_docks_n_-65537_4x.dds"},
	{0x502DA22CC598E757ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl04.mmh"},
	{0x502DB82CC5990CC9ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl04.msh"},
	{0x502DE3CAD7E0B205ULL, "6164062_m.fxe"},
	{0x502E3F43ED5C5710ULL, "ser120_ghyslain6164761.cl"},
	{0x502E681FBF1BBAFFULL, "art\\vfx\\cinematic\\vfx_flash_p.impa.evt"},
	{0x502EDC977C0D47CAULL, "63818677.ogg"},
	{0x502EE5C8145B6691ULL, "6161151_m.fxe"},
	{0x502EFCC4FB8C5C04ULL, "lowtown_male_beggar_sta6151045.cl"},
	{0x502F2CAA024F4CA7ULL, "cd_fen210_hadriana.wbd"},
	{0x502FAC7E9D3823FFULL, "ave150tr_start_ambush.ncs"},
	{0x502FB468FB46FBDFULL, "6162646_m.fxe"},
	{0x50302994C1BCEF6CULL, "565260468.ogg"},
	{0x5030B032E0CACB09ULL, "6025163_m.fxe"},
	{0x5030B7840C96A992ULL, "441514056.ogg"},
	{0x5030C491317D9491ULL, "118479898.ogg"},
	{0x5030D17B5888689FULL, "vxlt_undercityencounter_uj_0_0.vlm"},
	{0x5031DF427FB47AA7ULL, "vxht_0_hightown_a2_9cn_0_0.vlm"},
	{0x50331CB093FC3C15ULL, "748354811.ogg"},
	{0x5033A9B3EBFB72DBULL, "6163726_m.fxe"},
	{0x503425AC4238A6C7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v2.evt"},
	{0x5034B1271FDAB8BBULL, "6083460_m.fxe"},
	{0x5034B3BB8B65AEFFULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_avel.mmh"},
	{0x5034B9BB8B65B8A1ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_avel.msh"},
	{0x5034C7667497E899ULL, "vxlt_0_lowtown_a3_n_c_gw6_0_0.vlm"},
	{0x5034D49BD5C5AF86ULL, "seb000im_leliana_boots.uti"},
	{0x5035491FBF21A4E8ULL, "art\\vfx\\cinematic\\vfx_flash_p.impa.gad"},
	{0x5035BE6279BCD00CULL, "109662459.ogg"},
	{0x5035EB1F6C839619ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb13a.msh"},
	{0x5035F51F6C83A707ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb13a.mmh"},
	{0x503605474AB22F93ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_loop.evt"},
	{0x50360E8E40B70FB6ULL, "70162828.ogg"},
	{0x5036A1A97494B548ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_4y.dds"},
	{0x5036E381A5D74723ULL, "gen_im_npc_wep_two_car_fake.uti"},
	{0x5036EEEE815DA3A3ULL, "vxht_0_hightown_n_c3k_0_0.vlm"},
	{0x50377C9A0038D91BULL, "vxlt_gallowstemplar_f_5ru_0_0.vlm"},
	{0x5037E73C6E5388F8ULL, "211330460.ogg"},
	{0x5038D9467C9BCA75ULL, "345047914.ogg"},
	{0x503900CDDB293264ULL, "hlac_ht_chantry_2534_34x.dds"},
	{0x503930223A66C976ULL, "vxw_woundedcoast_1m1_0_0.vlm"},
	{0x5039AA0E0E0CBA3DULL, "mag291_torpor6056895.cl"},
	{0x5039AD07065905D9ULL, "vxlt_undercityserialkiller_1lq_0_1.vlm"},
	{0x5039B0E768591EB2ULL, "vxca_caverns2_9di_0_0.vlm"},
	{0x503A05DF75533718ULL, "dre180cr_spider_boss.ncs"},
	{0x503A322144CED680ULL, "6168984_m.fxe"},
	{0x503A4FC565D64F69ULL, "6131136_m.fxe"},
	{0x503A7D06E6BE2B75ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_0y.dds"},
	{0x503B46AC423EFD10ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v2.gad"},
	{0x503B4D0CF00E142DULL, "vxht_brothel_n_47u_0_0.vlm"},
	{0x503B8D1C2704B7DAULL, "696728293.ogg"},
	{0x503BA1655C5DD2DBULL, "gen00fl_dog6117799.cl"},
	{0x503BD1B8B7C71B92ULL, "6163882_m.fxe"},
	{0x503BF832D60DC034ULL, "hlac_ht_bartrand_n_2995_3y.dds"},
	{0x503BFE474AB68FD4ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_loop.gad"},
	{0x503CCA662F1DE1F9ULL, "401883104.ogg"},
	{0x503D4664877AD817ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_loop_v3.gad"},
	{0x503D5D4FF561DFBCULL, "vxlt_0_lowtown_a3_n_c_gck_0_0.vlm"},
	{0x503D87C1554A8BBBULL, "6091885_m.fxe"},
	{0x503D9D160FCA8070ULL, "358680815.ogg"},
	{0x503E2956FCF3E383ULL, "vxw_sundermount_34m_0_0.vlm"},
	{0x503E31B591792B17ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basick.mmh"},
	{0x503E35B591793224ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basick.mao"},
	{0x503E47B591795089ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basick.msh"},
	{0x503E79AF6EA672D6ULL, "mag100_cullen6140919.cl"},
	{0x503E9FA395089C7DULL, "vxlt_gallowstemplar_f_3on_0_0.vlm"},
	{0x503F3AA55FB836B0ULL, "644870575.ogg"},
	{0x503F47B6D974391FULL, "6153343_m.fxe"},
	{0x503F8EDE7DCEBA46ULL, "hlac_ht_mansion_n_3371_2x.dds"},
	{0x5040456483AC785BULL, "gen00fl_anders6096633.cl"},
	{0x5040537DD67682C6ULL, "mag320_orsino6071085.cl"},
	{0x50406A3B720C8BE4ULL, "gen00fl_anders6125633.cl"},
	{0x50409336383A8796ULL, "229349139.ogg"},
	{0x5040C797E6ECCF3EULL, "74246593.ogg"},
	{0x5041658E86E6CCB6ULL, "1058510514.ogg"},
	{0x50416D91446398FEULL, "vxht_0_hightown_a2_bst_0_0.vlm"},
	{0x504170078F69E372ULL, "vxw_sundermount_316_0_1.vlm"},
	{0x5041B3D648DF87C1ULL, "vxlt_0_lowtown_docks_b8w_0_0.vlm"},
	{0x504229254C7E6A2AULL, "vxht_mansion2_2o8_0_0.vlm"},
	{0x504232A93CD890C6ULL, "vxlt_gallowscourtyard_7bo_0_0.vlm"},
	{0x50426672F7B94AB2ULL, "6153759_m.fxe"},
	{0x50430D474ABCC7B2ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_loop.ani"},
	{0x50432ADC5E6BD506ULL, "vxca_caverns1_mines_ead_0_0.vlm"},
	{0x5043422BFDA4DE22ULL, "m_dae110ar_lowtown.bnk"},
	{0x504368F5FB98ADAAULL, "lt_0_lowtown_a2_n_c_32_0_ll.mmh"},
	{0x504460DD7A0BDE12ULL, "1003439866.ogg"},
	{0x5044B91E6193A731ULL, "6068857_f.fxe"},
	{0x5044F1CAF8E2253CULL, "hm_generic_ghl_a.mor"},
	{0x5045513F900E309AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ext.gad"},
	{0x504556D1186F4F18ULL, "vxlt_gallowscourtyard_5z3_0_0.vlm"},
	{0x5045581FF0110D2BULL, "vxca_caverns2_8mn_0_0.vlm"},
	{0x504560CB7353BC23ULL, "vxht_0_hightown_n_c8j_0_0.vlm"},
	{0x50456BD30F7FA5AEULL, "ser210_witness6168110.cl"},
	{0x50457AEFCBE57120ULL, "plt_lgt000pt_lite_key_01.nss"},
	{0x5045E3A097227922ULL, "vxlt_0_lowtown_docks_cop_0_0.vlm"},
	{0x5045EE76F22EF19BULL, "vxlt_gallowstemplar_f_5ut_0_0.vlm"},
	{0x5046501BCAA6A4E8ULL, "hlac_w_blightlands_1_0x.dds"},
	{0x50465A204ADDD6E5ULL, "6020436_m.fxe"},
	{0x50469F0C2BDF88E1ULL, "qun110_petrice6042202.cl"},
	{0x5046A56C59555FA6ULL, "6103219_m.fxe"},
	{0x5046AC42B7626A87ULL, "6143261_m.fxe"},
	{0x5046DCA47BECA0B3ULL, "zz_vault_debug6203210.cl"},
	{0x504716F9F55B6A25ULL, "vxw_sundermount_228_0_0.vlm"},
	{0x504723D6AF91B9F4ULL, "6046936_m.fxe"},
	{0x5047A1FF2C5CE1F6ULL, "vxw_woundedcoast_1h2_0_0.vlm"},
	{0x5047C00C6630A156ULL, "vxca_cavern1_haunted_j2j_0_0.vlm"},
	{0x504876A8462C9EB8ULL, "6897820.ogg"},
	{0x5048B9D979C15A1CULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.anb"},
	{0x5048EFF954DC0696ULL, "hlac_dr_dungeon_3214_6z.dds"},
	{0x5048F5550A2AE633ULL, "6082254_m.fxe"},
	{0x50498295E1DA6089ULL, "vxw_bonepit_26w_0_0.vlm"},
	{0x504A03E41B3CD9A8ULL, "6093528_f.fxe"},
	{0x504A2B3B8FFC701DULL, "lt_0_lowtown_docks_n_111_0.mao"},
	{0x504A2F3B8FFC7756ULL, "lt_0_lowtown_docks_n_111_0.mmh"},
	{0x504A352975328434ULL, "gen00fl_varric6099459.cl"},
	{0x504A393B8FFC8854ULL, "lt_0_lowtown_docks_n_111_0.msh"},
	{0x504ABC03A41E1D0BULL, "vis121_saemus6113524.cl"},
	{0x504B09343D536F6DULL, "dae000ip_crowd_lm3_talk15.utp"},
	{0x504B1062D1411C2BULL, "130759398.ogg"},
	{0x504B692A48E11913ULL, "970376087.ogg"},
	{0x504BE77C1C32E227ULL, "zz_followers_banter6159102.cl"},
	{0x504C292A7CEA6C57ULL, "vxlt_0_lowtown_docks_n_cs6_0_0.vlm"},
	{0x504C543F901454B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ext.evt"},
	{0x504C5EF6963ED030ULL, "vxlt_darktown_5ja_0_0.vlm"},
	{0x504C9B33EEE9FC03ULL, "vxw_sundermount_33l_0_0.vlm"},
	{0x504CD33F3568B768ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudempty.msh"},
	{0x504CE53F3568D5F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudempty.mao"},
	{0x504CF13F3568E992ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudempty.mmh"},
	{0x504DBF4FF5F9AFE0ULL, "dae000_katriela6162658.cl"},
	{0x504DC84F2013B22FULL, "vxw_bonepitdestroyed_a3_kn_0_0.vlm"},
	{0x504E1BEBD861E1CDULL, "541089420.ogg"},
	{0x504EE48A32A270E4ULL, "w_encounter2_camp_44_0.msh"},
	{0x504EF68A32A28F4DULL, "w_encounter2_camp_44_0.mao"},
	{0x504EFA8A32A29586ULL, "w_encounter2_camp_44_0.mmh"},
	{0x504F755C60EB5838ULL, "6125593_f.fxe"},
	{0x504F76190E38ABDFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1c.msh"},
	{0x504F7C190E38B581ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1c.mmh"},
	{0x504F88190E38CA22ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1c.mao"},
	{0x504F96E23321B41AULL, "vxlt_warehouse_n_x8_0_0.vlm"},
	{0x50503084B27E5284ULL, "dog_glow.mao"},
	{0x50507E63C2ED204FULL, "ss_ambient_male.bnk"},
	{0x5050C1E1EAB19964ULL, "da2terrain_terrain_p0.dsh.fxo.dx11.assembly"},
	{0x5050E791246B5F4EULL, "287625804.ogg"},
	{0x5050E8039A634F99ULL, "430931061.ogg"},
	{0x50513B59A8AAC45EULL, "dae000_arishok6070876.cl"},
	{0x50517364878BCD9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_loop_v3.ani"},
	{0x5051A2439AC19BCAULL, "mcr370_negotiation6080268.cl"},
	{0x505239E69BB22834ULL, "ser221_gamlen6131878.cl"},
	{0x50527A4116E0AB5AULL, "202800550.ogg"},
	{0x5053007E04C3DB3EULL, "dae110ip_at_lowtown.utp"},
	{0x5053A063C2EF8FB7ULL, "ss_ambient_male.cnv"},
	{0x5053AB88BFBE7C30ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_mere.mmh"},
	{0x5053B988BFBE944AULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_mere.msh"},
	{0x5053C0A735AB592BULL, "vxht_0_hightown_bge_0_0.vlm"},
	{0x505433E64D2AFFAFULL, "6144559_f.fxe"},
	{0x5054449694550B1AULL, "vxlt_darktown_4t0_0_0.vlm"},
	{0x5054A452FA309C60ULL, "mag121_idunna6084896.cl"},
	{0x5054E54918BA6EFCULL, "vxht_chantry_n_338_0_0.vlm"},
	{0x5054F2BC6F6A6197ULL, "6043701_m.fxe"},
	{0x505513FC1A4A12E7ULL, "6167115_m.fxe"},
	{0x505595B3A187D274ULL, "6041836_m.fxe"},
	{0x5055A18767EA6292ULL, "14957570.ogg"},
	{0x5055ACC81467F9ACULL, "qun131_taskmaster6019168.cl"},
	{0x5055F375D99CD588ULL, "29822041.ogg"},
	{0x505613DC1E52FA76ULL, "vxw_woundedcoast_1s3_0_0.vlm"},
	{0x50564252BDE86D59ULL, "follower_banter6155829.cl"},
	{0x505665FFC0E7CCAAULL, "vxw_bonepitdestroyed_a3_17l_0_0.vlm"},
	{0x5056A7572B0F6DEDULL, "6031993_m.fxe"},
	{0x5056FEA1BE187C26ULL, "6119630_m.fxe"},
	{0x50572455A3F4D219ULL, "6060490_m.fxe"},
	{0x505727A10D5B31FDULL, "382673484.ogg"},
	{0x50576E7A36B8F2F8ULL, "qun160_vashothlead6053919.cl"},
	{0x5057EE0810DB838FULL, "logininfosharing.gfx"},
	{0x50583CA045A8E6ABULL, "6110247_m.fxe"},
	{0x50584728CF3E9BB3ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga106a_0s.dds"},
	{0x5058A0CDB978DA46ULL, "vxdr_primevaltomb_86d_0_0.vlm"},
	{0x5058BF99B343597AULL, "hla_ht_0_hightown_a2_0_0z.dds"},
	{0x5058ED4BF2110AD2ULL, "vxht_mansion2b_2vs_0_0.vlm"},
	{0x5058F18A32AADB51ULL, "w_encounter2_camp_44_0.phy"},
	{0x505993C16873F704ULL, "7005400.ogg"},
	{0x505A333FDC0ECD6FULL, "hlac_lt_0_lowtown_docks_0_1x.dds"},
	{0x505A6C3F9020BA80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_ext.ani"},
	{0x505AD0D38834E8B0ULL, "vxlt_darktown_5on_0_0.vlm"},
	{0x505B0CB0DDBB1E36ULL, "6119017_f.fxe"},
	{0x505B16F13488832EULL, "w_bonepit_226_0.mmh"},
	{0x505B22F1348897C5ULL, "w_bonepit_226_0.mao"},
	{0x505B23BD8F69F713ULL, "280141557.ogg"},
	{0x505B30F13488AF1CULL, "w_bonepit_226_0.msh"},
	{0x505B49C51369D006ULL, "vxht_0_hightown_a2_bqf_0_0.vlm"},
	{0x505B7ABBAA1B27B2ULL, "6147888_m.fxe"},
	{0x505BD1708C7E4989ULL, "mag110_vincento6016309.cl"},
	{0x505DD8595B9DB648ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_2.pwk"},
	{0x505E514B284FC9FEULL, "vxht_0_hightown_a2_byv_0_0.vlm"},
	{0x505F166320C4C1C6ULL, "vxlt_gallowscourtyard_7hq_0_0.vlm"},
	{0x505F297E257CE816ULL, "w_bonepit_232_0.mao"},
	{0x505F357E257CFCB5ULL, "w_bonepit_232_0.mmh"},
	{0x505F377E257D002BULL, "w_bonepit_232_0.msh"},
	{0x505FC6D3F1A0A0B1ULL, "59580648.ogg"},
	{0x505FFA8A32B1094DULL, "w_encounter2_camp_44_0.rml"},
	{0x5060323EE766FD30ULL, "6034655_f.fxe"},
	{0x506091BF0AC29E03ULL, "284206694.ogg"},
	{0x5060B46721A376D5ULL, "sst_ss_pride_demon.gda"},
	{0x506181F11629EC10ULL, "zz_and_debug6134176.cl"},
	{0x5061DE1F6CA8BB3CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb13a.phy"},
	{0x506207352B71DFC5ULL, "zz_lgt_2_debug6211868.cl"},
	{0x5062C75A7B0EAA22ULL, "vxlt_0_lowtown_docks_cur_0_0.vlm"},
	{0x5062D230D61B229BULL, "vxlt_gallowstemplar_f_5cv_0_0.vlm"},
	{0x50630982E9E7BEB9ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_7z.dds"},
	{0x506355E50E4B932BULL, "6149775_m.fxe"},
	{0x50636A56AC2EB6CAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9rr_stair.mao"},
	{0x50636E56AC2EBD19ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9rr_stair.mmh"},
	{0x50637856AC2ECE07ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9rr_stair.msh"},
	{0x506397C6CB5B28C2ULL, "6183963_m.fxe"},
	{0x5063A6BD5238EEA9ULL, "6039243_m.fxe"},
	{0x5063D0F6DFCD1E2FULL, "follower_banter6164456.cl"},
	{0x5063DA6C03293BB8ULL, "6180498_m.fxe"},
	{0x5064D7DCB2DDE52AULL, "vxw_bonepitdestroyed_a3_18m_0_0.vlm"},
	{0x50650AD67A835292ULL, "vxlt_warehouse_hu_0_0.vlm"},
	{0x506519B5FE229E16ULL, "vxlt_gallowstemplar_3ni_0_0.vlm"},
	{0x50651CB432C47DD0ULL, "hlac_lt_0_lowtown_a2_-196607_0y.dds"},
	{0x506573030C85416EULL, "vxlt_0_lowtown_a2_n_c_ho7_0_0.vlm"},
	{0x5065F920E2749646ULL, "llac_ht_mansion2b_818_62.dds"},
	{0x50660667871121ABULL, "213877300.ogg"},
	{0x5066C74F465D582BULL, "249885771.ogg"},
	{0x5066F0AFD96CF1CFULL, "6121254_m.fxe"},
	{0x5067587CB44F46B3ULL, "mag100_cullen6140974.cl"},
	{0x5067A98E3FC46B97ULL, "vxht_keep_n_3xz_0_0.vlm"},
	{0x50683E64C0BC4AC0ULL, "162902695.ogg"},
	{0x5068B2E6CB44293CULL, "vxlt_0_lowtown_a3_n_c_gpn_0_0.vlm"},
	{0x50690FE76278E8D8ULL, "dae000cr_grand_cleric6148121.cl"},
	{0x50691372B9AD8D91ULL, "zz_dae_debug6029519.cl"},
	{0x50692C2BFDC58840ULL, "m_dae110ar_lowtown.wbd"},
	{0x50697E0BA4E22609ULL, "755237783.ogg"},
	{0x5069BBA2055FE886ULL, "vxht_0_hightown_a2_bzg_0_0.vlm"},
	{0x5069EC989C114032ULL, "6140988_m.fxe"},
	{0x506AA2BFCF8F0DB1ULL, "dae000_paivel6148956.cl"},
	{0x506ACCD584A7A705ULL, "hlac_lt_backalley_g_65536_0x.dds"},
	{0x506ACFDEFB162346ULL, "vxlt_0_lowtown_docks_baw_0_0.vlm"},
	{0x506ADF5F78BC9EC1ULL, "hlac_ca_caverns2_5802_68x.dds"},
	{0x506AF82F20934B6FULL, "vxlt_gallowstemplar_4yz_0_0.vlm"},
	{0x506B09C715FE2EA4ULL, "155902749.ogg"},
	{0x506B31ED900250A9ULL, "vxlt_0_lowtown_a2_n_c_gj0_0_0.vlm"},
	{0x506B4AF686ACE103ULL, "art\\characters\\creatures\\mubari\\animation\\com_pant_v2.evt"},
	{0x506B6227BC1788FBULL, "6129687_m.fxe"},
	{0x506B69A7354156B7ULL, "art\\characters\\playercharacter\\humanfemale\\boo_urva.phy"},
	{0x506C1ED76DD594FEULL, "hlac_lt_0_lowtown_a3_n_c_131072_4y.dds"},
	{0x506CF56E1D65C457ULL, "mag311_emile_de_launcet6043572.cl"},
	{0x506DAF38710FE9CBULL, "vxdr_primevaltomb_8o6_0_0.vlm"},
	{0x506DDB9FEF0AD305ULL, "6018530_m.fxe"},
	{0x506DDBB49CA113F5ULL, "bakedlowlod.psh.fxo"},
	{0x506DDE6FCAE186EAULL, "751359897.ogg"},
	{0x506E96F3AE5497BFULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.dura.gad"},
	{0x506EB465C2873A5FULL, "vxht_brothel_n_33h_0_0.vlm"},
	{0x506F8D2B6A5AD8C4ULL, "ser211ip_mothers_ring.ncs"},
	{0x506FAAC2E16A33DBULL, "572259377.ogg"},
	{0x506FDA9965186F28ULL, "333117271.ogg"},
	{0x507036AA0286F4DDULL, "cd_fen210_hadriana.bnk"},
	{0x5070D190B5295647ULL, "680679717.ogg"},
	{0x5070FC68C59A65C1ULL, "static_static_p2.vsh.fxo.dx11.assembly"},
	{0x50712565D99D07FCULL, "art\\levels\\races\\design\\_textures\\trap_floorlever_sa.dds"},
	{0x507156E9C6D45480ULL, "w_blightlands_77_0.phy"},
	{0x507223F686B2BD84ULL, "art\\characters\\creatures\\mubari\\animation\\com_pant_v2.gad"},
	{0x5072720E5DF24C2BULL, "898556952.ogg"},
	{0x5072C948FAC15A1EULL, "hlac_lt_hangedman_-1_1x.dds"},
	{0x5073E25BA204AD26ULL, "6113030_m.fxe"},
	{0x50740D1CD0371C71ULL, "99094633.ogg"},
	{0x5074A0414F1CD538ULL, "692431942.ogg"},
	{0x5074F7B32358974DULL, "vxht_0_hightown_4n6_0_0.vlm"},
	{0x507569E65F3A5F93ULL, "ser270_templar6112567.cl"},
	{0x50757DDE0F1D1902ULL, "vxht_bartrand_n_5rh_0_0.vlm"},
	{0x50758BF3AE5AA3ACULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.dura.evt"},
	{0x5075B69277F92F1FULL, "6189371_m.fxe"},
	{0x50763A94EFEEEFDFULL, "hlac_ht_0_hightown_a2_65537_1y.dds"},
	{0x5076FEBC101752CBULL, "vxlt_0_lowtown_gcd_0_0.vlm"},
	{0x50777944AEC0FA99ULL, "6116764_m.fxe"},
	{0x50780B6760552BE1ULL, "vxw_blightlands_bd_0_0.vlm"},
	{0x50788B2973C9E608ULL, "llac_ca_caverns1_24424_16.dds"},
	{0x5078966509A49596ULL, "326489367.ogg"},
	{0x5078C7D23DE9C64DULL, "ser221_gascard6134525.cl"},
	{0x5078E2398B7B5E76ULL, "555173116.ogg"},
	{0x5078F6F6A6FD7AE8ULL, "hlac_ht_keep_4519_6x.dds"},
	{0x507920C27A6B1CFFULL, "vxht_0_hightown_c71_0_1.vlm"},
	{0x507932F686B8F662ULL, "art\\characters\\creatures\\mubari\\animation\\com_pant_v2.ani"},
	{0x507941BBED0C3BC6ULL, "vxlt_0_lowtown_docks_bdx_0_0.vlm"},
	{0x507952E2F914D1ABULL, "vxlt_gallowstemplar_5bz_0_0.vlm"},
	{0x50795645AADB9529ULL, "gen_im_acc_rng_act3_11.uti"},
	{0x507A18933982F527ULL, "vxht_0_hightown_a2_9ze_0_0.vlm"},
	{0x507A4599C9A34D39ULL, "922007971.ogg"},
	{0x507A8C404A41596CULL, "6109798_m.fxe"},
	{0x507A9DA33E15EF6BULL, "6189421_m.fxe"},
	{0x507AA6A4393B6AF0ULL, "vxlt_0_lowtown_a2_gbx_0_0.vlm"},
	{0x507B35050C3BFAFEULL, "vxht_0_hightown_a2_bgh_0_0.vlm"},
	{0x507B7E059AC72922ULL, "vxdr_primevalentrance_2eg_0_0.vlm"},
	{0x507BAD440F3A27C7ULL, "6092471_m.fxe"},
	{0x507C0310D8849340ULL, "116096018.ogg"},
	{0x507C9EDECA562035ULL, "vrc221_bartrand6043276.cl"},
	{0x507CA02D86299EA6ULL, "zz_qun_debug6028234.cl"},
	{0x507CC41B07EB01C1ULL, "isa211_isabela6069399.cl"},
	{0x507CFDB2488E155AULL, "vxht_bartrand_n_645_0_0.vlm"},
	{0x507D353D9545B54CULL, "6075146_f.fxe"},
	{0x507D8C8FC345B228ULL, "6126523_f.fxe"},
	{0x507DA1EF6E02E8FAULL, "596984580.ogg"},
	{0x507E33163C9BAD4FULL, "6126616_f.fxe"},
	{0x507E5902F83F5263ULL, "hlac_ht_mansion2_818_32x.dds"},
	{0x507E7A37C16FCE42ULL, "425640388.ogg"},
	{0x507ED1746575CEE6ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\knockback.ani"},
	{0x507EF1192E10075CULL, "397075289.ogg"},
	{0x507F35F919A8415AULL, "llac_ht_0_hightown_a2_1_0.dds"},
	{0x507F9FD1D26F6D4EULL, "6036564_m.fxe"},
	{0x507FC942825F5428ULL, "600445728.ogg"},
	{0x507FEE015B5A743BULL, "171681326.ogg"},
	{0x508083AB9BB722CFULL, "llac_dr_slavers_3214_5.dds"},
	{0x5080E34C58E9AF93ULL, "vxlt_0_lowtown_docks_a3_n_d0s_0_0.vlm"},
	{0x5080E4E5E250EF72ULL, "gen00fl_anders6112893.cl"},
	{0x5081406A6A8EED7CULL, "6085652_m.fxe"},
	{0x508160A38386431CULL, "ply_clo.mao"},
	{0x50817BA4EA740C29ULL, "vrc321st_spooky_scene_4.stg"},
	{0x5081941D21A9D7C8ULL, "vxw_blightlands_ba_0_0.vlm"},
	{0x5081EE905E6F8392ULL, "vxlt_warehouse_vw_0_0.vlm"},
	{0x50820410D486C46AULL, "hlac_w_encounter3_-327677_0z.dds"},
	{0x50825485BF13A9D2ULL, "dr_0_deeproads_5849.anb"},
	{0x50825BC58C240AEDULL, "vxlt_0_lowtown_a2_n_c_hns_0_0.vlm"},
	{0x508273EFCC1913E3ULL, "plt_lgt000pt_lite_key_01.plo"},
	{0x5082A9ECC6646306ULL, "vxht_0_hightown_n_a78_0_0.vlm"},
	{0x5082BF2BAB80F2ECULL, "369580025.ogg"},
	{0x5082F5ACF7DC4AE4ULL, "6109053_m.fxe"},
	{0x508373F3AE66B805ULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.dura.ani"},
	{0x5083D0994FDE7B96ULL, "1020773895.ogg"},
	{0x5083DFCD8AE825B1ULL, "528869339.ogg"},
	{0x5083EFBB01133182ULL, "vxht_bartrand_n_5ui_0_0.vlm"},
	{0x508495DE39A97264ULL, "6144981_f.fxe"},
	{0x508526E6EE6DDADDULL, "6135712_m.fxe"},
	{0x50864F99CDC070E5ULL, "hlac_ht_bartrand_n_3373_2x.dds"},
	{0x508654A7D1AC6B8FULL, "373880129.ogg"},
	{0x5086D69CE5C883CAULL, "6116631_m.fxe"},
	{0x50873F7874E2ED51ULL, "476717241.ogg"},
	{0x5087C4B120B7D03DULL, "character_character_p231.psh.fxo.dx11"},
	{0x50881E1B976FA650ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_anda.mmh"},
	{0x50882BDEDADDCA05ULL, "6080258_m.fxe"},
	{0x50882C1B976FBE6AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_anda.msh"},
	{0x5088C7256F93E51AULL, "isa311_castillon.cnv"},
	{0x5088CF62068CD194ULL, "vxht_mansion_n_5a8_0_0.vlm"},
	{0x508A1EBC6065596AULL, "6049423_m.fxe"},
	{0x508A96BADAB50792ULL, "lt_gallowscourtyard_n_59_0_ll.mmh"},
	{0x508B3DEBCC11EA94ULL, "725143598.ogg"},
	{0x508BE91CD898BC6FULL, "vxht_0_hightown_a3_ajp_0_0.vlm"},
	{0x508BE9256F965562ULL, "isa311_castillon.bnk"},
	{0x508BFDF783694482ULL, "1001073040.ogg"},
	{0x508C01155FC37C99ULL, "gen_im_trs_krk_hvy_boo_r0.uti"},
	{0x508C3696FF9D1A59ULL, "6165642_f.fxe"},
	{0x508C627B0AF0BC64ULL, "gen00fl_aveline6053315.cl"},
	{0x508C634D20BE4439ULL, "tools_lightbulb.mao"},
	{0x508CB9746581E357ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\knockback.evt"},
	{0x508D68E2FE6B58CDULL, "vxlt_0_lowtown_f4r_0_0.vlm"},
	{0x508DD84960A3C726ULL, "dae000ip_crowd_ff1_talk15.utp"},
	{0x508DE5C96EF5443FULL, "hlac_lt_0_lowtown_-262144_4z.dds"},
	{0x508E2C3CEB1DA7D0ULL, "6070418_m.fxe"},
	{0x508EF5D5C928135AULL, "t1_mub_ro1.tnt"},
	{0x508EFD2E651645DFULL, "6000884_f.fxe"},
	{0x508F2D60B69CA081ULL, "6144572_m.fxe"},
	{0x508F561F6642FF15ULL, "dae000_orana6149029.cl"},
	{0x508F6370B9479032ULL, "hlac_lt_darktown_-1_46x.dds"},
	{0x508FA2C7CFEFE44CULL, "6119315_m.fxe"},
	{0x508FF81F655AD357ULL, "hlac_ht_brothel_n_3374_1x.dds"},
	{0x508FFF54D10A4CB5ULL, "6060443_m.fxe"},
	{0x50900F04A9E184E5ULL, "6163261_m.fxe"},
	{0x50912F74974C15E4ULL, "mag200_feynriel6024077.cl"},
	{0x50915D92E760D78CULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90l.ani"},
	{0x50916789E9D26364ULL, "6102153_m.fxe"},
	{0x50919C6E53F3115DULL, "ave260tr_long_road_bandits.ncs"},
	{0x50926197F3094A02ULL, "vxht_bartrand_n_5xj_0_0.vlm"},
	{0x5092D221DA963D2CULL, "6179326_m.fxe"},
	{0x5093B288BFF46C4BULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_mere.phy"},
	{0x5093BA74658803E0ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\knockback.gad"},
	{0x50949182F09B81D2ULL, "6071250_f.fxe"},
	{0x5094960F15F129ADULL, "vxdr_primevaltomb_3jp_0_0.vlm"},
	{0x50954A77AB6BB964ULL, "6104804_f.fxe"},
	{0x5095500D08F651FCULL, "art\\characters\\playercharacter\\humanmale\\hlh_masc.mao"},
	{0x50955C0D08F6659FULL, "art\\characters\\playercharacter\\humanmale\\hlh_masc.mmh"},
	{0x50955F9974AE5CB0ULL, "vxw_woundedcoast_1d1_0_1.vlm"},
	{0x5095620D08F67041ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masc.msh"},
	{0x5095842923B2B46BULL, "qun231_party_comment6077138.cl"},
	{0x5095A70E3029DA5EULL, "6099206_m.fxe"},
	{0x5095BFD6DEB6FEFFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupb.phy"},
	{0x5095F76C20F5A6C3ULL, "6066173_m.fxe"},
	{0x509623D31922E48AULL, "vxlt_gamlenshouse_n_5y_0_0.vlm"},
	{0x50965948211BD87BULL, "vxlt_0_lowtown_docks_b5s_0_1.vlm"},
	{0x509673E9C6F3A215ULL, "w_blightlands_77_0.msh"},
	{0x509675E9C6F3A58BULL, "w_blightlands_77_0.mmh"},
	{0x509681E9C6F3BA28ULL, "w_blightlands_77_0.mao"},
	{0x5096CBB132F6346CULL, "pro000_colonel_ewald6040596.cl"},
	{0x509712DAC1FB0A32ULL, "103217709.ogg"},
	{0x50977C651E977C05ULL, "follower_banter6136651.cl"},
	{0x5097BCBB8BBA3B44ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_avel.phy"},
	{0x50984AEEC6EA466EULL, "dae000_porfiria6155381.cl"},
	{0x50984BE6E2199EDAULL, "follower_banter6156164.cl"},
	{0x50985ABBB9BB6885ULL, "vxht_0_hightown_c10_0_0.vlm"},
	{0x50986998E80E257CULL, "vxlt_darktown_6hp_0_0.vlm"},
	{0x50986A92E7670CCEULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90l.gad"},
	{0x5098B5B9623132E2ULL, "725181873.ogg"},
	{0x50996BE81DF59802ULL, "6050447_m.fxe"},
	{0x509982B68CD93D7DULL, "717066792.ogg"},
	{0x50999D4AF7509660ULL, "790107064.ogg"},
	{0x5099D85724A402A6ULL, "dae000_arishok6067700.cl"},
	{0x509A2485EA0E9B8EULL, "vxdr_0_deeproads_ec8_0_0.vlm"},
	{0x509A4F0990BAF301ULL, "255066214.ogg"},
	{0x509B3986ED0A945BULL, "gen_im_arm_cht_comf.uti"},
	{0x509B6DEE96C98E14ULL, "vxw_bonepitdestroyed_a3_25c_0_0.vlm"},
	{0x509BB3EDFCFA9B18ULL, "gen00fl_fenris6107150.cl"},
	{0x509C44D2D092C2DDULL, "vxht_mansion2b_1rk_0_0.vlm"},
	{0x509C8F9636B8D41BULL, "140512770.ogg"},
	{0x509C9E19DD13C050ULL, "6075718_m.fxe"},
	{0x509D2768C6E1BDB0ULL, "wm_ico_playershome_roll.dds"},
	{0x509D5745416DBD3BULL, "6088331_m.fxe"},
	{0x509DE92E1319B014ULL, "mag160_feynriel6020083.cl"},
	{0x509DEB4DF8C3211EULL, "hlac_ht_mansion2_818_57x.dds"},
	{0x509E0E72660A633CULL, "gen00fl_fenris6113850.cl"},
	{0x509E14A4C1E5FCCCULL, "6114415_m.fxe"},
	{0x509E3225653C63EBULL, "6155868_m.fxe"},
	{0x509F0B09E323AE89ULL, "ser221ip_note_meredith.ncs"},
	{0x509F6592E76D22ADULL, "art\\characters\\creatures\\desiredemon\\animation\\tn90l.evt"},
	{0x509FC05DE1C1D504ULL, "hm_friedrich_thug.mor"},
	{0x509FD5AF12999F7FULL, "667072683.ogg"},
	{0x509FF8086B091AC4ULL, "vxht_chantry_n_2tm_0_1.vlm"},
	{0x50A065479B1F7690ULL, "58064494.ogg"},
	{0x50A0F2BF54B18A3DULL, "6162082_m.fxe"},
	{0x50A0FE8643156003ULL, "6089493_m.fxe"},
	{0x50A182D77B75D429ULL, "hf_gencr_raider_reaver.mor"},
	{0x50A1966561AA149BULL, "6079296_m.fxe"},
	{0x50A1A454C2670E73ULL, "972468238.ogg"},
	{0x50A1D09EA001BFB5ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb105a_0d.dds"},
	{0x50A1E4DBE66D8834ULL, "6199023_m.fxe"},
	{0x50A214178E22FD9AULL, "468586531.ogg"},
	{0x50A2BDBFFC37B735ULL, "vxca_caverns2prc_8q1_0_0.vlm"},
	{0x50A2F456722CF894ULL, "vxlt_hangedman_1r8_0_0.vlm"},
	{0x50A33485790D027CULL, "mrl000_marethari6142121.cl"},
	{0x50A3479BF88C397DULL, "vxca_caverns2_94b_0_0.vlm"},
	{0x50A35ED70E1BB10BULL, "110919556.ogg"},
	{0x50A3744912F8CC99ULL, "829313700.ogg"},
	{0x50A3D609C6197513ULL, "954271506.ogg"},
	{0x50A4AD4488FEBAE8ULL, "vxlt_blackemporium_8w_0_0.vlm"},
	{0x50A4DBB218DC14EBULL, "llac_ht_brothel_n_3370_1.dds"},
	{0x50A4FA9D92942910ULL, "one100dg_lia6013854.cl"},
	{0x50A5396904FEA55DULL, "6118753_m.fxe"},
	{0x50A5E1E068D0FA1AULL, "llac_ca_caverns1_24424_7.dds"},
	{0x50A611D71456AF46ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extbrick.dds"},
	{0x50A67FB990ACD677ULL, "gen00fl_aveline6137650.cl"},
	{0x50A6C0D6DEC57304ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupb.mmh"},
	{0x50A6C4D6DEC57A57ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupb.mao"},
	{0x50A6D6D6DEC598A6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupb.msh"},
	{0x50A706EFBAD889C4ULL, "448345011.ogg"},
	{0x50A796E3CF68F29DULL, "mag110_samson6049821.cl"},
	{0x50A7BB5D5DE05046ULL, "ave121_jeven6106046.cl"},
	{0x50A83AD1D683E190ULL, "hlac_dr_primevaltomb_658_4y.dds"},
	{0x50A89662DC04B40EULL, "vxdr_0_deeproads_el9_0_0.vlm"},
	{0x50A8F03F35B7468DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudempty.phy"},
	{0x50A9BF58A19492D5ULL, "gen00fl_aveline6139799.cl"},
	{0x50A9BFB6C1146319ULL, "gen00fl_anders6125425.cl"},
	{0x50A9C4C73547DB11ULL, "art\\levels\\races\\design\\_textures\\plc_ballista_01_0_n.dds"},
	{0x50A9E8516BB3BCA4ULL, "vxlt_undercityencounter_v4_0_0.vlm"},
	{0x50A9FBFEBA5BF02FULL, "vxht_bartrand_n_5w6_0_0.vlm"},
	{0x50AA0CF4CFFB1477ULL, "609800952.ogg"},
	{0x50AA1DE2B3820F77ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small.gad"},
	{0x50AA48C8B53E4287ULL, "vxdr_0_deeproads_d0y_0_0.vlm"},
	{0x50AA5C3689E0FDE7ULL, "hlac_lt_warehouse_-65535_1z.dds"},
	{0x50AAB6AFC288DB5DULL, "vxht_mansion2b_1uj_0_0.vlm"},
	{0x50AAC0C4DF6EC864ULL, "zz_followers_banter6157246.cl"},
	{0x50AAC1AEFF848FDBULL, "6163577_m.fxe"},
	{0x50AB0B33CAE1DFD5ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1ww_basic_1.mao"},
	{0x50AB3AA7491E1CEDULL, "275197570.ogg"},
	{0x50AC2AEFE768F75BULL, "842916782.ogg"},
	{0x50AC815728B2A411ULL, "em_mag_controlled_4.mor"},
	{0x50AD09849CA68675ULL, "hlac_ca_caverns1_mines_24424_7z.dds"},
	{0x50AD0A90D30B7A97ULL, "875428328.ogg"},
	{0x50AD3F3ABC20A493ULL, "6016633_m.fxe"},
	{0x50AD4E09AF18CDEEULL, "hlac_lt_0_lowtown_docks_n_-65539_0x.dds"},
	{0x50AD7C7095B760C1ULL, "dae200cr_viscount6087380.cl"},
	{0x50AE83AAEC5C3937ULL, "6069399_f.fxe"},
	{0x50AEADCBB9A7FDABULL, "follower_banter6163732.cl"},
	{0x50AFF409E331FB68ULL, "ser221ip_note_meredith.utp"},
	{0x50B010AD7DB2706EULL, "vxht_keep_n_61l_0_1.vlm"},
	{0x50B0924FBC47D7EAULL, "286895156.ogg"},
	{0x50B0DDFC1455F2C3ULL, "vxht_chantry_n_2by_0_0.vlm"},
	{0x50B1244777CA037EULL, "hlac_w_encounter2_196608_0z.dds"},
	{0x50B132E2B3885104ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small.evt"},
	{0x50B1663364231114ULL, "vxlt_hangedman_1y9_0_0.vlm"},
	{0x50B1AF8B96FAA6EDULL, "hlac_lt_0_lowtown_a2_65535_7y.dds"},
	{0x50B1DE9E30023CEEULL, "art\\levels\\ztools\\shaders\\blank_sa.dds"},
	{0x50B22F28FBCE4EA3ULL, "w_bonepitdestroyed_a3_4_0.rml"},
	{0x50B24344FD1B5DEFULL, "809302720.ogg"},
	{0x50B24AA7357D24ACULL, "art\\characters\\playercharacter\\humanfemale\\boo_urva.mmh"},
	{0x50B24D29F98DC508ULL, "896498813.ogg"},
	{0x50B250A7357D2E9EULL, "art\\characters\\playercharacter\\humanfemale\\boo_urva.msh"},
	{0x50B26FA1A690F447ULL, "mer221_sabin6080530.cl"},
	{0x50B2C07292F87404ULL, "lowtown_fem_prostitute_sta6152321.cl"},
	{0x50B2F38600F8D893ULL, "fen110_fenris6118110.cl"},
	{0x50B39DDC4DAD026FULL, "one350_terath6087131.cl"},
	{0x50B3C48668DAC29AULL, "100382127.ogg"},
	{0x50B4D847135B4587ULL, "6129746_m.fxe"},
	{0x50B55785BF3F080EULL, "dr_0_deeproads_5849.rml"},
	{0x50B5B55D829F2E03ULL, "532330331.ogg"},
	{0x50B626585886C9C5ULL, "bur_hole_clu.mao"},
	{0x50B674797DAD9523ULL, "6040814_m.fxe"},
	{0x50B6E03795777641ULL, "vxht_0_hightown_n_c62_0_0.vlm"},
	{0x50B7778D30FF3389ULL, "gen00fl_anders6125955.cl"},
	{0x50B7B6BBAF006143ULL, "lt_0_lowtown_docks_241_0.msh"},
	{0x50B7C4BBAF00789DULL, "lt_0_lowtown_docks_241_0.mmh"},
	{0x50B7C8BBAF007FAEULL, "lt_0_lowtown_docks_241_0.mao"},
	{0x50B86DDBAC5208AFULL, "vxht_bartrand_n_5z1_0_0.vlm"},
	{0x50B898288ABEB351ULL, "hlac_lt_0_lowtown_docks_a3_-131073_1y.dds"},
	{0x50B8C5594EBF37CEULL, "hlac_lt_blackemporium_1055_8z.dds"},
	{0x50B9223443053077ULL, "vxca_cavern1_haunted_e6h_0_0.vlm"},
	{0x50B97C5AE2A1DD44ULL, "754219978.ogg"},
	{0x50B99B1A875A3DBBULL, "vxht_keep_n_8bo_0_0.vlm"},
	{0x50B9C3B0BEDB47D4ULL, "docks_fem_sailor_mob6176418.cl"},
	{0x50BA0809E31734CFULL, "6060554_f.fxe"},
	{0x50BA1483FE1A3548ULL, "hlac_ht_keep_3731_0y.dds"},
	{0x50BA86108B042D08ULL, "character_character_p168.psh.fxo.dx11.assembly"},
	{0x50BA947D920E5B18ULL, "vxlt_0_lowtown_a2_es1_0_0.vlm"},
	{0x50BB34F7DC038612ULL, "599551992.ogg"},
	{0x50BBA9151673DFA5ULL, "hlac_dr_0_deeproads_5842_11z.dds"},
	{0x50BBD28E734982A0ULL, "vxht_brothel_n_3sk_0_0.vlm"},
	{0x50BBF097DCADAAA0ULL, "6071057_m.fxe"},
	{0x50BBFB1869A32E0CULL, "lowtown_male_alienage1_sta6162741.cl"},
	{0x50BC1819A062501DULL, "vxw_encounter3_4n_0_0.vlm"},
	{0x50BC77A07DDCDA35ULL, "701084146.ogg"},
	{0x50BC7DD0FBC3F9C8ULL, "vxw_sundermount_2ax_0_0.vlm"},
	{0x50BC9BDBB517822CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilec.msh"},
	{0x50BCA1DBB5178C1EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilec.mmh"},
	{0x50BCADDBB517A0B5ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilec.mao"},
	{0x50BCFD8123EA00C2ULL, "weapon_rendertoshadowmap_p0.vsh.fxo.dx11.assembly"},
	{0x50BD3D0A778224A4ULL, "vxdr_primevaltomb_wc_0_0.vlm"},
	{0x50BD4DD171B4F0ADULL, "vxht_brothel_n_4dl_0_0.vlm"},
	{0x50BDDE9E22446FCFULL, "hlac_lt_0_lowtown_a2_n_c_65537_6y.dds"},
	{0x50BDE24027019103ULL, "6087693_m.fxe"},
	{0x50BDF73E71CC9010ULL, "6083157_m.fxe"},
	{0x50BE4AE2B39304FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_small.ani"},
	{0x50BE5A9D167E6B10ULL, "6144476_m.fxe"},
	{0x50BF7C7421C56420ULL, "8484318.ogg"},
	{0x50BFDFEAD9FAB946ULL, "vxw_bonepitdestroyed_a3_1gp_0_0.vlm"},
	{0x50BFE3B738DFBA32ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_master_d.dds"},
	{0x50C0056D40D339C0ULL, "897276560.ogg"},
	{0x50C092B942355E00ULL, "vxw_woundedcoast_19s_0_0.vlm"},
	{0x50C0A605C6C11AE6ULL, "1072319061.ogg"},
	{0x50C126B9D9AE9AECULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stones.msh"},
	{0x50C12CB9D9AEA4DEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stones.mmh"},
	{0x50C138B9D9AEB975ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stones.mao"},
	{0x50C17F2A3243A24EULL, "vxht_mansion2_1n0_0_0.vlm"},
	{0x50C1AD79385E4A24ULL, "qun220cr_javaris6071512.cl"},
	{0x50C21D22E8EAD65DULL, "6112553_m.fxe"},
	{0x50C32E7B03125528ULL, "mcr371_scared_mage_f6086927.cl"},
	{0x50C3371121E264C8ULL, "6176424_m.fxe"},
	{0x50C42968E222649DULL, "nat380_nathaniel6066369.cl"},
	{0x50C43EAF35890769ULL, "642903195.ogg"},
	{0x50C4E01FADFD45FCULL, "208395513.ogg"},
	{0x50C55214876D8EC1ULL, "vxht_0_hightown_n_c31_0_0.vlm"},
	{0x50C56F56AC818906ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9rr_stair.phy"},
	{0x50C5939761A05DAEULL, "6100320_m.fxe"},
	{0x50C5A5D910D25B75ULL, "6206676_m.fxe"},
	{0x50C5B977A17621FFULL, "695764000.ogg"},
	{0x50C614EEEE7D3BC2ULL, "hlac_lt_0_lowtown_-3_5x.dds"},
	{0x50C61DDA95E079AEULL, "6127096_f.fxe"},
	{0x50C62319B67381AAULL, "6100085_m.fxe"},
	{0x50C638543630E95BULL, "dae000_bodhan6129147.cl"},
	{0x50C66C0DC5558490ULL, "vxlt_warehouse_111_0_0.vlm"},
	{0x50C6A97A23C60F89ULL, "230378431.ogg"},
	{0x50C6FC686D488EF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_exit.ani"},
	{0x50C71E7EB0D84CD6ULL, "art\\vfx\\combat\\vfx_c_rw_attack2_c.impa.ani"},
	{0x50C727020A4376A1ULL, "hlac_lt_hangedman_-1_18y.dds"},
	{0x50C72C3AB8753FAEULL, "6054058_m.fxe"},
	{0x50C796D61F3FEF8EULL, "6127107_m.fxe"},
	{0x50C7CD05E017B2C4ULL, "6122083_m.fxe"},
	{0x50C8251B97A5966BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_anda.phy"},
	{0x50C8639DB0E60B99ULL, "586036256.ogg"},
	{0x50C90CA09E3A0362ULL, "art\\vfx\\textures\\vfx_small_dots.dds"},
	{0x50C95668B55B5E62ULL, "147452559.ogg"},
	{0x50CA074C3C19F2E5ULL, "6174889_m.fxe"},
	{0x50CA89F69258689DULL, "vxw_encounter3_1m_0_0.vlm"},
	{0x50CA951645F2856DULL, "vxlt_0_lowtown_a2_n_c_hun_0_0.vlm"},
	{0x50CACC7647DCEB9FULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.mmh"},
	{0x50CACF11187B0B95ULL, "mag121_viveka6083372.cl"},
	{0x50CAEFADEDBA1248ULL, "vxw_sundermount_2jy_0_0.vlm"},
	{0x50CB1352171C34FEULL, "456575743.ogg"},
	{0x50CB218BBA2B9C42ULL, "hlac_ca_caverns1_mines_24424_22z.dds"},
	{0x50CB71CAC9F7A478ULL, "hlac_lt_0_lowtown_a2_n_c_131072_5y.dds"},
	{0x50CBB2A454DD3A3CULL, "art\\vfx\\materials\\gui\\vfx_gui_mov_point.mao"},
	{0x50CBDAF02139FC86ULL, "vxlt_0_lowtown_a2_hel_0_0.vlm"},
	{0x50CC2347E942B784ULL, "vxdr_0_deeproads_fi3_0_0.vlm"},
	{0x50CC2FB549299E46ULL, "vxdr_primevaltomb_8nl_0_0.vlm"},
	{0x50CD319EF9E12AEAULL, "615623155.ogg"},
	{0x50CD5E287150C78EULL, "vxht_chantry_33a_0_0.vlm"},
	{0x50CD7AC66E98DA32ULL, "vrc321cr_golem_shade.ncs"},
	{0x50CD86537DD1B42FULL, "285668488.ogg"},
	{0x50CDD7686D4E6E3BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_exit.gad"},
	{0x50CE51C7CBF0D1C6ULL, "vxw_bonepitdestroyed_a3_1hq_0_0.vlm"},
	{0x50CE71EB77766BF4ULL, "art\\vfx\\cinematic\\vfx_arishok_spittle_c.impa.ani"},
	{0x50CE9A6D2DACDF9FULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm04.msh"},
	{0x50CEA06D2DACEA41ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm04.mmh"},
	{0x50CFABE3FF76101DULL, "hlac_ht_chantry_2534_28x.dds"},
	{0x50CFBC9856C373B2ULL, "llac_ht_0_hightown_a2_-65538_1.dds"},
	{0x50CFD8D51E5B7EE1ULL, "vxw_sundermount_2os_0_0.vlm"},
	{0x50D09494EF045F6AULL, "292759231.ogg"},
	{0x50D0CDAAEF031051ULL, "289653699.ogg"},
	{0x50D21E9B9ECC5974ULL, "342400409.ogg"},
	{0x50D239699E638EBDULL, "zz_followers_banter6159540.cl"},
	{0x50D24BE2D08C725AULL, "782257395.ogg"},
	{0x50D2C1E1EED146FFULL, "zz_and_debug6134050.cl"},
	{0x50D32F51932F2B19ULL, "hlac_lt_gallowstemplar_n_2394_6z.dds"},
	{0x50D3C3F17963A741ULL, "vxht_0_hightown_n_c80_0_0.vlm"},
	{0x50D3F1CB85F2FC16ULL, "vxht_brothel_n_623_0_0.vlm"},
	{0x50D4136BD9BA8C83ULL, "hlac_w_sundermount_196611_0x.dds"},
	{0x50D4267EB0E2E4C7ULL, "art\\vfx\\combat\\vfx_c_rw_attack2_c.impa.evt"},
	{0x50D457305AFB7573ULL, "6051845_m.fxe"},
	{0x50D46EEB777AD256ULL, "art\\vfx\\cinematic\\vfx_arishok_spittle_c.impa.gad"},
	{0x50D494F6A8699A2AULL, "6105385_m.fxe"},
	{0x50D4F4686D54BE20ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_exit.evt"},
	{0x50D5472826E4ECEFULL, "dae011_conspiracy_mob.bnk"},
	{0x50D5DF179EE62F1BULL, "115207630.ogg"},
	{0x50D5F42E0DE8CC45ULL, "150053825.ogg"},
	{0x50D644B1BFC775AAULL, "vxdr_primevaltomb_86h_0_0.vlm"},
	{0x50D6B2D2418EF8F3ULL, "flame04.dds"},
	{0x50D6C1E9EE49BB65ULL, "qun111_ketojan6040937.cl"},
	{0x50D7D98F4967DEC2ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.dura.gad"},
	{0x50D8692826E75C57ULL, "dae011_conspiracy_mob.cnv"},
	{0x50D8A11A7DCE6763ULL, "799112099.ogg"},
	{0x50D8E2E0FF417A37ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_sa.dds"},
	{0x50D95E988EBA7828ULL, "6060406_m.fxe"},
	{0x50D98D6E5430AD20ULL, "ave260tr_long_road_bandits.utt"},
	{0x50D9C7FE08E1C387ULL, "87268846.ogg"},
	{0x50D9D6F9AF03A177ULL, "vxlt_hangedman_n_227_0_0.vlm"},
	{0x50D9E8E4E8396C8CULL, "gen00fl_varric6098540.cl"},
	{0x50DA4CCD13301506ULL, "vxlt_0_lowtown_a2_hbm_0_0.vlm"},
	{0x50DA63952A51D8EDULL, "hlac_lt_0_lowtown_a2_-131071_0z.dds"},
	{0x50DA68456414378DULL, "vxlt_gallowscourtyard_6w6_0_0.vlm"},
	{0x50DA9288F165653EULL, "vxlt_gallowscourtyard_693_0_0.vlm"},
	{0x50DA9B946746E309ULL, "6169847_m.fxe"},
	{0x50DAF0D1DF61C06EULL, "ca_cavern1_haunted_mmap_ph.dds"},
	{0x50DAF9F7856964EAULL, "vxlt_0_lowtown_a2_n_c_gxq_0_0.vlm"},
	{0x50DB477EB0E93B10ULL, "art\\vfx\\combat\\vfx_c_rw_attack2_c.impa.gad"},
	{0x50DB8212BB03F7F9ULL, "354390720.ogg"},
	{0x50DB89EB77811EB5ULL, "art\\vfx\\cinematic\\vfx_arishok_spittle_c.impa.evt"},
	{0x50DBC4F9D6743D92ULL, "vxlt_blackemporium_15w_0_0.vlm"},
	{0x50DC489A5F9A5F65ULL, "6149658_m.fxe"},
	{0x50DCAB3BBCC7F34FULL, "6017043_m.fxe"},
	{0x50DDAF7842A04E7EULL, "character_character_p131.psh.fxo.dx11.assembly"},
	{0x50DDE05F7978DBC0ULL, "436244210.ogg"},
	{0x50DE4AB210519761ULL, "vxw_sundermount_2pr_0_0.vlm"},
	{0x50DE6C399329BA1FULL, "175160528.ogg"},
	{0x50DEFC8F496E38B9ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.dura.evt"},
	{0x50DF2494395B6923ULL, "vxca_caverns2_4od_0_0.vlm"},
	{0x50DF7C476B1ADD07ULL, "vxca_caverns1_mines_hm4_0_0.vlm"},
	{0x50E0B61DEF2F7369ULL, "w_bonepitdestroyed_a3_240_0.mao"},
	{0x50E0C21DEF2F8702ULL, "w_bonepitdestroyed_a3_240_0.mmh"},
	{0x50E0C41DEF2F8AB8ULL, "w_bonepitdestroyed_a3_240_0.msh"},
	{0x50E148097F49A4D4ULL, "45421561.ogg"},
	{0x50E16B5D147B3EEBULL, "63900568.ogg"},
	{0x50E18487ED748E81ULL, "6092607_f.fxe"},
	{0x50E3B0EFD4491EDFULL, "mag100_karras6034740.cl"},
	{0x50E3B4956257FECAULL, "hlac_ht_chantry_2534_3x.dds"},
	{0x50E3BB4D59F7512EULL, "239520571.ogg"},
	{0x50E455C9389A2A22ULL, "6196575_f.fxe"},
	{0x50E4B68EB1BD8E2AULL, "vxdr_primevaltomb_81i_0_0.vlm"},
	{0x50E51D28FBF989BAULL, "w_bonepitdestroyed_a3_4_0.msh"},
	{0x50E52B28FBF9A1F3ULL, "w_bonepitdestroyed_a3_4_0.mao"},
	{0x50E52F28FBF9A800ULL, "w_bonepitdestroyed_a3_4_0.mmh"},
	{0x50E5769526384933ULL, "ef_one_mo_lia.mor"},
	{0x50E59E285890534CULL, "hlac_ca_caverns1_mines_24425_5z.dds"},
	{0x50E5A2934187AF24ULL, "vxca_caverns1_clj_0_0.vlm"},
	{0x50E5E0EF73308AB2ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.mmh"},
	{0x50E60285B74C46B0ULL, "l_dae380ar_deep_roads.wbd"},
	{0x50E60E6536FAC5ACULL, "465888555.ogg"},
	{0x50E675EA3AF97D88ULL, "vxlt_0_lowtown_a2_n_c_euf_0_0.vlm"},
	{0x50E6BABC743BF2EFULL, "6070578_m.fxe"},
	{0x50E6D1B343BD2BE4ULL, "vxlt_0_lowtown_a2_n_c_i0i_0_0.vlm"},
	{0x50E7223606805D45ULL, "538260993.ogg"},
	{0x50E799F617653C38ULL, "vxht_0_hightown_n_ajq_0_0.vlm"},
	{0x50E7A6D1D31AF0F0ULL, "vxlt_gallowsprison_n_6cf_0_0.vlm"},
	{0x50E7B61A49794954ULL, "821037588.ogg"},
	{0x50E7C6A1FC7BAA14ULL, "fxc_lactate_blend.mao"},
	{0x50E7D367D1A64348ULL, "vxw_sundermount_2pw_0_0.vlm"},
	{0x50E863844916A266ULL, "356564496.ogg"},
	{0x50E8E0A7ED4DCD46ULL, "fx_crescent02.mao"},
	{0x50E914DF854417C6ULL, "hlac_ht_chantry_2534_34z.dds"},
	{0x50E9210CF39C0158ULL, "hlac_ht_0_hightown_a3_65533_0z.dds"},
	{0x50EA119D6FBAAE9DULL, "vxht_mansion_n_616_0_0.vlm"},
	{0x50EA4498E06F10F6ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_death.ani"},
	{0x50EA533BC03B3A4FULL, "vxca_caverns1_mines_imi_0_0.vlm"},
	{0x50EA7ED582AF8F29ULL, "420079472.ogg"},
	{0x50EA864ADD70D702ULL, "hf_noble_gen_2.mor"},
	{0x50EACCADC5670746ULL, "6130103_m.fxe"},
	{0x50EB0512FEAD9C90ULL, "vxca_cavern1_haunted_gk7_0_1.vlm"},
	{0x50EB727608374AAEULL, "vxht_0_hightown_c3i_0_0.vlm"},
	{0x50EBC31E1245E8B5ULL, "6119645_f.fxe"},
	{0x50EBD49076285486ULL, "455724177.ogg"},
	{0x50EC048F4978D0A8ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.dura.ani"},
	{0x50EC0D9DD51D7815ULL, "llac_dr_primevaltomb_658_8.dds"},
	{0x50ECBDF431489A3EULL, "859445362.ogg"},
	{0x50ED2D7EE1086A8DULL, "vxht_mansion2_3dg_0_0.vlm"},
	{0x50EDC211E7EA68DDULL, "282602342.ogg"},
	{0x50EE490BB735B062ULL, "14423404.ogg"},
	{0x50EEC4234852204CULL, "llac_lt_0_lowtown_a2_-65535_4.dds"},
	{0x50EEE77FA4EE49C3ULL, "406176106.ogg"},
	{0x50F01833998F1BDFULL, "vxca_cavern1_haunted_jl1_0_0.vlm"},
	{0x50F04E7AC264724BULL, "vxca_cavern1_haunted_j5v_0_0.vlm"},
	{0x50F0DFE8E63C4D87ULL, "hlac_lt_0_lowtown_docks_a3_-65538_8x.dds"},
	{0x50F11786405D8E2AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed2.gad"},
	{0x50F1A3431A99D9C6ULL, "mag211_marethari6025356.cl"},
	{0x50F21036DD7708C8ULL, "hlac_ht_0_hightown_65535_1z.dds"},
	{0x50F25368AC6B8037ULL, "hlac_ca_caverns2prc_5802_50x.dds"},
	{0x50F3008662636523ULL, "24282762.ogg"},
	{0x50F33F1818A56003ULL, "331988388.ogg"},
	{0x50F3470029460374ULL, "llac_lt_0_lowtown_a2_n_c_65533_2.dds"},
	{0x50F3AEB9037B0D06ULL, "dae000ip_crowd_ff2_talk7.utp"},
	{0x50F3B96AA8FB54B7ULL, "hlac_w_encounter2_131077_0z.dds"},
	{0x50F3C87FA57385F4ULL, "208477875.ogg"},
	{0x50F3C88F9743C895ULL, "256337939.ogg"},
	{0x50F3CD7647FFF2DCULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.anb"},
	{0x50F400860B9994B1ULL, "78913193.ogg"},
	{0x50F491C846B570D5ULL, "179788265.ogg"},
	{0x50F4BCA7CB714AE9ULL, "647367600.ogg"},
	{0x50F4F6D4F38994B3ULL, "366358295.ogg"},
	{0x50F524207FBB2A98ULL, "634828672.ogg"},
	{0x50F55172FD82941EULL, "hlac_lt_0_lowtown_a2_n_c_131072_2z.dds"},
	{0x50F5DC8D48093010ULL, "6043918_f.fxe"},
	{0x50F6472B70F4CD7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_reaction.ani"},
	{0x50F650D46144854AULL, "mer221_sabin6080555.cl"},
	{0x50F69B693F8A0CBEULL, "lightprobeproj.psh.fxo"},
	{0x50F71545398D52ADULL, "vxht_brothel_n_4xh_0_0.vlm"},
	{0x50F74BFF4800688DULL, "vxlt_gallowscourtyard_6y4_0_0.vlm"},
	{0x50F74C98E079A8E7ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_death.evt"},
	{0x50F7C290E1DAE0E4ULL, "gen_im_cns_cft_wrn_08.uti"},
	{0x50F7CE95DEA714F0ULL, "vxlt_gallowstemplar_n_65x_0_0.vlm"},
	{0x50F7FA8640637BC1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed2.evt"},
	{0x50F81DEBFA1DBC34ULL, "fxa_cen_aoe_snow.mao"},
	{0x50F82BDBAFAC4DE5ULL, "mag311_emile_de_launcet6043714.cl"},
	{0x50F84D9D9341275BULL, "141724585.ogg"},
	{0x50F8650D094AF2E4ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masc.phy"},
	{0x50F8BA1309BE82DCULL, "210357655.ogg"},
	{0x50F8C518B23152CFULL, "vxca_caverns1_mines_ijh_0_0.vlm"},
	{0x50F8D3426EEB19FBULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\rightlegram.ani"},
	{0x50F953A38CEE9622ULL, "277686608.ogg"},
	{0x50F99980A6D25B2AULL, "follower_banter6153985.cl"},
	{0x50F9B37E35FF71E4ULL, "llac_lt_0_lowtown_a3_n_c_-65537_2.dds"},
	{0x50F9D6A2D9EC4795ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.mmh"},
	{0x50F9DAE9CCB45C95ULL, "578881972.ogg"},
	{0x50F9E452FA2D632EULL, "vxht_0_hightown_c6j_0_0.vlm"},
	{0x50F9E594630C38A5ULL, "362137818.ogg"},
	{0x50FA399986162498ULL, "vxdr_primevalentrance_394_0_0.vlm"},
	{0x50FA459F7AAD8FBEULL, "6052746_m.fxe"},
	{0x50FA5AAB725AD71AULL, "517548501.ogg"},
	{0x50FAA16D2DD226BEULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm04.phy"},
	{0x50FAE0351DB9AC86ULL, "vxht_brothel_n_47j_0_0.vlm"},
	{0x50FAF4E380F186E2ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef02.msh"},
	{0x50FAF6E380F18998ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef02.mmh"},
	{0x50FB15CB02E2019DULL, "6135147_m.fxe"},
	{0x50FB36C685BA5407ULL, "gen00fl_fenris6108763.cl"},
	{0x50FB7567F848C864ULL, "6051792_m.fxe"},
	{0x50FC05CAF614F87AULL, "mag210_arianni6031656.cl"},
	{0x50FDFA5596D5C99FULL, "static_static_p35.psh.fxo.dx11.assembly"},
	{0x50FE588D5469D86FULL, "43093626.ogg"},
	{0x50FE6D98E07FFF30ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_death.gad"},
	{0x50FEC1A8F2273970ULL, "956498900.ogg"},
	{0x50FF0436AC9156B1ULL, "760975551.ogg"},
	{0x51003B2B28D44D94ULL, "vxw_encounter2_camp_66_0_0.vlm"},
	{0x51004EDB31C2249FULL, "nat380_nathaniel6071024.cl"},
	{0x510059BD5D043263ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_pc.evt"},
	{0x510088B7D6BF154CULL, "6106661_m.fxe"},
	{0x5100D11CBB26DD24ULL, "vxht_brothel_n_62q_0_0.vlm"},
	{0x51016F66DDE5CD70ULL, "vxht_0_hightown_a3_b95_0_0.vlm"},
	{0x5101D77404E604ADULL, "vxlt_0_lowtown_a2_h2b_0_0.vlm"},
	{0x5102E59C6B992120ULL, "vxht_bartrand_n_6r5_0_0.vlm"},
	{0x5102F754B1817E7BULL, "vxdr_primevaltomb_6j7_0_0.vlm"},
	{0x5103172B740CB4B0ULL, "llac_ht_bartrand_n_3371_1.dds"},
	{0x51034C580A4CB42DULL, "dae000ip_crowd_mf4_talk10.utp"},
	{0x51038E38981FAD8EULL, "hlac_lt_undercityencounter_-1_11z.dds"},
	{0x5103CF89B800AD91ULL, "follower_banter6154057.cl"},
	{0x510440C513409439ULL, "279043419.ogg"},
	{0x51045B2D9935BE6CULL, "ser221_party_comments6128806.cl"},
	{0x51045F2B7101334FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_reaction.evt"},
	{0x5104CEFBD1B29748ULL, "static_static_p30.psh.fxo"},
	{0x5104E51FDD3F4A66ULL, "vxw_sundermount_35n_0_1.vlm"},
	{0x5104F286406DF810ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed2.ani"},
	{0x5104FDE380F9EA23ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef02.phy"},
	{0x51056B098846ED4EULL, "vxht_bartrand_n_5qq_0_0.vlm"},
	{0x5105BFB60F2DF908ULL, "394728766.ogg"},
	{0x5105C2D834397572ULL, "788305102.ogg"},
	{0x510644D519736A2CULL, "422804769.ogg"},
	{0x51068239047072C1ULL, "vxw_woundedcoast_as_0_0.vlm"},
	{0x5106DB426EF764F2ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\rightlegram.evt"},
	{0x510732BD5D0A0EE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_pc.gad"},
	{0x510736F5A4276B4FULL, "vxca_caverns1_mines_igk_0_0.vlm"},
	{0x5107C1F52DE2B11EULL, "6197957_m.fxe"},
	{0x5107CC98BBA92A5FULL, "533317470.ogg"},
	{0x5107D1462E6452EFULL, "6169188_m.fxe"},
	{0x5107D3A2D9F88022ULL, "art\\vfx\\environment\\vfx_pois_gas_area_lrg.anb"},
	{0x5107F1A7ED689AE8ULL, "fx_crescent02.dds"},
	{0x5108412EF8A3A15EULL, "vxht_mansion_n_6ii_0_0.vlm"},
	{0x5108EFE66D85DA5AULL, "vxlt_0_lowtown_docks_n_b5i_0_1.vlm"},
	{0x510925B64CAC7967ULL, "lgt310_mis_burn_body6180587.cl"},
	{0x5109600F929F0C0AULL, "vxlt_0_lowtown_docks_a3_9zr_0_0.vlm"},
	{0x5109DCE61C9C1AA1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_n.dds"},
	{0x510A33085F965A58ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_mountain_n.dds"},
	{0x510A7CA8658A3509ULL, "lt_0_lowtown_a2_n_c_mmi.gff"},
	{0x510AA14BAE34FFD1ULL, "hlac_w_encounter2_65537_0y.dds"},
	{0x510AB17DB069F4D5ULL, "vxlt_gallowstemplar_f_35h_0_0.vlm"},
	{0x510B602B710753F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_reaction.gad"},
	{0x510B98529D80A4A4ULL, "zz_vrc_debug6200167.cl"},
	{0x510C28931BAD911FULL, "49909619.ogg"},
	{0x510C9C9A29065A13ULL, "gen00fl_varric6103458.cl"},
	{0x510CFA426EFC0485ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\rightlegram.gad"},
	{0x510D30FBD20FB3E7ULL, "vxlt_gallowstemplar_n_4eu_0_0.vlm"},
	{0x510D34099095D753ULL, "hlac_lt_0_lowtown_65533_2x.dds"},
	{0x510D45B0A1B47619ULL, "ambient_movement_patterns.gda"},
	{0x510DA703E01D4534ULL, "lla_ht_0_hightown_65537_0.dds"},
	{0x510DE2D18528A6DAULL, "w_woundedcoast_184_0.mmh"},
	{0x510DE6D18528AD11ULL, "w_woundedcoast_184_0.mao"},
	{0x510DF4D18528C520ULL, "w_woundedcoast_184_0.msh"},
	{0x510E41BD5D1046C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_pc.ani"},
	{0x510E9204F166A7C7ULL, "vxht_chantry_n_2lk_0_0.vlm"},
	{0x510EAD081ACA6614ULL, "vxw_encounter2_camp_17_0_0.vlm"},
	{0x510EDFEF73538E59ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.anb"},
	{0x510FBC91B9FD1845ULL, "6020069_m.fxe"},
	{0x5110BE3C18F2BC0CULL, "zz_mcr_debug6080782.cl"},
	{0x511114483E3812EAULL, "6132783_m.fxe"},
	{0x51112CD70A7351FDULL, "524320122.ogg"},
	{0x51116931A37796FBULL, "vxdr_primevaltomb_6q6_0_0.vlm"},
	{0x51117AFA2B96BE89ULL, "335973366.ogg"},
	{0x5111ACDE3807836BULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.cess.ani"},
	{0x5111B082E4517740ULL, "llac_lt_0_lowtown_a3_n_c_0_9.dds"},
	{0x5111B70818B0906DULL, "6153428_m.fxe"},
	{0x51124C533D5E65CFULL, "vxht_mansion2b_31z_0_0.vlm"},
	{0x511284A8017765A3ULL, "gen00fl_anders6128367.cl"},
	{0x5112B489766D29D9ULL, "351227694.ogg"},
	{0x5112DE44E1F41D4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_exit.ani"},
	{0x5112E91F3D162635ULL, "hlac_lt_darktown_-1_16z.dds"},
	{0x51133D58AC9BB9E0ULL, "267679711.ogg"},
	{0x51139CCFBA34EEBFULL, "dae011_corff_sta6142569.cl"},
	{0x5113DCE67A3D05CEULL, "vxht_bartrand_n_5tr_0_0.vlm"},
	{0x5113FA217E33AEC6ULL, "756889487.ogg"},
	{0x5114690602F818C6ULL, "vxdr_primevaltomb_8uq_0_0.vlm"},
	{0x51152367A58B21F0ULL, "mag311_emile_de_launcet6043832.cl"},
	{0x5115C256947889D2ULL, "qun101ip_at_coast.utp"},
	{0x5115D0510443AED9ULL, "6170658_m.fxe"},
	{0x5116A45F04D0FCFCULL, "follower_barks6153687.cl"},
	{0x5116B30BEA99B9DEULL, "vxht_mansion_n_6lj_0_0.vlm"},
	{0x5116BA4B1E3884ADULL, "6125595_m.fxe"},
	{0x5116E5B06A58B222ULL, "vxht_keep_n_5dy_0_0.vlm"},
	{0x5117D6F1671BA813ULL, "follower_banter6173089.cl"},
	{0x51184BCAE7FC538FULL, "253834147.ogg"},
	{0x5118C751B49AD7E5ULL, "mag291_torpor6026957.cl"},
	{0x5118C881924096B0ULL, "6152750_m.fxe"},
	{0x5118F7A6763D2348ULL, "6201531_m.fxe"},
	{0x511924FBDE899DE8ULL, "70230288.ogg"},
	{0x5119593A495E6B51ULL, "6162279_m.fxe"},
	{0x511964A87500A048ULL, "6135201_m.fxe"},
	{0x5119C75F99CA0D32ULL, "llac_w_blightlands_65539_1.dds"},
	{0x5119EB44E1FA518EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_exit.gad"},
	{0x5119F964BA63A4FFULL, "862535862.ogg"},
	{0x511A549E6D1996AEULL, "vxlt_0_lowtown_docks_aq4_0_0.vlm"},
	{0x511B5D3D6E7599BDULL, "vxca_caverns1_mines_hzo_0_0.vlm"},
	{0x511CD259FA9E9322ULL, "vxdr_slavers_1by_0_0.vlm"},
	{0x511D1145A0BED49AULL, "vxht_mansion2_2ek_0_1.vlm"},
	{0x511D4641BF821B5BULL, "hlac_ht_mansion2b_818_0y.dds"},
	{0x511D73806D839F59ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_varric_.gad"},
	{0x511DCE08240AF0F0ULL, "vxht_mansion_n_4vd_0_0.vlm"},
	{0x511DECC7D586A4FEULL, "vxw_encounter1_im_0_1.vlm"},
	{0x511E4F5AA112A6DAULL, "6152352_m.fxe"},
	{0x511E94DE3811E4E2ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.cess.evt"},
	{0x511E9C1EE1421374ULL, "6096407_m.fxe"},
	{0x511EECC4C65713C1ULL, "157022149.ogg"},
	{0x511FCE7A2B5BDDCEULL, "717844922.ogg"},
	{0x511FE644E1FEB56DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_exit.evt"},
	{0x511FE79AF06AFC56ULL, "vxw_bonepitmining_a2_2ac_0_0.vlm"},
	{0x511FF2A19165185FULL, "854588216.ogg"},
	{0x512044CAA480A651ULL, "827983190.ogg"},
	{0x5120503434E3C1F9ULL, "478679073.ogg"},
	{0x51205C70B01E5251ULL, "hlac_lt_darktown_-1_23z.dds"},
	{0x512087D194DE2A6BULL, "vxht_bartrand_n_3p9_0_0.vlm"},
	{0x5120ECB635EA9E64ULL, "gen_im_arm_cht_comc_t5.uti"},
	{0x5121107291937764ULL, "vxdr_primevaltomb_5a7_0_1.vlm"},
	{0x51212FAC2CB1A24CULL, "dae000ip_crowd_mf4_talk8.utp"},
	{0x51214CC4EF738B21ULL, "254258234.ogg"},
	{0x51218ED068FD54BAULL, "nat380_nathaniel_banter6067285.cl"},
	{0x5121D1C3E7CB32CFULL, "vxlt_gallowstemplar_6sz_0_0.vlm"},
	{0x5121D941570319E1ULL, "6080801_m.fxe"},
	{0x5121DE6169BCA307ULL, "213947982.ogg"},
	{0x5121EAA539F64AE4ULL, "lit_aoe_01_l2.mao"},
	{0x5121F555D2CAE0B2ULL, "t3_wdg_tw1.tnt"},
	{0x5121F70048847B1DULL, "vxht_keep_n_6he_0_0.vlm"},
	{0x512352507B43BB8FULL, "701937031.ogg"},
	{0x5123DB8FA0C9FA91ULL, "zz_mer_debug6039440.cl"},
	{0x5123DE66347921E9ULL, "82215531.ogg"},
	{0x512484806D89DAA7ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_varric_.ani"},
	{0x5124B4B2DF2A73AFULL, "vxht_mansion2b_1bq_0_0.vlm"},
	{0x5124B500125083EFULL, "6163388_m.fxe"},
	{0x512538AA4A475BAEULL, "6152576_m.fxe"},
	{0x512593DE381801F5ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.cess.gad"},
	{0x5125E8FB59D66794ULL, "vxca_caverns1_j54_0_0.vlm"},
	{0x51268D023AC2C587ULL, "vxht_0_hightown_a3_c8t_0_0.vlm"},
	{0x51269A05A1268149ULL, "m_dae140ar_bone_pit.wbd"},
	{0x5126FFBF29DEEC24ULL, "w_encounter3_36_0.msh"},
	{0x512711BF29DF0A8DULL, "w_encounter3_36_0.mao"},
	{0x512715BF29DF11C6ULL, "w_encounter3_36_0.mmh"},
	{0x5127312C4EB07FCDULL, "162304834.ogg"},
	{0x5127D68566F6B8C8ULL, "6130101_m.fxe"},
	{0x5127F297C48BF62AULL, "1027088217.ogg"},
	{0x5128470A1BF7F377ULL, "6164433_m.fxe"},
	{0x51285078F7893FB1ULL, "374458187.ogg"},
	{0x512868181EBA8449ULL, "art\\cinematics\\props\\prp_barge02.mao"},
	{0x5128A3F03D380C03ULL, "vxlt_0_lowtown_a2_n_c_gl8_0_0.vlm"},
	{0x5129176A5E411E47ULL, "qun131cr_warrens_thug4.utc"},
	{0x51297B9E9E928A12ULL, "vxlt_house_e5_0_0.vlm"},
	{0x5129E55E0DD08E3FULL, "6171570_m.fxe"},
	{0x512A09511317DAD9ULL, "use_pressed.cur"},
	{0x512A44451905029CULL, "ser270_cut_motherreveal6131378.cl"},
	{0x512B614300120CA7ULL, "gen00fl_aveline6106939.cl"},
	{0x512B62CBFECCC39DULL, "vxlt_warehouse_9q_0_0.vlm"},
	{0x512BA6CA3AC3A1E2ULL, "mag291_torpor6056810.cl"},
	{0x512C00BBD4311179ULL, "gen00fl_anders6196683.cl"},
	{0x512C0E144161CE2BULL, "hlac_ht_0_hightown_-131073_0z.dds"},
	{0x512C23219729B9F8ULL, "hf_mystery_friend.mor"},
	{0x512C3710A9D30EE7ULL, "vxlt_gallowstemplar_f_43k_0_0.vlm"},
	{0x512CC0CCED0B77E9ULL, "hlac_w_bonepitdestroyed_a3_0_3x.dds"},
	{0x512CD41735A333DCULL, "6059569_m.fxe"},
	{0x512CD885B7880272ULL, "l_dae380ar_deep_roads.bnk"},
	{0x512D27455C3B0CAEULL, "363706252.ogg"},
	{0x512DAB2B408830AEULL, "zz_isa_debug6133151.cl"},
	{0x512DDAE682E890C5ULL, "gen00fl_varric6076055.cl"},
	{0x512DECD6EDAE0BE5ULL, "258773384.ogg"},
	{0x512DF2D71FE3D027ULL, "335288503.ogg"},
	{0x512DF90B2AF85732ULL, "6127051_m.fxe"},
	{0x512E730145DB96F5ULL, "6166820_m.fxe"},
	{0x512EEBBF65C18EA4ULL, "vxw_encounter2_ki_0_0.vlm"},
	{0x512F11F1EE533C2AULL, "vxw_bonepitdestroyed_a3_1bc_0_0.vlm"},
	{0x512FC11F9654C1A2ULL, "vxlt_0_lowtown_a3_n_c_epc_0_1.vlm"},
	{0x513068DD3A7A939DULL, "vxht_keep_n_6cd_0_0.vlm"},
	{0x5130B7785A95AB70ULL, "mcr371_post_harvester6083673.cl"},
	{0x5130B82CA0D62AEAULL, "vxdr_primevaltomb_7uc_0_1.vlm"},
	{0x5130BA61447DD641ULL, "t1_skn_h01.tnt"},
	{0x5130CEC7FD4826C5ULL, "seb321_leliana6197960.cl"},
	{0x5131A0ACDC7EC549ULL, "6174863_m.fxe"},
	{0x5131D7CFDA52BC41ULL, "vxw_woundedcoast_pp_0_0.vlm"},
	{0x51320CBF29E90991ULL, "w_encounter3_36_0.phy"},
	{0x513222C975A5B152ULL, "hlac_w_sundermount_131076_0y.dds"},
	{0x51322F8C0B02340AULL, "vxdr_0_deeproads_euj_0_0.vlm"},
	{0x51323755691AB3B4ULL, "vxht_mansion2_28g_0_0.vlm"},
	{0x513401D18549AAE5ULL, "w_woundedcoast_184_0.phy"},
	{0x5135F98290EC904EULL, "6104795_m.fxe"},
	{0x513639E885BC84CFULL, "tbe100ip_store_crafted_good.utp"},
	{0x5136515A2BFE33E2ULL, "6127983_m.fxe"},
	{0x5137511708039133ULL, "vxht_bartrand_n_5sn_0_0.vlm"},
	{0x513815BF29ED848DULL, "w_encounter3_36_0.rml"},
	{0x51386FE2EDBA1702ULL, "vxht_mansion2b_1t6_0_0.vlm"},
	{0x51388692A7F1FD67ULL, "vxlt_gallowstemplar_n_4tv_0_0.vlm"},
	{0x5138A38E9F81D51DULL, "vfx0.psh.fxo"},
	{0x5138D8FA10737F70ULL, "hlac_lt_0_lowtown_a2_n_65535_2z.dds"},
	{0x5138ECB009D24D60ULL, "gen_im_trs_krk_hvy_cht_r0.uti"},
	{0x51393E2337B75744ULL, "502339172.ogg"},
	{0x513A9E751CBB032FULL, "6208355_m.fxe"},
	{0x513ABBA17F9782AFULL, "vxlt_0_lowtown_a2_n_gye_0_0.vlm"},
	{0x513AD215BACB37DBULL, "6177819_m.fxe"},
	{0x513C0FE4A2B4AA7AULL, "follower_banter6160617.cl"},
	{0x513C42837E27AD5AULL, "755354818.ogg"},
	{0x513C9C8CDF768614ULL, "6067725_f.fxe"},
	{0x513CAD703838F8E2ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.mmh"},
	{0x513D2B4785AF9CEBULL, "191025391.ogg"},
	{0x513D5710B94B5212ULL, "hlac_lt_gallowscourtyard_n_-65535_2x.dds"},
	{0x513DA8E9FFBC52B2ULL, "6063908_m.fxe"},
	{0x513E5FCFF95499D0ULL, "417605287.ogg"},
	{0x513E73F586E294CAULL, "lgt110_arl_bann_teagan6115555.cl"},
	{0x513EC18AC3E39266ULL, "llac_lt_darktown_-1_57.dds"},
	{0x513EFB7281BC58C6ULL, "6085805_m.fxe"},
	{0x513F991C76F6FE5FULL, "dae000_gamlen6048442.cl"},
	{0x513FD4E8A67F388FULL, "872537166.ogg"},
	{0x5140ABC96F921E78ULL, "94599556.ogg"},
	{0x5140ED166AD20117ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_3y.dds"},
	{0x5141986CC316A4AFULL, "vxht_mansion2b_1xs_0_0.vlm"},
	{0x5141BC28984BBCE5ULL, "clf221_jalen6070342.cl"},
	{0x5142FD581F442AC5ULL, "6041855_m.fxe"},
	{0x514339AFF6A98065ULL, "llac_lt_0_lowtown_docks_a3_65533_0.dds"},
	{0x514370BC1EAEF687ULL, "vxht_0_hightown_a3_c6v_0_0.vlm"},
	{0x514374E8CE51527FULL, "lt_0_lowtown_docks_a3_31_0_lc.mmh"},
	{0x5143D31167D5FF03ULL, "6092585_m.fxe"},
	{0x51441C86C87032F9ULL, "6162554_m.fxe"},
	{0x5144829B7BB4C300ULL, "6139519_m.fxe"},
	{0x5144A3B908DF25B5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comb_0d.dds"},
	{0x5144A40306910965ULL, "zz_dae_debug6029182.cl"},
	{0x5145682A7D178F80ULL, "vxca_caverns1_mines_io0_0_0.vlm"},
	{0x5145F53C64C8A3D5ULL, "6053283_m.fxe"},
	{0x5146B2D44457E33DULL, "vxca_caverns1_mines_hkl_0_0.vlm"},
	{0x514723A39D3F95FDULL, "mer320_hubert6053711.cl"},
	{0x5147A43BF55D7640ULL, "hlac_ht_keep_n_4519_3x.dds"},
	{0x5147A5CA9B186515ULL, "vxdr_primevaltomb_53h_0_1.vlm"},
	{0x514864FE8F42D65EULL, "skin_depth_debug.vsh.fxo"},
	{0x5148B689F9A1BF13ULL, "vxht_mansion2b_327_0_0.vlm"},
	{0x5148BEE0668CA37EULL, "vxdr_primevaltomb_3in_0_0.vlm"},
	{0x51492D7E718D9B2FULL, "vxlt_0_lowtown_a2_n_gvd_0_0.vlm"},
	{0x51493D74B307B93CULL, "403342677.ogg"},
	{0x514A0A2C784D4F30ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dslb_0n.dds"},
	{0x514A4D9760FF9FD9ULL, "547733282.ogg"},
	{0x514A54579EA897A1ULL, "359563673.ogg"},
	{0x514A593B31A9CD5FULL, "vxlt_backalley_g_1l2_0_0.vlm"},
	{0x514A854274C579C0ULL, "zz_suhwan_test.cnv"},
	{0x514B825FBF4FEED0ULL, "6139933_m.fxe"},
	{0x514BCF7949ADBFA4ULL, "vxw_encounter2_qo_0_0.vlm"},
	{0x514C05D98B045E6DULL, "gen00fl_anders6202676.cl"},
	{0x514C28A599E4DA92ULL, "vxlt_warehouse_8e_0_0.vlm"},
	{0x514C6C05A147022BULL, "m_dae140ar_bone_pit.bnk"},
	{0x514C6EF889F533A1ULL, "dae011_corff_sta6163838.cl"},
	{0x514C93B4AACF8CA0ULL, "6001230_m.fxe"},
	{0x514CA657EC354880ULL, "llac_ht_playershouse_n_855_23.dds"},
	{0x514CA73C95778972ULL, "6165343_m.fxe"},
	{0x514CC2295376BC3AULL, "vxw_bonepitmining_a2_26j_0_0.vlm"},
	{0x514D9A9D70CB3A1FULL, "6028165_f.fxe"},
	{0x514DA45A421567CEULL, "vxht_bartrand_n_5hv_0_0.vlm"},
	{0x514E19ADCA572193ULL, "clf260_harley6093819.cl"},
	{0x514E3219D6AC2A18ULL, "443624072.ogg"},
	{0x514E89FA0D6AA80EULL, "6171122_m.fxe"},
	{0x514EA1FA4481DCB1ULL, "lt_0_lowtown_a2_n_44_0_ll.mmh"},
	{0x514ED943A0035B57ULL, "hlac_lt_0_lowtown_n_1_2x.dds"},
	{0x514F22A0D7BEE16EULL, "l_fen221ar_base.wbd"},
	{0x514FB1452DEAFAA7ULL, "6036495_m.fxe"},
	{0x514FE05CE24F9210ULL, "1068731428.ogg"},
	{0x5150FF3360AF0979ULL, "vxlt_0_lowtown_a2_feu_0_0.vlm"},
	{0x51511E9A7C082064ULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.mmh"},
	{0x51516E76E56C2E68ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bookcase_d.dds"},
	{0x51517873575FF4B0ULL, "vxca_caverns1_mines_ihu_0_0.vlm"},
	{0x515184D1482C7C20ULL, "llac_ca_caverns1_mines_24425_5.dds"},
	{0x5151885730E09B0CULL, "vxw_sundermount_34v_0_0.vlm"},
	{0x5151A868A4F8BEBDULL, "dae000_arishok6072758.cl"},
	{0x51520CB1A46693BEULL, "hlac_ht_chantry_n_2534_51z.dds"},
	{0x515246C5C3A2C8F5ULL, "hlac_lt_0_lowtown_docks_a3_-262143_0y.dds"},
	{0x51525D9E1040ACF4ULL, "dae000_porfiria6155286.cl"},
	{0x5152D703A2935CBBULL, "906443474.ogg"},
	{0x515335140FEA64E2ULL, "6129783_m.fxe"},
	{0x5153374CAC0266EDULL, "876952402.ogg"},
	{0x5153923DA9253823ULL, "vxlt_0_lowtown_docks_a3_cxo_0_0.vlm"},
	{0x5153966448248FF9ULL, "mrl310_party_comments6142898.cl"},
	{0x5153A4FBAECA1673ULL, "lgt220pt_bounty_hunter.ncs"},
	{0x5153B5E4DFD1204AULL, "vxlt_undercityencounter_1ls_0_0.vlm"},
	{0x5153DA076F0DA800ULL, "vxca_caverns1_mines_id1_0_0.vlm"},
	{0x51542937439CC206ULL, "vxdr_primevaltomb_3jo_0_0.vlm"},
	{0x515431C887710F8DULL, "vxw_encounter3_fh_0_0.vlm"},
	{0x51544CBF68B70006ULL, "60661243.ogg"},
	{0x515489FFBF8A63E5ULL, "fen110_fenris6097068.cl"},
	{0x5154A439E8F51794ULL, "vxdr_primevalentrance_22d_0_0.vlm"},
	{0x5154EA03960B7607ULL, "character_character_p114.psh.fxo"},
	{0x5155D39054B8180CULL, "393602674.ogg"},
	{0x5156030BDD00BC49ULL, "zz_followers_banter6159621.cl"},
	{0x51565D98BF5080AEULL, "6072318_f.fxe"},
	{0x5156AFBB94AD9FA4ULL, "6041480_f.fxe"},
	{0x5156EF84358B2B82ULL, "mag121cr_idunna_stunt.utc"},
	{0x51573D06DFCCCAADULL, "plt_lgt321pt_alistair_king.nss"},
	{0x515775796F24ECF4ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01.ani"},
	{0x5157839B1D7B59ACULL, "lt_0_lowtown_docks_a3_n_12_0_ll.mmh"},
	{0x51579C8A6C62AA9DULL, "568972633.ogg"},
	{0x5157D69448AFF68EULL, "6072285_m.fxe"},
	{0x515916DC6BAE1D17ULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urvb_0s.dds"},
	{0x5159EFB922359929ULL, "plt_gen00pt_achievement.plo"},
	{0x515AB7FD047B483EULL, "art\\levels\\races\\design\\prp_bodybagc_1.mao"},
	{0x515B16738C8908B4ULL, "6167985_m.fxe"},
	{0x515C0E5246DF2E35ULL, "art\\characters\\heraldry\\hld_city_of_chains.dds"},
	{0x515C5486230E3683ULL, "682022241.ogg"},
	{0x515C575F131F52EDULL, "dae011_corff_sta6164574.cl"},
	{0x515C7ABB51DF19A8ULL, "711796043.ogg"},
	{0x515CD0AEBF3579B4ULL, "vxlt_hangedman_s2_0_0.vlm"},
	{0x515CFE88D6A90BE7ULL, "mag110_arianni6016064.cl"},
	{0x515D4D97F073326EULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtwitchv1.ani"},
	{0x515D72796F295356ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01.gad"},
	{0x515DF2B321CDBFD5ULL, "hlac_lt_backalley_g_65537_2y.dds"},
	{0x515E4D997B9CD859ULL, "vxw_woundedcoast_5w_0_0.vlm"},
	{0x515EC2392ABEC7C9ULL, "dae000_gamlen6145134.cl"},
	{0x515EDA12A6E30984ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_carcass_sa.dds"},
	{0x515F0477837050C9ULL, "196722637.ogg"},
	{0x515F3CE21DC20FA3ULL, "vxw_woundedcoast_19_0_0.vlm"},
	{0x515FFA3422D6B38CULL, "vxw_sundermount_33w_0_0.vlm"},
	{0x5160323B1BF0691BULL, "w_bonepitmining_a2_98_0.mmh"},
	{0x51603E3B1BF07DB8ULL, "w_bonepitmining_a2_98_0.mao"},
	{0x5160503B1BF09C05ULL, "w_bonepitmining_a2_98_0.msh"},
	{0x5160EA48A6920022ULL, "6068039_m.fxe"},
	{0x51613F01584FD82BULL, "t3_shk_b03.tnt"},
	{0x516223DD0B687EEEULL, "530862042.ogg"},
	{0x516224EC2B87AF06ULL, "6150480_m.fxe"},
	{0x5162A38B1BB45060ULL, "6165309_m.fxe"},
	{0x5162A3A57967280DULL, "vxw_encounter3_cg_0_0.vlm"},
	{0x5162DCE827F1E103ULL, "6076121_m.fxe"},
	{0x516304C5D05A51FEULL, "vxht_brothel_n_4ba_0_0.vlm"},
	{0x516478A4E0D99B0DULL, "6154201_m.fxe"},
	{0x51648068E37E7531ULL, "6064146_m.fxe"},
	{0x51648B7EF26DE527ULL, "one100_solivitus6048139.cl"},
	{0x51648BF6A4698A46ULL, "6048815_m.fxe"},
	{0x51648D796F2F9FB5ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01.evt"},
	{0x5164A7FB1E9792B4ULL, "halfroundcombatanims.gda"},
	{0x5164E8134205A1E2ULL, "vxlt_darktown_50m_0_0.vlm"},
	{0x5164EFF36641ADA1ULL, "6067392_m.fxe"},
	{0x5165A00E7C098912ULL, "220142234.ogg"},
	{0x5165AC70385BFB89ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.anb"},
	{0x5165F58ACD4AD72CULL, "vxdr_primevaltomb_7yp_0_0.vlm"},
	{0x5166B75333646AAAULL, "vxdr_primevaltomb_8db_0_0.vlm"},
	{0x5167B3C1CCBF17EDULL, "narrative_varric6172502.cl"},
	{0x5168B06E40B0085DULL, "6033115_m.fxe"},
	{0x5168FCE6487D95C9ULL, "vxw_bonepitmining_a2_8k_0_0.vlm"},
	{0x5169424CF57D6431ULL, "6085428_m.fxe"},
	{0x5169546A62B7423FULL, "257257403.ogg"},
	{0x5169A5E33762ED3AULL, "vxw_bonepitmining_a2_28d_0_0.vlm"},
	{0x5169EC74F398D68BULL, "vxca_caverns1_mines_jg0_0_0.vlm"},
	{0x5169F1B596959356ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_4z.dds"},
	{0x516A35C0C60D03D7ULL, "hlac_ht_0_hightown_n_65536_1y.dds"},
	{0x516A4597F07DAEBFULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtwitchv1.evt"},
	{0x516A9C29C3EF2306ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_nat_p.impa.evt"},
	{0x516AA39D123904F4ULL, "vxlt_gallowstemplar_f_3p2_0_0.vlm"},
	{0x516AB64B11093CB3ULL, "vxht_keep_4b8_0_0.vlm"},
	{0x516AEA8B38B712F4ULL, "dm_vrc2_hugin.mor"},
	{0x516B1BB88C1CE2FCULL, "742085600.ogg"},
	{0x516B50B0B189C348ULL, "6045982_m.fxe"},
	{0x516BCF048C5B8B70ULL, "6101004_m.fxe"},
	{0x516BFF67F88CF243ULL, "vxlt_0_lowtown_a2_n_c_hcz_0_0.vlm"},
	{0x516CB4C747874F64ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.cess.ani"},
	{0x516CC1C2DAF29694ULL, "qcr220st_post_distraction.stg"},
	{0x516CDC3A6BAFCC40ULL, "zz_tbe_debug6196194.cl"},
	{0x516CDFE7E72F11C9ULL, "l_po_deca_f.mao"},
	{0x516CE2E856C29695ULL, "543961795.ogg"},
	{0x516CFA19F9A901FAULL, "6104438_m.fxe"},
	{0x516D0DC538BE30D3ULL, "429377678.ogg"},
	{0x516D2C7F2494898DULL, "lt_0_lowtown_docks_20_0_ll.mmh"},
	{0x516D785B87A5C3FDULL, "vxht_0_hightown_asp_0_0.vlm"},
	{0x516E0F97D8462F66ULL, "vxlt_darktown_2qx_0_1.vlm"},
	{0x516E5737F8A1F686ULL, "920333275.ogg"},
	{0x516E72C2BF284716ULL, "6039309_m.fxe"},
	{0x516E8795A3060CCBULL, "sst_ss_dalish_f_1.gda"},
	{0x516F41134E486282ULL, "hlac_lt_warehouse_n_1_1z.dds"},
	{0x51703D5CF0B8FDE9ULL, "follower_barks6153378.cl"},
	{0x517056BD6881AB01ULL, "hlac_w_bonepitmining_a2_196609_0x.dds"},
	{0x517076152A5A83AEULL, "645263105.ogg"},
	{0x51709413116FC695ULL, "lt_0_lowtown_docks_37_0.rml"},
	{0x5170C74F75758B2FULL, "w_blightlands_193_0.mao"},
	{0x5170D10AF86F8DEBULL, "vxlt_0_lowtown_docks_n_bfh_0_0.vlm"},
	{0x5170D34F75759FCCULL, "w_blightlands_193_0.mmh"},
	{0x5170D94F7575A9BEULL, "w_blightlands_193_0.msh"},
	{0x51712697F08398A8ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtwitchv1.gad"},
	{0x5171283AE075DE1BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junka.mmh"},
	{0x51712ECEDA0179C6ULL, "786486834.ogg"},
	{0x5171343AE075F2B8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junka.mao"},
	{0x5171463AE0761105ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junka.msh"},
	{0x5171B329C3F56951ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_nat_p.impa.gad"},
	{0x5172D96C40045351ULL, "vxht_0_hightown_bww_0_0.vlm"},
	{0x5172EC39C5101522ULL, "vxw_bonepit_29i_0_0.vlm"},
	{0x51732EA846BF881AULL, "mag320_bsc_banter6100760.cl"},
	{0x517359F033FBBA62ULL, "vxlt_darktown_55n_0_0.vlm"},
	{0x5173B1C7478D68C6ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.cess.gad"},
	{0x517404A3AC67A121ULL, "6071943_f.fxe"},
	{0x5174ECB361FA9493ULL, "art\\vfx\\creatures\\fxc_drg_brth_s_c.impa.ani"},
	{0x51752AC6A7E08395ULL, "vxlt_shop_in_0_0.vlm"},
	{0x5176114FF0136033ULL, "plt_cod_bks_kirkwall_guards.plo"},
	{0x5176F6FD3FEF6F19ULL, "gen00fl_aveline6114398.cl"},
	{0x5176F9EE89056A79ULL, "6086667_m.fxe"},
	{0x517768C5F4E5F4EBULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefish_0.pwk"},
	{0x5177B429C3F9D6FFULL, "art\\vfx\\combat\\vfx_m_amg_estorm_nat_p.impa.ani"},
	{0x5178056A321ACEABULL, "gen00fl_aveline6137489.cl"},
	{0x517867A147227024ULL, "6111121_f.fxe"},
	{0x5178AEB5A774583AULL, "gen00fl_aveline6127432.cl"},
	{0x51791B9DF6EF360EULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowa.msh"},
	{0x5179299DF6EF4E5FULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowa.mao"},
	{0x5179359DF6EF62FCULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowa.mmh"},
	{0x5179B4A502C76F14ULL, "6041116_f.fxe"},
	{0x5179D86B061398CEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_0.pwk"},
	{0x517A219A7C2B2A5BULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.anb"},
	{0x517A5FA958671081ULL, "529849488.ogg"},
	{0x517AACC747937EA5ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.cess.evt"},
	{0x517AB216F79F538DULL, "67136686.ogg"},
	{0x517AB6CBB74C9E66ULL, "64292425.ogg"},
	{0x517AE4F7EC56A137ULL, "6103669_m.fxe"},
	{0x517B18A0D7E40A9CULL, "l_fen221ar_base.bnk"},
	{0x517B4DE446DB21B5ULL, "6059915_m.fxe"},
	{0x517B800412169D2BULL, "vxlt_0_lowtown_a3_n_c_cw0_0_0.vlm"},
	{0x517C12BFAA461677ULL, "vxw_woundedcoast_1dk_0_0.vlm"},
	{0x517C75229496A8EDULL, "vrc221_varric6040917.cl"},
	{0x517CA8238C83BDEDULL, "917019819.ogg"},
	{0x517D534F5E6C1E7AULL, "6114417_m.fxe"},
	{0x517D54925D4BC16FULL, "688844632.ogg"},
	{0x517D78779E8216A9ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_act_c.mmh"},
	{0x517DEEBEDB900066ULL, "hlac_lt_gallowstemplar_2382_3z.dds"},
	{0x517DF59F0A878985ULL, "gen00fl_bethany6109227.cl"},
	{0x517DF7414E10430FULL, "gen00fl_merrill6129616.cl"},
	{0x517F42E7EA65A66BULL, "vxlt_0_lowtown_docks_n_bck_0_0.vlm"},
	{0x517F654F27FEA876ULL, "175242201.ogg"},
	{0x518016575C8640E2ULL, "ser210ip_blood_trail.utp"},
	{0x51806B6D2D6B8BD9ULL, "art\\vfx\\textures\\vfx_dot.dds"},
	{0x51809E828E957FD1ULL, "hlac_lt_0_lowtown_docks_n_262142_0y.dds"},
	{0x5180BC9FAFB005DCULL, "ave310_brennan6137004.cl"},
	{0x51810122388AA623ULL, "hlac_lt_0_lowtown_1_7z.dds"},
	{0x51814B4931FA6BD1ULL, "vxht_0_hightown_bpv_0_0.vlm"},
	{0x51815ECB52798BCCULL, "zz_qun_debug6085813.cl"},
	{0x5181624C23FD920DULL, "vxht_keep_6hz_0_0.vlm"},
	{0x5181B17AD34E6C92ULL, "mag100_thrask6032041.cl"},
	{0x5181FBFA95466454ULL, "642198284.ogg"},
	{0x51822514647563D6ULL, "6109434_f.fxe"},
	{0x51828E21E4CDC797ULL, "6187799_m.fxe"},
	{0x5182D77B21FAD514ULL, "vxht_0_hightown_n_a6i_0_0.vlm"},
	{0x5182E4B36206C4CAULL, "art\\vfx\\creatures\\fxc_drg_brth_s_c.impa.evt"},
	{0x51839156C4FAC64BULL, "134832644.ogg"},
	{0x5183A747E20E79E4ULL, "vxlt_0_lowtown_docks_a3_bt1_0_0.vlm"},
	{0x5183B94D2E62C277ULL, "6029019_m.fxe"},
	{0x518403AFE6CE3E3BULL, "vxw_woundedcoast_1ps_0_0.vlm"},
	{0x5184628FB9B9511DULL, "follower_banter6164050.cl"},
	{0x5184C73E77DD710CULL, "6005776_f.fxe"},
	{0x5184CEE777EDD645ULL, "6208472_m.fxe"},
	{0x5185553DE1A92C52ULL, "llac_lt_0_lowtown_a2_n_c_-1_0.dds"},
	{0x5185A85C8F048D15ULL, "6174368_m.fxe"},
	{0x5185ECCCCA977B94ULL, "vxdr_dungeon_2hi_0_1.vlm"},
	{0x518645AC4ADB8801ULL, "6201383_m.fxe"},
	{0x5186A3B960DFD10DULL, "vxht_0_hightown_n_bgp_0_0.vlm"},
	{0x5187360C7761AAAAULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.anb"},
	{0x518788876852F3D3ULL, "isa211_qunari_ambusher2.utc"},
	{0x5187CFD006B4A959ULL, "171887193.ogg"},
	{0x5187D7FB9192F719ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen2.ani"},
	{0x51886BE697223FB4ULL, "287289941.ogg"},
	{0x5188C94ABCE73664ULL, "gen00fl_anders6116110.cl"},
	{0x518999E101BA5908ULL, "18600663.ogg"},
	{0x5189BC23FA520439ULL, "gen00fl_anders6103674.cl"},
	{0x5189BFC138D13A35ULL, "6056815_m.fxe"},
	{0x5189C3B3620CAB3DULL, "art\\vfx\\creatures\\fxc_drg_brth_s_c.impa.gad"},
	{0x5189D4754192E991ULL, "vxlt_0_lowtown_eou_0_1.vlm"},
	{0x518A849C9C3C2EF7ULL, "vxw_woundedcoast_1oj_0_0.vlm"},
	{0x518A9429860A82A3ULL, "6020535_m.fxe"},
	{0x518AA5717581EE83ULL, "cd_dre000_mother.wbd"},
	{0x518AC7F7BEE70913ULL, "122413693.ogg"},
	{0x518B14E9D0753274ULL, "gen00fl_anders6142560.cl"},
	{0x518BC1B2920AE314ULL, "77940832.ogg"},
	{0x518BEDCA5050EA76ULL, "714285047.ogg"},
	{0x518BF2D0F90E421EULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb102a_0n.dds"},
	{0x518C5DEC61024BC5ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_6y.dds"},
	{0x518C953A81B786FEULL, "art\\characters\\creatures\\desiredemon\\c_desiredemon_head.mao"},
	{0x518D20618CA3788BULL, "hlac_lt_0_lowtown_65535_7y.dds"},
	{0x518DB5A2C92CB0B4ULL, "qun220cr_javaris6071862.cl"},
	{0x518E80DB8001AA1FULL, "vxw_encounter2_3v_0_0.vlm"},
	{0x518E902D3858317CULL, "hlac_lt_0_lowtown_docks_n_-131073_3x.dds"},
	{0x518EB2FB9198D75BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen2.gad"},
	{0x51900A710C2C1753ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v2.evt"},
	{0x519018D63F465958ULL, "567992606.ogg"},
	{0x5190715AB6C3EFE1ULL, "696571900.ogg"},
	{0x519096F1566B7C56ULL, "6102534_f.fxe"},
	{0x5190A4DDC590CA74ULL, "vxlt_0_lowtown_docks_n_b0i_0_0.vlm"},
	{0x5190FEE662C77B7BULL, "491489657.ogg"},
	{0x51911574DBE46B0BULL, "810276986.ogg"},
	{0x519131A3F8041AFCULL, "gen00fl_fenris6110910.cl"},
	{0x5191BD6FB991D157ULL, "one100dg_elren6060989.cl"},
	{0x5191E3B6D8CD1ADAULL, "zz_followers_banter6158973.cl"},
	{0x51920343BBB37F4AULL, "open_pressed.cur"},
	{0x519216A4C9247223ULL, "hlac_ht_0_hightown_n_-131071_0y.dds"},
	{0x51929A5AB3F996ADULL, "196888108.ogg"},
	{0x5192D587AB1ADBD4ULL, "145582830.ogg"},
	{0x51935D66DA423E36ULL, "954846550.ogg"},
	{0x5193B412FC723AF6ULL, "290481079.ogg"},
	{0x51941FB3A2017B74ULL, "isa211_isabela6069451.cl"},
	{0x51947B144688AB6DULL, "1069182154.ogg"},
	{0x5194B0BE88708F0FULL, "physics.mao"},
	{0x5194BB4C5461B9C6ULL, "vxw_woundedcoast_kz_0_0.vlm"},
	{0x5194FAA204BA9A6BULL, "6139839_m.fxe"},
	{0x519501E1B2793339ULL, "vxht_mansion2b_24r_0_0.vlm"},
	{0x51950428AF60F83AULL, "zz_approval_debug6003649.cl"},
	{0x51950B2C0771CC6DULL, "docks_male_overseer_mob6176583.cl"},
	{0x51958C0D0A228C9DULL, "dre000_dougal6129818.cl"},
	{0x5195CFFB919F2740ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closedtoopen2.evt"},
	{0x519641090BBDB107ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.anb"},
	{0x5196D8DADC55624FULL, "gen00fl_anders6119514.cl"},
	{0x519703710C322994ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v2.gad"},
	{0x51971242F5B8477DULL, "hlac_lt_0_lowtown_a2_n_c_-65537_3y.dds"},
	{0x5197295B1D3C49DAULL, "478336106.ogg"},
	{0x5197526DEE8B9D56ULL, "6025335_m.fxe"},
	{0x519769D3C8E9D5FDULL, "vxca_cavern1_haunted_dpa_0_0.vlm"},
	{0x5197982B5D30D506ULL, "fen310_fenris6117671.cl"},
	{0x519808EDB39D8F40ULL, "lt_gallowscourtyard_29_0.rml"},
	{0x5198928D0502775AULL, "gen_im_monster_weapon.uti"},
	{0x5198A29752AECA2DULL, "plt_prm_cod_evra_ring.plo"},
	{0x51991260E1AF3A1BULL, "hlac_lt_0_lowtown_docks_a3_196607_3x.dds"},
	{0x5199D58393BB1746ULL, "zz_mag_debug6050911.cl"},
	{0x519A5196ABB7918AULL, "95125545.ogg"},
	{0x519A899F5672666CULL, "vxca_cavern1_haunted_ckf_0_0.vlm"},
	{0x519B2F0203FF66E4ULL, "6028409_f.fxe"},
	{0x519B3C1E4F04D6EDULL, "hlac_ht_keep_4519_17y.dds"},
	{0x519C218D90978DBEULL, "vxlt_0_lowtown_eu9_0_0.vlm"},
	{0x519CCC399832B3FEULL, "vxht_brothel_n_4ve_0_0.vlm"},
	{0x519D3A91FDD132A4ULL, "6066358_m.fxe"},
	{0x519D92F6A35DDAEDULL, "970230321.ogg"},
	{0x519DC39D9FC50565ULL, "6075809_m.fxe"},
	{0x519E12710C386272ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v2.ani"},
	{0x519E55E7AEDB108CULL, "1000914443.ogg"},
	{0x519E602F2616F0ACULL, "621979187.ogg"},
	{0x519EE6C02004165EULL, "vxht_brothel_n_60e_0_0.vlm"},
	{0x519EF0584BB502DAULL, "art\\vfx\\materials\\gui\\vfx_gui_target_ret.mao"},
	{0x519FD513FFE70D92ULL, "261489516.ogg"},
	{0x519FF94C2C8C6342ULL, "vxca_cavern1_haunted_gk5_0_1.vlm"},
	{0x51A03C37C95F4E7EULL, "vxht_keep_n_5xy_0_0.vlm"},
	{0x51A080D9F66CCBA7ULL, "236882817.ogg"},
	{0x51A082FFC1F2EB0BULL, "lowtown_male_urchin_sta6164223.cl"},
	{0x51A0E183B79D81F1ULL, "hla_ht_0_hightown_n_65537_0z.dds"},
	{0x51A11661B84ED998ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsrock.phy"},
	{0x51A1BE5AF77AB4ADULL, "hlac_lt_gallowstemplar_n_2393_1z.dds"},
	{0x51A1F048A8F3CE65ULL, "vxlt_0_lowtown_a2_d4t_0_0.vlm"},
	{0x51A294E96EC760FCULL, "vxht_brothel_n_3s7_0_0.vlm"},
	{0x51A2E70BA78339DAULL, "98027592.ogg"},
	{0x51A2F4717FC66E6DULL, "688928082.ogg"},
	{0x51A327A9398D02C1ULL, "w_woundedcoast_27_0.rml"},
	{0x51A373BEA46F4BB9ULL, "vxht_mansion2b_21q_0_0.vlm"},
	{0x51A375DD094425BAULL, "mrl310_merrill6120110.cl"},
	{0x51A3D5058BE3A8DEULL, "t3_wdg_tr2.tnt"},
	{0x51A58A1D8BDFAA61ULL, "573462349.ogg"},
	{0x51A5EDF3CFDCA3ECULL, "623635166.ogg"},
	{0x51A5F1672046C69CULL, "6015971_f.fxe"},
	{0x51A60BE8D424DBE1ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_beyondpossession_n.dds"},
	{0x51A675779EA516DEULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_act_c.anb"},
	{0x51A707FE3DE6EF9EULL, "203875599.ogg"},
	{0x51A75C3E90E8496AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_loop.gad"},
	{0x51A83384DFC80023ULL, "qun211_braggart6075475.cl"},
	{0x51A834F36ED9EC30ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rock_002_d.dds"},
	{0x51A840E601412074ULL, "952531270.ogg"},
	{0x51A888B51AB5103DULL, "884096248.ogg"},
	{0x51A890B04F0A376DULL, "vxlt_hangedman_n_2ep_0_0.vlm"},
	{0x51A9252BB5ADF0CAULL, "6125325_m.fxe"},
	{0x51A9C78FE01A5746ULL, "llac_lt_0_lowtown_a3_n_c_-65537_0.dds"},
	{0x51A9ED2CD8399B38ULL, "6065324_m.fxe"},
	{0x51AA631197ED335DULL, "6053881_m.fxe"},
	{0x51AA79BEB8475246ULL, "zz_spy130_darktown_res_36055340.cl"},
	{0x51AB649563EDDB1FULL, "vxw_encounter2_9p_0_0.vlm"},
	{0x51AB64AB006C32E2ULL, "vxw_woundedcoast_1r4_0_0.vlm"},
	{0x51AB88058D4256B3ULL, "6129319_m.fxe"},
	{0x51AB8EB708FEFF8AULL, "hlac_w_blightlands_0_3x.dds"},
	{0x51ABD96833CD0537ULL, "pro000_meeran6048807.cl"},
	{0x51ABF48CD3B3FEB2ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.dura.gad"},
	{0x51AC04BDF4E246DBULL, "gen00fl_aveline6139146.cl"},
	{0x51AC26FFBF23B4DAULL, "and330st_base_justice_mid.stg"},
	{0x51AC5349AABD8636ULL, "254379998.ogg"},
	{0x51AD003B41C3B62CULL, "ser120_ghyslain6118388.cl"},
	{0x51AD17AE407816D6ULL, "hlac_lt_0_lowtown_a3_n_c_65537_0y.dds"},
	{0x51AD30F500E3B6AFULL, "6103742_m.fxe"},
	{0x51AD366BE8817C5FULL, "follower_barks6153509.cl"},
	{0x51ADD5F4DABEDE04ULL, "lt_0_lowtown_n_34_0.anb"},
	{0x51AE3F3E90EE3601ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_loop.evt"},
	{0x51AE58BE46B254FCULL, "726021886.ogg"},
	{0x51AEB3A3001D144AULL, "dae000_gamlen6147238.cl"},
	{0x51AEF568F9AE852DULL, "hlac_ht_0_hightown_n_65536_0x.dds"},
	{0x51AF3600C42A401FULL, "983262380.ogg"},
	{0x51AF3AFE8C31041DULL, "lgt110cr_gxa_coa_spy.utc"},
	{0x51AF977175A0F331ULL, "cd_dre000_mother.bnk"},
	{0x51AFAD02E097A286ULL, "mrl310_merrill6120011.cl"},
	{0x51B019DF1E0250E5ULL, "vxht_mansion2_23g_0_0.vlm"},
	{0x51B027564D1AF3C0ULL, "vxht_0_hightown_a3_bye_0_0.vlm"},
	{0x51B02C06E018D6BEULL, "plt_lgt321pt_alistair_king.plo"},
	{0x51B0390C7784B49DULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.mmh"},
	{0x51B0D39F939AA074ULL, "140314517.ogg"},
	{0x51B10A52B9C11664ULL, "hlac_w_sundermount_-327682_0x.dds"},
	{0x51B11960DD0B8880ULL, "59216719.ogg"},
	{0x51B19CD06B1E4981ULL, "41260587.ogg"},
	{0x51B235689AD1A3D3ULL, "480458976.ogg"},
	{0x51B24ACE1004ECEFULL, "qun211_party_comment6182711.cl"},
	{0x51B267DB91D070F4ULL, "vxca_caverns2_958_0_0.vlm"},
	{0x51B2A0D736B0E7A6ULL, "follower_banter6170469.cl"},
	{0x51B2BBDEAD52D3D4ULL, "6166999_m.fxe"},
	{0x51B2DC4B04555DFAULL, "264855237.ogg"},
	{0x51B2F78CD3BA22E9ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.dura.evt"},
	{0x51B3413E8BF55102ULL, "6106919_m.fxe"},
	{0x51B3A8FAE2675CE4ULL, "ave200_donnic6110619.cl"},
	{0x51B3DF34C8514ADEULL, "zz_tannertest.cut"},
	{0x51B4A6C3C3FDB5A3ULL, "gen_im_trs_apo_lgt_boo_r0.uti"},
	{0x51B5DA33721E7877ULL, "vxw_woundedcoast_1pg_0_0.vlm"},
	{0x51B6ADBDCC14EEACULL, "6174545_m.fxe"},
	{0x51B6C57262D8D363ULL, "6136005_m.fxe"},
	{0x51B6F9E6410A0788ULL, "vxlt_house_n_0_0.vlm"},
	{0x51B800557EB03B74ULL, "1008748488.ogg"},
	{0x51B8141FB9795511ULL, "6036448_m.fxe"},
	{0x51B8493AE0B21950ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junka.phy"},
	{0x51B87D9D63093B33ULL, "768260510.ogg"},
	{0x51B981490372A55AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_chest_out.gad"},
	{0x51B9A029FFE37C11ULL, "vxlt_gallowstemplar_f_46v_0_0.vlm"},
	{0x51BA0FC96F38FB76ULL, "vxlt_foundry_n_ym_0_0.vlm"},
	{0x51BA6A5DB771C2F0ULL, "mag370_meredith6044367.cl"},
	{0x51BA912DA15090EAULL, "mcr310_party_banter6087803.cl"},
	{0x51BA96F45B213244ULL, "gen00fl_carver6103686.cl"},
	{0x51BADC2350AF2DADULL, "gen00fl_isabela6106402.cl"},
	{0x51BB3481DBF94FF9ULL, "6162745_m.fxe"},
	{0x51BBF23707F44E97ULL, "6048238_m.fxe"},
	{0x51BC373E90FA6650ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_loop.ani"},
	{0x51BC53AE095DFC0FULL, "915901444.ogg"},
	{0x51BC84C85A2FB959ULL, "vxca_caverns1_e8q_0_0.vlm"},
	{0x51BC94B996DA0A9FULL, "957199069.ogg"},
	{0x51BCB9EB4A63DADFULL, "303215951.ogg"},
	{0x51BCF66964B49E4DULL, "lt_0_lowtown_docks_n_163_0.mmh"},
	{0x51BCFA6964B4A49EULL, "lt_0_lowtown_docks_n_163_0.mao"},
	{0x51BD086964B4BCD3ULL, "lt_0_lowtown_docks_n_163_0.msh"},
	{0x51BD10935E078B5CULL, "287051757.ogg"},
	{0x51BD4A143FD11386ULL, "6046840_m.fxe"},
	{0x51BD5EC4B5887150ULL, "6072312_f.fxe"},
	{0x51BD728A072118BDULL, "zz_vault_debug6203366.cl"},
	{0x51BDCE9E6316C86DULL, "474542328.ogg"},
	{0x51BDDA1E77F8018EULL, "art\\vfx\\combat\\vfx_c_rw_teleport_out_c.impa.ani"},
	{0x51BEE34B68FDB410ULL, "51170790.ogg"},
	{0x51BEF4895D00A8C7ULL, "ss_raider_f_1.wbd"},
	{0x51BEF9095F6E3CDBULL, "69536134.ogg"},
	{0x51BF3E090BE0B108ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.mmh"},
	{0x51BF46003FFF9CE3ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardleft.ani"},
	{0x51BF87DCAA0E89BBULL, "vxlt_darktown_5c5_0_0.vlm"},
	{0x51BFB628C0741244ULL, "mrl000_marethari6129197.cl"},
	{0x51BFD50FF9E77CF1ULL, "mag111st_cassandra2.stg"},
	{0x51BFDF8CD3C48458ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.dura.ani"},
	{0x51BFFEC0D68CF0EFULL, "6013332_f.fxe"},
	{0x51C079C738F46CE2ULL, "808662233.ogg"},
	{0x51C084490378C971ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_chest_out.evt"},
	{0x51C08641A5DC9A76ULL, "hlac_lt_0_lowtown_a3_n_c_-1_5x.dds"},
	{0x51C0D5D4BA10BD4BULL, "llac_lt_0_lowtown_n_65537_5.dds"},
	{0x51C0DB7EAE17889DULL, "vxht_keep_n_61z_0_0.vlm"},
	{0x51C13D9AE5D7CAB8ULL, "vxlt_0_lowtown_a2_h23_0_0.vlm"},
	{0x51C1BD576E6D8BC6ULL, "hlac_w_encounter2_camp_131077_0y.dds"},
	{0x51C1C5C73C8A2F71ULL, "vxw_sundermount_30z_0_1.vlm"},
	{0x51C1C76C3C8151E6ULL, "hlac_ht_brothel_n_2995_0x.dds"},
	{0x51C226C949861296ULL, "608427191.ogg"},
	{0x51C2816FA1AA90E6ULL, "vxht_bartrand_n_64i_0_0.vlm"},
	{0x51C2D04DA586370CULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnl.ani"},
	{0x51C30046ED055C67ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\mt_hed_masa_0d.dds"},
	{0x51C3037CF7648FC2ULL, "hlac_ht_0_hightown_-65535_0x.dds"},
	{0x51C31E8BF0865DB5ULL, "830397548.ogg"},
	{0x51C3EA1294EAC3E2ULL, "368787177.ogg"},
	{0x51C3F7D5D0031696ULL, "follower_banter6164499.cl"},
	{0x51C3F849DB1AAEE9ULL, "gen00fl_merrill6129607.cl"},
	{0x51C401095A077361ULL, "hlac_lt_warehouse_n_65536_2z.dds"},
	{0x51C47827975ED588ULL, "vxlt_undercityserialkiller_1i2_0_0.vlm"},
	{0x51C5374F54CEEBE3ULL, "6131105_m.fxe"},
	{0x51C5408BE3A13757ULL, "197362191.ogg"},
	{0x51C541435AE688BFULL, "vxca_cavern1_haunted_cki_0_0.vlm"},
	{0x51C5D3D16D68F022ULL, "333693702.ogg"},
	{0x51C63BE6DA93F7D4ULL, "llac_lt_gallowstemplar_n_2391_0.dds"},
	{0x51C698E97C7FD7F4ULL, "gen00fl_anders6125448.cl"},
	{0x51C737C0435E824FULL, "vxlt_0_lowtown_a2_n_eqs_0_0.vlm"},
	{0x51C7913AE0FE5530ULL, "lgt220_shp_mag_guard6149867.cl"},
	{0x51C7E5DEE0661189ULL, "fen300pt_alone.ncs"},
	{0x51C87C8C7E9BABB1ULL, "218953748.ogg"},
	{0x51C881A6612F13F6ULL, "vxlt_foundry_n_tn_0_0.vlm"},
	{0x51C8BFF327A53B50ULL, "6176746_f.fxe"},
	{0x51C8D8F1BC2C7F8FULL, "6111411_m.fxe"},
	{0x51C96D2676D7F734ULL, "hlac_lt_0_lowtown_docks_-196609_0z.dds"},
	{0x51C9AA4F890A29DAULL, "mag210_arianni6030462.cl"},
	{0x51C9DD4DA58C6C4EULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnl.gad"},
	{0x51C9E37DB67A8AB5ULL, "genip_exploding_crate_sm_wt.utp"},
	{0x51C9FCB703EB3D2BULL, "433657289.ogg"},
	{0x51CA0A19AA1C25B7ULL, "854472683.ogg"},
	{0x51CA5BCEE805AF83ULL, "gen00fl_anders6103913.cl"},
	{0x51CA6413F9EA6717ULL, "6043338_m.fxe"},
	{0x51CAD21E78027EDFULL, "art\\vfx\\combat\\vfx_c_rw_teleport_out_c.impa.evt"},
	{0x51CAFEFE374DDE63ULL, "vxlt_0_lowtown_docks_a3_d08_0_0.vlm"},
	{0x51CC300281661FD3ULL, "vxlt_gallowstemplar_n_2rv_0_0.vlm"},
	{0x51CC5B7632521C46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_loop_v1.evt"},
	{0x51CC5E00400A4FDAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardleft.evt"},
	{0x51CC9E6BE8407E7BULL, "lt_0_lowtown_docks_a3_12_0_ll.mmh"},
	{0x51CD05CB25DF2F9CULL, "6107276_f.fxe"},
	{0x51CD5614C5A01BA9ULL, "vxdr_0_deeproads_fib_0_0.vlm"},
	{0x51CD9C4903837C40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_chest_out.ani"},
	{0x51CDAAA7F3AEECA6ULL, "vxw_encounter2_ov_0_1.vlm"},
	{0x51CDDE2CEC67506BULL, "vxht_bartrand_n_345_0_0.vlm"},
	{0x51CE09575CC84BBFULL, "ser210ip_blood_trail.ncs"},
	{0x51CE3E0B51409451ULL, "6207702_f.fxe"},
	{0x51CF242D75790E4AULL, "hlac_w_blightlands_1_0z.dds"},
	{0x51D0632D23C102AFULL, "art\\vfx\\combat\\vfx_r_shd_decoy_dis_aoe_p.impa.gad"},
	{0x51D0D84DA592822DULL, "art\\characters\\creatures\\stonegolem\\animation\\rn_lnl.evt"},
	{0x51D12F0BDDAD712AULL, "vxht_0_hightown_n_c50_0_0.vlm"},
	{0x51D1B31E780868C8ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_out_c.impa.gad"},
	{0x51D1B440AD3B72EFULL, "gen00fl_bethany6097896.cl"},
	{0x51D1F4CC5CB37856ULL, "6174910_m.fxe"},
	{0x51D23E7479330C0FULL, "vxca_caverns1_mines_e7g_0_0.vlm"},
	{0x51D278CDA3F1C86DULL, "vxht_bartrand_n_47u_0_0.vlm"},
	{0x51D292935C7C162BULL, "gen00fl_anders6114489.cl"},
	{0x51D29AFFF66EC378ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_archesthin.phy"},
	{0x51D2EA048954EE08ULL, "vxlt_undercityserialkiller_1n3_0_0.vlm"},
	{0x51D2FD90F597CE17ULL, "51562383.ogg"},
	{0x51D32F7B57398712ULL, "vxht_keep_n_67s_0_0.vlm"},
	{0x51D35D0040106CCDULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardleft.gad"},
	{0x51D3727632586191ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_loop_v1.gad"},
	{0x51D3D97A1487CF41ULL, "vxw_bonepit_13w_0_0.vlm"},
	{0x51D40BC7C4B5FD0BULL, "485704793.ogg"},
	{0x51D42265E3007EDCULL, "vxht_mansion2b_33s_0_0.vlm"},
	{0x51D4EAFC25FF90C7ULL, "zz_and_debug6134061.cl"},
	{0x51D4EB92EBDF4BD5ULL, "vxlt_gallowstemplar_f_3sz_0_0.vlm"},
	{0x51D5613DFBB0457CULL, "6129051_m.fxe"},
	{0x51D606A939B82451ULL, "w_woundedcoast_27_0.anb"},
	{0x51D64EE273ADDEC9ULL, "llac_dr_dungeon_3216_12.dds"},
	{0x51D69220EB333C3FULL, "971380182.ogg"},
	{0x51D6BAC996B864E0ULL, "6094997_m.fxe"},
	{0x51D6BF5551E5731EULL, "one100dg_elren6058888.cl"},
	{0x51D7118B4C9709C8ULL, "49361308.ogg"},
	{0x51D7782D23C7455CULL, "art\\vfx\\combat\\vfx_r_shd_decoy_dis_aoe_p.impa.evt"},
	{0x51D7C8646EAF961AULL, "ico_potion_talent.dds"},
	{0x51D7E8D69C41B3C1ULL, "lt_gamlenshouse_n_3567.anb"},
	{0x51D82B0ECD74604EULL, "742411683.ogg"},
	{0x51D8582AB90B2D77ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1a.mmh"},
	{0x51D85905FC22CEB2ULL, "vrc321ip_spooky5_obj_8.utp"},
	{0x51D85C2AB90B3384ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1a.mao"},
	{0x51D86E2AB90B52E9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1a.msh"},
	{0x51D8AB2980E29379ULL, "vxw_encounter1_dp_0_0.vlm"},
	{0x51D8B660B7ED6FF9ULL, "678500540.ogg"},
	{0x51D90F362267A75DULL, "6085993_f.fxe"},
	{0x51D93B2B24ABDFA2ULL, "mer320_hubert6052982.cl"},
	{0x51D9A3849A1C4C5CULL, "one100dg_vanard6048227.cl"},
	{0x51DA7376325E823FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_loop_v1.ani"},
	{0x51DACCC0B84E4B0FULL, "llac_lt_gallowstemplar_2384_0.dds"},
	{0x51DB91A251CA9B14ULL, "6135160_m.fxe"},
	{0x51DBCB16350EA520ULL, "419339986.ogg"},
	{0x51DC10ABAA75A6A1ULL, "6133765_m.fxe"},
	{0x51DC43E3E9C03D67ULL, "vxlt_0_lowtown_a2_e3b_0_0.vlm"},
	{0x51DC5009DE5D68EBULL, "vxht_bartrand_n_334_0_0.vlm"},
	{0x51DD720E9CEBBE88ULL, "6076328_m.fxe"},
	{0x51DD74876293CF20ULL, "mag360_alain6063782.cl"},
	{0x51DD7944010B33CFULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_c.dura.gad"},
	{0x51DD8B1E6301EC8FULL, "6121603_m.fxe"},
	{0x51DD8BAA59199624ULL, "hlac_w_encounter2_camp_131074_0z.dds"},
	{0x51DDFF6105CD89C6ULL, "487721359.ogg"},
	{0x51DE0821B0D26556ULL, "6118989_m.fxe"},
	{0x51DE1B102877EE04ULL, "hlac_lt_0_lowtown_a3_n_c_-3_3z.dds"},
	{0x51DECB50A0706535ULL, "hlac_lt_hangedman_-1_17x.dds"},
	{0x51DF021CA4A7998CULL, "95273350.ogg"},
	{0x51DF2E2B31DBE241ULL, "vxw_woundedcoast_4l_0_0.vlm"},
	{0x51DF80DE4C35CEADULL, "vxw_sundermount_343_0_0.vlm"},
	{0x51DFC762734536FAULL, "6106663_m.fxe"},
	{0x51E0B5282B2CE54EULL, "mag121_idunna6084205.cl"},
	{0x51E116CF1929BB77ULL, "6118074_m.fxe"},
	{0x51E180AC7FA60E03ULL, "isa311_castillon6092134.cl"},
	{0x51E1F563AF352341ULL, "vxlt_0_lowtown_docks_a3_cx5_0_0.vlm"},
	{0x51E2AA388C42E194ULL, "6115349_f.fxe"},
	{0x51E303D8C28FD249ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_upperblank_d.dds"},
	{0x51E3C2EEB4FE9632ULL, "follower_banter6141511.cl"},
	{0x51E4078EC2A81B94ULL, "km_scion_1.mor"},
	{0x51E419A939C481B0ULL, "w_woundedcoast_27_0.msh"},
	{0x51E427A939C499CAULL, "w_woundedcoast_27_0.mmh"},
	{0x51E42BA939C4A001ULL, "w_woundedcoast_27_0.mao"},
	{0x51E48E440111767CULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_c.dura.evt"},
	{0x51E49F956BD2DDA8ULL, "687755114.ogg"},
	{0x51E4AF6063074D4AULL, "vxht_bartrand_n_5z8_0_0.vlm"},
	{0x51E4FFEA9313A8E9ULL, "185047624.ogg"},
	{0x51E5802D23D39075ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_dis_aoe_p.impa.ani"},
	{0x51E58DF65ADC9865ULL, "vxlt_gallowscourtyard_711_0_0.vlm"},
	{0x51E5E0D0A3F0DB65ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hgrass_da.dds"},
	{0x51E638E7CA9DE99EULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml101a_0n.dds"},
	{0x51E684B85CF8DF1AULL, "vxht_mansion2_1m9_0_0.vlm"},
	{0x51E71D0672D8ABF9ULL, "vxw_encounter1_iw_0_0.vlm"},
	{0x51E78B2BE004FA0CULL, "pro000st_merchant.stg"},
	{0x51E7C857A24305DEULL, "211446493.ogg"},
	{0x51E80D9610BF51CCULL, "vxw_bonepitmining_a2_5e_0_0.vlm"},
	{0x51E8481F6F9F6D9EULL, "vxlt_0_lowtown_docks_n_9zr_0_0.vlm"},
	{0x51E868362949189CULL, "569170767.ogg"},
	{0x51E8E7F220455CE3ULL, "6134523_f.fxe"},
	{0x51E9A7F76716FF7CULL, "vxht_bartrand_n_6ri_0_0.vlm"},
	{0x51E9F9D648B87161ULL, "dae000_loyal_mage6165242.cl"},
	{0x51EACB59EA346497ULL, "6119396_m.fxe"},
	{0x51EADD737FF0D33BULL, "vxlt_darktown_5r6_0_0.vlm"},
	{0x51EB46C897F3676AULL, "219892078.ogg"},
	{0x51EB4CF090E2C0B1ULL, "hlac_w_sundermount_327682_0x.dds"},
	{0x51EB7717F1438581ULL, "vxca_caverns1_mines_ia3_0_0.vlm"},
	{0x51EB8AA19CF99731ULL, "lt_0_lowtown_docks_a3_n_20_0_ll.mmh"},
	{0x51EBFE0FAE7C1C48ULL, "91444880.ogg"},
	{0x51EC159494BA87E3ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_7x.dds"},
	{0x51EC52D9210412C0ULL, "6170152_m.fxe"},
	{0x51EC86FE21223AA6ULL, "6069414_m.fxe"},
	{0x51EC8A4AB0FA2C2FULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl90pause.gad"},
	{0x51ECD2943EEC9DD8ULL, "988636824.ogg"},
	{0x51ECD73E2D919889ULL, "6118726_m.fxe"},
	{0x51ED10C17BBF4C83ULL, "6147768_m.fxe"},
	{0x51ED814D46CC569FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s2_c.impa.gad"},
	{0x51EDA00823D1FAC1ULL, "vxw_woundedcoast_1c_0_0.vlm"},
	{0x51EDB794477EE4A2ULL, "186707089.ogg"},
	{0x51EDF2BB3E2BE72DULL, "vxw_sundermount_332_0_0.vlm"},
	{0x51EE032E791396B6ULL, "gen00fl_aveline6121913.cl"},
	{0x51EE20121F35B86EULL, "474706750.ogg"},
	{0x51EE2FA62E26F267ULL, "vxht_keep_6fr_0_0.vlm"},
	{0x51EE8224A267BEB0ULL, "vfx_lightning_f_07.dds"},
	{0x51EEC95496A11649ULL, "6048437_f.fxe"},
	{0x51EEF4F4DAF6AA64ULL, "lt_0_lowtown_n_34_0.rml"},
	{0x51EF388565F653D2ULL, "lgt000im_ultra_pommel.uti"},
	{0x51EF4924EEB23ED5ULL, "6176098_m.fxe"},
	{0x51EFC530D4ABE51BULL, "hlac_w_sundermount_-262143_0y.dds"},
	{0x51F0179D7129AE61ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.anb"},
	{0x51F03040583E0DE6ULL, "mag110cr_sailor_5.utc"},
	{0x51F078B41FBA4CDBULL, "189954867.ogg"},
	{0x51F0DB6DFECE01B6ULL, "vxht_mansion2_28u_0_0.vlm"},
	{0x51F105ECF0373D63ULL, "one110cr_tree_minder.utc"},
	{0x51F163FA6983AAEBULL, "fx_tree.dds"},
	{0x51F19644011C0D95ULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_c.dura.ani"},
	{0x51F1B12C3974C82EULL, "vxht_0_hightown_n_c90_0_0.vlm"},
	{0x51F25ED0D1B8A182ULL, "vxlt_gallowstemplar_55r_0_0.vlm"},
	{0x51F31F8EE6C03ED8ULL, "hlac_lt_backalley_g_1_2y.dds"},
	{0x51F335463ED9BF1BULL, "6185498_m.fxe"},
	{0x51F36911E7D54A7FULL, "w_bonepitmining_a2_128_0.msh"},
	{0x51F36E84C9ED2B99ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecart_1.mao"},
	{0x51F36F11E7D55421ULL, "w_bonepitmining_a2_128_0.mmh"},
	{0x51F375C7EBC69732ULL, "hlac_lt_0_lowtown_a2_1_6z.dds"},
	{0x51F37B11E7D568C2ULL, "w_bonepitmining_a2_128_0.mao"},
	{0x51F39B56057DEBF6ULL, "hlac_lt_0_lowtown_a3_n_c_2_0y.dds"},
	{0x51F3A3EAC42FD32EULL, "806606152.ogg"},
	{0x51F40B6EAEFB1523ULL, "follower_banter6136730.cl"},
	{0x51F454A442F61E76ULL, "art\\levels\\races\\design\\blendtree\\beartrap.anb"},
	{0x51F460F1663BC76FULL, "qun221_viscount6069924.cl"},
	{0x51F46CCC9574C3CAULL, "art\\characters\\creatures\\dragon\\animation\\new25.gad"},
	{0x51F4764D46D2628CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s2_c.impa.evt"},
	{0x51F6A58ACAB30E5AULL, "sse_ss_harvester_1.gda"},
	{0x51F6EBBF2627DC2EULL, "art\\vfx\\cinematic\\vfx_anders_healing_c.mmh"},
	{0x51F7249C8CB82611ULL, "hlac_lt_0_lowtown_docks_n_131072_0z.dds"},
	{0x51F73346A406B098ULL, "mag291ip_to_caress.utp"},
	{0x51F81368F7E07774ULL, "652971905.ogg"},
	{0x51F832490E6F58E1ULL, "91694648.ogg"},
	{0x51F84D3167A10753ULL, "vxlt_undercityencounter_13r_0_1.vlm"},
	{0x51F8560BEC5C5A77ULL, "943984815.ogg"},
	{0x51F86F79DDC5F7B5ULL, "23321554.ogg"},
	{0x51F88ACC1C03602BULL, "112952872.ogg"},
	{0x51F8F4658E61D7A1ULL, "6139165_m.fxe"},
	{0x51F8FCB295AFD30BULL, "twopassblurbloom5.psh.fxo"},
	{0x51F950C9B1DB478FULL, "vxdr_dungeon_1lp_0_0.vlm"},
	{0x51FA3B2B529D6D61ULL, "fen210_orana6109687.cl"},
	{0x51FA81928D3000AEULL, "vxdr_primevaltomb_50n_0_1.vlm"},
	{0x51FA8357E241BB6EULL, "6138795_f.fxe"},
	{0x51FB491B1F87B109ULL, "6117626_m.fxe"},
	{0x51FB66897663976FULL, "979945994.ogg"},
	{0x51FBC0F9A35E6FC1ULL, "794247455.ogg"},
	{0x51FBE2E407E03A26ULL, "mag371_elsa6064666.cl"},
	{0x51FCA183201D0AE7ULL, "vxht_keep_6cm_0_0.vlm"},
	{0x51FCAA417DE6B8A2ULL, "241219438.ogg"},
	{0x51FCF89D713403DAULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.mmh"},
	{0x51FD0361B89D182DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsrock.msh"},
	{0x51FD1161B89D3060ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsrock.mao"},
	{0x51FD1561B89D36B3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsrock.mmh"},
	{0x51FD90ED5BD7FA7CULL, "57357146.ogg"},
	{0x51FDA52F5F6CA4C7ULL, "6125294_m.fxe"},
	{0x51FDF88BBB1BA512ULL, "darktown_male_sickly_sta6175307.cl"},
	{0x51FE350D17FB7D25ULL, "vxlt_foundry_n_v3_0_0.vlm"},
	{0x51FE410B8FEC038FULL, "dae300_cullen6135869.cl"},
	{0x51FE7C5B6CD48338ULL, "w_encounter3_19_0.rml"},
	{0x51FEFE895D369DFDULL, "ss_raider_f_1.bnk"},
	{0x51FF3E9E2267AF6EULL, "523523604.ogg"},
	{0x51FF59751AD685EDULL, "6050336_m.fxe"},
	{0x51FF8C9BD1A9E940ULL, "pro000_merchant6051598.cl"},
	{0x51FFC46E9A51CBD0ULL, "208445737.ogg"},
	{0x5200600485FAA324ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\basestaff\\blendtree\\devastator_overlay.anb"},
	{0x5200889BBF8DC6E4ULL, "31870755.ogg"},
	{0x520114EBBE8A77ECULL, "vxlt_0_lowtown_a2_n_c_e85_0_0.vlm"},
	{0x52011A39C8503FF0ULL, "634312294.ogg"},
	{0x52014911DA692EA7ULL, "6125346_m.fxe"},
	{0x5201A74AB10CB9F5ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnl90pause.ani"},
	{0x5201CB0851CFE44AULL, "hlac_ht_keep_4519_6z.dds"},
	{0x5201CCDA4ADCD073ULL, "482763951.ogg"},
	{0x5201E58BEC60F031ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_hatch.msh"},
	{0x5201E86E3B1DA5ADULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medc_0d.dds"},
	{0x5201F38BEC61084CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_hatch.mao"},
	{0x5201FF8BEC611CEFULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_hatch.mmh"},
	{0x52025E4D46DE77E5ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s2_c.impa.ani"},
	{0x5202616447F79803ULL, "llac_ht_mansion2b_818_49.dds"},
	{0x52027BB6919EBE21ULL, "hlac_ht_chantry_n_2534_38x.dds"},
	{0x5203ED5368BA3CAAULL, "vxw_woundedcoast_nu_0_0.vlm"},
	{0x52048CAB9A69E560ULL, "6144344_m.fxe"},
	{0x520505216FCF039AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\ledgedropland.gad"},
	{0x52050ABF26344E2DULL, "art\\vfx\\cinematic\\vfx_anders_healing_c.anb"},
	{0x520568F51C4EA680ULL, "vxlt_0_lowtown_a2_n_ei8_0_0.vlm"},
	{0x520633E22FA2D0B5ULL, "6037601_m.fxe"},
	{0x5206934DFA23A33CULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.cess.ani"},
	{0x52071BDAE48860A3ULL, "250101244.ogg"},
	{0x52073A38A91EBDE3ULL, "wm_ico_genericpin.dds"},
	{0x52077C898F56FD36ULL, "hlac_lt_backalley_g_0_4z.dds"},
	{0x5207A526F8DF2D37ULL, "6179973_f.fxe"},
	{0x5207A8DE532A7D39ULL, "411999110.ogg"},
	{0x5207EB0DBD7132A4ULL, "vxw_blightlands_bm_0_0.vlm"},
	{0x5208440BDA0D2DF5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbar.msh"},
	{0x5208460BDA0D316BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbar.mmh"},
	{0x5208520BDA0D4508ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbar.mao"},
	{0x5208C7A8E3D33223ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\diev1.ani"},
	{0x5208F84EB9208930ULL, "vxlt_0_lowtown_docks_n_azt_0_0.vlm"},
	{0x520947CC9586E0B0ULL, "art\\characters\\creatures\\dragon\\animation\\new25.ani"},
	{0x520A36C5F7D359F8ULL, "vxht_0_hightown_9bj_0_0.vlm"},
	{0x520ADAB568D1B96AULL, "plt_seb000pt_midstage.plo"},
	{0x520AFD4A330D30D9ULL, "4180271.ogg"},
	{0x520B136012132367ULL, "vxht_keep_6ll_0_0.vlm"},
	{0x520B546F8CBF9089ULL, "hlac_lt_undercityencounter_-1_42y.dds"},
	{0x520B675FB8B48346ULL, "vxw_bonepit_1ox_0_0.vlm"},
	{0x520B9330C8ACFD5BULL, "6133046_m.fxe"},
	{0x520B96F0F37C338CULL, "w_bonepitdestroyed_a3_35_0.msh"},
	{0x520B9CF0F37C3E7EULL, "w_bonepitdestroyed_a3_35_0.mmh"},
	{0x520BA8F0F37C5215ULL, "w_bonepitdestroyed_a3_35_0.mao"},
	{0x520C08216FD527B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\ledgedropland.evt"},
	{0x520CB942FAFC531EULL, "6044625_f.fxe"},
	{0x520CFA75490A3708ULL, "6160658_m.fxe"},
	{0x520DA04DFA29D87EULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.cess.gad"},
	{0x520E3ABBAD55C12AULL, "551936137.ogg"},
	{0x520EFB0FAEF16407ULL, "atl_loginnewac_dxt5_dat.xml"},
	{0x520F8F83D40320D5ULL, "art\\vfx\\textures\\vfx_distor_fake.dds"},
	{0x521037AC9DE56330ULL, "228712705.ogg"},
	{0x52107CFA699E85CDULL, "fx_tree.mao"},
	{0x5211C9D69C72F751ULL, "lt_gamlenshouse_n_3567.rml"},
	{0x5211DEB25B462308ULL, "hlac_lt_gallowstemplar_2383_4z.dds"},
	{0x5211ECC18F6C21C7ULL, "vxlt_0_lowtown_a2_n_c_htc_0_0.vlm"},
	{0x5211F4A1D187A827ULL, "gen00fl_sebastian6174932.cl"},
	{0x52123E6E998EC7E2ULL, "vxlt_darktown_5lq_0_0.vlm"},
	{0x52134624442E2F44ULL, "392687276.ogg"},
	{0x52136C036496285EULL, "llac_w_bonepit_393218_0.dds"},
	{0x52138ED5E9632908ULL, "342427763.ogg"},
	{0x5213CE3F96179DDBULL, "709264544.ogg"},
	{0x5213DAD20E44BF00ULL, "vxlt_0_lowtown_a2_n_er9_0_0.vlm"},
	{0x5213F70E90C9DDBAULL, "llac_dr_slavers_3207_8.dds"},
	{0x5214692CA00AC626ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\bowoverlay.anb"},
	{0x5214791EB3C0F762ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mase.phy"},
	{0x52147B4DFA2FB79DULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.cess.evt"},
	{0x5214F57C05D7A057ULL, "130928622.ogg"},
	{0x521529C9FB85BB7AULL, "genip_blocker_fire.ncs"},
	{0x5215DF3C5CEB9DFDULL, "vxdr_primevaltomb_73w_0_0.vlm"},
	{0x52162BD1791EC1DDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw5sm.mao"},
	{0x52162FD1791EC816ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw5sm.mmh"},
	{0x521639D1791ED914ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw5sm.msh"},
	{0x5216A57ED848106CULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm06.phy"},
	{0x5216BF6988F1DEE3ULL, "421424639.ogg"},
	{0x5216DFA8E3DF981AULL, "art\\characters\\creatures\\shriek\\animation\\combat\\diev1.evt"},
	{0x5217165AD47CC60DULL, "vxht_0_hightown_n_b5f_0_0.vlm"},
	{0x5217EB5BB6866E94ULL, "232325382.ogg"},
	{0x52189570B196DB37ULL, "vxdr_primevalentrance_2l7_0_2.vlm"},
	{0x5219A46A7059DC57ULL, "723736513.ogg"},
	{0x5219D0155396C5C3ULL, "llac_lt_backalley_n_0_8.dds"},
	{0x521A20216FE18D80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\ledgedropland.ani"},
	{0x521A27BD98D9ADD7ULL, "6126512_m.fxe"},
	{0x521A721CC10B94E2ULL, "627456020.ogg"},
	{0x521A8DE29CA37EEEULL, "hlac_w_blightlands_65539_1z.dds"},
	{0x521A9764B5E33863ULL, "6021275_m.fxe"},
	{0x521AB887C9377FD2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90runf.gad"},
	{0x521ACF92C32F0EDAULL, "vxht_0_hightown_a3_av5_0_0.vlm"},
	{0x521B193BE9D9D8B6ULL, "llac_ht_mansion2b_818_17.dds"},
	{0x521BFC0EF413F853ULL, "hlac_lt_gallowscourtyard_n_-65538_4x.dds"},
	{0x521C384DAE2D4BF9ULL, "art\\cinematics\\act1\\cs_fp_yell.ani"},
	{0x521C4E627BF9E55FULL, "6207659_m.fxe"},
	{0x521C7A5B6CED7D15ULL, "w_encounter3_19_0.msh"},
	{0x521C7C5B6CED808BULL, "w_encounter3_19_0.mmh"},
	{0x521C885B6CED9528ULL, "w_encounter3_19_0.mao"},
	{0x521C8C4AFE5BE637ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwlkb.phy"},
	{0x521D80046925E46BULL, "6077823_m.fxe"},
	{0x521D935A3AC82BC8ULL, "246068428.ogg"},
	{0x521D9F80B07A4BF3ULL, "57430558.ogg"},
	{0x521DD91009619AFCULL, "6126772_f.fxe"},
	{0x521DDEA8E3E5B50DULL, "art\\characters\\creatures\\shriek\\animation\\combat\\diev1.gad"},
	{0x521E3434B246C74FULL, "209423403.ogg"},
	{0x521E5E39344FFD3FULL, "gen00fl_merrill6129559.cl"},
	{0x521E70D0A80161A5ULL, "vxlt_gallowstemplar_f_5x9_0_0.vlm"},
	{0x521E76D42ADFAF4AULL, "vxht_bartrand_n_5n4_0_0.vlm"},
	{0x521ECDCE17C47AC6ULL, "vxht_0_hightown_a2_ca3_0_0.vlm"},
	{0x521F482B8592CA25ULL, "6164044_m.fxe"},
	{0x521F54AF50F14EB5ULL, "hlac_lt_gallowstemplar_f_2384_7x.dds"},
	{0x521F62CF129AD00BULL, "hlac_w_bonepitmining_a2_196609_0z.dds"},
	{0x521FEB4934AF586BULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.dura.ani"},
	{0x522003B92C22E6F5ULL, "zz_approval_debug6003781.cl"},
	{0x522041C2B46A709AULL, "llac_w_woundedcoast_1_0.dds"},
	{0x52206864634FB2F1ULL, "vxht_mansion2b_24j_0_0.vlm"},
	{0x52207ECBDCFDE6B1ULL, "hlac_w_woundedcoast_0_9y.dds"},
	{0x5220BB87C93BF009ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90runf.evt"},
	{0x5221A98CCA8323CCULL, "lowtown_fem_prostitute_sta6152310.cl"},
	{0x52223A5D1DD2CFC9ULL, "gen00fl_fenris6103346.cl"},
	{0x5222428BD704356FULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.dura.gad"},
	{0x522257841D2626E6ULL, "6107981_m.fxe"},
	{0x522310D2981AF57BULL, "docks_male_elven_worker_sta6175801.cl"},
	{0x5223134DAE332B3BULL, "art\\cinematics\\act1\\cs_fp_yell.gad"},
	{0x52231B75D95325B3ULL, "vxca_caverns1_mines_e59_0_0.vlm"},
	{0x5223A976DD8CE291ULL, "mage_tempest.dds"},
	{0x52241681253E7CF3ULL, "waterfall_dis.mao"},
	{0x5224674D544D8C16ULL, "vxw_sundermount_2j7_0_0.vlm"},
	{0x52249506DEE7E3ABULL, "gui_guipixelshader_p1.psh.fxo.dx11.assembly"},
	{0x5224BEFB7FCC4C93ULL, "vxca_caverns1_mines_i24_0_0.vlm"},
	{0x5225B691E2A3758FULL, "hlac_ht_mansion2b_818_4y.dds"},
	{0x5225D07F2AFE0ADAULL, "6177466_f.fxe"},
	{0x52261B6C61C843AFULL, "hlac_ht_brothel_n_3369_1z.dds"},
	{0x52266F08034B414DULL, "1051749606.ogg"},
	{0x52267247AB83676DULL, "l_mer141ar_caverns.bnk"},
	{0x5226EEABE2A5C6B5ULL, "vxw_bonepitdestroyed_a3_8e_0_0.vlm"},
	{0x52272581B54DAAF3ULL, "522056632.ogg"},
	{0x522725872E132346ULL, "p_distor.mao"},
	{0x52274F2432EC1B16ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallcelti.msh"},
	{0x5227592432EC2C14ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallcelti.mmh"},
	{0x52275D2432EC3327ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallcelti.mao"},
	{0x5227E4A49EECE1A7ULL, "w_bonepitdestroyed_a3_mmi.gff"},
	{0x5228396BD0997601ULL, "6061797_f.fxe"},
	{0x52289A28B62DFE86ULL, "gen_im_trs_cot_rng_r0.uti"},
	{0x5229578BD70A771CULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.dura.evt"},
	{0x522969DD550A9921ULL, "vxht_mansion2_2k3_0_0.vlm"},
	{0x52297D681B8C55DEULL, "one140cr_charade6128518.cl"},
	{0x5229819E218F38D8ULL, "llac_lt_undercityencounter_-1_13.dds"},
	{0x5229B54FAD4374A4ULL, "art\\characters\\playercharacter\\children\\kf_bdy_grl.msh"},
	{0x5229CB4FAD439A46ULL, "art\\characters\\playercharacter\\children\\kf_bdy_grl.mmh"},
	{0x522A02E6737B4957ULL, "6172706_m.fxe"},
	{0x522A2537E49857CEULL, "117756078.ogg"},
	{0x522A41402D436BAAULL, "one220_yevhen6103839.cl"},
	{0x522A4A956F8B2D72ULL, "royalarcher_rangedcombo.dds"},
	{0x522A5D5B6CF989F0ULL, "w_encounter3_19_0.anb"},
	{0x522AEE9247299CF2ULL, "dae300_vashothlead6062803.cl"},
	{0x522B43E2775C72A0ULL, "genip_door_highin_chantry.utp"},
	{0x522B48E3F7A3B74CULL, "hlac_ca_caverns1_mines_24427_2y.dds"},
	{0x522BB5EAA2AD5359ULL, "hlac_lt_gallowstemplar_7198_0x.dds"},
	{0x522C8BFF5FB3F0B7ULL, "gen00fl_isabela6103959.cl"},
	{0x522CA79C5C06EED5ULL, "634088271.ogg"},
	{0x522CD34934B9B9E2ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.dura.evt"},
	{0x522CE2AD99F77A25ULL, "vxlt_gallowstemplar_f_5c8_0_0.vlm"},
	{0x522D1CAB906AE7C0ULL, "zz_ser_debug6086476.cl"},
	{0x522D2709D06FF239ULL, "1010256415.ogg"},
	{0x522D3AB49C536E2BULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv1.ani"},
	{0x522D87FFF6BB4E0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_archesthin.msh"},
	{0x522D95FFF6BB6640ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_archesthin.mao"},
	{0x522D99FFF6BB6C93ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_archesthin.mmh"},
	{0x522E69187C65DB54ULL, "lgt000cr_beacon.utc"},
	{0x522EA387C9480578ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr90runf.ani"},
	{0x522EA50790EA30CCULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_foam01.dds"},
	{0x522EAB447B30DC78ULL, "llac_ca_caverns1_24424_23.dds"},
	{0x522EDA415545CB71ULL, "vxht_mansion2b_21i_0_0.vlm"},
	{0x522F5A3106AB8A90ULL, "clf260_harley6069640.cl"},
	{0x523040714B079647ULL, "gen00fl_merrill6111277.cl"},
	{0x52307E96645A54B6ULL, "zz_mageslate_transactual.cut"},
	{0x5230852C3E80BA0AULL, "vxlt_darktown_5b3_0_0.vlm"},
	{0x5230B646A438E71FULL, "hlac_w_sundermount_-2_2y.dds"},
	{0x52314D3E749C6E21ULL, "394444517.ogg"},
	{0x52315A0B3E57643BULL, "vxw_woundedcoast_14o_0_0.vlm"},
	{0x5231664BF264F326ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout_1.mao"},
	{0x52318144AFDF1A3AULL, "pn_eye_nrmi_0n.dds"},
	{0x52321ABAB19D9579ULL, "vxw_bonepitdestroyed_a3_26g_0_0.vlm"},
	{0x5232B7FFDB6E41A8ULL, "125951061.ogg"},
	{0x5232BB5759D731F4ULL, "vxlt_gallowstemplar_n_3r0_0_0.vlm"},
	{0x5232D92A4643A496ULL, "vxw_sundermount_2o8_0_0.vlm"},
	{0x5232D9D0719C3794ULL, "6028846_m.fxe"},
	{0x5232E3993C97CEF0ULL, "weapon_rendertoshadowmap_p0.psh.fxo.dx11.assembly"},
	{0x523327F2AF2F1865ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_f_c.anb"},
	{0x5233A77F7D38827EULL, "6126351_m.fxe"},
	{0x5233D24934BFD6F5ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.dura.gad"},
	{0x5233D6C83AD4AC2DULL, "and230st_mage_girl_2.stg"},
	{0x5234C24FAD4D9211ULL, "art\\characters\\playercharacter\\children\\kf_bdy_grl.phy"},
	{0x52350853A30A61EDULL, "527387009.ogg"},
	{0x5235149E6C4FD4BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_right.gad"},
	{0x5235220606551802ULL, "ser270_emeric6113433.cl"},
	{0x52353C50552B7E13ULL, "6097850_m.fxe"},
	{0x52355E63DA1653CCULL, "tools_way_pt02.phy"},
	{0x5235B31B96E13126ULL, "fx_dd_swipe.mao"},
	{0x5235C14F435762C0ULL, "hlac_lt_gallowscourtyard_131071_2y.dds"},
	{0x523601C3CA34C041ULL, "6208589_f.fxe"},
	{0x523606EFF37E0B60ULL, "198136905.ogg"},
	{0x52365F8BD7150F35ULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.dura.ani"},
	{0x5236709BBF123CD0ULL, "mcr310_sibling_st.stg"},
	{0x52368B56A1677F93ULL, "402021721.ogg"},
	{0x52379489312BECDDULL, "6128372_m.fxe"},
	{0x5237A16070F86C3BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud2high.msh"},
	{0x5237B36070F88A86ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud2high.mao"},
	{0x5237BF6070F89F25ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud2high.mmh"},
	{0x5238083A140D2CE8ULL, "144075401.ogg"},
	{0x523808F043C57F48ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1ww_basic.mmh"},
	{0x523814F043C593EBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1ww_basic.mao"},
	{0x52381D0280109179ULL, "vxlt_0_lowtown_a2_f5e_0_0.vlm"},
	{0x523826F043C5B272ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1ww_basic.msh"},
	{0x52388BEA2522E573ULL, "741590274.ogg"},
	{0x5238927EE4469ADFULL, "zz_dog_debug6127798.cl"},
	{0x5238B3F2B76484CAULL, "868527996.ogg"},
	{0x5238BA99722EDD51ULL, "496066637.ogg"},
	{0x5238E0025BCD7AB6ULL, "one220_yevhen6098704.cl"},
	{0x5238E77E96225387ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_2z.dds"},
	{0x5238EF677C135655ULL, "gen00fl_aveline6121967.cl"},
	{0x5239541EB3DFD416ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mase.mao"},
	{0x52395F5F54CD21C9ULL, "6033713_m.fxe"},
	{0x5239601EB3DFE8B5ULL, "art\\characters\\playercharacter\\humanmale\\hlh_mase.mmh"},
	{0x5239621EB3DFEC2BULL, "art\\characters\\playercharacter\\humanmale\\hlh_mase.msh"},
	{0x5239E37D8B453F22ULL, "vxlt_0_lowtown_docks_a3_app_0_0.vlm"},
	{0x5239E3B568F93961ULL, "plt_seb000pt_midstage.nss"},
	{0x523A15BFDAE58CC7ULL, "184688310.ogg"},
	{0x523A4E2299B1D794ULL, "seb321_leliana6180945.cl"},
	{0x523A6C5B644C2E73ULL, "vxlt_0_lowtown_e8k_0_0.vlm"},
	{0x523AB162AEA5496BULL, "mag291_ntbanter6056835.cl"},
	{0x523B22B49C5F82A2ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv1.evt"},
	{0x523B812662244810ULL, "w_bonepitmining_a2_211_0.msh"},
	{0x523B85EC5007D6ECULL, "combat_ar_r1_combo.gda"},
	{0x523B8F266224602AULL, "w_bonepitmining_a2_211_0.mmh"},
	{0x523B932662246761ULL, "w_bonepitmining_a2_211_0.mao"},
	{0x523C179E6C55F8D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_right.evt"},
	{0x523C1B7110F88A88ULL, "lt_0_lowtown_a2_26_0.anb"},
	{0x523C78C302FB71D1ULL, "6054981_m.fxe"},
	{0x523CAF11A5D2010FULL, "zz_seb_debug6183086.cl"},
	{0x523CF1B3C4181C8DULL, "6044683_m.fxe"},
	{0x523D2AA23D317B52ULL, "hlac_lt_0_lowtown_n_-262144_0z.dds"},
	{0x523D9A32916285DCULL, "5541980.ogg"},
	{0x523DB4C730929EAAULL, "vxw_woundedcoast_zq_0_0.vlm"},
	{0x523DB937B6195939ULL, "lgt120cr_orz_bln_messenger.utc"},
	{0x523E0AA2EF247C97ULL, "6195823_f.fxe"},
	{0x523E1E946FE0006AULL, "vxlt_0_lowtown_a2_gp8_0_0.vlm"},
	{0x523E467A5E538D8BULL, "gen00fl_anders6211960.cl"},
	{0x523E52B817A30489ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_loop.gad"},
	{0x523E8F0AE7467A64ULL, "gen00fl_isabela6075264.cl"},
	{0x523EC419C11883DAULL, "6147794_m.fxe"},
	{0x523F0C5F9E8A8DE3ULL, "llac_lt_gallowscourtyard_n_-65538_4.dds"},
	{0x523F21943FFBAE66ULL, "ave121_jeven6105638.cl"},
	{0x523F51F1BFC035F4ULL, "47570650.ogg"},
	{0x523FF2EC6B71C819ULL, "mag130tr_danzig.utt"},
	{0x52408C80919F45E5ULL, "vxht_mansion2_2am_0_0.vlm"},
	{0x52408C97A393ADF9ULL, "vxw_bonepitdestroyed_a3_21f_0_0.vlm"},
	{0x52414B073839BD16ULL, "vxw_sundermount_2p9_0_0.vlm"},
	{0x5241962849886240ULL, "42562810.ogg"},
	{0x5241A8D875E90549ULL, "vxw_bonepitmining_a2_mt_0_0.vlm"},
	{0x52420055CFB1902DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba03a.mmh"},
	{0x52421255CFB1AEB3ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba03a.msh"},
	{0x524221B49C659FB5ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv1.gad"},
	{0x52426160159F4382ULL, "sky_lt_0_lowtown_a2.mao"},
	{0x5242D4C1978B54A4ULL, "vxw_sundermount_2at_0_0.vlm"},
	{0x5243DF66C9476AC4ULL, "88912082.ogg"},
	{0x5243F9019A5EB6A0ULL, "dae000cr_grand_cleric6148110.cl"},
	{0x524427244537E1C9ULL, "dae000_arishok6174618.cl"},
	{0x52444B1C967C4902ULL, "223154638.ogg"},
	{0x524472C0B2D38107ULL, "218965097.ogg"},
	{0x524523B817A8D3D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_loop.ani"},
	{0x52459B8CBC7BC013ULL, "vxlt_gallowstemplar_f_378_0_0.vlm"},
	{0x5245B57C7AEA7D37ULL, "vxlt_gallowstemplar_5o9_0_0.vlm"},
	{0x5245E9E5BD572313ULL, "6039545_f.fxe"},
	{0x524600788FB6159BULL, "48412828.ogg"},
	{0x524600CB48E8EA53ULL, "w_bonepitdestroyed_a3_152_0.phy"},
	{0x5246360078C8198CULL, "hlac_lt_0_lowtown_docks_-65539_5z.dds"},
	{0x52464D9738F6CA21ULL, "vxht_mansion2_2q1_0_0.vlm"},
	{0x52465C6072C5D997ULL, "mrl310_merrill6120409.cl"},
	{0x5246C2FA7564D6E0ULL, "vxca_caverns2_7mf_0_0.vlm"},
	{0x5246C56FE3B1B6DCULL, "vxlt_0_lowtown_docks_n_aqi_0_0.vlm"},
	{0x5246E205C0E2994AULL, "792908627.ogg"},
	{0x52473903A916FDEDULL, "gen00fl_varric6075605.cl"},
	{0x52476AED15316AF0ULL, "384305030.ogg"},
	{0x5247D7DAC95C9C5FULL, "hlac_ht_chantry_2534_20z.dds"},
	{0x52481CB6D043F42FULL, "247678872.ogg"},
	{0x52490C7F783134BEULL, "551852253.ogg"},
	{0x52492F9E6C60ABA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_right.ani"},
	{0x5249905CEA5AFB5AULL, "art\\characters\\creatures\\strider\\cn_bdy_str.mmh"},
	{0x5249945CEA5B0191ULL, "art\\characters\\creatures\\strider\\cn_bdy_str.mao"},
	{0x5249A25CEA5B19A0ULL, "art\\characters\\creatures\\strider\\cn_bdy_str.msh"},
	{0x524A638FAF99B505ULL, "vxht_chantry_n_27p_0_0.vlm"},
	{0x524A8E0AEF030FFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_enter.ani"},
	{0x524B0599CAE3D0EFULL, "48970074.ogg"},
	{0x524B15A1BA3F19ACULL, "6069948_m.fxe"},
	{0x524B1A2919BD0C50ULL, "vxht_0_hightown_ai7_0_0.vlm"},
	{0x524B60FA34BF9F41ULL, "689487704.ogg"},
	{0x524BDB7A9CB77BD9ULL, "gen_im_trs_apo_lgt_hlm_r1.uti"},
	{0x524BE7984A971807ULL, "325486184.ogg"},
	{0x524C26A42288B72AULL, "vxw_woundedcoast_wr_0_0.vlm"},
	{0x524C27951A146FB9ULL, "6119185_m.fxe"},
	{0x524C4FC427A33F7AULL, "hlac_lt_0_lowtown_docks_n_-65535_0x.dds"},
	{0x524C79390F39627CULL, "6129240_m.fxe"},
	{0x524C7FD856180D3DULL, "26121382.ogg"},
	{0x524C907161D618EAULL, "vxlt_0_lowtown_a2_gw9_0_0.vlm"},
	{0x524CC2814A84B8CFULL, "6131628_m.fxe"},
	{0x524CCA22A0470E2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_enter.ani"},
	{0x524D1C8085031409ULL, "art\\characters\\playercharacter\\humanmale\\boo_masb.phy"},
	{0x524D23099DB508A5ULL, "gen_im_acc_rng_act2_05.uti"},
	{0x524D35F6B30E9C5AULL, "6140694_m.fxe"},
	{0x524DB4618EC18854ULL, "hlac_ht_keep_3740_0y.dds"},
	{0x524E16547DF62682ULL, "1052715791.ogg"},
	{0x524E7830E6399AD3ULL, "6081192_m.fxe"},
	{0x524E7833D2ECBFF8ULL, "t3_arm_rlr.tnt"},
	{0x524EF81CEBD886F6ULL, "6206626_m.fxe"},
	{0x524EFE5D83955E65ULL, "vxht_mansion2_2jl_0_0.vlm"},
	{0x524F5DB4391381CAULL, "862010779.ogg"},
	{0x524FDE54EAEA96B3ULL, "vxlt_hangedman_260_0_0.vlm"},
	{0x525009ED87999895ULL, "vxca_cavern1_haunted_ihd_0_0.vlm"},
	{0x525035E5AD675F38ULL, "gen00fl_isabela6096850.cl"},
	{0x52506C7A92B25C9EULL, "6174486_m.fxe"},
	{0x5250F63795DFB84AULL, "vxht_mansion2b_1bx_0_0.vlm"},
	{0x5251469E89816D24ULL, "vxw_sundermount_2ju_0_0.vlm"},
	{0x52519B0AEF09443EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_enter.gad"},
	{0x52522A46A4534D39ULL, "mag291ip_to_caress.ncs"},
	{0x52525A6034E76566ULL, "hlac_ca_cavern1_haunted_24427_1x.dds"},
	{0x52526847ABA89097ULL, "l_mer141ar_caverns.wbd"},
	{0x525270126C010D43ULL, "180272694.ogg"},
	{0x5252881994058C99ULL, "6063860_m.fxe"},
	{0x52528B847D604C43ULL, "llac_ca_caverns2prc_5802_43.dds"},
	{0x5252D5D6CEC28929ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robe_n.dds"},
	{0x5252F955CFBFF7EAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba03a.phy"},
	{0x52532B13C1F375E7ULL, "follower_banter6143183.cl"},
	{0x52537CF0F3B9BD35ULL, "w_bonepitdestroyed_a3_35_0.rml"},
	{0x52541A829365D36BULL, "llac_lt_0_lowtown_-65536_2.dds"},
	{0x5254A88290F9E06FULL, "151949130.ogg"},
	{0x5254D60D9247ADACULL, "vxlt_warehouse_n_u9_0_0.vlm"},
	{0x52551E49CF29E011ULL, "gen00fl_aveline6114813.cl"},
	{0x525521FA6D9C4976ULL, "hlac_lt_0_lowtown_docks_n_-65538_6y.dds"},
	{0x5255D5D48C3B392AULL, "gen00fl_sebastian6174993.cl"},
	{0x5255E366A34A8A88ULL, "6128012_m.fxe"},
	{0x5255EE7B1AE4C7C1ULL, "6126378_m.fxe"},
	{0x5256317BA90AA467ULL, "hlac_w_woundedcoast_327686_0x.dds"},
	{0x5256D31C910E47F2ULL, "6088316_m.fxe"},
	{0x5256D5C102EEEA89ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_corruptlyrium_n.dds"},
	{0x52577C7ED87F6137ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm06.mmh"},
	{0x5257927ED87F86A9ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm06.msh"},
	{0x5257B4E23DAE5F5AULL, "dre000_dougal6130967.cl"},
	{0x5257E8BC00148AE8ULL, "character_character_p5.psh.fxo"},
	{0x5257EC24C1D659BAULL, "987301657.ogg"},
	{0x525844557EB7CFD1ULL, "595558901.ogg"},
	{0x52586D4D365C2EF6ULL, "6125444_m.fxe"},
	{0x5258760AEF0F245DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_enter.evt"},
	{0x5258804B19B42891ULL, "dae210_gate_guard6085131.cl"},
	{0x525887F5AA4887D7ULL, "hlac_ht_chantry_2534_28z.dds"},
	{0x5258A7A2C3C6F738ULL, "6154778_m.fxe"},
	{0x5258EDD627345B3EULL, "llac_w_bonepitdestroyed_a3_4_0.dds"},
	{0x52594365DF8B7D43ULL, "6042751_m.fxe"},
	{0x5259AB4A62016292ULL, "1038245119.ogg"},
	{0x525A1C46B9675C3BULL, "gen_im_trs_rdr_one_mac_r2.uti"},
	{0x525A64F28D13F585ULL, "6018718_m.fxe"},
	{0x525AB222A05322A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_enter.evt"},
	{0x525ABBBF920A0648ULL, "hlac_w_blightlands_-65539_0x.dds"},
	{0x525B565347546168ULL, "6167053_m.fxe"},
	{0x525C28F2AF521EA6ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_f_c.mmh"},
	{0x525C43A9C31719F9ULL, "6028444_f.fxe"},
	{0x525CEA0DD82FB353ULL, "6084292_m.fxe"},
	{0x525D40037CE4DD96ULL, "llac_lt_blackemporium_1055_19.dds"},
	{0x525D488C2E47EA82ULL, "vxca_caverns2_8kq_0_0.vlm"},
	{0x525D528CDB909B57ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\256x128.dds"},
	{0x525D7061A15CA67EULL, "6110585_m.fxe"},
	{0x525DDC54CFA0E265ULL, "llac_lt_0_lowtown_n_65538_0.dds"},
	{0x525DF3EC6B8AC7DCULL, "mag130tr_danzig.ncs"},
	{0x525E60E79B4FEE3FULL, "6105476_m.fxe"},
	{0x525E8AE8F4C4FB54ULL, "735219726.ogg"},
	{0x525E925F493DC52BULL, "6076899_m.fxe"},
	{0x525EAC9102318E0EULL, "zz_followers_banter6158832.cl"},
	{0x525F0F57BD0D5A72ULL, "85255315.ogg"},
	{0x525FB87B7B7785A4ULL, "vxw_sundermount_2oz_0_0.vlm"},
	{0x52606CAAA3996A03ULL, "gen00fl_fenris6106668.cl"},
	{0x526086710ACD58AAULL, "646908892.ogg"},
	{0x5260C0F910D6D447ULL, "129255557.ogg"},
	{0x52611E6463DBFBC6ULL, "hlac_w_sundermount_-327682_0z.dds"},
	{0x5261415003815252ULL, "gen00fl_varric6077826.cl"},
	{0x52619DA7A8BA0A39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_catch_and_hold.ani"},
	{0x5261B122A0593FB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_enter.gad"},
	{0x5261E5CA18CDF75EULL, "6065285_m.fxe"},
	{0x526329579B0D359EULL, "456357086.ogg"},
	{0x5263614AFE979F8FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwlkb.mao"},
	{0x52636D4AFE97B42CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwlkb.mmh"},
	{0x52637158D56619D1ULL, "358074091.ogg"},
	{0x5263734AFE97BE1EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldwlkb.msh"},
	{0x526399C40C2A627CULL, "hlac_ht_0_hightown_-65538_0y.dds"},
	{0x5263D1980AF4370AULL, "isa111_isabela6067098.cl"},
	{0x526411AC31B91231ULL, "vxw_encounter1_dh_0_0.vlm"},
	{0x526443AB5BC170C3ULL, "95970652.ogg"},
	{0x5264BF46BAE129FDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a.phy"},
	{0x52652239C4783235ULL, "llac_w_encounter2_camp_-65535_0.dds"},
	{0x52653E59E110E6E1ULL, "6168145_m.fxe"},
	{0x5265641E1A06B4D2ULL, "6152529_m.fxe"},
	{0x5265CCE8A2155539ULL, "hightown_market_guard_sta6169665.cl"},
	{0x5265D3F9FC973C42ULL, "gen00fl_varric6018421.cl"},
	{0x5266315F3EBBB70DULL, "vxlt_0_lowtown_ey2_0_0.vlm"},
	{0x52666C382BA42F67ULL, "6141861_m.fxe"},
	{0x5266FD26364CC74AULL, "vxlt_0_lowtown_a2_n_gyl_0_0.vlm"},
	{0x52672CE349EB7162ULL, "vxw_woundedcoast_119_0_0.vlm"},
	{0x5267F59EE5711D2CULL, "vxlt_darktown_4t6_0_0.vlm"},
	{0x526806A68C2829D6ULL, "lgt150ip_ntb_wr_ing_chest_1.utp"},
	{0x526878A7A8BFEA7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_catch_and_hold.gad"},
	{0x526885BC49573100ULL, "gen_im_acc_blt_lore_03.uti"},
	{0x5268C5DBF18EC0FFULL, "dae300_petrice6187894.cl"},
	{0x5268DFF6631ECDC2ULL, "6089702_m.fxe"},
	{0x52692B1E077359F4ULL, "6182373_m.fxe"},
	{0x5269ABEC65F78911ULL, "6125375_m.fxe"},
	{0x526A0B8EF8BCD7FDULL, "6122077_m.fxe"},
	{0x526A1BDBAAD8E5EBULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunariamulet_n.dds"},
	{0x526A9476DEE4EC54ULL, "mag360_alain6047696.cl"},
	{0x526AF92EC5E1C8E6ULL, "dae000_merchant_hm_mob6174215.cl"},
	{0x526B1938B46BB5A7ULL, "zz_vault_debug6177784.cl"},
	{0x526B294FF60CBF43ULL, "6025082_m.fxe"},
	{0x526B6B3E12C4F169ULL, "vxlt_0_lowtown_docks_a3_n_d16_0_0.vlm"},
	{0x526BABCBFA5EB0E0ULL, "sky_ca_caverns2prc.mao"},
	{0x526BDB5301AC03FFULL, "22946729.ogg"},
	{0x526BE21767818F65ULL, "vxht_mansion2_2pj_0_0.vlm"},
	{0x526C2A1905146A5CULL, "vxlt_0_lowtown_a2_n_eid_0_0.vlm"},
	{0x526C5152A0E05DB4ULL, "follower_banter6156197.cl"},
	{0x526DD1123DD9B8F4ULL, "zz_vault_debug6178924.cl"},
	{0x526DD72405A82960ULL, "vxlt_0_lowtown_h27_0_0.vlm"},
	{0x526DD9F179CBE94AULL, "vxht_mansion2b_1xz_0_0.vlm"},
	{0x526E699A96924450ULL, "lgt160_ntb_elf_conflict6172097.cl"},
	{0x526E83C65569098EULL, "hlac_lt_0_lowtown_a2_n_196607_1z.dds"},
	{0x526E95A7A8C48760ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_catch_and_hold.evt"},
	{0x526EBC7411B0E928ULL, "65144385.ogg"},
	{0x526F1AE31B985CC2ULL, "zz_followers_banter6158418.cl"},
	{0x526F3499DD8560F3ULL, "337061641.ogg"},
	{0x526F53BC2B565AA0ULL, "hlac_ca_caverns2_5802_20y.dds"},
	{0x526F6A3187CDC73BULL, "6185814_m.fxe"},
	{0x526F964B0148A158ULL, "6036971_m.fxe"},
	{0x526FAF5CEA7BFF65ULL, "art\\characters\\creatures\\strider\\cn_bdy_str.phy"},
	{0x5270009D96B81C16ULL, "mcr310_argument6075123.cl"},
	{0x5270AAB01AEA7B83ULL, "hm_carver_h2s1.mor"},
	{0x5270D46E46112AEDULL, "6219236_m.fxe"},
	{0x5270EA7A96759096ULL, "211154691.ogg"},
	{0x52711D0583F0D28CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45win.phy"},
	{0x52716F3B03E86A6BULL, "vxca_caverns1_mines_coi_0_0.vlm"},
	{0x52717B2F4E680E91ULL, "hlac_lt_0_lowtown_a3_n_c_-131071_2x.dds"},
	{0x5272684FCF238611ULL, "vxht_0_hightown_n_bwt_0_0.vlm"},
	{0x527270DBFFC34A07ULL, "vxlt_hangedman_28n_0_0.vlm"},
	{0x5272838923AF2AB1ULL, "vxw_encounter1_io_0_0.vlm"},
	{0x5272B028E0D3251AULL, "zz_one_debug6077960.cl"},
	{0x52735F23F7800884ULL, "777129645.ogg"},
	{0x5273615EFD181142ULL, "6146784_m.fxe"},
	{0x5273E2C56685407FULL, "dae000_mother6136957.cl"},
	{0x527489E5F211E6BDULL, "hlac_lt_0_lowtown_a2_-131073_3y.dds"},
	{0x5274D1C329FFD484ULL, "vxht_0_hightown_a3_bya_0_0.vlm"},
	{0x52753563DA4BF197ULL, "tools_way_pt02.mmh"},
	{0x52754B63DA4C1709ULL, "tools_way_pt02.msh"},
	{0x52756F032842DFCAULL, "vxlt_0_lowtown_a2_n_gvm_0_0.vlm"},
	{0x527575D26BC9F329ULL, "6017610_m.fxe"},
	{0x5275B446BAEF8A69ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a.mao"},
	{0x5275BBAF1D74595DULL, "586649380.ogg"},
	{0x5275BF7AB597A4C1ULL, "6089864_m.fxe"},
	{0x5275C046BAEF9E02ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a.mmh"},
	{0x5275C246BAEFA1B8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a.msh"},
	{0x5276063EFE13C8F1ULL, "llac_lt_backalley_n_65537_4.dds"},
	{0x52764D02071C05A4ULL, "gen00fl_aveline6114018.cl"},
	{0x5276C7A9313F2AC3ULL, "ico_tome.dds"},
	{0x527824E49DDAE89DULL, "hlac_ca_caverns2_5802_9y.dds"},
	{0x527832CD4E41D929ULL, "6040161_f.fxe"},
	{0x5278D088E7F411B0ULL, "480297020.ogg"},
	{0x5278DBD9F39ABF16ULL, "6140618_m.fxe"},
	{0x5278FD93B0547879ULL, "6019031_f.fxe"},
	{0x527905808528282EULL, "art\\characters\\playercharacter\\humanmale\\boo_masb.mmh"},
	{0x5279118085283CC5ULL, "art\\characters\\playercharacter\\humanmale\\boo_masb.mao"},
	{0x52791F808528541CULL, "art\\characters\\playercharacter\\humanmale\\boo_masb.msh"},
	{0x5279BE48ED72A155ULL, "6171873_m.fxe"},
	{0x5279CE58C31F12B4ULL, "6121503_m.fxe"},
	{0x527A2FBF95B4EAB1ULL, "hlac_lt_gallowstemplar_n_2394_8x.dds"},
	{0x527A3AD3D08E35A7ULL, "383502598.ogg"},
	{0x527A710FB13CEEE7ULL, "123194225.ogg"},
	{0x527A9BF5F70A82DCULL, "vxlt_0_lowtown_a2_n_ere_0_0.vlm"},
	{0x527B8E42C0F14F74ULL, "mag321cr_servant_gossip_2b.utc"},
	{0x527BCC5014A50BC0ULL, "450891452.ogg"},
	{0x527BF77DE93CF8F4ULL, "ss_ambient_female.cnv"},
	{0x527C1A71112E6D40ULL, "lt_0_lowtown_a2_26_0.rml"},
	{0x527C1B11D4186CD6ULL, "llac_lt_0_lowtown_a2_n_c_65533_0.dds"},
	{0x527C1B41D16CC2EFULL, "378927272.ogg"},
	{0x527C2942E3AFF3D1ULL, "462832505.ogg"},
	{0x527CCF865E44B34BULL, "seb221_flora_harimann6177670.cl"},
	{0x527CCFF5E4673AC1ULL, "gen00fl_carver6102732.cl"},
	{0x527D03180D865CAEULL, "vxca_caverns2_4k9_0_0.vlm"},
	{0x527E13DA2370B337ULL, "dae021_sister_samea_sta6150143.cl"},
	{0x527E6D261BB65DD4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_noba_0s.dds"},
	{0x527E8C12A5E90CBDULL, "hlac_w_woundedcoast_-65529_0x.dds"},
	{0x527E986071345FD2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud2high.phy"},
	{0x527E9CA566CB6392ULL, "471126786.ogg"},
	{0x527E9E849056D0DAULL, "vxht_keep_5go_0_0.vlm"},
	{0x527ED97DE93EFB34ULL, "ss_ambient_female.bnk"},
	{0x527EDFCC677F1012ULL, "vxlt_gallowstemplar_3r1_0_0.vlm"},
	{0x527F58CD8107A8A1ULL, "vxlt_hangedman_5m_0_0.vlm"},
	{0x527F593382982A81ULL, "cd_mcr370_meredith.wbd"},
	{0x527FE966E8433A69ULL, "gen00flst_midfriendrival.stg"},
	{0x528045C6840A4F1CULL, "lt_0_lowtown_docks_n_20_0.anb"},
	{0x52805D57B304D8D1ULL, "vxdr_primevaltomb_6pw_0_0.vlm"},
	{0x5280647C3B65306EULL, "art\\vfx\\models\\vfx_moon.mmh"},
	{0x52807E7C3B655C5CULL, "art\\vfx\\models\\vfx_moon.msh"},
	{0x5280E2B8F1B96287ULL, "vxlt_hangedman_25i_0_0.vlm"},
	{0x5281350D707DE7D8ULL, "hlac_ca_caverns1_mines_24422_0x.dds"},
	{0x5281883B46C8D2E5ULL, "vxca_cavern1_haunted_ih4_0_0.vlm"},
	{0x52818B62671898DDULL, "6147726_m.fxe"},
	{0x5281A260008C2CF8ULL, "6024543_m.fxe"},
	{0x5281A976E741A403ULL, "6014253_m.fxe"},
	{0x5281B717FAB59189ULL, "299694415.ogg"},
	{0x5282378CC835C9E8ULL, "vxlt_gallowscourtyard_6xf_0_0.vlm"},
	{0x5282B01A32DC6CE9ULL, "zz_seb_debug6183097.cl"},
	{0x5282C82012202B41ULL, "vxw_bonepitmining_a2_1pp_0_0.vlm"},
	{0x528305BCD370549BULL, "art\\characters\\creatures\\strider\\animation\\attkfl_arms.ani"},
	{0x5283151922A7E80DULL, "vxlt_0_lowtown_eo0_0_0.vlm"},
	{0x528325F9B4D9A372ULL, "qcr220_orsino6061787.cl"},
	{0x52835824333A890FULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallcelti.phy"},
	{0x5283CB07004A1D1BULL, "53821619.ogg"},
	{0x528446ED691E9D7AULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyn.mmh"},
	{0x52844AED691EA3B1ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyn.mao"},
	{0x528458ED691EBBC0ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyn.msh"},
	{0x52848994BA600833ULL, "321616200.ogg"},
	{0x528497B77576AC80ULL, "qun220cr_javaris6071710.cl"},
	{0x5284A42F530B2A17ULL, "weapon1_s2y.psh.fxo"},
	{0x52851AF41DC69773ULL, "llac_dr_dungeon_3216_10.dds"},
	{0x5285B860DBCFFD43ULL, "pro000st_bethany.stg"},
	{0x5285C3B0470AFEC2ULL, "6087502_m.fxe"},
	{0x52863077F10FB866ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_v5.evt"},
	{0x52864DDDD84C39E2ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_nat_p.mmh"},
	{0x528696750F020AE5ULL, "follower_banter6161202.cl"},
	{0x528753B4E2FE7278ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_4z.dds"},
	{0x52877DC36F21449DULL, "vxw_encounter2_fc_0_0.vlm"},
	{0x5287D4D3FA1EE38CULL, "6060745_m.fxe"},
	{0x5287D7A2A7A0980AULL, "851416234.ogg"},
	{0x52884CEC02973739ULL, "lt_0_lowtown_docks_a3_n_144_0.msh"},
	{0x528856EC02974827ULL, "lt_0_lowtown_docks_a3_n_144_0.mmh"},
	{0x52885AEC02974F74ULL, "lt_0_lowtown_docks_a3_n_144_0.mao"},
	{0x528889A435B7589CULL, "hf_amb_pickpocket.mor"},
	{0x528894CD57CE6782ULL, "vxht_0_hightown_bwx_0_0.vlm"},
	{0x52889A60C20697F4ULL, "w_bonepitmining_a2_47_0.rml"},
	{0x5288D896DB8A75A4ULL, "6108160_m.fxe"},
	{0x5288E1E621C34851ULL, "llac_lt_ruinedhouse_n_-1_3.dds"},
	{0x52890DD2E9009B5CULL, "vxlt_0_lowtown_a2_n_ewz_0_0.vlm"},
	{0x5289253BA9ECDF1BULL, "ef_serendipity.mor"},
	{0x5289B8A7652C6AE2ULL, "hlac_ca_caverns1_mines_24423_4y.dds"},
	{0x5289BDC1E6915CF9ULL, "300286622.ogg"},
	{0x5289C5DAF7CA7959ULL, "1014434718.ogg"},
	{0x5289E4CBC39FB855ULL, "ser270_emeric6149358.cl"},
	{0x528A7A75384A6C22ULL, "433952725.ogg"},
	{0x528A94E8A4341A9DULL, "lgt200cr_gang_2_docks.utc"},
	{0x528AADAEAA266B69ULL, "dae000_mother6103065.cl"},
	{0x528ABDAB5DB81A4AULL, "vxht_mansion2b_1vt_0_0.vlm"},
	{0x528ACD3BAC1DE8CCULL, "9371084.ogg"},
	{0x528B3E5F5FFBFBF5ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_008a_0t.dds"},
	{0x528BF0704CE63C43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_down.evt"},
	{0x528C2B186EB33EF2ULL, "vxlt_gallowstemplar_5c4_0_0.vlm"},
	{0x528C3CC47EA056A1ULL, "6116283_m.fxe"},
	{0x528C52FE236EE0CCULL, "vxlt_gallowstemplar_486_0_0.vlm"},
	{0x528C5686851DA7ADULL, "l_lgt220ar_gang_hideout_ht.wbd"},
	{0x528C96DB334D9D6EULL, "557048942.ogg"},
	{0x528CB342B90CA00EULL, "hlac_lt_0_lowtown_a2_n_-3_1y.dds"},
	{0x528D12259AB3A0B3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_highlandplant01_n.dds"},
	{0x528D2F1D8D6749C1ULL, "vxht_playershouse_n_1d3_0_0.vlm"},
	{0x528D3D49DE137FD9ULL, "598536056.ogg"},
	{0x528D4777F115FDB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_v5.gad"},
	{0x528DF3F14B404A80ULL, "6096335_m.fxe"},
	{0x528E06ACC2B24A8FULL, "vxca_caverns1_mines_ehj_0_0.vlm"},
	{0x528E7EC46210380EULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffdeselectedexit.ani"},
	{0x528E9428AC8E6B81ULL, "vxlt_0_lowtown_a2_n_ei5_0_0.vlm"},
	{0x528ECF34A4FAF151ULL, "vxdr_primevaltomb_6wv_0_0.vlm"},
	{0x528EF26C54F60C91ULL, "950017542.ogg"},
	{0x528EFBEEE3C6FF32ULL, "324108866.ogg"},
	{0x528F5E359860C9ADULL, "one110_tree_minder.bnk"},
	{0x528FC1900CC08B7FULL, "780947161.ogg"},
	{0x528FF8294C5A30B4ULL, "dre000_cut2d_ragstoriches6025088.cl"},
	{0x52900DBCD37AEC92ULL, "art\\characters\\creatures\\strider\\animation\\attkfl_arms.evt"},
	{0x52904785706C8180ULL, "vxlt_0_lowtown_docks_n_bta_0_0.vlm"},
	{0x5290E1758EE1C448ULL, "and100_anders6206376.cl"},
	{0x52917EB8EEC3E928ULL, "vxlt_darktown_4u0_0_1.vlm"},
	{0x5291C6ACD986C1AFULL, "vxca_caverns2_8t9_0_0.vlm"},
	{0x5292165935E7B1CCULL, "vxht_0_hightown_c42_0_0.vlm"},
	{0x5292803598633915ULL, "one110_tree_minder.cnv"},
	{0x5292C9704CEC18C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_down.gad"},
	{0x5293F4B0C8C53861ULL, "vxht_bartrand_n_3qx_0_0.vlm"},
	{0x5294036BE6128E9AULL, "follower_banter6155424.cl"},
	{0x52944877F11C1E5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_v5.ani"},
	{0x52944E1ADE700FB2ULL, "6068435_m.fxe"},
	{0x52953F627C645952ULL, "ht_0_hightown_a3_lc.mmh"},
	{0x529590B1AFC4D3BFULL, "hlac_lt_0_lowtown_65535_3y.dds"},
	{0x5295B3AE975B309AULL, "seb201_lady_harimann6177569.cl"},
	{0x5295CE097953787FULL, "6185479_f.fxe"},
	{0x5295D8927C69EA14ULL, "art\\characters\\creatures\\spider\\animation\\cwkfl.ani"},
	{0x5296079C247DE8D5ULL, "71769663.ogg"},
	{0x52960DE553DB6BDFULL, "art\\characters\\weapons\\greatswords\\w_gsw_004a.msh"},
	{0x529613E553DB7581ULL, "art\\characters\\weapons\\greatswords\\w_gsw_004a.mmh"},
	{0x52961FE553DB8A22ULL, "art\\characters\\weapons\\greatswords\\w_gsw_004a.mao"},
	{0x5296A9F9B1FCDD0AULL, "hlac_lt_0_lowtown_-131072_5y.dds"},
	{0x5296FE52BEA0257CULL, "vxht_bartrand_n_66e_0_0.vlm"},
	{0x52972CBCD3814025ULL, "art\\characters\\creatures\\strider\\animation\\attkfl_arms.gad"},
	{0x5297D26768735426ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_gibbet_0.pwk"},
	{0x5298D8704CF09DA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_down.ani"},
	{0x52993ECE1723AF41ULL, "gen00fl_dog6166810.cl"},
	{0x52996EDC43050BB5ULL, "6125787_f.fxe"},
	{0x5299D3AEFE5D5EF8ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa_2.mao"},
	{0x529A0B61B421F9C9ULL, "gen00fl_anders6113132.cl"},
	{0x529A7E2C9843B676ULL, "6067755_m.fxe"},
	{0x529AB91E28736014ULL, "6147244_m.fxe"},
	{0x529B0BD771CB4920ULL, "llac_lt_0_lowtown_docks_a3_-65540_2.dds"},
	{0x529B2D77BB0EE615ULL, "vxlt_gallowscourtyard_7aq_0_0.vlm"},
	{0x529B76C4621AB55FULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffdeselectedexit.evt"},
	{0x529B9413A3B304C3ULL, "636053594.ogg"},
	{0x529C13F47AF11761ULL, "w_bonepitdestroyed_a3_183_0.phy"},
	{0x529C6B51BDD1CCD5ULL, "vxht_chantry_n_1yi_0_0.vlm"},
	{0x529CD5927C700476ULL, "art\\characters\\creatures\\spider\\animation\\cwkfl.gad"},
	{0x529D06059E848401ULL, "vxlt_0_lowtown_a2_n_er4_0_0.vlm"},
	{0x529D0E20C0CAB1E9ULL, "vxdr_primevaltomb_90z_0_2.vlm"},
	{0x529D411196F109D1ULL, "vxdr_primevaltomb_6zu_0_0.vlm"},
	{0x529D5AB6EC4F3B97ULL, "qun211cr_elf_warrior6072345.cl"},
	{0x529D77BB5A476B74ULL, "211156127.ogg"},
	{0x529DB1BB58FBD1F7ULL, "llac_w_woundedcoast_-196607_0.dds"},
	{0x529DC2ED68D7D8D9ULL, "6089108_f.fxe"},
	{0x529E08CEBD89B761ULL, "dae231ar_gallowsbasement.ncs"},
	{0x529E35E4667A68AEULL, "propalpha1_ds1.psh.fxo"},
	{0x529E675D25180565ULL, "6165173_m.fxe"},
	{0x529EB96262629A00ULL, "vxlt_0_lowtown_docks_n_bqn_0_0.vlm"},
	{0x529EE7052D2794B8ULL, "vxw_sundermount_34b_0_0.vlm"},
	{0x529EECD6A72323DAULL, "lowtown_male_beggar_sta6151069.cl"},
	{0x529EED91FE26CE60ULL, "vxca_cavern1_haunted_ec8_0_0.vlm"},
	{0x52A0D93423C3F89EULL, "396911526.ogg"},
	{0x52A1727EA6C2B007ULL, "lowtown_male_cityguard_mob6150487.cl"},
	{0x52A18B418B7FC1C0ULL, "6105686_m.fxe"},
	{0x52A192C0C0078C3BULL, "6018892_f.fxe"},
	{0x52A20C7F2A0C550BULL, "vxlt_warehouse_b6_0_0.vlm"},
	{0x52A23829600E9D3BULL, "gen00fl_bethany6101210.cl"},
	{0x52A23F296E0D810AULL, "vxca_caverns1_mines_jgu_0_0.vlm"},
	{0x52A240EB60E4485AULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravec_sa.dds"},
	{0x52A257C462209F48ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffdeselectedexit.gad"},
	{0x52A26ED251C739F7ULL, "gen00fl_carver6102679.cl"},
	{0x52A2CF8356D9DB19ULL, "rescalearea.ncs"},
	{0x52A33DE881AF8686ULL, "follower_banter6172852.cl"},
	{0x52A3F0927C7650D5ULL, "art\\characters\\creatures\\spider\\animation\\cwkfl.evt"},
	{0x52A4617D530D759DULL, "vxw_encounter2_la_0_0.vlm"},
	{0x52A4CDB85D521BF5ULL, "6077599_m.fxe"},
	{0x52A5227CE0D9B1D7ULL, "831250264.ogg"},
	{0x52A526A716E0AA3EULL, "6172985_m.fxe"},
	{0x52A5702FB0963DFCULL, "vxht_bartrand_n_63z_0_0.vlm"},
	{0x52A57B3382B93383ULL, "cd_mcr370_meredith.bnk"},
	{0x52A59D54775E9528ULL, "6076081_m.fxe"},
	{0x52A5BC50BF76A6A4ULL, "6106360_m.fxe"},
	{0x52A5CCA54692B08EULL, "vxht_mansion2_3c9_0_0.vlm"},
	{0x52A62BA1632C60F9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_enter.ani"},
	{0x52A7368469818498ULL, "6020653_m.fxe"},
	{0x52A7BED3201C2B30ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodmetal_sa.dds"},
	{0x52A7D58980D425C0ULL, "702898938.ogg"},
	{0x52A81DE2C05BDE5EULL, "85629396.ogg"},
	{0x52A81E89FF838B10ULL, "vxlt_gallowstemplar_f_45v_0_0.vlm"},
	{0x52A84741AA4D6FBEULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean_2.mao"},
	{0x52A8F9CB493D5BC8ULL, "w_bonepitdestroyed_a3_152_0.mmh"},
	{0x52A905CB493D706BULL, "w_bonepitdestroyed_a3_152_0.mao"},
	{0x52A917CB493D8EF2ULL, "w_bonepitdestroyed_a3_152_0.msh"},
	{0x52A92AFB1A697894ULL, "6140344_m.fxe"},
	{0x52A95E619BBD4A0EULL, "68637640.ogg"},
	{0x52A965ED693DED85ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyn.phy"},
	{0x52A9BABFF3C353FEULL, "6028912_m.fxe"},
	{0x52AAAD79D7FEEB36ULL, "781374341.ogg"},
	{0x52AAEBD766DF9C81ULL, "619959105.ogg"},
	{0x52AB68AAC5FA0DE1ULL, "vxlt_gallowscourtyard_n_4wg_0_0.vlm"},
	{0x52AB9265853ADA2AULL, "6128779_m.fxe"},
	{0x52ABECCFD0687024ULL, "6094683_m.fxe"},
	{0x52AC38D1EE448B2CULL, "vxlt_0_lowtown_a3_n_c_gof_0_1.vlm"},
	{0x52AC72ECFB4B7168ULL, "lgt110_arl_bann_teagan6115566.cl"},
	{0x52ACBBD4F4C23F53ULL, "plt_cod_itm_bianca.plo"},
	{0x52ACF09796EB258AULL, "314118624.ogg"},
	{0x52AD06A16332403BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_enter.gad"},
	{0x52AD58E21F1DAD38ULL, "vxw_sundermount_33c_0_0.vlm"},
	{0x52AD66DD81AAF75CULL, "437341725.ogg"},
	{0x52AD85E2D9888B43ULL, "hlac_lt_gallowstemplar_f_2383_1y.dds"},
	{0x52AE25A6CDD3CF81ULL, "6119589_m.fxe"},
	{0x52AE92D66F5279D1ULL, "vxlt_backalley_n_1xh_0_1.vlm"},
	{0x52AF4CDDD86F3C89ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_nat_p.anb"},
	{0x52AF7808DFB8B1BEULL, "llac_lt_0_lowtown_docks_a3_131071_4.dds"},
	{0x52AFA4A8EEBFD3B4ULL, "genip_bedroll.utp"},
	{0x52B02AD7692B0C5CULL, "vxca_caverns1_mines_i1r_0_0.vlm"},
	{0x52B06185375F6649ULL, "dae000_quintus6153103.cl"},
	{0x52B0A42D2DDFEF34ULL, "hlac_lt_0_lowtown_n_-65535_3z.dds"},
	{0x52B0ED2A288EA7E2ULL, "6106942_m.fxe"},
	{0x52B100E05103BC4DULL, "hlac_ca_cavern1_haunted_24422_7z.dds"},
	{0x52B159CF61670471ULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench3f_merillintro.gad"},
	{0x52B17F99840EA7ABULL, "terrain_blend7.psh.fxo"},
	{0x52B1C8028E568C59ULL, "vxlt_undercityserialkiller_17a_0_0.vlm"},
	{0x52B1C9B2931CD182ULL, "plt_prm000pt_air_ring.plo"},
	{0x52B1F051270CD430ULL, "hightown_male_noble1_sta6170869.cl"},
	{0x52B1F40584282457ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45win.mmh"},
	{0x52B1F80584282B64ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45win.mao"},
	{0x52B20A05842849C9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45win.msh"},
	{0x52B235F87AA3E395ULL, "6059704_m.fxe"},
	{0x52B2F8EC48398079ULL, "6105526_m.fxe"},
	{0x52B2FE3F39C8AC67ULL, "art\\characters\\creatures\\stonegolem\\animation\\dblw_dmg_v1.gad"},
	{0x52B3A3A2CFC4E7B6ULL, "907595346.ogg"},
	{0x52B3D95952F6E533ULL, "and121_karl6093902.cl"},
	{0x52B423A163389020ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_enter.evt"},
	{0x52B44C1FBF487761ULL, "dae000_friend_turn_in6179261.cl"},
	{0x52B49F5020CCAAB7ULL, "521801073.ogg"},
	{0x52B4B79840CE4F35ULL, "406663129.ogg"},
	{0x52B4DE97305CDD84ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maso.msh"},
	{0x52B4F097305CFCEDULL, "art\\characters\\playercharacter\\humanmale\\bdy_maso.mao"},
	{0x52B4F160874EB9C8ULL, "vxlt_gallowscourtyard_n_4wb_0_0.vlm"},
	{0x52B4F497305D0326ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maso.mmh"},
	{0x52B5689CA7A58A89ULL, "star_02.mao"},
	{0x52B594268AD90517ULL, "6027790_m.fxe"},
	{0x52B5A8615B779D18ULL, "6027753_m.fxe"},
	{0x52B5DFD0E429C44FULL, "6111073_m.fxe"},
	{0x52B6064FC204EE0EULL, "vxlt_gallowscourtyard_n_7h6_0_0.vlm"},
	{0x52B60FD2F0D8DB51ULL, "174533332.ogg"},
	{0x52B6134365C8F933ULL, "vxca_cavern1_haunted_icr_0_1.vlm"},
	{0x52B62B9B081A1137ULL, "557059484.ogg"},
	{0x52B6564E80677145ULL, "dae000_epic_merchant_1.bnk"},
	{0x52B6B579FC2F7412ULL, "vxw_encounter2_bt_0_0.vlm"},
	{0x52B6FDBC033B7470ULL, "755662595.ogg"},
	{0x52B706205483513DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mb.mao"},
	{0x52B70A2054835876ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mb.mmh"},
	{0x52B7142054836974ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mb.msh"},
	{0x52B77236CD6E8911ULL, "6150090_m.fxe"},
	{0x52B780AF44958872ULL, "vxlt_gallowstemplar_5r7_0_0.vlm"},
	{0x52B7AF338E63681FULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.cess.gad"},
	{0x52B80A04061C3FF0ULL, "vxlt_0_lowtown_docks_bc7_0_0.vlm"},
	{0x52B850799AEFC6DAULL, "hlac_lt_0_lowtown_docks_a3_n_65532_0y.dds"},
	{0x52B85ACF616D241FULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench3f_merillintro.ani"},
	{0x52B8C276E978B91CULL, "6113128_m.fxe"},
	{0x52B8CFAFF311B0AAULL, "vxca_caverns1_eaq_0_0.vlm"},
	{0x52B8F33F39CD0574ULL, "art\\characters\\creatures\\stonegolem\\animation\\dblw_dmg_v1.evt"},
	{0x52B90834B631879FULL, "hlac_ca_caverns2prc_5802_13y.dds"},
	{0x52B95C439894940FULL, "vxca_caverns1_mines_egw_0_0.vlm"},
	{0x52B975AA960D2DEBULL, "vrc321ip_spooky6_obj_2.utp"},
	{0x52BA16118A140DBAULL, "553258202.ogg"},
	{0x52BA35AC20DFCF4FULL, "default_mage_male.utc"},
	{0x52BA384E806B277DULL, "dae000_epic_merchant_1.cnv"},
	{0x52BA454FEB3B0179ULL, "vxca_caverns2prc_8oj_0_1.vlm"},
	{0x52BAA10A0FEBE8DBULL, "206339402.ogg"},
	{0x52BAA9C6204E4D30ULL, "vxlt_house_ga_0_0.vlm"},
	{0x52BACACBF8AC9442ULL, "6101460_m.fxe"},
	{0x52BB0D42F261C15EULL, "6176556_m.fxe"},
	{0x52BB9651ADD6E887ULL, "681900803.ogg"},
	{0x52BC12B386819823ULL, "749968642.ogg"},
	{0x52BC9EFDAEE033EAULL, "hlac_ht_0_hightown_a2_-65535_0z.dds"},
	{0x52BCB90E31E2E44EULL, "follower_banter6141447.cl"},
	{0x52BCE0E9392734ACULL, "dae300_petrice6097853.cl"},
	{0x52BCE12BC1F161E1ULL, "6037090_m.fxe"},
	{0x52BD69A07A551A7BULL, "134524002.ogg"},
	{0x52BD80147672C4B9ULL, "lt_0_lowtown_docks_n_21_0_ll.mmh"},
	{0x52BDC645A0DB0D45ULL, "vxdr_0_deeproads_d0s_0_0.vlm"},
	{0x52BE7DBA9A4328B6ULL, "hlac_lt_0_lowtown_a3_n_c_1_0x.dds"},
	{0x52BE9673F0149745ULL, "570626667.ogg"},
	{0x52BE97238DE55C36ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar3_mcr370_playerflees.ani"},
	{0x52BEA4338E69740CULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.cess.evt"},
	{0x52BEB3B3ACF25C68ULL, "gen00fl_fenris6115315.cl"},
	{0x52BEEB97306548F1ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maso.phy"},
	{0x52BF2ABFDA8718DCULL, "hlac_ht_keep_4519_1y.dds"},
	{0x52BF3F162EDC9900ULL, "6065926_m.fxe"},
	{0x52BF5AE05AC1DD68ULL, "keymap.gda"},
	{0x52BF79AA302642E8ULL, "6171920_m.fxe"},
	{0x52BFE286833FD034ULL, "l_mer231ar_arlan_hideout.wbd"},
	{0x52BFF37DE976BEE6ULL, "ss_ambient_female.wbd"},
	{0x52C064BA3289BB63ULL, "w_bonepitmining_a2_16_0_lc.mmh"},
	{0x52C143C908A3F1F5ULL, "159903839.ogg"},
	{0x52C144C68441E4ACULL, "lt_0_lowtown_docks_n_20_0.rml"},
	{0x52C1C486B6C1AE31ULL, "aip_undeadcorpse_base.gda"},
	{0x52C1C68CAA3B9948ULL, "543810797.ogg"},
	{0x52C1F7BD99B354D6ULL, "w_encounter3_221_0.msh"},
	{0x52C201BD99B365D4ULL, "w_encounter3_221_0.mmh"},
	{0x52C205BD99B36CE7ULL, "w_encounter3_221_0.mao"},
	{0x52C214E55400B2FEULL, "art\\characters\\weapons\\greatswords\\w_gsw_004a.phy"},
	{0x52C253E994826EFCULL, "vxht_bartrand_n_69x_0_0.vlm"},
	{0x52C27FE4D01407DFULL, "6089471_m.fxe"},
	{0x52C29A60C2380FB7ULL, "w_bonepitmining_a2_47_0.mmh"},
	{0x52C29E60C23816C4ULL, "w_bonepitmining_a2_47_0.mao"},
	{0x52C2B060C2383529ULL, "w_bonepitmining_a2_47_0.msh"},
	{0x52C2FDD312B377C4ULL, "dae000_arishok6067812.cl"},
	{0x52C3CAAD794C7895ULL, "vxlt_0_lowtown_docks_bby_0_0.vlm"},
	{0x52C3F61DEB8A06B5ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgtb_s.dds"},
	{0x52C4F3C9AB309CA6ULL, "228299217.ogg"},
	{0x52C52756EE258C92ULL, "vxw_encounter2_mu_0_0.vlm"},
	{0x52C565B7B6B2EE93ULL, "6156146_m.fxe"},
	{0x52C666A96556F3D6ULL, "lt_0_lowtown_42_0.rml"},
	{0x52C67BE0F8125870ULL, "vxlt_0_lowtown_docks_bf4_0_0.vlm"},
	{0x52C6EB3F39D934ADULL, "art\\characters\\creatures\\stonegolem\\animation\\dblw_dmg_v1.ani"},
	{0x52C712B641B5A7AEULL, "6059426_m.fxe"},
	{0x52C7A2389688CBAAULL, "6059343_m.fxe"},
	{0x52C7BAD89C971B57ULL, "vxht_keep_3zo_0_0.vlm"},
	{0x52C7F225FD8398CBULL, "hlac_ht_0_hightown_n_65533_1z.dds"},
	{0x52C7FD378AA79234ULL, "6150914_m.fxe"},
	{0x52C802205243187AULL, "vxca_caverns1_mines_hm5_0_0.vlm"},
	{0x52C865099668979EULL, "vxht_keep_5e5_0_0.vlm"},
	{0x52C8A0D78006C2EAULL, "6116110_m.fxe"},
	{0x52C8AAFA0045509FULL, "mag291_demon_marethari6028837.cl"},
	{0x52C8EEA419977F98ULL, "lgt310_mis_burn_dump6180593.cl"},
	{0x52C90FBBB3362631ULL, "dae000_orsino6085425.cl"},
	{0x52C9598A53D0B404ULL, "vxht_mansion2_2l3_0_0.vlm"},
	{0x52C95F225D3E9AEBULL, "static_static_p22.psh.fxo.dx11.assembly"},
	{0x52C9EF64A6968601ULL, "dae000_arishok6068432.cl"},
	{0x52CBA7565A0ADF59ULL, "6092378_m.fxe"},
	{0x52CBAFC74E5BD16DULL, "follower_banter6172282.cl"},
	{0x52CBE0E6A39E278FULL, "temploadhints.gda"},
	{0x52CC01BB916F9941ULL, "6028937_m.fxe"},
	{0x52CC4F96ED031090ULL, "6081201_m.fxe"},
	{0x52CC5A4FF8BCC7F9ULL, "6108891_m.fxe"},
	{0x52CC8115A31F5127ULL, "hlac_w_encounter1_-2_0y.dds"},
	{0x52CC8B32FFCC41FFULL, "287756866.ogg"},
	{0x52CC8C338E758965ULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.cess.ani"},
	{0x52CC957E1EB73C7CULL, "666977029.ogg"},
	{0x52CD43802E610458ULL, "267162516.ogg"},
	{0x52CD50DE5BF0EF82ULL, "01_fire.dds"},
	{0x52CD608685554FD7ULL, "l_lgt220ar_gang_hideout_ht.bnk"},
	{0x52CDBB74C197648DULL, "88548152.ogg"},
	{0x52CDC03840E7D6FFULL, "vxca_caverns2_8cx_0_0.vlm"},
	{0x52CE45A8A58D2DBFULL, "gen00fl_merrill6141207.cl"},
	{0x52CEB152A7716B5DULL, "vxht_keep_3xg_0_0.vlm"},
	{0x52CF0741B3C7F7E9ULL, "vxht_keep_63c_0_0.vlm"},
	{0x52CF321601A12E9EULL, "w_woundedcoast_26_0_ll.mmh"},
	{0x52CF521380D98427ULL, "286704774.ogg"},
	{0x52CF5B8B17AD9142ULL, "vxlt_gallowstemplar_f_46y_0_0.vlm"},
	{0x52CF854849E54BF6ULL, "6125633_m.fxe"},
	{0x52CFEACB5AE97003ULL, "182147943.ogg"},
	{0x52D01D68827C006AULL, "qcr220_meredith6054734.cl"},
	{0x52D02329E00EA917ULL, "6080537_m.fxe"},
	{0x52D0508BD38460DFULL, "llac_lt_0_lowtown_a3_n_c_1_3.dds"},
	{0x52D10768A5DA7FDCULL, "vxlt_gallowstemplar_f_34v_0_0.vlm"},
	{0x52D198E20BCB86F7ULL, "dae000_merchant_rare_em6177944.cl"},
	{0x52D1B371267F72C4ULL, "283595866.ogg"},
	{0x52D1C428DCA3742BULL, "947519916.ogg"},
	{0x52D1DB8F222C66FAULL, "mag120_macha6095528.cl"},
	{0x52D2061620467D49ULL, "293131255.ogg"},
	{0x52D29B0EF9C80337ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basicxend.phy"},
	{0x52D2C0238DF64B70ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar3_mcr370_playerflees.gad"},
	{0x52D2D1EF9D281461ULL, "141187429.ogg"},
	{0x52D2E1DC010FCB2DULL, "one220_yevhen6120154.cl"},
	{0x52D39933E01BA512ULL, "vxw_encounter2_hv_0_0.vlm"},
	{0x52D439A39A59FDF4ULL, "hlac_lt_0_lowtown_n_-65534_0x.dds"},
	{0x52D43D16DC9425F1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crnf.ani"},
	{0x52D470DF8297BEF3ULL, "vxlt_hangedman_n_1vp_0_0.vlm"},
	{0x52D494078F494F1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_approach.evt"},
	{0x52D4E4A7D37D6AFEULL, "131058327.ogg"},
	{0x52D55C315BF5501AULL, "hlac_lt_0_lowtown_n_-1_1y.dds"},
	{0x52D583676C3309E8ULL, "isa311_isabela6092313.cl"},
	{0x52D5C1A4597F233EULL, "gencr_gate_slave_l1.utc"},
	{0x52D5D9F4A33E1855ULL, "dae000_bodhan_letter6144325.cl"},
	{0x52D5FB898F93C9E8ULL, "mag360_samson6042068.cl"},
	{0x52D60C5154D3B4BDULL, "hlac_w_sundermount_0_6z.dds"},
	{0x52D6592D11FC6BDAULL, "llac_lt_gallowsprison_n_-1_12.dds"},
	{0x52D6923AE9292C1FULL, "w_encounter1_51_0.msh"},
	{0x52D6983AE92936C1ULL, "w_encounter1_51_0.mmh"},
	{0x52D69875B68A943BULL, "dae000_flemeth6013322.cl"},
	{0x52D6A43AE9294B62ULL, "w_encounter1_51_0.mao"},
	{0x52D713A75A64B377ULL, "hightown_male_noble_gossip6171296.cl"},
	{0x52D739BF9A391046ULL, "vxw_bonepitmining_a2_1ny_0_0.vlm"},
	{0x52D76B11CCA959D8ULL, "vxht_bartrand_n_5ph_0_0.vlm"},
	{0x52D7EF2DEBC1966DULL, "6103693_player_exaggerate.cut"},
	{0x52D7EF9CB58F6E89ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyn_0n.dds"},
	{0x52D816ADFDA97625ULL, "vxlt_gallowstemplar_n_5k8_0_0.vlm"},
	{0x52D83EC690924AA9ULL, "gen_im_npc_hum_arm_hlm_chna.uti"},
	{0x52D89442E834CA70ULL, "zz_vault_debug6200333.cl"},
	{0x52D89DF299223FD9ULL, "vxdr_primevalentrance_39k_0_2.vlm"},
	{0x52D8B74EA67CD3E9ULL, "vxw_encounter1_jc_0_1.vlm"},
	{0x52D95035A8A64D53ULL, "llac_lt_0_lowtown_a2_n_65534_0.dds"},
	{0x52D9DBD3BCC632E9ULL, "hlac_ht_0_hightown_a3_65534_0z.dds"},
	{0x52D9F0FD06633D2BULL, "6089413_m.fxe"},
	{0x52DA105F9BF7BD8FULL, "l_env_wal_lowtown_int_calm_p2p.wbd"},
	{0x52DA3FCE1DB3DB10ULL, "vfx_env_dust_02.mao"},
	{0x52DA649BD13C77AEULL, "hlac_ht_0_hightown_a3_-65536_1y.dds"},
	{0x52DAAB078F4DE249ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_approach.gad"},
	{0x52DAADBAE0599087ULL, "lgt101cr_mag_escp_terrie.utc"},
	{0x52DAC60EEBFE9D0AULL, "lt_0_lowtown_docks_a3_n_8_0_lc.mmh"},
	{0x52DB2E925E4A7851ULL, "zz_ave_debug6056690.cl"},
	{0x52DB2F0DDBC1B2D0ULL, "249343209.ogg"},
	{0x52DB4816DC9A5693ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crnf.gad"},
	{0x52DB7F41C39FDE18ULL, "vxht_brothel_n_6f1_0_0.vlm"},
	{0x52DBB37E29B4EC55ULL, "57481633.ogg"},
	{0x52DBBE831E668A65ULL, "art\\vfx\\combat\\vfx_w_wns_front_clear_c.mmh"},
	{0x52DBE84E329DD904ULL, "192467106.ogg"},
	{0x52DC0679F20018C5ULL, "character_character_p136.psh.fxo.dx11"},
	{0x52DD0312D5730DEAULL, "vxlt_0_lowtown_a2_g93_0_0.vlm"},
	{0x52DDDCB4A12DEF92ULL, "hlac_lt_undercitydungeon_-1_25z.dds"},
	{0x52DDE7093825FE3CULL, "ht_bartrand_n_3369.anb"},
	{0x52DDF7F21FFA023EULL, "272895391.ogg"},
	{0x52DE06643B55B2B7ULL, "6153857_m.fxe"},
	{0x52DE16A1C462F093ULL, "6087176_m.fxe"},
	{0x52DF4085450DB3A7ULL, "493276035.ogg"},
	{0x52DF639EB40038DFULL, "6087671_m.fxe"},
	{0x52E02643145197D5ULL, "isa311_isabela6092242.cl"},
	{0x52E0380F72CC819CULL, "mcr310_argument6074478.cl"},
	{0x52E0753B1BE7DCF2ULL, "gen00fl_carver6104887.cl"},
	{0x52E12A13CC57688AULL, "723165150.ogg"},
	{0x52E17C078F53B097ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_approach.ani"},
	{0x52E24516DCA070F8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_crnf.evt"},
	{0x52E24579AC27E30FULL, "6066498_m.fxe"},
	{0x52E279B177A1F36CULL, "isa311_isabela6095327.cl"},
	{0x52E28FAA4128F21FULL, "381639115.ogg"},
	{0x52E2BA21ADA42A49ULL, "69502870.ogg"},
	{0x52E2BD9C6ECD0710ULL, "llac_lt_darktown_-1_51.dds"},
	{0x52E2D327F7981532ULL, "6146977_m.fxe"},
	{0x52E3A64243DFED2AULL, "403291070.ogg"},
	{0x52E4EB94D0A5EC8FULL, "6120594_m.fxe"},
	{0x52E682A1190EE206ULL, "gen00fl_aveline6114412.cl"},
	{0x52E69213C03E32E5ULL, "gen00fl_anders6128390.cl"},
	{0x52E6E659BD6C2EBCULL, "scaleform_sf_main_p2.psh.fxo"},
	{0x52E7085282D88B45ULL, "883574792.ogg"},
	{0x52E735AFBEF2D382ULL, "436655294.ogg"},
	{0x52E75C15C2608304ULL, "mcr310cr_templar_loyalist.utc"},
	{0x52E783B4A4BA9573ULL, "6109010_m.fxe"},
	{0x52E79316CA6A027EULL, "dae000_gamlen6005763.cl"},
	{0x52E84D038947431EULL, "vxht_0_hightown_c0l_0_0.vlm"},
	{0x52E86183AE6CAAF5ULL, "hlac_ht_0_hightown_65532_0y.dds"},
	{0x52E8E20C899E870DULL, "llac_dr_primevalentrance_1726_9.dds"},
	{0x52E8F1741107254BULL, "hlac_w_encounter2_131071_0x.dds"},
	{0x52E98ABD9723CEF5ULL, "vxlt_0_lowtown_a3_n_c_d4t_0_0.vlm"},
	{0x52E9A06D85ED4298ULL, "ravenscare.mao"},
	{0x52E9BB831E72C232ULL, "art\\vfx\\combat\\vfx_w_wns_front_clear_c.anb"},
	{0x52EA07F3E898C797ULL, "404144264.ogg"},
	{0x52EA1BDC810B6AEDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_lute_0.pwk"},
	{0x52EA4F83E4DD679AULL, "plt_and200pt_opening.nss"},
	{0x52EA5F1AABD6298EULL, "vrc321tr_spooky_scenes_6.utt"},
	{0x52EB7762641CD748ULL, "art\\levels\\races\\ferelden\\props\\_textures\\brazier_sa.dds"},
	{0x52EBB30631EE53A3ULL, "vxdr_0_deeproads_err_0_0.vlm"},
	{0x52EBD0868364ECCEULL, "l_mer231ar_arlan_hideout.bnk"},
	{0x52EC7CB0D57F04C4ULL, "vxht_bartrand_n_5sm_0_0.vlm"},
	{0x52ED519027BB6C9DULL, "6027693_m.fxe"},
	{0x52ED8C98ABF10434ULL, "lgt220_shp_mag_guard6149504.cl"},
	{0x52ED96012CDA5C26ULL, "vxlt_foundry_n_zp_0_0.vlm"},
	{0x52EDE7EC432A4582ULL, "644333511.ogg"},
	{0x52EE62598DEEF77CULL, "6100087_m.fxe"},
	{0x52EE8E4113F4D237ULL, "dae000_arishok6068579.cl"},
	{0x52EEEE6BED2CA615ULL, "vxca_cavern1_haunted_i7i_0_0.vlm"},
	{0x52EF347EB7960D85ULL, "6174993_m.fxe"},
	{0x52EF3BA26D60C58BULL, "pro000_merchant6051609.cl"},
	{0x52EF54C39DF5F1B8ULL, "llac_dr_0_deeproads_5849_6.dds"},
	{0x52EF9F5BE94E8819ULL, "226744521.ogg"},
	{0x52EFB66159E34589ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\vulnerablefromopen.ani"},
	{0x52EFD27B3082CECAULL, "17219822.ogg"},
	{0x52EFF7AB4C39761DULL, "6167597_m.fxe"},
	{0x52F0A07930C0D1C6ULL, "ctplt_sm_rock.mao"},
	{0x52F154996683EFF3ULL, "vxlt_hangedman_n_1hr_0_0.vlm"},
	{0x52F2189864A0A554ULL, "6174967_m.fxe"},
	{0x52F229FBCA2E420CULL, "sst_ss_ambient_f.gda"},
	{0x52F28444367459DEULL, "103483243.ogg"},
	{0x52F2C60C24CE7810ULL, "vxht_mansion2_3bx_0_0.vlm"},
	{0x52F3B855ABF1B829ULL, "gen_im_trs_rdr_med_hlm_r2.uti"},
	{0x52F4A7035C51CC21ULL, "vxlt_gallowstemplar_46y_0_0.vlm"},
	{0x52F4C1544A9E282EULL, "ui_messages.gda"},
	{0x52F50520ECEBD48DULL, "dre180ip_primeval_door.utp"},
	{0x52F58C998EB6D91CULL, "vxca_caverns2_7ec_0_1.vlm"},
	{0x52F5A1DDF112963BULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun.msh"},
	{0x52F5B3DDF112B486ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun.mao"},
	{0x52F5BFDDF112C925ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun.mmh"},
	{0x52F5D03DC1B5687DULL, "hlac_lt_0_lowtown_docks_a3_n_-131071_0y.dds"},
	{0x52F6091B2A215A19ULL, "vxlt_0_lowtown_docks_n_cw4_0_0.vlm"},
	{0x52F68B78BA7244F4ULL, "lt_0_lowtown_a2_n_c_21_0.rml"},
	{0x52F6BEE07B3D5B9EULL, "vxht_0_hightown_c7m_0_0.vlm"},
	{0x52F6D16159E9920BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\vulnerablefromopen.gad"},
	{0x52F744F0DEF8EC76ULL, "zz_dae_debug6029327.cl"},
	{0x52F78AC820FEDE1BULL, "813586363.ogg"},
	{0x52F892C7ACDF7364ULL, "gui_reticle_p0.psh.fxo.dx11.assembly"},
	{0x52F8B0885774D8FBULL, "w_bonepitmining_a2_118_0.phy"},
	{0x52F8E3506E7C5781ULL, "6076228_m.fxe"},
	{0x52F915CF16CA207FULL, "vxca_caverns2_8pe_0_0.vlm"},
	{0x52F923E799AFEB94ULL, "d15_b_sp_d.dds"},
	{0x52F933C2048FC341ULL, "dae000_merchant_armor_hm6164283.cl"},
	{0x52F94BA8DFF4C1D5ULL, "one300_zevran6089118.cl"},
	{0x52F992121E3ABA74ULL, "gen00fl_sebastian6179790.cl"},
	{0x52F9C2E2F5BF3779ULL, "vxdr_primevalentrance_26i_0_0.vlm"},
	{0x52FA06E97D53B4DDULL, "vxht_keep_3id_0_0.vlm"},
	{0x52FA8629BD5E8726ULL, "20154729.ogg"},
	{0x52FA934AC752DECFULL, "221276373.ogg"},
	{0x52FA9E286AE1D68AULL, "1038368232.ogg"},
	{0x52FB17F95DBEF6E9ULL, "qcr000_arishok6060199.cl"},
	{0x52FBF2EE62F9263CULL, "139540019.ogg"},
	{0x52FC7ABFE1860FD2ULL, "505520928.ogg"},
	{0x52FC8EFF1CC01581ULL, "vxlt_0_lowtown_docks_n_bbn_0_0.vlm"},
	{0x52FCC7871E174152ULL, "673036995.ogg"},
	{0x52FCEBBB2BBB56C7ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.cess.gad"},
	{0x52FD0A92A46CEFC6ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.impa.ani"},
	{0x52FD34686BB90FE8ULL, "lgt000cr_hate_guild.utc"},
	{0x52FDBEA3E5332B94ULL, "539029941.ogg"},
	{0x52FDCBD7464D5978ULL, "store_and311cr_lirene.utm"},
	{0x52FDCE6159EFAC70ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\vulnerablefromopen.evt"},
	{0x52FE26E02138B63FULL, "vxlt_0_lowtown_docks_a3_n_cy6_0_1.vlm"},
	{0x52FE64A5027D6B5AULL, "ach_exp_codex_plot.dds"},
	{0x52FE7D18EFB92366ULL, "cor_bld_clo.mao"},
	{0x52FEF6573843209FULL, "6195147_m.fxe"},
	{0x52FF18F47B459C9DULL, "w_bonepitdestroyed_a3_183_0.mao"},
	{0x52FF1CF47B45A3D6ULL, "w_bonepitdestroyed_a3_183_0.mmh"},
	{0x52FF26F47B45B4D4ULL, "w_bonepitdestroyed_a3_183_0.msh"},
	{0x52FF5FE7CAEA2826ULL, "vxlt_gallowstemplar_f_5qy_0_0.vlm"},
	{0x52FF74441A6605CEULL, "449731532.ogg"},
	{0x52FF9A9E10113635ULL, "vxw_bonepitdestroyed_a3_mv_0_0.vlm"},
	{0x5300114CF772B764ULL, "29373796.ogg"},
	{0x530063A96588668AULL, "lt_0_lowtown_42_0.anb"},
	{0x530097706F8896BCULL, "dae000_worthy6166621.cl"},
	{0x5300EE5CAEF1C29EULL, "vxht_mansion2_2eo_0_1.vlm"},
	{0x53017E0B78276E7FULL, "6059676_m.fxe"},
	{0x53023F24958C0BAAULL, "hlac_ca_caverns2prc_5802_24z.dds"},
	{0x53024141BDF40A9BULL, "vxht_mansion_n_5bj_0_0.vlm"},
	{0x5302521AABEAA5B7ULL, "vrc321tr_spooky_scenes_6.ncs"},
	{0x530271D44A5C8EEFULL, "6131901_m.fxe"},
	{0x5302993AE94E733EULL, "w_encounter1_51_0.phy"},
	{0x5302AC7F12D7CCDAULL, "dae000_merchant_generic_hm6172588.cl"},
	{0x5302C0A8A28BA358ULL, "vxht_bartrand_n_5ag_0_0.vlm"},
	{0x5303E0BB2BC162D4ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.cess.evt"},
	{0x5303E8069FC6081CULL, "mag311_keran6063753.cl"},
	{0x5304009E2F2ED254ULL, "hlac_lt_0_lowtown_n_-2_5z.dds"},
	{0x530445156834989DULL, "llac_lt_gallowstemplar_2383_10.dds"},
	{0x5304E9B9E87600F7ULL, "vxca_caverns1_mines_in4_0_0.vlm"},
	{0x53058DB748B30DE5ULL, "6149891_f.fxe"},
	{0x5305B0DB4E639D7FULL, "6188646_m.fxe"},
	{0x5306B97C1C0FB91DULL, "vxht_brothel_n_6co_0_0.vlm"},
	{0x5306C015301D0D4FULL, "ave200_donnic6111855.cl"},
	{0x5307410FD6E28CB7ULL, "vxht_0_hightown_b00_0_0.vlm"},
	{0x53076AA8E7C8ED82ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_p.impa.gad"},
	{0x530781F9AE0588BBULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_monique.ani"},
	{0x53078A6F8E0C3036ULL, "107573225.ogg"},
	{0x53078BB338B805C5ULL, "and341ip_loot_corpse_2.utp"},
	{0x5307B1AF648DB475ULL, "one100_ht_banter6060905.cl"},
	{0x5308D26A5ADFAD25ULL, "vxlt_0_lowtown_docks_n_bdp_0_0.vlm"},
	{0x530911863BD6138BULL, "vxht_0_hightown_n_9ao_0_0.vlm"},
	{0x530948C56B081C74ULL, "6105589_m.fxe"},
	{0x53095E316F876968ULL, "hlac_lt_0_lowtown_a2_n_-1_4z.dds"},
	{0x530967314CD61DA0ULL, "follower_banter6143353.cl"},
	{0x5309CA3BF66F1129ULL, "6027455_m.fxe"},
	{0x5309DFCA211796C8ULL, "hlac_ca_cavern1_haunted_24421_7z.dds"},
	{0x5309FDAB70C77086ULL, "6105846_m.fxe"},
	{0x530A4BF200810D8DULL, "6094295_m.fxe"},
	{0x530A51C2B1906F7CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_amllcrstnew_d.dds"},
	{0x530A68DC663AF615ULL, "vxw_bonepitdestroyed_a3_1aq_0_0.vlm"},
	{0x530A7398537A36EBULL, "qcr220_meredith.bnk"},
	{0x530A8A91F052CDC8ULL, "6034740_m.fxe"},
	{0x530ADB38EDBCDC56ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsstair.msh"},
	{0x530AE538EDBCED54ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsstair.mmh"},
	{0x530AE938EDBCF467ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsstair.mao"},
	{0x530AF292A4790437ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.impa.evt"},
	{0x530B01FEE623E23DULL, "hlac_w_bonepit_196610_0z.dds"},
	{0x530B42581346A75FULL, "41755007.ogg"},
	{0x530B4C63BD9519DDULL, "vxdr_primevaltomb_50x_0_0.vlm"},
	{0x530BC8759B19A837ULL, "6142455_m.fxe"},
	{0x530C09AD83C4FEF5ULL, "w_bonepitmining_a2_225_0.msh"},
	{0x530C0BAD83C5026BULL, "w_bonepitmining_a2_225_0.mmh"},
	{0x530C17AD83C51608ULL, "w_bonepitmining_a2_225_0.mao"},
	{0x530C2FD73D0D8B39ULL, "w_bonepit_164_0.msh"},
	{0x530C39D73D0D9C27ULL, "w_bonepit_164_0.mmh"},
	{0x530C3DD73D0DA374ULL, "w_bonepit_164_0.mao"},
	{0x530C44BDF138A657ULL, "vxlt_warehouse_n_1a1_0_0.vlm"},
	{0x530C6783A49DD30EULL, "22413277.ogg"},
	{0x530C77AA42D1D87BULL, "hf_vrc3_housekeeper.mor"},
	{0x530C9D83D346D9CFULL, "character_character_p91.psh.fxo.dx11.assembly"},
	{0x530D359365750832ULL, "642522492.ogg"},
	{0x530DDA2CD8F29F30ULL, "art\\vfx\\combat\\vfx_g_spear_att_back_c.mmh"},
	{0x530DE2F3E512EA1FULL, "363425004.ogg"},
	{0x530E14AFCCDD8161ULL, "ave121_brennan6105916.cl"},
	{0x530E4341C4D3CB5BULL, "vxlt_undercityencounter_1hj_0_0.vlm"},
	{0x530E520BE90D02FAULL, "lt_0_lowtown_docks_a3_166_0.msh"},
	{0x530E600BE90D1A33ULL, "lt_0_lowtown_docks_a3_166_0.mao"},
	{0x530E640BE90D2140ULL, "lt_0_lowtown_docks_a3_166_0.mmh"},
	{0x530E7598537E22CBULL, "qcr220_meredith.cnv"},
	{0x530E78A79A1975DCULL, "6113916_f.fxe"},
	{0x530E808D293146BBULL, "ht_mansion2_818_lc.mmh"},
	{0x530E8DA8E7CF4879ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_p.impa.evt"},
	{0x530F12DB21EB00FAULL, "vxht_0_hightown_a3_c8u_0_0.vlm"},
	{0x530F49C19A1584ECULL, "follower_banter6163610.cl"},
	{0x530FA739A1E11CF1ULL, "6070351_f.fxe"},
	{0x530FD6B0EBCBF0B6ULL, "410295647.ogg"},
	{0x530FF87872F79483ULL, "448120749.ogg"},
	{0x53102CD83BE1B3FFULL, "vxlt_gallowstemplar_5ao_0_0.vlm"},
	{0x53104BA6064BB5EDULL, "6104449_f.fxe"},
	{0x5310927EE703635BULL, "596096462.ogg"},
	{0x5310D8BB2BCBDF0DULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.cess.ani"},
	{0x531114F0E6AC3352ULL, "nat320cr_delilah.utc"},
	{0x53114DFD43C7B2EDULL, "follower_banter6171939.cl"},
	{0x5311664307C304CFULL, "ht_mansion2_mmap_ph.dds"},
	{0x531170AE7A548CDDULL, "mcr370_assignments6081683.cl"},
	{0x53119BAA4489A3F8ULL, "vxht_chantry_n_2bv_0_0.vlm"},
	{0x5311F392A47F24C0ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.impa.gad"},
	{0x5311FE8E3C5AB3F1ULL, "6171891_m.fxe"},
	{0x531263CB0186A5E1ULL, "hlac_lt_0_lowtown_a2_-65536_5y.dds"},
	{0x53129093F12FA9F3ULL, "dre000_varric6208004.cl"},
	{0x5312B5E1B236F177ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyn_0t.dds"},
	{0x53133081FCA9C082ULL, "773735477.ogg"},
	{0x53134B936DADE842ULL, "mag291_vincento6027377.cl"},
	{0x53135B96DA6C1977ULL, "vxca_caverns1_mines_ik7_0_0.vlm"},
	{0x5313BDA11C7670EFULL, "rdr110dg_party_comments6094676.cl"},
	{0x5313C1DFC56125EAULL, "6198130_m.fxe"},
	{0x5313FDD4B6E7D920ULL, "6156552_m.fxe"},
	{0x531426080BD13366ULL, "6111278_m.fxe"},
	{0x5314457671FFD75AULL, "hlac_w_blightlands_65532_0x.dds"},
	{0x53145469D95DA651ULL, "6167311_m.fxe"},
	{0x531482F9C1E0E55FULL, "vxlt_gallowstemplar_f_2jb_0_0.vlm"},
	{0x5315ADA31188746DULL, "lt_0_lowtown_a3_n_c_53_0_lc.mmh"},
	{0x5315AE73D138B3B2ULL, "vxht_0_hightown_a3_57f_0_0.vlm"},
	{0x5315C50BEC63FD69ULL, "280510632.ogg"},
	{0x5315CBAD098DDA1EULL, "art\\vfx\\combat\\vfx_c_rw_teleport_in_c.impa.evt"},
	{0x5315F0334E5A84FFULL, "vxht_0_hightown_buc_0_0.vlm"},
	{0x5315F988FAB6517FULL, "vxca_caverns2_8zg_0_0.vlm"},
	{0x5315FE8E5ACDB04DULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_05brgdebris.msh"},
	{0x5316047B4C06BE90ULL, "6106452_m.fxe"},
	{0x53160C8E5ACDC780ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_05brgdebris.mao"},
	{0x5316108E5ACDCED3ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_05brgdebris.mmh"},
	{0x53164D26DA7EAD35ULL, "309731532.ogg"},
	{0x53167BADFBCBCA69ULL, "dre000_varric6035145.cl"},
	{0x53169F76CDB41C0DULL, "vxht_mansion_n_6ot_0_0.vlm"},
	{0x5316D85250181F88ULL, "vxca_caverns1_mines_jfx_0_0.vlm"},
	{0x5317444201A017F8ULL, "hlac_w_bonepitmining_a2_327683_0y.dds"},
	{0x5317C2B293737949ULL, "plt_prm000pt_air_ring.nss"},
	{0x5317D5ECFF0271BAULL, "w_encounter3_66_0.phy"},
	{0x5317DF507511E3FDULL, "vxlt_darktown_425_0_0.vlm"},
	{0x5317F9EF57139395ULL, "dae000_hate_follower_react6167492.cl"},
	{0x53185C6DB72BB83CULL, "6180379_m.fxe"},
	{0x5318871CC45298B6ULL, "vxht_0_hightown_c31_0_0.vlm"},
	{0x5318AE7F097B4A31ULL, "6144096_f.fxe"},
	{0x5319046D954417E5ULL, "art\\vfx\\combat\\vfx_c_rw_attack1_c.anb"},
	{0x531935F1813A977FULL, "vxht_0_hightown_a2_bk1_0_0.vlm"},
	{0x53196F01555CFF3FULL, "vfx_vfx_p2.vsh.fxo.dx11.assembly"},
	{0x5319700EFA03BC8FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basicxend.mao"},
	{0x53197C0EFA03D12CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basicxend.mmh"},
	{0x5319820EFA03DB1EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rr_basicxend.msh"},
	{0x53198236D2EE3F6FULL, "6139268_m.fxe"},
	{0x5319A7C86CAB1D47ULL, "gen00fl_anders6125409.cl"},
	{0x5319F7705072CDADULL, "llac_w_woundedcoast_65542_0.dds"},
	{0x531A2A5F9C2DCE65ULL, "l_env_wal_lowtown_int_calm_p2p.bnk"},
	{0x531A72473743596CULL, "6094957_m.fxe"},
	{0x531A916E6B9CB1AFULL, "6089848_m.fxe"},
	{0x531AFD2CD8FD654FULL, "art\\vfx\\combat\\vfx_g_spear_att_back_c.anb"},
	{0x531B78BD2BEA43C6ULL, "zz_vault_debug6178665.cl"},
	{0x531BA8F9AE167445ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_monique.gad"},
	{0x531BE2AD09926D49ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_in_c.impa.gad"},
	{0x531C0EC1F902986EULL, "vxlt_0_lowtown_a2_n_e89_0_0.vlm"},
	{0x531C24406FDFDAFCULL, "gen00fl_anders6119887.cl"},
	{0x531C66362994319BULL, "vxlt_hangedman_1qj_0_0.vlm"},
	{0x531C95A8E7DB9368ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_p.impa.ani"},
	{0x531CB51EB6C9E3DBULL, "vxlt_undercityencounter_1em_0_0.vlm"},
	{0x531D2692C0CFA5B6ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.dura.evt"},
	{0x531D406AF14C42F9ULL, "6144781_m.fxe"},
	{0x531DE2A6A84F0CDAULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_wkf.gad"},
	{0x531E0C92E27A8C61ULL, "6099326_m.fxe"},
	{0x531E191693D73571ULL, "6075051_f.fxe"},
	{0x531E2791822DA65AULL, "6119339_m.fxe"},
	{0x531E88A2BB82EBC0ULL, "vxlt_0_lowtown_docks_b8z_0_0.vlm"},
	{0x531E9EB52DD7CC7FULL, "vxlt_gallowstemplar_5jn_0_0.vlm"},
	{0x531EC43E41D8E0AFULL, "6156898_m.fxe"},
	{0x531EE609385D93CCULL, "ht_bartrand_n_3369.rml"},
	{0x531F24FBA1E03B9BULL, "vxht_mansion_n_5hl_0_0.vlm"},
	{0x531FA4628677D458ULL, "vxht_bartrand_n_5oe_0_0.vlm"},
	{0x531FC7B17F8D5969ULL, "6137563_m.fxe"},
	{0x531FC8F9E782922AULL, "6177410_m.fxe"},
	{0x53206ABE1E5D6B09ULL, "hlac_w_woundedcoast_327688_0x.dds"},
	{0x5320CDEA1C217DBCULL, "6201607_m.fxe"},
	{0x5320D1B3E613AC91ULL, "mag131_keran6086171.cl"},
	{0x5320F13F60990B5AULL, "264456625.ogg"},
	{0x5321040AD1D342D6ULL, "lowtown_male_urchin_sta6164192.cl"},
	{0x5321263141FCDC9BULL, "6083662_m.fxe"},
	{0x53212DE08D800993ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelstack.phy"},
	{0x5321D95E6CF960ACULL, "6027592_f.fxe"},
	{0x5321F6CCA5B0A7BFULL, "438775876.ogg"},
	{0x5322AC78BA97B7F4ULL, "lt_0_lowtown_a2_n_c_21_0.anb"},
	{0x5322B3AD09983B97ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_in_c.impa.ani"},
	{0x5322F4D6B3D6FDDFULL, "vxlt_gallowstemplar_f_2om_0_0.vlm"},
	{0x53235B2E77496FEEULL, "vxca_caverns1_mines_j5g_0_0.vlm"},
	{0x53239D248D1F457BULL, "6153413_m.fxe"},
	{0x53239D35FFFBEA1DULL, "vxht_brothel_n_6im_0_0.vlm"},
	{0x53241D92C0D5B541ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.dura.gad"},
	{0x53242E34082806B3ULL, "6152339_m.fxe"},
	{0x53247EF2DC793A5AULL, "vxht_brothel_n_5dv_0_0.vlm"},
	{0x53248950F50DF8AFULL, "6110588_m.fxe"},
	{0x532497885799E89AULL, "w_bonepitmining_a2_118_0.msh"},
	{0x5324A588579A00D3ULL, "w_bonepitmining_a2_118_0.mao"},
	{0x5324A988579A07E0ULL, "w_bonepitmining_a2_118_0.mmh"},
	{0x5324B2053FF93BC1ULL, "6121256_m.fxe"},
	{0x5324DE60484C7716ULL, "mag100_cullen6082257.cl"},
	{0x5324E5A6A85530F1ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_wkf.evt"},
	{0x532555485BF1266AULL, "qun120cr_javaris6042841.cl"},
	{0x53258D9F95DC8F4DULL, "vxw_bonepitdestroyed_a3_4y_0_0.vlm"},
	{0x5325FF9CE15CD2BCULL, "455209443.ogg"},
	{0x532636CDDC28E5FDULL, "mag210_arianni6025025.cl"},
	{0x5326512D6707FC7DULL, "vxlt_darktown_494_0_0.vlm"},
	{0x53265CAE52BDDE97ULL, "vxht_mansion2b_2hp_0_0.vlm"},
	{0x532677EB61693035ULL, "hlac_lt_hangedman_n_-1_7z.dds"},
	{0x53267B1932E40576ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_water.msh"},
	{0x5326851932E41674ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_water.mmh"},
	{0x5326891932E41C87ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_water.mao"},
	{0x5326F8F9B648B136ULL, "vxht_0_hightown_c62_0_0.vlm"},
	{0x5327306B650DFD28ULL, "567519225.ogg"},
	{0x53278679D3BD7584ULL, "326628362.ogg"},
	{0x5327A7CE7330AFFFULL, "vxht_0_hightown_a2_bl0_0_0.vlm"},
	{0x5327D9AEAACACBA0ULL, "vxlt_gallowstemplar_f_26e_0_0.vlm"},
	{0x53281C79926B9F37ULL, "vxht_brothel_n_4ck_0_0.vlm"},
	{0x5328A410B4E30168ULL, "hlac_lt_0_lowtown_a2_n_131072_4x.dds"},
	{0x5328C726A136A2BFULL, "hlac_ht_0_hightown_n_-65537_0x.dds"},
	{0x5328E93957764BB3ULL, "972025242.ogg"},
	{0x53290855058DC4BDULL, "6117280_m.fxe"},
	{0x53290C2CBEC451F9ULL, "6041785_m.fxe"},
	{0x5329B81B0E8C6191ULL, "vxht_mansion_n_62g_0_0.vlm"},
	{0x5329B9414B774D55ULL, "llac_lt_0_lowtown_n_65536_11.dds"},
	{0x532A8DB0B814240EULL, "vxca_caverns1_inp_0_0.vlm"},
	{0x532ABA4E0C8A21B7ULL, "vxdr_primevaltomb_3dh_0_0.vlm"},
	{0x532B1E92C0DBD5EFULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.dura.ani"},
	{0x532B51AD236FE9B8ULL, "hlac_lt_0_lowtown_docks_a3_n_65537_0x.dds"},
	{0x532B541FF4FC6A75ULL, "dae000_gamlen6005818.cl"},
	{0x532BF363BEF97114ULL, "lgt000ip_at_hybris.utp"},
	{0x532BFDDE559418D0ULL, "ef_variel.mor"},
	{0x532C05E578680FBFULL, "6132734_f.fxe"},
	{0x532C4C4E1FF11CAFULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.mmh"},
	{0x532C5B1D8BFA8C54ULL, "vxht_0_hightown_9zn_0_0.vlm"},
	{0x532C996E7423BEDAULL, "6114439_m.fxe"},
	{0x532C9F97E22E07FBULL, "419560857.ogg"},
	{0x532CA7B31C49275AULL, "6154260_m.fxe"},
	{0x532D10921FCDE4FFULL, "vxlt_gallowstemplar_5wi_0_0.vlm"},
	{0x532D1B9909B07E4EULL, "ser270_emeric6149333.cl"},
	{0x532D4A7F9599FFFDULL, "6018673_m.fxe"},
	{0x532DC239F66625DBULL, "53995930.ogg"},
	{0x532DCD545B69DF15ULL, "mag291cr_crowd_elf_5.utc"},
	{0x532E84284D49C521ULL, "462045284.ogg"},
	{0x532EF3C7FC272C5DULL, "dre000_varric6150374.cl"},
	{0x532F5FF1A4FF2917ULL, "39675937.ogg"},
	{0x532F98CBCB155548ULL, "hlac_lt_0_lowtown_docks_n_-65540_1y.dds"},
	{0x53300D851D7E46CFULL, "llac_lt_warehouse_n_-65536_1.dds"},
	{0x5330B81D6A8626A9ULL, "vxht_0_hightown_a3_a7u_0_0.vlm"},
	{0x5330DC47931AE9F2ULL, "vxca_caverns2_817_0_0.vlm"},
	{0x533146347A633DDBULL, "6019016_m.fxe"},
	{0x53316272CA2A6B6DULL, "vxht_mansion2b_35l_0_0.vlm"},
	{0x5331FDA6A85FE3C0ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_wkf.ani"},
	{0x53320B15991DF734ULL, "6005397_m.fxe"},
	{0x5332820DF1A43BEFULL, "vxlt_0_lowtown_docks_a3_cyd_0_0.vlm"},
	{0x53328C11CA78CB98ULL, "vxca_caverns2_7o5_0_1.vlm"},
	{0x5332BA2A6ED94A30ULL, "qun131_taskmaster6019212.cl"},
	{0x53334062E7F3C603ULL, "1034250224.ogg"},
	{0x5333C6464EF21587ULL, "mag100_thrask6031977.cl"},
	{0x5334CEB8EA367FA3ULL, "1555230.ogg"},
	{0x5334F417C77EE92EULL, "dre000_bodahn6031330.cl"},
	{0x5336191C4CDE3B92ULL, "art\\vfx\\materials\\vfx_distor_si_blend.mao"},
	{0x533646CB590C2EDCULL, "gen_im_trs_cot_lgt_glv_r0.uti"},
	{0x53365F6D8D9888A4ULL, "6170325_m.fxe"},
	{0x53368E568461B7B7ULL, "vxht_brothel_n_4hj_0_0.vlm"},
	{0x53370C9A54D955AEULL, "791706718.ogg"},
	{0x533779485981D9DDULL, "hlac_lt_0_lowtown_n_65536_5y.dds"},
	{0x5337D886F43ABDAFULL, "6140549_m.fxe"},
	{0x533818237673601BULL, "250219965.ogg"},
	{0x533830C535C411D9ULL, "hlac_dr_dungeon_3217_12y.dds"},
	{0x5338B9810F1624CBULL, "vxht_mansion2b_2jj_0_0.vlm"},
	{0x5338CCFE71A8C770ULL, "mag160_feynriel6020447.cl"},
	{0x5338DAF63C0F5AF8ULL, "crt_cycle_5.dds"},
	{0x53390D5DEA14FE03ULL, "889222130.ogg"},
	{0x53395AA367D2B666ULL, "plt_lgt000pt_ultra_4_02.nss"},
	{0x533A27C3C0E1281CULL, "6081692_m.fxe"},
	{0x533A97C35732C636ULL, "llac_dr_0_deeproads_5842_14.dds"},
	{0x533AF733FDC5EB6DULL, "230317533.ogg"},
	{0x533B13A9FEBD0AAFULL, "vxlt_0_lowtown_n_fep_0_0.vlm"},
	{0x533B32ED0C5244A4ULL, "lt_0_lowtown_docks_n_157_0.mmh"},
	{0x533B36ED0C524BF7ULL, "lt_0_lowtown_docks_n_157_0.mao"},
	{0x533B39CC23CF5F4CULL, "6207784_m.fxe"},
	{0x533B48ED0C526A46ULL, "lt_0_lowtown_docks_n_157_0.msh"},
	{0x533B5154D5FA3E5CULL, "508923845.ogg"},
	{0x533B6C569B997727ULL, "129472462.ogg"},
	{0x533B826F11C3FD7FULL, "vxlt_gallowstemplar_5ph_0_0.vlm"},
	{0x533BA6ED0FFFEF30ULL, "hlac_ca_cavern1_haunted_24425_11y.dds"},
	{0x533BD3775A575071ULL, "386475299.ogg"},
	{0x533BDA2D055A67DDULL, "vxw_bonepit_yz_0_0.vlm"},
	{0x533C73A504B2DBAFULL, "llac_ht_0_hightown_a2_0_0.dds"},
	{0x533C823DF73A45A2ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.impa.gad"},
	{0x533CB62AE963C961ULL, "127296941.ogg"},
	{0x533D1485BA27A63AULL, "lowtown_fem_lowlife_mob6168983.cl"},
	{0x533D79AC79D994B3ULL, "hlac_ca_cave_2193_1z.dds"},
	{0x533DFCCA10C317E6ULL, "233011848.ogg"},
	{0x533E0B57A3E566C9ULL, "6136915_m.fxe"},
	{0x533E467E3C7B563FULL, "529744923.ogg"},
	{0x533E90922ADA2F8AULL, "and100_anders6095800.cl"},
	{0x533ECCECB559892DULL, "277050523.ogg"},
	{0x533EE27BA732A281ULL, "llac_lt_0_lowtown_n_-65536_1.dds"},
	{0x533EED4F42979237ULL, "llac_ca_cave_2194_0.dds"},
	{0x533EF84332AFABF0ULL, "hf_amb_darktown_lowlife.mor"},
	{0x533F7488DBDBFF5EULL, "gen00fl_sebastian6174830.cl"},
	{0x533F895577A8A1ABULL, "qun211cr_crazy_woman6088450.cl"},
	{0x533FD44FBC2083EDULL, "vxht_mansion2b_38k_0_0.vlm"},
	{0x534034F6FA79F43EULL, "6164450_m.fxe"},
	{0x5340373A6022B5BFULL, "6180523_m.fxe"},
	{0x53403975847D844DULL, "tools_way_pt04.mmh"},
	{0x53404B75847DA2D3ULL, "tools_way_pt04.msh"},
	{0x5340D7684D9AF5F8ULL, "tools_bulb.mmh"},
	{0x53415AECAA650BE8ULL, "965273015.ogg"},
	{0x5341728F31B6ABD1ULL, "hlac_ht_mansion2b_818_44y.dds"},
	{0x5341ADBB65F188B6ULL, "art\\characters\\creatures\\spider\\animation\\atk_v1.evt"},
	{0x5341AEB01E1E8895ULL, "54621201.ogg"},
	{0x5342056D95671E26ULL, "art\\vfx\\combat\\vfx_c_rw_attack1_c.mmh"},
	{0x53423AA401F38159ULL, "vxlt_undercityserialkiller_1yo_0_0.vlm"},
	{0x53424877EF2E4E40ULL, "ico_boots_light_3.dds"},
	{0x53428AA01CCAE17FULL, "art\\vfx\\models\\vfx_twist_chunky.msh"},
	{0x534290A01CCAEB21ULL, "art\\vfx\\models\\vfx_twist_chunky.mmh"},
	{0x5342E25CA1168C2AULL, "seb201_lady_harimann6177312.cl"},
	{0x5343057A9189A549ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_basic_2.mao"},
	{0x53430BDA49A3B383ULL, "gen00fl_merrill6128707.cl"},
	{0x534332AF2AFEC0ACULL, "65023622.ogg"},
	{0x53433E007BF09734ULL, "vxht_0_hightown_n_c7t_0_0.vlm"},
	{0x53435044148D007FULL, "rock_small.dds"},
	{0x53436155AA066613ULL, "278770473.ogg"},
	{0x5343A0760458AFF2ULL, "llac_dr_0_deeproads_5845_8.dds"},
	{0x5343A53DF7409F99ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.impa.evt"},
	{0x534407922F59B76EULL, "art\\characters\\playercharacter\\humanmale\\boo_masd.msh"},
	{0x534415922F59CEBFULL, "art\\characters\\playercharacter\\humanmale\\boo_masd.mao"},
	{0x534421922F59E35CULL, "art\\characters\\playercharacter\\humanmale\\boo_masd.mmh"},
	{0x53444E9132D4BF04ULL, "vxlt_0_lowtown_docks_bss_0_0.vlm"},
	{0x53445441029F15C2ULL, "lgt110cr_urn_con_leader.utc"},
	{0x5344A4AF036B2729ULL, "6167870_m.fxe"},
	{0x5344AE3A60A40304ULL, "gen_im_trs_cot_rng_r2.uti"},
	{0x534549DAAAE30740ULL, "278113686.ogg"},
	{0x534628906BEB995CULL, "l_env_wal_hightown_crowd_bed.wbd"},
	{0x5346895D1D937711ULL, "184662604.ogg"},
	{0x5346C57D16A0D113ULL, "6114085_m.fxe"},
	{0x5346E282B9F397A7ULL, "llac_lt_0_lowtown_a2_n_-196609_3.dds"},
	{0x5346E7E0B44BB772ULL, "x_narrativeroom_mmi.gff"},
	{0x53470A23B2B7AE38ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.bl_atk_v1.evt"},
	{0x534769452934E4B6ULL, "vxlt_0_lowtown_docks_aql_0_0.vlm"},
	{0x53477BEEA8BEB700ULL, "6129261_f.fxe"},
	{0x534819CE7BE3C116ULL, "dae000_merchant_mage_hf6164462.cl"},
	{0x5348472213378D42ULL, "gen00fl_merrill6129888.cl"},
	{0x53487B2ED443ABFFULL, "vxlt_0_lowtown_a2_n_c_ep3_0_1.vlm"},
	{0x5348A4BB65F79841ULL, "art\\characters\\creatures\\spider\\animation\\atk_v1.gad"},
	{0x5348AFC9624A4F1AULL, "6038636_m.fxe"},
	{0x53499DEB546DCCABULL, "1046388683.ogg"},
	{0x5349EF2DD2B97592ULL, "79914.ogg"},
	{0x534A48B2057AED0AULL, "ht_0_hightown_30_0_lc.mmh"},
	{0x534A4C09F750805DULL, "vxw_bonepit_dy_0_0.vlm"},
	{0x534ADE684DA35583ULL, "tools_bulb.phy"},
	{0x534BC103E5EB5C23ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_aravell_dp.dds"},
	{0x534BFC45E42E0A51ULL, "88005480.ogg"},
	{0x534D783FCF0C8D09ULL, "hlac_ca_caverns2_5802_0z.dds"},
	{0x534D8B294D830B5FULL, "ach_exp_craft_nodes.dds"},
	{0x534E0DD83DC340A0ULL, "vxht_0_hightown_a3_at9_0_0.vlm"},
	{0x534E420E796E7A6EULL, "lt_house_probe_hq.dds"},
	{0x534E6FA367E535A1ULL, "plt_lgt000pt_ultra_4_02.plo"},
	{0x534EBA78909AB61BULL, "lt_backalley_n_2_0_lc.mmh"},
	{0x534ECFA4131DBC0EULL, "vxlt_gallowscourtyard_65g_0_0.vlm"},
	{0x534F31F4B1BC9B98ULL, "dae000_brothel_orana6160792.cl"},
	{0x534F51EF71A6CADEULL, "gen00fl_fenris6126852.cl"},
	{0x534F6C83E532990DULL, "plt_and200pt_opening.plo"},
	{0x534FA5BB65FDB8EFULL, "art\\characters\\creatures\\spider\\animation\\atk_v1.ani"},
	{0x534FFFA8C73261FFULL, "vxlt_0_lowtown_a2_hg3_0_0.vlm"},
	{0x5350AD3DF74B3788ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.impa.ani"},
	{0x5350DDDC8BF32AD5ULL, "6179316_m.fxe"},
	{0x53513275848BEB0AULL, "tools_way_pt04.phy"},
	{0x53519192B330FA73ULL, "6054851_m.fxe"},
	{0x535198BE2BA3633CULL, "dae000_bartrand6201347.cl"},
	{0x5351F0110D60ACF9ULL, "dae200_bran6144570.cl"},
	{0x5352327AF5347575ULL, "cli_army_ds_gatesdef.gda"},
	{0x5352413821AA6096ULL, "hlac_lt_0_lowtown_docks_-196609_0x.dds"},
	{0x53527D8DE15292F8ULL, "377337271.ogg"},
	{0x535280187E7CBA28ULL, "364784574.ogg"},
	{0x53530DC2B2717927ULL, "831031982.ogg"},
	{0x5353146ACABC0A2FULL, "hlac_ht_chantry_n_2534_14x.dds"},
	{0x535413079FE75CA5ULL, "zz_approval_debug6003688.cl"},
	{0x5354272FFE798288ULL, "follower_banter6171954.cl"},
	{0x5354DDA3A087ABAAULL, "vxw_bonepitmining_a2_1nu_0_0.vlm"},
	{0x5354E1F58D775261ULL, "534212130.ogg"},
	{0x53554D4E201423ECULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.anb"},
	{0x5355E2930204F87EULL, "808766744.ogg"},
	{0x5355FDC267D34C9FULL, "6067928_f.fxe"},
	{0x53562DA9F16C939BULL, "vxlt_hangedman_1mn_0_0.vlm"},
	{0x5356774E9226D8FEULL, "art\\levels\\races\\proxy\\props\\_textures\\logpile_n.dds"},
	{0x5356944A1DB09D0BULL, "1061232712.ogg"},
	{0x5356DB959B579D35ULL, "character_character_p239.psh.fxo.dx11"},
	{0x53576BA24FD44675ULL, "6118156_f.fxe"},
	{0x5357DE2504D52DF1ULL, "686266788.ogg"},
	{0x53582B2D0B00B91EULL, "6033376_m.fxe"},
	{0x53586A45974968E6ULL, "art\\characters\\creatures\\stonegolem\\animation\\awaken.ani"},
	{0x53588CC4C2921160ULL, "hlac_lt_0_lowtown_docks_a3_-2_0z.dds"},
	{0x53589E148CBD91D0ULL, "vxdr_primevaltomb_7t7_0_0.vlm"},
	{0x5358BDE6E94698DDULL, "vxw_bonepit_ox_0_0.vlm"},
	{0x5358CB1C9A7E2332ULL, "634833704.ogg"},
	{0x5358FE3025B4C795ULL, "511566966.ogg"},
	{0x5359C91276185B43ULL, "116306242.ogg"},
	{0x535A6036287BB09CULL, "hlac_lt_warehouse_n_0_6z.dds"},
	{0x535A71D50C857398ULL, "ave200_donnic6110718.cl"},
	{0x535B42587FE29074ULL, "hlac_ht_0_hightown_a3_0_0x.dds"},
	{0x535BAF3660CD68C4ULL, "6129464_m.fxe"},
	{0x535BC5B8EB832103ULL, "160068716.ogg"},
	{0x535BD41384D57B29ULL, "vxlt_0_lowtown_a2_n_c_ekg_0_2.vlm"},
	{0x535BDBFF273D6EA9ULL, "hlac_lt_gallowstemplar_2394_3z.dds"},
	{0x535C7CEE4FE49CA1ULL, "hlac_w_sundermount_327674_0y.dds"},
	{0x535D8A445915FD56ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hc1a.msh"},
	{0x535D944459160E54ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hc1a.mmh"},
	{0x535DD39311C43527ULL, "vxht_bartrand_n_64t_0_0.vlm"},
	{0x535DEE77CE16E3C8ULL, "w_bonepit_11_0_lc.mmh"},
	{0x535E740587C3941CULL, "hlac_ca_caverns2prc_5802_44z.dds"},
	{0x535F1E5DE5DFB259ULL, "vxlt_undercityserialkiller_1sm_0_0.vlm"},
	{0x536021BA5FDCC834ULL, "vxht_0_hightown_n_c9z_0_0.vlm"},
	{0x53603128A35E4470ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_sika_0d.dds"},
	{0x536053501C4B8A66ULL, "vxht_0_hightown_a3_avi_0_0.vlm"},
	{0x53607F7662C15C1EULL, "267940279.ogg"},
	{0x5360B12C9D26A1D4ULL, "hlac_lt_undercityserialkiller_-1_40x.dds"},
	{0x53615D4414A613DAULL, "rock_small.mmh"},
	{0x5361614414A61A11ULL, "rock_small.mao"},
	{0x53616F4414A63220ULL, "rock_small.msh"},
	{0x5361F063B91ED616ULL, "426097873.ogg"},
	{0x536209775AAB1B66ULL, "vxht_0_hightown_ccq_0_0.vlm"},
	{0x5362175DB34A8E5FULL, "zz_mrl_debug6184220.cl"},
	{0x536326A3A89E5B49ULL, "545415599.ogg"},
	{0x53636354CAD08CBDULL, "mcr310_party_banter6200737.cl"},
	{0x5363FA55CB8E91FAULL, "gen00fl_dog6167629.cl"},
	{0x5364E418F139FEF1ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.dura.ani"},
	{0x536505F6BAAA1CC3ULL, "hlac_ht_0_hightown_-1_0z.dds"},
	{0x536566EAF68215F7ULL, "35490242.ogg"},
	{0x53661F9A139E4ABFULL, "6060386_m.fxe"},
	{0x5366524597557D57ULL, "art\\characters\\creatures\\stonegolem\\animation\\awaken.evt"},
	{0x5366AFAD48666DA0ULL, "825449268.ogg"},
	{0x53675A015FC8D03FULL, "6050073_m.fxe"},
	{0x53678C9DB7EB3CBBULL, "art\\levels\\races\\design\\prp_bodybagcoins.mao"},
	{0x536824400B40FB7CULL, "mcr310_argument6072322.cl"},
	{0x53683C32BD2D3590ULL, "vxw_encounter3_ex_0_0.vlm"},
	{0x5368BFAEFF073DCCULL, "vxdr_primevaltomb_4zv_0_0.vlm"},
	{0x536900922F78C727ULL, "art\\characters\\playercharacter\\humanmale\\boo_masd.phy"},
	{0x53694825F002E193ULL, "vxht_0_hightown_a2_a6e_0_0.vlm"},
	{0x53697DF939454BAFULL, "mag110_arianni6020502.cl"},
	{0x5369CD2746E6FEDFULL, "6125833_f.fxe"},
	{0x536A6906882780CAULL, "6129915_f.fxe"},
	{0x536AAD4C0182AE96ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0cc_basic_1.mao"},
	{0x536AF0DE639BDDC1ULL, "ser191_emeric6108142.cl"},
	{0x536BC5794E898CC3ULL, "t3_har_gry.tnt"},
	{0x536BEF18F1402F93ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.dura.gad"},
	{0x536BF9646EEF61D9ULL, "llac_lt_darktown_-1_47.dds"},
	{0x536C8E624ED8E9EBULL, "vxw_encounter3_h_0_0.vlm"},
	{0x536CA7959ACA5432ULL, "matt_rogue_1.cut"},
	{0x536D5345975B9DE0ULL, "art\\characters\\creatures\\stonegolem\\animation\\awaken.gad"},
	{0x536D8B3DDA077879ULL, "174023459.ogg"},
	{0x536E9B928E82CC9DULL, "vxw_encounter2_6s_0_0.vlm"},
	{0x536F588CC0B9AD90ULL, "vxht_0_hightown_n_bsu_0_0.vlm"},
	{0x536F75282DBC49CEULL, "vxlt_0_lowtown_a2_hh3_0_0.vlm"},
	{0x53709B6ED74FAF83ULL, "205807786.ogg"},
	{0x5370E47F1F3A7B7EULL, "6126889_m.fxe"},
	{0x5370E952178BC0A0ULL, "6161240_m.fxe"},
	{0x537104AE9ABD1ADEULL, "ico_helm_light_4.dds"},
	{0x537130B15B24F3BFULL, "zz_followers_banter6159003.cl"},
	{0x5371452E87EADCCCULL, "hlac_lt_0_lowtown_65538_0x.dds"},
	{0x53715DAFC7D341B9ULL, "mrl310_merrill6144214.cl"},
	{0x53718E778B4D6AD2ULL, "hlac_w_encounter1_0_0y.dds"},
	{0x5371C58C83B3649DULL, "6031711_m.fxe"},
	{0x5371DF0AE62BB243ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_exit.ani"},
	{0x537226906C10D0D6ULL, "l_env_wal_hightown_crowd_bed.bnk"},
	{0x53724F9B5C84F8AFULL, "6107287_m.fxe"},
	{0x5372839337C058E3ULL, "385735332.ogg"},
	{0x5372DEF48C415971ULL, "287647379.ogg"},
	{0x5372EC18F14649F8ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.dura.evt"},
	{0x53731BDF6153D4F6ULL, "fen110_fenris6100291.cl"},
	{0x537325CB9CF3D69BULL, "188815985.ogg"},
	{0x5373CEECFF50C4E3ULL, "w_encounter3_66_0.msh"},
	{0x5373DCECFF50DC3DULL, "w_encounter3_66_0.mmh"},
	{0x5373E0ECFF50E34EULL, "w_encounter3_66_0.mao"},
	{0x53745CFBFA643AB4ULL, "hlac_lt_hangedman_-1_9z.dds"},
	{0x5374946E45C906D4ULL, "vxht_0_hightown_9as_0_0.vlm"},
	{0x5374F28E74A4CBBCULL, "195543900.ogg"},
	{0x5375D079EB6A67C1ULL, "gen00fl_fenris6128006.cl"},
	{0x53760490065D38A9ULL, "vxht_0_hightown_a3_by0_0_0.vlm"},
	{0x53766334BF389B7DULL, "892240932.ogg"},
	{0x5376B788B49F9DB1ULL, "vxdr_primevalentrance_23r_0_0.vlm"},
	{0x5376B9D91C51F004ULL, "hlac_lt_0_lowtown_n_65536_10z.dds"},
	{0x5377036EC48FC142ULL, "6207801_m.fxe"},
	{0x5377615B962F19B1ULL, "vxca_caverns2_5a9_0_0.vlm"},
	{0x53779F1F2F5EAB0AULL, "vxht_mansion2_3eg_0_0.vlm"},
	{0x5377C3ED057D0297ULL, "6155781_m.fxe"},
	{0x537888FC75CD8710ULL, "vxht_keep_6al_0_0.vlm"},
	{0x5379213EDF0D3E86ULL, "34597972.ogg"},
	{0x537AB23CFF853F2EULL, "hlac_lt_0_lowtown_a2_131070_5x.dds"},
	{0x537AE72BA33FF3E9ULL, "47719991.ogg"},
	{0x537AE97705E54C20ULL, "hlac_lt_gallowscourtyard_n_-131071_3x.dds"},
	{0x537B25DD6B6DEDBCULL, "mcr310_party_banter6182940.cl"},
	{0x537B2F6EC0B93388ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_5y.dds"},
	{0x537B81C9E308E17AULL, "6069722_m.fxe"},
	{0x537B897A74EE76F1ULL, "fadesaturatemask.dds"},
	{0x537BD69F3DBC5FD2ULL, "68172772.ogg"},
	{0x537BDBE36B54417AULL, "llac_lt_0_lowtown_-2_2.dds"},
	{0x537BFE2CD3593F1CULL, "art\\characters\\playercharacter\\humanfemale\\rob_maga.mao"},
	{0x537C0A2CD35953BFULL, "art\\characters\\playercharacter\\humanfemale\\rob_maga.mmh"},
	{0x537C102CD3595E61ULL, "art\\characters\\playercharacter\\humanfemale\\rob_maga.msh"},
	{0x537CC313F3604858ULL, "hlac_w_sundermount_-262146_0y.dds"},
	{0x537CC5882B23DC2BULL, "6092505_m.fxe"},
	{0x537D083510B7E48BULL, "oil_proj_tail.mao"},
	{0x537D0D6F8078E51DULL, "vxw_encounter2_1r_0_0.vlm"},
	{0x537DF8677EA0955FULL, "zz_followers_banter6159357.cl"},
	{0x537E0E0548B94734ULL, "vxht_0_hightown_a2_c1a_0_0.vlm"},
	{0x537E2041F23B7650ULL, "vxca_cave_23j_0_0.vlm"},
	{0x537E6D154AA42C28ULL, "w_blightlands_252_0.msh"},
	{0x537E7F154AA44AB9ULL, "w_blightlands_252_0.mao"},
	{0x537E8B154AA45F52ULL, "w_blightlands_252_0.mmh"},
	{0x537EADE93CC9A06DULL, "hlac_lt_0_lowtown_a3_n_c_0_7y.dds"},
	{0x537EF261D2BA2B6BULL, "art\\vfx\\combat\\vfx_c_pride_pois_aoe_p.impa.ani"},
	{0x537F1A79564EB079ULL, "ht_0_hightown_34_0.anb"},
	{0x537FF70AE638183AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_exit.evt"},
	{0x538002992F736413ULL, "vxca_caverns2_6zm_0_0.vlm"},
	{0x538030AE378AEF5BULL, "hlac_lt_0_lowtown_docks_n_131072_0x.dds"},
	{0x53805D0151F16DDCULL, "6162511_m.fxe"},
	{0x53807044504C296FULL, "6079674_m.fxe"},
	{0x5380B68FDD303779ULL, "character_character_p222.psh.fxo.dx11"},
	{0x538138E9FDC3FB69ULL, "hlac_lt_darktown_-1_12z.dds"},
	{0x53819B8A661172ADULL, "w_bonepit_81_0.msh"},
	{0x5381A98A66118AE0ULL, "w_bonepit_81_0.mao"},
	{0x5381AD8A66119133ULL, "w_bonepit_81_0.mmh"},
	{0x538208D9C40DF5ACULL, "dae300_alistair_king6134904.cl"},
	{0x53826E87904872A3ULL, "dae300_cullen6140627.cl"},
	{0x5382B1FE7FF7AE3BULL, "vxlt_hangedman_n_1yh_0_1.vlm"},
	{0x538326E08DD2C808ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelstack.mmh"},
	{0x538332E08DD2DCABULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelstack.mao"},
	{0x538344E08DD2FB32ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelstack.msh"},
	{0x53834D96167B9A10ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrelpile.mao"},
	{0x53835196167BA123ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrelpile.mmh"},
	{0x53835F96167BB97DULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrelpile.msh"},
	{0x538402B8369E052CULL, "lgt150ip_ntb_wr_ing_chest_3.utp"},
	{0x53842888A1037987ULL, "art\\characters\\creatures\\rubblegolem\\animation\\cindes\\dialogue\\ready_slow.gad"},
	{0x5384523F36822481ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_neutral_female.anb"},
	{0x53847435CF26ECE6ULL, "6033989_m.fxe"},
	{0x5384997CF50F2D58ULL, "6064728_f.fxe"},
	{0x5385271B48436CBBULL, "gen_im_acc_rng_act2_03.uti"},
	{0x53852965A695B631ULL, "vxdr_primevalentrance_26q_0_0.vlm"},
	{0x53852E3E9CD478C9ULL, "zz_mcr_debug6080828.cl"},
	{0x53853F0CA2CE5FB0ULL, "seb321cr_rebel_mage.ncs"},
	{0x538540F8841BE609ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubrush_d.dds"},
	{0x5385E2622E6B84FBULL, "6059997_m.fxe"},
	{0x5385F60AE63C822DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_exit.gad"},
	{0x5386621F45081C67ULL, "6088418_m.fxe"},
	{0x538669A68127B054ULL, "vxht_chantry_n_26v_0_0.vlm"},
	{0x5386A71E30EBE181ULL, "llac_lt_gallowstemplar_n_2384_7.dds"},
	{0x5386FAD967C39F90ULL, "vxht_keep_6jm_0_0.vlm"},
	{0x53881EE7D2319F1BULL, "vxw_encounter2_camp_4w_0_0.vlm"},
	{0x5388721EAB365369ULL, "ht_keep_n_3732.rml"},
	{0x5388F90BF46D0DAEULL, "6092193_m.fxe"},
	{0x53899141648807B3ULL, "806309991.ogg"},
	{0x5389F36B3C5658F7ULL, "1009393873.ogg"},
	{0x538A5B8A1CD5AD5DULL, "t3_psp_b00.tnt"},
	{0x538A83A078DC3BB5ULL, "hlac_lt_backalley_n_1_3x.dds"},
	{0x538A94FC04C573A0ULL, "124774481.ogg"},
	{0x538AFC25D6DA8F7BULL, "gen00fl_isabela_init6110244.cl"},
	{0x538B1FF4F7D2CACFULL, "mcr370_meredith6136112.cl"},
	{0x538B37651D19F4ABULL, "6097605_m.fxe"},
	{0x538B98930FAC25B7ULL, "646290930.ogg"},
	{0x538BDA61D2C48CE2ULL, "art\\vfx\\combat\\vfx_c_pride_pois_aoe_p.impa.evt"},
	{0x538BDCF8100B179AULL, "572829571.ogg"},
	{0x538C17BCF1E1A527ULL, "140867852.ogg"},
	{0x538C3C46A4A5DE90ULL, "vxht_0_hightown_n_bys_0_0.vlm"},
	{0x538CCEEA17F0739EULL, "6070935_m.fxe"},
	{0x538D55B129A2BA82ULL, "6057352_m.fxe"},
	{0x538D7D0BAC206322ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.mmh"},
	{0x538D9676B84D6E6EULL, "mag321_comtesse_de_launcet6172411.cl"},
	{0x538DE47BC05CC6FDULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesebrie.mao"},
	{0x538DE87BC05CCD36ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesebrie.mmh"},
	{0x538DF27BC05CDE34ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesebrie.msh"},
	{0x538DFA8C558F9015ULL, "20603433.ogg"},
	{0x538E43545A8406EAULL, "t1_mue_pi2.tnt"},
	{0x538EA3EAAF281156ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_4.ani"},
	{0x538EA51768600DAAULL, "vxw_bonepitmining_a2_1zq_0_0.vlm"},
	{0x538EDCCB707F511BULL, "hlac_w_encounter1_-65534_0x.dds"},
	{0x538F569E9389124EULL, "vxht_0_hightown_a2_b27_0_0.vlm"},
	{0x538F774B8C2D902EULL, "vxlt_0_lowtown_a2_n_c_emh_0_0.vlm"},
	{0x5390BFB7BE465F10ULL, "391521313.ogg"},
	{0x53913D31D5AD693EULL, "vxw_woundedcoast_1ea_0_0.vlm"},
	{0x539157028DF0AC97ULL, "6131394_m.fxe"},
	{0x5391822202147A90ULL, "806873031.ogg"},
	{0x539249A0C618FDE9ULL, "sst_ss_meredith.gda"},
	{0x53924D9457E997B6ULL, "hlac_w_sundermount_-131074_0z.dds"},
	{0x5392D961D2CAA9F5ULL, "art\\vfx\\combat\\vfx_c_pride_pois_aoe_p.impa.gad"},
	{0x5392FF610937F3D3ULL, "337303091.ogg"},
	{0x53936821635C86E2ULL, "hlac_lt_0_lowtown_a2_-3_0y.dds"},
	{0x5393CB7902AA7042ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_imp_c.mmh"},
	{0x5393F241720749A8ULL, "vxht_0_hightown_aki_0_0.vlm"},
	{0x539400713E67FCCFULL, "424740602.ogg"},
	{0x53945C09F2F7C93CULL, "115357006.ogg"},
	{0x53948C8438D61887ULL, "6149493_f.fxe"},
	{0x539494F35B99BF94ULL, "364305429.ogg"},
	{0x5394A6E2941DBF66ULL, "zz_spy133cr_slave.utc"},
	{0x5394AF3D39D94AA2ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig2_r.dds"},
	{0x5394D726BD8BA1AFULL, "hlac_lt_hangedman_n_-1_10z.dds"},
	{0x5394F242C0468997ULL, "ave000pt_main.ncs"},
	{0x539563433724630CULL, "6030670_m.fxe"},
	{0x5395CC4011F53515ULL, "6154577_m.fxe"},
	{0x5395DBEF6932DAA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_thumb_jab_back_and_side.gad"},
	{0x539618B8CACEED02ULL, "vxdr_primevaltomb_6px_0_0.vlm"},
	{0x53962DBF9CAA497EULL, "llac_ht_mansion2b_818_53.dds"},
	{0x539659BE06C33BDEULL, "vxlt_undercitydungeon_1ih_0_1.vlm"},
	{0x5396D748AC54EFE6ULL, "1047587293.ogg"},
	{0x53976AE8E663262EULL, "6097293_m.fxe"},
	{0x5398553AEDD2803EULL, "qcr000_arishok6060114.cl"},
	{0x53988D440F613634ULL, "84726317.ogg"},
	{0x5398AFD8A4939783ULL, "mcr310_argument6072288.cl"},
	{0x5398E0E3B02D6AC8ULL, "vxlt_gallowscourtyard_4xd_0_0.vlm"},
	{0x53991588A115B5CDULL, "art\\characters\\creatures\\rubblegolem\\animation\\cindes\\dialogue\\ready_slow.ani"},
	{0x53995830534AFF3DULL, "plt_cod_aow_injuries.plo"},
	{0x539968C4CDF58250ULL, "783862888.ogg"},
	{0x5399A0ED424843A4ULL, "gen00fl_carver6103394.cl"},
	{0x5399B5F259DE88E0ULL, "vxca_caverns2_7ap_0_2.vlm"},
	{0x5399D1C97A4C5ACBULL, "vxlt_blackemporium_w1_0_0.vlm"},
	{0x539AECFCF79E1238ULL, "llac_lt_0_lowtown_a2_65533_3.dds"},
	{0x539B25CEF24A2FB4ULL, "mrl250_pol.cnv"},
	{0x539BA73A0CE4C33EULL, "6029036_m.fxe"},
	{0x539BABEAAF32A947ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_4.evt"},
	{0x539CA23C7711C58BULL, "vxht_keep_n_6i4_0_0.vlm"},
	{0x539CFEEF69393499ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_thumb_jab_back_and_side.evt"},
	{0x539D439060984A8EULL, "vxw_encounter3_92_0_0.vlm"},
	{0x539DBC52F95221E5ULL, "vxdr_primevaltomb_2gg_0_1.vlm"},
	{0x539DCDE982DC83A8ULL, "mag311_keran6041410.cl"},
	{0x539E2D5D9B45414DULL, "621941800.ogg"},
	{0x539E6A69AAEB36B8ULL, "61181139.ogg"},
	{0x539E7D5366B3D872ULL, "gen00fl_anders6125648.cl"},
	{0x539EE0C9867F2480ULL, "hlac_w_bonepit_327680_0x.dds"},
	{0x539EF9AB8491D92CULL, "hlac_lt_0_lowtown_a2_n_-65535_0y.dds"},
	{0x539F07CEF24DE5F4ULL, "mrl250_pol.bnk"},
	{0x539F68D54810D71AULL, "llac_dr_0_deeproads_5849_4.dds"},
	{0x539F755A867224ACULL, "qcr000_aveline6041086.cl"},
	{0x539FCDA57891C3E9ULL, "721930016.ogg"},
	{0x53A03C464E7EF587ULL, "6161485_m.fxe"},
	{0x53A0508ABBC9AD67ULL, "vxht_mansion_n_6kj_0_0.vlm"},
	{0x53A0C6E6BAA3D1A8ULL, "lt_0_lowtown_docks_116_0.msh"},
	{0x53A0D2BD6CBE6963ULL, "6098751_f.fxe"},
	{0x53A0D8E6BAA3F039ULL, "lt_0_lowtown_docks_116_0.mao"},
	{0x53A0E4E6BAA404D2ULL, "lt_0_lowtown_docks_116_0.mmh"},
	{0x53A1707EADDA84E0ULL, "vxlt_0_lowtown_a2_n_c_hv2_0_0.vlm"},
	{0x53A18C1C6E98AAEBULL, "vxdr_primevaltomb_7e6_0_1.vlm"},
	{0x53A1C25BF552D162ULL, "114430256.ogg"},
	{0x53A1CBEE4707B8FCULL, "6019155_warren_lead_threat.cut"},
	{0x53A1EE9CC4F9D6BDULL, "vxlt_0_lowtown_n_cfv_0_0.vlm"},
	{0x53A2641E63FD6228ULL, "vxht_0_hightown_anf_0_0.vlm"},
	{0x53A270CAD7755B66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv2.evt"},
	{0x53A29CD868804537ULL, "hlac_ht_chantry_n_2534_26y.dds"},
	{0x53A2CCEAAF38FF90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_4.gad"},
	{0x53A2F3DCFF2A6440ULL, "374581819.ogg"},
	{0x53A2FDBCD980535EULL, "hlac_ht_0_hightown_a2_-65538_1z.dds"},
	{0x53A3A43D877B81FDULL, "vxlt_warehouse_n_p3_0_0.vlm"},
	{0x53A48A95BCC50582ULL, "vxdr_primevaltomb_6wy_0_0.vlm"},
	{0x53A4936FDF1A5E3DULL, "358568624.ogg"},
	{0x53A4C88C20908CDCULL, "6083922_m.fxe"},
	{0x53A51E1143212899ULL, "follower_banter6151230.cl"},
	{0x53A589A72E58AECAULL, "6173371_m.fxe"},
	{0x53A67FB75E2016DAULL, "vxht_brothel_n_57o_0_0.vlm"},
	{0x53A68EE16CEB0FDCULL, "6126810_m.fxe"},
	{0x53A7EE73C2F21950ULL, "6143819_m.fxe"},
	{0x53A8003829E3F72CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_012a_0d.dds"},
	{0x53A85E2F8DEC5ADBULL, "6019207_m.fxe"},
	{0x53A8B344D2D9FB0BULL, "vxlt_0_lowtown_a2_en1_0_0.vlm"},
	{0x53A8B662809F631DULL, "vxw_blightlands_lz_0_0.vlm"},
	{0x53A8F05CC2AFDC0BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyn_0s.dds"},
	{0x53A9200088920F90ULL, "vxht_0_hightown_n_boq_0_0.vlm"},
	{0x53A92F46030597C1ULL, "6143448_f.fxe"},
	{0x53A95735453D1CEBULL, "6131884_m.fxe"},
	{0x53A9728C7AD4CAF0ULL, "6083342_m.fxe"},
	{0x53A987CAD77BA0B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv2.gad"},
	{0x53A9F3FF10539198ULL, "vxw_sundermount_2e4_0_0.vlm"},
	{0x53A9FDE6F44387A1ULL, "vxht_0_hightown_n_c44_0_0.vlm"},
	{0x53AA06EF6943CC88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_thumb_jab_back_and_side.ani"},
	{0x53AA3CC89518C1D6ULL, "162206986.ogg"},
	{0x53AA4F16D866810AULL, "follower_banter6163133.cl"},
	{0x53AA93E82E9C2426ULL, "vxlt_gallowstemplar_n_5yy_0_0.vlm"},
	{0x53AAA52C910EA257ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masf_0s.dds"},
	{0x53AB60D8D20BCA39ULL, "vxlt_0_lowtown_n_e1q_0_0.vlm"},
	{0x53AC1F1E85BE509EULL, "6048473_m.fxe"},
	{0x53AC81636C9F8D6EULL, "vxlt_0_lowtown_a2_n_en3_0_0.vlm"},
	{0x53AC868FDBD16277ULL, "hightown_fem_cityguard_mob6168434.cl"},
	{0x53AC988237CB4912ULL, "vxw_blightlands_ar_0_0.vlm"},
	{0x53ADB371C76DF194ULL, "vxw_sundermount_30b_0_0.vlm"},
	{0x53AE20EBB9999A3EULL, "vxw_woundedcoast_1kc_0_0.vlm"},
	{0x53AE8F00C50F8C87ULL, "gen00fl_anders6119262.cl"},
	{0x53AEAC6A10FA752AULL, "435642958.ogg"},
	{0x53AEC267ADBFC5E7ULL, "vxht_mansion_n_6ne_0_0.vlm"},
	{0x53AED9D08BFC0459ULL, "dae000_fem_commoner2_mob6174589.cl"},
	{0x53AEEE105305130EULL, "110018113.ogg"},
	{0x53AEF024FAF5E61FULL, "hlac_ht_bartrand_n_3367_0y.dds"},
	{0x53AF056676BCCC5CULL, "vxdr_0_deeproads_ad6_0_1.vlm"},
	{0x53AF382C0374D345ULL, "vxht_keep_4ak_0_0.vlm"},
	{0x53AF6857D5925ECFULL, "861927941.ogg"},
	{0x53AFCC03CE359AA9ULL, "vxht_0_hightown_a3_be4_0_0.vlm"},
	{0x53AFD6D8DDC969B2ULL, "567368636.ogg"},
	{0x53AFE25B9FD09D60ULL, "vxlt_0_lowtown_a2_n_c_hq3_0_0.vlm"},
	{0x53AFF10751F99CFEULL, "6080811_m.fxe"},
	{0x53AFFEA8152EE08CULL, "zz_followers_banter6158184.cl"},
	{0x53B06681ADB01CECULL, "zz_otl_debug6060949.cl"},
	{0x53B06B2636DBF46AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_leaves01_e.dds"},
	{0x53B088CAD781C15FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv2.ani"},
	{0x53B099DD3EDC6E53ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkb.ani"},
	{0x53B0CE69529DA570ULL, "586999446.ogg"},
	{0x53B11ECE9DB4EC05ULL, "ef_arianni.mor"},
	{0x53B134F77547DD71ULL, "plt_lgt000pt_lite_darktown.nss"},
	{0x53B2018BBF7618C2ULL, "827487198.ogg"},
	{0x53B2161A79719A7DULL, "vxlt_warehouse_n_w2_0_0.vlm"},
	{0x53B254B21BAA2A26ULL, "hlac_lt_hangedman_n_-1_23y.dds"},
	{0x53B349CA554A755FULL, "vxw_bonepitmining_a2_63_0_0.vlm"},
	{0x53B40A3FC48C92E8ULL, "720021699.ogg"},
	{0x53B4100E7315AD4FULL, "48688397.ogg"},
	{0x53B44DA4B4CFEAF1ULL, "489175029.ogg"},
	{0x53B48932E4923028ULL, "6134772_m.fxe"},
	{0x53B4CB50D48A40A5ULL, "916260359.ogg"},
	{0x53B4D0EEA6661DDAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeic.phy"},
	{0x53B5525CF0132D33ULL, "gen00fl_varric6087596.cl"},
	{0x53B555220C2DADF3ULL, "6143276_m.fxe"},
	{0x53B5D225939E2F13ULL, "hla_ht_0_hightown_a2_-131070_0z.dds"},
	{0x53B5E311FFB3EE7BULL, "vxlt_gallowscourtyard_n_6xn_0_0.vlm"},
	{0x53B61C27215E1575ULL, "556353354.ogg"},
	{0x53B67C0BAC4366C9ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.anb"},
	{0x53B6CAEC803A98A3ULL, "36141965.ogg"},
	{0x53B6ECBEF7390733ULL, "vxht_keep_n_580_0_0.vlm"},
	{0x53B7283F72957B9DULL, "vxw_blightlands_qy_0_0.vlm"},
	{0x53B7C2AACADC3FFDULL, "art\\cinematics\\_crit"},
	{0x53B8006567ED4253ULL, "257452778.ogg"},
	{0x53B81B79567E7779ULL, "ht_0_hightown_34_0.rml"},
	{0x53B865DC0249AA18ULL, "vxw_sundermount_2n5_0_0.vlm"},
	{0x53B91846AC654B88ULL, "339771613.ogg"},
	{0x53B95371BC118880ULL, "follower_banter6161083.cl"},
	{0x53B9876E3B0BB06DULL, "mag370_meredith6172380.cl"},
	{0x53B9934459646B4FULL, "art\\characters\\playercharacter\\elfmale\\em_har_hc1a.phy"},
	{0x53B9D87232E502BAULL, "dae300_bran6143526.cl"},
	{0x53B9EEC21B68A2CBULL, "llac_lt_0_lowtown_n_-196609_1.dds"},
	{0x53BA728844849BC9ULL, "840757196.ogg"},
	{0x53BAB702FF7D2D12ULL, "vxw_encounter2_8f_0_0.vlm"},
	{0x53BAF3405E95A5EEULL, "vxlt_0_lowtown_a2_n_es4_0_0.vlm"},
	{0x53BB0A5F29C16192ULL, "vxw_blightlands_js_0_0.vlm"},
	{0x53BB4F5BAC362CA5ULL, "vxlt_hangedman_n_58_0_0.vlm"},
	{0x53BB67B904A8E7C0ULL, "724406371.ogg"},
	{0x53BB711EAB61AA89ULL, "ht_keep_n_3732.anb"},
	{0x53BC1689099C41C1ULL, "181223277.ogg"},
	{0x53BC22610A977DE4ULL, "gen00fl_dog6167069.cl"},
	{0x53BC58A841B95A66ULL, "6137186_m.fxe"},
	{0x53BC96BA4829E0EEULL, "vxlt_warehouse_ub_0_0.vlm"},
	{0x53BCADC1A18C8CCCULL, "character_character_p287.psh.fxo.dx11"},
	{0x53BCB97CC1AA3DA6ULL, "418269886.ogg"},
	{0x53BCCA7902CD73E9ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_imp_c.anb"},
	{0x53BCD81486269453ULL, "hlac_dr_slavers_3217_1y.dds"},
	{0x53BD6223F999AE97ULL, "675554195.ogg"},
	{0x53BD91DD3EE6EA8AULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkb.evt"},
	{0x53BDB8787A66ACA8ULL, "gen00fl_varric6180238.cl"},
	{0x53BF2C2549C504B5ULL, "6125263_m.fxe"},
	{0x53BF62EA8A0A550FULL, "lt_0_lowtown_docks_n_243_0.mao"},
	{0x53BF6EEA8A0A69ACULL, "lt_0_lowtown_docks_n_243_0.mmh"},
	{0x53BF74EA8A0A739EULL, "lt_0_lowtown_docks_n_243_0.msh"},
	{0x53BFCDF50A26BD7FULL, "t3_0000053.tnt"},
	{0x53C059BFCCE39898ULL, "163734666.ogg"},
	{0x53C071203D98EE3DULL, "hlac_lt_0_lowtown_a3_n_c_-196608_3x.dds"},
	{0x53C07B1628B09C71ULL, "dre000_dougal6129803.cl"},
	{0x53C087F76B67B2FDULL, "vxlt_warehouse_n_z1_0_0.vlm"},
	{0x53C1036DB17904BFULL, "character_character_p42.psh.fxo.dx11"},
	{0x53C1615C51BFF804ULL, "6179494_m.fxe"},
	{0x53C177D5CB05F97AULL, "llac_lt_0_lowtown_docks_a3_0_1.dds"},
	{0x53C1B73C249C9E25ULL, "6143098_m.fxe"},
	{0x53C1C1F922EAAED7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_9.pwk"},
	{0x53C1C937072187D6ULL, "lowtown_male_lowlife2_mob6169190.cl"},
	{0x53C3291235441F38ULL, "vxlt_0_lowtown_a2_n_c_gx3_0_0.vlm"},
	{0x53C470DD3EECD1FDULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkb.gad"},
	{0x53C4EDD81107D328ULL, "hlac_lt_0_lowtown_docks_n_131071_6x.dds"},
	{0x53C654FCE7463263ULL, "6095316_f.fxe"},
	{0x53C65845C33C3481ULL, "hlac_ht_0_hightown_a2_1_0y.dds"},
	{0x53C6734B9327A5C0ULL, "vxw_woundedcoast_op_0_0.vlm"},
	{0x53C6D7B8F43FC298ULL, "vxw_sundermount_2s2_0_0.vlm"},
	{0x53C7B766E38DDDC3ULL, "vxca_caverns1_mines_imm_0_0.vlm"},
	{0x53C8E90FC8C7369DULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluebase.phy"},
	{0x53C929378FAB8C09ULL, "6147771_m.fxe"},
	{0x53C92F893AEB293BULL, "6043737_f.fxe"},
	{0x53C93EBF5405F22EULL, "vxlt_0_lowtown_a2_n_c_eql_0_0.vlm"},
	{0x53C96A25EC3E56EDULL, "vxlt_0_lowtown_a2_n_eow_0_0.vlm"},
	{0x53C96DD6497C6BD4ULL, "165745211.ogg"},
	{0x53C9747E3B68FA4FULL, "669356134.ogg"},
	{0x53CA5A9E7DC5810FULL, "vxdr_primevaltomb_3lx_0_0.vlm"},
	{0x53CA79408480DFCDULL, "6161133_m.fxe"},
	{0x53CAC8979793CD9AULL, "vxca_caverns2_7z4_0_1.vlm"},
	{0x53CADF1FB43B09FEULL, "6070492_m.fxe"},
	{0x53CAE5A1498ECFC9ULL, "6121927_f.fxe"},
	{0x53CB090ACCE20B47ULL, "6121972_f.fxe"},
	{0x53CB69521160D450ULL, "hlac_w_bonepitdestroyed_a3_131078_0z.dds"},
	{0x53CC7A0E82321ECAULL, "vxca_caverns2_8s1_0_0.vlm"},
	{0x53CCBCA50BB96A13ULL, "dae000_gamlen6005030.cl"},
	{0x53CD1A1CD81D6836ULL, "dre000_sibling_has_blight6036526.cl"},
	{0x53CD2CCC509CC3BCULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls105a_0n.dds"},
	{0x53CD2EFD4AA84F83ULL, "hm_emile.mor"},
	{0x53CD9E023BBB1D35ULL, "6122363_m.fxe"},
	{0x53CDB5B14CA9F6AAULL, "one100dg_kelder6109127.cl"},
	{0x53CDD7129A58D5DAULL, "hlac_w_bonepit_65537_0x.dds"},
	{0x53CE0213B5164B84ULL, "6038839_f.fxe"},
	{0x53CE087DB1F2C2FEULL, "gen_im_trs_dal_hvy_boo_r0.uti"},
	{0x53CEDFC0D71F0516ULL, "6020388_f.fxe"},
	{0x53CFD4017E1D6BEEULL, "332635736.ogg"},
	{0x53D06FECB5CA4B45ULL, "6160897_m.fxe"},
	{0x53D090CA9BA75523ULL, "dm_gen_carta_assassin.mor"},
	{0x53D0EA4347B67F0AULL, "one300_nuncio6083478.cl"},
	{0x53D1119A49EB945EULL, "101512060.ogg"},
	{0x53D1235C8F45915AULL, "vxht_0_hightown_a2_c3a_0_0.vlm"},
	{0x53D175AEDEF0558DULL, "6163430_m.fxe"},
	{0x53D21BCEF2796153ULL, "mrl250_pol.stg"},
	{0x53D2762C82F58E08ULL, "224653503.ogg"},
	{0x53D2D388AB7F8E2AULL, "vxht_playershouse_n_1r2_0_0.vlm"},
	{0x53D30E03BCA65573ULL, "6093909_f.fxe"},
	{0x53D383946D71BDF0ULL, "vxw_woundedcoast_h5_0_0.vlm"},
	{0x53D3A4D88F8E89ABULL, "lgt311_shw_mer_intro6156411.cl"},
	{0x53D3E4E3EF1219F1ULL, "6138724_m.fxe"},
	{0x53D409A456BB38A6ULL, "dae000_mother6136918.cl"},
	{0x53D46891FBD643C4ULL, "6165341_m.fxe"},
	{0x53D491B077857AD7ULL, "6042827_m.fxe"},
	{0x53D4E528851DBE40ULL, "vxw_woundedcoast_dq_0_0.vlm"},
	{0x53D54995E635DB18ULL, "vxw_sundermount_2t3_0_0.vlm"},
	{0x53D5957CE3736785ULL, "6169843_m.fxe"},
	{0x53D5BEB9BAA63715ULL, "llac_ht_mansion_n_2995_4.dds"},
	{0x53D5E221CA9113C4ULL, "mrl000cr_shadow.ncs"},
	{0x53D6097F7CB396D7ULL, "fx_fade_flip_cloud.mao"},
	{0x53D62943D583F643ULL, "vxca_caverns1_mines_ijl_0_0.vlm"},
	{0x53D65987B8C12E92ULL, "hlac_lt_0_lowtown_a2_-2_6x.dds"},
	{0x53D673FF3D882BD9ULL, "326934046.ogg"},
	{0x53D686AC6925D26EULL, "800864127.ogg"},
	{0x53D842414755FDBEULL, "isa211_isabela6202298.cl"},
	{0x53D8A5659522B441ULL, "vxht_chantry_31t_0_0.vlm"},
	{0x53D8EF012373C10CULL, "vxlt_0_lowtown_a2_gpn_0_0.vlm"},
	{0x53D9C6CE24201D27ULL, "312393941.ogg"},
	{0x53D9DE0FC8D59709ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluebase.mao"},
	{0x53D9E973C18C4068ULL, "readonly-0-rev-3-xx_extshingle.dds"},
	{0x53D9EA0FC8D5ABA2ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluebase.mmh"},
	{0x53D9EC0FC8D5AF58ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluebase.msh"},
	{0x53DA2A8B1FC1E44BULL, "727220333.ogg"},
	{0x53DA38DA4A418794ULL, "6102579_m.fxe"},
	{0x53DA488C22098492ULL, "gen_im_wep_rog_arc_act0_01.uti"},
	{0x53DA61E7F84A897EULL, "527682811.ogg"},
	{0x53DA98A17712B40CULL, "401285228.ogg"},
	{0x53DAD6953840E1A9ULL, "549937668.ogg"},
	{0x53DB65BECD8ACF1AULL, "hlac_lt_0_lowtown_docks_a3_n_65537_0z.dds"},
	{0x53DBAF9570535B2FULL, "hlac_lt_darktown_-1_1z.dds"},
	{0x53DBF637AFEB471DULL, "548408747.ogg"},
	{0x53DBF8C10F4B074BULL, "6141858_m.fxe"},
	{0x53DBFFD08FB47276ULL, "vxca_caverns2_970_0_0.vlm"},
	{0x53DC00C26FEFFF41ULL, "gen00fl_anders6125832.cl"},
	{0x53DC0F80109CFD5CULL, "vxlt_0_lowtown_n_d3s_0_0.vlm"},
	{0x53DC119997A749BAULL, "vxlt_gallowstemplar_480_0_0.vlm"},
	{0x53DD5206CA28C02CULL, "vxht_mansion2_1g9_0_0.vlm"},
	{0x53DD88D0091493F2ULL, "vxdr_primevalentrance_1kb_0_0.vlm"},
	{0x53DD9F09367864C1ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_fin_c.anb"},
	{0x53DDE92AE3EB370EULL, "466479428.ogg"},
	{0x53DDF278738B6B2AULL, "6146804_f.fxe"},
	{0x53DE132CD3AC2D04ULL, "art\\characters\\playercharacter\\humanfemale\\rob_maga.phy"},
	{0x53DE34F7FE1E01BEULL, "hlac_lt_0_lowtown_docks_n_-65537_0z.dds"},
	{0x53DE451635AC2904ULL, "6177694_m.fxe"},
	{0x53DE6044E73736FBULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward_rockwraith.evt"},
	{0x53DEE9B7CE42CBDEULL, "mcr310_argument6075112.cl"},
	{0x53E0AABD82B52BF1ULL, "898975005.ogg"},
	{0x53E1813F9F07FE2FULL, "ts_generic_wpn_props.gda"},
	{0x53E19557323361EDULL, "mag311_keran6041585.cl"},
	{0x53E1D4F4AA201ABAULL, "vxht_0_hightown_c44_0_0.vlm"},
	{0x53E1E37F7CD4CE40ULL, "vxw_sundermount_1yk_0_0.vlm"},
	{0x53E1F5715F67D670ULL, "vxw_woundedcoast_e2_0_0.vlm"},
	{0x53E2012908F2FB0FULL, "follower_banter6160690.cl"},
	{0x53E213E54E28A5C6ULL, "qun221_petrice6077945.cl"},
	{0x53E261F757A515F6ULL, "vxlt_0_lowtown_n_ep4_0_1.vlm"},
	{0x53E2C69CC4969E83ULL, "724830562.ogg"},
	{0x53E318B21E3D88ECULL, "vxlt_undercityserialkiller_1fo_0_0.vlm"},
	{0x53E36AB1D80760CAULL, "64066594.ogg"},
	{0x53E377FBBC0F0DC4ULL, "art\\vfx\\models\\vfx_mer_sword_2.msh"},
	{0x53E38DFBBC0F3366ULL, "art\\vfx\\models\\vfx_mer_sword_2.mmh"},
	{0x53E3BB72D82BF398ULL, "vxw_sundermount_2y0_0_0.vlm"},
	{0x53E3E611DDA5AFA6ULL, "743922104.ogg"},
	{0x53E49B20C77A0EC3ULL, "vxca_caverns1_mines_igo_0_0.vlm"},
	{0x53E4B582A76AED67ULL, "11875946.ogg"},
	{0x53E4C31512D62A63ULL, "29095134.ogg"},
	{0x53E4D169385B503BULL, "hlac_lt_0_lowtown_-262144_1y.dds"},
	{0x53E5265F4B1ABE7CULL, "mcr371_hostile_merrill6082158.cl"},
	{0x53E5289090256DDFULL, "1052482258.ogg"},
	{0x53E55944E73D493CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward_rockwraith.gad"},
	{0x53E55B57E9106449ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hed_frsa_0n.dds"},
	{0x53E63C01702351C6ULL, "hlac_dr_0_deeproads_5842_6y.dds"},
	{0x53E64757820D3469ULL, "6168909_m.fxe"},
	{0x53E64DDFD02A87EDULL, "vxlt_0_lowtown_a2_n_euu_0_0.vlm"},
	{0x53E66EF4D24D759AULL, "512961556.ogg"},
	{0x53E6F4F0D78AF7E9ULL, "and100_anders6111675.cl"},
	{0x53E6FB179243EB44ULL, "vrc321_varric6097151.cl"},
	{0x53E75C809C21C68DULL, "hlac_ht_chantry_n_2534_27x.dds"},
	{0x53E79A7AE66AF189ULL, "vxlt_gallowstemplar_f_6qi_0_0.vlm"},
	{0x53E7A6BF6C25F600ULL, "58325272.ogg"},
	{0x53E8AD4D0D69982AULL, "171493959.ogg"},
	{0x53E8BB76EFB48115ULL, "6049460_m.fxe"},
	{0x53E9120DE2C9777CULL, "clf260_harley6069954.cl"},
	{0x53E9EF91B571E83DULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcratepile_0.pwk"},
	{0x53E9FF6CD8AE43E4ULL, "zz_use_ability.ncs"},
	{0x53EA0631F2F458B4ULL, "vxca_cavern1_haunted_cpc_0_0.vlm"},
	{0x53EA0E8B70E86890ULL, "364674141.ogg"},
	{0x53EA71AD81AA8AF6ULL, "vxca_caverns2_981_0_0.vlm"},
	{0x53EABCD3F7E55150ULL, "gen00fl_varric6033590.cl"},
	{0x53EB41675BC9CA19ULL, "165807547.ogg"},
	{0x53EB800936846D3AULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_fin_c.mmh"},
	{0x53EBD655D503CED3ULL, "6179461_f.fxe"},
	{0x53EBFB3D163E57A9ULL, "vxht_mansion_n_753_0_0.vlm"},
	{0x53EC16A764AC17DEULL, "lowtown_male_alienage2_mob6166540.cl"},
	{0x53EC47E5C116E84EULL, "179726402.ogg"},
	{0x53EC5844E743663AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward_rockwraith.ani"},
	{0x53ED107F7CC6810DULL, "fx_fade_flip_cloud.dds"},
	{0x53ED1545C7BD688AULL, "vxw_encounter3_bg_0_0.vlm"},
	{0x53ED24CD59355253ULL, "vxca_caverns2_aig_0_0.vlm"},
	{0x53ED68E7BBCFC321ULL, "6119871_m.fxe"},
	{0x53EDEE8D514BC78BULL, "llac_lt_0_lowtown_n_-65536_3.dds"},
	{0x53EE4BF7757BB25AULL, "plt_lgt000pt_lite_darktown.plo"},
	{0x53EEED884EABC137ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphanga.mmh"},
	{0x53EEF077E85916F8ULL, "vxlt_hangedman_n_24b_0_0.vlm"},
	{0x53EEF1884EABC844ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphanga.mao"},
	{0x53EF03884EABE6A9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarphanga.msh"},
	{0x53EF351A61B1A5D1ULL, "6100312_m.fxe"},
	{0x53EF9EAD539C7CCDULL, "6100306_f.fxe"},
	{0x53EFF68CEE773D96ULL, "follower_banter6163577.cl"},
	{0x53F01642A08BCC0CULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnr180walkf.ani"},
	{0x53F0674E515DEEF0ULL, "vxw_woundedcoast_b3_0_0.vlm"},
	{0x53F0983758858C06ULL, "mag131_keran6086273.cl"},
	{0x53F0B9442A2B521CULL, "prp_lightbeam.dds"},
	{0x53F0DB04F661B0B7ULL, "131520313.ogg"},
	{0x53F10716882EB02CULL, "vxlt_0_lowtown_docks_a3_n_bmy_0_0.vlm"},
	{0x53F207C44C0DFF4BULL, "follower_barks6153916.cl"},
	{0x53F210F72FB27CFFULL, "887543150.ogg"},
	{0x53F238B2CAC8A7A2ULL, "darktown_male_fereldan_sta6175154.cl"},
	{0x53F291B5BF2E64EEULL, "dae000_mother6121059.cl"},
	{0x53F293D55362D5FFULL, "vxht_brothel_n_4a9_0_0.vlm"},
	{0x53F32CA4F6939498ULL, "1059431335.ogg"},
	{0x53F4226F3518BD42ULL, "vxlt_0_lowtown_e11_0_0.vlm"},
	{0x53F43E76EDBA656AULL, "vxht_0_hightown_a3_bfr_0_0.vlm"},
	{0x53F4C8EEB625D913ULL, "hlac_ca_caverns1_mines_24421_9y.dds"},
	{0x53F4D9C471C348CEULL, "hlac_lt_0_lowtown_-131072_1y.dds"},
	{0x53F4FDA67DE20714ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightpath_da.dds"},
	{0x53F503D989F5E2D3ULL, "6069590_m.fxe"},
	{0x53F50F4389CAE4A8ULL, "vxlt_0_lowtown_n_gp9_0_0.vlm"},
	{0x53F522F23D57093EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_filecabinet_n.dds"},
	{0x53F5B83D7C72A865ULL, "6092288_m.fxe"},
	{0x53F5D2BB075FF20CULL, "vxlt_0_lowtown_a2_gjl_0_0.vlm"},
	{0x53F6323846E9BEF7ULL, "181282087.ogg"},
	{0x53F6682E75501CC1ULL, "839509984.ogg"},
	{0x53F6A1B5BF369AB4ULL, "qun110_petrice6018570.cl"},
	{0x53F6C3BF13164494ULL, "590672504.ogg"},
	{0x53F72342A092014EULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnr180walkf.gad"},
	{0x53F7990D2CF65FC6ULL, "hm_pickpocket.mor"},
	{0x53F84A04382578C8ULL, "hlac_ht_mansion_n_3369_1z.dds"},
	{0x53F8524652660F46ULL, "vxlt_blackemporium_te_0_0.vlm"},
	{0x53F87A58CEC07C33ULL, "57187380.ogg"},
	{0x53F90A0BE0282E2FULL, "272822510.ogg"},
	{0x53F923F9B975935EULL, "vxlt_foundry_n_xr_0_0.vlm"},
	{0x53F9622017D5F631ULL, "vxca_caverns2_520_0_0.vlm"},
	{0x53F99FE3B105878AULL, "vxht_mansion2_30x_0_0.vlm"},
	{0x53FA8740F93AD7DBULL, "hlac_lt_0_lowtown_a3_n_c_-131071_2z.dds"},
	{0x53FA8BC3283EE435ULL, "hlac_lt_gallowstemplar_n_2384_1z.dds"},
	{0x53FB1D326DEE1754ULL, "llac_lt_undercityserialkiller_-1_46.dds"},
	{0x53FB62FDFE52B403ULL, "hlac_lt_0_lowtown_65534_1x.dds"},
	{0x53FBCBAEE2898FEBULL, "vxlt_hangedman_170_0_0.vlm"},
	{0x53FBDE19C46F04F4ULL, "6099345_m.fxe"},
	{0x53FBE6E7851409E3ULL, "static_static_p39.psh.fxo.dx11.assembly"},
	{0x53FC114EB68C351CULL, "6111217_f.fxe"},
	{0x53FC40540E2937EBULL, "478020747.ogg"},
	{0x53FC566F7D78164BULL, "hlac_lt_darktown_-1_50z.dds"},
	{0x53FC795B76ADD0F3ULL, "vxlt_ruinedhouse_n_1x_0_0.vlm"},
	{0x53FCE3E6BAF261CDULL, "lt_0_lowtown_docks_116_0.phy"},
	{0x53FD03AD202602A9ULL, "art\\characters\\creatures\\desiredemon\\animation\\tal_scare.ani"},
	{0x53FD59DCE9F0A422ULL, "vxdr_0_deeproads_cmx_0_0.vlm"},
	{0x53FD7D40CB133232ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_nrmo_0n.dds"},
	{0x53FE1E42A098172DULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnr180walkf.evt"},
	{0x53FE34F121954019ULL, "lowtown_male_urchin_sta6161583.cl"},
	{0x53FE60FD30778134ULL, "920593390.ogg"},
	{0x53FE898A3EE26CEBULL, "follower_banter6167949.cl"},
	{0x53FED17DAAC95EB2ULL, "293486.ogg"},
	{0x53FF62F86B4D6C4EULL, "373615106.ogg"},
	{0x53FFA0084BA61143ULL, "424588712.ogg"},
	{0x540034B479EF4B62ULL, "llac_ht_mansion_n_3371_0.dds"},
	{0x540043FB1A51862CULL, "6071822_m.fxe"},
	{0x5400ADC99B4EE01AULL, "6075687_f.fxe"},
	{0x5400BD01319B1B0BULL, "darktown_fem_fereldan_sta6152809.cl"},
	{0x5400BFA577EDB00EULL, "llac_w_bonepit_262150_0.dds"},
	{0x5400E543F7391387ULL, "193775289.ogg"},
	{0x540119377C868CEBULL, "hm_teagan.mor"},
	{0x5401E21911A3E536ULL, "vxht_0_hightown_n_c7f_0_0.vlm"},
	{0x5401E3DCB89C666CULL, "6000927_m.fxe"},
	{0x54020F39566C0444ULL, "424656821.ogg"},
	{0x540214052E6F02CCULL, "gen00fl_merrill6150709.cl"},
	{0x540238C284640EE5ULL, "6119844_m.fxe"},
	{0x54029681B5D0E121ULL, "750913499.ogg"},
	{0x540357EB2FA20D40ULL, "vxw_bonepit_1wv_0_0.vlm"},
	{0x54041EAD202C4F2BULL, "art\\characters\\creatures\\desiredemon\\animation\\tal_scare.gad"},
	{0x5404C123E8B88F4AULL, "zz_vault_debug6178948.cl"},
	{0x5404FDF0EA54731FULL, "vxdr_0_deeproads_d3d_0_0.vlm"},
	{0x540509EEB8660A44ULL, "vxlt_hangedman_263_0_0.vlm"},
	{0x54053008D2AE6871ULL, "6125937_m.fxe"},
	{0x540608CD57237421ULL, "mer221_sabin6047335.cl"},
	{0x54068E292E62E2BFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.open2.gad"},
	{0x5406E9EBD6E089B4ULL, "vxca_cavern1_haunted_cna_0_0.vlm"},
	{0x54088B6907753028ULL, "6125226_m.fxe"},
	{0x54098216F61A65F0ULL, "art\\characters\\creatures\\arh_abm_rev\\c_rev_sword.mao"},
	{0x5409A5E306A074E4ULL, "mag210_arianni6021242.cl"},
	{0x5409D795DAE99C75ULL, "6085536_m.fxe"},
	{0x540A001CC12B202EULL, "6084424_m.fxe"},
	{0x540A1BAD2030B590ULL, "art\\characters\\creatures\\desiredemon\\animation\\tal_scare.evt"},
	{0x540A626BEF96FCE5ULL, "vxht_0_hightown_a2_c5x_0_0.vlm"},
	{0x540ADD83EB6D92CAULL, "lt_0_lowtown_docks_mmap_ph.dds"},
	{0x540B3ABB39131904ULL, "hlac_lt_0_lowtown_docks_-65538_4z.dds"},
	{0x540B60094123BB58ULL, "174026462.ogg"},
	{0x540B635834D97DE3ULL, "llac_w_woundedcoast_-196603_0.dds"},
	{0x540C504670632135ULL, "428131578.ogg"},
	{0x540C7D841F5D42CEULL, "mrl150_merrill6085846.cl"},
	{0x540C8A0170009D57ULL, "918347852.ogg"},
	{0x540CAB6EB53E174DULL, "vxw_bonepitdestroyed_a3_di_0_0.vlm"},
	{0x540CB01BE325BF5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforward.gad"},
	{0x540CD972F33AC3BCULL, "6068293_m.fxe"},
	{0x540CDFEAA30DD441ULL, "zz_dog_debug6127749.cl"},
	{0x540D04CCD05C7D3BULL, "hlac_lt_0_lowtown_a2_n_c_-131073_3x.dds"},
	{0x540D4303550B8DD1ULL, "6161586_m.fxe"},
	{0x540D4B08354A1FF0ULL, "vxw_woundedcoast_t1_0_0.vlm"},
	{0x540E0BB99184108AULL, "6125530_m.fxe"},
	{0x540E634CBAB33038ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10b.phy"},
	{0x540E80A6BC563C34ULL, "vxht_0_hightown_a2_cco_0_0.vlm"},
	{0x540FB545AFE34353ULL, "water_water_p2.psh.fxo"},
	{0x540FD2350A8F50CEULL, "art\\vfx\\combat\\vfx_r_shd_decoy_imp_aoe_p.impa.evt"},
	{0x54104E75970059B5ULL, "vxht_mansion2b_1py_0_0.vlm"},
	{0x54107434113CA119ULL, "follower_banter6169336.cl"},
	{0x5410C9EEA6B47003ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeic.msh"},
	{0x5410D5EA2DEDE991ULL, "vxht_mansion_n_6bw_0_0.vlm"},
	{0x5410D7EEA6B4885DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeic.mmh"},
	{0x5410DBEEA6B48F6EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeic.mao"},
	{0x5411139DFABD7BF3ULL, "rdr111dg_martin6057034.cl"},
	{0x54111559AF36118AULL, "vxht_0_hightown_n_bxn_0_0.vlm"},
	{0x54118024876026FFULL, "llac_lt_0_lowtown_docks_a3_n_-262144_1.dds"},
	{0x5411DBD5555FC3C4ULL, "315490040.ogg"},
	{0x5412231A73E47885ULL, "hlac_w_woundedcoast_0_6x.dds"},
	{0x54137BCBAA5C22C4ULL, "vxlt_hangedman_230_0_0.vlm"},
	{0x5413ED26DB044F31ULL, "zz_lgt_1_debug6114391.cl"},
	{0x54142D2CEA06C729ULL, "192808926.ogg"},
	{0x541441153D925AFBULL, "6214897_m.fxe"},
	{0x54144238C6D03D65ULL, "539243344.ogg"},
	{0x54149CA12FE41777ULL, "vxht_chantry_2tm_0_1.vlm"},
	{0x5414F2C0E74DB31FULL, "623177597.ogg"},
	{0x54154C985A5B9C62ULL, "6071947_m.fxe"},
	{0x54155879E67A8C72ULL, "lt_gallowscourtyard_n_39_0.rml"},
	{0x54157A511CA3D1EBULL, "589683372.ogg"},
	{0x541584D5E1820AB3ULL, "qun120_arishok_guard6060800.cl"},
	{0x5415AF16495E7495ULL, "315729989.ogg"},
	{0x54166D23C02E9AEEULL, "vxlt_blackemporium_16b_0_0.vlm"},
	{0x5416AD0C88B1CBF3ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.anb"},
	{0x5416C9350A956079ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_imp_aoe_p.impa.gad"},
	{0x5416E47E01A6ED54ULL, "849209682.ogg"},
	{0x54175674A540D748ULL, "plt_cod_ltr_broker_hunter.nss"},
	{0x5417E38745775035ULL, "vxlt_0_lowtown_a2_n_gnn_0_0.vlm"},
	{0x54186568E1FA342AULL, "hlac_ht_0_hightown_a2_-131071_1x.dds"},
	{0x5418D448E18D1565ULL, "vxht_0_hightown_a2_c0g_0_0.vlm"},
	{0x5418D8953D40A7F8ULL, "vxht_bartrand_n_4hi_0_0.vlm"},
	{0x5418E315C73F6121ULL, "lowtown_male_lowlife1_sta6164962.cl"},
	{0x5419109790646CFBULL, "art\\characters\\playercharacter\\humanmale\\boo_ric.mmh"},
	{0x54191C9790648098ULL, "art\\characters\\playercharacter\\humanmale\\boo_ric.mao"},
	{0x5419227355F47E2EULL, "6153889_m.fxe"},
	{0x54192E9790649FE5ULL, "art\\characters\\playercharacter\\humanmale\\boo_ric.msh"},
	{0x541AB712CCDDD657ULL, "6208427_m.fxe"},
	{0x541AED69F3156AEBULL, "llac_lt_warehouse_65537_4.dds"},
	{0x541B479441E1CB81ULL, "631713341.ogg"},
	{0x541B6B292E750305ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.open2.ani"},
	{0x541B6E2E3EC1CAEBULL, "vxht_mansion_n_504_0_0.vlm"},
	{0x541B716EAAA39A4EULL, "6102708_m.fxe"},
	{0x541B944BA7EB0F6CULL, "6174310_m.fxe"},
	{0x541BCC5017C657DDULL, "738122242.ogg"},
	{0x541C26BF4CF0E771ULL, "lit_aoe_05_l3.mao"},
	{0x541CB902D4F05C68ULL, "6030029_m.fxe"},
	{0x541CDA350A99E8C7ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_imp_aoe_p.impa.ani"},
	{0x541D0BEE8EE0BC60ULL, "t1_mul_ta1.tnt"},
	{0x541D459BA8BDBE70ULL, "hlac_lt_0_lowtown_n_65534_1z.dds"},
	{0x541E490BB39666F3ULL, "llac_w_encounter2_-65531_0.dds"},
	{0x541E5F74A547055BULL, "plt_cod_ltr_broker_hunter.plo"},
	{0x541E6EACAB269168ULL, "vxht_chantry_n_2y3_0_0.vlm"},
	{0x541E87E67BEC1F37ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dcast_end.gad"},
	{0x541ED2936D2553E9ULL, "6141132_f.fxe"},
	{0x541EE3A11EA7FB00ULL, "mrl150_merrill6093528.cl"},
	{0x541F045FB771594EULL, "133306882.ogg"},
	{0x541F25FE77336AEBULL, "gen00fl_varric6076084.cl"},
	{0x541F8B112CA362DFULL, "vxlt_0_lowtown_n_ewa_0_0.vlm"},
	{0x5420166136720D2FULL, "6202326_m.fxe"},
	{0x54204D06AC685DF8ULL, "gen00fl_isabela_init6077331.cl"},
	{0x5420785959B63E2CULL, "6059411_m.fxe"},
	{0x54208D1BE3362CA5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforward.ani"},
	{0x5420A9D0C9FA2152ULL, "6169906_f.fxe"},
	{0x5420AD22C41D141AULL, "vxca_caverns2_8pl_0_0.vlm"},
	{0x5420D7623D0D7909ULL, "1059128627.ogg"},
	{0x5420ED2C8367BB39ULL, "vxdr_primevaltomb_4he_0_0.vlm"},
	{0x5420FC9EBE3842E0ULL, "448915650.ogg"},
	{0x542122B8501EF64EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rip_out_heart.evt"},
	{0x5421F4CDB5854A07ULL, "mcr310_sibling6095350.cl"},
	{0x5421FBD343CD09EAULL, "vxht_mansion_n_73z_0_0.vlm"},
	{0x5422777FDECC964AULL, "6090366_m.fxe"},
	{0x5422BDC216CDDBF6ULL, "281367780.ogg"},
	{0x542303EB65CAAFA9ULL, "vxlt_hangedman_301_0_0.vlm"},
	{0x542305B164FD6CD7ULL, "vxlt_0_lowtown_gbw_0_0.vlm"},
	{0x542324ABDD9A61B0ULL, "gen00fl_aveline6139730.cl"},
	{0x54234111A5CC84DBULL, "292950749.ogg"},
	{0x54235D749F622F14ULL, "plt_genpt_encounters.plo"},
	{0x54236A8FDE25B340ULL, "6118565_m.fxe"},
	{0x54238126682688F3ULL, "6129737_m.fxe"},
	{0x5424915C48EBEF1EULL, "94146954.ogg"},
	{0x54249CE67BF0AEC4ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dcast_end.evt"},
	{0x5424CA0C88BE3A5CULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.mmh"},
	{0x5424E120580945A1ULL, "755255112.ogg"},
	{0x54255B65FDFAFAEBULL, "debugoptions.gfx"},
	{0x5425AD51CF6FA89AULL, "ave331_jeven6138765.cl"},
	{0x5426750D53CADF5FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdazed.gad"},
	{0x542747EC33FFC5D7ULL, "vxlt_0_lowtown_a2_n_c_i1e_0_0.vlm"},
	{0x5427968847880852ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shielda_d.dds"},
	{0x5427D86DEF6ADC7DULL, "6151350_m.fxe"},
	{0x5427EBD2437CB8BFULL, "ser221_gamlen6121603.cl"},
	{0x5428065C00991828ULL, "hlac_ht_keep_n_3739_0z.dds"},
	{0x542819B8502505F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rip_out_heart.gad"},
	{0x5428627CB4F0A718ULL, "6070025_m.fxe"},
	{0x5428A51B1880E04FULL, "dae000ip_crowd_em4_talk11.utp"},
	{0x54292D1E49593B5CULL, "hlac_lt_0_lowtown_a2_n_131070_4z.dds"},
	{0x54295E06CACDDDCBULL, "art\\vfx\\textures\\vfx_vollightgrad.dds"},
	{0x542977B5E99999A9ULL, "cd_sib100_father_contact.bnk"},
	{0x5429920B0826F666ULL, "220776266.ogg"},
	{0x5429F5E824AFE3DEULL, "998731364.ogg"},
	{0x542A5824E374F737ULL, "dae000cr_grand_cleric6147891.cl"},
	{0x542BECC5144DC2B5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpouta.msh"},
	{0x542BEEC5144DC62BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpouta.mmh"},
	{0x542BF4167D383AD7ULL, "471312759.ogg"},
	{0x542BFAC5144DDAC8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpouta.mao"},
	{0x542C24F8AD103934ULL, "439907558.ogg"},
	{0x542C70C5CE5740A5ULL, "6098933_f.fxe"},
	{0x542D074649E70D96ULL, "hlac_dr_dungeon_3215_2y.dds"},
	{0x542D145DFD096DBFULL, "art\\characters\\playercharacter\\humanmale\\bdy_rob_maga.mao"},
	{0x542D4262963C158EULL, "vxlt_0_lowtown_ewo_0_0.vlm"},
	{0x542D622C7C824ADCULL, "6206952_m.fxe"},
	{0x542D6A0D53D0EB4CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdazed.evt"},
	{0x542D72B9417460E0ULL, "zz_followers_banter6157310.cl"},
	{0x542DB9A411DA1A91ULL, "vxht_mansion_n_6hu_0_0.vlm"},
	{0x542E1007779B5FBDULL, "mcr371_angry_mage_m6087031.cl"},
	{0x542E16C23DE323A0ULL, "mcr370_meredith6136179.cl"},
	{0x542E2AB850298E47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rip_out_heart.ani"},
	{0x542E56C1C252EB40ULL, "hlac_dr_primevaltomb_658_5z.dds"},
	{0x542EA13DC3E12E1EULL, "hlac_lt_gallowscourtyard_-1_2x.dds"},
	{0x542EA667946CAA17ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3a.mmh"},
	{0x542EAA67946CB124ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3a.mao"},
	{0x542EBC67946CCF89ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf3a.msh"},
	{0x542FA4A23CB2732BULL, "zz_vrc_debug6200250.cl"},
	{0x5430778540C1B7C5ULL, "6168149_m.fxe"},
	{0x5430C2FF6527DEFCULL, "vxlt_gallowstemplar_4uq_0_0.vlm"},
	{0x5430E67635620D4CULL, "679771609.ogg"},
	{0x54310A7E0C63A666ULL, "6174730_m.fxe"},
	{0x54315AE804EDC2B4ULL, "6101394_m.fxe"},
	{0x5431805ACE9ADE56ULL, "6187934_m.fxe"},
	{0x5431F735644DB636ULL, "hlac_w_bonepitmining_a2_131072_0z.dds"},
	{0x5431F7C5ACD50F73ULL, "113435846.ogg"},
	{0x54326A891F2F6E62ULL, "6191246_f.fxe"},
	{0x5432702E3D308C92ULL, "fx_headsmoke.dds"},
	{0x54327D3897765A2EULL, "mrl000_marethari6129348.cl"},
	{0x543281AAD5225989ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardleft.ani"},
	{0x5432B4E67BFD14BDULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dcast_end.ani"},
	{0x54337A9D134B9225ULL, "6108082_m.fxe"},
	{0x54339987B475C685ULL, "6110968_m.fxe"},
	{0x5433B45DBEEA6E49ULL, "vxlt_0_lowtown_h2b_0_0.vlm"},
	{0x5433C9124E3473ABULL, "vxht_0_hightown_a3_4kf_0_0.vlm"},
	{0x5434EF49229F1C7FULL, "vxlt_undercityserialkiller_1ih_0_1.vlm"},
	{0x5435377B96E79C97ULL, "vxlt_hangedman_rn_0_0.vlm"},
	{0x5435B32FDB05068BULL, "llac_lt_gallowstemplar_n_2384_5.dds"},
	{0x5435D4C3E7E653E3ULL, "water_pan_blur2.mao"},
	{0x543626C98AAA5319ULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_fleshpod_01d.dds"},
	{0x5436B19B8E8A27EEULL, "vrc321_party_barks6095765.cl"},
	{0x5436C75B82E249F1ULL, "vxlt_0_lowtown_n_e1i_0_0.vlm"},
	{0x5436D92B50C4ACDBULL, "vxht_mansion2b_1wr_0_0.vlm"},
	{0x5436ED4BE5EF1718ULL, "zz_qcr_debug6058844.cl"},
	{0x5437669251F06A83ULL, "hlac_ht_0_hightown_n_-65538_0y.dds"},
	{0x54378E36C416BBE7ULL, "hlac_w_woundedcoast_-65530_0x.dds"},
	{0x54379FDEDDD51D43ULL, "hlac_lt_0_lowtown_a2_n_c_65533_1x.dds"},
	{0x5437F3B52CC37B0FULL, "vxlt_0_lowtown_docks_c6j_0_0.vlm"},
	{0x5437FD6E72EBCDAFULL, "6126884_m.fxe"},
	{0x54381EB5F1421604ULL, "6108106_m.fxe"},
	{0x543886CF30409F2BULL, "43740810.ogg"},
	{0x54388D1131B7EE71ULL, "782732721.ogg"},
	{0x5438B30B211477B0ULL, "598620561.ogg"},
	{0x5439470B3B9B6B17ULL, "dae000_orana6149126.cl"},
	{0x543952DA842A06B1ULL, "vxht_0_hightown_a2_b02_0_0.vlm"},
	{0x5439783ED8B25896ULL, "hlac_lt_0_lowtown_n_-65535_3x.dds"},
	{0x54399896E9C77639ULL, "hlac_ca_caverns1_mines_24421_8x.dds"},
	{0x54399CAAD528A60BULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardleft.gad"},
	{0x543A5026C138340FULL, "plt_lgt000pt_baz_shrift.plo"},
	{0x543A520D53DB4CA5ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdazed.ani"},
	{0x543B18B3E5835CAEULL, "6043835_f.fxe"},
	{0x543B476EB73202EEULL, "6156617_m.fxe"},
	{0x543C3B4C49B76DB2ULL, "vxht_bartrand_n_5sk_0_0.vlm"},
	{0x543C3F0A6D55A348ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masa_0d.dds"},
	{0x543C570B918CF755ULL, "one100dg_elren6014720.cl"},
	{0x543C79C941C265F8ULL, "art\\characters\\playercharacter\\humanmale\\glv_dksa.phy"},
	{0x543C856FC4E17FE1ULL, "character_character_p93.psh.fxo.dx11.assembly"},
	{0x543CB21DDE6C9536ULL, "vxht_0_hightown_a2_c1s_0_0.vlm"},
	{0x543CC706B6F84CE5ULL, "6093557_m.fxe"},
	{0x543D1F5EF77A6F40ULL, "vxw_bonepit_1kr_0_0.vlm"},
	{0x543D8CA305447939ULL, "762603526.ogg"},
	{0x543DC2C3BF34C3B1ULL, "hlac_ht_chantry_2534_51x.dds"},
	{0x543E18732D736F5CULL, "6016577_m.fxe"},
	{0x543EA1E461C0D46AULL, "540556209.ogg"},
	{0x543EE165031B88CCULL, "1072279408.ogg"},
	{0x543F737C265676C2ULL, "vxht_mansion_n_62h_0_0.vlm"},
	{0x543F9B3698FEDC16ULL, "gen00fl_sebastian6180420.cl"},
	{0x543F9F01D6B6811BULL, "vxw_sundermount_37q_0_1.vlm"},
	{0x54401479C34FB86BULL, "art\\vfx\\combat\\vfx_r_var_pinpoint_str_c.mmh"},
	{0x5440374710BD69C9ULL, "6080644_f.fxe"},
	{0x5440379CAF3A0540ULL, "lgt110_orz_harrowmont6109007.cl"},
	{0x54408FEB4A703992ULL, "6027768_m.fxe"},
	{0x544099AAD52EC070ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardleft.evt"},
	{0x5440E14FE037B612ULL, "6108669_m.fxe"},
	{0x5441182AA6E8E481ULL, "vrc321_housekeeper6087295.cl"},
	{0x5441DDEBA22625F1ULL, "pro000_guard6005645.cl"},
	{0x54424F29A48AC3C7ULL, "vxca_caverns1_mines_imq_0_0.vlm"},
	{0x54426C0C5218799BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_05ruinstower.phy"},
	{0x5442808D88FC218BULL, "6108967_m.fxe"},
	{0x54429D136E47559FULL, "59752889.ogg"},
	{0x54431EAF1D1DB408ULL, "299833680.ogg"},
	{0x544342102B37377CULL, "vxht_0_hightown_a3_aqz_0_0.vlm"},
	{0x544361A81EE9538CULL, "gen_im_npc_arm_hum_boo_rdrd.uti"},
	{0x5444F8376996C87CULL, "vxht_0_hightown_cdr_0_0.vlm"},
	{0x544503947EAA6103ULL, "vxlt_gallowsprison_n_64h_0_0.vlm"},
	{0x54452765ED27810FULL, "3480282.ogg"},
	{0x54455CD55C34146CULL, "6185149_m.fxe"},
	{0x54460081BE0BF41EULL, "6060093_m.fxe"},
	{0x54463608DB2AC01AULL, "vxlt_gallowstemplar_5jg_0_0.vlm"},
	{0x544665921EB9938FULL, "vxlt_0_lowtown_docks_c3u_0_0.vlm"},
	{0x544674475360C385ULL, "llac_lt_0_lowtown_a2_n_65534_6.dds"},
	{0x544740A9824945E1ULL, "32231068.ogg"},
	{0x54479CC008D0933BULL, "6215890_m.fxe"},
	{0x54483B98DE84ABA0ULL, "hlac_lt_warehouse_0_1z.dds"},
	{0x54485F119EC8EF87ULL, "gen00fl_fenris6112081.cl"},
	{0x5448ABB7B678702CULL, "zz_t_tvspot_spiders.cut"},
	{0x54492B9D9CEEA9DDULL, "305620775.ogg"},
	{0x5449970E3BE953E0ULL, "6016642_f.fxe"},
	{0x5449B86932AB2C3FULL, "102867865.ogg"},
	{0x544A5C86B20EB840ULL, "vxca_caverns2_4o3_0_0.vlm"},
	{0x544AC35081F47673ULL, "295636400.ogg"},
	{0x544AE2D098376BEFULL, "zz_vrc_debug6064254.cl"},
	{0x544B010EA1B8F79AULL, "llac_lt_0_lowtown_a2_65533_1.dds"},
	{0x544B913BE97087C0ULL, "vxw_bonepit_1ls_0_0.vlm"},
	{0x544BB98ED8FE5B08ULL, "904928916.ogg"},
	{0x544C3E67DC864111ULL, "mrl310_party_comments6142744.cl"},
	{0x544C57FC786F473EULL, "follower_banter6173374.cl"},
	{0x544CF579C35A0DD0ULL, "art\\vfx\\combat\\vfx_r_var_pinpoint_str_c.anb"},
	{0x544D0AF03D042D62ULL, "6166540_m.fxe"},
	{0x544D493A960960F7ULL, "dae000_keep_servant_1_sta6156873.cl"},
	{0x544D7DE92A71AA24ULL, "vxw_sundermount_2zt_0_1.vlm"},
	{0x544E2CA971B0C4CBULL, "21126360.ogg"},
	{0x544E6334C6462A30ULL, "dre000_cut_bartrandbetrayal6029370.cl"},
	{0x544EB30F952AC189ULL, "6140002_m.fxe"},
	{0x544EF85C868B9E08ULL, "vxlt_warehouse_cr_0_0.vlm"},
	{0x544F359AB62833D9ULL, "vxht_keep_n_639_0_1.vlm"},
	{0x544F72064B0EA918ULL, "948424087.ogg"},
	{0x544F77DD76BED4F5ULL, "vxdr_primevaltomb_7t6_0_0.vlm"},
	{0x544FB9EF31E3ADBEULL, "892325270.ogg"},
	{0x544FC2D14B3351CCULL, "rog_bleeddamage.dds"},
	{0x544FD41DB39C8D1CULL, "hlac_ht_0_hightown_a3_65536_0y.dds"},
	{0x545009C581B61742ULL, "hlac_lt_0_lowtown_n_-131071_1y.dds"},
	{0x5450C1069680DC47ULL, "vxca_caverns1_mines_ijp_0_0.vlm"},
	{0x54512B090118DD05ULL, "nat320_delilah6065886.cl"},
	{0x54513B7822370A10ULL, "mer120_hubert6077143.cl"},
	{0x545207FDCDFCBDC4ULL, "ht_0_hightownexplosion_n_35_0.rml"},
	{0x545258FD0DA6F61BULL, "vxw_encounter2_camp_ne_0_0.vlm"},
	{0x5452AA5E72506556ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_harrowing.phy"},
	{0x5452BC54B2C1D3E9ULL, "vxw_encounter2_camp_pp_0_1.vlm"},
	{0x5453B4260D282D48ULL, "hlac_w_encounter3_-327678_0y.dds"},
	{0x545431F353E9ED7BULL, "lt_0_lowtown_docks_n_141_0.msh"},
	{0x545443F353EA0BC6ULL, "lt_0_lowtown_docks_n_141_0.mao"},
	{0x54544FF353EA2065ULL, "lt_0_lowtown_docks_n_141_0.mmh"},
	{0x54546BA606D8254BULL, "chargen_stage2.gfx"},
	{0x54548C3469747675ULL, "hlac_lt_gallowstemplar_2394_7z.dds"},
	{0x5454F08AA58CEE80ULL, "vxca_caverns2_9fu_0_0.vlm"},
	{0x54556D6B32372DEFULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv2.gad"},
	{0x54557933A9AC94F8ULL, "6042122_f.fxe"},
	{0x5455EE7C2F79DB60ULL, "18796955.ogg"},
	{0x545640A7250583C5ULL, "mag110_arianni6036792.cl"},
	{0x545670BE1FED9274ULL, "dre000_varric6208052.cl"},
	{0x5456EC6D0EE99371ULL, "5244608.ogg"},
	{0x545717AA31D96EA0ULL, "dae000ip_crowd_em3_talk6.utp"},
	{0x5457C52E420B9FCCULL, "hlac_ca_caverns2_5802_45x.dds"},
	{0x5458249E0A021040ULL, "art\\vfx\\materials\\vfx_smoke_puffy2_add.mao"},
	{0x54587029CDBB50A9ULL, "41674058.ogg"},
	{0x54587669FA41D538ULL, "llac_dr_primevaltomb_659_11.dds"},
	{0x5458F096380FD72EULL, "6032195_m.fxe"},
	{0x54591D16650D36C3ULL, "899479049.ogg"},
	{0x54596CCC581D0D03ULL, "6155796_m.fxe"},
	{0x54597AC81CDA9997ULL, "follower_banter6164874.cl"},
	{0x54598D501ED588E2ULL, "6041536_f.fxe"},
	{0x545A32EF04CD3FDAULL, "vxht_keep_n_707_0_0.vlm"},
	{0x545AF6B4E6EB857AULL, "vxlt_hangedman_28o_0_0.vlm"},
	{0x545BD0BD4B06ACB6ULL, "6090432_m.fxe"},
	{0x545C826B323D6F9CULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv2.evt"},
	{0x545D0DB5452C6756ULL, "hlac_lt_0_lowtown_n_-65534_0z.dds"},
	{0x545D5824B6F2711DULL, "6021219_m.fxe"},
	{0x545D935E725A45F7ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_harrowing.msh"},
	{0x545DA55E725A645AULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_harrowing.mao"},
	{0x545DA63E72ABEF0DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl02.mmh"},
	{0x545DA95E725A6B69ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_harrowing.mmh"},
	{0x545DB83E72AC0D93ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl02.msh"},
	{0x545DFDB27432C5DAULL, "hlac_lt_hangedman_-1_2y.dds"},
	{0x545E0DBF69E4166CULL, "6050179_m.fxe"},
	{0x545ECBB56EDB1D13ULL, "follower_banter6169781.cl"},
	{0x545ED60B5DDC138DULL, "vxw_sundermount_27o_0_0.vlm"},
	{0x545F2388F46056CEULL, "docks_male_sailor_mob6176666.cl"},
	{0x545F32E38876F4C7ULL, "vxca_caverns1_mines_igs_0_0.vlm"},
	{0x545F824A7C3EB016ULL, "vxht_keep_63a_0_1.vlm"},
	{0x545F82E43360005BULL, "isa311_isabela6095388.cl"},
	{0x545F9C257FA296D8ULL, "ave331_jeven6137285.cl"},
	{0x5460319790A0A730ULL, "art\\characters\\playercharacter\\humanmale\\boo_ric.phy"},
	{0x5460CAD9FF9D0E9BULL, "vxw_encounter2_camp_id_0_0.vlm"},
	{0x5460D993F3C8E7ACULL, "mag121_idunna6083978.cl"},
	{0x5460DC6EB36F23A4ULL, "art\\characters\\playercharacter\\humanmale\\boo_bda.msh"},
	{0x5460EC39B2C90B61ULL, "vxw_encounter1_cm_0_0.vlm"},
	{0x5460F26EB36F4946ULL, "art\\characters\\playercharacter\\humanmale\\boo_bda.mmh"},
	{0x5460F4F0B9227261ULL, "vxlt_hangedman_n_rq_0_0.vlm"},
	{0x5461376D9171D151ULL, "427655835.ogg"},
	{0x5461AE643E1D24EBULL, "dae000_quintus6154575.cl"},
	{0x546223F239F7B380ULL, "6014998_f.fxe"},
	{0x546319C2BF16F11AULL, "vxlt_gallowstemplar_5pa_0_0.vlm"},
	{0x5463CE5BEC088ABBULL, "hlac_w_sundermount_-327684_0y.dds"},
	{0x5463D9867FD578F3ULL, "vxlt_undercitydungeon_1n9_0_0.vlm"},
	{0x5464446C5D7E81E9ULL, "vxht_mansion_n_60e_0_0.vlm"},
	{0x54646381C1E65CD8ULL, "6066611_m.fxe"},
	{0x54655CFF81388432ULL, "6059141_m.fxe"},
	{0x5465B0F463F4326AULL, "vxca_caverns2prc_9y7_0_0.vlm"},
	{0x5465C0A8C1CF6C7FULL, "187342921.ogg"},
	{0x5465E52D461FFE2FULL, "vxlt_warehouse_n_15e_0_0.vlm"},
	{0x5466D14DD6F5D384ULL, "vxht_0_hightown_n_c38_0_0.vlm"},
	{0x54674843A908A2FDULL, "392402665.ogg"},
	{0x54679D56CBBEBC45ULL, "vxca_caverns1_mines_e4l_0_0.vlm"},
	{0x5467F54038AB8031ULL, "follower_barks6153200.cl"},
	{0x54692D9469C65D83ULL, "lgt300cr_gang_5_ht.utc"},
	{0x5469304094CD8DF9ULL, "vxw_sundermount_34n_0_1.vlm"},
	{0x54696891D8E19DFAULL, "vxlt_hangedman_25h_0_0.vlm"},
	{0x5469872A6890A1B4ULL, "w_bonepitmining_a2_1_0.rml"},
	{0x54698A6B324807B5ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv2.ani"},
	{0x5469DE8D95C51F27ULL, "gen00fl_fenris6110797.cl"},
	{0x546A484121ADEF47ULL, "vxht_mansion2b_31b_0_0.vlm"},
	{0x546A504CBB016ECDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10b.msh"},
	{0x546A624CBB018D53ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10b.mmh"},
	{0x546A74EF741DB748ULL, "vxw_encounter1_ch_0_0.vlm"},
	{0x546A7DA67A771E48ULL, "vxlt_hangedman_n_rt_0_0.vlm"},
	{0x546AF05B638C4F3FULL, "ser120_ghyslain.bnk"},
	{0x546BB00C44C96DC0ULL, "km_sabina_brat.mor"},
	{0x546BE96EB3794111ULL, "art\\characters\\playercharacter\\humanmale\\boo_bda.phy"},
	{0x546C25E3349482EFULL, "gen00fl_aveline6002525.cl"},
	{0x546C4B30B3FCA2E8ULL, "6086244_m.fxe"},
	{0x546C7CBC29E69C82ULL, "v_c_blood_06.dds"},
	{0x546CB69AF56525E1ULL, "vxca_caverns1_mines_e97_0_0.vlm"},
	{0x546DD25B638E5247ULL, "ser120_ghyslain.cnv"},
	{0x546E5B644B57A3D8ULL, "6139951_m.fxe"},
	{0x546E6C8CEFD8FF85ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_min_c.anb"},
	{0x546E8DCEBC228B08ULL, "45925377.ogg"},
	{0x546E9F3E72BA56CAULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl02.phy"},
	{0x546F160A4964450AULL, "vxw_encounter3_7x_0_0.vlm"},
	{0x546F3CB6F193271BULL, "vxw_encounter2_camp_dg_0_0.vlm"},
	{0x546F530C523F3D3AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_05ruinstower.msh"},
	{0x546F58A3BDF5E39DULL, "seb221st_flora_2.stg"},
	{0x546F5E16A4BF23E1ULL, "vxw_encounter1_hl_0_0.vlm"},
	{0x546F610C523F5573ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_05ruinstower.mao"},
	{0x546F650C523F5B80ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_05ruinstower.mmh"},
	{0x546FB210AC2B7576ULL, "fx_snowvolume.mao"},
	{0x547010B5437A0ED6ULL, "77530929.ogg"},
	{0x54704896DA470E40ULL, "48409940.ogg"},
	{0x5470A8C3980CF543ULL, "886994348.ogg"},
	{0x5470CBA3DBFBCF46ULL, "lgt320_shw_ors_secret_thug.cnv"},
	{0x5470DFF6C773BADAULL, "6166811_f.fxe"},
	{0x547155B5E9D715FBULL, "cd_sib100_father_contact.wbd"},
	{0x5471DD52D024E5C7ULL, "vxlt_0_lowtown_a2_n_e5w_0_0.vlm"},
	{0x5471E4224FFDCACDULL, "w_bonepitdestroyed_a3_131_0.mmh"},
	{0x5471E8224FFDD11EULL, "w_bonepitdestroyed_a3_131_0.mao"},
	{0x5471F6224FFDE953ULL, "w_bonepitdestroyed_a3_131_0.msh"},
	{0x5472176131A82C62ULL, "vxlt_house_fp_0_0.vlm"},
	{0x5474BC7050842003ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_coward.anb"},
	{0x5474CDA3DBFFBB06ULL, "lgt320_shw_ors_secret_thug.bnk"},
	{0x5475019683466A16ULL, "lt_gallowscourtyard_51_0_lc.mmh"},
	{0x5475432AC8EBEC04ULL, "vxht_0_hightown_n_c89_0_0.vlm"},
	{0x5475858F0AE96ECCULL, "604880114.ogg"},
	{0x54758FC6B8664819ULL, "6109427_m.fxe"},
	{0x5475EE53D5D9FDC4ULL, "6083405_m.fxe"},
	{0x5475F62A58B4AAAAULL, "vxlt_blackemporium_ti_0_0.vlm"},
	{0x54763F50B6B9D1C8ULL, "vxca_caverns1_mines_e92_0_0.vlm"},
	{0x547693C122FBC653ULL, "441150237.ogg"},
	{0x5476A39607840177ULL, "llac_lt_0_lowtown_n_131071_4.dds"},
	{0x5476FFB9944F4E65ULL, "122619413.ogg"},
	{0x54770E9AE694F84CULL, "gen00fl_isabela6079084.cl"},
	{0x5477443E2CB1BADDULL, "158265804.ogg"},
	{0x5477A21D86C3A679ULL, "vxw_sundermount_33o_0_1.vlm"},
	{0x5477BC6D3E642C90ULL, "255901930.ogg"},
	{0x5477DA6ECAD7B67AULL, "vxlt_hangedman_22i_0_0.vlm"},
	{0x5478677619E5C0E2ULL, "w_sundermount_47_0.msh"},
	{0x5478697619E5C398ULL, "w_sundermount_47_0.mmh"},
	{0x5478757619E5D83BULL, "w_sundermount_47_0.mao"},
	{0x5478E6CC6613CFC8ULL, "vxw_encounter1_hi_0_0.vlm"},
	{0x5479795CC227419EULL, "6090829_m.fxe"},
	{0x547A2CE6E09823E1ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2_tool.phy"},
	{0x547A4DF35C6DE788ULL, "vxlt_warehouse_rq_0_0.vlm"},
	{0x547B697619E7F9ABULL, "w_sundermount_47_0.rml"},
	{0x547BAAE40BBC897CULL, "6060006_m.fxe"},
	{0x547C07924EC7A90FULL, "zz_facefx_test6088491.cl"},
	{0x547C485CE4B24879ULL, "vxht_keep_n_5k7_0_0.vlm"},
	{0x547C7178C4A94FDEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wax_d.dds"},
	{0x547D9BA52BEB6CC9ULL, "dae000_hate_enemies6167584.cl"},
	{0x547E0A645087393EULL, "6107454_f.fxe"},
	{0x547E7B40DDE595EEULL, "one110_meeran6089851.cl"},
	{0x547E99E6450C97D7ULL, "vxht_chantry_n_31u_0_0.vlm"},
	{0x547EACAD03354F87ULL, "349670044.ogg"},
	{0x547F3E70F2FD1324ULL, "6088138_m.fxe"},
	{0x547F6CD622A50F37ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelarge_0.pwk"},
	{0x547F98D64F168B9FULL, "6017898_m.fxe"},
	{0x54801DADF5A75003ULL, "6107279_f.fxe"},
	{0x54803D11AD795046ULL, "6120151_m.fxe"},
	{0x5480AB2BB6CE3D05ULL, "fen310_fenris6118217.cl"},
	{0x5480EB7125BA9FCBULL, "hlac_ca_caverns2_5802_68z.dds"},
	{0x54810AA305B975E4ULL, "6071930_m.fxe"},
	{0x54811478495AFFBDULL, "181345784.ogg"},
	{0x5481D96A5F6881F6ULL, "hlac_lt_0_lowtown_a2_131073_2x.dds"},
	{0x548221CAF5892F50ULL, "vxca_cave_2ix_0_0.vlm"},
	{0x5482C2AD59EBA791ULL, "194431238.ogg"},
	{0x5482EF4664CDD040ULL, "vxdr_primevaltomb_4h1_0_1.vlm"},
	{0x54830FD1381A0DBAULL, "dae300_alistair_king6134919.cl"},
	{0x548326EBC4E5CB70ULL, "vxca_cavern1_haunted_jfp_0_0.vlm"},
	{0x54835A392FA5F3A4ULL, "lgt210cr_eau_senestra.utc"},
	{0x548401A3AA5C6099ULL, "6102527_m.fxe"},
	{0x54846B3CA3D0F439ULL, "framebuffereffects_illuminationmask_p0.vsh.fxo.dx11"},
	{0x5485312EBFDBF7FCULL, "hlac_ca_cavern1_haunted_24424_4x.dds"},
	{0x54856F26C17854F4ULL, "plt_lgt000pt_baz_shrift.nss"},
	{0x54859A4722B71681ULL, "6111413_m.fxe"},
	{0x5485EE58EB6014C5ULL, "vxlt_hangedman_28d_0_0.vlm"},
	{0x548734A2F862BE6BULL, "238705161.ogg"},
	{0x548763A567BB091AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_partycall_d.dds"},
	{0x548785B9B6B551FBULL, "6064931_m.fxe"},
	{0x5487918A2BB4B61DULL, "385753154.ogg"},
	{0x5487A83D4CBEB862ULL, "llac_ht_chantry_2534_5.dds"},
	{0x5487E68605F068D8ULL, "6165723_m.fxe"},
	{0x5487ED52A4E2ABC6ULL, "sse_ss_arcanehorror_1.gda"},
	{0x54881F72F6DF95F4ULL, "6060608_m.fxe"},
	{0x54883E9707FBA1DFULL, "322351894.ogg"},
	{0x548868837F359385ULL, "73200951.ogg"},
	{0x5488A172451FB604ULL, "6083997_m.fxe"},
	{0x5488B31925A918B1ULL, "dae300_cullen6140593.cl"},
	{0x5488BFD04E640008ULL, "vxlt_warehouse_wn_0_0.vlm"},
	{0x54891FB772348CDAULL, "hlac_ht_mansion2_818_54y.dds"},
	{0x5489918A0619472EULL, "vxdr_0_deeproads_ev7_0_0.vlm"},
	{0x548A128B6B92A934ULL, "653122687.ogg"},
	{0x548A2659FA1B50FFULL, "vxht_bartrand_n_700_0_1.vlm"},
	{0x548A4C212BBB4DD7ULL, "lowtown_male_drunk_mob6151172.cl"},
	{0x548A54A2D57D6A29ULL, "lt_0_lowtown_docks_8_0.rml"},
	{0x548AA080AA1BF51DULL, "6145122_m.fxe"},
	{0x548AB2D5CCF1EEB2ULL, "gen_im_trs_crc_sta_spi_r2.uti"},
	{0x548B61D1517BAEE2ULL, "14134933.ogg"},
	{0x548BD6E0BF4D07EEULL, "vxht_keep_n_6am_0_0.vlm"},
	{0x548BD8249FACE642ULL, "6057118_m.fxe"},
	{0x548C6E9109EC5D26ULL, "6095813_m.fxe"},
	{0x548CAC32465BA4EAULL, "vxlt_gallowscourtyard_7fo_0_0.vlm"},
	{0x548CF521F5242D89ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair1_d.dds"},
	{0x548D212713071057ULL, "llac_lt_gallowstemplar_2383_12.dds"},
	{0x548DB04DE4F32BA4ULL, "6083038_m.fxe"},
	{0x548DB9AEADDB324DULL, "zz_one_debug6075022.cl"},
	{0x548DC60E03529680ULL, "gen00fl_varric6118983.cl"},
	{0x548EC310AC4642F8ULL, "fx_snowvolume.dds"},
	{0x548F1DFF2F308FFBULL, "mer220_hubert6059585.cl"},
	{0x548F2ADDCD2E2E84ULL, "gen00fl_isabela6080840.cl"},
	{0x548F58774A7C7871ULL, "6033699_f.fxe"},
	{0x548F8EA17F821FEBULL, "1054784336.ogg"},
	{0x548F900E05B95862ULL, "hlac_dr_0_deeproads_17639_1y.dds"},
	{0x548F9D15F0D70DFAULL, "hlac_ht_mansion2b_818_34y.dds"},
	{0x5490CD2455199698ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_meda.phy"},
	{0x5490EDBC62A15483ULL, "6151153_f.fxe"},
	{0x54911FD853CE2839ULL, "hlac_ht_0_hightown_n_-131072_0z.dds"},
	{0x549121EC06C8D8B8ULL, "llac_lt_0_lowtown_a2_n_131074_0.dds"},
	{0x5491D3669FD684D2ULL, "one130ar_booker_dungeon.are"},
	{0x549232431A59D2CAULL, "hlac_lt_0_lowtown_a2_n_-1_4x.dds"},
	{0x549256AC23DBB18FULL, "133869548.ogg"},
	{0x5492B1C27BCC4879ULL, "6144378_m.fxe"},
	{0x5492CDBD61F3A91AULL, "557974331.ogg"},
	{0x5492ECCD889F9CFCULL, "hlac_ca_cavern1_haunted_24421_10z.dds"},
	{0x54933456AF01385AULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_betb.phy"},
	{0x54933E527DBE3F65ULL, "vxca_cavern1_haunted_cjx_0_0.vlm"},
	{0x5493427194FBAE0AULL, "zz_followers_banter6158274.cl"},
	{0x549366A8818B9C8BULL, "6144457_m.fxe"},
	{0x5493977F63D2D07AULL, "plt_cod_ltr_raiders_martin.nss"},
	{0x5493B24598413C69ULL, "1032033926.ogg"},
	{0x5493DAE0E5F54301ULL, "hlac_lt_ruinedhouse_n_-1_9z.dds"},
	{0x5493E8AE2DF3081CULL, "hlac_w_sundermount_131071_0z.dds"},
	{0x54943541D906B22EULL, "vxlt_warehouse_18t_0_0.vlm"},
	{0x54946035DD562D45ULL, "vxlt_hangedman_25c_0_0.vlm"},
	{0x54947AB65871869AULL, "vxlt_0_lowtown_a2_hi4_0_0.vlm"},
	{0x549481B30BCAEE16ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_small.evt"},
	{0x5494DF25DF6CC1FAULL, "qcr220_orsino6060739.cl"},
	{0x549522413AED58FEULL, "vxdr_0_deeproads_cvy_0_0.vlm"},
	{0x5495547EB613C867ULL, "533906265.ogg"},
	{0x549599C1C4C38C23ULL, "6175831_m.fxe"},
	{0x5495BD99611BAF38ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_5z.dds"},
	{0x5495BFB8AF858665ULL, "mag100_thrask6034819.cl"},
	{0x54960D793257EE47ULL, "vxca_caverns1_mines_j27_0_0.vlm"},
	{0x549629574DBCC9E9ULL, "hlac_lt_0_lowtown_docks_n_131073_1z.dds"},
	{0x54963A8CAA1EC185ULL, "zz_prmpt_debug6198241.cl"},
	{0x54966D8CEFFA53C6ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_min_c.mmh"},
	{0x54968FA5546A0D39ULL, "6088889_m.fxe"},
	{0x549731AD405A1888ULL, "vxlt_warehouse_xo_0_0.vlm"},
	{0x549766C9420EF08DULL, "art\\characters\\playercharacter\\humanmale\\glv_dksa.msh"},
	{0x54977892463CBA47ULL, "hlac_ht_chantry_n_2534_27z.dds"},
	{0x549778C9420F0F13ULL, "art\\characters\\playercharacter\\humanmale\\glv_dksa.mmh"},
	{0x5498397ED0A43B3DULL, "285562580.ogg"},
	{0x549894703E2BEA9BULL, "follower_banter6155553.cl"},
	{0x54989E072FF92A45ULL, "233867738.ogg"},
	{0x549939B80B0385D4ULL, "6054579_m.fxe"},
	{0x549959A0B8C3451AULL, "lt_0_lowtown_docks_a3_n_211_0.mmh"},
	{0x54995DA0B8C34C51ULL, "lt_0_lowtown_docks_a3_n_211_0.mao"},
	{0x54996BA0B8C36460ULL, "lt_0_lowtown_docks_a3_n_211_0.msh"},
	{0x549A061C0FF314D9ULL, "lowtown_male_lowlife1_sta6164979.cl"},
	{0x549A33B8ED22CEC6ULL, "vrc321_varric6084984.cl"},
	{0x549A5BF8DA564F36ULL, "llac_ca_cavern1_haunted_24425_12.dds"},
	{0x549A795AFEB00E95ULL, "one100dg_vanard6048259.cl"},
	{0x549A88ADC1CDB303ULL, "251311079.ogg"},
	{0x549B3556159DD4C1ULL, "hlac_lt_gallowsprison_n_-1_4z.dds"},
	{0x549B78B30BD0FDA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_small.gad"},
	{0x549BCA669FDEC985ULL, "one130ar_booker_dungeon.lst"},
	{0x549C3C21EDDE32D5ULL, "dae000_leonato6162485.cl"},
	{0x549CACBD77E5D1B6ULL, "88604172.ogg"},
	{0x549CFC3EB5846AB7ULL, "6025349_m.fxe"},
	{0x549E25E453C5271FULL, "vxht_0_hightown_a3_c7m_0_0.vlm"},
	{0x549E2B149FE3546DULL, "152542714.ogg"},
	{0x549E5CB559F4E1B5ULL, "6049459_m.fxe"},
	{0x549FDDA01E9BFC49ULL, "6080157_m.fxe"},
	{0x54A023910041F5F4ULL, "6041839_f.fxe"},
	{0x54A030D0AF8F3BA6ULL, "sky_ca_caverns1.mao"},
	{0x54A075798CB848C7ULL, "l_and121ar_chantry.bnk"},
	{0x54A09B9A5F218E4BULL, "vxdr_primevalentrance_23r_0_2.vlm"},
	{0x54A12128C07CE13FULL, "llac_w_encounter2_camp_65541_0.dds"},
	{0x54A1284E75D5CFEBULL, "plt_gen00pt_combos.plo"},
	{0x54A130A4C3335127ULL, "6135927_m.fxe"},
	{0x54A14BD452FF5F6BULL, "vxw_encounter2_camp_b1_0_0.vlm"},
	{0x54A1673C427BF02DULL, "32941138.ogg"},
	{0x54A19AA4A49E89F8ULL, "dae011_drunk2_sta6178323.cl"},
	{0x54A1FFB29C41D096ULL, "703479154.ogg"},
	{0x54A279B30BD71E4FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_weightshift_small.ani"},
	{0x54A2C3E184C92166ULL, "llac_ca_caverns2prc_5802_19.dds"},
	{0x54A2C4D049B3677DULL, "ht_0_hightown_a3_37_0_ll.mmh"},
	{0x54A2D95220E63124ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.impa.ani"},
	{0x54A2E2669FE51114ULL, "one130ar_booker_dungeon.ncs"},
	{0x54A2F4C2673DABFAULL, "6085034_m.fxe"},
	{0x54A3872A68C21A77ULL, "w_bonepitmining_a2_1_0.mmh"},
	{0x54A38B2A68C22084ULL, "w_bonepitmining_a2_1_0.mao"},
	{0x54A39D2A68C23FE9ULL, "w_bonepitmining_a2_1_0.msh"},
	{0x54A467CCBC568151ULL, "d15_b_sp.dds"},
	{0x54A46A662E5FF089ULL, "vxlt_gallowscourtyard_75y_0_0.vlm"},
	{0x54A49BEB66B125C3ULL, "vxlt_0_lowtown_a2_n_eqw_0_0.vlm"},
	{0x54A4B9CF0A8871B8ULL, "gen_im_npc_arm_hum_boo_krkb.uti"},
	{0x54A5448CCB958E8FULL, "186602299.ogg"},
	{0x54A57F8D71F1E4EBULL, "657536739.ogg"},
	{0x54A5D62879BC0655ULL, "plt_dae000pt_amb_sunder.nss"},
	{0x54A5FF955DC6C7BEULL, "228384515.ogg"},
	{0x54A657448D20C7A0ULL, "456017975.ogg"},
	{0x54A68DEC59B9AF04ULL, "vxw_sundermount_1yo_0_0.vlm"},
	{0x54A7ADC43ABA0B3AULL, "657530673.ogg"},
	{0x54A7EF7E44AE0986ULL, "dae000_player_generic_find6179047.cl"},
	{0x54A8017253F8B784ULL, "vxw_woundedcoast_az_0_0.vlm"},
	{0x54A846B302AC6434ULL, "152067708.ogg"},
	{0x54A87B93BEA072BEULL, "374367792.ogg"},
	{0x54A8D65220EA9786ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.impa.gad"},
	{0x54A8D9DC53CC1834ULL, "6133782_m.fxe"},
	{0x54A946EA552A0A65ULL, "vxht_0_hightown_a2_cbm_0_0.vlm"},
	{0x54A95611ECBC0ED3ULL, "plt_dae000pt_home_ambients.plo"},
	{0x54AA1542075173FCULL, "vxht_0_hightown_c9u_0_0.vlm"},
	{0x54AA4B55AED6C0F2ULL, "6013389_m.fxe"},
	{0x54AA86515DE88AACULL, "337023536.ogg"},
	{0x54AA9330DABFE6BDULL, "vxlt_0_lowtown_docks_a3_n_cyl_0_1.vlm"},
	{0x54AAA12A9BFCABEFULL, "509415715.ogg"},
	{0x54AAA8C40793C98BULL, "dae210cr_elf_m_01.utc"},
	{0x54AAB8C1592E08EDULL, "and100_anders6096035.cl"},
	{0x54AB2C2A868535FAULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tn180l.evt"},
	{0x54ABBD9662F5DC54ULL, "seb321ip_at_hightown.utp"},
	{0x54AC0310D2C2FB69ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_coma.phy"},
	{0x54AC526D1D502E83ULL, "hlac_lt_gallowstemplar_n_2382_0y.dds"},
	{0x54AC971E22486F9FULL, "zz_qun_debug6028251.cl"},
	{0x54ACC13C7AA4B993ULL, "896413590.ogg"},
	{0x54AD2E2F6F97DA16ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default.phy"},
	{0x54AE145988F8725FULL, "mrl150_merrill6084468.cl"},
	{0x54AE568F0620FB43ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_bark01_n.dds"},
	{0x54AE8984904348F2ULL, "vxlt_backalley_n_1ko_0_0.vlm"},
	{0x54AF750B9D656F1DULL, "110831289.ogg"},
	{0x54AFCE19CD9CF3BBULL, "mcr370_templar16087627.cl"},
	{0x54AFD15220F0AE65ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.impa.evt"},
	{0x54AFF9D814C91F0AULL, "vxht_0_hightown_n_b1a_0_0.vlm"},
	{0x54B061E052D65E3BULL, "zz_test_james.stg"},
	{0x54B06E2C3CF9E604ULL, "6114838_f.fxe"},
	{0x54B1059F893AF89AULL, "zz_followers_banter6158199.cl"},
	{0x54B1151BDFC88EDFULL, "dae000_merchant_access_hf6162962.cl"},
	{0x54B11C9ED3601CB3ULL, "hlac_lt_0_lowtown_n_1_9y.dds"},
	{0x54B178225FB56ACAULL, "hlac_lt_0_lowtown_a2_n_131072_4z.dds"},
	{0x54B218B2938ECCDAULL, "vxlt_0_lowtown_a2_n_c_e83_0_0.vlm"},
	{0x54B219BFAA417F7CULL, "6038678_m.fxe"},
	{0x54B2A6E966A23596ULL, "art\\characters\\creatures\\spider\\animation\\tn90r.ani"},
	{0x54B30453A9CFA774ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholestack.phy"},
	{0x54B3735F38563E39ULL, "6086689_m.fxe"},
	{0x54B39A1EBD8466F0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_boosterprops.mao"},
	{0x54B39E1EBD846D03ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_boosterprops.mmh"},
	{0x54B3AC1EBD84855DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_boosterprops.msh"},
	{0x54B46560203F64F9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_lgtd_s.dds"},
	{0x54B48BF02210D71FULL, "tbe300im_promise_belt.uti"},
	{0x54B4B63C2A12C859ULL, "art\\vfx\\materials\\gui\\vfx_gui_target_ret_2.mao"},
	{0x54B501BB937303BCULL, "qcr000_aveline6054369.cl"},
	{0x54B56D97E3BB6C1EULL, "907161569.ogg"},
	{0x54B5805130F1DFA1ULL, "dae011cr_gossip_2.utc"},
	{0x54B6331391BD27F6ULL, "581303349.ogg"},
	{0x54B64B6975A7BEF7ULL, "mcr371_hostile_merrill.stg"},
	{0x54B66E16B22747D0ULL, "395048801.ogg"},
	{0x54B6D06A7EED0736ULL, "173960879.ogg"},
	{0x54B6FF51B68FE0E9ULL, "6065282_m.fxe"},
	{0x54B7172F6FA007B7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default.msh"},
	{0x54B7292F6FA0261AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default.mao"},
	{0x54B72D2F6FA02D29ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default.mmh"},
	{0x54B72D76C516CAFEULL, "857261218.ogg"},
	{0x54B731C1AE690E7EULL, "llac_lt_0_lowtown_docks_-65536_0.dds"},
	{0x54B7A4CDAC9447C5ULL, "815933323.ogg"},
	{0x54B7C403DAE712D5ULL, "ser270_templar6175489.cl"},
	{0x54B7CD184035D2C3ULL, "gen_im_trs_gen_sta_ele_r1.uti"},
	{0x54B7F137C1560557ULL, "mag110_samson6149961.cl"},
	{0x54B8443BAF0E855EULL, "vxlt_gallowstemplar_f_53g_0_0.vlm"},
	{0x54B8C8FAE86206E4ULL, "6166551_m.fxe"},
	{0x54BA1F1FD4D1E335ULL, "mag210_arianni6024563.cl"},
	{0x54BB5C0B089A6EC3ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_011a_0d.dds"},
	{0x54BBFD23BCDA4FA9ULL, "vxlt_0_lowtown_docks_azr_0_0.vlm"},
	{0x54BC754835174468ULL, "hlac_w_sundermount_-65533_0x.dds"},
	{0x54BCC2BAC7D9A20FULL, "6116362_m.fxe"},
	{0x54BCD6E4EE30D4FEULL, "vxht_chantry_2dy_0_0.vlm"},
	{0x54BCE153A9D7C071ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholestack.msh"},
	{0x54BCEF53A9D7D78CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholestack.mao"},
	{0x54BCFB53A9D7EC2FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholestack.mmh"},
	{0x54BD46E94F0901C8ULL, "6120153_m.fxe"},
	{0x54BD53A2D5A8C249ULL, "lt_0_lowtown_docks_8_0.anb"},
	{0x54BDBBE797C50792ULL, "vxlt_warehouse_n_b2_0_0.vlm"},
	{0x54BECA3021D88740ULL, "vxlt_0_lowtown_docks_a3_cx8_0_0.vlm"},
	{0x54BEF1ED1519069BULL, "gen00fl_merrill6119619.cl"},
	{0x54BF36EAD0794136ULL, "6034078_m.fxe"},
	{0x54BF469169B7F7F6ULL, "hlac_lt_0_lowtown_docks_-262143_0y.dds"},
	{0x54BFE5E9AC3D73C8ULL, "vxlt_hangedman_249_0_0.vlm"},
	{0x54C022EBCA4FAC07ULL, "520264453.ogg"},
	{0x54C067647BE4C1AEULL, "lt_0_lowtown_docks_a3_n_25_0_ll.mmh"},
	{0x54C0910182A22418ULL, "6169024_m.fxe"},
	{0x54C09DD90FCCB3F2ULL, "follower_banter6171079.cl"},
	{0x54C0AEE966AE8087ULL, "art\\characters\\creatures\\spider\\animation\\tn90r.evt"},
	{0x54C0CFBAC9E472D9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ready.ani"},
	{0x54C0F09134431DB6ULL, "character_character_p83.psh.fxo.dx11.assembly"},
	{0x54C11D418FB57FB9ULL, "lgt110tr_arl_connor.utt"},
	{0x54C1B1026DAD672DULL, "830572990.ogg"},
	{0x54C261CADB1FEFE5ULL, "dae000_gamlen6121507.cl"},
	{0x54C2A4FAE73C127BULL, "gen00fl_aveline6131098.cl"},
	{0x54C34EEC190B93A4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magb_0n.dds"},
	{0x54C3920404606B52ULL, "vxca_caverns2_7ap_0_0.vlm"},
	{0x54C3A7E8FA79D203ULL, "gen00fl_fenris6126796.cl"},
	{0x54C4839BA95123E4ULL, "6156152_m.fxe"},
	{0x54C493BDAD2D73DEULL, "vxlt_gallowstemplar_n_5rl_0_0.vlm"},
	{0x54C49F1EBD92E228ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\9pp_boosterprops.phy"},
	{0x54C4A0F8373381E1ULL, "vxht_mansion_n_5iy_0_0.vlm"},
	{0x54C4B7028FA76BC4ULL, "321881465.ogg"},
	{0x54C4CF10DCDEE061ULL, "vxlt_hangedman_213_0_0.vlm"},
	{0x54C502F9F208E2C6ULL, "6142567_f.fxe"},
	{0x54C511617397D307ULL, "448449765.ogg"},
	{0x54C54969374810DDULL, "lgt260_eau_tmp_roderick6143705.cl"},
	{0x54C58EDBED1EAF2FULL, "779273736.ogg"},
	{0x54C59C14A26A9C8EULL, "vxca_cave_19e_0_0.vlm"},
	{0x54C6083FFF9652B3ULL, "6059669_m.fxe"},
	{0x54C661D044F4461BULL, "146251664.ogg"},
	{0x54C67E53BA4C08F9ULL, "436904193.ogg"},
	{0x54C717934ED19B6BULL, "sib111_mother6098957.cl"},
	{0x54C7330D0FA7579BULL, "6053182_m.fxe"},
	{0x54C74AC67D024B90ULL, "6084369_m.fxe"},
	{0x54C777767508A198ULL, "6005326_m.fxe"},
	{0x54C7AABAC9EA521BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ready.gad"},
	{0x54C7CFE966B4D7D0ULL, "art\\characters\\creatures\\spider\\animation\\tn90r.gad"},
	{0x54C7D52879D8EAB6ULL, "plt_dae000pt_amb_sunder.plo"},
	{0x54C859E978CE6EB6ULL, "synccreatureanimations.gda"},
	{0x54C8D3200ED62514ULL, "142965676.ogg"},
	{0x54C8FFB1DE01A0ABULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyo_0d.dds"},
	{0x54C965EEE4EAF863ULL, "dae000_flemeth6000771.cl"},
	{0x54C988C80E8FFAA6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\blendtree\\doubledoor.anb"},
	{0x54C9BF42EE1FD71EULL, "plt_lgt000pt_ultra_1_01.nss"},
	{0x54C9FFE79591DC0AULL, "vxdr_primevalentrance_38a_0_0.vlm"},
	{0x54CA1B78C940EF67ULL, "6116847_m.fxe"},
	{0x54CA74D496755EC0ULL, "vxw_woundedcoast_1n_0_0.vlm"},
	{0x54CC2DC489BB2012ULL, "vxlt_warehouse_n_i3_0_0.vlm"},
	{0x54CC6B798CDD72FDULL, "l_and121ar_chantry.wbd"},
	{0x54CC6B9BE698C405ULL, "hlac_lt_0_lowtown_docks_a3_-65538_0z.dds"},
	{0x54CCFA727CB5B353ULL, "vxlt_gamlenshouse_n_q0_0_0.vlm"},
	{0x54CD067256731BFFULL, "ht_0_hightown_n_20_0_ll.mmh"},
	{0x54CD68F6CCF93AA7ULL, "703098873.ogg"},
	{0x54CD953580CEA7F7ULL, "lgt000_chanter6147183.cl"},
	{0x54CD9E47316E7758ULL, "vxht_0_hightown_boz_0_0.vlm"},
	{0x54CDA8C7C26F59B6ULL, "6031378_m.fxe"},
	{0x54CE57C69E338C48ULL, "vxlt_hangedman_216_0_0.vlm"},
	{0x54CE94E1BFA31F70ULL, "82729318.ogg"},
	{0x54CEAE9CB87EA315ULL, "hlac_lt_0_lowtown_a2_n_c_-1_1z.dds"},
	{0x54CEC7BAC9F0A200ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ready.evt"},
	{0x54CEE6C7752868E1ULL, "vxca_caverns1_mines_fmp_0_0.vlm"},
	{0x54CF0A883E5B99BAULL, "zz_ave_debug6056772.cl"},
	{0x54D00E358959C453ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_sylvesta.ani"},
	{0x54D01950E20D08FCULL, "6206146_f.fxe"},
	{0x54D0220860120EF6ULL, "vxca_cavern1_haunted_dq1_0_0.vlm"},
	{0x54D03EA6FFAF0A83ULL, "vxca_caverns2prc_9wx_0_0.vlm"},
	{0x54D04ECD6738FCA3ULL, "ts_profane_prog_junk1.gda"},
	{0x54D069D456B90BCFULL, "isa111_isabela6067112.cl"},
	{0x54D08E07EB81EB40ULL, "6087027_m.fxe"},
	{0x54D092001BA56B6AULL, "vxw_bonepitdestroyed_a3_1wl_0_0.vlm"},
	{0x54D09F2886653B59ULL, "711355627.ogg"},
	{0x54D0CEE9185218D5ULL, "default_rogue_male.utc"},
	{0x54D0CFF03378DB62ULL, "w_encounter2_camp_58_0.rml"},
	{0x54D135C68343D4EEULL, "8195762.ogg"},
	{0x54D13FE6A22A801AULL, "vxht_0_hightown_a2_b92_0_0.vlm"},
	{0x54D1677A14481FE6ULL, "hlac_lt_0_lowtown_a3_n_c_131071_3y.dds"},
	{0x54D18C0F830FA63EULL, "6048200_m.fxe"},
	{0x54D225415BEB8D2BULL, "qun120_arishok6049652.cl"},
	{0x54D3059A9F238C5EULL, "vxlt_gallowstemplar_n_5um_0_0.vlm"},
	{0x54D3570929DB0104ULL, "vxw_woundedcoast_py_0_0.vlm"},
	{0x54D379EE691F5FC8ULL, "wm_ico_hut_roll.dds"},
	{0x54D3FCE96E010CF3ULL, "171451365.ogg"},
	{0x54D4466A3F2DECE9ULL, "6164703_m.fxe"},
	{0x54D49E5AA39475B0ULL, "733087260.ogg"},
	{0x54D54971C8B58DD8ULL, "6079365_m.fxe"},
	{0x54D56FDB434B714AULL, "6082347_m.fxe"},
	{0x54D5A38E216C6657ULL, "6079146_m.fxe"},
	{0x54D607CDE9E55D1CULL, "vxw_bonepitmining_a2_ai_0_0.vlm"},
	{0x54D6165E20A56888ULL, "vxca_caverns1_i08_0_0.vlm"},
	{0x54D6444C3E29CA35ULL, "29174039.ogg"},
	{0x54D64E15F81E1D4FULL, "vxw_encounter2_25_0_0.vlm"},
	{0x54D67B283E0A51D2ULL, "vxlt_gamlenshouse_n_q5_0_0.vlm"},
	{0x54D7273200B97620ULL, "6101835_m.fxe"},
	{0x54D786561521F031ULL, "gen_im_wep_rog_arc_act1_01.uti"},
	{0x54D7D0F0337EFB16ULL, "w_encounter2_camp_58_0.phy"},
	{0x54D7EC10D2E80E8EULL, "art\\characters\\playercharacter\\humanfemale\\hlh_coma.mmh"},
	{0x54D7F810D2E82325ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_coma.mao"},
	{0x54D80610D2E83B7CULL, "art\\characters\\playercharacter\\humanfemale\\hlh_coma.msh"},
	{0x54D83F2247D39001ULL, "ser270_templar6149990.cl"},
	{0x54D86F7D367D14C8ULL, "vxca_caverns1_mines_fmu_0_0.vlm"},
	{0x54D8C70E8B1C0AC1ULL, "vxca_caverns1_mines_ghf_0_0.vlm"},
	{0x54D8E550BD5C9170ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_alchem_n.dds"},
	{0x54D8F51C80464A6AULL, "vxht_keep_n_4hd_0_0.vlm"},
	{0x54D93D11ECE50870ULL, "plt_dae000pt_home_ambients.nss"},
	{0x54D9840636574C8BULL, "50088095.ogg"},
	{0x54D9CD9A4D68DCE9ULL, "784323585.ogg"},
	{0x54DA1729FF1B8A4CULL, "gen00fl_varric6201269.cl"},
	{0x54DA422F9980ABFEULL, "6111209_m.fxe"},
	{0x54DA87AF1E742A76ULL, "w_bonepitdestroyed_a3_156_0.msh"},
	{0x54DA91AF1E743B74ULL, "w_bonepitdestroyed_a3_156_0.mmh"},
	{0x54DA95AF1E744187ULL, "w_bonepitdestroyed_a3_156_0.mao"},
	{0x54DB9A4CE7DE4DBAULL, "vxlt_0_lowtown_n_eo1_0_1.vlm"},
	{0x54DC2DDD0556BEB9ULL, "zz_and_debug6178548.cl"},
	{0x54DC52C200A7ACF6ULL, "264678418.ogg"},
	{0x54DC755B0D9C529AULL, "vxca_caverns2_83a_0_0.vlm"},
	{0x54DCA806D2C13407ULL, "276162342.ogg"},
	{0x54DD31E6E0ECA91DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2_tool.mao"},
	{0x54DD35E6E0ECB056ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2_tool.mmh"},
	{0x54DD3FE6E0ECC154ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2_tool.msh"},
	{0x54DDF3093281624AULL, "vxlt_undercityserialkiller_x0_0_0.vlm"},
	{0x54DE1BBE35CCB381ULL, "vxca_caverns2_9fx_0_0.vlm"},
	{0x54DE215010B5F96EULL, "306872914.ogg"},
	{0x54DE39450486725FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wmda_0n.dds"},
	{0x54DF1E418FCE7F7CULL, "lgt110tr_arl_connor.ncs"},
	{0x54DF702229C07701ULL, "914846939.ogg"},
	{0x54DF946480C257D9ULL, "vxw_bonepitmining_a2_b9_0_0.vlm"},
	{0x54E09722E2BB25C6ULL, "qcr220_meredith6060306.cl"},
	{0x54E0C336B66AB148ULL, "318174101.ogg"},
	{0x54E197199DEFA6B1ULL, "gen00fl_aveline6116675.cl"},
	{0x54E19A3E6B5E358BULL, "6066721_f.fxe"},
	{0x54E19C678D7B95EFULL, "6076446_m.fxe"},
	{0x54E1B9F0338728B7ULL, "w_encounter2_camp_58_0.msh"},
	{0x54E1CBF03387471AULL, "w_encounter2_camp_58_0.mao"},
	{0x54E1CFF033874E29ULL, "w_encounter2_camp_58_0.mmh"},
	{0x54E1E2B3AE0851CFULL, "6093674_m.fxe"},
	{0x54E1F341F130F628ULL, "777462755.ogg"},
	{0x54E21DB9FD3DB76BULL, "vxlt_0_lowtown_a3_n_c_gyf_0_0.vlm"},
	{0x54E24ECD4E636A5CULL, "mcr310_argument6127590.cl"},
	{0x54E29639C668BA08ULL, "vxlt_blackemporium_14n_0_0.vlm"},
	{0x54E2A39DFF20431BULL, "896492485.ogg"},
	{0x54E2D9572D62A759ULL, "thrice_enchanted_gloves.dds"},
	{0x54E2F0D60637954DULL, "6116683_m.fxe"},
	{0x54E303A1D10C0269ULL, "6027725-ifailedyouwesley.cut"},
	{0x54E3D7BAEB3EE94CULL, "hlac_w_sundermount_-327677_0x.dds"},
	{0x54E3E1B8354189CAULL, "6085247_m.fxe"},
	{0x54E3E535896A27FDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_sylvesta.gad"},
	{0x54E3E6CC78A8829DULL, "hlac_lt_backalley_g_65537_1x.dds"},
	{0x54E3E75C79C64C84ULL, "hlac_lt_darktown_-1_40x.dds"},
	{0x54E4166A2AB4F9D6ULL, "hlac_ht_0_hightown_a3_0_0z.dds"},
	{0x54E4BEA602E1D758ULL, "vxlt_gallowstemplar_5at_0_0.vlm"},
	{0x54E54EDFDDE7E0A2ULL, "310224244.ogg"},
	{0x54E5EAE34210B50FULL, "gen00fl_fenris6116562.cl"},
	{0x54E6596C54F9FA3EULL, "and100_anders6095967.cl"},
	{0x54E6EBD164EAB56AULL, "386882926.ogg"},
	{0x54E733F6DEC2DFB1ULL, "734212903.ogg"},
	{0x54E771A7EFA2A24BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyn_0d.dds"},
	{0x54E7D126A9BE1D05ULL, "6118722_m.fxe"},
	{0x54E7ED52E2AC80EDULL, "vxlt_gallowstemplar_n_35h_0_0.vlm"},
	{0x54E86CBB9AD8C59EULL, "hlac_lt_0_lowtown_a2_n_65535_6x.dds"},
	{0x54E8876FB38AC481ULL, "433873755.ogg"},
	{0x54E8C73E59A904D9ULL, "130595660.ogg"},
	{0x54E932EEEBA980F8ULL, "hlac_lt_0_lowtown_n_-131072_8y.dds"},
	{0x54E9A338EB751E28ULL, "356681894.ogg"},
	{0x54E9E39DB98F8B75ULL, "plt_vis100pt_rescue.nss"},
	{0x54E9F266555E49A9ULL, "456324821.ogg"},
	{0x54EA39BB46459699ULL, "6098952_m.fxe"},
	{0x54EAB54056C8B4D4ULL, "6118774_m.fxe"},
	{0x54EACE6E37A9605AULL, "ht_chantry_n_probe_2a4.dds"},
	{0x54EAF07F54EBC191ULL, "vxlt_undercitydungeon_1ij_0_0.vlm"},
	{0x54EB5D2814553FC7ULL, "vxlt_0_lowtown_docks_n_arw_0_0.vlm"},
	{0x54EB5EDEC3BE56E7ULL, "234946931.ogg"},
	{0x54EC3B9CF7DF6F77ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medh.phy"},
	{0x54EC656975D57A58ULL, "mcr371_hostile_merrill.cnv"},
	{0x54EC6BF156759F48ULL, "6131960_m.fxe"},
	{0x54EC77E59D44E397ULL, "6084549_f.fxe"},
	{0x54ECBA245567D52DULL, "art\\characters\\playercharacter\\dwarfmale\\arm_meda.msh"},
	{0x54ECC8245567ED60ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_meda.mao"},
	{0x54ECCC245567F3B3ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_meda.mmh"},
	{0x54EDAF91949C5348ULL, "gen00fl_carver6097402.cl"},
	{0x54EE064172B87059ULL, "vxw_bonepitmining_a2_i8_0_0.vlm"},
	{0x54EE0CD50BC2CC30ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_min_c.mmh"},
	{0x54EE9BD8D1C946D4ULL, "dae000ip_crowd_lm3_talk3.utp"},
	{0x54EEE449C9A0ECD6ULL, "art\\levels\\races\\proxy\\props\\_textures\\bigboat_d.dds"},
	{0x54EF2D56AF4F8A83ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_betb.msh"},
	{0x54EF3B56AF4FA2DDULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_betb.mmh"},
	{0x54EF48ACD179088DULL, "vxw_sundermount_2ee_0_0.vlm"},
	{0x54EF9B5CBB2AB1A7ULL, "zz_dae_debug6029485.cl"},
	{0x54EFE954830FBD5EULL, "vxlt_gallowstemplar_n_5co_0_0.vlm"},
	{0x54EFF5AC57274843ULL, "lt_0_lowtown_docks_a3_141_0.msh"},
	{0x54F003AC57275F9DULL, "lt_0_lowtown_docks_a3_141_0.mmh"},
	{0x54F007AC572766AEULL, "lt_0_lowtown_docks_a3_141_0.mao"},
	{0x54F030076B3F4DA9ULL, "clf260_harley6069694.cl"},
	{0x54F0476975D93060ULL, "mcr371_hostile_merrill.bnk"},
	{0x54F053130ED04F6FULL, "259471111.ogg"},
	{0x54F0ACCCE9D6863CULL, "one350_terath6087116.cl"},
	{0x54F15005EAF42770ULL, "542363827.ogg"},
	{0x54F167A42AD4DB99ULL, "follower_banter6172776.cl"},
	{0x54F1B65F8A8ABC36ULL, "art\\levels\\ztools\\default\\generic\\3pp_basic.msh"},
	{0x54F1C05F8A8ACD34ULL, "art\\levels\\ztools\\default\\generic\\3pp_basic.mmh"},
	{0x54F1C45F8A8AD447ULL, "art\\levels\\ztools\\default\\generic\\3pp_basic.mao"},
	{0x54F273BDE215EDF0ULL, "735588105.ogg"},
	{0x54F2E37A5B85FE3BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_leaves01_dp.dds"},
	{0x54F2EB87CDD18E1CULL, "vxw_bonepitmining_a2_ko_0_0.vlm"},
	{0x54F35FD7C0DF6967ULL, "hla_ht_0_hightown_n_131070_0z.dds"},
	{0x54F39C771A370037ULL, "additiveflies.mao"},
	{0x54F471F1227A8620ULL, "ts_slavers_wpn_props.gda"},
	{0x54F4B1593A61A854ULL, "zz_vault_debug6177806.cl"},
	{0x54F4CCCCE35228C9ULL, "6128710_m.fxe"},
	{0x54F5087D44260A47ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgtp_0s.dds"},
	{0x54F6FE2752F5F4B7ULL, "w_bonepit_216_0.mmh"},
	{0x54F7022752F5FBC4ULL, "w_bonepit_216_0.mao"},
	{0x54F7142752F61A29ULL, "w_bonepit_216_0.msh"},
	{0x54F754FA6D97D422ULL, "art\\levels\\races\\vfx_models\\prp_camploglrg.phy"},
	{0x54F7A78B56F5BDB7ULL, "517674083.ogg"},
	{0x54F827C2FC2D9115ULL, "92594310.ogg"},
	{0x54F837FD7FC4706EULL, "hlac_ht_keep_n_4519_7z.dds"},
	{0x54F850095297FCD3ULL, "vxlt_gamlenshouse_n_b5_0_0.vlm"},
	{0x54F8A27279A9212EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn04.phy"},
	{0x54F9A1CD42D4034FULL, "897667275.ogg"},
	{0x54F9B47F6429B4EDULL, "plt_cod_ltr_raiders_martin.plo"},
	{0x54F9C214E4D760F3ULL, "vxw_woundedcoast_1bq_0_0.vlm"},
	{0x54F9C5E4C8E8C03EULL, "zz_mrl_debug.cnv"},
	{0x54F9D71F579C23C2ULL, "169983419.ogg"},
	{0x54FA1D3E9897E737ULL, "art\\characters\\playercharacter\\humanmale\\glv_masc.phy"},
	{0x54FA25A1B9CB4445ULL, "6103920_m.fxe"},
	{0x54FA5147B89901BFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bedroll.phy"},
	{0x54FB2FD50BCD924FULL, "art\\vfx\\combat\\vfx_glb_nat_imp_min_c.anb"},
	{0x54FB67DB35B41E3EULL, "vxlt_0_lowtown_a2_n_c_bz2_0_0.vlm"},
	{0x54FB9489D1849A39ULL, "nat380_temmerin6067392.cl"},
	{0x54FBE796F187B4EAULL, "vxw_bonepitdestroyed_a3_1fo_0_0.vlm"},
	{0x54FC01DDBC1BF4A3ULL, "dae000cr_grand_cleric6147697.cl"},
	{0x54FC02D702B3140EULL, "vxlt_undercitydungeon_15u_0_0.vlm"},
	{0x54FC25416A339E8FULL, "616432333.ogg"},
	{0x54FC78ED2719690FULL, "gen_im_cns_msc_backpack_02.uti"},
	{0x54FC7C5DADFC919FULL, "llac_w_woundedcoast_65544_0.dds"},
	{0x54FCD4DBC6D68C27ULL, "m_dae260ar_wounded_coast.bnk"},
	{0x54FCF5699E5B5C67ULL, "852286547.ogg"},
	{0x54FD310DA536A416ULL, "hlac_lt_hangedman_-1_9x.dds"},
	{0x54FD33EE6CEFA721ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.anb"},
	{0x54FD5B9ACAC4BFA7ULL, "vxht_0_hightown_ccl_0_0.vlm"},
	{0x54FD7AD30ED69F41ULL, "66057898.ogg"},
	{0x54FD8E48B702BF05ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_5x.dds"},
	{0x54FDBA89C36F210DULL, "vxw_sundermount_2nd_0_0.vlm"},
	{0x54FE712C0F547918ULL, "332649005.ogg"},
	{0x54FED572D31E8ADCULL, "6161297_m.fxe"},
	{0x54FEFA3556778750ULL, "hlac_lt_0_lowtown_a2_131070_0y.dds"},
	{0x54FF65C84EFDCBEEULL, "vxht_keep_n_69e_0_0.vlm"},
	{0x54FFDE7C0AC7A72BULL, "hlac_lt_0_lowtown_a2_n_c_-131071_2y.dds"},
	{0x54FFF36141BF720BULL, "zz_dae_debug6029841.cl"},
	{0x55001B482B365F4BULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_019a_0n.dds"},
	{0x5500FBE7E5827AAEULL, "hlac_w_encounter2_262143_0x.dds"},
	{0x55017DE7D5ED700EULL, "6083843_m.fxe"},
	{0x5501A25FE6CE0858ULL, "vxlt_gallowstemplar_5wr_0_0.vlm"},
	{0x5501D0BF13EC9B52ULL, "vxlt_gamlenshouse_n_b0_0_0.vlm"},
	{0x5501FDD0DFF92302ULL, "sky_ht_keep.mao"},
	{0x55020B494EEF06AAULL, "6206657_m.fxe"},
	{0x550228FF9DAE8F0BULL, "vxw_encounter2_g6_0_0.vlm"},
	{0x55024FF443C1DAC7ULL, "vxlt_0_lowtown_a2_n_eki_0_0.vlm"},
	{0x550298FC59E85682ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_2.gad"},
	{0x5502E1C96AD28B33ULL, "168240778.ogg"},
	{0x55031045286496F8ULL, "6108568_m.fxe"},
	{0x55036E8A95FED33FULL, "6108611_m.fxe"},
	{0x5503FD6D72B2ED16ULL, "6188626_f.fxe"},
	{0x550465D884225085ULL, "vxlt_hangedman_n_2dk_0_0.vlm"},
	{0x55046C80A8133C70ULL, "dae000_madam_lusine6151634.cl"},
	{0x5504CFB8F1562C4CULL, "6061805_m.fxe"},
	{0x5505046A5EDB12AAULL, "gen00fl_dog6121267.cl"},
	{0x55051188CFD488B8ULL, "1036265034.ogg"},
	{0x5505325C481485C8ULL, "vxlt_hangedman_1lz_0_0.vlm"},
	{0x550533CC1684AD0CULL, "60431853.ogg"},
	{0x55056A2242A2E05AULL, "mcr371_scared_mage_m6087691.cl"},
	{0x5505D582F4172B6BULL, "gen00fl_merrill6128844.cl"},
	{0x5506B3E4CBF94FACULL, "hlac_lt_backalley_n_0_16x.dds"},
	{0x550798FA3AB4E5D4ULL, "6112574_m.fxe"},
	{0x5507AAEF724656BCULL, "6162807_m.fxe"},
	{0x550833F1D6CD7973ULL, "vxw_woundedcoast_1mp_0_0.vlm"},
	{0x55085CF20D4616D4ULL, "793437324.ogg"},
	{0x550882F463480AB3ULL, "vxw_bonepit_203_0_0.vlm"},
	{0x55088A7BA45CBA9AULL, "llac_dr_primevaltomb_659_13.dds"},
	{0x5508C9CBA830DC9FULL, "1053827603.ogg"},
	{0x5508E0579FA0B6F5ULL, "6127654_m.fxe"},
	{0x5509492685B30BFFULL, "505175192.ogg"},
	{0x55097DCFF54FD74FULL, "mrl000_marethari6069731.cl"},
	{0x5509BBFC59EEB179ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_2.evt"},
	{0x550A5973E37DCD6AULL, "vxw_bonepitdestroyed_a3_1kh_0_0.vlm"},
	{0x550A8C75E36F7964ULL, "vxht_mansion2_2n5_0_0.vlm"},
	{0x550ACE5A528E4CCCULL, "6171029_m.fxe"},
	{0x550AF0F7449F8AAEULL, "6125570_m.fxe"},
	{0x550B14EE6CFBAF9AULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.mmh"},
	{0x550B484567D9FD08ULL, "darktown_fem_lowlife_mob6174166.cl"},
	{0x550B5247B8A776C4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bedroll.mmh"},
	{0x550B5647B8A77D17ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bedroll.mao"},
	{0x550B6847B8A79C66ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bedroll.msh"},
	{0x550BE29DB9AC6FD6ULL, "plt_vis100pt_rescue.plo"},
	{0x550C2C66B565398DULL, "vxw_sundermount_2sc_0_0.vlm"},
	{0x550D03841D126F9BULL, "zz_fol_debug6057018.cl"},
	{0x550DD4E74B4DC4F8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tlfrbroom_01_0_d.dds"},
	{0x550DD7A540F3E46EULL, "vxht_keep_n_64f_0_0.vlm"},
	{0x550DEBCE7953FB76ULL, "hlac_lt_gallowscourtyard_n_-65536_3z.dds"},
	{0x550E73EAD2B3B22DULL, "zz_mcr_debug6080882.cl"},
	{0x550F3C7CDB70637DULL, "vxht_mansion2_1t5_0_0.vlm"},
	{0x550FCCAAD21EA984ULL, "6042550_m.fxe"},
	{0x550FDBFE37B17B47ULL, "6177815_m.fxe"},
	{0x551082E950280CEEULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.cess.evt"},
	{0x5510F83BF87B2F2CULL, "156163453.ogg"},
	{0x5511380C488D9C75ULL, "6086883_m.fxe"},
	{0x5511660E9C6E2C17ULL, "vxlt_0_lowtown_n_erv_0_0.vlm"},
	{0x5511FC99F2B3BD8DULL, "6127508_m.fxe"},
	{0x5512D88EACBCDD16ULL, "lt_0_lowtown_a3_n_c_28_0.rml"},
	{0x5513A71EFBAAF2A0ULL, "vxdr_primevalentrance_2q8_0_1.vlm"},
	{0x5513C6DC02B26A5AULL, "693049168.ogg"},
	{0x55141F1D653DC7D3ULL, "6120473_m.fxe"},
	{0x5514529BCDD64242ULL, "6095692_f.fxe"},
	{0x55146008265181CEULL, "6108993_m.fxe"},
	{0x551472EC6A45F841ULL, "6101188_m.fxe"},
	{0x55148090E32BC210ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs106a_0d.dds"},
	{0x55148CF7F28A8807ULL, "122390006.ogg"},
	{0x5514A442EE5F9569ULL, "plt_lgt000pt_ultra_1_01.plo"},
	{0x5515255A456099B8ULL, "vxlt_gallowstemplar_f_65x_0_0.vlm"},
	{0x55153C810E9B5267ULL, "6198827_m.fxe"},
	{0x55154116E088260CULL, "llac_lt_backalley_g_65537_0.dds"},
	{0x551597E495F1A36BULL, "hlac_ca_caverns2prc_5802_17y.dds"},
	{0x5516028BAAE9149CULL, "6035369_m.fxe"},
	{0x551646162ACE0B11ULL, "vxlt_undercitydungeon_1zo_0_0.vlm"},
	{0x551667A9C40AA734ULL, "pro000_cut2d_voyage6004226.cl"},
	{0x551679E9502C6899ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.cess.gad"},
	{0x5516A5CEC8C391F3ULL, "vxw_woundedcoast_1hs_0_0.vlm"},
	{0x5516E3AAF0CF5B5BULL, "gen00fl_sebastian6174929.cl"},
	{0x5517C3FC59FAFC68ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_2.ani"},
	{0x5517E4C3A808645DULL, "203799869.ogg"},
	{0x5517E5B4AD961EDCULL, "hlac_lt_gallowsprison_n_-1_12x.dds"},
	{0x5518394FFC2343CEULL, "vxlt_backalley_n_1dn_0_0.vlm"},
	{0x5518582B9DCA023BULL, "vxht_0_hightown_a2u_0_0.vlm"},
	{0x5518688538E96542ULL, "710904575.ogg"},
	{0x5518DEB614A487DCULL, "823840863.ogg"},
	{0x5518F527DED8BE0BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masd_0s.dds"},
	{0x55193073EAA12F38ULL, "hlac_ht_keep_3731_1z.dds"},
	{0x551962D43695A32EULL, "6122470_m.fxe"},
	{0x55198CE673E06AB4ULL, "vxht_playershouse_n_1gw_0_0.vlm"},
	{0x5519977E96A0B741ULL, "268457391.ogg"},
	{0x5519F216485A49EFULL, "835362068.ogg"},
	{0x551A477D0484933DULL, "llac_w_sundermount_262145_0.dds"},
	{0x551A6D9A9328BDB1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright_rockwraith.ani"},
	{0x551B1F7283FCAC6DULL, "vxca_caverns1_mines_ck3_0_0.vlm"},
	{0x551C3804117A2990ULL, "vxht_mansion_n_6iw_0_0.vlm"},
	{0x551CC16AE548654FULL, "hlac_ht_mansion2_818_38z.dds"},
	{0x551D2FFA6DB864D6ULL, "art\\levels\\races\\vfx_models\\prp_camploglrg.mao"},
	{0x551D3BFA6DB87975ULL, "art\\levels\\races\\vfx_models\\prp_camploglrg.mmh"},
	{0x551D3DFA6DB87CEBULL, "art\\levels\\races\\vfx_models\\prp_camploglrg.msh"},
	{0x551D6052760BC14DULL, "mag100_cullen6171213.cl"},
	{0x551D8AE95032A4E7ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.cess.ani"},
	{0x551DA87DCC8E3AB0ULL, "w_encounter2_3_0.mao"},
	{0x551DAC7DCC8E41C3ULL, "w_encounter2_3_0.mmh"},
	{0x551DBA7DCC8E591DULL, "w_encounter2_3_0.msh"},
	{0x551E6C6A8EFD30CFULL, "6119936_m.fxe"},
	{0x551E77768D958136ULL, "vxlt_0_lowtown_docks_8ue_0_0.vlm"},
	{0x551E829A261CF18BULL, "vxht_0_hightown_a2_asr_0_0.vlm"},
	{0x551E8568396B79B6ULL, "vxlt_0_lowtown_docks_a3_ctj_0_0.vlm"},
	{0x551F0DE70E40BFB4ULL, "zz_approval_debug6133647.cl"},
	{0x551F1CE20F6892E3ULL, "248686474.ogg"},
	{0x551F33AE27AE0BC7ULL, "vxlt_0_lowtown_a2_n_eqk_0_0.vlm"},
	{0x551F37D44BF4C5A4ULL, "vxlt_gallowscourtyard_4xh_0_0.vlm"},
	{0x551F96AB2B9026EFULL, "6165096_m.fxe"},
	{0x551FD7EB8E644497ULL, "vxlt_0_lowtown_n_eoq_0_0.vlm"},
	{0x551FF586238D2B51ULL, "6154906_m.fxe"},
	{0x551FFFD609842343ULL, "art\\characters\\creatures\\mubari\\animation\\cwkf.ani"},
	{0x5520524479EB043FULL, "6106411_m.fxe"},
	{0x55209B1DA5A3BD3BULL, "plt_mcr000pt_general.nss"},
	{0x5521330D5E19C9B5ULL, "6095840_f.fxe"},
	{0x55214C85DACCDEE8ULL, "zz_dae_debug6050732.cl"},
	{0x5521789A932EEF53ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright_rockwraith.gad"},
	{0x5521835056A9035CULL, "hlac_w_bonepitdestroyed_a3_196612_0y.dds"},
	{0x5521F47B30257DF0ULL, "hlac_lt_darktown_-1_72y.dds"},
	{0x5523587874654641ULL, "combat_in_nmv_default.gda"},
	{0x5523EC6A2FFEFA3FULL, "hlac_lt_0_lowtown_n_65535_8y.dds"},
	{0x55240D6A83F694DFULL, "781928615.ogg"},
	{0x552438B4A4BA9076ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.dura.evt"},
	{0x552439DA0B939CB7ULL, "75579194.ogg"},
	{0x5524E2AF0FDFBDB5ULL, "935046583.ogg"},
	{0x5524E7179D4FC6EDULL, "6078802_m.fxe"},
	{0x552517ABBAB9AA73ULL, "vxw_woundedcoast_1sr_0_0.vlm"},
	{0x55251B1B44A2CAEAULL, "vxdr_primevaltomb_8em_0_0.vlm"},
	{0x5525C411838CE7F5ULL, "6129854_m.fxe"},
	{0x5525F91596B7FFD8ULL, "319793413.ogg"},
	{0x55260E4C51234B21ULL, "llac_lt_0_lowtown_a2_n_c_131071_2.dds"},
	{0x55264C730CE62639ULL, "6178990.cut"},
	{0x55269B9807E9634FULL, "6198705_f.fxe"},
	{0x5526B2E4C90F8D99ULL, "zz_mrl_debug.ncs"},
	{0x5526E01828306A52ULL, "6101827_f.fxe"},
	{0x5527759A933355B8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkright_rockwraith.evt"},
	{0x5527A163A597092FULL, "follower_banter6170149.cl"},
	{0x5527A72789D35BBBULL, "zz_qcr_debug6045151.cl"},
	{0x5527E923B24CA0D3ULL, "art\\vfx\\combat\\vfx_c_debili_aoe_imp_p.anb"},
	{0x5527F3FBC04590C3ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall.msh"},
	{0x552801FBC045A81DULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall.mmh"},
	{0x552805FBC045AF2EULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_wall.mao"},
	{0x552813269598ED0EULL, "6108261_m.fxe"},
	{0x552831C16A1A4DE5ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_013a_0n.dds"},
	{0x5529102099516A8DULL, "vxw_sundermount_2ya_0_0.vlm"},
	{0x5529589C4DE29682ULL, "llac_w_encounter2_196608_0.dds"},
	{0x5529686FF988B42FULL, "6034713_f.fxe"},
	{0x55296B474CA84A92ULL, "6066345_m.fxe"},
	{0x552996611C314283ULL, "lowtown_fem_alienage1_sta6161573.cl"},
	{0x5529A31F32D17803ULL, "vxca_caverns1_hu1_0_0.vlm"},
	{0x5529CADBC6FD695DULL, "m_dae260ar_wounded_coast.wbd"},
	{0x5529E6D00D586C66ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_loop.ani"},
	{0x5529F605A120113DULL, "fx_glow01.dds"},
	{0x552A2FB4A4BEEC01ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.dura.gad"},
	{0x552A448CDEA0A90AULL, "vxw_sundermount_31n_0_1.vlm"},
	{0x552A9775EB6E85EEULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar5_.ani"},
	{0x552AA9E103704210ULL, "vxht_mansion_n_6lt_0_0.vlm"},
	{0x552ACEC564107216ULL, "zz_fol_debug6098100.cl"},
	{0x552B157B7820BE07ULL, "vxw_bonepit_26i_0_0.vlm"},
	{0x552B23AC62EF1CF9ULL, "art\\vfx\\combat\\vfx_g_spider_leap_act_c.impa.ani"},
	{0x552B9D1F044C7828ULL, "sky_x_narrativeroom.mao"},
	{0x552BDD857C400C6DULL, "t1_tat_brn.tnt"},
	{0x552C2036BF5C947DULL, "vxht_mansion2_1b3_0_0.vlm"},
	{0x552D59001F2C2CD8ULL, "125180153.ogg"},
	{0x552D86449F4467C9ULL, "vxlt_0_lowtown_a2_n_c_hjc_0_0.vlm"},
	{0x552DC60606D4A199ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushd.msh"},
	{0x552DD00606D4B287ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushd.mmh"},
	{0x552E17D60990893AULL, "art\\characters\\creatures\\mubari\\animation\\cwkf.evt"},
	{0x552E3CE0BD01A1DBULL, "371535043.ogg"},
	{0x552E3D53B8A935BDULL, "vxdr_primevalentrance_1ky_0_0.vlm"},
	{0x552E49C8805A5D17ULL, "vxlt_0_lowtown_n_ehp_0_0.vlm"},
	{0x552EAD7DCC9CB6E8ULL, "w_encounter2_3_0.phy"},
	{0x552EB105B8DF0524ULL, "584947591.ogg"},
	{0x552ED5C65A87EF31ULL, "92020879.ogg"},
	{0x552F01709D4CBF30ULL, "vxw_encounter2_40_0_0.vlm"},
	{0x552F14A37E179451ULL, "249159292.ogg"},
	{0x552F4AD3C93A2AE7ULL, "vxw_encounter2_54_0_0.vlm"},
	{0x552FA3DA6F9A43F0ULL, "81776388.ogg"},
	{0x552FF9ED823B3B55ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sarcophagus_sa.dds"},
	{0x55307F821EA3E482ULL, "llac_lt_0_lowtown_a2_-131072_6.dds"},
	{0x55311C6B13BBF861ULL, "hlac_lt_backalley_n_65536_2y.dds"},
	{0x55312299564ACAA3ULL, "vxht_mansion2_3dy_0_0.vlm"},
	{0x55312AFD5EB83BEEULL, "dae000_flemeth6001887.cl"},
	{0x553130B4A4C50CAFULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.dura.ani"},
	{0x553163698B6C6BFDULL, "vxlt_0_lowtown_docks_a3_n_but_0_0.vlm"},
	{0x55317AA8221069BAULL, "6090371_m.fxe"},
	{0x55319DE98039B82DULL, "ser221_gascard6111757.cl"},
	{0x5531FEAC62F4FC3BULL, "art\\vfx\\combat\\vfx_g_spider_leap_act_c.impa.gad"},
	{0x55320B54C14CAAFEULL, "172521.ogg"},
	{0x5532261D746C579AULL, "vxlt_undercityserialkiller_ys_0_0.vlm"},
	{0x553234BE4ABC8ACBULL, "vxlt_0_lowtown_n_eqk_0_0.vlm"},
	{0x55328073F4BF3639ULL, "vxlt_0_lowtown_a2_n_c_e17_0_0.vlm"},
	{0x55337C8F75DE42AAULL, "zz_dae_debug6029420.cl"},
	{0x55338FC554B0374BULL, "349152636.ogg"},
	{0x5533FA12FEEA7FDBULL, "w_bonepit_78_0.mmh"},
	{0x55340612FEEA9478ULL, "w_bonepit_78_0.mao"},
	{0x5534109CF81CDBCFULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medh.mao"},
	{0x553416D60994F32DULL, "art\\characters\\creatures\\mubari\\animation\\cwkf.gad"},
	{0x55341812FEEAB2C5ULL, "w_bonepit_78_0.msh"},
	{0x55341C9CF81CF06CULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medh.mmh"},
	{0x5534229CF81CFA5EULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medh.msh"},
	{0x55342322DB383F8FULL, "ser100_party_comments6108236.cl"},
	{0x55345D71FCC4A072ULL, "dae000_male_commoner1_mob6174463.cl"},
	{0x5534DC23BBA625B6ULL, "mag291_vincento6027702.cl"},
	{0x5535455A34D5FE55ULL, "ser121_jethann6104432.cl"},
	{0x553590AF1EC0E56FULL, "w_bonepitdestroyed_a3_156_0.phy"},
	{0x5535AC7DCCA2D3E0ULL, "w_encounter2_3_0.rml"},
	{0x5535E42D95F417F8ULL, "6071761_m.fxe"},
	{0x5535F5B7FA44C93EULL, "136960258.ogg"},
	{0x55360623B2590F3CULL, "art\\vfx\\combat\\vfx_c_debili_aoe_imp_p.mmh"},
	{0x5536429845F511E9ULL, "mag101_grace6053113.cl"},
	{0x5536EA993D41C1BFULL, "832325729.ogg"},
	{0x5537164683500718ULL, "6109758_m.fxe"},
	{0x5537B2DE99749979ULL, "gen00fl_fenris6107175.cl"},
	{0x5537D9D7AA6E4422ULL, "llac_lt_darktown_-1_84.dds"},
	{0x553894B0368882AFULL, "dae000_meredith6092559.cl"},
	{0x5538A17279DF03F1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn04.mmh"},
	{0x5538BB7279DF2FAFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn04.msh"},
	{0x5538CB2E1A416254ULL, "214970294.ogg"},
	{0x5538EFB847C86963ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightc_n.dds"},
	{0x55390C0039569EE2ULL, "t1_mul_ta3.tnt"},
	{0x55391BAC62FB4C20ULL, "art\\vfx\\combat\\vfx_g_spider_leap_act_c.impa.evt"},
	{0x55396E7729629BD3ULL, "861661890.ogg"},
	{0x553987586A16D687ULL, "vxw_bonepit_21h_0_0.vlm"},
	{0x5539A793EC279CC2ULL, "741468550.ogg"},
	{0x5539B29D2DCE0D5FULL, "557187264.ogg"},
	{0x553A74D96C257CA2ULL, "96894273.ogg"},
	{0x553B1C3CE19AF608ULL, "vxht_keep_n_6mw_0_0.vlm"},
	{0x553C059E59FCC7A7ULL, "6085260_m.fxe"},
	{0x553C5C868A2C122CULL, "854213812.ogg"},
	{0x553D0BAD4608C90DULL, "ht_0_hightown_a3_59_0.rml"},
	{0x553D60C80AA0DCB1ULL, "hlac_dr_primevalentrance_1724_4y.dds"},
	{0x553DC2624BEFF3EDULL, "653165690.ogg"},
	{0x553E26FC64C8D2DBULL, "6068023_m.fxe"},
	{0x553E7075EB7EEC28ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar5_.gad"},
	{0x553EBB51F9428743ULL, "6048722_m.fxe"},
	{0x553ECFD00D6AA160ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_loop.gad"},
	{0x553EDEE7DE8B6018ULL, "lgt320_shw_ors_marlein6156313.cl"},
	{0x553F1285A03C0A17ULL, "75463778.ogg"},
	{0x553F75A6A5E64E97ULL, "bad_large_bs.mao"},
	{0x553F9221681DC011ULL, "vxht_0_hightown_a3_9p0_0_0.vlm"},
	{0x553F9CF435AA5E39ULL, "vxlt_0_lowtown_h8l_0_0.vlm"},
	{0x553F9F4729A4C44FULL, "gen00fl_aveline6139241.cl"},
	{0x554055D5F4C1A71FULL, "6162138_m.fxe"},
	{0x55406A9D915A5B60ULL, "vxlt_0_lowtown_docks_a3_n_d0t_0_0.vlm"},
	{0x5540F23E98D3A08FULL, "art\\characters\\playercharacter\\humanmale\\glv_masc.mao"},
	{0x5540FE3E98D3B52CULL, "art\\characters\\playercharacter\\humanmale\\glv_masc.mmh"},
	{0x5541043E98D3BF1EULL, "art\\characters\\playercharacter\\humanmale\\glv_masc.msh"},
	{0x55418365F4D1A745ULL, "t3_hey_br1.tnt"},
	{0x5541BA32447FFD30ULL, "581155621.ogg"},
	{0x5541E33A7DB796E4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_x2keep.msh"},
	{0x5541F4E2065694FFULL, "dae310ip_store_weapons.utp"},
	{0x5541F53A7DB7B54DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_x2keep.mao"},
	{0x5541F93A7DB7BB86ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_x2keep.mmh"},
	{0x5541FE86FCB0763AULL, "272309993.ogg"},
	{0x554225B225385374ULL, "6179466_f.fxe"},
	{0x5542466B52D728B5ULL, "mrl000_marethari6129327.cl"},
	{0x55426006F9154655ULL, "dae000_serendipity.bnk"},
	{0x5542C0A600677463ULL, "6205415_m.fxe"},
	{0x5542E6F2900E680BULL, "hlac_lt_ruinedhouse_n_-1_9x.dds"},
	{0x554324A50E475B89ULL, "643827119.ogg"},
	{0x5543316903A6EBB8ULL, "vxlt_0_lowtown_a2_gjx_0_0.vlm"},
	{0x554353130D2370A2ULL, "vxht_mansion_n_4h8_0_0.vlm"},
	{0x5543750529A42FC2ULL, "vxht_mansion_n_6hz_0_0.vlm"},
	{0x55439D3F3F368EE4ULL, "142835483.ogg"},
	{0x5543DFA727B0BF12ULL, "one100dg_nabil6033096.cl"},
	{0x5543FCCB08C47EE8ULL, "vxlt_hangedman_n_268_0_0.vlm"},
	{0x554573DF9FB76991ULL, "6109764_m.fxe"},
	{0x55458206F917B5ADULL, "dae000_serendipity.cnv"},
	{0x5545BC0978362C6AULL, "vxlt_0_lowtown_docks_b7y_0_0.vlm"},
	{0x5545D1AB0B36949AULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_5x.dds"},
	{0x55462548CEA58A30ULL, "hlac_dr_primevalentrance_1726_5y.dds"},
	{0x554648DE98D288CEULL, "vxw_blightlands_139_0_0.vlm"},
	{0x554650B2A7D87CC7ULL, "vxca_caverns1_mines_i7c_0_0.vlm"},
	{0x5546962DDC5A2981ULL, "6110251_f.fxe"},
	{0x5547EB47505F4AA4ULL, "pro000_aveline_intro6109371.cl"},
	{0x55484017C7B64B22ULL, "33688534.ogg"},
	{0x5548CD9755013778ULL, "dae000_denier6162670.cl"},
	{0x554903F0A348C57DULL, "vxht_mansion2_1h1_0_0.vlm"},
	{0x55491D4E656F1323ULL, "hlac_lt_darktown_-1_60y.dds"},
	{0x55492580949036D0ULL, "vxlt_undercityencounter_13t_0_0.vlm"},
	{0x55493A0E6A6451D5ULL, "hm_refugee01.mor"},
	{0x55493EA62B827700ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_floortile_r.dds"},
	{0x55497A60A76D2EE3ULL, "lt_0_lowtown_a2_36_0.rml"},
	{0x55498E19D3910E88ULL, "vxht_keep_n_6ht_0_0.vlm"},
	{0x5549A4B136D221A5ULL, "6126975_f.fxe"},
	{0x5549FEB43EFF8B29ULL, "561585317.ogg"},
	{0x554A6749CA51E340ULL, "hlac_w_woundedcoast_131073_0z.dds"},
	{0x554AF52821827165ULL, "389524864.ogg"},
	{0x554B0B726EADFE1AULL, "vxdr_primevaltomb_6rj_0_0.vlm"},
	{0x554B767B197E74D9ULL, "fx_c_grad_in_add.mao"},
	{0x554B799FAA1E2F93ULL, "195750505.ogg"},
	{0x554B98C06BA9223FULL, "lowtown_male_alienage1_sta6162807.cl"},
	{0x554BCA55CB78E327ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_45.gad"},
	{0x554BF03A7DC00151ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_x2keep.phy"},
	{0x554C0D3DA62BC0D2ULL, "6089576_f.fxe"},
	{0x554C385B34B8B30CULL, "gen00fl_fenris6108032.cl"},
	{0x554CD58EACEE50CAULL, "lt_0_lowtown_a3_n_c_28_0.anb"},
	{0x554D28445FA7580BULL, "art\\vfx\\combat\\vfx_glb_ele_n_deathblow_c.mmh"},
	{0x554D2D2EEB389FC3ULL, "6043622_m.fxe"},
	{0x554D509DF2596A52ULL, "zz_personality_debug6064397.cl"},
	{0x554D9158FB5A362AULL, "6067746_m.fxe"},
	{0x554DBF5F8AD92A2FULL, "art\\levels\\ztools\\default\\generic\\3pp_basic.phy"},
	{0x554DC1E9BBDA3C8AULL, "hlac_lt_0_lowtown_docks_n_131071_6z.dds"},
	{0x554E122BFD32F647ULL, "gen_im_trs_drk_amu_r0.uti"},
	{0x554EA190BD196F35ULL, "6082592_f.fxe"},
	{0x554EBC5EF103DD66ULL, "mag210_arianni6030480.cl"},
	{0x555070B21A85146AULL, "vxdr_primevaltomb_8vh_0_0.vlm"},
	{0x5550A61DA5CCF3D4ULL, "plt_mcr000pt_general.plo"},
	{0x5550D6B711CD1452ULL, "llac_lt_0_lowtown_a3_n_c_-3_1.dds"},
	{0x5550FE540E100C2CULL, "vxdr_dungeon_1m8_0_0.vlm"},
	{0x55514E409CCAFD87ULL, "hlac_w_sundermount_1_0y.dds"},
	{0x555159A8E69CCCCAULL, "gen_im_trs_apo_junk2_03.uti"},
	{0x5551759552C79A3CULL, "face_rendertodepth_p0.psh.fxo"},
	{0x55517FC297E3E549ULL, "gen00fl_fenris6106653.cl"},
	{0x5551E94098413DC1ULL, "6041770_m.fxe"},
	{0x555200C3C3FBA5CEULL, "vxlt_backalley_n_1pb_0_0.vlm"},
	{0x55522F64960DF6A5ULL, "vxw_encounter2_camp_49_0_0.vlm"},
	{0x555246768AB710F4ULL, "gen_im_npc_arm_hum_hlm_krka.uti"},
	{0x5552BF55CB7EEF34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_45.evt"},
	{0x555338221FF9FEA0ULL, "mcr310_argument6136211.cl"},
	{0x5553CFDA6A09C17AULL, "lt_0_lowtown_docks_n_151_0.mmh"},
	{0x5553D3DA6A09C7B1ULL, "lt_0_lowtown_docks_n_151_0.mao"},
	{0x5553E1DA6A09DFC0ULL, "lt_0_lowtown_docks_n_151_0.msh"},
	{0x5554600FBA8041CCULL, "171175338.ogg"},
	{0x555475C41D69F16FULL, "hlac_lt_0_lowtown_docks_-131071_0y.dds"},
	{0x5554C88A2C9E7332ULL, "vxlt_hangedman_265_0_0.vlm"},
	{0x555511F68D965096ULL, "vxht_keep_6lp_0_1.vlm"},
	{0x5555BA27B8D19C3FULL, "456618395.ogg"},
	{0x5556290B9764FAF0ULL, "952761408.ogg"},
	{0x55565A89DA747B49ULL, "912515105.ogg"},
	{0x55576E6E9BDCADABULL, "ch1armtnt_s2.psh.fxo"},
	{0x55577F4577A2B515ULL, "vxht_playershouse_n_1i8_0_0.vlm"},
	{0x5557D289218F00FEULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_en.evt"},
	{0x55584AFE39FAC2EBULL, "vxht_0_hightown_a2_8ay_0_0.vlm"},
	{0x55586294B297188CULL, "vxlt_gallowsprison_n_64s_0_0.vlm"},
	{0x5558CD2E085C9F82ULL, "vxlt_0_lowtown_docks_a3_btg_0_0.vlm"},
	{0x555924B2610C437BULL, "vxht_0_hightown_a3_b8y_0_0.vlm"},
	{0x55596E3FEFAFBF32ULL, "6088895_m.fxe"},
	{0x555980706510447FULL, "character_rendertodepth_p7.vsh.fxo.dx11.assembly"},
	{0x5559923DC15D1F29ULL, "vxht_0_hightown_n_c6j_0_0.vlm"},
	{0x5559A54D0EF68DC2ULL, "hlac_ht_0_hightown_n_-131069_0y.dds"},
	{0x555A55EAC6B32DABULL, "vxht_bartrand_n_67u_0_0.vlm"},
	{0x555B0149328CFB77ULL, "6150917_m.fxe"},
	{0x555B09445FB36170ULL, "art\\vfx\\combat\\vfx_glb_ele_n_deathblow_c.anb"},
	{0x555B298719297F9DULL, "6118995_m.fxe"},
	{0x555B819B2371B955ULL, "vxlt_darktown_4ze_0_0.vlm"},
	{0x555BB3DC9FC7BA94ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_1x.dds"},
	{0x555C2F76ACCF376FULL, "llac_dr_primevaltomb_658_11.dds"},
	{0x555C34EDF1C1113BULL, "zz_isa_debug6175183.cl"},
	{0x555C59D74C96E4A2ULL, "gen00fl_merrill6127547.cl"},
	{0x555CCB976C32D331ULL, "373715245.ogg"},
	{0x555CEC947B08CF56ULL, "6128322_m.fxe"},
	{0x555D47B83D911A32ULL, "vxdr_0_deeproads_782_0_1.vlm"},
	{0x555D6143DA0281B0ULL, "vxlt_0_lowtown_a3_n_c_byx_0_0.vlm"},
	{0x555EB874D0985426ULL, "6043408_f.fxe"},
	{0x555EBBCBB27C934DULL, "vxlt_0_lowtown_a2_n_gcm_0_0.vlm"},
	{0x555EC73988CA61F7ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\death.gad"},
	{0x555EDFA72601ED5EULL, "follower_banter6163546.cl"},
	{0x555EE28F0C7B2CEAULL, "vxdr_primevaltomb_8qi_0_0.vlm"},
	{0x555EE98921954629ULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_en.gad"},
	{0x555F356DCD9B4722ULL, "706921056.ogg"},
	{0x55603A32DF92D021ULL, "6065794_m.fxe"},
	{0x5560A2EBC7708505ULL, "lla_ht_0_hightown_n_-65534_0.dds"},
	{0x5560B755CB8B1F6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_45.ani"},
	{0x5560E084ECB0AFE8ULL, "vxlt_hangedman_n_286_0_0.vlm"},
	{0x5561AAEE8FF9E703ULL, "vxlt_0_lowtown_a3_n_c_ekf_0_0.vlm"},
	{0x5561C2BEDC4FBF32ULL, "812446023.ogg"},
	{0x5561C476202ACCF9ULL, "19655896.ogg"},
	{0x5562265B310A3547ULL, "524398390.ogg"},
	{0x55633A671E948BB2ULL, "vxlt_hangedman_236_0_0.vlm"},
	{0x5563E28A4BF4C2E4ULL, "seb221_flora_harimann6191385.cl"},
	{0x5563F61BB675FE04ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith.anb"},
	{0x55645BF4E14FFA5CULL, "436349760.ogg"},
	{0x55654EF13657F81EULL, "118186441.ogg"},
	{0x556552842A8671EBULL, "6142041_f.fxe"},
	{0x5565B7A611CBFB3BULL, "gen_im_trs_qun_hvy_boo_r0.uti"},
	{0x5565BA89219B1577ULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_en.ani"},
	{0x5565D940888B7F60ULL, "727527473.ogg"},
	{0x5565D9BA5A3BE70EULL, "vxlt_0_lowtown_a3_n_c_hsr_0_0.vlm"},
	{0x5565DC3988D0A384ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\death.evt"},
	{0x5565E42E5A9ED70EULL, "vxht_brothel_n_638_0_0.vlm"},
	{0x55660AAB8DCE5AEDULL, "llac_lt_hangedman_n_-1_25.dds"},
	{0x556658741B5DB98DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_frame.mmh"},
	{0x55665C741B5DC0DEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_frame.mao"},
	{0x55666A741B5DD813ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_frame.msh"},
	{0x556671C16EA6322CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraitsb_d.dds"},
	{0x5566FF82B6A617F4ULL, "hlac_lt_undercitydungeon_-1_14x.dds"},
	{0x55674DB8671CC9C9ULL, "vxlt_0_lowtown_a2_n_c_hvg_0_0.vlm"},
	{0x5568041AB35337A9ULL, "vxht_0_hightown_n_c3i_0_0.vlm"},
	{0x55680B4AA4308A87ULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\pine_s.dds"},
	{0x55682B351BAEA462ULL, "vxca_caverns2_6ym_0_0.vlm"},
	{0x55684AB6E727A69BULL, "558766817.ogg"},
	{0x55685504592E517DULL, "65182320.ogg"},
	{0x5568DCB8F172A1A5ULL, "gen_im_trs_bst_drg_eye_02.uti"},
	{0x55698EEF75D7E202ULL, "hlac_lt_warehouse_1_2z.dds"},
	{0x5569FC80273FB462ULL, "6081610_m.fxe"},
	{0x556A000B7FD06660ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs106a_0t.dds"},
	{0x556ABE850AB237D8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masa_0t.dds"},
	{0x556B2C4357C98D7CULL, "6188700_m.fxe"},
	{0x556C041D525E81BFULL, "mcr370_cut_epilogue6086874.cl"},
	{0x556C1858C4B47AC4ULL, "vxht_keep_n_583_0_0.vlm"},
	{0x556C36E77C60D02CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt.ani"},
	{0x556C47BBCD616E10ULL, "w_bonepitmining_a2_90_0.mao"},
	{0x556C4BBBCD617523ULL, "w_bonepitmining_a2_90_0.mmh"},
	{0x556C4E343E9C892EULL, "288244037.ogg"},
	{0x556C59BBCD618D7DULL, "w_bonepitmining_a2_90_0.msh"},
	{0x556C6B198FDFE0E8ULL, "plt_cod_crt_mabari.plo"},
	{0x556CC30F14686C07ULL, "lgt200_eau_tome6180571.cl"},
	{0x556DAF93387E9EA8ULL, "961454663.ogg"},
	{0x556DBD2976C999BDULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolema_movement.anb"},
	{0x556E03AB2E424A75ULL, "486810983.ogg"},
	{0x556E593ABF55873FULL, "hlac_lt_0_lowtown_docks_a3_-196607_0x.dds"},
	{0x556ED7BA8C0E054EULL, "hlac_lt_foundry_n_320_12x.dds"},
	{0x556ED7D9E86C9F5CULL, "hlac_lt_0_lowtown_a2_n_131072_3y.dds"},
	{0x556EED430162A164ULL, "vxlt_hangedman_5d_0_0.vlm"},
	{0x556F4C95D9756F7AULL, "zz_approval_debug6003748.cl"},
	{0x556FB680020D21ECULL, "vxht_keep_64c_0_0.vlm"},
	{0x55708B89B517FC9CULL, "vxlt_0_lowtown_n_h4v_0_0.vlm"},
	{0x55713B23DC95B799ULL, "6000720_m.fxe"},
	{0x557143E0287EEE48ULL, "6032828_f.fxe"},
	{0x55715130FAC719F2ULL, "mag110_arianni6016913.cl"},
	{0x55715C82820F3AADULL, "6164787_m.fxe"},
	{0x55719733E9CBAC64ULL, "6097390_m.fxe"},
	{0x5571D3D925F5EA14ULL, "6171519_m.fxe"},
	{0x5571DAEFFA3D42A1ULL, "vxht_playershouse_n_1o6_0_0.vlm"},
	{0x55720374C0A793A3ULL, "conversation_wheel_pos.gda"},
	{0x5572B6901F9778BEULL, "vxlt_0_lowtown_docks_a3_n_col_0_0.vlm"},
	{0x5572F1D6102D3200ULL, "619888228.ogg"},
	{0x5572F43988DB577DULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\death.ani"},
	{0x5573105061C492E8ULL, "qcr220_grey_wardens6060452.cl"},
	{0x557343E77C67056EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt.gad"},
	{0x5573B251D12F12CAULL, "6180910_f.fxe"},
	{0x55756F3B2ED1F367ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge15.gad"},
	{0x557587C8B49A9524ULL, "171189207.ogg"},
	{0x5575927ACE648D75ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipb.msh"},
	{0x5575947ACE6490EBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipb.mmh"},
	{0x5575A07ACE64A488ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slaveflipb.mao"},
	{0x5575AACAB0D4125CULL, "dae000ip_crowd_ff2_talk5.utp"},
	{0x5575BF955912E249ULL, "vxlt_0_lowtown_a2_n_c_hqf_0_0.vlm"},
	{0x557651F9D39BF032ULL, "6086695_m.fxe"},
	{0x557675F7A5495029ULL, "vxht_0_hightown_n_c8h_0_0.vlm"},
	{0x55768473CD53DF2EULL, "vxht_mansion2_f5_0_0.vlm"},
	{0x55769CCE46FC86E3ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_act_c.impa.evt"},
	{0x55770B410CA1B1A5ULL, "lgt210_eau_symp_templar6144480.cl"},
	{0x557751741B6C214AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_frame.phy"},
	{0x55778E1D815F6E30ULL, "vxw_blightlands_w6_0_0.vlm"},
	{0x557846D9179CA794ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\blendtree\\trooper_overlay.anb"},
	{0x5578EEDA6A291185ULL, "lt_0_lowtown_docks_n_151_0.phy"},
	{0x5578FB93DF8A7275ULL, "vxca_cave_2iy_0_0.vlm"},
	{0x55792A9CFF5B85DCULL, "art\\characters\\playercharacter\\humanmale\\cth_homa.mao"},
	{0x557935C68050FE32ULL, "gen_im_npc_arm_hum_glv_tmpd.uti"},
	{0x5579369CFF5B9A7FULL, "art\\characters\\playercharacter\\humanmale\\cth_homa.mmh"},
	{0x55793C9CFF5BA421ULL, "art\\characters\\playercharacter\\humanmale\\cth_homa.msh"},
	{0x55795AA18BEBB2D4ULL, "vxht_keep_n_8b4_0_0.vlm"},
	{0x55797D50C56EC7FEULL, "6055141_m.fxe"},
	{0x5579FB50323FDE24ULL, "mcr370_assignments6081734.cl"},
	{0x557A1AABFC3596EBULL, "art\\vfx\\combat\\vfx_glb_ele_c_deathblow_c.impa.ani"},
	{0x557A33CC38145B53ULL, "6078466_m.fxe"},
	{0x557A3EE77C6D1B4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\arcane_bolt.evt"},
	{0x557B44D2B44001E8ULL, "vxlt_0_lowtown_docks_b4g_0_1.vlm"},
	{0x557B659A8A17EF44ULL, "hlac_lt_0_lowtown_a2_-1_4y.dds"},
	{0x557B72A0364797BFULL, "hlac_lt_0_lowtown_a2_n_-262144_1z.dds"},
	{0x557C2A9D6D4DBB2CULL, "742558717.ogg"},
	{0x557C4E4B97FE834AULL, "gen00fl_dog6117190.cl"},
	{0x557CC6C1424BCEFEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownloop.ani"},
	{0x557CCCFC81A7FA6FULL, "509072648.ogg"},
	{0x557CD40CB030104CULL, "gen00fl_anders6125291.cl"},
	{0x557D616EC09EA283ULL, "vxca_caverns1_k2o_0_0.vlm"},
	{0x557D758E84290748ULL, "sebcr_merc_archer.utc"},
	{0x557D75CE47026364ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_act_c.impa.gad"},
	{0x557DD0B9F280AE17ULL, "vxw_bonepitmining_a2_d1_0_0.vlm"},
	{0x557E8EA873E61803ULL, "vxlt_0_lowtown_a3_n_c_eqh_0_0.vlm"},
	{0x557F0626AAA07951ULL, "hlac_lt_0_lowtown_a3_n_c_65536_1z.dds"},
	{0x557F6B9F35D6708BULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.impa.ani"},
	{0x557F7A19BE89B7AFULL, "vxca_cavern1_haunted_gke_0_0.vlm"},
	{0x557F7AA4D10FF3B1ULL, "6071089_m.fxe"},
	{0x55801C62F17B8E47ULL, "hlac_lt_darktown_-1_28y.dds"},
	{0x5580AD6481642704ULL, "vxw_woundedcoast_4e_0_0.vlm"},
	{0x558170B14F463709ULL, "vxca_caverns2prc_9lg_0_0.vlm"},
	{0x5582577C46AF4A0FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_female.gad"},
	{0x558257D0D727802DULL, "hlac_lt_gallowscourtyard_-131072_3y.dds"},
	{0x55829874ACFBCBAAULL, "6125624_m.fxe"},
	{0x5582D6A33841B693ULL, "dre000_vampire6034277.cl"},
	{0x5582F80FDFC64771ULL, "mer120_hubert6059405.cl"},
	{0x5583DD2D65798F7CULL, "6053750_m.fxe"},
	{0x5583E6D143B849B3ULL, "vxlt_backalley_n_1ra_0_0.vlm"},
	{0x558484CE47089B42ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_act_c.impa.ani"},
	{0x5584E2DFEE5B5199ULL, "vxht_0_hightown_b86_0_0.vlm"},
	{0x5584E37DF06DB3B7ULL, "art\\characters\\creatures\\corpse\\hed_cor.mao"},
	{0x5584EEB02847336DULL, "6070154_m.fxe"},
	{0x558524A168DD75FBULL, "lgt311_shw_mer_mettin6156467.cl"},
	{0x55863B4477911227ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v6.gad"},
	{0x5586AB43DA5B8F01ULL, "hlac_lt_0_lowtown_docks_131071_4x.dds"},
	{0x5586B4491E867DBAULL, "mag321_comtesse_de_launcet6043359.cl"},
	{0x5586D731F95402D5ULL, "vxlt_darktown_4kb_0_0.vlm"},
	{0x5586FEE07D71D9D2ULL, "6022562_f.fxe"},
	{0x558702ABFC3FF862ULL, "art\\vfx\\combat\\vfx_glb_ele_c_deathblow_c.impa.evt"},
	{0x5587C7D188564ED1ULL, "mrl150_dalish_guard6068291.cl"},
	{0x5587E46194643FFDULL, "6082155_m.fxe"},
	{0x55883B6B38AC0A1EULL, "6121963_m.fxe"},
	{0x55886C7C46B3D9BCULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_female.evt"},
	{0x55889AEEF02CAAF8ULL, "vxlt_0_lowtown_docks_a3_n_d35_0_0.vlm"},
	{0x55891495063CDA8DULL, "6127319_m.fxe"},
	{0x55892635CA4E19C9ULL, "922101879.ogg"},
	{0x558943476F0C8094ULL, "vxlt_gallowscourtyard_n_50v_0_0.vlm"},
	{0x55895C3B2EE27BADULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge15.ani"},
	{0x55896FD527498258ULL, "vxca_caverns2_83s_0_0.vlm"},
	{0x558978D0A9C0453CULL, "6034689_m.fxe"},
	{0x55897AAC3DBDA578ULL, "zz_dae_debug6045446.cl"},
	{0x558A2478EBBD0ADDULL, "625413233.ogg"},
	{0x558A2AEF83922CBAULL, "6162728_m.fxe"},
	{0x558A5D60A7A4945FULL, "lt_0_lowtown_a2_36_0.anb"},
	{0x558A6B8144459E35ULL, "775030948.ogg"},
	{0x558A72FB681FB163ULL, "1031342351.ogg"},
	{0x558ADEC1425834CFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownloop.evt"},
	{0x558B753673D3713DULL, "vxlt_warehouse_n_126_0_0.vlm"},
	{0x558C124B5B127EDCULL, "mrl150_dalish_guard6170399.cl"},
	{0x558C4296E476C697ULL, "vxw_bonepitmining_a2_c0_0_0.vlm"},
	{0x558C971F7E9DA61BULL, "dae000_sandal6147812.cl"},
	{0x558CCB66A79502E6ULL, "6000914_m.fxe"},
	{0x558CEFCFF94C107DULL, "vxht_bartrand_n_5sp_0_0.vlm"},
	{0x558D304477971E34ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v6.evt"},
	{0x558D539F35E28502ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.impa.evt"},
	{0x558DCE3C3E61B4A7ULL, "vxht_0_hightown_c5f_0_0.vlm"},
	{0x558DDD4835841D2EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_2.ani"},
	{0x558DE86924125AF0ULL, "360794931.ogg"},
	{0x558E01ABFC461575ULL, "art\\vfx\\combat\\vfx_glb_ele_c_deathblow_c.impa.gad"},
	{0x558E29D0CE3D5953ULL, "llac_ht_0_hightown_131072_0.dds"},
	{0x558E35745D4D8DB6ULL, "51769929.ogg"},
	{0x558F00A07E46AD63ULL, "art\\characters\\creatures\\ashwraith\\animation\\atk_v2.evt"},
	{0x558F659A0C6FFF7FULL, "606891676.ogg"},
	{0x558F9A4A0383A9BEULL, "vxlt_0_lowtown_docks_a3_n_cun_0_0.vlm"},
	{0x5590575C9AA0F02DULL, "vxht_0_hightown_2xj_0_0.vlm"},
	{0x5590692F0A88FA64ULL, "art\\vfx\\materials\\chunky\\vfx_c_d_clear.mao"},
	{0x55906A82CABAB4B1ULL, "307485244.ogg"},
	{0x55908E22B1702C79ULL, "llac_lt_0_lowtown_a3_n_c_131070_0.dds"},
	{0x5590B07FC89495B1ULL, "s_ballista_02_l3.mao"},
	{0x55912F51301E308EULL, "vxlt_0_lowtown_a3_n_c_hbu_0_0.vlm"},
	{0x559197CE1523EB06ULL, "614394216.ogg"},
	{0x5591A81A7D2C6D54ULL, "vxht_bartrand_n_4nw_0_0.vlm"},
	{0x5591DFC1425E5578ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownloop.gad"},
	{0x559226B1FCD542CAULL, "vxlt_warehouse_n_15l_0_0.vlm"},
	{0x5592321DC701A0FBULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtlrg.mmh"},
	{0x55923E1DC701B498ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtlrg.mao"},
	{0x55924502E3F2ECCEULL, "6207778_f.fxe"},
	{0x55924BA54C691F74ULL, "gen00fl_isabela6110158.cl"},
	{0x5592501DC701D3E5ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtlrg.msh"},
	{0x55927EBF2656637FULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.impa.gad"},
	{0x55928212D34127D7ULL, "vxht_brothel_n_3er_0_0.vlm"},
	{0x5592EEFD3B8FB88BULL, "vxw_blightlands_97_0_0.vlm"},
	{0x5592F0AD0A53725FULL, "700257059.ogg"},
	{0x55933E21883D3CF4ULL, "dae100st_solivitus.utm"},
	{0x55938CD1D413CC56ULL, "176559980.ogg"},
	{0x5593E0EBD426E4C8ULL, "dae300_alistair_king6133592.cl"},
	{0x5594303C705327E1ULL, "removerecipe.ncs"},
	{0x5594529F35E8A215ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.impa.gad"},
	{0x55945938FF6EE7CAULL, "dae000_madam_lusine6152233.cl"},
	{0x5594C7B622170569ULL, "hlac_lt_hangedman_n_-1_3z.dds"},
	{0x5594E96E7D34F171ULL, "lt_0_lowtown_docks_a3_n_229_0.msh"},
	{0x5594F76E7D35088CULL, "lt_0_lowtown_docks_a3_n_229_0.mao"},
	{0x5595036E7D351D2FULL, "lt_0_lowtown_docks_a3_n_229_0.mmh"},
	{0x559564A9E9B31386ULL, "798411259.ogg"},
	{0x5595993DD0C58661ULL, "isa311_isabela6092365.cl"},
	{0x5595C19EBD396DC2ULL, "830971978.ogg"},
	{0x5595D9A07E4C89E4ULL, "art\\characters\\creatures\\ashwraith\\animation\\atk_v2.gad"},
	{0x5595E77490BF1FBEULL, "zz_dae_debug6049989.cl"},
	{0x55964632ED6689E8ULL, "vxlt_ruinedhouse_n_ky_0_0.vlm"},
	{0x55964D14E938EF45ULL, "da2terrain_terrain_tools_p1.psh.fxo.dx11.assembly"},
	{0x5596747C46C024D5ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_reaction_female.ani"},
	{0x5596997E3C09214AULL, "gen_im_trs_bst_drg_gland_04.uti"},
	{0x5596BA99EBC672D0ULL, "dae130ar_undercity_darktown.lst"},
	{0x559741FB4986D804ULL, "mag360_alain6044600.cl"},
	{0x5597F45E045FCACCULL, "dae120cr_amb_dwarf_2.utc"},
	{0x559810CDB949722DULL, "follower_banter6151326.cl"},
	{0x559822C0170D74CEULL, "gen00fl_anders6119540.cl"},
	{0x559826E2FCB36FA3ULL, "vxw_encounter3_fv_0_0.vlm"},
	{0x559863C333FD84C8ULL, "vxht_mansion_n_53n_0_0.vlm"},
	{0x55986DDE6064E7F1ULL, "hlac_lt_0_lowtown_a2_65536_1z.dds"},
	{0x559873BF265ABC6CULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.impa.evt"},
	{0x5599395090A5E4E4ULL, "one140cr_charade6134614.cl"},
	{0x5599396AA7D321DAULL, "6016913_m.fxe"},
	{0x55995A11EA7FC549ULL, "vxca_caverns2prc_a9p_0_0.vlm"},
	{0x559982B02732EE29ULL, "hlac_w_bonepitdestroyed_a3_196610_0z.dds"},
	{0x5599872C4D310520ULL, "6016008_f.fxe"},
	{0x5599A4BD1A5FE1DAULL, "6163491_m.fxe"},
	{0x559A01CFC35DA8A3ULL, "vxht_keep_6yc_0_0.vlm"},
	{0x559AB473D66CDF17ULL, "vxw_bonepitmining_a2_n3_0_0.vlm"},
	{0x559B284477A34E6DULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v6.ani"},
	{0x559B312915B98939ULL, "vxht_chantry_n_2in_0_0.vlm"},
	{0x559B48078ED8646EULL, "210214758.ogg"},
	{0x559BD54835904D7FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_2.evt"},
	{0x559CB2A7E8050F74ULL, "6121203_f.fxe"},
	{0x559CE8A07E52C1C2ULL, "art\\characters\\creatures\\ashwraith\\animation\\atk_v2.ani"},
	{0x559D3B3DE65BD754ULL, "zz_followers_banter6157800.cl"},
	{0x559D3C26DC36374DULL, "6116859_f.fxe"},
	{0x559D53A0224D09B3ULL, "6180472_f.fxe"},
	{0x559D61050FBEE322ULL, "vxlt_shop_cs_0_0.vlm"},
	{0x559D9A99EBCC5A81ULL, "dae130ar_undercity_darktown.ncs"},
	{0x559DAE42DC2B84EEULL, "dae000ip_crowd_ff3_talk2.utp"},
	{0x559DAE916A5BD1B9ULL, "9897691.ogg"},
	{0x559E079956A124B7ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.impa.gad"},
	{0x559E45B727C928B1ULL, "6066382_m.fxe"},
	{0x559E70139701BF36ULL, "gen00fl_varric6033699.cl"},
	{0x559E8DBC743D2213ULL, "zz_vrc_debug6200261.cl"},
	{0x559EA4BA2D467259ULL, "ave310_donnic6134731.cl"},
	{0x559EBA70F352DAB6ULL, "406943352.ogg"},
	{0x559F116BE4220F84ULL, "289861158.ogg"},
	{0x559F6FF3931CB0FBULL, "452614092.ogg"},
	{0x559FA12E2214490EULL, "vxlt_0_lowtown_a3_n_c_hov_0_0.vlm"},
	{0x559FB6AC5DE9BB66ULL, "vxht_keep_n_70k_0_0.vlm"},
	{0x55A04D38E96632F9ULL, "626625754.ogg"},
	{0x55A06622982A0703ULL, "dre000_cut_sibling_dies6036308.cl"},
	{0x55A075E63E78FF28ULL, "543327013.ogg"},
	{0x55A0B199EBCEB7BBULL, "dae130ar_undercity_darktown.are"},
	{0x55A0C2BDA3A0B851ULL, "vxht_bartrand_n_6pv_0_0.vlm"},
	{0x55A0CA8B27A47AB3ULL, "vxlt_backalley_n_1xc_0_0.vlm"},
	{0x55A0DC2944080133ULL, "532577952.ogg"},
	{0x55A1ACEA37308381ULL, "6120596_m.fxe"},
	{0x55A20C9269001CA2ULL, "347466554.ogg"},
	{0x55A241F81AF43195ULL, "755654447.ogg"},
	{0x55A289C0BBE3B4E1ULL, "hlac_ht_playershouse_n_855_22x.dds"},
	{0x55A2B64835963768ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_2.gad"},
	{0x55A2BFBE11C34C15ULL, "dae000_mother6103043.cl"},
	{0x55A2F14B5F96AB7BULL, "mcr370_templar16086967.cl"},
	{0x55A31B76F4A8FA96ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_enter.ani"},
	{0x55A3AC18C6EEE407ULL, "mag121_idunna6084220.cl"},
	{0x55A41C9956A5B444ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.impa.evt"},
	{0x55A434AC42F86452ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\blendtree\\trooper.anb"},
	{0x55A502B6DDC7EA37ULL, "qun110_petrice6037693.cl"},
	{0x55A52378F693B883ULL, "6167488_f.fxe"},
	{0x55A540B253526E25ULL, "6034074_f.fxe"},
	{0x55A572C94D3982B8ULL, "879422307.ogg"},
	{0x55A60539E26F936DULL, "6169524_m.fxe"},
	{0x55A6078C3527CD72ULL, "735016950.ogg"},
	{0x55A65BBF2666D0C5ULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.impa.ani"},
	{0x55A69684D0BC7A1BULL, "6081101_m.fxe"},
	{0x55A698BFEEA98823ULL, "vxw_encounter3_cy_0_0.vlm"},
	{0x55A6A6E77767BC90ULL, "gen_im_trs_crc_sta_spi_r0.uti"},
	{0x55A7CBEEDC75DDC9ULL, "vxca_caverns2prc_a4w_0_0.vlm"},
	{0x55A873ACB553C123ULL, "vxht_keep_6bb_0_0.vlm"},
	{0x55A8AEA50B281EDEULL, "fire_ball.dds"},
	{0x55AAE192229A33B3ULL, "243236841.ogg"},
	{0x55AB482BE56F8005ULL, "6033124_m.fxe"},
	{0x55ABBEC7569509E1ULL, "vxht_mansion_n_59i_0_0.vlm"},
	{0x55AD1AB2C0A4C546ULL, "268380670.ogg"},
	{0x55AD5B93C3BE3B32ULL, "mag370_meredith6039277.cl"},
	{0x55AD8CED50F5D194ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.impa.ani"},
	{0x55ADDE126F9D8236ULL, "6051310_m.fxe"},
	{0x55AE6973BD5DD909ULL, "694335661.ogg"},
	{0x55AE8AF847351994ULL, "6162458_m.fxe"},
	{0x55AF052B1F089AAEULL, "348584688.ogg"},
	{0x55AF77A256A9C612ULL, "follower_banter6026249.cl"},
	{0x55AFC301926DB1ACULL, "llac_blank_exterior_0_0x.dds"},
	{0x55B069C04A827C56ULL, "dre000_dougal6129770.cl"},
	{0x55B0DA6DAB6B818AULL, "hlac_ht_keep_n_3739_0x.dds"},
	{0x55B12376F4B54587ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_enter.evt"},
	{0x55B14BBFC8E0326BULL, "zz_followers_banter6158457.cl"},
	{0x55B15A32F17BB1F7ULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure_0.pwk"},
	{0x55B2349956B21A3DULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.impa.ani"},
	{0x55B2776ECA074513ULL, "6035304_m.fxe"},
	{0x55B2967E459103B3ULL, "vxlt_0_lowtown_docks_a3_aq6_0_0.vlm"},
	{0x55B33CE27719218CULL, "1000839360.ogg"},
	{0x55B3CE3AA8B37825ULL, "hlac_lt_gallowstemplar_f_2383_2z.dds"},
	{0x55B3D3620E61A476ULL, "fen310_party_comment6116871.cl"},
	{0x55B45530C8F33825ULL, "6075975_m.fxe"},
	{0x55B4579346F942A4ULL, "rdr111dg_martin6054508.cl"},
	{0x55B489ED50FBEBF6ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.impa.gad"},
	{0x55B5477D17E9B5C8ULL, "vxht_mansion_n_59l_0_0.vlm"},
	{0x55B58368EF147F96ULL, "vxlt_hangedman_n_w9_0_0.vlm"},
	{0x55B63DCBCE6BF649ULL, "vxca_caverns2prc_a3v_0_0.vlm"},
	{0x55B78A11A38182A2ULL, "6101543_m.fxe"},
	{0x55B84476F4BB9CD0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_enter.gad"},
	{0x55B84A3521150541ULL, "w_sundermount_126_0.rml"},
	{0x55B8827D59129902ULL, "139825954.ogg"},
	{0x55B8EBCE36D0E7ABULL, "art\\vfx\\combat\\vfx_c_mer_grab_stab_c.impa.evt"},
	{0x55B975FED55F1593ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03b.phy"},
	{0x55B9B32B1660DEA1ULL, "187190225.ogg"},
	{0x55B9B8DD84DD9EA7ULL, "one280im_control_rod.uti"},
	{0x55BA7441CAE8B545ULL, "hlac_ht_0_hightown_n_-131072_1y.dds"},
	{0x55BA9C7DB1FD595FULL, "6157192_m.fxe"},
	{0x55BB0292C6B3C4ACULL, "vxlt_gallowstemplar_53j_0_0.vlm"},
	{0x55BB0354CCDD42B1ULL, "hlac_lt_0_lowtown_a2_65535_0x.dds"},
	{0x55BB4FF7E94C6832ULL, "follower_banter6172830.cl"},
	{0x55BBA4ED51023855ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.impa.evt"},
	{0x55BBE38A0C03CD38ULL, "6207959_m.fxe"},
	{0x55BC3CCCCA8413D5ULL, "627533689.ogg"},
	{0x55BCA6F7E5EF5087ULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef_eyesm1.mao"},
	{0x55BD691F4C4D996BULL, "vxlt_undercityencounter_13j_0_1.vlm"},
	{0x55BDDC3E6BA2A822ULL, "401147243.ogg"},
	{0x55BDF8455A2AB050ULL, "llac_w_woundedcoast_262151_0.dds"},
	{0x55BE4D6FF3E39F52ULL, "vxlt_0_lowtown_a3_n_c_ep3_0_1.vlm"},
	{0x55BE86E1BFCF0F70ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01tree.msh"},
	{0x55BE94E1BFCF268AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01tree.mmh"},
	{0x55BE98E1BFCF2DC1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01tree.mao"},
	{0x55BEBB9905E71D03ULL, "vxlt_0_lowtown_docks_b05_0_0.vlm"},
	{0x55BEC4B574DED419ULL, "framebuffereffects_linearfade_p0.psh.fxo"},
	{0x55BEF71DD373E1BCULL, "6177221_m.fxe"},
	{0x55BEF8E23D3FADB1ULL, "6098831_m.fxe"},
	{0x55BF233BD0BB9ABBULL, "6105674_m.fxe"},
	{0x55BFCDBB7094B3EAULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluviancrack_d.dds"},
	{0x55C004CE36D7316CULL, "art\\vfx\\combat\\vfx_c_mer_grab_stab_c.impa.gad"},
	{0x55C01CBDA4419CEEULL, "6181752_m.fxe"},
	{0x55C07691E73AFAC8ULL, "dae000_gamlen6144347.cl"},
	{0x55C08B9578179D69ULL, "hlac_lt_0_lowtown_a2_n_c_65538_3y.dds"},
	{0x55C0E66A473310E4ULL, "vxht_0_hightown_n_ca2_0_0.vlm"},
	{0x55C19BC9C348D368ULL, "vxlt_ruinedhouse_n_tt_0_0.vlm"},
	{0x55C2290FF2373096ULL, "834391505.ogg"},
	{0x55C27B35D8993638ULL, "75542977.ogg"},
	{0x55C297A849C2A8E9ULL, "hlac_ht_mansion2_818_9z.dds"},
	{0x55C2A4A8949A3F63ULL, "hlac_ca_caverns1_mines_24421_8z.dds"},
	{0x55C50F2741036077ULL, "6005394_m.fxe"},
	{0x55C55766993FF223ULL, "vxht_keep_6hd_0_0.vlm"},
	{0x55C55950F41DC79AULL, "355621909.ogg"},
	{0x55C580E85A1FD7E6ULL, "character_character_p12.psh.fxo.dx11"},
	{0x55C6249D374CA050ULL, "vxw_bonepitdestroyed_a3_nb_0_0.vlm"},
	{0x55C65A82CFF2F112ULL, "vxht_0_hightown_a3_cda_0_0.vlm"},
	{0x55C6D929A1C9F883ULL, "narrative_banter6165545.cl"},
	{0x55C6F9DD91257FA5ULL, "6169825_m.fxe"},
	{0x55C70038356D205EULL, "361630528.ogg"},
	{0x55C703CE36DD4E6AULL, "art\\vfx\\combat\\vfx_c_mer_grab_stab_c.impa.ani"},
	{0x55C704EF5965EEE4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_keepdoor_1.mao"},
	{0x55C73D6CEF22795CULL, "follower_banter6163900.cl"},
	{0x55C7403D36FDAABCULL, "vxht_0_hightown_b8o_0_0.vlm"},
	{0x55C7AE6BA08006E4ULL, "llac_lt_0_lowtown_docks_n_196609_0.dds"},
	{0x55C7F25A2B69A8CCULL, "449037069.ogg"},
	{0x55C9448594BC149AULL, "hlac_ht_keep_3731_1x.dds"},
	{0x55C9A43FC352B3C1ULL, "vxlt_gallowscourtyard_6yp_0_0.vlm"},
	{0x55C9BE95D60D215AULL, "6207768_m.fxe"},
	{0x55C9C365A122FE0FULL, "vxlt_0_lowtown_docks_a3_bu5_0_0.vlm"},
	{0x55CA490884F62AB8ULL, "77345774.ogg"},
	{0x55CA55C328F6F41AULL, "6125431_m.fxe"},
	{0x55CA7D07B16A4D39ULL, "follower_banter6143325.cl"},
	{0x55CA98D3D6223123ULL, "vxht_0_hightown_n_bgn_0_0.vlm"},
	{0x55CB0376E680DDF1ULL, "vxlt_0_lowtown_h8t_0_0.vlm"},
	{0x55CB486F18D45740ULL, "gen00fl_fenris6126533.cl"},
	{0x55CBB2E659502FEEULL, "361414308.ogg"},
	{0x55CC4B6FD1B04E7BULL, "llac_lt_0_lowtown_a2_n_131070_3.dds"},
	{0x55CC98D606F95E75ULL, "isa211_isabela.bnk"},
	{0x55CCC605E92A81B8ULL, "zz_vault_debug6203170.cl"},
	{0x55CCF4D25A30FCE2ULL, "hlac_lt_gallowscourtyard_n_-65537_3y.dds"},
	{0x55CD2EB309F11A7EULL, "vxht_0_hightown_a3_b9k_0_0.vlm"},
	{0x55CDD40A1195A522ULL, "vxlt_gallowscourtyard_76w_0_0.vlm"},
	{0x55CDDE9250C14BF9ULL, "vxw_bonepit_2a6_0_0.vlm"},
	{0x55CE37774027F3A3ULL, "765878879.ogg"},
	{0x55CE53DBD90F8C63ULL, "follower_banter6136451.cl"},
	{0x55CEFB4FE73BA185ULL, "qcr000_arishok6166987.cl"},
	{0x55CF3F22B5035BF6ULL, "6201168_m.fxe"},
	{0x55CFAC24B3C6D8F6ULL, "6135658_m.fxe"},
	{0x55CFBAD606FBCDCDULL, "isa211_isabela.cnv"},
	{0x55CFC6C0931877B1ULL, "vxw_sundermount_31x_0_0.vlm"},
	{0x55D0571067663902ULL, "llac_w_blightlands_-65535_0.dds"},
	{0x55D0AEE23BA0738EULL, "172886051.ogg"},
	{0x55D1193CB04F984BULL, "vxlt_0_lowtown_n_e8d_0_0.vlm"},
	{0x55D13579030F80FFULL, "mrl310_merrill6119957.cl"},
	{0x55D14605757F099DULL, "zz_fen_invis_wall.utp"},
	{0x55D183E8AB7A1292ULL, "vxlt_gallowscourtyard_n_4iv_0_0.vlm"},
	{0x55D25082C6C02C9CULL, "vxlt_0_lowtown_docks_a3_n_d2s_0_1.vlm"},
	{0x55D33CBF4563D29DULL, "vxw_encounter2_pt_0_1.vlm"},
	{0x55D3D9FBDA48B91DULL, "character_character_p82.psh.fxo"},
	{0x55D473D520899C90ULL, "978302047.ogg"},
	{0x55D4967A2942B8D0ULL, "vxw_bonepitdestroyed_a3_ic_0_0.vlm"},
	{0x55D4B270162BE308ULL, "plt_lgt000pt_chanter.plo"},
	{0x55D57174AB007E54ULL, "728431162.ogg"},
	{0x55D58118CA47C4B5ULL, "hlac_dr_0_deeproads_17038_6z.dds"},
	{0x55D58B72AE42FB88ULL, "hlac_ca_caverns2_5802_30x.dds"},
	{0x55D5C99BEAB6AC64ULL, "vxlt_gallowstemplar_43s_0_0.vlm"},
	{0x55D6231BC2624B6BULL, "272452759.ogg"},
	{0x55D637DC5D214F32ULL, "dae000_gen_turn_in_36198687.cl"},
	{0x55D6A3C901FF98D0ULL, "6068587_m.fxe"},
	{0x55D75C9BE0C9786EULL, "593896260.ogg"},
	{0x55D790F9DD1606BCULL, "56988880.ogg"},
	{0x55D7AD2AFB39B3EFULL, "60600737.ogg"},
	{0x55D7C0951362E256ULL, "w_encounter1_18_0.msh"},
	{0x55D7CA951362F354ULL, "w_encounter1_18_0.mmh"},
	{0x55D7CE951362FA67ULL, "w_encounter1_18_0.mao"},
	{0x55D89EEA28D51EA7ULL, "vxht_bartrand_n_5pj_0_1.vlm"},
	{0x55D8BDB4E81359CFULL, "6208258_m.fxe"},
	{0x55D8C7A01AED0C9AULL, "6122531_m.fxe"},
	{0x55D90AB0C81849A3ULL, "vxht_0_hightown_n_bbq_0_0.vlm"},
	{0x55D94A957E0BE383ULL, "156046760.ogg"},
	{0x55D94C1971A80DEEULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boilerb.phy"},
	{0x55DA66E8233C205EULL, "art\\characters\\creatures\\mubari\\animation\\p_wkb.evt"},
	{0x55DA8E0F87799D7FULL, "6016901_m.fxe"},
	{0x55DAC1B7E3BD5376ULL, "372981561.ogg"},
	{0x55DAED1BEE994BD7ULL, "6084707_f.fxe"},
	{0x55DB55BE1AC964BEULL, "fx_magic_shield.mao"},
	{0x55DBD1D8F1ECDE46ULL, "mag291_demon_marethari6028878.cl"},
	{0x55DC3F9CFFB026C4ULL, "art\\characters\\playercharacter\\humanmale\\cth_homa.phy"},
	{0x55DC94469075DC42ULL, "997613135.ogg"},
	{0x55DCB464FFD79AAEULL, "6108218_m.fxe"},
	{0x55DCE4AF7AE87B87ULL, "847178642.ogg"},
	{0x55DF34D55DE4CDEAULL, "vxw_woundedcoast_gb_0_0.vlm"},
	{0x55DF81DD23EEF3D3ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.ani"},
	{0x55DF96781F6AD89FULL, "vxw_bonepitdestroyed_a3_25t_0_0.vlm"},
	{0x55E00EDC94A15DB7ULL, "llac_lt_0_lowtown_docks_a3_n_-131075_1.dds"},
	{0x55E02C13090E86F7ULL, "mag100_karras6034937.cl"},
	{0x55E04FFBD87702C6ULL, "hlac_w_sundermount_262147_0y.dds"},
	{0x55E0DFE40438E8F3ULL, "vxw_woundedcoast_12a_0_0.vlm"},
	{0x55E17DE823426589ULL, "art\\characters\\creatures\\mubari\\animation\\p_wkb.gad"},
	{0x55E24BD8CC3ED19DULL, "character_character_p73.psh.fxo"},
	{0x55E25A4E5BEBB9BDULL, "vxlt_backalley_n_1ev_0_0.vlm"},
	{0x55E27055D3FADA5EULL, "store_lgt_epic2_3.utm"},
	{0x55E2C2D6E4AA8268ULL, "depth_debug.psh.fxo"},
	{0x55E301BB86D0B733ULL, "gen_im_trs_bst_drag_gland_02.uti"},
	{0x55E33FF6F4CAF7ADULL, "vxht_0_hightown_a3_bze_0_0.vlm"},
	{0x55E3CE0DDFA2534CULL, "follower_banter6171529.cl"},
	{0x55E4038CC2EF6291ULL, "hlac_lt_0_lowtown_a2_-131072_6y.dds"},
	{0x55E40E638DDC4DFAULL, "6061807_m.fxe"},
	{0x55E4C25B6A59566FULL, "hlac_w_bonepitdestroyed_a3_65541_0x.dds"},
	{0x55E4DFE8BC909B23ULL, "129254958.ogg"},
	{0x55E54D0CAF07A66AULL, "one300_nuncio6070814.cl"},
	{0x55E619F7467B68E3ULL, "w_woundedcoast_234_0.mao"},
	{0x55E61DF7467B6F30ULL, "w_woundedcoast_234_0.mmh"},
	{0x55E62BF7467B874AULL, "w_woundedcoast_234_0.msh"},
	{0x55E62F7F485FCF39ULL, "dae000_flemeth6013261.cl"},
	{0x55E67D3D4D0C86B7ULL, "hlac_ca_caverns1_mines_24421_2x.dds"},
	{0x55E68D4B81A0C20AULL, "mag121_idunna6084245.cl"},
	{0x55E6F14442099B1DULL, "dae000_keep_guard_1_sta.bnk"},
	{0x55E71C869D0BF14BULL, "1033415143.ogg"},
	{0x55E73F84A3E7B76EULL, "6051822_m.fxe"},
	{0x55E7468835DB5755ULL, "vxlt_warehouse_n_8c_0_0.vlm"},
	{0x55E7E4F2DC45B38EULL, "vxht_brothel_n_6d1_0_0.vlm"},
	{0x55E7F8C4D06F29B4ULL, "follower_banter6160853.cl"},
	{0x55E84EE8234834D7ULL, "art\\characters\\creatures\\mubari\\animation\\p_wkb.ani"},
	{0x55E8729839241C08ULL, "vxht_keep_n_61k_0_0.vlm"},
	{0x55E908C67A921441ULL, "591191829.ogg"},
	{0x55E9631CB10423E1ULL, "vxw_encounter2_camp_rs_0_0.vlm"},
	{0x55EA91A5E1C79FF0ULL, "1037602090.ogg"},
	{0x55EAD344420D5125ULL, "dae000_keep_guard_1_sta.cnv"},
	{0x55EB49BDE5C30973ULL, "227024301.ogg"},
	{0x55EB619086F15483ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1g.mao"},
	{0x55EB659086F15BD0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1g.mmh"},
	{0x55EB739086F173EAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1g.msh"},
	{0x55EB9E59C429687EULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv2.evt"},
	{0x55EC05956A65D3BFULL, "fen110_fenris6118132.cl"},
	{0x55EC3F7DAB79A74DULL, "hlac_lt_0_lowtown_a3_n_c_65535_7y.dds"},
	{0x55EC73B0A6CD2C04ULL, "6108821_m.fxe"},
	{0x55EC79DD23F9700AULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.evt"},
	{0x55ECA047DCEF6EDEULL, "6121044_m.fxe"},
	{0x55ED0A3395B44630ULL, "507789105.ogg"},
	{0x55EDF8A921230F42ULL, "sib131_slaver_lord6165197.cl"},
	{0x55EE41D6A28812B8ULL, "6133139_m.fxe"},
	{0x55EE7AB8DCF97630ULL, "vxlt_darktown_43k_0_0.vlm"},
	{0x55EE849C63292981ULL, "gencr_templar_hunter.utc"},
	{0x55EFCD2F0E793BF6ULL, "144778852.ogg"},
	{0x55EFDAF1579BD3B9ULL, "6101470_m.fxe"},
	{0x55F020A5F79B7124ULL, "w_bonepitdestroyed_a3_145_0.mmh"},
	{0x55F024A5F79B7877ULL, "w_bonepitdestroyed_a3_145_0.mao"},
	{0x55F036A5F79B96C6ULL, "w_bonepitdestroyed_a3_145_0.msh"},
	{0x55F07FF3E3ADDFC5ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls102a_0d.dds"},
	{0x55F0BDB5BE34EA1DULL, "character_character_p20.psh.fxo"},
	{0x55F0CBE00FE51224ULL, "vxw_encounter1_cd_0_0.vlm"},
	{0x55F0D10AFB2E5A16ULL, "6171935_m.fxe"},
	{0x55F260C694B39AE1ULL, "vxca_caverns2prc_8bd_0_0.vlm"},
	{0x55F28B772B6149C9ULL, "481977575.ogg"},
	{0x55F29FBC0615F186ULL, "hlac_w_encounter2_camp_131074_0x.dds"},
	{0x55F2B559C42FADA9ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv2.gad"},
	{0x55F2D27E8273D056ULL, "218112528.ogg"},
	{0x55F2E4E078855D3AULL, "6085232_m.fxe"},
	{0x55F2EBD27258CFC8ULL, "vxw_encounter2_camp_rv_0_0.vlm"},
	{0x55F358DD23FF577DULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.gad"},
	{0x55F409E3C854915BULL, "209340423.ogg"},
	{0x55F502533E10DBDBULL, "t3_glk_h01.tnt"},
	{0x55F6020D2F68FB8FULL, "dre000_varric6034676.cl"},
	{0x55F6042E046C5C85ULL, "676174808.ogg"},
	{0x55F67B79A91EC233ULL, "770701234.ogg"},
	{0x55F6AF701648C4C7ULL, "plt_lgt000pt_chanter.nss"},
	{0x55F6D2EF3028AFC5ULL, "mrl150_merrill6078285.cl"},
	{0x55F74D96748F7A6CULL, "framebuffereffects_ssao_p0.psh.fxo"},
	{0x55F7A73709D18FA0ULL, "qun160_vashothbandits6069285.cl"},
	{0x55F7BF2C7B5EA27BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyo_0t.dds"},
	{0x55F7C8D827CEE76CULL, "hlac_lt_0_lowtown_a2_n_131070_5y.dds"},
	{0x55F7D14A48D86619ULL, "vxdr_primevalentrance_1mc_0_0.vlm"},
	{0x55F7D4F9A2FA3C61ULL, "vxw_encounter2_camp_mr_0_0.vlm"},
	{0x55F8303653DDDE83ULL, "vxlt_undercitydungeon_1jd_0_1.vlm"},
	{0x55F86E9506CBACD1ULL, "6044188_m.fxe"},
	{0x55F8978638DDC81AULL, "370141637.ogg"},
	{0x55F8B29A31D5EB08ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha13a.msh"},
	{0x55F8D09A31D61E32ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha13a.mmh"},
	{0x55F8EBF6858AAAF2ULL, "character_character_p96.psh.fxo.dx11"},
	{0x55F93C35214C8430ULL, "w_sundermount_126_0.msh"},
	{0x55F94A35214C9C4AULL, "w_sundermount_126_0.mmh"},
	{0x55F94E35214CA281ULL, "w_sundermount_126_0.mao"},
	{0x55F96A26E29F857EULL, "6217141_m.fxe"},
	{0x55F98659C4357CF7ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv2.ani"},
	{0x55FA7088E1DF68B0ULL, "dae000_oldmangossip6175772.cl"},
	{0x55FAB96A2577241FULL, "1038736083.ogg"},
	{0x55FAD426D2C59759ULL, "6131198_m.fxe"},
	{0x55FAF4BC0124A9E4ULL, "vxht_0_hightown_cat_0_1.vlm"},
	{0x55FB09A61F3A3CACULL, "373588132.ogg"},
	{0x55FB6B714736F82CULL, "120932368.ogg"},
	{0x55FBE64F089BF7F0ULL, "gen00fl_anders6113049.cl"},
	{0x55FC1E4EC3230D24ULL, "hlac_w_sundermount_262144_0z.dds"},
	{0x55FC5F103FF9FD1BULL, "6153967_m.fxe"},
	{0x55FC6CACCF2E8F35ULL, "hlac_w_bonepit_458755_0y.dds"},
	{0x55FD1EDB29AB3219ULL, "129375451.ogg"},
	{0x55FD726058D6769BULL, "804338074.ogg"},
	{0x55FDC39DE82519F3ULL, "vxw_woundedcoast_18c_0_0.vlm"},
	{0x55FED68C9B9B6C25ULL, "eye_eye_p0.vsh.fxo.dx11.assembly"},
	{0x55FF35BB1673046AULL, "vxw_bonepit_1qj_0_0.vlm"},
	{0x55FF3DBD01DB2AA4ULL, "vxw_encounter1_he_0_0.vlm"},
	{0x55FF3E083FD7EABDULL, "vxlt_backalley_n_1kt_0_0.vlm"},
	{0x55FFA19C4A5A5B11ULL, "explo01.dds"},
	{0x55FFCF5BF55953FDULL, "99557877.ogg"},
	{0x5600028ED40CCC3CULL, "308077845.ogg"},
	{0x5600529BDD2B0C0FULL, "467724588.ogg"},
	{0x5600ABE06CB5D6C2ULL, "vxlt_undercitydungeon_1ie_0_0.vlm"},
	{0x56015DAF644EE848ULL, "vxw_encounter2_camp_mw_0_0.vlm"},
	{0x560226146523A247ULL, "genip_trap_fire_jet_short.utp"},
	{0x560297FAFA9E2AADULL, "vxlt_0_lowtown_a2_n_hhk_0_0.vlm"},
	{0x5602DEBFCF3760BFULL, "121811564.ogg"},
	{0x56033C67340B1403ULL, "llac_lt_0_lowtown_docks_-2_0.dds"},
	{0x560342AFDB6474FCULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic_1.mao"},
	{0x56034A91F7813174ULL, "617762191.ogg"},
	{0x56038695DF760236ULL, "vxht_playershouse_n_1th_0_0.vlm"},
	{0x56045F134EE20B70ULL, "w_blightlands_132_0.mao"},
	{0x560463134EE21183ULL, "w_blightlands_132_0.mmh"},
	{0x560471134EE229DDULL, "w_blightlands_132_0.msh"},
	{0x5604B16499BE17C7ULL, "hlac_ca_caverns2_5802_10y.dds"},
	{0x5604C8ACC031E48EULL, "vxht_brothel_n_6j3_0_0.vlm"},
	{0x5604D6CFB5D280FCULL, "zz_fen_debug6215244.cl"},
	{0x560593AE480FFDB8ULL, "gen00fl_anders6103747.cl"},
	{0x560639FA256B89DBULL, "vxdr_0_deeproads_9gq_0_0.vlm"},
	{0x560646D694F054E1ULL, "vxw_encounter2_camp_hq_0_0.vlm"},
	{0x560677C97A514658ULL, "vxdr_primevaltomb_6yy_0_0.vlm"},
	{0x5606B615D577BC12ULL, "55811307.ogg"},
	{0x5606F3E8D90805BCULL, "vxlt_warehouse_n_1bc_0_0.vlm"},
	{0x56083DDB32126AE0ULL, "vxht_mansion2b_320_0_0.vlm"},
	{0x560891C9EEF6F94CULL, "vxlt_hangedman_n_24v_0_0.vlm"},
	{0x5608DEFF777F9E14ULL, "437124472.ogg"},
	{0x56096A70F3B07E7EULL, "vxlt_0_lowtown_a2_n_h1w_0_0.vlm"},
	{0x560A9BF9D754EE69ULL, "lt_house.arl"},
	{0x560A9F8F1D1D272EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_enter.evt"},
	{0x560ABD937B2F7696ULL, "vxht_mansion_n_5c0_0_0.vlm"},
	{0x560B2590867443B8ULL, "6139739_m.fxe"},
	{0x560B8A75C96F6802ULL, "6070409_m.fxe"},
	{0x560BD9791141A170ULL, "hlac_lt_gallowstemplar_f_2390_0y.dds"},
	{0x560C21A14493241CULL, "hlac_lt_0_lowtown_docks_n_131071_1y.dds"},
	{0x560C5513FD741949ULL, "follower_banter6141657.cl"},
	{0x560D450E9D66A23BULL, "659401224.ogg"},
	{0x560D6C664012F8FCULL, "hightown_fem_cityguard_mob6168480.cl"},
	{0x560D98577B864925ULL, "hlac_ht_playershouse_n_855_0x.dds"},
	{0x560DA79808691CEAULL, "vxw_bonepit_1zk_0_0.vlm"},
	{0x560DADFA52233005ULL, "6043707_m.fxe"},
	{0x560DAFE531CE033DULL, "vxlt_backalley_n_1vs_0_0.vlm"},
	{0x560DBB434309F4FAULL, "vxdr_primevalentrance_2cu_0_0.vlm"},
	{0x560DD672F01BB130ULL, "mcr310_sibling6076253.cl"},
	{0x560E0DBBF85069EAULL, "6000815_m.fxe"},
	{0x560E1641D7D58B0AULL, "vxca_caverns1_cjr_0_0.vlm"},
	{0x560E4C0211593359ULL, "dre000_sibling_has_blight6036512.cl"},
	{0x560E80411F44CEBDULL, "6119215_m.fxe"},
	{0x560EA49445D87964ULL, "6105894_f.fxe"},
	{0x560EE933DC5276BFULL, "6132830_m.fxe"},
	{0x560EF33E097E0B8CULL, "vxca_caverns2prc_9df_0_0.vlm"},
	{0x560F68825DD4439BULL, "hlac_lt_darktown_-1_23x.dds"},
	{0x560FA64F22FC1EC2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_boilerbase_d.dds"},
	{0x560FCDC9C20FEABEULL, "6102186_m.fxe"},
	{0x560FCF8C564500C8ULL, "vxw_encounter2_camp_ht_0_0.vlm"},
	{0x5610046F661F6933ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.anb"},
	{0x56100E6F73995C28ULL, "6072200_f.fxe"},
	{0x561098EDA62D68FAULL, "vxht_0_hightown_n_avw_0_0.vlm"},
	{0x5610B383FC4E9FA9ULL, "dae250ip_at_scaverns_b.utp"},
	{0x5610D4456E9543C3ULL, "419369610.ogg"},
	{0x5610D7C9FF06E29AULL, "zz_and_debug6134088.cl"},
	{0x5611968F1D2336D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_enter.gad"},
	{0x5611B3FF0540CF4CULL, "vxlt_0_lowtown_a2_n_c_htd_0_0.vlm"},
	{0x5611CA9CF9035D84ULL, "vxht_playershouse_n_1j4_0_0.vlm"},
	{0x5611D7BDB95377E1ULL, "531852520.ogg"},
	{0x5613989F78994EA0ULL, "vxca_caverns2_7ar_0_0.vlm"},
	{0x5613AE7134B5E091ULL, "fx_impact013.dds"},
	{0x5614B0226D8798C2ULL, "aip_qunarigeneral_base.gda"},
	{0x5614B8B386E66D61ULL, "vxw_encounter2_camp_cp_0_0.vlm"},
	{0x56157A3FF2556E31ULL, "6043884_m.fxe"},
	{0x5615F68C279C5173ULL, "6167209_m.fxe"},
	{0x561631228CFFA31BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyn_0t.dds"},
	{0x5616A0937B867038ULL, "173138393.ogg"},
	{0x5616EBEAD253B601ULL, "vxdr_primevaltomb_7yq_0_0.vlm"},
	{0x5617227889567579ULL, "zz_mer_debug6180865.cl"},
	{0x561734A7E5419F10ULL, "dre000_varric6037012.cl"},
	{0x56174D0854D5405CULL, "6116707_m.fxe"},
	{0x56183661D614E694ULL, "vxlt_0_lowtown_a2_gzd_0_0.vlm"},
	{0x561864468485973FULL, "6076485_f.fxe"},
	{0x5618731A00BB1B2FULL, "6111015_m.fxe"},
	{0x5618842230F2E61CULL, "zz_fol_debug6103813.cl"},
	{0x56189FF6C2BFBA2EULL, "gen_im_trs_prf_rng_r2.uti"},
	{0x5618A78F1D297227ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_enter.ani"},
	{0x56195D8912D0ED96ULL, "vxlt_hangedman_24w_0_0.vlm"},
	{0x5619DFFF0D651787ULL, "6119187_m.fxe"},
	{0x5619FC52BB658082ULL, "6075709_m.fxe"},
	{0x561A2BE9FEA0F163ULL, "hlac_ht_0_hightown_n_-131072_0x.dds"},
	{0x561A48FDA269CAA6ULL, "gen00fl_isabela6096809.cl"},
	{0x561A4B1971DFA2B1ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boilerb.mmh"},
	{0x561A571971DFB752ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boilerb.mao"},
	{0x561A651971DFCF6FULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boilerb.msh"},
	{0x561A7296C90E79C3ULL, "338502048.ogg"},
	{0x561A876D27A190C3ULL, "6095618_m.fxe"},
	{0x561A956F532FB6F2ULL, "art\\characters\\creatures\\spider\\animation\\ev_v1.gad"},
	{0x561ADA5E7C7ECBB4ULL, "gen00fl_dog6169320.cl"},
	{0x561AF3F3BB30B336ULL, "hlac_lt_0_lowtown_n_131074_0z.dds"},
	{0x561B6EFED5B1D408ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03b.mmh"},
	{0x561B8CFED5B20732ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03b.msh"},
	{0x561B9CA28610C319ULL, "6171547_m.fxe"},
	{0x561BAAD717B378E2ULL, "vxht_0_hightown_a2_aok_0_0.vlm"},
	{0x561BAF44ADC7DA1BULL, "6149425_m.fxe"},
	{0x561BCCC976E2A33EULL, "vxlt_gallowstemplar_65u_0_0.vlm"},
	{0x561C13528B3D9829ULL, "6168435_m.fxe"},
	{0x561C22475BF3A6F4ULL, "69570557.ogg"},
	{0x561C44B7F15664DAULL, "gen00fl_merrill6121686.cl"},
	{0x561CB33ABBCC1AEEULL, "vxw_encounter2_pb_0_0.vlm"},
	{0x561CCE54163CB8D5ULL, "gen_im_trs_rdr_hvy_glv_r1.uti"},
	{0x561CD2F8B7C9D839ULL, "character_character_p71.psh.fxo.dx11"},
	{0x561D0A404849DAE7ULL, "gen00fl_isabela6097603.cl"},
	{0x561D1D34F5F8B06DULL, "6169775_m.fxe"},
	{0x561D216F662A249CULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.mmh"},
	{0x561D96D18BF11D7BULL, "gen00fl_anders6125470.cl"},
	{0x561DB14CE6FF4E83ULL, "515832724.ogg"},
	{0x561DBC3A78901BA7ULL, "llac_ht_0_hightown_n_65534_0.dds"},
	{0x561DC3EBFB7D01B7ULL, "hlac_ca_caverns1_24426_13x.dds"},
	{0x561E4169483B1948ULL, "vxw_encounter2_camp_cu_0_0.vlm"},
	{0x561E6BED33FC9FB8ULL, "one300dg_lia6064519.cl"},
	{0x561F20B89269B47BULL, "6035151_m.fxe"},
	{0x561FFC3C27C60545ULL, "vxlt_0_lowtown_docks_cwk_0_0.vlm"},
	{0x5620344BAC11812CULL, "llac_lt_0_lowtown_a3_n_c_131071_0.dds"},
	{0x56209CDFE1E147E7ULL, "vxht_0_hightown_a2_c3q_0_1.vlm"},
	{0x56211CA958599901ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_grassa_d.dds"},
	{0x56218C4B4A34724FULL, "vxlt_warehouse_n_sl_0_0.vlm"},
	{0x5621986F5335DA29ULL, "art\\characters\\creatures\\spider\\animation\\ev_v1.evt"},
	{0x5621D091B8BD8057ULL, "255104156.ogg"},
	{0x56223686E2A37BECULL, "m_dae310ar_lowtown.bnk"},
	{0x562258AF8F5371B6ULL, "6041409_m.fxe"},
	{0x56227235CBCD9E20ULL, "demoral_07.mao"},
	{0x56229910E08039D9ULL, "6177264_m.fxe"},
	{0x5622FB2397F8A3F6ULL, "6067988_m.fxe"},
	{0x5623025E73EA7280ULL, "639130161.ogg"},
	{0x56233A80A59354D9ULL, "hlac_ht_brothel_n_3381_1z.dds"},
	{0x5624475961A87004ULL, "vxw_bonepitmining_a2_1py_0_0.vlm"},
	{0x562452C2F04F6BAAULL, "hlac_lt_0_lowtown_a2_-196609_4y.dds"},
	{0x56247363B700CFA0ULL, "static_lightblend_p0.psh.fxo.dx11.assembly"},
	{0x5624F8E1880D6F30ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08a.mmh"},
	{0x562506E1880D874AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08a.msh"},
	{0x562528D61B279E50ULL, "6146885_m.fxe"},
	{0x56257BA345D9286EULL, "showabilitylist.ncs"},
	{0x5625C10222E9D7BFULL, "w_encounter3_57_0.msh"},
	{0x5625C70222E9E261ULL, "w_encounter3_57_0.mmh"},
	{0x5625CCAB03E648CCULL, "mag100_cullen6095105.cl"},
	{0x5625D30222E9F602ULL, "w_encounter3_57_0.mao"},
	{0x562624F746B15F4BULL, "w_woundedcoast_234_0.phy"},
	{0x562697ABC69008F1ULL, "vxht_chantry_n_2iv_0_0.vlm"},
	{0x56277C9EF9EF658CULL, "vxlt_gallowscourtyard_7f9_0_0.vlm"},
	{0x56278757D2C73A80ULL, "372091631.ogg"},
	{0x5627A14D5F1BA796ULL, "vxht_mansion_n_5i2_0_0.vlm"},
	{0x5627CF1229510A86ULL, "6095758_m.fxe"},
	{0x5627CF6604C70616ULL, "vxlt_hangedman_21x_0_0.vlm"},
	{0x56281742B2C0B0DDULL, "dae011_conspiracy_mob6149569.cl"},
	{0x56285249F44716DDULL, "250331207.ogg"},
	{0x562864F5C96D93C1ULL, "vxht_keep_6y9_0_0.vlm"},
	{0x5628751513C18EF8ULL, "vxca_caverns1_mines_imb_0_0.vlm"},
	{0x562884FB525AEEEAULL, "6034917_m.fxe"},
	{0x562892FF91803E58ULL, "vxlt_0_lowtown_a2_h96_0_0.vlm"},
	{0x5628B7BD3D4BD7D8ULL, "hlac_lt_gallowstemplar_2392_0x.dds"},
	{0x56294A5C6ECAAC9AULL, "6069582_m.fxe"},
	{0x56294FF9DE5F0B48ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax106a_0n.dds"},
	{0x5629ADBB137C879AULL, "6166801_m.fxe"},
	{0x5629D0F35659CCBBULL, "6027784_f.fxe"},
	{0x5629D61D81B2BED4ULL, "ht_mansion_n_3368.rml"},
	{0x5629DE61DA987562ULL, "hlac_ca_caverns1_mines_24425_10z.dds"},
	{0x562A18CCC61E93F0ULL, "6034066_f.fxe"},
	{0x562A90CCAF45F10EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v2.evt"},
	{0x562B0C4F02441689ULL, "vxht_brothel_n_5i2_0_0.vlm"},
	{0x562C4949D183BBAAULL, "ave200_donnic6111210.cl"},
	{0x562CA1C6D34A9C8DULL, "vxca_cavern1_haunted_k2r_0_2.vlm"},
	{0x562CB819914ACEE3ULL, "259536290.ogg"},
	{0x562CCA9513AB3357ULL, "w_encounter1_18_0.rml"},
	{0x562DD93FD146DCBAULL, "vxlt_0_lowtown_h4g_0_0.vlm"},
	{0x562E6EB9B1A88DAFULL, "6204254_m.fxe"},
	{0x562E9745F7CE87BDULL, "6069422_m.fxe"},
	{0x562EC7921A6E077CULL, "hlac_ca_caverns2_5802_37y.dds"},
	{0x562F2DA07A3D31CCULL, "640958185.ogg"},
	{0x562F54F9C4FB3889ULL, "llac_lt_0_lowtown_docks_n_-4_1.dds"},
	{0x562F6BD82EB3AB81ULL, "77751002.ogg"},
	{0x562F806F5341EE98ULL, "art\\characters\\creatures\\spider\\animation\\ev_v1.ani"},
	{0x562FA9F62D9BDA65ULL, "677446418.ogg"},
	{0x562FFE283C2A8ACFULL, "vxlt_warehouse_n_vo_0_0.vlm"},
	{0x563060362D090603ULL, "825042107.ogg"},
	{0x5630C05498D5A773ULL, "plt_lgt000pt_ultra_1_03.plo"},
	{0x5630DA3615244C30ULL, "6042459_m.fxe"},
	{0x56316ADCCF9ACD8EULL, "vxw_bonepit_1wy_0_1.vlm"},
	{0x563187CCAF4C00B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v2.gad"},
	{0x5631A1587DDAD468ULL, "vxw_woundedcoast_xl_0_1.vlm"},
	{0x5631AA01D51D681BULL, "mag320cr_nobleman_1.utc"},
	{0x5631F1B3C2582C5BULL, "174308015.ogg"},
	{0x56326A7F0FA3BFE9ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgtp.phy"},
	{0x5633C99513B1504FULL, "w_encounter1_18_0.phy"},
	{0x5633D746BC4ABA6DULL, "zz_prmpt_debug6200951.cl"},
	{0x5634299EB8D452CFULL, "79008144.ogg"},
	{0x56345E2B1539262CULL, "vxca_caverns2_8pj_0_0.vlm"},
	{0x56346DE3D861F3DBULL, "llac_lt_hangedman_-1_28.dds"},
	{0x56350988B8862171ULL, "vxht_chantry_n_2du_0_0.vlm"},
	{0x563584712E9E69E9ULL, "118364422.ogg"},
	{0x5635B7558474B40BULL, "hlac_lt_0_lowtown_docks_131071_4z.dds"},
	{0x5635D4EDDBE610C1ULL, "vxca_caverns1_ihe_0_0.vlm"},
	{0x5635EB3E507B7306ULL, "gen00fl_anders6119694.cl"},
	{0x5636015790764D50ULL, "581713658.ogg"},
	{0x56362FE210820B0DULL, "vxlt_gallowstemplar_f_5cp_0_0.vlm"},
	{0x5636C3B8F1514887ULL, "6113787_m.fxe"},
	{0x5636D6D2BB63AC41ULL, "vxht_keep_6b8_0_0.vlm"},
	{0x5636E6F205B7A778ULL, "vxca_caverns1_mines_ijc_0_0.vlm"},
	{0x56372C021F1A7B5DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandright.mao"},
	{0x563730021F1A8196ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandright.mmh"},
	{0x56373A021F1A9294ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandright.msh"},
	{0x5637455EB536CAE0ULL, "6075264_m.fxe"},
	{0x56379CEFA2581FBEULL, "vxht_mansion2b_3dh_0_0.vlm"},
	{0x5637B6A8436AD18AULL, "vxht_brothel_n_61f_0_0.vlm"},
	{0x5637DC5A1A42D058ULL, "hlac_lt_0_lowtown_docks_a3_0_5x.dds"},
	{0x563829B2C3E7F4B2ULL, "9926574.ogg"},
	{0x5638334C06D73028ULL, "art\\vfx\\materials\\chunky\\vfx_c_falloff_add.mao"},
	{0x56388919D4DFB569ULL, "142071806.ogg"},
	{0x563898CCAF523C07ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v2.ani"},
	{0x56389CBDA143A114ULL, "6060712_m.fxe"},
	{0x56397E2BF43A2F09ULL, "vxht_brothel_n_5l1_0_0.vlm"},
	{0x563A710423F1AAE4ULL, "6148215_f.fxe"},
	{0x563A7BB41823398AULL, "vxw_encounter2_gs_0_0.vlm"},
	{0x563AFF57A0AF1F11ULL, "zz_dae_debug6047586.cl"},
	{0x563B0B4BBF1B78BEULL, "zz_followers_banter6158133.cl"},
	{0x563B1FC31E586AF4ULL, "qun160_vashothlead6061099.cl"},
	{0x563B6CFCAF9DBF04ULL, "ser270_templar6114151.cl"},
	{0x563BF5D25288C351ULL, "6189373_m.fxe"},
	{0x563C0FE0748879E4ULL, "art\\characters\\weapons\\maces\\w_mce_mc106.mao"},
	{0x563C305621B02E96ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v4.evt"},
	{0x563C4EB525B8F126ULL, "6048499_m.fxe"},
	{0x563C6E9063F82289ULL, "71465059.ogg"},
	{0x563D59CB9F5CA704ULL, "6144731_m.fxe"},
	{0x563DBDDF52077788ULL, "6060734_m.fxe"},
	{0x563DCA0F0A5AB5EAULL, "docks_male_laborer_mob6175385.cl"},
	{0x563E96B59DD8999BULL, "vxca_caverns1_gkd_0_0.vlm"},
	{0x563EABC09222FB19ULL, "zz_dre_debug6092006.cl"},
	{0x563FFF497E9179F5ULL, "vxlt_undercityencounter_13u_0_0.vlm"},
	{0x5640014F0674069EULL, "6041830_m.fxe"},
	{0x56400AAEF8AF1160ULL, "760545935.ogg"},
	{0x564013356FD0ECE8ULL, "vxw_woundedcoast_uk_0_1.vlm"},
	{0x56403B8C01C05BEEULL, "120253713.ogg"},
	{0x56411CABE74D0DC3ULL, "6130076_m.fxe"},
	{0x56417958A35F4F72ULL, "llac_ca_caverns1_24421_2.dds"},
	{0x56419058AAECD5A6ULL, "vxw_encounter2_camp_8z_0_0.vlm"},
	{0x5641DBFA6B19117EULL, "follower_banter6170292.cl"},
	{0x5641F93A35589FE1ULL, "vxlt_0_lowtown_docks_cv4_0_0.vlm"},
	{0x564200131777490AULL, "art\\vfx\\combat\\vfx_m_ent_sleep_act_c.anb"},
	{0x564202F623B71135ULL, "588426358.ogg"},
	{0x56423BA58A7A70DDULL, "plt_cod_crt_arcane_horror.plo"},
	{0x56428B352CA7FE9AULL, "w_encounter3_44_0.rml"},
	{0x5642B5F182A52084ULL, "6169972_m.fxe"},
	{0x564325ECEF279B71ULL, "gen_im_trs_apo_sta_fir_r2.uti"},
	{0x5643275621B63E21ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v4.gad"},
	{0x56433B54DCAE5B19ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml103a_0s.dds"},
	{0x5643B7B70B91F410ULL, "and121st_karl2.stg"},
	{0x5643BBD720915B8AULL, "583004123.ogg"},
	{0x5643BE51FD6DB914ULL, "lgt110_orz_harr_carta6109045.cl"},
	{0x5643BF56282D37DAULL, "llac_ca_caverns1_24424_21.dds"},
	{0x5643EC3F5068ECF2ULL, "vxlt_gallowscourtyard_n_6yp_0_0.vlm"},
	{0x56445727C112C50FULL, "gen_im_trs_rdr_two_2hs_r0.uti"},
	{0x564489DBA89A4D66ULL, "6896414.ogg"},
	{0x56454959DFE9ADCAULL, "hlac_w_sundermount_-65533_0z.dds"},
	{0x56463CDA4804CEA4ULL, "vxlt_hangedman_245_0_0.vlm"},
	{0x56467507CF694ECCULL, "mag321_comtesse_de_launcet6043410.cl"},
	{0x56470C1E5C394763ULL, "6141234_m.fxe"},
	{0x5647608D4FFAF5ECULL, "vxca_caverns1_igy_0_0.vlm"},
	{0x56476F0BD0417EEAULL, "vxw_bonepit_1ng_0_0.vlm"},
	{0x5647E0B807FA7039ULL, "vxlt_0_lowtown_a2_n_c_hol_0_0.vlm"},
	{0x5648241BE1DB81DCULL, "101207115.ogg"},
	{0x5648310232C2153AULL, "mag321_comtesse_de_launcet.cnv"},
	{0x56485237B5EB5341ULL, "vxlt_0_lowtown_a2_n_c_eop_0_1.vlm"},
	{0x5649698DB440E9D5ULL, "ave200_donnic6110506.cl"},
	{0x564982BCF4AA61E9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge02_n.dds"},
	{0x564A0DC94C97A575ULL, "6036538_m.fxe"},
	{0x564A285621BC5ECFULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v4.ani"},
	{0x564A6BB19E964E40ULL, "6082302_f.fxe"},
	{0x564A94B6D518B59AULL, "gen00fl_anders6103756.cl"},
	{0x564AA7022308C78AULL, "w_encounter3_57_0.rml"},
	{0x564AC09217AF09A6ULL, "6043599_m.fxe"},
	{0x564AE2334C3E27F0ULL, "hlac_ht_0_hightown_a3_-65538_0z.dds"},
	{0x564B15D49D599745ULL, "714273258.ogg"},
	{0x564B2C60442CBBB6ULL, "gen00fl_fenris6126517.cl"},
	{0x564B5742B32FC33FULL, "hlac_lt_ruinedhouse_n_-1_5x.dds"},
	{0x564B81EFF6AD4BC8ULL, "vxlt_0_lowtown_docks_cv1_0_0.vlm"},
	{0x564BE3C1317316EEULL, "241907788.ogg"},
	{0x564C02A9D7D05B01ULL, "hlac_lt_gallowstemplar_2384_2x.dds"},
	{0x564C154D96EE8620ULL, "llac_ht_mansion2b_818_11.dds"},
	{0x564C530232C63782ULL, "mag321_comtesse_de_launcet.bnk"},
	{0x564C91C0DCABF5C9ULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressuredown.ani"},
	{0x564D6F962A4741F5ULL, "llac_ca_caverns2prc_5802_45.dds"},
	{0x564D72727200A2FDULL, "hlac_lt_warehouse_65537_5y.dds"},
	{0x564E7137949E2F4DULL, "llac_lt_0_lowtown_docks_a3_n_65533_3.dds"},
	{0x564EB45E7263E3FBULL, "vxht_0_hightown_a3_bgh_0_0.vlm"},
	{0x564EDEFF7E5F57D7ULL, "ts_slavers_prog_junk2.gda"},
	{0x564F1857F47CF689ULL, "6131396_m.fxe"},
	{0x564FB5DBEAE192EFULL, "hlac_w_encounter2_-131068_0y.dds"},
	{0x564FE07D8E57CE52ULL, "zz_approval_debug6003825.cl"},
	{0x565029AC598CC2C7ULL, "6076499_m.fxe"},
	{0x56503A3B8ECC4C6EULL, "937705689.ogg"},
	{0x56506211520C1320ULL, "qcr000_arishok6083986.cl"},
	{0x56506E2C90E4BA78ULL, "vxht_mansion2b_31q_0_0.vlm"},
	{0x5651D75A88257EA0ULL, "vxlt_gamlenshouse_n_b2_0_0.vlm"},
	{0x5651FDD59DED4894ULL, "vxlt_0_lowtown_a2_gnh_0_0.vlm"},
	{0x5652ACC0DCB08F4BULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressuredown.gad"},
	{0x5652EFDD4F5C3016ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v5.ani"},
	{0x5652FEB7C6012962ULL, "922027970.ogg"},
	{0x56532EEF8B37D62FULL, "14851101.ogg"},
	{0x5653D1731E1E4F32ULL, "lowtown_fem_alien_elf_2_mob6166440.cl"},
	{0x565428BCB75EDB48ULL, "319794661.ogg"},
	{0x5654BD8396C729B7ULL, "root_war_2hand.dds"},
	{0x5654CF9A32247B2DULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha13a.phy"},
	{0x56555E45B88A3513ULL, "411850318.ogg"},
	{0x56556426B4F19694ULL, "vxht_0_hightown_a3_c3w_0_1.vlm"},
	{0x565596345E2984F8ULL, "587386971.ogg"},
	{0x5655F71D81D831D4ULL, "ht_mansion_n_3368.anb"},
	{0x56561BFE5D2ACD22ULL, "hlac_ca_cavern1_haunted_24424_15z.dds"},
	{0x5656490B4EABB5E1ULL, "343291944.ogg"},
	{0x5656CE0AADE8B16EULL, "vxw_blightlands_6k_0_0.vlm"},
	{0x565777154DDE852AULL, "hlac_lt_0_lowtown_a2_n_c_1_5x.dds"},
	{0x56577E2DBA5EDD9EULL, "mcr371_orsino6083348.cl"},
	{0x5658798AF131A32CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_003a_0s.dds"},
	{0x5658B4F2597399C8ULL, "890646696.ogg"},
	{0x565905CB5C713294ULL, "mcr370_meredith6136166.cl"},
	{0x565951DB65A0B189ULL, "lgt331ar_supply_depot.are"},
	{0x5659C417B66A7A4EULL, "hlac_lt_0_lowtown_a2_n_0_2x.dds"},
	{0x5659E7112C46D22CULL, "vxlt_gallowstemplar_5ja_0_0.vlm"},
	{0x565A08F90F709559ULL, "lt_0_lowtown_a2_n_c_1_0.msh"},
	{0x565A12F90F70A647ULL, "lt_0_lowtown_a2_n_c_1_0.mmh"},
	{0x565A16F90F70AC94ULL, "lt_0_lowtown_a2_n_c_1_0.mao"},
	{0x565A44D6588F4D10ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5open.mao"},
	{0x565A48D6588F5423ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5open.mmh"},
	{0x565A56D6588F6C7DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt5open.msh"},
	{0x565B0B7549BF2623ULL, "vxht_0_hightown_n_b5x_0_0.vlm"},
	{0x565B280C24978EA6ULL, "6035907_m.fxe"},
	{0x565B631E54F055F2ULL, "vxdr_0_deeproads_f5i_0_0.vlm"},
	{0x565B6D1D9494C9F7ULL, "hm_lgt_keran_father.mor"},
	{0x565BA7D1ADAD239FULL, "hlac_lt_0_lowtown_docks_65534_3y.dds"},
	{0x565BE337DA61DEFAULL, "dae000cr_harshal.utc"},
	{0x565C7E9B68636FC6ULL, "mag210_arianni6057331.cl"},
	{0x565E12E77B0AC77FULL, "lt_foundry_n_320_ll.mmh"},
	{0x565E17DDDB73BE3AULL, "vxht_chantry_n_2lj_0_0.vlm"},
	{0x565E537F0FC8D30EULL, "art\\characters\\playercharacter\\elffemale\\glv_lgtp.mmh"},
	{0x565E58EA293457DFULL, "war_offensedmgupgrade.dds"},
	{0x565E5F7F0FC8E7A5ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgtp.mao"},
	{0x565E6D7F0FC8FFFCULL, "art\\characters\\playercharacter\\elffemale\\glv_lgtp.msh"},
	{0x565E72C55CD26C08ULL, "26694838.ogg"},
	{0x565E8FE8A156CAD1ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_6z.dds"},
	{0x565EC66DD2446FA4ULL, "vxca_caverns1_mines_fmy_0_0.vlm"},
	{0x565F730B649DEDAEULL, "s_aoe_ro_f.mao"},
	{0x565FA75498FCED10ULL, "plt_lgt000pt_ultra_1_03.nss"},
	{0x565FA9F815F5D827ULL, "da2terrain_rendertodepth_p0.psh.fxo.dx11"},
	{0x565FAF8982EDCB0BULL, "hlac_lt_0_lowtown_65536_0y.dds"},
	{0x565FC806A269EE83ULL, "310247974.ogg"},
	{0x56603D6C8ED9BA56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_l.evt"},
	{0x5660EF6F95509BBBULL, "6034630_f.fxe"},
	{0x5660F65BC8935D49ULL, "zz_qun_debug6045380.cl"},
	{0x5660F7DD4F687B07ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v5.evt"},
	{0x56624086E2D9701EULL, "m_dae310ar_lowtown.wbd"},
	{0x566281B67F23D0A4ULL, "hlac_ht_0_hightown_n_65535_0x.dds"},
	{0x5662F27B8645A124ULL, "vxht_mansion_n_5rq_0_0.vlm"},
	{0x56639AAA955B58A2ULL, "vxlt_0_lowtown_a3_n_c_e28_0_0.vlm"},
	{0x5663E8A9BEC1F96EULL, "vxw_encounter1_gs_0_1.vlm"},
	{0x5663FC6285E31438ULL, "6005722_m.fxe"},
	{0x56646E98A1A22110ULL, "vxw_sundermount_21p_0_0.vlm"},
	{0x5664822C8CE8CEF6ULL, "ht_brothel_n_3381.anb"},
	{0x5664FE9624802CB1ULL, "dae000_arishok6096370.cl"},
	{0x5664FFE188435F4BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08a.phy"},
	{0x56653E83821F5905ULL, "vxca_caverns1_mines_ijz_0_0.vlm"},
	{0x566591A1CF6EA88FULL, "gen00fl_varric6033577.cl"},
	{0x5665C1B288E9377CULL, "vxht_keep_n_79z_0_0.vlm"},
	{0x566615A41253A929ULL, "rdr110dg_dockworker6015932.cl"},
	{0x5667346C8EDFC9E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_l.gad"},
	{0x566740F40932DCC4ULL, "lt_0_lowtown_a3_n_c.anb"},
	{0x566743C3D28BA907ULL, "6154122_m.fxe"},
	{0x56675CF409330C56ULL, "lt_0_lowtown_a3_n_c.arl"},
	{0x5667F27278263624ULL, "6127677_m.fxe"},
	{0x566818DD4F6ED250ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v5.gad"},
	{0x56687DB210BD7458ULL, "dae000_hate_enemies6167608.cl"},
	{0x5668D98C5A9C10C0ULL, "llac_lt_darktown_-1_42.dds"},
	{0x5668DB8DE8ECB969ULL, "hlac_lt_gallowstemplar_n_2384_5z.dds"},
	{0x56690DFC1A14BEB1ULL, "501255593.ogg"},
	{0x566985674492AACCULL, "419287922.ogg"},
	{0x5669B0404396C8A8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolem_movement.anb"},
	{0x5669CE4EFD81DFA2ULL, "vxca_cavern1_haunted_iit_0_0.vlm"},
	{0x566A03131798A0FDULL, "art\\vfx\\combat\\vfx_m_ent_sleep_act_c.mmh"},
	{0x566A7CEF32A584F2ULL, "hlac_ca_caverns1_mines_24426_11z.dds"},
	{0x566AB949FE0D3B54ULL, "6126989_f.fxe"},
	{0x566B31EA13120821ULL, "122057497.ogg"},
	{0x566B329D6EBC44D5ULL, "680536962.ogg"},
	{0x566B6C70340B21F5ULL, "6069634_f.fxe"},
	{0x566B6D8A68ECFE02ULL, "46572329.ogg"},
	{0x566B93CB1FE89503ULL, "dra_dot.mao"},
	{0x566CBD2F7B3CE80BULL, "vxw_sundermount_2w9_0_0.vlm"},
	{0x566CC2DE237AFA57ULL, "hlac_lt_backalley_g_65537_1z.dds"},
	{0x566D356C8EE4368FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_l.ani"},
	{0x566D36DDA3B493FBULL, "vxht_keep_58g_0_0.vlm"},
	{0x566D6DBA9112D148ULL, "gen_im_trs_apo_junk2_01.uti"},
	{0x566D72251EF06BA0ULL, "gen_im_arm_cht_comb_t2.uti"},
	{0x566D93604C750AA6ULL, "106154887.ogg"},
	{0x566E1CE34A334848ULL, "bd_02_f.mao"},
	{0x566F8EFB327FE747ULL, "gen_im_trs_dwv_med_glv_r1.uti"},
	{0x566FB7C1BBA32A95ULL, "mcr370_templar26087013.cl"},
	{0x5670258DBD7932E3ULL, "vxw_bonepit_58_0_0.vlm"},
	{0x567068EEE0B0BD0CULL, "196104018.ogg"},
	{0x56708444E29D4401ULL, "6052763_m.fxe"},
	{0x5670ADD3DEB10EEAULL, "17492876.ogg"},
	{0x5670DCD76F0347C7ULL, "6064548_m.fxe"},
	{0x56715BC9329CF836ULL, "6129376_m.fxe"},
	{0x5671BE56921B2961ULL, "mag200_feynriel6023866.cl"},
	{0x567218130A0C4A47ULL, "vxht_keep_n_625_0_0.vlm"},
	{0x5672D33955D720DAULL, "plt_cod_mgc_fade.plo"},
	{0x5673678923EDAD76ULL, "6152781_m.fxe"},
	{0x567375EC060107AFULL, "6075291_f.fxe"},
	{0x567386933D4294DCULL, "lgt320_shw_ors_secret_meet6156360.cl"},
	{0x5673B06074157185ULL, "vxca_caverns1_mines_igy_0_0.vlm"},
	{0x5674BE0C465031B5ULL, "ht_0_hightown_n_26_0_ll.mmh"},
	{0x5675AA71A2BFD8EAULL, "zz_isaac_test.cut"},
	{0x5675B5FF4E6AED1CULL, "318479677.ogg"},
	{0x5675EE707E3D1940ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bod.phy"},
	{0x5676358CE924B3A9ULL, "vxlt_0_lowtown_docks_a3_cun_0_0.vlm"},
	{0x5676508C0D630F0DULL, "hf_taraneh.mor"},
	{0x5676877770C4951FULL, "ogre_grunt.utc"},
	{0x5676B8B64CBD8FCCULL, "965155678.ogg"},
	{0x5676CACB1033032CULL, "vxlt_gallowstemplar_5pg_0_0.vlm"},
	{0x56773123CFB3DD0FULL, "pro000_friedrich6051063.cl"},
	{0x567738BF0F55ACDAULL, "vxlt_undercityserialkiller_1lo_0_0.vlm"},
	{0x5677E0F6B841AED9ULL, "framebuffereffects_screentexrect_p0.psh.fxo.dx11"},
	{0x5677FAE06152D6EDULL, "gen00fl_anders6125450.cl"},
	{0x567829C5525DAA34ULL, "vxht_0_hightown_a2_c9x_0_1.vlm"},
	{0x56788FBBCE9DC029ULL, "49251899.ogg"},
	{0x5678D6BE96D8814FULL, "one100dg_lia6055138.cl"},
	{0x5678DE036DDC01A3ULL, "1047313105.ogg"},
	{0x56795D7BF76F4373ULL, "mag110_vincento6017898.cl"},
	{0x5679655A834CF5F4ULL, "gen_im_cft_reg_a1.uti"},
	{0x567A3FA5EF631430ULL, "vxlt_warehouse_n_qi_0_0.vlm"},
	{0x567A6B9C38A05D71ULL, "vxlt_warehouse_1e0_0_0.vlm"},
	{0x567AA506F3264CD4ULL, "gen00fl_aveline6139510.cl"},
	{0x567AC32FB737DE23ULL, "w_bonepit_61_0.msh"},
	{0x567AC44BFB4973C7ULL, "378294042.ogg"},
	{0x567AD12FB737F67DULL, "w_bonepit_61_0.mmh"},
	{0x567AD52FB737FC8EULL, "w_bonepit_61_0.mao"},
	{0x567AFB97BF5FEF3AULL, "vxht_chantry_n_2bd_0_0.vlm"},
	{0x567B2F0C6D33008BULL, "vxw_sundermount_2x8_0_0.vlm"},
	{0x567C5346D50BD025ULL, "6039816_m.fxe"},
	{0x567CF8A6700F4F03ULL, "hlac_lt_0_lowtown_docks_a3_-196608_0y.dds"},
	{0x567CFA65284EB5A8ULL, "mer221_sabin6080104.cl"},
	{0x567D106082F5FC54ULL, "6128315_m.fxe"},
	{0x567D2F8E7674C590ULL, "vxca_caverns2_4wz_0_1.vlm"},
	{0x567D8EF6E6F36333ULL, "vxlt_blackemporium_rw_0_1.vlm"},
	{0x567DA804A3C4479EULL, "6172534_m.fxe"},
	{0x567DCA28284773F1ULL, "791490724.ogg"},
	{0x567E2B678A0D6401ULL, "551309228.ogg"},
	{0x567E30D5A9DE0A78ULL, "hlac_ht_mansion2b_818_24z.dds"},
	{0x567E42DB65BFB4DFULL, "lgt331ar_supply_depot.ncs"},
	{0x567E9F32A1F6AEFEULL, "585986322.ogg"},
	{0x567EC5BE0D4C6C05ULL, "vxht_mansion2b_31h_0_0.vlm"},
	{0x567F2F07C8D42F5BULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.impa.ani"},
	{0x567F79FE66D70047ULL, "6171387_m.fxe"},
	{0x567FE57FE043E202ULL, "hlac_lt_0_lowtown_-196609_3z.dds"},
	{0x56808A0E466530BFULL, "vxca_caverns1_ijq_0_0.vlm"},
	{0x5680E378F41A104DULL, "gen00fl_anders6096499.cl"},
	{0x5681089D7E1D3902ULL, "mcr310_party_banter6087797.cl"},
	{0x56817751F102C4BAULL, "zz_fol_debug6062220.cl"},
	{0x5681E0315DC18B29ULL, "vxw_woundedcoast_44_0_0.vlm"},
	{0x568254135185889CULL, "vxlt_0_lowtown_a3_n_c_f5l_0_0.vlm"},
	{0x5682B734F44CDEECULL, "6124712_f.fxe"},
	{0x56831DBFC9AEEE18ULL, "hlac_dr_0_deeproads_5838_2z.dds"},
	{0x5683320638A4A0F9ULL, "gen00fl_anders6125713.cl"},
	{0x568389FBC1780803ULL, "vxw_woundedcoast_y1_0_0.vlm"},
	{0x5683A5AFE76B000EULL, "mag131_taraneh6094551.cl"},
	{0x5683A7DDDC57855BULL, "6027589_m.fxe"},
	{0x5683C6D631F9CD80ULL, "776741349.ogg"},
	{0x5683DE1B4D2337D1ULL, "6061228_m.fxe"},
	{0x5683E128EC1BAC49ULL, "6047725_m.fxe"},
	{0x56843F560EA9021BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_iron.phy"},
	{0x5684405BC7717A0CULL, "6107068_m.fxe"},
	{0x56844827828A5806ULL, "vxca_cavern1_haunted_i1j_0_2.vlm"},
	{0x568474D1C9742AE5ULL, "llac_lt_0_lowtown_a2_0_0.dds"},
	{0x5684A769DB1ACC29ULL, "vxlt_0_lowtown_docks_a3_cxm_0_0.vlm"},
	{0x5684D62C5C126724ULL, "6129877_m.fxe"},
	{0x5684FA47252D72D2ULL, "vxdr_0_deeproads_eh1_0_0.vlm"},
	{0x56855B13C2AB78DBULL, "vxca_caverns2prc_4zz_0_0.vlm"},
	{0x568562DB65C609CEULL, "lgt331ar_supply_depot.lst"},
	{0x568591133145CFB4ULL, "llac_w_woundedcoast_-65528_0.dds"},
	{0x5685A97FD7824A48ULL, "dae000_worthy6166603.cl"},
	{0x568646E5720A9DD6ULL, "6119313_m.fxe"},
	{0x568649705BADD3A1ULL, "vxlt_warehouse_11l_0_0.vlm"},
	{0x5686FBF90F976D7CULL, "lt_0_lowtown_a2_n_c_1_0.phy"},
	{0x5686FF3052F4761EULL, "126006144.ogg"},
	{0x568765564BF3764EULL, "hlac_lt_foundry_n_320_1x.dds"},
	{0x568779B898D87F73ULL, "vxlt_0_lowtown_docks_n_be4_0_1.vlm"},
	{0x5688210BE10B0CF1ULL, "vxlt_0_lowtown_a2_n_c_ew9_0_0.vlm"},
	{0x5688558863C418AEULL, "plt_fen200pt_bitterpill.plo"},
	{0x568890F651A3EF25ULL, "vxlt_0_lowtown_a3_n_c_gza_0_0.vlm"},
	{0x5688B182E1592CB0ULL, "vxlt_warehouse_n_tv_0_0.vlm"},
	{0x56894D973A8F6F3AULL, "llac_ht_0_hightown_a3_-131073_0.dds"},
	{0x56897AAA121F44F1ULL, "85772051.ogg"},
	{0x568A6B352CE57E51ULL, "w_encounter3_44_0.mmh"},
	{0x568A77352CE592F2ULL, "w_encounter3_44_0.mao"},
	{0x568A85352CE5AA0FULL, "w_encounter3_44_0.msh"},
	{0x568A97C8BCE3043FULL, "dae000_arishok6072828.cl"},
	{0x568B1316F906E000ULL, "vxlt_warehouse_n_x2_0_0.vlm"},
	{0x568B43F7A7643F3BULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.dura.evt"},
	{0x568B4F656F7D9B71ULL, "582927006.ogg"},
	{0x568B7C998DDFE8FAULL, "6139086_f.fxe"},
	{0x568BD7133C600C39ULL, "follower_banter6149226.cl"},
	{0x568BEF514055CDF5ULL, "6059223_m.fxe"},
	{0x568C0469E940A9CDULL, "1022952336.ogg"},
	{0x568C3BA44880B23FULL, "vxht_mansion2_312_0_0.vlm"},
	{0x568C9319E01244CEULL, "gen00fl_varric6033686.cl"},
	{0x568C9D253B309DE4ULL, "ht_0_hightown_a2_57_0.rml"},
	{0x568CF7CF175B8072ULL, "6108571_m.fxe"},
	{0x568D3707C8E07A52ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.impa.evt"},
	{0x568E4B2EB7B93905ULL, "vxlt_gallowstemplar_6sh_0_0.vlm"},
	{0x568E559EA424CAD4ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_wf.gad"},
	{0x568E668AD231C387ULL, "6084252_m.fxe"},
	{0x568E73481083C5DCULL, "6156761_m.fxe"},
	{0x568E9E9926E86DBCULL, "550910040.ogg"},
	{0x568EA7314335BE01ULL, "1054380349.ogg"},
	{0x568EFBEB385B493FULL, "vxca_caverns1_imp_0_0.vlm"},
	{0x568F31E0BB226EFFULL, "gen00fl_sebastian6180391.cl"},
	{0x56904632E42F9627ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.cess.gad"},
	{0x5690A66279C31D8EULL, "art\\characters\\creatures\\spider\\animation\\tal_charge.evt"},
	{0x5690A7DDB67C964BULL, "880033405.ogg"},
	{0x5690C06EF48B5A0CULL, "vxca_caverns1_k2t_0_0.vlm"},
	{0x569133557C9B9F67ULL, "vxlt_foundry_n_zm_0_0.vlm"},
	{0x56913CF7A7689F7CULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.dura.gad"},
	{0x56919FF1E5575C46ULL, "6034363_f.fxe"},
	{0x5691EDA8A7D2CF8CULL, "vxlt_0_lowtown_a3_n_c_eqs_0_0.vlm"},
	{0x5691F53EAE741423ULL, "vxdr_primevalentrance_32z_0_0.vlm"},
	{0x5691FBD8B36E2083ULL, "vxw_woundedcoast_v0_0_0.vlm"},
	{0x5692BF6797D9BF9FULL, "w_encounter2_24_0.rml"},
	{0x569368A8B9CDB35BULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.dura.evt"},
	{0x5693761C42C23978ULL, "6103965_m.fxe"},
	{0x5693899432102EE6ULL, "dae000_maren6148425.cl"},
	{0x56945607C8E6CDE5ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.impa.gad"},
	{0x5694715C62EB6F7AULL, "gen_im_npc_elf_arm_boo_dald.uti"},
	{0x569475BB05E8E3F3ULL, "mag121_idunna6084181.cl"},
	{0x5694B8C26400B656ULL, "6114413_m.fxe"},
	{0x5694BB4D4DA3EC21ULL, "vxlt_warehouse_14c_0_0.vlm"},
	{0x5694BEC4675ED4C3ULL, "vxlt_gallowstemplar_3rd_0_0.vlm"},
	{0x56953A6BECEB3320ULL, "dae010cr_joker1.utc"},
	{0x5695649EA42B02B2ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_wf.ani"},
	{0x56957662E436F6E2ULL, "vxdr_primevaltomb_763_0_0.vlm"},
	{0x56959465C0C2825CULL, "326730276.ogg"},
	{0x5696311979300D44ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowbutress.msh"},
	{0x5696431979302BADULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowbutress.mao"},
	{0x56964719793032E6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowbutress.mmh"},
	{0x569680EE523E55E6ULL, "vxlt_0_lowtown_a2_n_c_eoz_0_0.vlm"},
	{0x56972078DE0D2DB5ULL, "llac_lt_0_lowtown_docks_-2_6.dds"},
	{0x56973B32E435A234ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.cess.evt"},
	{0x5697652C8D13F68AULL, "ht_brothel_n_3381.rml"},
	{0x5697668A4CA4FABCULL, "vxlt_warehouse_n_14y_0_0.vlm"},
	{0x56979D6279C92D39ULL, "art\\characters\\creatures\\spider\\animation\\tal_charge.gad"},
	{0x5697A3A2441B60C6ULL, "vxlt_0_lowtown_n_h7i_0_0.vlm"},
	{0x5697B96C1A4FE94DULL, "vxht_mansion_n_4x3_0_0.vlm"},
	{0x5697C279742926A2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_rust_n.dds"},
	{0x5698341F34F43547ULL, "dae000_mother6127186.cl"},
	{0x56983BF7A76EBC7AULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.dura.ani"},
	{0x569857E4586168A9ULL, "hlac_lt_0_lowtown_a2_-65536_6x.dds"},
	{0x56985E89E81956B0ULL, "6153793_m.fxe"},
	{0x5698CCC37196621FULL, "vxdr_primevalentrance_22s_0_0.vlm"},
	{0x5698E25B1DFD1BF0ULL, "6110671_m.fxe"},
	{0x5699036E82AB0177ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_3y.dds"},
	{0x5699CFCA75CA98CEULL, "fxa_cen_aoe_smoke.mao"},
	{0x5699DE6797E01337ULL, "w_encounter2_24_0.phy"},
	{0x5699E42AD2BADB9BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg03a.mmh"},
	{0x569A022AD2BB0E85ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg03a.msh"},
	{0x569A61A8B9D3C59CULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.dura.gad"},
	{0x569A6B43B3C81617ULL, "gen_im_trs_gem_cub_zirc_04.uti"},
	{0x569A8D16055AE6E6ULL, "6107655_m.fxe"},
	{0x569ACB628B4DB610ULL, "976088518.ogg"},
	{0x569AE4FBAFA37EDFULL, "vxlt_0_lowtown_h24_0_0.vlm"},
	{0x569B0B707E5C66D5ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bod.msh"},
	{0x569B0D707E5C6A4BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bod.mmh"},
	{0x569B0E547F1C4763ULL, "6176548_m.fxe"},
	{0x569B59A991384B4DULL, "vxlt_gallowsprison_n_6m1_0_0.vlm"},
	{0x569C19A66BA3E39AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_meda_0n.dds"},
	{0x569C2B1D928C2FD4ULL, "vxlt_darktown_5il_0_0.vlm"},
	{0x569C2BA8926D47E3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldb.mao"},
	{0x569C2FA8926D4E30ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldb.mmh"},
	{0x569C3DA8926D664AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldb.msh"},
	{0x569C67C68FFACDDAULL, "hlac_lt_0_lowtown_docks_a3_n_65535_4x.dds"},
	{0x569CEC5E1FBB1877ULL, "124951544.ogg"},
	{0x569D0137CEC4E98FULL, "672718886.ogg"},
	{0x569D0F034958C0F0ULL, "gen00fl_anders6119308.cl"},
	{0x569D1100F66361D0ULL, "vxht_mansion2b_1mk_0_0.vlm"},
	{0x569D22B01C718C16ULL, "6074478_m.fxe"},
	{0x569D2770AB3E30E0ULL, "fen110_party_comment6095231.cl"},
	{0x569D468863D65B3DULL, "plt_fen200pt_bitterpill.nss"},
	{0x569D76345BB47EADULL, "arc_mit_li.mao"},
	{0x569DF013AA7E15D6ULL, "vxlt_hangedman_n_1t7_0_0.vlm"},
	{0x569EAE6279CF6887ULL, "art\\characters\\creatures\\spider\\animation\\tal_charge.ani"},
	{0x569F2C2C93F3DF9CULL, "6093601_f.fxe"},
	{0x569FD502B6495291ULL, "vxca_caverns1_fmq_0_0.vlm"},
	{0x569FE71BAA5A928BULL, "6139308_f.fxe"},
	{0x56A03E19793877B1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galowbutress.phy"},
	{0x56A0B5D9E939BB1EULL, "hla_ht_0_hightown_n_-2_0z.dds"},
	{0x56A0E0CA9C0105BBULL, "gen_im_trs_bst_drg_eye_04.uti"},
	{0x56A130C7D9A7B0F0ULL, "6034237_f.fxe"},
	{0x56A160A8B9D9E29AULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.dura.ani"},
	{0x56A1B3D4F71C31F9ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobb.msh"},
	{0x56A1BDD4F71C42E7ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobb.mmh"},
	{0x56A1C1D4F71C4934ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobb.mao"},
	{0x56A1D2548B1BBFCBULL, "481496481.ogg"},
	{0x56A1FF4DDCA60F7CULL, "llac_lt_0_lowtown_docks_n_-3_3.dds"},
	{0x56A2706460D2FEA9ULL, "1021545847.ogg"},
	{0x56A2A61DD5BCBDB8ULL, "vxht_0_hightown_a3_bz4_0_0.vlm"},
	{0x56A3081BB15B6EACULL, "stringtokens.gda"},
	{0x56A38C32584EFF5DULL, "gen00fl_anders6125609.cl"},
	{0x56A455138E6BF5F4ULL, "hlac_dr_primevalentrance_1726_9y.dds"},
	{0x56A48FE784E772B1ULL, "hlac_ht_keep_4519_10x.dds"},
	{0x56A4FF86E8D9546CULL, "167494081.ogg"},
	{0x56A53332E441D26DULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.cess.ani"},
	{0x56A55A95857B186EULL, "315820896.ogg"},
	{0x56A5C142AEDE1238ULL, "vxlt_gallowsprison_n_64g_0_0.vlm"},
	{0x56A5D2388FEF65DDULL, "6073546_m.fxe"},
	{0x56A6157F36117946ULL, "vxlt_0_lowtown_n_h2j_0_0.vlm"},
	{0x56A66779BE87A8ACULL, "6067674_m.fxe"},
	{0x56A724321E33C7DCULL, "6087791_m.fxe"},
	{0x56A75FF40968F524ULL, "lt_0_lowtown_a3_n_c.rml"},
	{0x56A82BBEA98D98ACULL, "vxlt_0_lowtown_a2_hi2_0_0.vlm"},
	{0x56A82FA0666BEC2FULL, "gen_im_trs_slv_hvy_hlm_r0.uti"},
	{0x56A84EEF59438BFEULL, "vxlt_0_lowtown_a2_n_hfb_0_0.vlm"},
	{0x56A85BCF4ABC6BB0ULL, "1001461328.ogg"},
	{0x56A8E1B4497339E3ULL, "976839980.ogg"},
	{0x56A91A854EF50762ULL, "794788933.ogg"},
	{0x56A965F5757AFDB5ULL, "6102772_m.fxe"},
	{0x56A9D0E354CD4B4CULL, "demo_rockwraith_mage.cut"},
	{0x56A9D16B4281D3D7ULL, "lt_0_lowtown_docks_252_0.msh"},
	{0x56A9DB88FB47B172ULL, "6106771_m.fxe"},
	{0x56A9E36B4281F23AULL, "lt_0_lowtown_docks_252_0.mao"},
	{0x56A9E76B4281F949ULL, "lt_0_lowtown_docks_252_0.mmh"},
	{0x56AA9CFA84824854ULL, "vxlt_darktown_5rm_0_0.vlm"},
	{0x56AADF767A20FEA8ULL, "6129543_m.fxe"},
	{0x56AB5E53C0B085C6ULL, "6117907_f.fxe"},
	{0x56AB63641BF15F69ULL, "isa_trail_detection.ncs"},
	{0x56AB92AA5C791469ULL, "gen00fl_varric6033544.cl"},
	{0x56ABBB9E753EFC4AULL, "vxca_cavern1_haunted_gkl_0_0.vlm"},
	{0x56AC3255D53787B5ULL, "8711046.ogg"},
	{0x56AC44729EB9D39CULL, "mag210_arianni6152618.cl"},
	{0x56AC93EE13D8AF6DULL, "274808280.ogg"},
	{0x56AD753BC90234F8ULL, "hlac_dr_slavers_3207_4x.dds"},
	{0x56AE096771AB0824ULL, "gen_im_trs_tmp_med_hlm_r2.uti"},
	{0x56AE376E120D733CULL, "mag131ip_loot_corpse_2.utp"},
	{0x56AE6F8BAF3EA6E2ULL, "vxdr_primevalentrance_2qj_0_1.vlm"},
	{0x56AE71D6BC72D1CAULL, "538539548.ogg"},
	{0x56AEF86BA9CBE058ULL, "350234980.ogg"},
	{0x56AF14DCFA2DD808ULL, "vxw_encounter2_fr_0_0.vlm"},
	{0x56AF17AFC435F2DFULL, "vxca_cavern1_haunted_ei9_0_0.vlm"},
	{0x56AF3CB13ECD65C4ULL, "vxw_bonepit_253_0_0.vlm"},
	{0x56AF81C4B9B41034ULL, "art\\characters\\creatures\\shriek\\hed_ems.mao"},
	{0x56AFA2A4CB9DC970ULL, "6031137_f.fxe"},
	{0x56AFCDB086782245ULL, "6048733_m.fxe"},
	{0x56B0C2651D3AD8F2ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\varric\\dm_cth_vara_0n.dds"},
	{0x56B0E6FF6D9316D4ULL, "follower_banter6160624.cl"},
	{0x56B0FD4D2016FC9FULL, "hlac_ht_0_hightown_a2_131073_0y.dds"},
	{0x56B118846FB2B648ULL, "143253534.ogg"},
	{0x56B126560ECFC5BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_iron.msh"},
	{0x56B134560ECFDDF3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_iron.mao"},
	{0x56B138560ECFE400ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_iron.mmh"},
	{0x56B1EC19748791D3ULL, "mag100_karras6034713.cl"},
	{0x56B25F8C23C066DFULL, "dae000_mother6121023.cl"},
	{0x56B3ED09E9522F6BULL, "vxlt_darktown_2r5_0_0.vlm"},
	{0x56B404219116FCB6ULL, "6060318_m.fxe"},
	{0x56B4727209439DB4ULL, "qun111_ketojan6039246.cl"},
	{0x56B52488D568477AULL, "llac_lt_gallowstemplar_n_2388_4.dds"},
	{0x56B57A3207C61840ULL, "vxht_chantry_31y_0_0.vlm"},
	{0x56B5A46FD35BFB2CULL, "hlac_ht_chantry_n_2534_46x.dds"},
	{0x56B67E1EBB6ACC82ULL, "vxht_bartrand_n_6py_0_0.vlm"},
	{0x56B6A630C2349536ULL, "vxdr_0_deeproads_evo_0_0.vlm"},
	{0x56B6A8541179D927ULL, "320744946.ogg"},
	{0x56B6AE04541F8BD4ULL, "6175146_m.fxe"},
	{0x56B71012C842ED80ULL, "6144524_m.fxe"},
	{0x56B76D3FBD50515CULL, "739918364.ogg"},
	{0x56B8558F1C2918C6ULL, "6084895_m.fxe"},
	{0x56B89EB1B4A491F8ULL, "llac_w_encounter2_262143_0.dds"},
	{0x56B9BD6642202284ULL, "fx_center_snow_large.dds"},
	{0x56BA1D69F94922ABULL, "709130434.ogg"},
	{0x56BA2B5587093CA8ULL, "hlac_ca_caverns1_mines_24423_11y.dds"},
	{0x56BA5E9CC02B374EULL, "w_bonepitmining_a2_239_0.msh"},
	{0x56BA6C9CC02B4E9FULL, "w_bonepitmining_a2_239_0.mao"},
	{0x56BA789CC02B633CULL, "w_bonepitmining_a2_239_0.mmh"},
	{0x56BADA49A47406DEULL, "6163972_m.fxe"},
	{0x56BB1662301C8080ULL, "vxlt_darktown_4e9_0_0.vlm"},
	{0x56BB553159CE7691ULL, "hlac_lt_0_lowtown_a2_-65537_1z.dds"},
	{0x56BB89F02CD881D9ULL, "fxa_unholy_center.dds"},
	{0x56BBD6F438ECE3B2ULL, "vxht_keep_n_585_0_0.vlm"},
	{0x56BBFCE2F7A1E809ULL, "227881935.ogg"},
	{0x56BC6ECD352377B6ULL, "gen00fl_aveline6122732.cl"},
	{0x56BC8C18CAA4F126ULL, "qun221_viscount6095605.cl"},
	{0x56BCB03DE730117DULL, "vxdr_dungeon_1t2_0_0.vlm"},
	{0x56BCFE66214DE375ULL, "vxw_bonepitdestroyed_a3_nc_0_0.vlm"},
	{0x56BD7592BAF25C53ULL, "6055136_f.fxe"},
	{0x56BD86B9EC23F088ULL, "vxw_encounter2_as_0_0.vlm"},
	{0x56BDAE8E30C37E44ULL, "vxw_bonepit_200_0_0.vlm"},
	{0x56BE91755075DFBDULL, "follower_banter6170125.cl"},
	{0x56BF89D7B9A8EEB8ULL, "vxht_0_hightown_a3_bl2_0_0.vlm"},
	{0x56BF8BB61B2748DCULL, "llac_lt_0_lowtown_65533_1.dds"},
	{0x56C0A0527B62080DULL, "6014620_m.fxe"},
	{0x56C0E1469F15457BULL, "497339837.ogg"},
	{0x56C1326E58C7FCFAULL, "mag211ar_ariannis_house.ncs"},
	{0x56C177EE0720EE9CULL, "845341362.ogg"},
	{0x56C25EE6DB4847EBULL, "vxlt_darktown_2u4_0_0.vlm"},
	{0x56C27077C9EEE050ULL, "follower_banter6164008.cl"},
	{0x56C27440F9C53A55ULL, "6128691_m.fxe"},
	{0x56C294223907A39DULL, "72708922.ogg"},
	{0x56C29F171B8C276DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8ma.mmh"},
	{0x56C2A3171B8C2DBEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8ma.mao"},
	{0x56C2B1171B8C45F3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8ma.msh"},
	{0x56C2CA281D412312ULL, "mrl150_merrill6088669.cl"},
	{0x56C2DD41B30907BEULL, "32926541.ogg"},
	{0x56C2DF66849FF995ULL, "ts_reactive_health.gda"},
	{0x56C38BFE0181455EULL, "mer221_sabin6080582.cl"},
	{0x56C39DB2109383D2ULL, "6178486_m.fxe"},
	{0x56C3A9E832BD2334ULL, "134683832.ogg"},
	{0x56C3C23180E5F3C2ULL, "dae011_gossips6149757.cl"},
	{0x56C3D5C6AF67BED2ULL, "6169448_m.fxe"},
	{0x56C3E548C64EDC12ULL, "vxlt_gallowscourtyard_4uu_0_0.vlm"},
	{0x56C4197D4386D5ADULL, "gen_im_trs_tmp_hvy_glv_r0.uti"},
	{0x56C4497F7F0DD8DEULL, "hlac_ht_chantry_2534_16z.dds"},
	{0x56C4BD35B77D8877ULL, "6067372_m.fxe"},
	{0x56C4F6C6C68898A5ULL, "vxht_0_hightown_a2_8pd_0_0.vlm"},
	{0x56C5A8DE2FF54929ULL, "6114803_m.fxe"},
	{0x56C61CF1FE002507ULL, "hlac_ca_cavern1_haunted_24422_7x.dds"},
	{0x56C652A47D4655EAULL, "vxw_woundedcoast_7r_0_0.vlm"},
	{0x56C6A283841F000DULL, "vxlt_gallowstemplar_f_51f_0_0.vlm"},
	{0x56C6B3E46A5A1396ULL, "vxlt_hangedman_2xc_0_0.vlm"},
	{0x56C6F1D4F5BA382AULL, "qun220cr_javaris6071886.cl"},
	{0x56C72147723B7B19ULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\til_ruins01blight_sa.dds"},
	{0x56C76292D450C760ULL, "vxca_caverns1_mines_i85_0_0.vlm"},
	{0x56C7787D57658378ULL, "6149875_m.fxe"},
	{0x56C780B4686E7954ULL, "vxlt_darktown_5xs_0_0.vlm"},
	{0x56C787AF0F2CBA52ULL, "vxlt_darktown_58e_0_0.vlm"},
	{0x56C7B2F3A3CF4F9AULL, "zz_dae_debug6029165.cl"},
	{0x56C7DCBC80100DD7ULL, "vxht_mansion2b_2uc_0_0.vlm"},
	{0x56C7E971EEA19972ULL, "6163565_m.fxe"},
	{0x56C82A6E58CE0DEBULL, "mag211ar_ariannis_house.lst"},
	{0x56C85014945593E7ULL, "fen210_orana6133060.cl"},
	{0x56C8BCA5E6E53BB5ULL, "943881017.ogg"},
	{0x56C8CA6466C918E6ULL, "6187835_m.fxe"},
	{0x56C8DB0C4E3A90CEULL, "art\\characters\\creatures\\highdragon\\animation\\blendtree\\highdragon.anb"},
	{0x56C9146C25D8B215ULL, "mcr371_orsino6083460.cl"},
	{0x56C989F86833556CULL, "6085783_f.fxe"},
	{0x56C9914906DC9109ULL, "dae000_orana6149054.cl"},
	{0x56CA69896205990EULL, "gen_im_trs_gen_hvy_hlm_r0.uti"},
	{0x56CB41B1B604630AULL, "vxw_blightlands_9r_0_0.vlm"},
	{0x56CB70431343FBF5ULL, "vxw_bonepitdestroyed_a3_ib_0_0.vlm"},
	{0x56CB899C8FA2802DULL, "607285557.ogg"},
	{0x56CBF896DE1A0908ULL, "vxw_encounter2_lp_0_0.vlm"},
	{0x56CC18D6A6B4F3DBULL, "485279974.ogg"},
	{0x56CC20078FF0241CULL, "hlac_dr_0_deeproads_5842_8y.dds"},
	{0x56CD2F72515CC47BULL, "503328253.ogg"},
	{0x56CD3CE1F7B0B101ULL, "42231656.ogg"},
	{0x56CDA6D4F741561CULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobb.phy"},
	{0x56CDFBB4AB9F0738ULL, "vxht_0_hightown_a3_bk3_0_0.vlm"},
	{0x56CE10487B316A03ULL, "vxca_cavern1_haunted_jtm_0_0.vlm"},
	{0x56CE2CCCD327BF0BULL, "6108673_m.fxe"},
	{0x56CE931760465706ULL, "6104625_m.fxe"},
	{0x56CEC341960FA44BULL, "gen00fl_isabela6135635.cl"},
	{0x56CF122F6D58208DULL, "6019120_m.fxe"},
	{0x56CF64E7C777753EULL, "vxlt_gallowstemplar_f_3dt_0_0.vlm"},
	{0x56CF6FA5D104ADDBULL, "515557911.ogg"},
	{0x56CF7121075D8C1EULL, "vxlt_undercityserialkiller_1c6_0_0.vlm"},
	{0x56CF854F30C1DB63ULL, "lt_0_lowtown_docks_239_0.mao"},
	{0x56CF894F30C1E1B0ULL, "lt_0_lowtown_docks_239_0.mmh"},
	{0x56CF974F30C1F9CAULL, "lt_0_lowtown_docks_239_0.msh"},
	{0x56CFB4EAB4E342CAULL, "ave121tr_preload_brennan.ncs"},
	{0x56CFCAC293C6C3EFULL, "hla_ht_0_hightown_n_-131072_0y.dds"},
	{0x56D02AE8B4BF6222ULL, "672876792.ogg"},
	{0x56D0CBCCC3B5C353ULL, "6085523_m.fxe"},
	{0x56D0D0C3CD3E606BULL, "vxlt_darktown_2x7_0_0.vlm"},
	{0x56D12CB40DB1CF4BULL, "gen_im_wep_rog_arc_act0_04.uti"},
	{0x56D1660E6B97741AULL, "llac_ht_bartrand_n_3364_1.dds"},
	{0x56D177FEBAC32A92ULL, "105947651.ogg"},
	{0x56D1A063830C745AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4ocextru.mmh"},
	{0x56D1A463830C7A91ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4ocextru.mao"},
	{0x56D1B263830C92A0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4ocextru.msh"},
	{0x56D1C48E3264A50AULL, "mag110_arianni6032476.cl"},
	{0x56D2193B06FCC799ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3_0.pwk"},
	{0x56D2385C0CB9CEDDULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirecamp.phy"},
	{0x56D2843D8441118AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_boosterprops_d.dds"},
	{0x56D340C4CEE34774ULL, "196751717.ogg"},
	{0x56D38171A8FE64CCULL, "w_woundedcoast_181_0.phy"},
	{0x56D41ABB21EB61A9ULL, "6113903_m.fxe"},
	{0x56D41D152B61B98BULL, "dae100_alistair_drunk6135121.cl"},
	{0x56D44A8CD3D7DCD2ULL, "gen00fl_bethany6101190.cl"},
	{0x56D4CE77A69A50B9ULL, "llac_w_encounter2_camp_262144_0.dds"},
	{0x56D52DA82542AEBCULL, "gen_im_acc_rng_act1_01.uti"},
	{0x56D5317C3750D623ULL, "mrl310_merrill6202138.cl"},
	{0x56D53308AF448F09ULL, "250262194.ogg"},
	{0x56D5B288154090DEULL, "vxht_mansion_n_5ab_0_0.vlm"},
	{0x56D69B26A8FDDF0AULL, "vxht_brothel_n_6hy_0_0.vlm"},
	{0x56D6C2BC0221B29DULL, "1020119843.ogg"},
	{0x56D6CFA3602C568FULL, "lgt110_arl_bann_teagan6115941.cl"},
	{0x56D6D874BE3BC3D0ULL, "vxht_mansion2b_1qw_0_0.vlm"},
	{0x56D754E3FC8D249EULL, "hlac_dr_slavers_3207_5y.dds"},
	{0x56D75F204F9DF522ULL, "gen00fl_anders6136011.cl"},
	{0x56D780BAE68487B4ULL, "mrl210_merrill6102124.cl"},
	{0x56D7BD7C65B5B686ULL, "zz_followers_banter6158685.cl"},
	{0x56D8481CC8EFDC3CULL, "684389732.ogg"},
	{0x56D9EC7A747F8290ULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_0y.dds"},
	{0x56DA39B721B327EAULL, "zz_followers_banter6157529.cl"},
	{0x56DA3BF6DFEF977DULL, "hlac_ht_mansion2_818_0y.dds"},
	{0x56DA491B94ED9DD6ULL, "6108467_m.fxe"},
	{0x56DAEF8A88B84E15ULL, "570503487.ogg"},
	{0x56DB2D03269966D5ULL, "558235485.ogg"},
	{0x56DB8ADC979BD14FULL, "narrative_varric6172531.cl"},
	{0x56DBCC6FF81C9902ULL, "6068840_m.fxe"},
	{0x56DC2CCDA5BA8EB1ULL, "6126877_m.fxe"},
	{0x56DC2CDCC96141FCULL, "vxlt_0_lowtown_a2_n_ep2_0_1.vlm"},
	{0x56DC4033C0A53837ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_1z.dds"},
	{0x56DCBCA024D5626EULL, "449610282.ogg"},
	{0x56DCDA08FC6F7346ULL, "gen_im_cns_flu_aveline_01.uti"},
	{0x56DD45137CE0765DULL, "vxht_mansion2b_2pz_0_0.vlm"},
	{0x56DDD88F54E1D09AULL, "vxlt_gallowscourtyard_n_74w_0_1.vlm"},
	{0x56DE125BD3822BADULL, "vxlt_0_lowtown_docks_bet_0_0.vlm"},
	{0x56DE34BB46AEBB49ULL, "6114496.ogg"},
	{0x56DEBAFAECEB00D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ledgedropland.evt"},
	{0x56DF519CCC94F209ULL, "l_dae160ar_wounded_coast.wbd"},
	{0x56DFF3F35734E2D8ULL, "gen00fl_merrill6106324.cl"},
	{0x56E017603C546C29ULL, "vxca_caverns1_e96_0_0.vlm"},
	{0x56E0181905AAFF45ULL, "dae000_junar6147846.cl"},
	{0x56E0A9B790F2AA70ULL, "322345193.ogg"},
	{0x56E0B367981BCC8FULL, "w_encounter2_24_0.mao"},
	{0x56E0BF67981BE12CULL, "w_encounter2_24_0.mmh"},
	{0x56E0C567981BEB1EULL, "w_encounter2_24_0.msh"},
	{0x56E0EAA965C84CCCULL, "vxht_0_hightown_a3_c3o_0_1.vlm"},
	{0x56E1052AD2F716D0ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg03a.phy"},
	{0x56E173467C132855ULL, "vxht_0_hightown_a3_c55_0_0.vlm"},
	{0x56E1FE581897F90AULL, "gen00fl_aveline6121940.cl"},
	{0x56E27FC4B844BDFEULL, "vxlt_shop_6p_0_0.vlm"},
	{0x56E32D5C0CC82F49ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirecamp.mao"},
	{0x56E3365832969341ULL, "gen00fl_fenris6113862.cl"},
	{0x56E3395C0CC843E2ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirecamp.mmh"},
	{0x56E33B5C0CC84698ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirecamp.msh"},
	{0x56E3625DEFD68B5CULL, "vxlt_warehouse_n_ui_0_0.vlm"},
	{0x56E3A99BB5A426FCULL, "1004055713.ogg"},
	{0x56E3D34C3F6A712CULL, "follower_banter6149250.cl"},
	{0x56E421FEFA424A46ULL, "6049214_m.fxe"},
	{0x56E47E32B8023159ULL, "humanfemale.fxa"},
	{0x56E4DA566B430DADULL, "vxht_keep_n_4fy_0_0.vlm"},
	{0x56E4E52ED6995170ULL, "1066243169.ogg"},
	{0x56E4F0EEF346922CULL, "vxht_0_hightown_a2_b94_0_0.vlm"},
	{0x56E50D039AF3F78AULL, "vxht_brothel_n_6uz_0_0.vlm"},
	{0x56E5453AD8CB7435ULL, "vxht_keep_n_36z_0_0.vlm"},
	{0x56E5466F94E0A5D1ULL, "173733579.ogg"},
	{0x56E54A51B031DC50ULL, "vxht_mansion2b_1tt_0_0.vlm"},
	{0x56E5528351719C31ULL, "6125581_f.fxe"},
	{0x56E553E66E09DC4AULL, "6105658_f.fxe"},
	{0x56E56C03CE6165ECULL, "vxw_encounter3_32_0_0.vlm"},
	{0x56E5B1FAECF11061ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ledgedropland.gad"},
	{0x56E61B2C42BFBA89ULL, "454947286.ogg"},
	{0x56E61D96E854A4E6ULL, "vxlt_gamlenshouse_n_3g_0_0.vlm"},
	{0x56E61F556F1B77F2ULL, "fxa_cen_snow.dds"},
	{0x56E66F5D2348FD90ULL, "vxw_sundermount_2by_0_0.vlm"},
	{0x56E677BC8A040B68ULL, "6038627_m.fxe"},
	{0x56E68220A4CB44A1ULL, "57264960.ogg"},
	{0x56E6CD54292A5048ULL, "dae000_meredith6166316.cl"},
	{0x56E740AC2185A66EULL, "vxw_blightlands_de_0_0.vlm"},
	{0x56E7FB63B3FFF1B9ULL, "25962757.ogg"},
	{0x56E832D82B2F6BAAULL, "106224012.ogg"},
	{0x56E83367BA812DAAULL, "isa211_isabela6089918.cl"},
	{0x56E85114A57D4D5EULL, "vxw_bonepitdestroyed_a3_259_0_0.vlm"},
	{0x56E867D57ECB252BULL, "vxht_keep_587_0_0.vlm"},
	{0x56E8D1902DEA0EC3ULL, "dae000_fem_priest_mob6174191.cl"},
	{0x56E8D3F0C07B1BD1ULL, "vfxchunky0_scroll.vsh.fxo"},
	{0x56E8DC50C2063A08ULL, "vxw_encounter2_rn_0_0.vlm"},
	{0x56E8E68330E7111EULL, "dae011_corff_sta6149628.cl"},
	{0x56E947ABAC32540AULL, "vxw_sundermount_26o_0_0.vlm"},
	{0x56E95D49971A0675ULL, "dre000_mother6030229.cl"},
	{0x56EA2C5C56272B5FULL, "gen00fl_sebastian6188122.cl"},
	{0x56EA3B8BB72B222FULL, "fx_dust.dds"},
	{0x56EBAA746E432BA0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_kirkwallgates_da.dds"},
	{0x56EBB2FAECF57D0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ledgedropland.ani"},
	{0x56EBC66085A9C94FULL, "vxlt_warehouse_n_9r_0_0.vlm"},
	{0x56EBCD13E1C4D598ULL, "6165136_m.fxe"},
	{0x56EBEF65B7D51C19ULL, "hlac_lt_0_lowtown_a2_n_c_-3_3z.dds"},
	{0x56ECA0944CC3828DULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyo_t.dds"},
	{0x56ED183CA02FAD96ULL, "6105792_m.fxe"},
	{0x56EDB21C8A2EE127ULL, "dae000_oldmangossip6174895.cl"},
	{0x56EDDA5A9CB332CEULL, "290720655.ogg"},
	{0x56EE90A91C5D1CB5ULL, "unlock.cur"},
	{0x56EEF139AA680C15ULL, "891066876.ogg"},
	{0x56EFD39461788156ULL, "hlac_lt_undercitydungeon_-1_14z.dds"},
	{0x56F03D955EF94FB4ULL, "6080181_m.fxe"},
	{0x56F0724CE724802DULL, "6053180_m.fxe"},
	{0x56F0AFE5D02B2547ULL, "6042261_m.fxe"},
	{0x56F135F0AD5F3158ULL, "609782325.ogg"},
	{0x56F13B8BDCA7F1DCULL, "art\\toolset\\_textures\\lightbulb_d.dds"},
	{0x56F1429AA7F212F2ULL, "vxlt_gallowscourtyard_n_65d_0_0.vlm"},
	{0x56F18552FF71609DULL, "829171922.ogg"},
	{0x56F1D43AE1CCA3DCULL, "vxlt_warehouse_n_xn_0_0.vlm"},
	{0x56F2508857F9323EULL, "dae_warrior_modal.ncs"},
	{0x56F3468EC6829E22ULL, "vfx_lightnig_dot_ani.dds"},
	{0x56F3523F697882DBULL, "vxw_bonepit_lg_0_0.vlm"},
	{0x56F356D17D500955ULL, "6153374_m.fxe"},
	{0x56F3B5C7A7FC82A8ULL, "6031369_m.fxe"},
	{0x56F3DCA80F5EE6B5ULL, "vxca_cave_1fr_0_1.vlm"},
	{0x56F3DEC8CFE450FFULL, "art\\characters\\playercharacter\\elffemale\\boo_nuda.phy"},
	{0x56F3FB4FF6C2F1CAULL, "gen00fl_aveline6131220.cl"},
	{0x56F418F5D615F4BFULL, "vxca_caverns1_i2y_0_0.vlm"},
	{0x56F43397D15458C3ULL, "art\\vfx\\materials\\vfx_fire_ground_fb_add.mao"},
	{0x56F43B6E58F36508ULL, "mag211ar_ariannis_house.are"},
	{0x56F44E73B46AEA0DULL, "gen00fl_isabela6196847.cl"},
	{0x56F44FD5A44A58CEULL, "gen00fl_aveline6113981.cl"},
	{0x56F4C05C97BA72B4ULL, "6093523_m.fxe"},
	{0x56F4D9CE1E0F4484ULL, "457098135.ogg"},
	{0x56F500C2EC4A96B1ULL, "637048997.ogg"},
	{0x56F55F66BDB8CA48ULL, "6084897_m.fxe"},
	{0x56F5A91BE05EFE4FULL, "leave_01_l3.mao"},
	{0x56F5FB152945E97EULL, "hlac_lt_0_lowtown_-65536_5x.dds"},
	{0x56F65EBB5C4E6DBEULL, "898478640.ogg"},
	{0x56F6AF77B888004BULL, "vxht_chantry_n_28s_0_0.vlm"},
	{0x56F81B1E21C588DFULL, "6053339_m.fxe"},
	{0x56F827CB4154272BULL, "vxht_0_hightown_a3_auu_0_0.vlm"},
	{0x56F90545365A038DULL, "794886582.ogg"},
	{0x56F96B0EE2515832ULL, "hlac_lt_0_lowtown_docks_n_-196610_2y.dds"},
	{0x56F9E8B1C9589DB5ULL, "6085341_m.fxe"},
	{0x56F9F0794025F920ULL, "gen00fl_isabela6106290.cl"},
	{0x56F9F2985F57FD11ULL, "6174184_m.fxe"},
	{0x56FA054BC17D03C0ULL, "dae000_paivel.cnv"},
	{0x56FA0BE01C74FC6DULL, "art\\characters\\weapons\\shields\\w_shd_005a.mmh"},
	{0x56FA0FE01C7502BEULL, "art\\characters\\weapons\\shields\\w_shd_005a.mao"},
	{0x56FA1DE01C751AF3ULL, "art\\characters\\weapons\\shields\\w_shd_005a.msh"},
	{0x56FA54F5316B7F81ULL, "floorn_normals.dds"},
	{0x56FA8B0F9C231DC4ULL, "1057392348.ogg"},
	{0x56FAE4A3FDE60693ULL, "vxht_0_hightown_a3_6pm_0_0.vlm"},
	{0x56FB0EBB81E9800BULL, "hlac_lt_gallowstemplar_2384_2z.dds"},
	{0x56FB1744FFE137AAULL, "441863831.ogg"},
	{0x56FB2711332E4668ULL, "6172792_m.fxe"},
	{0x56FC1A01155FFCF6ULL, "lt_0_lowtown_a2_11_0.rml"},
	{0x56FCE74BC17F06F8ULL, "dae000_paivel.bnk"},
	{0x56FD0A81CB24F11EULL, "625004898.ogg"},
	{0x56FD51F242D14953ULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v1.ani"},
	{0x56FE2CBBDDDB80C5ULL, "zz_one_debug6113366.cl"},
	{0x56FE479C5820380DULL, "vxlt_gallowstemplar_n_5an_0_0.vlm"},
	{0x56FE495972D6D05EULL, "vxlt_hangedman_1yc_0_0.vlm"},
	{0x56FE6E42EF0C3FBEULL, "322210936.ogg"},
	{0x56FE7F48B4A32168ULL, "llac_ht_0_hightown_a2_-65537_0.dds"},
	{0x56FEA4E483F71A03ULL, "vxca_caverns1_eb1_0_0.vlm"},
	{0x56FED89DD363BEFAULL, "6064700_f.fxe"},
	{0x56FF1B25EA66829DULL, "1045508103.ogg"},
	{0x56FF593A558376CCULL, "43335080.ogg"},
	{0x56FFA87FC3EE1DC3ULL, "vxca_caverns2_62e_0_0.vlm"},
	{0x56FFC0C5BC707361ULL, "488151321.ogg"},
	{0x56FFDC89AC7FC963ULL, "lowtown_fem_prostitute_sta6161609.cl"},
	{0x57004D01885F930AULL, "dre000_cut_sibling_dies6036347.cl"},
	{0x570057511569F80DULL, "vxca_caverns2_907_0_1.vlm"},
	{0x57005AE19CD34042ULL, "hlac_lt_0_lowtown_n_-3_0x.dds"},
	{0x57008AD89DA0C528ULL, "vxw_bonepitdestroyed_a3_e7_0_0.vlm"},
	{0x570121F2EBB42165ULL, "art\\vfx\\combat\\vfx_itm_fel_bomb_aoe_p.mmh"},
	{0x5701B1CAE0DF6FBFULL, "vxw_encounter3_62_0_0.vlm"},
	{0x5701B89D6E022B54ULL, "zz_mer_debug6030398.cl"},
	{0x57024E850154FF35ULL, "vxca_cave_1as_0_1.vlm"},
	{0x5702B657C24265D9ULL, "vxca_cavern1_haunted_eb2_0_0.vlm"},
	{0x5702C7D85B59FE6BULL, "t3_hlk_h02.tnt"},
	{0x570376175CEB1DACULL, "76910209.ogg"},
	{0x5703F8EB66862FC9ULL, "follower_banter6170569.cl"},
	{0x5704239CCCB3C0EBULL, "l_dae160ar_wounded_coast.bnk"},
	{0x570424660571D76EULL, "vxw_blightlands_rg_0_0.vlm"},
	{0x570461F27C53BCA1ULL, "6082682_f.fxe"},
	{0x570485536171343CULL, "6129276_f.fxe"},
	{0x5704A85C620B3C5EULL, "vxlt_gallowsprison_n_5sg_0_0.vlm"},
	{0x5704B0568AB08EF9ULL, "hlac_ht_chantry_n_2534_16x.dds"},
	{0x5704DFC8CFF2C504ULL, "art\\characters\\playercharacter\\elffemale\\boo_nuda.mmh"},
	{0x5704F5C8CFF2EAA6ULL, "art\\characters\\playercharacter\\elffemale\\boo_nuda.msh"},
	{0x57052154AA7E18CBULL, "vxht_chantry_n_23r_0_0.vlm"},
	{0x5705599996E4D6F8ULL, "vxlt_0_lowtown_a2_n_eqx_0_0.vlm"},
	{0x570603B501F35BC0ULL, "345277222.ogg"},
	{0x57076B09366926F4ULL, "vxlt_0_lowtown_docks_n_ctl_0_0.vlm"},
	{0x57077B30B9CA1776ULL, "qun111_ketojan6039129.cl"},
	{0x570820864CF7578EULL, "hlac_lt_0_lowtown_docks_a3_-2_4x.dds"},
	{0x57083749620D765AULL, "mag330_walter6045978.cl"},
	{0x57088F3492E1A0F9ULL, "llac_ht_chantry_2534_42.dds"},
	{0x5708EE0F4F844FE0ULL, "vxht_0_hightown_n_9wf_0_0.vlm"},
	{0x5708FC6AFE142FCDULL, "gen_im_arm_cht_med_act2_03.uti"},
	{0x570A49F242DBC58AULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v1.evt"},
	{0x570B43143FC49908ULL, "vxlt_darktown_363_0_0.vlm"},
	{0x570B8F81F661A18AULL, "zz_6115810.cut"},
	{0x570C115A983D9D80ULL, "6128806_f.fxe"},
	{0x570C58766F3C320EULL, "vxx_partypicker_3c_0_0.vlm"},
	{0x570C65D2DD01B62FULL, "follower_banter6173320.cl"},
	{0x570CBB3664CCE8DEULL, "vxlt_hangedman_1dd_0_0.vlm"},
	{0x570D441EADC1137CULL, "vxlt_gamlenshouse_n_qk_0_0.vlm"},
	{0x570DD8222074C58DULL, "lt_0_lowtown_docks_a3_n_115_0.mmh"},
	{0x570DDC222074CCDEULL, "lt_0_lowtown_docks_a3_n_115_0.mao"},
	{0x570DEA222074E413ULL, "lt_0_lowtown_docks_a3_n_115_0.msh"},
	{0x570EF66E5694BF72ULL, "320794674.ogg"},
	{0x570EFD238F00F7F2ULL, "ser270_cut_aftermath6162266.cl"},
	{0x570F1EF2EBC05932ULL, "art\\vfx\\combat\\vfx_itm_fel_bomb_aoe_p.anb"},
	{0x570F32458EA7DD3BULL, "zz_approval_debug6003967.cl"},
	{0x570F52DC3D81F151ULL, "dae000cr_grand_cleric6183967.cl"},
	{0x570F5687285BBAC7ULL, "vxlt_gallowstemplar_3rx_0_0.vlm"},
	{0x571012699EC9CE6FULL, "vxht_mansion2_3en_0_0.vlm"},
	{0x5710D3799DE64A67ULL, "240730285.ogg"},
	{0x5710E3B8283F4A5BULL, "hlac_ht_chantry_n_2534_39y.dds"},
	{0x57112834B4387E59ULL, "vxca_cavern1_haunted_eg1_0_0.vlm"},
	{0x571128F242E1ACFDULL, "art\\characters\\creatures\\desiredemon\\animation\\attk_v1.gad"},
	{0x57114B503CAC1974ULL, "497242303.ogg"},
	{0x57118323CEA4F6A8ULL, "hlac_lt_0_lowtown_a2_n_c_131070_0x.dds"},
	{0x57118376AEAB2EECULL, "vxht_keep_n_6y7_0_0.vlm"},
	{0x5712E320E77E8C82ULL, "dre000_bodahn6031309.cl"},
	{0x5712E858399FC71EULL, "348800885.ogg"},
	{0x57134AA420875D56ULL, "dae000ip_crowd_ff8_talk9.utp"},
	{0x57135871A9340297ULL, "w_woundedcoast_181_0.mmh"},
	{0x57135C71A93409A4ULL, "w_woundedcoast_181_0.mao"},
	{0x57136E71A9342809ULL, "w_woundedcoast_181_0.msh"},
	{0x57137AD2E59A0077ULL, "vxdr_primevaltomb_7lt_0_0.vlm"},
	{0x57137BB5E5E201B4ULL, "hlac_ca_cavern1_haunted_24424_12y.dds"},
	{0x57139B545F48F97AULL, "vxw_bonepit_26h_0_0.vlm"},
	{0x5713DCB1247925B3ULL, "vxdr_0_deeproads_f6a_0_0.vlm"},
	{0x5713E05A8FDEF21BULL, "vxlt_undercityencounter_1pr_0_0.vlm"},
	{0x571420C2BA6A44BDULL, "6016424_m.fxe"},
	{0x57142D45C3C69EF2ULL, "267857849.ogg"},
	{0x5714AD23B7EE84F8ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_flys.dds"},
	{0x5714D662FA4754B0ULL, "llac_w_bonepitmining_a2_458753_0.dds"},
	{0x5715DCE6285F3F74ULL, "vxlt_0_lowtown_docks_n_com_0_0.vlm"},
	{0x5715FC800F58021DULL, "493507623.ogg"},
	{0x571641A97131C2E5ULL, "415499535.ogg"},
	{0x57167E6A2D38DE77ULL, "647260051.ogg"},
	{0x5716A9397BCABE1BULL, "w_encounter3_24_0.mmh"},
	{0x5716B5397BCAD2B8ULL, "w_encounter3_24_0.mao"},
	{0x5716BDFF5226368FULL, "vxdr_0_deeproads_cqm_0_0.vlm"},
	{0x5716C7397BCAF105ULL, "w_encounter3_24_0.msh"},
	{0x57175FEC417A6860ULL, "vxht_0_hightown_n_9xg_0_0.vlm"},
	{0x57177BCB9067D70EULL, "hlac_w_blightlands_196609_0z.dds"},
	{0x5718AB3B92545072ULL, "165057570.ogg"},
	{0x5718EBFEEA79FC01ULL, "vxlt_darktown_4t7_0_0.vlm"},
	{0x5719088430EEA9A6ULL, "vxca_cavern1_haunted_k2o_0_0.vlm"},
	{0x57199A345C42F4A3ULL, "llac_lt_0_lowtown_a3_n_c_131070_2.dds"},
	{0x5719D1EE937FB313ULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.dura.ani"},
	{0x5719E40A091F11C6ULL, "989611042.ogg"},
	{0x571B4ECE099C8BFFULL, "654445841.ogg"},
	{0x571B5B12F85100D0ULL, "6083099_f.fxe"},
	{0x571B6BC32222283BULL, "vxdr_primevaltomb_7xd_0_0.vlm"},
	{0x571B889E67E34B03ULL, "vxca_caverns1_eh3_0_0.vlm"},
	{0x571BDF4DC32FEB4DULL, "bw_cyc.mao"},
	{0x571C04FEE87E422FULL, "one140cr_charade6137052.cl"},
	{0x571C2CC5D8E2EF50ULL, "dae000_tranquil_16165100.cl"},
	{0x571C49AF8879522EULL, "llac_ht_mansion2b_818_40.dds"},
	{0x571C62A32A001B94ULL, "6041418_m.fxe"},
	{0x571CCE39EA441027ULL, "hlac_lt_0_lowtown_a2_n_-2_2z.dds"},
	{0x571D745E708B672EULL, "w_encounter2_33_0.msh"},
	{0x571D825E708B7F7FULL, "w_encounter2_33_0.mao"},
	{0x571D8E5E708B931CULL, "w_encounter2_33_0.mmh"},
	{0x571D9DD5F49E5CE4ULL, "784619375.ogg"},
	{0x571E30991C209335ULL, "vxht_0_hightown_a2_9bg_0_0.vlm"},
	{0x571E5A44ECF76E04ULL, "lt_0_lowtown_docks_a3_n_47_0.rml"},
	{0x571E9DE9E4B0342EULL, "6166207_m.fxe"},
	{0x571F330922ACB7EDULL, "zz_vault_debug6178934.cl"},
	{0x571F52DA820D2B1EULL, "dre000_dougal6129727.cl"},
	{0x572034DA006AE153ULL, "zz_followers_banter6158400.cl"},
	{0x57204DC3196C5D8CULL, "hlac_w_encounter2_camp_0_0x.dds"},
	{0x5720917BC360568DULL, "vxlt_0_lowtown_docks_n_ba7_0_0.vlm"},
	{0x5720D0334E4BACF1ULL, "vxlt_hangedman_n_21x_0_0.vlm"},
	{0x5720E1E564E6974CULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef05.phy"},
	{0x57211B9629EB5945ULL, "gen00fl_anders6111109.cl"},
	{0x57217AF089782C4EULL, "hlac_lt_0_lowtown_n_-4_0y.dds"},
	{0x5721B348A3BC246AULL, "vxw_bonepitmining_a2_1ja_0_0.vlm"},
	{0x5721CEFFD864B198ULL, "gen_im_npc_arm_hum_hlm_cota.uti"},
	{0x57220D31513F11FAULL, "vxw_bonepit_21i_0_0.vlm"},
	{0x5722842B4C81DB57ULL, "6122090_f.fxe"},
	{0x5722E0A8154BEEEAULL, "vxlt_backalley_g_1y4_0_0.vlm"},
	{0x5723B9633C6014AAULL, "vxlt_0_lowtown_n_h2f_0_0.vlm"},
	{0x57249158FE6DE8EAULL, "mag360cr_apostate.utc"},
	{0x5724CB0A05694821ULL, "230955410.ogg"},
	{0x5724EACB57396198ULL, "6076931_m.fxe"},
	{0x5724F57B3F15745EULL, "527994938.ogg"},
	{0x5725282D8B9BB875ULL, "1014912083.ogg"},
	{0x572538A075F1F1ABULL, "gen00fl_sebastian6195858.cl"},
	{0x5725559B70614F1CULL, "hlac_lt_0_lowtown_a2_n_-65535_1z.dds"},
	{0x57255CD72A841726ULL, "191709689.ogg"},
	{0x5725BCE40998AA97ULL, "515609386.ogg"},
	{0x5725C67862CA564CULL, "lt_0_lowtown_docks_206_0.msh"},
	{0x5725CC7862CA603EULL, "lt_0_lowtown_docks_206_0.mmh"},
	{0x5725D12A71B39E8EULL, "6108248_f.fxe"},
	{0x5725D87862CA74D5ULL, "lt_0_lowtown_docks_206_0.mao"},
	{0x57262CCBCE307515ULL, "hlac_ca_caverns1_24424_20z.dds"},
	{0x572649EF850F918CULL, "gen00fl_dog6167245.cl"},
	{0x5726863DA44ADE69ULL, "hlac_w_bonepitdestroyed_a3_65536_0x.dds"},
	{0x5726AB9A0A28A6F1ULL, "hlac_lt_gallowscourtyard_-131072_4x.dds"},
	{0x5726E2C80AEAAB05ULL, "vxlt_warehouse_v4_0_0.vlm"},
	{0x57274E15EA0AFFFBULL, "hlac_ht_0_hightown_n_65533_0y.dds"},
	{0x572777C217B8FC0AULL, "6129224_m.fxe"},
	{0x5727B2A2595F813DULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_0z.dds"},
	{0x5727C9EE938BE34AULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.dura.evt"},
	{0x5727F5526E1FFE22ULL, "weapon1.psh.fxo"},
	{0x572849083FC34B34ULL, "6126528_m.fxe"},
	{0x5728A04DCABB906FULL, "268474255.ogg"},
	{0x572911A3562F6E73ULL, "6156191_m.fxe"},
	{0x57297339498D2961ULL, "426150201.ogg"},
	{0x5729BDC251310A9AULL, "6094808_f.fxe"},
	{0x5729CB544C3A7AA6ULL, "902126235.ogg"},
	{0x5729DDA0141840BBULL, "vxdr_primevaltomb_7ug_0_0.vlm"},
	{0x572A99FF28191808ULL, "lgt160_ntb_elf_conflict6163385.cl"},
	{0x572AC4D4BEFF2D3EULL, "gen00fl_varric6107303.cl"},
	{0x572AC7EAB5300EDBULL, "ave121tr_preload_brennan.utt"},
	{0x572AFB252C221B0EULL, "zz_warrior2.cut"},
	{0x572B68692AFB55BFULL, "vxlt_0_lowtown_a2_ep3_0_0.vlm"},
	{0x572C13311698A120ULL, "6096076_m.fxe"},
	{0x572C769E170F47F8ULL, "vxca_caverns1_mines_i7p_0_0.vlm"},
	{0x572CB7B93C345220ULL, "267694703.ogg"},
	{0x572CC63D1EFB2D7AULL, "303591520.ogg"},
	{0x572D8B2E29EA05C7ULL, "vxw_sundermount_30l_0_1.vlm"},
	{0x572DDC3151D1DBACULL, "vxht_brothel_n_4a6_0_0.vlm"},
	{0x572E5F6857A61BA8ULL, "hlac_ht_keep_3736_2z.dds"},
	{0x572EA520A525BDC1ULL, "790410866.ogg"},
	{0x572EA8EE9391C9BDULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.dura.gad"},
	{0x572F2D9C42A79CBEULL, "vxlt_0_lowtown_docks_n_b03_0_0.vlm"},
	{0x572F5B74C9FD86E5ULL, "zz_approval_debug6003709.cl"},
	{0x57305E8CC9863177ULL, "vxdr_primevaltomb_7bv_0_0.vlm"},
	{0x57309657830CB5E7ULL, "33842607.ogg"},
	{0x5730FC552FE2BBB8ULL, "vxca_caverns2prc_9ww_0_0.vlm"},
	{0x5731257C413F3B63ULL, "6089800_m.fxe"},
	{0x57318B71356B62A2ULL, "plt_zz_mrlpt_debug.nss"},
	{0x5732053456DA64DAULL, "llac_w_woundedcoast_262148_0.dds"},
	{0x573259EAEF515905ULL, "hlac_w_encounter2_196613_0y.dds"},
	{0x57338433BF148400ULL, "6169910_m.fxe"},
	{0x57339BFFFC5F43E0ULL, "ser270_cut_aftermath6162307.cl"},
	{0x5733B12B134C8885ULL, "vxlt_0_lowtown_a2_n_eqa_0_0.vlm"},
	{0x5733BFB970EE247CULL, "6125477_m.fxe"},
	{0x5733C081C3FA4A26ULL, "418939503.ogg"},
	{0x573449AC8231E865ULL, "gen00fl_fenris6127054.cl"},
	{0x5734D6189512B0A8ULL, "6082204_m.fxe"},
	{0x5734EF282AFF9EA5ULL, "827739005.ogg"},
	{0x57351701158FBCAAULL, "lt_0_lowtown_a2_11_0.anb"},
	{0x5735892795EE1C36ULL, "gen00fl_merrill6130086.cl"},
	{0x573636CFCB5AFFB3ULL, "hlac_w_woundedcoast_327688_0z.dds"},
	{0x57363C5212580F54ULL, "702895865.ogg"},
	{0x573648A9F2A71323ULL, "219623554.ogg"},
	{0x5736501196625446ULL, "vxlt_0_lowtown_e1u_0_0.vlm"},
	{0x57365695BBAC0895ULL, "mag360_thrask6043541.cl"},
	{0x57366450519827BEULL, "271571924.ogg"},
	{0x5736A8DBEC262757ULL, "one100dg_kelder6055434.cl"},
	{0x5736B9A41BC93AD1ULL, "dre180cr_profane_vampire.utc"},
	{0x573753C77F50D4F3ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_3y.dds"},
	{0x573791DCE6010AEAULL, "6068291_m.fxe"},
	{0x5737EBB974BAD612ULL, "782620515.ogg"},
	{0x5737FE9135F8E536ULL, "vxht_bartrand_n_731_0_0.vlm"},
	{0x573808CB728480AEULL, "hlac_lt_hangedman_n_-1_26z.dds"},
	{0x5738118D139A5A22ULL, "w_encounter1_36_0.msh"},
	{0x5738138D139A5DD8ULL, "w_encounter1_36_0.mmh"},
	{0x57381BB00D59CF49ULL, "root_royalarcher.dds"},
	{0x57381F8D139A727BULL, "w_encounter1_36_0.mao"},
	{0x5739126B1133BD44ULL, "vxlt_backalley_n_1rb_0_0.vlm"},
	{0x573966CC0F3B971DULL, "zz_ave_debug6207690.cl"},
	{0x573A063BEFFDDAF5ULL, "clf260_fell_orden6070473.cl"},
	{0x573A5638B948AC17ULL, "vxca_caverns2_7qy_0_0.vlm"},
	{0x573A8EEC3A4CE39FULL, "vxht_0_hightown_n_anf_0_0.vlm"},
	{0x573B101AAB1A8518ULL, "vxlt_hangedman_n_t2_0_0.vlm"},
	{0x573B4DDDC14905B7ULL, "6101960_m.fxe"},
	{0x573BDDEDFBB18177ULL, "6133781_m.fxe"},
	{0x573C3B1AB31F513FULL, "vxht_keep_n_6y8_0_0.vlm"},
	{0x573C4673B13D2571ULL, "vxht_bartrand_n_4ff_0_0.vlm"},
	{0x573C4E0E43C7F42CULL, "vxht_brothel_n_4f7_0_0.vlm"},
	{0x573C780CD0B6BCB7ULL, "6168914_m.fxe"},
	{0x573C8E5E70A6446FULL, "w_encounter2_33_0.rml"},
	{0x573CC5FED7330043ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\aveline.anb"},
	{0x573CFE67792C7849ULL, "hlac_lt_0_lowtown_a2_n_c_-1_5z.dds"},
	{0x573DE836AD000A25ULL, "hlac_lt_0_lowtown_a2_n_c_65533_0y.dds"},
	{0x573E27A5A7960F3FULL, "hlac_lt_0_lowtown_docks_131071_0z.dds"},
	{0x573E72A30569B28FULL, "vxht_chantry_2cm_0_0.vlm"},
	{0x573E80E83BCD5406ULL, "plt_isa200pt_night_terrors.nss"},
	{0x573E92F863BD88C5ULL, "vxw_bonepit_26c_0_0.vlm"},
	{0x573E949E1E41B1D1ULL, "vxht_0_hightown_a3_a2v_0_0.vlm"},
	{0x573F44849A56E179ULL, "6035126_m.fxe"},
	{0x573F95E609296B68ULL, "vrc221_hugin6065250.cl"},
	{0x57404803F4843E6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2_reverse.evt"},
	{0x57405F94F5661E35ULL, "762154682.ogg"},
	{0x5740914F852E1A5DULL, "391931858.ogg"},
	{0x57419444A45CDA2DULL, "vxw_bonepitdestroyed_a3_1lz_0_0.vlm"},
	{0x5741D15D677490A4ULL, "vxht_brothel_n_70f_0_0.vlm"},
	{0x57426F5FDF994C03ULL, "vxw_bonepit_162_0_0.vlm"},
	{0x57427EBD1CF12ABEULL, "vxca_caverns2_90a_0_0.vlm"},
	{0x5742D8B7A5C3EA34ULL, "grid.dds"},
	{0x574329901F55A87EULL, "186882476.ogg"},
	{0x57436D5E70AC2AE7ULL, "w_encounter2_33_0.phy"},
	{0x5743A8B2EEE5145EULL, "64082938.ogg"},
	{0x574423275EAF92B8ULL, "llac_ht_keep_4519_0.dds"},
	{0x5744F915F76D5483ULL, "mag311_emile_de_launcet6043679.cl"},
	{0x5745D725C58869ACULL, "vxlt_undercityserialkiller_yu_0_0.vlm"},
	{0x5745D8CF805E0979ULL, "vxdr_0_deeproads_cv2_0_0.vlm"},
	{0x57463C21ADF7DFB1ULL, "art\\vfx\\models\\vfx_vol_light_wide_3.msh"},
	{0x57463F03F4889A19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2_reverse.gad"},
	{0x5746491E7D5E95FCULL, "1069434896.ogg"},
	{0x57465621ADF80C6FULL, "art\\vfx\\models\\vfx_vol_light_wide_3.mmh"},
	{0x574670682C86BED3ULL, "montage_subs_demohack6217144.cl"},
	{0x574687B9BB8C4D34ULL, "366121287.ogg"},
	{0x5746B34D7D0F6036ULL, "pro000st_mother_intro.stg"},
	{0x5746EC1839F20600ULL, "zz_dae_debug6049958.cl"},
	{0x57476021194693DEULL, "6122532_f.fxe"},
	{0x57476DE11C8AC008ULL, "gen00fl_aveline6139653.cl"},
	{0x5747BBFC6837C4BCULL, "mag100_thrask6031749.cl"},
	{0x5747DBE33009B96EULL, "307006000.ogg"},
	{0x574866A764E9861CULL, "6095976_m.fxe"},
	{0x5748714A5E6F97ADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_sandmid.msh"},
	{0x57487F4A5E6FAFE0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_sandmid.mao"},
	{0x5748834A5E6FB633ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_sandmid.mmh"},
	{0x57488A0FF7A9BAE8ULL, "llac_w_blightlands_65536_0.dds"},
	{0x5748B5486EFDD739ULL, "lt_0_lowtown_a3_n_c_45_0.anb"},
	{0x5748C1FC400034D6ULL, "gen00fl_aveline6114283.cl"},
	{0x5748C815AB3EC497ULL, "vxca_caverns2_7vx_0_0.vlm"},
	{0x5748D2EED4EDD82AULL, "vxlt_gallowstemplar_n_432_0_0.vlm"},
	{0x57494D7A511F84F4ULL, "942371785.ogg"},
	{0x57496B7283D7D928ULL, "vxlt_0_lowtown_docks_a3_n_bfl_0_0.vlm"},
	{0x5749DF331FF0CEF0ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha01a.mmh"},
	{0x5749ED331FF0E60AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha01a.msh"},
	{0x5749F2CEDF8531CDULL, "57934560.ogg"},
	{0x574A3A5781892F3DULL, "dae000_porfiria6162643.cl"},
	{0x574B895A67713A4BULL, "vxht_mansion_n_3k3_0_0.vlm"},
	{0x574B939DC660F3AEULL, "vxht_0_hightown_n_bu1_0_0.vlm"},
	{0x574C621A60347E3FULL, "203944896.ogg"},
	{0x574C6371D8EC4996ULL, "vxw_encounter2_camp_r8_0_0.vlm"},
	{0x574C663D1F7E297BULL, "gen00fl_aveline6120816.cl"},
	{0x574CBFD1A210F7D7ULL, "gen00fl_anders6128312.cl"},
	{0x574D003312C36593ULL, "6120805_f.fxe"},
	{0x574D04D555B3A145ULL, "vxw_bonepit_21b_0_0.vlm"},
	{0x574D1AD1F7B593AAULL, "hlac_lt_0_lowtown_docks_a3_131071_4z.dds"},
	{0x574D1E5137880CDEULL, "vxw_encounter2_camp_6l_0_0.vlm"},
	{0x574D27B809F5EDADULL, "30198457.ogg"},
	{0x574D2D1CEEC48DDCULL, "vxlt_darktown_4jx_0_0.vlm"},
	{0x574D5003F48ED667ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2_reverse.ani"},
	{0x574E1C4C0746D6B6ULL, "zz_approval_debug6003601.cl"},
	{0x574E799AC17510D3ULL, "vxlt_darktown_5nj_0_0.vlm"},
	{0x574E9E54B95080F7ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic_2.mao"},
	{0x574EC435DD46E06BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackright.evt"},
	{0x574F3669720BB9E3ULL, "155816029.ogg"},
	{0x574FAAE41B918F98ULL, "33905985.ogg"},
	{0x574FABF3F6F343CEULL, "character_character_p10.psh.fxo.dx11.assembly"},
	{0x57502C4A25594AA1ULL, "209020479.ogg"},
	{0x57502EBEEE33C793ULL, "hlac_ht_0_hightown_-131072_0y.dds"},
	{0x5750470597B54752ULL, "follower_banter6173345.cl"},
	{0x5750F09A0EE7433EULL, "vxca_caverns2_95b_0_0.vlm"},
	{0x5751A168A733D8F9ULL, "82342510.ogg"},
	{0x5751B60EB2D4BC21ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railnopost.phy"},
	{0x5752934B1F8B0A1CULL, "mag320_orsino6071069.cl"},
	{0x575295E83BDE2141ULL, "plt_isa200pt_night_terrors.plo"},
	{0x5752CD0FCF375D2AULL, "lt_0_lowtown_docks_a3_189_0.mao"},
	{0x5752D10FCF376479ULL, "lt_0_lowtown_docks_a3_189_0.mmh"},
	{0x5752DB0FCF377567ULL, "lt_0_lowtown_docks_a3_189_0.msh"},
	{0x57534EB46E1BE9D2ULL, "6170059_m.fxe"},
	{0x575397DF6680C557ULL, "follower_banter6172846.cl"},
	{0x5753AF20FDC75DCCULL, "279891057.ogg"},
	{0x5753C2D02C446CFBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a.msh"},
	{0x5753D4D02C448B46ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a.mao"},
	{0x5753E0D02C449FE5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a.mmh"},
	{0x57544AAC725421F9ULL, "vxdr_0_deeproads_cs1_0_0.vlm"},
	{0x57547D603C8848F2ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_cenotaphfire_night.mmh"},
	{0x5754C40D7B74CD2EULL, "vxht_keep_6c8_0_0.vlm"},
	{0x5754DD35DD4B762CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackright.gad"},
	{0x57554DA5E9F8CB99ULL, "883363857.ogg"},
	{0x57556D599DFA3341ULL, "vxlt_0_lowtown_eot_0_0.vlm"},
	{0x57557062C2B379DBULL, "43039200.ogg"},
	{0x5755A7D41998E25DULL, "llac_ht_0_hightown_-65536_2.dds"},
	{0x5755A8305F356BC4ULL, "hlac_lt_0_lowtown_a3_n_c_-2_0x.dds"},
	{0x575692D87D33A69CULL, "vxlt_hangedman_n_10q_0_0.vlm"},
	{0x5756A8F55A9FA7BFULL, "follower_banter6163765.cl"},
	{0x575721DF07FA45BCULL, "follower_banter6160822.cl"},
	{0x57578CF2CFAC0B47ULL, "369048735.ogg"},
	{0x5757D20D7D9C1FE7ULL, "vxht_0_hightown_a3_ca3_0_0.vlm"},
	{0x5757FA5082C9AF52ULL, "vxlt_darktown_5no_0_0.vlm"},
	{0x5757FD3364BBDA3EULL, "llac_lt_backalley_g_1_1.dds"},
	{0x57583D37D7DB8DE7ULL, "769580255.ogg"},
	{0x57592A2D95295671ULL, "vxht_bartrand_n_4pd_0_0.vlm"},
	{0x575977B70CDAC671ULL, "6051807_m.fxe"},
	{0x57599BB87C9BA608ULL, "vxlt_0_lowtown_a3_n_c_h23_0_0.vlm"},
	{0x5759A35306FC9A7AULL, "zz_seb_debug6215744.cl"},
	{0x5759B2413B926D78ULL, "hlac_lt_0_lowtown_docks_a3_n_262142_0y.dds"},
	{0x575A6246608DBC54ULL, "hlac_dr_primevaltomb_658_8y.dds"},
	{0x575A949142F67C23ULL, "vxht_mansion_n_6oj_0_0.vlm"},
	{0x575AB5EE39611AE3ULL, "842723911.ogg"},
	{0x575ACE2F9376F985ULL, "vxca_caverns1_mines_i7i_0_0.vlm"},
	{0x575AD54ECAE26216ULL, "vxw_encounter2_camp_m9_0_0.vlm"},
	{0x575B7B9D3EEBC0FFULL, "llac_w_woundedcoast_131078_0.dds"},
	{0x575BAC44F0CB7ECEULL, "493045298.ogg"},
	{0x575BDC35DD51932AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackright.ani"},
	{0x575BF4F95BB88188ULL, "6137720_m.fxe"},
	{0x575C8BF47D3DD1FCULL, "vxlt_0_lowtown_a2_n_c_ehu_0_0.vlm"},
	{0x575CDCD53299F821ULL, "mag100_cullen6095151.cl"},
	{0x575D39B20C46B003ULL, "6110834_m.fxe"},
	{0x575D3F8B4A4EFD2DULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_dotsoft.mao"},
	{0x575DE102FCAA0F37ULL, "mer141_jansen6077799.cl"},
	{0x575EBA93B9E5D603ULL, "qun110_petrice6037367.cl"},
	{0x575EBC7ABFCA102CULL, "vxdr_primevaltomb_6rl_0_0.vlm"},
	{0x575ECA70F6DE87E8ULL, "mag320_orsino6071106.cl"},
	{0x575F09F6F26D548AULL, "516498995.ogg"},
	{0x575F368231C3ABC1ULL, "gen_im_arm_cht_hvy_set_med.uti"},
	{0x575F675EA515612FULL, "character_character_p68.psh.fxo.dx11"},
	{0x575F795290307ABFULL, "vxht_0_hightown_n_b87_0_0.vlm"},
	{0x575FC53349B0E32CULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tool.phy"},
	{0x576008F81BD6C694ULL, "mcr370_cut_player_rules6098293.cl"},
	{0x576013205529082DULL, "vxlt_0_lowtown_docks_b0k_0_0.vlm"},
	{0x57601EBA6658AC33ULL, "437320271.ogg"},
	{0x576078713592B325ULL, "plt_zz_mrlpt_debug.plo"},
	{0x5760A30B3464C6CEULL, "w_woundedcoast_98_0.msh"},
	{0x5760A4F2F75974A1ULL, "vxht_keep_5x4_0_1.vlm"},
	{0x5760B10B3464DE1FULL, "w_woundedcoast_98_0.mao"},
	{0x5760B8E5651C3517ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef05.mmh"},
	{0x5760BD0B3464F2BCULL, "w_woundedcoast_98_0.mmh"},
	{0x5760CEE5651C5A89ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef05.msh"},
	{0x5760DD6FE60347AAULL, "dae000_fenarel6133377.cl"},
	{0x5761007352F8AB93ULL, "vxlt_0_lowtown_a2_n_c_ekf_0_1.vlm"},
	{0x576153396B8A5475ULL, "gen_im_trs_rdr_two_2hs_r2.uti"},
	{0x5761729A7979383AULL, "vxlt_0_lowtown_a2_n_c_htb_0_0.vlm"},
	{0x5761AE470CD43A4EULL, "6063550_f.fxe"},
	{0x5761D7F81F4F477CULL, "ave200_donnic6112129.cl"},
	{0x5761DCF229E3F398ULL, "vxlt_0_lowtown_fip_0_0.vlm"},
	{0x57628B569E2F5244ULL, "6015001_f.fxe"},
	{0x5762ED79BC1DEAF0ULL, "hlac_ht_0_hightown_1_0z.dds"},
	{0x57630441456AB717ULL, "llac_w_blightlands_65538_1.dds"},
	{0x576344517C0A21B3ULL, "hlac_ca_caverns2_5802_0x.dds"},
	{0x57635B34CE5C6DE1ULL, "dre000_vampire6034363.cl"},
	{0x5763F8380FBD2D30ULL, "hlac_lt_0_lowtown_docks_-2_2y.dds"},
	{0x57643EDC75C8546AULL, "6060393_f.fxe"},
	{0x5764A9397C0CDFA8ULL, "w_encounter3_24_0.rml"},
	{0x5764D502B5042564ULL, "vxht_0_hightown_a2_cf4_0_0.vlm"},
	{0x5765E022D6510846ULL, "modifyproperty.ncs"},
	{0x5766419DFD30F91DULL, "vxw_encounter1_bf_0_0.vlm"},
	{0x5766AA87BBF176D8ULL, "art\\vfx\\combat\\vfx_gib_ogr_p_1_1.mao"},
	{0x5766E872E565C1BFULL, "mag121_idunna6083939.cl"},
	{0x576776D222DE26D8ULL, "qun221_petrice6082680.cl"},
	{0x5767AB68625BAD66ULL, "lt_0_lowtown_a3_n_c_45_0_lc.mmh"},
	{0x5767BA3DD8EAC4AFULL, "art\\characters\\creatures\\corpse\\bdy_cor.mao"},
	{0x5768020F4C148171ULL, "22707263.ogg"},
	{0x576811C6CE14D590ULL, "6076953_m.fxe"},
	{0x57681CA94FB26C6EULL, "vxdr_primevalentrance_21f_0_1.vlm"},
	{0x57681FD9A60F263FULL, "hlac_lt_0_lowtown_65536_4y.dds"},
	{0x576852F6D4E95C83ULL, "6149902_m.fxe"},
	{0x57687D1EA92CB0CFULL, "897823390.ogg"},
	{0x5768A69BAD535E25ULL, "mag371_elsa6041741.cl"},
	{0x5769135047FF00ABULL, "686498932.ogg"},
	{0x57699DF765E012FEULL, "vxlt_gallowscourtyard_7er_0_0.vlm"},
	{0x576A08EBEF822634ULL, "vxlt_0_lowtown_docks_a3_n_d66_0_0.vlm"},
	{0x576A4F39595BCF58ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02barge_1.mao"},
	{0x576AE32DC7C285E6ULL, "hlac_w_woundedcoast_-196608_0x.dds"},
	{0x576B27DCC1E6CFC9ULL, "one200_gold_followers6120140.cl"},
	{0x576B61BC1660B43FULL, "6139844_m.fxe"},
	{0x576B93D5044773A0ULL, "llac_dr_0_deeproads_5842_12.dds"},
	{0x576BA057AC0428D7ULL, "plt_ave200pt_mother_dead.nss"},
	{0x576BC062EFFC4EDFULL, "zz_vault_debug6203294.cl"},
	{0x576BF01AC6FDA84EULL, "hlac_dr_0_deeproads_5842_3z.dds"},
	{0x576C0760E83F5C30ULL, "6133373_m.fxe"},
	{0x576C5844856B4BE7ULL, "mcr371_hostile_merrill6082164.cl"},
	{0x576C941E2582CED7ULL, "gen00fl_isabela6100064.cl"},
	{0x576CAAE07169FE9AULL, "207232602.ogg"},
	{0x576CADD8DE25250EULL, "art\\characters\\weapons\\greataxes\\w_gax_ds001a.phy"},
	{0x576CE82139534E12ULL, "vxw_bonepitdestroyed_a3_2aq_0_0.vlm"},
	{0x576D28C151E7F934ULL, "736628981.ogg"},
	{0x576D2E57B1C028ACULL, "vxdr_primevaltomb_6ym_0_0.vlm"},
	{0x576D3FDE881D0BE3ULL, "6112664_f.fxe"},
	{0x576DB3AAD68AB0D8ULL, "art\\characters\\playercharacter\\elfmale\\em_glv_lgte.phy"},
	{0x576E195E5ED6C004ULL, "6089502_m.fxe"},
	{0x576E6A872E9CE058ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12_0.pwk"},
	{0x576E7E36C94F9A9EULL, "gen_im_arm_cht_comb_t4.uti"},
	{0x576E84FD471F20ADULL, "vxlt_0_lowtown_docks_b7j_0_0.vlm"},
	{0x576EED3B65104F03ULL, "696401159.ogg"},
	{0x576FC78085BF42C0ULL, "599818947.ogg"},
	{0x576FD77F95BBBFFFULL, "vxdr_0_deeproads_fit_0_0.vlm"},
	{0x57707B7B5619E74DULL, "vxht_0_hightown_a2_c3s_0_1.vlm"},
	{0x577180464C27CC80ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunklow_dp.dds"},
	{0x577185C4B12956EEULL, "blood_conj_06.mao"},
	{0x57720C8F8798CB88ULL, "hlac_ht_mansion2b_818_25y.dds"},
	{0x5772BBC17492D620ULL, "6116058_f.fxe"},
	{0x5772CC21688A26DEULL, "vxca_caverns1_j5n_0_0.vlm"},
	{0x5773477E010B176BULL, "6087210_f.fxe"},
	{0x57735646FD5DC904ULL, "514864361.ogg"},
	{0x577369A231C2638EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1e.mmh"},
	{0x577375A231C27825ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1e.mao"},
	{0x577383A231C2907CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1e.msh"},
	{0x57746A443A9A6425ULL, "dae200_saemus6142366.cl"},
	{0x5774710513E28157ULL, "55436468.ogg"},
	{0x5774B37AEF27119DULL, "vxw_encounter1_ge_0_0.vlm"},
	{0x5774C3852CBE2713ULL, "vxdr_0_deeproads_d0u_0_0.vlm"},
	{0x5774C7C94F72D411ULL, "mrl211ip_woodenhalla.utp"},
	{0x5774DE0A66B5E052ULL, "vxlt_darktown_5ti_0_0.vlm"},
	{0x5775084E42F03186ULL, "vxlt_gallowscourtyard_7fc_0_0.vlm"},
	{0x5775708628FEF110ULL, "6131898_m.fxe"},
	{0x5775FCB932716814ULL, "576198325.ogg"},
	{0x5776554F44BD78CBULL, "dae000_bartrand6064037.cl"},
	{0x5777176C09BA2B24ULL, "llac_w_sundermount_0_1.dds"},
	{0x5777220AF6853AF1ULL, "vxdr_primevalentrance_2ol_0_2.vlm"},
	{0x5777EDF7CF79399BULL, "llac_lt_ruinedhouse_n_-1_1.dds"},
	{0x5777F6E69EFB27FDULL, "vxht_0_hightown_a2_bu5_0_0.vlm"},
	{0x57788A89249AE7C8ULL, "vxw_bonepit_226_0_0.vlm"},
	{0x5778F0929C116120ULL, "qun221_viscount6069752.cl"},
	{0x577906985DBECD61ULL, "ico_dagger_0.dds"},
	{0x577913D51C77DF23ULL, "dae000_surprises6157090.cl"},
	{0x577942C30E202AA4ULL, "vxw_encounter2_camp_rz_0_0.vlm"},
	{0x57794AB49ACBAFECULL, "157926961.ogg"},
	{0x577986D3543FB0E7ULL, "vxlt_gallowstemplar_n_5ja_0_0.vlm"},
	{0x5779BA4D1FCF490AULL, "vxw_sundermount_2da_0_0.vlm"},
	{0x577A054211788E8CULL, "vxw_woundedcoast_g4_0_0.vlm"},
	{0x577AC5622C0200D7ULL, "gen00fl_aveline6033288.cl"},
	{0x577AEDA32050C33DULL, "6116370_f.fxe"},
	{0x577C69FDA89102F3ULL, "499587744.ogg"},
	{0x577CB0564EB10DA8ULL, "477583479.ogg"},
	{0x577CFEF5F2F44E9DULL, "664138359.ogg"},
	{0x577D236FB2F42419ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_2x.dds"},
	{0x577D5D0D76E8567EULL, "6093613_m.fxe"},
	{0x577D8168342CF6E8ULL, "vxca_caverns1_mines_io8_0_0.vlm"},
	{0x577DA17CCEBDB1B0ULL, "gen00fl_varric6109206.cl"},
	{0x577E0C91A39C1A49ULL, "6136182_m.fxe"},
	{0x577E39CF5FDDE5A2ULL, "gen00fl_carver6097318.cl"},
	{0x577EFC7DBF10AE83ULL, "vxlt_backalley_g_1jt_0_0.vlm"},
	{0x577F510405C3FC4FULL, "221704692.ogg"},
	{0x577F993799EAD15CULL, "793136179.ogg"},
	{0x57803E5B119DB2A1ULL, "851938249.ogg"},
	{0x5780A5A9059DF683ULL, "vxca_caverns1_e58_0_0.vlm"},
	{0x5780F7B28688E689ULL, "mag320_orsino6071181.cl"},
	{0x5780F7D7224A8D91ULL, "294506302.ogg"},
	{0x57816B213EBD2B37ULL, "follower_barks6153596.cl"},
	{0x5782406CF1CFA1A4ULL, "vxca_caverns2prc_8bm_0_0.vlm"},
	{0x57829FDC9D8910B3ULL, "6169676_m.fxe"},
	{0x5782B6486F2F5139ULL, "lt_0_lowtown_a3_n_c_45_0.rml"},
	{0x5782C04F30A86AC1ULL, "6099476_m.fxe"},
	{0x578334EA2F153D7DULL, "270865919.ogg"},
	{0x578370CEDAAB06EAULL, "6048297_m.fxe"},
	{0x57839C252AC65FB5ULL, "vxlt_0_lowtown_a3_n_c_ep1_0_0.vlm"},
	{0x5783C95032278CFEULL, "vxlt_0_lowtown_docks_bci_0_0.vlm"},
	{0x5783EAD1581D3B2EULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.cess.ani"},
	{0x5784110C6D4C0D04ULL, "art\\characters\\creatures\\corpse\\textures\\cn_bdy_cor_0d.dds"},
	{0x57848B0B280FC0BFULL, "6109739_m.fxe"},
	{0x5784E7F5468B5553ULL, "vxlt_gallowscourtyard_n_761_0_0.vlm"},
	{0x5785046D036811ACULL, "817365534.ogg"},
	{0x57857AE992B34D7CULL, "dae000_junar6147935.cl"},
	{0x5785889037288BEFULL, "vxw_encounter3_bn_0_0.vlm"},
	{0x57859C0B3483D687ULL, "w_woundedcoast_98_0.phy"},
	{0x5785ACA41C0CF89FULL, "dre180cr_profane_vampire.ncs"},
	{0x57861088AB57D7D0ULL, "zz_ave_debug6207424.cl"},
	{0x57863F121B6FA7DAULL, "214747344.ogg"},
	{0x5786BC5E097E4825ULL, "6140907_m.fxe"},
	{0x57873C39CB9A1744ULL, "lgt311ar_rand_sewer_1_nt.are"},
	{0x5787E883B8E30A5BULL, "6077809_m.fxe"},
	{0x57882C2A11C5618AULL, "vxw_sundermount_2ib_0_0.vlm"},
	{0x57889AB0B259740DULL, "466493470.ogg"},
	{0x5788EB1ACC2B8D30ULL, "6139573_m.fxe"},
	{0x578944D0D52ADE62ULL, "hlac_ca_cave_2194_13y.dds"},
	{0x578955C006A60C2AULL, "619153283.ogg"},
	{0x578AE6332028710BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha01a.phy"},
	{0x578AE6D9A6281837ULL, "hlac_w_bonepit_6_0y.dds"},
	{0x578B23D7B273D4F7ULL, "6132723_f.fxe"},
	{0x578B99CBB4D72BC5ULL, "6095605_m.fxe"},
	{0x578D2F11924A52FBULL, "vxlt_0_lowtown_ent_0_1.vlm"},
	{0x578D4321009E81DBULL, "vxlt_0_lowtown_a3_n_c_ei8_0_0.vlm"},
	{0x578D62278CA61E7CULL, "dae000_arishok6075208.cl"},
	{0x578D6E5AB106C703ULL, "vxlt_backalley_g_1ww_0_0.vlm"},
	{0x578D8FCAE1989EB0ULL, "ht_0_hightown_a3_52_0_ll.mmh"},
	{0x578DC37E765EA114ULL, "6076782_m.fxe"},
	{0x578DF9B8CD732672ULL, "character_character_p148.psh.fxo.dx11.assembly"},
	{0x578E3DB1422B4AF1ULL, "w_bonepit_230_0.msh"},
	{0x578E4BB1422B620CULL, "w_bonepit_230_0.mao"},
	{0x578E57B1422B76AFULL, "w_bonepit_230_0.mmh"},
	{0x578E8E335FBE6404ULL, "llac_lt_0_lowtown_n_1_5.dds"},
	{0x578EB1ABB007E142ULL, "432501496.ogg"},
	{0x578EB4827D18044FULL, "vxht_0_hightown_a3_65f_0_0.vlm"},
	{0x578EB4BBB79C4F38ULL, "hlac_lt_0_lowtown_docks_n_-65540_0z.dds"},
	{0x579047A13FBE57B2ULL, "dae000_gamlen6121551.cl"},
	{0x57907F1BBBDDBF17ULL, "dae000ip_crowd_ff1_talk9.utp"},
	{0x579088508C1D992CULL, "mrl000_marethari6129311.cl"},
	{0x5790B15DE4E7806EULL, "920326570.ogg"},
	{0x579104B12BD29AF6ULL, "llac_lt_0_lowtown_n_-65538_0.dds"},
	{0x57914C70FF8167E1ULL, "mcr310_sibling6076395.cl"},
	{0x579170DB8AA5A62DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_xenon.msh"},
	{0x57917EDB8AA5BE60ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_xenon.mao"},
	{0x579182DB8AA5C4B3ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_xenon.mmh"},
	{0x5791B7E3015706F5ULL, "zz_dae_debug6206727.cl"},
	{0x5791BB0D5AEB198DULL, "w_bonepit_135_0.mmh"},
	{0x5791BF0D5AEB20DEULL, "w_bonepit_135_0.mao"},
	{0x5791CD0D5AEB3813ULL, "w_bonepit_135_0.msh"},
	{0x5791E2D158296B7FULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.cess.evt"},
	{0x5791E6C358989D5EULL, "ser221_gascard6113808.cl"},
	{0x57923B2D241DA57EULL, "vxlt_0_lowtown_docks_bfj_0_0.vlm"},
	{0x57924666CC0267B7ULL, "vxht_0_hightown_a3_b8m_0_0.vlm"},
	{0x57925769648CBA97ULL, "hlac_lt_0_lowtown_docks_a3_-65537_4y.dds"},
	{0x5792948DF29151C1ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masa.phy"},
	{0x5792A260493E824AULL, "862540874.ogg"},
	{0x57933A4E722EC9B6ULL, "zz_followers_banter6158748.cl"},
	{0x57939BDEE625FFC0ULL, "follower_banter6161102.cl"},
	{0x5793BF14C24D8E91ULL, "hlac_lt_undercitydungeon_-1_43y.dds"},
	{0x5793CCEE802F6E2EULL, "vxht_0_hightown_n_bl6_0_0.vlm"},
	{0x5793EAC9E064A4F5ULL, "vxht_mansion2b_1mj_0_0.vlm"},
	{0x5793EC7C8140E9EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforwardright.ani"},
	{0x5793F342F36CC59DULL, "6208379_m.fxe"},
	{0x57941C0D99CE8926ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.anb"},
	{0x5794A0C2FFFD8EBDULL, "vxlt_gallowscourtyard_n_6yk_0_0.vlm"},
	{0x5794D70F3E74CB73ULL, "isa211_qunari_ambush6081810.cl"},
	{0x5794DAA082E758FDULL, "vxht_0_hightown_a2_bc3_0_0.vlm"},
	{0x5794FE6721D564B9ULL, "hlac_ht_0_hightown_-131072_1x.dds"},
	{0x5795143CAB3A2256ULL, "862603100.ogg"},
	{0x57951810717DD9FBULL, "vxht_brothel_n_5hg_0_0.vlm"},
	{0x579623B5E0ACB9DCULL, "6099486_m.fxe"},
	{0x5796267CF20C5BA4ULL, "vxw_encounter2_camp_hx_0_0.vlm"},
	{0x579632AC198F6008ULL, "vxw_encounter2_6b_0_0.vlm"},
	{0x57969E0703BB7A0AULL, "vxw_sundermount_2rc_0_0.vlm"},
	{0x5797A300D1439C1FULL, "zz_followers_banter6159735.cl"},
	{0x57983C53556653BAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1a.phy"},
	{0x5798C3D1582F5568ULL, "art\\vfx\\environment\\vfx_trap_floor_spike_p.cess.gad"},
	{0x5798C52EC190897CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medp_0t.dds"},
	{0x57995FDC7ECE4994ULL, "vxlt_gallowstemplar_n_6sz_0_0.vlm"},
	{0x5799F299B55E3D64ULL, "vxca_cavern1_haunted_flz_0_0.vlm"},
	{0x579ABA3CE6AD341AULL, "6054018_m.fxe"},
	{0x579ACA6249152E62ULL, "hlac_lt_0_lowtown_docks_-65539_1x.dds"},
	{0x579ACD72D6CB4853ULL, "6061727_m.fxe"},
	{0x579ADB9BA3533EB0ULL, "gen00fl_varric6021085.cl"},
	{0x579B459F9797C7A6ULL, "al_varric_talents.gda"},
	{0x579B4C4534C5E80DULL, "341332717.ogg"},
	{0x579B97FA4D94D646ULL, "vxht_mansion_n_3of_0_0.vlm"},
	{0x579BB9D02C821292ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a.phy"},
	{0x579C3AACEEE233E5ULL, "406714649.ogg"},
	{0x579C4292062CBC5BULL, "vxht_keep_n_3im_0_0.vlm"},
	{0x579C4B9E82F3C913ULL, "hlac_lt_gallowstemplar_2391_9y.dds"},
	{0x579CBA82DF88A9DBULL, "fx_impact011.dds"},
	{0x579CCE4C74E41B1FULL, "vxdr_primevalentrance_2lm_0_0.vlm"},
	{0x579CD79F8C50159FULL, "965763638.ogg"},
	{0x579CF0C25D2D8478ULL, "920711673.ogg"},
	{0x579D334EA7D34675ULL, "sse_ss_carver.gda"},
	{0x579D4C8347168ACDULL, "alphagradient01.mao"},
	{0x579D8C3E2C1E66AFULL, "443025971.ogg"},
	{0x579DAE4C96D9AB09ULL, "20338924.ogg"},
	{0x579DC109F30BBA08ULL, "w_encounter3_203_0.mmh"},
	{0x579DCD09F30BCEABULL, "w_encounter3_203_0.mao"},
	{0x579DD38055319EA2ULL, "dae321st_alistair_king.stg"},
	{0x579DDF09F30BED32ULL, "w_encounter3_203_0.msh"},
	{0x579DE2A1938DF3D6ULL, "and100_anders6095826.cl"},
	{0x579E501472B6093FULL, "46572773.ogg"},
	{0x579E98248D516B94ULL, "hlac_w_sundermount_131070_0y.dds"},
	{0x579FA0EACB9AEFD5ULL, "378493292.ogg"},
	{0x579FCCA0A779693DULL, "gen00fl_dog6117708.cl"},
	{0x57A0183D9C4B62A7ULL, "6111987_m.fxe"},
	{0x57A07B31B411A1FEULL, "vxca_cavern1_haunted_i9x_0_0.vlm"},
	{0x57A09BF83BCFFFADULL, "300994581.ogg"},
	{0x57A09C3349E834F7ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tool.mmh"},
	{0x57A0A03349E83B04ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tool.mao"},
	{0x57A0B23349E85A69ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tool.msh"},
	{0x57A0B843BDF88037ULL, "vxht_0_hightown_a3_b7l_0_0.vlm"},
	{0x57A10593C038C530ULL, "dae000_mother6099181.cl"},
	{0x57A108871B244A9DULL, "247857152.ogg"},
	{0x57A114011B2F89A9ULL, "hlac_lt_gallowstemplar_n_7198_7z.dds"},
	{0x57A14B620A410899ULL, "6152113_m.fxe"},
	{0x57A173BF66DCFB0EULL, "6077259_m.fxe"},
	{0x57A1D0437694565CULL, "800251178.ogg"},
	{0x57A200C047D1C848ULL, "ico_helm_light_2.dds"},
	{0x57A21F0D99DACBD1ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.mmh"},
	{0x57A2B40D5AF9814AULL, "w_bonepit_135_0.phy"},
	{0x57A2FF36AC0D5D40ULL, "hlac_w_bonepitdestroyed_a3_262144_0y.dds"},
	{0x57A3331A64BF8478ULL, "qun211_braggart6075715.cl"},
	{0x57A3BD2D828C9045ULL, "6135367_m.fxe"},
	{0x57A41B35521B646AULL, "zz_fol_debug6096551.cl"},
	{0x57A49859E4027424ULL, "vxw_encounter2_camp_cy_0_0.vlm"},
	{0x57A4A4890B857888ULL, "vxw_encounter2_1c_0_0.vlm"},
	{0x57A4B45D94BA2635ULL, "90743354.ogg"},
	{0x57A54339CBB320E6ULL, "lgt311ar_rand_sewer_1_nt.ncs"},
	{0x57A5BE07624EED30ULL, "gen00fl_varric6078685.cl"},
	{0x57A5C0154497D62BULL, "405133797.ogg"},
	{0x57A64831E3FBEEC0ULL, "6163178_m.fxe"},
	{0x57A73959EB46AB0CULL, "w_bonepitmining_a2_17_0_lc.mmh"},
	{0x57A7C57C81515028ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforwardright.gad"},
	{0x57A7E912EAF183F3ULL, "plt_one100pt_loose_end_meer.plo"},
	{0x57A81C60D94CC0D7ULL, "198231670.ogg"},
	{0x57A8655C1A450C57ULL, "spill_alpha.mao"},
	{0x57A988A41EA134EFULL, "vxht_0_hightown_n_bxg_0_0.vlm"},
	{0x57A9F703E29AC193ULL, "vxca_caverns1_mines_hm8_0_0.vlm"},
	{0x57AA0844821584CFULL, "vxht_mansion2_1al_0_0.vlm"},
	{0x57AAE6F607C44CEFULL, "hm_amb_pickpocketee.mor"},
	{0x57AAF409030DB96DULL, "gen00fl_sebastian6179788.cl"},
	{0x57AB175CD5A047C3ULL, "6070790_m.fxe"},
	{0x57AC17A20484F7BDULL, "vxdr_0_deeproads_f5r_0_0.vlm"},
	{0x57AC2F5E786A9183ULL, "hlac_lt_0_lowtown_docks_n_131070_1y.dds"},
	{0x57AC3B39CBB931DFULL, "lgt311ar_rand_sewer_1_nt.lst"},
	{0x57AC74CBE3E4ABAEULL, "w_bonepitdestroyed_a3_244_0.mmh"},
	{0x57AC7FF21A4CF017ULL, "dre000_varric6036982.cl"},
	{0x57AC80CBE3E4C045ULL, "w_bonepitdestroyed_a3_244_0.mao"},
	{0x57AC8ECBE3E4D79CULL, "w_bonepitdestroyed_a3_244_0.msh"},
	{0x57ACA99D492D57D7ULL, "6072253_m.fxe"},
	{0x57AD3C6C64220097ULL, "756661055.ogg"},
	{0x57AD6D99523DFDF0ULL, "pro000_merchant6051613.cl"},
	{0x57ADACD8DE5CBAD1ULL, "art\\characters\\weapons\\greataxes\\w_gax_ds001a.mmh"},
	{0x57ADB8D8DE5CCF72ULL, "art\\characters\\weapons\\greataxes\\w_gax_ds001a.mao"},
	{0x57ADC6D8DE5CE68FULL, "art\\characters\\weapons\\greataxes\\w_gax_ds001a.msh"},
	{0x57ADD2F36EAF950FULL, "mcr370_assignments6176734.cl"},
	{0x57AEA7CA78740801ULL, "and100_anders6096004.cl"},
	{0x57AFC301476B45E1ULL, "lgt120_orz_bln_official6140179.cl"},
	{0x57B00CF401D1CA10ULL, "llac_w_encounter3_-65535_0.dds"},
	{0x57B08758D4D91096ULL, "w_bonepitdestroyed_a3_254_0.mao"},
	{0x57B09358D4D92535ULL, "w_bonepitdestroyed_a3_254_0.mmh"},
	{0x57B09558D4D928ABULL, "w_bonepitdestroyed_a3_254_0.msh"},
	{0x57B1F874BEE34B85ULL, "6114490_m.fxe"},
	{0x57B22F0A7482A8C5ULL, "6130067_m.fxe"},
	{0x57B2BD5CB339D048ULL, "mcr370_cut_flight_gallows.cnv"},
	{0x57B2D6E74CC2635EULL, "vxht_0_hightown_n_a81_0_0.vlm"},
	{0x57B2DF5B8AE4BE86ULL, "vxht_keep_n_4mg_0_0.vlm"},
	{0x57B3076FC498EADCULL, "6093686_m.fxe"},
	{0x57B369C227BFC588ULL, "vxw_blightlands_fm_0_0.vlm"},
	{0x57B3A7FE221C2214ULL, "hlac_lt_backalley_n_0_15y.dds"},
	{0x57B3AD2253C6C4D3ULL, "llac_lt_0_lowtown_docks_a3_196608_0.dds"},
	{0x57B3BB0EB3278F5DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railnopost.mao"},
	{0x57B3BF0EB3279596ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railnopost.mmh"},
	{0x57B3C90EB327A694ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railnopost.msh"},
	{0x57B3EF099B5B4554ULL, "vxlt_0_lowtown_a2_n_e1t_0_0.vlm"},
	{0x57B3F3F59CB0EFAAULL, "vxlt_0_lowtown_e1y_0_0.vlm"},
	{0x57B3F71985E2BE5BULL, "6140274_f.fxe"},
	{0x57B45B48C2E100C5ULL, "hlac_w_sundermount_-131070_0y.dds"},
	{0x57B4AAA0F13B957BULL, "6000771_m.fxe"},
	{0x57B52474C8A10DC5ULL, "6128924_f.fxe"},
	{0x57B592BD22156E33ULL, "621780850.ogg"},
	{0x57B59BDAFDBF6EC0ULL, "vxca_caverns1_mines_ghk_0_0.vlm"},
	{0x57B5DF5CB33C3FB0ULL, "mcr370_cut_flight_gallows.bnk"},
	{0x57B680E7048A6F72ULL, "cntr_flare.mao"},
	{0x57B68357AC43E2B8ULL, "plt_ave200pt_mother_dead.plo"},
	{0x57B683CF2D3EAE6FULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\prp_blightrock_d.dds"},
	{0x57B7A18C94CF1FDAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crn_rdy.gad"},
	{0x57B7B62A86F97FEDULL, "6139644_m.fxe"},
	{0x57B84847FFC37FC1ULL, "lgt100cr_gang_1_ht.utc"},
	{0x57B93BFD886F240CULL, "6090373_m.fxe"},
	{0x57B94991766D59FCULL, "6122726_m.fxe"},
	{0x57B95FBE99F25B29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_scared.ani"},
	{0x57B9AD5F265DBCC1ULL, "gen00fl_aveline6137555.cl"},
	{0x57B9C8AA1556A926ULL, "gst_none.gda"},
	{0x57B9E6C457DD357EULL, "6119915_m.fxe"},
	{0x57B9FFEBE25C47E1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brazier_sa.dds"},
	{0x57BAE7C75D027B4AULL, "6088352_m.fxe"},
	{0x57BB0B183F2C3FE7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchwall_d.dds"},
	{0x57BB416758BCF590ULL, "214795741.ogg"},
	{0x57BB62F415C66839ULL, "vxlt_darktown_5au_0_0.vlm"},
	{0x57BBA177FC98EBE4ULL, "mrl251ip_loot_b.utp"},
	{0x57BBBEFDBC6158FEULL, "ser211ip_note_scribbles2.utp"},
	{0x57BC8D103E997268ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_001a_0s.dds"},
	{0x57BCD7EC788C2CF9ULL, "196728679.ogg"},
	{0x57BE1D59BF68F927ULL, "6059975_m.fxe"},
	{0x57BE82D06204FBF2ULL, "hlac_w_blightlands_0_4y.dds"},
	{0x57BEA0AC0288F89DULL, "gen00fl_aveline6122122.cl"},
	{0x57BEA48C94D543F1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crn_rdy.evt"},
	{0x57BECCABBA20733CULL, "vxlt_0_lowtown_docks_a3_bv7_0_0.vlm"},
	{0x57BED47B9EFBDE6DULL, "hlac_lt_0_lowtown_docks_-131074_4x.dds"},
	{0x57BF110ACD74F6DAULL, "208421881.ogg"},
	{0x57BF35060E93BB4BULL, "493212998.ogg"},
	{0x57BF5E5AF4F69A08ULL, "lt_0_lowtown_docks_170_0.msh"},
	{0x57BF705AF4F6B899ULL, "lt_0_lowtown_docks_170_0.mao"},
	{0x57BF7C5AF4F6CD32ULL, "lt_0_lowtown_docks_170_0.mmh"},
	{0x57BFC9EB218C45E9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3_tool.phy"},
	{0x57BFE168157B33DEULL, "6173677_f.fxe"},
	{0x57BFF8E23EA909BFULL, "plt_cod_cha_divine_justina.plo"},
	{0x57C0404B794468CFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand_2.mao"},
	{0x57C070309D44C221ULL, "6115372_m.fxe"},
	{0x57C07ABE99F8A7ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_scared.gad"},
	{0x57C0F667A050C8DBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railsingle.msh"},
	{0x57C0FFFEACB2A725ULL, "mcr370_meredith6083074.cl"},
	{0x57C10867A050E726ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railsingle.mao"},
	{0x57C11467A050FBC5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railsingle.mmh"},
	{0x57C15F0E2944C67DULL, "mag101_alain6032195.cl"},
	{0x57C1939D973A530CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chaintiexa.msh"},
	{0x57C1999D973A5DFEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chaintiexa.mmh"},
	{0x57C1A59D973A7195ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chaintiexa.mao"},
	{0x57C1D6032417106EULL, "dae000cr_grand_cleric6124752.cl"},
	{0x57C208DFF399EA4AULL, "dae000_hate_follower_react6167475.cl"},
	{0x57C2689CC02D9838ULL, "vxlt_0_lowtown_a3_n_c_eki_0_0.vlm"},
	{0x57C273B06AA07B1DULL, "17452084.ogg"},
	{0x57C400B8FC5A7CFAULL, "678434998.ogg"},
	{0x57C474D3AF1F2AB0ULL, "enc_trips_sp.gda"},
	{0x57C59D2421CD647EULL, "vxlt_0_lowtown_a3_n_c_es7_0_0.vlm"},
	{0x57C677BE99FD0E10ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_scared.evt"},
	{0x57C6B5A4CC3A8AE4ULL, "vxca_caverns2_7ti_0_0.vlm"},
	{0x57C6C3B6D83A83B7ULL, "dae000_hate_enemies6167568.cl"},
	{0x57C6CCCD67A3D9D5ULL, "vxca_caverns1_mines_e88_0_0.vlm"},
	{0x57C6F64406695971ULL, "llac_lt_0_lowtown_a3_n_c_131072_6.dds"},
	{0x57C75FF90A301086ULL, "vxw_bonepitdestroyed_a3_1xa_0_0.vlm"},
	{0x57C79828DC0F05DBULL, "vxht_keep_n_3xn_0_0.vlm"},
	{0x57C7B98323D6BF80ULL, "framebuffereffects_twopassblurcochorizontl_p0.psh.fxo"},
	{0x57C7E4EEFA1E2942ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v3.gad"},
	{0x57C82B2A0E2657ADULL, "m_pro000ar_gallow_court.bnk"},
	{0x57C8312820DD851DULL, "294368865.ogg"},
	{0x57C9A0AAD6D8EF6DULL, "art\\characters\\playercharacter\\elfmale\\em_glv_lgte.msh"},
	{0x57C9B2AAD6D90DF3ULL, "art\\characters\\playercharacter\\elfmale\\em_glv_lgte.mmh"},
	{0x57C9C65E16A9CDF8ULL, "6078102_m.fxe"},
	{0x57C9CD1B26C8509FULL, "2179330.ogg"},
	{0x57C9D4D107BC80B9ULL, "vxlt_darktown_5jt_0_0.vlm"},
	{0x57C9FA3A28D77DAFULL, "face_face_p0.vsh.fxo"},
	{0x57CA06C44456A3A9ULL, "6076102_f.fxe"},
	{0x57CA66F028E68D08ULL, "fx_st_cloud_add.dds"},
	{0x57CB2E595A6840E6ULL, "vxlt_undercityserialkiller_1a4_0_0.vlm"},
	{0x57CB33B99DE0EF5EULL, "seb321_leliana6197859.cl"},
	{0x57CB64D4C00621E2ULL, "vxdr_0_deeproads_gfy_0_0.vlm"},
	{0x57CB95973BE3F353ULL, "lgt220_shp_mag_guard6149893.cl"},
	{0x57CBBC8C94DFF6C0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crn_rdy.ani"},
	{0x57CBE948CCFBD62FULL, "dae000_oldmangossip6174770.cl"},
	{0x57CBF3874BAAF487ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_action.gad"},
	{0x57CCD95D3323049EULL, "hlac_lt_0_lowtown_n_65534_5x.dds"},
	{0x57CD38C3F9989E66ULL, "hlac_lt_0_lowtown_docks_a3_n_0_6z.dds"},
	{0x57CD3CAF55E7242AULL, "mcr371_cut_orsino_defiance6077474.cl"},
	{0x57CD6942A3FF7928ULL, "gen00fl_sebastian6189397.cl"},
	{0x57CD75FFEE01A8C9ULL, "vxlt_0_lowtown_docks_n_b81_0_1.vlm"},
	{0x57CD9AFC77ABDEBEULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades.mmh"},
	{0x57CDAD96C84682B1ULL, "88782105.ogg"},
	{0x57CE1B14EB54AFC9ULL, "vxht_mansion2b_32e_0_0.vlm"},
	{0x57CEC6BAA12E713AULL, "6108136_f.fxe"},
	{0x57CEE42A6F5A4C97ULL, "765245851.ogg"},
	{0x57CF07EEFA248339ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v3.evt"},
	{0x57CF4F3A456B1E9BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wi_galowint.msh"},
	{0x57CF613A456B3DE6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wi_galowint.mao"},
	{0x57CF6D3A456B5185ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wi_galowint.mmh"},
	{0x57CFD48E886CDDD3ULL, "6127645_m.fxe"},
	{0x57CFE3C506C02001ULL, "dae150_radha.bnk"},
	{0x57D04D7C0BABF688ULL, "vxw_blightlands_lk_0_0.vlm"},
	{0x57D0692D9EB2E96FULL, "259223396.ogg"},
	{0x57D0B5FF78A15D00ULL, "6049134_f.fxe"},
	{0x57D0C39C2B3E07D9ULL, "380890951.ogg"},
	{0x57D14415D6EE043EULL, "vxlt_gallowstemplar_n_3rj_0_0.vlm"},
	{0x57D15CCB594EFC89ULL, "hlac_lt_0_lowtown_131069_0x.dds"},
	{0x57D218AE4B6EC76BULL, "lgt210_qun_blk_qunari6142392.cl"},
	{0x57D24215E641D78BULL, "hlac_w_sundermount_-327679_0y.dds"},
	{0x57D2E8874BB10094ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_action.evt"},
	{0x57D305C506C29071ULL, "dae150_radha.cnv"},
	{0x57D33FC141019A38ULL, "6076153_m.fxe"},
	{0x57D35ED345952235ULL, "mcr370_cut_player_survives6171838.cl"},
	{0x57D3E3D87CC26DC9ULL, "6041118_m.fxe"},
	{0x57D473605AB59BA4ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_imp_p.anb"},
	{0x57D4776B6C8F21ADULL, "w_bonepitdestroyed_a3_157_0.mmh"},
	{0x57D47B6B6C8F28FEULL, "w_bonepitdestroyed_a3_157_0.mao"},
	{0x57D4896B6C8F4033ULL, "w_bonepitdestroyed_a3_157_0.msh"},
	{0x57D532C94758BB05ULL, "hlac_lt_gallowscourtyard_-65538_3x.dds"},
	{0x57D5350E77F5A3DDULL, "206967302.ogg"},
	{0x57D5A7DB59D5F565ULL, "hlac_lt_0_lowtown_a2_-131072_9x.dds"},
	{0x57D630D5FE94E615ULL, "vxht_0_hightown_n_31g_0_0.vlm"},
	{0x57D6770D83D8CF4BULL, "6084461_m.fxe"},
	{0x57D6C614834998CCULL, "dae150_harshal6148845.cl"},
	{0x57D6D012EB18C990ULL, "plt_one100pt_loose_end_meer.nss"},
	{0x57D6E2BAE1CECD88ULL, "6176749_m.fxe"},
	{0x57D6FE6DF6919D45ULL, "lt_0_lowtown_a2_n_26_0_ll.mmh"},
	{0x57D70AA500E3DCDCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzardb_sa.dds"},
	{0x57D75989C491FD01ULL, "gen00fl_fenris6110922.cl"},
	{0x57D75D88F2A914CCULL, "vxw_encounter1_f6_0_1.vlm"},
	{0x57D766085161DD30ULL, "vxht_mansion2_1rj_0_0.vlm"},
	{0x57D78E295177C16CULL, "259546557.ogg"},
	{0x57D7E38E98B1B094ULL, "850748521.ogg"},
	{0x57D868D7E3E9A351ULL, "vxht_mansion2_2kc_0_0.vlm"},
	{0x57D878B3FD1C50F5ULL, "qun110cr_varnell.utc"},
	{0x57D905D762CAB8A8ULL, "lt_0_lowtown_docks_a3_216_0.msh"},
	{0x57D90A7EF006539BULL, "3992884.ogg"},
	{0x57D917D762CAD739ULL, "lt_0_lowtown_docks_a3_216_0.mao"},
	{0x57D91F0F92B7B9CCULL, "439200253.ogg"},
	{0x57D923D762CAEBD2ULL, "lt_0_lowtown_docks_a3_216_0.mmh"},
	{0x57D99A6A00EFF3D0ULL, "661312959.ogg"},
	{0x57D9C6E2BCF75F01ULL, "hlac_lt_undercityencounter_-1_9z.dds"},
	{0x57D9C79C690F9080ULL, "vxht_mansion2_1v6_0_0.vlm"},
	{0x57D9E9276116FE13ULL, "6124778_m.fxe"},
	{0x57DA0BB7C0269E62ULL, "hm_gxa_vgk_slvr_or_gd.mor"},
	{0x57DA1634BAD53BFDULL, "vxlt_0_lowtown_a2_hd0_0_0.vlm"},
	{0x57DA1B59CC6C2936ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospider.evt"},
	{0x57DA5CA61FDB01E7ULL, "6178487_m.fxe"},
	{0x57DA94BABEAF3CE7ULL, "6169449_m.fxe"},
	{0x57DA970D90672ECAULL, "dae300crst_cullenaveline.stg"},
	{0x57DAC49E9029F439ULL, "qun120_arishok_guard_nt6063079.cl"},
	{0x57DB246956F27D61ULL, "6149588_m.fxe"},
	{0x57DB60C7A5A65E94ULL, "vxdr_primevaltomb_505_0_0.vlm"},
	{0x57DBA3A070169C89ULL, "783784105.ogg"},
	{0x57DBE7DCDFF7C149ULL, "vxlt_0_lowtown_docks_n_b56_0_1.vlm"},
	{0x57DC0FEEFA2F1B28ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v3.ani"},
	{0x57DC16CC827B4534ULL, "vxht_keep_63b_0_0.vlm"},
	{0x57DC241A9A331F75ULL, "vxht_mansion2b_1tu_0_0.vlm"},
	{0x57DC6EA7B6296A76ULL, "aoe_core.ncs"},
	{0x57DCA3225FDC7664ULL, "vxlt_0_lowtown_n_bxp_0_0.vlm"},
	{0x57DCA89D22309EEFULL, "6135344_f.fxe"},
	{0x57DCBEC82E220BB0ULL, "dae000_unhappy_mage6165324.cl"},
	{0x57DCEC32E33E95EAULL, "w_encounter2_camp_5_0.mao"},
	{0x57DCF032E33E9C39ULL, "w_encounter2_camp_5_0.mmh"},
	{0x57DCF03349CBF2DFULL, "vxlt_0_lowtown_a2_n_c_hq0_0_0.vlm"},
	{0x57DCFA32E33EAD27ULL, "w_encounter2_camp_5_0.msh"},
	{0x57DD27079BB0457AULL, "llac_w_bonepitmining_a2_262150_0.dds"},
	{0x57DEEA834CB9E56EULL, "hlac_dr_primevaltomb_658_1x.dds"},
	{0x57DF26CB7C78BDC1ULL, "329923702.ogg"},
	{0x57DF3AAF59CB2F21ULL, "vxlt_0_lowtown_docks_a3_n_d4u_0_0.vlm"},
	{0x57DF4C56A419C938ULL, "vxlt_0_lowtown_a3_n_c_eqg_0_0.vlm"},
	{0x57DF83221B78CBDFULL, "fontshader.vsh.fxo"},
	{0x57DFD13C481AA10CULL, "6135915_m.fxe"},
	{0x57DFDAE5FAAF335CULL, "mag291_torpor6026983.cl"},
	{0x57DFFF4ACD442703ULL, "vxw_bonepitmining_a2_27h_0_0.vlm"},
	{0x57E07708C9041212ULL, "vxw_bonepitdestroyed_a3_29y_0_0.vlm"},
	{0x57E086242FB73532ULL, "145853697.ogg"},
	{0x57E093B2DF181AABULL, "dr_slavers_3215.rml"},
	{0x57E0992F92790CA6ULL, "6119128_m.fxe"},
	{0x57E0B0944A91E192ULL, "vxlt_undercityencounter_13w_0_1.vlm"},
	{0x57E0C9F3EA733C84ULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.mmh"},
	{0x57E0E0874BBD30CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_action.ani"},
	{0x57E11259CC7238C1ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospider.gad"},
	{0x57E117E6823E6D5DULL, "sho_dis.mao"},
	{0x57E11B46B6956639ULL, "hlac_lt_gallowstemplar_2391_8x.dds"},
	{0x57E28F858926D437ULL, "mag161tr_post_spiders.utt"},
	{0x57E292605AC20DA7ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_imp_p.mmh"},
	{0x57E2CB33EA4349DFULL, "14444231.ogg"},
	{0x57E3120BFD4ADF60ULL, "vxlt_gallowstemplar_37h_0_0.vlm"},
	{0x57E318092BD921D1ULL, "709089264.ogg"},
	{0x57E34F9F520B78BDULL, "6166284_m.fxe"},
	{0x57E3995EB026BBE4ULL, "vxca_caverns2_7no_0_0.vlm"},
	{0x57E3FE848DC303B0ULL, "gen_im_trs_gen_sta_phy_r1.uti"},
	{0x57E41C10CC8A4980ULL, "470314874.ogg"},
	{0x57E443B2EE1C4186ULL, "vxw_bonepitdestroyed_a3_1fc_0_0.vlm"},
	{0x57E47EB28072522FULL, "209771188.ogg"},
	{0x57E4BC40C14847F5ULL, "435804126.ogg"},
	{0x57E50280E874F9FAULL, "378218292.ogg"},
	{0x57E5706B6C9D896AULL, "w_bonepitdestroyed_a3_157_0.phy"},
	{0x57E5D7E54357F5B0ULL, "vxht_mansion2_1uk_0_0.vlm"},
	{0x57E5F33C71CBB35DULL, "dae000_bartrand6027562.cl"},
	{0x57E6B88AEBA8B1B9ULL, "vxlt_darktown_5pz_0_0.vlm"},
	{0x57E74B3903023B7EULL, "vxw_bonepitdestroyed_a3_1ns_0_0.vlm"},
	{0x57E7E1711B02F720ULL, "787131421.ogg"},
	{0x57E81224266D1CF9ULL, "nat380_nathaniel6068356.cl"},
	{0x57E81359CC78596FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospider.ani"},
	{0x57E830B1EBC807FDULL, "870689056.ogg"},
	{0x57E84A9006008104ULL, "vxlt_0_lowtown_docks_a3_n_9rc_0_0.vlm"},
	{0x57E8AAA1B8C14FF9ULL, "vxlt_0_lowtown_a2_gon_0_0.vlm"},
	{0x57E8F6510ACFFE9FULL, "gen00fl_isabela_init6175726.cl"},
	{0x57E8F7B749066741ULL, "vxht_mansion_n_6o0_0_0.vlm"},
	{0x57E970708A4DDE8FULL, "6197228_m.fxe"},
	{0x57E9BAE6A1B75A9AULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobd.phy"},
	{0x57E9E7A2025FE58FULL, "6176122_m.fxe"},
	{0x57EA01F7AD82A836ULL, "6152356_m.fxe"},
	{0x57EA11AFEFB6D558ULL, "6170862_m.fxe"},
	{0x57EA4875743FFA52ULL, "and100_anders6095978.cl"},
	{0x57EA69DDFD41EB5FULL, "715614081.ogg"},
	{0x57EA71C5A4FF75B4ULL, "hlac_lt_gallowstemplar_f_2395_1y.dds"},
	{0x57EAB03E0B782B24ULL, "dre000_hireling26038678.cl"},
	{0x57EABEDFD5C085EAULL, "vxdr_0_deeproads_bxf_0_0.vlm"},
	{0x57EB1A0F48E1912DULL, "vxw_woundedcoast_ms_0_0.vlm"},
	{0x57EBB2EB21B1590EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3_tool.mmh"},
	{0x57EBBEEB21B16DA5ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3_tool.mao"},
	{0x57EBCA79234530F3ULL, "hlac_lt_0_lowtown_a2_n_c_-1_5x.dds"},
	{0x57EBCCEB21B185FCULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3_tool.msh"},
	{0x57EC43E815030218ULL, "hlac_lt_0_lowtown_docks_a3_131068_0z.dds"},
	{0x57ECB8486C590ED3ULL, "6129845_m.fxe"},
	{0x57EDA630B4CD0FBFULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_thief_.gad"},
	{0x57EE3437EA5C9D02ULL, "cr_cycle_sp.mao"},
	{0x57EF3B372D477F99ULL, "299122316.ogg"},
	{0x57EF512FAAF230D2ULL, "hlac_lt_0_lowtown_a2_n_c_-2_1x.dds"},
	{0x57EF68ADC20B4F35ULL, "mer320_hubert6053676.cl"},
	{0x57EF75B246117633ULL, "gen00fl_merrill6116319.cl"},
	{0x57EFC72EDAD42A10ULL, "vxdr_primevaltomb_2po_0_1.vlm"},
	{0x57EFE125A1143A0CULL, "vxca_cavern1_haunted_jsw_0_0.vlm"},
	{0x57F107BEB242AA13ULL, "vxht_0_hightown_a3_c8x_0_0.vlm"},
	{0x57F15E05DD31A803ULL, "vxw_bonepitmining_a2_1s3_0_0.vlm"},
	{0x57F236D3CC5B2F81ULL, "hf_cgn_002.mor"},
	{0x57F236D3CC5B2F83ULL, "hf_cgn_002.mop"},
	{0x57F2AAF3981EC59CULL, "hlac_w_encounter2_camp_327680_0x.dds"},
	{0x57F2B58FE0125A06ULL, "vxw_bonepitdestroyed_a3_1kd_0_0.vlm"},
	{0x57F2B8CD19E6CCA5ULL, "10332031.ogg"},
	{0x57F33B81F7B404D5ULL, "283712770.ogg"},
	{0x57F360062C4F71FDULL, "572403158.ogg"},
	{0x57F3654E68A83A2BULL, "vxht_brothel_n_5b5_0_0.vlm"},
	{0x57F3A76AC6BCB84EULL, "w_encounter3_181_0.phy"},
	{0x57F3D606F5250C0FULL, "249187734.ogg"},
	{0x57F4212A0E4B81D7ULL, "m_pro000ar_gallow_court.wbd"},
	{0x57F4355355B4A6E3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1a.msh"},
	{0x57F4373908CA781AULL, "llac_ht_keep_4519_2.dds"},
	{0x57F4435355B4BE3DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1a.mmh"},
	{0x57F4475355B4C54EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1a.mao"},
	{0x57F48A65DA7264DBULL, "zz_dae_debug6029474.cl"},
	{0x57F4998DF2E424FDULL, "art\\characters\\playercharacter\\humanfemale\\arm_masa.mao"},
	{0x57F49D8DF2E42B36ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masa.mmh"},
	{0x57F4A78DF2E43C34ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masa.msh"},
	{0x57F4CCE5EA31035EULL, "vxca_caverns1_jfg_0_0.vlm"},
	{0x57F50E45B9AC9A76ULL, "zz_mrl_debug6185314.cl"},
	{0x57F54C91C7D5D451ULL, "vxht_mansion2_2qa_0_0.vlm"},
	{0x57F58E0A9C49099BULL, "6124709_m.fxe"},
	{0x57F5B5DAFC9E1F76ULL, "73052728.ogg"},
	{0x57F61CCDAB913717ULL, "6099548_m.fxe"},
	{0x57F69BAC878DB0E6ULL, "prty_merril_armor_2.dds"},
	{0x57F6A27B261BB268ULL, "ave100_donnic6106079.cl"},
	{0x57F6BCC559225689ULL, "mixgui.mao"},
	{0x57F71C7EAAB76879ULL, "vxlt_0_lowtown_a2_gbm_0_0.vlm"},
	{0x57F77874A29B6DE7ULL, "6163649_m.fxe"},
	{0x57F7E41BBFDAEE4DULL, "167070121.ogg"},
	{0x57F800ACA81FD57AULL, "499428558.ogg"},
	{0x57F847704D2D73DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\rdy.ani"},
	{0x57F9824AF258F888ULL, "5244296.ogg"},
	{0x57FA3E8B37A250C5ULL, "6140312_m.fxe"},
	{0x57FA5E895BCD3B30ULL, "art\\characters\\creatures\\shriek\\bdy_ems.mao"},
	{0x57FA62895BCD4243ULL, "art\\characters\\creatures\\shriek\\bdy_ems.mmh"},
	{0x57FA70895BCD599DULL, "art\\characters\\creatures\\shriek\\bdy_ems.msh"},
	{0x57FA73B240E20222ULL, "mrl150_marethaript2.stg"},
	{0x57FAAA909F796CA0ULL, "vxlt_darktown_52y_0_0.vlm"},
	{0x57FADC0318E49A86ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_bl.ani"},
	{0x57FBA032A1C0BD99ULL, "ser221_party_comments6163783.cl"},
	{0x57FC048368001243ULL, "pro000_opening_pc6001204.cl"},
	{0x57FCE304B1305803ULL, "vxw_bonepitmining_a2_29j_0_0.vlm"},
	{0x57FDA779FF365CBEULL, "756310556.ogg"},
	{0x57FDADEB29A38EAEULL, "llac_lt_0_lowtown_a2_n_-1_6.dds"},
	{0x57FE0B7FEE779831ULL, "vxlt_0_lowtown_a2_n_byz_0_0.vlm"},
	{0x57FE42A65D5D8EEAULL, "w_encounter2_54_0.mao"},
	{0x57FE46A65D5D9539ULL, "w_encounter2_54_0.mmh"},
	{0x57FE50A65D5DA627ULL, "w_encounter2_54_0.msh"},
	{0x57FE9B33AEA259AFULL, "6037687_m.fxe"},
	{0x57FEFB4CB305CEB2ULL, "vxw_bonepit_255_0_0.vlm"},
	{0x57FF548B58B3014FULL, "qun200_party_comment6090393.cl"},
	{0x58002171918C62CDULL, "cd_mcr371_hostile_aveline.bnk"},
	{0x580035AE8176DFBCULL, "hlac_w_bonepit_131076_0x.dds"},
	{0x580050BDA57CD1D6ULL, "vxca_cavern1_haunted_e65_0_0.vlm"},
	{0x5800FC55B6C5641CULL, "vxlt_hangedman_n_255_0_0.vlm"},
	{0x5801014C0870EB93ULL, "zz_one_debug6077413.cl"},
	{0x58011D2C229DDA5EULL, "112618975.ogg"},
	{0x580261D7E1929DB6ULL, "mer120_hubert6059387.cl"},
	{0x580264C9486E37B8ULL, "6063960_m.fxe"},
	{0x58028330B4DF3005ULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_thief_.ani"},
	{0x58028451D364218CULL, "215360340.ogg"},
	{0x5802BB4899ED88F0ULL, "6154297_m.fxe"},
	{0x5802BB9F274426B0ULL, "vxht_mansion2_1ci_0_0.vlm"},
	{0x5802CFB131ACF0A5ULL, "llac_ca_caverns2prc_5802_30.dds"},
	{0x58031A8A7E8FC079ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_activ_c.impa.ani"},
	{0x58037F0BA50ADB3FULL, "hlac_lt_gallowstemplar_2384_6z.dds"},
	{0x5803A55B0A41837EULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_hed_0s.dds"},
	{0x58051D333EF1DA00ULL, "vxht_mansion2_1g5_0_0.vlm"},
	{0x58059ABD2C748004ULL, "llac_ht_keep_n_3739_0.dds"},
	{0x580614B91112ADE0ULL, "6208687_f.fxe"},
	{0x58064F704D39BED2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\rdy.evt"},
	{0x5806716380E49623ULL, "6095391_m.fxe"},
	{0x5806B01C58382C3CULL, "t3_csp_b02.tnt"},
	{0x5806E0B05A1AF57EULL, "gen00fl_anders6125240.cl"},
	{0x5806E561426E6571ULL, "406493955.ogg"},
	{0x5806FCB123311E17ULL, "6103302_m.fxe"},
	{0x580748D43C7E394DULL, "962066570.ogg"},
	{0x5807FDC92CCDC22DULL, "vxw_woundedcoast_gq_0_0.vlm"},
	{0x5808411A451679B0ULL, "dae000_templar_knight6165181.cl"},
	{0x58084DE55EB590B6ULL, "vxlt_gallowscourtyard_n_73o_0_0.vlm"},
	{0x58088E7E7A115710ULL, "178178682.ogg"},
	{0x5808994BFD06855CULL, "dae150_radha6148927.cl"},
	{0x5808BA9FADBEE26FULL, "6173572_m.fxe"},
	{0x5808BD935552EDFAULL, "w_bonepitdestroyed_a3_143_0.mmh"},
	{0x5808C1935552F431ULL, "w_bonepitdestroyed_a3_143_0.mao"},
	{0x5808C40318F0AFF7ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_bl.evt"},
	{0x5808CCF3EA94947BULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.anb"},
	{0x5808CF9355530C40ULL, "w_bonepitdestroyed_a3_143_0.msh"},
	{0x5808D41900A2D7BCULL, "953646438.ogg"},
	{0x5808ED67A08E6F72ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railsingle.phy"},
	{0x58094C7915F343E0ULL, "351169007.ogg"},
	{0x5809F58A7E959FBBULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_activ_c.impa.gad"},
	{0x5809F634A36FB82FULL, "qun220cr_javaris6068906.cl"},
	{0x580A488AD2FCF550ULL, "313931196.ogg"},
	{0x580A5F03B706D0EDULL, "6076497_m.fxe"},
	{0x580A7AA28D5D9177ULL, "81341392.ogg"},
	{0x580B63895BDBB768ULL, "art\\characters\\creatures\\shriek\\bdy_ems.phy"},
	{0x580BEE90ED6BAC94ULL, "vxw_bonepit_y7_0_0.vlm"},
	{0x580C16F6325B387BULL, "vxht_keep_n_42s_0_0.vlm"},
	{0x580C597AE699AF81ULL, "236628277.ogg"},
	{0x580CE489562358A1ULL, "898531669.ogg"},
	{0x580CF8DB8722D2D1ULL, "6142083_f.fxe"},
	{0x580D4E31994CE285ULL, "gen00fl_fenris6108062.cl"},
	{0x580D6073E6899299ULL, "130464511.ogg"},
	{0x580D6D29A4FBE732ULL, "vxw_bonepit_206_0_0.vlm"},
	{0x580D6E704D401265ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\rdy.gad"},
	{0x580D95E506F5B4F4ULL, "hlac_lt_0_lowtown_a2_n_65536_3x.dds"},
	{0x580D9DBB9128E448ULL, "hlac_lt_0_lowtown_n_-65538_0x.dds"},
	{0x580D9E4DD82FAEBEULL, "art\\vfx\\cinematic\\vfx_dragon_saliva_c.mmh"},
	{0x580DEB78962EDB13ULL, "vxht_0_hightown_a3_c6z_0_0.vlm"},
	{0x580E71FC77E33059ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades.phy"},
	{0x580EC50318F51C80ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_bl.gad"},
	{0x580F5FADB9377EDFULL, "zz_mer_debug6045140.cl"},
	{0x581015B16766EA2FULL, "6148123_m.fxe"},
	{0x581049084C946B2BULL, "vxht_brothel_n_5h7_0_0.vlm"},
	{0x58106A8F5808D6B6ULL, "aip_gateguardian_plot.gda"},
	{0x5810D2879BAB7E17ULL, "vxlt_warehouse_1cl_0_0.vlm"},
	{0x5811128A7E9BEFA0ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_activ_c.impa.evt"},
	{0x5811FEF65A1185B6ULL, "6185244_m.fxe"},
	{0x581218E447157D5BULL, "732236703.ogg"},
	{0x58123BB31D7CD2EEULL, "651709578.ogg"},
	{0x5812442638F2424EULL, "gen00fl_bethany6104056.cl"},
	{0x58126C68195E41E9ULL, "hlac_ht_0_hightown_a2_-131072_0y.dds"},
	{0x581293A61F852A56ULL, "vxlt_hangedman_n_27i_0_0.vlm"},
	{0x581300878F7D6817ULL, "6093748_m.fxe"},
	{0x5813A0A4AE881842ULL, "vrc321_party_barks6095776.cl"},
	{0x58143BF1A5C481FEULL, "vxlt_0_lowtown_docks_b5c_0_0.vlm"},
	{0x58148DAD19489321ULL, "art\\vfx\\combat\\vfx_glb_foot_dust_c.impa.ani"},
	{0x58150CDD85987C25ULL, "mag131_keran6086184.cl"},
	{0x58152D1AE520003EULL, "853629562.ogg"},
	{0x581541951B5BF976ULL, "6059251_m.fxe"},
	{0x58156B0C924008ABULL, "hlac_ht_0_hightown_a3_-65535_1y.dds"},
	{0x58156E8E15273697ULL, "6104202_m.fxe"},
	{0x5816463A45A71232ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wi_galowint.phy"},
	{0x581670B369B39A39ULL, "ave200_aveline_bar6115506.cl"},
	{0x581804B7AA765FE4ULL, "character_character_p207.psh.fxo.dx11"},
	{0x58186B01D8A6AAA6ULL, "6121989_m.fxe"},
	{0x5818A1837C1889A5ULL, "llac_lt_0_lowtown_docks_n_-131073_0.dds"},
	{0x58190AF495B806B0ULL, "969975063.ogg"},
	{0x5819D5A73D5299DAULL, "hlac_lt_gallowscourtyard_-1_1y.dds"},
	{0x5819F2697CC04EB6ULL, "qun211_party_comment6072172.cl"},
	{0x581A47BEB976D8DDULL, "static_static_p15.psh.fxo.dx11.assembly"},
	{0x581A92B706A9550BULL, "gen00fl_varric6105430.cl"},
	{0x581AC7EF239A2411ULL, "vxlt_0_lowtown_docks_n_bn2_0_0.vlm"},
	{0x581B1AE9B1ECF472ULL, "6144939_m.fxe"},
	{0x581B3A3B6344734CULL, "lt_0_lowtown_n_32_0_ll.mmh"},
	{0x581B5AE787A5539BULL, "234865719.ogg"},
	{0x581B7D4DD83BB47DULL, "art\\vfx\\cinematic\\vfx_dragon_saliva_c.anb"},
	{0x581B98AD194EC4C3ULL, "art\\vfx\\combat\\vfx_glb_foot_dust_c.impa.gad"},
	{0x581BB725334D4482ULL, "gen00pt_vault.ncs"},
	{0x581BBB0C4BF0692BULL, "682987067.ogg"},
	{0x581C0C99EEF67A58ULL, "vxlt_0_lowtown_a2_n_c_gv9_0_0.vlm"},
	{0x581DC620CF44241AULL, "qun110_petrice6059144.cl"},
	{0x581EAB00C957A832ULL, "878975620.ogg"},
	{0x581EE4B5705BE790ULL, "vxlt_gallowstemplar_f_5wu_0_0.vlm"},
	{0x581EE60386BC4F70ULL, "6060306_f.fxe"},
	{0x581FD826EA778E47ULL, "vxw_encounter2_72_0_0.vlm"},
	{0x58207179FE8FBC99ULL, "1057541570.ogg"},
	{0x5820BA46D71176DAULL, "vxht_keep_64e_0_0.vlm"},
	{0x5820F8A6CB54A9FBULL, "ghostarmor.tnt"},
	{0x58210D1CEA7EA5A5ULL, "atl_prccontent_dxt5_00.dds"},
	{0x5821DEE462AE31FDULL, "6107291_f.fxe"},
	{0x582212DA4CD3E11DULL, "one140cr_charade6134587.cl"},
	{0x582275AD1954A7A8ULL, "art\\vfx\\combat\\vfx_glb_foot_dust_c.impa.evt"},
	{0x5822F4D89EF2DA90ULL, "llac_w_bonepitdestroyed_a3_196608_0.dds"},
	{0x58231BF93B6784F4ULL, "593271403.ogg"},
	{0x5823795D1FF04B71ULL, "6035203_m.fxe"},
	{0x5823CD32FD9A488CULL, "art\\characters\\playercharacter\\qunarimale\\bdy_stn.msh"},
	{0x5823D332FD9A537EULL, "art\\characters\\playercharacter\\qunarimale\\bdy_stn.mmh"},
	{0x5823DF32FD9A6715ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_stn.mao"},
	{0x582404565E9CF7BEULL, "445782637.ogg"},
	{0x5824C2C7886FACE2ULL, "mage_heroicaura.dds"},
	{0x5824D7530A8A3A31ULL, "6105508_m.fxe"},
	{0x5824E18310B9F32DULL, "vxw_woundedcoast_yo_0_0.vlm"},
	{0x58252ED6D1BF258EULL, "plt_vrc200pt_n_initial.plo"},
	{0x5826F5DB00858FB7ULL, "w_bonepitmining_a2_150_0.phy"},
	{0x582727F6EE74B055ULL, "725014358.ogg"},
	{0x582747908C8FFE6CULL, "vxlt_undercitydungeon_11s_0_0.vlm"},
	{0x582816085F748E96ULL, "gen00fl_varric6081979.cl"},
	{0x58282A33F03DAD22ULL, "hlac_ca_caverns1_mines_24425_1x.dds"},
	{0x58282B1D97E3E1B1ULL, "236140223.ogg"},
	{0x58287BE239B8CE54ULL, "mcr371_idunna6087277.cl"},
	{0x5828CF7728738D60ULL, "vxht_0_hightown_a2_a6b_0_0.vlm"},
	{0x5828D24AD157DD94ULL, "vxw_bonepit_o5_0_0.vlm"},
	{0x5829A4858962FAE6ULL, "mag161tr_post_spiders.ncs"},
	{0x5829AC5927E986F6ULL, "gen00fl_merrill6131346.cl"},
	{0x5829C06BB877A70EULL, "6077689_f.fxe"},
	{0x582A843C92F3E4E9ULL, "static_static_p99.psh.fxo.dx11"},
	{0x582A90286ADC7B5EULL, "666273471.ogg"},
	{0x582AC5FF61DB48D5ULL, "aip_spidermonstrous_base.gda"},
	{0x582AE4730BACF51CULL, "vxht_bartrand_n_4g0_0_0.vlm"},
	{0x582B21BD087DB197ULL, "zz_charge_blight.cut"},
	{0x582B424398BA96EBULL, "144050999.ogg"},
	{0x582B626A0A15DEE2ULL, "dae000_quintus6154789.cl"},
	{0x582B6299D724657FULL, "6027844_m.fxe"},
	{0x582B8407C2902B81ULL, "962448699.ogg"},
	{0x582BBCF70F4DC7CDULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef03.mmh"},
	{0x582BCEF70F4DE653ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef03.msh"},
	{0x582C041E33C698D0ULL, "gen_im_trs_slv_junk2_04.uti"},
	{0x582C177191B18CF7ULL, "cd_mcr371_hostile_aveline.wbd"},
	{0x582C19937D23CEDAULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden.mao"},
	{0x582C69A3F8AA727EULL, "t3_wdg_w18.tnt"},
	{0x582CA3AC00C11A59ULL, "fxe_splash_add.mao"},
	{0x582DB0EC80792077ULL, "vxlt_warehouse_n_o8_0_0.vlm"},
	{0x582E182574CAB906ULL, "vxlt_0_lowtown_docks_b7u_0_0.vlm"},
	{0x582EB749E2F05EB6ULL, "135578681.ogg"},
	{0x582F279D8F531E3BULL, "hlac_ca_caverns1_mines_24424_14z.dds"},
	{0x582F2C23C9078F5AULL, "vxht_keep_61f_0_0.vlm"},
	{0x582FC1A4359A3A14ULL, "vxca_caverns1_j20_0_1.vlm"},
	{0x58300E892FBFC494ULL, "vxlt_0_lowtown_n_f13_0_0.vlm"},
	{0x583022E828359FA6ULL, "hlac_lt_0_lowtown_n_131071_5z.dds"},
	{0x5830FC57909AD1BAULL, "vxlt_hangedman_n_2yg_0_0.vlm"},
	{0x583125C1E6229E46ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hallastatue_sa.dds"},
	{0x58312F6D79B88343ULL, "dae011cr_conspiracy_theor.utc"},
	{0x58313FF302EB4E6FULL, "vxlt_0_lowtown_a2_ewv_0_0.vlm"},
	{0x58319F4D96ECA20BULL, "gen00fl_aveline6139865.cl"},
	{0x5831AD4100BCF362ULL, "vxlt_warehouse_n_px_0_0.vlm"},
	{0x58321FD9D6598F8DULL, "dae300_petrice6097872.cl"},
	{0x583270F5CCEE6B35ULL, "hlac_ca_caverns1_24423_0y.dds"},
	{0x5832CAA8BD0A2A92ULL, "vxlt_0_lowtown_cmw_0_0.vlm"},
	{0x583308124252150DULL, "hightown_market_guard_sta6169610.cl"},
	{0x58333A29CD407BCDULL, "vxlt_warehouse_n_11s_0_0.vlm"},
	{0x5833536002B00BADULL, "vxw_woundedcoast_vn_0_0.vlm"},
	{0x583376DC63B4AF74ULL, "dae000_gamlen6149450.cl"},
	{0x5833A66AC6F29A11ULL, "w_encounter3_181_0.mmh"},
	{0x5833B26AC6F2AEB2ULL, "w_encounter3_181_0.mao"},
	{0x5833C06AC6F2C6CFULL, "w_encounter3_181_0.msh"},
	{0x583417FB75548EE7ULL, "dae000_fenarel6133451.cl"},
	{0x58347D1BA938EEBCULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_c.impa.ani"},
	{0x5834927A44B0D2D1ULL, "6125542_m.fxe"},
	{0x5834E9682696608DULL, "w_bonepitdestroyed_a3_227_0.msh"},
	{0x5834F768269678C0ULL, "w_bonepitdestroyed_a3_227_0.mao"},
	{0x5834FB6826967F13ULL, "w_bonepitdestroyed_a3_227_0.mmh"},
	{0x58355E02981EE935ULL, "zz_ave_debug6174806.cl"},
	{0x5836382B7C62E1ACULL, "hlac_ca_caverns1_24425_4z.dds"},
	{0x58364EF8AE993A7EULL, "32561957.ogg"},
	{0x583666CD5FFFA53DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangsign_sa.dds"},
	{0x5836708BD752112DULL, "vxw_bonepit_1zx_0_0.vlm"},
	{0x5836C60430E97830ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_inkwell.msh"},
	{0x5836D40430E9904AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_inkwell.mmh"},
	{0x5836D80430E99681ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_inkwell.mao"},
	{0x5837056A3072655FULL, "hlac_ht_bartrand_n_3364_0x.dds"},
	{0x58372F3419096F42ULL, "hlac_lt_backalley_g_0_17x.dds"},
	{0x583740C7CB35E654ULL, "llac_w_sundermount_-327675_0.dds"},
	{0x583790407B456DAFULL, "dae000_orana6149067.cl"},
	{0x5837F032E38B4792ULL, "w_encounter2_camp_5_0.rml"},
	{0x58380229971CC1F8ULL, "gen_im_wep_war_two_act3_02.uti"},
	{0x583859585C1139EFULL, "24966651.ogg"},
	{0x583879B915C1D57FULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd2a.msh"},
	{0x58387FB915C1DF21ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd2a.mmh"},
	{0x5838B3B2BE81ABEBULL, "687313649.ogg"},
	{0x5838C3376685801FULL, "62306082.ogg"},
	{0x5838F5F539C1AB05ULL, "static_static_p8.psh.fxo.dx11.assembly"},
	{0x5839564FFDA30D9CULL, "vxht_bartrand_n_4l1_0_0.vlm"},
	{0x58396F631807D151ULL, "hlac_w_sundermount_327679_0z.dds"},
	{0x5839AABCE28E4C7BULL, "vxw_woundedcoast_fx_0_0.vlm"},
	{0x5839BA7B1D4ACD00ULL, "6101813_m.fxe"},
	{0x5839D53FBCA2562DULL, "6110375_m.fxe"},
	{0x583A1FD6D1D1681DULL, "plt_vrc200pt_n_initial.nss"},
	{0x583A776639B5C0C0ULL, "qun221ip_at_cht_htown_nt.utp"},
	{0x583B0FD9944859B3ULL, "6116090_m.fxe"},
	{0x583B8A1BA93F23FEULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_c.impa.gad"},
	{0x583BC1EF537EF2C8ULL, "plt_zz_avept_debug.plo"},
	{0x583BF202CAAB2A7CULL, "6170361_m.fxe"},
	{0x583C0A1F60CF6D98ULL, "llac_ht_0_hightown_a3_-65534_0.dds"},
	{0x583CB5F70F5C2E8AULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef03.phy"},
	{0x583CCB47FA0E5107ULL, "6148829_f.fxe"},
	{0x583CFEDB06A1C29AULL, "mrl000_marethari6069624.cl"},
	{0x583D6371DFB3F8F8ULL, "221202526.ogg"},
	{0x583D8232D3640DE8ULL, "vxlt_0_lowtown_a2_n_ei0_0_0.vlm"},
	{0x583DE15387374E76ULL, "891492476.ogg"},
	{0x583E44EB81695A28ULL, "m_mcr372ar_gallows_templar_nt.bnk"},
	{0x583EB7423E06CDE0ULL, "6133074_m.fxe"},
	{0x583EC8CD61B7349AULL, "hlac_lt_0_lowtown_docks_n_-65540_0x.dds"},
	{0x583EF132E3916826ULL, "w_encounter2_camp_5_0.phy"},
	{0x583F5781B40B698BULL, "6155811_m.fxe"},
	{0x583FC8C1CA291894ULL, "vxca_caverns2prc_9so_0_0.vlm"},
	{0x584002ACB64D1A9EULL, "374149035.ogg"},
	{0x58401F1DF2B30BE2ULL, "vxlt_warehouse_n_wy_0_0.vlm"},
	{0x584088C051D5545CULL, "vxht_mansion2_1vj_0_0.vlm"},
	{0x5841478042635285ULL, "mag311_huon6154411.cl"},
	{0x5841B6DF8CC5910AULL, "6083685_m.fxe"},
	{0x5841C53CF4A6242DULL, "vxw_woundedcoast_sm_0_0.vlm"},
	{0x5841D9121F1A9BEDULL, "490733085.ogg"},
	{0x584216671A5F714CULL, "vxca_caverns1_mines_imv_0_0.vlm"},
	{0x5842651BA945031DULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_c.impa.evt"},
	{0x58436F3C401596F5ULL, "nat380_sibling6068017.cl"},
	{0x58437915995C821DULL, "llac_lt_0_lowtown_65537_0.dds"},
	{0x5843D0E38AF599E9ULL, "hlac_dr_0_deeproads_17038_2z.dds"},
	{0x58443DA301256E61ULL, "vxlt_darktown_2yj_0_0.vlm"},
	{0x5844614304A13FDBULL, "hlac_lt_0_lowtown_a2_-65537_1x.dds"},
	{0x5844E17BA8C1523EULL, "317840526.ogg"},
	{0x5844EE84854502ADULL, "6149427_m.fxe"},
	{0x5845AD83C43323AEULL, "6085452_m.fxe"},
	{0x5845B3E6A205ADC3ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobd.msh"},
	{0x5845C1E6A205C51DULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobd.mmh"},
	{0x5845C5E6A205CC2EULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nobd.mao"},
	{0x584643F2D32202C2ULL, "gen_im_trs_dal_med_glv_r0.uti"},
	{0x5846C976C69CE7F1ULL, "vxlt_darktown_5am_0_0.vlm"},
	{0x58470BBB150D1CA6ULL, "1046795491.ogg"},
	{0x584743CB7950607DULL, "304921934.ogg"},
	{0x584758A4B473E11CULL, "562481774.ogg"},
	{0x58479D787D8F2836ULL, "617973445.ogg"},
	{0x584810520373C9BAULL, "w_woundedcoast_118_0.phy"},
	{0x58482132B64726EDULL, "6174089_m.fxe"},
	{0x58483D5D22B507F0ULL, "6108490_m.fxe"},
	{0x58487F98CF1DC059ULL, "6117935_f.fxe"},
	{0x5848B888E2E11B0FULL, "343625008.ogg"},
	{0x5848EE232DCA5AE1ULL, "dae000_hate_follower_react6103252.cl"},
	{0x5849706721DE87DFULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mfollower1_scaredmage.gad"},
	{0x584A8BEC12493D81ULL, "vxht_0_hightown_a3_5gy_0_0.vlm"},
	{0x584AB6E32FA81469ULL, "vxlt_foundry_n_zs_0_0.vlm"},
	{0x584ABFF5A0CDD0AAULL, "hlac_lt_0_lowtown_docks_a3_-65539_1y.dds"},
	{0x584AF0F17EADA31FULL, "traps.gda"},
	{0x584AFDD50799CEE0ULL, "gen_im_wep_mag_sta_ico_fake.uti"},
	{0x584BCFBBCDB075D4ULL, "193091902.ogg"},
	{0x584C06D27B334225ULL, "6064887_m.fxe"},
	{0x584C9666CC945DF4ULL, "119225035.ogg"},
	{0x584CC8481D18051BULL, "6077324_m.fxe"},
	{0x584DD258B5F83102ULL, "gen00fl_anders6107709.cl"},
	{0x584E90FAE4A92462ULL, "vxlt_warehouse_n_zz_0_0.vlm"},
	{0x584E9185B783821DULL, "gen00fl_isabela_init6175665.cl"},
	{0x584EBEB1FD4A58A1ULL, "vxht_brothel_n_540_0_1.vlm"},
	{0x584EFA9D43CB6CDCULL, "vxht_mansion2_1yk_0_0.vlm"},
	{0x584F14408655DB31ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_sand01_sa.dds"},
	{0x584F1F9BA73F3187ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany_loop.gad"},
	{0x584F425AC598A877ULL, "vxlt_darktown_4u2_0_0.vlm"},
	{0x584FBDB401608C4AULL, "6185107_m.fxe"},
	{0x585028BC7EBBA98AULL, "6086985_m.fxe"},
	{0x585088440C5589CCULL, "vxca_caverns1_mines_ijw_0_0.vlm"},
	{0x5850A397ABBB4DD9ULL, "vxlt_undercitydungeon_1xm_0_0.vlm"},
	{0x5850C4EE01132DE5ULL, "gen00fl_fenris6107153.cl"},
	{0x5851796B9D709B01ULL, "s_ballista_05_l2.mao"},
	{0x585191A76E5BD88EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantrywind.phy"},
	{0x58531205C2AFD9D7ULL, "vxlt_gallowstemplar_n_2jb_0_0.vlm"},
	{0x5853A1130DB29748ULL, "776140959.ogg"},
	{0x5853BCF38E774EEEULL, "vxlt_darktown_4y3_0_0.vlm"},
	{0x5853C1FC2A29E76DULL, "vxlt_darktown_4vy_0_0.vlm"},
	{0x5853F6C238B96141ULL, "dae300_alistair_king6133542.cl"},
	{0x585404E9CF451F45ULL, "hlac_w_bonepitmining_a2_131076_0y.dds"},
	{0x585440A3B2B274FAULL, "vxdr_primevaltomb_2wo_0_2.vlm"},
	{0x5854670F640156FBULL, "dae000_madam_lusine6152289.cl"},
	{0x5854A8937F8C3A00ULL, "6153490_m.fxe"},
	{0x5854E60286CE9D1BULL, "6165539_m.fxe"},
	{0x58553B53B8930071ULL, "vxlt_darktown_5jl_0_0.vlm"},
	{0x5855F8216C97F8CFULL, "vxlt_gallowscourtyard_3vj_0_0.vlm"},
	{0x5856149BA7453D94ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany_loop.evt"},
	{0x585681BD1AD78D38ULL, "gen_im_arm_cht_elf_coma_t5.uti"},
	{0x58568E76C67A7D7BULL, "vxw_woundedcoast_xz_0_0.vlm"},
	{0x5856FA77C0E02395ULL, "693065983.ogg"},
	{0x58573091EF1F32F0ULL, "vfx_rendertoshadowmap_p2.vsh.fxo.dx11"},
	{0x5857900801652FF8ULL, "6043676_m.fxe"},
	{0x5857D6642EFD1D81ULL, "718884470.ogg"},
	{0x58586C4D5366A796ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_mera_0d.dds"},
	{0x58588463E668CCDFULL, "vxdr_0_deeproads_bp6_0_1.vlm"},
	{0x585884EDB8F9A5AAULL, "gen00fl_isabela6103974.cl"},
	{0x5858B53465D78294ULL, "49615722.ogg"},
	{0x5858CF233EBDA4B5ULL, "435678581.ogg"},
	{0x5858DCBAF3426E44ULL, "744941036.ogg"},
	{0x5858F7BA7209FB26ULL, "vxca_cave_2jw_0_0.vlm"},
	{0x58591A8016A9F52CULL, "isa211_isabela6202098.cl"},
	{0x585946A65DAA4092ULL, "w_encounter2_54_0.rml"},
	{0x58595F7D5858EC56ULL, "vxht_bartrand_n_4sf_0_0.vlm"},
	{0x5859F2F6DC9424C7ULL, "vxw_blightlands_o5_0_0.vlm"},
	{0x585A2868362E5AE2ULL, "vxw_sundermount_33x_0_1.vlm"},
	{0x585A8F2BA1E64CC3ULL, "hlac_lt_0_lowtown_docks_a3_n_-196607_0y.dds"},
	{0x585B9AFC21F2FEE0ULL, "6139674_m.fxe"},
	{0x585BB25F69FF7D7BULL, "hlac_lt_darktown_-1_51y.dds"},
	{0x585C413D3C32B003ULL, "653968659.ogg"},
	{0x585C425D2EA7ED80ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.mmh"},
	{0x585C6E63193F11D4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windownk.phy"},
	{0x585C80C9CDB7E069ULL, "vxlt_gallowstemplar_f_5qz_0_0.vlm"},
	{0x585CB857FA673116ULL, "gen00fl_bethany6206064.cl"},
	{0x585D3E6CFC513E0EULL, "117742602.ogg"},
	{0x585D4B96544DC944ULL, "mcr371_hostile_anders6082116.cl"},
	{0x585D4D6721EEF525ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mfollower1_scaredmage.ani"},
	{0x585D9C6C912E72B7ULL, "6150294_m.fxe"},
	{0x585DBEEF539BD387ULL, "plt_zz_avept_debug.nss"},
	{0x585E08F6EF632D4DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chanterb_n.dds"},
	{0x585E15E0F829AC3DULL, "zz_approval_debug6003107.cl"},
	{0x585E2B7CC4591786ULL, "6015989_m.fxe"},
	{0x585E46B9A776C08CULL, "gen00fl_anders6110053.cl"},
	{0x585EE819D37E4A7AULL, "76145797.ogg"},
	{0x585EFA20FE4BA24CULL, "vxca_caverns1_mines_igt_0_0.vlm"},
	{0x585F8ADF72E0C57BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairmid.msh"},
	{0x585F9CDF72E0E3C6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairmid.mao"},
	{0x585FA8DF72E0F865ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairmid.mmh"},
	{0x585FC674D7098A17ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbreadoven.phy"},
	{0x586047A65DB06126ULL, "w_encounter2_54_0.phy"},
	{0x586051892801C9A8ULL, "bld_mac_03.mao"},
	{0x586109D68B60CB6AULL, "store_dae110_access.utm"},
	{0x5861E312E85BA161ULL, "792675218.ogg"},
	{0x5862039CDC2C92B0ULL, "fxe_day_beam.mao"},
	{0x58629AE914A1D8A8ULL, "381748560.ogg"},
	{0x5862AF1A8C17BB41ULL, "vxlt_0_lowtown_a3_n_c_htd_0_0.vlm"},
	{0x5862DC1B0563FF2CULL, "qun110_petrice6042227.cl"},
	{0x5863013FDD689E91ULL, "vxlt_0_lowtown_docks_n_bw5_0_0.vlm"},
	{0x586309CF1558B5A1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\sister_pcopening.gad"},
	{0x5863285C5B39F5DEULL, "358951959.ogg"},
	{0x5863836A0E74381EULL, "136218784.ogg"},
	{0x5863AD30AA8918F1ULL, "vxlt_darktown_5oc_0_0.vlm"},
	{0x58640C9BA7516DCDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany_loop.ani"},
	{0x586436EB818A11D2ULL, "m_mcr372ar_gallows_templar_nt.wbd"},
	{0x586447701CE3C6FEULL, "follower_banter6143157.cl"},
	{0x5864AA32FDD1A419ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_stn.phy"},
	{0x586508DD8925D64DULL, "vxht_0_hightown_a2_9uv_0_0.vlm"},
	{0x586580B915E8CE9EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd2a.phy"},
	{0x58659F4777A1B631ULL, "vxlt_0_lowtown_a2_ghw_0_0.vlm"},
	{0x5865EBACCE839D59ULL, "aip_abomination_base.gda"},
	{0x5865ECCC78E15600ULL, "vxht_bartrand_n_5pp_0_0.vlm"},
	{0x58664DEADC7F06B4ULL, "default_mage_female.utc"},
	{0x586672592D33677BULL, "llac_w_encounter1_65539_0.dds"},
	{0x58669BFE3FA38E15ULL, "llac_lt_gallowstemplar_n_2391_5.dds"},
	{0x58671F198199B314ULL, "qcr000_arishok6059329.cl"},
	{0x586739BBA2BE0F99ULL, "vxdr_0_deeproads_f0e_0_0.vlm"},
	{0x58674B631948DDD1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windownk.msh"},
	{0x586759631948F5ECULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windownk.mao"},
	{0x586765631949098FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windownk.mmh"},
	{0x586789C23E80E616ULL, "vxht_keep_n_5jg_0_0.vlm"},
	{0x586864D3CE8A3D47ULL, "vxw_blightlands_p4_0_0.vlm"},
	{0x5869255D2EB2469FULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.anb"},
	{0x5869767BBB601453ULL, "dae000ip_crowd_em2_talk4.utp"},
	{0x586A0ACF155ED64FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\sister_pcopening.ani"},
	{0x586A0DA87D6367E7ULL, "6170058_m.fxe"},
	{0x586A0F7E97829CCCULL, "vxht_mansion2b_31e_0_0.vlm"},
	{0x586A12E28F64690DULL, "314289912.ogg"},
	{0x586A3BC2DF4912B2ULL, "6087700_m.fxe"},
	{0x586A4D8A782E06D8ULL, "87133938.ogg"},
	{0x586AB517235087F4ULL, "gen00fl_varric6118998.cl"},
	{0x586AF3B875756C2CULL, "vxdr_primevaltomb_6ni_0_1.vlm"},
	{0x586BDE5727B79DDCULL, "vxht_mansion2_1gi_0_0.vlm"},
	{0x586C027AFF44945AULL, "362334209.ogg"},
	{0x586C2C2E194D770CULL, "677958226.ogg"},
	{0x586C541B03AFFAA6ULL, "hlac_lt_undercitydungeon_-1_33y.dds"},
	{0x586D1D12273CD8D3ULL, "hlac_lt_0_lowtown_1_6y.dds"},
	{0x586D481CE13C6F9DULL, "atl_loginviewt_dxt5_00.dds"},
	{0x586D683B60F1936FULL, "vxht_keep_n_6ix_0_0.vlm"},
	{0x586DA1A55528FFB5ULL, "zz_followers_banter6158301.cl"},
	{0x586DC416A919CA5FULL, "366384588.ogg"},
	{0x586DCADB00C1490FULL, "w_bonepitmining_a2_150_0.mao"},
	{0x586DD6DB00C15DACULL, "w_bonepitmining_a2_150_0.mmh"},
	{0x586DDCDB00C1679EULL, "w_bonepitmining_a2_150_0.msh"},
	{0x586E4720A689B5B3ULL, "6166758_m.fxe"},
	{0x586E4F646D91D385ULL, "6027796_m.fxe"},
	{0x586EACD495E8F157ULL, "vxht_0_hightown_cb3_0_0.vlm"},
	{0x587037FF9F2A732DULL, "vxw_bonepit_1nt_0_0.vlm"},
	{0x5870A4170956C5E0ULL, "hm_raider_3.mor"},
	{0x5871351BE07F52D6ULL, "mrl000_marethari6129230.cl"},
	{0x5871795744C16628ULL, "790306982.ogg"},
	{0x58723060A0FABD1AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxb.mmh"},
	{0x58723460A0FAC451ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxb.mao"},
	{0x58724260A0FADC60ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxb.msh"},
	{0x5873AB9040A1F9ABULL, "vxht_0_hightown_n_c0j_0_0.vlm"},
	{0x5874150C1D840949ULL, "708244264.ogg"},
	{0x587435C1830276C4ULL, "hm_ave3_insurgent_guard2.mor"},
	{0x5874550525F10865ULL, "6142260_m.fxe"},
	{0x58747931E3D05AF7ULL, "isa211_isabela6202275.cl"},
	{0x5874C09CAC226F75ULL, "683864662.ogg"},
	{0x5874F02C065FE467ULL, "286431304.ogg"},
	{0x587557B98BFD5485ULL, "732847134.ogg"},
	{0x5876A107DCAC8FC3ULL, "hlac_lt_undercityencounter_-1_29x.dds"},
	{0x5876B5974ABC0B4FULL, "art\\vfx\\combat\\vfx_r_dwp_critical_strike_c.impa.gad"},
	{0x5876D6B0C08055C7ULL, "vxw_blightlands_u7_0_0.vlm"},
	{0x5876DA79B198F1AEULL, "116275772.ogg"},
	{0x587752AA6AB439A6ULL, "6095397_f.fxe"},
	{0x5877A6241EF9485FULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_c.impa.gad"},
	{0x5877CE96C9393DA9ULL, "6072981_m.fxe"},
	{0x5877EEEF2F181CD1ULL, "gen_im_arm_boo_med_champion.uti"},
	{0x587823B303B4AE85ULL, "vxht_keep_n_670_0_0.vlm"},
	{0x5878AFD3CE66D0D3ULL, "vxht_mansion2_3g4_0_0.vlm"},
	{0x58798A50ED5ED38DULL, "6102079_m.fxe"},
	{0x5879C2658C330E8DULL, "6131077_m.fxe"},
	{0x5879C62F64614D12ULL, "narrative_flemeth6002790.cl"},
	{0x5879F6BC877222A2ULL, "w_bonepitdestroyed_a3_178_0.msh"},
	{0x5879F8BC87722658ULL, "w_bonepitdestroyed_a3_178_0.mmh"},
	{0x587A04BC87723AFBULL, "w_bonepitdestroyed_a3_178_0.mao"},
	{0x587A10F4F04BD0E2ULL, "zz_spy130_darktown_res_16055315.cl"},
	{0x587A9940793C3797ULL, "gen_im_npc_wep_two_2hsa.uti"},
	{0x587B965287C4BC33ULL, "llac_x_narrativeroom_0_0.dds"},
	{0x587BCB5A7F325480ULL, "6087760_m.fxe"},
	{0x587C50AEC9367D9BULL, "6161074_m.fxe"},
	{0x587C5A413FE30F29ULL, "hlac_w_encounter3_-327681_0x.dds"},
	{0x587DCA974AC24DFCULL, "art\\vfx\\combat\\vfx_r_dwp_critical_strike_c.impa.evt"},
	{0x587E0A2801D70B29ULL, "73873098.ogg"},
	{0x587E9B241EFF544CULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_c.impa.evt"},
	{0x587F2D3629EF456CULL, "vxca_cavern1_haunted_huk_0_0.vlm"},
	{0x587F3F6E16E8FA5EULL, "atl_deathscree_dxt5_dat.xml"},
	{0x587FF889C8FA6B9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_still.gad"},
	{0x588050EF55B72251ULL, "vxht_0_hightown_n_cak_0_0.vlm"},
	{0x588064C86C3FFC0DULL, "vxlt_hangedman_n_tc_0_0.vlm"},
	{0x58811F2532AB2BE1ULL, "138853430.ogg"},
	{0x58813A5133F7722DULL, "6001046_m.fxe"},
	{0x58815D54BB3F4F4EULL, "zz_followers_banter6158775.cl"},
	{0x58816AB3A28A86F9ULL, "art\\vfx\\creatures\\fxc_arch_fb_p.impa.ani"},
	{0x588186E1BA877F94ULL, "93117669.ogg"},
	{0x58819D664D139043ULL, "vxlt_darktown_48i_0_0.vlm"},
	{0x588230898FBB6F52ULL, "vxht_mansion2_3g1_0_0.vlm"},
	{0x588283015B8DE731ULL, "vxlt_0_lowtown_a2_gbu_0_0.vlm"},
	{0x5882CC917EE913E1ULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v1.ani"},
	{0x5882FFCF52CEA41DULL, "vxlt_hangedman_10o_0_0.vlm"},
	{0x5883A288FA76BD15ULL, "6216523_m.fxe"},
	{0x5883E8D9D9D3A772ULL, "vxdr_0_deeproads_bpw_0_1.vlm"},
	{0x58843B262CF33506ULL, "6043832_m.fxe"},
	{0x58847E8AB028C2F6ULL, "696395103.ogg"},
	{0x588490BB69598658ULL, "nat380_nathaniel6066609.cl"},
	{0x5884A4B8494F1AE4ULL, "69038565.ogg"},
	{0x5884D29F81707B89ULL, "6067349_m.fxe"},
	{0x58850AA8B5F9997FULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_fin_c.impa.gad"},
	{0x5885162542469000ULL, "hlac_lt_0_lowtown_a2_131070_1z.dds"},
	{0x588519BCA94B5E01ULL, "vxlt_0_lowtown_docks_a3_n_bsz_0_0.vlm"},
	{0x58851D1A97773306ULL, "art\\vfx\\cinematic\\vfx_blood_dripping_wpool_c.mmh"},
	{0x588708B137CEDA56ULL, "gen00fl_isabela6090854.cl"},
	{0x5887535A5F758ACAULL, "llac_ht_0_hightown_a2_-65537_2.dds"},
	{0x588801B272AC3C31ULL, "hlac_lt_0_lowtown_a3_n_c_0_0x.dds"},
	{0x58884161EBBE0088ULL, "134505871.ogg"},
	{0x588845B3A290663BULL, "art\\vfx\\creatures\\fxc_arch_fb_p.impa.gad"},
	{0x5888D106856C2632ULL, "vxlt_backalley_n_1rd_0_0.vlm"},
	{0x5888D2F46710840BULL, "hlac_lt_undercityencounter_-1_9x.dds"},
	{0x5888D651338B2711ULL, "t1_skn_h12.tnt"},
	{0x5888EADD733D7728ULL, "vxlt_backalley_n_1oh_0_1.vlm"},
	{0x58893608BB2800ECULL, "14907326.ogg"},
	{0x588951793074FEFAULL, "hlac_lt_0_lowtown_65537_5x.dds"},
	{0x5889D7917EEF4483ULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v1.gad"},
	{0x5889E10BA4186EB0ULL, "dre000_vampire6034405.cl"},
	{0x5889E7DE22A37081ULL, "pro000_db_ogre_rogu.cut"},
	{0x588A6B4413E3C45DULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravec.phy"},
	{0x588A7CEEEC7FBE2FULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkl.gad"},
	{0x588AA2B6C3CD9AF1ULL, "6035297_f.fxe"},
	{0x588ABF06C274AB0FULL, "169906677.ogg"},
	{0x588AD2974ACCE515ULL, "art\\vfx\\combat\\vfx_r_dwp_critical_strike_c.impa.ani"},
	{0x588AFFA8B5FDF26CULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_fin_c.impa.evt"},
	{0x588B565E5366EC74ULL, "vxlt_0_lowtown_n_ch9_0_0.vlm"},
	{0x588B83241F09B5A5ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_c.impa.ani"},
	{0x588BF65546C715A0ULL, "433625104.ogg"},
	{0x588C0457635AB6C8ULL, "hlac_dr_slavers_3214_2z.dds"},
	{0x588C09F28A8A8DF5ULL, "llac_lt_0_lowtown_a2_-65534_0.dds"},
	{0x588C12D3BEC064C2ULL, "hlac_lt_gallowscourtyard_-65537_3z.dds"},
	{0x588C5B5B736ED534ULL, "6024583_m.fxe"},
	{0x588C5E457D6886D9ULL, "gen00fl_fenris6126545.cl"},
	{0x588D562E00BBDDD8ULL, "vxlt_0_lowtown_docks_n_3m7_0_0.vlm"},
	{0x588D896FF202496DULL, "vxlt_darktown_4bu_0_0.vlm"},
	{0x588DA44D17B51F76ULL, "lt_0_lowtown_docks_n_220_0.mao"},
	{0x588DB04D17B53315ULL, "lt_0_lowtown_docks_n_220_0.mmh"},
	{0x588DB24D17B5368BULL, "lt_0_lowtown_docks_n_220_0.msh"},
	{0x588DB6725521F25FULL, "vrc221_bartrand6043065.cl"},
	{0x588DB968DB2DA2CCULL, "gen00fl_aveline6106879.cl"},
	{0x588DBC6835835723ULL, "hlac_ht_chantry_n_2534_16z.dds"},
	{0x588DC0293B2B6E13ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.dura.evt"},
	{0x588E04DD0200B32FULL, "hlac_lt_gallowstemplar_f_2391_7x.dds"},
	{0x588E4866CF319F01ULL, "vxht_mansion2_1g8_0_0.vlm"},
	{0x588E56F70A335FF5ULL, "zz_area_signage13.utp"},
	{0x588E79C0E38A1FFEULL, "gen00fl_aveline6115420.cl"},
	{0x588E7E42B4A71259ULL, "6105715_m.fxe"},
	{0x588ED6A55E36148DULL, "vxlt_hangedman_n_sb_0_0.vlm"},
	{0x588F323580E57CABULL, "llac_ht_mansion2_818_21.dds"},
	{0x588F62B3A296B620ULL, "art\\vfx\\creatures\\fxc_arch_fb_p.impa.evt"},
	{0x588F72431A7937F9ULL, "vxlt_gallowsprison_n_6rp_0_0.vlm"},
	{0x588F8EA4BF4A0D56ULL, "6037314_m.fxe"},
	{0x588FB85632C1A146ULL, "630546945.ogg"},
	{0x588FCFC7B9A0BA92ULL, "730538404.ogg"},
	{0x589070692736A313ULL, "hlac_w_encounter2_262148_0x.dds"},
	{0x5890B61F50958388ULL, "813201306.ogg"},
	{0x5890C71D24C651CFULL, "dae000_epic_merchant_16179430.cl"},
	{0x5890CCCF6D209815ULL, "lt_0_lowtown_docks_1_0_lc.mmh"},
	{0x58915F87946E538CULL, "framebuffereffects_bloomblend_p0.psh.fxo.dx11"},
	{0x5891695A47857A58ULL, "plt_gen00pt_vault.plo"},
	{0x58918A8EE142ECADULL, "vxca_caverns1_egp_0_0.vlm"},
	{0x589191EEEC8600DCULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkl.evt"},
	{0x58919B463DB46923ULL, "llac_ht_chantry_2534_40.dds"},
	{0x5891F5368046F5F3ULL, "gen00fl_sebastian6188643.cl"},
	{0x58925B7DF2FD660CULL, "vxlt_backalley_g_1jo_0_0.vlm"},
	{0x589265CE931D75C3ULL, "217667754.ogg"},
	{0x589290A76E936E51ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantrywind.mmh"},
	{0x58929CA76E9382F2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantrywind.mao"},
	{0x5892AAA76E939A0FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantrywind.msh"},
	{0x5892F9BB2948EB6FULL, "298230120.ogg"},
	{0x5893099219CF5DCFULL, "ts_kirkwall_wpn_props.gda"},
	{0x5893B9293B2FCE54ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.dura.gad"},
	{0x589427B4DCD240A4ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_cracked01_n.dds"},
	{0x58943071BD10C69DULL, "vxht_mansion_n_52l_0_0.vlm"},
	{0x58945FEF7D5C8B21ULL, "gen_im_acc_amu_act3_06.uti"},
	{0x5894BDCFDFA7C4E0ULL, "art\\characters\\creatures\\corpse\\animation\\blendtree\\corpse_archer.anb"},
	{0x5894D589C90C8CE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_still.ani"},
	{0x5894EA9D2E07EB30ULL, "hlac_w_woundedcoast_-65528_0z.dds"},
	{0x5895104386986CE4ULL, "309211877.ogg"},
	{0x5895D4B6A82C9BCCULL, "vxca_caverns1_mines_j20_0_0.vlm"},
	{0x58961A9B976BCBFAULL, "210365919.ogg"},
	{0x589630833879F91DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobc_0t.dds"},
	{0x58967D2C7C1BB31BULL, "6156573_m.fxe"},
	{0x58974F60A11A0E25ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxb.phy"},
	{0x5897E6B51CB49FB7ULL, "vxht_keep_55p_0_0.vlm"},
	{0x5897FA83D570BF00ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_visc.phy"},
	{0x5898E7A8B60A06C5ULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_fin_c.impa.ani"},
	{0x5898EA04A02F71CBULL, "6096850_f.fxe"},
	{0x5899B5CD876B5BF2ULL, "sky_ht_brothel_n.mao"},
	{0x5899C4365D669175ULL, "hlac_lt_gallowstemplar_n_7198_3z.dds"},
	{0x5899D67DB0051FD2ULL, "zz_dae_debug6049992.cl"},
	{0x5899DFEEC2858DE5ULL, "vxlt_gallowstemplar_n_4uo_0_0.vlm"},
	{0x589A43E29586EE79ULL, "vxht_mansion_n_4vq_0_0.vlm"},
	{0x589A57357977600AULL, "hlac_lt_0_lowtown_a2_n_c_131070_0z.dds"},
	{0x589ABAE5182E9336ULL, "vxht_keep_63p_0_0.vlm"},
	{0x589AC8293B360632ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.dura.ani"},
	{0x589B22D28F1B90A8ULL, "6128028_f.fxe"},
	{0x589B39A4CA9B767BULL, "237502297.ogg"},
	{0x589B604413F224C9ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravec.mao"},
	{0x589B6C4413F23962ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravec.mmh"},
	{0x589B6E4413F23C18ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravec.msh"},
	{0x589B72FD3C6D4F3FULL, "vxca_cavern1_haunted_hze_0_0.vlm"},
	{0x589BA211CFC40104ULL, "zz_dae_debug6029916.cl"},
	{0x589C5D20123E6F77ULL, "6175967_m.fxe"},
	{0x589C785152303B7EULL, "6134734_m.fxe"},
	{0x589C8BB2ECAE1C95ULL, "ss_harlan.bnk"},
	{0x589C95D91474FD09ULL, "qun131ar_undercity_warrens.lst"},
	{0x589C9E3804060F3DULL, "lt_0_lowtown_docks_a3_n_8_0.anb"},
	{0x589D8969A1CE6200ULL, "268732493.ogg"},
	{0x589DB25F1E4C10D5ULL, "gen00fl_aveline6139218.cl"},
	{0x589DF056248E4943ULL, "gen00fl_fenris6126437.cl"},
	{0x589DF06DBF98EE1EULL, "l_mag111ar_warehouse.bnk"},
	{0x589E168AAB9A7210ULL, "venge_fuelledrage.dds"},
	{0x589E308E0018525FULL, "fer_ter_rock_001_diff.dds"},
	{0x589E413F4A004D4AULL, "733268740.ogg"},
	{0x589E78F1EDEB9D42ULL, "fx_litbolt01.mao"},
	{0x589EDF83B0F8037CULL, "vxdr_0_deeproads_ebi_0_0.vlm"},
	{0x589F42E0997D0AFEULL, "vxw_woundedcoast_hk_0_0.vlm"},
	{0x589F62439A288039ULL, "plt_cod_hst_mag_embrium.plo"},
	{0x589F99EEEC924BF5ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkl.ani"},
	{0x589FC43D94F1C997ULL, "one300_zevran6075731.cl"},
	{0x58A0CD5AE4F37E8CULL, "vxlt_backalley_g_1wl_0_0.vlm"},
	{0x58A0F22E3D56CAD7ULL, "mag131_keran6086285.cl"},
	{0x58A21ADA9DA97487ULL, "vxlt_0_lowtown_a3_n_c_eqe_0_1.vlm"},
	{0x58A24B589EC35A20ULL, "qun200_party_comment6090338.cl"},
	{0x58A2CB25D31E73A3ULL, "vxw_sundermount_27q_0_0.vlm"},
	{0x58A36DD9147AD840ULL, "qun131ar_undercity_warrens.ncs"},
	{0x58A3EB301B9135BCULL, "vxlt_darktown_69w_0_0.vlm"},
	{0x58A4095203C21CE3ULL, "w_woundedcoast_118_0.msh"},
	{0x58A4175203C2343DULL, "w_woundedcoast_118_0.mmh"},
	{0x58A41A00AD0BA09FULL, "w_blightlands_138_0.msh"},
	{0x58A41B5203C23B4EULL, "w_woundedcoast_118_0.mao"},
	{0x58A42000AD0BAB41ULL, "w_blightlands_138_0.mmh"},
	{0x58A429638BE2E7A5ULL, "vxlt_undercityencounter_1d0_0_0.vlm"},
	{0x58A42C00AD0BBFE2ULL, "w_blightlands_138_0.mao"},
	{0x58A4559C5AD255A9ULL, "287561059.ogg"},
	{0x58A4E92E938960F6ULL, "gen00fl_anders6125194.cl"},
	{0x58A55AB4F5FF7912ULL, "761669496.ogg"},
	{0x58A5690918F5D04CULL, "476517109.ogg"},
	{0x58A598C70F7C7B04ULL, "llac_lt_0_lowtown_a2_n_-65535_5.dds"},
	{0x58A5B4C069585732ULL, "vxlt_backalley_n_1xf_0_0.vlm"},
	{0x58A68244DC93DEE1ULL, "follower_banter6172884.cl"},
	{0x58A69B74D745446FULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbreadoven.mao"},
	{0x58A6A774D745580CULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbreadoven.mmh"},
	{0x58A6AD74D74562FEULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbreadoven.msh"},
	{0x58A781DF731E6B12ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairmid.phy"},
	{0x58A800EBF70292A0ULL, "vxlt_darktown_5nm_0_0.vlm"},
	{0x58A8B61F4C471F21ULL, "6054080_m.fxe"},
	{0x58A963A3C8C8C574ULL, "vxlt_gallowstemplar_5cb_0_0.vlm"},
	{0x58A97905FE7AE127ULL, "242549960.ogg"},
	{0x58A9962151390577ULL, "pro000_cut2d_thefirstyear6005304.cl"},
	{0x58A9A3A6CA3DC5DFULL, "835797976.ogg"},
	{0x58A9E4DA2E6367BFULL, "vxca_cavern1_haunted_hud_0_0.vlm"},
	{0x58AA2386717F3D49ULL, "vxca_caverns1_mines_i2f_0_0.vlm"},
	{0x58AA298186A2A74EULL, "6004406_m.fxe"},
	{0x58AAB92CC5DA0DCDULL, "dae000ip_crowd_em4_talk13.utp"},
	{0x58AAEEA971331AABULL, "t3_eye_dbl.tnt"},
	{0x58AB03A02C1B2273ULL, "vxca_caverns2_97q_0_0.vlm"},
	{0x58AB7417ED86B976ULL, "6119209_m.fxe"},
	{0x58ABB600B4BE11F3ULL, "276508980.ogg"},
	{0x58ABD6BBD973D426ULL, "606915429.ogg"},
	{0x58ABDCD9442A28ABULL, "dre000_varric6033321.cl"},
	{0x58AC000173C8B213ULL, "llac_lt_0_lowtown_a2_65537_0.dds"},
	{0x58AC79F060ED8C5FULL, "vxlt_hangedman_n_2z7_0_0.vlm"},
	{0x58AD8ED9148364AEULL, "qun131ar_undercity_warrens.are"},
	{0x58ADB8C2676C5BF2ULL, "345709011.ogg"},
	{0x58ADEBB109959745ULL, "683872962.ogg"},
	{0x58AE087801D131C3ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2b.mao"},
	{0x58AE0C7801D13810ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2b.mmh"},
	{0x58AE1A7801D1502AULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2b.msh"},
	{0x58AECADAE4452077ULL, "355896867.ogg"},
	{0x58AEF497F8C8C572ULL, "hlac_lt_warehouse_n_0_2x.dds"},
	{0x58AFB9F0F0CF35E3ULL, "vxht_bartrand_n_6ki_0_0.vlm"},
	{0x58B06B036A1F6728ULL, "9160857.ogg"},
	{0x58B0A1AFF4ABF5F6ULL, "858785397.ogg"},
	{0x58B13A7FD2D5B334ULL, "ser221_gascard6122451.cl"},
	{0x58B13D02C5148C23ULL, "vxw_sundermount_28p_0_0.vlm"},
	{0x58B14167C29BD5CBULL, "hlac_lt_gallowsprison_n_-1_4x.dds"},
	{0x58B169722D2411D4ULL, "17427254.ogg"},
	{0x58B16E41C1461F36ULL, "vxlt_gallowstemplar_f_43f_0_0.vlm"},
	{0x58B19E564A2C0429ULL, "731635620.ogg"},
	{0x58B1A9F8A3FDF19CULL, "vxw_woundedcoast_1ff_0_0.vlm"},
	{0x58B1AC1C58B93502ULL, "mer120_hubert6173911.cl"},
	{0x58B1DD63ABD10991ULL, "6050332_m.fxe"},
	{0x58B1E4079DA51543ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_forward.ani"},
	{0x58B22EF6B85A2C2EULL, "mag291_demon_marethari6027981.cl"},
	{0x58B26095FC33E9FAULL, "vxlt_0_lowtown_a2_n_c_gp9_0_0.vlm"},
	{0x58B2638E5D0C8B48ULL, "333948536.ogg"},
	{0x58B2AE3F07E8B7AAULL, "lt_gallowscourtyard_n_water_96.mao"},
	{0x58B2B23F07E8BEF9ULL, "lt_gallowscourtyard_n_water_96.mmh"},
	{0x58B2BC3F07E8CFE7ULL, "lt_gallowscourtyard_n_water_96.msh"},
	{0x58B308675810A8F2ULL, "hlac_ca_caverns2prc_5802_40x.dds"},
	{0x58B361EB363B4864ULL, "336493101.ogg"},
	{0x58B4E6D70F7A8CCFULL, "ave200_donnic6112168.cl"},
	{0x58B4ED85962A717AULL, "539568125.ogg"},
	{0x58B538E40AB637C9ULL, "vxht_mansion2b_3bu_0_0.vlm"},
	{0x58B5552463E21087ULL, "qun221_petrice6083651.cl"},
	{0x58B5C13AB846F2CDULL, "434709671.ogg"},
	{0x58B5E9C8A6E8C99AULL, "199630552.ogg"},
	{0x58B60A165C214293ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r.evt"},
	{0x58B645A3F28D5ABDULL, "856737645.ogg"},
	{0x58B6739C6269A881ULL, "6164439_m.fxe"},
	{0x58B68C25162F88D6ULL, "vxlt_darktown_56s_0_0.vlm"},
	{0x58B6C5A9700CC499ULL, "266700099.ogg"},
	{0x58B6E39BFB1ED0ADULL, "hlac_lt_0_lowtown_a2_n_c_-65535_0z.dds"},
	{0x58B705266F4E7A07ULL, "6200034_m.fxe"},
	{0x58B710846A96BB29ULL, "hlac_lt_backalley_n_65536_1x.dds"},
	{0x58B72AC62A95CDD5ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_005a_0s.dds"},
	{0x58B7337A0278850AULL, "hlac_ht_keep_3736_2x.dds"},
	{0x58B78222FEA053A2ULL, "gen00fl_varric6033677.cl"},
	{0x58B78E8311C8D41CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_left_walk.ani"},
	{0x58B8C08944920285ULL, "one100_party_comments6106771.cl"},
	{0x58B97E309EA48517ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.dura.gad"},
	{0x58BB81F2386895EBULL, "554261653.ogg"},
	{0x58BB866F2C0BFE00ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarp.phy"},
	{0x58BB953A68163411ULL, "llac_lt_darktown_-1_1.dds"},
	{0x58BBD617E88CF3BFULL, "vxlt_0_lowtown_a3_n_c_hws_0_0.vlm"},
	{0x58BC03165C25A2D4ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r.gad"},
	{0x58BCC91189705FAFULL, "vxlt_hangedman_su_0_0.vlm"},
	{0x58BCE025B725362DULL, "vxca_caverns1_ehm_0_0.vlm"},
	{0x58BD1783D5900C95ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_visc.msh"},
	{0x58BD1983D590100BULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_visc.mmh"},
	{0x58BD26C3B679B3F1ULL, "gen00fl_fenris6126691.cl"},
	{0x58BD5404D932E942ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14a.phy"},
	{0x58BD64278526F712ULL, "vxht_0_hightown_a2_bz3_0_0.vlm"},
	{0x58BD7411C7BB84F9ULL, "444541485.ogg"},
	{0x58BDA68EB6966651ULL, "plt_lgt311pt_shw_mer_base_1.plo"},
	{0x58BE9B8311CF095EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_left_walk.gad"},
	{0x58BEEEF816785370ULL, "darktown_fem_fereldan_sta6152813.cl"},
	{0x58BF24E776942675ULL, "6004052_m.fxe"},
	{0x58BF3A81098AA6B8ULL, "71636333.ogg"},
	{0x58BF5364AC10FE7AULL, "vxlt_warehouse_n_un_0_1.vlm"},
	{0x58BFC9DEE08D27EAULL, "vxht_chantry_32z_0_0.vlm"},
	{0x58BFFC079DB17B3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_forward.evt"},
	{0x58C01BD595F40A1CULL, "vxw_woundedcoast_1ag_0_0.vlm"},
	{0x58C0285FF6273313ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_dbl_c.impa.evt"},
	{0x58C074801CBC4845ULL, "vxlt_0_lowtown_a2_n_c_gz4_0_0.vlm"},
	{0x58C078895C7D078DULL, "6059029_m.fxe"},
	{0x58C0928433CBCC27ULL, "6050110_m.fxe"},
	{0x58C093309EAAC7A4ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.dura.evt"},
	{0x58C0E88F488671CBULL, "6144186_f.fxe"},
	{0x58C1694513DA0858ULL, "465359585.ogg"},
	{0x58C22F61143E5D3CULL, "hm_vrc3_fake_commander.mor"},
	{0x58C27A79B15E6344ULL, "vxht_chantry_207_0_0.vlm"},
	{0x58C2BF5964A73A59ULL, "llac_dr_dungeon_3214_9.dds"},
	{0x58C2C97F471A37E0ULL, "69471768.ogg"},
	{0x58C312165C2BDAB2ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r.ani"},
	{0x58C31B93DAF00175ULL, "gen00fl_aveline6132736.cl"},
	{0x58C341490CBDB7D6ULL, "vxw_woundedcoast_1d6_0_0.vlm"},
	{0x58C35855288A6E82ULL, "and100_anders6095794.cl"},
	{0x58C3AD61FA75DF01ULL, "mcr310_cut_chantry_blows_up6087551.cl"},
	{0x58C3B3A9EC50A076ULL, "vxlt_0_lowtown_docks_a3_n_aql_0_0.vlm"},
	{0x58C3FA188B25AD1BULL, "hlac_ht_playershouse_n_855_23y.dds"},
	{0x58C4E4A5DAEEC3A0ULL, "vxlt_darktown_5tk_0_0.vlm"},
	{0x58C5359712D2B4D1ULL, "6195145_m.fxe"},
	{0x58C56733B6CB887FULL, "402856796.ogg"},
	{0x58C56A8AC67033E3ULL, "dre000_bodahn6033407.cl"},
	{0x58C5768311D4E97DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_left_walk.evt"},
	{0x58C599796B6937F9ULL, "vxht_mansion_n_4gv_0_0.vlm"},
	{0x58C5F303C83A18CDULL, "515169.ogg"},
	{0x58C5FB079DB5E52DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_forward.gad"},
	{0x58C6215FF62B9354ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_dbl_c.impa.gad"},
	{0x58C71D276215D0CAULL, "art\\characters\\playercharacter\\humanfemale\\glv_medj.mao"},
	{0x58C721276215D719ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medj.mmh"},
	{0x58C72B276215E807ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medj.msh"},
	{0x58C7DE08FE2DD64DULL, "6139811_m.fxe"},
	{0x58C86A310C958A21ULL, "379188988.ogg"},
	{0x58C8B1B2ECD3981FULL, "ss_harlan.wbd"},
	{0x58C8F6824FC4D7EAULL, "27721359.ogg"},
	{0x58C9084AF1F49944ULL, "vxdr_0_deeproads_f6b_0_0.vlm"},
	{0x58C944475FFA4A6BULL, "338805060.ogg"},
	{0x58C97F17C964002AULL, "vxht_mansion_n_6t6_0_0.vlm"},
	{0x58C993F16DAF13CBULL, "vxlt_hangedman_n_25v_0_0.vlm"},
	{0x58C9A557462A5B7FULL, "rock_c_01.dds"},
	{0x58CA1129767E0D99ULL, "dae000_hate_follower_react6103249.cl"},
	{0x58CA240596780FFFULL, "6014697_m.fxe"},
	{0x58CAE5E7ABF1336CULL, "vxlt_0_lowtown_a3_n_c_hcp_0_0.vlm"},
	{0x58CB0017AE2EB0C9ULL, "vxlt_gallowstemplar_f_37j_0_0.vlm"},
	{0x58CB0EE232181354ULL, "vrc221_varric6043872.cl"},
	{0x58CB84C51A3DEA28ULL, "w_woundedcoast_250_0.msh"},
	{0x58CB96C51A3E08B9ULL, "w_woundedcoast_250_0.mao"},
	{0x58CBA2C51A3E1D52ULL, "w_woundedcoast_250_0.mmh"},
	{0x58CC4593F06E4EDDULL, "244916630.ogg"},
	{0x58CC9DAAD4BB66E3ULL, "vxht_bartrand_n_6qk_0_0.vlm"},
	{0x58CD305FF631CB32ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_dbl_c.impa.ani"},
	{0x58CDAB309EB57A9DULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.dura.ani"},
	{0x58CE0C0827CF7F1BULL, "mag101_grace6034237.cl"},
	{0x58CE7B3B76DD0E31ULL, "gen00fl_anders6125630.cl"},
	{0x58CE7ED8E098AA3DULL, "6095803_m.fxe"},
	{0x58CE8DB287EA229CULL, "vxw_woundedcoast_1ld_0_0.vlm"},
	{0x58CE9C7DDB4764CBULL, "mag320_orsino6071082.cl"},
	{0x58CF6895AFB3FB70ULL, "6149615_m.fxe"},
	{0x58CFC9D6ADAC8C57ULL, "vxlt_0_lowtown_n_euc_0_0.vlm"},
	{0x58CFDAC1D0E207A1ULL, "qun131_taskmaster6019174.cl"},
	{0x58D01CA5D555B93CULL, "33702537.ogg"},
	{0x58D02100AD30E7BEULL, "w_blightlands_138_0.phy"},
	{0x58D02E199DCE5F13ULL, "ts_generic_04.gda"},
	{0x58D0919AE84F7F76ULL, "montage_subs_demohack6217129.cl"},
	{0x58D0E54F83DB35DFULL, "t1_mue_bl2.tnt"},
	{0x58D1407F17C78682ULL, "vxht_mansion2_2eb_0_0.vlm"},
	{0x58D16DDD1E0E51D9ULL, "llac_lt_0_lowtown_a3_n_c_-131072_9.dds"},
	{0x58D1A67893D2E5B4ULL, "vxht_keep_6ls_0_0.vlm"},
	{0x58D1B325FEB3D056ULL, "vxw_woundedcoast_1o7_0_0.vlm"},
	{0x58D1BB6BB1A8FB98ULL, "6208666_m.fxe"},
	{0x58D1F02394F24F18ULL, "vxw_encounter2_camp_s5_0_0.vlm"},
	{0x58D25E3FFFADA3BBULL, "hlac_lt_0_lowtown_a2_n_-262144_4y.dds"},
	{0x58D26D9434D77EB6ULL, "6607678.ogg"},
	{0x58D27D6F1E304CA1ULL, "hlac_lt_gallowscourtyard_-65535_1y.dds"},
	{0x58D2F70A1ABA29E0ULL, "183934301.ogg"},
	{0x58D30066FECDA551ULL, "gen00fl_isabela6101357.cl"},
	{0x58D374D314501A93ULL, "ser210_witness6168117.cl"},
	{0x58D39ADF97A4CC43ULL, "vxlt_warehouse_n_bg_0_0.vlm"},
	{0x58D39F3051FFD03AULL, "1008111727.ogg"},
	{0x58D3A125144A315AULL, "hlac_lt_0_lowtown_a2_n_65535_5y.dds"},
	{0x58D3C1111BB321C9ULL, "framebuffereffects_twopassblurcocvertical_p0.psh.fxo"},
	{0x58D43CEB465BD7FFULL, "6057362_m.fxe"},
	{0x58D45BC4322A7818ULL, "169467333.ogg"},
	{0x58D49DEBE419246CULL, "vxlt_undercitydungeon_1mo_0_0.vlm"},
	{0x58D4B93A9EAB0EF4ULL, "vxlt_gallowstemplar_5ra_0_0.vlm"},
	{0x58D57A918F086654ULL, "828593786.ogg"},
	{0x58D586B753F1AB52ULL, "6139768_m.fxe"},
	{0x58D59EF2F85AF30DULL, "vxht_mansion_n_5bl_0_0.vlm"},
	{0x58D616234A85C12EULL, "lowtown_fem_alienage1_sta6162588.cl"},
	{0x58D69F3804378935ULL, "lt_0_lowtown_docks_a3_n_8_0.rml"},
	{0x58D6B93AA5E14A1AULL, "6171139_m.fxe"},
	{0x58D6CFE6CB42A9EDULL, "6056618_f.fxe"},
	{0x58D7E4A865B648DCULL, "hlac_ht_0_hightown_-65537_2z.dds"},
	{0x58D8466375E0519CULL, "seb221_flora_harimann6198463.cl"},
	{0x58D88A92C603FC10ULL, "vxlt_gallowstemplar_n_5rr_0_0.vlm"},
	{0x58D90A5461556CFEULL, "vxw_woundedcoast_to_0_0.vlm"},
	{0x58D92B5CD6F799D5ULL, "ca_caverns1_mines_24425.rml"},
	{0x58D930A228EA9C8EULL, "vxlt_0_lowtown_docks_n_bfi_0_0.vlm"},
	{0x58D9320F2477198FULL, "985487958.ogg"},
	{0x58D9AD4A4426A92EULL, "vxht_keep_n_60j_0_0.vlm"},
	{0x58DA138067238C1BULL, "character_rendertodepth_p3.psh.fxo"},
	{0x58DA5190ED8CA141ULL, "qun221_petrice6076680.cl"},
	{0x58DA6EF8ED602C1DULL, "gen00fl_aveline6114000.cl"},
	{0x58DB0C9CD720D517ULL, "vxlt_warehouse_19b_0_0.vlm"},
	{0x58DBCE6F7BAE0489ULL, "84437571.ogg"},
	{0x58DC374C788AF01CULL, "6137690_m.fxe"},
	{0x58DC8AA43037C651ULL, "one110tr_meer_ambush.utt"},
	{0x58DCA9EF49552425ULL, "6036682_m.fxe"},
	{0x58DD5247AB0DA1FDULL, "456358387.ogg"},
	{0x58DD5C434069388DULL, "6053229_m.fxe"},
	{0x58DD853C42C47C8AULL, "vxlt_gallowstemplar_f_5jh_0_0.vlm"},
	{0x58DDD6FD8B5C296FULL, "llac_lt_hangedman_n_-1_7.dds"},
	{0x58DDDA72A1AA13F0ULL, "6142715_m.fxe"},
	{0x58DEC26DBFD03724ULL, "l_mag111ar_warehouse.wbd"},
	{0x58DF2CE60F4A7E4EULL, "vxht_mansion2_21b_0_0.vlm"},
	{0x58DF35BCCCE98D22ULL, "161085582.ogg"},
	{0x58DFB05B7C5EDD69ULL, "gen_im_arm_hlm_lgt_act2_02.uti"},
	{0x58DFB25C09BD9F02ULL, "vxht_mansion2_2nc_0_0.vlm"},
	{0x58DFB41D74CF75D3ULL, "vxw_encounter3_g5_0_0.vlm"},
	{0x58DFEF3D490A3209ULL, "llac_lt_0_lowtown_a2_n_c_65536_6.dds"},
	{0x58E02502F0A9E8D6ULL, "vxw_woundedcoast_1j8_0_0.vlm"},
	{0x58E04761124206FBULL, "hlac_ca_caverns2_5802_69y.dds"},
	{0x58E0620086E86798ULL, "vxw_encounter2_camp_n2_0_0.vlm"},
	{0x58E0A36F2C2B4B95ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarp.msh"},
	{0x58E0A56F2C2B4F0BULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarp.mmh"},
	{0x58E0B16F2C2B63A8ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarp.mao"},
	{0x58E14E08274BA804ULL, "6036706_m.fxe"},
	{0x58E191A06426CB0BULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_candleflame.mmh"},
	{0x58E23B04D951DA95ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14a.mmh"},
	{0x58E23D04D951DE0BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14a.msh"},
	{0x58E2922AF917A477ULL, "zz_vault_debug6200369.cl"},
	{0x58E2A626E6D96356ULL, "238770487.ogg"},
	{0x58E2F552FF018BCDULL, "mag121_idunna6084895.cl"},
	{0x58E3A05882BC8D32ULL, "hlac_ht_chantry_2534_4y.dds"},
	{0x58E3C6E72EDB372DULL, "vxw_bonepit_16l_0_0.vlm"},
	{0x58E41909F6D4C1F5ULL, "vxca_caverns2_8a4_0_0.vlm"},
	{0x58E42101B155CA4CULL, "hlac_ht_mansion2b_818_36y.dds"},
	{0x58E474AB3E658B86ULL, "vxw_woundedcoast_yp_0_0.vlm"},
	{0x58E47FB848DB08A7ULL, "vxlt_undercityserialkiller_146_0_0.vlm"},
	{0x58E52B1797D7629AULL, "6174239_m.fxe"},
	{0x58E5A9B90B70FEA3ULL, "vxw_bonepitdestroyed_a3_1x1_0_1.vlm"},
	{0x58E5CB504D06D341ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow3base.phy"},
	{0x58E5FA7CFEBEE1BEULL, "dre000_varric6207929.cl"},
	{0x58E612D928451700ULL, "6048145_m.fxe"},
	{0x58E6C5FCA408915EULL, "6126706_m.fxe"},
	{0x58E6CC6B54091741ULL, "ico_1h_axe_4.dds"},
	{0x58E6D30F613B584CULL, "one300_zevran6075392.cl"},
	{0x58E6FC6FB7FA1490ULL, "vxlt_gallowstemplar_n_5us_0_0.vlm"},
	{0x58E734C87CC8C78EULL, "gen00fl_fenris6116532.cl"},
	{0x58E791D713079106ULL, "vxw_bonepit_1qf_0_0.vlm"},
	{0x58E83071BE3173E2ULL, "200605232.ogg"},
	{0x58E8AB939615B178ULL, "55655484.ogg"},
	{0x58E8E67D673C8E0FULL, "120345680.ogg"},
	{0x58E92C16D3FE5098ULL, "vxca_caverns1_jw2_0_0.vlm"},
	{0x58E934D336241452ULL, "vxw_encounter3_g0_0_0.vlm"},
	{0x58E996E0EAF49A25ULL, "6180148_f.fxe"},
	{0x58E9A2D05C75458FULL, "6173823_f.fxe"},
	{0x58E9AA6FD4719196ULL, "zz_vault_debug6203330.cl"},
	{0x58E9D80020038F72ULL, "mrl310_party_comments6143063.cl"},
	{0x58EA6AA1F0B2AED2ULL, "dae000_surprises6157163.cl"},
	{0x58EB6D00397A2565ULL, "6142031_m.fxe"},
	{0x58EB716C6BD6539CULL, "vxw_woundedcoast_1rz_0_0.vlm"},
	{0x58EBA11BC9E5E2FDULL, "vxdr_primevalentrance_1jf_0_0.vlm"},
	{0x58EC0E9C81630779ULL, "lt_0_lowtown_docks_n_161_0.msh"},
	{0x58EC189C81631867ULL, "lt_0_lowtown_docks_n_161_0.mmh"},
	{0x58EC1C9C81631EB4ULL, "lt_0_lowtown_docks_n_161_0.mao"},
	{0x58EC8B1CBD0FCC9FULL, "gen_im_cft_rec_arn_09.uti"},
	{0x58EC8E129ABAEDF5ULL, "46087277.ogg"},
	{0x58ECE14C70222C7DULL, "vxlt_0_lowtown_docks_c59_0_0.vlm"},
	{0x58ED067BA658E6A1ULL, "fen310_varania6118625.cl"},
	{0x58EDB657B3FC35A3ULL, "follower_banter6163417.cl"},
	{0x58EE2438FBB3B782ULL, "vxht_mansion2_2kl_0_0.vlm"},
	{0x58EEA2FF94DC34E5ULL, "hlac_lt_0_lowtown_65534_0y.dds"},
	{0x58EEE25A2421224EULL, "llac_lt_0_lowtown_docks_n_65534_2.dds"},
	{0x58EF1B984C2F4A2DULL, "vxw_woundedcoast_7a_0_0.vlm"},
	{0x58EF25ED0040A3D2ULL, "gen00fl_fenris6105598.cl"},
	{0x58EF34757CDBC292ULL, "qun231cr_varnell_minion8.utc"},
	{0x58EF513B001E7123ULL, "vxlt_0_lowtown_a3_n_c_enx_0_1.vlm"},
	{0x58EFC3E763FB1A0DULL, "vxlt_hangedman_1ql_0_0.vlm"},
	{0x58F06F9CCE6C2A24ULL, "vxw_bonepit_275_0_0.vlm"},
	{0x58F0956768DDF504ULL, "plt_cod_ltr_ser_alrik.nss"},
	{0x58F0E3B2F6972E3FULL, "vxlt_undercitydungeon_1ja_0_0.vlm"},
	{0x58F18B98214C42B1ULL, "llac_lt_gallowstemplar_f_2391_1.dds"},
	{0x58F19CF482973FF4ULL, "vxlt_gallowstemplar_5xg_0_0.vlm"},
	{0x58F1DBFB500F3C1FULL, "vxw_sundermount_30u_0_0.vlm"},
	{0x58F1F0E216363522ULL, "dae000_meredith6085336.cl"},
	{0x58F1F6349AAB7BB5ULL, "follower_banter6151278.cl"},
	{0x58F1FAB590C81554ULL, "vxlt_0_lowtown_docks_a3_n_cz8_0_1.vlm"},
	{0x58F282ACDC47240DULL, "vxht_mansion_n_5hj_0_0.vlm"},
	{0x58F2E688305BA406ULL, "vxw_woundedcoast_vq_0_0.vlm"},
	{0x58F36570579AE440ULL, "977122031.ogg"},
	{0x58F37C0EC296A4CBULL, "vxht_bartrand_n_4gs_0_0.vlm"},
	{0x58F39890C96DF249ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtd.phy"},
	{0x58F3A40A9C730D5DULL, "vxlt_0_lowtown_n_emy_0_0.vlm"},
	{0x58F3A554C7A06FCCULL, "vxlt_0_lowtown_a2_n_efi_0_0.vlm"},
	{0x58F442DD9415E23BULL, "w_woundedcoast_133_0.phy"},
	{0x58F481954E84074BULL, "783828184.ogg"},
	{0x58F4DC6E5281E27DULL, "1072675540.ogg"},
	{0x58F5232E52D27434ULL, "llac_lt_0_lowtown_65536_4.dds"},
	{0x58F56E4CA9F02D10ULL, "vxlt_gallowstemplar_n_5hp_0_0.vlm"},
	{0x58F57C72979285F8ULL, "vxlt_gallowstemplar_3rk_0_0.vlm"},
	{0x58F585C60A3BF5ABULL, "169791594.ogg"},
	{0x58F58C404310C971ULL, "6109229_m.fxe"},
	{0x58F5A7CCB2DB1F51ULL, "534555808.ogg"},
	{0x58F5AA096BC1500FULL, "vxlt_gallowscourtyard_77e_0_0.vlm"},
	{0x58F603B404FDA986ULL, "vxw_bonepit_1zg_0_0.vlm"},
	{0x58F6DE9483F52A83ULL, "282615951.ogg"},
	{0x58F6E6687707B618ULL, "321503521.ogg"},
	{0x58F6EEE4B335A81AULL, "qcr000_arishok6059679.cl"},
	{0x58F7ADCD9D2717DBULL, "6149047_m.fxe"},
	{0x58F80FC9DD0604F8ULL, "242390448.ogg"},
	{0x58F85ABDBA3BB281ULL, "6099021_m.fxe"},
	{0x58F8EBD790D3940BULL, "vxht_0_hightown_a3_c3m_0_0.vlm"},
	{0x58F95E265E99D1FDULL, "6076259_m.fxe"},
	{0x58F9EF62B363427DULL, "dae000_sandal6148239.cl"},
	{0x58FAAB8E189BE0ADULL, "21532051.ogg"},
	{0x58FAC029FB9A81D2ULL, "131649756.ogg"},
	{0x58FB0FC3515FE88EULL, "art\\vfx\\combat\\vfx_r_modal_act_c.mmh"},
	{0x58FB805C7129C9F7ULL, "324241894.ogg"},
	{0x58FB90FA13A85714ULL, "vxca_caverns2prc_92x_0_0.vlm"},
	{0x58FB9F7285C36E3FULL, "413172782.ogg"},
	{0x58FBD157A03FEC9FULL, "isa311_party_comments6117822.cl"},
	{0x58FC36D907713C58ULL, "hlac_lt_0_lowtown_a3_n_c_131072_5x.dds"},
	{0x58FD42D7760628B4ULL, "6176213_m.fxe"},
	{0x58FD45BA6AD49898ULL, "vxw_encounter2_camp_d0_0_0.vlm"},
	{0x58FDF50BE76385B1ULL, "vxlt_gallowsprison_n_6lj_0_0.vlm"},
	{0x58FDF7FFA0BBB2B1ULL, "124039925.ogg"},
	{0x58FE11CB2EB042AEULL, "309126898.ogg"},
	{0x58FE4765A9912EF7ULL, "465753139.ogg"},
	{0x58FE7C76F477C739ULL, "vxlt_0_lowtown_a2_n_e8b_0_0.vlm"},
	{0x58FEDDA6875E2B9AULL, "art\\characters\\weapons\\shields\\textures\\w_shd_elna_0s.dds"},
	{0x58FEF0766D8715C3ULL, "vxlt_warehouse_n_sh_0_0.vlm"},
	{0x58FFAAF300F99131ULL, "vxlt_gallowstemplar_n_65u_0_0.vlm"},
	{0x5900C18A75E5B234ULL, "hlac_lt_0_lowtown_a2_-1_5z.dds"},
	{0x5900C7472A09DF3AULL, "340671410.ogg"},
	{0x5900D00366AE4A48ULL, "mer340_bonepit_comments6090832.cl"},
	{0x5900E5B7167AF3FCULL, "mag291_torpor6027038.cl"},
	{0x590158652251BC86ULL, "vxw_woundedcoast_sr_0_0.vlm"},
	{0x590195214347505BULL, "1003797097.ogg"},
	{0x590197C35DD05285ULL, "6044474_m.fxe"},
	{0x590298EDFEA6FA71ULL, "vxlt_backalley_g_1vs_0_1.vlm"},
	{0x5902D1E121452605ULL, "hm_merchant_guard.mor"},
	{0x590318B512F0F334ULL, "6095015_m.fxe"},
	{0x5903468B913C622AULL, "vxht_mansion_n_6h2_0_0.vlm"},
	{0x59036B2EA3FFA492ULL, "hlac_lt_0_lowtown_a2_n_c_1_2y.dds"},
	{0x5903E0299BE64590ULL, "vxlt_gallowstemplar_n_5cq_0_0.vlm"},
	{0x59041BE65DB7688FULL, "vxlt_gallowscourtyard_78d_0_0.vlm"},
	{0x59042866B5EA2467ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy.gad"},
	{0x590434866D9BFCD5ULL, "480900550.ogg"},
	{0x5904DA2734BE30B7ULL, "6172250_m.fxe"},
	{0x590539C23C10F7C0ULL, "vxht_keep_6y4_0_0.vlm"},
	{0x5905B6150B2F9332ULL, "mag101_grace6034195.cl"},
	{0x5905F8279AE0DC67ULL, "6089619_m.fxe"},
	{0x5906A3286C4BCFA0ULL, "dae380st_sibling.stg"},
	{0x5906A9F1C7116B0BULL, "747991330.ogg"},
	{0x5906AAA7CF4BEFACULL, "6129068_m.fxe"},
	{0x5906D63116F5635AULL, "vxca_caverns1_ijx_0_0.vlm"},
	{0x5906DE7BFD81C491ULL, "hlac_ht_0_hightown_a2_65533_0x.dds"},
	{0x5906E4F20917882EULL, "vxht_mansion_n_6n0_0_0.vlm"},
	{0x5906FA5670E23090ULL, "qcr000_arishok6087515.cl"},
	{0x5907AE9E1459AB01ULL, "dae000_porfiria6155285.cl"},
	{0x5907B4CA3169DC69ULL, "vxlt_gallowstemplar_n_5yz_0_0.vlm"},
	{0x59081B9FD0FCCE4CULL, "6142046_f.fxe"},
	{0x5908E0EC0B904947ULL, "40255724.ogg"},
	{0x59092EC3516C5A8DULL, "art\\vfx\\combat\\vfx_r_modal_act_c.anb"},
	{0x59096353A594678CULL, "hlac_lt_0_lowtown_a2_n_-3_6z.dds"},
	{0x590972F10A2EF687ULL, "mcr371_idunna.bnk"},
	{0x590A1D66B5EE7D74ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy.evt"},
	{0x590A3733A56FC509ULL, "vxlt_undercityserialkiller_171_0_0.vlm"},
	{0x590A75ED1C9BACD4ULL, "vxlt_house_dl_0_0.vlm"},
	{0x590A76B6CD888536ULL, "6147790_m.fxe"},
	{0x590A77C4899A0B7DULL, "dr_slavers_3217.rml"},
	{0x590AAB53D1649C0FULL, "isa211st_foundry.stg"},
	{0x590B1EEE61077463ULL, "one120_athenril6088914.cl"},
	{0x590B2785011BB61FULL, "12821654.ogg"},
	{0x590B998088C588B2ULL, "700815872.ogg"},
	{0x590BACC25E252AF4ULL, "hla_ht_0_hightown_a2_131070_0z.dds"},
	{0x590C1563D2F35E6FULL, "llac_lt_gallowstemplar_f_2394_2.dds"},
	{0x590C775D8EA595C9ULL, "912233261.ogg"},
	{0x590C94F10A3166FFULL, "mcr371_idunna.cnv"},
	{0x590CEE0823199CC2ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb104.mao"},
	{0x590D1C1FE4FC7D7DULL, "gen00fl_anders6096532.cl"},
	{0x590D5E071DAA8D92ULL, "571659524.ogg"},
	{0x590D62535F7D2E43ULL, "vxlt_warehouse_n_vk_0_0.vlm"},
	{0x590DB1E33EE69A96ULL, "6114031_m.fxe"},
	{0x590DCDA8DCBF1B2FULL, "6126407_f.fxe"},
	{0x590E1A067EB62C42ULL, "dae000_flemeth6001898.cl"},
	{0x590E4BE371887396ULL, "w_encounter3_105_0.mao"},
	{0x590E57E371888835ULL, "w_encounter3_105_0.mmh"},
	{0x590E59E371888BABULL, "w_encounter3_105_0.msh"},
	{0x590E8B99425154BFULL, "hlac_lt_0_lowtown_a2_n_c_-131072_9y.dds"},
	{0x590EBF509CC1DBDCULL, "one300_nuncio6076717.cl"},
	{0x590F543719E44E8EULL, "6097437_m.fxe"},
	{0x590F990DB974FC9BULL, "6148196_m.fxe"},
	{0x590FA925FC3926DDULL, "124701394.ogg"},
	{0x590FCF0658BAFC2EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock5a.phy"},
	{0x590FF3D14235F069ULL, "6163975_m.fxe"},
	{0x59104204E51D0F8CULL, "vxw_bonepitmining_a2_29q_0_0.vlm"},
	{0x59105FC8A682D5CBULL, "vxht_bartrand_n_4qu_0_0.vlm"},
	{0x59110CD8E75C055BULL, "ability_core.ncs"},
	{0x591156EF8FF152A2ULL, "atl_prestartme_dxt5_00.dds"},
	{0x591167E0A7728482ULL, "504540383.ogg"},
	{0x5911F363B734284BULL, "dae200cr_viscount6087203.cl"},
	{0x59120BBC85ADD8AAULL, "336269508.ogg"},
	{0x59122C889520DBCCULL, "vxlt_warehouse_v9_0_0.vlm"},
	{0x59123C8F9E09B6FBULL, "6172975_m.fxe"},
	{0x5912966768FADDFFULL, "plt_cod_ltr_ser_alrik.plo"},
	{0x591306A4371BB70DULL, "vxlt_gallowsprison_n_5qf_0_0.vlm"},
	{0x5913282987A58124ULL, "588760796.ogg"},
	{0x59136A20465D69E0ULL, "gen00fl_bethany6101268.cl"},
	{0x5913AB9F2E071040ULL, "vxht_keep_6b5_0_0.vlm"},
	{0x5913AD3EAF0D8F39ULL, "hlac_lt_0_lowtown_n_131070_1z.dds"},
	{0x5913EC7DA7F08360ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_wic_0s.dds"},
	{0x59140F32118591BDULL, "6063077_m.fxe"},
	{0x5914A117006D7FB1ULL, "955206086.ogg"},
	{0x5914A36057DBA7E8ULL, "hlac_w_woundedcoast_-262144_0y.dds"},
	{0x5914BD643A8A1422ULL, "zz_slate1.cut"},
	{0x59150D2D0AAD0443ULL, "6048813_f.fxe"},
	{0x5915B18EB6E0EEB6ULL, "plt_lgt311pt_shw_mer_base_1.nss"},
	{0x5915D2B560FE8008ULL, "mag311_keran6041704.cl"},
	{0x5915F00F0DE8C3FAULL, "vxlt_gallowscourtyard_n_6w6_0_0.vlm"},
	{0x59165D358F5F3FACULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink_explore.ani"},
	{0x59165D7686980D84ULL, "vxlt_gallowstemplar_f_478_0_0.vlm"},
	{0x59167111A6E1773BULL, "604039953.ogg"},
	{0x5917FE44516593EFULL, "6128975_f.fxe"},
	{0x59181566B5FAACADULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_rdy.ani"},
	{0x5918673FDFE3DE7AULL, "vxht_chantry_26k_0_0.vlm"},
	{0x5918A55C1FBDEDD1ULL, "168239013.ogg"},
	{0x5918C1401C7E5108ULL, "638971758.ogg"},
	{0x5918E51B0CC7DD62ULL, "isa211_isabela6069398.cl"},
	{0x5919159CA28F2C9BULL, "vxca_caverns2_8o8_0_0.vlm"},
	{0x591985C0CD2699EEULL, "6127994_m.fxe"},
	{0x59198BEA16929D3FULL, "6026202_f.fxe"},
	{0x5919AA8450AB5C13ULL, "6083532_m.fxe"},
	{0x5919AB720678C9A8ULL, "ave121_brennan6105825.cl"},
	{0x591A066ABB8EC24AULL, "6107682_m.fxe"},
	{0x591B731DF3BCF0BBULL, "gen00fl_anders6111334.cl"},
	{0x591BACA0E8F85F8AULL, "67854251.ogg"},
	{0x591BED2F8D381CF4ULL, "589812274.ogg"},
	{0x591C464668502925ULL, "6175327_m.fxe"},
	{0x591C75C98E11786FULL, "vxlt_0_lowtown_byw_0_0.vlm"},
	{0x591C8477C12DAC1AULL, "one280_emrys6116814.cl"},
	{0x591D059CBF5FF9FBULL, "plt_seb221pt_repentance.nss"},
	{0x591D3A2058D1B969ULL, "llac_lt_0_lowtown_docks_a3_n_0_1.dds"},
	{0x591D6A358F6574EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink_explore.gad"},
	{0x591E0A78CCA82DE3ULL, "hlac_ht_0_hightown_-131072_1z.dds"},
	{0x591E0E3A0A48A17FULL, "zz_amb_joe.utc"},
	{0x591E2C85CD418BDCULL, "277620170.ogg"},
	{0x591E9BAC55D5A60FULL, "tal_thre_skef.mao"},
	{0x591ED7C88AFFF954ULL, "852223145.ogg"},
	{0x591EE9C7472CC55FULL, "585240553.ogg"},
	{0x591EFAEB9D82B94CULL, "vxlt_0_lowtown_a2_n_eql_0_0.vlm"},
	{0x591F56B160296FFAULL, "mag291_demon_marethari6029028.cl"},
	{0x591F73BB1CC00EA3ULL, "mrl000_marethari6068808.cl"},
	{0x591F7CA7E2932B71ULL, "vxlt_backalley_g_1xq_0_1.vlm"},
	{0x591F8375A1010B12ULL, "vxlt_0_lowtown_a2_he8_0_0.vlm"},
	{0x591F8A9C42F6F432ULL, "vxlt_0_lowtown_docks_a3_bsb_0_0.vlm"},
	{0x591FC20F2C9A47DBULL, "176214280.ogg"},
	{0x5920165DDDC19419ULL, "follower_banter6170752.cl"},
	{0x592029DD943AF2DAULL, "w_woundedcoast_133_0.msh"},
	{0x592037DD943B0A13ULL, "w_woundedcoast_133_0.mao"},
	{0x59203BDD943B1120ULL, "w_woundedcoast_133_0.mmh"},
	{0x59206328D435AF52ULL, "982157253.ogg"},
	{0x59208190C994B96EULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtd.mmh"},
	{0x59209B90C994E55CULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtd.msh"},
	{0x5920C88C01EFD47BULL, "vxw_woundedcoast_6h_0_0.vlm"},
	{0x592153E0C4B06848ULL, "131975745.ogg"},
	{0x5921ADB0A04BC7B9ULL, "689878269.ogg"},
	{0x5921AF62893D7010ULL, "vxdr_0_deeproads_d37_0_0.vlm"},
	{0x5921C9C822B63B92ULL, "898954933.ogg"},
	{0x592262550DFD0ED5ULL, "6125515_m.fxe"},
	{0x5922851EB4C0684EULL, "228160296.ogg"},
	{0x5922C720A469C25BULL, "614490599.ogg"},
	{0x5922C954B308DA53ULL, "lgt000im_ultra_finger.uti"},
	{0x5922FCAB5AD769DFULL, "dae000_pc_home_ambients6208337.cl"},
	{0x5923D40413FA3785ULL, "vxlt_gallowstemplar_3rr_0_0.vlm"},
	{0x5923F0CF9FFBC86EULL, "190715793.ogg"},
	{0x592408E151CEF3C1ULL, "1072022177.ogg"},
	{0x592410522C455204ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basicnb.msh"},
	{0x592422522C45716DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basicnb.mao"},
	{0x592426522C4577A6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basicnb.mmh"},
	{0x592465358F6B8ACDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\potion_drink_explore.evt"},
	{0x59249DE5D8264CE8ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.anb"},
	{0x5924A76032BBEE02ULL, "l_env_wal_lowtown_night_dist_oneshots.bnk"},
	{0x5924B2F1BDB0A30DULL, "6180903_m.fxe"},
	{0x5924FE11C1B84E5BULL, "vxca_cavern1_haunted_jg0_0_0.vlm"},
	{0x5925272CA3A121A4ULL, "zz_lgt_2_debug6211852.cl"},
	{0x5925AA65465D6D31ULL, "vxht_mansion_n_4vi_0_0.vlm"},
	{0x5925C9A82FA487DCULL, "873759944.ogg"},
	{0x5925CF1CC5F741EFULL, "32008482.ogg"},
	{0x5925DD464CA06FA9ULL, "6208542_m.fxe"},
	{0x592622A9D770EF1AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1a.mmh"},
	{0x592626A9D770F651ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1a.mao"},
	{0x592630CCD916C213ULL, "913699558.ogg"},
	{0x592634A9D7710E60ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1a.msh"},
	{0x592772CB51B03FF5ULL, "vxht_keep_36a_0_0.vlm"},
	{0x5927B5CED5CEA5E7ULL, "778915493.ogg"},
	{0x5927DBDD3782E610ULL, "622140777.ogg"},
	{0x592842473EE7B910ULL, "lt_0_lowtown_docks_a3_159_0.msh"},
	{0x5928458E684AFAE6ULL, "472797040.ogg"},
	{0x592850264D08D0CCULL, "zz_bet_debug.cnv"},
	{0x592850473EE7D12AULL, "lt_0_lowtown_docks_a3_159_0.mmh"},
	{0x592854473EE7D861ULL, "lt_0_lowtown_docks_a3_159_0.mao"},
	{0x59287DF9C234EF60ULL, "hlac_lt_house_-1_4x.dds"},
	{0x5928C46B88B05A76ULL, "hlac_ht_keep_3732_18y.dds"},
	{0x5928FCF2D0082CC8ULL, "hlac_lt_0_lowtown_docks_-3_3y.dds"},
	{0x5929131B66D51DC1ULL, "lgt310_mag_melson6152113.cl"},
	{0x592922276268A306ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medj.phy"},
	{0x59293AF82CE19E46ULL, "74939301.ogg"},
	{0x59296E51D2424745ULL, "mag371_elsa6041415.cl"},
	{0x5929FDF962D8370BULL, "vxdr_0_deeproads_dub_0_0.vlm"},
	{0x592A49DC60D7F70DULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyo_0n.dds"},
	{0x592A9EFCE2689B3BULL, "llac_lt_undercityencounter_-1_1.dds"},
	{0x592AD366A08B4EC6ULL, "dae000_bartrand6027534.cl"},
	{0x592C9F7BCD1345A5ULL, "6147128_m.fxe"},
	{0x592CC4380E74F814ULL, "gui_guipixelshader_p8.psh.fxo.dx11"},
	{0x592D418FD29E9B44ULL, "489262696.ogg"},
	{0x592D4E0ED39B0AC4ULL, "6118171_f.fxe"},
	{0x592DF55292F72392ULL, "vxlt_0_lowtown_a2_hb9_0_0.vlm"},
	{0x592E1D522C4DBD71ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basicnb.phy"},
	{0x592E811155BF0EEBULL, "hlac_lt_0_lowtown_docks_-1_1z.dds"},
	{0x592E99B80F2319E2ULL, "132661065.ogg"},
	{0x592EC1F3187FCCC5ULL, "mag100_thrask6031993.cl"},
	{0x592EFB9FAA8ADF48ULL, "dre000_carver6037627.cl"},
	{0x592F24FAC0BE67B3ULL, "983869457.ogg"},
	{0x592F3A68F3E5ECFBULL, "vxw_woundedcoast_3k_0_0.vlm"},
	{0x592F679E15756231ULL, "sib100im_giftcar_letters.uti"},
	{0x592F82040FB6A2CDULL, "500196736.ogg"},
	{0x592F8C95CDF42DE9ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyn_0n.dds"},
	{0x592FC0A71C7D59B0ULL, "qcr000_arishok6059963.cl"},
	{0x59307EA91CD46E24ULL, "874431913.ogg"},
	{0x59307F2AE835D105ULL, "6155787_m.fxe"},
	{0x59308F5911F34140ULL, "vxht_keep_6h3_0_0.vlm"},
	{0x59318BA4308007A4ULL, "one110tr_meer_ambush.ncs"},
	{0x5931F4F14D45DA4FULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_c.mmh"},
	{0x59320FBF19C7CCE8ULL, "gen00fl_anders6111289.cl"},
	{0x59321D39F84AF090ULL, "dae000_bodhan6146901.cl"},
	{0x5932590A0D398C9BULL, "fex_ruin_arch_lrg_d.mmh"},
	{0x59325E5CF4DC354AULL, "follower_barks6153379.cl"},
	{0x5932770A0D39BF85ULL, "fex_ruin_arch_lrg_d.msh"},
	{0x5932C70D711422C4ULL, "6169767_m.fxe"},
	{0x5933B1ACE0CCAF5FULL, "llac_dr_dungeon_3216_1.dds"},
	{0x5933FB1C7A6074B6ULL, "lt_0_lowtown_docks_n_18_0_lc.mmh"},
	{0x59349E896E214677ULL, "947053852.ogg"},
	{0x5934AF2424F1CC5BULL, "dae000_flemeth6000692.cl"},
	{0x5934B1531B91DF3EULL, "6033765_m.fxe"},
	{0x5935587BE6349CADULL, "lt_0_lowtown_docks_a3_n_207_0.msh"},
	{0x5935667BE634B4E0ULL, "lt_0_lowtown_docks_a3_n_207_0.mao"},
	{0x59356A7BE634BB33ULL, "lt_0_lowtown_docks_a3_n_207_0.mmh"},
	{0x5935D0C64ED58040ULL, "vxht_0_hightown_n_bg9_0_0.vlm"},
	{0x59361759651DBC28ULL, "vxca_caverns2_8gg_0_0.vlm"},
	{0x59361E325D79918CULL, "161137999.ogg"},
	{0x59365D851880E3D0ULL, "250552597.ogg"},
	{0x5936697AB112CAEEULL, "6129794_m.fxe"},
	{0x5936809A2F973965ULL, "707155369.ogg"},
	{0x5936BF9F24C2D33CULL, "6118294_f.fxe"},
	{0x5936CF9294859BEEULL, "exptable.gda"},
	{0x593713D7F44D92D9ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_anglel.ani"},
	{0x593725912352A533ULL, "llac_lt_darktown_-1_78.dds"},
	{0x59379D6F684CAC6DULL, "6019205_m.fxe"},
	{0x5937C0F2C2F938F8ULL, "w_encounter3_114_0.phy"},
	{0x5937DDFEDC4C0FFFULL, "vxlt_0_lowtown_a2_n_hg8_0_0.vlm"},
	{0x5938665F0C25D621ULL, "vxht_keep_6a1_0_0.vlm"},
	{0x593897EDB2762C6FULL, "hlac_ca_caverns1_mines_24424_10z.dds"},
	{0x5938D0D76F90B193ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc104a_0s.dds"},
	{0x5938DB5925253EC7ULL, "zz_vault_debug6203197.cl"},
	{0x5938E9E162DBD38DULL, "6143313_m.fxe"},
	{0x5939180EE9B47319ULL, "art\\vfx\\combat\\vfx_sg_quake_act_c.impa.ani"},
	{0x5939ECC984DC543CULL, "6097884_m.fxe"},
	{0x593A0C0105F1C93BULL, "character_character_p73.psh.fxo.dx11"},
	{0x593A3F85D2F05825ULL, "6134683_m.fxe"},
	{0x593AB8FC4A79B1A8ULL, "hlac_lt_0_lowtown_docks_-262144_0z.dds"},
	{0x593AD66BBAFC1E22ULL, "6102884_f.fxe"},
	{0x593AE2EB3DFCAD64ULL, "408872665.ogg"},
	{0x593AEE6E6D777995ULL, "vxht_keep_6m9_0_0.vlm"},
	{0x593BCF17CBD76658ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_grand_1.mao"},
	{0x593BDF2A14566F5EULL, "407664863.ogg"},
	{0x593C0A5F80AF002DULL, "vxw_bonepitmining_a2_1ms_0_0.vlm"},
	{0x593CD7EBB4FE5D67ULL, "1064789143.ogg"},
	{0x593D0D9C129E239EULL, "vxlt_warehouse_ct_0_0.vlm"},
	{0x593D76C489C56295ULL, "dr_slavers_3217.anb"},
	{0x593DC2C1B906E075ULL, "429016531.ogg"},
	{0x593DEED7F453721BULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_anglel.gad"},
	{0x593E5AEEC785441CULL, "399067635.ogg"},
	{0x593E707BB48DA0B5ULL, "hlac_w_bonepitdestroyed_a3_327683_0x.dds"},
	{0x593E7F433C04E76CULL, "vxht_brothel_n_6k2_0_0.vlm"},
	{0x593EC2F02A0B6FEFULL, "vxw_encounter2_camp_pv_0_1.vlm"},
	{0x593F2B324132AA25ULL, "65506210.ogg"},
	{0x593F57AABBD847DCULL, "w_bonepitmining_a2_229_0.mao"},
	{0x593F63AABBD85C7FULL, "w_bonepitmining_a2_229_0.mmh"},
	{0x593F69AABBD86621ULL, "w_bonepitmining_a2_229_0.msh"},
	{0x593F9623F2E1FD0DULL, "965978906.ogg"},
	{0x593FBD2F00577A5AULL, "genip_corpse2.utp"},
	{0x593FF30EE9BA535BULL, "art\\vfx\\combat\\vfx_sg_quake_act_c.impa.gad"},
	{0x593FFD01F8AE4CCEULL, "1025506067.ogg"},
	{0x594055234BFF5EECULL, "6125925_m.fxe"},
	{0x59405CBDFA085B17ULL, "898901581.ogg"},
	{0x59406E6F39C4F4FAULL, "58259928.ogg"},
	{0x5940D17297C8F751ULL, "lowtown_fem_prostitute_sta6152322.cl"},
	{0x5940DB59B9D93E08ULL, "llac_w_sundermount_262139_0.dds"},
	{0x59410F93B05D37D0ULL, "w_bonepit_70_0.mao"},
	{0x59411393B05D3EE3ULL, "w_bonepit_70_0.mmh"},
	{0x59412193B05D563DULL, "w_bonepit_70_0.msh"},
	{0x594187580E15957DULL, "6070489_f.fxe"},
	{0x5941F3835A934356ULL, "761045162.ogg"},
	{0x594206A2CD43FCC1ULL, "qun220cr_javaris6071861.cl"},
	{0x594236B6B08DE50FULL, "vxht_mansion2_2b6_0_0.vlm"},
	{0x594324A2B6B88C5EULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.impa.ani"},
	{0x59432873E874366DULL, "lgt000cr_dwf_turn_in_3.utc"},
	{0x59432AB3C69F4177ULL, "118506592.ogg"},
	{0x5943467F71B50369ULL, "626247637.ogg"},
	{0x5943E3C8F5D714EAULL, "6179272_m.fxe"},
	{0x594442A8478D631BULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_book_.ani"},
	{0x59444548823225C4ULL, "hlac_ca_caverns1_mines_24426_3y.dds"},
	{0x5944613F880E2F3BULL, "lowtown_male_lowlife1_sta6165000.cl"},
	{0x594461DDCC15A99BULL, "vxht_0_hightown_bk1_0_0.vlm"},
	{0x5944694D5CBBE358ULL, "vxdr_0_deeproads_fio_0_0.vlm"},
	{0x59446D9883D9446CULL, "566542794.ogg"},
	{0x594489365713D4A8ULL, "vxca_caverns2_8bd_0_0.vlm"},
	{0x5944EDD2F337C250ULL, "w_sundermount_62_0.rml"},
	{0x59453B7908B44714ULL, "476480054.ogg"},
	{0x5945B36031DE667EULL, "vxca_caverns1_edn_0_0.vlm"},
	{0x5945CD81B4CF50A5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobb_0n.dds"},
	{0x59460E686C1A79E5ULL, "754397404.ogg"},
	{0x594617F01DAD2A4CULL, "vxca_caverns1_mines_i7d_0_0.vlm"},
	{0x5946228A7BB872B3ULL, "fen310_varania6118605.cl"},
	{0x59467658755F84CCULL, "qcr220_qunari_hunter.cnv"},
	{0x5946D83BFE1BEEA1ULL, "vxht_keep_6j0_0_0.vlm"},
	{0x594708F1A7BE9196ULL, "fx_energy03.mao"},
	{0x5947100EE9C0A340ULL, "art\\vfx\\combat\\vfx_sg_quake_act_c.impa.evt"},
	{0x5947459214C9F2F8ULL, "6078857_f.fxe"},
	{0x59478D0CEB49BBC4ULL, "6071718_m.fxe"},
	{0x5947BA26C5E44C9CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_004a_0d.dds"},
	{0x5947D0504D59A67DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow3base.mao"},
	{0x5947D4504D59ACB6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow3base.mmh"},
	{0x5947DE504D59BDB4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow3base.msh"},
	{0x5947EC5067BCD60EULL, "585661900.ogg"},
	{0x5948454ED320A1E0ULL, "vxca_caverns2_4x3_0_1.vlm"},
	{0x5948ABAD4616A9E1ULL, "211558177.ogg"},
	{0x594918264C6B9D49ULL, "dae300_alistair_king6133627.cl"},
	{0x59495858756187FCULL, "qcr220_qunari_hunter.bnk"},
	{0x59495EF88D8F8013ULL, "art\\characters\\creatures\\strider\\textures\\c_stridera_0d.dds"},
	{0x594970B3A61A7981ULL, "isa211_isabela6069452.cl"},
	{0x5949916032DAE520ULL, "l_env_wal_lowtown_night_dist_oneshots.wbd"},
	{0x5949D97BE63FA166ULL, "hlac_lt_backalley_n_0_0y.dds"},
	{0x594A2476D92E1E24ULL, "vxw_encounter2_camp_qv_0_1.vlm"},
	{0x594A58187E72F8BCULL, "llac_w_sundermount_-65535_0.dds"},
	{0x594A7C3C72A518ADULL, "vxw_bonepitmining_a2_1jr_0_0.vlm"},
	{0x594A8EE4A446C880ULL, "6084686_m.fxe"},
	{0x594AA066C39E3476ULL, "lt_0_lowtown_a2_n_21_0.rml"},
	{0x594AB29357CBF5F8ULL, "792059448.ogg"},
	{0x594AB2B7BEED6E90ULL, "hlac_ht_mansion2_818_16y.dds"},
	{0x594ACF945245004CULL, "6171854_m.fxe"},
	{0x594B090F739B75C4ULL, "vxdr_0_deeproads_fak_0_0.vlm"},
	{0x594B41A9D7904025ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1a.phy"},
	{0x594B5D7569394713ULL, "vxlt_hangedman_25e_0_0.vlm"},
	{0x594BADD2B344A954ULL, "zz_dog_debug6127733.cl"},
	{0x594BAF94DD589D39ULL, "ht_0_hightown_a3_42_0.anb"},
	{0x594C109CBF877C94ULL, "plt_seb221pt_repentance.plo"},
	{0x594CFAB6F351E737ULL, "pri_dot.mao"},
	{0x594D0ABC1AA26F0DULL, "vxw_encounter2_camp_9r_0_0.vlm"},
	{0x594D117B7114AC22ULL, "gen_im_trs_krk_arc_sbw_r0.uti"},
	{0x594D3F92A872D1AFULL, "plt_dae00pt_premium.plo"},
	{0x594D9CE5D8494F83ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.mmh"},
	{0x594DB7FC2732B0DBULL, "6100282_m.fxe"},
	{0x594F180941457BEBULL, "vxlt_0_lowtown_docks_n_8rn_0_0.vlm"},
	{0x594F6079BBF34DE3ULL, "lgt110_urn_barker6132981.cl"},
	{0x594F890248BA82E5ULL, "vxht_keep_6gk_0_0.vlm"},
	{0x594FBA626EE82474ULL, "lt_0_lowtown_docks_a3_37_0.rml"},
	{0x594FC8315E8A69BEULL, "6097593_f.fxe"},
	{0x594FCE0658F0DEF1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock5a.mmh"},
	{0x594FDA0658F0F292ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock5a.mao"},
	{0x594FE80658F10AAFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock5a.msh"},
	{0x59500F438FBCE031ULL, "one110cr_coterie_thug_3.utc"},
	{0x59503CA2B6C33F2FULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.impa.evt"},
	{0x5950C7AC7B139A89ULL, "m_mag211ar_ariannis_house.bnk"},
	{0x5950E6974AD625DDULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm04.phy"},
	{0x59511DB70EEE8506ULL, "vxca_caverns1_ego_0_0.vlm"},
	{0x595133B2F69086B0ULL, "91037648.ogg"},
	{0x59513F7B07E2D875ULL, "mcr371_scared_mage_f6086924.cl"},
	{0x59519CD9008F76DDULL, "vxw_bonepitmining_a2_f9_0_0.vlm"},
	{0x5951D0C0D64E8645ULL, "vxht_chantry_23o_0_0.vlm"},
	{0x5951ED5BE630F879ULL, "hlac_lt_0_lowtown_a2_n_c_-4_0z.dds"},
	{0x59521F0392F13253ULL, "639488939.ogg"},
	{0x59525DA1F5BFDF06ULL, "hlac_lt_house_-1_5y.dds"},
	{0x595392691C30F40BULL, "70477446.ogg"},
	{0x595394BC3AAC3363ULL, "mag121_idunna6092340.cl"},
	{0x59542D264D2DD05FULL, "zz_bet_debug.ncs"},
	{0x5954C7FE67B2581DULL, "vxca_cave_255_0_0.vlm"},
	{0x59555742069E2C2FULL, "zz_crowd_1.utc"},
	{0x595613D60CAF19DEULL, "vxca_cave_2gh_0_0.vlm"},
	{0x59572C857FB54A86ULL, "92638439.ogg"},
	{0x59573DA2B6C95FD8ULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.impa.gad"},
	{0x59578B54B1DF38A6ULL, "683629989.ogg"},
	{0x5957E98555EFAEC8ULL, "815429928.ogg"},
	{0x59583BEC3E1B4B97ULL, "818911067.ogg"},
	{0x595869A8479E4EA5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_book_.gad"},
	{0x5958815F62C71380ULL, "fx_0014_diff.dds"},
	{0x5958FCE1832291D1ULL, "vxht_0_hightown_n_c4d_0_0.vlm"},
	{0x5959238F9A6BB6CAULL, "vxdr_dungeon_2hg_0_1.vlm"},
	{0x5959B7DF26DDC3D4ULL, "ser121_jethann6108794.cl"},
	{0x5959CF525B2F5F93ULL, "vxlt_hangedman_22d_0_0.vlm"},
	{0x595A54D079B424F1ULL, "vxca_caverns1_mines_ihx_0_0.vlm"},
	{0x595AF5F14D68E08CULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_c.anb"},
	{0x595B40F337A04DCDULL, "vxca_cavern1_haunted_ghr_0_0.vlm"},
	{0x595B7C990C98878DULL, "vxw_encounter2_camp_4q_0_0.vlm"},
	{0x595BA9B1A8AA9B85ULL, "vxw_encounter2_b7_0_0.vlm"},
	{0x595C0B6614304343ULL, "gen00fl_varric6080441.cl"},
	{0x595C827920AEF011ULL, "gen00fl_dog6165723.cl"},
	{0x595C869B4D61E41AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbench_n.dds"},
	{0x595D63A5D0942F0AULL, "vxlt_gallowstemplar_n_5qm_0_0.vlm"},
	{0x595D861DF60CBCEDULL, "hlac_ht_bartrand_n_3381_1x.dds"},
	{0x595EAC4A33F9A50AULL, "667398557.ogg"},
	{0x595F284EDD0012C1ULL, "hlac_dr_slavers_3217_8x.dds"},
	{0x595F5FD08C729F6FULL, "6174869_m.fxe"},
	{0x595F71F44B55FC11ULL, "647188524.ogg"},
	{0x595F8F9400E49D86ULL, "vxca_caverns1_ebp_0_0.vlm"},
	{0x595FD6E17132E350ULL, "vxdr_primevaltomb_6z8_0_0.vlm"},
	{0x596008F8CB6A48AEULL, "art\\characters\\playercharacter\\humanfemale\\cth_noba.phy"},
	{0x5960944474EE6242ULL, "159780369.ogg"},
	{0x5961865FA428C538ULL, "275999960.ogg"},
	{0x5961E7974AE49AE2ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm04.mmh"},
	{0x5961E9974AE49D98ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm04.msh"},
	{0x59623F26DDF2B135ULL, "121338904.ogg"},
	{0x59629889297ECD94ULL, "hlac_ht_chantry_n_2534_45y.dds"},
	{0x5962F05501994717ULL, "383755163.ogg"},
	{0x59634DB96E9319ABULL, "lt_0_lowtown_a3_n_c_33_0_lc.mmh"},
	{0x59635614039DFDE0ULL, "6165200_m.fxe"},
	{0x596369B5A3371C30ULL, "t3_wdg_w16.tnt"},
	{0x5963C59F2C19B94AULL, "vxlt_0_lowtown_a2_n_f4r_0_0.vlm"},
	{0x596473346FE49655ULL, "6067684_m.fxe"},
	{0x5964BCD53AC86888ULL, "hlac_w_encounter3_-196608_0x.dds"},
	{0x59650DC86802EBDEULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.cess.evt"},
	{0x59653CFB7BC6ADB9ULL, "vxlt_warehouse_a5_0_0.vlm"},
	{0x596600FCEB24DCCBULL, "871244193.ogg"},
	{0x596619F1A7D95F18ULL, "fx_energy03.dds"},
	{0x59661D86961A2E9EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1ww_basic_2.mao"},
	{0x59663E04302DF266ULL, "vxht_keep_649_0_0.vlm"},
	{0x596707A321FFF604ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_act_c.mmh"},
	{0x5967C05993EB318DULL, "mage_repulsionwave.dds"},
	{0x5967C5DF1A3ADDCBULL, "genip_gguardian_lowerhalf.utp"},
	{0x59680810D5284040ULL, "vxlt_0_lowtown_e5w_0_0.vlm"},
	{0x596815EE0050ED9DULL, "201627634.ogg"},
	{0x59683DB3E7E08EF9ULL, "dae000_pc_home_ambients6208304.cl"},
	{0x59686332E8806D1EULL, "vxlt_warehouse_rw_0_0.vlm"},
	{0x59687235E965C7C7ULL, "823207005.ogg"},
	{0x5968A98BA737AF6FULL, "178586491.ogg"},
	{0x5968AB92A568158CULL, "vxdr_0_deeproads_bb2_0_0.vlm"},
	{0x5969205863737CA5ULL, "6201649_f.fxe"},
	{0x59694EF495B93AE4ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.anb"},
	{0x59696A123C74F1F3ULL, "248561960.ogg"},
	{0x5969B8C637118A86ULL, "6135966_m.fxe"},
	{0x596A1AB05BD81311ULL, "vxlt_0_lowtown_a3_n_c_dbb_0_0.vlm"},
	{0x596A1B8E9AA0B405ULL, "vxw_encounter2_m6_0_0.vlm"},
	{0x596A275861682F63ULL, "hlac_lt_gallowstemplar_2391_8z.dds"},
	{0x596B237912DAF884ULL, "vxlt_gallowscourtyard_6yy_0_0.vlm"},
	{0x596B3597B7CAF0D2ULL, "vxw_encounter3_89_0_0.vlm"},
	{0x596B37EE6878A2BCULL, "6064679_m.fxe"},
	{0x596B628675DABDDBULL, "6062991_m.fxe"},
	{0x596B7F778B438878ULL, "llac_lt_0_lowtown_docks_a3_n_-3_0.dds"},
	{0x596BB64F8B251676ULL, "vxht_mansion2b_1xf_0_0.vlm"},
	{0x596C24C868093109ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.cess.gad"},
	{0x596C319B8540B62AULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.anb"},
	{0x596C7327D8D88B25ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrdb_0t.dds"},
	{0x596D08B954FDA055ULL, "hla_ht_0_hightown_n_-65537_0y.dds"},
	{0x596D32623006D488ULL, "w_bonepitmining_a2_73_0.msh"},
	{0x596D37D5655736FFULL, "140116662.ogg"},
	{0x596D44623006F319ULL, "w_bonepitmining_a2_73_0.mao"},
	{0x596D5062300707B2ULL, "w_bonepitmining_a2_73_0.mmh"},
	{0x596D677B9569D6CFULL, "6025341_m.fxe"},
	{0x596D84556F73A5F6ULL, "dae000_bodhan6166376.cl"},
	{0x596DBB85273C94E3ULL, "zz_lgt_1_debug6114223.cl"},
	{0x596E7165CF88AB78ULL, "223999796.ogg"},
	{0x596E8092E47BA7DDULL, "vxw_bonepitmining_a2_h7_0_0.vlm"},
	{0x596ECEF2841A5672ULL, "vxht_keep_5gg_0_0.vlm"},
	{0x596F0C794D4B3EDFULL, "hlac_w_encounter3_-1_0z.dds"},
	{0x596F4A39158D6FB9ULL, "673409486.ogg"},
	{0x596FB774A1F7F31BULL, "vxht_0_hightown_bz2_0_0.vlm"},
	{0x5970EED2F35CFEB8ULL, "w_sundermount_62_0.anb"},
	{0x597153F72C31CF34ULL, "vxht_mansion_n_60d_0_0.vlm"},
	{0x59715C7EF1F42FC6ULL, "6179298_m.fxe"},
	{0x5971A11EAFF30BEFULL, "6149678_m.fxe"},
	{0x59722D6755ECF6D3ULL, "6098293_m.fxe"},
	{0x59725649AF176E99ULL, "596804166.ogg"},
	{0x5972CEF644001526ULL, "84020154.ogg"},
	{0x5972D5053660512AULL, "lowtown_fem_cityguard_mob6164837.cl"},
	{0x5972F5C8680F0057ULL, "art\\vfx\\creatures\\fxc_shale_fire_r_c.cess.ani"},
	{0x5972F78FF09B4ADEULL, "vxca_cave_2mj_0_0.vlm"},
	{0x597305325A333B6BULL, "493905380.ogg"},
	{0x59730B13CDAA05A5ULL, "vxht_0_hightown_caq_0_1.vlm"},
	{0x5973E12634C3F52FULL, "167247758.ogg"},
	{0x5973F2589047E6A4ULL, "vxca_caverns1_gjx_0_0.vlm"},
	{0x5973FCA83793972AULL, "dre000_carver6037636.cl"},
	{0x597522C8009E1801ULL, "w_bonepit_17_0_ll.mmh"},
	{0x597584BA60B75FCDULL, "vxlt_darktown_6bw_0_0.vlm"},
	{0x5975BF32CF373CEDULL, "vxca_caverns1_mines_hvk_0_0.vlm"},
	{0x5975D1557ABD0995ULL, "vxca_caverns2_7en_0_1.vlm"},
	{0x5975DDE5DC08CC64ULL, "vxlt_warehouse_11e_0_0.vlm"},
	{0x59766DF495C3F9E7ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.mmh"},
	{0x597679EDC71E58C0ULL, "vxlt_0_lowtown_e0t_0_0.vlm"},
	{0x597689B46921D8B7ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_arrow_01_sa.dds"},
	{0x5976D50FDA76859EULL, "vxlt_warehouse_wh_0_0.vlm"},
	{0x5976E42469526479ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_joking.ani"},
	{0x5976E6250C060611ULL, "640742686.ogg"},
	{0x59774E813533944CULL, "6044909_m.fxe"},
	{0x5978142C4A7862E8ULL, "hlac_lt_0_lowtown_a2_131071_1y.dds"},
	{0x597886A8AF7030AEULL, "art\\levels\\races\\proxy\\props\\_textures\\cart_n.dds"},
	{0x59788D6B8C96CC85ULL, "vxw_encounter2_h5_0_0.vlm"},
	{0x597903AA69B0E3BEULL, "6182942_m.fxe"},
	{0x59792CE11EDFBF20ULL, "hlac_ht_chantry_2534_33y.dds"},
	{0x5979352F02B064CDULL, "838204723.ogg"},
	{0x597985FC08E44C36ULL, "vxlt_gallowstemplar_n_45h_0_0.vlm"},
	{0x597988392E83FBC7ULL, "vxlt_warehouse_n_st_0_0.vlm"},
	{0x5979A9CB5A6EBB3CULL, "6063779_m.fxe"},
	{0x5979F4062C448CE3ULL, "vxht_bartrand_n_65w_0_0.vlm"},
	{0x597A10D18AEDAA91ULL, "zz_kneelsister_v3.cut"},
	{0x597A117F4661B3D2ULL, "hlac_ht_0_hightown_n_131074_0x.dds"},
	{0x597A282C7D1B2EF6ULL, "vxht_mansion2b_1sg_0_0.vlm"},
	{0x597A3CEED45ECA1AULL, "558257727.ogg"},
	{0x597A650B9D3D3ED9ULL, "qcr220_grey_wardens.bnk"},
	{0x597A6518A6A6275AULL, "vxca_caverns1_i2p_0_0.vlm"},
	{0x597AC2171588EDF9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robb_0n.dds"},
	{0x597ACE2DAADD07B0ULL, "6084189_m.fxe"},
	{0x597B31F9EBB98DC1ULL, "vxlt_gallowscourtyard_n_7ac_0_1.vlm"},
	{0x597B4EB1041BF3E7ULL, "art\\characters\\creatures\\shade\\textures\\c_shadeb_0i.dds"},
	{0x597B7F9EF23FD4FFULL, "vxht_mansion2_2iw_0_0.vlm"},
	{0x597C12439580D7E4ULL, "hlac_ca_cavern1_haunted_24423_5z.dds"},
	{0x597C1A518C3D04CDULL, "231380631.ogg"},
	{0x597C734DE4D0CE86ULL, "vxca_caverns1_ehr_0_0.vlm"},
	{0x597C8B615FDEEF8BULL, "vxht_keep_4mq_0_0.vlm"},
	{0x597C98D4FDC8B8E2ULL, "gen_im_trs_gen_shd_met_r1.uti"},
	{0x597CFC00C5C62C68ULL, "6113978_f.fxe"},
	{0x597D23E3F9F63EC0ULL, "gen00fl_merrill6119603.cl"},
	{0x597DBF24695843BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_joking.gad"},
	{0x597DC61B35208909ULL, "vxlt_undercityserialkiller_1o9_0_0.vlm"},
	{0x597E870B9D416139ULL, "qcr220_grey_wardens.cnv"},
	{0x597F0B7AAF948628ULL, "w_blightlands_83_0.msh"},
	{0x597F1D7AAF94A4B9ULL, "w_blightlands_83_0.mao"},
	{0x597F297AAF94B952ULL, "w_blightlands_83_0.mmh"},
	{0x597F3F1959D2FFDBULL, "vxlt_backalley_g_1e5_0_0.vlm"},
	{0x597FEC431F6ABBCBULL, "mag211_marethari6025351.cl"},
	{0x5980170D3CB5EDB0ULL, "vxw_bonepitdestroyed_a3_262_0_0.vlm"},
	{0x598082E14553BDC3ULL, "6169489_m.fxe"},
	{0x5980A636712DA21BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9mw_pulleya.msh"},
	{0x5980B836712DC166ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9mw_pulleya.mao"},
	{0x5980C436712DD505ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9mw_pulleya.mmh"},
	{0x5980C63A51A4DFD6ULL, "vxlt_darktown_5xa_0_0.vlm"},
	{0x59812A693D9AA18DULL, "vxca_caverns1_j3o_0_0.vlm"},
	{0x598156D21C86122DULL, "vxw_bonepitmining_a2_276_0_0.vlm"},
	{0x598198E0B6325957ULL, "6133535_f.fxe"},
	{0x5981994EF12D0C02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_reaction_loop.gad"},
	{0x59822CC62E11618CULL, "llac_lt_0_lowtown_docks_a3_n_196606_0.dds"},
	{0x5982337FECBC5AFEULL, "6164032_m.fxe"},
	{0x5983196351F96D00ULL, "vxca_caverns2prc_3i2_0_1.vlm"},
	{0x59839A1ECF05C2A5ULL, "vxlt_0_lowtown_docks_n_cnk_0_0.vlm"},
	{0x59839CEBF8EF7B2BULL, "6070502_f.fxe"},
	{0x59839D66C3CDF42AULL, "lt_0_lowtown_a2_n_21_0.anb"},
	{0x59839F967F74D4C8ULL, "6000784_f.fxe"},
	{0x5983A99BC376AE42ULL, "hlac_lt_undercityserialkiller_-1_38y.dds"},
	{0x5983AB59FEC64148ULL, "6151070_m.fxe"},
	{0x5983C27CBDA40EEBULL, "vxht_mansion_n_4y6_0_0.vlm"},
	{0x5983D05D6FEE3164ULL, "404111504.ogg"},
	{0x598443B03A7D7AADULL, "vxw_bonepitmining_a2_1vn_0_0.vlm"},
	{0x59844FC2CDFEE4E4ULL, "vxlt_warehouse_14j_0_0.vlm"},
	{0x5984DC24695E93A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_joking.evt"},
	{0x5984E2D81F77FA6BULL, "vxht_0_hightown_brj_0_0.vlm"},
	{0x5984FFDECF8E8736ULL, "vrc221_bartrand6043277.cl"},
	{0x5985247762405849ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v2_c.impa.ani"},
	{0x598546ECCC6C9E1EULL, "vxlt_warehouse_xi_0_0.vlm"},
	{0x598563AB1C730AE1ULL, "hlac_lt_0_lowtown_docks_196608_1x.dds"},
	{0x59858A8D65D618D1ULL, "dae000ip_crowd_em2_talk6.utp"},
	{0x5985B094DD8A1739ULL, "ht_0_hightown_a3_42_0.rml"},
	{0x5986013876A58408ULL, "gen00fl_merrill6121129.cl"},
	{0x598619550B28E9B0ULL, "vxlt_0_lowtown_a2_f50_0_0.vlm"},
	{0x5986CF5A70543859ULL, "6208388_m.fxe"},
	{0x5986FF487E8CE505ULL, "vxw_encounter2_s4_0_0.vlm"},
	{0x5987668204B75DBBULL, "zz_mrl_debug6185296.cl"},
	{0x59879AB141760A96ULL, "vxlt_0_lowtown_docks_a3_d16_0_0.vlm"},
	{0x5987C46E3F9FDBF7ULL, "6093520_m.fxe"},
	{0x5987FA16207A1447ULL, "vxlt_warehouse_n_vw_0_0.vlm"},
	{0x5988085A43ECF017ULL, "follower_banter6163879.cl"},
	{0x5988106033DD9F8BULL, "vxht_keep_738_0_0.vlm"},
	{0x59889A096F114776ULL, "vxht_mansion2b_1vh_0_0.vlm"},
	{0x59899333BA49CA33ULL, "422135606.ogg"},
	{0x5989E785F24DDC49ULL, "qun231_varnell6078108.cl"},
	{0x598A657DF0986E4FULL, "hlac_ca_caverns2_5802_17x.dds"},
	{0x598A8F19D95FFD32ULL, "6045965_f.fxe"},
	{0x598B244913334BBCULL, "6116255_m.fxe"},
	{0x598B3F776244F1CBULL, "art\\vfx\\combat\\vfx_d_qun_attk_v2_c.impa.gad"},
	{0x598B920EF72BCAD7ULL, "vxlt_0_lowtown_n_e2v_0_0.vlm"},
	{0x598BC58E3E4621CCULL, "vxht_0_hightown_n_aok_0_1.vlm"},
	{0x598C26A2B222F96FULL, "6197963_m.fxe"},
	{0x598CE3556F4CDF28ULL, "6107723_m.fxe"},
	{0x598D3354712D3004ULL, "435107019.ogg"},
	{0x598D3BF01686FB2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_z_pos.gad"},
	{0x598DFFED19DA76B7ULL, "hlac_lt_gallowstemplar_2391_2x.dds"},
	{0x598E584E711DC6D4ULL, "hlac_lt_0_lowtown_a2_n_-65536_4x.dds"},
	{0x598E93BD9A5B850BULL, "309964790.ogg"},
	{0x598F0AA322214DFBULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_act_c.anb"},
	{0x598F4E314082A1D9ULL, "zz_approval_debug6003657.cl"},
	{0x598F8FAFB35864E3ULL, "vxht_mansion2b_2k1_0_0.vlm"},
	{0x598FBF7F40F65CE1ULL, "dae020cr_chantry1.utc"},
	{0x598FE5E35D5FBCBFULL, "hlac_lt_0_lowtown_docks_a3_-131075_2x.dds"},
	{0x5990044995B89A65ULL, "6070802_m.fxe"},
	{0x5990090FC4EEF52EULL, "ttx_arch_a_1.mmh"},
	{0x59900ED37B2D2336ULL, "6188636_m.fxe"},
	{0x5990230FC4EF211CULL, "ttx_arch_a_1.msh"},
	{0x59902F1B1D883DF6ULL, "vxlt_0_lowtown_a2_n_ghs_0_0.vlm"},
	{0x599039DCD72830B7ULL, "6096347_m.fxe"},
	{0x5990861E4DE3C04FULL, "221902511.ogg"},
	{0x5990F11262771275ULL, "6171594_m.fxe"},
	{0x599143448A31DF3FULL, "hlac_lt_undercityencounter_-1_5x.dds"},
	{0x5991D5028E97DA3AULL, "vxca_caverns1_mines_imp_0_0.vlm"},
	{0x59920397CC654BE3ULL, "6128755_f.fxe"},
	{0x599203D50DE40D36ULL, "hlac_dr_0_deeproads_5849_4z.dds"},
	{0x5992196B079B7451ULL, "hlac_ht_keep_3732_8y.dds"},
	{0x59921FD28E45895EULL, "hlac_ca_cavern1_haunted_24425_7x.dds"},
	{0x59923C77624B0B30ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v2_c.impa.evt"},
	{0x59928726DAEEE382ULL, "71120746.ogg"},
	{0x5992A2ECB3236DEDULL, "vxca_caverns1_mines_hli_0_0.vlm"},
	{0x5992A8B45CF0A869ULL, "664570540.ogg"},
	{0x5992ADF2C345C38DULL, "w_encounter3_114_0.msh"},
	{0x5992B5F48D0A41B8ULL, "6099987_m.fxe"},
	{0x5992BBF2C345DBC0ULL, "w_encounter3_114_0.mao"},
	{0x5992BFF2C345E213ULL, "w_encounter3_114_0.mmh"},
	{0x599312BB116FA03FULL, "6064023_m.fxe"},
	{0x59935743269C394BULL, "mer220_hubert6049423.cl"},
	{0x5993792AAF1E5071ULL, "root_war_berserker.dds"},
	{0x59947C968EDE4A49ULL, "hlac_ca_caverns1_24424_24z.dds"},
	{0x59948A6E391C2362ULL, "191924475.ogg"},
	{0x59948CEB26D021B2ULL, "6110718_m.fxe"},
	{0x5994CCA31A5FE236ULL, "fen110_fenris6100314.cl"},
	{0x5994D495290C6DE7ULL, "vxlt_0_lowtown_a2_f17_0_0.vlm"},
	{0x5994EC3601E6D6D2ULL, "vxlt_0_lowtown_a2_er8_0_0.vlm"},
	{0x5995349B8563C01DULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.mmh"},
	{0x5995700CDCDF0A7EULL, "247354630.ogg"},
	{0x5995A005AA12DBD4ULL, "vxlt_gallowstemplar_35f_0_0.vlm"},
	{0x5995A9233095F05FULL, "388427083.ogg"},
	{0x5995AE8591B25EBCULL, "da_vignette.dds"},
	{0x59960C506D813682ULL, "vxht_0_hightown_n_cad_0_0.vlm"},
	{0x59965DB3A74A891FULL, "vxlt_undercityencounter_1j8_0_0.vlm"},
	{0x59965FFF9AC0AD97ULL, "142553573.ogg"},
	{0x599669F6B1C81E56ULL, "hlac_lt_0_lowtown_a2_n_65536_3z.dds"},
	{0x59966FCAA365BADAULL, "vxw_encounter3_34_0_0.vlm"},
	{0x5996C44EF13FB1E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_reaction_loop.ani"},
	{0x5997AAA5D4C34B00ULL, "6115878_m.fxe"},
	{0x5997EFB87F76B0FEULL, "vxw_bonepit_15b_0_0.vlm"},
	{0x599821047556C1ABULL, "366303646.ogg"},
	{0x5998495DD542EEFFULL, "6072855_m.fxe"},
	{0x59985E92A8B2F194ULL, "plt_dae00pt_premium.nss"},
	{0x5998A5AC7B5116DBULL, "m_mag211ar_ariannis_house.wbd"},
	{0x5998C7C7C35F799CULL, "vxw_woundedcoast_16v_0_0.vlm"},
	{0x5998D22A57BE8EEFULL, "6051571_m.fxe"},
	{0x5998E128CBB837D6ULL, "ht_0_hightown_28_0.rml"},
	{0x5998EB0A4CBD4377ULL, "vxlt_0_lowtown_a2_n_c_f66_0_1.vlm"},
	{0x5999341F18C410A5ULL, "6096312_m.fxe"},
	{0x5999DBD2F37FE3CDULL, "w_sundermount_62_0.msh"},
	{0x5999E9D2F37FFB00ULL, "w_sundermount_62_0.mao"},
	{0x5999EDD2F3800253ULL, "w_sundermount_62_0.mmh"},
	{0x5999EF66AA2788A6ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_chrgdattk.evt"},
	{0x599A0E2B14D376F8ULL, "llac_lt_darktown_-1_33.dds"},
	{0x599A64B343EBDAAAULL, "377424351.ogg"},
	{0x599AA1351F06EADBULL, "vxlt_0_lowtown_docks_a3_n_d0j_0_1.vlm"},
	{0x599ADC077D1CDBBDULL, "vxht_mansion2_27s_0_0.vlm"},
	{0x599B40B29889ECE5ULL, "564473721.ogg"},
	{0x599BEBD843E46C57ULL, "follower_barks6153163.cl"},
	{0x599C06AF15A97306ULL, "389445061.ogg"},
	{0x599CB9AA9867674AULL, "1015779761.ogg"},
	{0x599D039A712D7D2FULL, "6099801_m.fxe"},
	{0x599D50541D0FF156ULL, "6116055_m.fxe"},
	{0x599D5FD3B2F58A61ULL, "hlac_lt_gallowscourtyard_-65536_3z.dds"},
	{0x599DBE6326BB79EFULL, "dae000ip_crowd_mf3_talk8.utp"},
	{0x599DF39C0EC1F310ULL, "vxht_0_hightown_a3_c90_0_0.vlm"},
	{0x599E45857E3A56A9ULL, "mrl000_marethari6142122.cl"},
	{0x599EB577696ECC55ULL, "vxht_0_hightown_a2_b7g_0_0.vlm"},
	{0x599EBC6E1B7C4C36ULL, "vxht_keep_6ib_0_0.vlm"},
	{0x599EDCA8E0DFEBF9ULL, "darktown_fem_elf_sta6174533.cl"},
	{0x599F046F439F0FB9ULL, "vxlt_warehouse_u9_0_0.vlm"},
	{0x599F1326DC33BA2CULL, "6085465_m.fxe"},
	{0x599F926EB3D9AF65ULL, "hlac_lt_gallowstemplar_n_2382_3z.dds"},
	{0x599F9D6523CF76C6ULL, "follower_banter6136650.cl"},
	{0x599FADDDDD59A2BFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_noba_0d.dds"},
	{0x59A00666AA2C1BF1ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_chrgdattk.gad"},
	{0x59A007F8CBA02B71ULL, "art\\characters\\playercharacter\\humanfemale\\cth_noba.mmh"},
	{0x59A013F8CBA03F12ULL, "art\\characters\\playercharacter\\humanfemale\\cth_noba.mao"},
	{0x59A021F8CBA0572FULL, "art\\characters\\playercharacter\\humanfemale\\cth_noba.msh"},
	{0x59A046DF808DF2BAULL, "vxca_caverns1_mines_ijq_0_0.vlm"},
	{0x59A116F016976510ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_z_pos.ani"},
	{0x59A129676E9B10F8ULL, "plt_tut00pt_tutorials.plo"},
	{0x59A15556F2A60EB3ULL, "438811282.ogg"},
	{0x59A16D16C54FE566ULL, "atl_startmenu2_dxt5_02.dds"},
	{0x59A1859502A41733ULL, "vxht_mansion2_2ak_0_0.vlm"},
	{0x59A1C69203642B6BULL, "vxht_0_hightown_bxl_0_0.vlm"},
	{0x59A2CE5DCC36AEEBULL, "fer_wood.mao"},
	{0x59A30A251D888D92ULL, "6059985_m.fxe"},
	{0x59A354F3EFD4D645ULL, "6031720_m.fxe"},
	{0x59A357FA5D87EC09ULL, "4383889.ogg"},
	{0x59A35E12F3DCEF52ULL, "vxlt_0_lowtown_a2_eu9_0_0.vlm"},
	{0x59A3FDDB585E8B36ULL, "vxht_0_hightown_n_bfl_0_0.vlm"},
	{0x59A47A7B8A68A041ULL, "6164017_m.fxe"},
	{0x59A4A4AA1F958146ULL, "war_attackspeedupgrade.dds"},
	{0x59A4F662DE014033ULL, "t1_skn_h10.tnt"},
	{0x59A5711AEC42A0CCULL, "331339536.ogg"},
	{0x59A57D0DB1550D3EULL, "hlac_ht_0_hightown_-3_0y.dds"},
	{0x59A604A41D57A6AFULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_ars_0n.dds"},
	{0x59A65D94B6A34E45ULL, "6149584_m.fxe"},
	{0x59A70766AA323B9FULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_chrgdattk.ani"},
	{0x59A739A4B555921CULL, "vxw_woundedcoast_11w_0_0.vlm"},
	{0x59A77D34DF085085ULL, "270770004.ogg"},
	{0x59A77EC6FE47EC7BULL, "vxw_bonepitmining_a2_1xz_0_0.vlm"},
	{0x59A78339649DD93FULL, "6177907_m.fxe"},
	{0x59A8F2EDEF6B71F2ULL, "ico_longbow_1.dds"},
	{0x59A8F854D13FBE14ULL, "ts_slavers_02.gda"},
	{0x59A9377838D3FAB4ULL, "gen00fl_fenris6126528.cl"},
	{0x59A9EBBF541523DBULL, "dae000cr_mage_jaken.utc"},
	{0x59AA0976CC82CD3AULL, "lgt150im_ntb_wr_ingrdnt_1.uti"},
	{0x59AA10D2CF1B02F1ULL, "t1_mub_te2.tnt"},
	{0x59AA187FAD95D886ULL, "594268938.ogg"},
	{0x59AA4FCAD40DD3C8ULL, "90182246.ogg"},
	{0x59AA7513E2E172EBULL, "vxlt_gallowstemplar_5qc_0_0.vlm"},
	{0x59AB6EACBB13FD3EULL, "6060233_m.fxe"},
	{0x59AB8BAA5E643F20ULL, "hlac_w_sundermount_-327680_0z.dds"},
	{0x59ABA700D66CA1FCULL, "vxlt_darktown_5t7_0_0.vlm"},
	{0x59AC7369974495E3ULL, "vxht_mansion2b_2a3_0_0.vlm"},
	{0x59AC7BEE8761C6D2ULL, "zz_fen_debug6134383.cl"},
	{0x59AD12D501746EF6ULL, "vxlt_0_lowtown_a2_n_gbu_0_0.vlm"},
	{0x59AD27545B64E4D5ULL, "vxht_0_hightown_a2_b8f_0_0.vlm"},
	{0x59ADA5A806528E5CULL, "gen00fl_anders6128366.cl"},
	{0x59ADC03AA2B26073ULL, "barr_01.mao"},
	{0x59ADD2334A97E36DULL, "154936553.ogg"},
	{0x59AE079FDE9B41DEULL, "t1_mul_te3.tnt"},
	{0x59AE528D598E7553ULL, "vxlt_house_hk_0_0.vlm"},
	{0x59AE81BF2D211038ULL, "vxw_woundedcoast_70_0_0.vlm"},
	{0x59AEB8BC72840B3AULL, "vxca_caverns1_mines_igr_0_0.vlm"},
	{0x59AF10BD4220F617ULL, "art\\characters\\weapons\\shields\\w_shd_srda.msh"},
	{0x59AF22BD4221157AULL, "art\\characters\\weapons\\shields\\w_shd_srda.mao"},
	{0x59AF26BD42211B89ULL, "art\\characters\\weapons\\shields\\w_shd_srda.mmh"},
	{0x59AFF771F49A2FB3ULL, "vxht_mansion2_2jj_0_0.vlm"},
	{0x59B06E1184ADF7A6ULL, "vxlt_undercityencounter_1bu_0_0.vlm"},
	{0x59B08E6425D2A6FCULL, "lt_0_lowtown_docks_184_0.mao"},
	{0x59B09A6425D2BA9FULL, "lt_0_lowtown_docks_184_0.mmh"},
	{0x59B0A06425D2C541ULL, "lt_0_lowtown_docks_184_0.msh"},
	{0x59B0A18CC67852A1ULL, "735035785.ogg"},
	{0x59B0E24201E91FDAULL, "348983887.ogg"},
	{0x59B13FF42D55A5E7ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_6y.dds"},
	{0x59B1560AF887CD37ULL, "vxht_brothel_n_5bq_0_0.vlm"},
	{0x59B1908584147407ULL, "321943041.ogg"},
	{0x59B1A410E46E5D86ULL, "296725768.ogg"},
	{0x59B1E5C4798F0AFBULL, "vxht_playershouse_n_1l9_0_0.vlm"},
	{0x59B26FB84A54A3B6ULL, "vxht_0_hightown_n_bam_0_0.vlm"},
	{0x59B273D991FA10A8ULL, "6148933_m.fxe"},
	{0x59B2822EFB49A1E5ULL, "vxht_mansion_n_322_0_0.vlm"},
	{0x59B32F900A4E8685ULL, "6111549_m.fxe"},
	{0x59B332C5627B7F80ULL, "gen00fl_anders6111292.cl"},
	{0x59B3A2AD1158F19EULL, "6055138_m.fxe"},
	{0x59B4090822C0034BULL, "vxw_woundedcoast_1r9_0_0.vlm"},
	{0x59B466991F95E66EULL, "hlac_ca_caverns2prc_5802_28z.dds"},
	{0x59B47454D9362954ULL, "6052605_m.fxe"},
	{0x59B47A0276A9EBA4ULL, "vxlt_0_lowtown_docks_a3_d1t_0_0.vlm"},
	{0x59B5005E85072712ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robc_0n.dds"},
	{0x59B5CE8A3BC53B48ULL, "678088357.ogg"},
	{0x59B7021924D44E0DULL, "6207385_m.fxe"},
	{0x59B742876043A64FULL, "6031159_m.fxe"},
	{0x59B7A3C22A7346D5ULL, "vxht_0_hightown_n_ap6_0_0.vlm"},
	{0x59B874E9E3626C5DULL, "qun231_varnell6099589.cl"},
	{0x59B8C719DB278078ULL, "957208623.ogg"},
	{0x59B9210B04D94151ULL, "360185970.ogg"},
	{0x59B939872731D0B1ULL, "art\\levels\\races\\design\\_textures\\trap_generic_n.dds"},
	{0x59B9CE1A0BBB05BAULL, "vxht_mansion2b_31c_0_0.vlm"},
	{0x59BA06FB4E13D53DULL, "271012664.ogg"},
	{0x59BA2102E9F20A7FULL, "192477672.ogg"},
	{0x59BA293A389E05E2ULL, "vxht_mansion2_1bx_0_0.vlm"},
	{0x59BA5956993D8AF7ULL, "vxdr_0_deeproads_bhf_0_1.vlm"},
	{0x59BA7D51E2DFD9E4ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_1y.dds"},
	{0x59BB2BC6AEED16F5ULL, "163735897.ogg"},
	{0x59BB78336538E050ULL, "hlac_lt_backalley_n_0_19y.dds"},
	{0x59BB8B9A38AB2225ULL, "hlac_w_woundedcoast_8_0z.dds"},
	{0x59BBB42B4558A593ULL, "zz_isa_debug6133156.cl"},
	{0x59BBF5992D387421ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkic.phy"},
	{0x59BC3BBA44649606ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\tools_flatspecular_s.dds"},
	{0x59BC551EC1A3DEE3ULL, "gen00fl_isabela6096789.cl"},
	{0x59BCA9B130E204D7ULL, "dae300_cullen6140652.cl"},
	{0x59BCCD14C583AF84ULL, "632681318.ogg"},
	{0x59BD401CC940A92EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv5.evt"},
	{0x59BD56E8A3574D5DULL, "6048139_m.fxe"},
	{0x59BD8F095B4192D5ULL, "246199682.ogg"},
	{0x59BDA8ACD3BDB121ULL, "6033667_m.fxe"},
	{0x59BDE33791AF1C48ULL, "1025672108.ogg"},
	{0x59BE22F38721AB0AULL, "ser121_jethann6108811.cl"},
	{0x59BE70252E29A1C8ULL, "641675874.ogg"},
	{0x59BE7BB98C9333A2ULL, "mag310cr_progenitor_2.utc"},
	{0x59BED4443FDF4D17ULL, "137665502.ogg"},
	{0x59BF481EC547671FULL, "6071003_m.fxe"},
	{0x59BFDB739459785AULL, "mcr310_argument6182452.cl"},
	{0x59C03CDB905CA731ULL, "311495688.ogg"},
	{0x59C10DB0C82282E6ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.cess.ani"},
	{0x59C11A7A7DF47BCDULL, "316471814.ogg"},
	{0x59C334F94C2FDECEULL, "134465551.ogg"},
	{0x59C42D76CADE222EULL, "839440042.ogg"},
	{0x59C4371CC946B8D9ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv5.gad"},
	{0x59C531A17796B57CULL, "vxlt_gallowscourtyard_n_76j_0_0.vlm"},
	{0x59C5AB6C6C4B5CA4ULL, "vxlt_darktown_2vb_0_0.vlm"},
	{0x59C5C537C55D9922ULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em.phy"},
	{0x59C6159F1C695F55ULL, "vxht_0_hightown_n_au5_0_0.vlm"},
	{0x59C64198032BAC12ULL, "354798608.ogg"},
	{0x59C6B3DD8ADEDBDCULL, "qun110_petrice6018732.cl"},
	{0x59C7B424B38542FEULL, "vxca_caverns1_e7g_0_0.vlm"},
	{0x59C84D43EAD0492FULL, "vxht_0_hightown_c6g_0_0.vlm"},
	{0x59C8C98D6A958A40ULL, "621890039.ogg"},
	{0x59C92A8BA8FC6CC3ULL, "vxlt_warehouse_n_9v_0_0.vlm"},
	{0x59C9747C1D57C06CULL, "primitives_primitive_p4.psh.fxo.dx11"},
	{0x59CA16533041F02BULL, "vxlt_0_lowtown_docks_a3_cob_0_0.vlm"},
	{0x59CA8B796BA13087ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\bsh_thorn_sa.dds"},
	{0x59CAB0972EF4AEA0ULL, "hlac_lt_0_lowtown_docks_65533_1y.dds"},
	{0x59CB481CC94CF427ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv5.ani"},
	{0x59CBDF8D66F93FF3ULL, "vxht_0_hightown_ani_0_0.vlm"},
	{0x59CC06F2EB555AD3ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_mera_0n.dds"},
	{0x59CC2D20BFCAC293ULL, "553742908.ogg"},
	{0x59CC5527442EF9D7ULL, "llac_lt_0_lowtown_65537_2.dds"},
	{0x59CCA0F892DCED62ULL, "6165527_m.fxe"},
	{0x59CCA15969A9F4E3ULL, "6121056_m.fxe"},
	{0x59CCA1AF3D24DD43ULL, "hlac_w_bonepitdestroyed_a3_327686_0y.dds"},
	{0x59CCBA35F1CF881BULL, "art\\vfx\\combat\\vfx_gui_target_ret_120_p.mmh"},
	{0x59CD52D4ACF62036ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die.ani"},
	{0x59CDCDA509D7ABFDULL, "vxw_blightlands_118_0_0.vlm"},
	{0x59CE39C4DC73FE37ULL, "vxht_brothel_n_5hs_0_0.vlm"},
	{0x59CEF5B0C82E9757ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.cess.evt"},
	{0x59CEF7B757C78E47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel.gad"},
	{0x59CF298FA3518250ULL, "llac_ht_0_hightown_a2_-196608_0.dds"},
	{0x59CF7D004FC3DD96ULL, "363541008.ogg"},
	{0x59CF9DBE20215EEEULL, "vxw_sundermount_32c_0_1.vlm"},
	{0x59D14278CBA27F0EULL, "6036461_m.fxe"},
	{0x59D1C25A7AA493FEULL, "art\\vfx\\combat\\vfx_m_sph_conj_c.impa.ani"},
	{0x59D1C472B781691EULL, "vxw_encounter3_4y_0_0.vlm"},
	{0x59D1D8A7590D4E61ULL, "spider2.dds"},
	{0x59D215D3DCAB4467ULL, "495617201.ogg"},
	{0x59D2372503F452A0ULL, "vxht_mansion2_3g3_0_0.vlm"},
	{0x59D27035C462232AULL, "mrl210_merrill_varterralpt1.stg"},
	{0x59D2DE28CBE9AA8AULL, "ht_0_hightown_28_0.anb"},
	{0x59D31E679173793BULL, "590290213.ogg"},
	{0x59D373E1A676FF5CULL, "442944232.ogg"},
	{0x59D3A37E698CCDFCULL, "vxlt_gallowscourtyard_n_79k_0_0.vlm"},
	{0x59D3A70FD60838B6ULL, "hightown_fem_noble1_sta6169246.cl"},
	{0x59D4AD5F8B669687ULL, "vxht_0_hightown_a3_bc8_0_0.vlm"},
	{0x59D4E98B91A0B8E7ULL, "vxdr_primevaltomb_915_0_0.vlm"},
	{0x59D4EA16F7292C41ULL, "418471947.ogg"},
	{0x59D5ECB757CD9A54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel.evt"},
	{0x59D5F6B0C834B7E0ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.cess.gad"},
	{0x59D5FE735EBA1EA4ULL, "gen_im_trs_crc_junk1_04.uti"},
	{0x59D6188490FB428DULL, "mag121_idunna6083955.cl"},
	{0x59D6246B274A2C37ULL, "hlac_lt_gallowstemplar_f_7198_3x.dds"},
	{0x59D7B1FCCA2F144AULL, "6081360_m.fxe"},
	{0x59D7D28D7A02A8E6ULL, "vxlt_gallowscourtyard_n_7dj_0_0.vlm"},
	{0x59D8560C0AEBF104ULL, "vxht_keep_6b1_0_0.vlm"},
	{0x59D87C3A6A37F0A3ULL, "darktown_male_lowlife_mob6174399.cl"},
	{0x59D953A567D5CF37ULL, "vxlt_0_lowtown_docks_b4e_0_0.vlm"},
	{0x59D958EBB7A82A4CULL, "345075051.ogg"},
	{0x59D9827926901A53ULL, "vxdr_primevalentrance_339_0_0.vlm"},
	{0x59D9AD60876C72E2ULL, "497398459.ogg"},
	{0x59D9DB35F1DA4A80ULL, "art\\vfx\\combat\\vfx_gui_target_ret_120_p.anb"},
	{0x59D9EB5E08A3B29BULL, "409862547.ogg"},
	{0x59DA68AE541431AAULL, "hlac_dr_dungeon_3216_11z.dds"},
	{0x59DA731ECC11C107ULL, "400812206.ogg"},
	{0x59DA9BED05D66315ULL, "vxht_mansion_n_674_0_0.vlm"},
	{0x59DB5AD4AD026B27ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die.evt"},
	{0x59DBB3AAC96FF6EEULL, "ave310_donnic6134752.cl"},
	{0x59DBCE9022D0B1EFULL, "hlac_lt_0_lowtown_a2_n_-262144_0y.dds"},
	{0x59DBFA3A101F7548ULL, "w_bonepit_93_0.msh"},
	{0x59DC0C3A101F93D9ULL, "w_bonepit_93_0.mao"},
	{0x59DC183A101FA872ULL, "w_bonepit_93_0.mmh"},
	{0x59DC2C7A25797F26ULL, "leave_si.mao"},
	{0x59DC397ACFD2B5C3ULL, "vxht_0_hightown_a3_cd4_0_0.vlm"},
	{0x59DC3F81FBCDC47DULL, "vxw_blightlands_147_0_0.vlm"},
	{0x59DD6A746A5B6B40ULL, "6050172_f.fxe"},
	{0x59DD7B1F8A43E183ULL, "6201565_m.fxe"},
	{0x59DDDCCE90892A94ULL, "6029373_m.fxe"},
	{0x59DDDDA17F3DE974ULL, "vxdr_primevalentrance_1q6_0_0.vlm"},
	{0x59DE4263E75445B3ULL, "313362409.ogg"},
	{0x59DF8AA46B9EE9DCULL, "vxdr_dungeon_1mh_0_0.vlm"},
	{0x59DF8DD3D2FED14FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_loop.gad"},
	{0x59DFCF0717E4E206ULL, "286011578.ogg"},
	{0x59DFDA5A7AB0F9CFULL, "art\\vfx\\combat\\vfx_m_sph_conj_c.impa.evt"},
	{0x59E01E36180A019EULL, "mrl000_marethari6129267.cl"},
	{0x59E0364FA977819EULL, "vxw_encounter3_1z_0_0.vlm"},
	{0x59E0E6340E7DFC25ULL, "weapon1_s1.psh.fxo"},
	{0x59E0F1265EED4E8DULL, "hlac_ht_mansion2b_818_67z.dds"},
	{0x59E1635735D1A7B8ULL, "vxca_caverns1_e57_0_0.vlm"},
	{0x59E17BD4AD070F70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die.gad"},
	{0x59E19058828B7A06ULL, "vxca_caverns1_e5y_0_0.vlm"},
	{0x59E1DB75E31277F6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunkhigh_n.dds"},
	{0x59E1EDB0439F184FULL, "qcr270_fake_meredith6066335.cl"},
	{0x59E21A4B7180DD76ULL, "hlac_w_bonepit_196614_0y.dds"},
	{0x59E226F66575A711ULL, "6040885_m.fxe"},
	{0x59E245EFA1A58047ULL, "828351006.ogg"},
	{0x59E2814FB7CDA433ULL, "vxca_cavern1_haunted_ih2_0_0.vlm"},
	{0x59E2AC389C0E7357ULL, "277819553.ogg"},
	{0x59E2E4B757D8168DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel.ani"},
	{0x59E32C27FD9AFED0ULL, "6078740_m.fxe"},
	{0x59E339EA46691A11ULL, "26744662.ogg"},
	{0x59E352F82D71FE47ULL, "vxlt_darktown_4he_0_0.vlm"},
	{0x59E3DA037DE01173ULL, "282264252.ogg"},
	{0x59E3DEE40BD465ACULL, "mcr370_meredith6122901.cl"},
	{0x59E4088D696853ABULL, "llac_dr_0_deeproads_5845_5.dds"},
	{0x59E438E0E7FF0089ULL, "hla_ht_0_hightown_n_-65534_0x.dds"},
	{0x59E46B310AC41D7DULL, "clf260_harley6069643.cl"},
	{0x59E47F76A8F2F3CFULL, "6025136_m.fxe"},
	{0x59E4B3D7CBBF2B8EULL, "6089473_f.fxe"},
	{0x59E4BBE7C06C0F50ULL, "253042350.ogg"},
	{0x59E50CC784246EB5ULL, "vxw_bonepitmining_a2_261_0_0.vlm"},
	{0x59E52DACFAAE18E7ULL, "vxw_encounter2_camp_8g_0_0.vlm"},
	{0x59E593521C6504BAULL, "vxca_caverns1_mines_j26_0_0.vlm"},
	{0x59E5A78ABE000240ULL, "plt_zz_mer130pt_racketeers.plo"},
	{0x59E5AA531E3C8AF8ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extblack.dds"},
	{0x59E5F60752C426C2ULL, "toolssubmissionqueue.sub.xml"},
	{0x59E62865F35810DCULL, "515720646.ogg"},
	{0x59E649B5295B391EULL, "6099343_m.fxe"},
	{0x59E68C069C16EC46ULL, "75237985.ogg"},
	{0x59E699FB5D5AC9A5ULL, "vxht_0_hightown_c9y_0_1.vlm"},
	{0x59E6A2D3D30513FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_loop.evt"},
	{0x59E6B7B1A8A3F73CULL, "vxht_bartrand_n_68o_0_0.vlm"},
	{0x59E6DB5A7AB71A78ULL, "art\\vfx\\combat\\vfx_m_sph_conj_c.impa.gad"},
	{0x59E725AC539D9F82ULL, "483091599.ogg"},
	{0x59E7276E8161387FULL, "ca_caverns1_mines_24423.rml"},
	{0x59E7B9754C40A015ULL, "394141979.ogg"},
	{0x59E80DFDBDB87956ULL, "6079796_m.fxe"},
	{0x59E8BD8F5DBC7CAFULL, "vxlt_0_lowtown_docks_aq9_0_0.vlm"},
	{0x59E8FB819FD59AE7ULL, "zz_one_debug6051413.cl"},
	{0x59E954396C0C48D7ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.impa.gad"},
	{0x59E972C6BAD1C6CDULL, "gen00fl_fenris6127025.cl"},
	{0x59E989D5098C24ADULL, "610572763.ogg"},
	{0x59EAA07C7A88A88BULL, "139094803.ogg"},
	{0x59EAE3AC7B52DB27ULL, "6143508_m.fxe"},
	{0x59EB39B57D18F08FULL, "vxlt_gallowscourtyard_7ht_0_0.vlm"},
	{0x59EBAC37C57E3E75ULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em.mmh"},
	{0x59EBAE37C57E41EBULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em.msh"},
	{0x59ECE0104130503AULL, "fx_smoke_vol2.dds"},
	{0x59ED094D468C4571ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_grenade_toss.ani"},
	{0x59EE2C08E13A7838ULL, "vxlt_backalley_g_1wx_0_0.vlm"},
	{0x59EE4B67D1C73A31ULL, "hlac_w_bonepitmining_a2_0_5z.dds"},
	{0x59EE590A2DF44A86ULL, "vxlt_backalley_g_1w6_0_0.vlm"},
	{0x59EE598071AFE9A4ULL, "6155658_m.fxe"},
	{0x59EED9CA03524D7DULL, "487038092.ogg"},
	{0x59EF0F5E8E679511ULL, "mer341_jansen6083002.cl"},
	{0x59EF42E041E4FF2FULL, "llac_w_woundedcoast_262149_0.dds"},
	{0x59EF5EB71DB45C0BULL, "vxw_encounter1_f4_0_0.vlm"},
	{0x59EFDDB086C6BC13ULL, "192149344.ogg"},
	{0x59F069396C128B64ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.impa.evt"},
	{0x59F08EDC94FAA1C4ULL, "hlac_lt_gallowscourtyard_0_0z.dds"},
	{0x59F16BB381612427ULL, "ave260cr_smuggler.utc"},
	{0x59F1BDDAB6A278B0ULL, "vxca_caverns2_5xk_0_0.vlm"},
	{0x59F1C829EC14D9ABULL, "6042135_m.fxe"},
	{0x59F1D373DD1571E7ULL, "lit_bt_flare.mao"},
	{0x59F20457446A56CBULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath2.ani"},
	{0x59F20FD79D530F0CULL, "479933005.ogg"},
	{0x59F2B4BCB9D7F2C8ULL, "6173517_m.fxe"},
	{0x59F36054637E2F65ULL, "hlac_ca_caverns1_24422_2x.dds"},
	{0x59F39F89ECA43167ULL, "vxw_encounter2_camp_3f_0_0.vlm"},
	{0x59F3AAD3D30FAB15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_loop.ani"},
	{0x59F4144D46927613ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_grenade_toss.gad"},
	{0x59F4354180FB7F13ULL, "zz_qun_debug6064343.cl"},
	{0x59F46092F468E6CEULL, "vxca_caverns1_hl3_0_0.vlm"},
	{0x59F4C9D959C928A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv2.ani"},
	{0x59F4F45025FD0DE3ULL, "dae000_tomwise6167336.cl"},
	{0x59F539C5EED82204ULL, "vxht_keep_6h7_0_0.vlm"},
	{0x59F553E7E0DBBB1CULL, "character_rendertodepth_p4.psh.fxo.dx11"},
	{0x59F578204A199094ULL, "1066068371.ogg"},
	{0x59F637C23A822718ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_laundry.phy"},
	{0x59F6517BC480645EULL, "vxlt_0_lowtown_a3_n_c_i1e_0_0.vlm"},
	{0x59F674A5D143DFFFULL, "vxlt_0_lowtown_docks_bv7_0_0.vlm"},
	{0x59F6995507E65514ULL, "63951969.ogg"},
	{0x59F6CCCAAFF70699ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_1y.dds"},
	{0x59F6D045C1A38F46ULL, "follower_banter6156230.cl"},
	{0x59F6F331576A520FULL, "vxlt_0_lowtown_docks_bar_0_0.vlm"},
	{0x59F7739B4FC93550ULL, "dae000_fenarel6128137.cl"},
	{0x59F787EA7FF3A33FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic.msh"},
	{0x59F78DEA7FF3ADE1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic.mmh"},
	{0x59F799EA7FF3C182ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic.mao"},
	{0x59F8DAD8BBA3C9B9ULL, "vxlt_blackemporium_169_0_0.vlm"},
	{0x59F93D862B4C51C2ULL, "6206886_m.fxe"},
	{0x59F97CD07AAC89A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_3.ani"},
	{0x59F9D77288102EAFULL, "6113645_f.fxe"},
	{0x59FA9697D621B0DEULL, "vxw_sundermount_308_0_0.vlm"},
	{0x59FAC15B632A1A74ULL, "vxdr_primevalentrance_1k4_0_0.vlm"},
	{0x59FAD0DBE7F4BE9AULL, "vxht_0_hightown_a3_a7j_0_0.vlm"},
	{0x59FB114D46989078ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_grenade_toss.evt"},
	{0x59FB7BFC4B3DBE23ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_quna_0d.dds"},
	{0x59FB7F7BAB009E40ULL, "hlac_lt_0_lowtown_a3_n_c_131071_2x.dds"},
	{0x59FB82BBA6D3A0FDULL, "461278919.ogg"},
	{0x59FBC2414DE32B7EULL, "vxlt_0_lowtown_docks_b52_0_1.vlm"},
	{0x59FBE4D959CF752BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv2.gad"},
	{0x59FBF14D8BB6652BULL, "vxlt_gamlenshouse_n_ax_0_0.vlm"},
	{0x59FC1B3E11876B79ULL, "6177817_m.fxe"},
	{0x59FC64B599CF1C48ULL, "6195135_m.fxe"},
	{0x59FD07D5254EA336ULL, "gen00fl_merrill6119627.cl"},
	{0x59FD9C269B64C708ULL, "6125441_f.fxe"},
	{0x59FDD0940FAA748BULL, "vxw_encounter1_k7_0_0.vlm"},
	{0x59FDE8C634748F35ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_lgta_0t.dds"},
	{0x59FDF9CD5D01F095ULL, "vxlt_hangedman_n_1yg_0_0.vlm"},
	{0x59FE81396C1EF15DULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.impa.ani"},
	{0x59FE822396C2826CULL, "vxw_sundermount_2m8_0_0.vlm"},
	{0x59FE852A24771CE9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_gallowpad.phy"},
	{0x59FEEBDB76C22C4CULL, "mag320_orsino6071201.cl"},
	{0x59FF17565C428C5DULL, "art\\characters\\weapons\\shields\\w_shd_dslb.mao"},
	{0x59FF1B565C429296ULL, "art\\characters\\weapons\\shields\\w_shd_dslb.mmh"},
	{0x59FF25565C42A394ULL, "art\\characters\\weapons\\shields\\w_shd_dslb.msh"},
	{0x59FF4585017494F2ULL, "llac_lt_warehouse_0_7.dds"},
	{0x5A0032CE73F06F56ULL, "11404491.ogg"},
	{0x5A0087D07AB2BB43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_3.gad"},
	{0x5A00A6A78721DA55ULL, "mag370_meredith6064964.cl"},
	{0x5A010689250524C0ULL, "vxlt_warehouse_um_0_1.vlm"},
	{0x5A013F0CB2D46550ULL, "460048217.ogg"},
	{0x5A01E1D959D3DB90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv2.evt"},
	{0x5A01F4B41318EB01ULL, "vxdr_dungeon_1m9_0_0.vlm"},
	{0x5A01FBB467F808A1ULL, "6065317_f.fxe"},
	{0x5A0209B54AE5A5BBULL, "vxht_bartrand_n_5tc_0_0.vlm"},
	{0x5A02454118DB9157ULL, "90647728.ogg"},
	{0x5A027351318A8D46ULL, "910287186.ogg"},
	{0x5A03C73833A0B3B2ULL, "vxht_chantry_270_0_0.vlm"},
	{0x5A03D454C5DC9ACEULL, "vxht_keep_5ii_0_0.vlm"},
	{0x5A0445D5A91DC0C5ULL, "lt_warehouse_n.rml"},
	{0x5A046B93CC42D8ECULL, "llac_lt_0_lowtown_a2_-131072_8.dds"},
	{0x5A049DE97D595C6BULL, "cod_hst_kirk_arch_1a_8.utp"},
	{0x5A055CE8D6C14CA6ULL, "6125659_m.fxe"},
	{0x5A059EE62A9EF5A1ULL, "240892545.ogg"},
	{0x5A05EB57447AD555ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath2.gad"},
	{0x5A05F30E2E2DC96BULL, "dae011_gossips6149790.cl"},
	{0x5A06739722AF872AULL, "vxlt_ruinedhouse_n_hb_0_0.vlm"},
	{0x5A067EBC2092A566ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dazed.evt"},
	{0x5A069A9F0DA571E5ULL, "zz_followers_banter6157322.cl"},
	{0x5A06FB5F4204D43AULL, "character_character_p261.psh.fxo.dx11.assembly"},
	{0x5A0722D5B03A655FULL, "dae100_bran6142995.cl"},
	{0x5A0764D07AB89E28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_3.evt"},
	{0x5A077AE0D0E361EBULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rtl.ani"},
	{0x5A07B6A3CF53E74EULL, "776768504.ogg"},
	{0x5A08480D50D73C68ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06a.msh"},
	{0x5A086284F20D9309ULL, "dae310ip_store_armor.utp"},
	{0x5A08660D50D76E92ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06a.mmh"},
	{0x5A08B8A7A55BB739ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomeb_n.dds"},
	{0x5A08D6FB268E3F3FULL, "mer231cr_brekker.utc"},
	{0x5A09B1B9F127EE26ULL, "vxw_sundermount_31b_0_1.vlm"},
	{0x5A09FAC54D4C2292ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_crnf.gad"},
	{0x5A0A0EEBD5E138DBULL, "6039784_m.fxe"},
	{0x5A0A0F64A050B106ULL, "vxw_bonepitmining_a2_1jm_0_0.vlm"},
	{0x5A0A7F6EC51CCB47ULL, "6076089_f.fxe"},
	{0x5A0B3CC411E07B86ULL, "vxlt_backalley_g_1y8_0_0.vlm"},
	{0x5A0B5E400A0620A9ULL, "6182065_f.fxe"},
	{0x5A0B9AA20EEF9A21ULL, "llac_ht_0_hightown_65534_1.dds"},
	{0x5A0BEEF20F78FAC5ULL, "430280316.ogg"},
	{0x5A0C2C52F3E8F06FULL, "vxht_0_hightown_a2_ai8_0_0.vlm"},
	{0x5A0C2D08E1214C8FULL, "vxht_keep_n_4av_0_0.vlm"},
	{0x5A0C535FE49D0CC6ULL, "605094725.ogg"},
	{0x5A0C6BAA4EF80915ULL, "vxlt_hangedman_n_1vf_0_0.vlm"},
	{0x5A0C9A373037609DULL, "6173314_m.fxe"},
	{0x5A0CA55C1DE539A4ULL, "llac_lt_0_lowtown_a2_-65536_3.dds"},
	{0x5A0CD962CC91F03AULL, "43855311.ogg"},
	{0x5A0D95BC2098EAB1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dazed.gad"},
	{0x5A0DF72E4A2D9CF3ULL, "w_dgr_dg06a_0n.dds"},
	{0x5A0F92AD25C45363ULL, "dae000ip_crowd_lm1_talk15.utp"},
	{0x5A10A15915BC5700ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drga_0s.dds"},
	{0x5A10FDC54D5246C9ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_crnf.evt"},
	{0x5A113260701F6886ULL, "hlac_w_sundermount_262144_0x.dds"},
	{0x5A11D4ABDB74B971ULL, "vxlt_0_lowtown_docks_n_azy_0_0.vlm"},
	{0x5A1236E3E6B19974ULL, "6162461_m.fxe"},
	{0x5A12518CDDC5D42BULL, "331051947.ogg"},
	{0x5A127A83924084BFULL, "6020499_f.fxe"},
	{0x5A12B48EA9B97750ULL, "6166270_m.fxe"},
	{0x5A12DEEC5C0BBD28ULL, "mag311_emile_de_launcet6043627.cl"},
	{0x5A133866DE62535AULL, "dae000_merchant_dm_mob6174310.cl"},
	{0x5A134A88F8E52753ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_left.phy"},
	{0x5A1385B3BE9C7007ULL, "vxca_cavern1_haunted_iok_0_0.vlm"},
	{0x5A140010ADC78185ULL, "ht_0_hightown_n_27_0_lc.mmh"},
	{0x5A1466489AC018A2ULL, "vxlt_blackemporium_nw_0_0.vlm"},
	{0x5A1496BC209F0B5FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dazed.ani"},
	{0x5A14973D06B11738ULL, "6182016_m.fxe"},
	{0x5A1498301319BBE6ULL, "6102788_m.fxe"},
	{0x5A14D153015D3F09ULL, "dae000_bartrand6030636.cl"},
	{0x5A1559D0CBD61450ULL, "6111175_m.fxe"},
	{0x5A1694B357ED360CULL, "879003210.ogg"},
	{0x5A16DA4AC1B7B8B1ULL, "gen00fl_anders6116113.cl"},
	{0x5A16EF5BA8EB88D9ULL, "vxlt_0_lowtown_a2_n_c_byp_0_0.vlm"},
	{0x5A16FB5918C15344ULL, "hlac_lt_0_lowtown_a2_131072_1x.dds"},
	{0x5A17233FB7E07F83ULL, "6190980_m.fxe"},
	{0x5A1773148B3E7B40ULL, "6126794_f.fxe"},
	{0x5A1814B33C10AAA3ULL, "6076085_f.fxe"},
	{0x5A18233F94E0BDE5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_hed_flma_0d.dds"},
	{0x5A187BEAF4C5D5EAULL, "llac_lt_gallowstemplar_f_2382_2.dds"},
	{0x5A18AF7E0551C616ULL, "hm_gxa_vgk_cnspr_4.mor"},
	{0x5A1987663C7A2825ULL, "mcr310_argument6075816.cl"},
	{0x5A19CDD1B33AE0D5ULL, "lt_0_lowtown_docks_247_0.msh"},
	{0x5A19CFD1B33AE44BULL, "lt_0_lowtown_docks_247_0.mmh"},
	{0x5A19DBD1B33AF8E8ULL, "lt_0_lowtown_docks_247_0.mao"},
	{0x5A1A24131E84DC45ULL, "llac_lt_0_lowtown_a2_65537_6.dds"},
	{0x5A1A2F77F1C13728ULL, "134080202.ogg"},
	{0x5A1A3AF11438FA39ULL, "plt_cod_ltr_meredith_thank.nss"},
	{0x5A1A5485AB50C379ULL, "vxlt_gallowscourtyard_7e9_0_0.vlm"},
	{0x5A1AD5429B6E2E21ULL, "508629667.ogg"},
	{0x5A1ADD8740EE2195ULL, "vxlt_hangedman_n_1se_0_0.vlm"},
	{0x5A1AE9EECD4E5FABULL, "vxdr_primevaltomb_2gh_0_0.vlm"},
	{0x5A1B61E0D0F3E075ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rtl.gad"},
	{0x5A1B83026DC91E17ULL, "6176038_m.fxe"},
	{0x5A1B9441DC5C0F41ULL, "420258660.ogg"},
	{0x5A1CD6EFE8AF088AULL, "347793872.ogg"},
	{0x5A1CF5224CB201F2ULL, "mrl000_marethari6129414.cl"},
	{0x5A1D0C9710D26F49ULL, "qcr000_arishok6085011.cl"},
	{0x5A1D8400F4CC0365ULL, "gen00fl_isabela6080537.cl"},
	{0x5A1DFA992D8B475DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkic.mao"},
	{0x5A1DFE992D8B4D96ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkic.mmh"},
	{0x5A1E08992D8B5E94ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkic.msh"},
	{0x5A1E0A0175EBE048ULL, "mcr370_cut_player_survives6189917.cl"},
	{0x5A1E1BA1EF473027ULL, "t3_wpn_blr.tnt"},
	{0x5A1E328D510F91EBULL, "6077728_m.fxe"},
	{0x5A1E4167BC363891ULL, "151911850.ogg"},
	{0x5A1EA175DF34027BULL, "hlac_lt_undercityencounter_-1_27z.dds"},
	{0x5A1EE5C54D5E5B38ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_crnf.ani"},
	{0x5A1F2F771B80B793ULL, "6208298_m.fxe"},
	{0x5A1F4457F0D9F15DULL, "gen00fl_dog6167013.cl"},
	{0x5A204B3499559EECULL, "393123850.ogg"},
	{0x5A20933DC457656AULL, "hlac_ht_0_hightown_n_-1_2z.dds"},
	{0x5A20A252A4F7A9C1ULL, "follower_banter6156194.cl"},
	{0x5A20A43C9A452167ULL, "lt_0_lowtown_n_ll.mmh"},
	{0x5A2121F1143EEE12ULL, "plt_cod_ltr_meredith_thank.plo"},
	{0x5A21352C82B0B8A3ULL, "6149630_m.fxe"},
	{0x5A21D931168C239CULL, "6068547_m.fxe"},
	{0x5A221E7474635B35ULL, "zz_lgtpt_3_debug.ncs"},
	{0x5A238EEA8018EA5EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic.phy"},
	{0x5A23D0491EA77EC0ULL, "vxlt_0_lowtown_etx_0_0.vlm"},
	{0x5A240C08F2DD44D9ULL, "w_bonepitdestroyed_a3_2_0_ll.mmh"},
	{0x5A242FC154BE929DULL, "48864849.ogg"},
	{0x5A245C0229C3E04EULL, "vxlt_0_lowtown_docks_bd0_0_0.vlm"},
	{0x5A2489894F4C83B3ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.dura.evt"},
	{0x5A256E5BE43FC777ULL, "vxht_bartrand_n_5qt_0_0.vlm"},
	{0x5A25B6AE480DDE8DULL, "vxw_encounter2_camp_nc_0_0.vlm"},
	{0x5A25DD23FF2EDFDAULL, "gen00fl_anders6103675.cl"},
	{0x5A25F97BE6AB0ACEULL, "vxht_0_hightown_n_cbe_0_0.vlm"},
	{0x5A26486DCA72F1C4ULL, "water_water_p5.psh.fxo.dx11"},
	{0x5A26BBB59900A619ULL, "6154616_m.fxe"},
	{0x5A26FB4FEC742824ULL, "follower_banter6172959.cl"},
	{0x5A270DC67B2B737EULL, "vxw_bonepitmining_a2_257_0_0.vlm"},
	{0x5A2795F7050393DAULL, "137698868.ogg"},
	{0x5A27AFA0822AC558ULL, "689566729.ogg"},
	{0x5A284A6E81990AE3ULL, "ca_caverns1_mines_24423.anb"},
	{0x5A28B6F6F8472568ULL, "291070933.ogg"},
	{0x5A28C34FFFC4C3DBULL, "vxlt_0_lowtown_a2_n_fex_0_0.vlm"},
	{0x5A28C9C5684B3D36ULL, "dae021_brothel_onenter6196699.cl"},
	{0x5A28DE18EFC4020EULL, "173007769.ogg"},
	{0x5A2939C79B36A4BFULL, "vxw_sundermount_2m7_0_0.vlm"},
	{0x5A29CAE48ED55AAAULL, "6127063_f.fxe"},
	{0x5A29E30985EB82EEULL, "vxlt_gallowscourtyard_4cc_0_0.vlm"},
	{0x5A2A6E2A249C300EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_gallowpad.mmh"},
	{0x5A2A7A2A249C44A5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_gallowpad.mao"},
	{0x5A2A882A249C5CFCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_gallowpad.msh"},
	{0x5A2AC1CF8765FD4CULL, "598046343.ogg"},
	{0x5A2B49DE53ECA5ADULL, "6118847_m.fxe"},
	{0x5A2B82894F5296F4ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.dura.gad"},
	{0x5A2C26553F9BBD46ULL, "hlac_ca_cavern1_haunted_24423_5x.dds"},
	{0x5A2C3D6B816F749EULL, "674073450.ogg"},
	{0x5A2D04077EF690DEULL, "6119537_m.fxe"},
	{0x5A2D16BC54A59C2BULL, "vxht_playershouse_n_1li_0_0.vlm"},
	{0x5A2E009569A70A32ULL, "dae000_mother6028529.cl"},
	{0x5A2F0727D65DA4E3ULL, "6111055_m.fxe"},
	{0x5A2F897A1601EF62ULL, "810092594.ogg"},
	{0x5A2FF80FC1E51D36ULL, "vxht_mansion_n_60v_0_0.vlm"},
	{0x5A30CD161950A50DULL, "hlac_w_woundedcoast_131072_0x.dds"},
	{0x5A30CDD97BE68D1FULL, "mer240_jansen6050435.cl"},
	{0x5A30FF4F68A2CB2FULL, "6093876_m.fxe"},
	{0x5A3197D173B30F7EULL, "915863356.ogg"},
	{0x5A31C5DF3EB4B04BULL, "dae000ip_crowd_mf2_talk1.utp"},
	{0x5A31EB88C916791DULL, "dae000_paivel6111929.cl"},
	{0x5A324226109D9740ULL, "vxlt_0_lowtown_eoy_0_0.vlm"},
	{0x5A3251E0439A5085ULL, "277610503.ogg"},
	{0x5A3291894F58CED2ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.dura.ani"},
	{0x5A32DEC43609238CULL, "6089530_f.fxe"},
	{0x5A33478236D8475AULL, "6109118_m.fxe"},
	{0x5A338F832E89AF1EULL, "6217138_m.fxe"},
	{0x5A33E038D635DFF7ULL, "vxht_bartrand_n_5tw_0_0.vlm"},
	{0x5A34288B3A03F70DULL, "vxw_encounter2_camp_ib_0_0.vlm"},
	{0x5A3455A3D6160B05ULL, "vxw_encounter2_78_0_0.vlm"},
	{0x5A346606733FCF21ULL, "6060199_m.fxe"},
	{0x5A34878E6D51D9EDULL, "gen00fl_sebastian6180836.cl"},
	{0x5A34BD3A77B4F8ADULL, "122048897.ogg"},
	{0x5A34EDA01CA58526ULL, "87457065.ogg"},
	{0x5A354BB2DC0DC05FULL, "vxht_mansion2b_323_0_0.vlm"},
	{0x5A35BA6245993EB3ULL, "vxht_keep_5kv_0_0.vlm"},
	{0x5A36736651E4BF30ULL, "vxca_caverns1_mines_edm_0_0.vlm"},
	{0x5A36C5DB3F18F0D8ULL, "mcr310_sibling6103603.cl"},
	{0x5A3738D5A9490447ULL, "lt_warehouse_n.arl"},
	{0x5A3744D5A94918EDULL, "lt_warehouse_n.anb"},
	{0x5A374C7FCBBF69C0ULL, "ser270_emeric6201565.cl"},
	{0x5A37A3595FA3941EULL, "vxht_keep_n_6hr_0_0.vlm"},
	{0x5A37D19FE2DD58DFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_ladderxa.phy"},
	{0x5A3829D7903C31CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_loop.gad"},
	{0x5A3836EA6A189642ULL, "459104054.ogg"},
	{0x5A384BF29E37A283ULL, "mer221_sabin6065654.cl"},
	{0x5A38909B0217F6F5ULL, "124317735.ogg"},
	{0x5A38E33B59264DB4ULL, "zz_qun_debug6083565.cl"},
	{0x5A3927E8DC0AF0FFULL, "327540301.ogg"},
	{0x5A3934AD7C95BED3ULL, "6104713_f.fxe"},
	{0x5A393B6EAA5CF7A0ULL, "vxw_encounter3_g2_0_0.vlm"},
	{0x5A397392F059C2B7ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_003a_n.dds"},
	{0x5A397CA5E705A453ULL, "vxlt_gallowstemplar_39a_0_0.vlm"},
	{0x5A3995234C645EFDULL, "follower_banter6161189.cl"},
	{0x5A39BBBB45E2BE2DULL, "6117947_m.fxe"},
	{0x5A3A03AB61CFE36EULL, "vxlt_0_lowtown_a2_fen_0_0.vlm"},
	{0x5A3A29671F5A0E08ULL, "6107073_m.fxe"},
	{0x5A3A5129C2BF5FD8ULL, "6089964_m.fxe"},
	{0x5A3AA3DD65E3E0C7ULL, "6084331_m.fxe"},
	{0x5A3AD0CFEAC91136ULL, "464974686.ogg"},
	{0x5A3AE300FE62563AULL, "vxlt_0_lowtown_docks_n_arv_0_0.vlm"},
	{0x5A3B01B8EE26007EULL, "6095431_m.fxe"},
	{0x5A3B245AA4033984ULL, "vxlt_0_lowtown_e0p_0_0.vlm"},
	{0x5A3B5F64C6DC0877ULL, "mer320_hubert6089956.cl"},
	{0x5A3BAC42723330F3ULL, "rdr110dg_isabela.bnk"},
	{0x5A3BD12912BE07BBULL, "vxht_bartrand_n_5hg_0_0.vlm"},
	{0x5A3BE24230C1B699ULL, "6065652_f.fxe"},
	{0x5A3BFE5CF7C54AC7ULL, "plt_cod_hst_mag_felandaris.plo"},
	{0x5A3C27F48FFFEC36ULL, "dae300_alistair_king6134951.cl"},
	{0x5A3C8791ED431A79ULL, "hm_anders_dark.mor"},
	{0x5A3D01A826281C45ULL, "hlac_lt_0_lowtown_n_131070_0y.dds"},
	{0x5A3D50575BE1AC01ULL, "llac_lt_warehouse_n_1_2.dds"},
	{0x5A3DB3367B0DEE89ULL, "621320036.ogg"},
	{0x5A3DDBE9FB1BE9F1ULL, "mag110_vincento6051739.cl"},
	{0x5A3E948ABE4C0ACFULL, "plt_zz_mer130pt_racketeers.nss"},
	{0x5A3EF0866267EAA4ULL, "seb221_flora_harimann6177671.cl"},
	{0x5A3F3ED79042747CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_loop.evt"},
	{0x5A3F459DA6160F00ULL, "vxca_caverns2_8nx_0_0.vlm"},
	{0x5A3F52FB3CC43D1EULL, "plt_cod_bks_fifth_blight_2.plo"},
	{0x5A3F6C9FA6B0A819ULL, "6081446_m.fxe"},
	{0x5A3FCE4272375343ULL, "rdr110dg_isabela.cnv"},
	{0x5A400BE883FCB92DULL, "vxw_bonepitmining_a2_17a_0_0.vlm"},
	{0x5A40107286DE78D2ULL, "104092000.ogg"},
	{0x5A403EF7EEF0F1C5ULL, "6121124_m.fxe"},
	{0x5A40AB1068F738EBULL, "vxht_0_hightown_b3y_0_0.vlm"},
	{0x5A40B4030293AFC0ULL, "vxlt_0_lowtown_ejf_0_0.vlm"},
	{0x5A40BCF8FD77D838ULL, "vxw_bonepitmining_a2_27g_0_0.vlm"},
	{0x5A40FF23A1D432E8ULL, "seb201_lady_harimann6177369.cl"},
	{0x5A41064F77B703DBULL, "6122757_f.fxe"},
	{0x5A429A682BFA0F8DULL, "vxw_encounter2_camp_da_0_0.vlm"},
	{0x5A4348A838F702F3ULL, "hlac_ca_caverns1_24424_24x.dds"},
	{0x5A43D6D868291306ULL, "vxw_bonepitmining_a2_1vq_0_0.vlm"},
	{0x5A441DC8E9910C5EULL, "vxlt_0_lowtown_docks_a3_d2g_0_0.vlm"},
	{0x5A44333D5DA7EA70ULL, "ave121tr_aveline_varric.ncs"},
	{0x5A4438E6BC3A71F1ULL, "765314139.ogg"},
	{0x5A444BCF494EA02FULL, "hlac_dr_primevaltomb_11985_1x.dds"},
	{0x5A44588B094578C1ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_onyxstatuette_d.dds"},
	{0x5A44A14C12E8FD5BULL, "llac_lt_darktown_-1_3.dds"},
	{0x5A44AD2BB4F1B03FULL, "art\\characters\\playercharacter\\humanfemale\\boo_ric.phy"},
	{0x5A44CFB155D4188BULL, "vxlt_hangedman_27g_0_0.vlm"},
	{0x5A45210B71F8CD33ULL, "llac_lt_0_lowtown_docks_n_-4_3.dds"},
	{0x5A4538AA9D515CB7ULL, "vxht_keep_n_42g_0_0.vlm"},
	{0x5A462DB62FE6D828ULL, "framebuffereffects_bloomblend_p0.psh.fxo"},
	{0x5A462F83559D46B8ULL, "6101425_f.fxe"},
	{0x5A463AAC6346BE53ULL, "lla_ht_0_hightown_n_2_0.dds"},
	{0x5A46A59F9D1E430BULL, "lgt160cr_ntb_elf_human.utc"},
	{0x5A471FF21EE648CCULL, "vxdr_slavers_1sq_0_1.vlm"},
	{0x5A47CAB4EAFF5A81ULL, "6048269_m.fxe"},
	{0x5A480256AA369EF5ULL, "1022963611.ogg"},
	{0x5A480898ACC4FB67ULL, "vxdr_primevalentrance_270_0_0.vlm"},
	{0x5A4883DC14A4719BULL, "401777759.ogg"},
	{0x5A48B7EDDCFB474CULL, "zz_approval_debug6003081.cl"},
	{0x5A48D29FE2EBCDE4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_ladderxa.mmh"},
	{0x5A48D69FE2EBD437ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_ladderxa.mao"},
	{0x5A48E89FE2EBF286ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_ladderxa.msh"},
	{0x5A49976BED5EC438ULL, "follower_barks6153508.cl"},
	{0x5A499F5A429DD241ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_5y.dds"},
	{0x5A49A95A56558486ULL, "gen00fl_varric6118750.cl"},
	{0x5A4A23A30EDC01F5ULL, "120608747.ogg"},
	{0x5A4A3FE96747B047ULL, "6020485_m.fxe"},
	{0x5A4A5286702C96B4ULL, "ico_gloves_medium_5.dds"},
	{0x5A4A80CD9E81DDB9ULL, "art\\vfx\\models\\vfx_vol_light_narrow_3.msh"},
	{0x5A4A8ACD9E81EEA7ULL, "art\\vfx\\models\\vfx_vol_light_narrow_3.mmh"},
	{0x5A4AD76DBEE65F70ULL, "vxlt_0_lowtown_h6s_0_0.vlm"},
	{0x5A4B20CCB0B430ECULL, "lt_0_lowtown_a2_n_c_34_0_ll.mmh"},
	{0x5A4B54D4EA00A9AAULL, "941346283.ogg"},
	{0x5A4B6ED422D8FBBBULL, "6035274_m.fxe"},
	{0x5A4B75F64651A73CULL, "6114073_m.fxe"},
	{0x5A4B775C03C86E23ULL, "vxlt_foundry_n_ve_0_0.vlm"},
	{0x5A4C033D364EE80FULL, "vxw_bonepitdestroyed_a3_1gu_0_0.vlm"},
	{0x5A4C11A42B27F6DBULL, "312505463.ogg"},
	{0x5A4C46D7904D0B95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_loop.ani"},
	{0x5A4CBBEF9FF340D1ULL, "vxlt_0_lowtown_a2_n_c_f6p_0_1.vlm"},
	{0x5A4DC728C5449491ULL, "mrl000_marethari6129194.cl"},
	{0x5A4DD70BC58B7AE8ULL, "w_woundedcoast_175_0.msh"},
	{0x5A4DD781D589B62CULL, "art\\levels\\races\\design\\prp_chesta_5.pwk"},
	{0x5A4DE90BC58B9979ULL, "w_woundedcoast_175_0.mao"},
	{0x5A4DF50BC58BAD12ULL, "w_woundedcoast_175_0.mmh"},
	{0x5A4E191CAC3EBF8FULL, "926243005.ogg"},
	{0x5A4F1B86993F7FAEULL, "vxw_sundermount_245_0_0.vlm"},
	{0x5A4FC074F330F770ULL, "vxht_chantry_n_2yk_0_0.vlm"},
	{0x5A4FC8768EA47179ULL, "vxw_bonepitdestroyed_a3_1q5_0_0.vlm"},
	{0x5A50E195B92A6197ULL, "vxlt_0_lowtown_docks_a3_n_d2e_0_0.vlm"},
	{0x5A50F3E31E906D80ULL, "vxht_0_hightown_a2_bx7_0_0.vlm"},
	{0x5A5131BFF9825279ULL, "vxca_cavern1_haunted_i93_0_0.vlm"},
	{0x5A518CB9E4467085ULL, "vxw_blightlands_a1_0_0.vlm"},
	{0x5A5224C23AD065ADULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_laundry.msh"},
	{0x5A522AB7347CD16FULL, "vxht_brothel_n_6bv_0_0.vlm"},
	{0x5A5232C23AD07DE0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_laundry.mao"},
	{0x5A5236C23AD08433ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_laundry.mmh"},
	{0x5A5268A901E4A461ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_reversal.ani"},
	{0x5A5296263634D868ULL, "vis121_saemus6101162.cl"},
	{0x5A5309250282081CULL, "736353896.ogg"},
	{0x5A5336F56BA453A4ULL, "em_zevran.mor"},
	{0x5A536A33A3DD8631ULL, "vxca_caverns2prc_9ut_0_0.vlm"},
	{0x5A537859057EA50DULL, "w_sundermount_2_0.mmh"},
	{0x5A537C59057EAC5EULL, "w_sundermount_2_0.mao"},
	{0x5A538A59057EC393ULL, "w_sundermount_2_0.msh"},
	{0x5A541083D4BB5767ULL, "follower_banter6163509.cl"},
	{0x5A5437E9859B2F20ULL, "vxht_mansion2_328_0_0.vlm"},
	{0x5A5452611DFA7821ULL, "6112135_m.fxe"},
	{0x5A54648052F6E1BBULL, "fadesky2.mao"},
	{0x5A54B1629526C1ABULL, "vxdr_primevaltomb_2sl_0_0.vlm"},
	{0x5A55821E1FB4532EULL, "vxca_cavern1_haunted_ei5_0_1.vlm"},
	{0x5A55AD37AF01F372ULL, "174446093.ogg"},
	{0x5A55AE2BB5002544ULL, "art\\characters\\playercharacter\\humanfemale\\boo_ric.mmh"},
	{0x5A55B650316138B2ULL, "470890518.ogg"},
	{0x5A55C42BB5004AE6ULL, "art\\characters\\playercharacter\\humanfemale\\boo_ric.msh"},
	{0x5A55C5630B923D1AULL, "zz_dre_debug6168221.cl"},
	{0x5A55CB065DB66F5FULL, "plt_cod_crt_genlock.plo"},
	{0x5A55F71FABFD51E5ULL, "645131005.ogg"},
	{0x5A56B0AA5B342675ULL, "vxdr_primevaltomb_6z9_0_0.vlm"},
	{0x5A56CF0F1BA03FE8ULL, "llac_lt_0_lowtown_a2_n_c_-65536_4.dds"},
	{0x5A57030C7FBA3FA2ULL, "plt_cod_mgc_qun.plo"},
	{0x5A57C8AA944501CEULL, "rune_armor_spirit.dds"},
	{0x5A57E4CC7A201973ULL, "vxca_caverns1_mines_icq_0_2.vlm"},
	{0x5A57FE427D8E2C4FULL, "vxlt_0_lowtown_a2_hei_0_0.vlm"},
	{0x5A580B84722D6066ULL, "vxht_mansion2_31g_0_0.vlm"},
	{0x5A58938A0BFC415EULL, "zz_vault_debug6203367.cl"},
	{0x5A58BE696ADCE07FULL, "hlac_lt_darktown_-1_54z.dds"},
	{0x5A592F48CE6056E1ULL, "629550284.ogg"},
	{0x5A5973A901EAD503ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_reversal.gad"},
	{0x5A59B6ECDBF8889CULL, "royalarcher_dmgreduction.dds"},
	{0x5A5A8A203CE3BAD6ULL, "vxca_caverns1_mines_ikj_0_0.vlm"},
	{0x5A5B0BDD18DAED30ULL, "6155724_m.fxe"},
	{0x5A5B33057CBACAB7ULL, "521213331.ogg"},
	{0x5A5B4B0823092D9FULL, "one200dg_solivitus6048175.cl"},
	{0x5A5BFDAE5D6647E3ULL, "vxht_0_hightown_a3_a6h_0_0.vlm"},
	{0x5A5C29578A024000ULL, "vxca_caverns2_8pf_0_0.vlm"},
	{0x5A5C359E1DE45269ULL, "hlac_lt_0_lowtown_a2_n_1_9y.dds"},
	{0x5A5D2AA71EE6D613ULL, "947406924.ogg"},
	{0x5A5D4142A1D83A23ULL, "vxlt_gallowstemplar_f_5ul_0_0.vlm"},
	{0x5A5D8D638B35982EULL, "vxw_sundermount_296_0_0.vlm"},
	{0x5A5DA0B2E1640938ULL, "vxw_bonepitmining_a2_29e_0_0.vlm"},
	{0x5A5DD42E06C96BD8ULL, "6142416_m.fxe"},
	{0x5A5DFF40DAF76789ULL, "6119718_m.fxe"},
	{0x5A5E3251E5270FF0ULL, "vxht_chantry_n_2th_0_0.vlm"},
	{0x5A5E387139584217ULL, "follower_banter6154060.cl"},
	{0x5A5E3A53809A89F9ULL, "vxw_bonepitdestroyed_a3_1z4_0_0.vlm"},
	{0x5A5E78590588AC86ULL, "w_sundermount_2_0.rml"},
	{0x5A5E934C386F7A0AULL, "vis160cr_winter_warrior.utc"},
	{0x5A5EA885D41CCD1BULL, "gen_im_npc_elf_arm_glv_dalb.uti"},
	{0x5A5F31AA50D4529FULL, "hlac_dr_primevaltomb_661_5y.dds"},
	{0x5A5FD146DDAB6576ULL, "6120747_f.fxe"},
	{0x5A5FD2B8C8CC1E39ULL, "vxht_keep_n_6f7_0_0.vlm"},
	{0x5A5FDA1AE58832D4ULL, "vxlt_gallowstemplar_3sp_0_0.vlm"},
	{0x5A5FFE96D63C8905ULL, "vxw_blightlands_j0_0_0.vlm"},
	{0x5A60086A604E8C40ULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_1z.dds"},
	{0x5A6050A901F0B8E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_reversal.evt"},
	{0x5A60872CF6A4C28EULL, "hlac_lt_0_lowtown_a2_n_65536_12x.dds"},
	{0x5A60CBFC6FCE4EDDULL, "vxca_caverns2_77m_0_0.vlm"},
	{0x5A615F637A49294BULL, "vxw_woundedcoast_165_0_0.vlm"},
	{0x5A6168632E05A111ULL, "vxht_0_hightown_al9_0_0.vlm"},
	{0x5A61E744724B2772ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arrow_act_c.mmh"},
	{0x5A62966D3F96EC00ULL, "697110729.ogg"},
	{0x5A62A279BB485AD0ULL, "mag100_karras6034518.cl"},
	{0x5A638D49419A9807ULL, "llac_lt_0_lowtown_docks_a3_n_65533_1.dds"},
	{0x5A63ED58F33275C7ULL, "fen310_party_comment6116847.cl"},
	{0x5A640ECAFB9F117FULL, "mag210_arianni6031659.cl"},
	{0x5A64650D5125CB8DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos06a.phy"},
	{0x5A646FFE0D4C0C19ULL, "vxht_mansion2_25e_0_0.vlm"},
	{0x5A64CCA0D1DD4FAEULL, "gen00fl_aveline6139884.cl"},
	{0x5A653B19D8B691CAULL, "783147606.ogg"},
	{0x5A6541F4D8D5739FULL, "1070698747.ogg"},
	{0x5A656B42F072C035ULL, "ave331cr_insurgent_m2.utc"},
	{0x5A66194759FFFA03ULL, "6048847_m.fxe"},
	{0x5A6639F2B603F37EULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_groundivy_n.dds"},
	{0x5A66701F6F8444CFULL, "vxlt_0_lowtown_a2_hbh_0_0.vlm"},
	{0x5A667CCEECE28B3CULL, "gen00fl_anders6103912.cl"},
	{0x5A66FB7DEFFAE20AULL, "344504248.ogg"},
	{0x5A67022F2517319DULL, "309558711.ogg"},
	{0x5A671E8304C0FBEFULL, "6033719_m.fxe"},
	{0x5A67924BB18ACEDEULL, "6093824_m.fxe"},
	{0x5A683AE81A16F79CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_boosterprops_sa.dds"},
	{0x5A685870D8A2D788ULL, "6147428_m.fxe"},
	{0x5A68E024FF9DF6ABULL, "6042302_m.fxe"},
	{0x5A69223BE68EC3E5ULL, "6112100_m.fxe"},
	{0x5A693E2F184D3D35ULL, "426312717.ogg"},
	{0x5A6A7F14AC21F849ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefruit.phy"},
	{0x5A6ABCDA0CA03F7BULL, "739477613.ogg"},
	{0x5A6ACD15B13A3546ULL, "6090043_m.fxe"},
	{0x5A6B28E40F22D5CAULL, "lt_0_lowtown_a3_n_c_ll.mmh"},
	{0x5A6B35D069090FC4ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.anb"},
	{0x5A6BB31F93CE52A3ULL, "vxlt_gallowstemplar_f_5xo_0_0.vlm"},
	{0x5A6BBF51623EBA66ULL, "919156975.ogg"},
	{0x5A6BC7EEADA6374DULL, "vxht_bartrand_n_755_0_1.vlm"},
	{0x5A6BF8F2CBBCA381ULL, "vxlt_gallowstemplar_53k_0_0.vlm"},
	{0x5A6C711DCCED8009ULL, "6114018_m.fxe"},
	{0x5A6C9D1962FB6DE7ULL, "6079152_f.fxe"},
	{0x5A6E123858566A9BULL, "hlac_lt_0_lowtown_a3_n_c_65536_1x.dds"},
	{0x5A6E1579DD6E8379ULL, "vxca_cavern1_haunted_i31_0_0.vlm"},
	{0x5A6E7C45F081BD7CULL, "6085515_m.fxe"},
	{0x5A6EB98711BB223AULL, "vxlt_0_lowtown_a2_n_eqj_0_0.vlm"},
	{0x5A6EBB0A4651DBF9ULL, "mag370_meredith6092169.cl"},
	{0x5A6F0E711869026FULL, "vxht_brothel_n_6hp_0_0.vlm"},
	{0x5A6F272708D947CEULL, "6125826_m.fxe"},
	{0x5A6F96795EFEE1E9ULL, "gen00fl_aveline6137502.cl"},
	{0x5A6FB5FA6F4EF344ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_ian.gad"},
	{0x5A6FD1406C3F41CBULL, "vxw_woundedcoast_114_0_0.vlm"},
	{0x5A6FDC4A82C4365FULL, "ser221_gascard6128276.cl"},
	{0x5A7024FBF3ECF9FCULL, "832407550.ogg"},
	{0x5A702BA421019A0BULL, "dae000cr_grand_cleric6199337.cl"},
	{0x5A702BBFC5FBD6AFULL, "55258509.ogg"},
	{0x5A706DFB15D36926ULL, "dae130st_morton.stg"},
	{0x5A708E9F94D2BF77ULL, "vxlt_gallowscourtyard_3w1_0_0.vlm"},
	{0x5A713E2DF57A82CDULL, "vxht_brothel_n_3r5_0_0.vlm"},
	{0x5A71DF507FE0F7CCULL, "vxht_keep_n_622_0_0.vlm"},
	{0x5A71F5C0759D2DDDULL, "6201523_m.fxe"},
	{0x5A7292482D813A66ULL, "vxlt_warehouse_qv_0_0.vlm"},
	{0x5A72CC714D8470B7ULL, "vxw_woundedcoast_fl_0_0.vlm"},
	{0x5A73E6F072E42DA6ULL, "vxht_mansion2b_1hf_0_0.vlm"},
	{0x5A73E824F39F12F2ULL, "fen110_fenris6101964.cl"},
	{0x5A740621B7183ABBULL, "58124230.ogg"},
	{0x5A74B022597E4ABAULL, "gen00fl_aveline6120846.cl"},
	{0x5A74BE5FD25832D4ULL, "art\\levels\\4cinematics\\ferelden\\prp_deadmeredith.phy"},
	{0x5A74CF45DA88A3BCULL, "hlac_dr_dungeon_3207_4z.dds"},
	{0x5A7536C18969FE0FULL, "vxdr_primevaltomb_8dk_0_0.vlm"},
	{0x5A754C7294C82C28ULL, "hlac_dr_primevaltomb_659_5y.dds"},
	{0x5A75500D0E7416ABULL, "vxht_playershouse_n_1wl_0_0.vlm"},
	{0x5A7555802F7F8834ULL, "vxht_mansion_n_6nz_0_0.vlm"},
	{0x5A75826673DFE0E6ULL, "6077787_m.fxe"},
	{0x5A75A030AAEA2651ULL, "vxlt_0_lowtown_docks_a3_cwx_0_0.vlm"},
	{0x5A75C4FA6F537822ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_ian.ani"},
	{0x5A75F1E7D894713CULL, "6121980_f.fxe"},
	{0x5A7629DDDD39DE97ULL, "124085506.ogg"},
	{0x5A764388F93998C8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_left.mmh"},
	{0x5A764F88F939AD6BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_left.mao"},
	{0x5A766188F939CBF2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_left.msh"},
	{0x5A766E86BBBF036CULL, "280222930.ogg"},
	{0x5A7688BA1F0B0E3AULL, "hlac_ht_keep_3732_11z.dds"},
	{0x5A7795C82C202C06ULL, "hlac_ht_0_hightown_n_65535_0z.dds"},
	{0x5A77C1B11F9F18DEULL, "6076278_m.fxe"},
	{0x5A77F8FCAEF06248ULL, "233071897.ogg"},
	{0x5A7854D06913CEC7ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.mmh"},
	{0x5A786109C2F725E9ULL, "art\\vfx\\combat\\vfx_c_rw_arc_blast_c.impa.ani"},
	{0x5A7876F3DD4933C1ULL, "6036955_f.fxe"},
	{0x5A78FC5630D24373ULL, "mag320_orsino6071172.cl"},
	{0x5A793BA2CECF5620ULL, "hlac_lt_foundry_n_320_16z.dds"},
	{0x5A795602368F22F3ULL, "375942481.ogg"},
	{0x5A7A24FC85C46B23ULL, "vxlt_gallowstemplar_f_5cn_0_0.vlm"},
	{0x5A7A42BCF59DEC19ULL, "tbe100st_crafted_goods.utm"},
	{0x5A7A708FE14568E8ULL, "6095972_m.fxe"},
	{0x5A7B0FC139FD84BFULL, "410364740.ogg"},
	{0x5A7B5276CAFC2159ULL, "ser270_cut_aftermath6162281.cl"},
	{0x5A7BCD9FB4DD5A09ULL, "ave310_brennan6137007.cl"},
	{0x5A7C0DE91969B738ULL, "llac_dr_primevaltomb_658_1.dds"},
	{0x5A7C43A9AAD837FDULL, "good_large.mao"},
	{0x5A7CEC15625C70A6ULL, "6106182_m.fxe"},
	{0x5A7D2A1F8CE2BE82ULL, "vxht_0_hightown_n_c1t_0_0.vlm"},
	{0x5A7D804E0A5F1AEFULL, "vxht_brothel_n_6us_0_0.vlm"},
	{0x5A7DF73FFDA4DD96ULL, "llac_lt_0_lowtown_65536_6.dds"},
	{0x5A7F1677188AE8FDULL, "51187025.ogg"},
	{0x5A7F49BAB2B8C57DULL, "116307746.ogg"},
	{0x5A7F4B1DDFC4604CULL, "6105916_m.fxe"},
	{0x5A7F7C09C2FD726BULL, "art\\vfx\\combat\\vfx_c_rw_arc_blast_c.impa.gad"},
	{0x5A7F9B5FD261FED1ULL, "art\\levels\\4cinematics\\ferelden\\prp_deadmeredith.msh"},
	{0x5A7FA95FD26216ECULL, "art\\levels\\4cinematics\\ferelden\\prp_deadmeredith.mao"},
	{0x5A7FB55FD2622A8FULL, "art\\levels\\4cinematics\\ferelden\\prp_deadmeredith.mmh"},
	{0x5A7FEA97FC93FF87ULL, "sse_ss_shadow_f_1.gda"},
	{0x5A8004BF224ADB4AULL, "lgt200cr_tome_corpse.utc"},
	{0x5A80BFF045764A96ULL, "art\\levels\\4cinematics\\ferelden\\letterback.mao"},
	{0x5A80D49F6656C35DULL, "6130113_m.fxe"},
	{0x5A8104251F7752E6ULL, "vxlt_warehouse_zw_0_0.vlm"},
	{0x5A810D9FFF036E84ULL, "39341093.ogg"},
	{0x5A813E4E3F7A8937ULL, "vxw_woundedcoast_co_0_0.vlm"},
	{0x5A8157C7DD4561A8ULL, "vxlt_undercityserialkiller_xb_0_0.vlm"},
	{0x5A81699B68666741ULL, "319693864.ogg"},
	{0x5A817F5B7F32C2F2ULL, "dae210_gate_guard6209976.cl"},
	{0x5A81B4D99A9C3857ULL, "6147247_m.fxe"},
	{0x5A81BAF61B9D776FULL, "vxw_sundermount_26f_0_0.vlm"},
	{0x5A81C33CB481F62CULL, "962637337.ogg"},
	{0x5A81C374C5B8D0D4ULL, "6180383_f.fxe"},
	{0x5A8210251DBBCF40ULL, "6143969_m.fxe"},
	{0x5A8258CD64DA4626ULL, "vxht_mansion2b_1cg_0_0.vlm"},
	{0x5A829F47FB04EE13ULL, "6071861_m.fxe"},
	{0x5A82CDE58DF8044FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_out.gad"},
	{0x5A830C7C4C0618FDULL, "6069628_m.fxe"},
	{0x5A8411C7F16AF23CULL, "w_sundermount_219_0.mao"},
	{0x5A841DC7F16B06DFULL, "w_sundermount_219_0.mmh"},
	{0x5A8423C7F16B1081ULL, "w_sundermount_219_0.msh"},
	{0x5A852C296B63089FULL, "vxht_mansion2b_1a4_0_0.vlm"},
	{0x5A8550BA799F1E99ULL, "fex_ruin_circ_c.mmh"},
	{0x5A855ABA799F2F87ULL, "fex_ruin_circ_c.msh"},
	{0x5A85732238DADDEAULL, "vxlt_gallowscourtyard_6uc_0_0.vlm"},
	{0x5A85E3B735E12CAEULL, "6036418_m.fxe"},
	{0x5A86169F0FBF8446ULL, "gen00fl_bethany6109226.cl"},
	{0x5A861AD64D954C02ULL, "dae000_loyal_mage6165243.cl"},
	{0x5A867909C3038BD0ULL, "art\\vfx\\combat\\vfx_c_rw_arc_blast_c.impa.evt"},
	{0x5A867FB04315BDC5ULL, "w_blightlands_7_0.rml"},
	{0x5A86D417A4CA6D66ULL, "6110920_f.fxe"},
	{0x5A876D5E2491ED2EULL, "vxw_encounter2_5a_0_0.vlm"},
	{0x5A87F2485D5F8915ULL, "vxht_mansion_n_6s8_0_0.vlm"},
	{0x5A88E666B1D24093ULL, "ht_0_hightown_n_51_0_ll.mmh"},
	{0x5A88EA3B4D567090ULL, "qun110_petrice6018673.cl"},
	{0x5A893234A05AAB0BULL, "630656715.ogg"},
	{0x5A89649A288E55CDULL, "117370749.ogg"},
	{0x5A89880FE0AB481DULL, "6090307_m.fxe"},
	{0x5A89959C20B81B96ULL, "hlac_lt_0_lowtown_a2_-1_5x.dds"},
	{0x5A89C658F843301CULL, "plt_dae000pt_location.nss"},
	{0x5A89CEDA85EA66E9ULL, "zz_and_debug6178661.cl"},
	{0x5A89DD602F98AC80ULL, "fen110_fenris6118010.cl"},
	{0x5A89E2E58DFE46FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_out.evt"},
	{0x5A89E644726C7719ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arrow_act_c.anb"},
	{0x5A89F8C9936A2507ULL, "dae000_meredith6085300.cl"},
	{0x5A8ABB56E668CF80ULL, "vxht_0_hightown_a2_bl3_0_0.vlm"},
	{0x5A8AF0037B66D760ULL, "6054976_f.fxe"},
	{0x5A8B67E5F6D6EA8AULL, "6050321_f.fxe"},
	{0x5A8BA185FE3C94AFULL, "ht_0_hightown_a2_mmap_ph.dds"},
	{0x5A8BCE95C639D561ULL, "zz_dre_debug6168286.cl"},
	{0x5A8C42A3FD316F29ULL, "gen00fl_fenris6110913.cl"},
	{0x5A8CE8C695A6C0BCULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2_0.pwk"},
	{0x5A8D396E01967CD2ULL, "6086930_m.fxe"},
	{0x5A8D564CDAC17BE4ULL, "vxlt_gallowstemplar_f_4o0_0_0.vlm"},
	{0x5A8D5A6041FAB6ACULL, "hlac_dr_slavers_3216_11z.dds"},
	{0x5A8DB25743FF5929ULL, "6135891_m.fxe"},
	{0x5A8DEE646F6CA5DDULL, "673723865.ogg"},
	{0x5A8DF7F16CB227CCULL, "follower_banter6173088.cl"},
	{0x5A8E29F52C6C8EF4ULL, "cod_shw_mer_notes_1.utp"},
	{0x5A8E3CF3F4C36D27ULL, "6154500_m.fxe"},
	{0x5A8E6EE16FF036FAULL, "239756019.ogg"},
	{0x5A8E8768A03E41BDULL, "6104302_f.fxe"},
	{0x5A8EC2184B103462ULL, "962315237.ogg"},
	{0x5A8EDC622F481190ULL, "488604688.ogg"},
	{0x5A8EE230A34ADBC8ULL, "dae000_gamlen6145106.cl"},
	{0x5A8EF0741DB4AF4CULL, "420178738.ogg"},
	{0x5A8F2C6EB3D63DDCULL, "follower_banter6136731.cl"},
	{0x5A8F3C678895739BULL, "qun231_varnell6077275.cl"},
	{0x5A8FE5D51152C2F2ULL, "mcr370_meredith6083080.cl"},
	{0x5A8FEA8DA8548DDBULL, "hlac_ht_0_hightown_a2_65533_0z.dds"},
	{0x5A9026B68C9250D7ULL, "6140347_m.fxe"},
	{0x5A90CAAA56D05EA6ULL, "vxht_mansion2b_1fh_0_0.vlm"},
	{0x5A910D3809084247ULL, "hlac_ht_mansion2b_818_67x.dds"},
	{0x5A91B4B760CB7F32ULL, "734728704.ogg"},
	{0x5A91DAE178BF9E9FULL, "dae011_wench_mob6149855.cl"},
	{0x5A922A3D5DE9FC7DULL, "ave121tr_aveline_varric.utt"},
	{0x5A923C1B7CABFFDAULL, "6085889_m.fxe"},
	{0x5A9245898CAF6944ULL, "zz_dae_debug6047606.cl"},
	{0x5A926B3D3686218AULL, "gen00fl_varric6033646.cl"},
	{0x5A92E04B4FF05E4BULL, "art\\characters\\creatures\\spider\\animation\\die_v1.ani"},
	{0x5A92E6887A53688DULL, "blood_conj_02_l2.mao"},
	{0x5A932356EDE8011CULL, "vxlt_0_lowtown_docks_a3_n_cnm_0_0.vlm"},
	{0x5A932BB5AB4AF1D2ULL, "ts_reactive_injurykit.gda"},
	{0x5A9344E756DA65DAULL, "hlac_ht_mansion2b_818_24x.dds"},
	{0x5A9395FC15007ED9ULL, "156659144.ogg"},
	{0x5A93DC15EE03FE21ULL, "vxht_mansion_n_6r7_0_0.vlm"},
	{0x5A93F9F6A8347257ULL, "mag121_idunna6084866.cl"},
	{0x5A940E71CA11362CULL, "ave121_jeven6105508.cl"},
	{0x5A942190C9B70DEEULL, "6106933_m.fxe"},
	{0x5A94603FC2162DD0ULL, "6134757_f.fxe"},
	{0x5A9480D408F79456ULL, "hla_ht_0_hightown_a2_131070_0x.dds"},
	{0x5A94BD25E0FE915BULL, "vxht_mansion2b_2eg_0_0.vlm"},
	{0x5A94D2E1FA81EFDFULL, "120755515.ogg"},
	{0x5A95134570FC1B3BULL, "hlac_lt_0_lowtown_docks_131071_5y.dds"},
	{0x5A958D5C86CB7E61ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyelids.gad"},
	{0x5A95DE4C975E40C3ULL, "vxht_mansion2_1dk_0_0.vlm"},
	{0x5A9663667FE8408FULL, "9709589.ogg"},
	{0x5A96EAE58E08DE15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_out.ani"},
	{0x5A96FC0BB6716EE4ULL, "vxht_bartrand_n_272_0_0.vlm"},
	{0x5A974D286DADDAE0ULL, "vxw_bonepitmining_a2_1wy_0_1.vlm"},
	{0x5A976814AC48BF6EULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefruit.mmh"},
	{0x5A977414AC48D305ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefruit.mao"},
	{0x5A978214AC48EB5CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefruit.msh"},
	{0x5A97BB30C2B5A315ULL, "292339858.ogg"},
	{0x5A9819CD8601AE65ULL, "480436978.ogg"},
	{0x5A985ABBB54075ADULL, "follower_banter6160869.cl"},
	{0x5A986E96FCB20EF2ULL, "213992376.ogg"},
	{0x5A98B4D67F6561F9ULL, "164982212.ogg"},
	{0x5A98D76487D47353ULL, "516070417.ogg"},
	{0x5A994F99CB9563E8ULL, "63083943.ogg"},
	{0x5A99B182B7ACBBCEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasea_sa.dds"},
	{0x5A9A0A64CFF425B9ULL, "6052777_f.fxe"},
	{0x5A9A4B9E5ECDD5FAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc01a.phy"},
	{0x5A9AB86096FD5BEEULL, "llac_lt_backalley_n_0_1.dds"},
	{0x5A9B7D49D2FDAD95ULL, "llac_lt_undercityserialkiller_-1_43.dds"},
	{0x5A9B8E5C86CFEB0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyelids.ani"},
	{0x5A9BC2354E10D769ULL, "6148110_m.fxe"},
	{0x5A9BC656B3CBFBE6ULL, "821748578.ogg"},
	{0x5A9BC829CCB79464ULL, "vxlt_gallowstemplar_f_4t1_0_0.vlm"},
	{0x5A9C2D7823F9AE2EULL, "dae020ip_merch_weapon.utp"},
	{0x5A9C558ADB233928ULL, "442689843.ogg"},
	{0x5A9C82219A6F1052ULL, "847701106.ogg"},
	{0x5A9CB95059E05863ULL, "hlac_lt_0_lowtown_n_131070_1x.dds"},
	{0x5A9CDB5459CEA23CULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_c.impa.ani"},
	{0x5A9CDC16CE82F323ULL, "dae000_gamlen6005764.cl"},
	{0x5A9D0C03315DAC39ULL, "vxlt_hangedman_2xq_0_1.vlm"},
	{0x5A9D3853FB43DBBDULL, "6035917_m.fxe"},
	{0x5A9DE80B1A3C40ABULL, "vis121_saemus6089641.cl"},
	{0x5A9E22082366BA37ULL, "vxw_woundedcoast_ui_0_0.vlm"},
	{0x5A9E95F784FC1D00ULL, "vxlt_gallowstemplar_5pk_0_0.vlm"},
	{0x5A9ECF40DCCEA1D7ULL, "lt_0_lowtown_docks_a3_n_214_0.mmh"},
	{0x5A9ED340DCCEA8E4ULL, "lt_0_lowtown_docks_a3_n_214_0.mao"},
	{0x5A9EE540DCCEC749ULL, "lt_0_lowtown_docks_a3_n_214_0.msh"},
	{0x5A9F18502D5CFDDEULL, "vxlt_undercityencounter_1ju_0_0.vlm"},
	{0x5A9F3C8748C67726ULL, "vxht_mansion2b_1yi_0_0.vlm"},
	{0x5A9FB044ED09BB22ULL, "isa311_isabela6092492.cl"},
	{0x5AA08C585F4C478FULL, "vxdr_primevaltomb_8ut_0_0.vlm"},
	{0x5AA0C84B4FFC72C2ULL, "art\\characters\\creatures\\spider\\animation\\die_v1.evt"},
	{0x5AA0D6C54C415872ULL, "hlac_lt_0_lowtown_a2_n_131073_0y.dds"},
	{0x5AA0D8A0846E986BULL, "vxdr_primevaltomb_6xm_0_0.vlm"},
	{0x5AA0DDEC3C324E33ULL, "hlac_lt_gallowstemplar_7198_1y.dds"},
	{0x5AA1DD98B0085041ULL, "lgt220_shp_mag_guard6149507.cl"},
	{0x5AA251C7DDD0CE54ULL, "vxht_0_hightown_at8_0_0.vlm"},
	{0x5AA25B1D8A2C9AD4ULL, "vxca_caverns2_9a4_0_0.vlm"},
	{0x5AA2A7395BB5C2E9ULL, "zz_approval_debug6003518.cl"},
	{0x5AA33663462B7D2FULL, "vxht_0_hightown_n_c73_0_0.vlm"},
	{0x5AA38B64B34CF7E0ULL, "zz_seb_debug6215768.cl"},
	{0x5AA3B34884181FE0ULL, "743115359.ogg"},
	{0x5AA3E85459D4D77EULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_c.impa.gad"},
	{0x5AA49ED40E05EC6FULL, "313270508.ogg"},
	{0x5AA4DAD526548872ULL, "822763008.ogg"},
	{0x5AA53B2F680CFAD1ULL, "lgt241ip_at_bonepit.utp"},
	{0x5AA54ACA96251B2FULL, "221907530.ogg"},
	{0x5AA56E75B139E5CCULL, "450566574.ogg"},
	{0x5AA57A935BA6B4A3ULL, "vxlt_gallowstemplar_f_5ls_0_0.vlm"},
	{0x5AA660295FEA7F46ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\bsh_wideleaf_e.dds"},
	{0x5AA68062442BF4ECULL, "hlac_ht_0_hightown_-65537_3y.dds"},
	{0x5AA6A8F768EA7C81ULL, "6176120_m.fxe"},
	{0x5AA6C72A3A9D19B1ULL, "391704962.ogg"},
	{0x5AA6C74B5000DCD5ULL, "art\\characters\\creatures\\spider\\animation\\die_v1.gad"},
	{0x5AA7BCD304D16C23ULL, "w_woundedcoast_72_0.mao"},
	{0x5AA7C0D304D17370ULL, "w_woundedcoast_72_0.mmh"},
	{0x5AA7CB5601A1EE58ULL, "vxht_0_hightown_a3_a7x_0_0.vlm"},
	{0x5AA7CED304D18A8AULL, "w_woundedcoast_72_0.msh"},
	{0x5AA7DF1340450E68ULL, "6082600_m.fxe"},
	{0x5AA80DA06C254274ULL, "517920798.ogg"},
	{0x5AA822E2A0F8147EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masd_0d.dds"},
	{0x5AA9432A156018DCULL, "vxht_0_hightown_a2_bwj_0_0.vlm"},
	{0x5AAA88B1A6B1F1D2ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkleft.gad"},
	{0x5AAAC35459DAB69DULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_c.impa.evt"},
	{0x5AAAE9BD114E242BULL, "6101075_f.fxe"},
	{0x5AAAED43C3D2063CULL, "99417075.ogg"},
	{0x5AAC4C5DD9CD5E22ULL, "zz_followers_banter6158766.cl"},
	{0x5AAC5FB0433656EEULL, "w_blightlands_7_0.mmh"},
	{0x5AAC6BB043366A85ULL, "w_blightlands_7_0.mao"},
	{0x5AAC79B0433682DCULL, "w_blightlands_7_0.msh"},
	{0x5AAC93E5155CD2B7ULL, "vxw_woundedcoast_rh_0_0.vlm"},
	{0x5AAD1A970949524BULL, "6077672_m.fxe"},
	{0x5AAD3EB0161C187BULL, "vxdr_0_deeproads_ew5_0_0.vlm"},
	{0x5AAEA662B92EBF32ULL, "vxlt_gallowscourtyard_n_64w_0_0.vlm"},
	{0x5AAEB841ACDD1CB1ULL, "dae011_corff_sta6152742.cl"},
	{0x5AAEE3122253B881ULL, "gen00fl_sebastian6179793.cl"},
	{0x5AAF4F2E9121EE75ULL, "hlac_lt_0_lowtown_docks_-65536_0y.dds"},
	{0x5AAFB634C0F2F42BULL, "786559371.ogg"},
	{0x5AB00710D1D4321CULL, "vxlt_0_lowtown_docks_a3_n_ctc_0_0.vlm"},
	{0x5AB031A99B8BBCC0ULL, "plt_lgt150pt_nat_werewolf.plo"},
	{0x5AB057E34827389DULL, "seb201_lady_harimann6177425.cl"},
	{0x5AB05BB22E9B7128ULL, "742208274.ogg"},
	{0x5AB08BB1A6B66209ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkleft.evt"},
	{0x5AB0B6077AB52F1BULL, "vxw_bonepit_mh_0_0.vlm"},
	{0x5AB0FA8CDACA2387ULL, "vxlt_backalley_g_1wz_0_1.vlm"},
	{0x5AB13176BA9F1C4DULL, "zz_fen_debug6165692.cl"},
	{0x5AB1416232E32B08ULL, "6107222_m.fxe"},
	{0x5AB22DB01A519712ULL, "301197847.ogg"},
	{0x5AB2547671651BA8ULL, "798264242.ogg"},
	{0x5AB29CBF71F8D34DULL, "ave310_donnic6062076.cl"},
	{0x5AB2AEB771926D0CULL, "art\\characters\\weapons\\staff\\w_stf_006a.msh"},
	{0x5AB2B4B7719277FEULL, "art\\characters\\weapons\\staff\\w_stf_006a.mmh"},
	{0x5AB2C0B771928B95ULL, "art\\characters\\weapons\\staff\\w_stf_006a.mao"},
	{0x5AB2E1ED9112FE72ULL, "vxht_0_hightown_a2_c4h_0_0.vlm"},
	{0x5AB2EFBFF41160FFULL, "hlac_lt_0_lowtown_docks_a3_n_-196608_0x.dds"},
	{0x5AB313313A6A8180ULL, "vxlt_0_lowtown_a2_hg0_0_0.vlm"},
	{0x5AB3165124B86B2EULL, "vxw_blightlands_hi_0_0.vlm"},
	{0x5AB388AFB54AED88ULL, "mag100_cullen6092276.cl"},
	{0x5AB3D13A1F9836A2ULL, "hlac_lt_0_lowtown_65538_4z.dds"},
	{0x5AB401CB92A9BEADULL, "6166176_m.fxe"},
	{0x5AB5310EE004D166ULL, "862494986.ogg"},
	{0x5AB5B46E7074ED35ULL, "hlac_lt_gallowstemplar_n_2394_11z.dds"},
	{0x5AB69259CC1371CFULL, "6111118_m.fxe"},
	{0x5AB6B57F16EE9E5FULL, "dae000_merchant_generic_hm6172587.cl"},
	{0x5AB6F56700BD0A49ULL, "919867342.ogg"},
	{0x5AB7B5070756315CULL, "vxht_0_hightown_a2_bxk_0_0.vlm"},
	{0x5AB7BBE28A71CD15ULL, "hm_amb_old_gossip2.mor"},
	{0x5AB7DA6F3F49B723ULL, "ht_mansion_n_3370_ll.mmh"},
	{0x5AB82B116D2E1365ULL, "804442331.ogg"},
	{0x5AB85548AD86B675ULL, "t1_stb_gry.tnt"},
	{0x5AB8F97FFF7230FCULL, "vxht_mansion2_3go_0_0.vlm"},
	{0x5AB923AC0ACEBD9FULL, "dae000_mother6004201.cl"},
	{0x5AB96661AF40DCA9ULL, "hlac_lt_0_lowtown_a2_n_131071_2z.dds"},
	{0x5AB9B756F7DC92D6ULL, "gen00fl_anders6111346.cl"},
	{0x5AB9C5BEC974CFAFULL, "6047672_f.fxe"},
	{0x5ABA0F6E90ABD088ULL, "em_mag_controlled_1.mor"},
	{0x5ABA80BC7BDB29A3ULL, "286489844.ogg"},
	{0x5ABB09BC23C6FE5DULL, "360725800.ogg"},
	{0x5ABB0CAD161ADBD6ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_5_c.impa.ani"},
	{0x5ABB28507BFAB345ULL, "zz_lgt_1_debug6114364.cl"},
	{0x5ABB6DAB953BD72DULL, "zz_ser_debug6086475.cl"},
	{0x5ABB899C7D5CACADULL, "6089741_m.fxe"},
	{0x5ABB9F515462CC3EULL, "lt_gamlenshouse_n_3567_lc.mmh"},
	{0x5ABBE165EDEDA8F9ULL, "176282515.ogg"},
	{0x5ABCA5D880AFECE7ULL, "vxca_cavern1_haunted_k2r_0_0.vlm"},
	{0x5ABCCB59A213E87BULL, "6042232_m.fxe"},
	{0x5ABD38F961E22E4AULL, "qcr000_arishok6060198.cl"},
	{0x5ABD742024FD1E54ULL, "6116000_m.fxe"},
	{0x5ABE73B1A6C27778ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkleft.ani"},
	{0x5ABEF39D333B6A9FULL, "vxht_mansion2b_1u0_0_0.vlm"},
	{0x5ABF09CB8324396DULL, "gen00fl_varric6073755.cl"},
	{0x5AC0E1FA1DCE22E5ULL, "gen_im_trs_gem_snowobsid_03.uti"},
	{0x5AC153CA830916F2ULL, "vxht_0_hightown_a2_c3i_0_0.vlm"},
	{0x5AC18939485A611FULL, "314749360.ogg"},
	{0x5AC1AFA182036D6EULL, "vxw_encounter1_dk_0_0.vlm"},
	{0x5AC222E636C465F2ULL, "hlac_w_encounter2_131076_0x.dds"},
	{0x5AC25F4D53B14B08ULL, "vxlt_0_lowtown_a3_n_c_eot_0_1.vlm"},
	{0x5AC2894D75FE905AULL, "hlac_dr_slavers_3207_4z.dds"},
	{0x5AC2BD58497F3A0BULL, "vxlt_foundry_n_10h_0_0.vlm"},
	{0x5AC2BF07086775E4ULL, "vxlt_darktown_582_0_1.vlm"},
	{0x5AC38D0DF54C1B0AULL, "hlac_lt_0_lowtown_docks_-262144_0x.dds"},
	{0x5AC41E5619DDD02BULL, "vxlt_undercityserialkiller_195_0_0.vlm"},
	{0x5AC53E42C35A998FULL, "mag360_thrask6044281.cl"},
	{0x5AC5ACF31DEE82F1ULL, "432835334.ogg"},
	{0x5AC624B534048A0BULL, "664161508.ogg"},
	{0x5AC645DD883556F2ULL, "hightown_fem_noble2_mob6169539.cl"},
	{0x5AC6B80BE49B4217ULL, "493406067.ogg"},
	{0x5AC6BA33C64AD98DULL, "mrl150_merrill6170448.cl"},
	{0x5AC6EC82780E77FAULL, "mer110_jansen6032599.cl"},
	{0x5AC7FD4B9B83DA6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_self.gad"},
	{0x5AC814AD162573C7ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_5_c.impa.evt"},
	{0x5AC8C75070801E7AULL, "zz_followers_banter6159408.cl"},
	{0x5AC93C0F2EA903A9ULL, "vxw_encounter3_ft_0_0.vlm"},
	{0x5AC9AE101DF9ECA0ULL, "vxw_encounter3_58_0_0.vlm"},
	{0x5AC9B135F9B9DF1FULL, "6108130_m.fxe"},
	{0x5AC9EF475AA29953ULL, "vxlt_gallowstemplar_3kk_0_0.vlm"},
	{0x5AC9F6265D79EBDDULL, "qun120_arishok6053235.cl"},
	{0x5ACA23A289602969ULL, "gen00fl_anders6114502.cl"},
	{0x5ACAE9C2FCB0AE65ULL, "21292614.ogg"},
	{0x5ACB0F5A837C4C41ULL, "llac_ht_chantry_n_2534_12.dds"},
	{0x5ACB16FBF5533CA9ULL, "vxht_keep_6ya_0_0.vlm"},
	{0x5ACB5503873E23C1ULL, "gen00fl_varric6019889.cl"},
	{0x5ACBC20FEA58CF74ULL, "vxlt_backalley_n_1k9_0_0.vlm"},
	{0x5ACC0C2482C95854ULL, "6082147_m.fxe"},
	{0x5ACC261C25ED3BBEULL, "hlac_lt_gallowstemplar_n_2383_6y.dds"},
	{0x5ACC59DF41F48CD9ULL, "6092091_m.fxe"},
	{0x5ACC6151B758B69CULL, "hlac_lt_0_lowtown_docks_196606_0x.dds"},
	{0x5ACC9FB9171A0B18ULL, "vxca_caverns1_gk3_0_0.vlm"},
	{0x5ACCA00AEC16FCB1ULL, "gen00fl_isabela6075267.cl"},
	{0x5ACCCC8D7495067EULL, "106588604.ogg"},
	{0x5ACCEACAB5C0631CULL, "vxlt_0_lowtown_docks_a3_n_cza_0_0.vlm"},
	{0x5ACD07AD0C1A2A48ULL, "188420634.ogg"},
	{0x5ACD1977BA76C21EULL, "hlac_dr_primevaltomb_657_2x.dds"},
	{0x5ACD2FC82BD65905ULL, "qun110_petrice_aftermath.stg"},
	{0x5ACD3A61BD516914ULL, "qun200_party_comment6090341.cl"},
	{0x5ACD81AB0A8C1DE6ULL, "vxw_encounter3_6g_0_0.vlm"},
	{0x5ACE51BD4458F646ULL, "vxw_woundedcoast_10j_0_0.vlm"},
	{0x5ACE583E1A64D4B1ULL, "vxlt_0_lowtown_docks_bf9_0_0.vlm"},
	{0x5ACE6D587AB6A38FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_noba_0t.dds"},
	{0x5ACEBD0F7364B9D0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb0col.mao"},
	{0x5ACEC10F7364C0E3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb0col.mmh"},
	{0x5ACECF0F7364D83DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb0col.msh"},
	{0x5ACED6B45E795B86ULL, "6070482_f.fxe"},
	{0x5ACEDBDCE3C63D4DULL, "6076945_m.fxe"},
	{0x5ACEEBFD4956649FULL, "910648281.ogg"},
	{0x5ACF124B9B8A1C1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_self.evt"},
	{0x5ACF2D577359B07DULL, "308956338.ogg"},
	{0x5ACF35AD162BCA10ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_5_c.impa.gad"},
	{0x5ACFFA0B08A49C2EULL, "vxw_blightlands_vk_0_0.vlm"},
	{0x5AD0217E73F985EEULL, "vxw_encounter1_il_0_0.vlm"},
	{0x5AD0C8DD9BA46FD4ULL, "6016678_m.fxe"},
	{0x5AD0D35F0BF7AA36ULL, "6044622_m.fxe"},
	{0x5AD0EB3A472BAA50ULL, "zz_followers_banter6158607.cl"},
	{0x5AD0ED26FBE75209ULL, "6141085_m.fxe"},
	{0x5AD0F74237BF5491ULL, "vxlt_0_lowtown_n_glh_0_0.vlm"},
	{0x5AD1599700D4C8EFULL, "749782572.ogg"},
	{0x5AD1A2DEA59E8B94ULL, "vxlt_backalley_g_1le_0_0.vlm"},
	{0x5AD1DC110833963DULL, "6072350_m.fxe"},
	{0x5AD24A019F2FA60DULL, "dae000cr_grand_cleric6148113.cl"},
	{0x5AD27F47DCA93922ULL, "dae021_keep_servant_male.utc"},
	{0x5AD28ABA47704856ULL, "6072601_m.fxe"},
	{0x5AD3D208AB9A83E8ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v1_c.anb"},
	{0x5AD400B7634CB19AULL, "14490239.ogg"},
	{0x5AD416C02272C65FULL, "qun221_petrice6029651.cl"},
	{0x5AD4EDF97F53F88EULL, "vxw_encounter3_i_0_0.vlm"},
	{0x5AD4F9FC8C41B4CBULL, "230795766.ogg"},
	{0x5AD508A7F15A41F5ULL, "6161142_m.fxe"},
	{0x5AD5DEA12099A848ULL, "rdr110dg_party_comments6094677.cl"},
	{0x5AD650408968E12CULL, "6127000_m.fxe"},
	{0x5AD687336D69FB7DULL, "7285858.ogg"},
	{0x5AD687D7C536A19EULL, "vxht_keep_n_61m_0_0.vlm"},
	{0x5AD689BF11159A69ULL, "33940103.ogg"},
	{0x5AD6C3DF4098C173ULL, "194038226.ogg"},
	{0x5AD7E892CAC58E30ULL, "858292437.ogg"},
	{0x5AD81FDC71D98750ULL, "zz_lgt_1_debug6114622.cl"},
	{0x5AD8275BEAA08D36ULL, "lt_0_lowtown_docks_a3_152_0.phy"},
	{0x5AD84B576194535AULL, "vxlt_undercitydungeon_11u_0_0.vlm"},
	{0x5AD89446EBFE61FBULL, "vxdr_0_deeproads_ef6_0_0.vlm"},
	{0x5AD89CADFFEF01CAULL, "dre000_varric6035144.cl"},
	{0x5AD8EF630839C9A4ULL, "hlac_lt_0_lowtown_a2_n_65538_3z.dds"},
	{0x5AD927371EFA4CFEULL, "follower_barks6153238.cl"},
	{0x5AD96378D9F9C573ULL, "vxdr_primevaltomb_7si_0_0.vlm"},
	{0x5AD9A5215D280BB0ULL, "gen_im_arm_glv_med_act1_01.uti"},
	{0x5AD9B17C897C33C4ULL, "vxlt_shop_g9_0_0.vlm"},
	{0x5ADA55B0C6548E2FULL, "403474757.ogg"},
	{0x5ADA7E0174BF70D4ULL, "6085047_m.fxe"},
	{0x5ADAA385EE4CEAABULL, "6089538_m.fxe"},
	{0x5ADAF96D9103C0A3ULL, "hlac_lt_0_lowtown_a2_n_c_-4_0x.dds"},
	{0x5ADB119609102398ULL, "vxca_caverns1_gf0_0_0.vlm"},
	{0x5ADBBD14CC37E3D2ULL, "311456578.ogg"},
	{0x5ADBC9B0857AC604ULL, "llac_ca_caverns2_5802_18.dds"},
	{0x5ADBF387FC823666ULL, "vxw_encounter3_3h_0_0.vlm"},
	{0x5ADC0CBC9670302EULL, "6108730_f.fxe"},
	{0x5ADC1A4B9B94B435ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_self.ani"},
	{0x5ADC33EE6FF3F948ULL, "follower_banter6160804.cl"},
	{0x5ADC6730179B293DULL, "vxht_keep_5mz_0_0.vlm"},
	{0x5ADCA2FDB567CACBULL, "6161342_m.fxe"},
	{0x5ADCBF01A63426EFULL, "zz_refugeehawke.utc"},
	{0x5ADD1844903A79AAULL, "isa111_isabela6066772.cl"},
	{0x5ADD328BB5F20CC1ULL, "vxlt_gallowstemplar_f_5r7_0_0.vlm"},
	{0x5ADD64FFFCC90AEBULL, "zz_followers_banter6157728.cl"},
	{0x5ADDA8801F79CDF5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic_5.mao"},
	{0x5ADE21967798EA31ULL, "717966337.ogg"},
	{0x5ADE821B1E802956ULL, "vxlt_0_lowtown_docks_a3_n_cx5_0_0.vlm"},
	{0x5ADEAB84394C75EBULL, "6149876_m.fxe"},
	{0x5ADEEC3A96E458AFULL, "dae000_epic_merchant_26179444.cl"},
	{0x5ADFAE7A33B3048FULL, "vxht_0_hightown_n_c59_0_0.vlm"},
	{0x5ADFF30B275ADA87ULL, "6025987_m.fxe"},
	{0x5AE0688F5086771DULL, "vxw_sundermount_32u_0_0.vlm"},
	{0x5AE0B4B573E554A0ULL, "6172918_m.fxe"},
	{0x5AE0E927C36B98C7ULL, "hlac_w_woundedcoast_131072_0z.dds"},
	{0x5AE16535D7B8FBC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\knockback.gad"},
	{0x5AE1B675B5916F18ULL, "302866131.ogg"},
	{0x5AE1C0D30502EB33ULL, "w_woundedcoast_72_0.rml"},
	{0x5AE1D758F88DC347ULL, "plt_dae000pt_location.plo"},
	{0x5AE2105BEAA8BAD7ULL, "lt_0_lowtown_docks_a3_152_0.msh"},
	{0x5AE2225BEAA8D93AULL, "lt_0_lowtown_docks_a3_152_0.mao"},
	{0x5AE225C94735FA3DULL, "763288337.ogg"},
	{0x5AE2265BEAA8E049ULL, "lt_0_lowtown_docks_a3_152_0.mmh"},
	{0x5AE25B147D474178ULL, "vxlt_darktown_48f_0_0.vlm"},
	{0x5AE26BD87E134D7DULL, "lgt000_chanter6147202.cl"},
	{0x5AE29845EDF4799CULL, "795552258.ogg"},
	{0x5AE2DD0626B000C6ULL, "532842132.ogg"},
	{0x5AE35A64E8C17654ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_2z.dds"},
	{0x5AE36E2F6764BA99ULL, "llac_lt_backalley_g_0_3.dds"},
	{0x5AE3EEB8D9368D4FULL, "852492258.ogg"},
	{0x5AE3F2B3EA3897F2ULL, "mag131_keran6086170.cl"},
	{0x5AE5548B1A420501ULL, "vxca_caverns2_8pk_0_0.vlm"},
	{0x5AE61CEBAF8F754AULL, "vxht_0_hightown_c3m_0_0.vlm"},
	{0x5AE626C7F1BDE024ULL, "w_sundermount_219_0.phy"},
	{0x5AE6951A33C5CFECULL, "113079180.ogg"},
	{0x5AE6B5250D72E7BAULL, "6081373_m.fxe"},
	{0x5AE6D6D72CB096EFULL, "art\\effects\\cloudshadows\\w_bonepit\\cloud2.dds"},
	{0x5AE751A0E72B785EULL, "vxlt_undercityencounter_1sj_0_0.vlm"},
	{0x5AE7B192041751F9ULL, "hlac_ca_caverns1_24425_12z.dds"},
	{0x5AE8397BE68AA03FULL, "6072693_m.fxe"},
	{0x5AE84BB0AFC9AA4DULL, "805023708.ogg"},
	{0x5AE85A35D7BF07D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\knockback.evt"},
	{0x5AE86757422D0390ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeceltic_0.pwk"},
	{0x5AE8736189D5AFDCULL, "vxlt_gallowstemplar_5j1_0_0.vlm"},
	{0x5AE892D2AF761C2DULL, "vxht_chantry_32i_0_0.vlm"},
	{0x5AE8D69158CA5D94ULL, "genip_spittoon.utp"},
	{0x5AE90E078883E017ULL, "plt_cod_crt_ogre.plo"},
	{0x5AE90E8C64BD17ECULL, "809133105.ogg"},
	{0x5AE94B77F57CE6E9ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sarcophagus_d.dds"},
	{0x5AE98927D90F8F70ULL, "fen210_fenris6109424.cl"},
	{0x5AEA0554D4BAC2C2ULL, "one280_merin6116750.cl"},
	{0x5AEA5A2EBA70B99AULL, "t3_glk_b02.tnt"},
	{0x5AEC27F2403340B6ULL, "487534416.ogg"},
	{0x5AEC34CA37DCEBF0ULL, "gen00fl_sebastian6179589.cl"},
	{0x5AEC5C6905FAD10BULL, "llac_lt_warehouse_n_1_0.dds"},
	{0x5AECB24552B07109ULL, "600722903.ogg"},
	{0x5AECBF95765B9950ULL, "hlac_w_bonepitmining_a2_2_0z.dds"},
	{0x5AECE1DE1F213B89ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_rf.ani"},
	{0x5AECE9D414D3D4ABULL, "vxw_bonepitmining_a2_168_0_0.vlm"},
	{0x5AED4C8C1FFEB312ULL, "dae000_pc_home_ambients6208187.cl"},
	{0x5AED88B448E780DDULL, "253697193.ogg"},
	{0x5AEDB5910703B6F3ULL, "zz_followers_banter6158835.cl"},
	{0x5AEDCEF5871B7E4AULL, "zz_dae_debug6029891.cl"},
	{0x5AEDDD4D1883BC3DULL, "vxht_0_hightown_c54_0_0.vlm"},
	{0x5AEDE208238028F0ULL, "l_fen121ar_mansion_nt.wbd"},
	{0x5AEE8698898ABC94ULL, "vxlt_backalley_g_1vk_0_0.vlm"},
	{0x5AEE9499800C19CEULL, "gen_im_trs_prf_sta_ele_r2.uti"},
	{0x5AEEF582593CB53BULL, "one000_ilen6037381.cl"},
	{0x5AEF3CACD598C1C8ULL, "syncdeathpositional.gda"},
	{0x5AEFBB7AB49E4307ULL, "vxht_brothel_n_5c2_0_0.vlm"},
	{0x5AF066A612FBD9A6ULL, "dae000ip_crowd_mf2_drunk2.utp"},
	{0x5AF0B602D78FBE68ULL, "gen00fl_fenris6110864.cl"},
	{0x5AF0ED55FB4D9972ULL, "751979901.ogg"},
	{0x5AF1072622FAF37EULL, "sound_movement_foley.gda"},
	{0x5AF121227FBEC897ULL, "6167854_m.fxe"},
	{0x5AF16B5D990304B0ULL, "692426304.ogg"},
	{0x5AF1B8DFDC085319ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_emo_0n.dds"},
	{0x5AF1DD376C954A49ULL, "6130082_f.fxe"},
	{0x5AF260CBDA7DE435ULL, "mer221_sabin6080565.cl"},
	{0x5AF2BC8BF9A87592ULL, "vxw_sundermount_34x_0_0.vlm"},
	{0x5AF3290A7790272BULL, "hlac_ht_keep_n_4519_17z.dds"},
	{0x5AF3FCDE1F27880BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_rf.gad"},
	{0x5AF3FEB025DC2A0FULL, "art\\vfx\\combat\\vfx_w_tmp_clean_lar_aoe_p.mmh"},
	{0x5AF48EC8A1858DCAULL, "vxht_0_hightown_c6n_0_0.vlm"},
	{0x5AF4C30B529D7472ULL, "mrl351cr_pridedemon.utc"},
	{0x5AF5449E5F1A7523ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc01a.msh"},
	{0x5AF55235D7C9840DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\knockback.ani"},
	{0x5AF5529E5F1A8D7DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc01a.mmh"},
	{0x5AF59346E15C4311ULL, "hlac_lt_0_lowtown_a3_n_c_65535_0x.dds"},
	{0x5AF5A40720AD752CULL, "636243987.ogg"},
	{0x5AF5BB479A08B889ULL, "dae000_tomwise6167309.cl"},
	{0x5AF66C92CB358629ULL, "vxht_keep_6hf_0_0.vlm"},
	{0x5AF670E8D139D9F6ULL, "448164586.ogg"},
	{0x5AF6C5A06EBFE01EULL, "58572153.ogg"},
	{0x5AF7A11AD61B604CULL, "6086370_m.fxe"},
	{0x5AF804797ACD3B4CULL, "6141893_m.fxe"},
	{0x5AF804DBBCB8EE45ULL, "6041077_m.fxe"},
	{0x5AF8A4CB50413399ULL, "vxht_keep_57p_0_0.vlm"},
	{0x5AF8A8D137D61041ULL, "hlac_w_woundedcoast_65540_0z.dds"},
	{0x5AF8C4AA3607B503ULL, "hlac_lt_0_lowtown_a2_n_c_131071_7y.dds"},
	{0x5AF9ACB0431C7008ULL, "vxlt_0_lowtown_gon_0_1.vlm"},
	{0x5AFA3965A37B9F95ULL, "gen00fl_player6207813.cl"},
	{0x5AFA4B74E75AD800ULL, "lgt230ip_at_tomecave.utp"},
	{0x5AFAF9DE1F2DA270ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_rf.evt"},
	{0x5AFB11F071D6D301ULL, "hm_one_mo_nabil.mor"},
	{0x5AFB1CAEB10B3A6FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robd_0s.dds"},
	{0x5AFB335607AB3ECCULL, "lt_0_lowtown_docks_a3_n_181_0.msh"},
	{0x5AFB395607AB48BEULL, "lt_0_lowtown_docks_a3_n_181_0.mmh"},
	{0x5AFB455607AB5D55ULL, "lt_0_lowtown_docks_a3_n_181_0.mao"},
	{0x5AFB73D69370E5D6ULL, "6061803_m.fxe"},
	{0x5AFBAE000817C529ULL, "vxht_0_hightown_n_bgl_0_0.vlm"},
	{0x5AFC29547D8F378BULL, "hlac_lt_0_lowtown_docks_-196610_2y.dds"},
	{0x5AFC465728FF4007ULL, "gen00fl_fenris6126569.cl"},
	{0x5AFC6B943F599881ULL, "vxlt_gallowstemplar_5aa_0_0.vlm"},
	{0x5AFCD108ABBD8683ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v1_c.mmh"},
	{0x5AFD0D28B4EB113FULL, "zz_approval_debug6003646.cl"},
	{0x5AFD2B21F5DBA4E4ULL, "gen00fl_aveline6113756.cl"},
	{0x5AFD78DB487A4F63ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_floordirt_n.dds"},
	{0x5AFDA25D3D9CB7CEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masd_0t.dds"},
	{0x5AFE25DE1CCAE322ULL, "784617018.ogg"},
	{0x5AFE562BCAC61F76ULL, "llac_lt_gallowstemplar_f_2388_0.dds"},
	{0x5AFEA56E14167A4EULL, "vxht_0_hightown_n_c5t_0_0.vlm"},
	{0x5AFF093D07AE37FBULL, "llac_w_blightlands_196604_0.dds"},
	{0x5AFFFB8ACCD64291ULL, "6163359_m.fxe"},
	{0x5B0009D34A57B525ULL, "vxca_caverns2_8qt_0_0.vlm"},
	{0x5B00C08F342E4647ULL, "gen_im_trs_gem_tigerseye_02.uti"},
	{0x5B012E68EB9E8E12ULL, "vxw_sundermount_33y_0_0.vlm"},
	{0x5B0184E3876DAAF8ULL, "1775044.ogg"},
	{0x5B01D68B5A0BEB20ULL, "vxlt_0_lowtown_a2_ex8_0_0.vlm"},
	{0x5B02043266650E47ULL, "vxht_mansion2_1at_0_0.vlm"},
	{0x5B022EEF51D26981ULL, "vxht_0_hightown_bum_0_0.vlm"},
	{0x5B0284FC39D1B328ULL, "zz_fol_debug6066147.cl"},
	{0x5B028EFAA99DAAAFULL, "llac_ht_keep_n_3733_1.dds"},
	{0x5B029FC2E7DFD345ULL, "llac_lt_gallowstemplar_n_2394_8.dds"},
	{0x5B0350828C773C6BULL, "hightown_fem_cityguard_mob6168526.cl"},
	{0x5B0370448019055EULL, "391792834.ogg"},
	{0x5B03BBBA9D4453EFULL, "vxw_sundermount_2ci_0_0.vlm"},
	{0x5B041F57B6B55E9BULL, "6037125_f.fxe"},
	{0x5B051CD12513485AULL, "6123388_quentin_purpose.cut"},
	{0x5B0526F0A1C3CD89ULL, "219970957.ogg"},
	{0x5B06C1F10041F736ULL, "vxht_keep_n_602_0_0.vlm"},
	{0x5B075105E0B83F5AULL, "art\\vfx\\combat\\vfx_q_ari_chop_act_c.anb"},
	{0x5B07C43FD8E8C263ULL, "lt_0_lowtown_docks_a3_139_0.mao"},
	{0x5B07C83FD8E8C8B0ULL, "lt_0_lowtown_docks_a3_139_0.mmh"},
	{0x5B07D63FD8E8E0CAULL, "lt_0_lowtown_docks_a3_139_0.msh"},
	{0x5B07D9DE8F1BF720ULL, "vxlt_house_hl_0_0.vlm"},
	{0x5B085ECB5E9CD284ULL, "zz_and_debug6178565.cl"},
	{0x5B086471765DC1DDULL, "clf221cr_jalen.utc"},
	{0x5B0864C112CA3599ULL, "vxht_0_hightown_a3_an2_0_0.vlm"},
	{0x5B0888228BD45641ULL, "vxlt_gallowstemplar_f_5c4_0_0.vlm"},
	{0x5B0905C5374F5EDDULL, "6043901_m.fxe"},
	{0x5B091EA99BD7C54FULL, "plt_lgt150pt_nat_werewolf.nss"},
	{0x5B0958818E9CEE47ULL, "t3_cth_cm6.tnt"},
	{0x5B0A5A4B201A7C4CULL, "mrl150_merrill6085835.cl"},
	{0x5B0A979DF9616023ULL, "vxlt_gallowstemplar_f_51x_0_0.vlm"},
	{0x5B0B6A526D76ED94ULL, "vxlt_backalley_g_1xi_0_0.vlm"},
	{0x5B0B93F1A9C6ABF4ULL, "6051378_m.fxe"},
	{0x5B0C5D01E0EDE613ULL, "106582607.ogg"},
	{0x5B0CD3214ED25993ULL, "6146828_m.fxe"},
	{0x5B0D174B060C92CEULL, "vxht_0_hightown_n_c2u_0_0.vlm"},
	{0x5B0D2762FD7C1018ULL, "vxlt_0_lowtown_docks_a3_d07_0_0.vlm"},
	{0x5B0D8858460FC628ULL, "417097655.ogg"},
	{0x5B0E854F7753D0A6ULL, "da2terrain_terrain_tools_p0.psh.fxo"},
	{0x5B0EBAF073551FE9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction2.ani"},
	{0x5B0F09C691496F8BULL, "vxlt_hangedman_2mu_0_0.vlm"},
	{0x5B0F644DC8C7FB93ULL, "115862541.ogg"},
	{0x5B0F9D110264BA98ULL, "6095297_m.fxe"},
	{0x5B1024CBB17152E3ULL, "103015007.ogg"},
	{0x5B1029D4504F1958ULL, "character_character_p47.psh.fxo.dx11"},
	{0x5B1043C56B628858ULL, "dae000_mother6136956.cl"},
	{0x5B10760F585B26C7ULL, "vxht_mansion2_1dw_0_0.vlm"},
	{0x5B10B0530659CE5DULL, "522591669.ogg"},
	{0x5B1105DCBC89D621ULL, "vxlt_undercityserialkiller_1j3_0_1.vlm"},
	{0x5B115D79143671ECULL, "gen00fl_aveline6132766.cl"},
	{0x5B11BCF7E8B2466CULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_imp_p.impa.ani"},
	{0x5B120058C3CEB0A3ULL, "abi_r_escape.wbd"},
	{0x5B122D978F3A6C6FULL, "vxw_sundermount_2dh_0_0.vlm"},
	{0x5B12BBA64368F5BFULL, "creatureranks.gda"},
	{0x5B137849403D2C03ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_016a_n.dds"},
	{0x5B14433E55927F74ULL, "227190328.ogg"},
	{0x5B1478030A53B0DBULL, "6140295_m.fxe"},
	{0x5B14B4E26FBE6A7BULL, "436964672.ogg"},
	{0x5B15BC3831767022ULL, "hlac_lt_darktown_-1_79z.dds"},
	{0x5B15D5F0735B6C6BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction2.gad"},
	{0x5B15DB39039A6E52ULL, "art\\vfx\\combat\\vfx_r_arc_critical_shot_c.impa.gad"},
	{0x5B167A93ABB7B7F1ULL, "79754429.ogg"},
	{0x5B169FD1E6A757F5ULL, "6152148_m.fxe"},
	{0x5B16A8CB00952E5CULL, "vxca_caverns2_55d_0_0.vlm"},
	{0x5B16AE2554507E5AULL, "gen00fl_aveline6137680.cl"},
	{0x5B16E71FF7804EA3ULL, "vxlt_gallowstemplar_n_5po_0_0.vlm"},
	{0x5B1790E927F6EEE4ULL, "clf260cr_guard1.utc"},
	{0x5B17C9F7E8B6C7AEULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_imp_p.impa.gad"},
	{0x5B185335A5ABF302ULL, "6071895_m.fxe"},
	{0x5B1883ACE5C6F4DEULL, "vxca_cave_2cu_0_1.vlm"},
	{0x5B189953803D3719ULL, "llac_ca_caverns2_5802_20.dds"},
	{0x5B18C86843499164ULL, "6083526_m.fxe"},
	{0x5B18E65BAF591517ULL, "283466925.ogg"},
	{0x5B18F0F5E9441662ULL, "gen00fl_carver6102733.cl"},
	{0x5B18FF561ADC95FFULL, "6155943_m.fxe"},
	{0x5B193F8EA632EE0AULL, "vxw_encounter1_ap_0_0.vlm"},
	{0x5B194F4E2345C981ULL, "vxlt_gallowstemplar_5wg_0_0.vlm"},
	{0x5B1966211F1E239AULL, "vxca_caverns1_hm4_0_0.vlm"},
	{0x5B1987EC80F338B1ULL, "6064968_m.fxe"},
	{0x5B19BBE8A51D8974ULL, "art\\levels\\atmosphere\\lights\\_textures\\gobotest_d.dds"},
	{0x5B19D4552330455CULL, "vxlt_0_lowtown_a2_hgl_0_0.vlm"},
	{0x5B1A0CBAD99CEB70ULL, "6007724_m.fxe"},
	{0x5B1A3FC66536A67BULL, "vxht_chantry_33p_0_0.vlm"},
	{0x5B1A51330D06060AULL, "710320005.ogg"},
	{0x5B1B18AAC19064A8ULL, "928682234.ogg"},
	{0x5B1B5E30DB662C64ULL, "qcr000_arishok6173708.cl"},
	{0x5B1B70B61F865F8FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.dura.gad"},
	{0x5B1BDE39039EDE89ULL, "art\\vfx\\combat\\vfx_r_arc_critical_shot_c.impa.evt"},
	{0x5B1BFA8CCFAF3279ULL, "lowtown_fem_prostitute_sta6152313.cl"},
	{0x5B1BFFB025FD7E4CULL, "art\\vfx\\combat\\vfx_w_tmp_clean_lar_aoe_p.anb"},
	{0x5B1C2D3A1401E726ULL, "491098500.ogg"},
	{0x5B1C58C81BF9CC59ULL, "258314259.ogg"},
	{0x5B1C9A461AEEE3CAULL, "w_bonepitdestroyed_a3_148_0.mao"},
	{0x5B1C9E461AEEEA19ULL, "w_bonepitdestroyed_a3_148_0.mmh"},
	{0x5B1CA8461AEEFB07ULL, "w_bonepitdestroyed_a3_148_0.msh"},
	{0x5B1CD2F0736185D0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction2.evt"},
	{0x5B1CDAECE73D776CULL, "gen00fl_bethany6097835.cl"},
	{0x5B1D3FB9851ED67AULL, "6125978_m.fxe"},
	{0x5B1DBB47DD236A4DULL, "vxlt_0_lowtown_a2_n_fhe_0_0.vlm"},
	{0x5B1DE7830D717AF1ULL, "249470801.ogg"},
	{0x5B1E168A9D2CDC7CULL, "vxht_mansion2_38f_0_0.vlm"},
	{0x5B1EC4F7E8BCDD8DULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_imp_p.impa.evt"},
	{0x5B1EE55B110E23ACULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocka.msh"},
	{0x5B1EEB5B110E2D9EULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocka.mmh"},
	{0x5B1EF75B110E4235ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverocka.mao"},
	{0x5B1F433A7094A16BULL, "vxlt_warehouse_n_10v_0_0.vlm"},
	{0x5B1F554AB952D72BULL, "6144314_m.fxe"},
	{0x5B1F5F6D1B543E77ULL, "6024580_m.fxe"},
	{0x5B1FDF9BAAECCBE9ULL, "509230549.ogg"},
	{0x5B209F74813084EFULL, "vxw_sundermount_2ik_0_0.vlm"},
	{0x5B2123870BD54F57ULL, "vxlt_0_lowtown_a2_n_c_e33_0_0.vlm"},
	{0x5B213EB0475B1F85ULL, "lt_gallowscourtyard_n_43_0.anb"},
	{0x5B2174E3ECD7DB33ULL, "479868229.ogg"},
	{0x5B2185B61F8AEF3CULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.dura.evt"},
	{0x5B21E38B95E7E925ULL, "6155769_m.fxe"},
	{0x5B21F590C3FE7AC5ULL, "dae000_pc_home_ambients6208638.cl"},
	{0x5B2225A30A1F8B56ULL, "6149990_m.fxe"},
	{0x5B2253119C328EAEULL, "hlac_lt_0_lowtown_docks_n_-2_6x.dds"},
	{0x5B225B3A42BBE9DDULL, "6121458_m.fxe"},
	{0x5B22A857E8368E96ULL, "vxdr_0_deeproads_cxc_0_0.vlm"},
	{0x5B22D9772F3E029FULL, "pro000_colonel_ewald6047390.cl"},
	{0x5B2351F215F33548ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\_3_-_default.mao"},
	{0x5B2431AC9E43D28AULL, "hlac_dr_primevalentrance_1725_4y.dds"},
	{0x5B2500630B4AEE3DULL, "6095105_f.fxe"},
	{0x5B2542B7EE57B671ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_p.impa.ani"},
	{0x5B25487AF6B66699ULL, "vxht_0_hightown_a3_ap0_0_0.vlm"},
	{0x5B256127916B3C65ULL, "311094456.ogg"},
	{0x5B25A6D39CAE0C1DULL, "69115993.ogg"},
	{0x5B25B7FB1363FEA7ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.cess.gad"},
	{0x5B26914BD7D9DA60ULL, "vxdr_0_deeproads_bwp_0_1.vlm"},
	{0x5B2794FD28402F90ULL, "hlac_ht_mansion2b_818_39x.dds"},
	{0x5B27B16B9829068AULL, "vxw_encounter1_fq_0_0.vlm"},
	{0x5B27C12B153BE201ULL, "vxlt_gallowstemplar_5pf_0_0.vlm"},
	{0x5B2829DC58F4DDEBULL, "lgt110_arl_bann_teagan6115633.cl"},
	{0x5B286FA7F132A9ABULL, "vxdr_0_deeproads_ewe_0_0.vlm"},
	{0x5B288C74EE315815ULL, "plt_cod_itm_ferryman.plo"},
	{0x5B289A600A98BFBEULL, "6093770_m.fxe"},
	{0x5B28A8933F644E7EULL, "vfx_vfx_p3.vsh.fxo.dx11.assembly"},
	{0x5B28A8D8ECD3D2BCULL, "603547753.ogg"},
	{0x5B28B1A3572CBEFBULL, "vxht_chantry_36s_0_0.vlm"},
	{0x5B297884AF72669AULL, "zz_and_debug6176459.cl"},
	{0x5B29A5D729DC25D3ULL, "arm_star.mao"},
	{0x5B29C63903AAF3F8ULL, "art\\vfx\\combat\\vfx_r_arc_critical_shot_c.impa.ani"},
	{0x5B29DFFC916FD255ULL, "135881252.ogg"},
	{0x5B2A557FF07CDED3ULL, "vxht_brothel_n_622_0_0.vlm"},
	{0x5B2AA0FDB6191701ULL, "vxca_caverns2_586_0_0.vlm"},
	{0x5B2AAEDAC2109B0BULL, "gen00fl_anders6128309.cl"},
	{0x5B2AC20B09310602ULL, "gen00fl_isabela6085676.cl"},
	{0x5B2B1CA801C6DD20ULL, "728468115.ogg"},
	{0x5B2B3B01A8F20B56ULL, "6180298_m.fxe"},
	{0x5B2B3DA92004B473ULL, "seb321_leliana6181752.cl"},
	{0x5B2B4932B9DD28E6ULL, "vxht_0_hightown_n_b8b_0_0.vlm"},
	{0x5B2B4A276CDB6959ULL, "hlac_ht_mansion2b_818_2y.dds"},
	{0x5B2B632CFAE95AADULL, "vxlt_0_lowtown_docks_a3_n_9fj_0_0.vlm"},
	{0x5B2BA6BED03A57E1ULL, "dae000ip_crowd_lm1_talk17.utp"},
	{0x5B2BB1967714EEFAULL, "6122878_m.fxe"},
	{0x5B2BB58793A5E0CEULL, "vxht_brothel_n_59i_0_0.vlm"},
	{0x5B2C21FAC3849C9AULL, "llac_dr_primevaltomb_658_3.dds"},
	{0x5B2C2BB1BEC6A5E9ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_pc_ogremage.gad"},
	{0x5B2C4DB7EE5DE713ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_p.impa.gad"},
	{0x5B2C9440F9DBE2E5ULL, "vxw_sundermount_2zq_0_1.vlm"},
	{0x5B2CA732BD93BC10ULL, "6071512_m.fxe"},
	{0x5B2CACFB136A0AB4ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.cess.evt"},
	{0x5B2D5F11C4FA3F09ULL, "906939012.ogg"},
	{0x5B2D5F3D2A1B7492ULL, "llac_w_sundermount_131068_0.dds"},
	{0x5B2D68B4C041F3DFULL, "art\\characters\\heraldry\\hld_grey_wardens.dds"},
	{0x5B2DA8C7DAE50E93ULL, "s_stand_02.dds"},
	{0x5B2DE996A0760721ULL, "vxlt_undercityserialkiller_1d1_0_1.vlm"},
	{0x5B2E339186D5B2F8ULL, "6034465_m.fxe"},
	{0x5B2ED1F01C858EFFULL, "vxw_woundedcoast_o3_0_0.vlm"},
	{0x5B2EFFA8A9C3D8D5ULL, "63218742.ogg"},
	{0x5B2F115173269D6FULL, "vxw_sundermount_2rj_0_0.vlm"},
	{0x5B2F459737A093CEULL, "6162279_f.fxe"},
	{0x5B2F5405E0D9960DULL, "art\\vfx\\combat\\vfx_q_ari_chop_act_c.mmh"},
	{0x5B2F8DB61F973A55ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.dura.ani"},
	{0x5B3030FCFEF39A5AULL, "vxca_cavern1_haunted_hum_0_0.vlm"},
	{0x5B30BD0881F186E7ULL, "6066865_m.fxe"},
	{0x5B311A34DA2CA716ULL, "vxdr_0_deeproads_cud_0_0.vlm"},
	{0x5B3163B59F2345CEULL, "vxw_woundedcoast_n1_0_0.vlm"},
	{0x5B3186E88680688BULL, "follower_banter6172855.cl"},
	{0x5B31F45099354E86ULL, "vxlt_0_lowtown_docks_n_b5u_0_1.vlm"},
	{0x5B3221B23FF3BB1FULL, "mrl150_merrill6071947.cl"},
	{0x5B32776EBF425711ULL, "219207240.ogg"},
	{0x5B32B9D89E305C6AULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.anb"},
	{0x5B32E3A593DB24D2ULL, "hlac_lt_gallowscourtyard_n_-1_0y.dds"},
	{0x5B32FCB1BECC7437ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_pc_ogremage.ani"},
	{0x5B332B04BCA189F5ULL, "hlac_ca_caverns2_5802_67y.dds"},
	{0x5B332DB2D553C758ULL, "vxht_mansion2_2cn_0_0.vlm"},
	{0x5B334AB7EE640178ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_p.impa.evt"},
	{0x5B335112BD89BF71ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_act_c.impa.ani"},
	{0x5B336F5B58177FC1ULL, "337391537.ogg"},
	{0x5B33861155748762ULL, "vxlt_hangedman_26e_0_0.vlm"},
	{0x5B33D635B1D17D52ULL, "vxht_brothel_n_627_0_0.vlm"},
	{0x5B33F998C532D636ULL, "vxht_mansion_n_6nh_0_0.vlm"},
	{0x5B34588DC7D87D78ULL, "vxlt_warehouse_n_bh_0_0.vlm"},
	{0x5B355FCE1C008AE2ULL, "gen00fl_dog6166811.cl"},
	{0x5B35B80823BD96B2ULL, "l_fen121ar_mansion_nt.bnk"},
	{0x5B35D81BA4B9E970ULL, "560147081.ogg"},
	{0x5B360FCFD7D51522ULL, "character_character_p56.psh.fxo"},
	{0x5B36189B67BB0C78ULL, "follower_banter6154036.cl"},
	{0x5B3623488A1F1F0AULL, "vxw_encounter1_kr_0_0.vlm"},
	{0x5B362380085ECA63ULL, "mag370_meredith6046957.cl"},
	{0x5B363777AD4A4070ULL, "vxlt_0_lowtown_a2_n_c_hku_0_0.vlm"},
	{0x5B3689AC5D290918ULL, "444997612.ogg"},
	{0x5B36B80F071C765CULL, "vxlt_0_lowtown_a2_hib_0_0.vlm"},
	{0x5B36BFEF97E85474ULL, "6134762_f.fxe"},
	{0x5B36DD6548AFCD66ULL, "vxdr_primevalentrance_34a_0_0.vlm"},
	{0x5B36E184E328C22BULL, "vxdr_0_deeproads_epd_0_0.vlm"},
	{0x5B37F258C3EF6851ULL, "abi_r_escape.bnk"},
	{0x5B37FDEB3F0E2374ULL, "follower_banter6156012.cl"},
	{0x5B38218423E5DAC1ULL, "782779386.ogg"},
	{0x5B383A3D663C1406ULL, "133701528.ogg"},
	{0x5B3854AAFFCF624EULL, "art\\vfx\\combat\\vfx_gib_qunc_1.mao"},
	{0x5B38555C222806D4ULL, "434237610.ogg"},
	{0x5B386A4485211CD8ULL, "mag170_hugh6043096.cl"},
	{0x5B38C3EC11B15162ULL, "738274056.ogg"},
	{0x5B38D84D90F2252DULL, "6148194_m.fxe"},
	{0x5B39129255937BDFULL, "hlac_ht_0_hightown_a2_131070_0x.dds"},
	{0x5B3929C01AAD9C8FULL, "742696797.ogg"},
	{0x5B397572E0DF2A4EULL, "vxht_0_hightown_a2_c7a_0_0.vlm"},
	{0x5B399EF6E4BDB019ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastwood_d.dds"},
	{0x5B39FAE8DB524B02ULL, "vxht_0_hightown_a2_cfb_0_0.vlm"},
	{0x5B3A357C23531B56ULL, "57044970.ogg"},
	{0x5B3A5C12BD8FF013ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_act_c.impa.gad"},
	{0x5B3AA4FB13763AEDULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.cess.ani"},
	{0x5B3B171E5FCBF376ULL, "vxdr_0_deeproads_9nc_0_0.vlm"},
	{0x5B3B303EAD48113AULL, "6057104_f.fxe"},
	{0x5B3B59CEEAB56ABBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangan.msh"},
	{0x5B3B6BCEEAB58906ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangan.mao"},
	{0x5B3B77CEEAB59DA5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangan.mmh"},
	{0x5B3BBDD41F3A40F2ULL, "vxca_cavern1_haunted_ckh_0_0.vlm"},
	{0x5B3BDF8441CA7AE9ULL, "6126701_m.fxe"},
	{0x5B3D0C2D1AC6E8A2ULL, "6142094_m.fxe"},
	{0x5B3D13BF5A365612ULL, "zz_mag_debug6025380.cl"},
	{0x5B3D43CD0E7BA77FULL, "vxw_woundedcoast_d2_0_0.vlm"},
	{0x5B3D73A875D834ADULL, "lgt260cr_way_apostate_1.utc"},
	{0x5B3D83C458B1792AULL, "vxw_woundedcoast_15e_0_0.vlm"},
	{0x5B3E3B92C51CE190ULL, "hlac_lt_0_lowtown_a2_65534_6x.dds"},
	{0x5B3E6B1D5144D18EULL, "dae000_madam_lusine6162379.cl"},
	{0x5B3ECD6201C52283ULL, "6093866_f.fxe"},
	{0x5B3EFC48D4C5BC01ULL, "dae000_mother6099238.cl"},
	{0x5B3F09DB86377700ULL, "lgt120cr_traitor.utc"},
	{0x5B3F0B920A706015ULL, "85719548.ogg"},
	{0x5B3F3DCF2C934E67ULL, "659126619.ogg"},
	{0x5B3F719A0E767E66ULL, "vxlt_0_lowtown_a2_eok_0_0.vlm"},
	{0x5B3F8C11CC22BF96ULL, "vxdr_0_deeproads_cre_0_0.vlm"},
	{0x5B3F9B2328599BCEULL, "art\\vfx\\cinematic\\vfx_blood_directional_l.mmh"},
	{0x5B3FE7D3141F71EBULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.impa.ani"},
	{0x5B40A4A84C2C04EEULL, "246455000.ogg"},
	{0x5B415067814BBD56ULL, "gen00fl_aveline6121966.cl"},
	{0x5B415912BD960A78ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_act_c.impa.evt"},
	{0x5B41DB2338E2B14EULL, "6037394_m.fxe"},
	{0x5B420AC6E581ED88ULL, "one300_zevran6093495.cl"},
	{0x5B42223AB91D265AULL, "vxlt_0_lowtown_a3_n_c_hwz_0_0.vlm"},
	{0x5B423CB6CD629823ULL, "vxlt_gallowstemplar_n_5ap_0_0.vlm"},
	{0x5B423D9FDEC03184ULL, "vxht_mansion_n_6p4_0_0.vlm"},
	{0x5B425FB10190F558ULL, "1075489.ogg"},
	{0x5B42931036DA3393ULL, "llac_dr_primevaltomb_659_7.dds"},
	{0x5B42CCBFA626BD97ULL, "161093445.ogg"},
	{0x5B434B2EF64135C2ULL, "mag371cr_elsa.utc"},
	{0x5B438964B9ED15F7ULL, "18759368.ogg"},
	{0x5B44C98BB832135BULL, "zz_followers_banter6159498.cl"},
	{0x5B45D3BCAC890A23ULL, "hlac_w_bonepitmining_a2_262149_0y.dds"},
	{0x5B4635279ABD6584ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicechair_sa.dds"},
	{0x5B46B2D892E0D7A2ULL, "sky_w_sundermount.mao"},
	{0x5B46D84CF7B012EAULL, "25286511.ogg"},
	{0x5B471515CCC20503ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_pause.ani"},
	{0x5B4771C13715AC30ULL, "vxdr_primevaltomb_3ip_0_0.vlm"},
	{0x5B47C6E692BCD798ULL, "gen00fl_anders6119230.cl"},
	{0x5B47CF3FD91EB8CBULL, "lt_0_lowtown_docks_a3_139_0.phy"},
	{0x5B4887759B233D46ULL, "544200452.ogg"},
	{0x5B48B8440CF599B6ULL, "954742902.ogg"},
	{0x5B49619054791D1AULL, "vxlt_0_lowtown_docks_bg9_0_0.vlm"},
	{0x5B4980A390E7B5C6ULL, "ss_warrior_ogredb.cut"},
	{0x5B4988FB51C20BF6ULL, "vxdr_0_deeproads_9cd_0_0.vlm"},
	{0x5B4A2FB111305972ULL, "vxca_cavern1_haunted_cli_0_0.vlm"},
	{0x5B4ABC0BF7E7ADD9ULL, "508784913.ogg"},
	{0x5B4B2DC5558FAF01ULL, "mag121_idunna6106083.cl"},
	{0x5B4BF50B5712CE6FULL, "vxw_sundermount_2xt_0_0.vlm"},
	{0x5B4BF5A14AA791AAULL, "vxw_woundedcoast_10f_0_0.vlm"},
	{0x5B4C34358698320FULL, "vxw_woundedcoast_zx_0_0.vlm"},
	{0x5B4CBA2328645ACDULL, "art\\vfx\\cinematic\\vfx_blood_directional_l.anb"},
	{0x5B4CCFD31429D362ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.impa.evt"},
	{0x5B4CD3C7CF9344CAULL, "42184805.ogg"},
	{0x5B4D063311B25D9CULL, "6085336_f.fxe"},
	{0x5B4D0E3901054275ULL, "hlac_dr_0_deeproads_17039_1x.dds"},
	{0x5B4D60F00EB29F59ULL, "6084500_m.fxe"},
	{0x5B4D9BFA4F0CBC1BULL, "hlac_lt_0_lowtown_docks_a3_n_65534_6x.dds"},
	{0x5B4DDA899D4FAFE1ULL, "vxht_mansion_n_4dc_0_0.vlm"},
	{0x5B4DEEFAA2D01484ULL, "hlac_lt_0_lowtown_docks_n_-196610_0y.dds"},
	{0x5B4E1195D84BC43CULL, "llac_w_sundermount_1_0.dds"},
	{0x5B4E958F4F91E5B7ULL, "vxw_bonepit_11k_0_0.vlm"},
	{0x5B4EADAF90CEA31AULL, "6107156_m.fxe"},
	{0x5B4F5BA03A027BBFULL, "6185128_m.fxe"},
	{0x5B4F8FF3E7076C38ULL, "1026705090.ogg"},
	{0x5B4FAD37745034EAULL, "vxlt_gallowscourtyard_63q_0_0.vlm"},
	{0x5B4FF5EB76C4542EULL, "art\\characters\\playercharacter\\humanmale\\boo_medp.phy"},
	{0x5B4FFAB66269EC80ULL, "759884715.ogg"},
	{0x5B5056D48F4AB954ULL, "framebuffereffects_twopassblurcochorizontl_p0.psh.fxo.dx11"},
	{0x5B5059029C0BB131ULL, "vxlt_0_lowtown_docks_b50_0_0.vlm"},
	{0x5B506D1FC4255202ULL, "dae000_friend_turn_in6179260.cl"},
	{0x5B5090640344149FULL, "art\\vfx\\combat\\vfx_r_arc_burst_sh_aoe_p.impa.gad"},
	{0x5B50DA5CDD6BDF82ULL, "ser200_aveline6201516.cl"},
	{0x5B50EF5722D59C90ULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm_facem1.mao"},
	{0x5B5115C39DD8E088ULL, "lla_ht_0_hightown_131070_0.dds"},
	{0x5B512D7B6092A689ULL, "and230_mage_girl6135804.cl"},
	{0x5B5171BDE9EBABA4ULL, "llac_w_bonepitdestroyed_a3_0_6.dds"},
	{0x5B523C767EE913B0ULL, "dae000_bartrand6033982.cl"},
	{0x5B526F8810C713BAULL, "lt_0_lowtown_37_0.rml"},
	{0x5B528D4B358F3CD2ULL, "309347530.ogg"},
	{0x5B52E6FD679BFF06ULL, "349077312.ogg"},
	{0x5B532B7DF6F056C0ULL, "6141523_m.fxe"},
	{0x5B53BC5A92E021B9ULL, "vxlt_gallowstemplar_3dt_0_0.vlm"},
	{0x5B53CED3142FF075ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.impa.gad"},
	{0x5B5418EB1F567E5CULL, "gen00fl_bethany6101022.cl"},
	{0x5B54A7A89CFFBFE5ULL, "rock_jagged_diffuse.dds"},
	{0x5B54FCD4BF22E060ULL, "686880978.ogg"},
	{0x5B5501368478F36AULL, "vxlt_0_lowtown_a3_n_c_fg0_0_0.vlm"},
	{0x5B5581083C6CE77CULL, "vxlt_warehouse_tw_0_0.vlm"},
	{0x5B5594B99A61016AULL, "6105752_m.fxe"},
	{0x5B55F3925141BC4BULL, "hlac_lt_backalley_n_65537_1y.dds"},
	{0x5B5662CA023FFAFFULL, "320963150.ogg"},
	{0x5B56B12ED01F8AC4ULL, "vxht_mansion2_2ah_0_0.vlm"},
	{0x5B56F2078DDB493DULL, "vxht_brothel_n_5af_0_0.vlm"},
	{0x5B56F61EF2FB21BBULL, "783730715.ogg"},
	{0x5B578564034A208CULL, "art\\vfx\\combat\\vfx_r_arc_burst_sh_aoe_p.impa.evt"},
	{0x5B57D827F6EA0CB1ULL, "948684491.ogg"},
	{0x5B57FAD843B82476ULL, "vxdr_0_deeproads_9de_0_0.vlm"},
	{0x5B5809B670A8A81CULL, "vxht_mansion2b_3ei_0_0.vlm"},
	{0x5B58714D4901C93EULL, "290474017.ogg"},
	{0x5B5913EF84A7C6FCULL, "one140cr_charade6137071.cl"},
	{0x5B597C1DA8EC2469ULL, "vxlt_0_lowtown_a3_n_c_eoi_0_1.vlm"},
	{0x5B598049E0DBD89BULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.impa.ani"},
	{0x5B59B63ACECC5782ULL, "hlac_lt_warehouse_n_-65535_1x.dds"},
	{0x5B59DCEFD41F11B4ULL, "hlac_lt_gallowscourtyard_n_0_1x.dds"},
	{0x5B59EAFA9BED2968ULL, "6013289_f.fxe"},
	{0x5B5A3FB0478AE68DULL, "lt_gallowscourtyard_n_43_0.rml"},
	{0x5B5A6AAB6E70E73BULL, "hlac_lt_gallowstemplar_2384_3y.dds"},
	{0x5B5AC12AA48FAC26ULL, "dae300_alistair_king6133796.cl"},
	{0x5B5B17FB0B42D72DULL, "50877554.ogg"},
	{0x5B5B2C15CCD2D5EDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_pause.gad"},
	{0x5B5BAF2ACB9615AAULL, "294797316.ogg"},
	{0x5B5BBCD89E53665DULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.mmh"},
	{0x5B5BDE6ACDEBE018ULL, "qun231cr_envoy.utc"},
	{0x5B5C1B2B9BEF58C4ULL, "w_bonepitdestroyed_a3_93_0.msh"},
	{0x5B5C2D2B9BEF772DULL, "w_bonepitdestroyed_a3_93_0.mao"},
	{0x5B5C312B9BEF7E66ULL, "w_bonepitdestroyed_a3_93_0.mmh"},
	{0x5B5CCEE8933800A9ULL, "85957896.ogg"},
	{0x5B5CF44ED83228C4ULL, "fires_d.dds"},
	{0x5B5D9CCF2DD51797ULL, "vxht_0_hightown_n_90d_0_0.vlm"},
	{0x5B5E05690AF13D4CULL, "lt_0_lowtown_docks_a3_106_0.msh"},
	{0x5B5E0B690AF1473EULL, "lt_0_lowtown_docks_a3_106_0.mmh"},
	{0x5B5E17690AF15BD5ULL, "lt_0_lowtown_docks_a3_106_0.mao"},
	{0x5B5E21DA7FEBE92CULL, "hlac_ht_0_hightown_a2_-1_1x.dds"},
	{0x5B5E4AB2B1140835ULL, "6187766_m.fxe"},
	{0x5B5E76377C61595EULL, "gui.mat.dx11"},
	{0x5B5EEE1AA9C8A0C4ULL, "gen_im_cns_flu_aveline_03.uti"},
	{0x5B5FAE249DBAC6F8ULL, "vxlt_warehouse_n_sg_0_0.vlm"},
	{0x5B5FB3831A1CE8B6ULL, "6049815_f.fxe"},
	{0x5B5FC8634F258FC3ULL, "character_character_p211.psh.fxo.dx11"},
	{0x5B60E0C0F698F000ULL, "6075677_m.fxe"},
	{0x5B60F57E2B4654E0ULL, "p_sh.mao"},
	{0x5B611526505E9510ULL, "vxw_encounter2_camp_6r_0_0.vlm"},
	{0x5B616A4FBEB67528ULL, "33005582.ogg"},
	{0x5B61964D0785A2D9ULL, "vxht_mansion2b_3f9_0_0.vlm"},
	{0x5B61A7A0074A8D9CULL, "6135195_m.fxe"},
	{0x5B61D23446E0014BULL, "6139750_m.fxe"},
	{0x5B623728748EA597ULL, "105325487.ogg"},
	{0x5B629D56A01F216EULL, "178958783.ogg"},
	{0x5B6359879A843401ULL, "649602763.ogg"},
	{0x5B63946771038C35ULL, "isa311_isabela6092310.cl"},
	{0x5B63D72C16CCFD4CULL, "vxw_encounter2_h8_0_0.vlm"},
	{0x5B63F2E52E62FFFCULL, "vxlt_warehouse_yt_0_0.vlm"},
	{0x5B640D943A443C93ULL, "121988569.ogg"},
	{0x5B6415420D442248ULL, "6051315_f.fxe"},
	{0x5B6427FFF92BCB20ULL, "hlac_lt_0_lowtown_a2_n_0_6z.dds"},
	{0x5B642B4613049E8EULL, "vxlt_hangedman_171_0_0.vlm"},
	{0x5B6433FB826D5114ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.mmh"},
	{0x5B6478C3DEC07259ULL, "743581377.ogg"},
	{0x5B65230BC215A344ULL, "vxht_mansion2_2ji_0_0.vlm"},
	{0x5B65318CC6CB7C76ULL, "vxlt_0_lowtown_n_fee_0_0.vlm"},
	{0x5B6561C2D64AD7C4ULL, "hlac_w_woundedcoast_131076_0y.dds"},
	{0x5B656D64035635E5ULL, "art\\vfx\\combat\\vfx_r_arc_burst_sh_aoe_p.impa.ani"},
	{0x5B6571C360E7C170ULL, "sib131ip_ogre_skull.utp"},
	{0x5B65BCBE0C5C7FB5ULL, "gen_im_npc_wep_arc_lbwa.uti"},
	{0x5B65D7F4EA879B10ULL, "hlac_w_encounter2_-65536_0z.dds"},
	{0x5B664787907839F9ULL, "561792085.ogg"},
	{0x5B668849E0E67092ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.impa.evt"},
	{0x5B67A5D83F58BFBCULL, "vxht_0_hightown_a3_api_0_0.vlm"},
	{0x5B67C5E1F743AF1DULL, "hlac_lt_0_lowtown_docks_-131075_6y.dds"},
	{0x5B6836793C29C612ULL, "follower_banter6143148.cl"},
	{0x5B686A0AA4AAFFFCULL, "t1_mue_gn2.tnt"},
	{0x5B68B1E753D99DC2ULL, "vxht_0_hightown_ai5_0_0.vlm"},
	{0x5B69133B42EAC751ULL, "vxht_playershouse_n_1vy_0_0.vlm"},
	{0x5B693228565CB69AULL, "vxlt_0_lowtown_docks_aps_0_0.vlm"},
	{0x5B697EFB8A071200ULL, "6085574_f.fxe"},
	{0x5B69BD573176C8E8ULL, "fen210_tevinter_ambusher26108645.cl"},
	{0x5B6A03E72F93E194ULL, "6090320_m.fxe"},
	{0x5B6A1A16A8333974ULL, "6026985_m.fxe"},
	{0x5B6A6879611980C0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sackclose_d.dds"},
	{0x5B6A68BE6C9B8BD0ULL, "hlac_ht_chantry_n_2534_49y.dds"},
	{0x5B6B2395BEC81B40ULL, "6137588_m.fxe"},
	{0x5B6B9893A26A49A6ULL, "gen00fl_anders6125338.cl"},
	{0x5B6BB929E1069532ULL, "hlac_lt_0_lowtown_a2_n_131070_0x.dds"},
	{0x5B6C885D8E1D22E8ULL, "783260567.ogg"},
	{0x5B6CB9E210A6AE90ULL, "dae000_merchant_rare_em6177945.cl"},
	{0x5B6CD811DE06E2ACULL, "cage_03.mao"},
	{0x5B6D126D2BF99FCEULL, "6207513_m.fxe"},
	{0x5B6D88C419F5FC07ULL, "gen00fl_aveline6122726.cl"},
	{0x5B6D9D4F806B437BULL, "869677970.ogg"},
	{0x5B6DA2523A518380ULL, "vxlt_undercityserialkiller_yy_0_0.vlm"},
	{0x5B6DA749E0ECC425ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.impa.gad"},
	{0x5B6DB1C84353CCE6ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\deadtospider.ani"},
	{0x5B6DC275D8FC767AULL, "6062129_m.fxe"},
	{0x5B6DC4B00661DE43ULL, "6076727_m.fxe"},
	{0x5B6E8C68A6A11A46ULL, "hlac_ht_bartrand_n_2995_5y.dds"},
	{0x5B6F6F0FC849FEA6ULL, "isa120_hayder6068792.cl"},
	{0x5B6F87034254AD90ULL, "vxw_encounter2_camp_1s_0_0.vlm"},
	{0x5B70349D8ABEAD58ULL, "hlac_ht_playershouse_n_855_12z.dds"},
	{0x5B70A8F8AB01242BULL, "vxdr_0_deeproads_elx_0_0.vlm"},
	{0x5B70CD7F358FA475ULL, "hlac_dr_slavers_3215_2x.dds"},
	{0x5B70E381B01E256EULL, "920492222.ogg"},
	{0x5B7130F1A84F6B34ULL, "hlac_lt_0_lowtown_65538_3y.dds"},
	{0x5B719708072428FEULL, "6126980_m.fxe"},
	{0x5B71B85644686AE6ULL, "art\\vfx\\textures\\vfx_flash.dds"},
	{0x5B721D8D2CFD7581ULL, "llac_ht_keep_n_3732_15.dds"},
	{0x5B7257BDADEF4FEAULL, "vxlt_0_lowtown_e2d_0_0.vlm"},
	{0x5B725DCA7D553F2AULL, "hlac_lt_backalley_n_0_9z.dds"},
	{0x5B737D3F50D56FDCULL, "vxht_0_hightown_a2_b9d_0_0.vlm"},
	{0x5B73856DBECB3780ULL, "6169187_m.fxe"},
	{0x5B74CBA11DDFC40BULL, "gen00fl_aveline6114415.cl"},
	{0x5B753C89D1360E69ULL, "774718429.ogg"},
	{0x5B7561EA35AB9A2AULL, "vxca_caverns1_fm6_0_0.vlm"},
	{0x5B75666C8DCC5E8BULL, "llac_w_bonepit_65541_0.dds"},
	{0x5B75BAFFA0AF2E90ULL, "819421454.ogg"},
	{0x5B7685724BBFB8C6ULL, "vxlt_foundry_n_w7_0_0.vlm"},
	{0x5B769487E8C1C7BEULL, "art\\characters\\creatures\\desiredemon\\animation\\hv_dmg.ani"},
	{0x5B77735CF436656CULL, "gen00fl_varric6087597.cl"},
	{0x5B77AD00145C80B3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowc_n.dds"},
	{0x5B78700BF24CD4B0ULL, "479565086.ogg"},
	{0x5B78D3DBDA384B01ULL, "trigger_core.ncs"},
	{0x5B78FD41499D3C17ULL, "hm_brekker.mor"},
	{0x5B7930B025DEAA9EULL, "dae000_merchant_access_hf6162990.cl"},
	{0x5B7975A1327DE681ULL, "lgt160cr_wk_soph_cultist.utc"},
	{0x5B797DFB28B5375FULL, "214424817.ogg"},
	{0x5B798A52220CD988ULL, "159127326.ogg"},
	{0x5B7990C6A6EDB9E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_up.ani"},
	{0x5B799D0658F1C04EULL, "vxw_woundedcoast_w6_0_0.vlm"},
	{0x5B7A286CAAF60EFFULL, "vxw_bonepit_1x4_0_0.vlm"},
	{0x5B7A3587B2515A8AULL, "vxlt_hangedman_1vj_0_0.vlm"},
	{0x5B7A7032100E2564ULL, "6048960_f.fxe"},
	{0x5B7AEDA34B91096AULL, "t1_skn_e07.tnt"},
	{0x5B7B00FCC66E510AULL, "438286344.ogg"},
	{0x5B7B43C114251364ULL, "art\\vfx\\zztest\\zzeventstest.anb"},
	{0x5B7B51AE24445E39ULL, "vxlt_0_lowtown_docks_cvm_0_1.vlm"},
	{0x5B7B765CAA68F63BULL, "hlac_ht_0_hightown_n_-131070_0z.dds"},
	{0x5B7B99C8435FE157ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\deadtospider.evt"},
	{0x5B7BB33BF6F7015AULL, "art\\levels\\races\\design\\blendtree\\floortrap.anb"},
	{0x5B7C0901830485A6ULL, "394256670.ogg"},
	{0x5B7C1169858649B6ULL, "687706804.ogg"},
	{0x5B7C7D03F74C2DC1ULL, "961491926.ogg"},
	{0x5B7CE95FABD77CA5ULL, "hlac_lt_undercityencounter_-1_28y.dds"},
	{0x5B7CF690EEBE5202ULL, "llac_ht_0_hightown_a2_-131072_0.dds"},
	{0x5B7D0B5AFCC49565ULL, "156047541.ogg"},
	{0x5B7DE6826F802075ULL, "759958099.ogg"},
	{0x5B7EB2B3BC47425CULL, "vxca_caverns2_837_0_0.vlm"},
	{0x5B7F4C3DCC203944ULL, "ach_oth_champion.dds"},
	{0x5B80000788B427F4ULL, "vxca_cavern1_haunted_gix_0_0.vlm"},
	{0x5B8093FFDF28F6CFULL, "6135714_m.fxe"},
	{0x5B80AEE7DED5F172ULL, "gen_im_trs_gen_junk2_03.uti"},
	{0x5B80C87744D73628ULL, "375127036.ogg"},
	{0x5B810182EC04F754ULL, "931710799.ogg"},
	{0x5B81C9B15808DE74ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appp_0s.dds"},
	{0x5B8206C5A601D444ULL, "vxht_mansion2_2po_0_0.vlm"},
	{0x5B82231CDC35ECBBULL, "dre000_sibling_has_blight6036521.cl"},
	{0x5B823829DAF8561FULL, "199912790.ogg"},
	{0x5B829AC8436601E0ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\deadtospider.gad"},
	{0x5B8300CA25A1707EULL, "w_encounter2_43_0.rml"},
	{0x5B83118BAFEAF1B0ULL, "6191939_m.fxe"},
	{0x5B831ABCBEB08627ULL, "vxca_caverns1_mines_j4a_0_0.vlm"},
	{0x5B831E7006CC3584ULL, "218176537.ogg"},
	{0x5B83AC87E8CC7A8FULL, "art\\characters\\creatures\\desiredemon\\animation\\hv_dmg.evt"},
	{0x5B83B6A7DE2C59BCULL, "llac_lt_0_lowtown_a2_-262144_3.dds"},
	{0x5B83CDC56F3CD98EULL, "vxht_mansion_n_505_0_0.vlm"},
	{0x5B841A83E6914738ULL, "hightown_male_noble1_sta6170803.cl"},
	{0x5B846535E9B051A3ULL, "6174498_m.fxe"},
	{0x5B84827E14154F1EULL, "seb221_random_servants6187839.cl"},
	{0x5B848DD4B5074E89ULL, "default_warrior_male.utc"},
	{0x5B84A483B8A91B63ULL, "gen00fl_bethany6101188.cl"},
	{0x5B84A8D6AB59CA10ULL, "vxlt_foundry_n_143_0_0.vlm"},
	{0x5B84CDCA49C1D123ULL, "6170780_m.fxe"},
	{0x5B857B9619DC2E1EULL, "280799595.ogg"},
	{0x5B858F660F85648EULL, "hlac_w_encounter3_0_10z.dds"},
	{0x5B8609D40B2F3D68ULL, "ave100_donnic6105940.cl"},
	{0x5B8621B5C7622825ULL, "338729371.ogg"},
	{0x5B8647770DE58D0BULL, "vxw_encounter2_camp_j9_0_0.vlm"},
	{0x5B86C2F68160AF10ULL, "vxca_caverns1_j5p_0_0.vlm"},
	{0x5B872A76BC242477ULL, "mag110_vincento6036914.cl"},
	{0x5B87EC99589C60E3ULL, "4687242.ogg"},
	{0x5B8862C1142FD267ULL, "art\\vfx\\zztest\\zzeventstest.mmh"},
	{0x5B89C1575FA4758DULL, "908287793.ogg"},
	{0x5B89D05AE7C2C063ULL, "vxlt_0_lowtown_a2_n_eqf_0_1.vlm"},
	{0x5B8A19CA25A7B942ULL, "w_encounter2_43_0.phy"},
	{0x5B8A76F0988EAC04ULL, "vxht_mansion_n_6o9_0_0.vlm"},
	{0x5B8AAD87E8D29B38ULL, "art\\characters\\creatures\\desiredemon\\animation\\hv_dmg.gad"},
	{0x5B8B08EBBE4D0045ULL, "vxlt_gallowscourtyard_70n_0_0.vlm"},
	{0x5B8B61365EA3937EULL, "542373604.ogg"},
	{0x5B8BBDB20AC8EA9AULL, "198484938.ogg"},
	{0x5B8BE8BFBF2EED1EULL, "vxlt_0_lowtown_a3_n_c_hq8_0_0.vlm"},
	{0x5B8C0FA4CDE9C292ULL, "gen00fl_isabela6080837.cl"},
	{0x5B8C20A17BB4BA29ULL, "vxht_0_hightown_n_b5z_0_0.vlm"},
	{0x5B8C2B7DB6B68686ULL, "llac_w_sundermount_0_3.dds"},
	{0x5B8C4C8810F850C6ULL, "lt_0_lowtown_37_0.anb"},
	{0x5B8C7DB2EDFD95FCULL, "hlac_lt_0_lowtown_a3_n_c_196607_0x.dds"},
	{0x5B8CE7A575C42D76ULL, "6206815_m.fxe"},
	{0x5B8CF5387AECDBB0ULL, "gen00fl_isabela6135627.cl"},
	{0x5B8D36FB82905BCBULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_lp_tag_c.anb"},
	{0x5B8D562ECCF676E1ULL, "380669496.ogg"},
	{0x5B8DD12A16330C6BULL, "llac_lt_0_lowtown_a2_131070_2.dds"},
	{0x5B8DD8B7883D8C88ULL, "vxw_encounter1_bw_0_0.vlm"},
	{0x5B8DFDB8CC473B71ULL, "80209330.ogg"},
	{0x5B8E2A26607D41A3ULL, "174284203.ogg"},
	{0x5B8E79C6A6FFEEE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_up.gad"},
	{0x5B8EFA617E36580BULL, "vxht_keep_n_53e_0_0.vlm"},
	{0x5B8F1484B7BAABF2ULL, "274047134.ogg"},
	{0x5B8F2012441767B4ULL, "dre000_cut_varric_intro6033474.cl"},
	{0x5B8F8409CCCA4374ULL, "6104714_f.fxe"},
	{0x5B8FBA9F82699FB4ULL, "6128727_f.fxe"},
	{0x5B8FF4EB76FA36F1ULL, "art\\characters\\playercharacter\\humanmale\\boo_medp.mmh"},
	{0x5B9000EB76FA4A92ULL, "art\\characters\\playercharacter\\humanmale\\boo_medp.mao"},
	{0x5B900EEB76FA62AFULL, "art\\characters\\playercharacter\\humanmale\\boo_medp.msh"},
	{0x5B901752C63C3808ULL, "hlac_ca_caverns1_24424_1z.dds"},
	{0x5B9046E91DEB91F7ULL, "6176210_m.fxe"},
	{0x5B90B52FEEC338E1ULL, "6081210_f.fxe"},
	{0x5B90C98098574969ULL, "vxlt_0_lowtown_gos_0_1.vlm"},
	{0x5B90FFECEC52C319ULL, "212451556.ogg"},
	{0x5B91C93E8EB4C720ULL, "hlac_lt_foundry_n_320_5z.dds"},
	{0x5B92324D34BB15F0ULL, "vxlt_hangedman_n_28n_0_0.vlm"},
	{0x5B9266FDC584AB2EULL, "art\\vfx\\combat\\vfx_gib_glv_meda.mao"},
	{0x5B92BDA21F926673ULL, "660148439.ogg"},
	{0x5B934BB665928F39ULL, "318838424.ogg"},
	{0x5B9380037152C0A6ULL, "6161291_m.fxe"},
	{0x5B93BAF0BA098560ULL, "hlac_lt_0_lowtown_docks_n_-131075_5x.dds"},
	{0x5B93BEA179B64595ULL, "982374725.ogg"},
	{0x5B94B953FFDBA58BULL, "vxw_encounter2_camp_e8_0_0.vlm"},
	{0x5B94BF10FFF1ECCDULL, "gen_im_fla_isabela_02.uti"},
	{0x5B94D1D31555B81BULL, "485288963.ogg"},
	{0x5B94DBE08803693EULL, "vxca_caverns1_mines_huy_0_0.vlm"},
	{0x5B94E35F2A2C0EF7ULL, "ca_caverns1_26037.rml"},
	{0x5B9554C2096C9EE2ULL, "dae000_merchant_armor_hm6164282.cl"},
	{0x5B95E74EF84F7101ULL, "vxht_0_hightown_a2_b95_0_0.vlm"},
	{0x5B95FD708084E658ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.anb"},
	{0x5B964D273AE17A60ULL, "mag200_feynriel6205364.cl"},
	{0x5B9689E0128FAFBFULL, "face_face_p15.psh.fxo.dx11"},
	{0x5B96C3955E1208F9ULL, "124338215.ogg"},
	{0x5B96E3F9AF19761FULL, "lowtown_fem_alienage1_sta6162554.cl"},
	{0x5B970437421CE282ULL, "follower_barks6153407.cl"},
	{0x5B971941963D8B8AULL, "vxlt_hangedman_1xd_0_0.vlm"},
	{0x5B97ACE9854917C4ULL, "vxca_cavern1_haunted_ih1_0_0.vlm"},
	{0x5B97B83151A70D0DULL, "follower_banter6143350.cl"},
	{0x5B984BF7004BF936ULL, "follower_banter6171923.cl"},
	{0x5B986779D7B90950ULL, "mag291_demon_marethari6166169.cl"},
	{0x5B98FF2DD30CD15EULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.cess.evt"},
	{0x5B997AC8B04318C5ULL, "vxlt_gallowscourtyard_75m_0_0.vlm"},
	{0x5B99CD6EB4AA5302ULL, "vxlt_0_lowtown_docks_a3_czp_0_0.vlm"},
	{0x5B9A5139D9B8F9A2ULL, "hlac_ca_caverns2_5802_41z.dds"},
	{0x5B9A99FC30B95DD8ULL, "6169420_m.fxe"},
	{0x5B9AC56BA37B3D95ULL, "6109225_m.fxe"},
	{0x5B9AFD6133084799ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he05.msh"},
	{0x5B9B076133085887ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he05.mmh"},
	{0x5B9B2A6F73BE603BULL, "vxca_caverns1_mines_eh6_0_0.vlm"},
	{0x5B9B4940A8DDD1CFULL, "qun211cr_braggart_ally_f2.utc"},
	{0x5B9C4A947A33A508ULL, "vxw_encounter1_gt_0_0.vlm"},
	{0x5B9C70D645CD54D7ULL, "dae300_alistair_king.bnk"},
	{0x5B9CB471124F893BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\prp_guardianlegs_1.mao"},
	{0x5B9CB6023DB8EC11ULL, "ch1alphemmultnt_vl.psh.fxo"},
	{0x5B9CDD2794E1CFADULL, "6127179_m.fxe"},
	{0x5B9CE49891EC3D4CULL, "mag311_emile_de_launcet6043685.cl"},
	{0x5B9D2A1C86405D9EULL, "vxw_encounter2_ft_0_0.vlm"},
	{0x5B9D9039DD087B06ULL, "lgt110ip_at_gang_hideout_d.utp"},
	{0x5B9E17EB10C4D1DDULL, "132069019.ogg"},
	{0x5B9E6B0AC83C856AULL, "6055574_m.fxe"},
	{0x5B9E8395941F296FULL, "hlac_lt_0_lowtown_docks_131071_1y.dds"},
	{0x5B9EB13AADBA8DE9ULL, "vxw_encounter2_camp_8y_0_0.vlm"},
	{0x5B9EE2690B2898D9ULL, "lt_0_lowtown_docks_a3_106_0.phy"},
	{0x5B9F367DFEDF0BE7ULL, "6065253_m.fxe"},
	{0x5B9F52D645CF57EFULL, "dae300_alistair_king.cnv"},
	{0x5B9FC764B69F9D37ULL, "6036711_f.fxe"},
	{0x5B9FF18BBD34E6C5ULL, "vxht_0_hightown_a3_bx7_0_0.vlm"},
	{0x5BA0162DD3131689ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.cess.gad"},
	{0x5BA0A681F5B9C14FULL, "6118755_m.fxe"},
	{0x5BA0FC556CAC33F9ULL, "w_bonepitdestroyed_a3_133_0.msh"},
	{0x5BA0FD8A4C929551ULL, "6179914_m.fxe"},
	{0x5BA106556CAC44E7ULL, "w_bonepitdestroyed_a3_133_0.mmh"},
	{0x5BA10A556CAC4B34ULL, "w_bonepitdestroyed_a3_133_0.mao"},
	{0x5BA11CC363C14381ULL, "vxca_caverns2_83f_0_0.vlm"},
	{0x5BA175215DD05F6DULL, "6143450_m.fxe"},
	{0x5BA19628C0C4B988ULL, "w_bonepit_244_0.msh"},
	{0x5BA1A828C0C4D819ULL, "w_bonepit_244_0.mao"},
	{0x5BA1B428C0C4ECB2ULL, "w_bonepit_244_0.mmh"},
	{0x5BA1C4E011FBE926ULL, "vxlt_0_lowtown_a2_gc1_0_0.vlm"},
	{0x5BA1FE8E2781B27CULL, "vxw_encounter3_8g_0_0.vlm"},
	{0x5BA24D4F20F58709ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.impa.ani"},
	{0x5BA255D6F3F48841ULL, "383420410.ogg"},
	{0x5BA2A47909376C9AULL, "6068372_m.fxe"},
	{0x5BA3D4053687C603ULL, "hlac_lt_0_lowtown_131070_0x.dds"},
	{0x5BA3F86210DDFAE7ULL, "6086931_m.fxe"},
	{0x5BA41F95886BE041ULL, "287508371.ogg"},
	{0x5BA43139F660C574ULL, "vxdr_primevaltomb_6po_0_1.vlm"},
	{0x5BA4FCD3DB6F8669ULL, "6102938_m.fxe"},
	{0x5BA506D5591BDEE1ULL, "23100768.ogg"},
	{0x5BA50C72CDF62FDCULL, "mag170_thrask6206388.cl"},
	{0x5BA54A45FD9039C1ULL, "vxlt_gallowstemplar_n_5p5_0_0.vlm"},
	{0x5BA56171509C6220ULL, "gen00fl_merrill6111276.cl"},
	{0x5BA57E0380D8587FULL, "vxw_bonepit_1k1_0_0.vlm"},
	{0x5BA60853635150AFULL, "dre000_varric6207953.cl"},
	{0x5BA64A953CAE4A51ULL, "431260355.ogg"},
	{0x5BA671499D2154DCULL, "413876535.ogg"},
	{0x5BA678F7748A99AEULL, "art\\characters\\playercharacter\\qunarimale\\uni_nuda.msh"},
	{0x5BA692F7748AC59CULL, "art\\characters\\playercharacter\\qunarimale\\uni_nuda.mmh"},
	{0x5BA6B3B675A355CCULL, "6057072_f.fxe"},
	{0x5BA6E68DA27BC876ULL, "906563580.ogg"},
	{0x5BA6E72DD318E5D7ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.cess.ani"},
	{0x5BA7604F1326E4F7ULL, "dre000_dougal6142208.cl"},
	{0x5BA7E96B379FC846ULL, "hlac_lt_0_lowtown_0_8y.dds"},
	{0x5BA81B03A833AEEAULL, "6037116_m.fxe"},
	{0x5BA825FADD21B486ULL, "vxht_chantry_32v_0_0.vlm"},
	{0x5BA8D2AC827BFDADULL, "254409450.ogg"},
	{0x5BA9374895715615ULL, "6102325_m.fxe"},
	{0x5BA9684F20FBD38BULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.impa.gad"},
	{0x5BA9914F39932256ULL, "624492144.ogg"},
	{0x5BAA24E802D768F1ULL, "167244177.ogg"},
	{0x5BAA920AA0C2267FULL, "555311744.ogg"},
	{0x5BAAF1207418A98FULL, "vxht_0_hightown_a2_ca6_0_0.vlm"},
	{0x5BAAF2B5C34DE6C1ULL, "qun110_petrice6018573.cl"},
	{0x5BAB0074FCD4CAE1ULL, "6060524_m.fxe"},
	{0x5BAB0BF7BB3FEDCBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationcast.ani"},
	{0x5BAB939EB1769F5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_lp.ani"},
	{0x5BAC0CCCB7B2516EULL, "266886547.ogg"},
	{0x5BAC2501F01BF8C4ULL, "gen_im_trs_drk_one_axe_r1.uti"},
	{0x5BAC8A056E0D609EULL, "mrl251im_amu_chandan.uti"},
	{0x5BAD23179FB0A669ULL, "vxw_encounter2_camp_3x_0_0.vlm"},
	{0x5BADB273E195D5A0ULL, "857238456.ogg"},
	{0x5BADE0963DCA31BAULL, "6127603_f.fxe"},
	{0x5BADED044E810543ULL, "6137550_m.fxe"},
	{0x5BAE6476A767A8E2ULL, "6172398_m.fxe"},
	{0x5BAE7E9AF233C89AULL, "vxlt_0_lowtown_docks_b80_0_0.vlm"},
	{0x5BAEA1E162D00B60ULL, "6082618_f.fxe"},
	{0x5BAEF4CA25C696F6ULL, "w_encounter2_43_0.mao"},
	{0x5BAF00CA25C6AA95ULL, "w_encounter2_43_0.mmh"},
	{0x5BAF02CA25C6AE0BULL, "w_encounter2_43_0.msh"},
	{0x5BAF3508D6FE79C1ULL, "6115435_m.fxe"},
	{0x5BAF873150817497ULL, "gen00fl_aveline6114824.cl"},
	{0x5BAF97245F420085ULL, "vxlt_darktown_4qd_0_0.vlm"},
	{0x5BAFD262B43A1544ULL, "mag291_ntbanter6056834.cl"},
	{0x5BB0654F2101EDF0ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.impa.evt"},
	{0x5BB0706B1977CAFCULL, "vxw_encounter3_5d_0_0.vlm"},
	{0x5BB07C44E0789DD4ULL, "gen00fl_aveline6132890.cl"},
	{0x5BB0D022E5A54395ULL, "gen00fl_anders6202359.cl"},
	{0x5BB10C35EF7435AAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robc_0s.dds"},
	{0x5BB1832113B5E4BEULL, "6032069_m.fxe"},
	{0x5BB21696F177E475ULL, "hlac_lt_0_lowtown_a2_n_131071_6z.dds"},
	{0x5BB236D11CC8FC23ULL, "one100dg_vanard6054645.cl"},
	{0x5BB28EF4EF1DC049ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.impa.ani"},
	{0x5BB2E9329B0BBD17ULL, "dae210_gate_guard6085146.cl"},
	{0x5BB352100B02067DULL, "static_static_p101.psh.fxo.dx11"},
	{0x5BB386E92618FC31ULL, "dae000_pc_home_ambients6208205.cl"},
	{0x5BB3BC22EF865241ULL, "vxlt_gallowstemplar_n_5k4_0_0.vlm"},
	{0x5BB3EFE072CE70FFULL, "vxw_bonepit_1l0_0_0.vlm"},
	{0x5BB428C4503136A4ULL, "follower_barks6153917.cl"},
	{0x5BB4BCB831D9C776ULL, "343347918.ogg"},
	{0x5BB4E9D86F54D52CULL, "vxht_0_hightown_a3_b02_0_0.vlm"},
	{0x5BB504EFDE4FAA70ULL, "gen00fl_sebastian6189047.cl"},
	{0x5BB559E9846F8532ULL, "6066424_m.fxe"},
	{0x5BB55B3D53C76B0FULL, "qun200_party_comment6095394.cl"},
	{0x5BB5D6FA337859D2ULL, "vxht_brothel_n_6g8_0_0.vlm"},
	{0x5BB681A5F56C27CEULL, "vxw_bonepit_1o0_0_0.vlm"},
	{0x5BB753C75C0C7262ULL, "hlac_w_woundedcoast_65538_0z.dds"},
	{0x5BB7685DD7191AC5ULL, "hm_nobleman_3.mor"},
	{0x5BB7AE37AE150BFDULL, "hlac_ca_cave_2194_0z.dds"},
	{0x5BB80DC6AC94705AULL, "hlac_lt_0_lowtown_n_65534_6y.dds"},
	{0x5BB831E93E54F659ULL, "dae300_petrice6097850.cl"},
	{0x5BB85F78C9848ACAULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcfem_fenrisrom.gad"},
	{0x5BB8650F23C92EFDULL, "hlac_ca_caverns1_24423_3x.dds"},
	{0x5BB8A6231FF6579AULL, "follower_banter6171879.cl"},
	{0x5BB8A9F4EF2259CBULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.impa.gad"},
	{0x5BB8DD4157212B8FULL, "vxca_caverns1_efy_0_0.vlm"},
	{0x5BB8F3F7BB4C0242ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationcast.evt"},
	{0x5BB905517AE5E1C9ULL, "618665410.ogg"},
	{0x5BB910A494A32533ULL, "6095806_f.fxe"},
	{0x5BB92A6DA8EAB1A9ULL, "6051551_m.fxe"},
	{0x5BB9442D832810BBULL, "qun221_petrice6083628.cl"},
	{0x5BB9BC2AABAFE681ULL, "dae000_elegant6166946.cl"},
	{0x5BB9CC77A1E4483DULL, "529411481.ogg"},
	{0x5BB9EB260E3568FFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02b.mmh"},
	{0x5BB9F1260E3572A1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02b.msh"},
	{0x5BBA0DD66A2C8E9EULL, "vxw_encounter2_lv_0_0.vlm"},
	{0x5BBA621275A74E4FULL, "hlac_ht_playershouse_n_855_27y.dds"},
	{0x5BBA652EFA50AE5EULL, "6151230_m.fxe"},
	{0x5BBA6BBA3C2C7C9CULL, "6107143_m.fxe"},
	{0x5BBAA50BE0C436E6ULL, "6174028_m.fxe"},
	{0x5BBB0A64B49E87E6ULL, "gen00fl_isabela6099574.cl"},
	{0x5BBB667A3E6E05CDULL, "752406377.ogg"},
	{0x5BBB77C6BDD5B361ULL, "vxlt_gallowscourtyard_746_0_0.vlm"},
	{0x5BBB9A346BC40B79ULL, "968904171.ogg"},
	{0x5BBC287B4AF45703ULL, "w_bonepitmining_a2_248_0.mao"},
	{0x5BBC2C7B4AF45E50ULL, "w_bonepitmining_a2_248_0.mmh"},
	{0x5BBC3A7B4AF4766AULL, "w_bonepitmining_a2_248_0.msh"},
	{0x5BBC9B6DC7A06046ULL, "ht_0_hightown_a3_36_0_ll.mmh"},
	{0x5BBCF6C233094E70ULL, "223471810.ogg"},
	{0x5BBD27C0811A3673ULL, "6173653_m.fxe"},
	{0x5BBD3338728ED1FAULL, "vxht_0_hightown_a3_bc9_0_0.vlm"},
	{0x5BBD8A3BD9A5F24FULL, "6112555_m.fxe"},
	{0x5BBD9A98ED947506ULL, "hlac_lt_0_lowtown_docks_n_-131075_4y.dds"},
	{0x5BBEF2F7BB506C55ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationcast.gad"},
	{0x5BBEFC7080A7E9F3ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.mmh"},
	{0x5BBFA6F4EF287330ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.impa.evt"},
	{0x5BBFAC9EB18772D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_lp.gad"},
	{0x5BBFB04E27A3D8AAULL, "414984749.ogg"},
	{0x5BC04A5E3EA77F44ULL, "322372263.ogg"},
	{0x5BC17B9629950251ULL, "and111_donation_box6200725.cl"},
	{0x5BC1ADC64E8384FAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera0col.msh"},
	{0x5BC1BBC64E839C33ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera0col.mao"},
	{0x5BC1BFC64E83A340ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera0col.mmh"},
	{0x5BC2024E4D6BED0CULL, "mag360_samson6042099.cl"},
	{0x5BC22F6F0E8E61F3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_female_rogue.anb"},
	{0x5BC2D2A82FF53369ULL, "6172326_m.fxe"},
	{0x5BC30A6AE069A829ULL, "vxdr_primevalentrance_32x_0_0.vlm"},
	{0x5BC349F461437F45ULL, "612523073.ogg"},
	{0x5BC37CB4D809BEB0ULL, "vxlt_0_lowtown_docks_c4o_0_0.vlm"},
	{0x5BC3967FA9503CEEULL, "hlac_w_bonepit_4_0z.dds"},
	{0x5BC3C2CA26BC400DULL, "219009316.ogg"},
	{0x5BC40D1BB3BD7113ULL, "vxdr_primevaltomb_353_0_0.vlm"},
	{0x5BC5007C7F2A5F48ULL, "vxlt_gallowscourtyard_743_0_0.vlm"},
	{0x5BC65A30097EEA7AULL, "hla_ht_0_hightown_a2_-65537_0y.dds"},
	{0x5BC6B5C846E0FBB2ULL, "gen00fl_isabela6106177.cl"},
	{0x5BC6F061332D6CBCULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he05.phy"},
	{0x5BC707A96979F8CFULL, "art\\vfx\\environment\\vfx_ash_falling_p.mmh"},
	{0x5BC720C1E9A54997ULL, "6053686_f.fxe"},
	{0x5BC7C5F4F40175CCULL, "hlac_lt_0_lowtown_a3_n_c_1_9y.dds"},
	{0x5BC8523D74F5F0E0ULL, "llac_lt_gallowstemplar_f_2388_6.dds"},
	{0x5BC8D771A125A478ULL, "274422102.ogg"},
	{0x5BC9073D52A2EA3CULL, "127514666.ogg"},
	{0x5BC94B6DA60DD8A1ULL, "gen00fl_aveline6139971.cl"},
	{0x5BC960E5AEE70519ULL, "349920051.ogg"},
	{0x5BC96B58C6E5E7D3ULL, "253032798.ogg"},
	{0x5BC9802BBD8AE67CULL, "framebuffereffects_smallblur_p0.psh.fxo.dx11.assembly"},
	{0x5BCA06E0C4B41ABAULL, "vxht_mansion2b_3gp_0_1.vlm"},
	{0x5BCA4DF507482AEDULL, "482505090.ogg"},
	{0x5BCA5EFACBCE3FE3ULL, "vxlt_0_lowtown_docks_aoc_0_0.vlm"},
	{0x5BCAA9F7392B3A44ULL, "fen110_fenris6097052.cl"},
	{0x5BCAC45021B662BAULL, "6129803_f.fxe"},
	{0x5BCAC8342EFD1F20ULL, "vxw_bonepitdestroyed_a3_1wg_0_1.vlm"},
	{0x5BCADA5EA156D1DCULL, "vxht_brothel_n_4fg_0_0.vlm"},
	{0x5BCADD64F701FF36ULL, "hlac_lt_undercityserialkiller_-1_31z.dds"},
	{0x5BCB10AEF7AD3981ULL, "cod_ltr_pro_meeran.ncs"},
	{0x5BCB406968D5372FULL, "6166992_m.fxe"},
	{0x5BCBB52C719D987CULL, "vxlt_0_lowtown_a2_erf_0_0.vlm"},
	{0x5BCBC0CADE784DCCULL, "6062813_m.fxe"},
	{0x5BCC611E27504D07ULL, "cd_one280_golem.wbd"},
	{0x5BCC71F774AB5D67ULL, "art\\characters\\playercharacter\\qunarimale\\uni_nuda.phy"},
	{0x5BCCAF341E2AA5ADULL, "dae000cr_follower_apr.ncs"},
	{0x5BCD3A78C996A7B0ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcfem_fenrisrom.ani"},
	{0x5BCD53143F382229ULL, "6079089_m.fxe"},
	{0x5BCD5C1888612827ULL, "6118037_f.fxe"},
	{0x5BCD7ADAD7E1704FULL, "6108595_f.fxe"},
	{0x5BCDA077407DE79FULL, "6077351_m.fxe"},
	{0x5BCDC787645B2560ULL, "366789690.ogg"},
	{0x5BCE254B07EA6714ULL, "fx_dot_add.mao"},
	{0x5BCE5C33F9524CE9ULL, "mer141_jansen6031852.cl"},
	{0x5BCEAD67C1CF29B0ULL, "hlac_ht_chantry_2534_12x.dds"},
	{0x5BCF007CA7BD308EULL, "6036448_f.fxe"},
	{0x5BCF6837BF2F401EULL, "vxlt_undercitydungeon_1s7_0_0.vlm"},
	{0x5BCF7A8924F9D254ULL, "hlac_w_sundermount_131077_0x.dds"},
	{0x5BCFACE3BD0B9D22ULL, "hlac_lt_gallowstemplar_2395_1x.dds"},
	{0x5BCFCCA46D745E83ULL, "728894278.ogg"},
	{0x5BD016CFB5A56D35ULL, "vxw_bonepitmining_a2_58_0_0.vlm"},
	{0x5BD036148C6BEF28ULL, "273778983.ogg"},
	{0x5BD0C2489378CE27ULL, "w_encounter2_camp_37_0_ll.mmh"},
	{0x5BD0CDE2E2FF007BULL, "447939436.ogg"},
	{0x5BD0FFCB27A6A568ULL, "llac_w_bonepitdestroyed_a3_393216_0.dds"},
	{0x5BD129AB70D81AA8ULL, "vxca_caverns1_mines_k2p_0_1.vlm"},
	{0x5BD132F934920D67ULL, "984130341.ogg"},
	{0x5BD2369B4B455E9AULL, "18828900.ogg"},
	{0x5BD2C859D5158266ULL, "463709827.ogg"},
	{0x5BD2CDCC6A5FB7F6ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_2_darren.ani"},
	{0x5BD2F77357A93C8EULL, "art\\vfx\\models\\vfx_debris_03.msh"},
	{0x5BD2F8EEA71635CAULL, "qun120cr_javaris6043051.cl"},
	{0x5BD2FE88DF032481ULL, "vxdr_primevaltomb_3m1_0_0.vlm"},
	{0x5BD3057357A954DFULL, "art\\vfx\\models\\vfx_debris_03.mao"},
	{0x5BD3117357A9697CULL, "art\\vfx\\models\\vfx_debris_03.mmh"},
	{0x5BD33B06C62AB53EULL, "6126548_m.fxe"},
	{0x5BD3413985E131B9ULL, "vxht_0_hightown_a2_aur_0_0.vlm"},
	{0x5BD37259712077C8ULL, "vxlt_gallowscourtyard_790_0_0.vlm"},
	{0x5BD3CB58E15F0039ULL, "vxw_encounter2_o4_0_0.vlm"},
	{0x5BD3D4A58951EF6AULL, "hlac_ht_0_hightown_n_-196607_0z.dds"},
	{0x5BD4557E04E9F8EEULL, "vxdr_primevalentrance_23q_0_0.vlm"},
	{0x5BD528F27BF80F2EULL, "dae000_mother6135753.cl"},
	{0x5BD5559DDC81E1D8ULL, "620266204.ogg"},
	{0x5BD56D2887B6CA95ULL, "isa211_isabela6202189.cl"},
	{0x5BD5FE88539BAD59ULL, "fen110_fenris6100360.cl"},
	{0x5BD67C3BB5603979ULL, "891227569.ogg"},
	{0x5BD69656C7989425ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masa_0n.dds"},
	{0x5BD6CECA08807046ULL, "character_character_p90.psh.fxo.dx11.assembly"},
	{0x5BD6F1904E18BF9EULL, "vxw_encounter2_rh_0_0.vlm"},
	{0x5BD721F06F01788DULL, "495852538.ogg"},
	{0x5BD74F2FC7E90613ULL, "6081710_m.fxe"},
	{0x5BD76B062E4A4A91ULL, "6078806_m.fxe"},
	{0x5BD7FC3F8A3DD163ULL, "441563428.ogg"},
	{0x5BD87E46B006D606ULL, "vxlt_blackemporium_4e_0_0.vlm"},
	{0x5BD917575862A31AULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgtb_rendermesh.mmh"},
	{0x5BD929575862C260ULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgtb_rendermesh.msh"},
	{0x5BD93A1120F337A0ULL, "vxw_bonepitdestroyed_a3_1xf_0_1.vlm"},
	{0x5BD9878372ED22BFULL, "6219018_m.fxe"},
	{0x5BD9E13E79323059ULL, "53189061.ogg"},
	{0x5BDA0D6B7898A198ULL, "character_character_p236.psh.fxo.dx11"},
	{0x5BDA11C9287AD4FCULL, "lowtown_fem_alienage1_sta.cnv"},
	{0x5BDA24D467E7AC7BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_basic.phy"},
	{0x5BDA27096393B0FCULL, "vxlt_0_lowtown_a2_eug_0_0.vlm"},
	{0x5BDA312E25BF0F1FULL, "qun160_vashothlead6053873.cl"},
	{0x5BDA73AFFA08E097ULL, "6105995_m.fxe"},
	{0x5BDA7F860254816CULL, "167653859.ogg"},
	{0x5BDA881C7537182AULL, "545458322.ogg"},
	{0x5BDA9C326CC999EFULL, "158946574.ogg"},
	{0x5BDB57C4D271C7C5ULL, "vxht_keep_538_0_0.vlm"},
	{0x5BDB5FF18FC43A31ULL, "vxw_woundedcoast_h8_0_0.vlm"},
	{0x5BDB6609CB0FE3F9ULL, "896589004.ogg"},
	{0x5BDB727CB06718D8ULL, "ach_exp_mansion.dds"},
	{0x5BDD1EF3F047715FULL, "6092479_m.fxe"},
	{0x5BDD33C9287D444CULL, "lowtown_fem_alienage1_sta.bnk"},
	{0x5BDE33205319FD78ULL, "gen00fl_aveline6139419.cl"},
	{0x5BDE3AF4A8767F56ULL, "dae000_bodhan_letter6144324.cl"},
	{0x5BDE6B9A255DA178ULL, "hm_gamlen_h3s2.mor"},
	{0x5BDEA335F2069A30ULL, "vxdr_slavers_1jc_0_0.vlm"},
	{0x5BDED29156DABA81ULL, "vxht_brothel_n_4a7_0_0.vlm"},
	{0x5BDF0C3B220DD55BULL, "987825585.ogg"},
	{0x5BDF2FD5CC32FBACULL, "973482067.ogg"},
	{0x5BDFA320C672A94AULL, "llac_lt_0_lowtown_a2_n_c_-65536_6.dds"},
	{0x5BDFB3BEAFCDD1CFULL, "llac_lt_0_lowtown_a2_n_c_131070_5.dds"},
	{0x5BE0A07B693E10B7ULL, "vxw_bonepitmining_a2_1xn_0_0.vlm"},
	{0x5BE0E4D6B1BAFDFEULL, "6090238_m.fxe"},
	{0x5BE0F964F890C06BULL, "6078101_m.fxe"},
	{0x5BE17799CE734673ULL, "fen110ar_abandoned_house.are"},
	{0x5BE1B31677D74A39ULL, "vxht_0_hightown_a2_apq_0_0.vlm"},
	{0x5BE2AF19CDE72758ULL, "gen00fl_aveline6120857.cl"},
	{0x5BE2C75AF6E0116EULL, "vxdr_primevalentrance_26r_0_0.vlm"},
	{0x5BE2CC3BD10DD4B6ULL, "gen00fl_anders6119113.cl"},
	{0x5BE2E06195FD2A57ULL, "llac_ca_caverns2prc_5802_65.dds"},
	{0x5BE3E99654B80735ULL, "vxlt_gallowstemplar_4ve_0_0.vlm"},
	{0x5BE41D7FDF2958F9ULL, "6014718_m.fxe"},
	{0x5BE42A683F82BCCEULL, "6207702_m.fxe"},
	{0x5BE4382B695DBB32ULL, "hlac_w_encounter2_131075_0y.dds"},
	{0x5BE53DAD5A896685ULL, "hlac_w_sundermount_-131075_0z.dds"},
	{0x5BE5DCB8BBB6CBD9ULL, "510696569.ogg"},
	{0x5BE5EE395AEEC80AULL, "vxw_encounter2_camp_5m_0_0.vlm"},
	{0x5BE6359620CF8EA1ULL, "w_woundedcoast_5_0.rml"},
	{0x5BE68231AB031A0CULL, "6185170_m.fxe"},
	{0x5BE6C2DAA1B6BF8EULL, "6098872_m.fxe"},
	{0x5BE6F6CC6A70A630ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_2_darren.gad"},
	{0x5BE6FB32A4A40755ULL, "774021864.ogg"},
	{0x5BE728DD064C9133ULL, "hlac_lt_0_lowtown_131069_0z.dds"},
	{0x5BE7333F39B1ECA0ULL, "vxht_0_hightown_a2_a7q_0_0.vlm"},
	{0x5BE7AC11E02B8327ULL, "vxca_caverns1_ij9_0_0.vlm"},
	{0x5BE7B2C710C66E96ULL, "hlac_lt_0_lowtown_docks_n_196607_1y.dds"},
	{0x5BE7B894BE6883B6ULL, "788332643.ogg"},
	{0x5BE805A0AC3FD88AULL, "178861692.ogg"},
	{0x5BE82ADC77036145ULL, "vxht_0_hightown_a3_bc2_0_0.vlm"},
	{0x5BE955DFD7F361AEULL, "vxlt_foundry_n_z8_0_0.vlm"},
	{0x5BE9A4713A33B6EFULL, "art\\characters\\weapons\\unique\\aveline\\textures\\w_shd_009a_0s.dds"},
	{0x5BE9C6F5212D997CULL, "dae000_arishok6081654.cl"},
	{0x5BE9E9861E04E1D2ULL, "dre000_varric6207932.cl"},
	{0x5BEA5F598083A44DULL, "hlac_lt_0_lowtown_a2_n_-65537_2x.dds"},
	{0x5BEAAED31B12D546ULL, "6046479_m.fxe"},
	{0x5BEAC64734446A49ULL, "6197796_m.fxe"},
	{0x5BEAF007E4C092A4ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgecrgrck.msh"},
	{0x5BEAF8D940B404D9ULL, "433993512.ogg"},
	{0x5BEB0207E4C0B10DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgecrgrck.mao"},
	{0x5BEB0607E4C0B846ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgecrgrck.mmh"},
	{0x5BEC5A38B016DDA2ULL, "prop2.vsh.fxo"},
	{0x5BEC5BA1E0CCF2CAULL, "vxht_0_hightown_n_c94_0_0.vlm"},
	{0x5BED25CBDD8B99B0ULL, "57923473.ogg"},
	{0x5BED370429F4E54EULL, "6120098_m.fxe"},
	{0x5BED446E48D0D301ULL, "vxht_brothel_n_4f6_0_0.vlm"},
	{0x5BED96CF80184924ULL, "war_stagger.dds"},
	{0x5BED99F12D4EF1BCULL, "vxht_0_hightown_n_cf3_0_0.vlm"},
	{0x5BEDE59B2AC5F278ULL, "212216745.ogg"},
	{0x5BEE9C398A99D57CULL, "vxlt_house_h0_0_0.vlm"},
	{0x5BEF0AAF9339A628ULL, "619216045.ogg"},
	{0x5BEF12585B342937ULL, "vxw_bonepitmining_a2_1ui_0_0.vlm"},
	{0x5BEF74176D2D5F54ULL, "dae000_gamlen6005822.cl"},
	{0x5BEF830A11093C5AULL, "vxht_brothel_n_6k4_0_0.vlm"},
	{0x5BEFB6E036A4442DULL, "hlac_lt_0_lowtown_docks_-131073_4y.dds"},
	{0x5BF01EC0AA16E63EULL, "6128748_m.fxe"},
	{0x5BF0349632ACC2ACULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopeb.msh"},
	{0x5BF03A9632ACCC9EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopeb.mmh"},
	{0x5BF0469632ACE135ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopeb.mao"},
	{0x5BF187A118838955ULL, "607075794.ogg"},
	{0x5BF19F6156873572ULL, "mag121_idunna6084487.cl"},
	{0x5BF1C230EB375627ULL, "473231168.ogg"},
	{0x5BF208345A8A512EULL, "hlac_lt_undercitydungeon_-1_36z.dds"},
	{0x5BF23C11552C805CULL, "1059835058.ogg"},
	{0x5BF27A1F1AE9B79AULL, "dae000ip_crowd_em3_talk15.utp"},
	{0x5BF309F3E21129B0ULL, "lt_0_lowtown_docks_127_0.msh"},
	{0x5BF317F3E21141CAULL, "lt_0_lowtown_docks_127_0.mmh"},
	{0x5BF31BF3E2114801ULL, "lt_0_lowtown_docks_127_0.mao"},
	{0x5BF35DA7E24704A7ULL, "vxht_mansion2_31z_0_0.vlm"},
	{0x5BF3993AD5012D59ULL, "148588335.ogg"},
	{0x5BF3DEA10D774CE9ULL, "711545523.ogg"},
	{0x5BF4DB58A0E218DBULL, "mag291_torpor6116238.cl"},
	{0x5BF4EBE2A14B5CAFULL, "203228257.ogg"},
	{0x5BF4ECDA803AC2B1ULL, "hlac_ht_mansion2_818_24y.dds"},
	{0x5BF5591734F005A7ULL, "6147701_f.fxe"},
	{0x5BF569074D6E3DE2ULL, "hlac_w_woundedcoast_-196602_0z.dds"},
	{0x5BF5880F527A88E8ULL, "hlac_lt_0_lowtown_a3_n_c_65537_3z.dds"},
	{0x5BF59037EFA0A293ULL, "vxw_bonepit_26e_0_0.vlm"},
	{0x5BF5BA25AAEEA708ULL, "vxw_bonepitdestroyed_a3_23e_0_0.vlm"},
	{0x5BF5FD07E4CAB011ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01hgecrgrck.phy"},
	{0x5BF61DEED2219BA7ULL, "vxca_caverns1_im8_0_0.vlm"},
	{0x5BF656984F2C41A0ULL, "lgt311_shw_mer_mettin6156479.cl"},
	{0x5BF66AC6645C55E5ULL, "6092381_m.fxe"},
	{0x5BF6CD85CA914881ULL, "vxlt_undercityserialkiller_yt_0_0.vlm"},
	{0x5BF6E325050456F9ULL, "301163230.ogg"},
	{0x5BF6FDFFE6011182ULL, "vxlt_gamlenshouse_n_my_0_0.vlm"},
	{0x5BF83261A2F06326ULL, "gen00fl_aveline6139192.cl"},
	{0x5BF839EA3604D461ULL, "vxca_cavern1_haunted_j20_0_1.vlm"},
	{0x5BF914246942DA46ULL, "6058866_f.fxe"},
	{0x5BF92A8A1C863B19ULL, "art\\levels\\ztools\\default\\generic\\0ci_basic.msh"},
	{0x5BF9348A1C864C07ULL, "art\\levels\\ztools\\default\\generic\\0ci_basic.mmh"},
	{0x5BF9388A1C865354ULL, "art\\levels\\ztools\\default\\generic\\0ci_basic.mao"},
	{0x5BF96B5FD437367FULL, "370647558.ogg"},
	{0x5BFA0A3F7FC18048ULL, "hlac_lt_0_lowtown_docks_n_-65539_6x.dds"},
	{0x5BFA128BAEBA0BDBULL, "combat_2h_r1_a5.gda"},
	{0x5BFA2557C4FF87DDULL, "117755018.ogg"},
	{0x5BFA6DE997C81102ULL, "6092175_m.fxe"},
	{0x5BFA8D3F9939B96DULL, "vxlt_gallowstemplar_2rw_0_0.vlm"},
	{0x5BFAAFECC3E3C09DULL, "6111268_m.fxe"},
	{0x5BFC0B8CA5C4AB7AULL, "vxca_cavern1_haunted_ioj_0_0.vlm"},
	{0x5BFC410D243F452FULL, "514532331.ogg"},
	{0x5BFC75B60E77544AULL, "6075091_m.fxe"},
	{0x5BFCB89915D134F1ULL, "hlac_lt_gallowstemplar_n_2391_3z.dds"},
	{0x5BFCDB326893FED8ULL, "6087118_f.fxe"},
	{0x5BFD5EDBFA2FF8C7ULL, "vxw_bonepitdestroyed_a3_2a8_0_0.vlm"},
	{0x5BFD6B8D01D01E91ULL, "134643972.ogg"},
	{0x5BFD84354D2A41B7ULL, "vxw_bonepitmining_a2_1rh_0_0.vlm"},
	{0x5BFDF4E702FF54DAULL, "vxht_brothel_n_6t5_0_0.vlm"},
	{0x5BFEC492A5AD19AAULL, "430502653.ogg"},
	{0x5BFEF906A4F35D49ULL, "mag311_keran6063750.cl"},
	{0x5BFF20EFB74149B9ULL, "vxw_encounter2_p9_0_0.vlm"},
	{0x5BFF913B59DBAA40ULL, "vxw_sundermount_229_0_0.vlm"},
	{0x5BFF9CC63661C8F9ULL, "vxca_cavern1_haunted_jfe_0_0.vlm"},
	{0x5BFFC6D8FCCA757AULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_dw_hammer_0_d.dds"},
	{0x5C005EFD6C073A50ULL, "hlac_lt_0_lowtown_-65536_1z.dds"},
	{0x5C00714C9637E23AULL, "vfx_dot_blur_add.mao"},
	{0x5C009D34CC9EB7CDULL, "zz_followers_banter6157818.cl"},
	{0x5C00BF3E9B34FA2FULL, "1005225707.ogg"},
	{0x5C0120ACB4604D30ULL, "6139887_m.fxe"},
	{0x5C0125B7F09ED0EEULL, "vxlt_0_lowtown_docks_9r9_0_0.vlm"},
	{0x5C0169714F945B24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_reaction.ani"},
	{0x5C0170070D0B5FDEULL, "6069649_m.fxe"},
	{0x5C018FD121C006A2ULL, "6106807_m.fxe"},
	{0x5C01E0DB2AD5728CULL, "463189427.ogg"},
	{0x5C01F446882EECF4ULL, "gen00fl_anders6125145.cl"},
	{0x5C025A39E05BA6A7ULL, "follower_banter6164720.cl"},
	{0x5C02FE14F1006F06ULL, "251837466.ogg"},
	{0x5C037522DF71EF17ULL, "gen00fl_isabela6015850.cl"},
	{0x5C03967140A7A03DULL, "vxht_0_hightown_a2_c4s_0_0.vlm"},
	{0x5C03BDE6F09A3118ULL, "hlac_ca_caverns1_mines_24425_9z.dds"},
	{0x5C03E98F8B62FCE0ULL, "sf_cxformmul.psh.fxo"},
	{0x5C03F37A7E4DF036ULL, "zz_qun_debug6028281.cl"},
	{0x5C040214E196BB13ULL, "vxw_bonepit_21d_0_0.vlm"},
	{0x5C04120E49D4D4EBULL, "173501552.ogg"},
	{0x5C04534922826064ULL, "848407751.ogg"},
	{0x5C0499A4A0B092BFULL, "hlac_lt_0_lowtown_docks_-196610_6y.dds"},
	{0x5C04B412DBD8CD85ULL, "lt_warehouse_n_4_0_ll.mmh"},
	{0x5C052C0C11FEEF0BULL, "6075260_f.fxe"},
	{0x5C052E607190CCF9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobb_0s.dds"},
	{0x5C0549122EEA160AULL, "6206679_m.fxe"},
	{0x5C06061CA3E8A08BULL, "vxdr_primevalentrance_2ol_0_0.vlm"},
	{0x5C060BD4680CBC1AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_basic.msh"},
	{0x5C0619D4680CD453ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_basic.mao"},
	{0x5C061DD4680CDB60ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_basic.mmh"},
	{0x5C06A8B5A3C64C8FULL, "dae000_bartrand6201377.cl"},
	{0x5C06D351CBBB9227ULL, "6062797_m.fxe"},
	{0x5C0766714F98C186ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_reaction.gad"},
	{0x5C081F00E91DEF20ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba05a.phy"},
	{0x5C084D261B5A69BAULL, "gen00fl_varric6072982.cl"},
	{0x5C0891AA98031DA3ULL, "vxlt_hangedman_n_5n_0_0.vlm"},
	{0x5C08B62D5B6F28C5ULL, "vxw_bonepitdestroyed_a3_15m_0_0.vlm"},
	{0x5C08C3BB03078B90ULL, "vxca_caverns1_jfy_0_0.vlm"},
	{0x5C08CEC62E79EB69ULL, "6053588_m.fxe"},
	{0x5C08D8FEC75DA3A4ULL, "6088220_m.fxe"},
	{0x5C08E37EAD8B7D1FULL, "6153624_m.fxe"},
	{0x5C0991A367F9462EULL, "vxlt_gallowstemplar_f_5t2_0_0.vlm"},
	{0x5C09A8C02B5BD387ULL, "6086874_m.fxe"},
	{0x5C09AB6B837A6E0EULL, "vxw_encounter2_camp_b0_0_0.vlm"},
	{0x5C09E88CCA6354EEULL, "hlac_ca_cavern1_haunted_24423_11y.dds"},
	{0x5C09FCE9D597A3F0ULL, "kf_urchin_1.mor"},
	{0x5C0A3FFEC62EDBB3ULL, "vxht_0_hightown_a2_cbk_0_0.vlm"},
	{0x5C0AB16FBA1DD6EAULL, "vxca_cave_1tb_0_0.vlm"},
	{0x5C0AF2F3F564B7F2ULL, "hlac_lt_0_lowtown_a2_n_-131072_8y.dds"},
	{0x5C0B4C9E064A9C3AULL, "609074604.ogg"},
	{0x5C0B670F415CCC19ULL, "w_encounter1_22_0.msh"},
	{0x5C0B710F415CDD07ULL, "w_encounter1_22_0.mmh"},
	{0x5C0B750F415CE454ULL, "w_encounter1_22_0.mao"},
	{0x5C0B98EAA9A01EB0ULL, "6083754_petrice_grandcleric.cut"},
	{0x5C0BAFE9EB93201EULL, "6176673_m.fxe"},
	{0x5C0C0B1EDB736FA0ULL, "vxdr_primevaltomb_8q1_0_2.vlm"},
	{0x5C0C44B355FC97C2ULL, "storage_sa.dds"},
	{0x5C0C61330B76131DULL, "em_comt_servant2b.mor"},
	{0x5C0CB56879B42E9CULL, "vxlt_gallowstemplar_f_3ok_0_0.vlm"},
	{0x5C0D03D09C5AA348ULL, "zz_vrc_debug6064255.cl"},
	{0x5C0D156FA3185CFFULL, "386352879.ogg"},
	{0x5C0D6B1E2787F53DULL, "cd_one280_golem.bnk"},
	{0x5C0D8099CE989038ULL, "fen110ar_abandoned_house.lst"},
	{0x5C0DE46B9D9ADEA7ULL, "vxlt_warehouse_qk_0_0.vlm"},
	{0x5C0E4134F4E6D2E8ULL, "6137694_m.fxe"},
	{0x5C0E447F6B3A1197ULL, "vxht_mansion2b_3g0_0_0.vlm"},
	{0x5C0E61714F9ED865ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\help_up_reaction.evt"},
	{0x5C0E7281AA9A3CB9ULL, "12073350.ogg"},
	{0x5C0F5F67E0AB2C72ULL, "mrl310_party_comments6142745.cl"},
	{0x5C0F6F60315A8C15ULL, "6098798_m.fxe"},
	{0x5C0FBA845DED2A1BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a_1.mao"},
	{0x5C0FD58E2DD9E1B2ULL, "370198367.ogg"},
	{0x5C10346C07482887ULL, "hlac_w_encounter3_-262143_0y.dds"},
	{0x5C105EB9EE79C8DCULL, "6102760_m.fxe"},
	{0x5C109F5659272C73ULL, "hlac_lt_0_lowtown_docks_a3_65534_2y.dds"},
	{0x5C10A135A43B82FDULL, "6125991_m.fxe"},
	{0x5C10EEFFC2C8A5BCULL, "vxlt_gallowstemplar_45q_0_0.vlm"},
	{0x5C1112AF69C7CE76ULL, "one100_ht_banter6060904.cl"},
	{0x5C11F466D83F820DULL, "vxca_caverns1_gkb_0_0.vlm"},
	{0x5C12084E329DB8BDULL, "vxht_0_hightown_a2_c3r_0_0.vlm"},
	{0x5C127B893C96339AULL, "755424953.ogg"},
	{0x5C12F83E558E876EULL, "6067852_m.fxe"},
	{0x5C1320C2877FA9E7ULL, "gen00fl_fenris6115332.cl"},
	{0x5C13228BC945669FULL, "6089714_m.fxe"},
	{0x5C1383B730C4E227ULL, "dre000_dougal6129768.cl"},
	{0x5C139CEC0107189BULL, "narrative_varric6172553.cl"},
	{0x5C13AFA90283EDB8ULL, "art\\levels\\races\\proxy\\props\\_textures\\shelf_sa.dds"},
	{0x5C14118F166F5CBCULL, "mag311_keran6054229.cl"},
	{0x5C14978A9E5CC117ULL, "738904558.ogg"},
	{0x5C14A099CE9EE529ULL, "fen110ar_abandoned_house.ncs"},
	{0x5C1541341EFBD0F4ULL, "art\\vfx\\combat\\vfx_glb_ele_c_major_c.impa.ani"},
	{0x5C15452EBDB1AAA7ULL, "6069697_m.fxe"},
	{0x5C15934056A9B766ULL, "mcr370_desten6087289.cl"},
	{0x5C15A88D626385A7ULL, "6166934_m.fxe"},
	{0x5C15FD28D0BAFEB2ULL, "vxht_mansion2_23d_0_0.vlm"},
	{0x5C163B92E8F6FBC4ULL, "vxht_keep_5tx_0_0.vlm"},
	{0x5C1650F579EE1CB3ULL, "gen_im_cns_flu_varric_01.uti"},
	{0x5C1664D6FCBEB51DULL, "614267904.ogg"},
	{0x5C16937EAC577CE0ULL, "lowtown_male_cityguard_mob6150486.cl"},
	{0x5C16DF67E5814CE2ULL, "w_bonepit_21_0.rml"},
	{0x5C174ADBB953BC24ULL, "6083320_m.fxe"},
	{0x5C17555B9F81959FULL, "6156524_m.fxe"},
	{0x5C18AFA6E140F184ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.cess.ani"},
	{0x5C18EF4689B0DC7AULL, "vxca_cavern1_haunted_iad_0_0.vlm"},
	{0x5C19CC25579B7A7CULL, "vxdr_primevalentrance_33b_0_0.vlm"},
	{0x5C1A228152C58264ULL, "6172389_m.fxe"},
	{0x5C1A346C11B0E20AULL, "hlac_dr_dungeon_3214_7y.dds"},
	{0x5C1A3B6205B0E1E8ULL, "aip_qunariwarrior_base.gda"},
	{0x5C1A8A20D95DA535ULL, "6108433_m.fxe"},
	{0x5C1B3E341F003756ULL, "art\\vfx\\combat\\vfx_glb_ele_c_major_c.impa.gad"},
	{0x5C1B9EB09326D34EULL, "vxw_bonepit_127_0_0.vlm"},
	{0x5C1C56488F90F727ULL, "vxlt_warehouse_zj_0_0.vlm"},
	{0x5C1C65D5332ED21EULL, "vxht_brothel_n_6cx_0_0.vlm"},
	{0x5C1CF4260E89F544ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02b.phy"},
	{0x5C1CFC429412BC94ULL, "6082301_m.fxe"},
	{0x5C1D018778C6BC4CULL, "gen_im_trs_krk_grd_junk2_01.uti"},
	{0x5C1D37081FB37063ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_bdy_snga_0s.dds"},
	{0x5C1D623838A71601ULL, "gen00fl_isabela_init6077916.cl"},
	{0x5C1E1B7C596877AAULL, "qun211_braggart6075469.cl"},
	{0x5C1E24B7B6C90781ULL, "vxdr_primevaltomb_6yl_0_0.vlm"},
	{0x5C1E590B6620E928ULL, "6114402_m.fxe"},
	{0x5C1EA2D478BE7A03ULL, "llac_lt_backalley_g_65536_2.dds"},
	{0x5C1F27BA57D67C8EULL, "vxlt_0_lowtown_a3_n_c_ghd_0_0.vlm"},
	{0x5C1FACA6E1470BE6ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.cess.gad"},
	{0x5C202A8859F1F06DULL, "vxca_caverns2_8r5_0_0.vlm"},
	{0x5C204F43A8FA7315ULL, "6102156_m.fxe"},
	{0x5C20E02EDEB55A06ULL, "limp_streak01.mao"},
	{0x5C21299ED7169A8BULL, "llac_ht_keep_n_3732_17.dds"},
	{0x5C2138F320C5EA7EULL, "zz_mag_debug6044334.cl"},
	{0x5C223AFB88AF1DE2ULL, "dre000_varric6034657.cl"},
	{0x5C2259341F0683B5ULL, "art\\vfx\\combat\\vfx_glb_ele_c_major_c.impa.evt"},
	{0x5C22677C7B6D64B3ULL, "gen00fl_aveline6139951.cl"},
	{0x5C22F826AEBBF8B5ULL, "6130988_m.fxe"},
	{0x5C235BAD8ED5CDF7ULL, "vxht_keep_670_0_1.vlm"},
	{0x5C2399424992B4B1ULL, "vxw_woundedcoast_q3_0_0.vlm"},
	{0x5C23AAFC3ECBB71AULL, "vxht_0_hightown_n_c6u_0_0.vlm"},
	{0x5C23BFF776FD996DULL, "621856681.ogg"},
	{0x5C23E70D5881D379ULL, "6015934_f.fxe"},
	{0x5C240C39ADA3F79AULL, "plt_prm_cod_resilient_ring.plo"},
	{0x5C245E6362241306ULL, "follower_barks6153515.cl"},
	{0x5C2473A778825472ULL, "6080998_m.fxe"},
	{0x5C24C12EB3900E7CULL, "vxht_keep_n_6br_0_0.vlm"},
	{0x5C251D8A1CAB603CULL, "art\\levels\\ztools\\default\\generic\\0ci_basic.phy"},
	{0x5C253A4B9933B981ULL, "24969793.ogg"},
	{0x5C2582853CF4322AULL, "dae000_quintus6153102.cl"},
	{0x5C25A7A6E14B6EC5ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.cess.evt"},
	{0x5C26279621055A90ULL, "w_woundedcoast_5_0.msh"},
	{0x5C263596210572AAULL, "w_woundedcoast_5_0.mmh"},
	{0x5C263996210579E1ULL, "w_woundedcoast_5_0.mao"},
	{0x5C2645C13645F0D8ULL, "llac_ht_mansion2b_818_46.dds"},
	{0x5C267C4FF6169817ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_stonestand.phy"},
	{0x5C27C967E58F9A37ULL, "w_bonepit_21_0.msh"},
	{0x5C27DB67E58FB89AULL, "w_bonepit_21_0.mao"},
	{0x5C27DF67E58FBFA9ULL, "w_bonepit_21_0.mmh"},
	{0x5C27FAD48C80CEE4ULL, "vxht_keep_6a8_0_0.vlm"},
	{0x5C2846E31B5D6A26ULL, "dae000_meredith6085264.cl"},
	{0x5C2888A12718839BULL, "793555610.ogg"},
	{0x5C28B1ED471A79F1ULL, "gen00fl_carver6103397.cl"},
	{0x5C292B94842FB87BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so_galow.mmh"},
	{0x5C293794842FCC18ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so_galow.mao"},
	{0x5C294994842FEB65ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so_galow.msh"},
	{0x5C2953D01ABED7F6ULL, "6023834_m.fxe"},
	{0x5C2A1B67A8CEA158ULL, "gen00fl_merrill6108519.cl"},
	{0x5C2A2CE4BF84D08FULL, "6139189_m.fxe"},
	{0x5C2AA4B3179F72D1ULL, "vxdr_0_deeproads_eqq_0_0.vlm"},
	{0x5C2B6888719D1F95ULL, "6179060_m.fxe"},
	{0x5C2CCAE8581701A8ULL, "6113502_m.fxe"},
	{0x5C2D3C00E93D3CB5ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba05a.msh"},
	{0x5C2D3E00E93D402BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba05a.mmh"},
	{0x5C2D61C53F7BBEF3ULL, "follower_banter6172738.cl"},
	{0x5C2D8C6656DFF617ULL, "662285088.ogg"},
	{0x5C2DB874D40D9455ULL, "gen00fl_anders6119167.cl"},
	{0x5C2DDE4D3902CFC9ULL, "340617970.ogg"},
	{0x5C2E0BEBE3095C2EULL, "6110390_m.fxe"},
	{0x5C2E15D3AA977E55ULL, "vxlt_0_lowtown_a2_n_c_gco_0_0.vlm"},
	{0x5C2E3091F1AD777CULL, "w_sundermount_68_0_ll.mmh"},
	{0x5C2EBC299E23ECD9ULL, "105634246.ogg"},
	{0x5C2EF7DB10009219ULL, "vxht_brothel_n_5jw_0_0.vlm"},
	{0x5C2F3820E1674824ULL, "6201532_f.fxe"},
	{0x5C2F81DEFFF39BABULL, "6219292_m.fxe"},
	{0x5C2F84764018EB5CULL, "6119147_f.fxe"},
	{0x5C2F9256875E1D2FULL, "949254181.ogg"},
	{0x5C2FD502E750BD54ULL, "vxlt_darktown_41q_0_0.vlm"},
	{0x5C2FE7020711E8F0ULL, "6093592_m.fxe"},
	{0x5C2FEA565F4DD5FAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_chimney.mmh"},
	{0x5C2FEE565F4DDC31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_chimney.mao"},
	{0x5C2FFC565F4DF440ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_chimney.msh"},
	{0x5C303F56DE2C936EULL, "6166571_m.fxe"},
	{0x5C3096C13A69FFF0ULL, "llac_lt_0_lowtown_a2_n_c_65538_1.dds"},
	{0x5C30A3BFD7775283ULL, "vxlt_0_lowtown_hae_0_0.vlm"},
	{0x5C31119632E41E39ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopeb.phy"},
	{0x5C31505A97B49A2CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_lp.gad"},
	{0x5C315580EE1EE2E2ULL, "vxht_0_hightown_a2_bdy_0_0.vlm"},
	{0x5C31986489029CF7ULL, "vxlt_0_lowtown_n_csg_0_0.vlm"},
	{0x5C31BA05714237C9ULL, "36893031.ogg"},
	{0x5C31BC35523A6F92ULL, "6096204_f.fxe"},
	{0x5C31BCDFB137AB15ULL, "vxw_woundedcoast_1bg_0_0.vlm"},
	{0x5C321CD930C1CF9AULL, "vxht_0_hightown_n_c3v_0_0.vlm"},
	{0x5C3299CF02DFBC31ULL, "6102688_carver_hit_someone.cut"},
	{0x5C32C3D5DBCBA306ULL, "6116569_m.fxe"},
	{0x5C3325B8F47D5A59ULL, "vxlt_gallowstemplar_f_3r9_0_0.vlm"},
	{0x5C335B7FEC1B3C68ULL, "mcr370_negotiation6080183.cl"},
	{0x5C337E77AA23835AULL, "art\\levels\\races\\design\\prp_chestc.mmh"},
	{0x5C338277AA238991ULL, "art\\levels\\races\\design\\prp_chestc.mao"},
	{0x5C338BD634C13D7DULL, "2174771.ogg"},
	{0x5C339077AA23A1A0ULL, "art\\levels\\races\\design\\prp_chestc.msh"},
	{0x5C33AE8C5FF495A0ULL, "lt_0_lowtown_53_0.anb"},
	{0x5C33AFB03BC359A4ULL, "qcr000_arishok6059978.cl"},
	{0x5C33B5C252AB2FD5ULL, "12115890.ogg"},
	{0x5C348746299AB030ULL, "dae000_bartrand6064029.cl"},
	{0x5C3492CA54B066A0ULL, "6118082_m.fxe"},
	{0x5C34D0D8A970733CULL, "mcr310_argument6072289.cl"},
	{0x5C34E6FAC02917EBULL, "hlac_lt_hangedman_n_-1_12x.dds"},
	{0x5C355F99CB7D4C44ULL, "gen00fl_aveline6152579.cl"},
	{0x5C358C4FAB7A7F49ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candlecluster_d.dds"},
	{0x5C35CA0D5BCD2B4FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_loop.gad"},
	{0x5C3617A2A52EFBA2ULL, "vxw_bonepitmining_a2_pd_0_0.vlm"},
	{0x5C366CB17E76E764ULL, "vxht_keep_6j9_0_0.vlm"},
	{0x5C3677F92B538D90ULL, "30044429.ogg"},
	{0x5C36AE888C78ADAEULL, "static_static_p104.psh.fxo.dx11"},
	{0x5C36F39969F35549ULL, "300105121.ogg"},
	{0x5C36F442B0900413ULL, "6071106_m.fxe"},
	{0x5C37729EB941B711ULL, "6070150_m.fxe"},
	{0x5C37ECBD4A2BCA1BULL, "zz_followers_banter6158358.cl"},
	{0x5C382CB73C3CCE68ULL, "ss_rogue2.cut"},
	{0x5C384F5A97BAB72AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_lp.ani"},
	{0x5C3889C0F92549CFULL, "art\\vfx\\combat\\vfx_r_var_pin_shot_act_c.mmh"},
	{0x5C38A3D395BF8F32ULL, "dae100_bran6072440.cl"},
	{0x5C38EA9E19C8018AULL, "vxlt_hangedman_2p7_0_0.vlm"},
	{0x5C39169009958B51ULL, "vxdr_0_deeproads_ezp_0_0.vlm"},
	{0x5C3925085AF7DBE1ULL, "vxw_bonepitdestroyed_a3_195_0_0.vlm"},
	{0x5C393BBFA95DE039ULL, "vxw_blightlands_60_0_0.vlm"},
	{0x5C39498F171B031EULL, "vxht_brothel_n_6iz_0_0.vlm"},
	{0x5C398C7C0B6459CBULL, "vxca_caverns2_9de_0_1.vlm"},
	{0x5C39E148FDF45D03ULL, "fx_fireflamefb03_add.mao"},
	{0x5C39E697CA22C87CULL, "dae020cr_hm_gossip4.utc"},
	{0x5C3A3245A664B987ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_female.gad"},
	{0x5C3A4C74FDACCB79ULL, "1011430427.ogg"},
	{0x5C3A799AFCD7F579ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_016a_0t.dds"},
	{0x5C3AD8F3C284DFA8ULL, "dae000_meredith6085356.cl"},
	{0x5C3CB9A00950FD45ULL, "vxca_caverns2prc_9pw_0_0.vlm"},
	{0x5C3CCBAEE7979924ULL, "llac_w_woundedcoast_-262138_0.dds"},
	{0x5C3D02F70970E0D8ULL, "zz_adjustplayer.ncs"},
	{0x5C3D3288660E2623ULL, "vxlt_0_lowtown_docks_n_b46_0_0.vlm"},
	{0x5C3D68119DC12371ULL, "mag370_meredith6046883.cl"},
	{0x5C3D7ABBA984313EULL, "924839166.ogg"},
	{0x5C3D7EA129CD3942ULL, "gen00fl_isabela6097293.cl"},
	{0x5C3E710F41883764ULL, "w_encounter1_22_0.rml"},
	{0x5C3E7FFAD66DFBDEULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_007a_0d.dds"},
	{0x5C3EB12331A9092DULL, "mag210_arianni6033879.cl"},
	{0x5C3EB47241E6D098ULL, "llac_lt_backalley_n_0_7.dds"},
	{0x5C3F1ED552B7C01CULL, "lgt210_eau_symp_templar6144476.cl"},
	{0x5C3F655565022085ULL, "6149913_m.fxe"},
	{0x5C3F8A14D8372DB1ULL, "35622489.ogg"},
	{0x5C3FA0C8C1F537F9ULL, "hlac_lt_gallowstemplar_f_2391_5x.dds"},
	{0x5C402EBCA32DC395ULL, "vxw_woundedcoast_1mf_0_0.vlm"},
	{0x5C4050822E80703EULL, "79170390.ogg"},
	{0x5C4073AEA15329D6ULL, "vxlt_gallowstemplar_38s_0_0.vlm"},
	{0x5C40837C58A074A3ULL, "ht_0_hightown_25_0.rml"},
	{0x5C408EB622B7E81AULL, "vxht_0_hightown_n_c8w_0_0.vlm"},
	{0x5C40FECF469511E4ULL, "ser270_templar6127481.cl"},
	{0x5C410C072E38054DULL, "6096870_f.fxe"},
	{0x5C412745A66AC594ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_female.evt"},
	{0x5C4140B9C626F777ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug_handsup.gad"},
	{0x5C41D6E9257443FFULL, "hla_ht_0_hightown_-1_0y.dds"},
	{0x5C426CB3B545D180ULL, "hlac_lt_hangedman_n_-1_22x.dds"},
	{0x5C4297755D09DDD1ULL, "zz_t_trailer_varric_v2.cut"},
	{0x5C42ADBE1C4C87C8ULL, "vxw_bonepitdestroyed_a3_190_0_0.vlm"},
	{0x5C4304A746A67F20ULL, "6117182_m.fxe"},
	{0x5C43078B6A05875EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maso_0n.dds"},
	{0x5C43489BD1BF31BAULL, "hlac_lt_0_lowtown_a2_n_c_131073_2y.dds"},
	{0x5C435590F01FDB77ULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_hed_ogra_0d.dds"},
	{0x5C435C25FCAB0B60ULL, "764345160.ogg"},
	{0x5C43878CF272459FULL, "lowtown_fem_alien_elf_2_mob.bnk"},
	{0x5C43BA47DC82F651ULL, "vxlt_warehouse_143_0_0.vlm"},
	{0x5C43E6EA8E256BC5ULL, "vxca_cavern1_haunted_iao_0_0.vlm"},
	{0x5C43F740D8485B47ULL, "525278790.ogg"},
	{0x5C4483E5C74E8217ULL, "llac_ht_0_hightown_-65536_0.dds"},
	{0x5C44A95075C0B64FULL, "668398146.ogg"},
	{0x5C44EBA3173AFD1CULL, "zz_prmpt_debug6200981.cl"},
	{0x5C44ED4827A8A40FULL, "599706087.ogg"},
	{0x5C45286524DA439DULL, "smallbits.mao"},
	{0x5C4541D93CFF66BAULL, "dae000_flemeth.cnv"},
	{0x5C45E3FDAC8B2B3BULL, "6208693_f.fxe"},
	{0x5C4611A02537D203ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_megatiki_d.dds"},
	{0x5C4699A823E149A8ULL, "hlac_lt_0_lowtown_0_1x.dds"},
	{0x5C4705469B7E62B5ULL, "750593161.ogg"},
	{0x5C4736C7E0260C1BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad.phy"},
	{0x5C47698CF275FBA7ULL, "lowtown_fem_alien_elf_2_mob.cnv"},
	{0x5C47886CFB8BA3D1ULL, "vxdr_0_deeproads_egw_0_0.vlm"},
	{0x5C47A471C0E278EDULL, "follower_banter6161080.cl"},
	{0x5C47A67352913851ULL, "hightown_fem_courtesan_mob6168915.cl"},
	{0x5C47A6ECF38BCE52ULL, "mag110_samson6049819.cl"},
	{0x5C47A78FE0AC8A10ULL, "hightown_fem_cityguard_mob6168435.cl"},
	{0x5C47B0F808AE2B9EULL, "vxlt_0_lowtown_a3_n_c_h25_0_0.vlm"},
	{0x5C47B29BACFC4CADULL, "709866173.ogg"},
	{0x5C4855B9C62D3904ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug_handsup.evt"},
	{0x5C485C0202521D37ULL, "w_bonepitdestroyed_a3_234_0.mmh"},
	{0x5C48600202522444ULL, "w_bonepitdestroyed_a3_234_0.mao"},
	{0x5C4860AC7E60B090ULL, "vxdr_slavers_149_0_0.vlm"},
	{0x5C487202025242A9ULL, "w_bonepitdestroyed_a3_234_0.msh"},
	{0x5C49002EB7CBD561ULL, "vxca_cavern1_haunted_in2_0_0.vlm"},
	{0x5C4963D93D038902ULL, "dae000_flemeth.bnk"},
	{0x5C49E70D5BDE0515ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_loop.ani"},
	{0x5C49FAD090D72CFAULL, "dae000_fem_commoner2_mob6174588.cl"},
	{0x5C4A10928B25FA03ULL, "vxw_encounter2_camp_p2_0_1.vlm"},
	{0x5C4AB51CB5DF0759ULL, "vxlt_0_lowtown_docks_n_b5w_0_0.vlm"},
	{0x5C4B0111C8F9998FULL, "ser120_ghyslain6104182.cl"},
	{0x5C4B1B2A05F9F079ULL, "6068744_m.fxe"},
	{0x5C4B73A771C63406ULL, "942932605.ogg"},
	{0x5C4CD9142406FE93ULL, "da2terrain_terrain_p3.psh.fxo.dx11"},
	{0x5C4CDF6740EF4DDEULL, "hlac_ht_0_hightown_65536_2x.dds"},
	{0x5C4CFD08384DBBF4ULL, "6126645_m.fxe"},
	{0x5C4D2E8A0C2D5ADBULL, "art\\vfx\\models\\vfx_isa_pose5.mmh"},
	{0x5C4D4C8A0C2D8DC5ULL, "art\\vfx\\models\\vfx_isa_pose5.msh"},
	{0x5C4D6ABE6F7AF45DULL, "vxca_cavern1_haunted_e93_0_0.vlm"},
	{0x5C4E0D53F46F5137ULL, "278881898.ogg"},
	{0x5C4E307045782D81ULL, "follower_barks6153473.cl"},
	{0x5C4E6CED23083D63ULL, "771948212.ogg"},
	{0x5C4E7FB5231EA228ULL, "vxlt_0_lowtown_docks_a3_brw_0_0.vlm"},
	{0x5C4E9C89989FD80FULL, "gen00fl_fenris6106628.cl"},
	{0x5C4EA0999523DC15ULL, "vxw_woundedcoast_1he_0_0.vlm"},
	{0x5C4EF429050EDCC6ULL, "gen00fl_anders6111089.cl"},
	{0x5C4F08DCFB369A2DULL, "6028524_m.fxe"},
	{0x5C4F0B3669EEAD8BULL, "560429531.ogg"},
	{0x5C4F1F45A676F5CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_female.ani"},
	{0x5C4FBE14EF716E02ULL, "vxdr_primevaltomb_2wu_0_0.vlm"},
	{0x5C5078CEECFC14F3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_wf.ani"},
	{0x5C50A39A5FE25209ULL, "6043934_m.fxe"},
	{0x5C511A1CCB42407AULL, "gen00fl_varric6093653.cl"},
	{0x5C5127FC23FE463BULL, "230597454.ogg"},
	{0x5C512ECB3B92175FULL, "w_encounter3_49_0.rml"},
	{0x5C5170F44433A255ULL, "hlac_ht_keep_n_4519_18y.dds"},
	{0x5C518C1903C8EBD0ULL, "6076144_m.fxe"},
	{0x5C51CBFBB15FF709ULL, "vxlt_0_lowtown_eqg_0_1.vlm"},
	{0x5C527C10BEA1870DULL, "vxlt_0_lowtown_docks_a3_d0f_0_0.vlm"},
	{0x5C527E5309270EE8ULL, "mag121_viveka6156507.cl"},
	{0x5C5288E479208148ULL, "vxca_cavern1_haunted_in7_0_0.vlm"},
	{0x5C52B90C6E9FF6EDULL, "hlac_dr_0_deeproads_5842_17x.dds"},
	{0x5C52C0768A4C0DC1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_inquisitionsoldier2_.ani"},
	{0x5C531E3A1821081AULL, "dae000cr_grand_cleric6147643.cl"},
	{0x5C53263516761344ULL, "gen00fl_bethany6101534.cl"},
	{0x5C532A4F45FD1D1FULL, "6078426_m.fxe"},
	{0x5C53310CABEA53DEULL, "6178374_m.fxe"},
	{0x5C53D1515FDBF16BULL, "hlac_w_sundermount_4_0z.dds"},
	{0x5C540ED31F095577ULL, "592015997.ogg"},
	{0x5C556D87B53CE249ULL, "llac_lt_backalley_n_1_3.dds"},
	{0x5C556DB9C637ECFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug_handsup.ani"},
	{0x5C564F5CAA67C30AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2craggyrock.mao"},
	{0x5C56535CAA67CA59ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2craggyrock.mmh"},
	{0x5C565D5CAA67DB47ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2craggyrock.msh"},
	{0x5C566E3A83963BA9ULL, "qun131_party_comment6117279.cl"},
	{0x5C566EFE8CCA7B99ULL, "hlac_lt_0_lowtown_docks_a3_65534_3x.dds"},
	{0x5C56F49F41F01C74ULL, "702622814.ogg"},
	{0x5C57405DD935E090ULL, "vxht_0_hightown_n_an5_0_0.vlm"},
	{0x5C57666F373D02E8ULL, "dae000_bodhan6146800.cl"},
	{0x5C57720BA9C1EDE1ULL, "vxca_cavern1_haunted_ik1_0_0.vlm"},
	{0x5C579231AB9D6327ULL, "zz_dog_debug6134426.cl"},
	{0x5C57A4D69C459D70ULL, "628349020.ogg"},
	{0x5C57B0F162702E34ULL, "hlac_ca_caverns2_5802_46y.dds"},
	{0x5C57D76BE4148781ULL, "hlac_lt_undercityserialkiller_-1_5x.dds"},
	{0x5C57E5494B433953ULL, "6171585_m.fxe"},
	{0x5C5816375762B805ULL, "follower_banter6166452.cl"},
	{0x5C5836F63DDBDAC2ULL, "dae000_sabina6161467.cl"},
	{0x5C584DCB3B986AF7ULL, "w_encounter3_49_0.phy"},
	{0x5C587188750B1069ULL, "hlac_lt_gallowscourtyard_-65535_2x.dds"},
	{0x5C592B51EC22F40BULL, "906932190.ogg"},
	{0x5C599D77AA448765ULL, "art\\levels\\races\\design\\prp_chestc.phy"},
	{0x5C59BA345268D24CULL, "vxw_blightlands_o2_0_0.vlm"},
	{0x5C59CB768A523F63ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_inquisitionsoldier2_.gad"},
	{0x5C59D4D98A82D0E3ULL, "190148518.ogg"},
	{0x5C5B2DA5E1583BA1ULL, "6028872_m.fxe"},
	{0x5C5B5709F6763581ULL, "hlac_ca_caverns2prc_5802_3x.dds"},
	{0x5C5B8C74882836FAULL, "66962551.ogg"},
	{0x5C5BB93DA9E51B41ULL, "llac_lt_0_lowtown_131071_3.dds"},
	{0x5C5C6C6270306CF0ULL, "vxlt_hangedman_n_2fp_0_0.vlm"},
	{0x5C5CC8ACC69501DDULL, "blood_clo.mao"},
	{0x5C5D12768719F495ULL, "vxw_woundedcoast_1sd_0_0.vlm"},
	{0x5C5D3407A5210A66ULL, "zz_approval_debug6003689.cl"},
	{0x5C5DA2D002E96408ULL, "365620325.ogg"},
	{0x5C5E5AEBCC606D76ULL, "6161192_m.fxe"},
	{0x5C5EDC601268D13AULL, "vxlt_gallowstemplar_3ry_0_0.vlm"},
	{0x5C5F661456D8FAE8ULL, "6180429_m.fxe"},
	{0x5C5F9C4A4A5A7113ULL, "6123318_f.fxe"},
	{0x5C5FD59093D7E814ULL, "dae100st_alistair_drunk.stg"},
	{0x5C5FD5CAB59AB442ULL, "809510575.ogg"},
	{0x5C6000D2BB23AA07ULL, "one300pt_murder_crows.ncs"},
	{0x5C605517B096578EULL, "vxlt_0_lowtown_eov_0_1.vlm"},
	{0x5C60784994EC2172ULL, "6169369_m.fxe"},
	{0x5C607CD3B2E1BC86ULL, "vxca_caverns2prc_78w_0_1.vlm"},
	{0x5C609615A6F7282CULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody6_.gad"},
	{0x5C60CBD85D5554E0ULL, "character_rendertodepth_p6.vsh.fxo.dx11"},
	{0x5C60FAC16B1699C8ULL, "vxca_cavern1_haunted_ik4_0_0.vlm"},
	{0x5C616D397B114522ULL, "vxw_bonepitmining_a2_ag_0_0.vlm"},
	{0x5C618AC0F948500CULL, "art\\vfx\\combat\\vfx_r_var_pin_shot_act_c.anb"},
	{0x5C61BDCF8D8E4BFAULL, "6016827_m.fxe"},
	{0x5C6235C09288A138ULL, "gen_im_npc_wep_shd_rdrb.uti"},
	{0x5C62AAEDB3DF40E9ULL, "gen_im_arm_cht_nobd_t3.uti"},
	{0x5C62BD10F54D1212ULL, "759888537.ogg"},
	{0x5C633B2D7066E565ULL, "vxw_woundedcoast_1b7_0_0.vlm"},
	{0x5C635D290B36F56EULL, "mer341_jansen6054937.cl"},
	{0x5C639A6680684FC7ULL, "6169610_m.fxe"},
	{0x5C63DAFB9193F56FULL, "hlac_lt_gallowstemplar_2384_7y.dds"},
	{0x5C63E964970DD0A8ULL, "6108476_m.fxe"},
	{0x5C644BB8435ADB1BULL, "6131215_m.fxe"},
	{0x5C644D56F270D909ULL, "6077989_f.fxe"},
	{0x5C644FCEED0C779DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_wf.gad"},
	{0x5C64EA402DE0E02EULL, "hlac_ht_mansion_n_3365_1x.dds"},
	{0x5C65ABF514B3DF55ULL, "cloth2.vsh.fxo"},
	{0x5C65DCA4FAF18781ULL, "644667363.ogg"},
	{0x5C660E9F8D537D1FULL, "zz_followers_banter6158196.cl"},
	{0x5C67278595507191ULL, "lt_0_lowtown_a2_n_c_51_0_ll.mmh"},
	{0x5C674CB976BB81B2ULL, "726870859.ogg"},
	{0x5C6775FC32BD6532ULL, "dae000_arishok6080440.cl"},
	{0x5C679515A6FD452AULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody6_.ani"},
	{0x5C67DEC4491F53E5ULL, "hlac_dr_primevaltomb_9621_3x.dds"},
	{0x5C67F49D8347928FULL, "6118470_m.fxe"},
	{0x5C682C11445EEACCULL, "vxw_blightlands_p3_0_0.vlm"},
	{0x5C682C1D7D183EB2ULL, "6101002_m.fxe"},
	{0x5C6895BB4BE39684ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardian.mmh"},
	{0x5C6899BB4BE39DD7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardian.mao"},
	{0x5C68ABBB4BE3BC26ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardian.msh"},
	{0x5C68B609FE22DB25ULL, "371909477.ogg"},
	{0x5C68BD437D191A69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_exit.ani"},
	{0x5C68D3CE7AA5C227ULL, "vxw_encounter3_6z_0_0.vlm"},
	{0x5C6909A6F4263D63ULL, "6125640_f.fxe"},
	{0x5C6938E756BAD3C4ULL, "6043669_m.fxe"},
	{0x5C69A757E4A88E79ULL, "6019209_m.fxe"},
	{0x5C6A5078E441DB5FULL, "dae000_bartrand6064042.cl"},
	{0x5C6AC6DFAD9D16E2ULL, "one280_iwan.cnv"},
	{0x5C6ADE3F62268570ULL, "vxlt_hangedman_n_2mq_0_0.vlm"},
	{0x5C6AF4CB485E66BCULL, "6041127_f.fxe"},
	{0x5C6B0FAFB542F621ULL, "vxw_encounter2_3x_0_0.vlm"},
	{0x5C6B39700E507C63ULL, "6076002_m.fxe"},
	{0x5C6B604101FE8691ULL, "gen00fl_fenris6108040.cl"},
	{0x5C6BAB0BA2FEDD50ULL, "qun221_viscount6069914.cl"},
	{0x5C6C71BF69CB2BD8ULL, "lt_0_lowtown_44_0.rml"},
	{0x5C6C81E298FD63E6ULL, "vxca_caverns1_ckq_0_0.vlm"},
	{0x5C6D031D72900B02ULL, "6081746_m.fxe"},
	{0x5C6D199928ADC9F1ULL, "vxht_0_hightown_a2_api_0_0.vlm"},
	{0x5C6D514FF652526FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_stonestand.mao"},
	{0x5C6D5D4FF652660CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_stonestand.mmh"},
	{0x5C6D634FF65270FEULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_stonestand.msh"},
	{0x5C6DABD48D83A88DULL, "vfx_distortionmask_p1.vsh.fxo"},
	{0x5C6DAEFA22D071DCULL, "mag100_cullen6140988.cl"},
	{0x5C6DB4A72344F59CULL, "llac_ht_0_hightown_n_-65534_0.dds"},
	{0x5C6DBE506364B536ULL, "vxw_sundermount_23l_0_0.vlm"},
	{0x5C6DC425FF2B62C9ULL, "58236744.ogg"},
	{0x5C6E8F351161C4A3ULL, "follower_banter6141619.cl"},
	{0x5C6E9D561D6D5426ULL, "vxdr_0_deeproads_9if_0_0.vlm"},
	{0x5C6EA8DFADA0CC1AULL, "one280_iwan.bnk"},
	{0x5C6F046C5544FE3FULL, "6053555_m.fxe"},
	{0x5C6FD8437D1F66EBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_exit.gad"},
	{0x5C700B73C9222F63ULL, "vxht_keep_n_7ai_0_0.vlm"},
	{0x5C704C94846BF2B0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so_galow.phy"},
	{0x5C705CBF68BEC302ULL, "vxht_playershouse_n_1su_0_0.vlm"},
	{0x5C705CE552F987CBULL, "qun221_petrice6077942.cl"},
	{0x5C705FF34580E203ULL, "hlac_lt_0_lowtown_a2_-65535_0y.dds"},
	{0x5C70BD383B036F8EULL, "gen00fl_bethany6101231.cl"},
	{0x5C70BDA3AEEA1A23ULL, "hlac_ca_caverns1_24425_12x.dds"},
	{0x5C71AD0A625CFDE5ULL, "vxw_woundedcoast_1m6_0_0.vlm"},
	{0x5C71ED991FE97CD1ULL, "6065273_m.fxe"},
	{0x5C7294C0A5C8F2F7ULL, "vxlt_0_lowtown_a2_hf4_0_0.vlm"},
	{0x5C729762CD6F3784ULL, "765308548.ogg"},
	{0x5C72A5CEAD440FB4ULL, "llac_ht_mansion2_818_39.dds"},
	{0x5C734B519FCDB87CULL, "hlac_ht_0_hightown_a3_131074_0z.dds"},
	{0x5C734CFDADC4B39AULL, "vxw_woundedcoast_a4_0_0.vlm"},
	{0x5C73D67724A15B32ULL, "6075229_f.fxe"},
	{0x5C741DC7E04CCFBAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad.msh"},
	{0x5C742BC7E04CE7F3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad.mao"},
	{0x5C742FC7E04CEE00ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad.mmh"},
	{0x5C748D8C602BF4C8ULL, "lt_0_lowtown_53_0.rml"},
	{0x5C74F9E7305A2BE9ULL, "82075401.ogg"},
	{0x5C750C1797146D91ULL, "vrc321_varric6097152.cl"},
	{0x5C756CE257564811ULL, "6163720_m.fxe"},
	{0x5C75B4BA581015E6ULL, "vxca_cavern1_haunted_hu1_0_0.vlm"},
	{0x5C75BFB0D63EB5B4ULL, "6043456_m.fxe"},
	{0x5C75D5630DA4F99CULL, "hlac_lt_0_lowtown_-65537_6x.dds"},
	{0x5C75F886B22F9FEFULL, "vxlt_gallowstemplar_f_5ja_0_0.vlm"},
	{0x5C7636DD709B42E9ULL, "mcr310_party_banter6182943.cl"},
	{0x5C76540309BEF920ULL, "hlac_dr_0_deeproads_5842_7x.dds"},
	{0x5C768113D1295F45ULL, "vxca_caverns2prc_9ls_0_0.vlm"},
	{0x5C769DEE3655034CULL, "vxw_blightlands_u0_0_0.vlm"},
	{0x5C76D5437D258050ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_exit.evt"},
	{0x5C76D975C65F2E90ULL, "871991427.ogg"},
	{0x5C76E3FBB1CA78ADULL, "hlac_ca_caverns1_24424_15z.dds"},
	{0x5C773A89EBFF474AULL, "hm_one_crow_2.mor"},
	{0x5C77484402A59B81ULL, "follower_banter6151296.cl"},
	{0x5C7752CEB1D295FBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_statuenaked.phy"},
	{0x5C7782AEE2475776ULL, "6030401_m.fxe"},
	{0x5C779ED8E6D7C18CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highwardrobe_n.dds"},
	{0x5C7821C274CCDAE2ULL, "gen00fl_anders6125833.cl"},
	{0x5C78F076739C72DEULL, "character_character_p3.psh.fxo.dx11"},
	{0x5C7985493E1B50AFULL, "vxht_playershouse_n_1g0_0_0.vlm"},
	{0x5C798E137ECE4D69ULL, "6099703_f.fxe"},
	{0x5C79E9FD40AE63F2ULL, "hlac_lt_0_lowtown_a2_n_-65535_5x.dds"},
	{0x5C7A976B19B31753ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ent.ani"},
	{0x5C7AB0EDF67EE82EULL, "6066338_m.fxe"},
	{0x5C7ACB01BED76ED3ULL, "vxw_sundermount_2j6_0_0.vlm"},
	{0x5C7B0DB4F5EEA73CULL, "t1_skn_e01.tnt"},
	{0x5C7B5A1C1D87442EULL, "llac_lt_0_lowtown_a2_-196608_0.dds"},
	{0x5C7B8C4B25E59311ULL, "vxca_cave_2dv_0_0.vlm"},
	{0x5C7B8F4919587E53ULL, "183263369.ogg"},
	{0x5C7BAC7D3E35D74DULL, "6208674_m.fxe"},
	{0x5C7BAD0F035A5E82ULL, "6095488_m.fxe"},
	{0x5C7BBF6972F0D16BULL, "hlac_lt_darktown_-1_3x.dds"},
	{0x5C7BEB44C57ABF95ULL, "6054897_m.fxe"},
	{0x5C7C302D555ACDB6ULL, "vxw_sundermount_24m_0_0.vlm"},
	{0x5C7C5BD93D2ED50DULL, "dae000_flemeth.stg"},
	{0x5C7C7608A4F80887ULL, "6080289_m.fxe"},
	{0x5C7C857B9F677C92ULL, "6163334_m.fxe"},
	{0x5C7D59D9C93BB759ULL, "dae300_alistair_king6134907.cl"},
	{0x5C7D5C0378D85272ULL, "6163169_m.fxe"},
	{0x5C7D67E188F62AE4ULL, "vxht_mansion2_3gf_0_1.vlm"},
	{0x5C7D8E9B21DBEABAULL, "135933774.ogg"},
	{0x5C7DD577B5983BDDULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masc.phy"},
	{0x5C7E53F304A1A3A1ULL, "vxlt_0_lowtown_a2_n_c_d9b_0_0.vlm"},
	{0x5C7E9F588C2F0C5BULL, "hlac_lt_0_lowtown_a3_n_c_65535_0z.dds"},
	{0x5C7ECE9C5AB4DB82ULL, "vxht_playershouse_n_1vv_0_0.vlm"},
	{0x5C7F33A8DEB200E3ULL, "gen_im_trs_dal_hvy_cht_r0.uti"},
	{0x5C7F5FA07BD6D547ULL, "vxw_bonepitdestroyed_a3_243_0_0.vlm"},
	{0x5C7FC2405452FDFFULL, "vxw_bonepitmining_a2_1o3_0_0.vlm"},
	{0x5C7FC6862AAEDEDBULL, "vxlt_0_lowtown_n_eoe_0_0.vlm"},
	{0x5C7FF72D35869913ULL, "hlac_lt_0_lowtown_docks_0_2y.dds"},
	{0x5C7FF9D071D2F87EULL, "6025115_m.fxe"},
	{0x5C801EE754531665ULL, "vxw_woundedcoast_1h5_0_0.vlm"},
	{0x5C804EB32A7FBD49ULL, "6135161_m.fxe"},
	{0x5C80667C58D6271FULL, "ht_0_hightown_25_0.anb"},
	{0x5C806A0E3619A779ULL, "hlac_lt_0_lowtown_docks_65536_0x.dds"},
	{0x5C80B60598240C48ULL, "qun221_petrice6083254.cl"},
	{0x5C80C555968795A9ULL, "6034359_f.fxe"},
	{0x5C80CD1E7AFA01A8ULL, "6106676_m.fxe"},
	{0x5C80F274CC710389ULL, "vxca_caverns1_i70_0_0.vlm"},
	{0x5C81069D97BF0B77ULL, "vxlt_0_lowtown_a2_hc7_0_0.vlm"},
	{0x5C811BFA92F82399ULL, "zz_lgt_2_debug6140445.cl"},
	{0x5C816C91CA62E3AFULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic_3.mao"},
	{0x5C8171AA83DCDEC8ULL, "6072157_m.fxe"},
	{0x5C81862688F5BF95ULL, "331653192.ogg"},
	{0x5C81B72550FB2C64ULL, "gen00fl_carver6102555.cl"},
	{0x5C8210C5D0E53966ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_escape_qunari.evt"},
	{0x5C822A9A5DD05C1DULL, "vxht_mansion2b_1z3_0_0.vlm"},
	{0x5C8261775E0A3E4EULL, "hlac_dr_primevalentrance_1725_0y.dds"},
	{0x5C82B2B5F21B959FULL, "6166376_m.fxe"},
	{0x5C83BF1F13B50C57ULL, "205074599.ogg"},
	{0x5C83DCD1260A60A0ULL, "vxht_brothel_n_625_0_0.vlm"},
	{0x5C8447EBA5A1E59EULL, "vxw_encounter2_6d_0_0.vlm"},
	{0x5C84E74E66ED8505ULL, "hlac_w_bonepitmining_a2_131073_0z.dds"},
	{0x5C85621A59341152ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_1y.dds"},
	{0x5C85A013A78C5EB4ULL, "hlac_lt_blackemporium_1055_20x.dds"},
	{0x5C85C428B65335E6ULL, "hm_raider_1.mor"},
	{0x5C862733154D7E63ULL, "vxlt_0_lowtown_docks_a0n_0_0.vlm"},
	{0x5C8752A2B0EBEEABULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.dura.ani"},
	{0x5C878F6B19BD938AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ent.evt"},
	{0x5C87DA1DB27C1EC4ULL, "891079666.ogg"},
	{0x5C87F50D2ED26EBEULL, "lgt331ip_at_darktown.utp"},
	{0x5C88179C50CB8167ULL, "mag291_torpor6027004.cl"},
	{0x5C8894DF5A4ED1C5ULL, "vxdr_primevaltomb_84i_0_0.vlm"},
	{0x5C889DC29C07079EULL, "6095660_m.fxe"},
	{0x5C88DA393FB28587ULL, "vxlt_0_lowtown_a2_n_c_fg0_0_0.vlm"},
	{0x5C8927C5D0EB7EB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_escape_qunari.gad"},
	{0x5C893CDEB0CD8753ULL, "vxw_sundermount_2o9_0_0.vlm"},
	{0x5C897B2067E6815BULL, "character_character_p268.psh.fxo.dx11.assembly"},
	{0x5C8995ECFE4382A0ULL, "hlac_w_encounter2_camp_-131067_0y.dds"},
	{0x5C8A1E0E1090DD4DULL, "vxlt_hangedman_wc_0_0.vlm"},
	{0x5C8A9750F94F8182ULL, "6153823_m.fxe"},
	{0x5C8AA20A4750E636ULL, "vxw_sundermount_29n_0_0.vlm"},
	{0x5C8ACC220ECC78C6ULL, "6060228_m.fxe"},
	{0x5C8B2A5D56089F2CULL, "6146922_m.fxe"},
	{0x5C8B811001598526ULL, "vxdr_0_deeproads_9gh_0_0.vlm"},
	{0x5C8BA9BD74EC1029ULL, "vxht_mansion_n_6oh_0_0.vlm"},
	{0x5C8C770F357D5EACULL, "dae000_flemeth6000914.cl"},
	{0x5C8CCF2070D68150ULL, "6150271_m.fxe"},
	{0x5C8DA08F8BD4B3FDULL, "hlac_lt_undercityencounter_-1_33y.dds"},
	{0x5C8DE5F52AD703E5ULL, "6128057_m.fxe"},
	{0x5C8DF82BADB2D1ADULL, "leave_03_l3.mao"},
	{0x5C8E37C514C8301CULL, "vxht_0_hightown_a7y_0_0.vlm"},
	{0x5C8E6E6B19C37AFDULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ent.gad"},
	{0x5C8E90C446492EE5ULL, "vxw_woundedcoast_1s4_0_0.vlm"},
	{0x5C8E9DFCD097DE2BULL, "885772820.ogg"},
	{0x5C8EB04D3069D46BULL, "one120_athenril6088305.cl"},
	{0x5C8EC635718EB424ULL, "6144472_m.fxe"},
	{0x5C8ECD6515DED155ULL, "6179947_m.fxe"},
	{0x5C8ED677B5A6B0E2ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masc.mmh"},
	{0x5C8ED877B5A6B398ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masc.msh"},
	{0x5C8F0CADA80D3A05ULL, "da2_prc_one_fr-fr.cre"},
	{0x5C8F74D807CD890AULL, "122177169.ogg"},
	{0x5C8F787A89B523F7ULL, "vxlt_0_lowtown_a2_hh6_0_0.vlm"},
	{0x5C8FA46B8907A70DULL, "647760155.ogg"},
	{0x5C8FA891F1CC4863ULL, "6139444_m.fxe"},
	{0x5C9003100E2E6FF6ULL, "vxlt_warehouse_n_15p_0_0.vlm"},
	{0x5C9028C5D0F19F5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_escape_qunari.ani"},
	{0x5C90341B68B30260ULL, "hlac_w_encounter2_262147_0z.dds"},
	{0x5C908F6C0EAE92E9ULL, "6114230_m.fxe"},
	{0x5C90C5E2C8E6CD4EULL, "vxw_bonepitmining_a2_1k2_0_0.vlm"},
	{0x5C90C72EE1F0FB41ULL, "6165743_f.fxe"},
	{0x5C91764982D96D8CULL, "559414046.ogg"},
	{0x5C91EF0629F95B69ULL, "6080165_f.fxe"},
	{0x5C921E063ABE6132ULL, "hlac_lt_0_lowtown_a2_n_-1_3y.dds"},
	{0x5C9232EDEFE1E8A2ULL, "6170604_m.fxe"},
	{0x5C9247E4E5248729ULL, "hlac_ht_keep_n_3732_9y.dds"},
	{0x5C925C906D5A5130ULL, "vxw_sundermount_315_0_0.vlm"},
	{0x5C925DB28F398CBFULL, "6143500_m.fxe"},
	{0x5C92B151F318E357ULL, "gen00cs_cs_end_music_int_0.ncs"},
	{0x5C92B9C89797FE1EULL, "vxw_encounter2_1e_0_0.vlm"},
	{0x5C92E59FB7380261ULL, "vxlt_0_lowtown_a2_byu_0_0.vlm"},
	{0x5C930FE256B0DA44ULL, "6178348_f.fxe"},
	{0x5C9311B089703398ULL, "529814955.ogg"},
	{0x5C936B84E6701A7AULL, "plt_lgt200pt_night.nss"},
	{0x5C939D087386343AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasec_d.dds"},
	{0x5C93AD76EACD15CBULL, "qun131_warrens_thug6019152.cl"},
	{0x5C9488759E00AEFEULL, "dae100_gamlen_letter6146983.cl"},
	{0x5C94B337D8815F15ULL, "mer110_jansen6032828.cl"},
	{0x5C94D65E15096B02ULL, "6089953_m.fxe"},
	{0x5C951758253D4B06ULL, "6056834_m.fxe"},
	{0x5C953AA2B0F80322ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.dura.evt"},
	{0x5C95F9C4E2E9547BULL, "mer120_hubert6070525.cl"},
	{0x5C96AE001460AD51ULL, "655259643.ogg"},
	{0x5C96F0CFD47461EEULL, "vxht_0_hightown_a2_b01_0_0.vlm"},
	{0x5C96F4D09E1308C1ULL, "vxlt_hangedman_n_54_0_0.vlm"},
	{0x5C9746ADEA76A4E5ULL, "317563396.ogg"},
	{0x5C9762729F09218BULL, "130411355.ogg"},
	{0x5C97A7B044A8D82FULL, "vxlt_0_lowtown_a2_n_c_eqa_0_0.vlm"},
	{0x5C97AEBBA2C39FD3ULL, "vxw_sundermount_2p8_0_0.vlm"},
	{0x5C97D7D21A64C4FBULL, "386860318.ogg"},
	{0x5C982B033D6E1AA8ULL, "537016916.ogg"},
	{0x5C984FB028FC6E87ULL, "vxw_woundedcoast_59_0_0.vlm"},
	{0x5C9852BF69F03190ULL, "lt_0_lowtown_44_0.anb"},
	{0x5C98A7CCAF8E0D82ULL, "pro000cr_soldier_leader.utc"},
	{0x5C9906F9196674DDULL, "6177354_m.fxe"},
	{0x5C99E358F5A0300EULL, "hlac_lt_0_lowtown_a3_n_c_-131073_2y.dds"},
	{0x5C9A27A4AD81CB22ULL, "796528575.ogg"},
	{0x5C9A947B61C2F7E2ULL, "vxht_keep_n_58e_0_0.vlm"},
	{0x5C9AA26C591138E3ULL, "and100_anders6095960.cl"},
	{0x5C9ACB2A7463709DULL, "qun131_taskmaster6019211.cl"},
	{0x5C9B19EF7D53742AULL, "160785678.ogg"},
	{0x5C9BB8002C5807C2ULL, "vxca_cave_23h_0_0.vlm"},
	{0x5C9BCCD985A138C5ULL, "s_stand_04.dds"},
	{0x5C9BE1184D44623AULL, "w_blightlands_247_0.msh"},
	{0x5C9BEF184D447A73ULL, "w_blightlands_247_0.mao"},
	{0x5C9BF0C263013D35ULL, "da2terrain_terrain_p1.dsh.fxo.dx11"},
	{0x5C9BF3184D448080ULL, "w_blightlands_247_0.mmh"},
	{0x5C9C39A2B0FE2035ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.dura.gad"},
	{0x5C9C637E25D7D43CULL, "1044387728.ogg"},
	{0x5C9C6E55788CAE48ULL, "vxlt_0_lowtown_a2_byp_0_0.vlm"},
	{0x5C9CA2F9894BBF50ULL, "gen_im_trs_gen_junk2_01.uti"},
	{0x5C9CC09EE22B2610ULL, "gen00fl_anders6125978.cl"},
	{0x5C9CCFF2E0680971ULL, "dae000_gamlen6048435.cl"},
	{0x5C9D2485BE65A10FULL, "vxw_bonepitmining_a2_1zx_0_0.vlm"},
	{0x5C9D38126384CCA4ULL, "6141772_m.fxe"},
	{0x5C9EA29483A6FD1AULL, "vxw_woundedcoast_p7_0_0.vlm"},
	{0x5C9ED423F09C3403ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow2win.mao"},
	{0x5C9ED823F09C3B50ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow2win.mmh"},
	{0x5C9EE0FBACA107DEULL, "vxlt_0_lowtown_a2_gn2_0_0.vlm"},
	{0x5C9EE623F09C536AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow2win.msh"},
	{0x5C9F6F600EF41265ULL, "hlac_lt_0_lowtown_docks_n_131070_2z.dds"},
	{0x5C9F95FA5DEA6CB4ULL, "vxlt_0_lowtown_a2_n_cif_0_0.vlm"},
	{0x5CA00C21EA037295ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_raillow.msh"},
	{0x5CA00E21EA03760BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_raillow.mmh"},
	{0x5CA01A21EA038AA8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_raillow.mao"},
	{0x5CA022CB3BD5D74FULL, "w_encounter3_49_0.mao"},
	{0x5CA02ECB3BD5EBECULL, "w_encounter3_49_0.mmh"},
	{0x5CA032FA9612B562ULL, "6146847_m.fxe"},
	{0x5CA034CB3BD5F5DEULL, "w_encounter3_49_0.msh"},
	{0x5CA07C67B0ED169DULL, "6174157_m.fxe"},
	{0x5CA0D0438EEFA6D8ULL, "6101703_m.fxe"},
	{0x5CA11F3537A0ECFAULL, "833807276.ogg"},
	{0x5CA17F01B3D24B1EULL, "vxw_blightlands_fk_0_0.vlm"},
	{0x5CA19A39C55B36DCULL, "hlac_lt_0_lowtown_docks_a3_n_-65536_0z.dds"},
	{0x5CA1A02EB2B3DABAULL, "375113855.ogg"},
	{0x5CA2245A311D28A5ULL, "934094354.ogg"},
	{0x5CA230CD3CFA3E9CULL, "art\\characters\\weapons\\staff\\textures\\w_stf_004a_0d.dds"},
	{0x5CA2DD6A9D114797ULL, "539938374.ogg"},
	{0x5CA339CEB1F7A59AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_statuenaked.msh"},
	{0x5CA347CEB1F7BDD3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_statuenaked.mao"},
	{0x5CA34BCEB1F7C4E0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_statuenaked.mmh"},
	{0x5CA432A767281683ULL, "vxlt_0_lowtown_h9m_0_0.vlm"},
	{0x5CA463F3A93B0FF6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabslarge.msh"},
	{0x5CA46DF3A93B20F4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabslarge.mmh"},
	{0x5CA471F3A93B2707ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabslarge.mao"},
	{0x5CA492459056F459ULL, "vxlt_gallowstemplar_n_3n5_0_0.vlm"},
	{0x5CA5200BF51AA96EULL, "al_fenris_talents.gda"},
	{0x5CA54204607B9C14ULL, "533675597.ogg"},
	{0x5CA595D0997FCFEFULL, "mcr370_varric6086849.cl"},
	{0x5CA5D02A88A84F4FULL, "6053887_m.fxe"},
	{0x5CA5F4BC7BA9E38AULL, "101838823.ogg"},
	{0x5CA66FCA4457F3B2ULL, "vxht_mansion2_2an_0_0.vlm"},
	{0x5CA69B09F324D6D0ULL, "hlac_lt_0_lowtown_docks_n_-2_3y.dds"},
	{0x5CA6C18D1AF28707ULL, "vxw_woundedcoast_28_0_0.vlm"},
	{0x5CA6D402EBB5031BULL, "hlac_lt_0_lowtown_docks_196608_0y.dds"},
	{0x5CA6D52CB7206B06ULL, "fen210_fenris6108717.cl"},
	{0x5CA72370C9A1A431ULL, "mag311_emile_de_launcet6043808.cl"},
	{0x5CA77C22E42319F7ULL, "hlac_ht_0_hightown_n_-65536_1x.dds"},
	{0x5CA7A87C4D3BC1A7ULL, "hlac_lt_0_lowtown_65533_1y.dds"},
	{0x5CA8059721E00337ULL, "706446084.ogg"},
	{0x5CA8E7465486E260ULL, "mag100_thrask6031976.cl"},
	{0x5CA99BD1AED6B816ULL, "static_staticreliefmap_p1.gsh.fxo.dx11.assembly"},
	{0x5CA9FD72C30640D4ULL, "6116228_m.fxe"},
	{0x5CAA34F571989D61ULL, "art\\characters\\creatures\\stonegolem\\animation\\statue.gad"},
	{0x5CAAEAF3C1C3ED66ULL, "character_character_p119.psh.fxo.dx11"},
	{0x5CABBB6223D0B76AULL, "253837402.ogg"},
	{0x5CAC3C5DB9A05E92ULL, "206300674.ogg"},
	{0x5CACB259DE12A843ULL, "vxca_caverns1_mines_eae_0_0.vlm"},
	{0x5CAD281307C442D4ULL, "vxlt_warehouse_n_hv_0_0.vlm"},
	{0x5CAD7698026D788DULL, "vxlt_undercityencounter_1dx_0_0.vlm"},
	{0x5CAE228C0FF5A95DULL, "6033628_f.fxe"},
	{0x5CAE3C1607C3D22DULL, "vxlt_gallowscourtyard_6up_0_0.vlm"},
	{0x5CAE5D2A50582363ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_c.impa.ani"},
	{0x5CAE70B7BBC6E87FULL, "z_dust2.dds"},
	{0x5CAE8A99CCCAD411ULL, "6070341_m.fxe"},
	{0x5CAEB3011E26119CULL, "749704609.ogg"},
	{0x5CAF59F0EF81EC77ULL, "vxht_0_hightown_a2_b22_0_0.vlm"},
	{0x5CAF89AD7D33EDE7ULL, "vxw_sundermount_385_0_0.vlm"},
	{0x5CAFF497D8F3E22EULL, "vxht_0_hightown_a3_caf_0_0.vlm"},
	{0x5CAFFEDDCF9027B0ULL, "fen110_fenris6117932.cl"},
	{0x5CB02525F55D1851ULL, "184792125.ogg"},
	{0x5CB0284E96641F0AULL, "vxw_encounter2_camp_oc_0_0.vlm"},
	{0x5CB035F5719D0A0FULL, "art\\characters\\creatures\\stonegolem\\animation\\statue.ani"},
	{0x5CB0B721E1965DC5ULL, "llac_dr_primevaltomb_659_1.dds"},
	{0x5CB1922941D4F8B7ULL, "6082162_m.fxe"},
	{0x5CB207F39660C056ULL, "follower_banter6172988.cl"},
	{0x5CB278F26B565A8EULL, "mag131cr_bloodmage3.utc"},
	{0x5CB286E622C093B5ULL, "llac_lt_backalley_g_65536_4.dds"},
	{0x5CB2EA11827CB6D0ULL, "874249138.ogg"},
	{0x5CB40415CC1A89C2ULL, "lowtown_male_lowlife1_sta6164963.cl"},
	{0x5CB44E91C7A877AEULL, "6059145_m.fxe"},
	{0x5CB48204E0AD3B36ULL, "vxw_encounter2_59_0_0.vlm"},
	{0x5CB4CAC4A07E1FDBULL, "aip_assassincommander.gda"},
	{0x5CB56AD28D0D27DCULL, "438178373.ogg"},
	{0x5CB595C8E8580CD1ULL, "90535278.ogg"},
	{0x5CB5AA557D3F2084ULL, "qun211cr_crazy_woman6088451.cl"},
	{0x5CB642C7B9F2F2C5ULL, "6187949_f.fxe"},
	{0x5CB6541ECC4B6657ULL, "vxlt_warehouse_10i_0_0.vlm"},
	{0x5CB6C7FAF0AEDAACULL, "vxht_mansion2_2ix_0_0.vlm"},
	{0x5CB6EB02FCC3C27CULL, "vxlt_0_lowtown_bza_0_0.vlm"},
	{0x5CB79DC3D77B5537ULL, "mag170_thrask6016940.cl"},
	{0x5CB802019134DBAAULL, "ch2armtnt_s2.psh.fxo"},
	{0x5CB807CB11A6FCEAULL, "vxca_cave_18v_0_0.vlm"},
	{0x5CB810D5A1D64EA1ULL, "isa_mod_item_acq.ncs"},
	{0x5CB849304DB98F3CULL, "6176095_m.fxe"},
	{0x5CB8D813098B9772ULL, "vxlt_0_lowtown_eny_0_1.vlm"},
	{0x5CB904AC414DD494ULL, "vxdr_0_deeproads_cph_0_1.vlm"},
	{0x5CB9334853C380A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_shrug_small.ani"},
	{0x5CB9545CAABC4946ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2craggyrock.phy"},
	{0x5CB971526CF487FAULL, "6069911_m.fxe"},
	{0x5CBB157913762AE6ULL, "hlac_ca_caverns1_mines_24424_23y.dds"},
	{0x5CBB6607843EB3A4ULL, "vrc321ip_spooky6_obj_9.utp"},
	{0x5CBB752A5062D65AULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_c.impa.evt"},
	{0x5CBB99EFF9BA5B54ULL, "vxlt_warehouse_n_ow_0_0.vlm"},
	{0x5CBBC15A30946E6BULL, "hlac_dr_primevaltomb_661_1y.dds"},
	{0x5CBBC3B94645504DULL, "zz_followers_banter6157313.cl"},
	{0x5CBC113C542AB714ULL, "6051402_m.fxe"},
	{0x5CBC1B799EC916CEULL, "vxw_bonepitmining_a2_1z5_0_0.vlm"},
	{0x5CBC2689F69BAC3AULL, "llac_w_encounter2_131074_0.dds"},
	{0x5CBCB3932043DC66ULL, "vxlt_undercitydungeon_1j4_0_0.vlm"},
	{0x5CBCE35FB764C703ULL, "6067291_m.fxe"},
	{0x5CBD6CD1B9BE61A9ULL, "gen00fl_aveline6139672.cl"},
	{0x5CBE36914349CBA4ULL, "1009016519.ogg"},
	{0x5CBE62BB97BE7C1EULL, "vxw_blightlands_lm_0_0.vlm"},
	{0x5CBED9CA1A508BDFULL, "309430323.ogg"},
	{0x5CBF22179FE45F62ULL, "vxlt_0_lowtown_docks_ctm_0_0.vlm"},
	{0x5CBFA03967A25F32ULL, "plt_rdr200pt_insider.plo"},
	{0x5CC04E4853C9CD2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_shrug_small.gad"},
	{0x5CC09AFCBA9FC4AAULL, "mcr370_negotiation6080307.cl"},
	{0x5CC1325E04EE8D88ULL, "6051424_m.fxe"},
	{0x5CC1545F504F7BBAULL, "655360940.ogg"},
	{0x5CC2742A5068F34DULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_c.impa.gad"},
	{0x5CC311F9FDFBD024ULL, "481155149.ogg"},
	{0x5CC3389ADA2A8E59ULL, "hlac_w_bonepitdestroyed_a3_0_5z.dds"},
	{0x5CC35384284424B2ULL, "vxht_mansion2_2wh_0_0.vlm"},
	{0x5CC43381C72531B7ULL, "251260542.ogg"},
	{0x5CC4354E52A4A80EULL, "993893256.ogg"},
	{0x5CC43CEBAF2D5BE6ULL, "vxw_encounter2_iu_0_0.vlm"},
	{0x5CC440AEA69B3099ULL, "6125821_m.fxe"},
	{0x5CC4BA9B935A9CD3ULL, "vrc321_party_barks6095762.cl"},
	{0x5CC4C6D569283639ULL, "hlac_lt_0_lowtown_docks_0_3x.dds"},
	{0x5CC552D224C4CB43ULL, "vxdr_dungeon_11w_0_0.vlm"},
	{0x5CC55F5F527AE3A5ULL, "mag131tr_eau_blades.ncs"},
	{0x5CC59AEC83C7C7B1ULL, "vxw_bonepit_102_0_0.vlm"},
	{0x5CC5D57D2B1198E6ULL, "1048188758.ogg"},
	{0x5CC64825A3197924ULL, "6043914_f.fxe"},
	{0x5CC64B4853CE3390ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_shrug_small.evt"},
	{0x5CC64CBB86CF6425ULL, "dae000_loyal_mage6165219.cl"},
	{0x5CC689BD4AABDBC7ULL, "6118683_m.fxe"},
	{0x5CC6C70E30C31763ULL, "6126718_m.fxe"},
	{0x5CC7630364C73D2CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml.dura.ani"},
	{0x5CC829FD02E790C6ULL, "gen_im_trs_gem_opal_02.uti"},
	{0x5CC8712DA8AAB544ULL, "lt_0_lowtown_docks_a3_n_244_0.mmh"},
	{0x5CC8752DA8AABB97ULL, "lt_0_lowtown_docks_a3_n_244_0.mao"},
	{0x5CC87A4A6364A8A1ULL, "mcr370_templar1.bnk"},
	{0x5CC8872DA8AADAE6ULL, "lt_0_lowtown_docks_a3_n_244_0.msh"},
	{0x5CC889F3EBC3C2ABULL, "isa311_castillon6092079.cl"},
	{0x5CC8C2996D3BDAD0ULL, "vxlt_0_lowtown_a2_n_fga_0_0.vlm"},
	{0x5CC8D91EEB3C6B5DULL, "957180542.ogg"},
	{0x5CC931077C76885EULL, "mcr371_angry_mage_m6087030.cl"},
	{0x5CC99D67BDA5A729ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb05a.phy"},
	{0x5CC9A1386DFFE8B4ULL, "hlac_ca_caverns1_mines_24426_2z.dds"},
	{0x5CC9AEA4618E297EULL, "war_buildup.dds"},
	{0x5CC9B95FDAF04458ULL, "6112901_m.fxe"},
	{0x5CCA0BCCEBB073D4ULL, "vxlt_warehouse_n_rt_0_0.vlm"},
	{0x5CCA3EAA165F33EDULL, "w_encounter3_206_0.mmh"},
	{0x5CCA40D50E41ED9BULL, "vxca_caverns2_83f_0_2.vlm"},
	{0x5CCA42AA165F3A3EULL, "w_encounter3_206_0.mao"},
	{0x5CCA50AA165F5273ULL, "w_encounter3_206_0.msh"},
	{0x5CCAE10000E67F2DULL, "324463379.ogg"},
	{0x5CCB0C606192D307ULL, "qun211_party_comment6072166.cl"},
	{0x5CCB6CC08424BA55ULL, "vxlt_gallowscourtyard_n_6zv_0_0.vlm"},
	{0x5CCB73946A0C7959ULL, "942804503.ogg"},
	{0x5CCBA8A6B464FC6FULL, "hlac_lt_gallowstemplar_2394_10z.dds"},
	{0x5CCBBC75C91BCEDDULL, "vxlt_warehouse_15d_0_0.vlm"},
	{0x5CCC71B672105B61ULL, "zz_flemeth_land.cut"},
	{0x5CCC9C4A6368CB11ULL, "mcr370_templar1.cnv"},
	{0x5CCCA21E053D5056ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud3high.msh"},
	{0x5CCCAC1E053D6154ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud3high.mmh"},
	{0x5CCCB01E053D6867ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud3high.mao"},
	{0x5CCCD49889B4949EULL, "vxw_blightlands_qn_0_0.vlm"},
	{0x5CCD2A67D0D6D099ULL, "vxht_0_hightown_a2_c6e_0_0.vlm"},
	{0x5CCD72F0E5227566ULL, "644372336.ogg"},
	{0x5CCDEDD00B8DB6C8ULL, "422705374.ogg"},
	{0x5CCDF96E87A2E6C9ULL, "llac_ca_caverns2_5802_33.dds"},
	{0x5CCE700364CD726EULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml.dura.gad"},
	{0x5CCEB2B522C7ED27ULL, "vxht_keep_n_6jm_0_0.vlm"},
	{0x5CCEBE2ACC7EB37AULL, "gen_im_arm_hlm_hvy_act1_02.uti"},
	{0x5CCF19B817FCD019ULL, "ts_dwarven_wpn_props.gda"},
	{0x5CCF2875DA384D8CULL, "art\\vfx\\environment\\vfx_fire_torch_p.dura.ani"},
	{0x5CCF6D507E99541BULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.dura.ani"},
	{0x5CD08D8B26B1725BULL, "546189912.ogg"},
	{0x5CD0A5765878E6D8ULL, "vxht_0_hightown_n_c6g_0_0.vlm"},
	{0x5CD0BF404D662DB0ULL, "429327324.ogg"},
	{0x5CD0CEA7AABC7840ULL, "hlac_ht_mansion2_818_17z.dds"},
	{0x5CD0D5716390361CULL, "qun221_petrice6088202.cl"},
	{0x5CD0E64F2ED5C8D3ULL, "6115485_f.fxe"},
	{0x5CD11B453D3071C2ULL, "qun211ip_barrel.utp"},
	{0x5CD145997E9A82CAULL, "vxca_caverns2_7xs_0_0.vlm"},
	{0x5CD1620E663BB097ULL, "vxlt_gallowstemplar_f_3t1_0_0.vlm"},
	{0x5CD1A95DAF97732EULL, "vxlt_gallowstemplar_n_5j4_0_0.vlm"},
	{0x5CD1BB9D6F6A4E5EULL, "336968533.ogg"},
	{0x5CD1C5611A3A3D32ULL, "vxht_mansion2_2pi_0_0.vlm"},
	{0x5CD21BCAD2EA80C3ULL, "hlac_w_bonepitdestroyed_a3_393217_0z.dds"},
	{0x5CD229E84C5409C4ULL, "168408379.ogg"},
	{0x5CD349A1610DE548ULL, "vxw_bonepit_4x_0_0.vlm"},
	{0x5CD45F61ED8B6F56ULL, "vxca_cavern1_haunted_gka_0_1.vlm"},
	{0x5CD4B0F6BB85B3B4ULL, "w_sundermount_75_0.phy"},
	{0x5CD4B31E54F4F028ULL, "gen_im_arm_boo_lgt_act1_03.uti"},
	{0x5CD4D2AB54F489E2ULL, "hlac_lt_undercityencounter_-1_10y.dds"},
	{0x5CD512AD28BAF687ULL, "dae011_gossips6143019.cl"},
	{0x5CD538EB22B92FE3ULL, "6123418_m.fxe"},
	{0x5CD54621F895D42FULL, "6116102_m.fxe"},
	{0x5CD5680B407846B0ULL, "dae000_orana6149127.cl"},
	{0x5CD56B0364D3884DULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml.dura.evt"},
	{0x5CD5DFAF7AD9D9B0ULL, "94620908.ogg"},
	{0x5CD63575DA3E82CEULL, "art\\vfx\\environment\\vfx_fire_torch_p.dura.gad"},
	{0x5CD64B24E3CD80E4ULL, "368471504.ogg"},
	{0x5CD64BE60345B6E8ULL, "vxlt_darktown_4s3_0_0.vlm"},
	{0x5CD6505F52893DF8ULL, "mag131tr_eau_blades.utt"},
	{0x5CD67A377A59466EULL, "hlac_lt_0_lowtown_a2_-196609_0y.dds"},
	{0x5CD6B3A7BAC97ACBULL, "892164713.ogg"},
	{0x5CD796E14A5159C9ULL, "6028847_m.fxe"},
	{0x5CD811BB4C9A358EULL, "hlac_lt_house_-1_0z.dds"},
	{0x5CD87DA9DDA68C54ULL, "vxlt_warehouse_n_yu_0_0.vlm"},
	{0x5CD8BFF9BD6E9F93ULL, "pro000cr_friedrich_thug.utc"},
	{0x5CDA337E35085B9EULL, "6098256_m.fxe"},
	{0x5CDA3A6190374338ULL, "narrative_flemeth6002778.cl"},
	{0x5CDA7435370E5898ULL, "lt_0_lowtown_docks_a3_30_0_lc.mmh"},
	{0x5CDA9323C4979AD9ULL, "6154119_m.fxe"},
	{0x5CDA957754CE7596ULL, "art\\characters\\playercharacter\\humanmale\\glv_medc.phy"},
	{0x5CDAA3047E47EC8BULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.impa.evt"},
	{0x5CDAA7F6BB8A105CULL, "w_sundermount_75_0.rml"},
	{0x5CDAC3BD7E9022A7ULL, "vxlt_0_lowtown_a2_eov_0_0.vlm"},
	{0x5CDAD71E4B803239ULL, "vxlt_0_lowtown_a3_n_c_bym_0_0.vlm"},
	{0x5CDAEC3300766C2AULL, "6110388_f.fxe"},
	{0x5CDB078DB042F5A1ULL, "mag121_idunna6083964.cl"},
	{0x5CDBB3D0CB986589ULL, "vxca_cave_18h_0_1.vlm"},
	{0x5CDBF82659EBDD1CULL, "491856092.ogg"},
	{0x5CDC5998B34E101AULL, "6162654_m.fxe"},
	{0x5CDC6BB72F1DFF0BULL, "vxw_encounter2_camp_i_0_0.vlm"},
	{0x5CDC71ADF9D93273ULL, "323152430.ogg"},
	{0x5CDC75507EA3EC12ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.dura.evt"},
	{0x5CDCE1DE13607941ULL, "158466847.ogg"},
	{0x5CDD3075DA4498ADULL, "art\\vfx\\environment\\vfx_fire_torch_p.dura.evt"},
	{0x5CDD3A5C3F855DA1ULL, "hlac_w_bonepitmining_a2_5_0z.dds"},
	{0x5CDD840FFA6D0813ULL, "m_dae_caverns_shared.bnk"},
	{0x5CDD84D6D51267C8ULL, "gen00fl_anders6125530.cl"},
	{0x5CDD8866076F941FULL, "vxlt_gallowstemplar_n_6sm_0_0.vlm"},
	{0x5CDDD564F0F3BB4CULL, "hlac_lt_0_lowtown_65537_3x.dds"},
	{0x5CDE06FDD9692D4DULL, "gen00fl_aveline6139508.cl"},
	{0x5CDE56DA281C9F76ULL, "l_mrl351ar_demon_cave.wbd"},
	{0x5CDE8DF6BB8DCCB1ULL, "w_sundermount_75_0.msh"},
	{0x5CDE9BF6BB8DE4CCULL, "w_sundermount_75_0.mao"},
	{0x5CDEA7F6BB8DF96FULL, "w_sundermount_75_0.mmh"},
	{0x5CDEDF23F0D22B6BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow2win.phy"},
	{0x5CDFAB40689B424EULL, "vxdr_primevaltomb_2vv_0_0.vlm"},
	{0x5CDFAED41E777315ULL, "6143238_m.fxe"},
	{0x5CDFD3EB5831C917ULL, "vxlt_gallowstemplar_f_3s0_0_0.vlm"},
	{0x5CDFDDFA18982FE7ULL, "67948357.ogg"},
	{0x5CDFE2ADFDBE6A89ULL, "vxlt_hangedman_n_1xp_0_0.vlm"},
	{0x5CE00E04901B2678ULL, "638891485.ogg"},
	{0x5CE02AF7E0D3B936ULL, "vxw_blightlands_h1_0_0.vlm"},
	{0x5CE09096EA455B8BULL, "vxw_bonepit_29d_0_0.vlm"},
	{0x5CE09D547EDAC018ULL, "dae000_tomwise6167287.cl"},
	{0x5CE109C6DBDD8035ULL, "6179059_m.fxe"},
	{0x5CE13528DA4A904EULL, "mcr370_negotiation6080215.cl"},
	{0x5CE1541DD698DA84ULL, "hlac_ht_keep_n_3732_19y.dds"},
	{0x5CE15432485F9FE5ULL, "36483746.ogg"},
	{0x5CE1BC047E4E364CULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.impa.gad"},
	{0x5CE1CDFC1C996815ULL, "conversation_tokens.gda"},
	{0x5CE234D72CFED6A6ULL, "68227582.ogg"},
	{0x5CE23570E0525E94ULL, "w_encounter2_36_0_ll.mmh"},
	{0x5CE247186089C139ULL, "416288705.ogg"},
	{0x5CE2B1017EF17B16ULL, "hlac_lt_gallowscourtyard_n_0_1z.dds"},
	{0x5CE394507EAA3FA5ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.dura.gad"},
	{0x5CE3AAC814AF4863ULL, "6128518_m.fxe"},
	{0x5CE3E1C9AAEBEA98ULL, "225412684.ogg"},
	{0x5CE3E55D2CBE2AE7ULL, "hlac_lt_0_lowtown_a2_n_c_65534_1y.dds"},
	{0x5CE4230DE7F6CBA9ULL, "clf260_harley6069957.cl"},
	{0x5CE47E7754D6A337ULL, "art\\characters\\playercharacter\\humanmale\\glv_medc.msh"},
	{0x5CE4907754D6C19AULL, "art\\characters\\playercharacter\\humanmale\\glv_medc.mao"},
	{0x5CE4947754D6C8A9ULL, "art\\characters\\playercharacter\\humanmale\\glv_medc.mmh"},
	{0x5CE55BA8CD728250ULL, "854025257.ogg"},
	{0x5CE588AB05DF8D52ULL, "character_character_p115.psh.fxo.dx11"},
	{0x5CE59E6FD401E3D1ULL, "vxht_keep_60w_0_1.vlm"},
	{0x5CE5A2D5E3CBC4A9ULL, "pro000st_dead_sibling.stg"},
	{0x5CE5B29D8FD556DCULL, "38689521.ogg"},
	{0x5CE5BBF3C881B22AULL, "aip_follower_fenris.gda"},
	{0x5CE63631B2923C79ULL, "vxw_bonepitmining_a2_20c_0_0.vlm"},
	{0x5CE68246722D8AB0ULL, "vxca_caverns1_dpe_0_0.vlm"},
	{0x5CE69FE06F0D0CB6ULL, "llac_ca_caverns1_mines_24425_10.dds"},
	{0x5CE7005377818720ULL, "360992283.ogg"},
	{0x5CE74459B1281CB1ULL, "zz_dog_debug6127932.cl"},
	{0x5CE76B84F98180B2ULL, "vxca_cavern1_haunted_ih7_0_0.vlm"},
	{0x5CE82A1DD59BD978ULL, "vxht_mansion2_1ag_0_0.vlm"},
	{0x5CE85598840DCC50ULL, "static_static_p1.vsh.fxo.dx11"},
	{0x5CE8BB047E54534AULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.impa.ani"},
	{0x5CE8C7B908818243ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_102a_d.dds"},
	{0x5CE9D1CE0DC578B6ULL, "207426407.ogg"},
	{0x5CE9E523BFBEAA3BULL, "mag291_vincento6027705.cl"},
	{0x5CEA0CEA2F080946ULL, "6100081_m.fxe"},
	{0x5CEA0E21B4C30199ULL, "vxht_0_hightown_a2_c8k_0_0.vlm"},
	{0x5CEA765497BAE676ULL, "gen00fl_isabela6075385.cl"},
	{0x5CEAA0401B5141D5ULL, "dae300_bran6143804.cl"},
	{0x5CEAB88433BC5320ULL, "1011143382.ogg"},
	{0x5CEACB516CDA42B9ULL, "plt_lgt000pt_ultra_5_05.plo"},
	{0x5CEAF2F61A09C83FULL, "vxdr_primevalentrance_1m4_0_1.vlm"},
	{0x5CEC316A460B3643ULL, "vxht_0_hightown_n_a79_0_0.vlm"},
	{0x5CEC84DD6ADF6511ULL, "6163571_m.fxe"},
	{0x5CED89303C6517D8ULL, "vxht_0_hightown_n_c8e_0_0.vlm"},
	{0x5CEDABB7C96687D4ULL, "919479344.ogg"},
	{0x5CEDF6642C7EEA99ULL, "co_len_flare.mao"},
	{0x5CEE38BA9C5302F6ULL, "ht_0_hightown_53_0_ll.mmh"},
	{0x5CEE50788E5D942BULL, "6040821_m.fxe"},
	{0x5CEE695AC5DDA65AULL, "lgt310cr_mis_burn_rival_thu.utc"},
	{0x5CEE9AA9F5D07476ULL, "6030610_m.fxe"},
	{0x5CEE9CD4D2C9D1B6ULL, "vxw_blightlands_m2_0_0.vlm"},
	{0x5CEECA0159FCD520ULL, "46676465.ogg"},
	{0x5CEF51122FF53B26ULL, "797520466.ogg"},
	{0x5CEFD1447BC9DA70ULL, "gen00fl_aveline6139877.cl"},
	{0x5CEFF7D063DB80E1ULL, "6093908_m.fxe"},
	{0x5CF0211B8989312AULL, "6179557_m.fxe"},
	{0x5CF0412330F4862DULL, "l_one300ar_crow_camp.wbd"},
	{0x5CF11B228B14AA48ULL, "mag311_keran6041487.cl"},
	{0x5CF148B524616864ULL, "861357565.ogg"},
	{0x5CF16CAF58BE99F0ULL, "6036290_m.fxe"},
	{0x5CF22CCF4392EC83ULL, "vxca_cavern1_haunted_edz_0_0.vlm"},
	{0x5CF2C0D960358F33ULL, "6180945_m.fxe"},
	{0x5CF3EFE7E2A3F2E5ULL, "lgt320_shw_ors_marlein6156310.cl"},
	{0x5CF406EE94E31FB8ULL, "and110_refugee6111611.cl"},
	{0x5CF48BF43FEB26BCULL, "571429570.ogg"},
	{0x5CF4977A9798E54EULL, "6127503_m.fxe"},
	{0x5CF5630279DB10F8ULL, "follower_banter6141841.cl"},
	{0x5CF58667BDCABB4EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb05a.mmh"},
	{0x5CF59C0EA7BC0764ULL, "llac_lt_undercitydungeon_-1_9.dds"},
	{0x5CF5A067BDCAE73CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb05a.msh"},
	{0x5CF5BB4EF607C6B3ULL, "llac_lt_0_lowtown_a2_n_c_65536_4.dds"},
	{0x5CF5D751CCCCE4EDULL, "isa321_velasco6117612.cl"},
	{0x5CF750C590511CC0ULL, "hlac_lt_0_lowtown_docks_a3_n_0_1y.dds"},
	{0x5CF7AA68F6F2AD9AULL, "follower_banter6173161.cl"},
	{0x5CF810D0EB01FA95ULL, "hlac_ht_0_hightown_a3_65538_0z.dds"},
	{0x5CF863984A18494AULL, "mag101_grace6053112.cl"},
	{0x5CF87A9FBE1CF792ULL, "6018545_f.fxe"},
	{0x5CF8BA9B09B43AB7ULL, "ser121cr_jethann.utc"},
	{0x5CF9700F919DB39FULL, "6139484_m.fxe"},
	{0x5CF98884E6C6FEEDULL, "plt_lgt200pt_night.plo"},
	{0x5CFB5DA5DE9E2998ULL, "6087756_f.fxe"},
	{0x5CFB94F98F9CD7F9ULL, "mer120_hubert6070224.cl"},
	{0x5CFBC1FEE9ABCF4FULL, "w_sundermount_mmap_ph.dds"},
	{0x5CFBF0F54643E4A2ULL, "vxlt_0_lowtown_n_h4x_0_0.vlm"},
	{0x5CFBFC7D93F30361ULL, "352123384.ogg"},
	{0x5CFBFF21F146A938ULL, "dae000_flemeth6013389.cl"},
	{0x5CFC76F3D5DD2F96ULL, "vxlt_shop_db_0_0.vlm"},
	{0x5CFC7799BBEBB4FFULL, "mag291_vincento6027363.cl"},
	{0x5CFC7835C2ADBBD3ULL, "one280_iwan6101124.cl"},
	{0x5CFC85993BC8E628ULL, "dae000_templar_recruit_16164439.cl"},
	{0x5CFCA2179B05E022ULL, "pro000_colonel_ewald6039954.cl"},
	{0x5CFD0EB1C4BFEA36ULL, "vxw_blightlands_v3_0_0.vlm"},
	{0x5CFD3BB41F169EC1ULL, "6111369_m.fxe"},
	{0x5CFD5A08B8F11B26ULL, "686089405.ogg"},
	{0x5CFD5BF4A7E4F97FULL, "w_blightlands_116_0.msh"},
	{0x5CFD61F4A7E50321ULL, "w_blightlands_116_0.mmh"},
	{0x5CFD6DF4A7E517C2ULL, "w_blightlands_116_0.mao"},
	{0x5CFD8B4834B51521ULL, "898847771.ogg"},
	{0x5CFE6F32373ED8DAULL, "mag130_danzig6018844.cl"},
	{0x5CFE84D1FF22FF3AULL, "w_encounter3_234_0.mmh"},
	{0x5CFE88D1FF230671ULL, "w_encounter3_234_0.mao"},
	{0x5CFE96D1FF231D80ULL, "w_encounter3_234_0.msh"},
	{0x5CFE97BBCD32C7CAULL, "537791378.ogg"},
	{0x5CFE98B65AE90543ULL, "vxht_0_hightown_a2_bek_0_0.vlm"},
	{0x5CFEAF762C130870ULL, "913587335.ogg"},
	{0x5CFF6C43EBDB00CDULL, "vxlt_gallowstemplar_4tw_0_0.vlm"},
	{0x5D0040F4FD6796A8ULL, "mcr370_meredith6136113.cl"},
	{0x5D00A56093783106ULL, "535318102.ogg"},
	{0x5D021534F15261DEULL, "hlac_lt_gallowscourtyard_131071_0y.dds"},
	{0x5D0223E5C40B0C4FULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody7_.gad"},
	{0x5D0274B84657B666ULL, "240280859.ogg"},
	{0x5D02B2101E7FF226ULL, "hlac_lt_0_lowtown_n_131072_2z.dds"},
	{0x5D0338F7A03B0DC5ULL, "vxht_brothel_n_5c8_0_0.vlm"},
	{0x5D033BE37587CB23ULL, "6033539_m.fxe"},
	{0x5D03463077134545ULL, "6139796_m.fxe"},
	{0x5D03635B3EC2A1F8ULL, "782317660.ogg"},
	{0x5D03E63D93CBD1C4ULL, "6160690_m.fxe"},
	{0x5D0439995F559AF3ULL, "llac_lt_backalley_n_1_1.dds"},
	{0x5D049B881690C0F4ULL, "one100_solivitus6048141.cl"},
	{0x5D04D4C1509A828CULL, "136144441.ogg"},
	{0x5D05001DC338BDB5ULL, "hlac_ca_caverns1_24426_5y.dds"},
	{0x5D050FE159A47917ULL, "vxw_encounter2_camp_rt_0_1.vlm"},
	{0x5D055584E3CDEFC6ULL, "138101284.ogg"},
	{0x5D0567A2AFF125C3ULL, "6091984_m.fxe"},
	{0x5D05E66D6E99B5F5ULL, "hlac_lt_undercityserialkiller_-1_21x.dds"},
	{0x5D067E91DCF2580BULL, "dae000_gamlen6121574.cl"},
	{0x5D06C4C82C3BF6CCULL, "hlac_w_bonepitdestroyed_a3_65537_0z.dds"},
	{0x5D06E37D8E2DAC8BULL, "hlac_lt_undercityserialkiller_-1_5z.dds"},
	{0x5D06F708E98F7B43ULL, "hlac_w_blightlands_393214_0z.dds"},
	{0x5D07081C6F31B463ULL, "art\\vfx\\models\\vfs_grenade_bottle.msh"},
	{0x5D07161C6F31CBBDULL, "art\\vfx\\models\\vfs_grenade_bottle.mmh"},
	{0x5D072F6C93A1A6CBULL, "172313707.ogg"},
	{0x5D07A9FCD7702D58ULL, "llac_lt_0_lowtown_a2_n_-1_0.dds"},
	{0x5D07D3860A561CF4ULL, "448481281.ogg"},
	{0x5D085549DD72320CULL, "gen_im_trs_cot_junk2_01.uti"},
	{0x5D0A631BA08F5A8BULL, "hlac_ca_caverns2prc_5802_3z.dds"},
	{0x5D0A68D1DEA31A95ULL, "vxw_woundedcoast_17h_0_0.vlm"},
	{0x5D0A7CDA28421AE4ULL, "l_mrl351ar_demon_cave.bnk"},
	{0x5D0B17412D323BA3ULL, "and330tr_worldmap.ncs"},
	{0x5D0B31A908E1F4EEULL, "vxht_brothel_n_4g2_0_0.vlm"},
	{0x5D0B3B8A90898E59ULL, "668288494.ogg"},
	{0x5D0B766CA740954BULL, "6144189_m.fxe"},
	{0x5D0C5B0060D5C302ULL, "dae000_flemeth6000934.cl"},
	{0x5D0CC6516CF720CEULL, "plt_lgt000pt_ultra_5_05.nss"},
	{0x5D0D282A6183108CULL, "vxlt_0_lowtown_e22_0_0.vlm"},
	{0x5D0D65D2014720A6ULL, "480991678.ogg"},
	{0x5D0DDB0F386F3131ULL, "llac_lt_hangedman_n_-1_9.dds"},
	{0x5D0E38B9392D5A03ULL, "vxht_chantry_n_2cj_0_0.vlm"},
	{0x5D0E6630C55F7800ULL, "dae000ip_crowd_em3_talk17.utp"},
	{0x5D0E6D708D07C721ULL, "6015686_f.fxe"},
	{0x5D0E7088091E7077ULL, "20720567.ogg"},
	{0x5D0EB4D1145DFC3BULL, "6016949_m.fxe"},
	{0x5D0FE42BFE66524EULL, "ht_0_hightown_16_0_lc.mmh"},
	{0x5D1005A44A8296FCULL, "vxca_caverns2_6t8_0_0.vlm"},
	{0x5D1083B4BF231F97ULL, "6110415_f.fxe"},
	{0x5D10FB756F6D9AA7ULL, "zz_refugeehawke2.utc"},
	{0x5D116CA4A19F6D93ULL, "art\\vfx\\cinematic\\vfx_arishok_spark_p.anb"},
	{0x5D119FC5C80482DEULL, "820550420.ogg"},
	{0x5D11D832CCB68C0BULL, "dae000ip_crowd_mf5_talk15.utp"},
	{0x5D12A9C190561A52ULL, "6084275_m.fxe"},
	{0x5D13828A2C389741ULL, "dae221ip_home_amb_merr_flw1.utp"},
	{0x5D13AA959A7BAAE8ULL, "556406653.ogg"},
	{0x5D13BE893CA74D27ULL, "365008388.ogg"},
	{0x5D13F9D38EC13A96ULL, "dae200_cullen6140261.cl"},
	{0x5D1464676907F2ABULL, "dae000_dalish_apprentices6147543.cl"},
	{0x5D14D63E5A67402BULL, "vxlt_0_lowtown_a2_elj_0_1.vlm"},
	{0x5D14F4387B47EA2FULL, "gen_im_arm_hlm_lgt_act3_03.uti"},
	{0x5D151180C8DF0B63ULL, "668288816.ogg"},
	{0x5D152D4C596787BCULL, "751928927.ogg"},
	{0x5D15B7551DDBDC95ULL, "art\\characters\\weapons\\axes\\w_axe_101a.mao"},
	{0x5D162AC9784960AEULL, "carpetnew_d.dds"},
	{0x5D1640E5C41BE615ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody7_.ani"},
	{0x5D16824A63A6D7D0ULL, "mcr370_templar1.utc"},
	{0x5D16E8CA676C4989ULL, "art\\characters\\creatures\\dragon\\animation\\choke.ani"},
	{0x5D1791B92E4BB0BEULL, "6104012_m.fxe"},
	{0x5D17C6C1AC17E977ULL, "vxdr_primevaltomb_3a3_0_0.vlm"},
	{0x5D17D9C69B1E1CD7ULL, "6105861_m.fxe"},
	{0x5D1815843670FE61ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chairxa.phy"},
	{0x5D18230D863986E6ULL, "vxht_brothel_n_6ef_0_0.vlm"},
	{0x5D18DAAED0993315ULL, "vxw_woundedcoast_12w_0_0.vlm"},
	{0x5D18E28500128AE4ULL, "seb201_lady_harimann6177592.cl"},
	{0x5D19009B347E1AD5ULL, "lt_warehouse_mmap_ph.dds"},
	{0x5D19025AD13690B6ULL, "156950130.ogg"},
	{0x5D191FA0CEBE9956ULL, "vxlt_gallowstemplar_3sb_0_0.vlm"},
	{0x5D1984EAF484D2CBULL, "6086908_m.fxe"},
	{0x5D19940D6D9D5323ULL, "vxht_mansion_n_5br_0_0.vlm"},
	{0x5D19A16EA513AB96ULL, "597502213.ogg"},
	{0x5D19A88E968AA5EDULL, "vxht_brothel_n_4cy_0_0.vlm"},
	{0x5D19D0C14554386AULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_nud_0d.dds"},
	{0x5D1A0A0D160F9624ULL, "hlac_w_encounter1_0_2y.dds"},
	{0x5D1A4749EE5E7E1EULL, "425328495.ogg"},
	{0x5D1A76AE1C7A5383ULL, "vfx.mat"},
	{0x5D1AAB66DE76BA8CULL, "6020526_f.fxe"},
	{0x5D1B3B7585362366ULL, "vxw_blightlands_7p_0_0.vlm"},
	{0x5D1B6236CF51E483ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_action.ani"},
	{0x5D1B780564E88E2CULL, "art\\characters\\playercharacter\\humanmale\\glv_bda.phy"},
	{0x5D1BDAE4DBD9A430ULL, "vxw_bonepit_2ac_0_0.vlm"},
	{0x5D1BE07B932AFED2ULL, "474678477.ogg"},
	{0x5D1C170E082386EDULL, "gen00fl_varric6118980.cl"},
	{0x5D1C52D3426EBAFCULL, "gen00fl_aveline6128981.cl"},
	{0x5D1C649104C02F3DULL, "lgt000pt_gangs.ncs"},
	{0x5D1CA15870CCA902ULL, "ef_athenril.mor"},
	{0x5D1CE8665A8220BBULL, "qun120_arishok6051323.cl"},
	{0x5D1E03CA6772960BULL, "art\\characters\\creatures\\dragon\\animation\\choke.gad"},
	{0x5D1E25FF52C52174ULL, "172469700.ogg"},
	{0x5D1E89A4A1AA29FCULL, "art\\vfx\\cinematic\\vfx_arishok_spark_p.mmh"},
	{0x5D1EB32B92D2F6EDULL, "6132649_m.fxe"},
	{0x5D1F0E80053AE0D5ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.mmh"},
	{0x5D1F16337931ABA8ULL, "552755744.ogg"},
	{0x5D1F31B85AA5D6D3ULL, "vxca_caverns2_79l_0_1.vlm"},
	{0x5D1F4F5F843FE6F2ULL, "llac_ca_caverns1_mines_24426_0.dds"},
	{0x5D1F7D5C6FCBE563ULL, "hlac_lt_0_lowtown_a2_0_9y.dds"},
	{0x5D1FF0E148BCF610ULL, "540932227.ogg"},
	{0x5D200E453D742F9FULL, "qun211ip_barrel.ncs"},
	{0x5D20793C966C8B19ULL, "663717689.ogg"},
	{0x5D20CF3AC86DB84BULL, "vxdr_dungeon_2id_0_0.vlm"},
	{0x5D21A2529917F3AFULL, "282295407.ogg"},
	{0x5D21B9C0CB53493AULL, "165825218.ogg"},
	{0x5D22E276DEE25B67ULL, "6188581_m.fxe"},
	{0x5D231DD8AEB7FBA6ULL, "6156767_m.fxe"},
	{0x5D2468AE1C8290C1ULL, "vfx.ncs"},
	{0x5D24A20FFAA93E71ULL, "m_dae_caverns_shared.wbd"},
	{0x5D24DF874114B674ULL, "one100_party_comments6133376.cl"},
	{0x5D2500CA6778B070ULL, "art\\characters\\creatures\\dragon\\animation\\choke.evt"},
	{0x5D263D3C3214A123ULL, "6202807_m.fxe"},
	{0x5D266D2130982970ULL, "lowtown_male_drunk_mob6151173.cl"},
	{0x5D26EFC41999795BULL, "6096480_m.fxe"},
	{0x5D26F898581C3D3FULL, "plt_genpt_app_carver.plo"},
	{0x5D270D991BA17293ULL, "947804621.ogg"},
	{0x5D27AA072A5B4712ULL, "dae000ip_crowd_mf2_talk4.utp"},
	{0x5D27B89C0824E465ULL, "pou_fire.mao"},
	{0x5D281C7FCCECF243ULL, "783431627.ogg"},
	{0x5D287A36CF5C987AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_action.evt"},
	{0x5D287B74C5BDC29BULL, "hlac_w_sundermount_327679_0x.dds"},
	{0x5D28AB1E058BBE4FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud3high.phy"},
	{0x5D28AED9A7A25EC0ULL, "pro000cs_real_surv.cut"},
	{0x5D28AFA37AE0196EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\blendtree\\meredith_override.anb"},
	{0x5D28D135ADA54F0FULL, "qcr000_arishok6060012.cl"},
	{0x5D29072AB010EF5CULL, "290618064.ogg"},
	{0x5D297233AF5F8B99ULL, "hlac_lt_0_lowtown_a2_n_c_65537_0y.dds"},
	{0x5D29AD52772C3BE6ULL, "vxw_blightlands_8q_0_0.vlm"},
	{0x5D29CB6810D58F50ULL, "mcr310tr_party_banter.utt"},
	{0x5D29CD9B5BE418ECULL, "577498565.ogg"},
	{0x5D29CEBAD8D4DFB7ULL, "tobrius.mor"},
	{0x5D2A1954D8152A4EULL, "art\\characters\\creatures\\stonegolem\\animation\\dormant.ani"},
	{0x5D2A7D1AA7C45589ULL, "llac_lt_0_lowtown_a2_n_c_65535_1.dds"},
	{0x5D2AB2C92DC5A208ULL, "6149898_f.fxe"},
	{0x5D2AFB74104B48DEULL, "277029807.ogg"},
	{0x5D2B69F84390A573ULL, "356790782.ogg"},
	{0x5D2BCECE9733288DULL, "274757612.ogg"},
	{0x5D2C0676536491BFULL, "art\\vfx\\combat\\vfx_c_mer_attk3_c.mmh"},
	{0x5D2C0B8005456662ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.anb"},
	{0x5D2C303BBD9CA104ULL, "hlac_lt_0_lowtown_a2_n_c_65536_9x.dds"},
	{0x5D2C6B7EFD655ABCULL, "vxht_0_hightown_a2_c82_0_0.vlm"},
	{0x5D2CC5AC0F2A7ADDULL, "lowtown_male_lowlife1_sta.bnk"},
	{0x5D2CCFFB0BB16C22ULL, "matdef_df_normal.dds"},
	{0x5D2D0A2EDE110237ULL, "follower_banter6172918.cl"},
	{0x5D2D131C0F8C6AB5ULL, "zz_mer_debug6062736.cl"},
	{0x5D2D1729703A35ACULL, "vxca_caverns1_hm2_0_0.vlm"},
	{0x5D2D3F03BAA80B5DULL, "vxw_woundedcoast_1g0_0_0.vlm"},
	{0x5D2D735B8EEABFEBULL, "6081981_m.fxe"},
	{0x5D2DA51E2096A96DULL, "vxdr_primevaltomb_6j1_0_0.vlm"},
	{0x5D2E167DB99BCCE7ULL, "gen00fl_isabela6099326.cl"},
	{0x5D2ECB38C7CEC7C8ULL, "401733121.ogg"},
	{0x5D2EFBBCB60FF4DDULL, "vxlt_darktown_54r_0_0.vlm"},
	{0x5D2F2178B01532F6ULL, "vxlt_backalley_n_1y5_0_0.vlm"},
	{0x5D2F41072A14B269ULL, "6080272_m.fxe"},
	{0x5D2F7936CF62B56DULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_action.gad"},
	{0x5D2FA7AC0F2C7DE5ULL, "lowtown_male_lowlife1_sta.cnv"},
	{0x5D3065B500DC0A89ULL, "553471286.ogg"},
	{0x5D30C39413991C39ULL, "w_encounter3_222_0.msh"},
	{0x5D30C592D3D95A77ULL, "follower_banter6141798.cl"},
	{0x5D30CD9413992D27ULL, "w_encounter3_222_0.mmh"},
	{0x5D30D19413993474ULL, "w_encounter3_222_0.mao"},
	{0x5D30F71E96841D15ULL, "6127979_m.fxe"},
	{0x5D3142A4C68E94E5ULL, "hlac_ca_caverns2_5802_55x.dds"},
	{0x5D314B23312C2E57ULL, "l_one300ar_crow_camp.bnk"},
	{0x5D31CA5EC477A151ULL, "lt_0_lowtown_docks_n_169_0.msh"},
	{0x5D31D85EC477B96CULL, "lt_0_lowtown_docks_n_169_0.mao"},
	{0x5D31E45EC477CD0FULL, "lt_0_lowtown_docks_n_169_0.mmh"},
	{0x5D32262A3EE0CE80ULL, "gencr_abomination.utc"},
	{0x5D3286C4AF569197ULL, "vfx_combat_blood_02.mao"},
	{0x5D329F42A39E9933ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv2.evt"},
	{0x5D32C6943396BEFCULL, "pro000_cut2d_real_story6013917.cl"},
	{0x5D32E0AD352A4E98ULL, "zz_dre_debug6056975.cl"},
	{0x5D3301A221E708CFULL, "6128374_m.fxe"},
	{0x5D33B901D93D7A23ULL, "vxlt_hangedman_1qr_0_0.vlm"},
	{0x5D33C48CEBEAC17CULL, "scaleform_sf_main_p2.vsh.fxo.dx11.assembly"},
	{0x5D340CDB14663F50ULL, "6172235_m.fxe"},
	{0x5D343D3EB91C6810ULL, "hlac_dr_0_deeproads_5839_2y.dds"},
	{0x5D34A0A72CA5DB99ULL, "mcr310_argument6075067.cl"},
	{0x5D3506C76A25B7E6ULL, "vxht_brothel_n_6kh_0_0.vlm"},
	{0x5D35626DA0E0CFC3ULL, "vxlt_0_lowtown_a2_hem_0_0.vlm"},
	{0x5D35889EC28DD348ULL, "6105383_m.fxe"},
	{0x5D35BBF99105A6C7ULL, "vxca_caverns1_mines_efx_0_0.vlm"},
	{0x5D35E225888FB1E1ULL, "vxlt_0_lowtown_ghs_0_0.vlm"},
	{0x5D361C2B181D622AULL, "mag311_keran6041496.cl"},
	{0x5D3677C751898423ULL, "vxht_mansion_n_5ht_0_0.vlm"},
	{0x5D36C1D9ED42D442ULL, "llac_w_bonepitdestroyed_a3_131072_0.dds"},
	{0x5D374E2D0CC664C5ULL, "81410227.ogg"},
	{0x5D376209113CFE27ULL, "799953706.ogg"},
	{0x5D379E923EEC97C0ULL, "27508251.ogg"},
	{0x5D37A822BE3C5B5AULL, "qun111_ketojan6040154.cl"},
	{0x5D37F40D8431AF6FULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_1x.dds"},
	{0x5D382444433B4984ULL, "hlac_ht_chantry_2534_6y.dds"},
	{0x5D38D773C3EF89CBULL, "988630331.ogg"},
	{0x5D38E42BBD59999CULL, "6106049_m.fxe"},
	{0x5D396331755F99ABULL, "830775256.ogg"},
	{0x5D399842A3A4AC74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv2.gad"},
	{0x5D3B0D2D62900BE1ULL, "45800338.ogg"},
	{0x5D3B2BAD90A70CE2ULL, "llac_w_blightlands_327683_0.dds"},
	{0x5D3B7F5ACAB4648AULL, "ser221_party_comments6166194.cl"},
	{0x5D3B8E206C51561CULL, "vxht_0_hightown_asu_0_0.vlm"},
	{0x5D3BA69AEE20B60EULL, "llac_dr_0_deeproads_5838_0.dds"},
	{0x5D3BA7A1D95159BEULL, "hlac_ca_caverns2_5802_22y.dds"},
	{0x5D3BCEC2BC67FDA6ULL, "6087797_m.fxe"},
	{0x5D3BD4C09F728827ULL, "qun110_varnell6047724.cl"},
	{0x5D3BE71F9DD254E5ULL, "vxw_woundedcoast_178_0_0.vlm"},
	{0x5D3C2AD8EB6BC345ULL, "643213529.ogg"},
	{0x5D3CBD630971D6FDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01a.phy"},
	{0x5D3D05C146DD2EA8ULL, "6206237_m.fxe"},
	{0x5D3DDE959B545943ULL, "169433296.ogg"},
	{0x5D3DE0A55D52DD23ULL, "hlac_lt_undercityencounter_-1_48y.dds"},
	{0x5D3E727C03C19DF0ULL, "vxlt_backalley_n_1r4_0_2.vlm"},
	{0x5D3E9E4C667CD795ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_015a_0s.dds"},
	{0x5D3EF254D8274388ULL, "art\\characters\\creatures\\stonegolem\\animation\\dormant.gad"},
	{0x5D3F066430810347ULL, "zz_followers_banter6159159.cl"},
	{0x5D3F6ADB49E45DC8ULL, "vxlt_0_lowtown_ghv_0_0.vlm"},
	{0x5D3FD0CA959F9619ULL, "217641107.ogg"},
	{0x5D3FEDC6F77EA1B3ULL, "fx_c_grad_add.mao"},
	{0x5D40284B0E4FB565ULL, "6119295_m.fxe"},
	{0x5D404E4B66934B9AULL, "vxlt_0_lowtown_docks_a3_cuq_0_0.vlm"},
	{0x5D40A742A3AAE452ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv2.ani"},
	{0x5D415D59E5027C13ULL, "6108017_m.fxe"},
	{0x5D4194235107E900ULL, "hlac_lt_backalley_n_0_18z.dds"},
	{0x5D41BF5E0663744CULL, "51203872.ogg"},
	{0x5D420005B8398173ULL, "m_pro000ar_blightlands_fake.bnk"},
	{0x5D4237BCC8CB4CABULL, "hlac_w_encounter2_196611_0y.dds"},
	{0x5D428611D0EC0F0AULL, "6206921_m.fxe"},
	{0x5D4293FC74C88999ULL, "794409824.ogg"},
	{0x5D431E8264EF42ACULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_p.gad"},
	{0x5D4339618886919AULL, "6044859_m.fxe"},
	{0x5D43492E356AB389ULL, "dae221ar_player_home_nt.lst"},
	{0x5D43D44A92D6E843ULL, "vxlt_0_lowtown_a2_hbl_0_0.vlm"},
	{0x5D44C79AD96EAE2BULL, "art\\vfx\\combat\\vfx_gib_ogre_bld_p.impa.evt"},
	{0x5D45458C0F5C1B13ULL, "qun210cr_guard6072075.cl"},
	{0x5D45AED512B4C6CBULL, "art\\vfx\\combat\\vfx_glb_charm_c.impa.ani"},
	{0x5D46DC3DC5904BFDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_intimidate_pose_en.evt"},
	{0x5D4712BFC5844AAFULL, "6025118_m.fxe"},
	{0x5D4732B7D3CD5103ULL, "mcr310_argument6075115.cl"},
	{0x5D47918D42978E84ULL, "llac_w_bonepitmining_a2_0_6.dds"},
	{0x5D47F81E27240578ULL, "zz_qun_debug6028250.cl"},
	{0x5D47FC3B08302FDEULL, "mcr370_meredith6083602.cl"},
	{0x5D494ED2D54A71DDULL, "vxlt_undercitydungeon_yk_0_0.vlm"},
	{0x5D49FAB3C92EE03EULL, "977218947.ogg"},
	{0x5D49FCF78632F234ULL, "lt_0_lowtown_a2_n_c_9_0.phy"},
	{0x5D4A1D8264F55FAAULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_p.ani"},
	{0x5D4A2BB4C391DA96ULL, "gen_im_trs_chn_rng_r1.uti"},
	{0x5D4A75598DD5BA38ULL, "mrl150_merrill6084469.cl"},
	{0x5D4B03CC0256F202ULL, "vxlt_warehouse_113_0_0.vlm"},
	{0x5D4B7025DA7B4DDBULL, "561153990.ogg"},
	{0x5D4BC7EECA6EE550ULL, "tm_war_d06.tnt"},
	{0x5D4BE09AD974F7ECULL, "art\\vfx\\combat\\vfx_gib_ogre_bld_p.impa.gad"},
	{0x5D4BEF19D279CF54ULL, "mcr370_templar16087626.cl"},
	{0x5D4C138C14159152ULL, "vxlt_warehouse_n_ak_0_0.vlm"},
	{0x5D4C2C2D95FB55C1ULL, "6169898_m.fxe"},
	{0x5D4C71F8D6D12DF9ULL, "1046430760.ogg"},
	{0x5D4CCB937365E6EDULL, "mrl150_dalish_guard6089416.cl"},
	{0x5D4D35A7DD5A3D97ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_sloped.phy"},
	{0x5D4D35D9DE88AF62ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkfront.gad"},
	{0x5D4D4C3EB5349499ULL, "w_bonepitdestroyed_a3_168_0.msh"},
	{0x5D4D563EB534A587ULL, "w_bonepitdestroyed_a3_168_0.mmh"},
	{0x5D4D5A3EB534ACD4ULL, "w_bonepitdestroyed_a3_168_0.mao"},
	{0x5D4D761BE4A5D6B8ULL, "dae000_merchant_access_hf6162963.cl"},
	{0x5D4DBC07F7CE5328ULL, "6016010_m.fxe"},
	{0x5D4DBE6309804B02ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01a.mmh"},
	{0x5D4DC06309804EB8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01a.msh"},
	{0x5D4DDAD87A704E15ULL, "6129779_m.fxe"},
	{0x5D4E19C530F682F1ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isab_0s.dds"},
	{0x5D4E27FDD323A786ULL, "qun131_warrens_thug6038837.cl"},
	{0x5D4E6AEC49C2E41EULL, "154195720.ogg"},
	{0x5D4E970968DCBB58ULL, "6200860_m.fxe"},
	{0x5D4EBA6810F48E0DULL, "mcr310tr_party_banter.ncs"},
	{0x5D4F1309DB1E9DE9ULL, "186918642.ogg"},
	{0x5D4F2B2C527DAD19ULL, "vxht_0_hightown_a2_cel_0_0.vlm"},
	{0x5D4F2C4597CDF8EAULL, "hlac_w_sundermount_0_3z.dds"},
	{0x5D4F6929AA04A3EEULL, "dae000_bartrand6032011.cl"},
	{0x5D4F7965DE52214FULL, "vxlt_0_lowtown_a2_n_c_gxx_0_0.vlm"},
	{0x5D4FEBCF9B335E2EULL, "258247417.ogg"},
	{0x5D5056C7D4A96E47ULL, "vxlt_gallowstemplar_2lk_0_0.vlm"},
	{0x5D5083A276A7BDAEULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtb.msh"},
	{0x5D5091A276A7D5FFULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtb.mao"},
	{0x5D509DA276A7E99CULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtb.mmh"},
	{0x5D50E612F2D629ECULL, "hlac_ht_mansion2_818_19x.dds"},
	{0x5D5141F451F28759ULL, "812220321.ogg"},
	{0x5D5218136B2D66D7ULL, "6145146_m.fxe"},
	{0x5D52742236AF210BULL, "822123814.ogg"},
	{0x5D52771D8430DC6DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\_textures\\prp_cot_d.dds"},
	{0x5D52920378177998ULL, "6061789_f.fxe"},
	{0x5D52DF9AD97B14EAULL, "art\\vfx\\combat\\vfx_gib_ogre_bld_p.impa.ani"},
	{0x5D52F18D9F4F585BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_flat.msh"},
	{0x5D53038D9F4F76A6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_flat.mao"},
	{0x5D530F8D9F4F8B45ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_flat.mmh"},
	{0x5D5358D9DE8D5659ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkfront.evt"},
	{0x5D5396D512C0DB42ULL, "art\\vfx\\combat\\vfx_glb_charm_c.impa.evt"},
	{0x5D53CDF8FD6B4285ULL, "25007799.ogg"},
	{0x5D53D9F7863B0B31ULL, "lt_0_lowtown_a2_n_c_9_0.msh"},
	{0x5D53E7F7863B234CULL, "lt_0_lowtown_a2_n_c_9_0.mao"},
	{0x5D53F3F7863B37EFULL, "lt_0_lowtown_a2_n_c_9_0.mmh"},
	{0x5D54422E35791B2EULL, "dae221ar_player_home_nt.are"},
	{0x5D5474600CE9774EULL, "art\\vfx\\combat\\material__26.mao"},
	{0x5D54BCAD4ABEE1EAULL, "lt_0_lowtown_docks_n_123_0.mao"},
	{0x5D54C0AD4ABEE839ULL, "lt_0_lowtown_docks_n_123_0.mmh"},
	{0x5D54CAAD4ABEF927ULL, "lt_0_lowtown_docks_n_123_0.msh"},
	{0x5D550776538798FCULL, "art\\vfx\\combat\\vfx_c_mer_attk3_c.anb"},
	{0x5D551B5DB64B3A84ULL, "follower_banter6156269.cl"},
	{0x5D557D12BB301C62ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basicnb.msh"},
	{0x5D557F12BB301F18ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basicnb.mmh"},
	{0x5D558B12BB3033BBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basicnb.mao"},
	{0x5D560A221E5201D9ULL, "272630427.ogg"},
	{0x5D5679D6D8F6E852ULL, "static1_ds1y.psh.fxo"},
	{0x5D567E9030B015FAULL, "llac_w_blightlands_-4_0.dds"},
	{0x5D56D638FE076AFEULL, "hlac_lt_0_lowtown_a3_n_c_-1_7z.dds"},
	{0x5D56FBBEAC5E1723ULL, "vxlt_gallowsprison_n_5qx_0_0.vlm"},
	{0x5D5743135A68CF31ULL, "vxlt_undercityserialkiller_20w_0_0.vlm"},
	{0x5D577727AD2B6494ULL, "vxw_bonepitmining_a2_8j_0_0.vlm"},
	{0x5D57C66DFE5B61B4ULL, "vxlt_foundry_n_zr_0_0.vlm"},
	{0x5D57F5728999B93FULL, "lig3.dds"},
	{0x5D581394551AD0FBULL, "vxlt_0_lowtown_a2_e25_0_0.vlm"},
	{0x5D5995D512C54555ULL, "art\\vfx\\combat\\vfx_glb_charm_c.impa.gad"},
	{0x5D599FDD4C254593ULL, "171421568.ogg"},
	{0x5D59CA5EBC9B2377ULL, "773199072.ogg"},
	{0x5D59D7457CA6DFF2ULL, "543462646.ogg"},
	{0x5D5B306F086F0DC8ULL, "and231tr_spiders.utt"},
	{0x5D5B632B306C90B5ULL, "vxht_mansion_n_5m8_0_1.vlm"},
	{0x5D5B83AECEBA7043ULL, "hlac_ht_mansion_n_3373_3x.dds"},
	{0x5D5B98CE338F8990ULL, "vxca_cave_2bi_0_0.vlm"},
	{0x5D5BBE37268FDA3EULL, "follower_banter6141942.cl"},
	{0x5D5BCE12AD570C7BULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1a.msh"},
	{0x5D5BE012AD572AC6ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1a.mao"},
	{0x5D5BEC12AD573F65ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadf1a.mmh"},
	{0x5D5C2A1D184DC9A4ULL, "64753777.ogg"},
	{0x5D5C3239B3BA97E1ULL, "953834076.ogg"},
	{0x5D5C4F05651FDFF7ULL, "art\\characters\\playercharacter\\humanmale\\glv_bda.mmh"},
	{0x5D5C650565200569ULL, "art\\characters\\playercharacter\\humanmale\\glv_bda.msh"},
	{0x5D5D230F3CDA4D37ULL, "gen_im_cns_pot_mana_01.uti"},
	{0x5D5D306AE87F9ECEULL, "lt_backalley_g_mmap_ph.dds"},
	{0x5D5D43CE1355B802ULL, "vxdr_0_deeproads_egx_0_0.vlm"},
	{0x5D5D6EE45628C1FFULL, "dae210ip_store_access.utp"},
	{0x5D5D818A06138F65ULL, "6107004_f.fxe"},
	{0x5D5DE6FE39869C01ULL, "hlac_ca_caverns2prc_5802_52x.dds"},
	{0x5D5DF9BE13097B08ULL, "ht_0_hightown_a3_18_0.rml"},
	{0x5D5ECC7390A40962ULL, "171209899.ogg"},
	{0x5D5ED322ECED94E4ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar3_mcr370_playerrules.ani"},
	{0x5D5EDD0BF9E74756ULL, "vxht_0_hightown_a2_arx_0_0.vlm"},
	{0x5D5F3500A792F126ULL, "hlac_lt_warehouse_n_1_0y.dds"},
	{0x5D5F8612BB387FA3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basicnb.phy"},
	{0x5D5FB254E6017B97ULL, "vxlt_gallowstemplar_n_3j7_0_0.vlm"},
	{0x5D5FFDBE63E5540EULL, "405753312.ogg"},
	{0x5D601C412D7A843AULL, "and330tr_worldmap.utt"},
	{0x5D6022637B33A742ULL, "6108690_m.fxe"},
	{0x5D602784831A2BACULL, "dae111ip_home_amb_brush.utp"},
	{0x5D6061423E92F84DULL, "662100259.ogg"},
	{0x5D60F321C9064D5DULL, "dre000_dougal6131176.cl"},
	{0x5D6160D9DE99A148ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkfront.ani"},
	{0x5D616A693C24EC7EULL, "lgt260_eau_tmp_roderick6143704.cl"},
	{0x5D61D357C222E337ULL, "vxht_chantry_36g_0_0.vlm"},
	{0x5D626BBD0E2198C3ULL, "6164730_m.fxe"},
	{0x5D62E3BF8A2FB52AULL, "zz_otl_debug6060590.cl"},
	{0x5D62F61EB603452FULL, "378283176.ogg"},
	{0x5D632AD77172A2B1ULL, "vxw_bonepitmining_a2_25d_0_0.vlm"},
	{0x5D6334DF484CAE9BULL, "543096514.ogg"},
	{0x5D633FD4401AA0A3ULL, "dae300_vashothlead6062794.cl"},
	{0x5D637D4594F91352ULL, "hlac_lt_blackemporium_1055_19z.dds"},
	{0x5D63D9523DBBF61BULL, "vxht_bartrand_n_6k1_0_0.vlm"},
	{0x5D641352AC21180BULL, "vxca_caverns1_gh9_0_0.vlm"},
	{0x5D64D9DE1E9D5C32ULL, "275558477.ogg"},
	{0x5D651C5AC2C3BDDEULL, "hlac_ht_keep_n_3732_10x.dds"},
	{0x5D657F98AD95681CULL, "vxca_cavern1_haunted_dpd_0_0.vlm"},
	{0x5D65D022ECF3AE46ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar3_mcr370_playerrules.gad"},
	{0x5D65E936466E540AULL, "dae000_arishok6082020.cl"},
	{0x5D6637D9EED492D5ULL, "plt_cod_mgc_circle_history.plo"},
	{0x5D663DA0FC56B1F1ULL, "vxlt_0_lowtown_a3_n_c_byu_0_0.vlm"},
	{0x5D664EF2D4382B2EULL, "6108100_m.fxe"},
	{0x5D66A25FD24AE95CULL, "899147905.ogg"},
	{0x5D675D47C68A2E48ULL, "qcr000_varric.cnv"},
	{0x5D67926D13265439ULL, "6078334_m.fxe"},
	{0x5D68F2C2865A1C0DULL, "852416120.ogg"},
	{0x5D68FD71BA03EA7BULL, "vxca_caverns1_ikn_0_0.vlm"},
	{0x5D69836735222159ULL, "vxlt_0_lowtown_n_df9_0_0.vlm"},
	{0x5D69969E6A971819ULL, "zz_approval_debug6003774.cl"},
	{0x5D69D4C3DF9E82E6ULL, "6072722_m.fxe"},
	{0x5D69DD17332753C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.ani"},
	{0x5D6A1A4146A18989ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_check_nails.ani"},
	{0x5D6A3A8774FE57A1ULL, "6094530_m.fxe"},
	{0x5D6A7A3EEAA55775ULL, "6108635_m.fxe"},
	{0x5D6A7F47C68C9DB0ULL, "qcr000_varric.bnk"},
	{0x5D6C03FA8646794CULL, "481798816.ogg"},
	{0x5D6C05360F1C658BULL, "gen00fl_isabela_init6076478.cl"},
	{0x5D6D3A81B8959F47ULL, "vxlt_gallowstemplar_2be_0_0.vlm"},
	{0x5D6D87D57B0C5110ULL, "6149156_m.fxe"},
	{0x5D6DA54047704210ULL, "722955262.ogg"},
	{0x5D6DCF3E356F6A86ULL, "vxlt_gallowscourtyard_6uo_0_0.vlm"},
	{0x5D6DDF3B158F20ABULL, "vxht_mansion2_3cx_0_0.vlm"},
	{0x5D6DE151B691BD8EULL, "t3_wdg_tb2.tnt"},
	{0x5D6E6E2FDD0C42DBULL, "248196386.ogg"},
	{0x5D6E78459BF0F823ULL, "153796556.ogg"},
	{0x5D6E912FE5505D50ULL, "one280_emrys6104262.cl"},
	{0x5D6F5DFFD484A4BAULL, "vxw_encounter2_73_0_0.vlm"},
	{0x5D6FA7089F0B45ABULL, "vxlt_gallowscourtyard_n_7et_0_0.vlm"},
	{0x5D6FB56023C72867ULL, "hlac_ht_chantry_2534_42z.dds"},
	{0x5D705DD52685902FULL, "llac_ht_mansion2_818_61.dds"},
	{0x5D70CC72A7D2F55DULL, "dae000_orana6149083.cl"},
	{0x5D71354146A7D60BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_check_nails.gad"},
	{0x5D71408E63CA41F4ULL, "22351049.ogg"},
	{0x5D715F9AEBC106F9ULL, "gen00fl_isabela6079087.cl"},
	{0x5D718D0F946EC5F6ULL, "253864671.ogg"},
	{0x5D71B3C0572E9885ULL, "6179759_f.fxe"},
	{0x5D7286620DDF9794ULL, "6171270_m.fxe"},
	{0x5D72F3406FC41A13ULL, "6154009_m.fxe"},
	{0x5D7310E4603EBA0DULL, "58999401.ogg"},
	{0x5D73406D8F2E6762ULL, "vxht_keep_n_5gt_0_0.vlm"},
	{0x5D7368210B78D969ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female.ani"},
	{0x5D73B1944ECBE2B7ULL, "dae000ip_wall_lean_left_s.utp"},
	{0x5D73BA5FBB0AC3C4ULL, "gen00fl_dog6116494.cl"},
	{0x5D74025BB1ACFE75ULL, "smoke04.dds"},
	{0x5D7427B9A3F9083AULL, "6120771_m.fxe"},
	{0x5D7460224CB06BA2ULL, "ser270_templar6149991.cl"},
	{0x5D748992816023B1ULL, "vxw_bonepitmining_a2_1q3_0_0.vlm"},
	{0x5D74F74E390701FBULL, "vxlt_0_lowtown_a2_e87_0_0.vlm"},
	{0x5D75211936BC0689ULL, "hlac_lt_gallowsprison_n_-1_1y.dds"},
	{0x5D75346089FE13CBULL, "hlac_dr_slavers_3217_8z.dds"},
	{0x5D75F7C361CFA1F1ULL, "6100064_m.fxe"},
	{0x5D767CA276C88167ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtb.phy"},
	{0x5D76D2814E1722F6ULL, "6116622_m.fxe"},
	{0x5D7771BFA89F6213ULL, "vxlt_0_lowtown_a2_n_c_gz2_0_0.vlm"},
	{0x5D779D99ACF2BAFFULL, "546416670.ogg"},
	{0x5D77F5173333B9BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.evt"},
	{0x5D7829D5BCF1E2A8ULL, "vxlt_0_lowtown_docks_ba8_0_1.vlm"},
	{0x5D78324146ADF070ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_a_female_check_nails.evt"},
	{0x5D786436D1774DCEULL, "hlac_dr_dungeon_3214_3y.dds"},
	{0x5D788849CE44732DULL, "181149614.ogg"},
	{0x5D78BE868E0495CDULL, "vxht_0_hightown_c8s_0_0.vlm"},
	{0x5D78C31BDBADE055ULL, "vxlt_gallowscourtyard_n_66z_0_0.vlm"},
	{0x5D792D171F744862ULL, "vxdr_slavers_1at_0_0.vlm"},
	{0x5D793F3EB559B9BCULL, "w_bonepitdestroyed_a3_168_0.phy"},
	{0x5D794700D7E6128FULL, "6031651_f.fxe"},
	{0x5D79FEE61E81B37DULL, "6116311_m.fxe"},
	{0x5D7A322CB0E687F2ULL, "221218677.ogg"},
	{0x5D7A60023AA1CEC2ULL, "6080594_m.fxe"},
	{0x5D7A83210B7F25EBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female.gad"},
	{0x5D7A9655EA01B305ULL, "953583648.ogg"},
	{0x5D7B1A8436C5839DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chairxa.mao"},
	{0x5D7B1E8436C58AD6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chairxa.mmh"},
	{0x5D7B288436C59BD4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chairxa.msh"},
	{0x5D7B9C4D660B2762ULL, "6068555_f.fxe"},
	{0x5D7C06FC959BA962ULL, "696639186.ogg"},
	{0x5D7CC58031BB9F73ULL, "llac_ca_caverns2_5802_31.dds"},
	{0x5D7CF87B57D3B4CFULL, "6147720_m.fxe"},
	{0x5D7D136CC788A065ULL, "6098810_f.fxe"},
	{0x5D7D6F02174F5B6CULL, "777931396.ogg"},
	{0x5D7D81DEF6E136ACULL, "hlac_w_bonepitdestroyed_a3_327680_0x.dds"},
	{0x5D7D8C2AB3E40D7BULL, "gen_im_wep_mag_sta_act3_06.uti"},
	{0x5D7D9480CDC4A646ULL, "sst_ss_midclass_thugs_f_1.gda"},
	{0x5D7DACC5740AC1CCULL, "one350_vinell6087345.cl"},
	{0x5D7DF417333823ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_twitch.gad"},
	{0x5D7E1B56460757C8ULL, "vxht_chantry_n_35z_0_0.vlm"},
	{0x5D7E2E40CC1FF144ULL, "843010675.ogg"},
	{0x5D7E5C20FD4CF24AULL, "hlac_lt_0_lowtown_a3_n_c_65537_3x.dds"},
	{0x5D7E7443F903C5CBULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls106a_0n.dds"},
	{0x5D7E7B4E46255F5AULL, "zz_mag_debug6054498.cl"},
	{0x5D7EFC09F452C16EULL, "6110556_m.fxe"},
	{0x5D809389383FE3DAULL, "llac_lt_0_lowtown_docks_a3_n_-3_2.dds"},
	{0x5D809AA8E2ED966DULL, "zz_dae_debug6036586.cl"},
	{0x5D80BD0C21A8271BULL, "vxht_bartrand_n_6q3_0_0.vlm"},
	{0x5D811DEF4C41A0A6ULL, "vxca_caverns2prc_9q3_0_0.vlm"},
	{0x5D8180210B853F50ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female.evt"},
	{0x5D81FA6B43A3F60DULL, "6170870_m.fxe"},
	{0x5D8263529181991CULL, "vxca_cavern1_haunted_dnz_0_0.vlm"},
	{0x5D834F347CC735B0ULL, "gen00fl_aveline6122086.cl"},
	{0x5D837045AA83E93DULL, "vxlt_0_lowtown_a2_n_c_ej6_0_0.vlm"},
	{0x5D8433A09EB2B6B1ULL, "847903549.ogg"},
	{0x5D849B9D01162063ULL, "201976738.ogg"},
	{0x5D858A37B1FBC05FULL, "891949571.ogg"},
	{0x5D85EFBCCC08BE60ULL, "440548928.ogg"},
	{0x5D86A46D3F910368ULL, "hlac_ht_mansion2_818_45y.dds"},
	{0x5D870FA8AF50668FULL, "hlac_lt_0_lowtown_a3_n_c_65535_4z.dds"},
	{0x5D875DF8CE918875ULL, "6106435_m.fxe"},
	{0x5D87A2560DDC03C7ULL, "vxht_0_hightown_a2_bnv_0_0.vlm"},
	{0x5D87A888C69BC295ULL, "gen00fl_aveline6137520.cl"},
	{0x5D87B895B781A68BULL, "826696330.ogg"},
	{0x5D87C153964EC1CCULL, "vxlt_warehouse_n_vp_0_0.vlm"},
	{0x5D8809FC8C03CE38ULL, "982626269.ogg"},
	{0x5D8815995BDFAE0DULL, "hlac_lt_0_lowtown_a2_0_2z.dds"},
	{0x5D8818E2BC8F26B8ULL, "hlac_lt_0_lowtown_docks_-3_2z.dds"},
	{0x5D889512DE31B412ULL, "vxlt_0_lowtown_a2_gwa_0_0.vlm"},
	{0x5D88CD505AC532F9ULL, "lgt110_gxa_vgk_conspirator16117065.cl"},
	{0x5D88D81408391B31ULL, "58073833.ogg"},
	{0x5D894EA9CC3A79D9ULL, "gen00fl_sebastian6177885.cl"},
	{0x5D89801FF9CA6046ULL, "6033383_m.fxe"},
	{0x5D8A0A7574D308B0ULL, "117293986.ogg"},
	{0x5D8A1E05B87769D1ULL, "m_pro000ar_blightlands_fake.wbd"},
	{0x5D8A21C13385BACAULL, "vxlt_0_lowtown_a2_n_c_eml_0_0.vlm"},
	{0x5D8A66AB96672174ULL, "vxht_bartrand_n_64k_0_0.vlm"},
	{0x5D8AC6A983B791EBULL, "hlac_lt_0_lowtown_docks_65537_0y.dds"},
	{0x5D8B35ABE324881DULL, "gen00fl_aveline6139733.cl"},
	{0x5D8B867BBAD0A057ULL, "gen00fl_anders6107669.cl"},
	{0x5D8BB15595B097ABULL, "6154382_m.fxe"},
	{0x5D8C2408104CB75FULL, "719162908.ogg"},
	{0x5D8C247DBA4FF5FBULL, "vxht_0_hightown_n_bz1_0_0.vlm"},
	{0x5D8C77EBDA640F3DULL, "231704235.ogg"},
	{0x5D8CAED48AE22BADULL, "vxlt_0_lowtown_e7x_0_0.vlm"},
	{0x5D8CB12376280845ULL, "158817098.ogg"},
	{0x5D8CCE1F9A895B8AULL, "6120469_m.fxe"},
	{0x5D8CF1276D11645FULL, "6081310_m.fxe"},
	{0x5D8D5F28B5B6B716ULL, "6035302_f.fxe"},
	{0x5D8E75AB0FB5D466ULL, "6147146_f.fxe"},
	{0x5D8E9AF2A24A26A8ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_1y.dds"},
	{0x5D8EDA0EDCF463A4ULL, "490406131.ogg"},
	{0x5D8EDED7228EC542ULL, "6115959_m.fxe"},
	{0x5D8EDF8EAAF19D5FULL, "zz_start_qcr6081927.cl"},
	{0x5D8F004C601B4EBFULL, "vxlt_hangedman_n_o4_0_0.vlm"},
	{0x5D8F3BBA443F73A2ULL, "6161505_m.fxe"},
	{0x5D8F6732F1571958ULL, "6133100_m.fxe"},
	{0x5D8F6BF4E3D27551ULL, "art\\characters\\creatures\\ragedemon\\animation\\electr.ani"},
	{0x5D8FA999D7616AB2ULL, "follower_banter6163494.cl"},
	{0x5D90515D668AB0AEULL, "mrl310_merrill6140758.cl"},
	{0x5D90FFD68FE4CF23ULL, "vxw_encounter2_camp_9l_0_0.vlm"},
	{0x5D92438AF64C8ECAULL, "1064887156.ogg"},
	{0x5D9266245976AC2EULL, "6162584_m.fxe"},
	{0x5D92B7F8579941BFULL, "661117878.ogg"},
	{0x5D933262DDAD7174ULL, "hlac_lt_undercityencounter_-1_19x.dds"},
	{0x5D933B13133BBB74ULL, "6119412_m.fxe"},
	{0x5D93F9B0E42E30DBULL, "w_bonepit_225_0.msh"},
	{0x5D940AA7DD95F7EFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_sloped.mao"},
	{0x5D940AA969A56AF1ULL, "hlac_lt_0_lowtown_docks_-131073_3x.dds"},
	{0x5D940BB0E42E4F26ULL, "w_bonepit_225_0.mao"},
	{0x5D9416A7DD960B8CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_sloped.mmh"},
	{0x5D9417B0E42E63C5ULL, "w_bonepit_225_0.mmh"},
	{0x5D941CA7DD96167EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_sloped.msh"},
	{0x5D943BCD239FDBE0ULL, "250145539.ogg"},
	{0x5D9446FE7CC836C4ULL, "gen00fl_varric6076085.cl"},
	{0x5D945BFB3AC5CED5ULL, "6127078_m.fxe"},
	{0x5D9576F4E3D6F3F3ULL, "art\\characters\\creatures\\ragedemon\\animation\\electr.gad"},
	{0x5D95955BA9037430ULL, "814146312.ogg"},
	{0x5D95A24071F0C6CDULL, "vxht_0_hightown_c2q_0_0.vlm"},
	{0x5D95E86F9F282E6BULL, "vxht_bartrand_n_6ik_0_0.vlm"},
	{0x5D96D762C0A46961ULL, "hlac_lt_0_lowtown_n_2_1z.dds"},
	{0x5D96FE15E23E78D0ULL, "416815955.ogg"},
	{0x5D9706EFD027CC92ULL, "vxlt_0_lowtown_a2_gjb_0_0.vlm"},
	{0x5D971266F8556A2EULL, "vxht_0_hightown_a3_c1v_0_0.vlm"},
	{0x5D97A80D1918EFEEULL, "hlac_ht_mansion_n_3381_1z.dds"},
	{0x5D988CCF3B83D848ULL, "mer140_bonepit_comments.cnv"},
	{0x5D98B9CECC22F07CULL, "vxw_encounter2_kq_0_0.vlm"},
	{0x5D9924D50B51B406ULL, "vxlt_gallowscourtyard_6zr_0_0.vlm"},
	{0x5D993632657576A2ULL, "vxht_0_hightown_9h2_0_0.vlm"},
	{0x5D993A1CA82C0C60ULL, "994271633.ogg"},
	{0x5D998E232FE244A6ULL, "758659037.ogg"},
	{0x5D9A016F41AAF7D5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\14gallowbird.msh"},
	{0x5D9A036F41AAFB4BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\14gallowbird.mmh"},
	{0x5D9A0F6F41AB0FE8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\14gallowbird.mao"},
	{0x5D9A1799400B8FEBULL, "follower_banter6165406.cl"},
	{0x5D9A59456CAC3355ULL, "darktown_fem_lowlife_mob6174165.cl"},
	{0x5D9A6FBC25EEF295ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_catacombs.phy"},
	{0x5D9A965AAC460E7BULL, "vxht_0_hightown_n_be0_0_0.vlm"},
	{0x5D9AAB02B1520C82ULL, "236538883.ogg"},
	{0x5D9AC029927ABD18ULL, "696361366.ogg"},
	{0x5D9AC842B202AEA2ULL, "vxht_mansion_n_33v_0_0.vlm"},
	{0x5D9AE88D9F8CFEF2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_flat.phy"},
	{0x5D9B0698EB094F2CULL, "gen00fl_fenris6115346.cl"},
	{0x5D9B075ABB3682E6ULL, "face_face_p23.psh.fxo.dx11"},
	{0x5D9B20B17CD8442DULL, "vxlt_0_lowtown_e2w_0_0.vlm"},
	{0x5D9B942A3AE668D7ULL, "vxlt_darktown_5br_0_0.vlm"},
	{0x5D9BAECF3B8647B0ULL, "mer140_bonepit_comments.bnk"},
	{0x5D9C485F9AB5A262ULL, "242266181.ogg"},
	{0x5D9C49C4D48675C1ULL, "follower_banter6160850.cl"},
	{0x5D9C73F4E3DD0D58ULL, "art\\characters\\creatures\\ragedemon\\animation\\electr.evt"},
	{0x5D9CD52849B00B2BULL, "hlac_ht_0_hightown_131070_0z.dds"},
	{0x5D9D0663B21B8297ULL, "408310977.ogg"},
	{0x5D9D83934E6CF3CCULL, "323594578.ogg"},
	{0x5D9DEAB546C3B06BULL, "106181071.ogg"},
	{0x5D9E3B28EE63011CULL, "427208731.ogg"},
	{0x5D9E3C87D3D026DAULL, "310107008.ogg"},
	{0x5D9E4C1D5F1756C3ULL, "6176522_m.fxe"},
	{0x5D9EE6136CAB233CULL, "lt_0_lowtown_n_45_0.rml"},
	{0x5D9F71B381DAE7A3ULL, "vxw_encounter2_camp_4o_0_0.vlm"},
	{0x5D9F848790B1EE9EULL, "hlac_lt_gallowstemplar_f_2393_0z.dds"},
	{0x5D9FA93DF6AC2C7AULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_movement.anb"},
	{0x5DA03C99DD4EC357ULL, "6034048_m.fxe"},
	{0x5DA04E33F857203CULL, "444664358.ogg"},
	{0x5DA09EC08641BBA6ULL, "gen_im_npc_wep_two_2hab.uti"},
	{0x5DA19E07EAED5FC6ULL, "hm_gencr_raider_assassin.mor"},
	{0x5DA19E2194435AD9ULL, "art\\characters\\playercharacter\\children\\km_har_ha01a.msh"},
	{0x5DA1A82194436BC7ULL, "art\\characters\\playercharacter\\children\\km_har_ha01a.mmh"},
	{0x5DA20FCD8F624F21ULL, "139296742.ogg"},
	{0x5DA25D4FFEDBBC05ULL, "hlac_lt_darktown_-1_63z.dds"},
	{0x5DA27789471FA271ULL, "6166558_m.fxe"},
	{0x5DA2D5748033E9B9ULL, "6156467_m.fxe"},
	{0x5DA2DA5B03E87596ULL, "one100dg_vanard6048258.cl"},
	{0x5DA3082F0C1FF4C1ULL, "6174070_m.fxe"},
	{0x5DA31DB11ED111F3ULL, "vxht_bartrand_n_5wb_0_0.vlm"},
	{0x5DA394EE10262051ULL, "6027038_f.fxe"},
	{0x5DA417CF627589B0ULL, "zz_fakemode.ncs"},
	{0x5DA445BD1108145EULL, "vxlt_0_lowtown_n_fdz_0_0.vlm"},
	{0x5DA45A4C911E46EBULL, "vxht_bartrand_n_6rj_0_0.vlm"},
	{0x5DA464267F6623DFULL, "hlac_lt_0_lowtown_docks_n_-4_7x.dds"},
	{0x5DA4BB50F502C203ULL, "vxca_caverns2prc_aev_0_0.vlm"},
	{0x5DA561E613D13214ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_rail.phy"},
	{0x5DA590600DDA5D26ULL, "6209901_f.fxe"},
	{0x5DA594DECB7DBFDCULL, "pro000_captain_wright6047485.cl"},
	{0x5DA5DD1AE4781EF3ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobd_0n.dds"},
	{0x5DA6066A41DF3344ULL, "w_encounter3_168_0.msh"},
	{0x5DA6186A41DF51ADULL, "w_encounter3_168_0.mao"},
	{0x5DA61C6A41DF58E6ULL, "w_encounter3_168_0.mmh"},
	{0x5DA6208C9F018037ULL, "w_encounter3_125_0.mmh"},
	{0x5DA6248C9F018744ULL, "w_encounter3_125_0.mao"},
	{0x5DA6368C9F01A5A9ULL, "w_encounter3_125_0.msh"},
	{0x5DA678E24966A6CDULL, "hlac_lt_undercityencounter_-1_43z.dds"},
	{0x5DA67C490ECC9A1CULL, "49064957.ogg"},
	{0x5DA6F2BE96F6C290ULL, "790385046.ogg"},
	{0x5DA71BAD42E3BE13ULL, "zz_mag_debug6025592.cl"},
	{0x5DA71FBD9CEEA84EULL, "lt_0_lowtown_docks_a3_171_0.msh"},
	{0x5DA724E2BDF5F218ULL, "6077752_m.fxe"},
	{0x5DA72DBD9CEEBF9FULL, "lt_0_lowtown_docks_a3_171_0.mao"},
	{0x5DA739BD9CEED43CULL, "lt_0_lowtown_docks_a3_171_0.mmh"},
	{0x5DA73AD684650A31ULL, "616332275.ogg"},
	{0x5DA7669F0A068987ULL, "275108198.ogg"},
	{0x5DA8B538194817DDULL, "286208375.ogg"},
	{0x5DA978CFC6A13B3AULL, "vxw_blightlands_o4_0_0.vlm"},
	{0x5DA997FC86982976ULL, "vxca_cavern1_haunted_gkp_0_0.vlm"},
	{0x5DAA7EF0FFBC13FDULL, "6153594_m.fxe"},
	{0x5DABAE16F8D31866ULL, "vxw_blightlands_ez_0_0.vlm"},
	{0x5DABC7F0AC4D8016ULL, "6095963_m.fxe"},
	{0x5DAC0083DFB3FB16ULL, "vxca_cavern1_haunted_iny_0_0.vlm"},
	{0x5DAC133D3567986EULL, "hlac_lt_0_lowtown_a3_n_c_131071_6z.dds"},
	{0x5DAC4AECD5434A58ULL, "6139700_m.fxe"},
	{0x5DAC59863F6547A0ULL, "vxlt_0_lowtown_a2_n_c_d1w_0_0.vlm"},
	{0x5DAC65D11BDDBC27ULL, "hlac_ca_caverns1_24427_1x.dds"},
	{0x5DAC6AB433D2D3F9ULL, "685724816.ogg"},
	{0x5DAC73862223EA26ULL, "vxca_caverns2prc_9f6_0_0.vlm"},
	{0x5DAC766B39782883ULL, "vxlt_0_lowtown_a2_n_c_hnm_0_0.vlm"},
	{0x5DAD2A8709573924ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_bdy_suca_0s.dds"},
	{0x5DAD81CDAF4264BFULL, "vxlt_backalley_n_1n0_0_1.vlm"},
	{0x5DADAA593D150978ULL, "606601438.ogg"},
	{0x5DAEAE76CF44DBD7ULL, "6031348_m.fxe"},
	{0x5DAEF277219B7476ULL, "ico_node.dds"},
	{0x5DAEF76C33E46B1CULL, "6044866_m.fxe"},
	{0x5DAF223CC1CFD25AULL, "llac_lt_darktown_-1_31.dds"},
	{0x5DAF31E43ADCD161ULL, "vxca_cavern1_haunted_j2a_0_0.vlm"},
	{0x5DAF3EE613D94B11ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_rail.msh"},
	{0x5DAF4CE613D9632CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_rail.mao"},
	{0x5DAF58E613D977CFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_rail.mmh"},
	{0x5DAFAF513B685DC5ULL, "hlac_w_sundermount_-5_0x.dds"},
	{0x5DAFFA3061DDB5C7ULL, "vxlt_0_lowtown_a2_heq_0_0.vlm"},
	{0x5DB00EA6311F0FA5ULL, "vxdr_0_deeproads_esf_0_1.vlm"},
	{0x5DB0528D8B65511CULL, "311136201.ogg"},
	{0x5DB086C5311BB18BULL, "440316720.ogg"},
	{0x5DB1496665D2D5FCULL, "hlac_ht_mansion_n_3370_2y.dds"},
	{0x5DB211BAA3C2BC40ULL, "898639140.ogg"},
	{0x5DB21336E2954EF4ULL, "framebuffereffects_dofblur_p0.psh.fxo.dx11"},
	{0x5DB22831B6A0D27DULL, "mrl350ip_shrine.ncs"},
	{0x5DB27E3AA19F7C71ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatestatues.msh"},
	{0x5DB28C3AA19F938CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatestatues.mao"},
	{0x5DB2983AA19FA82FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatestatues.mmh"},
	{0x5DB337B60BB23BCCULL, "140927661.ogg"},
	{0x5DB341EEE7CE73DAULL, "hlac_ca_caverns1_mines_24426_9y.dds"},
	{0x5DB3F28D5199E11DULL, "ts_kirk_guards_01.gda"},
	{0x5DB3F3B95FA7B00CULL, "vxca_cave_1o5_0_0.vlm"},
	{0x5DB4CF58D8C5B732ULL, "gen00fl_merrill6151111.cl"},
	{0x5DB5974E6C151D46ULL, "747728913.ogg"},
	{0x5DB598DC680450DCULL, "533821697.ogg"},
	{0x5DB59D88B00F217CULL, "vxw_encounter2_qw_0_0.vlm"},
	{0x5DB5C3D74A5BB71DULL, "hlac_ht_bartrand_n_3373_1y.dds"},
	{0x5DB6652BB4DA2313ULL, "227742573.ogg"},
	{0x5DB665BB69C1C4A4ULL, "mag321_comtesse_de_launcet6042794.cl"},
	{0x5DB67434CBD0509DULL, "dre000_cut_bartrandbetrayal6029373.cl"},
	{0x5DB6AF4E7C7BBD40ULL, "vxlt_foundry_n_v2_0_0.vlm"},
	{0x5DB6C1AD4B11B426ULL, "lt_0_lowtown_docks_n_123_0.phy"},
	{0x5DB6D1AF2815CB2EULL, "6083991_m.fxe"},
	{0x5DB6E79E159A02D8ULL, "mag100_cullen6086365.cl"},
	{0x5DB6FEEBA7BAF1D2ULL, "pro000_guard6005644.cl"},
	{0x5DB71F6F08BD4FF5ULL, "and231tr_spiders.ncs"},
	{0x5DB752EF7F9E87BFULL, "cod_mgc_chant_structure.utp"},
	{0x5DB764F47DF2D01AULL, "vxlt_gallowstemplar_n_5ux_0_1.vlm"},
	{0x5DB791A79B14CE0CULL, "vxlt_0_lowtown_h9v_0_0.vlm"},
	{0x5DB7CABFFD5877DFULL, "art\\characters\\creatures\\dog\\textures\\c_eye_doga_0d.dds"},
	{0x5DB7EAACB89753BAULL, "vxw_blightlands_p5_0_0.vlm"},
	{0x5DB866FD5135A013ULL, "835610493.ogg"},
	{0x5DB89BEACD509C71ULL, "189808658.ogg"},
	{0x5DB8BA99FC317D48ULL, "vxca_cavern1_haunted_j2d_0_0.vlm"},
	{0x5DB8C12755834FC7ULL, "6195860_m.fxe"},
	{0x5DB90D8C6FAAE684ULL, "art\\levels\\races\\design\\prp_orichaicum.mmh"},
	{0x5DB9118C6FAAEDD7ULL, "art\\levels\\races\\design\\prp_orichaicum.mao"},
	{0x5DB91EECE2966384ULL, "vxx_partypicker_5c_0_0.vlm"},
	{0x5DB9238C6FAB0C26ULL, "art\\levels\\races\\design\\prp_orichaicum.msh"},
	{0x5DB970ED526164DFULL, "6027484_m.fxe"},
	{0x5DB9861C5F620C5EULL, "6051390_m.fxe"},
	{0x5DB99BE33CBFB276ULL, "ser221_gascard6111749.cl"},
	{0x5DBA4E05794089E3ULL, "fen200ar_slavers_exterior.are"},
	{0x5DBA6CD92D1E7721ULL, "hlac_lt_0_lowtown_n_1_0x.dds"},
	{0x5DBACB63315B6020ULL, "vxlt_0_lowtown_a2_n_c_d4t_0_0.vlm"},
	{0x5DBB4695EFE8540FULL, "117107462.ogg"},
	{0x5DBB52EDA474CE19ULL, "vxht_playershouse_n_18o_0_0.vlm"},
	{0x5DBBA4F021AF1B73ULL, "6127115_m.fxe"},
	{0x5DBCD6E4CF9363BEULL, "498828425.ogg"},
	{0x5DBD55F0A2C9701FULL, "913938363.ogg"},
	{0x5DBD7532861DBE05ULL, "6144381_m.fxe"},
	{0x5DBD983405CCF520ULL, "gen00fl_aveline6120826.cl"},
	{0x5DBDB0B4C3439543ULL, "vxdr_primevaltomb_3jn_0_0.vlm"},
	{0x5DBDD4733C2B183EULL, "244225331.ogg"},
	{0x5DBE1556EE31E7FAULL, "6154930_m.fxe"},
	{0x5DBE6C0D53D3CE47ULL, "vxlt_0_lowtown_a2_hbp_0_0.vlm"},
	{0x5DBEB067427AD861ULL, "spt_branch.vsh.fxo"},
	{0x5DBF6B9893AA4F44ULL, "vxht_0_hightown_a2_cbh_0_0.vlm"},
	{0x5DBF6C90B6D03F50ULL, "6096326_m.fxe"},
	{0x5DBFA382B43B85AAULL, "953091254.ogg"},
	{0x5DC0B591AFEFA73EULL, "mag100_karras6034541.cl"},
	{0x5DC1AE6171B66907ULL, "mrl310_merrill.bnk"},
	{0x5DC1F087FD9CBEF9ULL, "one130_thug6131601.cl"},
	{0x5DC2801FDA0A4A36ULL, "mag210_arianni6024562.cl"},
	{0x5DC2B21549953EC2ULL, "gen00fl_merrill6108467.cl"},
	{0x5DC2C96B9A48840AULL, "vxdr_slavers_1b1_0_0.vlm"},
	{0x5DC328B74A26C14EULL, "dae000ip_crowd_ff5_talk8.utp"},
	{0x5DC3C2B7FE8A76DFULL, "627245306.ogg"},
	{0x5DC3E934FB5E1CCAULL, "vxlt_0_lowtown_a2_n_c_eqh_0_0.vlm"},
	{0x5DC41E019E75E374ULL, "598102843.ogg"},
	{0x5DC42CAD6A2C3354ULL, "472142958.ogg"},
	{0x5DC49D054104C487ULL, "6143063_m.fxe"},
	{0x5DC4D06171B8D97FULL, "mrl310_merrill.cnv"},
	{0x5DC510F406C84F6AULL, "vxht_0_hightown_a3_9za_0_0.vlm"},
	{0x5DC5167EF6DACFC5ULL, "65106012.ogg"},
	{0x5DC51F83B2DD01A3ULL, "vxca_caverns1_j3q_0_0.vlm"},
	{0x5DC56E75131DEF29ULL, "vxw_sundermount_22l_0_0.vlm"},
	{0x5DC5E076026ED820ULL, "vxw_sundermount_2a0_0_0.vlm"},
	{0x5DC5EA426370F56EULL, "6045786_m.fxe"},
	{0x5DC611AA351E273EULL, "98362042.ogg"},
	{0x5DC6314A6B10840FULL, "llac_lt_warehouse_65536_1.dds"},
	{0x5DC65C89AA8D6C3AULL, "vxw_blightlands_u6_0_0.vlm"},
	{0x5DC689F79B5A11C8ULL, "6033381_m.fxe"},
	{0x5DC690FACF0B5C84ULL, "6083346_m.fxe"},
	{0x5DC774B28492B994ULL, "art\\vfx\\materials\\vfx_crack_decal_add.mao"},
	{0x5DC840B31812B934ULL, "llac_ht_keep_3732_6.dds"},
	{0x5DC86A08C6B186BCULL, "610466751.ogg"},
	{0x5DC86ECB1B5CAE70ULL, "gen00fl_varric6078128.cl"},
	{0x5DC88730EF205D62ULL, "dae000_porfiria6155305.cl"},
	{0x5DC8ACDA6CC80023ULL, "hlac_lt_gallowstemplar_f_2391_5z.dds"},
	{0x5DC8DDF4289123D8ULL, "w_woundedcoast_probe_1ng.dds"},
	{0x5DC9574006101B26ULL, "vxca_caverns2prc_9h8_0_0.vlm"},
	{0x5DC9BCC3C7FB8CA8ULL, "vxdr_0_deeproads_fbr_0_0.vlm"},
	{0x5DC9DA6E4C1F642DULL, "116450761.ogg"},
	{0x5DC9F6E53C0E961CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hugekeg_n.dds"},
	{0x5DCA1335A69C3D59ULL, "642647833.ogg"},
	{0x5DCA285758DA3DFBULL, "mnd_imp_cycle.mao"},
	{0x5DCAE7136CD05F8CULL, "lt_0_lowtown_n_45_0.anb"},
	{0x5DCB69142011D0BEULL, "dae000_epic_merchant_26179375.cl"},
	{0x5DCB73E3D065B1D7ULL, "6206869_m.fxe"},
	{0x5DCBA996645263BDULL, "vxlt_0_lowtown_a2_n_c_es1_0_0.vlm"},
	{0x5DCC45A7E95A31FDULL, "dre000_varric6037011.cl"},
	{0x5DCC884DC430EC12ULL, "02_outersparks.mao"},
	{0x5DCC915B62CCBFFEULL, "vxdr_primevaltomb_4i1_0_0.vlm"},
	{0x5DCCAA0566A8FDC2ULL, "266710237.ogg"},
	{0x5DCCC5F0B1E8DF0DULL, "one100dg_lia6049829.cl"},
	{0x5DCCE992D38B0A9BULL, "dae000_oldmangossip6174925.cl"},
	{0x5DCD3045263DF107ULL, "art\\characters\\creatures\\desiredemon\\animation\\buffactivate.gad"},
	{0x5DCD650E681EEEA6ULL, "196260402.ogg"},
	{0x5DCD9C8B25566938ULL, "6110864_m.fxe"},
	{0x5DCDFC87E77E94AAULL, "347147185.ogg"},
	{0x5DCE00FA599F1C91ULL, "vxw_bonepitdestroyed_a3_no_0_0.vlm"},
	{0x5DCE204AD7C77488ULL, "zz_approval_debug6003841.cl"},
	{0x5DCE2573AF3CC67BULL, "567539530.ogg"},
	{0x5DCE2C700846F4C7ULL, "gen00fl_dog6167657.cl"},
	{0x5DCE9121946A32FCULL, "art\\characters\\playercharacter\\children\\km_har_ha01a.phy"},
	{0x5DCE9FDF1F40CE54ULL, "6093798_m.fxe"},
	{0x5DCEF22AEB357E36ULL, "498875580.ogg"},
	{0x5DCF00C924842830ULL, "6165158_m.fxe"},
	{0x5DCF2B5E809617C1ULL, "gen00fl_dog6169323.cl"},
	{0x5DCF35425759C33DULL, "315637875.ogg"},
	{0x5DCF6DB9CEB3B30AULL, "hlac_lt_0_lowtown_0_1z.dds"},
	{0x5DCF71D0BA0917BBULL, "6161629_m.fxe"},
	{0x5DCF8040C9F60E15ULL, "vxlt_foundry_n_t9_0_0.vlm"},
	{0x5DCFA54FFA72F705ULL, "vxlt_0_lowtown_a2_fi5_0_0.vlm"},
	{0x5DD0058D1E6F5600ULL, "gen_im_trs_krk_arc_sbw_r2.uti"},
	{0x5DD01FEC1CE882CBULL, "vxlt_gallowstemplar_3qq_0_0.vlm"},
	{0x5DD0BB2704CA7451ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closed.ani"},
	{0x5DD1B546D5117E89ULL, "247626424.ogg"},
	{0x5DD1C2427E77305CULL, "6180519_m.fxe"},
	{0x5DD1F5DB19096389ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffy2low_n.dds"},
	{0x5DD20D27AF7676C6ULL, "221541443.ogg"},
	{0x5DD22A4B767A6A94ULL, "vxlt_0_lowtown_n_ef5_0_0.vlm"},
	{0x5DD255A2C985FB55ULL, "656166135.ogg"},
	{0x5DD3448C3031622BULL, "vxlt_0_lowtown_a2_e2e_0_0.vlm"},
	{0x5DD3E052051407A9ULL, "vxw_sundermount_27s_0_0.vlm"},
	{0x5DD425452643FD14ULL, "art\\characters\\creatures\\desiredemon\\animation\\buffactivate.evt"},
	{0x5DD4A858B150934CULL, "vxlt_darktown_4mm_0_0.vlm"},
	{0x5DD54F40EA1A4531ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin_1.mao"},
	{0x5DD558C2BE0AF6E0ULL, "537304903.ogg"},
	{0x5DD58B32188FF04CULL, "786226753.ogg"},
	{0x5DD5F5846A5FFA2EULL, "603531615.ogg"},
	{0x5DD633CE1213C6C1ULL, "ser191_emeric6164746.cl"},
	{0x5DD63E361B1AAA49ULL, "cod_hst_kirk_alienage.utp"},
	{0x5DD66CCC2D8286AFULL, "870481474.ogg"},
	{0x5DD6C62704CEF2F3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closed.gad"},
	{0x5DD6EB120C81A1C0ULL, "vxlt_gallowstemplar_f_5x8_0_0.vlm"},
	{0x5DD73F6A00515827ULL, "vxca_cavern1_haunted_hi0_0_0.vlm"},
	{0x5DD789D5612E394CULL, "hlac_lt_0_lowtown_-131072_6z.dds"},
	{0x5DD7A66020756119ULL, "500134736.ogg"},
	{0x5DD7D4B754C7D8C9ULL, "lt_gallowscourtyard_43_0_lc.mmh"},
	{0x5DD887548FDA6B3AULL, "character_character_p164.psh.fxo.dx11"},
	{0x5DD89315FDB3D37AULL, "dots.mao"},
	{0x5DD8C0FC4C2BD8AAULL, "hlac_lt_backalley_g_0_5y.dds"},
	{0x5DD8E700E0D5C623ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.dura.evt"},
	{0x5DD919E7C157CFE3ULL, "6096220_m.fxe"},
	{0x5DD936740B4D2596ULL, "6082615_m.fxe"},
	{0x5DDA1B88178579DEULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrla_0d.dds"},
	{0x5DDB2187E6F9A92CULL, "6077070_f.fxe"},
	{0x5DDB22475E7F362EULL, "gen_im_arm_hlm_med_act2_02.uti"},
	{0x5DDBF77FEF9A1712ULL, "ave310_brennan6136878.cl"},
	{0x5DDC5B1E9F63BF52ULL, "191969803.ogg"},
	{0x5DDC5E66D799AF12ULL, "333130454.ogg"},
	{0x5DDC74E9E7D4D6D9ULL, "44870817.ogg"},
	{0x5DDCAB70828F73A5ULL, "vxlt_darktown_6oh_0_0.vlm"},
	{0x5DDD162852CF599AULL, "vxht_playershouse_n_1jz_0_0.vlm"},
	{0x5DDD26438F863D32ULL, "6124371_m.fxe"},
	{0x5DDD4A6D7B4AF7CFULL, "329218123.ogg"},
	{0x5DDDB8D3DBDBE2E2ULL, "vxw_bonepit_25e_0_0.vlm"},
	{0x5DDDE3ADABFF303BULL, "6166729_m.fxe"},
	{0x5DDE21E7901281E0ULL, "vxca_caverns1_mines_e4m_0_0.vlm"},
	{0x5DDE6DECC62D39C2ULL, "ttx_floor_large_1.msh"},
	{0x5DDE6FECC62D3D78ULL, "ttx_floor_large_1.mmh"},
	{0x5DDF12365656C314ULL, "139918814.ogg"},
	{0x5DDF353A8735CA8CULL, "framebuffereffects_twopassblur_p0.psh.fxo.dx11.assembly"},
	{0x5DDF35580573E046ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_deeproads_e.dds"},
	{0x5DDFC000E0DBA2A4ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.dura.gad"},
	{0x5DE00D260B6FD39EULL, "225824283.ogg"},
	{0x5DE030778BCD7E24ULL, "793704378.ogg"},
	{0x5DE0482E76492D99ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv3.ani"},
	{0x5DE04E5753116980ULL, "llac_lt_backalley_g_0_6.dds"},
	{0x5DE083787AB8C6CEULL, "419508124.ogg"},
	{0x5DE0E6BD9864F884ULL, "hlac_ca_cavern1_haunted_24426_9y.dds"},
	{0x5DE10B9D6DB5B96DULL, "335832788.ogg"},
	{0x5DE182BC262B15B0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_catacombs.msh"},
	{0x5DE190BC262B2DCAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_catacombs.mmh"},
	{0x5DE194BC262B3401ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_catacombs.mao"},
	{0x5DE1B7A7FD8D53CCULL, "hlac_w_sundermount_327685_0x.dds"},
	{0x5DE20E8BD61D4A0EULL, "65864434.ogg"},
	{0x5DE21D4526502D4DULL, "art\\characters\\creatures\\desiredemon\\animation\\buffactivate.ani"},
	{0x5DE2522EF70A2029ULL, "vxw_sundermount_28r_0_0.vlm"},
	{0x5DE2636659FAC033ULL, "dae200_saemus6091640.cl"},
	{0x5DE3137C400ACD7CULL, "hlac_lt_backalley_g_1_5x.dds"},
	{0x5DE35561ADEDDE2BULL, "vxw_encounter3_dx_0_0.vlm"},
	{0x5DE36DB17F4642C8ULL, "6039181_m.fxe"},
	{0x5DE3BB732B50AA70ULL, "lgt220_mag_apost_apost26156740.cl"},
	{0x5DE3EE3A6CFA8A9DULL, "vxw_sundermount_31m_0_1.vlm"},
	{0x5DE40BBE2D15C71AULL, "w_bonepitdestroyed_a3_248_0.mmh"},
	{0x5DE40FBE2D15CE51ULL, "w_bonepitdestroyed_a3_248_0.mao"},
	{0x5DE41DBE2D15E660ULL, "w_bonepitdestroyed_a3_248_0.msh"},
	{0x5DE450855CE8D710ULL, "408028635.ogg"},
	{0x5DE472903F5B6133ULL, "dae000_sketch.bnk"},
	{0x5DE4EBC2F569ED64ULL, "hlac_w_encounter1_-131067_0z.dds"},
	{0x5DE55CEEFE77BA40ULL, "vxlt_gallowstemplar_f_5c9_0_0.vlm"},
	{0x5DE5CEBA0773A65CULL, "6120860_f.fxe"},
	{0x5DE64364E2380F21ULL, "2177716.ogg"},
	{0x5DE657057965D3C8ULL, "fen200ar_slavers_exterior.lst"},
	{0x5DE6A07DABE7BDA8ULL, "vxdr_0_deeproads_fhp_0_0.vlm"},
	{0x5DE6A992975148F9ULL, "llac_dr_primevaltomb_9621_3.dds"},
	{0x5DE6B07B67DD2738ULL, "6169918_m.fxe"},
	{0x5DE6C575E3E7D79EULL, "w_encounter2_13_0_lc.mmh"},
	{0x5DE6CF00E0E1DA82ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.dura.ani"},
	{0x5DE6F6F7FAA2785AULL, "293793052.ogg"},
	{0x5DE6F9C07084B718ULL, "lowtown_male_alienage1_sta6162806.cl"},
	{0x5DE7107AF74EC037ULL, "demo_montage_f.cut"},
	{0x5DE7232E764F0DDBULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv3.gad"},
	{0x5DE729FCC4E549FAULL, "art\\levels\\ztools\\default\\generic\\9pp_pole.mmh"},
	{0x5DE72DFCC4E55031ULL, "art\\levels\\ztools\\default\\generic\\9pp_pole.mao"},
	{0x5DE73BFCC4E56840ULL, "art\\levels\\ztools\\default\\generic\\9pp_pole.msh"},
	{0x5DE751B206DDF4A4ULL, "vxca_cavern1_haunted_ik8_0_0.vlm"},
	{0x5DE794903F5DD083ULL, "dae000_sketch.cnv"},
	{0x5DE7A850FD433E16ULL, "6085715_m.fxe"},
	{0x5DE7AE496496538AULL, "pro000_varric6179439.cl"},
	{0x5DE81D77F01A343CULL, "hlac_ca_caverns1_24423_11x.dds"},
	{0x5DE835DB2C8DA228ULL, "w_bonepitdestroyed_a3_62_0.msh"},
	{0x5DE847DB2C8DC0B9ULL, "w_bonepitdestroyed_a3_62_0.mao"},
	{0x5DE853DB2C8DD552ULL, "w_bonepitdestroyed_a3_62_0.mmh"},
	{0x5DE873ED6F53C669ULL, "514616341.ogg"},
	{0x5DE876ECC6359D03ULL, "ttx_floor_large_1.phy"},
	{0x5DE8BE9AECB917BBULL, "w_bonepitdestroyed_a3_29_0_lc.mmh"},
	{0x5DE9048F49F82331ULL, "mcr370cr_corpse_1.utc"},
	{0x5DE99E5D2168B494ULL, "6171403_m.fxe"},
	{0x5DE9DCAE919BDE29ULL, "174545702.ogg"},
	{0x5DEAA38CDBBF66D2ULL, "6183970_m.fxe"},
	{0x5DEAD2555976A742ULL, "hlac_w_blightlands_65539_0y.dds"},
	{0x5DEADF5C7216ADF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die_loop.evt"},
	{0x5DEB18D13DA426BFULL, "dae300_alistair_king6134916.cl"},
	{0x5DEB1F6EC31D7B0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_hurloks.evt"},
	{0x5DEB2615DB8BB260ULL, "592769192.ogg"},
	{0x5DEB5701D89F8C73ULL, "vxht_bartrand_n_5ni_0_0.vlm"},
	{0x5DEB8EE33F1D639CULL, "hlac_w_bonepitdestroyed_a3_262149_0z.dds"},
	{0x5DEBA041C60F09B3ULL, "182240165.ogg"},
	{0x5DEBBDCCEF03DB69ULL, "one350_terath6087115.cl"},
	{0x5DEC20616A900A8EULL, "dae011_gossips6149739.cl"},
	{0x5DEC2AB0CDD1FB62ULL, "vxw_bonepit_20f_0_0.vlm"},
	{0x5DEC9FF0A32F1B7CULL, "vxht_brothel_n_6gf_0_0.vlm"},
	{0x5DECC09C70A619F7ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.dura.gad"},
	{0x5DED3BE25DF0B3DEULL, "1068405129.ogg"},
	{0x5DED4A9079554809ULL, "vxlt_undercitydungeon_192_0_0.vlm"},
	{0x5DED5705796BF219ULL, "fen200ar_slavers_exterior.ncs"},
	{0x5DEE402E76555DC0ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv3.evt"},
	{0x5DEE4D3002C51949ULL, "369577228.ogg"},
	{0x5DEF64A59BA7DFB1ULL, "6155682_m.fxe"},
	{0x5DEF70C422597885ULL, "llac_ht_0_hightown_n_-65538_1.dds"},
	{0x5DEF72124D51C591ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_1z.dds"},
	{0x5DEF92E4BB746C1AULL, "vxw_bonepitmining_a2_1p7_0_0.vlm"},
	{0x5DF023D35722EBE4ULL, "gen00fl_sebastian6179596.cl"},
	{0x5DF02846141D932BULL, "vxlt_0_lowtown_a2_e8g_0_0.vlm"},
	{0x5DF0D65C721B0981ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die_loop.gad"},
	{0x5DF132A8E56E0614ULL, "lgt320ar_gang_hideout_ht.are"},
	{0x5DF1B04EE16D92ECULL, "blank_exterior_0_0.msh"},
	{0x5DF1B64EE16D9CDEULL, "blank_exterior_0_0.mmh"},
	{0x5DF1C24EE16DB175ULL, "blank_exterior_0_0.mao"},
	{0x5DF1D8F8A3857708ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.anb"},
	{0x5DF200A5D31F89EEULL, "m_sib131ar_family_estate.wbd"},
	{0x5DF2166EC3238AB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_hurloks.gad"},
	{0x5DF28CB1972F2FE5ULL, "vxlt_foundry_n_yj_0_0.vlm"},
	{0x5DF2A0468710942BULL, "vxca_caverns1_in1_0_0.vlm"},
	{0x5DF2A549D27AA5E0ULL, "rand_forest.gda"},
	{0x5DF3D59C70AC5B84ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.dura.evt"},
	{0x5DF3E12FD36694E2ULL, "vxdr_primevaltomb_8q0_0_1.vlm"},
	{0x5DF3EA6D9CCA27C3ULL, "797297091.ogg"},
	{0x5DF44558C9CD85D0ULL, "494395695.ogg"},
	{0x5DF509A7C4E352E6ULL, "vxw_sundermount_2mb_0_0.vlm"},
	{0x5DF51F3083036902ULL, "6169417_m.fxe"},
	{0x5DF5346F861C08A8ULL, "zz_followers_banter6158706.cl"},
	{0x5DF5A667553C6146ULL, "sse_ss_templars_m_1.gda"},
	{0x5DF5C549E42AD768ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_win.msh"},
	{0x5DF5D749E42AF5F9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_win.mao"},
	{0x5DF5E349E42B0992ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_win.mmh"},
	{0x5DF5E7BB094324F9ULL, "hlac_dr_primevaltomb_11985_3x.dds"},
	{0x5DF6EF7F1E357C0AULL, "llac_w_encounter2_camp_327680_0.dds"},
	{0x5DF6EFB2716E0161ULL, "6096165_m.fxe"},
	{0x5DF7240E13294896ULL, "vxlt_gallowscourtyard_7cl_0_0.vlm"},
	{0x5DF7244DF3EC967DULL, "vxht_mansion2_2au_0_0.vlm"},
	{0x5DF76C15FDCE412CULL, "dots.dds"},
	{0x5DF7ACA8D4849643ULL, "hlac_w_woundedcoast_0_11z.dds"},
	{0x5DF7B3324FED59C4ULL, "138859183.ogg"},
	{0x5DF7D75C72212A2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_walk_and_die_loop.ani"},
	{0x5DF8103A135ECD14ULL, "6174703_m.fxe"},
	{0x5DF9276EC329C607ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage_hurloks.ani"},
	{0x5DFA04F294FC9533ULL, "hla_ht_0_hightown_n_-65534_0z.dds"},
	{0x5DFA1931B6DE6E44ULL, "mrl350ip_shrine.utp"},
	{0x5DFA434BF3EF9C4EULL, "337422775.ogg"},
	{0x5DFA574823913A21ULL, "qun120_arishok_guard6054849.cl"},
	{0x5DFA9558C0831A12ULL, "6166966_m.fxe"},
	{0x5DFBE4A9EBBB2BD2ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_combat_runf.gad"},
	{0x5DFC72E9AFB3DA60ULL, "vxlt_hangedman_28e_0_0.vlm"},
	{0x5DFCA32519F55212ULL, "749956927.ogg"},
	{0x5DFD1449B9C46C2FULL, "vxlt_undercitydungeon_xr_0_0.vlm"},
	{0x5DFDB8B16053FA95ULL, "t3_dey_am3.tnt"},
	{0x5DFDC0781B396498ULL, "vxht_0_hightown_bze_0_0.vlm"},
	{0x5DFE2057A4D40A72ULL, "zz_dae_debug6047587.cl"},
	{0x5DFEA0B3F319FEC7ULL, "mag291_vincento6027334.cl"},
	{0x5DFEA422012B3210ULL, "mcr370_meredith6085989.cl"},
	{0x5DFF937D964B568EULL, "llac_lt_0_lowtown_n_-3_2.dds"},
	{0x5DFFC0F1E0D7D139ULL, "6161115_m.fxe"},
	{0x5E005314D8227570ULL, "6207772_f.fxe"},
	{0x5E0058E4BE7446D5ULL, "691886753.ogg"},
	{0x5E00ED9C70B70F7DULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.dura.ani"},
	{0x5E010249BAD99187ULL, "gen00fl_isabela6015689.cl"},
	{0x5E01B62AE5AFC7CAULL, "ts_circleofmagi_prog_junk2.gda"},
	{0x5E01E7A9EBBF9C09ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_combat_runf.evt"},
	{0x5E023B81BC340DFEULL, "6129317_f.fxe"},
	{0x5E0245EE9FF77DE7ULL, "6154104_m.fxe"},
	{0x5E026C2482860D74ULL, "dae200_bran6094805.cl"},
	{0x5E0316F172FACE6CULL, "vxht_0_hightown_4n3_0_0.vlm"},
	{0x5E037B84B6D96B66ULL, "vxw_sundermount_2vc_0_0.vlm"},
	{0x5E039DD5621FAD62ULL, "tablewood_02.mao"},
	{0x5E04EC3026C45227ULL, "hlac_ca_caverns2prc_5802_49z.dds"},
	{0x5E051A9A8C241B2EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\cavevista\\8gr_cavern.msh"},
	{0x5E0521774FF3CE4EULL, "6100312_f.fxe"},
	{0x5E05289A8C24337FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\cavevista\\8gr_cavern.mao"},
	{0x5E05349A8C24471CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\cavevista\\8gr_cavern.mmh"},
	{0x5E05861240029FF3ULL, "501791475.ogg"},
	{0x5E0595EB051F6116ULL, "vxlt_gallowscourtyard_7dm_0_0.vlm"},
	{0x5E05962AE5E2AEFDULL, "vxht_mansion2_2jt_0_0.vlm"},
	{0x5E061167F7CDC782ULL, "691250399.ogg"},
	{0x5E067947FEF4F12CULL, "hlac_lt_hangedman_-1_0y.dds"},
	{0x5E067E2FA35D8CE1ULL, "static_rendertodepth_p0.psh.fxo.dx11"},
	{0x5E06A46EFC331A16ULL, "vxca_cavern1_haunted_k2n_0_1.vlm"},
	{0x5E06BF8C3FF153E7ULL, "dae000_orsino6081360.cl"},
	{0x5E06C3D47771DC97ULL, "644446379.ogg"},
	{0x5E076546E6D36B58ULL, "vxw_woundedcoast_16_0_0.vlm"},
	{0x5E07A9F8E3E5DEC0ULL, "443039270.ogg"},
	{0x5E07D40609170827ULL, "vxca_caverns1_ckl_0_0.vlm"},
	{0x5E080F4FB7367900ULL, "737884184.ogg"},
	{0x5E080FE38AD29A47ULL, "6097173_m.fxe"},
	{0x5E0822A9C38F1DE3ULL, "w_encounter3_38_0.rml"},
	{0x5E08342198FEE235ULL, "hla_ht_0_hightown_a2_-131073_0y.dds"},
	{0x5E085B2D19B9A79CULL, "fen210_tevinter_ambusher6108317.cl"},
	{0x5E085EA355D8B26DULL, "hlac_w_encounter2_-2_0x.dds"},
	{0x5E08E5A4BEBF30FEULL, "12726750.ogg"},
	{0x5E090C5037121471ULL, "mcr370_assignments6081737.cl"},
	{0x5E094539E0E0AC87ULL, "zz_fen_debug6218578.cl"},
	{0x5E09761FE0EC6FA3ULL, "hlac_lt_0_lowtown_docks_65536_0z.dds"},
	{0x5E09ABA4AA28B045ULL, "dae011_drunk2_sta6178320.cl"},
	{0x5E09F2F82C058F36ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_basic.phy"},
	{0x5E0A2293F422E5B5ULL, "hlac_ht_0_hightown_a2_65537_0x.dds"},
	{0x5E0A57686FB0A376ULL, "qun120_arishok6051698.cl"},
	{0x5E0A76D804C8A2B9ULL, "6037586_m.fxe"},
	{0x5E0AE4C6A1A9F2E0ULL, "vxlt_hangedman_25b_0_0.vlm"},
	{0x5E0AF69D2DE9231DULL, "vxht_mansion_n_4zi_0_0.vlm"},
	{0x5E0AFBA9B9ACB22DULL, "1030762563.ogg"},
	{0x5E0B15F3D60EBB94ULL, "hlac_ht_0_hightown_a2_-1_2y.dds"},
	{0x5E0B336B434C56E5ULL, "vxlt_0_lowtown_docks_n_b67_0_1.vlm"},
	{0x5E0B39F06AEBDB10ULL, "hlac_lt_0_lowtown_a3_n_c_1_6x.dds"},
	{0x5E0C48FCC5049A05ULL, "art\\levels\\ztools\\default\\generic\\9pp_pole.phy"},
	{0x5E0CA018BCEC526FULL, "6121957_m.fxe"},
	{0x5E0CF30FE39FC10BULL, "hlac_ca_caverns2prc_5802_52z.dds"},
	{0x5E0D4E276A732FD7ULL, "vxca_caverns1_mines_jg4_0_0.vlm"},
	{0x5E0D7AF3E6A5D486ULL, "gen00fl_player6207746.cl"},
	{0x5E0DB97EE2B170A2ULL, "vxlt_0_lowtown_a3_n_c_f5b_0_0.vlm"},
	{0x5E0DBE02BF78DE04ULL, "llac_lt_0_lowtown_n_-131073_0.dds"},
	{0x5E0DD6C09261792AULL, "vxca_caverns1_mines_cmv_0_0.vlm"},
	{0x5E0E3FE5E3DB9928ULL, "w_bonepit_198_0.msh"},
	{0x5E0E51E5E3DBB7B9ULL, "w_bonepit_198_0.mao"},
	{0x5E0E5DE5E3DBCC52ULL, "w_bonepit_198_0.mmh"},
	{0x5E0E616F206A88D2ULL, "seb000cr_change_loot.ncs"},
	{0x5E0E7425525EC816ULL, "hlac_lt_blackemporium_1055_20z.dds"},
	{0x5E0EA27F1066954BULL, "46079999.ogg"},
	{0x5E0EB807F56ADCDCULL, "gencr_mercenary_commander.utc"},
	{0x5E0EE6373AA18BD4ULL, "hlac_w_encounter1_1_0x.dds"},
	{0x5E0F50BDA1D344CDULL, "llac_lt_0_lowtown_a3_n_c_0_0.dds"},
	{0x5E0F53CDB14F07BFULL, "hlac_lt_undercityserialkiller_-1_1z.dds"},
	{0x5E0F9BBCC43C243FULL, "6132719_m.fxe"},
	{0x5E0FC46198496134ULL, "vxlt_undercitydungeon_1ir_0_1.vlm"},
	{0x5E0FCFA9EBCBB178ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_combat_runf.ani"},
	{0x5E0FD51263FE4C6DULL, "6149254_m.fxe"},
	{0x5E1012A86DF38F36ULL, "700819595.ogg"},
	{0x5E10494A7373A4A2ULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_enter.gad"},
	{0x5E1073F9B06EAA59ULL, "ef_mrl150cr_dalish_guard2.mor"},
	{0x5E108A3C61E56CDCULL, "269504936.ogg"},
	{0x5E119AB1C5861828ULL, "vxlt_0_lowtown_a2_n_fi3_0_0.vlm"},
	{0x5E11DC58E5CE7585ULL, "vxw_bonepitdestroyed_a3_2a2_0_0.vlm"},
	{0x5E1286F1590389D4ULL, "zz_lgt_1_debug6136983.cl"},
	{0x5E12D36BC3B0B5BFULL, "hlac_ca_caverns2prc_5802_7z.dds"},
	{0x5E131CEC5C3311D5ULL, "896389226.ogg"},
	{0x5E13B6FC2CFA701EULL, "98922927.ogg"},
	{0x5E13DBF82C0DBCD7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_basic.msh"},
	{0x5E13EDF82C0DDB3AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_basic.mao"},
	{0x5E13F1F82C0DE249ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_basic.mmh"},
	{0x5E1407C7F7157996ULL, "vxlt_gallowscourtyard_7in_0_0.vlm"},
	{0x5E1414BDA8CA0347ULL, "hlac_w_woundedcoast_327685_0y.dds"},
	{0x5E143BAFF39D816FULL, "hlac_w_encounter1_-131068_0x.dds"},
	{0x5E1440A15C22C6ACULL, "gen00fl_isabela6096775.cl"},
	{0x5E14D88C30E3BA51ULL, "vxlt_gallowstemplar_n_4qo_0_0.vlm"},
	{0x5E14F3C20EE043DCULL, "6131093_f.fxe"},
	{0x5E153C0D0B24F95CULL, "hlac_lt_0_lowtown_n_-1_2z.dds"},
	{0x5E1541B60CE20786ULL, "art\\vfx\\creatures\\fxc_drg_tswipe_c.impa.evt"},
	{0x5E15797B6BE9E600ULL, "llac_ca_caverns2_5802_25.dds"},
	{0x5E15DD57481130CBULL, "vxht_0_hightown_a2_atg_0_0.vlm"},
	{0x5E15EB26F6AD619FULL, "vxht_keep_n_3f8_0_0.vlm"},
	{0x5E16265A7427412BULL, "dae000_arishok6075003.cl"},
	{0x5E1631A8E58D208FULL, "lgt320ar_gang_hideout_ht.lst"},
	{0x5E16F52202B1F8A3ULL, "gen00fl_fenris6126723.cl"},
	{0x5E176C4A7379FE99ULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_enter.evt"},
	{0x5E18320078C32689ULL, "130994007.ogg"},
	{0x5E18597A2A62DB9BULL, "vxht_mansion2b_2xt_0_0.vlm"},
	{0x5E1861B115C81B49ULL, "zz_followers_banter6158781.cl"},
	{0x5E18E14F364D8B67ULL, "zz_gen_im_trs_krk_low_sword.uti"},
	{0x5E18EE986222F0A2ULL, "664964506.ogg"},
	{0x5E192775CC20057DULL, "mcr370_negotiation6080283.cl"},
	{0x5E19758B7AD9C77CULL, "llac_lt_undercityencounter_-1_22.dds"},
	{0x5E19861AF2DAEEB8ULL, "166739985.ogg"},
	{0x5E19C6A0AAA2AFD9ULL, "vxlt_0_lowtown_n_gcm_0_0.vlm"},
	{0x5E19D7F8A3A6C7A3ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.mmh"},
	{0x5E1A48C5B4DD8ADFULL, "179663887.ogg"},
	{0x5E1A841156882688ULL, "qcr000_arishok6060141.cl"},
	{0x5E1AA186668AD83BULL, "zz_mer_debug6017194.cl"},
	{0x5E1AD984A92C5CF0ULL, "427209493.ogg"},
	{0x5E1ADC31F9211167ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgta_0s.dds"},
	{0x5E1B236C1A61B869ULL, "mag291_ambient_2.cnv"},
	{0x5E1B39F60F8B3E43ULL, "6095531_m.fxe"},
	{0x5E1B88214E0D8FECULL, "vxlt_0_lowtown_f4w_0_0.vlm"},
	{0x5E1C4CA119CD3E0EULL, "vxlt_darktown_2r4_0_0.vlm"},
	{0x5E1C58B60CE84DD1ULL, "art\\vfx\\creatures\\fxc_drg_tswipe_c.impa.gad"},
	{0x5E1C5D8F7415ED3DULL, "509054013.ogg"},
	{0x5E1D28B481DDEA23ULL, "235770397.ogg"},
	{0x5E1DBC0F442EF767ULL, "666138386.ogg"},
	{0x5E1DF6A5D344B31CULL, "m_sib131ar_family_estate.bnk"},
	{0x5E1E363E8A3F79B4ULL, "vxlt_undercitydungeon_1lq_0_1.vlm"},
	{0x5E1F030AE78F3519ULL, "724693437.ogg"},
	{0x5E1F62FC95F4BA41ULL, "vxw_encounter2_camp_96_0_0.vlm"},
	{0x5E1F65FE8D47BDCEULL, "999838455.ogg"},
	{0x5E1FCCC2576A090CULL, "6049443_m.fxe"},
	{0x5E1FDCAD12330F17ULL, "533917869.ogg"},
	{0x5E201DA74EF10C69ULL, "gen00fl_bethany6083994.cl"},
	{0x5E2197B96D1A118CULL, "vxht_chantry_n_2cq_0_0.vlm"},
	{0x5E21B96DCAE19506ULL, "llac_lt_0_lowtown_a2_-65536_1.dds"},
	{0x5E21C51B4A4C3159ULL, "209629225.ogg"},
	{0x5E2217BA375D998DULL, "6156221_m.fxe"},
	{0x5E2259B60CECBB7FULL, "art\\vfx\\creatures\\fxc_drg_tswipe_c.impa.ani"},
	{0x5E2279E4C9CEDFFDULL, "vxht_mansion2_2pr_0_0.vlm"},
	{0x5E22BCD21FB32462ULL, "6059291_m.fxe"},
	{0x5E2349D8B0B814BBULL, "vxlt_backalley_n_1jq_0_0.vlm"},
	{0x5E24744A73849688ULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_enter.ani"},
	{0x5E24ADB4C7392A5DULL, "vxlt_0_lowtown_docks_a3_n_d0m_0_0.vlm"},
	{0x5E254F790C9DD41AULL, "vxht_playershouse_n_1uw_0_0.vlm"},
	{0x5E255F9E14D0F589ULL, "vxca_cavern1_haunted_e8l_0_0.vlm"},
	{0x5E2619546D8073E9ULL, "6103450_m.fxe"},
	{0x5E26482956982D3CULL, "454885761.ogg"},
	{0x5E26E574EC8FFCEDULL, "6141913_m.fxe"},
	{0x5E27322F4DAF3CD4ULL, "vxht_mansion2_1wq_0_0.vlm"},
	{0x5E275B87EE6A5A77ULL, "416283449.ogg"},
	{0x5E27C4261ABDC04EULL, "vxlt_undercityserialkiller_1ql_0_0.vlm"},
	{0x5E27DA5711D5541DULL, "vxht_mansion_n_4tg_0_0.vlm"},
	{0x5E28346D321E7BB6ULL, "vxlt_gallowstemplar_f_5qi_0_0.vlm"},
	{0x5E2843EDB45A1DDEULL, "vxdr_primevaltomb_50o_0_0.vlm"},
	{0x5E29388C41D6B5B8ULL, "6042147_m.fxe"},
	{0x5E296A753234D6B8ULL, "vxlt_gallowscourtyard_642_0_0.vlm"},
	{0x5E2981A97EBB230BULL, "6139670_f.fxe"},
	{0x5E299018CBD18EE7ULL, "hlac_lt_0_lowtown_a2_n_c_65536_6z.dds"},
	{0x5E2A2555E9349923ULL, "6087389_m.fxe"},
	{0x5E2A624CB2CF94B3ULL, "one130_thug6133304.cl"},
	{0x5E2A81F958983E05ULL, "6088918_f.fxe"},
	{0x5E2B49F2170A1CECULL, "hm_guard_3.mor"},
	{0x5E2B69FE5956A3FAULL, "vxdr_primevaltomb_34x_0_1.vlm"},
	{0x5E2B8A503F5F06ECULL, "llac_lt_undercitydungeon_-1_12.dds"},
	{0x5E2BA89DC5E5B176ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_aoe_imp_p.impa.ani"},
	{0x5E2BAF136199A602ULL, "gen00fl_varric6105482.cl"},
	{0x5E2C14F36498AA4EULL, "6054645_m.fxe"},
	{0x5E2C1D8879DBFAB8ULL, "lgt150im_ntb_wr_ingrdnt_3.uti"},
	{0x5E2C51EC4ABDD3E2ULL, "6053931_f.fxe"},
	{0x5E2C9C6C13BA44B5ULL, "llac_w_bonepit_262144_0.dds"},
	{0x5E2CBED39E5E1A85ULL, "mag320_mage_conspirator6040829.cl"},
	{0x5E2D708ADD2615CDULL, "466501430.ogg"},
	{0x5E2D7AF7024D9195ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_002a_d.dds"},
	{0x5E2D932F5F7CAA90ULL, "307182575.ogg"},
	{0x5E2DC9DD09C7005CULL, "202798062.ogg"},
	{0x5E2DD4D987EAD2C1ULL, "vxw_encounter2_camp_45_0_0.vlm"},
	{0x5E2E57ED07C10DE8ULL, "t3_hey_am2.tnt"},
	{0x5E2E827409F9D8F5ULL, "731357348.ogg"},
	{0x5E2F9B02B42A1D2EULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrla_0t.dds"},
	{0x5E3010376527D158ULL, "39005900.ogg"},
	{0x5E302319DA96D414ULL, "vxw_bonepitmining_a2_mu_0_0.vlm"},
	{0x5E3043CC5103189FULL, "mag311_emile_de_launcet6043730.cl"},
	{0x5E30B15ABC61BFECULL, "6109237_m.fxe"},
	{0x5E30F05F62000569ULL, "hm_gen_templar_archer.mor"},
	{0x5E318D4EE1A33B79ULL, "blank_exterior_0_0.phy"},
	{0x5E319B4CFCF6B82CULL, "hlac_ht_mansion2_818_56y.dds"},
	{0x5E31B866F08D69B2ULL, "6172952_m.fxe"},
	{0x5E32C1112BFD61CBULL, "vxht_0_hightown_a2_ajy_0_0.vlm"},
	{0x5E32F3C1AA1A2FAAULL, "vxht_0_hightown_a6l_0_0.vlm"},
	{0x5E33675133C29826ULL, "47391336.ogg"},
	{0x5E353D340E4F0C9BULL, "vxht_mansion2b_2nv_0_0.vlm"},
	{0x5E3561FA76C2AAF4ULL, "494911437.ogg"},
	{0x5E358D85C14CAB66ULL, "art\\vfx\\combat\\vfx_w_tmp_holy_smt_aoe_p.anb"},
	{0x5E361226A32E3964ULL, "6071112_m.fxe"},
	{0x5E361B2FA3800359ULL, "fxa_cen_light_sp.dds"},
	{0x5E361BBA7312CA04ULL, "6202151_f.fxe"},
	{0x5E367B87CC00ACAFULL, "vxht_mansion2_a9_0_0.vlm"},
	{0x5E37111C270DA1A1ULL, "mer120_hubert6017865.cl"},
	{0x5E37281DEDCEA7D8ULL, "6076812_m.fxe"},
	{0x5E372F2EFE785F12ULL, "follower_banter6170047.cl"},
	{0x5E375C048FE7C92CULL, "vxht_0_hightown_n_c6s_0_0.vlm"},
	{0x5E382CF466AA0C1AULL, "hlac_lt_0_lowtown_docks_-3_2x.dds"},
	{0x5E3831AB05FAA1C7ULL, "hlac_lt_0_lowtown_a2_0_2x.dds"},
	{0x5E383CA2581C1DF7ULL, "vxlt_backalley_n_1dk_0_0.vlm"},
	{0x5E38964EE1A96895ULL, "blank_exterior_0_0.rml"},
	{0x5E38B09DC5F04967ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_aoe_imp_p.impa.evt"},
	{0x5E38F3D64A8E5685ULL, "6083818_f.fxe"},
	{0x5E39305AFDB96F0EULL, "vxlt_darktown_2x6_0_0.vlm"},
	{0x5E3A74C686F7A528ULL, "14766220.ogg"},
	{0x5E3AAE83B349DA5EULL, "dae000cr_grand_cleric6187957.cl"},
	{0x5E3AD55ED4ECD1A5ULL, "vxlt_darktown_587_0_1.vlm"},
	{0x5E3AE6C52B91AE17ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyn_0t.dds"},
	{0x5E3B09DBE8730F16ULL, "vxw_bonepitdestroyed_a3_1pi_0_0.vlm"},
	{0x5E3B10A9C3BA58FAULL, "w_encounter3_38_0.msh"},
	{0x5E3B1EA9C3BA7033ULL, "w_encounter3_38_0.mao"},
	{0x5E3B22A9C3BA7740ULL, "w_encounter3_38_0.mmh"},
	{0x5E3B7E7BFB927AACULL, "mag110_vincento6017899.cl"},
	{0x5E3BEDE56A7CC83EULL, "339464519.ogg"},
	{0x5E3C20DBBE99C5FEULL, "6118772_m.fxe"},
	{0x5E3C288A5AC68329ULL, "443425858.ogg"},
	{0x5E3C863628B3C1C0ULL, "6116146_f.fxe"},
	{0x5E3C9B0A420E02DAULL, "6100306_m.fxe"},
	{0x5E3DA65A3A0E6556ULL, "ser121_jethann6104433.cl"},
	{0x5E3DB0CB4A7BDEBEULL, "vxdr_0_deeproads_cun_0_2.vlm"},
	{0x5E3E2A04FD5910D8ULL, "w_bonepitdestroyed_a3_120_0.phy"},
	{0x5E3E8C6A09E7E2AAULL, "6042493_m.fxe"},
	{0x5E3E94F6CC8CEC94ULL, "vxw_bonepitmining_a2_hz_0_0.vlm"},
	{0x5E3F124B64738614ULL, "mcr370_templar16086966.cl"},
	{0x5E3F7B285C6BE785ULL, "6029149_m.fxe"},
	{0x5E3FC6FB94FD407AULL, "hlac_dr_0_deeproads_5840_1z.dds"},
	{0x5E3FC839123DE5E2ULL, "hlac_lt_0_lowtown_a2_n_196607_0y.dds"},
	{0x5E3FD19DC5F69FB0ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_aoe_imp_p.impa.gad"},
	{0x5E402D9294A445BBULL, "vxlt_backalley_n_1ps_0_0.vlm"},
	{0x5E40500C97EC2738ULL, "438870789.ogg"},
	{0x5E40DF72F1FF5A85ULL, "llac_ht_chantry_2534_51.dds"},
	{0x5E40FC3C77FF3887ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_ian.gad"},
	{0x5E413E08B34CD13CULL, "hlac_lt_0_lowtown_docks_n_-65539_1y.dds"},
	{0x5E41B3BDC280B6FFULL, "t1_skn_007.tnt"},
	{0x5E41EC407AFCE3ADULL, "vxca_cave_1op_0_0.vlm"},
	{0x5E429085C1573A11ULL, "art\\vfx\\combat\\vfx_w_tmp_holy_smt_aoe_p.mmh"},
	{0x5E42AB553B5ACF27ULL, "vxw_encounter2_cn_0_0.vlm"},
	{0x5E42C3DF738892C8ULL, "zz_and_debug6134118.cl"},
	{0x5E42EA6D3DA521D9ULL, "6034298_f.fxe"},
	{0x5E441A5F76C7EF79ULL, "dre000_sibling_has_blight6037029.cl"},
	{0x5E445F8C8A92D709ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\ready.ani"},
	{0x5E448EC922FD103DULL, "vxlt_undercityserialkiller_14d_0_0.vlm"},
	{0x5E449C857BE29639ULL, "vxlt_darktown_4do_0_0.vlm"},
	{0x5E44D16A0CB976AFULL, "character_character_p105.psh.fxo.dx11"},
	{0x5E44D463A9ED43CDULL, "zz_isa_debug6133976.cl"},
	{0x5E45445FEB06067AULL, "777151947.ogg"},
	{0x5E454A1F759A4D52ULL, "559786312.ogg"},
	{0x5E458A2EB66A01ADULL, "vxlt_0_lowtown_h93_0_0.vlm"},
	{0x5E45AE7F2B35E4A0ULL, "zz_playersetup.ncs"},
	{0x5E45CDE181DDE1ACULL, "vxht_0_hightown_n_c3p_0_0.vlm"},
	{0x5E463FC55281B23FULL, "vxlt_0_lowtown_f4x_0_0.vlm"},
	{0x5E465F1FF622E890ULL, "lt_blackemporium.arl"},
	{0x5E468F15071C8B50ULL, "mer340_bonepit_comments6083146.cl"},
	{0x5E46AE7F4A123677ULL, "vxlt_backalley_n_1oj_0_0.vlm"},
	{0x5E475ED68F5538B5ULL, "6050435_m.fxe"},
	{0x5E47A89CD87448E2ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunarifort_d.dds"},
	{0x5E47B20062CD20A6ULL, "6093680_m.fxe"},
	{0x5E47DC176D595540ULL, "6028943_m.fxe"},
	{0x5E47EC6245DF97BEULL, "vxlt_gallowstemplar_5c8_0_0.vlm"},
	{0x5E48AC2597D793C8ULL, "366649666.ogg"},
	{0x5E49191B9F4866A0ULL, "6147594_m.fxe"},
	{0x5E497BB8DA692796ULL, "vxw_bonepitdestroyed_a3_1uj_0_0.vlm"},
	{0x5E4A4EA4016E1CBEULL, "37748964.ogg"},
	{0x5E4A9C997D9D4037ULL, "llac_lt_0_lowtown_docks_n_-131075_0.dds"},
	{0x5E4AF6C34DAEAD5AULL, "832162976.ogg"},
	{0x5E4B06D402A7AED4ULL, "618744646.ogg"},
	{0x5E4B60CCF28F99BBULL, "5246055.ogg"},
	{0x5E4B7A8C8A99238BULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\ready.gad"},
	{0x5E4BC8448F76B4C2ULL, "gen00fl_aveline6112285.cl"},
	{0x5E4C7D1169FF02E4ULL, "qcr000_arishok6167002.cl"},
	{0x5E4CB11BEC118DABULL, "vxdr_primevalentrance_3bb_0_0.vlm"},
	{0x5E4D565F3407D291ULL, "6133040_m.fxe"},
	{0x5E4D5CA41E9D66ACULL, "ht_0_hightown_n_lc.mmh"},
	{0x5E4D5D2A1D455083ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flaggallows_n.dds"},
	{0x5E4DD2E713FAFF63ULL, "hlac_lt_0_lowtown_docks_0_3z.dds"},
	{0x5E4E41BAE3533B0CULL, "77029430.ogg"},
	{0x5E4E4DB258358FA7ULL, "vxlt_0_lowtown_a3_n_c_hf4_0_0.vlm"},
	{0x5E4E8791EC0B7D15ULL, "dae000_gamlen6144344.cl"},
	{0x5E4F401709B1D2DFULL, "178137334.ogg"},
	{0x5E4F8729FE02E0D7ULL, "gen00fl_anders6219292.cl"},
	{0x5E4FF35FAC23238FULL, "dre000_dougal6130995.cl"},
	{0x5E4FF75C08A0B497ULL, "follower_banter6136648.cl"},
	{0x5E504DB0D1D422EEULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bed.phy"},
	{0x5E505D5EDD9A1715ULL, "484301528.ogg"},
	{0x5E50B534EAB33F09ULL, "vxca_cavern1_haunted_e7q_0_0.vlm"},
	{0x5E50E5BA7C8E85E0ULL, "hlac_dr_dungeon_3217_5x.dds"},
	{0x5E511CB4CA4E72E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish.ani"},
	{0x5E511D322D50E7A7ULL, "vxw_encounter2_ni_0_0.vlm"},
	{0x5E51B0A7E8A76CEBULL, "vxht_bartrand_n_66f_0_0.vlm"},
	{0x5E51E249E479668DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_win.phy"},
	{0x5E51E991FCF6E4D9ULL, "vxca_caverns2prc_7uf_0_0.vlm"},
	{0x5E520564570DFEAEULL, "216224830.ogg"},
	{0x5E520CC25B61C7D5ULL, "lt_0_lowtown_docks_a3_147_0.msh"},
	{0x5E520EC25B61CB4BULL, "lt_0_lowtown_docks_a3_147_0.mmh"},
	{0x5E521AC25B61DFE8ULL, "lt_0_lowtown_docks_a3_147_0.mao"},
	{0x5E523547FCECEA6EULL, "mag101tr_terrie_cower.utt"},
	{0x5E52754A18D25216ULL, "hlac_ca_caverns1_mines_24426_2x.dds"},
	{0x5E52778C8A9F3DF0ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\ready.evt"},
	{0x5E52E9C7FDEF3D81ULL, "6067378_m.fxe"},
	{0x5E540ADCACB45BCAULL, "gen00fl_anders6125802.cl"},
	{0x5E5436F4CA15E736ULL, "llac_ca_cavern1_haunted_24425_9.dds"},
	{0x5E543DB1B5BA6E46ULL, "vxca_caverns2_9zg_0_0.vlm"},
	{0x5E543FBE73D3FA2CULL, "vxht_0_hightown_n_c8q_0_0.vlm"},
	{0x5E5453C8E5E4283AULL, "vxdr_0_deeproads_gbu_0_0.vlm"},
	{0x5E55947DDA90E088ULL, "dae000_bodhan6129133.cl"},
	{0x5E55E93C781174CDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerwinning_2_ian.ani"},
	{0x5E561A8B169793CFULL, "hlac_lt_0_lowtown_docks_n_196608_0z.dds"},
	{0x5E561FCCA9969460ULL, "w_bonepitdestroyed_a3_10_0_ll.mmh"},
	{0x5E56703792A24DC9ULL, "570959181.ogg"},
	{0x5E56B819A3847292ULL, "gen00fl_aveline6116674.cl"},
	{0x5E56E8E9C65782ADULL, "vxlt_0_lowtown_eqj_0_0.vlm"},
	{0x5E5717D9C6D1F0BDULL, "6162507_m.fxe"},
	{0x5E573021B7DD44C8ULL, "lgt110_gxa_vgk_silver_order6117012.cl"},
	{0x5E57BD5124CB3FABULL, "vxca_caverns1_i36_0_0.vlm"},
	{0x5E5811536A8D7812ULL, "hlac_ht_mansion2b_818_33x.dds"},
	{0x5E58AB5A54EA52BEULL, "6120825_m.fxe"},
	{0x5E58D65F1DBD6D6FULL, "vxht_keep_73t_0_0.vlm"},
	{0x5E592FAB1F86FA0FULL, "6093528_m.fxe"},
	{0x5E5956DF373D2116ULL, "vxca_cavern1_haunted_i2e_0_0.vlm"},
	{0x5E59996F1DA546ADULL, "gen00fl_fenris6126530.cl"},
	{0x5E59BBD096F179FDULL, "258148191.ogg"},
	{0x5E5A04EDF2B13F10ULL, "hlac_lt_0_lowtown_-3_4y.dds"},
	{0x5E5A450B1D3A1A24ULL, "1011233555.ogg"},
	{0x5E5A6EFFA218C566ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_2_hands.evt"},
	{0x5E5A7EF39041E82DULL, "gen_im_trs_gen_sta_spi_r1.uti"},
	{0x5E5AA2FACAA0ACBAULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_fin_c.anb"},
	{0x5E5AA5E7B5730360ULL, "6068579_f.fxe"},
	{0x5E5AB3D9AA83DC86ULL, "vxlt_0_lowtown_e2h_0_0.vlm"},
	{0x5E5AB42645708BB5ULL, "560572673.ogg"},
	{0x5E5ADAA4BD520BBFULL, "hm_gen_act1.mor"},
	{0x5E5B45C2488947F1ULL, "6097399_m.fxe"},
	{0x5E5B82683B1D4077ULL, "follower_banter6168773.cl"},
	{0x5E5BCFD0300427CDULL, "6146951_f.fxe"},
	{0x5E5BD6A883191D62ULL, "vxw_blightlands_11s_0_0.vlm"},
	{0x5E5C02593FEA88E1ULL, "zz_vault_debug6177805.cl"},
	{0x5E5D2EF70F394091ULL, "653733071.ogg"},
	{0x5E5D8CB652429728ULL, "hlac_w_bonepit_196613_0z.dds"},
	{0x5E5DF7B575E17C62ULL, "hlac_lt_0_lowtown_a2_n_-3_2x.dds"},
	{0x5E5E67D75C460B41ULL, "6118757_m.fxe"},
	{0x5E5ECFFA5EE914ADULL, "vxca_cave_1en_0_0.vlm"},
	{0x5E5F04B4CA5A8757ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish.evt"},
	{0x5E5F3E5D9A6C3441ULL, "320868842.ogg"},
	{0x5E5F8F0F1F470027ULL, "vxw_encounter2_ih_0_0.vlm"},
	{0x5E6007D14A299C30ULL, "hlac_ht_0_hightown_n_-196610_0x.dds"},
	{0x5E602284DA9D856BULL, "vxht_bartrand_n_63y_0_0.vlm"},
	{0x5E60381B47F1F852ULL, "hlac_ht_0_hightown_n_-65534_0x.dds"},
	{0x5E60836975F41130ULL, "595074365.ogg"},
	{0x5E6092CAE431561BULL, "vxht_mansion2b_2ak_0_0.vlm"},
	{0x5E60B9CD9FC086ADULL, "6096644_m.fxe"},
	{0x5E60EC796FC59575ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\blendtree\\trooperassasin_override.anb"},
	{0x5E60F9A315879ED4ULL, "vxht_mansion2_1ku_0_0.vlm"},
	{0x5E610039E0044247ULL, "vxht_brothel_n_4dn_0_0.vlm"},
	{0x5E615EF8EE23C80EULL, "t3_nbl_n02.tnt"},
	{0x5E617C417062580BULL, "gen_im_trs_slv_sta_ele_r0.uti"},
	{0x5E61803F5FCEC739ULL, "vxlt_darktown_4nm_0_0.vlm"},
	{0x5E6185FFA21F0AB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_2_hands.gad"},
	{0x5E61A2694A05D151ULL, "vxht_mansion2_3bu_0_0.vlm"},
	{0x5E61AEB8F0F13DD3ULL, "art\\vfx\\combat\\vfx_str_att_front_r_arm_c.impa.ani"},
	{0x5E61B102730C7ED4ULL, "6176473_m.fxe"},
	{0x5E61F500B81C39B7ULL, "vxca_caverns1_mines_gcu_0_0.vlm"},
	{0x5E6214C08C6CAFEBULL, "w_encounter2_mmap.dds"},
	{0x5E62A638FB54BA46ULL, "vxht_0_hightown_arl_0_0.vlm"},
	{0x5E62D8D6FFD4FB4FULL, "darktown_male_fer_urch_sta6174964.cl"},
	{0x5E63154A620665E1ULL, "6125173_m.fxe"},
	{0x5E63358F61AEF92DULL, "mrl251ip_loot_g.utp"},
	{0x5E6375E7A42A7AD7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic_4.mao"},
	{0x5E6392392DFE6777ULL, "vxlt_backalley_n_1ud_0_0.vlm"},
	{0x5E6452F378B0FCA9ULL, "vxw_sundermount_2ey_0_0.vlm"},
	{0x5E64AEFA848FDBDEULL, "hlac_ca_cavern1_haunted_24426_0x.dds"},
	{0x5E6528459A747441ULL, "246284803.ogg"},
	{0x5E6556262F94BAABULL, "vxw_encounter3_9s_0_0.vlm"},
	{0x5E6560C537FBC9BCULL, "vrc221_hugin6040974.cl"},
	{0x5E6589B6B8C8093DULL, "6167607_m.fxe"},
	{0x5E6597591688CCCDULL, "6110047_m.fxe"},
	{0x5E65A8AAD83156C7ULL, "257306859.ogg"},
	{0x5E6605B4CA60A7E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish.gad"},
	{0x5E66436CA219F5F0ULL, "6013343_m.fxe"},
	{0x5E665F72BE555896ULL, "vxw_bonepitdestroyed_a3_1cl_0_0.vlm"},
	{0x5E66BD3ECEACC842ULL, "vxlt_undercityencounter_13v_0_0.vlm"},
	{0x5E67321A6CA00742ULL, "401141692.ogg"},
	{0x5E678E8DAE423A62ULL, "vxlt_backalley_n_1r4_0_0.vlm"},
	{0x5E67C80074E4C7D2ULL, "tm_war_d04.tnt"},
	{0x5E67EF0347E21C9AULL, "lt_gallowscourtyard_n_42_0_ll.mmh"},
	{0x5E6886FFA2252B5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_2_hands.ani"},
	{0x5E68B5939748DAFDULL, "gen_im_acc_blt_act3_10.uti"},
	{0x5E692847FCFFB297ULL, "mag101tr_terrie_cower.ncs"},
	{0x5E699DAC52AAE917ULL, "696368120.ogg"},
	{0x5E69F11F1857675EULL, "975256114.ogg"},
	{0x5E6A5CE5E42A294DULL, "w_bonepit_198_0.phy"},
	{0x5E6A74DBDDEC671CULL, "follower_banter6136450.cl"},
	{0x5E6B137406C85249ULL, "follower_banter6173076.cl"},
	{0x5E6C5B01B54CB634ULL, "vxdr_0_deeproads_d15_0_0.vlm"},
	{0x5E6CDD1BD9934EC4ULL, "194804125.ogg"},
	{0x5E6EA6B8F0FBBA0AULL, "art\\vfx\\combat\\vfx_str_att_front_r_arm_c.impa.evt"},
	{0x5E6EA98110B4C82FULL, "943597046.ogg"},
	{0x5E6F187AB96EDCB7ULL, "hlac_lt_undercityserialkiller_-1_11y.dds"},
	{0x5E6FE540B94DC76FULL, "vrc321ip_spooky5_obj_1.utp"},
	{0x5E70AAAB7F90E103ULL, "lt_0_lowtown_docks_200_0.mao"},
	{0x5E70AEAB7F90E850ULL, "lt_0_lowtown_docks_200_0.mmh"},
	{0x5E70BCAB7F91006AULL, "lt_0_lowtown_docks_200_0.msh"},
	{0x5E711815ED4AD2C6ULL, "vxht_0_hightown_aum_0_0.vlm"},
	{0x5E71621877055A83ULL, "6038831_m.fxe"},
	{0x5E7177CD393BFC3BULL, "gen00fl_aveline6122735.cl"},
	{0x5E718767BF4B347AULL, "6051848_m.fxe"},
	{0x5E71A258562A1D38ULL, "6013840_m.fxe"},
	{0x5E71B11F846EF4CBULL, "6078414_f.fxe"},
	{0x5E71B91D23D2947CULL, "vxw_blightlands_bu_0_0.vlm"},
	{0x5E71E8FFC834158BULL, "mag121_idunna6084875.cl"},
	{0x5E73554B5FA7901DULL, "6000904_m.fxe"},
	{0x5E73779E033228B6ULL, "6174309_m.fxe"},
	{0x5E738873FE37066EULL, "94179332.ogg"},
	{0x5E73B2B5607DA97BULL, "94003515.ogg"},
	{0x5E73C803218AD32BULL, "vxw_encounter3_2r_0_0.vlm"},
	{0x5E73F17D7CF81FC0ULL, "hlac_lt_backalley_n_0_1x.dds"},
	{0x5E7464019EB87D0CULL, "6206343_m.fxe"},
	{0x5E7469B90FF29359ULL, "one280_iwan6120109.cl"},
	{0x5E74D14FB04B7116ULL, "vxw_bonepitdestroyed_a3_1dm_0_0.vlm"},
	{0x5E74E7141D1D9ADEULL, "501743107.ogg"},
	{0x5E7522618BE338D8ULL, "6165442_m.fxe"},
	{0x5E7577A75BE97FA8ULL, "hlac_dr_dungeon_3215_1z.dds"},
	{0x5E7585B8F101A17DULL, "art\\vfx\\combat\\vfx_str_att_front_r_arm_c.impa.gad"},
	{0x5E75CC0DD67B5BBCULL, "383861131.ogg"},
	{0x5E76494CB3CC1CA9ULL, "llac_ht_0_hightown_a3_65536_4.dds"},
	{0x5E76A1BA23668B0DULL, "118006715.ogg"},
	{0x5E76E89DBC30CE4CULL, "hm_one_har_guard1.mor"},
	{0x5E7716723045B1E6ULL, "239827533.ogg"},
	{0x5E775574ACD59B46ULL, "6060000_m.fxe"},
	{0x5E77843778E7F19AULL, "bd_distor.mao"},
	{0x5E7958730EDFA3D0ULL, "6150059_m.fxe"},
	{0x5E799CC507CF8181ULL, "static_static_p91.psh.fxo.dx11"},
	{0x5E79CE9E90C243DFULL, "21423229.ogg"},
	{0x5E7A5D838D1487F0ULL, "6082100_f.fxe"},
	{0x5E7AACFE9CD3F0D7ULL, "833006933.ogg"},
	{0x5E7ABBE0D57F75B6ULL, "vxdr_0_deeproads_a6s_0_0.vlm"},
	{0x5E7AC562B0197586ULL, "hlac_dr_dungeon_3217_4y.dds"},
	{0x5E7B59BB59609E04ULL, "vxlt_foundry_n_v6_0_0.vlm"},
	{0x5E7C91F44B15AF7EULL, "785934219.ogg"},
	{0x5E7CF1AF3E7383DEULL, "vxlt_0_lowtown_n_f1i_0_0.vlm"},
	{0x5E7D49B9E1CB92BDULL, "6126860_m.fxe"},
	{0x5E7DBD31ECF23D09ULL, "vxlt_undercitydungeon_1o8_0_0.vlm"},
	{0x5E7E439E3C33F2A9ULL, "isa311_isabela6175479.cl"},
	{0x5E7E9ECFFAE4A328ULL, "mrl000_marethari6069730.cl"},
	{0x5E7EBCE1C1A65794ULL, "178313823.ogg"},
	{0x5E7F3020F80C6B6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_death.gad"},
	{0x5E7FDA1CC3FED9AFULL, "mag100_cullen6082302.cl"},
	{0x5E7FE9F71FA77E7BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\varric\\dm_cth_vara_0s.dds"},
	{0x5E7FF73AB4E63712ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_011a_0d.dds"},
	{0x5E80D5891024A9FAULL, "vxw_woundedcoast_58_0_0.vlm"},
	{0x5E81211A1F7A5616ULL, "vxdr_primevaltomb_8mi_0_0.vlm"},
	{0x5E81EE056C372592ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.mmh"},
	{0x5E82600A60259CD6ULL, "211862459.ogg"},
	{0x5E82BD2132D6964FULL, "421912457.ogg"},
	{0x5E831C0311B74471ULL, "vxlt_warehouse_n_qd_0_0.vlm"},
	{0x5E83957DD0012811ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirer.mmh"},
	{0x5E83A17DD0013CB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirer.mao"},
	{0x5E83A5FACAC3B76DULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_fin_c.mmh"},
	{0x5E83AF7DD00154CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirer.msh"},
	{0x5E83B5EAAE580F0BULL, "6133603_f.fxe"},
	{0x5E83CCCE840A621CULL, "6043763_f.fxe"},
	{0x5E843AA9B28CFBADULL, "vxht_chantry_n_2r7_0_0.vlm"},
	{0x5E843BFFAD7171CAULL, "vxdr_0_deeproads_ev3_0_0.vlm"},
	{0x5E847247922E6B62ULL, "vxlt_backalley_n_1x6_0_0.vlm"},
	{0x5E847BBA2BBCE7EDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperc.mmh"},
	{0x5E847FBA2BBCEE3EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperc.mao"},
	{0x5E8485B66645D001ULL, "6110880_f.fxe"},
	{0x5E848DBA2BBD0673ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paperc.msh"},
	{0x5E84EBDE2F1CD945ULL, "6096891_m.fxe"},
	{0x5E85688707CFD898ULL, "479672322.ogg"},
	{0x5E857B405F35679DULL, "1047621497.ogg"},
	{0x5E85952865EAC55EULL, "6110945_m.fxe"},
	{0x5E859F2407AED382ULL, "931061367.ogg"},
	{0x5E85E32B0CCC5129ULL, "6148705_m.fxe"},
	{0x5E85F3060BC8287EULL, "hlac_lt_gallowstemplar_2383_8x.dds"},
	{0x5E861320F8125801ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_death.evt"},
	{0x5E86442DE1188F3AULL, "hlac_lt_0_lowtown_a2_131071_0x.dds"},
	{0x5E86AB941B1EF36FULL, "161382381.ogg"},
	{0x5E86FE05FEE0C5ACULL, "vxw_woundedcoast_a2_0_0.vlm"},
	{0x5E875972C5172C61ULL, "vxw_bonepitmining_a2_2ah_0_0.vlm"},
	{0x5E87B170CB4733B3ULL, "vxlt_0_lowtown_docks_a3_n_czk_0_1.vlm"},
	{0x5E87BC1A0D45D28CULL, "87962200.ogg"},
	{0x5E87E42AF907330FULL, "6038642_m.fxe"},
	{0x5E888D356EF4A2C5ULL, "ch2notnt_ds2.psh.fxo"},
	{0x5E8A537E103E76A8ULL, "85902165.ogg"},
	{0x5E8BB7B4FF5AA4F0ULL, "vxlt_0_lowtown_a2_n_ewv_0_0.vlm"},
	{0x5E8BD1A5FE3221A9ULL, "mag311_keran6041721.cl"},
	{0x5E8BD1BAEA9BD3C1ULL, "mrl210_merrill6102127.cl"},
	{0x5E8C4D8DAB7C3B11ULL, "gen_im_acc_amu_act2_01.uti"},
	{0x5E8C60D922268FD8ULL, "ts_templars_02.gda"},
	{0x5E8C6915EF8F740FULL, "617733851.ogg"},
	{0x5E8CAC00D8D840C4ULL, "6035291_m.fxe"},
	{0x5E8CB856C78FD28FULL, "6126523_m.fxe"},
	{0x5E8CC840C4AE3797ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_rf.gad"},
	{0x5E8D40E0C7CD1AB3ULL, "gestures_core_settings.gda"},
	{0x5E8D4E914367C443ULL, "831373516.ogg"},
	{0x5E8DAADB8F96BB43ULL, "art\\characters\\creatures\\mubari\\animation\\p.evt"},
	{0x5E8DCD71576C9A9CULL, "hlac_lt_0_lowtown_0_6y.dds"},
	{0x5E8DD42FDA04305DULL, "6105074_m.fxe"},
	{0x5E8E10CBFF399C44ULL, "gencr_dragon.utc"},
	{0x5E8EC2CE8C738695ULL, "130015216.ogg"},
	{0x5E8EDFB7805EDCCAULL, "vxlt_warehouse_18p_0_0.vlm"},
	{0x5E8F0CC479B0A0ABULL, "vxlt_0_lowtown_a2_eqp_0_0.vlm"},
	{0x5E8F2163BAF8CFEEULL, "ter_default_n.dds"},
	{0x5E8F4766021AC27AULL, "vxw_woundedcoast_29_0_0.vlm"},
	{0x5E90E228866BD848ULL, "vxw_bonepitmining_a2_2am_0_0.vlm"},
	{0x5E914CB0D20BB7B1ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bed.mmh"},
	{0x5E9158B0D20BCC52ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bed.mao"},
	{0x5E91616E07AB03B8ULL, "vxlt_0_lowtown_haj_0_0.vlm"},
	{0x5E9166753B443F63ULL, "6013421_m.fxe"},
	{0x5E9166B0D20BE46FULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bed.msh"},
	{0x5E927D2BC7113965ULL, "6090393_f.fxe"},
	{0x5E938394F6CEEA86ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_2.evt"},
	{0x5E9394A25B56627CULL, "584846366.ogg"},
	{0x5E93CB5149EB45EFULL, "6162952_m.fxe"},
	{0x5E93DD1D67755205ULL, "vxw_bonepitdestroyed_a3_249_0_0.vlm"},
	{0x5E93DD40C4B47A24ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_rf.evt"},
	{0x5E93ED78A89E4E7BULL, "follower_banter6170761.cl"},
	{0x5E940B20F81E8850ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_death.ani"},
	{0x5E9483DB8F9C97C4ULL, "art\\characters\\creatures\\mubari\\animation\\p.gad"},
	{0x5E94FFD49F84D2C6ULL, "hlac_w_encounter1_-131067_0x.dds"},
	{0x5E95751B5C7EA859ULL, "hlac_w_sundermount_-196610_0x.dds"},
	{0x5E95C7395F023D8AULL, "6037044_m.fxe"},
	{0x5E95DF00FDB69C1DULL, "mag311_emile_de_launcet6043453.cl"},
	{0x5E95E8C44553CB72ULL, "mag131_taraneh6088745.cl"},
	{0x5E9625AE3F71B46CULL, "416442344.ogg"},
	{0x5E96A600BB98EDF9ULL, "405378521.ogg"},
	{0x5E9707BA6E0E257BULL, "vxlt_0_lowtown_emp_0_0.vlm"},
	{0x5E9731DB6111EDAAULL, "t3_wdg_dn2.tnt"},
	{0x5E975DC8799B65B2ULL, "6088244_m.fxe"},
	{0x5E9822968C6F8D03ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingb_d.dds"},
	{0x5E9853C77CBA2243ULL, "hlac_ht_0_hightown_a2_65532_0y.dds"},
	{0x5E9913E61342B880ULL, "24802226.ogg"},
	{0x5E9956DB2951A1F2ULL, "87628454.ogg"},
	{0x5E99A438BF4F6318ULL, "6118345_m.fxe"},
	{0x5E99EC1D303CB055ULL, "532516285.ogg"},
	{0x5E99F8BFB9EEE2F2ULL, "vxca_caverns1_mines_gjw_0_0.vlm"},
	{0x5E9A1704FDA74F6DULL, "w_bonepitdestroyed_a3_120_0.msh"},
	{0x5E9A2504FDA766A0ULL, "w_bonepitdestroyed_a3_120_0.mao"},
	{0x5E9A2904FDA76DF3ULL, "w_bonepitdestroyed_a3_120_0.mmh"},
	{0x5E9A2BEE92808CA8ULL, "623362460.ogg"},
	{0x5E9A5DD34B89CE31ULL, "hlac_lt_0_lowtown_a2_-131073_2z.dds"},
	{0x5E9A92DB8FA11CA2ULL, "art\\characters\\creatures\\mubari\\animation\\p.ani"},
	{0x5E9A9A94F6D530D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_2.gad"},
	{0x5E9AC7ADA7DCA447ULL, "vxht_brothel_n_4xj_0_0.vlm"},
	{0x5E9B5490E61AE7F6ULL, "gen00fl_merrill6150871.cl"},
	{0x5E9C0AF938A5665CULL, "6102874_m.fxe"},
	{0x5E9C1A27884E74A0ULL, "vxlt_warehouse_n_ai_0_0.vlm"},
	{0x5E9C83BAD106DA30ULL, "hlac_ca_cavern1_haunted_24425_3z.dds"},
	{0x5E9CA4F680CC7218ULL, "vxht_0_hightown_b5x_0_0.vlm"},
	{0x5E9CD492DBD6DAADULL, "636589938.ogg"},
	{0x5E9D3849204F2310ULL, "1048658561.ogg"},
	{0x5E9D7EA16BA6B92BULL, "vxlt_0_lowtown_a2_ets_0_0.vlm"},
	{0x5E9D902D72EEA0E3ULL, "vxlt_gamlenshouse_n_2u_0_0.vlm"},
	{0x5E9DD1897A042A13ULL, "40891397.ogg"},
	{0x5E9E52F3F755ACDCULL, "follower_banter6170557.cl"},
	{0x5E9F11C6F108D679ULL, "64857669.ogg"},
	{0x5E9F811C4AA04186ULL, "702388828.ogg"},
	{0x5EA062D933B5D5D2ULL, "follower_banter6154104.cl"},
	{0x5EA09B94F6D99E7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_long_2.ani"},
	{0x5EA0F540C4BF2D1DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_rf.ani"},
	{0x5EA10FC68FB3F626ULL, "genpt_follower_approval.ncs"},
	{0x5EA1109293FF642BULL, "dae000_fenarel6134326.cl"},
	{0x5EA1658C5331F614ULL, "6132856_f.fxe"},
	{0x5EA175ADEE6E9BCFULL, "vxlt_gallowstemplar_f_653_0_0.vlm"},
	{0x5EA1B2E53247D031ULL, "1035188465.ogg"},
	{0x5EA1EE3D12C11122ULL, "6154985_m.fxe"},
	{0x5EA26CCDD109FDD1ULL, "gen00fl_fenris6112007.cl"},
	{0x5EA26D7FCC431EEEULL, "hlac_lt_0_lowtown_n_65538_3z.dds"},
	{0x5EA299E664E65EF2ULL, "311986420.ogg"},
	{0x5EA33FBCF91B82F6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabs.phy"},
	{0x5EA35581A9AA622FULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliff02_darelief.dds"},
	{0x5EA35DB4789FF99DULL, "mag291_ambient_26218722.cl"},
	{0x5EA3B4200960CD66ULL, "6095128_f.fxe"},
	{0x5EA3F09C88EC54B2ULL, "character_character_p225.psh.fxo.dx11.assembly"},
	{0x5EA473BD10AF99F1ULL, "445919386.ogg"},
	{0x5EA4AD7156D6CC71ULL, "vxlt_darktown_4wr_0_0.vlm"},
	{0x5EA4B9CCB50F9503ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_lrgchair_1.mao"},
	{0x5EA56F1C00D6DE31ULL, "llac_lt_0_lowtown_n_65534_0.dds"},
	{0x5EA5FC05ABDC3E34ULL, "llac_w_woundedcoast_3_0.dds"},
	{0x5EA60A757DAA56D2ULL, "227922345.ogg"},
	{0x5EA74D7759C4C623ULL, "6082200_f.fxe"},
	{0x5EA7B022B5593929ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_reaction_dead.ani"},
	{0x5EA8064B04E3543AULL, "6124141_f.fxe"},
	{0x5EA8148F48E76712ULL, "zz_warriorslate_baddy.cut"},
	{0x5EA87F930ACE79BCULL, "hlac_dr_0_deeproads_5845_17z.dds"},
	{0x5EA892CB6A248120ULL, "916694010.ogg"},
	{0x5EA904A5D3B12B0CULL, "470183766.ogg"},
	{0x5EA93BEB906938F3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_help_up_goodbye_bethany.evt"},
	{0x5EA99C10AB7C038FULL, "6128723_m.fxe"},
	{0x5EA9EC2C934EDED7ULL, "659444205.ogg"},
	{0x5EAA4422E0CEBF68ULL, "ser100_party_comments6108237.cl"},
	{0x5EAA447699532BF4ULL, "mer220_hubert6079890.cl"},
	{0x5EAAE7E8C3446BC1ULL, "590714582.ogg"},
	{0x5EAAED056C5A2939ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.anb"},
	{0x5EAB175E6AF3C2C3ULL, "static_static_p8.vsh.fxo.dx11"},
	{0x5EABB00D215D7195ULL, "6069872_f.fxe"},
	{0x5EABC371644B0DCAULL, "vxlt_warehouse_12r_0_0.vlm"},
	{0x5EABCD2D14993945ULL, "vxw_woundedcoast_53_0_0.vlm"},
	{0x5EABE0C89AC9037DULL, "vxlt_0_lowtown_docks_bb1_0_0.vlm"},
	{0x5EAC020A64E4B963ULL, "vxlt_gamlenshouse_n_5t_0_0.vlm"},
	{0x5EAC32D1D4F69EECULL, "vxlt_0_lowtown_n_gkl_0_0.vlm"},
	{0x5EAC9888F05E8189ULL, "vxca_caverns2prc_a0k_0_0.vlm"},
	{0x5EACA2D0FDECB67CULL, "vxw_sundermount_2al_0_0.vlm"},
	{0x5EACAAE9DBF9FC72ULL, "427488201.ogg"},
	{0x5EAD28BCF923AF97ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabs.msh"},
	{0x5EAD3ABCF923CEFAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabs.mao"},
	{0x5EAD3EBCF923D509ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_slabs.mmh"},
	{0x5EAD42E1ABCC7F01ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medc.phy"},
	{0x5EADEF7A80E77E00ULL, "404530017.ogg"},
	{0x5EAEAE867350FF83ULL, "68480577.ogg"},
	{0x5EAEC5B0205F7D0FULL, "dae200cr_bodahn.utc"},
	{0x5EAECB22B55F85ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_reaction_dead.gad"},
	{0x5EAF4026DAF3D1A2ULL, "gen00fl_isabela_init6127645.cl"},
	{0x5EAF9040886F452DULL, "vxht_chantry_n_2c4_0_0.vlm"},
	{0x5EAFB71A49736B52ULL, "6095344_f.fxe"},
	{0x5EB034EB906F4B34ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_help_up_goodbye_bethany.gad"},
	{0x5EB0F36DF10BE160ULL, "793251968.ogg"},
	{0x5EB111B2C2AD17EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_enter_meredith.evt"},
	{0x5EB15E0667EA8DA2ULL, "follower_banter6171901.cl"},
	{0x5EB16BF3FC88D1F1ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.cess.ani"},
	{0x5EB1B07E53A5AB0BULL, "ico_mana_1.dds"},
	{0x5EB1B2D9159E116FULL, "vxw_encounter1_ay_0_0.vlm"},
	{0x5EB255B4D7A450EDULL, "98393364.ogg"},
	{0x5EB2802924F04FE4ULL, "319525209.ogg"},
	{0x5EB2C27D352EF4C9ULL, "448097478.ogg"},
	{0x5EB335EFFD2CF9A6ULL, "442183518.ogg"},
	{0x5EB343788E33501AULL, "zz_mer_debug6180864.cl"},
	{0x5EB397E7812ED18EULL, "ave121_jeven6105692.cl"},
	{0x5EB3B9E3FEF7CB60ULL, "6149264_m.fxe"},
	{0x5EB3C55E287AE44CULL, "gen00fl_isabela6075282.cl"},
	{0x5EB3D822ED105E0DULL, "hlac_lt_undercityserialkiller_-1_10x.dds"},
	{0x5EB4C32788DBDCF2ULL, "728734485.ogg"},
	{0x5EB53616FE1FE5EEULL, "mcr310_argument6127659.cl"},
	{0x5EB5638B4E806D87ULL, "static_static_p39.psh.fxo.dx11"},
	{0x5EB5EF4E4176EB9CULL, "6041685_m.fxe"},
	{0x5EB643EB9073D012ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_help_up_goodbye_bethany.ani"},
	{0x5EB687F2A33B9418ULL, "6112145_m.fxe"},
	{0x5EB68D98F54FC7A7ULL, "vxw_blightlands_7m_0_0.vlm"},
	{0x5EB774C7CE106E19ULL, "65403957.ogg"},
	{0x5EB7B3555697FE3AULL, "hlac_ca_caverns2prc_5802_37z.dds"},
	{0x5EB7F88B358AFC1EULL, "vxlt_gallowscourtyard_n_4mv_0_0.vlm"},
	{0x5EB81A59576446AFULL, "hlac_lt_darktown_-1_55y.dds"},
	{0x5EB84DE133E82DDAULL, "hlac_dr_primevaltomb_657_1y.dds"},
	{0x5EB876F3FC8F0293ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.cess.gad"},
	{0x5EB88C3661E9E929ULL, "903113646.ogg"},
	{0x5EB8F96E354AEBA1ULL, "189535653.ogg"},
	{0x5EB91F1E137C4CBDULL, "hlac_ca_caverns2_5802_64x.dds"},
	{0x5EB9509E41C07746ULL, "mag101_grace6034072.cl"},
	{0x5EB9BDE9005C9D80ULL, "seb221_ruxton_harimann6197721.cl"},
	{0x5EB9EF60AE6813B2ULL, "vrc221_bartrand6043048.cl"},
	{0x5EBA2891190D4EE3ULL, "6080249_m.fxe"},
	{0x5EBA3F0A068F51C5ULL, "vxw_woundedcoast_22_0_0.vlm"},
	{0x5EBAD9D8F8BD42B7ULL, "hm_rdr_guard_warehouse2.mor"},
	{0x5EBB14ADEFE2CEFCULL, "vxw_sundermount_2jm_0_0.vlm"},
	{0x5EBB30FBC38C51C5ULL, "6093904_m.fxe"},
	{0x5EBB59F2983FD411ULL, "hlac_w_sundermount_196606_0z.dds"},
	{0x5EBBCF061A3316F3ULL, "vxlt_0_lowtown_docks_n_c50_0_0.vlm"},
	{0x5EBC4E574FAC1AD4ULL, "896491432.ogg"},
	{0x5EBC5DC6063E5C1CULL, "tinydrops.mao"},
	{0x5EBCA50B05184662ULL, "198378583.ogg"},
	{0x5EBCC4CD54715A9AULL, "art\\characters\\weapons\\zzdao\\w_lsw_ls01a.phy"},
	{0x5EBD29B545CA5CD6ULL, "hlac_lt_0_lowtown_a2_1_7y.dds"},
	{0x5EBE21598D0EFD0FULL, "ht_0_hightown_n_43_0_lc.mmh"},
	{0x5EBF73F3FC951CF8ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.cess.evt"},
	{0x5EBFD56B6D64E00CULL, "vxlt_0_lowtown_a2_n_c_hnv_0_0.vlm"},
	{0x5EC024B6079429EFULL, "vxw_encounter1_fx_0_0.vlm"},
	{0x5EC037C1086B5869ULL, "rdr111dg_martin6015001.cl"},
	{0x5EC055730619F434ULL, "6123310_m.fxe"},
	{0x5EC0A14FC8B971B0ULL, "vxlt_0_lowtown_a2_n_by7_0_0.vlm"},
	{0x5EC0EAB2F4A6DD67ULL, "vxlt_0_lowtown_a2_n_bz9_0_0.vlm"},
	{0x5EC11AA0247FFDFDULL, "vxht_keep_n_4b5_0_0.vlm"},
	{0x5EC14CABBF5FED40ULL, "vxlt_0_lowtown_docks_n_bdq_0_0.vlm"},
	{0x5EC154778A83A1ECULL, "6179018_m.fxe"},
	{0x5EC1C93C4F771AD4ULL, "vxht_bartrand_n_5t8_0_0.vlm"},
	{0x5EC1DEE9F16E823AULL, "222439985.ogg"},
	{0x5EC22CEAD4FA81D0ULL, "zz_vrc_debug6064264.cl"},
	{0x5EC24E6CF44FCD89ULL, "follower_banter6163903.cl"},
	{0x5EC26AE551699DB4ULL, "hf_zmother_h2s3.mor"},
	{0x5EC2EDCB4B7399C6ULL, "fx_twirl01.mao"},
	{0x5EC2FB8FA736F5D4ULL, "water_pan1.mao"},
	{0x5EC41CBF4C038276ULL, "zz_vault_debug6203226.cl"},
	{0x5EC4BC609A65BD80ULL, "vxht_0_hightown_n_7g4_0_0.vlm"},
	{0x5EC4FF75E745E027ULL, "vxw_blightlands_8l_0_0.vlm"},
	{0x5EC5418F0B0735BFULL, "hlac_ht_mansion2_818_36x.dds"},
	{0x5EC603E30980F148ULL, "909232582.ogg"},
	{0x5EC65BEBCFCE0CC0ULL, "157904745.ogg"},
	{0x5EC66A97C965AFB9ULL, "54996822.ogg"},
	{0x5EC6D6A2E9B4ACA9ULL, "hlac_ht_mansion2b_818_18x.dds"},
	{0x5EC707BA64A812D1ULL, "vxlt_0_lowtown_a2_e1r_0_0.vlm"},
	{0x5EC785B584E873B1ULL, "245221759.ogg"},
	{0x5EC78F910E13EE1DULL, "ser221ip_to_hightown.utp"},
	{0x5EC8F151ED859C5EULL, "91685230.ogg"},
	{0x5EC90E121891123AULL, "vxlt_warehouse_n_su_0_0.vlm"},
	{0x5EC9868AE1D8E77CULL, "vxw_sundermount_2ob_0_0.vlm"},
	{0x5EC9ACE8F8E0162AULL, "character_character_p67.psh.fxo"},
	{0x5EC9AD537981B936ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_0x.dds"},
	{0x5EC9DC0CF044D0ADULL, "vxht_mansion_n_4d6_0_1.vlm"},
	{0x5EC9FBA2B0AFF630ULL, "vxht_chantry_2ad_0_0.vlm"},
	{0x5ECA40E30C292F73ULL, "vxlt_0_lowtown_docks_n_c03_0_0.vlm"},
	{0x5ECABF4816DF54BBULL, "6054900_f.fxe"},
	{0x5ECAF1EAFA2BB338ULL, "mrl150tr_mtn_undead_1.ncs"},
	{0x5ECB0ECF8AA20553ULL, "vxht_mansion2_1sn_0_0.vlm"},
	{0x5ECB80FAFBE3D0E6ULL, "lowtown_argument6169444.cl"},
	{0x5ECC237020E3025FULL, "6078804_m.fxe"},
	{0x5ECC5B6539654F67ULL, "6001013_m.fxe"},
	{0x5ECCD2828370BE58ULL, "ht_store_mmi.gff"},
	{0x5ECD520C1EEEF547ULL, "spt_branchshadow.psh.fxo"},
	{0x5ECEDF55BB116C8BULL, "6108798_m.fxe"},
	{0x5ECF1004A4CB95BEULL, "vxdr_primevalentrance_15p_0_0.vlm"},
	{0x5ECF7F2E10EF401DULL, "6050196_m.fxe"},
	{0x5ECF8C7D1676167DULL, "vxht_keep_n_4i4_0_0.vlm"},
	{0x5ECFD1E1EDDCB5FAULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_90_c.impa.gad"},
	{0x5ED00C8DB00D7662ULL, "dae000_flemeth6013358.cl"},
	{0x5ED066DA9B8B7DC3ULL, "hightown_fem_noble1_sta6169340.cl"},
	{0x5ED088D48E3B31A6ULL, "dae000_cora.cnv"},
	{0x5ED0A0C47A62F8D1ULL, "vxlt_gallowstemplar_n_42z_0_0.vlm"},
	{0x5ED14039E4133AF9ULL, "vxw_bonepitmining_a2_7j_0_0.vlm"},
	{0x5ED1838A1DB951FFULL, "water_pan_add_blur.dds"},
	{0x5ED1B4398AECA591ULL, "1073077264.ogg"},
	{0x5ED1C8DF8DB68CB7ULL, "lt_0_lowtown_docks_n_239_0.mmh"},
	{0x5ED1CCDF8DB693C4ULL, "lt_0_lowtown_docks_n_239_0.mao"},
	{0x5ED1DEDF8DB6B229ULL, "lt_0_lowtown_docks_n_239_0.msh"},
	{0x5ED1FAEB9D006CA3ULL, "470437660.ogg"},
	{0x5ED249F3264B03D8ULL, "vxht_playershouse_n_1ue_0_0.vlm"},
	{0x5ED2AB55FC5E5503ULL, "vxlt_hangedman_n_2en_0_0.vlm"},
	{0x5ED2EB0808F3DF80ULL, "vxdr_0_deeproads_fiw_0_0.vlm"},
	{0x5ED32838578B4016ULL, "364643505.ogg"},
	{0x5ED38AD48E3D6B66ULL, "dae000_cora.bnk"},
	{0x5ED3B14ABC348970ULL, "vxlt_gallowstemplar_f_34z_0_0.vlm"},
	{0x5ED425FF55EB5698ULL, "194148597.ogg"},
	{0x5ED48F854BF6A3D2ULL, "vxht_mansion2_1sk_0_0.vlm"},
	{0x5ED4B0074305EB18ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_meda.phy"},
	{0x5ED4BD332C42D7E2ULL, "6108074_m.fxe"},
	{0x5ED4FD09C101EED1ULL, "6154128_m.fxe"},
	{0x5ED549D23F945E17ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_punch_into_palm.gad"},
	{0x5ED58E00B7C1F84CULL, "954561097.ogg"},
	{0x5ED5D3453D70E023ULL, "vxlt_0_lowtown_a3_n_c_gz7_0_0.vlm"},
	{0x5ED5D4E1EDE12611ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_90_c.impa.evt"},
	{0x5ED6555629FD186CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair2_sa.dds"},
	{0x5ED7188FE2648ED1ULL, "vxw_encounter3_eu_0_0.vlm"},
	{0x5ED71C4FEC739C46ULL, "qcr000_arishok6166986.cl"},
	{0x5ED74615337F1B2BULL, "vxlt_0_lowtown_a2_ehw_0_0.vlm"},
	{0x5ED75595DEFF887FULL, "zz_approval_debug6003747.cl"},
	{0x5ED77FEF0A872ABAULL, "vxlt_warehouse_n_vv_0_0.vlm"},
	{0x5ED792F5AC9AD196ULL, "vxht_chantry_n_354_0_0.vlm"},
	{0x5ED7D9B405490BA2ULL, "6177923_f.fxe"},
	{0x5ED7F194C14EBFACULL, "620766566.ogg"},
	{0x5ED7F867D3CEFFFCULL, "vxw_sundermount_2pc_0_0.vlm"},
	{0x5ED86D7FA2A60EB0ULL, "vxht_chantry_2ze_0_0.vlm"},
	{0x5ED8732BABB324AFULL, "vxlt_0_lowtown_docks_a3_n_d1p_0_0.vlm"},
	{0x5ED882FAC3FB1BAEULL, "vxw_bonepitdestroyed_a3_27o_0_0.vlm"},
	{0x5ED8BA248C09B0DCULL, "6187954_m.fxe"},
	{0x5ED9D9F58A4A7152ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_1x.dds"},
	{0x5EDA14C421CD600AULL, "l_env_wal_hightown_int_crowd_bed.wbd"},
	{0x5EDA82602565C4AAULL, "160582909.ogg"},
	{0x5EDAC18F3A811883ULL, "fen310_fenris6175081.cl"},
	{0x5EDACCC096FFD6BAULL, "zz_dre_debug6092007.cl"},
	{0x5EDB1F19BFA9E2EBULL, "6075443_m.fxe"},
	{0x5EDB4B624C8A774AULL, "dae000_bartrand6027484.cl"},
	{0x5EDBBA2D7F378A1CULL, "nat380_temmerin6067346.cl"},
	{0x5EDC5ED23F9AA0A4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_punch_into_palm.evt"},
	{0x5EDC7924CB13A70BULL, "650402624.ogg"},
	{0x5EDCCED18BA01DF0ULL, "vxlt_0_lowtown_docks_n_bb3_0_0.vlm"},
	{0x5EDDF1E6C81CD3C7ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_bigtreasure_n.dds"},
	{0x5EDDFD8186757291ULL, "target_circle_invalid.cur"},
	{0x5EDE34F5D3EB8E8FULL, "vxlt_0_lowtown_docks_n_b5p_0_1.vlm"},
	{0x5EDE38316E6FD2ECULL, "6177218_m.fxe"},
	{0x5EDE59C16A2EC675ULL, "lt_0_lowtown_docks_a3_n_25_0_lc.mmh"},
	{0x5EDEB4D1E08E4011ULL, "vxht_mansion2b_1qz_0_0.vlm"},
	{0x5EDECAC7CC3574A7ULL, "vxht_brothel_n_6tt_0_0.vlm"},
	{0x5EDF0DD4C3166BE3ULL, "gen00fl_varric6107304.cl"},
	{0x5EDF12A16C591151ULL, "vxlt_gallowstemplar_n_47y_0_0.vlm"},
	{0x5EDF369CF2FCD955ULL, "6059300_m.fxe"},
	{0x5EDF50574D1D53A0ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_basket_d.dds"},
	{0x5EDFF5384AF0CCE1ULL, "lt_0_lowtown_52_0.rml"},
	{0x5EE0D0BEE08FF12DULL, "vxca_cave_18e_0_0.vlm"},
	{0x5EE0E898C0531737ULL, "llac_lt_0_lowtown_a2_n_c_-131071_1.dds"},
	{0x5EE0F5E8F15CD66BULL, "dae000cr_grand_cleric6124764.cl"},
	{0x5EE11B6C307A4D4BULL, "llac_ht_chantry_n_2534_10.dds"},
	{0x5EE152142B7D2091ULL, "6139051_m.fxe"},
	{0x5EE17DBE9895119FULL, "vxw_sundermount_317_0_1.vlm"},
	{0x5EE18CC51F95BF21ULL, "vxdr_0_deeproads_d3j_0_0.vlm"},
	{0x5EE1B3DF01BB6A61ULL, "217701355.ogg"},
	{0x5EE1FECB4B8E6768ULL, "fx_twirl01.dds"},
	{0x5EE308EAFA406C35ULL, "mrl150tr_mtn_undead_1.utt"},
	{0x5EE30D89D9DF5A11ULL, "6208596_f.fxe"},
	{0x5EE318A5CF7FDE49ULL, "ave221_brennan6110469.cl"},
	{0x5EE3EBD0CEDBEC45ULL, "ser191st_emeric_turn_in.stg"},
	{0x5EE3ECE1EDED8CE0ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_90_c.impa.ani"},
	{0x5EE3EEA62AC0AE48ULL, "896712133.ogg"},
	{0x5EE440C32D2901E6ULL, "llac_ht_0_hightown_n_-131072_1.dds"},
	{0x5EE448DA8E715435ULL, "6100239_m.fxe"},
	{0x5EE4529DB481CE75ULL, "pro000_refugee_male6052113.cl"},
	{0x5EE4989784662876ULL, "mag100_cullen6140942.cl"},
	{0x5EE567B0E8201EC8ULL, "154156419.ogg"},
	{0x5EE5EF17D7BC64ADULL, "vxca_caverns2prc_62n_0_0.vlm"},
	{0x5EE5F51559BD761CULL, "vxlt_darktown_420_0_0.vlm"},
	{0x5EE66A259A99C657ULL, "lyriumghost_lyriumghost.dds"},
	{0x5EE765599B04E051ULL, "vxht_keep_6aa_0_0.vlm"},
	{0x5EE782F7143A5B1BULL, "38162713.ogg"},
	{0x5EE78C35FE5BA58AULL, "vxca_caverns1_dr1_0_0.vlm"},
	{0x5EE7C7988C8109ADULL, "1067426870.ogg"},
	{0x5EE7F09F1A2B5835ULL, "vxht_chantry_n_2q2_0_0.vlm"},
	{0x5EE8D4DBC01DB5D5ULL, "mag110_sailor6107414.cl"},
	{0x5EE97265C863A202ULL, "6078522_m.fxe"},
	{0x5EE976D23FA5539DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_punch_into_palm.ani"},
	{0x5EE9BDB9A1652F25ULL, "hlac_ht_keep_3732_7x.dds"},
	{0x5EE9C34C9F81CB7CULL, "287702200.ogg"},
	{0x5EEA49BEE7531257ULL, "seb321_leliana6197957.cl"},
	{0x5EEA9F805DB09A3AULL, "vxlt_darktown_4qo_0_0.vlm"},
	{0x5EEB8C090F1D5346ULL, "6059347_m.fxe"},
	{0x5EEBE95C8B2B984AULL, "756798985.ogg"},
	{0x5EEC38B5AD9A325EULL, "llac_lt_blackemporium_1055_4.dds"},
	{0x5EEC75A2B8885F68ULL, "6107147_m.fxe"},
	{0x5EEC7F9CC9AE8318ULL, "6165019_m.fxe"},
	{0x5EEC8C85C8AEBDB4ULL, "273808901.ogg"},
	{0x5EEC8EE8FB9AB757ULL, "vxlt_0_lowtown_ekl_0_2.vlm"},
	{0x5EEC9BD90AD0C28FULL, "mag120_macha6087850.cl"},
	{0x5EECC3C81906EE7CULL, "527260602.ogg"},
	{0x5EED9C8B72A1EF4DULL, "vxht_0_hightown_9x9_0_0.vlm"},
	{0x5EEDA3CDF4FC1895ULL, "6086029_m.fxe"},
	{0x5EEE0374B5362506ULL, "hlac_lt_0_lowtown_a2_n_65538_3x.dds"},
	{0x5EEEBBF9F3996E2AULL, "vxht_0_hightown_aua_0_0.vlm"},
	{0x5EEECF544F83DBBEULL, "hlac_lt_0_lowtown_a2_65534_2z.dds"},
	{0x5EEF5C30B914DFBAULL, "lowtown_male_cityguard_mob6150501.cl"},
	{0x5EEFBE3E917FD021ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsbuild.phy"},
	{0x5EEFBFF894694BEFULL, "dae000ip_crowd_mf4_talk7.utp"},
	{0x5EEFC53BDB6530C5ULL, "art\\levels\\races\\design\\_textures\\prp_chestc_n.dds"},
	{0x5EF084AE586A9480ULL, "hlac_ht_chantry_n_2534_48z.dds"},
	{0x5EF0AA661BF1A33FULL, "6108556_m.fxe"},
	{0x5EF0E6BE285A5FDFULL, "6153137_f.fxe"},
	{0x5EF254DC4A5F3200ULL, "vxlt_0_lowtown_docks_a3_8uh_0_0.vlm"},
	{0x5EF27EF787B780FDULL, "vxw_bonepitdestroyed_a3_1a9_0_0.vlm"},
	{0x5EF2884460B3B76FULL, "67805010.ogg"},
	{0x5EF2ADA59E6CDE41ULL, "bink.vsh.fxo"},
	{0x5EF2B5E8D50D71E1ULL, "6075859_f.fxe"},
	{0x5EF2BF801A3602C3ULL, "582010489.ogg"},
	{0x5EF3A95CA9AD1486ULL, "391640647.ogg"},
	{0x5EF400CF0E09C81BULL, "character_character_p55.psh.fxo.dx11"},
	{0x5EF43B718C103FBBULL, "art\\vfx\\combat\\vfx_ha_ar_armrestor_act_c.impa.evt"},
	{0x5EF44081FD7B70CEULL, "hlac_ht_bartrand_n_2995_0z.dds"},
	{0x5EF466F24BB38E9CULL, "vxlt_darktown_491_0_0.vlm"},
	{0x5EF4A459C373CB95ULL, "336864494.ogg"},
	{0x5EF4F7944390704CULL, "vxlt_gallowstemplar_n_6dj_0_1.vlm"},
	{0x5EF535FB055654E5ULL, "lowtown_male_lowlife1_sta6164939.cl"},
	{0x5EF61A93E46DE119ULL, "mcr371_hostile_aveline6082202.cl"},
	{0x5EF6258566CB364EULL, "art\\vfx\\combat\\vfx_f_avl_bdygrd_imp_c.impa.ani"},
	{0x5EF6EAF6CE9F983AULL, "585069865.ogg"},
	{0x5EF735649EE25251ULL, "hf_lgt_urn_guard.mor"},
	{0x5EF79CE5DDF9EC7AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medj_0d.dds"},
	{0x5EF8D9532F266842ULL, "774465290.ogg"},
	{0x5EF8E2AF89B889D0ULL, "dae000_brothel_orana6161218.cl"},
	{0x5EF962A453D056C7ULL, "llac_ht_playershouse_n_855_28.dds"},
	{0x5EF98DCBE99CF208ULL, "vxw_sundermount_37g_0_0.vlm"},
	{0x5EF9B28B6F8C4EF0ULL, "vxlt_0_lowtown_docks_n_bt1_0_0.vlm"},
	{0x5EF9C826288DFD39ULL, "gen_im_wep_mag_sta_act1_02.uti"},
	{0x5EFA34718C149FFCULL, "art\\vfx\\combat\\vfx_ha_ar_armrestor_act_c.impa.gad"},
	{0x5EFA3BC0EA9E9ED0ULL, "lgt311_shw_mer_mettin6156570.cl"},
	{0x5EFA5AFC7EB31343ULL, "zz_followers_banter6159645.cl"},
	{0x5EFADD35E615E591ULL, "6165043_m.fxe"},
	{0x5EFBB488E249E46AULL, "ht_0_hightown_a2_12_0_lc.mmh"},
	{0x5EFBDE4B38AC6922ULL, "gen00fl_anders6119131.cl"},
	{0x5EFC1A56D6E90A55ULL, "6053100_m.fxe"},
	{0x5EFC8E6D6D55DF1EULL, "def_rock_d.dds"},
	{0x5EFC9B3CDEA47BC6ULL, "vxdr_primevaltomb_77d_0_0.vlm"},
	{0x5EFCA4AD784433BFULL, "300603836.ogg"},
	{0x5EFCD2869C43D198ULL, "vxdr_primevaltomb_6vd_0_0.vlm"},
	{0x5EFCE76A2AD7AB6EULL, "hlac_lt_0_lowtown_docks_a3_-65539_5y.dds"},
	{0x5EFD6C2182D5E184ULL, "1015493850.ogg"},
	{0x5EFE2C615800B4EFULL, "cd_dre000_cut_arrivaldeeproads.wbd"},
	{0x5EFE48236BCD0862ULL, "vxlt_0_lowtown_docks_a3_bs2_0_0.vlm"},
	{0x5EFF2F3B8634F991ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\set_tarp_n.dds"},
	{0x5EFF4FBE5DDDCB64ULL, "529595883.ogg"},
	{0x5EFF8E3056B333B1ULL, "art\\vfx\\combat\\vfx_glb_ele_p_minor_c.impa.ani"},
	{0x5EFFB0E2C1362819ULL, "vxlt_undercityserialkiller_11w_0_0.vlm"},
	{0x5EFFC20AE3223EE6ULL, "6180881_m.fxe"},
	{0x5F000258335F737AULL, "6087347_f.fxe"},
	{0x5F002492A48CCC2DULL, "985144837.ogg"},
	{0x5F003993712DF259ULL, "vxlt_0_lowtown_a3_n_c_gwv_0_0.vlm"},
	{0x5F0069763AD32173ULL, "hf_amb_unhappy_mage.mor"},
	{0x5F0074FFE92273B0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_tmpa_0n.dds"},
	{0x5F0084ECAFFD4F33ULL, "95210268.ogg"},
	{0x5F00AC5609B54AAAULL, "99046798.ogg"},
	{0x5F00D90B2D0A2DDAULL, "one280_emrys6137190.cl"},
	{0x5F00F0D479AD997DULL, "vxw_bonepitdestroyed_a3_1j8_0_0.vlm"},
	{0x5F010DBD01380636ULL, "234000522.ogg"},
	{0x5F0133718C1ABCFAULL, "art\\vfx\\combat\\vfx_ha_ar_armrestor_act_c.impa.ani"},
	{0x5F015442FDB12417ULL, "441267014.ogg"},
	{0x5F017A1ADC4ADDE9ULL, "6018732_m.fxe"},
	{0x5F017F68A8A87C06ULL, "vxca_cave_1oo_0_0.vlm"},
	{0x5F0194A4CC232890ULL, "6061396_f.fxe"},
	{0x5F01CB6FDC39BBCCULL, "vxht_mansion2_1as_0_0.vlm"},
	{0x5F03254C0B5F5B3BULL, "zz_approval_debug6003606.cl"},
	{0x5F039D8D74B21946ULL, "mrl150cr_dalish_hunter.utc"},
	{0x5F03A04D4A72DED4ULL, "vxht_keep_4ng_0_0.vlm"},
	{0x5F03CAC3F4D47EFEULL, "qcr220cr_alistar_gw.utc"},
	{0x5F041D04F162CDBDULL, "6206962_m.fxe"},
	{0x5F041D8566D7659FULL, "art\\vfx\\combat\\vfx_f_avl_bdygrd_imp_c.impa.evt"},
	{0x5F04259A2B476C3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforward.gad"},
	{0x5F042A88CD2ABFEAULL, "233640004.ogg"},
	{0x5F048C3DEBE4B7E1ULL, "zz_followers_banter6157803.cl"},
	{0x5F04AE312474B41AULL, "70380790.ogg"},
	{0x5F04B4641B13DCD6ULL, "w_encounter3_mmi.gff"},
	{0x5F04F055975BC7B8ULL, "vxlt_0_lowtown_h9b_0_0.vlm"},
	{0x5F050FBF21EE86F4ULL, "47571734.ogg"},
	{0x5F0594602135C1CAULL, "vxht_bartrand_n_73m_0_0.vlm"},
	{0x5F05F881D9C5C553ULL, "6060378_f.fxe"},
	{0x5F069660C1829D6FULL, "977307170.ogg"},
	{0x5F06993056B96553ULL, "art\\vfx\\combat\\vfx_glb_ele_p_minor_c.impa.gad"},
	{0x5F06DE8F973540F6ULL, "hlac_lt_0_lowtown_n_-262144_1y.dds"},
	{0x5F06EA6B460B125CULL, "llac_lt_undercitydungeon_-1_27.dds"},
	{0x5F0729F687F3B595ULL, "357068703.ogg"},
	{0x5F077DED41D77ED4ULL, "follower_banter6163188.cl"},
	{0x5F079379226D05B2ULL, "vxca_cavern1_haunted_cod_0_0.vlm"},
	{0x5F07A1BD98CCC4E3ULL, "6072214_f.fxe"},
	{0x5F07FFA8DB930A88ULL, "vxw_sundermount_32d_0_0.vlm"},
	{0x5F080EBC27335A8CULL, "vxlt_0_lowtown_a2_n_c_huk_0_0.vlm"},
	{0x5F087D6645404D29ULL, "hightown_fem_cityguard_mob6168483.cl"},
	{0x5F0905540EC358FFULL, "ttx_roof_flat_2.msh"},
	{0x5F090B540EC362A1ULL, "ttx_roof_flat_2.mmh"},
	{0x5F097E7EF2F73CBCULL, "6116708_f.fxe"},
	{0x5F0992566E10A827ULL, "6165261_m.fxe"},
	{0x5F09A3C6AA21E668ULL, "vxw_sundermount_2z8_0_1.vlm"},
	{0x5F0A045BF32C1A45ULL, "vxca_caverns1_e85_0_0.vlm"},
	{0x5F0ABB7CA4FD650CULL, "prp_str_bark_n.dds"},
	{0x5F0AC9FBD2F59F75ULL, "6122086_m.fxe"},
	{0x5F0AFE8566DD4F88ULL, "art\\vfx\\combat\\vfx_f_avl_bdygrd_imp_c.impa.gad"},
	{0x5F0B0D19D09A9446ULL, "vxdr_primevaltomb_70e_0_0.vlm"},
	{0x5F0B249A2B4D893AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforward.ani"},
	{0x5F0B469E3FBCE3CFULL, "dae000_gamlen6026204.cl"},
	{0x5F0B5F75D34E3C22ULL, "zz_qun_debug6206597.cl"},
	{0x5F0BA903BC510E1BULL, "69643009.ogg"},
	{0x5F0BAD1EF514979BULL, "927277850.ogg"},
	{0x5F0BCAB5C7257AD3ULL, "vxlt_gallowstemplar_n_536_0_0.vlm"},
	{0x5F0C1C590806963FULL, "810465812.ogg"},
	{0x5F0C2431F8D89F22ULL, "mcr370_negotiation6080206.cl"},
	{0x5F0C7B14BEE91153ULL, "1560418.ogg"},
	{0x5F0C95E173ED72C0ULL, "576069410.ogg"},
	{0x5F0C963056BDCBB8ULL, "art\\vfx\\combat\\vfx_glb_ele_p_minor_c.impa.evt"},
	{0x5F0C9C6050C5820CULL, "548634259.ogg"},
	{0x5F0C9EE9F99B3FB6ULL, "290836280.ogg"},
	{0x5F0CC025993191E6ULL, "545800792.ogg"},
	{0x5F0CD18184D0B9D2ULL, "hlac_lt_warehouse_n_-65536_1y.dds"},
	{0x5F0D2FE15DA8D9B6ULL, "vxlt_gallowstemplar_n_5mm_0_0.vlm"},
	{0x5F0D5E6402DF9D0CULL, "vxlt_hangedman_n_20z_0_0.vlm"},
	{0x5F0D70080E465978ULL, "vxca_caverns1_mines_eaj_0_0.vlm"},
	{0x5F0D8EF3877BAFE9ULL, "vxlt_undercityencounter_1bv_0_0.vlm"},
	{0x5F0DA961B0009FFFULL, "hlac_lt_gallowsprison_n_-1_8x.dds"},
	{0x5F0DEEC8BE1F41ECULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em04.msh"},
	{0x5F0DF4C8BE1F4BDEULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em04.mmh"},
	{0x5F0E35B40B342EA2ULL, "vxca_caverns2_62q_0_1.vlm"},
	{0x5F0E3D03F007B321ULL, "hlac_lt_gallowstemplar_n_2395_1x.dds"},
	{0x5F0E441B33275D00ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypoucha_sa.dds"},
	{0x5F0E8B19525D596AULL, "zz_vault_debug6200344.cl"},
	{0x5F0EB4E2E4D4114BULL, "hlac_w_woundedcoast_65540_0x.dds"},
	{0x5F0F2A3407E2B832ULL, "vxht_0_hightown_a2_cbn_0_0.vlm"},
	{0x5F0F96B6D4846AA1ULL, "765521061.ogg"},
	{0x5F0FDCE4268C9717ULL, "bu_dark_f2.mao"},
	{0x5F0FF7BE6FE948EEULL, "6084208_m.fxe"},
	{0x5F1047E1AC21053DULL, "art\\characters\\playercharacter\\humanmale\\hlh_medc.mao"},
	{0x5F104BE1AC210C76ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medc.mmh"},
	{0x5F104C3E48F2541FULL, "mcr310_sibling6076331.cl"},
	{0x5F1055E1AC211D74ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medc.msh"},
	{0x5F10B03CA997B3F6ULL, "6106407_m.fxe"},
	{0x5F10C6782F5436B8ULL, "dae011_badpoet_sta6145175.cl"},
	{0x5F10EAB4CAD4FC72ULL, "vxht_keep_64m_0_0.vlm"},
	{0x5F114FB5A7535AB6ULL, "ico_helm_medium_5.dds"},
	{0x5F11C076BC2728ECULL, "261739369.ogg"},
	{0x5F1213E8440EC6B8ULL, "6089568_m.fxe"},
	{0x5F124B35612B6587ULL, "233363364.ogg"},
	{0x5F1263EB3DC77CD5ULL, "vxlt_gallowstemplar_n_6t6_0_0.vlm"},
	{0x5F129DB018A1732AULL, "rune_weapon_fire.dds"},
	{0x5F12A8D221BC7DEFULL, "vxlt_hangedman_1vc_0_0.vlm"},
	{0x5F12D4384B1BEE71ULL, "lt_0_lowtown_52_0.anb"},
	{0x5F12E423D3C4CB24ULL, "vxht_chantry_n_32w_0_0.vlm"},
	{0x5F12FCAFC2166EDBULL, "lt_0_lowtown_docks_232_0.mmh"},
	{0x5F1308AFC2168378ULL, "lt_0_lowtown_docks_232_0.mao"},
	{0x5F1316B68B3D2D99ULL, "6118431_f.fxe"},
	{0x5F131AAFC216A1C5ULL, "lt_0_lowtown_docks_232_0.msh"},
	{0x5F13A5C051F42509ULL, "vxw_woundedcoast_1cp_0_0.vlm"},
	{0x5F13BA7FDC52CC95ULL, "dae000_worthy6166600.cl"},
	{0x5F1488E88B1D2606ULL, "gen00fl_aveline6132827.cl"},
	{0x5F14B3F7B4BAD2FFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\abomination_movement.anb"},
	{0x5F14D5908C442172ULL, "6149055_m.fxe"},
	{0x5F14ECC2DBE56533ULL, "6086162_m.fxe"},
	{0x5F15354493DB2DC3ULL, "854239200.ogg"},
	{0x5F154B6B887A1952ULL, "vxlt_gallowstemplar_n_533_0_0.vlm"},
	{0x5F15636006C11C3FULL, "hlac_ca_caverns2prc_5802_56z.dds"},
	{0x5F15C231256D4D2BULL, "6111675_m.fxe"},
	{0x5F160149DEF49289ULL, "6591796.ogg"},
	{0x5F1644A0DD647867ULL, "254346593.ogg"},
	{0x5F167C11F4031B06ULL, "vxlt_0_lowtown_equ_0_0.vlm"},
	{0x5F1680991929730CULL, "vxlt_0_lowtown_a2_n_c_hph_0_0.vlm"},
	{0x5F16E2C61DC37212ULL, "755657258.ogg"},
	{0x5F180C6B296B276CULL, "vxw_encounter2_camp_p8_0_0.vlm"},
	{0x5F1872C0500C955BULL, "dre000_dougal6129777.cl"},
	{0x5F18BDCD54BFADC3ULL, "art\\characters\\weapons\\zzdao\\w_lsw_ls01a.msh"},
	{0x5F18CBCD54BFC51DULL, "art\\characters\\weapons\\zzdao\\w_lsw_ls01a.mmh"},
	{0x5F18CFCD54BFCC2EULL, "art\\characters\\weapons\\zzdao\\w_lsw_ls01a.mao"},
	{0x5F19C6AF1F120274ULL, "vxht_0_hightown_a3_aij_0_0.vlm"},
	{0x5F1A8645509885DEULL, "6001334_f.fxe"},
	{0x5F1A8A5D7FB01344ULL, "364071070.ogg"},
	{0x5F1A9C19E4E2B9B3ULL, "gen00fl_varric6033681.cl"},
	{0x5F1B7FC11F3C447EULL, "vxht_chantry_n_2y5_0_0.vlm"},
	{0x5F1C0674887FDAD6ULL, "hlac_lt_undercityencounter_-1_19z.dds"},
	{0x5F1CCC36D44E74C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\facial"},
	{0x5F1D601F3AE05638ULL, "character_character_p199.psh.fxo.dx11.assembly"},
	{0x5F1E0C2EDD8772E5ULL, "6074363_m.fxe"},
	{0x5F1E18CA23415894ULL, "6103562_f.fxe"},
	{0x5F1E3113CA61AFF9ULL, "zz_seb_debug6216546.cl"},
	{0x5F1E479A248A4B18ULL, "207154854.ogg"},
	{0x5F1E6AE9E1F86B04ULL, "29928101.ogg"},
	{0x5F1E894C26E2F9FFULL, "w_sundermount_111_0.rml"},
	{0x5F1F5C91BCCB14F2ULL, "vxht_keep_61n_0_0.vlm"},
	{0x5F1FA2E804245310ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_appd_0d.dds"},
	{0x5F1FB2223AFAF976ULL, "250915231.ogg"},
	{0x5F20F2C422092920ULL, "l_env_wal_hightown_int_crowd_bed.bnk"},
	{0x5F213CBC56CC4CD2ULL, "232648346.ogg"},
	{0x5F21D29436E110EBULL, "dae000_maren6148422.cl"},
	{0x5F21F9FFD526D3F3ULL, "6068531_m.fxe"},
	{0x5F220B46CD56E620ULL, "mag320_orsino6071151.cl"},
	{0x5F22179D43EA3D89ULL, "vxw_woundedcoast_1nw_0_0.vlm"},
	{0x5F22343BCC0E29B5ULL, "llac_w_woundedcoast_-262140_0.dds"},
	{0x5F2245074D928F24ULL, "6084714_m.fxe"},
	{0x5F22665DCF03CD7CULL, "6037122_f.fxe"},
	{0x5F2278198360F981ULL, "vxlt_0_lowtown_n_eki_0_2.vlm"},
	{0x5F23476D7E3A39F2ULL, "6142155_m.fxe"},
	{0x5F23BA4269BC5E5BULL, "gen00fl_fenris6126925.cl"},
	{0x5F23F00EB8DEBA1FULL, "6080281_m.fxe"},
	{0x5F2466F42588FC3AULL, "vxlt_darktown_4mk_0_0.vlm"},
	{0x5F250BDED3799454ULL, "6085048_f.fxe"},
	{0x5F252F2EDDE00886ULL, "6108737_m.fxe"},
	{0x5F259796DD170DE5ULL, "6131161_m.fxe"},
	{0x5F259797CB91E9AFULL, "560609715.ogg"},
	{0x5F25EB82F8125950ULL, "ts_templars_acc_table.gda"},
	{0x5F25F8C8C1BE9918ULL, "dae000_arishok6072829.cl"},
	{0x5F26007A0085F4A9ULL, "6126536_m.fxe"},
	{0x5F26108054F87C78ULL, "vxdr_dungeon_11x_0_0.vlm"},
	{0x5F265634B14FAC6DULL, "vxlt_0_lowtown_a2_n_c_gos_0_0.vlm"},
	{0x5F267E481B613FECULL, "vxw_encounter2_camp_k9_0_0.vlm"},
	{0x5F26A191BBD2FBBEULL, "and230_mage_girl6061847.cl"},
	{0x5F26B90A6B0730F0ULL, "1002818873.ogg"},
	{0x5F26F09D1F302428ULL, "6116040_m.fxe"},
	{0x5F277FD3EDF6CCA3ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.impa.ani"},
	{0x5F27D017E44D600BULL, "6095294_m.fxe"},
	{0x5F27F1815495C7D7ULL, "follower_banter6163462.cl"},
	{0x5F28191614C7F171ULL, "158704617.ogg"},
	{0x5F2827C96DB5ECCFULL, "one100pt_gamlen.ncs"},
	{0x5F28388C11081AF4ULL, "vxht_0_hightown_a3_avk_0_0.vlm"},
	{0x5F287E7A7DD0FB67ULL, "vxlt_0_lowtown_a2_gi2_0_0.vlm"},
	{0x5F2895E0B0AD04C8ULL, "6059345_m.fxe"},
	{0x5F2951CA5C17ACACULL, "594446581.ogg"},
	{0x5F29E5E39997AFD5ULL, "6152618_m.fxe"},
	{0x5F29EEB34F67ABF4ULL, "vxht_0_hightown_ccs_0_0.vlm"},
	{0x5F2AD1DF1280E860ULL, "6149017_m.fxe"},
	{0x5F2BB25287A56EE5ULL, "art\\vfx\\environment\\vfx_trap_pot_p.anb"},
	{0x5F2BCF786549AEC0ULL, "470525115.ogg"},
	{0x5F2C00CF44B5A668ULL, "vxlt_0_lowtown_n_ekl_0_2.vlm"},
	{0x5F2C1284F07214BAULL, "6064523_m.fxe"},
	{0x5F2CB6483C84F882ULL, "vxlt_gallowstemplar_f_45t_0_0.vlm"},
	{0x5F2CCD7C66B4C959ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_dksa_0n.dds"},
	{0x5F2D00C98E3F391BULL, "vxht_0_hightown_a3_a82_0_0.vlm"},
	{0x5F2D099E96B7360CULL, "ave331_jeven6137308.cl"},
	{0x5F2D26E85071F8D7ULL, "6030642_m.fxe"},
	{0x5F2D8636804C9014ULL, "673052274.ogg"},
	{0x5F2D96977E34E52DULL, "206351701.ogg"},
	{0x5F2E7944898E8340ULL, "mcr371_hostile_merrill6082165.cl"},
	{0x5F2F5B57208E45FAULL, "hlac_ca_caverns1_mines_24423_10x.dds"},
	{0x5F2F66935B1FAAE0ULL, "gen00fl_aveline6006065.cl"},
	{0x5F2F69FE5FAFCF5AULL, "vxht_bartrand_n_326_0_0.vlm"},
	{0x5F2F8C8C05A8AEEFULL, "vxlt_hangedman_1xm_0_0.vlm"},
	{0x5F30719C0EEFA6CCULL, "qun110_petrice6041271.cl"},
	{0x5F309D07435429ADULL, "art\\characters\\playercharacter\\humanfemale\\hlf_meda.msh"},
	{0x5F30AF0743544833ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_meda.mmh"},
	{0x5F31496822AA3F63ULL, "20132203.ogg"},
	{0x5F315D6CF37F79F6ULL, "866417226.ogg"},
	{0x5F319BA364726282ULL, "dae000_gamlen6149427.cl"},
	{0x5F31D86440C4005BULL, "6136777_m.fxe"},
	{0x5F326C61133EB101ULL, "829033844.ogg"},
	{0x5F33C01963103B7AULL, "dae000_arishok6067638.cl"},
	{0x5F34863C9C5F896FULL, "dae000cr_grand_cleric6147723.cl"},
	{0x5F34AA867A8A29C4ULL, "llac_ca_cavern1_haunted_24424_22.dds"},
	{0x5F34C617F0D9A661ULL, "6018677_m.fxe"},
	{0x5F34C811A345C4EDULL, "vxlt_0_lowtown_a2_n_c_ghr_0_0.vlm"},
	{0x5F34D705EEB4A805ULL, "zz_vault_debug6203173.cl"},
	{0x5F353D2015E4823AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_basic.mmh"},
	{0x5F35412015E48971ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_basic.mao"},
	{0x5F354F2015E4A080ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_basic.msh"},
	{0x5F3597D3EE03329AULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.impa.evt"},
	{0x5F3678C6F6581276ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_exit.ani"},
	{0x5F37C9AB517994F1ULL, "426901184.ogg"},
	{0x5F37D170996B268DULL, "6092110_m.fxe"},
	{0x5F37EF1FCF1E80DFULL, "821800848.ogg"},
	{0x5F384F0129DAB453ULL, "vxht_0_hightown_n_5w5_0_0.vlm"},
	{0x5F3851BFE9CA595EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_exit.evt"},
	{0x5F38BBDF43302482ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottlec.msh"},
	{0x5F38BDDF43302838ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottlec.mmh"},
	{0x5F38BED9332A7DDBULL, "6074075_m.fxe"},
	{0x5F38C9DF43303CDBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottlec.mao"},
	{0x5F38E0D46D9E81C1ULL, "vxlt_0_lowtown_a3_n_c_gmh_0_0.vlm"},
	{0x5F3943BC047700E0ULL, "6142117_m.fxe"},
	{0x5F396213A4F9C821ULL, "6052103_m.fxe"},
	{0x5F39A247E2AF8382ULL, "art\\characters\\weapons\\staff\\w_stf_017a.msh"},
	{0x5F39A447E2AF8738ULL, "art\\characters\\weapons\\staff\\w_stf_017a.mmh"},
	{0x5F39B047E2AF9BDBULL, "art\\characters\\weapons\\staff\\w_stf_017a.mao"},
	{0x5F39F32EA57411EAULL, "hlac_lt_gallowscourtyard_-2_1x.dds"},
	{0x5F3A0DF8DA74645DULL, "vxlt_0_lowtown_docks_a3_n_cp2_0_0.vlm"},
	{0x5F3A1A18EA15A9F0ULL, "w_woundedcoast_126_0.msh"},
	{0x5F3A2818EA15C10AULL, "w_woundedcoast_126_0.mmh"},
	{0x5F3A2C18EA15C841ULL, "w_woundedcoast_126_0.mao"},
	{0x5F3A6A336E6001C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_loop.ani"},
	{0x5F3A729804A2DA7BULL, "vxlt_undercitydungeon_12e_0_0.vlm"},
	{0x5F3B0ADD0E4A77EAULL, "one000_ilen6058709.cl"},
	{0x5F3B1220A30A18A8ULL, "6125311_f.fxe"},
	{0x5F3BC6F6869CEDFCULL, "221882982.ogg"},
	{0x5F3C96D3EE094F8DULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.impa.gad"},
	{0x5F3D42EED620CD9EULL, "6121101_m.fxe"},
	{0x5F3D7AE6E4507FCCULL, "888949707.ogg"},
	{0x5F3D8CE483BEA519ULL, "vxlt_0_lowtown_a2_ekf_0_2.vlm"},
	{0x5F3E2C405E6EB3A4ULL, "dae100_viscount6118519.cl"},
	{0x5F3E46615836C4C5ULL, "cd_dre000_cut_arrivaldeeproads.bnk"},
	{0x5F3EA47DB3D18345ULL, "vxlt_blackemporium_rq_0_1.vlm"},
	{0x5F3EDBD995070CCCULL, "dae000ip_crowd_mf1_talk12.utp"},
	{0x5F3F68BFE9D09E89ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_exit.gad"},
	{0x5F403BD127CAB6EBULL, "706949521.ogg"},
	{0x5F4043654DACF55DULL, "mag311_emile_de_launcet6171507.cl"},
	{0x5F40BD2E6FEAD16FULL, "vxht_mansion_n_5zn_0_0.vlm"},
	{0x5F40C27294C5166FULL, "zz_followers_banter6159201.cl"},
	{0x5F417C3A85E91762ULL, "dae000_gamlen6139550.cl"},
	{0x5F423B939DD42C59ULL, "6124761_m.fxe"},
	{0x5F424A64783F2E93ULL, "964642848.ogg"},
	{0x5F4274E7DC0B0749ULL, "hlac_lt_0_lowtown_docks_a3_-131076_1z.dds"},
	{0x5F4289EEF7DB4AE5ULL, "gen00fl_merrill6106998.cl"},
	{0x5F42C40F2DDF49BFULL, "vxw_encounter2_camp_p7_0_0.vlm"},
	{0x5F431525DC5EDB8DULL, "vxht_0_hightown_bz4_0_0.vlm"},
	{0x5F434EB2446ECB8EULL, "hlac_lt_0_lowtown_docks_n_-131075_1z.dds"},
	{0x5F4380C6F662AA67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_exit.evt"},
	{0x5F43AB47E2B7E7C3ULL, "art\\characters\\weapons\\staff\\w_stf_017a.phy"},
	{0x5F43CAD6E794AAD6ULL, "zz_and_debug6134105.cl"},
	{0x5F43E9D54ED9E9F3ULL, "gen_im_trs_slv_hvy_glv_r0.uti"},
	{0x5F44030410EB1AD5ULL, "clf360_evets6072211.cl"},
	{0x5F44FF562A216C42ULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_dwn_c.impa.gad"},
	{0x5F450EAB09AB82A0ULL, "gen_im_arm_cht_elf_comb_t4.uti"},
	{0x5F451361AFD3653AULL, "898044719.ogg"},
	{0x5F456E31A0E6AE57ULL, "vxlt_warehouse_b2_0_0.vlm"},
	{0x5F45A8B555F9CD74ULL, "6104432_m.fxe"},
	{0x5F4639BFE9D66DD7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_exit.ani"},
	{0x5F466F89B4A70EEAULL, "dae000ip_crowd_mf3_talk12.utp"},
	{0x5F468B2DD31449D8ULL, "6113285_m.fxe"},
	{0x5F46964FCF07EED1ULL, "6162987_m.fxe"},
	{0x5F46C7B0CCAB9B18ULL, "6087247_m.fxe"},
	{0x5F47003A68ECBCA7ULL, "vxw_blightlands_eg_0_0.vlm"},
	{0x5F47B83CF19062BAULL, "6129314_m.fxe"},
	{0x5F47FE28F9FD975AULL, "w_sundermount_45_0.mmh"},
	{0x5F480228F9FD9D91ULL, "w_sundermount_45_0.mao"},
	{0x5F480DC356C408E3ULL, "vxdr_0_deeproads_eug_0_0.vlm"},
	{0x5F481028F9FDB5A0ULL, "w_sundermount_45_0.msh"},
	{0x5F48E474F698F2FBULL, "vxlt_undercitydungeon_15d_0_0.vlm"},
	{0x5F497344CB3566B7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45winup.phy"},
	{0x5F4995550DB9EF97ULL, "ef_mrl351cr_fakefemaleelf.mor"},
	{0x5F49E6568D014614ULL, "6148164_m.fxe"},
	{0x5F4A39767CA42385ULL, "vxca_caverns1_mines_efr_0_0.vlm"},
	{0x5F4A8B9C43EA0D00ULL, "6109436_f.fxe"},
	{0x5F4AA1C6F66900B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action_exit.gad"},
	{0x5F4AAC71E9A2EF29ULL, "hlac_dr_slavers_3216_9x.dds"},
	{0x5F4AB6FDEADE06CCULL, "mcr370cr_gate_guardian.ncs"},
	{0x5F4B0632BCB2A1B5ULL, "one120_athenril6088233.cl"},
	{0x5F4B6385FB5CC55BULL, "6168125_m.fxe"},
	{0x5F4BA756514C0587ULL, "vrc221_hugin6042726.cl"},
	{0x5F4BC21447FBF3AEULL, "vxw_encounter2_camp_33_0_0.vlm"},
	{0x5F4BD6A00D6E1A37ULL, "hlac_ht_mansion2_818_61x.dds"},
	{0x5F4BE3CE703328A1ULL, "359271749.ogg"},
	{0x5F4BEC35F74FE15BULL, "328701219.ogg"},
	{0x5F4C22562A27C639ULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_dwn_c.impa.evt"},
	{0x5F4C78909175F0B0ULL, "qun210cr_guard.cnv"},
	{0x5F4CD85B2CE3743AULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_app_p.anb"},
	{0x5F4CE319609B9CE8ULL, "gen_im_cns_cft_pot_06.uti"},
	{0x5F4CE3AB9765E145ULL, "312609781.ogg"},
	{0x5F4D508F1DE27A80ULL, "6076449_m.fxe"},
	{0x5F4D7C424831EDF0ULL, "vxdr_0_deeproads_cfu_0_2.vlm"},
	{0x5F4DCBC8BE54EA79ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em04.phy"},
	{0x5F4E65B377B1D8F1ULL, "6089827_f.fxe"},
	{0x5F4EE924D37C7A77ULL, "6209855_m.fxe"},
	{0x5F4F2C28841E3A1BULL, "nat300ip_barrel_lever.utp"},
	{0x5F4F3599B41D8B32ULL, "dae000_pc_home_ambients6208356.cl"},
	{0x5F4F36D3F89E39AEULL, "1061944568.ogg"},
	{0x5F4F53336E7236C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_loop.gad"},
	{0x5F4F7CCB1EDDB0AEULL, "518454193.ogg"},
	{0x5F4F9A9091786008ULL, "qun210cr_guard.bnk"},
	{0x5F5114C4C2E52296ULL, "llac_ht_keep_3732_4.dds"},
	{0x5F5135EC1FD5623FULL, "vxw_encounter2_camp_k6_0_0.vlm"},
	{0x5F5176E0245BCE06ULL, "6173029_m.fxe"},
	{0x5F51C33E91D2A35DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsbuild.mao"},
	{0x5F51C73E91D2A996ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsbuild.mmh"},
	{0x5F51D13E91D2BA94ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsbuild.msh"},
	{0x5F51E53D3EE484E7ULL, "6004120_m.fxe"},
	{0x5F520DFDD41FBCABULL, "6075234_m.fxe"},
	{0x5F52AE5535AF6E4BULL, "651774856.ogg"},
	{0x5F52AEC87849717AULL, "6166565_f.fxe"},
	{0x5F52C28BF5CDB72FULL, "vxlt_gallowstemplar_f_433_0_0.vlm"},
	{0x5F52EA7D73C69DB8ULL, "vxca_cavern1_haunted_edu_0_0.vlm"},
	{0x5F538B8AC7286E12ULL, "6168487_m.fxe"},
	{0x5F53CE30FEE8F503ULL, "6114830_m.fxe"},
	{0x5F540752D323EBE2ULL, "222371302.ogg"},
	{0x5F54618118A98038ULL, "hlac_ht_chantry_2534_11y.dds"},
	{0x5F54B35287C87526ULL, "art\\vfx\\environment\\vfx_trap_pot_p.mmh"},
	{0x5F55413EC6D0EF55ULL, "fen210_orana6133035.cl"},
	{0x5F5572175AE2D527ULL, "vxw_blightlands_nf_0_0.vlm"},
	{0x5F5579EFA2AAF8FDULL, "vxdr_primevaltomb_8r7_0_0.vlm"},
	{0x5F56001598340927ULL, "87923785.ogg"},
	{0x5F56F37BFB989310ULL, "hlac_ht_0_hightown_-65535_1y.dds"},
	{0x5F574C8FB194413DULL, "scrolladditive.mao"},
	{0x5F57E7B3B5E645D7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetal_sa.dds"},
	{0x5F57FBBBFF0490D6ULL, "vxlt_warehouse_n_ec_0_0.vlm"},
	{0x5F58293C05DF06E8ULL, "631921251.ogg"},
	{0x5F584F9E178BE832ULL, "gen00fl_merrill6122214.cl"},
	{0x5F58CA7AA38FFA59ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_gesture_left.gad"},
	{0x5F58FD7935E02580ULL, "6102536_f.fxe"},
	{0x5F5905537E411F26ULL, "llac_w_sundermount_262146_0.dds"},
	{0x5F590E5A6C62574AULL, "6127411_m.fxe"},
	{0x5F5927562ED94AB3ULL, "6162044_m.fxe"},
	{0x5F592A562A325E28ULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_dwn_c.impa.ani"},
	{0x5F59BA4B620C9C60ULL, "rdr111dg_martin6054311.cl"},
	{0x5F59BE9CDD479F66ULL, "mag110_samson6020784.cl"},
	{0x5F5A5C201603D345ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ri_basic.phy"},
	{0x5F5AA425359DEEE4ULL, "vxw_encounter2_3q_0_0.vlm"},
	{0x5F5AC12728D95CEDULL, "sky_lt_ruinedhouse_n.mao"},
	{0x5F5B225F9A74F9EFULL, "9131364.ogg"},
	{0x5F5B26E7A3A52E9BULL, "926018098.ogg"},
	{0x5F5C163B9933A481ULL, "vxdr_0_deeproads_fiz_0_0.vlm"},
	{0x5F5C6447185EB403ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.dura.evt"},
	{0x5F5CA6BC94E9B9B1ULL, "vxw_bonepitmining_a2_7r_0_0.vlm"},
	{0x5F5CF2ADE8A87F98ULL, "character_character_p117.psh.fxo"},
	{0x5F5D193EA57E04FBULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_enter.ani"},
	{0x5F5D3AAB78C16E51ULL, "gen_im_fla_sebastian_01.uti"},
	{0x5F5D5240010F7F90ULL, "gen00fl_dog6167029.cl"},
	{0x5F5D6D56E68D7031ULL, "and111_lirene6094268.cl"},
	{0x5F5E4D35B2FB7451ULL, "hlac_lt_0_lowtown_n_65536_4z.dds"},
	{0x5F5E838FDA09AC32ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower_mcr370_playerrules.gad"},
	{0x5F5EBF9767285F88ULL, "qcr000_arishok6042501.cl"},
	{0x5F5F3991CAAAE36BULL, "14233507.ogg"},
	{0x5F5F4B3627BD9CA8ULL, "killcreature.ncs"},
	{0x5F5F92C8681999F3ULL, "vxca_caverns2_8gh_0_0.vlm"},
	{0x5F5FA7C911CB7ABFULL, "vxw_encounter2_camp_f1_0_0.vlm"},
	{0x5F5FDB7AA39635A7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_gesture_left.ani"},
	{0x5F604128842CD13EULL, "nat300ip_barrel_lever.ncs"},
	{0x5F604F3B7EA9363FULL, "follower_banter6141835.cl"},
	{0x5F613468E7C3CFAFULL, "vxlt_gallowstemplar_f_482_0_0.vlm"},
	{0x5F6186BC8FAA42C9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartalentwhirlwind.ani"},
	{0x5F61AF12BC67C1D0ULL, "6131348_m.fxe"},
	{0x5F627130A9A5762CULL, "vxlt_0_lowtown_docks_a3_cxt_0_0.vlm"},
	{0x5F628188E7698F1DULL, "dae000_oldmangossip6175771.cl"},
	{0x5F62AED0F05D96E9ULL, "khg.cut"},
	{0x5F630D2F8582A718ULL, "166353354.ogg"},
	{0x5F633D4718649084ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.dura.gad"},
	{0x5F634B2BD62BD479ULL, "zz_dae_debug6131009.cl"},
	{0x5F63917FEF349941ULL, "6053885_m.fxe"},
	{0x5F63E3F44CD8EDA7ULL, "vxw_blightlands_sa_0_0.vlm"},
	{0x5F63F434972B51C7ULL, "hlac_lt_undercityserialkiller_-1_10z.dds"},
	{0x5F63FDA382E81BD8ULL, "6086958_m.fxe"},
	{0x5F6425255A6F9877ULL, "289693770.ogg"},
	{0x5F6435B821BC56D8ULL, "6093996_m.fxe"},
	{0x5F643E0CF6D4AFD5ULL, "vxht_0_hightown_a3_bal_0_0.vlm"},
	{0x5F643EE9B67A42C8ULL, "gen00fl_merrill6128720.cl"},
	{0x5F6488C41004A227ULL, "war_warcry.dds"},
	{0x5F64AEE1AD6B3657ULL, "vxht_mansion2_309_0_0.vlm"},
	{0x5F64C27D9957243FULL, "6120789_m.fxe"},
	{0x5F65882D6B7BE323ULL, "6046922_m.fxe"},
	{0x5F65D64D454DEA84ULL, "563689382.ogg"},
	{0x5F66067C6A86988BULL, "zz_followers_banter6158682.cl"},
	{0x5F663BCDC572AFD0ULL, "vxlt_gallowstemplar_46d_0_0.vlm"},
	{0x5F6649E257297C86ULL, "t1_mub_bk1.tnt"},
	{0x5F66742CBBEF65CFULL, "375289430.ogg"},
	{0x5F6685CF96E80706ULL, "llac_w_bonepitdestroyed_a3_0_4.dds"},
	{0x5F66C37F135E1EBEULL, "rot_brain.mao"},
	{0x5F671EC3EAF05CEFULL, "llac_dr_slavers_3207_1.dds"},
	{0x5F678ED6AF699412ULL, "326695337.ogg"},
	{0x5F67A1BC8FAEDC4BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartalentwhirlwind.gad"},
	{0x5F681DEBCB9DEA29ULL, "hlac_ht_0_hightown_-65536_4z.dds"},
	{0x5F6882B7268409EFULL, "zz_followers_banter6157526.cl"},
	{0x5F68BC03B897D471ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxd.mmh"},
	{0x5F68C803B897E812ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxd.mao"},
	{0x5F68D603B898002FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxd.msh"},
	{0x5F68E077B2863B94ULL, "vxlt_0_lowtown_a3_n_c_fip_0_0.vlm"},
	{0x5F699F74BE10BCFCULL, "em_hireling_1.mor"},
	{0x5F69DFC28E30FE73ULL, "6083745_m.fxe"},
	{0x5F6A4C47186AC962ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.dura.ani"},
	{0x5F6A742F63784D55ULL, "lt_0_lowtown_n_22_0.rml"},
	{0x5F6AB1AC6CA644ADULL, "vxlt_0_lowtown_a2_n_c_hd1_0_0.vlm"},
	{0x5F6AEA91EAC7BD9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_down_to_up.gad"},
	{0x5F6AFE3A0F25EA8BULL, "vxdr_0_deeproads_cmu_0_0.vlm"},
	{0x5F6B213EA58A4FF2ULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_enter.evt"},
	{0x5F6B34527D7DF51EULL, "llac_ht_0_hightown_a3_-65536_0.dds"},
	{0x5F6B348AB40A5484ULL, "6068792_m.fxe"},
	{0x5F6B77521147B45DULL, "6135342_f.fxe"},
	{0x5F6B8A088252BC2EULL, "6144737_m.fxe"},
	{0x5F6BCD1175F390EAULL, "6155991_m.fxe"},
	{0x5F6C7D4C272506EFULL, "w_sundermount_111_0.mao"},
	{0x5F6C894C27251A8CULL, "w_sundermount_111_0.mmh"},
	{0x5F6C8F4C2725257EULL, "w_sundermount_111_0.msh"},
	{0x5F6D651EA65B48F0ULL, "6021135_m.fxe"},
	{0x5F6DA4AE4D9A2405ULL, "gen00fl_anders6103744.cl"},
	{0x5F6E04A55A0FB273ULL, "vxca_caverns2_8bk_0_0.vlm"},
	{0x5F6E19A603C1933FULL, "vxw_encounter2_camp_a0_0_0.vlm"},
	{0x5F6E1D28FA1E9B65ULL, "w_sundermount_45_0.phy"},
	{0x5F6E9793426FE909ULL, "lgt320_shw_ors_secret_meet6156363.cl"},
	{0x5F6E9EBC8FB4F5B0ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartalentwhirlwind.evt"},
	{0x5F6EB417E4AAEDCAULL, "mcr370_meredith6083116.cl"},
	{0x5F6EF66769610B38ULL, "vxht_chantry_n_2ce_0_0.vlm"},
	{0x5F6F0D7D8BD90CA5ULL, "1031755744.ogg"},
	{0x5F6F126DDB5B70F5ULL, "w_bonepitmining_a2_87_0.msh"},
	{0x5F6F146DDB5B746BULL, "w_bonepitmining_a2_87_0.mmh"},
	{0x5F6F206DDB5B8808ULL, "w_bonepitmining_a2_87_0.mao"},
	{0x5F6F8B9B06FD72A8ULL, "vxlt_hangedman_m0_0_0.vlm"},
	{0x5F6F8BBA66EB8350ULL, "vxlt_0_lowtown_docks_as0_0_0.vlm"},
	{0x5F6FB5A442241123ULL, "llac_dr_primevaltomb_9621_1.dds"},
	{0x5F6FEB7115A29848ULL, "572002754.ogg"},
	{0x5F70844C36D7785FULL, "one000_ilen6056876.cl"},
	{0x5F70EB9D0E0BB7CDULL, "vxca_cavern1_haunted_fmd_0_0.vlm"},
	{0x5F713E7D90ABEA69ULL, "vxlt_0_lowtown_docks_a3_n_azr_0_0.vlm"},
	{0x5F71403EA58EEF85ULL, "art\\characters\\creatures\\desiredemon\\animation\\kd_enter.gad"},
	{0x5F7149F2DB900F06ULL, "111204106.ogg"},
	{0x5F7150F81128D733ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.impa.evt"},
	{0x5F7198547130BE73ULL, "dae000_bodhan6146834.cl"},
	{0x5F71BA026C27141CULL, "vxlt_warehouse_n_p6_0_0.vlm"},
	{0x5F71F530F99F0291ULL, "6165234_m.fxe"},
	{0x5F72112462624F64ULL, "6154033_m.fxe"},
	{0x5F721BE35AF58E1FULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_cb.phy"},
	{0x5F726E8FDA1A31D8ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower_mcr370_playerrules.ani"},
	{0x5F727E175B2E3552ULL, "46389466.ogg"},
	{0x5F72B8CC297D1938ULL, "gen_im_arm_hlm_hvy_act2_01.uti"},
	{0x5F72F5955BCE5D41ULL, "sky_lt_gallowstemplar.mao"},
	{0x5F733251EA722755ULL, "6053711_m.fxe"},
	{0x5F7373A35F1E9029ULL, "309025183.ogg"},
	{0x5F73C99D10B66CC0ULL, "vxlt_hangedman_n_59_0_0.vlm"},
	{0x5F742BEA4386CCCFULL, "dae000cr_arishok_duel.ncs"},
	{0x5F747E9F2D2D1476ULL, "382603000.ogg"},
	{0x5F74DF75E2F0C1D6ULL, "vxlt_warehouse_n_oe_0_0.vlm"},
	{0x5F74FE28FA248571ULL, "w_sundermount_45_0.rml"},
	{0x5F752262A0C8F760ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_appd_0t.dds"},
	{0x5F75372339CE87DFULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdie.gad"},
	{0x5F7573AF0A9348D7ULL, "ht_store_187.anb"},
	{0x5F75CF23DE7D358BULL, "vxht_chantry_2is_0_0.vlm"},
	{0x5F75DB5B2D067EEDULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_app_p.mmh"},
	{0x5F75E772F5A5D79DULL, "mcr310_sibling6076250.cl"},
	{0x5F7676180E05C419ULL, "517387578.ogg"},
	{0x5F76BEA8A0C55750ULL, "hlac_dr_primevaltomb_665_1z.dds"},
	{0x5F77127D15B41D62ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy_flrsh.gad"},
	{0x5F77311F8C55FAACULL, "art\\vfx\\environment\\vfx_trap_glow_p.dura.ani"},
	{0x5F775DCA61CFE582ULL, "vxht_mansion2_3bz_0_0.vlm"},
	{0x5F77C8831E76D215ULL, "w_encounter3_19_0_lc.mmh"},
	{0x5F7818C1C0DB9196ULL, "mcr370_templar26087012.cl"},
	{0x5F7819E40C1CBC6CULL, "mag100_cullen6095170.cl"},
	{0x5F78449B3DEC8A11ULL, "1045058941.ogg"},
	{0x5F7849F8112EEA74ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.impa.gad"},
	{0x5F7875BE7C442940ULL, "118980315.ogg"},
	{0x5F787E23F8248EDBULL, "hlac_lt_0_lowtown_a2_n_c_-196609_1x.dds"},
	{0x5F78F812B1780B3BULL, "w_bonepit_12_0.mmh"},
	{0x5F790412B1781FD8ULL, "w_bonepit_12_0.mao"},
	{0x5F791612B1783E25ULL, "w_bonepit_12_0.msh"},
	{0x5F797017011C030BULL, "vxdr_0_deeproads_cjt_0_0.vlm"},
	{0x5F798A7678D5EAB1ULL, "vxw_bonepitmining_a2_9p_0_0.vlm"},
	{0x5F79D858243589D3ULL, "hlac_ht_chantry_n_2534_17y.dds"},
	{0x5F7A2E33AB17571EULL, "6081701_f.fxe"},
	{0x5F7A3F20E4F3172AULL, "vxdr_primevaltomb_77h_0_0.vlm"},
	{0x5F7A481D3FDBAFB5ULL, "6195817_m.fxe"},
	{0x5F7A6E74F100C529ULL, "6068498_m.fxe"},
	{0x5F7A997252644A53ULL, "566082439.ogg"},
	{0x5F7AB4B889067321ULL, "vxht_0_hightown_a3_c7c_0_0.vlm"},
	{0x5F7AF7AC3E2B5A65ULL, "6021211_m.fxe"},
	{0x5F7B800FDA9EF701ULL, "vxlt_0_lowtown_docks_a3_8ue_0_0.vlm"},
	{0x5F7B9BD0C40F95D7ULL, "262269947.ogg"},
	{0x5F7C2C2339D493CCULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdie.evt"},
	{0x5F7C665FFC54BBBEULL, "vxdr_primevalentrance_1ql_0_0.vlm"},
	{0x5F7C6A71CC55596DULL, "6033586_m.fxe"},
	{0x5F7C6D2C9FA9FE73ULL, "vxht_0_hightown_bfn_0_0.vlm"},
	{0x5F7C8290919F25E5ULL, "qun210cr_guard.utc"},
	{0x5F7C8C8835BE0229ULL, "vxlt_foundry_n_vg_0_0.vlm"},
	{0x5F7D357D15B8C459ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy_flrsh.evt"},
	{0x5F7D3C49939CC9E5ULL, "106523976.ogg"},
	{0x5F7DAC8C509416CFULL, "vfx_steamfast.wbd"},
	{0x5F7E3E1F8C5C2FEEULL, "art\\vfx\\environment\\vfx_trap_glow_p.dura.gad"},
	{0x5F7ECA51C88AB740ULL, "hlac_w_bonepit_393222_0y.dds"},
	{0x5F7EDDD25DDC9A84ULL, "al_anders_talents.gda"},
	{0x5F7EF3CCB415ED23ULL, "hlac_dr_primevaltomb_11985_3z.dds"},
	{0x5F7EF683CF965369ULL, "llac_w_sundermount_-327681_0.dds"},
	{0x5F7EF758436E712FULL, "647024206.ogg"},
	{0x5F7F58F811352252ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.impa.ani"},
	{0x5F7FE491E1D7B5A9ULL, "vxlt_gallowstemplar_f_5ro_0_0.vlm"},
	{0x5F800591EADA4780ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_down_to_up.ani"},
	{0x5F802BDF5E1D2C9CULL, "vxlt_warehouse_n_w7_0_0.vlm"},
	{0x5F806A1067F67952ULL, "788255948.ogg"},
	{0x5F80BCA201A6CB48ULL, "6177416_m.fxe"},
	{0x5F819664803E352BULL, "6163744_m.fxe"},
	{0x5F822E194F505729ULL, "389766136.ogg"},
	{0x5F82794D89498093ULL, "850842961.ogg"},
	{0x5F828E9091A3A5FFULL, "qun210cr_guard.stg"},
	{0x5F82DC8283CC52FCULL, "vxlt_warehouse_n_a5_0_0.vlm"},
	{0x5F831CE35B040324ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_cb.mmh"},
	{0x5F8320E35B040A77ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_cb.mao"},
	{0x5F8332E35B0428C6ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_cb.msh"},
	{0x5F833346BA852047ULL, "vxw_sundermount_341_0_0.vlm"},
	{0x5F835164D41B80E4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_treebark_n.dds"},
	{0x5F838C1C72D3246AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforward.gad"},
	{0x5F84BE9CCCF134FBULL, "hlac_lt_0_lowtown_a2_n_c_65535_5z.dds"},
	{0x5F84D0DF5C09E042ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgta_0t.dds"},
	{0x5F84E45DD6EE0A86ULL, "882424648.ogg"},
	{0x5F84F26B38599C10ULL, "329825751.ogg"},
	{0x5F84F29032C6EA7CULL, "dae000_fem_priest_mob6174190.cl"},
	{0x5F84F946A9252A70ULL, "207425581.ogg"},
	{0x5F85391F8C6245CDULL, "art\\vfx\\environment\\vfx_trap_glow_p.dura.evt"},
	{0x5F87409E1FBE9AF0ULL, "qun120_arishok6051483.cl"},
	{0x5F87B9EE0305029FULL, "833624823.ogg"},
	{0x5F87ED3C6E9D6B75ULL, "hf_nobelwoman_2.mor"},
	{0x5F88295BC57FEA42ULL, "vxht_bartrand_n_5i5_0_0.vlm"},
	{0x5F88328551525BFEULL, "vxlt_hangedman_n_16y_0_0.vlm"},
	{0x5F88B0FDD6E92FAAULL, "vxdr_primevaltomb_70i_0_0.vlm"},
	{0x5F88C21435091FFEULL, "vxdr_primevaltomb_4xo_0_1.vlm"},
	{0x5F88E532F5B1AB70ULL, "follower_banter6161281.cl"},
	{0x5F88E54C8FB6EB27ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratequest_sa.dds"},
	{0x5F89142339DEF525ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdie.ani"},
	{0x5F894A6864CD482AULL, "217785437.ogg"},
	{0x5F895E2B55C4C55EULL, "dae110ip_store_access.utp"},
	{0x5F897A85EB9B354FULL, "6206252_m.fxe"},
	{0x5F8A17EC7148B913ULL, "pro000_cut_innocent_blood6004078.cl"},
	{0x5F8A6F1C72D91101ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforward.evt"},
	{0x5F8A7E33DC55CB44ULL, "vxlt_blackemporium_v9_0_0.vlm"},
	{0x5F8AC45432C2BDDBULL, "art\\vfx\\cinematic\\prerendered\\vfx_fog_ground_p.dura.ani"},
	{0x5F8B3D7D15C50F48ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy_flrsh.ani"},
	{0x5F8B4AD2BBF70B80ULL, "6165099_m.fxe"},
	{0x5F8B80F7240B724EULL, "361032379.ogg"},
	{0x5F8BAEA4C547D4C4ULL, "41977103.ogg"},
	{0x5F8BF8A4A28FFF53ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.cess.ani"},
	{0x5F8C681C94205D33ULL, "6141850_m.fxe"},
	{0x5F8CFAE0025CAB0AULL, "276338373.ogg"},
	{0x5F8D04FDBEB8960CULL, "character_character_p113.psh.fxo"},
	{0x5F8D6C67CA6E8FEFULL, "6096475_m.fxe"},
	{0x5F8DC31657640B02ULL, "lt_0_lowtown_a2_n_32_0.rml"},
	{0x5F8E0622AFD8BF72ULL, "329858788.ogg"},
	{0x5F8E566ED3CDCE29ULL, "vxlt_gallowstemplar_f_5un_0_0.vlm"},
	{0x5F8E9DBC5013451CULL, "vxlt_warehouse_n_z4_0_0.vlm"},
	{0x5F8EB7DC85DA0653ULL, "112119118.ogg"},
	{0x5F8EFD4BB9B2DAB2ULL, "hlac_w_bonepitmining_a2_0_13x.dds"},
	{0x5F8F3AA2E3553E8FULL, "vxlt_gallowstemplar_f_419_0_0.vlm"},
	{0x5F8F68D698B52196ULL, "847672437.ogg"},
	{0x5F902BBDB5F215F0ULL, "709952187.ogg"},
	{0x5F904844CB71200FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45winup.mao"},
	{0x5F905444CB7134ACULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45winup.mmh"},
	{0x5F905A44CB713E9EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45winup.msh"},
	{0x5F91291763AC4245ULL, "6090344_f.fxe"},
	{0x5F918213786E1795ULL, "art\\vfx\\cinematic\\vfx_bld_punch_ds_p.anb"},
	{0x5F91A523AC7B38C7ULL, "vxw_sundermount_330_0_0.vlm"},
	{0x5F91C32FC6DCF2D6ULL, "vxlt_warehouse_n_yg_0_0.vlm"},
	{0x5F91D0B84D7FD0A0ULL, "mag100_cullen6086354.cl"},
	{0x5F9343B8EE59B512ULL, "vxca_caverns2prc_9rn_0_0.vlm"},
	{0x5F934B3E19B629FFULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.dura.gad"},
	{0x5F936BBE1855ED10ULL, "vxht_mansion2b_1nf_0_0.vlm"},
	{0x5F939B0B390B700FULL, "310590962.ogg"},
	{0x5F9467B1CDE062A1ULL, "t1_mub_pi2.tnt"},
	{0x5F9468F58E3AE684ULL, "vxlt_0_lowtown_docks_n_baz_0_0.vlm"},
	{0x5F9509D09DEC7ED7ULL, "6060328_m.fxe"},
	{0x5F9595DBE298CA5DULL, "art\\characters\\creatures\\shriek\\textures\\cn_bdy_ems_0d.dds"},
	{0x5F95A86E5AEC9B1FULL, "hlac_lt_hangedman_-1_22z.dds"},
	{0x5F96201BD11480F8ULL, "vxlt_0_lowtown_a2_heb_0_0.vlm"},
	{0x5F963A5E1307ACBEULL, "lowtown_male_urchin_sta6164174.cl"},
	{0x5F9670E4281049D6ULL, "vxlt_darktown_4uo_0_0.vlm"},
	{0x5F9680CCEC19CAE0ULL, "vxlt_0_lowtown_docks_cwj_0_0.vlm"},
	{0x5F97305EAC61C25EULL, "art\\characters\\creatures\\dragon\\animation\\tal_cst_s.evt"},
	{0x5F973169EC9366C3ULL, "1041349878.ogg"},
	{0x5F975133023AD22BULL, "vxht_mansion_n_6jm_0_0.vlm"},
	{0x5F97628ACD15552CULL, "6042561_m.fxe"},
	{0x5F97D055D815EC20ULL, "6062818_f.fxe"},
	{0x5F980A5019544114ULL, "676243028.ogg"},
	{0x5F983B04019FF01DULL, "vxlt_0_lowtown_a2_bxn_0_0.vlm"},
	{0x5F983B46A2875D20ULL, "23632390.ogg"},
	{0x5F98671C72E54150ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforward.ani"},
	{0x5F98CC5432CF08D2ULL, "art\\vfx\\cinematic\\prerendered\\vfx_fog_ground_p.dura.evt"},
	{0x5F98D5FDEB205CDEULL, "mcr370cr_gate_guardian.utc"},
	{0x5F98F0A4A29A7B8AULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.cess.evt"},
	{0x5F99403E19BA82ECULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.dura.evt"},
	{0x5F994A19E040ECBEULL, "vxdr_primevalentrance_1kn_0_0.vlm"},
	{0x5F9AF2197D602E6AULL, "6122122_m.fxe"},
	{0x5F9B1F4BB98D4753ULL, "one280_merin6116748.cl"},
	{0x5F9B3E6F43A0CAEAULL, "133269301.ogg"},
	{0x5F9B79DBFEA9DB2CULL, "ico_2h_axe_4.dds"},
	{0x5F9BBBE2191086A5ULL, "dae000cr_grand_cleric6147905.cl"},
	{0x5F9BC326AC7B7225ULL, "573344098.ogg"},
	{0x5F9BF1290F260696ULL, "zz_isa_debug6175463.cl"},
	{0x5F9BFD89B15CA41CULL, "lowtown_fem_prostitute_sta6161608.cl"},
	{0x5F9D79C950849FB0ULL, "gen00fl_merrill6128926.cl"},
	{0x5F9E43FE3B0DECF2ULL, "lowtown_male_alienage1_sta6162714.cl"},
	{0x5F9E475EAC680789ULL, "art\\characters\\creatures\\dragon\\animation\\tal_cst_s.gad"},
	{0x5F9EA0D8EA5C7C10ULL, "6077037_m.fxe"},
	{0x5F9EC3DACE5567B6ULL, "gen_im_trs_rdr_hvy_hlm_r2.uti"},
	{0x5F9FACB5B88DF017ULL, "art\\characters\\heraldry\\hld_orlais.dds"},
	{0x5F9FCFA4A2A062FDULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.cess.gad"},
	{0x5F9FD7D0D791A52AULL, "94066274.ogg"},
	{0x5F9FEB5432D55C65ULL, "art\\vfx\\cinematic\\prerendered\\vfx_fog_ground_p.dura.gad"},
	{0x5FA00A91695EEFA4ULL, "6189660_m.fxe"},
	{0x5FA01AD41EB10D4BULL, "vxlt_darktown_47p_0_0.vlm"},
	{0x5FA02B23081CBDAEULL, "vxdr_primevalentrance_27e_0_0.vlm"},
	{0x5FA036E5C181B738ULL, "vxca_cave_184_0_0.vlm"},
	{0x5FA04BD2BC4F6AD2ULL, "mer100pt_missing_miners.ncs"},
	{0x5FA063E70E3B8986ULL, "vxca_cave_18z_0_0.vlm"},
	{0x5FA13A1D4F468762ULL, "llac_ht_0_hightown_65536_0.dds"},
	{0x5FA16CB548659E5BULL, "mag370_meredith6046830.cl"},
	{0x5FA170AF0AB87EEBULL, "ht_store_187.rml"},
	{0x5FA1C73BA2BAEBFBULL, "6033795_m.fxe"},
	{0x5FA1DD9B0A4C0590ULL, "vxht_mansion2b_1ag_0_0.vlm"},
	{0x5FA220D75363C583ULL, "dae200_bran6094772.cl"},
	{0x5FA35D6C682368C5ULL, "mer120_hubert6071883.cl"},
	{0x5FA3BC7754DFA6EEULL, "vxca_cavern1_haunted_j3t_0_1.vlm"},
	{0x5FA3C5D0D2F082DFULL, "lt_gallowstemplar_n_2390.rml"},
	{0x5FA40F53FB5C056FULL, "vxca_caverns1_j2f_0_0.vlm"},
	{0x5FA491F8C30A9978ULL, "vxlt_0_lowtown_a2_hbc_0_0.vlm"},
	{0x5FA5185EAC6DD6D7ULL, "art\\characters\\creatures\\dragon\\animation\\tal_cst_s.ani"},
	{0x5FA52F0EF065E848ULL, "530008686.ogg"},
	{0x5FA582400567FD06ULL, "vxca_caverns2prc_62q_0_0.vlm"},
	{0x5FA584BAC1F35848ULL, "hlac_ht_0_hightown_n_-196608_0y.dds"},
	{0x5FA5C30FF430EAABULL, "vxht_mansion_n_6ml_0_0.vlm"},
	{0x5FA5D88E98A31D96ULL, "6171113_m.fxe"},
	{0x5FA5E413E8F649A9ULL, "gen00fl_anders6214961.cl"},
	{0x5FA6182330926F13ULL, "vxht_0_hightown_a3_2xa_0_0.vlm"},
	{0x5FA6A5214AA0005CULL, "850757356.ogg"},
	{0x5FA7283E19C69745ULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.dura.ani"},
	{0x5FA7C0FCC3163EDBULL, "character_character_p86.psh.fxo.dx11.assembly"},
	{0x5FA86122C7CE807AULL, "6072964_m.fxe"},
	{0x5FA86A7ADEC17540ULL, "vxlt_0_lowtown_docks_n_b4a_0_0.vlm"},
	{0x5FA873C7CA87BBA0ULL, "554501576.ogg"},
	{0x5FA968EE600E4F8BULL, "6019253_f.fxe"},
	{0x5FA995B951340712ULL, "6127244_m.fxe"},
	{0x5FAB3A28B7B9FF29ULL, "vxlt_gallowstemplar_f_5cl_0_0.vlm"},
	{0x5FAB53459384B8D4ULL, "zz_approval_debug6003966.cl"},
	{0x5FABBC64530E529EULL, "6126957_m.fxe"},
	{0x5FABCBB5538FF46FULL, "6136188_m.fxe"},
	{0x5FACC167F6162847ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_bridge8x_1.mao"},
	{0x5FAD12B5DC529490ULL, "6078137_m.fxe"},
	{0x5FAD845D4C65D8D1ULL, "6004333_m.fxe"},
	{0x5FAFB5653AFBC262ULL, "6015682_m.fxe"},
	{0x5FAFFB055FD7D994ULL, "6050171_m.fxe"},
	{0x5FB0910316F02737ULL, "vxca_caverns1_inu_0_0.vlm"},
	{0x5FB09976E56FB93BULL, "lla_ht_0_hightown_a3_131070_0.dds"},
	{0x5FB103B244C61618ULL, "ht_mansion2_probe_1ng.dds"},
	{0x5FB146A03DF28FABULL, "719616469.ogg"},
	{0x5FB179BABE18F6BBULL, "6066887_m.fxe"},
	{0x5FB192BB950574D6ULL, "dae000_meredith6166291.cl"},
	{0x5FB196B24A34AE6CULL, "gen00fl_merrill6116318.cl"},
	{0x5FB1C9C45EB80A9BULL, "104519142.ogg"},
	{0x5FB24DF283A76498ULL, "vxw_bonepit_57_0_0.vlm"},
	{0x5FB2BB584E21C5FBULL, "vxlt_0_lowtown_a3_n_c_hsc_0_0.vlm"},
	{0x5FB34A03C1DB1E72ULL, "hm_guard_1.mor"},
	{0x5FB34C39C2CC1741ULL, "267579445.ogg"},
	{0x5FB60023750C3316ULL, "26738705.ogg"},
	{0x5FB658F62D854EBEULL, "6042483_f.fxe"},
	{0x5FB6CD1F82836C8EULL, "251712255.ogg"},
	{0x5FB6D312EEA6C934ULL, "764084113.ogg"},
	{0x5FB6E5A81FE6F900ULL, "store_dae210_tailor.utm"},
	{0x5FB76F0AC327325FULL, "vxdr_0_deeproads_cuc_0_2.vlm"},
	{0x5FB799D61DCA661DULL, "weapon_vfx_p0.vsh.fxo"},
	{0x5FB79A4C11EBCD90ULL, "413794505.ogg"},
	{0x5FB7F53A2F8EE827ULL, "244084896.ogg"},
	{0x5FB875AD9659E0EBULL, "111984533.ogg"},
	{0x5FB8902CDBBE0C24ULL, "hlac_ca_caverns1_mines_24423_7z.dds"},
	{0x5FB94E36597C9F23ULL, "llac_lt_0_lowtown_docks_n_-131071_0.dds"},
	{0x5FB9B312A1731916ULL, "595099227.ogg"},
	{0x5FB9D595E4FCE77FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thin.msh"},
	{0x5FB9DB95E4FCF121ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thin.mmh"},
	{0x5FB9E795E4FD05C2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thin.mao"},
	{0x5FB9EBA36A16FE5DULL, "em_ser_jethann.mor"},
	{0x5FBA6154C304A0F4ULL, "vxht_0_hightown_c5y_0_0.vlm"},
	{0x5FBA831378911ED6ULL, "art\\vfx\\cinematic\\vfx_bld_punch_ds_p.mmh"},
	{0x5FBACF7A60F805A6ULL, "6125255_m.fxe"},
	{0x5FBB06B9DFF2082DULL, "138326291.ogg"},
	{0x5FBB98E4AB5A5C66ULL, "6109204_m.fxe"},
	{0x5FBBBFC2BC094CC8ULL, "6061399_m.fxe"},
	{0x5FBBCC94A22E7523ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_loop.ani"},
	{0x5FBBE7E9DB581A58ULL, "6101124_f.fxe"},
	{0x5FBBF026A1FE8595ULL, "6060281_m.fxe"},
	{0x5FBC34C742AD905FULL, "combat_deathblow_11.gda"},
	{0x5FBCA373B5AECCCEULL, "vxlt_gallowstemplar_f_46u_0_0.vlm"},
	{0x5FBCBC0720B086D5ULL, "402699087.ogg"},
	{0x5FBCE1A1DCD00995ULL, "art\\characters\\creatures\\spider\\textures\\c_corspidra_0s.dds"},
	{0x5FBE7D4E267A31CAULL, "296630400.ogg"},
	{0x5FBEAD54675A5F57ULL, "291796841.ogg"},
	{0x5FBEAF16796129EEULL, "58453906.ogg"},
	{0x5FBEB2CEC21F7BEFULL, "548950509.ogg"},
	{0x5FBEC68C50CBD9A5ULL, "vfx_steamfast.bnk"},
	{0x5FBF657675A47700ULL, "mag101_grace6035851.cl"},
	{0x5FBF82A74ED28A07ULL, "and121_karl6093887.cl"},
	{0x5FBF86D82BEBB475ULL, "hlac_ht_mansion2b_818_14x.dds"},
	{0x5FBF9750745BA527ULL, "dae021_keep_guard_4.utc"},
	{0x5FBFD4EE4697373CULL, "hlac_w_encounter3_-393212_0x.dds"},
	{0x5FC0129CECB85523ULL, "6175426_m.fxe"},
	{0x5FC03D15A1855184ULL, "hlac_lt_0_lowtown_a2_n_131070_6x.dds"},
	{0x5FC071988C993DBAULL, "6165291_m.fxe"},
	{0x5FC1ED000130334DULL, "ser270_cut_aftermath6162304.cl"},
	{0x5FC1F78418E68728ULL, "w_woundedcoast_17_0.anb"},
	{0x5FC2146413281618ULL, "397887309.ogg"},
	{0x5FC2502368F61F49ULL, "vxlt_hangedman_280_0_0.vlm"},
	{0x5FC27F4E9CFA4536ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzchair_0.pwk"},
	{0x5FC2A024A197872BULL, "hlac_w_woundedcoast_327680_0z.dds"},
	{0x5FC2A722AB959209ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf04.phy"},
	{0x5FC2FD182F7C44E7ULL, "ch1_s2.psh.fxo"},
	{0x5FC3CA685F473D4AULL, "998362373.ogg"},
	{0x5FC3F995A746FCC2ULL, "llac_ca_caverns1_24426_12.dds"},
	{0x5FC434EB1A3E62B9ULL, "6095376_f.fxe"},
	{0x5FC4629F089BB263ULL, "vxlt_darktown_517_0_0.vlm"},
	{0x5FC532F356CEB94EULL, "llac_lt_gallowstemplar_n_2383_7.dds"},
	{0x5FC56A36B5B2C3B6ULL, "558119416.ogg"},
	{0x5FC591DB057E6C78ULL, "hlac_ca_caverns2prc_5802_62z.dds"},
	{0x5FC634136778D254ULL, "vxdr_dungeon_1lw_0_0.vlm"},
	{0x5FC65522F612BF1CULL, "dae000_flemeth6013293.cl"},
	{0x5FC6D5521A19D2C6ULL, "vxdr_primevaltomb_7qr_0_0.vlm"},
	{0x5FC6F812B1BA2CC8ULL, "w_bonepit_12_0.rml"},
	{0x5FC6F8E73724292AULL, "473141417.ogg"},
	{0x5FC7C01657957EFEULL, "lt_0_lowtown_a2_n_32_0.anb"},
	{0x5FC7D9F5322596F4ULL, "6173117_m.fxe"},
	{0x5FC7DEB6DA78AEF2ULL, "6070984_m.fxe"},
	{0x5FC81DE29BA63029ULL, "vxlt_gallowstemplar_f_5ir_0_0.vlm"},
	{0x5FC82D2DA4E2B2D7ULL, "1005426887.ogg"},
	{0x5FC83767AD137F59ULL, "zz_gtest.ncs"},
	{0x5FC8B0D3C2B74012ULL, "vxlt_gallowscourtyard_7f7_0_0.vlm"},
	{0x5FC9F3057C552113ULL, "6005776_m.fxe"},
	{0x5FCB5ED019DD9297ULL, "233815870.ogg"},
	{0x5FCC1734A792EF88ULL, "qun220cr_javaris6068907.cl"},
	{0x5FCC363573B8C8B1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoorcobble_n.dds"},
	{0x5FCC49F1265D0215ULL, "vxca_caverns1_mines_cjx_0_0.vlm"},
	{0x5FCCC1B4E57F32D2ULL, "736769925.ogg"},
	{0x5FCCC5707A6E6D9DULL, "sto_dust.mao"},
	{0x5FCD214B52CA0380ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_cliffwall_d.dds"},
	{0x5FCDA6E60DF9EDB5ULL, "vrc221_hugin6065253.cl"},
	{0x5FCE1DA590A911C3ULL, "strmt02_d.dds"},
	{0x5FCE1DABA6C413E1ULL, "zz_mer_debug6021016.cl"},
	{0x5FCE293F967AF490ULL, "6161547_m.fxe"},
	{0x5FCEFB24A522E1F0ULL, "vxca_caverns1_mines_cn5_0_0.vlm"},
	{0x5FCF536E3FF410D2ULL, "142258507.ogg"},
	{0x5FCF8B4DD632DBE4ULL, "w_encounter3_243_0.mmh"},
	{0x5FCF8F4DD632E237ULL, "w_encounter3_243_0.mao"},
	{0x5FCFA14DD6330086ULL, "w_encounter3_243_0.msh"},
	{0x5FD02372756521BBULL, "vxdr_0_deeproads_f4s_0_0.vlm"},
	{0x5FD04361BEFB329AULL, "6161127_m.fxe"},
	{0x5FD0997D6C81E1E6ULL, "vxlt_gallowscourtyard_n_4uv_0_0.vlm"},
	{0x5FD09EE4190B97E7ULL, "6135300_f.fxe"},
	{0x5FD0C2005AEC37C9ULL, "vxlt_hangedman_257_0_0.vlm"},
	{0x5FD0E394A240F80DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_loop.gad"},
	{0x5FD107D275B1E5E9ULL, "vxht_mansion2b_1he_0_0.vlm"},
	{0x5FD14688CE759759ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_rack_n.dds"},
	{0x5FD26D706B073875ULL, "6139820_m.fxe"},
	{0x5FD28749C3A59CABULL, "vxlt_undercitydungeon_187_0_0.vlm"},
	{0x5FD2C0755D58E8CCULL, "6155481_m.fxe"},
	{0x5FD2E990253BEA00ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga101a_0s.dds"},
	{0x5FD306172967304DULL, "6070172_m.fxe"},
	{0x5FD37F175650D072ULL, "hlac_ht_0_hightown_a3_-196608_2y.dds"},
	{0x5FD3C3A29D28C3B1ULL, "605837930.ogg"},
	{0x5FD3E085089F563FULL, "ht_0_hightown_a2_21_0_ll.mmh"},
	{0x5FD3F1FF5170A3C5ULL, "vxlt_undercityserialkiller_188_0_0.vlm"},
	{0x5FD40931E259F24BULL, "vxdr_0_deeproads_gft_0_0.vlm"},
	{0x5FD49D9B3D1458A8ULL, "6051208_m.fxe"},
	{0x5FD4D2F671F9BF37ULL, "vxlt_gallowscourtyard_65b_0_0.vlm"},
	{0x5FD50E59A01D75C4ULL, "vxlt_undercityserialkiller_1kt_0_0.vlm"},
	{0x5FD55602895A6C25ULL, "gen00fl_isabela_init6110286.cl"},
	{0x5FD67EE1215CF655ULL, "gen00fl_aveline6139650.cl"},
	{0x5FD73747FE1ECCC2ULL, "dr_0_deeproads_5838.rml"},
	{0x5FD740C006AF0098ULL, "llac_w_blightlands_262142_0.dds"},
	{0x5FD7BB73F5E416E0ULL, "dae000ip_crowd_lm4_talk8.utp"},
	{0x5FD83049801ABBA6ULL, "dae000cr_grand_cleric6147661.cl"},
	{0x5FD848EDF43BE655ULL, "crt_abomination.bnk"},
	{0x5FD8BF8ABFF73E3BULL, "61694038.ogg"},
	{0x5FD8DCAB83730F0BULL, "6127493_f.fxe"},
	{0x5FD94AF8165DA75AULL, "vxht_keep_708_0_0.vlm"},
	{0x5FD9CE869FC7ACE1ULL, "6139311_f.fxe"},
	{0x5FD9E3E0576A1190ULL, "6153560_m.fxe"},
	{0x5FD9F0C05ECAD519ULL, "6072686_m.fxe"},
	{0x5FD9F20E96DA0552ULL, "6161022_m.fxe"},
	{0x5FDA5394D5CB9CE6ULL, "hlac_lt_darktown_-1_31z.dds"},
	{0x5FDA7F6C7D30126BULL, "vxlt_gallowstemplar_46z_0_1.vlm"},
	{0x5FDB2198B7FD2543ULL, "water_water_p4.psh.fxo.dx11.assembly"},
	{0x5FDB48955944A279ULL, "ts_coterie_prog_junk3.gda"},
	{0x5FDBA50ED2246790ULL, "vxht_mansion2b_1uc_0_0.vlm"},
	{0x5FDBED18597A134EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_ex.evt"},
	{0x5FDC4730F4979B32ULL, "6042211_m.fxe"},
	{0x5FDCB3B70ACC8D64ULL, "gen00fl_varric6105431.cl"},
	{0x5FDCDA45BB725DB5ULL, "w_bonepit_67_0.msh"},
	{0x5FDCDC45BB72612BULL, "w_bonepit_67_0.mmh"},
	{0x5FDCE845BB7275C8ULL, "w_bonepit_67_0.mao"},
	{0x5FDD919D77E0FAD4ULL, "6017664_m.fxe"},
	{0x5FDD926A52B62D80ULL, "zz_daept_debug.ncs"},
	{0x5FDDE15C787633E2ULL, "llac_lt_0_lowtown_docks_196606_0.dds"},
	{0x5FDE21F99D43C19AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_mushroom.msh"},
	{0x5FDE33F99D43E0E0ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_mushroom.mmh"},
	{0x5FDE8F870786E8DCULL, "442307327.ogg"},
	{0x5FDF33DD4CE25049ULL, "vxlt_hangedman_226_0_0.vlm"},
	{0x5FDF601C557852F3ULL, "ht_0_hightownexplosion_n_21_0_ll.mmh"},
	{0x5FDF79AF67A7FE69ULL, "vxht_mansion2b_1cd_0_0.vlm"},
	{0x5FDF9543E27B7800ULL, "960265732.ogg"},
	{0x5FE0426573B60D71ULL, "vxht_mansion2_1rg_0_0.vlm"},
	{0x5FE04AAEF80BD4F5ULL, "215697300.ogg"},
	{0x5FE04B2BA69AC427ULL, "vxht_0_hightown_n_c3f_0_1.vlm"},
	{0x5FE0AD6932A3277FULL, "101574656.ogg"},
	{0x5FE153D625008D2DULL, "gen00fl_aveline6101770.cl"},
	{0x5FE177F41B5D48CDULL, "6075272_m.fxe"},
	{0x5FE1F6B21A3F77B6ULL, "vxlt_hangedman_n_4l_0_0.vlm"},
	{0x5FE2AB6B10BA7D09ULL, "cs_pro000_db_ogre_rog_war.wbd"},
	{0x5FE2C086AFC82489ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masb_0t.dds"},
	{0x5FE2D26783406A0BULL, "352696808.ogg"},
	{0x5FE2D5F35CC4BE86ULL, "1025984942.ogg"},
	{0x5FE2E418598022F9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_ex.gad"},
	{0x5FE3C5D72F36D5F9ULL, "gen00fl_anders6119215.cl"},
	{0x5FE3D6C01675BA28ULL, "vxht_mansion2_2mp_0_0.vlm"},
	{0x5FE3D7C6E99AA56CULL, "hlac_lt_0_lowtown_n_-1_3y.dds"},
	{0x5FE414FBD7D065E5ULL, "6041028_m.fxe"},
	{0x5FE62F57455F1AEFULL, "518623398.ogg"},
	{0x5FE6450D9710D5B1ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_4y.dds"},
	{0x5FE660C7FD15008AULL, "hlac_w_bonepit_196613_0x.dds"},
	{0x5FE673E2341FB30AULL, "gen00fl_varric6076124.cl"},
	{0x5FE6B71827FB58C6ULL, "zz_fen_debug6134350.cl"},
	{0x5FE6DC95E523E09EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thin.phy"},
	{0x5FE74B0C68C0740DULL, "6149504_m.fxe"},
	{0x5FE765C95D09696AULL, "vxht_store_bu_0_0.vlm"},
	{0x5FE7C760D429CF75ULL, "vxlt_hangedman_5q_0_0.vlm"},
	{0x5FE7DB10AE0B7185ULL, "clf221_jalen6070395.cl"},
	{0x5FE855BD49602A10ULL, "6156660_m.fxe"},
	{0x5FE8873D245B0414ULL, "gen00fl_aveline6120817.cl"},
	{0x5FE8C7DC67CDAD81ULL, "dae000_gamlen6149453.cl"},
	{0x5FE8F5185984AB47ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_whirlwind_ex.ani"},
	{0x5FE939190AE4AC06ULL, "dae000_junar6147847.cl"},
	{0x5FE9C65B901DAC8EULL, "art\\vfx\\combat\\vfx_c_single_2nd_stun_c.impa.ani"},
	{0x5FE9DF9D3136A60CULL, "gen00fl_merrill6128874.cl"},
	{0x5FE9EC6743A12482ULL, "6133137_m.fxe"},
	{0x5FEA036766098202ULL, "plt_prm000pt_parth_staff_.plo"},
	{0x5FEA3883CAA7BD5AULL, "6116859_m.fxe"},
	{0x5FEA7FB41696D0CFULL, "vxht_0_hightown_2es_0_0.vlm"},
	{0x5FEAA573C4ED6706ULL, "dae210_gate_guard6084313.cl"},
	{0x5FEAAD43698384BDULL, "vxca_caverns1_mines_gjl_0_0.vlm"},
	{0x5FEAF284190983B0ULL, "w_woundedcoast_17_0.mao"},
	{0x5FEAF68419098AC3ULL, "w_woundedcoast_17_0.mmh"},
	{0x5FEAF7259003F932ULL, "hlac_lt_0_lowtown_a2_n_-196608_2y.dds"},
	{0x5FEB04841909A21DULL, "w_woundedcoast_17_0.msh"},
	{0x5FEB58D1223CB80DULL, "339369925.ogg"},
	{0x5FEC04ABCC9135E5ULL, "hlac_lt_0_lowtown_a2_n_c_131071_4z.dds"},
	{0x5FEC350E4263169CULL, "hlac_dr_primevaltomb_664_6x.dds"},
	{0x5FEC44D3A0B07B2FULL, "918085929.ogg"},
	{0x5FEC82CC15FA27FBULL, "vxlt_0_lowtown_a3_n_c_hog_0_0.vlm"},
	{0x5FEC8EC8A5FD8EB7ULL, "gen00fl_merrill6107058.cl"},
	{0x5FECB588A71425D0ULL, "follower_banter6155772.cl"},
	{0x5FECD3F0FA2EA302ULL, "vxw_encounter3_ez_0_0.vlm"},
	{0x5FED3B2D3956C1C8ULL, "hlac_w_encounter3_0_5z.dds"},
	{0x5FED4BED7522D09CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1l.mao"},
	{0x5FED57ED7522E53FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1l.mmh"},
	{0x5FED5DED7522EFE1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1l.msh"},
	{0x5FEDB62CCC47CDEDULL, "6181749_f.fxe"},
	{0x5FEDD9E7D5B143ADULL, "vxht_0_hightown_a3_9zr_0_0.vlm"},
	{0x5FEDEB8C599E16E9ULL, "vxht_mansion2b_1fk_0_0.vlm"},
	{0x5FEE09ECF499E637ULL, "126207715.ogg"},
	{0x5FEE22ED77B314C3ULL, "6013299_f.fxe"},
	{0x5FEE9022ABBAA62EULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf04.mmh"},
	{0x5FEEAA22ABBAD21CULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf04.msh"},
	{0x5FEEB44265AC25F1ULL, "vxht_mansion2_1uf_0_0.vlm"},
	{0x5FEF266FEAD429AFULL, "dae000_fenarel6133378.cl"},
	{0x5FEF82886348666BULL, "vxht_0_hightown_a3_atf_0_0.vlm"},
	{0x5FEF99D63D01F53EULL, "terrain_blend2.psh.fxo"},
	{0x5FEFD16F288E29C3ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv4.evt"},
	{0x5FF074A9B3765680ULL, "fen110_party_comment6095244.cl"},
	{0x5FF091809D777557ULL, "gen00fl_sebastian6179922.cl"},
	{0x5FF0D8AA29C5905EULL, "237015761.ogg"},
	{0x5FF0D905C1399E96ULL, "91248156.ogg"},
	{0x5FF0E6635B3EE087ULL, "mag321_comtesse_de_launcet6043327.cl"},
	{0x5FF0F8D4408A6884ULL, "318083247.ogg"},
	{0x5FF140F98623BFF3ULL, "hlac_lt_0_lowtown_docks_a3_-131076_1x.dds"},
	{0x5FF15A2E1BEC6244ULL, "llac_w_encounter1_-65533_0.dds"},
	{0x5FF18E34D38EEBFEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_side_down.ani"},
	{0x5FF2398FDBE24B74ULL, "dae310cr_elf_m_01.utc"},
	{0x5FF25EAD17E24C89ULL, "dae200_cullen6140512.cl"},
	{0x5FF2969A52F6412FULL, "fx_tent_64f.mao"},
	{0x5FF2E8FC543B0D68ULL, "zz_dog_debug6127764.cl"},
	{0x5FF3161F03416530ULL, "683728428.ogg"},
	{0x5FF33A451ABAD9B8ULL, "mrl150_merrill6071847.cl"},
	{0x5FF3F04AFDBB1B02ULL, "vxht_0_hightown_n_bsw_0_0.vlm"},
	{0x5FF44200F41C881EULL, "vxlt_0_lowtown_a2_fgt_0_0.vlm"},
	{0x5FF46FD503AC4760ULL, "57105266.ogg"},
	{0x5FF50207C48B4C07ULL, "633840955.ogg"},
	{0x5FF57178C3D61AC7ULL, "hlac_w_woundedcoast_-65533_0y.dds"},
	{0x5FF595C2780F337CULL, "6171573_m.fxe"},
	{0x5FF5CD1BACA2D07DULL, "vxw_bonepit_1p6_0_0.vlm"},
	{0x5FF5CEE6C14E8F8CULL, "6082028_m.fxe"},
	{0x5FF62FC2FA29BB7FULL, "17777217.ogg"},
	{0x5FF63E1DBCA5BA18ULL, "hlac_lt_gallowstemplar_n_2383_7x.dds"},
	{0x5FF6AA6F28940644ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv4.gad"},
	{0x5FF6B96C68DB3F0FULL, "108865999.ogg"},
	{0x5FF6BA18C202C28BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_exit.ani"},
	{0x5FF6BE5B902829DFULL, "art\\vfx\\combat\\vfx_c_single_2nd_stun_c.impa.evt"},
	{0x5FF6CD09875FB05CULL, "dae000tr_assassin.utt"},
	{0x5FF6EF54E5F3EA87ULL, "zz_isa_debug6133181.cl"},
	{0x5FF6F03EC34BD56EULL, "406544685.ogg"},
	{0x5FF7069F0AB8E8F2ULL, "vxca_caverns1_ijp_0_0.vlm"},
	{0x5FF741F6AA6A4394ULL, "follower_banter6170328.cl"},
	{0x5FF76CC2EE72FE37ULL, "429045252.ogg"},
	{0x5FF77B6DB5C490C5ULL, "829159559.ogg"},
	{0x5FF7846D8F031662ULL, "vxdr_0_deeproads_f64_0_0.vlm"},
	{0x5FF7A35C02B0384CULL, "qun231cr_viscount_guard2.utc"},
	{0x5FF7FDF8EE002BA2ULL, "554827469.ogg"},
	{0x5FF821A0EF11DB34ULL, "6119221_m.fxe"},
	{0x5FF823E43D0DA4A5ULL, "vxht_bartrand_n_65a_0_0.vlm"},
	{0x5FF86EDDA4FE02DBULL, "one140cr_charade6134307.cl"},
	{0x5FF8ACF8E4D5E001ULL, "one100dg_vanard6017488.cl"},
	{0x5FF991FFBB06F46FULL, "6125687_m.fxe"},
	{0x5FF9BD38AA26EAB5ULL, "vxlt_0_lowtown_a2_n_c_ht9_0_0.vlm"},
	{0x5FF9DEBA7D74582FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_walls_sa.dds"},
	{0x5FFA4468C356B6A4ULL, "gen00fl_isabela6079062.cl"},
	{0x5FFA747EDAF7FDF6ULL, "mer_glow.mao"},
	{0x5FFAC4B48F3AD924ULL, "368589993.ogg"},
	{0x5FFAD3B6D6D2737EULL, "gen00fl_anders6125921.cl"},
	{0x5FFAE42BD288C29DULL, "mer110pt_back_to_work.ncs"},
	{0x5FFB4044291EA0C7ULL, "568486811.ogg"},
	{0x5FFB87BABAD98148ULL, "746799363.ogg"},
	{0x5FFBD11651BF7FE7ULL, "dae000_orana6149030.cl"},
	{0x5FFBD341BB59A921ULL, "art\\characters\\creatures\\hurlock\\animation\\blendtree\\hrlkoverlay.anb"},
	{0x5FFBEEEDCCE6066CULL, "705762726.ogg"},
	{0x5FFC5D694B942F69ULL, "vxht_mansion2b_1yj_0_0.vlm"},
	{0x5FFCB96F28988B22ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atkv4.ani"},
	{0x5FFD20BAB2CEF482ULL, "vxht_keep_6an_0_0.vlm"},
	{0x5FFD261F57A23E71ULL, "vxht_mansion2_1xe_0_0.vlm"},
	{0x5FFD70D971AB17B0ULL, "6087812_m.fxe"},
	{0x5FFD9F5B902E13C8ULL, "art\\vfx\\combat\\vfx_c_single_2nd_stun_c.impa.gad"},
	{0x5FFE4AF56CDD395DULL, "996622645.ogg"},
	{0x5FFE4BB906BBE90AULL, "hlac_dr_dungeon_3215_1x.dds"},
	{0x5FFEBB172AF1368DULL, "738089329.ogg"},
	{0x5FFEEDB800F23546ULL, "399990104.ogg"},
	{0x5FFF4DB0D22B6425ULL, "vxht_mansion2b_2n8_0_0.vlm"},
	{0x5FFF54F77663836CULL, "dae000_fenarel6127240.cl"},
	{0x5FFF7B6B2D800D07ULL, "hlac_lt_0_lowtown_a2_n_-65537_2z.dds"},
	{0x5FFFA634D39B51CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_side_down.evt"},
	{0x6000FD3291DAA09BULL, "6018820_m.fxe"},
	{0x600106A757102C10ULL, "6087030_m.fxe"},
	{0x60011F0B754F049EULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bed_1.mao"},
	{0x6001A2FC2CB647EFULL, "6041112_m.fxe"},
	{0x6001E76196DD7F24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_reaction.ani"},
	{0x60022AE365D87E73ULL, "6120113_f.fxe"},
	{0x600261EE275AC0C9ULL, "6126969_m.fxe"},
	{0x6002AC6BC19FC424ULL, "6131963_f.fxe"},
	{0x6002B8350CD46A72ULL, "vxht_mansion2_313_0_0.vlm"},
	{0x6002E20FEA4E6DC2ULL, "vxht_mansion2b_1tv_0_0.vlm"},
	{0x6002F684191E1CE0ULL, "w_woundedcoast_17_0.rml"},
	{0x600343ED0C907E2CULL, "vxw_bonepitmining_a2_1p1_0_0.vlm"},
	{0x60036B7F77D0C379ULL, "vxlt_0_lowtown_a2_cmt_0_0.vlm"},
	{0x600375869F49E6DFULL, "256558213.ogg"},
	{0x6003B20513461559ULL, "mag210_arianni6024556.cl"},
	{0x6003E79F814972A9ULL, "142353678.ogg"},
	{0x6003EA26E38C7A9DULL, "vxlt_0_lowtown_a3_n_c_emz_0_0.vlm"},
	{0x60043EF89E98E8FDULL, "vxw_bonepit_1u5_0_0.vlm"},
	{0x600440C3B344A80CULL, "6085128_m.fxe"},
	{0x60046EEDF46161DFULL, "crt_abomination.wbd"},
	{0x6004739DB5729183ULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.cess.ani"},
	{0x6004A218C20ED702ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_exit.evt"},
	{0x6004B930585C02FFULL, "vxht_mansion2_3dm_0_0.vlm"},
	{0x6005787BFCAF0172ULL, "vxca_caverns1_imq_0_0.vlm"},
	{0x600578FF25367338ULL, "vxca_caverns2prc_aey_0_0.vlm"},
	{0x6005A2E9B00C34C5ULL, "298106719.ogg"},
	{0x6005A60071F04586ULL, "vxca_caverns2prc_ae7_0_0.vlm"},
	{0x6006797A1478EAD8ULL, "6047232_m.fxe"},
	{0x6006A3CC24836C40ULL, "gen00fl_fenris6126894.cl"},
	{0x6006A734D3A17278ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_side_down.gad"},
	{0x600708D46B1B0F85ULL, "str_jaggedrock01_d.dds"},
	{0x600733125B6DCCD3ULL, "lt_0_lowtown_a2_21_0.anb"},
	{0x60074DBBE3152D86ULL, "zz_one_debug6113367.cl"},
	{0x60077D6B10D94BEBULL, "cs_pro000_db_ogre_rog_war.bnk"},
	{0x6007D7F223AB9C84ULL, "zz_approval_debug6133719.cl"},
	{0x6007E46196E1E586ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_reaction.gad"},
	{0x6008199FBBF09CFEULL, "6175140_m.fxe"},
	{0x60088A5B17DEFF0FULL, "hlac_ht_0_hightown_n_65537_0y.dds"},
	{0x6008909331436A82ULL, "662200190.ogg"},
	{0x6008BC5AB85ED6D2ULL, "dae000ip_crowd_mf5_talk10.utp"},
	{0x6009AA6636080E6AULL, "w_bonepitdestroyed_a3_171_0.mao"},
	{0x6009AC97E8B2D07AULL, "art\\characters\\playercharacter\\children\\textures\\km_cth_boya_0s.dds"},
	{0x6009AE66360814B9ULL, "w_bonepitdestroyed_a3_171_0.mmh"},
	{0x6009B866360825A7ULL, "w_bonepitdestroyed_a3_171_0.msh"},
	{0x600B3F9DB2A4B473ULL, "vxdr_primevaltomb_34u_0_1.vlm"},
	{0x600B4973C587B9A3ULL, "6174013_f.fxe"},
	{0x600B49FF1326A102ULL, "1044817263.ogg"},
	{0x600B97FC499856F1ULL, "vxht_mansion2_1cd_0_0.vlm"},
	{0x600BA118C214F415ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_exit.gad"},
	{0x600C44E1A5F6CE1AULL, "gen_im_trs_rdr_hvy_cht_r0.uti"},
	{0x600C96A198010692ULL, "vxw_bonepitdestroyed_a3_1w4_0_0.vlm"},
	{0x600CE3EFB69FBF2AULL, "vxlt_0_lowtown_docks_n_b9m_0_1.vlm"},
	{0x600D03C23105BC3DULL, "ser221ip_blood_vials.utp"},
	{0x600D38CFAAF39F2DULL, "383277940.ogg"},
	{0x600D6DAD853FCB6BULL, "ss_merrillcu2.cut"},
	{0x600DA7358A25FFFAULL, "docks_male_human_worker_sta6176068.cl"},
	{0x600E03DF0C523145ULL, "6151170_m.fxe"},
	{0x600E306DF59B6101ULL, "hlac_ht_chantry_n_2534_20y.dds"},
	{0x600E5AF1F50DD54EULL, "6044188_f.fxe"},
	{0x600E9FA3FD23F5E9ULL, "6051397_m.fxe"},
	{0x600EDF6196E7FC65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_reaction.evt"},
	{0x600EFC074D2E7A4BULL, "vxlt_warehouse_n_cp_0_0.vlm"},
	{0x600FD968D1B9B0C9ULL, "6145144_m.fxe"},
	{0x600FE908E085FD05ULL, "vxlt_backalley_n_1x4_0_1.vlm"},
	{0x6010D404E1A74C02ULL, "vxht_0_hightown_n_byq_0_0.vlm"},
	{0x60112979D216256CULL, "6063952_f.fxe"},
	{0x60113181C7427EDDULL, "gen00fl_carver6180608.cl"},
	{0x60113447FE503FBEULL, "dr_0_deeproads_5838.anb"},
	{0x60118B9DB57D457AULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.cess.evt"},
	{0x6011AE792CA7825CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightr_1.mao"},
	{0x6011D61171F5255BULL, "vxlt_0_lowtown_a2_n_c_hx3_0_0.vlm"},
	{0x6011DEC85BBE56D6ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha06a.msh"},
	{0x6011E8C85BBE67D4ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha06a.mmh"},
	{0x6011EA08A6DD43CCULL, "299133445.ogg"},
	{0x6012C1ABBCB7213EULL, "752734855.ogg"},
	{0x60131C7D455CD9DEULL, "vxdr_primevaltomb_6z7_0_1.vlm"},
	{0x60142DF7C3801B95ULL, "6097860_f.fxe"},
	{0x60159489A902F0AFULL, "dae000_fenarel6128118.cl"},
	{0x6015A3B51417DAD8ULL, "gen00fl_fenris6127048.cl"},
	{0x601658F51BA32991ULL, "gen00fl_isabela6096831.cl"},
	{0x6016654522966C72ULL, "mrl150_hunter6071291.cl"},
	{0x601688791B3D25DFULL, "vfx0_dba.psh.fxo"},
	{0x6016D1DF93E013A6ULL, "rdr111dg_harbormaster6015973.cl"},
	{0x6016F4D36F03B483ULL, "hlac_lt_warehouse_65536_3z.dds"},
	{0x6017391922333324ULL, "vxw_bonepitmining_a2_2aa_0_0.vlm"},
	{0x60176544D6ACFC5DULL, "526468213.ogg"},
	{0x6018086F6FDDE817ULL, "6119600_m.fxe"},
	{0x60181260ABEE9F16ULL, "sound_bodyfall_foley.gda"},
	{0x60182961B793527EULL, "hlac_lt_0_lowtown_a2_131073_0z.dds"},
	{0x60182E80CF651BC6ULL, "67562.ogg"},
	{0x60183613FCE235B3ULL, "llac_lt_0_lowtown_a2_n_65535_9.dds"},
	{0x601837FF731BF761ULL, "follower_banter6160627.cl"},
	{0x60184FE61E44544BULL, "471874596.ogg"},
	{0x601856EB0366BAFBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_plain.msh"},
	{0x601868EB0366D946ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_plain.mao"},
	{0x601874EB0366EDE5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_plain.mmh"},
	{0x60188A9DB583626DULL, "art\\vfx\\cinematic\\vfx_lyrium_idol_c.cess.gad"},
	{0x6018C250789E5FB0ULL, "6190972_m.fxe"},
	{0x6019033EA427B5E8ULL, "6088923_m.fxe"},
	{0x601941232F806069ULL, "vxht_mansion2b_1wh_0_0.vlm"},
	{0x601972E8FA7298D1ULL, "666066704.ogg"},
	{0x60199EE05E17594DULL, "6166969_m.fxe"},
	{0x601A27867AD0AFDEULL, "6039301_m.fxe"},
	{0x601A7DE4C93B5B99ULL, "vxw_encounter3_fu_0_1.vlm"},
	{0x601AA89C1DDA5FC4ULL, "llac_lt_0_lowtown_a2_n_c_1_5.dds"},
	{0x601ABD26B376F948ULL, "llac_ca_cavern1_haunted_24427_1.dds"},
	{0x601B087E89F71F12ULL, "vxw_bonepitdestroyed_a3_1x5_0_0.vlm"},
	{0x601B0BD517166B53ULL, "vxht_keep_42h_0_0.vlm"},
	{0x601B55CCA895D7AAULL, "vxlt_0_lowtown_docks_n_b6l_0_1.vlm"},
	{0x601B562F3A5FC737ULL, "art\\vfx\\textures\\vfx_drain_fb.dds"},
	{0x601B77EFD3B4AF2AULL, "dae000_arishok6067928.cl"},
	{0x601BA5F65BA1F238ULL, "vxlt_0_lowtown_a2_n_c_hic_0_0.vlm"},
	{0x601CDDD7A88FD8E9ULL, "ts_profane_acc_table.gda"},
	{0x601CFA5DB224BBF7ULL, "1043323223.ogg"},
	{0x601D10201220FD6EULL, "661578304.ogg"},
	{0x601D138AD01334EEULL, "plt_cod_mgc_apostates.plo"},
	{0x601D47C5A49A52C4ULL, "167458878.ogg"},
	{0x601DE2525E1664F9ULL, "llac_lt_0_lowtown_docks_-131075_5.dds"},
	{0x601ECC34F9320243ULL, "685980528.ogg"},
	{0x6020469457FF5A1BULL, "vxlt_0_lowtown_n_ext_0_0.vlm"},
	{0x6020F7D89C278869ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlb.phy"},
	{0x602110CBF49AB9E4ULL, "556518169.ogg"},
	{0x602131004B786853ULL, "lowtown_male_beggar_sta6151039.cl"},
	{0x60222E9F23983D14ULL, "vxlt_0_lowtown_h2c_0_0.vlm"},
	{0x6022A756020C74B6ULL, "6072843_m.fxe"},
	{0x6022DFACD37F3AFDULL, "art\\characters\\playercharacter\\humanmale\\hlh_wic.phy"},
	{0x6022EE95A9A0BF65ULL, "vxca_caverns2_986_0_0.vlm"},
	{0x60232CFE6C17B32FULL, "vxht_keep_6c5_0_0.vlm"},
	{0x602346E2BF744DB7ULL, "vxw_encounter3_97_0_0.vlm"},
	{0x6023920BA99D534EULL, "lt_warehouse_n_5_0_ll.mmh"},
	{0x6023C6F4A06F8F8FULL, "vxlt_0_lowtown_n_h7l_0_0.vlm"},
	{0x6023C770D78F7489ULL, "ca_caverns1_26035.rml"},
	{0x60240577E4CC6C5BULL, "79977.ogg"},
	{0x60241EBE322B25A1ULL, "41820772.ogg"},
	{0x60248C8AD86B09D2ULL, "vxht_keep_42m_0_0.vlm"},
	{0x60249620C02F8720ULL, "vxht_mansion2_1si_0_0.vlm"},
	{0x60249C585FEEF5D1ULL, "6004104_m.fxe"},
	{0x6024F41B0EA88462ULL, "da2terrain_terrain_p0.vsh.fxo.dx11.assembly"},
	{0x602502C6F0A124C9ULL, "6174618_m.fxe"},
	{0x60257DBC6B968F79ULL, "341208034.ogg"},
	{0x6025BB65736907D8ULL, "6036694_m.fxe"},
	{0x602610187D02E1EFULL, "vxca_caverns1_jgi_0_0.vlm"},
	{0x60267A4C61D40097ULL, "6114100_m.fxe"},
	{0x60269240ECA9B8BEULL, "gen00fl_fenris6105604.cl"},
	{0x6026A1EB8BA22CE7ULL, "one350_terath.bnk"},
	{0x6026C7DD1DB207E8ULL, "llac_ht_keep_n_3740_0.dds"},
	{0x6026E0E17F0DB38AULL, "6133057_f.fxe"},
	{0x60270C2DCF26F61EULL, "896894749.ogg"},
	{0x60270D7F8DC3B775ULL, "117956956.ogg"},
	{0x6027751B961DCE68ULL, "6083823_m.fxe"},
	{0x602786A5B6B12249ULL, "666365429.ogg"},
	{0x602791E256EC4E09ULL, "235780374.ogg"},
	{0x6027D3F7C32E3B79ULL, "6125847_m.fxe"},
	{0x6028085820180687ULL, "vxw_bonepitmining_a2_iv_0_0.vlm"},
	{0x602811CA4FBCEB27ULL, "6074461_f.fxe"},
	{0x60281B01D7E9C939ULL, "298117234.ogg"},
	{0x60293ACEA89856E5ULL, "vxw_encounter2_camp_qs_0_1.vlm"},
	{0x6029407F4C95270BULL, "art\\vfx\\creatures\\fxc_rage_crust_c.mmh"},
	{0x6029C3EB8BA49C5FULL, "one350_terath.cnv"},
	{0x6029D14C762A62BEULL, "gen00fl_merrill6115753.cl"},
	{0x6029DE9532237722ULL, "hla_ht_0_hightown_n_65536_0z.dds"},
	{0x602B04C2311EBBFCULL, "ser221ip_blood_vials.ncs"},
	{0x602B48E48A3810B0ULL, "lt_0_lowtown_docks_a3_227_0.msh"},
	{0x602B56E48A3828CAULL, "lt_0_lowtown_docks_a3_227_0.mmh"},
	{0x602B5AE48A382F01ULL, "lt_0_lowtown_docks_a3_227_0.mao"},
	{0x602C18CB989BB8DCULL, "lt_0_lowtown_docks_a3_161_0.mao"},
	{0x602C24CB989BCD7FULL, "lt_0_lowtown_docks_a3_161_0.mmh"},
	{0x602C2ACB989BD721ULL, "lt_0_lowtown_docks_a3_161_0.msh"},
	{0x602CB6F9DEF3DF80ULL, "hlac_lt_0_lowtown_docks_n_-2_2z.dds"},
	{0x602DB2CFD5A688B5ULL, "crt_sf_center.mao"},
	{0x602DB7BEC5937D02ULL, "vxht_0_hightown_n_bos_0_0.vlm"},
	{0x602DF3914823F875ULL, "vxlt_0_lowtown_docks_n_b5r_0_1.vlm"},
	{0x602E2BB1377FAE7DULL, "hlac_w_blightlands_262147_0x.dds"},
	{0x602E88676E0FBF20ULL, "6163260_m.fxe"},
	{0x602EB87149F5729BULL, "vxlt_0_lowtown_n_euw_0_0.vlm"},
	{0x602F9072A5C26586ULL, "qcr221_cut_diplomatic_sol.cnv"},
	{0x602FBAD8DD298A88ULL, "6202267_f.fxe"},
	{0x60305C079550F8B4ULL, "6125796_m.fxe"},
	{0x6030FF65F05ED233ULL, "6077081_m.fxe"},
	{0x60312E25BB8AFB57ULL, "163561247.ogg"},
	{0x60313825CB2B4AB9ULL, "3553764.ogg"},
	{0x60314E5EBD21D522ULL, "13561712.ogg"},
	{0x60316E75D3717E53ULL, "6108642_m.fxe"},
	{0x60319211FBF60657ULL, "vxht_keep_n_6n1_0_0.vlm"},
	{0x6031B63D2F3272D1ULL, "894638803.ogg"},
	{0x6031B8BFB16A6637ULL, "vxw_encounter3_26_0_0.vlm"},
	{0x60320586D8E1CC1BULL, "gen00fl_isabela6099335.cl"},
	{0x60322119D27D7CA8ULL, "w_encounter3_177_0.msh"},
	{0x60323319D27D9B39ULL, "w_encounter3_177_0.mao"},
	{0x60323F19D27DAFD2ULL, "w_encounter3_177_0.mmh"},
	{0x60327492127778FAULL, "mcr310_argument6074487.cl"},
	{0x603278B8EBA551A2ULL, "t3_arm_pl2.tnt"},
	{0x60329272A5C49F46ULL, "qcr221_cut_diplomatic_sol.bnk"},
	{0x6032E6D52B9AA4AEULL, "hlac_lt_0_lowtown_a2_n_c_-65539_0y.dds"},
	{0x6032FE20A7826B3FULL, "hlac_lt_0_lowtown_a3_n_c_-65535_3x.dds"},
	{0x60330C7D387F4117ULL, "hlac_lt_0_lowtown_docks_-131074_5y.dds"},
	{0x6033321E38A4D9E3ULL, "6178366_m.fxe"},
	{0x603350125B93388FULL, "lt_0_lowtown_a2_21_0.rml"},
	{0x60338400FA5B5D83ULL, "vxca_caverns1_mines_iia_0_0.vlm"},
	{0x6033B5576B73300DULL, "vxdr_primevaltomb_6s6_0_0.vlm"},
	{0x6033E0ACD38DAF02ULL, "art\\characters\\playercharacter\\humanmale\\hlh_wic.mmh"},
	{0x6033E2ACD38DB2B8ULL, "art\\characters\\playercharacter\\humanmale\\hlh_wic.msh"},
	{0x6034018CB4EA489FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk03.mmh"},
	{0x6034078CB4EA5341ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk03.msh"},
	{0x60348C6936830552ULL, "6020421_f.fxe"},
	{0x6034F346BF9CCF59ULL, "em_fenris.mor"},
	{0x60357B9009F60CC1ULL, "dr_primevaltomb_664_lc.mmh"},
	{0x603618195C6BF0ABULL, "narrative_varric6124280.cl"},
	{0x6036A5EFF7416273ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masg_0n.dds"},
	{0x6036CA5E5D11DFC1ULL, "vxca_cave_2e1_0_0.vlm"},
	{0x6036D0ED88BCDF77ULL, "gen00fl_anders6183812.cl"},
	{0x60370AD023EBC3A7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackr_1.mao"},
	{0x6037217F4CA13070ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.anb"},
	{0x6037A8C05D737C56ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg01a.phy"},
	{0x6037C0224E9083F9ULL, "hlac_lt_gallowstemplar_n_2388_1z.dds"},
	{0x603854DF7BE713DFULL, "gen00fl_fenris6112067.cl"},
	{0x60388717FDE13770ULL, "vrc321_housekeeper6085870.cl"},
	{0x6039020E91538904ULL, "gen00fl_sebastian6178348.cl"},
	{0x6039161430BA16F8ULL, "mag110_samson6025385.cl"},
	{0x603920F253ECB268ULL, "1049365141.ogg"},
	{0x6039F575CABC781BULL, "qun200_party_comment6090410.cl"},
	{0x603A865A4ACD2F24ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candleoff_d.dds"},
	{0x603B4264E5513076ULL, "hlac_w_encounter1_0_3x.dds"},
	{0x603BACCA2B56F429ULL, "vxlt_gallowstemplar_f_51z_0_0.vlm"},
	{0x603BF6F8A01641CEULL, "llac_lt_0_lowtown_-196609_3.dds"},
	{0x603C5518D42E2067ULL, "vxlt_0_lowtown_n_ffu_0_0.vlm"},
	{0x603C5E2D1B2E6FABULL, "vxlt_0_lowtown_a3_n_c_hx2_0_0.vlm"},
	{0x603D0F643529D9A2ULL, "vxw_woundedcoast_1fh_0_0.vlm"},
	{0x603D2A4E3BEB8B1BULL, "vxlt_0_lowtown_n_env_0_0.vlm"},
	{0x603D60168E36D671ULL, "w_encounter3_107_0.msh"},
	{0x603D6DC2FD3AC403ULL, "25619895.ogg"},
	{0x603D6E168E36ED8CULL, "w_encounter3_107_0.mao"},
	{0x603D7A168E37022FULL, "w_encounter3_107_0.mmh"},
	{0x603DD135234A7E2FULL, "llac_lt_0_lowtown_docks_a3_n_-2_2.dds"},
	{0x603E066C66714AFDULL, "vxw_bonepit_1i1_0_0.vlm"},
	{0x603E2F6F789D0A45ULL, "504384070.ogg"},
	{0x603E977E06A3AF36ULL, "hlac_lt_gallowstemplar_f_2394_10z.dds"},
	{0x603F1955FE05AECAULL, "140495809.ogg"},
	{0x603F24FBA4DCED20ULL, "949893893.ogg"},
	{0x603F2E5733D6C341ULL, "596241134.ogg"},
	{0x603FAE8D5D207D8EULL, "6059981_m.fxe"},
	{0x603FB2490C7310EAULL, "dae000_orana6149055.cl"},
	{0x603FD73BA0480789ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lamp.phy"},
	{0x603FDC9892A7DA4FULL, "vxlt_0_lowtown_a2_n_hev_0_0.vlm"},
	{0x603FFDB2DB00A5E6ULL, "hlac_lt_0_lowtown_n_-131071_2x.dds"},
	{0x604003EEEDEC1ED7ULL, "vxht_keep_n_6i0_0_0.vlm"},
	{0x604010B85003E42FULL, "vxht_keep_6i7_0_0.vlm"},
	{0x6040457150612186ULL, "gen00fl_isabela6079059.cl"},
	{0x6040C57BE71E7D3BULL, "vxdr_0_deeproads_ck3_0_0.vlm"},
	{0x60414B8F228403EEULL, "6166815_f.fxe"},
	{0x6041D6920910D252ULL, "6101455_m.fxe"},
	{0x60424084935883ECULL, "hlac_w_sundermount_-65537_0x.dds"},
	{0x604291C05D7D5CF7ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg01a.msh"},
	{0x6042A7C05D7D8269ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg01a.mmh"},
	{0x6042C84BFF1AC56BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_gesture_head.evt"},
	{0x6042DB107CE30851ULL, "vxht_mansion_n_6rg_0_0.vlm"},
	{0x6042E12F452FB7DEULL, "vxht_brothel_n_33e_0_0.vlm"},
	{0x6042F56D4FC7037BULL, "zz_followers_banter6159120.cl"},
	{0x60434BD0D5AE690CULL, "741311908.ogg"},
	{0x60435CA72BD9F6F5ULL, "hlac_dr_slavers_3216_5x.dds"},
	{0x6043B929BFEAC3C1ULL, "23762483.ogg"},
	{0x6043BD26FD43869DULL, "1071665946.ogg"},
	{0x6044512CD36F239CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave1statue.mao"},
	{0x60445D2CD36F383FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave1statue.mmh"},
	{0x6044632CD36F42E1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave1statue.msh"},
	{0x6044660443129D5BULL, "hlac_w_sundermount_196606_0x.dds"},
	{0x6044E35C1ABCF92CULL, "hlac_lt_0_lowtown_docks_a3_n_65535_2x.dds"},
	{0x604582B92035C06FULL, "vxlt_0_lowtown_a3_n_c_eoo_0_1.vlm"},
	{0x6045BBA60E4A391AULL, "isa120_hayder6066916.cl"},
	{0x6045D956AD3B9EFDULL, "llac_lt_0_lowtown_docks_-131074_0.dds"},
	{0x6045F4833E51D238ULL, "art\\characters\\weapons\\staff\\w_stf_104a.mao"},
	{0x6046039ABB5FBC49ULL, "10926292.ogg"},
	{0x60461ABCB40AB19EULL, "gen_im_arm_cht_elf_comb_t2.uti"},
	{0x60463AE3606ACE78ULL, "6144102_m.fxe"},
	{0x6046587C1C892845ULL, "dae000cr_leonato.utc"},
	{0x6046ADE9174CC06BULL, "382451743.ogg"},
	{0x60470A4CC09AF1CEULL, "gen00fl_anders6121213.cl"},
	{0x6047288E7E3E0E06ULL, "vxlt_0_lowtown_a2_n_c_hz0_0_0.vlm"},
	{0x6047FC4C2975E2A9ULL, "vxlt_gallowstemplar_n_5pm_0_0.vlm"},
	{0x6048C37B1506DC4BULL, "vxlt_warehouse_n_wt_0_0.vlm"},
	{0x6048E14BFF1F5B2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_gesture_head.gad"},
	{0x60492BADB0561741ULL, "6140622_m.fxe"},
	{0x604A3A2E8F34E347ULL, "hf_templar_officer01.mor"},
	{0x604A680D8D0F148FULL, "163340446.ogg"},
	{0x604A69039C9ED300ULL, "68752675.ogg"},
	{0x604A8B71D0688758ULL, "470273629.ogg"},
	{0x604AD914E6C589EAULL, "hlac_lt_0_lowtown_a2_65537_3x.dds"},
	{0x604AD92BAE7299C2ULL, "art\\vfx\\models\\vfx_bomb_commander.mmh"},
	{0x604ADA7DA4A58E4DULL, "541609000.ogg"},
	{0x604ADB2BAE729D78ULL, "art\\vfx\\models\\vfx_bomb_commander.msh"},
	{0x604AE2C19923940FULL, "sst_ss_aveline.gda"},
	{0x604B0C9BED87D3A9ULL, "645190651.ogg"},
	{0x604B25B06C704D0BULL, "cth_cm6.tnt"},
	{0x604B42BD1CD040E2ULL, "vxdr_0_deeproads_cwj_0_0.vlm"},
	{0x604B8141271FF222ULL, "vxw_woundedcoast_1ai_0_0.vlm"},
	{0x604B9665B675C83BULL, "vxht_chantry_2c1_0_0.vlm"},
	{0x604B9C2B2DE1A39BULL, "vxlt_0_lowtown_n_eki_0_0.vlm"},
	{0x604BE4014C482082ULL, "lgt120_orz_bln_official6140178.cl"},
	{0x604C1351A9E0ED07ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_exit.gad"},
	{0x604C7CD012514EE5ULL, "vxlt_0_lowtown_a3_n_c_deg_0_0.vlm"},
	{0x604C845E7E5188D6ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_mered_chant_ex2.ani"},
	{0x604CDA61D4D1868CULL, "cd_act1_rockwraith_intro_demo_warrior.bnk"},
	{0x604CE0D89C4C9B8EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlb.mmh"},
	{0x604CECD89C4CB025ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlb.mao"},
	{0x604CFAD89C4CC87CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlb.msh"},
	{0x604D6714838370BEULL, "vxlt_backalley_n_1ea_0_0.vlm"},
	{0x604D817E5FA5E3AEULL, "vxdr_primevalentrance_2sq_0_0.vlm"},
	{0x604D90BD20D394B8ULL, "ave200_donnic6110721.cl"},
	{0x604E522FB75DAF53ULL, "6106442_m.fxe"},
	{0x604E5D80BD18D4DDULL, "6084881_m.fxe"},
	{0x604EA3B9BD7DDA75ULL, "vxlt_gallowstemplar_43f_0_0.vlm"},
	{0x604EE42814468FE6ULL, "mag321_comtesse_de_launcet6043399.cl"},
	{0x604FC5D90CC04F9DULL, "6119563_m.fxe"},
	{0x604FDF0328E78553ULL, "dae000cr_grand_cleric6124755.cl"},
	{0x604FE04BFF25782AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_gesture_head.ani"},
	{0x604FFC67666029C9ULL, "plt_prm000pt_parth_staff_.nss"},
	{0x605078083C263293ULL, "vxlt_gallowscourtyard_75k_0_0.vlm"},
	{0x60509F60B67D02E2ULL, "738462888.ogg"},
	{0x60511B8DCBED728DULL, "prp_str_dbone_s.dds"},
	{0x60513DDC819D8C81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\dm_injured_movement.anb"},
	{0x6051690DF0CA8ED0ULL, "323043982.ogg"},
	{0x60518A0B507224BAULL, "vxht_mansion2_1au_0_0.vlm"},
	{0x6052A19226D1B123ULL, "qcr220_meredith6054746.cl"},
	{0x6052C40987ADFFF9ULL, "dae000tr_assassin.ncs"},
	{0x60530524EC32B062ULL, "6134828_m.fxe"},
	{0x60530851A9E6F914ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_exit.evt"},
	{0x60531640E53A6D6CULL, "vxdr_slavers_211_0_0.vlm"},
	{0x6053ECE814563927ULL, "mag110_samson6016171.cl"},
	{0x6054ACC05260E6F8ULL, "6141202_m.fxe"},
	{0x6054C3A585F5EEBDULL, "mag160cr_varian_ilithis.utc"},
	{0x6054CFF251CF8112ULL, "vxw_bonepitdestroyed_a3_1l1_0_0.vlm"},
	{0x6055163B5745D3AFULL, "593942454.ogg"},
	{0x6055557B6E71AC1FULL, "686365223.ogg"},
	{0x60557D95966389D6ULL, "66432812.ogg"},
	{0x6055FEF841CB6474ULL, "zz_qun_debug6038057.cl"},
	{0x6056426E297C12FBULL, "576422997.ogg"},
	{0x605642C348436D75ULL, "227232449.ogg"},
	{0x60569FFAC415FCDFULL, "cd_dre000_cut_bartrandbetrayal.bnk"},
	{0x6056C53F4E6511E7ULL, "dae000_madam_lusine6152227.cl"},
	{0x6056FA3F27E1EEA8ULL, "427624940.ogg"},
	{0x60570D75F7C55AEFULL, "vxw_encounter2_camp_ju_0_0.vlm"},
	{0x6057355806FCF4CBULL, "vxlt_warehouse_n_zw_0_0.vlm"},
	{0x6057C7E62FB81C3CULL, "505478533.ogg"},
	{0x6057DEF15757A05BULL, "vxlt_warehouse_13h_0_0.vlm"},
	{0x6058180EA415EAC1ULL, "hlac_ht_0_hightown_n_-65536_3x.dds"},
	{0x6058F22E97A0327BULL, "gen00fl_anders6125193.cl"},
	{0x60591EEC9B8A1CEDULL, "40527793.ogg"},
	{0x6059949157CB1E2CULL, "db_sh.mao"},
	{0x60599642FC0845F7ULL, "265080385.ogg"},
	{0x6059C5CC7A63CD23ULL, "lt_0_lowtown_docks_227_0.msh"},
	{0x6059D3CC7A63E57DULL, "lt_0_lowtown_docks_227_0.mmh"},
	{0x6059D7CC7A63EB8EULL, "lt_0_lowtown_docks_227_0.mao"},
	{0x6059F31E19160AA2ULL, "vxw_woundedcoast_1lj_0_0.vlm"},
	{0x605ABD11FF2696B8ULL, "vxlt_hangedman_n_20n_0_0.vlm"},
	{0x605AC1AC0765D43EULL, "gen00fl_aveline6122123.cl"},
	{0x605AC62203357B81ULL, "420896826.ogg"},
	{0x605AE9F6784FD554ULL, "vxca_caverns1_mines_ehm_0_0.vlm"},
	{0x605B178DF3250857ULL, "6105686_f.fxe"},
	{0x605C0A162ECAFDC0ULL, "lowtown_fem_urchin_mob6165914.cl"},
	{0x605C4774C5BB6714ULL, "6038227_m.fxe"},
	{0x605C8282FBE42BA5ULL, "vxw_encounter3_gb_0_1.vlm"},
	{0x605CB413EEE4C564ULL, "mage_wintersgrasp.dds"},
	{0x605D1596AF73F2F5ULL, "vxlt_gallowstemplar_46e_0_0.vlm"},
	{0x605DF19960C662FEULL, "vxlt_hangedman_n_280_0_0.vlm"},
	{0x605E4EDD770B728CULL, "6116174_m.fxe"},
	{0x605EA2FA702433F1ULL, "6070446_m.fxe"},
	{0x605F3A6FC265107FULL, "6071847_m.fxe"},
	{0x605F98EF648EDFF4ULL, "fx_icefloor.mao"},
	{0x605FD5E6F94FAA49ULL, "gen_im_trs_slv_hvy_glv_r2.uti"},
	{0x605FEF72F0F3BDEAULL, "6095084_m.fxe"},
	{0x605FFBE842683D3AULL, "vxht_mansion2_1dv_0_0.vlm"},
	{0x60603BCEE55F5FC5ULL, "one140cr_charade6137641.cl"},
	{0x6060450EFC429DD6ULL, "dre000_vampire6034281.cl"},
	{0x60607D28873F738EULL, "one300_nuncio6083405.cl"},
	{0x6060AD5E7E627710ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_mered_chant_ex2.gad"},
	{0x6060CE210DD4F1F5ULL, "hlac_ht_0_hightown_-65536_0z.dds"},
	{0x60610051A9F3294DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_exit.ani"},
	{0x60612147E5B95953ULL, "vxht_mansion_n_5c1_0_0.vlm"},
	{0x606199F1CD8D905CULL, "llac_ht_keep_n_4519_1.dds"},
	{0x6061A1FD3D72E676ULL, "vxlt_0_lowtown_a2_n_f4n_0_0.vlm"},
	{0x6061CC069DE55D25ULL, "hlac_ht_mansion_n_3373_2y.dds"},
	{0x606226551C2D2E4AULL, "829126217.ogg"},
	{0x60622A951F7FECC3ULL, "vxlt_gallowscourtyard_n_5a6_0_0.vlm"},
	{0x60625B9B5F1CCF50ULL, "dae000ip_crowd_mf3_talk10.utp"},
	{0x6062958E1571C76AULL, "6139980_m.fxe"},
	{0x60633B2C3FC5F5FEULL, "gen_im_cut_dagger.uti"},
	{0x606356273D991EEAULL, "mag291st_feynriel.stg"},
	{0x60636687C578556AULL, "534028853.ogg"},
	{0x60637800C5F9AE65ULL, "mcr370_meredith6083588.cl"},
	{0x606394CF7B02E9DDULL, "6051898_m.fxe"},
	{0x60640C48622A3F06ULL, "vxlt_0_lowtown_a2_n_c_hp2_0_0.vlm"},
	{0x606452473347D45EULL, "540767908.ogg"},
	{0x606456CFB74C387BULL, "vxw_sundermount_200_0_0.vlm"},
	{0x60649398A2B45EF3ULL, "725542972.ogg"},
	{0x6064C7973158BB6BULL, "vxht_0_hightown_c9v_0_0.vlm"},
	{0x6064E93B9D2AB29DULL, "vxca_caverns1_mines_k2r_0_2.vlm"},
	{0x606559725B6AEB92ULL, "lgt260_eau_tmp_roderick6143735.cl"},
	{0x60657F52E9BB736FULL, "vxw_encounter2_camp_et_0_0.vlm"},
	{0x606650CE494DB8DBULL, "vxlt_warehouse_16k_0_0.vlm"},
	{0x6066658350ECC675ULL, "vxlt_0_lowtown_docks_as1_0_0.vlm"},
	{0x60667BADC3EAF9E9ULL, "abi_r_stealth.wbd"},
	{0x6067B3C3F110B92BULL, "vxlt_0_lowtown_a3_n_c_ho7_0_0.vlm"},
	{0x6067B68C71F9769FULL, "vxca_caverns1_mines_icu_0_2.vlm"},
	{0x6068267700BC71E2ULL, "vxdr_0_deeproads_cit_0_0.vlm"},
	{0x60687A1F9A61F93BULL, "vxht_chantry_2y3_0_0.vlm"},
	{0x6068ECFE6B355DE8ULL, "vxca_caverns2_8cs_0_0.vlm"},
	{0x6069EFDA68CB09D7ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em02.phy"},
	{0x606A4ACE676FA1BEULL, "vxlt_backalley_n_1kc_0_0.vlm"},
	{0x606A9976839A346FULL, "hlac_lt_0_lowtown_a2_n_c_65534_6x.dds"},
	{0x606A9B0A62A61063ULL, "6115411_m.fxe"},
	{0x606AA1FDA70DF7D2ULL, "vxht_mansion_n_5c4_0_0.vlm"},
	{0x606AE65EE65C6965ULL, "hlac_w_bonepitmining_a2_196614_0z.dds"},
	{0x606B001D5007A6FAULL, "vrc321_varric6087053.cl"},
	{0x606B01A064872C0EULL, "rules_core.ncs"},
	{0x606B1B2A8C9B3668ULL, "hlac_ca_cavern1_haunted_24425_12x.dds"},
	{0x606B21F5D92F8818ULL, "dae000pt_amb_anywhere.ncs"},
	{0x606B51019065C9CCULL, "t1_skn_d08.tnt"},
	{0x606B8B63A4D71E44ULL, "art\\levels\\ztools\\default\\generic\\9ff_cylinder.msh"},
	{0x606B9D63A4D73CADULL, "art\\levels\\ztools\\default\\generic\\9ff_cylinder.mao"},
	{0x606BA163A4D743E6ULL, "art\\levels\\ztools\\default\\generic\\9ff_cylinder.mmh"},
	{0x606BAF66365AE0A6ULL, "w_bonepitdestroyed_a3_171_0.phy"},
	{0x606BBD79FEC4A7FFULL, "vxw_encounter3_fb_0_0.vlm"},
	{0x606BC03BA06D1BAEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lamp.mmh"},
	{0x606BCC3BA06D3045ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lamp.mao"},
	{0x606BDA3BA06D479CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lamp.msh"},
	{0x606C1C3F76293595ULL, "6087167_m.fxe"},
	{0x606C72F4171A649CULL, "vxdr_primevalentrance_2ni_0_1.vlm"},
	{0x606D14D7621A4C71ULL, "6075746_m.fxe"},
	{0x606D14E7BBA65604ULL, "vxlt_0_lowtown_docks_n_b4e_0_0.vlm"},
	{0x606DE7C85C0CC4CFULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha06a.phy"},
	{0x606E16F24B95530FULL, "vxht_keep_6n2_0_0.vlm"},
	{0x606E2B33DD3C31F7ULL, "gen00fl_aveline6122814.cl"},
	{0x606E71F15E7F5F64ULL, "vxca_caverns1_mines_k2o_0_2.vlm"},
	{0x606F1CF1C6D9CD9EULL, "vxht_0_hightown_a3_c71_0_1.vlm"},
	{0x606F514B3D5F38E1ULL, "617584417.ogg"},
	{0x606FC13D7D5AD053ULL, "vxlt_hangedman_24v_0_0.vlm"},
	{0x607006C052261D62ULL, "zz_and_debug6178611.cl"},
	{0x6071C54F56E31C4BULL, "llac_lt_0_lowtown_131071_1.dds"},
	{0x6072DC1088FD1F7FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8x4holebase.phy"},
	{0x6072DE9B2C9AD4F2ULL, "llac_lt_0_lowtown_n_131072_2.dds"},
	{0x607337C6215A0DE1ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_angrydebate_1_ian.gad"},
	{0x607351E2FF582C29ULL, "vxlt_gallowstemplar_n_5ar_0_0.vlm"},
	{0x60738D8FF4F8F9A4ULL, "314281777.ogg"},
	{0x6073F89BF89B1964ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplarfemale6_.ani"},
	{0x607455A7EEF50B4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v2.ani"},
	{0x607468CF32237FE5ULL, "gen00fl_anders6111192.cl"},
	{0x6074704C7F432242ULL, "mag110_samson6018703.cl"},
	{0x60749CDFAB84B4B9ULL, "472292788.ogg"},
	{0x6074C4D0AF6F2747ULL, "6208496_m.fxe"},
	{0x60751907256111B4ULL, "6127186_f.fxe"},
	{0x6075493B83AF5E94ULL, "111454950.ogg"},
	{0x60756B9984EED6CEULL, "vxht_0_hightown_n_bww_0_0.vlm"},
	{0x60757DB6CBA2BF05ULL, "vxht_brothel_n_4dd_0_0.vlm"},
	{0x60759863A4DF88B1ULL, "art\\levels\\ztools\\default\\generic\\9ff_cylinder.phy"},
	{0x6075A1DBC325FF0DULL, "vxw_bonepit_bo_0_0.vlm"},
	{0x6075C38E28EC6C62ULL, "6108257_f.fxe"},
	{0x6075DFEB3F93BDB2ULL, "dae000ip_crowd_mf1_talk14.utp"},
	{0x60769853F2B28A62ULL, "vxdr_0_deeproads_cfu_0_0.vlm"},
	{0x6076ADBD7256CB4DULL, "llac_lt_undercityserialkiller_-1_32.dds"},
	{0x6076D6D7FD023BA2ULL, "vxw_woundedcoast_1rt_0_0.vlm"},
	{0x6076E59E9506CFDAULL, "qun120_arishok_guard_nt6063078.cl"},
	{0x60771BE004F8771EULL, "mag120_macha6080654.cl"},
	{0x60782E5BA53327F2ULL, "vxw_encounter3_63_0_0.vlm"},
	{0x607938C6215E7A8FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_angrydebate_1_ian.ani"},
	{0x60794617A6E5ADDCULL, "68693464.ogg"},
	{0x6079CDA837B841BBULL, "284040957.ogg"},
	{0x607A2F56F0BAC07FULL, "vxw_encounter3_cm_0_0.vlm"},
	{0x607A640C0330653EULL, "148531585.ogg"},
	{0x607AB716D356FCD2ULL, "vrc221_bartrand6043394.cl"},
	{0x607AEF9219627DA2ULL, "6169347_m.fxe"},
	{0x607AF59BF8A132C6ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplarfemale6_.gad"},
	{0x607B096FCC98CBAFULL, "hlac_lt_undercityencounter_-1_23z.dds"},
	{0x607B8104E6B77AA9ULL, "llac_lt_0_lowtown_a2_n_c_-1_7.dds"},
	{0x607BAF24DCEA6569ULL, "100651163.ogg"},
	{0x607BB4BF98ABA12EULL, "vxht_0_hightown_n_aqv_0_0.vlm"},
	{0x607BEDF71450121EULL, "vxw_encounter1_bq_0_0.vlm"},
	{0x607C01A8B606FE59ULL, "933874162.ogg"},
	{0x607C0A43B764F97FULL, "vxht_keep_6bv_0_0.vlm"},
	{0x607C3F077559546EULL, "401001022.ogg"},
	{0x607CBC785EEC6126ULL, "follower_banter6173107.cl"},
	{0x607D8A8A113EA2DCULL, "848281140.ogg"},
	{0x607D8B443FD411E6ULL, "dae200_saemus6142367.cl"},
	{0x607D97489CB883BBULL, "fake_warrior_male.utc"},
	{0x607E0501C9A58A53ULL, "vxht_mansion_n_5i3_0_0.vlm"},
	{0x607E6E06DAF0998AULL, "366842595.ogg"},
	{0x607EC2120592C1BFULL, "6036971_f.fxe"},
	{0x607F26DDF1AB82CCULL, "vxw_bonepitdestroyed_a3_244_0_0.vlm"},
	{0x607F30899A4C887CULL, "vxht_bartrand_n_3jh_0_0.vlm"},
	{0x607FEC696F0F5A9CULL, "mcr371_orsino6081270.cl"},
	{0x607FEFFE45223848ULL, "hlac_ht_0_hightown_a3_131071_0y.dds"},
	{0x6080681FE66FC855ULL, "760991661.ogg"},
	{0x608162693523726EULL, "hlac_ht_mansion2_818_13x.dds"},
	{0x6081E066AFB586CDULL, "vxlt_0_lowtown_a2_n_c_ewm_0_0.vlm"},
	{0x60822FF0D8DC57F1ULL, "6084261_m.fxe"},
	{0x60824DA7EF013A9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v2.evt"},
	{0x608285FAC43B0B15ULL, "cd_dre000_cut_bartrandbetrayal.wbd"},
	{0x6083946CEC2E1CE0ULL, "vxdr_primevaltomb_354_0_0.vlm"},
	{0x6083DD10890B9384ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8x4holebase.mmh"},
	{0x6083E110890B9AD7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8x4holebase.mao"},
	{0x6083F310890BB926ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8x4holebase.msh"},
	{0x608413B8B51C178DULL, "vxw_bonepit_mn_0_0.vlm"},
	{0x60850A575E5F51B7ULL, "6083351_f.fxe"},
	{0x60853A7A5DF01D96ULL, "6068574_m.fxe"},
	{0x60855D65D274D863ULL, "421577980.ogg"},
	{0x60856D6BE2127C4DULL, "llac_lt_0_lowtown_docks_a3_-65538_2.dds"},
	{0x6085A3CE6F6B19ADULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urva_0t.dds"},
	{0x6085B61DCF81CE31ULL, "scaleform_sf_main_p1.psh.fxo.dx11.assembly"},
	{0x60861E536A311C19ULL, "vxlt_gallowstemplar_f_4qo_0_0.vlm"},
	{0x60863F8D4E321ED4ULL, "vxca_caverns1_mines_egp_0_0.vlm"},
	{0x6086A6DA3B153995ULL, "dae000ip_crowd_lm2_talk3.utp"},
	{0x6086EA6C73FA15D2ULL, "601791143.ogg"},
	{0x6087009E4156F0ABULL, "llac_lt_0_lowtown_a2_65536_1.dds"},
	{0x608708926094A0C3ULL, "hlac_lt_0_lowtown_a3_n_c_65534_0y.dds"},
	{0x608709870B726247ULL, "mrl150ip_altar.utp"},
	{0x608785B78AFA28D2ULL, "vxht_mansion_n_5i6_0_0.vlm"},
	{0x608786BE56D880F3ULL, "isa211_isabela6070584.cl"},
	{0x60886F59DBF6BF32ULL, "73991933.ogg"},
	{0x6088BDD8720F481BULL, "6153798_m.fxe"},
	{0x6088E6D4221845B6ULL, "m_qun101ar_talvashoth_cave.wbd"},
	{0x6088F19D98455AAAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv6.gad"},
	{0x60892EA7EF072488ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d1_v2.gad"},
	{0x608990EB2081DA1EULL, "6125555_f.fxe"},
	{0x6089D805252E363CULL, "one110_pryce6088794.cl"},
	{0x608A5FD406462A9EULL, "vxw_encounter1_gr_0_0.vlm"},
	{0x608AA2FA3252256DULL, "6156346_m.fxe"},
	{0x608AB9F01F07E568ULL, "758897145.ogg"},
	{0x608ADA3ADEF3A71AULL, "dre000_stroud6036682.cl"},
	{0x608AE776F0EB9928ULL, "6063555_m.fxe"},
	{0x608AF4371A13817DULL, "hlac_ca_caverns1_24426_6x.dds"},
	{0x608B146136B182D1ULL, "vxht_mansion_n_6iz_0_0.vlm"},
	{0x608B4BD8FB3CAF45ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.mmh"},
	{0x608B8E83C647B583ULL, "llac_lt_0_lowtown_docks_a3_n_196607_2.dds"},
	{0x608BD950914B5AD9ULL, "mrl000_marethari6129312.cl"},
	{0x608C22F19EA51491ULL, "fen110_party_comment6095074.cl"},
	{0x608C43C4F633F4E6ULL, "6188113_m.fxe"},
	{0x608C4DADC40B7ACBULL, "abi_r_stealth.bnk"},
	{0x608C61A4CB437A63ULL, "6018709_m.fxe"},
	{0x608C9B7828D239A5ULL, "71976742.ogg"},
	{0x608CA1D37D9F9E85ULL, "6095826_f.fxe"},
	{0x608CA29CFD97FD37ULL, "vxdr_primevalentrance_379_0_0.vlm"},
	{0x608CE461D5077BBEULL, "cd_act1_rockwraith_intro_demo_warrior.wbd"},
	{0x608DA88A75317FE4ULL, "6112996_m.fxe"},
	{0x608E3E19D2CC0CCDULL, "w_encounter3_177_0.phy"},
	{0x608ED20228A74231ULL, "vxlt_0_lowtown_a2_cml_0_0.vlm"},
	{0x608F2138D2DFA424ULL, "gen_im_wep_war_shd_act0_02.uti"},
	{0x608FA8627DB47E77ULL, "372730706.ogg"},
	{0x608FAB65DF4D8D74ULL, "zz_dae_debug6029475.cl"},
	{0x608FD49D984B4841ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv6.evt"},
	{0x608FE2CBB51FA109ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_basic.mao"},
	{0x608FEECBB51FB5A2ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_basic.mmh"},
	{0x608FF075D4497351ULL, "6108132_m.fxe"},
	{0x608FF0CBB51FB958ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_basic.msh"},
	{0x609038ABC71E6602ULL, "779986189.ogg"},
	{0x6090546BD338AD24ULL, "480312778.ogg"},
	{0x609070FB36ADDE41ULL, "zz_mageslate_white.cut"},
	{0x609099E2303AC669ULL, "static_motionvectors_p0.vsh.fxo"},
	{0x6090B7648CA09F01ULL, "gen00fl_anders6096631.cl"},
	{0x609187CBC47BA6DDULL, "649854713.ogg"},
	{0x6091AD524018B302ULL, "6076865_m.fxe"},
	{0x60926C97745250F7ULL, "vxdr_0_deeproads_fjy_0_0.vlm"},
	{0x60930E23AE0D6327ULL, "hlac_w_blightlands_0_11z.dds"},
	{0x6093439413DDE055ULL, "6107098_m.fxe"},
	{0x6093C34FA7B611F5ULL, "173173007.ogg"},
	{0x6093D1C2554FE017ULL, "835188219.ogg"},
	{0x609442954117A768ULL, "vxca_caverns2_8pn_0_0.vlm"},
	{0x6094E74F941B95E7ULL, "6042508_m.fxe"},
	{0x6094EC46DCD42523ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_broken.mao"},
	{0x6094F046DCD42C70ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_broken.mmh"},
	{0x6094FE46DCD4438AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_broken.msh"},
	{0x6094FF978EC55390ULL, "31595155.ogg"},
	{0x609656B365065335ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian.anb"},
	{0x6096712B33E3E723ULL, "vxlt_0_lowtown_docks_a3_d0x_0_0.vlm"},
	{0x60968B3C48617C68ULL, "114859141.ogg"},
	{0x6097004A41150D6EULL, "30053650.ogg"},
	{0x60970A8CB53ED5E4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk03.phy"},
	{0x60971406CD1146E5ULL, "hlac_lt_0_lowtown_131070_1y.dds"},
	{0x60972FB56405609BULL, "6156498_m.fxe"},
	{0x60979589CD9B2071ULL, "800485385.ogg"},
	{0x6097C132A69D993AULL, "ser221_party_comments6163782.cl"},
	{0x6097CEA51AEDBD72ULL, "6208317_f.fxe"},
	{0x6097E3B0AE572939ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtc.msh"},
	{0x6097EAFEEEC4AE61ULL, "hlac_lt_0_lowtown_docks_a3_-65537_2y.dds"},
	{0x6097EDB0AE573A27ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtc.mmh"},
	{0x6097F1B0AE574174ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtc.mao"},
	{0x609802C81277F29EULL, "6122455_f.fxe"},
	{0x60983D0727470609ULL, "gen_im_cns_flu_varric_03.uti"},
	{0x6098C0D63C2F814EULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.cess.ani"},
	{0x609948D8FB48E712ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.anb"},
	{0x60994F2870D134CAULL, "and331ip_loot_1.utp"},
	{0x6099863E28A79B51ULL, "vxht_mansion_n_6ly_0_0.vlm"},
	{0x609A153DCD9F6E60ULL, "6118244_m.fxe"},
	{0x609A2BA38C29A179ULL, "vxdr_primevaltomb_4gx_0_0.vlm"},
	{0x609AE73088D6C812ULL, "vxdr_primevaltomb_8q1_0_0.vlm"},
	{0x609B18E3069120F6ULL, "zz_dae_debug6206726.cl"},
	{0x609BC1DF5DE89901ULL, "narrative_banter6149655.cl"},
	{0x609C2FE203F88D55ULL, "vxht_brothel_n_5kp_0_0.vlm"},
	{0x609C5287E5B5F97FULL, "6053670_m.fxe"},
	{0x609C7CC58B6D61F7ULL, "611707887.ogg"},
	{0x609CCC9D9855C490ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv6.ani"},
	{0x609E09FC4FF26816ULL, "vxlt_undercityencounter_1bt_0_1.vlm"},
	{0x609F281C02E67B13ULL, "vxht_0_hightown_a3_bc4_0_0.vlm"},
	{0x609FAFFB66FBC78AULL, "vxht_chantry_2l5_0_0.vlm"},
	{0x609FB7F0F67C7643ULL, "521301766.ogg"},
	{0x60A009630432F17EULL, "hlac_ca_cavern1_haunted_24424_9y.dds"},
	{0x60A026DACF239528ULL, "framebuffereffects_screencolorclear_p0.psh.fxo.dx11.assembly"},
	{0x60A0A2092BC6D4A8ULL, "vxlt_gallowstemplar_419_0_0.vlm"},
	{0x60A0CC0C2D99E154ULL, "hlac_lt_gallowstemplar_f_2394_3x.dds"},
	{0x60A1360FC92B05BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_exit.ani"},
	{0x60A1B2A8F4BC33A1ULL, "872954453.ogg"},
	{0x60A1E4F9EF4F0DF2ULL, "vxlt_0_lowtown_a2_ep2_0_0.vlm"},
	{0x60A25055DA6E31DFULL, "105345658.ogg"},
	{0x60A25A073E989998ULL, "mcr370_meredith6083068.cl"},
	{0x60A261FEB48E07F0ULL, "mag160_feynriel6016565.cl"},
	{0x60A290C89F8A83B0ULL, "primitives_primitive_p6.vsh.fxo.dx11"},
	{0x60A2EAD3A2471125ULL, "6115897_m.fxe"},
	{0x60A411348B8E3808ULL, "l_po_s.mao"},
	{0x60A41D567D789C39ULL, "vxw_encounter1_d0_0_0.vlm"},
	{0x60A42C19760C8F07ULL, "mcr370_negotiation6080230.cl"},
	{0x60A471EEF19DEBECULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_exit.ani"},
	{0x60A472B6EF594EEEULL, "dae000_porfiria6156623.cl"},
	{0x60A498A4B0FB5102ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masd_0d.dds"},
	{0x60A52FDB1FA0F685ULL, "6109399_m.fxe"},
	{0x60A5CFE5698AC12EULL, "ht_0_hightown_a2_mmap.dds"},
	{0x60A5F1DEC80EBDA5ULL, "6201574_m.fxe"},
	{0x60A6014F58007C10ULL, "6076382_m.fxe"},
	{0x60A6156B934EA69DULL, "vxca_cavern1_haunted_j4c_0_0.vlm"},
	{0x60A6B12FC5AAD325ULL, "215934269.ogg"},
	{0x60A6B8D63C3BB09FULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.cess.evt"},
	{0x60A6CCF9407CA95DULL, "6128973_f.fxe"},
	{0x60A72C7727E014A8ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.anb"},
	{0x60A7DAFD1BEE288AULL, "m_dae270ar_gallows.wbd"},
	{0x60A8AEF67543E6F8ULL, "6176576_m.fxe"},
	{0x60A8F951D096EC27ULL, "mag121_idunna6084255.cl"},
	{0x60A950FA5614A31BULL, "49075651.ogg"},
	{0x60A957139500777FULL, "47591689.ogg"},
	{0x60AA2B941F1C0D73ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtc_0d.dds"},
	{0x60AA5B907765109DULL, "ser270_emeric6113638.cl"},
	{0x60AA7EEEF1A26D2EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_exit.gad"},
	{0x60AAC06F98B94B97ULL, "vxca_cave_nm_0_1.vlm"},
	{0x60AADB38E28D337EULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_disp_p.mmh"},
	{0x60ABC0ADBE14C6B8ULL, "zz_mer_debug6045141.cl"},
	{0x60ABC0C57C6DDB34ULL, "vxw_encounter2_camp_8x_0_0.vlm"},
	{0x60ABF7E4B74E2C9FULL, "qcr000_arishok6059676.cl"},
	{0x60ACF10C36BBB3FCULL, "6208891_m.fxe"},
	{0x60AD337B30827406ULL, "vxlt_ruinedhouse_n_ia_0_0.vlm"},
	{0x60AD5DD9811CC3AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_sigh_big.ani"},
	{0x60AD6D10035CDC06ULL, "vxht_0_hightown_a3_9zm_0_0.vlm"},
	{0x60AD99D63C419A88ULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.cess.gad"},
	{0x60ADA75285DAF413ULL, "llac_lt_0_lowtown_n_-262144_1.dds"},
	{0x60ADC7DFF481A7F6ULL, "585182168.ogg"},
	{0x60AE0A65DBB2BF26ULL, "vxlt_0_lowtown_a2_n_fea_0_0.vlm"},
	{0x60AE199CD93DA58CULL, "art\\characters\\playercharacter\\humanfemale\\boo_wrba.msh"},
	{0x60AE1F9CD93DB07EULL, "art\\characters\\playercharacter\\humanfemale\\boo_wrba.mmh"},
	{0x60AE2B9CD93DC415ULL, "art\\characters\\playercharacter\\humanfemale\\boo_wrba.mao"},
	{0x60AE315389AB07DFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_keeplogo_da.dds"},
	{0x60AE3E0FC9359DB2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_exit.evt"},
	{0x60AE93DCC293C15AULL, "vxca_caverns1_mines_fo9_0_0.vlm"},
	{0x60AEB0E5E79E43ADULL, "6076483_f.fxe"},
	{0x60AF12B2EC9A2E55ULL, "dre000_dougal6206175.cl"},
	{0x60AF90CE0B51A364ULL, "dae000_bartrand6027477.cl"},
	{0x60AFF788AB4223ABULL, "vxht_brothel_n_6dp_0_0.vlm"},
	{0x60B0DBD887643FD3ULL, "hlac_w_bonepit_327683_0z.dds"},
	{0x60B0ED0BC33C13D7ULL, "ico_armor_medium_5.dds"},
	{0x60B0F7703258B70BULL, "vxht_mansion_n_4h5_0_0.vlm"},
	{0x60B15CB0943D29A5ULL, "6110797_m.fxe"},
	{0x60B16598227F0A82ULL, "clf260_harley6089679.cl"},
	{0x60B179EEF1A8830DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\bless_exit.evt"},
	{0x60B1C2CB7B90AD8AULL, "vxdr_0_deeproads_cj1_0_0.vlm"},
	{0x60B1D0DA69088ACCULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em02.mmh"},
	{0x60B1D6DA690894BEULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em02.msh"},
	{0x60B1ECCC365945D7ULL, "vxlt_gallowstemplar_34z_0_0.vlm"},
	{0x60B22B1DC01EC1A0ULL, "plt_cod_aow_spellcombo2.plo"},
	{0x60B286DA5D7C5B7DULL, "vxca_caverns2_8ea_0_1.vlm"},
	{0x60B28F336F6EB4B9ULL, "vxw_encounter1_i7_0_0.vlm"},
	{0x60B2A2D208EB1206ULL, "648575186.ogg"},
	{0x60B2B3F11F1F599EULL, "zz_and_debug6134174.cl"},
	{0x60B2B7AB77730972ULL, "vxht_keep_n_6y9_0_0.vlm"},
	{0x60B30D0B9A658124ULL, "6070252_m.fxe"},
	{0x60B345B4F51818B7ULL, "vxlt_0_lowtown_ewj_0_0.vlm"},
	{0x60B3A2307A801800ULL, "304764871.ogg"},
	{0x60B3DF0C55C4450DULL, "llac_dr_primevaltomb_661_0.dds"},
	{0x60B4D1E2405CC68DULL, "art\\characters\\playercharacter\\humanfemale\\cth_beta.mmh"},
	{0x60B4E3E2405CE513ULL, "art\\characters\\playercharacter\\humanfemale\\cth_beta.msh"},
	{0x60B50CD4223DC124ULL, "m_qun101ar_talvashoth_cave.bnk"},
	{0x60B5178928114E60ULL, "vxw_bonepit_26b_0_0.vlm"},
	{0x60B51AF06F6EA626ULL, "707694117.ogg"},
	{0x60B55ADCE789D54AULL, "hlac_ht_0_hightown_a3_0_1y.dds"},
	{0x60B55D0FC93BF145ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_exit.gad"},
	{0x60B57B2036C0488EULL, "dre000_cut2d_ragstoriches6025091.cl"},
	{0x60B5A163A0213E81ULL, "gen00cs_cs_end_music_int_2.ncs"},
	{0x60B60B36D7EB78C7ULL, "6060510_f.fxe"},
	{0x60B65702756D9EA4ULL, "450812581.ogg"},
	{0x60B6790ED5175F3BULL, "255462708.ogg"},
	{0x60B68F27014AA7B9ULL, "238371678.ogg"},
	{0x60B6D7514D5DE620ULL, "follower_banter6163232.cl"},
	{0x60B6F8F7B18B9F60ULL, "6112044_m.fxe"},
	{0x60B73196CA4AF980ULL, "character_character_p124.psh.fxo.dx11.assembly"},
	{0x60B7E7B9008BA10BULL, "vxw_bonepit_1wx_0_1.vlm"},
	{0x60B7EBB688443366ULL, "246078737.ogg"},
	{0x60B8030D3F669F24ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masd_0d.dds"},
	{0x60B813FD11269AADULL, "vxht_0_hightown_a3_90d_0_0.vlm"},
	{0x60B81D6F31502645ULL, "dae000_bartrand6027508.cl"},
	{0x60B863D2BF35C0A8ULL, "vxlt_0_lowtown_docks_aqb_0_0.vlm"},
	{0x60B8BA38E299383DULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_disp_p.anb"},
	{0x60B8CE4EB7CB348DULL, "llac_w_bonepit_0_7.dds"},
	{0x60B96D94E3392D88ULL, "hlac_ca_caverns2prc_5802_63y.dds"},
	{0x60B9923455EC356EULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf02.msh"},
	{0x60B9AC3455EC615CULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf02.mmh"},
	{0x60B9CB7B58255368ULL, "vxlt_gallowstemplar_5je_0_0.vlm"},
	{0x60BA07595ED3D305ULL, "vxlt_0_lowtown_a2_h2j_0_0.vlm"},
	{0x60BA32A26E63F3B4ULL, "vxw_encounter2_camp_3y_0_0.vlm"},
	{0x60BA592EA4EB71C0ULL, "hlac_dr_0_deeproads_5839_3z.dds"},
	{0x60BAEA55DEFCE98AULL, "dae000ip_craft_runes.utp"},
	{0x60BB48C6E4039DF6ULL, "6102145_f.fxe"},
	{0x60BB55D98128F3FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_sigh_big.evt"},
	{0x60BB65B23292AC31ULL, "vxlt_gallowscourtyard_6x3_0_0.vlm"},
	{0x60BC254C54AC24A1ULL, "vxw_bonepitdestroyed_a3_25z_0_0.vlm"},
	{0x60BC5F4A1CA8BED1ULL, "vxlt_undercityencounter_u4_0_0.vlm"},
	{0x60BC8CA6E8FF6364ULL, "vxlt_foundry_n_x8_0_0.vlm"},
	{0x60BC93B54AE7F88AULL, "vxht_chantry_2b7_0_0.vlm"},
	{0x60BCE5E17AF37471ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_aveline_arishokresponse.gad"},
	{0x60BD484F6F4EF8EEULL, "illuminationmask.vsh.fxo"},
	{0x60BE69659D383C2BULL, "vxht_brothel_n_6as_0_0.vlm"},
	{0x60BE9932C1AD6079ULL, "715134013.ogg"},
	{0x60BE9C82454A1E11ULL, "plt_ave000pt_midstage.nss"},
	{0x60BF31B579580623ULL, "6021110_m.fxe"},
	{0x60BF78443E19AF81ULL, "vxht_0_hightown_a3_a7m_0_0.vlm"},
	{0x60C034A86D86C60AULL, "vxdr_0_deeproads_cg2_0_0.vlm"},
	{0x60C03AE9C1D14700ULL, "hlac_lt_0_lowtown_n_131071_2y.dds"},
	{0x60C1B791E70E3137ULL, "vxlt_0_lowtown_ere_0_0.vlm"},
	{0x60C236D9812EDDE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_sigh_big.gad"},
	{0x60C23FA6D9EE2824ULL, "hlac_lt_0_lowtown_a2_-3_3z.dds"},
	{0x60C27363F2B6830FULL, "vxlt_0_lowtown_e1p_0_0.vlm"},
	{0x60C28D441D08F166ULL, "926978713.ogg"},
	{0x60C3386C267CD99CULL, "vxdr_primevaltomb_6nj_0_2.vlm"},
	{0x60C389661A0766E0ULL, "vxw_bonepit_21c_0_0.vlm"},
	{0x60C3D6B0AE7C4E5CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtc.phy"},
	{0x60C3E6E17AF9941FULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_aveline_arishokresponse.ani"},
	{0x60C41DE5020693CAULL, "6100059_f.fxe"},
	{0x60C4938B3D1775ECULL, "6100278_m.fxe"},
	{0x60C4FB0F64DA167EULL, "283437373.ogg"},
	{0x60C53662B1D41DA2ULL, "6020101_m.fxe"},
	{0x60C5523B147A2697ULL, "lowtown_fem_lowlife_mob6168979.cl"},
	{0x60C5CAE2406B2E4AULL, "art\\characters\\playercharacter\\humanfemale\\cth_beta.phy"},
	{0x60C5FDB52521086AULL, "isa321_velasco6091883.cl"},
	{0x60C70433E26E7156ULL, "cr_cycle.mao"},
	{0x60C75667B2293FD7ULL, "hlac_w_encounter2_camp_2_0y.dds"},
	{0x60C789B466B81A12ULL, "961127206.ogg"},
	{0x60C7FDA9E2715FAEULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firemed.dura.evt"},
	{0x60C853500BF63781ULL, "llac_lt_0_lowtown_docks_n_-65540_1.dds"},
	{0x60C859F821615A21ULL, "mcr370_cut_player_rules6098290.cl"},
	{0x60C87C703C9798C2ULL, "hlac_ht_keep_n_4519_6y.dds"},
	{0x60C89B06B99D219CULL, "llac_ca_caverns1_24424_12.dds"},
	{0x60C8F1D246DD7293ULL, "ser221_gascard6134523.cl"},
	{0x60C94658502D4E0AULL, "plt_dae270pt_gallows.plo"},
	{0x60C9F45723E3EF64ULL, "6078717_m.fxe"},
	{0x60CAD1CE48616F3EULL, "6152131_f.fxe"},
	{0x60CAF4A072137976ULL, "331887879.ogg"},
	{0x60CB178EE3EB3DC9ULL, "660990010.ogg"},
	{0x60CB41F5B725DA21ULL, "dae000_ineria6136231.cl"},
	{0x60CB4E8BE4E060A7ULL, "mrl150_merrill6117764.cl"},
	{0x60CB7FD5E6F1202CULL, "258179653.ogg"},
	{0x60CB9C589826FCEAULL, "vxlt_0_lowtown_docks_n_bgb_0_0.vlm"},
	{0x60CC8BE82311E4DAULL, "vxlt_0_lowtown_f4q_0_0.vlm"},
	{0x60CCDB428F2E54ABULL, "vxht_brothel_n_6jr_0_0.vlm"},
	{0x60CD0054CC17B76CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xa.msh"},
	{0x60CD0654CC17C15EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xa.mmh"},
	{0x60CD1254CC17D5F5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xa.mao"},
	{0x60CD6DDAFF2ABE5CULL, "691887975.ogg"},
	{0x60CDDB2A1644E80BULL, "vxht_mansion_n_4b7_0_0.vlm"},
	{0x60CDF6870BAE44E2ULL, "mrl150ip_altar.ncs"},
	{0x60CE3A7F7B401708ULL, "vxlt_0_lowtown_a2_n_gn7_0_0.vlm"},
	{0x60CE6934DF7156B1ULL, "w_encounter3_62_0.mmh"},
	{0x60CE7534DF716B52ULL, "w_encounter3_62_0.mao"},
	{0x60CE7D9493F82B65ULL, "vxlt_0_lowtown_a3_n_c_d2n_0_0.vlm"},
	{0x60CE8334DF71836FULL, "w_encounter3_62_0.msh"},
	{0x60CEF4A9E2776F59ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firemed.dura.gad"},
	{0x60CF07F9C08BF3BCULL, "1019526286.ogg"},
	{0x60CF9CA99EC6C1CDULL, "694204257.ogg"},
	{0x60CFA795DE2D4402ULL, "6165435_m.fxe"},
	{0x60CFDF2249A50443ULL, "6121070_m.fxe"},
	{0x60D02B7728031843ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.mmh"},
	{0x60D050C3702D962DULL, "416479163.ogg"},
	{0x60D0708F7FCAEE07ULL, "6170234_m.fxe"},
	{0x60D0A17D9C7BA45DULL, "fen310_varania6118580.cl"},
	{0x60D1D0974A93DF33ULL, "lgt321cr_templar.utc"},
	{0x60D2F960747F7004ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfalll_1.mao"},
	{0x60D349AAA19861A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_one_knee_loop.ani"},
	{0x60D3A83FA3CA3622ULL, "6027001_m.fxe"},
	{0x60D3E9FA93F1D03BULL, "35518582.ogg"},
	{0x60D44D0885C43E4CULL, "w_bonepitmining_a2_28_0.msh"},
	{0x60D4530885C4483EULL, "w_bonepitmining_a2_28_0.mmh"},
	{0x60D45F0885C45CD5ULL, "w_bonepitmining_a2_28_0.mao"},
	{0x60D57AB12A4A6A02ULL, "814865750.ogg"},
	{0x60D5F5E3548432F2ULL, "vxht_0_hightown_n_b86_0_0.vlm"},
	{0x60D5F746DD0BCE8BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_broken.phy"},
	{0x60D605A9E27DAAA7ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firemed.dura.ani"},
	{0x60D670B5DC0061A0ULL, "follower_banter6172757.cl"},
	{0x60D6AF353C118468ULL, "vxlt_gallowstemplar_5pc_0_0.vlm"},
	{0x60D6B0762E829780ULL, "war_flankimmune.dds"},
	{0x60D6C7657092370AULL, "gen00fl_anders6119140.cl"},
	{0x60D6ED0197776593ULL, "llac_lt_gallowscourtyard_-196608_0.dds"},
	{0x60D729C427DA9B67ULL, "6202057_f.fxe"},
	{0x60D76E0B89D87006ULL, "lgt331im_shw_mer_map.uti"},
	{0x60D7C56186A6BF08ULL, "vxdr_primevaltomb_4hy_0_1.vlm"},
	{0x60D8107F7195DBB7ULL, "vxca_cave_1ah_0_0.vlm"},
	{0x60D825E9CB087BD4ULL, "vxht_0_hightown_a2_boz_0_0.vlm"},
	{0x60D837E2D25CAD14ULL, "fxa_cen_red_c.dds"},
	{0x60D85FF4DBD42D7AULL, "art\\vfx\\textures\\vfx_wt_talent.dds"},
	{0x60D903BC24FDDE5AULL, "hlac_w_encounter1_131070_0y.dds"},
	{0x60D946AAA19CC806ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_one_knee_loop.gad"},
	{0x60D9477A11115D34ULL, "6046873_m.fxe"},
	{0x60DA1CBFFD78A720ULL, "hlac_lt_gallowscourtyard_131069_1y.dds"},
	{0x60DA9C148DA98E01ULL, "6208646_f.fxe"},
	{0x60DA9D39B9400A6EULL, "fxa_pois_calt.dds"},
	{0x60DAD81CAE39A960ULL, "36364631.ogg"},
	{0x60DADE75D53A0A31ULL, "482270734.ogg"},
	{0x60DB0297653B5365ULL, "one130_thug6131627.cl"},
	{0x60DB0704E04BAA32ULL, "67559448.ogg"},
	{0x60DB47507DD80531ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforwardleft.ani"},
	{0x60DB5413189B299FULL, "m_lgt331ar_hybris.bnk"},
	{0x60DBCD91E97B1428ULL, "6027865_m.fxe"},
	{0x60DCCC4F4569DD3DULL, "6119829_f.fxe"},
	{0x60DD3EC2B6EBE16EULL, "508765111.ogg"},
	{0x60DD5241C01BF59CULL, "vxht_mansion2_1t0_0_0.vlm"},
	{0x60DD58D58331A5B1ULL, "vxdr_primevaltomb_4xm_0_0.vlm"},
	{0x60DDA8F0C6910D90ULL, "321207869.ogg"},
	{0x60DDB754699CAF8EULL, "zz_approval_debug6003942.cl"},
	{0x60DDB9276199CC79ULL, "178351161.ogg"},
	{0x60DDCE26F3E75AEDULL, "370894965.ogg"},
	{0x60DE0070361C5493ULL, "vxca_cavern1_haunted_iog_0_0.vlm"},
	{0x60DE8B34560B4527ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf02.phy"},
	{0x60DF1096B6EBA348ULL, "target_creaturelock_pressed.cur"},
	{0x60DF5F7BA7A77124ULL, "fire_ball_red.mao"},
	{0x60E020AC030919DAULL, "art\\vfx\\combat\\vfx_g_spear_mod_c.impa.gad"},
	{0x60E020F6ABD8426EULL, "vxlt_0_lowtown_a2_gbv_0_0.vlm"},
	{0x60E037B670F18C66ULL, "6177010_m.fxe"},
	{0x60E0733228271351ULL, "6083981_m.fxe"},
	{0x60E077B536E5A356ULL, "vxht_mansion2_1wc_0_0.vlm"},
	{0x60E07CCE9803A7BFULL, "82681446.ogg"},
	{0x60E0D41B4E35936BULL, "115933315.ogg"},
	{0x60E0EC9797715036ULL, "170504953.ogg"},
	{0x60E10154B0404DE8ULL, "6044862_m.fxe"},
	{0x60E13407AE8CB805ULL, "10236348.ogg"},
	{0x60E159FF8E2C52E9ULL, "1053482011.ogg"},
	{0x60E1C79E8AF692D9ULL, "dae000_brothel_orana.bnk"},
	{0x60E1ECA5B31CF15BULL, "lgt110_orz_harrowmont6108984.cl"},
	{0x60E2320885D04DFDULL, "w_bonepitmining_a2_28_0.anb"},
	{0x60E252507DDE36D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforwardleft.gad"},
	{0x60E2AF62EC7BA651ULL, "291595776.ogg"},
	{0x60E2BE88E644255AULL, "6114794_m.fxe"},
	{0x60E2EB4EC652B541ULL, "one100dg_nabil6060505.cl"},
	{0x60E2F9D585B9C023ULL, "vxlt_undercityencounter_1og_0_0.vlm"},
	{0x60E317E683F6BE9BULL, "6039380_m.fxe"},
	{0x60E329CD98D28DBDULL, "vxca_caverns1_mines_j43_0_0.vlm"},
	{0x60E345C590927B76ULL, "6177277_f.fxe"},
	{0x60E382B85611713EULL, "6083101_f.fxe"},
	{0x60E3CF6537012C9EULL, "vxw_bonepitdestroyed_a3_23c_0_0.vlm"},
	{0x60E3E60618E94FEFULL, "6202332_f.fxe"},
	{0x60E40CE1B3078922ULL, "6149518_m.fxe"},
	{0x60E458D36F0220DFULL, "dae100cr_viscount.utc"},
	{0x60E47D4FED9E8786ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend_da.dds"},
	{0x60E4D51D466683E4ULL, "hlac_lt_0_lowtown_a2_-2_0x.dds"},
	{0x60E4E919FE6D47AEULL, "hlac_lt_warehouse_n_1_5z.dds"},
	{0x60E4EDFD6505CBB8ULL, "6043975_m.fxe"},
	{0x60E50F2332A73582ULL, "dae000_hate_follower_react6103253.cl"},
	{0x60E512EEA79C5C03ULL, "vxlt_0_lowtown_a2_n_ejf_0_0.vlm"},
	{0x60E52663FE16EEA5ULL, "llac_lt_gallowscourtyard_n_-131072_3.dds"},
	{0x60E5478B3600A8F5ULL, "art\\characters\\weapons\\shields\\w_shd_003a.msh"},
	{0x60E5498B3600AC6BULL, "art\\characters\\weapons\\shields\\w_shd_003a.mmh"},
	{0x60E54DF7DF9A7BD3ULL, "mcr310_loyalist_mages6076868.cl"},
	{0x60E5558B3600C008ULL, "art\\characters\\weapons\\shields\\w_shd_003a.mao"},
	{0x60E5B27CD447D81DULL, "gen00fl_varric6109205.cl"},
	{0x60E5E99E8AFAB539ULL, "dae000_brothel_orana.cnv"},
	{0x60E63B143361252CULL, "art\\levels\\races\\ferelden\\props\\_textures\\brazier_n.dds"},
	{0x60E64F9D5B41754BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medd_0s.dds"},
	{0x60E6A1F1FB3D1D84ULL, "hlac_lt_0_lowtown_a2_n_-1_1y.dds"},
	{0x60E6E0CA8F08F029ULL, "vxlt_gallowstemplar_n_59z_0_0.vlm"},
	{0x60E70D4C30BE6D0AULL, "mcr370_negotiation6080255.cl"},
	{0x60E723AC030F3DF1ULL, "art\\vfx\\combat\\vfx_g_spear_mod_c.impa.evt"},
	{0x60E72E39FC63837DULL, "dae000_bodhan6146902.cl"},
	{0x60E74122E4B22603ULL, "gen_im_flw_sta_merril_01.uti"},
	{0x60E7432794816673ULL, "plt_one000pt_herbalist.nss"},
	{0x60E958195DBAD603ULL, "6103210_m.fxe"},
	{0x60E99F7CCA97627EULL, "ser270st_aftermath.stg"},
	{0x60EA80A99374B11BULL, "vxlt_0_lowtown_n_e2b_0_0.vlm"},
	{0x60EAB285BC605DBEULL, "gen00fl_isabela_init6175664.cl"},
	{0x60EB2C6EF17FE8E8ULL, "vxlt_0_lowtown_a2_hg8_0_0.vlm"},
	{0x60EB50550C9162C5ULL, "lgt000ip_store_epic3_3.utp"},
	{0x60EB5C96591E4B91ULL, "mcr371_hostile_anders6082115.cl"},
	{0x60EC1F1145A3E167ULL, "restore.ncs"},
	{0x60EC2213EE1F1D8EULL, "vxht_mansion_n_4y7_0_0.vlm"},
	{0x60EC2278881AE922ULL, "vxlt_darktown_5q2_0_0.vlm"},
	{0x60EC3D241EB227FEULL, "464793180.ogg"},
	{0x60EC794165B4A67BULL, "hlac_ht_keep_n_3743_0x.dds"},
	{0x60ECD34FEB2FDAF4ULL, "93044468.ogg"},
	{0x60ED426F4FF3090EULL, "vxht_0_hightown_brk_0_0.vlm"},
	{0x60ED84174EB265F5ULL, "llac_dr_0_deeproads_5842_8.dds"},
	{0x60ED9C7D396F1A77ULL, "w_bonepitdestroyed_a3_58_0.msh"},
	{0x60EDAE7D396F38DAULL, "w_bonepitdestroyed_a3_58_0.mao"},
	{0x60EDB27D396F3FE9ULL, "w_bonepitdestroyed_a3_58_0.mmh"},
	{0x60EDD0EB95D65E1BULL, "193481225.ogg"},
	{0x60EE21C210BE31DEULL, "6118022_m.fxe"},
	{0x60EE7F21CC303C92ULL, "6165211_m.fxe"},
	{0x60EE8ABFEF0F36E5ULL, "ht_0_hightown_a2_53_0_lc.mmh"},
	{0x60EEA709B876BEB3ULL, "move_indicator.mao"},
	{0x60EEAB954BC67C20ULL, "1065098555.ogg"},
	{0x60EEB68B903CE700ULL, "gen_im_fla_fenris_01.uti"},
	{0x60EEB8FD1C29F1A0ULL, "m_dae270ar_gallows.bnk"},
	{0x60EEF69CD9750119ULL, "art\\characters\\playercharacter\\humanfemale\\boo_wrba.phy"},
	{0x60EF0C28A06565AFULL, "zz_followers_banter6158628.cl"},
	{0x60EFFBF2D54CEFF2ULL, "vxdr_primevalentrance_316_0_0.vlm"},
	{0x60F018D4AFA62D1DULL, "6049437_m.fxe"},
	{0x60F0B102A6218367ULL, "6161023_m.fxe"},
	{0x60F0EC9FA8C7FEE4ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm4.ani"},
	{0x60F10B6248A76184ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_roba_s.dds"},
	{0x60F11983B5D96F25ULL, "455177099.ogg"},
	{0x60F188635C62D633ULL, "zz_followers_banter6157584.cl"},
	{0x60F1EC72D72FC8E8ULL, "rdr111cr_harbormaster.utc"},
	{0x60F26941CEA37CB7ULL, "ser221_party_comments6163604.cl"},
	{0x60F287327156110DULL, "lt_0_lowtown_a3_n_c_27_0_ll.mmh"},
	{0x60F28C4D4E1D7B58ULL, "vxca_caverns1_mines_gh3_0_0.vlm"},
	{0x60F2F0CE5593B4E8ULL, "551269358.ogg"},
	{0x60F384CB99927483ULL, "vxlt_0_lowtown_a2_n_eoi_0_0.vlm"},
	{0x60F3B6DF95341713ULL, "vxca_caverns2prc_9pq_0_0.vlm"},
	{0x60F3CEF8539852E5ULL, "dae000_arishok6067679.cl"},
	{0x60F43BAC0319F0C0ULL, "art\\vfx\\combat\\vfx_g_spear_mod_c.impa.ani"},
	{0x60F4E6205BCA3943ULL, "6134986_m.fxe"},
	{0x60F5795B646F2C08ULL, "graphicsdebugger.gfx"},
	{0x60F59352C7F6D5FAULL, "mag110_samson6018718.cl"},
	{0x60F63E599E7BC714ULL, "hlac_w_bonepitdestroyed_a3_262148_0y.dds"},
	{0x60F69B621EE85629ULL, "dae021_keep_guard_2.utc"},
	{0x60F718B28C1F666AULL, "mag320_orsino6071180.cl"},
	{0x60F732733ED746AEULL, "venge_vengeance.dds"},
	{0x60F7E99FA8CE1846ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm4.gad"},
	{0x60F86622BFE4D12FULL, "433819353.ogg"},
	{0x60F93E74B56EB673ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_big_p.impa.ani"},
	{0x60F94F5850568101ULL, "plt_dae270pt_gallows.nss"},
	{0x60F9CD672826FA4FULL, "vxca_caverns1_hkk_0_0.vlm"},
	{0x60F9D23FA440A129ULL, "character_character_p125.psh.fxo.dx11.assembly"},
	{0x60F9F99DD6BE4F94ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_c.mmh"},
	{0x60FA30B756E38605ULL, "6179984_f.fxe"},
	{0x60FA35FBA408269CULL, "vxht_mansion2_1b6_0_0.vlm"},
	{0x60FA91F9326E91ACULL, "vxdr_0_deeproads_eri_0_0.vlm"},
	{0x60FAD39096B5BCEBULL, "ave310_donnic6134764.cl"},
	{0x60FAE42A1A088593ULL, "vxca_cavern1_haunted_iai_0_0.vlm"},
	{0x60FAEB5270E34923ULL, "1034779363.ogg"},
	{0x60FB76C220C0AC5CULL, "dog_slaver.dds"},
	{0x60FBB382457DF3FAULL, "plt_ave000pt_midstage.plo"},
	{0x60FBB9CC96328D0FULL, "mrl210_merrill6102147.cl"},
	{0x60FC6C1C9D4BA200ULL, "hlac_lt_undercitydungeon_-1_32x.dds"},
	{0x60FCA6A4A3FE6784ULL, "zz_prmpt_debug6198219.cl"},
	{0x60FD283E40402752ULL, "643915792.ogg"},
	{0x60FD79E83AF498BAULL, "vxlt_gallowscourtyard_3vs_0_0.vlm"},
	{0x60FE425D1CB2789BULL, "6164406_m.fxe"},
	{0x60FE740227999858ULL, "749420031.ogg"},
	{0x60FEE5C214C9A169ULL, "vxlt_0_lowtown_a2_gc2_0_0.vlm"},
	{0x60FF89566F238507ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_bda_0e.dds"},
	{0x610058A47FA6E4ABULL, "gen_im_trs_tmp_hvy_hlm_r2.uti"},
	{0x6100A83530ACEAF4ULL, "6102815_f.fxe"},
	{0x6100D116C0E79117ULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statueb.msh"},
	{0x6100E316C0E7B07AULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statueb.mao"},
	{0x6100E33A53DB6960ULL, "522856521.ogg"},
	{0x6100E716C0E7B689ULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statueb.mmh"},
	{0x61012BA09F44614FULL, "vxlt_gallowscourtyard_7ff_0_0.vlm"},
	{0x6102821F71080D77ULL, "6142166_f.fxe"},
	{0x61028A6AC4A4E55DULL, "gen00fl_merrill6127630.cl"},
	{0x6102F2729BEC2EB2ULL, "lowtown_fem_prostitute_sta6152323.cl"},
	{0x610362C660C12097ULL, "6076863_f.fxe"},
	{0x610370294A17C6BBULL, "513994578.ogg"},
	{0x6103AA4C0233D989ULL, "dae150_radha6148924.cl"},
	{0x61044A23B99CCAECULL, "vxca_caverns1_gi9_0_0.vlm"},
	{0x61045DA7FBAB3CC8ULL, "742935338.ogg"},
	{0x610496C1EDD30573ULL, "5252609.ogg"},
	{0x61053DCFC9041D97ULL, "vxht_0_hightown_a3_ceo_0_0.vlm"},
	{0x610597A1801669B6ULL, "vxw_bonepitdestroyed_a3_277_0_0.vlm"},
	{0x61063674B57932AAULL, "art\\vfx\\cinematic\\vfx_lyrium_c_big_p.impa.evt"},
	{0x610672F08A5134B4ULL, "6209859_f.fxe"},
	{0x6106AEFF750E0A75ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgtp_0n.dds"},
	{0x61073A1318C038D5ULL, "m_lgt331ar_hybris.wbd"},
	{0x6107ECAD79A59753ULL, "6189678_m.fxe"},
	{0x6107FD5252428F3CULL, "vxca_caverns1_mines_jfl_0_0.vlm"},
	{0x61081294B1D9B84EULL, "vxca_caverns1_mines_gki_0_1.vlm"},
	{0x61085E7BDA8C5F59ULL, "vxw_bonepitdestroyed_a3_lm_0_0.vlm"},
	{0x61089EF088084365ULL, "vxca_cavern1_haunted_dq6_0_0.vlm"},
	{0x610925A2F1C52DD5ULL, "16093598.ogg"},
	{0x6109D8CC47F5D7FDULL, "vxw_bonepitmining_a2_1l0_0_0.vlm"},
	{0x610A262933DF3A0EULL, "vxht_0_hightown_bxm_0_0.vlm"},
	{0x610AA2F2DD54212DULL, "mag110_samson6063964.cl"},
	{0x610AA9A5FA8E36C0ULL, "vxw_encounter2_camp_48_0_0.vlm"},
	{0x610CC287DC09FCB4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masd_0t.dds"},
	{0x610CDD54CC4D5FF9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xa.phy"},
	{0x610CF8C26E9BB78BULL, "vxlt_gallowstemplar_n_6dh_0_0.vlm"},
	{0x610D1574B57F191DULL, "art\\vfx\\cinematic\\vfx_lyrium_c_big_p.impa.gad"},
	{0x610D1E74D5B9BBDEULL, "465835548.ogg"},
	{0x610D43239214C8B9ULL, "573171614.ogg"},
	{0x610D579F06BFB9E9ULL, "vxlt_0_lowtown_a2_gf1_0_0.vlm"},
	{0x610D72B358E2F69AULL, "gen00fl_sebastian6179983.cl"},
	{0x610D763E97F37519ULL, "831227440.ogg"},
	{0x610DA0C7B27AD3CEULL, "709814931.ogg"},
	{0x610DA4425BD0B598ULL, "496267787.ogg"},
	{0x610DE18910F5F544ULL, "art\\characters\\creatures\\mubari\\animation\\dieloop_v1.gad"},
	{0x610E327DBC37772DULL, "qun160_vashothlead6061104.cl"},
	{0x610E3B927B3AAA23ULL, "dae300_cullen.bnk"},
	{0x610EAB4D0EB1D4F8ULL, "llac_lt_0_lowtown_a2_n_c_-3_4.dds"},
	{0x610EC56ECAE9DAB4ULL, "6182294_m.fxe"},
	{0x610EEFA111004654ULL, "vxlt_0_lowtown_a2_hcx_0_0.vlm"},
	{0x610F6654C00FC433ULL, "zz_followers_banter6158772.cl"},
	{0x6110257A30C7BBC0ULL, "299768195.ogg"},
	{0x61109A9979204813ULL, "vxca_caverns2prc_9zs_0_0.vlm"},
	{0x6110F925B8DB63CAULL, "6065883_m.fxe"},
	{0x6110FF93573A7ACFULL, "6120042_m.fxe"},
	{0x6111014C3F73EBA5ULL, "807658742.ogg"},
	{0x6111040E7A71DE85ULL, "6197868_m.fxe"},
	{0x611177EDB951067EULL, "vxht_chantry_n_2tg_0_1.vlm"},
	{0x6111839181194C4EULL, "llac_ht_chantry_2534_49.dds"},
	{0x6111A316D8164956ULL, "clf260_harley6069673.cl"},
	{0x61121D927B3E6053ULL, "dae300_cullen.cnv"},
	{0x611299075726430AULL, "6036740_m.fxe"},
	{0x6112D4AB135C818EULL, "vxlt_gallowstemplar_5qb_0_0.vlm"},
	{0x61130F8F96571C0EULL, "lgt220_shp_mag_guard6149515.cl"},
	{0x611312F3EE0E0C19ULL, "848161506.ogg"},
	{0x6113E636E0671941ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sis_flemeth_fake.ani"},
	{0x6113F08910FA7A22ULL, "art\\characters\\creatures\\mubari\\animation\\dieloop_v1.ani"},
	{0x6113FBFCA1DFEE49ULL, "gen00fl_isabela6080447.cl"},
	{0x6114097E720C8236ULL, "vxw_bonepitdestroyed_a3_228_0_0.vlm"},
	{0x61141E69E94FBBFEULL, "426169732.ogg"},
	{0x6114D969C0BD367CULL, "w_encounter3_15_0.mao"},
	{0x6114E569C0BD4A1FULL, "w_encounter3_15_0.mmh"},
	{0x6114EB69C0BD54C1ULL, "w_encounter3_15_0.msh"},
	{0x611506191E1A1767ULL, "atl_fonttest_dxt5_dat.xml"},
	{0x61161EC0BD4D0F5EULL, "6147710_m.fxe"},
	{0x6116AA543F251001ULL, "dae000_bodhan6129145.cl"},
	{0x6116EAC3D962F2F4ULL, "35587707.ogg"},
	{0x61172CCC699DD02AULL, "ser191_emeric6107876.cl"},
	{0x61175BCE2243B2C2ULL, "6115784_m.fxe"},
	{0x6117EAF914408CE5ULL, "6177919_f.fxe"},
	{0x61183890FCF206D3ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.cess.evt"},
	{0x611921CB42252279ULL, "mcr371_emile_de_launcet6086949.cl"},
	{0x611948DC38E628C2ULL, "vxca_caverns2_8l0_0_0.vlm"},
	{0x6119773F26EA420BULL, "vxlt_shop_fq_0_0.vlm"},
	{0x611A156C4861CA86ULL, "dae200cr_viscount6087239.cl"},
	{0x611A2E22C680D7C2ULL, "hlac_lt_0_lowtown_a2_65536_8z.dds"},
	{0x611A3F28FEBE0556ULL, "vxht_mansion2_1kg_0_0.vlm"},
	{0x611A65BF334E68BFULL, "gen_im_cns_cft_arn_01.uti"},
	{0x611AF0E3762F0944ULL, "19574062.ogg"},
	{0x611AF136E06D4AE3ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_sis_flemeth_fake.gad"},
	{0x611B0E8C5BC7F39AULL, "6019120_f.fxe"},
	{0x611B3308860014F5ULL, "w_bonepitmining_a2_28_0.rml"},
	{0x611B54D01D098666ULL, "vxlt_0_lowtown_a2_fhu_0_0.vlm"},
	{0x611BEB8CCCAB50B7ULL, "zz_followers_banter6159252.cl"},
	{0x611C008603F23694ULL, "vxlt_0_lowtown_a2_n_c_hjb_0_0.vlm"},
	{0x611C36A99527C167ULL, "vxw_bonepit_lc_0_0.vlm"},
	{0x611C658BE46B413DULL, "vxca_caverns1_iyn_0_0.vlm"},
	{0x611CDD08122E752FULL, "art\\vfx\\combat\\vfx_w_modal_act_c.mmh"},
	{0x611CE69109658653ULL, "hlac_lt_0_lowtown_docks_a3_-131074_8y.dds"},
	{0x611D05AEBDCC513DULL, "468540286.ogg"},
	{0x611D1798D4E3F653ULL, "vxlt_hangedman_2xb_0_0.vlm"},
	{0x611D20E658A966DFULL, "vxdr_0_deeproads_dsd_0_0.vlm"},
	{0x611D617E02F65ED4ULL, "vxlt_0_lowtown_a2_hhy_0_0.vlm"},
	{0x611D857E1392E063ULL, "6135124_f.fxe"},
	{0x611DC7D241AB4230ULL, "w_woundedcoast_221_0.msh"},
	{0x611DD5D241AB5A4AULL, "w_woundedcoast_221_0.mmh"},
	{0x611DD9D241AB6081ULL, "w_woundedcoast_221_0.mao"},
	{0x611E2812BD2B0B23ULL, "hlac_ht_0_hightown_a2_65536_4y.dds"},
	{0x611ECEB56F7A22E6ULL, "vxht_keep_70d_0_0.vlm"},
	{0x611F3190FCF81914ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.cess.gad"},
	{0x611FA37924D3DB72ULL, "gen00fl_dog6165722.cl"},
	{0x612095E605FD7CB4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexint_n.dds"},
	{0x6120AA43428300F8ULL, "llac_lt_0_lowtown_n_-131072_5.dds"},
	{0x6120DE57395BC6F5ULL, "vxht_0_hightown_n_c17_0_0.vlm"},
	{0x612100ACE6A2561FULL, "129807228.ogg"},
	{0x61218590C010B72AULL, "771493423.ogg"},
	{0x6121E599399A3840ULL, "vxca_caverns1_j69_0_0.vlm"},
	{0x6122FC9DD6E15A4BULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_c.anb"},
	{0x612314B8D3A522FEULL, "162278208.ogg"},
	{0x612333E1F7C56DA5ULL, "zz_and_debug6134052.cl"},
	{0x61242AD5160EEF9BULL, "vxht_keep_n_7a1_0_0.vlm"},
	{0x61242D33548B61A2ULL, "vxw_woundedcoast_16x_0_0.vlm"},
	{0x6124418E831D595CULL, "qun231_varnell6078114.cl"},
	{0x61255A653FCBAF66ULL, "90944624.ogg"},
	{0x612582AA6BF47465ULL, "vxca_cavern1_haunted_do4_0_0.vlm"},
	{0x6125AC378F171422ULL, "mcr371st_orsino_3.stg"},
	{0x61264090FCFE51F2ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.cess.ani"},
	{0x612641E76BC97566ULL, "dae000cr_grand_cleric6148123.cl"},
	{0x6126C207FDF159E6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_die_v1.evt"},
	{0x612793F18DF186CDULL, "vxht_bartrand_n_4a3_0_0.vlm"},
	{0x61280CB1B2DE73FBULL, "955091396.ogg"},
	{0x612819F1B755049AULL, "llac_w_sundermount_-393216_0.dds"},
	{0x6128BFB63243FD49ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinpoint_strike.ani"},
	{0x6128D297B3B54B13ULL, "68521698.ogg"},
	{0x612913AD0BB2FE03ULL, "vxca_caverns1_mines_i6p_0_0.vlm"},
	{0x61296E553E0163F3ULL, "920338730.ogg"},
	{0x6129B7F3F26A1FEDULL, "6189041_m.fxe"},
	{0x6129D2F4F1CB37BFULL, "fx_aoe_radius.mao"},
	{0x612AB04F0344C08FULL, "431601565.ogg"},
	{0x612B33678DD1F22DULL, "vxca_cavern1_haunted_j24_0_1.vlm"},
	{0x612B486484DEA757ULL, "vxca_caverns2_6sz_0_0.vlm"},
	{0x612B7B23662550C3ULL, "6033971_m.fxe"},
	{0x612B9FA04A7A084FULL, "hlac_ht_0_hightown_n_65534_0x.dds"},
	{0x612BC1CF024336EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_tilt_confused.evt"},
	{0x612BC3B1A6C3B502ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female.gad"},
	{0x612C0AC01F277D08ULL, "lt_0_lowtown_docks_32_0.anb"},
	{0x612C29D15121340BULL, "lt_0_lowtown_a3_n_c_36_0_ll.mmh"},
	{0x612C508405AEF69AULL, "vxht_mansion2_3dd_0_0.vlm"},
	{0x612C954E0C965A9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hold_object_front.ani"},
	{0x612CF7E1A648635EULL, "vxlt_backalley_g_1i0_0_0.vlm"},
	{0x612D5D4D7EDBD5C3ULL, "lt_0_lowtown_docks_a3_97_0.msh"},
	{0x612D6B4D7EDBED1DULL, "lt_0_lowtown_docks_a3_97_0.mmh"},
	{0x612D6F4D7EDBF42EULL, "lt_0_lowtown_docks_a3_97_0.mao"},
	{0x612DD907FDF79F31ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_die_v1.gad"},
	{0x612DEBE75A26572BULL, "163195929.ogg"},
	{0x612E13BBB55C32DAULL, "786471300.ogg"},
	{0x612E344B58D9E70CULL, "896377778.ogg"},
	{0x612E4BEBE4FCDDF3ULL, "6179790_m.fxe"},
	{0x612E84F82EB42BA0ULL, "235750965.ogg"},
	{0x612EA20C4BD05E7FULL, "vxht_0_hightown_n_b5l_0_0.vlm"},
	{0x612EDAB6324896CBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinpoint_strike.gad"},
	{0x612EE69263B69689ULL, "vxlt_0_lowtown_docks_a3_n_cyq_0_0.vlm"},
	{0x612F10E3EC702A4AULL, "vxlt_0_lowtown_a2_e33_0_0.vlm"},
	{0x612F2BE38C7E00EFULL, "qun111_ketojan6040715.cl"},
	{0x612F328716955880ULL, "zz_dog_debug6127896.cl"},
	{0x613006411CEB4179ULL, "6170711_m.fxe"},
	{0x613012FC6327A9A6ULL, "6021116_f.fxe"},
	{0x61303B2E79CFBC80ULL, "497994819.ogg"},
	{0x61312FB9EB1E9335ULL, "6014677_m.fxe"},
	{0x6131B8CF02479299ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_tilt_confused.gad"},
	{0x61329F1046817A22ULL, "vxw_woundedcoast_11y_0_0.vlm"},
	{0x6132E6B1A6CA0FF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female.evt"},
	{0x61330ECAF21CE44EULL, "ser270_templar6122509.cl"},
	{0x613321515BB603D4ULL, "295511416.ogg"},
	{0x61333AD03A7F1635ULL, "vxlt_undercityencounter_1c4_0_1.vlm"},
	{0x613345E495D007F3ULL, "6143078_m.fxe"},
	{0x61340739A98369DAULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_act_c.impa.gad"},
	{0x6134DA07FDFDBFDFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_die_v1.ani"},
	{0x613551FB97644DF9ULL, "hlac_ht_0_hightown_n_65533_2y.dds"},
	{0x6135D7B6324EB030ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinpoint_strike.evt"},
	{0x613602F42C4AC2B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.ani"},
	{0x613605CE7FE79F4DULL, "vxht_bartrand_n_4f2_0_0.vlm"},
	{0x61360DCA81272687ULL, "vxht_mansion_n_34j_0_0.vlm"},
	{0x61363C466856B810ULL, "gen00fl_sebastian6195882.cl"},
	{0x6136E3B1DBA1FCE2ULL, "art\\characters\\playercharacter\\humanmale\\cth_noba.msh"},
	{0x6136E5B1DBA1FF98ULL, "art\\characters\\playercharacter\\humanmale\\cth_noba.mmh"},
	{0x6136F1B1DBA2143BULL, "art\\characters\\playercharacter\\humanmale\\cth_noba.mao"},
	{0x6137704F82D4470CULL, "16554225.ogg"},
	{0x613788385F0E4E5BULL, "248496640.ogg"},
	{0x6137D1E825A46227ULL, "prop1_dy.psh.fxo"},
	{0x6137DF79A76CE4A0ULL, "678429213.ogg"},
	{0x6137E0BC98C7E05CULL, "444225362.ogg"},
	{0x6138008AE6F67E96ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp13_galowint.msh"},
	{0x61380A8AE6F68F94ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp13_galowint.mmh"},
	{0x61380E8AE6F696A7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp13_galowint.mao"},
	{0x613820D6DD6C4DE2ULL, "follower_banter6151338.cl"},
	{0x6138C9CF024DCEE7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_tilt_confused.ani"},
	{0x61394549175C0B38ULL, "fxe_waterfall.mao"},
	{0x61394945C857723DULL, "vxca_caverns1_igl_0_0.vlm"},
	{0x61395407DB98EF52ULL, "art\\vfx\\materials\\vfx_d_sphere.mao"},
	{0x6139868A5815388CULL, "mag291_vincento6027380.cl"},
	{0x61399D4E0CA0F292ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hold_object_front.evt"},
	{0x6139A6D6C9F8B1A9ULL, "vxlt_hangedman_n_5l_0_0.vlm"},
	{0x613A58E49B1AA107ULL, "t3_com_c01.tnt"},
	{0x613AC3397737F08BULL, "gen_im_arm_glv_hvy_set_high.uti"},
	{0x613B0A39A9898DF1ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_act_c.impa.evt"},
	{0x613B3A5C92CE9F92ULL, "vxw_bonepit_1q2_0_0.vlm"},
	{0x613B786736CAEB8BULL, "vxlt_gallowscourtyard_76z_0_0.vlm"},
	{0x613C4A447EE94F51ULL, "plt_ser000pt_debug.plo"},
	{0x613C5DA7BDA31611ULL, "vxdr_primevaltomb_6m0_0_0.vlm"},
	{0x613CD775E4707456ULL, "6170482_m.fxe"},
	{0x613CE34DFD75FCCCULL, "622729424.ogg"},
	{0x613CFFF42C50DC16ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.gad"},
	{0x613D132E4233A670ULL, "mag131_keran6086284.cl"},
	{0x613D49EF0F816852ULL, "vxht_0_hightown_a2_brk_0_0.vlm"},
	{0x613D8CEBDAADDF32ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2.gad"},
	{0x613DD201D265C97EULL, "gen00fl_isabela6075275.cl"},
	{0x613E4A9D77A5223DULL, "art\\vfx\\materials\\vfx_link_smo_dark.mao"},
	{0x613EF94D98645206ULL, "6095403_f.fxe"},
	{0x613FEA3BE1DD8AC7ULL, "cod_ltr_sk_shipment.utp"},
	{0x6140492C54C1EA33ULL, "llac_lt_0_lowtown_a2_n_c_65535_3.dds"},
	{0x61405A5CE9D53135ULL, "gencr_coterie_thug_2w.utc"},
	{0x61405F4B3994E1F1ULL, "169101861.ogg"},
	{0x6140AB5910644946ULL, "gen00fl_merrill6141136.cl"},
	{0x6140BC4E0CA74625ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hold_object_front.gad"},
	{0x6140ECB1DBAA6023ULL, "art\\characters\\playercharacter\\humanmale\\cth_noba.phy"},
	{0x6140EEB1A6D65AE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female.ani"},
	{0x61414A09A2A66DE5ULL, "170872211.ogg"},
	{0x6141DEFAE05A2E79ULL, "6149258_m.fxe"},
	{0x61420B4F4401E77BULL, "one280ip_locked_door.utp"},
	{0x61441AF42C572975ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_exit.evt"},
	{0x6144622E25AD3325ULL, "vxht_keep_n_6lp_0_1.vlm"},
	{0x61448FEBDAB40369ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2.evt"},
	{0x6144A413DD143037ULL, "hlac_lt_warehouse_n_65537_0y.dds"},
	{0x6144D5591F251368ULL, "vxht_0_hightown_a2_b90_0_0.vlm"},
	{0x6145297AFD3028BBULL, "6036918_m.fxe"},
	{0x6145B56B8EAE7D44ULL, "6201468_f.fxe"},
	{0x6145DE0812517C6CULL, "art\\vfx\\combat\\vfx_w_modal_act_c.anb"},
	{0x61463E136DB91E4AULL, "vxht_0_hightown_n_bu5_0_0.vlm"},
	{0x6146A7DD77221850ULL, "vxlt_0_lowtown_docks_a3_cz2_0_0.vlm"},
	{0x6146A95F8EE9FAA3ULL, "vxlt_0_lowtown_docks_a3_n_aoc_0_0.vlm"},
	{0x6146E3012E48150CULL, "vxca_caverns1_mines_iiz_0_0.vlm"},
	{0x61471BCAF42128CAULL, "vxlt_gallowstemplar_n_454_0_0.vlm"},
	{0x61474C7C9BD8BF58ULL, "284106270.ogg"},
	{0x614777285DC956F6ULL, "dae000ip_crowd_lm5_talk3.utp"},
	{0x61479190D043F092ULL, "character_character_p250.psh.fxo.dx11.assembly"},
	{0x6147BB22BA4D8ABDULL, "vxca_caverns1_ijk_0_0.vlm"},
	{0x61482239A99440C0ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_act_c.impa.ani"},
	{0x61483C6BD3D52BD1ULL, "fen110_fenris6118145.cl"},
	{0x614849A94B9FB34CULL, "vxlt_0_lowtown_docks_9zs_0_0.vlm"},
	{0x61487E00DDD987F0ULL, "6043284_f.fxe"},
	{0x6149000A063B4F4AULL, "llac_w_encounter3_-65533_0.dds"},
	{0x61492B929FA36FC8ULL, "llac_ca_cave_2194_10.dds"},
	{0x614947C742A459D7ULL, "dae000_friend_turn_in6179829.cl"},
	{0x614979AC2E1A6981ULL, "6126521_m.fxe"},
	{0x6149AC3984C4B812ULL, "vxw_bonepit_1z3_0_0.vlm"},
	{0x6149B5B779F8BAE8ULL, "6001344_m.fxe"},
	{0x614A4AFCD41BD2C9ULL, "9546445.ogg"},
	{0x614A53A77AEF7D0CULL, "zz_one_debug6051513.cl"},
	{0x614ACF84AF992E91ULL, "vxdr_primevaltomb_6p7_0_0.vlm"},
	{0x614ADE3200DF8F6DULL, "hlac_lt_gallowstemplar_f_2383_4x.dds"},
	{0x614B1CB8BE0E8E17ULL, "6091690_m.fxe"},
	{0x614B4054EEF95FCDULL, "gen00fl_sebastian.bnk"},
	{0x614B47BA726D4731ULL, "6169905_f.fxe"},
	{0x614B75667EEE07D0ULL, "34350443.ogg"},
	{0x614BA53ADF99BA3CULL, "docks_male_human_worker_sta6175940.cl"},
	{0x614BB3E90403F77BULL, "hlac_w_bonepit_65542_0y.dds"},
	{0x614BE7EFCC494D25ULL, "920364733.ogg"},
	{0x614C0294C881D232ULL, "601318478.ogg"},
	{0x614C2FD471A530BDULL, "196967175.ogg"},
	{0x614C58A44C9B4581ULL, "sky_sky2_p0.psh.fxo.dx11"},
	{0x614C98A2B4D1B600ULL, "gen00fl_aveline6132713.cl"},
	{0x614CE4B4E43D0F3AULL, "vxw_bonepitdestroyed_a3_2a9_0_0.vlm"},
	{0x614D3E2794D810DCULL, "plt_one000pt_herbalist.plo"},
	{0x614D5D16539199B7ULL, "6121456_m.fxe"},
	{0x614D748AD92F06CFULL, "vxht_keep_n_4hm_0_0.vlm"},
	{0x614DB8A39D482DD5ULL, "863573553.ogg"},
	{0x614DC92C8B1658BAULL, "441709401.ogg"},
	{0x614E30C0A1CB8E42ULL, "vxht_keep_36b_0_0.vlm"},
	{0x614E360DB1CF4919ULL, "vxlt_gallowstemplar_n_4sm_0_0.vlm"},
	{0x614E6254EEFBCF35ULL, "gen00fl_sebastian.cnv"},
	{0x614F5537E1EA1710ULL, "6153199_m.fxe"},
	{0x614F910464F4F15FULL, "hlac_ht_keep_n_4519_13z.dds"},
	{0x615079BF5D380986ULL, "vxlt_hangedman_n_2o1_0_0.vlm"},
	{0x61507D67497FBFEEULL, "6124700_m.fxe"},
	{0x61511932A547DAAAULL, "vxca_cave_2d1_0_0.vlm"},
	{0x615130B046D1B308ULL, "6108735_m.fxe"},
	{0x615165E17C06F29EULL, "6075998_m.fxe"},
	{0x6151701C82082C3FULL, "619087132.ogg"},
	{0x615177EBDABE64D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim2.ani"},
	{0x6151949FE68E7B9BULL, "1051091734.ogg"},
	{0x6151F9C315911358ULL, "235085698.ogg"},
	{0x615250D045053865ULL, "llac_w_woundedcoast_-65536_0.dds"},
	{0x615253D906115587ULL, "vxlt_0_lowtown_a3_n_c_fhx_0_0.vlm"},
	{0x61528989FF8A6E3EULL, "ser270_emeric6113378.cl"},
	{0x6152B6B1D1AF0985ULL, "6177174_m.fxe"},
	{0x6152F1733C585D6AULL, "vxlt_gallowstemplar_f_3e3_0_0.vlm"},
	{0x6153354F20A08B4AULL, "6059659_m.fxe"},
	{0x6153834B6684404FULL, "vxw_bonepitdestroyed_a3_1xd_0_0.vlm"},
	{0x615448B28E6E9D09ULL, "6060120_m.fxe"},
	{0x6154661D0945F3B0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblelarge.mao"},
	{0x61546A1D0945FAC3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblelarge.mmh"},
	{0x6154781D0946121DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblelarge.msh"},
	{0x615485B20D1F6D9EULL, "6026983_m.fxe"},
	{0x6154BC7C688C9710ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_die_c.mmh"},
	{0x6155665AA7E5C59DULL, "6016312_m.fxe"},
	{0x61556C6E0BEC2529ULL, "229578017.ogg"},
	{0x61556CA686BE5BC2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb04.phy"},
	{0x6155C7F9CBCA9894ULL, "vxlt_0_lowtown_a2_n_c_hvf_0_0.vlm"},
	{0x6155DF57588981D2ULL, "hightown_fem_elfservant_mob6169041.cl"},
	{0x615602362AC6CE69ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomea.mao"},
	{0x61560E362AC6E202ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomea.mmh"},
	{0x615610362AC6E5B8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomea.msh"},
	{0x61561BF2F806901CULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_008a_0s.dds"},
	{0x61562CFFAC43A33DULL, "vxca_caverns1_imj_0_0.vlm"},
	{0x6156978CFA875458ULL, "zz_mag_debug6039840.cl"},
	{0x6156BC43936E303DULL, "6153477_m.fxe"},
	{0x6156CBF73299403CULL, "vxlt_gallowstemplar_n_4o0_0_0.vlm"},
	{0x615719F8BE8F33FFULL, "vxdr_primevalentrance_39o_0_0.vlm"},
	{0x6157269A2CCE4C30ULL, "vxlt_0_lowtown_docks_brh_0_0.vlm"},
	{0x615792F32E02643EULL, "6104885_m.fxe"},
	{0x6158198AC2B99642ULL, "hlac_lt_0_lowtown_n_0_6z.dds"},
	{0x61582FA01F3496DEULL, "6045881_m.fxe"},
	{0x61584D787C2AACDEULL, "vxlt_backalley_g_1x3_0_0.vlm"},
	{0x6158967194AD1C82ULL, "vxht_mansion_n_6rh_0_0.vlm"},
	{0x6158D54EE897C739ULL, "hlac_lt_0_lowtown_-262144_3y.dds"},
	{0x6158F7B603758E1CULL, "plt_cod_prc_pickled_apples.plo"},
	{0x61594161A18F4711ULL, "vxdr_primevaltomb_6w6_0_0.vlm"},
	{0x615954D508BC0501ULL, "6172735_m.fxe"},
	{0x615961405CCB5F94ULL, "427984547.ogg"},
	{0x61597FE15CEFBFABULL, "6133024_m.fxe"},
	{0x615982DCCF2270A6ULL, "hlac_lt_warehouse_1_3y.dds"},
	{0x6159AB92BC4C9459ULL, "hlac_w_bonepitmining_a2_196612_0x.dds"},
	{0x6159B728ADEA5C62ULL, "783992244.ogg"},
	{0x6159E4C8A1F2AADAULL, "6154732_m.fxe"},
	{0x6159FF6418CC1AF5ULL, "731220297.ogg"},
	{0x615A1EA2A02CD1D5ULL, "6214068_m.fxe"},
	{0x615A20BCA913C936ULL, "vxdr_primevaltomb_918_0_0.vlm"},
	{0x615A5409EA564DE5ULL, "452868222.ogg"},
	{0x615AF59A453295F2ULL, "vfxchunky0_cs.psh.fxo"},
	{0x615B0594DDF7A29CULL, "vxht_bartrand_n_5su_0_0.vlm"},
	{0x615B3CB609B3A76FULL, "vxdr_slavers_1b8_0_0.vlm"},
	{0x615B930137ADE0EEULL, "vxlt_gallowsprison_n_6li_0_0.vlm"},
	{0x615BB088218A38BFULL, "hlac_ht_keep_n_3736_3y.dds"},
	{0x615BBD6FAA02E369ULL, "hlac_ht_0_hightown_n_65536_4x.dds"},
	{0x615BE5859C61537DULL, "docks_scene_dockworkers6176492.cl"},
	{0x615C0D36C004E6F8ULL, "6140599_m.fxe"},
	{0x615C1E524768D0A5ULL, "3264118.ogg"},
	{0x615C246526B5E9C7ULL, "142455749.ogg"},
	{0x615C66AF72354D9BULL, "6133378_m.fxe"},
	{0x615C7A84B6EDB4F7ULL, "151253338.ogg"},
	{0x615C7C1461D06675ULL, "vxca_caverns2prc_78p_0_1.vlm"},
	{0x615DAB24609F125FULL, "755681308.ogg"},
	{0x615DFD12FCBCA82CULL, "vis121_saemus6086704.cl"},
	{0x615E1832F749AC73ULL, "653528185.ogg"},
	{0x615E2D1B0A91C0D9ULL, "qun110_petrice6042224.cl"},
	{0x615E9200D6E0BDDBULL, "915098315.ogg"},
	{0x615F07B2178D62DEULL, "6107556_m.fxe"},
	{0x615F52F2F4D39A7CULL, "978946122.ogg"},
	{0x615F9B6AA1835625ULL, "gen00fl_fenris6108017.cl"},
	{0x615FB772BD1612E1ULL, "hlac_w_encounter2_camp_262147_0y.dds"},
	{0x61602038A37EC76AULL, "gen00fl_merrill6106810.cl"},
	{0x61606685825D8E0AULL, "mrl000_marethari6142123.cl"},
	{0x616116B2E14D8A21ULL, "454211532.ogg"},
	{0x616178D74658D6C2ULL, "vxht_0_hightown_bp9_0_0.vlm"},
	{0x6162C3C413047AC4ULL, "221147586.ogg"},
	{0x6162DF7C6899102FULL, "art\\vfx\\combat\\vfx_m_arc_bolt_die_c.anb"},
	{0x6162FEF169133FC8ULL, "6206651_m.fxe"},
	{0x616350C8E51839C3ULL, "hlac_lt_foundry_n_320_8z.dds"},
	{0x61636B0943E28EF2ULL, "dae220st_merch_tailor.utm"},
	{0x61639FF4DC0A62E5ULL, "hlac_lt_0_lowtown_a2_-65535_3z.dds"},
	{0x6163B2BA12C279F7ULL, "6135138_m.fxe"},
	{0x6163C6BB1234460CULL, "vxca_caverns1_mines_icx_0_0.vlm"},
	{0x6163FB1DFFDF38C7ULL, "llac_dr_primevaltomb_661_2.dds"},
	{0x616439D6BDC0B114ULL, "vxlt_0_lowtown_a2_n_c_hqg_0_0.vlm"},
	{0x61645C91BDBFE936ULL, "llac_lt_0_lowtown_docks_n_0_3.dds"},
	{0x61647848F493398BULL, "character_character_p91.psh.fxo.dx11"},
	{0x6164D33DB2957487ULL, "6166606_m.fxe"},
	{0x6164EF3508F06999ULL, "62988237.ogg"},
	{0x6164F061F85D6840ULL, "plt_lgt220pt_lights_out.nss"},
	{0x61651DB7745B5B7EULL, "222814049.ogg"},
	{0x61656B1D09546FE8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblelarge.phy"},
	{0x61658BD5B0854C7FULL, "vxdr_primevalentrance_32n_0_0.vlm"},
	{0x6165FAE7A27D88FCULL, "dre000_sibling_has_blight6036461.cl"},
	{0x61661530497B8AB3ULL, "6083917_f.fxe"},
	{0x61663F6D7AB513BBULL, "hlac_lt_undercityserialkiller_-1_4y.dds"},
	{0x61665D81DBA8350BULL, "gen00fl_aveline6152527.cl"},
	{0x61666E4B8BC42D0FULL, "pro000_meeran6050114.cl"},
	{0x61667C17202FB927ULL, "952112076.ogg"},
	{0x6166D7466EE0B4ADULL, "756699546.ogg"},
	{0x6166FE65ECC72744ULL, "hlac_w_blightlands_0_6y.dds"},
	{0x616793CDC2198572ULL, "364739208.ogg"},
	{0x61679BE852A64BF4ULL, "vxht_keep_n_70i_0_0.vlm"},
	{0x6167ADBEF0612C9FULL, "6066410_m.fxe"},
	{0x6167B33E93855F91ULL, "vxdr_primevaltomb_6z5_0_0.vlm"},
	{0x6167BE5AF5B7B981ULL, "vxht_mansion2_2iy_0_0.vlm"},
	{0x6167DB8595D60DB3ULL, "6126865_f.fxe"},
	{0x6168000342DDDE70ULL, "695612603.ogg"},
	{0x6168E3E8AE6B689CULL, "w_bonepitdestroyed_a3_3_0.mao"},
	{0x6168EA6061E62847ULL, "llac_w_bonepit_0_5.dds"},
	{0x6168EFE8AE6B7D3FULL, "w_bonepitdestroyed_a3_3_0.mmh"},
	{0x6168F5E8AE6B87E1ULL, "w_bonepitdestroyed_a3_3_0.msh"},
	{0x616929C48BCBF486ULL, "follower_banter6161351.cl"},
	{0x61693EF2F0DB228FULL, "vxca_cave_1ny_0_0.vlm"},
	{0x6169BF0B8D16C1BBULL, "hlac_ca_caverns2prc_5802_2y.dds"},
	{0x616A6A2EE76800DEULL, "hlac_lt_0_lowtown_a2_n_-1_8x.dds"},
	{0x616A764E7CF990B3ULL, "85074832.ogg"},
	{0x616A80212927B079ULL, "vxca_cave_188_0_1.vlm"},
	{0x616B1DB407E05F3BULL, "6002882_f.fxe"},
	{0x616B2DC3613BCC74ULL, "727243834.ogg"},
	{0x616C09C01F5D5FC0ULL, "lt_0_lowtown_docks_32_0.rml"},
	{0x616C4986F7DBC801ULL, "vxht_mansion_n_704_0_0.vlm"},
	{0x616C7421B396E7C8ULL, "6125473_m.fxe"},
	{0x616D0ED9A0D09CB0ULL, "6081380_m.fxe"},
	{0x616D64644AF8B7D2ULL, "follower_banter6166682.cl"},
	{0x616D69CF13667E0EULL, "ss_city_guard_m_1.wbd"},
	{0x616D908C533B27BCULL, "vxht_keep_629_0_0.vlm"},
	{0x616DF37C77BCE63DULL, "vxlt_warehouse_q9_0_0.vlm"},
	{0x616E736CE2AF2731ULL, "43615181.ogg"},
	{0x616EA2FAE5DF5DC2ULL, "573185445.ogg"},
	{0x616EC735B9674502ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd_basic.phy"},
	{0x616EC868D26A46FCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ff_basic.mao"},
	{0x616ED468D26A5A9FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ff_basic.mmh"},
	{0x616EDA68D26A6541ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ff_basic.msh"},
	{0x616F9355DCB16ECEULL, "lt_0_lowtown_docks_198_0.msh"},
	{0x616FA155DCB1861FULL, "lt_0_lowtown_docks_198_0.mao"},
	{0x616FAD55DCB19ABCULL, "lt_0_lowtown_docks_198_0.mmh"},
	{0x6170521A4AA0A01DULL, "dae000_templar_knight6165182.cl"},
	{0x61705EA9A8F85B97ULL, "mag370_meredith6046922.cl"},
	{0x61707EE724095276ULL, "525009111.ogg"},
	{0x6170E61BB87BDCB3ULL, "mrl150_merrill6150866.cl"},
	{0x61723898042A5E8CULL, "vxca_caverns1_mines_ixy_0_0.vlm"},
	{0x61727DCEDCBB1F11ULL, "art\\vfx\\materials\\vfx_blood_sh_si_add.mao"},
	{0x6172951264BCAD4DULL, "6168488_m.fxe"},
	{0x6172DCA6A7A4C128ULL, "vxht_0_hightown_n_c7h_0_0.vlm"},
	{0x6172F5AC3ED27CB7ULL, "hlac_w_blightlands_196608_0x.dds"},
	{0x617302E4348F0DADULL, "6137264_m.fxe"},
	{0x6174758B5E47CD28ULL, "qun200_party_comment6090392.cl"},
	{0x61749A64A3BC12E6ULL, "6167002_m.fxe"},
	{0x6174ECD9D0048684ULL, "art\\vfx\\materials\\vfx_crack_decal.mao"},
	{0x617545D51E44BB03ULL, "vxlt_gallowstemplar_f_38y_0_0.vlm"},
	{0x617677C854095172ULL, "t3_wpn_hst.tnt"},
	{0x61769C2967DB189CULL, "zz_start_mcr6081962.cl"},
	{0x6177C4053DB750ACULL, "fx_icebling.mao"},
	{0x617876BB1B9A11EEULL, "vxlt_gallowsprison_n_6rk_0_0.vlm"},
	{0x6178D6BB041741B6ULL, "vxw_encounter2_camp_3k_0_0.vlm"},
	{0x6179120AD3C26D25ULL, "vxht_0_hightown_a3_amc_0_0.vlm"},
	{0x617923429322E3D6ULL, "6016823_m.fxe"},
	{0x6179918D42294DC8ULL, "179114883.ogg"},
	{0x6179E51F8894E1FCULL, "art\\characters\\playercharacter\\humanfemale\\arm_dksa.mao"},
	{0x6179F11F8894F59FULL, "art\\characters\\playercharacter\\humanfemale\\arm_dksa.mmh"},
	{0x6179F71F88950041ULL, "art\\characters\\playercharacter\\humanfemale\\arm_dksa.msh"},
	{0x617A099D5D6D9897ULL, "5089027.ogg"},
	{0x617A2292960D9F2DULL, "185695597.ogg"},
	{0x617A53A686DD4D15ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb04.mmh"},
	{0x617A55A686DD508BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb04.msh"},
	{0x617BD9FCAD226673ULL, "mag360_samson6044917.cl"},
	{0x617C655969B2FEBDULL, "vxlt_warehouse_z8_0_0.vlm"},
	{0x617C9C81B2A8BBBAULL, "289919963.ogg"},
	{0x617CD1E21AFA21EDULL, "hlac_lt_0_lowtown_docks_a3_196606_0x.dds"},
	{0x617D219DBE0804DCULL, "334446659.ogg"},
	{0x617D28F61147C891ULL, "787837919.ogg"},
	{0x617D5F160B15D8C3ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maso.mmh"},
	{0x617D651379127591ULL, "472439503.ogg"},
	{0x617D6D160B15F01DULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maso.msh"},
	{0x617DB32AFDF2CC10ULL, "zz_vault_debug6200368.cl"},
	{0x617DDF0A93144ABDULL, "6087787_m.fxe"},
	{0x617DE569C11633ECULL, "w_encounter3_15_0.rml"},
	{0x617DFB4E96690ABFULL, "vxlt_0_lowtown_docks_a3_d75_0_0.vlm"},
	{0x617E5854EF24E264ULL, "gen00fl_sebastian.utc"},
	{0x617E5A8503CE3BDAULL, "331634299.ogg"},
	{0x617ED8E23C6689CFULL, "vxw_bonepitdestroyed_a3_1ka_0_0.vlm"},
	{0x617F27FF7F9678D5ULL, "hlac_lt_0_lowtown_a2_-131072_7z.dds"},
	{0x617FFD6459DDCCAEULL, "hlac_ht_0_hightown_a2_-196610_0y.dds"},
	{0x6180A0BF0DEB8049ULL, "6079265_m.fxe"},
	{0x6181056CDAB59EC8ULL, "isa311_isabela6091961.cl"},
	{0x61814AA98DAA8D11ULL, "272092542.ogg"},
	{0x61817B83E5C0B512ULL, "529195005.ogg"},
	{0x61820629618F1737ULL, "vxlt_undercitydungeon_15p_0_0.vlm"},
	{0x61824C25619D9473ULL, "mag110_arianni6016846.cl"},
	{0x61827C3102B095B0ULL, "vxlt_0_lowtown_docks_bcw_0_0.vlm"},
	{0x6182A8D2FD047C07ULL, "342696410.ogg"},
	{0x6183485A96A7C86EULL, "vxht_mansion_n_4vj_0_0.vlm"},
	{0x6183A308D40CE675ULL, "gen00fl_isabela6088245.cl"},
	{0x6183C78A55827A75ULL, "vxca_caverns2_73i_0_0.vlm"},
	{0x6183F31BB29475ECULL, "6141592_m.fxe"},
	{0x618444889A50EAFDULL, "6147559_m.fxe"},
	{0x61844F804F672264ULL, "dae011_corff_sta6164362.cl"},
	{0x61848A16294E3DBAULL, "vxdr_primevaltomb_9e8_0_0.vlm"},
	{0x6184C561A0F8A2C8ULL, "214305061.ogg"},
	{0x61857E33562022EFULL, "gen00fl_merrill6131262.cl"},
	{0x618582C86A6BF8B4ULL, "6151706_m.fxe"},
	{0x61861D0C5950DF66ULL, "vxw_encounter1_ep_0_0.vlm"},
	{0x6186C028FD4CA2C1ULL, "sw_imp_cyc.mao"},
	{0x61874FC12AB5C43CULL, "6071817_m.fxe"},
	{0x61876DE20F7CDD65ULL, "1016932533.ogg"},
	{0x6187D4534F2D9414ULL, "vxca_caverns2prc_a9q_0_0.vlm"},
	{0x618886B90B6CBB14ULL, "hlac_lt_gallowstemplar_2383_0z.dds"},
	{0x618963BC10B5B18DULL, "hlac_lt_warehouse_n_65537_1x.dds"},
	{0x618AADCE7F2BD1AAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ingots_d.dds"},
	{0x618AC49DCAC10E75ULL, "lt_warehouse_n_mmap.dds"},
	{0x618AD0BFAE10C8CAULL, "169526656.ogg"},
	{0x618B39EEF5FDE163ULL, "176231955.ogg"},
	{0x618B45D50CE32C7EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\revenant_overlay.anb"},
	{0x618B5E5CA7F0AE61ULL, "851966080.ogg"},
	{0x618B955442A5AE3EULL, "6142995_m.fxe"},
	{0x618BDD01CF4551CAULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bookcase_sa.dds"},
	{0x618BF90DE4659C35ULL, "6129937_m.fxe"},
	{0x618C304361E1E30EULL, "vxca_caverns2_82b_0_0.vlm"},
	{0x618C717422A37838ULL, "548354690.ogg"},
	{0x618CCF2B72339895ULL, "dae000_bartrand6107545.cl"},
	{0x618CD537EAC4671CULL, "hlac_lt_hangedman_-1_1z.dds"},
	{0x618CDFEAA7127067ULL, "hlac_lt_0_lowtown_docks_a3_-3_0z.dds"},
	{0x618D7D567B207FE3ULL, "vxht_mansion2_3ej_0_0.vlm"},
	{0x618DD0F3EE5BE8C1ULL, "263002022.ogg"},
	{0x618DE6582CB27A31ULL, "gen00fl_anders6113055.cl"},
	{0x618DFDF802EECE27ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\ready.gad"},
	{0x618E284DA23B9186ULL, "6085131_m.fxe"},
	{0x618E3B99CB9378FBULL, "vxca_caverns1_fo9_0_0.vlm"},
	{0x618E56D105065987ULL, "597885081.ogg"},
	{0x618E60160B244DE8ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maso.phy"},
	{0x618E7764278798CAULL, "vxht_0_hightown_n_bl2_0_0.vlm"},
	{0x618EED4D47ACD9F7ULL, "vxca_caverns1_mines_k2r_0_0.vlm"},
	{0x618F104DDCF1E57AULL, "hlac_lt_0_lowtown_docks_n_-131074_1y.dds"},
	{0x618F58234F2C078BULL, "380153704.ogg"},
	{0x618F5F9CE05E2AB3ULL, "995704412.ogg"},
	{0x618F73DED41CC2E0ULL, "674161148.ogg"},
	{0x618FE26CB0242237ULL, "qun131_party_comment6117291.cl"},
	{0x618FE57B3FFA919DULL, "t3_wdg_bn3.tnt"},
	{0x6190EE0DF4A6AE30ULL, "vxlt_0_lowtown_docks_bft_0_0.vlm"},
	{0x6191155A69DC1492ULL, "vxca_cavern1_haunted_jsy_0_0.vlm"},
	{0x6191B67D4A66E4DAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutc.mmh"},
	{0x6191BA7D4A66EB11ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutc.mao"},
	{0x6191C87D4A670320ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutc.msh"},
	{0x61921C7F8719DDD4ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_imp_p.anb"},
	{0x61932EEDDCC1AF05ULL, "6105993_m.fxe"},
	{0x619354CDAC179B9AULL, "vxw_encounter3_fk_0_0.vlm"},
	{0x6193C543F9594DBFULL, "6064728_m.fxe"},
	{0x6194098AE744EC8FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0pp13_galowint.phy"},
	{0x61943C75716B54BFULL, "6087222_m.fxe"},
	{0x619455447F33D7B6ULL, "plt_ser000pt_debug.nss"},
	{0x61945F227588431CULL, "mag100_karras6034918.cl"},
	{0x61948EE94B46F7E6ULL, "vxw_encounter1_jq_0_0.vlm"},
	{0x6194A235B987D5B6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd_basic.mao"},
	{0x6194AE35B987EA55ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd_basic.mmh"},
	{0x6194B035B987EDCBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd_basic.msh"},
	{0x6194E92EF0816946ULL, "hlac_lt_0_lowtown_a2_-2_0z.dds"},
	{0x6194F2F802F4DA34ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\ready.evt"},
	{0x61951016399D93A6ULL, "114833499.ogg"},
	{0x619627B71D96BC4DULL, "104019745.ogg"},
	{0x619642C821A094ECULL, "ser221_gamlen6135236.cl"},
	{0x619646304123AC94ULL, "vxca_caverns2prc_a4v_0_0.vlm"},
	{0x61964BAA30588608ULL, "rdr111dg_martin6054900.cl"},
	{0x61964C25F0AEA619ULL, "gen00fl_dog6166421.cl"},
	{0x6196DD053DD22BBEULL, "fx_icebling.dds"},
	{0x6196DD234E4D4B07ULL, "823626668.ogg"},
	{0x61973257A51B8278ULL, "isa311_party_comments6117823.cl"},
	{0x61977800850F8CCBULL, "vxht_mansion2_1qp_0_0.vlm"},
	{0x6197849642BBDC26ULL, "652462131.ogg"},
	{0x6197ABEBF60E2F25ULL, "vxht_0_hightown_cbi_0_0.vlm"},
	{0x619897BB26E096E2ULL, "gen00fl_varric6133858.cl"},
	{0x6198AC2869EDBD1EULL, "vxlt_hangedman_10x_0_0.vlm"},
	{0x619907AA6FAA9F71ULL, "llac_lt_0_lowtown_-2_5.dds"},
	{0x61995FCF138BA83CULL, "ss_city_guard_m_1.bnk"},
	{0x61996BF2B6815ECEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_frameb.msh"},
	{0x619979F2B681761FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_frameb.mao"},
	{0x619985F2B6818ABCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_frameb.mmh"},
	{0x6199B69C4A375603ULL, "949862784.ogg"},
	{0x619A47BD1D1D9177ULL, "6032143_m.fxe"},
	{0x619ACAAD04E04E6CULL, "vxw_bonepitdestroyed_a3_dl_0_0.vlm"},
	{0x619B396D8199267FULL, "6154746_m.fxe"},
	{0x619B45B86767C88DULL, "6075240_f.fxe"},
	{0x619BEDA6A9982A69ULL, "gen_im_wep_war_one_default.uti"},
	{0x619C1BA082265B36ULL, "6132739_f.fxe"},
	{0x619DFA8273AB6802ULL, "gen00fl_merrill6108526.cl"},
	{0x619E0493A486E0A0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsblock.phy"},
	{0x619E4670C8C7AE66ULL, "352319238.ogg"},
	{0x619E669F3648A15CULL, "6106603_m.fxe"},
	{0x619E7099047D6BA5ULL, "896024958.ogg"},
	{0x619E8009EF536A6FULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyn_0n.dds"},
	{0x619EFB7F87242F97ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_imp_p.mmh"},
	{0x619F2E12849AAAB0ULL, "vxca_cavern1_haunted_i9f_0_0.vlm"},
	{0x61A0CFC24E7B9702ULL, "vxht_mansion_n_6iu_0_0.vlm"},
	{0x61A0D2969C59C56AULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_5z.dds"},
	{0x61A152027E3737DDULL, "vxw_sundermount_36x_0_1.vlm"},
	{0x61A1C6AA9E0DB41AULL, "vxw_encounter3_cd_0_0.vlm"},
	{0x61A20E83092695DFULL, "ser221_gascard6165648.cl"},
	{0x61A26A2457A9FD11ULL, "vxlt_0_lowtown_n_f0u_0_0.vlm"},
	{0x61A2C72643762C53ULL, "vxw_encounter2_camp_r9_0_0.vlm"},
	{0x61A2E961F892BD03ULL, "plt_lgt220pt_lights_out.plo"},
	{0x61A2EAF803010A6DULL, "art\\characters\\creatures\\corpse\\animation\\combat\\ready.ani"},
	{0x61A33450DBCD3620ULL, "hlac_w_encounter2_camp_-65536_0x.dds"},
	{0x61A38ED8DFDD6BA1ULL, "vxht_0_hightown_n_a6p_0_0.vlm"},
	{0x61A3FCDFBCA39EABULL, "hlac_w_woundedcoast_-65536_0z.dds"},
	{0x61A424D199130EF4ULL, "vxht_0_hightown_n_cas_0_1.vlm"},
	{0x61A45DF2DAC1F66BULL, "621629830.ogg"},
	{0x61A4DB74F4861C70ULL, "19343668.ogg"},
	{0x61A4F0849D87D82FULL, "mrl310_party_comments6142632.cl"},
	{0x61A4FBAA3B1F7CFEULL, "audio_animation_events.gda"},
	{0x61A58AB21488E857ULL, "vxw_encounter2_g2_0_0.vlm"},
	{0x61A58B32560C5BA4ULL, "6114841_f.fxe"},
	{0x61A5BD8218558C12ULL, "318441368.ogg"},
	{0x61A5CFB8F5A2FFDDULL, "6176168_m.fxe"},
	{0x61A5E9DD7705A54BULL, "vxht_mansion2_1ts_0_0.vlm"},
	{0x61A5F4F901A06547ULL, "gen00fl_anders6045369.cl"},
	{0x61A60FA1A97F8C76ULL, "123341066.ogg"},
	{0x61A6B44511A78495ULL, "hlac_lt_0_lowtown_docks_a3_-131073_0z.dds"},
	{0x61A6CF0BAEA31100ULL, "6110693_m.fxe"},
	{0x61A6E4AC0D217EB9ULL, "vxlt_hangedman_2ys_0_0.vlm"},
	{0x61A6EE6408E92D23ULL, "llac_lt_0_lowtown_docks_-131075_7.dds"},
	{0x61A6F4F4F4F3B83BULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.cess.evt"},
	{0x61A7545995492678ULL, "6151135_m.fxe"},
	{0x61A7790217F021C4ULL, "vxca_caverns1_j23_0_2.vlm"},
	{0x61A7841F6A1A7C3CULL, "6059402_m.fxe"},
	{0x61A795625B9C2BC2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basic.msh"},
	{0x61A797625B9C2F78ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basic.mmh"},
	{0x61A7A3625B9C431BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basic.mao"},
	{0x61A7EEF24307E08AULL, "ave330tr_jeven_init.utt"},
	{0x61A8098D45979C7EULL, "6204963_m.fxe"},
	{0x61A810EDEA7D0C33ULL, "725206982.ogg"},
	{0x61A8163B34C98F2FULL, "vxht_keep_n_60g_0_0.vlm"},
	{0x61A88A2CA9E92045ULL, "666413528.ogg"},
	{0x61A9155783B2A96DULL, "llac_lt_0_lowtown_2_0.dds"},
	{0x61A92BBE30A11AA1ULL, "hlac_lt_gallowscourtyard_-131072_2z.dds"},
	{0x61A93D1FE9D9581EULL, "gen00fl_anders6096533.cl"},
	{0x61A97207597E5BC4ULL, "83017243.ogg"},
	{0x61A9CAB537E88410ULL, "vxlt_0_lowtown_n_fag_0_0.vlm"},
	{0x61A9D78222D43055ULL, "6162569_m.fxe"},
	{0x61A9EA18F3947CCCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05a.phy"},
	{0x61AAB3CBE1F6FB8BULL, "vxw_bonepitdestroyed_a3_16z_0_0.vlm"},
	{0x61AB93A12CA8C309ULL, "vxht_mansion2b_3cj_0_0.vlm"},
	{0x61AB93C6BEF4FE2EULL, "gen00fl_fenris6127024.cl"},
	{0x61ABC99ADF28D04FULL, "vxlt_0_lowtown_docks_b0q_0_0.vlm"},
	{0x61AC3DA31B245DA0ULL, "6139151_m.fxe"},
	{0x61AC849B53E1C73DULL, "dae000_fenarel6128134.cl"},
	{0x61ACEDF4F4F8187CULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.cess.gad"},
	{0x61AD43AB28B6C716ULL, "vxlt_blackemporium_ub_0_0.vlm"},
	{0x61ADA0AFF86F1F5EULL, "mag360_thrask6043517.cl"},
	{0x61ADA5330CA4AD86ULL, "lla_ht_0_hightown_65538_1.dds"},
	{0x61ADC4AB1B9AB2A9ULL, "6096455_m.fxe"},
	{0x61AE56EC3A2D1AC5ULL, "qun211_braggart6095431.cl"},
	{0x61AE6BE4B66C9634ULL, "6120904_m.fxe"},
	{0x61AE8501157E6F48ULL, "6067740_m.fxe"},
	{0x61AE9341973FA6C6ULL, "6041175_m.fxe"},
	{0x61AECE8D8655CAEDULL, "6097315_m.fxe"},
	{0x61AF419F4071AF82ULL, "vxht_mansion_n_6lv_0_0.vlm"},
	{0x61AF4F2DAA5B88FCULL, "rune_weapon_impact.dds"},
	{0x61AF7F6D47918728ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg106a_0n.dds"},
	{0x61AFB3B5DAE77916ULL, "hm_raider_leader.mor"},
	{0x61AFBB6E249D5474ULL, "one000_ilen6060494.cl"},
	{0x61AFC16DD7B2634CULL, "vxlt_0_lowtown_a2_hev_0_0.vlm"},
	{0x61AFD55F3405536EULL, "6165530_m.fxe"},
	{0x61AFE07B21FECA02ULL, "6092579_m.fxe"},
	{0x61B0049D0105D2C4ULL, "vxlt_gallowstemplar_n_4uj_0_0.vlm"},
	{0x61B01F04E4228981ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\meredeth_.gad"},
	{0x61B02E3FBB1488CEULL, "dae000_mother6099200.cl"},
	{0x61B05DD9F73B3F34ULL, "vxht_0_hightown_n_am8_0_0.vlm"},
	{0x61B08BACB3F4BC8AULL, "227541687.ogg"},
	{0x61B08D5BA454FD8EULL, "266094171.ogg"},
	{0x61B0D0305574FBDBULL, "vxlt_0_lowtown_docks_a3_n_d2q_0_0.vlm"},
	{0x61B13903356C44D3ULL, "vxw_encounter2_camp_m8_0_0.vlm"},
	{0x61B167C8BD7B8683ULL, "art\\vfx\\combat\\vfx_g_spider_att_f1_c.anb"},
	{0x61B19B5C3A9A0C5FULL, "6076491_m.fxe"},
	{0x61B19E625BA48F03ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basic.phy"},
	{0x61B1D7753B4BAADCULL, "vxlt_0_lowtown_docks_bsk_0_0.vlm"},
	{0x61B227AF17E48BD4ULL, "mag121_pricia_corthwaite.stg"},
	{0x61B23BD7C6781DCBULL, "dae000_male_templar_mob6174367.cl"},
	{0x61B240F7E383B648ULL, "6138831_m.fxe"},
	{0x61B25B8C3EA6867FULL, "hlac_lt_backalley_n_65537_5y.dds"},
	{0x61B2BA27B78884EAULL, "vxlt_0_lowtown_docks_n_b7r_0_0.vlm"},
	{0x61B301EC337DC563ULL, "hm_ave3_coteriem.mor"},
	{0x61B31E9E5FB4DA62ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk05.phy"},
	{0x61B38DCF77495A8BULL, "856413815.ogg"},
	{0x61B3E940C3626FB9ULL, "vxca_caverns1_mines_edx_0_0.vlm"},
	{0x61B3ECF4F4FE357AULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.cess.ani"},
	{0x61B3FD0F48027424ULL, "6113941_f.fxe"},
	{0x61B431A9BC95C9E5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junka_d.dds"},
	{0x61B45B7A883870FCULL, "vxlt_gallowscourtyard_4iw_0_0.vlm"},
	{0x61B472BBE4194C55ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_mushroom_d.dds"},
	{0x61B5162C12DABD94ULL, "follower_banter6151242.cl"},
	{0x61B519B4E397134DULL, "vxht_bartrand_n_72q_0_0.vlm"},
	{0x61B5FA26CABCB332ULL, "122733109.ogg"},
	{0x61B63EC11473722CULL, "follower_banter6156045.cl"},
	{0x61B7032DAE00F2B2ULL, "zz_gateguardian.cut"},
	{0x61B72004E428AA2FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\meredeth_.ani"},
	{0x61B75BA5FA79E33BULL, "hlac_ht_keep_n_3740_0y.dds"},
	{0x61B76D05E1EEA3B9ULL, "vxca_caverns2_5aa_0_0.vlm"},
	{0x61B774EF526B7329ULL, "70963767.ogg"},
	{0x61B7F73CE8C62E1CULL, "hlac_ht_mansion2_818_57z.dds"},
	{0x61B8B853AFF06631ULL, "vxht_chantry_n_20i_0_0.vlm"},
	{0x61B8DDEFE43D5F29ULL, "6021347_m.fxe"},
	{0x61B8F6F5F3E34B79ULL, "6149049_m.fxe"},
	{0x61B9519DA11AA0DCULL, "art\\characters\\creatures\\desiredemon\\animation\\ua_dconjure_complete.ani"},
	{0x61B9DD7821BD84DBULL, "art\\characters\\playercharacter\\humanfemale\\rob_appa.mmh"},
	{0x61B9E97821BD9978ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appa.mao"},
	{0x61B9FB7821BDB7C5ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appa.msh"},
	{0x61BA057E1E9EDB89ULL, "vxht_mansion2b_3fi_0_0.vlm"},
	{0x61BA3B77D11EE8CFULL, "vxlt_0_lowtown_docks_b7p_0_0.vlm"},
	{0x61BABC00FED9E9F0ULL, "vxlt_0_lowtown_a3_n_c_bxw_0_0.vlm"},
	{0x61BAD4D4C4860B56ULL, "character_character_p182.psh.fxo.dx11"},
	{0x61BB4A0A49FE4EBDULL, "vxca_caverns1_i2c_0_0.vlm"},
	{0x61BB552344CECEF1ULL, "6125494_m.fxe"},
	{0x61BB94F3F6630F87ULL, "hlac_lt_undercityencounter_-1_43x.dds"},
	{0x61BBA379CDC9135CULL, "vxca_caverns1_hmb_0_0.vlm"},
	{0x61BBCDB1F552D53EULL, "hlac_lt_0_lowtown_a3_n_c_1_2z.dds"},
	{0x61BC56DCDC5759BCULL, "6153379_m.fxe"},
	{0x61BCB62CCC3101F5ULL, "1066935744.ogg"},
	{0x61BD0C9FAF5B6195ULL, "dre000_carver6037624.cl"},
	{0x61BD11DCA20B46E8ULL, "vrc221_bartrand6065280.cl"},
	{0x61BD140D79D74005ULL, "133412056.ogg"},
	{0x61BE174FB4470BCAULL, "443787456.ogg"},
	{0x61BE64F2B6A06E87ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_frameb.phy"},
	{0x61BE98BEBE94B300ULL, "nat320_delilah.cnv"},
	{0x61BED0EEADE172B7ULL, "ave200_donnic_post_patrol.stg"},
	{0x61BEF317FAEFC20EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\righthand_clench_fist.ani"},
	{0x61BF44C8BD8788ACULL, "art\\vfx\\combat\\vfx_g_spider_att_f1_c.mmh"},
	{0x61BF5DAB5FB4B1B8ULL, "dae000_pc_home_ambients6208336.cl"},
	{0x61BF5E9DA11F221EULL, "art\\characters\\creatures\\desiredemon\\animation\\ua_dconjure_complete.gad"},
	{0x61BFC9F2431C3473ULL, "ave330tr_jeven_init.ncs"},
	{0x61C0324E3054D813ULL, "437293209.ogg"},
	{0x61C03AA679CD51BDULL, "hlac_w_woundedcoast_-131071_0x.dds"},
	{0x61C049522D41C35CULL, "vxlt_0_lowtown_docks_bvh_0_0.vlm"},
	{0x61C0855148A41E1AULL, "vxht_keep_6hy_0_0.vlm"},
	{0x61C0A75CF9AD174FULL, "follower_barks6153376.cl"},
	{0x61C0CC33F9634C23ULL, "hlac_lt_gallowstemplar_n_2388_1x.dds"},
	{0x61C1233C6328F7A7ULL, "6035278_m.fxe"},
	{0x61C175939404C807ULL, "vxlt_hangedman_n_1xn_0_0.vlm"},
	{0x61C190FE94751897ULL, "rdr111dg_assistant6055116.cl"},
	{0x61C1E44C9CF74AD7ULL, "vxht_0_hightown_a2_asv_0_0.vlm"},
	{0x61C1FBA5435AA98DULL, "6170085_m.fxe"},
	{0x61C20BC2AAFDC3D3ULL, "156794065.ogg"},
	{0x61C233AF21554929ULL, "hf_amb_elegant.mor"},
	{0x61C2513CC814C52CULL, "hlac_lt_gallowscourtyard_-1_3y.dds"},
	{0x61C264E880930756ULL, "6062815_m.fxe"},
	{0x61C27ABEBE986938ULL, "nat320_delilah.bnk"},
	{0x61C2B0FA594E582AULL, "fen100pt_bait.ncs"},
	{0x61C2CD975AF1D64BULL, "vxht_mansion2_1bu_0_0.vlm"},
	{0x61C2E6B54DEEF68BULL, "gen00fl_aveline6101826.cl"},
	{0x61C32193A4A62E35ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsblock.msh"},
	{0x61C32393A4A631ABULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsblock.mmh"},
	{0x61C32F93A4A64648ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsblock.mao"},
	{0x61C34C55C9CBADD3ULL, "6125193_m.fxe"},
	{0x61C3681DC72BF451ULL, "6161003_m.fxe"},
	{0x61C3949A406507CAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medb_0s.dds"},
	{0x61C4825FA9D7D0BFULL, "gen00fl_merrill6111219.cl"},
	{0x61C4B39C496BC32DULL, "617929485.ogg"},
	{0x61C4D88A02B202D7ULL, "6029370_m.fxe"},
	{0x61C526FC31DBFAE7ULL, "vxht_keep_n_3ii_0_0.vlm"},
	{0x61C53895C9725BBDULL, "llac_lt_0_lowtown_a2_n_1_11.dds"},
	{0x61C56BBFAB9A1076ULL, "6082350_m.fxe"},
	{0x61C57A060263EB3BULL, "6209892_f.fxe"},
	{0x61C58251095470BFULL, "vxw_bonepitdestroyed_a3_dc_0_0.vlm"},
	{0x61C5B42DA09D0B17ULL, "6125742_m.fxe"},
	{0x61C5EE617D5E788EULL, "vxht_0_hightown_b9z_0_0.vlm"},
	{0x61C5F1379620ACD4ULL, "gen00fl_varric6093660.cl"},
	{0x61C6399DA125023DULL, "art\\characters\\creatures\\desiredemon\\animation\\ua_dconjure_complete.evt"},
	{0x61C6CD4D7625ACC1ULL, "860827585.ogg"},
	{0x61C6DC0607C455AFULL, "vxht_mansion_n_6k2_0_0.vlm"},
	{0x61C71D7989488E17ULL, "6178436_m.fxe"},
	{0x61C7337D805BA133ULL, "6167475_f.fxe"},
	{0x61C768D2E5D963F9ULL, "6142149_m.fxe"},
	{0x61C778968EAABC09ULL, "6747797.ogg"},
	{0x61C810F02EB04B3AULL, "493713494.ogg"},
	{0x61C8274BD6E38B92ULL, "dae011_conspiracy_mob6149551.cl"},
	{0x61C85F78AA3CB990ULL, "dae000ip_crowd_ff1_talk2.utp"},
	{0x61C8B12CDC8DE495ULL, "202281004.ogg"},
	{0x61C924F2EC8B9515ULL, "vxdr_primevaltomb_7jt_0_0.vlm"},
	{0x61C95088DC6FDA46ULL, "dae011_corff_sta6164359.cl"},
	{0x61C99C684FBD2725ULL, "art\\vfx\\combat\\vfx_c_rw_s2_s1_trans_c.anb"},
	{0x61CA37AC30F3100CULL, "318695299.ogg"},
	{0x61CA69C4DDF20E72ULL, "hlac_lt_0_lowtown_-65537_2z.dds"},
	{0x61CCCC9EEB128475ULL, "675272251.ogg"},
	{0x61CCFF2FED21E045ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_06a_0d.dds"},
	{0x61CD0E1F52FEB879ULL, "vxca_cavern1_haunted_ehl_0_0.vlm"},
	{0x61CD59B4DFF31E3BULL, "vxht_mansion2_21s_0_0.vlm"},
	{0x61CDD4C1188D1213ULL, "659824867.ogg"},
	{0x61CDDD5E813DC4FEULL, "583079004.ogg"},
	{0x61CE1176C47A5CD6ULL, "art\\characters\\weapons\\longswords\\w_lsw_003a.msh"},
	{0x61CE1B76C47A6DD4ULL, "art\\characters\\weapons\\longswords\\w_lsw_003a.mmh"},
	{0x61CE1F76C47A74E7ULL, "art\\characters\\weapons\\longswords\\w_lsw_003a.mao"},
	{0x61CE662B5EB24EE2ULL, "vxlt_shop_sc_0_0.vlm"},
	{0x61CEE57965E3EBBAULL, "vxw_bonepitdestroyed_a3_242_0_0.vlm"},
	{0x61CEF22C692D4BB3ULL, "gen00fl_aveline6139274.cl"},
	{0x61CF1368E12DEF42ULL, "6076911_m.fxe"},
	{0x61CF14F311E500C0ULL, "lgt260cr_eau_templar_capt.utc"},
	{0x61CF2075C5F10593ULL, "40649163.ogg"},
	{0x61CF22BE881B1151ULL, "6153626_m.fxe"},
	{0x61CF40A6EAD826F5ULL, "fen110_anso6094982.cl"},
	{0x61CFE0BF943C2336ULL, "87350447.ogg"},
	{0x61D069117560AB85ULL, "863586908.ogg"},
	{0x61D096AD17AF4FD2ULL, "6014137_m.fxe"},
	{0x61D09D913275E77AULL, "6207001_m.fxe"},
	{0x61D0B338AB19C5BCULL, "vxlt_gallowscourtyard_n_7gi_0_0.vlm"},
	{0x61D1213A9FF0B7E4ULL, "vxdr_0_deeproads_d3c_0_0.vlm"},
	{0x61D13F07BA909F0BULL, "vxlt_0_lowtown_a3_n_c_e25_0_0.vlm"},
	{0x61D1BE1F8F901F97ULL, "hlac_ht_bartrand_n_3381_0y.dds"},
	{0x61D1DD68D2BEE7E4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ff_basic.phy"},
	{0x61D1EFE8AEC4660CULL, "w_bonepitdestroyed_a3_3_0.rml"},
	{0x61D2C31835FE9887ULL, "417673634.ogg"},
	{0x61D2CC17FB002948ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\righthand_clench_fist.gad"},
	{0x61D3D65F0E21888AULL, "follower_barks6153685.cl"},
	{0x61D3F3A178212628ULL, "art\\characters\\creatures\\ashwraith\\c_ashwraith.mmh"},
	{0x61D3FFA178213ACBULL, "art\\characters\\creatures\\ashwraith\\c_ashwraith.mao"},
	{0x61D411A178215952ULL, "art\\characters\\creatures\\ashwraith\\c_ashwraith.msh"},
	{0x61D47070D5A33D5DULL, "6096504_f.fxe"},
	{0x61D4905076857AFBULL, "hlac_dr_slavers_3217_9y.dds"},
	{0x61D4A8317DB8668CULL, "828594354.ogg"},
	{0x61D4CC8FC2AD6B63ULL, "follower_banter6164056.cl"},
	{0x61D4D0C96A4BFA81ULL, "762190790.ogg"},
	{0x61D4D76F7656977CULL, "and330st_base_justice_post2.stg"},
	{0x61D535C3F289B1AAULL, "mer220_hubert6081396.cl"},
	{0x61D557B822317448ULL, "445355644.ogg"},
	{0x61D78C52FF330C20ULL, "vxht_mansion_n_5i4_0_0.vlm"},
	{0x61D7BDA0C6521A8FULL, "framebuffereffects_cocdownsample_p0.vsh.fxo"},
	{0x61D7D4155AE1E9C8ULL, "cd_isa211_walleyedsam.wbd"},
	{0x61D8059E5FD3CBB5ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk05.mmh"},
	{0x61D8079E5FD3CF2BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk05.msh"},
	{0x61D8116BA3620A43ULL, "6174766_m.fxe"},
	{0x61D82F41F8194A87ULL, "6037584_m.fxe"},
	{0x61D8A7C1B513A1ABULL, "hlac_ca_caverns1_mines_24424_12x.dds"},
	{0x61D8F4ECA34E6469ULL, "vxlt_0_lowtown_n_fhe_0_0.vlm"},
	{0x61D94AFE1F5C2658ULL, "vxht_mansion2_3dv_0_0.vlm"},
	{0x61D9F05F8B6F9CBCULL, "ico_armor_heavy_4.dds"},
	{0x61DA59DCB2C42A24ULL, "6101325_m.fxe"},
	{0x61DAB6E158145A42ULL, "465931716.ogg"},
	{0x61DB342EBE85DBD8ULL, "6090193_m.fxe"},
	{0x61DB7B3D2629D007ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_bdy_0d.dds"},
	{0x61DBEB38C61E9187ULL, "6166017_m.fxe"},
	{0x61DC1843295E2F8DULL, "6117824_m.fxe"},
	{0x61DC88E863BB0C44ULL, "vxht_0_hightown_n_65g_0_0.vlm"},
	{0x61DC9E04311D8616ULL, "vxht_mansion2b_267_0_0.vlm"},
	{0x61DCD544CDB98291ULL, "952423903.ogg"},
	{0x61DCFA1F88E982E4ULL, "art\\characters\\playercharacter\\humanfemale\\arm_dksa.phy"},
	{0x61DD5091AE0DE33FULL, "218667546.ogg"},
	{0x61DD793A130BB463ULL, "1016451074.ogg"},
	{0x61DE0175D5D78222ULL, "hlac_lt_blackemporium_1055_9y.dds"},
	{0x61DE0D8975431481ULL, "vxca_caverns1_hm3_0_0.vlm"},
	{0x61DE27A2D220D862ULL, "qun220cr_javaris6071860.cl"},
	{0x61DE594D77F0F907ULL, "vxlt_hangedman_n_1rh_0_0.vlm"},
	{0x61DE788531BFAB0AULL, "vxlt_gallowsprison_n_6pq_0_0.vlm"},
	{0x61DEA39CCF5C5771ULL, "llac_lt_0_lowtown_docks_a3_0_6.dds"},
	{0x61DFB4452F3E3767ULL, "377882655.ogg"},
	{0x61E020DDDE8048C0ULL, "hlac_lt_0_lowtown_-3_5z.dds"},
	{0x61E0A683351E5DDAULL, "fen210_hadriana6110382.cl"},
	{0x61E1595E2C8FF372ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_rnf.gad"},
	{0x61E1B3C4E7EC6D33ULL, "6153167_m.fxe"},
	{0x61E1BCB93884614BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_loop.ani"},
	{0x61E1DA3BC8406A6BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bazaar_dp.dds"},
	{0x61E25D5376C28780ULL, "ser221_party_comments6122875.cl"},
	{0x61E32577ABA944FFULL, "zz_airshock_forest_cu_v3.cut"},
	{0x61E33E381DA825C1ULL, "one300_nuncio6083996.cl"},
	{0x61E350C9788F0470ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaclosed.mao"},
	{0x61E354C9788F0A83ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaclosed.mmh"},
	{0x61E362C9788F22DDULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomeaclosed.msh"},
	{0x61E38C1CC72004C8ULL, "6125622_m.fxe"},
	{0x61E3936CEC13FFB9ULL, "vxlt_0_lowtown_docks_9rb_0_0.vlm"},
	{0x61E3F949C0E8A276ULL, "vxht_0_hightown_c3q_0_0.vlm"},
	{0x61E476334946CA64ULL, "6018809_m.fxe"},
	{0x61E4B3191C2E2F79ULL, "643560962.ogg"},
	{0x61E4E3275A9181A9ULL, "6067701_m.fxe"},
	{0x61E52D7AAA2EA940ULL, "hlac_ca_caverns1_mines_24424_22x.dds"},
	{0x61E58457454731D0ULL, "109190200.ogg"},
	{0x61E5F3EBE25BAD5AULL, "vxw_encounter2_camp_f8_0_0.vlm"},
	{0x61E65DC5926B06E6ULL, "6120018_m.fxe"},
	{0x61E702AAACFDCDA2ULL, "vxlt_0_lowtown_a2_n_e38_0_0.vlm"},
	{0x61E774424ACAE01DULL, "vxlt_backalley_g_1dp_0_0.vlm"},
	{0x61E7EB89045FA230ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernlava_e.dds"},
	{0x61E7F668E93FE8E6ULL, "6034180_m.fxe"},
	{0x61E85C5E2C9616A9ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_rnf.evt"},
	{0x61E8E0790F4774D9ULL, "215182062.ogg"},
	{0x61E8F8C81CDE203AULL, "w_woundedcoast_241_0.msh"},
	{0x61E906C81CDE3873ULL, "w_woundedcoast_241_0.mao"},
	{0x61E90AC81CDE3E80ULL, "w_woundedcoast_241_0.mmh"},
	{0x61E97DB3A6B5E245ULL, "mcr310_argument6072682.cl"},
	{0x61E9C118F3CA1A97ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05a.mmh"},
	{0x61E9CD225830F065ULL, "6037404_f.fxe"},
	{0x61E9D718F3CA4009ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear05a.msh"},
	{0x61E9D8DF0AE8A2CAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cenotaph_d.dds"},
	{0x61EA8B2B7ABF32B6ULL, "6094458_m.fxe"},
	{0x61EB0FE123139E96ULL, "vxht_mansion2b_238_0_0.vlm"},
	{0x61EB1F80B1ABEB7DULL, "hlac_lt_undercityserialkiller_-1_9x.dds"},
	{0x61EB7C53408CC565ULL, "vxca_caverns2prc_7qv_0_0.vlm"},
	{0x61EB884EAB5F5C2EULL, "hlac_ht_mansion_n_2995_2y.dds"},
	{0x61EBCAC081AF61DEULL, "vxht_mansion2b_2wl_0_0.vlm"},
	{0x61EBDD0CB917EA37ULL, "vxlt_0_lowtown_a3_n_c_hsw_0_0.vlm"},
	{0x61EBE59296777D2EULL, "follower_banter6173136.cl"},
	{0x61EC547DD0FD760EULL, "vxdr_primevaltomb_6nj_0_0.vlm"},
	{0x61EDBDC8A79D6C47ULL, "370174269.ogg"},
	{0x61EE089785F14203ULL, "157304043.ogg"},
	{0x61EE22C19E7CD00BULL, "vxlt_0_lowtown_a3_n_c_e87_0_0.vlm"},
	{0x61EE3259E8C03FC6ULL, "qun210cr_guard6072094.cl"},
	{0x61EE5CAE4EA3CA6AULL, "character_character_p38.psh.fxo.dx11"},
	{0x61EE8AE0C49CC19FULL, "504825740.ogg"},
	{0x61EE8C57BF665D93ULL, "177806565.ogg"},
	{0x61EEF30F3379F6C8ULL, "vxdr_primevaltomb_768_0_1.vlm"},
	{0x61EF153F7A9C6B59ULL, "6118013_f.fxe"},
	{0x61EF87C876ECA1A6ULL, "cod_ltr_ser_alriks_pape.utp"},
	{0x61EF959FD8AE9D5FULL, "6165041_m.fxe"},
	{0x61EFA4B9389075C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_loop.evt"},
	{0x61EFB5BC3F890E1CULL, "mag121_idunna6092341.cl"},
	{0x61EFC81D5DE66A7DULL, "vxdr_0_deeproads_bt1_0_0.vlm"},
	{0x61F084A69744B29DULL, "250576190.ogg"},
	{0x61F1267FC133A2AAULL, "6081452_m.fxe"},
	{0x61F17E3B0D9C1FBBULL, "getflag.ncs"},
	{0x61F1C0455242A6EAULL, "hlac_ht_0_hightown_n_-131073_0z.dds"},
	{0x61F2381E3AACB6EBULL, "hlac_lt_hangedman_-1_26z.dds"},
	{0x61F26B26B2DEBAF6ULL, "vxht_0_hightown_c6r_0_0.vlm"},
	{0x61F29D684FE02E66ULL, "art\\vfx\\combat\\vfx_c_rw_s2_s1_trans_c.mmh"},
	{0x61F2A19260CF3AE1ULL, "montage_subs_demohack6217111.cl"},
	{0x61F2A38677446FBFULL, "6156623_m.fxe"},
	{0x61F31BAFB949DFE4ULL, "llac_lt_0_lowtown_docks_a3_-65539_6.dds"},
	{0x61F3C19E5D88491DULL, "6103403_m.fxe"},
	{0x61F3EF3C376D9FCAULL, "hlac_ca_cavern1_haunted_24425_12z.dds"},
	{0x61F3FC9B50C2E81FULL, "mag160_varian_ilithis6016500.cl"},
	{0x61F465C8D451C5DAULL, "vxw_encounter2_camp_a9_0_0.vlm"},
	{0x61F4E2400355C48FULL, "vxht_mansion_n_6b9_0_0.vlm"},
	{0x61F5A3B93894DFD5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_loop.gad"},
	{0x61F5D8A6873A9569ULL, "vxlt_0_lowtown_n_ffk_0_0.vlm"},
	{0x61F5F3234FA8F82EULL, "6064912_m.fxe"},
	{0x61F6445E2CA22B18ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180l_rnf.ani"},
	{0x61F66845DB360166ULL, "6031280_m.fxe"},
	{0x61F6CE6BC3CE9606ULL, "6082260_m.fxe"},
	{0x61F6FA8F25C8FAD0ULL, "6141651_m.fxe"},
	{0x61F6FCF22B0B9EE3ULL, "449490986.ogg"},
	{0x61F75DC04E42B439ULL, "vxht_0_hightown_a2_bgc_0_0.vlm"},
	{0x61F82C45F4B646F1ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_barge.mmh"},
	{0x61F834FCFC53F056ULL, "217287331.ogg"},
	{0x61F83845F4B65A92ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_barge.mao"},
	{0x61F84645F4B672AFULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_barge.msh"},
	{0x61F8C8EE0DD1CA25ULL, "vxlt_0_lowtown_n_eu9_0_0.vlm"},
	{0x61F9C51EC604F6D8ULL, "493549479.ogg"},
	{0x61F9EE303282DDE5ULL, "vxca_caverns2prc_7tu_0_0.vlm"},
	{0x61FA3C9D73A57A5EULL, "vxht_mansion2b_2pm_0_0.vlm"},
	{0x61FAE3BEC13B6BA4ULL, "zz_spy130_darktown_res_36055346.cl"},
	{0x61FB90E883A9CD49ULL, "6137707_m.fxe"},
	{0x61FC76BDFDD66D81ULL, "gen00fl_aveline6139144.cl"},
	{0x61FCE571CEF1E774ULL, "556466079.ogg"},
	{0x61FD9B6BF1A76052ULL, "vrc321st_spooky_scene_3.stg"},
	{0x61FDA0A2F1219F78ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_abma_0s.dds"},
	{0x61FDDD37358CB371ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb3.mmh"},
	{0x61FDE0A8D8D5CD96ULL, "6071194_m.fxe"},
	{0x61FDF737358CDF2FULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb3.msh"},
	{0x61FE38416CEAA307ULL, "dre000_bodahn6031342.cl"},
	{0x61FED2D74798756FULL, "dae220st_seldrek.stg"},
	{0x61FEE045D7D51B50ULL, "6032204_f.fxe"},
	{0x61FF06636F185FA9ULL, "follower_banter6155919.cl"},
	{0x61FF12C33075F3F8ULL, "465221996.ogg"},
	{0x61FF30386B435877ULL, "vxht_mansion2_21g_0_0.vlm"},
	{0x61FF551229E97235ULL, "6072871_m.fxe"},
	{0x61FF9FB1B70CF790ULL, "vxht_keep_3xn_0_0.vlm"},
	{0x62001702E98BBCE4ULL, "mrl310_merrill6120017.cl"},
	{0x6200797C06F3CC0EULL, "hlac_dr_dungeon_3217_1z.dds"},
	{0x6200DDD8993734A9ULL, "vxht_0_hightown_7j1_0_0.vlm"},
	{0x62011E053B31332FULL, "lowtown_fem_cityguard_mob6164838.cl"},
	{0x6201419B50BABA5DULL, "6131179_f.fxe"},
	{0x62015F4C7F6C96DCULL, "4634553.ogg"},
	{0x6201859D7C2C1698ULL, "791970005.ogg"},
	{0x6201FE7821FB7210ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appa.phy"},
	{0x62027CCE18F8A695ULL, "qun211_party_comment6182713.cl"},
	{0x6202AFE4A2404892ULL, "6089491_f.fxe"},
	{0x6202CDCC1CCD65BBULL, "vxlt_gallowscourtyard_779_0_0.vlm"},
	{0x6202DEEBC312317BULL, "llac_lt_0_lowtown_a2_n_-262144_3.dds"},
	{0x6202F37D137F3D63ULL, "vxw_encounter3_bj_0_0.vlm"},
	{0x620467533E5E315DULL, "mag170_hugh6148615.cl"},
	{0x6204D6FEE26105D4ULL, "vxlt_0_lowtown_docks_c6m_0_0.vlm"},
	{0x6204E7C3E1148CDEULL, "lt_foundry_n_probe_uj.dds"},
	{0x62052B3AFF5CA728ULL, "fenris_1.cut"},
	{0x6205E81D2A5B1DA8ULL, "dae000_epic_merchant_16179431.cl"},
	{0x6205FDC32C8556F9ULL, "vxht_chantry_n_2eb_0_0.vlm"},
	{0x620671E44AC149A0ULL, "6133540_m.fxe"},
	{0x6206E3CBCD7F2C0CULL, "vxlt_0_lowtown_a2_n_eor_0_0.vlm"},
	{0x620702AE67B360CDULL, "979180753.ogg"},
	{0x620878FE0B3BB22DULL, "dae011_wench_mob6144205.cl"},
	{0x6208C0C69D041B37ULL, "vxlt_0_lowtown_a3_n_c_heq_0_0.vlm"},
	{0x6208EC5D4D80E6D8ULL, "one300_nuncio6073787.cl"},
	{0x6209075F140ED24AULL, "zz_ave_debug6207471.cl"},
	{0x62093837B4E9A70EULL, "vxdr_primevaltomb_6xl_0_0.vlm"},
	{0x620A05F11C64DD61ULL, "382214560.ogg"},
	{0x620A1A83DE2F303DULL, "6127698_m.fxe"},
	{0x620A2D594F284182ULL, "6150202_m.fxe"},
	{0x620A7136AEC74863ULL, "ave100_donnic6105989.cl"},
	{0x620AC429F3631ABBULL, "6060132_f.fxe"},
	{0x620B0F82F6CE17E5ULL, "6137516_m.fxe"},
	{0x620B2FBF3A0DD75FULL, "vxca_caverns1_mines_e4n_0_0.vlm"},
	{0x620B4572063F7A4DULL, "mrl310_merrill6144364.cl"},
	{0x620B47A24670226EULL, "6028156_m.fxe"},
	{0x620B66C13D67ED73ULL, "mag170_thrask6016030.cl"},
	{0x620B74E3FEC72E2DULL, "gen00fl_merrill6119600.cl"},
	{0x620B8F4BCDB89A82ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.cess.gad"},
	{0x620D1897DFAA6263ULL, "ht_0_hightown_a3_57_0.rml"},
	{0x620DA3B6028E924AULL, "6130987_m.fxe"},
	{0x620E07EBCDB10AC6ULL, "vxlt_undercityencounter_1n9_0_0.vlm"},
	{0x620E6904802346D4ULL, "6086341_m.fxe"},
	{0x620E8241BADDA10CULL, "571415748.ogg"},
	{0x620EC60F9AE99898ULL, "vxht_bartrand_n_6qd_0_0.vlm"},
	{0x620ECD785177B03EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_action.ani"},
	{0x620F16697EC0E0F3ULL, "hlac_w_blightlands_65536_0y.dds"},
	{0x620F5D61ED85A11CULL, "hlac_lt_0_lowtown_a2_n_c_131072_2x.dds"},
	{0x620F7E1AFCC75C63ULL, "vxht_mansion2_30u_0_0.vlm"},
	{0x620FDDA5B4188118ULL, "1055764149.ogg"},
	{0x620FF168E24BB302ULL, "seb222at_to_estate.utp"},
	{0x62106C1147D5BE86ULL, "382819607.ogg"},
	{0x62110DDEBB9E7424ULL, "llac_w_bonepitmining_a2_131078_0.dds"},
	{0x62120FED4FD11E8DULL, "6160978_m.fxe"},
	{0x6212136B04F18EC1ULL, "hlac_lt_0_lowtown_a3_n_c_65535_6z.dds"},
	{0x6212450D35EFEC10ULL, "236077696.ogg"},
	{0x621251F8807D03B7ULL, "hlac_ca_cavern1_haunted_24426_12x.dds"},
	{0x62129F05D4315FBBULL, "ser210_witness6168179.cl"},
	{0x6212B24BCDBEF579ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.cess.evt"},
	{0x62131F630B60E70EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_action.evt"},
	{0x621348DBD4571E54ULL, "vxlt_0_lowtown_docks_c3r_0_0.vlm"},
	{0x621380D38C1611ACULL, "1062894017.ogg"},
	{0x6213DBC8FA981855ULL, "art\\characters\\playercharacter\\humanfemale\\arm_ric.phy"},
	{0x6213DE3BC6BB7E4DULL, "gen00fl_aveline6132888.cl"},
	{0x6214A8C52A5376FFULL, "789406443.ogg"},
	{0x6214E69A1310AEF3ULL, "al_rogue_t03.gda"},
	{0x62153B13DA10062CULL, "vxdr_0_deeproads_esk_0_1.vlm"},
	{0x621627D2280CC29CULL, "crt_sf_02.mao"},
	{0x62167D5393BD4879ULL, "lt_0_lowtown_docks_a3_n_19_0.anb"},
	{0x6216C18963AB650AULL, "m_dae280ar_deep_roads.wbd"},
	{0x62171DC06F8E2FA1ULL, "vxht_0_hightown_n_anh_0_0.vlm"},
	{0x621724682C1CAEBEULL, "vxdr_0_deeproads_f5e_0_0.vlm"},
	{0x6217301C3332EA12ULL, "6042992_m.fxe"},
	{0x62174B7EF8AF67C1ULL, "ht_bartrand_n_4152.rml"},
	{0x62176587EEB43075ULL, "6165344_m.fxe"},
	{0x62176985B7EEF353ULL, "mrl150_dalish_guard.bnk"},
	{0x6217899A1B02E74DULL, "ht_0_hightown_a3_25_0_lc.mmh"},
	{0x6217B7300BBF508FULL, "6125226_f.fxe"},
	{0x621831C30837175BULL, "ht_0_hightown_n_36_0_lc.mmh"},
	{0x62186E1A4D9A2495ULL, "gen_im_trs_rdr_arc_lbw_r0.uti"},
	{0x6219698F208CB151ULL, "6078424_m.fxe"},
	{0x6219F49F5C5DD3E8ULL, "dae000_flemeth6013374.cl"},
	{0x6219FAB4E2D9B5EAULL, "vxht_0_hightown_7fz_0_0.vlm"},
	{0x621A16630B66F6B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_action.gad"},
	{0x621A8B85B7F162A3ULL, "mrl150_dalish_guard.cnv"},
	{0x621ABC3D811F0162ULL, "el_lp_outside_day.mmh"},
	{0x621B7E17468BFB50ULL, "921074939.ogg"},
	{0x621BB7264810FA93ULL, "6172298_m.fxe"},
	{0x621BE5785182630FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_action.evt"},
	{0x621BF902F1CC2B04ULL, "406762847.ogg"},
	{0x621D06DE40B57CA0ULL, "vxlt_hangedman_1k3_0_0.vlm"},
	{0x621DA37B189C72B0ULL, "vxdr_primevaltomb_8xl_0_0.vlm"},
	{0x621DA6B1410BE493ULL, "mer221_sabin6048387.cl"},
	{0x621E1EAC24F97936ULL, "hlac_lt_0_lowtown_docks_n_-3_0y.dds"},
	{0x621EF256FFED0A36ULL, "w_encounter3_254_0.msh"},
	{0x621EFA155B1E2C8AULL, "cd_isa211_walleyedsam.bnk"},
	{0x621EFC56FFED1B34ULL, "w_encounter3_254_0.mmh"},
	{0x621F0056FFED2247ULL, "w_encounter3_254_0.mao"},
	{0x621F2F16AE01A7D5ULL, "6083959_f.fxe"},
	{0x621FD1928C86D44CULL, "cod_bks_dalish_history.utp"},
	{0x6220616AF2B71FF2ULL, "vxca_cavern1_haunted_hue_0_0.vlm"},
	{0x62207407B2F07EEBULL, "vxht_keep_n_79y_0_0.vlm"},
	{0x6220BA4BCDCB4068ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.cess.ani"},
	{0x62210ECC2005FD5DULL, "132846885.ogg"},
	{0x6221125FE1D25F4FULL, "6051269_m.fxe"},
	{0x622127630B6D3207ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_action.ani"},
	{0x6222292B2B7D3BCAULL, "7138319.ogg"},
	{0x622255FEEA62A625ULL, "81957629.ogg"},
	{0x6222E5387564FB4EULL, "6142083_m.fxe"},
	{0x6222E678518883B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_action.gad"},
	{0x62231F43ACD9695FULL, "gen_im_cns_cft_psn_03.uti"},
	{0x62239D3094E1524BULL, "vxht_bartrand_n_5s6_0_0.vlm"},
	{0x6223C785B16B5D0CULL, "vxlt_0_lowtown_a2_n_eup_0_0.vlm"},
	{0x622431D89AD51E9DULL, "6069611_m.fxe"},
	{0x622435614EDB39E2ULL, "gen00fl_varric6076577.cl"},
	{0x62243EB61BDB86B7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_maso_0d.dds"},
	{0x62244C520FED88CFULL, "239222630.ogg"},
	{0x62246CCF4ECFAD87ULL, "llac_lt_0_lowtown_a3_n_c_0_2.dds"},
	{0x6224853F075F19EEULL, "hlac_lt_backalley_n_0_5z.dds"},
	{0x6224CAFA0FF9C204ULL, "vxdr_0_deeproads_dq4_0_1.vlm"},
	{0x6225841B19216159ULL, "6140661_f.fxe"},
	{0x6225A90D0D1C9C16ULL, "dae000ip_crowd_ff4_talk5.utp"},
	{0x6225FFF81C00C6DDULL, "darktown_fem_fereldan_sta6152810.cl"},
	{0x6226232302281CF9ULL, "hightown_fem_courtesan_mob6169017.cl"},
	{0x6226F390FAF7E663ULL, "vxht_0_hightown_n_bxc_0_0.vlm"},
	{0x622794AE575242EFULL, "vxlt_0_lowtown_a3_n_c_i0o_0_0.vlm"},
	{0x62285C1FFE292FA9ULL, "6122763_m.fxe"},
	{0x62286EA4F4528E66ULL, "6060989_m.fxe"},
	{0x62289D834BB3F9EFULL, "6137604_m.fxe"},
	{0x6228B2C3C9201D78ULL, "combat_smr_r4_a1.gda"},
	{0x6228C46A1A7D21A0ULL, "llac_lt_0_lowtown_a2_n_c_65536_10.dds"},
	{0x622939A13BA9F632ULL, "vxlt_gallowstemplar_f_4na_0_0.vlm"},
	{0x622961049CA24153ULL, "6075146_m.fxe"},
	{0x622980C159AF2FDFULL, "vxlt_hangedman_28f_0_0.vlm"},
	{0x62299DD6DCD86499ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypouchb_sa.dds"},
	{0x6229AE451B76695CULL, "gen00fl_carver6103787.cl"},
	{0x6229E5FA9DAB8E79ULL, "dae000cr_grand_cleric6124748.cl"},
	{0x622A07851BCF0D5FULL, "character_character_p62.psh.fxo"},
	{0x622A07D7150F58A8ULL, "ave200_donnic6112169.cl"},
	{0x622A1A76C4C8CACFULL, "art\\characters\\weapons\\longswords\\w_lsw_003a.phy"},
	{0x622A29033A071313ULL, "6175398_m.fxe"},
	{0x622A3C726957C92DULL, "6018822_m.fxe"},
	{0x622A4B12294DFAB1ULL, "vxlt_0_lowtown_ex1_0_0.vlm"},
	{0x622A7AC27AB8F3C1ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.cess.ani"},
	{0x622A95F9B9895916ULL, "343324359.ogg"},
	{0x622ABEE471CA2A2EULL, "6059674_f.fxe"},
	{0x622AF5488CEF4110ULL, "vxht_keep_3im_0_0.vlm"},
	{0x622B60D68DAA8B59ULL, "hlac_w_sundermount_-393211_0x.dds"},
	{0x622B7B22BAB5FD61ULL, "hlac_lt_0_lowtown_a2_65535_6z.dds"},
	{0x622B888CBF1EBF97ULL, "fake_im_arm_cht_med_champ.uti"},
	{0x622BA90B1D2CA0D0ULL, "vxht_playershouse_n_1v4_0_0.vlm"},
	{0x622C38E39991F7E1ULL, "dre000_varric6034681.cl"},
	{0x622CB3C18894726FULL, "184783675.ogg"},
	{0x622D6DA4BCE8AEF1ULL, "art\\vfx\\combat\\vfx_gib_glv_masc.mao"},
	{0x622DE66528A058CBULL, "follower_banter6136657.cl"},
	{0x622EBB3E7F9FAD68ULL, "art\\characters\\playercharacter\\humanmale\\glv_urva.msh"},
	{0x622ED93E7F9FDF92ULL, "art\\characters\\playercharacter\\humanmale\\glv_urva.mmh"},
	{0x622F194201E891D9ULL, "vxlt_0_lowtown_a2_n_gdd_0_0.vlm"},
	{0x622FAD6FF6144DA8ULL, "vxht_0_hightown_a2_cdr_0_0.vlm"},
	{0x622FC26315E60268ULL, "273915751.ogg"},
	{0x622FEEDAEDBB7B81ULL, "dae000_arishok6081666.cl"},
	{0x6230BCEDC7A56B4EULL, "vxht_mansion_n_6hv_0_0.vlm"},
	{0x6230FEB5A82DD893ULL, "217749195.ogg"},
	{0x623185C27ABF2563ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.cess.gad"},
	{0x623252A8ACE0ABF2ULL, "vxlt_0_lowtown_a3_n_c_hwr_0_0.vlm"},
	{0x62325FBE9C7F04E1ULL, "w_bonepit_13_0.rml"},
	{0x62326A1D52E22B3CULL, "6037936_m.fxe"},
	{0x62328DB6B8F5B688ULL, "vxht_mansion_n_4zx_0_0.vlm"},
	{0x62330A65CA07757CULL, "6050070_m.fxe"},
	{0x62338BC55022C7A2ULL, "mag100_cullen6206079.cl"},
	{0x62345057909E443EULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.cess.ani"},
	{0x6235656DECEDFEE3ULL, "vxht_0_hightown_n_bcb_0_0.vlm"},
	{0x623586A319FF15A5ULL, "6082981_f.fxe"},
	{0x6235BA1D72E8B254ULL, "984411937.ogg"},
	{0x6235EC803EDAFC05ULL, "462506259.ogg"},
	{0x623615B076F685FAULL, "1029694986.ogg"},
	{0x62362DF42FDE1888ULL, "vxlt_gallowsprison_n_6kv_0_0.vlm"},
	{0x623650AFB26A11D7ULL, "vxw_blightlands_93_0_0.vlm"},
	{0x6236672B8676EF0AULL, "qcr210cr_guardsman_m.utc"},
	{0x62369C9D362C25ADULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubbucket_d.dds"},
	{0x6236A6D88942B450ULL, "w_woundedcoast_237_0.mao"},
	{0x6236AAD88942BB63ULL, "w_woundedcoast_237_0.mmh"},
	{0x6236B8D88942D2BDULL, "w_woundedcoast_237_0.msh"},
	{0x6236F910E1100DE0ULL, "vxlt_0_lowtown_a2_n_c_gz5_0_0.vlm"},
	{0x62371819AEBFAA25ULL, "mer231ip_to_undercity.utp"},
	{0x6237222E5134C5DCULL, "804911148.ogg"},
	{0x623724DCE9CDF26EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldc_n.dds"},
	{0x623755FCA3E4621DULL, "snow_t_02.mao"},
	{0x6237607DF1E95E1EULL, "qun211cr_elf_warrior6072333.cl"},
	{0x6237755375DC0C02ULL, "vxht_brothel_n_60y_0_0.vlm"},
	{0x6237F29E4BA5485FULL, "vxlt_hangedman_25a_0_0.vlm"},
	{0x62380CD848C147F0ULL, "follower_barks6153162.cl"},
	{0x6238442859171014ULL, "dae000_mother6127194.cl"},
	{0x623862C27AC50848ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.cess.evt"},
	{0x623865E7467F1085ULL, "gen00fl_aveline6120762.cl"},
	{0x6238A2D055CCEEF8ULL, "6002009_f.fxe"},
	{0x6238BCEF1B441331ULL, "vxlt_0_lowtown_es0_0_0.vlm"},
	{0x6238EF48B374F082ULL, "llac_ca_caverns1_mines_24424_26.dds"},
	{0x62397FCDBAD0A547ULL, "hlac_lt_warehouse_n_65537_1z.dds"},
	{0x623A0ED772253EC5ULL, "vxdr_0_deeproads_cua_0_2.vlm"},
	{0x623A4AEA9182B474ULL, "hlac_lt_0_lowtown_docks_n_-196610_1z.dds"},
	{0x623A6B5AEFE7B51AULL, "mag170_thrask6054031.cl"},
	{0x623A7DF4750B9FB5ULL, "6000792_f.fxe"},
	{0x623AC343359E3EC8ULL, "mrl000_marethari_act2pt1.stg"},
	{0x623AFDA8E5BCC69AULL, "darktown_fem_elf_sta6174532.cl"},
	{0x623B1D5634F57AACULL, "mag320_orsino6071173.cl"},
	{0x623B48CF172D940AULL, "6125282_m.fxe"},
	{0x623B63231AE1A699ULL, "479928439.ogg"},
	{0x623B6C7C63001A22ULL, "vxlt_0_lowtown_docks_a3_n_ctm_0_0.vlm"},
	{0x623BE8FF7DE65006ULL, "qun110_petrice6037391.cl"},
	{0x623C933CAF6CF62AULL, "208226433.ogg"},
	{0x623D51E0E9B9451DULL, "ht_keep_n_3743.rml"},
	{0x623D81A661235B75ULL, "vxlt_0_lowtown_docks_a3_cz3_0_0.vlm"},
	{0x623DFB3B516F037DULL, "qun110_petrice6018670.cl"},
	{0x623E1F4CE80A6628ULL, "vxht_0_hightown_a2_ccs_0_0.vlm"},
	{0x623F4CA6C772D376ULL, "hlac_lt_0_lowtown_a2_65533_1z.dds"},
	{0x623F70E41D694538ULL, "6034726_m.fxe"},
	{0x623FCB3DE19C48B7ULL, "dae000_serendipity6161422.cl"},
	{0x62401D5C88B60DBCULL, "vxlt_warehouse_cf_0_0.vlm"},
	{0x62404F47C5C4CB58ULL, "vxw_encounter3_fy_0_0.vlm"},
	{0x62409954A5887191ULL, "841831163.ogg"},
	{0x6240AFC77293D3E7ULL, "hlac_ht_store_187_0z.dds"},
	{0x6241685790A8F70FULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.cess.evt"},
	{0x624170D07AC201C8ULL, "qcr210cr_converted_elf.utc"},
	{0x6242018FADBB5912ULL, "vxlt_backalley_g_1wb_0_0.vlm"},
	{0x624229B5B135A771ULL, "vxdr_primevalentrance_232_0_0.vlm"},
	{0x62432CDAEB6E8C38ULL, "6167897_m.fxe"},
	{0x62439C3B7C71DA12ULL, "gen00fl_anders6125631.cl"},
	{0x6244497E9427BF64ULL, "w_bonepit_mmap_ph.dds"},
	{0x624463C942F6FDA1ULL, "636078270.ogg"},
	{0x62447AC115085991ULL, "zz_qun_debug6045410.cl"},
	{0x624492F16F2F8115ULL, "qun221_viscount6069926.cl"},
	{0x62454266C4D771CFULL, "vxw_woundedcoast_jh_0_0.vlm"},
	{0x6245AFB6B9167D41ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_dis_imp_p.impa.ani"},
	{0x6245DDB4FE4CB841ULL, "448852653.ogg"},
	{0x6245E6D58F9A197DULL, "6143296_m.fxe"},
	{0x62460FEBB6755BC6ULL, "hlac_ca_cavern1_haunted_24424_14y.dds"},
	{0x624617892471C983ULL, "lt_0_lowtown_docks_a3_n_187_0.mao"},
	{0x62461B892471D0D0ULL, "lt_0_lowtown_docks_a3_n_187_0.mmh"},
	{0x624629892471E8EAULL, "lt_0_lowtown_docks_a3_n_187_0.msh"},
	{0x624670EC2FFBE368ULL, "6107456_f.fxe"},
	{0x6246F559BA44AF64ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8mb.mmh"},
	{0x6246F959BA44B5B7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8mb.mao"},
	{0x62470B59BA44D406ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8mb.msh"},
	{0x6248695790AF17B8ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.cess.gad"},
	{0x62488ED8DC9757A5ULL, "vxlt_0_lowtown_docks_aou_0_0.vlm"},
	{0x6248F1F5CCAB29D1ULL, "6166374_m.fxe"},
	{0x624933F736767BACULL, "hlac_ht_0_hightown_a2_65534_0x.dds"},
	{0x624957BA97A2FA5CULL, "667395400.ogg"},
	{0x6249D7918AFBE787ULL, "619733871.ogg"},
	{0x6249EF8CFAD3390CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexd.msh"},
	{0x6249F58CFAD343FEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexd.mmh"},
	{0x6249F84541EE2F65ULL, "and100_anders6095896.cl"},
	{0x624A018CFAD35795ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexd.mao"},
	{0x624A4974AC657E9EULL, "follower_banter6165632.cl"},
	{0x624A63D92C06E180ULL, "zz_followers_banter6159684.cl"},
	{0x624A78C47697662DULL, "6042450_m.fxe"},
	{0x624AA9399F92E729ULL, "vxht_mansion_n_5bp_0_0.vlm"},
	{0x624ADE4C858E691FULL, "6092418_f.fxe"},
	{0x624BA8628023C2C8ULL, "6207769_f.fxe"},
	{0x624BADA3065F08E9ULL, "vxht_keep_3el_0_0.vlm"},
	{0x624C30C98AA9C27CULL, "vxlt_hangedman_n_wc_0_0.vlm"},
	{0x624C68953B814D42ULL, "6035849_f.fxe"},
	{0x624CA82052A6BBCFULL, "cd_mag160_feynriel.wbd"},
	{0x624CBAB6B91CAEE3ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_dis_imp_p.impa.gad"},
	{0x624D24A1DBAD7F14ULL, "608261711.ogg"},
	{0x624D3B2EB37E1AF0ULL, "111009185.ogg"},
	{0x624D79219F085E8FULL, "llac_lt_0_lowtown_docks_n_-262144_0.dds"},
	{0x624DC1FB53047A7FULL, "gen_im_trs_krk_low_junk2_03.uti"},
	{0x624E8F397AAC263CULL, "vxlt_warehouse_dg_0_0.vlm"},
	{0x624EB5839427FA17ULL, "zz_prmpt_debug6198258.cl"},
	{0x624EC124B7BAE3D8ULL, "vxw_encounter3_cv_0_0.vlm"},
	{0x624ECF1CE2FF21DCULL, "hlac_ht_keep_3733_1x.dds"},
	{0x624EFAAE873D7A4DULL, "vxlt_darktown_5j8_0_0.vlm"},
	{0x624F7B152C101266ULL, "6100261_m.fxe"},
	{0x624F7E5393ED0F79ULL, "lt_0_lowtown_docks_a3_n_19_0.rml"},
	{0x624F9A2C9EE78158ULL, "vxca_caverns1_ih0_0_0.vlm"},
	{0x624FA00C40D76944ULL, "6070986_m.fxe"},
	{0x6250362A6263768CULL, "139363300.ogg"},
	{0x62503810E25EDADDULL, "vxlt_0_lowtown_a2_n_c_dej_0_0.vlm"},
	{0x6250603E4212D719ULL, "57050064.ogg"},
	{0x62508185B81EE120ULL, "mrl150_dalish_guard.stg"},
	{0x625088F69C1BAE37ULL, "aip_devastatorcommander.gda"},
	{0x6250931FCFD7FA59ULL, "qcr000_arishok6059331.cl"},
	{0x62509C117E651CD8ULL, "vxht_keep_6bm_0_0.vlm"},
	{0x6250B62108297581ULL, "llac_lt_0_lowtown_a2_-65538_0.dds"},
	{0x6250BBF3A0DEB85CULL, "vxlt_0_lowtown_docks_b8f_0_0.vlm"},
	{0x6250E52EBF92158BULL, "524101772.ogg"},
	{0x6250E534376691A1ULL, "vxht_0_hightown_n_arl_0_0.vlm"},
	{0x62518AD30EDAF9A6ULL, "hlac_lt_0_lowtown_n_-131072_7z.dds"},
	{0x6251B36FD9FBAA9BULL, "zz_vault_debug6203337.cl"},
	{0x6251DB212D99743DULL, "6121992_f.fxe"},
	{0x6251FB11A5BA0786ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_act_c.anb"},
	{0x62520539F7B27383ULL, "6068293_f.fxe"},
	{0x62520B8AFFC653DBULL, "gen00fl_anders6142419.cl"},
	{0x625210D2DA27A281ULL, "hlac_ca_caverns1_mines_24421_0x.dds"},
	{0x6252167F7076D258ULL, "544413962.ogg"},
	{0x625218B83C68B835ULL, "vxht_mansion2b_318_0_0.vlm"},
	{0x62521D1166B20581ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.dura.ani"},
	{0x62531A91ABD3A818ULL, "resize.dds"},
	{0x625383C772460969ULL, "gen00fl_anders6125515.cl"},
	{0x625397B6B92291C8ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_dis_imp_p.impa.evt"},
	{0x6253D3CB6A0013F8ULL, "vxlt_0_lowtown_a2_n_c_glw_0_0.vlm"},
	{0x62544D4098DBE9ECULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_3x.dds"},
	{0x6254E611B2643019ULL, "hlac_w_blightlands_65536_1x.dds"},
	{0x6255086C1017F3E2ULL, "plasma.dds"},
	{0x62550C43A52749A8ULL, "mcr370_negotiation6080266.cl"},
	{0x62552D79BCBF9CDBULL, "6208016_m.fxe"},
	{0x6255A0A8FF304431ULL, "vxlt_0_lowtown_ei6_0_0.vlm"},
	{0x6255B97809AC0453ULL, "lt_0_lowtown_a2_n_11_0.anb"},
	{0x6255C7692CCF5638ULL, "lgt110_gxa_vgk_silver_order6116994.cl"},
	{0x62561E84040DFC14ULL, "6044523_m.fxe"},
	{0x6256A0B11BD32EC4ULL, "976427319.ogg"},
	{0x6256A1A1086D1D88ULL, "llac_dr_slavers_3216_8.dds"},
	{0x6256A1B0B1CF06DDULL, "6021209_f.fxe"},
	{0x62571206915C0E81ULL, "w_encounter3_50_0.phy"},
	{0x62572E67CF5A17FBULL, "darktown_fem_fer_urch_sta.bnk"},
	{0x6257702D838AF0C4ULL, "hlac_lt_warehouse_n_0_4x.dds"},
	{0x62579CD2158373B1ULL, "vxht_chantry_n_2yf_0_0.vlm"},
	{0x6257B4EB987243C2ULL, "l_mag100ar_encounter.bnk"},
	{0x6257F96A482A753AULL, "6187754_m.fxe"},
	{0x62580A273980218BULL, "vxw_bonepitdestroyed_a3_1rv_0_0.vlm"},
	{0x6258281166B68423ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.dura.gad"},
	{0x625867441806FD56ULL, "vxca_caverns1_mines_e7r_0_0.vlm"},
	{0x625874A203D53731ULL, "vxca_cavern1_haunted_eht_0_0.vlm"},
	{0x6258AB1C601E05C3ULL, "and100_anders6096738.cl"},
	{0x6258E96B48C0B9D3ULL, "6163216_m.fxe"},
	{0x625998FB649F5E88ULL, "zz_daggerstab.cut"},
	{0x625A536D1AA2AA57ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillarc.msh"},
	{0x625A696D1AA2CFC9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillarc.mmh"},
	{0x625A805715322836ULL, "vxca_caverns2prc_8h1_0_0.vlm"},
	{0x625AA07B0D1CF276ULL, "mcr371_scared_mage_f6086925.cl"},
	{0x625AEEC8FAD43B70ULL, "art\\characters\\playercharacter\\humanfemale\\arm_ric.msh"},
	{0x625AFCC8FAD4528AULL, "art\\characters\\playercharacter\\humanfemale\\arm_ric.mmh"},
	{0x625B2289EE1503EFULL, "mag100_cullen6082283.cl"},
	{0x625B2CE4393D3720ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_medj_0n.dds"},
	{0x625B3067CF5E03BBULL, "darktown_fem_fer_urch_sta.cnv"},
	{0x625B83FCE2D2D444ULL, "hlac_ca_caverns2prc_5802_46x.dds"},
	{0x625CC0C3861FF861ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobc.phy"},
	{0x625CE864B8B253E2ULL, "189415066.ogg"},
	{0x625D3FA1E29C3929ULL, "gen00fl_merrill6116007.cl"},
	{0x625D5EDFB9FF4A19ULL, "art\\characters\\creatures\\mubari\\animation\\pp_p_rst.ani"},
	{0x625D678F2380A702ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\8gr_galowalcove.msh"},
	{0x625D698F2380AAB8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\8gr_galowalcove.mmh"},
	{0x625D6C8B793392CDULL, "vxlt_darktown_5o7_0_0.vlm"},
	{0x625D758F2380BF5BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\8gr_galowalcove.mao"},
	{0x625D94690A6D834CULL, "105013343.ogg"},
	{0x625D9F8963E72E20ULL, "m_dae280ar_deep_roads.bnk"},
	{0x625DB7EA2017BF92ULL, "vxw_bonepitmining_a2_1j9_0_0.vlm"},
	{0x625DB8A72F071072ULL, "693477514.ogg"},
	{0x625DCC565390133BULL, "vxht_mansion2_2om_0_0.vlm"},
	{0x625DE62460D81E0DULL, "6089666_m.fxe"},
	{0x625E1B0691623C9DULL, "w_encounter3_50_0.rml"},
	{0x625E662A46FAE674ULL, "vxlt_0_lowtown_a2_n_c_eox_0_0.vlm"},
	{0x625E753F52F3AE23ULL, "character_character_p59.psh.fxo.dx11.assembly"},
	{0x625E7B2F11802EB5ULL, "145523247.ogg"},
	{0x625E86B6EEB31958ULL, "6142579_m.fxe"},
	{0x625EE50120ADAE33ULL, "821840220.ogg"},
	{0x625F051166BC6708ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.dura.evt"},
	{0x625F0C307C90E92DULL, "combat_smr_r3_a1.gda"},
	{0x625FA0AD475BFF50ULL, "6144740_m.fxe"},
	{0x625FFE11A5C64A31ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_act_c.mmh"},
	{0x626084494E00E26AULL, "618881404.ogg"},
	{0x62608AE5C76343C2ULL, "vxdr_primevaltomb_6p8_0_0.vlm"},
	{0x6260C77755D8EFCCULL, "vxht_brothel_n_4di_0_0.vlm"},
	{0x6261099F01B1F6BAULL, "weapon_weapon_p1.psh.fxo.dx11"},
	{0x62613A2140C8BBD0ULL, "hlac_lt_0_lowtown_a3_n_c_-1_3x.dds"},
	{0x626148FB6E05B2E1ULL, "vxlt_0_lowtown_a3_n_c_f9h_0_1.vlm"},
	{0x62616F25E6F385B2ULL, "665857865.ogg"},
	{0x6261A21B2EDD650DULL, "6112975_m.fxe"},
	{0x6262244549EBF1B7ULL, "hlac_ht_chantry_n_2534_7z.dds"},
	{0x6263040E7A3EDA1BULL, "art\\characters\\creatures\\highdragon\\cn_hed_hdr.mmh"},
	{0x6263100E7A3EEEB8ULL, "art\\characters\\creatures\\highdragon\\cn_hed_hdr.mao"},
	{0x6263220E7A3F0D05ULL, "art\\characters\\creatures\\highdragon\\cn_hed_hdr.msh"},
	{0x62640C5C1AE7AD9BULL, "vxlt_0_lowtown_docks_a3_cy8_0_0.vlm"},
	{0x626413EE20FD0957ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chandelier_dp.dds"},
	{0x626414724CAF765CULL, "6125802_m.fxe"},
	{0x626436EC04146307ULL, "placeable_statecontrollers.gda"},
	{0x626439DFBA052A5BULL, "art\\characters\\creatures\\mubari\\animation\\pp_p_rst.gad"},
	{0x6264481D219F8F46ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_ltrb.ani"},
	{0x62644C255793D1C7ULL, "hlac_lt_0_lowtown_docks_a3_262142_0y.dds"},
	{0x6264BD92B9FAE813ULL, "qun120_arishok6052911.cl"},
	{0x6264CE2E0B330E29ULL, "vxlt_hangedman_1qp_0_0.vlm"},
	{0x6264F4EA2B4AEA3DULL, "722960375.ogg"},
	{0x62656DC5D932417BULL, "vxht_bartrand_n_5oc_0_1.vlm"},
	{0x626598DD0D551DD4ULL, "hlac_lt_0_lowtown_a2_65536_11x.dds"},
	{0x6265ADB0873AC7C4ULL, "vxht_keep_6gn_0_0.vlm"},
	{0x6265BFA37009D818ULL, "hlac_ca_caverns2_5802_23x.dds"},
	{0x6265C23A020E4F32ULL, "675400307.ogg"},
	{0x6265D6FBDF677E6FULL, "dae200cr_orsino.utc"},
	{0x6265E0891B2D02F1ULL, "vxht_0_hightown_a2_bye_0_0.vlm"},
	{0x6265F87EE8FFD445ULL, "6134997_m.fxe"},
	{0x62660EAF07798C31ULL, "vxht_chantry_n_2te_0_0.vlm"},
	{0x62661A7858C60B1BULL, "vxca_caverns2_7uq_0_0.vlm"},
	{0x626667F89C83FE3BULL, "6218602_m.fxe"},
	{0x6266DE698619D565ULL, "ser000pt_debug.ncs"},
	{0x6267139FD4726315ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_3windows.phy"},
	{0x6267360A2DFFBB30ULL, "mcr310_confused_guard6077114.cl"},
	{0x62678CF3837F1829ULL, "vxht_mansion_n_5hv_0_0.vlm"},
	{0x6267E716931B8C09ULL, "hlac_lt_0_lowtown_a3_n_c_-131071_1y.dds"},
	{0x6268626CD12C469CULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.ani"},
	{0x6268CFA6DDE16BEAULL, "385556596.ogg"},
	{0x6268EAD75637A360ULL, "14574449.ogg"},
	{0x6269894734246773ULL, "vxlt_0_lowtown_docks_a3_bmy_0_0.vlm"},
	{0x6269B6451758CB79ULL, "hlac_lt_darktown_-1_53y.dds"},
	{0x6269BC971E622841ULL, "6049741_m.fxe"},
	{0x6269BCAACE406BD3ULL, "ave310_donnic6134755.cl"},
	{0x6269C94B21552819ULL, "mag291cr_crowd.ncs"},
	{0x626A27C407891900ULL, "6202794_m.fxe"},
	{0x626A492E09616BE4ULL, "vxht_0_hightown_a3_c7j_0_0.vlm"},
	{0x626AF3DAB3116817ULL, "w_bonepitmining_a2_136_0.phy"},
	{0x626B56DFBA0B7A40ULL, "art\\characters\\creatures\\mubari\\animation\\pp_p_rst.evt"},
	{0x626B6273E9B7D7CBULL, "840758388.ogg"},
	{0x626B72F35E98573CULL, "vxlt_warehouse_re_0_0.vlm"},
	{0x626BDE686B29AB4DULL, "vxlt_darktown_5p6_0_0.vlm"},
	{0x626C12A285D06071ULL, "vxlt_hangedman_2eo_0_0.vlm"},
	{0x626C18A09469A0BCULL, "hlac_ht_chantry_n_2534_51x.dds"},
	{0x626C3C82DF041114ULL, "64261649.ogg"},
	{0x626C4EFFD027283BULL, "hlac_ca_caverns2prc_5802_53y.dds"},
	{0x626CBEDDE9666827ULL, "vxlt_gallowscourtyard_4lk_0_0.vlm"},
	{0x626CCAB203E33318ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_flat_1.mao"},
	{0x626D7FCB62514DD8ULL, "vxht_keep_6hk_0_0.vlm"},
	{0x626DC92E369D4CBBULL, "hlac_w_bonepit_65541_0x.dds"},
	{0x626DE8453B21B6B7ULL, "vxht_brothel_n_6d4_0_0.vlm"},
	{0x626E1880E1061727ULL, "ave121_brennan6105802.cl"},
	{0x626EFCC2B9595C42ULL, "vxdr_primevaltomb_6w9_0_0.vlm"},
	{0x626EFDD020FF722EULL, "gen00fl_fenris6110808.cl"},
	{0x626F410E18929969ULL, "6039306_m.fxe"},
	{0x626F5B4E415D6980ULL, "lowtown_male_lowlife1_sta6164990.cl"},
	{0x626F6F6CD1327BDEULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.gad"},
	{0x626F804305850EB0ULL, "hlac_ht_keep_n_3732_14z.dds"},
	{0x626F8F33E4665520ULL, "6117779_m.fxe"},
	{0x626FA061B98327B1ULL, "315736167.ogg"},
	{0x626FA5A3A3DAE530ULL, "6087679_f.fxe"},
	{0x626FAFBD9DD821EFULL, "hlac_lt_undercityserialkiller_-1_0y.dds"},
	{0x626FD96D398E70A6ULL, "hlac_lt_0_lowtown_-196609_2y.dds"},
	{0x627097FD9AB9BB4FULL, "vxw_woundedcoast_yu_0_0.vlm"},
	{0x6270AE654ED59311ULL, "vxlt_0_lowtown_a2_n_c_i0p_0_0.vlm"},
	{0x6271C9C318FB2B48ULL, "223089505.ogg"},
	{0x6271D4007FC31F44ULL, "gen00fl_aveline6132818.cl"},
	{0x6271D57D80284636ULL, "383412610.ogg"},
	{0x6271DC0F97112003ULL, "vxlt_0_lowtown_n_d8u_0_0.vlm"},
	{0x6271F82C4AB7784BULL, "6153659_m.fxe"},
	{0x62720930C7F975BBULL, "vxw_bonepitdestroyed_a3_179_0_0.vlm"},
	{0x627265B71F70B23FULL, "gen00fl_anders6142503.cl"},
	{0x6272D2D886F8B71CULL, "133998193.ogg"},
	{0x6272F899F06E7439ULL, "6020469_m.fxe"},
	{0x62732F5BB039CFEFULL, "hlac_ca_caverns2prc_5802_6y.dds"},
	{0x62733D9743B5779EULL, "6060505_m.fxe"},
	{0x627351BE9CB683D0ULL, "w_bonepit_13_0.msh"},
	{0x62735FBE9CB69BEAULL, "w_bonepit_13_0.mmh"},
	{0x627363BE9CB6A221ULL, "w_bonepit_13_0.mao"},
	{0x627378204C6D148CULL, "gen00fl_isabela6101703.cl"},
	{0x6273B5296F5BC0CFULL, "vxw_bonepit_1qc_0_0.vlm"},
	{0x6273CCA62EE0A21DULL, "vxw_bonepitmining_a2_26y_0_0.vlm"},
	{0x6274A70A52FC9E44ULL, "6065030_m.fxe"},
	{0x6274BDBF761C0AAEULL, "ave310_donnic6062077.cl"},
	{0x6274CBD9761BBC9EULL, "128813575.ogg"},
	{0x6275258D8B8F87D9ULL, "6146831_m.fxe"},
	{0x627560FB7C62EAEFULL, "vxlt_0_lowtown_docks_a3_d3i_0_0.vlm"},
	{0x627599864B40D168ULL, "hlac_dr_primevalentrance_1726_6x.dds"},
	{0x6275D220ACE0FEF2ULL, "vxlt_undercitydungeon_1ma_0_0.vlm"},
	{0x627666AC84025A7FULL, "hlac_lt_0_lowtown_docks_a3_131070_0x.dds"},
	{0x6276ADD5132E3186ULL, "12576160.ogg"},
	{0x6277715488AE877FULL, "vfxchunkydd.psh.fxo"},
	{0x62783039080095A4ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.anb"},
	{0x62789B2BD355ABA6ULL, "gen00fl_isabela6089588.cl"},
	{0x62789BCC22FED973ULL, "vxca_cavern1_haunted_gk1_0_0.vlm"},
	{0x6279311D21B1C440ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_ltrb.gad"},
	{0x627935715826A049ULL, "429374826.ogg"},
	{0x62796227E1790EEDULL, "vxht_0_hightown_a3_bgn_0_0.vlm"},
	{0x6279E4D0508E6FBCULL, "vxlt_warehouse_wz_0_0.vlm"},
	{0x6279ECB1539AB286ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm6.ani"},
	{0x627A477136343F33ULL, "dae000ip_crowd_em3_talk1.utp"},
	{0x627A555D92014509ULL, "vfx_lavajet_med.wbd"},
	{0x627AB010377C443BULL, "vxht_mansion2_2uo_0_0.vlm"},
	{0x627AC9DE44C44F0DULL, "6087466_m.fxe"},
	{0x627AD890185E202DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masd_0n.dds"},
	{0x627AF4C2E1539349ULL, "vxw_bonepit_267_0_0.vlm"},
	{0x627B064DC5FD34B4ULL, "qun110_petrice6042289.cl"},
	{0x627C06EF0BCF9314ULL, "265433443.ogg"},
	{0x627C1CBA9F0C5CACULL, "hlac_lt_0_lowtown_a2_n_65535_7y.dds"},
	{0x627C3739667161CBULL, "6089769_m.fxe"},
	{0x627C4297070800F2ULL, "dae000_madam_lusine6152079.cl"},
	{0x627D30073BE1FFEDULL, "613272109.ogg"},
	{0x627D9EEB9892EDE0ULL, "l_mag100ar_encounter.wbd"},
	{0x627DDDF04D12B735ULL, "vxca_caverns1_mines_j2m_0_0.vlm"},
	{0x627EAE799D4A051CULL, "6041259_f.fxe"},
	{0x627F8E2E93151290ULL, "109407175.ogg"},
	{0x62806A80904C553AULL, "160074929.ogg"},
	{0x628073A714C8D1A7ULL, "198744156.ogg"},
	{0x62807B0DB9EF8E3BULL, "vxw_bonepitdestroyed_a3_188_0_0.vlm"},
	{0x6280F2F7D7A63851ULL, "mrl250cr_spiderhunt.utc"},
	{0x62810A2D9090FF84ULL, "zz_qun_debug6028232.cl"},
	{0x628126FCDD86B91EULL, "hlac_lt_0_lowtown_a2_n_-196607_0z.dds"},
	{0x6281D67809D1700FULL, "lt_0_lowtown_a2_n_11_0.rml"},
	{0x62821DAD6CFB06C6ULL, "vxht_bartrand_n_5rl_0_0.vlm"},
	{0x628227066151D94FULL, "vxw_bonepit_1zb_0_0.vlm"},
	{0x628229B85FB25445ULL, "hlac_lt_0_lowtown_-262144_4x.dds"},
	{0x62825F9D07417D89ULL, "hlac_w_blightlands_65535_0z.dds"},
	{0x6282B6F26EDF5416ULL, "persistent.gda"},
	{0x6282C75FDAB215ABULL, "crt_dragon.wbd"},
	{0x6282E24C09409D78ULL, "648770880.ogg"},
	{0x6283D2D86E59036FULL, "vxlt_0_lowtown_docks_a3_d4h_0_0.vlm"},
	{0x6283EA0CA4F88ADDULL, "6145110_f.fxe"},
	{0x6284771982392507ULL, "aura.mao"},
	{0x6284782C9FF7D35FULL, "346143927.ogg"},
	{0x6284F441952619E3ULL, "vxw_encounter3_7u_0_0.vlm"},
	{0x628541D27B12BD3AULL, "vxca_caverns1_mines_efy_0_0.vlm"},
	{0x62857226992078DCULL, "6077842_m.fxe"},
	{0x6285FC3A4B443C3DULL, "zz_followers_banter6158604.cl"},
	{0x62864F39080D07A7ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.mmh"},
	{0x628780F8D9B6BA16ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2.ani"},
	{0x62879600C0F721B7ULL, "149295524.ogg"},
	{0x6287A2D1A10FB00EULL, "vxlt_warehouse_n_10w_0_0.vlm"},
	{0x6287A304E1EC7F70ULL, "hlac_lt_0_lowtown_a2_n_-131072_7z.dds"},
	{0x6287F2F3E700B113ULL, "hlac_lt_0_lowtown_a3_n_c_-3_4x.dds"},
	{0x628801C8FA952185ULL, "6131943_m.fxe"},
	{0x628810EADE53AB29ULL, "vxlt_gallowsprison_n_5qz_0_0.vlm"},
	{0x62881FA9BACA2CC7ULL, "dae300_alistair_king6133513.cl"},
	{0x6289578188981D81ULL, "hlac_ca_caverns1_24426_11x.dds"},
	{0x6289669FD349ABC9ULL, "vxw_bonepit_216_0_0.vlm"},
	{0x62898FCF46656A69ULL, "412770110.ogg"},
	{0x6289B9098B30F6D6ULL, "854709465.ogg"},
	{0x6289D51B149AB1DFULL, "6205969_m.fxe"},
	{0x6289E64F4BB0DB16ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\taunt.ani"},
	{0x628A3A98C894F59DULL, "6090408_m.fxe"},
	{0x628AC602CBDBA39CULL, "vxdr_dungeon_1t7_0_0.vlm"},
	{0x628ACBFF1F0DE7B7ULL, "vxht_brothel_n_6j6_0_0.vlm"},
	{0x628ACC8CFB0A9499ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexd.phy"},
	{0x628AD83E7FEE3C8DULL, "art\\characters\\playercharacter\\humanmale\\glv_urva.phy"},
	{0x628B63B945227944ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal_shorter.ani"},
	{0x628B94EA0B969309ULL, "vxlt_gallowstemplar_f_3ri_0_0.vlm"},
	{0x628BB55D622CBAD9ULL, "vxlt_0_lowtown_docks_n_c1f_0_0.vlm"},
	{0x628BE72678A12F27ULL, "hlac_w_sundermount_-327681_0z.dds"},
	{0x628C899C447E73C7ULL, "6071704_m.fxe"},
	{0x628CFDED8DB5C0E3ULL, "69353072.ogg"},
	{0x628D508A83BA4BA0ULL, "vxht_0_hightown_a2_bri_0_0.vlm"},
	{0x628D7BB77EA5EC4FULL, "vxw_woundedcoast_sw_0_0.vlm"},
	{0x628DA6B8A2870552ULL, "vxht_keep_n_5vg_0_0.vlm"},
	{0x628DC22052DE7EA5ULL, "cd_mag160_feynriel.bnk"},
	{0x628DD4D450D2877BULL, "378638215.ogg"},
	{0x628DD5B153AB3480ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm6.gad"},
	{0x628E74B92EDDD6E2ULL, "6207662_f.fxe"},
	{0x628F7FF5CF6C55A8ULL, "vxlt_0_lowtown_docks_a3_ctd_0_0.vlm"},
	{0x628F947B15560A92ULL, "gen00fl_aveline6053317.cl"},
	{0x628F9975A5482A9BULL, "6048256_m.fxe"},
	{0x628FA2D9DEE04D77ULL, "vxht_mansion2_2oy_0_0.vlm"},
	{0x628FEE28D300404FULL, "gen00fl_varric6118973.cl"},
	{0x629048DAA6E2BE03ULL, "51890406.ogg"},
	{0x629084234ECEE2EFULL, "hlac_lt_0_lowtown_-196608_1z.dds"},
	{0x62908F8A5EF11F46ULL, "vxht_bartrand_n_5um_0_0.vlm"},
	{0x6290A8E47841E08AULL, "6035132_m.fxe"},
	{0x6290DEC798CA942BULL, "zz_getplot.ncs"},
	{0x6290E22F6479E49DULL, "6082196_f.fxe"},
	{0x62916445DD5BD5B1ULL, "vxht_chantry_n_2ez_0_0.vlm"},
	{0x629182C4E20833A3ULL, "37857357.ogg"},
	{0x62919F880AB39917ULL, "351646590.ogg"},
	{0x6291BD4981927884ULL, "672550487.ogg"},
	{0x6291C4D3124FFBFDULL, "gen00fl_bethany.utc"},
	{0x6291CEBE24640516ULL, "832723461.ogg"},
	{0x6291E5788FBDAE80ULL, "vxw_woundedcoast_o0_0_0.vlm"},
	{0x6291FEC4E6A0FC7DULL, "117382726.ogg"},
	{0x629207C624E666CEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic_1.mao"},
	{0x62924442EB9D25D9ULL, "vxlt_0_lowtown_a3_n_c_eo9_0_1.vlm"},
	{0x629260B9452892A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal_shorter.gad"},
	{0x629276D5E4F6FA19ULL, "vxw_encounter3_42_0_0.vlm"},
	{0x629441AF2A8F0EB8ULL, "vxca_caverns1_mines_iin_0_0.vlm"},
	{0x6294D46ED6312FBAULL, "hlac_ht_mansion2_818_44x.dds"},
	{0x6294F498697D9AA2ULL, "6182443_m.fxe"},
	{0x62952E7F7E273C74ULL, "6051060_m.fxe"},
	{0x62955D93C2A45487ULL, "359939597.ogg"},
	{0x629588F8D9C30507ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2.evt"},
	{0x6297124D0A3A893FULL, "6165723_f.fxe"},
	{0x6297E4A712C9BEEDULL, "vxht_keep_58a_0_0.vlm"},
	{0x6297EE4F4BBD2607ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\taunt.evt"},
	{0x6298A0409A27A5E2ULL, "vxlt_shop_5q_0_0.vlm"},
	{0x6299095296FB97EBULL, "21214052.ogg"},
	{0x62995BB9452EA885ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal_shorter.evt"},
	{0x6299B338E18BFA28ULL, "319868999.ogg"},
	{0x6299ECC75AC2F098ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic.phy"},
	{0x6299FDE4611C902BULL, "vxlt_gallowstemplar_f_5po_0_0.vlm"},
	{0x629A3E7F71571DF3ULL, "gen00fl_dog6167679.cl"},
	{0x629A752C50EB80BBULL, "6068569_f.fxe"},
	{0x629A8EEB1DB151CCULL, "vxht_brothel_n_4xm_0_0.vlm"},
	{0x629AD3754C754B62ULL, "vxca_cavern1_haunted_iyd_0_0.vlm"},
	{0x629AF1E5C6F1C306ULL, "2789334.ogg"},
	{0x629B285CF6899B95ULL, "art\\vfx\\combat\\vfx_sg_blast_act_c.mmh"},
	{0x629B612D5E8EB6B4ULL, "hlac_lt_blackemporium_1055_0x.dds"},
	{0x629B690FE5C001DCULL, "llac_lt_gallowscourtyard_n_-65535_2.dds"},
	{0x629C3EBA8AD14271ULL, "643357777.ogg"},
	{0x629C43A4FF2BE9FEULL, "follower_banter6155994.cl"},
	{0x629C6EB4C86B323FULL, "vxht_mansion2b_1kj_0_0.vlm"},
	{0x629C7AF083F9357BULL, "mer221_sabin6047427.cl"},
	{0x629C7FC12E6F7422ULL, "gender.gda"},
	{0x629C843AB8FE097CULL, "rand_plains.gda"},
	{0x629CA9F8D9C95C50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2.gad"},
	{0x629DF5C16A0613B0ULL, "vxw_woundedcoast_hu_0_0.vlm"},
	{0x629E1C3BECB18B06ULL, "834718317.ogg"},
	{0x629E8526788B9329ULL, "static_static_p31.psh.fxo"},
	{0x629F016750E737C6ULL, "vxht_bartrand_n_5xn_0_0.vlm"},
	{0x629F0F16CA53F685ULL, "vxlt_0_lowtown_a2_n_c_gbl_0_0.vlm"},
	{0x629F0F4F4BC37D50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\taunt.gad"},
	{0x629F23293ECB55BAULL, "406698506.ogg"},
	{0x629F275D922013EBULL, "vfx_lavajet_med.bnk"},
	{0x629F5A6149402490ULL, "vxlt_0_lowtown_n_f4t_0_0.vlm"},
	{0x629FA47C1C44F398ULL, "war_shieldbash.dds"},
	{0x629FDCA0BEB684BAULL, "vxlt_gallowstemplar_2lj_0_0.vlm"},
	{0x62A0575581B3C700ULL, "vxw_woundedcoast_d1_0_0.vlm"},
	{0x62A0DA6783E46472ULL, "6081676_m.fxe"},
	{0x62A0E8B2D6ED1299ULL, "vxw_encounter3_11_0_0.vlm"},
	{0x62A16AF416C4B97CULL, "vxht_mansion_n_5cj_0_0.vlm"},
	{0x62A1F598847893A9ULL, "gen00fl_isabela6099359.cl"},
	{0x62A20C44E7F9370EULL, "gr_sp_02.mao"},
	{0x62A283103E65AB55ULL, "558133280.ogg"},
	{0x62A2FB1577C7C3A3ULL, "6147785_f.fxe"},
	{0x62A32FC16364C546ULL, "llac_lt_0_lowtown_docks_a3_-65539_4.dds"},
	{0x62A36774306B0F35ULL, "921953123.ogg"},
	{0x62A383D5B517AD38ULL, "dae100_bran6142994.cl"},
	{0x62A3D992384F92DEULL, "117876657.ogg"},
	{0x62A431CFA6D48518ULL, "6182207_m.fxe"},
	{0x62A4AB65EA3F5356ULL, "239624570.ogg"},
	{0x62A55119E90A03CCULL, "vxlt_0_lowtown_a2_h2g_0_0.vlm"},
	{0x62A554646362431AULL, "vxlt_0_lowtown_ey1_0_0.vlm"},
	{0x62A622B426B1FAAAULL, "art\\vfx\\combat\\vfx_r_ibl_fore_act_c.anb"},
	{0x62A6F2CDCA9EC281ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_enter.ani"},
	{0x62A72410BC1B5182ULL, "6205645_m.fxe"},
	{0x62A7A95FDAD0FF79ULL, "crt_dragon.bnk"},
	{0x62A8303C22FD595AULL, "zz_mcr_debug6080780.cl"},
	{0x62A8C658DBE89883ULL, "6178447_m.fxe"},
	{0x62A8E99C27B7C803ULL, "lt_0_lowtown_docks_a3_100_0.mao"},
	{0x62A8ED9C27B7CF50ULL, "lt_0_lowtown_docks_a3_100_0.mmh"},
	{0x62A8FB9C27B7E76AULL, "lt_0_lowtown_docks_a3_100_0.msh"},
	{0x62A9255CF695D422ULL, "art\\vfx\\combat\\vfx_sg_blast_act_c.anb"},
	{0x62A957261474ACF4ULL, "6079798_m.fxe"},
	{0x62A9C2D8A828E22DULL, "character_character_p231.psh.fxo.dx11.assembly"},
	{0x62AA250E7A7B1550ULL, "art\\characters\\creatures\\highdragon\\cn_hed_hdr.phy"},
	{0x62AA31DF58E57B87ULL, "gen00fl_fenris6106725.cl"},
	{0x62AA4A657DF4CF61ULL, "hlac_w_encounter2_-65531_0z.dds"},
	{0x62AB842402484E94ULL, "773142897.ogg"},
	{0x62ABF5FCD377C891ULL, "hlac_w_sundermount_-196605_0z.dds"},
	{0x62AC5D74E0875E6DULL, "565154412.ogg"},
	{0x62ACB47EDE1BA153ULL, "6106661_f.fxe"},
	{0x62ACFDCDCAA34123ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_enter.gad"},
	{0x62AD1FC204D7CE6AULL, "vxht_0_hightown_avl_0_0.vlm"},
	{0x62AD8C66BE9820E2ULL, "6092522_m.fxe"},
	{0x62AE269FD4AE8630ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_3windows.msh"},
	{0x62AE349FD4AE9E4AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_3windows.mmh"},
	{0x62AE389FD4AEA481ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_3windows.mao"},
	{0x62AEAADFDF41FFE3ULL, "vxlt_0_lowtown_a2_ewr_0_0.vlm"},
	{0x62AEF6CCFDF81542ULL, "6144324_m.fxe"},
	{0x62AF3151BEA3B3ABULL, "mag291_torpor6026951.cl"},
	{0x62AFB42BC81B3910ULL, "gen_im_wep_war_shd_act1_05.uti"},
	{0x62AFBEB35413EB53ULL, "hlac_lt_0_lowtown_docks_a3_-196611_1x.dds"},
	{0x62B03B594EA125ADULL, "qun131cr_warrens_thug6.utc"},
	{0x62B082E843105162ULL, "vxht_mansion2_2p9_0_0.vlm"},
	{0x62B105E5D75D2041ULL, "468841426.ogg"},
	{0x62B125690E7B3FB8ULL, "vxca_caverns1_mines_icl_0_0.vlm"},
	{0x62B1323403CF8D8DULL, "active.ncs"},
	{0x62B18C0494ADAA1EULL, "zz_mage_whiteout.cut"},
	{0x62B19EC76BF64E2CULL, "hlac_lt_0_lowtown_docks_n_-196608_0x.dds"},
	{0x62B1C8DAB34D226FULL, "w_bonepitmining_a2_136_0.mao"},
	{0x62B1D48AE1990654ULL, "art\\characters\\creatures\\dragonling\\textures\\c_dragling_0s.dds"},
	{0x62B1D4DAB34D360CULL, "w_bonepitmining_a2_136_0.mmh"},
	{0x62B1DADAB34D40FEULL, "w_bonepitmining_a2_136_0.msh"},
	{0x62B2A1C3190BA33CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_table_01_1_d.dds"},
	{0x62B2FED2B8CD89E1ULL, "zz_dog_debug6127730.cl"},
	{0x62B35A61BDDD5792ULL, "ico_mace_3.dds"},
	{0x62B3C5A729CE98E7ULL, "1064732170.ogg"},
	{0x62B3DACDCAA92408ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_enter.evt"},
	{0x62B4030A0F142F88ULL, "102917369.ogg"},
	{0x62B48458D0E53AD9ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_pot.msh"},
	{0x62B48E58D0E54BC7ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_pot.mmh"},
	{0x62B49258D0E55214ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_pot.mao"},
	{0x62B5E00BBC7E0A82ULL, "mag320_noblewoman_2.cnv"},
	{0x62B5FE81F17DFA64ULL, "plt_zz_prmpt_debug__.nss"},
	{0x62B675BEB69D4439ULL, "vxdr_primevaltomb_3tf_0_0.vlm"},
	{0x62B6AEFD29123EAEULL, "llac_lt_gallowstemplar_n_7198_6.dds"},
	{0x62B6D768E2E3F299ULL, "80488506.ogg"},
	{0x62B6F0BFDDD5ACC3ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_injured_1_ian.ani"},
	{0x62B702267F9984E1ULL, "vxlt_0_lowtown_a2_csf_0_0.vlm"},
	{0x62B70DFEA2A4E244ULL, "106949490.ogg"},
	{0x62B72185164675F2ULL, "117672549.ogg"},
	{0x62B72D95E18173A4ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryglow.mmh"},
	{0x62B76330BD443F8BULL, "vxca_caverns2prc_98a_0_0.vlm"},
	{0x62B880F9D1DECA1CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0cc_basic_1.mao"},
	{0x62B8B18491E5F539ULL, "475079382.ogg"},
	{0x62B8EB49C8B7EA75ULL, "6083400_m.fxe"},
	{0x62B933D40FF7D1C4ULL, "ico_boots_heavy_2.dds"},
	{0x62B93F4BCCF298E9ULL, "577415439.ogg"},
	{0x62B96995EAEEA806ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.impa.ani"},
	{0x62B99E198DC581E5ULL, "art\\levels\\races\\proxy\\_textures\\ter_1_d.dds"},
	{0x62B9C20BBC81C0BAULL, "mag320_noblewoman_2.bnk"},
	{0x62B9C68F372B69BFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02b.phy"},
	{0x62B9E303746F9A42ULL, "art\\characters\\weapons\\axes\\w_axe_001.msh"},
	{0x62B9E503746F9DF8ULL, "art\\characters\\weapons\\axes\\w_axe_001.mmh"},
	{0x62BA06766657ECB4ULL, "llac_ca_cavern1_haunted_24424_17.dds"},
	{0x62BA170691B094BDULL, "w_encounter3_50_0.mao"},
	{0x62BA1B0691B09BF6ULL, "w_encounter3_50_0.mmh"},
	{0x62BA250691B0ACF4ULL, "w_encounter3_50_0.msh"},
	{0x62BA5773D721AA0AULL, "gen_im_trs_apo_lgt_cht_r0.uti"},
	{0x62BA715F548174C7ULL, "qun120_arishok6051680.cl"},
	{0x62BAA75B95215E48ULL, "6085739_m.fxe"},
	{0x62BAB6A9304E4B70ULL, "vxlt_warehouse_n_df_0_0.vlm"},
	{0x62BAD97B4DF244B0ULL, "vxw_woundedcoast_bs_0_0.vlm"},
	{0x62BAEE0E5C8298DCULL, "llac_w_woundedcoast_-196604_0.dds"},
	{0x62BAF51B482D5742ULL, "vxht_keep_n_30k_0_0.vlm"},
	{0x62BBEACC89BC4F24ULL, "character_character_p65.psh.fxo"},
	{0x62BBFE43B08E3962ULL, "6097622_m.fxe"},
	{0x62BC1C6AF272BE98ULL, "vxht_bartrand_n_65x_0_0.vlm"},
	{0x62BC8B38BD61906DULL, "zz_vault_debug6177786.cl"},
	{0x62BD36862D1BFFDEULL, "llac_lt_0_lowtown_-65539_0.dds"},
	{0x62BD62879DC3AE75ULL, "445151412.ogg"},
	{0x62BD6677B8153B59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_gesture_right_hand.ani"},
	{0x62BD9357921133DDULL, "llac_ht_keep_n_3732_2.dds"},
	{0x62BDCA515F07E37BULL, "fen110_fenris6118037.cl"},
	{0x62BDED76A90AC8D3ULL, "vxht_keep_n_5jf_0_0.vlm"},
	{0x62BE39265200692AULL, "dae300_alistair_king6133626.cl"},
	{0x62BEA3B9E7DA52D3ULL, "117343583.ogg"},
	{0x62BED83AC43FF2A0ULL, "vxlt_hangedman_2an_0_0.vlm"},
	{0x62BED8F711E7AE60ULL, "6126375_m.fxe"},
	{0x62BF34677086E9BDULL, "vxht_keep_n_685_0_0.vlm"},
	{0x62BF431246CF92A2ULL, "zz_vault_debug6178926.cl"},
	{0x62BF5703428FD926ULL, "216448690.ogg"},
	{0x62BFC5C386747D9DULL, "art\\characters\\playercharacter\\humanmale\\cth_nobc.mao"},
	{0x62BFC9C3867484D6ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobc.mmh"},
	{0x62BFD3C3867495D4ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobc.msh"},
	{0x62C0491C932E210BULL, "6117246_m.fxe"},
	{0x62C06538BB38F541ULL, "vxw_encounter2_camp_ph_0_1.vlm"},
	{0x62C07059B44F5899ULL, "vxht_mansion_n_61j_0_0.vlm"},
	{0x62C07C4638655A7BULL, "hm_gen_apostate.mor"},
	{0x62C099BE53957563ULL, "186096506.ogg"},
	{0x62C0ED0ECF976E94ULL, "gen00fl_dog6166390.cl"},
	{0x62C118071E4D5CCAULL, "hlac_lt_undercitydungeon_-1_15y.dds"},
	{0x62C148D9C2E7880AULL, "27260579.ogg"},
	{0x62C1C4DD900C3009ULL, "qun110_petrice6018731.cl"},
	{0x62C21D4BAA72F04DULL, "fen210_tevinter_ambusher6108295.cl"},
	{0x62C22F626485D588ULL, "364528646.ogg"},
	{0x62C236F9E613F7DBULL, "mag321_comtesse_de_launcet6064706.cl"},
	{0x62C287788F644414ULL, "30902763.ogg"},
	{0x62C2B599B272CEF1ULL, "6124769_m.fxe"},
	{0x62C2DDE51222FCB2ULL, "86069331.ogg"},
	{0x62C32495664D117EULL, "698817308.ogg"},
	{0x62C38A5C3694E98CULL, "pro000_captain_wright6047478.cl"},
	{0x62C3A4C52F90A044ULL, "372405308.ogg"},
	{0x62C3EC037477FD83ULL, "art\\characters\\weapons\\axes\\w_axe_001.phy"},
	{0x62C436A116A736D7ULL, "6125108_m.fxe"},
	{0x62C44177B81B1A9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_gesture_right_hand.gad"},
	{0x62C454C56F796725ULL, "dae000_mother6136955.cl"},
	{0x62C4748F6C972DC4ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_db_c.mmh"},
	{0x62C48F055CD8E054ULL, "llac_ht_chantry_n_2534_26.dds"},
	{0x62C49AFD18DE1E91ULL, "6149279_f.fxe"},
	{0x62C4B14417B6D88CULL, "121851905.ogg"},
	{0x62C4BFB2CB0D44DDULL, "6125524_m.fxe"},
	{0x62C4C70645E61AE0ULL, "lt_0_lowtown_28_0_ll.mmh"},
	{0x62C5BEC27AA2E19AULL, "6131077_f.fxe"},
	{0x62C5DF7942C54C52ULL, "qun220cr_javaris6071510.cl"},
	{0x62C5EE326DCFB2DBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05a.msh"},
	{0x62C60C326DCFE5C5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05a.mmh"},
	{0x62C656A2448CC680ULL, "535249949.ogg"},
	{0x62C75195EAFABD77ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.impa.evt"},
	{0x62C76AF3270D5968ULL, "6077275_m.fxe"},
	{0x62C76E2C6A5F6752ULL, "vxht_keep_n_5jc_0_0.vlm"},
	{0x62C7E398020D015FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masc_0d.dds"},
	{0x62C855519AA24BE0ULL, "vxht_chantry_23n_0_0.vlm"},
	{0x62C85AE827A87B71ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2a.msh"},
	{0x62C868E827A8928CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2a.mao"},
	{0x62C874E827A8A72FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2a.mmh"},
	{0x62C896208131ACD7ULL, "art\\levels\\races\\proxy\\props\\_textures\\tables_n.dds"},
	{0x62C8ADA8265DC472ULL, "6089823_m.fxe"},
	{0x62C8C49CA75F301CULL, "698769243.ogg"},
	{0x62C8C9C09A14E50EULL, "6050332_f.fxe"},
	{0x62C8D4DAC3BF035BULL, "hlac_lt_0_lowtown_n_1_1y.dds"},
	{0x62C8F0CC152B9CA4ULL, "6133522_m.fxe"},
	{0x62C910BF87DA4344ULL, "115599240.ogg"},
	{0x62C94393F35A1750ULL, "gen00fl_aveline6139709.cl"},
	{0x62C9FFD3E5FE30C6ULL, "gen00fl_sebastian6188688.cl"},
	{0x62CA1F2EDFE0061AULL, "vxca_cave_1c7_0_0.vlm"},
	{0x62CAAFF5D53DCF85ULL, "116011458.ogg"},
	{0x62CAC78F3739DEC4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02b.mmh"},
	{0x62CADD8F373A0466ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02b.msh"},
	{0x62CB07BFDDE67CADULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_injured_1_ian.gad"},
	{0x62CB3CA6F7FE18D1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mb.mmh"},
	{0x62CB48A6F7FE2D72ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mb.mao"},
	{0x62CB56A6F7FE448FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mb.msh"},
	{0x62CB5E77B8216A80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_gesture_right_hand.evt"},
	{0x62CB993953A1AFB2ULL, "ht_0_hightown_12_0_lc.mmh"},
	{0x62CC63F552279740ULL, "6188624_f.fxe"},
	{0x62CC6884148A62B0ULL, "vxw_bonepit_1sj_0_0.vlm"},
	{0x62CCB5C061F1B897ULL, "722891389.ogg"},
	{0x62CCC08AECA702FCULL, "vxht_mansion_n_5ri_0_0.vlm"},
	{0x62CD0649393CDC26ULL, "6177854_m.fxe"},
	{0x62CD5295EAFF2A00ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.impa.gad"},
	{0x62CDA644627D023DULL, "vxht_keep_n_634_0_0.vlm"},
	{0x62CE23389AE04107ULL, "gen00fl_bethany6169919.cl"},
	{0x62CE4E7CDC11A8B2ULL, "vxht_keep_67x_0_0.vlm"},
	{0x62CEC81F16EE75EEULL, "w_encounter2_4_0.msh"},
	{0x62CECAD31283B398ULL, "gen00fl_bethany.cnv"},
	{0x62CED61F16EE8D3FULL, "w_encounter2_4_0.mao"},
	{0x62CED7B2DF41B4C7ULL, "6165733_m.fxe"},
	{0x62CEE21F16EEA1DCULL, "w_encounter2_4_0.mmh"},
	{0x62CF25B426D5049DULL, "art\\vfx\\combat\\vfx_r_ibl_fore_act_c.mmh"},
	{0x62CF85B2C9B6F18AULL, "vxdr_primevaltomb_56i_0_0.vlm"},
	{0x62CFD0E412D5AEBFULL, "684924646.ogg"},
	{0x62CFF7501E32996CULL, "vxlt_0_lowtown_a2_n_gij_0_0.vlm"},
	{0x62D007B35601454EULL, "gen_im_trs_apo_sta_cld_r1.uti"},
	{0x62D02955ADC8E2BEULL, "623717259.ogg"},
	{0x62D086DFB3D07BA9ULL, "vxw_encounter2_camp_4m_0_0.vlm"},
	{0x62D09830DD15FC21ULL, "vxlt_0_lowtown_docks_a3_n_cyx_0_1.vlm"},
	{0x62D0B124E8F2B76CULL, "382381379.ogg"},
	{0x62D0C8A32AC99E35ULL, "llac_w_blightlands_2_0.dds"},
	{0x62D1ACD31285B6A0ULL, "gen00fl_bethany.bnk"},
	{0x62D1BC5197D0F306ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.dura.evt"},
	{0x62D24BE5D1FC161AULL, "hlac_ca_cavern1_haunted_24425_4y.dds"},
	{0x62D2560DD3A05B6AULL, "6054032_m.fxe"},
	{0x62D2A87E089D4F57ULL, "6122008_m.fxe"},
	{0x62D2C3F5C9D4DFB9ULL, "724307952.ogg"},
	{0x62D2C67FE5711B0BULL, "vxlt_0_lowtown_gp9_0_0.vlm"},
	{0x62D2D1DE334FC72FULL, "dae300_petrice6097904.cl"},
	{0x62D33D432F59F5D6ULL, "6072708_m.fxe"},
	{0x62D3609D391EB47FULL, "genip_pp_hl_one_shot.utp"},
	{0x62D3CAF579C4E68AULL, "s_ballista_03_l3.mao"},
	{0x62D504ED318F47E6ULL, "6105730_m.fxe"},
	{0x62D5A369A8C93C83ULL, "zz_followers_banter6159546.cl"},
	{0x62D68E8CD08C1D9FULL, "mag311_nyssa6042215.cl"},
	{0x62D6B8508CA3D24BULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_ice_0d.dds"},
	{0x62D6C0D14F8A2F90ULL, "6163185_m.fxe"},
	{0x62D6F88CAB277AFBULL, "dae200cr_viscount6087180.cl"},
	{0x62D7282EF7E91A3AULL, "vxht_0_hightown_a2_bnw_0_0.vlm"},
	{0x62D77605C6739AA2ULL, "68089423.ogg"},
	{0x62D79C1C9CE48E5CULL, "373939705.ogg"},
	{0x62D7BD3531DE75B0ULL, "vxw_woundedcoast_tq_0_0.vlm"},
	{0x62D7F25A5B26668BULL, "gen00fl_varric6118757.cl"},
	{0x62D7FF81F19AE25FULL, "plt_zz_prmpt_debug__.plo"},
	{0x62D82628A25498ABULL, "clf221_jalen6070344.cl"},
	{0x62D8D35197D73951ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.dura.gad"},
	{0x62D8EB1CEEAEB355ULL, "hlac_ht_playershouse_n_855_31x.dds"},
	{0x62D9E2EAF7244653ULL, "art\\vfx\\combat\\vfx_r_ibl_activation_c.impa.ani"},
	{0x62DA0710C98B1C38ULL, "mist_fl_li.mao"},
	{0x62DB0E031099E52BULL, "zz_dae_debug6029184.cl"},
	{0x62DB63E2B138ECF7ULL, "art\\characters\\playercharacter\\humanmale\\hoo_robb.phy"},
	{0x62DB9C8EBFBEEAE1ULL, "983510957.ogg"},
	{0x62DBEE428DAFDA01ULL, "244148765.ogg"},
	{0x62DCC059CE07C132ULL, "vxht_keep_60y_0_0.vlm"},
	{0x62DD0F6BE7679CC0ULL, "hlac_ht_0_hightown_-65535_0z.dds"},
	{0x62DD3BF51E2E2E80ULL, "vxw_bonepit_1x7_0_0.vlm"},
	{0x62DD41096B42B4CFULL, "vxca_caverns1_egj_0_0.vlm"},
	{0x62DE3ED14F0F58E9ULL, "vxca_caverns2prc_9q0_0_0.vlm"},
	{0x62DE61FFF6BA6440ULL, "377583825.ogg"},
	{0x62DE73DDB7DB4CC5ULL, "566605508.ogg"},
	{0x62DEA9F6AC9A8B60ULL, "6089869_m.fxe"},
	{0x62DED45197DBA6FFULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.dura.ani"},
	{0x62DFA3EE6ECDB568ULL, "574048761.ogg"},
	{0x62E0153F2E78400BULL, "t3_wdg_d01.tnt"},
	{0x62E050E2A23943AEULL, "zz_anders_spin_tst.cut"},
	{0x62E1673E5BCF0220ULL, "6172833_m.fxe"},
	{0x62E17758D10C12FCULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_pot.phy"},
	{0x62E1AFAB5BCD7FE8ULL, "zz_spy130_darktown_res_2.cnv"},
	{0x62E1B275FECD882CULL, "gen00fl_anders6127275.cl"},
	{0x62E1CB558C7F8DB9ULL, "vxdr_primevaltomb_3gk_0_0.vlm"},
	{0x62E35228072CA3BFULL, "t1_stb_dbl.tnt"},
	{0x62E36F2201D49022ULL, "vxht_mansion2b_3eg_0_0.vlm"},
	{0x62E50C5AD67AD6A7ULL, "lgt000ip_ultra_bottledscar.utp"},
	{0x62E59278063DF10BULL, "200054619.ogg"},
	{0x62E5B046B165C458ULL, "face_face_p10.psh.fxo.dx11"},
	{0x62E5D485F917CDA4ULL, "6139322_m.fxe"},
	{0x62E5DBE669FCEDF9ULL, "ser270_templar6112565.cl"},
	{0x62E60C40063094F0ULL, "vxlt_warehouse_n_ue_0_0.vlm"},
	{0x62E6A81004240633ULL, "174702611.ogg"},
	{0x62E6DAEAF72EC28AULL, "art\\vfx\\combat\\vfx_r_ibl_activation_c.impa.evt"},
	{0x62E73B2465C5B62AULL, "lt_0_lowtown_a2_n_water_21225.mao"},
	{0x62E73F2465C5BD79ULL, "lt_0_lowtown_a2_n_water_21225.mmh"},
	{0x62E7492465C5CE67ULL, "lt_0_lowtown_a2_n_water_21225.msh"},
	{0x62E768DC7C5E194EULL, "6106876_m.fxe"},
	{0x62E7B74D24D40A12ULL, "ser221_gascard6134249.cl"},
	{0x62E7CF956D7049B4ULL, "hlac_lt_0_lowtown_docks_-65537_5x.dds"},
	{0x62E8526494F0A801ULL, "vxht_0_hightown_n_c6r_0_0.vlm"},
	{0x62E8703AF75E6DB9ULL, "mag360cr_grace.utc"},
	{0x62E87A3AE8F4684EULL, "follower_banter6151266.cl"},
	{0x62E88A5829C88FB9ULL, "vxca_cavern1_haunted_j28_0_1.vlm"},
	{0x62E8A9195CCAEE54ULL, "lgt160ip_wk_avern_deaddrop.ncs"},
	{0x62E8F49C27EDBF6BULL, "lt_0_lowtown_docks_a3_100_0.phy"},
	{0x62E90656DEA38E82ULL, "vxw_sundermount_21r_0_0.vlm"},
	{0x62E95744815FB5ECULL, "6156149_m.fxe"},
	{0x62E9F63D350AA458ULL, "6121099_m.fxe"},
	{0x62EA619B741AC530ULL, "vxdr_primevaltomb_8zg_0_1.vlm"},
	{0x62EA7C742E4D357AULL, "hlac_dr_primevaltomb_659_4x.dds"},
	{0x62EC2F4C3042E905ULL, "art\\vfx\\combat\\vfx_gib_hlm_masc.mao"},
	{0x62EC4BE7298DCEE2ULL, "seb321_rebel_mage.cnv"},
	{0x62EC7BE1D2E35216ULL, "6119270_m.fxe"},
	{0x62ECB62CFDF5A526ULL, "931975488.ogg"},
	{0x62ECDB0A021ECA6CULL, "vxlt_0_lowtown_a2_n_gch_0_0.vlm"},
	{0x62ECE0DF5779A4CBULL, "hlac_ca_cavern1_haunted_24422_4y.dds"},
	{0x62ECE21F17079F2FULL, "w_encounter2_4_0.rml"},
	{0x62ED09CBD6017484ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\death.ani"},
	{0x62ED778F6CBA37BBULL, "art\\vfx\\combat\\vfx_m_arc_bolt_db_c.anb"},
	{0x62EDB9EAF734A9FDULL, "art\\vfx\\combat\\vfx_r_ibl_activation_c.impa.gad"},
	{0x62EDC8976709A794ULL, "6173416_m.fxe"},
	{0x62EDF3C90AB7124EULL, "vxlt_undercitydungeon_1ii_0_0.vlm"},
	{0x62EE273ED5BDDEF7ULL, "hightown_fem_elfservant_mob6169095.cl"},
	{0x62EE38BCCB865D3DULL, "hlac_lt_backalley_g_65536_3y.dds"},
	{0x62EE657483D0A017ULL, "hm_amb_templar.mor"},
	{0x62EF26D1B3BF0D47ULL, "887422606.ogg"},
	{0x62EFAB7CC810B978ULL, "lt_0_lowtown_a2_n_27_0.rml"},
	{0x62F02DE72991841AULL, "seb321_rebel_mage.bnk"},
	{0x62F07DD00F241BCFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_loop.gad"},
	{0x62F0DE91ABBB72A5ULL, "hlac_lt_0_lowtown_docks_n_65535_0x.dds"},
	{0x62F13FCB8649347FULL, "llac_lt_house_-1_1.dds"},
	{0x62F1D77570DFCB94ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_1y.dds"},
	{0x62F1E2F168A62DCDULL, "6028447_m.fxe"},
	{0x62F21178BAB957E8ULL, "hlac_w_sundermount_131074_0y.dds"},
	{0x62F294A3D54BFDCBULL, "ts_kirk_hightown_prog_junk2.gda"},
	{0x62F29906D824CCADULL, "42965267.ogg"},
	{0x62F2AD7EBA566309ULL, "6113002_m.fxe"},
	{0x62F2B37F0C2CA296ULL, "dae000_arishok6078443.cl"},
	{0x62F2D5A0D6ADC493ULL, "gen00fl_aveline6139883.cl"},
	{0x62F3724C9BEA7675ULL, "vxdr_primevaltomb_2w6_0_0.vlm"},
	{0x62F3A8D80A1B8EB6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_catch_and_hold_loop.ani"},
	{0x62F3C11F170D85A7ULL, "w_encounter2_4_0.phy"},
	{0x62F3FF4FA0F8DA7EULL, "845102639.ogg"},
	{0x62F406CBD6078EE6ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\death.gad"},
	{0x62F41ADBF5257A91ULL, "pro000_captain_wright6004414.cl"},
	{0x62F47E1CF826AD70ULL, "vxlt_warehouse_n_xb_0_0.vlm"},
	{0x62F4A0C0A6218395ULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_fin_c.mmh"},
	{0x62F4EF7ECF0DF268ULL, "625958692.ogg"},
	{0x62F4F626F81CC693ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action.ani"},
	{0x62F51E6D1B5A2144ULL, "vxca_caverns1_i9s_0_0.vlm"},
	{0x62F562ACC129374CULL, "266952763.ogg"},
	{0x62F5D9C75B112F2DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic.msh"},
	{0x62F5E7C75B114760ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic.mao"},
	{0x62F5EBC75B114DB3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic.mmh"},
	{0x62F61C050EF6B9B5ULL, "mag320_orsino6060226.cl"},
	{0x62F634AFE4D295A9ULL, "vxca_caverns1_j3s_0_0.vlm"},
	{0x62F638B2F7F66DCBULL, "llac_ca_cavern1_haunted_24424_9.dds"},
	{0x62F69A09AD69E311ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\hm_arm_nud_0d.dds"},
	{0x62F6C44186E6C081ULL, "vxht_0_hightown_n_c3q_0_0.vlm"},
	{0x62F6E43099107A63ULL, "vxlt_0_lowtown_a2_eny_0_0.vlm"},
	{0x62F6FB2FA171A78CULL, "gen00fl_varric6079578.cl"},
	{0x62F71F78EC647B05ULL, "gen00fl_aveline6020567.cl"},
	{0x62F792D00F2A5E7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_loop.evt"},
	{0x62F7CAAE104E1E52ULL, "ser221pt_remains.ncs"},
	{0x62F7CD2E3CE85C52ULL, "hlac_lt_0_lowtown_a2_65537_4y.dds"},
	{0x62F7DE3AC4CEB750ULL, "vxlt_warehouse_11q_0_0.vlm"},
	{0x62F7F7A5025FA9C9ULL, "107915389.ogg"},
	{0x62F855BC8D192519ULL, "vxht_keep_n_6h9_0_0.vlm"},
	{0x62F858E9E55DFB03ULL, "hlac_w_encounter3_-196603_0z.dds"},
	{0x62F87CD936B5C518ULL, "lgt000ip_safe_crate.utp"},
	{0x62F894D9DB8A31B8ULL, "vxdr_primevalentrance_2q3_0_2.vlm"},
	{0x62F93D2E417BEF81ULL, "gen00fl_carver6104907.cl"},
	{0x62FA01CBD60BF1C5ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\death.evt"},
	{0x62FA086F848EBD96ULL, "38994007.ogg"},
	{0x62FA0FF3160E952DULL, "650096627.ogg"},
	{0x62FA8D802436F0B8ULL, "t1_mub_re1.tnt"},
	{0x62FAAD0449A9A42FULL, "lit_tep_01.mao"},
	{0x62FAB1C5C0160ADAULL, "6169453_m.fxe"},
	{0x62FAF8E0F5E0A584ULL, "6142711_m.fxe"},
	{0x62FB90D910728A95ULL, "5249724.ogg"},
	{0x62FBCC57D0C29E24ULL, "hlac_lt_gallowscourtyard_n_-65537_0z.dds"},
	{0x62FBFB7710E8B3CFULL, "gen00fl_dog6117773.cl"},
	{0x62FC1FCCF076AF55ULL, "combat_smcf_r4_a1.gda"},
	{0x62FCE9E65AE2AF32ULL, "115599620.ogg"},
	{0x62FCF4D328FBBAEBULL, "hf_mother.mor"},
	{0x62FD3B2D8892E910ULL, "vxw_bonepit_7y_0_0.vlm"},
	{0x62FDB8C438B00C07ULL, "hlac_ca_caverns2_5802_3y.dds"},
	{0x62FEF5C533F42EC7ULL, "vrc321_party_barks6095735.cl"},
	{0x62FF18618AAA3C32ULL, "6154291_m.fxe"},
	{0x62FF80094BEACC3AULL, "vxlt_0_lowtown_a2_hep_0_0.vlm"},
	{0x62FF8D6AA6AF07B3ULL, "dae270ip_weapon_store.utp"},
	{0x62FFC338753E3BB2ULL, "vxlt_gallowstemplar_n_4ul_0_0.vlm"},
	{0x63001FCFA29B6BE2ULL, "weapon_distortionmask_p0.vsh.fxo.dx11"},
	{0x63002F83B9B3F444ULL, "6170927_m.fxe"},
	{0x6300640FD5D86AABULL, "vxlt_gallowscourtyard_n_3x3_0_0.vlm"},
	{0x630088195CDF4785ULL, "lgt160ip_wk_avern_deaddrop.utp"},
	{0x6300FC7E88BC4263ULL, "637699094.ogg"},
	{0x63011CE48440C78BULL, "hlac_ca_caverns1_mines_24421_0z.dds"},
	{0x63015FBEFA8E7094ULL, "follower_banter6172744.cl"},
	{0x63018F67B06FB2CAULL, "vxht_mansion_n_6n4_0_0.vlm"},
	{0x6302082E8F0FB7ECULL, "6087179_m.fxe"},
	{0x63029DC0A62DBC22ULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_fin_c.anb"},
	{0x6302ADFA941AEEE8ULL, "6153726_m.fxe"},
	{0x6302B7832C8A48B2ULL, "dae210_gate_guard6084331.cl"},
	{0x6302BA019D611BD5ULL, "877739367.ogg"},
	{0x6302CD47C0832EEDULL, "w_bonepit_3_0_lc.mmh"},
	{0x6302CDB6FCF1B1BCULL, "vxht_mansion_n_613_0_0.vlm"},
	{0x6302EE26F828F6CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action.evt"},
	{0x63035452A3AE1D4BULL, "610004378.ogg"},
	{0x6303AA7907B2C0D4ULL, "6098600_f.fxe"},
	{0x6303C0DB3C27E0C5ULL, "nat380_nathaniel6071026.cl"},
	{0x63049AD00F34F595ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_loop.ani"},
	{0x6305361E78DCD901ULL, "vxht_0_hightown_n_c8p_0_0.vlm"},
	{0x63057EC946796D98ULL, "hlac_lt_0_lowtown_docks_a3_-4_2z.dds"},
	{0x6305B5A493091004ULL, "vxlt_0_lowtown_a3_n_c_gzd_0_0.vlm"},
	{0x6305FB9599362025ULL, "684109883.ogg"},
	{0x63065D209B078E7CULL, "vxlt_0_lowtown_cow_0_0.vlm"},
	{0x6306A21925002F7DULL, "6031855_f.fxe"},
	{0x6306AB68355BC297ULL, "gen_im_trs_dem_junk1_03.uti"},
	{0x6306C7997F0F3D99ULL, "vxht_keep_n_6c8_0_0.vlm"},
	{0x630718796A798ADCULL, "vxw_woundedcoast_dm_0_0.vlm"},
	{0x63072979F572C4AFULL, "projdecal.mat"},
	{0x6307D1D80A2C7DF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_catch_and_hold_loop.gad"},
	{0x630810AB7374679CULL, "mag291_vincento6023821.cl"},
	{0x6308244ADB029EF6ULL, "vxht_0_hightown_a2_c13_0_0.vlm"},
	{0x630830615ACE5C3BULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsaltar.msh"},
	{0x630842615ACE7A86ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsaltar.mao"},
	{0x63084D98247A89E5ULL, "543063457.ogg"},
	{0x63084E615ACE8F25ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsaltar.mmh"},
	{0x6308918BF4107800ULL, "vxw_bonepit_1k2_0_0.vlm"},
	{0x630896A04124FE4FULL, "vxca_caverns1_ehw_0_0.vlm"},
	{0x6309946824F1A269ULL, "vxca_caverns2prc_9f5_0_0.vlm"},
	{0x6309C2F08FFC2243ULL, "926232690.ogg"},
	{0x6309CD26F82EDD3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_action.gad"},
	{0x6309DA414F15E648ULL, "920985685.ogg"},
	{0x630A7AAAED1EEF37ULL, "6081212_f.fxe"},
	{0x630AF2E8A8AD3D64ULL, "222878694.ogg"},
	{0x630C12986AF9AAB0ULL, "llac_w_blightlands_131073_0.dds"},
	{0x630C467A91B05974ULL, "6051213_m.fxe"},
	{0x630CA1F5359AFF50ULL, "6113527_f.fxe"},
	{0x630CE30E3AD6916CULL, "follower_banter6141441.cl"},
	{0x630D1AE18E2EC8F3ULL, "zz_dog_debug6127750.cl"},
	{0x630D3AAE253C5AF9ULL, "vxht_bartrand_n_4es_0_0.vlm"},
	{0x630D3B01FAAF423EULL, "6043692_m.fxe"},
	{0x630D61509F8FDA2EULL, "art\\vfx\\combat\\vfx_glb_restore_health_c.mmh"},
	{0x630D88E2BA0A17E2ULL, "135752401.ogg"},
	{0x630DBBDD9F0565C0ULL, "mcr370_meredith6085901.cl"},
	{0x630DC3AA000BD7BBULL, "6031696_m.fxe"},
	{0x630DE5326E0D5972ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05a.phy"},
	{0x630DF1E63DE0E4BAULL, "vxlt_0_lowtown_a2_hbq_0_0.vlm"},
	{0x630DF316196FEF39ULL, "vxht_keep_n_5ws_0_0.vlm"},
	{0x630E93EF578BF7E6ULL, "vxw_encounter2_camp_pb_0_0.vlm"},
	{0x630E9ADFA04C33F7ULL, "hlac_lt_0_lowtown_docks_a3_-131074_3x.dds"},
	{0x630EA191379A7C8AULL, "342371460.ogg"},
	{0x630EBB9F12DD6BA6ULL, "zz_followers_banter6157323.cl"},
	{0x630EBD666A0CCF43ULL, "374349256.ogg"},
	{0x630F17289B2463A6ULL, "6051145_m.fxe"},
	{0x630F9AE863E86BF5ULL, "hlac_w_woundedcoast_0_8z.dds"},
	{0x630FCA52A7E7A6D2ULL, "zz_vrc_debug6200165.cl"},
	{0x63100410A4F8A15DULL, "dre000_varric6035185.cl"},
	{0x63104F996D41B8D9ULL, "362856597.ogg"},
	{0x63106E93AEFA6ABAULL, "613441551.ogg"},
	{0x6310AB211840F6DBULL, "vxlt_0_lowtown_a2_n_eh8_0_0.vlm"},
	{0x63114C9655CCE262ULL, "vxlt_gallowscourtyard_7ar_0_0.vlm"},
	{0x63114EBE6F5ABCC8ULL, "vxht_0_hightown_621_0_0.vlm"},
	{0x63119B30D1DE9A66ULL, "s_po_splat_dot.mao"},
	{0x6311DDDA197550F1ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_2z.dds"},
	{0x631256238B4792FAULL, "hlac_dr_0_deeproads_5843_6z.dds"},
	{0x6312B230AE2B8270ULL, "bd_01.mao"},
	{0x6312B77CCA591C6BULL, "hlac_w_encounter3_-262146_0z.dds"},
	{0x6312D3C4B65375F2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_lava_d.dds"},
	{0x63134FD77C7411A7ULL, "pin_w.mao"},
	{0x63138F6DAC6C73F8ULL, "6040017_m.fxe"},
	{0x63138FF1C3B51C1EULL, "dae021_bureaucrat_2.utc"},
	{0x6313AAD15658080FULL, "art\\vfx\\combat\\vfx_str_att_front_l_arm_c.impa.gad"},
	{0x6313D25697ECE734ULL, "gen00fl_aveline6139222.cl"},
	{0x6313D95EB6F0EB02ULL, "670069804.ogg"},
	{0x6313E574E411C98EULL, "hla_ht_0_hightown_131073_0y.dds"},
	{0x63142C0CE8FCA9A5ULL, "gen00fl_sebastian6188612.cl"},
	{0x63142FD31BA8E572ULL, "dae000_arishok6067810.cl"},
	{0x63143A9CCFC2C740ULL, "361221591.ogg"},
	{0x6314A1D4CE58DD30ULL, "vxw_bonepit_1hw_0_0.vlm"},
	{0x6314F30578B51A8FULL, "vxlt_ruinedhouse_n_hk_0_0.vlm"},
	{0x6314F8FD30B77759ULL, "llac_ht_0_hightown_n_65536_0.dds"},
	{0x631507B572C02D61ULL, "hlac_ht_keep_4519_16z.dds"},
	{0x63158A565C6FA35CULL, "vxw_woundedcoast_ab_0_0.vlm"},
	{0x6316205C55808BF7ULL, "6128918_m.fxe"},
	{0x631654D5A74D916CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1_0.pwk"},
	{0x6316772C6AF3677FULL, "hlac_ht_0_hightown_a3_65532_0z.dds"},
	{0x631678A914EE74CBULL, "dae000_merchant_armor_hm6151546.cl"},
	{0x6316B9DD45097974ULL, "hlac_lt_0_lowtown_n_1_10x.dds"},
	{0x63170368E6069080ULL, "vxw_bonepit_1l3_0_0.vlm"},
	{0x631725CE6BA125DDULL, "22421266.ogg"},
	{0x6317345920DEEDBAULL, "265660339.ogg"},
	{0x6317805D7B3B528EULL, "289793123.ogg"},
	{0x631792AE0692ECBFULL, "vxlt_0_lowtown_a2_n_gcg_0_0.vlm"},
	{0x63182E6034699CEDULL, "fen110_fenris6118013.cl"},
	{0x6318BFD8717A02CAULL, "181919490.ogg"},
	{0x6318F4F4DA396D39ULL, "vxht_brothel_n_4vn_0_0.vlm"},
	{0x6319263098ED0084ULL, "28148414.ogg"},
	{0x631929E3DEFF6C8EULL, "6163330_m.fxe"},
	{0x63195CEEA17C78A6ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_hed_suca_0e.dds"},
	{0x63199ACCF34227F1ULL, "1028251468.ogg"},
	{0x63199F70343F32BBULL, "15102796.ogg"},
	{0x6319BFD1565C97BCULL, "art\\vfx\\combat\\vfx_str_att_front_l_arm_c.impa.evt"},
	{0x631A2FFA66B9B98DULL, "art\\vfx\\models\\vfx_rockarmor_torso.msh"},
	{0x631A3A2C877BEF54ULL, "w_bonepit_18_0_ll.mmh"},
	{0x631A41FA66B9D813ULL, "art\\vfx\\models\\vfx_rockarmor_torso.mmh"},
	{0x631A601F8B1299DFULL, "art\\characters\\weapons\\longbows\\w_lbw_lb101.mao"},
	{0x631AA202F274CEAEULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_act_c.mmh"},
	{0x631B7630C9E7D8A1ULL, "vxca_cave_2b4_0_0.vlm"},
	{0x631B80509F9C4C2DULL, "art\\vfx\\combat\\vfx_glb_restore_health_c.anb"},
	{0x631B857B985EC3DCULL, "art\\characters\\playercharacter\\humanmale\\hn_bdy_tmpa.mao"},
	{0x631BC3E09D2F6D2DULL, "ht_0_hightown_a2_41_0.rml"},
	{0x631C3E996B4CA723ULL, "755589257.ogg"},
	{0x631C8C7CC8377230ULL, "lt_0_lowtown_a2_n_27_0.anb"},
	{0x631CA54B5735D32AULL, "vxht_bartrand_n_5xb_0_0.vlm"},
	{0x631CE8D4858F3E5AULL, "xx_wood.dds"},
	{0x631D05CC49821066ULL, "vxw_encounter2_camp_kc_0_0.vlm"},
	{0x631D34C75AAEDFC8ULL, "llac_lt_blackemporium_1055_2.dds"},
	{0x631D56EA68799C93ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candle_n.dds"},
	{0x631D675D17C1A94EULL, "w_bonepitmining_a2_97_0.msh"},
	{0x631D755D17C1C09FULL, "w_bonepitmining_a2_97_0.mao"},
	{0x631D815D17C1D53CULL, "w_bonepitmining_a2_97_0.mmh"},
	{0x631D845678B9176EULL, "vxlt_0_lowtown_docks_bc9_0_0.vlm"},
	{0x631D847789E265AFULL, "vxw_sundermount_248_0_0.vlm"},
	{0x631F5F8C1F829BBBULL, "vxw_bonepitdestroyed_a3_1s5_0_0.vlm"},
	{0x631FD9D68A73B633ULL, "vxlt_undercitydungeon_1kv_0_0.vlm"},
	{0x6320A89FB4C4F012ULL, "270762844.ogg"},
	{0x6320ABFF95C37A9FULL, "228484366.ogg"},
	{0x6320C0A7C1DBF10BULL, "vxca_cavern1_haunted_j56_0_0.vlm"},
	{0x63216E949555FB89ULL, "6655490.ogg"},
	{0x63216F2FC96A83A7ULL, "vxw_encounter1_em_0_0.vlm"},
	{0x6321F24D7FD66136ULL, "6177322_m.fxe"},
	{0x63220D919CF86271ULL, "6165559_m.fxe"},
	{0x63227DCEB15A85E9ULL, "w_blightlands_146_0.mao"},
	{0x632282D4072DE3F5ULL, "vxht_playershouse_n_1v5_0_0.vlm"},
	{0x632289CEB15A9982ULL, "w_blightlands_146_0.mmh"},
	{0x63228BCEB15A9D38ULL, "w_blightlands_146_0.msh"},
	{0x632313B1C04EF5B0ULL, "vxw_bonepit_1mt_0_0.vlm"},
	{0x632338E2B176594FULL, "art\\characters\\playercharacter\\humanmale\\hoo_robb.mao"},
	{0x632344E2B1766DECULL, "art\\characters\\playercharacter\\humanmale\\hoo_robb.mmh"},
	{0x63234AE2B17677DEULL, "art\\characters\\playercharacter\\humanmale\\hoo_robb.msh"},
	{0x63238B1D63BC0409ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_6x.dds"},
	{0x632397F3BAED23E3ULL, "35292404.ogg"},
	{0x6323A390FB72F43DULL, "91006722.ogg"},
	{0x6323E3410D738E1AULL, "lgt310cr_mag_melson_guard2.utc"},
	{0x632478D52182E212ULL, "vxca_caverns2_73k_0_1.vlm"},
	{0x6324833D227F2554ULL, "6152549_m.fxe"},
	{0x6324B64879226F96ULL, "6125769_m.fxe"},
	{0x632536582C251DDDULL, "zz_dae_debug6029858.cl"},
	{0x632541086BE14DA5ULL, "dae000_hate_follower_react6103209.cl"},
	{0x632612401CD07B7BULL, "vxlt_0_lowtown_docks_n_bfx_0_0.vlm"},
	{0x632630164B7BD86EULL, "6076667_m.fxe"},
	{0x63263B5AF6316DFFULL, "art\\levels\\ztools\\default\\generic\\4sc_basic.msh"},
	{0x6326415AF63177A1ULL, "art\\levels\\ztools\\default\\generic\\4sc_basic.mmh"},
	{0x63264D5AF6318C42ULL, "art\\levels\\ztools\\default\\generic\\4sc_basic.mao"},
	{0x63268254C472E3B2ULL, "6071496_f.fxe"},
	{0x632701F6B9B9AE50ULL, "vxdr_0_deeproads_em8_0_0.vlm"},
	{0x63271DFCB7DEC76CULL, "song_clo.mao"},
	{0x6327819CF42FA12DULL, "6164404_m.fxe"},
	{0x6327A225F6E2933DULL, "vxlt_undercitydungeon_14e_0_0.vlm"},
	{0x6327C7D15668E2D5ULL, "art\\vfx\\combat\\vfx_str_att_front_l_arm_c.impa.ani"},
	{0x6327FE702EB180F6ULL, "6129453_m.fxe"},
	{0x632831443F7852B6ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.impa.evt"},
	{0x63283A4D23348B3EULL, "llac_lt_0_lowtown_a2_-131073_3.dds"},
	{0x6328702EC9521002ULL, "w_encounter2_28_0.msh"},
	{0x632871446E59699BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pickaxe_sa.dds"},
	{0x6328722EC95213B8ULL, "w_encounter2_28_0.mmh"},
	{0x63287D22C860501AULL, "vxlt_0_lowtown_docks_a3_d2r_0_1.vlm"},
	{0x63287E2EC952285BULL, "w_encounter2_28_0.mao"},
	{0x632881FA14DF19D1ULL, "291789599.ogg"},
	{0x6328B656919F56AFULL, "62032992.ogg"},
	{0x6328C102F28140ADULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_act_c.anb"},
	{0x6328EF6A137DF3BEULL, "vxht_bartrand_n_641_0_0.vlm"},
	{0x6329CD9BBB701197ULL, "vxlt_hangedman_20v_0_0.vlm"},
	{0x632A115057407E95ULL, "866341115.ogg"},
	{0x632A427C752E6007ULL, "vxlt_0_lowtown_a3_n_c_hr0_0_0.vlm"},
	{0x632B0FAF3230EFEDULL, "6143168_m.fxe"},
	{0x632B5D678D724F34ULL, "qun231_varnell6077274.cl"},
	{0x632B5DF902264016ULL, "6094542_m.fxe"},
	{0x632B722EC95449CBULL, "w_encounter2_28_0.rml"},
	{0x632BA6B91D6297C8ULL, "hlac_dr_primevaltomb_658_7x.dds"},
	{0x632BADCA5E92A1C0ULL, "6076074_m.fxe"},
	{0x632BE5AC6091D109ULL, "161416291.ogg"},
	{0x632BF95AD6B6B942ULL, "lgt000ip_ultra_bottledscar.ncs"},
	{0x632C936B7E76C26CULL, "vxlt_0_lowtown_docks_n_c4t_0_0.vlm"},
	{0x632CB28827A7FF71ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_maj_c.anb"},
	{0x632CFF659768C521ULL, "6136206_f.fxe"},
	{0x632DBDE647BF3CFCULL, "cd_act1_rockwraith_intro_demo_mage.wbd"},
	{0x632E009A4ACAC3A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_reaction.ani"},
	{0x632E061086F92963ULL, "mrl310_party_comments6142834.cl"},
	{0x632E0F9965F22293ULL, "219324358.ogg"},
	{0x632E46D37AECDA2CULL, "53978819.ogg"},
	{0x632E4BB37C69CEB3ULL, "vxlt_undercitydungeon_1ny_0_0.vlm"},
	{0x632E9AEFCF47969DULL, "6116122_m.fxe"},
	{0x632EE3877EF4A4D5ULL, "hlac_lt_undercitydungeon_-1_42z.dds"},
	{0x632F28443F7E6241ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.impa.gad"},
	{0x632F465B12C2E46AULL, "plt_dae010pt_amb_docks.plo"},
	{0x632F4BE4EA20CD26ULL, "lgt110_arl_bann_teagan6115609.cl"},
	{0x632F84BD77B46B9DULL, "dae000_keep_noblewoman_sta6162071.cl"},
	{0x632FBA82367F5282ULL, "zz_followers_banter6158814.cl"},
	{0x632FDF6062034467ULL, "dae000cr_grand_cleric6187918.cl"},
	{0x632FE10CBB609C27ULL, "vxw_encounter1_jl_0_0.vlm"},
	{0x63303D792A60A276ULL, "vxca_caverns2_71d_0_0.vlm"},
	{0x633079C0EF088C87ULL, "17093878.ogg"},
	{0x63308CB92D9E2EA0ULL, "511325559.ogg"},
	{0x6330AEDAA9D9B89CULL, "6116072_f.fxe"},
	{0x633111CE8224FF7EULL, "hlac_lt_undercityserialkiller_-1_51x.dds"},
	{0x6331363F9AFF5BA8ULL, "6147688_m.fxe"},
	{0x6331A8FF6EFC30AFULL, "gen_im_trs_cot_arc_lbw_r0.uti"},
	{0x6331D52B4AEF254CULL, "zz_isa_debug6133157.cl"},
	{0x6332481A80F99C87ULL, "vxca_cavern1_haunted_e7o_0_0.vlm"},
	{0x6332792EC95A7443ULL, "w_encounter2_28_0.phy"},
	{0x6332A87EA2F64860ULL, "gen_im_acc_amu_act3_01"},
	{0x6332BF1615422708ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_whoresign_sa.dds"},
	{0x6333345FCC0B9EB2ULL, "158367460.ogg"},
	{0x633352E66FA819F1ULL, "6060699_f.fxe"},
	{0x63339C17AA097B5BULL, "778356494.ogg"},
	{0x6333ECFB19994A42ULL, "24008082.ogg"},
	{0x63341EDFFCD2F8DFULL, "744647601.ogg"},
	{0x633457DF3D1FF51AULL, "lowtown_male_beggar_sta6151076.cl"},
	{0x63346A886CE58F9BULL, "467050339.ogg"},
	{0x6334707605C2C2ADULL, "vxdr_0_deeproads_boq_0_1.vlm"},
	{0x6334E9D4CD6E5B47ULL, "llac_w_bonepitdestroyed_a3_262146_0.dds"},
	{0x63350ACAC5CBB5F0ULL, "hla_ht_0_hightown_65538_1x.dds"},
	{0x63350B9A4AD0F543ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_reaction.gad"},
	{0x6335152995B2E939ULL, "hlac_lt_0_lowtown_n_-131073_3z.dds"},
	{0x6335182C53342EA9ULL, "952552175.ogg"},
	{0x63353C70DAC33D2AULL, "dog_sl_01.mao"},
	{0x633550C0F944B4E4ULL, "6132786_f.fxe"},
	{0x63358B3C14FDD9DCULL, "vxw_bonepit_1wj_0_0.vlm"},
	{0x633629443F8482EFULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.impa.ani"},
	{0x633658BA8B275631ULL, "vxca_caverns2prc_afu_0_0.vlm"},
	{0x6336D594853478EAULL, "8621526.ogg"},
	{0x6336ECA12202FC64ULL, "gen00fl_aveline6114414.cl"},
	{0x6337485708C427C4ULL, "6034659_f.fxe"},
	{0x63374E7ADA735FBDULL, "vxca_cavern1_haunted_hib_0_0.vlm"},
	{0x6337755D7711635CULL, "62332648.ogg"},
	{0x63379ACE91B0A19CULL, "vxht_0_hightown_a2_bp3_0_0.vlm"},
	{0x6337F46603E9A481ULL, "vxw_woundedcoast_a3_0_0.vlm"},
	{0x63380C9A17D23E73ULL, "217206489.ogg"},
	{0x6338639332B1428BULL, "hlac_ca_caverns1_24426_11z.dds"},
	{0x6339E9CC140A74E2ULL, "vxlt_0_lowtown_docks_bbz_0_0.vlm"},
	{0x633A1DA377856FD8ULL, "vxdr_primevalentrance_32u_0_0.vlm"},
	{0x633A622A66DA9EE6ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.dura.ani"},
	{0x633A93474779C0FEULL, "vxlt_0_lowtown_a2_n_ei6_0_0.vlm"},
	{0x633AD38827B474EAULL, "art\\vfx\\combat\\vfx_glb_lit_imp_maj_c.mmh"},
	{0x633B3911233D8B9BULL, "mag121_viveka6083374.cl"},
	{0x633B5F75CB95E8C5ULL, "11514810.ogg"},
	{0x633BB6529B2C065BULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgte.phy"},
	{0x633BD7ADDE9661DDULL, "6125775_m.fxe"},
	{0x633BE89A4AD6D828ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_reaction.evt"},
	{0x633BE9F64C83C4BAULL, "hlac_lt_darktown_-1_85y.dds"},
	{0x633C066A9FBC867FULL, "6169441_m.fxe"},
	{0x633C2A6E77C857CBULL, "vxlt_hangedman_26h_0_0.vlm"},
	{0x633C5D0287D56A37ULL, "vxw_sundermount_2k9_0_0.vlm"},
	{0x633CBD906E5FE733ULL, "vxlt_undercitydungeon_1qx_0_0.vlm"},
	{0x633D3427302F2754ULL, "6083579_m.fxe"},
	{0x633D898E2E3E3522ULL, "vxlt_house_f0_0_0.vlm"},
	{0x633DD523B31B6AEEULL, "367519896.ogg"},
	{0x633E010B25694D0DULL, "mcr310_argument6074503.cl"},
	{0x633E13D4CF9CBCB1ULL, "vxlt_0_lowtown_a2_n_c_hkx_0_0.vlm"},
	{0x633E306CC6E258F8ULL, "qcr000_varric6208240.cl"},
	{0x633EAF172F5AF0A7ULL, "gen00fl_anders6201792.cl"},
	{0x633F1B55574F4485ULL, "gen_im_cns_cft_psn_01.uti"},
	{0x633F59D004362FECULL, "w_bonepitmining_a2_188_0.phy"},
	{0x6340018C56CB08ACULL, "vxht_mansion2_3db_0_0.vlm"},
	{0x634060982783F0F8ULL, "6149440_m.fxe"},
	{0x6340665481D189DFULL, "llac_lt_0_lowtown_a3_n_c_131073_1.dds"},
	{0x6340B7AB29E0C609ULL, "368480358.ogg"},
	{0x6340DFED3251ECDCULL, "vxw_woundedcoast_pa_0_0.vlm"},
	{0x6341CC9A01A08916ULL, "1008829688.ogg"},
	{0x6341EFEFEA0A335AULL, "165837020.ogg"},
	{0x63423BF8C30AD1D3ULL, "6101357_m.fxe"},
	{0x63424BC754680820ULL, "ico_helm_medium_3.dds"},
	{0x6342A51EB7942B6CULL, "dae000ip_crowd_ff4_talk7.utp"},
	{0x63432DE3C19E825DULL, "tow_smok.mao"},
	{0x63432F26E3B6FBC8ULL, "llac_lt_0_lowtown_docks_-1_0.dds"},
	{0x63439713654346B0ULL, "763800107.ogg"},
	{0x6343C0717491CE50ULL, "6108453_f.fxe"},
	{0x6343D460F14C2178ULL, "leave_07_l2.mao"},
	{0x6343EBA2D037F622ULL, "1067967811.ogg"},
	{0x6343F57CE0634B20ULL, "369390004.ogg"},
	{0x63441EE58C03C6BCULL, "vxlt_warehouse_9t_0_0.vlm"},
	{0x634433A829306B0DULL, "6139716_m.fxe"},
	{0x6344403F75B2DD86ULL, "gen_im_fla_merril_01.uti"},
	{0x63445F5485BCBA77ULL, "vxlt_0_lowtown_docks_bd5_0_0.vlm"},
	{0x6344664102676F30ULL, "vxca_caverns1_edp_0_0.vlm"},
	{0x63449EB8ED55569FULL, "6040822_f.fxe"},
	{0x63450FB24B9804C6ULL, "vxht_mansion2b_1t2_0_0.vlm"},
	{0x63451E0E92EE0279ULL, "6112078_m.fxe"},
	{0x634531E4C4CE8C2BULL, "vxlt_gallowstemplar_n_5xo_0_0.vlm"},
	{0x634534F9FF45F38DULL, "fxa_pois_smoke_add.mao"},
	{0x6345D09CD7D00D38ULL, "vxlt_0_lowtown_a2_n_eji_0_0.vlm"},
	{0x634604EC47F1D8FAULL, "pro000cs_flemeth_fake.cut"},
	{0x63460CAB83A6BA1CULL, "vxht_0_hightown_a2_bu0_0_0.vlm"},
	{0x6346869B5988B219ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em03.mmh"},
	{0x6346909B5988C307ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em03.msh"},
	{0x6346BF6A834F9CC7ULL, "one110_meeran6089881.cl"},
	{0x6346C4223183B3BCULL, "vxlt_0_lowtown_a2_emo_0_0.vlm"},
	{0x6347022FB014FC6EULL, "270571745.ogg"},
	{0x63472636591A9107ULL, "vxlt_0_lowtown_a3_n_c_hd2_0_0.vlm"},
	{0x6347330DBA255482ULL, "vrc221_bartrand_fake6064784.cl"},
	{0x634767361D92D4C3ULL, "mrl000_marethari6129260.cl"},
	{0x63476EF0E7EC9894ULL, "zz_dae_debug6029321.cl"},
	{0x63479E4A02F5309DULL, "43575955.ogg"},
	{0x63480FED9FF3F735ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerc.msh"},
	{0x634811ED9FF3FAABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerc.mmh"},
	{0x63481DED9FF40F48ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerc.mao"},
	{0x63481FC0D2D4CD0CULL, "6152336_m.fxe"},
	{0x63484A2A66E6B357ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.dura.evt"},
	{0x6348B588546EA0ADULL, "llac_ht_mansion2b_818_18.dds"},
	{0x63490524396FD97EULL, "vxlt_0_lowtown_a2_n_er7_0_0.vlm"},
	{0x63491BD4CB7B743CULL, "816428999.ogg"},
	{0x634924F6C7B196BCULL, "vxht_keep_n_6ca_0_0.vlm"},
	{0x634A200CBDC306ACULL, "vxlt_gallowscourtyard_6y1_0_0.vlm"},
	{0x634A5A9F6F2FCD16ULL, "6088914_m.fxe"},
	{0x634AB522F564E53BULL, "vxw_bonepitdestroyed_a3_1b6_0_0.vlm"},
	{0x634ACEDF79CB82B7ULL, "vxw_sundermount_2l8_0_0.vlm"},
	{0x634B10F5C9811652ULL, "112852092.ogg"},
	{0x634B18333C72E52CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicd.msh"},
	{0x634B1E333C72EF1EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicd.mmh"},
	{0x634B2596E77F3ADFULL, "6079286_f.fxe"},
	{0x634B2A333C7303B5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicd.mao"},
	{0x634B3BA06A3D55DBULL, "lt_0_lowtown_docks_a3_132_0.mmh"},
	{0x634B47A06A3D6A78ULL, "lt_0_lowtown_docks_a3_132_0.mao"},
	{0x634B4A142BE09A7DULL, "56967266.ogg"},
	{0x634B59A06A3D88C5ULL, "lt_0_lowtown_docks_a3_132_0.msh"},
	{0x634BB2D6B937FD5DULL, "6173501_m.fxe"},
	{0x634BBBB9D0267B56ULL, "307922980.ogg"},
	{0x634C8FFB3E6FCBA4ULL, "dre000_varric6036998.cl"},
	{0x634CDCD7D533FA69ULL, "144438946.ogg"},
	{0x634D32937B2F2938ULL, "6111607_m.fxe"},
	{0x634D821FD844167CULL, "vxca_caverns1_mines_ei6_0_0.vlm"},
	{0x634D9C4392E7F687ULL, "1009146820.ogg"},
	{0x634DE27BED678CE7ULL, "vxca_cavern1_haunted_jg4_0_0.vlm"},
	{0x634E1392CAC69696ULL, "6097287_f.fxe"},
	{0x634E3831CF3514D8ULL, "410016189.ogg"},
	{0x634E3EF024B6DC9AULL, "6066378_m.fxe"},
	{0x634E5A2479066A94ULL, "vxlt_0_lowtown_docks_a3_n_c3r_0_0.vlm"},
	{0x634E68092C6B9F56ULL, "770168863.ogg"},
	{0x634F089FDC1FF37DULL, "97319938.ogg"},
	{0x634F27615B0A4FD2ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsaltar.phy"},
	{0x634F4B2A66ECD3E0ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.dura.gad"},
	{0x634F58C667A98EACULL, "6064925_m.fxe"},
	{0x634FB2BA5F156711ULL, "140109516.ogg"},
	{0x634FBF372DC52347ULL, "mag110_arianni6015779.cl"},
	{0x635022C08B71545EULL, "zz_followers_banter6159189.cl"},
	{0x635043E8BB3E8748ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutb.msh"},
	{0x635055E8BB3EA5D9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutb.mao"},
	{0x635061E8BB3EBA72ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutb.mmh"},
	{0x63507329EF58D5EDULL, "seb201_lady_harimann6177371.cl"},
	{0x635080C576D57421ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed.anb"},
	{0x6350FFA2A1BCB9D7ULL, "6090323_m.fxe"},
	{0x635116768C1E1F1EULL, "849381629.ogg"},
	{0x635186774B5DCE0AULL, "vxdr_primevaltomb_5cb_0_0.vlm"},
	{0x635186E6D9AA39EBULL, "one140cr_charade6137618.cl"},
	{0x6351FE1ECC123D6BULL, "hlac_lt_gallowstemplar_f_2383_9y.dds"},
	{0x63520FF826140EE5ULL, "pro000_merchant6051801.cl"},
	{0x63533A817BC1E169ULL, "qun230cr_guild_contact6068860.cl"},
	{0x6353425AF658671EULL, "art\\levels\\ztools\\default\\generic\\4sc_basic.phy"},
	{0x6354030D0D6049C6ULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.cess.evt"},
	{0x63541ABD5C0F0F8DULL, "vxht_bartrand_n_6q5_0_0.vlm"},
	{0x6354815DFE37915BULL, "327463631.ogg"},
	{0x6354ADE92B073DB2ULL, "qun110_petrice6172089.cl"},
	{0x6354C1DD7B3B6E76ULL, "813933168.ogg"},
	{0x6355C27BF25AA81AULL, "vxht_mansion_n_6ow_0_0.vlm"},
	{0x6355D64F1534D32FULL, "vxw_encounter2_5l_0_0.vlm"},
	{0x63569B310B31CEC8ULL, "vxht_0_hightown_9zr_0_0.vlm"},
	{0x6356DA6BA1BA153BULL, "6068878_m.fxe"},
	{0x63572051630941B4ULL, "llac_ht_bartrand_n_3366_1.dds"},
	{0x635774BC6001F8C3ULL, "fxp_pillar_sym04.mao"},
	{0x6357F9A4E541FA0DULL, "llac_lt_0_lowtown_196607_1.dds"},
	{0x635842412165C1B0ULL, "1026739747.ogg"},
	{0x63589918653D818CULL, "vxca_caverns1_mines_ck6_0_0.vlm"},
	{0x6358C0559C463D56ULL, "1058855187.ogg"},
	{0x6358CC7C6125E596ULL, "6083814_m.fxe"},
	{0x6358CFCAF1FCAD03ULL, "art\\vfx\\combat\\vfx_m_cre_conj_c.anb"},
	{0x63594E20807075C2ULL, "llac_w_sundermount_-65537_0.dds"},
	{0x6359F4A477E70702ULL, "57504744.ogg"},
	{0x6359FA95F874D4C9ULL, "llac_w_encounter2_camp_131071_0.dds"},
	{0x635ABA8A2156CD0EULL, "244832984.ogg"},
	{0x635ABBE647E62776ULL, "cd_act1_rockwraith_intro_demo_mage.bnk"},
	{0x635AF01D81A9E069ULL, "6201453_m.fxe"},
	{0x635AF3E751FEA368ULL, "782342482.ogg"},
	{0x635B171AA60D76E0ULL, "mag291_vincento6027717.cl"},
	{0x635B1A0D0D668F11ULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.cess.gad"},
	{0x635B3DEB1447F8B1ULL, "qun111_ketojan6039803.cl"},
	{0x635BF7D64A3B1B25ULL, "ico_belt_5.dds"},
	{0x635C011FA96187D6ULL, "580021132.ogg"},
	{0x635CBB9F42DCE09BULL, "6081282_m.fxe"},
	{0x635CBE812D1F0F6AULL, "follower_banter6141775.cl"},
	{0x635CED62A9DD17C5ULL, "fxe_water_pan_add.dds"},
	{0x635D021826BDDB1BULL, "art\\vfx\\creatures\\vfxroot.impa.ani"},
	{0x635D614505EF567CULL, "vxlt_hangedman_2x9_0_0.vlm"},
	{0x635E101FD715BAF7ULL, "mag110_samson6149958.cl"},
	{0x635E259E02062295ULL, "311483751.ogg"},
	{0x635E4F5B12EA6461ULL, "plt_dae010pt_amb_docks.nss"},
	{0x635E4FAF69EF5217ULL, "t3_wpn_ibk.tnt"},
	{0x635E61C576E17C9AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed.mmh"},
	{0x635EFAB7AE94408BULL, "6087244_m.fxe"},
	{0x635FC3A209080AB3ULL, "6154759_m.fxe"},
	{0x63606F335417E8B7ULL, "dae011_conspiracy_mob6149585.cl"},
	{0x6360A402C3C28077ULL, "hlac_w_encounter2_65536_0x.dds"},
	{0x63610A8DCBD8D895ULL, "mcr370_meredith6081098.cl"},
	{0x63621B0D0D6CAFBFULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.cess.ani"},
	{0x636242F83C62042CULL, "6133833_f.fxe"},
	{0x6362CA6CA07C5313ULL, "vxw_woundedcoast_55_0_0.vlm"},
	{0x6362F0656792EB1CULL, "vxht_0_hightown_a2_bc6_0_0.vlm"},
	{0x6363E1657A1C432EULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_act_c.impa.evt"},
	{0x6363EFA1256A2A95ULL, "rdr110dg_party_comments6094674.cl"},
	{0x636409F999A2A0F3ULL, "plt_cod_ltr_waywrd_feynriel.nss"},
	{0x63648B70A7CDC598ULL, "6136024_m.fxe"},
	{0x63650D8A87EAA262ULL, "vxca_cavern1_haunted_i7j_0_0.vlm"},
	{0x6365240F5FE51EA8ULL, "43091454.ogg"},
	{0x63655630982DC98AULL, "lt_0_lowtown_n_44_0_lc.mmh"},
	{0x636576CA5689E2BBULL, "vxdr_primevaltomb_8n5_0_0.vlm"},
	{0x6365D155F1AE2045ULL, "vxlt_0_lowtown_a3_n_c_fhr_0_0.vlm"},
	{0x6365D9C67142061FULL, "vxw_bonepitmining_a2_fc_0_0.vlm"},
	{0x63660F468380C5A8ULL, "vxht_playershouse_n_1qi_0_0.vlm"},
	{0x63663BCAB705CBAAULL, "ca_cave_mmap.dds"},
	{0x6366ACCAF208AF2CULL, "art\\vfx\\combat\\vfx_m_cre_conj_c.mmh"},
	{0x6366BE9A900C8FCDULL, "vxlt_0_lowtown_a2_n_e20_0_0.vlm"},
	{0x63674C1A9E392D67ULL, "6061781_m.fxe"},
	{0x636774DCA71F9459ULL, "6089994_m.fxe"},
	{0x63679D529B5116FAULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgte.msh"},
	{0x6367AB529B512E33ULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgte.mao"},
	{0x6367AF529B513540ULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgte.mmh"},
	{0x6367DCFF794C8DE5ULL, "6032476_f.fxe"},
	{0x6368579722E45481ULL, "246766448.ogg"},
	{0x6368F3FD86C9BBEBULL, "416625096.ogg"},
	{0x63691105519DCFCEULL, "353168505.ogg"},
	{0x6369475BC1D9C155ULL, "117302115.ogg"},
	{0x63699A097F81E59FULL, "344542179.ogg"},
	{0x6369AA952691B517ULL, "mag101_grace6034066.cl"},
	{0x636A097794EA17CCULL, "vxlt_blackemporium_nx_0_1.vlm"},
	{0x636A0A1826C87312ULL, "art\\vfx\\creatures\\vfxroot.impa.evt"},
	{0x636A44EE74C47B95ULL, "follower_banter6160807.cl"},
	{0x636A56418691FECCULL, "zz_qun_debug6064342.cl"},
	{0x636A65D9BC30477CULL, "vxca_caverns1_mines_ec4_0_0.vlm"},
	{0x636A983F6693D324ULL, "and230st_mage_girl_anders.stg"},
	{0x636AD8657A2252D9ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_act_c.impa.gad"},
	{0x636B38485B0EF8E9ULL, "6147717_m.fxe"},
	{0x636B3F5289C724BAULL, "zz_followers_banter6159174.cl"},
	{0x636B43BB3ACBE125ULL, "353395548.ogg"},
	{0x636B4D53B5240EFBULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rtlb.ani"},
	{0x636B71E59A344318ULL, "6074326_m.fxe"},
	{0x636BB456830F1AB0ULL, "533911680.ogg"},
	{0x636C640C5D6EC090ULL, "vxht_keep_n_3zw_0_0.vlm"},
	{0x636DAF23240EDC1FULL, "follower_banter6171876.cl"},
	{0x636E028B69D69D1FULL, "960428815.ogg"},
	{0x636E0B54E6B4EEE8ULL, "6161154_m.fxe"},
	{0x636E15EE78BE4E02ULL, "6065920_m.fxe"},
	{0x636E620B08022AC1ULL, "art\\vfx\\combat\\vfx_gib_boo_meda.mao"},
	{0x636E9E4DB1A4652BULL, "mcr310_sibling6076353.cl"},
	{0x636F12E2C7512CB0ULL, "hlac_ca_cavern1_haunted_24426_4z.dds"},
	{0x636F28776C3C7DD2ULL, "847868667.ogg"},
	{0x636F52AFDCD63BDCULL, "vxw_bonepit_1kn_0_0.vlm"},
	{0x636F83BC6016A479ULL, "fxp_pillar_sym04.dds"},
	{0x636FE774D96AEC91ULL, "character_character_p88.psh.fxo.dx11"},
	{0x63707BA6178F4804ULL, "610981038.ogg"},
	{0x63712633ADB256B8ULL, "vxlt_0_lowtown_a2_n_eud_0_0.vlm"},
	{0x6371291826CEC6A5ULL, "art\\vfx\\creatures\\vfxroot.impa.gad"},
	{0x63714A2BC3E0B305ULL, "art\\characters\\playercharacter\\qunarimale\\har_com.mao"},
	{0x63714B595C47DE19ULL, "and121_karl6093900.cl"},
	{0x6371538AFD86FB96ULL, "913856105.ogg"},
	{0x6371721DF83DE441ULL, "528899112.ogg"},
	{0x6371A336E79AFB36ULL, "6062378_m.fxe"},
	{0x6371C887B1ED4E36ULL, "289631810.ogg"},
	{0x6371E9657A288E27ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_act_c.impa.ani"},
	{0x6372C0C4C59F0033ULL, "697069909.ogg"},
	{0x6372C1C783B1C88FULL, "zz_and_debug6134120.cl"},
	{0x6372CCBD26DB1166ULL, "dae000_sandal.cnv"},
	{0x6372ED30997CE2EAULL, "art\\characters\\playercharacter\\humanmale\\cth_comb.mao"},
	{0x6372F130997CE939ULL, "art\\characters\\playercharacter\\humanmale\\cth_comb.mmh"},
	{0x6372FB30997CFA27ULL, "art\\characters\\playercharacter\\humanmale\\cth_comb.msh"},
	{0x63730D65236538AFULL, "vxw_blightlands_ge_0_0.vlm"},
	{0x63732F68ED97E44DULL, "121939295.ogg"},
	{0x637391276389CAD1ULL, "lla_ht_0_hightown_-65535_0.dds"},
	{0x63744BA363381E9FULL, "vxw_bonepitmining_a2_mb_0_0.vlm"},
	{0x6374795599992001ULL, "694409436.ogg"},
	{0x637487B5D052B156ULL, "vxht_0_hightown_a2_baf_0_0.vlm"},
	{0x6374C6E5AD716943ULL, "207506959.ogg"},
	{0x6374E5715EA03064ULL, "gen00fl_carver6103888.cl"},
	{0x63758870B9070C5BULL, "vxlt_gallowstemplar_f_356_0_0.vlm"},
	{0x6375BF4FF34A366FULL, "hlac_ca_caverns2prc_5802_57y.dds"},
	{0x6375CA4F8B7A8858ULL, "llac_w_sundermount_-196606_0.dds"},
	{0x637661BCD52FABFFULL, "vxw_sundermount_27e_0_0.vlm"},
	{0x637677A2C739AFA4ULL, "6165367_m.fxe"},
	{0x6376CEBD26DEFD26ULL, "dae000_sandal.bnk"},
	{0x637734B48CCE9063ULL, "llac_lt_0_lowtown_docks_n_196606_0.dds"},
	{0x63775AE96C41DF7EULL, "6217703_m.fxe"},
	{0x63778894D255B289ULL, "944993814.ogg"},
	{0x6377BDBDC0D0F80FULL, "w_blightlands_126_0.mao"},
	{0x6377C9BDC0D10CACULL, "w_blightlands_126_0.mmh"},
	{0x6377CF4816580CAFULL, "one300_nuncio6076727.cl"},
	{0x6377CFBDC0D1169EULL, "w_blightlands_126_0.msh"},
	{0x63780BF94A7FFB4EULL, "6170613_m.fxe"},
	{0x637813CA75055467ULL, "dae200_cullen6140257.cl"},
	{0x6378252A7067598DULL, "gen00fl_merrill6150962.cl"},
	{0x63783716DCAB8937ULL, "zz_dae_debug6029791.cl"},
	{0x637894BDCD58453FULL, "484328318.ogg"},
	{0x6378AE73E8A56157ULL, "72919515.ogg"},
	{0x6378D8407C002F17ULL, "mi_wispy.mao"},
	{0x63793CCA97FB00E7ULL, "hlac_w_woundedcoast_-131068_0x.dds"},
	{0x63795E16F4B2E0EAULL, "hlac_lt_0_lowtown_docks_n_-65538_5x.dds"},
	{0x637A7C5C8A487A4FULL, "vxlt_backalley_g_1w3_0_0.vlm"},
	{0x637AFEF3898EFEC1ULL, "gen_im_trs_trs_two_2hs_r1.uti"},
	{0x637BD1DA1EB966A5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldb_1.mao"},
	{0x637BD5523CEAE6DEULL, "674756890.ogg"},
	{0x637C022635F08212ULL, "one350_variel6087045.cl"},
	{0x637C2A3CFCC88DB2ULL, "901363934.ogg"},
	{0x637C3264FFA273DDULL, "gen00fl_anders6061152.cl"},
	{0x637C56CD193B28D9ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd04a.msh"},
	{0x637C60CD193B39C7ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd04a.mmh"},
	{0x637C93C17E092E3DULL, "422449266.ogg"},
	{0x637CB90444A75571ULL, "6140860_m.fxe"},
	{0x637D706699AE71DBULL, "art\\vfx\\combat\\vfx_c_pride_s_p.cess.evt"},
	{0x637DA6AE7827EF9EULL, "6119414_m.fxe"},
	{0x637DC48CCECC545CULL, "vxw_bonepit_1lo_0_0.vlm"},
	{0x637DDABF84BA2FF9ULL, "161336858.ogg"},
	{0x637E19AB634F83B6ULL, "dae000_fenarel.cnv"},
	{0x637F1C14A824A13DULL, "hlac_lt_0_lowtown_a2_n_c_0_3z.dds"},
	{0x637F7453B534F985ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rtlb.gad"},
	{0x638030D0046D80B7ULL, "w_bonepitmining_a2_188_0.mmh"},
	{0x638034D0046D87C4ULL, "w_bonepitmining_a2_188_0.mao"},
	{0x638046D0046DA629ULL, "w_bonepitmining_a2_188_0.msh"},
	{0x638084FA95213D18ULL, "w_blightlands_11_0.rml"},
	{0x638087FB00F8F262ULL, "vxw_bonepitdestroyed_a3_1ar_0_0.vlm"},
	{0x63811BAB6351BC96ULL, "dae000_fenarel.bnk"},
	{0x63817F42155B512FULL, "vxw_blightlands_hd_0_0.vlm"},
	{0x6381A3C79B8E2809ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic_2.mao"},
	{0x63821EF283D18DC5ULL, "6034301_m.fxe"},
	{0x63825A843A7613BBULL, "vxdr_primevaltomb_8d7_0_0.vlm"},
	{0x63825E8D38DCE623ULL, "6119306_f.fxe"},
	{0x63829CEF2D44C89FULL, "art\\characters\\playercharacter\\qunarimale\\uni_magb.mmh"},
	{0x6382A2EF2D44D341ULL, "art\\characters\\playercharacter\\qunarimale\\uni_magb.msh"},
	{0x6382DC8A686E348DULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srdb_0s.dds"},
	{0x6382DE0A65790A63ULL, "hm_gencr_merc_warrior.mor"},
	{0x6382FA843059CDCCULL, "vxw_sundermount_346_0_0.vlm"},
	{0x6383B3802E25D93DULL, "908921305.ogg"},
	{0x6383E274C2F372D0ULL, "vxlt_undercitydungeon_zw_0_0.vlm"},
	{0x6384696699B4841CULL, "art\\vfx\\combat\\vfx_c_pride_s_p.cess.gad"},
	{0x6384907C7DD36325ULL, "80431223.ogg"},
	{0x6384D399C725C47FULL, "vxw_sundermount_28d_0_0.vlm"},
	{0x63864D25896B3731ULL, "vxlt_0_lowtown_a2_n_c_hrs_0_0.vlm"},
	{0x6386E66908AE2533ULL, "6084469_m.fxe"},
	{0x6386F25F631A4947ULL, "vxht_0_hightown_a3_bzg_0_0.vlm"},
	{0x63873317FC9BD8E8ULL, "6030447_m.fxe"},
	{0x6387C57F059508DBULL, "vxht_0_hightown_n_9zn_0_0.vlm"},
	{0x6387CE506E4D91BAULL, "vxdr_primevaltomb_3js_0_0.vlm"},
	{0x638888C33626417CULL, "vxht_mansion_n_53z_0_0.vlm"},
	{0x6388ED9849C6DABCULL, "460024088.ogg"},
	{0x638911509D65AE6CULL, "hlac_w_woundedcoast_-196600_0z.dds"},
	{0x63892202B53727EBULL, "follower_banter6171029.cl"},
	{0x63899CDA7775F97CULL, "6154411_m.fxe"},
	{0x6389B38D4661649BULL, "111173852.ogg"},
	{0x6389C487F2132144ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_coma_0d.dds"},
	{0x638A16D8A774C7B1ULL, "w_sundermount_248_0.mmh"},
	{0x638A22D8A774DC52ULL, "w_sundermount_248_0.mao"},
	{0x638A30D8A774F46FULL, "w_sundermount_248_0.msh"},
	{0x638A3D6EB9039209ULL, "follower_banter6136732.cl"},
	{0x638AAC8267B4F62CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdmgfront.ani"},
	{0x638AEEC0F23F984DULL, "vxca_caverns1_mines_j5p_0_0.vlm"},
	{0x638B62D524BD6C8FULL, "147826733.ogg"},
	{0x638B686699BAA11AULL, "art\\vfx\\combat\\vfx_c_pride_s_p.cess.ani"},
	{0x638B6B37C45F9695ULL, "516410071.ogg"},
	{0x638BF5333CAA40B9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicd.phy"},
	{0x638C3B058E355EB7ULL, "6055625_m.fxe"},
	{0x638C8C437C8637F9ULL, "gen00fl_aveline6139926.cl"},
	{0x638D4C3A49BFB9A8ULL, "966197248.ogg"},
	{0x638D50EA2FBF8DA5ULL, "hlac_lt_0_lowtown_a3_n_c_65534_3z.dds"},
	{0x638DC318C65C3759ULL, "hlac_lt_0_lowtown_-131073_0z.dds"},
	{0x638DF0BAA08EDC6DULL, "6180929_m.fxe"},
	{0x638EFA784E73758FULL, "6127182_m.fxe"},
	{0x638F222ACD41A028ULL, "mag110_vincento6032991.cl"},
	{0x638F41813A555E00ULL, "ts_demonic_prog_junk2.gda"},
	{0x638F85D5B36A0CC2ULL, "art\\characters\\playercharacter\\humanfemale\\boo_mede.phy"},
	{0x638FF11F075169AFULL, "vxw_blightlands_mg_0_0.vlm"},
	{0x63905CDBB3F4CC0AULL, "vxw_bonepitdestroyed_a3_1r3_0_0.vlm"},
	{0x639095FC3EA23675ULL, "zz_fol_debug6066144.cl"},
	{0x6391447B6D944472ULL, "vxw_sundermount_2h5_0_0.vlm"},
	{0x63916C61224FE64CULL, "vxw_sundermount_337_0_0.vlm"},
	{0x63918A88B6B05FE7ULL, "character_character_p109.psh.fxo.dx11.assembly"},
	{0x6391B98267BB2B6EULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdmgfront.gad"},
	{0x63921E3917835A30ULL, "781985208.ogg"},
	{0x6392717EB32A8C37ULL, "708595258.ogg"},
	{0x6392F21192022467ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.anb"},
	{0x639313954F24D174ULL, "vxlt_undercityserialkiller_1a6_0_0.vlm"},
	{0x63935CA06A7B4310ULL, "lt_0_lowtown_docks_a3_132_0.phy"},
	{0x6393874D55BE4713ULL, "6122212_f.fxe"},
	{0x6393A6A9EA0792FCULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.impa.ani"},
	{0x639424874C7624C9ULL, "hlac_ht_chantry_n_2534_23x.dds"},
	{0x6394319D48E1688BULL, "llac_w_bonepitdestroyed_a3_327685_0.dds"},
	{0x6394B28E6936F9C7ULL, "mag100_cullen6140956.cl"},
	{0x6394E4F3B7AB7552ULL, "73499615.ogg"},
	{0x639542592D6D2E35ULL, "6180972_m.fxe"},
	{0x63954BCC03AB9D61ULL, "vxht_0_hightown_9aj_0_0.vlm"},
	{0x6395643C551061C7ULL, "vxht_0_hightown_a3_bqf_0_0.vlm"},
	{0x63958189FE5238C2ULL, "w_bonepitdestroyed_a3_188_0.phy"},
	{0x6395E70B8252030FULL, "llac_ca_caverns1_24424_28.dds"},
	{0x63966ADCC1DA3E6CULL, "6097387_m.fxe"},
	{0x63966FCB636D55D1ULL, "zz_and_debug6178566.cl"},
	{0x639688A0D950CE0CULL, "6095382_f.fxe"},
	{0x6396D1DECF2AC9FDULL, "vxlt_gallowstemplar_f_4xx_0_0.vlm"},
	{0x6397124C785BD204ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mcn.mmh"},
	{0x6397164C785BD957ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mcn.mao"},
	{0x6397284C785BF7A6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mcn.msh"},
	{0x6397992678ADF83BULL, "hlac_lt_0_lowtown_a2_-4_0y.dds"},
	{0x6397D88676F763C7ULL, "vxlt_0_lowtown_a2_n_hen_0_0.vlm"},
	{0x6397E131944A88F6ULL, "6177915_m.fxe"},
	{0x6397E680B24E6508ULL, "sky_sky2_p1.vsh.fxo"},
	{0x63986E28B9C6A618ULL, "zz_approval_debug6003647.cl"},
	{0x639870397A969A06ULL, "vxca_caverns1_mines_i32_0_0.vlm"},
	{0x6398B48267C1414DULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdmgfront.evt"},
	{0x6398BC22E6F226C7ULL, "gen_im_wep_rog_dua_act2_08.uti"},
	{0x6398FA8D3791C1D0ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shielda.mao"},
	{0x6398FE8D3791C8E3ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shielda.mmh"},
	{0x63990C8D3791E03DULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shielda.msh"},
	{0x6399E0B66A6829E3ULL, "vxlt_0_lowtown_bys_0_0.vlm"},
	{0x6399E9FE5A68A689ULL, "dae110ip_store_weapons.utp"},
	{0x639A0EAA570D9231ULL, "vxlt_hangedman_n_28c_0_0.vlm"},
	{0x639A59A64F546E0DULL, "319448596.ogg"},
	{0x639A6E8B40986897ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_homa_0d.dds"},
	{0x639AB3A9EA0DC73EULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.impa.gad"},
	{0x639ACA60909E9425ULL, "seb221_flora_harimann6177733.cl"},
	{0x639B069028C2B122ULL, "6180918_f.fxe"},
	{0x639BD1D7ECCA97B5ULL, "226982815.ogg"},
	{0x639C53364FDAC2E7ULL, "6127627_m.fxe"},
	{0x639C86466A3F9606ULL, "genlock01.mao"},
	{0x639C912CAE97539BULL, "gen_im_wep_rog_dua_lore_02.uti"},
	{0x639D43DBF18A9086ULL, "6050098_m.fxe"},
	{0x639D5E2C1AD2BACCULL, "follower_banter6141812.cl"},
	{0x639DD45301067D63ULL, "230807423.ogg"},
	{0x639DE8A52608303AULL, "352149185.ogg"},
	{0x639E42BB9D1FA1AEULL, "6075617_m.fxe"},
	{0x639E62FBF947822FULL, "vxw_blightlands_vf_0_0.vlm"},
	{0x639E6961C08E4ED1ULL, "969470206.ogg"},
	{0x639EC8EC43CF8578ULL, "6033553_f.fxe"},
	{0x639ED481C5004948ULL, "vxht_0_hightown_9ao_0_0.vlm"},
	{0x639EF152B590EF5AULL, "228123200.ogg"},
	{0x639F82FA953BEAF5ULL, "w_blightlands_11_0.msh"},
	{0x639F84FA953BEE6BULL, "w_blightlands_11_0.mmh"},
	{0x639F90FA953C0208ULL, "w_blightlands_11_0.mao"},
	{0x639F9395B9AB49B2ULL, "mag170_thrask6171437.cl"},
	{0x639FB6585F8A5CF2ULL, "vxw_sundermount_2m6_0_0.vlm"},
	{0x63A03118D1F47C51ULL, "vxht_0_hightown_a3_3xr_0_0.vlm"},
	{0x63A03583A30AC106ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartodual.ani"},
	{0x63A04CD2AD6D97C4ULL, "6156376_m.fxe"},
	{0x63A0F515968098CEULL, "6141815_m.fxe"},
	{0x63A169CE9F3304FEULL, "vxht_0_hightown_n_2xj_0_0.vlm"},
	{0x63A18EA9EA13A75DULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.impa.evt"},
	{0x63A1C98072CF43F7ULL, "6182297_m.fxe"},
	{0x63A227AA45BA0CEDULL, "6143339_m.fxe"},
	{0x63A2B63CEF0E5CADULL, "562369736.ogg"},
	{0x63A33997E1630C02ULL, "6146995_m.fxe"},
	{0x63A426A32EC64CC6ULL, "46287816.ogg"},
	{0x63A45A4DBBDF23DEULL, "vxw_bonepitmining_a2_80_0_0.vlm"},
	{0x63A45B778B0FECF1ULL, "vxht_brothel_n_4vv_0_0.vlm"},
	{0x63A4B4E699F5F7FAULL, "6162104_m.fxe"},
	{0x63A567F3FA5EC872ULL, "follower_banter6143416.cl"},
	{0x63A56C7D1A12727CULL, "vxht_mansion_n_59x_0_0.vlm"},
	{0x63A5E4C8167F2EFAULL, "865192892.ogg"},
	{0x63A5F8AFB0333432ULL, "392225195.ogg"},
	{0x63A67EE17282BA33ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.anb"},
	{0x63A688A9BEE675D9ULL, "cli_army_ds_gatesatt.gda"},
	{0x63A6AFDC44405F2DULL, "69894944.ogg"},
	{0x63A705D5FD33ADACULL, "vxw_encounter3_fm_0_0.vlm"},
	{0x63A7184EE0637AD1ULL, "926888804.ogg"},
	{0x63A721043F0A46E1ULL, "vxht_chantry_n_26o_0_0.vlm"},
	{0x63A7588D9794F934ULL, "669133615.ogg"},
	{0x63A7CB75BC3C8771ULL, "vxht_keep_n_5xj_0_0.vlm"},
	{0x63A7E5EFE03C2B74ULL, "vxlt_0_lowtown_docks_a3_czg_0_1.vlm"},
	{0x63A82BF99F80CE1CULL, "vxlt_0_lowtown_a2_hd5_0_0.vlm"},
	{0x63A8879B59DB7E06ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em03.phy"},
	{0x63A88939FBA16F40ULL, "13616054.ogg"},
	{0x63A88E1DB7AB6AFEULL, "vxlt_0_lowtown_docks_n_bbm_0_0.vlm"},
	{0x63A8CD14045C716CULL, "character_character_p151.psh.fxo.dx11.assembly"},
	{0x63A913B0CAAEF3EDULL, "character_character_p91.psh.fxo"},
	{0x63A928F61F4AE0BBULL, "6113014_f.fxe"},
	{0x63A949CD196200FCULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd04a.phy"},
	{0x63A97C5F237765F7ULL, "vxlt_0_lowtown_docks_b92_0_0.vlm"},
	{0x63A9F83901368F41ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.anb"},
	{0x63A9FB6C7B2D037AULL, "vxlt_hangedman_n_1xo_0_0.vlm"},
	{0x63AB85736492F4B1ULL, "561176052.ogg"},
	{0x63ABB9C1D5071D64ULL, "vxw_bonepitdestroyed_a3_25s_0_0.vlm"},
	{0x63ABC2451A830F68ULL, "vxht_mansion_n_502_0_1.vlm"},
	{0x63AC17AA6B80C301ULL, "zz_prm_debug.ncs"},
	{0x63AC317BB0F94374ULL, "zz_add_all_warrior.ncs"},
	{0x63ADF3D8D7050156ULL, "6044172_m.fxe"},
	{0x63ADF6F7EC560C6EULL, "vxlt_0_lowtown_docks_b53_0_0.vlm"},
	{0x63AE020744153304ULL, "6075840_m.fxe"},
	{0x63AE19FC5947DE5AULL, "fen110_anso6206191.cl"},
	{0x63AE1D83A316D677ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartodual.evt"},
	{0x63AE532AC382D5DBULL, "static_static_p84.psh.fxo.dx11"},
	{0x63AE5CC33257973DULL, "vxlt_0_lowtown_a3_n_c_hff_0_0.vlm"},
	{0x63AEC8C977B9EE4CULL, "6100029_m.fxe"},
	{0x63AF9D35A92CBFB0ULL, "art\\characters\\playercharacter\\humanmale\\w_stf_001a.mmh"},
	{0x63AFAB35A92CD7CAULL, "art\\characters\\playercharacter\\humanmale\\w_stf_001a.msh"},
	{0x63AFD0393E2CE9FEULL, "zz_ave_debug6207530.cl"},
	{0x63AFD88979AB49ADULL, "6026204_f.fxe"},
	{0x63AFDBD11B2D240FULL, "dre000_vampire6150368.cl"},
	{0x63AFEF9EF5B1D2E6ULL, "319446557.ogg"},
	{0x63B05F1F1F6FC26AULL, "429918353.ogg"},
	{0x63B07714774CCA8FULL, "876578215.ogg"},
	{0x63B08C6E3A572002ULL, "hlac_lt_0_lowtown_n_-65537_3z.dds"},
	{0x63B0DD887F33063AULL, "zz_followers_banter6158829.cl"},
	{0x63B11F0C468FE193ULL, "hlac_lt_gallowstemplar_2383_10z.dds"},
	{0x63B132A6378A7763ULL, "qun131_taskmaster6059059.cl"},
	{0x63B1AB74D3592482ULL, "6143695_m.fxe"},
	{0x63B1F646186C62ECULL, "follower_banter6155505.cl"},
	{0x63B25A28E9737B14ULL, "446264316.ogg"},
	{0x63B276DEFF9C979FULL, "dae220st_merch_weapon.utm"},
	{0x63B2A94541C1A1A8ULL, "plt_zz_isapt_debug.nss"},
	{0x63B2C6324755C59BULL, "144595919.ogg"},
	{0x63B2CCD51D123464ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgtb.msh"},
	{0x63B2E2D51D125906ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgtb.mmh"},
	{0x63B2FCEA00B0B5D2ULL, "mag110_vincento6051738.cl"},
	{0x63B390AA7C685281ULL, "422417445.ogg"},
	{0x63B39BE1728D759CULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.mmh"},
	{0x63B3F87494BB8986ULL, "vxlt_0_lowtown_docks_n_bgn_0_0.vlm"},
	{0x63B41E83A31B4300ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartodual.gad"},
	{0x63B460D5B388EA76ULL, "art\\characters\\playercharacter\\humanfemale\\boo_mede.mao"},
	{0x63B46CD5B388FE15ULL, "art\\characters\\playercharacter\\humanfemale\\boo_mede.mmh"},
	{0x63B46ED5B389018BULL, "art\\characters\\playercharacter\\humanfemale\\boo_mede.msh"},
	{0x63B47EB4994092AEULL, "w_blightlands_135_0.msh"},
	{0x63B48CB49940AAFFULL, "w_blightlands_135_0.mao"},
	{0x63B498B49940BE9CULL, "w_blightlands_135_0.mmh"},
	{0x63B4C56AD5B83417ULL, "106572297.ogg"},
	{0x63B521D733759402ULL, "951286838.ogg"},
	{0x63B56C4BFB7330B2ULL, "vxht_keep_6gh_0_0.vlm"},
	{0x63B577B2EF29C62CULL, "vxw_encounter3_cb_0_0.vlm"},
	{0x63B590EBAC2241A5ULL, "dae000_meredith6085258.cl"},
	{0x63B5A9288A0DF9CCULL, "6144109_f.fxe"},
	{0x63B5DE9A8C60ACBEULL, "llac_lt_0_lowtown_docks_131070_2.dds"},
	{0x63B5E747D62820DCULL, "hlac_lt_darktown_-1_39x.dds"},
	{0x63B64A27BBF27A64ULL, "hf_agnes.mor"},
	{0x63B67340F79BB5C6ULL, "6037026_m.fxe"},
	{0x63B6C956F543C124ULL, "348653852.ogg"},
	{0x63B6FC7632B7716FULL, "dae210ip_store5.utp"},
	{0x63B754351626F421ULL, "385452839.ogg"},
	{0x63B7627C60B691E3ULL, "388544128.ogg"},
	{0x63B7858DBCA50C6DULL, "character_character_p46.psh.fxo"},
	{0x63B7D55CC72115C2ULL, "172068716.ogg"},
	{0x63B7D939014297BAULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.mmh"},
	{0x63B80B242D46A3CFULL, "vxht_mansion2_2er_0_1.vlm"},
	{0x63B81B8CD48C0D1AULL, "lowtown_fem_prostitute_sta6152312.cl"},
	{0x63B81CA9095DB376ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_noble_female.anb"},
	{0x63B853FADDE0E6B3ULL, "dae300_alistair_king6134949.cl"},
	{0x63B899F28FC71EC7ULL, "545149043.ogg"},
	{0x63B8DE6AF59F9653ULL, "399816391.ogg"},
	{0x63B8FC7A826674BBULL, "zz_qun_debug6028286.cl"},
	{0x63B936F033C6F3E7ULL, "6129827_m.fxe"},
	{0x63B989766D1BD513ULL, "96842983.ogg"},
	{0x63B99331B883A726ULL, "follower_banter6155868.cl"},
	{0x63B9B24541C7CFBBULL, "plt_zz_isapt_debug.plo"},
	{0x63B9C34A08409456ULL, "vxlt_0_lowtown_a2_hfi_0_0.vlm"},
	{0x63B9EAC99805FC16ULL, "6167536_m.fxe"},
	{0x63B9FCFAD014B61CULL, "6066369_m.fxe"},
	{0x63BA5C89FE711676ULL, "w_bonepitdestroyed_a3_188_0.mao"},
	{0x63BA6889FE712A15ULL, "w_bonepitdestroyed_a3_188_0.mmh"},
	{0x63BA6A89FE712D8BULL, "w_bonepitdestroyed_a3_188_0.msh"},
	{0x63BA977E22DD043BULL, "zz_mcr_debug6071549.cl"},
	{0x63BAA3AD77D853EAULL, "vxdr_primevaltomb_7kl_0_0.vlm"},
	{0x63BAEF1192237168ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.mmh"},
	{0x63BB9865ECDB2727ULL, "6208697_f.fxe"},
	{0x63BC4C111C30EC92ULL, "6180841_m.fxe"},
	{0x63BC6E113ADAD18EULL, "one220_yevhen6098726.cl"},
	{0x63BC7424E57CA410ULL, "752991722.ogg"},
	{0x63BCD9D51D1A9ED1ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_lgtb.phy"},
	{0x63BDE577995BDAE5ULL, "614178138.ogg"},
	{0x63BE9E0A8EE0F535ULL, "hlac_w_sundermount_196602_0z.dds"},
	{0x63BECB8E58486041ULL, "vxca_caverns2_8oz_0_0.vlm"},
	{0x63BFE81564919A28ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_7y.dds"},
	{0x63C057A6D637EF6CULL, "38547416.ogg"},
	{0x63C0F5D2F6820493ULL, "hlac_ht_0_hightown_131071_0y.dds"},
	{0x63C0F7CED1BEABBEULL, "dae000_hate_enemies6167552.cl"},
	{0x63C1019B7F0F3413ULL, "vxca_caverns1_e83_0_0.vlm"},
	{0x63C17F3E7AE92B2AULL, "hm_gen_street_thug.mor"},
	{0x63C245659BE5632EULL, "hm_cgn_004.mop"},
	{0x63C272CA98608753ULL, "abi_m_repwave.bnk"},
	{0x63C2C081FD93ECECULL, "hlac_w_encounter2_camp_196608_0z.dds"},
	{0x63C302F6C8620EC2ULL, "mrl210_merrill6166779.cl"},
	{0x63C31EFC3C551DD6ULL, "hlac_lt_0_lowtown_docks_n_-196610_1x.dds"},
	{0x63C345DF6A478114ULL, "fen110_fenris6100297.cl"},
	{0x63C38EA49615205FULL, "vxlt_0_lowtown_docks_cwi_0_0.vlm"},
	{0x63C3A5BC100BFEA8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowlarge.mmh"},
	{0x63C3B1BC100C134BULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowlarge.mao"},
	{0x63C3C3BC100C31D2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowlarge.msh"},
	{0x63C3FFE76D6463B4ULL, "630418194.ogg"},
	{0x63C43441E16BB6B3ULL, "vxw_woundedcoast_1b1_0_0.vlm"},
	{0x63C455DCF862965BULL, "176806973.ogg"},
	{0x63C46FD7757C4E35ULL, "vxlt_warehouse_8v_0_0.vlm"},
	{0x63C4AE8833F27106ULL, "6054515_m.fxe"},
	{0x63C5099776B2624DULL, "545354856.ogg"},
	{0x63C56086B105A34EULL, "716246651.ogg"},
	{0x63C5F76AAE9B24EDULL, "character_character_p37.psh.fxo"},
	{0x63C60FCFDB88C878ULL, "qun210ip_ambient_throne.utp"},
	{0x63C66705840E4BB0ULL, "vxca_caverns1_e7y_0_0.vlm"},
	{0x63C686A45C75A0BDULL, "vxlt_undercitydungeon_1mz_0_0.vlm"},
	{0x63C6DF265F19347AULL, "vxlt_hangedman_n_1ri_0_0.vlm"},
	{0x63C737755286A7A3ULL, "dae000cr_grand_cleric6178471.cl"},
	{0x63C83526FA36ACD6ULL, "vxlt_0_lowtown_a2_hcj_0_0.vlm"},
	{0x63C9158A69CE6C6AULL, "vxdr_primevaltomb_7dm_0_0.vlm"},
	{0x63C983F7CFBBBEBCULL, "llac_lt_0_lowtown_docks_a3_131071_6.dds"},
	{0x63C9BB0594B71F4AULL, "295035125.ogg"},
	{0x63CA04D710BE26C3ULL, "6126772_m.fxe"},
	{0x63CA04F999F94B5CULL, "plt_cod_ltr_waywrd_feynriel.plo"},
	{0x63CA9CAC665763C0ULL, "gen00fl_varric6102799.cl"},
	{0x63CADFBA6BF4D5EDULL, "vxlt_0_lowtown_docks_b4w_0_0.vlm"},
	{0x63CB0A733D4E811EULL, "6154699_m.fxe"},
	{0x63CB7F585E70B17FULL, "6170050_m.fxe"},
	{0x63CB936E77C48229ULL, "dae000_bodhan6129112.cl"},
	{0x63CC34B72E1E2650ULL, "vxlt_0_lowtown_a2_n_c_eom_0_1.vlm"},
	{0x63CC45E57A4C6755ULL, "hlac_w_woundedcoast_-131065_0y.dds"},
	{0x63CC5911FA5DEEF1ULL, "vxlt_0_lowtown_docks_c4b_0_0.vlm"},
	{0x63CCA8316BB267E3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerc_sa.dds"},
	{0x63CDA43E88BA6786ULL, "hlac_ca_caverns1_mines_24423_7x.dds"},
	{0x63CDB00DC54B7A74ULL, "hlac_ht_keep_n_3732_18z.dds"},
	{0x63CE0E746BE64908ULL, "vxlt_warehouse_n_138_0_0.vlm"},
	{0x63CE63B159D3FE3DULL, "283903073.ogg"},
	{0x63CE910B08131998ULL, "240327633.ogg"},
	{0x63CEE9FFBCDA0701ULL, "6120804_m.fxe"},
	{0x63CF21F78FF3AFF6ULL, "vxca_caverns2_7xo_0_0.vlm"},
	{0x63CF9C2A5C760023ULL, "vxca_caverns1_mines_icr_0_1.vlm"},
	{0x63CFACCDFC60827DULL, "6112569_f.fxe"},
	{0x63CFADED2A9DD431ULL, "vxht_0_hightown_a3_9x8_0_0.vlm"},
	{0x63D07A646AE597CEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lgchainlink.mmh"},
	{0x63D08300C15C1D12ULL, "6143471_m.fxe"},
	{0x63D086646AE5AC65ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lgchainlink.mao"},
	{0x63D094646AE5C3BCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_lgchainlink.msh"},
	{0x63D0AE13958A19C3ULL, "6118179_m.fxe"},
	{0x63D0C69EDDF23FA6ULL, "art\\vfx\\cinematic\\vfx_meredith_sword_c.impa.evt"},
	{0x63D0DCB44A72C4B5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chab_n.dds"},
	{0x63D1452AE6E05D8FULL, "hlac_lt_0_lowtown_n_1_5y.dds"},
	{0x63D184E7EA0C5F03ULL, "mag170_hugh6080737.cl"},
	{0x63D1CCD73FA13A04ULL, "war_groundstab.dds"},
	{0x63D27A055A68251CULL, "gut_sh.mao"},
	{0x63D2A61ED361CF33ULL, "vxw_woundedcoast_1m0_0_0.vlm"},
	{0x63D2D252D7EDBD2CULL, "gen00fl_isabela6101079.cl"},
	{0x63D320B6CF29CAFFULL, "llac_lt_0_lowtown_docks_a3_n_-196609_1.dds"},
	{0x63D3210C921ED0F1ULL, "vxht_keep_n_5ko_0_0.vlm"},
	{0x63D322D1833FCF7CULL, "656367980.ogg"},
	{0x63D340B68F375198ULL, "dae000ip_crowd_mf7_talk7.utp"},
	{0x63D3517FF3448F5EULL, "vxlt_hangedman_4q_0_0.vlm"},
	{0x63D469E2A64423F2ULL, "vxlt_0_lowtown_a2_n_cfi_0_0.vlm"},
	{0x63D4E9B8A23789C3ULL, "art\\characters\\weapons\\spears\\w_spr_002a_p.msh"},
	{0x63D4F23099CFB526ULL, "art\\characters\\playercharacter\\humanmale\\cth_comb.phy"},
	{0x63D4F3107FA192C5ULL, "vxlt_hangedman_n_1xd_0_0.vlm"},
	{0x63D4F7B8A237A11DULL, "art\\characters\\weapons\\spears\\w_spr_002a_p.mmh"},
	{0x63D603834E786C38ULL, "vxlt_gallowstemplar_f_38v_0_0.vlm"},
	{0x63D66A611CD4E829ULL, "zz_followers_banter6159579.cl"},
	{0x63D67419EAFC130AULL, "vxdr_primevaltomb_8o3_0_0.vlm"},
	{0x63D681F324DEDB23ULL, "zz_mag_debug6044333.cl"},
	{0x63D6A139A4230077ULL, "hlac_lt_0_lowtown_docks_n_262143_0x.dds"},
	{0x63D6A703EC2CC556ULL, "vxlt_0_lowtown_a2_hhk_0_0.vlm"},
	{0x63D6BDA242656420ULL, "6080134_m.fxe"},
	{0x63D6DD9EDDF6D2F1ULL, "art\\vfx\\cinematic\\vfx_meredith_sword_c.impa.gad"},
	{0x63D84997AC9A2484ULL, "vxca_cave_2e8_0_0.vlm"},
	{0x63D85CFC30654C0DULL, "zz_and_debug6134063.cl"},
	{0x63D8A54D70474DF8ULL, "89416772.ogg"},
	{0x63D8FB59D7E7FC61ULL, "6102916_m.fxe"},
	{0x63D90B4B498BBFD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_left.gad"},
	{0x63DA77B499615667ULL, "w_blightlands_135_0.phy"},
	{0x63DAA5ECB27AC7DAULL, "hlac_ca_caverns2prc_5802_62x.dds"},
	{0x63DAAE91519CE2F2ULL, "pro000_opening6207388.cl"},
	{0x63DACAEEEC540771ULL, "vxlt_0_lowtown_docks_c1a_0_0.vlm"},
	{0x63DB1F18535806E7ULL, "fxe_sp_vol_add.mao"},
	{0x63DB831B8671792DULL, "vxlt_0_lowtown_docks_n_b7a_0_0.vlm"},
	{0x63DC80515DDC6188ULL, "vxlt_warehouse_n_189_0_0.vlm"},
	{0x63DCB899B2127AB9ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpb.msh"},
	{0x63DCB9225F0863BFULL, "gen00fl_aveline6120849.cl"},
	{0x63DCC299B2128BA7ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpb.mmh"},
	{0x63DD2D0FEA53002FULL, "895685996.ogg"},
	{0x63DD44DE239F04E3ULL, "6033135_m.fxe"},
	{0x63DD70C91BE9D014ULL, "90522792.ogg"},
	{0x63DDDE9EDDFCF29FULL, "art\\vfx\\cinematic\\vfx_meredith_sword_c.impa.ani"},
	{0x63DE74B41BC9981AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha04a.mmh"},
	{0x63DE84028EB67ED4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_coma_0t.dds"},
	{0x63DE86B41BC9B760ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha04a.msh"},
	{0x63DE8FCBC29704C3ULL, "vxlt_gallowscourtyard_7fb_0_0.vlm"},
	{0x63DF0E4B49903009ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_left.evt"},
	{0x63DF5FA656119E90ULL, "6149887_f.fxe"},
	{0x63DFC39D8A037C0AULL, "vxlt_0_lowtown_gca_0_0.vlm"},
	{0x63DFFBE335CB70ACULL, "6036526_m.fxe"},
	{0x63E03B426BF4A6EFULL, "hlac_ht_0_hightown_a2_-2_0y.dds"},
	{0x63E06099F980FD2CULL, "art\\levels\\races\\darkspawn\\props\\prp_ogreskull.phy"},
	{0x63E117FBC557E7B3ULL, "vxw_woundedcoast_1h3_0_0.vlm"},
	{0x63E11FD52C99D424ULL, "105158006.ogg"},
	{0x63E1EC394E999580ULL, "gen00fl_fenris6126937.cl"},
	{0x63E2C43F951BAB19ULL, "vxlt_undercitydungeon_17p_0_0.vlm"},
	{0x63E2D5C8C78B1D80ULL, "vxw_bonepitmining_a2_1o0_0_0.vlm"},
	{0x63E310D274818C7CULL, "vxw_encounter2_camp_rb_0_0.vlm"},
	{0x63E36469E95808C7ULL, "ht_mansion_n_3367.rml"},
	{0x63E3816E487CB404ULL, "337953499.ogg"},
	{0x63E4236A16B014FFULL, "hlac_w_sundermount_393213_0y.dds"},
	{0x63E492D544830FB4ULL, "gen00fl_fenris6126865.cl"},
	{0x63E4B548AD1A0E26ULL, "6026969_m.fxe"},
	{0x63E4DEBBF93E99B4ULL, "147067441.ogg"},
	{0x63E501F5CA408395ULL, "vxlt_undercityencounter_140_0_0.vlm"},
	{0x63E515883C2DBCA4ULL, "dae000_pc_home_ambients6208602.cl"},
	{0x63E51A6595ADBC58ULL, "811808396.ogg"},
	{0x63E56EF307D9326DULL, "985237140.ogg"},
	{0x63E5A5EF2D9955E4ULL, "art\\characters\\playercharacter\\qunarimale\\uni_magb.phy"},
	{0x63E77F1B6D548BD2ULL, "vxlt_0_lowtown_docks_a3_d12_0_0.vlm"},
	{0x63E7A30B75AF779EULL, "vxw_sundermount_37a_0_0.vlm"},
	{0x63E7AAB7AB678592ULL, "hlac_w_bonepitmining_a2_1_0y.dds"},
	{0x63E888722E4ED7ACULL, "xx_black.dds"},
	{0x63E8B212D9F7E12CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_basic.msh"},
	{0x63E8B812D9F7EB1EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_basic.mmh"},
	{0x63E8C412D9F7FFB5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_basic.mao"},
	{0x63E8F30D6CB89ACEULL, "597786255.ogg"},
	{0x63E9B2EA4E0CF70EULL, "6132765_m.fxe"},
	{0x63EA20D805DCA41FULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb105a_0d.dds"},
	{0x63EA4D8691E9BABAULL, "vxht_brothel_n_4xk_0_0.vlm"},
	{0x63EA62D8FB47A74DULL, "vxlt_warehouse_ay_0_0.vlm"},
	{0x63EB62606D045F3CULL, "6125518_m.fxe"},
	{0x63EC5D9CDD72958EULL, "hlac_lt_0_lowtown_a2_n_c_-262144_1y.dds"},
	{0x63EC8429FFAE3EF4ULL, "6180296_m.fxe"},
	{0x63EC8CF873AA8FD6ULL, "vxlt_0_lowtown_a2_n_c_eqe_0_1.vlm"},
	{0x63EC9D95327767E9ULL, "6059681_m.fxe"},
	{0x63ECD23CE4BF1A3BULL, "vxht_0_hightown_n_c5e_0_0.vlm"},
	{0x63ECF64B499C4578ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_left.ani"},
	{0x63ED3242880997EFULL, "1001206814.ogg"},
	{0x63ED8A0C64DB4A09ULL, "6120946_f.fxe"},
	{0x63EDBFB86796F011ULL, "450377474.ogg"},
	{0x63EDBFE85C93EB86ULL, "vxlt_0_lowtown_docks_n_bsr_0_0.vlm"},
	{0x63EEA18EDBA65DD2ULL, "gen00fl_isabela_init6175655.cl"},
	{0x63EEB803AECFFA75ULL, "vxlt_warehouse_11p_0_0.vlm"},
	{0x63EEBF5BDC639563ULL, "6144299_m.fxe"},
	{0x63EEE611A1D382B0ULL, "vxw_bonepitmining_a2_1hu_0_0.vlm"},
	{0x63EF059BBB2E0DCDULL, "vxdr_primevaltomb_9d9_0_1.vlm"},
	{0x63EF23F45E09184EULL, "6170402_m.fxe"},
	{0x63EF81048CC9DF15ULL, "472603856.ogg"},
	{0x63EF89D8B74E0033ULL, "vxw_woundedcoast_1s2_0_0.vlm"},
	{0x63F11116C0FD9DAFULL, "331520720.ogg"},
	{0x63F175120160D94CULL, "64442187.ogg"},
	{0x63F182AF6677A4FCULL, "vxw_encounter2_camp_mc_0_0.vlm"},
	{0x63F1F07BC1963F11ULL, "hlac_ca_caverns2_5802_8z.dds"},
	{0x63F2FCFB50B01D16ULL, "vxlt_0_lowtown_n_e38_0_0.vlm"},
	{0x63F3447D3FAC1723ULL, "6106672_f.fxe"},
	{0x63F34B28292E1F00ULL, "waypoint.gda"},
	{0x63F35883EE7C1A6AULL, "hlac_ht_mansion_n_2995_6y.dds"},
	{0x63F37D18992B6748ULL, "6037024_m.fxe"},
	{0x63F3B7DAAF4BE05EULL, "vxlt_0_lowtown_n_ewl_0_0.vlm"},
	{0x63F40880864DA788ULL, "m_qun221ar_chantry.wbd"},
	{0x63F4154C2F65CEC3ULL, "vxca_caverns1_huq_0_0.vlm"},
	{0x63F4BEF678F665F2ULL, "mrl150_merrill6094013.cl"},
	{0x63F4CBB6E2083589ULL, "w_encounter1_49_0.phy"},
	{0x63F50FB55D426CBAULL, "vxlt_0_lowtown_a2_h2a_0_0.vlm"},
	{0x63F538A5E83A8287ULL, "vxw_woundedcoast_1cn_0_0.vlm"},
	{0x63F57F14F1D9F4FCULL, "463447732.ogg"},
	{0x63F597566B6B461EULL, "zz_one_debug6025839.cl"},
	{0x63F5B3E1CAE3C8FAULL, "llac_ht_chantry_2534_16.dds"},
	{0x63F5E74CBEC4A447ULL, "6086184_m.fxe"},
	{0x63F614E867A5901EULL, "vxw_sundermount_32b_0_0.vlm"},
	{0x63F68CA88AAA8B41ULL, "lgt311_shw_mer_mettin6156393.cl"},
	{0x63F6A3129F6A022FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masc_0t.dds"},
	{0x63F8A5001695EE43ULL, "6129240_f.fxe"},
	{0x63F8C6F47CFBA834ULL, "246097534.ogg"},
	{0x63F8DABC76B617B5ULL, "art\\vfx\\models\\vfx_defense.msh"},
	{0x63F8DCBC76B61B2BULL, "art\\vfx\\models\\vfx_defense.mmh"},
	{0x63F8E8BC76B62FC8ULL, "art\\vfx\\models\\vfx_defense.mao"},
	{0x63F9C9899D881E54ULL, "static_static_p32.psh.fxo.dx11.assembly"},
	{0x63F9D32863389E61ULL, "hlac_lt_0_lowtown_a2_131074_0x.dds"},
	{0x63FA9EE22763A8E1ULL, "286579899.ogg"},
	{0x63FB4419D6B532BBULL, "vxht_0_hightown_n_c2d_0_0.vlm"},
	{0x63FB69A7FD4E1D78ULL, "6005826_f.fxe"},
	{0x63FBE575B3818546ULL, "6097587_m.fxe"},
	{0x63FCA8257BD12B5CULL, "vxw_woundedcoast_1r_0_0.vlm"},
	{0x63FD7EF5F13FD893ULL, "mag311_emile_de_launcet6043763.cl"},
	{0x63FE0907BB7F1EC2ULL, "390489733.ogg"},
	{0x63FE6D97F6133ACAULL, "hlac_dr_primevalentrance_1726_6z.dds"},
	{0x63FE9BCCC592888AULL, "494277166.ogg"},
	{0x63FEA46A403CC1A0ULL, "hlac_ht_bartrand_n_2995_4x.dds"},
	{0x63FEBE324A3BEAFBULL, "vxlt_0_lowtown_docks_n_b9i_0_0.vlm"},
	{0x63FEDA8E862509D0ULL, "6197506_m.fxe"},
	{0x63FFA0B85854C202ULL, "hlac_ht_0_hightown_a3_-65536_2x.dds"},
	{0x6400151000918630ULL, "w_sundermount_174_0.mao"},
	{0x6400191000918D43ULL, "w_sundermount_174_0.mmh"},
	{0x640027100091A49DULL, "w_sundermount_174_0.msh"},
	{0x64003332DD4D1D86ULL, "p_imp_clo.mao"},
	{0x64004CD8A572B0E6ULL, "855314332.ogg"},
	{0x6400EF7983002BD1ULL, "one300_zevran6089738.cl"},
	{0x640132E6BC171D6FULL, "dae200_bran6094714.cl"},
	{0x640148B5448B164DULL, "6215056_m.fxe"},
	{0x6402220142A0D5C7ULL, "352642175.ogg"},
	{0x64028729215BE743ULL, "vxca_caverns1_hzp_0_0.vlm"},
	{0x64029A65A8B40696ULL, "gen00fl_player6207812.cl"},
	{0x6402BCF60C07D7D8ULL, "vxht_mansion_n_6oe_0_0.vlm"},
	{0x640303DB2E5025F4ULL, "6106666_f.fxe"},
	{0x64032F7E50B525AAULL, "ht_brothel_n_9156.rml"},
	{0x640393B41BE8E925ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha04a.phy"},
	{0x6403AA82DA309B07ULL, "vxw_woundedcoast_1ni_0_0.vlm"},
	{0x6403D9B301F7B795ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_keepthin_1.mao"},
	{0x6403F7FE063B94D7ULL, "6082302_m.fxe"},
	{0x6404A0AAAE7A5238ULL, "hlac_ht_0_hightown_n_-196608_1z.dds"},
	{0x6404CE958AB7020CULL, "lgt110_gxa_vgk_conspirator4.cnv"},
	{0x6405C57B2A23A1B9ULL, "hlac_ht_0_hightown_131071_1x.dds"},
	{0x6405E23AC247AD94ULL, "vxlt_gallowstemplar_6dh_0_0.vlm"},
	{0x64068E466B8AE224ULL, "cod_hst_kirk_arch_1a_3.utp"},
	{0x6406D45E48CCA0FFULL, "vxw_sundermount_2eo_0_0.vlm"},
	{0x6406E8716F667429ULL, "vxlt_0_lowtown_a3_n_c_gz5_0_0.vlm"},
	{0x64071CEF340F0FC3ULL, "6140127_m.fxe"},
	{0x64074692DF33D84DULL, "vxlt_warehouse_ow_0_0.vlm"},
	{0x6407A3671836D4BCULL, "w_bonepitdestroyed_a3_25_0_ll.mmh"},
	{0x6407CB250D2320FFULL, "6016493_m.fxe"},
	{0x6408A2DAB909D1A0ULL, "36277508.ogg"},
	{0x6408AB99B2379FDCULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpb.phy"},
	{0x6408B0958ABAB83CULL, "lgt110_gxa_vgk_conspirator4.bnk"},
	{0x6408E6C12E1E8302ULL, "gen00fl_merrill6129484.cl"},
	{0x6408FEC8BB34AD66ULL, "824193601.ogg"},
	{0x6409145A7FA3E4A3ULL, "vxca_caverns2_801_0_0.vlm"},
	{0x64098D34A76222BFULL, "hlac_ca_caverns1_mines_24421_4z.dds"},
	{0x640990CA989CBCB1ULL, "abi_m_repwave.wbd"},
	{0x6409A102A943C256ULL, "6087053_f.fxe"},
	{0x6409A54A07AC0609ULL, "vxdr_primevaltomb_7zq_0_1.vlm"},
	{0x6409C77356B46FB2ULL, "hightown_fem_courtesan_mob6168914.cl"},
	{0x640A176A09FC9DE3ULL, "sheild.dds"},
	{0x640A183A9A42537CULL, "al_isabela_talents.gda"},
	{0x640AB73C8F643A3AULL, "art\\characters\\playercharacter\\children\\mt_bas_km.mmh"},
	{0x640AC93C8F645880ULL, "art\\characters\\playercharacter\\children\\mt_bas_km.msh"},
	{0x640AE6D251D05B29ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggera_d.dds"},
	{0x640B16EC100531EFULL, "llac_w_blightlands_393212_0.dds"},
	{0x640B4A078AD60677ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_diagonal.mmh"},
	{0x640B4E078AD60C84ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_diagonal.mao"},
	{0x640B579ABBCB02D8ULL, "one200ip_st_ilen.utp"},
	{0x640B60078AD62BE9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_diagonal.msh"},
	{0x640CAE7919643299ULL, "gen00fl_aveline6132765.cl"},
	{0x640CE11CA34E853EULL, "6059259_m.fxe"},
	{0x640DDD4B38E80E87ULL, "17243855.ogg"},
	{0x640E25E6FF577ECCULL, "392521941.ogg"},
	{0x640E66694A63D5FCULL, "vxw_encounter2_camp_ca_0_0.vlm"},
	{0x640ED2E0A923E3AAULL, "6139845_m.fxe"},
	{0x640EDFBB8CCE2F7EULL, "6116006_m.fxe"},
	{0x640EEAA1407EF1C8ULL, "vxlt_hangedman_n_1t9_0_0.vlm"},
	{0x640F16C71E303014ULL, "6116145_m.fxe"},
	{0x640F440EB94C6D3BULL, "qun231_party_comment6077008.cl"},
	{0x640FAA0B453855C1ULL, "vxlt_gallowstemplar_5vt_0_0.vlm"},
	{0x641041CDA7CA2902ULL, "6182018_m.fxe"},
	{0x64104527191B45FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_sprint.evt"},
	{0x64109B949338115EULL, "vxlt_0_lowtown_n_emn_0_0.vlm"},
	{0x6410C6FAB7565EC3ULL, "264299634.ogg"},
	{0x6410D2947C88C8BBULL, "337321161.ogg"},
	{0x6411CFAC2E9368ABULL, "477427217.ogg"},
	{0x6411D2C881AB0A90ULL, "vxlt_hangedman_n_30o_0_0.vlm"},
	{0x64121C5FCC26B387ULL, "vxw_woundedcoast_1ih_0_0.vlm"},
	{0x6412FFD62CAA54F1ULL, "470302072.ogg"},
	{0x64136790AEAE5501ULL, "vxlt_0_lowtown_docks_a3_cxu_0_0.vlm"},
	{0x6413BD8405F71981ULL, "vxw_bonepit_146_0_0.vlm"},
	{0x6413CE70D2353DB8ULL, "238408317.ogg"},
	{0x6413D3C871205E61ULL, "vxlt_hangedman_n_1q3_0_0.vlm"},
	{0x64141AD40FFFBFB5ULL, "ave100_donnic6105943.cl"},
	{0x64148C1DA10567F4ULL, "671389413.ogg"},
	{0x6414CBCF914C909FULL, "22818408.ogg"},
	{0x6415040FD25559C9ULL, "hlac_lt_0_lowtown_docks_a3_-131073_4z.dds"},
	{0x6415463B3AC2B97FULL, "vxw_sundermount_2nn_0_0.vlm"},
	{0x6415608BF26D3066ULL, "6064659_m.fxe"},
	{0x6415CF5932450621ULL, "708677983.ogg"},
	{0x6415FE77D9AA459EULL, "62723176.ogg"},
	{0x64170CC23AAC0D8DULL, "vxht_mansion2b_2ko_0_0.vlm"},
	{0x641710FDC355D107ULL, "vxlt_0_lowtown_a2_n_eod_0_1.vlm"},
	{0x64174FEB1690F089ULL, "6096792_m.fxe"},
	{0x64175987B0A88482ULL, "hlac_lt_0_lowtown_a2_n_65534_1z.dds"},
	{0x64175C2719218B29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_sprint.gad"},
	{0x641774C7DE984AA0ULL, "vxht_keep_n_5ja_0_0.vlm"},
	{0x6418077C91364ED5ULL, "gen00fl_merrill6128662.cl"},
	{0x64185536A051626BULL, "gencr_raider_reaver_f.utc"},
	{0x641886DA125161F7ULL, "6151705_m.fxe"},
	{0x64188F3A87B9730AULL, "qun131_party_comment6117278.cl"},
	{0x6418E755E63C5A18ULL, "hlac_lt_0_lowtown_a2_65534_5y.dds"},
	{0x64191EE5A83011B3ULL, "317736404.ogg"},
	{0x6419BA047E3E83C9ULL, "vxlt_0_lowtown_a2_n_c_gpf_0_0.vlm"},
	{0x6419CF9703C0E88DULL, "6133627_f.fxe"},
	{0x641A9B404B69289BULL, "lt_0_lowtown_docks_n_27_0_ll.mmh"},
	{0x641A9E7E562EFD34ULL, "vxlt_undercityencounter_1bw_0_0.vlm"},
	{0x641AC3041BFE2B77ULL, "hlac_w_encounter3_-196612_0x.dds"},
	{0x641B007EC8D97A65ULL, "6029358_m.fxe"},
	{0x641B661C9A716088ULL, "hlac_lt_0_lowtown_a2_131070_3x.dds"},
	{0x641C778F6159DB69ULL, "dae000ip_crowd_lm5_beg.utp"},
	{0x641C8CD7FC0DC4D7ULL, "ave221_brennan6110481.cl"},
	{0x641CA50586A83F5BULL, "hlac_lt_gallowstemplar_n_2395_0y.dds"},
	{0x641CCDAC78A602BCULL, "w_bonepit_30_0.mao"},
	{0x641CD9AC78A6175FULL, "w_bonepit_30_0.mmh"},
	{0x641CDFAC78A62101ULL, "w_bonepit_30_0.msh"},
	{0x641D3B824034921CULL, "342238527.ogg"},
	{0x641D624B70B081FFULL, "6020653_f.fxe"},
	{0x641DDF18459F7129ULL, "hlac_lt_0_lowtown_docks_a3_-65537_1x.dds"},
	{0x641E2D2719275A77ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_sprint.ani"},
	{0x641E8D3D221D3F4CULL, "428985415.ogg"},
	{0x641EA8C0700F5477ULL, "vxht_0_hightown_n_c5q_0_0.vlm"},
	{0x641EC5B1B4EFAB31ULL, "ach_oth_smuggler.dds"},
	{0x641F03008BA64B7DULL, "vxht_keep_67c_0_0.vlm"},
	{0x641F3A3677E37C8FULL, "6072200_m.fxe"},
	{0x6420054DB908F56AULL, "vxlt_0_lowtown_a3_n_c_gnv_0_0.vlm"},
	{0x6420617088A11522ULL, "hlac_lt_0_lowtown_65537_7z.dds"},
	{0x6420B4B6E22D49AEULL, "w_encounter1_49_0.mmh"},
	{0x6420BB0C4D75AB65ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_ulm_basa.dds"},
	{0x6420BC0BA7176F3DULL, "qun221_viscount6069917.cl"},
	{0x6420C0B6E22D5E45ULL, "w_encounter1_49_0.mao"},
	{0x6420CEB6E22D759CULL, "w_encounter1_49_0.msh"},
	{0x64213799F9B84EF7ULL, "art\\levels\\races\\darkspawn\\props\\prp_ogreskull.mmh"},
	{0x64213B99F9B85504ULL, "art\\levels\\races\\darkspawn\\props\\prp_ogreskull.mao"},
	{0x64214D99F9B87469ULL, "art\\levels\\races\\darkspawn\\props\\prp_ogreskull.msh"},
	{0x64215243838C16B8ULL, "6131906_m.fxe"},
	{0x6423B8182CB8D1FFULL, "vxw_sundermount_2si_0_0.vlm"},
	{0x6423D268E46348E6ULL, "6063759_m.fxe"},
	{0x6423F849D3EFCE2FULL, "6041246_m.fxe"},
	{0x6424353F09612016ULL, "hlac_lt_blackemporium_1055_0z.dds"},
	{0x6424474926B93F09ULL, "6151057_m.fxe"},
	{0x64246769E98FA4DBULL, "ht_mansion_n_3367.anb"},
	{0x6424803431EDE974ULL, "hlac_ht_chantry_2534_7z.dds"},
	{0x6424F2207B65D665ULL, "vxlt_undercityencounter_11o_0_0.vlm"},
	{0x642556401491870AULL, "mcr310_argument6072320.cl"},
	{0x6425D8AA69F2206CULL, "vxht_mansion_n_4x6_0_0.vlm"},
	{0x64261CE045C777CEULL, "dae000ip_crowd_lm1_talk7.utp"},
	{0x6426F8282D0950FDULL, "177688655.ogg"},
	{0x6427099B0B59355BULL, "934645455.ogg"},
	{0x64272A1E74F317CDULL, "zz_qcr_debug6045148.cl"},
	{0x64277242ADC82EA2ULL, "126780165.ogg"},
	{0x642773F420E4FDF5ULL, "hlac_dr_slavers_3217_7y.dds"},
	{0x6427880647B1EA7FULL, "lgt260_eau_tmp_conrad6143798.cl"},
	{0x6427A241B187CA3BULL, "vxht_0_hightown_a2_c7p_0_0.vlm"},
	{0x64292FD35554D870ULL, "84429753.ogg"},
	{0x642972030DA56CFBULL, "dae300_cullen6135872.cl"},
	{0x642978E7E0DA826CULL, "vxlt_gallowsprison_n_6m4_0_0.vlm"},
	{0x64298F12DA2F3CB9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_basic.phy"},
	{0x6429DEA7D23E133FULL, "1007053988.ogg"},
	{0x642A277052FBFB93ULL, "dae000cr_grand_cleric6124676.cl"},
	{0x642AD8B5FC81AA43ULL, "vxht_keep_n_4hi_0_0.vlm"},
	{0x642AF94FA29D7B52ULL, "6093562_f.fxe"},
	{0x642B1690C9382786ULL, "dae000_pc_home_ambients6208639.cl"},
	{0x642B3DB2AE909BB4ULL, "6127080_m.fxe"},
	{0x642B49F02EAAB21EULL, "t3_wdg_db2.tnt"},
	{0x642B81406E990E08ULL, "vxlt_0_lowtown_a3_n_c_emk_0_0.vlm"},
	{0x642B88011E7C9ACBULL, "vxht_0_hightown_a2_bcu_0_0.vlm"},
	{0x642B8C7060AAF1CCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_mash_0d.dds"},
	{0x642BD42862A69A7FULL, "6027753_f.fxe"},
	{0x642BE3B09B85E4CAULL, "6110300_f.fxe"},
	{0x642BE99CDC9081FEULL, "6085041_m.fxe"},
	{0x642C12F16F302966ULL, "qun110_petrice6042258.cl"},
	{0x642C5956D696C6ECULL, "6110561_m.fxe"},
	{0x642CB0A5A27F7A4AULL, "6163468_m.fxe"},
	{0x642D1A9D62056CF7ULL, "vxht_0_hightown_n_c2p_0_0.vlm"},
	{0x642D37B500F844F2ULL, "8005621.ogg"},
	{0x642D5D340A1F5D36ULL, "gen_im_trs_chn_light_doc_03.uti"},
	{0x642E192E30D10156ULL, "gen00fl_aveline6114100.cl"},
	{0x642E1CD8DEC40525ULL, "vxlt_0_lowtown_docks_a3_cyv_0_0.vlm"},
	{0x642E66FE6A17588AULL, "6068104_varric_teasing.cut"},
	{0x642F8EABB89F7D15ULL, "hlac_ht_mansion2b_818_45z.dds"},
	{0x642FD63C8F838B45ULL, "art\\characters\\playercharacter\\children\\mt_bas_km.phy"},
	{0x64306085EDE44ED2ULL, "one100cr_ilen.ncs"},
	{0x64306F18333D4AD2ULL, "7339967.ogg"},
	{0x64307325DF112FE3ULL, "vxca_caverns2prc_8j7_0_1.vlm"},
	{0x6431305C81352B91ULL, "6033567_f.fxe"},
	{0x6431BBE897D35512ULL, "gen00fl_sebastian6179882.cl"},
	{0x6431CEE1F67EB246ULL, "227384325.ogg"},
	{0x6431DB474278F1A9ULL, "789490419.ogg"},
	{0x6431F6B51A137801ULL, "212658218.ogg"},
	{0x643229F51EAEEA7FULL, "vxw_sundermount_2th_0_0.vlm"},
	{0x64327DE5571CBF24ULL, "qun221_petrice6077943.cl"},
	{0x6432DC22BEFB3D5AULL, "zz_isa_debug6133168.cl"},
	{0x6433B271556D518DULL, "gen00fl_merrill6111275.cl"},
	{0x6433B272359DF97BULL, "6080447_m.fxe"},
	{0x6433C0B1EBFAE8B6ULL, "llac_ca_caverns1_24425_13.dds"},
	{0x6433CD1F15E4AE1EULL, "6172097_m.fxe"},
	{0x6433F07C1E983E8DULL, "vxht_mansion2b_2am_0_0.vlm"},
	{0x643401EBCDE1C4F6ULL, "lgt200ip_eau_tome04.utp"},
	{0x64342EAB03E84129ULL, "hlac_ht_chantry_2534_52y.dds"},
	{0x643432843DAA94B0ULL, "vxw_bonepitmining_a2_206_0_0.vlm"},
	{0x6434430658B7F6F6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edge.phy"},
	{0x64349182DF750D14ULL, "6126798_f.fxe"},
	{0x6434C0DE94A6FE36ULL, "vxw_sundermount_388_0_0.vlm"},
	{0x6434D057E81BA64DULL, "6093481_m.fxe"},
	{0x6435020E7E4A91DBULL, "hlac_w_sundermount_-196605_0x.dds"},
	{0x64354C5D9502DF81ULL, "220138157.ogg"},
	{0x64359C94A7E0AFFFULL, "pl.cut"},
	{0x6435A9A495A4E3E2ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation4.gad"},
	{0x6436141EA37DE2BBULL, "vxht_0_hightown_a2_c2s_0_0.vlm"},
	{0x6436DB92F49CB436ULL, "145930172.ogg"},
	{0x6436ED39ED07F832ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_0z.dds"},
	{0x6436F1A2BAA44223ULL, "335109004.ogg"},
	{0x6436FCF56428914AULL, "6149412_f.fxe"},
	{0x6437293A76FF868AULL, "3791374.ogg"},
	{0x64374CD2016273BFULL, "vxlt_gallowsprison_n_6g9_0_0.vlm"},
	{0x6438016CD13B6FA0ULL, "hlac_lt_0_lowtown_0_9x.dds"},
	{0x643857DD74BE7A4AULL, "mcr310_party_banter6182942.cl"},
	{0x6438D1AB2AC0FBF9ULL, "180872687.ogg"},
	{0x6438F764AC338DAFULL, "vxca_caverns1_fn6_0_0.vlm"},
	{0x64393FD25F4B83D0ULL, "86672335.ogg"},
	{0x6439559D3C3D8FD0ULL, "470206553.ogg"},
	{0x64395E4A3F3C50DFULL, "6129082_f.fxe"},
	{0x643969AFB59C2163ULL, "lt_0_lowtown_docks_22_0.rml"},
	{0x6439B0524B2914A5ULL, "mer210_coterie_contact6036763.cl"},
	{0x6439B4A13249F173ULL, "609990626.ogg"},
	{0x6439F9DE1072B34BULL, "vxht_0_hightown_a2_bdt_0_0.vlm"},
	{0x643ADC23C6DCE6D7ULL, "gen00fl_aveline6062112.cl"},
	{0x643AEED400D90BDEULL, "gen00fl_varric6033592.cl"},
	{0x643B1BC9BA17F965ULL, "hlac_ht_chantry_2534_20x.dds"},
	{0x643B227EE6A8C8E0ULL, "6080835_m.fxe"},
	{0x643B2E808689EB4AULL, "m_qun221ar_chantry.bnk"},
	{0x643BB1C4D7263BC0ULL, "ht_0_hightown_n_mmap_ph.dds"},
	{0x643BB9BEF1C2E9E3ULL, "art\\vfx\\combat\\vfx_m_modal_act_c.impa.evt"},
	{0x643BC07CE19C912CULL, "hlac_ht_mansion2b_818_22x.dds"},
	{0x643BD0AC6A80AA29ULL, "521378795.ogg"},
	{0x643C0E97412F7E72ULL, "llac_lt_0_lowtown_n_65536_1.dds"},
	{0x643D216373CE90C0ULL, "gen00fl_merrill6150914.cl"},
	{0x643D7FF8CAF46476ULL, "m_dae120ar_hightown_nt.wbd"},
	{0x643DF887E1EBD519ULL, "6139929_m.fxe"},
	{0x643E2C0658C02397ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edge.msh"},
	{0x643E3E0658C042FAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edge.mao"},
	{0x643E420658C04909ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edge.mmh"},
	{0x643E8527C812FE53ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.dura.evt"},
	{0x643EB2F6D978BC0FULL, "gen00fl_aveline6122780.cl"},
	{0x643ED9FA61591526ULL, "hlac_dr_primevalentrance_1726_13z.dds"},
	{0x643F7B115082B584ULL, "hlac_lt_0_lowtown_a2_n_-196608_0y.dds"},
	{0x643F99F3218B7ACDULL, "446948699.ogg"},
	{0x643FAAAC9052B4E3ULL, "6043730_m.fxe"},
	{0x643FBEBB8846AEC1ULL, "6082168_m.fxe"},
	{0x643FCD68547F3423ULL, "mag200_feynriel6030428.cl"},
	{0x64403113D6E940CFULL, "975368694.ogg"},
	{0x6440678B6AD0FF81ULL, "6075200_f.fxe"},
	{0x64409BD210A502FFULL, "vxw_sundermount_2yk_0_0.vlm"},
	{0x6440B1311F0F23D9ULL, "mag360_thrask6043660.cl"},
	{0x6440E8AA5EDCA54CULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01brgdebris_1.mao"},
	{0x644146DF5F3EBC37ULL, "6120143_m.fxe"},
	{0x6441918D3ABCC000ULL, "fx_white.dds"},
	{0x6441C5D0578FE036ULL, "w_bonepitmining_a2_18_0_ll.mmh"},
	{0x644266714AED3FF0ULL, "ser221_quentin6123363.cl"},
	{0x644292BEF1C8C664ULL, "art\\vfx\\combat\\vfx_m_modal_act_c.impa.gad"},
	{0x6442AC48C128EAB2ULL, "lt_0_lowtown_n_35_0_lc.mmh"},
	{0x6442DA88112A5616ULL, "llac_ca_cavern1_haunted_24424_15.dds"},
	{0x6442E4ED05545AB6ULL, "lgt110_arl_bann_teagan6115564.cl"},
	{0x6442F098F68EDD73ULL, "hlac_ht_chantry_n_2534_23z.dds"},
	{0x644332A2682F5DF0ULL, "vxht_mansion2_1vf_0_0.vlm"},
	{0x64436150808F905BULL, "14349318.ogg"},
	{0x6443760D0D4DC789ULL, "hlac_w_woundedcoast_-131066_0x.dds"},
	{0x644466EED76C4B9FULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml106a_0d.dds"},
	{0x64457E27C8191094ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.dura.gad"},
	{0x644580C6CCDABE3EULL, "hlac_ht_chantry_2534_31y.dds"},
	{0x644587990F6708DBULL, "6075469_m.fxe"},
	{0x64460DD314B3523CULL, "6108021_m.fxe"},
	{0x644663976375495AULL, "art\\characters\\creatures\\dragon\\animation\\k_bk.gad"},
	{0x644673A9175DC9CCULL, "370987130.ogg"},
	{0x644678E8BD9082B9ULL, "621703646.ogg"},
	{0x6446A7BAE20F9244ULL, "18105445.ogg"},
	{0x6446AEEA18287B10ULL, "166462614.ogg"},
	{0x644701E8D8ADD94AULL, "fxe_water_dis.mao"},
	{0x6447027E16DB62A9ULL, "llac_lt_0_lowtown_a3_n_c_-3_6.dds"},
	{0x644704A8A5438804ULL, "6093681_f.fxe"},
	{0x644761F821802947ULL, "877011300.ogg"},
	{0x64479EDCCAF0B3B8ULL, "132597324.ogg"},
	{0x6447FD5079F91B47ULL, "vxlt_0_lowtown_docks_bqz_0_0.vlm"},
	{0x644826C965D70575ULL, "6122135_f.fxe"},
	{0x6448C3C0DAE35023ULL, "804823493.ogg"},
	{0x6448E03BFF291B52ULL, "6143004_m.fxe"},
	{0x644905A89413EA9EULL, "llac_ht_mansion2b_818_26.dds"},
	{0x644930AFE2D02975ULL, "776085932.ogg"},
	{0x6449525AD70383C4ULL, "hlac_lt_0_lowtown_a2_n_131073_2y.dds"},
	{0x6449A1BEF1CEFE42ULL, "art\\vfx\\combat\\vfx_m_modal_act_c.impa.ani"},
	{0x6449D4A495B5D5C8ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation4.ani"},
	{0x6449DC93E3DA47E8ULL, "6035949_m.fxe"},
	{0x644A3615EAFBD0EFULL, "vxdr_0_deeproads_cj8_0_0.vlm"},
	{0x644A4C291A6C943DULL, "ser270_templar6164151.cl"},
	{0x644A5F3FFFD7CE2DULL, "6045802_m.fxe"},
	{0x644A7B839B48D408ULL, "primitives_primitive_p4.psh.fxo"},
	{0x644AB05DCBE749FFULL, "225516576.ogg"},
	{0x644B12A2171A840FULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar1_scaredmage.gad"},
	{0x644BCE73D9E4A35AULL, "vxw_bonepitdestroyed_a3_dj_0_0.vlm"},
	{0x644C0332AEB70C8FULL, "vxw_woundedcoast_15l_0_0.vlm"},
	{0x644C3B179F76E5FAULL, "hlac_dr_dungeon_3216_1y.dds"},
	{0x644C4816DE0BFF29ULL, "69543709.ogg"},
	{0x644C8D27C81F4972ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.dura.ani"},
	{0x644C9F3E10FE4727ULL, "vxlt_hangedman_18u_0_0.vlm"},
	{0x644D4848A85ED040ULL, "follower_banter6155967.cl"},
	{0x644D594760B1659BULL, "hlac_lt_0_lowtown_n_65536_4x.dds"},
	{0x644D6697637B6D71ULL, "art\\characters\\creatures\\dragon\\animation\\k_bk.evt"},
	{0x644DCB306DC314C3ULL, "vxw_bonepitdestroyed_a3_1fb_0_0.vlm"},
	{0x644F16B67671D782ULL, "903938776.ogg"},
	{0x645035EB5B3EC6C5ULL, "vxca_cavern1_haunted_ebv_0_0.vlm"},
	{0x6450550CEAF15B59ULL, "vxht_0_hightown_b8v_0_0.vlm"},
	{0x645059EA736C582BULL, "mrl000_marethari6069605.cl"},
	{0x64509A27DE9802DDULL, "fen210_fenris6109427.cl"},
	{0x6450A0F7CB6AE10EULL, "6070179_m.fxe"},
	{0x6450B88E0CD04508ULL, "w_encounter3_151_0.msh"},
	{0x6450CA8E0CD06399ULL, "w_encounter3_151_0.mao"},
	{0x6450CF79C2EC12D0ULL, "vxlt_0_lowtown_docks_a3_n_d4h_0_0.vlm"},
	{0x6450D68E0CD07832ULL, "w_encounter3_151_0.mmh"},
	{0x6450E45F3C66BF19ULL, "6068869_m.fxe"},
	{0x64516707FB1C1FF7ULL, "vxw_woundedcoast_n4_0_0.vlm"},
	{0x645243582C0C4818ULL, "dae000ip_home_ambients.ncs"},
	{0x64524B3BD1699985ULL, "6118388_m.fxe"},
	{0x64534395E22CEE36ULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_0z.dds"},
	{0x6454B274F87A8BDBULL, "6148094_f.fxe"},
	{0x64558E16F512CAE8ULL, "vxlt_0_lowtown_a3_n_c_gob_0_1.vlm"},
	{0x6455A9A1410A0BC2ULL, "976524276.ogg"},
	{0x64561F675C05375FULL, "dre000_carver6037583.cl"},
	{0x64581C95596E6A7AULL, "w_bonepit_184_0.phy"},
	{0x64584B4A6B7D1ACEULL, "631955751.ogg"},
	{0x645886D97C35009FULL, "6144251_m.fxe"},
	{0x64588C03338B3835ULL, "zz_one_debug6058966.cl"},
	{0x6458CF27829C89CBULL, "mag101_aombanter6052748.cl"},
	{0x6459135C3C509B8AULL, "vxca_cavern1_haunted_j5s_0_0.vlm"},
	{0x645926482F525922ULL, "vxlt_gallowstemplar_3e0_0_0.vlm"},
	{0x64596EB0A16B0D17ULL, "1061554469.ogg"},
	{0x645978C53CD80477ULL, "vxht_0_hightown_a2_c7d_0_0.vlm"},
	{0x6459E88A24C71335ULL, "vxht_keep_n_62o_0_0.vlm"},
	{0x6459F05778EEAF1BULL, "hlac_lt_0_lowtown_a3_n_c_-4_0y.dds"},
	{0x6459F6A1B99E6B04ULL, "follower_banter6169876.cl"},
	{0x645A4050CBDABBDAULL, "vxw_bonepitdestroyed_a3_ok_0_0.vlm"},
	{0x645A5078E94DD45CULL, "pro000_colonel_ewald6040615.cl"},
	{0x645A750FA0AD250FULL, "vxw_woundedcoast_10o_0_0.vlm"},
	{0x645A7E9763862040ULL, "art\\characters\\creatures\\dragon\\animation\\k_bk.ani"},
	{0x645AAAA6FE6C7D2CULL, "vxht_mansion2_3gn_0_1.vlm"},
	{0x645ADDD073290AB9ULL, "6110894_m.fxe"},
	{0x645B767E9A2FBFC6ULL, "vxca_caverns2_8lt_0_0.vlm"},
	{0x645BA472AB734918ULL, "vxht_0_hightown_cbp_0_0.vlm"},
	{0x645BD78BA7DF2F3EULL, "67136047.ogg"},
	{0x645BF7B8E1EB0949ULL, "mag210_arianni6021216.cl"},
	{0x645C158F74BD5CAFULL, "898080233.ogg"},
	{0x645C3EDED1BC86EDULL, "366005056.ogg"},
	{0x645CA3A4F06327C0ULL, "vxlt_undercityencounter_1d1_0_0.vlm"},
	{0x645CA7842850EFECULL, "mrl150_merrill6085840.cl"},
	{0x645CED5064E1C93FULL, "static_static_p13.psh.fxo.dx11.assembly"},
	{0x645D80774068E661ULL, "6043255_m.fxe"},
	{0x645DB2A1D83EA94FULL, "hlac_ht_0_hightown_n_-131074_0z.dds"},
	{0x645DF6323877B1EFULL, "769977446.ogg"},
	{0x645E05F4FA990B6AULL, "6060149_m.fxe"},
	{0x645E464CDA944C7FULL, "llac_ht_0_hightown_a3_-65537_4.dds"},
	{0x645E5C450D41D0FEULL, "hlac_lt_0_lowtown_docks_-196608_1x.dds"},
	{0x645E69930CCD7645ULL, "hlac_lt_0_lowtown_n_-131073_2y.dds"},
	{0x645EA45BA2EABECBULL, "229047843.ogg"},
	{0x645EA7C84D34DF45ULL, "vxca_cavern1_haunted_egu_0_0.vlm"},
	{0x645ED5B328ACD0DBULL, "dae021_fem_worshipper_sta6131074.cl"},
	{0x645F0DD53B694966ULL, "361381930.ogg"},
	{0x645F1F6B3A714EB8ULL, "gen00fl_isabela6135681.cl"},
	{0x645F2FA2172B5ED5ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar1_scaredmage.ani"},
	{0x645F975B1A08BAF7ULL, "hlac_lt_0_lowtown_a2_n_-65537_3y.dds"},
	{0x645FC59DFAB686DDULL, "hlac_lt_0_lowtown_a2_n_-196609_3z.dds"},
	{0x645FF6F4A96BE11AULL, "vxht_brothel_n_6fc_0_0.vlm"},
	{0x646066E9BC777BA7ULL, "gen00fl_anders6119532.cl"},
	{0x64616A4E7B7306B4ULL, "vxlt_0_lowtown_docks_a3_n_apm_0_0.vlm"},
	{0x6461864D2BC75589ULL, "hf_qmq_patronf3.mor"},
	{0x646210173B45CDFFULL, "6136687_m.fxe"},
	{0x64624F9A5A673A48ULL, "591359343.ogg"},
	{0x64628A1B69AED29AULL, "cd_gen00fl_fenris.bnk"},
	{0x6462998E37AC1207ULL, "6027743_m.fxe"},
	{0x6462CDF08296BDBBULL, "618873747.ogg"},
	{0x6463184A51F382A0ULL, "148639469.ogg"},
	{0x6463A1A5F75604C9ULL, "mcr370_assignments6081698.cl"},
	{0x6463F820E3AB3B39ULL, "vxca_caverns1_dpp_0_0.vlm"},
	{0x6464172CC96D3E6EULL, "6152752_m.fxe"},
	{0x6464F00D2732A6E4ULL, "vxdr_primevalentrance_22t_0_0.vlm"},
	{0x6465CA49C875218DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masf_0t.dds"},
	{0x64670F2B10A9719DULL, "1256838.ogg"},
	{0x646715307DFDF528ULL, "ring_of_whispers.dds"},
	{0x64671FAF0603A00AULL, "lgt260cr_way_bloodmage_1.utc"},
	{0x64678660D3C43F2AULL, "348638606.ogg"},
	{0x6468083B1348D997ULL, "6129261_m.fxe"},
	{0x64695A582C1F4E11ULL, "dae000ip_home_ambients.utp"},
	{0x646978C0FE584996ULL, "vxlt_0_lowtown_n_h8h_0_0.vlm"},
	{0x64698CE32130CB99ULL, "6149184_m.fxe"},
	{0x6469A5F8CB19DFE4ULL, "m_dae120ar_hightown_nt.bnk"},
	{0x6469C226E540ADD5ULL, "176256802.ogg"},
	{0x6469DF05DFCA0BC6ULL, "vxht_0_hightown_bdy_0_0.vlm"},
	{0x6469E85B8C25D846ULL, "vxca_caverns2_8ku_0_0.vlm"},
	{0x6469EA684A3F642CULL, "vxht_0_hightown_n_b5c_0_0.vlm"},
	{0x646A30F7700FFC32ULL, "pro000_colonel_ewald6005416.cl"},
	{0x646B039063960DCEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_landscapes_n.dds"},
	{0x646C1BBD785E147BULL, "6071519_m.fxe"},
	{0x646D103265AA1537ULL, "hlac_lt_0_lowtown_n_65537_7z.dds"},
	{0x646D4379A51299E3ULL, "198483732.ogg"},
	{0x646D751D709DF64CULL, "vxlt_0_lowtown_a2_n_c_glc_0_0.vlm"},
	{0x646E24205113EBCBULL, "hlac_ht_0_hightown_n_-65536_3z.dds"},
	{0x646E569B207AEBA4ULL, "kj.cut"},
	{0x646E88393E11A770ULL, "vxht_mansion2_1ge_0_0.vlm"},
	{0x646EAD0EF062B4C7ULL, "vxca_caverns1_hum_0_0.vlm"},
	{0x646ED7A09DFD4549ULL, "vxlt_gallowscourtyard_n_6rj_0_0.vlm"},
	{0x646EF49FBDE25F04ULL, "hlac_ht_0_hightown_-65537_0x.dds"},
	{0x646F72AADA67EA04ULL, "llac_w_bonepitmining_a2_458757_0.dds"},
	{0x646F7325F315CA3CULL, "zz_fol_debug6096575.cl"},
	{0x646F7BE23FE6B363ULL, "1034814734.ogg"},
	{0x64704B4DD00E0344ULL, "dre000_sibling_has_blight6037044.cl"},
	{0x6470A144854B6BD3ULL, "art\\vfx\\textures\\vfx_fire_ground_fb.dds"},
	{0x6470A3A71842B316ULL, "hlac_lt_0_lowtown_docks_-65537_5z.dds"},
	{0x6470C395566A7DCBULL, "vxlt_0_lowtown_a2_hdv_0_0.vlm"},
	{0x64711910CE279473ULL, "6102139_f.fxe"},
	{0x6471B7818593410EULL, "6169509_m.fxe"},
	{0x6472589A37EA3445ULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.anb"},
	{0x647264E0DA48B02AULL, "1012281069.ogg"},
	{0x64726816EEFA099BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medd_0n.dds"},
	{0x64727653F2BCD505ULL, "gen00flst_isa_gift.stg"},
	{0x6472B93B17C0E623ULL, "vxlt_0_lowtown_a2_n_eox_0_1.vlm"},
	{0x6472EC87DC8BD521ULL, "vxlt_undercityencounter_1j6_0_0.vlm"},
	{0x6472F2A80CEE2206ULL, "vxlt_gallowstemplar_n_357_0_0.vlm"},
	{0x647392DC73CBD1AFULL, "6120064_m.fxe"},
	{0x6473FA40234A9148ULL, "6104898_m.fxe"},
	{0x647420A8EB4D7C50ULL, "llac_w_sundermount_65538_0.dds"},
	{0x647454D4192607E2ULL, "607403033.ogg"},
	{0x6474A23C069C6C7CULL, "hlac_lt_0_lowtown_a2_131070_4y.dds"},
	{0x64750FFFBB85815BULL, "6163159_m.fxe"},
	{0x647514C59EDD864EULL, "6161534_m.fxe"},
	{0x6475956AEE1B86D4ULL, "218575011.ogg"},
	{0x647612EA5A4E7A30ULL, "llac_lt_gallowstemplar_2394_4.dds"},
	{0x6476D3870F014FB5ULL, "6139914_m.fxe"},
	{0x64771F75E2D7C46FULL, "92946896.ogg"},
	{0x6477E498F65262F3ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_act_c.impa.ani"},
	{0x6478342C6AC61A4DULL, "6172674_m.fxe"},
	{0x64787182920C0844ULL, "hightown_fem_cityguard_mob6168527.cl"},
	{0x647881B343B6F35FULL, "follower_banter6167109.cl"},
	{0x64789472D5CBB841ULL, "508354479.ogg"},
	{0x647944F4D442F11CULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemb_hed_0i.dds"},
	{0x6479EF2A25527C2BULL, "vxlt_0_lowtown_docks_n_b99_0_0.vlm"},
	{0x647A3ABE368148CBULL, "vxlt_backalley_n_1ww_0_1.vlm"},
	{0x647A3E471D60E03EULL, "mcr370_assignments6081725.cl"},
	{0x647A52C9DCF74997ULL, "vxht_0_hightown_n_aka_0_0.vlm"},
	{0x647A7662994F095BULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_multistrike.evt"},
	{0x647ABC3DACF4B5F5ULL, "vxlt_undercitydungeon_zv_0_0.vlm"},
	{0x647AD2FFBE06434EULL, "zz_dae_debug6029301.cl"},
	{0x647AD4C6D90E3157ULL, "mer120_hubert6070240.cl"},
	{0x647AFEA7732B562BULL, "vxht_mansion_n_53h_0_1.vlm"},
	{0x647B2919D231E01DULL, "588512981.ogg"},
	{0x647B7A6B68DC734EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_vine.mmh"},
	{0x647B946B68DC9F3CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_vine.msh"},
	{0x647C06C1180AF366ULL, "hlac_lt_gallowscourtyard_-65537_2y.dds"},
	{0x647C21347470D628ULL, "102940493.ogg"},
	{0x647C46CFFD93D0D1ULL, "6215100_m.fxe"},
	{0x647C55C2AFBD6DA6ULL, "795730742.ogg"},
	{0x647C74594F192BD8ULL, "vxlt_0_lowtown_a3_n_c_ht1_0_0.vlm"},
	{0x647CBC7AFAAD0EC7ULL, "203503156.ogg"},
	{0x647CD6F5830053CFULL, "vxht_0_hightown_n_cat_0_1.vlm"},
	{0x647CEEA060533CA0ULL, "hlac_ht_0_hightown_n_1_2y.dds"},
	{0x647CFC13CA470EABULL, "gen00fl_anders6128396.cl"},
	{0x647D1648366A74A2ULL, "vxlt_gallowstemplar_n_3sa_0_0.vlm"},
	{0x647E0DB2F4CE6EA4ULL, "hlac_lt_0_lowtown_a2_n_65536_9z.dds"},
	{0x647E8050F0CA0268ULL, "6072368_m.fxe"},
	{0x647F624CDC980F6BULL, "216299035.ogg"},
	{0x647FB21BDBAC6BB5ULL, "vxdr_primevaltomb_6dv_0_1.vlm"},
	{0x64808EF9D617D527ULL, "ser121_jethann6108809.cl"},
	{0x6480C058C90EE954ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls104a_0s.dds"},
	{0x6480C24C284DE681ULL, "gen00fl_aveline6106942.cl"},
	{0x64812893300423B1ULL, "dae000_merchant_weapons_dm6164403.cl"},
	{0x64815E64CE81EDA1ULL, "vxlt_undercityencounter_1g5_0_0.vlm"},
	{0x64816F6299551B9CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_multistrike.gad"},
	{0x6481A2BA7DDD4FEFULL, "6173947_m.fxe"},
	{0x6481C4C441D17D47ULL, "vxlt_0_lowtown_docks_bev_0_0.vlm"},
	{0x648208E889BCAB68ULL, "follower_banter6151314.cl"},
	{0x648217D3CBEA17CAULL, "vxw_encounter2_5e_0_0.vlm"},
	{0x6482B0BAD815082AULL, "273738237.ogg"},
	{0x6482ED48EDBDA3F0ULL, "6135375_m.fxe"},
	{0x648350FC67956B16ULL, "6077691_m.fxe"},
	{0x64836761373B1B6EULL, "6146850_m.fxe"},
	{0x6483758307295509ULL, "6125740_m.fxe"},
	{0x6483B7C4CA9814C8ULL, "6064636_f.fxe"},
	{0x648415B600146361ULL, "6148837_m.fxe"},
	{0x6484163C36F2E3F0ULL, "nat320_delilah6065935.cl"},
	{0x64846E6DD776DBE6ULL, "dae_rogue_aoe_instant.ncs"},
	{0x6484A4510C318DA4ULL, "273227531.ogg"},
	{0x6484B41B2CCCD807ULL, "932836888.ogg"},
	{0x6484C93930B49909ULL, "hurlock_alpha_axe_shield.utc"},
	{0x6484D62D018B9A70ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_meda.mmh"},
	{0x6484DC98F65CDF2AULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_act_c.impa.evt"},
	{0x6484DD054BC49653ULL, "tbe_dog_act3.ncs"},
	{0x6484DECEEFD4D809ULL, "6178434_m.fxe"},
	{0x6484E42D018BB18AULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_meda.msh"},
	{0x648547E697705597ULL, "dae000_bartrand6031141.cl"},
	{0x6485D9AC78FF002CULL, "w_bonepit_30_0.rml"},
	{0x64867B61A7C1452BULL, "gen00fl_aveline6139195.cl"},
	{0x6486F0C900A6A8AAULL, "6085269_m.fxe"},
	{0x64874AAACF022B4BULL, "34621249.ogg"},
	{0x6487575B0017478BULL, "6124698_f.fxe"},
	{0x6487A41B69CE1B38ULL, "cd_gen00fl_fenris.wbd"},
	{0x6487FE1F44DEE7F6ULL, "vxlt_0_lowtown_a2_n_c_emp_0_0.vlm"},
	{0x648809B77FDC68CEULL, "qun220cr_javaris6071712.cl"},
	{0x648820FB69841286ULL, "t1_mul_re1.tnt"},
	{0x648856D2660E4A97ULL, "vxdr_slavers_13w_0_0.vlm"},
	{0x64886E62995B389AULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_multistrike.ani"},
	{0x64889DBB36D7B168ULL, "360310764.ogg"},
	{0x64898FDFCEE4F7F9ULL, "vxw_woundedcoast_130_0_0.vlm"},
	{0x6489B2FC28357D43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bc_transition.evt"},
	{0x6489BFDF7C7AD9A4ULL, "156222626.ogg"},
	{0x6489FBB409D2AD54ULL, "6102455_f.fxe"},
	{0x648A0F2B192081E5ULL, "vxlt_undercityencounter_1lh_0_0.vlm"},
	{0x648A22D8E54604E4ULL, "6040978_m.fxe"},
	{0x648AC7D400320378ULL, "794955764.ogg"},
	{0x648AE5EE953A0963ULL, "23357210.ogg"},
	{0x648AE870D635B36EULL, "hlac_lt_backalley_g_0_1y.dds"},
	{0x648B2E7BC2F481F7ULL, "vxw_woundedcoast_z0_0_0.vlm"},
	{0x648BBB98F662C59DULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_act_c.impa.gad"},
	{0x648BC9CF1F533407ULL, "llac_lt_undercityserialkiller_-1_30.dds"},
	{0x648BE74E12A4E358ULL, "6018561_f.fxe"},
	{0x648C4825C1207082ULL, "hlac_w_bonepit_327685_0y.dds"},
	{0x648C889281A36746ULL, "6027984_m.fxe"},
	{0x648CEFB87E085466ULL, "6087182_m.fxe"},
	{0x648D51EB50B292A6ULL, "vxlt_0_lowtown_a3_n_c_fe6_0_0.vlm"},
	{0x648D56516341335CULL, "227872451.ogg"},
	{0x648D82A70F42A72FULL, "342621104.ogg"},
	{0x648E08E5FBF53AF9ULL, "follower_banter6163741.cl"},
	{0x648E0951D2D778F5ULL, "410381109.ogg"},
	{0x648E19FA873A8B36ULL, "gen_im_wep_war_shd_default.uti"},
	{0x648E6B0CAF09EBBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_enter.ani"},
	{0x648E91F7E7E424D7ULL, "gen_im_trs_dem_junk2_02.uti"},
	{0x648F4B6412D70215ULL, "follower_banter6164775.cl"},
	{0x648F6A993323C2B8ULL, "442063157.ogg"},
	{0x648F8733908C871AULL, "vxw_sundermount_22s_0_0.vlm"},
	{0x648FE81CACCFB229ULL, "vxdr_0_deeproads_dqe_0_1.vlm"},
	{0x649014976FD3686FULL, "6176647_m.fxe"},
	{0x649029ED3D85209FULL, "216346069.ogg"},
	{0x64908BFC283B59C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bc_transition.gad"},
	{0x6490D7A14D48BD77ULL, "6191778_m.fxe"},
	{0x6490D7A99FD26CA2ULL, "861659647.ogg"},
	{0x64916A829C9879CBULL, "276358309.ogg"},
	{0x64919199025B8F44ULL, "llac_lt_0_lowtown_a2_n_c_-262144_2.dds"},
	{0x6491B28D0F4D8902ULL, "vxlt_foundry_n_xn_0_0.vlm"},
	{0x6491D03D6A1F5BDCULL, "vxw_bonepitmining_a2_1pa_0_0.vlm"},
	{0x649204BD228AB423ULL, "lt_0_lowtown_docks_a3_127_0.msh"},
	{0x649212BD228ACC7DULL, "lt_0_lowtown_docks_a3_127_0.mmh"},
	{0x6492140AF7045693ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drunk_drink.ani"},
	{0x649216BD228AD28EULL, "lt_0_lowtown_docks_a3_127_0.mao"},
	{0x6492DC1863840229ULL, "375717447.ogg"},
	{0x6492E80320834AEFULL, "hlac_w_sundermount_-262144_0x.dds"},
	{0x649308B7EC9EB097ULL, "6063079_m.fxe"},
	{0x6493234E60386464ULL, "vxht_0_hightown_n_a6y_0_0.vlm"},
	{0x6493314B51EACCB1ULL, "hlac_lt_0_lowtown_docks_a3_n_-131072_0x.dds"},
	{0x64936C758C587468ULL, "hlac_lt_0_lowtown_docks_-2_1x.dds"},
	{0x64938728E4DB12DFULL, "art\\levels\\4cinematics\\ferelden\\prp_bookinquisb.msh"},
	{0x64938D28E4DB1C81ULL, "art\\levels\\4cinematics\\ferelden\\prp_bookinquisb.mmh"},
	{0x64939928E4DB3122ULL, "art\\levels\\4cinematics\\ferelden\\prp_bookinquisb.mao"},
	{0x6493A435F4D234E4ULL, "865916551.ogg"},
	{0x6493D9580EE2E454ULL, "6084862_m.fxe"},
	{0x6493E317E296A86FULL, "337799812.ogg"},
	{0x64944AE0A1074390ULL, "hlac_lt_0_lowtown_n_-3_1y.dds"},
	{0x6494D473A672F899ULL, "vxlt_0_lowtown_a3_n_c_hjv_0_0.vlm"},
	{0x64954CFC8612BA6AULL, "w_blightlands_196_0.mao"},
	{0x649550FC8612C0B9ULL, "w_blightlands_196_0.mmh"},
	{0x64955AFC8612D1A7ULL, "w_blightlands_196_0.msh"},
	{0x6495AE5482F352E5ULL, "dae000_tomwise6167284.cl"},
	{0x64969AFC283FDEA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bc_transition.ani"},
	{0x6496A675F5E0D53EULL, "270284951.ogg"},
	{0x6497047CDCEFADBEULL, "hlac_lt_0_lowtown_docks_65533_3y.dds"},
	{0x6497110F0CCC9697ULL, "zz_fen_debug6134348.cl"},
	{0x6497778085B3CFC1ULL, "vxca_caverns2_80k_0_0.vlm"},
	{0x64979FA2F4718C43ULL, "vxlt_0_lowtown_docks_b7t_0_0.vlm"},
	{0x6498074276E168DDULL, "23251899.ogg"},
	{0x64986848B46EE771ULL, "6076682_m.fxe"},
	{0x64987C477BC533FEULL, "one140cr_charade6134605.cl"},
	{0x64989DC7D62D0C14ULL, "vxca_caverns1_mines_i2g_0_0.vlm"},
	{0x649915551C5A2B3BULL, "6129188_m.fxe"},
	{0x64993D1D18687F10ULL, "99762319.ogg"},
	{0x649942BA23811CF6ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.impa.ani"},
	{0x64994B15FE633363ULL, "1049259263.ogg"},
	{0x64998E421199F540ULL, "6096216_m.fxe"},
	{0x64998EE88D7604D6ULL, "vxht_keep_n_5zw_0_0.vlm"},
	{0x6499A058B4EA9A77ULL, "vxw_woundedcoast_w3_0_0.vlm"},
	{0x649A7F9512D7AFFAULL, "vxlt_undercitydungeon_133_0_0.vlm"},
	{0x649A897D8F0EE507ULL, "llac_lt_0_lowtown_docks_a3_-65538_0.dds"},
	{0x649ABA70D12E2EFEULL, "vxca_caverns1_mines_cnk_0_0.vlm"},
	{0x649AFA47F1CA0B9EULL, "128530777.ogg"},
	{0x649B38ED8FAE8171ULL, "vxw_sundermount_318_0_0.vlm"},
	{0x649B599A380D3A86ULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.mmh"},
	{0x649B5FCC33B99EABULL, "317727136.ogg"},
	{0x649B730CAF1483B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_enter.evt"},
	{0x649C81D3C3EABC07ULL, "127103500.ogg"},
	{0x649CC357A4247405ULL, "vxlt_gallowscourtyard_7ap_0_1.vlm"},
	{0x649D0A57032383ACULL, "255690331.ogg"},
	{0x649D4ADC5E89A9C4ULL, "lgt110_arl_bann_teagan6115632.cl"},
	{0x649DCAC846DB493BULL, "55945582.ogg"},
	{0x649DD4D876D57C05ULL, "6196690_m.fxe"},
	{0x649DF91082829F9AULL, "vxw_sundermount_27l_0_0.vlm"},
	{0x649E21D5552DA511ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_up.ani"},
	{0x649E29F308D85136ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh_additive.evt"},
	{0x649E39F5BC75E3B0ULL, "6098194_f.fxe"},
	{0x649EB867DB06D891ULL, "552376492.ogg"},
	{0x649F4EE9DA1A7D4AULL, "vxw_blightlands_gl_0_0.vlm"},
	{0x649F7A16A9DC9CFBULL, "face_face_p8.psh.fxo.dx11"},
	{0x649F86DCE33650D2ULL, "222390307.ogg"},
	{0x649F8DEEEFDBE2D6ULL, "hlac_lt_0_lowtown_n_1_10z.dds"},
	{0x64A00C0AF71086CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drunk_drink.evt"},
	{0x64A00F35ECDF8658ULL, "llac_ht_0_hightown_-65538_1.dds"},
	{0x64A039DCFC5847BCULL, "vxw_encounter2_ff_0_0.vlm"},
	{0x64A03B94FB5B6867ULL, "llac_lt_0_lowtown_n_-3_7.dds"},
	{0x64A0B8B9B73A0970ULL, "593311486.ogg"},
	{0x64A0BDD628C19EE7ULL, "vxht_mansion2b_2oa_0_0.vlm"},
	{0x64A12EF098D6C41DULL, "6174038_m.fxe"},
	{0x64A1A5DEE4726917ULL, "6174414_m.fxe"},
	{0x64A1AC8C6043118AULL, "gen_im_trs_slv_hvy_boo_r2.uti"},
	{0x64A1E055E5EA8770ULL, "gen00fl_aveline6140150.cl"},
	{0x64A1F96103707EDBULL, "6153862_m.fxe"},
	{0x64A2544B96C07443ULL, "vxdr_primevaltomb_8ve_0_0.vlm"},
	{0x64A2920CAF1AD745ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_enter.gad"},
	{0x64A2CEB7BC893DE2ULL, "977720213.ogg"},
	{0x64A2D84C6C14A34EULL, "vxca_caverns1_fmr_0_0.vlm"},
	{0x64A2EC652C5E799EULL, "195169173.ogg"},
	{0x64A305BA6F4877C9ULL, "155886880.ogg"},
	{0x64A30CF7C6538F57ULL, "6051401_m.fxe"},
	{0x64A338D88416C84BULL, "vxht_keep_n_58h_0_0.vlm"},
	{0x64A3B3256F640694ULL, "vxht_mansion2b_3bt_0_0.vlm"},
	{0x64A3BF283FB275EEULL, "6164083_m.fxe"},
	{0x64A404277D0BF016ULL, "6067122_m.fxe"},
	{0x64A4A2D4B5249060ULL, "6182306_m.fxe"},
	{0x64A4AB3071C509FCULL, "905075376.ogg"},
	{0x64A515B5933D1419ULL, "art\\characters\\weapons\\axes\\w_awe_dsaa.mmh"},
	{0x64A51FB5933D2507ULL, "art\\characters\\weapons\\axes\\w_awe_dsaa.msh"},
	{0x64A520F308DE60C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh_additive.gad"},
	{0x64A523417D73A9C5ULL, "441236924.ogg"},
	{0x64A52CD55533D6B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_up.gad"},
	{0x64A59733D91E0F91ULL, "468781698.ogg"},
	{0x64A5AA296A745961ULL, "gen00fl_bethany6101212.cl"},
	{0x64A5C8FD9F46A921ULL, "gen00fl_sebastian6180498.cl"},
	{0x64A615D14581FAD0ULL, "mrl310_merrill_pt1.stg"},
	{0x64A624C7AE3E4D86ULL, "vxca_caverns1_mines_ckl_0_0.vlm"},
	{0x64A64ABA238BB4E7ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.impa.evt"},
	{0x64A67399B2D128F9ULL, "vxw_woundedcoast_196_0_0.vlm"},
	{0x64A6EB0AF7166D3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drunk_drink.gad"},
	{0x64A70F4B430EC43DULL, "vxw_blightlands_e5_0_0.vlm"},
	{0x64A71F72EA2FE92AULL, "one100dg_nabil6173820.cl"},
	{0x64A7719AF8A5DC7FULL, "6078703_m.fxe"},
	{0x64A79ABA78B74461ULL, "dre000_mother6037116.cl"},
	{0x64A7B87C8B83C191ULL, "vxht_0_hightown_cbd_0_1.vlm"},
	{0x64A7FD67D58494B9ULL, "dae200_cullen6140295.cl"},
	{0x64A81944E7206017ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicb.phy"},
	{0x64A83DFCAA8A0968ULL, "follower_banter6143230.cl"},
	{0x64A917CC1C47A553ULL, "mag100_cullen6095148.cl"},
	{0x64A9295324DEFE8DULL, "w_encounter1_62_0.msh"},
	{0x64A9375324DF16C0ULL, "w_encounter1_62_0.mao"},
	{0x64A93B5324DF1D13ULL, "w_encounter1_62_0.mmh"},
	{0x64A9AFC621BFAA59ULL, "6108064_m.fxe"},
	{0x64AA4D76089F7FD4ULL, "6128683_m.fxe"},
	{0x64AAA9E0417D5C22ULL, "dr_slavers_probe_23x.dds"},
	{0x64AAE3ABC5511A76ULL, "6088323_m.fxe"},
	{0x64ABA275896486E2ULL, "llac_dr_0_deeproads_5841_4.dds"},
	{0x64AC21F308E4816FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh_additive.ani"},
	{0x64AC29D55539F018ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_front_up.evt"},
	{0x64AC647C5E3959AFULL, "qun211_braggart6075466.cl"},
	{0x64AC6AED7478B81AULL, "vxw_sundermount_28m_0_0.vlm"},
	{0x64ACB3774CACDD39ULL, "lgt160_wk_avern_body6180469.cl"},
	{0x64ACF33BA448A713ULL, "6053353_m.fxe"},
	{0x64AD0174A9CC5178ULL, "6156393_f.fxe"},
	{0x64AD09013B27861FULL, "vxlt_0_lowtown_a3_n_c_fig_0_0.vlm"},
	{0x64AD6BBA23920B30ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.impa.gad"},
	{0x64ADA8ED4A0900B1ULL, "6115080_m.fxe"},
	{0x64ADB4CB3E80BA64ULL, "hlac_ca_caverns1_mines_24427_1x.dds"},
	{0x64ADC0C6CC1095CAULL, "vxw_blightlands_hm_0_0.vlm"},
	{0x64AE13979DB13D5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_stroud.evt"},
	{0x64AE65714D4C0CC3ULL, "qun221_petrice6077447.cl"},
	{0x64AF4D88FD5403ECULL, "docks_male_sailor_mob6176660.cl"},
	{0x64AF858835B1321CULL, "6099193_f.fxe"},
	{0x64B02C775BEC34EEULL, "482537665.ogg"},
	{0x64B1FAB1473BB2BEULL, "450865466.ogg"},
	{0x64B24C0F14416281ULL, "vxlt_0_lowtown_docks_a3_c1b_0_0.vlm"},
	{0x64B26157A4DA2DA3ULL, "vxlt_0_lowtown_n_f9h_0_0.vlm"},
	{0x64B3159559BB09A3ULL, "w_bonepit_184_0.msh"},
	{0x64B3239559BB21FDULL, "w_bonepit_184_0.mmh"},
	{0x64B3279559BB280EULL, "w_bonepit_184_0.mao"},
	{0x64B36E514AFA6AA0ULL, "6167025_f.fxe"},
	{0x64B4273E2DFF008FULL, "vxlt_0_lowtown_a2_n_gob_0_1.vlm"},
	{0x64B43A4D11995D01ULL, "vxw_bonepitmining_a2_1p0_0_0.vlm"},
	{0x64B455C34157DAC1ULL, "ser270_emeric6149361.cl"},
	{0x64B49A8A321A44BEULL, "rdr111dg_martin6054511.cl"},
	{0x64B4B4BC5895379AULL, "hlac_ht_0_hightown_n_-196608_1x.dds"},
	{0x64B4C8CA5B7A32D3ULL, "hlac_lt_gallowstemplar_f_2391_4y.dds"},
	{0x64B4D861114ED973ULL, "art\\levels\\atmosphere\\sky\\sb03_cloud.dds"},
	{0x64B51267F0318C47ULL, "llac_w_bonepit_327685_0.dds"},
	{0x64B5129EF79BDC13ULL, "hlac_lt_0_lowtown_docks_-65538_3x.dds"},
	{0x64B52A979DB78289ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_stroud.gad"},
	{0x64B56A5E9795CEC1ULL, "art\\characters\\creatures\\mubari\\animation\\com_lunge.ani"},
	{0x64B581283504DCBDULL, "vxw_blightlands_n4_0_0.vlm"},
	{0x64B695B3A3E7BCF0ULL, "weapon2_s1.psh.fxo"},
	{0x64B6A0D489DB3A00ULL, "one140cr_charade6134319.cl"},
	{0x64B6AD1DAD1E1423ULL, "lt_0_lowtown_docks_172_0.mao"},
	{0x64B6B11DAD1E1B70ULL, "lt_0_lowtown_docks_172_0.mmh"},
	{0x64B6BF1DAD1E328AULL, "lt_0_lowtown_docks_172_0.msh"},
	{0x64B6F444899E371EULL, "1020647763.ogg"},
	{0x64B71B5D4915E939ULL, "water_water_p2.psh.fxo.dx11.assembly"},
	{0x64B71DAFC388322EULL, "6101006_m.fxe"},
	{0x64B7B599FF4F97CBULL, "sto_cl_flash.mao"},
	{0x64B7BED0C89647AFULL, "vxlt_foundry_n_z5_0_0.vlm"},
	{0x64B7EDC8E1668855ULL, "6156719_m.fxe"},
	{0x64B814E32524BF3BULL, "vxht_0_hightown_a2_caj_0_0.vlm"},
	{0x64B81C564BBC7FA4ULL, "hlac_ca_cave_2194_10z.dds"},
	{0x64B8866A73E50DFDULL, "hlac_lt_0_lowtown_a2_65536_5z.dds"},
	{0x64B8DDFC7AEC83D6ULL, "qun221cr_cleric.utc"},
	{0x64B902FAA8EAC36BULL, "vxlt_darktown_588_0_0.vlm"},
	{0x64B93122EADDAA96ULL, "gen00fl_anders6202358.cl"},
	{0x64B94B14513A3D73ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc105a_0n.dds"},
	{0x64B960A8D24A4AF9ULL, "zz_t_trailer_shot.cut"},
	{0x64B9759F426980A5ULL, "atl_startmenu_dxt5_03.dds"},
	{0x64B996F638685F23ULL, "vxw_bonepit_bq_0_0.vlm"},
	{0x64B9B2317CE8C598ULL, "6124694_m.fxe"},
	{0x64B9B6E0AE91751CULL, "vxdr_primevalentrance_1jc_0_0.vlm"},
	{0x64BA57D5CA998AC4ULL, "6176569_m.fxe"},
	{0x64BBCD4F7ECA6F5CULL, "hlac_dr_primevalentrance_1726_1y.dds"},
	{0x64BBFB979DBD51D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_stroud.ani"},
	{0x64BC32A3BE06AE4AULL, "vxw_blightlands_mn_0_0.vlm"},
	{0x64BC3344B1DEE182ULL, "581018487.ogg"},
	{0x64BC5EDD07F61A36ULL, "vxlt_undercityencounter_1hc_0_0.vlm"},
	{0x64BC72BA496A909EULL, "gen_im_trs_cot_lgt_glv_r2.uti"},
	{0x64BC755E979C0063ULL, "art\\characters\\creatures\\mubari\\animation\\com_lunge.gad"},
	{0x64BD1D96E04478BCULL, "vxw_encounter2_ld_0_0.vlm"},
	{0x64BD457E114CF74DULL, "vxlt_house_g8_0_0.vlm"},
	{0x64BD7C94213A8C66ULL, "77657884.ogg"},
	{0x64BDA1900CADCFE7ULL, "vxht_mansion2b_2ec_0_0.vlm"},
	{0x64BE213B4085B263ULL, "hlac_lt_0_lowtown_n_-131073_3x.dds"},
	{0x64BE2414FF965F7CULL, "112182720.ogg"},
	{0x64BEF9BEC5DFEC85ULL, "hlac_ht_0_hightown_n_131072_0y.dds"},
	{0x64BEFC3ED9AA4465ULL, "one100dg_kelder6060649.cl"},
	{0x64BF503D7BDA0B1CULL, "nat380_nathaniel6066182.cl"},
	{0x64BF6D91527F7C9AULL, "t3_wdg_w29.tnt"},
	{0x64BF92284C57CB64ULL, "docks_male_overseer_mob6176618.cl"},
	{0x64BFAA567B6C6BB5ULL, "zz_dae_debug6029498.cl"},
	{0x64BFE1FD572AD96BULL, "6176575_m.fxe"},
	{0x64C06145FAF624E5ULL, "397241256.ogg"},
	{0x64C09BE0890A95DFULL, "llac_lt_0_lowtown_docks_a3_n_-65537_2.dds"},
	{0x64C13CE2595D8BB1ULL, "147576526.ogg"},
	{0x64C1C5930CB749F6ULL, "vxlt_0_lowtown_a2_n_c_eqt_0_0.vlm"},
	{0x64C21C4DC2497CE0ULL, "genip_trap_spike.utp"},
	{0x64C25900F4F8D011ULL, "6149250_m.fxe"},
	{0x64C30C8793C833D1ULL, "6150173_m.fxe"},
	{0x64C3525E97A1E348ULL, "art\\characters\\creatures\\mubari\\animation\\com_lunge.evt"},
	{0x64C37673B98F9539ULL, "35255667.ogg"},
	{0x64C3947AC43FBDE5ULL, "llac_lt_0_lowtown_a3_n_c_1_1.dds"},
	{0x64C3BCAEB1D0AF71ULL, "6150336_f.fxe"},
	{0x64C3D0217C6E1273ULL, "hlac_lt_0_lowtown_docks_a3_-131073_4x.dds"},
	{0x64C3EA3E25A82824ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_noble1_arishokaddress.gad"},
	{0x64C3F30526FAF53DULL, "vxw_blightlands_s3_0_0.vlm"},
	{0x64C4397E2A65DAC2ULL, "dae000ip_crowd_lm4_talk17.utp"},
	{0x64C52068FEB2EF70ULL, "ser121_jethann6106558.cl"},
	{0x64C5DD2D01C33C8BULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_meda.phy"},
	{0x64C652BC3704D4A4ULL, "3499618.ogg"},
	{0x64C6C9CCAB91CD27ULL, "vxw_encounter2_camp_fy_0_0.vlm"},
	{0x64C6F10E08235997ULL, "hlac_dr_0_deeproads_5842_16y.dds"},
	{0x64C7FBB6170A5EA8ULL, "hlac_lt_0_lowtown_n_-262143_0y.dds"},
	{0x64C808D32A5E77A3ULL, "vxw_bonepit_mp_0_0.vlm"},
	{0x64C957D0E4E4C882ULL, "812340594.ogg"},
	{0x64C9FDA54856C305ULL, "hlac_ca_caverns1_24425_13y.dds"},
	{0x64CA278743E6B94DULL, "2960976.ogg"},
	{0x64CA2BFDF7D96A98ULL, "548664269.ogg"},
	{0x64CA3C45E9FCFCDCULL, "967715533.ogg"},
	{0x64CAA480AFFCC6CAULL, "vxw_blightlands_vo_0_0.vlm"},
	{0x64CADEA8E37D2697ULL, "vxca_cavern1_haunted_j2f_0_1.vlm"},
	{0x64CAF93E25AE6002ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_noble1_arishokaddress.ani"},
	{0x64CC2A50661AEC50ULL, "6201588_m.fxe"},
	{0x64CCC65A1E4888B8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a_1.mao"},
	{0x64CD3931F79BA537ULL, "368814331.ogg"},
	{0x64CD42AD87179874ULL, "hlac_ca_cavern1_haunted_24426_8z.dds"},
	{0x64CD4D951AB5FDA3ULL, "follower_banter6173038.cl"},
	{0x64CDAC851E5B3A08ULL, "vxca_cavern1_haunted_j4r_0_0.vlm"},
	{0x64CE08EE9D327D19ULL, "6105647_m.fxe"},
	{0x64CE1C08BFAAFB90ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_maga_0d.dds"},
	{0x64CE22D39D78B7E3ULL, "6118034_m.fxe"},
	{0x64CE285AA531DF46ULL, "295381156.ogg"},
	{0x64CE71D7B5252D40ULL, "640663407.ogg"},
	{0x64CEBDF61E44746EULL, "vxdr_primevaltomb_4zu_0_1.vlm"},
	{0x64CEF469E243EDCAULL, "6100236_m.fxe"},
	{0x64CF39387A03EF7BULL, "vxca_caverns2_83u_0_1.vlm"},
	{0x64CF6F800473602CULL, "41528374.ogg"},
	{0x64CF95A410D382E3ULL, "vxht_brothel_n_6br_0_0.vlm"},
	{0x64CF9ED48FBE1D9FULL, "6144426_m.fxe"},
	{0x64D0102F6626AC96ULL, "primitives_primitive_p0.psh.fxo.dx11.assembly"},
	{0x64D147FCED99B2A4ULL, "6126599_m.fxe"},
	{0x64D16FAA92F77C4FULL, "939932857.ogg"},
	{0x64D1A7021EC73CB9ULL, "514705452.ogg"},
	{0x64D2074AC1CBB686ULL, "6152323_m.fxe"},
	{0x64D2504FD9BEA571ULL, "llac_lt_gallowstemplar_n_2388_3.dds"},
	{0x64D25F9B25AC272BULL, "182964221.ogg"},
	{0x64D264E218F10DBDULL, "vxw_blightlands_t2_0_0.vlm"},
	{0x64D27936260A287CULL, "936119898.ogg"},
	{0x64D2B91FA49236BAULL, "vxw_sundermount_340_0_0.vlm"},
	{0x64D2D2AA9527E080ULL, "6033716_m.fxe"},
	{0x64D2E25DD2923004ULL, "6118631_m.fxe"},
	{0x64D3200EA307F882ULL, "808604084.ogg"},
	{0x64D323A6B6CF4295ULL, "lgt311_shw_mer_mettin6156542.cl"},
	{0x64D32CA2F0919A55ULL, "559151069.ogg"},
	{0x64D37C1D0760B7F3ULL, "6166677_m.fxe"},
	{0x64D446324A82EFA3ULL, "80343773.ogg"},
	{0x64D47415F1DB58A1ULL, "6107234_m.fxe"},
	{0x64D5057057EF29C4ULL, "vxlt_undercitydungeon_1ku_0_0.vlm"},
	{0x64D53327801505BEULL, "vxht_0_hightown_a3_as2_0_0.vlm"},
	{0x64D53BA99D87E5A7ULL, "vxw_encounter2_camp_ax_0_0.vlm"},
	{0x64D586B6B186A12BULL, "vxw_encounter2_camp_rp_0_1.vlm"},
	{0x64D5967F10951DEBULL, "llac_lt_gallowscourtyard_n_-2_3.dds"},
	{0x64D5DF12BFC548BAULL, "850328064.ogg"},
	{0x64D61020A9EB8275ULL, "6053571_m.fxe"},
	{0x64D63851B42E484BULL, "vxlt_0_lowtown_docks_cqc_0_0.vlm"},
	{0x64D65E12F6C8B094ULL, "vxlt_gallowstemplar_f_39i_0_0.vlm"},
	{0x64D6DBF11DFAEE61ULL, "158047327.ogg"},
	{0x64D753B886EA8386ULL, "hlac_lt_0_lowtown_a2_-3_3x.dds"},
	{0x64D7B81A9291F3F6ULL, "llac_ht_chantry_2534_1.dds"},
	{0x64D7B9944102E579ULL, "art\\vfx\\materials\\vfx_soulloop_add.mao"},
	{0x64D910924DE2A8C2ULL, "follower_banter6160767.cl"},
	{0x64D96164952A6478ULL, "rob_dal.tnt"},
	{0x64D990B6FEFCBDD0ULL, "vxlt_gallowstemplar_f_456_0_0.vlm"},
	{0x64D9C00E094753D6ULL, "vxdr_primevalentrance_1a4_0_0.vlm"},
	{0x64D9EFD700C9D40FULL, "5155681.ogg"},
	{0x64D9F60ABB5D0A7BULL, "869293334.ogg"},
	{0x64DA0150C430A9BCULL, "vxw_encounter2_rz_0_0.vlm"},
	{0x64DA0A75C0F05061ULL, "dae000_flemeth6013320.cl"},
	{0x64DA3087452DB481ULL, "one100_party_comments6133375.cl"},
	{0x64DA999846911BFCULL, "6132833_m.fxe"},
	{0x64DB187BEB315C15ULL, "vxht_bartrand_n_5wt_0_0.vlm"},
	{0x64DB5F3B1E9D5024ULL, "hlac_lt_gallowscourtyard_n_-1_2y.dds"},
	{0x64DB91A6C164D21AULL, "llac_lt_undercityserialkiller_-1_48.dds"},
	{0x64DB9824A272F704ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_4z.dds"},
	{0x64DBC642F29A197EULL, "zz_vault_debug6200331.cl"},
	{0x64DBCE5A41577001ULL, "lgt311_shw_mer_agatha.bnk"},
	{0x64DC1E6210515288ULL, "vxca_cavern1_haunted_j3s_0_0.vlm"},
	{0x64DC82BDD26553D4ULL, "might_of_the_sten.dds"},
	{0x64DD183860A46319ULL, "vxht_brothel_n_6c3_0_0.vlm"},
	{0x64DD72901433205DULL, "mag121_viveka6082616.cl"},
	{0x64DDC128115849B4ULL, "hlac_lt_0_lowtown_65537_0y.dds"},
	{0x64DDE3AEB6CEDE93ULL, "zz_one_debug6075024.cl"},
	{0x64DE078102C99B63ULL, "vxht_brothel_n_6ou_0_0.vlm"},
	{0x64DE10B181B4361FULL, "6141726_m.fxe"},
	{0x64DEE4B1606FA85BULL, "1049970348.ogg"},
	{0x64DEF05A4159E071ULL, "lgt311_shw_mer_agatha.cnv"},
	{0x64DF0BEE31A93238ULL, "hlac_ht_0_hightown_a3_131071_1z.dds"},
	{0x64DF0D4B9AB98202ULL, "6131359_m.fxe"},
	{0x64DF36F625FCC72EULL, "hlac_lt_warehouse_1_6z.dds"},
	{0x64DF7B00431F2721ULL, "zz_dre_debug6054659.cl"},
	{0x64DFE5B7D8BD0385ULL, "6032527_m.fxe"},
	{0x64E12AFC96884F3AULL, "vxw_sundermount_331_0_0.vlm"},
	{0x64E18119F1F0A5F1ULL, "137339991.ogg"},
	{0x64E1D5737285FA54ULL, "zz_followers_banter6158085.cl"},
	{0x64E22560D20CA3DFULL, "vxw_bonepit_26a_0_0.vlm"},
	{0x64E286A523DE4BB8ULL, "23301360.ogg"},
	{0x64E2A27ABDBC500DULL, "hm_amb_templar_wander.mor"},
	{0x64E2CDBD6BBF011BULL, "6121557_m.fxe"},
	{0x64E3774D49E54244ULL, "vxlt_undercitydungeon_1nz_0_0.vlm"},
	{0x64E4A77C2EE5D411ULL, "136770676.ogg"},
	{0x64E4E0E8BD71CCA2ULL, "lgt150_nat_were_supporter6110974.cl"},
	{0x64E5578B938A595BULL, "w_bonepitdestroyed_a3_247_0.msh"},
	{0x64E5698B938A77A6ULL, "w_bonepitdestroyed_a3_247_0.mao"},
	{0x64E5758B938A8C45ULL, "w_bonepitdestroyed_a3_247_0.mmh"},
	{0x64E5776F3C0D8535ULL, "dae000_bodhan6146803.cl"},
	{0x64E612BDD754FA36ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_9whitelyrium.msh"},
	{0x64E61CBDD7550B34ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_9whitelyrium.mmh"},
	{0x64E620BDD7551247ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_9whitelyrium.mao"},
	{0x64E6746DF189E47BULL, "6136104_m.fxe"},
	{0x64E67621C021E40BULL, "6172665_m.fxe"},
	{0x64E6AF52EA9CDB0BULL, "6165688_f.fxe"},
	{0x64E6BFBB19CAD3BBULL, "and100_anders6096028.cl"},
	{0x64E6F3C5435DF12CULL, "6035145_m.fxe"},
	{0x64E782E9E618A72AULL, "vxht_0_hightown_bdu_0_0.vlm"},
	{0x64E790BBD6993569ULL, "6128003_f.fxe"},
	{0x64E7B480D440389DULL, "915870104.ogg"},
	{0x64E89C2E27A3C50DULL, "vxlt_gallowstemplar_f_2go_0_0.vlm"},
	{0x64E8F914F003609FULL, "atl_slideshow_dxt5_00.dds"},
	{0x64E91E2D6D8091D1ULL, "art\\vfx\\cinematic\\vfx_blood_dripping_c.anb"},
	{0x64E96EE6B681CCCEULL, "2049083.ogg"},
	{0x64EA2F4E936E4E51ULL, "6080283_m.fxe"},
	{0x64EA51704A550922ULL, "follower_barks6153472.cl"},
	{0x64EAAF05E22A058CULL, "782736261.ogg"},
	{0x64EACCCDA453036BULL, "gen_im_npc_dwf_arm_cht_lgtb.uti"},
	{0x64EB15256FBE016BULL, "gen00fl_anders6135985.cl"},
	{0x64EB5C3C2F710EA2ULL, "vxlt_0_lowtown_n_deg_0_0.vlm"},
	{0x64EB67B552C92D0EULL, "t1_mue_pu2.tnt"},
	{0x64EB6E1FDFD13DFAULL, "6179873_f.fxe"},
	{0x64EBEB8430527108ULL, "765819420.ogg"},
	{0x64EC1E62D35B10BDULL, "6109438_m.fxe"},
	{0x64EC5A55D894EA34ULL, "hlac_w_blightlands_0_7z.dds"},
	{0x64EC65767177AA0FULL, "gen00fl_anders6095500.cl"},
	{0x64EC795DF4BFB3E3ULL, "vxht_brothel_n_6ht_0_0.vlm"},
	{0x64EC893A365B4BB4ULL, "vxht_mansion2b_1ce_0_0.vlm"},
	{0x64ED5C640B4F3AD8ULL, "772977841.ogg"},
	{0x64EEEE44E75C1A6FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicb.mao"},
	{0x64EEFA44E75C2E0CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicb.mmh"},
	{0x64EF0044E75C38FEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicb.msh"},
	{0x64EF204340CCF46AULL, "6053600_m.fxe"},
	{0x64EF5EA39481BAF1ULL, "vxca_caverns1_dph_0_0.vlm"},
	{0x64EF785C923F8E7AULL, "vxca_cavern1_haunted_e8s_0_0.vlm"},
	{0x64EFB5F18B82DB8EULL, "6005056_m.fxe"},
	{0x64F0D2CB48524739ULL, "hlac_w_encounter2_196610_0x.dds"},
	{0x64F0E5043EC15022ULL, "vxca_cavern1_haunted_dpj_0_0.vlm"},
	{0x64F123F13B3DE899ULL, "lgt000im_ultra_kirkmap.uti"},
	{0x64F2E9A509EBAC5BULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgta.mmh"},
	{0x64F2F5A509EBC0F8ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgta.mao"},
	{0x64F307A509EBDF45ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgta.msh"},
	{0x64F33D57BB2234F8ULL, "983035266.ogg"},
	{0x64F37C0C0D40BF74ULL, "6187815_m.fxe"},
	{0x64F41F4BEF3FA3A6ULL, "zz_qcr_debug6058846.cl"},
	{0x64F4629789CABAE2ULL, "6150013_m.fxe"},
	{0x64F548C00C5B67C0ULL, "640238335.ogg"},
	{0x64F5E556AE9B69CAULL, "6134904_f.fxe"},
	{0x64F5F4C6D80EBFAAULL, "vxht_0_hightown_bav_0_0.vlm"},
	{0x64F6E7AAC61AF99DULL, "57910092.ogg"},
	{0x64F73F2D6D8D074AULL, "art\\vfx\\cinematic\\vfx_blood_dripping_c.mmh"},
	{0x64F79978E57D9F43ULL, "dae000_flemeth6013403.cl"},
	{0x64F7AEF2B911BAFDULL, "vxht_keep_6zt_0_0.vlm"},
	{0x64F8037135B748FBULL, "vxht_mansion_n_5m7_0_0.vlm"},
	{0x64F840E6EE210EBCULL, "hlac_lt_0_lowtown_docks_n_-65537_0x.dds"},
	{0x64F879806084F6E1ULL, "hlac_lt_0_lowtown_a3_n_c_0_6z.dds"},
	{0x64F917EF4E0972DFULL, "mcr370_meredith6083050.cl"},
	{0x64F9600C15C7655DULL, "dae100_alistair_drunk6135138.cl"},
	{0x64F9FBF244909419ULL, "vxht_brothel_n_6i1_0_0.vlm"},
	{0x64FA13FF2769A8E5ULL, "6151149_m.fxe"},
	{0x64FABAE166E8055CULL, "character_character_p83.psh.fxo"},
	{0x64FACC39E4372336ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinswall.phy"},
	{0x64FAEB3AE6B5CC63ULL, "vxht_brothel_n_6uw_0_0.vlm"},
	{0x64FAFB1728516434ULL, "vxht_mansion2b_1fj_0_0.vlm"},
	{0x64FB884BC5752436ULL, "hlac_lt_undercitydungeon_-1_24y.dds"},
	{0x64FB94BFCD2CF4C9ULL, "6132858_m.fxe"},
	{0x64FC30921446D4E5ULL, "vxlt_0_lowtown_docks_a3_n_aqk_0_0.vlm"},
	{0x64FC7822B61D925FULL, "w_bonepitmining_a2_106_0.msh"},
	{0x64FC7E22B61D9C01ULL, "w_bonepitmining_a2_106_0.mmh"},
	{0x64FC8A22B61DB0A2ULL, "w_bonepitmining_a2_106_0.mao"},
	{0x64FD1E8917B9E35BULL, "6045422_m.fxe"},
	{0x64FE4806CFA33D0EULL, "vxht_bartrand_n_6ij_0_0.vlm"},
	{0x64FE58F35A10A769ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_wf_injured.gad"},
	{0x64FEB763E54ACDA0ULL, "64528823.ogg"},
	{0x64FEF5C0A3977280ULL, "qun110_varnell6047725.cl"},
	{0x64FF130FA68E81B1ULL, "combat_dw_r1_a2.gda"},
	{0x64FF6C6C3B14C560ULL, "w_encounter3_33_0.phy"},
	{0x64FF9F4551BE7E9BULL, "fxa_gas_smoke_blend.mao"},
	{0x65004159EEFD0954ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_0y.dds"},
	{0x65004EB8BA6E3131ULL, "gencr_profane_b.utc"},
	{0x65005E7AC8A6218DULL, "vxht_0_hightown_a3_a84_0_0.vlm"},
	{0x6500909F272B6AB3ULL, "6169478_m.fxe"},
	{0x6500DB8BD2BF4A3CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow1win.mao"},
	{0x6500E78BD2BF5EDFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow1win.mmh"},
	{0x6500ED8BD2BF6881ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow1win.msh"},
	{0x650115C73DC13068ULL, "lt_0_lowtown_a3_n_c_35_0.anb"},
	{0x65018D7C82A0C5E0ULL, "vxlt_gallowscourtyard_70o_0_0.vlm"},
	{0x65018DAACB7D429CULL, "790806613.ogg"},
	{0x6501B980339261B9ULL, "hlac_lt_0_lowtown_docks_-196607_0y.dds"},
	{0x6501DDEF870A10B8ULL, "hlac_lt_0_lowtown_docks_262143_0x.dds"},
	{0x650256D0DF6E4286ULL, "vxht_0_hightown_7fv_0_0.vlm"},
	{0x6502892E4C05B7F0ULL, "dae000cr_grand_cleric6182022.cl"},
	{0x6502A8C4A813E78FULL, "vxlt_foundry_n_w2_0_0.vlm"},
	{0x650311F39BBE1FB0ULL, "6033679_m.fxe"},
	{0x650393E20D301A4DULL, "6096767_m.fxe"},
	{0x65039C5D67131869ULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180l.ani"},
	{0x6503C588C37859C7ULL, "vxlt_0_lowtown_docks_b0i_0_0.vlm"},
	{0x6503D10FB1831DF1ULL, "6172380_m.fxe"},
	{0x65046B966C841B40ULL, "hlac_lt_darktown_-1_36y.dds"},
	{0x650491AF3BFDADE3ULL, "6140602_m.fxe"},
	{0x6504B539E43F50D7ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinswall.msh"},
	{0x6504C739E43F6F3AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinswall.mao"},
	{0x6504CB39E43F7649ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinswall.mmh"},
	{0x650500A7D3DB1CE3ULL, "t1_skn_cas.tnt"},
	{0x650529F35A1675B7ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_wf_injured.ani"},
	{0x65056F8F2595AA69ULL, "m_rdr370ar_finale.bnk"},
	{0x650583466AC04C01ULL, "258416087.ogg"},
	{0x6505B3367D088F0EULL, "6163518_m.fxe"},
	{0x6505D785959AC599ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightpath_d.dds"},
	{0x6505E105344B5F7CULL, "vxlt_0_lowtown_docks_c07_0_0.vlm"},
	{0x650620CFAB07D37DULL, "vxht_keep_6gs_0_0.vlm"},
	{0x65063784BA2FF9DFULL, "gen00fl_carver6097384.cl"},
	{0x650668489D60BC65ULL, "746078344.ogg"},
	{0x650685D3E4426AA0ULL, "565699244.ogg"},
	{0x65068B6C3B1B1818ULL, "w_encounter3_33_0.rml"},
	{0x6506B235EFE92864ULL, "vxht_0_hightown_n_ana_0_0.vlm"},
	{0x6506E9CFDD582C70ULL, "dae100_viscount6118487.cl"},
	{0x650716B5938FE006ULL, "art\\characters\\weapons\\axes\\w_awe_dsaa.phy"},
	{0x6507362C664E2128ULL, "gen00fl_fenris.cnv"},
	{0x6507A51835B30852ULL, "670314513.ogg"},
	{0x65087DE1A0595164ULL, "genim_arm_boo_inq.uti"},
	{0x65092CBE58DE1DDCULL, "character_character_p72.psh.fxo"},
	{0x65096CF41A477CB4ULL, "vxht_mansion2b_1yk_0_0.vlm"},
	{0x650A582C66509090ULL, "gen00fl_fenris.bnk"},
	{0x650A984C75784098ULL, "6160984_m.fxe"},
	{0x650AB75D671964EBULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180l.gad"},
	{0x650AC16A2CA9EEFAULL, "537469453.ogg"},
	{0x650AD1F8A0FBDABFULL, "ht_0_hightown_n_37_0.rml"},
	{0x650AF2B4C30C97BFULL, "184337597.ogg"},
	{0x650B55ECBB073D54ULL, "vxht_chantry_2cj_0_0.vlm"},
	{0x650B61B6BFDA9758ULL, "6144577_m.fxe"},
	{0x650B6BADAC700EEAULL, "209228931.ogg"},
	{0x650CB9E3C199558EULL, "vxht_bartrand_n_6rk_0_0.vlm"},
	{0x650CBF28D2ADB246ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb102a_0t.dds"},
	{0x650CF79A6635A598ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12a.phy"},
	{0x650D9CC91DD3554CULL, "w_sundermount_40_0.phy"},
	{0x650DC8BE22AD8122ULL, "vxca_cavern1_haunted_dnt_0_0.vlm"},
	{0x650DD9B9B4809B02ULL, "177350896.ogg"},
	{0x650E5FCA79DC2CECULL, "967060402.ogg"},
	{0x650F225A59486947ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobb_0t.dds"},
	{0x650FFF597496DE60ULL, "vxlt_gallowscourtyard_75l_0_0.vlm"},
	{0x651007372AA718E3ULL, "follower_banter6141945.cl"},
	{0x651073ABD7FBC1D2ULL, "485159279.ogg"},
	{0x6510B98A5C4A7983ULL, "6167292_m.fxe"},
	{0x651142A78F0C1396ULL, "255541653.ogg"},
	{0x65115705675E07A7ULL, "77670976.ogg"},
	{0x65117C42D1384CFAULL, "vxw_bonepit_1wm_0_1.vlm"},
	{0x65119D408850C311ULL, "491021738.ogg"},
	{0x6511B09C35BA2CC3ULL, "dae010_urchin_m1.utc"},
	{0x6511B45D671F7E50ULL, "art\\characters\\creatures\\desiredemon\\animation\\ctn180l.evt"},
	{0x6511FC3737C5D2DFULL, "zz_cave_accolades_cinetest.cut"},
	{0x65121F7BE75EE6DFULL, "gen00fl_aveline6137463.cl"},
	{0x651226F37F0569D4ULL, "vxht_0_hightown_n_7fw_0_0.vlm"},
	{0x65123765B56E7247ULL, "vxlt_0_lowtown_docks_b7h_0_0.vlm"},
	{0x65123A88D315B851ULL, "llac_lt_hangedman_n_-1_21.dds"},
	{0x6512B89E8913B546ULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_imp_p.impa.evt"},
	{0x6512F469D3499069ULL, "376899094.ogg"},
	{0x651301BA3E43BC08ULL, "llac_ht_mansion2b_818_20.dds"},
	{0x65130368B4C42DCBULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_browsing_store.ani"},
	{0x65133629B688430EULL, "art\\vfx\\combat\\vfx_gib_hm_masc_ltr.anb"},
	{0x65135EB6E49FD096ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.ani"},
	{0x6513A5E687E105C8ULL, "vxht_mansion2_1bs_0_1.vlm"},
	{0x6513D74469E137A9ULL, "fen210_hadriana6110585.cl"},
	{0x65140089B072F804ULL, "art\\vfx\\combat\\vfx_d_cast_aoe_c.anb"},
	{0x651468FF33E3A92EULL, "sky_ht_0_hightownexplosion_n.mao"},
	{0x651473C91DD92E54ULL, "w_sundermount_40_0.rml"},
	{0x65147EB407C8E581ULL, "6085763_m.fxe"},
	{0x65147F74D8BB93C1ULL, "vxdr_primevaltomb_6vq_0_0.vlm"},
	{0x651521ADFCA8B48CULL, "724857463.ogg"},
	{0x651546A44C13A2BBULL, "mag371_elsa6065019.cl"},
	{0x651583C4AC27D43EULL, "55744043.ogg"},
	{0x6516594A1A1043A6ULL, "319023839.ogg"},
	{0x6516761C8188AC60ULL, "vxht_0_hightown_a3_bx6_0_0.vlm"},
	{0x65179E9B4AD4365CULL, "character_character_p21.psh.fxo"},
	{0x6517CD44E6017E61ULL, "gen00fl_aveline6132893.cl"},
	{0x65180E86F82D6823ULL, "213350939.ogg"},
	{0x65187AD9CE16DFFAULL, "dae300_alistair_king6134906.cl"},
	{0x6518D5DC20911114ULL, "244451314.ogg"},
	{0x6518E052A339A5EFULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb105a_0t.dds"},
	{0x6518EEA5DC019CC4ULL, "1070537544.ogg"},
	{0x6519C036AABAC95FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edge.msh"},
	{0x6519C636AABAD301ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edge.mmh"},
	{0x6519CF9E8919FA91ULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_imp_p.impa.gad"},
	{0x6519D236AABAE7A2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edge.mao"},
	{0x6519D393B1D0AFD8ULL, "6141677_m.fxe"},
	{0x6519DEF981CCF238ULL, "917892468.ogg"},
	{0x651A15E0F8DB9322ULL, "hlac_lt_0_lowtown_-131072_2x.dds"},
	{0x651A23B4B1A75EE6ULL, "llac_lt_0_lowtown_docks_a3_-131074_6.dds"},
	{0x651A40A369F25860ULL, "llac_ht_mansion_n_3371_2.dds"},
	{0x651A700096B41DC5ULL, "vxca_cavern1_haunted_e8x_0_0.vlm"},
	{0x651A7E55F5EF8C73ULL, "1003799811.ogg"},
	{0x651AA3FEC0BF036BULL, "618141987.ogg"},
	{0x651AACB0628098AEULL, "vxlt_gallowstemplar_n_64y_0_0.vlm"},
	{0x651ABC49366822DDULL, "6019887_m.fxe"},
	{0x651ACDF36821D7FAULL, "vxca_cavern1_haunted_e7n_0_0.vlm"},
	{0x651AE066286C58FAULL, "llac_lt_blackemporium_1055_15.dds"},
	{0x651AF2B21D473C36ULL, "61376054.ogg"},
	{0x651B4002A5627E03ULL, "gen_im_cns_pot_attribute_02.uti"},
	{0x651C0D535DAC479EULL, "toolset_audio_events.gda"},
	{0x651C15EFE3D81DDDULL, "gen00fl_sebastian6189044.cl"},
	{0x651C6E1F2CD78B92ULL, "1043336974.ogg"},
	{0x651D0B1591F3B2C5ULL, "hlac_ht_0_hightown_n_-65537_0z.dds"},
	{0x651D415E1D6EB477ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magb_0t.dds"},
	{0x651D6C307228371CULL, "zz_lgt_debug6114334.cl"},
	{0x651D74590B179BB3ULL, "6163678_m.fxe"},
	{0x651D76EF41E20DEAULL, "57983459.ogg"},
	{0x651E0C4F6F2A1269ULL, "vxlt_0_lowtown_a2_n_c_ci5_0_0.vlm"},
	{0x651EE771028BAD1AULL, "625666089.ogg"},
	{0x651EEBD7E0434823ULL, "864672132.ogg"},
	{0x651EF2E05EF5FE62ULL, "6036457_f.fxe"},
	{0x651F0DF305D9A67FULL, "vxlt_foundry_n_vq_0_0.vlm"},
	{0x651F76BC6E78F8DEULL, "829470172.ogg"},
	{0x651F8BE9318141A6ULL, "vxca_caverns2_7qt_0_0.vlm"},
	{0x651FA11B87C1D22BULL, "lgt160cr_wk_soph_sophia.utc"},
	{0x651FE064A00ED6DEULL, "6016647_m.fxe"},
	{0x65208F63B7F0F9F7ULL, "321754506.ogg"},
	{0x6520C7C13548C759ULL, "6108235_m.fxe"},
	{0x6520D09E89201B3FULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_imp_p.impa.ani"},
	{0x6520EB68B4D04242ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_browsing_store.evt"},
	{0x652128865407DC96ULL, "hlac_ca_caverns1_mines_24426_10y.dds"},
	{0x65214A5DADF1092AULL, "vxht_0_hightown_bpq_0_0.vlm"},
	{0x652166B6E4AC1B87ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.evt"},
	{0x6521CF977339D77BULL, "vxw_encounter3_x_0_0.vlm"},
	{0x65221F89B07F6A07ULL, "art\\vfx\\combat\\vfx_d_cast_aoe_c.mmh"},
	{0x652257E010D377E2ULL, "506957011.ogg"},
	{0x6522D948DD371493ULL, "loadhints_base.gda"},
	{0x6522F2DCB8B780E7ULL, "6096062_m.fxe"},
	{0x652338E1D41222EEULL, "vxw_bonepit_101_0_0.vlm"},
	{0x65239B835C4F9FE0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_maga_0t.dds"},
	{0x6523A22FE968EF3DULL, "one280_emrys6104261.cl"},
	{0x6523BCA05DACE441ULL, "one300_zevran6089121.cl"},
	{0x6523BF5EBBAE305AULL, "llac_lt_0_lowtown_a2_n_c_-3_6.dds"},
	{0x6524B1BE2872C196ULL, "fx_dust_bl.dds"},
	{0x6524B741FE2DD471ULL, "vxw_bonepit_2an_0_0.vlm"},
	{0x65252639F5A79322ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobd_0n.dds"},
	{0x6525896C3B35C5F5ULL, "w_encounter3_33_0.msh"},
	{0x65258B6C3B35C96BULL, "w_encounter3_33_0.mmh"},
	{0x6525976C3B35DD08ULL, "w_encounter3_33_0.mao"},
	{0x6525A601B9C23F22ULL, "429756300.ogg"},
	{0x6525CF5FF4B3300BULL, "hlac_ht_keep_n_3735_0y.dds"},
	{0x65263C7ED845C0A6ULL, "953394902.ogg"},
	{0x652650ADFE33ADB4ULL, "vxht_mansion2b_1wi_0_0.vlm"},
	{0x6526BBF1F62B93F8ULL, "6118820_m.fxe"},
	{0x6526EA68B4D4AC55ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_browsing_store.gad"},
	{0x65270449AF98CA6CULL, "756535045.ogg"},
	{0x6527E17DDB8F6979ULL, "302530152.ogg"},
	{0x652839A69EF36E54ULL, "vxht_chantry_2yh_0_0.vlm"},
	{0x65283F51B007B0C7ULL, "plt_zz_demo_plot_flags.plo"},
	{0x65287F22B642D97EULL, "w_bonepitmining_a2_106_0.phy"},
	{0x652887B6E4B272D0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.gad"},
	{0x6528D0B0A60859F8ULL, "335261725.ogg"},
	{0x652914B53CD460FCULL, "gen00fl_fenris6128078.cl"},
	{0x65295B9714F95974ULL, "169818713.ogg"},
	{0x652961EE0F564A48ULL, "6054232_m.fxe"},
	{0x6529A02D6B7ABBE2ULL, "lt_backalley_g_1_0_ll.mmh"},
	{0x6529B3B317E57D51ULL, "vxht_0_hightown_a3_c73_0_0.vlm"},
	{0x652A043EBF80235DULL, "hlac_ht_playershouse_n_855_3y.dds"},
	{0x652A31E875ADD125ULL, "vxlt_gamlenshouse_n_5b_0_0.vlm"},
	{0x652AC3DB0A6736A4ULL, "953719723.ogg"},
	{0x652AD98580438E01ULL, "71068711.ogg"},
	{0x652B4A2E756B463AULL, "qcr220_orsino6060726.cl"},
	{0x652BC6BB5A4DC607ULL, "318008616.ogg"},
	{0x652C414CF3EB93A5ULL, "1045095937.ogg"},
	{0x652C44C6D59A33B5ULL, "art\\characters\\playercharacter\\humanmale\\boo_wic.phy"},
	{0x652C506C0B99B790ULL, "1022904521.ogg"},
	{0x652C5A04F8C2CEDBULL, "6066909_f.fxe"},
	{0x652C712DF91597D8ULL, "324660038.ogg"},
	{0x652CB3773DE50BF9ULL, "6150740_m.fxe"},
	{0x652D4295E6A5292BULL, "dae000_orsino6085504.cl"},
	{0x652D91CCA3327439ULL, "6042366_m.fxe"},
	{0x652DFDC623775A26ULL, "vxca_caverns2_7vu_0_0.vlm"},
	{0x652E12AF50870116ULL, "t3_eye_grn.tnt"},
	{0x652E40B72F86AD29ULL, "853502934.ogg"},
	{0x652E6BB1F61F949AULL, "vxw_sundermount_2ef_0_0.vlm"},
	{0x652E7AFE704875C2ULL, "sf_d_strip.vsh.fxo"},
	{0x652F254BC04105B6ULL, "743690764.ogg"},
	{0x652F317A28AC6842ULL, "docks_male_sailor_mob6176640.cl"},
	{0x652F33748DCD8E3BULL, "hlac_w_bonepitmining_a2_262148_0z.dds"},
	{0x652F52AA55949562ULL, "pro000cr_meeran.utc"},
	{0x652FAC0F4042CC7AULL, "vxlt_0_lowtown_a3_n_c_hd3_0_0.vlm"},
	{0x6530D7D9A3E9727FULL, "vxlt_gallowstemplar_f_5ux_0_0.vlm"},
	{0x653130366AB0397CULL, "hlac_w_encounter2_196608_0x.dds"},
	{0x65317D26D0611168ULL, "hlac_w_encounter1_2_0y.dds"},
	{0x6532844E4A3C30DFULL, "zz_mag_debug6054497.cl"},
	{0x6533430339206D91ULL, "6118689_m.fxe"},
	{0x6533CCAF4F545A29ULL, "hlac_dr_primevalentrance_1724_1z.dds"},
	{0x65342174930EA6DEULL, "llac_w_encounter3_-262145_0.dds"},
	{0x6534268496518F73ULL, "dae000_flemeth6013340.cl"},
	{0x6534494D3975E44FULL, "vxht_chantry_32s_0_0.vlm"},
	{0x65346008B8EBEFD9ULL, "art\\vfx\\models\\vfx_lightplane.msh"},
	{0x65346A08B8EC00C7ULL, "art\\vfx\\models\\vfx_lightplane.mmh"},
	{0x65346E08B8EC0714ULL, "art\\vfx\\models\\vfx_lightplane.mao"},
	{0x6534C5CDF339BEAAULL, "combat_ws_mv_default.gda"},
	{0x65353FE98CE720F6ULL, "mag311_keran6041412.cl"},
	{0x65354BD1D6798DA0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_thin.phy"},
	{0x6535AEB5861F26AEULL, "follower_banter6171130.cl"},
	{0x6535BDF580FD7A10ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_firestatue_dp.dds"},
	{0x6535BE54EF7F5C5AULL, "llac_lt_0_lowtown_n_-131072_7.dds"},
	{0x653689F62724455DULL, "143651551.ogg"},
	{0x6536EC2A79404C3EULL, "qun131_taskmaster6019210.cl"},
	{0x6536F31D349503C8ULL, "vxlt_ruinedhouse_n_31_0_0.vlm"},
	{0x6537502C6677712DULL, "gen00fl_fenris.utc"},
	{0x6537532CE774DCA3ULL, "430526140.ogg"},
	{0x65377F4A628EED69ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft.ani"},
	{0x6537C80261177623ULL, "lgt260cr_way_bloodmage_lead.utc"},
	{0x6538C5894AD58A67ULL, "vxlt_0_lowtown_a2_n_f5l_0_0.vlm"},
	{0x6538CC51BCA2DD24ULL, "vxdr_primevaltomb_6zn_0_2.vlm"},
	{0x6539150A8E2E258FULL, "llac_lt_0_lowtown_a2_131071_7.dds"},
	{0x65391E17D1E5DDCCULL, "dre000_bodahn6031336.cl"},
	{0x653970FC0881ED11ULL, "6149041_m.fxe"},
	{0x6539F59060257CC5ULL, "137068075.ogg"},
	{0x653A2B2D59D4FF20ULL, "dae000ip_crowd_ff2_talk9.utp"},
	{0x653A2FD557E51549ULL, "lgt210_eau_symp_templar6144475.cl"},
	{0x653A4543E538C7A0ULL, "vxlt_darktown_2s9_0_0.vlm"},
	{0x653A7294FF84D7A0ULL, "vxlt_0_lowtown_docks_a3_arx_0_0.vlm"},
	{0x653B0AA50A299990ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_lgta.phy"},
	{0x653B3737E8070BADULL, "hlac_lt_hangedman_-1_14y.dds"},
	{0x653B487B73368DBAULL, "6016935_m.fxe"},
	{0x653B72A2B9F39D28ULL, "one000_ilen6205651.cl"},
	{0x653BFEFE7C0E167EULL, "mag160_feynriel6020445.cl"},
	{0x653C1460338602C0ULL, "vxlt_0_lowtown_docks_n_cnj_0_0.vlm"},
	{0x653C594F8D32ED3CULL, "vxht_brothel_n_6ih_0_0.vlm"},
	{0x653C6FA3156D72A6ULL, "vxca_caverns2_7kv_0_0.vlm"},
	{0x653C81ADAA39B6D8ULL, "vxw_sundermount_22a_0_0.vlm"},
	{0x653C951D16E09B74ULL, "vxht_keep_618_0_0.vlm"},
	{0x653CA24A6D9A8B2AULL, "923869212.ogg"},
	{0x653CB64F105E2576ULL, "gen_im_npc_dwf_arm_glv_lgta.uti"},
	{0x653CDD8EE815AD1AULL, "vxw_sundermount_2ng_0_0.vlm"},
	{0x653D0AC46A97F2BDULL, "517304885.ogg"},
	{0x653DF3A8203FCDBFULL, "vfx_vfx_p0.psh.fxo.dx11"},
	{0x653E22F5A7D7476DULL, "lt_0_lowtown_docks_n_2_0_lc.mmh"},
	{0x653E9A4A629539EBULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft.gad"},
	{0x653ED9C0BBFC6C56ULL, "mattscreen_2.cut"},
	{0x653F08C3A447E8E3ULL, "6107548_m.fxe"},
	{0x653F3A37AB1F6333ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_c.impa.ani"},
	{0x654021A453B0EC6DULL, "vxw_bonepit_11y_0_0.vlm"},
	{0x654044C839F1922EULL, "dae000ip_crowd_mf7_talk9.utp"},
	{0x654094C91DFEA154ULL, "w_sundermount_40_0.anb"},
	{0x6540C97012B02B72ULL, "hlac_dr_primevaltomb_664_2z.dds"},
	{0x65410D5CC4208C6DULL, "zz_dae_debug6029487.cl"},
	{0x654114C73DF71220ULL, "lt_0_lowtown_a3_n_c_35_0.rml"},
	{0x65412A1BA775E4ADULL, "zz_mer_debug6032452.cl"},
	{0x65412F6C327ADA3EULL, "lgt210_qun_blk_qunari6142315.cl"},
	{0x65427CE7527F9DF2ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_hunch_en.gad"},
	{0x654296A72165CEF4ULL, "hlac_ht_keep_4519_2x.dds"},
	{0x6542E245CF9D546EULL, "vxlt_backalley_g_1fq_0_0.vlm"},
	{0x65430EF9E883AE15ULL, "gen_im_trs_gen_junk1_01.uti"},
	{0x654367A4F968CBA4ULL, "vxlt_0_lowtown_n_elj_0_1.vlm"},
	{0x65439696E8461A35ULL, "plt_lgt100pt_night.nss"},
	{0x6543F13977EA06EFULL, "w_blightlands_232_0.mao"},
	{0x6543FD3977EA1A8CULL, "w_blightlands_232_0.mmh"},
	{0x654401DF995CD10FULL, "6082204_f.fxe"},
	{0x6544033977EA257EULL, "w_blightlands_232_0.msh"},
	{0x6544EA2BD94F5273ULL, "720152842.ogg"},
	{0x654505D30E7EDCDFULL, "art\\vfx\\models\\vfx_ground_ring.msh"},
	{0x65450BD30E7EE681ULL, "art\\vfx\\models\\vfx_ground_ring.mmh"},
	{0x654517D30E7EFB22ULL, "art\\vfx\\models\\vfx_ground_ring.mao"},
	{0x654572878E1C6624ULL, "vxht_chantry_n_21g_0_0.vlm"},
	{0x6545974A629B5350ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft.evt"},
	{0x6545C5976C966745ULL, "vxca_cavern1_haunted_e7e_0_0.vlm"},
	{0x6545C736AAE0107EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edge.phy"},
	{0x654643F4AE00985BULL, "dae000_bodhan_letter6144323.cl"},
	{0x6546562FC069940CULL, "art\\levels\\ztools\\default\\generic\\_textures\\generic_d.dds"},
	{0x6546F7AB3FD70CE7ULL, "vxlt_gallowstemplar_f_2pd_0_0.vlm"},
	{0x654712ECA681F670ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wmda_0d.dds"},
	{0x65473A18334B2A04ULL, "678462102.ogg"},
	{0x6547A942ACED55F5ULL, "vxlt_0_lowtown_docks_a3_n_d4i_0_0.vlm"},
	{0x6547BB9019700418ULL, "52977183.ogg"},
	{0x6548AD2CB3E43B1CULL, "hlac_lt_gallowscourtyard_-1_2z.dds"},
	{0x6548AE5A38A155DFULL, "540327879.ogg"},
	{0x6549637C4CB7B9E4ULL, "lt_0_lowtown_docks_a3_n_170_0.msh"},
	{0x6549757C4CB7D84DULL, "lt_0_lowtown_docks_a3_n_170_0.mao"},
	{0x6549797C4CB7DE86ULL, "lt_0_lowtown_docks_a3_n_170_0.mmh"},
	{0x65497FE75285C129ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_hunch_en.evt"},
	{0x6549B11A65E8E403ULL, "6129276_m.fxe"},
	{0x6549D4F441D7EEA7ULL, "hlac_lt_0_lowtown_docks_a3_65533_2x.dds"},
	{0x654AF4F8A131FA23ULL, "ht_0_hightown_n_37_0.anb"},
	{0x654B457746B817B0ULL, "hm_cgn_006.mop"},
	{0x654B457746B817B2ULL, "hm_cgn_006.mor"},
	{0x654C33795E699394ULL, "hlac_ca_caverns2_5802_34x.dds"},
	{0x654C3696875514D4ULL, "6103046_m.fxe"},
	{0x654C390AC0796B59ULL, "761878565.ogg"},
	{0x654C4D8F25D172BBULL, "m_rdr370ar_finale.wbd"},
	{0x654C98A4FB787680ULL, "6118424_m.fxe"},
	{0x654D3237AB2B936AULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_c.impa.evt"},
	{0x654D73C91E08F317ULL, "w_sundermount_40_0.mmh"},
	{0x654D77C91E08FA24ULL, "w_sundermount_40_0.mao"},
	{0x654D89C91E091889ULL, "w_sundermount_40_0.msh"},
	{0x654DBB9387D5A37FULL, "vxlt_gallowstemplar_f_5cz_0_0.vlm"},
	{0x654DFC28BE0C5EC3ULL, "vxw_woundedcoast_yq_0_0.vlm"},
	{0x654F37044CA109F5ULL, "445009270.ogg"},
	{0x654FB9FC33BEA551ULL, "213714832.ogg"},
	{0x65506EFD8CBC7EC4ULL, "6089821_m.fxe"},
	{0x6550DB9DFC0FAD89ULL, "6016428_m.fxe"},
	{0x6550E7BCDB43540FULL, "290825551.ogg"},
	{0x6551195492AE6443ULL, "vxw_bonepit_1qg_0_0.vlm"},
	{0x65517D5392F8517EULL, "449764448.ogg"},
	{0x6551813D3606657FULL, "sw_imp_up.mao"},
	{0x6551B71AAC5F6F1CULL, "267988031.ogg"},
	{0x65525D49D7992D68ULL, "gen00fl_isabela6110213.cl"},
	{0x65528F83F5B50F1CULL, "6109765_m.fxe"},
	{0x6552DD32A75386BEULL, "6106083_m.fxe"},
	{0x655385C608340133ULL, "vxlt_foundry_n_yl_0_0.vlm"},
	{0x6553BBEF7C0E2BBCULL, "gen00fl_fenris6126854.cl"},
	{0x65541137AB3179DDULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_c.impa.gad"},
	{0x6554D743857C4BA7ULL, "gen00fl_varric6033634.cl"},
	{0x6554EED120BA38DAULL, "lgt300cr_gang_2_ht.utc"},
	{0x65559A15AC917F5AULL, "mag311_keran6041508.cl"},
	{0x6555B53F6F16D02FULL, "lt_0_lowtown_32_0_ll.mmh"},
	{0x6555D1300FFFDF95ULL, "vxht_mansion2b_1b3_0_0.vlm"},
	{0x6555D5F22E558CCFULL, "487862097.ogg"},
	{0x6556841F701DE3D6ULL, "hlac_ht_keep_n_3732_18x.dds"},
	{0x6556C51772B63FE1ULL, "dae000_gamlen6005821.cl"},
	{0x6557414A2F694500ULL, "160095382.ogg"},
	{0x655767E75291D598ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_hunch_en.ani"},
	{0x655790E25E082333ULL, "445690610.ogg"},
	{0x6557C818A2D25332ULL, "hlac_lt_0_lowtown_n_-1_6x.dds"},
	{0x655804CD1B516580ULL, "art\\characters\\creatures\\qunari\\textures\\qm_har_stn_0d.dds"},
	{0x6558242FA45ABA8EULL, "hlac_lt_0_lowtown_n_65533_1y.dds"},
	{0x65586A82775F2D76ULL, "6050120_m.fxe"},
	{0x655887052EEA44E9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_loop.gad"},
	{0x655920028103E156ULL, "141951054.ogg"},
	{0x6559821D80086FD7ULL, "6171273_m.fxe"},
	{0x6559E83262FB17EDULL, "vxw_woundedcoast_3m_0_0.vlm"},
	{0x655A68D1D698DB35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_thin.msh"},
	{0x655A6AD1D698DEABULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_thin.mmh"},
	{0x655A76D1D698F348ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_thin.mao"},
	{0x655A7745E28C40CCULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_grindstone.phy"},
	{0x655A7D415229037DULL, "vxlt_0_lowtown_e85_0_0.vlm"},
	{0x655AA3B344B75BC5ULL, "vxlt_0_lowtown_a3_n_c_hd8_0_0.vlm"},
	{0x655AE775B6E47597ULL, "835052574.ogg"},
	{0x655B008B8C54AE43ULL, "6076244_m.fxe"},
	{0x655B34F7AC0EB00EULL, "art\\characters\\creatures\\qunari\\textures\\qm_eye_mag_0d.dds"},
	{0x655B8CBD38D1F738ULL, "128897797.ogg"},
	{0x655C503AD9C2436AULL, "6118586_f.fxe"},
	{0x655C5F18DEAA591BULL, "11931703.ogg"},
	{0x655C6E05B0027743ULL, "vxw_woundedcoast_vp_0_0.vlm"},
	{0x655D079DDA494142ULL, "llac_w_sundermount_3_0.dds"},
	{0x655D225103056982ULL, "vxlt_gallowstemplar_n_5rp_0_0.vlm"},
	{0x655D29FA6CEA5733ULL, "pro000_friedrich6050194.cl"},
	{0x655D5511CE79E387ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.dura.gad"},
	{0x655DBD8CFE6C66E1ULL, "art\\characters\\weapons\\longswords\\w_lsw_005a.phy"},
	{0x655DC8DCE89B9FC6ULL, "hlac_ca_caverns1_mines_24427_1z.dds"},
	{0x655EAF6D3B5726E0ULL, "vxht_0_hightown_a3_bc3_0_0.vlm"},
	{0x655EFAD68655BDBBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp21_arch.msh"},
	{0x655F0CD68655DC06ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp21_arch.mao"},
	{0x655F18D68655F0A5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp21_arch.mmh"},
	{0x655F33F487C69FB7ULL, "vxlt_0_lowtown_docks_n_bfl_0_0.vlm"},
	{0x655F425818932971ULL, "static_static_p37.psh.fxo.dx11.assembly"},
	{0x655F58052EF01337ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_loop.ani"},
	{0x655F8B3184A47CC3ULL, "vxw_bonepit_1zf_0_0.vlm"},
	{0x655F968EF93BF1A6ULL, "875009597.ogg"},
	{0x65604CAC7D7C7D40ULL, "6018531_m.fxe"},
	{0x6560895C37569BACULL, "hlac_lt_0_lowtown_n_65534_4y.dds"},
	{0x6560B3C068448836ULL, "6205970_f.fxe"},
	{0x6561D478F7819728ULL, "vxht_bartrand_n_21x_0_0.vlm"},
	{0x6561F7A2FA2A19B3ULL, "vxlt_foundry_n_to_0_0.vlm"},
	{0x656210AA48A5D6BEULL, "6162607_m.fxe"},
	{0x656224B28F2A45FFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_right.phy"},
	{0x6562C53C86FFF83CULL, "w_bonepit_7_0.mao"},
	{0x6562C793154352F9ULL, "vxlt_gallowstemplar_6qj_0_0.vlm"},
	{0x6562D13C87000CDFULL, "w_bonepit_7_0.mmh"},
	{0x6562D73C87001681ULL, "w_bonepit_7_0.msh"},
	{0x6562F08BD3123824ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow1win.phy"},
	{0x6562FCF11BD5DE73ULL, "164023267.ogg"},
	{0x6563237C93CE2540ULL, "6129768_f.fxe"},
	{0x65634847CAB8E8AAULL, "hlac_dr_0_deeproads_5849_5y.dds"},
	{0x6563E32C0433385EULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingb_n.dds"},
	{0x6563EF3E4EE14CC8ULL, "one101st_kelder.stg"},
	{0x65644A11CE7FEF94ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.dura.evt"},
	{0x6564C85A755DFDE3ULL, "art\\vfx\\creatures\\vfx_c_hd_tswipe_c.impa.ani"},
	{0x65656DE9116D4789ULL, "pro000cs_prologueend.ncs"},
	{0x65659596E862FE96ULL, "plt_lgt100pt_night.plo"},
	{0x6566D2EE4B899BE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_90.evt"},
	{0x65671B44AFC70C7EULL, "framebuffereffects_hslmatrix_p0.psh.fxo.dx11"},
	{0x65674513366A4795ULL, "6129112_f.fxe"},
	{0x6567AE696E9CF9CAULL, "hlac_lt_0_lowtown_a2_n_65536_2y.dds"},
	{0x65683325BDF7F69AULL, "vxw_sundermount_2yb_0_0.vlm"},
	{0x65683F6638CFA086ULL, "361771850.ogg"},
	{0x65686240A7126B96ULL, "vxlt_hangedman_n_1tw_0_0.vlm"},
	{0x6568E49A6683E42DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12a.msh"},
	{0x6568EF1E441F1BFDULL, "vxlt_0_lowtown_e34_0_0.vlm"},
	{0x6568F1962A805FE4ULL, "vxdr_primevalentrance_2lj_0_0.vlm"},
	{0x6568F69A668402B3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos12a.mmh"},
	{0x65696ECF31BEAEE8ULL, "dae000ip_crowd_mf5_talk1.utp"},
	{0x656973844376BA2CULL, "vxht_mansion_n_6oq_0_0.vlm"},
	{0x6569AEF511C4765DULL, "mcr370_cut_player_survives6098278.cl"},
	{0x656A10479FAFF8BFULL, "vxlt_warehouse_n_1dv_0_0.vlm"},
	{0x656A4EA55BF58F3DULL, "780287988.ogg"},
	{0x656A874CC7883066ULL, "6072829_m.fxe"},
	{0x656ACF06D9C993CDULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_smoke02.dds"},
	{0x656ADFE2A1F88FC3ULL, "vxw_woundedcoast_ss_0_0.vlm"},
	{0x656B218D2BD22F54ULL, "6051643_f.fxe"},
	{0x656B942DF4FB8202ULL, "vxlt_gallowstemplar_n_5uq_0_0.vlm"},
	{0x656BB87AB70271ABULL, "hlac_lt_gallowstemplar_2394_12x.dds"},
	{0x656C5F889FAF0822ULL, "ca_caverns1.arl"},
	{0x656D033E032F23FEULL, "6061854_m.fxe"},
	{0x656D184AD2511478ULL, "vxht_0_hightown_a3_bzt_0_0.vlm"},
	{0x656D37CD5D90B263ULL, "6147113_m.fxe"},
	{0x656D62D9665849CFULL, "6201525_m.fxe"},
	{0x656D64B52A3DE6A6ULL, "6187952_m.fxe"},
	{0x656D8E71CB437BE8ULL, "6061027_m.fxe"},
	{0x656DA1290093D249ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrock_n.dds"},
	{0x656DA5D179BCB837ULL, "vxlt_0_lowtown_docks_n_bco_0_0.vlm"},
	{0x656DE9EE4B8FE131ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_90.gad"},
	{0x656F75016FFF9337ULL, "hlac_lt_undercitydungeon_-1_0z.dds"},
	{0x656F7FE71734E762ULL, "zz_approval_debug6133645.cl"},
	{0x657014582D615AC0ULL, "gencr_shade.utc"},
	{0x657075A39881D02BULL, "vxlt_hangedman_1te_0_0.vlm"},
	{0x6570AF93FBF9D275ULL, "art\\vfx\\combat\\vfx_str_att_front_l_leg_c.anb"},
	{0x6570CEBA017C9892ULL, "104886012.ogg"},
	{0x6570D7F7D139437CULL, "6173399_m.fxe"},
	{0x6570F925C712B123ULL, "vxht_keep_n_3hf_0_0.vlm"},
	{0x65713B827F09814BULL, "866489985.ogg"},
	{0x657148172211DB74ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowb_d.dds"},
	{0x6571D5D7628BE40BULL, "6096750_m.fxe"},
	{0x6571E05A7568B0DAULL, "art\\vfx\\creatures\\vfx_c_hd_tswipe_c.impa.evt"},
	{0x65724211CE8C1FCDULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.dura.ani"},
	{0x6572B4E9F3EC1095ULL, "vxht_mansion2b_1x1_0_0.vlm"},
	{0x6572DA8250EDE8BCULL, "hlac_dr_primevaltomb_658_0y.dds"},
	{0x657325B28F38BA04ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_right.mmh"},
	{0x6573266974C94D6FULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml106a_0t.dds"},
	{0x657329B28F38C157ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_right.mao"},
	{0x6573346910CDDA2BULL, "vxht_mansion_n_5mg_0_0.vlm"},
	{0x65733BB28F38DFA6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_right.msh"},
	{0x65736E52A813E960ULL, "6131627_m.fxe"},
	{0x65738850E532066EULL, "llac_dr_0_deeproads_5842_7.dds"},
	{0x6573F0A4A99211BAULL, "llac_ht_0_hightown_-196608_0.dds"},
	{0x657457C6D5D809D0ULL, "art\\characters\\playercharacter\\humanmale\\boo_wic.msh"},
	{0x657465C6D5D821EAULL, "art\\characters\\playercharacter\\humanmale\\boo_wic.mmh"},
	{0x657469C6D5D82821ULL, "art\\characters\\playercharacter\\humanmale\\boo_wic.mao"},
	{0x6574C9059E2FA01FULL, "vxlt_0_lowtown_h7w_0_0.vlm"},
	{0x6574EAEE4B9601DFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_90.ani"},
	{0x6574FE91B406E5E3ULL, "mag100_karras6034546.cl"},
	{0x6576294E6684ED04ULL, "6101523_m.fxe"},
	{0x6576D41D99088416ULL, "vxlt_hangedman_n_1qx_0_0.vlm"},
	{0x6576E9F434E7F8B6ULL, "def_cobble_d.dds"},
	{0x6576EBCFA12ACE6FULL, "vxlt_gallowsprison_n_6px_0_0.vlm"},
	{0x6576FD45E03661E2ULL, "40319887.ogg"},
	{0x6577891FDE22CEBBULL, "mag210_arianni6024565.cl"},
	{0x65782F4DC144AAC5ULL, "6140150_m.fxe"},
	{0x65786ACF3ED57596ULL, "rdr111ip_crate_cargo.ncs"},
	{0x657889AD0CB62298ULL, "vxlt_0_lowtown_n_e25_0_0.vlm"},
	{0x6578DF5A756ECDCDULL, "art\\vfx\\creatures\\vfx_c_hd_tswipe_c.impa.gad"},
	{0x65790C0C52C07153ULL, "vxdr_0_deeproads_cxb_0_0.vlm"},
	{0x65795FDBAEE72A44ULL, "vxw_woundedcoast_1b2_0_0.vlm"},
	{0x657963F20CFF03ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_keepwall_d.dds"},
	{0x657985D7EF89BCB4ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_roots_n.dds"},
	{0x6579CFEAD63B2DA3ULL, "49524953.ogg"},
	{0x657A057DF4315A61ULL, "vxlt_hangedman_n_27b_0_0.vlm"},
	{0x657A7C2E0CA28CB5ULL, "222628285.ogg"},
	{0x657AB4FC205544CEULL, "gen_im_trs_dem_sta_fir_r1.uti"},
	{0x657AFC8D6C69085BULL, "hlac_ca_caverns2_5802_8x.dds"},
	{0x657BC0969887F52DULL, "hlac_lt_gallowstemplar_f_2384_4y.dds"},
	{0x657C2A0F0CFADBB5ULL, "fx_1018_diff.dds"},
	{0x657CC2BDC5ABEEE3ULL, "hlac_ht_0_hightown_n_-65537_3y.dds"},
	{0x657D5D686148DEE0ULL, "6101185_m.fxe"},
	{0x657D9AE04DB1680CULL, "6084398_m.fxe"},
	{0x657E0E750051065AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_2rubblelarge.msh"},
	{0x657E1C7500511D93ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_2rubblelarge.mao"},
	{0x657E2075005124A0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_2rubblelarge.mmh"},
	{0x657E47E41DB180B2ULL, "hlac_lt_0_lowtown_a3_n_c_65537_4y.dds"},
	{0x657E57ED79B596EDULL, "vxca_caverns1_e6s_0_0.vlm"},
	{0x657E5D88C996C82FULL, "vxht_0_hightown_a3_cak_0_0.vlm"},
	{0x657EEB8E2F2FB184ULL, "hlac_lt_gallowstemplar_f_2383_12z.dds"},
	{0x657F0904DB78BEF6ULL, "gen_im_cft_reg_k1.uti"},
	{0x657F2C228FE52C95ULL, "mag311_keran6041484.cl"},
	{0x657FE3F6A19DFCEBULL, "vxlt_gallowscourtyard_n_76i_0_0.vlm"},
	{0x6580303D905AAF3FULL, "follower_barks6153413.cl"},
	{0x658062FDFF33969AULL, "mag321_nella.cnv"},
	{0x6580941A5416D681ULL, "916966280.ogg"},
	{0x6580BEE861F70571ULL, "6091393_m.fxe"},
	{0x6580F04BB7FF2342ULL, "vxht_bartrand_n_6p9_0_0.vlm"},
	{0x6580F1BF8AA90C50ULL, "rog_impactmode.dds"},
	{0x65811988BC7D20C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_lhand.ani"},
	{0x658126A7799B0563ULL, "vxw_sundermount_2cm_0_0.vlm"},
	{0x658126C6E5E22915ULL, "vxht_mansion2b_1s0_0_0.vlm"},
	{0x6581C3E0CEB1A50AULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.dura.gad"},
	{0x6581D17441C818CBULL, "vxdr_slavers_1ay_0_0.vlm"},
	{0x6582B03EA481D617ULL, "6014998_m.fxe"},
	{0x6582D58728538FE0ULL, "atl_guiachieve_dxt5_dat.xml"},
	{0x6583012EAF25C7E0ULL, "hlac_lt_0_lowtown_a2_n_c_-2_0y.dds"},
	{0x65832F093AF904D5ULL, "913641742.ogg"},
	{0x658384FDFF3606E2ULL, "mag321_nella.bnk"},
	{0x65838E33B5860648ULL, "vxlt_hangedman_n_27g_0_0.vlm"},
	{0x6583D64B62FE54BAULL, "plt_vrc100pt_n_initial.nss"},
	{0x6584AB0BDC690EC2ULL, "hlac_ca_caverns2prc_5802_29y.dds"},
	{0x6585585B3513D01AULL, "t3_hey_te1.tnt"},
	{0x6585A2CC8A3EDCC4ULL, "ave200_donnic6110703.cl"},
	{0x6586159F37EDE8A6ULL, "6056922_m.fxe"},
	{0x65868FD62714EC42ULL, "mag311_nyssa6042533.cl"},
	{0x6586AB2D970AC67BULL, "vxca_caverns1_j40_0_0.vlm"},
	{0x6586E9B4F03FCCB6ULL, "llac_lt_darktown_-1_80.dds"},
	{0x65871FAF4D31147DULL, "vxca_caverns1_i6o_0_0.vlm"},
	{0x658757AC1BB9CFBCULL, "vxw_encounter2_6v_0_0.vlm"},
	{0x65877DE944B689D3ULL, "vxdr_0_deeproads_cue_0_0.vlm"},
	{0x65878B7709BC83C0ULL, "212908031.ogg"},
	{0x6587C80F3AAB2059ULL, "dae000_flemeth6000917.cl"},
	{0x6587D1B8A0DD42C4ULL, "vxw_woundedcoast_1m3_0_0.vlm"},
	{0x65881688BC833A26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_lhand.gad"},
	{0x65886CB1A33006C4ULL, "vxw_bonepit_1xb_0_1.vlm"},
	{0x6588775AE62772E1ULL, "vxlt_hangedman_n_22a_0_0.vlm"},
	{0x658877E7D8E7B302ULL, "vxlt_gallowstemplar_n_5cs_0_0.vlm"},
	{0x6588903B700995B2ULL, "vxht_0_hightown_a2_c3x_0_1.vlm"},
	{0x6588983EA73A26AFULL, "zz_mcr_debug6080826.cl"},
	{0x6588A6E0CEB792A1ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.dura.evt"},
	{0x65893E59DB320BA8ULL, "6067069_m.fxe"},
	{0x65896CCDCE4ABFB6ULL, "zz_mer_debug6045635.cl"},
	{0x658982CFDEF4A182ULL, "gen00fl_fenris6127015.cl"},
	{0x65899A7E1A3BC40AULL, "weapontrail_streaks.dds"},
	{0x6589DF12F6CFBEC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_goodbye_wesley.gad"},
	{0x658A6E06EB57B15EULL, "vxlt_0_lowtown_a3_n_c_ff7_0_0.vlm"},
	{0x658B09FE81378B6AULL, "880850191.ogg"},
	{0x658B7F3F6CF01988ULL, "mcr310_confused_guard6077037.cl"},
	{0x658BABFF41142193ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_act_c.impa.evt"},
	{0x658C280614708072ULL, "and111_lirene6084512.cl"},
	{0x658CCDD433A3F008ULL, "171709322.ogg"},
	{0x658D0B7A84A5DA9AULL, "hlac_ht_0_hightown_65536_1y.dds"},
	{0x658DE6D0C6CBD2DEULL, "gen00fl_anders6119430.cl"},
	{0x658ED18CD4F66AE3ULL, "hlac_ht_0_hightown_131071_1z.dds"},
	{0x658F1188BC895005ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_lhand.evt"},
	{0x658F17758DFBDC23ULL, "vxlt_0_lowtown_a2_n_c_i4f_0_0.vlm"},
	{0x658F1FFFDBC4179AULL, "hlac_ht_0_hightown_a3_131071_1x.dds"},
	{0x658F38B4873DFB91ULL, "ca_ciga_smo.mao"},
	{0x658F46029D122D27ULL, "dae000_meredith6085375.cl"},
	{0x658F7B2A928D8611ULL, "6100284_m.fxe"},
	{0x658F98846B911DE3ULL, "vxw_sundermount_2dl_0_0.vlm"},
	{0x6590342726EF7D77ULL, "6210350_m.fxe"},
	{0x6590B0A970E4C581ULL, "359180137.ogg"},
	{0x6590B49F68E742EEULL, "hlac_lt_0_lowtown_-3_1x.dds"},
	{0x65910212F6D618B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_goodbye_wesley.evt"},
	{0x65917315EC6D2073ULL, "dae300_cullen6133849.cl"},
	{0x6591A4FF411881D4ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_act_c.impa.gad"},
	{0x6591EB3DD8B087F1ULL, "178347614.ogg"},
	{0x65920010A77C1EC8ULL, "vxlt_hangedman_n_22d_0_0.vlm"},
	{0x65921BF8196E342AULL, "vxlt_0_lowtown_docks_n_cvi_0_1.vlm"},
	{0x65922D368AEE0938ULL, "hlac_lt_gallowscourtyard_131071_1x.dds"},
	{0x6592CA11B81B9980ULL, "hlac_lt_0_lowtown_n_131072_5y.dds"},
	{0x6593370E804B9145ULL, "vxca_caverns2prc_a5p_0_0.vlm"},
	{0x659344FDD33B8D1AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglebody_d.dds"},
	{0x6593826982802CD9ULL, "6016486_m.fxe"},
	{0x659468D9055C3563ULL, "6053498_m.fxe"},
	{0x65946CA68DE1E75DULL, "hlac_lt_darktown_-1_27z.dds"},
	{0x6594D227209224DDULL, "vxht_0_hightown_a3_c8f_0_0.vlm"},
	{0x6594E452173278E0ULL, "513189658.ogg"},
	{0x6594FDE9B19A04D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants3a.msh"},
	{0x659507E9B19A15D4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants3a.mmh"},
	{0x65950BE9B19A1CE7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants3a.mao"},
	{0x659513A3B39E4786ULL, "gen_im_arm_boo_hvy_set_med.uti"},
	{0x659558A36AD496A0ULL, "llac_lt_0_lowtown_docks_n_0_5.dds"},
	{0x659559EFF6021257ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgtp_0t.dds"},
	{0x65959EE0CEC20FF0ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.dura.ani"},
	{0x6595C13C3495A36EULL, "hlac_ca_caverns1_mines_24424_26z.dds"},
	{0x6595C3F9B5D3A88FULL, "follower_banter6161154.cl"},
	{0x6595CEB0AF417E34ULL, "mag110_samson6053555.cl"},
	{0x6595CFB248655B19ULL, "hlac_lt_0_lowtown_a2_-65536_0z.dds"},
	{0x6595DBE176C68F1DULL, "zz_one_debug6099752.cl"},
	{0x6595DE112836C6B4ULL, "197221921.ogg"},
	{0x6595EFC636ACA253ULL, "vxdr_0_deeproads_crd_0_0.vlm"},
	{0x6596439592D35B44ULL, "vxw_woundedcoast_1h0_0_0.vlm"},
	{0x659657C463C3BFEEULL, "865451304.ogg"},
	{0x659659309FE330DDULL, "gen_im_trs_qun_amu_r1.uti"},
	{0x6597D5BA2E066820ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comc_0n.dds"},
	{0x6597F3CC167308F2ULL, "and110_refugee6111594.cl"},
	{0x6598451020614CC6ULL, "vxlt_undercityserialkiller_zv_0_0.vlm"},
	{0x65985383C5FC16F5ULL, "755411451.ogg"},
	{0x6598B3FF411EB9B2ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_act_c.impa.ani"},
	{0x6598FB684F9F01B7ULL, "vxlt_0_lowtown_docks_n_brh_0_0.vlm"},
	{0x6599297227B6429CULL, "vxlt_backalley_n_1we_0_0.vlm"},
	{0x65993A30BBC07CD5ULL, "6088684_f.fxe"},
	{0x6599AE5FBCA66850ULL, "vxw_encounter3_e8_0_0.vlm"},
	{0x6599B093FC1CD8B6ULL, "art\\vfx\\combat\\vfx_str_att_front_l_leg_c.mmh"},
	{0x6599B25E857EF531ULL, "6125829_m.fxe"},
	{0x659A4E45E2C1DE97ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_grindstone.mmh"},
	{0x659A5245E2C1E5A4ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_grindstone.mao"},
	{0x659A5D35173C0F53ULL, "9448015.ogg"},
	{0x659A6445E2C20409ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_grindstone.msh"},
	{0x659A65630B148F3FULL, "immortals.ncs"},
	{0x659AA76284859E1AULL, "art\\characters\\weapons\\longswords\\w_lsw_001a.mmh"},
	{0x659AAB628485A551ULL, "art\\characters\\weapons\\longswords\\w_lsw_001a.mao"},
	{0x659AB9628485BD60ULL, "art\\characters\\weapons\\longswords\\w_lsw_001a.msh"},
	{0x659B65AA74C06474ULL, "6037315_f.fxe"},
	{0x659B92674324E6C0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wmda_0t.dds"},
	{0x659C10FA896E1C9CULL, "579494196.ogg"},
	{0x659C29249CDA1766ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.dura.evt"},
	{0x659C4D8A07ADED57ULL, "vxht_0_hightown_a3_c3i_0_0.vlm"},
	{0x659CBB75DA188136ULL, "vxht_mansion2b_1hv_0_0.vlm"},
	{0x659CE852AF65B18BULL, "25164872.ogg"},
	{0x659D8DB0854DC425ULL, "711157821.ogg"},
	{0x659D96424DE37BF5ULL, "dae300_bran6142865.cl"},
	{0x659DDED0D7231F5AULL, "vxlt_0_lowtown_a2_n_gcn_0_0.vlm"},
	{0x659E0A12F6E0B0A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_goodbye_wesley.ani"},
	{0x659E0A615D873663ULL, "vxw_sundermount_2io_0_0.vlm"},
	{0x659E0CB69F71F76BULL, "236439468.ogg"},
	{0x659E75CF3EF657A3ULL, "rdr111ip_crate_cargo.utp"},
	{0x659ED661244CE269ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw101a.mao"},
	{0x659F020039DFB29DULL, "t1_skn_e08.tnt"},
	{0x659F047418777EC8ULL, "vxht_bartrand_n_4dm_0_0.vlm"},
	{0x659F707BF80140D2ULL, "vxdr_0_deeproads_cra_0_0.vlm"},
	{0x659FB297B048211BULL, "plt_cod_ltr_gamlen_charade.nss"},
	{0x659FC40E24F73766ULL, "vxw_bonepitdestroyed_a3_o7_0_0.vlm"},
	{0x65A0124EA03A592FULL, "vxlt_gallowstemplar_n_5j9_0_0.vlm"},
	{0x65A01EB6D357AA1BULL, "259940853.ogg"},
	{0x65A04FAB94CDA115ULL, "6082696_m.fxe"},
	{0x65A05AA84AE155BBULL, "6143812_m.fxe"},
	{0x65A19D1951434E79ULL, "202565843.ogg"},
	{0x65A1A871469AEAD0ULL, "vxlt_gallowstemplar_n_474_0_0.vlm"},
	{0x65A1B42C3CC1C560ULL, "llac_ht_chantry_2534_7.dds"},
	{0x65A2172737AAA229ULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_c.impa.ani"},
	{0x65A22CDB4B7C7A8BULL, "583759684.ogg"},
	{0x65A26C6AB2231D77ULL, "dae000ip_crowd_mf6_talk9.utp"},
	{0x65A31A36C42CDDDFULL, "6096807_f.fxe"},
	{0x65A340249CE05CB1ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.dura.gad"},
	{0x65A34589ABD1B720ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_2y.dds"},
	{0x65A37939883B6785ULL, "321916834.ogg"},
	{0x65A3BAB7F409A241ULL, "6174399_m.fxe"},
	{0x65A4426310FC5C7BULL, "704922605.ogg"},
	{0x65A4B57284C973C4ULL, "vxw_woundedcoast_1s1_0_0.vlm"},
	{0x65A4E1FA26997284ULL, "243653353.ogg"},
	{0x65A586CF17CDBFCAULL, "hla_ht_0_hightown_a3_-65534_0y.dds"},
	{0x65A59023D70F974DULL, "522121374.ogg"},
	{0x65A5A3B09BA79022ULL, "one100dg_kelder6175890.cl"},
	{0x65A5BD36F58032DDULL, "534561146.ogg"},
	{0x65A5F1D68691B152ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp21_arch.phy"},
	{0x65A5FFD74049E900ULL, "47973804.ogg"},
	{0x65A64720C63EDD82ULL, "vxdr_0_deeproads_d35_0_0.vlm"},
	{0x65A6DFBE9A297678ULL, "vxht_0_hightown_a3_bfp_0_0.vlm"},
	{0x65A70DE47A6ACB6AULL, "6077065_f.fxe"},
	{0x65A90172879B8CF9ULL, "vxlt_0_lowtown_docks_n_aor_0_0.vlm"},
	{0x65A91E45DD4D5527ULL, "6065019_m.fxe"},
	{0x65A92C76E57EDBF8ULL, "332763994.ogg"},
	{0x65A9322737B0EEABULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_c.impa.gad"},
	{0x65A99FCF75A43AF6ULL, "vxlt_warehouse_12n_0_0.vlm"},
	{0x65A9C0366807463FULL, "695091041.ogg"},
	{0x65AA26E25727F8F5ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_dream_n.dds"},
	{0x65AA41249CE67D5FULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.dura.ani"},
	{0x65AA5546E079569CULL, "vxw_blightlands_142_0_0.vlm"},
	{0x65AA8713CCC1E9E9ULL, "6085929_m.fxe"},
	{0x65AAA2E8A55E5498ULL, "hlac_lt_gallowstemplar_n_2390_0y.dds"},
	{0x65AABF66F9A405D7ULL, "vxht_0_hightown_a3_c4h_0_0.vlm"},
	{0x65AAC0C2907608D6ULL, "vxlt_backalley_n_1u9_0_0.vlm"},
	{0x65AACF642DD04901ULL, "hlac_lt_0_lowtown_-65538_0z.dds"},
	{0x65ABC84961924165ULL, "vxlt_0_lowtown_a2_n_gvv_0_0.vlm"},
	{0x65ABFF54528444CEULL, "hlac_w_encounter2_327684_0x.dds"},
	{0x65AC84E4BC676846ULL, "554670294.ogg"},
	{0x65AD5445DCC052D6ULL, "hlac_ht_chantry_2534_7x.dds"},
	{0x65AD73FE576A043FULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_512x256occ.msh"},
	{0x65AD79FE576A0EE1ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_512x256occ.mmh"},
	{0x65AD7ABA57430456ULL, "vxw_blightlands_13a_0_0.vlm"},
	{0x65AD85FE576A2282ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_512x256occ.mao"},
	{0x65AE9D4AA16BFA63ULL, "mag161ip_loot_1.utp"},
	{0x65AE9F4DDE966C11ULL, "618825531.ogg"},
	{0x65AF2F2737B55510ULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_c.impa.evt"},
	{0x65AF353DDB15BF4EULL, "lla_ht_0_hightown_n_65536_0.dds"},
	{0x65AFD4A26187408EULL, "345647577.ogg"},
	{0x65B0296B4A103AF3ULL, "hlac_ca_caverns2_5802_14y.dds"},
	{0x65B03AFBD611A867ULL, "zz_dae_debug6029889.cl"},
	{0x65B047E857CF384CULL, "6059323_m.fxe"},
	{0x65B094FE19B249A1ULL, "qun221_petrice_cleric_2.stg"},
	{0x65B0C7A6077E91A2ULL, "mer341_jansen6082981.cl"},
	{0x65B0C93524DFA777ULL, "vxw_bonepitmining_a2_1k7_0_0.vlm"},
	{0x65B16F3C1190C029ULL, "hlac_lt_0_lowtown_a2_65535_3y.dds"},
	{0x65B1F2013124F61AULL, "hlac_lt_0_lowtown_docks_262143_0z.dds"},
	{0x65B203D0966145FFULL, "dae000_fem_commoner2_mob6174587.cl"},
	{0x65B233D88F506118ULL, "vxht_keep_n_7af_0_0.vlm"},
	{0x65B271603E724AA9ULL, "602547915.ogg"},
	{0x65B276B742590E38ULL, "gen00fl_anders6111146.cl"},
	{0x65B28911A355EF52ULL, "mag370_meredith6046882.cl"},
	{0x65B33E30CDA4C2E0ULL, "vxca_caverns2prc_9pv_0_0.vlm"},
	{0x65B4C04E20C2C02FULL, "6037100_m.fxe"},
	{0x65B4D2C73D2470CBULL, "6147676_m.fxe"},
	{0x65B4EC8307A39D05ULL, "893453298.ogg"},
	{0x65B51254A84DD330ULL, "6126454_m.fxe"},
	{0x65B514CD6337031EULL, "gen00fl_merrill6108690.cl"},
	{0x65B51BF412F674F7ULL, "81805649.ogg"},
	{0x65B61DEA9FACADD5ULL, "6086884_f.fxe"},
	{0x65B62E62E622DCDAULL, "mer220_hubert6065084.cl"},
	{0x65B635105935D439ULL, "6094985_m.fxe"},
	{0x65B69BB33CF33861ULL, "323926523.ogg"},
	{0x65B6A5E286E54EB4ULL, "vrc321_varric6084292.cl"},
	{0x65B6E9E73AD4BDCEULL, "6179914_f.fxe"},
	{0x65B76E42A91D0435ULL, "6086173_m.fxe"},
	{0x65B7915BE9E78866ULL, "hlac_lt_shop_-1_2z.dds"},
	{0x65B7F0D9E935AA5EULL, "343879253.ogg"},
	{0x65B8AEF452504AABULL, "vxlt_hangedman_1ox_0_0.vlm"},
	{0x65B8D7BBF20D9A9AULL, "757153.ogg"},
	{0x65B8DACECF6B9B95ULL, "6124748_f.fxe"},
	{0x65B95C565C9E370DULL, "vfx0_addv.psh.fxo"},
	{0x65B975455C961AE0ULL, "zz_followers_banter6158751.cl"},
	{0x65B97CD6A8604B15ULL, "190460269.ogg"},
	{0x65BA103F19227B8EULL, "vxht_bartrand_n_66g_0_0.vlm"},
	{0x65BA729B343E0750ULL, "187746618.ogg"},
	{0x65BA82BB55A874FEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_hangtarp_dp.dds"},
	{0x65BA93E8F3C3A5B4ULL, "ser121_jethann6107100.cl"},
	{0x65BAB3280E1901B2ULL, "772494106.ogg"},
	{0x65BC10704F4DEDA1ULL, "64123997.ogg"},
	{0x65BC4B5A5A795360ULL, "187477255.ogg"},
	{0x65BC4FBF07529C9EULL, "vxlt_0_lowtown_a2_n_gn1_0_0.vlm"},
	{0x65BC51A5BB6A9A93ULL, "vxw_sundermount_34l_0_1.vlm"},
	{0x65BCB16E4570F133ULL, "mag370_meredith6172386.cl"},
	{0x65BCE2501D0CBF32ULL, "140181454.ogg"},
	{0x65BCE5B105D100DCULL, "62120348.ogg"},
	{0x65BD6659F9085E43ULL, "6080263_m.fxe"},
	{0x65BDAE73A41F1150ULL, "vxht_0_hightown_n_by3_0_0.vlm"},
	{0x65BE54D6AC2D4A8FULL, "mrl310_merrill6144043.cl"},
	{0x65BE617F4BA3CB42ULL, "6094484_f.fxe"},
	{0x65BEC892AFE0F078ULL, "854002610.ogg"},
	{0x65BF6DE7362FEF82ULL, "isa211_isabela6072549.cl"},
	{0x65BF7D6A57E168A6ULL, "mag170_thrask6054057.cl"},
	{0x65BFC66284A4EF25ULL, "art\\characters\\weapons\\longswords\\w_lsw_001a.phy"},
	{0x65BFEFB2B3D5C544ULL, "407530115.ogg"},
	{0x65C06B33F50D744FULL, "follower_banter6163978.cl"},
	{0x65C0875D22B21DDAULL, "413351063.ogg"},
	{0x65C0C28CFEC0EC1DULL, "art\\characters\\weapons\\longswords\\w_lsw_005a.mao"},
	{0x65C0C68CFEC0F356ULL, "art\\characters\\weapons\\longswords\\w_lsw_005a.mmh"},
	{0x65C0C6F227F098BDULL, "lgt310_mis_burn_murder16152753.cl"},
	{0x65C0C8CCC1BCFDEEULL, "vxw_bonepitmining_a2_25g_0_0.vlm"},
	{0x65C0D08CFEC10454ULL, "art\\characters\\weapons\\longswords\\w_lsw_005a.msh"},
	{0x65C0D1552D051C61ULL, "vxht_bartrand_n_4sm_0_0.vlm"},
	{0x65C0E62D1B739076ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleftcrn_1.mao"},
	{0x65C116F197077BE5ULL, "vxht_mansion2b_1vo_0_0.vlm"},
	{0x65C11DB47DE30352ULL, "47769285.ogg"},
	{0x65C14FBD164FDEE5ULL, "739781938.ogg"},
	{0x65C159681CE2B2ECULL, "vxlt_warehouse_n_11v_0_0.vlm"},
	{0x65C1747F7AFCEEEFULL, "555622868.ogg"},
	{0x65C29B7AD17F92B1ULL, "6132667_m.fxe"},
	{0x65C2D77674B50D8BULL, "gen_im_trs_slv_junk1_02.uti"},
	{0x65C3462812272B7EULL, "160416442.ogg"},
	{0x65C3B0A00F019380ULL, "6118657_m.fxe"},
	{0x65C3E62A4E51E23FULL, "llac_lt_0_lowtown_a2_n_-2_4.dds"},
	{0x65C3F485660E6B30ULL, "plt_cod_prc_invitations.plo"},
	{0x65C45D0BDEC13947ULL, "dae320ip_chant_board_ht.utp"},
	{0x65C47F08FD988C1CULL, "vxlt_backalley_n_1xx_0_0.vlm"},
	{0x65C4A111413C80F6ULL, "558877994.ogg"},
	{0x65C4C39EDE1EC334ULL, "905171788.ogg"},
	{0x65C4DEE17F13FEAEULL, "lgt000_chanter6147405.cl"},
	{0x65C4FC8148C98DE7ULL, "gen00fl_varric6081587.cl"},
	{0x65C59B2107B294EEULL, "art\\vfx\\combat\\vfx_w_wns_right_close1_c.mmh"},
	{0x65C5C92883561B17ULL, "gen00fl_fenris6108095.cl"},
	{0x65C6005E7C2DFEDFULL, "vxlt_0_lowtown_docks_a3_aoh_0_0.vlm"},
	{0x65C615F2914C7F83ULL, "vxlt_darktown_4ij_0_0.vlm"},
	{0x65C6B7A220C7528CULL, "754994143.ogg"},
	{0x65C6DDD8F1D73721ULL, "gen_im_wep_war_one_act3_05.uti"},
	{0x65C739A7A374D36EULL, "art\\characters\\creatures\\dragon\\animation\\p_rst.evt"},
	{0x65C78504186FC914ULL, "gen_im_arm_boo_hvy_set_high.uti"},
	{0x65C7865765D97B29ULL, "vxlt_0_lowtown_docks_a3_d0z_0_0.vlm"},
	{0x65C8821C0B18940EULL, "vxht_bartrand_n_63x_0_0.vlm"},
	{0x65C893EB7F0944C7ULL, "vxw_woundedcoast_ym_0_0.vlm"},
	{0x65C8E08C44148B40ULL, "dae000_orsino6081361.cl"},
	{0x65C8F0731DDE583EULL, "6044824_m.fxe"},
	{0x65C91711808B70C3ULL, "vxca_caverns1_ebq_0_0.vlm"},
	{0x65C95FF833697FE3ULL, "vxw_sundermount_2xp_0_0.vlm"},
	{0x65C96492FF73A256ULL, "6155487_m.fxe"},
	{0x65CA4FD232B9FAE7ULL, "abi_w_stun.bnk"},
	{0x65CA5A0AEE59C848ULL, "vxht_bartrand_n_4sh_0_0.vlm"},
	{0x65CAA39846B2552EULL, "449237656.ogg"},
	{0x65CAC382AD60B313ULL, "vxw_sundermount_33m_0_1.vlm"},
	{0x65CB1F331D3D7290ULL, "hlac_lt_backalley_g_0_16y.dds"},
	{0x65CBC4050D4FB24EULL, "hlac_ca_cavern1_haunted_24424_11z.dds"},
	{0x65CBD13C8758F5ACULL, "w_bonepit_7_0.rml"},
	{0x65CD08E815A6B12CULL, "gen00fl_fenris6110851.cl"},
	{0x65CD12D525205378ULL, "6080674_f.fxe"},
	{0x65CD15F22AB3CB8AULL, "hlac_lt_0_lowtown_a3_n_c_-196609_4y.dds"},
	{0x65CD30A7A3792F19ULL, "art\\characters\\creatures\\dragon\\animation\\p_rst.gad"},
	{0x65CD4667872E0DE4ULL, "w_bonepit_150_0.msh"},
	{0x65CD5867872E2C4DULL, "w_bonepit_150_0.mao"},
	{0x65CD5C67872E3286ULL, "w_bonepit_150_0.mmh"},
	{0x65CD7B82254C254AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gwalla.mao"},
	{0x65CD7F82254C2B99ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gwalla.mmh"},
	{0x65CD8982254C3C87ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gwalla.msh"},
	{0x65CDC59A5F1CF0AEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot_v2.ani"},
	{0x65CE8DBAF6C19F84ULL, "6033558_m.fxe"},
	{0x65CEBD97B06FA4B4ULL, "plt_cod_ltr_gamlen_charade.plo"},
	{0x65D02074E4F3A173ULL, "qcr220_grey_wardens6135953.cl"},
	{0x65D021EAB190F3E0ULL, "vxca_caverns2prc_9zt_0_0.vlm"},
	{0x65D06A7FE8FE00F5ULL, "vxlt_gallowstemplar_f_457_0_0.vlm"},
	{0x65D09A1893918CD7ULL, "6171400_m.fxe"},
	{0x65D0C3AD5296B685ULL, "6119606_m.fxe"},
	{0x65D1C828AF11FEDEULL, "6067328_m.fxe"},
	{0x65D1E10A1EA4FF26ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkright.evt"},
	{0x65D1F60E8C3A0430ULL, "6128654_m.fxe"},
	{0x65D22787D1AA7EEEULL, "vxw_bonepitmining_a2_1q0_0_0.vlm"},
	{0x65D2586270E8182CULL, "mag110_arianni6036801.cl"},
	{0x65D2BA2107BD53EDULL, "art\\vfx\\combat\\vfx_w_wns_right_close1_c.anb"},
	{0x65D2BD88F60B84B0ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medc.mao"},
	{0x65D2C188F60B8BC3ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medc.mmh"},
	{0x65D2C65375EF0311ULL, "193600572.ogg"},
	{0x65D2CF88F60BA31DULL, "art\\characters\\playercharacter\\humanmale\\bdy_medc.msh"},
	{0x65D3A4A6417A92E1ULL, "art\\vfx\\combat\\vfx_a_stealth_att_1_c.anb"},
	{0x65D415A7AE573907ULL, "6155820_m.fxe"},
	{0x65D430FB78504189ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.cess.ani"},
	{0x65D435B192311C01ULL, "763704297.ogg"},
	{0x65D441A7A37F6B67ULL, "art\\characters\\creatures\\dragon\\animation\\p_rst.ani"},
	{0x65D484539BC7A75EULL, "llac_lt_0_lowtown_docks_a3_-196610_1.dds"},
	{0x65D4B6458D7877DFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_vael_0n.dds"},
	{0x65D50ABABF8534F0ULL, "hlac_lt_0_lowtown_131071_3y.dds"},
	{0x65D54854520B1E48ULL, "weapon_weapon_p1.vsh.fxo.dx11"},
	{0x65D5834A0B1CB7E3ULL, "6119381_m.fxe"},
	{0x65D6113CECCFC939ULL, "llac_ht_0_hightown_a3_-131071_1.dds"},
	{0x65D705C870FF5D47ULL, "vxw_woundedcoast_vl_0_0.vlm"},
	{0x65D7536787367851ULL, "w_bonepit_150_0.phy"},
	{0x65D7F80A1EA99271ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkright.gad"},
	{0x65D87F27DBA77EE6ULL, "lowtown_male_alienage1_sta6162761.cl"},
	{0x65D96DF2239C61F3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_quill_n.dds"},
	{0x65D9CD51633C821DULL, "304632694.ogg"},
	{0x65DA22F445A162C7ULL, "vxw_bonepit_1zz_0_0.vlm"},
	{0x65DA2EA69248F3DAULL, "art\\vfx\\textures\\vfx_blood_splatter_s.dds"},
	{0x65DA6F7407A2E009ULL, "hlac_lt_0_lowtown_a2_-131072_3z.dds"},
	{0x65DAB7691B0BD8C8ULL, "dae000_arishok6070892.cl"},
	{0x65DACCBD1E829E6FULL, "hlac_w_encounter1_-65533_0y.dds"},
	{0x65DB119C91C800AAULL, "6129444_m.fxe"},
	{0x65DB4BFB78568E0BULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.cess.gad"},
	{0x65DB662C0FCCC458ULL, "vxw_sundermount_2et_0_0.vlm"},
	{0x65DBBD9A5F2920FFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot_v2.evt"},
	{0x65DBC5832CA482A9ULL, "da2terrain_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0x65DC1ECBFAC1F0F7ULL, "vxw_bonepitmining_a2_1z0_0_0.vlm"},
	{0x65DC2E55411F3DABULL, "vxw_encounter2_camp_7n_0_0.vlm"},
	{0x65DC6DBD6D504F9DULL, "hlac_lt_0_lowtown_a2_-65537_5z.dds"},
	{0x65DC97EE5C745F6AULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarp_0.pwk"},
	{0x65DCA1E649EC8E01ULL, "lla_ht_0_hightown_65536_0.dds"},
	{0x65DD7E8F92F9612AULL, "one100_party_comments6112656.cl"},
	{0x65DD9732537B175AULL, "vxlt_0_lowtown_docks_n_c4r_0_0.vlm"},
	{0x65DDB52EBB3AD586ULL, "vxht_0_hightown_ame_0_0.vlm"},
	{0x65DDBE7ECF62BDFAULL, "vxw_woundedcoast_1co_0_0.vlm"},
	{0x65DDC53303F3CC32ULL, "vxca_cavern1_haunted_hvp_0_0.vlm"},
	{0x65DDD4B5421D0FBFULL, "vxlt_0_lowtown_a2_bz2_0_0.vlm"},
	{0x65DE27D45AD24E49ULL, "follower_banter6155652.cl"},
	{0x65DE2BBB5142305DULL, "6129230_m.fxe"},
	{0x65DE57C1F2DB6FA8ULL, "786769827.ogg"},
	{0x65DE9BB3D5B132B6ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg.anb"},
	{0x65DEA8A51BAD04B1ULL, "vxht_mansion2b_33r_0_0.vlm"},
	{0x65DEF90A1EAFB21FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkright.ani"},
	{0x65DF1B9D455EB247ULL, "qun110_petrice6018544.cl"},
	{0x65DF53362C70A425ULL, "vxlt_warehouse_103_0_0.vlm"},
	{0x65DF7A09ACAB96A6ULL, "222619394.ogg"},
	{0x65DFD7381AE61354ULL, "vrc321_varric6087060.cl"},
	{0x65DFDDBF7F3A2A02ULL, "vxlt_0_lowtown_a2_n_c_eoo_0_1.vlm"},
	{0x65E03C5F1FE82AC2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityenter.gad"},
	{0x65E0700C0F11E315ULL, "6179802_m.fxe"},
	{0x65E0811CA050A669ULL, "llac_w_encounter3_-196607_0.dds"},
	{0x65E085A64184E85AULL, "art\\vfx\\combat\\vfx_a_stealth_att_1_c.mmh"},
	{0x65E09CAD19E6BB5BULL, "514163179.ogg"},
	{0x65E0A1B9CB99AD15ULL, "w_encounter3_25_0_ll.mmh"},
	{0x65E0B289A919A7B2ULL, "qun160_vashothlead6053939.cl"},
	{0x65E12696BA55FE2FULL, "hlac_lt_0_lowtown_a2_n_c_65535_1z.dds"},
	{0x65E2122603B37142ULL, "hlac_w_encounter2_327683_0y.dds"},
	{0x65E2348681F8F370ULL, "vxdr_dungeon_1md_0_0.vlm"},
	{0x65E248FB785CA870ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.cess.evt"},
	{0x65E28DC3E0C4EF0AULL, "1057824786.ogg"},
	{0x65E29E9A5F2F0AE8ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot_v2.gad"},
	{0x65E3057C3BA0F0C9ULL, "6111069_f.fxe"},
	{0x65E32BBCB2B079A3ULL, "6130050_m.fxe"},
	{0x65E344E04F4C3E38ULL, "6047409_m.fxe"},
	{0x65E378A58A261A0FULL, "zz_lgt_2_debug.cnv"},
	{0x65E394399D70A1CCULL, "hlac_lt_0_lowtown_docks_n_-131074_3y.dds"},
	{0x65E3C288F61A00E8ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medc.phy"},
	{0x65E3C2C958271F70ULL, "w_sundermount_251_0.mao"},
	{0x65E3C6C958272583ULL, "w_sundermount_251_0.mmh"},
	{0x65E3D4C958273DDDULL, "w_sundermount_251_0.msh"},
	{0x65E3EC576742B795ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_002a_s.dds"},
	{0x65E3F526FD12D063ULL, "6114481_m.fxe"},
	{0x65E3F81B9D9BE103ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\vista_cmountain.mao"},
	{0x65E3FC1B9D9BE850ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\vista_cmountain.mmh"},
	{0x65E40A1B9D9C006AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\vista_cmountain.msh"},
	{0x65E4C5BB8CAA8D38ULL, "115085350.ogg"},
	{0x65E577A562F575C7ULL, "vxw_woundedcoast_so_0_0.vlm"},
	{0x65E58FBD296AA365ULL, "vxlt_0_lowtown_a2_n_gbr_0_0.vlm"},
	{0x65E5FAC1E371993BULL, "355846178.ogg"},
	{0x65E5FACB6477A1C3ULL, "vxca_caverns1_ehs_0_0.vlm"},
	{0x65E6696F9F08AE23ULL, "vxht_chantry_n_34v_0_0.vlm"},
	{0x65E69B2A6C865D26ULL, "character_character_p113.psh.fxo.dx11.assembly"},
	{0x65E6A4BC97A3FF61ULL, "ser100_party_comments.bnk"},
	{0x65E6C0CF4E698093ULL, "hlac_w_woundedcoast_327681_0y.dds"},
	{0x65E6CD5DCE00DEFCULL, "vxw_blightlands_124_0_0.vlm"},
	{0x65E6EE3A97A31258ULL, "hlac_lt_0_lowtown_a3_n_c_-1_0y.dds"},
	{0x65E75F5F1FEE84B9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityenter.evt"},
	{0x65E77037B62B7E4DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_wallcap.msh"},
	{0x65E77E37B62B9580ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_wallcap.mao"},
	{0x65E78237B62B9CD3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_wallcap.mmh"},
	{0x65E7BC8567C6AB80ULL, "150159133.ogg"},
	{0x65E7EAA8F6156C58ULL, "isa311_isabela6095316.cl"},
	{0x65E813994A835119ULL, "6180693_m.fxe"},
	{0x65E914CF0E8239FDULL, "mer120_hubert6077588.cl"},
	{0x65E92EA2C332ED28ULL, "llac_lt_0_lowtown_docks_a3_65532_0.dds"},
	{0x65E92F652E497CC8ULL, "hlac_ht_brothel_n_2995_5y.dds"},
	{0x65E9C6BC97A66ED1ULL, "ser100_party_comments.cnv"},
	{0x65E9D80901C2DCD8ULL, "vxw_sundermount_2nu_0_0.vlm"},
	{0x65E9F34B6355392DULL, "plt_vrc100pt_n_initial.plo"},
	{0x65EA36572F263922ULL, "6083656_m.fxe"},
	{0x65EAA0323315562BULL, "vxw_encounter2_camp_2a_0_0.vlm"},
	{0x65EB9EB3D5BBC221ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg.mmh"},
	{0x65EBC4315C815D0CULL, "hlac_lt_0_lowtown_a2_n_c_131073_0y.dds"},
	{0x65EC03652BA27B47ULL, "dae000_merchant_armor_hm6135892.cl"},
	{0x65EC110C2156D53FULL, "vxlt_warehouse_n_11y_0_0.vlm"},
	{0x65EC305BC158D67AULL, "vxw_woundedcoast_1nh_0_0.vlm"},
	{0x65EC7FC95916FA6EULL, "653737872.ogg"},
	{0x65ED53253C9144B6ULL, "dae000_mother6135739.cl"},
	{0x65ED958A57F8C54BULL, "hlac_lt_darktown_-1_68y.dds"},
	{0x65EE0C3CBC7F60E5ULL, "cod_bks_sermons.utp"},
	{0x65EE259F3C3CAF40ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_hightown_windows_d.dds"},
	{0x65EE35F0C416FC8BULL, "6033572_f.fxe"},
	{0x65EE76B9B291373BULL, "vxht_playershouse_n_1rg_0_0.vlm"},
	{0x65EEC517547C6E2BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_encn_0t.dds"},
	{0x65EEECAD1775D06AULL, "356829351.ogg"},
	{0x65EF104A5149486DULL, "vxw_bonepitmining_a2_1rz_0_0.vlm"},
	{0x65EF4510D6D7C6FEULL, "ts_darkspawn_arm_props.gda"},
	{0x65EF4D02412E9880ULL, "hlac_lt_warehouse_n_1_1x.dds"},
	{0x65EF82767B7550EDULL, "hlac_dr_0_deeproads_17039_2y.dds"},
	{0x65EF99F2E7CEF591ULL, "6023987_m.fxe"},
	{0x65F02FF306C05E0EULL, "6168483_m.fxe"},
	{0x65F3281BD8C80D6CULL, "vxht_0_hightown_a2_9us_0_0.vlm"},
	{0x65F32D90DE818901ULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.anb"},
	{0x65F3A4459266428DULL, "vxlt_0_lowtown_n_exr_0_0.vlm"},
	{0x65F4208CF8DE3234ULL, "follower_banter6163571.cl"},
	{0x65F4675F1FF91CA8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\immunityenter.ani"},
	{0x65F4CF0DF025CBCBULL, "vxw_bonepit_20k_0_0.vlm"},
	{0x65F4DB4C90FEC6A3ULL, "vxht_chantry_n_31y_0_0.vlm"},
	{0x65F5C8BE139AB1A1ULL, "665443320.ogg"},
	{0x65F5F9137DB38A19ULL, "hlac_ca_caverns2_5802_15x.dds"},
	{0x65F601013614624FULL, "6048807_m.fxe"},
	{0x65F60E9D34428199ULL, "6141812_m.fxe"},
	{0x65F645D232DF241DULL, "abi_w_stun.wbd"},
	{0x65F64983C7771FA8ULL, "oil_proj_dot.mao"},
	{0x65F65A1EF7CA6D71ULL, "6096207_f.fxe"},
	{0x65F66AEC065398D4ULL, "6129126_f.fxe"},
	{0x65F67EBA6FCC2147ULL, "690732389.ogg"},
	{0x65F69DF81C18148AULL, "6100759_m.fxe"},
	{0x65F7492393478B52ULL, "6101059_m.fxe"},
	{0x65F777BFCF61C61AULL, "gen00fl_anders6111175.cl"},
	{0x65F7B9DC32A72745ULL, "6049474_f.fxe"},
	{0x65F862C62DE2F465ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_c.mmh"},
	{0x65F87021682BFEF6ULL, "lt_0_lowtown_docks_n_19_0_ll.mmh"},
	{0x65F91920DD11745AULL, "70672213.ogg"},
	{0x65F986E853C37B41ULL, "w_bonepitdestroyed_a3_53_0.rml"},
	{0x65F99382A6F59281ULL, "hlac_lt_gallowstemplar_2391_0x.dds"},
	{0x65F9CA3C9F568656ULL, "440795539.ogg"},
	{0x65FAA238B34EEEFAULL, "vxw_woundedcoast_1ii_0_0.vlm"},
	{0x65FAFBA30E02A48BULL, "framebuffereffects_illuminationmask_p0.vsh.fxo"},
	{0x65FBA0150B351D3DULL, "mer340_bonepit_comments6083145.cl"},
	{0x65FCA296826512D7ULL, "vxlt_0_lowtown_a2_ff2_0_0.vlm"},
	{0x65FCB5C4A7E78DD1ULL, "ico_maul_3.dds"},
	{0x65FD033A37BA531EULL, "6018652_m.fxe"},
	{0x65FD2439D7FE0E99ULL, "vxlt_0_lowtown_a2_n_cmk_0_0.vlm"},
	{0x65FD7226320667F1ULL, "6086807_m.fxe"},
	{0x65FDE013A9ABDC7CULL, "68938093.ogg"},
	{0x65FE67C48CE6D413ULL, "6090373_f.fxe"},
	{0x65FE7EDA9B9D53AFULL, "vxca_caverns2_938_0_1.vlm"},
	{0x65FE82F0E49F9708ULL, "and111_lirene6084763.cl"},
	{0x65FF19701EA29760ULL, "vxdr_primevaltomb_84h_0_0.vlm"},
	{0x65FF5E0291D824E0ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc103a_0s.dds"},
	{0x66010E90DE8D927AULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.mmh"},
	{0x6601559349A1A8C5ULL, "ht_0_hightown_a2_probe_9x4.dds"},
	{0x66021622845C5B0DULL, "vxlt_0_lowtown_n_euq_0_0.vlm"},
	{0x660278AA8ADBD29CULL, "hlac_lt_0_lowtown_a2_n_65535_6z.dds"},
	{0x6602B31B7DCE3593ULL, "gen_im_trs_rdr_hvy_boo_r0.uti"},
	{0x6602CA238C896551ULL, "833496343.ogg"},
	{0x660307A0CC4F11B0ULL, "470422117.ogg"},
	{0x6603B34A53727E52ULL, "llac_dr_primevaltomb_664_6.dds"},
	{0x6603CAF43BCF2FFAULL, "hlac_lt_0_lowtown_a2_n_c_-65536_0x.dds"},
	{0x66043C190F5170CAULL, "vxlt_gallowstemplar_f_5t6_0_0.vlm"},
	{0x660472DE280A7ACFULL, "6043907_m.fxe"},
	{0x6604BC2DA7923868ULL, "gen_im_npc_arm_hum_cht_slvb.uti"},
	{0x6604C413781B8305ULL, "llac_lt_0_lowtown_docks_a3_n_-262144_3.dds"},
	{0x6604DD5B078CBBC4ULL, "fx_mfx_def.dds"},
	{0x6604ED4D705886C7ULL, "clf360cr_evets_thug.utc"},
	{0x66050FD50E0E2D0AULL, "6105459_m.fxe"},
	{0x660539DF0B0C89C0ULL, "181331122.ogg"},
	{0x6605E2CE7380F2CAULL, "6185482_m.fxe"},
	{0x66065FC62DEF2C32ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_c.anb"},
	{0x66069839F9771179ULL, "6136151_m.fxe"},
	{0x6606BBC2E5AF0DD8ULL, "vxw_sundermount_2ts_0_0.vlm"},
	{0x6606E2FA381D591EULL, "llac_lt_gallowscourtyard_-65536_2.dds"},
	{0x66076107AB90FE57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_weightshift.gad"},
	{0x66076D0C51FF953AULL, "6075100_f.fxe"},
	{0x6607A89E05325735ULL, "vxw_encounter2_7h_0_0.vlm"},
	{0x6607E797B189FCACULL, "589605074.ogg"},
	{0x66082B23F31E3D28ULL, "zz_vault_debug6178946.cl"},
	{0x6608B622D3D74D45ULL, "vxw_woundedcoast_1cd_0_0.vlm"},
	{0x660914974EF599A7ULL, "llac_w_bonepitmining_a2_1_0.dds"},
	{0x6609343A62376432ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_forward.anb"},
	{0x6609890AC6FE1C4EULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\threaten.ani"},
	{0x6609CBB3BFFD0EF0ULL, "6075493_m.fxe"},
	{0x660A03B0B9096183ULL, "756746789.ogg"},
	{0x660A1AB2EE070235ULL, "6167351_m.fxe"},
	{0x660AC7F9FB7DA460ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_meda_0s.dds"},
	{0x660AEAFCDFB37BD7ULL, "mag170_thrask6016935.cl"},
	{0x660B2B47DE807769ULL, "vxlt_0_lowtown_a2_n_feb_0_0.vlm"},
	{0x660B3800DD68C554ULL, "gen00fl_aveline6120992.cl"},
	{0x660B4A0BDEFD1BE2ULL, "dae320ip_chant_board_ht.ncs"},
	{0x660B4F11C731A8B9ULL, "992500936.ogg"},
	{0x660BF54CE37CEEB4ULL, "vxlt_0_lowtown_a2_gc3_0_0.vlm"},
	{0x660C7947C74FE836ULL, "w_bonepitmining_a2_231_0.msh"},
	{0x660C8347C74FF934ULL, "w_bonepitmining_a2_231_0.mmh"},
	{0x660C8747C7500047ULL, "w_bonepitmining_a2_231_0.mao"},
	{0x660E124605FD4459ULL, "vxht_0_hightown_a2_91n_0_0.vlm"},
	{0x660E33728C7C0290ULL, "vxht_mansion2b_2ps_0_0.vlm"},
	{0x660E7607AB9740E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_weightshift.evt"},
	{0x660E86755A422DAEULL, "vxlt_gallowsprison_n_68m_0_0.vlm"},
	{0x660E872E74501AE4ULL, "vxlt_0_lowtown_docks_cs3_0_1.vlm"},
	{0x660EC0F8084903C9ULL, "6090321_m.fxe"},
	{0x661020FB588A7C45ULL, "317202550.ogg"},
	{0x66107977F9FCF407ULL, "vxlt_0_lowtown_docks_a3_n_cyo_0_0.vlm"},
	{0x6610A5D89151C946ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas_drelief.dds"},
	{0x66113FD25FC1E572ULL, "hlac_dr_dungeon_3215_6y.dds"},
	{0x66114AFC7CF41239ULL, "vxdr_0_deeproads_cvr_0_0.vlm"},
	{0x66117B36E4C33057ULL, "vxht_keep_4hz_0_0.vlm"},
	{0x6611997B956F0B83ULL, "dae000_mother6127173.cl"},
	{0x66119CD4162E9420ULL, "6013307_m.fxe"},
	{0x6611E6C431A4F9E4ULL, "hlac_lt_0_lowtown_a2_n_c_1_0y.dds"},
	{0x66125441E0919913ULL, "hlac_ca_caverns2_5802_73z.dds"},
	{0x6612D96A6FB00403ULL, "ch1notnt_ds1.psh.fxo"},
	{0x6612DF0991C771EFULL, "hlac_ht_0_hightown_n_131071_2x.dds"},
	{0x6612F714C6364DB2ULL, "861826188.ogg"},
	{0x661303A27C067AE3ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\256x256.dds"},
	{0x661341BCAFFFB2EDULL, "hlac_dr_slavers_3215_1y.dds"},
	{0x66134AADFB17F137ULL, "6114870_m.fxe"},
	{0x66138617FDD2C23BULL, "442188284.ogg"},
	{0x6613C66E47DD822AULL, "382678309.ogg"},
	{0x6613EA680D79C4C7ULL, "vxw_bonepit_1nv_0_0.vlm"},
	{0x66142CCCDBE6A4F2ULL, "hlac_w_sundermount_65533_1z.dds"},
	{0x66146360632D32B7ULL, "hlac_dr_dungeon_3207_8y.dds"},
	{0x661473C650A2A687ULL, "428480261.ogg"},
	{0x66152D9FD7A52658ULL, "vxw_sundermount_2yp_0_0.vlm"},
	{0x66155ABFCAB03D86ULL, "dae000_arishok6096326.cl"},
	{0x6615798269715B53ULL, "1007818336.ogg"},
	{0x66161A7AF7286FB5ULL, "vxw_encounter2_2w_0_0.vlm"},
	{0x661669D3CA599EB2ULL, "6007722_m.fxe"},
	{0x661692CD59BF552EULL, "art\\characters\\playercharacter\\humanmale\\arm_maso.phy"},
	{0x6616DB90CA96BA36ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.impa.evt"},
	{0x6616F6DFD80609E1ULL, "465512194.ogg"},
	{0x6617152BC0D719CBULL, "fen310_fenris6118211.cl"},
	{0x661727FFC5CD65C5ULL, "vxw_woundedcoast_1nc_0_0.vlm"},
	{0x6617810AC70A4B9FULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\threaten.evt"},
	{0x6617C95F226BBFDDULL, "6174463_m.fxe"},
	{0x66191673256BBBF7ULL, "mer110_jansen6032897.cl"},
	{0x6619CD62D2B0EB18ULL, "and100_anders6173584.cl"},
	{0x6619FA8DC888B538ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxa_n.dds"},
	{0x661A095D6D7B1516ULL, "6179988_m.fxe"},
	{0x661ABF42EC0B1683ULL, "6111802_m.fxe"},
	{0x661B3BCCC435763AULL, "303844781.ogg"},
	{0x661BDB6A5B59B594ULL, "297075877.ogg"},
	{0x661C1208AB17F23BULL, "854003100.ogg"},
	{0x661C31D9D65CFA97ULL, "6035919_m.fxe"},
	{0x661C4087372ADDCAULL, "hlac_lt_0_lowtown_docks_-2_1z.dds"},
	{0x661C6DA6B29C4D77ULL, "dre000st_carver.stg"},
	{0x661C8E07ABA3A6DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_weightshift.ani"},
	{0x661CA54F7E721B10ULL, "vxht_mansion2b_2mp_0_0.vlm"},
	{0x661D3AE32626B9DCULL, "153002487.ogg"},
	{0x661D5A20C8BCDD91ULL, "32455569.ogg"},
	{0x661DA323D8958509ULL, "6017827_m.fxe"},
	{0x661DD290CA9CC9C1ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.impa.gad"},
	{0x661DDEE56D2EB381ULL, "vxlt_0_lowtown_n_eov_0_1.vlm"},
	{0x661E4811DD0B4D7CULL, "vxlt_0_lowtown_docks_a3_d1l_0_0.vlm"},
	{0x661E620AC7103588ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\threaten.gad"},
	{0x661E667FB75BE9EBULL, "34033569.ogg"},
	{0x661F12849F7E7287ULL, "525639780.ogg"},
	{0x661FBCD96EEA2AB9ULL, "vxdr_0_deeproads_csq_0_0.vlm"},
	{0x661FBECD931F8554ULL, "6128298_m.fxe"},
	{0x661FF0C066E9F0F4ULL, "vxht_mansion2_31e_0_0.vlm"},
	{0x6621079EE0B6A5FCULL, "vxlt_darktown_2se_0_0.vlm"},
	{0x66211FD2F33DA1CAULL, "vxlt_gallowstemplar_f_5b0_0_0.vlm"},
	{0x662146A526F52A08ULL, "739260403.ogg"},
	{0x6621A0E80FD39BE0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratetarp_d.dds"},
	{0x6621BC8AED0E09B7ULL, "6113970_m.fxe"},
	{0x66226592CB708F44ULL, "hlac_lt_0_lowtown_a2_n_-65535_3x.dds"},
	{0x662268CF5746FB28ULL, "vxlt_gallowstemplar_f_43h_0_0.vlm"},
	{0x66228B5A4BC36082ULL, "vxht_0_hightown_a3_c92_0_0.vlm"},
	{0x66232762FD60368FULL, "383196201.ogg"},
	{0x66235D901A64338EULL, "329234487.ogg"},
	{0x662409176BD12091ULL, "hlac_w_bonepit_393219_0z.dds"},
	{0x662463A24883E2C5ULL, "character_character_p204.psh.fxo.dx11"},
	{0x6624AC46561B58C0ULL, "zz_followers_banter6159507.cl"},
	{0x6624C40BCC2792B2ULL, "vxlt_undercitydungeon_1ks_0_0.vlm"},
	{0x6624D390CAA2EA6FULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.impa.ani"},
	{0x6624DFB41362BCB5ULL, "vxw_blightlands_aa_0_0.vlm"},
	{0x66250F2CAB83137EULL, "129802680.ogg"},
	{0x662599DCB7C37E45ULL, "vxw_woundedcoast_1ib_0_0.vlm"},
	{0x6625CC2AC978F392ULL, "hlac_lt_0_lowtown_n_65535_2z.dds"},
	{0x6625DE1B729B4545ULL, "78480950.ogg"},
	{0x662652304AD5B352ULL, "hlac_lt_0_lowtown_docks_n_-65538_2y.dds"},
	{0x66269A4BC2B4B233ULL, "6136676_m.fxe"},
	{0x6626A94033E95829ULL, "gen_im_trs_slv_lgt_hlm_r0.uti"},
	{0x6626AAA7780B05FFULL, "gen00fl_isabela6097287.cl"},
	{0x6626E7B0C4CE04AEULL, "vxdr_primevaltomb_536_0_0.vlm"},
	{0x6626EFA2D39F4155ULL, "6036765_m.fxe"},
	{0x662752E24C04D2F4ULL, "zz_followers_banter6159675.cl"},
	{0x662801ED13CDE040ULL, "blood_conj_06_l3.mao"},
	{0x66281F7CB1EF8FCBULL, "hlac_lt_0_lowtown_a3_n_c_65535_6x.dds"},
	{0x662A71DCBC495C5BULL, "art\\characters\\playercharacter\\humanfemale\\glv_medc.phy"},
	{0x662B7C6140841DCEULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash_p.impa.ani"},
	{0x662BB0B1FCCBB7EFULL, "6202065_m.fxe"},
	{0x662BC02941FACBB6ULL, "qun231ar_varnell_hideout.are"},
	{0x662BF2BA538F5E32ULL, "249366413.ogg"},
	{0x662C7CE4918B61F5ULL, "hlac_dr_primevalentrance_1724_5z.dds"},
	{0x662C99AC90991628ULL, "332110536.ogg"},
	{0x662D0A38041B1D12ULL, "6170774_m.fxe"},
	{0x662D7DF9F59C5023ULL, "6107382_m.fxe"},
	{0x662D9E2B54B0414AULL, "11677749.ogg"},
	{0x662DB695733147EBULL, "hlac_lt_0_lowtown_a3_n_c_65536_12z.dds"},
	{0x662DD79438C41329ULL, "pro000_cut2d_real_story6013914.cl"},
	{0x662DDDAFE3F63CA4ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_3y.dds"},
	{0x662DF5918CBC6123ULL, "6207779_f.fxe"},
	{0x662E726685D57AC7ULL, "qun120cr_javaris6044057.cl"},
	{0x662ECE4920F18EAAULL, "vxdr_primevaltomb_6wq_0_0.vlm"},
	{0x662F0EAE27A08AB8ULL, "332032715.ogg"},
	{0x662F4A93F1011379ULL, "vxca_caverns1_mines_i1k_0_0.vlm"},
	{0x662F4FC557F7936FULL, "23587367.ogg"},
	{0x662F797BD2ACBE7CULL, "vxlt_darktown_2vz_0_0.vlm"},
	{0x662F7F0EF4B98A30ULL, "hlac_lt_0_lowtown_a2_-262144_3z.dds"},
	{0x662FC9DE3817CCFBULL, "dae000_sandal6147759.cl"},
	{0x662FDC1D2B819580ULL, "follower_banter6163126.cl"},
	{0x6630232822C8E0A8ULL, "hlac_lt_0_lowtown_docks_n_196607_2z.dds"},
	{0x6631590020E2C3EDULL, "6100074_f.fxe"},
	{0x6631A10575EDF2E2ULL, "659052795.ogg"},
	{0x6631CE2BF111022AULL, "hlac_ca_cave_2194_5z.dds"},
	{0x6631F7C2883FE273ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_5.ani"},
	{0x663215778A090062ULL, "rdr110dg_party_comments6094680.cl"},
	{0x663227EB006A92EAULL, "hlac_lt_gallowscourtyard_n_-65536_2y.dds"},
	{0x6632564B2F7B0A51ULL, "6143069_m.fxe"},
	{0x6632C5380F033AA1ULL, "303983546.ogg"},
	{0x6632FE34DB14A0B5ULL, "vxw_encounter2_8u_0_0.vlm"},
	{0x6633D6A247197551ULL, "zz_vrc_debug6200252.cl"},
	{0x663409FFCC574DE4ULL, "mag121_idunna6084874.cl"},
	{0x66343BD6A2BE93C8ULL, "vxw_woundedcoast_1d8_0_0.vlm"},
	{0x6634AB63EF2AB6C2ULL, "art\\vfx\\combat\\vfx_glb_ele_f_minor_c.mmh"},
	{0x66355636C7F480C9ULL, "6135943_m.fxe"},
	{0x6635741C14C4D1B6ULL, "zz_mer_debug6062737.cl"},
	{0x66358F63115532E7ULL, "mcr371_orsino6083457.cl"},
	{0x6635AD11E17364E1ULL, "llac_dr_0_deeproads_19793_0.dds"},
	{0x6636087141E819B5ULL, "vxlt_0_lowtown_a3_n_c_bz0_0_0.vlm"},
	{0x6636EE1A1678D023ULL, "gen00fl_carver6104848.cl"},
	{0x6636FF5B88371D24ULL, "lgt160_ntb_elf_conflict6163330.cl"},
	{0x66371220809F597CULL, "6139847_m.fxe"},
	{0x663736C66C51C59EULL, "hlac_lt_0_lowtown_a2_n_-131072_3x.dds"},
	{0x6637E44EDCD829B9ULL, "s_deca_cr4.mao"},
	{0x66385DAA8114B06AULL, "364027242.ogg"},
	{0x6638774911DF6BE9ULL, "633195897.ogg"},
	{0x66388336164ACA6AULL, "6110493_m.fxe"},
	{0x6638A993007C5990ULL, "263223895.ogg"},
	{0x6638FD22D2794A33ULL, "t3_wdg_w02.tnt"},
	{0x6638FFA030378E39ULL, "vxht_chantry_2dr_0_0.vlm"},
	{0x663938749258C1CAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_2.gad"},
	{0x663957CB291A4FCAULL, "character_character_p179.psh.fxo.dx11.assembly"},
	{0x6639746140904D1FULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash_p.impa.evt"},
	{0x6639C8D01C6D5AB7ULL, "gen00fl_varric6099155.cl"},
	{0x6639DB9B98EF688CULL, "vrc321_party_barks6095763.cl"},
	{0x663A7784223DCAF4ULL, "vxlt_warehouse_qt_0_0.vlm"},
	{0x663A78E853FAFA30ULL, "w_bonepitdestroyed_a3_53_0.msh"},
	{0x663A86E853FB124AULL, "w_bonepitdestroyed_a3_53_0.mmh"},
	{0x663A8AE853FB1881ULL, "w_bonepitdestroyed_a3_53_0.mao"},
	{0x663A91C028A721AEULL, "vxw_bonepit_p2_0_0.vlm"},
	{0x663B031C3989BC7DULL, "6114058_m.fxe"},
	{0x663B3EA3700A46D6ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb102a_0d.dds"},
	{0x663B75A2C3E12C95ULL, "123427895.ogg"},
	{0x663BD4F988ADB1E4ULL, "follower_banter6171882.cl"},
	{0x663C0AC441FEBFCDULL, "86043783.ogg"},
	{0x663CEB369876DB53ULL, "387621916.ogg"},
	{0x663CFB62228955E8ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_encn_0s.dds"},
	{0x663D01BCC2FBCA11ULL, "vxlt_hangedman_tb_0_0.vlm"},
	{0x663D2EA5167BF879ULL, "dae000_gamlen6005032.cl"},
	{0x663D402612E7A72AULL, "vxdr_primevaltomb_6zr_0_0.vlm"},
	{0x663D790B46026D1DULL, "dae000_orana6149124.cl"},
	{0x663E44687ADCEF96ULL, "6089467_m.fxe"},
	{0x663E4FB662AC8220ULL, "mag100_karras6035382.cl"},
	{0x663EEFC2884A5EAAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_5.evt"},
	{0x663F42BF854650A5ULL, "6087087_f.fxe"},
	{0x663F4D6757AED14CULL, "37525855.ogg"},
	{0x663F4D79F2CAB5F8ULL, "vxlt_gallowscourtyard_7fm_0_0.vlm"},
	{0x663F52416337A303ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_min_c.impa.evt"},
	{0x663F99600BF19389ULL, "1021002030.ogg"},
	{0x663FF9D0D00B7BD2ULL, "596990299.ogg"},
	{0x664005980080903FULL, "vxw_bonepit_87_0_0.vlm"},
	{0x66401B74925EAF61ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_2.evt"},
	{0x66402D6067279FE0ULL, "qun211_party_comment6072167.cl"},
	{0x6640556140963708ULL, "art\\vfx\\cinematic\\vfx_isa_intro_glass_smash_p.impa.gad"},
	{0x6640649E40572A0AULL, "isa311_isabela6175478.cl"},
	{0x66407D21EBA32398ULL, "6141552_m.fxe"},
	{0x6640E1FB3EB9B4CFULL, "lt_gallowscourtyard_35_0_ll.mmh"},
	{0x66419CB7C80D401BULL, "vxht_mansion2_3e2_0_0.vlm"},
	{0x6641B800B4654801ULL, "hlac_lt_gallowstemplar_f_7198_1x.dds"},
	{0x6641C36DF74EEDB5ULL, "vxw_blightlands_oo_0_0.vlm"},
	{0x664241483508EE9AULL, "hlac_lt_gallowscourtyard_131071_1z.dds"},
	{0x6642ADB394B4AC48ULL, "vxw_woundedcoast_1o9_0_0.vlm"},
	{0x6643A99CB99FC08EULL, "lgt110_orz_harrowmont6109005.cl"},
	{0x6643C536A365D0B4ULL, "gen00fl_sebastian6180426.cl"},
	{0x6643E24AD620DEEFULL, "597229108.ogg"},
	{0x664434B669C0E028ULL, "257372969.ogg"},
	{0x66443D7BFBCB0119ULL, "6170405_m.fxe"},
	{0x6645319CA895805CULL, "618703078.ogg"},
	{0x6645CEC28850451DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_5.gad"},
	{0x6645E08FE6EA0A80ULL, "gen00fl_fenris6106632.cl"},
	{0x664621E40C19EDB6ULL, "vxlt_0_lowtown_a2_n_bz4_0_0.vlm"},
	{0x66462B41633D7F84ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_min_c.impa.gad"},
	{0x66463A2190CBF70EULL, "6127240_m.fxe"},
	{0x66465797467E2584ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.mmh"},
	{0x6646A546B32C9A1AULL, "vxlt_foundry_n_vx_0_0.vlm"},
	{0x6646C808C7D2DE77ULL, "311867971.ogg"},
	{0x6647F7315B673375ULL, "446665453.ogg"},
	{0x6647FD9BA208A9DEULL, "vxca_caverns2_7mh_0_0.vlm"},
	{0x6648D7C5E7493B33ULL, "6087428_m.fxe"},
	{0x66491C599567FBFCULL, "gen00fl_isabela6122906.cl"},
	{0x664945553CD0A6CBULL, "6189744_m.fxe"},
	{0x664945C4A3CE7ABBULL, "6076046_m.fxe"},
	{0x6649FD505F464D9AULL, "vxlt_gallowstemplar_f_5rp_0_0.vlm"},
	{0x6649FDF7DB1A437CULL, "6068987_m.fxe"},
	{0x664A00E845992BA6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ent.evt"},
	{0x664A7F596EE9E14DULL, "vxw_encounter2_dy_0_0.vlm"},
	{0x664A9802DBBDD39EULL, "6188598_f.fxe"},
	{0x664AF1855E348C2BULL, "805124322.ogg"},
	{0x664AF45DF76108AAULL, "6177906_m.fxe"},
	{0x664B5A6C20BD115CULL, "27484876.ogg"},
	{0x664BA0E263B8896DULL, "vxdr_primevaltomb_3vu_0_1.vlm"},
	{0x664BD111A92EABCDULL, "gen00fl_fenris6112083.cl"},
	{0x664BE9340A9DE48DULL, "gen00fl_aveline6120825.cl"},
	{0x664C2D0CFAFF732AULL, "559550155.ogg"},
	{0x664C38C53A9CABF0ULL, "vxlt_gallowstemplar_536_0_0.vlm"},
	{0x664CA167FA7079FCULL, "6054006_f.fxe"},
	{0x664CECD940FF1582ULL, "gen00fl_fenris6110831.cl"},
	{0x664D3A416343B862ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_min_c.impa.ani"},
	{0x664E1374926ADEB0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_2.ani"},
	{0x664E5EF04B930249ULL, "t3_shk_b01.tnt"},
	{0x664EBF167D296E44ULL, "6041501_f.fxe"},
	{0x664EF0353D89D9F6ULL, "469699768.ogg"},
	{0x664FA72F34ED99B2ULL, "745586607.ogg"},
	{0x664FFDF6A58EAB2FULL, "98633459.ogg"},
	{0x665017E8459DBEF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ent.gad"},
	{0x665027252DB086F9ULL, "vxlt_gallowstemplar_f_34o_0_0.vlm"},
	{0x665033798FDB2D28ULL, "vxlt_undercitydungeon_1in_0_1.vlm"},
	{0x6650354AE9450635ULL, "vxw_blightlands_pn_0_0.vlm"},
	{0x665053B474AC5634ULL, "hlac_w_sundermount_-327678_0y.dds"},
	{0x665064B67CB099B9ULL, "t3_dey_bl3.tnt"},
	{0x66509B82908C4378ULL, "284884015.ogg"},
	{0x6650C7294219F381ULL, "qun231ar_varnell_hideout.lst"},
	{0x6650DE494D2911C6ULL, "ave331_jeven6138754.cl"},
	{0x66517505A140CE4BULL, "combat_deathblow_02.gda"},
	{0x665182BF2B031807ULL, "follower_banter6155481.cl"},
	{0x6651A10E4C20B94DULL, "gen_im_cft_rec_wrn_05.uti"},
	{0x66524E2B009065ECULL, "mer120_hubert6017844.cl"},
	{0x6652834C9654CD71ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_orsino_.ani"},
	{0x66528E6F347B1CE8ULL, "zz_approval_debug6003913.cl"},
	{0x6652A80FDC7975A0ULL, "244721437.ogg"},
	{0x665366E77B8F14C4ULL, "one300_nuncio6075327.cl"},
	{0x665379F6ABB72DA8ULL, "6167314_m.fxe"},
	{0x66543F69282B4859ULL, "6161326_m.fxe"},
	{0x665475CD261CB39FULL, "878203247.ogg"},
	{0x6654CF586C56FB2FULL, "6099188_m.fxe"},
	{0x66559C4A6EE2C342ULL, "12757574.ogg"},
	{0x6655F354501BBA92ULL, "6141936_m.fxe"},
	{0x665616BF31EA01D6ULL, "hlac_ca_cavern1_haunted_24426_8x.dds"},
	{0x665658DCBC6E6CFAULL, "art\\characters\\playercharacter\\humanfemale\\glv_medc.msh"},
	{0x665666DCBC6E8433ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medc.mao"},
	{0x66566ADCBC6E8B40ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medc.mmh"},
	{0x665691CD59F537F1ULL, "art\\characters\\playercharacter\\humanmale\\arm_maso.mmh"},
	{0x66569DCD59F54B92ULL, "art\\characters\\playercharacter\\humanmale\\arm_maso.mao"},
	{0x6656AA0648109BD8ULL, "6115451_f.fxe"},
	{0x6656ABCD59F563AFULL, "art\\characters\\playercharacter\\humanmale\\arm_maso.msh"},
	{0x665718E845A3DE9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ent.ani"},
	{0x665792ACA13DD6CEULL, "vxlt_0_lowtown_docks_n_bz5_0_0.vlm"},
	{0x6657BF2942200578ULL, "qun231ar_varnell_hideout.ncs"},
	{0x6657E087AD9B1DC3ULL, "vxlt_backalley_g_1w7_0_0.vlm"},
	{0x665802E3E2B585B1ULL, "qun221_petrice6185230.cl"},
	{0x665866ACF08BD6BAULL, "hm_lgt_den_rev.mor"},
	{0x66586F2D513C661AULL, "vxlt_gallowstemplar_f_5uq_0_0.vlm"},
	{0x665885BA534E7251ULL, "vxht_mansion_n_5aq_0_0.vlm"},
	{0x6658F13660DFF9CDULL, "vxw_encounter2_ox_0_0.vlm"},
	{0x66598E4C965AFE13ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_orsino_.gad"},
	{0x6659DAB3AE1CF54CULL, "vrc221_cassandra6040885.cl"},
	{0x6659DB75D7E96E0BULL, "hlac_lt_0_lowtown_-65538_0x.dds"},
	{0x6659F9D34E1C7A4EULL, "dae000_gamlen6149449.cl"},
	{0x665AD9E9048402E9ULL, "gen00fl_fenris6126794.cl"},
	{0x665ADA9AD74D59C0ULL, "6151254_m.fxe"},
	{0x665B154CD91050D2ULL, "599341528.ogg"},
	{0x665B256D63C8B92BULL, "vxw_bonepit_220_0_1.vlm"},
	{0x665B28225EC92096ULL, "6087667_m.fxe"},
	{0x665B32B8C3AA28B7ULL, "hlac_lt_0_lowtown_a2_n_65537_5z.dds"},
	{0x665B917E73E95D94ULL, "589770525.ogg"},
	{0x665BD3F57AEC554EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardright.ani"},
	{0x665C4852717DCEA9ULL, "vxlt_undercityencounter_1ph_0_0.vlm"},
	{0x665C58A232151121ULL, "850086233.ogg"},
	{0x665C6559B6BCE892ULL, "zz_dog_debug6127933.cl"},
	{0x665DAA63EF4DBA69ULL, "art\\vfx\\combat\\vfx_glb_ele_f_minor_c.anb"},
	{0x665DB2639C63D911ULL, "hlac_lt_backalley_g_65537_6y.dds"},
	{0x665DEFD19BD5E738ULL, "6070040_f.fxe"},
	{0x665EA727DB3B1EB5ULL, "vxw_blightlands_um_0_0.vlm"},
	{0x665F78E60DBAD4C7ULL, "w_sundermount_29_0_ll.mmh"},
	{0x6660290F8C93811FULL, "vxlt_0_lowtown_a2_n_c_hju_0_0.vlm"},
	{0x66607B3EEADF051AULL, "ss_crow_m_1.bnk"},
	{0x666087DD68620731ULL, "qun210cr_guard6087756.cl"},
	{0x6660928E257487C7ULL, "vxca_caverns1_eho_0_0.vlm"},
	{0x6660F8C5ED5D6ECDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appb_0t.dds"},
	{0x666141CA07FAFCF8ULL, "zz_and_debug6134086.cl"},
	{0x666155BA3DA58852ULL, "6013311_f.fxe"},
	{0x6661C1B5D4480849ULL, "190305364.ogg"},
	{0x6662B1461A2160D9ULL, "6166226_m.fxe"},
	{0x66630C46073315A7ULL, "hlac_ca_caverns1_24425_8y.dds"},
	{0x6663204CCEE6A4F0ULL, "vxht_0_hightown_a2_7iy_0_0.vlm"},
	{0x666406E88B0B635FULL, "vxlt_0_lowtown_a2_n_c_gz6_0_0.vlm"},
	{0x66640C76338B70A6ULL, "vxca_caverns2prc_ab2_0_0.vlm"},
	{0x666429F357B1A998ULL, "hlac_w_encounter2_camp_131076_0x.dds"},
	{0x66647A94A94249E1ULL, "vxw_woundedcoast_1p1_0_0.vlm"},
	{0x66648124E82E4EA3ULL, "w_bonepitmining_a2_71_0.mao"},
	{0x66648524E82E55F0ULL, "w_bonepitmining_a2_71_0.mmh"},
	{0x66649324E82E6D0AULL, "w_bonepitmining_a2_71_0.msh"},
	{0x6664BA0FB9C1DC2EULL, "hlac_lt_0_lowtown_docks_a3_-65540_0z.dds"},
	{0x6664BD28D78D3381ULL, "6139610_f.fxe"},
	{0x6664CF20C77CAD03ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small.mao"},
	{0x6664D320C77CB450ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small.mmh"},
	{0x6664E06448827311ULL, "dae000_quintus6154577.cl"},
	{0x6664E120C77CCC6AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small.msh"},
	{0x6664E2B923C7C739ULL, "6080108_m.fxe"},
	{0x6664FA2C930E74B4ULL, "dae000_epic_merchant_16179413.cl"},
	{0x666525052E016BC3ULL, "vxw_bonepitmining_a2_1mm_0_0.vlm"},
	{0x666632A3CB0C8D93ULL, "162982977.ogg"},
	{0x66668E7454681288ULL, "character_character_p283.psh.fxo.dx11"},
	{0x666701C5C0FF84EBULL, "vxlt_gallowscourtyard_n_7fy_0_0.vlm"},
	{0x6667631352D6124DULL, "vxw_encounter2_jw_0_0.vlm"},
	{0x666866A7EE370C9EULL, "dre000_varric6037010.cl"},
	{0x6668B83189C3D637ULL, "901013159.ogg"},
	{0x6668CE256287581AULL, "static_static_p9.psh.fxo.dx11"},
	{0x66691A3C1DD68C14ULL, "dre00st_thewayout.stg"},
	{0x66694C9C7A7E1432ULL, "886893534.ogg"},
	{0x6669845FBAD4C7B8ULL, "genip_cin_tmp_badge.utp"},
	{0x6669CBF57AF8849FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardright.evt"},
	{0x666AB54BC1B07BC1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comc_0s.dds"},
	{0x666AF851D261B0CEULL, "isa321_velasco6117613.cl"},
	{0x666AFF2E8029B44BULL, "550442850.ogg"},
	{0x666B0FBCA2C81FD3ULL, "hlac_dr_primevaltomb_11985_2y.dds"},
	{0x666C09584BBC5B7BULL, "vxlt_0_lowtown_a3_n_c_ghu_0_0.vlm"},
	{0x666DA50B6F326785ULL, "vxlt_gallowscourtyard_7ft_0_0.vlm"},
	{0x666DD88BBC611BBDULL, "6137641_f.fxe"},
	{0x666E034A6A96F5C8ULL, "vxw_woundedcoast_1p4_0_0.vlm"},
	{0x666E5A97469F7D7BULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.anb"},
	{0x666E664E64EF51E8ULL, "vxht_0_hightown_a3_a7h_0_0.vlm"},
	{0x666EA337683F573EULL, "vxht_0_hightown_a2_c4d_0_0.vlm"},
	{0x666F11C5324BFC5EULL, "vxlt_0_lowtown_h29_0_0.vlm"},
	{0x666F670CD4B908BEULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv3.ani"},
	{0x667032EAD0960ACBULL, "vxlt_backalley_n_1r9_0_0.vlm"},
	{0x6670A1CDECC5B154ULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn.phy"},
	{0x6670ACF57AFE6E88ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforwardright.gad"},
	{0x6670C105BF95CDE2ULL, "vxlt_blackemporium_ib_0_0.vlm"},
	{0x66711A5DEBDDF1A3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_enter.ani"},
	{0x66714A7DF031ACAFULL, "6097890_m.fxe"},
	{0x667271239A1EA51FULL, "mcr370_negotiation6080151.cl"},
	{0x6672A9AB8F235791ULL, "349918614.ogg"},
	{0x6672B5335BCFDF21ULL, "vxlt_0_lowtown_a2_gnq_0_0.vlm"},
	{0x66730A973030F434ULL, "s_pool_glow.mao"},
	{0x667396E21FF78443ULL, "vxw_bonepitmining_a2_1jl_0_0.vlm"},
	{0x6673ABD52979786BULL, "vxht_mansion2_3gl_0_0.vlm"},
	{0x6673CA8899D24CC9ULL, "665214245.ogg"},
	{0x6673FE3F4586F022ULL, "one100dg_lia6060519.cl"},
	{0x6674C44191874EC3ULL, "vxlt_backalley_g_1y9_0_0.vlm"},
	{0x66752E6783675396ULL, "hlac_w_blightlands_0_7x.dds"},
	{0x667552E73528971AULL, "vxlt_gallowstemplar_f_5cs_0_0.vlm"},
	{0x6675BFC7C2294E3EULL, "dae000ip_crowd_lm3_talk1.utp"},
	{0x6675D5BFFA4052B4ULL, "vxlt_undercityencounter_13a_0_1.vlm"},
	{0x6676C978F309587EULL, "demo_montage_old.cut"},
	{0x66772DD528B7CDFEULL, "dr_primevalentrance_1726.rml"},
	{0x66773DEAD91313BDULL, "zz_vrc_debug6064267.cl"},
	{0x667814268F763186ULL, "6188643_m.fxe"},
	{0x667825BF501A53FBULL, "zz_vault_debug6203221.cl"},
	{0x66782CE2AF43AF50ULL, "6020547_m.fxe"},
	{0x66786592298F3240ULL, "599835735.ogg"},
	{0x66790B371A6A8280ULL, "w_bonepitdestroyed_a3_36_0_ll.mmh"},
	{0x6679C25926AF9048ULL, "hlac_ca_cavern1_haunted_24427_4y.dds"},
	{0x6679DEDCF3251063ULL, "hlac_w_encounter2_196610_0z.dds"},
	{0x6679FFBA8C3E4684ULL, "dr_primevaltomb_9621.rml"},
	{0x667A601007BF3E2DULL, "281556543.ogg"},
	{0x667A88D9DF005AEDULL, "one140cr_charade6173696.cl"},
	{0x667B7ECDECCF7D51ULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn.msh"},
	{0x667B98CDECCFA90FULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn.mmh"},
	{0x667BEA03255ACF25ULL, "6126419_m.fxe"},
	{0x667BED84212AF7B5ULL, "453756638.ogg"},
	{0x667C4517E96994C3ULL, "6206146_m.fxe"},
	{0x667C47A2A917C99DULL, "6201537_f.fxe"},
	{0x667C7DADA22FD22CULL, "6000765_m.fxe"},
	{0x667C7F0CD4C3BB8FULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv3.evt"},
	{0x667CE1429ECDEE0BULL, "smoke_water.mao"},
	{0x667D15145A356FBEULL, "vxht_0_hightown_a2_c3e_0_0.vlm"},
	{0x667D24873ABF42F8ULL, "vxht_keep_n_4mi_0_0.vlm"},
	{0x667E6AEA392C869DULL, "levelparty.ncs"},
	{0x667E8F29E3DADE67ULL, "6136449_m.fxe"},
	{0x667EB26738D3D0A7ULL, "vxw_sundermount_35b_0_0.vlm"},
	{0x667EE161868A3E20ULL, "889407336.ogg"},
	{0x667F187C9C28DF08ULL, "qun131_taskmaster6059007.cl"},
	{0x667F325DEBEA579AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_enter.evt"},
	{0x667F40E0F28759BAULL, "481687343.ogg"},
	{0x667F5D8DBD3ED88CULL, "6177243_m.fxe"},
	{0x667FA8740AE05749ULL, "6102691_m.fxe"},
	{0x668022031F401767ULL, "826848177.ogg"},
	{0x6680A6B62428C5EDULL, "287016757.ogg"},
	{0x6680A9F936FD5D2EULL, "6172956_m.fxe"},
	{0x6680F7EBC1411D59ULL, "49234317.ogg"},
	{0x6681397B8BD0812AULL, "6172893_m.fxe"},
	{0x6681A540E84AD68CULL, "one110_meeran6089857.cl"},
	{0x6681A5C000DFE2E5ULL, "6034532_m.fxe"},
	{0x66823CB9FC970F37ULL, "vxht_0_hightown_7vk_0_0.vlm"},
	{0x668258C10C8E8FCAULL, "rdr111dg_martin6015000.cl"},
	{0x66825957C788F907ULL, "follower_barks6153444.cl"},
	{0x6682B69AF173EDE3ULL, "489779520.ogg"},
	{0x66833347C8147C78ULL, "vxw_bonepitdestroyed_a3_1wn_0_0.vlm"},
	{0x6683447946DE705EULL, "vxca_cavern1_haunted_ei4_0_0.vlm"},
	{0x6683690A5087EF0EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww22_basic.phy"},
	{0x66836CDD8F8A628FULL, "hlac_ca_caverns2_5802_4x.dds"},
	{0x6683800CD4C9DC38ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv3.gad"},
	{0x6683B54AA92773EAULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_player_scaredmage.gad"},
	{0x6683B8B114A56050ULL, "gen00fl_varric6073886.cl"},
	{0x6683D6C8414B0E30ULL, "70523195.ogg"},
	{0x668406288C24C90EULL, "docks_scene_dockworkers6176224.cl"},
	{0x668446CD36C2434DULL, "vxw_encounter2_pu_0_0.vlm"},
	{0x66850A1E7473F457ULL, "vxht_keep_40r_0_0.vlm"},
	{0x66858DB021D461F0ULL, "6149578_m.fxe"},
	{0x6685953EEAFE4DB8ULL, "ss_crow_m_1.wbd"},
	{0x6686068528E960B3ULL, "shock_proj_light.mao"},
	{0x6686315DEBF0748DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_enter.gad"},
	{0x66875ACCE6BBC03CULL, "character_character_p117.psh.fxo.dx11"},
	{0x6687794F9D2A88C7ULL, "vxw_bonepit_16n_0_0.vlm"},
	{0x6687F403D3913C76ULL, "752211251.ogg"},
	{0x668932721B796E3BULL, "ave200_party_comment.bnk"},
	{0x668A0FA564890C5DULL, "834462138.ogg"},
	{0x668AA85BFA7C3BB2ULL, "6048428_m.fxe"},
	{0x668B0F46522CA8AEULL, "355162181.ogg"},
	{0x668B96642CB55B78ULL, "vxht_keep_n_4hf_0_0.vlm"},
	{0x668BCD04A98A6614ULL, "499279291.ogg"},
	{0x668BD2B76CEE5A47ULL, "6116048_f.fxe"},
	{0x668C1F32BD14F2CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_exit.evt"},
	{0x668C21ED7A511F0BULL, "96657644.ogg"},
	{0x668C34721B7BA7FBULL, "ave200_party_comment.cnv"},
	{0x668CEC2109E08429ULL, "206132991.ogg"},
	{0x668CFF4B695DA416ULL, "615500786.ogg"},
	{0x668D254767CA3603ULL, "hlac_ht_brothel_n_3370_1x.dds"},
	{0x668D2686C2EF2087ULL, "vxht_mansion2b_3ct_0_0.vlm"},
	{0x668D7E2136209CDDULL, "lowtown_male_drunk_mob6151170.cl"},
	{0x668D849D936D8AC3ULL, "module_core.ncs"},
	{0x668D955F1D13E346ULL, "vxht_keep_n_3b9_0_0.vlm"},
	{0x668D973289F79962ULL, "mcr370_varric6098324.cl"},
	{0x668DA32D23D381C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_gesture_self.ani"},
	{0x668E08641CA98B5AULL, "6096870_m.fxe"},
	{0x668E3D8102C1B12DULL, "6131378_m.fxe"},
	{0x668E4B9CDC028DA9ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar4_.ani"},
	{0x668E89AA82F3B7B6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpilexa.msh"},
	{0x668E93AA82F3C8B4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpilexa.mmh"},
	{0x668E97AA82F3CFC7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpilexa.mao"},
	{0x668EAD80CCB43814ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mede_0n.dds"},
	{0x668F8692F45559FAULL, "6059321_m.fxe"},
	{0x668FCF9E5CCA0BC6ULL, "128779765.ogg"},
	{0x66901C44D1D7A82FULL, "hlac_lt_gallowscourtyard_n_-65538_0x.dds"},
	{0x66901E20B3A7FD29ULL, "gen00fl_merrill6106868.cl"},
	{0x66905C564EAD3A28ULL, "gen_im_trs_gen_lgt_glv_r0.uti"},
	{0x66908828A6A7AB75ULL, "vxw_encounter3_e9_0_0.vlm"},
	{0x6690C5109D263386ULL, "6053473_m.fxe"},
	{0x66921DB7FBC9BD97ULL, "gencr_city_guard_archer_f.utc"},
	{0x66922B3EA1058BBDULL, "6171558_m.fxe"},
	{0x6692C069948EB557ULL, "vxw_encounter1_f0_0_0.vlm"},
	{0x6692CE00BA7E9BC7ULL, "vxw_woundedcoast_7c_0_0.vlm"},
	{0x66933832BD1B3B8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_exit.gad"},
	{0x6693B972515DDA0DULL, "6024124_m.fxe"},
	{0x6694A02D23D99B26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_gesture_self.gad"},
	{0x6694B105D1FEDD91ULL, "mrl000_marethari6068899.cl"},
	{0x6694EC04DA38F348ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1i.msh"},
	{0x6694FE04DA3911D9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1i.mao"},
	{0x66950A04DA392672ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1i.mmh"},
	{0x66950E08CFC99D17ULL, "782556789.ogg"},
	{0x6695669CDC08DA2BULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar4_.gad"},
	{0x669566E6FF48AE74ULL, "vxw_encounter3_6e_0_0.vlm"},
	{0x66957848286C62C2ULL, "qun120_arishok_guard6054848.cl"},
	{0x66958F01BAC32EBDULL, "hlac_w_woundedcoast_0_5y.dds"},
	{0x6695F27200407E6AULL, "hlac_w_sundermount_-327676_0y.dds"},
	{0x6696D701BC9C2AA2ULL, "vxdr_0_deeproads_gf9_0_0.vlm"},
	{0x6696F0F3BEFDF20AULL, "6050205_m.fxe"},
	{0x6697EE97D83D6848ULL, "515456668.ogg"},
	{0x669828FA1E4879BFULL, "dae000_bartrand6016670.cl"},
	{0x66982BAEB43F1233ULL, "hlac_w_blightlands_65535_0x.dds"},
	{0x6698823A309C2DF5ULL, "vxlt_gallowstemplar_n_475_0_0.vlm"},
	{0x6698904AA93990D0ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_player_scaredmage.ani"},
	{0x6698B8EA654084EDULL, "1062434623.ogg"},
	{0x6698D0EED6ABD7A6ULL, "6128380_m.fxe"},
	{0x669930577E23274DULL, "t3_hey.tnt"},
	{0x6699317C3B826FD2ULL, "hlac_lt_0_lowtown_a2_-65539_0x.dds"},
	{0x66993732BD1FA58AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_exit.ani"},
	{0x66994925E9D422D8ULL, "qcr220_orsino6060737.cl"},
	{0x669967FBE3E239A3ULL, "6116044_f.fxe"},
	{0x669A3913646804E7ULL, "vxlt_0_lowtown_n_erf_0_0.vlm"},
	{0x669AFBCEC99AF029ULL, "hlac_ht_keep_4519_13y.dds"},
	{0x669B503209F7C91AULL, "6174194_m.fxe"},
	{0x669B63BFB9ED3588ULL, "vxw_bonepitmining_a2_26h_0_0.vlm"},
	{0x669B9B2D23DFB105ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_gesture_self.evt"},
	{0x669BA05ACAEBC440ULL, "vxca_caverns2_8ow_0_0.vlm"},
	{0x669BAC55111D427FULL, "680547559.ogg"},
	{0x669BB2A913477576ULL, "art\\vfx\\combat\\vfx_c_mer_whirlwind_c.impa.evt"},
	{0x669BDEE3B4CE9DF0ULL, "dae000_arishok6080437.cl"},
	{0x669C2BB01E2FD813ULL, "dae000_orsino6085535.cl"},
	{0x669C4FD279D81C10ULL, "500542178.ogg"},
	{0x669C56604D221BD4ULL, "gen00fl_fenris6126511.cl"},
	{0x669C66A5A8D26E6DULL, "vxw_bonepitmining_a2_16n_0_0.vlm"},
	{0x669CFC0CD5377E4BULL, "79029905.ogg"},
	{0x669CFDEA1BE97875ULL, "549063820.ogg"},
	{0x669D0999CF930591ULL, "6170931_m.fxe"},
	{0x669D0B59119038E3ULL, "dae000_arishok6068526.cl"},
	{0x669D4933E7F3E7F8ULL, "6139472_m.fxe"},
	{0x669D56EF9E6E8889ULL, "campaign_base_ru-ru.cre"},
	{0x669DDE76794D5249ULL, "hlac_ht_mansion2b_818_41z.dds"},
	{0x669E079DE7285E55ULL, "6036516_m.fxe"},
	{0x669E3CDC3EB77615ULL, "725715381.ogg"},
	{0x669E92D386F2D5F0ULL, "gen00fl_anders6125772.cl"},
	{0x669ECCC1CDFF2D43ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_5z.dds"},
	{0x669EE297D26EC4CBULL, "vxlt_0_lowtown_docks_a3_aqn_0_0.vlm"},
	{0x66A028332ACAA15EULL, "vxca_cavern1_haunted_ec6_0_0.vlm"},
	{0x66A08C2B117E3154ULL, "vxlt_shop_eg_0_0.vlm"},
	{0x66A12651BB104C70ULL, "6015828_m.fxe"},
	{0x66A132468684CDD7ULL, "vxw_encounter1_k3_0_0.vlm"},
	{0x66A1A9A9134BD101ULL, "art\\vfx\\combat\\vfx_c_mer_whirlwind_c.impa.gad"},
	{0x66A1E4A7DF3BFB31ULL, "6170733_m.fxe"},
	{0x66A20506A4706CEFULL, "vxlt_0_lowtown_a3_n_c_e2y_0_0.vlm"},
	{0x66A2BA4571E49A56ULL, "darktown_fem_lowlife_mob6174164.cl"},
	{0x66A3A5A33A276054ULL, "atl_itemupgrad_dxt1_dat.xml"},
	{0x66A3C52778E2C546ULL, "hlac_lt_0_lowtown_65538_5y.dds"},
	{0x66A459D4842C5482ULL, "6069398_f.fxe"},
	{0x66A4910259A51646ULL, "narrative_varric_demo6170813.cl"},
	{0x66A4D45BCBEDA6CEULL, "530011985.ogg"},
	{0x66A4DA20C7B2A46BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small.phy"},
	{0x66A4F06ECA11D73BULL, "dae000ip_crowd_em2_talk17.utp"},
	{0x66A4FE0D2A5B2C50ULL, "6169190_m.fxe"},
	{0x66A5167A8405E10CULL, "1027993128.ogg"},
	{0x66A5486E067A4A9AULL, "hlac_ht_keep_n_3732_13y.dds"},
	{0x66A6F8A372DD7B6FULL, "vxw_bonepit_na_0_0.vlm"},
	{0x66A73794B83D3DE6ULL, "738576709.ogg"},
	{0x66A7B0220271A115ULL, "6083523_m.fxe"},
	{0x66A80037D0A8B7EBULL, "llac_lt_0_lowtown_n_0_7.dds"},
	{0x66A89F94510EB78BULL, "hlac_lt_gallowstemplar_2391_0z.dds"},
	{0x66A8AAA91351F1AFULL, "art\\vfx\\combat\\vfx_c_mer_whirlwind_c.impa.ani"},
	{0x66A8B45AC3A07311ULL, "6118270_m.fxe"},
	{0x66A91015C9BDE424ULL, "134433213.ogg"},
	{0x66A917C382770F19ULL, "6078143_m.fxe"},
	{0x66A987772DD74148ULL, "vxlt_darktown_5cr_0_0.vlm"},
	{0x66A995115B5A5CD5ULL, "qcr000_arishok6060142.cl"},
	{0x66A9C5FD68B1802FULL, "qcr220cr_nobleman.utc"},
	{0x66AA4A7B29DFEADDULL, "854651455.ogg"},
	{0x66AA4E999365B10FULL, "6128028_m.fxe"},
	{0x66AA6A3EEBCF6D7FULL, "vxlt_0_lowtown_a3_n_c_ep2_0_1.vlm"},
	{0x66AA6E9530B07660ULL, "812657151.ogg"},
	{0x66AA7F9E14029DD9ULL, "493119611.ogg"},
	{0x66AA9F0D804C0CCAULL, "vxht_0_hightown_a3_cab_0_0.vlm"},
	{0x66AAB92D94851369ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_manhole.phy"},
	{0x66AAE34C42FC9BBFULL, "901199309.ogg"},
	{0x66AB613ECC18CCF3ULL, "dae120ip_at_store1.utp"},
	{0x66AB7C18B726F485ULL, "vxht_keep_n_4mp_0_0.vlm"},
	{0x66ABED0C2BBA2270ULL, "vxht_0_hightown_a2_bwf_0_0.vlm"},
	{0x66AC26E83CB45BD2ULL, "6110992_m.fxe"},
	{0x66AC5D208B4CFCA0ULL, "369391087.ogg"},
	{0x66AC76688972349DULL, "vxht_chantry_n_27h_0_0.vlm"},
	{0x66AC993AA3E7036BULL, "art\\cinematics\\deathblows\\meatgolem\\cs_golem_meatwarr.ani"},
	{0x66AD5E55E7CFE643ULL, "vxw_bonepitmining_a2_1vp_0_0.vlm"},
	{0x66AD664B56845550ULL, "821045978.ogg"},
	{0x66ADE2C1639349B3ULL, "and100_anders6096033.cl"},
	{0x66AE709E5E78ADE1ULL, "vxlt_darktown_5dv_0_0.vlm"},
	{0x66AE88DE9DF6C5F8ULL, "vxw_bonepitdestroyed_a3_1fm_0_0.vlm"},
	{0x66AEA7B1FF006EADULL, "hlac_lt_0_lowtown_a2_n_c_65538_2z.dds"},
	{0x66AF4CC0EEB730BDULL, "lgt311_shw_mer_mettin6156573.cl"},
	{0x66B0091E2CAE2BC5ULL, "zz_qun_debug6028253.cl"},
	{0x66B0453B0DBAB503ULL, "mcr370_meredith6083605.cl"},
	{0x66B06ABE6DCC2250ULL, "209290552.ogg"},
	{0x66B0FB7FF8B2B853ULL, "6061805_f.fxe"},
	{0x66B0FC78933A825EULL, "llac_w_encounter3_-393216_0.dds"},
	{0x66B18AD9447C2E05ULL, "vxw_bonepitdestroyed_a3_1ww_0_0.vlm"},
	{0x66B1AA18D39EB1ACULL, "vxw_sundermount_27j_0_0.vlm"},
	{0x66B21276B1B8B63DULL, "6082076_m.fxe"},
	{0x66B23A7491BA4E3FULL, "gen00fl_varric6021134.cl"},
	{0x66B2A366B6FF9F2EULL, "qun110_petrice6036864.cl"},
	{0x66B2BDB63B8977B6ULL, "w_woundedcoast_90_0.msh"},
	{0x66B2C1EF3AE35DF7ULL, "691939105.ogg"},
	{0x66B2C7B63B8988B4ULL, "w_woundedcoast_90_0.mmh"},
	{0x66B2CBB63B898FC7ULL, "w_woundedcoast_90_0.mao"},
	{0x66B2FD195D4F5273ULL, "vxht_0_hightown_c30_0_0.vlm"},
	{0x66B3322EEB110244ULL, "78897670.ogg"},
	{0x66B34019D802AFE1ULL, "mcr370_templar16087625.cl"},
	{0x66B362ECDBA3930CULL, "llac_w_encounter3_-196611_0.dds"},
	{0x66B371D14821E182ULL, "vfx_vfx_p1.psh.fxo.dx11.assembly"},
	{0x66B39EDCF08AB350ULL, "66781075.ogg"},
	{0x66B3A085DB52373EULL, "lt_0_lowtown_a3_n_c_2_0.rml"},
	{0x66B3AB4925DD6B99ULL, "llac_lt_0_lowtown_a2_-131072_1.dds"},
	{0x66B436F52D2ABD78ULL, "226350833.ogg"},
	{0x66B45707A8CD98DEULL, "hm_amb_keep_noble.mor"},
	{0x66B467F76C0A4895ULL, "vxht_0_hightown_bfx_0_0.vlm"},
	{0x66B473700FAD267DULL, "gencr_corpse.utc"},
	{0x66B4883C8E205475ULL, "vxht_0_hightown_n_by2_0_0.vlm"},
	{0x66B48DA80F7BE46DULL, "97614383.ogg"},
	{0x66B54C2EF4C164A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_exit.evt"},
	{0x66B64072C58E8094ULL, "89622896.ogg"},
	{0x66B6CAB2149EEC9DULL, "128723624.ogg"},
	{0x66B6D12562800E0BULL, "hlac_ht_mansion2b_818_64y.dds"},
	{0x66B712DE606BFD9BULL, "vxw_encounter3_b2_0_0.vlm"},
	{0x66B771D7BC08EEE9ULL, "llac_w_sundermount_393212_0.dds"},
	{0x66B7BDD209D8856CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sluce_d.dds"},
	{0x66B847799DD96688ULL, "vxw_bonepitmining_a2_28f_0_0.vlm"},
	{0x66B88414AED7F540ULL, "vxca_caverns2_8qu_0_0.vlm"},
	{0x66B8FA5679644862ULL, "vxlt_blackemporium_vo_0_0.vlm"},
	{0x66B9397D66645881ULL, "art\\toolset\\_textures\\iconrgby_da.dds"},
	{0x66B95A918B92566BULL, "hlac_w_blightlands_327679_0z.dds"},
	{0x66B95D2914FA515EULL, "framebuffereffects_screencolorclear_p0.psh.fxo.dx11"},
	{0x66B9B985DB56CD02ULL, "lt_0_lowtown_a3_n_c_2_0.phy"},
	{0x66B9CF2490E24AD4ULL, "6098706_m.fxe"},
	{0x66BA5EE91DB03AF0ULL, "vxht_0_hightown_a2_bxg_0_0.vlm"},
	{0x66BAB116A1D79231ULL, "vxca_caverns1_mines_i1s_0_0.vlm"},
	{0x66BB632EF4C5F7F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_exit.gad"},
	{0x66BBC4DF25C4A56AULL, "796191496.ogg"},
	{0x66BCA967694AB4DFULL, "vxht_mansion2_1gg_0_1.vlm"},
	{0x66BCE27B506EC661ULL, "vxlt_darktown_5iu_0_0.vlm"},
	{0x66BCFABB8FECDE78ULL, "vxw_bonepitdestroyed_a3_1kj_0_0.vlm"},
	{0x66BD61A6593A9A0DULL, "6096773_f.fxe"},
	{0x66BDDE5ACDDB998DULL, "vxlt_0_lowtown_n_e2d_0_0.vlm"},
	{0x66BDFF4505AB80E4ULL, "vxw_sundermount_30r_0_0.vlm"},
	{0x66BE1FC65C90D390ULL, "llac_lt_0_lowtown_docks_a3_-131074_0.dds"},
	{0x66BE317D82B7E13BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_5.pwk"},
	{0x66BE7345F19F0661ULL, "6108583_m.fxe"},
	{0x66BED1432B543984ULL, "w_bonepitmining_a2_46_0.msh"},
	{0x66BEE3432B5458EDULL, "w_bonepitmining_a2_46_0.mao"},
	{0x66BEE7432B545F26ULL, "w_bonepitmining_a2_46_0.mmh"},
	{0x66BF1D8549649474ULL, "vxlt_0_lowtown_a2_epd_0_0.vlm"},
	{0x66BF36CB603F04C1ULL, "hlac_lt_0_lowtown_docks_a3_-131074_1x.dds"},
	{0x66C01BF5C594CA2CULL, "vxw_sundermount_28k_0_0.vlm"},
	{0x66C048738AD51626ULL, "l_dae310ar_lowtown_nt.bnk"},
	{0x66C0803AA3F781F5ULL, "art\\cinematics\\deathblows\\meatgolem\\cs_golem_meatwarr.gad"},
	{0x66C08453A3AECEBDULL, "6085176_m.fxe"},
	{0x66C0C7FC80559361ULL, "6167140_m.fxe"},
	{0x66C10FF1DE1CED3CULL, "dr_primevaltomb_664.rml"},
	{0x66C122CABF135EE5ULL, "729018440.ogg"},
	{0x66C13104AD17689FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_lp.gad"},
	{0x66C1EC9378FAB2CEULL, "mrl150_dalish_guard6089417.cl"},
	{0x66C2642EF4CC179FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_exit.ani"},
	{0x66C28D53215EDAD4ULL, "vxlt_gallowscourtyard_77b_0_0.vlm"},
	{0x66C2B0842E6A3843ULL, "6104545_m.fxe"},
	{0x66C2EB096D3F0B7EULL, "vxdr_primevaltomb_7yr_0_0.vlm"},
	{0x66C2F9F0F4F59422ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_003a_n.dds"},
	{0x66C2FA1980166CF5ULL, "vxht_0_hightown_n_bd1_0_0.vlm"},
	{0x66C2FF895144E0B3ULL, "6207938_m.fxe"},
	{0x66C39D9B5C0796C3ULL, "mag170_thrask6020538.cl"},
	{0x66C3D320BFDD936DULL, "304241775.ogg"},
	{0x66C4680A50BF84D1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww22_basic.mmh"},
	{0x66C4740A50BF9972ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww22_basic.mao"},
	{0x66C4820A50BFB08FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww22_basic.msh"},
	{0x66C5171EB12920AEULL, "vxca_caverns2prc_9hp_0_0.vlm"},
	{0x66C5945D89220D70ULL, "vxca_caverns2_8p6_0_0.vlm"},
	{0x66C59814A78CA411ULL, "6112070_m.fxe"},
	{0x66C5B2851FDCB83AULL, "hlac_lt_0_lowtown_n_-196608_1z.dds"},
	{0x66C60B318103166DULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4whitelyrium.msh"},
	{0x66C6193181032DA0ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4whitelyrium.mao"},
	{0x66C61D31810334F3ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4whitelyrium.mmh"},
	{0x66C66B3111C37248ULL, "vxlt_darktown_5ip_0_0.vlm"},
	{0x66C6C1B5F39EEE92ULL, "1018429485.ogg"},
	{0x66C7BFD17F404230ULL, "gen_im_npc_dwf_arm_boo_meda.uti"},
	{0x66C8071603ACDB71ULL, "vxdr_0_deeproads_ceo_0_0.vlm"},
	{0x66C8A8DA2CA75D5AULL, "6208674_f.fxe"},
	{0x66C91E77231F9332ULL, "vxw_woundedcoast_1b4_0_0.vlm"},
	{0x66C950EFE80B1B03ULL, "one280_emrys6120034.cl"},
	{0x66C97A18430E0040ULL, "6105700_f.fxe"},
	{0x66CA1D782EC2B273ULL, "vxlt_gallowstemplar_416_0_0.vlm"},
	{0x66CB231D1997F8BBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2b.mmh"},
	{0x66CB2F1D19980D58ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2b.mao"},
	{0x66CB411D19982BA5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2b.msh"},
	{0x66CB51D70598E480ULL, "340097726.ogg"},
	{0x66CB54584264DEE1ULL, "vxlt_darktown_5rt_0_0.vlm"},
	{0x66CB6AB8CC383856ULL, "hlac_ht_keep_4519_2z.dds"},
	{0x66CB6C9881E2F6F8ULL, "vxw_bonepitdestroyed_a3_1lk_0_0.vlm"},
	{0x66CBC5BAD22C0FCFULL, "329632062.ogg"},
	{0x66CC53D356EF9DCAULL, "vxlt_gallowstemplar_n_5j0_0_0.vlm"},
	{0x66CC7A0FC7213ED6ULL, "898022265.ogg"},
	{0x66CC9736767E1AEAULL, "vxht_bartrand_n_5vl_0_0.vlm"},
	{0x66CD9BDECCF1315FULL, "6110529_f.fxe"},
	{0x66CE3567BFA8992AULL, "95670469.ogg"},
	{0x66CE4EEF21B2A743ULL, "llac_ca_caverns1_mines_24421_6.dds"},
	{0x66CE661A2B9FBA29ULL, "6135107_m.fxe"},
	{0x66CEA0ACE464699DULL, "6045984_f.fxe"},
	{0x66CEE25DC244BB3DULL, "6162419_m.fxe"},
	{0x66CFB2440AAB5752ULL, "lgt220_shp_mag_guard6149876.cl"},
	{0x66CFE7432B62D2EDULL, "w_bonepitmining_a2_46_0.rml"},
	{0x66D015D77E76BBD4ULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.mmh"},
	{0x66D03D367A3E40C7ULL, "gen_im_npc_wep_shd_quna.uti"},
	{0x66D0E4F1D0B3AF48ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_abma_0s.dds"},
	{0x66D0ED0FC86F243EULL, "6105815_m.fxe"},
	{0x66D0FF301354F354ULL, "vxlt_gallowscourtyard_78c_0_0.vlm"},
	{0x66D14D4A53D3F326ULL, "6032270_m.fxe"},
	{0x66D1BE015276DE2BULL, "vxw_encounter2_camp_bq_0_0.vlm"},
	{0x66D2321C2BE8CA42ULL, "mer120_hubert6017864.cl"},
	{0x66D24004FA799034ULL, "hlac_ht_0_hightown_a3_65536_3x.dds"},
	{0x66D2935D884DF320ULL, "vxdr_0_deeproads_cxe_0_0.vlm"},
	{0x66D33DA326CEE242ULL, "6070104_m.fxe"},
	{0x66D39BD4611E5AF5ULL, "isa111_isabela6067110.cl"},
	{0x66D41BBBE098E724ULL, "6013361_m.fxe"},
	{0x66D467B026EAB6F8ULL, "6126726_m.fxe"},
	{0x66D4C40E37BE4EA7ULL, "6156313_m.fxe"},
	{0x66D4F0B13E498266ULL, "vxlt_0_lowtown_a2_n_gf1_0_0.vlm"},
	{0x66D54C3C909AC542ULL, "qcr220_meredith6041530.cl"},
	{0x66D5D367B9E687CCULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinscol.msh"},
	{0x66D5D967B9E691BEULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinscol.mmh"},
	{0x66D5E567B9E6A655ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinscol.mao"},
	{0x66D5E8EDFD8E8C19ULL, "6153536_m.fxe"},
	{0x66D6049E2A064E6CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_andraste_d.dds"},
	{0x66D60E04AD2989E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_lp.ani"},
	{0x66D66214AADBAF91ULL, "lt_0_lowtown_docks_245_0.mmh"},
	{0x66D66E14AADBC432ULL, "lt_0_lowtown_docks_245_0.mao"},
	{0x66D67C14AADBDC4FULL, "lt_0_lowtown_docks_245_0.msh"},
	{0x66D6A22D94AA268EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_manhole.mmh"},
	{0x66D6AE2D94AA3B25ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_manhole.mao"},
	{0x66D6BC2D94AA537CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff22_manhole.msh"},
	{0x66D6DC713EF6755AULL, "vxht_mansion_n_4x0_0_0.vlm"},
	{0x66D707BD5E1F1A1AULL, "gen00fl_bethany6028813.cl"},
	{0x66D71C4B2447A4B4ULL, "781679332.ogg"},
	{0x66D77BE112CD42E6ULL, "161964368.ogg"},
	{0x66D7EC798531436DULL, "vxdr_0_deeproads_ew3_0_0.vlm"},
	{0x66D8A02A2E958799ULL, "6166433_m.fxe"},
	{0x66D8AD09DE9D59C6ULL, "953125975.ogg"},
	{0x66D8CDC2EEEE59B4ULL, "vxdr_0_deeproads_9id_0_0.vlm"},
	{0x66D9697FF2A487DEULL, "ser120_ghyslain6118427.cl"},
	{0x66D974F2C539370CULL, "vxlt_darktown_4iq_0_0.vlm"},
	{0x66D9A03EBFE2125CULL, "hlac_lt_0_lowtown_docks_-2_6y.dds"},
	{0x66DA7CAEB5DED75AULL, "vxlt_gallowsprison_n_6m2_0_0.vlm"},
	{0x66DA8E116ECF8531ULL, "qcr000_arishok6167001.cl"},
	{0x66DAA6DD804D6DA9ULL, "vxlt_0_lowtown_a3_n_c_ekf_0_2.vlm"},
	{0x66DAB01ECFE21D6BULL, "vxw_encounter3_gm_0_0.vlm"},
	{0x66DADFAE3DCDADE5ULL, "vxlt_gallowstemplar_f_1ij_0_0.vlm"},
	{0x66DB2AC0F9C3E2E0ULL, "759704956.ogg"},
	{0x66DB41EF690CFC61ULL, "53346956.ogg"},
	{0x66DB8DA7DE8705B1ULL, "vxlt_gallowstemplar_f_34w_0_0.vlm"},
	{0x66DBCD3800BC60FEULL, "gen_im_trs_gen_lgt_boo_r1.uti"},
	{0x66DBFB8BDE6E1381ULL, "6072202_m.fxe"},
	{0x66DC28419A1455FFULL, "vxlt_hangedman_n_5z_0_0.vlm"},
	{0x66DC2981238DFE3BULL, "vxlt_undercityserialkiller_1my_0_0.vlm"},
	{0x66DC53563ACAF149ULL, "vxht_0_hightown_a3_bxc_0_0.vlm"},
	{0x66DC9FCFF92493AAULL, "6152235_m.fxe"},
	{0x66DCDD7FE8979616ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.anb"},
	{0x66DD260733459BA6ULL, "vxlt_warehouse_13q_0_0.vlm"},
	{0x66DD4DF2630D5C32ULL, "661156296.ogg"},
	{0x66DDD41DE5AA1C41ULL, "mrl310_merrill6143970.cl"},
	{0x66DEA98166938107ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_c.anb"},
	{0x66DF2BC9E2FF82ADULL, "260936281.ogg"},
	{0x66DF74E309AD0CBCULL, "dae011_wench_mob6147247.cl"},
	{0x66DF8B011B083271ULL, "hlac_lt_hangedman_-1_13x.dds"},
	{0x66DF9485DB775DB6ULL, "lt_0_lowtown_a3_n_c_2_0.mao"},
	{0x66DFA085DB777255ULL, "lt_0_lowtown_a3_n_c_2_0.mmh"},
	{0x66DFA285DB7775CBULL, "lt_0_lowtown_a3_n_c_2_0.msh"},
	{0x66DFC4CF9D34275AULL, "205922356.ogg"},
	{0x66DFEC71EFD6CBFCULL, "6101361_m.fxe"},
	{0x66E002999E0F01FCULL, "342615333.ogg"},
	{0x66E1053A7A440BA0ULL, "vxdr_0_deeproads_cub_0_0.vlm"},
	{0x66E149DCC4FA3EF1ULL, "vxht_chantry_2zh_0_0.vlm"},
	{0x66E16999BB117BF2ULL, "hlac_lt_hangedman_-1_5x.dds"},
	{0x66E16A6EBA1D78AAULL, "656540402.ogg"},
	{0x66E1DF41C2319E73ULL, "vxlt_0_lowtown_docks_aqm_0_0.vlm"},
	{0x66E26EC1C0B271F0ULL, "vxht_0_hightown_bu0_0_0.vlm"},
	{0x66E3315E1D833335ULL, "345387414.ogg"},
	{0x66E3628E303F9AE6ULL, "vxlt_0_lowtown_a2_n_gc2_0_0.vlm"},
	{0x66E3AA5B3EC3505AULL, "gen00fl_fenris6108030.cl"},
	{0x66E3F24ED965F152ULL, "6120844_f.fxe"},
	{0x66E46164DB9ADEA8ULL, "vxht_keep_6fq_0_0.vlm"},
	{0x66E536F4D42BE492ULL, "vxlt_0_lowtown_hap_0_0.vlm"},
	{0x66E53B8EE10289FEULL, "6189023_f.fxe"},
	{0x66E5BEE2C7A458B1ULL, "hlac_w_blightlands_327683_0y.dds"},
	{0x66E5E5B38C0B5398ULL, "hlac_lt_0_lowtown_a3_n_c_1_5y.dds"},
	{0x66E60231070BC432ULL, "vxw_woundedcoast_1h6_0_0.vlm"},
	{0x66E66CD4001395FDULL, "mag330_walter6044828.cl"},
	{0x66E67DE0A21EDB88ULL, "ts_generic_03.gda"},
	{0x66E6D1F3E1872B0BULL, "vxlt_0_lowtown_eow_0_1.vlm"},
	{0x66E6EF87C072E22DULL, "hlac_lt_backalley_n_1_0y.dds"},
	{0x66E7410F531C35BDULL, "gen_im_delete_me_10.uti"},
	{0x66E7430B6A36AC06ULL, "art\\vfx\\combat\\vfx_r_ibl_disappear_imp_p.anb"},
	{0x66E7917E8B7B44ADULL, "406443622.ogg"},
	{0x66E7CF2BE31B3B5CULL, "vxca_caverns1_mines_ghw_0_0.vlm"},
	{0x66E7DCDF33DAB5A7ULL, "vxlt_darktown_2qt_0_0.vlm"},
	{0x66E7E6D7DB7C3F59ULL, "903337534.ogg"},
	{0x66E7FD82659D10CAULL, "6166562_m.fxe"},
	{0x66E80304380A7F56ULL, "6207005_m.fxe"},
	{0x66E8172754ED4861ULL, "gen00fl_anders6121173.cl"},
	{0x66E83163CCAE71D7ULL, "vxlt_hangedman_27c_0_0.vlm"},
	{0x66E886580DFF137FULL, "6088739_m.fxe"},
	{0x66E8B49637BF6394ULL, "hlac_ht_mansion2b_818_21y.dds"},
	{0x66E96031874284ECULL, "vxca_cavern1_haunted_ghw_0_0.vlm"},
	{0x66E9E767FDF0257DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_basic.phy"},
	{0x66E9EB7432AF24F8ULL, "vxlt_0_lowtown_docks_b0z_0_0.vlm"},
	{0x66EA1AE5D91A6FEBULL, "115946975.ogg"},
	{0x66EA92AA834225AFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpilexa.phy"},
	{0x66EAC5332CC109C9ULL, "vxht_0_hightown_a3_bwb_0_0.vlm"},
	{0x66EAE07FE8A3D881ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.mmh"},
	{0x66EB2AA260170435ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_floor_n.dds"},
	{0x66EB752145DFE334ULL, "gen_im_arm_glv_med_act3_03.uti"},
	{0x66EB7DC7D0326DF7ULL, "one100_solivitus6059915.cl"},
	{0x66EC185C0D7D9030ULL, "follower_banter6136649.cl"},
	{0x66ECBF7391E1757AULL, "t3_shd_wht.tnt"},
	{0x66ECD2007036B909ULL, "6167287_m.fxe"},
	{0x66ECFA65FC7C2EFFULL, "208303953.ogg"},
	{0x66ED051CED26B68AULL, "44130790.ogg"},
	{0x66EE2EA4E0F46A47ULL, "vxw_bonepitmining_a2_1jp_0_0.vlm"},
	{0x66EF77176C3A2420ULL, "vxdr_0_deeproads_crc_0_0.vlm"},
	{0x66EFACAB17CB7F66ULL, "276239816.ogg"},
	{0x66EFC016FB37FB50ULL, "407969320.ogg"},
	{0x66F060E087C40E3DULL, "288217009.ogg"},
	{0x66F0C4125E7E6D0BULL, "hlac_lt_gallowstemplar_f_7198_1z.dds"},
	{0x66F0CF32E1B3F180ULL, "6209948_m.fxe"},
	{0x66F12F34FABBBF0EULL, "w_bonepit_77_0.msh"},
	{0x66F13D34FABBD75FULL, "w_bonepit_77_0.mao"},
	{0x66F14934FABBEBFCULL, "w_bonepit_77_0.mmh"},
	{0x66F14B6A0AD6E7F8ULL, "6128526_m.fxe"},
	{0x66F19FC086873821ULL, "296933137.ogg"},
	{0x66F1AF2C3916EA63ULL, "166365877.ogg"},
	{0x66F1B5D6C53A5E32ULL, "6119887_m.fxe"},
	{0x66F2155F94291F18ULL, "vxlt_0_lowtown_docks_a3_cyo_0_0.vlm"},
	{0x66F218D941C36C64ULL, "571622043.ogg"},
	{0x66F220686065D62CULL, "fen310_party_comment6116868.cl"},
	{0x66F363738E865163ULL, "vxlt_gallowstemplar_f_5je_0_0.vlm"},
	{0x66F37FBCFAD22F29ULL, "vxlt_0_lowtown_docks_n_cs8_0_0.vlm"},
	{0x66F38343B08C94EEULL, "ser121_jethann6105706.cl"},
	{0x66F39FC0CCD9C9D8ULL, "vxlt_foundry_n_vj_0_0.vlm"},
	{0x66F3B823268C2DCAULL, "6156107_m.fxe"},
	{0x66F41F3181CB7DF6ULL, "vxlt_0_lowtown_docks_a3_n_d6d_0_0.vlm"},
	{0x66F4740DF901DCB2ULL, "vxw_woundedcoast_1s7_0_0.vlm"},
	{0x66F4F6946BB02002ULL, "dre000_sibling_has_blight6036489.cl"},
	{0x66F5300B56782EA6ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack02.ani"},
	{0x66F5460B6A42EEB1ULL, "art\\vfx\\combat\\vfx_r_ibl_disappear_imp_p.mmh"},
	{0x66F56329449DA15FULL, "vxht_chantry_23m_0_0.vlm"},
	{0x66F5AC99AB2054EBULL, "vxht_mansion2_32g_0_0.vlm"},
	{0x66F5B17CD2DA8AB4ULL, "vxdr_0_deeproads_9gj_0_0.vlm"},
	{0x66F5FE3F64940CF9ULL, "189738762.ogg"},
	{0x66F64108D51153DCULL, "vxca_caverns1_mines_get_0_0.vlm"},
	{0x66F65D5DB0F22929ULL, "vxca_caverns1_k2o_0_2.vlm"},
	{0x66F6F7CA78F37D58ULL, "vxlt_gallowstemplar_f_5rb_0_0.vlm"},
	{0x66F775F857D97CF5ULL, "al_rogue_s02.gda"},
	{0x66F77F1DF66699BEULL, "vxca_cavern1_haunted_ckd_0_0.vlm"},
	{0x66F788865075D80DULL, "vxht_keep_n_7a7_0_0.vlm"},
	{0x66F78A9764399EA9ULL, "vxlt_0_lowtown_a3_n_c_eqh_0_2.vlm"},
	{0x66F7D20E79389D6CULL, "vxca_cavern1_haunted_get_0_0.vlm"},
	{0x66F7E0C06D95DAD4ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcfem_fenrisrom_alt.ani"},
	{0x66F9136EA72F335CULL, "560910607.ogg"},
	{0x66F914F9C770A7DEULL, "6116217_m.fxe"},
	{0x66F918D77E99C58BULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.anb"},
	{0x66F93988238F3CCAULL, "607271409.ogg"},
	{0x66F96A7FCE49FB37ULL, "art\\vfx\\combat\\vfx_r_var_grenade_aoe_c.anb"},
	{0x66F9C429FE4AA885ULL, "vxw_bonepitdestroyed_a3_1lr_0_0.vlm"},
	{0x66F9EF8CC3A76986ULL, "vxlt_0_lowtown_a3_n_c_gky_0_0.vlm"},
	{0x66F9FF4C65E234B1ULL, "485079750.ogg"},
	{0x66FA38F5EAB0116EULL, "493562535.ogg"},
	{0x66FA51753F746164ULL, "6201239_f.fxe"},
	{0x66FA8A9DAC5A2F92ULL, "6077105_m.fxe"},
	{0x66FADC67FDFE85E9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_basic.mao"},
	{0x66FAE141B3997B54ULL, "humanmale_icon.fxa"},
	{0x66FAE867FDFE9982ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_basic.mmh"},
	{0x66FAEA67FDFE9D38ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_basic.msh"},
	{0x66FC26EE38C82453ULL, "1046454096.ogg"},
	{0x66FCA6FABDE27352ULL, "17583837.ogg"},
	{0x66FD12CB18A40524ULL, "6071516_m.fxe"},
	{0x66FD946711FF766DULL, "217377387.ogg"},
	{0x66FE08CC2ABD6108ULL, "6152321_m.fxe"},
	{0x66FE2D96A8E66FF1ULL, "vxht_chantry_2pn_0_0.vlm"},
	{0x66FEC787055821BEULL, "llac_w_woundedcoast_196612_0.dds"},
	{0x66FECCE8E6550B9BULL, "w_bonepit_189_0.phy"},
	{0x66FEDDC06D9BF436ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcfem_fenrisrom_alt.gad"},
	{0x66FEF88137527F18ULL, "6208334_m.fxe"},
	{0x66FFD83895AF6A2EULL, "vxlt_0_lowtown_a3_n_c_efh_0_0.vlm"},
	{0x66FFDC3433A477E1ULL, "562501444.ogg"},
	{0x67004A738B0AFDCCULL, "l_dae310ar_lowtown_nt.wbd"},
	{0x6701C180747E166BULL, "vxlt_warehouse_su_0_0.vlm"},
	{0x6701F80B6BFD0177ULL, "one130_thug6131579.cl"},
	{0x6702180B56829017ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack02.evt"},
	{0x67023C47AF7F4F09ULL, "ser221_quentin6123315.cl"},
	{0x6702776478BF75ABULL, "822155819.ogg"},
	{0x6702910E73C19676ULL, "vxlt_0_lowtown_docks_a3_n_d1e_0_0.vlm"},
	{0x670296AAF438840AULL, "vxlt_blackemporium_w4_0_0.vlm"},
	{0x67030DEB58458778ULL, "6147015_m.fxe"},
	{0x670311F9C6F42578ULL, "vxdr_primevaltomb_8vj_0_0.vlm"},
	{0x67036C2D1EE6FCC9ULL, "fen210_tevinter_ambusher6108314.cl"},
	{0x6703D75B239D05E5ULL, "lt_0_lowtown_a2_n_c_20_0.anb"},
	{0x6703D7F6A0C75C85ULL, "6091696_m.fxe"},
	{0x6704193096217080ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_ice_c.mmh"},
	{0x67047651858D7D2CULL, "lgt320_shw_ors_marlein6154746.cl"},
	{0x6704E5CA78D3E21DULL, "6048811_f.fxe"},
	{0x6704E7F66BE7C0E0ULL, "vxdr_primevaltomb_7ei_0_1.vlm"},
	{0x67052F033A8253E7ULL, "mer341_jansen6054871.cl"},
	{0x670545B7C9883462ULL, "1056776774.ogg"},
	{0x670568EE6687980CULL, "art\\characters\\playercharacter\\humanfemale\\glv_meda.msh"},
	{0x67056EEE6687A2FEULL, "art\\characters\\playercharacter\\humanfemale\\glv_meda.mmh"},
	{0x67057AEE6687B695ULL, "art\\characters\\playercharacter\\humanfemale\\glv_meda.mao"},
	{0x6705A9DE2F2DD83AULL, "w_bonepitdestroyed_a3_77_0.msh"},
	{0x6705B0258AC0FEC6ULL, "hlac_lt_darktown_-1_78y.dds"},
	{0x6705B7DE2F2DF073ULL, "w_bonepitdestroyed_a3_77_0.mao"},
	{0x6705BBDE2F2DF680ULL, "w_bonepitdestroyed_a3_77_0.mmh"},
	{0x6705C588A8C58C1AULL, "vxlt_gallowstemplar_f_51e_0_0.vlm"},
	{0x6705F882DC0AC75DULL, "qun131_taskmaster6039778.cl"},
	{0x670624A6CFBD1E3EULL, "vxht_keep_n_703_0_0.vlm"},
	{0x6706294D54D867A9ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.dura.ani"},
	{0x67066AF3235D881EULL, "llac_w_blightlands_262140_0.dds"},
	{0x6706D96BE424CDFCULL, "pro000_varric6001262.cl"},
	{0x6706F19EA291132FULL, "llac_lt_warehouse_n_0_5.dds"},
	{0x67070E3CB555A29CULL, "6174185_m.fxe"},
	{0x67071A40D440D761ULL, "w_sundermount_203_0.phy"},
	{0x67074B2F410B4A06ULL, "147109963.ogg"},
	{0x670756D603330EF4ULL, "vxlt_0_lowtown_a2_eoi_0_0.vlm"},
	{0x67075C9BA6D199D7ULL, "435273973.ogg"},
	{0x6707826A3CA85186ULL, "6084381_m.fxe"},
	{0x6707A68166B68108ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_c.mmh"},
	{0x6707F3FA67589F68ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_rug_d.dds"},
	{0x6708B15FD5AF74C4ULL, "vxlt_foundry_n_yo_0_0.vlm"},
	{0x6708BC6388F2066CULL, "lt_0_lowtown_docks_n_107_0.phy"},
	{0x6709190B5688B0A0ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_attack02.gad"},
	{0x670942D78A52F1CEULL, "699127288.ogg"},
	{0x67094384BBE67766ULL, "qcr000_varric6208219.cl"},
	{0x670982506BCD7C36ULL, "qcr220_grey_wardens6060450.cl"},
	{0x6709E524514D6124ULL, "vxlt_hangedman_n_27k_0_0.vlm"},
	{0x670A092537F44B80ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_painting_n.dds"},
	{0x670B35CFE5DE83E5ULL, "655099546.ogg"},
	{0x670B394323627EAAULL, "vxht_keep_n_3b5_0_0.vlm"},
	{0x670B914AEF4914A2ULL, "vxlt_warehouse_n_w9_0_0.vlm"},
	{0x670C0CFD433C29ADULL, "6121555_m.fxe"},
	{0x670C8B9EB91342F2ULL, "hlac_ht_mansion2_818_53x.dds"},
	{0x670CDB5A6113E4ECULL, "llac_ca_caverns1_mines_24424_28.dds"},
	{0x670D444D54DEB42BULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.dura.gad"},
	{0x670DCCDAA4F7C354ULL, "818332641.ogg"},
	{0x670DE823DA7B0583ULL, "ht_0_hightownexplosion_n_43_0.rml"},
	{0x670E0004AEDE3877ULL, "vxlt_undercityserialkiller_1mm_0_0.vlm"},
	{0x670EE841BEE1644EULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.impa.ani"},
	{0x670F126BD552C7A4ULL, "967480097.ogg"},
	{0x670F1A00662B6FD4ULL, "vxht_mansion2_2b1_0_0.vlm"},
	{0x670FE566BED8E2C0ULL, "zz_vault_debug6203329.cl"},
	{0x670FF60140190956ULL, "216618301.ogg"},
	{0x6710335D66742EEBULL, "vxlt_warehouse_tt_0_0.vlm"},
	{0x6710C1B1960E6703ULL, "220409659.ogg"},
	{0x6710E0697DBEF986ULL, "hlac_lt_gallowscourtyard_n_-65537_0x.dds"},
	{0x6710FC30962BC99FULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_ice_c.anb"},
	{0x671137E6C4E37342ULL, "6163774_m.fxe"},
	{0x671183D6B8EA3DF8ULL, "vxdr_primevaltomb_8qk_0_0.vlm"},
	{0x67123AA2BDCD7D03ULL, "908028585.ogg"},
	{0x6712441D19D433F0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2b.phy"},
	{0x671249D392BD7505ULL, "6096796_m.fxe"},
	{0x6712D6F8BEB90A75ULL, "54911294.ogg"},
	{0x6712FB7E5387CE0AULL, "6149030_f.fxe"},
	{0x6713414D54E31A90ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.dura.evt"},
	{0x67137F1C9272A06FULL, "542052274.ogg"},
	{0x67138EF1E872ECCDULL, "283760296.ogg"},
	{0x6713A030BF7265D9ULL, "sky_lt_0_lowtown_n.mao"},
	{0x671407A41AC5957EULL, "6103358_m.fxe"},
	{0x671417D58BB6A73FULL, "vxca_cavern1_haunted_ghx_0_0.vlm"},
	{0x6714D087304C956AULL, "vxht_bartrand_n_5mi_0_0.vlm"},
	{0x6714F51D685CB8C8ULL, "llac_lt_0_lowtown_a2_n_c_196607_1.dds"},
	{0x6715D747F2B6E6D3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishock_movement.anb"},
	{0x671646EA1207971CULL, "6031996_m.fxe"},
	{0x6716B067BA1DE359ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinscol.phy"},
	{0x6717233CC7A58D44ULL, "vxlt_foundry_n_tl_0_0.vlm"},
	{0x67175BD4FEAE5188ULL, "qun220cr_javaris6071888.cl"},
	{0x671768FBF2479A73ULL, "ht_0_hightown_n_4_0.rml"},
	{0x671774F1D9AD428CULL, "227996704.ogg"},
	{0x67178462901B7B18ULL, "llac_dr_0_deeproads_5842_1.dds"},
	{0x6717AD0FFAA1DFBCULL, "6082576_m.fxe"},
	{0x6717B5F6B29023F2ULL, "347950440.ogg"},
	{0x67182EB93DFF13E8ULL, "6097591_f.fxe"},
	{0x67184305AF16D685ULL, "vxlt_0_lowtown_docks_b0c_0_0.vlm"},
	{0x67184F744B1CA9E5ULL, "6111260_m.fxe"},
	{0x671888DA181B1398ULL, "gen00fl_varric6075651.cl"},
	{0x6718CF77E816C280ULL, "zz_followers_banter6158247.cl"},
	{0x671A624D148E74E6ULL, "6122140_m.fxe"},
	{0x671B931BB41DE080ULL, "117617274.ogg"},
	{0x671BB521D31B0B4AULL, "6189015_m.fxe"},
	{0x671BC9629162369BULL, "835708464.ogg"},
	{0x671BED66CD53BC98ULL, "guardian_indomitable.dds"},
	{0x671BEDF3B0B3FD8BULL, "504398709.ogg"},
	{0x671C3F6497B46869ULL, "tbe100_xenon_urchin.bnk"},
	{0x671CBAD90EAF7923ULL, "311036348.ogg"},
	{0x671CBBF2799B9B2EULL, "vxlt_0_lowtown_a3_n_c_elj_0_0.vlm"},
	{0x671CDBB7C99C6324ULL, "27487617.ogg"},
	{0x671CE041BEED939FULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.impa.evt"},
	{0x671D8491F1D96FFBULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_encn_0d.dds"},
	{0x671D8BDD58218854ULL, "vxht_mansion2_2o6_0_0.vlm"},
	{0x671EA53A586A476BULL, "vxlt_warehouse_yw_0_0.vlm"},
	{0x671ED371C80845C8ULL, "679823311.ogg"},
	{0x671F17DA9332B6A5ULL, "6154912_f.fxe"},
	{0x671F216497B66AA9ULL, "tbe100_xenon_urchin.cnv"},
	{0x671F4341319BFD50ULL, "447255366.ogg"},
	{0x6720145FCB891BF1ULL, "hlac_lt_gallowstemplar_f_2384_3x.dds"},
	{0x67204323AA905AA5ULL, "862571644.ogg"},
	{0x6720C7F17072C382ULL, "llac_ca_cavern1_haunted_24422_6.dds"},
	{0x672124882F1E5658ULL, "gen_im_trs_tmp_med_glv_r2.uti"},
	{0x672147A3AA79EC0FULL, "mag110_arianni6016055.cl"},
	{0x672219297A0A5200ULL, "614575059.ogg"},
	{0x67224D614ED5C6D8ULL, "vxlt_gallowstemplar_f_5ca_0_0.vlm"},
	{0x6722677FCE6CFB38ULL, "art\\vfx\\combat\\vfx_r_var_grenade_aoe_c.mmh"},
	{0x672279810CBBADFEULL, "6104258_m.fxe"},
	{0x6722D491E019759DULL, "hlac_ca_caverns1_24422_1y.dds"},
	{0x6722EB708C2FE0CEULL, "796322650.ogg"},
	{0x6723C141BEF37D88ULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.impa.gad"},
	{0x6723F690E2FED59FULL, "vxdr_0_deeproads_bpv_0_1.vlm"},
	{0x672430D7716AD1BFULL, "6071616_m.fxe"},
	{0x67244333BBF09E1EULL, "vxlt_gallowsprison_n_6kp_0_0.vlm"},
	{0x6724A7CD22041181ULL, "235575215.ogg"},
	{0x6725093A8AD563C7ULL, "vxw_bonepitmining_a2_274_0_0.vlm"},
	{0x672518224CCFE0B8ULL, "vxlt_0_lowtown_a3_n_c_gx2_0_0.vlm"},
	{0x6725C99079D52389ULL, "6129263_m.fxe"},
	{0x6725F42EEED4F4CAULL, "6057345_m.fxe"},
	{0x6726B4E2A10CEF05ULL, "vxlt_0_lowtown_docks_b7b_0_0.vlm"},
	{0x6727063929CD3609ULL, "vxw_bonepitdestroyed_a3_p9_0_0.vlm"},
	{0x6727528314EC0A46ULL, "fen110_fenris6100242.cl"},
	{0x6727722715D254D9ULL, "93100116.ogg"},
	{0x672787969784D601ULL, "m_fen200ar_slaver_ambush2.bnk"},
	{0x67279C5E889A04FAULL, "art\\characters\\weapons\\shields\\w_shd_lrda.msh"},
	{0x6727AA5E889A1C33ULL, "art\\characters\\weapons\\shields\\w_shd_lrda.mao"},
	{0x6727AE5E889A2340ULL, "art\\characters\\weapons\\shields\\w_shd_lrda.mmh"},
	{0x6727F2BAEF78AF62ULL, "mrl210_merrill6102126.cl"},
	{0x6727F618A8CCCC47ULL, "vxw_bonepitmining_a2_1vl_0_0.vlm"},
	{0x672866E00793D1C0ULL, "mcr370_negotiation6095516.cl"},
	{0x6728A8ADCC42987AULL, "mag320_mage_conspirator6040907.cl"},
	{0x67292BD872D63B8CULL, "spt_branch_low.psh.fxo"},
	{0x6729A79DAEC5117AULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_up_c.anb"},
	{0x6729EFF5481C8256ULL, "997373347.ogg"},
	{0x672A55F376CAD826ULL, "hlac_lt_0_lowtown_docks_n_65534_4x.dds"},
	{0x672A560036EA0483ULL, "gen_im_trs_bst_spd_gland_03.uti"},
	{0x672B0F64B2E835C0ULL, "6115401_m.fxe"},
	{0x672B2DCF6B91B3AEULL, "vxlt_0_lowtown_a3_n_c_ewk_0_0.vlm"},
	{0x672B45BCEA7A374EULL, "hlac_lt_0_lowtown_a3_n_c_-196609_0y.dds"},
	{0x672BB3E8E67BCF3AULL, "w_bonepit_189_0.msh"},
	{0x672BC1E8E67BE773ULL, "w_bonepit_189_0.mao"},
	{0x672BC5E8E67BED80ULL, "w_bonepit_189_0.mmh"},
	{0x672D01A0F1F216E5ULL, "6121036_m.fxe"},
	{0x672D116092C64436ULL, "6156033_m.fxe"},
	{0x672D487CEA6350DAULL, "805702325.ogg"},
	{0x672DECFECAF0C0D1ULL, "6066412_m.fxe"},
	{0x672E3D7803718CC9ULL, "llac_dr_0_deeproads_5843_5.dds"},
	{0x672E5AA7340B2D28ULL, "6100317_m.fxe"},
	{0x672E79B6B3886EFDULL, "vxca_cavern1_haunted_hvk_0_0.vlm"},
	{0x672EBCFAE058379BULL, "6139174_m.fxe"},
	{0x672F101C55ADCD3DULL, "136330593.ogg"},
	{0x672FA5B8CC881033ULL, "gen00fl_fenris6107234.cl"},
	{0x67303D8D41CA0A60ULL, "vxdr_primevaltomb_7tv_0_1.vlm"},
	{0x6730E211FF2A1E57ULL, "6148167_m.fxe"},
	{0x673127F72A5C4A76ULL, "mag101_grace6053066.cl"},
	{0x67319D288E4B694DULL, "vxw_encounter2_4i_0_0.vlm"},
	{0x6731C8B7CC8AEAD4ULL, "vxw_bonepitdestroyed_a3_17b_0_0.vlm"},
	{0x6731F103FC9F430AULL, "vxht_mansion2b_28a_0_0.vlm"},
	{0x67324571B9124F59ULL, "6005763_m.fxe"},
	{0x67326E214D1A9932ULL, "t3_dog_or3.tnt"},
	{0x6733217BCA63DF66ULL, "6075767_m.fxe"},
	{0x6734CD0C98946828ULL, "genip_bear_trap.utp"},
	{0x67358D875F0EEB60ULL, "263136139.ogg"},
	{0x6735C7E98AA0ACF3ULL, "ser221_gascard6111751.cl"},
	{0x67360B505103D496ULL, "w_woundedcoast_42_0.rml"},
	{0x67364D3ECA986DA5ULL, "vxca_caverns1_mines_jgn_0_0.vlm"},
	{0x67364E9FB871DAC1ULL, "6103331_m.fxe"},
	{0x6737D258E627EBDBULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.dura.evt"},
	{0x6739572F10B998B3ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_6z.dds"},
	{0x673A408F63CB015CULL, "hlac_lt_0_lowtown_docks_a3_n_-262143_0y.dds"},
	{0x673A490604CB779AULL, "dae000cr_flemeth2.utc"},
	{0x673A84FA885FC138ULL, "42314457.ogg"},
	{0x673B939496B826A9ULL, "4040562.ogg"},
	{0x673BE22F6052FCD7ULL, "lt_0_lowtown_a2_n_mmap.dds"},
	{0x673C6B85847AD788ULL, "gen_im_trs_apo_lgt_cht_r2.uti"},
	{0x673CFAF6A34C4A1FULL, "582028679.ogg"},
	{0x673DD106D232A170ULL, "157594542.ogg"},
	{0x673DD85B23CE7FEDULL, "lt_0_lowtown_a2_n_c_20_0.rml"},
	{0x673EC4D2C409B17EULL, "llac_lt_0_lowtown_0_1.dds"},
	{0x673ECB58E62DFE1CULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.dura.gad"},
	{0x673F005FEECBF1CCULL, "hlac_dr_primevaltomb_659_2x.dds"},
	{0x673F347FA2012165ULL, "mi_cyc_f.mao"},
	{0x673F5E75E4A37E67ULL, "139653682.ogg"},
	{0x673FB3C0E997BEB2ULL, "418603352.ogg"},
	{0x673FE71C9CAB8277ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_isaa_0s.dds"},
	{0x6740335D3FAB2DDEULL, "vxca_caverns2_73g_0_1.vlm"},
	{0x67403A94BE810354ULL, "vxw_bonepitdestroyed_a3_18c_0_0.vlm"},
	{0x67406261AA85B64DULL, "vxw_blightlands_cs_0_0.vlm"},
	{0x674062E0EE955B8AULL, "vxht_mansion2b_25b_0_0.vlm"},
	{0x67417330001FC30FULL, "vxht_0_hightown_a6e_0_0.vlm"},
	{0x67417BF12B5A9A9FULL, "vxw_woundedcoast_1gj_0_0.vlm"},
	{0x6741D238CBFAA707ULL, "vxlt_gallowstemplar_f_3mz_0_0.vlm"},
	{0x674212E4FB38BF1CULL, "vxlt_0_lowtown_docks_b69_0_1.vlm"},
	{0x67421F7CAE2DF310ULL, "6208316_m.fxe"},
	{0x6742694FEADF2BCEULL, "6027669_m.fxe"},
	{0x6742A0E78600FA73ULL, "ave121_jeven6105695.cl"},
	{0x6742B15ACC809A61ULL, "107777772.ogg"},
	{0x67434D700DDBC0A0ULL, "gen00fl_dog6167656.cl"},
	{0x67438C171FDEF9CBULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_aoe_p.impa.evt"},
	{0x6743C1414B3306CFULL, "hlac_lt_0_lowtown_n_65536_8x.dds"},
	{0x6743D2A368192FDEULL, "6079958_m.fxe"},
	{0x6743EA50CA9B8260ULL, "hlac_ca_caverns1_mines_24425_10x.dds"},
	{0x67440F13AC51D820ULL, "ht_0_hightown_a3_10_0.rml"},
	{0x674479D5AD3E05ECULL, "gen00fl_aveline6113987.cl"},
	{0x6744FADE01BD4975ULL, "vxht_0_hightown_n_b78_0_0.vlm"},
	{0x67451289AE1B38EDULL, "319144120.ogg"},
	{0x6745294716D1B51BULL, "vxht_0_hightown_n_9us_0_0.vlm"},
	{0x67455D0212974CC3ULL, "530102217.ogg"},
	{0x67457150295DFA8EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_black_d.dds"},
	{0x6745CA58E6341B1AULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.dura.ani"},
	{0x674645EE66BEF399ULL, "art\\characters\\playercharacter\\humanfemale\\glv_meda.phy"},
	{0x674688BA0F8DB0D3ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv2.evt"},
	{0x6746D4CD260AED9EULL, "zz_followers_banter6158586.cl"},
	{0x6746E0E21BE9807AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerb_d.dds"},
	{0x6746F8A555695A0EULL, "403434763.ogg"},
	{0x67471F83A9922B97ULL, "w_bonepitdestroyed_a3_mmap.dds"},
	{0x674753DE98D88AA9ULL, "504453596.ogg"},
	{0x674757F9B1DB201DULL, "hm_lord_harimann.mor"},
	{0x674779A6E4E5E734ULL, "80733399.ogg"},
	{0x6747C77876B69A77ULL, "vxlt_undercityserialkiller_1qi_0_0.vlm"},
	{0x67481CF6052759ABULL, "vfx_distortionmask_p1.vsh.fxo.dx11"},
	{0x67487FEF2ABE75A6ULL, "lowtown_male_alienage2_mob6166687.cl"},
	{0x6748EDCBD3F04ABAULL, "llac_lt_0_lowtown_docks_65535_4.dds"},
	{0x6749936389295737ULL, "lt_0_lowtown_docks_n_107_0.mmh"},
	{0x6749976389295E44ULL, "lt_0_lowtown_docks_n_107_0.mao"},
	{0x6749A41D52AA3E73ULL, "w_encounter1_50_0.phy"},
	{0x6749A96389297CA9ULL, "lt_0_lowtown_docks_n_107_0.msh"},
	{0x6749F0497BBA8DB2ULL, "mag160_varian_ilithis6018897.cl"},
	{0x674AA5171FE5428CULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_aoe_p.impa.gad"},
	{0x674B2A53C99FFE26ULL, "gen00fl_isabela6079288.cl"},
	{0x674B4501D794D512ULL, "sky_lt_0_lowtown_a3_n_c.mao"},
	{0x674B7D3D5F490A18ULL, "mag311_nyssa6042122.cl"},
	{0x674C053182B33A28ULL, "6062800_m.fxe"},
	{0x674C689027651528ULL, "6161583_m.fxe"},
	{0x674C6E2B5138646DULL, "508998860.ogg"},
	{0x674CA9163ED54D82ULL, "833909695.ogg"},
	{0x674CAA8823660AF3ULL, "hlac_ht_mansion2b_818_41x.dds"},
	{0x674CB08E7172C0A1ULL, "6083080_m.fxe"},
	{0x674CBAE5FD7B092DULL, "6041297_m.fxe"},
	{0x674CE031480957B9ULL, "75073650.ogg"},
	{0x674D81BA0F93C314ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv2.gad"},
	{0x674D87356D0E9B49ULL, "gen_im_home_clothing.uti"},
	{0x674E031805CC2800ULL, "6076540_m.fxe"},
	{0x674ED43E9C7BCECDULL, "vxw_blightlands_dr_0_0.vlm"},
	{0x674F2F5D972DB71EULL, "llac_dr_dungeon_3207_2.dds"},
	{0x674FB2762407968DULL, "mrl210_merrill6166868.cl"},
	{0x67508795BA890615ULL, "vxdr_0_deeproads_9c3_0_0.vlm"},
	{0x67508E973931BF2CULL, "vxw_sundermount_2na_0_0.vlm"},
	{0x6750A4171FE9AC8AULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_aoe_p.impa.ani"},
	{0x6750B0FB8A212324ULL, "vxht_bartrand_n_4sd_0_0.vlm"},
	{0x6752AA9DAEE81B2DULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_up_c.mmh"},
	{0x6752D25B2350719BULL, "follower_banter6170152.cl"},
	{0x6752DAA840671253ULL, "vxlt_shop_dc_0_0.vlm"},
	{0x675340C3593C3A84ULL, "gen_im_trs_apo_junk1_04.uti"},
	{0x67535857DF94BE0EULL, "hlac_ca_caverns1_mines_24423_8y.dds"},
	{0x67535F6CC641E9EDULL, "vxht_chantry_36u_0_0.vlm"},
	{0x675490BA0F99FBF2ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv2.ani"},
	{0x6754C911803A65EEULL, "723510415.ogg"},
	{0x6754D18DA3C80EEFULL, "395281119.ogg"},
	{0x6755A887154A5D19ULL, "27503277.ogg"},
	{0x675602B65312BAC4ULL, "760292633.ogg"},
	{0x67563B3175AE8611ULL, "6140293_m.fxe"},
	{0x67567957C417B91AULL, "793070006.ogg"},
	{0x6756D342126AC47FULL, "205973250.ogg"},
	{0x6756D3D8CC5A40E2ULL, "hlac_w_encounter3_0_11y.dds"},
	{0x67572DB30FED319FULL, "mag210_arianni6031660.cl"},
	{0x67574EAFAF096DACULL, "vxw_encounter2_qv_0_1.vlm"},
	{0x67577CC2EF756770ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_4.dds"},
	{0x67578079F1013D11ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.dura.ani"},
	{0x6757939F096AE36BULL, "6149443_m.fxe"},
	{0x67579DD24C91FDACULL, "llac_lt_darktown_-1_11.dds"},
	{0x6757E21F93C3750EULL, "vxht_0_hightown_a3_atg_0_0.vlm"},
	{0x6758C5DC63DCA892ULL, "vxlt_0_lowtown_h9x_0_0.vlm"},
	{0x6759E2B0F202D5F7ULL, "6176755_m.fxe"},
	{0x675A107D75461EAAULL, "art\\characters\\creatures\\mubari\\animation\\turnr_loop.gad"},
	{0x675A3BDC0AC4A92DULL, "6150920_m.fxe"},
	{0x675AF14BDE2FFBB7ULL, "vxw_encounter3_l_0_0.vlm"},
	{0x675B1447C0E23E4CULL, "hlac_dr_dungeon_3214_4z.dds"},
	{0x675B58CD9130C4F0ULL, "222592064.ogg"},
	{0x675B7F558AB1BF30ULL, "vxht_mansion2_1b8_0_2.vlm"},
	{0x675C5B5E01BBB0D2ULL, "vxlt_shop_df_0_0.vlm"},
	{0x675C702EF7A362EFULL, "gen_im_trs_rdr_two_2hm_r2.uti"},
	{0x675CB0E35188F9EBULL, "vxw_encounter3_8d_0_0.vlm"},
	{0x675D770A163D4336ULL, "dae000_arishok6078177.cl"},
	{0x675D98937C2094CBULL, "545379587.ogg"},
	{0x675DB4BF1163DCB5ULL, "6144095_m.fxe"},
	{0x675E0A6A5E7EAB3EULL, "vxlt_0_lowtown_a2_n_c_e87_0_0.vlm"},
	{0x675E469FEFDACAE6ULL, "gen00fl_merrill6129938.cl"},
	{0x675E8B79F1076EB3ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.dura.gad"},
	{0x675F00742B27D7ACULL, "vxw_sundermount_2sf_0_0.vlm"},
	{0x675F18B905DC3BA2ULL, "172646462.ogg"},
	{0x675F67344AFEB309ULL, "6174416_m.fxe"},
	{0x675F767CD8433B08ULL, "335331918.ogg"},
	{0x675FA64DF5CF1391ULL, "vxlt_gallowstemplar_43j_0_0.vlm"},
	{0x676023DB51668D1BULL, "715787136.ogg"},
	{0x676047E2DBE90362ULL, "ht_0_hightown_a2_30_0_lc.mmh"},
	{0x676078D470C82EBBULL, "hlac_ca_caverns1_mines_24421_1y.dds"},
	{0x676091C1683DDF25ULL, "427345859.ogg"},
	{0x6760C89D7400904EULL, "6169336_m.fxe"},
	{0x6760E64B98B7ED5AULL, "lt_gallowstemplar_n.arl"},
	{0x6760F5D44220199FULL, "gen00fl_aveline6139573.cl"},
	{0x67613D61826657CEULL, "451529002.ogg"},
	{0x67629D29B0129745ULL, "hlac_lt_0_lowtown_docks_a3_-196607_0z.dds"},
	{0x6762A1A704D0C8B4ULL, "6059190_m.fxe"},
	{0x6762C4DF035D5F10ULL, "787954705.ogg"},
	{0x676324DAFC167265ULL, "dae300_bran6144863.cl"},
	{0x6763942AE791BAB2ULL, "dae000_tomwise6167270.cl"},
	{0x6763F1C3BDD2B5F2ULL, "hlac_lt_0_lowtown_a2_n_c_131072_6z.dds"},
	{0x6764383CFA484A83ULL, "mer320_hubert6053375.cl"},
	{0x67645B010E63B77DULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_bdy_ggg_0s.dds"},
	{0x67646EEF41E7E9C5ULL, "vxlt_0_lowtown_a2_n_c_by4_0_0.vlm"},
	{0x67648365482E3155ULL, "gen00fl_merrill6162084.cl"},
	{0x67648B7105A447C8ULL, "6179779_m.fxe"},
	{0x6764E2FE6DB2134BULL, "6117186_f.fxe"},
	{0x676549CF1822C757ULL, "hlac_lt_0_lowtown_a2_-65537_5x.dds"},
	{0x676555F2DBBF2EDBULL, "gen00fl_anders6119529.cl"},
	{0x67657431D66E5ADFULL, "568377032.ogg"},
	{0x67658879F10D8818ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.dura.evt"},
	{0x67661F0AC0599357ULL, "vxlt_foundry_n_10l_0_0.vlm"},
	{0x67662D1214C49CF6ULL, "vxht_keep_630_0_0.vlm"},
	{0x67668863C39C548EULL, "1061839765.ogg"},
	{0x6766B8B25509A7B1ULL, "hlac_ht_0_hightown_65537_1y.dds"},
	{0x6766EBED8B7B2B7BULL, "cod_ltr_forgamlen.utp"},
	{0x6767969A880CB5ADULL, "206062452.ogg"},
	{0x6767CCB8CB942280ULL, "vxht_mansion2_3dn_0_0.vlm"},
	{0x6767D2DE9286AD06ULL, "188683314.ogg"},
	{0x6768902B1C75F025ULL, "299983010.ogg"},
	{0x6769300637FE5651ULL, "6124145_m.fxe"},
	{0x6769E634CC451E8AULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_quanarithrone_sa.dds"},
	{0x676A1F40D4955C9DULL, "w_sundermount_203_0.mao"},
	{0x676A231558E4B441ULL, "145317725.ogg"},
	{0x676A2340D49563D6ULL, "w_sundermount_203_0.mmh"},
	{0x676A2D40D49574D4ULL, "w_sundermount_203_0.msh"},
	{0x676A4C825A5BDC99ULL, "401248208.ogg"},
	{0x676AE789B1ED303EULL, "plt_cod_itm_longbow_jackal.plo"},
	{0x676AE84FCEAF3466ULL, "fxa_cen_aoe_volt_mesh.mao"},
	{0x676B22C0437F126BULL, "vxw_encounter3_5g_0_0.vlm"},
	{0x676BB7F88067FFCDULL, "vxw_blightlands_rp_0_0.vlm"},
	{0x676C30272F927695ULL, "vxlt_0_lowtown_n_fe2_0_0.vlm"},
	{0x676C37E57A45916EULL, "6110215_m.fxe"},
	{0x676C3CD323E9D9AEULL, "hlac_ht_mansion2b_818_28x.dds"},
	{0x676C436A63A66E89ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbasic.phy"},
	{0x676C5B58C1B8073EULL, "fen210_orana6109747.cl"},
	{0x676C87DAA068597CULL, "hightown_fem_noble1_sta6169341.cl"},
	{0x676D27CFA1DCF087ULL, "vxlt_gallowstemplar_f_3rw_0_0.vlm"},
	{0x676D72511D1DF02CULL, "vxw_sundermount_2tg_0_0.vlm"},
	{0x676D8C72A5574301ULL, "6019040_f.fxe"},
	{0x676DC9F8FE267206ULL, "hlac_ca_caverns1_mines_24425_11y.dds"},
	{0x676DEB7D75568890ULL, "art\\characters\\creatures\\mubari\\animation\\turnr_loop.ani"},
	{0x676E182AE7C52C11ULL, "vxlt_gallowstemplar_46i_0_0.vlm"},
	{0x676E959697C0F0B3ULL, "m_fen200ar_slaver_ambush2.wbd"},
	{0x676EFAEAF0845C14ULL, "6034924_m.fxe"},
	{0x676FA17242F1D4E4ULL, "200626527.ogg"},
	{0x676FA4BFB7AFF92AULL, "zz_followers_banter6157536.cl"},
	{0x676FB68F2704A7C5ULL, "hlac_lt_0_lowtown_a2_n_-262144_1x.dds"},
	{0x676FEAA91FA641D2ULL, "vxlt_gallowscourtyard_73x_0_0.vlm"},
	{0x677008505135484AULL, "w_woundedcoast_42_0.anb"},
	{0x677065CE1B3ED6FDULL, "6034661_f.fxe"},
	{0x6770C63D0779996AULL, "6108023_m.fxe"},
	{0x67712C609270C2A8ULL, "366744859.ogg"},
	{0x6771B7387345B8D1ULL, "95567401.ogg"},
	{0x67721842A38ED2D6ULL, "art\\characters\\weapons\\unique\\arishok\\w_gax_002a.mao"},
	{0x67722442A38EE775ULL, "art\\characters\\weapons\\unique\\arishok\\w_gax_002a.mmh"},
	{0x67722642A38EEAEBULL, "art\\characters\\weapons\\unique\\arishok\\w_gax_002a.msh"},
	{0x677291E4084890BAULL, "6097824_m.fxe"},
	{0x6773940687C227E5ULL, "vxdr_0_deeproads_9nd_0_0.vlm"},
	{0x6773C1B3F8AECAA0ULL, "mag291_vincento6027335.cl"},
	{0x6773D907F7D64927ULL, "6189076_m.fxe"},
	{0x67744C4CF85733C0ULL, "vxca_caverns2_80f_0_0.vlm"},
	{0x677458C8041ABB02ULL, "6141622_m.fxe"},
	{0x677499D454845875ULL, "6106776_m.fxe"},
	{0x6774BD0DBE58AF27ULL, "clf260_harley6069685.cl"},
	{0x677515526389E540ULL, "521992231.ogg"},
	{0x67755C314EAF169AULL, "zz_lgt_3_debug6176136.cl"},
	{0x677575CD28AF7833ULL, "vxca_caverns2_971_0_0.vlm"},
	{0x6776215A4B2C10ECULL, "hlac_w_bonepitdestroyed_a3_458753_0z.dds"},
	{0x67762721672F7970ULL, "vxht_0_hightown_a2_b9h_0_0.vlm"},
	{0x6776397027968AE3ULL, "qun220cr_javaris6071508.cl"},
	{0x6776444F584B8BEFULL, "507454013.ogg"},
	{0x6777A018283EE4D1ULL, "vxca_caverns1_jft_0_0.vlm"},
	{0x6777EC1D5D83951FULL, "vxlt_hangedman_n_2zw_0_0.vlm"},
	{0x677836E7F96E9F6FULL, "vxlt_0_lowtown_gch_0_0.vlm"},
	{0x6778375A6B87797FULL, "6080362_m.fxe"},
	{0x6778496FD47C25BBULL, "follower_banter6167879.cl"},
	{0x6778C8DC1E5463F9ULL, "6071706_m.fxe"},
	{0x6779146F65F8D3DFULL, "follower_banter6156242.cl"},
	{0x677955FF97A1083AULL, "47949649.ogg"},
	{0x6779B95823D74EC3ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_act_c.impa.ani"},
	{0x6779B9D7F15832E3ULL, "6206065_m.fxe"},
	{0x6779F770AD77B41BULL, "vxht_0_hightown_n_b14_0_0.vlm"},
	{0x677A29D5725E184DULL, "vxw_blightlands_wo_0_0.vlm"},
	{0x677B2ED4C7F3469CULL, "gen00fl_varric6107305.cl"},
	{0x677B95D8C51A4707ULL, "llac_lt_darktown_-1_49.dds"},
	{0x677BE42E0F1408ACULL, "vxw_sundermount_2yd_0_0.vlm"},
	{0x677BEB66293A8303ULL, "6118294_m.fxe"},
	{0x677D252BDF610984ULL, "vxht_0_hightown_bnu_0_0.vlm"},
	{0x677DAC5757ECC14EULL, "6163136_m.fxe"},
	{0x677DB79A5DCBBC0EULL, "6054582_m.fxe"},
	{0x677DC6B8310700A5ULL, "224799671.ogg"},
	{0x677DD52386ED3E8AULL, "hlac_lt_0_lowtown_65536_12y.dds"},
	{0x677E5C86119C5A52ULL, "vxlt_gallowscourtyard_74y_0_0.vlm"},
	{0x677EB59002C4DD37ULL, "hlac_ht_0_hightown_-65534_1y.dds"},
	{0x677EF519F6EA9850ULL, "vxlt_undercityencounter_uy_0_0.vlm"},
	{0x677F062E2B50FF40ULL, "6090265_m.fxe"},
	{0x677F3AD61143DBAEULL, "ser121_jethann6108785.cl"},
	{0x6780A743E6B89F77ULL, "one110_meeran6090191.cl"},
	{0x6780FA2FFDE1753DULL, "hlac_lt_0_lowtown_docks_a3_n_-3_6y.dds"},
	{0x67812094C9867075ULL, "561350498.ogg"},
	{0x67814C36F1F85DB8ULL, "10656602.ogg"},
	{0x6781D4E1C6D5B570ULL, "6111715_m.fxe"},
	{0x678205E379B84065ULL, "vxdr_0_deeproads_9cc_0_0.vlm"},
	{0x6783965E494FFC4FULL, "vxdr_primevaltomb_2w6_0_2.vlm"},
	{0x6783DE240FD3D22CULL, "6001898_m.fxe"},
	{0x6783E3BB24909BCDULL, "mag130cr_slaver2.utc"},
	{0x6783E7AA1AA590B3ULL, "vxca_caverns2_980_0_0.vlm"},
	{0x6783EA6C6DB24DFDULL, "vxlt_undercitydungeon_1ng_0_0.vlm"},
	{0x6784734E8F7F3949ULL, "hlac_w_encounter3_-65534_0x.dds"},
	{0x67847B07F0F5E24AULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_isabella_isabellasreturn.gad"},
	{0x6784B733633E9D87ULL, "vxlt_0_lowtown_docks_n_b59_0_0.vlm"},
	{0x67855E5EBB07B82CULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredidies.ani"},
	{0x67860D71564B7211ULL, "202756797.ogg"},
	{0x678667819BA4DFEBULL, "vxlt_0_lowtown_a2_ere_0_0.vlm"},
	{0x6786701A1C185DE8ULL, "6085505_f.fxe"},
	{0x6786B3279EE37C54ULL, "gen00fl_merrill6130080.cl"},
	{0x67870062945F65CFULL, "fake_rogue_male.utc"},
	{0x67877464973D5B1AULL, "mer320_hubert6052069.cl"},
	{0x6787A3F192A3DE22ULL, "197560903.ogg"},
	{0x6787D15823E3B4BAULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_act_c.impa.evt"},
	{0x6788989404F90F67ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_spear.anb"},
	{0x6788FE656AA2C2D5ULL, "rob_sla.tnt"},
	{0x67893C32E26463EDULL, "6084518_f.fxe"},
	{0x678948268A2B88F9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_walls_n.dds"},
	{0x678A1E4928EC2B98ULL, "mag321_comtesse_de_launcet6043357.cl"},
	{0x678A2AFC07010C09ULL, "hlac_lt_undercitydungeon_-1_46z.dds"},
	{0x678A452733038846ULL, "hlac_ca_caverns2_5802_40y.dds"},
	{0x678B9708D1572204ULL, "vxht_0_hightown_bkz_0_0.vlm"},
	{0x678C32020B811F68ULL, "930233888.ogg"},
	{0x678C6B5EBB0DED6EULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredidies.gad"},
	{0x678CFB939C18B6E8ULL, "vxca_caverns1_hm6_0_0.vlm"},
	{0x678D3D0FB5D9E5F3ULL, "ser221_gascard6123440.cl"},
	{0x678D727E56D15641ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose5.anb"},
	{0x678DB37609520D96ULL, "vxw_woundedcoast_1dv_0_0.vlm"},
	{0x678DD05823E81EADULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_act_c.impa.gad"},
	{0x678FADAF0E92C055ULL, "vxlt_0_lowtown_euf_0_0.vlm"},
	{0x678FC62051C37FD1ULL, "596781331.ogg"},
	{0x679014A24D52934BULL, "plt_lgt000pt_civ_citizens.plo"},
	{0x67904137C4CAA39FULL, "art\\characters\\creatures\\dragonling\\animation\\blendtree\\dragonling_overlay.anb"},
	{0x679069B4782C4CC8ULL, "lt_0_lowtown_docks_a3_n_130_0.msh"},
	{0x679074D513F7D740ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archleft.phy"},
	{0x679077C06BAE58E5ULL, "vxdr_0_deeproads_9db_0_0.vlm"},
	{0x67907BB4782C6B59ULL, "lt_0_lowtown_docks_a3_n_130_0.mao"},
	{0x679087B4782C7FF2ULL, "lt_0_lowtown_docks_a3_n_130_0.mmh"},
	{0x6790CD20AC25EA41ULL, "docks_male_human_worker_sta6175972.cl"},
	{0x6790E59EAC32F8D7ULL, "304722148.ogg"},
	{0x679107493FFE0C12ULL, "1902103.ogg"},
	{0x679188A21C556D9EULL, "lgt250_way_dalish_conflict6147137.cl"},
	{0x6791BCDC9FAD94CAULL, "vxht_0_hightown_a3_c1r_0_0.vlm"},
	{0x6791C1118CA5D834ULL, "168632826.ogg"},
	{0x67923B93E94ABCBAULL, "mcr371_hostile_aveline6082203.cl"},
	{0x6792A7BD47DE4075ULL, "6163562_m.fxe"},
	{0x6792C6E16A7F3674ULL, "gen00fl_aveline_generic_3.stg"},
	{0x679349DCBB9CF94EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsrock.msh"},
	{0x679357DCBB9D109FULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsrock.mao"},
	{0x679362B3DDD5AADEULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.mmh"},
	{0x679363DCBB9D253CULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsrock.mmh"},
	{0x6794112DCC10B2BBULL, "zz_mrl_debug6086517.cl"},
	{0x67949DFE30528531ULL, "mcr370_cut_player_survives6098263.cl"},
	{0x6794D95E8D9AF86BULL, "vxlt_0_lowtown_a2_eud_0_0.vlm"},
	{0x67957A9339343394ULL, "61314034.ogg"},
	{0x679584E5C65EC99DULL, "63470579.ogg"},
	{0x6795B0281249F23EULL, "hlac_lt_0_lowtown_a2_n_c_-65534_0z.dds"},
	{0x67964333DB5E075FULL, "hlac_lt_0_lowtown_65534_5x.dds"},
	{0x6796F6C0D200B1D7ULL, "6050172_m.fxe"},
	{0x67971D655DD50888ULL, "6201650_m.fxe"},
	{0x679771DB84A1A7B6ULL, "vxdr_0_deeproads_9iv_0_0.vlm"},
	{0x67979FC3ECCC114CULL, "vxlt_0_lowtown_a2_n_hei_0_0.vlm"},
	{0x6797A7C3A9C4938CULL, "6118625_f.fxe"},
	{0x6797BF831F38A9BBULL, "hlac_ca_caverns1_24426_10y.dds"},
	{0x6797F079E393947FULL, "vxlt_hangedman_n_ti_0_0.vlm"},
	{0x67982C6A63CB82AEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbasic.mmh"},
	{0x67983239B6D0DBC2ULL, "ch1notnt.psh.fxo"},
	{0x6798386A63CB9745ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbasic.mao"},
	{0x6798466A63CBAE9CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbasic.msh"},
	{0x6798469EA40663CCULL, "vxlt_gallowsprison_n_610_0_0.vlm"},
	{0x6798966092050BDDULL, "275884942.ogg"},
	{0x6798DF712F7A5EFEULL, "392530570.ogg"},
	{0x6798DFF2B3EFD27DULL, "lt_0_lowtown_n_38_0_lc.mmh"},
	{0x6798FB7433D32D41ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_end.ani"},
	{0x6798FD5051583AE3ULL, "w_woundedcoast_42_0.msh"},
	{0x67990B505158523DULL, "w_woundedcoast_42_0.mmh"},
	{0x67990F505158594EULL, "w_woundedcoast_42_0.mao"},
	{0x67995607F107FF30ULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_isabella_isabellasreturn.ani"},
	{0x679963F3FC830109ULL, "follower_banter6170554.cl"},
	{0x679A2A4C0CFDE6F4ULL, "vxht_keep_n_63y_0_0.vlm"},
	{0x679A6CE6516FF868ULL, "lt_0_lowtown_docks_a3_n_36_0.rml"},
	{0x679A80B427487365ULL, "6148363_m.fxe"},
	{0x679A9625D9289351ULL, "742228664.ogg"},
	{0x679B7306A621508FULL, "6094514_m.fxe"},
	{0x679C21D7B3D6DAAAULL, "hlac_lt_undercityserialkiller_-1_30y.dds"},
	{0x679C2552FB482616ULL, "vxw_woundedcoast_1ow_0_0.vlm"},
	{0x679C3ACDC3B066F3ULL, "follower_banter6151320.cl"},
	{0x679C5815AAA09EF3ULL, "vxht_0_hightown_n_c7g_0_0.vlm"},
	{0x679CE6CEF0FA27D6ULL, "vxdr_0_deeproads_cqx_0_0.vlm"},
	{0x679D1AC771E2BFCDULL, "zz_dre_debug6056947.cl"},
	{0x679D3C6C318BE453ULL, "6024548_m.fxe"},
	{0x679DAE74EEC1B0E5ULL, "vxlt_0_lowtown_n_feb_0_0.vlm"},
	{0x679DD23AA062924EULL, "hlac_ca_caverns1_mines_24425_5x.dds"},
	{0x679DD250859BE389ULL, "dae000_arishok6068519.cl"},
	{0x679E1FC891FD191BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_right.mmh"},
	{0x679E2BC891FD2DB8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_right.mao"},
	{0x679E3DC891FD4C05ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_right.msh"},
	{0x67A0067433D95EE3ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_end.gad"},
	{0x67A0392839280DA8ULL, "zz_one_debug6079322.cl"},
	{0x67A041B3DDDFFC9DULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.anb"},
	{0x67A04F8957014BC7ULL, "374617854.ogg"},
	{0x67A07CAB4DD8F9C2ULL, "69806346.ogg"},
	{0x67A0F29D57E5F928ULL, "vxht_mansion_n_60x_0_0.vlm"},
	{0x67A206149AC9D51DULL, "vxht_chantry_n_2bw_0_0.vlm"},
	{0x67A27C3F79EC4466ULL, "one300_murder_followers6166954.cl"},
	{0x67A2AE2DF621BEDCULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_revive.ani"},
	{0x67A319BB681182ABULL, "6208477_f.fxe"},
	{0x67A34B3B7F9110EBULL, "vxlt_0_lowtown_a2_exg_0_0.vlm"},
	{0x67A3BBFFCA64420BULL, "vxw_bonepitmining_a2_pi_0_0.vlm"},
	{0x67A4606277028028ULL, "hlac_lt_darktown_-1_71x.dds"},
	{0x67A4A4BD508713E0ULL, "536136149.ogg"},
	{0x67A4CC7E5BED1470ULL, "vxdr_primevaltomb_3ho_0_0.vlm"},
	{0x67A4DB0DC846689AULL, "hlac_ca_cavern1_haunted_24426_3y.dds"},
	{0x67A502DD12CFD1A5ULL, "130676837.ogg"},
	{0x67A5B34227DA9B8AULL, "160377183.ogg"},
	{0x67A60CD7F2E5C360ULL, "447558500.ogg"},
	{0x67A611A0DEC229CCULL, "vxlt_0_lowtown_a2_n_hhn_0_0.vlm"},
	{0x67A63D8A787327E9ULL, "591910206.ogg"},
	{0x67A648C43D69A7BAULL, "hlac_lt_gallowstemplar_2388_4y.dds"},
	{0x67A66256D589ACFFULL, "vxlt_hangedman_n_sh_0_0.vlm"},
	{0x67A6794586462BAFULL, "mrl000_marethari6129242.cl"},
	{0x67A67A312F7E4407ULL, "vxca_caverns2_96n_0_0.vlm"},
	{0x67A6C00440477E92ULL, "vxca_cavern1_haunted_edo_0_0.vlm"},
	{0x67A6E37433DF41C8ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_end.evt"},
	{0x67A6E982292A0E3DULL, "dre000_hireling16038593.cl"},
	{0x67A6EBEBB49E01AFULL, "vxw_bonepitdestroyed_a3_27b_0_0.vlm"},
	{0x67A6F83590245F4EULL, "llac_lt_0_lowtown_a2_-65535_6.dds"},
	{0x67A7634B69FE19A1ULL, "mcr370_templar16086965.cl"},
	{0x67A803D15AB5A7ECULL, "6048191_f.fxe"},
	{0x67A85AAAE909C46DULL, "hlac_lt_undercityserialkiller_-1_22y.dds"},
	{0x67A8BB2DF626401EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_revive.gad"},
	{0x67A8D49659E9D237ULL, "6103250_m.fxe"},
	{0x67A99FB310589AAFULL, "mer320_hubert6053731.cl"},
	{0x67A9AB2240C062D4ULL, "vxlt_gallowscourtyard_7gr_0_0.vlm"},
	{0x67AA535BD48ABE74ULL, "vxlt_0_lowtown_bxg_0_0.vlm"},
	{0x67AA972FED3E3E96ULL, "vxw_woundedcoast_1jx_0_0.vlm"},
	{0x67AB9D1D52FCFCE8ULL, "w_encounter1_50_0.mmh"},
	{0x67ABA91D52FD108BULL, "w_encounter1_50_0.mao"},
	{0x67ABBB1D52FD2F12ULL, "w_encounter1_50_0.msh"},
	{0x67AC6F693FC6D397ULL, "llac_ht_keep_n_3732_0.dds"},
	{0x67ACB6A5F92DF35AULL, "6205681_m.fxe"},
	{0x67AD152916A3E9DBULL, "hlac_w_bonepit_393219_0x.dds"},
	{0x67AD3514CD46FC95ULL, "6208285_m.fxe"},
	{0x67AD680D77BFC0BDULL, "hlac_dr_slavers_3217_4x.dds"},
	{0x67ADC758AD344A2DULL, "lgt210_magu_selby6142021.cl"},
	{0x67AE165E2DA6F2F9ULL, "gen00fl_isabela6075281.cl"},
	{0x67AE3339EF17CBB7ULL, "vxca_caverns1_mines_gkl_0_1.vlm"},
	{0x67AE3FB564409635ULL, "vxlt_darktown_4f1_0_0.vlm"},
	{0x67AE777FCC93D450ULL, "zz_dre_debug6168317.cl"},
	{0x67AF223158109D37ULL, "24727306.ogg"},
	{0x67AF2C6FB323376CULL, "lt_0_lowtown_docks_43_0.rml"},
	{0x67AF8F5E3FCDD858ULL, "hlac_lt_0_lowtown_docks_65532_1z.dds"},
	{0x67B01CB6DAEB6423ULL, "gen00fl_anders6125926.cl"},
	{0x67B071CBBD59D2F2ULL, "crafting_recipe_types.gda"},
	{0x67B08E2417F7C317ULL, "gen00fl_varric6078815.cl"},
	{0x67B0F4DE1005330FULL, "vxht_mansion_n_2ny_0_0.vlm"},
	{0x67B10FE4743012BFULL, "hlac_lt_gallowstemplar_2391_4z.dds"},
	{0x67B17F90584DFB5DULL, "720332596.ogg"},
	{0x67B1A769CB6A93D4ULL, "vxdr_primevaltomb_86f_0_0.vlm"},
	{0x67B1CDEE11B15D71ULL, "274886952.ogg"},
	{0x67B1CED89B85B225ULL, "6103227_m.fxe"},
	{0x67B1D59236E0ACF2ULL, "6076000_f.fxe"},
	{0x67B1D702C2E21578ULL, "vxw_bonepit_1qh_0_0.vlm"},
	{0x67B26B790823C8D7ULL, "vxlt_hangedman_2mq_0_0.vlm"},
	{0x67B2AD473BBC6C0EULL, "lt_0_lowtown_a2_n_53_0_lc.mmh"},
	{0x67B2B810790A41B7ULL, "hlac_ht_0_hightown_0_0z.dds"},
	{0x67B2CCFED09CAE95ULL, "hlac_lt_0_lowtown_a3_n_c_65538_1x.dds"},
	{0x67B338D8B078120CULL, "vxca_caverns1_mines_hvn_0_0.vlm"},
	{0x67B3451C32D4C5C7ULL, "65963550.ogg"},
	{0x67B38A1B1F99D261ULL, "vxca_caverns1_mines_inb_0_0.vlm"},
	{0x67B393C054E7BDF4ULL, "dre000_dougal6129776.cl"},
	{0x67B5651521BB6F96ULL, "546469516.ogg"},
	{0x67B582C50F7A40DDULL, "hlac_ht_mansion2b_818_48y.dds"},
	{0x67B591D5141724D5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archleft.msh"},
	{0x67B593D51417284BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archleft.mmh"},
	{0x67B59FD514173CE8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archleft.mao"},
	{0x67B6426C37FC76D0ULL, "zz_followers_banter6158988.cl"},
	{0x67B66E41B33A2EABULL, "gen00fl_sebastian6180799.cl"},
	{0x67B69E0DB59ADFC9ULL, "vxht_keep_5dy_0_0.vlm"},
	{0x67B72BAADE56C7EFULL, "vrc321_varric6087267.cl"},
	{0x67B73D09A39809D1ULL, "vxdr_slavers_144_0_0.vlm"},
	{0x67B7640CC10773E4ULL, "llac_w_blightlands_0_3.dds"},
	{0x67B81CFF32B67B54ULL, "vxlt_gallowscourtyard_7hs_0_0.vlm"},
	{0x67B842DCBBBC0907ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsrock.phy"},
	{0x67B85C9EAADDBF2EULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.dura.evt"},
	{0x67B8F739CD9B4A0AULL, "hlac_lt_0_lowtown_docks_n_196607_2x.dds"},
	{0x67B955B883CBD17CULL, "vxdr_0_deeproads_cx9_0_0.vlm"},
	{0x67B9814AFE6848CEULL, "6027038_m.fxe"},
	{0x67B989B2590226C5ULL, "llac_dr_0_deeproads_17035_1.dds"},
	{0x67BA3E9413C3FD01ULL, "qun231_varnell6098591.cl"},
	{0x67BA92C73E85EEA4ULL, "vxw_woundedcoast_1d4_0_0.vlm"},
	{0x67BAA73FE42B1DFFULL, "fire_u_02.dds"},
	{0x67BAB1547E3E3DE8ULL, "seb221_random_servants6187842.cl"},
	{0x67BAC1C8D93CC81CULL, "hla_ht_0_hightown_n_-2_0x.dds"},
	{0x67BAC77CBC8A860DULL, "vxlt_0_lowtown_a2_n_c_eot_0_1.vlm"},
	{0x67BACD7D587D7103ULL, "6095786_m.fxe"},
	{0x67BAEE2A7156530CULL, "132572659.ogg"},
	{0x67BB355412D2BA0EULL, "hlac_w_bonepit_327682_0x.dds"},
	{0x67BB661CFE8C1479ULL, "6201274_f.fxe"},
	{0x67BB7D0F9C860A9DULL, "hf_ser_agatha.mor"},
	{0x67BBC8A33765B35CULL, "vxlt_gallowscourtyard_6wc_0_0.vlm"},
	{0x67BBE07570FA49F8ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsbuild.phy"},
	{0x67BC3EAB0DEF321AULL, "mag100_cullen6095107.cl"},
	{0x67BC85FD7122AADDULL, "dre000_sandal6028400.cl"},
	{0x67BCAB430B998271ULL, "vxlt_gallowstemplar_n_686_0_0.vlm"},
	{0x67BCBC5616D56A72ULL, "vxlt_0_lowtown_f4y_0_0.vlm"},
	{0x67BD05E9DCAE4E7CULL, "cs_pro000cs_opening_scene.wbd"},
	{0x67BD12D0E0EE7E48ULL, "vxca_caverns1_mines_ing_0_0.vlm"},
	{0x67BD4C30512F2C04ULL, "hlac_lt_gallowstemplar_f_2394_9z.dds"},
	{0x67BD94C730528173ULL, "807928588.ogg"},
	{0x67BF12AB7E941AC4ULL, "qun210tr_follower_amb_qun.utt"},
	{0x67BF417585A1693FULL, "hlac_ht_mansion2b_818_60y.dds"},
	{0x67BF539EAAE3CED9ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.dura.gad"},
	{0x67BF8D156865E93DULL, "6127214_m.fxe"},
	{0x67BFE8AD12EEA9B0ULL, "6175708_m.fxe"},
	{0x67C01946BD60AC54ULL, "vxdr_primevaltomb_81g_0_0.vlm"},
	{0x67C02DAE86BB9085ULL, "vxw_woundedcoast_gy_0_0.vlm"},
	{0x67C048DFB4D82DF8ULL, "vxw_bonepit_1zi_0_0.vlm"},
	{0x67C0659935399711ULL, "gen_im_arm_cht_comd.uti"},
	{0x67C0AB0ED6DEDD58ULL, "llac_lt_gallowstemplar_n_7198_0.dds"},
	{0x67C0D56D06BCB652ULL, "6172909_m.fxe"},
	{0x67C1615C49C7EB1CULL, "318818221.ogg"},
	{0x67C164C6CD32B716ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgte.phy"},
	{0x67C189E46F93F25BULL, "lgt000ip_landmark.utp"},
	{0x67C18C49EBA130B6ULL, "vxlt_0_lowtown_n_ffx_0_0.vlm"},
	{0x67C1A79C5A6115EFULL, "6121175_m.fxe"},
	{0x67C1FBF8118FEAE1ULL, "vxca_caverns1_mines_ika_0_0.vlm"},
	{0x67C24538C2484DABULL, "569958595.ogg"},
	{0x67C26FE1E47DC4F3ULL, "follower_banter6143204.cl"},
	{0x67C2E858E59520B3ULL, "mcr310_cut_chantry_blows_up6087548.cl"},
	{0x67C358180137A2E1ULL, "hm_comt_servant_1a.mor"},
	{0x67C3753836665E8DULL, "hlac_ht_0_hightown_-65534_0x.dds"},
	{0x67C391830D4C9FE4ULL, "559626261.ogg"},
	{0x67C395222E7FC405ULL, "355583367.ogg"},
	{0x67C3A4AB3E4C1BDEULL, "hlac_lt_0_lowtown_docks_a3_n_0_3y.dds"},
	{0x67C3D67B0D07DCD4ULL, "vxlt_0_lowtown_docks_bau_0_0.vlm"},
	{0x67C42CA5CD38F14FULL, "vxlt_gallowscourtyard_63x_0_0.vlm"},
	{0x67C435A42D67907FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfirelrg.mmh"},
	{0x67C49B72A43DA3D6ULL, "vxht_chantry_2cx_0_0.vlm"},
	{0x67C4A8DBB52B70FBULL, "6104875_m.fxe"},
	{0x67C5A088E554E41CULL, "742423774.ogg"},
	{0x67C5DFC93C0C8EA3ULL, "6034431_m.fxe"},
	{0x67C60786E1880091ULL, "face_face_p4.psh.fxo"},
	{0x67C64D7FD297475CULL, "422554217.ogg"},
	{0x67C6649EAAEA0A27ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.dura.ani"},
	{0x67C6E76132FA256BULL, "plt_cod_hst_mgc_silverite.plo"},
	{0x67C77AE9D12A6F96ULL, "vxw_woundedcoast_1pz_0_0.vlm"},
	{0x67C7BFECD51F7870ULL, "55891961.ogg"},
	{0x67C7E73D347BD2CAULL, "25031018.ogg"},
	{0x67C82A4B7755A575ULL, "476068660.ogg"},
	{0x67C83E0995BDAC71ULL, "hlac_lt_warehouse_65537_4z.dds"},
	{0x67C85C0B13CD4812ULL, "6039763_m.fxe"},
	{0x67C8CEE38CCA600CULL, "6050427_m.fxe"},
	{0x67C904A4307C0724ULL, "vxw_woundedcoast_1o5_0_0.vlm"},
	{0x67C9111C9B045C43ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_3z.dds"},
	{0x67C9659BD4E0094FULL, "654839929.ogg"},
	{0x67C97CE69183221FULL, "318928693.ogg"},
	{0x67C998566ADF9656ULL, "363197959.ogg"},
	{0x67C9AECEBFCA1319ULL, "vxdr_0_deeproads_74p_0_0.vlm"},
	{0x67CB4DC6CD3AE4B7ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgte.msh"},
	{0x67CB5D2C23834EA6ULL, "6173348_m.fxe"},
	{0x67CB5FC6CD3B031AULL, "art\\characters\\playercharacter\\elfmale\\boo_lgte.mao"},
	{0x67CB63C6CD3B0A29ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgte.mmh"},
	{0x67CB7EEB3DF67AD0ULL, "6111618_m.fxe"},
	{0x67CB84ADD2E496C8ULL, "vxca_caverns1_mines_ikd_0_0.vlm"},
	{0x67CBDED50FF45FB6ULL, "hlac_ca_caverns2prc_5802_20z.dds"},
	{0x67CC6E037EEC876FULL, "sib131_slaver_lord6165200.cl"},
	{0x67CC9403BA63B151ULL, "1023706213.ogg"},
	{0x67CCA24F8C94C28AULL, "3167184.ogg"},
	{0x67CD272E61D3CA0EULL, "vxht_0_hightown_c9w_0_0.vlm"},
	{0x67CD8440B7B0DB70ULL, "vxht_brothel_n_4fk_0_0.vlm"},
	{0x67CDA53F13FB0917ULL, "181355031.ogg"},
	{0x67CDAD14D458D533ULL, "6054385_m.fxe"},
	{0x67CDEEF13F5B5F21ULL, "vxht_keep_3g2_0_0.vlm"},
	{0x67CDF56530AF5C2BULL, "494460316.ogg"},
	{0x67CE63BBF927718CULL, "plt_ser270pt_suspect.plo"},
	{0x67CEA0D255735A6BULL, "vxlt_0_lowtown_a2_eix_0_0.vlm"},
	{0x67CED10271FCC780ULL, "vxw_encounter3_fa_0_0.vlm"},
	{0x67CF1196A0468B8BULL, "vxw_bonepitmining_a2_aj_0_0.vlm"},
	{0x67CF24C2A10BD632ULL, "406174189.ogg"},
	{0x67CF4749F8B2CED2ULL, "6175809_m.fxe"},
	{0x67CF7BEF66C17167ULL, "mag100_karras6035318.cl"},
	{0x67CF87035E304B39ULL, "6151634_f.fxe"},
	{0x67CF95EF0E1B9380ULL, "62676962.ogg"},
	{0x67CFA60747625F08ULL, "lgt260_eau_tmp_lyrium_mess6143643.cl"},
	{0x67CFE2C6100CAFD8ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzfence_0.pwk"},
	{0x67D0FF60F2E3C1D7ULL, "6126794_m.fxe"},
	{0x67D10BEF2AB70967ULL, "626323829.ogg"},
	{0x67D15166DE282E55ULL, "llac_lt_gallowstemplar_2391_8.dds"},
	{0x67D15ED62F3181ECULL, "612507027.ogg"},
	{0x67D17DF1C23BD900ULL, "ser221_gamlen6135243.cl"},
	{0x67D240FFD037BB92ULL, "188569012.ogg"},
	{0x67D24857FEFDF554ULL, "vxlt_0_lowtown_docks_bdz_0_0.vlm"},
	{0x67D29D2147B298BDULL, "follower_barks6153594.cl"},
	{0x67D2FDE1E9EFBD74ULL, "hlac_lt_0_lowtown_a2_n_-1_0z.dds"},
	{0x67D303D5D5ED8EA2ULL, "art\\characters\\creatures\\spider\\animation\\p_rnf.gad"},
	{0x67D35CFA775B69FEULL, "vxlt_darktown_4s5_0_0.vlm"},
	{0x67D37F7A9B7FEAD4ULL, "hlac_ca_cavern1_haunted_24421_3z.dds"},
	{0x67D3CB1CD97DCA6FULL, "vxht_chantry_n_30n_0_0.vlm"},
	{0x67D460608F122782ULL, "6149958_m.fxe"},
	{0x67D4A667593DDBE1ULL, "vxlt_0_lowtown_a2_n_e1m_0_0.vlm"},
	{0x67D5200033695C93ULL, "616307714.ogg"},
	{0x67D54437F969F6C6ULL, "590123699.ogg"},
	{0x67D548BBF92D62C3ULL, "plt_ser270pt_suspect.nss"},
	{0x67D6002270B7B7A0ULL, "73775773.ogg"},
	{0x67D655CB0752791DULL, "293636585.ogg"},
	{0x67D6993101EC3030ULL, "hlac_w_woundedcoast_-131070_0z.dds"},
	{0x67D6C17C963FC444ULL, "6136767_m.fxe"},
	{0x67D6CEE0095424F6ULL, "889930043.ogg"},
	{0x67D7281A77694EF3ULL, "vxht_0_hightown_a2_c1r_0_0.vlm"},
	{0x67D79C5F32CD1017ULL, "6144524_f.fxe"},
	{0x67D7E6596646B416ULL, "vfx_bird_ra_up.dds"},
	{0x67D7F9D70569C728ULL, "darktown_male_fer_urch_sta6174965.cl"},
	{0x67D859170D15386CULL, "gen00flst_varric_preact2.stg"},
	{0x67D85B4F91344A00ULL, "gen00fl_aveline6139781.cl"},
	{0x67D8AC5D1B51E45CULL, "vxlt_gallowscourtyard_6ya_0_0.vlm"},
	{0x67D9105296BED754ULL, "vxht_keep_n_4aq_0_0.vlm"},
	{0x67D9E4DFA52D3249ULL, "98460428.ogg"},
	{0x67DA26D5D5F3E899ULL, "art\\characters\\creatures\\spider\\animation\\p_rnf.evt"},
	{0x67DA2CFF87367D42ULL, "hlac_ca_cavern1_haunted_24423_12x.dds"},
	{0x67DA318F2D2DC3B9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_3.pwk"},
	{0x67DA8A7FEDD3E9DAULL, "693047658.ogg"},
	{0x67DB20936E76E7AFULL, "6103355_m.fxe"},
	{0x67DB368198461A76ULL, "zz_testsetget.ncs"},
	{0x67DB93B7CED1EF3EULL, "hlac_w_encounter2_65539_0x.dds"},
	{0x67DBD93DF30ABACFULL, "634943819.ogg"},
	{0x67DC357474A15670ULL, "llac_ca_caverns2_5802_47.dds"},
	{0x67DC4FEADDC5066CULL, "vxlt_darktown_67m_0_0.vlm"},
	{0x67DD0989AD8A4573ULL, "llac_dr_0_deeproads_5843_7.dds"},
	{0x67DD11686AA7C185ULL, "vxw_woundedcoast_yg_0_0.vlm"},
	{0x67DD42DF63F2E000ULL, "vxw_encounter3_cn_0_0.vlm"},
	{0x67DD54CF78C8F203ULL, "art\\characters\\creatures\\hurlock\\textures\\default_gray50.dds"},
	{0x67DD8373923CA40BULL, "vxw_bonepitmining_a2_lm_0_0.vlm"},
	{0x67DE2F1D1A9287C8ULL, "vxlt_0_lowtown_a2_n_e1h_0_0.vlm"},
	{0x67DE99A7A4DCE30FULL, "vxlt_0_lowtown_a3_n_c_d46_0_0.vlm"},
	{0x67DF1DCC2A9D1900ULL, "vxht_keep_6bu_0_0.vlm"},
	{0x67DF7383DD9755EBULL, "vxht_keep_n_6bq_0_0.vlm"},
	{0x67DF76421D447559ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_roll_shoulders.ani"},
	{0x67DF7BECCF9FC879ULL, "266465486.ogg"},
	{0x67DFB892391CD298ULL, "vxlt_gamlenshouse_n_2z_0_0.vlm"},
	{0x67DFFEE39A4865EDULL, "art\\vfx\\textures\\vfx_web_head.dds"},
	{0x67E02CF534EC4129ULL, "art\\vfx\\combat\\vfx_gib_qun_ltr.ani"},
	{0x67E02E943F49C705ULL, "vxw_bonepit_1qq_0_0.vlm"},
	{0x67E0B00410E19DE0ULL, "23165214.ogg"},
	{0x67E10EE6E39F20CEULL, "gen00fl_anders6125392.cl"},
	{0x67E2E13D59E24D92ULL, "6092214_m.fxe"},
	{0x67E2E4D0E0EAD230ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredround_d.dds"},
	{0x67E37D7E8E0D90CCULL, "6004170_m.fxe"},
	{0x67E3E9EABE15A49DULL, "vxca_caverns2prc_5yx_0_0.vlm"},
	{0x67E49642FF979CDFULL, "hlac_ht_bartrand_n_3371_2x.dds"},
	{0x67E4DA4985A23D55ULL, "mer210pt_pickaxes.ncs"},
	{0x67E529AB7EB51171ULL, "qun210tr_follower_amb_qun.ncs"},
	{0x67E540C892395450ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4so_right.phy"},
	{0x67E5426FF2DC039BULL, "w_blightlands_4_0.msh"},
	{0x67E5546FF2DC22E6ULL, "w_blightlands_4_0.mao"},
	{0x67E5606FF2DC3685ULL, "w_blightlands_4_0.mmh"},
	{0x67E5ACE48D600282ULL, "vxw_encounter2_camp_6p_0_0.vlm"},
	{0x67E5B755284B594EULL, "art\\vfx\\combat\\vfx_sg_awake_act_c.impa.evt"},
	{0x67E646175648D8F7ULL, "294385211.ogg"},
	{0x67E651421D4A549BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_roll_shoulders.gad"},
	{0x67E66B86DD08A0E3ULL, "253717224.ogg"},
	{0x67E671B4766AC51CULL, "6100106_m.fxe"},
	{0x67E6C24D2569B0E4ULL, "mag211_marethari6030693.cl"},
	{0x67E6D72C06D80C4FULL, "vxlt_0_lowtown_e2m_0_0.vlm"},
	{0x67E6D9E21ECF51C7ULL, "6092130_m.fxe"},
	{0x67E6E39059CDA241ULL, "195546029.ogg"},
	{0x67E72469DAC331AFULL, "6093793_m.fxe"},
	{0x67E72ED5D5FE8088ULL, "art\\characters\\creatures\\spider\\animation\\p_rnf.ani"},
	{0x67E747F534F28DABULL, "art\\vfx\\combat\\vfx_gib_qun_ltr.gad"},
	{0x67E7FD7D414776F0ULL, "hlac_ca_caverns1_24421_5y.dds"},
	{0x67E8079CE218816BULL, "mag110_samson6020783.cl"},
	{0x67E828C6FAE0AC07ULL, "816458223.ogg"},
	{0x67E85AEEA6A22A5BULL, "852216850.ogg"},
	{0x67E8B3FFBFE0D2DEULL, "vxlt_warehouse_15s_0_0.vlm"},
	{0x67E9069A6E239F5DULL, "810470610.ogg"},
	{0x67E96AE729EBDA9DULL, "6048283_m.fxe"},
	{0x67E99270606D002FULL, "6104255_m.fxe"},
	{0x67EA03E9DCD538F6ULL, "cs_pro000cs_opening_scene.bnk"},
	{0x67EA2E9871E0A1B7ULL, "vxlt_hangedman_174_0_0.vlm"},
	{0x67EA395FD7D8A27BULL, "6149865_m.fxe"},
	{0x67EA78127ED41960ULL, "art\\vfx\\combat\\vfx_d_arc_aoe_cast_c.mmh"},
	{0x67EAE772D748B6FFULL, "dae000_gamlen6102717.cl"},
	{0x67EB291C7ECA984BULL, "vxht_chantry_27m_0_0.vlm"},
	{0x67EB492F77F588ACULL, "303290326.ogg"},
	{0x67EB4E5C1DC2A634ULL, "vxca_caverns2prc_9q1_0_0.vlm"},
	{0x67EB5C64A3D0E86BULL, "623368065.ogg"},
	{0x67EB78E3186FD86CULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_hed_ogra_0s.dds"},
	{0x67EB83455C9DDA05ULL, "vxw_woundedcoast_vf_0_0.vlm"},
	{0x67EC1433AB7AF23BULL, "1016619678.ogg"},
	{0x67ECAE55285168F9ULL, "art\\vfx\\combat\\vfx_sg_awake_act_c.impa.gad"},
	{0x67ECD783FF4CF523ULL, "dae000_flemeth6001013.cl"},
	{0x67ED04F3A24C6F1CULL, "t1_mul_wh1.tnt"},
	{0x67ED2AA46FAE8B7FULL, "vxht_mansion_n_5bf_0_0.vlm"},
	{0x67ED36A42D8A96BCULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfirelrg.anb"},
	{0x67ED6E421D50A480ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_roll_shoulders.evt"},
	{0x67EDC9C8497B4D32ULL, "6036296_m.fxe"},
	{0x67EDE45DA549EB35ULL, "80286036.ogg"},
	{0x67EE0B4AD8D46EFFULL, "6205697_m.fxe"},
	{0x67EE622DD33CB03CULL, "w_encounter1_13_0.rml"},
	{0x67EE7D75A60CA972ULL, "991519888.ogg"},
	{0x67EEA071313FDF85ULL, "vxw_bonepit_1zp_0_0.vlm"},
	{0x67EF9DEF8588F44EULL, "llac_dr_primevaltomb_9621_8.dds"},
	{0x67F03D98607EB836ULL, "lt_0_lowtown_docks_a3_4_0.rml"},
	{0x67F040B45B479AFEULL, "vxlt_darktown_4e7_0_0.vlm"},
	{0x67F0709C7E095466ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.anb"},
	{0x67F0F7EC5BD3E781ULL, "vxht_mansion2_3dc_0_0.vlm"},
	{0x67F1482DD33EB991ULL, "w_encounter1_13_0.msh"},
	{0x67F1489A91592D41ULL, "llac_w_bonepit_262148_0.dds"},
	{0x67F1562DD33ED1ACULL, "w_encounter1_13_0.mao"},
	{0x67F1622DD33EE64FULL, "w_encounter1_13_0.mmh"},
	{0x67F243C12F4B5574ULL, "hlac_lt_0_lowtown_a2_n_131072_0x.dds"},
	{0x67F2BF552855F147ULL, "art\\vfx\\combat\\vfx_sg_awake_act_c.impa.ani"},
	{0x67F2C71873FC7370ULL, "dae000ip_crowd_ff8_talk7.utp"},
	{0x67F364DE901ACBAEULL, "6014695_f.fxe"},
	{0x67F4015DBEDEB6EBULL, "736130380.ogg"},
	{0x67F41EC17F561B02ULL, "vxw_encounter2_camp_1q_0_0.vlm"},
	{0x67F45A3B065E50A4ULL, "vxw_woundedcoast_1p8_0_0.vlm"},
	{0x67F47A5B008739CFULL, "126697725.ogg"},
	{0x67F4AC96CDFC20EBULL, "6101159_m.fxe"},
	{0x67F4F2C0D04EA4A0ULL, "682655170.ogg"},
	{0x67F4F433001E3555ULL, "1066753463.ogg"},
	{0x67F525C9EA7F6C9BULL, "640404687.ogg"},
	{0x67F54BBF10C56A47ULL, "6097230_m.fxe"},
	{0x67F62BA24DA96D18ULL, "plt_lgt000pt_civ_citizens.nss"},
	{0x67F6A72FCC628B1DULL, "197502625.ogg"},
	{0x67F6C22C6E7D063DULL, "art\\levels\\races\\design\\prp_loot.phy"},
	{0x67F85B127EE0257FULL, "art\\vfx\\combat\\vfx_d_arc_aoe_cast_c.anb"},
	{0x67F8A15E0FB77545ULL, "vxlt_0_lowtown_docks_a3_n_cvx_0_0.vlm"},
	{0x67F8ADC91888213BULL, "vxht_bartrand_n_29w_0_0.vlm"},
	{0x67F8FF50E1252F7AULL, "vxlt_0_lowtown_docks_a3_n_cyn_0_0.vlm"},
	{0x67F93462819FC83FULL, "hlac_lt_gallowstemplar_f_7198_5z.dds"},
	{0x67F99E4B6651A0E0ULL, "mcr310cr_confused_guard.utc"},
	{0x67F9F5224E93F285ULL, "vxw_woundedcoast_se_0_0.vlm"},
	{0x67FA10537CB08FF8ULL, "vxw_bonepit_1ne_0_0.vlm"},
	{0x67FA82DF405FFC9DULL, "6143487_m.fxe"},
	{0x67FAACAC48817886ULL, "zz_dae_debug6045444.cl"},
	{0x67FABBD38727AA6CULL, "gen00fl_aveline6122749.cl"},
	{0x67FAFED602A2A1FFULL, "vxlt_0_lowtown_a3_n_c_d5u_0_0.vlm"},
	{0x67FB0A290FAF73F2ULL, "vxlt_undercitydungeon_24m_0_0.vlm"},
	{0x67FB166CC2CBE581ULL, "vxlt_gallowscourtyard_6y0_0_0.vlm"},
	{0x67FB6FC9465A95F6ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha12a.msh"},
	{0x67FB79C9465AA6F4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha12a.mmh"},
	{0x67FB8F5688A2F3FBULL, "656793618.ogg"},
	{0x67FC01860E894A00ULL, "vxht_keep_6hs_0_0.vlm"},
	{0x67FC09C7AEBE82D2ULL, "vxw_bonepitdestroyed_a3_19z_0_0.vlm"},
	{0x67FC32BB92B41573ULL, "230291853.ogg"},
	{0x67FC9792EDDEEA76ULL, "hlac_lt_gallowscourtyard_n_-1_3x.dds"},
	{0x67FD6C1672DE22D4ULL, "vxca_cavern1_haunted_i1j_0_0.vlm"},
	{0x67FD739C7E13E311ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.mmh"},
	{0x67FD80218473C2AEULL, "hlac_lt_gallowscourtyard_n_-131072_2x.dds"},
	{0x67FE648A25A90FF0ULL, "vxdr_primevaltomb_6y1_0_0.vlm"},
	{0x67FE7CD322FB18C4ULL, "123518297.ogg"},
	{0x67FEA288EC466ABEULL, "dae000_oldmangossip6175770.cl"},
	{0x67FEC0151B32FF86ULL, "506541144.ogg"},
	{0x67FFA5EB7B445431ULL, "474083346.ogg"},
	{0x67FFE09B73084102ULL, "lt_0_lowtown_a3_n_c_46_0.rml"},
	{0x680028CAB156E6CBULL, "vxdr_0_deeproads_f69_0_0.vlm"},
	{0x68005B0BD7DEF9B5ULL, "vrc221_varric6043884.cl"},
	{0x68008886F8A70B13ULL, "vxht_mansion_n_34f_0_0.vlm"},
	{0x6800B09E4877F39DULL, "557843548.ogg"},
	{0x6800C326A7B49ED6ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluvian_n.dds"},
	{0x6800D9573D6C85ECULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cratequest.msh"},
	{0x6800DF573D6C8FDEULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cratequest.mmh"},
	{0x6800EB573D6CA475ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cratequest.mao"},
	{0x6800F630A47FEA07ULL, "915370024.ogg"},
	{0x68016B27ED179018ULL, "nat320_delilah6090034.cl"},
	{0x6801761819CAE7D6ULL, "zz_followers_banter6157921.cl"},
	{0x6801DC1E1B46C6BEULL, "use.cur"},
	{0x680264384A3DDDD3ULL, "vxca_cavern1_haunted_inx_0_0.vlm"},
	{0x6803106899D40DE1ULL, "and100_anders6095856.cl"},
	{0x6803D49BE8403F44ULL, "6155553_m.fxe"},
	{0x6804091B36C3DB0CULL, "str_cl_2.mao"},
	{0x68045C8AF8292A20ULL, "6149361_m.fxe"},
	{0x6804814671D9C6DBULL, "ser191_emeric6120418.cl"},
	{0x680585913226A1C4ULL, "914132448.ogg"},
	{0x6806760709FD4FC7ULL, "166725110.ogg"},
	{0x6806E106151D3EDBULL, "637365582.ogg"},
	{0x6806E89D4AF14290ULL, "287289765.ogg"},
	{0x6806F798C145DB80ULL, "451289638.ogg"},
	{0x6807482893BEBB2BULL, "lt_0_lowtown_a3_n_c_42_0_lc.mmh"},
	{0x68074F98DB4EB27FULL, "vxlt_hangedman_1qf_0_0.vlm"},
	{0x6807610537C34667ULL, "vxca_caverns2_70z_0_0.vlm"},
	{0x6807610CDFD47F2FULL, "456575294.ogg"},
	{0x680785E2D81B2928ULL, "dre000_hireling16038657.cl"},
	{0x6807AEAFC8226890ULL, "vxlt_0_lowtown_n_ewr_0_0.vlm"},
	{0x6807B72C6E8B66A9ULL, "art\\levels\\races\\design\\prp_loot.mao"},
	{0x6807C32C6E8B7B42ULL, "art\\levels\\races\\design\\prp_loot.mmh"},
	{0x6807C52C6E8B7EF8ULL, "art\\levels\\races\\design\\prp_loot.msh"},
	{0x6807CAB7ECA8EA50ULL, "6075299_m.fxe"},
	{0x680833E5251023C9ULL, "gen00fl_anders6135770.cl"},
	{0x6809172B21389239ULL, "hlac_lt_0_lowtown_docks_131071_7y.dds"},
	{0x680960BF4E382CD0ULL, "130982742.ogg"},
	{0x6809CC12DA74D627ULL, "gen00fl_anders6116821.cl"},
	{0x6809D117CE18DCB7ULL, "vxht_mansion_n_500_0_0.vlm"},
	{0x6809E4BF11BF1A0AULL, "687915132.ogg"},
	{0x6809E8E6C85A28D3ULL, "hm_amb_merchant_armour.mor"},
	{0x6809FB9A3D2AA732ULL, "hlac_w_woundedcoast_65544_0y.dds"},
	{0x680A74A9F9F65A54ULL, "vxw_bonepitdestroyed_a3_1bm_0_0.vlm"},
	{0x680AB9A7BBF897D9ULL, "lgt311_shw_mer_mettin6156458.cl"},
	{0x680B592F8FE240B6ULL, "6180819_f.fxe"},
	{0x680B5AB4BE621C82ULL, "vxca_caverns1_j5r_0_0.vlm"},
	{0x680C326447D4DDA8ULL, "345676687.ogg"},
	{0x680CDBAB732EEBDEULL, "356027982.ogg"},
	{0x680D77EB15A1EE35ULL, "6180486_m.fxe"},
	{0x680E855624451DC2ULL, "vxlt_hangedman_5r_0_0.vlm"},
	{0x680EFFBC7EFFB079ULL, "zz_vrc_debug6200263.cl"},
	{0x680F2C161074F32EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic_1.mao"},
	{0x680F49B45A5CCD9CULL, "ht_0_hightown_a2_18_0_lc.mmh"},
	{0x680FA131C06844B5ULL, "nat380_nathaniel6066501.cl"},
	{0x68103346D9FB8BC1ULL, "hlac_lt_0_lowtown_a2_n_-65537_5y.dds"},
	{0x6810DB66D2A22C1BULL, "vxlt_gallowstemplar_34n_0_0.vlm"},
	{0x6811AD785CC887C3ULL, "vxht_chantry_32w_0_0.vlm"},
	{0x6811B38FB75EEB5EULL, "6103321_m.fxe"},
	{0x68126DD376945A54ULL, "6170039_m.fxe"},
	{0x6812CDC6DB9F7EBDULL, "vxlt_undercityserialkiller_1ir_0_1.vlm"},
	{0x6812E04831E10696ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogcorkc.mao"},
	{0x6812EC4831E11B35ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogcorkc.mmh"},
	{0x6812EE4831E11EABULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogcorkc.msh"},
	{0x68139951392BD9DFULL, "837406582.ogg"},
	{0x6813DFE0E6F6E660ULL, "vxw_woundedcoast_85_0_0.vlm"},
	{0x68144C9FB2A2A0C8ULL, "w_bonepit_110_0.msh"},
	{0x68145E9FB2A2BF59ULL, "w_bonepit_110_0.mao"},
	{0x68146A9FB2A2D3F2ULL, "w_bonepit_110_0.mmh"},
	{0x68146C65325952F6ULL, "mer221_sabin6080106.cl"},
	{0x6814CB3AF850B18AULL, "158470326.ogg"},
	{0x6815341E48C5277CULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath5.ani"},
	{0x681556EE171FD146ULL, "6108298_m.fxe"},
	{0x6815DFA227ADE39FULL, "6132885_m.fxe"},
	{0x6815F95AD71369C0ULL, "lt_0_lowtown_a2_n_52_0.anb"},
	{0x681639AE8F907F7CULL, "llac_ht_mansion2b_818_51.dds"},
	{0x68168ACD901DDEA1ULL, "hlac_lt_0_lowtown_docks_-131073_5z.dds"},
	{0x6816A3F2F3A4EFB4ULL, "vxca_caverns2prc_9f4_0_0.vlm"},
	{0x6816CD757146D48DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsbuild.msh"},
	{0x6816DB757146ECC0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsbuild.mao"},
	{0x6816DF757146F313ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsbuild.mmh"},
	{0x6817A4C471085CD1ULL, "581380130.ogg"},
	{0x681862CC72BA6996ULL, "qun220cr_javaris6071895.cl"},
	{0x6818803B4590D4FFULL, "vxht_mansion_n_5uc_0_0.vlm"},
	{0x68188051F458F70DULL, "128053451.ogg"},
	{0x6818E686EBEC72D4ULL, "vxw_bonepitdestroyed_a3_1gr_0_0.vlm"},
	{0x68192B6F415D8888ULL, "llac_dr_dungeon_3207_4.dds"},
	{0x68195F408C0B6AC7ULL, "ravenfly.mao"},
	{0x6819A5BBDC3AC11DULL, "382109628.ogg"},
	{0x6819E4603F171C24ULL, "nat380_temmerin6067328.cl"},
	{0x681A632DD361EC8CULL, "w_encounter1_13_0.anb"},
	{0x681A66D08CD64D40ULL, "hlac_lt_0_lowtown_n_-3_0z.dds"},
	{0x681AFBFBB99B0311ULL, "art\\characters\\creatures\\stonegolem\\animation\\crn_rdy_lf.ani"},
	{0x681B2533554F25F8ULL, "472242159.ogg"},
	{0x681B4D239D9DEDC1ULL, "ave310_brennan6136888.cl"},
	{0x681B9C74AB10BE2FULL, "dae000cr_grand_cleric6144102.cl"},
	{0x681B9C8875882441ULL, "dae200pt_find_mansion.ncs"},
	{0x681BA5BABE8551D9ULL, "669648994.ogg"},
	{0x681BFAF8EC90F7ADULL, "face0_s1.psh.fxo"},
	{0x681C1B7FE18B3396ULL, "dae000_worthy6166601.cl"},
	{0x681C411E48CB5BBEULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath5.gad"},
	{0x681C602E9812F0E6ULL, "vxht_mansion_n_4xl_0_0.vlm"},
	{0x681DED51DF15A7E6ULL, "877212900.ogg"},
	{0x681E24C31C33DD32ULL, "mag101_alain6042755.cl"},
	{0x681E4D13B987AC6AULL, "6170020_m.fxe"},
	{0x681EA995FF6156C5ULL, "vxlt_gallowscourtyard_n_3t9_0_0.vlm"},
	{0x681F053F71BE5D2BULL, "6042687_m.fxe"},
	{0x681F2C0FCF72303EULL, "21081794.ogg"},
	{0x681F50D78655BD42ULL, "592612744.ogg"},
	{0x681FE774CA88E825ULL, "llac_w_bonepitmining_a2_131075_0.dds"},
	{0x681FEE997A3A1E6FULL, "zz_vault_debug6203342.cl"},
	{0x681FF6C311A51E74ULL, "533107294.ogg"},
	{0x6820006C0EFB52E6ULL, "vxlt_gallowsprison_n_6mn_0_0.vlm"},
	{0x6820DBA4E3626460ULL, "6164632_m.fxe"},
	{0x682124204E5CE4C2ULL, "w_blightlands_185_0.phy"},
	{0x6821338C226EC167ULL, "506606866.ogg"},
	{0x682206FBB9A134B3ULL, "art\\characters\\creatures\\stonegolem\\animation\\crn_rdy_lf.gad"},
	{0x6822140D6F2E5BE8ULL, "vxlt_0_lowtown_docks_b82_0_0.vlm"},
	{0x682251BDD8ECFEE0ULL, "vxw_woundedcoast_52_0_0.vlm"},
	{0x6822799F9D10A6DAULL, "vxlt_0_lowtown_a3_n_c_e33_0_0.vlm"},
	{0x682309ABCC66895CULL, "hlac_ht_0_hightown_a2_-65538_1x.dds"},
	{0x682331876B8F0F5FULL, "843183781.ogg"},
	{0x682342CC0FB5F460ULL, "gen00fl_merrill6119658.cl"},
	{0x682354966BC73173ULL, "979603009.ogg"},
	{0x682382F58B209E44ULL, "146814271.ogg"},
	{0x6823C7876EC41AFBULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm.mmh"},
	{0x6823E5876EC44DE5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm.msh"},
	{0x6823F6F4E599BEC5ULL, "vxlt_0_lowtown_docks_a3_n_cye_0_0.vlm"},
	{0x682439886B862987ULL, "vxlt_0_lowtown_fe7_0_0.vlm"},
	{0x682455F8064F8F28ULL, "qainsf2.arl"},
	{0x68249269AC0E9990ULL, "vxlt_0_lowtown_n_emp_0_0.vlm"},
	{0x6824A8D22C2EB366ULL, "qun221_petrice6082682.cl"},
	{0x6824EE4FDCE3DA78ULL, "3019324.ogg"},
	{0x6825357DB57A3A23ULL, "10753429.ogg"},
	{0x682536515B1B57D4ULL, "vxdr_primevaltomb_8nn_0_0.vlm"},
	{0x6825606FF3121A8EULL, "w_blightlands_4_0.rml"},
	{0x6825F5A802C26A19ULL, "hlac_ht_playershouse_n_855_21y.dds"},
	{0x682605CF5AF6D3D2ULL, "74715499.ogg"},
	{0x682672FDD7615AC1ULL, "vxht_bartrand_n_6qq_0_0.vlm"},
	{0x6826C20D59B90465ULL, "6171428_m.fxe"},
	{0x6826CDACF7850507ULL, "vxlt_gallowstemplar_n_3qv_0_0.vlm"},
	{0x6826D3A0C18030B8ULL, "vxlt_darktown_4ie_0_0.vlm"},
	{0x6826D6FF84E77B0AULL, "llac_ht_0_hightown_a2_-131073_1.dds"},
	{0x6826E888085E1C5DULL, "6032008_m.fxe"},
	{0x6827168EB0F589A8ULL, "hlac_lt_0_lowtown_-2_5y.dds"},
	{0x68273345291FFCC0ULL, "529600563.ogg"},
	{0x6827D3A9D527CA8BULL, "vxht_0_hightown_n_3xr_0_0.vlm"},
	{0x68285BDD56601B3BULL, "lt_0_lowtown_docks_n_242_0.msh"},
	{0x68286DDD56603986ULL, "lt_0_lowtown_docks_n_242_0.mao"},
	{0x682879DD56604E25ULL, "lt_0_lowtown_docks_n_242_0.mmh"},
	{0x6828963A6679B287ULL, "6110896_m.fxe"},
	{0x6828F4684355F057ULL, "6065812_cassandra_legends.cut"},
	{0x6828F5641F4509A8ULL, "hlac_ca_caverns2_5802_49x.dds"},
	{0x682903FBB9A74E18ULL, "art\\characters\\creatures\\stonegolem\\animation\\crn_rdy_lf.evt"},
	{0x6829180CE7E25D1EULL, "6183968_f.fxe"},
	{0x682943ACF0DDC7FCULL, "vxht_chantry_2s2_0_0.vlm"},
	{0x6829532E7C638D4DULL, "gen00fl_varric6093678.cl"},
	{0x6829723CDB489603ULL, "6140273_m.fxe"},
	{0x682ABB08F4186D6FULL, "llac_lt_0_lowtown_docks_n_-2_5.dds"},
	{0x682AC4DDCA2D28BBULL, "1018664513.ogg"},
	{0x682AEE9E991DA15FULL, "qun120_arishok_guard_nt6063077.cl"},
	{0x682B064CA136DB0DULL, "vxca_caverns2_7px_0_0.vlm"},
	{0x682B38C33B078EAEULL, "6042215_m.fxe"},
	{0x682B92C584B41BB5ULL, "vxlt_warehouse_n_90_0_0.vlm"},
	{0x682BB35146BA1A8BULL, "6020055_f.fxe"},
	{0x682C563877D6D479ULL, "vxht_0_hightown_ap5_0_0.vlm"},
	{0x682C9DA702C7C017ULL, "6128806_m.fxe"},
	{0x682CA220322329BEULL, "644544753.ogg"},
	{0x682CCAE6FDF96C5FULL, "462463376.ogg"},
	{0x682D541A6D859EA4ULL, "701816737.ogg"},
	{0x682D6C57A1CB9D6DULL, "hlac_ht_keep_n_3732_8z.dds"},
	{0x682DD9FC09062AD5ULL, "hlac_w_bonepit_196612_0x.dds"},
	{0x682DEBA782224353ULL, "vxlt_0_lowtown_docks_byn_0_0.vlm"},
	{0x682E30553F12530EULL, "bad_med_bs.mao"},
	{0x682E5C2A90AC4446ULL, "hm_cor_blimey.mor"},
	{0x682EBFC36BDABA8CULL, "hlac_lt_0_lowtown_a2_131071_6x.dds"},
	{0x682F3BC72E0A8174ULL, "6044901_m.fxe"},
	{0x682F50C0894A38D2ULL, "hlac_lt_foundry_n_320_23x.dds"},
	{0x68302FD7B9944C35ULL, "vxht_0_hightown_a3_ccs_0_0.vlm"},
	{0x683044030926CFD6ULL, "lt_0_lowtown_docks_a3_n_23_0.rml"},
	{0x683096F752BFB6EFULL, "mag170_hugh6080711.cl"},
	{0x6830C39ACAE31760ULL, "vxw_woundedcoast_23_0_0.vlm"},
	{0x6830EEA8D01BD227ULL, "hlac_w_bonepitmining_a2_458754_0x.dds"},
	{0x6832088E25AE0D6BULL, "6076218_f.fxe"},
	{0x68325C13F510B157ULL, "vxht_keep_n_53a_0_0.vlm"},
	{0x6832C59CE13B5AEFULL, "tbe100ar_emporium.are"},
	{0x6833171BCA008ADAULL, "dae000_bodhan_letter6144290.cl"},
	{0x68333F603997F1F3ULL, "hlac_w_encounter3_-65534_0z.dds"},
	{0x683376AF05CDC7B6ULL, "vxca_caverns2prc_4zs_0_0.vlm"},
	{0x6833A82E4D117054ULL, "vxdr_primevaltomb_8io_0_0.vlm"},
	{0x68340246C5F575D5ULL, "lgt000_chanter6147394.cl"},
	{0x683441BF48D2447EULL, "tbe200im_barbarian_2h.uti"},
	{0x6834F44ADDB4817FULL, "gen00fl_aveline6112295.cl"},
	{0x6835500A2672A5D2ULL, "6144272_m.fxe"},
	{0x683617463162F726ULL, "vxlt_0_lowtown_docks_a3_n_d2h_0_0.vlm"},
	{0x6836398B8D5B166EULL, "6134819_m.fxe"},
	{0x68366F362D59BE70ULL, "w_encounter2_camp_63_0.mao"},
	{0x683673362D59C483ULL, "w_encounter2_camp_63_0.mmh"},
	{0x68367B60E68AC4A7ULL, "vxlt_0_lowtown_a2_n_gz0_0_0.vlm"},
	{0x683681362D59DCDDULL, "w_encounter2_camp_63_0.msh"},
	{0x6836BBD9B7107D92ULL, "189559850.ogg"},
	{0x68370146C6AFFB33ULL, "zz_prmpt_debug6200957.cl"},
	{0x683755966F46B6D0ULL, "656830934.ogg"},
	{0x68378BD65D667816ULL, "116900562.ogg"},
	{0x6837F9181809AD90ULL, "6141237_m.fxe"},
	{0x68382CBCB0207700ULL, "vxdr_primevalentrance_3dg_0_0.vlm"},
	{0x6838EAA237EBB99FULL, "vxlt_0_lowtown_docks_a3_n_d07_0_0.vlm"},
	{0x683914835D6506C8ULL, "6102420_f.fxe"},
	{0x6839AF27EA74E59AULL, "690169492.ogg"},
	{0x6839DD9B7339B4FEULL, "lt_0_lowtown_a3_n_c_46_0.anb"},
	{0x683A0F44C26714A3ULL, "follower_banter6172279.cl"},
	{0x683A5C3626112A81ULL, "976207826.ogg"},
	{0x683A8C57A99257E3ULL, "6041206_m.fxe"},
	{0x683ABC1C2BB6427BULL, "6042034_m.fxe"},
	{0x683B9C71856115D7ULL, "6143969_f.fxe"},
	{0x683CD5580A404350ULL, "mrl250_pol6101325.cl"},
	{0x683D987158EAA0C2ULL, "353212545.ogg"},
	{0x683D99F64CCCA674ULL, "vxca_cavern1_haunted_hu3_0_0.vlm"},
	{0x683E154DB75BF5F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose.ani"},
	{0x683E1E22CC023CA3ULL, "vxht_mansion2b_2uo_0_0.vlm"},
	{0x683E72B1C18B9523ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\run.ani"},
	{0x683EA1B4AB8A64B5ULL, "vxht_0_hightown_a3_cdr_0_0.vlm"},
	{0x683EC802734A8DD3ULL, "159064288.ogg"},
	{0x683ED12A805EFB6DULL, "rdr111dg_martin6058284.cl"},
	{0x683F167E619F22DAULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.impa.gad"},
	{0x683F8DE94DB3D59BULL, "6001348_f.fxe"},
	{0x683F9AC21DD11849ULL, "vxht_brothel_n_5cl_0_0.vlm"},
	{0x683F9B31A3ECD345ULL, "hlac_lt_ruinedhouse_n_-1_5z.dds"},
	{0x683FB4FDD80BEFACULL, "6088237_m.fxe"},
	{0x683FC08EE7711C3CULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgta_0d.dds"},
	{0x683FEE079B541592ULL, "mag160_feynriel6020456.cl"},
	{0x68407EBBFFF3AB86ULL, "zz_and_debug6178589.cl"},
	{0x6840A859A58FF810ULL, "mcr310_confused_guard6077046.cl"},
	{0x6840B6573DA22E79ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_cratequest.phy"},
	{0x6841CF0E3077AB2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf_injured.ani"},
	{0x68421A0B3F0788D4ULL, "vxdr_primevaltomb_8dl_0_0.vlm"},
	{0x6842B3FB09D5929FULL, "6126465_f.fxe"},
	{0x6842DD962F0934E4ULL, "vxht_mansion2b_1n2_0_0.vlm"},
	{0x684379BE7835F91AULL, "dae000_fenarel6127237.cl"},
	{0x6843BE9CE149C1ACULL, "tbe100ar_emporium.lst"},
	{0x684404034CB8A3DCULL, "vxw_encounter3_c2_0_0.vlm"},
	{0x6844103C77BC30A0ULL, "mag370_meredith6092180.cl"},
	{0x684438BF148CB9DBULL, "387348969.ogg"},
	{0x6844DD0B33E5415CULL, "vxca_caverns1_iht_0_0.vlm"},
	{0x6844F04DB761D53BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose.gad"},
	{0x68451E7551C0DA52ULL, "lgt210_qun_blk_qunari6142306.cl"},
	{0x68453845C2A04694ULL, "zz_mrl_debug6185312.cl"},
	{0x6845610C5EB24D15ULL, "577247474.ogg"},
	{0x6845663328DA95A3ULL, "vxht_bartrand_n_657_0_0.vlm"},
	{0x6845772B9C44AB4CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medc_0d.dds"},
	{0x6845FF204E7BC276ULL, "w_blightlands_185_0.mao"},
	{0x68460B204E7BD615ULL, "w_blightlands_185_0.mmh"},
	{0x68460D204E7BD98BULL, "w_blightlands_185_0.msh"},
	{0x6846197E61A546F1ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.impa.evt"},
	{0x68464211E15FC519ULL, "gen_im_trs_bst_spd_gland_01.uti"},
	{0x68469458F8A75ACBULL, "648544330.ogg"},
	{0x684774362D6839A8ULL, "w_encounter2_camp_63_0.phy"},
	{0x6847EA0685230C0DULL, "vxca_caverns2_7jv_0_0.vlm"},
	{0x68480787B3D03468ULL, "lt_0_lowtown_n_33_0_lc.mmh"},
	{0x6848A06DBAD2ECDDULL, "6126569_m.fxe"},
	{0x6848DC0E307DE06EULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf_injured.gad"},
	{0x68493D156CCCC948ULL, "vxlt_0_lowtown_docks_n_akk_0_0.vlm"},
	{0x684978293919A36AULL, "vxlt_gallowstemplar_51f_0_0.vlm"},
	{0x68497CF0C709803FULL, "6089964_f.fxe"},
	{0x6849BA743B30BB05ULL, "vxca_caverns1_egx_0_0.vlm"},
	{0x684A7680BD161E90ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_distort1.mao"},
	{0x684A9E9CE14FAA55ULL, "tbe100ar_emporium.ncs"},
	{0x684AAAEEFD1344A6ULL, "gen00fl_merrill6106999.cl"},
	{0x684AC75397D03039ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_act_c.impa.ani"},
	{0x684B33774B7DCA46ULL, "6109415_m.fxe"},
	{0x684B3D6C33E6DD07ULL, "gen_im_trs_tmp_hvy_glv_r2.uti"},
	{0x684B940AAA187DA3ULL, "441048613.ogg"},
	{0x684BE02398F4237BULL, "dae000_ineria6137186.cl"},
	{0x684C0A5FF96FB5ADULL, "6124299_m.fxe"},
	{0x684C0D4DB7682520ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose.evt"},
	{0x684C1FB0A4AE0016ULL, "hlac_w_bonepitdestroyed_a3_131077_0y.dds"},
	{0x684C43233E460150ULL, "668075109.ogg"},
	{0x684C55293EBDAC59ULL, "hlac_ht_keep_n_3733_0y.dds"},
	{0x684C8AB1C197FB1AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\run.evt"},
	{0x684D301E1FBACF9AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08b.mmh"},
	{0x684D421E1FBAEEE0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08b.msh"},
	{0x684D73362D6CA3A0ULL, "w_encounter2_camp_63_0.rml"},
	{0x684D7DAF10897778ULL, "dae000ip_crowd_ff2_talk12.utp"},
	{0x684D874FFA61B432ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_loop.gad"},
	{0x684D87C27DB32B53ULL, "vxlt_gallowscourtyard_7cm_0_0.vlm"},
	{0x684D884A57F6B963ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidright.mao"},
	{0x684D8C4A57F6BFB0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidright.mmh"},
	{0x684D9A4A57F6D7CAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidright.msh"},
	{0x684F43F1A2C9FFF2ULL, "fen110_party_comment6095075.cl"},
	{0x685017C309CBFDF5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_3.pwk"},
	{0x68514F7320FF4D64ULL, "vxht_mansion2b_1a3_0_0.vlm"},
	{0x6851A25397D6107BULL, "art\\vfx\\combat\\vfx_w_jug_lineback_act_c.impa.gad"},
	{0x6851A2A9F8F79F73ULL, "mag321_comtesse_de_launcet6042917.cl"},
	{0x685222D9B3763A70ULL, "vxdr_primevaltomb_34s_0_0.vlm"},
	{0x68528C0A65CA1EF7ULL, "819043606.ogg"},
	{0x6852C92911514FF3ULL, "ca_caverns1_24427_lc.mmh"},
	{0x685324DC9DBCF7C9ULL, "775057091.ogg"},
	{0x6853317E61AFF9C0ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.impa.ani"},
	{0x685368C26CF95D12ULL, "hlac_ht_0_hightown_a3_-65537_1z.dds"},
	{0x685389B1C19E180DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\run.gad"},
	{0x68538BD7EA87FDD4ULL, "mer120_hubert6059381.cl"},
	{0x6853BCA1177479E9ULL, "vxlt_0_lowtown_n_erx_0_0.vlm"},
	{0x6853E71A3F24114BULL, "vxdr_0_deeproads_cwg_0_0.vlm"},
	{0x68541080299EA697ULL, "6169910_f.fxe"},
	{0x68543E69D321EB0CULL, "6077187_m.fxe"},
	{0x68545E35F8636B10ULL, "225461615.ogg"},
	{0x68546C5C1CC1FFBEULL, "6152229_f.fxe"},
	{0x68548A4FFA67D869ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_loop.evt"},
	{0x6855078E03ACA0F1ULL, "hlac_ht_mansion2_818_1z.dds"},
	{0x6855EFCF185D3AE7ULL, "static_depth.vsh.fxo"},
	{0x6855F9CFF1DAD94EULL, "vxht_0_hightown_ai9_0_0.vlm"},
	{0x685678C946A750EFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha12a.phy"},
	{0x6856D85AD74AC8E8ULL, "lt_0_lowtown_a2_n_52_0.rml"},
	{0x6856E9B0333218D7ULL, "vxlt_0_lowtown_docks_coy_0_0.vlm"},
	{0x685778B407B92051ULL, "423309961.ogg"},
	{0x68579D1C3684931EULL, "6082138_m.fxe"},
	{0x6857BF5397DAAD60ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_act_c.impa.evt"},
	{0x6857F5E8CA593211ULL, "6218015_m.fxe"},
	{0x6857FC36023C8C81ULL, "vxw_encounter3_fl_0_0.vlm"},
	{0x6858254026EE02ADULL, "6092265_f.fxe"},
	{0x68584F8A94950670ULL, "w_woundedcoast_105_0.mao"},
	{0x6858538A94950C83ULL, "w_woundedcoast_105_0.mmh"},
	{0x6858618A949524DDULL, "w_woundedcoast_105_0.msh"},
	{0x68586FFB49E7DDB2ULL, "vxlt_foundry_n_yi_0_0.vlm"},
	{0x6858C2FA6C55088DULL, "625385150.ogg"},
	{0x6858C9E0CDC50845ULL, "6111484_m.fxe"},
	{0x6858EECBF968984CULL, "vxlt_gallowscourtyard_7fy_0_0.vlm"},
	{0x6859421EBA4B5C33ULL, "hlac_w_woundedcoast_-131066_0z.dds"},
	{0x68594585C7A32910ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magb.mao"},
	{0x68594985C7A33023ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magb.mmh"},
	{0x68594A9D02EF21B0ULL, "hlac_lt_gallowscourtyard_65534_1z.dds"},
	{0x68594E85F1B75412ULL, "vxca_caverns2prc_aec_0_0.vlm"},
	{0x68595785C7A3487DULL, "art\\characters\\playercharacter\\humanfemale\\boo_magb.msh"},
	{0x6859C143AA95A7E0ULL, "vxca_cave_2i9_0_1.vlm"},
	{0x685A13A4C9675E87ULL, "llac_w_sundermount_196602_0.dds"},
	{0x685A1AD6140F6303ULL, "hlac_ht_mansion_n_3364_1z.dds"},
	{0x685A1DC309D45C3AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_3.mao"},
	{0x685A81763989ECD2ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bed_sa.dds"},
	{0x685AFD6E4C1E203EULL, "vxlt_gamlenshouse_n_60_0_0.vlm"},
	{0x685B01DCAFEE6DA3ULL, "vxht_mansion2b_2kq_0_0.vlm"},
	{0x685B1416A54BF996ULL, "hlac_ht_0_hightown_a3_65536_3z.dds"},
	{0x685B44949DEDA888ULL, "321538475.ogg"},
	{0x685B599715A1117EULL, "gen_im_acc_rng_act1_03.uti"},
	{0x685B81EA416C47D2ULL, "6108931_m.fxe"},
	{0x685B84B4AAA90B6DULL, "6208205_m.fxe"},
	{0x685BA7BC874D4AF8ULL, "245197220.ogg"},
	{0x685BF99F6FA943D3ULL, "vxlt_gallowscourtyard_7dl_0_0.vlm"},
	{0x685C7E7C01BD4949ULL, "vxht_brothel_n_5ir_0_0.vlm"},
	{0x685C99D0B294FF54ULL, "835980205.ogg"},
	{0x685CFB42DF17252EULL, "39337704.ogg"},
	{0x685D3AE9B81C7171ULL, "774090914.ogg"},
	{0x685D44E7F99CCB02ULL, "hlac_lt_hangedman_n_-1_27y.dds"},
	{0x685D5C0699CD1F3CULL, "6127075_m.fxe"},
	{0x685E6F017EDE2706ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas2_drelief.dds"},
	{0x685E8EF27CA2C4C2ULL, "vxht_mansion2b_1hs_0_1.vlm"},
	{0x685EC31F7BAE457CULL, "6064020_m.fxe"},
	{0x685F1B12F5E6B2A8ULL, "6154903_m.fxe"},
	{0x685F4B797AF6EE1EULL, "vxdr_0_deeproads_9ek_0_1.vlm"},
	{0x685FA64F06684389ULL, "6199322_f.fxe"},
	{0x685FAF2D8CC5D2B0ULL, "llac_lt_0_lowtown_docks_a3_131070_3.dds"},
	{0x685FBF5D39E09842ULL, "6078676_m.fxe"},
	{0x68608B896A69B33BULL, "hlac_ht_0_hightown_a2_0_2z.dds"},
	{0x6860C976BEAA592CULL, "hlac_dr_primevaltomb_657_3y.dds"},
	{0x6861724FFA7239D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_hm_loop.ani"},
	{0x6861ABAF47B1B49FULL, "hlac_lt_warehouse_0_0z.dds"},
	{0x6861D295AF3D78A3ULL, "llac_ht_0_hightown_a3_-2_0.dds"},
	{0x6862BC76A64A408CULL, "ave331_jeven6151137.cl"},
	{0x68631D1614C8D076ULL, "342029143.ogg"},
	{0x6863C84C8E3D708EULL, "945649115.ogg"},
	{0x6863E6962102DC1BULL, "6172852_m.fxe"},
	{0x6864299111DCC96CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wo_galowint.phy"},
	{0x686442BC224BAE0EULL, "6133040_f.fxe"},
	{0x6864966D91F6077FULL, "vxw_encounter2_camp_9x_0_0.vlm"},
	{0x68649B637FF5930AULL, "6154435_m.fxe"},
	{0x6864A928D4C2A6D6ULL, "230291805.ogg"},
	{0x6864F4E728EAACD3ULL, "6071502_m.fxe"},
	{0x68666E12F432A501ULL, "vxw_encounter3_cc_0_0.vlm"},
	{0x6866EB19B1989F34ULL, "6126832_m.fxe"},
	{0x686728F030A0C49AULL, "art\\characters\\creatures\\desiredemon\\animation\\d_cst.gad"},
	{0x6867781E6B225946ULL, "llac_w_blightlands_0_1.dds"},
	{0x6867B662490E763AULL, "sib100im_tobriusdocs.uti"},
	{0x6867F9A3118BE42BULL, "mcr371_post_harvester6077598.cl"},
	{0x68682C04C1020DE9ULL, "art\\characters\\creatures\\stonegolem\\animation\\wk_p_lf.ani"},
	{0x6868927BC2A6F732ULL, "922906752.ogg"},
	{0x6869548A94A381A8ULL, "w_woundedcoast_105_0.phy"},
	{0x6869A5DB9176FE24ULL, "vxlt_darktown_5wb_0_0.vlm"},
	{0x6869E92493EB3CEFULL, "hlac_ca_caverns1_mines_24421_5y.dds"},
	{0x686A35AB20E81615ULL, "bd_clo.mao"},
	{0x686A4A85C7B1A548ULL, "art\\characters\\playercharacter\\humanfemale\\boo_magb.phy"},
	{0x686A6B7C619F5C53ULL, "vxlt_gallowscourtyard_7io_0_0.vlm"},
	{0x686AC4B70F9D0FB1ULL, "gen00fl_varric6105432.cl"},
	{0x686AC97EA407A501ULL, "dae000_bodhan6146861.cl"},
	{0x686ADAAB503F836BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_exit.evt"},
	{0x686AE8876F005530ULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm.phy"},
	{0x686B0409B53625A2ULL, "232165434.ogg"},
	{0x686B27CBB2750C62ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_1x.dds"},
	{0x686B7D80CF482398ULL, "dae000_arishok6067773.cl"},
	{0x686C0BDCACE641DCULL, "vxlt_gallowscourtyard_5z7_0_0.vlm"},
	{0x686C2035A258782CULL, "vxlt_gallowstemplar_f_5uw_0_0.vlm"},
	{0x686C4024A3226D9EULL, "vxlt_0_lowtown_a3_n_c_e1q_0_0.vlm"},
	{0x686CFFF50B138C3FULL, "6142416_f.fxe"},
	{0x686D0CF1E8D279AAULL, "follower_banter6170289.cl"},
	{0x686D4C851D6ED5BFULL, "1017048334.ogg"},
	{0x686D6FA214E9D254ULL, "vxdr_primevaltomb_8us_0_0.vlm"},
	{0x686D77BEBD575594ULL, "gen00fl_dog6166835.cl"},
	{0x686DA8B4E4253A4EULL, "gen00fl_aveline6139174.cl"},
	{0x686DB162AA634D87ULL, "tbe000im_wisdom_ring.uti"},
	{0x686DF40CB996981EULL, "57249441.ogg"},
	{0x686E2BF030A6E8B1ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_cst.evt"},
	{0x686E6D49010BEC25ULL, "vxlt_warehouse_bh_0_0.vlm"},
	{0x686E923DA381C32FULL, "603595638.ogg"},
	{0x686E9413F8BFA3AFULL, "vxdr_primevalentrance_27h_0_0.vlm"},
	{0x686E98A5D82DD062ULL, "lgt110_orz_harrowmont6109016.cl"},
	{0x686F1223CB09A9CEULL, "c_footstep.wbd"},
	{0x686F1EA195AF17B1ULL, "llac_lt_0_lowtown_docks_n_196607_3.dds"},
	{0x686F3E6DBFFDBB74ULL, "vxw_encounter2_cq_0_0.vlm"},
	{0x686F4704C1085A6BULL, "art\\characters\\creatures\\stonegolem\\animation\\wk_p_lf.gad"},
	{0x686FCC815C97A7D9ULL, "vxht_bartrand_n_68v_0_0.vlm"},
	{0x68702D61E5D7D59BULL, "371472049.ogg"},
	{0x68705DDBE06568B6ULL, "follower_banter6156072.cl"},
	{0x6870BFFCE0E5B7D2ULL, "hlac_w_bonepitmining_a2_4_0z.dds"},
	{0x6870CAD42310424BULL, "vxdr_0_deeproads_ciy_0_0.vlm"},
	{0x6870F3AB5044192CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_exit.gad"},
	{0x68712C628484786AULL, "493432344.ogg"},
	{0x6872092254BEE73CULL, "llac_lt_backalley_g_1_3.dds"},
	{0x6872318A0BD333F9ULL, "99855762.ogg"},
	{0x68724F1E1FDA20A5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08b.phy"},
	{0x6872587312F8F49BULL, "6175952_m.fxe"},
	{0x68725D2EF1E707DAULL, "vxlt_warehouse_n_11p_0_0.vlm"},
	{0x68727CDA55C7FB63ULL, "one140cr_charade6134581.cl"},
	{0x6872B57F7D512AABULL, "zz_followers_banter6159384.cl"},
	{0x687333445F28B408ULL, "lt_0_lowtown_a2_n_c_35_0_lc.mmh"},
	{0x68736AE4114A7D19ULL, "mag100_cullen6095173.cl"},
	{0x68739149E0815247ULL, "hlac_ht_0_hightown_-65534_0z.dds"},
	{0x6873F1CBA781170BULL, "81679420.ogg"},
	{0x6874324021B9F0A0ULL, "qcr220_orsino6060746.cl"},
	{0x687432E33B97ED5BULL, "art\\vfx\\combat\\vfx_glb_ent_imp_maj_c.impa.ani"},
	{0x6874BC1989AD852CULL, "6108959_m.fxe"},
	{0x6874D14528B39596ULL, "vxw_woundedcoast_14f_0_0.vlm"},
	{0x6874F1DDD9B2FE01ULL, "cod_ltr_danzig_wayward.utp"},
	{0x6874FCDBEDAFF9CDULL, "6141939_m.fxe"},
	{0x687500182CCC3ACBULL, "zz_fen_debug6134357.cl"},
	{0x6875AC5FAA175BFAULL, "vxht_mansion2b_3cu_0_0.vlm"},
	{0x6875CEE2E0EBDB75ULL, "vxlt_undercityencounter_ux_0_0.vlm"},
	{0x68764404C10E73D0ULL, "art\\characters\\creatures\\stonegolem\\animation\\wk_p_lf.evt"},
	{0x6876948B40372F69ULL, "dae320ar_hightown.lst"},
	{0x68769527F734CF47ULL, "540890955.ogg"},
	{0x6877427C304F5304ULL, "qcr000_varric6208222.cl"},
	{0x6877E59693DA9EA3ULL, "vxht_mansion2b_2as_0_0.vlm"},
	{0x6877E664A9E78291ULL, "vxca_caverns2_79g_0_0.vlm"},
	{0x6877F2AB504A362AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_exit.ani"},
	{0x68780A8B3A80D393ULL, "6089530_m.fxe"},
	{0x6878D2AAC76D1AF7ULL, "6125319_m.fxe"},
	{0x6878D87AEBEDBAA0ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_hed_suca_0s.dds"},
	{0x6879325A1018A3E3ULL, "521376607.ogg"},
	{0x6879851AFE5A33CDULL, "6015182_m.fxe"},
	{0x687A358D2E8EE22FULL, "6176478_m.fxe"},
	{0x687A4A5D82A7CCE3ULL, "gen00fl_isabela6090314.cl"},
	{0x687A59FC3B2EA056ULL, "720313064.ogg"},
	{0x687AD414C2F179E8ULL, "6162748_m.fxe"},
	{0x687AF48168AF9A32ULL, "949403999.ogg"},
	{0x687B24E06BB817B3ULL, "gen00fl_anders6125456.cl"},
	{0x687B7DCBCE1F45E5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookmisc_d.dds"},
	{0x687B9C43951D4902ULL, "22939708.ogg"},
	{0x687B9CF7771F28DFULL, "795305475.ogg"},
	{0x687C43F030B34E80ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_cst.ani"},
	{0x687C6903E93D5DF0ULL, "6044182_f.fxe"},
	{0x687CCDAFA50BCD9FULL, "6070031_m.fxe"},
	{0x687CDB5DE061BE4DULL, "vxlt_warehouse_n_q0_0_0.vlm"},
	{0x687D5DFA2C326236ULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urvb_0n.dds"},
	{0x687D6C8B403D09A0ULL, "dae320ar_hightown.ncs"},
	{0x687DB04AB1F3D3F4ULL, "vxw_encounter2_nv_0_0.vlm"},
	{0x687E25BCC22CD3E4ULL, "593585287.ogg"},
	{0x687E6D99B75FB9A9ULL, "hlac_lt_0_lowtown_a3_n_c_0_3y.dds"},
	{0x687EC5A9B84746EDULL, "fen110_party_comment6095247.cl"},
	{0x687F24FCCED52F84ULL, "llac_lt_0_lowtown_a2_n_-131072_7.dds"},
	{0x687F3CB115065ACBULL, "vxdr_0_deeproads_cfx_0_0.vlm"},
	{0x687F9A1AE313709EULL, "6051066_m.fxe"},
	{0x687FAA2AC34DE55BULL, "406292325.ogg"},
	{0x6880039037F43EA9ULL, "dae000_fem_priest_mob6174193.cl"},
	{0x688088ECB77FDFE0ULL, "vxca_caverns1_e84_0_0.vlm"},
	{0x68810C62A0124AA0ULL, "vxlt_shop_ky_0_0.vlm"},
	{0x68811D7137557B84ULL, "6208442_m.fxe"},
	{0x6882298F32FDC34BULL, "vxdr_0_deeproads_fa0_0_0.vlm"},
	{0x68823AE33BA43852ULL, "art\\vfx\\combat\\vfx_glb_ent_imp_maj_c.impa.evt"},
	{0x6882B38D978F72ACULL, "689259289.ogg"},
	{0x68830D7DD17D3EBBULL, "469578967.ogg"},
	{0x688357E704342C37ULL, "lgt210_eau_symp_templar6144459.cl"},
	{0x6883D38BFF8F4142ULL, "vxlt_gallowstemplar_46f_0_0.vlm"},
	{0x6883ECB8FFE59513ULL, "vxw_encounter2_i_0_0.vlm"},
	{0x688432124B477553ULL, "hightown_market_guard_sta6169616.cl"},
	{0x688498CF15006CF4ULL, "vxlt_undercitydungeon_1j6_0_0.vlm"},
	{0x6884A84F7D454B56ULL, "mcr310_argument6072306.cl"},
	{0x6884B67319CD1F9EULL, "mag210_arianni6024584.cl"},
	{0x6884E8DC6CAA8922ULL, "dae000_gamlen6149452.cl"},
	{0x68853E37F24AF012ULL, "llac_w_encounter3_0_10.dds"},
	{0x68854017B24A39DBULL, "one300_nuncio6084340.cl"},
	{0x688553EA63723CEAULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower2_meredithsintro.gad"},
	{0x6885756091C8D2B3ULL, "6091957_m.fxe"},
	{0x68858E74E659AF0AULL, "397227174.ogg"},
	{0x6885903CBFE600AEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaritrooper_ranged.anb"},
	{0x68859B6735340EA2ULL, "vxht_mansion2_1b8_0_0.vlm"},
	{0x6885D6105615B9B4ULL, "6030677_f.fxe"},
	{0x68860FBEF4374F66ULL, "257907028.ogg"},
	{0x68864580DF4280EAULL, "6080984_f.fxe"},
	{0x68869C749FE9A1F2ULL, "llac_lt_0_lowtown_n_65535_6.dds"},
	{0x6886A261FD80D00EULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grableft.ani"},
	{0x68870EB5E787A10FULL, "2351240.ogg"},
	{0x68878D8B4045960EULL, "dae320ar_hightown.are"},
	{0x6887F2633A606B68ULL, "w_sundermount_16_0.msh"},
	{0x688804633A6089F9ULL, "w_sundermount_16_0.mao"},
	{0x688810633A609D92ULL, "w_sundermount_16_0.mmh"},
	{0x6888B6223BB5072AULL, "zz_fol_debug6103811.cl"},
	{0x6888BAC8B2E515C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_varian.ani"},
	{0x688903EF8644A92CULL, "vxlt_gallowstemplar_f_5cu_0_0.vlm"},
	{0x688959E33BAA8BE5ULL, "art\\vfx\\combat\\vfx_glb_ent_imp_maj_c.impa.gad"},
	{0x688A36281CF23A98ULL, "vxdr_0_deeproads_euh_0_0.vlm"},
	{0x688A6C877F901A21ULL, "w_sundermount_108_0.phy"},
	{0x688AA91752618F04ULL, "vxdr_primevaltomb_6iu_0_0.vlm"},
	{0x688B16E6E8D7AF64ULL, "vxht_mansion2b_1u7_0_0.vlm"},
	{0x688B21012595AB50ULL, "gen00fl_aveline6116683.cl"},
	{0x688B4350F37408F1ULL, "hlac_lt_backalley_n_1_7x.dds"},
	{0x688BCA2DAE9411BBULL, "vxlt_gallowstemplar_n_4xv_0_0.vlm"},
	{0x688C2227A3E9EC74ULL, "vxw_encounter2_iw_0_0.vlm"},
	{0x688C6BCB8795F5D6ULL, "6129035_m.fxe"},
	{0x688C9790617B38B5ULL, "vxdr_0_deeproads_d2y_0_0.vlm"},
	{0x688CD3D61763D711ULL, "prm000im_apostate_staff.uti"},
	{0x688D2378A10814C2ULL, "vxlt_0_lowtown_docks_as2_0_0.vlm"},
	{0x688D934A582CAFCBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidright.phy"},
	{0x688E8D888B702734ULL, "hlac_lt_0_lowtown_docks_n_131071_2x.dds"},
	{0x688EACAFA69A9EEFULL, "6069607_f.fxe"},
	{0x688EB688C32D0A11ULL, "286841617.ogg"},
	{0x688F04BC3558F24FULL, "168090364.ogg"},
	{0x688FAEBDE2567137ULL, "vxw_encounter2_camp_b5_0_0.vlm"},
	{0x688FBB3CC02C7DAAULL, "lla_ht_0_hightown_n_131073_0.dds"},
	{0x68900E502028FC65ULL, "zz_followers_banter6158046.cl"},
	{0x68902059EE743727ULL, "gen_im_arm_cht_med_act2_01.uti"},
	{0x689034A3102F0BBFULL, "6104992_m.fxe"},
	{0x6890681A7B81C422ULL, "6071293_m.fxe"},
	{0x68912C374E15F2A2ULL, "hlac_dr_0_deeproads_5842_2y.dds"},
	{0x68916D9052FCF1D3ULL, "vxw_bonepitdestroyed_a3_1ph_0_0.vlm"},
	{0x689175877F96483DULL, "w_sundermount_108_0.rml"},
	{0x6891D54ECC84A10CULL, "6069834_m.fxe"},
	{0x6891F81A7E3E6432ULL, "dae000cr_orana.utc"},
	{0x689205E4E0BB934DULL, "vxht_chantry_10q_0_0.vlm"},
	{0x689238AD71367C0CULL, "6166757_m.fxe"},
	{0x68930AAC06F68574ULL, "vxlt_undercitydungeon_1m7_0_0.vlm"},
	{0x68930D47B1D24CE6ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetal_e.dds"},
	{0x689321C4A1CACA06ULL, "hlac_lt_0_lowtown_a2_n_65536_9x.dds"},
	{0x689392E263C90F77ULL, "tree_02.dds"},
	{0x68939A61FD8B4D5FULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grableft.evt"},
	{0x6893AEA98E76DC84ULL, "dae000_hate_enemies6167613.cl"},
	{0x6893FB26AD8921A2ULL, "661828104.ogg"},
	{0x6894350BBC59E05BULL, "dae000_meredith6085384.cl"},
	{0x6894678C0342B20BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext_0.pwk"},
	{0x68949375F3F671CCULL, "240622286.ogg"},
	{0x6894CA1E8DD3B391ULL, "vxca_caverns2_73e_0_0.vlm"},
	{0x689599D87C054367ULL, "6042401_m.fxe"},
	{0x6895E0C5A11B53D0ULL, "vxw_sundermount_210_0_0.vlm"},
	{0x6895F79EA5FE2631ULL, "280385892.ogg"},
	{0x68968249693D99A8ULL, "vxht_mansion_n_6ki_0_0.vlm"},
	{0x6896C11FAF7FDCC3ULL, "6024979_m.fxe"},
	{0x6896CB8B4BDADE2DULL, "dae000st_gamlen.stg"},
	{0x6896D2C8B2F17BBAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_varian.evt"},
	{0x68975BDCDF92A3D6ULL, "530440635.ogg"},
	{0x68977E2AB1BCBB6AULL, "vxca_caverns2_8mk_0_0.vlm"},
	{0x6898E4EB78EC33A3ULL, "w_bonepit_246_0.mao"},
	{0x6898E8EB78EC3AF0ULL, "w_bonepit_246_0.mmh"},
	{0x6898F6EB78EC520AULL, "w_bonepit_246_0.msh"},
	{0x68991AF44457A784ULL, "vxdr_primevaltomb_6lz_0_0.vlm"},
	{0x689968D9FF9C618AULL, "196063632.ogg"},
	{0x68996AE97D64B863ULL, "655351787.ogg"},
	{0x689A2EEA638459D0ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower2_meredithsintro.ani"},
	{0x689A7B61FD913748ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grableft.gad"},
	{0x689ACA2E03CA0263ULL, "614299349.ogg"},
	{0x689AEE4614519052ULL, "vxw_bonepitdestroyed_a3_1pm_0_0.vlm"},
	{0x689AF4A6C6C1154FULL, "mcr310_argument6072701.cl"},
	{0x689AF99FC4A8B40AULL, "vxlt_0_lowtown_n_eqn_0_0.vlm"},
	{0x689B0823CB2ED3FCULL, "c_footstep.bnk"},
	{0x689B16925E685EFFULL, "6076931_f.fxe"},
	{0x689B67A27F960AFFULL, "lt_0_lowtown_docks_n_probe_ctr.dds"},
	{0x689B8DFD04473116ULL, "164331044.ogg"},
	{0x689BF6CC88C90585ULL, "vxw_bonepit_16d_0_0.vlm"},
	{0x689C97EEE19F3E78ULL, "406157441.ogg"},
	{0x689CD1C8B2F5E5ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_varian.gad"},
	{0x689CD499CEB25ECCULL, "vxw_bonepitdestroyed_a3_1wz_0_0.vlm"},
	{0x689CDFD497B8492DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_broken.mmh"},
	{0x689CE3D497B8507EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_broken.mao"},
	{0x689CF1D497B867B3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_broken.msh"},
	{0x689D7146B5438BE6ULL, "6054999_m.fxe"},
	{0x689E67F1D254EFF2ULL, "animation_tuneblendspec.gda"},
	{0x689F52D9D7A50F63ULL, "hlac_lt_ruinedhouse_n_-1_6y.dds"},
	{0x689FB41BCE93F5F1ULL, "272766416.ogg"},
	{0x689FDF6D44F30A53ULL, "vxw_bonepitdestroyed_a3_1uk_0_0.vlm"},
	{0x689FEF94A95B429FULL, "gen00fl_merrill6128866.cl"},
	{0x68A00C58711EA647ULL, "m_dae210ar_lowtown.wbd"},
	{0x68A02771B0FE9680ULL, "vxht_0_hightown_n_b0l_0_0.vlm"},
	{0x68A03214C989FFAEULL, "isa120_hayder6068833.cl"},
	{0x68A07A41BBAB1E2DULL, "plt_cod_cha_orsino.plo"},
	{0x68A0BF035FF5A24CULL, "hlac_dr_dungeon_3216_3y.dds"},
	{0x68A12FD3425BB7EFULL, "hlac_ca_caverns1_24426_14y.dds"},
	{0x68A16C505A2AAC71ULL, "vxlt_0_lowtown_a2_n_hcr_0_0.vlm"},
	{0x68A29F6450574ACFULL, "vxlt_0_lowtown_eqp_0_0.vlm"},
	{0x68A2BAF664599F07ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.anb"},
	{0x68A33486538A5367ULL, "467886570.ogg"},
	{0x68A3368AAD02F76EULL, "ht_brothel_n_3374.anb"},
	{0x68A363A1CA2F2F49ULL, "dr_primevaltomb_9621_lc.mmh"},
	{0x68A370FFAE1EC0D4ULL, "176305040.ogg"},
	{0x68A42BE2ADCC625AULL, "vxht_0_hightown_a2_9uu_0_0.vlm"},
	{0x68A4F4265B33B228ULL, "vxht_mansion_n_6nf_0_0.vlm"},
	{0x68A5009112141A37ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wo_galowint.mmh"},
	{0x68A5049112142144ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wo_galowint.mao"},
	{0x68A5169112143FA9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3wo_galowint.msh"},
	{0x68A5236F0BC0FE25ULL, "vxht_0_hightown_n_bxu_0_0.vlm"},
	{0x68A562A780D4C18EULL, "6047696_m.fxe"},
	{0x68A5914D3F80FB1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_dead_loop.gad"},
	{0x68A5A89B8F2D1185ULL, "148425755.ogg"},
	{0x68A5B24BADBC62E8ULL, "vxht_0_hightown_n_c3t_0_0.vlm"},
	{0x68A5DD417FBA8E86ULL, "zz_slate5.cut"},
	{0x68A6050648BF4F2EULL, "vxca_cavern1_haunted_jga_0_0.vlm"},
	{0x68A605922766F78FULL, "366559409.ogg"},
	{0x68A6B636EA39FDB2ULL, "vfx_lightning_f_05.dds"},
	{0x68A6BEAA372C6EC2ULL, "hlac_lt_0_lowtown_a2_-196609_3x.dds"},
	{0x68A752D2D7BE7508ULL, "gen_im_npc_arm_hum_glv_krka.uti"},
	{0x68A78EC8E677451EULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.impa.ani"},
	{0x68A7AA571E9299D9ULL, "vxdr_primevaltomb_4h4_0_1.vlm"},
	{0x68A8CC2D1E75ADDAULL, "6146951_m.fxe"},
	{0x68A942F0EFFB3512ULL, "vxlt_0_lowtown_a2_n_c_hue_0_0.vlm"},
	{0x68A94A9BF28EB715ULL, "6143735_m.fxe"},
	{0x68A9562E096F11E2ULL, "hlac_lt_house_-1_1y.dds"},
	{0x68A967CF1AC2F9D7ULL, "vxdr_primevalentrance_2oh_0_0.vlm"},
	{0x68AAD04286576CBBULL, "vxlt_gallowstemplar_f_46d_0_0.vlm"},
	{0x68AB9684DA4741ECULL, "27265718.ogg"},
	{0x68AC61F975F49420ULL, "vxlt_shop_dd_0_0.vlm"},
	{0x68AC680539FBC2D2ULL, "6144401_m.fxe"},
	{0x68ACD2ADCB5C3ABBULL, "mer320_hubert6053670.cl"},
	{0x68ACF05BCAE038A0ULL, "vxlt_0_lowtown_docks_n_b7y_0_1.vlm"},
	{0x68AD3315ABFD8F0BULL, "em_dalish_apprentice2.mor"},
	{0x68ADC5575B995504ULL, "6044299_m.fxe"},
	{0x68ADD8D497C6B0EAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_broken.phy"},
	{0x68AE6F35B8FBE94DULL, "6042641_m.fxe"},
	{0x68AE7F737A843A9EULL, "vxlt_gallowstemplar_n_6s1_0_1.vlm"},
	{0x68AE8523086A4B95ULL, "667097824.ogg"},
	{0x68AE9298CBA77792ULL, "vxlt_hangedman_n_20t_0_0.vlm"},
	{0x68AF3AA118052828ULL, "ef_maren.mor"},
	{0x68AF4E551C53C04BULL, "171627458.ogg"},
	{0x68AFBA93A1C703D3ULL, "vxca_cavern1_haunted_i2d_0_0.vlm"},
	{0x68B041A77F4910CDULL, "117953469.ogg"},
	{0x68B04FFBC5C4635FULL, "6069895_m.fxe"},
	{0x68B0ED2A4D692AE9ULL, "hlac_w_blightlands_-3_0x.dds"},
	{0x68B170FC700F8EA6ULL, "one000_ilen6057362.cl"},
	{0x68B1D4CB56AD631FULL, "6150997_m.fxe"},
	{0x68B1D5CB1F33FA39ULL, "571672681.ogg"},
	{0x68B1F8EFABF6E279ULL, "gen00fl_varric6075151.cl"},
	{0x68B3850892297D71ULL, "hlac_lt_0_lowtown_docks_n_-1_0x.dds"},
	{0x68B3F5743BC6568AULL, "hlac_lt_gallowscourtyard_n_0_0y.dds"},
	{0x68B424289FB27B68ULL, "vxht_0_hightown_n_c8u_0_0.vlm"},
	{0x68B4B17025D7E05CULL, "follower_banter6143151.cl"},
	{0x68B4C142C397EB30ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomea.mao"},
	{0x68B4C542C397F243ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomea.mmh"},
	{0x68B4D342C398099DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_tomea.msh"},
	{0x68B4D65139183B60ULL, "hlac_lt_0_lowtown_docks_-65539_1z.dds"},
	{0x68B4E46F29248EDCULL, "6053206_m.fxe"},
	{0x68B4EDF8B555E9CCULL, "vxlt_0_lowtown_a3_n_c_en2_0_0.vlm"},
	{0x68B529E9CFDD1760ULL, "172767220.ogg"},
	{0x68B5436E5D3AD802ULL, "llac_ca_caverns2prc_5802_8.dds"},
	{0x68B5A6C8E683ABEFULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.impa.evt"},
	{0x68B5AA309CB71BAEULL, "639888048.ogg"},
	{0x68B5BBBC191404B3ULL, "gen00fl_dog6169758.cl"},
	{0x68B5FEAE2843D884ULL, "vxdr_primevaltomb_6vx_0_0.vlm"},
	{0x68B6CCF8EB950CACULL, "6088066_m.fxe"},
	{0x68B6E5D1EFC28E2FULL, "6135119_f.fxe"},
	{0x68B6F45CE6E5D5EFULL, "6206861_m.fxe"},
	{0x68B70FAB897E06B4ULL, "vxht_keep_n_3xu_0_0.vlm"},
	{0x68B73E0A106BC65AULL, "6143651_m.fxe"},
	{0x68B7B4CDE1F14D92ULL, "vxlt_0_lowtown_a2_n_c_hpf_0_0.vlm"},
	{0x68B7CBCD0464BA03ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_017a_0t.dds"},
	{0x68B7DD7AB9574C49ULL, "6140131_f.fxe"},
	{0x68B8812A7112B962ULL, "6115753_m.fxe"},
	{0x68B89E35DDCECEF8ULL, "vxlt_backalley_g_1w8_0_0.vlm"},
	{0x68B8FB64D2683373ULL, "vxlt_0_lowtown_docks_a3_ctk_0_0.vlm"},
	{0x68B93B49631BA252ULL, "vxca_cavern1_haunted_i2a_0_0.vlm"},
	{0x68B9409FF23301F2ULL, "vxw_encounter2_camp_p6_0_0.vlm"},
	{0x68B9A6A8A79B2070ULL, "pro000_athenril6048704.cl"},
	{0x68B9D329E09C7955ULL, "vxlt_0_lowtown_a3_n_c_hpt_0_0.vlm"},
	{0x68B9FD1311EE6304ULL, "vxlt_undercityserialkiller_ww_0_0.vlm"},
	{0x68BA2D8131C75380ULL, "vxdr_primevalentrance_39l_0_0.vlm"},
	{0x68BA6E4D3F931C65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_dead_loop.ani"},
	{0x68BABA43A999B29DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic_4.mao"},
	{0x68BADD0A9405233DULL, "349957215.ogg"},
	{0x68BADDE33D8024A7ULL, "6036914_m.fxe"},
	{0x68BB1DAEDD0C2103ULL, "6121980_m.fxe"},
	{0x68BB58FDA4F61EBEULL, "638787070.ogg"},
	{0x68BB76A09553964EULL, "984958085.ogg"},
	{0x68BB826987641033ULL, "6083600_m.fxe"},
	{0x68BBA7C8E6881898ULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.impa.gad"},
	{0x68BBF02901D90ABEULL, "mcr310ar_docks_nt.ncs"},
	{0x68BBF9CB272D7C14ULL, "art\\levels\\races\\proxy\\props\\_textures\\coins_d.dds"},
	{0x68BC1A599F8A8FD6ULL, "6162100_m.fxe"},
	{0x68BCC32728DF3B53ULL, "vxw_bonepitdestroyed_a3_1cm_0_0.vlm"},
	{0x68BD2C4130EDFDB7ULL, "art\\levels\\races\\design\\prp_bodybagb_1.mao"},
	{0x68BD3FBC0DD5B6CDULL, "llac_dr_dungeon_3214_5.dds"},
	{0x68BD9B622B511EBDULL, "w_sundermount_117_0.phy"},
	{0x68BDBBEF8F1A2EDAULL, "gen00fl_dog6167247.cl"},
	{0x68BDF653F0FA3CCEULL, "dae000cr_brothel_harlan.utc"},
	{0x68BE12A7B0862753ULL, "primitivetext_d.mao"},
	{0x68BEA937D79EB204ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_002a_0n.dds"},
	{0x68BEC8A6D03DD534ULL, "w_bonepitdestroyed_a3_37_0_lc.mmh"},
	{0x68BF2CA5CCAD80B2ULL, "6046920_f.fxe"},
	{0x68BF3CF473703CC1ULL, "vxca_caverns2prc_4x7_0_1.vlm"},
	{0x68BFF14493DF2520ULL, "vxlt_gallowscourtyard_73z_0_0.vlm"},
	{0x68C02F9A466E10DDULL, "6107450_m.fxe"},
	{0x68C0F618D50C5204ULL, "qun221_viscount6095603.cl"},
	{0x68C1166D2DBAF18AULL, "gen00fl_merrill6128685.cl"},
	{0x68C1A8983AA405AAULL, "art\\cinematics\\act1\\cs_fp_openstare.gad"},
	{0x68C20728EFAD2F25ULL, "vxht_0_hightown_n_bns_0_0.vlm"},
	{0x68C212F65FCFBACDULL, "vxht_mansion2b_1kq_0_1.vlm"},
	{0x68C2C82901DEE5F7ULL, "mcr310ar_docks_nt.lst"},
	{0x68C3293D814284E7ULL, "vxca_caverns2_7qi_0_0.vlm"},
	{0x68C33EE368204892ULL, "6113819_f.fxe"},
	{0x68C3D90A0B7E2181ULL, "dae000ip_crowd_em4_talk17.utp"},
	{0x68C474A2C8E840DDULL, "296306830.ogg"},
	{0x68C4935C89A46D89ULL, "llac_lt_0_lowtown_docks_-65537_2.dds"},
	{0x68C49C622B573F49ULL, "w_sundermount_117_0.rml"},
	{0x68C4D5120A792A4EULL, "dae000_light_tombs_boss6179466.cl"},
	{0x68C50824050D2B6EULL, "hf_amb_shopper.mor"},
	{0x68C55C1ACCFFF1B3ULL, "417762309.ogg"},
	{0x68C5A234823D002CULL, "6144469_m.fxe"},
	{0x68C5B7C3D35D0E93ULL, "vxlt_hangedman_n_1df_0_0.vlm"},
	{0x68C5E2B35E351CF8ULL, "vxw_bonepitmining_a2_1mb_0_0.vlm"},
	{0x68C5F7F257ACA159ULL, "6139550_m.fxe"},
	{0x68C627DD6D59458EULL, "art\\characters\\weapons\\staff\\w_stf_015a.mmh"},
	{0x68C633DD6D595A25ULL, "art\\characters\\weapons\\staff\\w_stf_015a.mao"},
	{0x68C641DD6D59727CULL, "art\\characters\\weapons\\staff\\w_stf_015a.msh"},
	{0x68C675B68385A658ULL, "vxht_0_hightown_a3_caq_0_1.vlm"},
	{0x68C70D0F8ACFDDB4ULL, "93355948.ogg"},
	{0x68C728F43601C76CULL, "vfx_vfx_p5.vsh.fxo.dx11.assembly"},
	{0x68C7346E4A79343EULL, "vxlt_0_lowtown_docks_b5r_0_1.vlm"},
	{0x68C7785FD742E097ULL, "vxlt_0_lowtown_a2_gvo_0_0.vlm"},
	{0x68C7B27CE4291A72ULL, "vxw_encounter2_camp_k7_0_0.vlm"},
	{0x68C7B7AB9389B942ULL, "496426568.ogg"},
	{0x68C82A30A494A84CULL, "vxw_bonepitdestroyed_a3_1fy_0_0.vlm"},
	{0x68C87FF83624A465ULL, "vxlt_warehouse_n_15w_0_0.vlm"},
	{0x68C89F5E23BD6C00ULL, "vxdr_primevalentrance_32m_0_0.vlm"},
	{0x68C9FEB646D315E8ULL, "vxw_woundedcoast_o8_0_0.vlm"},
	{0x68CA045534144512ULL, "6149900_f.fxe"},
	{0x68CA06DBE2F35AC2ULL, "6033633_m.fxe"},
	{0x68CA57A51E0C03B3ULL, "6141281_m.fxe"},
	{0x68CA8823B7BB7D19ULL, "37009262.ogg"},
	{0x68CB35041AD553D3ULL, "vxw_bonepitdestroyed_a3_1dl_0_0.vlm"},
	{0x68CB46DA5FB781E2ULL, "6113878_m.fxe"},
	{0x68CB540499BC3917ULL, "289806114.ogg"},
	{0x68CBB7F6647C9F08ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.mmh"},
	{0x68CBEE3FC73CCE4DULL, "llac_lt_0_lowtown_a3_n_c_-131072_5.dds"},
	{0x68CC705B26457F7BULL, "74508726.ogg"},
	{0x68CD3067F8B8DB06ULL, "hlac_ca_cave_2194_10x.dds"},
	{0x68CD80103D9B52FBULL, "hlac_ht_0_hightown_n_-65536_2y.dds"},
	{0x68CDF9AA82DCA09AULL, "vxlt_hangedman_n_sa_0_0.vlm"},
	{0x68CE6E911EAD5E39ULL, "hlac_lt_gallowstemplar_2384_1y.dds"},
	{0x68CE90622B5F7F29ULL, "w_sundermount_117_0.mao"},
	{0x68CE9C622B5F93C2ULL, "w_sundermount_117_0.mmh"},
	{0x68CE9E622B5F9778ULL, "w_sundermount_117_0.msh"},
	{0x68CEB6EEC289132AULL, "347665610.ogg"},
	{0x68CEF4CDED700205ULL, "965305081.ogg"},
	{0x68CF15956938AC47ULL, "vxlt_0_lowtown_docks_a3_n_9fh_0_0.vlm"},
	{0x68CFC39B1B32EE83ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9mw_windarkupper.mao"},
	{0x68CFC79B1B32F5D0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9mw_windarkupper.mmh"},
	{0x68CFD59B1B330DEAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9mw_windarkupper.msh"},
	{0x68D01C2DCA908A96ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.rearup.evt"},
	{0x68D078E12361FF16ULL, "6103783_m.fxe"},
	{0x68D0F6FEA3E87C73ULL, "vxlt_0_lowtown_n_bz9_0_0.vlm"},
	{0x68D19B1A73389D67ULL, "vxca_caverns2_7vh_0_0.vlm"},
	{0x68D1E5385ED16B0BULL, "6124697_m.fxe"},
	{0x68D2EF4699FFB195ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube_tool.msh"},
	{0x68D2F14699FFB50BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube_tool.mmh"},
	{0x68D2FD4699FFC9A8ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube_tool.mao"},
	{0x68D32D706B53464AULL, "vxw_bonepitdestroyed_a3_27k_0_0.vlm"},
	{0x68D38140A74B02F9ULL, "667366289.ogg"},
	{0x68D3A6D5FEEA01C2ULL, "m_mer200ar_exterior.bnk"},
	{0x68D3C92901ED5A0CULL, "mcr310ar_docks_nt.are"},
	{0x68D3C9592EEA80C1ULL, "vxht_bartrand_n_65m_0_0.vlm"},
	{0x68D41411743318ACULL, "6141769_m.fxe"},
	{0x68D424B52B3E032FULL, "355029322.ogg"},
	{0x68D445210BD73F2DULL, "221743606.ogg"},
	{0x68D45490502B3578ULL, "vxw_bonepitmining_a2_1jc_0_0.vlm"},
	{0x68D4B5B9DC29F252ULL, "vxw_bonepitdestroyed_a3_1di_0_0.vlm"},
	{0x68D4D9691EFF62DDULL, "art\\characters\\playercharacter\\humanmale\\hlf_meda.phy"},
	{0x68D5047478A7664DULL, "6149023_m.fxe"},
	{0x68D5780514BC5939ULL, "752221673.ogg"},
	{0x68D581EFC1BAFFF8ULL, "vxlt_backalley_g_1y6_0_0.vlm"},
	{0x68D583983AB46F90ULL, "art\\cinematics\\act1\\cs_fp_openstare.ani"},
	{0x68D583AEA7CA3B0FULL, "68434251.ogg"},
	{0x68D5EA3CC938F917ULL, "vxlt_0_lowtown_a2_gmn_0_0.vlm"},
	{0x68D61675CF9953B4ULL, "qun200_party_comment6090411.cl"},
	{0x68D62459D61F32F2ULL, "vxw_encounter2_camp_f0_0_0.vlm"},
	{0x68D62E8E3CCC06E8ULL, "mag110_arianni6032478.cl"},
	{0x68D6398AAD2E55D2ULL, "ht_brothel_n_3374.rml"},
	{0x68D6414BFA141EE3ULL, "vxht_keep_73p_0_0.vlm"},
	{0x68D6558D52CEB103ULL, "mag120tr_eau_macha.utt"},
	{0x68D65EC0D12A7721ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6bluelyrium.phy"},
	{0x68D67D6A5889E9FCULL, "vxw_bonepitmining_a2_1x8_0_1.vlm"},
	{0x68D6C80542E53A58ULL, "546199705.ogg"},
	{0x68D6F7944F00C94BULL, "417459092.ogg"},
	{0x68D784CE8A0438D3ULL, "vxdr_primevaltomb_8mh_0_0.vlm"},
	{0x68D7B8F1C1FB4275ULL, "hlac_lt_0_lowtown_1_0y.dds"},
	{0x68D7F1631D3604D9ULL, "gentr_ambient.utt"},
	{0x68D8709338C92E68ULL, "vxw_woundedcoast_d9_0_0.vlm"},
	{0x68D9EACAACB6986AULL, "vxlt_gallowstemplar_5ox_0_0.vlm"},
	{0x68DAAD78E7BAC4FBULL, "plt_lgt110pt_arl_connor.nss"},
	{0x68DAF326E16EAFCBULL, "241232083.ogg"},
	{0x68DB3695AA303D47ULL, "hlac_lt_0_lowtown_a2_n_c_-1_6y.dds"},
	{0x68DBC3A4B48CBB91ULL, "94941219.ogg"},
	{0x68DBF820BCA4E7A8ULL, "6116862_m.fxe"},
	{0x68DCA47A1E63D9D0ULL, "vxw_encounter2_3e_0_0.vlm"},
	{0x68DCD4FE77CB5620ULL, "vxlt_gallowscourtyard_79x_0_0.vlm"},
	{0x68DCF6C8B31BDEBAULL, "zz_followers_banter6157370.cl"},
	{0x68DD3B4F4CB2C4E6ULL, "zz_addpoints.ncs"},
	{0x68DD505736F16CF6ULL, "35543675.ogg"},
	{0x68DDFC90709EFD80ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hanginglight_n.dds"},
	{0x68DDFFBD84CFDAF6ULL, "6070382_m.fxe"},
	{0x68DE377C6BCF4A88ULL, "6135938_f.fxe"},
	{0x68DE77028E9419E6ULL, "gen00fl_isabela_init6110287.cl"},
	{0x68DE9F1A54734917ULL, "llac_lt_0_lowtown_a2_n_c_1_11.dds"},
	{0x68DEC5FFB7EC2086ULL, "gen_im_arm_cht_coma_t1.uti"},
	{0x68DEE3C0EC7E395CULL, "gen00fl_aveline6115426.cl"},
	{0x68DEEB7A86342D98ULL, "6089861_m.fxe"},
	{0x68DF12E6132788B8ULL, "86716491.ogg"},
	{0x68DF495E07818C7DULL, "475580644.ogg"},
	{0x68DFF842FDAA035DULL, "866555484.ogg"},
	{0x68E008497AF61774ULL, "vxht_brothel_n_6ed_0_0.vlm"},
	{0x68E0165871549B7DULL, "m_dae210ar_lowtown.bnk"},
	{0x68E021C1C665AA9BULL, "mcr370_templar26087015.cl"},
	{0x68E034C1482B3732ULL, "mag320_orsino6071254.cl"},
	{0x68E0500F3AEC11FBULL, "hlac_ca_caverns1_24425_3x.dds"},
	{0x68E0D46ED176C39BULL, "6091841_m.fxe"},
	{0x68E1085E4B9E52F6ULL, "6108526_m.fxe"},
	{0x68E1B7118BE08346ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.impa.evt"},
	{0x68E1C28C434ABE3FULL, "sib100_father_contact6108443.cl"},
	{0x68E21CF116EE6626ULL, "6121133_f.fxe"},
	{0x68E2268C989A9479ULL, "6086186_m.fxe"},
	{0x68E320B179F2B3F4ULL, "vxw_blightlands_7r_0_0.vlm"},
	{0x68E376516A9D7ECDULL, "6142123_m.fxe"},
	{0x68E397FFC90ACE18ULL, "6027348_m.fxe"},
	{0x68E3A1DC6CA83BBFULL, "hlac_lt_0_lowtown_n_131070_5x.dds"},
	{0x68E3ED8C1DFB2E01ULL, "119421713.ogg"},
	{0x68E4525CDA67AD5FULL, "6023985_m.fxe"},
	{0x68E46FCCC32FC6D5ULL, "gen_im_acc_amu_act3_02.uti"},
	{0x68E49636C8154B72ULL, "vxw_encounter2_camp_a1_0_0.vlm"},
	{0x68E4F0A99855BA0FULL, "static_static_p80.psh.fxo.dx11"},
	{0x68E55AED2F1F81D8ULL, "vxlt_0_lowtown_docks_n_b49_0_0.vlm"},
	{0x68E5BD84140786B7ULL, "275302604.ogg"},
	{0x68E5C6B2A7E6CC39ULL, "vxlt_gallowscourtyard_7ey_0_0.vlm"},
	{0x68E5CE691F0DC349ULL, "art\\characters\\playercharacter\\humanmale\\hlf_meda.mao"},
	{0x68E5DA691F0DD7E2ULL, "art\\characters\\playercharacter\\humanmale\\hlf_meda.mmh"},
	{0x68E5DC691F0DDA98ULL, "art\\characters\\playercharacter\\humanmale\\hlf_meda.msh"},
	{0x68E6063E07FA4CE7ULL, "6175808_m.fxe"},
	{0x68E63A317DC3003EULL, "6030618_m.fxe"},
	{0x68E653CAD5FDCCB0ULL, "llac_lt_0_lowtown_docks_-196608_1.dds"},
	{0x68E6E2702ABF46E8ULL, "vxw_woundedcoast_a6_0_0.vlm"},
	{0x68E6E54C2326119CULL, "hlac_lt_0_lowtown_n_65534_5z.dds"},
	{0x68E74BC8A20E488AULL, "6110633_m.fxe"},
	{0x68E7D98CF1F7C0E6ULL, "bee_groud.dds"},
	{0x68E82E09A727B41BULL, "lgt000pt_tombs_first.ncs"},
	{0x68E89153672C8202ULL, "6101815_m.fxe"},
	{0x68E8CE118BE6C891ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.impa.gad"},
	{0x68E92DEE8A6F5B86ULL, "zz_followers_banter6157231.cl"},
	{0x68E9E4D30DCAF9D5ULL, "hlac_lt_0_lowtown_docks_n_65533_1y.dds"},
	{0x68EA5B16339BED2DULL, "lowtown_fem_urchin_mob6165917.cl"},
	{0x68EA5ED4786E4F30ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_07skeleton.mao"},
	{0x68EA62D4786E5643ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_07skeleton.mmh"},
	{0x68EA70D4786E6D9DULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_07skeleton.msh"},
	{0x68EA99AA9969AF59ULL, "193860482.ogg"},
	{0x68EAE9A976D77930ULL, "framebuffereffects_diffusiondof_p2.psh.fxo.dx11"},
	{0x68EB10633AB51B39ULL, "w_sundermount_16_0.rml"},
	{0x68EB49C615EADF32ULL, "6020079_m.fxe"},
	{0x68EBD411D8B7BBB2ULL, "zz_dae_debug6029914.cl"},
	{0x68EC2A8F2F40F4A6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\blendtree\\em_exp_movement.anb"},
	{0x68EC6FF778F19D6EULL, "vxlt_0_lowtown_a2_cmo_0_0.vlm"},
	{0x68EC71877FE2ED5DULL, "w_sundermount_108_0.mao"},
	{0x68EC75877FE2F396ULL, "w_sundermount_108_0.mmh"},
	{0x68EC784B6BED4EF5ULL, "47801568.ogg"},
	{0x68EC7F877FE30494ULL, "w_sundermount_108_0.msh"},
	{0x68ED65DD05718D4BULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.mmh"},
	{0x68ED70A5B49A2044ULL, "138027087.ogg"},
	{0x68EDC1E866537354ULL, "vxw_woundedcoast_kx_0_0.vlm"},
	{0x68EE5D5F5BF5A0B7ULL, "vxca_caverns1_ics_0_1.vlm"},
	{0x68EE9A6C6760EC9CULL, "708092383.ogg"},
	{0x68EF694348049D3BULL, "6066274_m.fxe"},
	{0x68EF8524954DCF63ULL, "732042074.ogg"},
	{0x68EF86288C119C73ULL, "one300_nuncio6083402.cl"},
	{0x68EFCF118BECE93FULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.impa.ani"},
	{0x68EFDD5B59187E99ULL, "mag110_samson6018724.cl"},
	{0x68F036AF872C739EULL, "470883114.ogg"},
	{0x68F1424BE32B7F97ULL, "6049134_m.fxe"},
	{0x68F171D2EC654D0BULL, "hlac_ht_brothel_n_3372_0y.dds"},
	{0x68F1928E6BE8CC74ULL, "vxw_blightlands_8s_0_0.vlm"},
	{0x68F1D5620F7F6BA7ULL, "art\\characters\\creatures\\stonegolem\\animation\\wkf.gad"},
	{0x68F1DF4C59342F01ULL, "6135132_m.fxe"},
	{0x68F1F598E5723091ULL, "6027374_m.fxe"},
	{0x68F31B01D67EBA23ULL, "gen00fl_isabela6075272.cl"},
	{0x68F33509BC121E84ULL, "w_bonepit_217_0.msh"},
	{0x68F34709BC123DEDULL, "w_bonepit_217_0.mao"},
	{0x68F34B09BC124426ULL, "w_bonepit_217_0.mmh"},
	{0x68F35443F6146C6EULL, "hlac_dr_primevaltomb_658_12x.dds"},
	{0x68F36291B31E8DA5ULL, "mag100_karras6052917.cl"},
	{0x68F4AB96E0465CFDULL, "86612163.ogg"},
	{0x68F53C7100E77136ULL, "mag320_orsino6071104.cl"},
	{0x68F69D0395F34B33ULL, "509574339.ogg"},
	{0x68F6AA3A183BA84AULL, "fonts_fr-fr.gfx"},
	{0x68F6F2631D50B69CULL, "gentr_ambient.ncs"},
	{0x68F70B5CE535D0BFULL, "mag200_feynriel6023873.cl"},
	{0x68F8CA620F8577B4ULL, "art\\characters\\creatures\\stonegolem\\animation\\wkf.evt"},
	{0x68F8F8E030100952ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_meda_0s.dds"},
	{0x68F96E5BA507631CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault.mao"},
	{0x68F97A5BA50777BFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault.mmh"},
	{0x68F9805BA5078261ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault.msh"},
	{0x68F9883402500AD0ULL, "vxw_encounter2_9c_0_0.vlm"},
	{0x68F990D5FF0AABE0ULL, "m_mer200ar_exterior.wbd"},
	{0x68F991C84722A39EULL, "628532031.ogg"},
	{0x68F9DE1F500564CAULL, "375643756.ogg"},
	{0x68F9F9E6C823358EULL, "6208596_m.fxe"},
	{0x68FA36A5AF1D0F95ULL, "6143718_f.fxe"},
	{0x68FA3AD0C3DC6B15ULL, "hlac_w_encounter3_-65537_0y.dds"},
	{0x68FA426B6C868C14ULL, "222662299.ogg"},
	{0x68FA4853F6A575D3ULL, "elf_female_heads.gda"},
	{0x68FAA5F77B914419ULL, "dae000_fenarel6127243.cl"},
	{0x68FB46DD057D95B0ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.anb"},
	{0x68FB4A3696831C07ULL, "6042364_m.fxe"},
	{0x68FB57B1AC3A0503ULL, "90251983.ogg"},
	{0x68FBFF21B8C0FE5FULL, "genip_chest_normal2.utp"},
	{0x68FC01E8AE1B51B5ULL, "215038976.ogg"},
	{0x68FC3B1DA66BF95CULL, "6180960_m.fxe"},
	{0x68FCEC035EE24874ULL, "vxht_brothel_n_6kj_0_0.vlm"},
	{0x68FD4A6342354209ULL, "hlac_ht_0_hightown_a2_65533_3y.dds"},
	{0x68FD8CBF4D0EE50CULL, "379669602.ogg"},
	{0x68FDB4530D33A000ULL, "vxlt_0_lowtown_eo9_0_0.vlm"},
	{0x68FEB29E1F52FB12ULL, "169268255.ogg"},
	{0x68FF0D858CB47D29ULL, "art\\levels\\races\\design\\prp_ballista.phy"},
	{0x68FF3623F7EF29E9ULL, "w_bonepit_44_0.rml"},
	{0x68FF3D3A8985F31DULL, "6044618_m.fxe"},
	{0x68FFF20B742E4639ULL, "vxlt_0_lowtown_docks_bcb_0_0.vlm"},
	{0x6900109BB75DEDEBULL, "mag371_elsa6041747.cl"},
	{0x6900460DE0DA00AEULL, "vxlt_backalley_g_1ed_0_0.vlm"},
	{0x690051EADC6662D2ULL, "63431904.ogg"},
	{0x69008884A047132DULL, "hlac_ht_0_hightown_a2_-131072_1z.dds"},
	{0x6900CA438F2B7BEEULL, "6039851_m.fxe"},
	{0x6900DC00B0498292ULL, "vxlt_ruinedhouse_n_i5_0_0.vlm"},
	{0x6900EE8DAB15FC08ULL, "6139871_f.fxe"},
	{0x6901B440DC19669BULL, "vxht_brothel_n_613_0_0.vlm"},
	{0x6901C274BC456EF9ULL, "6043726_f.fxe"},
	{0x6901CDCF2309A93AULL, "vxht_0_hightown_a3_bka_0_0.vlm"},
	{0x6901F9623DA976CAULL, "489743072.ogg"},
	{0x6902CB4026809891ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_ars_0d.dds"},
	{0x6902D6259324CA8AULL, "6084734_m.fxe"},
	{0x6903241FA5EF31A5ULL, "vxw_encounter3_7c_0_0.vlm"},
	{0x690358546D92E704ULL, "6066174_m.fxe"},
	{0x69038CC6394D074CULL, "vxlt_0_lowtown_docks_b0n_0_0.vlm"},
	{0x6903E8700C852A9EULL, "hlac_w_sundermount_-131071_0y.dds"},
	{0x6903F53C4C2C35B8ULL, "static_static_p25.psh.fxo.dx11.assembly"},
	{0x6903F6F4E4F8F417ULL, "art\\vfx\\combat\\vfx_gib_hlk_p_5.mao"},
	{0x6903FA2914B01F9BULL, "zz_isa_debug6175464.cl"},
	{0x690478389C403A95ULL, "fx_cloud01.dds"},
	{0x6904B15FD9714BC1ULL, "6024116_f.fxe"},
	{0x690523E8FE2F46A8ULL, "42476805.ogg"},
	{0x690595A3D1369A98ULL, "mcr310_party_banter6087782.cl"},
	{0x69059A1657CCCEE8ULL, "hlac_lt_0_lowtown_a2_1_4z.dds"},
	{0x6905E0996D474933ULL, "vxlt_0_lowtown_n_gjx_0_0.vlm"},
	{0x69061BBFAC7C8E3CULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_2.mao"},
	{0x69063C1719247256ULL, "follower_banter6151468.cl"},
	{0x69063DC2F40F4098ULL, "228254605.ogg"},
	{0x69068B03CEC9E418ULL, "vxlt_darktown_518_0_0.vlm"},
	{0x6906941639B53F0DULL, "ico_crossbow_2.dds"},
	{0x6906AB3EC4520A6DULL, "lgt320_mis_resp_swordhunter6150262.cl"},
	{0x6906C2620F91A7EDULL, "art\\characters\\creatures\\stonegolem\\animation\\wkf.ani"},
	{0x6906E7DD878C1675ULL, "hlac_ca_caverns2prc_5802_12x.dds"},
	{0x6906E89599C45A5CULL, "vxht_0_hightown_n_b00_0_0.vlm"},
	{0x69072DB2EC199782ULL, "vxht_0_hightown_n_ay2_0_0.vlm"},
	{0x6907332AE721E093ULL, "vxca_cavern1_haunted_ebp_0_0.vlm"},
	{0x6907AB500EFBFBB7ULL, "6151082_m.fxe"},
	{0x6907CE141E5AC35DULL, "hlac_w_encounter2_camp_196611_0y.dds"},
	{0x6907EF5FDCDC69EBULL, "308212364.ogg"},
	{0x69092A23F7F769C9ULL, "w_bonepit_44_0.mao"},
	{0x69093623F7F77E62ULL, "w_bonepit_44_0.mmh"},
	{0x69093823F7F78118ULL, "w_bonepit_44_0.msh"},
	{0x6909B878E7E24794ULL, "plt_lgt110pt_arl_connor.plo"},
	{0x6909C49BE77C0530ULL, "vxlt_0_lowtown_enz_0_0.vlm"},
	{0x6909E21710620B00ULL, "gen00fl_isabela6099705.cl"},
	{0x690B4BB03F7B1230ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowc.msh"},
	{0x690B59B03F7B2A4AULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowc.mmh"},
	{0x690B5DB03F7B3081ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowc.mao"},
	{0x690C92264C8E0078ULL, "17375407.ogg"},
	{0x690CF8CAC9645D82ULL, "mer341_jansen6054883.cl"},
	{0x690D70D9C66A0C3DULL, "vxlt_0_lowtown_f4b_0_0.vlm"},
	{0x690DACE76A480120ULL, "mer221_sabin6047439.cl"},
	{0x690E46865567CA83ULL, "6128877_m.fxe"},
	{0x690E4963F40A0EA8ULL, "01_noise.mao"},
	{0x690E63E866245EB9ULL, "vxlt_0_lowtown_docks_bfa_0_0.vlm"},
	{0x690EF1EE943AAAFEULL, "vxht_mansion2_1v8_0_0.vlm"},
	{0x690F61F597F301FFULL, "sky_lt_undercityserialkiller.mao"},
	{0x690F96D3D2A22F95ULL, "215991831.ogg"},
	{0x690FDE934962CEB7ULL, "zz_followers_banter6159591.cl"},
	{0x690FF1F879D12A66ULL, "mag371tr_close_meredith.utt"},
	{0x69101CF8D8CA6F34ULL, "vfx_distortionmask_p0.vsh.fxo"},
	{0x6910458BF3631CCDULL, "hlac_lt_gallowsprison_n_-1_0z.dds"},
	{0x69104CDF890470F5ULL, "255638332.ogg"},
	{0x6910720D4B9729DCULL, "6143370_m.fxe"},
	{0x6910899B0A3E9E93ULL, "81886073.ogg"},
	{0x6910D794B1F98D2DULL, "356580062.ogg"},
	{0x6911241F451FA930ULL, "follower_banter6168321.cl"},
	{0x69112EFEA9BD4D56ULL, "vxlt_0_lowtown_a2_n_h2f_0_0.vlm"},
	{0x6911FEA32B431FCCULL, "vxlt_0_lowtown_docks_b7o_0_0.vlm"},
	{0x691311FB116CE297ULL, "zz_dae_debug6142944.cl"},
	{0x69132C741A27C61CULL, "dae000_bartrand6030679.cl"},
	{0x69139A57D91EFDFBULL, "art\\characters\\playercharacter\\humanfemale\\arm_bda.mmh"},
	{0x6913B857D91F30E5ULL, "art\\characters\\playercharacter\\humanfemale\\arm_bda.msh"},
	{0x6914C1F81D017F1AULL, "vxht_mansion_n_5bo_0_0.vlm"},
	{0x691586648FBB85D3ULL, "gen00fl_aveline6139968.cl"},
	{0x6915A507D917F913ULL, "vxca_cavern1_haunted_egs_0_0.vlm"},
	{0x6915AC0DB9C9FDDDULL, "llac_ht_bartrand_n_3370_1.dds"},
	{0x6916510DEC7277BCULL, "593789142.ogg"},
	{0x691667506FE48219ULL, "vxht_0_hightown_n_577_0_0.vlm"},
	{0x6916BFF3569350BEULL, "327548612.ogg"},
	{0x6917A62B40F0B85AULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\idletospear.gad"},
	{0x6917E18A22C23050ULL, "vxw_sundermount_2b9_0_0.vlm"},
	{0x6917ED086ADFC346ULL, "384218459.ogg"},
	{0x6918EB42F59983F2ULL, "art\\vfx\\materials\\vfx_fire_ring_fb.mao"},
	{0x691911D193DDA887ULL, "vxdr_primevaltomb_75c_0_0.vlm"},
	{0x69193130AE959F8FULL, "vxdr_primevaltomb_7ek_0_0.vlm"},
	{0x69193AC476F6069DULL, "6109005_m.fxe"},
	{0x69197B5B1F518081ULL, "388368441.ogg"},
	{0x69197BB945D107D9ULL, "character_character_p132.psh.fxo.dx11.assembly"},
	{0x6919967FFAAB939EULL, "vxlt_0_lowtown_a3_n_c_emm_0_0.vlm"},
	{0x691A1CB73754AEA6ULL, "6077844_m.fxe"},
	{0x691A32D73690995FULL, "6195868_m.fxe"},
	{0x691A62ADAA011AD0ULL, "hlac_lt_gallowscourtyard_131070_0x.dds"},
	{0x691A81043D2F2DFDULL, "llac_lt_gallowscourtyard_-65535_2.dds"},
	{0x691A81BC78EC0757ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viald.mmh"},
	{0x691A826510C10E78ULL, "llac_lt_hangedman_-1_12.dds"},
	{0x691A85BC78EC0E64ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viald.mao"},
	{0x691A97BC78EC2CC9ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viald.msh"},
	{0x691A9983CC7D2782ULL, "1014198411.ogg"},
	{0x691AA26173CE699DULL, "251173981.ogg"},
	{0x691AF4E4A268C5E6ULL, "335288515.ogg"},
	{0x691B0A71DE1710B0ULL, "vxca_cavern1_haunted_eh9_0_0.vlm"},
	{0x691B1943CF3641EBULL, "hlac_lt_undercityencounter_-1_21x.dds"},
	{0x691B2ED8F8503D60ULL, "171262780.ogg"},
	{0x691B673A12CD400AULL, "follower_banter6154207.cl"},
	{0x691C34AB10D7C254ULL, "vxw_bonepit_1vq_0_0.vlm"},
	{0x691C71630274D541ULL, "6065955_f.fxe"},
	{0x691C73B9CDCBAB96ULL, "gen00fl_isabela6090869.cl"},
	{0x691CA3ECF6185B39ULL, "vxca_cavern1_haunted_i9s_0_0.vlm"},
	{0x691D29C7C4C631AEULL, "vxlt_backalley_g_1of_0_0.vlm"},
	{0x691D63CB8630C37EULL, "vxht_mansion2_1y9_0_0.vlm"},
	{0x691D7FA93DC35F28ULL, "gen00fl_sebastian6180804.cl"},
	{0x691D9D0B4C4302D0ULL, "6079664_m.fxe"},
	{0x691E428D530C46A2ULL, "mag120tr_eau_macha.ncs"},
	{0x691E7E0830C3A6C1ULL, "mag101_grace6034235.cl"},
	{0x691EA92B40F6DC71ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\idletospear.evt"},
	{0x691EB3E4A3DF52A6ULL, "ser191_emeric6104781.cl"},
	{0x691EDC60A0738FF1ULL, "vxca_caverns2prc_9u4_0_0.vlm"},
	{0x691F01E67C571C8EULL, "163463540.ogg"},
	{0x691F14B00D0DA9E1ULL, "6120112_m.fxe"},
	{0x691FAA20A0ABCCCEULL, "845427963.ogg"},
	{0x691FB609E9E80072ULL, "6096222_f.fxe"},
	{0x6920756792836EB9ULL, "llac_ca_caverns2_5802_55.dds"},
	{0x69208B259F1B4AB4ULL, "6085238_m.fxe"},
	{0x69209476AEAE2DA8ULL, "trigger_trap_pressed.cur"},
	{0x6920D7F6F43C0CD6ULL, "34284722.ogg"},
	{0x692204AE1EA07B88ULL, "vxht_0_hightown_c19_0_0.vlm"},
	{0x69221275FD0FA190ULL, "6089803_m.fxe"},
	{0x69225D092D11F8B3ULL, "zz_vault_debug6178932.cl"},
	{0x69225E197F4A1F39ULL, "mag100_karras6034711.cl"},
	{0x6922BCDA8C72D9FCULL, "dre000_dougal6129721.cl"},
	{0x6922CB535760E60BULL, "6160987_m.fxe"},
	{0x6922F4E5E7E48A70ULL, "llac_lt_0_lowtown_a2_n_c_-65537_5.dds"},
	{0x6923A6CA4FE894DFULL, "6033580_m.fxe"},
	{0x6923C09E0D9C3F08ULL, "gen_im_trs_slv_hvy_boo_r0.uti"},
	{0x6924AFE2266D61D7ULL, "6137588_f.fxe"},
	{0x69252E82CCAF9315ULL, "hlac_lt_0_lowtown_a2_n_65533_2y.dds"},
	{0x6925356C7F6CB725ULL, "632229818.ogg"},
	{0x6926669413BFDE7CULL, "6091882_m.fxe"},
	{0x69269C373EDEFF62ULL, "ser221_gascard6148831.cl"},
	{0x69279B0EE6EE6A38ULL, "6173575_m.fxe"},
	{0x6927B68B9450A5CAULL, "vxdr_primevaltomb_91d_0_0.vlm"},
	{0x69287E81DFCB6D64ULL, "gen00fl_aveline6152526.cl"},
	{0x6929018DFAE6BF23ULL, "6041040_m.fxe"},
	{0x69293ED5052751D2ULL, "vxdr_primevaltomb_8tj_0_0.vlm"},
	{0x6929559481E49B08ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mugpewter01_0_sa.dds"},
	{0x692971D2CB6C0B54ULL, "vxca_caverns2_4x6_0_0.vlm"},
	{0x6929AC807D3092B9ULL, "vxw_bonepitdestroyed_a3_1zt_0_0.vlm"},
	{0x692AA166F62AEBC4ULL, "vxca_caverns2_972_0_0.vlm"},
	{0x692AB32FFE1DC308ULL, "vxlt_gallowstemplar_482_0_0.vlm"},
	{0x692AF6858CD9914EULL, "art\\levels\\races\\design\\prp_ballista.mmh"},
	{0x692B02858CD9A5E5ULL, "art\\levels\\races\\design\\prp_ballista.mao"},
	{0x692B0BA55E79EB3BULL, "zz_followers_banter6158307.cl"},
	{0x692B10858CD9BD3CULL, "art\\levels\\races\\design\\prp_ballista.msh"},
	{0x692B19A23FA5141FULL, "hlac_w_sundermount_196608_0z.dds"},
	{0x692BC12B41018F40ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\idletospear.ani"},
	{0x692BF65113C78D73ULL, "179465630.ogg"},
	{0x692D347421D4E98AULL, "hlac_lt_darktown_-1_71z.dds"},
	{0x692DA5B5FF08DB5CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston_1.mao"},
	{0x692DF6AA1724B453ULL, "vxht_chantry_n_355_0_0.vlm"},
	{0x692E045CA622A2FCULL, "gen00fl_fenris6126977.cl"},
	{0x692E498E5B972376ULL, "vxlt_warehouse_5z_0_0.vlm"},
	{0x692E7DAC59E6D4A0ULL, "6070803_m.fxe"},
	{0x692E9F1AC704A796ULL, "572998914.ogg"},
	{0x692EE6EC88A1A61AULL, "vxlt_hangedman_1qo_0_0.vlm"},
	{0x692EFC484556AF20ULL, "fen110_fenris6118029.cl"},
	{0x692EFD38DFB20BFEULL, "6017662_m.fxe"},
	{0x692F2083CEFBFABFULL, "llac_ht_playershouse_n_855_6.dds"},
	{0x692F5173F4F0D3C4ULL, "50036259.ogg"},
	{0x692F60820C832FDCULL, "vxdr_primevalentrance_321_0_0.vlm"},
	{0x69301FB0AF9275E7ULL, "gen00fl_varric6033555.cl"},
	{0x693021649DCE98D8ULL, "hlac_ca_cavern1_haunted_24424_8x.dds"},
	{0x6930768B10969408ULL, "vxht_0_hightown_c46_0_0.vlm"},
	{0x69309A7639D12864ULL, "static_static_p89.psh.fxo.dx11"},
	{0x6930A96702550738ULL, "709344793.ogg"},
	{0x6930B0C7ECECFBE0ULL, "6126840_f.fxe"},
	{0x6930D188C076C32AULL, "659622497.ogg"},
	{0x6930E78BFA98D631ULL, "qun131_taskmaster6039763.cl"},
	{0x6931658D9F880051ULL, "154946657.ogg"},
	{0x69319C53CC3BA3BEULL, "vxlt_undercityserialkiller_1al_0_0.vlm"},
	{0x6931A5B200EDB01AULL, "vxht_mansion_n_5hi_0_0.vlm"},
	{0x6931B5F3CC813D77ULL, "llac_lt_darktown_-1_74.dds"},
	{0x69320B115D22750DULL, "6106383_m.fxe"},
	{0x693235FDE1E38349ULL, "fxa_cen_aoe_dis.mao"},
	{0x69326B9498531F45ULL, "hlac_lt_0_lowtown_docks_131071_0x.dds"},
	{0x6932D70432B373B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\circlemage\\blendtree\\circlemage_override.anb"},
	{0x693308D0EF2E60D0ULL, "742356443.ogg"},
	{0x6933CDF23EA2A8CCULL, "326090107.ogg"},
	{0x6933EEDC4DA2B183ULL, "666875337.ogg"},
	{0x6933F14BB03C125BULL, "338198623.ogg"},
	{0x69349EDE2332C5B6ULL, "rmtst.arl"},
	{0x6934A28F02627825ULL, "394693911.ogg"},
	{0x6934B6009BD2EF48ULL, "vxlt_house_dp_0_0.vlm"},
	{0x6934E11432A1FE5DULL, "t3_eey_gn1.tnt"},
	{0x693586AFFB155AECULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_ash.dds"},
	{0x6936E016A4D8BD0AULL, "lgt110_arl_bann_teagan6115599.cl"},
	{0x693704E50B155B64ULL, "6113869_m.fxe"},
	{0x6937775FD87952D2ULL, "vxht_chantry_n_350_0_0.vlm"},
	{0x6937EC45A56CB72EULL, "vfx_vfx_p3.psh.fxo"},
	{0x69384C0329E7BE56ULL, "vxca_caverns1_mines_dpn_0_0.vlm"},
	{0x69385EC885BE27ACULL, "gen00fl_fenris6116534.cl"},
	{0x693863C0D17D4A5DULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6bluelyrium.mao"},
	{0x693867C0D17D5096ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6bluelyrium.mmh"},
	{0x693871C0D17D6194ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6bluelyrium.msh"},
	{0x6938A012C2B734E1ULL, "hlac_lt_0_lowtown_docks_-131074_3y.dds"},
	{0x69391343E8210444ULL, "vxca_caverns2_983_0_0.vlm"},
	{0x69398398F5A03AFCULL, "143343297.ogg"},
	{0x6939B55053729801ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_fireplace.phy"},
	{0x6939DA23EAD759F9ULL, "vxdr_primevaltomb_7m5_0_0.vlm"},
	{0x6939DADBF68B765DULL, "one100dg_kelder6055436.cl"},
	{0x6939DD6A81928082ULL, "1394821.ogg"},
	{0x6939DE1E2A3BC18AULL, "vxlt_0_lowtown_n_e8a_0_0.vlm"},
	{0x693AF6CAD37C00EEULL, "vxdr_primevaltomb_4xn_0_0.vlm"},
	{0x693B8BFD4331E643ULL, "1005882711.ogg"},
	{0x693BC0DD64680B25ULL, "dae000ip_crowd_ff5_talk16.utp"},
	{0x693C7395946B4905ULL, "vxw_bonepitmining_a2_1vf_0_0.vlm"},
	{0x693C9A3128EA472BULL, "vxlt_undercityencounter_1fh_0_0.vlm"},
	{0x693D0D311E9233DFULL, "vxlt_gallowscourtyard_75o_0_0.vlm"},
	{0x693D37705D4419A1ULL, "263346452.ogg"},
	{0x693D39C29F123067ULL, "6088292_m.fxe"},
	{0x693D425FBF6176FFULL, "vxw_encounter2_camp_ni_0_0.vlm"},
	{0x693D9305A09F654FULL, "zz_dog_debug6204118.cl"},
	{0x693DA1CE102225B1ULL, "dae000_bartrand6027474.cl"},
	{0x693DAD6F782F1325ULL, "vxht_keep_n_6xm_0_0.vlm"},
	{0x693DB6A11D2C621BULL, "vxht_0_hightown_a2_aid_0_0.vlm"},
	{0x693DD78399CEE76EULL, "485364093.ogg"},
	{0x693DEEF8D812866BULL, "6118823_m.fxe"},
	{0x693E8BECC2152B10ULL, "lt_0_lowtown_docks_9_0.rml"},
	{0x693EF58CE79F2D94ULL, "156628260.ogg"},
	{0x693F329E2984F821ULL, "llac_lt_gallowstemplar_n_2390_0.dds"},
	{0x693F5C944371BDBBULL, "vxlt_0_lowtown_n_h23_0_0.vlm"},
	{0x693F97FD1FB527D3ULL, "crt_aoe_cloud.mao"},
	{0x69400E30BE31BC3EULL, "vxlt_undercityserialkiller_1fm_0_0.vlm"},
	{0x6940721C38371E75ULL, "542254579.ogg"},
	{0x69408A8433D4C5DFULL, "6039780_m.fxe"},
	{0x6940C1D21F9755AAULL, "vxht_bartrand_n_6pq_0_0.vlm"},
	{0x6941155D386FED3FULL, "33921337.ogg"},
	{0x6941A53EBDFE2E5EULL, "vxlt_0_lowtown_docks_a3_n_d0z_0_0.vlm"},
	{0x6941B5188FFD345BULL, "6140927_m.fxe"},
	{0x6941C4CC39AAA3F9ULL, "1037716875.ogg"},
	{0x6941E1134EFDC3C5ULL, "fen110_fenris6096416.cl"},
	{0x694383F42B8401DFULL, "vxht_0_hightown_a3_bx5_0_0.vlm"},
	{0x6943ABA35B75A995ULL, "6092197_f.fxe"},
	{0x6943EB4CC164CC64ULL, "hf_bethany_h3s2.mor"},
	{0x6944362B0AA1B651ULL, "54596246.ogg"},
	{0x6944AC83865D54A8ULL, "6153344_m.fxe"},
	{0x6944C36A995833FAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornate.mmh"},
	{0x6944C76A99583A31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornate.mao"},
	{0x6944D56A99585240ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornate.msh"},
	{0x6944FC12082465D4ULL, "hlac_ht_chantry_n_2534_8x.dds"},
	{0x69457AC0E2B5F8CBULL, "37280261.ogg"},
	{0x6945E4F80C2446AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_pose_loop.gad"},
	{0x6946057DC8F314E4ULL, "6083074_m.fxe"},
	{0x69463D9161101B62ULL, "756134927.ogg"},
	{0x69466C89EF541E4AULL, "vxw_encounter2_camp_37_0_0.vlm"},
	{0x6946A1FB4B6BF8CFULL, "dae000_gamlen6026187.cl"},
	{0x6947253EE3AADA2EULL, "6164177_m.fxe"},
	{0x694769563E9F2DA6ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.dura.ani"},
	{0x69479C4C89771542ULL, "vxht_0_hightown_n_c14_0_0.vlm"},
	{0x69481EA2598B49C7ULL, "6155424_m.fxe"},
	{0x694835C0EC4729B6ULL, "764846481.ogg"},
	{0x69489FD4DAF0D0ADULL, "mage_heal.dds"},
	{0x6948CEB3C78CCA09ULL, "art\\levels\\races\\proxy\\props\\_textures\\storage_sa.dds"},
	{0x694948D944DE6EFFULL, "hlac_ca_cavern1_haunted_24422_0y.dds"},
	{0x69494F37D4010858ULL, "dae011_gossips6149742.cl"},
	{0x694A5E2D494F0B36ULL, "vxht_keep_n_3id_0_0.vlm"},
	{0x694A769015C46960ULL, "vxlt_undercityserialkiller_189_0_0.vlm"},
	{0x694A7D0A5F31DDB9ULL, "dae200_bran6092479.cl"},
	{0x694A807609EDF3FFULL, "vxca_cavern1_haunted_k6v_0_0.vlm"},
	{0x694AFF28CF8E6F78ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyo_0s.dds"},
	{0x694B0C0E1AE05FABULL, "vxlt_undercityencounter_1ck_0_0.vlm"},
	{0x694B123733BBBB99ULL, "vxw_encounter1_g9_0_0.vlm"},
	{0x694BAB9F910848E7ULL, "6083392_m.fxe"},
	{0x694BB43CB1578F7FULL, "vxw_encounter2_camp_ih_0_0.vlm"},
	{0x694C106322F76620ULL, "vxw_bonepitdestroyed_a3_19x_0_0.vlm"},
	{0x694C2C27AB7C3DADULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloora.mmh"},
	{0x694C3027AB7C44FEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloora.mao"},
	{0x694C3E27AB7C5C33ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloora.msh"},
	{0x694C4A3984B03FDBULL, "vxht_0_hightown_a3_c5k_0_0.vlm"},
	{0x694C60D5CA089EEBULL, "6117923_m.fxe"},
	{0x694C654AC812B2A4ULL, "6142052_m.fxe"},
	{0x694CF15AC88C3186ULL, "6016904_m.fxe"},
	{0x694D15F50FBB26E9ULL, "vxw_encounter2_p8_0_1.vlm"},
	{0x694DB41762BBC242ULL, "llac_ca_caverns2prc_5802_59.dds"},
	{0x694DE94F4DB7611BULL, "692728379.ogg"},
	{0x694E2E8165329169ULL, "6116210_m.fxe"},
	{0x694EA1AFA86C2697ULL, "hlac_lt_0_lowtown_a2_n_-196609_3x.dds"},
	{0x694F1607B337D6F6ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_2x.dds"},
	{0x694FA9A0157AFC9DULL, "6177237_m.fxe"},
	{0x695008C3C1F5E024ULL, "107212678.ogg"},
	{0x69503580D0F2BCFFULL, "6052992_m.fxe"},
	{0x69507FDD5314A7A6ULL, "227852734.ogg"},
	{0x69513B23090C7C40ULL, "hlac_lt_backalley_g_0_17z.dds"},
	{0x6951BD8AF00D1E9CULL, "6177367_f.fxe"},
	{0x6951DB9E6EABF1A4ULL, "vxca_caverns1_mines_ikh_0_0.vlm"},
	{0x695201C31D82736EULL, "6170262_m.fxe"},
	{0x695229A95BC2431AULL, "vxlt_gallowsprison_n_5qe_0_0.vlm"},
	{0x6952C0DD492C16CCULL, "6077341_m.fxe"},
	{0x6952F8A4766E018FULL, "vxdr_primevaltomb_7qw_0_0.vlm"},
	{0x69536C23F62A37EBULL, "vxw_encounter2_kr_0_0.vlm"},
	{0x695382C2DF33202FULL, "lgt210_shp_kill_qunari26156878.cl"},
	{0x6953D525CDF0BA20ULL, "423718823.ogg"},
	{0x6953FA15B79DAF30ULL, "vxht_chantry_n_2e7_0_0.vlm"},
	{0x6954394B832E82CFULL, "236462932.ogg"},
	{0x695451563EA98F17ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.dura.evt"},
	{0x6954E2EE27435A74ULL, "6069023_m.fxe"},
	{0x695502175312DC39ULL, "vxw_bonepitdestroyed_a3_1ey_0_0.vlm"},
	{0x6955ADA889052BFAULL, "vxlt_gallowstemplar_f_3rv_0_0.vlm"},
	{0x69561E42F18D6025ULL, "gen00fl_isabela6079146.cl"},
	{0x6957037C8EF80945ULL, "12800204.ogg"},
	{0x6957323B13EAD0E7ULL, "art\\levels\\ztools\\default\\generic\\_textures\\setprops_d.dds"},
	{0x695755162E668E07ULL, "138454107.ogg"},
	{0x6957F82B110C9C80ULL, "323016250.ogg"},
	{0x6957FB9C6A23FB2BULL, "6039722_m.fxe"},
	{0x695803DC3D26C289ULL, "6118285_m.fxe"},
	{0x69582C5679D79E1EULL, "qcr000_arishok6087517.cl"},
	{0x695869FEB6BD36EBULL, "mcr370_meredith6083072.cl"},
	{0x6959648D5FB247F8ULL, "lowtown_fem_alien_elf_2_mob6166491.cl"},
	{0x69598CD129F80B4DULL, "qun100_mage_party_comment6019093.cl"},
	{0x6959D1B0FAB497D6ULL, "vxlt_shop_eu_0_0.vlm"},
	{0x6959D5B4CDD2E594ULL, "830035359.ogg"},
	{0x695A195C29897BC7ULL, "6048383_f.fxe"},
	{0x695A6601F693A4E4ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.impa.ani"},
	{0x695A8AB085D276CEULL, "550054276.ogg"},
	{0x695A9306AA830412ULL, "411269762.ogg"},
	{0x695ABB57D95B3830ULL, "art\\characters\\playercharacter\\humanfemale\\arm_bda.phy"},
	{0x695AC095A5D10D4DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphangb.mmh"},
	{0x695AC495A5D1139EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphangb.mao"},
	{0x695AD295A5D12BD3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphangb.msh"},
	{0x695AF1B538FB96ECULL, "mrl000_marethari6069705.cl"},
	{0x695B01F80C36D475ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_pose_loop.ani"},
	{0x695B08E3A7A617CFULL, "859734469.ogg"},
	{0x695B107429FAAF68ULL, "ht_0_hightown_n_water_15346.msh"},
	{0x695B16F1433835D7ULL, "summons.gda"},
	{0x695B227429FACDF9ULL, "ht_0_hightown_n_water_15346.mao"},
	{0x695B2E7429FAE192ULL, "ht_0_hightown_n_water_15346.mmh"},
	{0x695B32620A634076ULL, "486499547.ogg"},
	{0x695B52563EAFAFA0ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.dura.gad"},
	{0x695B835BA55A5104ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault.phy"},
	{0x695BD1F394C226D6ULL, "hlac_lt_0_lowtown_a2_n_-1_0x.dds"},
	{0x695C17C22518DC4FULL, "static3_alpha.psh.fxo"},
	{0x695C560E36493986ULL, "153319075.ogg"},
	{0x695CFC3BC78CF60AULL, "art\\vfx\\textures\\vfx_web_debris.dds"},
	{0x695DB7FADDA97B13ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_2y.dds"},
	{0x695DBFC1405A611BULL, "6093577_m.fxe"},
	{0x695DE4F87A13353FULL, "mag371tr_close_meredith.ncs"},
	{0x695E37CACD482EF9ULL, "246040311.ogg"},
	{0x695E49589B840CE6ULL, "dae000_bartrand_deeproads.stg"},
	{0x695E56FB18B537BBULL, "vxht_keep_6xl_0_0.vlm"},
	{0x695EA3A7502507B7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appa_0s.dds"},
	{0x695EBF21A3FF8710ULL, "965984373.ogg"},
	{0x695F3292EF22088FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\blendtree\\doorsimple.anb"},
	{0x695F7337A1EBB48AULL, "395440274.ogg"},
	{0x696002F74F227C71ULL, "vxlt_0_lowtown_docks_bre_0_0.vlm"},
	{0x69605A84A59AE32FULL, "gen00fl_bethany6102905.cl"},
	{0x6960B498D6139109ULL, "em_cerimon.mor"},
	{0x6961043781363C47ULL, "ser221_party_comments6209872.cl"},
	{0x696151E1C241C139ULL, "vxdr_primevaltomb_8lu_0_0.vlm"},
	{0x69616301F699BE46ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.impa.gad"},
	{0x696190DFF6633DF6ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanr.evt"},
	{0x6961E5194C2EFFCCULL, "llac_ht_bartrand_n_3367_0.dds"},
	{0x696249A8DC2AD7E4ULL, "vxlt_0_lowtown_a2_gnx_0_0.vlm"},
	{0x6962A92E84F17961ULL, "hlac_lt_0_lowtown_docks_n_-196611_1x.dds"},
	{0x6962E891F26A03BEULL, "plt_cod_itm_blade_mercy.plo"},
	{0x696373F44508F4B9ULL, "vxw_bonepitdestroyed_a3_1nx_0_0.vlm"},
	{0x696379D681CEC680ULL, "6205320_m.fxe"},
	{0x69637DBB64DB5ECEULL, "vxdr_primevaltomb_7t9_0_0.vlm"},
	{0x69639868559776BBULL, "vxht_0_hightown_n_bwf_0_0.vlm"},
	{0x6963A2D0B1F32C78ULL, "70402571.ogg"},
	{0x6964300EE451A7D6ULL, "tools_charpos.phy"},
	{0x696446715F20570CULL, "qcr000_varric6065809.cl"},
	{0x69646DFBCAA63CD4ULL, "vxw_bonepit_1oz_0_0.vlm"},
	{0x69647576F5F963DCULL, "vxlt_0_lowtown_eoe_0_0.vlm"},
	{0x6964C1B5E0D1510DULL, "follower_banter6172754.cl"},
	{0x6964EF1CC1E87695ULL, "llac_ht_chantry_n_2534_23.dds"},
	{0x69650FB5750287A5ULL, "vxdr_slavers_2hf_0_1.vlm"},
	{0x69656A5D152CE453ULL, "32826309.ogg"},
	{0x6965715A41728CE7ULL, "470516118.ogg"},
	{0x6966076360D3AD60ULL, "mag321_comtesse_de_launcet6043326.cl"},
	{0x69663DBE0A7326E3ULL, "vxca_caverns1_mines_e64_0_0.vlm"},
	{0x696655205E46013CULL, "6036470_m.fxe"},
	{0x696662F5BC0102C2ULL, "dae000_ineria6136230.cl"},
	{0x6966BA6094C18AF2ULL, "782964132.ogg"},
	{0x6966D8139A80CC00ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha3a.phy"},
	{0x6967151B66D87C18ULL, "521225621.ogg"},
	{0x6967628699AC65AFULL, "711448734.ogg"},
	{0x696779D123B33639ULL, "564749787.ogg"},
	{0x696787DFF6679981ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanr.gad"},
	{0x69679E60D80B6765ULL, "vxlt_backalley_n_1y2_0_0.vlm"},
	{0x6967C386AA6C8C49ULL, "static_static_p105.psh.fxo.dx11"},
	{0x6967E358AB18F79DULL, "6049709_m.fxe"},
	{0x6967EC7FD5EC1842ULL, "hlac_ht_mansion_n_3381_0y.dds"},
	{0x696831BFACCF7E7BULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_2.pwk"},
	{0x69685E01F69FD425ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.impa.evt"},
	{0x6968AE60507A081AULL, "vxlt_hangedman_1mk_0_0.vlm"},
	{0x6968DD6707109C81ULL, "729097676.ogg"},
	{0x696901E9763448FAULL, "6163990_m.fxe"},
	{0x69692ED8D5C8B403ULL, "635661608.ogg"},
	{0x6969E26A99778405ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornate.phy"},
	{0x696A9AE7E9E7E071ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeright.mmh"},
	{0x696AA6E7E9E7F412ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeright.mao"},
	{0x696AB4E7E9E80C2FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_edgeright.msh"},
	{0x696AFFA43D9693D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c.ani"},
	{0x696B8CECC23C1B78ULL, "lt_0_lowtown_docks_9_0.anb"},
	{0x696C4D3A290FA281ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforwardright.ani"},
	{0x696C63496840AD39ULL, "6189913_m.fxe"},
	{0x696CC0546E6ED873ULL, "zz_approval_debug6003945.cl"},
	{0x696CFB769B0D2937ULL, "vfx_vfx_p8.psh.fxo.dx11"},
	{0x696D3D0C4A66D8FAULL, "vxlt_0_lowtown_docks_n_b58_0_0.vlm"},
	{0x696D75840CA2CC3EULL, "570224011.ogg"},
	{0x696D7610BB3F2311ULL, "vxca_caverns2_7nv_0_0.vlm"},
	{0x696E125B705A6821ULL, "cd_qun131_taskmaster.bnk"},
	{0x696E40C1400DA379ULL, "6128344_f.fxe"},
	{0x696E6904975D70B3ULL, "tbe300im_clean_chest.uti"},
	{0x696E88DFF66DBA2FULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanr.ani"},
	{0x696ED98B01664B5FULL, "vxht_0_hightown_a3_bi6_0_0.vlm"},
	{0x696EE6780E54130FULL, "lt_0_lowtown_docks_a3_n_97_0.mao"},
	{0x696EF2780E5427ACULL, "lt_0_lowtown_docks_a3_n_97_0.mmh"},
	{0x696EF8780E54319EULL, "lt_0_lowtown_docks_a3_n_97_0.msh"},
	{0x696F190EE45B8877ULL, "tools_charpos.msh"},
	{0x696F24432CDE41D2ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d06_0t.dds"},
	{0x696F2F0EE45BADE9ULL, "tools_charpos.mmh"},
	{0x696F98C863933C98ULL, "hlac_ht_chantry_2534_30x.dds"},
	{0x697047026D635EFAULL, "828395159.ogg"},
	{0x69704FDDDA1668EBULL, "vxw_encounter2_qt_0_0.vlm"},
	{0x6970A9F9A8421604ULL, "one300dg_vanard.stg"},
	{0x6970E20DCB10E366ULL, "art\\levels\\ztools\\shaders\\blank_n.dds"},
	{0x69712D3558BD4BF1ULL, "vxlt_gallowscourtyard_7ea_0_0.vlm"},
	{0x697139C3B0DDCC58ULL, "hightown_male_noble_gossip6171312.cl"},
	{0x69714C7041F37EDCULL, "973013652.ogg"},
	{0x6971FCA43D9CAD36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c.gad"},
	{0x697229E9323E1210ULL, "484897228.ogg"},
	{0x69723F7BA855838EULL, "hlac_lt_undercitydungeon_-1_11y.dds"},
	{0x6972583A29142123ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforwardright.gad"},
	{0x6972EBF86C20E395ULL, "gen00fl_fenris6126772.cl"},
	{0x6972EFC153869464ULL, "qun210cr_guard6072117.cl"},
	{0x6972F4D23191C483ULL, "479929852.ogg"},
	{0x6973A197B2AFBBF9ULL, "vxdr_primevaltomb_7y9_0_0.vlm"},
	{0x69745CD7CA9B5524ULL, "dae000_male_templar_mob6174366.cl"},
	{0x69747D51FCB7C27AULL, "261773196.ogg"},
	{0x69748A790AEF5FEBULL, "148567569.ogg"},
	{0x6974A8795F78FBBEULL, "lt_0_lowtown_a2_n_c_36_0_ll.mmh"},
	{0x6974E7962112F389ULL, "6112085_m.fxe"},
	{0x697506F45F98A65BULL, "894842644.ogg"},
	{0x69754CED8BA5C066ULL, "6167309_m.fxe"},
	{0x69759517E35C8F19ULL, "hlac_dr_slavers_3215_8z.dds"},
	{0x6975B3C41F3154B6ULL, "vxht_keep_n_3xg_0_0.vlm"},
	{0x69771BBF72D69F1CULL, "dae000_bodhan_letter6144263.cl"},
	{0x6977771CEF141F56ULL, "gen_im_arm_cht_hvy_act3_02.uti"},
	{0x6977D46A29B63454ULL, "vxht_0_hightown_a2_ca1_0_0.vlm"},
	{0x697817A43DA14695ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c.evt"},
	{0x697872951C8932BBULL, "6151266_m.fxe"},
	{0x6978E6F0743C396BULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.dura.ani"},
	{0x69791019B0DC1404ULL, "287896150.ogg"},
	{0x697A35F11BF4D85CULL, "one100_party_comments6133453.cl"},
	{0x697A6721F3619142ULL, "6127430_f.fxe"},
	{0x697B17D2DA1DBED6ULL, "hlac_lt_0_lowtown_a2_n_131072_0z.dds"},
	{0x697B3D66A97F6824ULL, "dae000_bartrand6027532.cl"},
	{0x697BCF8EE22E67A9ULL, "hlac_w_blightlands_-4_0y.dds"},
	{0x697BFAACE112FBC4ULL, "hlac_ht_0_hightown_a3_-196608_0y.dds"},
	{0x697C57AFF23AC0C2ULL, "hlac_lt_0_lowtown_a3_n_c_131071_7y.dds"},
	{0x697C8B835980ECD6ULL, "6124422_f.fxe"},
	{0x697D3A88DE2F130FULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.cess.gad"},
	{0x697D4B67F35C63DFULL, "vxht_0_hightown_a3_bd1_0_0.vlm"},
	{0x697D939CCB2E8B40ULL, "hlac_lt_0_lowtown_a2_65538_1y.dds"},
	{0x697DFBD74C287593ULL, "768325569.ogg"},
	{0x697EA0CDD33EDA22ULL, "6121097_m.fxe"},
	{0x697EFDB78A67FC3CULL, "vxw_encounter1_bc_0_0.vlm"},
	{0x697F52A84D02C666ULL, "hlac_ht_0_hightown_n_65538_0y.dds"},
	{0x697FA10019AC2DC3ULL, "mcr310_argument6074453.cl"},
	{0x697FDD4C1C9D9297ULL, "zz_followers_banter6157629.cl"},
	{0x6980493B958651C1ULL, "245103246.ogg"},
	{0x69808A4327A905A0ULL, "6188578_f.fxe"},
	{0x6980EA9A88DDD71DULL, "atl_logininfos_dxt5_02.dds"},
	{0x6980F9437186D09BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gibbet_n.dds"},
	{0x69815501444D89EEULL, "m_dae151ar_mount_cave.wbd"},
	{0x6981702F5C9FC72CULL, "630759079.ogg"},
	{0x6981BCD89BEE038DULL, "llac_w_sundermount_-196610_0.dds"},
	{0x6981D37CD924B3BEULL, "gen00fl_varric6109204.cl"},
	{0x69821374A4A5D479ULL, "vxdr_primevaltomb_7r8_0_0.vlm"},
	{0x6982C15DC42FFBEDULL, "vxlt_0_lowtown_a2_e23_0_0.vlm"},
	{0x69831FC20B3E3981ULL, "6059270_m.fxe"},
	{0x69834F88DE33A2BCULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.cess.evt"},
	{0x6983B09BCB869B55ULL, "gen00fl_isabela6127627.cl"},
	{0x6983F63AAD0A5FFBULL, "vxlt_0_lowtown_docks_n_cvm_0_0.vlm"},
	{0x6984034C22C2A383ULL, "vxht_0_hightown_by3_0_0.vlm"},
	{0x698415FB4793C8B4ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hpath2_da.dds"},
	{0x698426CA0D5E55E0ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ichm.phy"},
	{0x69843560F91E8492ULL, "647815683.ogg"},
	{0x698446354C83E0EDULL, "339435684.ogg"},
	{0x6984AF249910F620ULL, "6103918_m.fxe"},
	{0x698507B552122EE4ULL, "gen00fl_aveline6101827.cl"},
	{0x698516294EA10E10ULL, "296087510.ogg"},
	{0x6985CEF074469AE2ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.dura.evt"},
	{0x6985E4003CABAFB1ULL, "w_encounter2_camp.arl"},
	{0x69861E615CBAD728ULL, "ht_0_hightown_35_0.anb"},
	{0x69866A00A87C4C4CULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statuesworda.msh"},
	{0x69867000A87C563EULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statuesworda.mmh"},
	{0x69867C00A87C6AD5ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statuesworda.mao"},
	{0x69869D8640D576A5ULL, "vxlt_0_lowtown_docks_cs6_0_1.vlm"},
	{0x6986E56621DB64B0ULL, "vxlt_hangedman_n_16g_0_0.vlm"},
	{0x698720888B9725A5ULL, "vxca_caverns2_3nv_0_0.vlm"},
	{0x69872155BE8C0797ULL, "311161350.ogg"},
	{0x69873379C691A699ULL, "45830791.ogg"},
	{0x6987E7598DED9004ULL, "vxlt_gallowstemplar_f_6s3_0_1.vlm"},
	{0x69884EDF84DDA69AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed_b.anb"},
	{0x69888BC85070C098ULL, "vxlt_darktown_5b1_0_0.vlm"},
	{0x6988C23E1ABC61E6ULL, "6124744_m.fxe"},
	{0x6988FFE8CA59434EULL, "vxca_cave_23d_0_0.vlm"},
	{0x698920FA968C4D98ULL, "53651727.ogg"},
	{0x6989A95EAE2B91A5ULL, "qun131tr_warrens_thug.utt"},
	{0x6989C2D35CE54FC5ULL, "vxlt_0_lowtown_docks_n_b84_0_0.vlm"},
	{0x698A086DA3B06CBFULL, "183679999.ogg"},
	{0x698A44B5FC8B59D6ULL, "vxlt_gallowscourtyard_78q_0_0.vlm"},
	{0x698A6A5FC9D4418FULL, "6105836_m.fxe"},
	{0x698AC210C47F4DDFULL, "hlac_ht_0_hightown_a3_65535_2y.dds"},
	{0x698AF0F83A0929F8ULL, "hlac_ht_0_hightown_196607_0x.dds"},
	{0x698B35E0B460BA11ULL, "6079356_f.fxe"},
	{0x698B588E2504C5F1ULL, "vxlt_0_lowtown_docks_bcz_0_0.vlm"},
	{0x698BBD44E5527C5FULL, "vxht_0_hightown_a3_bc0_0_0.vlm"},
	{0x698BEDEC588BF695ULL, "mag110_vincento6031816.cl"},
	{0x698BF5139AA01995ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha3a.msh"},
	{0x698BF7139AA01D0BULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_ha3a.mmh"},
	{0x698C7AA2BA7270B1ULL, "mag311_emile_de_launcet6043787.cl"},
	{0x698C9E71C4EFC2AEULL, "vxw_bonepitmining_a2_21k_0_0.vlm"},
	{0x698CBA8E8B1C96F5ULL, "vxw_sundermount_211_0_0.vlm"},
	{0x698CCDF0744CB7F5ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.dura.gad"},
	{0x698D048116B88F6AULL, "vxht_0_hightown_b7p_0_0.vlm"},
	{0x698E4D3B622D3A5BULL, "796565331.ogg"},
	{0x698E8A41D3805850ULL, "ser221_party_comments6163605.cl"},
	{0x698E92379E587D0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_on_ground_loop.ani"},
	{0x698EC6973604F701ULL, "vxca_cave_1c0_0_0.vlm"},
	{0x698ED04D3521080AULL, "w_woundedcoast_82_0.mao"},
	{0x698ED44D35210F59ULL, "w_woundedcoast_82_0.mmh"},
	{0x698EDE4D35212047ULL, "w_woundedcoast_82_0.msh"},
	{0x698F000A16A67F7AULL, "vxca_caverns2_96o_0_0.vlm"},
	{0x698F7163E7AAD2C2ULL, "ave100pt_promotion.ncs"},
	{0x69900A04163C559DULL, "888656439.ogg"},
	{0x699013BEFE132F63ULL, "vxlt_hangedman_1vg_0_0.vlm"},
	{0x69902DF0A80A353CULL, "character_character_p54.psh.fxo"},
	{0x699064ACE7CB3B39ULL, "vxlt_0_lowtown_docks_b5h_0_0.vlm"},
	{0x6990BB20E740A7CCULL, "6166111_m.fxe"},
	{0x6991132C6D5083ECULL, "gen00fl_aveline6139275.cl"},
	{0x69915788DE3FEDD5ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.cess.ani"},
	{0x69916BC7E60CEF9CULL, "86252643.ogg"},
	{0x6991B1F916ABB67DULL, "w_blightlands_111_0.mao"},
	{0x6991B5F916ABBCB6ULL, "w_blightlands_111_0.mmh"},
	{0x6991BFF916ABCDB4ULL, "w_blightlands_111_0.msh"},
	{0x6991C952CE44754FULL, "vxca_caverns2prc_ae2_0_0.vlm"},
	{0x69927154BE9C0701ULL, "one100dg_kelder6055601.cl"},
	{0x6992D2B99E7D6B9EULL, "gen00fl_aveline6137653.cl"},
	{0x699321018B070EA0ULL, "6104818_m.fxe"},
	{0x6993844893501E78ULL, "t3_cth_cm1.tnt"},
	{0x6993AB3B2919156DULL, "vxca_caverns1_ikh_0_0.vlm"},
	{0x6993C74BE1C14C19ULL, "zz_approval_debug6003958.cl"},
	{0x699429D3244944E2ULL, "vxht_keep_6g8_0_0.vlm"},
	{0x69949A7B14045556ULL, "fx_ray_line.dds"},
	{0x6994A289A8B1DAFAULL, "592185820.ogg"},
	{0x6994C9273A62C203ULL, "mist_wh.mao"},
	{0x6994CC6F3882B62FULL, "llac_ht_keep_3732_14.dds"},
	{0x6994F2A6EA065EFDULL, "mcr371_hostile_fenris6082229.cl"},
	{0x6995822C9EC1EB7CULL, "6180018_m.fxe"},
	{0x69964CA3674FBB77ULL, "vxlt_gallowstemplar_n_43s_0_0.vlm"},
	{0x6996AB8E82A83D61ULL, "290836342.ogg"},
	{0x6996C473E60F3854ULL, "6103049_f.fxe"},
	{0x6996E3A029A47618ULL, "340099232.ogg"},
	{0x6996EECB8B51BE24ULL, "gen00fl_merrill6147015.cl"},
	{0x69975301C5BA9A1DULL, "hlac_dr_primevaltomb_9621_0y.dds"},
	{0x69975F858D8D2ACFULL, "dae000_gen_player_turn_in6177843.cl"},
	{0x699834B04EDB6845ULL, "vxlt_0_lowtown_docks_n_b53_0_0.vlm"},
	{0x6998EBDEC7456646ULL, "gen00fl_aveline6122149.cl"},
	{0x69995246CE9B8E1BULL, "969365171.ogg"},
	{0x6999CA6B16FADE71ULL, "vxlt_0_lowtown_docks_bfy_0_0.vlm"},
	{0x6999E8117D54FE72ULL, "follower_banter6171857.cl"},
	{0x699A0BECB3CF58B6ULL, "lgt351ip_disarm_door.ncs"},
	{0x699A63F85C46D9DAULL, "vxca_cavern1_haunted_ghq_0_0.vlm"},
	{0x699A7D92180191FFULL, "mcr310_argument6074488.cl"},
	{0x699A7DB97B7550C3ULL, "vxw_encounter1_ja_0_1.vlm"},
	{0x699A9A2CAE771843ULL, "art\\characters\\playercharacter\\humanmale\\glv_medj.mao"},
	{0x699A9E2CAE771E90ULL, "art\\characters\\playercharacter\\humanmale\\glv_medj.mmh"},
	{0x699A9F9E313F139AULL, "vxw_encounter2_camp_4r_0_0.vlm"},
	{0x699AAC2CAE7736AAULL, "art\\characters\\playercharacter\\humanmale\\glv_medj.msh"},
	{0x699B19558A1A2339ULL, "vxdr_primevaltomb_8xy_0_0.vlm"},
	{0x699B82391B13F42CULL, "gen00fl_carver6092287.cl"},
	{0x699BC16D01E71A86ULL, "hlac_w_sundermount_-262148_0y.dds"},
	{0x699BFEE89FF8DCCCULL, "6112893_m.fxe"},
	{0x699C081000765C1FULL, "564372228.ogg"},
	{0x699C57A6E9BC7717ULL, "6087760_f.fxe"},
	{0x699CBA5053C71E3DULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_fireplace.mao"},
	{0x699CBE5053C72576ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_fireplace.mmh"},
	{0x699CC85053C73674ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_fireplace.msh"},
	{0x699D29F84C50B6FFULL, "6071234_m.fxe"},
	{0x699DD5CE5ACE6209ULL, "vxlt_0_lowtown_n_cmw_0_0.vlm"},
	{0x699DDA91B61D4FF4ULL, "art\\vfx\\combat\\vfx_impact_c.impa.ani"},
	{0x699DFC927872BBF3ULL, "vxht_playershouse_n_1ti_0_0.vlm"},
	{0x699E1B8A4D531B3DULL, "vxlt_0_lowtown_n_gky_0_0.vlm"},
	{0x699E3BA6F2DD6E80ULL, "937032684.ogg"},
	{0x699E567D5812A9DAULL, "98839968.ogg"},
	{0x699E69C1AF4E7E06ULL, "377511716.ogg"},
	{0x699EB14E3D1DC4B1ULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_open.msh"},
	{0x699EBF4E3D1DDCCCULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_open.mao"},
	{0x699ECB4E3D1DF16FULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_open.mmh"},
	{0x699F3B3F9B141273ULL, "vxlt_0_lowtown_a2_n_c_ev4_0_0.vlm"},
	{0x699F48CE7B316698ULL, "417949408.ogg"},
	{0x699FA517A81C2CEDULL, "vxlt_0_lowtown_a2_e81_0_0.vlm"},
	{0x699FAD879493464EULL, "gen00fl_fenris6105550.cl"},
	{0x69A00B29D5B3B274ULL, "vxw_sundermount_2gb_0_0.vlm"},
	{0x69A036BDA88C0C48ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_8xholegrill.msh"},
	{0x69A0387D79AC788DULL, "612195317.ogg"},
	{0x69A048BDA88C2AD9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_8xholegrill.mao"},
	{0x69A054BDA88C3F72ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_8xholegrill.mmh"},
	{0x69A06FA48A1A753AULL, "zz_vault_debug6203213.cl"},
	{0x69A1C83C39990AA7ULL, "241252045.ogg"},
	{0x69A2188F9B2944F3ULL, "lgt220_shp_mag_guard6149512.cl"},
	{0x69A21D181B0F2DEDULL, "vxca_caverns1_ing_0_0.vlm"},
	{0x69A26B379E68E448ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_on_ground_loop.gad"},
	{0x69A287A3114B1839ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_3x.dds"},
	{0x69A3921118E9491EULL, "hlac_w_bonepitmining_a2_131077_0y.dds"},
	{0x69A3A1F9A86D620FULL, "one300dg_vanard.bnk"},
	{0x69A3B831A355BF4BULL, "gen00fl_fenris6108064.cl"},
	{0x69A3D791B621B656ULL, "art\\vfx\\combat\\vfx_impact_c.impa.gad"},
	{0x69A4C669EF1A217DULL, "676984964.ogg"},
	{0x69A5375CD6EE7149ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.cess.ani"},
	{0x69A541FECEB879C8ULL, "55641283.ogg"},
	{0x69A562DEFEABE169ULL, "6134247_m.fxe"},
	{0x69A5A815C0986C3FULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupb.msh"},
	{0x69A5AE15C09876E1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupb.mmh"},
	{0x69A5BA15C0988A82ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupb.mao"},
	{0x69A5C92DD8F338E0ULL, "area_transition.cur"},
	{0x69A5D0FA7F378CA0ULL, "vxlt_0_lowtown_a2_by8_0_0.vlm"},
	{0x69A5F50655FC4A3AULL, "vxlt_0_lowtown_docks_btt_0_0.vlm"},
	{0x69A61EC2AEC371C0ULL, "hlac_lt_gallowscourtyard_-65537_3x.dds"},
	{0x69A67888B38645B7ULL, "fen310_fenris6118699.cl"},
	{0x69A6A00E374AB30DULL, "787714791.ogg"},
	{0x69A6A28D0679579FULL, "follower_barks6153545.cl"},
	{0x69A6A37B654379E9ULL, "sst_ss_wryme.gda"},
	{0x69A6E3539DFDC79FULL, "and110_refugee6214656.cl"},
	{0x69A70F13B340AD20ULL, "vxlt_gallowscourtyard_7cj_0_0.vlm"},
	{0x69A75245CD3662EBULL, "vxht_brothel_n_6ge_0_0.vlm"},
	{0x69A767A3F3921735ULL, "115142890.ogg"},
	{0x69A7C3F9A8718577ULL, "one300dg_vanard.cnv"},
	{0x69A8107CC5970BC2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2right.msh"},
	{0x69A8127CC5970F78ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2right.mmh"},
	{0x69A81E7CC597231BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2right.mao"},
	{0x69A871BDF64E2571ULL, "6134830_m.fxe"},
	{0x69A8F4481B147F16ULL, "vxlt_0_lowtown_docks_a3_n_cxu_0_0.vlm"},
	{0x69A9247EF97C807DULL, "vxlt_foundry_n_yr_0_0.vlm"},
	{0x69A9C5E13FE1BD4BULL, "267423753.ogg"},
	{0x69AA43CA0D7F5675ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ichm.msh"},
	{0x69AA45CA0D7F59EBULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_ichm.mmh"},
	{0x69AAF291B62802B5ULL, "art\\vfx\\combat\\vfx_impact_c.impa.evt"},
	{0x69AB525CD6F30ACBULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.cess.gad"},
	{0x69AB776F10F1C14CULL, "vxw_woundedcoast_gt_0_0.vlm"},
	{0x69ABCCE9DFECE07BULL, "character_character_p220.psh.fxo.dx11"},
	{0x69ABCDAEFD390B2AULL, "hla_ht_0_hightown_n_-196608_0z.dds"},
	{0x69AC1320A813332AULL, "character_character_p122.psh.fxo.dx11.assembly"},
	{0x69AC4C0463510F84ULL, "hlac_lt_0_lowtown_n_-1_0x.dds"},
	{0x69AC985D64E66F82ULL, "555186383.ogg"},
	{0x69ACA17B1418F284ULL, "fx_ray_line.mao"},
	{0x69ACF778E1FF6063ULL, "vxlt_hangedman_1xi_0_0.vlm"},
	{0x69AD21F948588C9BULL, "984730264.ogg"},
	{0x69AD27A3A50C996CULL, "hlac_w_sundermount_262139_0z.dds"},
	{0x69AD4B014472B31CULL, "m_dae151ar_mount_cave.bnk"},
	{0x69ADC6F122DF438EULL, "ave100_donnic6057739.cl"},
	{0x69AE4354BAD15417ULL, "6069565_m.fxe"},
	{0x69AE7D06C7A9CAF4ULL, "vxw_sundermount_2hc_0_0.vlm"},
	{0x69AF0442A0B1D8F1ULL, "vxht_keep_n_4ne_0_0.vlm"},
	{0x69AF543445B159D4ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.dura.ani"},
	{0x69AFA4578D5A757EULL, "lowtown_guard_chase6169506.cl"},
	{0x69AFF5EEF27ACA15ULL, "vxlt_0_lowtown_docks_a3_brn_0_0.vlm"},
	{0x69AFFAEFF67C636FULL, "404872187.ogg"},
	{0x69B0679DADB89489ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext_6.pwk"},
	{0x69B08A4848245E83ULL, "6172553_m.fxe"},
	{0x69B08B71118A1A80ULL, "llac_lt_0_lowtown_a2_-131072_4.dds"},
	{0x69B09F995CA34CFEULL, "pro000_merchant6051611.cl"},
	{0x69B128DEF60DE89AULL, "37609858.ogg"},
	{0x69B151DF8500B14DULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed_b.mmh"},
	{0x69B19B2A8E492099ULL, "564250140.ogg"},
	{0x69B1BF5C5037FC87ULL, "6144336_m.fxe"},
	{0x69B1C975CA17730AULL, "hlac_ca_caverns2_5802_49z.dds"},
	{0x69B1DCA5E8106393ULL, "646127595.ogg"},
	{0x69B1E5003CD0EC17ULL, "w_encounter2_camp.rml"},
	{0x69B1FA895AFBBA61ULL, "6153590_m.fxe"},
	{0x69B1FCEC5C54B40DULL, "llac_w_bonepit_458756_0.dds"},
	{0x69B2197CC59F6F03ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2right.phy"},
	{0x69B24F5CD6F92430ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.cess.evt"},
	{0x69B2A362A25F3146ULL, "ch1alphemnotnt_vl.psh.fxo"},
	{0x69B2CAC7998B293BULL, "58439147.ogg"},
	{0x69B2CB544401EBA2ULL, "dae000_bodhan6129144.cl"},
	{0x69B30319641DF7FEULL, "6171863_m.fxe"},
	{0x69B316E9A9E0153FULL, "one100dg_elren6064004.cl"},
	{0x69B3CEE40BDF590BULL, "203454843.ogg"},
	{0x69B3E3BDE38DDEE8ULL, "vxlt_0_lowtown_docks_a3_cus_0_0.vlm"},
	{0x69B3FA21ED450A37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickexit.gad"},
	{0x69B4A4F0A4719BB5ULL, "mrl250cr_reptilerock.utc"},
	{0x69B4BE3F43EFD2ACULL, "358713355.ogg"},
	{0x69B4DDDCEC88602BULL, "6118214_m.fxe"},
	{0x69B513032E071271ULL, "vxw_bonepitdestroyed_a3_1zl_0_0.vlm"},
	{0x69B526ABB5DFC6CCULL, "6087522_m.fxe"},
	{0x69B542CB4701FD1AULL, "mcr371_emile_de_launcet6086948.cl"},
	{0x69B580F0A536C5A0ULL, "vxlt_gallowscourtyard_7dk_0_0.vlm"},
	{0x69B5D301600C3BC1ULL, "vxht_0_hightown_a3_a6r_0_0.vlm"},
	{0x69B61F2487ADC164ULL, "6043392_f.fxe"},
	{0x69B6205B709835D3ULL, "cd_qun131_taskmaster.wbd"},
	{0x69B6513445B77336ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.dura.gad"},
	{0x69B6B46BBADC1D6CULL, "hlac_lt_0_lowtown_a3_n_c_-65537_6z.dds"},
	{0x69B6B8DA73927C6CULL, "vxht_mansion2b_218_0_0.vlm"},
	{0x69B6F3910BDCCF2DULL, "675535318.ogg"},
	{0x69B75897CDD103D7ULL, "framebuffereffects_diffusiondof_p7.psh.fxo.dx11"},
	{0x69B75C34385B14BAULL, "204572727.ogg"},
	{0x69B78412FC8436E0ULL, "6206906_f.fxe"},
	{0x69B7B252082C9527ULL, "vxht_mansion_n_4xq_0_0.vlm"},
	{0x69B7E5C69D547A20ULL, "llac_lt_darktown_-1_86.dds"},
	{0x69B80A2AC583CAC2ULL, "hlac_ht_0_hightown_65534_0x.dds"},
	{0x69B8CF73B7350454ULL, "llac_ca_caverns1_mines_24424_19.dds"},
	{0x69B8EB5856867F97ULL, "mag291_demon_marethari6028947.cl"},
	{0x69B908C184E4601EULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_alchemic_0.pwk"},
	{0x69B9F7AE1B1B0EC5ULL, "vxca_caverns2_96d_0_0.vlm"},
	{0x69BA0F21ED4999C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickexit.evt"},
	{0x69BA1E592298A529ULL, "vxdr_primevaltomb_57q_0_0.vlm"},
	{0x69BA507E9A255BDFULL, "6112072_m.fxe"},
	{0x69BA692657D6CA11ULL, "504252005.ogg"},
	{0x69BB2BB4AC4BB323ULL, "6072830_f.fxe"},
	{0x69BB4351B6101CF3ULL, "combat_st_range.gda"},
	{0x69BB528F7F14F727ULL, "vxlt_gallowsprison_n_6ms_0_0.vlm"},
	{0x69BB6955D3F578E3ULL, "vxlt_hangedman_1gh_0_0.vlm"},
	{0x69BBC1C4AA6DBF9EULL, "vxca_cavern1_haunted_j4t_0_0.vlm"},
	{0x69BBDCE27E9E4CB5ULL, "lt_0_lowtown_docks_221_0.msh"},
	{0x69BBDEE27E9E502BULL, "lt_0_lowtown_docks_221_0.mmh"},
	{0x69BBEAE27E9E64C8ULL, "lt_0_lowtown_docks_221_0.mao"},
	{0x69BC1E7C99014E52ULL, "5769353.ogg"},
	{0x69BC46200967F22DULL, "hlac_lt_0_lowtown_docks_a3_-131071_0x.dds"},
	{0x69BC6C3445BC0C95ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.dura.evt"},
	{0x69BC818706ADA16AULL, "52774696.ogg"},
	{0x69BCFD5D5197C5A5ULL, "one300_nuncio6073784.cl"},
	{0x69BDA311C143AAF2ULL, "narrative_varric_demo6170831.cl"},
	{0x69BE0729FEC88CF0ULL, "6076724_m.fxe"},
	{0x69BE0B435FE9D89EULL, "6206984_f.fxe"},
	{0x69BE32E7DA6FCB3AULL, "vxca_caverns1_hul_0_0.vlm"},
	{0x69BE91F933ED6DD3ULL, "463140838.ogg"},
	{0x69BED186B3D411CCULL, "661199109.ogg"},
	{0x69BF16339A62F217ULL, "hlac_lt_gallowstemplar_f_2383_5y.dds"},
	{0x69BF16ECB3EE87C3ULL, "lgt351ip_disarm_door.utp"},
	{0x69BF29FFBFF7E00BULL, "vxlt_hangedman_n_1fx_0_0.vlm"},
	{0x69C0259978475C5CULL, "vxht_mansion2b_3gf_0_1.vlm"},
	{0x69C054552CC4B09BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_5.pwk"},
	{0x69C0CA76D2BA35E7ULL, "461118392.ogg"},
	{0x69C0DAF8CC0BC9A3ULL, "6031119_m.fxe"},
	{0x69C0EB7CC27D1B3AULL, "mag100_cullen6140977.cl"},
	{0x69C0FF46F755326AULL, "l_env_wal_ltdocks_wal_calm_p2p_act3.wbd"},
	{0x69C11C7E73A82C18ULL, "hlac_lt_0_lowtown_docks_65533_2x.dds"},
	{0x69C121236D4613B4ULL, "sto_cl_li.mao"},
	{0x69C2559AD583F768ULL, "vxlt_0_lowtown_docks_a3_cxp_0_0.vlm"},
	{0x69C2985E212C6631ULL, "vxlt_0_lowtown_a3_n_c_bxz_0_0.vlm"},
	{0x69C2AB8E2CCCBBE8ULL, "vxw_bonepit_143_0_0.vlm"},
	{0x69C2DA7FE810F0A7ULL, "6155907_m.fxe"},
	{0x69C2DC58F1BD7B50ULL, "mag200_feynriel6030447.cl"},
	{0x69C3332B40378F3BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp04.mmh"},
	{0x69C3512B4037C225ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp04.msh"},
	{0x69C359DC407FDFF7ULL, "6095580_m.fxe"},
	{0x69C3652F6508C006ULL, "vxht_bartrand_n_5ya_0_0.vlm"},
	{0x69C370BA3AA8014EULL, "6004333_f.fxe"},
	{0x69C3B47F17FFBF43ULL, "gencr_dalish_warrior_ws_f.utc"},
	{0x69C3D39C1A17C743ULL, "241520162.ogg"},
	{0x69C40BC13D847DE3ULL, "hm_alistair_drunk.mor"},
	{0x69C40E7FBE32C1C6ULL, "gen_im_trs_dwv_junk4_02.uti"},
	{0x69C464FB2C1123E7ULL, "6047459_m.fxe"},
	{0x69C46CD680B2CAC2ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.mmh"},
	{0x69C4BE52ACD2C4C8ULL, "hlac_lt_0_lowtown_docks_65535_2z.dds"},
	{0x69C544241B5B396DULL, "vxht_0_hightown_n_be6_0_0.vlm"},
	{0x69C57D9E4DDD9B82ULL, "293950103.ogg"},
	{0x69C5A498C6E836C3ULL, "hla_ht_0_hightown_a2_131073_0x.dds"},
	{0x69C6F0FB279EEDD1ULL, "vxlt_0_lowtown_docks_a3_d2t_0_0.vlm"},
	{0x69C71D615CF26CE0ULL, "ht_0_hightown_35_0.rml"},
	{0x69C72501B16345C1ULL, "vxlt_0_lowtown_a3_n_c_eoq_0_1.vlm"},
	{0x69C74700A8B3A7D9ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statuesworda.phy"},
	{0x69C82721ED55FFBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\quickexit.ani"},
	{0x69C84E0F01CCB6DBULL, "dre000_vampire6034286.cl"},
	{0x69C85B28F4DDF24CULL, "vxw_woundedcoast_yj_0_0.vlm"},
	{0x69C892F364A32419ULL, "vxlt_0_lowtown_f1u_0_0.vlm"},
	{0x69C9B9E9B5489FD5ULL, "93814921.ogg"},
	{0x69CA207FEBF270F5ULL, "6152723_f.fxe"},
	{0x69CA88DB29484D86ULL, "271198280.ogg"},
	{0x69CA9599188855DFULL, "vxca_cavern1_haunted_ioc_0_0.vlm"},
	{0x69CACF595DCDFE14ULL, "6129368_f.fxe"},
	{0x69CAD2B4DEE834C3ULL, "hlac_w_woundedcoast_196608_0x.dds"},
	{0x69CB407B33FC129BULL, "art\\vfx\\combat\\vfx_w_gn1_battle_activ_c.impa.ani"},
	{0x69CB5E94DBDB093DULL, "llac_lt_0_lowtown_-65536_4.dds"},
	{0x69CBDB0ABEC1F327ULL, "llac_lt_gallowstemplar_n_2383_2.dds"},
	{0x69CD5722C358DE21ULL, "231116476.ogg"},
	{0x69CD87C1418B24ACULL, "mag170_thrask6016031.cl"},
	{0x69CDA98708059361ULL, "812963282.ogg"},
	{0x69CDC6D603EAD53AULL, "vxlt_gallowscourtyard_722_0_1.vlm"},
	{0x69CDEEE90C89979DULL, "58982409.ogg"},
	{0x69CE06B9F73828FBULL, "132045756.ogg"},
	{0x69CE5C2CEEBD9714ULL, "art\\characters\\playercharacter\\humanmale\\bdy_magb.phy"},
	{0x69CE7903F332B33DULL, "gen00fl_merrill6128770.cl"},
	{0x69CF08543C5AFA40ULL, "face_face_p1.psh.fxo.dx11"},
	{0x69CF60D6B681E5F6ULL, "190500731.ogg"},
	{0x69CF7DFBE1FCD386ULL, "867732638.ogg"},
	{0x69CF85C39DA5021EULL, "490880333.ogg"},
	{0x69CF8AD5DE17B18DULL, "vxdr_0_deeproads_eu9_0_0.vlm"},
	{0x69CF8C8B0B80E82BULL, "hlac_lt_0_lowtown_docks_a3_-2_5z.dds"},
	{0x69D08C9E45049B3CULL, "lt_gamlenshouse_n_mmap_ph.dds"},
	{0x69D0A3CDC01EC104ULL, "408033626.ogg"},
	{0x69D0B242084D2A7BULL, "433143524.ogg"},
	{0x69D14A9D2BDE93BAULL, "vxlt_0_lowtown_docks_bew_0_0.vlm"},
	{0x69D1528E51262DECULL, "hlac_lt_backalley_g_0_19x.dds"},
	{0x69D198D63B395967ULL, "6141700_m.fxe"},
	{0x69D2341DFCD800F6ULL, "vxht_keep_n_60r_0_0.vlm"},
	{0x69D2AF4022DA2BC5ULL, "542306629.ogg"},
	{0x69D2CBCFDE303D04ULL, "hlac_lt_0_lowtown_a3_n_c_65537_6y.dds"},
	{0x69D31199D99FA616ULL, "w_encounter1_54_0.msh"},
	{0x69D31B99D99FB714ULL, "w_encounter1_54_0.mmh"},
	{0x69D31F99D99FBE27ULL, "w_encounter1_54_0.mao"},
	{0x69D34BE0DA923DCFULL, "vxlt_0_lowtown_a2_n_c_hz5_0_0.vlm"},
	{0x69D37E4308651CF5ULL, "vxw_encounter2_3d_0_0.vlm"},
	{0x69D3BA6EED68132CULL, "415504650.ogg"},
	{0x69D475467966D5F1ULL, "170609523.ogg"},
	{0x69D4CC5CE2F90934ULL, "477576718.ogg"},
	{0x69D50D9455D8C6FBULL, "480902200.ogg"},
	{0x69D51AE57513217BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_lp.ani"},
	{0x69D54A1501CA3D41ULL, "6171806_m.fxe"},
	{0x69D59001C10B836DULL, "lowtown_male_urchin_sta6164184.cl"},
	{0x69D5D39950486024ULL, "vxlt_0_lowtown_docks_a3_n_cx7_0_0.vlm"},
	{0x69D66B813BF5CE7EULL, "416852475.ogg"},
	{0x69D6CD05E6D40ACCULL, "vxw_woundedcoast_vk_0_0.vlm"},
	{0x69D7283B5680EAD8ULL, "446276719.ogg"},
	{0x69D788FD51568630ULL, "w_blightlands_253_0.mao"},
	{0x69D78CFD51568D43ULL, "w_blightlands_253_0.mmh"},
	{0x69D7930C00A972FCULL, "6000989_m.fxe"},
	{0x69D79AFD5156A49DULL, "w_blightlands_253_0.msh"},
	{0x69D7A05A77397DAFULL, "41731096.ogg"},
	{0x69D7C54900982123ULL, "vxlt_gallowstemplar_n_2oq_0_0.vlm"},
	{0x69D8392CEEC5B011ULL, "art\\characters\\playercharacter\\humanmale\\bdy_magb.msh"},
	{0x69D8487B3406AA92ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_activ_c.impa.evt"},
	{0x69D8532CEEC5DCCFULL, "art\\characters\\playercharacter\\humanmale\\bdy_magb.mmh"},
	{0x69D866FC5C196745ULL, "731566311.ogg"},
	{0x69D87B7AB9E60FEAULL, "186663736.ogg"},
	{0x69D8E1A538B0840DULL, "34987453.ogg"},
	{0x69D91480B498AAD5ULL, "follower_banter6153988.cl"},
	{0x69D92A39B228B6CEULL, "hlac_ht_0_hightown_65533_0y.dds"},
	{0x69DA04CEC8BEDFE0ULL, "6049817_f.fxe"},
	{0x69DA4233382FC740ULL, "6118723_m.fxe"},
	{0x69DA610B67881BFAULL, "6108051_f.fxe"},
	{0x69DB09108F349246ULL, "896782307.ogg"},
	{0x69DB54270F220137ULL, "6135822_m.fxe"},
	{0x69DB680D33F4B2EAULL, "dae000_keep_guard_2_sta.cnv"},
	{0x69DBA52CAEAEC1ABULL, "art\\characters\\playercharacter\\humanmale\\glv_medj.phy"},
	{0x69DC05E7D530AFDEULL, "dae000_bartrand6201392.cl"},
	{0x69DC4C43B7BE0B9EULL, "6142898_m.fxe"},
	{0x69DC6CFBEDBF7B0FULL, "zz_enable.ncs"},
	{0x69DCC6515ECE569FULL, "vxca_caverns1_mines_i2p_0_0.vlm"},
	{0x69DD1F279B7864A3ULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.dura.ani"},
	{0x69DD55E69266B9A2ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchwall.phy"},
	{0x69DD836C9BDCB4D4ULL, "llac_lt_0_lowtown_a2_n_65536_5.dds"},
	{0x69DD9DBB78D1B6AAULL, "dae011_corff_sta6164580.cl"},
	{0x69DDE0566984C234ULL, "character_character_p27.psh.fxo.dx11.assembly"},
	{0x69DE4A0D33F6B512ULL, "dae000_keep_guard_2_sta.bnk"},
	{0x69DE80A4E96206A8ULL, "p_dot.mao"},
	{0x69DEB25EAE73E168ULL, "qun131tr_warrens_thug.ncs"},
	{0x69DEC0344039D66AULL, "vxdr_primevaltomb_6qc_0_0.vlm"},
	{0x69DEDC8F02618C66ULL, "265554939.ogg"},
	{0x69DF17D900BDE536ULL, "ave331_post_fight6152560.cl"},
	{0x69DF2AADD273B2BAULL, "6080104_f.fxe"},
	{0x69DF677B340CFE25ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_activ_c.impa.gad"},
	{0x69DF7363A449C11DULL, "225108740.ogg"},
	{0x69DF799A8422ADBEULL, "752319537.ogg"},
	{0x69DFB8980F33A38AULL, "hlac_lt_blackemporium_1055_21y.dds"},
	{0x69DFCCD163471FD0ULL, "zz_followers_banter6157556.cl"},
	{0x69DFDA36443D5B3FULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb15a.msh"},
	{0x69DFE036443D65E1ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb15a.mmh"},
	{0x69E0308C7090F702ULL, "vxca_cave_2bk_0_0.vlm"},
	{0x69E04B710D7EECD3ULL, "docks_male_sailor_mob6176658.cl"},
	{0x69E04C0877A0185FULL, "vxca_caverns2prc_9pu_0_0.vlm"},
	{0x69E0F0C838890DBEULL, "dae000_unhappy_mage6165326.cl"},
	{0x69E1396DE1712B5BULL, "lt_gallowstemplar_2395.rml"},
	{0x69E1BDBDCC88564FULL, "vxlt_0_lowtown_a2_n_c_hu4_0_0.vlm"},
	{0x69E275205C4BBB8FULL, "mag110_vincento6167362.cl"},
	{0x69E322E5751F6C72ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_lp.evt"},
	{0x69E44E6950BDE125ULL, "6135895_m.fxe"},
	{0x69E4833C93E12817ULL, "6167609_m.fxe"},
	{0x69E499A4B08E3659ULL, "47120607.ogg"},
	{0x69E4E4326D3804BFULL, "6153228_m.fxe"},
	{0x69E4E46CCE101E82ULL, "zz_isa_debug6133968.cl"},
	{0x69E50B90E4F7BBEFULL, "l_qcr210ar_lowtown_nt.wbd"},
	{0x69E52F5314EAC1B5ULL, "vxw_bonepitdestroyed_a3_2ab_0_0.vlm"},
	{0x69E53EE2D8CA234CULL, "vxw_woundedcoast_sh_0_0.vlm"},
	{0x69E543EF44A18BC5ULL, "vxht_0_hightown_a4m_0_0.vlm"},
	{0x69E559E071B7F247ULL, "mag311_keran6041406.cl"},
	{0x69E575DA12CAD47BULL, "6076389_m.fxe"},
	{0x69E5AF7F3D6A4085ULL, "594501256.ogg"},
	{0x69E5F6A323554254ULL, "fen110_fenris6100312.cl"},
	{0x69E6BA43FBDA97C3ULL, "ser120_ghyslain6164764.cl"},
	{0x69E6CAF0096E3699ULL, "6202204_f.fxe"},
	{0x69E82EF7922F43C9ULL, "vxw_woundedcoast_5g_0_0.vlm"},
	{0x69E848572E1B474CULL, "6076077_f.fxe"},
	{0x69E84FC4E13C3FE5ULL, "follower_banter6169761.cl"},
	{0x69E8739D34569FCDULL, "vxht_mansion2_1u2_0_0.vlm"},
	{0x69E8ABBFA5BC7BE8ULL, "6041788_m.fxe"},
	{0x69E8B4102A25DFC0ULL, "6117623_m.fxe"},
	{0x69E941E575240C05ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_hr_lp.gad"},
	{0x69E95219ACC1AD28ULL, "ts_circleofmagi_03.gda"},
	{0x69E99ADE68BE1800ULL, "hlac_lt_warehouse_1_2x.dds"},
	{0x69E9B5563DF8A21FULL, "dae300_cullen6135335.cl"},
	{0x69EA2DC767815C4CULL, "6101484_m.fxe"},
	{0x69EA3CFCC893D049ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic.phy"},
	{0x69EAB002996DD554ULL, "6167277_m.fxe"},
	{0x69EAD44D356F6DB2ULL, "w_woundedcoast_82_0.rml"},
	{0x69EAD6148724B20BULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v4.evt"},
	{0x69EAF4E1888A73A0ULL, "vxw_bonepitdestroyed_a3_1po_0_0.vlm"},
	{0x69EB0E3DB5C15156ULL, "vxw_bonepitdestroyed_a3_17p_0_0.vlm"},
	{0x69EB37279B84CA9AULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.dura.evt"},
	{0x69EB48D7DB451D1BULL, "354711794.ogg"},
	{0x69EB942BAA8DD11CULL, "hlac_lt_darktown_-1_76y.dds"},
	{0x69EBBC493A2B8DBEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_silence.ani"},
	{0x69EC398708FFA072ULL, "6052833_f.fxe"},
	{0x69EC51F7D3E622A3ULL, "llac_w_woundedcoast_-262137_0.dds"},
	{0x69EC78703ACC1D5AULL, "6001005_m.fxe"},
	{0x69EC8D8382A3895FULL, "518953972.ogg"},
	{0x69ECB864731840F1ULL, "vxdr_primevaltomb_8lm_0_0.vlm"},
	{0x69ECB9674C9ACAB2ULL, "734652122.ogg"},
	{0x69ED06B0285D4D7FULL, "hf_gxa_vgk_cnspr_2.mor"},
	{0x69ED514CE9F2F03EULL, "llac_lt_0_lowtown_a3_n_c_65535_7.dds"},
	{0x69ED545CBDB75148ULL, "6105808_m.fxe"},
	{0x69ED6BD680D5CE69ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.anb"},
	{0x69EDD027FBA1091BULL, "6087819_m.fxe"},
	{0x69EDDD94FB40A657ULL, "gen00fl_sebastian6174693.cl"},
	{0x69EE3C7CA60FBE2EULL, "w_blightlands_142_0.mmh"},
	{0x69EE487CA60FD2C5ULL, "w_blightlands_142_0.mao"},
	{0x69EE567CA60FEA1CULL, "w_blightlands_142_0.msh"},
	{0x69EE90DC798DA102ULL, "377698501.ogg"},
	{0x69EEA2CFB0C0DF27ULL, "81489443.ogg"},
	{0x69EF11BA8FED8578ULL, "6069665_m.fxe"},
	{0x69EF53CA9D8368B8ULL, "142292675.ogg"},
	{0x69EF6FB9F9A03710ULL, "818334588.ogg"},
	{0x69EF91F27DD2AC8EULL, "600748229.ogg"},
	{0x69F007652CC39024ULL, "follower_banter6136656.cl"},
	{0x69F01B18E1E2FB86ULL, "vxht_brothel_n_5m4_0_1.vlm"},
	{0x69F02F9ABE7E6ECFULL, "vxlt_0_lowtown_a2_n_c_hp7_0_0.vlm"},
	{0x69F03B85B4A074B3ULL, "hlac_lt_0_lowtown_a2_-131072_3x.dds"},
	{0x69F08FA9AE7A721CULL, "vxht_mansion2_2pw_0_0.vlm"},
	{0x69F0DBB2563F7CFDULL, "llac_ht_brothel_n_3365_1.dds"},
	{0x69F125685985D1B9ULL, "pro000_opening6178539.cl"},
	{0x69F1630578F8C123ULL, "character_character_p99.psh.fxo.dx11"},
	{0x69F1EF14872AFBCCULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v4.gad"},
	{0x69F21A7D660DDFC2ULL, "6028417_m.fxe"},
	{0x69F21F12113C16EEULL, "6156491_m.fxe"},
	{0x69F236279B8AE78DULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.dura.gad"},
	{0x69F2D7A2F469BCB0ULL, "hlac_ca_caverns1_24424_18x.dds"},
	{0x69F2E5EEF89F003CULL, "6152615_m.fxe"},
	{0x69F2F56A92D2B8A5ULL, "196423680.ogg"},
	{0x69F3165C88640E5DULL, "hlac_lt_0_lowtown_a3_n_c_65536_5z.dds"},
	{0x69F37252D96B1B23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\dash.evt"},
	{0x69F3BFF41BCB4E2BULL, "6095706_m.fxe"},
	{0x69F43D5464035846ULL, "gen00fl_aveline6101425.cl"},
	{0x69F47C5FBC2409B6ULL, "gen_im_npc_elf_arm_cht_dala.uti"},
	{0x69F484BBDFF17035ULL, "133354222.ogg"},
	{0x69F55A6C63D312C3ULL, "739972176.ogg"},
	{0x69F5B8B9B1815DDAULL, "gen00fl_anders6110051.cl"},
	{0x69F5EF206CBB270FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_coma_0s.dds"},
	{0x69F61F38DB6B17B6ULL, "ss_cityelf_f_1.wbd"},
	{0x69F6A0D484255C49ULL, "vxw_woundedcoast_2f_0_0.vlm"},
	{0x69F6E57A264CB84DULL, "vxht_mansion2_1x1_0_0.vlm"},
	{0x69F6E98EB41F32D0ULL, "art\\characters\\heraldry\\hld_resolutionists.dds"},
	{0x69F722FFD6D863E2ULL, "vxht_mansion_n_32a_0_0.vlm"},
	{0x69F768A8FB2EBFF2ULL, "6164050_m.fxe"},
	{0x69F7C2FA95C7EB45ULL, "hlac_lt_gallowstemplar_2384_6x.dds"},
	{0x69F8D4493A36408FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_silence.evt"},
	{0x69F8EE14873118CAULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v4.ani"},
	{0x69F91DFF5C540BA9ULL, "vxw_bonepit_mr_0_0.vlm"},
	{0x69F965BCEE287D82ULL, "6084680_m.fxe"},
	{0x69F966BE7A808C20ULL, "vxw_bonepitdestroyed_a3_1ul_0_0.vlm"},
	{0x69F9801AA7B769D6ULL, "vxw_bonepitdestroyed_a3_18q_0_0.vlm"},
	{0x69F9E2CEC1562BB0ULL, "lt_0_lowtown_docks_117_0.mao"},
	{0x69F9E6CEC15632C3ULL, "lt_0_lowtown_docks_117_0.mmh"},
	{0x69F9F4CEC1564A1DULL, "lt_0_lowtown_docks_117_0.msh"},
	{0x69FA0A05BFE15C22ULL, "art\\vfx\\materials\\vfx_flash.mao"},
	{0x69FA471D46420D7FULL, "art\\characters\\weapons\\shields\\textures\\w_shd_005a_0t.dds"},
	{0x69FA4B52D970F7A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\dash.gad"},
	{0x69FA56CC28408505ULL, "547833258.ogg"},
	{0x69FADA1162622504ULL, "gen_im_arm_cht_coma_t3.uti"},
	{0x69FAE045D6F45026ULL, "slimebone.mao"},
	{0x69FB41282347EA47ULL, "258580683.ogg"},
	{0x69FB5D93BEF98562ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_vael_0d.dds"},
	{0x69FBD2FDC769D26EULL, "vxca_caverns1_mines_j23_0_1.vlm"},
	{0x69FCCFCDCC5074F5ULL, "hlac_ht_bartrand_n_3367_1x.dds"},
	{0x69FCD4E7DC82C2D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\pp_anders.anb"},
	{0x69FD2CA32CE12206ULL, "vxht_bartrand_n_5me_0_0.vlm"},
	{0x69FD5F0D1BE4E34BULL, "vxht_mansion2_239_0_0.vlm"},
	{0x69FDBCAE2ABD45D0ULL, "lt_0_lowtown_a3_n_c_25_0_lc.mmh"},
	{0x69FE6F316075A8AAULL, "zz_combatbalance.ncs"},
	{0x69FE811C747283B7ULL, "w_bonepitdestroyed_a3_191_0.mmh"},
	{0x69FE851C74728AC4ULL, "w_bonepitdestroyed_a3_191_0.mao"},
	{0x69FE971C7472A929ULL, "w_bonepitdestroyed_a3_191_0.msh"},
	{0x69FEAA21B850F8A1ULL, "360159805.ogg"},
	{0x69FFD5493A3C6138ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_silence.gad"},
	{0x6A0078D16AA87064ULL, "151751026.ogg"},
	{0x6A00B8ECBD250E6EULL, "t3_har_blk.tnt"},
	{0x6A013E4E5A49196FULL, "fen210_fenris6109393.cl"},
	{0x6A01469A80CBA99BULL, "llac_lt_hangedman_n_-1_23.dds"},
	{0x6A01552E22B38D02ULL, "gen_im_trs_dem_amu_r0.uti"},
	{0x6A015A52D9772F82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\dash.ani"},
	{0x6A016DAA2F12D2F7ULL, "vxlt_0_lowtown_e18_0_0.vlm"},
	{0x6A018E7597F270C8ULL, "6146811.cut"},
	{0x6A01C142B52EC9D7ULL, "182929190.ogg"},
	{0x6A01DE850B69AB39ULL, "6115318_m.fxe"},
	{0x6A01F65BC4E41D90ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mor.mmh"},
	{0x6A02045BC4E435AAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mor.msh"},
	{0x6A02868B63185075ULL, "qun200_party_comment6090391.cl"},
	{0x6A02AC528AF88743ULL, "vxlt_gallowscourtyard_6zs_0_0.vlm"},
	{0x6A02DAE7E1630702ULL, "ht_0_hightown_a3_41_0.rml"},
	{0x6A0330E692874A56ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchwall.mao"},
	{0x6A033CE692875EF5ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchwall.mmh"},
	{0x6A033EE69287626BULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchwall.msh"},
	{0x6A036FFBAC2A7EB7ULL, "war_bloodfrenzyupgrade.dds"},
	{0x6A038A958B56F740ULL, "hlac_ht_keep_n_3731_1y.dds"},
	{0x6A03A795C18897E5ULL, "172902958.ogg"},
	{0x6A03AA30CB862B24ULL, "hlac_lt_undercityencounter_-1_13z.dds"},
	{0x6A03DE8F3939F5DFULL, "180261852.ogg"},
	{0x6A041FC523F494CEULL, "one110_meeran6113244.cl"},
	{0x6A04DD71997E9174ULL, "510171070.ogg"},
	{0x6A0557571842D0CDULL, "vxht_mansion2_1c0_0_0.vlm"},
	{0x6A05A352E837EAA2ULL, "262077851.ogg"},
	{0x6A05E1933B68BCBDULL, "82988301.ogg"},
	{0x6A062DA5D5124793ULL, "6045978_m.fxe"},
	{0x6A06FFECE769A243ULL, "and100_anders6097144.cl"},
	{0x6A070A730645FF48ULL, "m_dae160ar_wounded_coast.bnk"},
	{0x6A071FE0D9B62008ULL, "vxdr_primevaltomb_4zr_0_0.vlm"},
	{0x6A0736FD78311F58ULL, "vxlt_hangedman_1fx_0_0.vlm"},
	{0x6A07D31BFAC9C4CCULL, "6166300_m.fxe"},
	{0x6A082F6688E84169ULL, "327544152.ogg"},
	{0x6A0846171497EF7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_reaction.ani"},
	{0x6A08465A7FB26B35ULL, "t3_arm_spb.tnt"},
	{0x6A08AB53BD73B531ULL, "792931596.ogg"},
	{0x6A08DD46F792AD80ULL, "l_env_wal_ltdocks_wal_calm_p2p_act3.bnk"},
	{0x6A09219F0294511BULL, "6190979_m.fxe"},
	{0x6A0924E1EFF70319ULL, "72877099.ogg"},
	{0x6A0934420191E020ULL, "6042099_m.fxe"},
	{0x6A093EB15CC583E2ULL, "166508718.ogg"},
	{0x6A0941E4D75485A0ULL, "vxca_cavern1_haunted_i2c_0_0.vlm"},
	{0x6A09AFB8E794F635ULL, "co_dots.mao"},
	{0x6A0A2232457EAF79ULL, "sky_lt_house.mao"},
	{0x6A0A542B4073CA70ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp04.phy"},
	{0x6A0AE06CDB2E9E86ULL, "317412460.ogg"},
	{0x6A0AE7CEC164A7E8ULL, "lt_0_lowtown_docks_117_0.phy"},
	{0x6A0B1D73129E5903ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extshingle.dds"},
	{0x6A0B26615FAC1B4BULL, "6042438_m.fxe"},
	{0x6A0BA2DBF78C4F21ULL, "breadoven_n.dds"},
	{0x6A0BC0118AA5F137ULL, "mag170_hugh6080742.cl"},
	{0x6A0BC117119CEAC9ULL, "6076270_f.fxe"},
	{0x6A0BD5782A2D3E66ULL, "vxw_bonepit_8b_0_0.vlm"},
	{0x6A0BE1364462A25EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb15a.phy"},
	{0x6A0C06F96CDE1F3FULL, "6087118_m.fxe"},
	{0x6A0C30DC5BD92316ULL, "one350_terath6087132.cl"},
	{0x6A0CBF6152AE64FAULL, "vxlt_0_lowtown_fe6_0_0.vlm"},
	{0x6A0CC22A350F05A8ULL, "54577134.ogg"},
	{0x6A0CCB264B7F86F4ULL, "follower_banter6163117.cl"},
	{0x6A0D3391F0F3D05FULL, "vxht_brothel_n_4ah_0_1.vlm"},
	{0x6A0DD3B9B98489E0ULL, "377967328.ogg"},
	{0x6A0DF828091D1155ULL, "zz_harvester_v4.cut"},
	{0x6A0EA97B1133C3FBULL, "mcr371_scared_mage_f6086922.cl"},
	{0x6A0EAAD58E268A95ULL, "550727135.ogg"},
	{0x6A0F43C4492A2A35ULL, "283485450.ogg"},
	{0x6A0F5385DEAD407AULL, "vxlt_gallowstemplar_n_3qw_0_0.vlm"},
	{0x6A0FB3AD65D891F9ULL, "6100385_m.fxe"},
	{0x6A100F03B2A6A392ULL, "vis121_saemus6113527.cl"},
	{0x6A106B59DBCB4E12ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_3z.dds"},
	{0x6A10BAE9B2958DF5ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb103a_0t.dds"},
	{0x6A10D1FC48DE814BULL, "vxlt_shop_gb_0_0.vlm"},
	{0x6A11C1D47DAA8DEDULL, "402820481.ogg"},
	{0x6A120474AA70CC17ULL, "vxlt_gallowscourtyard_72r_0_0.vlm"},
	{0x6A12577AF076EDA1ULL, "6126781_m.fxe"},
	{0x6A126EF7015A8B8FULL, "6170292_m.fxe"},
	{0x6A1289F6239796A0ULL, "6119190_f.fxe"},
	{0x6A12D4DDD284CA5AULL, "755276862.ogg"},
	{0x6A135E02B8A0BBACULL, "dae210ip_at_playerhouse.utp"},
	{0x6A1364A5BCA59E8CULL, "6156321_m.fxe"},
	{0x6A1365D4EA909108ULL, "527917784.ogg"},
	{0x6A13BD8BB8BCF973ULL, "qun120cr_javaris6043845.cl"},
	{0x6A1448FD7C43ABE8ULL, "vxdr_primevaltomb_6nm_0_1.vlm"},
	{0x6A144A44F7C425C6ULL, "19362397.ogg"},
	{0x6A15104D4345BAB7ULL, "lgt320cr_mis_resp_swordhntr.utc"},
	{0x6A15B2CA442A3EE1ULL, "vxht_0_hightown_a85_0_0.vlm"},
	{0x6A15C41A6805922EULL, "vxht_mansion2b_2e6_0_0.vlm"},
	{0x6A15ED0D84DFF745ULL, "llac_w_bonepitmining_a2_65536_0.dds"},
	{0x6A164A785E6CBD20ULL, "vxw_bonepitdestroyed_a3_1cj_0_0.vlm"},
	{0x6A164E1714A43A72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_reaction.evt"},
	{0x6A16740AC83EE0C0ULL, "dae150_chandan6148775.cl"},
	{0x6A1677EA2B54A4CBULL, "229614702.ogg"},
	{0x6A16BB18C22095A1ULL, "153385998.ogg"},
	{0x6A16EA82ACD64E7EULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_conj_c.impa.ani"},
	{0x6A1725FCC8BA976EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic.mmh"},
	{0x6A1731FCC8BAAB05ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic.mao"},
	{0x6A173FFCC8BAC35CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic.msh"},
	{0x6A17619A36429096ULL, "hlac_lt_0_lowtown_docks_n_131071_2z.dds"},
	{0x6A1810947B3141A5ULL, "vxlt_0_lowtown_bye_0_0.vlm"},
	{0x6A18972E3EC298A2ULL, "one280_iwan6120908.cl"},
	{0x6A18B0C2BCBFEC22ULL, "209511581.ogg"},
	{0x6A190559466CDC1DULL, "vxlt_0_lowtown_docks_a3_n_d34_0_0.vlm"},
	{0x6A19203D7D3416D0ULL, "43933329.ogg"},
	{0x6A193A61F8326C99ULL, "6183818_m.fxe"},
	{0x6A19E2723DD06BB0ULL, "748878578.ogg"},
	{0x6A19F376F42B0234ULL, "gen00fl_anders6121242.cl"},
	{0x6A19FFB5E7AFA436ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a.ani"},
	{0x6A1A0C3592CFF505ULL, "hlac_lt_gallowstemplar_n_2388_0y.dds"},
	{0x6A1A137C3F787A5FULL, "vxca_caverns2prc_9lq_0_0.vlm"},
	{0x6A1A69E4487F9901ULL, "vxht_mansion_n_6op_0_0.vlm"},
	{0x6A1ADA01E18853C7ULL, "66837757.ogg"},
	{0x6A1AE8339500523BULL, "vxht_mansion_n_62u_0_0.vlm"},
	{0x6A1B1934311EDE03ULL, "art\\vfx\\combat\\vfx_gib_qunw.msh"},
	{0x6A1B2734311EF65DULL, "art\\vfx\\combat\\vfx_gib_qunw.mmh"},
	{0x6A1B2B34311EFD6EULL, "art\\vfx\\combat\\vfx_gib_qunw.mao"},
	{0x6A1B73DE6DBC77ABULL, "gen_im_acc_amu_act3_04.uti"},
	{0x6A1B8DB3B46ADBF8ULL, "bookback.gfx"},
	{0x6A1BC6D9BC6A1472ULL, "774714264.ogg"},
	{0x6A1BCB2F98A1BAF1ULL, "vxlt_0_lowtown_docks_b5p_0_0.vlm"},
	{0x6A1C6D1714A8DA05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_reaction.gad"},
	{0x6A1D078227FC6105ULL, "6072170_m.fxe"},
	{0x6A1DB725E22CEEA4ULL, "765377951.ogg"},
	{0x6A1DE140406E7A58ULL, "6120593_m.fxe"},
	{0x6A1E62BAB212B9D1ULL, "6083372_m.fxe"},
	{0x6A1ECC4DA3FD90CDULL, "6083366_f.fxe"},
	{0x6A1F1600343CBD2CULL, "6134928_f.fxe"},
	{0x6A1F2CE2D3E0A834ULL, "58902945.ogg"},
	{0x6A1F3B80057EEAC8ULL, "vxht_0_hightown_a80_0_0.vlm"},
	{0x6A1F60A1E6BF708AULL, "gen00fl_merrill6116006.cl"},
	{0x6A20809C0B48D93AULL, "803183595.ogg"},
	{0x6A2081FE6701E78CULL, "825499852.ogg"},
	{0x6A209AFB66F0C5B8ULL, "mag160_feynriel6020065.cl"},
	{0x6A20C378F601A805ULL, "6129267_m.fxe"},
	{0x6A215CA4489C7C84ULL, "dae000ip_crowd_mf6_talk6.utp"},
	{0x6A217A40D7B2B6E3ULL, "vxca_caverns1_j2b_0_0.vlm"},
	{0x6A21EB826B6ADF51ULL, "vxlt_undercityencounter_1jf_0_0.vlm"},
	{0x6A224538DB909324ULL, "ss_cityelf_f_1.bnk"},
	{0x6A224A65E754C3ECULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1a.phy"},
	{0x6A226B2FE5C3315AULL, "hlac_lt_0_lowtown_a2_n_-131072_0y.dds"},
	{0x6A227C6E23B21C24ULL, "76700105.ogg"},
	{0x6A22BA2E7361DA75ULL, "274276863.ogg"},
	{0x6A22BADA6E39C468ULL, "vxdr_primevaltomb_6ul_0_1.vlm"},
	{0x6A22C945A0483EC6ULL, "50003086.ogg"},
	{0x6A22F25C6BAD292EULL, "gen_im_arm_cht_nobb_t6.uti"},
	{0x6A23208A0DC5E6AFULL, "vxca_cave_2j2_0_0.vlm"},
	{0x6A23838EC192FCCBULL, "608814776.ogg"},
	{0x6A23BBAAA1CE195EULL, "gen_im_trs_qun_sta_nat_r1.uti"},
	{0x6A24ED4CB47AD5B6ULL, "714718389.ogg"},
	{0x6A24FFDCD8D84F40ULL, "6077280_m.fxe"},
	{0x6A250282ACE2B44FULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_conj_c.impa.evt"},
	{0x6A252590E52DCBC5ULL, "l_qcr210ar_lowtown_nt.bnk"},
	{0x6A255E5D98FE71BDULL, "lt_ruinedhouse_n_probe_tv.dds"},
	{0x6A257C9276AE1728ULL, "vxw_encounter2_57_0_0.vlm"},
	{0x6A25BAE95C2FA1FDULL, "6069585_m.fxe"},
	{0x6A25D17332CCDC98ULL, "462172482.ogg"},
	{0x6A265AA90AC329BAULL, "vxw_bonepitdestroyed_a3_1kx_0_0.vlm"},
	{0x6A26C38FCBC777A0ULL, "character_character_p180.psh.fxo.dx11"},
	{0x6A26C6B6890340C8ULL, "6086510_m.fxe"},
	{0x6A270BFAD5169330ULL, "6120939_m.fxe"},
	{0x6A2731CB59DB2B18ULL, "follower_banter6169803.cl"},
	{0x6A27428499F0A2D3ULL, "mag121_idunna6083953.cl"},
	{0x6A27826A7D19FE03ULL, "6206605_m.fxe"},
	{0x6A27C421CF5F1201ULL, "6114528_m.fxe"},
	{0x6A283C36404D51BBULL, "mag121_eau_banter6121087.cl"},
	{0x6A287128F5C1A3D1ULL, "character_character_p279.psh.fxo.dx11"},
	{0x6A28DD5302C784B9ULL, "dae000_pc_home_ambients6208700.cl"},
	{0x6A291F0C022C761EULL, "147552862.ogg"},
	{0x6A294F6518BE71A2ULL, "w_encounter3_92_0.msh"},
	{0x6A29516518BE7558ULL, "w_encounter3_92_0.mmh"},
	{0x6A295D6518BE89FBULL, "w_encounter3_92_0.mao"},
	{0x6A297940CFBD1BCEULL, "hlac_lt_0_lowtown_65538_0z.dds"},
	{0x6A29DD78CD79DCB9ULL, "vxw_woundedcoast_eg_0_0.vlm"},
	{0x6A2A4959212DC265ULL, "hlac_ht_bartrand_n_3364_0z.dds"},
	{0x6A2B7275F6CF4667ULL, "6118096_m.fxe"},
	{0x6A2BC975092BD0E0ULL, "zz_mcr_debug6071553.cl"},
	{0x6A2BC9CD2C977007ULL, "vxlt_0_lowtown_a3_n_c_gp9_0_0.vlm"},
	{0x6A2BFC73066503F2ULL, "m_dae160ar_wounded_coast.wbd"},
	{0x6A2C0382ACE8D4F8ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_conj_c.impa.gad"},
	{0x6A2C158FCA55A08CULL, "6087351_m.fxe"},
	{0x6A2C4CAD6502D516ULL, "120150443.ogg"},
	{0x6A2C54E1FCFF1B66ULL, "zz_and_debug6134053.cl"},
	{0x6A2C6B724DD9DBEBULL, "vxw_blightlands_bv_0_0.vlm"},
	{0x6A2C74922B99C826ULL, "dae000ip_crowd_em1_talk13.utp"},
	{0x6A2D3132EEFCD8F6ULL, "hlac_lt_blackemporium_1055_18y.dds"},
	{0x6A2D50BA342EBE18ULL, "mag110_samson6018693.cl"},
	{0x6A2DC679FCA32E1DULL, "hlac_w_encounter2_-65537_0x.dds"},
	{0x6A2E28B5E7C09370ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a.gad"},
	{0x6A2E92A14063C847ULL, "primitives_primitive_p1.vsh.fxo"},
	{0x6A2F2FE6AD97011AULL, "379635191.ogg"},
	{0x6A2FEE6BD301957EULL, "6177379_m.fxe"},
	{0x6A301DAF4F154BB7ULL, "chu_bl_sh.mao"},
	{0x6A305D5F5D60F7D1ULL, "vxlt_undercityencounter_1ge_0_0.vlm"},
	{0x6A307D0D6CF6DEA6ULL, "685499685.ogg"},
	{0x6A30BB6713294F0AULL, "vxht_0_hightown_9uy_0_0.vlm"},
	{0x6A31796B9F47B24EULL, "379255975.ogg"},
	{0x6A319D320CE05E28ULL, "177012563.ogg"},
	{0x6A31A3C07040F15BULL, "vxca_caverns1_mines_e8f_0_0.vlm"},
	{0x6A31DF52B5E560DFULL, "6156340_m.fxe"},
	{0x6A31DF74030CC6DCULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.impa.ani"},
	{0x6A3385683F7F1D11ULL, "vxht_playershouse_n_1v9_0_0.vlm"},
	{0x6A33D03DD21D7E85ULL, "one120_athenril6088329.cl"},
	{0x6A33D8036048251BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\exit.evt"},
	{0x6A33F249C41B2414ULL, "134514172.ogg"},
	{0x6A34D78A5D42FA39ULL, "mag291_vincento6027383.cl"},
	{0x6A350363AE72D13AULL, "730205243.ogg"},
	{0x6A350434C566EBCCULL, "vxca_caverns1_egu_0_0.vlm"},
	{0x6A354608C687C59BULL, "vxw_sundermount_2c5_0_0.vlm"},
	{0x6A35EA194D16C867ULL, "dae150_radha6148940.cl"},
	{0x6A360215CE7038A9ULL, "6152801_m.fxe"},
	{0x6A362ED2A6076430ULL, "llac_lt_0_lowtown_65534_6.dds"},
	{0x6A365ED274838F05ULL, "lt_0_lowtown_docks_n_10_0_ll.mmh"},
	{0x6A36B8B6C6F9A0ECULL, "560114702.ogg"},
	{0x6A3723ACFCEA6EDFULL, "rune_weapon_electric.dds"},
	{0x6A37996007A8A2A3ULL, "vxlt_0_lowtown_n_exl_0_0.vlm"},
	{0x6A37C00D5DEE1A07ULL, "6185384_m.fxe"},
	{0x6A37EC740311481EULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.impa.gad"},
	{0x6A384F55BF6FF539ULL, "vxw_woundedcoast_bf_0_0.vlm"},
	{0x6A39283B58F62AD8ULL, "vxw_encounter2_camp_9c_0_0.vlm"},
	{0x6A3956CD8CD6A3CFULL, "vxlt_ruinedhouse_n_id_0_0.vlm"},
	{0x6A396D81A8C9C1ADULL, "zz_one_debug6051411.cl"},
	{0x6A3990625FB10BCFULL, "vxht_0_hightown_a3_9zh_0_0.vlm"},
	{0x6A3996572518789AULL, "6081306_m.fxe"},
	{0x6A39F1ACF86D716FULL, "vxht_mansion2b_2jf_0_0.vlm"},
	{0x6A3A155206DCFEA0ULL, "534662248.ogg"},
	{0x6A3A3BAA1E8D8887ULL, "vxlt_0_lowtown_a3_n_c_gw8_0_0.vlm"},
	{0x6A3A5BA35E9923A0ULL, "vxlt_0_lowtown_gm1_0_0.vlm"},
	{0x6A3A67619175B994ULL, "hlac_lt_0_lowtown_docks_n_-262144_3z.dds"},
	{0x6A3AD103604E385CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\exit.gad"},
	{0x6A3ADCFE62562F8AULL, "6097360_f.fxe"},
	{0x6A3B2110CFC15D38ULL, "hurlock_alpha_greataxe.utc"},
	{0x6A3B40F6880234E1ULL, "6140337_f.fxe"},
	{0x6A3B6E0D131654DEULL, "dae300_bran6142925.cl"},
	{0x6A3BAA19EE9DE9B6ULL, "6129758_m.fxe"},
	{0x6A3C079746A680B9ULL, "lgt220_shp_mag_guard6149891.cl"},
	{0x6A3C672A867BAB34ULL, "194676463.ogg"},
	{0x6A3C97C1782B69CAULL, "vxht_0_hightown_n_amv_0_0.vlm"},
	{0x6A3D164E90C5B864ULL, "gen00fl_fenris6127989.cl"},
	{0x6A3D4FA11549C3BCULL, "gen00fl_isabela6099367.cl"},
	{0x6A3D6721E1D6C17FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloor.phy"},
	{0x6A3E0743312D74EEULL, "6039389_m.fxe"},
	{0x6A3E1F97CE247A85ULL, "1012250778.ogg"},
	{0x6A3E594D1F881040ULL, "loadscreens.gda"},
	{0x6A3E846BFAEE0A4EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_basic.mmh"},
	{0x6A3E906BFAEE1EE5ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_basic.mao"},
	{0x6A3E9E6BFAEE363CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_basic.msh"},
	{0x6A3EB2389E20E254ULL, "vxw_bonepitmining_a2_1kx_0_0.vlm"},
	{0x6A3EC7740317283DULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.impa.evt"},
	{0x6A3F45E12D4A2FCCULL, "988968040.ogg"},
	{0x6A3FAAA3ABECCB3CULL, "339809343.ogg"},
	{0x6A3FF46BCB5C5A86ULL, "100489426.ogg"},
	{0x6A404B284B7AFF12ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_u_c.mmh"},
	{0x6A40C0745746948CULL, "vxca_caverns2_8h2_0_0.vlm"},
	{0x6A40D0036052A25AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\exit.ani"},
	{0x6A40ED9F22DE6C03ULL, "gen_im_arm_hlm_elf_comb.uti"},
	{0x6A41B83091CC2CF9ULL, "hlac_lt_gallowscourtyard_n_-65538_2x.dds"},
	{0x6A41DF1075B68C82ULL, "01_smsglow.mao"},
	{0x6A42386EFF8FF0C1ULL, "vxlt_gallowstemplar_4yd_0_0.vlm"},
	{0x6A424E32E3E80A1AULL, "vxca_caverns1_jde_0_0.vlm"},
	{0x6A426579ED6826EAULL, "6082574_m.fxe"},
	{0x6A428FFCFCEBA0E4ULL, "423571046.ogg"},
	{0x6A42E98E460B8108ULL, "llac_lt_0_lowtown_docks_196607_3.dds"},
	{0x6A42F4F064A736BFULL, "vxlt_hangedman_wi_0_0.vlm"},
	{0x6A42FD5BC51BC0ABULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mor.phy"},
	{0x6A432103E62D7EECULL, "6208297_m.fxe"},
	{0x6A4330C1F7EA553AULL, "138415224.ogg"},
	{0x6A4369C52692C698ULL, "gen00fl_fenris6106715.cl"},
	{0x6A437611B75D044CULL, "vxca_caverns1_ebj_0_0.vlm"},
	{0x6A43A8FD93AF8888ULL, "gen_im_trs_dal_amu_r1.uti"},
	{0x6A43B7E5B87DDE1BULL, "vxw_sundermount_2d4_0_0.vlm"},
	{0x6A4447D333C5EC4DULL, "6118252_m.fxe"},
	{0x6A447C9AC2654E89ULL, "6188723_m.fxe"},
	{0x6A45618C63A90DCDULL, "6108904_f.fxe"},
	{0x6A45EC4F88747AA6ULL, "6139391_m.fxe"},
	{0x6A460B3CF99EBB23ULL, "vxlt_0_lowtown_n_euo_0_0.vlm"},
	{0x6A46662670C224C9ULL, "vxca_caverns1_e8a_0_0.vlm"},
	{0x6A46A0E1B2A2C2C4ULL, "ser221st_quentin.stg"},
	{0x6A46B63AE4C70F69ULL, "docks_male_human_worker_sta6175943.cl"},
	{0x6A473A45CA97A8A4ULL, "follower_banter6156236.cl"},
	{0x6A47E732B8D7F8FFULL, "6208666_f.fxe"},
	{0x6A481B721F516B6CULL, "113087013.ogg"},
	{0x6A481DB5DF5A374BULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ds001a_0d.dds"},
	{0x6A482BF406FE0E4AULL, "vxdr_primevalentrance_2su_0_0.vlm"},
	{0x6A484B551AD076B8ULL, "hlac_ht_brothel_n_2995_4z.dds"},
	{0x6A4854F8CE3DCA16ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_enter.ani"},
	{0x6A48562FDABB7767ULL, "6112296_m.fxe"},
	{0x6A48A75EF8240EBBULL, "6115455_m.fxe"},
	{0x6A494D8214610139ULL, "qun221_viscount6092418.cl"},
	{0x6A49808474A4D4AFULL, "6109551_m.fxe"},
	{0x6A4A0791A1F47086ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90.evt"},
	{0x6A4A344E0C2EAB13ULL, "vxca_caverns2prc_a5v_0_0.vlm"},
	{0x6A4A6D34F10D5A0AULL, "6122089_m.fxe"},
	{0x6A4C2E960B45DF68ULL, "llac_lt_0_lowtown_n_1_9.dds"},
	{0x6A4C8178CE8E05F8ULL, "ser121_jethann6104574.cl"},
	{0x6A4CB66899F2A237ULL, "follower_banter6167196.cl"},
	{0x6A4CBEB720508C77ULL, "vxht_mansion_n_62a_0_0.vlm"},
	{0x6A4D5CCC9895A7E0ULL, "6018508_m.fxe"},
	{0x6A4DBB63A573073FULL, "844903580.ogg"},
	{0x6A4E2DCC311061B3ULL, "123262180.ogg"},
	{0x6A4E31D52E440354ULL, "gen00fl_merrill6119621.cl"},
	{0x6A4E6821E1E53584ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloor.mmh"},
	{0x6A4E6C21E1E53CD7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloor.mao"},
	{0x6A4E7E21E1E55B26ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloor.msh"},
	{0x6A4EA0F6A9924C1CULL, "vxlt_0_lowtown_docks_bu1_0_0.vlm"},
	{0x6A4EA59ADCF395B0ULL, "6074348_m.fxe"},
	{0x6A4EFEBC96EB13FAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright_5.mao"},
	{0x6A4F2A894E9ADF4BULL, "one100_party_comments6106777.cl"},
	{0x6A4F35D962F92146ULL, "llac_ht_mansion2_818_59.dds"},
	{0x6A4F9CA13394D50BULL, "vxlt_hangedman_n_10r_0_0.vlm"},
	{0x6A4FADF19D331425ULL, "lgt000cr_elf_turn_in_3.utc"},
	{0x6A4FF72D28FDE232ULL, "1070082686.ogg"},
	{0x6A505C85B7F444D0ULL, "follower_banter6173057.cl"},
	{0x6A50AB9ACFC9ECE8ULL, "vxht_playershouse_n_1jx_0_0.vlm"},
	{0x6A50D756DF5E3F6AULL, "6085474_m.fxe"},
	{0x6A511C2D5B314EF3ULL, "vxdr_0_deeproads_evn_0_0.vlm"},
	{0x6A511E91A1FAB6D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90.gad"},
	{0x6A51A4C1ABAD8871ULL, "6163838_m.fxe"},
	{0x6A51B515EA798E5EULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.cess.evt"},
	{0x6A51C6416ACEA1E6ULL, "6054050_m.fxe"},
	{0x6A51EE7E761323C7ULL, "vxlt_gallowscourtyard_6ur_0_0.vlm"},
	{0x6A51F7C0D3B62965ULL, "w_bonepit_40_0.rml"},
	{0x6A520F2B86FA0050ULL, "1006751116.ogg"},
	{0x6A5229C2AA73F69BULL, "vxw_sundermount_2i7_0_0.vlm"},
	{0x6A52B052B44FBF73ULL, "6109753_m.fxe"},
	{0x6A52DA76F5358148ULL, "mag211im_reward_book.uti"},
	{0x6A53411B129C66E3ULL, "ts_demonic_wpn_table.gda"},
	{0x6A534B61AD85703AULL, "vxlt_0_lowtown_docks_b0h_0_0.vlm"},
	{0x6A53FBCC7FA1F2CDULL, "mcr371pt_harvester.ncs"},
	{0x6A541209DD546E7DULL, "dre000_dougal6131129.cl"},
	{0x6A54B77FC59F4941ULL, "vxht_0_hightown_a3_amo_0_0.vlm"},
	{0x6A54E112AE46A0F2ULL, "vxw_bonepitdestroyed_a3_4r_0_0.vlm"},
	{0x6A560639E3F9D0C6ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.impa.ani"},
	{0x6A561AC0E332964AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic.mao"},
	{0x6A561EC0E3329C99ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic.mmh"},
	{0x6A5628C0E332AD87ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic.msh"},
	{0x6A565CF8CE4A1507ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_enter.evt"},
	{0x6A56705981E54483ULL, "657811274.ogg"},
	{0x6A56E34B6875E642ULL, "513837939.ogg"},
	{0x6A5709A1BD853C4FULL, "143529898.ogg"},
	{0x6A571ED6AA5CF354ULL, "ht_0_hightown_37_0_ll.mmh"},
	{0x6A571F91A1FF247FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90.ani"},
	{0x6A575484D233F252ULL, "6112113_f.fxe"},
	{0x6A57777DD26966AEULL, "6180958_f.fxe"},
	{0x6A58345C0CABE781ULL, "6094516_m.fxe"},
	{0x6A584BA92B5E651AULL, "w_encounter1_45_0.mmh"},
	{0x6A584EDFFC4DEC4CULL, "hlac_lt_0_lowtown_a2_n_c_-65536_6x.dds"},
	{0x6A584FA92B5E6C51ULL, "w_encounter1_45_0.mao"},
	{0x6A585DA92B5E8460ULL, "w_encounter1_45_0.msh"},
	{0x6A589916D1879712ULL, "hair0_pass1.psh.fxo"},
	{0x6A589EB08E37713DULL, "hlac_lt_gallowscourtyard_-65538_0y.dds"},
	{0x6A58BE3AA49B85B5ULL, "vxw_bonepitdestroyed_a3_29j_0_0.vlm"},
	{0x6A58CBFEF4B75393ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1_b.ani"},
	{0x6A58CC15EA7FD389ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.cess.gad"},
	{0x6A58F0795BC8DC8EULL, "6180845_m.fxe"},
	{0x6A58F29CC7189882ULL, "scaleform_default.mao"},
	{0x6A594558272BEB42ULL, "vxdr_0_deeproads_9is_0_1.vlm"},
	{0x6A594E1301EA69D9ULL, "vis121_saemus6086707.cl"},
	{0x6A59A7F5E5854E8EULL, "27051106.ogg"},
	{0x6A59D01EC3F49484ULL, "vxlt_darktown_6om_0_0.vlm"},
	{0x6A5A28F6B8D03903ULL, "zz_crowd_plc_test_1.utp"},
	{0x6A5A60B9D2AD813AULL, "llac_lt_0_lowtown_docks_-196611_1.dds"},
	{0x6A5A611CC72B3785ULL, "6174412_m.fxe"},
	{0x6A5A9AAA20FB42B8ULL, "hlac_w_encounter2_camp_-131070_0y.dds"},
	{0x6A5AA6EF93698820ULL, "521155173.ogg"},
	{0x6A5B12F82F1785E6ULL, "vxlt_0_lowtown_gjb_0_0.vlm"},
	{0x6A5B87A24C105F88ULL, "287558641.ogg"},
	{0x6A5B92C6EBA124CBULL, "6175813_m.fxe"},
	{0x6A5BABA1DE25B30EULL, "vxlt_gamlenshouse_n_bl_0_0.vlm"},
	{0x6A5BE3B4BB287825ULL, "gen00fl_anders6125375.cl"},
	{0x6A5D330709F098E6ULL, "6209844_f.fxe"},
	{0x6A5D45367435CF54ULL, "6174444_m.fxe"},
	{0x6A5D7DF8CE506C50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_enter.gad"},
	{0x6A5D9AD88ED6F90CULL, "vxht_0_hightown_bm4_0_0.vlm"},
	{0x6A5DA42E3B32C58CULL, "vxca_caverns2_8r0_0_0.vlm"},
	{0x6A5DF5985F726D8BULL, "hlac_lt_0_lowtown_a2_0_1y.dds"},
	{0x6A5E08205B7FC0D2ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appc_0t.dds"},
	{0x6A5E1EFFF5720AFAULL, "6182488_f.fxe"},
	{0x6A5E35B540717D7FULL, "298988104.ogg"},
	{0x6A5E5B9D0F68C4E9ULL, "vxca_cavern1_haunted_fna_0_0.vlm"},
	{0x6A5E6C221B73DBACULL, "ch1emarmtnt_vl.psh.fxo"},
	{0x6A5EC8BF61CB32D0ULL, "vxlt_gallowstemplar_f_5z2_0_0.vlm"},
	{0x6A5ED60B89F844D8ULL, "vxca_caverns1_j6a_0_0.vlm"},
	{0x6A5F98A0FB4D6E0FULL, "vxca_cave_2d8_0_0.vlm"},
	{0x6A5F9D15EA85A2D7ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.cess.ani"},
	{0x6A5FEA36CF656BC5ULL, "ico_hold.dds"},
	{0x6A6025F96C630833ULL, "6068474_m.fxe"},
	{0x6A6059CB9B49354CULL, "vxca_caverns1_ehh_0_0.vlm"},
	{0x6A609B9F9C6A0F1BULL, "vxw_sundermount_2r6_0_0.vlm"},
	{0x6A60B6E24329DB1DULL, "hlac_ca_caverns2_5802_56y.dds"},
	{0x6A60E25804CE25B4ULL, "gen00fl_bethany6206062.cl"},
	{0x6A60EBE6FA00B1B4ULL, "6145097_m.fxe"},
	{0x6A615FF19173917BULL, "724018502.ogg"},
	{0x6A617FE1028F5A03ULL, "zz_approval_debug6003101.cl"},
	{0x6A6187EE91D0B0FAULL, "6161013_m.fxe"},
	{0x6A61913812138FD0ULL, "6047412_m.fxe"},
	{0x6A61A6A1F8F94202ULL, "105262149.ogg"},
	{0x6A61BD3E9F7B88BAULL, "vxlt_0_lowtown_docks_b7i_0_0.vlm"},
	{0x6A61CFDFDA59EA20ULL, "6118510_m.fxe"},
	{0x6A61DC204E01CF89ULL, "art\\characters\\playercharacter\\children\\mt_bas_km_scalpm1.mao"},
	{0x6A61DE4FCEA7C9A2ULL, "626037910.ogg"},
	{0x6A61DF75884ABBFEULL, "mrl351_pridedemon6170228.cl"},
	{0x6A61FB15A65BC2EDULL, "hlac_ht_mansion2b_818_17y.dds"},
	{0x6A620154F9E4982BULL, "6087392_m.fxe"},
	{0x6A62705919C33F48ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyn_0t.dds"},
	{0x6A62E7F99361E245ULL, "139715555.ogg"},
	{0x6A632165E78C14B7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1a.mmh"},
	{0x6A632565E78C1BC4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1a.mao"},
	{0x6A632857386EFF68ULL, "vxw_bonepitmining_a2_1p5_0_0.vlm"},
	{0x6A63295CB79561C1ULL, "vxht_0_hightown_a3_ajn_0_0.vlm"},
	{0x6A633765E78C3A29ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_mrtchtble1a.msh"},
	{0x6A637EC3F75A93AFULL, "mer220_hubert6081399.cl"},
	{0x6A63A4EC95523EB9ULL, "vxw_woundedcoast_qk_0_0.vlm"},
	{0x6A63EE39E405E537ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.impa.evt"},
	{0x6A63F56F0A8AB767ULL, "hlac_ht_0_hightown_a2_-65536_2x.dds"},
	{0x6A643F6DDC5C2265ULL, "6140169_m.fxe"},
	{0x6A643F9BB13EA9C2ULL, "vxlt_0_lowtown_docks_a3_cz0_0_0.vlm"},
	{0x6A647F9D7A1C9226ULL, "mag360cr_templar.utc"},
	{0x6A6505D480B733A8ULL, "285436482.ogg"},
	{0x6A65409A52C96F14ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_004a_0s.dds"},
	{0x6A65411EE10CD2FDULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_0d.dds"},
	{0x6A654739F5AEC61AULL, "hlac_lt_0_lowtown_a2_-262144_4y.dds"},
	{0x6A666787D6B3896AULL, "819586588.ogg"},
	{0x6A66C3FEF4C383CAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1_b.evt"},
	{0x6A66CD175520D4E9ULL, "6103326_m.fxe"},
	{0x6A67301796919E35ULL, "vxw_bonepitdestroyed_a3_24i_0_0.vlm"},
	{0x6A6741DD6C884B5BULL, "1777373.ogg"},
	{0x6A678D686074BB34ULL, "qcr220_grey_wardens6060409.cl"},
	{0x6A6792DD0DAD7005ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brazier_d.dds"},
	{0x6A67B9987EAF749AULL, "w_woundedcoast_68_0.rml"},
	{0x6A694A284B9E02B9ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_u_c.anb"},
	{0x6A6953743280A02FULL, "6128939_f.fxe"},
	{0x6A698AC24E943031ULL, "6070600_m.fxe"},
	{0x6A69A0EEFCF86385ULL, "mag321_comte_de_launcet.bnk"},
	{0x6A69B0F7E967532DULL, "7689991.ogg"},
	{0x6A69C2379717FC32ULL, "6021133_m.fxe"},
	{0x6A6A055FAB77FC1CULL, "tal_sha_dot.mao"},
	{0x6A6A60F60D1EB0A7ULL, "6207394_m.fxe"},
	{0x6A6A6E3AE6F08E90ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfa.mao"},
	{0x6A6A723AE6F095A3ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfa.mmh"},
	{0x6A6A803AE6F0ADFDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfa.msh"},
	{0x6A6AB56032BBD2C2ULL, "337826353.ogg"},
	{0x6A6AEF39E40C05C0ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.impa.gad"},
	{0x6A6AFED2B4098C58ULL, "aip_fleshgolem_plot.gda"},
	{0x6A6B0630844D1656ULL, "vxht_keep_5i1_0_0.vlm"},
	{0x6A6B1B310ED0EB17ULL, "6084686_f.fxe"},
	{0x6A6B617A51031F9AULL, "6110053_m.fxe"},
	{0x6A6C82EEFCFA66BDULL, "mag321_comte_de_launcet.cnv"},
	{0x6A6C94BD6982CF2FULL, "6095096_m.fxe"},
	{0x6A6CE68EC7C2C2B6ULL, "217225586.ogg"},
	{0x6A6D1560E4600C76ULL, "follower_banter6155532.cl"},
	{0x6A6D367480D34448ULL, "vxlt_gallowstemplar_3yn_0_0.vlm"},
	{0x6A6D5744059A3996ULL, "mcr370_cut_player_rules6171811.cl"},
	{0x6A6DA2FEF4C96A3DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1_b.gad"},
	{0x6A6DE636246A420AULL, "vxht_chantry_n_23w_0_0.vlm"},
	{0x6A6E3883E73474A2ULL, "6039965_m.fxe"},
	{0x6A6E7F89B8036A92ULL, "6027351_m.fxe"},
	{0x6A6F3AD88E00D8AFULL, "6139992_m.fxe"},
	{0x6A6FC5DD556206C0ULL, "425208513.ogg"},
	{0x6A6FF8BEE698DA58ULL, "273213936.ogg"},
	{0x6A6FF9CB83C6C97AULL, "6166113_m.fxe"},
	{0x6A704A61E0FE0752ULL, "fen210_orana6109758.cl"},
	{0x6A70AE537B9377EDULL, "ser221_party_comments6122876.cl"},
	{0x6A70F3F769A2AAF9ULL, "vxlt_warehouse_n_xw_0_0.vlm"},
	{0x6A71734C133379F9ULL, "zz_one_debug6077411.cl"},
	{0x6A71DD9B69C89BEEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_and_walk.ani"},
	{0x6A726B268CFC38F8ULL, "vxht_chantry_32x_0_0.vlm"},
	{0x6A72B1F289164742ULL, "hlac_lt_0_lowtown_n_65538_2y.dds"},
	{0x6A72BCEF0EF4AEB3ULL, "dae000ip_crowd_ff5_talk18.utp"},
	{0x6A72C25EE4174A52ULL, "llac_ca_caverns1_24426_1.dds"},
	{0x6A72EE16AEB35DC9ULL, "sky_w_encounter2_camp.mao"},
	{0x6A734B905EAA831AULL, "vxw_encounter2_camp_ia_0_0.vlm"},
	{0x6A735160E7EB0B41ULL, "vxht_0_hightown_cbe_0_0.vlm"},
	{0x6A73B8043D63F180ULL, "6122788_m.fxe"},
	{0x6A745F1BEAE49474ULL, "mrl000_marethari6129236.cl"},
	{0x6A747193B352DDD9ULL, "vxht_0_hightown_c0g_0_0.vlm"},
	{0x6A7492E21CF57A4DULL, "vxlt_gallowscourtyard_4mv_0_0.vlm"},
	{0x6A7577E7023CF134ULL, "one111ip_ggt_crate_8.utp"},
	{0x6A75C70319EA166BULL, "vxw_sundermount_2jn_0_0.vlm"},
	{0x6A769857B99E6B0CULL, "390429654.ogg"},
	{0x6A76A80C473D06CEULL, "220601964.ogg"},
	{0x6A76D7C0D3D50E8EULL, "w_bonepit_40_0.mmh"},
	{0x6A76E3C0D3D52325ULL, "w_bonepit_40_0.mao"},
	{0x6A76F1C0D3D53B7CULL, "w_bonepit_40_0.msh"},
	{0x6A772CD3A556E69CULL, "dae000_bartrand6206971.cl"},
	{0x6A774A0243806840ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls102a.phy"},
	{0x6A77819C6E79B77AULL, "6089720_m.fxe"},
	{0x6A77C223531B8A41ULL, "vxca_cavern1_haunted_j1y_0_1.vlm"},
	{0x6A786A6AE5E32F5BULL, "art\\characters\\creatures\\spider\\animation\\crnf.evt"},
	{0x6A78762896C59356ULL, "vxca_caverns1_hvm_0_0.vlm"},
	{0x6A790D92865CBA84ULL, "619070537.ogg"},
	{0x6A792F6581784420ULL, "qcr220_grey_wardens6135930.cl"},
	{0x6A79457079603520ULL, "vxdr_primevaltomb_8th_0_0.vlm"},
	{0x6A797E9E92CA1F79ULL, "900581909.ogg"},
	{0x6A7997C2F4B7A439ULL, "155644725.ogg"},
	{0x6A79B9C20699863AULL, "6155937_m.fxe"},
	{0x6A79BF7DB91D299DULL, "fake_mage_male.utc"},
	{0x6A79F68D7F74959CULL, "vxlt_0_lowtown_docks_bb4_0_0.vlm"},
	{0x6A7A3628E15B9F1DULL, "vxht_chantry_n_2tp_0_1.vlm"},
	{0x6A7A77458174416FULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_meda_0s.dds"},
	{0x6A7B1514DC7EE405ULL, "cd_mag311_huon.wbd"},
	{0x6A7B2F624ECA870DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgta_0n.dds"},
	{0x6A7B3D933C7EFE7AULL, "6112891_m.fxe"},
	{0x6A7B733AE6FF0AC8ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfa.phy"},
	{0x6A7BB46C44B5E67AULL, "6156510_m.fxe"},
	{0x6A7CA1223B2FC3C5ULL, "326790550.ogg"},
	{0x6A7D10E919B448BDULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_overlay.anb"},
	{0x6A7D359E0811785DULL, "vxlt_0_lowtown_docks_a3_aor_0_0.vlm"},
	{0x6A7D6DB5BCE9D940ULL, "vxw_woundedcoast_h4_0_1.vlm"},
	{0x6A7E956BADFB8AB9ULL, "818336990.ogg"},
	{0x6A7EA59DEA7E0DD8ULL, "art\\levels\\ztools\\default\\generic\\9ff_bar.phy"},
	{0x6A7EB399BE461B20ULL, "6112710_m.fxe"},
	{0x6A7ED59B69D3183FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_and_walk.evt"},
	{0x6A7F1B701C616E10ULL, "w_encounter2_19_0_lc.mmh"},
	{0x6A7F5F3822850162ULL, "one300_nuncio6083997.cl"},
	{0x6A7F636AE5E9419CULL, "art\\characters\\creatures\\spider\\animation\\crnf.gad"},
	{0x6A7F9DE0D7C0280EULL, "vxlt_0_lowtown_a2_n_c_ehw_0_0.vlm"},
	{0x6A8035878E844595ULL, "hlac_ht_keep_3732_9z.dds"},
	{0x6A803B1504C5BA99ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden8.msh"},
	{0x6A80451504C5CB87ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden8.mmh"},
	{0x6A80BE881735779CULL, "vxht_keep_n_4g3_0_0.vlm"},
	{0x6A80C35B8306DEE7ULL, "vxlt_warehouse_13l_0_0.vlm"},
	{0x6A81011F86331E7EULL, "11666844.ogg"},
	{0x6A817B8D5D0962E4ULL, "804516383.ogg"},
	{0x6A81A613C755A8CDULL, "dre000_sibling_has_blight6036538.cl"},
	{0x6A81BD6D50A09B9AULL, "vxw_encounter2_camp_db_0_0.vlm"},
	{0x6A822CBCCD96BB4BULL, "6121572_f.fxe"},
	{0x6A82650BA162D55AULL, "166066296.ogg"},
	{0x6A828AC8E9B60562ULL, "vxlt_hangedman_n_1ve_0_0.vlm"},
	{0x6A82BC97C1BB2098ULL, "hlac_w_sundermount_262146_0x.dds"},
	{0x6A82E370A548F659ULL, "vxht_0_hightown_c7f_0_0.vlm"},
	{0x6A8348B83B978717ULL, "hlac_lt_darktown_-1_27x.dds"},
	{0x6A83629BBD1620D7ULL, "829569635.ogg"},
	{0x6A8381FBC3DC4AD2ULL, "vxht_brothel_n_4d6_0_1.vlm"},
	{0x6A83AD232321E63DULL, "40020841.ogg"},
	{0x6A83B49C65DE1C9FULL, "llac_lt_0_lowtown_a3_n_c_65536_5.dds"},
	{0x6A8438E00BE02EEBULL, "vxw_sundermount_2oa_0_0.vlm"},
	{0x6A85B69B69D90228ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_and_walk.gad"},
	{0x6A862061EF2492EAULL, "mcr310_argument6075787.cl"},
	{0x6A86626AE5EF5E9AULL, "art\\characters\\creatures\\spider\\animation\\crnf.ani"},
	{0x6A86C25A8B92F6EFULL, "187796586.ogg"},
	{0x6A871A20E30FEEA7ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_aoe_p.anb"},
	{0x6A87408D12FF364CULL, "vxw_bonepit_16i_0_0.vlm"},
	{0x6A8765E6911657E4ULL, "gen00fl_varric6076052.cl"},
	{0x6A877DFB4CB23620ULL, "vxht_chantry_n_352_0_0.vlm"},
	{0x6A87B86C5D2265D4ULL, "art\\characters\\creatures\\spider\\animation\\die_lp_v1.ani"},
	{0x6A87E907799855EEULL, "vxlt_0_lowtown_ex2_0_0.vlm"},
	{0x6A883B0C83E78E85ULL, "6015151_m.fxe"},
	{0x6A88528DC525A309ULL, "230423104.ogg"},
	{0x6A887514AB4520BBULL, "177916156.ogg"},
	{0x6A88D1DF5C596D32ULL, "vxca_caverns2_985_0_0.vlm"},
	{0x6A891CFCA774BA2AULL, "gen00fl_isabela6080446.cl"},
	{0x6A89ED5968B2812CULL, "llac_lt_undercityencounter_-1_44.dds"},
	{0x6A8A1B463DEE9E7BULL, "vxht_0_hightown_8pn_0_0.vlm"},
	{0x6A8A1E5637AD7C50ULL, "vxlt_gallowstemplar_f_5k1_0_0.vlm"},
	{0x6A8A65FE64A02E5CULL, "458035775.ogg"},
	{0x6A8A8E77F2D66E8DULL, "mag360_samson6042049.cl"},
	{0x6A8B1EA826451269ULL, "vxca_caverns2_7vv_0_0.vlm"},
	{0x6A8B32BAC8F27644ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basic.msh"},
	{0x6A8B44BAC8F294ADULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basic.mao"},
	{0x6A8B48BAC8F29BE6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basic.mmh"},
	{0x6A8B7508FAE7CD21ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.impa.ani"},
	{0x6A8BB77DDA917D13ULL, "6040829_m.fxe"},
	{0x6A8BDF92AEDFF1C0ULL, "vxw_woundedcoast_e3_0_1.vlm"},
	{0x6A8BFE64137E8BD7ULL, "guiquad.mmh"},
	{0x6A8C012F0E1F6FFEULL, "hlac_ca_caverns1_mines_24422_6x.dds"},
	{0x6A8C1464137EB149ULL, "guiquad.msh"},
	{0x6A8C3899D30FD33FULL, "746416711.ogg"},
	{0x6A8C821885BAFC6BULL, "llac_w_encounter3_0_6.dds"},
	{0x6A8C93557AB56AFCULL, "vxlt_0_lowtown_a2_byd_0_0.vlm"},
	{0x6A8D0171CDA6FC4EULL, "vxca_cave_2iv_0_0.vlm"},
	{0x6A8D0408D9470076ULL, "gen00fl_isabela6088244.cl"},
	{0x6A8D51F498F54C54ULL, "dae300_alistair_king6134957.cl"},
	{0x6A8D53B1B2C95B5AULL, "vxlt_darktown_67k_0_0.vlm"},
	{0x6A8D53B1D01D178AULL, "761335775.ogg"},
	{0x6A8DAFF35C7A7B19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_2.ani"},
	{0x6A8DD16EAEBE8B7CULL, "vxlt_gallowscourtyard_7c6_0_0.vlm"},
	{0x6A8DD1AE8F81D84BULL, "vxht_mansion2_2a3_0_0.vlm"},
	{0x6A8E2802EE5C3F31ULL, "mrl310_merrill6120014.cl"},
	{0x6A8E2E20B4B56D35ULL, "vxlt_0_lowtown_csf_0_0.vlm"},
	{0x6A8E3C03EDE43340ULL, "gen00fl_anders6112910.cl"},
	{0x6A8E6E28029F384AULL, "hlac_lt_0_lowtown_a2_1_4x.dds"},
	{0x6A8EA62B2F49B476ULL, "isa311_isabela6092294.cl"},
	{0x6A8EB56C5D287F36ULL, "art\\characters\\creatures\\spider\\animation\\die_lp_v1.gad"},
	{0x6A8EC0B659700BC9ULL, "6126796_m.fxe"},
	{0x6A8EC30161961FB9ULL, "art\\vfx\\materials\\vfx_d_forcevert_add.mao"},
	{0x6A8EF110C4839B3AULL, "6180948_m.fxe"},
	{0x6A8F0EC948D23B5EULL, "character_character_p144.psh.fxo.dx11.assembly"},
	{0x6A8F3065092B901CULL, "vxht_keep_n_4b0_0_0.vlm"},
	{0x6A8F3CE50247C7A5ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.mmh"},
	{0x6A8FF8E57F44220FULL, "6106676_f.fxe"},
	{0x6A90443DD6460B75ULL, "6019212_m.fxe"},
	{0x6A90869E9D6C10F4ULL, "521059196.ogg"},
	{0x6A90E813A62D4CAEULL, "vxlt_0_lowtown_a2_n_f59_0_0.vlm"},
	{0x6A90FCA5DBAC1DE2ULL, "vxlt_hangedman_n_1sf_0_0.vlm"},
	{0x6A916566A76D7D1DULL, "737239455.ogg"},
	{0x6A920A28320545B0ULL, "vxlt_ruinedhouse_n_ka_0_0.vlm"},
	{0x6A9228725F896085ULL, "vrc221_bartrand6043067.cl"},
	{0x6A92291242229E0FULL, "6060746_m.fxe"},
	{0x6A9235078B5F0E5BULL, "art\\characters\\playercharacter\\humanmale\\rob_chaa.mmh"},
	{0x6A9241078B5F22F8ULL, "art\\characters\\playercharacter\\humanmale\\rob_chaa.mao"},
	{0x6A9253078B5F4145ULL, "art\\characters\\playercharacter\\humanmale\\rob_chaa.msh"},
	{0x6A9255D87FF53DD6ULL, "vxht_keep_n_4ac_0_0.vlm"},
	{0x6A928008FAEDFEC3ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.impa.gad"},
	{0x6A92828E0386B290ULL, "zz_mag_debug6040691.cl"},
	{0x6A92953E3053494CULL, "vxw_woundedcoast_7d_0_0.vlm"},
	{0x6A92959D0EC022E2ULL, "6048194_m.fxe"},
	{0x6A92EDBB4D5C8DC8ULL, "6016808_f.fxe"},
	{0x6A93F91D2F2BA0F5ULL, "dae000_epic_merchant_16179432.cl"},
	{0x6A9465565265E87AULL, "6087520_m.fxe"},
	{0x6A948AF35C805B5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_2.gad"},
	{0x6A953F259680C99CULL, "vxw_bonepitdestroyed_a3_1v9_0_0.vlm"},
	{0x6A953FBAC8FAE0B1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basic.phy"},
	{0x6A95AB165CF088D8ULL, "510804042.ogg"},
	{0x6A95E72FF598DE3EULL, "vxlt_backalley_g_1r4_0_1.vlm"},
	{0x6A95F32A3FEC7E32ULL, "isa311_hawke_comments6117915.cl"},
	{0x6A96077E3C6B7162ULL, "vxht_mansion_n_3yt_0_0.vlm"},
	{0x6A962C9EB3EA2081ULL, "sky_ht_chantry_n.mao"},
	{0x6A965AE46B8E6E6EULL, "vxlt_0_lowtown_es3_0_0.vlm"},
	{0x6A96732F1B4B637DULL, "follower_banter6156095.cl"},
	{0x6A96F6F1196A35DDULL, "869926607.ogg"},
	{0x6A978FDF61532809ULL, "68998344.ogg"},
	{0x6A979AAE836CD8A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_enter.gad"},
	{0x6A97DB5CD42D0520ULL, "6087283_m.fxe"},
	{0x6A97FD7B50F19162ULL, "llac_lt_0_lowtown_a2_n_131071_7.dds"},
	{0x6A982C68419D9B5EULL, "pro000_meeran6048804.cl"},
	{0x6A98C41537A904BEULL, "20967631.ogg"},
	{0x6A98E247D7230B52ULL, "79628691.ogg"},
	{0x6A98E4EB89986768ULL, "vxht_playershouse_n_1uu_0_0.vlm"},
	{0x6A9925DB6AB75B47ULL, "159864547.ogg"},
	{0x6A995D08FAF3E1A8ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.impa.evt"},
	{0x6A998FB92B57B1E8ULL, "zz_prmpt_debug6207827.cl"},
	{0x6A9A0246C16B19F8ULL, "6018711_m.fxe"},
	{0x6A9A03CF76BC5659ULL, "vxlt_gallowstemplar_45h_0_0.vlm"},
	{0x6A9A7CCD607CA812ULL, "6015938_m.fxe"},
	{0x6A9AA46D280B0947ULL, "lgt000ip_finger_sack.utp"},
	{0x6A9B4A23E6CB01D3ULL, "art\\vfx\\combat\\vfx_w_off_accuracy_activ_c.impa.ani"},
	{0x6A9B73296F510D00ULL, "vxht_brothel_n_6fi_0_0.vlm"},
	{0x6A9B97B2EA761D2DULL, "6175950_m.fxe"},
	{0x6A9BA7F35C86AB40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_2.evt"},
	{0x6A9BBBE8E005A73CULL, "vxlt_hangedman_10j_0_0.vlm"},
	{0x6A9C39E502524D72ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.anb"},
	{0x6A9C438B8177F0CBULL, "vxht_mansion2_2j2_0_0.vlm"},
	{0x6A9C6702439FB5D5ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls102a.msh"},
	{0x6A9C6902439FB94BULL, "art\\characters\\weapons\\longswords\\w_lsw_ls102a.mmh"},
	{0x6A9D3B9F8EAA87FBULL, "6156292_m.fxe"},
	{0x6A9D8E6F1F7C58F8ULL, "385310312.ogg"},
	{0x6A9DA82B80D77596ULL, "6004198_m.fxe"},
	{0x6A9DC7C1177ACF00ULL, "395205956.ogg"},
	{0x6A9E51A42B9E5F2DULL, "844182127.ogg"},
	{0x6A9E8FAE8372E4B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_enter.evt"},
	{0x6A9E97C256B6316EULL, "384161898.ogg"},
	{0x6A9F1A359B2ECF2EULL, "vxlt_gallowstemplar_f_2p1_0_0.vlm"},
	{0x6A9F8012240C1BC8ULL, "6042495_m.fxe"},
	{0x6AA0B599F6D5C630ULL, "vxht_0_hightown_a3_9ur_0_0.vlm"},
	{0x6AA11C99EFCC5FEBULL, "vxw_sundermount_2uc_0_0.vlm"},
	{0x6AA17BB01100EE10ULL, "6110158_m.fxe"},
	{0x6AA1BE9228A6D4EAULL, "framebuffereffects_dofblur_p0.psh.fxo.dx11.assembly"},
	{0x6AA24FF378A94137ULL, "vxlt_darktown_2r1_0_0.vlm"},
	{0x6AA250EE6B11A449ULL, "one120_athenril6088916.cl"},
	{0x6AA27E24BEB3D43EULL, "demoral_05.mao"},
	{0x6AA290917ABE0CAEULL, "and111_refugee_barks.cnv"},
	{0x6AA2DF93C9CF479EULL, "vxca_cavern1_haunted_jdd_0_0.vlm"},
	{0x6AA32BA0DCE39A7EULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs104a_0t.dds"},
	{0x6AA32D8ADDDEBD8CULL, "6087140_m.fxe"},
	{0x6AA377D26EC91BCBULL, "vxca_cavern1_haunted_iyi_0_0.vlm"},
	{0x6AA37B0559599363ULL, "vxca_caverns1_jgm_0_0.vlm"},
	{0x6AA39897EB2C87FBULL, "6191254_f.fxe"},
	{0x6AA421AA35C36055ULL, "plt_lgt000pt_ultra_3_04.nss"},
	{0x6AA4242E47BC19DDULL, "mag131_keran6086287.cl"},
	{0x6AA4483003F50006ULL, "887742688.ogg"},
	{0x6AA46813D9980003ULL, "hlac_lt_0_lowtown_docks_n_-4_4y.dds"},
	{0x6AA4D1C9F9371F6DULL, "vxlt_0_lowtown_ewx_0_0.vlm"},
	{0x6AA58F6805585B30ULL, "369739940.ogg"},
	{0x6AA5B2917AC07C1EULL, "and111_refugee_barks.bnk"},
	{0x6AA5D595584DD681ULL, "892077767.ogg"},
	{0x6AA6368AB89ECC5EULL, "611179067.ogg"},
	{0x6AA67BE74414E779ULL, "character_character_p66.psh.fxo.dx11.assembly"},
	{0x6AA69F7544E7266EULL, "336085579.ogg"},
	{0x6AA6D21E8C5A43F1ULL, "vxw_woundedcoast_hx_0_0.vlm"},
	{0x6AA8059F612E5635ULL, "dae000_flemeth6013377.cl"},
	{0x6AA84223E6D57E0AULL, "art\\vfx\\combat\\vfx_w_off_accuracy_activ_c.impa.evt"},
	{0x6AA89B9CD7D4C5C6ULL, "6208897_m.fxe"},
	{0x6AA8B25E3D4C6AD8ULL, "qun221_petrice6076665.cl"},
	{0x6AA8B489E10E0829ULL, "6017815_f.fxe"},
	{0x6AA90AB7A535BCC3ULL, "ico_helm_heavy_4.dds"},
	{0x6AA981793D5637E3ULL, "llac_ca_caverns2_5802_57.dds"},
	{0x6AA9ABAFBD0BD8E4ULL, "vxlt_gallowstemplar_n_6sj_0_0.vlm"},
	{0x6AA9BD9ADEAE4E2CULL, "cod_hst_antiva.utp"},
	{0x6AA9D1494DFB41BFULL, "vxlt_gallowstemplar_4qr_0_0.vlm"},
	{0x6AA9DBC92C1E03E1ULL, "94381958.ogg"},
	{0x6AAA093CAB02182AULL, "gen_im_npc_arm_hum_boo_slva.uti"},
	{0x6AAAB52892AABC7CULL, "vxlt_gallowscourtyard_7i4_0_0.vlm"},
	{0x6AAB64A38173E7F6ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyn_0n.dds"},
	{0x6AAC87AE837F14EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_enter.ani"},
	{0x6AACB3955454A602ULL, "346429208.ogg"},
	{0x6AACB3D1F61618EEULL, "vxca_cave_1b0_0_0.vlm"},
	{0x6AACC550ADD7A120ULL, "qcr000_arishok6084349.cl"},
	{0x6AAD44B56B091D56ULL, "mag311_keran6041706.cl"},
	{0x6AAD9B63459A283EULL, "6143515_f.fxe"},
	{0x6AADE127C8F7ECC2ULL, "404937599.ogg"},
	{0x6AADFEB9965D0697ULL, "vxca_cavern1_haunted_imq_0_0.vlm"},
	{0x6AAE19E5EB78AB45ULL, "vxca_cavern1_haunted_drc_0_0.vlm"},
	{0x6AAE38173D2DDFD4ULL, "340754779.ogg"},
	{0x6AAF2123E6DB657DULL, "art\\vfx\\combat\\vfx_w_off_accuracy_activ_c.impa.gad"},
	{0x6AAF99987EECF451ULL, "w_woundedcoast_68_0.mmh"},
	{0x6AAFA39EDF7958E8ULL, "hlac_w_blightlands_-131069_0x.dds"},
	{0x6AAFA5987EED08F2ULL, "w_woundedcoast_68_0.mao"},
	{0x6AAFB3987EED200FULL, "w_woundedcoast_68_0.msh"},
	{0x6AB01720E332EEA8ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_aoe_p.mmh"},
	{0x6AB020924CD61683ULL, "6041127_m.fxe"},
	{0x6AB0918C7D612156ULL, "7812015.ogg"},
	{0x6AB09C9CE3BA75ACULL, "hlac_w_encounter3_-65535_0z.dds"},
	{0x6AB0BDA79A269A3FULL, "dae000ip_crowd_mf6_talk18.utp"},
	{0x6AB11D72108366F6ULL, "ave121_brennan6105827.cl"},
	{0x6AB122107C91EFD2ULL, "zz_bet_debug6113401.cl"},
	{0x6AB1CB994B295134ULL, "894154636.ogg"},
	{0x6AB26019C509FCB2ULL, "vxlt_0_lowtown_a3_n_c_ded_0_0.vlm"},
	{0x6AB272F44ED780D8ULL, "dre000_cut_rock_wraith6150004.cl"},
	{0x6AB2AA3932AF6F10ULL, "art\\vfx\\combat\\vfx_glb_ele_p_minor_c.mmh"},
	{0x6AB2C2432879C155ULL, "6177829_f.fxe"},
	{0x6AB33E9E4F7A9F6EULL, "vxlt_0_lowtown_ei5_0_0.vlm"},
	{0x6AB343A6EB2D37EDULL, "vxlt_0_lowtown_erw_0_0.vlm"},
	{0x6AB3B101C8672976ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_rnf.ani"},
	{0x6AB3C63EC26C6A9CULL, "6084013_m.fxe"},
	{0x6AB41EB4484AEE99ULL, "140150217.ogg"},
	{0x6AB42AEBEB6C685FULL, "350774144.ogg"},
	{0x6AB443800AD8F09BULL, "qun221cr_petrice_mob4.utc"},
	{0x6AB44766D8595A53ULL, "6142046_m.fxe"},
	{0x6AB53AC765CDCEEEULL, "vxht_chantry_n_2ye_0_0.vlm"},
	{0x6AB5584C303687EFULL, "rock_large_specular.dds"},
	{0x6AB5AFF70D4F73FCULL, "t3_cor_h02.tnt"},
	{0x6AB61297541F926EULL, "vxca_cavern1_haunted_ehw_0_0.vlm"},
	{0x6AB6B19C5C148956ULL, "vxca_cavern1_haunted_i1x_0_0.vlm"},
	{0x6AB6DD42CCAE54C4ULL, "6119971_m.fxe"},
	{0x6AB715D94CEA1C50ULL, "185322522.ogg"},
	{0x6AB7776CE37D251FULL, "823835781.ogg"},
	{0x6AB7C655E9E259E6ULL, "hlac_lt_gallowstemplar_n_2393_0z.dds"},
	{0x6AB7E7E0AA3A4D6BULL, "lt_0_lowtown_a2_25_0_ll.mmh"},
	{0x6AB803D40B0CA9D8ULL, "434681552.ogg"},
	{0x6AB81FC0E3856886ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic.phy"},
	{0x6AB836F23F4D6D97ULL, "6115878_f.fxe"},
	{0x6AB8C7246E47CCCDULL, "qun221_petrice6083653.cl"},
	{0x6AB8CB872936F211ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_1_sylvesta.gad"},
	{0x6AB8DF7F6A01C01DULL, "vxca_caverns1_mines_imc_0_0.vlm"},
	{0x6ABA125E11C5D768ULL, "dae000cr_grand_cleric6173592.cl"},
	{0x6ABAE740CE1FE39BULL, "vxht_mansion2_1pn_0_1.vlm"},
	{0x6ABB8E44E85D65A6ULL, "58721810.ogg"},
	{0x6ABBA0290D778CACULL, "llac_ca_caverns2prc_5802_57.dds"},
	{0x6ABBEF14DCB63AAFULL, "cd_mag311_huon.bnk"},
	{0x6ABC20BA638FBA9CULL, "vxht_chantry_358_0_1.vlm"},
	{0x6ABC709688531F17ULL, "vxca_cavern1_haunted_ijp_0_0.vlm"},
	{0x6ABCDF711C55203FULL, "665484037.ogg"},
	{0x6ABCE94636E5B785ULL, "rock_c_03.dds"},
	{0x6ABD03D74ECF9411ULL, "mrl250cr_frosthorror.utc"},
	{0x6ABD6007375BCFB9ULL, "6163535_m.fxe"},
	{0x6ABE27B1E6F8BFCFULL, "guardian_bodyguard.dds"},
	{0x6ABE3B5F2771D06BULL, "hlac_lt_warehouse_0_4z.dds"},
	{0x6ABE7A41E7120323ULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.cess.ani"},
	{0x6ABF33AD5C957237ULL, "vxlt_darktown_2x3_0_0.vlm"},
	{0x6ABF919741252DB7ULL, "gen00fl_varric6087471.cl"},
	{0x6ABFB3625FBA6713ULL, "vxw_woundedcoast_1cb_0_0.vlm"},
	{0x6ABFCC87293D12BFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_1_sylvesta.ani"},
	{0x6ABFE50A4F473B1FULL, "mag370_meredith6092167.cl"},
	{0x6AC0310DF54710F1ULL, "284757172.ogg"},
	{0x6AC05CBACB728205ULL, "dae000_merchant_tailor_hm6164348.cl"},
	{0x6AC0B901C871C167ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_rnf.evt"},
	{0x6AC0CD3932BBE82FULL, "art\\vfx\\combat\\vfx_glb_ele_p_minor_c.anb"},
	{0x6AC0EF7CBDCCC73AULL, "18627091.ogg"},
	{0x6AC0FB47F08ED157ULL, "mer240_jansen6050383.cl"},
	{0x6AC0FD1641B00E68ULL, "demo_weapon_swap6207144.cl"},
	{0x6AC19D09AF44D10CULL, "6080667_f.fxe"},
	{0x6AC1B583DD23506DULL, "vxlt_0_lowtown_emv_0_0.vlm"},
	{0x6AC1DC1B73E51FB9ULL, "vxca_caverns1_ede_0_0.vlm"},
	{0x6AC27091C1C7494DULL, "703984604.ogg"},
	{0x6AC2A926EA52F385ULL, "vxlt_blackemporium_3s_0_0.vlm"},
	{0x6AC2D4B3ACBDA6A7ULL, "pro000_guard6013437.cl"},
	{0x6AC3079328B28C68ULL, "497865508.ogg"},
	{0x6AC30ACF1E6DDB02ULL, "prm000im_isaac_boots.uti"},
	{0x6AC322BFFE4C309CULL, "one100dg_nabil6054978.cl"},
	{0x6AC3ACA457C3E76EULL, "vxht_chantry_n_2tf_0_0.vlm"},
	{0x6AC3EDA5C907B779ULL, "vxca_caverns1_mines_k2n_0_2.vlm"},
	{0x6AC43B572A2AAFC3ULL, "vxlt_0_lowtown_e2i_0_0.vlm"},
	{0x6AC487414D336167ULL, "hlac_w_encounter2_262146_0z.dds"},
	{0x6AC4AF2449678EF0ULL, "6168979_m.fxe"},
	{0x6AC53C1B53E69DA6ULL, "gen00fl_anders6215196.cl"},
	{0x6AC5B4A4FFDBE031ULL, "6080492_m.fxe"},
	{0x6AC5FC860DB88DC1ULL, "vxlt_0_lowtown_n_ex6_0_0.vlm"},
	{0x6AC5FF6C95CFC22DULL, "vxlt_gallowstemplar_2r7_0_0.vlm"},
	{0x6AC620AA35E044B6ULL, "plt_lgt000pt_ultra_3_04.plo"},
	{0x6AC65275E73C6757ULL, "6178515_m.fxe"},
	{0x6AC66B4E78BA5C22ULL, "hlac_ca_caverns1_mines_24423_12z.dds"},
	{0x6AC6F6CAAD507D44ULL, "249995626.ogg"},
	{0x6AC7515C5BF7D89DULL, "vxca_caverns1_mines_ijb_0_0.vlm"},
	{0x6AC7C9DBB5A32713ULL, "772366622.ogg"},
	{0x6AC7DA01C87817B0ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90l_rnf.gad"},
	{0x6AC7F2C64C85BED7ULL, "917325648.ogg"},
	{0x6AC8018A78B8A364ULL, "follower_banter6163478.cl"},
	{0x6AC83B6F39AFC9A3ULL, "6030682_m.fxe"},
	{0x6AC8D1AE89A60904ULL, "vxca_caverns2prc_3td_0_0.vlm"},
	{0x6AC93300CE999CF5ULL, "llac_ca_caverns1_mines_24421_0.dds"},
	{0x6AC99C9475B4E26DULL, "vxca_cave_1az_0_0.vlm"},
	{0x6ACA131EB92AF8E7ULL, "499182202.ogg"},
	{0x6ACA28C64112C134ULL, "mcr370_messenger6071817.cl"},
	{0x6ACA5C43E2E52FA9ULL, "vxlt_0_lowtown_gby_0_0.vlm"},
	{0x6ACA644E6C0A666FULL, "vxht_mansion2b_28h_0_0.vlm"},
	{0x6ACA8CE4A53EB24FULL, "903628711.ogg"},
	{0x6ACAE2737A493797ULL, "vxca_cavern1_haunted_igs_0_0.vlm"},
	{0x6ACBD8CE7CC0F4C9ULL, "6140183_m.fxe"},
	{0x6ACBF543B1AC7064ULL, "vxht_keep_3b8_0_0.vlm"},
	{0x6ACC1AF0A478AC41ULL, "llac_ht_0_hightown_-65535_1.dds"},
	{0x6ACC440D42A44118ULL, "xx_extfloor.dds"},
	{0x6ACC8ECD69E73C64ULL, "6104289_m.fxe"},
	{0x6ACC8ED3E20A1127ULL, "6068874_m.fxe"},
	{0x6ACC9241E71E691AULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.cess.evt"},
	{0x6ACCABFEA201553EULL, "vxht_keep_64a_0_0.vlm"},
	{0x6ACCBE767531F084ULL, "6160881_m.fxe"},
	{0x6ACCE7AA4A7D941DULL, "6120008_m.fxe"},
	{0x6ACD2BDCE7AAD3C2ULL, "hlac_lt_0_lowtown_docks_a3_-65539_6x.dds"},
	{0x6ACD37DE4FAF8FB3ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.dura.ani"},
	{0x6ACE253F51B07F93ULL, "vxw_woundedcoast_1ne_0_0.vlm"},
	{0x6ACE89C974AE2625ULL, "vxw_encounter2_gh_0_0.vlm"},
	{0x6ACEB8B65B356C0EULL, "lgt310_mis_burn_body6180584.cl"},
	{0x6ACED7909F59E02BULL, "hlac_ht_0_hightown_n_131070_0y.dds"},
	{0x6ACF62F796C8C115ULL, "6108060_m.fxe"},
	{0x6ACFA5BB414EB198ULL, "vxht_mansion2_3ee_0_0.vlm"},
	{0x6AD011AE0DD2A0E5ULL, "316864841.ogg"},
	{0x6AD02760CF1968EDULL, "vxlt_0_lowtown_ehu_0_0.vlm"},
	{0x6AD033BF6861833FULL, "framebuffereffects_linearfade_p0.psh.fxo.dx11"},
	{0x6AD07DCF3552EDEEULL, "6064840_m.fxe"},
	{0x6AD097708D9E32B7ULL, "t3_har_dbr.tnt"},
	{0x6AD0F44C75FD5B5FULL, "6189047_m.fxe"},
	{0x6AD152A2998986BAULL, "668149037.ogg"},
	{0x6AD1B73C840851FCULL, "vxw_bonepitdestroyed_a3_1p3_0_0.vlm"},
	{0x6AD2019547808E35ULL, "6016908_f.fxe"},
	{0x6AD21E2F237CDE70ULL, "538415107.ogg"},
	{0x6AD2258C4B8DCF1CULL, "6108744_m.fxe"},
	{0x6AD22B67F6F77902ULL, "llac_lt_0_lowtown_docks_a3_n_65535_2.dds"},
	{0x6AD25830145D2B04ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowera.mmh"},
	{0x6AD25C30145D3257ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowera.mao"},
	{0x6AD26E30145D50A6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowera.msh"},
	{0x6AD2748DCEF59EF3ULL, "vxht_bartrand_n_730_0_0.vlm"},
	{0x6AD2A01E7F1A5FDDULL, "hlac_lt_0_lowtown_docks_a3_n_65534_2z.dds"},
	{0x6AD2E3E8DF0C316FULL, "480782721.ogg"},
	{0x6AD319AD04DA9D36ULL, "vxca_cavern1_haunted_jg9_0_0.vlm"},
	{0x6AD34619CEFE0A6EULL, "s_ballista_dust_l2.mao"},
	{0x6AD36AF88607709BULL, "mag100_karras6035307.cl"},
	{0x6AD39141E724860DULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.cess.gad"},
	{0x6AD473DC016E9096ULL, "106512661.ogg"},
	{0x6AD47862D41C0C63ULL, "vxw_encounter2_camp_jq_0_0.vlm"},
	{0x6AD4C452D9327FD7ULL, "6173415_m.fxe"},
	{0x6AD4D79A840D61D7ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.cess.gad"},
	{0x6AD4F187EA7E2484ULL, "365411033.ogg"},
	{0x6AD542DD0D3D05CBULL, "hlac_lt_0_lowtown_docks_a3_-131074_1z.dds"},
	{0x6AD544E136DE92F1ULL, "vxw_bonepit_1sg_0_0.vlm"},
	{0x6AD56C6A41957C55ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_templdesk_n.dds"},
	{0x6AD5891DCA47355CULL, "zz_lgt_1_debug6114389.cl"},
	{0x6AD5DEB6F51E783EULL, "587017004.ogg"},
	{0x6AD627ED50110320ULL, "hlac_ca_cavern1_haunted_24424_15x.dds"},
	{0x6AD6567D92A54BEEULL, "qcr270_fake_meredith6066358.cl"},
	{0x6AD6809F18B39DA4ULL, "gen00fl_bethany6109220.cl"},
	{0x6AD73F00EECF155BULL, "6175307_m.fxe"},
	{0x6AD7507B127E389EULL, "hlac_lt_0_lowtown_a2_n_c_65536_10x.dds"},
	{0x6AD7A4307F75FFA2ULL, "vxlt_hangedman_2i4_0_0.vlm"},
	{0x6AD842EA80C81080ULL, "follower_banter6172991.cl"},
	{0x6AD88A3842A9FF84ULL, "6025085_m.fxe"},
	{0x6AD89FE9A559CDFAULL, "6161202_m.fxe"},
	{0x6AD8B4D930D7D1EDULL, "zz_followers_banter6159687.cl"},
	{0x6AD8D62B5E007EEFULL, "vxht_mansion2b_25k_0_0.vlm"},
	{0x6AD8DE4913368A38ULL, "hlac_ca_cavern1_haunted_24427_5z.dds"},
	{0x6AD975454867CB80ULL, "vxw_sundermount_27f_0_0.vlm"},
	{0x6ADA1BE7B23E13F7ULL, "vrc321_party_barks6096312.cl"},
	{0x6ADA56078B9CFB90ULL, "art\\characters\\playercharacter\\humanmale\\rob_chaa.phy"},
	{0x6ADA929DEACC4C6DULL, "art\\levels\\ztools\\default\\generic\\9ff_bar.msh"},
	{0x6ADAA09DEACC63A0ULL, "art\\levels\\ztools\\default\\generic\\9ff_bar.mao"},
	{0x6ADAA49DEACC6AF3ULL, "art\\levels\\ztools\\default\\generic\\9ff_bar.mmh"},
	{0x6ADAAC104689F881ULL, "hlac_lt_undercityserialkiller_-1_17y.dds"},
	{0x6ADAD165939B32B7ULL, "120890579.ogg"},
	{0x6ADAFE4E768EB5C0ULL, "follower_banter6141711.cl"},
	{0x6ADB1DDB93F76DBEULL, "vxht_keep_61b_0_0.vlm"},
	{0x6ADB2FDE4FBBBFEAULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.dura.evt"},
	{0x6ADB335991B744DFULL, "6057317_m.fxe"},
	{0x6ADB52107A6D8093ULL, "6121687_m.fxe"},
	{0x6ADBBE8F830B776BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_exit.ani"},
	{0x6ADBEC9A8413A464ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.cess.evt"},
	{0x6ADC7EF41B53748DULL, "llac_dr_0_deeproads_5845_16.dds"},
	{0x6ADC971C43A69813ULL, "vxw_woundedcoast_1id_0_0.vlm"},
	{0x6ADD5665B4A5D05AULL, "6108126_m.fxe"},
	{0x6ADD7876504F441CULL, "vxw_bonepitdestroyed_a3_1o2_0_0.vlm"},
	{0x6ADD959A13B15BCAULL, "ser221_gascard6122462.cl"},
	{0x6ADE136E5ECC5001ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6_tool.phy"},
	{0x6ADE5250A13DF93AULL, "vxw_woundedcoast_mp_0_0.vlm"},
	{0x6ADE93B29A384F2AULL, "vxht_0_hightown_a3_a2q_0_0.vlm"},
	{0x6ADEB88810F4C942ULL, "seb320ip_at_keep.utp"},
	{0x6ADEEE50991629B4ULL, "llac_lt_0_lowtown_n_-196607_0.dds"},
	{0x6ADEF30B6CF43431ULL, "hlac_ca_caverns2_5802_63y.dds"},
	{0x6ADF95EBEAA7DB6EULL, "527217963.ogg"},
	{0x6ADFAE6791899B41ULL, "qun231_varnell6077277.cl"},
	{0x6ADFD33D4E709E13ULL, "vxlt_0_lowtown_a3_n_c_ekg_0_1.vlm"},
	{0x6ADFF1880142F384ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\rearup.ani"},
	{0x6AE07E0CAA100BA7ULL, "547144547.ogg"},
	{0x6AE098E2D960FA17ULL, "vxca_caverns2prc_9xe_0_0.vlm"},
	{0x6AE0A912CFDDD702ULL, "one100dg_vanard6020937.cl"},
	{0x6AE0BBD3D1BFEBBEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1_lp.ani"},
	{0x6AE1A2D4A876A100ULL, "hlac_lt_0_lowtown_n_-131072_0y.dds"},
	{0x6AE1A95D80A006C0ULL, "hlac_lt_shop_-1_5y.dds"},
	{0x6AE20EDE4FC1A65DULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.dura.gad"},
	{0x6AE2722D0E8D3943ULL, "385809244.ogg"},
	{0x6AE2EA3FC61224E3ULL, "vxw_encounter2_camp_ep_0_0.vlm"},
	{0x6AE3778995A4D6F2ULL, "zz_dae_debug6047604.cl"},
	{0x6AE3C89410133CA9ULL, "414800424.ogg"},
	{0x6AE51F16493181B9ULL, "555055112.ogg"},
	{0x6AE55656C8531C51ULL, "631533055.ogg"},
	{0x6AE5D0E4A64A056DULL, "hlac_dr_slavers_3216_6y.dds"},
	{0x6AE5F9EDD424701CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.ani"},
	{0x6AE6108F7C405AC4ULL, "6126739_m.fxe"},
	{0x6AE6546CA424236BULL, "dae000ip_crowd_ff5_talk1.utp"},
	{0x6AE67BCDF491AE4CULL, "character_character_p38.psh.fxo.dx11.assembly"},
	{0x6AE6EE8801490DE6ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\rearup.gad"},
	{0x6AE7659511CF39DDULL, "one100dg_lia6060998.cl"},
	{0x6AE81735F70FEF92ULL, "vxca_caverns1_mines_iit_0_0.vlm"},
	{0x6AE84B18581EBA1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ab_transition.ani"},
	{0x6AE8A8EC40CD67CFULL, "22655049.ogg"},
	{0x6AE8AC6EFA8CE6EFULL, "gencr_gate_slave_r2.utc"},
	{0x6AE8C7ADE5485076ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hanginga.msh"},
	{0x6AE8D1ADE5486174ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hanginga.mmh"},
	{0x6AE8D5ADE5486787ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hanginga.mao"},
	{0x6AE8DBEB437B149AULL, "gen00fl_anders6111391.cl"},
	{0x6AE932786913EB23ULL, "mcr371_post_harvester6083676.cl"},
	{0x6AE954621DF57FE0ULL, "6110460_m.fxe"},
	{0x6AE991711C6A9DA8ULL, "593631519.ogg"},
	{0x6AE9A71E22714F83ULL, "hlac_lt_0_lowtown_n_196607_1y.dds"},
	{0x6AE9B1C02AA8804CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_pc.ani"},
	{0x6AEA049A84200A5DULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.cess.ani"},
	{0x6AEA5B4DFE776D3DULL, "hlac_w_encounter2_-131070_0z.dds"},
	{0x6AEA974FE1971AD8ULL, "557667798.ogg"},
	{0x6AEAC6BC44B66349ULL, "mag121_idunna6092342.cl"},
	{0x6AEAD6839904D5B0ULL, "zz_prmpt_debug6198259.cl"},
	{0x6AEB2F1115C9F130ULL, "vfx_ring01.dds"},
	{0x6AEB40FF3732D477ULL, "344183882.ogg"},
	{0x6AEB7B33FFFA1261ULL, "vxca_caverns2_61j_0_0.vlm"},
	{0x6AEBC1D6559D7F1BULL, "ico_maul_1.dds"},
	{0x6AEBF82B1AA08359ULL, "vxht_brothel_n_5g0_0_0.vlm"},
	{0x6AEBFF2E260FC089ULL, "hlac_lt_0_lowtown_a3_n_c_65535_3y.dds"},
	{0x6AEC522B679B8CC6ULL, "vxht_mansion2_2ng_0_0.vlm"},
	{0x6AEC6E0F680F53B6ULL, "6136695_m.fxe"},
	{0x6AECB89AEBDB3AD8ULL, "6052766_m.fxe"},
	{0x6AECE988014D70C5ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\rearup.evt"},
	{0x6AED06EDD42AA55EULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.gad"},
	{0x6AED4AE74554B64AULL, "hlac_lt_undercityencounter_-1_18y.dds"},
	{0x6AEDC9D247CAC2C9ULL, "697477276.ogg"},
	{0x6AEDE0EEE3E8E29AULL, "gen00fl_sebastian6189337.cl"},
	{0x6AEDE89D28831131ULL, "vxlt_warehouse_n_uh_0_0.vlm"},
	{0x6AEE2F0610995010ULL, "fxe_watervol_add.mao"},
	{0x6AEE55D0F8BED322ULL, "vxht_0_hightown_a2_bc8_0_0.vlm"},
	{0x6AEF05BA101A24E2ULL, "gen_im_trs_drk_junk1_03.uti"},
	{0x6AEF0FC6B90F0A56ULL, "vxw_bonepitdestroyed_a3_1mb_0_0.vlm"},
	{0x6AEF273BD44DEF34ULL, "vxlt_undercityserialkiller_242_0_0.vlm"},
	{0x6AEF4F5DF07691CEULL, "69411302.ogg"},
	{0x6AEFA58F831BF5F5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_exit.gad"},
	{0x6AF01BC08F1FB327ULL, "crack_ye.dds"},
	{0x6AF0AAE885EF0756ULL, "vxlt_0_lowtown_docks_b9o_0_0.vlm"},
	{0x6AF0BEC02AAEB48EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_pc.gad"},
	{0x6AF0F97A67DF9C6DULL, "360473055.ogg"},
	{0x6AF178A4BD8F20DFULL, "gen00fl_mrl_act3_gift2_ph.stg"},
	{0x6AF1BB1CA0FD2269ULL, "vxca_cavern1_haunted_ea7_0_0.vlm"},
	{0x6AF1D0E7F4BBE76AULL, "6167089_f.fxe"},
	{0x6AF1E73F0B27833AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc0col.mmh"},
	{0x6AF1EB3F0B278A71ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc0col.mao"},
	{0x6AF1F93F0B27A180ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterc0col.msh"},
	{0x6AF227C7CCF375CFULL, "6113978_m.fxe"},
	{0x6AF2707EE1CBFC68ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_chant2door.msh"},
	{0x6AF274798FCD30BFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_catacombs_1.mao"},
	{0x6AF2827EE1CC1AF9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_chant2door.mao"},
	{0x6AF28E7EE1CC2E92ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_chant2door.mmh"},
	{0x6AF29EB3ABEF8F06ULL, "mcr310_argument6072683.cl"},
	{0x6AF2A4D114A6F802ULL, "fex_ruin_arch_lrg_c.msh"},
	{0x6AF2A6D114A6FBB8ULL, "fex_ruin_arch_lrg_c.mmh"},
	{0x6AF3113964A9DC4FULL, "zz_approval_debug6003516.cl"},
	{0x6AF359459049EEC2ULL, "544148101.ogg"},
	{0x6AF38F237CE4ED7EULL, "333688759.ogg"},
	{0x6AF3B9DD3ADC3914ULL, "llac_lt_undercitydungeon_-1_45.dds"},
	{0x6AF3BB4DDB41EDF7ULL, "gen00fl_varric6107281.cl"},
	{0x6AF3D80AAA8AC26FULL, "lt_0_lowtown_docks_a3_26_0_lc.mmh"},
	{0x6AF3E1EDD430857DULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.evt"},
	{0x6AF45D82C52B0231ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drunk_loop.ani"},
	{0x6AF47C0767C831ABULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_paperlamp_e.dds"},
	{0x6AF49C0FFFE708A6ULL, "vxca_caverns1_mines_i1i_0_1.vlm"},
	{0x6AF4D4D3D1D0BF38ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1_lp.gad"},
	{0x6AF516E01433C891ULL, "vxdr_0_deeproads_eq1_0_0.vlm"},
	{0x6AF530BEF4DBBF8CULL, "442923015.ogg"},
	{0x6AF55BB8C7EA0152ULL, "977116939.ogg"},
	{0x6AF59BCE528746B0ULL, "hlac_ht_0_hightown_-2_0x.dds"},
	{0x6AF5CD52CD2AF6D8ULL, "follower_banter6155826.cl"},
	{0x6AF5F81D6FD4F529ULL, "mrl150_merrill_pt2.stg"},
	{0x6AF66318582B20EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ab_transition.evt"},
	{0x6AF66990665D036EULL, "vis110_commoner1.cnv"},
	{0x6AF690A60D8A8725ULL, "275748420.ogg"},
	{0x6AF69B9EDC9853FFULL, "w_blightlands_23_0.rml"},
	{0x6AF6B9C02AB3186DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_pc.evt"},
	{0x6AF6B9D6E0CFB650ULL, "vxht_0_hightown_a3_cba_0_0.vlm"},
	{0x6AF6ECE5EA2C02E6ULL, "dae000_worthy.cnv"},
	{0x6AF7075B26B2E7AFULL, "6177374_m.fxe"},
	{0x6AF72B7BD115F55EULL, "6109374_m.fxe"},
	{0x6AF85F66C4EB5C1AULL, "hlac_ht_brothel_n_2995_4x.dds"},
	{0x6AF8A8DE91686BDDULL, "hlac_w_encounter2_-65538_0y.dds"},
	{0x6AF8AB1B79883DCEULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs104a_0d.dds"},
	{0x6AF9774A8907DBC6ULL, "gst_sitting_chair.gda"},
	{0x6AF9EF2B597E0877ULL, "6126831_m.fxe"},
	{0x6AFA061E6D26928BULL, "vxlt_hangedman_n_256_0_0.vlm"},
	{0x6AFA6A080C969BD9ULL, "vxht_brothel_n_5j7_0_0.vlm"},
	{0x6AFA8B90666125DEULL, "vis110_commoner1.bnk"},
	{0x6AFAB3B344A15513ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_spi_p.impa.evt"},
	{0x6AFAC4085991A546ULL, "vxht_mansion2_2kh_0_0.vlm"},
	{0x6AFAECE38C359D87ULL, "dae000_madam_lusine6154931.cl"},
	{0x6AFAEDF0F7DFF7DDULL, "lt_backalley_g.rml"},
	{0x6AFAEEE5EA2FEEA6ULL, "dae000_worthy.bnk"},
	{0x6AFAF8122759CC3BULL, "6188717_f.fxe"},
	{0x6AFB360A852A2A3AULL, "vxw_woundedcoast_gr_0_0.vlm"},
	{0x6AFB6882C53133D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_drunk_loop.gad"},
	{0x6AFB89BD7F155993ULL, "vxca_caverns1_k2n_0_1.vlm"},
	{0x6AFBDC68D6C9E954ULL, "394057343.ogg"},
	{0x6AFC5A7A1A7929B1ULL, "vxlt_warehouse_n_xo_0_0.vlm"},
	{0x6AFC6418582F8D98ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ab_transition.gad"},
	{0x6AFCC842AE5EE378ULL, "463782617.ogg"},
	{0x6AFCCEB203AD4669ULL, "t3_wdg_dw2.tnt"},
	{0x6AFCDFABD6C78A7AULL, "vxdr_primevalentrance_397_0_1.vlm"},
	{0x6AFD0CD359EA9B7CULL, "vxw_bonepitdestroyed_a3_1c6_0_0.vlm"},
	{0x6AFD209B61E9F91EULL, "vxlt_backalley_g_1dg_0_0.vlm"},
	{0x6AFD9F7F9847C9F8ULL, "one220_yevhen6119971.cl"},
	{0x6AFDC0B88968DD55ULL, "6088225_m.fxe"},
	{0x6AFE1AA27EC16D9EULL, "die_dot.mao"},
	{0x6AFE53A7BA1DFC64ULL, "6048185_m.fxe"},
	{0x6AFE55B88F3581E7ULL, "vxlt_undercitydungeon_ww_0_0.vlm"},
	{0x6AFEF5D5D2E9C1F4ULL, "llac_lt_gallowstemplar_f_2384_6.dds"},
	{0x6AFFA619EB55103BULL, "6170098_m.fxe"},
	{0x6AFFB1E52E358063ULL, "mag170tr_eau_end.utt"},
	{0x6AFFBE67B0F74153ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_tr180_rfoot.ani"},
	{0x6AFFC200F0AFD8E6ULL, "222281557.ogg"},
	{0x6AFFF16ED32A1800ULL, "hlac_lt_0_lowtown_-196609_3x.dds"},
	{0x6AFFF6B81615E014ULL, "694381582.ogg"},
	{0x6B0058CB4FB9646EULL, "6116687_f.fxe"},
	{0x6B00ACB344A5B554ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_spi_p.impa.gad"},
	{0x6B00F8DE218A5F15ULL, "6154036_m.fxe"},
	{0x6B0149F21BF49D8FULL, "llac_lt_darktown_-1_38.dds"},
	{0x6B015B76C392C893ULL, "zz_fen_debug6165694.cl"},
	{0x6B0197AA7B05E3FAULL, "vxdr_primevaltomb_75b_0_0.vlm"},
	{0x6B01A1017F0D0A44ULL, "vxca_caverns2prc_7qs_0_0.vlm"},
	{0x6B01D375431DF9A3ULL, "vxlt_0_lowtown_n_e2z_0_0.vlm"},
	{0x6B01FA35E92C95C5ULL, "267444020.ogg"},
	{0x6B020602B461DA3BULL, "vxht_mansion_n_6be_0_0.vlm"},
	{0x6B02344D5816D0DCULL, "vxht_brothel_n_6f5_0_0.vlm"},
	{0x6B02B3BE7B9F8CADULL, "vxlt_gallowstemplar_5ok_0_0.vlm"},
	{0x6B02DF83D88E44E0ULL, "6142452_m.fxe"},
	{0x6B04859B31748F77ULL, "962587215.ogg"},
	{0x6B04A95D065F19EEULL, "llac_ca_caverns1_mines_24424_2.dds"},
	{0x6B04ED853146C4C9ULL, "zz_lgt_1_debug6114221.cl"},
	{0x6B055358D324DE59ULL, "isa321_velasco6091855.cl"},
	{0x6B05A696CF8751B9ULL, "vxw_bonepit_10j_0_0.vlm"},
	{0x6B068F855737723AULL, "891704179.ogg"},
	{0x6B06BF597662D014ULL, "hlac_ca_caverns1_24424_5z.dds"},
	{0x6B06C2D30C657C75ULL, "6097009_m.fxe"},
	{0x6B06DDC4DDA6CCECULL, "vxht_0_hightown_c8v_0_0.vlm"},
	{0x6B0729E6D61D0FC8ULL, "dr_primevalentrance_1724.rml"},
	{0x6B072BFA541D5EEAULL, "367924724.ogg"},
	{0x6B0748C0988E3BB8ULL, "6208607_f.fxe"},
	{0x6B079BF2E53CC0D6ULL, "vxlt_0_lowtown_d5t_0_0.vlm"},
	{0x6B07A9DD4B65CD5BULL, "vxlt_foundry_n_11w_0_0.vlm"},
	{0x6B07AA0918F265E3ULL, "vxw_bonepitdestroyed_a3_1wp_0_1.vlm"},
	{0x6B07B4E55E37190BULL, "dae000_orsino6085456.cl"},
	{0x6B07BBB344ABED32ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_spi_p.impa.ani"},
	{0x6B07BDCFD02543D4ULL, "vxdr_0_deeproads_bwu_0_0.vlm"},
	{0x6B07FB843D961EBBULL, "hlac_lt_gallowstemplar_2391_1y.dds"},
	{0x6B08540B1D57C2C6ULL, "6166868_f.fxe"},
	{0x6B08DBE4FE8CB459ULL, "vxht_brothel_n_5m6_0_0.vlm"},
	{0x6B092E572717A2BEULL, "follower_banter6165481.cl"},
	{0x6B0946B473764871ULL, "851603189.ogg"},
	{0x6B099CE445070D56ULL, "character_character_p254.psh.fxo.dx11.assembly"},
	{0x6B09C6AC33B04ED7ULL, "atl_logininfos_dxt5_00.dds"},
	{0x6B09FCBFC57D2ED0ULL, "vxlt_gallowstemplar_n_5r2_0_0.vlm"},
	{0x6B0AB45A0B0C25D3ULL, "hlac_ht_0_hightown_a3_-1_1z.dds"},
	{0x6B0ADDCE1E310D96ULL, "qun211_party_comment6182712.cl"},
	{0x6B0AE4CC90F62E17ULL, "6016710_m.fxe"},
	{0x6B0B38B820907B27ULL, "6136742_m.fxe"},
	{0x6B0BB1FAE411A3FFULL, "1017906818.ogg"},
	{0x6B0C0427674FA171ULL, "w_blightlands_101_0.mmh"},
	{0x6B0C1027674FB512ULL, "w_blightlands_101_0.mao"},
	{0x6B0C1E27674FCD2FULL, "w_blightlands_101_0.msh"},
	{0x6B0C3903F2132BB5ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef03.phy"},
	{0x6B0C4D20F605AAF4ULL, "1057224074.ogg"},
	{0x6B0CEE95E63C4947ULL, "6089682_m.fxe"},
	{0x6B0CFAC5B57DA031ULL, "6043151_f.fxe"},
	{0x6B0D02F16B1995C2ULL, "6076340_m.fxe"},
	{0x6B0D8BBCB0817449ULL, "loadscreens_one.gda"},
	{0x6B0DB2AAE400AD64ULL, "582715282.ogg"},
	{0x6B0DB6B6BF926247ULL, "vxlt_gallowscourtyard_64a_0_0.vlm"},
	{0x6B0DC9981EEE5EFFULL, "vxlt_0_lowtown_a3_n_c_i4c_0_0.vlm"},
	{0x6B0DD71040BCCDFBULL, "mcr310cr_loyalist_mage.utc"},
	{0x6B0DDCB8345058FAULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.br_atk_v1.evt"},
	{0x6B0E2C74D9174C11ULL, "vxlt_warehouse_14s_0_0.vlm"},
	{0x6B0EA10553029691ULL, "lt_0_lowtown_docks_a3_145_0.mmh"},
	{0x6B0EAD055302AB32ULL, "lt_0_lowtown_docks_a3_145_0.mao"},
	{0x6B0EBB055302C34FULL, "lt_0_lowtown_docks_a3_145_0.msh"},
	{0x6B0ECACA03FCB4C5ULL, "6097068_f.fxe"},
	{0x6B0F3DF6476D5321ULL, "vxlt_backalley_g_1vr_0_0.vlm"},
	{0x6B0FF94089DDDECEULL, "hlac_ca_caverns2_5802_45z.dds"},
	{0x6B1053C9CF06AE49ULL, "vxlt_undercityserialkiller_18l_0_0.vlm"},
	{0x6B10B9F82F835ED1ULL, "vxht_mansion2b_1aj_0_0.vlm"},
	{0x6B10FA1E93D3470CULL, "797940978.ogg"},
	{0x6B11259B6D95A52DULL, "vxlt_gallowstemplar_5hj_0_0.vlm"},
	{0x6B11D42D86619A58ULL, "vxw_encounter2_camp_nr_0_0.vlm"},
	{0x6B12DE9421107402ULL, "73680475.ogg"},
	{0x6B12FEDB7A709FBFULL, "6084555_m.fxe"},
	{0x6B139567B107A4FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_tr180_rfoot.gad"},
	{0x6B13F3D7DC29E0DAULL, "329601128.ogg"},
	{0x6B141873C17D6A39ULL, "vxw_bonepit_15i_0_0.vlm"},
	{0x6B142BF07F24E747ULL, "66843732.ogg"},
	{0x6B1438BB82379268ULL, "795951066.ogg"},
	{0x6B1482BF87BA50C3ULL, "vxw_encounter2_camp_pr_0_1.vlm"},
	{0x6B1493250143A7E2ULL, "hlac_lt_0_lowtown_docks_n_-131075_0y.dds"},
	{0x6B149F753FA089C4ULL, "438262025.ogg"},
	{0x6B1569462052D16CULL, "lt_0_lowtown_docks_a3_n_228_0.msh"},
	{0x6B156F462052DB5EULL, "lt_0_lowtown_docks_a3_n_228_0.mmh"},
	{0x6B157B462052EFF5ULL, "lt_0_lowtown_docks_a3_n_228_0.mao"},
	{0x6B159D6FC8E11480ULL, "338133111.ogg"},
	{0x6B166C284A4E283DULL, "22811463.ogg"},
	{0x6B16BE33F6CA4DFDULL, "ht_0_hightown_n_22_0.rml"},
	{0x6B16D1074F5F5845ULL, "47447607.ogg"},
	{0x6B16EBE70BF75C25ULL, "ash_02.dds"},
	{0x6B172B88193B248BULL, "6208337_m.fxe"},
	{0x6B1749FFB4B18CC6ULL, "6100773_m.fxe"},
	{0x6B176DA747690D22ULL, "pro000_varric_cass.cnv"},
	{0x6B17F81166A7F8FEULL, "vxlt_gallowstemplar_53h_0_0.vlm"},
	{0x6B1816CBD8457902ULL, "hlac_w_woundedcoast_131075_0z.dds"},
	{0x6B182A79A85E5D0FULL, "vxlt_undercityencounter_195_0_0.vlm"},
	{0x6B183AD2AD33C4B4ULL, "vxlt_0_lowtown_a2_n_fec_0_0.vlm"},
	{0x6B198136D9A2214DULL, "fex_bd_d_mid.mmh"},
	{0x6B199336D9A23FD3ULL, "fex_bd_d_mid.msh"},
	{0x6B19989C83580CA0ULL, "vxw_bonepit_1ri_0_0.vlm"},
	{0x6B19AA95128693A7ULL, "6131642_m.fxe"},
	{0x6B19B7F15E1AEDA1ULL, "6125216_m.fxe"},
	{0x6B1A0D08127A088EULL, "1016620002.ogg"},
	{0x6B1A3041EC05C4C2ULL, "vxdr_primevaltomb_2w5_0_0.vlm"},
	{0x6B1A4FA7476B105AULL, "pro000_varric_cass.bnk"},
	{0x6B1AB293D8B70642ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.mmh"},
	{0x6B1AE275014E38EFULL, "6031839_f.fxe"},
	{0x6B1B15F907353120ULL, "gen00fl_anders6045368.cl"},
	{0x6B1C7C3C8F96A6FDULL, "static_static_p13.vsh.fxo.dx11"},
	{0x6B1C9585E5E24EB1ULL, "hlac_dr_slavers_3215_6x.dds"},
	{0x6B1CC7462FE10641ULL, "art\\vfx\\cinematic\\vfx_dust_impact_p.anb"},
	{0x6B1CD09381A28114ULL, "935928606.ogg"},
	{0x6B1CD12529515AFDULL, "hlac_lt_gallowstemplar_n_2391_7z.dds"},
	{0x6B1D05900B364480ULL, "llac_w_blightlands_196607_0.dds"},
	{0x6B1D7E31F0AD0D71ULL, "vxw_bonepit_1hz_0_0.vlm"},
	{0x6B1E47BBDB976282ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fe_transition.gad"},
	{0x6B1F1243D0EE1758ULL, "vxca_cavern1_haunted_k6m_0_0.vlm"},
	{0x6B1F2691AD48228BULL, "llac_ht_playershouse_n_855_19.dds"},
	{0x6B1FB28CBC07145AULL, "464348315.ogg"},
	{0x6B1FCF452CBA0C01ULL, "mrl351_phantomelves6170356.cl"},
	{0x6B1FF33F30B3F637ULL, "nat380_temmerin6067369.cl"},
	{0x6B2044F9C5D6FCD1ULL, "501282618.ogg"},
	{0x6B20460A7857B2D8ULL, "vxw_encounter2_camp_is_0_0.vlm"},
	{0x6B2072E9306FD9F2ULL, "28085391.ogg"},
	{0x6B2116C5C846D936ULL, "hlac_lt_0_lowtown_a2_n_c_1_1x.dds"},
	{0x6B214FC82D236616ULL, "dae000ip_crowd_ff6_talk10.utp"},
	{0x6B21BF56863A0107ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_enter.gad"},
	{0x6B2288BF9E94C597ULL, "t1_mub_br3.tnt"},
	{0x6B229188E531FE8EULL, "mag110_arianni6016067.cl"},
	{0x6B22943BFF070ACDULL, "641399900.ogg"},
	{0x6B22966DB55216B5ULL, "511711782.ogg"},
	{0x6B234139057C6FDEULL, "vxlt_gallowstemplar_f_393_0_0.vlm"},
	{0x6B237A483C902053ULL, "6044909_f.fxe"},
	{0x6B238FB487568B1CULL, "vxdr_primevaltomb_8r2_0_0.vlm"},
	{0x6B23C0B16A90D0D8ULL, "mag291_demon_marethari6029026.cl"},
	{0x6B23C17EC192FDECULL, "vxht_0_hightown_c2t_0_0.vlm"},
	{0x6B23C6C8CB771361ULL, "445447405.ogg"},
	{0x6B23CEE644C66603ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyn_0n.dds"},
	{0x6B23D24D0ADE7005ULL, "ave331_jeven6151149.cl"},
	{0x6B23D85151458437ULL, "765025334.ogg"},
	{0x6B23EE3BC20E238BULL, "vxht_bartrand_n_4g3_0_0.vlm"},
	{0x6B248D8FF8271350ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01smallcairn.msh"},
	{0x6B249B8FF8272B6AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01smallcairn.mmh"},
	{0x6B249E5CFF90D201ULL, "vxht_brothel_n_6fd_0_0.vlm"},
	{0x6B249F8FF82731A1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_01smallcairn.mao"},
	{0x6B24A2FB1991A6C0ULL, "864439378.ogg"},
	{0x6B24BF4520A3BD89ULL, "gen00fl_carver6103784.cl"},
	{0x6B24D07800AA114CULL, "vxw_bonepitmining_a2_273_0_0.vlm"},
	{0x6B24E52888402F34ULL, "nat380_nathaniel6068261.cl"},
	{0x6B255AD372048EADULL, "vxht_0_hightown_bym_0_0.vlm"},
	{0x6B256ABBDB9DBD79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fe_transition.evt"},
	{0x6B25F18C47591946ULL, "vxw_encounter2_ql_0_1.vlm"},
	{0x6B268BA15B0C73BAULL, "vxw_woundedcoast_vm_0_0.vlm"},
	{0x6B26F59BD1880C65ULL, "llac_dr_dungeon_3216_3.dds"},
	{0x6B276EAB653ED805ULL, "dae000_pc_home_ambients6208335.cl"},
	{0x6B27BD561EA179CCULL, "vxw_bonepitmining_a2_1vk_0_0.vlm"},
	{0x6B27BFF292BEBE22ULL, "mag360_thrask6046972.cl"},
	{0x6B27E480776C8119ULL, "dae000ip_crowd_em2_talk15.utp"},
	{0x6B28251C361476D8ULL, "65382262.ogg"},
	{0x6B2851DC7ACD41DEULL, "zz_lgt_1_debug6114620.cl"},
	{0x6B2873006E1D912CULL, "vxht_mansion_n_5bi_0_0.vlm"},
	{0x6B289F866B9F008FULL, "6030422_m.fxe"},
	{0x6B28B45686400D14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_enter.evt"},
	{0x6B293D8E136DD8ECULL, "45170519.ogg"},
	{0x6B2980184F931A7FULL, "vxca_caverns1_gkh_0_0.vlm"},
	{0x6B299AACA5C5728DULL, "vfx_darkcloud01.dds"},
	{0x6B2A69C495FD97A8ULL, "llac_lt_0_lowtown_a2_1_0.dds"},
	{0x6B2AA8462FED0EBAULL, "art\\vfx\\cinematic\\vfx_dust_impact_p.mmh"},
	{0x6B2B32FD9757B495ULL, "vxdr_0_deeproads_fbk_0_0.vlm"},
	{0x6B2B93B4BC1DE163ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_3z.dds"},
	{0x6B2BD7013F355574ULL, "hlac_lt_0_lowtown_a2_n_-196608_1z.dds"},
	{0x6B2C21B02B598421ULL, "vxlt_backalley_g_1xp_0_0.vlm"},
	{0x6B2C535733DBE6D1ULL, "dae000_arishok6067705.cl"},
	{0x6B2C5680057D7DC3ULL, "501312485.ogg"},
	{0x6B2C5DC5518ECA31ULL, "vxlt_darktown_4ed_0_0.vlm"},
	{0x6B2C8F4E7F7A7345ULL, "vxdr_primevaltomb_75i_0_0.vlm"},
	{0x6B2CEBCDB9CAEA6AULL, "288926331.ogg"},
	{0x6B2CFEC1DB4790ECULL, "narrative_varric6172505.cl"},
	{0x6B2D247AC901303BULL, "hlac_lt_0_lowtown_docks_n_0_4z.dds"},
	{0x6B2DCD0B288BEE2AULL, "780947564.ogg"},
	{0x6B2DECF0F80B4FF5ULL, "lt_backalley_g.anb"},
	{0x6B2DF0F0F80B560FULL, "lt_backalley_g.arl"},
	{0x6B2E0A2804FC285EULL, "660541802.ogg"},
	{0x6B2E17B2DFDBD237ULL, "278958020.ogg"},
	{0x6B2E3061A3D6748DULL, "w_encounter1_probe_jm.dds"},
	{0x6B2E33E5F05547B9ULL, "fx_link.mao"},
	{0x6B2E768B48B8EF3FULL, "gen00fl_aveline6139120.cl"},
	{0x6B2FE80E8F752B8EULL, "71043856.ogg"},
	{0x6B30429AE9733586ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_enter.evt"},
	{0x6B30613CF9317D47ULL, "vxlt_0_lowtown_e7z_0_0.vlm"},
	{0x6B3093F7516C582BULL, "hlac_lt_0_lowtown_-196608_3x.dds"},
	{0x6B30FCFA14D35A82ULL, "587037666.ogg"},
	{0x6B319568E21AEF3FULL, "vxht_0_hightown_c8y_0_0.vlm"},
	{0x6B31D44535AE3006ULL, "vxlt_gallowstemplar_51j_0_0.vlm"},
	{0x6B32B4C4BFE20F47ULL, "hlac_ca_caverns1_24424_27y.dds"},
	{0x6B32B9E0A2ADC9BCULL, "w_sundermount_84_0.mao"},
	{0x6B32C0A496BA55A4ULL, "6117472_m.fxe"},
	{0x6B32C5E0A2ADDE5FULL, "w_sundermount_84_0.mmh"},
	{0x6B32C7D43B0A72D5ULL, "6108947_m.fxe"},
	{0x6B32CBE0A2ADE801ULL, "w_sundermount_84_0.msh"},
	{0x6B334254F2A029CCULL, "vxw_bonepitmining_a2_220_0_0.vlm"},
	{0x6B3372BBDBAA0868ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fe_transition.ani"},
	{0x6B337F906690D6FBULL, "vis110_commoner1.utc"},
	{0x6B33817C1D9F00D8ULL, "5753646.ogg"},
	{0x6B345D648A7995C3ULL, "896947224.ogg"},
	{0x6B34A5A89D30B8F2ULL, "6164968_m.fxe"},
	{0x6B34FD7E4D028C3AULL, "vxw_woundedcoast_sn_0_0.vlm"},
	{0x6B3552569B5F7850ULL, "vxlt_gallowstemplar_n_5c1_0_0.vlm"},
	{0x6B35923F618F0783ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_c.impa.evt"},
	{0x6B35C4E70C11CACFULL, "ash_02.mao"},
	{0x6B35C85CFF41E328ULL, "follower_barks6153377.cl"},
	{0x6B35CD090961B8B0ULL, "440829632.ogg"},
	{0x6B35D299B13FE0E4ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyo_0d.dds"},
	{0x6B35F242C96ED229ULL, "lt_0_lowtown_docks_n_18_0.rml"},
	{0x6B3604F9E4229F9DULL, "6100257_m.fxe"},
	{0x6B362F331097924CULL, "vxw_bonepitmining_a2_1sh_0_0.vlm"},
	{0x6B3631083C825C6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_l_open.evt"},
	{0x6B36AC56864C3D4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1_enter.ani"},
	{0x6B37599AE9797BD1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_enter.gad"},
	{0x6B37C52CD31B60A4ULL, "vxlt_gallowstemplar_38q_0_0.vlm"},
	{0x6B37D2722D281B98ULL, "lt_0_lowtown_docks_29_0.rml"},
	{0x6B37E59F33D6EBDAULL, "993207880.ogg"},
	{0x6B3815526D0A2ADEULL, "961786348.ogg"},
	{0x6B389AEC0AA89653ULL, "lt_undercityserialkiller_3567_ll.mmh"},
	{0x6B390E8036ED9937ULL, "art\\vfx\\combat\\vfx_c_rw_arc_blast_c.anb"},
	{0x6B39183A889A6F2FULL, "6051647_m.fxe"},
	{0x6B39C7219F99C3EFULL, "follower_banter6172785.cl"},
	{0x6B3A34CA79288BC0ULL, "dae200_cullen6140256.cl"},
	{0x6B3ACEEC655319C2ULL, "vxht_playershouse_n_1s5_0_0.vlm"},
	{0x6B3B3E416C3490DFULL, "gen00fl_aveline6139454.cl"},
	{0x6B3B9453822757FCULL, "6083828_f.fxe"},
	{0x6B3BA960A4E8F7C9ULL, "vxlt_undercityserialkiller_17q_0_0.vlm"},
	{0x6B3C1A6AF0F31B1FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright.phy"},
	{0x6B3C4A083C86F22CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_l_open.gad"},
	{0x6B3C5744C82A87BCULL, "6071039_m.fxe"},
	{0x6B3C6B3F6194E404ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_c.impa.gad"},
	{0x6B3C8B5817380E42ULL, "6160630_m.fxe"},
	{0x6B3CCA06D17FD942ULL, "vxlt_warehouse_1dl_0_0.vlm"},
	{0x6B3CFE997492C0C9ULL, "gen00fl_aveline6140005.cl"},
	{0x6B3D2B7C0565A66DULL, "vxca_cave_19r_0_0.vlm"},
	{0x6B3D5A9AE97DE97FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_enter.ani"},
	{0x6B3D941298120DBEULL, "774448206.ogg"},
	{0x6B3DE6B01818D4D4ULL, "6083690_m.fxe"},
	{0x6B3DE862CEFA0F4EULL, "hlac_lt_blackemporium_1055_25y.dds"},
	{0x6B3ED319EB2795C7ULL, "vxlt_0_lowtown_e2u_0_0.vlm"},
	{0x6B3F095EC9C768D7ULL, "w_woundedcoast_176_0.msh"},
	{0x6B3F0CA6F6F5D228ULL, "68626019.ogg"},
	{0x6B3F1B5EC9C7873AULL, "w_woundedcoast_176_0.mao"},
	{0x6B3F1F5EC9C78E49ULL, "w_woundedcoast_176_0.mmh"},
	{0x6B3F368B0B75396CULL, "6143727_m.fxe"},
	{0x6B3F78B7EBF85EA5ULL, "416256049.ogg"},
	{0x6B3FB99EB8434D76ULL, "narrative_flemeth6002808.cl"},
	{0x6B3FBE5C1DA11DD9ULL, "669736390.ogg"},
	{0x6B4027C72FCCCCB5ULL, "vxht_0_hightown_n_a2p_0_0.vlm"},
	{0x6B403846D1DFFC1FULL, "6165104_m.fxe"},
	{0x6B4078C4F6DF6DC3ULL, "181289240.ogg"},
	{0x6B4086E74BB70B46ULL, "gen00fl_aveline6120763.cl"},
	{0x6B40988A21A1C16AULL, "ave200_donnic6128939.cl"},
	{0x6B41186E5F20D63DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6_tool.mao"},
	{0x6B411C6E5F20DD76ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6_tool.mmh"},
	{0x6B41266E5F20EE74ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6_tool.msh"},
	{0x6B422B96B1C9B95DULL, "dae000_keep_bureaucrat_1_st.bnk"},
	{0x6B428E29365E454BULL, "vxlt_0_lowtown_docks_bbw_0_0.vlm"},
	{0x6B4297F4D9BDB82CULL, "vxlt_hangedman_1qi_0_0.vlm"},
	{0x6B4349083C8D0F2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_l_open.ani"},
	{0x6B437A3F619B1CE2ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_c.impa.ani"},
	{0x6B43B193D8DA09E9ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.anb"},
	{0x6B43D0ADE5950B6FULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hanginga.phy"},
	{0x6B43DCDFF58BFC39ULL, "vxca_caverns1_e7l_0_0.vlm"},
	{0x6B44556E71F40EDCULL, "hlac_ht_chantry_2534_16x.dds"},
	{0x6B448F9EDCDA60EFULL, "w_blightlands_23_0.mao"},
	{0x6B449B9EDCDA748CULL, "w_blightlands_23_0.mmh"},
	{0x6B44A19EDCDA7F7EULL, "w_blightlands_23_0.msh"},
	{0x6B450D96B1CBBC65ULL, "dae000_keep_bureaucrat_1_st.cnv"},
	{0x6B451BE1D9AD6F18ULL, "vxw_encounter3_be_0_0.vlm"},
	{0x6B4556BA5209C22CULL, "vxht_mansion_n_5ho_0_0.vlm"},
	{0x6B4560D4B65FDA22ULL, "6161080_m.fxe"},
	{0x6B45BE25473DD2B7ULL, "6019889_m.fxe"},
	{0x6B4642E5F069F347ULL, "fx_link.dds"},
	{0x6B46923F0D3EE753ULL, "qun120_arishok_guard6054850.cl"},
	{0x6B46A889F44F616CULL, "hlac_lt_0_lowtown_docks_-3_4x.dds"},
	{0x6B479EE52E731502ULL, "mag170tr_eau_end.ncs"},
	{0x6B47B66F1D2F26E6ULL, "vxht_keep_n_5xp_0_1.vlm"},
	{0x6B485BFDE285D3F4ULL, "1870125.ogg"},
	{0x6B48657FDA9FA583ULL, "205269410.ogg"},
	{0x6B48D10462587856ULL, "624825982.ogg"},
	{0x6B4940C957493242ULL, "vxht_playershouse_n_1v6_0_0.vlm"},
	{0x6B49AC5A647192BBULL, "w_encounter2_25_0.msh"},
	{0x6B49BE5A6471B106ULL, "w_encounter2_25_0.mao"},
	{0x6B49CA5A6471C5A5ULL, "w_encounter2_25_0.mmh"},
	{0x6B4A51BAA4CD23ADULL, "hlac_lt_0_lowtown_a2_n_1_8z.dds"},
	{0x6B4A6D34C6953034ULL, "vxht_mansion2b_2yx_0_0.vlm"},
	{0x6B4A816BF75BC0D1ULL, "vxht_mansion2b_1un_0_0.vlm"},
	{0x6B4AFD35092E26C2ULL, "6165930_m.fxe"},
	{0x6B4B9FCA765E5E37ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_infb_0d.dds"},
	{0x6B4C822656D14B2FULL, "dae300_alistair_king6133629.cl"},
	{0x6B4C8EE3C882FD81ULL, "201323686.ogg"},
	{0x6B4CAE52BCF1E9DFULL, "6218017_m.fxe"},
	{0x6B4CB3A9D9722AC5ULL, "6116319_m.fxe"},
	{0x6B4D1B6AF1019024ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright.mmh"},
	{0x6B4D1F6AF1019777ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright.mao"},
	{0x6B4D316AF101B5C6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright.msh"},
	{0x6B4D7F28BDDF38E5ULL, "zz_approval_debug6003644.cl"},
	{0x6B4DC582219C9AA5ULL, "vxht_brothel_n_6bd_0_0.vlm"},
	{0x6B4DD09FDEB69A9DULL, "482480739.ogg"},
	{0x6B4E3F2A36DABFD2ULL, "gen00fl_merrill6124403.cl"},
	{0x6B4E6CA04C28710CULL, "dae120ip_chant_board_ht.ncs"},
	{0x6B4EACC245F735D5ULL, "6088208_f.fxe"},
	{0x6B4F1FD2BDAA6482ULL, "zz_dog_debug6127731.cl"},
	{0x6B4F6C49749028B7ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_9x.dds"},
	{0x6B5020024B05D43BULL, "hlac_lt_gallowstemplar_f_7198_0y.dds"},
	{0x6B50B2F8858EB4C6ULL, "vxlt_hangedman_56_0_0.vlm"},
	{0x6B50E09F4FA236B8ULL, "747867780.ogg"},
	{0x6B512F1D66B224DCULL, "645704820.ogg"},
	{0x6B5158E26457C8D4ULL, "vxlt_blackemporium_tg_0_0.vlm"},
	{0x6B51823C9D4B03ABULL, "433915923.ogg"},
	{0x6B5195EE82E7FAFCULL, "vxdr_primevaltomb_8y7_0_0.vlm"},
	{0x6B51A67FC2F58E18ULL, "vxht_mansion2_30z_0_0.vlm"},
	{0x6B51D6EE8D08CF77ULL, "832775896.ogg"},
	{0x6B52194547262926ULL, "and100_anders6095897.cl"},
	{0x6B5226DC3891EA3AULL, "w_sundermount_183_0.phy"},
	{0x6B52551937B24378ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_c.anb"},
	{0x6B52D8174569AEF6ULL, "6097903_m.fxe"},
	{0x6B5304DC0193D528ULL, "6155496_m.fxe"},
	{0x6B53291A16B1CBE9ULL, "vxlt_warehouse_1cb_0_0.vlm"},
	{0x6B53330FBB1D27A7ULL, "vxlt_warehouse_n_111_0_0.vlm"},
	{0x6B5335DC731CC387ULL, "zz_followers_banter6157401.cl"},
	{0x6B53490526D9AB2FULL, "66503542.ogg"},
	{0x6B539DA9FC09BA64ULL, "594765352.ogg"},
	{0x6B53D2B1A855F2A2ULL, "6166180_m.fxe"},
	{0x6B53DDF62A7D4FD8ULL, "ave200_donnic6110754.cl"},
	{0x6B5427C823FAFCFFULL, "qun131_taskmaster6019165.cl"},
	{0x6B542E9BE6B17801ULL, "vxw_sundermount_22t_0_0.vlm"},
	{0x6B544C03F25101D0ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef03.msh"},
	{0x6B5453D756413E48ULL, "6100771_m.fxe"},
	{0x6B545A03F25119EAULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef03.mmh"},
	{0x6B548BF59B151881ULL, "llac_w_woundedcoast_-65529_0.dds"},
	{0x6B55133FC5C29A97ULL, "dae000ip_crowd_lm5_talk6.utp"},
	{0x6B55717329EB2314ULL, "153931873.ogg"},
	{0x6B55A2884BCC75F5ULL, "vxlt_gallowstemplar_f_5z3_0_0.vlm"},
	{0x6B55C2441FE31593ULL, "542918244.ogg"},
	{0x6B560AE609E9065FULL, "6097313_m.fxe"},
	{0x6B560F86D3AB0BA7ULL, "hlac_lt_hangedman_n_-1_16x.dds"},
	{0x6B56A707DA9D52A3ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_bda.mmh"},
	{0x6B56B507DA9D6AFDULL, "art\\characters\\playercharacter\\humanfemale\\hlf_bda.msh"},
	{0x6B570D0C5B28E772ULL, "hlac_lt_0_lowtown_a2_n_65535_2x.dds"},
	{0x6B57655F6623B3D5ULL, "vxht_store_ca_0_0.vlm"},
	{0x6B57848B56A84F92ULL, "hlac_lt_0_lowtown_docks_n_-131074_6x.dds"},
	{0x6B57B35C277113B1ULL, "vxlt_darktown_4jy_0_0.vlm"},
	{0x6B57C800C0153FCBULL, "vxht_keep_n_4bs_0_0.vlm"},
	{0x6B57F1B20019A060ULL, "and100_anders6096050.cl"},
	{0x6B57F5024922777CULL, "367681317.ogg"},
	{0x6B5807B3D1EFC8DFULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\bsh_thorn_dp.dds"},
	{0x6B582352206C1272ULL, "615190039.ogg"},
	{0x6B589CA3B94008F1ULL, "6166635_m.fxe"},
	{0x6B58E5CA12B6C59FULL, "vxlt_0_lowtown_enx_0_1.vlm"},
	{0x6B5963B185B3E5A2ULL, "823246555.ogg"},
	{0x6B5AA32C751700F2ULL, "gen_im_trs_gen_two_2hm_r1.uti"},
	{0x6B5AD111E757C3E6ULL, "606241111.ogg"},
	{0x6B5B22571EE3BCBCULL, "6066350_m.fxe"},
	{0x6B5B8BF97A827494ULL, "162556247.ogg"},
	{0x6B5C0E2B379969B5ULL, "llac_w_woundedcoast_262147_0.dds"},
	{0x6B5D29DF31437F02ULL, "ser121_jethann6108796.cl"},
	{0x6B5DF19CB2B782C2ULL, "9863039.ogg"},
	{0x6B5E0E16F97831FCULL, "6143503_m.fxe"},
	{0x6B5E8B0CC125F63FULL, "dae000_meredith6085296.cl"},
	{0x6B5EC9E5BBCB89C9ULL, "dae_rogue_decoy.ncs"},
	{0x6B5F39F27B973653ULL, "ht_0_hightown_a2_20_0_lc.mmh"},
	{0x6B5F4FC7128F4FA5ULL, "vxw_blightlands_9i_0_0.vlm"},
	{0x6B5FB6490742607BULL, "vxlt_gallowstemplar_5qs_0_0.vlm"},
	{0x6B5FD72DE2EE6A93ULL, "6020526_m.fxe"},
	{0x6B5FF0A268C60C63ULL, "hlac_lt_0_lowtown_a2_1_11y.dds"},
	{0x6B60DB2F5B6CDC55ULL, "935606615.ogg"},
	{0x6B61A4ECAD134027ULL, "vxlt_warehouse_n_160_0_0.vlm"},
	{0x6B620B8037109938ULL, "art\\vfx\\combat\\vfx_c_rw_arc_blast_c.mmh"},
	{0x6B62216124C0C51DULL, "6069862_m.fxe"},
	{0x6B6250683E069BE5ULL, "hlac_w_encounter3_-1_0x.dds"},
	{0x6B62A078D8A79081ULL, "vxw_sundermount_27k_0_0.vlm"},
	{0x6B62AF273C415678ULL, "hlac_lt_0_lowtown_n_131071_1x.dds"},
	{0x6B62C5B4483756C8ULL, "6105471_m.fxe"},
	{0x6B62EFFA7AE6822AULL, "hlac_w_sundermount_327681_0z.dds"},
	{0x6B62FBAE150E4517ULL, "6166383_m.fxe"},
	{0x6B63237BA8FB9EAAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardright.gad"},
	{0x6B637040F786DB3AULL, "vxw_bonepit_1nw_0_0.vlm"},
	{0x6B63B3722D4D2250ULL, "lt_0_lowtown_docks_29_0.anb"},
	{0x6B6482239A9B9988ULL, "mag360_thrask6055233.cl"},
	{0x6B648DDDE83718D7ULL, "86644287.ogg"},
	{0x6B649C583B069EB7ULL, "sib131_slaver_lord.bnk"},
	{0x6B658B1E07AF5962ULL, "6087807_m.fxe"},
	{0x6B65B52B07C97A18ULL, "6111225_m.fxe"},
	{0x6B65B8B0A62D695CULL, "one100_party_comments6112618.cl"},
	{0x6B65FE12DF4B10FEULL, "vxca_caverns2_8cu_0_0.vlm"},
	{0x6B661E528DC1AA8EULL, "seb321_leliana6180929.cl"},
	{0x6B6665E88293C8BFULL, "hlac_lt_0_lowtown_a2_0_5y.dds"},
	{0x6B668CE8F832CD64ULL, "hlac_ht_mansion2b_818_35x.dds"},
	{0x6B675F4C358E5BC9ULL, "792336340.ogg"},
	{0x6B678BD4F08DE542ULL, "6028808_m.fxe"},
	{0x6B67A807DAABC7C8ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_bda.phy"},
	{0x6B67FACC033227CFULL, "llac_ht_mansion_n_2995_6.dds"},
	{0x6B683C436F16376AULL, "752964863.ogg"},
	{0x6B684E180EB02038ULL, "fxc_puke_ani.mao"},
	{0x6B687E583B0A54CFULL, "sib131_slaver_lord.cnv"},
	{0x6B68C2832609A321ULL, "one200dg_solivitus6060328.cl"},
	{0x6B68F142C99A2A49ULL, "lt_0_lowtown_docks_n_18_0.anb"},
	{0x6B691BF5C1261818ULL, "vxht_0_hightown_n_bxl_0_0.vlm"},
	{0x6B69466EF673CC66ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\armreplace.evt"},
	{0x6B69F60F6DEA282AULL, "vxht_mansion2_2nk_0_0.vlm"},
	{0x6B6A067BA9018C41ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardright.evt"},
	{0x6B6A07C235F51F7FULL, "vxlt_0_lowtown_docks_a3_d0l_0_0.vlm"},
	{0x6B6A17D922957A0EULL, "art\\levels\\ztools\\default\\generic\\9ff_signmarker.phy"},
	{0x6B6A45B2C479F989ULL, "hlac_w_encounter2_camp_-65532_0y.dds"},
	{0x6B6A57EB8DDE9190ULL, "917753318.ogg"},
	{0x6B6A66DA3141495EULL, "dae021_sister_samea_sta6150140.cl"},
	{0x6B6AA93C0588CBA5ULL, "vxht_brothel_n_6hb_0_0.vlm"},
	{0x6B6B06E6284F4F43ULL, "vxca_cave_1on_0_0.vlm"},
	{0x6B6B1EC401D77E8CULL, "vxdr_primevaltomb_6j4_0_0.vlm"},
	{0x6B6BCD170DBF1E6BULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.mmh"},
	{0x6B6BF579DE73EAB5ULL, "gen_im_acc_rng_act3_02.uti"},
	{0x6B6D1AB3270FCD90ULL, "20318457.ogg"},
	{0x6B6D60F8E0221DB0ULL, "6092471_f.fxe"},
	{0x6B6DB7FB59DBC1DBULL, "6024100_m.fxe"},
	{0x6B6DBB61139AAF9BULL, "vxlt_gallowstemplar_f_5v9_0_0.vlm"},
	{0x6B6DED8BAFA001ACULL, "vxlt_hangedman_1fl_0_0.vlm"},
	{0x6B6DEDB60B2E1EF7ULL, "ter_stoney2_n.dds"},
	{0x6B6E823A3AC0F132ULL, "vxca_cavern1_haunted_dq5_0_0.vlm"},
	{0x6B6F9493BD40D9A7ULL, "hlac_lt_0_lowtown_n_-65535_6z.dds"},
	{0x6B6FAB36D18272DEULL, "6051673_m.fxe"},
	{0x6B6FE14C27A76412ULL, "vxdr_0_deeproads_cub_0_2.vlm"},
	{0x6B705D6EF67A11B1ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\armreplace.gad"},
	{0x6B707E737F30724BULL, "564295725.ogg"},
	{0x6B70B487A12F5DB1ULL, "344269975.ogg"},
	{0x6B70EEFDEDE74BE2ULL, "gen_im_wep_war_two_act1_02.uti"},
	{0x6B711255CA9DA901ULL, "vxw_sundermount_28j_0_0.vlm"},
	{0x6B716869BC5B2F86ULL, "vxca_caverns2_8ff_0_0.vlm"},
	{0x6B716BBB73A37BB3ULL, "vxdr_0_deeproads_cf1_0_1.vlm"},
	{0x6B716E22C7309621ULL, "vxlt_backalley_g_203_0_0.vlm"},
	{0x6B71D2FC453FAA53ULL, "vxdr_0_deeproads_bon_0_0.vlm"},
	{0x6B7277DE89D7118EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtopen.mmh"},
	{0x6B7283DE89D72625ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtopen.mao"},
	{0x6B7291DE89D73E7CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtopen.msh"},
	{0x6B72F354E86C863BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_smelt_sa.dds"},
	{0x6B7375B152B6CDD7ULL, "169223509.ogg"},
	{0x6B73923E65AE1DEDULL, "vxlt_undercitydungeon_15b_0_0.vlm"},
	{0x6B7403CA7D288BAEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter.ani"},
	{0x6B743C7D2E9B8A8AULL, "305566177.ogg"},
	{0x6B746ACF1828176DULL, "ico_gloves_light_1.dds"},
	{0x6B7486CBD84B39DCULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgtb_n.dds"},
	{0x6B74DE8CFFDD12EEULL, "dae210ar_lowtown.ncs"},
	{0x6B752CBAD0C04A2EULL, "6084345_m.fxe"},
	{0x6B7597098BED224FULL, "6049224_m.fxe"},
	{0x6B75ADAF95088F47ULL, "vxlt_0_lowtown_ha9_0_0.vlm"},
	{0x6B75E5FA3D2CDB42ULL, "705785666.ogg"},
	{0x6B7607041C2C1F7CULL, "hlac_lt_0_lowtown_-65536_5z.dds"},
	{0x6B768D4443EF824BULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.dura.evt"},
	{0x6B76B3E8171D7EC5ULL, "982992316.ogg"},
	{0x6B76C05642D20866ULL, "mcr310_sibling6076369.cl"},
	{0x6B76DD307CB7381BULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ds001a_0t.dds"},
	{0x6B76FE7BA90C0890ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardright.ani"},
	{0x6B77070D6E576630ULL, "336651912.ogg"},
	{0x6B772F146F87743DULL, "6132818_m.fxe"},
	{0x6B775E6EF680325FULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\armreplace.ani"},
	{0x6B783D4D90F6681AULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_orsa_0n.dds"},
	{0x6B783E3C76F511A2ULL, "6087595_m.fxe"},
	{0x6B7859C3ADFAD900ULL, "vxw_sundermount_2nm_0_0.vlm"},
	{0x6B7867EC5FE040AAULL, "vxht_mansion2_2kd_0_0.vlm"},
	{0x6B78954308DAE82BULL, "6020063_m.fxe"},
	{0x6B78AE170DC973D0ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.anb"},
	{0x6B78D33A03920189ULL, "zz_settint.ncs"},
	{0x6B78EA8B17B1E409ULL, "vxlt_warehouse_n_140_0_0.vlm"},
	{0x6B7955FE4754881FULL, "6076925_m.fxe"},
	{0x6B797C7F0C2D2F89ULL, "pro000_athenril6048750.cl"},
	{0x6B798EB93A50AD14ULL, "zz_approval_debug6003786.cl"},
	{0x6B79931E913FB2B0ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_stn_0n.dds"},
	{0x6B79E284B86680F8ULL, "zz_and_debug6176457.cl"},
	{0x6B7A03AB95D3A013ULL, "6075605_m.fxe"},
	{0x6B7A541937D39313ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_c.mmh"},
	{0x6B7A96E12069E32FULL, "vxlt_undercityencounter_1b0_0_0.vlm"},
	{0x6B7ABBD53CEA57F2ULL, "gen00fl_varric6108126.cl"},
	{0x6B7AD68CFFE170C7ULL, "dae210ar_lowtown.lst"},
	{0x6B7AF58185578CBAULL, "hlac_lt_blackemporium_1055_11z.dds"},
	{0x6B7B5EA4E212E2F5ULL, "mag360_thrask6043901.cl"},
	{0x6B7BC209459AE92EULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbed.phy"},
	{0x6B7BDDF675C47081ULL, "6131148_f.fxe"},
	{0x6B7C1C788344D5EEULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medd_s.dds"},
	{0x6B7C2D3E0590C81BULL, "vxlt_gallowstemplar_f_5y8_0_0.vlm"},
	{0x6B7C4EB89B91A5DDULL, "6204461_m.fxe"},
	{0x6B7C5F68A1961A2CULL, "vxlt_hangedman_1mm_0_0.vlm"},
	{0x6B7C6659471CFD3AULL, "vxlt_0_lowtown_docks_b6o_0_1.vlm"},
	{0x6B7CF166FC565E74ULL, "349272539.ogg"},
	{0x6B7D25848E6AAD46ULL, "257019334.ogg"},
	{0x6B7D2669C81086C0ULL, "245550607.ogg"},
	{0x6B7D58A6B29DBBDDULL, "355265362.ogg"},
	{0x6B7DA64443F5CB0CULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.dura.gad"},
	{0x6B7DFBAA06823BD8ULL, "vxlt_0_lowtown_a3_n_c_gz8_0_0.vlm"},
	{0x6B7EC91B4A503818ULL, "6168179_m.fxe"},
	{0x6B7F4BBF50A9342DULL, "vxht_0_hightown_a2_b19_0_0.vlm"},
	{0x6B7FD4D767372360ULL, "character_character_p122.psh.fxo"},
	{0x6B7FE0343D036E02ULL, "vxlt_undercityencounter_1sv_0_0.vlm"},
	{0x6B7FF431DD4344F6ULL, "vxw_encounter2_5y_0_0.vlm"},
	{0x6B80038F73A9EE43ULL, "vxlt_0_lowtown_eqt_0_0.vlm"},
	{0x6B8017CC2F44FDA4ULL, "gen00fl_aveline6140068.cl"},
	{0x6B801B205B63FB18ULL, "231987725.ogg"},
	{0x6B801B5148C43EC1ULL, "310411779.ogg"},
	{0x6B802C4445D667FEULL, "6129155_f.fxe"},
	{0x6B803E2B57A07D6AULL, "ts_circleofmagi_01.gda"},
	{0x6B806548FE53B6E5ULL, "hlac_ht_brothel_n_3370_0y.dds"},
	{0x6B8071959B7E35BAULL, "gen_im_trs_chn_shd_wdn_r1.uti"},
	{0x6B8095E4045BFA0EULL, "gen00fl_merrill6119601.cl"},
	{0x6B80F81F21AEBF75ULL, "vxlt_gallowstemplar_f_5k0_0_0.vlm"},
	{0x6B8199FA5091F8BFULL, "193195357.ogg"},
	{0x6B81D0F709A9D478ULL, "gen00fl_fenris6105875.cl"},
	{0x6B81FBCA7D34BBFFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter.evt"},
	{0x6B81FCE66ED9C89AULL, "ser270_templar6112564.cl"},
	{0x6B82041B57A4366DULL, "vxlt_undercitydungeon_18a_0_0.vlm"},
	{0x6B82B2C52DB497CDULL, "vxw_bonepitdestroyed_a3_26k_0_0.vlm"},
	{0x6B834A908EE812D5ULL, "6083156_m.fxe"},
	{0x6B834D81DB8F8E83ULL, "59476389.ogg"},
	{0x6B83A54443FA350AULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.dura.ani"},
	{0x6B8429D24E1CCCC5ULL, "hlac_lt_0_lowtown_docks_a3_-131075_2z.dds"},
	{0x6B843AC35D9E3486ULL, "6097833_m.fxe"},
	{0x6B844F9D197FD60DULL, "6033596_f.fxe"},
	{0x6B84B2EB680D028CULL, "art\\levels\\races\\ferelden\\props\\_textures\\brazier_dp.dds"},
	{0x6B84C2D0399F54BAULL, "vxca_caverns1_egs_0_0.vlm"},
	{0x6B8501C377028227ULL, "vxht_0_hightown_a2_9u4_0_0.vlm"},
	{0x6B8577FE82DF4B5BULL, "6065961_m.fxe"},
	{0x6B8587337AEEEF45ULL, "hlac_lt_undercityencounter_-1_5z.dds"},
	{0x6B862E43396ABD73ULL, "alphacloudanim01.mao"},
	{0x6B865F180EC93ABAULL, "fxc_puke_ani.dds"},
	{0x6B866633756D49D3ULL, "qcr220_meredith6041528.cl"},
	{0x6B86CBA09FF0F180ULL, "vxw_sundermount_2sj_0_0.vlm"},
	{0x6B875C6809A7FC89ULL, "vxlt_warehouse_n_197_0_0.vlm"},
	{0x6B8765142A0F8BA8ULL, "vxlt_foundry_n_zn_0_0.vlm"},
	{0x6B87769A8E78C52BULL, "496649739.ogg"},
	{0x6B8780A8CCE34F43ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03bricks.mao"},
	{0x6B8784A8CCE35590ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03bricks.mmh"},
	{0x6B8792A8CCE36DAAULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_03bricks.msh"},
	{0x6B87963F5BCA092BULL, "vfx_vfx_p0.psh.fxo.dx11.assembly"},
	{0x6B87CBD4707AB675ULL, "mer221_sabin6080558.cl"},
	{0x6B87E75CF09796DCULL, "6180674_m.fxe"},
	{0x6B88B50CFC2793CDULL, "vxlt_0_lowtown_a2_f5i_0_0.vlm"},
	{0x6B88DCCA7D3AA5E8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter.gad"},
	{0x6B88F25ABD516F9AULL, "hlac_ca_cavern1_haunted_24427_5x.dds"},
	{0x6B89483EDA990690ULL, "hightown_fem_elfservant_mob6169094.cl"},
	{0x6B8978FF45FFBB1EULL, "vxw_bonepitmining_a2_26n_0_0.vlm"},
	{0x6B89CA5A64A7A9AEULL, "w_encounter2_25_0.rml"},
	{0x6B8A102E0E026A82ULL, "cod_hst_tevinter.utp"},
	{0x6B8A3D46F9D1C1C7ULL, "6148079_m.fxe"},
	{0x6B8AC59F4CBFB2BFULL, "vxw_encounter2_rt_0_1.vlm"},
	{0x6B8AD8820F1D19E1ULL, "zz_mrl_debug6185294.cl"},
	{0x6B8AFB2D730F1D64ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masb.msh"},
	{0x6B8B0D2D730F3BCDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masb.mao"},
	{0x6B8B112D730F4206ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masb.mmh"},
	{0x6B8B92144DE4F174ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyo_0t.dds"},
	{0x6B8BD26595EB3C73ULL, "364903813.ogg"},
	{0x6B8BD78CFFEFE5FCULL, "dae210ar_lowtown.are"},
	{0x6B8C265030073DC9ULL, "dae000_tomwise6167334.cl"},
	{0x6B8C8ECF6FCC461EULL, "hlac_lt_0_lowtown_n_131071_0y.dds"},
	{0x6B8C90D7C4A4E931ULL, "466377118.ogg"},
	{0x6B8C9B05C56E6847ULL, "llac_dr_0_deeproads_5845_14.dds"},
	{0x6B8CBC02E64FA6F1ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_nud_0s.dds"},
	{0x6B8D26B466B79A91ULL, "6168123_m.fxe"},
	{0x6B8D5185FCB38A2FULL, "qun231_varnell6078106.cl"},
	{0x6B8D8E9D2704ED79ULL, "6208209_m.fxe"},
	{0x6B8DB50CF6A60A50ULL, "6134307_m.fxe"},
	{0x6B8DBF6801EF51FAULL, "38834442.ogg"},
	{0x6B8E42CDF3EB9406ULL, "vxht_0_hightown_bgl_0_0.vlm"},
	{0x6B8E4F6276264B7AULL, "hlac_lt_0_lowtown_a2_196606_0y.dds"},
	{0x6B8EAB016182C59FULL, "hm_qfq_mob3.mor"},
	{0x6B8F0FFE368FDF9CULL, "vxht_0_hightown_a2_c5e_0_0.vlm"},
	{0x6B8F71DD598681D4ULL, "momnt_star.mao"},
	{0x6B906778455848C5ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.mmh"},
	{0x6B90A35A64AD8652ULL, "w_encounter2_25_0.phy"},
	{0x6B90EFD3B71FD3D8ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyn_0d.dds"},
	{0x6B910D19805DD171ULL, "vxw_bonepit_10r_0_0.vlm"},
	{0x6B9124A21FAAB04DULL, "vxw_bonepitdestroyed_a3_21j_0_0.vlm"},
	{0x6B913534F547E2E4ULL, "hlac_lt_0_lowtown_a2_n_196606_0x.dds"},
	{0x6B9153A9B52D5A7EULL, "vxca_caverns2_8ph_0_0.vlm"},
	{0x6B91C8DD4E74702AULL, "weapon1_s1y.psh.fxo"},
	{0x6B92317C3B8AA0BDULL, "llac_lt_warehouse_65537_2.dds"},
	{0x6B925CC322114CA6ULL, "llac_lt_gallowstemplar_2388_1.dds"},
	{0x6B92910851627D0AULL, "isa211_isabela6202221.cl"},
	{0x6B930885D3059F21ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v3.ani"},
	{0x6B933AA3C41A322EULL, "6027571_m.fxe"},
	{0x6B93F6AFDB71B16EULL, "vxca_caverns2prc_afv_0_0.vlm"},
	{0x6B9411BA7CFF444EULL, "910578790.ogg"},
	{0x6B94866B86162974ULL, "6072706_m.fxe"},
	{0x6B9490351B3BF678ULL, "6151474_m.fxe"},
	{0x6B94CEE0A300B7A4ULL, "w_sundermount_84_0.phy"},
	{0x6B94E11D12CC9454ULL, "874167971.ogg"},
	{0x6B95082D731787D1ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masb.phy"},
	{0x6B95093B84AEF94CULL, "mcr370_meredith6084407.cl"},
	{0x6B952319D1F3659DULL, "qcr220_meredith6060318.cl"},
	{0x6B953D7D91E70A00ULL, "vxw_sundermount_2tk_0_0.vlm"},
	{0x6B9583C8DFF2A0BCULL, "6091879-guardsdragaway.cut"},
	{0x6B95E260DA56C75CULL, "6141835_m.fxe"},
	{0x6B9623C239640F1DULL, "6023972_m.fxe"},
	{0x6B964D67460DA38DULL, "512858985.ogg"},
	{0x6B96D52C0EDD3A98ULL, "121036036.ogg"},
	{0x6B96D8EB9CDE36B8ULL, "lowtown_argument6169420.cl"},
	{0x6B973C2186B33D9BULL, "547189580.ogg"},
	{0x6B97E3B917D1B7BDULL, "vxht_keep_n_6h4_0_1.vlm"},
	{0x6B987EB9613E7274ULL, "vxlt_gallowscourtyard_n_4ut_0_0.vlm"},
	{0x6B9895915D93C156ULL, "vxw_woundedcoast_ni_0_0.vlm"},
	{0x6B99054743A27CEDULL, "6117953_f.fxe"},
	{0x6B9922006114AC0EULL, "zz_firestaff_2_v2.cut"},
	{0x6B9923DA8D4A6929ULL, "6105430_m.fxe"},
	{0x6B992EFAC81F57A8ULL, "vxlt_gallowstemplar_f_5qg_0_0.vlm"},
	{0x6B9970859B2408BCULL, "259917641.ogg"},
	{0x6B99CC4902212BD8ULL, "361035192.ogg"},
	{0x6B99D15F652AF3EFULL, "vxlt_gallowstemplar_n_2pd_0_0.vlm"},
	{0x6B9A1385D30BD0C3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v3.gad"},
	{0x6B9AB4013B1E7D1BULL, "plt_lgt110_gxa_vgk_consprcy.nss"},
	{0x6B9ABCC67F32A555ULL, "vxlt_gallowscourtyard_n_3vs_0_0.vlm"},
	{0x6B9AE8A567877085ULL, "hlac_w_bonepit_262150_0y.dds"},
	{0x6B9BC5E0A306C72CULL, "w_sundermount_84_0.rml"},
	{0x6B9BFB1371A2B949ULL, "llac_lt_foundry_n_320_17.dds"},
	{0x6B9C0BCD3311284EULL, "hlac_w_sundermount_-327677_0z.dds"},
	{0x6B9C7EE6E6F8A269ULL, "hlac_lt_0_lowtown_docks_-131073_0y.dds"},
	{0x6B9C7F576A0351ADULL, "hlac_lt_0_lowtown_docks_n_131070_4x.dds"},
	{0x6B9C8BA04C6AC70DULL, "dae120ip_chant_board_ht.utp"},
	{0x6B9CBCDE78C2AA6AULL, "vxdr_primevaltomb_779_0_1.vlm"},
	{0x6B9CE7F58672C527ULL, "hlac_lt_darktown_-1_7x.dds"},
	{0x6B9D00C07D2B2733ULL, "99777978.ogg"},
	{0x6B9D0F2133E81C1BULL, "vxht_0_hightown_a3_3w1_0_0.vlm"},
	{0x6B9D81DB2885F81CULL, "vxht_0_hightown_a2_c0z_0_0.vlm"},
	{0x6B9DC5E51836F986ULL, "6027841_m.fxe"},
	{0x6B9DD4E3C69F2C6AULL, "169650717.ogg"},
	{0x6B9E268BDE8B025DULL, "vxht_chantry_26x_0_0.vlm"},
	{0x6B9E43ABFC6D3C92ULL, "art\\characters\\creatures\\mubari\\animation\\rst_p.gad"},
	{0x6B9E647845648092ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.anb"},
	{0x6B9ECD77CB0E00BBULL, "follower_banter6148458.cl"},
	{0x6B9F1DD84E49BB5DULL, "follower_barks6153161.cl"},
	{0x6B9FB7817F281D8FULL, "6030673_m.fxe"},
	{0x6B9FF70DB3FEA0B3ULL, "hlac_lt_undercitydungeon_-1_46x.dds"},
	{0x6BA0A74E9F4FA5D6ULL, "vxht_0_hightown_a2_c79_0_0.vlm"},
	{0x6BA0C66B42A3A349ULL, "vxlt_undercityserialkiller_1jz_0_0.vlm"},
	{0x6BA0D498506B6FA8ULL, "dae000_meredith6092561.cl"},
	{0x6BA0F085D311B3A8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v3.evt"},
	{0x6BA1182F8A52E0C4ULL, "6054969_f.fxe"},
	{0x6BA1EF4E67D52A73ULL, "lowtown_male_lowlife1_sta6165026.cl"},
	{0x6BA2135BB77EA9A6ULL, "127469731.ogg"},
	{0x6BA215B559125324ULL, "337408197.ogg"},
	{0x6BA23A21967DB112ULL, "6085855_f.fxe"},
	{0x6BA2C5570DFB7F58ULL, "311042429.ogg"},
	{0x6BA2F95BB32F22F8ULL, "851918505.ogg"},
	{0x6BA327F00E8F3B86ULL, "gen_im_trs_dwv_junk1_01.uti"},
	{0x6BA3AF5A83DD2280ULL, "vxw_sundermount_2yh_0_0.vlm"},
	{0x6BA3FA1334241FC7ULL, "6207992_m.fxe"},
	{0x6BA3FB6C467B68F4ULL, "art\\vfx\\combat\\vfx_g_spider_att_f2_c.impa.ani"},
	{0x6BA4BF696C1340C8ULL, "vxdr_0_deeproads_cwa_0_1.vlm"},
	{0x6BA5058C6704C263ULL, "llac_lt_0_lowtown_docks_a3_-196611_1.dds"},
	{0x6BA546ABFC7360C9ULL, "art\\characters\\creatures\\mubari\\animation\\rst_p.evt"},
	{0x6BA5B2280748AADCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_2_ian.ani"},
	{0x6BA5E2F8C8AEDF32ULL, "6042754_m.fxe"},
	{0x6BA5E8576FA97D8DULL, "vxht_0_hightown_a3_7fx_0_0.vlm"},
	{0x6BA6491E39CA7FAEULL, "gen00fl_anders6121169.cl"},
	{0x6BA67E2E8791D761ULL, "gen00fl_aveline6121916.cl"},
	{0x6BA6CC04C334F1EBULL, "336606843.ogg"},
	{0x6BA736529AB36CA3ULL, "6060539_m.fxe"},
	{0x6BA7823671734AABULL, "6042878_m.fxe"},
	{0x6BA8DF99FE5A39FCULL, "vxca_cavern1_haunted_j2p_0_0.vlm"},
	{0x6BA9F86C467FCF56ULL, "art\\vfx\\combat\\vfx_g_spider_att_f2_c.impa.gad"},
	{0x6BAA4F78B27A5AA4ULL, "6153564_m.fxe"},
	{0x6BAA50EC66165A76ULL, "mag311_emile_de_launcet6043625.cl"},
	{0x6BAAB3C15A5F200BULL, "vxw_woundedcoast_1fe_0_0.vlm"},
	{0x6BAAD5753A45504DULL, "706337234.ogg"},
	{0x6BAAEF96E6452122ULL, "714457464.ogg"},
	{0x6BAB053D3AB34C3EULL, "44796028.ogg"},
	{0x6BAB16D922CD0FD1ULL, "art\\levels\\ztools\\default\\generic\\9ff_signmarker.mmh"},
	{0x6BAB1ED5D02620D6ULL, "6188575_m.fxe"},
	{0x6BAB22D922CD2472ULL, "art\\levels\\ztools\\default\\generic\\9ff_signmarker.mao"},
	{0x6BAB30D922CD3B8FULL, "art\\levels\\ztools\\default\\generic\\9ff_signmarker.msh"},
	{0x6BAB404E7429E1FAULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkf.gad"},
	{0x6BAB9D24DD69C2F6ULL, "vxw_blightlands_hq_0_0.vlm"},
	{0x6BABBF28074D2C1EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_2_ian.gad"},
	{0x6BAC0AD054A0D5E7ULL, "vrc221_bartrand6044972.cl"},
	{0x6BAC2510A9D57CFEULL, "dre000_varric6035184.cl"},
	{0x6BAC2FC9EABE487CULL, "6214656_m.fxe"},
	{0x6BAC9868D0811ADDULL, "vxht_chantry_23w_0_0.vlm"},
	{0x6BACAD386AE1789AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel.mmh"},
	{0x6BACB1386AE17FD1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel.mao"},
	{0x6BACB97441EA3788ULL, "vxht_chantry_n_2lh_0_0.vlm"},
	{0x6BACBF386AE197E0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel.msh"},
	{0x6BAD06B4E3815624ULL, "6166443_m.fxe"},
	{0x6BADD7BCABA2B463ULL, "gen00fl_merrill6129847.cl"},
	{0x6BAE1FDC38E03D63ULL, "w_sundermount_183_0.msh"},
	{0x6BAE2DDC38E054BDULL, "w_sundermount_183_0.mmh"},
	{0x6BAE2E0DECBDBEFEULL, "vxht_0_hightown_bun_0_0.vlm"},
	{0x6BAE31DC38E05BCEULL, "w_sundermount_183_0.mao"},
	{0x6BAE50A6825B25ACULL, "vxw_encounter2_camp_4t_0_0.vlm"},
	{0x6BAEE92C14A242C2ULL, "56424245.ogg"},
	{0x6BAF03A149B7C989ULL, "6049136_m.fxe"},
	{0x6BAF0777E8B94383ULL, "6120415_m.fxe"},
	{0x6BAF199D84415A51ULL, "1042806480.ogg"},
	{0x6BAF38483499BBC9ULL, "vxlt_undercityserialkiller_1oy_0_0.vlm"},
	{0x6BB018670F819E3AULL, "vxca_caverns1_ehn_0_0.vlm"},
	{0x6BB027B33BD15EB2ULL, "339268070.ogg"},
	{0x6BB06FC23DCDDFAEULL, "6106692_m.fxe"},
	{0x6BB0CA64CDA3EE14ULL, "vxlt_hangedman_281_0_0.vlm"},
	{0x6BB0CC202AC70A44ULL, "vxca_cave_1ch_0_1.vlm"},
	{0x6BB0F1664684B7EBULL, "mcr310_argument6075810.cl"},
	{0x6BB1136C46861BB5ULL, "art\\vfx\\combat\\vfx_g_spider_att_f2_c.impa.evt"},
	{0x6BB116788732F16EULL, "6147167_m.fxe"},
	{0x6BB1434E742E5211ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkf.evt"},
	{0x6BB1FF7F0A3C2112ULL, "410243284.ogg"},
	{0x6BB2AE3706884770ULL, "6164733_m.fxe"},
	{0x6BB32EABFC7F7538ULL, "art\\characters\\creatures\\mubari\\animation\\rst_p.ani"},
	{0x6BB3B6AEBF7D90B5ULL, "vxht_0_hightown_n_ajx_0_0.vlm"},
	{0x6BB3D509B7550C86ULL, "vxlt_gallowstemplar_5bs_0_0.vlm"},
	{0x6BB406A1BFCE7567ULL, "character_character_p260.psh.fxo.dx11"},
	{0x6BB41D59D53D5B25ULL, "t3_brt_b01.tnt"},
	{0x6BB564160CBB50C2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_roof4ocextru.msh"},
	{0x6BB566160CBB5478ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_roof4ocextru.mmh"},
	{0x6BB572160CBB681BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_roof4ocextru.mao"},
	{0x6BB58E9484556703ULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench3_sebastianintro.ani"},
	{0x6BB5CD74E5A183ADULL, "vxht_0_hightown_b7c_0_0.vlm"},
	{0x6BB5F2A70EF88C62ULL, "cli_army_ds_alienage.gda"},
	{0x6BB6402DA534554BULL, "397710776.ogg"},
	{0x6BB6906C87CAF87FULL, "vxlt_undercityencounter_1os_0_0.vlm"},
	{0x6BB6942C9A49BAC0ULL, "6147774_m.fxe"},
	{0x6BB6A0B51568185EULL, "6084005_f.fxe"},
	{0x6BB6A28BA775A5D7ULL, "hlac_w_encounter2_-65537_0z.dds"},
	{0x6BB6DC8FCC3B067EULL, "vxlt_gallowstemplar_5jc_0_0.vlm"},
	{0x6BB76A359A893601ULL, "hlac_ht_mansion2_818_63x.dds"},
	{0x6BB76DE33E12D1F4ULL, "pro000_cut2d_real_story6001253.cl"},
	{0x6BB7FA0D6555B2A9ULL, "mag100_cullen6082322.cl"},
	{0x6BB87452ECBEB5BBULL, "6087459_m.fxe"},
	{0x6BB9259E4C55388BULL, "vxw_woundedcoast_1ad_0_0.vlm"},
	{0x6BB9359AF7894395ULL, "6016482_m.fxe"},
	{0x6BBA01DA403D51B3ULL, "gen_im_trs_rdr_med_cht_r2.uti"},
	{0x6BBA0F01CF5FDB76ULL, "vxw_blightlands_mr_0_0.vlm"},
	{0x6BBB76B3D98D6ABFULL, "6139700_f.fxe"},
	{0x6BBBC10945D0CBF1ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbed.mmh"},
	{0x6BBBCD0945D0DF92ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbed.mao"},
	{0x6BBBDB0945D0F7AFULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzbed.msh"},
	{0x6BBBE1DF920CAA5CULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredlong_n.dds"},
	{0x6BBC3478B1464F65ULL, "92713528.ogg"},
	{0x6BBD0668496E94DBULL, "152263389.ogg"},
	{0x6BBD7132F6F8C4E3ULL, "939706346.ogg"},
	{0x6BBDA9BE6C2FF635ULL, "320032254.ogg"},
	{0x6BBDC4C7857F6764ULL, "w_bonepitmining_a2_147_0.mmh"},
	{0x6BBDC8C7857F6DB7ULL, "w_bonepitmining_a2_147_0.mao"},
	{0x6BBDD86932597C85ULL, "lgt110_gxa_vgk_silver_order6116997.cl"},
	{0x6BBDDAC7857F8C06ULL, "w_bonepitmining_a2_147_0.msh"},
	{0x6BBE345E76E9712EULL, "365232980.ogg"},
	{0x6BBE4B9181A94E29ULL, "708899031.ogg"},
	{0x6BBE5D51589248E4ULL, "6110730_m.fxe"},
	{0x6BBEA5F4A12EFAECULL, "vxlt_0_lowtown_n_e18_0_0.vlm"},
	{0x6BBEC44A2BF014D9ULL, "6066315_m.fxe"},
	{0x6BBF17B8AFC60527ULL, "vxlt_0_lowtown_a3_n_c_era_0_0.vlm"},
	{0x6BBF3C41BF9A0694ULL, "vxlt_hangedman_256_0_0.vlm"},
	{0x6BBF5B4E743AB8E0ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkf.ani"},
	{0x6BBF6D160CC3B403ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0co_roof4ocextru.phy"},
	{0x6BBF86861098419AULL, "fen110_fenris6118113.cl"},
	{0x6BBFA87CD9AB5B79ULL, "vxht_mansion2_1vs_0_0.vlm"},
	{0x6BBFDBB13B4753DDULL, "dae300_cullen6140650.cl"},
	{0x6BC02A9811E8B953ULL, "782792246.ogg"},
	{0x6BC04BE31140AFBFULL, "dae000_maren6205671.cl"},
	{0x6BC1FCE80F19CAE2ULL, "follower_banter6161139.cl"},
	{0x6BC246E6A94B2506ULL, "vxlt_gallowstemplar_5ot_0_0.vlm"},
	{0x6BC292671D7454CEULL, "vxw_bonepitdestroyed_a3_ck_0_0.vlm"},
	{0x6BC2AE2B59EC931CULL, "vxca_caverns2_4ss_0_0.vlm"},
	{0x6BC2F47945C5A0F9ULL, "vxca_caverns1_mines_hin_0_0.vlm"},
	{0x6BC306F1371EB008ULL, "hlac_lt_0_lowtown_docks_n_-2_0x.dds"},
	{0x6BC3107987236332ULL, "one300_zevran6089739.cl"},
	{0x6BC3A1FB7CA1D235ULL, "art\\characters\\playercharacter\\humanmale\\boo_wmda.phy"},
	{0x6BC3E2E0914E629CULL, "vxw_bonepit_1p3_0_0.vlm"},
	{0x6BC423E8952B0050ULL, "6164311_m.fxe"},
	{0x6BC46199631BD8AFULL, "6129122_m.fxe"},
	{0x6BC47E7BA312262BULL, "art\\characters\\creatures\\gibgolem\\hed_gib.mao"},
	{0x6BC4D6A3F802A17CULL, "dae300_bran6143498.cl"},
	{0x6BC4FDA126D37EB1ULL, "gen00fl_aveline6114417.cl"},
	{0x6BC5024979C110FFULL, "vxlt_undercityencounter_1dr_0_0.vlm"},
	{0x6BC5382EB143CC61ULL, "6134584_m.fxe"},
	{0x6BC5A03C63B38A31ULL, "117275543.ogg"},
	{0x6BC695C84A6E5B99ULL, "hlac_w_encounter3_-196604_0y.dds"},
	{0x6BC6D5114C17EB81ULL, "427145048.ogg"},
	{0x6BC6F1AC3079648BULL, "gen00fl_aveline6128830.cl"},
	{0x6BC71CA65B544B98ULL, "vxw_encounter3_7z_0_0.vlm"},
	{0x6BC72909156652DAULL, "6016584_m.fxe"},
	{0x6BC736962251B938ULL, "777360114.ogg"},
	{0x6BC7977B3E4B510BULL, "vxw_woundedcoast_1lg_0_0.vlm"},
	{0x6BC7F23726E99F84ULL, "vxht_bartrand_n_6qx_0_0.vlm"},
	{0x6BC880DEC155F3F6ULL, "vxw_blightlands_vs_0_0.vlm"},
	{0x6BC8B49514CB29D4ULL, "wm_ico_keep.dds"},
	{0x6BC8F645DB279E19ULL, "lowtown_male_lowlife1_sta6165019.cl"},
	{0x6BC93743549570A0ULL, "song_dis.mao"},
	{0x6BC9A594846637EDULL, "art\\cinematics\\follower plots\\sebastian\\cs_hench3_sebastianintro.gad"},
	{0x6BC9BF013B4600B4ULL, "plt_lgt110_gxa_vgk_consprcy.plo"},
	{0x6BCA2E0BC5B22C77ULL, "atl_armycontro_dxt5_00.dds"},
	{0x6BCA8A664D39C8AAULL, "6170125_m.fxe"},
	{0x6BCA9767565E52FBULL, "493525009.ogg"},
	{0x6BCAB3CCAD3D0A45ULL, "w_bonepitmining_a2_11_0_ll.mmh"},
	{0x6BCABA4A1DC362F2ULL, "6178389_m.fxe"},
	{0x6BCAC4423C9EF523ULL, "hlac_lt_gallowscourtyard_n_-65538_2z.dds"},
	{0x6BCADFABD5375EBAULL, "ave331cr_insurgent_guard3.utc"},
	{0x6BCAE0466983B27BULL, "6067821_f.fxe"},
	{0x6BCAFA92173D476EULL, "hlac_lt_0_lowtown_n_-3_4x.dds"},
	{0x6BCB7539CE1C427BULL, "494869625.ogg"},
	{0x6BCBB24C43D74569ULL, "dae000_gamlen6165969.cl"},
	{0x6BCC3B049BE6F484ULL, "857901941.ogg"},
	{0x6BCC6C3B4E0A4280ULL, "1040400867.ogg"},
	{0x6BCDAE1EB1901F14ULL, "vxlt_hangedman_227_0_0.vlm"},
	{0x6BCDE060EB21F5E6ULL, "192787680.ogg"},
	{0x6BCE0B6F4F4E2062ULL, "176177852.ogg"},
	{0x6BCE1A59CBA173F9ULL, "vxht_mansion2_1yr_0_0.vlm"},
	{0x6BCE1AA08BE189ABULL, "llac_lt_0_lowtown_a2_n_c_0_7.dds"},
	{0x6BCE1E9F05CC24C8ULL, "hlac_ht_mansion2_818_15x.dds"},
	{0x6BCE8DAF55DB682DULL, "6208601_m.fxe"},
	{0x6BCEC8ABB976C530ULL, "8892616.ogg"},
	{0x6BCEF744CB40D86DULL, "vxlt_0_lowtown_a3_n_c_hvf_0_0.vlm"},
	{0x6BCF4362A6627BF1ULL, "vxca_caverns1_e7t_0_0.vlm"},
	{0x6BCF7371A5B07241ULL, "6120149_m.fxe"},
	{0x6BCF7DF2A841D369ULL, "mer221_sabin6065656.cl"},
	{0x6BCFEEF59618C5F5ULL, "lgt110_arl_bann_teagan6115558.cl"},
	{0x6BD0A90AA4CA845CULL, "vxw_sundermount_2e0_0_0.vlm"},
	{0x6BD123F2BE7C784AULL, "art\\vfx\\materials\\vfx_twirl.mao"},
	{0x6BD1925C3D7C00FEULL, "hlac_w_encounter1_0_1z.dds"},
	{0x6BD19712DF13052AULL, "457480932.ogg"},
	{0x6BD1BE5E5BDEEB47ULL, "61865115.ogg"},
	{0x6BD1CC386B00C9A5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel.phy"},
	{0x6BD25D05256C2356ULL, "vxw_woundedcoast_zm_0_0.vlm"},
	{0x6BD2EEAE160AC70BULL, "364442952.ogg"},
	{0x6BD3409FEC6DD270ULL, "hlac_dr_dungeon_3214_13x.dds"},
	{0x6BD387D593AB254AULL, "463341849.ogg"},
	{0x6BD3C049B027377EULL, "vxlt_gallowstemplar_5pe_0_0.vlm"},
	{0x6BD4FEC1345E0B49ULL, "6119545_m.fxe"},
	{0x6BD5174A0E10173EULL, "6107000_m.fxe"},
	{0x6BD52C676462BD25ULL, "258562263.ogg"},
	{0x6BD56D7E30C9620DULL, "117888088.ogg"},
	{0x6BD57A30FA0E28D6ULL, "vxw_bonepit_1vc_0_0.vlm"},
	{0x6BD5FB350E786D41ULL, "953696845.ogg"},
	{0x6BD657B25A921B17ULL, "6037001_m.fxe"},
	{0x6BD6856EDED47F76ULL, "hlac_lt_warehouse_n_-65536_2x.dds"},
	{0x6BD6FF2887379F3AULL, "vxw_bonepit_16o_0_0.vlm"},
	{0x6BD78C3259FD5030ULL, "vxw_woundedcoast_13e_0_0.vlm"},
	{0x6BD799D694358981ULL, "387642368.ogg"},
	{0x6BD79BF69EC2013EULL, "hlac_lt_0_lowtown_docks_n_-131074_8z.dds"},
	{0x6BD8074A3991E0D0ULL, "6103235_m.fxe"},
	{0x6BD85CD9E15A790AULL, "12752986.ogg"},
	{0x6BD8CC27E748B833ULL, "6071054_m.fxe"},
	{0x6BD8D03863E1831AULL, "6070140_m.fxe"},
	{0x6BD901368F515966ULL, "vxht_brothel_n_3hh_0_0.vlm"},
	{0x6BD926A520A1DF92ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_enter.gad"},
	{0x6BD9284CB120D899ULL, "zz_eroc.cut"},
	{0x6BD98031ACEE4915ULL, "6004229_m.fxe"},
	{0x6BD9B49A81243A67ULL, "vxlt_0_lowtown_docks_a3_n_d2u_0_0.vlm"},
	{0x6BDA0B4204DA6C02ULL, "679181978.ogg"},
	{0x6BDA53F5BB5BC57CULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_0.pwk"},
	{0x6BDA5503F1E69131ULL, "6095347_f.fxe"},
	{0x6BDA9735079A2F0AULL, "264892056.ogg"},
	{0x6BDADE92BF9167CCULL, "qun120_arishok6052910.cl"},
	{0x6BDB22D4490D4AB4ULL, "hlac_lt_0_lowtown_a2_n_c_131070_4x.dds"},
	{0x6BDC0C1451CB4C71ULL, "vxlt_backalley_g_1u7_0_0.vlm"},
	{0x6BDCCD261B529718ULL, "6111074_m.fxe"},
	{0x6BDD6921BD36F0EDULL, "vxlt_0_lowtown_a3_n_c_hse_0_0.vlm"},
	{0x6BDDA03591E1729CULL, "6132745_m.fxe"},
	{0x6BDDD13F9FB290E3ULL, "6164271_m.fxe"},
	{0x6BDDD7A600313E50ULL, "vxht_0_hightown_a3_c2q_0_0.vlm"},
	{0x6BDE417FF1C04509ULL, "vxlt_backalley_n_1nw_0_0.vlm"},
	{0x6BDE4BFB0985FAE0ULL, "805416860.ogg"},
	{0x6BDEC17812F5D83DULL, "vxw_sundermount_350_0_0.vlm"},
	{0x6BDECF6C9D3E6A28ULL, "isa321_velasco6117620.cl"},
	{0x6BDF1AE796C09CDCULL, "vxw_sundermount_2n1_0_0.vlm"},
	{0x6BDF29EDE704309AULL, "zz_approval_debug6003083.cl"},
	{0x6BDF2AA08D375606ULL, "vxlt_gallowstemplar_5uv_0_0.vlm"},
	{0x6BE029A520A803C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_enter.evt"},
	{0x6BE07589960A97A2ULL, "hlac_lt_0_lowtown_docks_n_-131076_0x.dds"},
	{0x6BE0CEE217623BD6ULL, "vxw_woundedcoast_wn_0_0.vlm"},
	{0x6BE12FCFC967AED3ULL, "694416948.ogg"},
	{0x6BE1C273A5CC2860ULL, "vxca_caverns2_8kr_0_1.vlm"},
	{0x6BE1CF6989A763F9ULL, "vxlt_0_lowtown_docks_a3_n_bs2_0_0.vlm"},
	{0x6BE2539592F3B33CULL, "qun221_viscount6084717.cl"},
	{0x6BE253D9A48BB23AULL, "vxw_woundedcoast_7b_0_0.vlm"},
	{0x6BE268730B6E4403ULL, "art\\vfx\\environment\\vfx_trap_radialblade_p.impa.ani"},
	{0x6BE37B206470E417ULL, "271104558.ogg"},
	{0x6BE38836226FB07CULL, "mrl000_marethari6129261.cl"},
	{0x6BE3D3D616B53FD7ULL, "881387464.ogg"},
	{0x6BE3E17A81D508E3ULL, "hlac_lt_0_lowtown_docks_a3_-131075_1y.dds"},
	{0x6BE3E9C0CDC9AEAEULL, "vxlt_0_lowtown_n_ey2_0_0.vlm"},
	{0x6BE4575F4034BA61ULL, "6031749_f.fxe"},
	{0x6BE46F7BF90040ABULL, "348703677.ogg"},
	{0x6BE47B352237820BULL, "vxw_woundedcoast_1ry_0_0.vlm"},
	{0x6BE4EDFB9FB02706ULL, "vxht_0_hightown_byr_0_0.vlm"},
	{0x6BE524484F9C6DFAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\enter_walk_female.gad"},
	{0x6BE539CA88675317ULL, "mag320_mage_conspirator6040830.cl"},
	{0x6BE53EDBAEA72B63ULL, "hlac_lt_undercityencounter_-1_6y.dds"},
	{0x6BE5848A3896113EULL, "119667004.ogg"},
	{0x6BE58BE26D355D06ULL, "follower_banter6169171.cl"},
	{0x6BE5EBC80EC81AAAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_ian.gad"},
	{0x6BE61B87988DD791ULL, "401362382.ogg"},
	{0x6BE640386965D67BULL, "558018141.ogg"},
	{0x6BE6E522C3120EDFULL, "zz_isa_debug6133167.cl"},
	{0x6BE72566AA79CF1CULL, "hlac_dr_primevaltomb_657_2z.dds"},
	{0x6BE7421555D79B9AULL, "6075240_m.fxe"},
	{0x6BE7C55D2C5BB508ULL, "gen00fl_fenris6103349.cl"},
	{0x6BE7E66AE3F54398ULL, "6148326_m.fxe"},
	{0x6BE86929738D4569ULL, "one100_solivitus6065675.cl"},
	{0x6BE8AFA92955AD59ULL, "seb321_leliana6181750.cl"},
	{0x6BE8BB16C4719B60ULL, "mag210_arianni6024577.cl"},
	{0x6BE8E2315A9C6D53ULL, "follower_banter6143356.cl"},
	{0x6BE8ED0C33FC8F0CULL, "gen00fl_merrill6108490.cl"},
	{0x6BE93C9724B95347ULL, "vxlt_0_lowtown_h91_0_0.vlm"},
	{0x6BE94B1E79164F2EULL, "zz_qcr_debug6045149.cl"},
	{0x6BE95D98A5A31D3FULL, "vxlt_0_lowtown_n_e17_0_0.vlm"},
	{0x6BEA7DF143C164F1ULL, "vxlt_backalley_g_1r6_0_0.vlm"},
	{0x6BEA8C9E0B823A1EULL, "vrc221ip_at_hightown.utp"},
	{0x6BEAC49F2E45D9E9ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6.phy"},
	{0x6BEAFE13AF8DA4F9ULL, "vxht_mansion2_1gp_0_0.vlm"},
	{0x6BEB083F8A255333ULL, "vxlt_undercityencounter_1ln_0_0.vlm"},
	{0x6BEB27484FA0DE11ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\enter_walk_female.evt"},
	{0x6BEB58C1C4EE4029ULL, "hlac_ht_mansion2_818_21z.dds"},
	{0x6BEB69805738298DULL, "752672714.ogg"},
	{0x6BEBAAE55F63D55BULL, "one220_dera.bnk"},
	{0x6BEBCC21E075DCFAULL, "6025999_m.fxe"},
	{0x6BEC5362E69B89A6ULL, "hlac_w_blightlands_65538_0z.dds"},
	{0x6BEC6270EC2DAFFCULL, "824179162.ogg"},
	{0x6BED8CC488B6B55CULL, "vxw_sundermount_2s6_0_0.vlm"},
	{0x6BEDA798C6CBA62BULL, "6208478_m.fxe"},
	{0x6BEDBE446B7DC881ULL, "6069376_m.fxe"},
	{0x6BEE03235F5DAD74ULL, "560297647.ogg"},
	{0x6BEE11A520B41838ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_enter.ani"},
	{0x6BEE7417957FAF4DULL, "6092073_m.fxe"},
	{0x6BEE9AF0F0099E4DULL, "239780778.ogg"},
	{0x6BEEB304D27DE92EULL, "llac_lt_0_lowtown_docks_a3_n_-4_5.dds"},
	{0x6BEF80730B78F7FAULL, "art\\vfx\\environment\\vfx_trap_radialblade_p.impa.evt"},
	{0x6BEF9C56E872B24FULL, "6095972_f.fxe"},
	{0x6BEFACE55F67C11BULL, "one220_dera.cnv"},
	{0x6BF003C88D321F93ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.impa.evt"},
	{0x6BF008E4B393CA67ULL, "vfx_flys.wbd"},
	{0x6BF043EFE3E887C6ULL, "mag100_karras6034743.cl"},
	{0x6BF06245B2C8D31BULL, "gen_im_cns_cft_wrn_03.uti"},
	{0x6BF1161B9434F910ULL, "llac_dr_slavers_3214_13.dds"},
	{0x6BF14A66279016C5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wrba_0t.dds"},
	{0x6BF1CB64CADA98FFULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpb.mmh"},
	{0x6BF1D164CADAA2A1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpb.msh"},
	{0x6BF1E27B184A3C49ULL, "6113345_m.fxe"},
	{0x6BF1FDC4BE5DB85BULL, "636428376.ogg"},
	{0x6BF24CA3D08A1ACAULL, "669533961.ogg"},
	{0x6BF2505EDCB6E715ULL, "hlac_lt_undercityserialkiller_-1_18x.dds"},
	{0x6BF25B9DBFBFC72EULL, "vxlt_0_lowtown_n_er3_0_0.vlm"},
	{0x6BF2CEEC55F27596ULL, "6016508_m.fxe"},
	{0x6BF3131A4C448581ULL, "vxw_sundermount_2ea_0_0.vlm"},
	{0x6BF341CEDA62603DULL, "ht_0_hightown_a2_17_0_lc.mmh"},
	{0x6BF38C988516C72DULL, "vxht_brothel_n_4h8_0_0.vlm"},
	{0x6BF3E81A4C258F60ULL, "hlac_ht_0_hightown_n_-65534_1y.dds"},
	{0x6BF4094F1317AE37ULL, "715790143.ogg"},
	{0x6BF43EE444833B66ULL, "1042923117.ogg"},
	{0x6BF46FEC3DE98130ULL, "vxw_woundedcoast_19c_0_0.vlm"},
	{0x6BF4A192271894B7ULL, "6086261_m.fxe"},
	{0x6BF4D3CB459B81A3ULL, "446970580.ogg"},
	{0x6BF57DC599BCC917ULL, "118978463.ogg"},
	{0x6BF5B86637A22FD8ULL, "6116365_m.fxe"},
	{0x6BF5FCC88D367FD4ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.impa.gad"},
	{0x6BF6128290590857ULL, "6116146_m.fxe"},
	{0x6BF67F730B7F14EDULL, "art\\vfx\\environment\\vfx_trap_radialblade_p.impa.gad"},
	{0x6BF6D37159923CEEULL, "gen00fl_merrill6111274.cl"},
	{0x6BF72324697B6CF8ULL, "63670846.ogg"},
	{0x6BF7AA6C50C3BCECULL, "60372145.ogg"},
	{0x6BF7C26EC427B2F1ULL, "vxw_bonepitmining_a2_1hx_0_0.vlm"},
	{0x6BF82C7324259D74ULL, "vxca_cavern1_haunted_k2o_0_2.vlm"},
	{0x6BF863690CD51344ULL, "6096201_m.fxe"},
	{0x6BF89C5B3C9B9B15ULL, "171404000.ogg"},
	{0x6BF8EACB73704D9AULL, "hlac_ht_0_hightown_a3_-131072_0y.dds"},
	{0x6BF93F484FAD44E0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\enter_walk_female.ani"},
	{0x6BF95560ACCF08B6ULL, "6127517_f.fxe"},
	{0x6BF993AB239A52AAULL, "w_encounter2_18_0.mao"},
	{0x6BF997AB239A59F9ULL, "w_encounter2_18_0.mmh"},
	{0x6BF9A1AB239A6AE7ULL, "w_encounter2_18_0.msh"},
	{0x6BF9B00D8601BA19ULL, "6170089_m.fxe"},
	{0x6BF9C6C80ED88490ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_ian.ani"},
	{0x6BF9D56810E2089BULL, "plt_qun200pt_following_qun.nss"},
	{0x6BFA9B5234A6D447ULL, "vxlt_0_lowtown_eqh_0_0.vlm"},
	{0x6BFB5354B4AA1202ULL, "594920957.ogg"},
	{0x6BFC903358F4C450ULL, "dae011_conspiracy_mob6149584.cl"},
	{0x6BFCAAD4D8B75B10ULL, "vxlt_warehouse_15m_0_0.vlm"},
	{0x6BFD0BC88D3CB7B2ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.impa.ani"},
	{0x6BFD4D6AD27F2D95ULL, "431532861.ogg"},
	{0x6BFD852204EDC7D0ULL, "cd_one100dg_kelder.bnk"},
	{0x6BFD9BB90517E6E7ULL, "6027725_m.fxe"},
	{0x6BFDD35B7DC9E2F3ULL, "288272910.ogg"},
	{0x6BFE3590E3D0292DULL, "and230_mage_girl6109115.cl"},
	{0x6BFEAFF7854238CEULL, "574747060.ogg"},
	{0x6BFF853E74B9D3F7ULL, "vxlt_undercitydungeon_1fm_0_0.vlm"},
	{0x6BFFD9F5B7E7B8A7ULL, "249492757.ogg"},
	{0x6C00B79887476021ULL, "gen_im_trs_gen_lgt_hlm_r1.uti"},
	{0x6C010C910FB2F59DULL, "6113944_f.fxe"},
	{0x6C01105B9DF539E3ULL, "vxdr_slavers_20y_0_0.vlm"},
	{0x6C020B5BFC6837EFULL, "6119446_m.fxe"},
	{0x6C02487E4E99F149ULL, "dae000ip_crowd_ff5_talk3.utp"},
	{0x6C02725BC56D0123ULL, "417708749.ogg"},
	{0x6C028BB27BF3FED5ULL, "framebuffereffects_cocdownsample_p0.vsh.fxo.dx11"},
	{0x6C0378D8BCA7E1F6ULL, "513958100.ogg"},
	{0x6C03841E078D0036ULL, "llac_lt_0_lowtown_a2_-196609_1.dds"},
	{0x6C03A75F9340A9D3ULL, "vxlt_blackemporium_uc_0_0.vlm"},
	{0x6C0482FBCA534825ULL, "mag110_arianni6016839.cl"},
	{0x6C04B6DBBF4BA884ULL, "mag311_emile_de_launcet6043713.cl"},
	{0x6C05213D4981CCFEULL, "vxlt_0_lowtown_a2_hi0_0_0.vlm"},
	{0x6C060006176B8148ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopleft.msh"},
	{0x6C0602F293B02A4CULL, "gen00fl_merrill6127422.cl"},
	{0x6C061206176B9FD9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopleft.mao"},
	{0x6C061E06176BB472ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatetopleft.mmh"},
	{0x6C0789280F22CAADULL, "vxdr_primevaltomb_6qp_0_0.vlm"},
	{0x6C07AB81AC97C06DULL, "static_static_p46.psh.fxo.dx11"},
	{0x6C082EBB20BD84B1ULL, "qcr000_arishok6059681.cl"},
	{0x6C09114D5C487E36ULL, "vxht_mansion2b_2yj_0_0.vlm"},
	{0x6C091A8D0645BA8AULL, "qun221_viscount6084728.cl"},
	{0x6C094D824CFB9736ULL, "336150541.ogg"},
	{0x6C096785512E36E6ULL, "hairalpha_hair_p1.psh.fxo.dx11.assembly"},
	{0x6C097FBF18B3A433ULL, "rdr111dg_martin6057078.cl"},
	{0x6C0AF9CBBA8FF2C0ULL, "gen_im_trs_drk_junk1_01.uti"},
	{0x6C0B0B2F55FB1961ULL, "gen_im_cft_reg_c.uti"},
	{0x6C0B94D5BAA1D385ULL, "dae100_bran6142997.cl"},
	{0x6C0BB4FB7CDFA850ULL, "art\\characters\\playercharacter\\humanmale\\boo_wmda.msh"},
	{0x6C0BC2FB7CDFC06AULL, "art\\characters\\playercharacter\\humanmale\\boo_wmda.mmh"},
	{0x6C0BC6FB7CDFC6A1ULL, "art\\characters\\playercharacter\\humanmale\\boo_wmda.mao"},
	{0x6C0C04B99CAF7EB2ULL, "6168159_m.fxe"},
	{0x6C0C1C314B1CDD1CULL, "vxw_bonepit_1i4_0_0.vlm"},
	{0x6C0CF9A5A08D6741ULL, "6016895_m.fxe"},
	{0x6C0D281554954852ULL, "vxlt_blackemporium_uf_0_0.vlm"},
	{0x6C0D5C68AD546832ULL, "4330996.ogg"},
	{0x6C0DC8627C5A35A5ULL, "vxw_sundermount_2dz_0_0.vlm"},
	{0x6C0DDEAC02055E4AULL, "tbe100_xenon6190975.cl"},
	{0x6C0DE6C28E37F49AULL, "lt_0_lowtown_docks_n_197_0.msh"},
	{0x6C0DF4C28E380CD3ULL, "lt_0_lowtown_docks_n_197_0.mao"},
	{0x6C0DF71B66AFEC77ULL, "vxlt_undercitydungeon_1il_0_0.vlm"},
	{0x6C0DF8C28E3813E0ULL, "lt_0_lowtown_docks_n_197_0.mmh"},
	{0x6C0E018A380D9705ULL, "vxlt_hangedman_n_1vg_0_1.vlm"},
	{0x6C0E40FFC94F358FULL, "vxht_bartrand_n_5ri_0_0.vlm"},
	{0x6C0E5FEC860043D7ULL, "vxdr_0_deeproads_cmq_0_0.vlm"},
	{0x6C0E7FC795C915C8ULL, "hlac_ca_caverns2_5802_29z.dds"},
	{0x6C0F09BD850ED043ULL, "vxca_caverns2prc_62p_0_0.vlm"},
	{0x6C0F41A4C1E68AD6ULL, "vxw_bonepit_1jg_0_0.vlm"},
	{0x6C0F7D084C62C81BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho03.mmh"},
	{0x6C0F9B084C62FB05ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho03.msh"},
	{0x6C0FDBF0C3788AAAULL, "hlac_lt_gallowstemplar_f_2394_11y.dds"},
	{0x6C0FF6D43030B681ULL, "vxw_sundermount_2sg_0_0.vlm"},
	{0x6C106A99DB510DA3ULL, "hlac_lt_0_lowtown_docks_a3_131070_3y.dds"},
	{0x6C107D38EE5E506FULL, "6114346_m.fxe"},
	{0x6C10B43DD6A3E8A2ULL, "qun221_viscount6070099.cl"},
	{0x6C115C2E811D29B3ULL, "vxw_bonepitdestroyed_a3_kz_0_0.vlm"},
	{0x6C116BD460B92CEFULL, "hlac_lt_gallowstemplar_2391_5y.dds"},
	{0x6C11991B5B97D21CULL, "394684389.ogg"},
	{0x6C11FF67FDBF5E72ULL, "vxlt_0_lowtown_docks_n_cuu_0_0.vlm"},
	{0x6C1253D19239E21FULL, "131603254.ogg"},
	{0x6C12BBA93E75743DULL, "6154216_m.fxe"},
	{0x6C12CFFF7144577BULL, "vxht_0_hightown_b9k_0_0.vlm"},
	{0x6C12EB458CBEEA72ULL, "vxht_mansion2b_1kk_0_0.vlm"},
	{0x6C130AAF7B8B611CULL, "seb221_brett_harimann.cnv"},
	{0x6C131E6AD8DEBA53ULL, "mrl150cr_shadow_assasin.utc"},
	{0x6C13DBF0B7879AA9ULL, "w_bonepitmining_a2_130_0.mao"},
	{0x6C13E7F0B787AF42ULL, "w_bonepitmining_a2_130_0.mmh"},
	{0x6C13E9F0B787B2F8ULL, "w_bonepitmining_a2_130_0.msh"},
	{0x6C13F1B34251F145ULL, "vxlt_0_lowtown_a3_n_c_gc6_0_0.vlm"},
	{0x6C14FA690274288DULL, "vxht_mansion2_3e4_0_0.vlm"},
	{0x6C156FBC54C2C36EULL, "hlac_dr_0_deeproads_5849_1y.dds"},
	{0x6C1575F8FEEB48C2ULL, "vxlt_warehouse_11s_0_0.vlm"},
	{0x6C15D3C0D596930CULL, "hlac_dr_0_deeproads_5845_8x.dds"},
	{0x6C15E293BFEB9292ULL, "gui_reticle_p0.psh.fxo.dx11"},
	{0x6C162CAF7B8DD16CULL, "seb221_brett_harimann.bnk"},
	{0x6C1649C5A1A9387EULL, "6100359_m.fxe"},
	{0x6C16A0FAA24DB2C6ULL, "hlac_ht_mansion2b_818_35z.dds"},
	{0x6C16AD9F2E6AED0EULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball6.mmh"},
	{0x6C16BA5F8B4E52F8ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_dsaxa_0n.dds"},
	{0x6C16BA82721DE7B6ULL, "hlac_lt_0_lowtown_docks_a3_131071_0z.dds"},
	{0x6C1737D0ADA56ACCULL, "419304850.ogg"},
	{0x6C176D86242F8C75ULL, "33912153.ogg"},
	{0x6C177D8CD2D078B2ULL, "708209448.ogg"},
	{0x6C17979DD1A57216ULL, "vxht_0_hightown_aqi_0_0.vlm"},
	{0x6C179A2D4123820DULL, "6027969_m.fxe"},
	{0x6C17BE1581BED057ULL, "hlac_dr_0_deeproads_5841_2y.dds"},
	{0x6C17F79C366FC959ULL, "hlac_ht_brothel_n_3374_0y.dds"},
	{0x6C17F927BBA736DFULL, "6036965_m.fxe"},
	{0x6C18037C55575A82ULL, "vxht_mansion2b_1nd_0_0.vlm"},
	{0x6C181F84CE016060ULL, "6115438_m.fxe"},
	{0x6C1857A21F8695E0ULL, "518263142.ogg"},
	{0x6C18BE13D07C688EULL, "follower_banter6143180.cl"},
	{0x6C18D5DDF8A5208EULL, "hlac_lt_0_lowtown_a2_n_65536_6y.dds"},
	{0x6C191FDE4EB4A404ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_002a_0n.dds"},
	{0x6C197369084A67DDULL, "art\\characters\\weapons\\staff\\textures\\w_stf_014a_0n.dds"},
	{0x6C19F47A8F0E29BFULL, "90448990.ogg"},
	{0x6C1A3FF690FB32FFULL, "vxlt_0_lowtown_docks_a3_n_aow_0_0.vlm"},
	{0x6C1A906F817EBDB7ULL, "ts_generic_prog_junk3.gda"},
	{0x6C1AF95163BC9D0DULL, "6112060_f.fxe"},
	{0x6C1B12BCA6EE2E6FULL, "494629887.ogg"},
	{0x6C1B6C2ECDCB0393ULL, "vxlt_0_lowtown_a2_n_c_by2_0_0.vlm"},
	{0x6C1BD830DB3F21C7ULL, "lt_gallowscourtyard_35_0.rml"},
	{0x6C1C3A3F6E504E25ULL, "vxw_sundermount_2iy_0_0.vlm"},
	{0x6C1C68F858A604F7ULL, "vxlt_undercitydungeon_1lo_0_0.vlm"},
	{0x6C1C904434598F95ULL, "6112993_m.fxe"},
	{0x6C1C9DCF47F12736ULL, "hm_cullen.mor"},
	{0x6C1CD1C977F65C57ULL, "vxdr_0_deeproads_cjp_0_0.vlm"},
	{0x6C1D332AD2122375ULL, "mag110_vincento6032992.cl"},
	{0x6C1D3DEA64DB5553ULL, "hlac_lt_0_lowtown_a2_n_-2_6z.dds"},
	{0x6C1DF50D9E9EF3EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_loop.gad"},
	{0x6C1E0D0CBE818300ULL, "inventory.gfx"},
	{0x6C1E68B12226CF01ULL, "vxw_sundermount_2tf_0_0.vlm"},
	{0x6C1E9411048AFC3FULL, "316842397.ogg"},
	{0x6C1EE22F5AF910ADULL, "vxht_brothel_n_4w5_0_0.vlm"},
	{0x6C1F52DF5E7A4860ULL, "gen00fl_fenris6106724.cl"},
	{0x6C1FECBF5FCD2A0FULL, "6072309_m.fxe"},
	{0x6C1FF04139ADAACBULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb101a_0t.dds"},
	{0x6C205003D1ADBFAAULL, "vxht_0_hightown_n_cal_0_0.vlm"},
	{0x6C21289AA71D142AULL, "wm_ico_gallows.dds"},
	{0x6C216291D965D20EULL, "vxlt_darktown_589_0_0.vlm"},
	{0x6C219E42545471CCULL, "480079149.ogg"},
	{0x6C21ABE4C9A3989EULL, "209239955.ogg"},
	{0x6C21C756E5E95199ULL, "6133822_m.fxe"},
	{0x6C21FF629B8AC67CULL, "lt_0_lowtown_docks_n_42_0.rml"},
	{0x6C226390344809C5ULL, "vxlt_0_lowtown_a3_n_c_gf5_0_0.vlm"},
	{0x6C226722050CB09AULL, "cd_one100dg_kelder.wbd"},
	{0x6C230DA896E71AFFULL, "lt_0_lowtown_a3_n_c_15_0.msh"},
	{0x6C2313A896E724A1ULL, "lt_0_lowtown_a3_n_c_15_0.mmh"},
	{0x6C231E0D7CC12F14ULL, "6016634_m.fxe"},
	{0x6C231FA896E73942ULL, "lt_0_lowtown_a3_n_c_15_0.mao"},
	{0x6C235E09EC7CD483ULL, "175967356.ogg"},
	{0x6C2383DDED03DFD7ULL, "6077799_m.fxe"},
	{0x6C23CEA276A55E55ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\blendtree\\meredith.anb"},
	{0x6C246B79FF4990C2ULL, "222045748.ogg"},
	{0x6C253337714014FAULL, "484797043.ogg"},
	{0x6C255430AD53C516ULL, "dae000_gamlen6145104.cl"},
	{0x6C2567421DE981A1ULL, "6189912-alwaysbeyourfriend.cut"},
	{0x6C257810A52CF133ULL, "ef_nyssa.mor"},
	{0x6C25DC94C00CD44AULL, "dae000_arishok6075189.cl"},
	{0x6C25F5074034AF36ULL, "vxht_mansion2b_2ol_0_0.vlm"},
	{0x6C26504E71AD1B70ULL, "hm_nat_elf_human.mor"},
	{0x6C265648EBC4E3DCULL, "761890351.ogg"},
	{0x6C267559474D7302ULL, "vxht_mansion2b_1ae_0_0.vlm"},
	{0x6C27A2B3474DA757ULL, "vxht_chantry_2bv_0_0.vlm"},
	{0x6C2805343121A722ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_action.gad"},
	{0x6C287B9DA5365CC1ULL, "mcr310_argument6075126.cl"},
	{0x6C28AD438163129AULL, "gen00fl_aveline6139927.cl"},
	{0x6C28E06811098C34ULL, "plt_qun200pt_following_qun.plo"},
	{0x6C2923218366A3BAULL, "nat380_sibling6068023.cl"},
	{0x6C29291DD8D6BAAEULL, "6142717_m.fxe"},
	{0x6C293E0ED5220221ULL, "gen00fl_dog6166393.cl"},
	{0x6C2A0A185517605EULL, "vxht_0_hightown_a2_a6l_0_0.vlm"},
	{0x6C2A6E4D94346ADBULL, "w_encounter2_8_0.rml"},
	{0x6C2A768C8DD63384ULL, "6149840_m.fxe"},
	{0x6C2B24B9AD3B668FULL, "vxht_bartrand_n_5xk_0_0.vlm"},
	{0x6C2B291120535255ULL, "vxht_mansion2b_3d7_0_0.vlm"},
	{0x6C2B36D86439E83EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_temp_rock_d.dds"},
	{0x6C2B43A669EC74D7ULL, "vxdr_0_deeproads_cgs_0_0.vlm"},
	{0x6C2B61B5EE636C91ULL, "489481652.ogg"},
	{0x6C2B97B533DDD08EULL, "105106722.ogg"},
	{0x6C2BC49A92461D81ULL, "809667566.ogg"},
	{0x6C2BF8411586DF70ULL, "339663136.ogg"},
	{0x6C2C070254DA4E57ULL, "wm_ico_caravan_roll.dds"},
	{0x6C2C262B501B33F9ULL, "zz_isa_debug6133154.cl"},
	{0x6C2C2C1F8560BB75ULL, "vxlt_gallowstemplar_n_5c0_0_0.vlm"},
	{0x6C2D196BE61E8628ULL, "gen_im_cns_tom_attribute_02.uti"},
	{0x6C2D7A8F1EF601C6ULL, "6060383_m.fxe"},
	{0x6C2D88DEED4F2B34ULL, "6037613_m.fxe"},
	{0x6C2DBF9FBF006BCFULL, "follower_banter6155892.cl"},
	{0x6C2DE2F6DBBE9311ULL, "455079102.ogg"},
	{0x6C2DF47C9C22ED1CULL, "vxht_0_hightown_a2_cbp_0_0.vlm"},
	{0x6C2E1AC75334E273ULL, "art\\vfx\\combat\\vfx_glb_stun_c.cess.ani"},
	{0x6C2E60ADE9E3285FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallojutflat.phy"},
	{0x6C2E629A46047062ULL, "ser191_emeric6107890.cl"},
	{0x6C2E874721781EE3ULL, "mcr370_assignments6081722.cl"},
	{0x6C2EB4F66B208746ULL, "6050076_m.fxe"},
	{0x6C2F283431280119ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_action.evt"},
	{0x6C2F55EC0ECED9CEULL, "6078424_f.fxe"},
	{0x6C3012E4B3C9BE9DULL, "vfx_flys.bnk"},
	{0x6C305F6F0C5A6FA8ULL, "art\\vfx\\combat\\vfx_c_mer_rush_strike_c.anb"},
	{0x6C30754D9438E233ULL, "w_encounter2_8_0.phy"},
	{0x6C3102CDDF26DC22ULL, "vxw_bonepitdestroyed_a3_1wu_0_1.vlm"},
	{0x6C316790409DD99DULL, "249161668.ogg"},
	{0x6C31BD160C0B08DDULL, "w_bonepitmining_a2_101_0.mao"},
	{0x6C31C1160C0B0F16ULL, "w_bonepitmining_a2_101_0.mmh"},
	{0x6C31CB160C0B2014ULL, "w_bonepitmining_a2_101_0.msh"},
	{0x6C32120D9EAFCDB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_loop.ani"},
	{0x6C322EDA2C5B8CD7ULL, "llac_w_encounter1_-65538_0.dds"},
	{0x6C323527A426A123ULL, "333496247.ogg"},
	{0x6C3241653560ABC9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medd_0s.dds"},
	{0x6C3258D2E79B3D0AULL, "vxlt_hangedman_n_25s_0_0.vlm"},
	{0x6C32973750489CD0ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killervase_0.pwk"},
	{0x6C32EB42A2C27C85ULL, "6135060_m.fxe"},
	{0x6C3301B89BA768D3ULL, "vxht_mansion2b_266_0_0.vlm"},
	{0x6C332D0C7B3C408BULL, "art\\characters\\creatures\\spider\\textures\\c_giaspidra_0n.dds"},
	{0x6C334498ADC60058ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gurney_d.dds"},
	{0x6C336CEA3413D76FULL, "vxlt_0_lowtown_n_eqg_0_0.vlm"},
	{0x6C348A7AAE878BCDULL, "443582670.ogg"},
	{0x6C34E0057DF334AAULL, "one100dg_nabil6033166.cl"},
	{0x6C34E73639438B82ULL, "vxht_mansion2b_1df_0_0.vlm"},
	{0x6C351816FE16A37AULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_6x.dds"},
	{0x6C3530E1B12B4CF3ULL, "vxht_keep_63q_0_0.vlm"},
	{0x6C35828D392DB1FBULL, "one100dg_kelder6173947.cl"},
	{0x6C364F81E3CD2961ULL, "907794515.ogg"},
	{0x6C3651842655BAA6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightownbooks_n.dds"},
	{0x6C366329C4865202ULL, "503100458.ogg"},
	{0x6C367318B463871FULL, "art\\vfx\\combat\\vfx_q_ari_wwind_act_c.impa.gad"},
	{0x6C3694CAEC243E6FULL, "hlac_lt_0_lowtown_docks_n_0_0z.dds"},
	{0x6C36B9CD5AB49479ULL, "375979485.ogg"},
	{0x6C375B5D813FCC0EULL, "art\\characters\\creatures\\spider\\c_poispidr.phy"},
	{0x6C3877B08A4BC24AULL, "hlac_w_blightlands_-131069_0z.dds"},
	{0x6C390C955F3EAC18ULL, "vxlt_0_lowtown_a2_env_0_0.vlm"},
	{0x6C391DF9523C7F25ULL, "vxw_sundermount_2wg_0_0.vlm"},
	{0x6C394CB23C9235F7ULL, "vxlt_undercitydungeon_1zi_0_0.vlm"},
	{0x6C3996969F317F0FULL, "vxht_bartrand_n_5cj_0_0.vlm"},
	{0x6C3A46E9361353E9ULL, "vxlt_warehouse_13r_0_0.vlm"},
	{0x6C3ACBAAC9873838ULL, "vxdr_primevalentrance_14u_0_0.vlm"},
	{0x6C3B12C7533F5EAAULL, "art\\vfx\\combat\\vfx_glb_stun_c.cess.evt"},
	{0x6C3BE9107FE0214DULL, "vxdr_primevalentrance_23f_0_0.vlm"},
	{0x6C3C08457EECB7A9ULL, "6148082_m.fxe"},
	{0x6C3C303431329908ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_standing_action.ani"},
	{0x6C3C3C652A64E604ULL, "dae000_merchant_generic_hm6172638.cl"},
	{0x6C3C41D6DC275206ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_injured_loop.ani"},
	{0x6C3C7762480CE55DULL, "hlac_lt_warehouse_0_8x.dds"},
	{0x6C3C826E5CFC0752ULL, "vxht_mansion2b_263_0_0.vlm"},
	{0x6C3CC5874C5BAC07ULL, "103333275.ogg"},
	{0x6C3CE70DEEBADCCBULL, "gencr_city_guard_2h.utc"},
	{0x6C3D1B403824F361ULL, "hla_ca_cave_2194_0y.dds"},
	{0x6C3D500460F35B0CULL, "hlac_ht_mansion2_818_42x.dds"},
	{0x6C3D62E081948674ULL, "341951704.ogg"},
	{0x6C3D6818B469930CULL, "art\\vfx\\combat\\vfx_q_ari_wwind_act_c.impa.evt"},
	{0x6C3DDFAD6F5F9455ULL, "555283012.ogg"},
	{0x6C3E3301E10CB3A8ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyo.msh"},
	{0x6C3E4501E10CD239ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyo.mao"},
	{0x6C3E48516A28A455ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg102a_0s.dds"},
	{0x6C3E4BD9D79AF56CULL, "dae000ip_crowd_ff6_talk12.utp"},
	{0x6C3E5101E10CE6D2ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyo.mmh"},
	{0x6C3ED76B8C879E02ULL, "6167899_m.fxe"},
	{0x6C3F61ADE9F19D64ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallojutflat.mmh"},
	{0x6C3F65ADE9F1A3B7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallojutflat.mao"},
	{0x6C3F77ADE9F1C206ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallojutflat.msh"},
	{0x6C3FEA1E999E6B5FULL, "vxlt_0_lowtown_a2_n_c_fg8_0_0.vlm"},
	{0x6C3FF3AF7F50CD2EULL, "11718547.ogg"},
	{0x6C402519B1BCD11FULL, "6152289_m.fxe"},
	{0x6C402EA30F534575ULL, "dae000_gamlen6147235.cl"},
	{0x6C40663C5A399F87ULL, "art\\vfx\\combat\\vfx_str_att_back_l_leg_c.impa.gad"},
	{0x6C406F1D396288D2ULL, "6121563_m.fxe"},
	{0x6C40B5731FA711BDULL, "gen00fl_aveline6049661.cl"},
	{0x6C40EDB890F23BDFULL, "vxw_woundedcoast_86_0_0.vlm"},
	{0x6C4149469F62C846ULL, "hlac_lt_0_lowtown_a2_n_196606_0z.dds"},
	{0x6C4173958D9D8153ULL, "vxht_mansion2b_239_0_0.vlm"},
	{0x6C41A39F7B507151ULL, "hlac_w_bonepitdestroyed_a3_131075_0x.dds"},
	{0x6C41A8065B612DA2ULL, "6139405_m.fxe"},
	{0x6C41DEC72609EFEFULL, "vxlt_0_lowtown_n_ejf_0_0.vlm"},
	{0x6C41E66D933AC67EULL, "vxht_0_hightown_a2_b96_0_0.vlm"},
	{0x6C41F1C75345451DULL, "art\\vfx\\combat\\vfx_glb_stun_c.cess.gad"},
	{0x6C4294494F1FCB35ULL, "art\\vfx\\models\\vfx_rockarmor_chest.mmh"},
	{0x6C4296494F1FCEABULL, "art\\vfx\\models\\vfx_rockarmor_chest.msh"},
	{0x6C42BC2A05EB4BCDULL, "vxht_0_hightown_a2_bs3_0_0.vlm"},
	{0x6C42D8C12420E036ULL, "vxht_mansion2b_2en_0_0.vlm"},
	{0x6C4322866CCEECD2ULL, "seb221_flora_harimann6177673.cl"},
	{0x6C434438A0750DE0ULL, "gen00fl_bethany6169918.cl"},
	{0x6C438218D4128FDDULL, "6120453_m.fxe"},
	{0x6C448E430F3F5C9CULL, "6070151_m.fxe"},
	{0x6C44A48D67AE4F83ULL, "6166398_m.fxe"},
	{0x6C44C19295648719ULL, "vxlt_0_lowtown_docks_n_cs9_0_1.vlm"},
	{0x6C4573AAC74192ABULL, "hm_gen_circle_mage.mor"},
	{0x6C46200AE33DBC70ULL, "vxht_0_hightown_a3_b0n_0_0.vlm"},
	{0x6C4627419CA1943DULL, "6094722_f.fxe"},
	{0x6C4653E4F218C83FULL, "6076812_f.fxe"},
	{0x6C465C7CE5F9A6A2ULL, "vxlt_0_lowtown_docks_c52_0_0.vlm"},
	{0x6C4680E95D3AD964ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01exit.ani"},
	{0x6C46BE94B0B4C411ULL, "vxlt_gallowscourtyard_n_741_0_0.vlm"},
	{0x6C475B3C5A3FAB94ULL, "art\\vfx\\combat\\vfx_str_att_back_l_leg_c.impa.evt"},
	{0x6C477AA38157BBB0ULL, "475351494.ogg"},
	{0x6C4796FDE46C37AAULL, "vxht_mansion2b_2b6_0_1.vlm"},
	{0x6C479D8227BF930CULL, "hf_shopkeeper.mor"},
	{0x6C47A7F8B6D96D7BULL, "vxdr_primevaltomb_6uz_0_0.vlm"},
	{0x6C480E715EDCA856ULL, "54505285.ogg"},
	{0x6C483B1A41655228ULL, "zz_seb_debug6183098.cl"},
	{0x6C48C758E35451D2ULL, "69608809.ogg"},
	{0x6C49B2EE4483698CULL, "35640484.ogg"},
	{0x6C49C45592782F19ULL, "6160974_m.fxe"},
	{0x6C49CE6A61F448A4ULL, "llac_ca_caverns1_mines_24423_3.dds"},
	{0x6C49FB5872B7E5F5ULL, "vxdr_0_deeproads_d3v_0_0.vlm"},
	{0x6C4A36031B9BFFF4ULL, "6051842_m.fxe"},
	{0x6C4A6E01781B508FULL, "608551615.ogg"},
	{0x6C4A8A9F7C41C9D5ULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_0x.dds"},
	{0x6C4A94D6D4D0D5FBULL, "vxdr_primevaltomb_542_0_0.vlm"},
	{0x6C4AC7251BBB71F3ULL, "llac_lt_foundry_n_320_15.dds"},
	{0x6C4ACF3182EFFFCCULL, "6129815_m.fxe"},
	{0x6C4AFAAFBF558AD6ULL, "mag100_cullen6092274.cl"},
	{0x6C4B0EFF8A81DA31ULL, "vxht_mansion2_1vk_0_0.vlm"},
	{0x6C4B5018B475A865ULL, "art\\vfx\\combat\\vfx_q_ari_wwind_act_c.impa.ani"},
	{0x6C4B73AEBDC7CD88ULL, "lotheringslament.dds"},
	{0x6C4B77113FACFA73ULL, "one220_yevhen6098721.cl"},
	{0x6C4BB4873C4F95EAULL, "follower_banter6164068.cl"},
	{0x6C4BCDACE97D20F9ULL, "vxht_bartrand_n_5p5_0_0.vlm"},
	{0x6C4BD63FAB0A8125ULL, "vxht_chantry_1yn_0_0.vlm"},
	{0x6C4BFACED22C119AULL, "art\\vfx\\materials\\vfx_dust_particles.mao"},
	{0x6C4C02585B983C85ULL, "hlac_lt_darktown_-1_54x.dds"},
	{0x6C4C080B3F5A91EAULL, "mcr370_negotiation6078748.cl"},
	{0x6C4C865ED71E51C0ULL, "dae120ar_hightown.are"},
	{0x6C4D1BE1FEAC98FEULL, "vxca_caverns2_83e_0_0.vlm"},
	{0x6C4D24777DBA88A2ULL, "vxht_keep_n_5h5_0_0.vlm"},
	{0x6C4D4D7EA02B4683ULL, "w_bonepitdestroyed_a3_90_0.msh"},
	{0x6C4D5B7EA02B5EDDULL, "w_bonepitdestroyed_a3_90_0.mmh"},
	{0x6C4D5F7EA02B65EEULL, "w_bonepitdestroyed_a3_90_0.mao"},
	{0x6C4D7DE95D40F2C6ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01exit.gad"},
	{0x6C4DC5477FDE24A3ULL, "one140cr_charade6134602.cl"},
	{0x6C4DCD3A6BB96DC2ULL, "296327159.ogg"},
	{0x6C4DFE8B8D9658ACULL, "6189426_m.fxe"},
	{0x6C4E03CC2B100C28ULL, "vxht_0_hightown_a3_ca0_0_0.vlm"},
	{0x6C4E3D458017229BULL, "6098197_f.fxe"},
	{0x6C4E86D67E4C4968ULL, "mcr310_argument6074401.cl"},
	{0x6C4EB66430C2F981ULL, "hlac_ht_0_hightown_n_-131074_2z.dds"},
	{0x6C4F112BB5F4DCDAULL, "follower_banter6171866.cl"},
	{0x6C4F20F03D716D09ULL, "6067767_m.fxe"},
	{0x6C4F26FFB0D2D758ULL, "ts_kirk_lowtown_prog_junk1.gda"},
	{0x6C4F5F9582E8545FULL, "vxw_woundedcoast_51_0_0.vlm"},
	{0x6C4F7D5ED7207842ULL, "dae120ar_hightown.ncs"},
	{0x6C4F9CEBE500BFB1ULL, "heat_dis.mao"},
	{0x6C5014A8970E141EULL, "lt_0_lowtown_a3_n_c_15_0.phy"},
	{0x6C502AD6DC37D400ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_injured_loop.gad"},
	{0x6C50408EA233F679ULL, "365865.ogg"},
	{0x6C504F59EAB363D0ULL, "llac_ht_0_hightown_a2_65532_0.dds"},
	{0x6C50AC05578EF297ULL, "terrain3.psh.fxo"},
	{0x6C51176BFCE60E1AULL, "vxca_caverns1_gka_0_0.vlm"},
	{0x6C51B20F6539C5DBULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medg_0s.dds"},
	{0x6C52149B95177DF3ULL, "vxht_keep_69s_0_0.vlm"},
	{0x6C522F26DB6769CFULL, "493394407.ogg"},
	{0x6C524B3D2941C5B9ULL, "vxdr_0_deeproads_bpa_0_0.vlm"},
	{0x6C52ECCB86B70D5DULL, "397256814.ogg"},
	{0x6C5300200135751CULL, "6075251_m.fxe"},
	{0x6C53038842DD18D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v3.ani"},
	{0x6C5342A8F6BE96C7ULL, "vxht_chantry_n_2c6_0_0.vlm"},
	{0x6C53A639D5E384D0ULL, "hlac_lt_warehouse_n_-65535_0y.dds"},
	{0x6C5478E95D4708A5ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01exit.evt"},
	{0x6C54D464CB2F2544ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpb.phy"},
	{0x6C554197A34CCF4DULL, "hlac_blank_exterior_0_0y.dds"},
	{0x6C55533C5A4BDBCDULL, "art\\vfx\\combat\\vfx_str_att_back_l_leg_c.impa.ani"},
	{0x6C5597AB23E8B852ULL, "w_encounter2_18_0.rml"},
	{0x6C55BE86323CC60AULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_monique.gad"},
	{0x6C55F9EC939153AAULL, "6155285_m.fxe"},
	{0x6C561A768105920DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgtb_0d.dds"},
	{0x6C562E8A2BAF7E24ULL, "vxht_0_hightown_a2_arz_0_0.vlm"},
	{0x6C56955ED726BFD3ULL, "dae120ar_hightown.lst"},
	{0x6C569E084C9F0350ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho03.phy"},
	{0x6C56A27A66D759D9ULL, "108629919.ogg"},
	{0x6C56FCF742B654CFULL, "vxht_mansion2b_381_0_0.vlm"},
	{0x6C57384ECBCF6D35ULL, "862305443.ogg"},
	{0x6C576B6853AF9304ULL, "142124994.ogg"},
	{0x6C57ADA2295B4560ULL, "art\\vfx\\materials\\chunky\\vfx_c_falloff.mao"},
	{0x6C57D8D2C8FC3CA5ULL, "6118443_m.fxe"},
	{0x6C580468C08D59B3ULL, "606520451.ogg"},
	{0x6C581BEDF0A86B1FULL, "ht_0_hightown_a2_18_0_ll.mmh"},
	{0x6C588638DBBCB786ULL, "vxca_caverns2_86v_0_0.vlm"},
	{0x6C58A5AE54CA55F7ULL, "var_module.gda"},
	{0x6C58ABD6E53E8F11ULL, "vxlt_0_lowtown_docks_a3_n_d4e_0_0.vlm"},
	{0x6C58AC0CBB262B7CULL, "hightown_male_noble_gossip6171428.cl"},
	{0x6C58BE8732695E0BULL, "826728286.ogg"},
	{0x6C58E2A24966919AULL, "6146822.cut"},
	{0x6C595E6F0C7D7343ULL, "art\\vfx\\combat\\vfx_c_mer_rush_strike_c.mmh"},
	{0x6C595F6C1C760083ULL, "564780140.ogg"},
	{0x6C5965CABDEEFE07ULL, "zz_spy130cr_darktown_res_1.utc"},
	{0x6C5980DC7C77F2B1ULL, "vxht_mansion2_1yj_0_0.vlm"},
	{0x6C5990526E28E26FULL, "hlac_lt_gallowstemplar_f_7198_4y.dds"},
	{0x6C59FE4FBA93830CULL, "hm_icarver_h2s3.mor"},
	{0x6C5A15166687FFC8ULL, "vxlt_darktown_41m_0_0.vlm"},
	{0x6C5A3244D6C25C8BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_aveline.ani"},
	{0x6C5A45BED7B7CCBDULL, "hlac_lt_0_lowtown_docks_a3_65535_5y.dds"},
	{0x6C5ABA5350582679ULL, "mag130_danzig6018809.cl"},
	{0x6C5AE5CF7B3CCB6CULL, "6071109_m.fxe"},
	{0x6C5BA461C676AE30ULL, "183994116.ogg"},
	{0x6C5C40F03C82CE0AULL, "vxht_bartrand_n_4gv_0_0.vlm"},
	{0x6C5C7DBDB56051ADULL, "hlac_dr_0_deeproads_5841_3x.dds"},
	{0x6C5C98AB23EED8E6ULL, "w_encounter2_18_0.phy"},
	{0x6C5CA33CA3A35EAFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_escape.gad"},
	{0x6C5CDB30DB76BDDBULL, "lt_gallowscourtyard_35_0.anb"},
	{0x6C5CFC88CF8C038AULL, "104205356.ogg"},
	{0x6C5D29A6854BA633ULL, "vxlt_0_lowtown_docks_a3_n_aqm_0_0.vlm"},
	{0x6C5D33E3DB66C3FDULL, "737613288.ogg"},
	{0x6C5D7D0A63D5FB12ULL, "403732908.ogg"},
	{0x6C5D7EDF9A01A484ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_basic_2.mao"},
	{0x6C5DAD0B18D6053DULL, "hlac_lt_0_lowtown_1_3x.dds"},
	{0x6C5DD17274DE6CDFULL, "vxw_woundedcoast_20_0_0.vlm"},
	{0x6C5E210EC7C175FBULL, "175202828.ogg"},
	{0x6C5E2D62CA439790ULL, "vxw_woundedcoast_1g9_0_0.vlm"},
	{0x6C5EE2B75099E283ULL, "vxlt_0_lowtown_a2_n_c_ghl_0_0.vlm"},
	{0x6C5F6EBE599B2D3AULL, "plt_cod_aow_inventory.plo"},
	{0x6C5F9FE3E9D77CCDULL, "vxht_0_hightown_a2_by1_0_0.vlm"},
	{0x6C5FD179E53B8BB2ULL, "vxca_caverns2prc_7tv_0_0.vlm"},
	{0x6C603CCD0F25D502ULL, "vxht_mansion2b_1ua_0_0.vlm"},
	{0x6C60EAAF010A93CFULL, "hlac_lt_0_lowtown_a2_n_c_-1_1x.dds"},
	{0x6C6103DF74CB4070ULL, "gen00fl_aveline6132858.cl"},
	{0x6C61B719D42B12A8ULL, "6129749_m.fxe"},
	{0x6C61BC97467DA91EULL, "6140795_m.fxe"},
	{0x6C627CCCE8A96168ULL, "ave200_aveline_bar6116412.cl"},
	{0x6C62DB3516D4B63BULL, "vxlt_0_lowtown_a3_n_c_h1y_0_0.vlm"},
	{0x6C63A415E08930BCULL, "6106368_m.fxe"},
	{0x6C63B67DC6ECACA8ULL, "lgt210cr_eau_thug3.utc"},
	{0x6C63B82878C2AE8DULL, "vxlt_0_lowtown_a2_eoe_0_1.vlm"},
	{0x6C63B83CA3A9A15CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_escape.evt"},
	{0x6C6437F155F47EDFULL, "6097552_f.fxe"},
	{0x6C646A12C37D818EULL, "hlac_lt_0_lowtown_docks_-65539_5x.dds"},
	{0x6C657ABA317C86AFULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_screen_d.dds"},
	{0x6C66764744A25B0BULL, "hlac_ht_mansion2_818_63z.dds"},
	{0x6C66CAD6C3DD2420ULL, "648386078.ogg"},
	{0x6C66DA8842ED7C7DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v3.gad"},
	{0x6C67CEAC9BCFC4C0ULL, "72453108.ogg"},
	{0x6C681A44D6CE7102ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_aveline.evt"},
	{0x6C68977466306607ULL, "6110063_m.fxe"},
	{0x6C68B56E1B71FACCULL, "239021887.ogg"},
	{0x6C68E38A36318363ULL, "rdr111dg_martin6054516.cl"},
	{0x6C690B8C780E2C4DULL, "6126381_m.fxe"},
	{0x6C694474771CFEE5ULL, "1016934449.ogg"},
	{0x6C699986324D30F0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheerlosing_1_monique.ani"},
	{0x6C69B0BF00190421ULL, "follower_banter6172747.cl"},
	{0x6C69B2858E8E5335ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st105a_0n.dds"},
	{0x6C69DB0054C24981ULL, "672867375.ogg"},
	{0x6C6A1C483633644AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb.mao"},
	{0x6C6A204836336A99ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb.mmh"},
	{0x6C6A2A4836337B87ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb.msh"},
	{0x6C6A8466C085B348ULL, "6060381_m.fxe"},
	{0x6C6B2393FF5BCB7AULL, "dre000_varric6208007.cl"},
	{0x6C6C50A12AA29196ULL, "rdr110dg_party_comments6094675.cl"},
	{0x6C6CC9327D45091FULL, "hlac_w_blightlands_262146_0y.dds"},
	{0x6C6CE79174F14109ULL, "511657431.ogg"},
	{0x6C6CEB5F4B59C0CCULL, "hlac_w_bonepit_196611_0x.dds"},
	{0x6C6D0166A4CEE217ULL, "6116713_m.fxe"},
	{0x6C6D0DE48A34750EULL, "6144863_m.fxe"},
	{0x6C6D382607C861AEULL, "vxw_bonepitmining_a2_cu_0_0.vlm"},
	{0x6C6D89F48E7E0860ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_0y.dds"},
	{0x6C6DD1DC4E1DE792ULL, "173488117.ogg"},
	{0x6C6E5BF8B7A19990ULL, "vxlt_darktown_59x_0_0.vlm"},
	{0x6C6E78F54E9F623EULL, "vxht_keep_n_6y5_0_0.vlm"},
	{0x6C6EC946226EDAC1ULL, "478393438.ogg"},
	{0x6C6F1944D6D48E15ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_aveline.gad"},
	{0x6C6F2D7DF19718C3ULL, "vxca_caverns2prc_ae6_0_0.vlm"},
	{0x6C6FB89201B39D28ULL, "vxlt_gallowstemplar_n_5jb_0_0.vlm"},
	{0x6C6FD8EC64A5BC9DULL, "hlac_lt_0_lowtown_docks_n_65533_2x.dds"},
	{0x6C6FE17E9D3F7805ULL, "6112081_m.fxe"},
	{0x6C703D39536A85EDULL, "gen00fl_fenris6126934.cl"},
	{0x6C7085D84CA9B9B8ULL, "llac_w_bonepitmining_a2_196613_0.dds"},
	{0x6C70AE9CD8B45EF7ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hanginga_d.dds"},
	{0x6C70AEBBB13E4C14ULL, "sse_ss_shadow_m_1.gda"},
	{0x6C70FEBD3193001AULL, "6169707_m.fxe"},
	{0x6C713CE418D6A910ULL, "gen00fl_sebastian6177903.cl"},
	{0x6C714044D694170CULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd03a.phy"},
	{0x6C716DB3DC36D5F2ULL, "259870579.ogg"},
	{0x6C71C03CA3B5EC75ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_escape.ani"},
	{0x6C72CC2BF9E3895AULL, "narrative_varric_demo6170840.cl"},
	{0x6C72D78D81DB80BCULL, "6162564_m.fxe"},
	{0x6C7346967D1B3AF3ULL, "zz_tanner_trial.cut"},
	{0x6C73DAA3BC4A6731ULL, "753370170.ogg"},
	{0x6C7416C53988FAA0ULL, "vrc321_party_barks6095734.cl"},
	{0x6C74268840FFF2F1ULL, "dae000_pc_home_ambients6208601.cl"},
	{0x6C748F1374E27A3AULL, "vxw_bonepitmining_a2_275_0_0.vlm"},
	{0x6C74C8636B182D64ULL, "follower_barks6153513.cl"},
	{0x6C75173847CE852FULL, "6174546_m.fxe"},
	{0x6C7541C512FEBDFBULL, "6114046_f.fxe"},
	{0x6C7548927E72C584ULL, "vxht_bartrand_n_65d_0_0.vlm"},
	{0x6C75A0E3138565ABULL, "276392101.ogg"},
	{0x6C7607C649E19509ULL, "dae000_hate_enemies6167548.cl"},
	{0x6C76509891C86135ULL, "224967153.ogg"},
	{0x6C766CE7B310C524ULL, "6076386_m.fxe"},
	{0x6C774707FEFD8187ULL, "7075655.ogg"},
	{0x6C777BF192D9E2BAULL, "vxw_bonepitmining_a2_1vm_0_0.vlm"},
	{0x6C77ED20D0DA6A6BULL, "llac_lt_0_lowtown_docks_n_131071_7.dds"},
	{0x6C785A5D817761D1ULL, "art\\characters\\creatures\\spider\\c_poispidr.mmh"},
	{0x6C785E21DE805843ULL, "static_static_p93.psh.fxo.dx11"},
	{0x6C78665D81777672ULL, "art\\characters\\creatures\\spider\\c_poispidr.mao"},
	{0x6C78745D81778D8FULL, "art\\characters\\creatures\\spider\\c_poispidr.msh"},
	{0x6C7877BA7D30E00EULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.dura.ani"},
	{0x6C788572F8F14C7AULL, "324187210.ogg"},
	{0x6C788ABBBDD07CACULL, "vxw_encounter2_camp_nf_0_0.vlm"},
	{0x6C78969917D1EC19ULL, "ser270_emeric6149336.cl"},
	{0x6C78E263361BEF69ULL, "zz_followers_banter6159225.cl"},
	{0x6C78EB3DDC15C40BULL, "996503409.ogg"},
	{0x6C794E442610178AULL, "81372852.ogg"},
	{0x6C797004A7950EBBULL, "zz_mer100pt_racketeers.ncs"},
	{0x6C799A38AE9638BBULL, "6163861_m.fxe"},
	{0x6C79F29A9D345F15ULL, "6161322_m.fxe"},
	{0x6C7A36580B3D524BULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml102a_0t.dds"},
	{0x6C7A53E14B83E6DBULL, "gen00fl_aveline6122065.cl"},
	{0x6C7A9E46FACC3927ULL, "6067634_f.fxe"},
	{0x6C7B165CB3826D8FULL, "and330tr_act3_opening.ncs"},
	{0x6C7B22A01C9623FBULL, "vxw_bonepit_e0_0_0.vlm"},
	{0x6C7C183E167C3252ULL, "6128062_f.fxe"},
	{0x6C7C32572EA15AB1ULL, "vxw_woundedcoast_1ci_0_1.vlm"},
	{0x6C7C7499ECDD7A6BULL, "vxht_0_hightown_a3_31g_0_0.vlm"},
	{0x6C7D24250332E8CEULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_p.mmh"},
	{0x6C7D91A58F460B8DULL, "874649496.ogg"},
	{0x6C7DA66C449141A2ULL, "147157436.ogg"},
	{0x6C7DE660ED9D255DULL, "365785942.ogg"},
	{0x6C7DFB802604991CULL, "240626970.ogg"},
	{0x6C7E5437D1C2A6BBULL, "913541356.ogg"},
	{0x6C7EFDA779EE79FEULL, "vxdr_primevaltomb_3m2_0_0.vlm"},
	{0x6C7F10C23DB3D9A5ULL, "art\\vfx\\combat\\vfx_w_wns_atk_close_c.mmh"},
	{0x6C7FC2CEB7B577AAULL, "6105671_m.fxe"},
	{0x6C7FD9966313D571ULL, "258386818.ogg"},
	{0x6C800826F39632C4ULL, "llac_w_sundermount_-131069_0.dds"},
	{0x6C816A02006E2281ULL, "6145175_m.fxe"},
	{0x6C816D0D61231297ULL, "6075677_f.fxe"},
	{0x6C81A1D8BAEED86DULL, "6077461_m.fxe"},
	{0x6C81DE1CD12AD1BBULL, "6139729_m.fxe"},
	{0x6C81FBF164614541ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeball.ani"},
	{0x6C82217B7911C600ULL, "vxlt_foundry_n_vr_0_0.vlm"},
	{0x6C829824C2976D35ULL, "vxht_brothel_n_4x0_0_0.vlm"},
	{0x6C82F3310BB54636ULL, "dae000_mother6121697.cl"},
	{0x6C838CFE98719F81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_reaction.ani"},
	{0x6C8407313D397355ULL, "lt_0_lowtown_docks_a3_33_0_lc.mmh"},
	{0x6C8419865433719BULL, "hlac_lt_0_lowtown_docks_n_-196611_0y.dds"},
	{0x6C843EFA99DB4495ULL, "285793630.ogg"},
	{0x6C8483274EDF42C7ULL, "vxca_cave_18g_0_0.vlm"},
	{0x6C84AE5AEA3D4BE5ULL, "lla_ht_0_hightown_a2_-131070_0.dds"},
	{0x6C84B425EE1EC8EBULL, "6164068_m.fxe"},
	{0x6C84E9359201E8C2ULL, "ht_0_hightown_46_0.rml"},
	{0x6C855BA9E434B616ULL, "573057306.ogg"},
	{0x6C856FBA7D3B5D5FULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.dura.evt"},
	{0x6C859F5E1B493785ULL, "866827059.ogg"},
	{0x6C85C8D80993B7FFULL, "6095297_f.fxe"},
	{0x6C85F25C19FEAA8FULL, "hm_hugh.mor"},
	{0x6C86E400CE09EFAAULL, "cage_01.mao"},
	{0x6C86FC98AFC6952CULL, "vxw_encounter2_camp_ig_0_0.vlm"},
	{0x6C872802E198A7B6ULL, "gen00fl_fenris6110866.cl"},
	{0x6C87562363C06A25ULL, "487183905.ogg"},
	{0x6C87C9E77DBC2B56ULL, "llac_lt_gallowstemplar_f_2384_4.dds"},
	{0x6C88290327DD1404ULL, "623378661.ogg"},
	{0x6C88DC8C12C1094DULL, "6089815_m.fxe"},
	{0x6C8906F1646776E3ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeball.gad"},
	{0x6C892E42A52B5518ULL, "art\\levels\\ztools\\default\\generic\\0oi_basic.phy"},
	{0x6C89882846751705ULL, "vxht_mansion_n_6s9_0_1.vlm"},
	{0x6C8997FE98761E23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_reaction.gad"},
	{0x6C89A48C94B55E04ULL, "248227476.ogg"},
	{0x6C89C9EEEAC0AC46ULL, "vxht_0_hightown_n_c1p_0_0.vlm"},
	{0x6C8A0B47F4913497ULL, "6085574_m.fxe"},
	{0x6C8A4325033DA7CDULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_p.anb"},
	{0x6C8A708F9AD2E60DULL, "vxw_encounter3_b4_0_0.vlm"},
	{0x6C8AAAEA8D08C8B8ULL, "6136918_m.fxe"},
	{0x6C8AD561CD2FB264ULL, "6040813_m.fxe"},
	{0x6C8B0F35A6AC07FEULL, "1005094054.ogg"},
	{0x6C8B270C43EF859CULL, "6166946_m.fxe"},
	{0x6C8BAF5BBD87EB21ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_act_c.anb"},
	{0x6C8C0C78939A1B4CULL, "vxht_chantry_32l_0_0.vlm"},
	{0x6C8C0D46D1F6A419ULL, "vxw_bonepitdestroyed_a3_8q_0_0.vlm"},
	{0x6C8C0DC23DBE5F72ULL, "art\\vfx\\combat\\vfx_w_wns_atk_close_c.anb"},
	{0x6C8C390860C25609ULL, "6041154_m.fxe"},
	{0x6C8C50BA7D414748ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_proj_p.dura.gad"},
	{0x6C8C7A7C4838124BULL, "vxw_bonepit_1pp_0_0.vlm"},
	{0x6C8C992F6826740FULL, "fen210_orana6133051.cl"},
	{0x6C8CAE92B2F38A4AULL, "zz_followers_banter6158238.cl"},
	{0x6C8CB3657D3CCABDULL, "vxw_bonepitdestroyed_a3_dy_0_0.vlm"},
	{0x6C8D1A0EBF15E258ULL, "vfx_vfx_p5.psh.fxo.dx11"},
	{0x6C8D44DE44AA3D15ULL, "gen00fl_sebastian6180018.cl"},
	{0x6C8DDD3BC559AB1CULL, "isa211_isabela6070578.cl"},
	{0x6C8DEFE34C49ACF0ULL, "hlac_dr_dungeon_3215_9z.dds"},
	{0x6C8E0F3ECAE6A4F2ULL, "vxlt_0_lowtown_a2_n_gcf_0_0.vlm"},
	{0x6C8EF614B8B3F892ULL, "vxlt_gallowstemplar_f_3ek_0_0.vlm"},
	{0x6C8FCC9BECD34A0BULL, "578598643.ogg"},
	{0x6C8FDE53146F5F7DULL, "71658534.ogg"},
	{0x6C8FE3F1646D59C8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformtospikeball.evt"},
	{0x6C905D63DD120013ULL, "hlac_lt_0_lowtown_a2_n_c_65534_5y.dds"},
	{0x6C906CB4DCA14612ULL, "850812495.ogg"},
	{0x6C906DAC5859738BULL, "6075559_m.fxe"},
	{0x6C9074FE987C0108ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_reaction.evt"},
	{0x6C90881AA1F54554ULL, "ht_0_hightownexplosion_n_4_0.phy"},
	{0x6C90AFC05DF99259ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_take.ani"},
	{0x6C90FAC75CE44A6CULL, "follower_banter6172285.cl"},
	{0x6C910F42F6B30A23ULL, "zz_vault_debug6200336.cl"},
	{0x6C911145FB295AE6ULL, "vfx_chunkyparticle_p5.psh.fxo"},
	{0x6C919F15E348131AULL, "vxlt_0_lowtown_docks_a3_d0e_0_0.vlm"},
	{0x6C91AEF726A80A77ULL, "428510803.ogg"},
	{0x6C91DFDBE8963505ULL, "6095797_m.fxe"},
	{0x6C92B11704B97981ULL, "253512384.ogg"},
	{0x6C92D9BD08500940ULL, "hlac_ht_chantry_2534_19y.dds"},
	{0x6C92DE9FBC3C3093ULL, "6101072_m.fxe"},
	{0x6C930A56B65A5BE4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_chab_n.dds"},
	{0x6C933853A2902E2EULL, "387161497.ogg"},
	{0x6C936E4D948D53A8ULL, "w_encounter2_8_0.mmh"},
	{0x6C937A4D948D684BULL, "w_encounter2_8_0.mao"},
	{0x6C9380D879B376B6ULL, "6054362_m.fxe"},
	{0x6C938C4D948D86D2ULL, "w_encounter2_8_0.msh"},
	{0x6C9406FDA34B9B79ULL, "vxht_bartrand_n_5o8_0_0.vlm"},
	{0x6C9415F35979A9B4ULL, "430421916.ogg"},
	{0x6C948F7BCAFC6F50ULL, "656380712.ogg"},
	{0x6C9512E803798D75ULL, "6108265_f.fxe"},
	{0x6C956E75A1BCADACULL, "vxw_encounter2_camp_dd_0_0.vlm"},
	{0x6C95F7A45B5EDD40ULL, "6119391_m.fxe"},
	{0x6C961032F8BB115AULL, "mcr370_meredith6084438.cl"},
	{0x6C9642655D17B2A7ULL, "follower_barks6153695.cl"},
	{0x6C9643C566474F72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_p.gad"},
	{0x6C96D53D40ED8368ULL, "gen00fl_varric6033648.cl"},
	{0x6C96EDF09A1018F9ULL, "vxlt_gallowscourtyard_3tz_0_0.vlm"},
	{0x6C97019D32BEF716ULL, "928332269.ogg"},
	{0x6C973546EF6B6859ULL, "6182758_m.fxe"},
	{0x6C977F1AA1FB54FCULL, "ht_0_hightownexplosion_n_4_0.rml"},
	{0x6C978AC05DFF719BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_take.gad"},
	{0x6C97D70BE7C229DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction2.ani"},
	{0x6C97E48143F72A17ULL, "353190543.ogg"},
	{0x6C97E763D3027937ULL, "vxht_0_hightown_8ay_0_0.vlm"},
	{0x6C98065A008254FBULL, "vxw_bonepit_k2_0_0.vlm"},
	{0x6C9854C650271BB2ULL, "983852876.ogg"},
	{0x6C989F83A630576CULL, "vxht_0_hightown_9v2_0_0.vlm"},
	{0x6C99262B69B8F71FULL, "865544047.ogg"},
	{0x6C99438CE3BC287EULL, "vxht_brothel_n_4f5_0_0.vlm"},
	{0x6C995356D542C1FBULL, "328914242.ogg"},
	{0x6C99905BBD93F39AULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_act_c.mmh"},
	{0x6C99F6FC43DC5076ULL, "zz_fol_debug6066145.cl"},
	{0x6C9A035CB39CFE36ULL, "and330tr_act3_opening.utt"},
	{0x6C9A5001E15B43CDULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyo.phy"},
	{0x6C9AFEAD36232A66ULL, "lt_0_lowtown_a2_52_0_lc.mmh"},
	{0x6C9B24EAB4F30ADDULL, "6077586_m.fxe"},
	{0x6C9B25426F32E33DULL, "vxw_bonepitdestroyed_a3_ox_0_0.vlm"},
	{0x6C9B2D7744947D01ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2.mmh"},
	{0x6C9B557E3B5F4EABULL, "vxca_cavern1_haunted_j2b_0_1.vlm"},
	{0x6C9B651AA1FF1151ULL, "ht_0_hightownexplosion_n_4_0.msh"},
	{0x6C9B731AA1FF296CULL, "ht_0_hightownexplosion_n_4_0.mao"},
	{0x6C9B7F1AA1FF3D0FULL, "ht_0_hightownexplosion_n_4_0.mmh"},
	{0x6C9C5EDA57D11BCAULL, "gen00fl_merrill6128704.cl"},
	{0x6C9CB6CE0C349AB2ULL, "546247950.ogg"},
	{0x6C9D46C5664D72A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_p.evt"},
	{0x6C9D4B5FBA83303EULL, "vxdr_primevalentrance_349_0_0.vlm"},
	{0x6C9DB32F633799E2ULL, "hlac_w_bonepitmining_a2_327684_0z.dds"},
	{0x6C9DCA10D21009D8ULL, "vxdr_primevalentrance_2gb_0_1.vlm"},
	{0x6C9DCD949E8CB9C7ULL, "pro000_refugee_male6052106.cl"},
	{0x6C9DE40BE7C6AB1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction2.gad"},
	{0x6C9E0C02C177E32EULL, "llac_w_bonepit_65539_0.dds"},
	{0x6C9EA7C05E05C180ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exchange_take.evt"},
	{0x6C9FB816ACF9B7F0ULL, "vxht_0_hightown_a3_a70_0_0.vlm"},
	{0x6C9FC9B04D8B3E87ULL, "334067334.ogg"},
	{0x6C9FE119D737B3E6ULL, "gen00fl_aveline6120855.cl"},
	{0x6CA0436D6BB6C08BULL, "vxw_woundedcoast_11t_0_0.vlm"},
	{0x6CA048C3C7AA87A4ULL, "hlac_lt_0_lowtown_a2_65537_6y.dds"},
	{0x6CA07096BBCF3EAFULL, "llac_ht_0_hightown_a2_131071_1.dds"},
	{0x6CA0CC4B2A23659AULL, "mrl150_merrill6085837.cl"},
	{0x6CA0EC74F5697DC4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww22_viscgate.msh"},
	{0x6CA0FE74F5699C2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww22_viscgate.mao"},
	{0x6CA10274F569A366ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww22_viscgate.mmh"},
	{0x6CA1594681266654ULL, "gen_im_trs_gen_blt_r1.uti"},
	{0x6CA1C0D7CA254EB3ULL, "312845460.ogg"},
	{0x6CA2CCA4657C9DFEULL, "vxlt_undercityserialkiller_yw_0_0.vlm"},
	{0x6CA33390F740BD35ULL, "6072674_f.fxe"},
	{0x6CA36BDAEBF4CB76ULL, "gen00fl_anders6119517.cl"},
	{0x6CA3CCE1DD77EC53ULL, "6086370_f.fxe"},
	{0x6CA3FADC5E687B8FULL, "character_character_p17.psh.fxo"},
	{0x6CA442EFC3B704F2ULL, "6025152_f.fxe"},
	{0x6CA484951C2D0477ULL, "1054718993.ogg"},
	{0x6CA48CEC821E9BDDULL, "598861042.ogg"},
	{0x6CA4BF0BE7CC8B3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_dieaction2.evt"},
	{0x6CA4FF9B33A8F463ULL, "vxlt_0_lowtown_a3_n_c_i0k_0_0.vlm"},
	{0x6CA633AEE4CCBD05ULL, "hlac_lt_0_lowtown_docks_a3_n_-196608_0z.dds"},
	{0x6CA6416CCC6C7F45ULL, "qcr000_varric6208243.cl"},
	{0x6CA69B10B663C9E6ULL, "6088666_m.fxe"},
	{0x6CA6D83A8C8A550FULL, "qun131_party_comment6117277.cl"},
	{0x6CA6F85DD289DA00ULL, "vxw_encounter2_camp_e6_0_1.vlm"},
	{0x6CA6F8F52A7E250AULL, "dae000_arishok6081656.cl"},
	{0x6CA71D26FA5B5ABEULL, "607827206.ogg"},
	{0x6CA790CC63119D93ULL, "6155604_m.fxe"},
	{0x6CA7B4FCB1A238F4ULL, "qa_useobject.ncs"},
	{0x6CA7DBFE4BF5A762ULL, "vxw_woundedcoast_1bd_0_0.vlm"},
	{0x6CA7F0A7CE812ECEULL, "vxht_mansion2b_1me_0_0.vlm"},
	{0x6CA81FA62F20172DULL, "189813799.ogg"},
	{0x6CA83622E323E320ULL, "hlac_ca_caverns1_mines_24425_1z.dds"},
	{0x6CA8A56EB1488E98ULL, "llac_ca_caverns1_mines_24424_4.dds"},
	{0x6CA9877A7F07B275ULL, "hlac_lt_0_lowtown_docks_65534_2x.dds"},
	{0x6CAA2B44112EC5D4ULL, "lgt210_eau_symp_templar.cnv"},
	{0x6CAAF974F571E831ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww22_viscgate.phy"},
	{0x6CAB01E5D62312B6ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.impa.evt"},
	{0x6CAB229C238A56DDULL, "hlac_lt_0_lowtown_a2_n_65533_1x.dds"},
	{0x6CAB2EC566598718ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l90_p.ani"},
	{0x6CAB650EBE2948D4ULL, "qun231_party_comment6077009.cl"},
	{0x6CAC1E1B1BF713E7ULL, "869094243.ogg"},
	{0x6CAD0D441130C814ULL, "lgt210_eau_symp_templar.bnk"},
	{0x6CAD0EB45BD2205AULL, "dae210ip_at_mrl_home.utp"},
	{0x6CAD96C4617B355FULL, "vxca_caverns1_e87_0_0.vlm"},
	{0x6CAE714729F44439ULL, "6064586_m.fxe"},
	{0x6CAF4FDDA002518BULL, "738380255.ogg"},
	{0x6CAF6C98AC4294D5ULL, "691295553.ogg"},
	{0x6CB014BAB4A17693ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_enter.evt"},
	{0x6CB0A330D1CC6F43ULL, "961720547.ogg"},
	{0x6CB0AFF686C182C4ULL, "6013440_m.fxe"},
	{0x6CB0B00BFDE4E143ULL, "vxlt_0_lowtown_a2_n_c_hz1_0_0.vlm"},
	{0x6CB0CF1BFB0DC300ULL, "vxlt_gallowstemplar_f_5cy_0_0.vlm"},
	{0x6CB1AE741EEC6031ULL, "hlac_lt_undercityserialkiller_-1_25x.dds"},
	{0x6CB1F8E5D6292241ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.impa.gad"},
	{0x6CB21744D6CB68D7ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd03a.mmh"},
	{0x6CB22D44D6CB8E49ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd03a.msh"},
	{0x6CB2443E72579763ULL, "vxca_caverns1_ghq_0_0.vlm"},
	{0x6CB292BAC53D0E7AULL, "553371796.ogg"},
	{0x6CB2F2BF66088D38ULL, "vxht_brothel_n_4hi_0_0.vlm"},
	{0x6CB3166C0E1C5588ULL, "gen00fl_fenris6126991.cl"},
	{0x6CB33DD640D0010AULL, "llac_lt_0_lowtown_a2_1_2.dds"},
	{0x6CB425C4A89F76BBULL, "vxca_caverns1_mines_j21_0_2.vlm"},
	{0x6CB44CECF14813BAULL, "gen00fl_bethany6097837.cl"},
	{0x6CB4AB12EA07BED6ULL, "hlac_lt_0_lowtown_a2_n_-196608_1x.dds"},
	{0x6CB50CEDA859E266ULL, "6083766_m.fxe"},
	{0x6CB558FABF559CF4ULL, "weapon1_ds2.psh.fxo"},
	{0x6CB5E996D9190DA9ULL, "550603178.ogg"},
	{0x6CB60DBAB4A5D6D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_enter.gad"},
	{0x6CB64DDB3DEBBFE2ULL, "vxw_woundedcoast_1me_0_0.vlm"},
	{0x6CB6E4D9DF5445ABULL, "203998927.ogg"},
	{0x6CB7EA1842F7E71CULL, "554441232.ogg"},
	{0x6CB8705D95E22766ULL, "hlac_ht_keep_n_4519_5x.dds"},
	{0x6CB8F9E5D62F42EFULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.impa.ani"},
	{0x6CB92EC911405733ULL, "art\\vfx\\combat\\vfx_r_gn3_embu_appear_p.mmh"},
	{0x6CBAB7C0DEE4254DULL, "vxca_caverns2prc_8cs_0_0.vlm"},
	{0x6CBB195D9F7DFBA3ULL, "783188516.ogg"},
	{0x6CBB4162F33A297BULL, "650395908.ogg"},
	{0x6CBB7DE5E96C1ACAULL, "721293376.ogg"},
	{0x6CBBAC380877BD66ULL, "6146814_m.fxe"},
	{0x6CBBB6645B99D0C3ULL, "499557190.ogg"},
	{0x6CBBEFF69E3D067CULL, "vxht_brothel_n_3ue_0_0.vlm"},
	{0x6CBC0CF4A8338DB7ULL, "hlac_lt_0_lowtown_n_-2_1y.dds"},
	{0x6CBC208D557118CFULL, "23936473.ogg"},
	{0x6CBC49BCFF54C8EFULL, "239979824.ogg"},
	{0x6CBC95AC153479C5ULL, "dae000_mother6004203.cl"},
	{0x6CBCE1570241D474ULL, "gen00fl_anders6111340.cl"},
	{0x6CBD1CBAB4AC0EB2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_enter.ani"},
	{0x6CBE43F820323AC2ULL, "hlac_ht_mansion_n_3368_0x.dds"},
	{0x6CBE86404D296346ULL, "6036498_m.fxe"},
	{0x6CBE942C1991F7A9ULL, "hlac_lt_0_lowtown_docks_-131074_0x.dds"},
	{0x6CBED96B6F2CDC16ULL, "579261457.ogg"},
	{0x6CBF093245AB39E1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medj_0s.dds"},
	{0x6CBF113AA94D49C0ULL, "atl_chargen_dxt5_01.dds"},
	{0x6CBF2848A5139C9EULL, "186928475.ogg"},
	{0x6CBF3C01B9054004ULL, "gen_im_trs_dwv_junk1_03.uti"},
	{0x6CBF97AB9FA2CBF3ULL, "zz_ser_debug6086473.cl"},
	{0x6CBF9ABE7B2BF87FULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumplrg.msh"},
	{0x6CBFA0BE7B2C0221ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumplrg.mmh"},
	{0x6CBFC1A10AFC31EBULL, "art\\vfx\\combat\\vfx_gib_hm_arm_ltr.ani"},
	{0x6CBFFA110D519821ULL, "102244000.ogg"},
	{0x6CBFFE7EA6EB4C16ULL, "vxdr_primevaltomb_71b_0_0.vlm"},
	{0x6CC02F4866AA5308ULL, "qcr220_orsino6061352.cl"},
	{0x6CC1472528B7B0F6ULL, "hightown_fem_courtesan_mob6168738.cl"},
	{0x6CC16A7F2D6B2B9EULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.cess.ani"},
	{0x6CC17771EEC19BF2ULL, "910411622.ogg"},
	{0x6CC182CED8F27263ULL, "6129420_f.fxe"},
	{0x6CC18EE55BEEF571ULL, "qun221_petrice6077940.cl"},
	{0x6CC19D8597CE334EULL, "6088745_m.fxe"},
	{0x6CC1B20046D9B62DULL, "song_dot.mao"},
	{0x6CC1C1FD53E19AA4ULL, "6198669_f.fxe"},
	{0x6CC1FC1E5928095BULL, "6149469_m.fxe"},
	{0x6CC242B82AFEE70DULL, "6119592_m.fxe"},
	{0x6CC2C2304F504FB3ULL, "dae000_flemeth6000959.cl"},
	{0x6CC34C35642B0595ULL, "vxht_mansion2_28b_0_0.vlm"},
	{0x6CC3EF294FB243E5ULL, "dae000ip_crowd_lm4_talk1.utp"},
	{0x6CC3F9F6C59FC39EULL, "381402110.ogg"},
	{0x6CC47B360503966EULL, "character_character_p234.psh.fxo.dx11"},
	{0x6CC4BFB82FE1D862ULL, "vxw_woundedcoast_1hf_0_0.vlm"},
	{0x6CC50B984139D05EULL, "6179889_m.fxe"},
	{0x6CC5616F3C82539CULL, "6060144_f.fxe"},
	{0x6CC5D44C47F490F6ULL, "hlac_w_sundermount_0_7z.dds"},
	{0x6CC5E232761D72FDULL, "hlac_ht_brothel_n_3368_0x.dds"},
	{0x6CC6258BAA5E2C91ULL, "483612748.ogg"},
	{0x6CC71C503CCE6306ULL, "vxdr_primevaltomb_6qo_0_0.vlm"},
	{0x6CC72743F10538F8ULL, "gen_im_npc_arm_hum_boo_cotb.uti"},
	{0x6CC72E7744B9BA7EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2.phy"},
	{0x6CC7EF8475D6161BULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfireout.dura.evt"},
	{0x6CC819328A9DF956ULL, "6128883_m.fxe"},
	{0x6CC91BCB8FEC5135ULL, "6136667_m.fxe"},
	{0x6CC98E3B9B9323FBULL, "891057777.ogg"},
	{0x6CC9B2DA2D8F5304ULL, "llac_lt_0_lowtown_a2_n_-131071_0.dds"},
	{0x6CC9C3B169511526ULL, "zz_followers_banter6159000.cl"},
	{0x6CC9EC933BE8AA19ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_0z.dds"},
	{0x6CCA23D651B45CFEULL, "vxdr_primevaltomb_6yo_0_0.vlm"},
	{0x6CCA74020BA6BBBDULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appp_0n.dds"},
	{0x6CCB210500E0B166ULL, "874367166.ogg"},
	{0x6CCB70DE8FFC8678ULL, "d15_b_shot_a.dds"},
	{0x6CCB7263789F802DULL, "gen00fl_merrill6150917.cl"},
	{0x6CCBECF0961323A5ULL, "hla_ht_0_hightown_a2_131073_1y.dds"},
	{0x6CCC214836863686ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quarterb.phy"},
	{0x6CCC46E49A792B23ULL, "670950716.ogg"},
	{0x6CCCD714D2E91F4FULL, "1048101833.ogg"},
	{0x6CCD0B41897C8879ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_0z.dds"},
	{0x6CCD58F2A8F9842FULL, "gen_im_trs_qun_hvy_cht_r1.uti"},
	{0x6CCD93C5E1D11243ULL, "vxlt_0_lowtown_a2_n_c_hp3_0_0.vlm"},
	{0x6CCE925DE8C0E0ACULL, "6084310_m.fxe"},
	{0x6CCEB54DA9CE6214ULL, "hlac_ht_mansion_n_3370_1x.dds"},
	{0x6CCEB5AB6120E260ULL, "6174945_m.fxe"},
	{0x6CCEE88475DC295CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfireout.dura.gad"},
	{0x6CCF16C1A31A72CFULL, "6013289_m.fxe"},
	{0x6CCF3C2527E041ECULL, "96949018.ogg"},
	{0x6CCF4C30E36A74C2ULL, "974077228.ogg"},
	{0x6CCF5685FE4CA2DEULL, "hlac_ht_0_hightown_-196608_2x.dds"},
	{0x6CCF78CE05ACFE57ULL, "vxlt_gallowstemplar_n_6sd_0_1.vlm"},
	{0x6CCF827F2D77926FULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.cess.evt"},
	{0x6CD052D4262257D1ULL, "6027749_m.fxe"},
	{0x6CD0573C318F49CFULL, "zz_mer_debug6023993.cl"},
	{0x6CD06DFB4B1E21DBULL, "6142569_m.fxe"},
	{0x6CD0B101C91A9A20ULL, "729013062.ogg"},
	{0x6CD13304A7EC338FULL, "vxht_mansion2b_1t7_0_0.vlm"},
	{0x6CD153CA0F4A47FFULL, "mag200_feynriel6023954.cl"},
	{0x6CD15BB501630440ULL, "dae000_mother6103057.cl"},
	{0x6CD15FB64C0F7F60ULL, "hlac_w_woundedcoast_65538_0x.dds"},
	{0x6CD2516899A6E4EFULL, "vxlt_0_lowtown_n_e8h_0_0.vlm"},
	{0x6CD25BCDBAD21F87ULL, "llac_dr_dungeon_3214_7.dds"},
	{0x6CD3319521D7F0E2ULL, "vxw_woundedcoast_1sg_0_0.vlm"},
	{0x6CD3A831D9A1C966ULL, "277906542.ogg"},
	{0x6CD3A8A10B0CB075ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_ltr.gad"},
	{0x6CD3CB4209E17755ULL, "vlowlevelloadhints.gda"},
	{0x6CD411FA3DED50A5ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\gibgolem_overlay.anb"},
	{0x6CD42EFD72A00D38ULL, "gen00fl_merrill_sundermount.stg"},
	{0x6CD4846F94F58EFBULL, "vxlt_warehouse_n_10f_0_0.vlm"},
	{0x6CD4B16EEA328175ULL, "vxht_0_hightown_a3_c2p_0_0.vlm"},
	{0x6CD4E78475E0935AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfireout.dura.ani"},
	{0x6CD4FFAC271B4F15ULL, "art\\levels\\ztools\\default\\generic\\0di_basic.phy"},
	{0x6CD52D0C10B39D39ULL, "hlac_lt_0_lowtown_a2_n_c_65534_4x.dds"},
	{0x6CD5307CABAF779DULL, "6004198.cut"},
	{0x6CD5837F2D7BFF18ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.cess.gad"},
	{0x6CD6C3CDCD170F3BULL, "w_bonepitmining_a2_151_0.msh"},
	{0x6CD6D5CDCD172D86ULL, "w_bonepitmining_a2_151_0.mao"},
	{0x6CD6D8C071185159ULL, "hlac_lt_0_lowtown_n_1_3y.dds"},
	{0x6CD6E1CDCD174225ULL, "w_bonepitmining_a2_151_0.mmh"},
	{0x6CD6EE8D55B36CD3ULL, "6077009_m.fxe"},
	{0x6CD796FE2AF320EEULL, "6150004_m.fxe"},
	{0x6CD7C681FFA39EC6ULL, "mer_smo.mao"},
	{0x6CD87B75B87F9F4BULL, "art\\characters\\creatures\\ragedemon\\animation\\crnf.ani"},
	{0x6CD8E0BFF1E01B27ULL, "6179857_m.fxe"},
	{0x6CD9483F6AAF9199ULL, "vxlt_undercityencounter_17m_0_0.vlm"},
	{0x6CD97CC6EF8C8AD6ULL, "one300_zevran6093497.cl"},
	{0x6CD97F0AE0518281ULL, "lowtown_male_urchin_sta6164197.cl"},
	{0x6CD988EBE7914A72ULL, "118963668.ogg"},
	{0x6CD98C27C281190DULL, "785353107.ogg"},
	{0x6CD9EB805A4A9EDEULL, "90557563.ogg"},
	{0x6CDA15292AFF209FULL, "6080102_m.fxe"},
	{0x6CDA5CA1B403C3A7ULL, "189294026.ogg"},
	{0x6CDA5F6E36A20233ULL, "llac_lt_0_lowtown_docks_-65537_0.dds"},
	{0x6CDA9B2ADFE7DD0EULL, "gen_im_cns_pot_respec_01.uti"},
	{0x6CDAD431750E2684ULL, "zz_seb_debug6190076.cl"},
	{0x6CDAE5440FEBEF4BULL, "art\\characters\\creatures\\pridedemon\\animation\\atk_v1.ani"},
	{0x6CDAF434717CD893ULL, "vxlt_0_lowtown_docks_a3_n_cyc_0_0.vlm"},
	{0x6CDAFA92C986CBF7ULL, "lt_0_lowtown_docks_n_181_0.msh"},
	{0x6CDB0C92C986EA5AULL, "lt_0_lowtown_docks_n_181_0.mao"},
	{0x6CDB1092C986F169ULL, "lt_0_lowtown_docks_n_181_0.mmh"},
	{0x6CDB59EE48421111ULL, "6042069_m.fxe"},
	{0x6CDB5FCC966702F0ULL, "584271649.ogg"},
	{0x6CDB875A0721FA0EULL, "6149279_m.fxe"},
	{0x6CDB9948AD2FBFADULL, "follower_banter6155964.cl"},
	{0x6CDBACA6DB4A048BULL, "461413695.ogg"},
	{0x6CDBB8314A5434DFULL, "w_woundedcoast_204_0.msh"},
	{0x6CDBBE314A543E81ULL, "w_woundedcoast_204_0.mmh"},
	{0x6CDBCA314A545322ULL, "w_woundedcoast_204_0.mao"},
	{0x6CDCA89983A6B54CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\.msh"},
	{0x6CDCAE9983A6BF3EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\.mmh"},
	{0x6CDD343B060E1FC6ULL, "fall_dust.mao"},
	{0x6CDD7E9662DFFC7DULL, "vxht_0_hightown_ak8_0_0.vlm"},
	{0x6CDDEE58BDD2D6E0ULL, "6031156_m.fxe"},
	{0x6CDE48563BEAD6B8ULL, "vxht_brothel_n_4wd_0_0.vlm"},
	{0x6CDEBF8489603B8DULL, "vxht_0_hightown_n_9ay_0_0.vlm"},
	{0x6CDEC6084C2ACB6CULL, "ht_0_hightown_n_25_0_ll.mmh"},
	{0x6CDEDF38DF78125AULL, "909649779.ogg"},
	{0x6CDEFC39F11BE0EBULL, "6080252_m.fxe"},
	{0x6CDF20B854C8A852ULL, "6075121_m.fxe"},
	{0x6CDF5E7E3F19498BULL, "6040605_m.fxe"},
	{0x6CDFE4A247E14F9AULL, "ave310_donnic6134749.cl"},
	{0x6CE0637D3FF7F57EULL, "1055955871.ogg"},
	{0x6CE073D62BA7A272ULL, "200275305.ogg"},
	{0x6CE10A5174393707ULL, "llac_lt_0_lowtown_a3_n_c_-131072_7.dds"},
	{0x6CE17770DF435BA6ULL, "1065528144.ogg"},
	{0x6CE191791AFF0591ULL, "hlac_lt_0_lowtown_docks_-131075_7z.dds"},
	{0x6CE1B81B965990CEULL, "vxht_mansion2b_1qy_0_0.vlm"},
	{0x6CE1F4D8EE0CA742ULL, "6102141_f.fxe"},
	{0x6CE2DC120D8CB717ULL, "162771872.ogg"},
	{0x6CE396DAE55CFDEEULL, "plt_genpt_dog.plo"},
	{0x6CE3D03AFBF3EEC9ULL, "qcr000_arishok6060111.cl"},
	{0x6CE3F5A07B1ADCEEULL, "hlac_ca_cave_2194_9z.dds"},
	{0x6CE408F7072547F5ULL, "hlac_ht_mansion2_818_25z.dds"},
	{0x6CE41961B857FF17ULL, "6116520_m.fxe"},
	{0x6CE4658BD745F00CULL, "hlac_lt_darktown_-1_87y.dds"},
	{0x6CE4E1A6B8BEDCC9ULL, "442402953.ogg"},
	{0x6CE4F030B6E4FB44ULL, "gen00fl_merrill6129543.cl"},
	{0x6CE5048948D0C9F0ULL, "vxht_0_hightown_a3_byu_0_0.vlm"},
	{0x6CE508C692F52BCFULL, "228938814.ogg"},
	{0x6CE51B42A57993ADULL, "art\\levels\\ztools\\default\\generic\\0oi_basic.msh"},
	{0x6CE52942A579ABE0ULL, "art\\levels\\ztools\\default\\generic\\0oi_basic.mao"},
	{0x6CE52D42A579B233ULL, "art\\levels\\ztools\\default\\generic\\0oi_basic.mmh"},
	{0x6CE57E1BAFF00277ULL, "6148394_m.fxe"},
	{0x6CE66375B88BB3C2ULL, "art\\characters\\creatures\\ragedemon\\animation\\crnf.evt"},
	{0x6CE6934C97A07272ULL, "6015620_m.fxe"},
	{0x6CE6B3F16436B374ULL, "vxht_keep_70f_0_0.vlm"},
	{0x6CE743DA056C9C7BULL, "vxlt_0_lowtown_g92_0_0.vlm"},
	{0x6CE7762E40A185AFULL, "vxlt_0_lowtown_a2_n_bza_0_0.vlm"},
	{0x6CE7E1BB05B69CABULL, "vxw_bonepit_za_0_0.vlm"},
	{0x6CE7EDA14608DADAULL, "6169952_m.fxe"},
	{0x6CE814CCB2643633ULL, "348745877.ogg"},
	{0x6CE8CD440FF803C2ULL, "art\\characters\\creatures\\pridedemon\\animation\\atk_v1.evt"},
	{0x6CE926A6F4EA40E2ULL, "mag110_vincento6166829.cl"},
	{0x6CE94F13F8E88543ULL, "vxw_sundermount_35o_0_1.vlm"},
	{0x6CE98263A65E2914ULL, "6188628_f.fxe"},
	{0x6CE996856AE81E4CULL, "389780069.ogg"},
	{0x6CE99C28893729E9ULL, "clf260_harley6069658.cl"},
	{0x6CE9AFD15C33C8E0ULL, "gen00fl_anders6142555.cl"},
	{0x6CE9DD5B96B13AC4ULL, "ttx_arch_inner_2.mmh"},
	{0x6CE9E90AAB13CD2CULL, "6070249_m.fxe"},
	{0x6CE9F35B96B16066ULL, "ttx_arch_inner_2.msh"},
	{0x6CEA930CD04502B5ULL, "hlac_lt_0_lowtown_docks_0_4y.dds"},
	{0x6CEAC32AA9930A61ULL, "vxlt_warehouse_n_rm_0_0.vlm"},
	{0x6CEB6F7EA4D8CD3EULL, "zz_playanimation.ncs"},
	{0x6CEC2DAE01BA22A3ULL, "6093915_m.fxe"},
	{0x6CEC6275B8901DD5ULL, "art\\characters\\creatures\\ragedemon\\animation\\crnf.gad"},
	{0x6CECA1BE7B52F19EULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumplrg.phy"},
	{0x6CECB0F944C04161ULL, "245742470.ogg"},
	{0x6CED082580F1BFC1ULL, "ser221_quentin6131253.cl"},
	{0x6CED51FCD8BDE097ULL, "zz_vault_debug6203166.cl"},
	{0x6CED7531872DE764ULL, "82820657.ogg"},
	{0x6CEDE58EBA45CB95ULL, "vxlt_hangedman_s7_0_0.vlm"},
	{0x6CEECC440FFC6DD5ULL, "art\\characters\\creatures\\pridedemon\\animation\\atk_v1.gad"},
	{0x6CEEF55D6D96EA4BULL, "vxca_cavern1_haunted_co9_0_0.vlm"},
	{0x6CEF124FDCD26030ULL, "6043562_m.fxe"},
	{0x6CEF22530B4C0E48ULL, "gen00fl_anders6104012.cl"},
	{0x6CEF3346D73E62ACULL, "vxca_caverns1_mines_ei7_0_1.vlm"},
	{0x6CF029F8884FA94EULL, "vxht_mansion2b_1tz_0_0.vlm"},
	{0x6CF0396A27E560E2ULL, "669974317.ogg"},
	{0x6CF04BB19F365EF7ULL, "llac_lt_0_lowtown_a2_n_c_65536_8.dds"},
	{0x6CF05592E7DDE489ULL, "mag321_comte_de_launcet6064925.cl"},
	{0x6CF097F181E9F690ULL, "zz_bet_debug6113397.cl"},
	{0x6CF0FE00C3466B57ULL, "zz_vault_debug6177605.cl"},
	{0x6CF21443C0259CA6ULL, "vxlt_0_lowtown_a2_n_c_d5u_0_0.vlm"},
	{0x6CF2353D9721F9D9ULL, "hlac_ht_0_hightown_n_65536_2z.dds"},
	{0x6CF26FE198FA2434ULL, "6196721_m.fxe"},
	{0x6CF29659A151BF73ULL, "6094994_m.fxe"},
	{0x6CF2BA9B2DC011E0ULL, "isa211_isabela6069402.cl"},
	{0x6CF2CBE858B99C63ULL, "vxlt_0_lowtown_docks_a3_d3m_0_0.vlm"},
	{0x6CF386BA010487E9ULL, "art\\characters\\heraldry\\hld_slavers.dds"},
	{0x6CF3EFF8A1E61AF7ULL, "6145094_m.fxe"},
	{0x6CF44BE06AE7B648ULL, "vxlt_warehouse_n_rh_0_0.vlm"},
	{0x6CF47F34A6BC874DULL, "vxca_caverns2prc_8wo_0_0.vlm"},
	{0x6CF4895354F43566ULL, "35900155.ogg"},
	{0x6CF4AADFF4F3EBB1ULL, "mrl350ip_pile_a.utp"},
	{0x6CF4CA83235A3FE5ULL, "vxht_mansion2_282_0_0.vlm"},
	{0x6CF5578D932756BAULL, "6162049_m.fxe"},
	{0x6CF5786C369218E3ULL, "lgt210_qun_blk_qunari6142312.cl"},
	{0x6CF68C18530C9CCBULL, "vxht_0_hightown_a2_cb3_0_0.vlm"},
	{0x6CF7130CBB4FC79AULL, "hlac_w_encounter2_camp_196607_0y.dds"},
	{0x6CF787DAE56D8D7DULL, "plt_genpt_dog.nss"},
	{0x6CF7D03B09A603C2ULL, "217440408.ogg"},
	{0x6CF7DEE0CAE4015CULL, "zz_followers_banter6157578.cl"},
	{0x6CF7F7BE875748D6ULL, "vxw_bonepitmining_a2_1kj_0_0.vlm"},
	{0x6CF8277A89C28574ULL, "lt_0_lowtown_docks_30_0_ll.mmh"},
	{0x6CF85EE4742D23F9ULL, "6070767_m.fxe"},
	{0x6CF8FBF5FF192F2AULL, "vxht_0_hightown_n_c4c_0_0.vlm"},
	{0x6CF924A8DDBD7A3CULL, "vxw_sundermount_32p_0_0.vlm"},
	{0x6CF935079B8922E1ULL, "vxlt_warehouse_n_yl_0_0.vlm"},
	{0x6CF93A68A4135936ULL, "lt_0_lowtown_docks_a3_179_0.msh"},
	{0x6CF94468A4136A34ULL, "lt_0_lowtown_docks_a3_179_0.mmh"},
	{0x6CF94868A4137147ULL, "lt_0_lowtown_docks_a3_179_0.mao"},
	{0x6CF9CD4D2E055C12ULL, "vxca_caverns1_mines_ck8_0_0.vlm"},
	{0x6CF9CE4592BDA7F4ULL, "vxlt_gallowstemplar_f_5nk_0_1.vlm"},
	{0x6CF9F076BE75009EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_loop.ani"},
	{0x6CFA0830B1C0305CULL, "788234.ogg"},
	{0x6CFA2547A4701A6FULL, "dae000_tomwise6167307.cl"},
	{0x6CFA633387B104D2ULL, "hlac_lt_0_lowtown_docks_-4_3z.dds"},
	{0x6CFABF100B4F75D2ULL, "hlac_dr_primevaltomb_663_1z.dds"},
	{0x6CFB1620A5341F79ULL, "6027604_m.fxe"},
	{0x6CFB9F74522732B6ULL, "238977219.ogg"},
	{0x6CFBB7387AAE2689ULL, "6106170_m.fxe"},
	{0x6CFBBFFDA1FF8C82ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hanginglite.msh"},
	{0x6CFBC1FDA1FF9038ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hanginglite.mmh"},
	{0x6CFBCDFDA1FFA4DBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hanginglite.mao"},
	{0x6CFC6E6B9D1F3200ULL, "t1_mue_gr2.tnt"},
	{0x6CFCF328AE012AF6ULL, "6000899_m.fxe"},
	{0x6CFD995BDC67C588ULL, "6050125_f.fxe"},
	{0x6CFDC275DADC1E8BULL, "hlac_ht_0_hightown_n_-131074_2x.dds"},
	{0x6CFDE2BBFCBC79E1ULL, "vxlt_darktown_4ud_0_0.vlm"},
	{0x6CFE5C4DD4DE8591ULL, "dre000_sibling_has_blight6037047.cl"},
	{0x6CFF0880CA5D5FB9ULL, "312018499.ogg"},
	{0x6CFFB8464A020104ULL, "art\\characters\\creatures\\strider\\animation\\disappear.ani"},
	{0x6D0055CF04FF4C2BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_loop.evt"},
	{0x6D005E6EBE9A11EAULL, "follower_banter6136733.cl"},
	{0x6D008D319CFA5B78ULL, "gencr_coterie_archer.utc"},
	{0x6D00CC9CDBD50F0DULL, "hlac_lt_0_lowtown_n_-2_0x.dds"},
	{0x6D00F2F041F75975ULL, "vxlt_hangedman_n_61_0_0.vlm"},
	{0x6D013DC54AAFB4E3ULL, "vxlt_0_lowtown_docks_a3_d4l_0_0.vlm"},
	{0x6D01569CCCE0BEB4ULL, "hlac_ht_0_hightown_a3_-196608_1z.dds"},
	{0x6D01CEE49373766EULL, "hlac_lt_backalley_g_0_13x.dds"},
	{0x6D01D14AC6D509D9ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_arh_0n.dds"},
	{0x6D024E7F20BF45F8ULL, "vxlt_gallowscourtyard_63s_0_0.vlm"},
	{0x6D028581533828EBULL, "6117627_f.fxe"},
	{0x6D02BDBD5CDDCEC8ULL, "vxlt_warehouse_n_yi_0_0.vlm"},
	{0x6D02C757EB2D8885ULL, "6107407_m.fxe"},
	{0x6D02D15A2CAEE5DDULL, "6126746_f.fxe"},
	{0x6D02F6A23715731BULL, "sst_ss_abomination_female.gda"},
	{0x6D0386CE024AB396ULL, "460280778.ogg"},
	{0x6D05C8FDA207F0C3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hanginglite.phy"},
	{0x6D05FB16552CC4A3ULL, "hla_ht_0_hightown_a3_131073_0y.dds"},
	{0x6D06B5464A081B66ULL, "art\\characters\\creatures\\strider\\animation\\disappear.gad"},
	{0x6D06DB2BB9A5506EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkleft.ani"},
	{0x6D070FF9F280BD7AULL, "6083748_m.fxe"},
	{0x6D0724363B1DE5F5ULL, "art\\vfx\\combat\\vfx_glb_ele_p_additive_c.mmh"},
	{0x6D076B71BE1125C8ULL, "vxlt_darktown_4ua_0_0.vlm"},
	{0x6D076ECF050595ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_loop.gad"},
	{0x6D077C716BC221F9ULL, "dae000_arishok6067752.cl"},
	{0x6D0782294859489DULL, "661391960.ogg"},
	{0x6D07FD42E2FF9584ULL, "dae000ip_crowd_ff3_talk15.utp"},
	{0x6D085B1816E19FE4ULL, "609489112.ogg"},
	{0x6D08DBFFC2D6B833ULL, "zz_dae_debug6029306.cl"},
	{0x6D0970B47388D7D4ULL, "637510617.ogg"},
	{0x6D09C24DD0A0E141ULL, "vxlt_0_lowtown_docks_a3_cyz_0_0.vlm"},
	{0x6D0A22101FCB6666ULL, "hlac_lt_0_lowtown_a2_65536_9y.dds"},
	{0x6D0A7282502236B9ULL, "vxht_chantry_30o_0_0.vlm"},
	{0x6D0AA4320A0C44A8ULL, "1002272307.ogg"},
	{0x6D0AACA916558306ULL, "6072850_m.fxe"},
	{0x6D0B2B4806822460ULL, "6025139_m.fxe"},
	{0x6D0B38AFB0B19AAEULL, "115869325.ogg"},
	{0x6D0BA386856C1F2EULL, "678546475.ogg"},
	{0x6D0CB0464A0C7E45ULL, "art\\characters\\creatures\\strider\\animation\\disappear.evt"},
	{0x6D0D2A223B89F07AULL, "6127617_m.fxe"},
	{0x6D0D362599D8C97FULL, "w_encounter2_camp_ll.mmh"},
	{0x6D0D5508E019C107ULL, "zz_dae_debug6049973.cl"},
	{0x6D0DB8DF2BAD35F3ULL, "33497292.ogg"},
	{0x6D0E0976BE85D418ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_loop.gad"},
	{0x6D0E6414DAF59647ULL, "vxlt_0_lowtown_a2_n_c_hd3_0_0.vlm"},
	{0x6D0E6DCF050BB2EAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_loop.ani"},
	{0x6D0E8F8EB97F8E50ULL, "hlac_ca_caverns1_24424_9z.dds"},
	{0x6D0EBD88E5492AECULL, "w_bonepit_231_0.msh"},
	{0x6D0EC388E54934DEULL, "w_bonepit_231_0.mmh"},
	{0x6D0ECF88E5494975ULL, "w_bonepit_231_0.mao"},
	{0x6D0EE3727F800826ULL, "hlac_w_bonepitmining_a2_262150_0z.dds"},
	{0x6D0EE55CFE4101BAULL, "gen00fl_varric6087595.cl"},
	{0x6D0F27FED718037FULL, "art\\vfx\\environment\\vfx_fire_cndl_p.mmh"},
	{0x6D0F74F63D714AE0ULL, "hlac_w_woundedcoast_-196602_0x.dds"},
	{0x6D0F8138E53D469FULL, "llac_ht_mansion2b_818_67.dds"},
	{0x6D0FA1520575C518ULL, "gen00fl_carver6103408.cl"},
	{0x6D104CB5103A1CF3ULL, "6069580_f.fxe"},
	{0x6D10C5D87CAA19D5ULL, "dae300_bran6144948.cl"},
	{0x6D10F5F4ADCA4992ULL, "676031511.ogg"},
	{0x6D1212746D8028FAULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\large_n.dds"},
	{0x6D139E29161634B3ULL, "vxht_mansion2b_1sf_0_0.vlm"},
	{0x6D13D32BB9AFCCBFULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkleft.evt"},
	{0x6D13D3CAB2BA70ADULL, "6139957_f.fxe"},
	{0x6D1421363B286A82ULL, "art\\vfx\\combat\\vfx_glb_ele_p_additive_c.anb"},
	{0x6D147A1235A47084ULL, "vxlt_0_lowtown_a2_n_f66_0_0.vlm"},
	{0x6D148E83AB58F4FAULL, "6156485_f.fxe"},
	{0x6D15CE113FC57466ULL, "1064406296.ogg"},
	{0x6D15EAC608149543ULL, "follower_banter6136334.cl"},
	{0x6D1612B784ACDDB3ULL, "qun220cr_javaris6071715.cl"},
	{0x6D161C6CD6C5ABA1ULL, "6173422_m.fxe"},
	{0x6D161D00C6EC1222ULL, "vxlt_backalley_n_1lj_0_0.vlm"},
	{0x6D16212AA16B3302ULL, "llac_lt_undercityserialkiller_-1_26.dds"},
	{0x6D165036CFDF7C89ULL, "6137469_f.fxe"},
	{0x6D16820DCDD3EC76ULL, "6152742_f.fxe"},
	{0x6D16AB24F1458D5EULL, "dae000cr_grand_cleric6147892.cl"},
	{0x6D16F48D8303AF43ULL, "vxht_0_hightown_a3_9zl_0_0.vlm"},
	{0x6D172BE3168CB2F3ULL, "w_blightlands_139_0.phy"},
	{0x6D174450E1A29B82ULL, "hlac_lt_undercitydungeon_-1_37y.dds"},
	{0x6D17D3E55D5574E9ULL, "hla_ht_0_hightown_n_-131070_0y.dds"},
	{0x6D1886F851389D8DULL, "vxht_0_hightown_n_9uu_0_0.vlm"},
	{0x6D192CBD5DB5BE3DULL, "setvaultflag.ncs"},
	{0x6D19756E2E5583F9ULL, "6169357_m.fxe"},
	{0x6D19DD240B7A7CCCULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.impa.ani"},
	{0x6D1A73903A00CD2BULL, "vxlt_0_lowtown_n_f5l_0_0.vlm"},
	{0x6D1AB42BB9B5B6A8ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkleft.gad"},
	{0x6D1AD9A5DD35247AULL, "dae300_cullen6135008.cl"},
	{0x6D1BD266817D4383ULL, "vxht_0_hightown_a3_4n1_0_0.vlm"},
	{0x6D1BDE2181D93297ULL, "lt_0_lowtown_36_0.anb"},
	{0x6D1C12AC27577230ULL, "art\\levels\\ztools\\default\\generic\\0di_basic.msh"},
	{0x6D1C20AC27578A4AULL, "art\\levels\\ztools\\default\\generic\\0di_basic.mmh"},
	{0x6D1C24AC27579081ULL, "art\\levels\\ztools\\default\\generic\\0di_basic.mao"},
	{0x6D1C2D66D41AFD93ULL, "hlac_ht_chantry_2534_46z.dds"},
	{0x6D1C4EB0D08CD4F3ULL, "vxht_keep_6ca_0_0.vlm"},
	{0x6D1CA93B95429C96ULL, "6141491_m.fxe"},
	{0x6D1CE34C2D2AC222ULL, "gen00fl_aveline6106943.cl"},
	{0x6D1CF7ED2CE8CDAAULL, "gen_im_wep_mag_sta_act1_05.uti"},
	{0x6D1D1695A7615701ULL, "vxlt_gallowstemplar_f_5uv_0_0.vlm"},
	{0x6D1D859983DE10D9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\.phy"},
	{0x6D1DBACDCD5302D2ULL, "w_bonepitmining_a2_151_0.phy"},
	{0x6D1DBCCE47814FF0ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medf.mmh"},
	{0x6D1DCACE4781670AULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medf.msh"},
	{0x6D1E31EC58431057ULL, "vxlt_gallowsprison_n_6a7_0_0.vlm"},
	{0x6D1E6B7823DCE4B6ULL, "lt_0_lowtown_docks_n_24_0_ll.mmh"},
	{0x6D1E6E50DD79933CULL, "6042572_m.fxe"},
	{0x6D1E837C89724631ULL, "6160813_m.fxe"},
	{0x6D1F1C6EC5D1EAC1ULL, "character_character_p15.psh.fxo"},
	{0x6D1FA7D7FFB6411CULL, "w_blightlands_240_0.mao"},
	{0x6D1FB3D7FFB655BFULL, "w_blightlands_240_0.mmh"},
	{0x6D1FB9D7FFB66061ULL, "w_blightlands_240_0.msh"},
	{0x6D203C1D0B81E773ULL, "vxlt_0_lowtown_eoo_0_1.vlm"},
	{0x6D20683BA5033FF3ULL, "6144347_m.fxe"},
	{0x6D2071E41D76D233ULL, "gen00fl_varric6099542.cl"},
	{0x6D20A8366793FE30ULL, "6039710_m.fxe"},
	{0x6D20EA240B80B10EULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.impa.gad"},
	{0x6D216FF88E5A3080ULL, "hlac_lt_0_lowtown_a2_n_2_1y.dds"},
	{0x6D2177D2C537ED07ULL, "follower_banter6156066.cl"},
	{0x6D21D9ABC626BDC7ULL, "art\\levels\\races\\design\\_textures\\trap_generic_sa.dds"},
	{0x6D221006080C4D33ULL, "vxht_mansion2b_1vi_0_0.vlm"},
	{0x6D221EC82B7D069FULL, "905949403.ogg"},
	{0x6D22D049E41D5A5CULL, "6092357_m.fxe"},
	{0x6D22EBEF279A8904ULL, "vxlt_0_lowtown_a2_n_f17_0_0.vlm"},
	{0x6D232E918FDA45DCULL, "zz_approval_debug6003056.cl"},
	{0x6D233530A2698949ULL, "613068197.ogg"},
	{0x6D234D555D399256ULL, "vxw_bonepitmining_a2_1zm_0_0.vlm"},
	{0x6D23844759BD88A6ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.cess.ani"},
	{0x6D2417B629F24125ULL, "30743598.ogg"},
	{0x6D242506F153ECACULL, "1067717250.ogg"},
	{0x6D248EDDB8E22AA2ULL, "vxlt_backalley_n_1wk_0_0.vlm"},
	{0x6D24938BD6EAC9CFULL, "vxca_caverns1_mines_i37_0_0.vlm"},
	{0x6D24F9055A5E60F8ULL, "vxlt_0_lowtown_e2f_0_0.vlm"},
	{0x6D251B8944A5F796ULL, "vxdr_primevaltomb_7li_0_0.vlm"},
	{0x6D251B94005CEB3DULL, "dae010cr_laborer.utc"},
	{0x6D252F558FAEE320ULL, "plt_prm_cod_fade_sword.plo"},
	{0x6D25619DA05F8587ULL, "hlac_lt_gallowsprison_n_-1_0x.dds"},
	{0x6D25ABFC1E15BA46ULL, "llac_w_blightlands_-131072_0.dds"},
	{0x6D25C429E4196AB8ULL, "vxlt_undercityserialkiller_1is_0_1.vlm"},
	{0x6D25EBCB1FF0E2E2ULL, "qcr220cr_gray_warden.utc"},
	{0x6D2615497B56412BULL, "zz_isa_debug6133927.cl"},
	{0x6D26333059151A79ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_4craggyrock.msh"},
	{0x6D263D3059152B67ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_4craggyrock.mmh"},
	{0x6D264130591531B4ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_4craggyrock.mao"},
	{0x6D266110EEEC669AULL, "vxlt_undercityencounter_1pw_0_0.vlm"},
	{0x6D2693582365E3A1ULL, "pet_smoke.mao"},
	{0x6D26E5240B8514EDULL, "art\\vfx\\combat\\vfx_r_gn3_paraly_brittle_c.impa.evt"},
	{0x6D26F74A8F9AC002ULL, "6167694_f.fxe"},
	{0x6D27E34E49285E2CULL, "6163219_m.fxe"},
	{0x6D280A81DFB3AEB1ULL, "74268419.ogg"},
	{0x6D29A517B68343B3ULL, "vxlt_0_lowtown_a2_n_ghr_0_0.vlm"},
	{0x6D29C27F6222113FULL, "vxca_cavern1_haunted_fmn_0_0.vlm"},
	{0x6D2AAE17D9937511ULL, "vxlt_gallowscourtyard_513_0_0.vlm"},
	{0x6D2B20F6A59FF724ULL, "gen00fl_sebastian6178371.cl"},
	{0x6D2BB9CB34367534ULL, "hlac_w_sundermount_-393216_0x.dds"},
	{0x6D2BEA950C7D7B47ULL, "llac_lt_gallowstemplar_f_2394_11.dds"},
	{0x6D2BEFA4B1AB51A6ULL, "116334039.ogg"},
	{0x6D2C5681464ED8E5ULL, "hlac_ht_0_hightown_a2_131070_0z.dds"},
	{0x6D2D00E42C3D96AEULL, "follower_banter6163191.cl"},
	{0x6D2DBA16BDCD0850ULL, "vxw_bonepitdestroyed_a3_25g_0_0.vlm"},
	{0x6D2DE62D3B1EBE66ULL, "154519207.ogg"},
	{0x6D2DFEAA00EF5A70ULL, "548120831.ogg"},
	{0x6D2E04A2535A20FFULL, "221053006.ogg"},
	{0x6D2E64757DE1BEADULL, "751564250.ogg"},
	{0x6D2E83BE731345F6ULL, "553636394.ogg"},
	{0x6D2EA038FA66B381ULL, "zz_fol_debug6103797.cl"},
	{0x6D2EFB7F4C55B342ULL, "gen00fl_bethany6031693.cl"},
	{0x6D2F2E7A27FF0228ULL, "64641041.ogg"},
	{0x6D2F40FF16311891ULL, "294427195.ogg"},
	{0x6D2F9676BDBD69DAULL, "llac_lt_hangedman_-1_10.dds"},
	{0x6D30323348B38BFBULL, "w_woundedcoast_53_0.mmh"},
	{0x6D303E3348B39F98ULL, "w_woundedcoast_53_0.mao"},
	{0x6D30503348B3BEE5ULL, "w_woundedcoast_53_0.msh"},
	{0x6D306C4759C7EA17ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.cess.evt"},
	{0x6D30753F9A36EB91ULL, "gen_im_trs_dwv_heavy_helm.uti"},
	{0x6D30A15F593C2C0FULL, "mcr371_hostile_merrill6082155.cl"},
	{0x6D31538A588D1A3FULL, "65224305.ogg"},
	{0x6D315F28E33E3D6CULL, "mcr370_negotiation6080213.cl"},
	{0x6D31BF324F2FAAD6ULL, "vxw_bonepitmining_a2_1wn_0_0.vlm"},
	{0x6D31C3662C1E0955ULL, "follower_banter6164029.cl"},
	{0x6D322A970891F866ULL, "317260351.ogg"},
	{0x6D32E3DB53942823ULL, "534882135.ogg"},
	{0x6D3307911369BEB9ULL, "gen_im_cft_rec_arn_07.uti"},
	{0x6D3330A647863871ULL, "hlac_ca_caverns1_24423_4y.dds"},
	{0x6D334429DC59213AULL, "vxw_bonepitmining_a2_17b_0_0.vlm"},
	{0x6D334F203A86F6BCULL, "274417256.ogg"},
	{0x6D341A4D815F554CULL, "mag200_feynriel6023890.cl"},
	{0x6D3437FCF6ADBCA1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robc_0t.dds"},
	{0x6D3443705EB4CD81ULL, "w_sundermount_27_0.rml"},
	{0x6D3452FC0812F2B5ULL, "one300dg_vanard6063948.cl"},
	{0x6D353CDF683E999CULL, "follower_banter6167232.cl"},
	{0x6D35573862C79625ULL, "180646361.ogg"},
	{0x6D355A231B355459ULL, "vxlt_0_lowtown_a2_n_c_eff_0_0.vlm"},
	{0x6D3593B24A597745ULL, "mrl150_merrill6071945.cl"},
	{0x6D361AD0023E50D3ULL, "606878408.ogg"},
	{0x6D364D9A5B1EBE17ULL, "lgt120_orz_bln_off_messenge6140227.cl"},
	{0x6D364EFCAF5A8BB5ULL, "follower_banter6143233.cl"},
	{0x6D36C0C5E90385DCULL, "391782576.ogg"},
	{0x6D376D4759CE0AA0ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.cess.gad"},
	{0x6D3772856280AE5DULL, "vxlt_gallowstemplar_f_6qe_0_0.vlm"},
	{0x6D37B92443FB7E54ULL, "hightown_fem_noble1_sta.cnv"},
	{0x6D37CF169D22A04DULL, "6121267_m.fxe"},
	{0x6D37F4C7EBA699EBULL, "42834237.ogg"},
	{0x6D38F0AE67878B32ULL, "781873734.ogg"},
	{0x6D39326AB47905F3ULL, "vxht_keep_6ic_0_0.vlm"},
	{0x6D399C3AA3D9F61EULL, "dae000_keep_servant_1_sta6156870.cl"},
	{0x6D39FA4F8B4D8801ULL, "vxlt_gallowstemplar_f_5ct_0_0.vlm"},
	{0x6D3A124771AE09E2ULL, "lt_0_lowtown_n_25_0.rml"},
	{0x6D3A9B2443FD8094ULL, "hightown_fem_noble1_sta.bnk"},
	{0x6D3ACF4AC18FF93DULL, "6125832_m.fxe"},
	{0x6D3AD69CCE85DC5FULL, "251307641.ogg"},
	{0x6D3B5D14E24D8682ULL, "444279732.ogg"},
	{0x6D3B8FDA19B5F8A0ULL, "896914906.ogg"},
	{0x6D3BC7B97186763BULL, "w_encounter3_56_0.rml"},
	{0x6D3D43EF7C631797ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_act_c.anb"},
	{0x6D3D9C3A3A39FADEULL, "935926202.ogg"},
	{0x6D3DADB6FB93193EULL, "qun211cr_elf_warrior6072346.cl"},
	{0x6D3E9B987AA3977DULL, "vxlt_0_lowtown_n_gph_0_0.vlm"},
	{0x6D3ED14E01EA986BULL, "6077284_f.fxe"},
	{0x6D3FABE4493EBEF3ULL, "318481127.ogg"},
	{0x6D401BC6F10FB94FULL, "vxlt_0_lowtown_docks_n_bgk_0_0.vlm"},
	{0x6D40AA881C40EC3DULL, "379049654.ogg"},
	{0x6D4174F0DA9CFFF0ULL, "108075580.ogg"},
	{0x6D41BB64082C27ECULL, "vxlt_0_lowtown_n_byo_0_0.vlm"},
	{0x6D41FF4328922896ULL, "vxdr_primevaltomb_7bk_0_0.vlm"},
	{0x6D424080D05222ECULL, "zz_qun_debug6028298.cl"},
	{0x6D42C405FC8327CAULL, "hm_raider_5.mor"},
	{0x6D4341B1765485B6ULL, "vxlt_hangedman_rs_0_0.vlm"},
	{0x6D43740A385640E0ULL, "vxw_encounter2_n_0_0.vlm"},
	{0x6D447A678BB0FEF4ULL, "gen00fl_aveline6121960.cl"},
	{0x6D455356BE5715E4ULL, "488202035.ogg"},
	{0x6D45CC9B0BD9ECB6ULL, "119003699.ogg"},
	{0x6D45E778E1992590ULL, "11177474.ogg"},
	{0x6D45F534BCDD4672ULL, "hlac_dr_slavers_3207_9y.dds"},
	{0x6D4605418DD27B95ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masc_0t.dds"},
	{0x6D463DBAC96D1FA5ULL, "vxlt_gallowstemplar_f_5yv_0_0.vlm"},
	{0x6D4688D19A6F74B3ULL, "vxlt_0_lowtown_a2_n_gbt_0_0.vlm"},
	{0x6D4798BA9DB86A30ULL, "dae000_arishok6067867.cl"},
	{0x6D47C0908AA41D66ULL, "642627124.ogg"},
	{0x6D47DB2181FE68ABULL, "lt_0_lowtown_36_0.rml"},
	{0x6D480D781C0424E6ULL, "hlac_lt_0_lowtown_n_0_7y.dds"},
	{0x6D48A166931F8ED2ULL, "6174899_m.fxe"},
	{0x6D48BCCA329871F0ULL, "w_encounter2_20_0_lc.mmh"},
	{0x6D48DD241C483B02ULL, "6111609_m.fxe"},
	{0x6D494211B5716A36ULL, "zz_seb_debug6183085.cl"},
	{0x6D498C430F0CF0A5ULL, "465872999.ogg"},
	{0x6D4997D4059B1163ULL, "6085831_m.fxe"},
	{0x6D49CA89506EB225ULL, "169588538.ogg"},
	{0x6D4A2EA6E615EFDCULL, "6082315_f.fxe"},
	{0x6D4A3DE7A754390FULL, "6125311_m.fxe"},
	{0x6D4A5C8B739E852CULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_act_c.impa.ani"},
	{0x6D4A7F39BFF2D6E4ULL, "6218704_m.fxe"},
	{0x6D4A85CE25FF0F0DULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialb.mmh"},
	{0x6D4A89CE25FF165EULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialb.mao"},
	{0x6D4A97CE25FF2D93ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialb.msh"},
	{0x6D4B492437237436ULL, "6091974_m.fxe"},
	{0x6D4B4DF7A94E40E2ULL, "plt_prm000pt_promo_core_a.plo"},
	{0x6D4B77A143F55860ULL, "vxlt_hangedman_n_1xe_0_0.vlm"},
	{0x6D4C70707AC7E0CEULL, "one120_athenril6173941.cl"},
	{0x6D4C7B32988B71DDULL, "6037050_m.fxe"},
	{0x6D4CE5B4E3515D73ULL, "vxht_0_hightown_n_bal_0_0.vlm"},
	{0x6D4CF9E051DB9872ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.cess.gad"},
	{0x6D4D04A2C5CF95A6ULL, "mag110cr_arianni.utc"},
	{0x6D4D12EC37A22416ULL, "1023882336.ogg"},
	{0x6D4DAA3EA17DF26EULL, "6036865_m.fxe"},
	{0x6D4E02484E0D6FF4ULL, "vxw_encounter1_er_0_0.vlm"},
	{0x6D4E15741135C5C1ULL, "138009437.ogg"},
	{0x6D4E8B1E51C2C921ULL, "752264861.ogg"},
	{0x6D4EAA45F76EEC83ULL, "6116708_m.fxe"},
	{0x6D4EAFBBD70AAB9BULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb101a_0d.dds"},
	{0x6D4EB0E5BCFA624BULL, "gen00fl_isabela6096855.cl"},
	{0x6D4F09F5C9CBA3AFULL, "vxlt_0_lowtown_a2_gij_0_0.vlm"},
	{0x6D4FB567700C202BULL, "vxlt_warehouse_n_106_0_0.vlm"},
	{0x6D4FC40E18FAB97FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\swqoverlay.anb"},
	{0x6D4FE4748EC47422ULL, "vxlt_backalley_n_1xv_0_0.vlm"},
	{0x6D50262D5A8E6F40ULL, "vxht_0_hightown_cbh_0_0.vlm"},
	{0x6D50751AB9403A04ULL, "dae211ar_gamlens_house_nt.are"},
	{0x6D51698B73A4BA6EULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_act_c.impa.gad"},
	{0x6D523DDCFF218559ULL, "vxlt_0_lowtown_a2_n_c_epd_0_0.vlm"},
	{0x6D52DF7D7CFFE1D3ULL, "6057072_m.fxe"},
	{0x6D53541D36714AF2ULL, "hlac_lt_0_lowtown_a2_196607_1y.dds"},
	{0x6D53BED6D7C7FA95ULL, "vxht_mansion2_2vh_0_0.vlm"},
	{0x6D53CA362CC42964ULL, "27676431.ogg"},
	{0x6D53FCE051E1BBA9ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.cess.evt"},
	{0x6D540C89E98B340AULL, "6074042_f.fxe"},
	{0x6D5477EE6F7D7CE4ULL, "186311282.ogg"},
	{0x6D5483512EC77B5CULL, "cod_ltr_broker_hunter_1.utp"},
	{0x6D5483929641B106ULL, "combat_ws_nmv_default.gda"},
	{0x6D5516EDAD07ED04ULL, "mag100cr_templar2.utc"},
	{0x6D559C13D6D4B690ULL, "vxlt_0_lowtown_docks_a3_n_d0d_0_0.vlm"},
	{0x6D55A407DDC30AD4ULL, "hlac_dr_slavers_3214_6z.dds"},
	{0x6D55EB1E0591BAABULL, "6080807_m.fxe"},
	{0x6D562918D6CAF8A5ULL, "836668644.ogg"},
	{0x6D5656C804F7F42BULL, "ss_grace.bnk"},
	{0x6D56D4EDFC82C40EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sneaking_peek_lean.ani"},
	{0x6D57A1433A6F0513ULL, "vxht_mansion2b_3cx_0_0.vlm"},
	{0x6D58648B73AAD04DULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_act_c.impa.evt"},
	{0x6D58810210018741ULL, "vxlt_gamlenshouse_n_ir_0_0.vlm"},
	{0x6D58AFC4994A518FULL, "6122211_m.fxe"},
	{0x6D593A3EA631BD4FULL, "pro000_meeran6048850.cl"},
	{0x6D59BB0129198CB6ULL, "6096874_m.fxe"},
	{0x6D5A153D07F28996ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobe_0d.dds"},
	{0x6D5A3DF11833E86BULL, "llac_w_bonepitmining_a2_327685_0.dds"},
	{0x6D5A5AC07144EFF0ULL, "qcr000_arishok6057121.cl"},
	{0x6D5ADBA6D510CC8CULL, "246574655.ogg"},
	{0x6D5B63A7175CB87AULL, "853252164.ogg"},
	{0x6D5BA21BC343A438ULL, "zz_dre_debug6201308.cl"},
	{0x6D5BDD819175ACDBULL, "6121089_m.fxe"},
	{0x6D5C21945C30DACBULL, "619914258.ogg"},
	{0x6D5C742540038874ULL, "vxw_encounter1_js_0_0.vlm"},
	{0x6D5CF0826E90E79CULL, "223311725.ogg"},
	{0x6D5D16764232C3F4ULL, "vxdr_primevaltomb_9du_0_0.vlm"},
	{0x6D5D2EB0C8CE2BA0ULL, "vxlt_blackemporium_ud_0_0.vlm"},
	{0x6D5DB4B5EFB11B0FULL, "dae000_flemeth6000708.cl"},
	{0x6D5EC3CE47B8F20BULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medf.phy"},
	{0x6D5EDA3F34F80D7CULL, "hlac_ht_mansion_n_3365_0y.dds"},
	{0x6D5EF511D58C9FE4ULL, "6020532_m.fxe"},
	{0x6D5F2B2AAEF70D04ULL, "dae300_alistair_king6133790.cl"},
	{0x6D5FE293369ACE4CULL, "269593792.ogg"},
	{0x6D5FE51B9ED64F68ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comb_0n.dds"},
	{0x6D5FF8A3715263A8ULL, "538554657.ogg"},
	{0x6D60443C0B968C0CULL, "ht_0_hightown_22_0.rml"},
	{0x6D60CC2BE627C734ULL, "vxlt_0_lowtown_n_ery_0_0.vlm"},
	{0x6D61856AEF87810EULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\flurry.ani"},
	{0x6D61E4E051EDD018ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.cess.ani"},
	{0x6D62481B1A3EAF36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_small.evt"},
	{0x6D62B3EBAD54003DULL, "vxlt_0_lowtown_docks_n_cun_0_0.vlm"},
	{0x6D639FC93C538C3DULL, "vxht_mansion2b_1kp_0_0.vlm"},
	{0x6D63A9AF5620B744ULL, "364597727.ogg"},
	{0x6D63B8670D9096BDULL, "6132670_m.fxe"},
	{0x6D63CCEDFC8D415FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sneaking_peek_lean.evt"},
	{0x6D63F6E5F3DFB416ULL, "hlac_lt_0_lowtown_a2_n_c_131070_4z.dds"},
	{0x6D640D47697F3504ULL, "143810586.ogg"},
	{0x6D649BBEEDE51330ULL, "vxlt_0_lowtown_a2_n_ex7_0_0.vlm"},
	{0x6D652DEB7311BD65ULL, "1053105835.ogg"},
	{0x6D652F563F4F5B86ULL, "375492608.ogg"},
	{0x6D654F966900C95AULL, "gen_im_arm_boo_med_act1_02.uti"},
	{0x6D655F7802DE36D0ULL, "vxdr_primevaltomb_99f_0_0.vlm"},
	{0x6D6587828CAC0EC4ULL, "831725167.ogg"},
	{0x6D659DE81860590EULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_nud_0n.dds"},
	{0x6D661CDDE9C59198ULL, "vxca_caverns1_cjp_0_0.vlm"},
	{0x6D66263016198560ULL, "6111364_m.fxe"},
	{0x6D6640EF7C861798ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_act_c.mmh"},
	{0x6D66AE85C940F387ULL, "one280_iwan6096905.cl"},
	{0x6D66B136B1BD7B8AULL, "6168112_f.fxe"},
	{0x6D670483CC469E70ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_010a_0t.dds"},
	{0x6D6734EB173F03AAULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_quill.mao"},
	{0x6D6738EB173F0AF9ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_quill.mmh"},
	{0x6D6742EB173F1BE7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_quill.msh"},
	{0x6D6798D251B93ABBULL, "vxw_woundedcoast_103_0_0.vlm"},
	{0x6D685B5B27E18960ULL, "vxlt_hangedman_n_1rc_0_0.vlm"},
	{0x6D68ED4551A48A4AULL, "726567935.ogg"},
	{0x6D691634CB8FEA51ULL, "vxlt_gallowstemplar_n_6s3_0_0.vlm"},
	{0x6D693F1B1A44BEC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_small.gad"},
	{0x6D6A5D26A54CCB9BULL, "mrl251cr_pol.utc"},
	{0x6D6AA365AE3E1F9BULL, "gen00fl_player6207815.cl"},
	{0x6D6AADEDFC932B48ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sneaking_peek_lean.gad"},
	{0x6D6AF6E28AFC9AC1ULL, "vrc321_varric6084291.cl"},
	{0x6D6B03BDD53FE935ULL, "vxw_sundermount_34k_0_0.vlm"},
	{0x6D6B0CA8A2C706D7ULL, "6137297_m.fxe"},
	{0x6D6C888ADFAEFC3EULL, "6109727_m.fxe"},
	{0x6D6CCBF738F06A43ULL, "mag110cr_pre_abomination.utc"},
	{0x6D6D75D118E71931ULL, "zz_one000_barber6179408.cl"},
	{0x6D6E6737FD08A442ULL, "vxlt_0_lowtown_docks_a3_n_d4j_0_0.vlm"},
	{0x6D6E7D6AEF91FE5FULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\flurry.evt"},
	{0x6D6EE5B9092475F8ULL, "and111_lirene6174752.cl"},
	{0x6D6EE69767C3B63FULL, "hlac_ht_mansion2_818_67z.dds"},
	{0x6D6F2858284842E3ULL, "6048821_m.fxe"},
	{0x6D6F2F3715FF02BCULL, "one110_pryce6088871.cl"},
	{0x6D6F334EFDF7D0A9ULL, "vxht_mansion2b_2um_0_0.vlm"},
	{0x6D6F94B3CC579F03ULL, "205071981.ogg"},
	{0x6D6FC387A6CCCB11ULL, "art\\vfx\\materials\\vfx_flare01_blend.mao"},
	{0x6D6FFFBABDA2134CULL, "943806581.ogg"},
	{0x6D70401B1A4ADF6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_small.ani"},
	{0x6D70426E2CC127D6ULL, "205688630.ogg"},
	{0x6D705424EDC42742ULL, "831589792.ogg"},
	{0x6D7055D45E90F0E7ULL, "6070013_m.fxe"},
	{0x6D7069D1810C169BULL, "vxht_keep_4hn_0_0.vlm"},
	{0x6D707D33C856893FULL, "vxlt_0_lowtown_a2_n_c_e1q_0_0.vlm"},
	{0x6D7098A0C87964C7ULL, "6119499_m.fxe"},
	{0x6D71CA1F144CB9A8ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_swordb_1.mao"},
	{0x6D7225664CFB20B7ULL, "vxht_mansion_n_4y2_0_0.vlm"},
	{0x6D7241B784869A0CULL, "vxlt_0_lowtown_a2_n_c_ghw_0_0.vlm"},
	{0x6D725847C3BB5833ULL, "603697831.ogg"},
	{0x6D7272FFB4D66D17ULL, "vxlt_backalley_g_1ko_0_0.vlm"},
	{0x6D72A9D662E02AF9ULL, "door_wispy.mao"},
	{0x6D735EE823E6676BULL, "6096624_m.fxe"},
	{0x6D739D6E2F80D943ULL, "153821388.ogg"},
	{0x6D740F4771DF7CDEULL, "lt_0_lowtown_n_25_0.anb"},
	{0x6D7435705EEA9A70ULL, "w_sundermount_27_0.msh"},
	{0x6D7443705EEAB18AULL, "w_sundermount_27_0.mmh"},
	{0x6D7447705EEAB8C1ULL, "w_sundermount_27_0.mao"},
	{0x6D7483077780B99EULL, "vxw_bonepitmining_a2_5g_0_0.vlm"},
	{0x6D748F13C379A958ULL, "6099024_m.fxe"},
	{0x6D74924BFF617E40ULL, "954079204.ogg"},
	{0x6D74F02E93C7BEC9ULL, "6055615_f.fxe"},
	{0x6D7535D043D8A90AULL, "vxlt_darktown_5kx_0_0.vlm"},
	{0x6D753B6A7EBB9787ULL, "isa321_velasco6091879.cl"},
	{0x6D755E6AEF97E848ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\flurry.gad"},
	{0x6D75B210DECB60F2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grill_d.dds"},
	{0x6D75CD0E5A9D720DULL, "vxdr_primevalentrance_351_0_0.vlm"},
	{0x6D7651874A0A9022ULL, "one100_party_comments6133374.cl"},
	{0x6D76741AB961079FULL, "dae211ar_gamlens_house_nt.lst"},
	{0x6D767B5F5AD029A9ULL, "vxht_bartrand_n_655_0_0.vlm"},
	{0x6D76828DB6BCDAEEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_01tent.mmh"},
	{0x6D768E8DB6BCEE85ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_01tent.mao"},
	{0x6D769C8DB6BD06DCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_01tent.msh"},
	{0x6D76D4319BE82A86ULL, "art\\vfx\\combat\\vfx_blank_c.dura.evt"},
	{0x6D770665B4258987ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.cess.gad"},
	{0x6D77085D14D94558ULL, "dae000_tomwise6167298.cl"},
	{0x6D77EC91E5D4D9D3ULL, "6062813_f.fxe"},
	{0x6D782E856B0DFB21ULL, "vxw_bonepitmining_a2_8s_0_0.vlm"},
	{0x6D78630C427BF97DULL, "vxlt_0_lowtown_n_gdd_0_0.vlm"},
	{0x6D78B00FC9024F55ULL, "art\\characters\\creatures\\hurlock\\w_stf_001a.mao"},
	{0x6D7924E316DF7168ULL, "w_blightlands_139_0.mmh"},
	{0x6D7930E316DF850BULL, "w_blightlands_139_0.mao"},
	{0x6D7942E316DFA392ULL, "w_blightlands_139_0.msh"},
	{0x6D79759AC73601B5ULL, "vxw_sundermount_33j_0_0.vlm"},
	{0x6D7A1809935920B9ULL, "hlac_ca_caverns1_24426_9y.dds"},
	{0x6D7A5F8C9FADED2FULL, "vxlt_0_lowtown_a2_gfw_0_0.vlm"},
	{0x6D7A6051C6F6629FULL, "clf260_harley6088781.cl"},
	{0x6D7A9341D540279EULL, "6162269_m.fxe"},
	{0x6D7A9C0047FA4FC2ULL, "zz_dre_debug6054658.cl"},
	{0x6D7AD29CC4967838ULL, "static3.vsh.fxo"},
	{0x6D7B4F74513921C0ULL, "6080530_m.fxe"},
	{0x6D7BD2AE2A2D0DF9ULL, "814748436.ogg"},
	{0x6D7BD4232A74AD22ULL, "6106923_f.fxe"},
	{0x6D7C621B6D36AED3ULL, "516326824.ogg"},
	{0x6D7D68EF457A0B96ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3.phy"},
	{0x6D7DB5466CA17A30ULL, "ser_effect_aoe_corpse.ncs"},
	{0x6D7DEB319BEE70D1ULL, "art\\vfx\\combat\\vfx_blank_c.dura.gad"},
	{0x6D7DFB65B42B9594ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.cess.evt"},
	{0x6D7E323348F5AC88ULL, "w_woundedcoast_53_0.rml"},
	{0x6D7E76E7B3F2EFC0ULL, "mer120_hubert6077284.cl"},
	{0x6D7E887DC28FF3ADULL, "gen00fl_isabela6099324.cl"},
	{0x6D7F146DADAA4415ULL, "vxht_mansion2_2yu_0_0.vlm"},
	{0x6D7FAF8B1BF41DC7ULL, "589658109.ogg"},
	{0x6D8087F0796116E2ULL, "vxw_woundedcoast_17k_0_0.vlm"},
	{0x6D80921ABC28B93BULL, "6166682_m.fxe"},
	{0x6D815459A4ACCBACULL, "6132473_m.fxe"},
	{0x6D817AF30FFC92B8ULL, "958823265.ogg"},
	{0x6D81EC6185B475E8ULL, "plt_cod_shw_mer_notes.nss"},
	{0x6D821A45CD81E3DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\ht_gallowsvista_sa.dds"},
	{0x6D821B73E6B9D9E3ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.cess.evt"},
	{0x6D82684BF41085ABULL, "zz_qcr_debug6058841.cl"},
	{0x6D82700252C2655FULL, "631424784.ogg"},
	{0x6D829CF319034726ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_wng_0n.dds"},
	{0x6D82BBB66786628FULL, "dae000_porfiria.bnk"},
	{0x6D8321823D3CF32CULL, "hlac_ca_caverns2prc_5802_64x.dds"},
	{0x6D8333313A9C3A93ULL, "337611345.ogg"},
	{0x6D8354253C27CF62ULL, "hlac_lt_0_lowtown_a2_131071_2z.dds"},
	{0x6D835696064D05A6ULL, "621579093.ogg"},
	{0x6D83A7AD35CEC18AULL, "vxlt_darktown_5ly_0_0.vlm"},
	{0x6D83AE9332559BCFULL, "338332153.ogg"},
	{0x6D83B7050B29228CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\mother\\mother_kneeling_loop.ani"},
	{0x6D83EC319BF2DE7FULL, "art\\vfx\\combat\\vfx_blank_c.dura.ani"},
	{0x6D841E6829F4B761ULL, "llac_lt_0_lowtown_docks_n_-131076_1.dds"},
	{0x6D843CC6DF41DBDEULL, "399222382.ogg"},
	{0x6D843D0F2C8431BEULL, "vxlt_undercitydungeon_11q_0_0.vlm"},
	{0x6D843D37397935C3ULL, "662446335.ogg"},
	{0x6D846A3C9E1B44D8ULL, "6047430_m.fxe"},
	{0x6D8527704BFCEDE2ULL, "follower_banter6155550.cl"},
	{0x6D853D2496F734E5ULL, "vxht_mansion2_2v8_0_0.vlm"},
	{0x6D853F150BCDBA60ULL, "vxlt_hangedman_n_1da_0_0.vlm"},
	{0x6D85476B188F0B46ULL, "6154417_m.fxe"},
	{0x6D85E6DCEDD2B71DULL, "vxlt_gallowstemplar_3rj_0_0.vlm"},
	{0x6D85F8DCBF6EABD1ULL, "bla_cl.mao"},
	{0x6D863500F4BF64E2ULL, "845035216.ogg"},
	{0x6D86DDB6678A85F7ULL, "dae000_porfiria.cnv"},
	{0x6D8767EF45825EA9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3.mmh"},
	{0x6D878132F2FB8726ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robb_0d.dds"},
	{0x6D87B608A994FDA8ULL, "hlac_w_sundermount_196610_0x.dds"},
	{0x6D87BCCC94D249E6ULL, "103580119.ogg"},
	{0x6D87F56185B8F0FBULL, "plt_cod_shw_mer_notes.plo"},
	{0x6D883ECFD2FE77EFULL, "art\\characters\\weapons\\axes\\w_axe_dsaxa.mao"},
	{0x6D88F473E6BFB664ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.cess.gad"},
	{0x6D88F77793511A31ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robc_0d.dds"},
	{0x6D89051EC736401EULL, "6167169_m.fxe"},
	{0x6D892F800AD9CC6CULL, "llac_ca_caverns2prc_5802_6.dds"},
	{0x6D8A19EF21A307C2ULL, "1007960351.ogg"},
	{0x6D8A2B2C80A70C05ULL, "vxlt_0_lowtown_ha3_0_0.vlm"},
	{0x6D8A37B081A7E4F6ULL, "6036326_m.fxe"},
	{0x6D8A4B3806DF8AACULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolem_s2.anb"},
	{0x6D8A6B9A5603C0A7ULL, "vxlt_hangedman_15s_0_1.vlm"},
	{0x6D8AC4050B2F57CEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\mother\\mother_kneeling_loop.gad"},
	{0x6D8ADE1EE1317F8BULL, "900303666.ogg"},
	{0x6D8B1B19B66D5209ULL, "vxdr_0_deeproads_f0u_0_0.vlm"},
	{0x6D8B21F28BEABE0FULL, "vxlt_gallowstemplar_55s_0_0.vlm"},
	{0x6D8B5FC8C03081D1ULL, "vxht_keep_3ba_0_0.vlm"},
	{0x6D8BF365B437C5CDULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.cess.ani"},
	{0x6D8C7275E7018AC2ULL, "aip_templarseeker_base.gda"},
	{0x6D8C762725C1736EULL, "gen00fl_sebastian6188622.cl"},
	{0x6D8D20BD05B556E8ULL, "vxlt_gallowscourtyard_6w7_0_0.vlm"},
	{0x6D8D46C0A86862EDULL, "qun110_varnell6047726.cl"},
	{0x6D8DE9E12AA3230FULL, "vxht_brothel_n_5j0_0_1.vlm"},
	{0x6D8E02DD755CBFE1ULL, "6102138_m.fxe"},
	{0x6D8E313DDC5C6F7EULL, "art\\vfx\\combat\\vfx_w_off_stagger_activ_c.mmh"},
	{0x6D8E644A79CC6F7DULL, "6130129_f.fxe"},
	{0x6D8EB7C267B2DE63ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.dura.ani"},
	{0x6D8ECC0EA1CD928AULL, "279873820.ogg"},
	{0x6D8EF9CD6B572F62ULL, "vxw_woundedcoast_12t_0_0.vlm"},
	{0x6D8F0FE5A35407FFULL, "lt_0_lowtown_a2_n_c_27_0_lc.mmh"},
	{0x6D8F3EE6DFAAD0DFULL, "w_blightlands_39_0.msh"},
	{0x6D8F44E6DFAADA81ULL, "w_blightlands_39_0.mmh"},
	{0x6D8F50E6DFAAEF22ULL, "w_blightlands_39_0.mao"},
	{0x6D8F5C759E74C69BULL, "and100_anders6206373.cl"},
	{0x6D8FBD44555AD818ULL, "vxht_brothel_n_61d_0_0.vlm"},
	{0x6D900373E6C5EE42ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.cess.ani"},
	{0x6D90297B92BB3D37ULL, "vxht_0_hightown_bxh_0_0.vlm"},
	{0x6D90715A10F49D45ULL, "6134588_m.fxe"},
	{0x6D908899C66EBBDAULL, "vxlt_0_lowtown_docks_n_97m_0_0.vlm"},
	{0x6D90902658CE8E35ULL, "1019029577.ogg"},
	{0x6D90F168DED9EEFCULL, "vxdr_primevalentrance_30n_0_1.vlm"},
	{0x6D910D5BAC4FBA62ULL, "seb321_leliana6180932.cl"},
	{0x6D9127388553DDBEULL, "gen00fl_isabela6135625.cl"},
	{0x6D9162DE82D788A0ULL, "atl_prestartme_dxt5_02.dds"},
	{0x6D9179E7613C6891ULL, "vxlt_0_lowtown_a2_e12_0_0.vlm"},
	{0x6D918C5044009547ULL, "vxht_0_hightown_a3_9zp_0_0.vlm"},
	{0x6D91BE35AD07CBEBULL, "6136373_m.fxe"},
	{0x6D91BF050B356DADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\mother\\mother_kneeling_loop.evt"},
	{0x6D92198A27C4DA0AULL, "vxlt_darktown_5qz_0_0.vlm"},
	{0x6D9282502A134018ULL, "vxw_encounter2_gq_0_0.vlm"},
	{0x6D9285A83FBEC9C9ULL, "6173417_m.fxe"},
	{0x6D9287534FD7C7D0ULL, "vxw_encounter2_camp_62_0_0.vlm"},
	{0x6D930A4587399321ULL, "6090435_m.fxe"},
	{0x6D9321751EA16584ULL, "follower_banter6161205.cl"},
	{0x6D9338283430AF69ULL, "vxlt_0_lowtown_docks_a3_n_d2k_0_0.vlm"},
	{0x6D950B3A55866B65ULL, "294731513.ogg"},
	{0x6D9588FFD6F62093ULL, "art\\levels\\ztools\\default\\generic\\2wl_basic.phy"},
	{0x6D95ADBEE37960D6ULL, "vxdr_0_deeproads_bfd_0_0.vlm"},
	{0x6D95D90500F8B671ULL, "vxht_chantry_30w_0_0.vlm"},
	{0x6D965EEC71C0118CULL, "mrl150_merrill6199330.cl"},
	{0x6D968A71C0ACCC3DULL, "vxlt_0_lowtown_a2_n_gzb_0_0.vlm"},
	{0x6D96E4CD3A70E16CULL, "6154006_m.fxe"},
	{0x6D970744D8445F6DULL, "lt_0_lowtown_docks_a3_36_0.rml"},
	{0x6D97581A61727276ULL, "269765803.ogg"},
	{0x6D97AA3D45959B86ULL, "47445066.ogg"},
	{0x6D97D2FC8690B5C9ULL, "follower_banter6173371.cl"},
	{0x6D97DC34B4F99B76ULL, "gen00fl_varric6033655.cl"},
	{0x6D97DF55B500C535ULL, "6133145_m.fxe"},
	{0x6D97FB6120922846ULL, "vxlt_0_lowtown_docks_a3_cws_0_0.vlm"},
	{0x6D9814EDD4B844A8ULL, "dae021_sister_lorena_sta6142660.cl"},
	{0x6D98881507381B6CULL, "qun231_party_comment6077012.cl"},
	{0x6D98CF21963E5B30ULL, "vxlt_0_lowtown_docks_n_ber_0_0.vlm"},
	{0x6D98D7FD723884B7ULL, "vxlt_gallowstemplar_5qg_0_0.vlm"},
	{0x6D996A1331024168ULL, "llac_ca_caverns2_5802_14.dds"},
	{0x6D9970893515E3AAULL, "48890089.ogg"},
	{0x6D9A3D92CDD1CE68ULL, "fen110_fenris6096194.cl"},
	{0x6D9A81C6B4B36605ULL, "dae200_cullen6140321.cl"},
	{0x6D9B36BA5981A8B1ULL, "6177387_m.fxe"},
	{0x6D9B52DAA4B3240CULL, "hlac_w_woundedcoast_131080_0x.dds"},
	{0x6D9BCFC267BD915AULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.dura.evt"},
	{0x6D9C103DDC68743DULL, "art\\vfx\\combat\\vfx_w_off_stagger_activ_c.anb"},
	{0x6D9C12256369BEEEULL, "w_woundedcoast_10_0.msh"},
	{0x6D9C20256369D63FULL, "w_woundedcoast_10_0.mao"},
	{0x6D9C2C256369EADCULL, "w_woundedcoast_10_0.mmh"},
	{0x6D9CC443A0D16024ULL, "322957718.ogg"},
	{0x6D9E53B8AB42E2C0ULL, "mcr370_templar26087007.cl"},
	{0x6D9E6BAFDE5E1080ULL, "rdr111dg_martin6054188.cl"},
	{0x6D9E6CDAE4B64AE7ULL, "vxlt_gallowstemplar_f_352_0_0.vlm"},
	{0x6D9EA4359116310CULL, "639883638.ogg"},
	{0x6D9F1705AA334552ULL, "328699270.ogg"},
	{0x6D9FA841A669E554ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownenter.ani"},
	{0x6D9FE5F682A989E8ULL, "vxht_mansion2_3df_0_0.vlm"},
	{0x6DA00509E705855AULL, "hlac_ht_0_hightown_196607_0z.dds"},
	{0x6DA0117E3CF7F16EULL, "398531580.ogg"},
	{0x6DA030129EFDE46BULL, "6131273_m.fxe"},
	{0x6DA033D957A0C6C4ULL, "vxlt_undercityencounter_1lm_0_0.vlm"},
	{0x6DA055081A652978ULL, "mag320_mage_conspirator.cnv"},
	{0x6DA0FC2C8A9A564AULL, "tent_d.dds"},
	{0x6DA18ABCDEF8E491ULL, "vxw_encounter3_e5_0_0.vlm"},
	{0x6DA1C6D111D27D10ULL, "786916197.ogg"},
	{0x6DA1FCA660603E59ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgta_0s.dds"},
	{0x6DA2288338FB12F4ULL, "596658803.ogg"},
	{0x6DA26DAFB345332EULL, "llac_w_sundermount_327681_0.dds"},
	{0x6DA2CEC267C3AE4DULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.dura.gad"},
	{0x6DA2D8D1BD103FF3ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.dura.ani"},
	{0x6DA337D406639103ULL, "gen00fl_varric6033595.cl"},
	{0x6DA34BB402887389ULL, "art\\vfx\\combat\\vfx_m_bld_wound_act_c.impa.ani"},
	{0x6DA3625281E7894DULL, "one300_murder_followers6086734.cl"},
	{0x6DA36689C43984D1ULL, "qun110tr_petrice_mugger.utt"},
	{0x6DA375D0A6638C96ULL, "zz_vrc_debug6064257.cl"},
	{0x6DA3C38B9145E508ULL, "684760430.ogg"},
	{0x6DA423DD51E77389ULL, "hla_ht_0_hightown_n_0_0x.dds"},
	{0x6DA42FAB301A8FFFULL, "art\\characters\\creatures\\highdragon\\cn_hed_fdr.msh"},
	{0x6DA435AB301A99A1ULL, "art\\characters\\creatures\\highdragon\\cn_hed_fdr.mmh"},
	{0x6DA437081A68DE80ULL, "mag320_mage_conspirator.bnk"},
	{0x6DA441AB301AAE42ULL, "art\\characters\\creatures\\highdragon\\cn_hed_fdr.mao"},
	{0x6DA48843E0C3360DULL, "w_bonepitmining_a2_173_0.mmh"},
	{0x6DA48C43E0C33D5EULL, "w_bonepitmining_a2_173_0.mao"},
	{0x6DA49A43E0C35493ULL, "w_bonepitmining_a2_173_0.msh"},
	{0x6DA4C7B971DF5F08ULL, "w_encounter3_56_0.mmh"},
	{0x6DA4D3B971DF73ABULL, "w_encounter3_56_0.mao"},
	{0x6DA4E5B971DF9232ULL, "w_encounter3_56_0.msh"},
	{0x6DA4FC4EB2A2E4BDULL, "vxlt_0_lowtown_a2_n_gwa_0_0.vlm"},
	{0x6DA4FDD4E5D1E26CULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robd.phy"},
	{0x6DA53AB6361EB6B0ULL, "follower_banter6170427.cl"},
	{0x6DA58AC5754C2FCCULL, "993872149.ogg"},
	{0x6DA5CE4747741236ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_ogre_victim_sis.ani"},
	{0x6DA5ECBA94917485ULL, "6000730_f.fxe"},
	{0x6DA6A541A66FFEB6ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownenter.gad"},
	{0x6DA6D63F7D49F80AULL, "384599866.ogg"},
	{0x6DA6DB8D8928CA44ULL, "852049406.ogg"},
	{0x6DA6E05740BBDAECULL, "mag311_keran6041582.cl"},
	{0x6DA6FFD917D223DAULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchhang_1.mao"},
	{0x6DA740FE883473B0ULL, "vxlt_0_lowtown_docks_n_bbs_0_0.vlm"},
	{0x6DA79D93E52288A2ULL, "hlac_dr_primevalentrance_1725_3x.dds"},
	{0x6DA7B129D0D421F5ULL, "lgt210pt_qun_blkpwdr.ncs"},
	{0x6DA7B54DFE006CD3ULL, "537435084.ogg"},
	{0x6DA8359CECE9C831ULL, "dae000_junar.bnk"},
	{0x6DA8F5D2A89A531BULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml102a_0d.dds"},
	{0x6DA8FAC2C5D032A9ULL, "vxht_mansion2b_2aq_0_0.vlm"},
	{0x6DA95DAF6F6792DCULL, "619054535.ogg"},
	{0x6DA977715075B35DULL, "ser221_quentin6123360.cl"},
	{0x6DA984CE68ED6711ULL, "vxlt_gallowstemplar_n_479_0_0.vlm"},
	{0x6DAA66B4028EC00BULL, "art\\vfx\\combat\\vfx_m_bld_wound_act_c.impa.gad"},
	{0x6DAA83A10AAF53CDULL, "80343623.ogg"},
	{0x6DAB1BF893A41E80ULL, "6054298_m.fxe"},
	{0x6DAB7BBAAAE013B5ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_act_c.anb"},
	{0x6DAC179CECED7E41ULL, "dae000_junar.cnv"},
	{0x6DAC1C3A45DFC784ULL, "6048725_m.fxe"},
	{0x6DAC7BF5885EDBEAULL, "vxht_bartrand_n_67p_0_0.vlm"},
	{0x6DAC8C19EDD33028ULL, "57333922.ogg"},
	{0x6DACC041A6749815ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownenter.evt"},
	{0x6DACD4D2693EEACFULL, "550406461.ogg"},
	{0x6DAD27E49B3748C8ULL, "6125349_m.fxe"},
	{0x6DAD405EA1B34CE6ULL, "vxht_mansion2b_21b_0_0.vlm"},
	{0x6DADB04D10531BC4ULL, "6943084.ogg"},
	{0x6DADD434AEC6D234ULL, "llac_w_sundermount_-393214_0.dds"},
	{0x6DAE02ADFCBBBAAAULL, "192744670.ogg"},
	{0x6DAE6714ABCE82D7ULL, "vxlt_shop_cz_0_0.vlm"},
	{0x6DAEDBEAB281A298ULL, "vxht_0_hightown_a2_as5_0_0.vlm"},
	{0x6DAEFBC6515343A2ULL, "242016917.ogg"},
	{0x6DAF6C05509BE41CULL, "vxdr_0_deeproads_by7_0_0.vlm"},
	{0x6DAFCBA733334AE4ULL, "mag110_arianni6036795.cl"},
	{0x6DAFD0D1BD1ABC2AULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.dura.evt"},
	{0x6DAFDD0D9B50FBDEULL, "278105258.ogg"},
	{0x6DB03C44C68FDF08ULL, "6097831_m.fxe"},
	{0x6DB065B07B3A949FULL, "gen_im_npc_ammo_quiver_qunb.uti"},
	{0x6DB08E1DC51A9D05ULL, "6202166_m.fxe"},
	{0x6DB0A9E931444767ULL, "vxht_0_hightown_n_9zj_0_0.vlm"},
	{0x6DB0DD64AD91487CULL, "597884342.ogg"},
	{0x6DB0E8AE85F002C7ULL, "hlac_lt_0_lowtown_n_-2_0z.dds"},
	{0x6DB1432F72520A40ULL, "vxlt_0_lowtown_n_eu8_0_0.vlm"},
	{0x6DB146F7A9A4E7A9ULL, "plt_prm000pt_promo_core_a.nss"},
	{0x6DB14A7853FB36BBULL, "w_bonepit_191_0.mmh"},
	{0x6DB1567853FB4B58ULL, "w_bonepit_191_0.mao"},
	{0x6DB163B40294DA70ULL, "art\\vfx\\combat\\vfx_m_bld_wound_act_c.impa.evt"},
	{0x6DB1687853FB69A5ULL, "w_bonepit_191_0.msh"},
	{0x6DB1B0C5A949A9A9ULL, "fxe_lava_stones.mao"},
	{0x6DB1ED460FCB74CDULL, "w_encounter3_192_0.mmh"},
	{0x6DB1F1460FCB7B1EULL, "w_encounter3_192_0.mao"},
	{0x6DB1FF460FCB9353ULL, "w_encounter3_192_0.msh"},
	{0x6DB25D21D1FC19A3ULL, "dre000_dougal6131170.cl"},
	{0x6DB3CA35D6F442D5ULL, "vxlt_0_lowtown_fgw_0_0.vlm"},
	{0x6DB424E6DFC9C0AAULL, "w_blightlands_39_0.rml"},
	{0x6DB4981D2453831FULL, "static_depth_debug.vsh.fxo"},
	{0x6DB4C30F98ECF635ULL, "6139745_m.fxe"},
	{0x6DB58143E0D19DCAULL, "w_bonepitmining_a2_173_0.phy"},
	{0x6DB5BE85008AFD2DULL, "qcr000_cut2d_thechampion6025136.cl"},
	{0x6DB64217502A9675ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs105a_0n.dds"},
	{0x6DB652E17D4E43E9ULL, "6054974_m.fxe"},
	{0x6DB6AFD1BD20A29DULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.dura.gad"},
	{0x6DB7E18BB2AB21DAULL, "vxht_0_hightown_c8p_0_0.vlm"},
	{0x6DB85AC45A9838D2ULL, "follower_barks6153915.cl"},
	{0x6DB95229BF61D07CULL, "6070379_m.fxe"},
	{0x6DB9E2E4B209B045ULL, "gen_im_trs_dem_sta_spi_r1.uti"},
	{0x6DB9F74747850170ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_ogre_victim_sis.gad"},
	{0x6DBA013CDCDEDCC3ULL, "vxlt_hangedman_n_2o0_0_0.vlm"},
	{0x6DBA2C256382E82FULL, "w_woundedcoast_10_0.rml"},
	{0x6DBA2EB494A6B886ULL, "928832475.ogg"},
	{0x6DBA8E1737D5E004ULL, "6043625_m.fxe"},
	{0x6DBAC5B54516D423ULL, "gen00fl_varric6088845.cl"},
	{0x6DBB0281CC0FDD5DULL, "665121819.ogg"},
	{0x6DBB83FE68E98EC0ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_010a_0d.dds"},
	{0x6DBB8645772DD1E0ULL, "6041494_m.fxe"},
	{0x6DBC0BFED0C6BC21ULL, "6129559_m.fxe"},
	{0x6DBC2EA09C044066ULL, "226728872.ogg"},
	{0x6DBC6C8E193B5130ULL, "6095734_m.fxe"},
	{0x6DBCBECD828D5975ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupa_n.dds"},
	{0x6DBCD8F19DC49B57ULL, "vxlt_shop_fu_0_0.vlm"},
	{0x6DBD1E355FA55BD0ULL, "follower_banter6172256.cl"},
	{0x6DBD4DC7A477BB18ULL, "vxht_0_hightown_a2_an2_0_0.vlm"},
	{0x6DBE28F71F1D736FULL, "6122000_m.fxe"},
	{0x6DBE50CE4DBABA3FULL, "disarm.cur"},
	{0x6DBECC3B3BDB1427ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_1_ian.gad"},
	{0x6DBEFFABBC0F4469ULL, "6017904_m.fxe"},
	{0x6DBF048D7545FC22ULL, "6119582_m.fxe"},
	{0x6DC02C1E49617876ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_wipe_eyes.evt"},
	{0x6DC11ECE74DE8C20ULL, "671670209.ogg"},
	{0x6DC15FD210359A06ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rocks_d.dds"},
	{0x6DC1705E4A252D77ULL, "6060647_m.fxe"},
	{0x6DC18D91455220CFULL, "mag101_grace6035861.cl"},
	{0x6DC1F1259020C1EBULL, "6164950_m.fxe"},
	{0x6DC22160383C4576ULL, "694206825.ogg"},
	{0x6DC257A0E73EEF95ULL, "vxlt_0_lowtown_docks_b64_0_1.vlm"},
	{0x6DC2718EB0DA9769ULL, "24651834.ogg"},
	{0x6DC2899AF4B6661FULL, "gen00fl_isabela6079089.cl"},
	{0x6DC2CF3CBB811674ULL, "120066469.ogg"},
	{0x6DC2DB9EC80D7D77ULL, "vxca_caverns1_fmw_0_0.vlm"},
	{0x6DC2F334439BCCF0ULL, "vxlt_0_lowtown_docks_bvd_0_0.vlm"},
	{0x6DC38C5E996307E7ULL, "662607374.ogg"},
	{0x6DC3E73F2A42F2FEULL, "6116824_m.fxe"},
	{0x6DC4272CCBBDB471ULL, "77917288.ogg"},
	{0x6DC4437FF582FBECULL, "6086789_m.fxe"},
	{0x6DC4AA5BACB0446CULL, "812867598.ogg"},
	{0x6DC4D03754E936D5ULL, "vxlt_gallowstemplar_n_3r5_0_0.vlm"},
	{0x6DC4EC5FC4003172ULL, "gen00fl_dog6116496.cl"},
	{0x6DC5164742570CAFULL, "616026080.ogg"},
	{0x6DC524DC6BA062A4ULL, "853147626.ogg"},
	{0x6DC54AEB2A1A526AULL, "gen00fl_bethany6101020.cl"},
	{0x6DC54BF921DAF55EULL, "6148470_m.fxe"},
	{0x6DC5A0D2E8A45781ULL, "vxht_bartrand_n_3vm_0_0.vlm"},
	{0x6DC5C436CCA9B126ULL, "art\\vfx\\textures\\rockarmor\\vfx_rockarmor_0s.dds"},
	{0x6DC5D94BBD230FC1ULL, "6060111_m.fxe"},
	{0x6DC5F4936C21918BULL, "vxlt_0_lowtown_docks_a3_n_czb_0_0.vlm"},
	{0x6DC6186CC90868CDULL, "vxlt_0_lowtown_docks_c46_0_0.vlm"},
	{0x6DC6231E4965D401ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_wipe_eyes.gad"},
	{0x6DC687C84E989D44ULL, "vxw_bonepitdestroyed_a3_ky_0_0.vlm"},
	{0x6DC6B86827560AF6ULL, "21277869.ogg"},
	{0x6DC6BA7C1F928C60ULL, "vxca_cavern1_haunted_ebw_0_0.vlm"},
	{0x6DC6FD4AD49165BEULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_2z.dds"},
	{0x6DC70CDF32B9FAF2ULL, "vxht_mansion_n_4x8_0_0.vlm"},
	{0x6DC7C6544589436CULL, "vxlt_0_lowtown_a2_n_c_enw_0_0.vlm"},
	{0x6DC8B8DD1327B3B8ULL, "zz_and_debug6178547.cl"},
	{0x6DC9242AB7E0DAC7ULL, "1068079682.ogg"},
	{0x6DC92EB18294726FULL, "vxht_0_hightown_9up_0_0.vlm"},
	{0x6DC989797A771982ULL, "hlac_ca_caverns1_mines_24424_2z.dds"},
	{0x6DC9BA5068631122ULL, "vxlt_0_lowtown_docks_a3_bsr_0_0.vlm"},
	{0x6DC9BFC5A95E5557ULL, "fxe_lava_stones.dds"},
	{0x6DC9CD7FF62425B6ULL, "mcr370_negotiation6080181.cl"},
	{0x6DC9E5994941D783ULL, "dae000_arishok6078475.cl"},
	{0x6DCA2444F435528AULL, "vxw_woundedcoast_160_0_0.vlm"},
	{0x6DCA24EC6328A39BULL, "vxdr_primevalentrance_2q7_0_1.vlm"},
	{0x6DCA6E0C5AE1D6F3ULL, "vxht_mansion_n_60w_0_0.vlm"},
	{0x6DCA881B9AF8F8B8ULL, "lt_warehouse_n_5_0.rml"},
	{0x6DCABCBD7DAC4481ULL, "static1_vbl.psh.fxo"},
	{0x6DCAC97FA1568A46ULL, "6108614_m.fxe"},
	{0x6DCB13B0DFA23ABFULL, "6101124_m.fxe"},
	{0x6DCB509D216B3247ULL, "810202949.ogg"},
	{0x6DCB7D242582B4BEULL, "6085258_m.fxe"},
	{0x6DCB9224AACCC1BAULL, "lowtown_male_lowlife1_sta6164945.cl"},
	{0x6DCBF145AFDA4EB4ULL, "6099891_m.fxe"},
	{0x6DCC577C634050E4ULL, "6037113_f.fxe"},
	{0x6DCC6208442B83ACULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_maso_0s.dds"},
	{0x6DCC84A3925B4D61ULL, "hlac_dr_0_deeproads_5842_10y.dds"},
	{0x6DCCE17B46EFE33BULL, "lt_0_lowtown_docks_a3_n_121_0.phy"},
	{0x6DCD241E496BF4AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_wipe_eyes.ani"},
	{0x6DCD4E5D2F05BED2ULL, "hlac_lt_0_lowtown_a2_n_c_-196607_0x.dds"},
	{0x6DCDCC3B19BB2E72ULL, "gen00fl_aveline6114085.cl"},
	{0x6DCDDE0D91ABD3BCULL, "373994357.ogg"},
	{0x6DCE3C73B4CCBCA0ULL, "gen_im_cns_arn_torpor.uti"},
	{0x6DCE704F8E123123ULL, "526012892.ogg"},
	{0x6DCEBA4CFDAD1CE4ULL, "llac_w_encounter3_-65531_0.dds"},
	{0x6DCEFED326A5E620ULL, "lt_0_lowtown_docks_a3_121_0.phy"},
	{0x6DCF615CF17A938EULL, "vxlt_gallowscourtyard_n_7fx_0_0.vlm"},
	{0x6DCFCAEBD89382A3ULL, "524904238.ogg"},
	{0x6DD118E450641DE1ULL, "art\\characters\\playercharacter\\children\\textures\\km_cth_boya_0d.dds"},
	{0x6DD136AB3041891EULL, "art\\characters\\creatures\\highdragon\\cn_hed_fdr.phy"},
	{0x6DD1A9FB73AA8066ULL, "6125784_f.fxe"},
	{0x6DD1B5F8F15B79B2ULL, "follower_banter6148729.cl"},
	{0x6DD1BDEEC3C8A862ULL, "hlac_dr_dungeon_3217_0y.dds"},
	{0x6DD247EA4C499BDFULL, "leave_001.dds"},
	{0x6DD25D686EDEE628ULL, "vxca_cave_2cs_0_1.vlm"},
	{0x6DD2EDBCAF98D928ULL, "dr_0_deeproads_20354.anb"},
	{0x6DD3B93B3BED506DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_1_ian.ani"},
	{0x6DD3C165F0270354ULL, "sst_ss_anders.gda"},
	{0x6DD447BEBB110394ULL, "hla_ht_0_hightown_a3_-131070_0z.dds"},
	{0x6DD47CBAAB031AF6ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_act_c.mmh"},
	{0x6DD491DB5E06444FULL, "6180429_f.fxe"},
	{0x6DD4C545969752DAULL, "vxht_0_hightown_c2r_0_0.vlm"},
	{0x6DD52C591188A4E0ULL, "vxca_cavern1_haunted_egt_0_0.vlm"},
	{0x6DD544CD6C159EE9ULL, "493391419.ogg"},
	{0x6DD603B97033F1ABULL, "vxlt_0_lowtown_gaz_0_0.vlm"},
	{0x6DD60632790EC987ULL, "6070116_m.fxe"},
	{0x6DD6081F7C256413ULL, "hlac_lt_0_lowtown_a2_n_c_65536_2z.dds"},
	{0x6DD61082FB32C22AULL, "llac_lt_undercitydungeon_-1_7.dds"},
	{0x6DD643D2D9195637ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_thank_the_maker.gad"},
	{0x6DD64F75AEB17F8AULL, "339452806.ogg"},
	{0x6DD727B47261E37AULL, "lt_warehouse_n_1_0_ll.mmh"},
	{0x6DD74024A9D71064ULL, "sf_text.psh.fxo"},
	{0x6DD75F2996EA38C3ULL, "llac_dr_0_deeproads_5839_0.dds"},
	{0x6DD7D00487177B44ULL, "359890181.ogg"},
	{0x6DD82872CAEB1843ULL, "dae011_drunk2_sta6144332.cl"},
	{0x6DD89375CA138FF8ULL, "545962829.ogg"},
	{0x6DD89621E62B6B0AULL, "vxw_woundedcoast_111_0_0.vlm"},
	{0x6DD900DA5CBB74C2ULL, "817479972.ogg"},
	{0x6DD9D039E1F07496ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lowdoor2_d.dds"},
	{0x6DDA95D5FD79E051ULL, "vxca_caverns1_mines_j2c_0_2.vlm"},
	{0x6DDAA85BCAB336C3ULL, "6034651_m.fxe"},
	{0x6DDB3187996B52B5ULL, "988733520.ogg"},
	{0x6DDB72ED819A1887ULL, "219657593.ogg"},
	{0x6DDB84A6BC07A996ULL, "lgt311_shw_mer_mettin6156543.cl"},
	{0x6DDB8DCAA007D48FULL, "one200_gold_followers6117250.cl"},
	{0x6DDBBAC803B5E10FULL, "static_static_p26.psh.fxo.dx11.assembly"},
	{0x6DDC0B6C59F4870EULL, "vxht_mansion2_3gm_0_0.vlm"},
	{0x6DDC28788858C21DULL, "6114061_m.fxe"},
	{0x6DDC58D2D91DE5C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_thank_the_maker.evt"},
	{0x6DDC98BA0859868BULL, "vxlt_0_lowtown_docks_n_as4_0_0.vlm"},
	{0x6DDCB83B783581D6ULL, "53452810.ogg"},
	{0x6DDCBDDF65553F12ULL, "96680537.ogg"},
	{0x6DDCEC3595F9619EULL, "43572171.ogg"},
	{0x6DDD5C82CA2569A1ULL, "6085103_m.fxe"},
	{0x6DDE2851DDEDE945ULL, "ave331_jeven6138768.cl"},
	{0x6DDE4F03C3576CD3ULL, "6076392_m.fxe"},
	{0x6DDE80C422F08D5EULL, "734996776.ogg"},
	{0x6DDEB30F825F84AFULL, "mcr310_argument6074475.cl"},
	{0x6DDED5578C845FFEULL, "6126485_m.fxe"},
	{0x6DDEF09F37DA54D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_loop.ani"},
	{0x6DDFB9669AA2CE57ULL, "art\\levels\\races\\design\\prp_chesta.phy"},
	{0x6DE02D79E201C1A8ULL, "160806328.ogg"},
	{0x6DE06C54F4079F0CULL, "492834679.ogg"},
	{0x6DE088BA39F52509ULL, "combat_dw_range.gda"},
	{0x6DE0E47E0C33ECACULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyehigh.ani"},
	{0x6DE164F4B2DBC0F4ULL, "dae000_bodhan_letter6144322.cl"},
	{0x6DE16E816ABE5AB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\ef_injured_movement.anb"},
	{0x6DE216EF9C2B43C9ULL, "699194858.ogg"},
	{0x6DE258E14FDC8F4BULL, "hlac_lt_undercityencounter_-1_40y.dds"},
	{0x6DE332625E19C38CULL, "hlac_ht_0_hightown_a3_0_2z.dds"},
	{0x6DE3FA69D69D8048ULL, "hlac_w_encounter2_65540_0z.dds"},
	{0x6DE3FE29B958E431ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanr.ani"},
	{0x6DE4697E9C9C9E95ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blendinside_1.mao"},
	{0x6DE4A6AF84BC34A5ULL, "hightown_fem_noble1_sta6174601.cl"},
	{0x6DE4F4716EF0A2F0ULL, "6052959_m.fxe"},
	{0x6DE560DBD2299781ULL, "6030420_m.fxe"},
	{0x6DE578FE85BDA472ULL, "gen00fl_varric6076087.cl"},
	{0x6DE5ADFC40913D5BULL, "6115503_m.fxe"},
	{0x6DE5D0273EC84D68ULL, "6084017_m.fxe"},
	{0x6DE5D4D4E6093337ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robd.mmh"},
	{0x6DE5D8D4E6093A44ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robd.mao"},
	{0x6DE5EAD4E60958A9ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robd.msh"},
	{0x6DE6E4BBB3E99756ULL, "l_and331ar_sewers.wbd"},
	{0x6DE7488C88AE6EE2ULL, "qcr000_varric6066271.cl"},
	{0x6DE777105B33A312ULL, "790107871.ogg"},
	{0x6DE787CDC40E3DAEULL, "mcr310_sibling6095353.cl"},
	{0x6DE799D0F683EE8FULL, "vxca_caverns2_8lq_0_0.vlm"},
	{0x6DE7F17E0C3A21EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyehigh.gad"},
	{0x6DE89382A7B99E2AULL, "cn_hed_abma_0t.dds"},
	{0x6DE9246381336992ULL, "vxht_0_hightown_av4_0_0.vlm"},
	{0x6DE94BE347C4282BULL, "nv.cut"},
	{0x6DE97B66BACAF260ULL, "follower_banter6167863.cl"},
	{0x6DEA10E3A44B6FB1ULL, "vxdr_dungeon_1mi_0_0.vlm"},
	{0x6DEA151E85265CCDULL, "gen00fl_merrill6130098.cl"},
	{0x6DEA236EA2ED9E7AULL, "hm_qfq_thug.mor"},
	{0x6DEA5EDA099C4B55ULL, "isa311_isabela6091823.cl"},
	{0x6DEA70D2D92A4BBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_thank_the_maker.ani"},
	{0x6DEAD5CC7BB55B77ULL, "6108083_f.fxe"},
	{0x6DEAD9A727E1C749ULL, "llac_ht_keep_n_4519_8.dds"},
	{0x6DEB0929B95F15D3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanr.gad"},
	{0x6DEB0E26DEA23EC3ULL, "llac_lt_0_lowtown_n_65533_1.dds"},
	{0x6DEB273C3C7B575DULL, "nat320_delilah6065936.cl"},
	{0x6DEBEC4F0144F398ULL, "6099473_m.fxe"},
	{0x6DEC17953B3FF1C1ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.impa.ani"},
	{0x6DEC8498E89F1B74ULL, "6180615_f.fxe"},
	{0x6DECD5D3EC6B695DULL, "1325141.ogg"},
	{0x6DECDAB4BB1EEC9AULL, "gen_im_arm_cht_nobd_t4.uti"},
	{0x6DED4D997AA0E0DFULL, "736135336.ogg"},
	{0x6DEE48CB197E1670ULL, "vxlt_0_lowtown_docks_bgc_0_0.vlm"},
	{0x6DEE7D9BAA32C08CULL, "gen_im_trs_gen_hvy_hlm_r2.uti"},
	{0x6DEF15E0DCD72152ULL, "937920725.ogg"},
	{0x6DEF631EE1CE54AAULL, "hlac_lt_0_lowtown_docks_n_-3_3x.dds"},
	{0x6DF00FAB5DD63BA0ULL, "w_bonepit_182_0.phy"},
	{0x6DF012AD3E7ADB26ULL, "zz_dre_debug6056977.cl"},
	{0x6DF0AA3E560D1A54ULL, "vxlt_0_lowtown_n_h7k_0_0.vlm"},
	{0x6DF0D0984A62001DULL, "vxca_caverns2_62k_0_0.vlm"},
	{0x6DF0E0ECB8A64DACULL, "6077824_m.fxe"},
	{0x6DF10629B9637C38ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_leanr.evt"},
	{0x6DF15E30F42D6FF2ULL, "322180604.ogg"},
	{0x6DF15E984399D486ULL, "dae300_alistair_king6134982.cl"},
	{0x6DF16D5D4D94FD7FULL, "hlac_w_sundermount_-131069_0y.dds"},
	{0x6DF171BB2D1F1BA5ULL, "vxlt_gallowstemplar_n_5qv_0_0.vlm"},
	{0x6DF26DD677622539ULL, "12571067.ogg"},
	{0x6DF27DF849FD65BFULL, "vxlt_0_lowtown_a2_n_c_enx_0_0.vlm"},
	{0x6DF28780B0F5528EULL, "art\\vfx\\models\\vfx_crater.msh"},
	{0x6DF29CA34A7080D4ULL, "hlac_lt_0_lowtown_docks_-3_7y.dds"},
	{0x6DF2A180B0F57F7CULL, "art\\vfx\\models\\vfx_crater.mmh"},
	{0x6DF2F9D5C998CC1EULL, "gen00fl_merrill6150673.cl"},
	{0x6DF3199F37EB4310ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_loop.gad"},
	{0x6DF31D13CFDD8D84ULL, "gen00fl_anders6128397.cl"},
	{0x6DF322953B462363ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.impa.gad"},
	{0x6DF325900F4CA323ULL, "6096196_m.fxe"},
	{0x6DF3A02CE0F59C81ULL, "vxlt_gallowstemplar_n_5nk_0_0.vlm"},
	{0x6DF41BD326C533B5ULL, "lt_0_lowtown_docks_a3_121_0.msh"},
	{0x6DF41DD326C5372BULL, "lt_0_lowtown_docks_a3_121_0.mmh"},
	{0x6DF429D326C54BC8ULL, "lt_0_lowtown_docks_a3_121_0.mao"},
	{0x6DF42B64F6F81C08ULL, "lt_0_lowtown_docks_n_133_0.msh"},
	{0x6DF43D64F6F83A99ULL, "lt_0_lowtown_docks_n_133_0.mao"},
	{0x6DF44964F6F84F32ULL, "lt_0_lowtown_docks_n_133_0.mmh"},
	{0x6DF497D60FAC85BDULL, "142729965.ogg"},
	{0x6DF4C9B1C4324242ULL, "hlac_lt_backalley_n_0_4y.dds"},
	{0x6DF649933598EF92ULL, "dae000_merchant_weapons_dm6164402.cl"},
	{0x6DF6691B9B1DFF70ULL, "lt_warehouse_n_5_0.anb"},
	{0x6DF6DE3363D612A1ULL, "follower_barks6153381.cl"},
	{0x6DF71EDDCBA188EDULL, "m_dae310ar_docks_nt.bnk"},
	{0x6DF7287539194F42ULL, "vxlt_gallowstemplar_f_454_0_0.vlm"},
	{0x6DF74C6D5D76617CULL, "6163549_m.fxe"},
	{0x6DF781FFD748DF08ULL, "art\\levels\\ztools\\default\\generic\\2wl_basic.mmh"},
	{0x6DF78DFFD748F3ABULL, "art\\levels\\ztools\\default\\generic\\2wl_basic.mao"},
	{0x6DF79FFFD7491232ULL, "art\\levels\\ztools\\default\\generic\\2wl_basic.msh"},
	{0x6DF810BB83D324E1ULL, "6051739_m.fxe"},
	{0x6DF8237A697487CAULL, "6180489_m.fxe"},
	{0x6DF83C08EB1EE56AULL, "hlac_lt_0_lowtown_a2_n_c_0_6z.dds"},
	{0x6DF86789C481C624ULL, "qun110tr_petrice_mugger.ncs"},
	{0x6DF86E0209FCDF00ULL, "dae000_gamlen6102415.cl"},
	{0x6DF8B4FE0F783457ULL, "hlac_lt_0_lowtown_docks_n_65533_2z.dds"},
	{0x6DF8C87B4714F3DAULL, "lt_0_lowtown_docks_a3_n_121_0.msh"},
	{0x6DF8D67B47150B13ULL, "lt_0_lowtown_docks_a3_n_121_0.mao"},
	{0x6DF8DA7B47151220ULL, "lt_0_lowtown_docks_a3_n_121_0.mmh"},
	{0x6DF98AE11DBCB426ULL, "337664810.ogg"},
	{0x6DF997C3396992A9ULL, "6032061_f.fxe"},
	{0x6DF9B602A79BB391ULL, "gen_im_trs_dal_arc_lbw_r1.uti"},
	{0x6DF9FF953B4C0648ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.impa.evt"},
	{0x6DFA54E94AF3F411ULL, "llac_ca_caverns1_mines_24425_8.dds"},
	{0x6DFB12539ED0386DULL, "1051966063.ogg"},
	{0x6DFB51B21F914478ULL, "llac_lt_gallowstemplar_2391_1.dds"},
	{0x6DFB620066E28DC9ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_ripple01.dds"},
	{0x6DFBB911707090FEULL, "6128685_m.fxe"},
	{0x6DFC013B53EAA597ULL, "9884129.ogg"},
	{0x6DFC5FFDD28B34F8ULL, "6144281_m.fxe"},
	{0x6DFD0E7FB9C0B1DFULL, "107498277.ogg"},
	{0x6DFDD4C80A8780AFULL, "dae000cr_junar.utc"},
	{0x6DFE072B33DBDFF5ULL, "143884506.ogg"},
	{0x6DFF15FA694251CFULL, "vxlt_gallowstemplar_n_352_0_0.vlm"},
	{0x6DFF1C1B480332D4ULL, "vxlt_0_lowtown_n_h2h_0_0.vlm"},
	{0x6DFF39886F95F1D2ULL, "qcr220_grey_wardens6060375.cl"},
	{0x6DFF803796F07774ULL, "ch1emnotnt.psh.fxo"},
	{0x6DFF8992FB83D31CULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.dura.ani"},
	{0x6DFF917EEED27C19ULL, "lt_0_lowtown_a2_n_c_19_0.anb"},
	{0x6DFFECD78ACEFAABULL, "133769004.ogg"},
	{0x6E000365A718689AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationcast.gad"},
	{0x6E002AC10EC91BD1ULL, "85333562.ogg"},
	{0x6E01091A411797C0ULL, "vxlt_0_lowtown_docks_aot_0_0.vlm"},
	{0x6E01185ECD492268ULL, "plt_cod_bks_chantry_hist_2.plo"},
	{0x6E012F5FAE94CB38ULL, "6076548_m.fxe"},
	{0x6E01499C9F314358ULL, "vxw_sundermount_36y_0_1.vlm"},
	{0x6E018DC80D61A56CULL, "vxlt_0_lowtown_a2_n_c_erk_0_0.vlm"},
	{0x6E024A3556ED0E6CULL, "vxlt_warehouse_1ea_0_0.vlm"},
	{0x6E025265F9E950AAULL, "344123377.ogg"},
	{0x6E030B894FC2A24DULL, "hlac_lt_0_lowtown_a3_n_c_-131071_0z.dds"},
	{0x6E030CDB94D1FBA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_writing.ani"},
	{0x6E03AE6D070C2F5EULL, "llac_lt_0_lowtown_docks_65533_2.dds"},
	{0x6E03F1E181A90B17ULL, "422935912.ogg"},
	{0x6E03FEFF5449AD7FULL, "vxca_caverns2_8o4_0_0.vlm"},
	{0x6E0418179E53AFEDULL, "vxlt_0_lowtown_a2_n_c_bxo_0_0.vlm"},
	{0x6E042F74C7E9B1E8ULL, "6108740_m.fxe"},
	{0x6E04BB63B647D845ULL, "vxht_bartrand_n_54m_0_0.vlm"},
	{0x6E0513BF759A3396ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_e03_0t.dds"},
	{0x6E05B7C8DA805FB5ULL, "vxdr_primevaltomb_6py_0_2.vlm"},
	{0x6E05D75D678AAD41ULL, "vxlt_gamlenshouse_n_5n_0_0.vlm"},
	{0x6E05DF3C94A1772FULL, "6180400_m.fxe"},
	{0x6E0629A0E55F095DULL, "hlac_ht_keep_3732_4y.dds"},
	{0x6E069692FB8A085EULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.dura.gad"},
	{0x6E06EA4029122EE8ULL, "vxw_encounter3_fi_0_0.vlm"},
	{0x6E070665A71E8CB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationcast.evt"},
	{0x6E0806CC41075EEFULL, "6068297_f.fxe"},
	{0x6E085D2336765A7BULL, "vxlt_darktown_6hc_0_0.vlm"},
	{0x6E086C26AF91BD71ULL, "hlac_lt_undercityencounter_-1_32z.dds"},
	{0x6E0893C76B2CB717ULL, "art\\levels\\races\\design\\prp_prpbodybagc.mmh"},
	{0x6E08A9C76B2CDC89ULL, "art\\levels\\races\\design\\prp_prpbodybagc.msh"},
	{0x6E08CBE6F431150FULL, "gen00fl_fenris6126846.cl"},
	{0x6E090DA5169D7638ULL, "gen00fl_anders6114462.cl"},
	{0x6E0983DD53300240ULL, "gen00fl_isabela6098203.cl"},
	{0x6E09BECB31A32E10ULL, "15148375.ogg"},
	{0x6E0A4FBF606F19E5ULL, "6109690_m.fxe"},
	{0x6E0A6B1D3660C2EAULL, "162160860.ogg"},
	{0x6E0A874EB04D503AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\runfront.gad"},
	{0x6E0A9E684D8E2F56ULL, "vxlt_blackemporium_tu_0_0.vlm"},
	{0x6E0B25EF123E792DULL, "532259506.ogg"},
	{0x6E0B97B2C293CF9AULL, "6139650_m.fxe"},
	{0x6E0C0FE5C9C8B1E5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\blendtree\\doorsimple.anb"},
	{0x6E0CAEFBE0868CC5ULL, "character_character_p62.psh.fxo.dx11.assembly"},
	{0x6E0D02F62F2EF0D2ULL, "614520532.ogg"},
	{0x6E0D7192FB8FE87DULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.dura.evt"},
	{0x6E0DAC2B9609BDFFULL, "llac_lt_blackemporium_1055_21.dds"},
	{0x6E0E36BE6E777DD1ULL, "vxlt_0_lowtown_n_f4y_0_0.vlm"},
	{0x6E0E5575110B4CA5ULL, "vxlt_gallowstemplar_n_5ot_0_0.vlm"},
	{0x6E0F18CF20455105ULL, "vxlt_0_lowtown_eqb_0_0.vlm"},
	{0x6E0F756F63C519ABULL, "dae000_bodhan.bnk"},
	{0x6E0F984DC867EE17ULL, "163023325.ogg"},
	{0x6E0FFDA02159EB49ULL, "534439975.ogg"},
	{0x6E0FFFA4FF57BDECULL, "vxlt_0_lowtown_a2_n_c_ewh_0_0.vlm"},
	{0x6E1092AF9CA74D06ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.cess.evt"},
	{0x6E1114A81FB821AEULL, "6088412_f.fxe"},
	{0x6E1124DB94DE619AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_writing.evt"},
	{0x6E1135DBC372762CULL, "6101358_m.fxe"},
	{0x6E1162651118631FULL, "ser120_ghyslain6104818.cl"},
	{0x6E118A4EB0537451ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\runfront.evt"},
	{0x6E11B98F54B16D40ULL, "spt_leafcardsshadow.psh.fxo"},
	{0x6E11CE541C10A630ULL, "follower_barks6172004.cl"},
	{0x6E11F413F903C6B1ULL, "6106879_m.fxe"},
	{0x6E121AA14ADA2F7DULL, "follower_banter6160959.cl"},
	{0x6E122BC1812BBF5CULL, "hlac_ca_cavern1_haunted_24425_7z.dds"},
	{0x6E124B2689A47689ULL, "6161281_m.fxe"},
	{0x6E12554052C719C4ULL, "vxlt_0_lowtown_docks_a3_n_aqn_0_0.vlm"},
	{0x6E1270DC463FC5FFULL, "vxca_caverns2_8j7_0_0.vlm"},
	{0x6E12776F63C7528BULL, "dae000_bodhan.cnv"},
	{0x6E130ABBB40F12C4ULL, "l_and331ar_sewers.bnk"},
	{0x6E131D40BD67429AULL, "zz_followers_banter6158021.cl"},
	{0x6E13B3CC8F0F5F11ULL, "ave200_donnic6110700.cl"},
	{0x6E13ECBCAFD06EE0ULL, "dr_0_deeproads_20354.rml"},
	{0x6E13FA1FB719AE92ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_lgtd_d.dds"},
	{0x6E14225DD384196DULL, "cod_hst_kirk_arch_1a_6.utp"},
	{0x6E14AF55192D5012ULL, "hlac_dr_dungeon_3216_6x.dds"},
	{0x6E14C7DDD29E5F73ULL, "321931817.ogg"},
	{0x6E14D09A37735004ULL, "vxht_0_hightown_cbl_0_0.vlm"},
	{0x6E14D8377C04525BULL, "seb201cr_allure.utc"},
	{0x6E14F94C1C81B04EULL, "vxht_0_hightown_bd1_0_0.vlm"},
	{0x6E150A66667E79C4ULL, "437021252.ogg"},
	{0x6E151E65A72AF280ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationcast.ani"},
	{0x6E1523A348FE4F50ULL, "vxht_0_hightown_n_a6m_0_0.vlm"},
	{0x6E152CAB5DF58935ULL, "w_bonepit_182_0.msh"},
	{0x6E152EAB5DF58CABULL, "w_bonepit_182_0.mmh"},
	{0x6E153AAB5DF5A148ULL, "w_bonepit_182_0.mao"},
	{0x6E154E60687B7283ULL, "707467721.ogg"},
	{0x6E155268395775BCULL, "hlac_ht_mansion2_818_12y.dds"},
	{0x6E155C1D1B084768ULL, "vxw_encounter3_cf_0_0.vlm"},
	{0x6E155F6CF98A3E67ULL, "plt_cod_cha_isolde.plo"},
	{0x6E163A0A353DEE30ULL, "325328986.ogg"},
	{0x6E1645E8C67E4014ULL, "llac_ca_caverns1_24423_2.dds"},
	{0x6E16ACE70E9DB133ULL, "6161052_m.fxe"},
	{0x6E16BE4105B6395FULL, "6139945_m.fxe"},
	{0x6E17297CC94AA802ULL, "6133461_m.fxe"},
	{0x6E17A9AF9CAD9351ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.cess.gad"},
	{0x6E17EE90E15A6326ULL, "26286974.ogg"},
	{0x6E1823DB94E47E8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_writing.gad"},
	{0x6E187D3077558C49ULL, "zz_lgt_debug6114337.cl"},
	{0x6E18BE4278828686ULL, "hlac_lt_undercityencounter_-1_13x.dds"},
	{0x6E18C095897AF310ULL, "527496289.ogg"},
	{0x6E1A903F71C24FD5ULL, "mcr310_confused_guard6077034.cl"},
	{0x6E1AD55A1BF71D3FULL, "llac_ht_0_hightown_n_65535_1.dds"},
	{0x6E1B14E6577B72DEULL, "982605440.ogg"},
	{0x6E1B913D95364418ULL, "follower_barks6153412.cl"},
	{0x6E1B95FCF39CF0D3ULL, "plt_dre000pt_extra_funds2.nss"},
	{0x6E1BDF939852F6CDULL, "486934850.ogg"},
	{0x6E1C496A3273DDECULL, "124866472.ogg"},
	{0x6E1C5DB7F6CAA152ULL, "6075819_m.fxe"},
	{0x6E1C702CA7E124EAULL, "and230cr_mage_girl.utc"},
	{0x6E1C7631E8DED6E5ULL, "vxht_0_hightown_c3v_0_0.vlm"},
	{0x6E1CC75203016525ULL, "vxlt_gallowstemplar_n_5bs_0_0.vlm"},
	{0x6E1D199BDA936D82ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.mmh"},
	{0x6E1DAAAF9CB200FFULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.cess.ani"},
	{0x6E1DB51314DD1BF7ULL, "gen00fl_sebastian6180508.cl"},
	{0x6E1E04283A98D2EAULL, "351776511.ogg"},
	{0x6E1E0A815009EE8CULL, "vxlt_gallowscourtyard_n_6xm_0_0.vlm"},
	{0x6E1E7C23F6D33EBDULL, "vxlt_0_lowtown_docks_n_c4a_0_0.vlm"},
	{0x6E1E8BCA9CA9199EULL, "vxht_0_hightown_c40_0_0.vlm"},
	{0x6E1E999CD4FDC723ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.dura.evt"},
	{0x6E1EA24EB05E2720ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\runfront.ani"},
	{0x6E1F70447BDB1536ULL, "vxlt_0_lowtown_n_erk_0_0.vlm"},
	{0x6E1FABBBE4404F1AULL, "292184284.ogg"},
	{0x6E205FC64464D833ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_meredeth_.ani"},
	{0x6E20BD037B9A8F09ULL, "6166381_m.fxe"},
	{0x6E20CB2C9DD99963ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clasphands.ani"},
	{0x6E20CF3D5E73E737ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_conj_c.anb"},
	{0x6E20E2B93835DE7FULL, "vxca_caverns2_8q6_0_0.vlm"},
	{0x6E20FA96096F0F09ULL, "vxlt_gallowscourtyard_n_62j_0_0.vlm"},
	{0x6E2157EF1200840FULL, "vxht_0_hightown_bwy_0_0.vlm"},
	{0x6E21BA6AF16623C7ULL, "logpile_n.dds"},
	{0x6E21C0CAAB25263FULL, "zz_ave_debug6059572.cl"},
	{0x6E21F200584D4149ULL, "vxca_caverns1_iaj_0_0.vlm"},
	{0x6E22A272F6A8DE79ULL, "vxca_caverns2_864_0_1.vlm"},
	{0x6E2314DDCBC6B217ULL, "m_dae310ar_docks_nt.wbd"},
	{0x6E23484DC7F46998ULL, "vxw_blightlands_9p_0_0.vlm"},
	{0x6E238B24646EE7AEULL, "506503748.ogg"},
	{0x6E23B159A84A9177ULL, "250741336.ogg"},
	{0x6E24112218D6DC2AULL, "vxlt_0_lowtown_docks_a3_czx_0_0.vlm"},
	{0x6E2493DFA7CE4B75ULL, "vxw_bonepitdestroyed_a3_25f_0_0.vlm"},
	{0x6E24AE77252C436CULL, "6149772_m.fxe"},
	{0x6E256813102E9EE7ULL, "6163270_m.fxe"},
	{0x6E25729CD503A3A4ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.dura.gad"},
	{0x6E259BAD36A18A06ULL, "lt_gallowscourtyard_41_0.rml"},
	{0x6E25AA258EA8D9BFULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_act_c.impa.gad"},
	{0x6E2617FF45927B44ULL, "734518031.ogg"},
	{0x6E26D85EED092E71ULL, "art\\characters\\playercharacter\\children\\textures\\km_cth_boya_0t.dds"},
	{0x6E270DC025CCA196ULL, "dae000_paivel6111573.cl"},
	{0x6E273379445B4AE5ULL, "6107890_m.fxe"},
	{0x6E2745C2763AA126ULL, "656280915.ogg"},
	{0x6E278E669AE03AAFULL, "art\\levels\\races\\design\\prp_chesta.mao"},
	{0x6E279A669AE04F4CULL, "art\\levels\\races\\design\\prp_chesta.mmh"},
	{0x6E27A0669AE0593EULL, "art\\levels\\races\\design\\prp_chesta.msh"},
	{0x6E27DB664BE566BEULL, "vxlt_0_lowtown_a3_n_c_hcr_0_0.vlm"},
	{0x6E2801644CA75E72ULL, "dae000_quintus6154576.cl"},
	{0x6E2818E9917FC371ULL, "vxw_bonepitmining_a2_72_0_0.vlm"},
	{0x6E28DAAB61914F00ULL, "122855476.ogg"},
	{0x6E295C5ECAB89791ULL, "art\\vfx\\combat\\vfx_v_bld_cast_c.impa.ani"},
	{0x6E29892319365198ULL, "mag311_emile_de_launcet6064725.cl"},
	{0x6E29984AEFE0278EULL, "507748272.ogg"},
	{0x6E29E52C69B477D1ULL, "6121582_m.fxe"},
	{0x6E2A8F25E61D2642ULL, "gen00fl_isabela_init6110247.cl"},
	{0x6E2AA8A58787D28FULL, "225943321.ogg"},
	{0x6E2AE80EDAD4EF65ULL, "vxht_0_hightown_c6u_0_0.vlm"},
	{0x6E2AEE76CB9C463BULL, "6092563_m.fxe"},
	{0x6E2AFB8747770F29ULL, "vxht_mansion2b_22x_0_0.vlm"},
	{0x6E2B74727E52F089ULL, "657662364.ogg"},
	{0x6E2BADB7AC7E0029ULL, "164984049.ogg"},
	{0x6E2C819CD509DB82ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.dura.ani"},
	{0x6E2C9585183D9523ULL, "vxlt_warehouse_10e_0_0.vlm"},
	{0x6E2C9F258EAEE5ACULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_act_c.impa.evt"},
	{0x6E2CB0E325C4CB04ULL, "dae000_meredith6085262.cl"},
	{0x6E2D33BFCC82BC99ULL, "vxht_mansion2b_1ij_0_0.vlm"},
	{0x6E2D9E0D1775FB00ULL, "419361297.ogg"},
	{0x6E2DE32C9DE44C5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clasphands.evt"},
	{0x6E2E02A972619ABEULL, "247518827.ogg"},
	{0x6E2E1893BA6C5CA3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkin.mao"},
	{0x6E2E1C93BA6C63F0ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkin.mmh"},
	{0x6E2E2A93BA6C7B0AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkin.msh"},
	{0x6E2E4F60C90A948AULL, "gen00fl_anders6125176.cl"},
	{0x6E2FC44E02A90495ULL, "darktown_fem_lowlife_mob6174158.cl"},
	{0x6E3063DD4A4359C9ULL, "vxca_caverns1_idi_0_0.vlm"},
	{0x6E30675ECABEC933ULL, "art\\vfx\\combat\\vfx_v_bld_cast_c.impa.gad"},
	{0x6E30718DA44FC5CEULL, "gen00fl_fenris6110794.cl"},
	{0x6E3085569B4C508CULL, "2428750.ogg"},
	{0x6E30A2B10FE6ED10ULL, "130159849.ogg"},
	{0x6E30B2A7E15DE034ULL, "562359172.ogg"},
	{0x6E3152EBB9FCA125ULL, "6091978_f.fxe"},
	{0x6E31936A840D57BEULL, "vxlt_undercitydungeon_1mm_0_0.vlm"},
	{0x6E31FCE17BA36ABEULL, "zz_one_debug6099753.cl"},
	{0x6E3221BF697D12B0ULL, "lt_0_lowtown_docks_a3_217_0.mao"},
	{0x6E3225BF697D19C3ULL, "lt_0_lowtown_docks_a3_217_0.mmh"},
	{0x6E3233BF697D311DULL, "lt_0_lowtown_docks_a3_217_0.msh"},
	{0x6E32481A0A927D4FULL, "vxca_caverns1_mines_cki_0_0.vlm"},
	{0x6E329ACD8A4799ECULL, "6126448_f.fxe"},
	{0x6E32B3222DC04E53ULL, "mcr310_argument6136214.cl"},
	{0x6E3333BD227668EEULL, "vxca_caverns2_7mz_0_2.vlm"},
	{0x6E33388B32BA89B8ULL, "one300_zevran6089755.cl"},
	{0x6E3352BF9A4D3905ULL, "vxw_woundedcoast_1bf_0_1.vlm"},
	{0x6E338E9909408E90ULL, "474804306.ogg"},
	{0x6E33D1E4591AFC93ULL, "character_character_p66.psh.fxo"},
	{0x6E3463CEC32AF538ULL, "6172709_m.fxe"},
	{0x6E346C0878352EF8ULL, "667801307.ogg"},
	{0x6E34C62EE8195BB6ULL, "101498425.ogg"},
	{0x6E34E22C9DEA694DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clasphands.gad"},
	{0x6E34EC4A54761076ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masf_0n.dds"},
	{0x6E3536C64476EEDDULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_meredeth_.gad"},
	{0x6E35B5E7C734728EULL, "65094257.ogg"},
	{0x6E36AA3E1B52136FULL, "6173165_m.fxe"},
	{0x6E36D4C38F2C3879ULL, "vxht_0_hightown_a2_bxr_0_0.vlm"},
	{0x6E3729FB334AF1B8ULL, "207695324.ogg"},
	{0x6E372D8272BDF1E2ULL, "vxlt_foundry_n_y9_0_0.vlm"},
	{0x6E37645ECAC4E298ULL, "art\\vfx\\combat\\vfx_v_bld_cast_c.impa.evt"},
	{0x6E386072D541741BULL, "6053332_f.fxe"},
	{0x6E3868602A27DB0DULL, "vxht_0_hightown_a2_a4q_0_0.vlm"},
	{0x6E38C9D631A06358ULL, "sse_ss_hurlock_1.gda"},
	{0x6E38ECCBA0DC312EULL, "zz_followers_banter6159057.cl"},
	{0x6E38F0120111B69CULL, "zz_one_debug6069468.cl"},
	{0x6E391A00719F847FULL, "716292116.ogg"},
	{0x6E3950A91DD204F9ULL, "vxlt_gallowstemplar_2jc_0_0.vlm"},
	{0x6E396B59A703F05FULL, "lowtown_male_lowlife2_mob6166747.cl"},
	{0x6E3983CA87EF3ACCULL, "vxca_caverns2_9dv_0_1.vlm"},
	{0x6E39927EEF03F619ULL, "lt_0_lowtown_a2_n_c_19_0.rml"},
	{0x6E3A02A33916BCB7ULL, "dae000_bartrand6031394.cl"},
	{0x6E3A276CD4A0729CULL, "6127228_m.fxe"},
	{0x6E3A54602E48C3D6ULL, "art\\characters\\creatures\\buzzard\\textures\\c_buzzard_0d.dds"},
	{0x6E3A72C1C41EAA2CULL, "6119587_f.fxe"},
	{0x6E3A87258EBAFA05ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_act_c.impa.ani"},
	{0x6E3AB74903CBE03FULL, "vxlt_0_lowtown_a2_n_c_ewg_0_0.vlm"},
	{0x6E3AFAADC5C63F93ULL, "6082028_f.fxe"},
	{0x6E3B79ADF3732672ULL, "hlac_lt_0_lowtown_n_65534_1x.dds"},
	{0x6E3C469A6BF75FEEULL, "vxht_mansion2b_33q_0_0.vlm"},
	{0x6E3C49E6552F50A2ULL, "achievements_seb.gda"},
	{0x6E3C53FE5FC74636ULL, "vxlt_0_lowtown_n_ehm_0_0.vlm"},
	{0x6E3C5DE4364A39F4ULL, "tbe300im_battle_1h.uti"},
	{0x6E3C7238DD897ADEULL, "6180456_m.fxe"},
	{0x6E3CACDFE6BAA50BULL, "86304402.ogg"},
	{0x6E3D0BF42304CCFBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_varian.evt"},
	{0x6E3D139FACE3FE18ULL, "vxht_brothel_n_6ux_0_0.vlm"},
	{0x6E3DBDF7E0D5925CULL, "727685195.ogg"},
	{0x6E3E4F2DF6186CBAULL, "6152536_m.fxe"},
	{0x6E3E885518C7FDA7ULL, "915130659.ogg"},
	{0x6E3E9591F36F7A26ULL, "831331177.ogg"},
	{0x6E3EBC438FD14068ULL, "hf_iconicfemale_normark.mop"},
	{0x6E3EBC438FD1406AULL, "hf_iconicfemale_normark.mor"},
	{0x6E3F00914A800223ULL, "6151161_m.fxe"},
	{0x6E3F475931808BA9ULL, "6113262_m.fxe"},
	{0x6E3FA040B17853B5ULL, "6076857_m.fxe"},
	{0x6E40C012E8C86993ULL, "vxlt_0_lowtown_docks_n_b82_0_0.vlm"},
	{0x6E40EB27645C3A75ULL, "vxlt_gallowsprison_n_5s8_0_0.vlm"},
	{0x6E414E47BE225A98ULL, "vxw_sundermount_26m_0_0.vlm"},
	{0x6E415B4FD2112272ULL, "vxht_0_hightown_a2_4of_0_0.vlm"},
	{0x6E419E0335DB0897ULL, "lgt200ip_eau_tome01.utp"},
	{0x6E41A30AE4703777ULL, "vxlt_foundry_n_wz_0_0.vlm"},
	{0x6E41B3DCC4E0B6F4ULL, "hlac_lt_0_lowtown_a2_131071_5y.dds"},
	{0x6E41B5DEF765F7F0ULL, "zz_followers_banter6157769.cl"},
	{0x6E42056F08A53B9BULL, "424300354.ogg"},
	{0x6E42689AA6835221ULL, "vxw_bonepitmining_a2_fm_0_0.vlm"},
	{0x6E42A59AFB036C97ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01b.phy"},
	{0x6E42A87BB7E4587EULL, "hlac_lt_0_lowtown_docks_a3_0_3x.dds"},
	{0x6E42E73E371EC29DULL, "6104279_m.fxe"},
	{0x6E4301941B4CFF3DULL, "6170724_m.fxe"},
	{0x6E4302F83AA16AAAULL, "vxca_caverns2_7n1_0_0.vlm"},
	{0x6E430FB6005D2C0EULL, "vxw_encounter3_gl_0_0.vlm"},
	{0x6E438F3649A369C9ULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_p.impa.ani"},
	{0x6E43FDEA5C81FB7FULL, "240973642.ogg"},
	{0x6E4404F4230ADF3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_varian.gad"},
	{0x6E4426150EE796D3ULL, "478529369.ogg"},
	{0x6E4444792498C394ULL, "6109061_f.fxe"},
	{0x6E4454156C9C5840ULL, "840295310.ogg"},
	{0x6E4484799F7F8186ULL, "vis121_saemus6099979.cl"},
	{0x6E44FCA3756BF471ULL, "vxw_bonepitmining_a2_90_0_0.vlm"},
	{0x6E45189BDAB4BE29ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.anb"},
	{0x6E4587A4CE7DDF31ULL, "hlac_ht_0_hightown_a3_65537_2y.dds"},
	{0x6E4602A0EBB86E21ULL, "6082274_m.fxe"},
	{0x6E46AE95ACCA029EULL, "llac_dr_0_deeproads_5840_1.dds"},
	{0x6E471FFE68F7581BULL, "vxlt_hangedman_27w_0_0.vlm"},
	{0x6E47CA0E58C17BB2ULL, "132284868.ogg"},
	{0x6E47E98D0B91D361ULL, "vxht_mansion2_1wh_0_0.vlm"},
	{0x6E481CB96F79070EULL, "6033567_m.fxe"},
	{0x6E486C0D9BFE95D6ULL, "vxlt_0_lowtown_docks_a3_n_95h_0_0.vlm"},
	{0x6E48A1DB37A57398ULL, "ht_mansion_n_4152.rml"},
	{0x6E48E49EB614C2ACULL, "6114287_f.fxe"},
	{0x6E490C82CA148DEAULL, "vxlt_undercityencounter_13o_0_1.vlm"},
	{0x6E494CB2C445DEECULL, "w_blightlands_172_0.phy"},
	{0x6E4960AFBB39FFD3ULL, "vxlt_0_lowtown_n_e39_0_0.vlm"},
	{0x6E49AA3649A8034BULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_p.impa.gad"},
	{0x6E49C6F47462B2B7ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.impa.gad"},
	{0x6E49CC3D5E96E738ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_conj_c.mmh"},
	{0x6E4ACE71CB47B9B3ULL, "follower_banter6161086.cl"},
	{0x6E4B03F42310FC3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_varian.ani"},
	{0x6E4B069E046D3C09ULL, "fa_dust_f.mao"},
	{0x6E4C6E569CE7BA01ULL, "6127089_m.fxe"},
	{0x6E4CCFC800C67104ULL, "173122048.ogg"},
	{0x6E4D6766885DCAB2ULL, "arc_mit_06.mao"},
	{0x6E4D7BF368273882ULL, "929435149.ogg"},
	{0x6E4DC66A5CB24AABULL, "mag170_thrask6054050.cl"},
	{0x6E4DC7A16C04B167ULL, "one121_trap6112788.cl"},
	{0x6E4DEC075B594346ULL, "vxdr_primevaltomb_2xp_0_0.vlm"},
	{0x6E4E494C3DE0B56EULL, "hlac_lt_undercityserialkiller_-1_34y.dds"},
	{0x6E4E57B239BAA320ULL, "593235753.ogg"},
	{0x6E4E64D09B3E8DD8ULL, "dae000_fem_commoner2_mob6174586.cl"},
	{0x6E4EC0BFE45A124EULL, "vxht_0_hightown_bp5_0_0.vlm"},
	{0x6E4ECCEF813B80E9ULL, "gen00fl_fenris6126857.cl"},
	{0x6E4F31EFDABE8213ULL, "vxlt_0_lowtown_docks_n_b55_0_0.vlm"},
	{0x6E4F3311D3609BB5ULL, "ser120_ghyslain6104180.cl"},
	{0x6E4F4D3A30CCC7A4ULL, "vxlt_warehouse_n_ea_0_0.vlm"},
	{0x6E4FA0BA87E08526ULL, "794791485.ogg"},
	{0x6E4FB80498CC5D10ULL, "vxht_mansion2b_2ei_0_1.vlm"},
	{0x6E4FDBF474674244ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.impa.evt"},
	{0x6E50020DC3F48EA9ULL, "vxlt_0_lowtown_docks_b60_0_1.vlm"},
	{0x6E504864F746AC2DULL, "lt_0_lowtown_docks_n_133_0.phy"},
	{0x6E504E7950ADAA11ULL, "395270078.ogg"},
	{0x6E508DE3263C72ACULL, "144502206.ogg"},
	{0x6E50A73649AE1CB0ULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_p.impa.evt"},
	{0x6E50B63F0B8644BAULL, "sse_ss_meeran.gda"},
	{0x6E51137F2709D68BULL, "vxht_mansion2b_3ej_0_0.vlm"},
	{0x6E51552ADCF05D1EULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearmode.ani"},
	{0x6E5157978E000A3CULL, "6169994_m.fxe"},
	{0x6E517242CCE67F48ULL, "vxht_mansion2_1wm_0_0.vlm"},
	{0x6E51E28ACC0D9EB7ULL, "llac_w_bonepitmining_a2_131073_0.dds"},
	{0x6E520382FAFA1EDBULL, "vxlt_undercityserialkiller_1d3_0_1.vlm"},
	{0x6E520914BF7CE02AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3a.mao"},
	{0x6E520D14BF7CE779ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3a.mmh"},
	{0x6E521714BF7CF867ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3a.msh"},
	{0x6E523F20B7CB348AULL, "gen00fl_merrill6106869.cl"},
	{0x6E534FA7E2E3079FULL, "llac_lt_0_lowtown_a3_n_c_-65537_5.dds"},
	{0x6E53A2300E0E38DBULL, "follower_banter6171951.cl"},
	{0x6E53B87D73186979ULL, "vxht_0_hightown_a2_bfp_0_0.vlm"},
	{0x6E53EA68C2F2F6E3ULL, "hlac_ht_0_hightown_a2_-196608_0y.dds"},
	{0x6E53F57485037B14ULL, "6087171_m.fxe"},
	{0x6E5473C44AA7C5F2ULL, "6039806_m.fxe"},
	{0x6E553CCB424814CEULL, "lt_0_lowtown_docks_n_11_0_ll.mmh"},
	{0x6E55FD723A0E8291ULL, "vxht_0_hightown_n_b74_0_0.vlm"},
	{0x6E561753344F50A3ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_0x.dds"},
	{0x6E5662F43E56A671ULL, "6027712_m.fxe"},
	{0x6E57380E043485A0ULL, "199882979.ogg"},
	{0x6E57AE12D147B753ULL, "872449506.ogg"},
	{0x6E57D205D623C8F2ULL, "mrl000_marethari6068898.cl"},
	{0x6E58C287E2628152ULL, "lt_gallowscourtyard_52_0.rml"},
	{0x6E59294FB31D2DF6ULL, "284589523.ogg"},
	{0x6E59A1F5501015D6ULL, "one300cr_crow_2.utc"},
	{0x6E59C9705AABE16BULL, "hlac_lt_0_lowtown_docks_n_-262143_0y.dds"},
	{0x6E59FDFE8E45F166ULL, "zz_aveline_guardcaptain.utc"},
	{0x6E5AB1240C6BD028ULL, "gen_im_wep_rog_dua_act1_06.uti"},
	{0x6E5C1D4940CC00EBULL, "vxht_mansion_n_5ci_0_0.vlm"},
	{0x6E5C3940ECDF79F5ULL, "vxdr_primevaltomb_99g_0_0.vlm"},
	{0x6E5C5DE44D4F5BC6ULL, "vxdr_primevaltomb_2wq_0_0.vlm"},
	{0x6E5C9B41E1846B22ULL, "zz_vault_debug6177798.cl"},
	{0x6E5CD241B75566BBULL, "783862274.ogg"},
	{0x6E5CF1E3802898A9ULL, "6093975_m.fxe"},
	{0x6E5CF8E286B13D65ULL, "ser210_witness6168133.cl"},
	{0x6E5D77959EE23521ULL, "257888182.ogg"},
	{0x6E5DE4F3EF99EEE4ULL, "gen00fl_player6207740.cl"},
	{0x6E5DF3F47473A83DULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.impa.ani"},
	{0x6E5E0C30DB9B638EULL, "vxht_mansion2_32f_0_0.vlm"},
	{0x6E5E391DBB866663ULL, "hlac_lt_0_lowtown_a2_n_c_65534_4z.dds"},
	{0x6E5E43D37C060116ULL, "ht_keep_n_3739.rml"},
	{0x6E5E68A76CB3138FULL, "zz_mrl_debug6184185.cl"},
	{0x6E5EAA063DFF4872ULL, "6018897_f.fxe"},
	{0x6E5EB749BF346DD9ULL, "6108806_m.fxe"},
	{0x6E5EDCC2E834E1C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_confused.ani"},
	{0x6E5F4B27FFE507A3ULL, "6125148_m.fxe"},
	{0x6E5F4C548A6F8321ULL, "vxw_bonepitmining_a2_hc_0_0.vlm"},
	{0x6E5F605AA2670CE7ULL, "6174898_m.fxe"},
	{0x6E5F6D2ADCFCC3EFULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearmode.evt"},
	{0x6E5F746E142CCACAULL, "matt_warrior_1.cut"},
	{0x6E5F948DE75A08A8ULL, "pro000cr_refugee3.utc"},
	{0x6E6011F15C79503FULL, "w_sundermount_194_0.msh"},
	{0x6E6017F15C795AE1ULL, "w_sundermount_194_0.mmh"},
	{0x6E601CBACDD2A92FULL, "17558937.ogg"},
	{0x6E6023F15C796E82ULL, "w_sundermount_194_0.mao"},
	{0x6E6035B5C94F5AF9ULL, "llac_w_bonepitdestroyed_a3_3_0.dds"},
	{0x6E603A0BCCDC3507ULL, "6149339_m.fxe"},
	{0x6E608C04CD4DF448ULL, "6084754_m.fxe"},
	{0x6E60A0299D64BC49ULL, "hlac_lt_undercityserialkiller_-1_14x.dds"},
	{0x6E61768A04C3C513ULL, "mrl150_merrill6085715.cl"},
	{0x6E62D44163341A57ULL, "lt_0_lowtown_docks_a3_8_0_lc.mmh"},
	{0x6E6341523D45EBFDULL, "6053939_m.fxe"},
	{0x6E635B9DAF8C823DULL, "hlac_w_bonepitmining_a2_0_1z.dds"},
	{0x6E6396F84EA05FBCULL, "97440819.ogg"},
	{0x6E640D0C11E3F6A3ULL, "6104426_m.fxe"},
	{0x6E64AD6D28AC1C6EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothd.phy"},
	{0x6E64B593C0B15C35ULL, "764911791.ogg"},
	{0x6E64C8744E02202CULL, "vxca_caverns1_gkg_0_0.vlm"},
	{0x6E64F28CAB4BC0C0ULL, "mag360_thrask6043559.cl"},
	{0x6E64F7C2E8397B4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_confused.gad"},
	{0x6E656E2ADD013098ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearmode.gad"},
	{0x6E657017F6BBEAFDULL, "446551657.ogg"},
	{0x6E65C22CFBFAFFE3ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgte.mmh"},
	{0x6E65CBF3F230BB9AULL, "6133627_m.fxe"},
	{0x6E65D02CFBFB173DULL, "art\\characters\\playercharacter\\elffemale\\glv_lgte.msh"},
	{0x6E664B49753CCB6EULL, "mag310cr_scion_3.utc"},
	{0x6E666D5A367374A1ULL, "1007357499.ogg"},
	{0x6E66C827E07860EBULL, "lowtown_male_alienage1_sta6162766.cl"},
	{0x6E679610E0A6D26BULL, "ser270_templar6127408.cl"},
	{0x6E6853FC9EB78D5FULL, "6136773_m.fxe"},
	{0x6E6862AAD4A62D08ULL, "6189035_m.fxe"},
	{0x6E6890F895EA99B1ULL, "6129191_m.fxe"},
	{0x6E68C8691FDC5B15ULL, "dae000_arishok6070891.cl"},
	{0x6E68ECD60A31B67BULL, "vxca_caverns2prc_3ri_0_0.vlm"},
	{0x6E690A330EA9A0EFULL, "41427713.ogg"},
	{0x6E694458A642D265ULL, "6085327_m.fxe"},
	{0x6E696F0446016214ULL, "vxw_bonepit_266_0_0.vlm"},
	{0x6E69928748EF5E6AULL, "477826880.ogg"},
	{0x6E6A2117A4F9250EULL, "vxlt_warehouse_st_0_0.vlm"},
	{0x6E6A576FE995AE43ULL, "767802326.ogg"},
	{0x6E6A9D8AC5D1CEDBULL, "hlac_lt_0_lowtown_docks_-131075_7x.dds"},
	{0x6E6AA912BE160DF8ULL, "vxca_cave_18t_0_0.vlm"},
	{0x6E6AAB3F73281CAFULL, "6060175_m.fxe"},
	{0x6E6AC25742E07F15ULL, "26963735.ogg"},
	{0x6E6B14A7AAA9082EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof2extrude.phy"},
	{0x6E6B56FFAA50A23EULL, "ser191_emeric6218594.cl"},
	{0x6E6BD46457513874ULL, "6102117_m.fxe"},
	{0x6E6BF4C2E83F94B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_confused.evt"},
	{0x6E6C280582A23422ULL, "6080200_m.fxe"},
	{0x6E6C6281A79B66D7ULL, "vxca_cavern1_haunted_ilb_0_0.vlm"},
	{0x6E6D7F3F1BF724F4ULL, "vxht_mansion_n_54k_0_0.vlm"},
	{0x6E6DEB1E78062EC1ULL, "llac_ht_0_hightown_a2_2_0.dds"},
	{0x6E6DFACFE2E09FDDULL, "dae100_viscount6118484.cl"},
	{0x6E6EDB04EE6DA5B7ULL, "llac_w_encounter2_camp_327678_0.dds"},
	{0x6E6F2393BAA4060BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkin.phy"},
	{0x6E6FD26D8EDD3DA4ULL, "hlac_lt_0_lowtown_docks_a3_-131072_0x.dds"},
	{0x6E7076DAFAF2DD73ULL, "gen00fl_varric6089644.cl"},
	{0x6E7099370C7DDC28ULL, "pssmcanopy.vsh.fxo"},
	{0x6E70D3BEA128CD1CULL, "lt_gallowstemplar_f_2384.rml"},
	{0x6E70F4EDA6FE2893ULL, "vxca_caverns2_96b_0_0.vlm"},
	{0x6E71078E935BF6A5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_freize_d.dds"},
	{0x6E713916619E07B1ULL, "467218779.ogg"},
	{0x6E7171B18666BC04ULL, "kf_amb_darktown_urchin.mor"},
	{0x6E71896CE5CD26FFULL, "6087756_m.fxe"},
	{0x6E719B7D76274DB9ULL, "6160620_m.fxe"},
	{0x6E71AA3E4CDCD280ULL, "llac_lt_gallowstemplar_n_2395_0.dds"},
	{0x6E71E3B12BF32F3AULL, "hf_bethany_h2s1.mor"},
	{0x6E7257D02FD97937ULL, "vxdr_primevaltomb_6no_0_0.vlm"},
	{0x6E729E8F770A7B2EULL, "vxht_chantry_n_2zh_0_0.vlm"},
	{0x6E72A1B7F0601688ULL, "6057365_m.fxe"},
	{0x6E72C27B39E1F034ULL, "6154558_m.fxe"},
	{0x6E72CFD65837822AULL, "vxlt_gamlenshouse_n_jp_0_0.vlm"},
	{0x6E734BC6536AA513ULL, "166915241.ogg"},
	{0x6E7355CA0ABEEAC6ULL, "hm_circlemage03.mor"},
	{0x6E73DDFFA09F26F0ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_3.dds"},
	{0x6E740B78EAB72D5CULL, "805920365.ogg"},
	{0x6E745C2EB2BD2600ULL, "243947995.ogg"},
	{0x6E747B07D905C746ULL, "art\\vfx\\combat\\vfx_m_sph_groupheal_c.impa.evt"},
	{0x6E74A0BA316AF857ULL, "vxlt_0_lowtown_a3_n_c_e21_0_0.vlm"},
	{0x6E74D25FA45234A5ULL, "6090879_m.fxe"},
	{0x6E74EFE768EBC413ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.dura.evt"},
	{0x6E74F3694E8136EDULL, "62461712.ogg"},
	{0x6E756D1086513719ULL, "vxht_mansion2b_1pm_0_0.vlm"},
	{0x6E7629F273BD0150ULL, "gen00fl_isabela6075251.cl"},
	{0x6E76922F161745A8ULL, "vxw_bonepitdestroyed_a3_279_0_0.vlm"},
	{0x6E76B7881B8524A7ULL, "997860732.ogg"},
	{0x6E76C32CFC097408ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgte.phy"},
	{0x6E76D1E1E6993A53ULL, "03_smoke.mao"},
	{0x6E76FAAF32726708ULL, "6019190_m.fxe"},
	{0x6E777182505E87DAULL, "mag100_karras6052930.cl"},
	{0x6E77C3A63C7970EAULL, "537474583.ogg"},
	{0x6E77E0E137F77A94ULL, "vxw_bonepit_217_0_0.vlm"},
	{0x6E7815D415F01250ULL, "6088418_f.fxe"},
	{0x6E78171D26AAD06FULL, "vxht_mansion_n_3z8_0_0.vlm"},
	{0x6E787D00F8DA3350ULL, "dae200_bran6094724.cl"},
	{0x6E7892F496EF3D8EULL, "vxlt_warehouse_tu_0_0.vlm"},
	{0x6E79010324B831EBULL, "vxht_mansion_n_5ik_0_0.vlm"},
	{0x6E790F008D5B83AFULL, "zz_and_debug6134155.cl"},
	{0x6E793217347DC603ULL, "llac_ht_keep_n_3736_0.dds"},
	{0x6E796562E1E99E4BULL, "vxlt_blackemporium_io_0_0.vlm"},
	{0x6E7AD7DA0F4F3D51ULL, "188482064.ogg"},
	{0x6E7AE0D97EAEC28DULL, "6096033_m.fxe"},
	{0x6E7AE8E768F02454ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.dura.gad"},
	{0x6E7B9207D90C0C91ULL, "art\\vfx\\combat\\vfx_m_sph_groupheal_c.impa.gad"},
	{0x6E7C51F3C782F454ULL, "zz_ave_debug6207727.cl"},
	{0x6E7CAB66818B47DAULL, "w_bonepit_101_0.msh"},
	{0x6E7CB966818B5F13ULL, "w_bonepit_101_0.mao"},
	{0x6E7CBD66818B6620ULL, "w_bonepit_101_0.mmh"},
	{0x6E7CF5717F68030DULL, "m_dae310ar_lowtown_nt.wbd"},
	{0x6E7D022234C82D99ULL, "vxlt_0_lowtown_a2_n_gnz_0_0.vlm"},
	{0x6E7D3640EE263046ULL, "607597329.ogg"},
	{0x6E7DE58E0D981F81ULL, "hlac_ca_cavern1_haunted_24426_10x.dds"},
	{0x6E7DFDB0D14EC36FULL, "dae000_bodhan6146889.cl"},
	{0x6E7E2AC8DDC80350ULL, "6115067_m.fxe"},
	{0x6E803576C7CF4D67ULL, "dae000_worthy6166637.cl"},
	{0x6E80C9AD21CF91B7ULL, "vxdr_primevaltomb_6un_0_0.vlm"},
	{0x6E81366ABCF3746BULL, "hlac_lt_0_lowtown_a2_n_c_65535_3x.dds"},
	{0x6E8141B34A2D9AAAULL, "vxlt_gamlenshouse_n_mq_0_0.vlm"},
	{0x6E814E9330341923ULL, "ht_0_hightown_a2_61_0.rml"},
	{0x6E817E21B7E1956FULL, "vxlt_hangedman_n_1i5_0_0.vlm"},
	{0x6E818C6B86794BA2ULL, "hlac_dr_primevaltomb_659_6z.dds"},
	{0x6E8190FCF3F39B3CULL, "plt_dre000pt_extra_funds2.plo"},
	{0x6E81B100D36A3561ULL, "vxht_mansion2_1kl_0_0.vlm"},
	{0x6E81F7E768F65C32ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.dura.ani"},
	{0x6E826674C0738E3BULL, "679534386.ogg"},
	{0x6E829307D9122D3FULL, "art\\vfx\\combat\\vfx_m_sph_groupheal_c.impa.ani"},
	{0x6E82E9612A5EAACAULL, "lowtown_fem_alienage1_sta6161570.cl"},
	{0x6E8346D262AD18EBULL, "vxca_caverns1_mines_fmb_0_0.vlm"},
	{0x6E834A08C2776047ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gear_d.dds"},
	{0x6E8374CB6C446ADDULL, "342794277.ogg"},
	{0x6E83870A9E5A136BULL, "384033543.ogg"},
	{0x6E83BD24663D610EULL, "w_sundermount_169_0.phy"},
	{0x6E843992749F2900ULL, "6197793_m.fxe"},
	{0x6E8451E380CAAA86ULL, "57925257.ogg"},
	{0x6E852D608FF8F58DULL, "480716573.ogg"},
	{0x6E85C0E88E8B267CULL, "hightown_male_noble_gossip6174089.cl"},
	{0x6E8624F42A6E19A2ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_lightbeam_2.mao"},
	{0x6E86403909BE0F70ULL, "6149493_m.fxe"},
	{0x6E8645BF670BEDB0ULL, "40641689.ogg"},
	{0x6E8681175760F0BAULL, "56687770.ogg"},
	{0x6E8692839DB4C87FULL, "fen110_anso6094944.cl"},
	{0x6E86BBDC1035745EULL, "388252853.ogg"},
	{0x6E86FC0965AB77EAULL, "6107097_m.fxe"},
	{0x6E8704D188E5560EULL, "vxlt_warehouse_yv_0_0.vlm"},
	{0x6E89051DFEF1DB23ULL, "vrc321ip_spooky5_obj_5.utp"},
	{0x6E8914ADF8A529A4ULL, "vxlt_warehouse_n_ye_0_0.vlm"},
	{0x6E89611C39D90AF5ULL, "vxw_encounter2_camp_63_0_0.vlm"},
	{0x6E89869AFB3F3A8CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01b.mmh"},
	{0x6E898C9AFB3F457EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01b.msh"},
	{0x6E89E48BAE2353C4ULL, "6115470_m.fxe"},
	{0x6E8A23B2C47D2FB7ULL, "w_blightlands_172_0.mmh"},
	{0x6E8A27B2C47D36C4ULL, "w_blightlands_172_0.mao"},
	{0x6E8A2AAE77B32816ULL, "hlac_ht_0_hightown_a3_-196608_1x.dds"},
	{0x6E8A39B2C47D5529ULL, "w_blightlands_172_0.msh"},
	{0x6E8A59E7ECACCEABULL, "lgt320_shw_ors_marlein6156316.cl"},
	{0x6E8A8B678E3FFE46ULL, "zz_followers_banter6159354.cl"},
	{0x6E8AD56A2E1E8432ULL, "vxlt_undercityencounter_11l_0_0.vlm"},
	{0x6E8B050F787F8EDCULL, "lgt260_way_blood_mage6148354.cl"},
	{0x6E8B39B694BEE148ULL, "vxht_mansion2_1ki_0_0.vlm"},
	{0x6E8B62A15BEF2A06ULL, "6155946_m.fxe"},
	{0x6E8B69C220D72255ULL, "962608877.ogg"},
	{0x6E8B6E6BB892C946ULL, "vxht_playershouse_n_1vr_0_0.vlm"},
	{0x6E8B73FF26BE4619ULL, "vxlt_0_lowtown_a2_n_gky_0_0.vlm"},
	{0x6E8BD08E62648E1CULL, "vxlt_undercityserialkiller_1ta_0_0.vlm"},
	{0x6E8C18F15C9E975EULL, "w_sundermount_194_0.phy"},
	{0x6E8C5E5AADFBBF52ULL, "6139122_m.fxe"},
	{0x6E8C730C26C7D338ULL, "fen200cr_tev_slavermage_2.utc"},
	{0x6E8D06837D08FFB5ULL, "442009916.ogg"},
	{0x6E8D6A935D5BD4F2ULL, "sib111_mother6098954.cl"},
	{0x6E8D93F0E971CD55ULL, "and111_lirene6084760.cl"},
	{0x6E8DA04D67C7B873ULL, "mer320_hubert6053209.cl"},
	{0x6E8DC4E22226CE33ULL, "6161243_m.fxe"},
	{0x6E8DFFB9BCABD510ULL, "gen00fl_anders6128348.cl"},
	{0x6E8E2030356480C2ULL, "420794008.ogg"},
	{0x6E8EC1879FE76F6AULL, "dae300_cullen6140624.cl"},
	{0x6E8F244D2520147DULL, "527701223.ogg"},
	{0x6E8F3B8A13C5AA37ULL, "vxdr_primevaltomb_6xi_0_0.vlm"},
	{0x6E8F82495AF6AC2EULL, "vxht_chantry_n_2pj_0_0.vlm"},
	{0x6E8F993AB106C1F6ULL, "vxht_0_hightown_a3_c1n_0_0.vlm"},
	{0x6E900358577C82BEULL, "gen00fl_bethany6121637.cl"},
	{0x6E901DA64C818FF4ULL, "hlac_dr_dungeon_3217_9x.dds"},
	{0x6E90691AC5343AF7ULL, "68561489.ogg"},
	{0x6E908603AF94CDB7ULL, "6128767_m.fxe"},
	{0x6E90F4915B07E360ULL, "vxca_cavern1_haunted_e8y_0_0.vlm"},
	{0x6E912BB4D55A434CULL, "hlac_w_blightlands_-2_0z.dds"},
	{0x6E9155B6EB98805EULL, "qun110_petrice6037690.cl"},
	{0x6E91847415572957ULL, "vxlt_0_lowtown_a3_n_c_e83_0_0.vlm"},
	{0x6E91905AB0C731B9ULL, "llac_lt_0_lowtown_a2_131072_6.dds"},
	{0x6E91FC83AEDA65E4ULL, "6101066_m.fxe"},
	{0x6E91FFF80383042BULL, "vxlt_darktown_6c2_0_0.vlm"},
	{0x6E9205EA88BC66EEULL, "vxlt_0_lowtown_docks_n_aoi_0_0.vlm"},
	{0x6E9208E313DD43F5ULL, "78872977.ogg"},
	{0x6E92ABB937FC5F1FULL, "hlac_lt_foundry_n_320_17z.dds"},
	{0x6E93385627D61968ULL, "6152133_f.fxe"},
	{0x6E93E9295B75B647ULL, "176411793.ogg"},
	{0x6E944F37B712F937ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_stala1a.msh"},
	{0x6E946137B713179AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_stala1a.mao"},
	{0x6E946537B7131EA9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_stala1a.mmh"},
	{0x6E94C5192066FE0FULL, "6095588_m.fxe"},
	{0x6E94C6C0614701BEULL, "hlac_dr_dungeon_3216_8z.dds"},
	{0x6E94DF3A748C7C6EULL, "hlac_ca_caverns1_mines_24421_12x.dds"},
	{0x6E94F8485E58FB4DULL, "vxw_sundermount_31l_0_0.vlm"},
	{0x6E94FE63B2E0F013ULL, "zz_isa_debug6133970.cl"},
	{0x6E959E15D3C8FD09ULL, "6099422_f.fxe"},
	{0x6E95B00A2B928FC1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_pota_0.pwk"},
	{0x6E95C4FE70586EE3ULL, "6000963_f.fxe"},
	{0x6E963C23F7EEC075ULL, "zz_vault_debug6178945.cl"},
	{0x6E966D088F2746F7ULL, "454307400.ogg"},
	{0x6E96C115101045DEULL, "mer340_bonepit_comments6083144.cl"},
	{0x6E96CAB00E6B2785ULL, "6109631_f.fxe"},
	{0x6E96D0E77A9A6B7CULL, "vxw_woundedcoast_y_0_0.vlm"},
	{0x6E96ED7B795C117BULL, "w_bonepitdestroyed_a3_105_0.phy"},
	{0x6E96F0A8281A7211ULL, "6208010_m.fxe"},
	{0x6E978148038391DDULL, "vxca_caverns2prc_9po_0_0.vlm"},
	{0x6E97D2F92BCF2375ULL, "vxw_encounter2_camp_12_0_0.vlm"},
	{0x6E9840D37C3774CAULL, "ht_keep_n_3739.anb"},
	{0x6E9867E7F839C721ULL, "pro000cr_merchants_guard.utc"},
	{0x6E986D3C0FB85010ULL, "278689820.ogg"},
	{0x6E9875148495F48EULL, "566087873.ogg"},
	{0x6E987795A32B7880ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_107a_d.dds"},
	{0x6E9878416A75C6B2ULL, "qun120_arishok6049651.cl"},
	{0x6E98BADF386DCAF7ULL, "6034281_f.fxe"},
	{0x6E98F411AB064C79ULL, "vxlt_0_lowtown_a2_hgu_0_0.vlm"},
	{0x6E9900A43A7DBF85ULL, "vxca_cave_18m_0_0.vlm"},
	{0x6E9924F1CB4F9DB4ULL, "hlac_ca_caverns1_24421_9y.dds"},
	{0x6E9950368F9F87ABULL, "6168143_m.fxe"},
	{0x6E998045F130ADA5ULL, "6128732_m.fxe"},
	{0x6E9998ACE28D476FULL, "971542065.ogg"},
	{0x6E99AB9386B4F9C8ULL, "vxht_mansion2_1nf_0_0.vlm"},
	{0x6E99BE25F6B1CD29ULL, "572087305.ogg"},
	{0x6E99CCC1D45C01B2ULL, "6043627_m.fxe"},
	{0x6E9A133DF535B98FULL, "6063555_f.fxe"},
	{0x6E9B31AF4784A8ABULL, "cs_dae340_db_drag_warr.wbd"},
	{0x6E9B74C1DBF91ABEULL, "vxht_brothel_n_6k0_0_0.vlm"},
	{0x6E9B7AE10E21EEE6ULL, "199422070.ogg"},
	{0x6E9BC246C82690A7ULL, "gen00fl_anders6119895.cl"},
	{0x6E9CE65A70208C6CULL, "395587564.ogg"},
	{0x6E9D228E7D198B9EULL, "cocdownsample.vsh.fxo"},
	{0x6E9D56DF59087695ULL, "hlac_w_encounter3_-262139_0z.dds"},
	{0x6E9DA053588F9822ULL, "hlac_dr_dungeon_3214_0x.dds"},
	{0x6E9DA231EF2625FEULL, "6103040_m.fxe"},
	{0x6E9E1EFD31AA3305ULL, "vxca_caverns2prc_62f_0_0.vlm"},
	{0x6E9E469A86B6D69BULL, "584509114.ogg"},
	{0x6E9EB8203D557C33ULL, "zz_pro_debug.ncs"},
	{0x6E9EE7600D8CE971ULL, "872855376.ogg"},
	{0x6E9F78AE78E2BD90ULL, "caustic_distortion.mao"},
	{0x6E9F8FD8285FDD1BULL, "character_character_p224.psh.fxo.dx11.assembly"},
	{0x6E9FB8C65FE2CADCULL, "6171202_f.fxe"},
	{0x6EA01B3B1D107199ULL, "lt_0_lowtown_docks_24_0_ll.mmh"},
	{0x6EA0396D55345459ULL, "6113868_m.fxe"},
	{0x6EA05390C698D514ULL, "lgt000ip_pommel_sack.utp"},
	{0x6EA137ED993EB0F3ULL, "vxht_keep_n_603_0_0.vlm"},
	{0x6EA13E3BBC10205EULL, "6060101_m.fxe"},
	{0x6EA189CE9CE2A9E2ULL, "386982709.ogg"},
	{0x6EA23757A158D843ULL, "mag130_danzig6095048.cl"},
	{0x6EA29868905F1D32ULL, "vxht_0_hightown_a3_5i8_0_0.vlm"},
	{0x6EA36395EC3BA804ULL, "dae000_orsino6085505.cl"},
	{0x6EA3BBABA873365FULL, "852720584.ogg"},
	{0x6EA3EEC9521CB275ULL, "ico_helm_heavy_2.dds"},
	{0x6EA454FAF3FEF3D7ULL, "hlac_lt_0_lowtown_a2_n_c_131071_3y.dds"},
	{0x6EA4598B497B022EULL, "6101525_m.fxe"},
	{0x6EA4938CBDF16779ULL, "1064288620.ogg"},
	{0x6EA4C17C7E7EA7F6ULL, "art\\vfx\\combat\\vfx_glb_ele_p_deathblow_c.anb"},
	{0x6EA4EEBC1A140F42ULL, "prx_lower.dds"},
	{0x6EA5501E3464252CULL, "vxlt_0_lowtown_docks_a3_d0c_0_0.vlm"},
	{0x6EA5AC6D28E3B131ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothd.mmh"},
	{0x6EA5B86D28E3C5D2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothd.mao"},
	{0x6EA5C66D28E3DDEFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothd.msh"},
	{0x6EA5F1A5BDB87C49ULL, "genip_trap_arm_lever.utp"},
	{0x6EA6044F7C247F29ULL, "w_blightlands_234_0.mao"},
	{0x6EA6104F7C2493C2ULL, "w_blightlands_234_0.mmh"},
	{0x6EA6124F7C249778ULL, "w_blightlands_234_0.msh"},
	{0x6EA77BBB48AE2165ULL, "381566985.ogg"},
	{0x6EA875E997D46DF1ULL, "253534911.ogg"},
	{0x6EAA3DB02CCB9B24ULL, "fen210_hadriana6110577.cl"},
	{0x6EAAF241B237BDAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv1.evt"},
	{0x6EAB13A7AADEEAF1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof2extrude.mmh"},
	{0x6EAB1FA7AADEFE92ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof2extrude.mao"},
	{0x6EAB2DA7AADF16AFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof2extrude.msh"},
	{0x6EAB874EE081B843ULL, "966131243.ogg"},
	{0x6EABE94035861B68ULL, "6083103_f.fxe"},
	{0x6EAC27F1E313FB25ULL, "gen00fl_anders6107472.cl"},
	{0x6EAC4C9EDA37F668ULL, "6144480_f.fxe"},
	{0x6EAD467518524797ULL, "165777666.ogg"},
	{0x6EADED217D758C03ULL, "vxht_0_hightown_a3_b8a_0_0.vlm"},
	{0x6EAE178B457EDEDDULL, "419705445.ogg"},
	{0x6EAE6A49D23561D1ULL, "6020461_m.fxe"},
	{0x6EAEDC1FC1EAAE56ULL, "dae000ip_crowd_mf8_talk13.utp"},
	{0x6EAEE9A46CA897EEULL, "vxlt_0_lowtown_docks_n_aqk_0_0.vlm"},
	{0x6EAF0EEC6ADB24F5ULL, "vxca_caverns1_j51_0_0.vlm"},
	{0x6EAF8D06C9F0AE82ULL, "vxht_keep_3io_0_0.vlm"},
	{0x6EAFE181BD44D23FULL, "zz_otl_debug6060944.cl"},
	{0x6EAFEBD1F39D5D1BULL, "vxw_encounter2_camp_58_0_0.vlm"},
	{0x6EAFF597A5CF7349ULL, "172530836.ogg"},
	{0x6EB0419122ED531CULL, "295809317.ogg"},
	{0x6EB04D5F108EB632ULL, "sf_d_xy16icf32_t2.vsh.fxo"},
	{0x6EB0CF26DF48E0CAULL, "art\\vfx\\combat\\vfx_m_sph_conj_c.anb"},
	{0x6EB12E888ED77B38ULL, "vxlt_0_lowtown_a2_n_c_gob_0_0.vlm"},
	{0x6EB1E941B23DCD59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv1.gad"},
	{0x6EB23B3DD3AF59DFULL, "288201314.ogg"},
	{0x6EB264A1CBEA3649ULL, "vxca_caverns1_i7p_0_0.vlm"},
	{0x6EB281CFB7B52E09ULL, "6093622_f.fxe"},
	{0x6EB2C47C7E8AEA61ULL, "art\\vfx\\combat\\vfx_glb_ele_p_deathblow_c.mmh"},
	{0x6EB2EF03859A467EULL, "dr_primevalentrance_1724_ll.mmh"},
	{0x6EB2FD465AEC482DULL, "zz_followers_banter6159504.cl"},
	{0x6EB308A876492861ULL, "6162349_m.fxe"},
	{0x6EB315E627F82CC5ULL, "hlac_ht_0_hightown_a2_65538_0y.dds"},
	{0x6EB39428E7F2F646ULL, "art\\characters\\weapons\\dagger\\w_dgr_110.mao"},
	{0x6EB3AD42325B5668ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs05a_0n.dds"},
	{0x6EB3BEFDB497D355ULL, "432975179.ogg"},
	{0x6EB3C02F10BBC2A6ULL, "258722253.ogg"},
	{0x6EB3C9694E0C9EB1ULL, "vxht_0_hightown_a2_bwk_0_0.vlm"},
	{0x6EB3FB4591996A35ULL, "lla_ht_0_hightown_a3_131073_1.dds"},
	{0x6EB4616B8681E422ULL, "462435226.ogg"},
	{0x6EB4623890E66E0EULL, "294660450.ogg"},
	{0x6EB48DDCDF08DE96ULL, "hlac_w_sundermount_-393216_0z.dds"},
	{0x6EB49A066125D119ULL, "art\\vfx\\combat\\vfx_gui_target_ret_p.mmh"},
	{0x6EB51DA321FD1E2AULL, "zz_prmpt_debug6200983.cl"},
	{0x6EB573317F8DA23DULL, "559922940.ogg"},
	{0x6EB5B32FDD93BDECULL, "vxht_bartrand_n_4a6_0_0.vlm"},
	{0x6EB5D0809720C6DBULL, "vxlt_0_lowtown_a2_n_c_gy3_0_0.vlm"},
	{0x6EB600AD90581BB6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robb_0t.dds"},
	{0x6EB64B1BAD0AB08EULL, "zz_mer_debug6032453.cl"},
	{0x6EB657797CE07C6DULL, "zz_ave_debug6063269.cl"},
	{0x6EB65C7B62EFF4E6ULL, "231928430.ogg"},
	{0x6EB65CEC3E349A38ULL, "574428860.ogg"},
	{0x6EB75D787FFE031AULL, "518504498.ogg"},
	{0x6EB7BAC185BF12BBULL, "hlac_lt_0_lowtown_n_-65536_5z.dds"},
	{0x6EB80DB3040EDEA8ULL, "371876950.ogg"},
	{0x6EB85C19C5423876ULL, "6059292_f.fxe"},
	{0x6EB868DE16BC65CFULL, "hlac_ca_caverns1_24424_20x.dds"},
	{0x6EB8FA41B24408A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv1.ani"},
	{0x6EB921F7BCF0EEE1ULL, "6135811_m.fxe"},
	{0x6EB9638BB23D0161ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala0a_1.mao"},
	{0x6EB97873BC36AE3BULL, "character_rendertodepth_p3.vsh.fxo.dx11.assembly"},
	{0x6EB9907680B4FB9CULL, "hlac_lt_gallowstemplar_f_2393_0x.dds"},
	{0x6EBAC5F10CFB2105ULL, "mer120_hubert6070215.cl"},
	{0x6EBAD7E030D8F5AEULL, "vxht_chantry_n_2am_0_0.vlm"},
	{0x6EBC4A2830EA2CE0ULL, "vxca_cavern1_haunted_e7d_0_0.vlm"},
	{0x6EBC68B254CBDA5EULL, "6135687_m.fxe"},
	{0x6EBC8591A1B5AD20ULL, "6074362_m.fxe"},
	{0x6EBCBCBF9DCA0130ULL, "6121972_m.fxe"},
	{0x6EBCEF80D4AE927FULL, "setmapswapenabled.ncs"},
	{0x6EBCFF717F9DF837ULL, "m_dae310ar_lowtown_nt.bnk"},
	{0x6EBD503C42D16DD2ULL, "llac_ca_caverns1_24425_6.dds"},
	{0x6EBD5DEA5E1A6C4DULL, "qcr221_cut_qunari_defeated6205147.cl"},
	{0x6EBE52C54080AF94ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier1_arishokresponse.ani"},
	{0x6EBE8B506FE600BBULL, "qcr220_grey_wardens6060457.cl"},
	{0x6EBF108F73CC8813ULL, "18777085.ogg"},
	{0x6EBF402F80B77C42ULL, "vxlt_gallowstemplar_n_476_0_0.vlm"},
	{0x6EBFDC7568E22C53ULL, "vxlt_0_lowtown_n_h8i_0_0.vlm"},
	{0x6EC013AF47A39279ULL, "cs_dae340_db_drag_warr.bnk"},
	{0x6EC01B246725D401ULL, "6112020_f.fxe"},
	{0x6EC05F6D8A8FD919ULL, "6066896_m.fxe"},
	{0x6EC08CF095349477ULL, "6037610_m.fxe"},
	{0x6EC0D0E54CC6872AULL, "1037711701.ogg"},
	{0x6EC0D2810A321382ULL, "vxdr_primevaltomb_50s_0_0.vlm"},
	{0x6EC16B67AB3D3E2EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc4a.msh"},
	{0x6EC18567AB3D6A1CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc4a.mmh"},
	{0x6EC19D8EBB699E57ULL, "enc_trips_sp_cli.gda"},
	{0x6EC19EB39BB818CBULL, "vxlt_blackemporium_vb_0_0.vlm"},
	{0x6EC1ECF7A0128644ULL, "vxlt_0_lowtown_a2_n_gvs_0_0.vlm"},
	{0x6EC22939AF8751FEULL, "qapath1.arl"},
	{0x6EC297269489F721ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopeningmage.ani"},
	{0x6EC2D200C421E32FULL, "6128688_m.fxe"},
	{0x6EC2D47B7981211AULL, "w_bonepitdestroyed_a3_105_0.msh"},
	{0x6EC2E27B79813953ULL, "w_bonepitdestroyed_a3_105_0.mao"},
	{0x6EC2E67B79814060ULL, "w_bonepitdestroyed_a3_105_0.mmh"},
	{0x6EC30BA77CE84DD8ULL, "gen00fl_isabela6097286.cl"},
	{0x6EC330BBF6552979ULL, "follower_banter6169822.cl"},
	{0x6EC39EF39289C42EULL, "605581171.ogg"},
	{0x6EC39F541D99EE06ULL, "and121_karl6093862.cl"},
	{0x6EC3E3203D74E1F8ULL, "zz_pro_debug.cnv"},
	{0x6EC4250CCF89D66CULL, "vxht_bartrand_n_4f7_0_0.vlm"},
	{0x6EC4971D608AA326ULL, "mcr370_cut_epilogue6086877.cl"},
	{0x6EC4BC246674F6D1ULL, "w_sundermount_169_0.mmh"},
	{0x6EC4C82466750B72ULL, "w_sundermount_169_0.mao"},
	{0x6EC4D6246675228FULL, "w_sundermount_169_0.msh"},
	{0x6EC5107A8204088EULL, "vxw_encounter3_8e_0_0.vlm"},
	{0x6EC52EAE54DFF751ULL, "230288659.ogg"},
	{0x6EC54CA167D32B25ULL, "vxht_0_hightown_a3_c33_0_0.vlm"},
	{0x6EC54FC54086C9F6ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier1_arishokresponse.gad"},
	{0x6EC56029D0E59322ULL, "dre000_varric6207983.cl"},
	{0x6EC5923BBF8EB2F4ULL, "gen_im_wep_mag_sta_default.uti"},
	{0x6EC5A068544B0656ULL, "6084232_f.fxe"},
	{0x6EC5B3871DFC7884ULL, "vxlt_0_lowtown_docks_aop_0_0.vlm"},
	{0x6EC6A6244F8BF7F0ULL, "dae000_fenarel6147440.cl"},
	{0x6EC7706573D184F4ULL, "6054856_m.fxe"},
	{0x6EC80EECD7E11A34ULL, "framebuffereffects_diffusiondof_p6.psh.fxo.dx11"},
	{0x6EC866FD9D8C3A81ULL, "vxlt_hangedman_n_5t_0_0.vlm"},
	{0x6EC8BDCA7F4676FFULL, "6061789_m.fxe"},
	{0x6EC8C9C2E391A844ULL, "vxht_mansion2b_1se_0_0.vlm"},
	{0x6EC8FD3615CCC649ULL, "6105618_m.fxe"},
	{0x6EC95D2B2A36CAD2ULL, "vxlt_0_lowtown_n_h8l_0_0.vlm"},
	{0x6EC9A226949028C3ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopeningmage.gad"},
	{0x6EC9CA8D1AA3989DULL, "6036992_m.fxe"},
	{0x6ECA0CC58CFB13F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy_flrsh.evt"},
	{0x6ECA42B22726C850ULL, "vxw_encounter1_jf_0_1.vlm"},
	{0x6ECA92EF36FE95A0ULL, "1066824375.ogg"},
	{0x6ECAF76E93ABC5A0ULL, "6079062_m.fxe"},
	{0x6ECB1B4A3F14A528ULL, "vxdr_primevaltomb_916_0_0.vlm"},
	{0x6ECB8A860FA09453ULL, "183980019.ogg"},
	{0x6ECB9BCAC2B5472DULL, "vxht_0_hightown_a3_4an_0_0.vlm"},
	{0x6ECBFB4AE334A272ULL, "626342054.ogg"},
	{0x6ECC7F1316DF1C1EULL, "mag100_karras6034934.cl"},
	{0x6ECD5739699C45D0ULL, "art\\vfx\\materials\\vfx_blood_dec1.mao"},
	{0x6ECD61706E6AE53BULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkl.evt"},
	{0x6ECD80F383466603ULL, "6099489_m.fxe"},
	{0x6ECDA9627615D9D9ULL, "mag110_arianni6036802.cl"},
	{0x6ECE48B5EB1E4B36ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockxxl.msh"},
	{0x6ECE52B5EB1E5C34ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockxxl.mmh"},
	{0x6ECE56B5EB1E6347ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockxxl.mao"},
	{0x6ECE6F955E2F8FF3ULL, "vxht_mansion_n_6ni_0_0.vlm"},
	{0x6ECE7E606BF88F4DULL, "qun211_party_comment6072164.cl"},
	{0x6ECEAD9E45280C0FULL, "isa311_isabela6175477.cl"},
	{0x6ECEF642DF319401ULL, "366687346.ogg"},
	{0x6ECF12A4CE2FEE6DULL, "gen_im_wep_war_two_act2_08.uti"},
	{0x6ECF3D55E3A79DD6ULL, "dae000cr_dalish_apprentice2.utc"},
	{0x6ECF86736E41A6DEULL, "gen00fl_anders6113119.cl"},
	{0x6ECF94C69BD74752ULL, "dae000_keep_noblewoman_sta6162049.cl"},
	{0x6ECF9D9952F70E89ULL, "113417703.ogg"},
	{0x6ECFEB2C21CAC9F8ULL, "vxca_caverns2prc_ae9_0_0.vlm"},
	{0x6ED003C58CFF6F81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy_flrsh.gad"},
	{0x6ED0761BFA1FC6FAULL, "6143075_m.fxe"},
	{0x6ED0C6C81739385BULL, "mcr370_varric6086852.cl"},
	{0x6ED12844090B2160ULL, "vxlt_0_lowtown_a3_n_c_hd9_0_0.vlm"},
	{0x6ED148BBCB5BF3DDULL, "vxca_caverns2prc_9lk_0_0.vlm"},
	{0x6ED168777D8F37A2ULL, "vxdr_0_deeproads_f9u_0_0.vlm"},
	{0x6ED1C0E931BF4F98ULL, "vxw_sundermount_2dc_0_0.vlm"},
	{0x6ED2B29CBE71E973ULL, "lgt110_orz_harrowmont6109002.cl"},
	{0x6ED34F93D8462A37ULL, "mag131_taraneh6088728.cl"},
	{0x6ED35A706E6F457CULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkl.gad"},
	{0x6ED3825773FA210EULL, "vxw_encounter3_5f_0_0.vlm"},
	{0x6ED3BE7E59C943A5ULL, "vxht_0_hightown_a3_c42_0_0.vlm"},
	{0x6ED42C26DF8593BBULL, "6147746_m.fxe"},
	{0x6ED452E07700E915ULL, "vxht_0_hightown_n_c7q_0_0.vlm"},
	{0x6ED4D21969DBE562ULL, "llac_lt_gallowstemplar_n_2392_1.dds"},
	{0x6ED511F216804F0DULL, "vxht_brothel_n_615_0_0.vlm"},
	{0x6ED5784AE8011B20ULL, "576860444.ogg"},
	{0x6ED704C58D05902FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rdy_flrsh.ani"},
	{0x6ED714785502F0B7ULL, "ca_caverns2_mmap.dds"},
	{0x6ED7168B1D8E5E22ULL, "hlac_ca_cavern1_haunted_24423_9x.dds"},
	{0x6ED722371797DF53ULL, "qun160_vashothbandits6069280.cl"},
	{0x6ED73B9FD587C0C4ULL, "vxht_mansion2b_1vj_0_0.vlm"},
	{0x6ED8BBF99239DCBCULL, "lgt300cr_jakeson_hall.utc"},
	{0x6ED8EFA39CA0EBC8ULL, "ht_0_hightown_a2_17_0.anb"},
	{0x6ED9439C2C760E3DULL, "338981894.ogg"},
	{0x6ED953FB352DC1DAULL, "6129839_m.fxe"},
	{0x6ED97E110D402504ULL, "6098269_m.fxe"},
	{0x6ED99A0B4ADF48BEULL, "dae000_orana6149125.cl"},
	{0x6ED9C08F78D17528ULL, "one110_pryce6088404.cl"},
	{0x6ED9D226DF6BEABDULL, "art\\vfx\\combat\\vfx_m_sph_conj_c.mmh"},
	{0x6EDA43532AF18AB9ULL, "hlac_lt_undercityserialkiller_-1_6y.dds"},
	{0x6EDA59706E75627AULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkl.ani"},
	{0x6EDA89F6AEB29B6BULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw107a.mao"},
	{0x6EDB1424F545BC4EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keep.phy"},
	{0x6EDB4631DE3BA2C2ULL, "vxht_0_hightown_n_by1_0_0.vlm"},
	{0x6EDB5B27726B3DC0ULL, "character_character_p229.psh.fxo.dx11"},
	{0x6EDBC37593A05778ULL, "6118002_m.fxe"},
	{0x6EDBE929AB0199E1ULL, "86143599.ogg"},
	{0x6EDBF497CDF8C91AULL, "169318387.ogg"},
	{0x6EDC66A360244D09ULL, "gen00fl_aveline6139728.cl"},
	{0x6EDCA763DBB8F70BULL, "and111_lirene6084494.cl"},
	{0x6EDD3B265E268FFDULL, "character_character_p40.psh.fxo.dx11"},
	{0x6EDD43276E7C0A19ULL, "6068696_m.fxe"},
	{0x6EDD76810C4CBE3EULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgtb_0d.dds"},
	{0x6EDD97066148D10EULL, "art\\vfx\\combat\\vfx_gui_target_ret_p.anb"},
	{0x6EDDC2F13D5338B9ULL, "hlac_ca_caverns2prc_5802_0y.dds"},
	{0x6EDDCA2AF9F9D4F5ULL, "vxdr_primevalentrance_22a_0_0.vlm"},
	{0x6EDEC66DB9B5CE0EULL, "6112059_m.fxe"},
	{0x6EDED0B183FEB744ULL, "vxlt_0_lowtown_a2_n_ghq_0_0.vlm"},
	{0x6EDFEADC21724B11ULL, "t3_nbl_c00.tnt"},
	{0x6EE06AD3E207E03FULL, "vxht_bartrand_n_4a9_0_0.vlm"},
	{0x6EE07698F50E5388ULL, "vxca_caverns1_mines_imr_0_0.vlm"},
	{0x6EE0D03900D5B403ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_act_c.impa.evt"},
	{0x6EE19A501BC25E37ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2b.msh"},
	{0x6EE1AC501BC27C9AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2b.mao"},
	{0x6EE1B0501BC283A9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2b.mmh"},
	{0x6EE1F26E3619AE17ULL, "hlac_lt_0_lowtown_a3_n_c_65536_5x.dds"},
	{0x6EE221E93FB35C51ULL, "6005316_m.fxe"},
	{0x6EE29E03D17BAC3BULL, "6140646_m.fxe"},
	{0x6EE2C02A800163EEULL, "vxlt_0_lowtown_a2_n_erg_0_0.vlm"},
	{0x6EE2FEF7EABD11B3ULL, "vxdr_0_deeproads_9ce_0_0.vlm"},
	{0x6EE39DC0252EE299ULL, "gen00fl_anders6119545.cl"},
	{0x6EE3B4B844AEA169ULL, "6064238_m.fxe"},
	{0x6EE437F66C90A844ULL, "hlac_lt_0_lowtown_docks_a3_n_65532_1x.dds"},
	{0x6EE438A3AD053024ULL, "lt_0_lowtown_22_0.rml"},
	{0x6EE484DF9935A92EULL, "vrc221_varric6043588.cl"},
	{0x6EE490A1E42D4D7AULL, "hlac_lt_gallowstemplar_f_2388_0z.dds"},
	{0x6EE4EB083E3EA0F4ULL, "vxw_encounter2_camp_ka_0_0.vlm"},
	{0x6EE5179AABD068A1ULL, "6051317_m.fxe"},
	{0x6EE547886076B6A6ULL, "6114469_f.fxe"},
	{0x6EE584F0EAFDDD67ULL, "540030603.ogg"},
	{0x6EE5E2615D4EB871ULL, "gst_wall_lean_left.gda"},
	{0x6EE61A412FE5ED3DULL, "6116631_f.fxe"},
	{0x6EE76467AB5E01E7ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc4a.phy"},
	{0x6EE7A93900DB9084ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_act_c.impa.gad"},
	{0x6EE7D8E310E6D2BDULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_015a_0d.dds"},
	{0x6EE812134F356DA2ULL, "lt_gallowscourtyard_mmap_ph.dds"},
	{0x6EE81B380513DD84ULL, "6147222_m.fxe"},
	{0x6EE8619B5BB0B72EULL, "gen00fl_merrill6107783.cl"},
	{0x6EE86A0B28791AC3ULL, "6111590_m.fxe"},
	{0x6EE893490DCBDB34ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic_1.mao"},
	{0x6EE8D633D78D0038ULL, "6149340_f.fxe"},
	{0x6EE906E01EE0062EULL, "art\\characters\\creatures\\dragon\\c_dragonb.msh"},
	{0x6EE914A327A0752DULL, "hlac_lt_0_lowtown_a2_n_c_131071_2x.dds"},
	{0x6EE914E01EE01E7FULL, "art\\characters\\creatures\\dragon\\c_dragonb.mao"},
	{0x6EE920E01EE0321CULL, "art\\characters\\creatures\\dragon\\c_dragonb.mmh"},
	{0x6EE95513E438E228ULL, "6188655_f.fxe"},
	{0x6EE9D1489A181DA8ULL, "vxw_encounter2_camp_3e_0_0.vlm"},
	{0x6EEA64AD67329703ULL, "6093689_m.fxe"},
	{0x6EEA7BF949F0AE5AULL, "dae021_sister_alema_mob6144847.cl"},
	{0x6EEAA1B2604A714BULL, "fen120ip_to_base.utp"},
	{0x6EEC7EDCAE299055ULL, "789092923.ogg"},
	{0x6EECCEDBEA3A069DULL, "16603733.ogg"},
	{0x6EED90DD15E500B1ULL, "vxht_0_hightown_a2_bko_0_0.vlm"},
	{0x6EEDDE62800CE196ULL, "art\\vfx\\materials\\vfx_distor_si.mao"},
	{0x6EEE0AC324C04ABDULL, "284890864.ogg"},
	{0x6EEE49D9622EACAEULL, "vxca_caverns2prc_5uk_0_0.vlm"},
	{0x6EEE942918926436ULL, "follower_banter6160693.cl"},
	{0x6EEEA4A315AB8098ULL, "vxw_sundermount_2ra_0_0.vlm"},
	{0x6EEEB83900E1C962ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_act_c.impa.ani"},
	{0x6EEEE875E7046C08ULL, "vxca_caverns1_mines_ijs_0_0.vlm"},
	{0x6EEF63697A0DFCF3ULL, "6040850_m.fxe"},
	{0x6EEFA9AABD598A06ULL, "follower_banter6167115.cl"},
	{0x6EEFDBDEC3F0AFACULL, "llac_w_bonepitdestroyed_a3_2_0.dds"},
	{0x6EF11BDF832FC801ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha14a.phy"},
	{0x6EF1320771F77C6EULL, "vxlt_0_lowtown_a2_n_ewx_0_0.vlm"},
	{0x6EF148696F04E993ULL, "vxlt_gallowstemplar_n_3qz_0_0.vlm"},
	{0x6EF1604A51F76182ULL, "vxw_bonepit_yq_0_0.vlm"},
	{0x6EF170D4DCB32A33ULL, "vxdr_0_deeproads_9dd_0_0.vlm"},
	{0x6EF1BF661F84230AULL, "mag000pt_otl_debug.ncs"},
	{0x6EF1FD4F4414687BULL, "6143971_m.fxe"},
	{0x6EF2A2B8817C64EFULL, "369626629.ogg"},
	{0x6EF2F9A0C28235B1ULL, "vxlt_gallowstemplar_5j0_0_0.vlm"},
	{0x6EF35CE53034B974ULL, "vxw_encounter2_camp_ff_0_0.vlm"},
	{0x6EF3B49147ACB22FULL, "6153387_m.fxe"},
	{0x6EF4485B1E4295F7ULL, "w_encounter3_240_0.mmh"},
	{0x6EF44C5B1E429C04ULL, "w_encounter3_240_0.mao"},
	{0x6EF45E5B1E42BB69ULL, "w_encounter3_240_0.msh"},
	{0x6EF4AEB658CD25EDULL, "vxlt_gallowstemplar_n_5x9_0_0.vlm"},
	{0x6EF5352065C8A59EULL, "vxdr_primevaltomb_4zt_0_0.vlm"},
	{0x6EF680BF56F1DDBDULL, "1014198979.ogg"},
	{0x6EF73C2EDFA30E10ULL, "667707807.ogg"},
	{0x6EF758594D15B9B1ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killbag.msh"},
	{0x6EF766594D15D1CCULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killbag.mao"},
	{0x6EF772594D15E66FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_killbag.mmh"},
	{0x6EF8127A665EFCB2ULL, "940707191.ogg"},
	{0x6EF81F2E31390FD1ULL, "6163254_m.fxe"},
	{0x6EF93F3C28D9F392ULL, "59276311.ogg"},
	{0x6EFA38CCDED7589AULL, "lt_0_lowtown_19_0.anb"},
	{0x6EFA3F7FF79401C9ULL, "hf_iconicfemale.mor"},
	{0x6EFA3F7FF79401CBULL, "hf_iconicfemale.mop"},
	{0x6EFA55EAA98E8C59ULL, "526357979.ogg"},
	{0x6EFAB3972961D3C3ULL, "752536128.ogg"},
	{0x6EFB040256CBB400ULL, "vxlt_hangedman_n_tj_0_0.vlm"},
	{0x6EFB6BD9B05A7AD7ULL, "vxw_bonepitdestroyed_a3_1rr_0_0.vlm"},
	{0x6EFB715C7E523639ULL, "vxlt_gallowscourtyard_6xk_0_0.vlm"},
	{0x6EFBB46B67EAE844ULL, "vxlt_0_lowtown_a2_n_gbw_0_0.vlm"},
	{0x6EFC03461446F2F8ULL, "dae150ip_at_scaverns_a.ncs"},
	{0x6EFC422042E50D4FULL, "6090055_m.fxe"},
	{0x6EFC7E5F79ECD7D9ULL, "mag110_ws_banter.bnk"},
	{0x6EFC8529144127E9ULL, "dae000_gamlen.bnk"},
	{0x6EFCA4E74680E399ULL, "w_encounter3_86_0.msh"},
	{0x6EFCAEE74680F487ULL, "w_encounter3_86_0.mmh"},
	{0x6EFCB2E74680FBD4ULL, "w_encounter3_86_0.mao"},
	{0x6EFCB312276D1116ULL, "lgt000cr_mage_turn_in_5.utc"},
	{0x6EFCBBB65424C52EULL, "vxca_caverns2prc_5xl_0_0.vlm"},
	{0x6EFD5A52D8FA8488ULL, "vxca_caverns1_mines_igp_0_0.vlm"},
	{0x6EFD9F9A6ADD3176ULL, "6130990_m.fxe"},
	{0x6EFDD2EC9E8C8696ULL, "400483270.ogg"},
	{0x6EFEA7895FC68233ULL, "vxlt_0_lowtown_n_fed_0_0.vlm"},
	{0x6EFECA77E445D031ULL, "551282420.ogg"},
	{0x6EFF40583E03DC18ULL, "llac_ca_caverns1_24421_11.dds"},
	{0x6EFF4FB50E0E7183ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_exit.ani"},
	{0x6EFF672914432A29ULL, "dae000_gamlen.cnv"},
	{0x6EFFA3E463ED94EEULL, "vxlt_0_lowtown_a2_n_exy_0_0.vlm"},
	{0x6F0007B24AB9E7C9ULL, "gen00fl_carver6103605.cl"},
	{0x6F0059820819B35DULL, "vxht_keep_5u2_0_0.vlm"},
	{0x6F00A05F79F0FA39ULL, "mag110_ws_banter.cnv"},
	{0x6F00C3BC2E787269ULL, "6174513_m.fxe"},
	{0x6F013B73465A8519ULL, "dae_item_singletarget.ncs"},
	{0x6F01B250A956035BULL, "ser221_quentin6128860.cl"},
	{0x6F01BB605EF7E6EFULL, "dae100_bran6118314.cl"},
	{0x6F020A8F562C154CULL, "follower_banter6161234.cl"},
	{0x6F025FCAB36384FCULL, "6127025_f.fxe"},
	{0x6F02A275659D320DULL, "vxht_0_hightown_a2_anc_0_0.vlm"},
	{0x6F039F847ECADC0FULL, "vxht_keep_3f3_0_0.vlm"},
	{0x6F03E236CF23D148ULL, "6125557_f.fxe"},
	{0x6F04AC3DC6AB23C2ULL, "vxht_mansion2b_2ek_0_1.vlm"},
	{0x6F0556623C2E18BEULL, "one300cr_zevran.utc"},
	{0x6F0624A87B08C96AULL, "6004158_m.fxe"},
	{0x6F078F3EA02E8B5DULL, "6086898_m.fxe"},
	{0x6F0791B3F6110BB1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\blendtree\\trooper_overlay.anb"},
	{0x6F07DA3AB522BFE7ULL, "vxlt_undercityencounter_wd_0_0.vlm"},
	{0x6F07F8DA3C138B1CULL, "6165042_m.fxe"},
	{0x6F083475024B5DEBULL, "vxca_caverns1_mines_ei5_0_0.vlm"},
	{0x6F08383A7FA867E6ULL, "plt_cod_cha_anders2.plo"},
	{0x6F08DA75328DB529ULL, "vxlt_gallowstemplar_n_2os_0_0.vlm"},
	{0x6F08DFD5DA0D5FA3ULL, "dae000ip_crowd_mf4_talk3.utp"},
	{0x6F094D13F9EB0FE1ULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.anb"},
	{0x6F0975DF48C1CC80ULL, "vxlt_hangedman_n_sk_0_0.vlm"},
	{0x6F09BA9E4AB490A4ULL, "mag101_grace6034074.cl"},
	{0x6F09C25B80D39470ULL, "vxw_woundedcoast_da_0_0.vlm"},
	{0x6F09F633F20299C9ULL, "ca_caverns1_mines_24425_ll.mmh"},
	{0x6F0AEE1304C101D4ULL, "6125406_m.fxe"},
	{0x6F0B907C44DB0FF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_exit.ani"},
	{0x6F0BAAD2B00C1F8EULL, "vxlt_0_lowtown_a2_exf_0_0.vlm"},
	{0x6F0BFD6C32FF9275ULL, "vxht_0_hightown_n_a6l_0_0.vlm"},
	{0x6F0C67B50E19257AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_exit.evt"},
	{0x6F0CC27E64CE4080ULL, "zz_followers_banter6158724.cl"},
	{0x6F0D1A2A829410E6ULL, "6106201_m.fxe"},
	{0x6F0DA623FFEBB337ULL, "vxlt_warehouse_n_10r_0_0.vlm"},
	{0x6F0DD6ECC1AB35A9ULL, "qun230ip_to_wilderness.utp"},
	{0x6F0E297CA0FB1555ULL, "qun131_taskmaster6059004.cl"},
	{0x6F0EFFE01F00C9E7ULL, "art\\characters\\creatures\\dragon\\c_dragonb.phy"},
	{0x6F0F22E543C99915ULL, "vxht_0_hightown_a2_c1d_0_0.vlm"},
	{0x6F0F6C3B477D74F3ULL, "hlac_lt_undercityserialkiller_-1_14z.dds"},
	{0x6F0F8DF95FCC7E1BULL, "6088866_f.fxe"},
	{0x6F0FA5642F91027FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge_n.dds"},
	{0x6F108A1A5467670AULL, "hlac_w_sundermount_196610_0z.dds"},
	{0x6F11145257934A8DULL, "vxht_0_hightown_a2_aib_0_0.vlm"},
	{0x6F115D48E2F96852ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.dura.gad"},
	{0x6F1192D5EA9A827FULL, "vxht_keep_3zw_0_0.vlm"},
	{0x6F11AE25E51C2B8EULL, "vxlt_house_h2_0_0.vlm"},
	{0x6F125413C119E9C8ULL, "6122457_f.fxe"},
	{0x6F129B7C44E14093ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_exit.gad"},
	{0x6F12E12BF2AC0A4DULL, "hlac_w_sundermount_-196603_0x.dds"},
	{0x6F130F2890F6F1F3ULL, "docks_scene_dockworkers6176223.cl"},
	{0x6F1352CAA43ECE9AULL, "6165224_m.fxe"},
	{0x6F1366B50E1F426DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_exit.gad"},
	{0x6F13B661BA6257E6ULL, "839866015.ogg"},
	{0x6F13DBF0426F250FULL, "art\\cinematics\\act3\\romances\\isabella\\cs_pcmale_isarom.gad"},
	{0x6F140CAA24FE44B1ULL, "hlac_lt_backalley_g_65536_2z.dds"},
	{0x6F141A46145BABF1ULL, "dae150ip_at_scaverns_a.utp"},
	{0x6F141EC57E13AB1AULL, "one350_vinell6087347.cl"},
	{0x6F14663E5E400FB1ULL, "gen00fl_anders6119691.cl"},
	{0x6F14AB062B733118ULL, "mag291_demon_marethari6027961.cl"},
	{0x6F14DCD124B6DB1CULL, "314885029.ogg"},
	{0x6F14F05320C6BE74ULL, "llac_lt_0_lowtown_1_2.dds"},
	{0x6F158042944B7982ULL, "vxlt_0_lowtown_n_f1u_0_0.vlm"},
	{0x6F160A4D2D773DEEULL, "plt_dae000pt_amb_anywhere.plo"},
	{0x6F162E13F9F5655AULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.mmh"},
	{0x6F164C17A718D867ULL, "vxlt_undercityencounter_rg_0_0.vlm"},
	{0x6F17155EE06CC7D4ULL, "dr_dungeon_probe_23x.dds"},
	{0x6F176048E2FDD889ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.dura.evt"},
	{0x6F1823D399282F34ULL, "dae200_cullen6140267.cl"},
	{0x6F18343872C9ACF0ULL, "vxw_woundedcoast_an_0_0.vlm"},
	{0x6F184F0BF221ED6EULL, "vxca_caverns1_mines_i1p_0_0.vlm"},
	{0x6F18765AE518CBC8ULL, "517950447.ogg"},
	{0x6F1892BAB8FAA3F1ULL, "6155703_m.fxe"},
	{0x6F189A8BBC3194C7ULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gax_002a_d.dds"},
	{0x6F189FB42D42A9BDULL, "6165388_m.fxe"},
	{0x6F18C856BEC92B63ULL, "vxlt_blackemporium_3t_0_1.vlm"},
	{0x6F193B09916D5880ULL, "hlac_w_sundermount_65535_0y.dds"},
	{0x6F198463E4F1B226ULL, "ave200_donnic.cnv"},
	{0x6F19987C44E75AF8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_exit.evt"},
	{0x6F19A0D76ABB3272ULL, "gen00fl_aveline6137709.cl"},
	{0x6F19EEA39CD88080ULL, "ht_0_hightown_a2_17_0.rml"},
	{0x6F1A114AB2E71F9CULL, "mag311_emile_st.stg"},
	{0x6F1A1EEA8D197E70ULL, "gen00fl_aveline6115453.cl"},
	{0x6F1A2099BFA00034ULL, "399535860.ogg"},
	{0x6F1A3E0CBCE6B588ULL, "vxca_caverns1_mines_iyn_0_0.vlm"},
	{0x6F1AA330944F8B99ULL, "gen_im_npc_wep_dua_genb.uti"},
	{0x6F1B1324F57B9E11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keep.mmh"},
	{0x6F1B1F24F57BB2B2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keep.mao"},
	{0x6F1B2D24F57BCACFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keep.msh"},
	{0x6F1BAFEE36163751ULL, "mag100_cullen6095090.cl"},
	{0x6F1C7F5014B62EFEULL, "mag320ip_to_launcet.utp"},
	{0x6F1C8663E4F3EBE6ULL, "ave200_donnic.bnk"},
	{0x6F1D2F57C930DDC7ULL, "follower_banter6155526.cl"},
	{0x6F1D3FC29DDFFDD0ULL, "6142760_m.fxe"},
	{0x6F1D54E2C182EE89ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_diagonal.phy"},
	{0x6F1E15FDB46D7FFDULL, "dre000_dougal6129836.cl"},
	{0x6F1EC650A633446FULL, "923636332.ogg"},
	{0x6F1ECF7FAA4F5B4BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_templdesk_sa.dds"},
	{0x6F1EED0B3A3139F8ULL, "zz_followers_banter6157862.cl"},
	{0x6F1F410E1288C7B3ULL, "gen00fl_varric6118986.cl"},
	{0x6F1F77677906E5ABULL, "6055157_m.fxe"},
	{0x6F1FEC93C6E59A98ULL, "611662846.ogg"},
	{0x6F20219C420BB2F0ULL, "6121073_m.fxe"},
	{0x6F20F7E6007DC388ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_tmpa_0n.dds"},
	{0x6F20FD7060C07E70ULL, "751670268.ogg"},
	{0x6F214258CD6911F9ULL, "vxdr_primevaltomb_3hz_0_0.vlm"},
	{0x6F218157CA466D7FULL, "follower_banter6171270.cl"},
	{0x6F223E24DBD4AACAULL, "llac_ca_caverns2_5802_16.dds"},
	{0x6F2280E0326ABBAFULL, "293153914.ogg"},
	{0x6F22B4DDB0E3D8BEULL, "479774149.ogg"},
	{0x6F22DD4D72287358ULL, "6176125_m.fxe"},
	{0x6F22E191C6941305ULL, "vxlt_0_lowtown_a2_n_c_hd9_0_0.vlm"},
	{0x6F22F4E11E03AD36ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_6y.dds"},
	{0x6F233F6169C22B0FULL, "lt_0_lowtown_docks_35_0.rml"},
	{0x6F234E3EAFD94684ULL, "271579438.ogg"},
	{0x6F23F21F86419202ULL, "vxlt_0_lowtown_n_f4v_0_0.vlm"},
	{0x6F241748CF304C4AULL, "dae151ip_at_sundermount_b.utp"},
	{0x6F245A945BDCCB31ULL, "vxlt_0_lowtown_a2_hgm_0_0.vlm"},
	{0x6F2460E187EBA4EDULL, "hlac_lt_0_lowtown_a3_n_c_65534_5x.dds"},
	{0x6F24A0AD20F04430ULL, "w_woundedcoast_251_0.mao"},
	{0x6F24A4AD20F04B43ULL, "w_woundedcoast_251_0.mmh"},
	{0x6F24B2AD20F0629DULL, "w_woundedcoast_251_0.msh"},
	{0x6F25182EE6378EEBULL, "vxca_caverns1_mines_ec7_0_0.vlm"},
	{0x6F254848E309EDF8ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.dura.ani"},
	{0x6F256BCE8D426409ULL, "mer210_seldrek6078859.cl"},
	{0x6F25C30B90E56588ULL, "vxca_caverns1_mines_j77_0_0.vlm"},
	{0x6F25F9BD11BDA9B6ULL, "6068453_m.fxe"},
	{0x6F2621683F98F7EBULL, "hlac_ca_caverns1_24427_2y.dds"},
	{0x6F262310B18E7FA2ULL, "zz_mrl150ip_torchmarker.utp"},
	{0x6F26F7346F8D7847ULL, "378887053.ogg"},
	{0x6F27453A257CD720ULL, "638428809.ogg"},
	{0x6F27C795E7F3AF25ULL, "zz_approval_debug6003745.cl"},
	{0x6F27F8F0427FFFD5ULL, "art\\cinematics\\act3\\romances\\isabella\\cs_pcmale_isarom.ani"},
	{0x6F2834732BD94E05ULL, "471940738.ogg"},
	{0x6F28B55D58B87A18ULL, "lt_gallowscourtyard_n_46_0.rml"},
	{0x6F28E8F73C69F086ULL, "6167118_m.fxe"},
	{0x6F28F2B95205E214ULL, "680381071.ogg"},
	{0x6F28FD0FC76FB493ULL, "107534193.ogg"},
	{0x6F299F213AFD787EULL, "lowtown_male_drunk_mob6151171.cl"},
	{0x6F29FB4D2D87CD7DULL, "plt_dae000pt_amb_anywhere.nss"},
	{0x6F2A646CF4190A03ULL, "565382091.ogg"},
	{0x6F2A653152C79B1FULL, "zz_lgt_3_debug6176139.cl"},
	{0x6F2AFAB27A74EEE1ULL, "6099206_f.fxe"},
	{0x6F2B131E8F46CEFBULL, "744392653.ogg"},
	{0x6F2B9F25E3A4AD18ULL, "902286955.ogg"},
	{0x6F2BE0B18158F189ULL, "360061764.ogg"},
	{0x6F2C1B0901DAB64FULL, "643520241.ogg"},
	{0x6F2C1ECF665201BDULL, "mnd_con_cycle.mao"},
	{0x6F2CF19FB7B1448BULL, "hlac_ca_cavern1_haunted_24426_10z.dds"},
	{0x6F2D24B8D7F14D91ULL, "gen_im_arm_cht_lgt_act0_02.uti"},
	{0x6F2D3BCCDF02B7C6ULL, "lt_0_lowtown_19_0.rml"},
	{0x6F2DB7B9F609426BULL, "6085390_m.fxe"},
	{0x6F2DD7F966DB7D17ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc107a_0d.dds"},
	{0x6F2DFFD1511C98AEULL, "gen00fl_merrill6106238.cl"},
	{0x6F2E372DF8E97999ULL, "189375765.ogg"},
	{0x6F2EDBDA8501098FULL, "vxdr_primevaltomb_6py_0_0.vlm"},
	{0x6F2F215BA0A40E2EULL, "ca_caverns1_24424_lc.mmh"},
	{0x6F2F6BFC5B02261DULL, "follower_banter6166720.cl"},
	{0x6F2F8F6E6E170D91ULL, "hlac_ht_bartrand_n_3373_0z.dds"},
	{0x6F3007FF49B381ABULL, "framebuffereffects_linearfade_p0.psh.fxo.dx11.assembly"},
	{0x6F301479B4225273ULL, "635872251.ogg"},
	{0x6F3029C9B7691B24ULL, "w_encounter3_184_0.mmh"},
	{0x6F302DC9B7692277ULL, "w_encounter3_184_0.mao"},
	{0x6F303FC9B76940C6ULL, "w_encounter3_184_0.msh"},
	{0x6F3087406F7AFF87ULL, "follower_banter6172301.cl"},
	{0x6F317898F512397AULL, "gen00fl_fenris6115344.cl"},
	{0x6F31F5FBA8EFAE8EULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgtb_0t.dds"},
	{0x6F32A65EB6E1042AULL, "hlac_lt_0_lowtown_a2_65536_0z.dds"},
	{0x6F333AFC92F48985ULL, "459157493.ogg"},
	{0x6F3375292B3D8E0EULL, "llac_lt_backalley_g_65537_2.dds"},
	{0x6F33DD29127E6CA0ULL, "hlac_lt_0_lowtown_65538_4x.dds"},
	{0x6F3434E882DB7E08ULL, "vxca_caverns1_mines_j24_0_0.vlm"},
	{0x6F343B03E655A27BULL, "llac_lt_0_lowtown_docks_a3_n_65536_0.dds"},
	{0x6F3461226608F843ULL, "vxht_mansion2b_385_0_0.vlm"},
	{0x6F348EAE9F2D393BULL, "gen00fl_aveline6139168.cl"},
	{0x6F349ECCF9C46CFBULL, "hlac_lt_0_lowtown_docks_a3_-131074_0y.dds"},
	{0x6F34AEEB45EEF9F9ULL, "vxdr_primevaltomb_2vo_0_2.vlm"},
	{0x6F34BA077BEB4BCCULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkr.ani"},
	{0x6F34C2978D59D494ULL, "mag100_cullen6140944.cl"},
	{0x6F35CD90C4490768ULL, "vxlt_0_lowtown_eo1_0_0.vlm"},
	{0x6F3668BD50F68A35ULL, "6096820_m.fxe"},
	{0x6F37369A81943038ULL, "628594297.ogg"},
	{0x6F37B538606F731FULL, "one110_meeran6089867.cl"},
	{0x6F381ACA97CEE453ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_90_c.impa.evt"},
	{0x6F38351E2B57E370ULL, "vxw_bonepit_1wf_0_0.vlm"},
	{0x6F39911C1F359E58ULL, "lowtown_male_lowlife1_sta6164976.cl"},
	{0x6F3A8DE04076D36AULL, "vxht_keep_582_0_0.vlm"},
	{0x6F3B258CBEAB98CAULL, "6146861_m.fxe"},
	{0x6F3BC7077BF1800EULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkr.gad"},
	{0x6F3CCECC0AD3E74EULL, "6154152_m.fxe"},
	{0x6F3CD03513E21732ULL, "6130052_f.fxe"},
	{0x6F3D024ED5AD62F3ULL, "453915727.ogg"},
	{0x6F3D0D23EBB1E698ULL, "vxlt_0_lowtown_a2_n_c_eqh_0_2.vlm"},
	{0x6F3D215FD499073FULL, "988577388.ogg"},
	{0x6F3D4DB776F7220FULL, "vxdr_primevaltomb_6wx_0_0.vlm"},
	{0x6F3DD59A8895FE82ULL, "699411923.ogg"},
	{0x6F3DE129503267D7ULL, "vxca_caverns2prc_96k_0_0.vlm"},
	{0x6F3E3A575A38E7A8ULL, "vxdr_primevalentrance_273_0_0.vlm"},
	{0x6F3E8E069B515C46ULL, "vxdr_0_deeproads_ety_0_0.vlm"},
	{0x6F3E93E852CA84BEULL, "vxw_sundermount_2g8_0_0.vlm"},
	{0x6F3E9A5760FC434CULL, "vxlt_0_lowtown_e2r_0_0.vlm"},
	{0x6F3EA72BDABE3610ULL, "6172316_m.fxe"},
	{0x6F3EB43B1393195FULL, "hlac_lt_0_lowtown_131070_4x.dds"},
	{0x6F3EDC894A7CE713ULL, "montage_subs_demohack6217108.cl"},
	{0x6F3EE19FAE026992ULL, "vxdr_primevaltomb_7p1_0_0.vlm"},
	{0x6F3EEC230F308034ULL, "hlac_dr_primevaltomb_659_1y.dds"},
	{0x6F3F13CA97D4F694ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_90_c.impa.gad"},
	{0x6F3FDB3EB31D46E6ULL, "vxlt_gallowstemplar_475_0_0.vlm"},
	{0x6F3FF583D976C345ULL, "6127983_f.fxe"},
	{0x6F401F906141944DULL, "llac_ht_chantry_n_2534_52.dds"},
	{0x6F40530D0C39B3A3ULL, "522976655.ogg"},
	{0x6F40A13302F8D365ULL, "vxht_0_hightown_a2_c14_0_0.vlm"},
	{0x6F40AF3C866CC0B7ULL, "qun221_petrice6083817.cl"},
	{0x6F413E4E3FC8FC31ULL, "vxlt_0_lowtown_a2_hic_0_0.vlm"},
	{0x6F4158C51082D2EAULL, "782373080.ogg"},
	{0x6F418317E761C1D7ULL, "vxdr_primevaltomb_8or_0_0.vlm"},
	{0x6F4195C2F67C31F2ULL, "6089496_m.fxe"},
	{0x6F41C2077BF5E3EDULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkr.evt"},
	{0x6F41F9D80CD7C327ULL, "gen00fl_aveline6139667.cl"},
	{0x6F420A80EF489322ULL, "vxdr_0_deeproads_cf5_0_0.vlm"},
	{0x6F42CDA2D83903A6ULL, "196904728.ogg"},
	{0x6F42DCCEFB9FDA4DULL, "llac_w_encounter2_327678_0.dds"},
	{0x6F433B18602D88EBULL, "vxht_mansion_n_53y_0_0.vlm"},
	{0x6F4442B1414D6C8DULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw41_raillow.mmh"},
	{0x6F4446B1414D73DEULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw41_raillow.mao"},
	{0x6F4454B1414D8B13ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw41_raillow.msh"},
	{0x6F446339E6FCEC26ULL, "hlac_lt_0_lowtown_a2_-131071_1z.dds"},
	{0x6F44916016C30E0DULL, "477856057.ogg"},
	{0x6F44ABA91E69E67DULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpb.phy"},
	{0x6F45289A5430E7F0ULL, "6134985_m.fxe"},
	{0x6F460AA8ECF4DD33ULL, "vxht_keep_3yj_0_0.vlm"},
	{0x6F4622CA97DB2F72ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_90_c.impa.ani"},
	{0x6F46A6FB1D4DFBF0ULL, "vxw_bonepit_1xg_0_0.vlm"},
	{0x6F46D833CBC00AF4ULL, "mrl210_merrill6128872.cl"},
	{0x6F474F706B236084ULL, "mag131_keran6086200.cl"},
	{0x6F47D31B0E12648EULL, "vxht_keep_n_6bp_0_0.vlm"},
	{0x6F483CBBE3397F80ULL, "plt_lgt000pt_ultra_3_02.plo"},
	{0x6F49376A2485AFB3ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_frnt.evt"},
	{0x6F493DE2C1A802AEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_diagonal.mmh"},
	{0x6F4942860E4E3BBEULL, "hlac_ht_0_hightown_n_1_0y.dds"},
	{0x6F4949E2C1A81745ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_diagonal.mao"},
	{0x6F4957E2C1A82E9CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_diagonal.msh"},
	{0x6F4964FAEB46D056ULL, "6077347_m.fxe"},
	{0x6F498EFFC758C4A6ULL, "6053200_m.fxe"},
	{0x6F49C0607A3AF2EDULL, "vxw_bonepitmining_a2_1wp_0_1.vlm"},
	{0x6F4A39E598210BAEULL, "157717745.ogg"},
	{0x6F4A4DD602FFD56AULL, "vxht_0_hightown_a3_aup_0_0.vlm"},
	{0x6F4A69B731076BB4ULL, "6098917_m.fxe"},
	{0x6F4A7D4613843297ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv4.gad"},
	{0x6F4ACA2BE3177845ULL, "6109034_m.fxe"},
	{0x6F4B1A01234C1C43ULL, "sky_sky2_p2.psh.fxo.dx11"},
	{0x6F4B1E3C7E33987DULL, "6107151_m.fxe"},
	{0x6F4B4421F30EB0BBULL, "6147157_m.fxe"},
	{0x6F4B46BEE7FD8C2FULL, "vxca_caverns2prc_9qv_0_0.vlm"},
	{0x6F4B7C4E2D32FCF9ULL, "ser270st_emeric.stg"},
	{0x6F4C1B220B629CE2ULL, "15043915.ogg"},
	{0x6F4C1D81473004C7ULL, "hlac_ht_0_hightown_n_-65535_1y.dds"},
	{0x6F4C40696E9BD2CAULL, "672283981.ogg"},
	{0x6F4C4B4364EFF3F7ULL, "715010850.ogg"},
	{0x6F4C663B129790BCULL, "mcr370_meredith6083604.cl"},
	{0x6F4CB70267C2C02EULL, "805346677.ogg"},
	{0x6F4CFEF266A88954ULL, "6067797_m.fxe"},
	{0x6F4D05C544C09D3EULL, "vxw_sundermount_2h9_0_0.vlm"},
	{0x6F4D1EAF75DA2A32ULL, "hlac_w_woundedcoast_-65528_0x.dds"},
	{0x6F4D81B816D69F7DULL, "659041378.ogg"},
	{0x6F4D951F67D4568EULL, "218350043.ogg"},
	{0x6F4D9B749695E318ULL, "gen00fl_varric6021135.cl"},
	{0x6F4DAB41C437DA84ULL, "290024924.ogg"},
	{0x6F4DD2CE65647F4EULL, "vxdr_primevaltomb_6p4_0_0.vlm"},
	{0x6F4E3CCB9F370EA2ULL, "484822143.ogg"},
	{0x6F4E5D447B4F70D9ULL, "6018901_m.fxe"},
	{0x6F4EC9EE300149CFULL, "68312311.ogg"},
	{0x6F4F2ECBA33042C0ULL, "lgt311_shw_mer_mettin.cnv"},
	{0x6F4F6119DCDF8A82ULL, "mcr370_templar16087624.cl"},
	{0x6F4F9BEF05946AC3ULL, "152181836.ogg"},
	{0x6F4FAC485E3CF8E3ULL, "6051607_m.fxe"},
	{0x6F4FF5937DCB27B3ULL, "mrl150_dalish_guard6089410.cl"},
	{0x6F50306A248BC2F4ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_frnt.gad"},
	{0x6F503E2CE2E6AB00ULL, "vxht_mansion_n_5be_0_0.vlm"},
	{0x6F50BBD17761C6B3ULL, "6118188_m.fxe"},
	{0x6F50DF37646236F8ULL, "83996386.ogg"},
	{0x6F514C6BED97F47FULL, "w_woundedcoast_125_0.msh"},
	{0x6F51526BED97FE21ULL, "w_woundedcoast_125_0.mmh"},
	{0x6F515E6BED9812C2ULL, "w_woundedcoast_125_0.mao"},
	{0x6F519246138A7524ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv4.evt"},
	{0x6F5210CBA33245F8ULL, "lgt311_shw_mer_mettin.bnk"},
	{0x6F521E6F91092D2BULL, "6099320_m.fxe"},
	{0x6F52938EDBB76F4BULL, "799484132.ogg"},
	{0x6F534505B6D803C6ULL, "48779849.ogg"},
	{0x6F53BDC2F9660373ULL, "261647409.ogg"},
	{0x6F5402327F9B218FULL, "ht_mansion_n_9156_ll.mmh"},
	{0x6F540ADAB5D16C86ULL, "vxlt_0_lowtown_docks_a3_b9d_0_0.vlm"},
	{0x6F54160F9A2A6707ULL, "6077760_m.fxe"},
	{0x6F5424DF83845576ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha14a.mmh"},
	{0x6F542EDF83846674ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha14a.msh"},
	{0x6F543C9D4AF37E20ULL, "qun110_petrice6018545.cl"},
	{0x6F5451460670C7F2ULL, "vxlt_0_lowtown_a2_bz3_0_0.vlm"},
	{0x6F54712A0ADD36A5ULL, "vxht_keep_73f_0_0.vlm"},
	{0x6F55ACA91E785A82ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpb.mmh"},
	{0x6F55AEA91E785E38ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpb.msh"},
	{0x6F55F6A662AD3457ULL, "6034927_m.fxe"},
	{0x6F55FA3C967D1B91ULL, "159946951.ogg"},
	{0x6F561F2EF1922D52ULL, "6076906_m.fxe"},
	{0x6F56F14D8D8AE03CULL, "182566233.ogg"},
	{0x6F573F6A2491FAD2ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_frnt.ani"},
	{0x6F579A66A646539BULL, "vxca_caverns1_jg5_0_0.vlm"},
	{0x6F584F442EF4C233ULL, "174897205.ogg"},
	{0x6F58872E58D242E6ULL, "gen00fl_aveline6107012.cl"},
	{0x6F589CABC8178309ULL, "6073890_m.fxe"},
	{0x6F58B2D28A956434ULL, "zz_isa_debug6175289.cl"},
	{0x6F58F4B9434828F8ULL, "hlac_dr_dungeon_3214_10y.dds"},
	{0x6F5915923123C42EULL, "vxlt_0_lowtown_a2_n_c_hln_0_0.vlm"},
	{0x6F5994A02FCB05F3ULL, "llac_ht_bartrand_n_3369_1.dds"},
	{0x6F59A20C9C878EBDULL, "858195995.ogg"},
	{0x6F5A1ACB311C9991ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.dura.ani"},
	{0x6F5A67ABADFE6480ULL, "ht_0_hightown_a3_35_0_lc.mmh"},
	{0x6F5C44AB575A97CEULL, "vxdr_primevaltomb_6w5_0_0.vlm"},
	{0x6F5CD93DBD62F445ULL, "6129783_f.fxe"},
	{0x6F5CE79534EF2A59ULL, "6209872_f.fxe"},
	{0x6F5CF05B2041AB75ULL, "252528274.ogg"},
	{0x6F5DEA1412EE6998ULL, "260181912.ogg"},
	{0x6F5EAA461395281DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv4.ani"},
	{0x6F5EDEA295D04C55ULL, "hla_ht_0_hightown_-65537_0z.dds"},
	{0x6F5F5494035BEAFFULL, "mag121_idunna6083975.cl"},
	{0x6F600721A0651512ULL, "ss_fadeswordshield.cut"},
	{0x6F601C89B1E3CF58ULL, "505058571.ogg"},
	{0x6F60979EDF9D372EULL, "character_character_p45.psh.fxo.dx11"},
	{0x6F610543CBB59623ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagkeep_sa.dds"},
	{0x6F6125CB3122CB33ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.dura.gad"},
	{0x6F619AAFDF156E37ULL, "vxw_bonepit_98_0_0.vlm"},
	{0x6F61BCD533B8C399ULL, "1055773180.ogg"},
	{0x6F61C5263F9177DCULL, "vxlt_hangedman_n_t6_0_0.vlm"},
	{0x6F628D9CE5AA8D72ULL, "vxlt_warehouse_n_11x_0_0.vlm"},
	{0x6F6293035A69565EULL, "842248136.ogg"},
	{0x6F6321F274706524ULL, "891670850.ogg"},
	{0x6F63887FD0AC663DULL, "zz_dre_debug6168314.cl"},
	{0x6F6392831D3E1891ULL, "art\\characters\\creatures\\dragon\\animation\\tal_fire.ani"},
	{0x6F6438EE9FA540C6ULL, "and110_refugee6111613.cl"},
	{0x6F64F2E894114064ULL, "gen00fl_aveline6132821.cl"},
	{0x6F65924AE1D211D6ULL, "zz_approval_debug6003843.cl"},
	{0x6F65B4CBBC7C5015ULL, "vxlt_backalley_g_1nw_0_1.vlm"},
	{0x6F65B5CC554F2F31ULL, "prty_isabela_armor_1.dds"},
	{0x6F661578522AF846ULL, "vxlt_warehouse_148_0_0.vlm"},
	{0x6F6822CB3128E498ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.dura.evt"},
	{0x6F688A1DD957D907ULL, "187655180.ogg"},
	{0x6F68A5B7E8597C9BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brk.mmh"},
	{0x6F68AE8C399E36A9ULL, "vxlt_0_lowtown_n_exn_0_0.vlm"},
	{0x6F68B1B7E8599138ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brk.mao"},
	{0x6F68BA615256D6E5ULL, "vxca_caverns2_71c_0_0.vlm"},
	{0x6F68C3B7E859AF85ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brk.msh"},
	{0x6F68E29706CBE384ULL, "sky.mat"},
	{0x6F693F18D577D7AEULL, "mag121_idunna6084223.cl"},
	{0x6F6941AC4D30D0B6ULL, "6136095_player_toolate.cut"},
	{0x6F695F2EC951291AULL, "206212018.ogg"},
	{0x6F6998E7660A4EA8ULL, "hlac_ca_caverns1_mines_24426_13z.dds"},
	{0x6F69BF24A28C0F22ULL, "vxht_chantry_2t5_0_0.vlm"},
	{0x6F69D532F0DB73DCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoortrim_d.dds"},
	{0x6F6A2D76E450BC69ULL, "vxht_0_hightown_n_9zt_0_0.vlm"},
	{0x6F6A63214E86D200ULL, "vxlt_hangedman_1qe_0_0.vlm"},
	{0x6F6A9D831D444A33ULL, "art\\characters\\creatures\\dragon\\animation\\tal_fire.gad"},
	{0x6F6AB6884950B04EULL, "vxdr_primevaltomb_6z6_0_0.vlm"},
	{0x6F6ADE4926C703CDULL, "494153944.ogg"},
	{0x6F6B4EDEDC3E5EB1ULL, "hlac_ht_mansion2b_818_10x.dds"},
	{0x6F6BCDE056C23979ULL, "153034418.ogg"},
	{0x6F6D091674515DF8ULL, "dae000_arishok6080491.cl"},
	{0x6F6D21E6C6D2DC00ULL, "vxht_mansion_n_5hc_0_0.vlm"},
	{0x6F6EB7C592A933FBULL, "487927509.ogg"},
	{0x6F6F101A38395EE5ULL, "50927387.ogg"},
	{0x6F6F27D06AEF8972ULL, "615976156.ogg"},
	{0x6F6FBBDC7552B8FFULL, "mag311_nyssa6042527.cl"},
	{0x6F7037033187905CULL, "vxlt_hangedman_n_s7_0_0.vlm"},
	{0x6F704CADA8DC88D5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxa.phy"},
	{0x6F7057EC88557C5DULL, "6132783_f.fxe"},
	{0x6F70CA49C4BCD5F1ULL, "vxx_narrativeroom_3k_0_0.vlm"},
	{0x6F70E3DFFDDB5140ULL, "796405593.ogg"},
	{0x6F7148FEC67CFE28ULL, "llac_lt_0_lowtown_a2_n_c_1_9.dds"},
	{0x6F719A831D4A6398ULL, "art\\characters\\creatures\\dragon\\animation\\tal_fire.evt"},
	{0x6F723D8A0364E018ULL, "6103666_m.fxe"},
	{0x6F72B9ACBED49221ULL, "503840681.ogg"},
	{0x6F72C57ED387729DULL, "vxlt_darktown_48g_0_0.vlm"},
	{0x6F7362A928E4E867ULL, "l_dae120ar_hightown_nt.bnk"},
	{0x6F740BBB201385D0ULL, "175260777.ogg"},
	{0x6F747ECA2AE5AF36ULL, "6146955_m.fxe"},
	{0x6F74D92A055704E0ULL, "6140977_m.fxe"},
	{0x6F74FBBF66ACFC6FULL, "mag360_thrask6043532.cl"},
	{0x6F7525089E93F264ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_dead.gad"},
	{0x6F76916AD4DD70BFULL, "qun211cr_braggart_ally_m3.utc"},
	{0x6F769C55BDDFD5AFULL, "vxca_caverns2prc_9fs_0_0.vlm"},
	{0x6F76FB05A538233DULL, "hlac_lt_0_lowtown_-65534_0z.dds"},
	{0x6F7727C26B1FCC79ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand.msh"},
	{0x6F772D726E48A7C8ULL, "hlac_w_bonepitmining_a2_262144_0y.dds"},
	{0x6F7731C26B1FDD67ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand.mmh"},
	{0x6F7735C26B1FE3B4ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand.mao"},
	{0x6F779EBFD76A6BF2ULL, "zz_followers_banter6158454.cl"},
	{0x6F77B23BB3376078ULL, "vxht_0_hightown_a3_9zc_0_0.vlm"},
	{0x6F77C338E93DB1DAULL, "hlac_lt_0_lowtown_n_131071_1z.dds"},
	{0x6F77E42B378B54ADULL, "vxdr_primevaltomb_7xs_0_1.vlm"},
	{0x6F7803A465E48960ULL, "hlac_lt_0_lowtown_a2_n_-3_2z.dds"},
	{0x6F7862134A310138ULL, "follower_banter6149229.cl"},
	{0x6F78818A19557214ULL, "6156879_m.fxe"},
	{0x6F78F2F4B0C16EF4ULL, "865941263.ogg"},
	{0x6F78FB3F85EFE394ULL, "6027380_m.fxe"},
	{0x6F7978E5588D2814ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen2.ani"},
	{0x6F79F51DEA86F7E2ULL, "mrl310_merrill6143971.cl"},
	{0x6F7A25E72F12B41BULL, "vxw_encounter2_camp_o6_0_0.vlm"},
	{0x6F7A3033C0ED56DEULL, "6139653_f.fxe"},
	{0x6F7A4280EAB93424ULL, "art\\characters\\playercharacter\\humanfemale\\arm_cas.mao"},
	{0x6F7A43394A176B53ULL, "68852475.ogg"},
	{0x6F7A4DE41FAE0C5EULL, "cod_shw_mer_notes_3.utp"},
	{0x6F7A8ABD25E40D2DULL, "mcr370_messenger6071825.cl"},
	{0x6F7B1A4C71030D71ULL, "star_blue.dds"},
	{0x6F7B35AD3E3E8FAEULL, "6093590_m.fxe"},
	{0x6F7B9954E54CB629ULL, "390452802.ogg"},
	{0x6F7BA270B902CC93ULL, "fen110_party_comment6095234.cl"},
	{0x6F7C34089E9A2A42ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_dead.ani"},
	{0x6F7C74CFEE8C969AULL, "vxlt_gallowstemplar_5v3_0_0.vlm"},
	{0x6F7CDE9CDF019FEEULL, "hm_qsw_warrens7.mor"},
	{0x6F7D98166258D41FULL, "hlac_lt_0_lowtown_a2_n_0_7z.dds"},
	{0x6F7DBA7CCF29436EULL, "6124292_m.fxe"},
	{0x6F7E469CCC8A8A14ULL, "6110569_m.fxe"},
	{0x6F7EC2BAA5BC991CULL, "gen00fl_fenris6126689.cl"},
	{0x6F7ED7D04B92790EULL, "lt_gallowstemplar_f_2382.rml"},
	{0x6F7F0233091FD16DULL, "vxht_keep_7a8_0_0.vlm"},
	{0x6F7F10D54F9CA81CULL, "6165235_m.fxe"},
	{0x6F7F3AB3982B12E0ULL, "vxw_woundedcoast_1ce_0_0.vlm"},
	{0x6F7F4A61DE7A5732ULL, "6135373_m.fxe"},
	{0x6F800B8A851D2CDDULL, "vxlt_0_lowtown_d7q_0_0.vlm"},
	{0x6F8075E558934276ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen2.gad"},
	{0x6F80AF66F55AF898ULL, "6104766_m.fxe"},
	{0x6F817EA5C03D4AF6ULL, "281337037.ogg"},
	{0x6F819306D02D2755ULL, "vxlt_0_lowtown_a3_n_c_gja_0_0.vlm"},
	{0x6F81994014B09CDCULL, "ach_kill_meredith.dds"},
	{0x6F81D34134AAAF2DULL, "llac_ca_cavern1_haunted_24425_6.dds"},
	{0x6F8229F0701C7075ULL, "716754241.ogg"},
	{0x6F82A154CECE16CAULL, "isa311_isabela6092269.cl"},
	{0x6F82B16CBFED889BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirel.mmh"},
	{0x6F82BD6CBFED9D38ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirel.mao"},
	{0x6F82CF6CBFEDBB85ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallospirel.msh"},
	{0x6F836957E00D403CULL, "fonts_ru-ru.gfx"},
	{0x6F838F8DB6655EA3ULL, "vxca_caverns1_mines_jg8_0_0.vlm"},
	{0x6F83C4A8ED52D733ULL, "zz_dae_debug6036580.cl"},
	{0x6F84382759CA2302ULL, "gen00fl_anders6121172.cl"},
	{0x6F845DBFBF90695AULL, "lt_0_lowtown_a2_n_25_0.rml"},
	{0x6F845DEA076793F0ULL, "mag100_thrask6031987.cl"},
	{0x6F850DA1435F6D92ULL, "6083586_m.fxe"},
	{0x6F85B080A7FCC0F1ULL, "gencr_gate_slave_r4.utc"},
	{0x6F863492A4299E36ULL, "403482959.ogg"},
	{0x6F86864BCD3CE749ULL, "zz_followers_banter6158136.cl"},
	{0x6F869EC7D50D9590ULL, "one100_solivitus6059914.cl"},
	{0x6F86D177193D1491ULL, "vxlt_undercityencounter_ut_0_0.vlm"},
	{0x6F86D7DF2F28B5F1ULL, "vxlt_gallowscourtyard_6xs_0_0.vlm"},
	{0x6F87418DCB295E8BULL, "vxht_0_hightown_a7z_0_0.vlm"},
	{0x6F8790E558998ED5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen2.evt"},
	{0x6F87C00874A042CCULL, "172933542.ogg"},
	{0x6F88515C20FB4E89ULL, "vxht_mansion2_25u_0_0.vlm"},
	{0x6F887B48BC07CC2EULL, "6071845_f.fxe"},
	{0x6F8891FDBE5AE65CULL, "vxlt_0_lowtown_fit_0_0.vlm"},
	{0x6F88CDBDA14111B0ULL, "vxca_caverns2_9fe_0_0.vlm"},
	{0x6F89007319C25659ULL, "hlac_ca_caverns2_5802_6x.dds"},
	{0x6F891F5663A944BFULL, "vrc221_varric6040865.cl"},
	{0x6F8938F2F4F3C831ULL, "vxca_caverns2_836_0_0.vlm"},
	{0x6F897982188EA97CULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkb.ani"},
	{0x6F89849764D8CF88ULL, "vxlt_0_lowtown_docks_n_art_0_0.vlm"},
	{0x6F898F2A4BF2F818ULL, "hlac_lt_0_lowtown_a2_-196608_2z.dds"},
	{0x6F89BC5603024734ULL, "vxca_caverns2_7qv_0_0.vlm"},
	{0x6F89FC6E30666238ULL, "111103711.ogg"},
	{0x6F8A4A07A13719A2ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheel.phy"},
	{0x6F8A877D53BE086DULL, "vxlt_0_lowtown_docks_n_brz_0_0.vlm"},
	{0x6F8AA887EA54D7DBULL, "lgt250_way_dalish_conflict6147167.cl"},
	{0x6F8AAE4AD1880FC8ULL, "205043222.ogg"},
	{0x6F8AD55ABD4E89C8ULL, "6173598_f.fxe"},
	{0x6F8BBCEDEFD29130ULL, "637364498.ogg"},
	{0x6F8BE3B75826AD07ULL, "312313242.ogg"},
	{0x6F8C98364F3656B6ULL, "vxca_caverns2_70w_0_0.vlm"},
	{0x6F8DAC908A212B60ULL, "vxw_woundedcoast_1nb_0_0.vlm"},
	{0x6F8E4239D5301DA1ULL, "649423760.ogg"},
	{0x6F8EB481875E7FF2ULL, "gencr_circle_highmage.utc"},
	{0x6F8F3916A32796B8ULL, "696167645.ogg"},
	{0x6F8FE84BB4CFF877ULL, "w_bonepitdestroyed_a3_117_0.msh"},
	{0x6F8FFA4BB4D016DAULL, "w_bonepitdestroyed_a3_117_0.mao"},
	{0x6F8FFE4BB4D01DE9ULL, "w_bonepitdestroyed_a3_117_0.mmh"},
	{0x6F9072A1252B10A5ULL, "vxlt_hangedman_n_ny_0_0.vlm"},
	{0x6F9086821894DDBEULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkb.gad"},
	{0x6F9087FED5DB7AE9ULL, "215142581.ogg"},
	{0x6F90E5AE44A02F26ULL, "6180966_m.fxe"},
	{0x6F90F2A43A176518ULL, "follower_banter6172779.cl"},
	{0x6F91122206363FD9ULL, "vxca_caverns1_mines_jfu_0_0.vlm"},
	{0x6F915DF45D59F783ULL, "vxht_mansion2b_6t_0_0.vlm"},
	{0x6F9195C26D791E06ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobe_0t.dds"},
	{0x6F91B6619EB12AE4ULL, "528424085.ogg"},
	{0x6F9257B9DB887D82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkback.gad"},
	{0x6F92A112D9019181ULL, "vxlt_hangedman_n_sf_0_0.vlm"},
	{0x6F93089BAD0B43F8ULL, "one280_emrys6104293.cl"},
	{0x6F930C675CFA4B4AULL, "6092148_m.fxe"},
	{0x6F9319DF4A4A7E56ULL, "6156514_m.fxe"},
	{0x6F93615FFE472549ULL, "static_static_p17.psh.fxo"},
	{0x6F93D25F128EE8EAULL, "hlac_w_sundermount_327684_0y.dds"},
	{0x6F93D53695520AC3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_mugpewter.mao"},
	{0x6F93D93695521110ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_mugpewter.mmh"},
	{0x6F93E7369552292AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_mugpewter.msh"},
	{0x6F9419728FA6BE52ULL, "6165336_m.fxe"},
	{0x6F943FEF18980329ULL, "lt_0_lowtown_docks_a3_29_0_ll.mmh"},
	{0x6F9443EDEF15CB01ULL, "339570240.ogg"},
	{0x6F956201824990FEULL, "433768182.ogg"},
	{0x6F95C827245CA620ULL, "hlac_lt_darktown_-1_79x.dds"},
	{0x6F96651750E0C0D0ULL, "hlac_ca_caverns1_mines_24422_2z.dds"},
	{0x6F96D7C660466138ULL, "650874330.ogg"},
	{0x6F96E6324043DD6AULL, "llac_w_bonepitdestroyed_a3_0_8.dds"},
	{0x6F976182189ABDDDULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\walkb.evt"},
	{0x6F97737ED87CB72EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basicwood.msh"},
	{0x6F97817ED87CCF7FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basicwood.mao"},
	{0x6F978D7ED87CE31CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basicwood.mmh"},
	{0x6F97A815702180E5ULL, "hlac_lt_0_lowtown_docks_n_-4_7z.dds"},
	{0x6F982E32F4F85FB4ULL, "vxca_caverns2_7vw_0_0.vlm"},
	{0x6F995FF928C71374ULL, "6129730_m.fxe"},
	{0x6F997AB9DB8ED879ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkback.evt"},
	{0x6F99B2EC173B8F46ULL, "6178889_f.fxe"},
	{0x6F99C10A7AA64D14ULL, "qun221_viscount_seamus_dead.stg"},
	{0x6F9A3C0D5B9FBF72ULL, "369151511.ogg"},
	{0x6F9B06A3A15F8373ULL, "849994905.ogg"},
	{0x6F9B316E7C5F65DBULL, "follower_barks6153664.cl"},
	{0x6F9B71CE98E622C5ULL, "107043279.ogg"},
	{0x6F9B9F6A2C6FAFFFULL, "vxw_bonepit_md_0_0.vlm"},
	{0x6F9BED4FADF9DE30ULL, "6045568_m.fxe"},
	{0x6F9C1E6D7C1743E0ULL, "vxw_woundedcoast_1ic_0_0.vlm"},
	{0x6F9C3D9C44C8B7E0ULL, "zz_followers_banter6158319.cl"},
	{0x6F9C4E583D737128ULL, "gen00fl_anders6125167.cl"},
	{0x6F9C918CDD26A235ULL, "hm_amb_keep_garol.mor"},
	{0x6F9CD83E72935A8EULL, "vxlt_gallowstemplar_398_0_0.vlm"},
	{0x6F9D8652AF535A24ULL, "401604058.ogg"},
	{0x6F9DC3279E88D3EBULL, "vxw_encounter2_camp_ra_0_0.vlm"},
	{0x6F9DE72D87033148ULL, "w_bonepit_208_0.msh"},
	{0x6F9DF92D87034FD9ULL, "w_bonepit_208_0.mao"},
	{0x6F9E052D87036472ULL, "w_bonepit_208_0.mmh"},
	{0x6F9ECE0D5347AEAAULL, "693718264.ogg"},
	{0x6F9F58A9290A119DULL, "l_dae120ar_hightown_nt.wbd"},
	{0x6F9F9586B7668E15ULL, "vxht_0_hightown_a2_ccj_0_0.vlm"},
	{0x6FA029BBE383D50FULL, "plt_lgt000pt_ultra_3_02.nss"},
	{0x6FA0776394456A88ULL, "vxlt_0_lowtown_a2_n_ekj_0_0.vlm"},
	{0x6FA0805B93A1AEA6ULL, "hlac_lt_0_lowtown_n_-65537_2y.dds"},
	{0x6FA14EE2EAF23EAEULL, "vxlt_0_lowtown_a2_n_c_hsc_0_0.vlm"},
	{0x6FA1745760203A3AULL, "vxlt_gallowscourtyard_6us_0_0.vlm"},
	{0x6FA1BC9DBDD2B9FBULL, "pro000_refugee_male6052115.cl"},
	{0x6FA1D33CF03D3DC9ULL, "static_static_p44.psh.fxo"},
	{0x6FA1F27E72CCB525ULL, "6202148_m.fxe"},
	{0x6FA1FB11029402F6ULL, "694214871.ogg"},
	{0x6FA2D0147388A5B2ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml103a_0d.dds"},
	{0x6FA2E621A6AB31ADULL, "333784827.ogg"},
	{0x6FA304C428B64200ULL, "6105656_f.fxe"},
	{0x6FA3204F4C5EAD74ULL, "1025602789.ogg"},
	{0x6FA37CCD8ECD2C83ULL, "vxht_0_hightown_a3_bgp_0_0.vlm"},
	{0x6FA386882E60143CULL, "279292577.ogg"},
	{0x6FA43CAAEA030468ULL, "hm_gamlen_h2s3.mor"},
	{0x6FA476532652234CULL, "391911934.ogg"},
	{0x6FA4E1CFD5071E54ULL, "vxht_mansion_n_6t8_0_0.vlm"},
	{0x6FA4F06DF4409E74ULL, "vxlt_hangedman_1yh_0_1.vlm"},
	{0x6FA51C9DE9CB5FA6ULL, "6143376_m.fxe"},
	{0x6FA56D6E0D754DCEULL, "vxlt_0_lowtown_n_f0v_0_0.vlm"},
	{0x6FA58812A5C8AD35ULL, "isa000pt_gift_talisman.ncs"},
	{0x6FA5A9351395911BULL, "vxdr_slavers_12t_0_0.vlm"},
	{0x6FA617097C74A89BULL, "vxht_0_hightown_n_aoy_0_0.vlm"},
	{0x6FA63F809DB954F4ULL, "81089884.ogg"},
	{0x6FA6A00FE6EE7834ULL, "vxca_caverns2_7kt_0_0.vlm"},
	{0x6FA6DAA8D2364E56ULL, "lt_0_lowtown_docks_20_0_lc.mmh"},
	{0x6FA72077ECE7B2EDULL, "zz_followers_banter6158244.cl"},
	{0x6FA782B9DB9B2368ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkback.ani"},
	{0x6FA88E39391ABDFFULL, "242784701.ogg"},
	{0x6FA899ED846DBC3FULL, "isa211_isabela6072551.cl"},
	{0x6FA9BCC3F9D73F95ULL, "6172285_m.fxe"},
	{0x6FA9EEF480FC627BULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeloop.evt"},
	{0x6FAA05E665BE31D9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_visa_0n.dds"},
	{0x6FAA3EB00017170DULL, "ts_qunari_arm_table.gda"},
	{0x6FAAAC018D77B655ULL, "6027554_m.fxe"},
	{0x6FAAB02248E93998ULL, "genip_hammer.utp"},
	{0x6FAB3954F7E1D01FULL, "hlac_lt_foundry_n_320_4z.dds"},
	{0x6FAB560FE7C17662ULL, "llac_lt_0_lowtown_docks_a3_65534_0.dds"},
	{0x6FAB5FE7CCCD71AEULL, "554091961.ogg"},
	{0x6FABC33BFD22E8EBULL, "6151477_m.fxe"},
	{0x6FAD017EFD048590ULL, "aip_follower_carver.gda"},
	{0x6FAD0285963DCEA0ULL, "551747506.ogg"},
	{0x6FAD4DA402BDDAA5ULL, "vxw_encounter1_ak_0_0.vlm"},
	{0x6FAD642E65587411ULL, "191251217.ogg"},
	{0x6FADF12698E54AADULL, "follower_banter6170385.cl"},
	{0x6FAE35621892EC3BULL, "494228509.ogg"},
	{0x6FAE6E96423E775CULL, "6027050_m.fxe"},
	{0x6FAFAFDA2A14B0D5ULL, "6122081_f.fxe"},
	{0x6FAFC6B7E895B7D0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brk.phy"},
	{0x6FB00233EEE6E48BULL, "zz_ser_debug6078159.cl"},
	{0x6FB018915277F03AULL, "149746179.ogg"},
	{0x6FB01B2294DA2CADULL, "6149346_m.fxe"},
	{0x6FB02507A157AA56ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheel.mao"},
	{0x6FB03107A157BEF5ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheel.mmh"},
	{0x6FB03307A157C26BULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_bigwheel.msh"},
	{0x6FB0AB2A3791E6D0ULL, "qcr220_cut_meredith_arrival6041456.cl"},
	{0x6FB0C2FB40BD0A6FULL, "734678413.ogg"},
	{0x6FB0E7F4810274BCULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeloop.gad"},
	{0x6FB11325F2430428ULL, "6065276_m.fxe"},
	{0x6FB1C198C3C14682ULL, "568337770.ogg"},
	{0x6FB2ABC0831ACDB8ULL, "6162955_m.fxe"},
	{0x6FB2BE0DD9E22BBAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_tarp_n.dds"},
	{0x6FB2D7972E23D928ULL, "vxca_cavern1_haunted_k2q_0_0.vlm"},
	{0x6FB2F77CA1F5797AULL, "669024654.ogg"},
	{0x6FB34B7E21140983ULL, "229471806.ogg"},
	{0x6FB3DAF8E45FC55EULL, "790867034.ogg"},
	{0x6FB40B2DAD5D0CCDULL, "hlac_w_encounter2_camp_131073_0z.dds"},
	{0x6FB42425A7B8CCDAULL, "242031253.ogg"},
	{0x6FB4752BDF21339FULL, "zz_dae_debug6131007.cl"},
	{0x6FB47BAA21952A85ULL, "vxht_0_hightown_a3_9wy_0_0.vlm"},
	{0x6FB47D032150E80AULL, "vxdr_primevaltomb_34h_0_1.vlm"},
	{0x6FB4FE7D9F9D8005ULL, "6126868_m.fxe"},
	{0x6FB5D84ACF7AFF1BULL, "vxlt_0_lowtown_docks_com_0_0.vlm"},
	{0x6FB68228C4D44872ULL, "vxw_bonepit_86_0_0.vlm"},
	{0x6FB6DBA1E860BCBFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2b.msh"},
	{0x6FB6E1A1E860C761ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2b.mmh"},
	{0x6FB6EDA1E860DB02ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2b.mao"},
	{0x6FB6FA764DCD4022ULL, "l_env_wal_gallows_whis_murm_p2p.wbd"},
	{0x6FB6FF50CBAC6EDCULL, "vxht_mansion_n_5b9_0_0.vlm"},
	{0x6FB7306515B0592BULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.cess.evt"},
	{0x6FB75FADA918ABF0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxa.msh"},
	{0x6FB76DADA918C30AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxa.mmh"},
	{0x6FB771ADA918CA41ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\tre_timberbeamxa.mao"},
	{0x6FB7A925FFFB1E48ULL, "hlac_dr_0_deeproads_5839_1x.dds"},
	{0x6FB7E6F4810891BAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeloop.ani"},
	{0x6FB82EA0EBCD5815ULL, "6175385_m.fxe"},
	{0x6FB8BC4AEA2A3930ULL, "cod_ltr_black_powder_1.ncs"},
	{0x6FB9696522309F92ULL, "651069397.ogg"},
	{0x6FB9B994AA291093ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.cess.evt"},
	{0x6FB9BB67EA8410C4ULL, "hlac_dr_dungeon_3215_4y.dds"},
	{0x6FBA3658F89F8451ULL, "6042475_f.fxe"},
	{0x6FBAA6E1827504EBULL, "vxw_encounter2_camp_hc_0_0.vlm"},
	{0x6FBB35565D2ED4DFULL, "genip_mallet.utp"},
	{0x6FBBB9B735CC3933ULL, "zz_lgt_1_debug6148338.cl"},
	{0x6FBBBF80F4B3F325ULL, "vxw_encounter1_fj_0_0.vlm"},
	{0x6FBBE568396C2E65ULL, "6106919_f.fxe"},
	{0x6FBC5ABD62E71137ULL, "lt_0_lowtown_a2_29_0.rml"},
	{0x6FBCEEAC47B16564ULL, "121817446.ogg"},
	{0x6FBD30F4C04B6868ULL, "6166942_m.fxe"},
	{0x6FBD533AB23A8A4DULL, "6052091_m.fxe"},
	{0x6FBD57484110F18AULL, "327109589.ogg"},
	{0x6FBD5B1D78319B88ULL, "vxlt_0_lowtown_a2_n_eqh_0_0.vlm"},
	{0x6FBD6C7ED89D7AE7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_basicwood.phy"},
	{0x6FBD89169F18033FULL, "6143816_m.fxe"},
	{0x6FBDA20F9A4771E2ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_2x.dds"},
	{0x6FBDADAD1BB9CEF5ULL, "40957206.ogg"},
	{0x6FBDDEA6D66574EFULL, "823326942.ogg"},
	{0x6FBDDEEBBB3DEC6EULL, "ht_0_hightown_n_10_0.rml"},
	{0x6FBDF2826D04F911ULL, "191889250.ogg"},
	{0x6FBE028875E35FA6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_left.ani"},
	{0x6FBE3ABFBFC1A666ULL, "lt_0_lowtown_a2_n_25_0.anb"},
	{0x6FBE3AF276673EEBULL, "6114495_f.fxe"},
	{0x6FBE3E3DE87AF2DEULL, "hlac_lt_gallowscourtyard_n_-131074_0x.dds"},
	{0x6FBE496515B6A2ECULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.cess.gad"},
	{0x6FBEC5F5119C4E53ULL, "vxlt_hangedman_n_1tv_0_0.vlm"},
	{0x6FBF18C56EDE7E75ULL, "6160949_m.fxe"},
	{0x6FBFB294AA2D70D4ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.cess.gad"},
	{0x6FBFD91BECB32D61ULL, "vxca_caverns1_mines_e7y_0_0.vlm"},
	{0x6FBFE8D2F2E5F0EDULL, "cscript.ldf"},
	{0x6FC09427A7ECFDA3ULL, "llac_w_bonepitmining_a2_131077_0.dds"},
	{0x6FC09B35FAD85730ULL, "6175403_m.fxe"},
	{0x6FC12B11A8E220EFULL, "hlac_lt_0_lowtown_n_-65536_1z.dds"},
	{0x6FC17C302CCFFF97ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_2x.dds"},
	{0x6FC1811B96EC21F6ULL, "hlac_lt_0_lowtown_a2_65537_7x.dds"},
	{0x6FC1FA940BD7B2D9ULL, "fx_liquid_add.mao"},
	{0x6FC263CF5D6060C4ULL, "6065396_f.fxe"},
	{0x6FC2F7E9BA3ACEDFULL, "267111449.ogg"},
	{0x6FC31F726EA822A6ULL, "vxht_0_hightown_a3_c0q_0_0.vlm"},
	{0x6FC3AB5D33EFABD2ULL, "983979883.ogg"},
	{0x6FC3C685D412D31FULL, "rdr110dg_guard_warehouse6016713.cl"},
	{0x6FC3D2C8F56DA05DULL, "690837491.ogg"},
	{0x6FC3D798193C737DULL, "6143054_m.fxe"},
	{0x6FC4225A38B64A57ULL, "vxca_cavern1_haunted_j52_0_0.vlm"},
	{0x6FC5486515BCBFEAULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.cess.ani"},
	{0x6FC5B461833D9951ULL, "vxht_brothel_n_33v_0_0.vlm"},
	{0x6FC5E6DC43145170ULL, "dae000_orsino6085466.cl"},
	{0x6FC6220F09AA675AULL, "6175081_m.fxe"},
	{0x6FC622164FA90236ULL, "dae000_oldmangossip6174889.cl"},
	{0x6FC62E6CA6072A93ULL, "645085784.ogg"},
	{0x6FC65172FE99F1E3ULL, "mcr310_sibling6076256.cl"},
	{0x6FC6B4BF92EF9A7DULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingc.phy"},
	{0x6FC6C194AA33A8B2ULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.cess.ani"},
	{0x6FC6DEE8A551934BULL, "216426790.ogg"},
	{0x6FC6E9EA38E61564ULL, "554240813.ogg"},
	{0x6FC6F217C89B49CCULL, "hlac_ht_0_hightown_a2_-131069_0x.dds"},
	{0x6FC7265EEE4DFAB6ULL, "art\\vfx\\combat\\cn_bdy_corm.mao"},
	{0x6FC77FC4F97C29A9ULL, "873505252.ogg"},
	{0x6FC7D9578FBF03CCULL, "ht_0_hightown_a3_50_0.rml"},
	{0x6FC80096409D4E64ULL, "face_face_p1.vsh.fxo.dx11.assembly"},
	{0x6FC84B365B5DC586ULL, "art\\characters\\weapons\\greataxes\\w_gax_ga105a.mao"},
	{0x6FC88B0CD46EC95EULL, "6113610_m.fxe"},
	{0x6FC9027AA30D3F24ULL, "649114467.ogg"},
	{0x6FC918BE746B1D6BULL, "vxw_encounter2_camp_cb_0_0.vlm"},
	{0x6FC952CA0D852345ULL, "zz_and_debug6134085.cl"},
	{0x6FC957EE6F553B68ULL, "vxht_mansion_n_6ou_0_0.vlm"},
	{0x6FC95F669B3FA2C9ULL, "78569426.ogg"},
	{0x6FC97D19912015E5ULL, "vxca_cavern1_haunted_cn4_0_0.vlm"},
	{0x6FC9F1EBBB93BD4FULL, "cs_mcr310cs_chantry_explosion.wbd"},
	{0x6FCA315DE6AA0BA5ULL, "vxw_encounter1_ki_0_0.vlm"},
	{0x6FCA42AD6AE470F7ULL, "zz_fol_debug6092325.cl"},
	{0x6FCA88C0FC5AE4D8ULL, "288540885.ogg"},
	{0x6FCA9875F58F494EULL, "6149406_m.fxe"},
	{0x6FCACC1D5706606FULL, "vxlt_hangedman_n_25z_0_0.vlm"},
	{0x6FCAD8316C6089ACULL, "dae000ip_crowd_mf8_talk11.utp"},
	{0x6FCAEA8875EDC117ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_left.evt"},
	{0x6FCAEDF30D906BC6ULL, "vxw_encounter1_cg_0_1.vlm"},
	{0x6FCBE74D0951EE4CULL, "lt_0_lowtown_docks_38_0_ll.mmh"},
	{0x6FCC64361ABEDA9DULL, "989974276.ogg"},
	{0x6FCCA3783FE3BF67ULL, "t3_hey_br3.tnt"},
	{0x6FCCE2257362156EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p.ani"},
	{0x6FCCF0C088C7772FULL, "dae310ip_store_tailor.utp"},
	{0x6FCD37D2039266D3ULL, "vxlt_hangedman_n_1qy_0_0.vlm"},
	{0x6FCDBB4F7FD880A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_exit.evt"},
	{0x6FCDEE32924B02EDULL, "gen_im_trs_cot_lgt_boo_r2.uti"},
	{0x6FCE8C4798E9E998ULL, "crt_cycle.dds"},
	{0x6FCE91D69471FA7DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic_2.mao"},
	{0x6FCE9C0D8951611CULL, "48437237.ogg"},
	{0x6FCECDC5D597DEA4ULL, "hlac_lt_0_lowtown_docks_n_-65538_0y.dds"},
	{0x6FCEEC1116C351DCULL, "l_dae111ar_tavern.wbd"},
	{0x6FCF531DC08F576DULL, "hlac_ht_chantry_2534_53z.dds"},
	{0x6FD0BF7C1A7F1A5EULL, "hlac_lt_darktown_-1_34y.dds"},
	{0x6FD0FF316CE6F2EDULL, "232924962.ogg"},
	{0x6FD126E6958B8813ULL, "gen_im_trs_slv_lgt_cht_r0.uti"},
	{0x6FD18AA47D47B622ULL, "1259263.ogg"},
	{0x6FD1914F609E3B26ULL, "vxht_0_hightown_a3_c5r_0_0.vlm"},
	{0x6FD1EB8875F3E1A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\order_left.gad"},
	{0x6FD2BC04B367301BULL, "vxlt_0_lowtown_docks_cuo_0_0.vlm"},
	{0x6FD2DB7B5FBF79F2ULL, "follower_barks6153568.cl"},
	{0x6FD3D24F7FDD13F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_exit.gad"},
	{0x6FD413CAB7C5DBE7ULL, "6044530_f.fxe"},
	{0x6FD51DF864C110D4ULL, "340250934.ogg"},
	{0x6FD544E95747D02DULL, "vxht_bartrand_n_67c_0_0.vlm"},
	{0x6FD5609DDA0357B1ULL, "vxht_brothel_n_4ff_0_0.vlm"},
	{0x6FD605F517A13C03ULL, "lt_backalley_n_water_1875.msh"},
	{0x6FD611B839C907DDULL, "qun210cr_guard6072125.cl"},
	{0x6FD613F517A1545DULL, "lt_backalley_n_water_1875.mmh"},
	{0x6FD617F517A15B6EULL, "lt_backalley_n_water_1875.mao"},
	{0x6FD61BD2F7243F80ULL, "vxw_encounter2_camp_4x_0_0.vlm"},
	{0x6FD61EC59EEC28E4ULL, "crt_so.mao"},
	{0x6FD63B4B2D63067FULL, "261179012.ogg"},
	{0x6FD683B9F7701699ULL, "vxlt_gallowsprison_n_6n4_0_0.vlm"},
	{0x6FD70C164B1F6F76ULL, "vxlt_0_lowtown_a2_f4y_0_0.vlm"},
	{0x6FD732D875713DF2ULL, "6043713_m.fxe"},
	{0x6FD76A9C7D4FEBEDULL, "lt_0_lowtown_docks_8_0_ll.mmh"},
	{0x6FD79BF7FE965416ULL, "689860443.ogg"},
	{0x6FD7A9BF92FDFAE9ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingc.mao"},
	{0x6FD7B5BF92FE0E82ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingc.mmh"},
	{0x6FD7B7BF92FE1238ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingc.msh"},
	{0x6FD7C93368ADDA24ULL, "vxht_mansion2_1wa_0_0.vlm"},
	{0x6FD7ED285FBCAB65ULL, "38519201.ogg"},
	{0x6FD80E19EE33B299ULL, "gen00fl_varric6033683.cl"},
	{0x6FD82762B646E030ULL, "w_blightlands_82_0.mao"},
	{0x6FD82B62B646E743ULL, "w_blightlands_82_0.mmh"},
	{0x6FD83962B646FE9DULL, "w_blightlands_82_0.msh"},
	{0x6FD8B778E9E90889ULL, "ser191_emeric6104775.cl"},
	{0x6FD8C63E1EB25BA8ULL, "art\\characters\\playercharacter\\humanmale\\hm_hat_medc.mao"},
	{0x6FD8E5E5DE5EA12EULL, "vxlt_warehouse_139_0_0.vlm"},
	{0x6FD944068F325A24ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p_injured.ani"},
	{0x6FD9696073267001ULL, "vxht_mansion_n_5bh_0_0.vlm"},
	{0x6FD9C229C1B28D80ULL, "hlac_lt_warehouse_n_-65535_1z.dds"},
	{0x6FD9DA25736C91BFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p.evt"},
	{0x6FD9EA8867CE6B96ULL, "6084910_m.fxe"},
	{0x6FDA5E45F02E4A10ULL, "fex_spt_md.mmh"},
	{0x6FDA62D977E5E5B7ULL, "31751561.ogg"},
	{0x6FDA6AF590FBA92AULL, "and231cr_lyrium_archer.utc"},
	{0x6FDA6C45F02E622AULL, "fex_spt_md.msh"},
	{0x6FDA85E30EDA61E9ULL, "dae011_wench_mob6147244.cl"},
	{0x6FDAD34F7FE3339FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_exit.ani"},
	{0x6FDCB0F4DBC45329ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_1z.dds"},
	{0x6FDCB306784868AFULL, "vxlt_0_lowtown_a3_n_c_eid_0_0.vlm"},
	{0x6FDCBE41D3DBFF42ULL, "484740344.ogg"},
	{0x6FDCE4C5AA0B7007ULL, "gencr_coterie_alchemist.utc"},
	{0x6FDDB4D03881245CULL, "mrl250_pol6101835.cl"},
	{0x6FDE3FBEF32C0282ULL, "follower_banter6151341.cl"},
	{0x6FDE537C1B863135ULL, "6053275_f.fxe"},
	{0x6FDEB5A024A5D7E9ULL, "mag120_macha.bnk"},
	{0x6FDF41068F36C086ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p_injured.gad"},
	{0x6FDF7BAB6EA28BF6ULL, "vxw_bonepitdestroyed_a3_22x_0_0.vlm"},
	{0x6FDFA0424B62588BULL, "crt_pro_01.mao"},
	{0x6FDFA3066996C611ULL, "vxca_caverns2prc_1za_0_0.vlm"},
	{0x6FDFE73F9AE1BB4AULL, "wm_ico_ruins_roll.dds"},
	{0x6FE01A579DFFED73ULL, "follower_banter6168687.cl"},
	{0x6FE052E580639F39ULL, "hlac_ca_caverns2prc_5802_51y.dds"},
	{0x6FE0579602AB3668ULL, "gen00fl_varric6118855.cl"},
	{0x6FE0BB2573727BA8ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p.gad"},
	{0x6FE0C52259F75218ULL, "1015597978.ogg"},
	{0x6FE0C77B5FA16B6CULL, "hlac_ca_cavern1_haunted_24425_6y.dds"},
	{0x6FE14AD2F01C12CDULL, "614729718.ogg"},
	{0x6FE18541CB28DF20ULL, "114547967.ogg"},
	{0x6FE197A024A7DA29ULL, "mag120_macha.cnv"},
	{0x6FE1CC9CF8F80C32ULL, "2334486.ogg"},
	{0x6FE2DFF116A600DEULL, "gen00fl_anders6125561.cl"},
	{0x6FE2E2A1E88603DEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2b.phy"},
	{0x6FE33D514900796BULL, "vxdr_slavers_2hi_0_0.vlm"},
	{0x6FE3809FCDE8260AULL, "vxlt_house_fx_0_0.vlm"},
	{0x6FE38C8E0B5D8D13ULL, "vxdr_primevaltomb_75o_0_0.vlm"},
	{0x6FE3B6C6493DE8ADULL, "vxht_bartrand_n_60b_0_0.vlm"},
	{0x6FE3B95C587993AAULL, "vxlt_warehouse_144_0_0.vlm"},
	{0x6FE44F0E1FC55CBCULL, "6056620_f.fxe"},
	{0x6FE47437BEF3F375ULL, "hlac_ht_chantry_n_2534_11y.dds"},
	{0x6FE5469451CB346DULL, "290057026.ogg"},
	{0x6FE572A1B56A2173ULL, "zz_demo_plot_flags.ncs"},
	{0x6FE5BDB61D2E0C8FULL, "vxlt_0_lowtown_a2_n_c_epc_0_1.vlm"},
	{0x6FE5EF08FA854BD5ULL, "6163207_m.fxe"},
	{0x6FE63C068F3CD765ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\move_p_injured.evt"},
	{0x6FE6971A43444540ULL, "vxlt_0_lowtown_docks_a3_cyw_0_0.vlm"},
	{0x6FE6C679D3A13B8EULL, "w_woundedcoast_230_0.msh"},
	{0x6FE6D479D3A153DFULL, "w_woundedcoast_230_0.mao"},
	{0x6FE6E079D3A1687CULL, "w_woundedcoast_230_0.mmh"},
	{0x6FE715040080CEEBULL, "mer140ip_at_caverns_ledge.utp"},
	{0x6FE72E31A98E514BULL, "6149652_m.fxe"},
	{0x6FE787817782412AULL, "vxdr_0_deeproads_egp_0_0.vlm"},
	{0x6FE7FE938927CB59ULL, "vfx_woodbits01.dds"},
	{0x6FE837DD345926D3ULL, "lowtown_fem_urchin_mob6165964.cl"},
	{0x6FE85EE7DE83928CULL, "t3_nbl_n00.tnt"},
	{0x6FE903D49A8E00CCULL, "vxlt_0_lowtown_h96_0_0.vlm"},
	{0x6FE95D319E766850ULL, "6041582_m.fxe"},
	{0x6FE995C30E5EE77DULL, "513825423.ogg"},
	{0x6FE99E13B3458800ULL, "6107345_m.fxe"},
	{0x6FE9F7A324CBDB26ULL, "ico_2h_sword_5.dds"},
	{0x6FEA435BC848CA23ULL, "6081977_f.fxe"},
	{0x6FEAB4E5C31EFC23ULL, "ts_prog_cubic_zirc.gda"},
	{0x6FEAD10A6FBC8B94ULL, "112609526.ogg"},
	{0x6FEADD705373B68CULL, "vxca_caverns1_mines_dnz_0_0.vlm"},
	{0x6FEB508BFF002194ULL, "vxlt_0_lowtown_a2_n_c_gza_0_0.vlm"},
	{0x6FEB8BC9C8D5C9ECULL, "vxlt_darktown_5o2_0_0.vlm"},
	{0x6FEC19A41E141CFCULL, "6139069_m.fxe"},
	{0x6FECD4841897C26DULL, "6131203_f.fxe"},
	{0x6FED067DE49B7DD6ULL, "dae000_bodhan6129131.cl"},
	{0x6FED18C57051DD9CULL, "vxht_keep_67f_0_0.vlm"},
	{0x6FED486A9397145DULL, "483998765.ogg"},
	{0x6FEDB0269F245C68ULL, "6070344_m.fxe"},
	{0x6FEE05B8D776E0F5ULL, "6142152_m.fxe"},
	{0x6FEE4949F535CFB8ULL, "97979373.ogg"},
	{0x6FEEEF8C94DD4B0CULL, "hlac_lt_undercitydungeon_-1_16z.dds"},
	{0x6FEF24ADDC2CA131ULL, "hlac_dr_slavers_3216_1x.dds"},
	{0x6FEF683A33EC4D88ULL, "112162241.ogg"},
	{0x6FF0188EF3FAF81DULL, "vxlt_warehouse_n_sf_0_0.vlm"},
	{0x6FF05C613D542E47ULL, "14310792.ogg"},
	{0x6FF097488664CF59ULL, "460717576.ogg"},
	{0x6FF0B8244C9E680AULL, "6093654_f.fxe"},
	{0x6FF0E0DE30884917ULL, "gen00fl_aveline6120795.cl"},
	{0x6FF0FBCA9A702E02ULL, "vxht_keep_n_3zu_0_0.vlm"},
	{0x6FF149FA9519C2F6ULL, "llac_ht_0_hightown_65536_4.dds"},
	{0x6FF1841B83B646A6ULL, "205574997.ogg"},
	{0x6FF1A547371069CFULL, "271936748.ogg"},
	{0x6FF1E6580E58D53DULL, "mrl250_pol6101326.cl"},
	{0x6FF209D0403A8960ULL, "lgt150_nat_were_supporter6110983.cl"},
	{0x6FF28EB314C8C778ULL, "mag210_arianni6031661.cl"},
	{0x6FF2BF2ADD307CDAULL, "vxlt_0_lowtown_n_byi_0_0.vlm"},
	{0x6FF2E30DF79969DBULL, "292146258.ogg"},
	{0x6FF38E54DEC69701ULL, "vxlt_hangedman_1qh_0_0.vlm"},
	{0x6FF3D2BB381CDFB3ULL, "gen_im_arm_boo_hvy_act2_03.uti"},
	{0x6FF44B46F77BE879ULL, "6111358_m.fxe"},
	{0x6FF48E3FB53D0B9CULL, "6125724_f.fxe"},
	{0x6FF4907289A5648EULL, "dae200_saemus6142806.cl"},
	{0x6FF5563C5958C7CEULL, "t3_wdg_dr2.tnt"},
	{0x6FF579B1C635FB8AULL, "hlac_lt_blackemporium_1055_1y.dds"},
	{0x6FF5A00E9B8469CBULL, "6142752_m.fxe"},
	{0x6FF5D0683D11F81AULL, "vxca_caverns2_853_0_0.vlm"},
	{0x6FF623E830E932A5ULL, "art\\levels\\races\\design\\_textures\\prp_rubble_n.dds"},
	{0x6FF63215289059D5ULL, "160408385.ogg"},
	{0x6FF64D1A5712A101ULL, "vxht_mansion_n_5hn_0_0.vlm"},
	{0x6FF6C95698E6BBDAULL, "vxlt_0_lowtown_a2_n_c_e1x_0_0.vlm"},
	{0x6FF77A57CD1DC5E0ULL, "follower_barks6153445.cl"},
	{0x6FF814E0EBE4BB98ULL, "llac_dr_primevaltomb_11985_2.dds"},
	{0x6FF82E2C651F3421ULL, "6080709_f.fxe"},
	{0x6FF8D810882F7ECFULL, "llac_lt_gallowstemplar_n_2391_7.dds"},
	{0x6FF94B4A660E0DA2ULL, "gen00fl_fenris6107185.cl"},
	{0x6FF971430477EBF3ULL, "370521498.ogg"},
	{0x6FF9F8C12E4D8A94ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.cess.ani"},
	{0x6FF9FDA6BACBE26CULL, "vxlt_darktown_5p3_0_0.vlm"},
	{0x6FFA628FAA7B81CCULL, "vxlt_0_lowtown_eqo_0_0.vlm"},
	{0x6FFA72487BE7718DULL, "228436093.ogg"},
	{0x6FFAEA1116E88956ULL, "l_dae111ar_tavern.bnk"},
	{0x6FFB8AA26247F61CULL, "vxht_keep_60g_0_0.vlm"},
	{0x6FFBC7EE0103A802ULL, "zz_isa_debug6175180.cl"},
	{0x6FFC0AEDCBE86433ULL, "t3_nbl_c02.tnt"},
	{0x6FFC2203911A74E8ULL, "6075044_m.fxe"},
	{0x6FFD0C3CFDD23310ULL, "683749061.ogg"},
	{0x6FFD20452049C6F6ULL, "vxht_0_hightown_a2_2xa_0_0.vlm"},
	{0x6FFD4F442DE7A681ULL, "6103735_m.fxe"},
	{0x6FFD518FECE02967ULL, "6133550_m.fxe"},
	{0x6FFDEDBFBC80DB2FULL, "zz_followers_banter6157539.cl"},
	{0x6FFE08764E095AD8ULL, "l_env_wal_gallows_whis_murm_p2p.bnk"},
	{0x6FFE77399CE47E56ULL, "6166306_m.fxe"},
	{0x6FFF45BE72BF02A7ULL, "vxca_caverns1_gkq_0_1.vlm"},
	{0x6FFFE11A82424841ULL, "art\\characters\\playercharacter\\humanmale\\hlm_ric.mao"},
	{0x6FFFE258E3FD3C90ULL, "zz_mcr_debug6080856.cl"},
	{0x6FFFEB4200DEAEA7ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r180_rf.gad"},
	{0x7000DF707A1C7D64ULL, "227437795.ogg"},
	{0x7000F5C12E53A4F6ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.cess.gad"},
	{0x700106049D2ACAEAULL, "qun210cr_guard6083832.cl"},
	{0x70018E15CAFC4CA8ULL, "dre000_varric6034668.cl"},
	{0x70018FB308997130ULL, "6177815_f.fxe"},
	{0x7001D90394E249C9ULL, "vxlt_undercitydungeon_13l_0_0.vlm"},
	{0x7001D950CD526019ULL, "vxlt_gallowsprison_n_6q9_0_0.vlm"},
	{0x700250D1CAA07F27ULL, "373454241.ogg"},
	{0x70027AF0A165657AULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.cess.gad"},
	{0x7002CC6C00E82A17ULL, "6173060_m.fxe"},
	{0x7002EFE3BA57115DULL, "dae000_arishok6080434.cl"},
	{0x7003143FDB090C44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_loop.ani"},
	{0x700339B62254D6B2ULL, "vxlt_0_lowtown_a2_n_gyt_0_0.vlm"},
	{0x70037424DB58B2F2ULL, "6076124_m.fxe"},
	{0x700397BCA20C17C5ULL, "hlac_lt_undercityserialkiller_-1_1x.dds"},
	{0x7003A76052AAFC61ULL, "gen00fl_fenris6126512.cl"},
	{0x7004CB2214D3686BULL, "vxlt_0_lowtown_docks_cs9_0_0.vlm"},
	{0x70050449DFD67AB9ULL, "hla_ht_0_hightown_a3_131070_0z.dds"},
	{0x70050944DCE4965BULL, "931556679.ogg"},
	{0x70055D7B1C270436ULL, "character_character_p78.psh.fxo.dx11"},
	{0x70057647C959007DULL, "llac_ht_mansion2_818_27.dds"},
	{0x70057D4E1A80BE09ULL, "llac_ca_caverns1_mines_24426_12.dds"},
	{0x7005C8DEFE3049BAULL, "ave121_jeven6105689.cl"},
	{0x7006C179311CE745ULL, "479179882.ogg"},
	{0x7006D716C0D6487AULL, "788489514.ogg"},
	{0x7006E04200E4BAB4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r180_rf.evt"},
	{0x7007175AB46DC5C5ULL, "hlac_ca_caverns2prc_5802_7x.dds"},
	{0x70074A76630A483DULL, "6097144_m.fxe"},
	{0x70074D7509F31326ULL, "hlac_lt_0_lowtown_a2_n_65534_0y.dds"},
	{0x700776B3E950E2FFULL, "6175097_m.fxe"},
	{0x70079E4315C9ABB1ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_die_c.impa.ani"},
	{0x70079EDFB7E9E465ULL, "vxca_caverns2_7xh_0_0.vlm"},
	{0x7007D816BC9F410EULL, "art\\vfx\\combat\\vfx_gib_hm_masc_ltrb.anb"},
	{0x700810C12E59F155ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.cess.evt"},
	{0x7008155A341EF9FEULL, "vxdr_0_deeproads_fiy_0_0.vlm"},
	{0x70083445E2EC5294ULL, "vxlt_0_lowtown_a2_n_c_gpg_0_0.vlm"},
	{0x7008344CFC616EF3ULL, "vxw_sundermount_23m_0_0.vlm"},
	{0x7008418474E22F3BULL, "lgt310cr_mis_burn_murder2.utc"},
	{0x700843002CDB0DA1ULL, "clf260_harley.bnk"},
	{0x700854DDCBBAF16AULL, "dae000cr_grand_cleric6147694.cl"},
	{0x70087DF0A169D591ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.cess.evt"},
	{0x7008826EA6C6CEE9ULL, "6033483_m.fxe"},
	{0x7008A3BA00EEAE0DULL, "vxw_sundermount_2c3_0_0.vlm"},
	{0x7008DB433F86E8BCULL, "hlac_lt_0_lowtown_a2_65534_2x.dds"},
	{0x7009F35C5887F59FULL, "plt_lgt120pt_bhelen.nss"},
	{0x700A113FDB0F25A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_loop.gad"},
	{0x700A168165D3F8B2ULL, "vxht_mansion2b_1xb_0_0.vlm"},
	{0x700AA43550B2FC63ULL, "6162097_m.fxe"},
	{0x700AC5C84C0058A1ULL, "364775070.ogg"},
	{0x700AFF8CDAE84DF2ULL, "6133326_m.fxe"},
	{0x700B09D0F3D3E66CULL, "6118792_m.fxe"},
	{0x700B0BEBBBCB8025ULL, "cs_mcr310cs_chantry_explosion.bnk"},
	{0x700B24E8702306C6ULL, "6032134_f.fxe"},
	{0x700B2C1D4B80229FULL, "6118217_f.fxe"},
	{0x700B3ABFF8B35BFEULL, "6098700_m.fxe"},
	{0x700C65002CDF3011ULL, "clf260_harley.cnv"},
	{0x700CBF79D3C1FF47ULL, "w_woundedcoast_230_0.phy"},
	{0x700D45E91481647BULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.cess.ani"},
	{0x700DBB827BC11391ULL, "221209056.ogg"},
	{0x700E33B458BF1DEAULL, "37885034.ogg"},
	{0x700E43CE74018819ULL, "mag291_torpor6021686.cl"},
	{0x700EA94315CFDD53ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_die_c.impa.gad"},
	{0x700ED7A04CE714B5ULL, "fxa_power_smoke.dds"},
	{0x700F2811A1DBB9B1ULL, "vxht_brothel_n_4zz_0_0.vlm"},
	{0x700F4E21ABBF4035ULL, "vxlt_0_lowtown_a2_gjq_0_0.vlm"},
	{0x700F733B94260713ULL, "zz_qcr_debug6106184.cl"},
	{0x700FCFFFC67E607FULL, "6118345_f.fxe"},
	{0x701022E99ED5F9EAULL, "883165227.ogg"},
	{0x701042F27D1E2029ULL, "vxlt_backalley_g_1pq_0_1.vlm"},
	{0x7010555DE1EBA3C6ULL, "hlac_lt_0_lowtown_65537_6y.dds"},
	{0x7010FC818EAE5F03ULL, "279929362.ogg"},
	{0x701108FE22466DD4ULL, "512240460.ogg"},
	{0x7011117D1DB6930FULL, "776952384.ogg"},
	{0x7011F2B3BC7549C1ULL, "vxca_caverns1_mines_jgb_0_0.vlm"},
	{0x7012721A6C5365C6ULL, "6129668_m.fxe"},
	{0x7012C676A610CDF7ULL, "llac_lt_0_lowtown_a2_n_65535_5.dds"},
	{0x7013B7E3F4FFA5D4ULL, "426340319.ogg"},
	{0x70142F656C5BC860ULL, "6098947_f.fxe"},
	{0x70144B8CC88C98CAULL, "6095493_m.fxe"},
	{0x70146A6FD9590154ULL, "follower_banter6167878.cl"},
	{0x7014A64315D443B8ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_die_c.impa.evt"},
	{0x7014D84200F0EAEDULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r180_rf.ani"},
	{0x701563B9FEFE72FBULL, "vxht_bartrand_n_63i_0_0.vlm"},
	{0x70157A58D09E710DULL, "ser221_gamlen6133138.cl"},
	{0x7015EB6E66D7780AULL, "6024570_m.fxe"},
	{0x7015F8F163694502ULL, "zz_lgt_1_debug6136981.cl"},
	{0x701616455B096631ULL, "nat320_delilah6065929.cl"},
	{0x70162199EDCD457EULL, "59522478.ogg"},
	{0x70167ACACFAC4A30ULL, "gen00fl_sebastian6188691.cl"},
	{0x701695F0A1763C60ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.cess.ani"},
	{0x7016A629EE578773ULL, "vxw_sundermount_24l_0_0.vlm"},
	{0x7016B639D940FC1FULL, "84800412.ogg"},
	{0x7016D111B1E73401ULL, "vxlt_gallowsprison_n_5qb_0_0.vlm"},
	{0x7016F4D3509EDBF4ULL, "gen_im_wep_war_two_start.uti"},
	{0x70172FF6B54E6323ULL, "lowtown_male_urchin_sta6164238.cl"},
	{0x7017EF753AC9F762ULL, "135941209.ogg"},
	{0x70184250834F3EE0ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtb.phy"},
	{0x70184AEDE6B7C881ULL, "346011925.ogg"},
	{0x70187845223D8804ULL, "6070797_m.fxe"},
	{0x7018885E57CA1132ULL, "vxht_mansion2b_1sc_0_0.vlm"},
	{0x7018F1B7F753F956ULL, "hlac_dr_dungeon_3217_9z.dds"},
	{0x701963C69E6FAE20ULL, "vxlt_0_lowtown_n_h8n_0_0.vlm"},
	{0x701971E2894850B9ULL, "480583296.ogg"},
	{0x70197C9E8493BB22ULL, "hlac_ht_bartrand_n_2995_1y.dds"},
	{0x7019CD1ACD8673FDULL, "6119259_m.fxe"},
	{0x701A5D28CC9F2FA0ULL, "hlac_ca_caverns2_5802_41x.dds"},
	{0x701A9C6C5B99FAE3ULL, "llac_lt_0_lowtown_a2_131072_4.dds"},
	{0x701AE3CFD5A504FCULL, "9797394.ogg"},
	{0x701AF9398F3A6D92ULL, "6055447_f.fxe"},
	{0x701B4DE9148DAF72ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.cess.evt"},
	{0x701B655DCFE56C78ULL, "genip_cin_barge2.utp"},
	{0x701BC843EB93C710ULL, "one110_meeran6090190.cl"},
	{0x701BF044037820D7ULL, "hlac_lt_0_lowtown_a2_-65535_4y.dds"},
	{0x701CBE8D75FE204FULL, "da2_prc_one_it-it.cre"},
	{0x701CEA494F32700CULL, "w_encounter1_40_0.phy"},
	{0x701DE65C589888B0ULL, "plt_lgt120pt_bhelen.plo"},
	{0x701DF6D8B6640187ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_1.mao"},
	{0x701F08719845727BULL, "363257757.ogg"},
	{0x701F90F42EAA5F83ULL, "zz_fen_debug6165669.cl"},
	{0x70200822DF60ECE9ULL, "vxlt_gallowstemplar_n_372_0_0.vlm"},
	{0x70201AFFB5F0602DULL, "634289076.ogg"},
	{0x70208E80D360B665ULL, "dae000_arishok6067770.cl"},
	{0x70216CE914924F05ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.cess.gad"},
	{0x70239B3DE0456C9EULL, "242888525.ogg"},
	{0x7023B6E580B900F9ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_alis.mmh"},
	{0x7023C0E580B911E7ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_alis.msh"},
	{0x702488089F01C9AAULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_1y.dds"},
	{0x70249F8A10697157ULL, "vxlt_backalley_g_1vu_0_1.vlm"},
	{0x7024B7A178200CC2ULL, "lgt311_shw_mer_mettin6156464.cl"},
	{0x702557EFBB933124ULL, "dae000_bartrand6031139.cl"},
	{0x70258773E4DADF0DULL, "vxw_sundermount_2i1_0_0.vlm"},
	{0x702599ADEE5D64C8ULL, "619712792.ogg"},
	{0x7025AA02377A88B7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basic.mmh"},
	{0x7025AE02377A8FC4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basic.mao"},
	{0x7025C002377AAE29ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wo_basic.msh"},
	{0x70264E2951E45151ULL, "art\\cinematics\\act1\\cs_fp_gritteeth.gad"},
	{0x7026FA3B49C029B2ULL, "vxht_mansion2b_1vl_0_0.vlm"},
	{0x702752F43208883CULL, "vxlt_gallowsprison_n_6kb_0_0.vlm"},
	{0x70277F2E758A4C7EULL, "vxlt_0_lowtown_docks_a3_8uf_0_0.vlm"},
	{0x702781981147A2FDULL, "404666667.ogg"},
	{0x702787F0791A3CB6ULL, "6071204_m.fxe"},
	{0x702808AC8A7C8F67ULL, "136310917.ogg"},
	{0x702853E59BC8B8F0ULL, "6174765_m.fxe"},
	{0x7028991E4A8CC63BULL, "782948979.ogg"},
	{0x70294BB815B38CEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ent.gad"},
	{0x70296BFBB95115EDULL, "vxlt_0_lowtown_docks_n_b9o_0_0.vlm"},
	{0x7029790CA72975B3ULL, "vxlt_warehouse_n_15q_0_0.vlm"},
	{0x7029D17B5CABD02FULL, "6123393_m.fxe"},
	{0x7029F5A2FA424E2BULL, "918777459.ogg"},
	{0x702A7890AC99E8DCULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_com_0d.dds"},
	{0x702B1AAC1E4BD86AULL, "hlac_lt_0_lowtown_n_-262144_2x.dds"},
	{0x702B2CB5EC09CBE4ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb104a_0t.dds"},
	{0x702B3C4DED8D383CULL, "llac_ca_caverns1_24425_8.dds"},
	{0x702BAA6A3D38ED68ULL, "6163574_m.fxe"},
	{0x702BF484F8E5B25FULL, "hlac_lt_0_lowtown_docks_a3_-2_1z.dds"},
	{0x702C4F2951E8BEFFULL, "art\\cinematics\\act1\\cs_fp_gritteeth.ani"},
	{0x702D55C8A69EF901ULL, "vxlt_hangedman_1ml_0_0.vlm"},
	{0x702D81F7D301E659ULL, "6091829_m.fxe"},
	{0x702DB9532AE4BD90ULL, "6085260_f.fxe"},
	{0x702DED4E3312108CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_keepdoor.phy"},
	{0x702E0C7ADD86759CULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv2.ani"},
	{0x702E146FA8A9D9ADULL, "625583389.ogg"},
	{0x702E33883712B3B5ULL, "6036659_m.fxe"},
	{0x702E4684BA80135AULL, "mer340_cowan6083526.cl"},
	{0x702E8BF51FA0C47AULL, "515361078.ogg"},
	{0x702E8F4CF8372032ULL, "vxlt_0_lowtown_a2_n_ghw_0_0.vlm"},
	{0x702ECD1B2E50C1FEULL, "6071131_f.fxe"},
	{0x702F8FC367D14FBEULL, "6111995_m.fxe"},
	{0x70302EB815B97981ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ent.evt"},
	{0x703083B479FFEACEULL, "vxlt_gallowstemplar_5lx_0_1.vlm"},
	{0x7030B3869DDD58D0ULL, "pssmskinmesh.vsh.fxo"},
	{0x7030F4B5053459C1ULL, "131806309.ogg"},
	{0x7031415C139852BDULL, "vxlt_warehouse_n_1dl_0_0.vlm"},
	{0x70319D2BD73A4C63ULL, "873337734.ogg"},
	{0x7031EF78B17345EFULL, "632031011.ogg"},
	{0x70324773E2EAA3FBULL, "vxht_bartrand_n_69k_0_0.vlm"},
	{0x703275B5A6E588F1ULL, "cod_bks_black_fox.utp"},
	{0x7032E8A175138DA9ULL, "vxlt_0_lowtown_n_e2x_0_0.vlm"},
	{0x7033F950D6D0F78DULL, "vxw_sundermount_2r0_0_0.vlm"},
	{0x7034592131B8E79CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_quick_modal_activate.ani"},
	{0x703497E922B2848BULL, "vxht_0_hightown_asv_0_0.vlm"},
	{0x70349D5942B13814ULL, "157551941.ogg"},
	{0x7035197ADD8CAADEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv2.gad"},
	{0x703520D9FA1F3B19ULL, "vxlt_0_lowtown_n_fij_0_0.vlm"},
	{0x70353C35CFFF4C4AULL, "vxw_bonepit_p6_0_0.vlm"},
	{0x70353E4F328BA8F8ULL, "hlac_lt_0_lowtown_n_131072_6x.dds"},
	{0x70353EE0F28DC87DULL, "w_encounter3_69_0.phy"},
	{0x7035520617731120ULL, "llac_ht_mansion2_818_48.dds"},
	{0x703561EFDAD29FBFULL, "hlac_ca_cavern1_haunted_24426_14z.dds"},
	{0x70359B55811BB817ULL, "293145584.ogg"},
	{0x7035D8D76583935BULL, "vxlt_0_lowtown_a2_gmz_0_0.vlm"},
	{0x703646CC4B4D0141ULL, "l_and211ar_andersshop.bnk"},
	{0x703700DDA4D120F6ULL, "art\\characters\\creatures\\mubari\\animation\\exit.ani"},
	{0x70370F12434BBE96ULL, "vxht_bartrand_n_5qi_0_0.vlm"},
	{0x70376B1D7DE59D4AULL, "vxht_0_hightown_n_4as_0_0.vlm"},
	{0x7037DDD8AB472E6DULL, "vxlt_0_lowtown_docks_n_b6n_0_0.vlm"},
	{0x703843584EA1E8AFULL, "6128293_m.fxe"},
	{0x703AC2959FA3F384ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomecclosed.mmh"},
	{0x703AC6959FA3FAD7ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomecclosed.mao"},
	{0x703AD8959FA41926ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomecclosed.msh"},
	{0x703B662131BF1CDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_quick_modal_activate.gad"},
	{0x703BEDEBED963B09ULL, "gen_im_trs_rdr_med_cht_r0.uti"},
	{0x703BF3D4C4F7FED9ULL, "6096929_m.fxe"},
	{0x703BF47ADD928AFDULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv2.evt"},
	{0x703BF4BE22C164ABULL, "fx_dot01.dds"},
	{0x703C48E3BDA3DDA7ULL, "vxlt_gallowstemplar_4n7_0_0.vlm"},
	{0x703D45E06FDB4FECULL, "gen00fl_anders6125457.cl"},
	{0x703E00ED6DEDE890ULL, "vxht_0_hightown_a2_a6r_0_0.vlm"},
	{0x703E26B815C5A9D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gptrl_ent.ani"},
	{0x703E5F5083703F75ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtb.msh"},
	{0x703E611934D85FA2ULL, "6018740_m.fxe"},
	{0x703E6150837042EBULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtb.mmh"},
	{0x703E6D5083705688ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtb.mao"},
	{0x703E75CE9F1A4B52ULL, "6091835_m.fxe"},
	{0x703EA94819D27813ULL, "119343512.ogg"},
	{0x703EC0F3C21C045BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_y_pos.ani"},
	{0x703ED65374F0F2B3ULL, "834021388.ogg"},
	{0x703F1883BDB13B3CULL, "dae000cr_grand_cleric6187951.cl"},
	{0x703F2C7380393198ULL, "6013400_m.fxe"},
	{0x703FCAD5B26BC699ULL, "gen00fl_aveline6113984.cl"},
	{0x703FFF99EE716BD0ULL, "vxw_sundermount_30f_0_0.vlm"},
	{0x7040320129AE3D3DULL, "gen00fl_aveline6116680.cl"},
	{0x704037A73F4D556BULL, "298345844.ogg"},
	{0x704064FBA438945DULL, "6036545_m.fxe"},
	{0x704073992D623748ULL, "gen00fl_aveline6132670.cl"},
	{0x7040A6F87B6DD27DULL, "vxlt_undercityencounter_1lp_0_0.vlm"},
	{0x7040B622EAD7A8B6ULL, "ser100_party_comments6108235.cl"},
	{0x7040D05A4473A6F4ULL, "ser121_jethann6104435.cl"},
	{0x7040D7432F6C99F0ULL, "499219604.ogg"},
	{0x704111E535F771C9ULL, "876668352.ogg"},
	{0x7041199971019FA4ULL, "gen00fl_anders6126001.cl"},
	{0x704149049B7C4650ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden4.mmh"},
	{0x704157049B7C5E6AULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden4.msh"},
	{0x7041E819D5178E8AULL, "hlac_lt_0_lowtown_docks_-65537_4y.dds"},
	{0x7042211B84FDA1A2ULL, "vxlt_blackemporium_3q_0_1.vlm"},
	{0x704224903C17760AULL, "dae000_fem_priest_mob6174192.cl"},
	{0x7042412131C4FCFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_quick_modal_activate.evt"},
	{0x7042431BE5C1C3E2ULL, "6002808_m.fxe"},
	{0x7042503A0EAEF53EULL, "mag211_marethari6025349.cl"},
	{0x7042AEB6B69EDA47ULL, "6071281_m.fxe"},
	{0x70437AC9B7263889ULL, "240438952.ogg"},
	{0x704408DDA4DBB8E7ULL, "art\\characters\\creatures\\mubari\\animation\\exit.evt"},
	{0x704447B8336BC850ULL, "vxw_encounter2_camp_pu_0_1.vlm"},
	{0x704492E46B6E4704ULL, "w_bonepitdestroyed_a3_235_0.msh"},
	{0x7044A4E46B6E666DULL, "w_bonepitdestroyed_a3_235_0.mao"},
	{0x7044A8E46B6E6CA6ULL, "w_bonepitdestroyed_a3_235_0.mmh"},
	{0x70450F20295DE9A4ULL, "1040945427.ogg"},
	{0x704580EF3541D716ULL, "vxht_bartrand_n_5tj_0_0.vlm"},
	{0x704633E0F29C28E9ULL, "w_encounter3_69_0.mao"},
	{0x70463FE0F29C3C82ULL, "w_encounter3_69_0.mmh"},
	{0x704641E0F29C4038ULL, "w_encounter3_69_0.msh"},
	{0x704700984919C9C7ULL, "vxlt_0_lowtown_a2_n_ep4_0_0.vlm"},
	{0x70477AC8D3099963ULL, "308234939.ogg"},
	{0x70477C9555ABF096ULL, "one100dg_vanard6173862.cl"},
	{0x7048082E98F6DBE2ULL, "6048518_m.fxe"},
	{0x704833C68E6AB36FULL, "hlac_ht_0_hightown_a3_2_1z.dds"},
	{0x70483964B9913244ULL, "6033815_m.fxe"},
	{0x7048B58905FB4DFBULL, "172760903.ogg"},
	{0x7049664A92DC838CULL, "767113148.ogg"},
	{0x704A2C848FEF31A3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic_2.mao"},
	{0x704A705839428A2CULL, "gen00fl_varric6076580.cl"},
	{0x704AB2F6335747ADULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medc.mmh"},
	{0x704AC4F633576633ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medc.msh"},
	{0x704AFAD947EA5A7BULL, "mag110_samson6016157.cl"},
	{0x704B29DDA4E20F30ULL, "art\\characters\\creatures\\mubari\\animation\\exit.gad"},
	{0x704B7306DC235132ULL, "vxlt_0_lowtown_a2_n_gbq_0_0.vlm"},
	{0x704B87C3BCFE6A24ULL, "dae000_arishok6067874.cl"},
	{0x704C43165AB3A6ADULL, "dae000_orana6149032.cl"},
	{0x704C880F6025DF85ULL, "6016565_m.fxe"},
	{0x704D94DF8AEEE63EULL, "6059963_m.fxe"},
	{0x704EDC6E03B5059FULL, "220501960.ogg"},
	{0x704EF80179A9F4F3ULL, "vxw_encounter2_58_0_0.vlm"},
	{0x704F55145D20C6FEULL, "hlac_lt_0_lowtown_a2_n_c_-2_2y.dds"},
	{0x704F824F821D2B1CULL, "96368655.ogg"},
	{0x704FAF1174645112ULL, "qcr000_arishok6167000.cl"},
	{0x704FB8D2AFF1D320ULL, "hlac_lt_gallowstemplar_2395_1z.dds"},
	{0x704FDF9F37C1BE53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\turn_left_180.ani"},
	{0x7050B4F8D7B7DC62ULL, "6005908_m.fxe"},
	{0x7051300620F23828ULL, "gen00fl_merrill6106508.cl"},
	{0x70513127EA02BD65ULL, "og_flsh_02.mao"},
	{0x705157D9D370EC21ULL, "vxlt_0_lowtown_h7y_0_0.vlm"},
	{0x70517FF5B6CE23A9ULL, "103412750.ogg"},
	{0x7051993887E6A737ULL, "494664721.ogg"},
	{0x7051D1A1AFC06CA6ULL, "6163598_m.fxe"},
	{0x705224C8BA1B8C91ULL, "91182152.ogg"},
	{0x70526B7C96B8B04FULL, "w_bonepitdestroyed_a3_20_0_ll.mmh"},
	{0x705309E66C93444CULL, "vxht_chantry_n_2c1_0_0.vlm"},
	{0x705356DBDD9A3F4AULL, "8121945.ogg"},
	{0x7053BCACD7E7C338ULL, "hlac_lt_0_lowtown_docks_a3_n_0_2x.dds"},
	{0x7053E7F3C22EA2E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_y_pos.gad"},
	{0x7053FD361C4A00EAULL, "vxht_brothel_n_5e1_0_0.vlm"},
	{0x7054255FB68A25B5ULL, "dre000_dougal6130997.cl"},
	{0x7054F64BBACB9430ULL, "prty_aveline_armor_3.dds"},
	{0x70550512EABA1B84ULL, "zz_mag_debug6039787.cl"},
	{0x705507428D79F087ULL, "6117259_m.fxe"},
	{0x705552C4A701AB8AULL, "470197066.ogg"},
	{0x705572753B0FE247ULL, "vxlt_0_lowtown_a2_n_eu7_0_0.vlm"},
	{0x7055DDA5BDDE32F9ULL, "21623311.ogg"},
	{0x70563745E0A73DAFULL, "w_bonepit_10_0_lc.mmh"},
	{0x70568EE27F3862EFULL, "dae100_alistair_drunk6135143.cl"},
	{0x7056D24B8AABADB6ULL, "vxca_caverns2_7ve_0_0.vlm"},
	{0x705702816C7E3629ULL, "6173145_m.fxe"},
	{0x705712DC60FA58CAULL, "llac_lt_0_lowtown_docks_a3_-131072_0.dds"},
	{0x70574765DC3ABEC3ULL, "vxlt_0_lowtown_docks_n_bss_0_0.vlm"},
	{0x70576F4EDE029051ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_wall_lean_left.anb"},
	{0x7057E39504C5B928ULL, "817700500.ogg"},
	{0x70581720C8389222ULL, "1044146262.ogg"},
	{0x7058562E83D8C6BBULL, "6116534_f.fxe"},
	{0x7059009954A6550BULL, "vxlt_0_lowtown_n_dej_0_0.vlm"},
	{0x705942A991C07C78ULL, "6118957_f.fxe"},
	{0x70595A42EC4C544DULL, "6133735_f.fxe"},
	{0x7059B91BBD6D1A81ULL, "docks_fem_sailor_mob6176303.cl"},
	{0x705A9F623B5239FEULL, "425251504.ogg"},
	{0x705ADE2429DC41F6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_baserocks.mao"},
	{0x705AEA2429DC5595ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_baserocks.mmh"},
	{0x705AEC2429DC590BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_baserocks.msh"},
	{0x705AFD6639E65AE2ULL, "761384542.ogg"},
	{0x705B09B5146F521AULL, "200936165.ogg"},
	{0x705B2CE39308DBC2ULL, "hlac_lt_backalley_g_0_2x.dds"},
	{0x705BABF63365AF6AULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medc.phy"},
	{0x705BC2FF6900B747ULL, "6045336_m.fxe"},
	{0x705BC9DD31A0BB04ULL, "dae000ip_crowd_ff4_talk14.utp"},
	{0x705BCC70A1463532ULL, "6189396_m.fxe"},
	{0x705CD79F37CC3A8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\turn_left_180.evt"},
	{0x705D241730871826ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop_low.evt"},
	{0x705D28B8FE03E904ULL, "6129438_f.fxe"},
	{0x705D8194CF112A1EULL, "952015938.ogg"},
	{0x705D8E48CD7C55B7ULL, "zz_dae_debug6029874.cl"},
	{0x705D8F884D038CFCULL, "vxw_sundermount_2zl_0_1.vlm"},
	{0x705DC1494F69C1D7ULL, "w_encounter1_40_0.mmh"},
	{0x705DC36F280C3B73ULL, "gen00fl_fenris6126536.cl"},
	{0x705DC5494F69C8E4ULL, "w_encounter1_40_0.mao"},
	{0x705DD7494F69E749ULL, "w_encounter1_40_0.msh"},
	{0x705E477BF97CE2FFULL, "60936085.ogg"},
	{0x705E5AA73596A369ULL, "follower_banter6163543.cl"},
	{0x705EE27C0EF0A406ULL, "llac_ca_caverns1_mines_24423_1.dds"},
	{0x705F193D4BF74645ULL, "la_sp_dot.mao"},
	{0x705F470687640481ULL, "vxw_encounter2_camp_4u_0_0.vlm"},
	{0x7060AFEC3719A22DULL, "hlac_lt_0_lowtown_a2_-65535_5x.dds"},
	{0x70614F0DEF02A170ULL, "6034317_m.fxe"},
	{0x70617C362BB9B084ULL, "hlac_ht_chantry_n_2534_2z.dds"},
	{0x706256F53853F4F9ULL, "follower_banter6165376.cl"},
	{0x70626F130E40196AULL, "vxht_brothel_n_5h2_0_0.vlm"},
	{0x706309C375D439DAULL, "lgt351ip_shw_mer_to_bmage.utp"},
	{0x70633B17308BAB71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop_low.gad"},
	{0x706388DC1F710AB5ULL, "6208177_m.fxe"},
	{0x7063B69F37D221FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\turn_left_180.gad"},
	{0x7063CCC61C15815EULL, "510493453.ogg"},
	{0x706420009DF44F35ULL, "244480887.ogg"},
	{0x706472936734690EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masf_0s.dds"},
	{0x70650387F6535057ULL, "6068261_m.fxe"},
	{0x7065F8D2EB2F8AD6ULL, "follower_banter6173323.cl"},
	{0x70660AD70A3A4A75ULL, "darktown_male_fer_urch_sta6174966.cl"},
	{0x7066AE9C551060A2ULL, "6110520_f.fxe"},
	{0x706709E83F39DAF9ULL, "art\\levels\\races\\design\\animation\\gasbarrel.inactive.gad"},
	{0x706713A04F661663ULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown3.mmh"},
	{0x70671CCE31C03A03ULL, "qun110_petrice6018755.cl"},
	{0x706721A04F662DBDULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown3.msh"},
	{0x70674E83FF865EF8ULL, "gen00fl_varric6118723.cl"},
	{0x706777319553A2B2ULL, "6121817_m.fxe"},
	{0x706798DAA595ADA9ULL, "hightown_fem_noble1_sta6169342.cl"},
	{0x7067A82AED20C163ULL, "vxlt_shop_f9_0_0.vlm"},
	{0x706847B6A4EDF0FAULL, "dae121ar_keep.ncs"},
	{0x706853DB7C812CE4ULL, "art\\vfx\\cinematic\\vfx_m_fire_pillar_imp_p.anb"},
	{0x7069161D0FFBD50BULL, "gen00fl_carver6102688.cl"},
	{0x7069164596F38FF2ULL, "hlac_ht_0_hightown_a3_-65538_0x.dds"},
	{0x70693827EA175AA3ULL, "og_flsh_02.dds"},
	{0x70699350D888B345ULL, "1698722.ogg"},
	{0x7069CD03F94881DCULL, "llac_ca_cavern1_haunted_24427_5.dds"},
	{0x7069FD105E000ABDULL, "hlac_ht_mansion2_818_28y.dds"},
	{0x706A3C173091CB1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop_low.ani"},
	{0x706A56BC750D84B9ULL, "vxlt_0_lowtown_exi_0_0.vlm"},
	{0x706A819A8E855B39ULL, "vxht_mansion_n_3uw_0_0.vlm"},
	{0x706B17E8AE742F97ULL, "6109436_m.fxe"},
	{0x706B63420B050579ULL, "6111149_m.fxe"},
	{0x706C0BF779A58B98ULL, "6142028_f.fxe"},
	{0x706C77595788D7DCULL, "855127781.ogg"},
	{0x706CDDAEB300D3FFULL, "vxlt_0_lowtown_gbo_0_0.vlm"},
	{0x706D1AE83F3E6347ULL, "art\\levels\\races\\design\\animation\\gasbarrel.inactive.ani"},
	{0x706DBBB88A388078ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravea_sa.dds"},
	{0x706DBE214C8DC15EULL, "follower_barks6153595.cl"},
	{0x706DCAF54F2D09D9ULL, "zz_cutscene_crash_test.ncs"},
	{0x706DDCDB886C9D17ULL, "6076449_f.fxe"},
	{0x706E508B4AA2F2DEULL, "lt_undercityencounter_mmap_ph.dds"},
	{0x706EC44E33496257ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_keepdoor.mmh"},
	{0x706EC49C4DB0BEEFULL, "vxlt_darktown_50l_0_0.vlm"},
	{0x706EC84E33496964ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_keepdoor.mao"},
	{0x706EDA4E334987C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_keepdoor.msh"},
	{0x706EEAEE564E4CC1ULL, "6175012_f.fxe"},
	{0x706F03A997D7AE4AULL, "vxw_bonepit_l2_0_0.vlm"},
	{0x706F17E6E86F95B3ULL, "gen00fl_anders6125395.cl"},
	{0x706F3FB6A4F401EBULL, "dae121ar_keep.lst"},
	{0x706F492D9C24D6F7ULL, "vxht_0_hightown_7u2_0_0.vlm"},
	{0x706F4F98E8F6D046ULL, "lgt200cr_gang_1_lt.utc"},
	{0x706F96FE95A917B4ULL, "vxht_0_hightown_n_ca3_0_1.vlm"},
	{0x706F996F59CCB63EULL, "hlac_lt_0_lowtown_a3_n_c_-262144_0y.dds"},
	{0x706FD4BF522390BEULL, "6030500_m.fxe"},
	{0x706FD5FF0A1C7696ULL, "6087736_f.fxe"},
	{0x70706CEA20AA2200ULL, "gen00fl_aveline6102079.cl"},
	{0x7070B1D51DB5BA39ULL, "6105846_f.fxe"},
	{0x7070E1A01D122B0CULL, "6048731_f.fxe"},
	{0x70713EBBC8AC8A35ULL, "qcr000_arishok6042450.cl"},
	{0x7071F58D0656BA08ULL, "6085898_m.fxe"},
	{0x7072434DF064F6E6ULL, "vxht_0_hightown_9xf_0_0.vlm"},
	{0x7072D85C9019BC1AULL, "vxdr_primevaltomb_6s5_0_0.vlm"},
	{0x7073B1AF04A0E3CEULL, "vxlt_0_lowtown_a2_n_c_i0s_0_0.vlm"},
	{0x707405A22D7EC30BULL, "157664891.ogg"},
	{0x707572DB7C8BEBE7ULL, "art\\vfx\\cinematic\\vfx_m_fire_pillar_imp_p.mmh"},
	{0x7075F5FDC86095B7ULL, "1044022415.ogg"},
	{0x70763FD4CD209BC9ULL, "gen00fl_varric6107306.cl"},
	{0x70765D2E8D998D5FULL, "w_woundedcoast_65_0.rml"},
	{0x7076DFE91D8DEEB1ULL, "552730199.ogg"},
	{0x707710F512A1049FULL, "6115312_m.fxe"},
	{0x70777E3742C1943CULL, "vxlt_0_lowtown_n_fi3_0_0.vlm"},
	{0x707795E45759DB03ULL, "6037936_f.fxe"},
	{0x7077FC494F00575BULL, "940892416.ogg"},
	{0x70782FF12597C7D9ULL, "vxca_caverns1_mines_j6e_0_0.vlm"},
	{0x7078E6233E478296ULL, "vxht_keep_4i4_0_0.vlm"},
	{0x7078ECD8B6B09CFCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_1.pwk"},
	{0x707981AE5F2E66B9ULL, "gen_im_trs_crc_lgt_hlm_r1.uti"},
	{0x707989C5A06A4817ULL, "6102536_m.fxe"},
	{0x7079DAFDA6176E80ULL, "97783554.ogg"},
	{0x7079E43F4FB19AADULL, "vxlt_darktown_4u1_0_1.vlm"},
	{0x7079E5D317AC98DCULL, "static_static_p102.psh.fxo.dx11"},
	{0x7079FC4EB779B4A5ULL, "vxw_encounter2_camp_5v_0_0.vlm"},
	{0x707A13BCB0DC18FFULL, "hlac_lt_0_lowtown_n_131069_0y.dds"},
	{0x707A3B4C7F1E7C1CULL, "gen00fl_merrill6115755.cl"},
	{0x707A8558F8B63C47ULL, "lgt220_mag_tmplr_templar6140562.cl"},
	{0x707AA0F479D072F3ULL, "vxw_blightlands_h0_0_0.vlm"},
	{0x707B77C3F67D7479ULL, "vxlt_0_lowtown_a2_n_c_gon_0_1.vlm"},
	{0x707B87DADB16F1D0ULL, "731875348.ogg"},
	{0x707BA15436326BE1ULL, "vxca_caverns1_mines_eht_0_0.vlm"},
	{0x707BB178F949B217ULL, "hightown_fem_noble2_mob6169468.cl"},
	{0x707C2600EA39335FULL, "art\\vfx\\combat\\vfx_glb_ele_s_extreme_c.impa.gad"},
	{0x707D36793FA6D76FULL, "vxlt_darktown_55o_0_0.vlm"},
	{0x707D39692B7AF663ULL, "6171606_m.fxe"},
	{0x707D54CC4B891BF3ULL, "l_and211ar_andersshop.wbd"},
	{0x707DE38400EC4B15ULL, "llac_lt_0_lowtown_a2_n_65536_0.dds"},
	{0x707E51BD86760780ULL, "716077798.ogg"},
	{0x707E5A0C62073DD9ULL, "807148445.ogg"},
	{0x707EA1C8BA7DF6BCULL, "250561388.ogg"},
	{0x707EBD55A809A2CEULL, "llac_w_encounter2_camp_131076_0.dds"},
	{0x707F9D1A6045AD8EULL, "character_character_p157.psh.fxo.dx11.assembly"},
	{0x707FE923A6080B20ULL, "548890592.ogg"},
	{0x7080465BFD5B23C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_loop.ani"},
	{0x708070CFEA33CBC9ULL, "6064740_m.fxe"},
	{0x7080EC3088AEDC74ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb104a_0d.dds"},
	{0x7081155BFF546616ULL, "hlac_lt_0_lowtown_-65537_3y.dds"},
	{0x7081BE3E898DFE44ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkplain.msh"},
	{0x7081C7761C03F779ULL, "6200890_m.fxe"},
	{0x7081D03E898E1CADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkplain.mao"},
	{0x7081D43E898E23E6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkplain.mmh"},
	{0x7081F43219F9FE4AULL, "744488096.ogg"},
	{0x7082468397F6BCDDULL, "6059238_m.fxe"},
	{0x70826CD2A9989B43ULL, "139257664.ogg"},
	{0x7082B2601DC306C8ULL, "6068983_m.fxe"},
	{0x70831B00EA3F3F4CULL, "art\\vfx\\combat\\vfx_glb_ele_s_extreme_c.impa.evt"},
	{0x7084B8603881FA29ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_dsmla_0d.dds"},
	{0x7084FA0CD59302D2ULL, "299526544.ogg"},
	{0x708540416D81021CULL, "llac_ht_0_hightown_a3_-65536_2.dds"},
	{0x7085C24E6610158EULL, "252354405.ogg"},
	{0x7085DE1A526C2859ULL, "vxw_encounter2_ph_0_1.vlm"},
	{0x7085F041B9CCDDD7ULL, "6188624_m.fxe"},
	{0x7086007D0C9CF132ULL, "gen_im_cft_reg_b2.uti"},
	{0x70860657B710CA8CULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.cess.ani"},
	{0x708608340D6FB2EEULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.cess.ani"},
	{0x708619D678D20785ULL, "6120422_m.fxe"},
	{0x70862FDBEF116BAEULL, "art\\characters\\creatures\\mubari\\animation\\eat2ready.evt"},
	{0x7086615BFD5FBD4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_loop.gad"},
	{0x7086A362711F06A6ULL, "240193018.ogg"},
	{0x7086B7E5810D7FE6ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_alis.phy"},
	{0x7086C57C05CC1160ULL, "222424404.ogg"},
	{0x7086E61F20D94EAAULL, "vxca_caverns2_7f8_0_1.vlm"},
	{0x7086EB834B6D4F99ULL, "128822266.ogg"},
	{0x7086EBE4F405BFAEULL, "hla_ht_0_hightown_1_0y.dds"},
	{0x70875800303D9B16ULL, "vxht_keep_4n5_0_0.vlm"},
	{0x70878FCFB0049EA9ULL, "37701322.ogg"},
	{0x708912D16BC68B73ULL, "vxw_blightlands_m3_0_0.vlm"},
	{0x70893B3C7E4A3B44ULL, "138369970.ogg"},
	{0x708968BDBBE063CCULL, "vxca_caverns2prc_62k_0_0.vlm"},
	{0x70899C02BE955FCCULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_0x.dds"},
	{0x7089AA5BEC455FFCULL, "vxlt_0_lowtown_n_e3b_0_0.vlm"},
	{0x7089E52F45D3A744ULL, "hlac_lt_hangedman_-1_3x.dds"},
	{0x708A2D4229770317ULL, "al_isabela_s01.gda"},
	{0x708A69E2CAB3DD6BULL, "6110962_m.fxe"},
	{0x708AEDC9DF8CAAA7ULL, "313527426.ogg"},
	{0x708B4FC95167A53BULL, "6182020_m.fxe"},
	{0x708BA111ADCA0C6BULL, "lowtown_fem_alienage1_sta6162569.cl"},
	{0x708BCB3E899668B1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkplain.phy"},
	{0x708BFFBC42D9C046ULL, "vxlt_gallowstemplar_f_4yd_0_0.vlm"},
	{0x708CACC954ED0A8DULL, "240757950.ogg"},
	{0x708CB7B7EE83E09FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopening_rog.gad"},
	{0x708CC70432F7191BULL, "vxdr_slavers_1bd_0_0.vlm"},
	{0x708D1357B716FFCEULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.cess.gad"},
	{0x708D18BEFF9FEE6EULL, "dae300_alistair_king6133806.cl"},
	{0x708D26DBEF177B59ULL, "art\\characters\\creatures\\mubari\\animation\\eat2ready.gad"},
	{0x708D5E5BFD65D6B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_female_loop.evt"},
	{0x708DD8F718EE743DULL, "6174359_m.fxe"},
	{0x708DE314A2DE868EULL, "fen210_orana6133063.cl"},
	{0x708E1F7C16189B01ULL, "qun221_petrice6077351.cl"},
	{0x708E58F2D534AC3AULL, "vxlt_0_lowtown_e2t_0_0.vlm"},
	{0x708EAC46997EC0A9ULL, "6206626_f.fxe"},
	{0x708F1E22207ED125ULL, "vxlt_foundry_n_106_0_0.vlm"},
	{0x708F90A488682A27ULL, "6110401_m.fxe"},
	{0x70900300EA49A0A5ULL, "art\\vfx\\combat\\vfx_glb_ele_s_extreme_c.impa.ani"},
	{0x7090779F111A49A6ULL, "art\\vfx\\combat\\vfx_c_buff_act_c.impa.ani"},
	{0x709094B1DA15042BULL, "vxht_bartrand_n_639_0_0.vlm"},
	{0x7090BFBC482127FEULL, "47649712.ogg"},
	{0x7091792156B73A8DULL, "166791924.ogg"},
	{0x70918BC07B4B6469ULL, "6180451_m.fxe"},
	{0x7091B435AEDE60BDULL, "20785300.ogg"},
	{0x7091C3FC2A5FE999ULL, "480519314.ogg"},
	{0x709225E0730B5FF1ULL, "vxht_0_hightown_6pb_0_0.vlm"},
	{0x70928EE49027C1CAULL, "cs_pro000cs_real_surv.wbd"},
	{0x70929D0233B555F5ULL, "6070983_m.fxe"},
	{0x709300340D7A2F3FULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.cess.evt"},
	{0x7093240E30297C4DULL, "403671387.ogg"},
	{0x7093C8C71336609FULL, "hlac_w_bonepit_393217_0z.dds"},
	{0x70940E57B71D15ADULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.cess.evt"},
	{0x709437DBEF1DB6A7ULL, "art\\characters\\creatures\\mubari\\animation\\eat2ready.ani"},
	{0x70944BDB1B2E1D0AULL, "llac_w_bonepitdestroyed_a3_458755_0.dds"},
	{0x70947A1F80796570ULL, "vxw_bonepitmining_a2_1pm_0_0.vlm"},
	{0x7094806CF4DD88FFULL, "mrl150_hunter6071010.cl"},
	{0x70948230ACB2FC39ULL, "732155996.ogg"},
	{0x70948937701C3A31ULL, "dae000cr_grand_cleric6182018.cl"},
	{0x70948D6738AF167CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02plank.mao"},
	{0x70948F221C0412E9ULL, "465071644.ogg"},
	{0x7094996738AF2A1FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02plank.mmh"},
	{0x70949F6738AF34C1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02plank.msh"},
	{0x70951641A499DDD2ULL, "gen00fl_isabela6135636.cl"},
	{0x709596E2DCEBAC75ULL, "dre000_hireling16038654.cl"},
	{0x7095ADD4FE79C95AULL, "vxlt_0_lowtown_a2_n_c_exs_0_0.vlm"},
	{0x7095DC0920EE093EULL, "90183874.ogg"},
	{0x709682C14A18E761ULL, "dae000_merchant_mage_hf.bnk"},
	{0x7096A2D7981C2122ULL, "245519589.ogg"},
	{0x7096E93078BDBEFBULL, "548731388.ogg"},
	{0x7097057DC24D6649ULL, "vxlt_0_lowtown_a3_n_c_hdl_0_0.vlm"},
	{0x70975A24DCBFC654ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney_n.dds"},
	{0x709784AE5DBCA3F3ULL, "vxw_blightlands_v2_0_0.vlm"},
	{0x7097A87A7AEF5B74ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston.phy"},
	{0x70982A91B8388544ULL, "vxdr_0_deeproads_9cf_0_0.vlm"},
	{0x709879D27C6AEC12ULL, "hlac_lt_0_lowtown_docks_131072_1y.dds"},
	{0x7098850E1A1E9CE1ULL, "vxca_caverns1_mines_ebj_0_0.vlm"},
	{0x7098B35B5A0538F6ULL, "fxa_web_03.mao"},
	{0x7098B76C7BE141A7ULL, "vxht_0_hightown_c81_0_0.vlm"},
	{0x70990643420077F2ULL, "vxca_cavern1_haunted_gez_0_0.vlm"},
	{0x709970C9A5E2F807ULL, "6165005_m.fxe"},
	{0x7099A4C14A1B56D1ULL, "dae000_merchant_mage_hf.cnv"},
	{0x7099E1340D801928ULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.cess.gad"},
	{0x709A155B970777E7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_green_d.dds"},
	{0x709AA7F7D6AB05E3ULL, "genip_cin_glass.utp"},
	{0x709B07B921ABD716ULL, "6103648_player_intense.cut"},
	{0x709B50B6A5195908ULL, "dae121ar_keep.are"},
	{0x709B573837EF7412ULL, "hlac_lt_0_lowtown_docks_n_-3_4y.dds"},
	{0x709B93ADBA276E31ULL, "t3_rob_apo.tnt"},
	{0x709BB3D68185C2B4ULL, "6082020_m.fxe"},
	{0x709C2D0807A7ED25ULL, "190018998.ogg"},
	{0x709C46904A7A92FAULL, "hlac_lt_0_lowtown_-2_4x.dds"},
	{0x709D5F9F1124AB17ULL, "art\\vfx\\combat\\vfx_c_buff_act_c.impa.evt"},
	{0x709DB414138004FEULL, "65695094.ogg"},
	{0x709DC464CB9927D6ULL, "llac_lt_0_lowtown_1_0.dds"},
	{0x709DC614EA0F8BBCULL, "fen210_tevinter_ambusher2.cnv"},
	{0x709DE69F05578FEDULL, "370510841.ogg"},
	{0x709DF2448BE89F15ULL, "6068435_f.fxe"},
	{0x709E23944179549DULL, "gen00fl_merrill6123726.cl"},
	{0x709E9FEB8CA359B6ULL, "6071073_m.fxe"},
	{0x709EA59BB48B2333ULL, "628735176.ogg"},
	{0x709EA79829C87046ULL, "vxlt_gallowscourtyard_n_74z_0_0.vlm"},
	{0x709EDAC53E7FCB35ULL, "m_one100ar_ruin_int.bnk"},
	{0x709F31689EB0E882ULL, "and100_anders6095857.cl"},
	{0x709F3C14FEAC0009ULL, "108734921.ogg"},
	{0x709F9FBDE36C4AD4ULL, "847009563.ogg"},
	{0x709FA24676B4EF74ULL, "ser191_emeric6120419.cl"},
	{0x709FAA1EE87BCEA5ULL, "vxca_caverns1_j2t_0_0.vlm"},
	{0x709FD049238B332DULL, "93545212.ogg"},
	{0x70A04F36C0E3E495ULL, "one140cr_charade6139434.cl"},
	{0x70A055E832999680ULL, "vxw_encounter2_camp_nj_0_0.vlm"},
	{0x70A158F4F7788178ULL, "6150068_m.fxe"},
	{0x70A1857A7AF77471ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston.msh"},
	{0x70A1937A7AF78B8CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston.mao"},
	{0x70A194B7EE9601E5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopening_rog.ani"},
	{0x70A19F7A7AF7A02FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston.mmh"},
	{0x70A1A6222645E46FULL, "mag100_karras6035372.cl"},
	{0x70A1E814EA13AE0CULL, "fen210_tevinter_ambusher2.bnk"},
	{0x70A2DDDA74951A4AULL, "vxca_caverns2prc_9fz_0_0.vlm"},
	{0x70A420789DB07E15ULL, "static_static_p77.psh.fxo.dx11"},
	{0x70A4609F112ACBA0ULL, "art\\vfx\\combat\\vfx_c_buff_act_c.impa.gad"},
	{0x70A59886E9A81CADULL, "6033407_m.fxe"},
	{0x70A5B26D323882C2ULL, "lgt360pt_nt_rescue.ncs"},
	{0x70A5F262726A5E79ULL, "llac_lt_0_lowtown_a2_n_0_7.dds"},
	{0x70A5F7EE1AFA0408ULL, "6043988_m.fxe"},
	{0x70A6305B6128DC2BULL, "vxht_mansion2b_2ic_0_0.vlm"},
	{0x70A69C6EAA2E9DC4ULL, "vxdr_0_deeproads_9dg_0_0.vlm"},
	{0x70A6E6D5D98F9895ULL, "vxdr_primevaltomb_2up_0_0.vlm"},
	{0x70A732C3ABDB5660ULL, "282004105.ogg"},
	{0x70A7330227812840ULL, "134445713.ogg"},
	{0x70A73B8183228B3EULL, "911416910.ogg"},
	{0x70A76BA5F09C9623ULL, "124552905.ogg"},
	{0x70A770F84AC13E22ULL, "zz_qun_debug6038055.cl"},
	{0x70A8A934E9B8F74BULL, "6075803_m.fxe"},
	{0x70A8C492B05C9873ULL, "6189906_m.fxe"},
	{0x70A9191DE2414888ULL, "w_bonepitdestroyed_a3_165_0.msh"},
	{0x70A92B1DE2416719ULL, "w_bonepitdestroyed_a3_165_0.mao"},
	{0x70A935122FD73ABDULL, "mag210_arianni6024989.cl"},
	{0x70A9371DE2417BB2ULL, "w_bonepitdestroyed_a3_165_0.mmh"},
	{0x70A99C06667D11A8ULL, "1027907147.ogg"},
	{0x70AA457104FFF5BBULL, "mag320_orsino6071103.cl"},
	{0x70AAC50F9E396A25ULL, "vxht_chantry_n_35b_0_1.vlm"},
	{0x70AADA1C204BA3B8ULL, "vxht_keep_n_6h5_0_1.vlm"},
	{0x70AB2BAC1059ED9CULL, "gen00fl_aveline6122125.cl"},
	{0x70AB418720292604ULL, "vxlt_0_lowtown_docks_a3_cys_0_0.vlm"},
	{0x70ABA4243F39AF50ULL, "w_bonepit_190_0.msh"},
	{0x70ABB12684A87A36ULL, "371162125.ogg"},
	{0x70ABB2243F39C76AULL, "w_bonepit_190_0.mmh"},
	{0x70ABB6243F39CDA1ULL, "w_bonepit_190_0.mao"},
	{0x70AC4CD7039E17B8ULL, "6096472_m.fxe"},
	{0x70AC7A25E5CEE8A6ULL, "326231182.ogg"},
	{0x70AC9218C181E0AAULL, "1010596314.ogg"},
	{0x70AD1153222EE175ULL, "hlac_dr_primevaltomb_661_4x.dds"},
	{0x70AD628B7AEA5CF8ULL, "zz_amb_moe.utc"},
	{0x70AD93FE7D261380ULL, "vxca_cavern1_haunted_k6u_0_0.vlm"},
	{0x70ADC3AA6991F500ULL, "rdr110dg_dockworker6015926.cl"},
	{0x70ADDA1D751BC930ULL, "832536333.ogg"},
	{0x70AE3EDCE20C548AULL, "vxca_caverns2_6hj_0_0.vlm"},
	{0x70AE41BB472E9209ULL, "223141510.ogg"},
	{0x70AE589004B35352ULL, "6047537_m.fxe"},
	{0x70AEBAFEA4E47F81ULL, "mag291_ntbanter6166385.cl"},
	{0x70AEC7C5248FAF00ULL, "vxw_encounter2_camp_ik_0_0.vlm"},
	{0x70AF9E911C36A0F8ULL, "410099377.ogg"},
	{0x70AFCAD1E96E99F8ULL, "6157168_m.fxe"},
	{0x70B0538284124EA9ULL, "746813237.ogg"},
	{0x70B07FA0CD696148ULL, "vxca_caverns1_mines_egl_0_0.vlm"},
	{0x70B1094FC951E4CFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_noba_0t.dds"},
	{0x70B1C9B1175728DEULL, "vxht_0_hightown_a3_c8q_0_0.vlm"},
	{0x70B1CF47EC2281BEULL, "6180009_m.fxe"},
	{0x70B2166BAC9CD58FULL, "813931581.ogg"},
	{0x70B2A2949DC01852ULL, "967132710.ogg"},
	{0x70B32E45258ACEFAULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonb_head.mao"},
	{0x70B34220953F2F3DULL, "vxlt_warehouse_n_11c_0_0.vlm"},
	{0x70B3585107A96D04ULL, "241531299.ogg"},
	{0x70B3925B9DF7D320ULL, "65203121.ogg"},
	{0x70B3D3232D41F5C4ULL, "vxlt_0_lowtown_n_feg_0_0.vlm"},
	{0x70B3E937A6399749ULL, "vxlt_0_lowtown_a3_n_c_hnr_0_0.vlm"},
	{0x70B415EF37C4EB0BULL, "hlac_w_bonepitdestroyed_a3_196608_0y.dds"},
	{0x70B44338A62ECE6BULL, "vxw_sundermount_30x_0_1.vlm"},
	{0x70B4497397952A3AULL, "901634933.ogg"},
	{0x70B507031BC79443ULL, "6067314_m.fxe"},
	{0x70B5073443D56E44ULL, "hlac_ht_mansion2_818_55x.dds"},
	{0x70B541340C174FEFULL, "vxw_woundedcoast_1r5_0_0.vlm"},
	{0x70B56A41E2843598ULL, "vxlt_gallowstemplar_n_5qo_0_0.vlm"},
	{0x70B59B26A9BA99F5ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyo_0s.dds"},
	{0x70B6DBCDC2B9E40CULL, "vxht_0_hightown_a3_bgk_0_0.vlm"},
	{0x70B6FD8163D9B37FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ready.gad"},
	{0x70B74E9188698B2DULL, "vxht_0_hightown_a3_a4p_0_0.vlm"},
	{0x70B76B1D7A204AF8ULL, "880119759.ogg"},
	{0x70B7B34BECF264C0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04b.phy"},
	{0x70B89B8018E9D6DBULL, "hlac_ht_bartrand_n_3373_0x.dds"},
	{0x70B913A360E6D515ULL, "vxht_playershouse_n_1zq_0_0.vlm"},
	{0x70B91A43039DF47AULL, "llac_ht_mansion2_818_37.dds"},
	{0x70B9424974D94429ULL, "ter_brickround01_d.dds"},
	{0x70B968B5920762F9ULL, "art\\characters\\playercharacter\\elfmale\\rob_enco.msh"},
	{0x70B972B5920773E7ULL, "art\\characters\\playercharacter\\elfmale\\rob_enco.mmh"},
	{0x70B976B592077A34ULL, "art\\characters\\playercharacter\\elfmale\\rob_enco.mao"},
	{0x70B9857DFBC445C5ULL, "hlac_ht_mansion2_818_36z.dds"},
	{0x70BA1A14442DE202ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_fern.msh"},
	{0x70BA1C14442DE5B8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_fern.mmh"},
	{0x70BA52A42F704F1DULL, "vxlt_warehouse_n_9x_0_0.vlm"},
	{0x70BA5E12FCD34E7EULL, "dae000_male_commoner2_mob6174422.cl"},
	{0x70BA7961FDA34A14ULL, "793756034.ogg"},
	{0x70BAA6A3D54F2D65ULL, "mcr310_party_banter6087781.cl"},
	{0x70BAC46D6A641D01ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medp_t.dds"},
	{0x70BB460C3F499460ULL, "llac_ht_0_hightown_65536_2.dds"},
	{0x70BB6AA7C9838389ULL, "vxht_keep_5hm_0_0.vlm"},
	{0x70BBB98447EE0B74ULL, "469143500.ogg"},
	{0x70BBE7D840EF88E2ULL, "gen_im_wep_rog_dua_act2_01.uti"},
	{0x70BC96F01711FA36ULL, "vxht_chantry_n_2p2_0_0.vlm"},
	{0x70BCF09DFA31C9A5ULL, "gencr_mercenary_warrior_2hf.utc"},
	{0x70BCF28163DE0C6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ready.evt"},
	{0x70BD0A2F68E0D33BULL, "vfxchunky0.vsh.fxo"},
	{0x70BD29D19B33B442ULL, "gen00fl_anders6125473.cl"},
	{0x70BD4CB441394103ULL, "llac_ht_0_hightown_a3_65534_0.dds"},
	{0x70BD6A9445248127ULL, "6071160_m.fxe"},
	{0x70BE3E7DE0555182ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_8y.dds"},
	{0x70BE53EAE18314FCULL, "6156620_m.fxe"},
	{0x70BE75C47BF5FC83ULL, "lt_warehouse_1_0_ll.mmh"},
	{0x70BEF191F6CDFD1BULL, "dae000_gamlen6144342.cl"},
	{0x70BEF7C7053F942EULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_act_c.impa.evt"},
	{0x70BF27BFCCE2DCF4ULL, "hlac_lt_0_lowtown_a3_n_c_65537_7z.dds"},
	{0x70BFCB0613165DB4ULL, "vxca_caverns2prc_ag3_0_0.vlm"},
	{0x70C07B076CD51C65ULL, "vxlt_0_lowtown_a2_n_c_emw_0_0.vlm"},
	{0x70C0935A5E52A6EBULL, "vxlt_0_lowtown_docks_c04_0_0.vlm"},
	{0x70C0B249708BB605ULL, "mag330_walter6045975.cl"},
	{0x70C19DEA3DC81430ULL, "vxdr_primevaltomb_7yl_0_0.vlm"},
	{0x70C1AE00E3D61FB6ULL, "lt_0_lowtown_docks_190_0.msh"},
	{0x70C1B800E3D630B4ULL, "lt_0_lowtown_docks_190_0.mmh"},
	{0x70C1BC00E3D637C7ULL, "lt_0_lowtown_docks_190_0.mao"},
	{0x70C25B14982FAFC9ULL, "vxlt_0_lowtown_a3_n_c_hkq_0_0.vlm"},
	{0x70C27C40CD944DADULL, "6039607_m.fxe"},
	{0x70C336BDD9C2A39AULL, "vxw_bonepit_mm_0_0.vlm"},
	{0x70C33D07B9B81BCDULL, "vxlt_0_lowtown_a2_n_byn_0_0.vlm"},
	{0x70C39500F0BC7FDBULL, "vxlt_gallowstemplar_f_6si_0_0.vlm"},
	{0x70C3A00A73AEFC85ULL, "6016716_m.fxe"},
	{0x70C3A9EFFA29D454ULL, "hlac_lt_0_lowtown_docks_n_-65539_2x.dds"},
	{0x70C48A2C5EB54B37ULL, "party_stash.utm"},
	{0x70C533887F15A5BAULL, "vxlt_0_lowtown_ha8_0_0.vlm"},
	{0x70C5512E8DDD4D4FULL, "w_woundedcoast_65_0.mao"},
	{0x70C55D2E8DDD61ECULL, "w_woundedcoast_65_0.mmh"},
	{0x70C5632E8DDD6BDEULL, "w_woundedcoast_65_0.msh"},
	{0x70C5D2373129C23BULL, "hlac_ht_mansion2_818_62y.dds"},
	{0x70C5EEC70545A3D9ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_act_c.impa.gad"},
	{0x70C60CD2C1BD932DULL, "628539868.ogg"},
	{0x70C6867E873A5E5EULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tn90r.evt"},
	{0x70C6B65F593A498CULL, "woodrope_n.dds"},
	{0x70C6DFA79A909CCEULL, "6111083_m.fxe"},
	{0x70C6EC7002147B8CULL, "vxht_keep_55g_0_0.vlm"},
	{0x70C740624AC3CCF4ULL, "994846040.ogg"},
	{0x70C760C0A2826E6CULL, "6023890_m.fxe"},
	{0x70C7858052DCED95ULL, "vxht_playershouse_n_1qp_0_0.vlm"},
	{0x70C8324F61EED7F7ULL, "plt_zz_mag200pt_nt_debug.nss"},
	{0x70C89775491038B3ULL, "vxlt_backalley_n_1y4_0_0.vlm"},
	{0x70C8C391A8D26DF5ULL, "hlac_w_encounter2_65541_0x.dds"},
	{0x70C90872419BC1D0ULL, "vxlt_blackemporium_3_0_0.vlm"},
	{0x70C983EB750F1508ULL, "follower_banter6170566.cl"},
	{0x70CA335D41F3B484ULL, "vxca_caverns2_7o8_0_0.vlm"},
	{0x70CA933D5F06E6C2ULL, "6164577_m.fxe"},
	{0x70CA980F6ED623B0ULL, "hlac_dr_0_deeproads_5845_7y.dds"},
	{0x70CAD2B8B3A82EC4ULL, "dae000_arishok6080510.cl"},
	{0x70CADA8163EA20C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\ready.ani"},
	{0x70CB00C53EA546BFULL, "m_one100ar_ruin_int.wbd"},
	{0x70CB59F3070C04F7ULL, "dae000ip_rename.ncs"},
	{0x70CB6DDA9404EA99ULL, "6024128_m.fxe"},
	{0x70CBF1EDEB34F21DULL, "164807487.ogg"},
	{0x70CCBC7F5ADF6124ULL, "rdr110dg_guard_warehouse6174113.cl"},
	{0x70CCCCBE8249FFAAULL, "hlac_lt_undercitydungeon_-1_27x.dds"},
	{0x70CCDC3D3C3B53DEULL, "gen00fl_aveline6113619.cl"},
	{0x70CCEB5633D583DCULL, "hlac_ht_0_hightown_65536_2z.dds"},
	{0x70CCFFC7054BDF27ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_act_c.impa.ani"},
	{0x70CD23AB503CCE6EULL, "505312260.ogg"},
	{0x70CD3F7F68EFD4B7ULL, "517087847.ogg"},
	{0x70CE9224D6605943ULL, "6037122_m.fxe"},
	{0x70CEA1D31D6FFB73ULL, "699749009.ogg"},
	{0x70CEAD6AFB4359DEULL, "vxht_0_hightown_a3_c6s_0_0.vlm"},
	{0x70CF0A580151027FULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.dura.gad"},
	{0x70D008AB0DEA9A51ULL, "6133574_m.fxe"},
	{0x70D021EB19AA36BDULL, "6027806_m.fxe"},
	{0x70D04F907831108CULL, "6088404_m.fxe"},
	{0x70D07921C721CD8AULL, "llac_lt_0_lowtown_docks_n_-196608_0.dds"},
	{0x70D0F5E9A0D4BB0FULL, "dae000_junar6147938.cl"},
	{0x70D2BF209831827EULL, "6125718_m.fxe"},
	{0x70D2E70B80091696ULL, "follower_banter6141645.cl"},
	{0x70D328B7A83A4EFEULL, "ser270_emeric6149276.cl"},
	{0x70D436948FB3659CULL, "vxht_keep_6gv_0_0.vlm"},
	{0x70D4913610D1382FULL, "6132931_m.fxe"},
	{0x70D4A7C5E7612F03ULL, "dae000_tranquil_16165105.cl"},
	{0x70D4FF5801555B6CULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.dura.evt"},
	{0x70D55E4CF40A940CULL, "vxht_keep_58d_0_0.vlm"},
	{0x70D55EF65263F198ULL, "6153858_m.fxe"},
	{0x70D5751493E67BB0ULL, "544863381.ogg"},
	{0x70D57EFBF7C19E3CULL, "ai_disable.ncs"},
	{0x70D58324D61C4E13ULL, "art\\vfx\\combat\\vfx_r_var_bayonet_v5_c.impa.ani"},
	{0x70D58FB7F1D3CCAFULL, "dae000_arishok6068492.cl"},
	{0x70D69C3129953640ULL, "gen00fl_carver6108940.cl"},
	{0x70D737A203D6ACF6ULL, "271530242.ogg"},
	{0x70D751DED0886086ULL, "6129145_m.fxe"},
	{0x70D75E34EB1B2380ULL, "vxdr_slavers_2ie_0_0.vlm"},
	{0x70D777ECCC92A67EULL, "gen00fl_merrill6127279.cl"},
	{0x70D846E9FDEDE1E7ULL, "6172801_m.fxe"},
	{0x70D8AF750A91778EULL, "vxw_encounter2_qu_0_0.vlm"},
	{0x70D8CA25A0B3057DULL, "6162653_m.fxe"},
	{0x70D8DD279AC78C0FULL, "dae000_arishok6075205.cl"},
	{0x70D8E604FFC38701ULL, "53138278.ogg"},
	{0x70D998D441FD4C4FULL, "6129236_m.fxe"},
	{0x70DA29B6F245F78EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow8base.phy"},
	{0x70DA325645E9F029ULL, "6067755_f.fxe"},
	{0x70DA6CE490653DE0ULL, "cs_pro000cs_real_surv.bnk"},
	{0x70DAC6A3AF620AEBULL, "187437717.ogg"},
	{0x70DB0E496E2C5B33ULL, "6076297_m.fxe"},
	{0x70DB52A1BC622F09ULL, "vxht_0_hightown_c76_0_0.vlm"},
	{0x70DB970290EFDDC6ULL, "6101367_m.fxe"},
	{0x70DBF3D5E484E28EULL, "60902894.ogg"},
	{0x70DBF98DA4FEE161ULL, "vxht_0_hightown_a2_bam_0_0.vlm"},
	{0x70DC442D68911EFFULL, "hlac_lt_0_lowtown_a2_-262143_0y.dds"},
	{0x70DC7A258C94F543ULL, "vxlt_gallowstemplar_n_356_0_0.vlm"},
	{0x70DC847195D07F8EULL, "6112070_f.fxe"},
	{0x70DCD04BED11B255ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04b.msh"},
	{0x70DCD24BED11B5CBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04b.mmh"},
	{0x70DD70D5398FF548ULL, "hlac_lt_0_lowtown_-3_7x.dds"},
	{0x70DDEBBA7842CC02ULL, "vxw_encounter1_jd_0_1.vlm"},
	{0x70DE0584974A09C0ULL, "llac_ca_caverns1_24426_10.dds"},
	{0x70DE22F27B0F616AULL, "309747831.ogg"},
	{0x70DE36C0A1658498ULL, "zz_dre_debug6092009.cl"},
	{0x70DEBC0442413E9BULL, "vxw_encounter2_camp_y_0_0.vlm"},
	{0x70DF24CC0ED9040CULL, "6144087_m.fxe"},
	{0x70DF6762A0CA47A4ULL, "llac_lt_0_lowtown_a2_-1_1.dds"},
	{0x70DFAA79AA39356CULL, "hlac_ht_0_hightown_a3_-196610_0x.dds"},
	{0x70E01D6ABBC00DA6ULL, "6037308_f.fxe"},
	{0x70E0E4571B7B875FULL, "war_bloodfrenzy.dds"},
	{0x70E0EE545FB63ED2ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.gad"},
	{0x70E11D65E8436722ULL, "zz_dae_debug6029477.cl"},
	{0x70E1D64F2E52A554ULL, "vxlt_ruinedhouse_n_hl_0_0.vlm"},
	{0x70E2008BEB8BDB04ULL, "6179599_m.fxe"},
	{0x70E2D3F610D89C3EULL, "426302144.ogg"},
	{0x70E2E0E3FD689B1BULL, "vxw_bonepit_29t_0_0.vlm"},
	{0x70E2E75801616FC5ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_lp_aoe_p.dura.ani"},
	{0x70E2F757443697ADULL, "6052598_m.fxe"},
	{0x70E3497E016323B8ULL, "707469401.ogg"},
	{0x70E36C1979A34E6DULL, "6096370_m.fxe"},
	{0x70E37B24D6287E4AULL, "art\\vfx\\combat\\vfx_r_var_bayonet_v5_c.impa.evt"},
	{0x70E3BB6FFC741991ULL, "art\\vfx\\models\\vfx_spin.mmh"},
	{0x70E3D56FFC74464FULL, "art\\vfx\\models\\vfx_spin.msh"},
	{0x70E41D892F41EEA4ULL, "357088285.ogg"},
	{0x70E4563C308FD53EULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_newmetal_n.dds"},
	{0x70E480D740B65611ULL, "vxw_encounter2_3h_0_0.vlm"},
	{0x70E515F4D14E3BB5ULL, "617477904.ogg"},
	{0x70E55BB5922C871CULL, "art\\characters\\playercharacter\\elfmale\\rob_enco.phy"},
	{0x70E563EA1578C663ULL, "vxdr_primevaltomb_56d_0_0.vlm"},
	{0x70E570DCC9D06EABULL, "141480036.ogg"},
	{0x70E5883F4DECFC2FULL, "one140cr_charade6139429.cl"},
	{0x70E5C849FFDB0519ULL, "209095747.ogg"},
	{0x70E5CD0222331645ULL, "dre000_vampire6034301.cl"},
	{0x70E667B5E5B873DDULL, "mag130cr_slaver.utc"},
	{0x70E6A27AA42D2CD0ULL, "799127290.ogg"},
	{0x70E6F1545FBAAF09ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.evt"},
	{0x70E7251CC068204BULL, "vxlt_gallowstemplar_n_4u1_0_0.vlm"},
	{0x70E74F0EA81A758FULL, "347966725.ogg"},
	{0x70E752A97DC673C1ULL, "vxht_mansion2_3ep_0_0.vlm"},
	{0x70E7579B93BE8C5EULL, "6175782_m.fxe"},
	{0x70E801239DD0FE14ULL, "dae000_ineria6137187.cl"},
	{0x70E872DC492E4659ULL, "vxht_mansion2_33r_0_0.vlm"},
	{0x70E8CE3EB58CFA15ULL, "964112464.ogg"},
	{0x70E94D118AA775D2ULL, "229723218.ogg"},
	{0x70E95296D8E1507DULL, "qun131_taskmaster6059075.cl"},
	{0x70EA1F09FB19BD89ULL, "hlac_lt_0_lowtown_a2_n_-65537_6x.dds"},
	{0x70EA5A24D62E64BDULL, "art\\vfx\\combat\\vfx_r_var_bayonet_v5_c.impa.gad"},
	{0x70EA8B4E7A5BA463ULL, "25988485.ogg"},
	{0x70EC4C29B7890AF8ULL, "root_rog_dual.dds"},
	{0x70ED1DAFDB130C5FULL, "vxht_mansion_n_34b_0_0.vlm"},
	{0x70ED44E9ECF1799AULL, "gen_im_trs_chn_blt_r2.uti"},
	{0x70ED6A0F4CA0ECFAULL, "isa211_qunari_ambush6081813.cl"},
	{0x70EDDCA615CBC258ULL, "art\\characters\\creatures\\rubblegolem\\c_profane.mao"},
	{0x70EDFB9C8A853D91ULL, "6089914_m.fxe"},
	{0x70EE0BFDDCCC1AC2ULL, "hlac_ht_0_hightown_a3_65534_1z.dds"},
	{0x70EE5EA0D23EA8F8ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_nrmi_0n.dds"},
	{0x70EE71FD7BF642D3ULL, "6131116_m.fxe"},
	{0x70EF3B6917BC7AEBULL, "hlac_lt_foundry_n_320_13z.dds"},
	{0x70EFB59E105B7142ULL, "vxlt_0_lowtown_n_eoi_0_1.vlm"},
	{0x70EFEFAE32A691BCULL, "6036768_m.fxe"},
	{0x70F0B3EB7F382CCDULL, "dae300_alistair_king6134965.cl"},
	{0x70F0DB6CCC518501ULL, "641413271.ogg"},
	{0x70F1537AE4BBEBDAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basicwood.mmh"},
	{0x70F1577AE4BBF211ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basicwood.mao"},
	{0x70F1655BB893C6ACULL, "915164600.ogg"},
	{0x70F1657AE4BC0A20ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basicwood.msh"},
	{0x70F221733E78A258ULL, "677078506.ogg"},
	{0x70F24B8BA6A08A36ULL, "gen00fl_fenris6128025.cl"},
	{0x70F25AA65CF11929ULL, "ca_caverns1_mines_26035_lc.mmh"},
	{0x70F26CF910DCB80AULL, "hlac_ca_caverns1_mines_24426_13x.dds"},
	{0x70F273008DD58F09ULL, "vxlt_gallowstemplar_458_0_0.vlm"},
	{0x70F2D0D2E722D7F7ULL, "54084235.ogg"},
	{0x70F3334C1002B218ULL, "llac_ca_caverns1_mines_24421_9.dds"},
	{0x70F466C3DDEC1EF2ULL, "6175315_m.fxe"},
	{0x70F4A63713B55B27ULL, "zz_one_debug6079301.cl"},
	{0x70F4B2B040FE67A2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicwood.phy"},
	{0x70F4D9545FC6C478ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.ani"},
	{0x70F4DDD8DFAEC6ECULL, "vxlt_0_lowtown_a2_n_e25_0_0.vlm"},
	{0x70F4DE77B2C1885FULL, "hlac_w_blightlands_262145_0x.dds"},
	{0x70F4FA601948F383ULL, "vxlt_0_lowtown_docks_b4i_0_0.vlm"},
	{0x70F4FD9AE5096175ULL, "6141517_m.fxe"},
	{0x70F554F700E2D70EULL, "war_ignoreminorhits.dds"},
	{0x70F5BD3F25E40471ULL, "vxlt_0_lowtown_exq_0_0.vlm"},
	{0x70F68195F4B3788EULL, "6172596_m.fxe"},
	{0x70F6895ADDB485FFULL, "clf260_fell_orden6088780.cl"},
	{0x70F6A3212EB89255ULL, "806313943.ogg"},
	{0x70F6B56693410DC8ULL, "24303862.ogg"},
	{0x70F74A0902492FC4ULL, "lgt000cr_elf_turn_in_6.utc"},
	{0x70F7DE5ED5678CD6ULL, "456567973.ogg"},
	{0x70F7E0A464840277ULL, "zz_dre_debug6069204.cl"},
	{0x70F84995673F0910ULL, "art\\vfx\\combat\\vfx_r_dwp_right_close1_c.mmh"},
	{0x70F84A619C0D3EFDULL, "260170463.ogg"},
	{0x70F8AF16804A87D8ULL, "llac_lt_0_lowtown_docks_a3_n_-4_3.dds"},
	{0x70F8E4B7A82D7AE3ULL, "vxw_bonepitdestroyed_a3_162_0_0.vlm"},
	{0x70F8F05FADD220A4ULL, "llac_lt_0_lowtown_docks_131071_1.dds"},
	{0x70F922043954C4BFULL, "6121099_f.fxe"},
	{0x70FA427B34AD7E65ULL, "vxlt_0_lowtown_a2_n_c_eqs_0_0.vlm"},
	{0x70FB3642BA7050C4ULL, "vxht_keep_3yi_0_0.vlm"},
	{0x70FB55F1EDA35BAFULL, "follower_banter6170286.cl"},
	{0x70FCA49521F90D5DULL, "hlac_w_encounter2_camp_327678_0z.dds"},
	{0x70FCFDE23755C1CDULL, "308913373.ogg"},
	{0x70FD18AAA9C64D94ULL, "535154375.ogg"},
	{0x70FDD8031CD499FAULL, "634402780.ogg"},
	{0x70FDE9994B7CE210ULL, "dae300_cullen6140608.cl"},
	{0x70FE28180E2C13CEULL, "art\\characters\\weapons\\maces\\w_mce_mc102a.msh"},
	{0x70FE42180E2C3FBCULL, "art\\characters\\weapons\\maces\\w_mce_mc102a.mmh"},
	{0x70FE73A5454A0BE5ULL, "670237415.ogg"},
	{0x70FE73FA000D1487ULL, "235343140.ogg"},
	{0x70FF9EB784222321ULL, "1966506.ogg"},
	{0x70FFC8D026E1E876ULL, "6169659_m.fxe"},
	{0x7100363748E06A6EULL, "393913121.ogg"},
	{0x7100A127470BBDD6ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_1y.dds"},
	{0x7100ED50A0BBFC05ULL, "hlac_lt_gallowsprison_n_-1_8z.dds"},
	{0x71015E27641C7AF7ULL, "607923271.ogg"},
	{0x7101CDA56899C188ULL, "vxht_keep_n_679_0_0.vlm"},
	{0x71020C4B6A7504B7ULL, "9363066.ogg"},
	{0x71028340268AE00DULL, "qcr220_orsino6060745.cl"},
	{0x7102EFCF4C01C752ULL, "6208637_m.fxe"},
	{0x7102F9A004822059ULL, "vxlt_warehouse_wc_0_0.vlm"},
	{0x7102FBB377848F38ULL, "ave200_aveline_bar6115509.cl"},
	{0x71031CA60D871A68ULL, "vxw_sundermount_22q_0_0.vlm"},
	{0x7103A1632F36779AULL, "816733220.ogg"},
	{0x7103A97FDA5FFC2DULL, "mag320_orsino6071127.cl"},
	{0x7103D0BE8202A89AULL, "hlac_lt_0_lowtown_docks_a3_n_0_2z.dds"},
	{0x7103E519BB62682EULL, "6150509_m.fxe"},
	{0x7103FEC86753B8EFULL, "6051030_m.fxe"},
	{0x71040EAE57A300CBULL, "gen00fl_anders6103742.cl"},
	{0x710483BBC0F90FDDULL, "dae030cr_lowlife_fem.utc"},
	{0x71066C95674B822FULL, "art\\vfx\\combat\\vfx_r_dwp_right_close1_c.anb"},
	{0x7106974C15C51761ULL, "zz_approval_debug6003604.cl"},
	{0x7106A5203FB5A8ACULL, "dre000_cut2d_ragstoriches6025097.cl"},
	{0x7106EA7EA8E480A2ULL, "dae000_bodhan6146860.cl"},
	{0x710702FAA512C47EULL, "hlac_lt_0_lowtown_a3_n_c_131072_3x.dds"},
	{0x7107170C1B5F5A5CULL, "vxw_encounter2_camp_n6_0_0.vlm"},
	{0x7107838CD9C409B3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_back_to_relaxed.evt"},
	{0x71081C73349F2B40ULL, "vxht_chantry_n_2ti_0_1.vlm"},
	{0x71095A88B3E8CEC3ULL, "follower_barks6153655.cl"},
	{0x71095E0347AA6D1AULL, "6096773_m.fxe"},
	{0x7109710D66959C14ULL, "dae230ip_at_to_sewers2.utp"},
	{0x7109A74EF77E2C45ULL, "hlac_ca_caverns2prc_5802_56x.dds"},
	{0x710A74BADBF2D246ULL, "6114533_f.fxe"},
	{0x710AD67C07C181A8ULL, "555298031.ogg"},
	{0x710B145D7DEE7012ULL, "llac_ht_keep_4519_13.dds"},
	{0x710C47A7DE46FA78ULL, "gen00fl_varric6093581.cl"},
	{0x710C6C5CF46B156AULL, "6105554_m.fxe"},
	{0x710E2B5250CA43CCULL, "vxlt_0_lowtown_a2_n_c_hd4_0_0.vlm"},
	{0x710E7C8CD9CA1CF4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_back_to_relaxed.gad"},
	{0x710EAACFADA93D38ULL, "vxht_0_hightown_a3_b8n_0_0.vlm"},
	{0x710EC421A969AAD5ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st01a_0s.dds"},
	{0x710F83E807077A91ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medp_d.dds"},
	{0x710FAC452DE520C3ULL, "vxca_caverns1_mines_ckm_0_0.vlm"},
	{0x710FB1DCFDB1718EULL, "vxht_brothel_n_6gd_0_0.vlm"},
	{0x710FCC233EEEE18BULL, "wa_dust.mao"},
	{0x710FE3F86EB30E8FULL, "mcr370_meredith6085931.cl"},
	{0x71102928B2DB92D4ULL, "6161231_m.fxe"},
	{0x711039589E499DB4ULL, "hlac_lt_0_lowtown_0_5x.dds"},
	{0x71103A0C8F0ADB9EULL, "vxca_caverns2_7my_0_1.vlm"},
	{0x71103BCE3845FB8EULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1whitelyrium.msh"},
	{0x71103F825A8FDA08ULL, "vxht_keep_n_626_0_0.vlm"},
	{0x711049CE384613DFULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1whitelyrium.mao"},
	{0x711055CE3846287CULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1whitelyrium.mmh"},
	{0x7110CF6E64377595ULL, "657905483.ogg"},
	{0x711107670B8F4D6AULL, "gen00fl_bethany6112226.cl"},
	{0x71111F28F8631626ULL, "vxht_bartrand_n_5p7_0_1.vlm"},
	{0x71112F18FE795437ULL, "52064248.ogg"},
	{0x71116B7CF67838D9ULL, "vxlt_warehouse_xb_0_0.vlm"},
	{0x71118E82FF7D32E8ULL, "vxw_sundermount_27n_0_0.vlm"},
	{0x7111DB1BEE26680DULL, "128022741.ogg"},
	{0x7111DF570F49097EULL, "vxht_0_hightown_a3_b00_0_0.vlm"},
	{0x7112A0F909D03571ULL, "vxlt_0_lowtown_enw_0_0.vlm"},
	{0x7112D0AC19F2F22EULL, "6024530_m.fxe"},
	{0x7112DDB4F3368486ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_up.ani"},
	{0x7113154F622E91D8ULL, "plt_zz_mag200pt_nt_debug.plo"},
	{0x711336781326DBCFULL, "6174465_m.fxe"},
	{0x711353A797D26FDAULL, "dae000_orsino6085529.cl"},
	{0x7114604CC06D0D76ULL, "ch1armtnt_vl.psh.fxo"},
	{0x71146FF7B87DA1F4ULL, "vxlt_darktown_2qk_0_0.vlm"},
	{0x7114A5F202079CCEULL, "6106388_m.fxe"},
	{0x7114AFE9A9A6111EULL, "291959564.ogg"},
	{0x71150F82CA8C28F0ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_dink.mao"},
	{0x71151382CA8C2F03ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_dink.mmh"},
	{0x71152182CA8C475DULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_dink.msh"},
	{0x711552457C8E7807ULL, "m_dae130ar_undercity.wbd"},
	{0x71158B8CD9D054D2ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_back_to_relaxed.ani"},
	{0x711611F4FFF42C8DULL, "6137242_m.fxe"},
	{0x71163B972012A4E5ULL, "613517109.ogg"},
	{0x711653609E3FA8B5ULL, "vxca_caverns1_mines_dq6_0_0.vlm"},
	{0x7116F6B209030789ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas2_d.dds"},
	{0x7117727AE4DCEFE5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basicwood.phy"},
	{0x71177A9D4C805D76ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_1z.dds"},
	{0x711819FC9E5C81C4ULL, "vxht_keep_3co_0_0.vlm"},
	{0x7118A661D555F22BULL, "6048739_m.fxe"},
	{0x7118CC8852542D88ULL, "dae000_flemeth6013424.cl"},
	{0x7118E697CDE8EAC6ULL, "6113433_f.fxe"},
	{0x7119149C123D0383ULL, "vxw_encounter2_ow_0_1.vlm"},
	{0x71193D262F7C81E3ULL, "hlac_ht_mansion2_818_35y.dds"},
	{0x7119B5B65DDC8142ULL, "541483414.ogg"},
	{0x711A8DB0411EF856ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicwood.mao"},
	{0x711A9192F90C9CB9ULL, "377792473.ogg"},
	{0x711A99B0411F0CF5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicwood.mmh"},
	{0x711A9BB0411F106BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicwood.msh"},
	{0x711B28B6F27D8D51ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow8base.mmh"},
	{0x711B34B6F27DA1F2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow8base.mao"},
	{0x711B42B6F27DB90FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow8base.msh"},
	{0x711BA9AAADDC0ABAULL, "925779566.ogg"},
	{0x711BADD12E1B42AEULL, "qun100_mage_party_comment6019092.cl"},
	{0x711BB6EBF7C71385ULL, "6166385_m.fxe"},
	{0x711C1AF7A539E4EFULL, "vxlt_darktown_5lp_0_0.vlm"},
	{0x711CAC8A0ACE1976ULL, "6163459_m.fxe"},
	{0x711CC4152B9A235BULL, "950098422.ogg"},
	{0x711CD4A09FBAA953ULL, "vxdr_primevaltomb_507_0_1.vlm"},
	{0x711CF2C5E5DA55F6ULL, "gen00fl_fenris6126889.cl"},
	{0x711D3A3700660D04ULL, "em_lgt_baz_sketch.mor"},
	{0x711DDB81CDDC3AE3ULL, "dae000_bartrand6015442.cl"},
	{0x711EA62EF5BBC7D4ULL, "vxca_caverns2prc_62q_0_2.vlm"},
	{0x711EDCBD7B186736ULL, "vxlt_warehouse_1cq_0_0.vlm"},
	{0x711F068917E051F6ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_anders_chant_ex.ani"},
	{0x711F0D6460595644ULL, "vxlt_0_lowtown_docks_n_b4t_0_1.vlm"},
	{0x711F26C989E80CBDULL, "vxlt_0_lowtown_a3_n_c_er3_0_0.vlm"},
	{0x711FBDE74514AC91ULL, "vxw_sundermount_2b4_0_0.vlm"},
	{0x711FE1C124F582CDULL, "character_character_p253.psh.fxo.dx11"},
	{0x71207AEE0408176FULL, "448532896.ogg"},
	{0x7120C5B4F34299F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_up.evt"},
	{0x7120C79EE3BDAC46ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier3_arishokresponse.ani"},
	{0x7120CBF566F25631ULL, "6153125_m.fxe"},
	{0x7121269F0C9B0A53ULL, "vxca_cavern1_haunted_e6t_0_0.vlm"},
	{0x71213521DCECC461ULL, "vxlt_0_lowtown_a2_hfb_0_0.vlm"},
	{0x71213B7FE9B58EA1ULL, "6152351_m.fxe"},
	{0x7121B111F36976BEULL, "fx_lightningbolt02.dds"},
	{0x7121BF690AA210BEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ferelden_jimmy_sa.dds"},
	{0x71220B7DA57170A3ULL, "fen310_varania6118586.cl"},
	{0x712240B39AA875F2ULL, "6014256_m.fxe"},
	{0x71224114916B069FULL, "dae150_harshal6148848.cl"},
	{0x7122C800ED0269D6ULL, "gst_standing_back.gda"},
	{0x7122C86D7F7F07BDULL, "257048933.ogg"},
	{0x7122FCB3322271A3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_x4broken_1.mao"},
	{0x71231A8A2D8EE55BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornatelrg.mmh"},
	{0x712321180E4B2387ULL, "art\\characters\\weapons\\maces\\w_mce_mc102a.phy"},
	{0x7123268A2D8EF9F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornatelrg.mao"},
	{0x7123388A2D8F1845ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornatelrg.msh"},
	{0x71234C5C98BACC33ULL, "6207568_f.fxe"},
	{0x7123788923D74530ULL, "ts_demonic_01.gda"},
	{0x71238C58E0F6C13CULL, "one220_yevhen.cnv"},
	{0x7123B98CEF185909ULL, "149093359.ogg"},
	{0x7123E6CF006B40F1ULL, "vxca_cavern1_haunted_eh4_0_0.vlm"},
	{0x7123FAC5FF4B8B5CULL, "vxw_encounter2_camp_d4_0_0.vlm"},
	{0x7124387D2705E593ULL, "vxht_0_hightown_a2_8q1_0_0.vlm"},
	{0x71245D1C62A99D29ULL, "ht_0_hightown_n_45_0_lc.mmh"},
	{0x712463904F342CE3ULL, "vxw_bonepit_ne_0_0.vlm"},
	{0x7125231D45B14552ULL, "mcr310_argument6074335.cl"},
	{0x71259592C865B5A3ULL, "weapon1_ds1.psh.fxo"},
	{0x7125A397E8EDA563ULL, "vxca_cavern1_haunted_ii1_0_0.vlm"},
	{0x712621BC965CEAA6ULL, "238312285.ogg"},
	{0x7126BD1D2AD11E40ULL, "6110820_m.fxe"},
	{0x7126C6B4F3470680ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_up.gad"},
	{0x7126E44BE62C94B8ULL, "6122780_m.fxe"},
	{0x7126F75DF46C6BDEULL, "vxht_keep_4ad_0_0.vlm"},
	{0x712726A3AA1D07C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_longrange_v3.gad"},
	{0x7127AE58E0FAE38CULL, "one220_yevhen.bnk"},
	{0x71284C67A988B6DEULL, "82899453.ogg"},
	{0x7128930636D4959CULL, "vxlt_0_lowtown_e80_0_0.vlm"},
	{0x7129506AB4B5F94DULL, "371331847.ogg"},
	{0x712968A2612F96E2ULL, "hlac_lt_0_lowtown_n_65533_2x.dds"},
	{0x7129811BC2D95B81ULL, "vxw_encounter2_camp_ng_0_0.vlm"},
	{0x712AA5AB2A516FC8ULL, "zz_firestaff2_v3.cut"},
	{0x712AAE89063C30D8ULL, "vxlt_warehouse_n_1a2_0_0.vlm"},
	{0x712ABDD79E417048ULL, "vxlt_0_lowtown_a2_hfg_0_0.vlm"},
	{0x712B07E0BB0882AEULL, "6080679_m.fxe"},
	{0x712B5373F5C28517ULL, "hlac_lt_warehouse_0_8z.dds"},
	{0x712C89E58C0B9253ULL, "162473989.ogg"},
	{0x712CBAA2FF7B3350ULL, "vxlt_undercityencounter_1s4_0_0.vlm"},
	{0x712CBE771D343C3EULL, "645771075.ogg"},
	{0x712D18076318C254ULL, "lowtown_fem_urchin_mob6165932.cl"},
	{0x712DEA2BE83A74DBULL, "6016164_f.fxe"},
	{0x712E1BA3AA2313D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_longrange_v3.evt"},
	{0x712E7545ECFC0F32ULL, "hlac_w_bonepitmining_a2_393219_0y.dds"},
	{0x712EC66E44398937ULL, "vxht_brothel_n_3i6_0_0.vlm"},
	{0x712EE15472920FACULL, "zz_approval_debug6003944.cl"},
	{0x712F33DC39FC7C2EULL, "t_table_02.dds"},
	{0x712FA6FECEE2DCE1ULL, "vxlt_0_lowtown_a2_hca_0_0.vlm"},
	{0x71301E0FA8DD2331ULL, "849759814.ogg"},
	{0x7130AFB12906629BULL, "hlac_w_bonepitdestroyed_a3_131075_0z.dds"},
	{0x7130E76A0E5BF047ULL, "6121085_m.fxe"},
	{0x7131957DD8D3F4A7ULL, "mag370_meredith6040634.cl"},
	{0x7131AAD6FAA1CFFFULL, "hlac_lt_0_lowtown_docks_a3_-131074_5z.dds"},
	{0x7131D2E0A23A0A78ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliff2_darelief.dds"},
	{0x7131EB4493F43E8EULL, "mcr371_hostile_merrill6082167.cl"},
	{0x7131FD1026DE4AE4ULL, "vxw_bonepitmining_a2_fd_0_0.vlm"},
	{0x713318BACA2D9573ULL, "774396216.ogg"},
	{0x71331BC7998AAF25ULL, "lt_0_lowtown_docks_28_0_ll.mmh"},
	{0x71332F8917F14030ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_anders_chant_ex.gad"},
	{0x713416DF9B0B36BDULL, "718464461.ogg"},
	{0x713466FA8D791360ULL, "gen00fl_varric6078704.cl"},
	{0x7134BF1ED24653A8ULL, "6188560_m.fxe"},
	{0x7134D8830EDA2189ULL, "dae000_pc_home_ambients6208173.cl"},
	{0x7134FCC97D5F66A1ULL, "vxca_caverns2prc_aap_0_0.vlm"},
	{0x7135152DF430D3D7ULL, "178508008.ogg"},
	{0x71357C5191667675ULL, "vxlt_0_lowtown_a2_n_c_epa_0_1.vlm"},
	{0x7135B09EE3CFE140ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier3_arishokresponse.gad"},
	{0x71360B5AB249DA65ULL, "6152364_m.fxe"},
	{0x71362BC2C10A0FC9ULL, "hlac_lt_warehouse_n_65537_5x.dds"},
	{0x71364A38A540C42AULL, "440972404.ogg"},
	{0x71366D57ADDA4C1BULL, "6076301_m.fxe"},
	{0x7136D962D872AEF5ULL, "vxw_sundermount_30g_0_0.vlm"},
	{0x7136DEF010DC7914ULL, "t3_arm_drb.tnt"},
	{0x713704E328CAAE1CULL, "vxlt_0_lowtown_e31_0_0.vlm"},
	{0x713733BCA68BBC46ULL, "vxlt_gallowstemplar_n_4qd_0_0.vlm"},
	{0x71376675FC57A112ULL, "zz_followers_banter6158160.cl"},
	{0x7137F2F8B4CF7401ULL, "vxw_encounter2_camp_if_0_0.vlm"},
	{0x71382A196D779C58ULL, "dae000_arishok6067636.cl"},
	{0x71383C1E4FBAB038ULL, "vxca_caverns1_mines_dnn_0_0.vlm"},
	{0x713885D830611E00ULL, "gen00fl_aveline6122077.cl"},
	{0x7138C2700EC669BAULL, "vxlt_0_lowtown_h90_0_0.vlm"},
	{0x7138E4CD96D4AC0AULL, "vxlt_gallowstemplar_3qt_0_0.vlm"},
	{0x7139003887C2B9C7ULL, "100196239.ogg"},
	{0x71392FB4903788C8ULL, "vxlt_0_lowtown_a2_hcd_0_0.vlm"},
	{0x713942AC34D8524DULL, "vxlt_0_lowtown_a2_gob_0_0.vlm"},
	{0x713956179A62E646ULL, "vxlt_gallowstemplar_f_45p_0_0.vlm"},
	{0x713957F157664593ULL, "hlac_lt_0_lowtown_docks_a3_-3_4z.dds"},
	{0x7139A6EA31C75177ULL, "469827120.ogg"},
	{0x7139AD9E68C7BBE8ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_lrgchair_1.mao"},
	{0x713A619C4DB2698AULL, "243343234.ogg"},
	{0x713A717B3D5703C4ULL, "61630259.ogg"},
	{0x713B13A3AA2D900DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_longrange_v3.ani"},
	{0x713B5FDB07CB38C4ULL, "gen00fl_sebastian6188578.cl"},
	{0x713B7F39675009A5ULL, "6067705_m.fxe"},
	{0x713B88D73DE697AEULL, "57226073.ogg"},
	{0x713BC49C7D1B53B4ULL, "121623700.ogg"},
	{0x713C65764F95EB15ULL, "vxht_chantry_355_0_1.vlm"},
	{0x713D2BDE374320BEULL, "elfmale.fxa"},
	{0x713DB4E54CBB5436ULL, "6004110_m.fxe"},
	{0x713E0E643E0CB481ULL, "6028912_f.fxe"},
	{0x713E4281D2E28B06ULL, "gen00fl_anders6113021.cl"},
	{0x713E4CD7E3E767FAULL, "gen00fl_anders6125468.cl"},
	{0x713E81D56421C58EULL, "hlac_w_encounter2_camp_0_0z.dds"},
	{0x713E8B88B9D61783ULL, "zz_ave_debug6207421.cl"},
	{0x713EDBC1B5E018EFULL, "6094954_m.fxe"},
	{0x713F709C47F8D532ULL, "6142497_f.fxe"},
	{0x713F783907A83B90ULL, "6096638_m.fxe"},
	{0x713FB9FC74E070ABULL, "one000_ilen6057365.cl"},
	{0x713FCF4CFECF7348ULL, "llac_ht_chantry_n_2534_39.dds"},
	{0x7140483EE569E340ULL, "6142812_m.fxe"},
	{0x7140491A6A7E3C72ULL, "181245991.ogg"},
	{0x71406EED18D46364ULL, "vxw_bonepitmining_a2_me_0_0.vlm"},
	{0x7140968551D61FD9ULL, "6113030_f.fxe"},
	{0x714162ED4D4AFEDFULL, "qcr000_arishok6059669.cl"},
	{0x71416E8F43261341ULL, "216266916.ogg"},
	{0x71418221737CC082ULL, "6164362_m.fxe"},
	{0x71426D5122AED5B3ULL, "vxlt_0_lowtown_docks_a3_br8_0_0.vlm"},
	{0x7142EAF693F57BFEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_male.evt"},
	{0x71430DF91F932F33ULL, "vxca_cavern1_haunted_gkq_0_0.vlm"},
	{0x71432A13A7923D17ULL, "mrl150_merrill6078544.cl"},
	{0x71439B41D798EA3DULL, "ser221_party_comments6163606.cl"},
	{0x7143D95BB932E94EULL, "6066412_f.fxe"},
	{0x7143DFD7F86EB501ULL, "435201273.ogg"},
	{0x714401F82B0AD024ULL, "lt_0_lowtown_docks_a3_n_215_0.msh"},
	{0x714413F82B0AEE8DULL, "lt_0_lowtown_docks_a3_n_215_0.mao"},
	{0x714417F82B0AF5C6ULL, "lt_0_lowtown_docks_a3_n_215_0.mmh"},
	{0x71443663F2EF0BA5ULL, "vxw_encounter2_camp_ox_0_0.vlm"},
	{0x714449FFB8B44FCFULL, "1036854707.ogg"},
	{0x7144566026D2C9CCULL, "qun110_petrice6042199.cl"},
	{0x7144800DD8B7BB06ULL, "zz_lgt_3_debug.cnv"},
	{0x714517ED516D57C9ULL, "vxw_woundedcoast_1c0_0_0.vlm"},
	{0x7145A5999881D4C6ULL, "vxlt_gallowstemplar_n_4ve_0_0.vlm"},
	{0x7146085218573B76ULL, "vxdr_primevalentrance_2si_0_0.vlm"},
	{0x714627FA613A42A7ULL, "6176796_f.fxe"},
	{0x714664D5A6C58C81ULL, "vxw_encounter2_camp_de_0_0.vlm"},
	{0x7146C0310D5A8B7CULL, "vxht_0_hightown_9zf_0_0.vlm"},
	{0x7146F4D6F1F9EC74ULL, "zz_and_debug6134103.cl"},
	{0x7147567404C6A134ULL, "vxlt_warehouse_13s_0_0.vlm"},
	{0x714789372560FDC4ULL, "6117612_m.fxe"},
	{0x7147A191822DA148ULL, "vxlt_0_lowtown_a2_hhe_0_0.vlm"},
	{0x7147E8F1C59DB25CULL, "vxlt_0_lowtown_a3_n_c_gmy_0_0.vlm"},
	{0x71480750C4F7BF23ULL, "w_encounter3_153_0.mao"},
	{0x71480B50C4F7C670ULL, "w_encounter3_153_0.mmh"},
	{0x71481950C4F7DD8AULL, "w_encounter3_153_0.msh"},
	{0x7148C6B7152097FAULL, "w_encounter3_182_0.mmh"},
	{0x7148CAB715209E31ULL, "w_encounter3_182_0.mao"},
	{0x7148D8B71520B640ULL, "w_encounter3_182_0.msh"},
	{0x7149AACE617664AFULL, "320342535.ogg"},
	{0x7149CA81B65EED17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\chargenidle.gad"},
	{0x714A01F693FBC129ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_male.gad"},
	{0x714BDAE5DD135115ULL, "lgt260_way_blood_mage6148388.cl"},
	{0x714C81273EB080E5ULL, "dre000_bodahn6031313.cl"},
	{0x714C93AECE37D66DULL, "hlac_lt_0_lowtown_a2_65535_2z.dds"},
	{0x714D650351660901ULL, "468456040.ogg"},
	{0x714DA9936B21BB42ULL, "135230028.ogg"},
	{0x714E19A4030FA971ULL, "llac_lt_gallowstemplar_2395_0.dds"},
	{0x714E19EFB0D3BD1AULL, "gen00fl_varric6075150.cl"},
	{0x714E2A7A8350FBF3ULL, "230851002.ogg"},
	{0x714E5BCCBF288536ULL, "lt_0_lowtown_a3_n_c_53_0_ll.mmh"},
	{0x714E82EB09B3EB2BULL, "6047374_m.fxe"},
	{0x714EB135100AF343ULL, "hlac_lt_gallowstemplar_2390_1z.dds"},
	{0x714EC459C82BD30CULL, "hlac_lt_gallowstemplar_2388_6y.dds"},
	{0x714F0718D9DCD551ULL, "qun221_viscount6095600.cl"},
	{0x714F5F231ACAAA3CULL, "lt_0_lowtown_docks_a3_n_2_0_lc.mmh"},
	{0x714FA1367E42B1ECULL, "hlac_lt_undercityserialkiller_-1_42z.dds"},
	{0x714FB2EBA300A52EULL, "hlac_ht_0_hightown_a2_-65535_1y.dds"},
	{0x714FB8C249CF3424ULL, "vxca_caverns1_mines_e7p_0_0.vlm"},
	{0x71505C28FD64AA1DULL, "6177318_m.fxe"},
	{0x7150666E91BC22A3ULL, "vxw_encounter3_75_0_0.vlm"},
	{0x715089B8146AB0F2ULL, "vxht_0_hightown_n_c3o_0_1.vlm"},
	{0x7150D2F694019077ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_male.ani"},
	{0x7150D82C18F9BDD1ULL, "hlac_ht_playershouse_n_855_2z.dds"},
	{0x7150FD503E33A0BFULL, "hlac_lt_0_lowtown_a2_n_c_-131073_1z.dds"},
	{0x715191DC6966D073ULL, "art\\vfx\\combat\\vfx_gui_target_ret_20_p.anb"},
	{0x7151A77B8888E664ULL, "art\\characters\\creatures\\ragedemon\\textures\\c_ragedmon_0e.dds"},
	{0x715278BCE9361029ULL, "hlac_dr_0_deeproads_5842_13x.dds"},
	{0x7152A840E4E52425ULL, "vxw_encounter2_camp_jg_0_0.vlm"},
	{0x7152D93923AFD489ULL, "vxht_brothel_n_5jg_0_0.vlm"},
	{0x7152DE120F499F33ULL, "dae000_light_tombs_boss6179461.cl"},
	{0x715320EEBFD02573ULL, "6075267_m.fxe"},
	{0x715334BED00865F9ULL, "da2terrain_terrain_p1.psh.fxo.dx11.assembly"},
	{0x71533FF42F652833ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_min_c.anb"},
	{0x71539F3E6EBED867ULL, "vxca_caverns2_78s_0_1.vlm"},
	{0x7153A97C85FC674EULL, "hlac_lt_blackemporium_1055_5y.dds"},
	{0x7154118F41D2874EULL, "6039954_m.fxe"},
	{0x7154AE4451298CAAULL, "w_encounter3_8_0.mao"},
	{0x7154B244512993F9ULL, "w_encounter3_8_0.mmh"},
	{0x7154BC445129A4E7ULL, "w_encounter3_8_0.msh"},
	{0x7154EF1C92CCAED4ULL, "gen00fl_isabela6081322.cl"},
	{0x7154F48DC3DE9274ULL, "ave200_donnic6110501.cl"},
	{0x71555783762D798AULL, "pro000_opening_pc6001207.cl"},
	{0x715560D2F99FC829ULL, "lgt000ip_talwain_crate.utp"},
	{0x7156175DDF44032EULL, "llac_ht_bartrand_n_3372_0.dds"},
	{0x715627553D6D0B14ULL, "hlac_w_sundermount_393211_0x.dds"},
	{0x71563FE498FE72AFULL, "one280_iwan6096736.cl"},
	{0x71565C457CC6203DULL, "m_dae130ar_undercity.bnk"},
	{0x7156719064ACD74BULL, "6118405_m.fxe"},
	{0x7156E21404FCC657ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelogs.phy"},
	{0x715800818BB7286EULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\ready.ani"},
	{0x7158930810AA033AULL, "vxlt_0_lowtown_elj_0_0.vlm"},
	{0x7159303BF2426EEEULL, "art\\levels\\ztools\\default\\generic\\9ff_pyramid.msh"},
	{0x71593E3BF242863FULL, "art\\levels\\ztools\\default\\generic\\9ff_pyramid.mao"},
	{0x71593EDBCAE035DBULL, "mag110_sailor6107412.cl"},
	{0x71594A3BF2429ADCULL, "art\\levels\\ztools\\default\\generic\\9ff_pyramid.mmh"},
	{0x71598093CEB705E3ULL, "dae000_mother6099184.cl"},
	{0x7159C363519A7BF5ULL, "vxlt_gamlenshouse_n_1w_0_1.vlm"},
	{0x715A4C9FB1DBD77DULL, "llac_dr_0_deeproads_5845_3.dds"},
	{0x715A53C54369CFCCULL, "6033126_f.fxe"},
	{0x715ACB0515689B7DULL, "vxht_mansion2b_1xy_0_0.vlm"},
	{0x715B31DFB1108260ULL, "927022934.ogg"},
	{0x715B9AC2FB66C03CULL, "gen00fl_merrill6116048.cl"},
	{0x715BA090C8EF056CULL, "729085133.ogg"},
	{0x715CB3890DABFB56ULL, "hlac_w_encounter2_camp_196610_0y.dds"},
	{0x715CC8D016253141ULL, "vxw_encounter3_bp_0_0.vlm"},
	{0x715CD0097C86DA59ULL, "vxlt_blackemporium_14c_0_0.vlm"},
	{0x715D5EA43E745869ULL, "665299772.ogg"},
	{0x715DC4DC05071771ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs101a_0n.dds"},
	{0x715DC54918FE3281ULL, "636448031.ogg"},
	{0x715DD10C52FF0582ULL, "612104864.ogg"},
	{0x715DE902E18D03D9ULL, "vxw_encounter3_4r_0_0.vlm"},
	{0x715DF781B66FE29DULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\chargenidle.ani"},
	{0x715E1C755FB8693CULL, "follower_banter6170122.cl"},
	{0x715E37243C136B62ULL, "target.cur"},
	{0x715ED11D0BF1F8FFULL, "lgt000ip_knife_sack.utp"},
	{0x715F6CDA44DD4E6DULL, "6173192_m.fxe"},
	{0x715FAEDC69733EDCULL, "art\\vfx\\combat\\vfx_gui_target_ret_20_p.mmh"},
	{0x715FD7271E681936ULL, "vxlt_backalley_n_1nu_0_1.vlm"},
	{0x715FEFFCC96C3515ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medh_0t.dds"},
	{0x71605CF42F6FE39CULL, "art\\vfx\\combat\\vfx_glb_ice_imp_min_c.mmh"},
	{0x7160977D13E7343AULL, "554690003.ogg"},
	{0x71609EBA2D9ADCC2ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.mmh"},
	{0x71611A1DD6DB3CA5ULL, "vxw_encounter2_camp_ef_0_0.vlm"},
	{0x71613839510C4701ULL, "116079504.ogg"},
	{0x71614B1615A5ED09ULL, "vxht_brothel_n_5mf_0_0.vlm"},
	{0x71615C02A6598813ULL, "plt_cod_tomes_tome.nss"},
	{0x7161FBA7355988C9ULL, "vxw_woundedcoast_1i6_0_0.vlm"},
	{0x71620B254180A621ULL, "hlac_lt_hangedman_-1_15z.dds"},
	{0x71638077DB6B825CULL, "pro000_friedrich6050168.cl"},
	{0x716423407BD6B98AULL, "plt_cod_hst_talvashoth.plo"},
	{0x71644267C47EC314ULL, "107089427.ogg"},
	{0x7164F8818BC1A4BFULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\ready.evt"},
	{0x7165915DAFBB5FE4ULL, "6076401_m.fxe"},
	{0x7165C59B88026CF3ULL, "6169517_m.fxe"},
	{0x7165F5BE59F8C49BULL, "6160671_m.fxe"},
	{0x716605AEF168CBC2ULL, "hlac_ca_caverns1_mines_24424_27y.dds"},
	{0x716618296042034CULL, "710627676.ogg"},
	{0x7166470AE585AECAULL, "and100_anders6149705.cl"},
	{0x7166D4B9150E2218ULL, "6103900_f.fxe"},
	{0x71678C12D755BF37ULL, "6125610_m.fxe"},
	{0x716798CDCAEBB8C5ULL, "dae000cr_grand_cleric6199340.cl"},
	{0x71679A6827A3768EULL, "ht_mansion2_mmap.dds"},
	{0x7167A94A19B3C7F4ULL, "vxw_bonepitdestroyed_a3_o5_0_0.vlm"},
	{0x7167E3C5A5CBE5A9ULL, "qun110_petrice6018748.cl"},
	{0x71693CE2075EB3FDULL, "vxht_mansion2b_1sx_0_0.vlm"},
	{0x716980B95593B741ULL, "6135828_m.fxe"},
	{0x7169BED0BBB8C697ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslb_0d.dds"},
	{0x7169D7D3BFA02BB0ULL, "154710550.ogg"},
	{0x716A10494D2D9C90ULL, "164604346.ogg"},
	{0x716A8394235EBA72ULL, "352881915.ogg"},
	{0x716A900736B33F9DULL, "hlac_ht_chantry_2534_23y.dds"},
	{0x716B3B8A2DCCD290ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_ornatelrg.phy"},
	{0x716B9E863C811D5AULL, "hightown_male_cityguard_mob6170028.cl"},
	{0x716BAE9F7CCF5DA3ULL, "6171440_m.fxe"},
	{0x716BC6CDEE19BBA7ULL, "vxlt_0_lowtown_docks_n_cs7_0_1.vlm"},
	{0x716BD9818BC78EA8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\ready.gad"},
	{0x716BDE4E5F1515C3ULL, "vxht_0_hightown_7fw_0_0.vlm"},
	{0x716C5ADFD3831C59ULL, "vxw_encounter3_1q_0_0.vlm"},
	{0x716C725A2BC70050ULL, "168878097.ogg"},
	{0x716CC0029364FBEBULL, "gen00fl_isabela_init6110280.cl"},
	{0x716CEA9D9D272F0AULL, "vxht_0_hightown_as3_0_0.vlm"},
	{0x716D278E63F2618EULL, "6140293_f.fxe"},
	{0x716D69015CAC4B32ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\arcanehorror_movement.anb"},
	{0x716D7C1EC1A1E52EULL, "6041849_m.fxe"},
	{0x716D9D9122895266ULL, "876189603.ogg"},
	{0x716DBFE9D3EB1952ULL, "dae000_player_generic_find.cnv"},
	{0x716E1AA2F07BEC25ULL, "spt_billboards0.psh.fxo"},
	{0x716E434DE714D36DULL, "gen00fl_anders6111358.cl"},
	{0x716F5FE0BBDA27B5ULL, "27827841.ogg"},
	{0x7170451C17075743ULL, "197016423.ogg"},
	{0x71706D84274FA149ULL, "vxw_woundedcoast_1t5_0_0.vlm"},
	{0x717078786A37E216ULL, "6163341_m.fxe"},
	{0x71709D117237AE13ULL, "928250025.ogg"},
	{0x7170E1E9D3ED88AAULL, "dae000_player_generic_find.bnk"},
	{0x7170EBC23E0E0AC8ULL, "76858840.ogg"},
	{0x7171096C55968E6FULL, "vxw_woundedcoast_118_0_0.vlm"},
	{0x71710C2947450982ULL, "hlac_lt_gallowstemplar_2394_2z.dds"},
	{0x7171651945D0AAE9ULL, "147971413.ogg"},
	{0x7171A4030FCEDE5EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_keepwall_sa.dds"},
	{0x7171DDFB56F666B5ULL, "vxlt_gallowsprison_n_64n_0_0.vlm"},
	{0x7171F30F46BB665BULL, "vxlt_gallowscourtyard_72f_0_0.vlm"},
	{0x717224A1DF0E11B6ULL, "gen00fl_varric6033574.cl"},
	{0x71728E63132B42B3ULL, "6139388_f.fxe"},
	{0x7172AA39BDC70DD0ULL, "134182823.ogg"},
	{0x7172B49300044E90ULL, "vxlt_0_lowtown_a2_n_c_dbb_0_0.vlm"},
	{0x7173E384BD716DF3ULL, "955158226.ogg"},
	{0x7174679B4BEEDD1BULL, "6165971_m.fxe"},
	{0x717471D4C8CA6399ULL, "hf_ave_brennan.mor"},
	{0x71748F61704788E7ULL, "567624780.ogg"},
	{0x7174EFBCC4F5D1C9ULL, "qun111_ketojan6041022.cl"},
	{0x717528D3A48E2B5AULL, "vxht_bartrand_n_4f1_0_0.vlm"},
	{0x7175469607043A98ULL, "6104800_f.fxe"},
	{0x71757562571926DFULL, "art\\vfx\\combat\\vfx_glb_bld_imp_maj_c.mmh"},
	{0x7175D89AA3EB6AC1ULL, "gen_im_trs_cot_junk1_01.uti"},
	{0x7176507CF8CBFCEAULL, "6083032_m.fxe"},
	{0x7176B760745C8D1AULL, "202363497.ogg"},
	{0x717757FCB27D0C76ULL, "6142182_m.fxe"},
	{0x7177AEBEF954CC7DULL, "vxht_mansion2b_1vw_0_0.vlm"},
	{0x7177BFC74E2093ABULL, "vxlt_gallowstemplar_f_2ec_0_0.vlm"},
	{0x7177ED58F5D60D3DULL, "mag200_feynriel6030444.cl"},
	{0x71788578EDF0ADDFULL, "vxlt_hangedman_n_1xf_0_0.vlm"},
	{0x717894882BDC0A9DULL, "vxht_mansion2_1af_0_0.vlm"},
	{0x7178C56D7F691424ULL, "vxlt_gallowstemplar_26f_0_0.vlm"},
	{0x717919E0F9B64786ULL, "hlac_w_encounter2_camp_327685_0y.dds"},
	{0x7179ED0FC1BBC7AFULL, "vxht_chantry_2te_0_1.vlm"},
	{0x717A2987B12D55E7ULL, "sebev_item_acquired.ncs"},
	{0x717A4A695F593F02ULL, "807284128.ogg"},
	{0x717B52EEAB4D1DDAULL, "6066406_m.fxe"},
	{0x717BB94AEAA541F3ULL, "gen_im_arm_boo_hvy_act1_02.uti"},
	{0x717BF5153FA65A22ULL, "vxht_mansion2_2py_0_0.vlm"},
	{0x717C20842E2A38A6ULL, "97274038.ogg"},
	{0x717C42C1CB428634ULL, "mcr370_templar26087014.cl"},
	{0x717C706DA4DB8B6DULL, "square.arl"},
	{0x717CD617256E5F99ULL, "vxlt_warehouse_n_13u_0_0.vlm"},
	{0x717D1A981787C34FULL, "6054103_m.fxe"},
	{0x717D238C48265318ULL, "sib100_father_contact6108442.cl"},
	{0x717D5533CAF3F9BDULL, "6019155_m.fxe"},
	{0x717DDB1C1796F36CULL, "6064837_m.fxe"},
	{0x717E293BF2617EA7ULL, "art\\levels\\ztools\\default\\generic\\9ff_pyramid.phy"},
	{0x717E4510B8144E4BULL, "clf221_jalen6070393.cl"},
	{0x717EB7E2CC0D60C0ULL, "vxdr_0_deeproads_eup_0_0.vlm"},
	{0x717EC0C8C8790BC1ULL, "301898873.ogg"},
	{0x717F1C6326407ECEULL, "6124145_f.fxe"},
	{0x717F3583F4C0CAC7ULL, "vxlt_0_lowtown_docks_n_b7c_0_0.vlm"},
	{0x717F43D0A7CBC2FDULL, "vxlt_0_lowtown_a2_n_ghl_0_0.vlm"},
	{0x717F6B321C8063E2ULL, "llac_w_encounter1_1_0.dds"},
	{0x717F6C1F0822DCA5ULL, "147481606.ogg"},
	{0x717F9CA3AC925CF1ULL, "zz_jp_aveline.cut"},
	{0x717FE4475D82F487ULL, "vxlt_darktown_6i0_0_0.vlm"},
	{0x7180014F0D017C83ULL, "vxw_woundedcoast_ia_0_0.vlm"},
	{0x71806093EBE20A75ULL, "vxlt_0_lowtown_docks_a3_d28_0_0.vlm"},
	{0x71806BCC7844829BULL, "qun220cr_javaris6071892.cl"},
	{0x7180777E05F5376BULL, "ef_vinell.mor"},
	{0x71808453489AE67CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_medj_0n.dds"},
	{0x7181218C06D352F2ULL, "hlac_ht_0_hightown_1_0x.dds"},
	{0x718152723BE6729FULL, "vxlt_foundry_n_122_0_0.vlm"},
	{0x71816942A3A72851ULL, "821444741.ogg"},
	{0x7181B4AE1E2D702FULL, "6069477_m.fxe"},
	{0x71820F09CDA9A600ULL, "966085006.ogg"},
	{0x71829712D70860AAULL, "mage_revival.dds"},
	{0x7182A9556BEE9DF3ULL, "6163717_m.fxe"},
	{0x7182F8F4496FA5C3ULL, "356593855.ogg"},
	{0x71832A392252EF87ULL, "art\\vfx\\environment\\vfx_water_drips_p.dura.gad"},
	{0x718370894F803598ULL, "6072262_f.fxe"},
	{0x7183C423DF534636ULL, "pro000_friedrich6051060.cl"},
	{0x7183E8E8A7E786E6ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_groundivy_dp.dds"},
	{0x7184247FE7154C9BULL, "dae000_worthy6166606.cl"},
	{0x71846B3B6A7CCDE9ULL, "5738068.ogg"},
	{0x7185D578F59C2C7AULL, "160563113.ogg"},
	{0x71862A38ED3F1236ULL, "vxlt_0_lowtown_a2_n_c_hl6_0_0.vlm"},
	{0x71862C57708CCB61ULL, "follower_banter6166500.cl"},
	{0x718706651DD2231DULL, "vxht_mansion2_1de_0_0.vlm"},
	{0x7187354699033C63ULL, "vxdr_primevaltomb_829_0_0.vlm"},
	{0x71877756602B612CULL, "vxht_0_hightown_bq0_0_0.vlm"},
	{0x7187BC6D2A888BF2ULL, "zz_mrl_debug6184201.cl"},
	{0x71889BDF07614A70ULL, "6145206_m.fxe"},
	{0x7188AEE22C14208FULL, "6094015_m.fxe"},
	{0x71890F4E85033AC0ULL, "6084414_f.fxe"},
	{0x718916F827919915ULL, "vxdr_0_deeproads_evx_0_0.vlm"},
	{0x71899DBA2DBDE069ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_c.anb"},
	{0x7189BBED6CCC2ED9ULL, "static_static_p4.vsh.fxo"},
	{0x718A1F392258FB94ULL, "art\\vfx\\environment\\vfx_water_drips_p.dura.evt"},
	{0x718A4407EEE206C7ULL, "vxca_caverns1_mines_ckq_0_0.vlm"},
	{0x718AB32F883A2DD5ULL, "77391107.ogg"},
	{0x718B0ACE2C5A73C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_heal.ani"},
	{0x718B3A9D8332B919ULL, "dae000_mother6083985.cl"},
	{0x718B47F417647819ULL, "vxlt_warehouse_n_18t_0_0.vlm"},
	{0x718B533A42DB9A90ULL, "6151221_m.fxe"},
	{0x718BAB26D51DD76DULL, "vxw_encounter2_camp_78_0_0.vlm"},
	{0x718BACBB385AACDBULL, "vxlt_gallowstemplar_n_6qk_0_0.vlm"},
	{0x718C43C7DF1B0431ULL, "434565301.ogg"},
	{0x718C6EDAB3C70D79ULL, "393627919.ogg"},
	{0x718CF0F32753C8A6ULL, "vxlt_foundry_n_11n_0_0.vlm"},
	{0x718CF1FDFB83C3FDULL, "6100360_f.fxe"},
	{0x718D0B6DA38B1705ULL, "vxca_caverns1_mines_ii7_0_0.vlm"},
	{0x718D2E92633D40AFULL, "character_character_p277.psh.fxo.dx11"},
	{0x718DC830C07F5E92ULL, "6173098_m.fxe"},
	{0x718DF529BBFBBD77ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_qu101a_0n.dds"},
	{0x718E060198F1734AULL, "llac_w_bonepitmining_a2_0_8.dds"},
	{0x718E105B8AA943E3ULL, "hla_ht_0_hightown_a3_131070_0x.dds"},
	{0x718E4385CA4EC161ULL, "823771333.ogg"},
	{0x718FC1C3442F94F1ULL, "vxlt_hangedman_n_16j_0_0.vlm"},
	{0x718FF0B66370511CULL, "cd_pro000_merchant.wbd"},
	{0x718FFC86AB5BCBCDULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxb_sa.dds"},
	{0x719025D83FA8A8ACULL, "vxlt_gallowstemplar_5v5_0_0.vlm"},
	{0x71904EA2B8575A7EULL, "664394305.ogg"},
	{0x71906F350BCA4179ULL, "vxlt_0_lowtown_a2_n_c_gvd_0_0.vlm"},
	{0x7190BBA205FBBE29ULL, "622864636.ogg"},
	{0x7190CCB20056AD04ULL, "vxca_caverns1_icl_0_1.vlm"},
	{0x719111C85EA7DA74ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_rack_0.pwk"},
	{0x7191234089734256ULL, "dae000_orana6149064.cl"},
	{0x719125CE2C5F0D4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_heal.gad"},
	{0x719132427A360914ULL, "396417594.ogg"},
	{0x7191579C3FA192A8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_meda_0t.dds"},
	{0x7191C3068C354914ULL, "6102416_m.fxe"},
	{0x7191E46631406FE2ULL, "hlac_ht_mansion2_818_46z.dds"},
	{0x7191E853A895D019ULL, "6154155_m.fxe"},
	{0x71923C837863344EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_walls_e.dds"},
	{0x71926C5CEA116598ULL, "mag200_feynriel6023872.cl"},
	{0x7192BD36207558B7ULL, "vxht_bartrand_n_6rn_0_0.vlm"},
	{0x71948A3745D7D22EULL, "690440155.ogg"},
	{0x7194CB097E54EA8EULL, "vxw_blightlands_bw_0_0.vlm"},
	{0x7194DBC78CDDD211ULL, "llac_w_encounter1_-65532_0.dds"},
	{0x7194DDD1675FB1BDULL, "zz_followers_banter6157555.cl"},
	{0x7194E13AB379B0D0ULL, "t3_shd_irn.tnt"},
	{0x7194F3A74344B512ULL, "767171649.ogg"},
	{0x71952E26C3B51C60ULL, "vxlt_gallowscourtyard_n_3t8_0_0.vlm"},
	{0x719569141DC9E342ULL, "hlac_lt_darktown_-1_13z.dds"},
	{0x7195E933522179ACULL, "vxht_0_hightown_bz1_0_0.vlm"},
	{0x7195F15AE2CC82E6ULL, "6118787_m.fxe"},
	{0x7195F9DAD71997E5ULL, "6057352_f.fxe"},
	{0x719622C74964E423ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v4.evt"},
	{0x71964EB075CE17D1ULL, "6038990_m.fxe"},
	{0x719651E095B33B7BULL, "vxht_0_hightown_a3_aow_0_0.vlm"},
	{0x719674E322F053E4ULL, "gen00fl_anders6125228.cl"},
	{0x7196767C6381CFADULL, "6118326_m.fxe"},
	{0x71968FB58F3E7DC2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic.phy"},
	{0x7196BDEBF5FFEA4FULL, "588794020.ogg"},
	{0x719711E08147CF83ULL, "one220_yevhen6116311.cl"},
	{0x7197591A18F74EB4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sandrune_d.dds"},
	{0x7197CD7790F0E440ULL, "vxlt_warehouse_102_0_0.vlm"},
	{0x7198173922652BCDULL, "art\\vfx\\environment\\vfx_water_drips_p.dura.ani"},
	{0x719822CE2C6526B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_heal.evt"},
	{0x71983B6FE574F85CULL, "llac_ca_cave_2194_14.dds"},
	{0x7198B59E4EAA293EULL, "5251964.ogg"},
	{0x7199D80FEBE2D4AFULL, "6125715_m.fxe"},
	{0x7199FEADD13FF697ULL, "hlac_lt_0_lowtown_a2_n_65533_1z.dds"},
	{0x719A071109787A78ULL, "vxht_store_bw_0_0.vlm"},
	{0x719A15B2B5DE6C89ULL, "l_dae150ar_sundermount.bnk"},
	{0x719A30251FD5D3F6ULL, "gen00fl_isabela6081319.cl"},
	{0x719A7B85A9ED8460ULL, "vxlt_0_lowtown_docks_a3_n_cyd_0_0.vlm"},
	{0x719A9C1B24896EACULL, "54747355.ogg"},
	{0x719AB869512D33F6ULL, "vxca_caverns1_i1k_0_2.vlm"},
	{0x719ADAEB93FB4622ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_2z.dds"},
	{0x719AFA0375CA3D99ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03a.mmh"},
	{0x719B040375CA4E87ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03a.msh"},
	{0x719B49E7C5ACAE96ULL, "477807896.ogg"},
	{0x719BCFFC08050C3AULL, "6040781_m.fxe"},
	{0x719CA0E29D4B17CFULL, "6083823_f.fxe"},
	{0x719CC63637CA91A8ULL, "dae200_bran6094827.cl"},
	{0x719CCE30469D154BULL, "vxht_keep_6g5_0_0.vlm"},
	{0x719CE508F0EDAD83ULL, "vxw_woundedcoast_cc_0_0.vlm"},
	{0x719CEE2714E6D491ULL, "6121186_m.fxe"},
	{0x719CFBC7496AC0A4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v4.gad"},
	{0x719D7662573A7A1CULL, "art\\vfx\\combat\\vfx_glb_bld_imp_maj_c.anb"},
	{0x719D8BE603C471A0ULL, "lt_0_lowtown_docks_a3_n_10_0.anb"},
	{0x719DB6004B8F3982ULL, "hlac_w_encounter3_-131072_0x.dds"},
	{0x719DD4A083EDBC45ULL, "vxht_0_hightown_b82_0_0.vlm"},
	{0x719E521182BFC19DULL, "88501173.ogg"},
	{0x719EB714053A32AFULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelogs.mao"},
	{0x719EC314053A474CULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelogs.mmh"},
	{0x719EC914053A513EULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelogs.msh"},
	{0x719EF3501B0C4704ULL, "6111711_m.fxe"},
	{0x719F8933710AF29BULL, "dae370ip_at_templar.utp"},
	{0x71A017E494DDFA4EULL, "hlac_lt_0_lowtown_docks_-65537_0y.dds"},
	{0x71A0B689B39013FBULL, "vxlt_gallowstemplar_4ub_0_0.vlm"},
	{0x71A0FA37F929860CULL, "448509765.ogg"},
	{0x71A17FB2F808266BULL, "vxw_blightlands_127_0_0.vlm"},
	{0x71A1A909E2B50946ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\varric_crossbow_combat.anb"},
	{0x71A1ABC95AEBA1BEULL, "gen00fl_merrill6128924.cl"},
	{0x71A20124731A27E8ULL, "vxw_sundermount_2ed_0_0.vlm"},
	{0x71A21FDBCE919FAEULL, "6080428_m.fxe"},
	{0x71A2576F081DAA04ULL, "w_encounter3_124_0.msh"},
	{0x71A2696F081DC96DULL, "w_encounter3_124_0.mao"},
	{0x71A26D6F081DCFA6ULL, "w_encounter3_124_0.mmh"},
	{0x71A29483F2018FA1ULL, "hlac_lt_gallowstemplar_f_2384_5z.dds"},
	{0x71A2C881E0CBAD3AULL, "vxht_chantry_n_2c7_0_0.vlm"},
	{0x71A30D448DC5462EULL, "1031604746.ogg"},
	{0x71A3BF69298F7DDBULL, "vxht_0_hightown_a3_bp1_0_0.vlm"},
	{0x71A3D7AE20CC27ECULL, "vxlt_warehouse_ps_0_0.vlm"},
	{0x71A40AC74970F882ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v4.ani"},
	{0x71A43CF2BF5C97B3ULL, "lgt210_shp_kill_qunari16176796.cl"},
	{0x71A47BA1586DCC34ULL, "136709215.ogg"},
	{0x71A50154906F6F8AULL, "dae000ip_crowd_ff3_talk13.utp"},
	{0x71A50319730BC7A6ULL, "6144944_m.fxe"},
	{0x71A521C1670B657BULL, "6089739_f.fxe"},
	{0x71A5490ED0297ECEULL, "6170385_m.fxe"},
	{0x71A59CFD2F30BE50ULL, "500946277.ogg"},
	{0x71A604BBD408826EULL, "196641807.ogg"},
	{0x71A633B34B767FDBULL, "vxlt_0_lowtown_a2_n_hd9_0_0.vlm"},
	{0x71A6632B4F082798ULL, "888308805.ogg"},
	{0x71A679E4C608E4FBULL, "vxht_0_hightown_can_0_0.vlm"},
	{0x71A6EB49A5A65C9BULL, "vxlt_0_lowtown_docks_a3_d35_0_0.vlm"},
	{0x71A718125A221E81ULL, "50799498.ogg"},
	{0x71A7491B82D6C802ULL, "pro000_merchant6051848.cl"},
	{0x71A7AD29EC3CF65CULL, "lgt220_shp_mag_guard6175910.cl"},
	{0x71A7ED9AF9A4E2E3ULL, "lt_0_lowtown_docks_142_0.msh"},
	{0x71A7FB9AF9A4FA3DULL, "lt_0_lowtown_docks_142_0.mmh"},
	{0x71A7FF9AF9A5014EULL, "lt_0_lowtown_docks_142_0.mao"},
	{0x71A852CE3F0A74F9ULL, "6165472_m.fxe"},
	{0x71A86F5A78DFA4E2ULL, "29215120.ogg"},
	{0x71A8F3A3AF80E5F1ULL, "23122975.ogg"},
	{0x71A932FA20EF805EULL, "6093492_f.fxe"},
	{0x71A94CF46F0BF74BULL, "6165179_m.fxe"},
	{0x71A9D827C39DC51BULL, "vxlt_0_lowtown_docks_a3_apm_0_0.vlm"},
	{0x71A9EEFA08296225ULL, "vxht_0_hightown_n_as3_0_0.vlm"},
	{0x71AA09AB33C79956ULL, "450201974.ogg"},
	{0x71AA9C46099B035EULL, "6042986_f.fxe"},
	{0x71AAB6D08B1D413DULL, "sst_ss_wilmod.gda"},
	{0x71AB86571A984B9FULL, "gen00fl_dog6165633.cl"},
	{0x71AB90A94293E275ULL, "gen00fl_sebastian6180807.cl"},
	{0x71AB9589CB443E0BULL, "art\\vfx\\cinematic\\vfx_merrill_reflect.mmh"},
	{0x71ABD53C44AE697DULL, "vxdr_0_deeproads_6qv_0_0.vlm"},
	{0x71AC16DB90E52E79ULL, "136507456.ogg"},
	{0x71AC486B10D76244ULL, "367263447.ogg"},
	{0x71AC5F0B33226E82ULL, "486182456.ogg"},
	{0x71AC6285F5295A28ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2plank.msh"},
	{0x71AC7485F52978B9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2plank.mao"},
	{0x71AC8085F5298D52ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2plank.mmh"},
	{0x71ACAF7C4A1EF96DULL, "dae020ip_merch_armor.utp"},
	{0x71ACE2C0B8D18A95ULL, "gen00fl_carver6102659.cl"},
	{0x71AD2ECA45B7B3E2ULL, "717286272.ogg"},
	{0x71AD8945C82ADA21ULL, "zz_mrl_debug6185311.cl"},
	{0x71ADA04CF8179927ULL, "llac_lt_0_lowtown_a2_-65535_3.dds"},
	{0x71ADE41C48BCF656ULL, "vxht_mansion_n_3ov_0_0.vlm"},
	{0x71AE82703A64E7A9ULL, "vxca_caverns2_7uk_0_0.vlm"},
	{0x71AEC5459DEA7402ULL, "zz_approval_debug6003964.cl"},
	{0x71AEDEAFB439181BULL, "6128077_m.fxe"},
	{0x71AF13C0D68003D1ULL, "hlac_w_bonepitdestroyed_a3_393218_0y.dds"},
	{0x71AF32FB1649BE30ULL, "zz_dae_debug6142945.cl"},
	{0x71AFA22A479A9069ULL, "art\\levels\\races\\vfx_models\\prp_brazlogbits.mao"},
	{0x71AFAE2A479AA402ULL, "art\\levels\\races\\vfx_models\\prp_brazlogbits.mmh"},
	{0x71AFB02A479AA7B8ULL, "art\\levels\\races\\vfx_models\\prp_brazlogbits.msh"},
	{0x71AFF9C3A21A44E2ULL, "6019093_m.fxe"},
	{0x71B01AE33C70B993ULL, "zz_mer_debug6073699.cl"},
	{0x71B0730165104068ULL, "vxw_sundermount_2ne_0_0.vlm"},
	{0x71B07E915283F85BULL, "6199752_m.fxe"},
	{0x71B09E69DE6B3515ULL, "vxlt_warehouse_129_0_0.vlm"},
	{0x71B0B2445177F252ULL, "w_encounter3_8_0.rml"},
	{0x71B0C0B3C6772887ULL, "912801930.ogg"},
	{0x71B0E0832D0B6973ULL, "vxlt_foundry_n_za_0_0.vlm"},
	{0x71B0F71D8ED6F33CULL, "vxlt_0_lowtown_a2_bxk_0_0.vlm"},
	{0x71B134B14AF1C16EULL, "w_sundermount_173_0.msh"},
	{0x71B142B14AF1D8BFULL, "w_sundermount_173_0.mao"},
	{0x71B1494E20523687ULL, "fen210_tevinter_ambusher26108639.cl"},
	{0x71B14EB14AF1ED5CULL, "w_sundermount_173_0.mmh"},
	{0x71B189CE970A3A9BULL, "vxca_caverns1_drd_0_0.vlm"},
	{0x71B19637BEB66510ULL, "mcr370_assignments6081701.cl"},
	{0x71B1D698966E7647ULL, "character_character_p202.psh.fxo.dx11.assembly"},
	{0x71B24CECB5C90FDFULL, "vxlt_hangedman_n_1db_0_0.vlm"},
	{0x71B282930990E8F2ULL, "qcr210cr_elven_supporters.utc"},
	{0x71B2C11A71BE004DULL, "6093647_m.fxe"},
	{0x71B30E750F3E88A4ULL, "hlac_ht_keep_4519_8z.dds"},
	{0x71B32B6F65792D3CULL, "lt_0_lowtown_docks_25_0.anb"},
	{0x71B3359A799F6C7BULL, "vxht_0_hightown_a3_aqi_0_0.vlm"},
	{0x71B35A36476E00ADULL, "6112126_m.fxe"},
	{0x71B37883DB37E542ULL, "vxw_sundermount_212_0_0.vlm"},
	{0x71B3D018E1EF04FEULL, "394583710.ogg"},
	{0x71B3F81EF7ED0754ULL, "464707705.ogg"},
	{0x71B3FE7A7DBC84D0ULL, "608525189.ogg"},
	{0x71B4333D07953EC5ULL, "llac_w_encounter3_-327681_0.dds"},
	{0x71B46D041BAD9ADBULL, "clf360_evets6072217.cl"},
	{0x71B4D7A0ACDA57AAULL, "vxlt_gallowstemplar_n_4qh_0_0.vlm"},
	{0x71B4EBC1B7FEFD7BULL, "vxht_0_hightown_cda_0_0.vlm"},
	{0x71B500CD73F7C574ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.dura.ani"},
	{0x71B5395158D5910FULL, "t3_arm_rst.tnt"},
	{0x71B5472AB41C353EULL, "hlac_w_encounter3_0_9x.dds"},
	{0x71B5560C97B91659ULL, "vxht_keep_n_6mz_0_0.vlm"},
	{0x71B55D26979C751BULL, "vxlt_0_lowtown_docks_a3_d44_0_0.vlm"},
	{0x71B5832C96FC06A5ULL, "vxdr_primevalentrance_2e4_0_0.vlm"},
	{0x71B5941F7D896CA6ULL, "vxht_0_hightown_a2_7g5_0_0.vlm"},
	{0x71B5BFAC67AA2883ULL, "gst_standing_right.gda"},
	{0x71B628A9B78BC418ULL, "dae000_gamlen6149432.cl"},
	{0x71B68F81E49BEFB1ULL, "gen00fl_aveline6152525.cl"},
	{0x71B6E822470DA7F9ULL, "fen210_orana6109690.cl"},
	{0x71B780D7753879F9ULL, "6130103_f.fxe"},
	{0x71B7864E5D9A0558ULL, "6031273_m.fxe"},
	{0x71B7BFD7FE1147A5ULL, "gencr_corpse_archer.utc"},
	{0x71B8DA22900EFAF0ULL, "art\\levels\\ztools\\default\\generic\\0do_angle.mao"},
	{0x71B8DE22900F0103ULL, "art\\levels\\ztools\\default\\generic\\0do_angle.mmh"},
	{0x71B8EB9B5A3AF485ULL, "qun160_vashothlead6053956.cl"},
	{0x71B8EC22900F195DULL, "art\\levels\\ztools\\default\\generic\\0do_angle.msh"},
	{0x71B96664D94A573DULL, "215493926.ogg"},
	{0x71B97689CB504770ULL, "art\\vfx\\cinematic\\vfx_merrill_reflect.anb"},
	{0x71B9B1EA2A89464BULL, "vxht_keep_6m7_0_0.vlm"},
	{0x71B9C8C2D4D9DE83ULL, "vxw_woundedcoast_ue_0_0.vlm"},
	{0x71B9CA830E463228ULL, "vxlt_undercityencounter_10u_0_0.vlm"},
	{0x71B9EFEEFEE50833ULL, "hla_ht_0_hightown_n_0_0z.dds"},
	{0x71BA0375A9715741ULL, "6090057_m.fxe"},
	{0x71BA281C1B9DCFF6ULL, "40446495.ogg"},
	{0x71BAFDCD73FC2BD6ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.dura.gad"},
	{0x71BB527ACE3C598CULL, "6140310_f.fxe"},
	{0x71BB6AB58F5D5B76ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic.mao"},
	{0x71BB76B58F5D6F15ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic.mmh"},
	{0x71BB78B58F5D728BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4so_basic.msh"},
	{0x71BBB628F9945A66ULL, "vxlt_0_lowtown_docks_b5j_0_1.vlm"},
	{0x71BBEEB663958896ULL, "cd_pro000_merchant.bnk"},
	{0x71BC88376F0FAA21ULL, "vxlt_warehouse_150_0_0.vlm"},
	{0x71BC947DAC79C2BBULL, "gen00fl_aveline6114631.cl"},
	{0x71BD1A2691273CDCULL, "one100dg_nabil6033128.cl"},
	{0x71BD369ABEE0EB39ULL, "vxlt_hangedman_n_265_0_0.vlm"},
	{0x71BD47468669B3E4ULL, "90686988.ogg"},
	{0x71BD73708193E834ULL, "895097459.ogg"},
	{0x71BDEECC79A35AD6ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.dura.ani"},
	{0x71BDF419E880E23AULL, "418968268.ogg"},
	{0x71BEB61579A1740EULL, "68193207.ogg"},
	{0x71BECDB7FE36C03AULL, "425916566.ogg"},
	{0x71BEE54F88C1BE70ULL, "follower_banter6170005.cl"},
	{0x71BEF28C335FCFC6ULL, "dae000_mother6121020.cl"},
	{0x71BFB1F105469CA7ULL, "hlac_lt_0_lowtown_a2_-1_0z.dds"},
	{0x71C00B1BD30867F2ULL, "327537009.ogg"},
	{0x71C0887310CBE645ULL, "dae000ip_crowd_ff1_talk12.utp"},
	{0x71C0B108F2421E30ULL, "6031659_m.fxe"},
	{0x71C19A8FE253CE0BULL, "fall_cloud.mao"},
	{0x71C1D3B11E594013ULL, "art\\characters\\creatures\\stonegolem\\animation\\dazed.ani"},
	{0x71C218CD74027835ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.dura.evt"},
	{0x71C21C264F988501ULL, "follower_banter6163114.cl"},
	{0x71C251B31B016B71ULL, "hlac_ca_caverns1_24421_10x.dds"},
	{0x71C2AA69CB1B3573ULL, "vxlt_gallowstemplar_f_5qh_0_0.vlm"},
	{0x71C2E5008FA9310BULL, "6090449_f.fxe"},
	{0x71C2F4A75938464DULL, "and121_karl6093885.cl"},
	{0x71C3B6AD2C83F558ULL, "398389020.ogg"},
	{0x71C3C7E989AF2ED9ULL, "vxht_keep_n_6hy_0_0.vlm"},
	{0x71C3D3CFB0D31684ULL, "art\\vfx\\combat\\vfx_w_wns_right_b_close1_c.anb"},
	{0x71C3D7606C5B5E82ULL, "ave331cr_insurgent_f.utc"},
	{0x71C487F042507292ULL, "one220_yevhen6115735.cl"},
	{0x71C4ED7D3E2E27BEULL, "llac_lt_0_lowtown_docks_n_-65538_7.dds"},
	{0x71C52916A9A99F0FULL, "lgt110_arl_bann_teagan6115596.cl"},
	{0x71C538FDB9AE3E65ULL, "vxlt_warehouse_18k_0_0.vlm"},
	{0x71C5AB3ED1936F5BULL, "fen210_orana6133033.cl"},
	{0x71C5ECBCA50202B6ULL, "6106322_m.fxe"},
	{0x71C60BEFF0C560A7ULL, "vxdr_primevaltomb_4hk_0_0.vlm"},
	{0x71C662F84AA3CE2CULL, "305549147.ogg"},
	{0x71C66820B3609889ULL, "vxht_brothel_n_52o_0_0.vlm"},
	{0x71C75702A6B0337CULL, "plt_cod_tomes_tome.plo"},
	{0x71C76950D853EBDEULL, "hlac_lt_gallowscourtyard_n_65534_0z.dds"},
	{0x71C77ABEC471DB00ULL, "zz_otl_debug6060659.cl"},
	{0x71C7A4F02448A769ULL, "hlac_w_bonepitdestroyed_a3_0_8y.dds"},
	{0x71C8A3143FF9FBEDULL, "plt_lgt000pt_baz_sketchy.nss"},
	{0x71C8FCF29F4F21A4ULL, "item_ranks.gda"},
	{0x71C94D16464D4E0EULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden6.mmh"},
	{0x71C96716464D7AFCULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden6.msh"},
	{0x71C9ACAAD244EA3DULL, "6129915_m.fxe"},
	{0x71C9DF22901D7628ULL, "art\\levels\\ztools\\default\\generic\\0do_angle.phy"},
	{0x71CA0FE32C4B8A29ULL, "vxht_chantry_351_0_1.vlm"},
	{0x71CAE0228E506032ULL, "6121137_f.fxe"},
	{0x71CAF6CC79ADF2C7ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.dura.evt"},
	{0x71CBAB60557209F8ULL, "6060342_f.fxe"},
	{0x71CC00E870BFC733ULL, "6175066_m.fxe"},
	{0x71CC56DFFB0A25AAULL, "vxht_0_hightown_a3_7u2_0_0.vlm"},
	{0x71CC675E91142511ULL, "6041489_f.fxe"},
	{0x71CCEB96ED8675E6ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_1z.dds"},
	{0x71CD00972E6D08D2ULL, "625997750.ogg"},
	{0x71CD4B8334BCE80DULL, "6018848_m.fxe"},
	{0x71CD95CAFA0AAB07ULL, "ave100_donnic6105956.cl"},
	{0x71CEE46F13F0C539ULL, "genip_cin_qun_control.utp"},
	{0x71CF2B0A0A21F5F4ULL, "lt_0_lowtown_a2_n_37_0_lc.mmh"},
	{0x71CF4287544CD06FULL, "hlac_ht_mansion2_818_66y.dds"},
	{0x71CF96481C35FD2EULL, "dae000_gamlen6022562.cl"},
	{0x71CFCBB11E65704AULL, "art\\characters\\creatures\\stonegolem\\animation\\dazed.evt"},
	{0x71CFD8F1E5776459ULL, "vxlt_0_lowtown_a2_emv_0_0.vlm"},
	{0x71CFE6769320F77EULL, "ice_shard_pet.mao"},
	{0x71D00E18BA48DA9AULL, "448959252.ogg"},
	{0x71D03831036A5508ULL, "hlac_lt_0_lowtown_docks_n_-131073_5z.dds"},
	{0x71D112E4C1ED57DCULL, "6082796_m.fxe"},
	{0x71D13B10544B66C5ULL, "one100dg_kelder6060537.cl"},
	{0x71D13D26E0B2DA3CULL, "lt_0_lowtown_docks_a3_n_197_0.mao"},
	{0x71D14926E0B2EEDFULL, "lt_0_lowtown_docks_a3_n_197_0.mmh"},
	{0x71D14F26E0B2F881ULL, "lt_0_lowtown_docks_a3_n_197_0.msh"},
	{0x71D1F2CFB0DF8887ULL, "art\\vfx\\combat\\vfx_w_wns_right_b_close1_c.mmh"},
	{0x71D2078A88B86B80ULL, "6087051_f.fxe"},
	{0x71D217CC79B44910ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.dura.gad"},
	{0x71D2459F5CD04E29ULL, "cth_cm4.tnt"},
	{0x71D2A8DE8A8BC69BULL, "6044375_m.fxe"},
	{0x71D3736BEADE6EB7ULL, "1038900387.ogg"},
	{0x71D376C889489888ULL, "109433950.ogg"},
	{0x71D3982BBB33AC48ULL, "675695859.ogg"},
	{0x71D4090038EC593AULL, "vxca_cave_18f_0_0.vlm"},
	{0x71D432F91DCF78FCULL, "135546493.ogg"},
	{0x71D43F3786B2A258ULL, "669417625.ogg"},
	{0x71D59B51B7917794ULL, "17571777.ogg"},
	{0x71D62DB14B10D127ULL, "w_sundermount_173_0.phy"},
	{0x71D630F8C3BDA02DULL, "art\\characters\\playercharacter\\humanmale\\hoo_chab.mmh"},
	{0x71D634F8C3BDA77EULL, "art\\characters\\playercharacter\\humanmale\\hoo_chab.mao"},
	{0x71D64120DDC264C5ULL, "qun110_petrice6059149.cl"},
	{0x71D642F8C3BDBEB3ULL, "art\\characters\\playercharacter\\humanmale\\hoo_chab.msh"},
	{0x71D64F190D9B07D8ULL, "lt_0_lowtown_docks_a3_n_25_0.rml"},
	{0x71D65B805F1801A8ULL, "em_huon.mor"},
	{0x71D6AAB11E6B56BDULL, "art\\characters\\creatures\\stonegolem\\animation\\dazed.gad"},
	{0x71D6E227F0EDA0FCULL, "hlac_lt_0_lowtown_a3_n_c_-1_7x.dds"},
	{0x71D7DBE9EB6556C0ULL, "gen_im_wep_rog_dua_act2_03.uti"},
	{0x71D845555F93D182ULL, "llac_w_encounter3_-131070_0.dds"},
	{0x71D8A8F744C9E23DULL, "vxlt_darktown_6li_0_0.vlm"},
	{0x71D8FA67071294EBULL, "vxlt_0_lowtown_docks_a3_d1o_0_0.vlm"},
	{0x71D954E8FAF71B5EULL, "66619130.ogg"},
	{0x71DA27C64218EFEBULL, "hlac_lt_0_lowtown_a2_n_0_3z.dds"},
	{0x71DB5BB62A9FADBEULL, "vxht_bartrand_n_322_0_0.vlm"},
	{0x71DBB467DDC87E13ULL, "6095382_m.fxe"},
	{0x71DBB9F40ABDED6BULL, "vxlt_darktown_2sf_0_0.vlm"},
	{0x71DBCCB645DEBA83ULL, "420483246.ogg"},
	{0x71DBDFEC080FEBFDULL, "vxlt_0_lowtown_docks_n_c5n_0_0.vlm"},
	{0x71DC2DC0605D6F5DULL, "195811435.ogg"},
	{0x71DC42CB5DF3BDE5ULL, "follower_banter6169800.cl"},
	{0x71DC8FE6669F0126ULL, "6171446_f.fxe"},
	{0x71DC9A498A821C84ULL, "dae000cr_grand_cleric6147667.cl"},
	{0x71DCA2DEE95EE0EAULL, "llac_lt_0_lowtown_a2_n_1_1.dds"},
	{0x71DCB58FED659CEEULL, "vxlt_0_lowtown_a2_f66_0_1.vlm"},
	{0x71DE05794D0EA93EULL, "vxlt_gallowstemplar_n_519_0_0.vlm"},
	{0x71DE1AA317226304ULL, "mcr371_post_harvester6077599.cl"},
	{0x71DE269A4A65250EULL, "vxw_sundermount_2jo_0_0.vlm"},
	{0x71DE4B5DBCE0C1DDULL, "357762076.ogg"},
	{0x71DE5B01EBDBD397ULL, "vxlt_0_lowtown_a3_n_c_fag_0_0.vlm"},
	{0x71DE621E09A3933AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_busha.mmh"},
	{0x71DE6AE603FBD0C8ULL, "lt_0_lowtown_docks_a3_n_10_0.rml"},
	{0x71DE6E3A2DCDD1F0ULL, "gen00fl_aveline6116636.cl"},
	{0x71DE741E09A3B180ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_busha.msh"},
	{0x71DEA4A674A4E944ULL, "vxlt_warehouse_n_15r_0_0.vlm"},
	{0x71DEB0BDB5CA499CULL, "194843164.ogg"},
	{0x71DEE4231100136BULL, "lt_0_lowtown_a2_n_19_0.anb"},
	{0x71DF35BBC459519EULL, "t3_shl_fir.tnt"},
	{0x71DF84C1B3E3705CULL, "6085696_m.fxe"},
	{0x71DF9021C3996889ULL, "vxca_caverns1_ei7_0_1.vlm"},
	{0x71DFBF92AB9310BEULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_lgta_0d.dds"},
	{0x71DFD71A29CC7428ULL, "mer221_sabin6047453.cl"},
	{0x71DFFAAECA3A37D4ULL, "dre000_varric6208034.cl"},
	{0x71E0CB5F68B6B784ULL, "vxlt_0_lowtown_docks_a3_9fh_0_0.vlm"},
	{0x71E19FC9B6392979ULL, "390396061.ogg"},
	{0x71E1F3B2B61BE8DBULL, "l_dae150ar_sundermount.wbd"},
	{0x71E20E43F4C85E31ULL, "vxlt_0_lowtown_a2_n_c_gjx_0_0.vlm"},
	{0x71E21CB79D9A6F65ULL, "vxlt_warehouse_12i_0_0.vlm"},
	{0x71E24902BC718F57ULL, "240604915.ogg"},
	{0x71E281B3A50CE716ULL, "543148649.ogg"},
	{0x71E305E1A78AE421ULL, "937654337.ogg"},
	{0x71E33A84F3C204B7ULL, "gen_im_trs_qun_two_2hs_r0.uti"},
	{0x71E34BDA974CC989ULL, "vxht_brothel_n_58m_0_0.vlm"},
	{0x71E4172A73DBF053ULL, "vxw_woundedcoast_1dw_0_0.vlm"},
	{0x71E44385A0E60E6CULL, "80437592.ogg"},
	{0x71E4C18295CBCB8BULL, "plt_genpt_amb_notif.plo"},
	{0x71E4F7D46B22C873ULL, "hlac_lt_warehouse_n_65537_5z.dds"},
	{0x71E52849B5A9F3DCULL, "hlac_ht_keep_n_3732_10z.dds"},
	{0x71E5485438BD529BULL, "dae000_meredith6166313.cl"},
	{0x71E5C277571A7B45ULL, "w_bonepitdestroyed_a3_2_0.rml"},
	{0x71E5E1946CF0E127ULL, "6169405_m.fxe"},
	{0x71E68CB85DADDA69ULL, "hlac_lt_0_lowtown_docks_131073_1y.dds"},
	{0x71E694AC124BBB1EULL, "ip_restrict_style_twohanded.ncs"},
	{0x71E69E30E578F844ULL, "6140937_m.fxe"},
	{0x71E729F8C3CC07EAULL, "art\\characters\\playercharacter\\humanmale\\hoo_chab.phy"},
	{0x71E73463D3AD8403ULL, "hlac_ht_playershouse_n_855_17z.dds"},
	{0x71E772E6732FA5B4ULL, "gen00fl_isabela6098218.cl"},
	{0x71E8703A3B09DC59ULL, "328675183.ogg"},
	{0x71E8ADAB3C8F908AULL, "vxdr_primevalentrance_2ok_0_2.vlm"},
	{0x71E9128C94EDE1B2ULL, "869384439.ogg"},
	{0x71E93B7D1A7DE6D0ULL, "art\\vfx\\combat\\vfx_gib_ogr_p_1.mao"},
	{0x71E962EADA5014C6ULL, "vxlt_0_lowtown_a2_e1y_0_0.vlm"},
	{0x71EA0D3C4F0A276CULL, "llac_lt_undercityserialkiller_-1_28.dds"},
	{0x71EAE6045E8E6686ULL, "511491735.ogg"},
	{0x71EB58FBF761E17AULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.dura.gad"},
	{0x71EB6232A99935ACULL, "lt_0_lowtown_docks_n_mmap.dds"},
	{0x71EBE5A218BE7C6FULL, "sib111_mother.bnk"},
	{0x71EBFF731F55F2C3ULL, "mag210_arianni6024583.cl"},
	{0x71ED27CE22E7A913ULL, "hlac_ht_store_187_4z.dds"},
	{0x71ED70BB0214BD0BULL, "vxlt_darktown_492_0_0.vlm"},
	{0x71ED71AC8365B032ULL, "394796620.ogg"},
	{0x71ED97E035308ED2ULL, "vxw_woundedcoast_1dr_0_0.vlm"},
	{0x71EE0A3DA7E55B3BULL, "llac_w_bonepit_0_1.dds"},
	{0x71EE6AA4DD7D98E5ULL, "gen00fl_aveline6128555.cl"},
	{0x71EECDA2CFDDC18CULL, "vxht_brothel_n_5eg_0_0.vlm"},
	{0x71EF07A218C0EBD7ULL, "sib111_mother.cnv"},
	{0x71EF136F7AB05640ULL, "566694415.ogg"},
	{0x71EF23F0338181D1ULL, "6147141_f.fxe"},
	{0x71EF88798F7CF2D6ULL, "6061775_m.fxe"},
	{0x71F048BD47A13430ULL, "art\\vfx\\combat\\vfx_gib_qunc_3.mao"},
	{0x71F0BE0F4CE7DD41ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junka_sa.dds"},
	{0x71F10BC5E4861309ULL, "vxht_0_hightown_a2_c6u_0_0.vlm"},
	{0x71F1370C41D8C390ULL, "art\\vfx\\models\\vfx_trap_spikes.msh"},
	{0x71F1450C41D8DBAAULL, "art\\vfx\\models\\vfx_trap_spikes.mmh"},
	{0x71F1490C41D8E2E1ULL, "art\\vfx\\models\\vfx_trap_spikes.mao"},
	{0x71F15BFBF7665191ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.dura.evt"},
	{0x71F1D59A9C1773DEULL, "hlac_ht_mansion2b_818_57z.dds"},
	{0x71F24FF304144834ULL, "zz_arishok_headlook_test.cut"},
	{0x71F2890765D208D3ULL, "vxw_woundedcoast_1ov_0_0.vlm"},
	{0x71F2CCE43EA5AAF8ULL, "6111540_m.fxe"},
	{0x71F37D605A0E1BE1ULL, "256396527.ogg"},
	{0x71F38022D084BCFCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_2.pwk"},
	{0x71F3FE15DD543D9FULL, "29057323.ogg"},
	{0x71F409EA174EAB59ULL, "6040812_m.fxe"},
	{0x71F42A6F65B0C2CCULL, "lt_0_lowtown_docks_25_0.rml"},
	{0x71F43CBA083D9F5AULL, "vxht_bartrand_n_72r_0_0.vlm"},
	{0x71F44F303C462E5FULL, "6205503_m.fxe"},
	{0x71F456F7FDBD63EDULL, "ter_stoney_d.dds"},
	{0x71F4586365A49CCDULL, "mag321_comtesse_de_launcet6043325.cl"},
	{0x71F53FD70C7AAF18ULL, "6090445_m.fxe"},
	{0x71F5A1D4D5C550E3ULL, "vxlt_0_lowtown_gcl_0_0.vlm"},
	{0x71F5CFC702731B63ULL, "6043495_m.fxe"},
	{0x71F62284A67D45D5ULL, "mrl310_party_comments6142630.cl"},
	{0x71F65B0A1E6706B2ULL, "vxdr_0_deeproads_9da_0_0.vlm"},
	{0x71F7435632BBB08AULL, "gen00fl_fenris6126434.cl"},
	{0x71F76A4A028B1DA9ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_ferelden_jimmy_n.dds"},
	{0x71F7A2629165CCF6ULL, "55200857.ogg"},
	{0x71F7D289225F6C43ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basicnb.msh"},
	{0x71F7E089225F839DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basicnb.mmh"},
	{0x71F7E489225F8AAEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1pp_basicnb.mao"},
	{0x71F7FF1FEA2F018DULL, "fly.dds"},
	{0x71F809F62A828D5CULL, "6165617_m.fxe"},
	{0x71F8152C813AB10DULL, "vxht_bartrand_n_3r5_0_0.vlm"},
	{0x71F826EA71F0E61BULL, "vrc221_bartrand6047717.cl"},
	{0x71F82D51662ED68BULL, "and111ip_store_lirene.utp"},
	{0x71F8957DDF69AEBCULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.impa.ani"},
	{0x71F8A6611D89A837ULL, "6170379_m.fxe"},
	{0x71F8C941F57206F0ULL, "vxht_chantry_34v_0_0.vlm"},
	{0x71F8D6D6590AEF55ULL, "hlac_lt_backalley_g_65537_7z.dds"},
	{0x71F940BC6F2EF949ULL, "6000867_m.fxe"},
	{0x71F9AA627D340C2EULL, "vxht_mansion2b_30d_0_0.vlm"},
	{0x71F9C1507F77BAF3ULL, "6163524_m.fxe"},
	{0x71F9FA27F175C2C1ULL, "201072829.ogg"},
	{0x71FA6154F059A6CDULL, "zz_isa_debug6133183.cl"},
	{0x71FA6D7743AADC4DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wrba_0n.dds"},
	{0x71FB0A542E51560EULL, "vxw_sundermount_2pa_0_0.vlm"},
	{0x71FB1CA960AD6940ULL, "plt_mer200pt_stolen_goods.plo"},
	{0x71FB66A7B3C2C8A4ULL, "247240793.ogg"},
	{0x71FC09BD2726A752ULL, "vxw_woundedcoast_1os_0_0.vlm"},
	{0x71FCFB03761D0986ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03a.phy"},
	{0x71FCFE603A77F797ULL, "vxca_caverns1_i14_0_0.vlm"},
	{0x71FD3F7FC1D3DA0CULL, "vxht_brothel_n_5hd_0_0.vlm"},
	{0x71FDAE1798D0E1ECULL, "vxlt_blackemporium_13s_0_0.vlm"},
	{0x71FDE74AB614DF4AULL, "1072751270.ogg"},
	{0x71FDEFB54A16516BULL, "313467195.ogg"},
	{0x71FE33874CDFDFCDULL, "6150366_m.fxe"},
	{0x71FE56AC190E7A93ULL, "hlac_w_sundermount_-262142_0z.dds"},
	{0x71FE72D4FA4D8818ULL, "vxht_keep_4h9_0_0.vlm"},
	{0x71FE9A7FC1448B98ULL, "dae000cr_grand_cleric6124691.cl"},
	{0x71FF035B81132E43ULL, "vxht_0_hightown_ava_0_0.vlm"},
	{0x71FF329065B353A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_head_up.gad"},
	{0x71FF73FBF772B860ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.dura.ani"},
	{0x71FFA27DDF6FE3FEULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.impa.gad"},
	{0x71FFFA0C046BD0A0ULL, "163299282.ogg"},
	{0x7200583FC40835ECULL, "dae000_mother6099206.cl"},
	{0x72008E2560A1CEA8ULL, "22150835.ogg"},
	{0x7200C0DA8E4481E9ULL, "6119017_m.fxe"},
	{0x7200CD3DB737CCC1ULL, "llac_lt_0_lowtown_docks_a3_n_-65538_7.dds"},
	{0x7200FAE457C82153ULL, "vxw_woundedcoast_1jy_0_0.vlm"},
	{0x720120C2215C5943ULL, "hlac_ht_brothel_n_3382_0z.dds"},
	{0x72012D6E2D932E22ULL, "one000_ilen6060496.cl"},
	{0x7201A17DC5FC6EAAULL, "vxht_mansion2_3br_0_0.vlm"},
	{0x7201F2FE298696EBULL, "185687439.ogg"},
	{0x720230190DC00D90ULL, "lt_0_lowtown_docks_a3_n_25_0.anb"},
	{0x72023A30BB83FA5FULL, "vxca_cavern1_haunted_egw_0_0.vlm"},
	{0x7202869F3B6A23F0ULL, "716135885.ogg"},
	{0x7202B7AEB79707D0ULL, "6148717_m.fxe"},
	{0x7202C54E50DD1227ULL, "6163605_m.fxe"},
	{0x720391BEA17A5EB2ULL, "vxlt_0_lowtown_n_fea_0_0.vlm"},
	{0x7203BD70ED6E95F1ULL, "gen_im_trs_krk_one_mac_r2.uti"},
	{0x7203D1C449703579ULL, "plt_cod_ltr_sk3_scribbles3.plo"},
	{0x7203F2843B2F5399ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.cess.ani"},
	{0x720435ACA0F73A35ULL, "6071250_m.fxe"},
	{0x7204677D0BBAB006ULL, "vxht_0_hightown_a3_bg1_0_0.vlm"},
	{0x720538236983DC01ULL, "6126889_f.fxe"},
	{0x720543BF0D7B22E6ULL, "zz_iconicarmor.ncs"},
	{0x720567B57CBB8805ULL, "6060469_m.fxe"},
	{0x72056D85BC0CD6BDULL, "follower_banter6173054.cl"},
	{0x7205795552530DE1ULL, "dae000_arishok6068458.cl"},
	{0x7205A551B299338CULL, "6153168_m.fxe"},
	{0x7205BFCE3DD08AF1ULL, "hlac_ca_caverns1_24426_1y.dds"},
	{0x72061CCAA78B0464ULL, "6035357_f.fxe"},
	{0x72067D7DDF75C31DULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.impa.evt"},
	{0x7206A2F847C66884ULL, "6034029_m.fxe"},
	{0x7206AC356F7831C6ULL, "262268534.ogg"},
	{0x7206D9564EAD42ADULL, "gen00fl_aveline6131218.cl"},
	{0x7207C904D7A1EBEBULL, "hlac_lt_foundry_n_320_0z.dds"},
	{0x72088032E1E7F3C1ULL, "vxlt_0_lowtown_a2_ewh_0_0.vlm"},
	{0x7208A5F5FBAE0D52ULL, "726219834.ogg"},
	{0x7208FA9A2F6D62CCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgta_0t.dds"},
	{0x7209108017D7710DULL, "ss_circlemages_m_2.bnk"},
	{0x720915661D38718BULL, "738176029.ogg"},
	{0x72093849FE0603B2ULL, "752966780.ogg"},
	{0x720971397BB5843BULL, "vxdr_primevaltomb_7o5_0_0.vlm"},
	{0x72097C3120476E8EULL, "vxw_sundermount_2ub_0_0.vlm"},
	{0x720A15CC2EE9278EULL, "gen00fl_fenris6126896.cl"},
	{0x720A253685859202ULL, "hlac_lt_warehouse_n_1_4y.dds"},
	{0x720A89762376E0C8ULL, "vxht_keep_n_5zy_0_0.vlm"},
	{0x720AB83E29A6C6D2ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.mmh"},
	{0x720ACD843B3533DBULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.cess.gad"},
	{0x720AE123112548B7ULL, "lt_0_lowtown_a2_n_19_0.rml"},
	{0x720B88EF3F7A7C6AULL, "6170507_m.fxe"},
	{0x720BA277573B146EULL, "w_bonepitdestroyed_a3_2_0.mmh"},
	{0x720BAE77573B2805ULL, "w_bonepitdestroyed_a3_2_0.mao"},
	{0x720BBC77573B405CULL, "w_bonepitdestroyed_a3_2_0.msh"},
	{0x720C1F67CE4DC5DEULL, "vxlt_hangedman_25l_0_0.vlm"},
	{0x720CB58961F8D582ULL, "6036501_f.fxe"},
	{0x720D4A78EFB54CDDULL, "hlac_lt_0_lowtown_docks_n_65535_3y.dds"},
	{0x720E08D3D7501F8DULL, "171522289.ogg"},
	{0x720E93EA84C53FC9ULL, "6020388_m.fxe"},
	{0x720EB5276E6C9DFAULL, "llac_ht_keep_3732_8.dds"},
	{0x720FCB80BD597D07ULL, "6076352_m.fxe"},
	{0x721023EB6C4BD2F0ULL, "570230319.ogg"},
	{0x721065C3CC8D9FDAULL, "llac_lt_gallowstemplar_2391_3.dds"},
	{0x7211109A30149966ULL, "165490698.ogg"},
	{0x72114B8952BE16A4ULL, "one100_party_comments6106776.cl"},
	{0x72115FB6691DDDE3ULL, "one110st_gustav.stg"},
	{0x7211A25F75F1FD59ULL, "179751151.ogg"},
	{0x7211EA843B3B83C0ULL, "art\\vfx\\combat\\vfx_v_fir_imp_aoe_p.cess.evt"},
	{0x7211F1008DF364D7ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_aoe_p.impa.gad"},
	{0x7212DFA8CF3A26F1ULL, "llac_ht_chantry_2534_11.dds"},
	{0x72131AF8DE8B4758ULL, "514961897.ogg"},
	{0x72132F943D684D9DULL, "vxdr_primevaltomb_7y4_0_1.vlm"},
	{0x7213A8ABC83AF44CULL, "gen_im_trs_prf_sta_ele_r0.uti"},
	{0x7213B6599A63163CULL, "zz_ave_debug6064128.cl"},
	{0x7213BD7990A30EFEULL, "vxht_mansion2_2iz_0_0.vlm"},
	{0x7213C98D03D7E0E6ULL, "zz_mag_debug6039842.cl"},
	{0x72141F9065C58FEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_head_up.ani"},
	{0x72144B1DFE5B7538ULL, "vxht_keep_55s_0_0.vlm"},
	{0x7214AA0BE33D058EULL, "vxlt_0_lowtown_docks_a3_cyx_0_1.vlm"},
	{0x7215335808E571C1ULL, "gen00fl_bethany6206061.cl"},
	{0x721537C29BC5BA77ULL, "197182252.ogg"},
	{0x72157113A475726DULL, "46035489.ogg"},
	{0x7215726CD7A8EA63ULL, "vxlt_0_lowtown_d46_0_0.vlm"},
	{0x7215F856C67AC56DULL, "6045338_m.fxe"},
	{0x721608C3CFC2987BULL, "vxlt_0_lowtown_n_es1_0_0.vlm"},
	{0x7216954C0D90005CULL, "934752631.ogg"},
	{0x7216B151021189B8ULL, "art\\characters\\weapons\\greatswords\\w_gsw_006a.phy"},
	{0x7216DBD4C0DBD482ULL, "zz_followers_banter6159048.cl"},
	{0x72170DB4E4CFCE92ULL, "vxht_bartrand_n_5y5_0_0.vlm"},
	{0x7217C6C8EF58CA20ULL, "413273831.ogg"},
	{0x7218279AFCBF0B07ULL, "hlac_lt_0_lowtown_a3_n_c_-131071_0x.dds"},
	{0x7218326702A2FAC6ULL, "vxht_keep_6jk_0_0.vlm"},
	{0x721848A592C71D7EULL, "vxht_mansion_n_707_0_0.vlm"},
	{0x7218E67C64549754ULL, "vxw_woundedcoast_15k_0_0.vlm"},
	{0x721906008DF9A764ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_aoe_p.impa.evt"},
	{0x7219140CA499FCD0ULL, "vxw_sundermount_35h_0_1.vlm"},
	{0x7219AEC35B494567ULL, "6208636_m.fxe"},
	{0x7219DBF518E1ACC6ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.dura.evt"},
	{0x7219DD2C8F15C148ULL, "and111cr_refugee_1.utc"},
	{0x7219F2E31437A593ULL, "cd_nat380_nathaniel.bnk"},
	{0x721A3D0CC58C90C6ULL, "mag370_meredith6037675.cl"},
	{0x721A441992BB4364ULL, "dre180cs_ragstoriches_mag.cut"},
	{0x721A928F2DA058E7ULL, "vxlt_darktown_4k8_0_0.vlm"},
	{0x721B8F6F960FFAD0ULL, "vxlt_gallowscourtyard_n_50r_0_0.vlm"},
	{0x721BABADF2B3F299ULL, "hlac_w_bonepitmining_a2_393222_0x.dds"},
	{0x721BAF206AD21D6DULL, "6066580_m.fxe"},
	{0x721BAFE481D5613AULL, "391364718.ogg"},
	{0x721C3C5DF1ED9E6AULL, "vxlt_0_lowtown_docks_a3_cog_0_0.vlm"},
	{0x721C4CF5585BF6B1ULL, "dae021_fem_worshipper_sta6141127.cl"},
	{0x721C5F31249DB95BULL, "vxca_cavern1_haunted_ijd_0_0.vlm"},
	{0x721CA693E91DC794ULL, "gen00fl_aveline6132731.cl"},
	{0x721CD6CA81042C65ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_leaves01_n.dds"},
	{0x721D73D43F8539C6ULL, "vxht_0_hightown_n_byu_0_0.vlm"},
	{0x721DD1B014CDA36BULL, "char2_vertexlit.vsh.fxo"},
	{0x721E0AC5609BE6A1ULL, "vxw_woundedcoast_1gd_0_0.vlm"},
	{0x721EF0C207D42BC5ULL, "6071947_f.fxe"},
	{0x721F7BA4B4DF3735ULL, "dae000_orsino6154297.cl"},
	{0x72209214404454FEULL, "plt_lgt000pt_baz_sketchy.plo"},
	{0x7220A49127ACB1A2ULL, "vxw_bonepitdestroyed_a3_1v7_0_0.vlm"},
	{0x7220A74A7B1C70CDULL, "lt_0_lowtown_docks_n_244_0.msh"},
	{0x7220B54A7B1C8800ULL, "lt_0_lowtown_docks_n_244_0.mao"},
	{0x7220B94A7B1C8F53ULL, "lt_0_lowtown_docks_n_244_0.mmh"},
	{0x7220F2F518E7F211ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.dura.gad"},
	{0x72211E60E876DDFBULL, "follower_banter6155535.cl"},
	{0x72217A43D48ED2F1ULL, "6099173_m.fxe"},
	{0x7221FDC941707944ULL, "651140238.ogg"},
	{0x7222B568758E9773ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_aslt.evt"},
	{0x72241F17F4754010ULL, "6178365_m.fxe"},
	{0x72248064202C4B39ULL, "514060409.ogg"},
	{0x72252F6BE3059A51ULL, "6127584_m.fxe"},
	{0x722535D7109098A5ULL, "342842969.ogg"},
	{0x72258E5CA0439C61ULL, "vxlt_0_lowtown_docks_c6t_0_0.vlm"},
	{0x7225CCC4498D128EULL, "plt_cod_ltr_sk3_scribbles3.nss"},
	{0x7225DCDF192AB801ULL, "463362863.ogg"},
	{0x722604AE548BDBAAULL, "gen00fl_anders6125918.cl"},
	{0x72262D96DDF3E9B5ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm03.phy"},
	{0x722687E4B8D4D0C3ULL, "dae000cr_grand_cleric6173669.cl"},
	{0x72271E008E060D5DULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_aoe_p.impa.ani"},
	{0x7227420F983F7B45ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_3z.dds"},
	{0x72275859564AAFD4ULL, "vxw_woundedcoast_10h_0_0.vlm"},
	{0x72278B8277A78FF0ULL, "c_body.wbd"},
	{0x7227F3F518EE12BFULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.dura.ani"},
	{0x72282E32824B74A3ULL, "vxlt_gallowstemplar_34v_0_0.vlm"},
	{0x7229AE687594A9B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_aslt.gad"},
	{0x7229DC647653AB54ULL, "vxw_woundedcoast_1hq_0_1.vlm"},
	{0x722A014C88061350ULL, "vxlt_gallowscourtyard_n_57s_0_0.vlm"},
	{0x722A0C307C772298ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupb.phy"},
	{0x722A14FF37ADCE5BULL, "lgt160_ntb_elf_conflict6163380.cl"},
	{0x722A63E8E278B8E6ULL, "6156455_m.fxe"},
	{0x722B8E3091E7DFB4ULL, "279459673.ogg"},
	{0x722BE5B1317B5246ULL, "vxht_0_hightown_n_bdv_0_0.vlm"},
	{0x722C89F8C6866944ULL, "55800822.ogg"},
	{0x722CC7DA100A4A5AULL, "zz_followers_banter6158403.cl"},
	{0x722D0F3C38B15390ULL, "801002779.ogg"},
	{0x722DB95C1B899E4FULL, "llac_lt_undercityserialkiller_-1_41.dds"},
	{0x722E1976B29B7322ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_cap.msh"},
	{0x722E1B76B29B76D8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_cap.mmh"},
	{0x722E2776B29B8B7BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_cap.mao"},
	{0x722E30C6EA99D815ULL, "vxlt_undercitydungeon_x0_0_0.vlm"},
	{0x722E9165DBE58F0DULL, "lla_ht_0_hightown_-131070_0.dds"},
	{0x722EFE741D3D26A3ULL, "llac_lt_0_lowtown_a2_n_0_5.dds"},
	{0x722F171261984848ULL, "vxlt_0_lowtown_docks_c6q_0_0.vlm"},
	{0x722FC19B711D0BFEULL, "244234719.ogg"},
	{0x722FD9351C502DC4ULL, "mag101_aombanter6052598.cl"},
	{0x722FDAC21ECBAF15ULL, "425571687.ogg"},
	{0x722FE5BE99A72474ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mere_hairm2.mao"},
	{0x72305C58561141FFULL, "lgt110_orz_harr_carta6109053.cl"},
	{0x72308E27717C0F03ULL, "6125518_f.fxe"},
	{0x723092DCAB095831ULL, "gen00fl_anders6125748.cl"},
	{0x72310A7420B1ACBEULL, "255986133.ogg"},
	{0x72312E7158700E75ULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunariamulet_sa.dds"},
	{0x72313275352E140BULL, "6076024_m.fxe"},
	{0x7231714A941991DCULL, "377689848.ogg"},
	{0x723188915088FC2EULL, "6066172_m.fxe"},
	{0x7231B7FCB71E9D37ULL, "dae100_alistair_drunk6135114.cl"},
	{0x7231F6662CF576FDULL, "hlac_lt_0_lowtown_a2_n_c_-65537_6z.dds"},
	{0x7231F7D815138238ULL, "vxht_mansion_n_72g_0_0.vlm"},
	{0x72326F106A9288AFULL, "one100ip_at_ruin_int.utp"},
	{0x7232EC7DB3AEC97BULL, "vxlt_0_lowtown_n_ei3_0_0.vlm"},
	{0x72337D147CFDE46EULL, "206588676.ogg"},
	{0x7233B73E29C9CA79ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.anb"},
	{0x723421BAE8C79BADULL, "vxlt_0_lowtown_docks_arv_0_0.vlm"},
	{0x72343A11B683E4C0ULL, "hlac_lt_0_lowtown_a2_65536_8x.dds"},
	{0x7234AA7EB420DCC8ULL, "llac_lt_0_lowtown_docks_65533_4.dds"},
	{0x7234D218C6855C46ULL, "ser111it_ninette_ring.uti"},
	{0x72351BA272416B20ULL, "mrl210_merrill6102111.cl"},
	{0x7235498FD7C22577ULL, "718583827.ogg"},
	{0x7235B008323E1373ULL, "6107154_f.fxe"},
	{0x7236068017FE4E37ULL, "ss_circlemages_m_2.wbd"},
	{0x723620418642066DULL, "hlac_ht_keep_4519_12z.dds"},
	{0x72362F980D224B7EULL, "gen00fl_aveline6114405.cl"},
	{0x7236872300D4E02AULL, "dae000_flemeth6013291.cl"},
	{0x7236C407B35CBE9AULL, "455967527.ogg"},
	{0x7236DE921CDED9D8ULL, "mcr310_argument6074489.cl"},
	{0x7236F480D1ED09CEULL, "6153694_m.fxe"},
	{0x72375A61444FEAA3ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_brkbrgmst.mao"},
	{0x72375E61444FF1F0ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_brkbrgmst.mmh"},
	{0x72376C614450090AULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_brkbrgmst.msh"},
	{0x72382B9DDBC20D47ULL, "142026592.ogg"},
	{0x72384E416849C3D4ULL, "vxw_woundedcoast_1sp_0_1.vlm"},
	{0x7238A2F6EB3F3D93ULL, "hlac_ht_0_hightown_n_131071_1y.dds"},
	{0x7238B6E348DBC5B2ULL, "qun211cr_crazy_woman6072188.cl"},
	{0x7238CB3B7292D62BULL, "vxlt_gallowstemplar_4o4_0_0.vlm"},
	{0x7238E05C892E31C8ULL, "w_encounter3_216_0.msh"},
	{0x7238F25C892E5059ULL, "w_encounter3_216_0.mao"},
	{0x7238FE5C892E64F2ULL, "w_encounter3_216_0.mmh"},
	{0x72390D9C5660187AULL, "war_resistmagic.dds"},
	{0x72392276B2A58A63ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_cap.phy"},
	{0x72397C946E196115ULL, "lt_0_lowtown_docks_a3_n_236_0.msh"},
	{0x72397E946E19648BULL, "lt_0_lowtown_docks_a3_n_236_0.mmh"},
	{0x72398A946E197928ULL, "lt_0_lowtown_docks_a3_n_236_0.mao"},
	{0x7239ABBD02F4054EULL, "6166197_m.fxe"},
	{0x723A26B7785581EBULL, "6164464_m.fxe"},
	{0x723A578E23716AC6ULL, "vxht_0_hightown_n_bow_0_0.vlm"},
	{0x723B02303613D08CULL, "15487650.ogg"},
	{0x723B44F1353FD49EULL, "hlac_lt_0_lowtown_131073_2z.dds"},
	{0x723BDBA0B9762A76ULL, "6106915_m.fxe"},
	{0x723C001E647E7F1CULL, "static_static_p7.vsh.fxo.dx11.assembly"},
	{0x723C0EB2CCA7FDC5ULL, "6191246_m.fxe"},
	{0x723C80687BDA58D3ULL, "435749665.ogg"},
	{0x723CC6DF864E8205ULL, "gen00fl_fenris6112065.cl"},
	{0x723D3B82A052B389ULL, "lt_0_lowtown_a3_n_c_27_0.rml"},
	{0x723DFB2873869256ULL, "art\\levels\\races\\design\\animation\\gasbarrel.activetoinactive.ani"},
	{0x723E2017F77B6481ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.impa.ani"},
	{0x723EBD7865EB0282ULL, "769350961.ogg"},
	{0x723ED425C3AE17F5ULL, "vxlt_undercityserialkiller_1c8_0_1.vlm"},
	{0x723ED841C31BD93EULL, "llac_lt_0_lowtown_a2_n_c_65537_4.dds"},
	{0x723F441187499162ULL, "vxdr_primevaltomb_7jw_0_0.vlm"},
	{0x723F45E20C89997CULL, "87429840.ogg"},
	{0x723F83B1464D1A01ULL, "gen00fl_isabela6090851.cl"},
	{0x7240139177FE7EB7ULL, "vxht_bartrand_n_66b_0_0.vlm"},
	{0x72405B35DFFD7CC5ULL, "vxlt_backalley_n_1xt_0_1.vlm"},
	{0x7240870C4CEEBA3FULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.mmh"},
	{0x7240877D75AB945BULL, "vxca_caverns2prc_9xq_0_0.vlm"},
	{0x7240F64420E0B676ULL, "6065007_m.fxe"},
	{0x7241017ED9456C96ULL, "6132736_m.fxe"},
	{0x72412B54806C0048ULL, "mag330_walter6045860.cl"},
	{0x72413E4B16B63206ULL, "6087098_m.fxe"},
	{0x72416F7C7A00F758ULL, "vxlt_0_lowtown_gbt_0_0.vlm"},
	{0x724187ED3538ABE2ULL, "961919063.ogg"},
	{0x724247A6A9C055C1ULL, "vxlt_0_lowtown_a2_ekl_0_0.vlm"},
	{0x7242E195CB4B6434ULL, "mag360_thrask6043546.cl"},
	{0x7243391AE80E9416ULL, "vxlt_0_lowtown_n_csz_0_0.vlm"},
	{0x7243C713CB78E02EULL, "dre000_sibling_has_blight6036539.cl"},
	{0x724405B115E44DABULL, "1032108925.ogg"},
	{0x7244225986A0A04BULL, "1030873751.ogg"},
	{0x72442B17F77FE323ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.impa.gad"},
	{0x72456CE1ED47B25CULL, "6079291_m.fxe"},
	{0x7245B1B749035058ULL, "vxlt_0_lowtown_a2_n_c_i1f_0_0.vlm"},
	{0x7245C04930EE284EULL, "525410975.ogg"},
	{0x7245F92F7052FCB2ULL, "615641656.ogg"},
	{0x724608389CB64FABULL, "6098275_m.fxe"},
	{0x7246A1D0837E077DULL, "6041693_m.fxe"},
	{0x72473407068EAD82ULL, "6115995_m.fxe"},
	{0x7247FBD177B54656ULL, "hlac_lt_0_lowtown_a3_n_c_65537_7x.dds"},
	{0x724861F4A77C2415ULL, "671952740.ogg"},
	{0x72487B2BC901E0F7ULL, "6164990_m.fxe"},
	{0x7248AAD2F3EBD680ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack_1.mao"},
	{0x7248C96B15678346ULL, "vxht_0_hightown_n_bjx_0_0.vlm"},
	{0x7248ED8C3A7FA5F0ULL, "6147024_m.fxe"},
	{0x72496E86DD7929D8ULL, "lgt210_mag_escp_rescued6156644.cl"},
	{0x724AD8829622A658ULL, "plt_genpt_amb_notif.nss"},
	{0x724B0817F785C608ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.impa.evt"},
	{0x724B1A2840B2FCF3ULL, "art\\vfx\\combat\\vfx_glb_ele_p_deathblow_c.impa.evt"},
	{0x724B239427768139ULL, "llac_lt_0_lowtown_docks_65534_2.dds"},
	{0x724C1999B048D00CULL, "chest.cur"},
	{0x724CDB31E909882CULL, "vxht_keep_3zx_0_0.vlm"},
	{0x724D0EBAF58D7BF2ULL, "vxlt_0_lowtown_a2_n_fhn_0_0.vlm"},
	{0x724D4A5ADEF8C28FULL, "dae000ip_crowd_ff3_talk7.utp"},
	{0x724D73B7F7C40035ULL, "6110362_m.fxe"},
	{0x724DA367B88C520CULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzchair.phy"},
	{0x724DAF77F6FB59EEULL, "mag360_samson6042048.cl"},
	{0x724E0F55807F5560ULL, "6018603_m.fxe"},
	{0x724E184400A6DD55ULL, "6064061_m.fxe"},
	{0x724E1A9004B3DA6CULL, "6175397_m.fxe"},
	{0x724EE59600843DCEULL, "atl_loginconne_dxt5_01.dds"},
	{0x724F4140759CE90FULL, "vxht_bartrand_n_6px_0_0.vlm"},
	{0x724F8F99A49F7B51ULL, "lgt120_orz_bln_official6137014.cl"},
	{0x724FF80D4B6E1086ULL, "6152525_m.fxe"},
	{0x725006AA9F2929C9ULL, "llac_w_bonepitdestroyed_a3_0_12.dds"},
	{0x72500C2D3ED3E920ULL, "6208370_m.fxe"},
	{0x72500D0ED4E629ECULL, "hlac_w_blightlands_131074_1x.dds"},
	{0x7250441DB9F95774ULL, "hlac_ht_mansion2_818_41y.dds"},
	{0x7250C4A2414EB854ULL, "vxca_caverns1_fm8_0_0.vlm"},
	{0x72518B82B079B372ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\set_warmwindow_d.dds"},
	{0x7251E3741169CF2BULL, "238778929.ogg"},
	{0x725200F5B892CDF8ULL, "242257226.ogg"},
	{0x7252132840B90F34ULL, "art\\vfx\\combat\\vfx_glb_ele_p_deathblow_c.impa.gad"},
	{0x7252242873978090ULL, "art\\levels\\races\\design\\animation\\gasbarrel.activetoinactive.gad"},
	{0x7252988FAE6F4739ULL, "836649476.ogg"},
	{0x725303683C2627DAULL, "montage_subs_demohack6217147.cl"},
	{0x7253734B8A831735ULL, "art\\vfx\\creatures\\vfx_c_drag_rake_c.mmh"},
	{0x7253969444440785ULL, "6073894_m.fxe"},
	{0x7253F9F2E8F5E285ULL, "6196575_m.fxe"},
	{0x725409A960F971CFULL, "plt_mer200pt_stolen_goods.nss"},
	{0x7254354532A41A61ULL, "hlac_lt_0_lowtown_a2_n_-196609_1x.dds"},
	{0x7254A55F948A082BULL, "6147218_m.fxe"},
	{0x7254BF3A986C4CCCULL, "6144450_m.fxe"},
	{0x7254EF90FBFCE876ULL, "6195829_m.fxe"},
	{0x72550C585B635B30ULL, "mag291_demon_marethari6028946.cl"},
	{0x7255479BCD589C52ULL, "one280_iwan6096898.cl"},
	{0x725642B53E295799ULL, "mrl000_marethari6069706.cl"},
	{0x72567D860B1D7DB5ULL, "vxca_caverns1_hu7_0_0.vlm"},
	{0x72570D6474FC8D94ULL, "522462180.ogg"},
	{0x7258222840BD9412ULL, "art\\vfx\\combat\\vfx_glb_ele_p_deathblow_c.impa.ani"},
	{0x72583AA31CB27AB4ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_5y.dds"},
	{0x7258529F38A2333EULL, "hlac_lt_gallowscourtyard_n_-65535_1y.dds"},
	{0x725891C9C14EAB97ULL, "6034084_m.fxe"},
	{0x7259B8CA17080234ULL, "6064529_m.fxe"},
	{0x725A470000431668ULL, "541826602.ogg"},
	{0x725A659DD2C4D13CULL, "108524080.ogg"},
	{0x725A6DB01CC75746ULL, "6064026_m.fxe"},
	{0x725A87F6B4584856ULL, "364719484.ogg"},
	{0x725ADD9ADB0DD6A6ULL, "llac_ht_brothel_n_3371_1.dds"},
	{0x725B0762B592A8AFULL, "vxw_bonepitmining_a2_21f_0_0.vlm"},
	{0x725B4F47D04F9BAEULL, "914626846.ogg"},
	{0x725B75715B0CF4B9ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcc.msh"},
	{0x725B7F715B0D05A7ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcc.mmh"},
	{0x725B83715B0D0CF4ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcc.mao"},
	{0x725C668AF49FCFEEULL, "t1_skn_h06.tnt"},
	{0x725C77E46DD5C504ULL, "vxlt_warehouse_106_0_0.vlm"},
	{0x725CCB64B65A2A66ULL, "6072310_f.fxe"},
	{0x725D3F690233E4D5ULL, "gen00fl_isabela6122923.cl"},
	{0x725D9B1F99A7E95CULL, "173172812.ogg"},
	{0x725DE9EDC502ACBAULL, "vxlt_0_lowtown_a2_n_c_hd2_0_0.vlm"},
	{0x725E2A1BB0C2C5D7ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.anb"},
	{0x725E9EEF8C8B17A1ULL, "17726741.ogg"},
	{0x725EB36BA76D9D26ULL, "sib131cr_slave_b.utc"},
	{0x725F4581AE17D9D8ULL, "plt_cod_itm_spiral_eye.plo"},
	{0x72600EB878E092A6ULL, "lt_0_lowtown_a2_n_probe_gp2.dds"},
	{0x726110E31473DBF1ULL, "cd_nat380_nathaniel.wbd"},
	{0x726114B444C8889CULL, "6143455_m.fxe"},
	{0x7261704B8A8F4FC2ULL, "art\\vfx\\creatures\\vfx_c_drag_rake_c.anb"},
	{0x7262485650CFF9DDULL, "883231606.ogg"},
	{0x72626B0A168C961FULL, "vxdr_primevalentrance_19y_0_0.vlm"},
	{0x7262AC67EF423C59ULL, "vxht_keep_n_61f_0_0.vlm"},
	{0x7262C7B2D431DD81ULL, "plt_cod_crt_spider.plo"},
	{0x7262E380C0493C52ULL, "vxlt_darktown_5c8_0_0.vlm"},
	{0x7262E4BF377E146BULL, "vxca_caverns1_mines_j2a_0_2.vlm"},
	{0x7263173C868020ABULL, "6140318_m.fxe"},
	{0x72633179CC4E1845ULL, "6041536_m.fxe"},
	{0x72636C45257CFAC6ULL, "mrl150_merrill6071845.cl"},
	{0x7263C71081D1E02FULL, "combat_ar_range.gda"},
	{0x726401C9B5145E0EULL, "1050236612.ogg"},
	{0x72640804D48B0F3FULL, "vxht_keep_n_6zu_0_0.vlm"},
	{0x726419184AB4238EULL, "lgt220_mag_tmplr_templar6140240.cl"},
	{0x726449CAAD83BBC7ULL, "llac_lt_gallowstemplar_f_2392_0.dds"},
	{0x7264A5F7A5F9366CULL, "vxht_keep_n_6ky_0_0.vlm"},
	{0x72652672149A387BULL, "ave121_brennan6105820.cl"},
	{0x72652E2D21E53CB8ULL, "fx_glyph_nuet.mao"},
	{0x726537FEF4BB88E3ULL, "vxlt_0_lowtown_eqe_0_1.vlm"},
	{0x726539570430BF1AULL, "880168380.ogg"},
	{0x72654ADA0CA6FE20ULL, "dae000_quintus6154615.cl"},
	{0x726552E366FE32A3ULL, "vxlt_0_lowtown_by3_0_0.vlm"},
	{0x726571F8DFF2230BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnright.evt"},
	{0x7266A6CDC0ED3A33ULL, "llac_ca_caverns1_24424_19.dds"},
	{0x726703A6B344C417ULL, "6031384_m.fxe"},
	{0x726706CEE09EB02AULL, "vxlt_0_lowtown_a2_e1u_0_0.vlm"},
	{0x7267D2186597A086ULL, "dae000cr_templar_wandering.utc"},
	{0x7267E4E82AC11207ULL, "qun160cr_leader.utc"},
	{0x72687DBEB364AADAULL, "574631224.ogg"},
	{0x726926F8A4E68210ULL, "ss_orzammar.dds"},
	{0x72694780F18D2AEBULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.dura.evt"},
	{0x7269880C4D11C17CULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.anb"},
	{0x726A82381F82C823ULL, "static_rendertodepth_p2.psh.fxo.dx11"},
	{0x726AC3CB2288428BULL, "vxlt_warehouse_n_p5_0_0.vlm"},
	{0x726B5FEF6C996993ULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.cess.ani"},
	{0x726BD934021AE824ULL, "vxht_0_hightown_a2_bad_0_0.vlm"},
	{0x726C68234777D404ULL, "hlac_lt_0_lowtown_n_65537_2z.dds"},
	{0x726C8AF8DFF86CCCULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnright.gad"},
	{0x726CBF87CF4FFF1CULL, "w_woundedcoast_246_0.mao"},
	{0x726CC8B3F9D0F9CCULL, "287519000.ogg"},
	{0x726CCB87CF5013BFULL, "w_woundedcoast_246_0.mmh"},
	{0x726CD187CF501E61ULL, "w_woundedcoast_246_0.msh"},
	{0x726CE49EC8A00931ULL, "gen_im_wep_rog_arc_lore_02.uti"},
	{0x726CF4876784D931ULL, "6027666_m.fxe"},
	{0x726D8C82DE46E026ULL, "vxdr_slavers_13z_0_0.vlm"},
	{0x726DB543DF91FDC6ULL, "isa311_party_comments.cnv"},
	{0x726DFA574725F7A6ULL, "vxht_0_hightown_n_alm_0_0.vlm"},
	{0x726E4096DE31BFD0ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm03.msh"},
	{0x726E4E96DE31D7EAULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm03.mmh"},
	{0x726E92166E975C0FULL, "gen00fl_dog6167091.cl"},
	{0x726EF2BECFD6B072ULL, "6034618_m.fxe"},
	{0x726EFCD4BFE812FFULL, "narrative_banter6136813.cl"},
	{0x726F01292EB11395ULL, "6059141_f.fxe"},
	{0x726F0E6742052580ULL, "6105866_m.fxe"},
	{0x726F6080F191C0ACULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.dura.gad"},
	{0x726F618277E4FDB2ULL, "c_body.bnk"},
	{0x72701C665F4D114FULL, "dae000_arishok6067682.cl"},
	{0x72703A82A07E0BA9ULL, "lt_0_lowtown_a3_n_c_27_0.anb"},
	{0x7270588E1F3A0601ULL, "80300136.ogg"},
	{0x7271462AB822F498ULL, "vxw_encounter1_ar_0_0.vlm"},
	{0x7271496C8969CEBCULL, "hlac_lt_warehouse_-65536_2y.dds"},
	{0x727153EBFDCF17F4ULL, "50413496.ogg"},
	{0x72716855886F8302ULL, "6095350_m.fxe"},
	{0x7271836C6BCAE94EULL, "art\\vfx\\materials\\vfx_fire_ring_fb_add.mao"},
	{0x7271A28B4A210934ULL, "6072081_m.fxe"},
	{0x7271B743DF95E986ULL, "isa311_party_comments.bnk"},
	{0x7271C1A49EEFC91EULL, "1028572160.ogg"},
	{0x7271F7CC1D6DF650ULL, "893859309.ogg"},
	{0x727214DBBB52DEC7ULL, "6034361_m.fxe"},
	{0x72722579B2BCA340ULL, "hlac_lt_0_lowtown_n_0_6x.dds"},
	{0x72722FD952B46232ULL, "dre000_varric6033322.cl"},
	{0x72729E51025FC84DULL, "art\\characters\\weapons\\greatswords\\w_gsw_006a.msh"},
	{0x7272AC51025FDF80ULL, "art\\characters\\weapons\\greatswords\\w_gsw_006a.mao"},
	{0x7272B051025FE6D3ULL, "art\\characters\\weapons\\greatswords\\w_gsw_006a.mmh"},
	{0x7272F5916ACEEAF5ULL, "490683870.ogg"},
	{0x72730F7CDCD7AB7DULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_013a_0t.dds"},
	{0x7273250F631CB3B6ULL, "mcr370_meredith6083105.cl"},
	{0x72736489B5C456D3ULL, "638139894.ogg"},
	{0x727389F8DFFE89CAULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnright.ani"},
	{0x727456EF1D654E50ULL, "mag160_feynriel6020461.cl"},
	{0x72752DAC50AA1122ULL, "vxdr_primevaltomb_8f0_0_0.vlm"},
	{0x727590FB1E591D03ULL, "vxw_woundedcoast_6p_0_0.vlm"},
	{0x7275A65330889B4DULL, "w_bonepit_58_0.mmh"},
	{0x7275A6D57A65D92BULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_p.impa.ani"},
	{0x7275AA533088A19EULL, "w_bonepit_58_0.mao"},
	{0x7275B8533088B9D3ULL, "w_bonepit_58_0.msh"},
	{0x7275EEBCCD42CFF4ULL, "6067832_m.fxe"},
	{0x72765F80F197DDAAULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.dura.ani"},
	{0x72770AB3F23CBAE7ULL, "gen00fl_anders6119936.cl"},
	{0x727717A45E5B2C2AULL, "vxw_encounter3_er_0_0.vlm"},
	{0x72772CB79B01E33BULL, "vxca_cave_23u_0_0.vlm"},
	{0x7277841C2F10CC63ULL, "6070393_f.fxe"},
	{0x7277B00E44627B4FULL, "779306173.ogg"},
	{0x72785362FF9BB7C6ULL, "zz_vault_debug6203297.cl"},
	{0x7278B75C348FD6F8ULL, "art\\vfx\\materials\\chunky\\vfx_c_lit_arc_blend.mao"},
	{0x7279340658EC1C18ULL, "hlac_dr_dungeon_3214_8z.dds"},
	{0x727935A8147E5B0BULL, "vxlt_warehouse_n_w4_0_0.vlm"},
	{0x727957EF6CA599CAULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.cess.evt"},
	{0x7279F4D2748D5CD1ULL, "vxlt_gallowstemplar_f_2os_0_0.vlm"},
	{0x727A819B873390D6ULL, "133958397.ogg"},
	{0x727AADF042288C20ULL, "vxca_cavern1_haunted_e6s_0_0.vlm"},
	{0x727B01E56A27AD67ULL, "vxlt_warehouse_n_89_0_0.vlm"},
	{0x727B60AAA16296E0ULL, "gen00fl_fenris6115368.cl"},
	{0x727B75EA9F869A2EULL, "dae000_denier6162694.cl"},
	{0x727C1BA6E6B3FDF5ULL, "art\\levels\\ztools\\default\\generic\\0fo_basic.phy"},
	{0x727D29C370CAA384ULL, "vxdr_0_deeproads_eax_0_0.vlm"},
	{0x727D6160A4B5EFC7ULL, "6118805_m.fxe"},
	{0x727D729F1EE0DAB9ULL, "hlac_ht_0_hightown_n_131071_0x.dds"},
	{0x727DC0DD6A97D3E4ULL, "27406409.ogg"},
	{0x727DC30F168BAC44ULL, "vxlt_backalley_n_1y7_0_0.vlm"},
	{0x727E3D73CE78A182ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_coffer.phy"},
	{0x727EA711EC3BC5B2ULL, "follower_banter6141861.cl"},
	{0x727EFBB4A0D60B8AULL, "859664453.ogg"},
	{0x727F3CC716514A6AULL, "vxht_playershouse_n_1ll_0_0.vlm"},
	{0x727F66859CF1E4C0ULL, "456875492.ogg"},
	{0x727FB807AA190D18ULL, "vxw_encounter1_fs_0_0.vlm"},
	{0x728036EF6CAB803DULL, "art\\vfx\\combat\\vfx_q_ari_impale_tar_lp_c.cess.gad"},
	{0x7280B926E09E2814ULL, "vxlt_0_lowtown_a2_d8u_0_0.vlm"},
	{0x7281098CDA42F536ULL, "vxlt_hangedman_n_2v6_0_0.vlm"},
	{0x728126786953C204ULL, "follower_banter6173101.cl"},
	{0x72813F4D48CAA663ULL, "6060135_m.fxe"},
	{0x72818DC511E65F8FULL, "qun211_party_comment6072243.cl"},
	{0x7281BB2428E9FFBDULL, "6090366_f.fxe"},
	{0x7283213D3AE556DCULL, "85759611.ogg"},
	{0x72835F18ACB06B3DULL, "294338576.ogg"},
	{0x72836477B8EFB08BULL, "ico_ring_2.dds"},
	{0x72838CE9A6A89236ULL, "vxlt_undercitydungeon_wz_0_0.vlm"},
	{0x72838ED57A71EDA2ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_p.impa.evt"},
	{0x7283911551DDE72EULL, "hlac_w_encounter2_camp_393215_0y.dds"},
	{0x72839561350180E8ULL, "t3_shk_h02.tnt"},
	{0x728402D8104F3583ULL, "vxw_woundedcoast_3s_0_0.vlm"},
	{0x72842BEBAC005FE4ULL, "6095081_f.fxe"},
	{0x728447EE0F991E63ULL, "6118399_m.fxe"},
	{0x72845953D7875782ULL, "vxht_mansion2b_2wp_0_0.vlm"},
	{0x7284E67F3BD99906ULL, "hlac_lt_0_lowtown_docks_a3_-131072_0z.dds"},
	{0x7285959643A8E5F0ULL, "hlac_w_encounter3_-327679_0z.dds"},
	{0x7285A414B8489F4BULL, "l_var221ar_mansion_real.wbd"},
	{0x7285BBDCA1617628ULL, "6162982_m.fxe"},
	{0x7285F9307CC5612DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupb.msh"},
	{0x728607307CC57960ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupb.mao"},
	{0x72860B307CC57FB3ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupb.mmh"},
	{0x7286271BB0E412D8ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.mmh"},
	{0x72868964E135CE2CULL, "vxdr_primevaltomb_6s3_0_0.vlm"},
	{0x7286E17A09A168B2ULL, "hlac_ht_chantry_2534_12z.dds"},
	{0x7286F9B7CBF7B763ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_0y.dds"},
	{0x7287443041A7A817ULL, "vxlt_hangedman_n_1mq_0_0.vlm"},
	{0x7287646E16FB7DAAULL, "vxht_keep_6af_0_0.vlm"},
	{0x7287675937E0C699ULL, "83727226.ogg"},
	{0x728768715B3219DCULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcc.phy"},
	{0x728779505E5DAB5AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_ornate.phy"},
	{0x7287A7850674738BULL, "vxlt_warehouse_n_z7_0_0.vlm"},
	{0x7288228E093A59E7ULL, "872772512.ogg"},
	{0x72884CF326AD93E4ULL, "mag100_thrask6031994.cl"},
	{0x728987D6093F0A7BULL, "324509033.ogg"},
	{0x7289A966B422DD71ULL, "733618177.ogg"},
	{0x7289AAF572C22F88ULL, "ser270_emeric6149388.cl"},
	{0x728A31F7A1FA4A4FULL, "art\\vfx\\creatures\\fxc_rage_burst_p.mmh"},
	{0x728A66BE0372611DULL, "6100085_f.fxe"},
	{0x728A8DD57A780AB5ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_f_p.impa.gad"},
	{0x728B2F260C09D5CBULL, "vxlt_hangedman_n_1vh_0_0.vlm"},
	{0x728B399A192B0203ULL, "6108021_f.fxe"},
	{0x728B3C2D2B70568AULL, "6162294_m.fxe"},
	{0x728B9BA062C0BC04ULL, "vxdr_0_deeproads_ejy_0_0.vlm"},
	{0x728BBA720DDA75B4ULL, "6108725_f.fxe"},
	{0x728BBD0B7C241E46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_escape.ani"},
	{0x728BD33D96AC0847ULL, "6117905_m.fxe"},
	{0x728BED155198AF7AULL, "dae000_meredith6085286.cl"},
	{0x728C534BC07435D5ULL, "241811321.ogg"},
	{0x728C93A0C2CB3A00ULL, "vxca_caverns1_gkk_0_0.vlm"},
	{0x728CC65F1C31CA9FULL, "adie_dot.mao"},
	{0x728CD76933DF5C23ULL, "109554399.ogg"},
	{0x728CE8441CC23F81ULL, "streak01_blend.mao"},
	{0x728CEC33B89660ACULL, "zz_slate2_baddy.cut"},
	{0x728CFF5FF996B4D7ULL, "zz_dog_debug6204280.cl"},
	{0x728D56513184BFB7ULL, "mag311cr_emile.utc"},
	{0x728D7C9426C2C9D0ULL, "lt_0_lowtown_a2_19_0_ll.mmh"},
	{0x728E29E49C0F2598ULL, "vxw_encounter1_kp_0_0.vlm"},
	{0x728E59EB2303C6B2ULL, "hlac_lt_0_lowtown_a2_1_3y.dds"},
	{0x728E6410410FFA74ULL, "vxlt_0_lowtown_a2_n_gi2_0_0.vlm"},
	{0x728E7A67B8C3A3D7ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzchair.mmh"},
	{0x728E7E67B8C3AAE4ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzchair.mao"},
	{0x728E9067B8C3C949ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzchair.msh"},
	{0x728EA23ED9A442F1ULL, "6149903_f.fxe"},
	{0x728ED98751153250ULL, "hlac_lt_darktown_-1_88x.dds"},
	{0x728EF668A71B5DE6ULL, "vxw_woundedcoast_qi_0_1.vlm"},
	{0x728FAD526E85C273ULL, "91720053.ogg"},
	{0x728FF0C2FA1A0086ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_we_stand_here.ani"},
	{0x72909BAD166E8222ULL, "hlac_w_encounter2_camp_-65533_0y.dds"},
	{0x72909C553C786439ULL, "hlac_ht_chantry_n_2534_25z.dds"},
	{0x7291CBF26D335890ULL, "6111987_f.fxe"},
	{0x7291D5D610EDE434ULL, "vxlt_0_lowtown_docks_a3_n_d7e_0_0.vlm"},
	{0x7292054C670D71E6ULL, "6037596_m.fxe"},
	{0x729244096A242A6CULL, "llac_lt_0_lowtown_a2_n_65537_4.dds"},
	{0x7292B9CB018F36E3ULL, "6117299_m.fxe"},
	{0x7292BA8D9375D858ULL, "lt_0_lowtown_a2_n_41_0.rml"},
	{0x7292DBA621D65436ULL, "6016580_m.fxe"},
	{0x72931E638334B9CEULL, "6101402_m.fxe"},
	{0x7293342AC1753AB9ULL, "6149214_m.fxe"},
	{0x7293804794839C6EULL, "ht_0_hightownexplosion_n_30_0_lc.mmh"},
	{0x7293C7D44F071858ULL, "900497229.ogg"},
	{0x7293E8479AC6D4D9ULL, "gen00fl_isabela6079670.cl"},
	{0x729410717EEE143BULL, "vxca_cave_29w_0_0.vlm"},
	{0x729445A1E57CE0DCULL, "vxw_bonepitdestroyed_a3_7u_0_0.vlm"},
	{0x7294FB41D32BE6ACULL, "vxdr_primevaltomb_6v0_0_0.vlm"},
	{0x729539C430D03A2BULL, "vxht_0_hightown_c96_0_0.vlm"},
	{0x729585EEE52CE80CULL, "hlac_lt_0_lowtown_a2_n_65536_1z.dds"},
	{0x7295B5190BE9DB03ULL, "707756957.ogg"},
	{0x7295B60D339DC097ULL, "vxlt_hangedman_n_1jp_0_0.vlm"},
	{0x7295C265D9F3834CULL, "611551531.ogg"},
	{0x72966471D9A5C573ULL, "vxlt_0_lowtown_docks_n_b4w_0_1.vlm"},
	{0x72967243B9E4CD26ULL, "vxdr_primevalentrance_2zn_0_0.vlm"},
	{0x7296B721BAD1866DULL, "vxlt_0_lowtown_a3_n_c_ghs_0_0.vlm"},
	{0x7296CF89232D37B1ULL, "dae000_keep_bureaucrat_1_st6165345.cl"},
	{0x7296DF6BDC310369ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev2.ani"},
	{0x7298442C27DAFB7BULL, "ht_mansion_n_3370.rml"},
	{0x72986F3BE5442E1EULL, "499951067.ogg"},
	{0x729881CF54FB9489ULL, "vxw_bonepitmining_a2_1x1_0_1.vlm"},
	{0x72989094C46DDE33ULL, "827478636.ogg"},
	{0x72990D6A491C0716ULL, "hlac_lt_0_lowtown_0_5z.dds"},
	{0x72991D4B1CE36F68ULL, "one110_meeran6089493.cl"},
	{0x72992A5744969CEEULL, "179142767.ogg"},
	{0x729951B6C27114D3ULL, "1022489931.ogg"},
	{0x72997FE9317B2BD5ULL, "vxlt_0_lowtown_docks_n_cti_0_0.vlm"},
	{0x7299A50B7C3032B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_escape.evt"},
	{0x7299DC4379E71295ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_trapdoorlid_0.pwk"},
	{0x729A2349823151D5ULL, "213543306.ogg"},
	{0x729A46A603B74492ULL, "280806297.ogg"},
	{0x729B131516EF03BFULL, "6018561_m.fxe"},
	{0x729B336642AE16C1ULL, "1624912.ogg"},
	{0x729B8710124B2FECULL, "hlac_ht_0_hightown_65536_3y.dds"},
	{0x729B96A77C0BBFBAULL, "6106241_m.fxe"},
	{0x729C8F909DDA1EDAULL, "vxlt_darktown_5o4_0_0.vlm"},
	{0x729CAE832D7C343FULL, "dae000_arishok6072402.cl"},
	{0x729CAF2B890D13CAULL, "hlac_ht_keep_4519_3y.dds"},
	{0x729CC0D68B7976BAULL, "6169475_m.fxe"},
	{0x729CD49F9B10CD85ULL, "6116526_m.fxe"},
	{0x729D03B15A102258ULL, "vxht_chantry_n_31v_0_0.vlm"},
	{0x729D2424590ECFAAULL, "vxht_0_hightown_n_6pb_0_0.vlm"},
	{0x729D89BF89F07959ULL, "423252881.ogg"},
	{0x729DCF789C63713FULL, "vxht_keep_n_6fq_0_0.vlm"},
	{0x729DD8C2FA2615F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_we_stand_here.evt"},
	{0x729DFA6BDC374FEBULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev2.gad"},
	{0x729E22B28112DD30ULL, "703115746.ogg"},
	{0x729E86BD3EE2DD0AULL, "mag100_cullen_karen_resolve.stg"},
	{0x729F26848FF34070ULL, "778276140.ogg"},
	{0x729F36509AB92E66ULL, "vxlt_hangedman_n_21s_0_0.vlm"},
	{0x729F5DFAEAE902CAULL, "21881192.ogg"},
	{0x729FC6BC7E208552ULL, "ser121_jethann6165793.cl"},
	{0x72A047B302E3FCB4ULL, "vxlt_0_lowtown_docks_a3_n_d2j_0_0.vlm"},
	{0x72A0A60B7C365340ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_escape.gad"},
	{0x72A0BDFBF7798BCDULL, "6080110_f.fxe"},
	{0x72A13DAC578B7EF4ULL, "vxw_sundermount_303_0_1.vlm"},
	{0x72A1CBF2C2F80249ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint.phy"},
	{0x72A2F3515B051FA9ULL, "329302921.ogg"},
	{0x72A336D1396CA832ULL, "hlac_ht_0_hightown_2_0y.dds"},
	{0x72A34AA6610F34B5ULL, "6004373_m.fxe"},
	{0x72A37DD9E2F777CDULL, "vxca_caverns2_78q_0_1.vlm"},
	{0x72A3CCFCE023F80CULL, "280130749.ogg"},
	{0x72A3D9C2FA2A8280ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_we_stand_here.gad"},
	{0x72A41873CE993236ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_coffer.mao"},
	{0x72A42473CE9946D5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_coffer.mmh"},
	{0x72A42673CE994A4BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_coffer.msh"},
	{0x72A4E24D5160510FULL, "plt_cod_hst_tevinter.plo"},
	{0x72A68028B1D41DECULL, "one300_zevran6076546.cl"},
	{0x72A79A14C7A26DBAULL, "llac_ca_caverns2_5802_63.dds"},
	{0x72A7C1CB0EFE59A6ULL, "vxht_0_hightown_n_apq_0_0.vlm"},
	{0x72A7F1C623714455ULL, "vxlt_0_lowtown_docks_n_coh_0_0.vlm"},
	{0x72A8378ED4926EB1ULL, "6163139_m.fxe"},
	{0x72A872932709B483ULL, "vxca_caverns1_e6m_0_0.vlm"},
	{0x72A8D83D446A4182ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_003a_0n.dds"},
	{0x72A9033B26521D77ULL, "vxca_cave_23a_0_0.vlm"},
	{0x72A94CB412146F80ULL, "6145141_m.fxe"},
	{0x72A9534A808A9E81ULL, "6129616_m.fxe"},
	{0x72A981F1C4FA3F06ULL, "145529130.ogg"},
	{0x72A9899537C9EDAAULL, "vxht_0_hightown_n_bdz_0_0.vlm"},
	{0x72A9CB9439B93ECFULL, "vxca_caverns1_mines_hl3_0_0.vlm"},
	{0x72AA148D0D5B5867ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_reaction.gad"},
	{0x72AA59816DC2B1C2ULL, "hlac_lt_0_lowtown_docks_n_131073_0z.dds"},
	{0x72AA8CAC94D0947BULL, "llac_lt_0_lowtown_a3_n_c_0_6.dds"},
	{0x72AACA7B16109E94ULL, "mcr371_scared_mage_f6086923.cl"},
	{0x72AB016D8FD0375AULL, "vxlt_darktown_5p5_0_0.vlm"},
	{0x72AB35E0F5B2BCB5ULL, "894424149.ogg"},
	{0x72AB47CA24FC2B74ULL, "vxlt_0_lowtown_a2_n_gc0_0_0.vlm"},
	{0x72AB690F4C321DB5ULL, "vxlt_0_lowtown_a3_n_c_eqn_0_0.vlm"},
	{0x72AB8614B8693B19ULL, "l_var221ar_mansion_real.bnk"},
	{0x72ABB4A6C2F7F18FULL, "1054193891.ogg"},
	{0x72ABD5061A9AFB0BULL, "6031152_f.fxe"},
	{0x72ABE806118541C8ULL, "6177900_m.fxe"},
	{0x72AC41558E5989BFULL, "vxht_keep_n_6ap_0_0.vlm"},
	{0x72AC4204E51A799FULL, "gen00fl_varric6093603.cl"},
	{0x72AC68AE895234ACULL, "312267168.ogg"},
	{0x72ACE03EC0E6A6CCULL, "ser270_cut_motherreveal6131558.cl"},
	{0x72ACF84407D9DA36ULL, "6077361_m.fxe"},
	{0x72AD0FFA2BC4F5E0ULL, "hlac_w_encounter2_65534_0y.dds"},
	{0x72AD1E4E522CCB6BULL, "vfx_lightning_f_08.dds"},
	{0x72AF316200B42EB3ULL, "ef_porfiria.mor"},
	{0x72AF69E4C1B3E8C5ULL, "qcr000_arishok6059674.cl"},
	{0x72AFE63F148F6840ULL, "6144380_m.fxe"},
	{0x72AFF3ED788D9E78ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medb_0t.dds"},
	{0x72B0098D0D5FB174ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_reaction.evt"},
	{0x72B0440924DA650CULL, "vxw_woundedcoast_cx_0_0.vlm"},
	{0x72B0615A5198D1B8ULL, "dae300_bran6143511.cl"},
	{0x72B114C83A08CB24ULL, "vxlt_0_lowtown_a2_hfk_0_0.vlm"},
	{0x72B1D35442872BECULL, "plt_prm_cod_lothering_tome.plo"},
	{0x72B2235113A5A49CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_loop.gad"},
	{0x72B228FB945B96C0ULL, "6103305_m.fxe"},
	{0x72B2AA3E1DEAE218ULL, "hlac_ca_cavern1_haunted_24423_1z.dds"},
	{0x72B2B02685F246D0ULL, "6172712_m.fxe"},
	{0x72B2BCB46EBBCB76ULL, "548393532.ogg"},
	{0x72B332F7A21D508CULL, "art\\vfx\\creatures\\fxc_rage_burst_p.anb"},
	{0x72B355FD9DD0FE26ULL, "vxdr_primevalentrance_2dp_0_0.vlm"},
	{0x72B3629209E87647ULL, "vxht_0_hightown_ajy_0_0.vlm"},
	{0x72B39ADB9EBDB76DULL, "vxlt_0_lowtown_a3_n_c_gbq_0_0.vlm"},
	{0x72B3C75AE3BEEACAULL, "6136668_m.fxe"},
	{0x72B43D6EBD105F7EULL, "6077674_f.fxe"},
	{0x72B46063F0490E97ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushlineglw.msh"},
	{0x72B47263F0492DFAULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushlineglw.mao"},
	{0x72B47663F0493409ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushlineglw.mmh"},
	{0x72B5046B216114A5ULL, "6016738_m.fxe"},
	{0x72B61DF12985087CULL, "zz_and_debug6134172.cl"},
	{0x72B633A800F47226ULL, "vxht_0_hightown_n_aur_0_0.vlm"},
	{0x72B636C48A267B59ULL, "329884272.ogg"},
	{0x72B6E0322B5127D3ULL, "hlac_lt_gallowscourtyard_n_-65538_3y.dds"},
	{0x72B819E38BCBB1C0ULL, "601688139.ogg"},
	{0x72B82588232A80E8ULL, "hlac_w_bonepitdestroyed_a3_131076_0x.dds"},
	{0x72B8350D0A8A255DULL, "282714085.ogg"},
	{0x72B88627B335628AULL, "424895846.ogg"},
	{0x72B8CBE8798FAA79ULL, "vxdr_0_deeproads_fi2_0_0.vlm"},
	{0x72B8F135497232BAULL, "hlac_lt_gallowscourtyard_n_131071_1y.dds"},
	{0x72B9043ADE29AC6AULL, "vxht_playershouse_n_1ph_0_0.vlm"},
	{0x72B9225113ABC19AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_loop.ani"},
	{0x72B93B27D0F3AD59ULL, "6069414_f.fxe"},
	{0x72B9487789FDD10DULL, "106909419.ogg"},
	{0x72B96B221DCED0FDULL, "vxca_caverns2_8s2_0_0.vlm"},
	{0x72B997B00F288AADULL, "6045243_f.fxe"},
	{0x72B9A4906F65A7BAULL, "6163675_m.fxe"},
	{0x72BA3AB930272D68ULL, "vxlt_foundry_n_vz_0_0.vlm"},
	{0x72BA65F37E4EFE36ULL, "mcr370_assignments6176737.cl"},
	{0x72BABD2CD45C7F9CULL, "628837693.ogg"},
	{0x72BB62EF143B1631ULL, "hlac_w_woundedcoast_0_4z.dds"},
	{0x72BBB8A3422AF213ULL, "vxht_bartrand_n_54k_0_0.vlm"},
	{0x72BBDEB5BD5E2F1EULL, "486359726.ogg"},
	{0x72BD92C2DC40E0E8ULL, "vxlt_gallowstemplar_f_5rr_0_0.vlm"},
	{0x72BE018D0D6BE0ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_reaction.ani"},
	{0x72BE0269050338DFULL, "vxca_cavern1_haunted_e8z_0_0.vlm"},
	{0x72BEB4FF361E9ACCULL, "hlac_lt_0_lowtown_65536_11z.dds"},
	{0x72BEB5E616D07D8CULL, "vxw_woundedcoast_xy_0_0.vlm"},
	{0x72BED32FB9900A71ULL, "jobboards.gda"},
	{0x72BEE17E8126044FULL, "6001344_f.fxe"},
	{0x72BF2C48BF402D5DULL, "vxlt_undercityserialkiller_1j6_0_0.vlm"},
	{0x72BF516A71E21943ULL, "art\\vfx\\textures\\vfx_wt_base.dds"},
	{0x72BF86A52BFEE3A4ULL, "vxlt_0_lowtown_a2_hch_0_0.vlm"},
	{0x72BFD716DCFE2738ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_meda_0d.dds"},
	{0x72C012DA7702B10CULL, "dae000_arishok6072803.cl"},
	{0x72C01DDD765BF396ULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.dura.ani"},
	{0x72C048238BC5BE79ULL, "w_sundermount_40_0_lc.mmh"},
	{0x72C09AD88651AF40ULL, "6104205_m.fxe"},
	{0x72C09F201A5917BDULL, "lgt240ip_at_tomecave.utp"},
	{0x72C0ED6D39B3BC2CULL, "m_mag321ar_mansion.bnk"},
	{0x72C17466F28FA8A1ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage2f_.ani"},
	{0x72C1C7DA8FC716A6ULL, "vxdr_primevalentrance_2kq_0_0.vlm"},
	{0x72C1D46EFBDE8EC7ULL, "vxht_0_hightown_amx_0_0.vlm"},
	{0x72C2F755A9F6A031ULL, "lt_0_lowtown_a2_n.arl"},
	{0x72C32EA6E6F02010ULL, "art\\levels\\ztools\\default\\generic\\0fo_basic.msh"},
	{0x72C33CA6E6F0382AULL, "art\\levels\\ztools\\default\\generic\\0fo_basic.mmh"},
	{0x72C340A6E6F03F61ULL, "art\\levels\\ztools\\default\\generic\\0fo_basic.mao"},
	{0x72C456D49286F214ULL, "vxlt_gallowscourtyard_758_0_0.vlm"},
	{0x72C45D36452A2D54ULL, "mag121_eau_banter6121086.cl"},
	{0x72C492D07399961FULL, "464125670.ogg"},
	{0x72C4A9C9249AE5E1ULL, "hlac_lt_gallowstemplar_f_2383_3y.dds"},
	{0x72C5137761C98751ULL, "602569519.ogg"},
	{0x72C5C6BA66FFCF93ULL, "108809698.ogg"},
	{0x72C5ECFC77CFEF5DULL, "one200it_blush.uti"},
	{0x72C61E5D56D17366ULL, "one300_nuncio6073785.cl"},
	{0x72C626C6C5C8BAB7ULL, "6077588_m.fxe"},
	{0x72C6760F61A46F2BULL, "44253375.ogg"},
	{0x72C68177ECCBB9DAULL, "6015182_f.fxe"},
	{0x72C6E7B8A3765A50ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlm_meda_0d.dds"},
	{0x72C716E7DB130403ULL, "vxlt_gamlenshouse_n_n1_0_0.vlm"},
	{0x72C73D6087D61F00ULL, "496998779.ogg"},
	{0x72C7495174BC472FULL, "vxht_mansion2b_2o9_0_0.vlm"},
	{0x72C749ECE80F4093ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_taunt.ani"},
	{0x72C75B58192D3397ULL, "6110693_f.fxe"},
	{0x72C7FFECA6E3D449ULL, "269955157.ogg"},
	{0x72C83FD0B6467B2FULL, "367993910.ogg"},
	{0x72C87F66F295DA43ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage2f_.gad"},
	{0x72C8FA0BC62ACE4BULL, "vxw_woundedcoast_1b8_0_1.vlm"},
	{0x72C91CD5BF4D6759ULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.mmh"},
	{0x72C91E6DEFA0BF66ULL, "vxht_bartrand_n_4kn_0_0.vlm"},
	{0x72C94D38E588A300ULL, "fxa_unholy_beams.dds"},
	{0x72C9549E3CF4096EULL, "art\\cinematics\\additional animation\\cs_malenoblev_cinematicanim.ani"},
	{0x72C9DD85479C0B12ULL, "6040701_m.fxe"},
	{0x72C9F463AEDA914DULL, "llac_lt_0_lowtown_a3_n_c_131071_5.dds"},
	{0x72CA2D3BFC3F630AULL, "6071712_m.fxe"},
	{0x72CACD979F78C014ULL, "vxht_0_hightown_a3_bxb_0_0.vlm"},
	{0x72CB1D2FA76F3257ULL, "259732630.ogg"},
	{0x72CB220153C2D7D7ULL, "1700515.ogg"},
	{0x72CB34F9933D7853ULL, "vxw_woundedcoast_14g_0_0.vlm"},
	{0x72CC016091234D4CULL, "227768574.ogg"},
	{0x72CC049FCE36F968ULL, "vxlt_gallowstemplar_f_5us_0_0.vlm"},
	{0x72CCA5E4433361CFULL, "628985999.ogg"},
	{0x72CD05B5B9D72F18ULL, "mrl000_marethari6072455.cl"},
	{0x72CD1B4AA92D3F00ULL, "vxlt_0_lowtown_docks_a3_d0o_0_0.vlm"},
	{0x72CD1DB513490A6CULL, "dre000_varric6208049.cl"},
	{0x72CD2D68883824B3ULL, "lt_0_lowtown_a3_n_c_32_0.anb"},
	{0x72CD48D7A6570084ULL, "vxlt_0_lowtown_epa_0_1.vlm"},
	{0x72CD9E25B13645DDULL, "vxlt_undercityserialkiller_1o5_0_0.vlm"},
	{0x72CDC4BC41432F9FULL, "306581284.ogg"},
	{0x72CDF8821DF4FC24ULL, "vxlt_0_lowtown_a2_hhi_0_0.vlm"},
	{0x72CDFFC1F6D2B684ULL, "lt_0_lowtown_docks_n_238_0.msh"},
	{0x72CE11C1F6D2D5EDULL, "lt_0_lowtown_docks_n_238_0.mao"},
	{0x72CE15C1F6D2DC26ULL, "lt_0_lowtown_docks_n_238_0.mmh"},
	{0x72CE25DD76683E87ULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.dura.evt"},
	{0x72CE38A3562182BDULL, "vxlt_darktown_67x_0_0.vlm"},
	{0x72CEB4F2C31EC96EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint.mmh"},
	{0x72CEBFE1FC82D7D8ULL, "6101398_m.fxe"},
	{0x72CEC0F2C31EDD05ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint.mao"},
	{0x72CECEF2C31EF55CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint.msh"},
	{0x72CF37B2F83EC789ULL, "6163368_m.fxe"},
	{0x72CF5704336230CDULL, "hlac_ht_playershouse_n_855_32y.dds"},
	{0x72CF8C0E5DB17254ULL, "vxlt_foundry_n_w5_0_0.vlm"},
	{0x72CFCA1AE235E54DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8mb.mmh"},
	{0x72CFCE1AE235EB9EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8mb.mao"},
	{0x72CFDC1AE23603D3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8mb.msh"},
	{0x72D05F97688E5C40ULL, "vxlt_gallowscourtyard_n_70p_0_0.vlm"},
	{0x72D077739846EC37ULL, "zz_mag_debug6054592.cl"},
	{0x72D130CBE0EBC2F3ULL, "6078659_m.fxe"},
	{0x72D1395B91412077ULL, "109321407.ogg"},
	{0x72D13C2D9984BA32ULL, "zz_qun_debug6028230.cl"},
	{0x72D145599532D8EDULL, "w_encounter3_53_0.mmh"},
	{0x72D149599532DF3EULL, "w_encounter3_53_0.mao"},
	{0x72D157599532F773ULL, "w_encounter3_53_0.msh"},
	{0x72D22EE6F329C3B6ULL, "zz_vrc_debug6200198.cl"},
	{0x72D23C23C4416118ULL, "vxlt_hangedman_1vh_0_0.vlm"},
	{0x72D2931B83917DA2ULL, "297051125.ogg"},
	{0x72D2EEF3037A4622ULL, "206950703.ogg"},
	{0x72D2F8A7CF9CC6ACULL, "abi_m_hex_damage.bnk"},
	{0x72D488436FC1FB76ULL, "fx_blizz_flake.dds"},
	{0x72D4B5AF549216D2ULL, "vxw_woundedcoast_14b_0_0.vlm"},
	{0x72D51DAEB07EA01AULL, "90338719.ogg"},
	{0x72D541ECE81B70CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_taunt.evt"},
	{0x72D546DD766E95D0ULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.dura.gad"},
	{0x72D5D9A0E555E23CULL, "hlac_lt_0_lowtown_a3_n_c_1_2x.dds"},
	{0x72D5E4F37F6EEF00ULL, "541357687.ogg"},
	{0x72D7287236329069ULL, "405470414.ogg"},
	{0x72D78AADA4731692ULL, "vxw_sundermount_30e_0_1.vlm"},
	{0x72D79F37621CDBA7ULL, "vxlt_0_lowtown_a2_n_c_e19_0_0.vlm"},
	{0x72D7C78E3B104F9EULL, "6103097_m.fxe"},
	{0x72D7D1B16EA7AFA5ULL, "mag291_demon_marethari6029025.cl"},
	{0x72D84663C5BE61D2ULL, "dae200cr_viscount6087200.cl"},
	{0x72D904F588CBD9B3ULL, "581021420.ogg"},
	{0x72D93F74916ED894ULL, "vxht_0_hightown_a3_bwc_0_0.vlm"},
	{0x72D9472C28129747ULL, "ht_mansion_n_3370.anb"},
	{0x72D952728D437B7AULL, "6161109_m.fxe"},
	{0x72D9D8AF476CFCFEULL, "vxlt_0_lowtown_docks_apw_0_0.vlm"},
	{0x72DABEACC4129495ULL, "vxlt_gallowscourtyard_n_73x_0_0.vlm"},
	{0x72DADA93B0382D19ULL, "hm_gen_act3.mor"},
	{0x72DB36AC492B8388ULL, "vxw_blightlands_a8_0_0.vlm"},
	{0x72DB4559953B2D66ULL, "w_encounter3_53_0.rml"},
	{0x72DB83F5EAF4A0F9ULL, "hlac_ht_0_hightown_n_-65536_0y.dds"},
	{0x72DC1002A32C5E5DULL, "vxlt_undercityserialkiller_1d4_0_0.vlm"},
	{0x72DC20ECE821573DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_taunt.gad"},
	{0x72DC493E90E566A3ULL, "140646903.ogg"},
	{0x72DC664C7C774428ULL, "isa311_isabela6095547.cl"},
	{0x72DCE5F67702CA4EULL, "166921203.ogg"},
	{0x72DD0BB2D053D48DULL, "vxht_0_hightown_a3_c4z_0_0.vlm"},
	{0x72DD2D9E3D046FA8ULL, "art\\cinematics\\additional animation\\cs_malenoblev_cinematicanim.gad"},
	{0x72DD6FE042D87470ULL, "vxlt_gallowscourtyard_n_775_0_0.vlm"},
	{0x72DDF1A0C745DA2AULL, "819078274.ogg"},
	{0x72DE27A0D16E1A70ULL, "483270569.ogg"},
	{0x72DE3D0742459C28ULL, "cli600cr_ogre.utc"},
	{0x72DE8A5A7F401174ULL, "follower_banter6165605.cl"},
	{0x72DE936E9C1733E7ULL, "123015045.ogg"},
	{0x72DE9EA7F76E6983ULL, "6126934_m.fxe"},
	{0x72DEBE0A80D83EC4ULL, "zz_qun_debug6083716.cl"},
	{0x72DF7A2BFBAB0397ULL, "dae000_flemeth6000861.cl"},
	{0x72E112292E0DABBAULL, "vxht_0_hightown_a3_9zq_0_0.vlm"},
	{0x72E1C5A784FA1A5AULL, "zz_one_debug6051511.cl"},
	{0x72E22D0155F22A13ULL, "vxlt_0_lowtown_docks_a3_n_byn_0_0.vlm"},
	{0x72E2317B6B540E48ULL, "isa311_castillon6094757.cl"},
	{0x72E23E5995413FAAULL, "w_encounter3_53_0.phy"},
	{0x72E264AABB5D1C1FULL, "6170696_m.fxe"},
	{0x72E2889F94B4B28AULL, "pro000cr_refugee1.utc"},
	{0x72E317CA3C62F37AULL, "6108962_m.fxe"},
	{0x72E34F51AD50D6F6ULL, "vxw_sundermount_36v_0_0.vlm"},
	{0x72E3596AE1025871ULL, "vxht_0_hightown_a3_9as_0_0.vlm"},
	{0x72E372505EABFD83ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_ornate.msh"},
	{0x72E380505EAC15DDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_ornate.mmh"},
	{0x72E384505EAC1CEEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wi_ornate.mao"},
	{0x72E449FDC4F0C2D0ULL, "vxlt_0_lowtown_a2_gnl_0_0.vlm"},
	{0x72E4BAE9777611DCULL, "hlac_ca_cavern1_haunted_24426_0z.dds"},
	{0x72E4BBB7A30803A7ULL, "vxlt_0_lowtown_h8b_0_0.vlm"},
	{0x72E524462BED05AFULL, "mag100_thrask6206098.cl"},
	{0x72E62241564BCD34ULL, "143486426.ogg"},
	{0x72E6396B2D06681EULL, "6104197_m.fxe"},
	{0x72E65CA588D62185ULL, "vxlt_0_lowtown_docks_n_bse_0_0.vlm"},
	{0x72E6661BFE470FBBULL, "6103693_f.fxe"},
	{0x72E6BD00AFBAB53EULL, "121910309.ogg"},
	{0x72E80A6EC4FB7F6CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_isaa_0d.dds"},
	{0x72E849D52ED80CFFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_noba_0d.dds"},
	{0x72E88593CC6D7EDEULL, "307121435.ogg"},
	{0x72E88C560189126EULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_deadmeredith_d.dds"},
	{0x72E8E859B2232A68ULL, "vxlt_gallowstemplar_f_5cq_0_0.vlm"},
	{0x72E957FFDAE5825FULL, "vxca_cavern1_haunted_e7g_0_0.vlm"},
	{0x72E95B42B06BC8AAULL, "532043917.ogg"},
	{0x72E9A8893B219C08ULL, "vxw_blightlands_j9_0_0.vlm"},
	{0x72EA435D3BBB6F42ULL, "851838999.ogg"},
	{0x72EA5276D9F5BEFBULL, "hlac_ht_0_hightown_a2_-196609_0z.dds"},
	{0x72EAE7CE0DF0C0E0ULL, "instancing1.vsh.fxo"},
	{0x72EB4D5AB4F1E0B3ULL, "mag370_meredith6040920.cl"},
	{0x72EB8F436FD4E5A4ULL, "fx_blizz_flake.mao"},
	{0x72EB95FFDE3322B5ULL, "vxlt_0_lowtown_docks_b03_0_0.vlm"},
	{0x72ECE06E247F424BULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v1.evt"},
	{0x72ECF7E1A8526878ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_black_s.dds"},
	{0x72ED45BAB8CA61ECULL, "chardebug.vsh.fxo"},
	{0x72EDC208AE24C429ULL, "6116089_m.fxe"},
	{0x72EE0F6BAF6A899FULL, "6118351_m.fxe"},
	{0x72EEC7802F18ADC0ULL, "vxca_caverns1_mines_jgo_0_0.vlm"},
	{0x72EEF855AA1BDC97ULL, "lt_0_lowtown_a2_n.rml"},
	{0x72EF0BA5F6831443ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pick_up_from_ground_loop.ani"},
	{0x72EF139C9D2DB071ULL, "6042317_m.fxe"},
	{0x72EF1FDDA82D9218ULL, "vxlt_hangedman_1xf_0_0.vlm"},
	{0x72EF3A0E8E8BACAEULL, "vxht_mansion2b_3gu_0_0.vlm"},
	{0x72EFC27F482A7242ULL, "mrl210_merrill6166872.cl"},
	{0x72F0061DA4D978F6ULL, "hlac_lt_0_lowtown_docks_n_-65538_1x.dds"},
	{0x72F03309E231491EULL, "dre000_dougal6131128.cl"},
	{0x72F0500B327B54C1ULL, "llac_ca_caverns2prc_5802_12.dds"},
	{0x72F057BD8036071BULL, "694507981.ogg"},
	{0x72F10EB0BD9F5EDBULL, "98047003.ogg"},
	{0x72F1154C6B49E418ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_5y.dds"},
	{0x72F1534271EDDD76ULL, "art\\vfx\\models\\vfx_lightning_arc.msh"},
	{0x72F15D4271EDEE74ULL, "art\\vfx\\models\\vfx_lightning_arc.mmh"},
	{0x72F219D5BF70674EULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.anb"},
	{0x72F2949C315A71A9ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_arm_cas_0n.dds"},
	{0x72F2A44A49931A7AULL, "plt_seb000pt_exiled_prince.nss"},
	{0x72F354C4AB90FDDCULL, "vxca_caverns1_gk7_0_0.vlm"},
	{0x72F3B6096D0E1DC1ULL, "vxlt_0_lowtown_byi_0_0.vlm"},
	{0x72F3F96E24858B0CULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v1.gad"},
	{0x72F46F1306C5927AULL, "vis121_saemus6086706.cl"},
	{0x72F4D7F8F03D283BULL, "6104793_f.fxe"},
	{0x72F4DAB77C36202DULL, "6149510_f.fxe"},
	{0x72F5189919962B6CULL, "vxht_keep_n_69w_0_0.vlm"},
	{0x72F51D7F45BB6195ULL, "qun221_petrice6185133.cl"},
	{0x72F562A8C1EF9EB9ULL, "dae000_merchant_generic_hm6172578.cl"},
	{0x72F57EE09761FE18ULL, "vxlt_gallowsprison_n_5rb_0_0.vlm"},
	{0x72F595A6476B88BBULL, "6015089_m.fxe"},
	{0x72F60617E5DFE0B5ULL, "tal_thre_com.mao"},
	{0x72F6152BDFCC8E4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_run.ani"},
	{0x72F6232E755B0994ULL, "vxht_0_hightown_a3_bia_0_0.vlm"},
	{0x72F6C79D31836D3FULL, "306357326.ogg"},
	{0x72F80B622C0C38F5ULL, "41596994.ogg"},
	{0x72F81A662D17B488ULL, "vxw_blightlands_o6_0_0.vlm"},
	{0x72F83992ED0EA2C4ULL, "vxca_cavern1_haunted_gkr_0_0.vlm"},
	{0x72F86E16646C212CULL, "vxca_caverns1_mines_j22_0_2.vlm"},
	{0x72F8AC061AB64155ULL, "vxlt_0_lowtown_a2_n_gjz_0_0.vlm"},
	{0x72F8EAEAEB6B5D70ULL, "vxca_caverns1_mines_gex_0_0.vlm"},
	{0x72F93C7D9E9FF489ULL, "art\\vfx\\creatures\\fxc_ohurl_imp_p.impa.ani"},
	{0x72F94A68885D916FULL, "lt_0_lowtown_a3_n_c_32_0.rml"},
	{0x72F990B4396EBD8EULL, "zz_followers_banter6158322.cl"},
	{0x72F997F22CB4CEEFULL, "6140189_m.fxe"},
	{0x72F9D10AD9CB1690ULL, "6085384_m.fxe"},
	{0x72F9E57B1539BD72ULL, "67397347.ogg"},
	{0x72F9F86E2489F50AULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v1.ani"},
	{0x72F9FD52D25E36D8ULL, "mag110_samson6018716.cl"},
	{0x72FA31407443D779ULL, "zz_drept_debug.ncs"},
	{0x72FA5A09C4E8591AULL, "hightown_market_guard_sta6169628.cl"},
	{0x72FABA2BBDA5A2DCULL, "201584563.ogg"},
	{0x72FB9CFA1BAAED57ULL, "gen_im_trs_gen_rng_r1.uti"},
	{0x72FC1EAF0ECA4FB3ULL, "zz_slate5_baddy.cut"},
	{0x72FC47361AB13503ULL, "vxw_bonepitmining_a2_1xb_0_0.vlm"},
	{0x72FC69A9CD2B2DEEULL, "vrc321_varric6173848.cl"},
	{0x72FC8148A160A21FULL, "6117251_m.fxe"},
	{0x72FC8F7B9834011CULL, "vxca_cavern1_haunted_hvn_0_0.vlm"},
	{0x72FD73D455028BCCULL, "115054420.ogg"},
	{0x72FE0C82E5968F58ULL, "272868762.ogg"},
	{0x72FE43A1888C65A2ULL, "vxht_brothel_n_50b_0_0.vlm"},
	{0x72FE7E3AA0DC3F79ULL, "art\\vfx\\combat\\vfx_c_single_tag_stun_c.mmh"},
	{0x72FE91A1C7997166ULL, "mer221_sabin6047680.cl"},
	{0x72FE963403223291ULL, "vxht_0_hightown_a3_cbl_0_0.vlm"},
	{0x72FED2BC4941E273ULL, "llac_w_bonepitdestroyed_a3_0_10.dds"},
	{0x72FF6E5933AFE56EULL, "zz_vault_debug6203194.cl"},
	{0x72FFA9DA79511DBFULL, "305932155.ogg"},
	{0x72FFEBCCA0998F35ULL, "mrl210_merrill6102145.cl"},
	{0x730029147C85DD4FULL, "877322105.ogg"},
	{0x7300515A071CF559ULL, "llac_w_bonepitmining_a2_3_0.dds"},
	{0x7300577D9EA6410BULL, "art\\vfx\\creatures\\fxc_ohurl_imp_p.impa.gad"},
	{0x730089A071B91B46ULL, "6152231_m.fxe"},
	{0x73013F80C0B4CD83ULL, "6174627_m.fxe"},
	{0x7301E184BA9F99E9ULL, "w_sundermount_243_0.mao"},
	{0x7301ED84BA9FAD82ULL, "w_sundermount_243_0.mmh"},
	{0x7301EF84BA9FB138ULL, "w_sundermount_243_0.msh"},
	{0x7301F76D39EB645EULL, "m_mag321ar_mansion.wbd"},
	{0x73021B3CAB903D5CULL, "602324591.ogg"},
	{0x7302F0306CC4C96FULL, "pro000_athenril6049452.cl"},
	{0x7302FFC427FC5FE4ULL, "1002163601.ogg"},
	{0x730322A5F693E42DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pick_up_from_ground_loop.gad"},
	{0x73032F05D2B5A8CAULL, "6169813_m.fxe"},
	{0x73035360397A0B08ULL, "6129147_m.fxe"},
	{0x7303D06FC51E7EB8ULL, "6172607_f.fxe"},
	{0x73040D2BDFD8BD9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_run.evt"},
	{0x7305153C5EA8423DULL, "llac_lt_0_lowtown_a2_131070_4.dds"},
	{0x73055ABEBADE4B46ULL, "plt_one100pt_ggt_notify.nss"},
	{0x73058B817DF6A43AULL, "w_encounter3_47_0.mmh"},
	{0x73058F817DF6AB71ULL, "w_encounter3_47_0.mao"},
	{0x730592BAE8C76E78ULL, "vxlt_gallowstemplar_f_3e1_0_0.vlm"},
	{0x73059D817DF6C280ULL, "w_encounter3_47_0.msh"},
	{0x7305C3947549B89DULL, "254091305.ogg"},
	{0x7305ED071BC0DA00ULL, "14836912.ogg"},
	{0x7305FEF4FEAE5E7CULL, "hlac_lt_gallowstemplar_2383_8z.dds"},
	{0x73068C431F0DCD08ULL, "vxw_blightlands_p7_0_0.vlm"},
	{0x730693FD5CD7B387ULL, "6096743_m.fxe"},
	{0x7306956FD9518B80ULL, "vxca_cavern1_haunted_icu_0_1.vlm"},
	{0x7306B5C41A7155CBULL, "vxht_mansion_n_6s6_0_0.vlm"},
	{0x7306E4D9FA52793BULL, "6074426_m.fxe"},
	{0x7307547D9EAC5B70ULL, "art\\vfx\\creatures\\fxc_ohurl_imp_p.impa.evt"},
	{0x7307586A5CEC122EULL, "zz_followers_banter6158076.cl"},
	{0x73075DDF32DF1CBCULL, "6044964_f.fxe"},
	{0x730780658649338DULL, "qcr220_grey_wardens6135933.cl"},
	{0x7307EEF62E942EA5ULL, "ave200_donnic6110757.cl"},
	{0x73082AFF892F55D2ULL, "337100422.ogg"},
	{0x73098FA8DC889095ULL, "vxlt_0_lowtown_docks_a3_n_d12_0_0.vlm"},
	{0x730A96520D565C02ULL, "637644088.ogg"},
	{0x730A9891EEBB0A5BULL, "6131259_m.fxe"},
	{0x730AB9130CA74D83ULL, "vxw_bonepitmining_a2_1ue_0_0.vlm"},
	{0x730AEE2BDFDEA788ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_run.gad"},
	{0x730B192FAC4B08A1ULL, "vrc321ip_spooky5_obj_7.utp"},
	{0x730BFE766C9CE155ULL, "493282597.ogg"},
	{0x730C138D445A4F15ULL, "gen00fl_anders6125959.cl"},
	{0x730CE465F78E91EDULL, "vxht_0_hightown_a2_8rr_0_0.vlm"},
	{0x730DC37519EC7CA1ULL, "hlac_lt_backalley_n_1_1z.dds"},
	{0x730E63A83C1FA864ULL, "6215088_m.fxe"},
	{0x730E8E86110654EBULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.cess.ani"},
	{0x730EB941132F2DE0ULL, "vxlt_0_lowtown_docks_n_b52_0_0.vlm"},
	{0x730F5341B4403245ULL, "179110961.ogg"},
	{0x730FA2088EDD52AAULL, "gen00fl_anders6209984.cl"},
	{0x730FAD2B8D1636B3ULL, "6018792_m.fxe"},
	{0x731009C6238733B1ULL, "6015850_m.fxe"},
	{0x7310735915720AADULL, "6015840_f.fxe"},
	{0x7310C46B3277A682ULL, "hlac_ht_mansion_n_2995_1x.dds"},
	{0x731163E4E8F76B03ULL, "996230068.ogg"},
	{0x7312329A652238E1ULL, "vxlt_backalley_n_1u2_0_0.vlm"},
	{0x73125320F55EC0BFULL, "6217971_m.fxe"},
	{0x7312920F03F0C0B2ULL, "420282264.ogg"},
	{0x7312F6A1F75B59C6ULL, "mcr370_varric6208288.cl"},
	{0x7312FEE79916596FULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_aveline_andersintro.gad"},
	{0x73133E570838EB06ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_103a_0t.dds"},
	{0x731380236D55F377ULL, "art\\vfx\\materials\\vfx_smoke_puffy_temp.mao"},
	{0x731402A7CFD46EDEULL, "abi_m_hex_damage.wbd"},
	{0x7314FE201103E588ULL, "vxw_blightlands_u4_0_0.vlm"},
	{0x731500F81727C7A7ULL, "6112964_m.fxe"},
	{0x731579766D2D05A1ULL, "vxlt_0_lowtown_docks_a3_n_d09_0_0.vlm"},
	{0x7315BED4530AFF01ULL, "vxca_caverns1_gkf_0_0.vlm"},
	{0x7315EDBCDC46E26AULL, "6197593_m.fxe"},
	{0x731617829CAC33E8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_low.msh"},
	{0x731629829CAC5279ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_low.mao"},
	{0x731635829CAC6612ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_low.mmh"},
	{0x73164460DB897E6AULL, "vxlt_undercityserialkiller_190_0_0.vlm"},
	{0x7316470E053552ACULL, "gen00fl_bethany6097879.cl"},
	{0x73166B69EB4D391DULL, "dre000_cut2d_ragstoriches6025115.cl"},
	{0x7316D1294C7F3CF9ULL, "197962986.ogg"},
	{0x731749F547D53A3EULL, "ico_1h_sword_0.dds"},
	{0x731765FCAC459C2FULL, "gen00fl_isabela6080449.cl"},
	{0x73180D4FE93A87D9ULL, "vxlt_0_lowtown_a3_n_c_hq3_0_0.vlm"},
	{0x7318154AB835EC47ULL, "904087469.ogg"},
	{0x7318F4684B9BD11CULL, "zz_mcr_debug6080878.cl"},
	{0x7319276D08BDF9E1ULL, "l_fen311ar_hanged_man.bnk"},
	{0x73192AEFFE9D6603ULL, "vxw_bonepitmining_a2_1rd_0_0.vlm"},
	{0x73196FBEBAEF1781ULL, "plt_one100pt_ggt_notify.plo"},
	{0x731A1D1704FAF740ULL, "vxca_caverns1_mines_jtj_0_0.vlm"},
	{0x731A2E94EDA8947EULL, "791376278.ogg"},
	{0x731A3B5F52BDD364ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxa.msh"},
	{0x731A4D5F52BDF1CDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxa.mao"},
	{0x731A515F52BDF806ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxa.mmh"},
	{0x731A7569C7BBC7A2ULL, "vxht_0_hightown_a2_c5k_0_0.vlm"},
	{0x731ABDE36A4D56A2ULL, "llac_lt_0_lowtown_docks_n_-131074_0.dds"},
	{0x731ACB098E793030ULL, "6099352_m.fxe"},
	{0x731AD13FBF0D1A0DULL, "gen00fl_fenris6113831.cl"},
	{0x731AEE645A2B1473ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtb_0s.dds"},
	{0x731B76861110B662ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.cess.evt"},
	{0x731BBB502676E4C8ULL, "vxlt_backalley_n_1u7_0_0.vlm"},
	{0x731BF2ADC8D6E7C6ULL, "zz_mer_debug6045143.cl"},
	{0x731C18D6A88B9779ULL, "seb321_leliana6191368.cl"},
	{0x731C2E9B8FF1BBCDULL, "6103674_m.fxe"},
	{0x731CAFC8FB736BD7ULL, "6129768_m.fxe"},
	{0x731DCC9DB909036AULL, "ser120ip_poster.utp"},
	{0x731E0614331AA1B9ULL, "1057970729.ogg"},
	{0x731E7547ED43E048ULL, "vxw_blightlands_13o_0_0.vlm"},
	{0x731F2521E3440EFEULL, "vxlt_0_lowtown_docks_b84_0_0.vlm"},
	{0x731F8413AF0D7F54ULL, "151569188.ogg"},
	{0x73202696C0098D7FULL, "vxht_mansion2b_2nz_0_0.vlm"},
	{0x732030FBA2EB22EEULL, "vxlt_warehouse_n_dh_0_0.vlm"},
	{0x7320422E95007941ULL, "6129043_m.fxe"},
	{0x732078297E5F19B6ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.dura.ani"},
	{0x73209A0EB05F04D4ULL, "893419693.ogg"},
	{0x73210A2CF003846DULL, "hlac_w_sundermount_-2_0y.dds"},
	{0x73217BD56E1B89C1ULL, "641850380.ogg"},
	{0x732197DD9537D7C4ULL, "mag131_keran6086183.cl"},
	{0x732275861116D375ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.cess.gad"},
	{0x732423CD9CB4FB0EULL, "6131409_m.fxe"},
	{0x7324485F52C63DD1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_brazierxa.phy"},
	{0x7324742324920C94ULL, "6092080_f.fxe"},
	{0x7324921E190B24FDULL, "twirl_air.dds"},
	{0x7324A49E6D1B463CULL, "hlac_lt_0_lowtown_-3_0y.dds"},
	{0x7324F6F462AFCF39ULL, "llac_lt_0_lowtown_docks_a3_196607_1.dds"},
	{0x7325667687499B60ULL, "hlac_lt_0_lowtown_a3_n_c_-1_8y.dds"},
	{0x73259D29A308F16FULL, "6102947_m.fxe"},
	{0x7325B50E29FCBACAULL, "6075351_f.fxe"},
	{0x7325BEB048CDD435ULL, "782006990.ogg"},
	{0x7325D647CD0B885BULL, "dre000_stroud6036708.cl"},
	{0x7325FF557DF1E5A1ULL, "dae000_bodhan6166373.cl"},
	{0x73261BA65928A198ULL, "qun110_petrice6150072.cl"},
	{0x7326AD50A64B27ACULL, "289095386.ogg"},
	{0x7326E1B13EEA9573ULL, "817211133.ogg"},
	{0x73271BE799273335ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_aveline_andersintro.ani"},
	{0x732759CB5BE444F2ULL, "w_encounter2_camp_12_0_ll.mmh"},
	{0x73277B3AA0FF3F6EULL, "art\\vfx\\combat\\vfx_c_single_tag_stun_c.anb"},
	{0x73278111B5FF404EULL, "6139434_m.fxe"},
	{0x73279765F049CF8EULL, "hlac_lt_0_lowtown_a2_n_-3_6x.dds"},
	{0x7327D2A4F0D3BFA8ULL, "character_character_p281.psh.fxo.dx11.assembly"},
	{0x7327FD3D9FA87307ULL, "hlac_w_sundermount_-196603_0z.dds"},
	{0x732815D1ECF93080ULL, "vxlt_undercityencounter_1dq_0_0.vlm"},
	{0x7328245A641A2131ULL, "gen00fl_varric6118755.cl"},
	{0x73287B3A2F942D0DULL, "29024820.ogg"},
	{0x7328D93E15007989ULL, "6109434_m.fxe"},
	{0x732945D495218496ULL, "vfx_vfx_p1.vsh.fxo.dx11"},
	{0x73294B79E7DC263AULL, "vxht_0_hightown_a3_9cn_0_0.vlm"},
	{0x7329534155B45DD8ULL, "qcr220_qunari_hunter6054470.cl"},
	{0x73297A539A097302ULL, "506866715.ogg"},
	{0x732A3E5C6DEC911FULL, "art\\characters\\playercharacter\\elffemale\\rob_appc.msh"},
	{0x732A445C6DEC9BC1ULL, "art\\characters\\playercharacter\\elffemale\\rob_appc.mmh"},
	{0x732AA43B03982850ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_activ_c.mmh"},
	{0x732B09433129F45DULL, "850732936.ogg"},
	{0x732B8BFDBDF69878ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_c.anb"},
	{0x732BEDA8217FBC13ULL, "207999308.ogg"},
	{0x732C495E8EDCA999ULL, "434288781.ogg"},
	{0x732CF254DEA761C6ULL, "6059978_m.fxe"},
	{0x732CF2FFBE12939AULL, "vxlt_0_lowtown_docks_a3_n_cw1_0_1.vlm"},
	{0x732E6341E8DD5232ULL, "vxlt_warehouse_n_15t_0_0.vlm"},
	{0x732E63AD9D7E201CULL, "gen00fl_anders6211891.cl"},
	{0x732E7EFCCAF647F0ULL, "6148829_m.fxe"},
	{0x732E80297E6B64A7ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.dura.evt"},
	{0x732E88692806C882ULL, "66062336.ogg"},
	{0x732EB7E8296B2BF7ULL, "6082023_m.fxe"},
	{0x732F0F525411218CULL, "591284272.ogg"},
	{0x732F1C914BC4C160ULL, "vfx_lavacom02.dds"},
	{0x732F5D0ACC52736EULL, "dre100cr_stroud.utc"},
	{0x732F84C2EFFDE483ULL, "6162404_m.fxe"},
	{0x732F93D864EC0F21ULL, "6154263_m.fxe"},
	{0x733034B04F4CFAB3ULL, "840983041.ogg"},
	{0x733072AE590F04DAULL, "6142123_f.fxe"},
	{0x73307C3EDF6ED460ULL, "vxca_caverns2_96e_0_0.vlm"},
	{0x733087C1BAC5AD29ULL, "isa111_isabela6067068.cl"},
	{0x73310BFE06C5DCAEULL, "6016491_f.fxe"},
	{0x73318B817E1BDF51ULL, "w_encounter3_47_0.rml"},
	{0x733268FED0E660F7ULL, "one280_iwan6096747.cl"},
	{0x7332BD1F26835715ULL, "llac_ht_0_hightown_131070_0.dds"},
	{0x73332BAC17B33E17ULL, "51623187.ogg"},
	{0x733331F4CD3F53D2ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\harvester.anb"},
	{0x7333D70BEEE7E4C0ULL, "zz_dog_debug6204102.cl"},
	{0x7333DC6E91F302DCULL, "vxlt_0_lowtown_docks_a3_d03_0_0.vlm"},
	{0x73343577F21A44FBULL, "178432617.ogg"},
	{0x73344045B4434AE3ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.anb"},
	{0x733440B722EFF908ULL, "ht_0_hightown_28_0_ll.mmh"},
	{0x73346447DEDD65F7ULL, "561907133.ogg"},
	{0x7334A1297E7008F0ULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.dura.gad"},
	{0x7334B1B6D2C9D321ULL, "art\\vfx\\combat\\vfx_c_rw_crust_c.anb"},
	{0x7334BF9DB91BCB97ULL, "ser120ip_poster.ncs"},
	{0x7334D21072745EF7ULL, "6120778_m.fxe"},
	{0x7334F37F1555FA4BULL, "123475555.ogg"},
	{0x733516DFCEDFA1DBULL, "vxlt_undercitydungeon_1d2_0_1.vlm"},
	{0x7335271EF0B3CDE2ULL, "plt_dre120pt_reqs_ready.nss"},
	{0x7335DC0409EE7232ULL, "mag311_keran6041566.cl"},
	{0x73367DA7E4D15EF4ULL, "zz_dog_debug6204098.cl"},
	{0x73374B1B06F69209ULL, "6102534_m.fxe"},
	{0x7337C73B03A2EE6FULL, "art\\vfx\\combat\\vfx_w_bsk_charge_activ_c.anb"},
	{0x7337F81851959ABAULL, "water_water_p5.psh.fxo.dx11.assembly"},
	{0x733806833A2DB89CULL, "gencr_raider_reaver.utc"},
	{0x73380C2D2CC2B6F1ULL, "llac_lt_blackemporium_1055_12.dds"},
	{0x7338502622E12F1DULL, "6054040_m.fxe"},
	{0x7338A43AF1E8156CULL, "follower_banter6151260.cl"},
	{0x7338A4A9320A12E3ULL, "980892896.ogg"},
	{0x7338CAEF3120302EULL, "art\\vfx\\models\\vfx_arrow.msh"},
	{0x7338E4EF31205C1CULL, "art\\vfx\\models\\vfx_arrow.mmh"},
	{0x73390B9A45FBA6A8ULL, "vxht_mansion2b_1yw_0_0.vlm"},
	{0x73390F3DEEC7B6E6ULL, "dae200cr_viscount6087367.cl"},
	{0x73392DEA6366B514ULL, "fx_snow_fall.dds"},
	{0x7339A38D2F6BBD79ULL, "ach_soc_respect.dds"},
	{0x733B0DF69BB7CAE5ULL, "vxlt_0_lowtown_docks_a3_n_d1b_0_0.vlm"},
	{0x733B34302546AB8EULL, "6139962_m.fxe"},
	{0x733B522058A23223ULL, "247455985.ogg"},
	{0x733BF13DD7557946ULL, "one120_athenril6088328.cl"},
	{0x733C2C50CC24BB38ULL, "vxht_brothel_n_5es_0_0.vlm"},
	{0x733C46DC6A11057CULL, "6083961_f.fxe"},
	{0x733C7123318075E9ULL, "6175785_m.fxe"},
	{0x733C9097D0C667A3ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockdxxl.mao"},
	{0x733C9497D0C66EF0ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockdxxl.mmh"},
	{0x733CA297D0C6860AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockdxxl.msh"},
	{0x733CCE135C461B8AULL, "vxdr_0_deeproads_gfq_0_0.vlm"},
	{0x733D37D3BCC8A678ULL, "vxlt_gallowstemplar_n_359_0_0.vlm"},
	{0x733D93781348CCC8ULL, "6152233_m.fxe"},
	{0x733DF4D56CDECE98ULL, "vxw_encounter2_camp_5o_0_0.vlm"},
	{0x733E3D9906B5D4B0ULL, "vxlt_0_lowtown_docks_a3_n_bvd_0_0.vlm"},
	{0x733E593EDEB1997DULL, "hightown_fem_elfservant_mob6169097.cl"},
	{0x733F13D42CB9ABFAULL, "vxlt_0_lowtown_docks_a3_bgf_0_0.vlm"},
	{0x733F1757D651D3B4ULL, "hlac_dr_dungeon_3215_5z.dds"},
	{0x733F1D311138C015ULL, "hla_ht_0_hightown_a2_-196608_0x.dds"},
	{0x734014DE120DBD31ULL, "716912410.ogg"},
	{0x734034C3B9CB3E93ULL, "714726464.ogg"},
	{0x7340452E4C98F57EULL, "mag131_keran6086286.cl"},
	{0x734058CE9B2CFEC1ULL, "dre000_dougal6131198.cl"},
	{0x73408E6FAFEC213AULL, "mer320cr_hubert.utc"},
	{0x7340D1749CBD8652ULL, "6095260_m.fxe"},
	{0x7340EC4F53CBE4E4ULL, "vxlt_0_lowtown_h7p_0_0.vlm"},
	{0x734116A4CF66D74FULL, "vxht_bartrand_n_5ve_0_0.vlm"},
	{0x73411D45B44D990CULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.mmh"},
	{0x73418B23F00F41B9ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_001a_0s.dds"},
	{0x73424E4B83E91B5CULL, "vxlt_0_lowtown_docks_a3_d50_0_0.vlm"},
	{0x73426FC226E886FAULL, "6141752_m.fxe"},
	{0x734292B6D2D5DB9AULL, "art\\vfx\\combat\\vfx_c_rw_crust_c.mmh"},
	{0x734297A43DDF7491ULL, "6078565_m.fxe"},
	{0x7342D4F355B04E65ULL, "hlac_ht_keep_n_4519_13x.dds"},
	{0x7342EE2F45DC9A45ULL, "199587912.ogg"},
	{0x7343255C7F5EF33DULL, "6085247_f.fxe"},
	{0x7343460518A3F591ULL, "6101958_f.fxe"},
	{0x7343B197205927C9ULL, "6188626_m.fxe"},
	{0x7343FB5F03183AC2ULL, "vxca_caverns2_7em_0_1.vlm"},
	{0x73443349DAA93F6BULL, "237790007.ogg"},
	{0x73444EE7DABB3503ULL, "dae000_bartrand6201395.cl"},
	{0x734456D9C2389B04ULL, "w_sundermount_mmi.gff"},
	{0x73448C7325F20B9CULL, "hlac_ht_bartrand_n_3370_1y.dds"},
	{0x7344EE20A93D8D55ULL, "gen_im_trs_dem_lgt_glv_r1.uti"},
	{0x73464D74EDE9580EULL, "6160850_m.fxe"},
	{0x73465D749A9E939DULL, "qun120cr_javaris6042439.cl"},
	{0x7346A3876BA91440ULL, "vxht_0_hightown_a3_7m0_0_0.vlm"},
	{0x7346FA0D5FD4E952ULL, "973762354.ogg"},
	{0x734736C352AE0A4BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bazaar_sa.dds"},
	{0x7347680CF2FAED68ULL, "462260321.ogg"},
	{0x7347D05F85E479BAULL, "hlac_w_encounter1_65538_0z.dds"},
	{0x7348481AB69AABB5ULL, "art\\characters\\creatures\\dragon\\textures\\cn_eye_draa_0d.dds"},
	{0x7349A3E680E217ACULL, "hlac_lt_0_lowtown_docks_a3_n_131071_3z.dds"},
	{0x734A2A605691BE18ULL, "qun221_petrice6076139.cl"},
	{0x734AB101C6A04F4EULL, "lowtown_male_urchin_sta6164185.cl"},
	{0x734AB73427ABF57BULL, "qun210ip_at_exit.utp"},
	{0x734B7C2D95EBD6FFULL, "vxht_mansion2b_2ag_0_0.vlm"},
	{0x734B869278CD6C6EULL, "vxlt_warehouse_n_uk_0_0.vlm"},
	{0x734BBF65DF6F1E9DULL, "llac_lt_0_lowtown_docks_n_-3_6.dds"},
	{0x734C1033045AA9F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_butt.evt"},
	{0x734C10589B5F8AA0ULL, "vxw_woundedcoast_1oq_0_0.vlm"},
	{0x734C513C9DE6CDB6ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_conjure.ani"},
	{0x734D47A328DE22E1ULL, "fen110_fenris6100311.cl"},
	{0x734E8F7C367C5E66ULL, "6097406_m.fxe"},
	{0x734ECA5AA7BC943AULL, "830127340.ogg"},
	{0x734F0FB01AB5361DULL, "one100ar_ruin_ext.lst"},
	{0x734F2391C53D226DULL, "gen00fl_anders6214891.cl"},
	{0x734F8881C15CEFCFULL, "vxht_bartrand_n_5yd_0_0.vlm"},
	{0x735034EA6379A042ULL, "fx_snow_fall.mao"},
	{0x7350EDAAC9744893ULL, "6084353_m.fxe"},
	{0x73510AA2DFF87563ULL, "6116189_f.fxe"},
	{0x7351CA60FEED1ED5ULL, "386019125.ogg"},
	{0x7351D4C3B112E9D9ULL, "vxlt_0_lowtown_a3_n_c_hm7_0_0.vlm"},
	{0x7352395FAD69178DULL, "vxht_keep_4mg_0_0.vlm"},
	{0x73523DF3495151BAULL, "966644632.ogg"},
	{0x73524F20DDCE144EULL, "265207505.ogg"},
	{0x735270AF2FFACF78ULL, "774937111.ogg"},
	{0x7352947274A1A63EULL, "121050405.ogg"},
	{0x7352C096E38826BBULL, "204178258.ogg"},
	{0x735309330460BC34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_butt.gad"},
	{0x735383DCAC143036ULL, "vrc221_bartrand6065282.cl"},
	{0x73538AFDBE17E813ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_c.mmh"},
	{0x73540D146246AC32ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.mmh"},
	{0x735507B01AB9940CULL, "one100ar_ruin_ext.ncs"},
	{0x73553BA20E3DFD07ULL, "llac_ht_chantry_n_2534_50.dds"},
	{0x735541E03509AFEFULL, "gen00fl_anders6119411.cl"},
	{0x73555008405264A8ULL, "6001201_m.fxe"},
	{0x73556C711313B88CULL, "docks_male_sailor_mob6176659.cl"},
	{0x7355B80E497CA4B8ULL, "vxlt_0_lowtown_docks_b4f_0_0.vlm"},
	{0x735630CD764B712EULL, "266522832.ogg"},
	{0x7356446220F8CEC8ULL, "6083463_m.fxe"},
	{0x7356455C6E11D83EULL, "art\\characters\\playercharacter\\elffemale\\rob_appc.phy"},
	{0x7356467E396D0401ULL, "vxlt_0_lowtown_docks_a3_d0b_0_0.vlm"},
	{0x7357E09C035B06E4ULL, "6165918_m.fxe"},
	{0x7357EBB5869605B9ULL, "758803659.ogg"},
	{0x7357ED0F3C5A56F2ULL, "vxht_mansion2b_216_0_0.vlm"},
	{0x7358004658C7EBAFULL, "gen00fl_isabela6118371.cl"},
	{0x73580C401D883933ULL, "1033918147.ogg"},
	{0x73589C0716980842ULL, "character_character_p232.psh.fxo.dx11.assembly"},
	{0x7358C14A49E9FEEDULL, "plt_seb000pt_exiled_prince.plo"},
	{0x7359151077A27DD6ULL, "mcr310_cut_chantry_blows_up6116289.cl"},
	{0x7359183304654112ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_butt.ani"},
	{0x73599FFBE3C3B827ULL, "vxlt_0_lowtown_a2_n_c_en0_0_0.vlm"},
	{0x7359F86F6AC384EEULL, "vxlt_warehouse_n_xl_0_0.vlm"},
	{0x735A21BE06AE6BB5ULL, "582265250.ogg"},
	{0x735A593C9DF318A7ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_conjure.evt"},
	{0x735B142941B3C17CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_iron.mao"},
	{0x735B202941B3D51FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_iron.mmh"},
	{0x735B262941B3DFC1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_iron.msh"},
	{0x735BB3B2301351B8ULL, "gen00fl_sebastian6195873.cl"},
	{0x735D33BBC5462F47ULL, "framebuffereffects_cocdownsample_p0.psh.fxo.dx11.assembly"},
	{0x735D4FEE10B28884ULL, "gen00fl_fenris6107154.cl"},
	{0x735DAAB14A332929ULL, "vxlt_warehouse_10g_0_0.vlm"},
	{0x735DB84CD7D6A226ULL, "and211ar_andersshop.lst"},
	{0x735E59EC6A2D2BFBULL, "mag311_emile_de_launcet6043622.cl"},
	{0x735EA3E4F5F608F6ULL, "6020938_m.fxe"},
	{0x735EEBCB98974AB6ULL, "vxlt_0_lowtown_docks_a3_n_d2x_0_0.vlm"},
	{0x735F214F68F925C4ULL, "w_bonepitmining_a2_222_0.mmh"},
	{0x735F21B933BEB642ULL, "77278762.ogg"},
	{0x735F254F68F92C17ULL, "w_bonepitmining_a2_222_0.mao"},
	{0x735F374F68F94B66ULL, "w_bonepitmining_a2_222_0.msh"},
	{0x735F510496D032AFULL, "vxlt_0_lowtown_a2_n_f54_0_0.vlm"},
	{0x735F57B5337AE9FCULL, "ser120_ghyslain6105650.cl"},
	{0x735FA4D864569B17ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack01.gad"},
	{0x735FB2D2ED5C1836ULL, "zz_followers_banter6159699.cl"},
	{0x73600C7012EF01BEULL, "vxht_0_hightown_4n1_0_0.vlm"},
	{0x7360356D08FA1393ULL, "l_fen311ar_hanged_man.wbd"},
	{0x736043F70176252EULL, "art\\characters\\creatures\\pridedemon\\animation\\d_lp.ani"},
	{0x73607A3C9DF7BCF0ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_conjure.gad"},
	{0x736082B585150036ULL, "293177063.ogg"},
	{0x7360855BE412D4F0ULL, "6119526_m.fxe"},
	{0x7360A77AD658FA9BULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_106a_sa.dds"},
	{0x736113E23E52EA97ULL, "gen_im_trs_gen_coinbag_03.uti"},
	{0x73620A82B9778836ULL, "6166360_m.fxe"},
	{0x7362785A4760C10DULL, "6205026_m.fxe"},
	{0x7363162078245A59ULL, "vxw_encounter2_rc_0_0.vlm"},
	{0x736321988ED087DDULL, "mnd_imp_dust.mao"},
	{0x73634871A660B897ULL, "310907571.ogg"},
	{0x736348B7D996D37DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgecrgrck.phy"},
	{0x736388C34E896AEAULL, "hlac_lt_0_lowtown_a2_65533_0y.dds"},
	{0x7363A2A59DE0E898ULL, "vxca_caverns1_j2m_0_0.vlm"},
	{0x7363DC6C7C567AEBULL, "6107575_m.fxe"},
	{0x7363EDCD55713DD4ULL, "6122509_f.fxe"},
	{0x7364B09175D60DACULL, "vxlt_0_lowtown_a2_n_ekl_0_2.vlm"},
	{0x7365141EF0DCD165ULL, "plt_dre120pt_reqs_ready.plo"},
	{0x73652785A03518B0ULL, "art\\vfx\\models\\vfx_debris_01.msh"},
	{0x73653585A03530CAULL, "art\\vfx\\models\\vfx_debris_01.mmh"},
	{0x73653B6E06174B64ULL, "6198687_f.fxe"},
	{0x73655B6F63F795F8ULL, "vxlt_0_lowtown_docks_n_b7p_0_0.vlm"},
	{0x73660758B9C18FD0ULL, "678468801.ogg"},
	{0x7366B9D8645CDDA4ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack01.evt"},
	{0x7366BCC6BE6965E7ULL, "vxht_0_hightown_a3_c0l_0_0.vlm"},
	{0x73672EFB8400A0E0ULL, "hlac_w_encounter3_-196610_0y.dds"},
	{0x7367B89BE5809277ULL, "82729623.ogg"},
	{0x7367C17DD9BC4049ULL, "character_character_p4.psh.fxo"},
	{0x7367C74CD7DF1031ULL, "and211ar_andersshop.are"},
	{0x73685D89456DDD58ULL, "637000714.ogg"},
	{0x73686A2044AB2FFAULL, "vxlt_darktown_6i1_0_0.vlm"},
	{0x736914927285FB93ULL, "ht_0_hightownexplosion_n_37_0_lc.mmh"},
	{0x73695E6C896EC70DULL, "dae000ip_crowd_ff3_talk5.utp"},
	{0x73696AB5E7CAA2F1ULL, "hlac_ca_caverns2prc_5802_8y.dds"},
	{0x7369932FDC981E30ULL, "vxlt_0_lowtown_docks_a3_n_bgc_0_0.vlm"},
	{0x736A685115CEB63DULL, "hlac_ht_chantry_n_2534_12x.dds"},
	{0x736A90F3D8F6F017ULL, "vxca_cavern1_haunted_hm4_0_0.vlm"},
	{0x736C03E77D6D3E5EULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_act_c.mmh"},
	{0x736C1EA949B904D3ULL, "7974985.ogg"},
	{0x736C2E400A1E97F0ULL, "w_bonepitdestroyed_a3_1_0_ll.mmh"},
	{0x736C44A4A080685CULL, "llac_lt_blackemporium_1055_6.dds"},
	{0x736C960EB149B068ULL, "hlac_lt_gallowsprison_n_-1_14z.dds"},
	{0x736C9CA2A0F3BF98ULL, "one140cr_charade6137080.cl"},
	{0x736CA09334DEE2E3ULL, "art\\characters\\weapons\\mauls\\w_mal_ml104a.mmh"},
	{0x736CAE9334DEFA3DULL, "art\\characters\\weapons\\mauls\\w_mal_ml104a.msh"},
	{0x736CEF466348014AULL, "vxlt_undercityserialkiller_21c_0_0.vlm"},
	{0x736D15C1E7EC2176ULL, "6027838_m.fxe"},
	{0x736D2BAC7D8B9715ULL, "vxht_keep_63g_0_0.vlm"},
	{0x736D4E8722492A1BULL, "vxdr_0_deeproads_cme_0_0.vlm"},
	{0x736E22A74D5DB657ULL, "ht_0_hightown_a3_53_0_ll.mmh"},
	{0x736ECCF251CD5204ULL, "hlac_lt_undercitydungeon_-1_5z.dds"},
	{0x737014EAACEE8FBDULL, "6143137_m.fxe"},
	{0x7370328D58956B67ULL, "867932444.ogg"},
	{0x737069F35E18D1F8ULL, "vxca_caverns1_mines_ckb_0_0.vlm"},
	{0x7370E9A82570578AULL, "204915917.ogg"},
	{0x73718EE42C0C71E8ULL, "mag291_ntbanter6056724.cl"},
	{0x7371E5C6AF6731B9ULL, "vxw_woundedcoast_19v_0_0.vlm"},
	{0x737234829CFAC30DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_low.phy"},
	{0x7372BD2460ED85A3ULL, "6118153_m.fxe"},
	{0x73739943B4E9BF1FULL, "6117426_m.fxe"},
	{0x7373B333B7AEC2FCULL, "vxlt_hangedman_n_27s_0_0.vlm"},
	{0x7373D1D86467909DULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack01.ani"},
	{0x7373F94B026B2CD7ULL, "and121_karl6093876.cl"},
	{0x73743DB7D9A533E9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgecrgrck.mao"},
	{0x737449B7D9A54782ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgecrgrck.mmh"},
	{0x73744BB7D9A54B38ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgecrgrck.msh"},
	{0x737490A9B35F5D9DULL, "mag370_meredith6046920.cl"},
	{0x73749BC31FB6E7C1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_firestatue.phy"},
	{0x7374AF3E98487C19ULL, "1057629256.ogg"},
	{0x7374F6E815AB8701ULL, "llac_w_woundedcoast_2_0.dds"},
	{0x7375132E0ADD50FEULL, "6129215_m.fxe"},
	{0x737515ECEDB38E21ULL, "6108446_m.fxe"},
	{0x73751CF701883F68ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_lp.gad"},
	{0x73752EA3B05F7E67ULL, "vxht_0_hightown_a3_c5o_0_0.vlm"},
	{0x73753CE7B7193B90ULL, "vrc321_party_barks6096313.cl"},
	{0x7375497B60734093ULL, "636875169.ogg"},
	{0x737579DBA673DC1BULL, "follower_barks6153725.cl"},
	{0x73758620BC08BE26ULL, "77591825.ogg"},
	{0x7375D0FBE759A5F7ULL, "dae000_orana6149100.cl"},
	{0x73762432E9F2ED06ULL, "26777159.ogg"},
	{0x73765A089CC96A9AULL, "vxht_bartrand_n_6k4_0_0.vlm"},
	{0x73765F106DEA0B67ULL, "hlac_lt_0_lowtown_a3_n_c_-3_7y.dds"},
	{0x7376CD19BA538200ULL, "6060716_m.fxe"},
	{0x73775FA8CC4A1A53ULL, "llac_lt_house_-1_5.dds"},
	{0x7377C0E8C71AA38AULL, "vxlt_warehouse_n_cu_0_0.vlm"},
	{0x73781F70F1A0751BULL, "vxht_chantry_2ic_0_0.vlm"},
	{0x73787D9FEE01C33EULL, "vxlt_0_lowtown_f4u_0_0.vlm"},
	{0x7378C0D7D86B1F12ULL, "hlac_lt_undercitydungeon_-1_20y.dds"},
	{0x7378E2E77D77901DULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_act_c.anb"},
	{0x7379A7DBD4B0DB11ULL, "6089688_m.fxe"},
	{0x737AEDEFFD9F7EC1ULL, "127175327.ogg"},
	{0x737B18B01ADA80FAULL, "one100ar_ruin_ext.are"},
	{0x737B2E409E4F9C3BULL, "vxca_cave_2ig_0_0.vlm"},
	{0x737BC064143F429BULL, "vxdr_0_deeproads_cjd_0_0.vlm"},
	{0x737C607D9BE7C442ULL, "vxlt_gallowstemplar_f_5z0_0_0.vlm"},
	{0x737CB46E2D9C9472ULL, "6111559_m.fxe"},
	{0x737CC9E4CA6A3607ULL, "750759839.ogg"},
	{0x737CCFA3D7E92681ULL, "ht_playershouse_n_mmap_ph.dds"},
	{0x737CD03E9BD686D1ULL, "1038683033.ogg"},
	{0x737D0C146269AFD9ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.anb"},
	{0x737D11FCD7932962ULL, "856732410.ogg"},
	{0x737D537F04335280ULL, "vxlt_0_lowtown_docks_a3_n_aot_0_0.vlm"},
	{0x737D9B78D7D9034FULL, "gen_im_wep_rog_dua_act3_07.uti"},
	{0x737D9B97D0FE110BULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockdxxl.phy"},
	{0x737DA19334ED5708ULL, "art\\characters\\weapons\\mauls\\w_mal_ml104a.phy"},
	{0x737E316E55129F27ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_medium.gad"},
	{0x737EC13245040F0AULL, "ts_profane_arm_props.gda"},
	{0x737F6DA35B78CCA5ULL, "vxlt_0_lowtown_docks_a3_d4z_0_0.vlm"},
	{0x7380976BE3461BBDULL, "vxw_bonepitmining_a2_243_0_0.vlm"},
	{0x7380C7050C4E0C9CULL, "450328493.ogg"},
	{0x738106FC3513E118ULL, "703873491.ogg"},
	{0x73822510A9A4DB7CULL, "vxlt_hangedman_n_22p_0_0.vlm"},
	{0x7382E8B039B99BF6ULL, "6064854_f.fxe"},
	{0x7383473D74815C3FULL, "vxlt_gallowstemplar_3o3_0_0.vlm"},
	{0x7383B52ED3B4D10CULL, "6154756_m.fxe"},
	{0x7383DF0926A32315ULL, "vxlt_gallowstemplar_f_43q_0_0.vlm"},
	{0x7384009E527F01ADULL, "141788985.ogg"},
	{0x7385250F2C14935EULL, "734761637.ogg"},
	{0x7385266E5518AB34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_medium.evt"},
	{0x73854CF68901B1F6ULL, "6087371_m.fxe"},
	{0x73859ABFA2B85A64ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit_1.mao"},
	{0x7385C66C3B239C2FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpouta.mao"},
	{0x7385D26C3B23B0CCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpouta.mmh"},
	{0x7385D86C3B23BABEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpouta.msh"},
	{0x738631CE1EB2608FULL, "vxca_caverns1_mines_k6v_0_0.vlm"},
	{0x738667D5F825A63AULL, "lt_0_lowtown_docks_a3_n_28_0_lc.mmh"},
	{0x73870206B3AB022EULL, "vxlt_gallowstemplar_f_3gq_0_0.vlm"},
	{0x73871C8CFDE13966ULL, "6125308_f.fxe"},
	{0x73874844BA18015DULL, "6171561_m.fxe"},
	{0x738748E76C47ED4AULL, "vxw_bonepitmining_a2_21o_0_0.vlm"},
	{0x7387E270D79D6A27ULL, "vxlt_0_lowtown_a2_n_c_h2e_0_0.vlm"},
	{0x7388795B596C427DULL, "6087642_m.fxe"},
	{0x7388DC4DC5CF7AE1ULL, "hlac_lt_darktown_-1_25x.dds"},
	{0x7388EFFB5245D48CULL, "vxw_woundedcoast_6k_0_0.vlm"},
	{0x7388F0CCF92C9E35ULL, "vxw_encounter2_qk_0_1.vlm"},
	{0x738914181028BB42ULL, "6089841_m.fxe"},
	{0x7389190F05459DB1ULL, "320620383.ogg"},
	{0x73898081A018AE03ULL, "dae000_fenarel6133352.cl"},
	{0x738985FE671787EDULL, "6018810_f.fxe"},
	{0x7389FCF23EC9C126ULL, "hlac_w_sundermount_5_0x.dds"},
	{0x7389FEEBF8699A57ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcs.mmh"},
	{0x738A02EBF869A164ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcs.mao"},
	{0x738A0F666177C815ULL, "vxht_keep_69e_0_0.vlm"},
	{0x738A105951CA7619ULL, "71842236.ogg"},
	{0x738A14EBF869BFC9ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcs.msh"},
	{0x738A622A50FBD71CULL, "vxlt_darktown_38m_0_0.vlm"},
	{0x738A7090E54D858BULL, "6103665_m.fxe"},
	{0x738AB1ECE6CE4796ULL, "ser270_emeric6149395.cl"},
	{0x738AF3E787668D21ULL, "dae000ip_crowd_mf4_talk1.utp"},
	{0x738AF584D4F598E5ULL, "gen00fl_aveline6139683.cl"},
	{0x738B10728396B2CBULL, "6172760_m.fxe"},
	{0x738B562F557B91AAULL, "6115296_m.fxe"},
	{0x738B9F0F6D19B736ULL, "zz_one000_barker6179349.cl"},
	{0x738BA8647D3D1E63ULL, "6119002_m.fxe"},
	{0x738C4D8FA438ABBBULL, "hlac_ca_cavern1_haunted_24426_11y.dds"},
	{0x738C5CE2F2FE2DC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_action.gad"},
	{0x738C6933A363367DULL, "vxca_caverns2prc_4oc_0_0.vlm"},
	{0x738C6EB1FFD1AEBBULL, "6079577_m.fxe"},
	{0x738CD8533715E07DULL, "367839652.ogg"},
	{0x738D018E60CD69AEULL, "hlac_lt_0_lowtown_a2_n_65534_5z.dds"},
	{0x738D46E7E664EA42ULL, "6156891_m.fxe"},
	{0x738F3A3522A985B7ULL, "gen00fl_aveline6139850.cl"},
	{0x7390210A45B3A577ULL, "vxca_cave_2cq_0_0.vlm"},
	{0x73902F6B0DD702D2ULL, "810314319.ogg"},
	{0x739030373299D1A4ULL, "640361240.ogg"},
	{0x7390699291F63FB4ULL, "6148946_m.fxe"},
	{0x73908BD7C86D53C6ULL, "vxht_mansion_n_6co_0_0.vlm"},
	{0x7390E066BF3E3BA4ULL, "vxht_keep_n_5zu_0_0.vlm"},
	{0x7391F82D751BB28FULL, "6037618_m.fxe"},
	{0x7391FD9EE70609D5ULL, "plt_cod_ltr_sk_meredith.nss"},
	{0x73928A12408E05B3ULL, "213751763.ogg"},
	{0x7392BADE7F9FBB6FULL, "77977316.ogg"},
	{0x7392E78A4609BAFEULL, "hlac_lt_gallowscourtyard_n_-1_1z.dds"},
	{0x739316363022E24CULL, "art\\levels\\races\\design\\prp_smalltreasur_0.pwk"},
	{0x73931E6E5524DB6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_medium.ani"},
	{0x73933DC280B59B9AULL, "vxht_bartrand_n_6q6_0_0.vlm"},
	{0x739351E2F30439D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_action.evt"},
	{0x7393B060B0AE7D0EULL, "6074056_m.fxe"},
	{0x7393B300E05F4785ULL, "vxlt_0_lowtown_docks_n_b7i_0_0.vlm"},
	{0x739451832FDBFD7EULL, "vxht_0_hightown_n_c6q_0_0.vlm"},
	{0x73950A0445733D04ULL, "vxlt_gallowscourtyard_n_70t_0_0.vlm"},
	{0x7396097738C5FA4EULL, "hlac_lt_0_lowtown_65537_3z.dds"},
	{0x7396BE504AF215F7ULL, "gen00fl_isabela6079016.cl"},
	{0x73974226B451E05FULL, "46299050.ogg"},
	{0x739761D8443BED0CULL, "vxw_woundedcoast_3h_0_0.vlm"},
	{0x73978CCFC562ACD4ULL, "6067798_f.fxe"},
	{0x73978F18EBED06BCULL, "6101022_m.fxe"},
	{0x73979B4486753D15ULL, "hlac_dr_dungeon_3207_1x.dds"},
	{0x7397E953A6D0AEC2ULL, "6147440_m.fxe"},
	{0x7397EB7D1C24FD5BULL, "dae000ip_crowd_lm3_talk8.utp"},
	{0x73986A20DF7C2C0FULL, "6067069_f.fxe"},
	{0x739895B96F03EC9BULL, "344086262.ogg"},
	{0x7398AF79D43D20C4ULL, "one110_gustav6089806.cl"},
	{0x7398EE33A1085D36ULL, "ht_0_hightown_a2_42_0.rml"},
	{0x739A5C985E940608ULL, "vxlt_darktown_48v_0_0.vlm"},
	{0x739B13A016A712F0ULL, "6093813_m.fxe"},
	{0x739B32E3BCC6C19FULL, "6087257_m.fxe"},
	{0x739B32EBCD89C61CULL, "hlac_lt_0_lowtown_a2_n_-196607_0x.dds"},
	{0x739C02F698A838F8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masf_0t.dds"},
	{0x739C84899CA8945CULL, "art\\characters\\playercharacter\\humanmale\\glv_meda.mao"},
	{0x739C90899CA8A8FFULL, "art\\characters\\playercharacter\\humanmale\\glv_meda.mmh"},
	{0x739C96899CA8B2A1ULL, "art\\characters\\playercharacter\\humanmale\\glv_meda.msh"},
	{0x739D2347CCC93B2AULL, "hlac_ca_caverns2prc_5802_21y.dds"},
	{0x739D278197825F01ULL, "747806601.ogg"},
	{0x739D75FFD1769BF6ULL, "788615423.ogg"},
	{0x739DA6DB236C05ADULL, "6165973_m.fxe"},
	{0x739DC097DD6DB00AULL, "lt_0_lowtown_docks_182_0.mao"},
	{0x739DC33EA5221DBDULL, "vxlt_0_lowtown_a2_n_fhu_0_0.vlm"},
	{0x739DC497DD6DB759ULL, "lt_0_lowtown_docks_182_0.mmh"},
	{0x739DCE97DD6DC847ULL, "lt_0_lowtown_docks_182_0.msh"},
	{0x739DDA7264868A9DULL, "vxca_caverns1_mines_eak_0_0.vlm"},
	{0x739E2DD58CDC4830ULL, "gen_im_arm_cht_elf_comc_t3.uti"},
	{0x739E6DFAA0FC6FEEULL, "112916521.ogg"},
	{0x739EA9FA3CBAD165ULL, "vxht_keep_637_0_0.vlm"},
	{0x739EE140CE5EBA0DULL, "llac_lt_0_lowtown_n_131073_0.dds"},
	{0x739EF42DC502B843ULL, "w_bonepitmining_a2_48_0.rml"},
	{0x739EFED0E7F3D689ULL, "static_static_p55.psh.fxo.dx11"},
	{0x739FAE1FFD64E887ULL, "follower_banter6160665.cl"},
	{0x73A049E2F30EB60DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\tackle_punch_action.ani"},
	{0x73A0739FF2245829ULL, "mi_gla_sht.mao"},
	{0x73A07945F0117C9FULL, "xx_extblack.dds"},
	{0x73A0EA48F3102058ULL, "6118477_m.fxe"},
	{0x73A15C52E719A3F5ULL, "vxlt_0_lowtown_a2_gza_0_0.vlm"},
	{0x73A2223DA2A4958EULL, "6079356_m.fxe"},
	{0x73A25CE9A89084DEULL, "lt_gallowscourtyard_n_18_0_lc.mmh"},
	{0x73A2C36021D215FEULL, "vxht_0_hightown_n_c3r_0_0.vlm"},
	{0x73A2F32EDD222D94ULL, "dae311ar_gamlens_house_nt.lst"},
	{0x73A2F6390DEFA0B4ULL, "gen_im_arm_glv_lgt_lore_01.uti"},
	{0x73A3167F9CFCED0AULL, "vxlt_warehouse_n_pp_0_0.vlm"},
	{0x73A32F5911A9627AULL, "llac_lt_gallowstemplar_7198_3.dds"},
	{0x73A36015979D3A27ULL, "28530779.ogg"},
	{0x73A39975EF435190ULL, "qun160ip_chest_1.utp"},
	{0x73A3E0FE831DAEB4ULL, "hlac_ht_keep_n_3734_0y.dds"},
	{0x73A4806440FC930DULL, "partycombat.cut"},
	{0x73A4D131FB1BFAA7ULL, "346066561.ogg"},
	{0x73A50379B04F9B1AULL, "llac_ht_chantry_n_2534_28.dds"},
	{0x73A5184F1A48AE94ULL, "vxht_keep_5dx_0_0.vlm"},
	{0x73A5CFE33CF448F7ULL, "6082250_m.fxe"},
	{0x73A5E19FB4B31BDAULL, "vxlt_0_lowtown_a2_n_e23_0_0.vlm"},
	{0x73A5EB4988C0A16AULL, "vxdr_0_deeproads_efc_0_0.vlm"},
	{0x73A60039B84D477EULL, "96170179.ogg"},
	{0x73A652788BC5D1C5ULL, "813712308.ogg"},
	{0x73A6550D78E0C48EULL, "6168106_m.fxe"},
	{0x73A6CE1D57595CB1ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_7z.dds"},
	{0x73A76B8CA791CCC0ULL, "6135263_m.fxe"},
	{0x73A7B61471CA0DC2ULL, "vxlt_gallowstemplar_f_5k3_0_0.vlm"},
	{0x73A7C3755F8F0215ULL, "6042533_m.fxe"},
	{0x73A834A3495C03DBULL, "w_woundedcoast_196_0.mmh"},
	{0x73A840A3495C1878ULL, "w_woundedcoast_196_0.mao"},
	{0x73A852A3495C36C5ULL, "w_woundedcoast_196_0.msh"},
	{0x73A88A655D3A6CD9ULL, "vxht_brothel_n_6cs_0_0.vlm"},
	{0x73A8B6EBCAEBE19AULL, "367324043.ogg"},
	{0x73A8BD3F542A57CEULL, "911268243.ogg"},
	{0x73A8BEF187E4D6B8ULL, "467024605.ogg"},
	{0x73A8D7FB03DC8472ULL, "174490124.ogg"},
	{0x73A8F7E0421CED42ULL, "hlac_lt_hangedman_n_-1_2z.dds"},
	{0x73A95C1B01FC24EDULL, "hlac_ca_caverns2prc_5802_11y.dds"},
	{0x73AA2F3E668B0241ULL, "vxw_encounter2_gd_0_0.vlm"},
	{0x73AA88B884B0CF43ULL, "695561247.ogg"},
	{0x73AAAD0E163ABE09ULL, "vxlt_0_lowtown_a2_n_c_ep4_0_0.vlm"},
	{0x73AABDDF9E2F915BULL, "183830293.ogg"},
	{0x73AB891AAF019DAEULL, "mag291_vincento6027715.cl"},
	{0x73AC488B4226685FULL, "vxw_woundedcoast_1cf_0_0.vlm"},
	{0x73AD47EE38C3D561ULL, "hlac_lt_0_lowtown_a2_-65536_8z.dds"},
	{0x73AD7B23FAB76916ULL, "506382770.ogg"},
	{0x73AD7FF4DBDF6FD3ULL, "6144620_m.fxe"},
	{0x73ADBDEB5D97B538ULL, "6171199_m.fxe"},
	{0x73ADE5F258E089A7ULL, "character_character_p60.psh.fxo.dx11"},
	{0x73ADFBD009DB0150ULL, "65114762.ogg"},
	{0x73AE617FB2ABD236ULL, "gen00fl_isabela6201252.cl"},
	{0x73AE7F1D86044871ULL, "isa211_isabela6202057.cl"},
	{0x73AE9A27EDA1E734ULL, "art\\vfx\\combat\\vfx_impact_c.anb"},
	{0x73AEB0D1DC06E07EULL, "6174442_m.fxe"},
	{0x73AEF8249FEED821ULL, "w_woundedcoast_203_0.phy"},
	{0x73B0646299C40DB5ULL, "vxw_woundedcoast_y7_0_0.vlm"},
	{0x73B10013F9C15768ULL, "vxlt_gallowstemplar_n_5ao_0_0.vlm"},
	{0x73B1353D13C82E7EULL, "vxht_0_hightown_n_c8s_0_0.vlm"},
	{0x73B1885C8EF3058AULL, "vxlt_warehouse_n_wq_0_0.vlm"},
	{0x73B1C32D47D922A2ULL, "vxca_cavern1_haunted_cot_0_0.vlm"},
	{0x73B2AE53895186CBULL, "abi_m_arcane_bolt.wbd"},
	{0x73B34FC39B9990DEULL, "vxht_0_hightown_n_avs_0_0.vlm"},
	{0x73B3A03E5278DC36ULL, "6169956_m.fxe"},
	{0x73B3BDC8E7AE1C76ULL, "884526236.ogg"},
	{0x73B3C403941F9FECULL, "hlac_ht_keep_n_3732_11y.dds"},
	{0x73B3FC9EE722EE36ULL, "plt_cod_ltr_sk_meredith.plo"},
	{0x73B4240862C3A179ULL, "vxht_0_hightown_a3_b5t_0_0.vlm"},
	{0x73B473BEB35A315BULL, "674338701.ogg"},
	{0x73B4B3C2945B3B26ULL, "gen00fl_isabela6101460.cl"},
	{0x73B4B61FD9729A09ULL, "lt_0_lowtown_docks_10_0_lc.mmh"},
	{0x73B5515177BB6EEEULL, "vxlt_darktown_4jz_0_0.vlm"},
	{0x73B5F01874EC11D1ULL, "fen110_fenris6067904.cl"},
	{0x73B6320CEA396E09ULL, "vxlt_0_lowtown_a2_n_c_f4s_0_0.vlm"},
	{0x73B678ABCDCB397FULL, "ch1emmultnt.psh.fxo"},
	{0x73B681266B4AE0DCULL, "qun120_arishok6053232.cl"},
	{0x73B686B59CC22C63ULL, "qcr000_arishok6096400.cl"},
	{0x73B82075EF57490DULL, "119109411.ogg"},
	{0x73B86403BD1176AFULL, "292861658.ogg"},
	{0x73B879DFF15EFECDULL, "vxlt_0_lowtown_a3_n_c_ehq_0_0.vlm"},
	{0x73B9F0BEB392F145ULL, "hlac_ht_chantry_n_2534_24y.dds"},
	{0x73BABA68341C80DFULL, "vxw_woundedcoast_1na_0_0.vlm"},
	{0x73BB9CB337292CCFULL, "6107456_m.fxe"},
	{0x73BBD18E5B5D808DULL, "251344765.ogg"},
	{0x73BBE16E9E4F9D46ULL, "vxht_mansion_n_6lr_0_0.vlm"},
	{0x73BBE179A4AB16DEULL, "hlac_lt_0_lowtown_a2_n_1_1z.dds"},
	{0x73BBF1D1CDD58853ULL, "6141520_m.fxe"},
	{0x73BC1B0FBAAE7799ULL, "llac_ca_caverns1_mines_24424_21.dds"},
	{0x73BC4466E5224F4CULL, "ttx_window_small_a_1.msh"},
	{0x73BC4A66E522593EULL, "ttx_window_small_a_1.mmh"},
	{0x73BC7927EDADECF7ULL, "art\\vfx\\combat\\vfx_impact_c.mmh"},
	{0x73BDCE02D6FC5158ULL, "6112677_m.fxe"},
	{0x73BE35043EFCC636ULL, "42172817.ogg"},
	{0x73BEAD8C96248175ULL, "follower_banter6161071.cl"},
	{0x73BED63F8BBA2635ULL, "vxw_woundedcoast_v6_0_0.vlm"},
	{0x73BEF8DA740E35CEULL, "hlac_w_bonepitdestroyed_a3_65537_0x.dds"},
	{0x73C12676D80FC674ULL, "630904343.ogg"},
	{0x73C22EA9EB92E880ULL, "207514414.ogg"},
	{0x73C25D71F9364236ULL, "685739822.ogg"},
	{0x73C26D725113763BULL, "lit_aoe_distor_l3.mao"},
	{0x73C2A196C8C3E13CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_talk_2.gad"},
	{0x73C2A545A9650DF0ULL, "hlac_dr_slavers_3207_6z.dds"},
	{0x73C318FCDD144D56ULL, "6068310_m.fxe"},
	{0x73C3D95D04126675ULL, "follower_barks6153374.cl"},
	{0x73C42D44C0324620ULL, "mer210_smith6051399.cl"},
	{0x73C46B380B7A9E1DULL, "vxht_0_hightown_a2_bzh_0_0.vlm"},
	{0x73C4BB7975137B02ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xhole.msh"},
	{0x73C4BD7975137EB8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xhole.mmh"},
	{0x73C4C9797513935BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xhole.mao"},
	{0x73C4D816730B5D68ULL, "vxlt_0_lowtown_docks_n_bd9_0_0.vlm"},
	{0x73C4F49EEC1F0797ULL, "aip_kirkwallcommander_base.gda"},
	{0x73C52F94130A6966ULL, "vxlt_darktown_5af_0_0.vlm"},
	{0x73C56E1F41269DD9ULL, "vxht_brothel_n_6iq_0_0.vlm"},
	{0x73C612BC0B11EC83ULL, "vxca_cavern1_haunted_icq_0_2.vlm"},
	{0x73C6814B56268C03ULL, "vxw_bonepitmining_a2_16p_0_0.vlm"},
	{0x73C6A026D413917FULL, "vxlt_0_lowtown_a2_n_f1p_0_0.vlm"},
	{0x73C6E26C94B7BEB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_0.pwk"},
	{0x73C7528309DBD522ULL, "6164432_m.fxe"},
	{0x73C76151CBB5E9B0ULL, "6116545_m.fxe"},
	{0x73C87DBAD0AA7CC6ULL, "dae000_merchant_tailor_hm6164349.cl"},
	{0x73C885CA7DF97B2DULL, "dae200_cullen6140255.cl"},
	{0x73C8F23159BCB24EULL, "vxlt_gallowstemplar_f_2rw_0_0.vlm"},
	{0x73C8FA2EDD430967ULL, "dae311ar_gamlens_house_nt.are"},
	{0x73C9273F5A597587ULL, "hlac_w_encounter2_camp_131073_0x.dds"},
	{0x73C93BEA40E4440AULL, "weapon_rendertoshadowmap_p0.psh.fxo"},
	{0x73C956B7312F51CAULL, "fx_lightning_4.mao"},
	{0x73C96337A0866802ULL, "gen00fl_varric6093662.cl"},
	{0x73C96916E5A0F6BDULL, "zz_dae_debug6029793.cl"},
	{0x73C9A096C8C9FE3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_talk_2.ani"},
	{0x73C9A8D4678D9632ULL, "vxca_cave_22z_0_1.vlm"},
	{0x73CA11C616E6245AULL, "vxca_caverns1_mines_j5s_0_0.vlm"},
	{0x73CAE6C570855CACULL, "hlac_lt_0_lowtown_a2_n_-131072_2y.dds"},
	{0x73CAF403541A6BDFULL, "6040151_m.fxe"},
	{0x73CB2F3F039D7F3FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerbunch01_0_d.dds"},
	{0x73CB7DD000EB33EBULL, "6048234_m.fxe"},
	{0x73CBA5B69CB387D5ULL, "lt_0_lowtown_docks_a3_n_138_0.phy"},
	{0x73CBBE470C2F17C3ULL, "zz_fol_debug6096538.cl"},
	{0x73CC460E3CB802F2ULL, "dae011_gossips6149793.cl"},
	{0x73CC979CE8AD0A17ULL, "131312138.ogg"},
	{0x73CC9BFC8F48C6F8ULL, "vxht_0_hightown_7fx_0_0.vlm"},
	{0x73CD350EAEF2158CULL, "zz_endwhite_pt1.cut"},
	{0x73CD481C7DB03EB5ULL, "vxw_woundedcoast_s5_0_0.vlm"},
	{0x73CD7CF21792B37FULL, "vxht_mansion2b_22n_0_0.vlm"},
	{0x73CD81AA8AC41532ULL, "vxlt_backalley_n_1y1_0_0.vlm"},
	{0x73CDD7203EDEF0B8ULL, "mag210_arianni6031720.cl"},
	{0x73CE10731C4DEB3FULL, "zz_isa_debug6133952.cl"},
	{0x73CE8BD88120D91EULL, "vxca_caverns1_mines_imt_0_0.vlm"},
	{0x73CF01931B8D38E7ULL, "m_dae321ar_brothel_nt.wbd"},
	{0x73CF32056ACC568CULL, "llac_lt_0_lowtown_a3_n_c_-1_1.dds"},
	{0x73CF380CE944C8C8ULL, "65359226.ogg"},
	{0x73CFFB19398FBE7AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowflat.phy"},
	{0x73D00EF070EE4AB8ULL, "vxht_keep_n_49e_0_0.vlm"},
	{0x73D0B273F248982EULL, "6078701_f.fxe"},
	{0x73D0F7CD79B325C1ULL, "6174467_m.fxe"},
	{0x73D10C12FD459BE4ULL, "hlac_dr_0_deeproads_5842_1z.dds"},
	{0x73D1118C3B0F8AA4ULL, "vxdr_primevalentrance_1a6_0_0.vlm"},
	{0x73D18DB269E3630AULL, "6070885_m.fxe"},
	{0x73D19DBB0377A486ULL, "6071842_m.fxe"},
	{0x73D1BD8AD60F2FAAULL, "dre000_bodahn6033404.cl"},
	{0x73D1E22DC52DF35AULL, "w_bonepitmining_a2_48_0.msh"},
	{0x73D1F02DC52E0A93ULL, "w_bonepitmining_a2_48_0.mao"},
	{0x73D1F42DC52E11A0ULL, "w_bonepitmining_a2_48_0.mmh"},
	{0x73D20578025440C9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackl.mao"},
	{0x73D2117802545562ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackl.mmh"},
	{0x73D2137802545818ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackl.msh"},
	{0x73D245B05BB7B139ULL, "470141242.ogg"},
	{0x73D25D6B90BF095DULL, "hlac_lt_0_lowtown_n_65536_0z.dds"},
	{0x73D26BB928622670ULL, "79874962.ogg"},
	{0x73D29DD2176683CDULL, "hlac_w_sundermount_262143_0z.dds"},
	{0x73D2EB33A139D0EAULL, "ht_0_hightown_a2_42_0.anb"},
	{0x73D2F5F61B4C7C3DULL, "6180910_m.fxe"},
	{0x73D33BFAD137A831ULL, "hlac_dr_slavers_3217_3y.dds"},
	{0x73D350A668F23490ULL, "vxlt_0_lowtown_docks_a3_n_b99_0_0.vlm"},
	{0x73D3A171050081E6ULL, "vxlt_darktown_5jg_0_0.vlm"},
	{0x73D41DFCB984F8C4ULL, "art\\vfx\\combat\\vfx_glb_ele_s_major_c.anb"},
	{0x73D42F609E77082BULL, "650376987.ogg"},
	{0x73D42FA5D2494A63ULL, "llac_lt_0_lowtown_docks_65534_0.dds"},
	{0x73D504C64B90868DULL, "ser211ip_to_foundry.utp"},
	{0x73D57AB50C1F40B5ULL, "448717527.ogg"},
	{0x73D61B42D914E4B0ULL, "797702930.ogg"},
	{0x73D63F38FF4A37ACULL, "29685909.ogg"},
	{0x73D65C0A767D240BULL, "vxht_mansion2_1t3_0_0.vlm"},
	{0x73D65E896BC4B7D9ULL, "641019769.ogg"},
	{0x73D6748CAB61A134ULL, "fake_rogue_female.utc"},
	{0x73D6A0C32009BAFDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_firestatue.mao"},
	{0x73D6A4C32009C136ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_firestatue.mmh"},
	{0x73D6AEC32009D234ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_firestatue.msh"},
	{0x73D6C7D269C890C2ULL, "hlac_ht_mansion2_818_8z.dds"},
	{0x73D6D1BC11746573ULL, "324214100.ogg"},
	{0x73D7F966CDEBAFF6ULL, "6166513_m.fxe"},
	{0x73D8905389722299ULL, "abi_m_arcane_bolt.bnk"},
	{0x73D93A3207123AEBULL, "837238628.ogg"},
	{0x73D972F17AD3959FULL, "follower_banter6173085.cl"},
	{0x73DA143D7B6450DEULL, "vxlt_hangedman_1w0_0_1.vlm"},
	{0x73DA43A31989F73BULL, "static1_s1.psh.fxo"},
	{0x73DA885A2DC4F3A7ULL, "938481754.ogg"},
	{0x73DAE69E199DD9F6ULL, "360272462.ogg"},
	{0x73DB23C6AEF205F5ULL, "vxlt_0_lowtown_a2_gnm_0_0.vlm"},
	{0x73DB563F31ABEDF0ULL, "vxlt_gallowscourtyard_6ym_0_0.vlm"},
	{0x73DB5C530FE7D848ULL, "dae000_bartrand6030639.cl"},
	{0x73DC56B268F3DB51ULL, "hlac_ht_mansion2b_818_49z.dds"},
	{0x73DCE563770D0017ULL, "vxlt_undercitydungeon_1c2_0_0.vlm"},
	{0x73DCFDB57316F19EULL, "vxca_caverns1_mines_iju_0_0.vlm"},
	{0x73DD6B5CC254FD17ULL, "59310115.ogg"},
	{0x73DDD34061C36A27ULL, "follower_banter6154213.cl"},
	{0x73DE3FB30A646920ULL, "vxlt_darktown_5r9_0_0.vlm"},
	{0x73DE58240DAD1F85ULL, "gen00fl_anders6103678.cl"},
	{0x73DE95B01AED67E6ULL, "llac_ht_brothel_n_3369_2.dds"},
	{0x73DF7AA79AA6EE48ULL, "624479585.ogg"},
	{0x73DF9B9930FF0937ULL, "vxht_0_hightown_caz_0_0.vlm"},
	{0x73DFCCD113C59CBDULL, "lowtown_male_alienage2_mob6166530.cl"},
	{0x73E007D323EF839AULL, "ser210_witness6168114.cl"},
	{0x73E053BEB25E8EBCULL, "vxht_brothel_n_3jh_0_0.vlm"},
	{0x73E062454BF70B2BULL, "and100_anders6095890.cl"},
	{0x73E0B6432ACBC591ULL, "948514248.ogg"},
	{0x73E0FBCFA1BE1E2BULL, "6095427_m.fxe"},
	{0x73E1296D3AFA55E8ULL, "gen_im_acc_blt_act3_02.uti"},
	{0x73E13CFCB98FB7C7ULL, "art\\vfx\\combat\\vfx_glb_ele_s_major_c.mmh"},
	{0x73E1A2B80B862C25ULL, "6099397_m.fxe"},
	{0x73E1F03EED1D2C8CULL, "red50.dds"},
	{0x73E203502E63F3B0ULL, "vxht_bartrand_n_5om_0_0.vlm"},
	{0x73E2134DF6F69A66ULL, "vxlt_darktown_5oh_0_0.vlm"},
	{0x73E274D717AB9F8AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2c.mao"},
	{0x73E278D717ABA6D9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2c.mmh"},
	{0x73E282D717ABB7C7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2c.msh"},
	{0x73E2B4067D1DF55FULL, "6089916_m.fxe"},
	{0x73E2CC3764333F71ULL, "94894867.ogg"},
	{0x73E2E1964EE134D6ULL, "6179799_m.fxe"},
	{0x73E3933390D853A6ULL, "hlac_w_encounter2_327682_0z.dds"},
	{0x73E3B4078EDACCF6ULL, "6156110_m.fxe"},
	{0x73E3F3849ED4F9FEULL, "vxw_woundedcoast_a0_0_0.vlm"},
	{0x73E3F43F9750C602ULL, "lowtown_male_lowlife1_sta6165003.cl"},
	{0x73E3F6021AE50F6BULL, "vxht_chantry_2s1_0_0.vlm"},
	{0x73E42B4D3111AC4EULL, "gen_im_trs_slv_lgt_boo_r0.uti"},
	{0x73E4456A8CCF8174ULL, "vxht_mansion_n_4xn_0_0.vlm"},
	{0x73E5A36A1F87C518ULL, "vxca_caverns1_jgb_0_0.vlm"},
	{0x73E5CF53812843CEULL, "ser221_party_comments6122877.cl"},
	{0x73E5DBB1C24711FDULL, "6030451_m.fxe"},
	{0x73E63B4898199372ULL, "44062174.ogg"},
	{0x73E642B204EA8FCDULL, "and100_anders6096053.cl"},
	{0x73E7E5A803B7E374ULL, "vxlt_gallowsprison_n_6ml_0_0.vlm"},
	{0x73E831BB1FE7AA69ULL, "gen_im_arm_cht_noba_t4.uti"},
	{0x73E880E687D2EC66ULL, "hm_ave1_banditm.mor"},
	{0x73E8A750ECC9E9AEULL, "hm_dumar.mor"},
	{0x73EA3612A8E3FA5EULL, "gen00fl_fenris6126742.cl"},
	{0x73EAF948EC3077D7ULL, "qcr000_arishok6057087.cl"},
	{0x73EB098530779CA3ULL, "nat380cr_nathaniel.utc"},
	{0x73EB3273D8B0EDB0ULL, "424259864.ogg"},
	{0x73EB342B708A3D92ULL, "fen310_fenris6117675.cl"},
	{0x73EB807780B9320EULL, "894051645.ogg"},
	{0x73EC40BB687C48DEULL, "gen_im_cft_reg_c1.uti"},
	{0x73EC7C01931A0F16ULL, "lgt160ip_wk_avern_messenger.ncs"},
	{0x73ECECAE7192A3AFULL, "vxlt_gallowscourtyard_7ab_0_1.vlm"},
	{0x73ECFF7B755815A7ULL, "vxlt_0_lowtown_a2_n_c_hob_0_0.vlm"},
	{0x73ED779C7477032DULL, "47799274.ogg"},
	{0x73EDB965D39127ABULL, "308432250.ogg"},
	{0x73EDE316A7D23832ULL, "245556005.ogg"},
	{0x73EDFE687EF3947FULL, "plt_and300pt_opening.nss"},
	{0x73EF1A1728A68A44ULL, "follower_banter6143304.cl"},
	{0x73EF7CD0A4077BD2ULL, "hlac_lt_0_lowtown_a2_n_-131073_1y.dds"},
	{0x73F055A34999F110ULL, "w_woundedcoast_196_0.phy"},
	{0x73F084D842231D01ULL, "gen_im_trs_krk_grd_junk1_01.uti"},
	{0x73F0FCCB9AF6EF5BULL, "sst_ss_meeran.gda"},
	{0x73F13E6D0ED7EB69ULL, "870441704.ogg"},
	{0x73F28E17D69B4BE3ULL, "ser221_gascard6128298.cl"},
	{0x73F298F6BB3D9D7BULL, "vxht_keep_n_73k_0_0.vlm"},
	{0x73F33FC45A69550BULL, "vxht_mansion2_1b5_0_0.vlm"},
	{0x73F35AA255629E28ULL, "vxlt_0_lowtown_a2_gx2_0_0.vlm"},
	{0x73F36713E38B1F17ULL, "6204157_m.fxe"},
	{0x73F3F62D1D44FB8EULL, "442361534.ogg"},
	{0x73F425E01479D311ULL, "vxht_0_hightown_a3_c7s_0_0.vlm"},
	{0x73F4A2F49E7E2CE1ULL, "dae300_alistair_king6134954.cl"},
	{0x73F4F7E9E3BF0AF0ULL, "6157090_m.fxe"},
	{0x73F4F83246FA7AECULL, "vxht_0_hightown_a2_ap8_0_0.vlm"},
	{0x73F54EB26D901A0CULL, "458731576.ogg"},
	{0x73F55AC6D590B9CDULL, "hlac_lt_0_lowtown_-65536_8x.dds"},
	{0x73F58251A278496CULL, "vxw_encounter2_jr_0_0.vlm"},
	{0x73F684EFFE44049EULL, "vxht_mansion2b_31g_0_0.vlm"},
	{0x73F6F73B8E4F97A1ULL, "dae011cr_sabina.utc"},
	{0x73F76E9B5B28A445ULL, "mag160_varian_ilithis6016502.cl"},
	{0x73F7750734128316ULL, "hlac_dr_primevaltomb_664_3y.dds"},
	{0x73F7ABDC6BC8D250ULL, "6089619_f.fxe"},
	{0x73F7BF7CB36FA97AULL, "waterflowing0.psh.fxo"},
	{0x73F7C52DCF31E8DFULL, "6149524_m.fxe"},
	{0x73F8D3D8B2B1F98DULL, "409760586.ogg"},
	{0x73F9810AEE842607ULL, "254448569.ogg"},
	{0x73F9AFAE944271AFULL, "lt_0_lowtown_43_0_ll.mmh"},
	{0x73FA0957C39B326AULL, "follower_banter6163414.cl"},
	{0x73FA37BEDA53957EULL, "700571569.ogg"},
	{0x73FA62966DEB6799ULL, "6170382_m.fxe"},
	{0x73FAC931B62950B2ULL, "ht_brothel_n_probe_54o.dds"},
	{0x73FAEEB3582805AEULL, "6160887_m.fxe"},
	{0x73FB17F9D66DEC33ULL, "vxlt_0_lowtown_a2_n_f4o_0_0.vlm"},
	{0x73FB39693703C1A3ULL, "875905779.ogg"},
	{0x73FB983A7E50AA24ULL, "6044818_f.fxe"},
	{0x73FBAB0AF0254CBAULL, "9238640.ogg"},
	{0x73FC08D4003E8126ULL, "428930448.ogg"},
	{0x73FC0B2BCA0BD10CULL, "6071130_m.fxe"},
	{0x73FC64F19C8ACEF6ULL, "6087120_m.fxe"},
	{0x73FD4C49603DB171ULL, "vxw_woundedcoast_19n_0_0.vlm"},
	{0x73FDA2B721215EB8ULL, "vxw_woundedcoast_cl_0_0.vlm"},
	{0x73FE3E7627938396ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spike_n.dds"},
	{0x73FE61CD15800DB1ULL, "6218851_m.fxe"},
	{0x73FEA1E6D1DB276AULL, "703252937.ogg"},
	{0x73FED1A33E791DF9ULL, "vxlt_0_lowtown_a2_n_h8w_0_0.vlm"},
	{0x73FF57FC859E61D3ULL, "vxlt_0_lowtown_docks_a3_n_cxt_0_0.vlm"},
	{0x73FF99899CFD3544ULL, "art\\characters\\playercharacter\\humanmale\\glv_meda.phy"},
	{0x73FFA558F56E549AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ik_transition.gad"},
	{0x73FFC98A29CC47B9ULL, "6033546_m.fxe"},
	{0x73FFE723AAF1FD9EULL, "389286439.ogg"},
	{0x74001AA4FFE05612ULL, "lgt120_orz_bln_official6140189.cl"},
	{0x740024ED5C4C6676ULL, "325325723.ogg"},
	{0x7400305BD801E33CULL, "vxlt_backalley_g_1du_0_0.vlm"},
	{0x7400AA1F5F07B854ULL, "6083943_f.fxe"},
	{0x74011CF66AD413EAULL, "6064934_m.fxe"},
	{0x7401F1687F042690ULL, "plt_and300pt_opening.plo"},
	{0x74022E51C5C8EF40ULL, "67746596.ogg"},
	{0x7402B18F3352406FULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredacti.gad"},
	{0x7402C05F90C63B63ULL, "6119613_m.fxe"},
	{0x74031905E11447F8ULL, "464355988.ogg"},
	{0x74031A2BBA0BAE5FULL, "follower_banter6171869.cl"},
	{0x74034902F3F10B12ULL, "mrl310_merrill6120015.cl"},
	{0x74036A0F38F0936CULL, "vxht_0_hightown_a2_ak9_0_0.vlm"},
	{0x7403F46A5763307EULL, "vxw_bonepit_1w8_0_0.vlm"},
	{0x74043FF7DC3CAA26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_enter.evt"},
	{0x74047F25C258195BULL, "6058868_m.fxe"},
	{0x7404878CC56B1402ULL, "235151100.ogg"},
	{0x7404C8D85793E1ACULL, "6143143_m.fxe"},
	{0x7404D2FCD9BDF38EULL, "dae300_cullen6187889.cl"},
	{0x7405E49280719623ULL, "art\\characters\\creatures\\dog\\textures\\c_eye_dogma_0m.dds"},
	{0x74060A60A8B0D19EULL, "vxlt_gallowstemplar_6sg_0_0.vlm"},
	{0x740615C3CCA2D691ULL, "plt_prm000pt_blood_dragon_.nss"},
	{0x74067EB0C9B3A3E3ULL, "hlac_ht_0_hightown_n_131071_0z.dds"},
	{0x7406A858F57478B1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ik_transition.evt"},
	{0x7406DD87A62C9159ULL, "llac_lt_darktown_-1_18.dds"},
	{0x74075B2D7A52BD15ULL, "695769179.ogg"},
	{0x74080CE7482A90A1ULL, "6044625_m.fxe"},
	{0x74082EEAA8542598ULL, "vxw_encounter2_camp_oa_0_0.vlm"},
	{0x74085D009F617133ULL, "6021688_m.fxe"},
	{0x74086457AFDF5586ULL, "isa311_party_comments6117821.cl"},
	{0x740898B21229664BULL, "123574168.ogg"},
	{0x7408D8B246F30052ULL, "vxlt_0_lowtown_docks_a3_n_cxq_0_0.vlm"},
	{0x7409285074EAD23EULL, "vxlt_0_lowtown_n_gkn_0_0.vlm"},
	{0x740A56F7DC413D71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_enter.gad"},
	{0x740A92351EA7EB14ULL, "6028169_m.fxe"},
	{0x740A945107265418ULL, "gen_im_flw_arc_sebastian_01.uti"},
	{0x740ACED19FA4C8A4ULL, "gui_reticle_p0.vsh.fxo"},
	{0x740AD9FE12F42C88ULL, "6077099_m.fxe"},
	{0x740B71CAB1A8C971ULL, "729131073.ogg"},
	{0x740C07FD990DF73EULL, "vxht_mansion_n_541_0_0.vlm"},
	{0x740D4A658EDDCC8EULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyn.phy"},
	{0x740D613204D75D69ULL, "6069444_m.fxe"},
	{0x740DDAD8E4DE5A6BULL, "49328964.ogg"},
	{0x740E26C2F39DE8B7ULL, "6116748_m.fxe"},
	{0x740E2FBBCEBBEF2AULL, "vxht_mansion_n_6cc_0_0.vlm"},
	{0x740E707D8AAD90EDULL, "hla_ht_0_hightown_-65535_0x.dds"},
	{0x740EA542BF64444DULL, "36073103.ogg"},
	{0x740ED6360659610FULL, "lt_0_lowtown_a2_n_20_0_ll.mmh"},
	{0x740EF67D8BAE713EULL, "llac_w_encounter2_camp_-65536_0.dds"},
	{0x740F0B931BC32D1DULL, "m_dae321ar_brothel_nt.bnk"},
	{0x740F9EE4092E22F3ULL, "gen00fl_merrill6119606.cl"},
	{0x740FA6364E9DEC8CULL, "vxw_bonepitmining_a2_1xy_0_0.vlm"},
	{0x74102F246A436FE2ULL, "226079853.ogg"},
	{0x74106C7BCFAD8059ULL, "vxw_encounter2_6o_0_0.vlm"},
	{0x74106E900F5DBCCAULL, "vxht_mansion2b_2e2_0_0.vlm"},
	{0x7410A16193050B69ULL, "gen00fl_anders6119848.cl"},
	{0x7410FD24A041AB5DULL, "w_woundedcoast_203_0.mao"},
	{0x74110124A041B196ULL, "w_woundedcoast_203_0.mmh"},
	{0x74110B24A041C294ULL, "w_woundedcoast_203_0.msh"},
	{0x741157F7DC475D1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_enter.ani"},
	{0x7411C71CBBC76058ULL, "cod_mcg_resolutionists.utp"},
	{0x7411C818B4F6533FULL, "vxw_encounter2_or_0_0.vlm"},
	{0x7411E062C92AE8A3ULL, "6187867_m.fxe"},
	{0x7412660B86647A6CULL, "vxw_encounter2_pp_0_0.vlm"},
	{0x74126647495948FEULL, "vxw_bonepit_1x9_0_0.vlm"},
	{0x74128701933AF123ULL, "lgt160ip_wk_avern_messenger.utp"},
	{0x7412B8B69CEFAAF0ULL, "lt_0_lowtown_docks_a3_n_138_0.msh"},
	{0x7412C6B69CEFC20AULL, "lt_0_lowtown_docks_a3_n_138_0.mmh"},
	{0x7412CAB69CEFC941ULL, "lt_0_lowtown_docks_a3_n_138_0.mao"},
	{0x7412E7ED5363BE52ULL, "qun231cr_qunari1.utc"},
	{0x741401F0707CC58AULL, "6033695_m.fxe"},
	{0x741451951AEC47C2ULL, "vxht_0_hightown_a3_cbc_0_0.vlm"},
	{0x7414B3508A287A64ULL, "zz_lgt_1_debug6114363.cl"},
	{0x7414C058F580DE80ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ik_transition.ani"},
	{0x741604FDEA7E9B19ULL, "llac_lt_0_lowtown_a3_n_c_-196607_0.dds"},
	{0x7416126C14301AF7ULL, "gen00fl_bethany6101452.cl"},
	{0x7416CE8F33631A35ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow1_meredacti.ani"},
	{0x7419920C7406CFEDULL, "hlac_lt_0_lowtown_a3_n_c_0_2z.dds"},
	{0x7419A866E74B2D63ULL, "hlac_ht_chantry_n_2534_25x.dds"},
	{0x7419E2ECEA173555ULL, "vxlt_0_lowtown_docks_n_8pf_0_0.vlm"},
	{0x741A203E9E547BB1ULL, "197910684.ogg"},
	{0x741A3BFE73952EAEULL, "vxht_0_hightown_a2_a82_0_0.vlm"},
	{0x741A490A9935E476ULL, "6019149_m.fxe"},
	{0x741A7DAFA35E408DULL, "vxlt_hangedman_27q_0_0.vlm"},
	{0x741A8671050D8FB8ULL, "vxw_woundedcoast_uj_0_0.vlm"},
	{0x741B495FC77E52B3ULL, "llac_ca_caverns1_mines_24426_10.dds"},
	{0x741B4DEC9B1E4584ULL, "6137557_m.fxe"},
	{0x741BC54D804160FBULL, "806616974.ogg"},
	{0x741C87D9F7EA96B1ULL, "587566909.ogg"},
	{0x741D8D897A0DF4E1ULL, "6061358_m.fxe"},
	{0x741DE631F7BE49F7ULL, "hlac_lt_0_lowtown_docks_n_1_0y.dds"},
	{0x741EDE58C1A398D9ULL, "vxw_encounter2_1n_0_0.vlm"},
	{0x741F2E4C04402366ULL, "477247450.ogg"},
	{0x741FAFD64B6E9D3FULL, "vxht_0_hightown_a2_ap7_0_0.vlm"},
	{0x74210B782EE48ED4ULL, "art\\vfx\\combat\\vfx_q_ari_wwind_act_c.mmh"},
	{0x7421DFE79AE16732ULL, "6030680_f.fxe"},
	{0x74224A280EEE039EULL, "vxca_caverns1_mines_j22_0_0.vlm"},
	{0x7423B73B47B4910FULL, "422476990.ogg"},
	{0x7423F71EA5FFC1B1ULL, "art\\vfx\\combat\\vfx_r_dwp_right_close1_c.impa.ani"},
	{0x7423FA107C138831ULL, "zz_isa_debug6133998.cl"},
	{0x7424913EFDBA5F64ULL, "vxca_caverns2prc_aay_0_0.vlm"},
	{0x7424EABADD2AC4C2ULL, "vxht_bartrand_n_3s9_0_0.vlm"},
	{0x7425D9823CEDCE8BULL, "6164011_m.fxe"},
	{0x7426218570FDA408ULL, "vxlt_warehouse_n_vt_0_0.vlm"},
	{0x74264DE41CEF3BFDULL, "gen00fl_sebastian6177900.cl"},
	{0x74268A862754B21DULL, "vxlt_0_lowtown_a2_hec_0_0.vlm"},
	{0x742776A0CE70D217ULL, "6180501_m.fxe"},
	{0x7427947DFF99C042ULL, "vxlt_gallowstemplar_n_5r0_0_0.vlm"},
	{0x7428029D6D12B854ULL, "gen_im_npc_wep_shd_slva.uti"},
	{0x742854C9DC0D4DD5ULL, "vxlt_0_lowtown_docks_n_8ue_0_0.vlm"},
	{0x74296AF7137976C1ULL, "vxht_brothel_n_6bh_0_0.vlm"},
	{0x742985DE1FEE335AULL, "vxlt_warehouse_az_0_0.vlm"},
	{0x742986FC395AAC53ULL, "zz_and_debug6134065.cl"},
	{0x742998F3E5D236D7ULL, "6069806_m.fxe"},
	{0x7429AC6E02C489BCULL, "6037290_f.fxe"},
	{0x7429EF03A328C52EULL, "lt_0_lowtown_docks_a3_n_11_0_lc.mmh"},
	{0x7429FCDE52B79E88ULL, "291265850.ogg"},
	{0x742A4BD0054AED4AULL, "6076858_m.fxe"},
	{0x742A5F002205264CULL, "hlac_lt_0_lowtown_a2_n_c_1_8y.dds"},
	{0x742A6EF5A9084CA5ULL, "6101460_f.fxe"},
	{0x742A805901BD1C6EULL, "fen310_party_comment6116844.cl"},
	{0x742AB54728A7B3D4ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_magb_0n.dds"},
	{0x742AF41939DC5DA3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowflat.msh"},
	{0x742B021939DC75FDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowflat.mmh"},
	{0x742B021EA605F353ULL, "art\\vfx\\combat\\vfx_r_dwp_right_close1_c.impa.gad"},
	{0x742B061939DC7C0EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowflat.mao"},
	{0x742B7F4B5C431FB5ULL, "vxca_caverns1_eb7_0_0.vlm"},
	{0x742C4C59422FC5E4ULL, "vxlt_0_lowtown_a2_n_c_hjr_0_0.vlm"},
	{0x742C58896220D50AULL, "hlac_ht_0_hightown_a3_65536_2y.dds"},
	{0x742CA39D0603AB3EULL, "37139500.ogg"},
	{0x742CB2A8969BE773ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtc.phy"},
	{0x742D1FC5E178934BULL, "qcr270cr_fake_templar.utc"},
	{0x742E70B63587DA15ULL, "combat_deathblow_13.gda"},
	{0x742ED9314FC52B58ULL, "zz_approval_debug6003658.cl"},
	{0x742F54076CFD5458ULL, "819008198.ogg"},
	{0x742F8FAB6A76D2C6ULL, "dae000_pc_home_ambients6208334.cl"},
	{0x742FA1D3ADE12331ULL, "791167475.ogg"},
	{0x742FC1323CBD3CB3ULL, "302790338.ogg"},
	{0x742FE290CBD07102ULL, "lt_0_lowtown_a2_38_0_lc.mmh"},
	{0x743015F97C23136DULL, "538949939.ogg"},
	{0x7430FF1EA60A59B8ULL, "art\\vfx\\combat\\vfx_r_dwp_right_close1_c.impa.evt"},
	{0x743155EEFF3EC7DBULL, "darktown_male_fer_urch_sta6174959.cl"},
	{0x74315B78B32F0A6CULL, "hlac_w_bonepitmining_a2_327686_0z.dds"},
	{0x7431B65B5A4105C7ULL, "6118122_m.fxe"},
	{0x74323607175CFE2EULL, "vxlt_gallowstemplar_n_3oq_0_0.vlm"},
	{0x7432400CA9BA19D1ULL, "671307177.ogg"},
	{0x7432A0387E9C157EULL, "gen00fl_varric6033729.cl"},
	{0x7432E6EEDB5A245AULL, "dae000_porfiria6155385.cl"},
	{0x743379E52147B78AULL, "w_encounter3_20_0_ll.mmh"},
	{0x74344F17A054E24EULL, "6083372_f.fxe"},
	{0x74349673CD3D3ED0ULL, "gen00fl_merrill6121154.cl"},
	{0x7434FC63194ACA9DULL, "vxlt_0_lowtown_a2_hbb_0_0.vlm"},
	{0x74357EA42F8A2092ULL, "dae000cr_grand_cleric6199334.cl"},
	{0x7435859A794387B2ULL, "dae000cr_commoner_d2.utc"},
	{0x7435A595A4D52553ULL, "hla_ht_0_hightown_a3_-2_0y.dds"},
	{0x7437ADFC9C5EB539ULL, "6137549_m.fxe"},
	{0x7437BB27E7B0E9D8ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgecrgrck.phy"},
	{0x7437E860E84E9BCDULL, "ach_oth_mercenary.dds"},
	{0x74382013E2859AE6ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v1.ani"},
	{0x74388DD6E2A18CB2ULL, "str_deca_hole.mao"},
	{0x7438930CACABC17BULL, "6085286_m.fxe"},
	{0x74394116C4271388ULL, "vxlt_darktown_4qm_0_0.vlm"},
	{0x7439CEEC1EA5C23BULL, "hlac_ca_cavern1_haunted_24424_27z.dds"},
	{0x743A014C14463760ULL, "gen00fl_anders6096647.cl"},
	{0x743A220A0028F330ULL, "zz_vanessa_test.stg"},
	{0x743A4038A4879B04ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexcity_sa.dds"},
	{0x743A88740AD2D2A7ULL, "vxlt_hangedman_n_21n_0_0.vlm"},
	{0x743AF09838941474ULL, "vxht_mansion_n_4fp_0_0.vlm"},
	{0x743B038698F0AB11ULL, "1039697929.ogg"},
	{0x743B5AF155E53E25ULL, "qcr000_arishok6060098.cl"},
	{0x743B7883C3DACA55ULL, "vxlt_gallowstemplar_5cg_0_0.vlm"},
	{0x743BCE92FA5DC303ULL, "vxdr_0_deeproads_ec1_0_0.vlm"},
	{0x743C2C1D5E8EA32FULL, "6129218_m.fxe"},
	{0x743C3102E5B12C3BULL, "gen00fl_fenris6110861.cl"},
	{0x743C96BA2E3AA2CFULL, "6077275_f.fxe"},
	{0x743D33C30D36F9DEULL, "299646657.ogg"},
	{0x743D5822AB3E779DULL, "fen310_fenris6118208.cl"},
	{0x743D98EAC04DDFDDULL, "296047564.ogg"},
	{0x743E5A27336F715AULL, "6094894_m.fxe"},
	{0x743E94420CA8551DULL, "304292902.ogg"},
	{0x743EC0DB3D41597FULL, "hlac_lt_undercityencounter_-1_44y.dds"},
	{0x743EE80E8A49F144ULL, "868788461.ogg"},
	{0x743F8A8B139A2031ULL, "vxht_0_hightown_a3_b5l_0_0.vlm"},
	{0x743F8A8BAEFE30DFULL, "w_blightlands_70_0.phy"},
	{0x743FA8CFA63153B1ULL, "6169432_m.fxe"},
	{0x743FBA97C7F42044ULL, "890327838.ogg"},
	{0x7440050678C71ACEULL, "vxlt_hangedman_ta_0_0.vlm"},
	{0x744010968D13BA36ULL, "848622974.ogg"},
	{0x744035CF0EAD7989ULL, "ht_0_hightown_a3_24_0.rml"},
	{0x74408E8E53E3676BULL, "6110824_f.fxe"},
	{0x7440A7E4095316AEULL, "vxlt_gallowstemplar_n_3jr_0_0.vlm"},
	{0x7440F999CDFCEA4AULL, "hlac_w_bonepitdestroyed_a3_131076_0z.dds"},
	{0x744151C3333FFC30ULL, "w_bonepitdestroyed_a3_63_0.mao"},
	{0x744155C333400343ULL, "w_bonepitdestroyed_a3_63_0.mmh"},
	{0x744163C333401A9DULL, "w_bonepitdestroyed_a3_63_0.msh"},
	{0x74421CC8FA89BBD0ULL, "mcr310_loyalist_templars6076899.cl"},
	{0x744231F63D56A3BBULL, "vxlt_undercityencounter_13k_0_0.vlm"},
	{0x7442463C1337983CULL, "llac_lt_0_lowtown_a2_-131073_1.dds"},
	{0x74432CC3CCD6AC7AULL, "plt_prm000pt_blood_dragon_.plo"},
	{0x74443A5D7248B071ULL, "6050387_m.fxe"},
	{0x74445517700CF73BULL, "6110812_m.fxe"},
	{0x744473F009893750ULL, "1048679087.ogg"},
	{0x74450551C0D692ADULL, "vxlt_0_lowtown_docks_a3_cot_0_0.vlm"},
	{0x7445BD240EF37B44ULL, "hlac_lt_gallowscourtyard_-1_0x.dds"},
	{0x7445CF2046469DD6ULL, "ca_caverns1_24423.anb"},
	{0x7445D79EAAF70695ULL, "vxht_keep_6fh_0_0.vlm"},
	{0x74460813E291AF57ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v1.evt"},
	{0x74464EB0F765A7C1ULL, "vxht_brothel_n_6hn_0_0.vlm"},
	{0x74466826A149AB83ULL, "and121_karl6174498.cl"},
	{0x7446699803DA645AULL, "vxlt_warehouse_ot_0_0.vlm"},
	{0x7446B227517777EFULL, "vxca_caverns2_6hc_0_0.vlm"},
	{0x7446C7D5BEEE774CULL, "vxlt_0_lowtown_a2_n_ep3_0_0.vlm"},
	{0x7446E07896C6B9F1ULL, "6176483_m.fxe"},
	{0x74471DB2413E9B57ULL, "vxw_bonepitmining_a2_pm_0_0.vlm"},
	{0x744786ED4A3F4AF0ULL, "vxca_caverns1_ihx_0_0.vlm"},
	{0x74478EFBDAE442E7ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_eye_suca_0d.dds"},
	{0x74478F404092B8A0ULL, "12630819.ogg"},
	{0x7447BDEA6D5C811AULL, "vxht_mansion2b_2ns_0_0.vlm"},
	{0x74486305402F50B5ULL, "vxca_caverns1_eh5_0_0.vlm"},
	{0x744872746ECFD021ULL, "720110525.ogg"},
	{0x74499A345D82B865ULL, "307226627.ogg"},
	{0x7449D291EEC0A8E1ULL, "lt_0_lowtown_docks_a3_n_11_0.rml"},
	{0x7449DD58B7281516ULL, "vxlt_backalley_n_1dv_0_0.vlm"},
	{0x7449E78279298A5EULL, "hlac_lt_0_lowtown_a2_65538_3y.dds"},
	{0x744A0CFEF76F63A3ULL, "l_lgt110ar_gang_hideout_ht.wbd"},
	{0x744A0E782F07988BULL, "art\\vfx\\combat\\vfx_q_ari_wwind_act_c.anb"},
	{0x744A5260DF88045AULL, "hlac_lt_0_lowtown_n_131072_6z.dds"},
	{0x744A5B597059F9CBULL, "6119176_m.fxe"},
	{0x744A874BC1CA16D9ULL, "vxw_blightlands_lf_0_0.vlm"},
	{0x744A9E1B43AA4A6EULL, "689299056.ogg"},
	{0x744AF8BD666B1814ULL, "vxht_brothel_n_5is_0_0.vlm"},
	{0x744BEBCBE8ABAB54ULL, "mer221_sabin6080562.cl"},
	{0x744C80DB788110ECULL, "vxw_blightlands_6y_0_0.vlm"},
	{0x744D0913E297CFE0ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v1.gad"},
	{0x744D1B47D654F222ULL, "llac_lt_0_lowtown_-3_3.dds"},
	{0x744D38D552E759A6ULL, "hlac_lt_hangedman_n_-1_24x.dds"},
	{0x744D9A708ACD89DFULL, "gen00fl_anders6119679.cl"},
	{0x744DEB544CB39E2AULL, "18441307.ogg"},
	{0x744E49658F156251ULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyn.mmh"},
	{0x744E55658F1576F2ULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyn.mao"},
	{0x744E63658F158E0FULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyn.msh"},
	{0x744EE8C71A1278AAULL, "gen_im_cft_reg_j2.uti"},
	{0x744EFD527879ADC7ULL, "llac_lt_0_lowtown_n_131073_2.dds"},
	{0x744F88A1D729DBD5ULL, "w_bonepit_118_0.phy"},
	{0x744F8BC9A1F318AEULL, "dae000_meredith6085303.cl"},
	{0x744FB4033207C049ULL, "478789705.ogg"},
	{0x7450285F7EF6AB99ULL, "6033989_f.fxe"},
	{0x745073AE5195686AULL, "210036639.ogg"},
	{0x745080D519BB2D7FULL, "108226273.ogg"},
	{0x74508B8BAF0CA5E4ULL, "w_blightlands_70_0.mmh"},
	{0x74508F8BAF0CAC37ULL, "w_blightlands_70_0.mao"},
	{0x7450A18BAF0CCA86ULL, "w_blightlands_70_0.msh"},
	{0x7450F239A8ECE2EBULL, "788006886.ogg"},
	{0x7451CBE0F0D82AE0ULL, "llac_w_bonepit_131078_0.dds"},
	{0x745285C68C80B56EULL, "862400684.ogg"},
	{0x7452EA14D57C09C2ULL, "vxlt_gallowstemplar_n_5c3_0_0.vlm"},
	{0x74540F5D9458D10BULL, "vxht_bartrand_n_5sv_0_0.vlm"},
	{0x745423E1CFD9FCBCULL, "797696833.ogg"},
	{0x745442F488649E11ULL, "vxw_encounter2_camp_ou_0_1.vlm"},
	{0x7454C08DE95BC041ULL, "vxht_brothel_n_6um_0_0.vlm"},
	{0x745523E10354EE7BULL, "321846496.ogg"},
	{0x7455FEB46DC44B27ULL, "vxht_bartrand_n_4fp_0_0.vlm"},
	{0x74562FC75F52999AULL, "vxht_mansion2b_2kl_0_0.vlm"},
	{0x74563FFFBC3D6C9EULL, "6102111_m.fxe"},
	{0x74569F1A936515B4ULL, "124478884.ogg"},
	{0x7456F8FB1C90111DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\basestaff\\blendtree\\devastator.anb"},
	{0x74575F7A6221D70BULL, "6013403_m.fxe"},
	{0x7458258250E276FDULL, "llac_ca_caverns1_24421_9.dds"},
	{0x74582F3D4FDF965EULL, "6095822_m.fxe"},
	{0x7458FB35FA9A1CC1ULL, "vxw_sundermount_26x_0_0.vlm"},
	{0x74596365613C9D00ULL, "follower_barks6153694.cl"},
	{0x745977617E02954EULL, "varricfight.cut"},
	{0x745A35E5F3C517C5ULL, "vxht_chantry_2t7_0_1.vlm"},
	{0x745A882A9BF02FE7ULL, "vxca_caverns1_mines_e8b_0_0.vlm"},
	{0x745A8AA9A577E784ULL, "hlac_lt_0_lowtown_a2_n_65537_0z.dds"},
	{0x745B211464FE7FE2ULL, "vxlt_undercityencounter_143_0_0.vlm"},
	{0x745BAF12FF2C22E0ULL, "zz_mer_debug6062748.cl"},
	{0x745BE7C6482A720DULL, "6048272_m.fxe"},
	{0x745BFCB140242F7EULL, "dae300_cullen6140651.cl"},
	{0x745CA369DABC8DCAULL, "w_encounter3_189_0.mao"},
	{0x745CA769DABC9419ULL, "w_encounter3_189_0.mmh"},
	{0x745CB169DABCA507ULL, "w_encounter3_189_0.msh"},
	{0x745CBB9CF5C89190ULL, "gen00fl_aveline6106383.cl"},
	{0x745CC843D39E1A0BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_exit.ani"},
	{0x745D7CA157ED50AFULL, "6085466_m.fxe"},
	{0x745D9DB070413E40ULL, "vxht_0_hightown_n_bxt_0_0.vlm"},
	{0x745E48403C6CC1B8ULL, "6054290_m.fxe"},
	{0x745E5D50CFF25265ULL, "152014949.ogg"},
	{0x745E7A40D604755EULL, "fen210_orana6109718.cl"},
	{0x745EB3BC41D6A0BAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5_tool.mmh"},
	{0x745EB7BC41D6A7F1ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5_tool.mao"},
	{0x745EC5BC41D6BF00ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5_tool.msh"},
	{0x745ED90DC8C92BECULL, "6136222_m.fxe"},
	{0x745F294425998FFAULL, "dae000_katriela.cnv"},
	{0x745F2FC0B592ECBAULL, "6207935_m.fxe"},
	{0x745FC109B146DF78ULL, "vxht_0_hightown_avn_0_0.vlm"},
	{0x745FCA223F35528DULL, "vxlt_hangedman_1i4_0_0.vlm"},
	{0x74607D7E0B91FE36ULL, "mag291cr_crowd_elf_2.utc"},
	{0x7460A8C284DBBFD9ULL, "419972881.ogg"},
	{0x74615BA97F88A23FULL, "art\\vfx\\combat\\vfx_gib_cora_2.mao"},
	{0x7461A575E8F63340ULL, "vxlt_gallowscourtyard_5yf_0_0.vlm"},
	{0x7461B99FBF7E10AFULL, "gen00fl_merrill6128970.cl"},
	{0x746221C64C4A4D51ULL, "gen00fl_anders6103739.cl"},
	{0x74624B44259BFF42ULL, "dae000_katriela.bnk"},
	{0x7462BB588EE33795ULL, "vxht_keep_6lv_0_0.vlm"},
	{0x74642C6E555EFA98ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_lrgchair.phy"},
	{0x74645FEDD612CD41ULL, "genip_node_a1.utp"},
	{0x746462FBE5042F65ULL, "dae000ip_crowd_ff7_talk9.utp"},
	{0x746470915FBA63A7ULL, "vxht_bartrand_n_4ks_0_0.vlm"},
	{0x7464C78921EA0F6DULL, "dre000_dougal6205761.cl"},
	{0x7465D5CD4C5ABC3FULL, "6115451_m.fxe"},
	{0x7465F20E7DD20E12ULL, "6118967_m.fxe"},
	{0x7465F60FF00D2A1AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_hightown_fabric_d.dds"},
	{0x746605E3D9341F8CULL, "613547505.ogg"},
	{0x746613CD0A0827E4ULL, "vxlt_0_lowtown_a2_n_c_hvv_0_0.vlm"},
	{0x7466582BCDA61557ULL, "dae210ar_lowtown_nt.ncs"},
	{0x7466C1129B144616ULL, "vxlt_backalley_n_1jx_0_0.vlm"},
	{0x7466CE1A99BD13D5ULL, "vxlt_gallowstemplar_5rd_0_0.vlm"},
	{0x7467573535514E19ULL, "vxlt_0_lowtown_a2_n_ffg_0_0.vlm"},
	{0x74675BCDEA8DF330ULL, "w_bonepit_199_0.mao"},
	{0x74675FCDEA8DFA43ULL, "w_bonepit_199_0.mmh"},
	{0x74676B05A5B647D9ULL, "vxw_blightlands_zd_0_0.vlm"},
	{0x74676DCDEA8E119DULL, "w_bonepit_199_0.msh"},
	{0x7467D50F6A8397B1ULL, "216857723.ogg"},
	{0x7467EF4A635A2D2BULL, "195656324.ogg"},
	{0x7467FD3480F74D4CULL, "art\\characters\\playercharacter\\humanmale\\arm_dksa.msh"},
	{0x7468033480F7573EULL, "art\\characters\\playercharacter\\humanmale\\arm_dksa.mmh"},
	{0x74680F3480F76BD5ULL, "art\\characters\\playercharacter\\humanmale\\arm_dksa.mao"},
	{0x74682DA2D6AE8AF6ULL, "vxht_mansion2b_1h6_0_0.vlm"},
	{0x74686E6F2A54EF2DULL, "vxlt_gamlenshouse_n_no_0_0.vlm"},
	{0x7469176686651592ULL, "6031143_m.fxe"},
	{0x7469AA9B74BDB685ULL, "hlac_lt_0_lowtown_docks_a3_131070_0z.dds"},
	{0x7469E441B741AC07ULL, "vxht_mansion2_25k_0_0.vlm"},
	{0x746A000A73FB9416ULL, "6094818_m.fxe"},
	{0x746A03A0D1F9A30FULL, "vxht_mansion2_2ec_0_0.vlm"},
	{0x746A82B08B6BD94EULL, "llac_ht_0_hightown_n_-196610_0.dds"},
	{0x746ACF561B21BFAFULL, "6207744_f.fxe"},
	{0x746BAA02E4F2E4CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_exit.evt"},
	{0x746BC8AA926F16DAULL, "6083366_m.fxe"},
	{0x746BFAAA2ADB0971ULL, "hlac_ca_caverns2prc_5802_59y.dds"},
	{0x746C0F8D623756C0ULL, "vxht_0_hightown_n_bcu_0_0.vlm"},
	{0x746C7EAD1424D980ULL, "6027659_f.fxe"},
	{0x746CE06CA20EED75ULL, "isa321_velasco6117623.cl"},
	{0x746D5160426DD35DULL, "6000815_f.fxe"},
	{0x746D582BCDAC33A6ULL, "dae210ar_lowtown_nt.lst"},
	{0x746D804A534A247EULL, "vxca_caverns1_ee1_0_0.vlm"},
	{0x746DC0E2E6E0BE77ULL, "framebuffereffects_ps3zcullreload_p0.psh.fxo"},
	{0x746DE4C39C0F7039ULL, "6170975_m.fxe"},
	{0x746DF710635A5B15ULL, "mcr370_meredith6085968.cl"},
	{0x746EBC39194B924CULL, "514643074.ogg"},
	{0x746EE3FE837BF217ULL, "6042298_m.fxe"},
	{0x746F49C6B7E551B5ULL, "vxw_bonepit_164_0_0.vlm"},
	{0x746F52D7A6B01C1FULL, "888856686.ogg"},
	{0x746FFEFEF7901B51ULL, "l_lgt110ar_gang_hideout_ht.bnk"},
	{0x74700DE391CA6A60ULL, "dae000_madam_lusine6154930.cl"},
	{0x747065F6A68CB107ULL, "dre180ar_deep_roads.ncs"},
	{0x7470B94DC2D33442ULL, "vxdr_0_deeproads_d3u_0_0.vlm"},
	{0x7470EB73649E3DA9ULL, "684816831.ogg"},
	{0x7471345D7CC9031CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_enter.ani"},
	{0x74716D93B5D4D50DULL, "778207686.ogg"},
	{0x74716E0D520448BEULL, "art\\vfx\\creatures\\fxc_arc_shd_c.impa.evt"},
	{0x74716E5B8EBAD168ULL, "lgt160_wk_soph_sophia6119831.cl"},
	{0x747183EFA11188A4ULL, "126664481.ogg"},
	{0x747187BC067D80A5ULL, "150892228.ogg"},
	{0x7471AF43D3B04B95ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_exit.gad"},
	{0x7471E31EA3D0355BULL, "vxlt_0_lowtown_a2_n_c_cms_0_0.vlm"},
	{0x747207BE2759C16FULL, "vxca_caverns2_6yl_0_0.vlm"},
	{0x74723257065AD281ULL, "gen00fl_anders6111343.cl"},
	{0x7472496E1887C1ECULL, "art\\characters\\playercharacter\\elffemale\\rob_appa.phy"},
	{0x747273491720E4D7ULL, "vxw_bonepitmining_a2_an_0_0.vlm"},
	{0x7472C302E4F92D8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_exit.gad"},
	{0x7472D2534BB208DFULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic.msh"},
	{0x7472D8534BB21281ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic.mmh"},
	{0x7472E4534BB22722ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic.mao"},
	{0x74731381433ECA9AULL, "vxht_mansion2b_2an_0_0.vlm"},
	{0x74735A87A506BA18ULL, "hlac_ht_0_hightown_n_1_1x.dds"},
	{0x74737D76C9F4B99EULL, "mag110_vincento6036917.cl"},
	{0x7473CD79FE9A23B0ULL, "fen110_fenris6100257.cl"},
	{0x74742ECF97776523ULL, "vxlt_0_lowtown_docks_b4x_0_1.vlm"},
	{0x747485A9FBFE4064ULL, "vxlt_0_lowtown_a2_n_c_hqw_0_0.vlm"},
	{0x74757A6C2A1EE7CAULL, "follower_banter6165629.cl"},
	{0x74758AC9A035A970ULL, "lt_0_lowtown_docks_a3_195_0.msh"},
	{0x747598C9A035C08AULL, "lt_0_lowtown_docks_a3_195_0.mmh"},
	{0x74759CC9A035C7C1ULL, "lt_0_lowtown_docks_a3_195_0.mao"},
	{0x7475BECCE575C909ULL, "gen_im_arm_boo_hvy_act2_01.uti"},
	{0x7475EB90C69490D2ULL, "hlac_lt_0_lowtown_docks_a3_-65538_5z.dds"},
	{0x747601D5CE8C6DAFULL, "6165481_m.fxe"},
	{0x74769F7FC8A4A376ULL, "vxht_mansion2b_1c7_0_0.vlm"},
	{0x74773D01B30409B4ULL, "vxht_0_hightown_n_9zu_0_0.vlm"},
	{0x747745F6A69299B6ULL, "dre180ar_deep_roads.lst"},
	{0x74774FE25D71AC52ULL, "557325410.ogg"},
	{0x74775507C461DCE8ULL, "vxlt_hangedman_n_5q_0_0.vlm"},
	{0x747755EC6A2640E5ULL, "vxht_keep_6f8_0_0.vlm"},
	{0x7477672BCDB4A1B1ULL, "dae210ar_lowtown_nt.are"},
	{0x747799D6F0EB551EULL, "vxlt_0_lowtown_docks_n_arr_0_0.vlm"},
	{0x7477A75E6004F50FULL, "gen00fl_isabela6135701.cl"},
	{0x7478415D7CCF385EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_enter.gad"},
	{0x747860CDEA9C6F68ULL, "w_bonepit_199_0.phy"},
	{0x7478850D520A8EE9ULL, "art\\vfx\\creatures\\fxc_arc_shd_c.impa.gad"},
	{0x7478B2204671C66AULL, "ca_caverns1_24423.rml"},
	{0x7478C202E4FD978AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_exit.ani"},
	{0x7478C6A8DABE2993ULL, "plt_lgt310pt_reining.plo"},
	{0x74791FD63C816EF0ULL, "vxht_0_hightown_n_bb6_0_0.vlm"},
	{0x7479773BF3DB417CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishok_pcoverride.anb"},
	{0x74798F84901CDDB9ULL, "6118234_m.fxe"},
	{0x7479EAD066BB3902ULL, "6174860_m.fxe"},
	{0x747A3ABD973EFB70ULL, "vxlt_gallowscourtyard_66p_0_0.vlm"},
	{0x747A9E77D53964B5ULL, "vxw_woundedcoast_79_0_0.vlm"},
	{0x747ABBC20FBC9D13ULL, "vxht_bartrand_n_47j_0_1.vlm"},
	{0x747BF98EB45FBB08ULL, "6152527_m.fxe"},
	{0x747C4461F47B98D2ULL, "158825009.ogg"},
	{0x747C6E8BB64E999AULL, "403571958.ogg"},
	{0x747C782364671B59ULL, "gen00fl_isabela6106406.cl"},
	{0x747CB191EEEBCA71ULL, "lt_0_lowtown_docks_a3_n_11_0.anb"},
	{0x747D0BB808F2A142ULL, "hlac_lt_gallowstemplar_n_2384_4z.dds"},
	{0x747D740925CF1F2EULL, "llac_w_encounter2_camp_393214_0.dds"},
	{0x747D7A1273960783ULL, "vxht_chantry_36k_0_0.vlm"},
	{0x747DDD7ED2EDA613ULL, "339329826.ogg"},
	{0x747E3B6C5E513D54ULL, "435219758.ogg"},
	{0x747E4381B5488879ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_nud_0t.dds"},
	{0x747F17C59871F4FDULL, "vxlt_warehouse_n_15o_0_0.vlm"},
	{0x747F1C5D7CD5187DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_enter.evt"},
	{0x747F560D52105D37ULL, "art\\vfx\\creatures\\fxc_arc_shd_c.impa.ani"},
	{0x747F66FB283E4588ULL, "hlac_w_blightlands_196611_1z.dds"},
	{0x74805367AEBFD599ULL, "6161090_m.fxe"},
	{0x74805CF0EA79D733ULL, "vxlt_hangedman_sa_0_0.vlm"},
	{0x7480E5260916FD57ULL, "vxw_bonepitmining_a2_li_0_0.vlm"},
	{0x7480FE9E67AF741CULL, "art\\vfx\\creatures\\fxc_footstep_l_c.impa.ani"},
	{0x74812024A7608146ULL, "hlac_dr_0_deeproads_5842_1x.dds"},
	{0x748142507EFE5E25ULL, "zz_followers_banter6159405.cl"},
	{0x74814601967AD01BULL, "vxw_bonepitdestroyed_a3_1wh_0_1.vlm"},
	{0x748154F6A69B07A1ULL, "dre180ar_deep_roads.are"},
	{0x74816AA5B505855FULL, "604229352.ogg"},
	{0x74817A677DF58E08ULL, "vxlt_darktown_4hf_0_0.vlm"},
	{0x7481B16936F17F15ULL, "308934396.ogg"},
	{0x7481F72E73C32439ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef03.msh"},
	{0x7481FD5ED38C6F3AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha10a.mmh"},
	{0x7482012E73C33527ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef03.mmh"},
	{0x74820F5ED38C8D80ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha10a.msh"},
	{0x748254C2170BD234ULL, "art\\characters\\playercharacter\\humanmale\\cth_inv.phy"},
	{0x74830664AAB346D7ULL, "6105700_m.fxe"},
	{0x7483B1D47DA944D5ULL, "vxlt_gallowstemplar_5xb_0_0.vlm"},
	{0x7483D0A92E30D5FAULL, "seb321_leliana6181751.cl"},
	{0x7483D2BC41F5F1C5ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5_tool.phy"},
	{0x7483E54493428B6BULL, "mag170_hugh6043093.cl"},
	{0x7484065A3970D4A4ULL, "art\\vfx\\materials\\vfx_lightning_arc_fb.mao"},
	{0x74845B88F7590180ULL, "6030645_m.fxe"},
	{0x748484B0AD89977BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02b.phy"},
	{0x7485AB5E42EBE224ULL, "hlac_w_encounter3_0_12z.dds"},
	{0x7486389BD019A488ULL, "6088417_m.fxe"},
	{0x748660483E097023ULL, "vxw_sundermount_2w1_0_0.vlm"},
	{0x7486B17C7D69392CULL, "llac_ht_mansion2b_818_33.dds"},
	{0x7486CF8AC71FD5A8ULL, "rog_duelistdefensebonus.dds"},
	{0x748767F5B17F8483ULL, "6026960_m.fxe"},
	{0x748791B32E778770ULL, "vxht_0_hightown_n_bm7_0_0.vlm"},
	{0x7487CFCF5FC4E455ULL, "gen00fl_aveline6114079.cl"},
	{0x748802B990BCB9AFULL, "cd_and111_donation_box.wbd"},
	{0x7488042A61C52D59ULL, "313598448.ogg"},
	{0x74880B9E67B5A95EULL, "art\\vfx\\creatures\\fxc_footstep_l_c.impa.gad"},
	{0x7488116016AAEFAEULL, "mer320_hubert6090368.cl"},
	{0x748849394D75E52EULL, "vxht_chantry_n_1yf_0_0.vlm"},
	{0x7488C40160606062ULL, "359996362.ogg"},
	{0x7488D1E2CD4EB884ULL, "vxht_mansion2_3ey_0_0.vlm"},
	{0x7488DB95FBFD8B89ULL, "302209203.ogg"},
	{0x748943BFBAF97654ULL, "6206933_f.fxe"},
	{0x748A4CE3338DEB4FULL, "hlac_ht_0_hightown_a3_65538_0x.dds"},
	{0x748A8B7EB2476AACULL, "442077022.ogg"},
	{0x748AEAFB1AA12699ULL, "327476169.ogg"},
	{0x748B38F6290767A7ULL, "lla_ht_0_hightown_a2_-2_0.dds"},
	{0x748BE5DCE387AB1DULL, "6059343_f.fxe"},
	{0x748C0E086FAEF299ULL, "6153693_m.fxe"},
	{0x748C31C21713EB31ULL, "art\\characters\\playercharacter\\humanmale\\cth_inv.msh"},
	{0x748C3A6332610E02ULL, "6181828_m.fxe"},
	{0x748C3FC21714034CULL, "art\\characters\\playercharacter\\humanmale\\cth_inv.mao"},
	{0x748C466E0523D602ULL, "vxlt_0_lowtown_n_ewp_0_0.vlm"},
	{0x748C4BC2171417EFULL, "art\\characters\\playercharacter\\humanmale\\cth_inv.mmh"},
	{0x748CBB4321ECFBDAULL, "face_face_p7.psh.fxo.dx11"},
	{0x748CEB0C7DF597C1ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_crew_0n.dds"},
	{0x748D06B1DED4E5F8ULL, "character_character_p45.psh.fxo.dx11.assembly"},
	{0x748D49085C854C93ULL, "and_templar_4.utc"},
	{0x748D7C026F289602ULL, "one220_yevhen6098700.cl"},
	{0x748E189B2DAE9105ULL, "vxht_0_hightown_avw_0_0.vlm"},
	{0x748E30A96DCB18DFULL, "gen_im_wep_rog_arc_act2_06.uti"},
	{0x748E3A96CF24C73BULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_aoe_p.impa.ani"},
	{0x748E8CA32057F01EULL, "vxlt_0_lowtown_a2_n_ekl_0_0.vlm"},
	{0x748EABA896EEA5E8ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtc.mmh"},
	{0x748EB7A896EEB98BULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtc.mao"},
	{0x748EC544A0B5EE19ULL, "6166190_m.fxe"},
	{0x748EC9A896EED812ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtc.msh"},
	{0x748EE69E67BB897DULL, "art\\vfx\\creatures\\fxc_footstep_l_c.impa.evt"},
	{0x748F33FA5440F2CCULL, "6027757_m.fxe"},
	{0x748F3AF1143E0F9EULL, "427764945.ogg"},
	{0x748F83AE18F2816CULL, "vxlt_0_lowtown_a2_el9_0_1.vlm"},
	{0x748F855EEA086F4CULL, "6108674_m.fxe"},
	{0x748FBDE7A900784EULL, "hlac_lt_0_lowtown_-131072_6x.dds"},
	{0x748FD4AADB87DC0EULL, "lt_0_lowtown_a2_n_27_0_ll.mmh"},
	{0x7490264A7BA45DDDULL, "zz_amb_chair.utp"},
	{0x7490B2C6D6C90BA0ULL, "rdr111ar_warehouse.lst"},
	{0x749102BE981724C8ULL, "heroicpartystats.gda"},
	{0x749199B64011CC06ULL, "6113965_f.fxe"},
	{0x74925040C0F1AC9BULL, "gen00fl_bethany6097892.cl"},
	{0x74925DE9521A93DAULL, "6108904_m.fxe"},
	{0x74929EBA7527E377ULL, "w_sundermount_164_0.phy"},
	{0x7492C7DBC41C2BD1ULL, "mag311_emile_de_launcet6043710.cl"},
	{0x7492D7DCA946744CULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_lava.mao"},
	{0x749341E6A0DE462BULL, "gen00fl_anders6119235.cl"},
	{0x74938339AC90D476ULL, "vxht_mansion2b_1y9_0_0.vlm"},
	{0x7493A827E7FF286DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgecrgrck.msh"},
	{0x7493B627E7FF3FA0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgecrgrck.mao"},
	{0x7493BA27E7FF46F3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgecrgrck.mmh"},
	{0x749404C0015CC2FAULL, "w_woundedcoast_183_0.phy"},
	{0x74940EED0E49BAD4ULL, "lgt110_arl_bann_teagan6115562.cl"},
	{0x74940F9DA0D5545EULL, "6085434_m.fxe"},
	{0x7494606C09B3B425ULL, "967924457.ogg"},
	{0x74946BE8323517ACULL, "655235464.ogg"},
	{0x7494AA78C20FBD08ULL, "6083517_m.fxe"},
	{0x7494C4D52D4A6D2BULL, "6091994_f.fxe"},
	{0x7494E22BC0F79226ULL, "116179368.ogg"},
	{0x7494F9237F2B31D5ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_spider.anb"},
	{0x7495D47C71A9A4DBULL, "vxw_bonepit_179_0_0.vlm"},
	{0x74960155A52B46D4ULL, "llac_lt_undercityserialkiller_-1_19.dds"},
	{0x74964721358FE383ULL, "hlac_lt_0_lowtown_65534_6y.dds"},
	{0x74964AD797417F3EULL, "6062785_m.fxe"},
	{0x74969BA1D765FEF0ULL, "w_bonepit_118_0.msh"},
	{0x7496A9A1D766160AULL, "w_bonepit_118_0.mmh"},
	{0x7496ADA1D7661D41ULL, "w_bonepit_118_0.mao"},
	{0x7497212864962E89ULL, "vxlt_undercityserialkiller_1io_0_0.vlm"},
	{0x7497B2C6D6CF2AF1ULL, "rdr111ar_warehouse.ncs"},
	{0x7497B650D3FAB0E3ULL, "t3_bow_yew.tnt"},
	{0x7497BAF8F6F1EF72ULL, "lla_ht_0_hightown_a2_-65536_0.dds"},
	{0x7497D71BDA376222ULL, "6201487_m.fxe"},
	{0x74983550AA79887CULL, "hlac_lt_0_lowtown_a2_131073_0x.dds"},
	{0x74984C2B6FF802D8ULL, "6172868_m.fxe"},
	{0x74986ED5FEDC584DULL, "hlac_ht_0_hightown_a2_65533_2z.dds"},
	{0x7499978A824EDEB5ULL, "gen_im_wep_rog_dua_act3_05.uti"},
	{0x7499AC772FFFFDFCULL, "vxlt_darktown_5cf_0_0.vlm"},
	{0x7499B42C19B6A625ULL, "932455786.ogg"},
	{0x7499EBBA455F92CEULL, "hlac_w_sundermount_327685_0z.dds"},
	{0x749A9DFB7C31BE02ULL, "6142771_m.fxe"},
	{0x749AA9C6D6D1516BULL, "rdr111ar_warehouse.are"},
	{0x749B4296CF2F5F32ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_aoe_p.impa.evt"},
	{0x749BB7FFF63C857CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackr.mao"},
	{0x749BC3FFF63C991FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackr.mmh"},
	{0x749BC9FFF63CA3C1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackr.msh"},
	{0x749C716BB02371A1ULL, "vxca_caverns1_mines_i8x_0_0.vlm"},
	{0x749C7DCD07ECDBDAULL, "hf_phylias.mor"},
	{0x749C83C2148D2DF7ULL, "gen00fl_anders6125269.cl"},
	{0x749CAEB4E10C9949ULL, "w_blightlands_25_0.rml"},
	{0x749CF48FCB475F87ULL, "dae200_bran6144778.cl"},
	{0x749E3AB9146F26D0ULL, "vxw_bonepit_y3_0_0.vlm"},
	{0x749EA03C4CBEAA4DULL, "6149548_m.fxe"},
	{0x749ED9534BD74FFEULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic.phy"},
	{0x749F0FC903AECBBFULL, "352500532.ogg"},
	{0x749F39E7373B77EEULL, "gen_im_arm_cht_elf_comc_t5.uti"},
	{0x749F3EF49866AF9AULL, "155999149.ogg"},
	{0x749FD67CC73CC76AULL, "6150322_f.fxe"},
	{0x749FF1188C8D43BBULL, "6042153_m.fxe"},
	{0x74A00D9E19132373ULL, "vxw_bonepitdestroyed_a3_276_0_0.vlm"},
	{0x74A0C48F60E264EFULL, "ser120_ghyslain6118402.cl"},
	{0x74A161080197CF9EULL, "6188171_m.fxe"},
	{0x74A18E66EDA80CFEULL, "vxlt_0_lowtown_docks_bgl_0_1.vlm"},
	{0x74A1B38541188970ULL, "hlac_lt_0_lowtown_docks_a3_-196610_2y.dds"},
	{0x74A20427E78DCF23ULL, "fen210_fenris6109421.cl"},
	{0x74A26196CF35B2C5ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_aoe_p.impa.gad"},
	{0x74A313DF43CE38E0ULL, "vxdr_primevaltomb_75h_0_0.vlm"},
	{0x74A336B21F4085ABULL, "6096894_f.fxe"},
	{0x74A35C770CFFA33FULL, "sib100_father_contact6108951.cl"},
	{0x74A397DA3682A287ULL, "gen00fl_fenris6116557.cl"},
	{0x74A39B6357DE194EULL, "lgt200cr_gang_1_docks.utc"},
	{0x74A3A6C366DE583AULL, "113133253.ogg"},
	{0x74A430FAF28F80EAULL, "gen_im_trs_gen_hvy_cht_r2.uti"},
	{0x74A44F8CCD038690ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_4rockpile.msh"},
	{0x74A45D8CCD039EAAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_4rockpile.mmh"},
	{0x74A4618CCD03A5E1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_4rockpile.mao"},
	{0x74A4BCB4893797BEULL, "6069685_m.fxe"},
	{0x74A67C04F5E2C61AULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.dura.gad"},
	{0x74A6A2B4E114D929ULL, "w_blightlands_25_0.mao"},
	{0x74A6AEB4E114EDC2ULL, "w_blightlands_25_0.mmh"},
	{0x74A6B0B4E114F178ULL, "w_blightlands_25_0.msh"},
	{0x74A6B8FC3BE95A20ULL, "hlac_ht_mansion2b_818_32y.dds"},
	{0x74A6BE0861EE1B58ULL, "6175968_m.fxe"},
	{0x74A6C2BC101B0C9EULL, "6085962_m.fxe"},
	{0x74A71453178E4C98ULL, "6201243_m.fxe"},
	{0x74A71C5ED3ABC045ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha10a.phy"},
	{0x74A788B17E8BD5CCULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_escape.ani"},
	{0x74A7904DF2278E79ULL, "vxht_mansion2_2c3_0_0.vlm"},
	{0x74A7E5F4838E882CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_loop.ani"},
	{0x74A80BDFB74FE471ULL, "6110739_m.fxe"},
	{0x74A811F9A66EA772ULL, "vxht_0_hightown_c8x_0_0.vlm"},
	{0x74A85DBA565EFF12ULL, "417920766.ogg"},
	{0x74A86710556350B1ULL, "dae000cr_templar_recruit.utc"},
	{0x74A8A9377F499560ULL, "6082165_m.fxe"},
	{0x74A8ACF4DACC356EULL, "vxht_mansion2_1vh_0_0.vlm"},
	{0x74A8ADA8DAE72330ULL, "plt_lgt310pt_reining.nss"},
	{0x74A8B65B2AEF8544ULL, "712170002.ogg"},
	{0x74A8CE051C94E2C4ULL, "str_link_smo.mao"},
	{0x74A8DA34812EA8D9ULL, "art\\characters\\playercharacter\\humanmale\\arm_dksa.phy"},
	{0x74A92A27E9100702ULL, "vxlt_0_lowtown_n_emr_0_0.vlm"},
	{0x74A975C7A5FB062CULL, "gen00fl_sebastian_act1.stg"},
	{0x74A9DD3D98AC944AULL, "6156482_m.fxe"},
	{0x74A9DF28E5FE22FDULL, "567441684.ogg"},
	{0x74A9EA842509AD20ULL, "hla_ht_0_hightown_n_65536_0x.dds"},
	{0x74A9EC3309774414ULL, "one111_thug.cnv"},
	{0x74AA28D3D0697BDDULL, "955389321.ogg"},
	{0x74AA3E6BF5C97A77ULL, "art\\vfx\\creatures\\fxc_gol_proj_c.anb"},
	{0x74AABB8D66775195ULL, "vxlt_0_lowtown_ep4_0_1.vlm"},
	{0x74AAD671CBCB07AEULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb02a.phy"},
	{0x74AB705D0444211EULL, "vxlt_0_lowtown_a2_n_eqn_0_0.vlm"},
	{0x74ABB798313B4DFAULL, "w_woundedcoast_46_0.rml"},
	{0x74ABD6E4975CB435ULL, "vxlt_backalley_g_1jr_0_0.vlm"},
	{0x74AC1CACEF6AA2E4ULL, "vxlt_warehouse_159_0_0.vlm"},
	{0x74AC415F70FE1AE3ULL, "gen00fl_bethany6006076.cl"},
	{0x74AC56C0DA73FFA1ULL, "756442221.ogg"},
	{0x74AC9565A3E74A17ULL, "vxw_woundedcoast_1kv_0_0.vlm"},
	{0x74ACB85E76722430ULL, "107558264.ogg"},
	{0x74AD51F155A148DDULL, "6100763_m.fxe"},
	{0x74AD61E41F8C04B7ULL, "6054200_m.fxe"},
	{0x74AD7F04F5E8EA31ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.dura.evt"},
	{0x74AD9367C30C22DEULL, "208657648.ogg"},
	{0x74ADCE33097AFA54ULL, "one111_thug.bnk"},
	{0x74ADEA2E73E8495CULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef03.phy"},
	{0x74AE03060D829863ULL, "llac_lt_0_lowtown_docks_a3_196607_3.dds"},
	{0x74AE2F78AA1F40D6ULL, "follower_banter6141744.cl"},
	{0x74AE8506C5497587ULL, "352184972.ogg"},
	{0x74AE8FF3386F2BDDULL, "6083314_m.fxe"},
	{0x74AE95B17E920A0EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_escape.gad"},
	{0x74AE9CDD6828C78EULL, "ave200_donnic6112170.cl"},
	{0x74AEB1DE6DD536DAULL, "vxlt_blackemporium_13u_0_0.vlm"},
	{0x74AEB8EFEBEB311BULL, "324572390.ogg"},
	{0x74AEC76D07034758ULL, "fxa_web_01.mao"},
	{0x74AEE751E2E5EC53ULL, "art\\characters\\creatures\\dragon\\animation\\rdy_crnf.evt"},
	{0x74AEF2F48394BD6EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_loop.gad"},
	{0x74B0489E9DF012ADULL, "dae000_meredith6092579.cl"},
	{0x74B060A51548E186ULL, "lt_warehouse_mmi.gff"},
	{0x74B06BB0ADAEA71AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02b.msh"},
	{0x74B07DB0ADAEC660ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02b.mmh"},
	{0x74B0805B6E4977CBULL, "vxw_woundedcoast_1bi_0_0.vlm"},
	{0x74B0E2190C3BDF14ULL, "347842271.ogg"},
	{0x74B0F4D257A391FEULL, "vxlt_0_lowtown_docks_a3_cuu_0_0.vlm"},
	{0x74B1482A7D74D5A9ULL, "dae010cr_sailor_f.utc"},
	{0x74B1DC00D8BDB91EULL, "6054850_m.fxe"},
	{0x74B2547980D93116ULL, "vxlt_0_lowtown_docks_a3_as1_0_0.vlm"},
	{0x74B2830FC0064EEDULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2groupcrag.mmh"},
	{0x74B2870FC006553EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2groupcrag.mao"},
	{0x74B2950FC0066D73ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_2groupcrag.msh"},
	{0x74B2B9F1B0D7F4F4ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\blendtree\\trooper.anb"},
	{0x74B2BAA6E6553548ULL, "hlac_lt_gallowscourtyard_n_65536_1y.dds"},
	{0x74B2E5037C0D1356ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.dura.evt"},
	{0x74B3206E18BF12B7ULL, "art\\characters\\playercharacter\\elffemale\\rob_appa.mmh"},
	{0x74B3366E18BF3829ULL, "art\\characters\\playercharacter\\elffemale\\rob_appa.msh"},
	{0x74B33E4F8DD0757EULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls103a_0n.dds"},
	{0x74B356A355EBBDECULL, "hlac_ca_cavern1_haunted_24426_1y.dds"},
	{0x74B38841CA3EC88FULL, "14301181.ogg"},
	{0x74B490B17E966DEDULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_escape.evt"},
	{0x74B58C1037108944ULL, "66627223.ogg"},
	{0x74B5C0DDCD0767BBULL, "vxlt_undercityencounter_1kc_0_0.vlm"},
	{0x74B5E051E2EBFE94ULL, "art\\characters\\creatures\\dragon\\animation\\rdy_crnf.gad"},
	{0x74B6447FB63B7AC0ULL, "gen00fl_aveline6137516.cl"},
	{0x74B6522A54B5215EULL, "vxw_bonepit_26l_0_0.vlm"},
	{0x74B6756E22B8D449ULL, "33005693.ogg"},
	{0x74B6DADDB3BF15D4ULL, "hlac_w_encounter3_0_1z.dds"},
	{0x74B71ED1CCC24DEEULL, "vxht_mansion2_1yi_0_0.vlm"},
	{0x74B75C99D4C422E3ULL, "vxht_mansion2b_2jb_0_0.vlm"},
	{0x74B7FECE592C8749ULL, "6179011_m.fxe"},
	{0x74B85D84D7FD446EULL, "6081382_m.fxe"},
	{0x74B8AD6338962223ULL, "569441965.ogg"},
	{0x74B8CE498BAFD96EULL, "hla_ht_0_hightown_2_0x.dds"},
	{0x74B8E6D02E82F9E3ULL, "6172694_m.fxe"},
	{0x74B92E66DBD55035ULL, "6161043_m.fxe"},
	{0x74B9CA0DDF05A3B2ULL, "929685751.ogg"},
	{0x74B9DC037C1322E1ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.dura.gad"},
	{0x74B9DC112323D51DULL, "pro000cs_loth_dest.cut"},
	{0x74BAB71E0B5AA909ULL, "6126925_m.fxe"},
	{0x74BACB7BB9EFCB7FULL, "391148261.ogg"},
	{0x74BB1E72F85B57D0ULL, "vxw_bonepit_o1_0_0.vlm"},
	{0x74BB9704F5F55000ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.dura.ani"},
	{0x74BB9E610A6F71C0ULL, "vxw_encounter3_7b_0_0.vlm"},
	{0x74BBB33F12D3B30CULL, "qun120_arishok_guard6054851.cl"},
	{0x74BBC3CE4797615DULL, "6105463_m.fxe"},
	{0x74BCC35F5BDE5243ULL, "dae340_db_drag_warr.cut"},
	{0x74BCEF51E2F23772ULL, "art\\characters\\creatures\\dragon\\animation\\rdy_crnf.ani"},
	{0x74BD918772517513ULL, "566120726.ogg"},
	{0x74BDD4C1BFE4DBB6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkb_d.dds"},
	{0x74BDF1CB68BAC36FULL, "vxlt_undercityencounter_zg_0_0.vlm"},
	{0x74BE89E327399C51ULL, "lt_0_lowtown_docks_168_0.mmh"},
	{0x74BE95E32739B0F2ULL, "lt_0_lowtown_docks_168_0.mao"},
	{0x74BEA3E32739C80FULL, "lt_0_lowtown_docks_168_0.msh"},
	{0x74BEF238603F904BULL, "vxw_woundedcoast_1mh_0_0.vlm"},
	{0x74BF2304AD3D6C3FULL, "968401766.ogg"},
	{0x74BFA068C5834A3DULL, "6160804_m.fxe"},
	{0x74BFBC66DDEDA06CULL, "vxw_encounter2_4l_0_0.vlm"},
	{0x74BFBCA2A0E26EFEULL, "vxw_bonepit_145_0_0.vlm"},
	{0x74BFD85C0C7BA6BBULL, "vxlt_0_lowtown_docks_a3_n_bt1_0_0.vlm"},
	{0x74BFDD037C178F8FULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.dura.ani"},
	{0x74C0196E55AD392DULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_lrgchair.msh"},
	{0x74C0276E55AD5160ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_lrgchair.mao"},
	{0x74C02B6E55AD57B3ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_lrgchair.mmh"},
	{0x74C0C7424E7B1B98ULL, "6119284_f.fxe"},
	{0x74C119462A52F429ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_simple.mao"},
	{0x74C125462A5308C2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_simple.mmh"},
	{0x74C127462A530C78ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_simple.msh"},
	{0x74C1494EA0D94228ULL, "236313790.ogg"},
	{0x74C1693AAB8CD8C3ULL, "6122116_m.fxe"},
	{0x74C1B906DA08952AULL, "vxht_brothel_n_60q_0_0.vlm"},
	{0x74C223304AB37270ULL, "t1_tat_dgn.tnt"},
	{0x74C25AB156C33ED0ULL, "vxlt_0_lowtown_n_f17_0_0.vlm"},
	{0x74C276BF3A787809ULL, "vxlt_undercityserialkiller_1xl_0_0.vlm"},
	{0x74C2B4AF89FB893CULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_fenris_fenrintro.gad"},
	{0x74C3290582C416AFULL, "one100dg_nabil6033169.cl"},
	{0x74C33489211639CCULL, "vxlt_0_lowtown_docks_n_bvk_0_0.vlm"},
	{0x74C3765368628921ULL, "hlac_lt_gallowstemplar_n_2382_2y.dds"},
	{0x74C3EFBD5F2AEEDBULL, "vxlt_backalley_g_1nx_0_0.vlm"},
	{0x74C46A832D469D21ULL, "vxht_mansion2b_3bm_0_0.vlm"},
	{0x74C4C40746AB39DEULL, "vxw_bonepit_21m_0_0.vlm"},
	{0x74C4F5B38A5AD872ULL, "vxht_0_hightown_c2z_0_0.vlm"},
	{0x74C4FCADCD9A7A5CULL, "6083665_m.fxe"},
	{0x74C5020E05E2477CULL, "vxlt_darktown_5re_0_0.vlm"},
	{0x74C571D8B610335EULL, "10357622.ogg"},
	{0x74C59DC587C5B6DFULL, "732421082.ogg"},
	{0x74C5BE2129237DDEULL, "6179980_m.fxe"},
	{0x74C5C086302DB110ULL, "vxht_0_hightown_a3_c8o_0_0.vlm"},
	{0x74C5CE76C6BA3B63ULL, "vxht_mansion2b_2ge_0_0.vlm"},
	{0x74C6D9328AEE1B88ULL, "hlac_ca_cavern1_haunted_24424_2z.dds"},
	{0x74C758AD20791263ULL, "6175794_m.fxe"},
	{0x74C7E00EF586F305ULL, "vxht_0_hightown_ajs_0_0.vlm"},
	{0x74C7E87CC425C4DFULL, "886567944.ogg"},
	{0x74C7E986C9219EAFULL, "lt_0_lowtown_docks_213_0.mao"},
	{0x74C7EB6981243D16ULL, "hlac_dr_dungeon_3215_5x.dds"},
	{0x74C7F586C921B34CULL, "lt_0_lowtown_docks_213_0.mmh"},
	{0x74C7FB86C921BD3EULL, "lt_0_lowtown_docks_213_0.msh"},
	{0x74C8121FB4A5F3F1ULL, "vxlt_gallowstemplar_6t1_0_0.vlm"},
	{0x74C91CB990F47C85ULL, "cd_and111_donation_box.bnk"},
	{0x74C9791F87D37B17ULL, "vxw_woundedcoast_1qp_0_0.vlm"},
	{0x74C98D104CB67C7FULL, "hlac_lt_gallowstemplar_f_2391_9z.dds"},
	{0x74C9A3829B0175F2ULL, "hightown_fem_cityguard_mob6168525.cl"},
	{0x74C9ABB1E934FA00ULL, "dae000_gamlen6147214.cl"},
	{0x74C9B3AF8A01A63AULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_fenris_fenrintro.ani"},
	{0x74C9E342B04C04A6ULL, "147413714.ogg"},
	{0x74CA2D984DE06383ULL, "6148093_m.fxe"},
	{0x74CAD863FD48FC96ULL, "hlac_w_sundermount_65533_0y.dds"},
	{0x74CAF9F83138BB38ULL, "vxw_bonepit_11h_0_0.vlm"},
	{0x74CB347B63736B56ULL, "t3_cor_b02.tnt"},
	{0x74CBDE494057DF93ULL, "gen_im_trs_tmp_one_axe_r0.uti"},
	{0x74CBF3DA8F6E8200ULL, "6076786_m.fxe"},
	{0x74CC7E05315396BAULL, "6131150_f.fxe"},
	{0x74CC83FC0FB2D327ULL, "489564836.ogg"},
	{0x74CC8791E1D04370ULL, "6206145_m.fxe"},
	{0x74CCDC92368DE66AULL, "332583053.ogg"},
	{0x74CCDEB94B01E4CEULL, "50449826.ogg"},
	{0x74CD64155235A8CBULL, "vxw_woundedcoast_1hk_0_0.vlm"},
	{0x74CD7AEA555F995BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tallnb.mmh"},
	{0x74CD7BAD2FAD361CULL, "vxlt_0_lowtown_a2_n_cox_0_0.vlm"},
	{0x74CD86EA555FADF8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tallnb.mao"},
	{0x74CD98EA555FCC45ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tallnb.msh"},
	{0x74CEE619E2AAF489ULL, "6171554_m.fxe"},
	{0x74CEF8781F0C38ACULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_wax.msh"},
	{0x74CEFE781F0C429EULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_wax.mmh"},
	{0x74CF043C05890CCDULL, "vxw_bonepit_1s3_0_0.vlm"},
	{0x74CF0A781F0C5735ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_wax.mao"},
	{0x74CF366BA258C5AFULL, "vxlt_0_lowtown_n_eo9_0_0.vlm"},
	{0x74CF657ADDD98D3AULL, "art\\vfx\\textures\\vfx_wis_smo_single.dds"},
	{0x74CF74A54AF6759CULL, "6149251_m.fxe"},
	{0x74CFB6EC0800876FULL, "752082964.ogg"},
	{0x74CFCF1E2E725548ULL, "587959054.ogg"},
	{0x74CFDE7B3951F65EULL, "hlac_ht_keep_n_3731_3y.dds"},
	{0x74CFEA897C63B9BCULL, "llac_lt_0_lowtown_docks_131070_0.dds"},
	{0x74CFEF292B58A9B0ULL, "rdr111dg_martin6055064.cl"},
	{0x74D00FAD5A630874ULL, "tbe_place_tome.ncs"},
	{0x74D027B4F692EAB3ULL, "ht_0_hightown_a3_17_0.anb"},
	{0x74D02D57AD132BD5ULL, "856475178.ogg"},
	{0x74D039C74940145CULL, "follower_banter6147605.cl"},
	{0x74D07294C40AE6CDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basicnb.mmh"},
	{0x74D07694C40AED1EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basicnb.mao"},
	{0x74D08494C40B0553ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basicnb.msh"},
	{0x74D08E2AB4A31D10ULL, "214152856.ogg"},
	{0x74D10AE0A85D4ED4ULL, "llac_w_encounter1_-2_0.dds"},
	{0x74D1905BFCCFDC12ULL, "63799226.ogg"},
	{0x74D2D817BBDCF852ULL, "mer320_hubert6053431.cl"},
	{0x74D33B6BF5EC7A78ULL, "art\\vfx\\creatures\\fxc_gol_proj_c.mmh"},
	{0x74D36E8ABF9092DCULL, "6086965_m.fxe"},
	{0x74D381CC0D899238ULL, "558592842.ogg"},
	{0x74D3D301404AEAAAULL, "hiteffects.gda"},
	{0x74D3E07AF1DB05B0ULL, "gen00fl_varric6079152.cl"},
	{0x74D4028BB0AE7EEEULL, "vxht_mansion2_1gk_0_0.vlm"},
	{0x74D47CBA2104A5B9ULL, "hlac_ca_caverns1_mines_24421_3y.dds"},
	{0x74D4967ACC7172C9ULL, "vxlt_0_lowtown_docks_n_b5g_0_0.vlm"},
	{0x74D49B778A419657ULL, "llac_lt_0_lowtown_docks_n_-3_4.dds"},
	{0x74D4ED37881B2495ULL, "6083071_m.fxe"},
	{0x74D52F92C4BD7679ULL, "qun120_arishok6052913.cl"},
	{0x74D67BAA9B46A086ULL, "vxw_woundedcoast_61_0_0.vlm"},
	{0x74D69B1FBB2F147BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dmg_v1.ani"},
	{0x74D6A49EEEC1B3ABULL, "hlac_lt_0_lowtown_docks_n_0_2x.dds"},
	{0x74D8B67FF2018241ULL, "6033397_m.fxe"},
	{0x74D8BAEDC621A666ULL, "vxca_caverns1_mines_ihs_0_0.vlm"},
	{0x74D93337777AD82CULL, "gen00fl_aveline6100114.cl"},
	{0x74D954BF4670DE24ULL, "6095550_m.fxe"},
	{0x74D9E35D037FB35CULL, "6120752_f.fxe"},
	{0x74DA1DEC1FDF2DCEULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\ready.evt"},
	{0x74DA2EAAD902F939ULL, "ave310_donnic6134757.cl"},
	{0x74DA317CA8A381CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l180_rf.evt"},
	{0x74DA5741A7A45EFAULL, "49038882.ogg"},
	{0x74DA64ACE326EF9BULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rblt.ani"},
	{0x74DA73BA75654FCFULL, "w_sundermount_164_0.mao"},
	{0x74DA7FBA7565646CULL, "w_sundermount_164_0.mmh"},
	{0x74DA85BA75656E5EULL, "w_sundermount_164_0.msh"},
	{0x74DB6A8CB58EE921ULL, "dae200cr_viscount6087182.cl"},
	{0x74DBC584A0AB3757ULL, "gen00fl_anders6189371.cl"},
	{0x74DBD5F2442BC14BULL, "vxw_woundedcoast_1sj_0_0.vlm"},
	{0x74DC30ECAF6449C8ULL, "6068456_f.fxe"},
	{0x74DCF85C31D9FB46ULL, "plotactiondefines.gda"},
	{0x74DD1D2AA9C80FD7ULL, "154245318.ogg"},
	{0x74DD2071D243CDABULL, "screencolorclear.vsh.fxo"},
	{0x74DD55C570080010ULL, "cd_qcr220_cut_distraction.bnk"},
	{0x74DD649598210869ULL, "qun221_viscount6084714.cl"},
	{0x74DDA848944EDE2FULL, "vxlt_0_lowtown_n_eh8_0_0.vlm"},
	{0x74DE544C3951C762ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomea_n.dds"},
	{0x74DE872AAB295277ULL, "mag211_marethari6025364.cl"},
	{0x74DF07ED25AE6F26ULL, "2119566.ogg"},
	{0x74DFB3EAADC5D4CDULL, "zz_ave_debug6207715.cl"},
	{0x74DFCA4B78FB72C7ULL, "vxlt_darktown_4x5_0_0.vlm"},
	{0x74DFE961D5A9E92BULL, "107665994.ogg"},
	{0x74E008EC378E4C37ULL, "6098055_m.fxe"},
	{0x74E09DDF3F811656ULL, "vxw_woundedcoast_10z_0_0.vlm"},
	{0x74E0DC0E759507B9ULL, "98724762.ogg"},
	{0x74E105792C79DDF9ULL, "w_encounter3_190_0.msh"},
	{0x74E10F792C79EEE7ULL, "w_encounter3_190_0.mmh"},
	{0x74E113792C79F534ULL, "w_encounter3_190_0.mao"},
	{0x74E114EC1FE53D79ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\ready.gad"},
	{0x74E14A7CA8A9CA8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l180_rf.gad"},
	{0x74E1AA647230BD77ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottle.msh"},
	{0x74E1BC647230DBDAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottle.mao"},
	{0x74E1C0647230E2E9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grogbottle.mmh"},
	{0x74E1CCA835B59AA7ULL, "gen00fl_merrill6116074.cl"},
	{0x74E1CF3BD3FA8534ULL, "pro000_friedrich6051039.cl"},
	{0x74E226372A537A55ULL, "dr_dungeon_3207_lc.mmh"},
	{0x74E253CD24A6061BULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornlrg.msh"},
	{0x74E265CD24A62566ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornlrg.mao"},
	{0x74E266BF51AD3745ULL, "bl_spa_cloud.mao"},
	{0x74E271CD24A63905ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornlrg.mmh"},
	{0x74E292E4219A0A6CULL, "gen00fl_varric6099543.cl"},
	{0x74E2A4401419E210ULL, "vxht_0_hightown_a3_c6m_0_0.vlm"},
	{0x74E3694740417D6DULL, "hlac_lt_0_lowtown_n_-196609_3x.dds"},
	{0x74E3D8BBAA7CFFBCULL, "aip_dalishwarrior2w_base.gda"},
	{0x74E3F36B3CDF5770ULL, "vxlt_0_lowtown_a2_es9_0_0.vlm"},
	{0x74E4220FFD9D2C44ULL, "dae000ip_crowd_lm5_talk12.utp"},
	{0x74E4226ABC6FA696ULL, "art\\characters\\weapons\\dagger\\w_dger_109a.phy"},
	{0x74E43E0C392A50B9ULL, "gen00fl_merrill6108493.cl"},
	{0x74E4662A20953253ULL, "llac_ht_0_hightown_a3_-65537_0.dds"},
	{0x74E47D428B008BD2ULL, "6149312_m.fxe"},
	{0x74E483BA77049085ULL, "llac_lt_house_-1_3.dds"},
	{0x74E49A3059C97903ULL, "hlac_ht_bartrand_n_3371_1y.dds"},
	{0x74E4A31FBB3B5F72ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dmg_v1.evt"},
	{0x74E52B842E16DD7DULL, "6116690_m.fxe"},
	{0x74E554FF548B57C0ULL, "vxlt_0_lowtown_a2_ewe_0_0.vlm"},
	{0x74E565FF0B62A77EULL, "hlac_lt_0_lowtown_docks_a3_n_65533_0z.dds"},
	{0x74E60AC788DFA9B8ULL, "w_encounter2_23_0.phy"},
	{0x74E60BB9063AC6A9ULL, "dae000cr_gamlen.utc"},
	{0x74E66C39D63A7AE8ULL, "character_character_p70.psh.fxo.dx11.assembly"},
	{0x74E68507D388261CULL, "6141916_m.fxe"},
	{0x74E6AF1CC9E3628FULL, "22131086.ogg"},
	{0x74E6CA044051B520ULL, "one220_yevhen6102953.cl"},
	{0x74E725EC1FE9C5C7ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\ready.ani"},
	{0x74E73155446B1AE5ULL, "plt_cod_aow_world_map.plo"},
	{0x74E7497CA8AE348AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_l180_rf.ani"},
	{0x74E77472DE1432E8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medb_0d.dds"},
	{0x74E7CAA573421EFCULL, "gencr_raider_assassin.utc"},
	{0x74E7FF3E985B69E9ULL, "hlac_lt_hangedman_-1_10y.dds"},
	{0x74E85F98A00DAB1FULL, "mag311_emile_de_launcet6043688.cl"},
	{0x74E862D73889A79CULL, "character_character_p164.psh.fxo.dx11.assembly"},
	{0x74E879F8301CEAABULL, "pro000_merchant6051807.cl"},
	{0x74E87CABAD3CD93EULL, "lt_gallowscourtyard_49_0.rml"},
	{0x74E8C102D699DC2BULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hyvo_0s.dds"},
	{0x74E8F8CB4F97046EULL, "vxw_sundermount_345_0_1.vlm"},
	{0x74E98DC746106018ULL, "vxht_keep_n_5u4_0_0.vlm"},
	{0x74E9E1F70F33FAC5ULL, "gen00fl_fenris6105876.cl"},
	{0x74EA740AE261C2BFULL, "vxw_encounter2_4c_0_0.vlm"},
	{0x74EA87EF4DC450F2ULL, "vxlt_gallowstemplar_47a_0_0.vlm"},
	{0x74EA8BF406C1DE1AULL, "ht_bartrand_n_probe_54o.dds"},
	{0x74EAC21FBB3FFF05ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_dmg_v1.gad"},
	{0x74EAD571CC00EA71ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb02a.mmh"},
	{0x74EAEF71CC01162FULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb02a.msh"},
	{0x74EB0DE428A5FCC5ULL, "hf_lady_harimann.mor"},
	{0x74EBABA2B0FE91FCULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potb.mao"},
	{0x74EBB31851B569EFULL, "199439408.ogg"},
	{0x74EBB7A2B0FEA59FULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potb.mmh"},
	{0x74EBBDA2B0FEB041ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potb.msh"},
	{0x74EBEA48BAAD811CULL, "vxw_woundedcoast_188_0_0.vlm"},
	{0x74ECE0E66B234631ULL, "vxht_0_hightown_a3_byx_0_0.vlm"},
	{0x74ED09C788E5C6D0ULL, "w_encounter2_23_0.rml"},
	{0x74ED46DABD83845BULL, "6036302_f.fxe"},
	{0x74ED8E97F1EBF158ULL, "vxht_bartrand_n_658_0_0.vlm"},
	{0x74EDCC56105AC189ULL, "vxlt_undercityserialkiller_1wq_0_0.vlm"},
	{0x74EE0B6ABC77D437ULL, "art\\characters\\weapons\\dagger\\w_dger_109a.msh"},
	{0x74EE216ABC77F9A9ULL, "art\\characters\\weapons\\dagger\\w_dger_109a.mmh"},
	{0x74EE8BACE337DB25ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_rblt.gad"},
	{0x74EEEC5D4E8B2125ULL, "hlac_lt_hangedman_-1_22x.dds"},
	{0x74EEF8A3D518C3D0ULL, "vxca_caverns2prc_9ei_0_0.vlm"},
	{0x74EEFDC001A96223ULL, "w_woundedcoast_183_0.msh"},
	{0x74EF07C58CC6A369ULL, "159923374.ogg"},
	{0x74EF0BC001A97A7DULL, "w_woundedcoast_183_0.mmh"},
	{0x74EF0FC001A9808EULL, "w_woundedcoast_183_0.mao"},
	{0x74F07F7423F78702ULL, "llac_lt_0_lowtown_a2_n_c_-65534_0.dds"},
	{0x74F0CA7C0AF94F4CULL, "hlac_lt_0_lowtown_-2_2x.dds"},
	{0x74F0E645949AAFEEULL, "vxht_mansion2_1mm_0_0.vlm"},
	{0x74F107966B634777ULL, "vxw_bonepitdestroyed_a3_ni_0_0.vlm"},
	{0x74F1551638180FBEULL, "hlac_lt_0_lowtown_a2_-2_5y.dds"},
	{0x74F18002721F918FULL, "373679286.ogg"},
	{0x74F1D070A931DC0CULL, "rune_weapon_cold.dds"},
	{0x74F2B52D381E44B5ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb104a_0s.dds"},
	{0x74F2D0E36D188D06ULL, "595360011.ogg"},
	{0x74F375C5A65E16AFULL, "242835454.ogg"},
	{0x74F397983178CCB1ULL, "w_woundedcoast_46_0.mmh"},
	{0x74F3A3983178E152ULL, "w_woundedcoast_46_0.mao"},
	{0x74F3B1983178F96FULL, "w_woundedcoast_46_0.msh"},
	{0x74F3F3B4D29879D0ULL, "ico_2h_sword_3.dds"},
	{0x74F4097083F7362AULL, "6065010_m.fxe"},
	{0x74F423DD4689D2F0ULL, "280614144.ogg"},
	{0x74F446E78BAA892AULL, "character_character_p238.psh.fxo.dx11"},
	{0x74F47B744DC6A306ULL, "llac_lt_0_lowtown_a2_-1_3.dds"},
	{0x74F496B9421082F3ULL, "vxdr_primevaltomb_919_0_0.vlm"},
	{0x74F50E7EBA1C1893ULL, "lowtown_male_cityguard_mob6150483.cl"},
	{0x74F557CB5911DA92ULL, "393435218.ogg"},
	{0x74F56A205C1C19CAULL, "hlac_lt_gallowsprison_n_-1_14x.dds"},
	{0x74F5A6263C817C46ULL, "vxw_sundermount_21v_0_0.vlm"},
	{0x74F5D59ED8D45496ULL, "6152129_m.fxe"},
	{0x74F627B0DCAF5340ULL, "vxht_keep_n_6xl_0_0.vlm"},
	{0x74F62A19C11260C2ULL, "921453686.ogg"},
	{0x74F63EC6406BDAC8ULL, "40426789.ogg"},
	{0x74F6597898B11949ULL, "vxca_caverns2_960_0_0.vlm"},
	{0x74F6E4F13BFAE9E9ULL, "336412389.ogg"},
	{0x74F7DA9C30267B44ULL, "165415325.ogg"},
	{0x74F7DFB4602D2A41ULL, "vxw_sundermount_2iu_0_0.vlm"},
	{0x74F8CDEEDF10950AULL, "dae000ip_crowd_ff4_talk12.utp"},
	{0x74F93943FE9C0FBFULL, "vxw_blightlands_cy_0_0.vlm"},
	{0x74F990B75F6BEF50ULL, "hlac_ht_0_hightown_a3_-65536_3y.dds"},
	{0x74F9CB678FC9FD01ULL, "gen00fl_aveline6121963.cl"},
	{0x74F9D736D00A36ECULL, "vxw_blightlands_ru_0_0.vlm"},
	{0x74F9DF115880CCA9ULL, "llac_dr_primevaltomb_664_1.dds"},
	{0x74FA19F4A2CEF955ULL, "llac_lt_undercitydungeon_-1_40.dds"},
	{0x74FA38D217D3514EULL, "745477143.ogg"},
	{0x74FA74AD109DC2EEULL, "mcr310_scared_mage.cnv"},
	{0x74FA8C9737A46732ULL, "6137555_m.fxe"},
	{0x74FAD830CE2CC1F8ULL, "6125162_f.fxe"},
	{0x74FB6F643A4DCFC2ULL, "vxht_0_hightown_a3_c2s_0_0.vlm"},
	{0x74FB8C1AEC981C06ULL, "6071838_f.fxe"},
	{0x74FB91285C0E27E3ULL, "vxlt_blackemporium_y7_0_0.vlm"},
	{0x74FBB252CF37C276ULL, "llac_dr_0_deeproads_5841_0.dds"},
	{0x74FBBD17136D4E7CULL, "91978633.ogg"},
	{0x74FBDCEA640B3F82ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_3.pwk"},
	{0x74FBE7651296ACB6ULL, "6087726_m.fxe"},
	{0x74FBEA160B4F86D0ULL, "vxlt_0_lowtown_g60_0_0.vlm"},
	{0x74FC33239604C130ULL, "6076089_m.fxe"},
	{0x74FC396D5D73704DULL, "hlac_lt_0_lowtown_docks_n_65532_0y.dds"},
	{0x74FC44B4F6B8576FULL, "ht_0_hightown_a3_17_0.rml"},
	{0x74FC4C3E1E14FEC7ULL, "gen00fl_aveline6139256.cl"},
	{0x74FCBEA7BC336519ULL, "6174730_f.fxe"},
	{0x74FCEAFAEEC2F747ULL, "gen00fl_anders6116818.cl"},
	{0x74FD072B696C9FBBULL, "702317313.ogg"},
	{0x74FD8BBC18AF5F4FULL, "vxca_caverns2_8fc_0_0.vlm"},
	{0x74FE0F7216EA2E24ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr90runf.ani"},
	{0x74FE480965AD7604ULL, "vxw_encounter3_by_0_0.vlm"},
	{0x74FE4B11B989EEBBULL, "zz_seb_debug6183082.cl"},
	{0x74FE8B71489C1FBEULL, "follower_banter6154063.cl"},
	{0x74FE96AD10A1E55EULL, "mcr310_scared_mage.bnk"},
	{0x74FECCFFF6912664ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackr.phy"},
	{0x74FEE71331152D6DULL, "507075447.ogg"},
	{0x74FEE9EBA2685D05ULL, "lowtown_argument6169423.cl"},
	{0x74FF6F7F9504EB48ULL, "vxdr_0_deeproads_cty_0_0.vlm"},
	{0x74FF88A1E5EB12C4ULL, "hlac_lt_0_lowtown_a2_n_65535_4x.dds"},
	{0x74FF9B53717F2556ULL, "dre000_varric6207950.cl"},
	{0x74FFC6A15B661FA1ULL, "x_narrativeroom_mmap_ph.dds"},
	{0x74FFEB1BA8175233ULL, "hlac_lt_0_lowtown_a2_n_-65537_6z.dds"},
	{0x75002FEC0ED81BE2ULL, "narrative_varric6172550.cl"},
	{0x75007AB8AC7865C2ULL, "6070496_m.fxe"},
	{0x750093B27F5B54F7ULL, "vrc321tr_projectile_1.utt"},
	{0x7500C841909AE81AULL, "zz_qun_debug6064340.cl"},
	{0x7500F007BD2FE86FULL, "hm_qmq_patronm1.mor"},
	{0x7501056F6DC5EEF9ULL, "ht_0_hightownexplosion_n_45_0_ll.mmh"},
	{0x75014750359C09AAULL, "dae000_tomwise6167335.cl"},
	{0x75016E272181525FULL, "mrl000_marethari6142487.cl"},
	{0x7501B5D65A66AD5FULL, "6218835_m.fxe"},
	{0x750267C53E59EA0DULL, "vrc321_party_barks6095737.cl"},
	{0x75028DAA99857BA8ULL, "6161006_m.fxe"},
	{0x7502DDFC1C27E84FULL, "follower_banter6141660.cl"},
	{0x750337C570289CDAULL, "cd_qcr220_cut_distraction.wbd"},
	{0x75035CD3AA1D4F76ULL, "6084543_m.fxe"},
	{0x7503B839D66354D2ULL, "917262900.ogg"},
	{0x75040C5D42130463ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear4.ani"},
	{0x75040C7216EE9486ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr90runf.gad"},
	{0x750443E132907761ULL, "306287551.ogg"},
	{0x750458A6C5A657E1ULL, "vxdr_0_deeproads_cqs_0_0.vlm"},
	{0x75049879EE822B06ULL, "vxlt_0_lowtown_docks_a3_cok_0_0.vlm"},
	{0x7504C30FE92CC7A5ULL, "209351993.ogg"},
	{0x7504DADAAB1DE2F3ULL, "vxw_encounter2_camp_8k_0_0.vlm"},
	{0x7505557D2139E8C0ULL, "86181457.ogg"},
	{0x75056E266939446EULL, "24408244.ogg"},
	{0x750600C12C75CA87ULL, "6197996_m.fxe"},
	{0x75065191522342C1ULL, "vxw_sundermount_2rt_0_0.vlm"},
	{0x75067A8E497324F0ULL, "dae000_denier6162667.cl"},
	{0x7506C6F9787504ACULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appa_0d.dds"},
	{0x7506C84632896E12ULL, "vxht_brothel_n_4hs_0_0.vlm"},
	{0x7507051ABD4E9348ULL, "81664692.ogg"},
	{0x750722F64D2C7082ULL, "vxw_bonepitmining_a2_f2_0_0.vlm"},
	{0x75081FB4A760FC88ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerd.mmh"},
	{0x75082BB4A761112BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerd.mao"},
	{0x75083DB4A7612FB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerd.msh"},
	{0x75084913C2004F6CULL, "vxw_blightlands_wj_0_0.vlm"},
	{0x750857731AAB0CC3ULL, "6035366_m.fxe"},
	{0x7508818AAD0A2749ULL, "6054880_m.fxe"},
	{0x7508E9B72E66CC50ULL, "688633932.ogg"},
	{0x7509BB1B5BAEC385ULL, "hlac_ht_0_hightown_a3_65532_0x.dds"},
	{0x7509FF93B3D0AF7AULL, "gen00fl_isabela6110178.cl"},
	{0x750B1FE24EDDBC47ULL, "vxlt_darktown_4i6_0_0.vlm"},
	{0x750B2FF709B9DAFFULL, "vxlt_0_lowtown_n_exx_0_0.vlm"},
	{0x750B6472466D1674ULL, "1018270278.ogg"},
	{0x750B9D3903EA2409ULL, "6162415_m.fxe"},
	{0x750BC368CF7520B9ULL, "hlac_ca_caverns1_24426_12y.dds"},
	{0x750C883AF6A5F16DULL, "6144561_m.fxe"},
	{0x750C963BEDDC614BULL, "6118305_f.fxe"},
	{0x750CA38F7DD396A7ULL, "6147812_m.fxe"},
	{0x750CCC5754DFA8FEULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em_scalpm1.mao"},
	{0x750CF853DCF21BE1ULL, "801062013.ogg"},
	{0x750CFD0E1BDBA68EULL, "684388153.ogg"},
	{0x750D9CB8C2ABBA86ULL, "6154743_m.fxe"},
	{0x750D9F3AE0F6DF64ULL, "vxw_woundedcoast_1gm_0_0.vlm"},
	{0x750DE15C86FB03C8ULL, "vxdr_0_deeproads_cqv_0_0.vlm"},
	{0x750E04715ACE7C06ULL, "llac_lt_0_lowtown_docks_131071_3.dds"},
	{0x750E484B5A7498A8ULL, "6153623_m.fxe"},
	{0x750E55814ABF6784ULL, "fonts_de-de.gfx"},
	{0x750E5679B5BD4DD4ULL, "vxht_0_hightown_a6b_0_0.vlm"},
	{0x750E718CA7BC1E2FULL, "vxht_brothel_n_502_0_1.vlm"},
	{0x750E74647E19C943ULL, "385533141.ogg"},
	{0x750EAEBEF0F1CE8AULL, "vxht_mansion2_1tv_0_0.vlm"},
	{0x750EEC959E6E7E42ULL, "6075796_m.fxe"},
	{0x750FC526E0BDF6A4ULL, "6140586_m.fxe"},
	{0x750FC68FC3879535ULL, "vxw_bonepitmining_a2_1v6_0_0.vlm"},
	{0x750FD627540AF892ULL, "gen00fl_isabela6100078.cl"},
	{0x75101625615AFD44ULL, "663476594.ogg"},
	{0x75108628C5D082E9ULL, "ico_mace_4.dds"},
	{0x7510D1E5C11D99A4ULL, "gen00fl_isabela6096854.cl"},
	{0x7511203250153109ULL, "351071145.ogg"},
	{0x7511E9AC65541AFBULL, "llac_lt_undercityserialkiller_-1_34.dds"},
	{0x751255F6FD04D2A4ULL, "309614508.ogg"},
	{0x75129738C7C1C1A1ULL, "vxlt_0_lowtown_a2_n_c_el9_0_1.vlm"},
	{0x7512ED3288E9196FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.closed.gad"},
	{0x7512F2FE26D0E6F0ULL, "6104546_m.fxe"},
	{0x75134CB79D13FB73ULL, "vxw_encounter2_camp_3j_0_0.vlm"},
	{0x7513B5D5BFD9CE46ULL, "dae100_bran6142996.cl"},
	{0x7514987DEB1457D6ULL, "225777512.ogg"},
	{0x7514C36E44195B41ULL, "vxw_sundermount_2wk_0_0.vlm"},
	{0x7514E35E1BF2A998ULL, "vxht_keep_n_5d3_0_0.vlm"},
	{0x75152E665711F096ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\basedualwield\\blendtree\\assassin.anb"},
	{0x751542EAAD168C25ULL, "233210749.ogg"},
	{0x75154B81F1876C25ULL, "826061066.ogg"},
	{0x7515661F6672188CULL, "vxdr_0_deeproads_ew6_0_0.vlm"},
	{0x751574E1505F0F74ULL, "gen00fl_aveline6122064.cl"},
	{0x751585425D13B402ULL, "553440209.ogg"},
	{0x75159BEA559D8690ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_tallnb.phy"},
	{0x7515A31D9DB5DBA8ULL, "vxlt_0_lowtown_n_fhx_0_0.vlm"},
	{0x7515AD60CD180A37ULL, "6088696_m.fxe"},
	{0x7515F1E111F7DB32ULL, "6126618_m.fxe"},
	{0x7517D51A0A9FBBC5ULL, "6038639_m.fxe"},
	{0x7518235D4223D44DULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear4.gad"},
	{0x751837A6A6A65855ULL, "711049959.ogg"},
	{0x75183BE391FBDB32ULL, "hlac_ht_0_hightown_n_-196610_0z.dds"},
	{0x751882BC75036A3EULL, "llac_ca_cavern1_haunted_24423_3.dds"},
	{0x75188758762E8228ULL, "6089713_m.fxe"},
	{0x7518AF8BD79B0847ULL, "6043239_m.fxe"},
	{0x7519E22E140FA7D1ULL, "900842644.ogg"},
	{0x751A0F15F5E03C89ULL, "6167715_m.fxe"},
	{0x751AA13AF251127AULL, "65044304.ogg"},
	{0x751AD007AC388A98ULL, "vxdr_slavers_1b3_0_0.vlm"},
	{0x751AD77075C27DE7ULL, "art\\vfx\\creatures\\fxc_drg_brth_e_c.anb"},
	{0x751B3E0A96F6037BULL, "hlac_ht_brothel_n_3367_1x.dds"},
	{0x751B64C27A5B4B3BULL, "vxht_bartrand_n_5r5_0_0.vlm"},
	{0x751BC78C3A0D819BULL, "vxlt_shop_fa_0_0.vlm"},
	{0x751C1B27936BC149ULL, "mag291_vincento6027673.cl"},
	{0x751C7FB51676C6D0ULL, "mag100_cullen6093738.cl"},
	{0x751D51DC27732FF2ULL, "485453869.ogg"},
	{0x751D9289E139BE39ULL, "vxht_0_hightown_n_2xa_0_0.vlm"},
	{0x751E33D994A8C656ULL, "689950003.ogg"},
	{0x751E386CB57DADB5ULL, "vxw_bonepitmining_a2_1s5_0_0.vlm"},
	{0x751E67A78C1DD85EULL, "414242517.ogg"},
	{0x751EC0ACE888163BULL, "928007571.ogg"},
	{0x751F5E468468C19AULL, "6077138_m.fxe"},
	{0x751F64A9F6389C75ULL, "hlac_lt_gallowstemplar_f_2391_2y.dds"},
	{0x751FED5540C06E24ULL, "zz_followers_banter6157651.cl"},
	{0x75200149B1229DF9ULL, "vxlt_0_lowtown_docks_b4j_0_1.vlm"},
	{0x752040547175DAE9ULL, "vxht_0_hightown_a3_c0r_0_0.vlm"},
	{0x75209873BE64FFCCULL, "hlac_lt_blackemporium_1055_22z.dds"},
	{0x7520A95AD50FCBFBULL, "llac_lt_0_lowtown_docks_a3_-65538_4.dds"},
	{0x7520B8C537AA603FULL, "dae000_brat6156940.cl"},
	{0x7520BEB02347FC6DULL, "hlac_lt_0_lowtown_docks_a3_n_131072_1x.dds"},
	{0x7520D85B30E958D2ULL, "481669549.ogg"},
	{0x7521361BD502B8F0ULL, "251039259.ogg"},
	{0x7521795FA76FB2F3ULL, "802524376.ogg"},
	{0x752179ABAD6C99E2ULL, "lt_gallowscourtyard_49_0.anb"},
	{0x7521DE1A9267E33AULL, "6021220_f.fxe"},
	{0x7521DFFE8563B609ULL, "mag131_keran6091386.cl"},
	{0x75222DA0F5FC231BULL, "678871602.ogg"},
	{0x75229D59271DA54DULL, "art\\characters\\creatures\\dog\\textures\\c_dog_tatb.dds"},
	{0x7522CFCA6DF63B3AULL, "943634742.ogg"},
	{0x7522F7E74FE2C811ULL, "vxht_mansion_n_5a1_0_0.vlm"},
	{0x7523354B360F73C1ULL, "vxw_sundermount_2xj_0_0.vlm"},
	{0x7523D60F3B159B8EULL, "hlac_w_encounter3_-131075_0y.dds"},
	{0x75241B9B6CF3A3DAULL, "vxht_keep_n_6fx_0_0.vlm"},
	{0x75248EDAD47E593FULL, "vxw_blightlands_rz_0_0.vlm"},
	{0x7524B87E9C011552ULL, "853709004.ogg"},
	{0x752534BDBCDFDC63ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_boo_cas_0n.dds"},
	{0x752534E21F24EF43ULL, "dae000_merchant_rare_em6177940.cl"},
	{0x7525492ECC8D4126ULL, "one100dg_lia6176018.cl"},
	{0x7525A7EE4BF848DDULL, "6083105_m.fxe"},
	{0x7525DB92F4A675CEULL, "15482167.ogg"},
	{0x75270A3288F9F335ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.closed.ani"},
	{0x7527200531932D13ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeend.phy"},
	{0x75272DB7B308F2AEULL, "vxht_chantry_n_10q_0_0.vlm"},
	{0x75277CA3484AAC5BULL, "vxlt_undercityserialkiller_25z_0_0.vlm"},
	{0x75278178B1E11D00ULL, "6126678_m.fxe"},
	{0x7527B0E59212A7C2ULL, "vxlt_hangedman_n_62_0_0.vlm"},
	{0x752813B0EDA60BFFULL, "vxlt_0_lowtown_n_enz_0_0.vlm"},
	{0x752845513A1048D0ULL, "6120863_m.fxe"},
	{0x75287AD33B565E8BULL, "6015809_m.fxe"},
	{0x7528889D48DC5369ULL, "hlac_lt_gallowstemplar_f_2384_0y.dds"},
	{0x7529BD6B897C5DB5ULL, "vxw_bonepitmining_a2_29l_0_0.vlm"},
	{0x7529D69F6C5163BBULL, "vxht_bartrand_n_5u4_0_0.vlm"},
	{0x752A5B76180FFDA1ULL, "6037339_m.fxe"},
	{0x752A5E7E036E0ACFULL, "hlac_lt_0_lowtown_docks_a3_196607_0y.dds"},
	{0x752AC394600CC6F8ULL, "lt_0_lowtown_n_36_0_ll.mmh"},
	{0x752AF817E0AD3C0DULL, "vxca_caverns1_jg3_0_0.vlm"},
	{0x752B9278D4DDFF8AULL, "vxht_mansion2_1bp_0_0.vlm"},
	{0x752C0D4A78E9D3E1ULL, "vxht_chantry_2cs_0_0.vlm"},
	{0x752C9855974334C3ULL, "38907756.ogg"},
	{0x752CE415AAC669CCULL, "874312177.ogg"},
	{0x752D17ADAE143A34ULL, "vxlt_gallowstemplar_n_373_0_0.vlm"},
	{0x752D2F3A06752F2FULL, "vxw_bonepitmining_a2_dy_0_0.vlm"},
	{0x752D5FDFE562F416ULL, "6076487_m.fxe"},
	{0x752D8BB22FEFF9DBULL, "960448268.ogg"},
	{0x752E2A19DC0895EBULL, "gen00fl_aveline6120852.cl"},
	{0x752E8A982E045838ULL, "601065309.ogg"},
	{0x752EAD5A880C1BD0ULL, "vxw_bonepit_79_0_0.vlm"},
	{0x752EB231636BF369ULL, "vxht_0_hightown_a3_c5q_0_0.vlm"},
	{0x752ECF7227B3530FULL, "vxht_0_hightown_n_cae_0_0.vlm"},
	{0x752F137CD8DCBA0FULL, "damagetarget.ncs"},
	{0x752F758B5BA436AFULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypouchb_n.dds"},
	{0x752F815F1C92494CULL, "461651731.ogg"},
	{0x7530D5ABA0EDD70AULL, "vxlt_0_lowtown_docks_a3_n_bu7_0_0.vlm"},
	{0x7530F731C321B1F9ULL, "vxlt_0_lowtown_docks_bsr_0_0.vlm"},
	{0x753116E073061F6EULL, "6045027_m.fxe"},
	{0x7531794377830DD2ULL, "865102761.ogg"},
	{0x75327982C6C7641DULL, "6067903_m.fxe"},
	{0x75328D785EE9BC5AULL, "vxht_keep_n_6ay_0_0.vlm"},
	{0x7532911AB1043B37ULL, "6086896_m.fxe"},
	{0x7532B9DA0380848BULL, "1957892.ogg"},
	{0x7532C6D822999243ULL, "sse_ss_city_guard_m_1.gda"},
	{0x7532E2325F7371A3ULL, "159292481.ogg"},
	{0x753300B7C67471BFULL, "vxw_blightlands_we_0_0.vlm"},
	{0x75330568ACA07103ULL, "280551220.ogg"},
	{0x753316454E1F172EULL, "6104856_m.fxe"},
	{0x75331D400DE11B68ULL, "vxlt_hangedman_n_tb_0_0.vlm"},
	{0x75334CF6B7769CFEULL, "mag121_idunna6084865.cl"},
	{0x7533F93D1554AE6FULL, "vxht_0_hightown_a2_aju_0_0.vlm"},
	{0x7533FDDFE9B5CCC4ULL, "130498962.ogg"},
	{0x75340CA67CF18EA1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloorb_1.mao"},
	{0x75342DD84B6C5AF4ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.anb"},
	{0x753482B283E0702AULL, "450202805.ogg"},
	{0x75349485F5764005ULL, "6125249_m.fxe"},
	{0x75349571E6DE1F72ULL, "647413227.ogg"},
	{0x7534BF7FC251D506ULL, "6084473_m.fxe"},
	{0x7534F94972EAE59BULL, "6163350_f.fxe"},
	{0x75356B125A1AB328ULL, "6087513_m.fxe"},
	{0x75358419588AF320ULL, "6202307_m.fxe"},
	{0x753596003A3E7FC8ULL, "vxht_chantry_2cv_0_0.vlm"},
	{0x75359673E01F8491ULL, "vxw_bonepitdestroyed_a3_25j_0_0.vlm"},
	{0x75359F3C87A0E44BULL, "6042801_m.fxe"},
	{0x7535E11A48B757B6ULL, "vxht_0_hightown_n_9zg_0_0.vlm"},
	{0x7536852E7CE295B4ULL, "519262283.ogg"},
	{0x753699CDE166FF06ULL, "398222945.ogg"},
	{0x75370DC4C8B81291ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_harrowing_r.dds"},
	{0x7537B3E176897B8DULL, "6095124_m.fxe"},
	{0x75382DC1F39C1347ULL, "art\\vfx\\combat\\vfx_glb_confused_c.impa.gad"},
	{0x7538487C5E477C3BULL, "vxht_bartrand_n_5x7_0_0.vlm"},
	{0x75385E5D6B3B315FULL, "static1_ds2y.psh.fxo"},
	{0x75385FB8BEAE9DEFULL, "hla_ht_0_hightown_a3_-65536_0z.dds"},
	{0x753865FA432CF0A9ULL, "vxlt_0_lowtown_ep0_0_1.vlm"},
	{0x753918394DB0564DULL, "662351071.ogg"},
	{0x7539202B24CF7DF9ULL, "character_character_p279.psh.fxo.dx11.assembly"},
	{0x7539A02EE8525AD2ULL, "mrl350ip_grave_e.utp"},
	{0x753A1EA12C684A92ULL, "gen00fl_aveline6114416.cl"},
	{0x753A22D2427BBC89ULL, "391542449.ogg"},
	{0x753A435EEE21D7FCULL, "59304867.ogg"},
	{0x753AD170E1C7CC22ULL, "vxdr_0_deeproads_by9_0_0.vlm"},
	{0x753B8CE78A493095ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_2y.dds"},
	{0x753BF39E382FA630ULL, "6020485_f.fxe"},
	{0x753C53A69592ACBDULL, "6177390_m.fxe"},
	{0x753DC8B449EE2B1BULL, "vxlt_gallowscourtyard_76j_0_0.vlm"},
	{0x753EB9BA41098084ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgta.msh"},
	{0x753ECBBA41099FEDULL, "art\\characters\\playercharacter\\humanmale\\glv_lgta.mao"},
	{0x753ECFBA4109A626ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgta.mmh"},
	{0x753EF85136143507ULL, "6151572_m.fxe"},
	{0x753F22C1F3A21F54ULL, "art\\vfx\\combat\\vfx_glb_confused_c.impa.evt"},
	{0x753F2BD731125004ULL, "vxht_0_hightown_n_bnv_0_0.vlm"},
	{0x753F817FED4CAD9AULL, "vxht_0_hightown_a3_a87_0_0.vlm"},
	{0x753FF85C13A93C4FULL, "vxlt_gallowstemplar_5bv_0_0.vlm"},
	{0x7540BBB63C54620CULL, "vxdr_0_deeproads_ef5_0_0.vlm"},
	{0x75418F1CFFD733E8ULL, "vxlt_hangedman_n_sc_0_0.vlm"},
	{0x7541F7C7892DE84DULL, "w_encounter2_23_0.msh"},
	{0x754205C7892DFF80ULL, "w_encounter2_23_0.mao"},
	{0x754206FEB0ED03D9ULL, "vxca_caverns1_j22_0_2.vlm"},
	{0x754209C7892E06D3ULL, "w_encounter2_23_0.mmh"},
	{0x75420CD84B785FB7ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.mmh"},
	{0x754212A0CA6CEA5FULL, "191143082.ogg"},
	{0x7542276C12EE8BD5ULL, "gen00fl_fenris6126992.cl"},
	{0x7542D47075E3CAE8ULL, "art\\vfx\\creatures\\fxc_drg_brth_e_c.mmh"},
	{0x75432C32CFE20FE9ULL, "qun110_petrice6018649.cl"},
	{0x75433F3C41C8D06FULL, "hlac_ca_cavern1_haunted_24424_23z.dds"},
	{0x754383EE652A5FE2ULL, "hlac_lt_undercitydungeon_-1_12x.dds"},
	{0x75439D258F677E40ULL, "6095028_m.fxe"},
	{0x7543F391FE119B13ULL, "hlac_lt_0_lowtown_docks_a3_n_65536_0x.dds"},
	{0x75444685D0ECF9E3ULL, "6096783_m.fxe"},
	{0x7544978AB82745BDULL, "6105658_m.fxe"},
	{0x7544A2ED3036821AULL, "vxlt_gallowstemplar_f_2mn_0_0.vlm"},
	{0x7544A5D436A6849AULL, "mag100_thrask6148432.cl"},
	{0x75457038896B1C63ULL, "gen00fl_isabela6135622.cl"},
	{0x7545920FBEFB8F1FULL, "qun111_ketojan6120462.cl"},
	{0x7545B1C061483F6CULL, "gen_im_arm_cht_hvy_act1_02.uti"},
	{0x75462F083C054A1EULL, "vxlt_house_ey_0_0.vlm"},
	{0x7546CABB6C5A9749ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.cess.ani"},
	{0x7547A8B27F977AA6ULL, "vrc321tr_projectile_1.ncs"},
	{0x7547BE313F618FE0ULL, "6051266_m.fxe"},
	{0x7547CF3B486117E3ULL, "vxht_mansion2b_28l_0_0.vlm"},
	{0x75485EA9C3982F23ULL, "art\\vfx\\combat\\vfx_glb_charm_c.cess.evt"},
	{0x75487632B8CA308AULL, "vxht_mansion2_1hr_0_0.vlm"},
	{0x75489FB98172249FULL, "265410946.ogg"},
	{0x7548C6BA4111EBF1ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgta.phy"},
	{0x7548F1045CD604E1ULL, "vxht_chantry_2yq_0_0.vlm"},
	{0x754AE0F0F25254C2ULL, "zz_dae_debug6029323.cl"},
	{0x754BAB093FEF37E0ULL, "vxca_caverns2_7c9_0_1.vlm"},
	{0x754BC1EC0D046318ULL, "6128972_f.fxe"},
	{0x754BE534B9106CFBULL, "gen00fl_varric6033652.cl"},
	{0x754C1AC1F3AC9B8DULL, "art\\vfx\\combat\\vfx_glb_confused_c.impa.ani"},
	{0x754C5C4596F29A4AULL, "hlac_lt_0_lowtown_a2_n_131072_1y.dds"},
	{0x754CE5BB6C5F30CBULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.cess.gad"},
	{0x754CFB4409ADEC35ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml107a_0s.dds"},
	{0x754D08B012251E7EULL, "art\\vfx\\materials\\vfx_crack_add.mao"},
	{0x754D77D5A1057D6BULL, "452239879.ogg"},
	{0x754D839869799C8DULL, "vxw_encounter2_camp_o0_0_0.vlm"},
	{0x754E3A2E271F692CULL, "6042727_m.fxe"},
	{0x754E4BA85B3FB561ULL, "222937374.ogg"},
	{0x754E6A39059F54CFULL, "vxlt_gallowstemplar_5oq_0_0.vlm"},
	{0x754EB3F971185F75ULL, "qcr000_arishok6060195.cl"},
	{0x754EC0A2B15332E4ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potb.phy"},
	{0x754EC630A1AA2755ULL, "666347304.ogg"},
	{0x754F318E7C7B2984ULL, "ser221_quentin6123428.cl"},
	{0x754F37A9C39E0BA4ULL, "art\\vfx\\combat\\vfx_glb_charm_c.cess.gad"},
	{0x754F9DE8E6DAB4DCULL, "6016149_f.fxe"},
	{0x755034D0B278CB4DULL, "vxca_cavern1_haunted_jfi_0_0.vlm"},
	{0x75507822FD87EB40ULL, "538126300.ogg"},
	{0x75508DE2C8971F40ULL, "dae300_alistair_king6133589.cl"},
	{0x7550C6F70ECA39E1ULL, "follower_banter6171926.cl"},
	{0x7551110F4EFAA8D2ULL, "zz_followers_banter6158502.cl"},
	{0x7551379A85EF5862ULL, "vxht_bartrand_n_5ww_0_0.vlm"},
	{0x7551C699B2536BF9ULL, "173444046.ogg"},
	{0x7551F73AA85C0ED1ULL, "mcr371_alain.bnk"},
	{0x7552215F80B412F1ULL, "w_bonepitdestroyed_a3_3_0_lc.mmh"},
	{0x75522392333C1687ULL, "331521590.ogg"},
	{0x755250A31A68E20AULL, "sky_w_woundedcoast.mao"},
	{0x755279BA1E2AB0C8ULL, "vxht_chantry_2yt_0_0.vlm"},
	{0x7552B0B1BAF28D97ULL, "darktown_male_elf_sta6174552.cl"},
	{0x7552B93DCD40FF0DULL, "crt_rock_wraith.bnk"},
	{0x7553215F89F06E97ULL, "128603191.ogg"},
	{0x75537ED1582EF6A1ULL, "6033376_f.fxe"},
	{0x755389525BE22A0AULL, "snow_02.mao"},
	{0x7553E2890C453596ULL, "dae000_orsino6085482.cl"},
	{0x7553E2BB6C654A30ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.cess.evt"},
	{0x75542D94181825CFULL, "vxht_0_hightown_bsu_0_0.vlm"},
	{0x7554B7CD18D91932ULL, "hlac_ca_cavern1_haunted_24425_3x.dds"},
	{0x7554D93AA85E11E1ULL, "mcr371_alain.cnv"},
	{0x755517B1F9E3453EULL, "6085140_m.fxe"},
	{0x7555B528AA129EC7ULL, "669370350.ogg"},
	{0x755619EEC504CE4CULL, "758970927.ogg"},
	{0x75563F16BDDE0130ULL, "vxca_caverns2_8dj_0_0.vlm"},
	{0x755646A9C3A44382ULL, "art\\vfx\\combat\\vfx_glb_charm_c.cess.ani"},
	{0x75567617B08C7E87ULL, "static_rendertodepth_p0.vsh.fxo.dx11"},
	{0x7556C66C5652C471ULL, "ht_0_hightown_a3_28_0_ll.mmh"},
	{0x7557F219B951B4B6ULL, "l_lgt241ar_tome_cave.wbd"},
	{0x75580E8A1B90F769ULL, "zz_vault_debug6203362.cl"},
	{0x75588F1BF2E638A8ULL, "6078421_m.fxe"},
	{0x7558CF5BFE426092ULL, "mag170_hugh6080811.cl"},
	{0x75592446F474AA3EULL, "79898904.ogg"},
	{0x755954A57F958B12ULL, "dae000ip_crowd_mf7_talk5.utp"},
	{0x75596F8998ECD82DULL, "908190837.ogg"},
	{0x75597C37B32259A7ULL, "gen00fl_varric6118951.cl"},
	{0x7559EB442BD44966ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_right_3.mao"},
	{0x755A076BBD072FAEULL, "804656326.ogg"},
	{0x755A3615FBC000FCULL, "vxw_encounter2_camp_qn_0_1.vlm"},
	{0x755AF619904255BDULL, "llac_ca_caverns2_5802_73.dds"},
	{0x755B460B4CB585CCULL, "lt_0_lowtown_docks_n_168_0.msh"},
	{0x755B4C0B4CB58FBEULL, "lt_0_lowtown_docks_n_168_0.mmh"},
	{0x755B580B4CB5A455ULL, "lt_0_lowtown_docks_n_168_0.mao"},
	{0x755B9F98726B37DDULL, "6030639_m.fxe"},
	{0x755C06F200CADF92ULL, "hlac_w_woundedcoast_262149_0y.dds"},
	{0x755C6A32BE0485C4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_stump01_n.dds"},
	{0x755CDC15F7956D4FULL, "vxlt_gallowstemplar_5hp_0_0.vlm"},
	{0x755D04E44AE4E638ULL, "vxw_bonepitmining_a2_1xm_0_0.vlm"},
	{0x755D32640D8ADC65ULL, "mrl000_marethari6142377.cl"},
	{0x755DFADCCBD0EC6DULL, "mcr310_argument6074419.cl"},
	{0x755E4B2DAF2CB7FEULL, "vxlt_0_lowtown_docks_a3_c1a_0_0.vlm"},
	{0x755E50CE33E363D0ULL, "6079152_m.fxe"},
	{0x755EE8A4158680A1ULL, "697371086.ogg"},
	{0x755FACB50633F3ADULL, "dae000_mother6103054.cl"},
	{0x755FDA2969F6D572ULL, "vxlt_0_lowtown_n_e16_0_0.vlm"},
	{0x7560396BAC84B27EULL, "vxw_bonepitmining_a2_1p3_0_0.vlm"},
	{0x756066586C1B73A6ULL, "308383504.ogg"},
	{0x756079E65ED9E84CULL, "llac_lt_darktown_-1_64.dds"},
	{0x75611B625F6231BCULL, "6172972_m.fxe"},
	{0x7561D4EF916A85D3ULL, "6144109_m.fxe"},
	{0x756257199F5C935DULL, "6006065_m.fxe"},
	{0x75625D97563C9720ULL, "452537765.ogg"},
	{0x7562A997AB9A644EULL, "vxlt_gallowstemplar_43i_0_0.vlm"},
	{0x756311978416F79CULL, "llac_ht_mansion2b_818_24.dds"},
	{0x7563974269B5B9F6ULL, "follower_banter6155958.cl"},
	{0x7563B254706A5ED5ULL, "525664605.ogg"},
	{0x756486B8B6AB1BF5ULL, "vxdr_primevaltomb_500_0_0.vlm"},
	{0x7564C320086B1692ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobb_0d.dds"},
	{0x75657945F6FF207DULL, "w_woundedcoast_117_0.phy"},
	{0x75663B6BCB0D34CEULL, "633291095.ogg"},
	{0x75670C832AC9768DULL, "art\\levels\\4cinematics\\ferelden\\prp_templarbadge.mmh"},
	{0x756710832AC97DDEULL, "art\\levels\\4cinematics\\ferelden\\prp_templarbadge.mao"},
	{0x75671287C781CD00ULL, "vxca_caverns2_8o7_0_0.vlm"},
	{0x75671B7884679FA2ULL, "6043362_m.fxe"},
	{0x75671E832AC99513ULL, "art\\levels\\4cinematics\\ferelden\\prp_templarbadge.msh"},
	{0x75671F590C536D70ULL, "vxlt_0_lowtown_eoi_0_0.vlm"},
	{0x75673B8D9C99B242ULL, "255809562.ogg"},
	{0x7567E0F0890FC6B9ULL, "308236782.ogg"},
	{0x7567FD1B0E149DC1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_wic_0n.dds"},
	{0x7568136C6C64224FULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_conj_c.impa.gad"},
	{0x756872F9F97EB0C6ULL, "ht_0_hightown_a3_37_0_lc.mmh"},
	{0x7568B6E7F4168FE9ULL, "mag170_hugh6080735.cl"},
	{0x75693202A322B679ULL, "mcr310_argument6074518.cl"},
	{0x7569DD4C38A6E13AULL, "6054388_m.fxe"},
	{0x7569EB0362C8224FULL, "852960362.ogg"},
	{0x756A071A907FC152ULL, "6142021_m.fxe"},
	{0x756A44F3E1A71385ULL, "pet_01.mao"},
	{0x756B12E1F9606123ULL, "mer210_seldrek6078798.cl"},
	{0x756B4DF2E98B85CFULL, "vxlt_gallowstemplar_5us_0_0.vlm"},
	{0x756B76C13CDAFEB8ULL, "vxw_bonepitmining_a2_1uj_0_0.vlm"},
	{0x756C09CF658497F5ULL, "76655273.ogg"},
	{0x756C4B9D4B854F66ULL, "hlac_lt_0_lowtown_a3_n_c_131071_4x.dds"},
	{0x756CAEB19DD68DFDULL, "vxht_0_hightown_a2_90e_0_0.vlm"},
	{0x756CB3F32EE75809ULL, "zz_mag_debug6044331.cl"},
	{0x756D44F5244764F8ULL, "vxdr_primevaltomb_7kn_0_0.vlm"},
	{0x756E1CFE8B6A8090ULL, "hlac_ca_cavern1_haunted_24423_13y.dds"},
	{0x756E72A4F0C7147AULL, "6027755_m.fxe"},
	{0x756E8C25BC2634A4ULL, "ch2alphemnotnt.psh.fxo"},
	{0x756EB1AB66FB3D14ULL, "ico_armor_light_1.dds"},
	{0x756EC409868E90FAULL, "6108676_m.fxe"},
	{0x756EF394751A2627ULL, "vxlt_darktown_5jz_0_0.vlm"},
	{0x756F286C6C6A64FCULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_conj_c.impa.evt"},
	{0x756F42B5E52C188EULL, "vxlt_hangedman_n_wa_0_0.vlm"},
	{0x756F46720F1FDA6FULL, "lt_0_lowtown_docks_n_96_0.mao"},
	{0x756F52720F1FEE0CULL, "lt_0_lowtown_docks_n_96_0.mmh"},
	{0x756F58720F1FF8FEULL, "lt_0_lowtown_docks_n_96_0.msh"},
	{0x756F696EBD0CDFAAULL, "571308860.ogg"},
	{0x756F8D3F51584A3CULL, "6175872_m.fxe"},
	{0x756FBB6CFA677AD3ULL, "vxlt_gallowstemplar_355_0_0.vlm"},
	{0x756FF2817A4FEE55ULL, "vxdr_primevalentrance_1q3_0_0.vlm"},
	{0x75702A74DC3A91E0ULL, "864783172.ogg"},
	{0x757046CC83609E53ULL, "6141209_m.fxe"},
	{0x7571025DC1E39BBDULL, "vxlt_gallowstemplar_4qh_0_0.vlm"},
	{0x7571114DFC0456CFULL, "vxht_0_hightown_byw_0_0.vlm"},
	{0x75711B749D907CCEULL, "vxlt_gallowstemplar_46j_0_0.vlm"},
	{0x7571A7BB4A0A3B33ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd12_basic.phy"},
	{0x75729C3C9492419EULL, "mcr310_argument6173618.cl"},
	{0x7572A2525BFD051CULL, "snow_02.dds"},
	{0x75732D1FD73C932BULL, "6034035_m.fxe"},
	{0x75734A0A42D017AAULL, "vxw_sundermount_21z_0_0.vlm"},
	{0x7573570F43F609E6ULL, "vxlt_0_lowtown_docks_n_app_0_0.vlm"},
	{0x75736939022C66C8ULL, "6053552_m.fxe"},
	{0x75736F51FAA73052ULL, "327721967.ogg"},
	{0x75749372D58B6F51ULL, "951040200.ogg"},
	{0x75757B0F0D1C1900ULL, "vxht_0_hightown_be0_0_0.vlm"},
	{0x7575A00AE52E5E22ULL, "lowtown_male_urchin_sta6164196.cl"},
	{0x7575A59D01E6129BULL, "hlac_lt_0_lowtown_a2_n_-196609_0y.dds"},
	{0x7575C7EA7DA04096ULL, "zz_finalshot.cut"},
	{0x75766E45F70D80E9ULL, "w_woundedcoast_117_0.mao"},
	{0x75767A45F70D9482ULL, "w_woundedcoast_117_0.mmh"},
	{0x75767C45F70D9838ULL, "w_woundedcoast_117_0.msh"},
	{0x7576F3C78E18CAB5ULL, "zz_and_debug6134122.cl"},
	{0x7576FC1E2F85E65BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02b.mmh"},
	{0x75771A1E2F861945ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02b.msh"},
	{0x75773A43DE28129CULL, "vxlt_foundry_n_yw_0_0.vlm"},
	{0x7577A8B63B3A0F0DULL, "zz_vault_debug6203238.cl"},
	{0x75781430026BC6BBULL, "art\\characters\\playercharacter\\humanmale\\arm_masc.msh"},
	{0x75782630026BE506ULL, "art\\characters\\playercharacter\\humanmale\\arm_masc.mao"},
	{0x75783230026BF9A5ULL, "art\\characters\\playercharacter\\humanmale\\arm_masc.mmh"},
	{0x7578513CD5114EEDULL, "mrl251cr_spider.utc"},
	{0x7578CAB8BBD55C8FULL, "6081372_m.fxe"},
	{0x757906E64DFC798DULL, "vxdr_slavers_12r_0_0.vlm"},
	{0x75793C22BBBC1952ULL, "vxlt_gallowstemplar_350_0_0.vlm"},
	{0x7579EFD04178C3CFULL, "vxw_sundermount_304_0_1.vlm"},
	{0x757BB6D2163D7D78ULL, "vxdr_primevaltomb_7do_0_0.vlm"},
	{0x757C306C6C74FC15ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_conj_c.impa.ani"},
	{0x757D0E5C3032A149ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunkxlo.mao"},
	{0x757D1A5C3032B5E2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunkxlo.mmh"},
	{0x757D1C5C3032B898ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunkxlo.msh"},
	{0x757D1E31FD29A398ULL, "6020786_f.fxe"},
	{0x757D6F4659E3A833ULL, "vxht_0_hightown_c3p_0_0.vlm"},
	{0x757ED67FFA3CAA3BULL, "mcr370_negotiation6080186.cl"},
	{0x757EFCC16A95784CULL, "vxlt_0_lowtown_docks_n_b7d_0_0.vlm"},
	{0x757F08BA7782E854ULL, "147376073.ogg"},
	{0x757F3D98CB6AC7A5ULL, "240123573.ogg"},
	{0x757FAF3DCD67DC37ULL, "crt_rock_wraith.wbd"},
	{0x757FF2689B30F2EEULL, "vxlt_gallowscourtyard_3ta_0_0.vlm"},
	{0x7580045F8407B162ULL, "29155749.ogg"},
	{0x75815F1A03D19B9EULL, "6135878_m.fxe"},
	{0x7581A982DF099582ULL, "6033710_m.fxe"},
	{0x758265C6BE40CBF5ULL, "385411221.ogg"},
	{0x7584171222E6B078ULL, "hlac_lt_0_lowtown_n_-131072_3x.dds"},
	{0x75841819B9773024ULL, "l_lgt241ar_tome_cave.bnk"},
	{0x75849A91D24895CAULL, "pro000_aveline6046486.cl"},
	{0x7584D935064839B7ULL, "w_woundedcoast_137_0.phy"},
	{0x7585700860FBBA13ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_3x.dds"},
	{0x758595A79F72908DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabs.msh"},
	{0x7585A0522DD176B9ULL, "631597880.ogg"},
	{0x7585A30BFC2E8EB7ULL, "pro000_merchant6051274.cl"},
	{0x7585A3A79F72A8C0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabs.mao"},
	{0x7585A7A79F72AF13ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_slabs.mmh"},
	{0x758623439E6D9BF1ULL, "vxca_caverns1_mines_i13_0_0.vlm"},
	{0x75865A1A91FACF4BULL, "406997482.ogg"},
	{0x75874CFBA8EA5D30ULL, "vxlt_gallowstemplar_2j6_0_0.vlm"},
	{0x758766C7D96DDCE3ULL, "prm000im_parth_staff.uti"},
	{0x75878ADA409E075CULL, "gen00fl_anders6201340.cl"},
	{0x7587EFF6664F2431ULL, "6087462_m.fxe"},
	{0x7588576A99FC1268ULL, "vxht_mansion_n_5bm_0_0.vlm"},
	{0x7588919EFD60D327ULL, "hlac_ca_caverns1_mines_24426_7z.dds"},
	{0x7588A0A62B255C9BULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig_e.dds"},
	{0x7589190531E5EB88ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeend.mmh"},
	{0x7589250531E6002BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeend.mao"},
	{0x758935FC09E1E78DULL, "vxlt_0_lowtown_docks_cok_0_0.vlm"},
	{0x7589370531E61EB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeend.msh"},
	{0x758966E5D3EE93D2ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_quanarithrone_n.dds"},
	{0x7589F6D7C852F335ULL, "321373028.ogg"},
	{0x758A0B10EDE57CACULL, "vxht_bartrand_n_6k2_0_0.vlm"},
	{0x758A764406A5B6E0ULL, "vxlt_0_lowtown_a3_n_c_gc7_0_0.vlm"},
	{0x758AA73EFF8CDF0FULL, "536731615.ogg"},
	{0x758AAFF871465F07ULL, "hlac_lt_0_lowtown_a2_-131072_0y.dds"},
	{0x758B348729D83221ULL, "292406729.ogg"},
	{0x758B535E217277C6ULL, "56766732.ogg"},
	{0x758B788196C77017ULL, "hightown_male_noble_gossip6171393.cl"},
	{0x758BA846DBABDC1DULL, "qcr000_arishok6166998.cl"},
	{0x758BB907D3AFCDF3ULL, "832112620.ogg"},
	{0x758BD74E59065727ULL, "vxlt_darktown_5pt_0_0.vlm"},
	{0x758BE1234BD9C0B3ULL, "vxht_0_hightown_c6s_0_0.vlm"},
	{0x758C1E11266EDC83ULL, "gen_im_arm_hlm_lgt_champion.uti"},
	{0x758C62120ECD7B8AULL, "vxht_bartrand_n_5s3_0_0.vlm"},
	{0x758CEBF6A4321DE6ULL, "hlac_ca_caverns1_24424_14z.dds"},
	{0x758CF2FD5D3649EDULL, "w_sundermount_239_0.mmh"},
	{0x758CF6FD5D36503EULL, "w_sundermount_239_0.mao"},
	{0x758D04FD5D366873ULL, "w_sundermount_239_0.msh"},
	{0x758D2CF29F0F331DULL, "root_war_templar.dds"},
	{0x758D5D30B2DDDE1BULL, "dae000_gamlen6145103.cl"},
	{0x758E36D998487B46ULL, "dae000cr_grand_cleric6124789.cl"},
	{0x758F33895366FC86ULL, "vxdr_primevaltomb_2se_0_0.vlm"},
	{0x7590068A85B66330ULL, "vxca_caverns2_8pv_0_0.vlm"},
	{0x75902F5C5882E2BFULL, "295641411.ogg"},
	{0x7590385687169684ULL, "6180967_f.fxe"},
	{0x75903AC927E23AE6ULL, "vxlt_0_lowtown_docks_n_azr_0_0.vlm"},
	{0x75904DFA7A55D9B5ULL, "6169906_m.fxe"},
	{0x7590785E08FE2A9BULL, "primitives_primitive_p1.psh.fxo.dx11"},
	{0x7590D912A782BF0CULL, "vxht_chantry_36p_0_0.vlm"},
	{0x7590FD6C88EBDD48ULL, "6185817_m.fxe"},
	{0x759106C6E729362FULL, "hlac_lt_0_lowtown_docks_a3_-131074_4y.dds"},
	{0x7591B723A1DA5579ULL, "6142567_m.fxe"},
	{0x7591B992F17D9A02ULL, "hlac_lt_0_lowtown_n_131073_2y.dds"},
	{0x759270BE94F1796BULL, "6136298_m.fxe"},
	{0x7592714DD7AD7F9CULL, "6177451_f.fxe"},
	{0x7592EB31C85ADF19ULL, "llac_lt_0_lowtown_a3_n_c_-65538_0.dds"},
	{0x759308DE584CE413ULL, "mrl150_merrill6070801.cl"},
	{0x759397B40C3A0AABULL, "plt_cod_cha_bartrand.plo"},
	{0x7594B9BB052277AFULL, "6127003_m.fxe"},
	{0x75950C36CAEA0D84ULL, "371562787.ogg"},
	{0x75951481FE7069F5ULL, "vxw_bonepit_y2_0_0.vlm"},
	{0x75955F4204AEBE7BULL, "127269242.ogg"},
	{0x7595676A81A45C95ULL, "hro_anvil_crust01d.dds"},
	{0x7595F5FFFDAE1A7EULL, "vxht_playershouse_n_1us_0_0.vlm"},
	{0x75960B9275C64676ULL, "vxlt_0_lowtown_a2_gca_0_0.vlm"},
	{0x75966E18BD2AA0EFULL, "art\\characters\\playercharacter\\humanmale\\hn_boo_lgta_1.mao"},
	{0x759686713326E03AULL, "vxlt_0_lowtown_a2_n_ep5_0_0.vlm"},
	{0x7596B5F4B6F4BF01ULL, "dae000_bodhan_letter6144321.cl"},
	{0x7596D581301E745DULL, "6068291_f.fxe"},
	{0x7596F1196C4D512AULL, "hlac_lt_gallowstemplar_n_2391_2z.dds"},
	{0x7596FADE5A1300C1ULL, "6029036_f.fxe"},
	{0x75979E2A8DAAFFCBULL, "vxw_woundedcoast_12y_0_0.vlm"},
	{0x759816D5EB493D65ULL, "hlac_ca_cavern1_haunted_24424_28y.dds"},
	{0x75986DC46CF6314BULL, "vrc221_varric6040700.cl"},
	{0x7598ACB210DBAE7DULL, "vxca_cavern1_haunted_gko_0_0.vlm"},
	{0x7598B7F54688766AULL, "6035154_m.fxe"},
	{0x75995273EA02AE8AULL, "6156644_m.fxe"},
	{0x7599B170F0853498ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_1x.dds"},
	{0x7599BB3325903EAEULL, "599966837.ogg"},
	{0x759AEC9F31DA79E5ULL, "6041266_m.fxe"},
	{0x759B473BDF8C1561ULL, "gen00fl_anders6119116.cl"},
	{0x759C258FF07609A1ULL, "6016910_m.fxe"},
	{0x759D3617B4389C17ULL, "182476988.ogg"},
	{0x759D4833094944F9ULL, "ave331_post_fight6138789.cl"},
	{0x759D5DC57C0A8A97ULL, "vxht_mansion2_1kk_0_1.vlm"},
	{0x759E3ED310D053AEULL, "hla_ht_0_hightown_a3_0_0y.dds"},
	{0x759E6DE1885F0786ULL, "dae011_wench_mob6149856.cl"},
	{0x759EB1BB2AB97527ULL, "329786347.ogg"},
	{0x759EB9E5C402259BULL, "hf_qun_crazed_warrior.mor"},
	{0x759ED96B7C745052ULL, "vxca_caverns2_91r_0_0.vlm"},
	{0x759F399549560E90ULL, "ser221_gamlen6135070.cl"},
	{0x759F52EBEFE213EFULL, "vxdr_primevaltomb_762_0_0.vlm"},
	{0x759F7963E9E4B669ULL, "llac_ht_mansion2b_818_58.dds"},
	{0x75A02B71A57E5340ULL, "plt_cod_ltr_servantletter.plo"},
	{0x75A02BA1B6E89E08ULL, "vxht_mansion2_1aw_0_0.vlm"},
	{0x75A0685B11A7D7C5ULL, "vxht_bartrand_n_624_0_0.vlm"},
	{0x75A13EC4B2C36364ULL, "6071493_m.fxe"},
	{0x75A16056DABE3906ULL, "vxht_playershouse_n_1pd_0_0.vlm"},
	{0x75A1AE31F0F22C74ULL, "ca_caverns1_24421.rml"},
	{0x75A2361CC08A4A9DULL, "6146804_m.fxe"},
	{0x75A27C5F059C3968ULL, "vxlt_hangedman_1qm_0_0.vlm"},
	{0x75A353021C6DAFF1ULL, "vxlt_undercityserialkiller_1jc_0_1.vlm"},
	{0x75A37EBB7E853667ULL, "hlac_lt_0_lowtown_docks_-65538_0y.dds"},
	{0x75A3F0C3E54BEB5EULL, "mag170_thrask6016943.cl"},
	{0x75A4131154FB760EULL, "6168123_f.fxe"},
	{0x75A421ADF334CE13ULL, "602731846.ogg"},
	{0x75A46F02F0378DE6ULL, "6153988_m.fxe"},
	{0x75A4F84E251CF8BAULL, "vxlt_0_lowtown_a2_n_eu6_0_0.vlm"},
	{0x75A53B247DE84368ULL, "vxht_mansion_n_5hk_0_0.vlm"},
	{0x75A55D12570839E6ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_act_c.anb"},
	{0x75A59BE336ECD98FULL, "6108721_m.fxe"},
	{0x75A619B5EDCE188DULL, "vxlt_0_lowtown_docks_cui_0_0.vlm"},
	{0x75A61A4A818BC198ULL, "116841744.ogg"},
	{0x75A61C5FEE9A805AULL, "vxht_keep_n_69q_0_0.vlm"},
	{0x75A649438E72266DULL, "gen00fl_varric6033636.cl"},
	{0x75A68E4533D64CF2ULL, "597594126.ogg"},
	{0x75A6AB9F0DA292E8ULL, "558960670.ogg"},
	{0x75A6B34DFD31C138ULL, "w_encounter2_camp_42_0_lc.mmh"},
	{0x75A6EECAD1D1ADACULL, "vxht_bartrand_n_6q0_0_0.vlm"},
	{0x75A70B6968290CD4ULL, "1038462817.ogg"},
	{0x75A70C68EC1FC6F8ULL, "vxdr_primevaltomb_7wj_0_0.vlm"},
	{0x75A78719609D6DEFULL, "character_character_p77.psh.fxo.dx11"},
	{0x75A79A1CB3D8ED6BULL, "6121324_m.fxe"},
	{0x75A8ACE485A7FA5FULL, "6182459_m.fxe"},
	{0x75A8D4368746E3FBULL, "46920763.ogg"},
	{0x75A9568842B18AA6ULL, "lgt150pt_nat_werewolf.ncs"},
	{0x75A95FFEE46AE170ULL, "hlac_ht_keep_n_3732_5x.dds"},
	{0x75A9C75F473F7929ULL, "6053513_m.fxe"},
	{0x75A9ED53220E79BFULL, "vxlt_darktown_438_0_0.vlm"},
	{0x75AA215DDB3DDFEEULL, "mag100cr_cullen.utc"},
	{0x75AA2CBB1840F405ULL, "gen_im_wep_rog_arc_act2_04.uti"},
	{0x75AC361E8949942EULL, "gen00fl_merrill6130099.cl"},
	{0x75AC7C9C1E301198ULL, "zz_dre_debug6168292.cl"},
	{0x75ACE0EEBD46C7B0ULL, "vxht_0_hightown_bup_0_0.vlm"},
	{0x75ACE5D39C9EF1F6ULL, "gen00fl_fenris6107202.cl"},
	{0x75AD37D0E3173AFCULL, "vxca_caverns1_mines_ins_0_0.vlm"},
	{0x75ADF6BA5671A01EULL, "hlac_lt_0_lowtown_n_-131072_2y.dds"},
	{0x75ADFD432E9B05DEULL, "hlac_lt_shop_-1_7y.dds"},
	{0x75AEA56DD56BA25EULL, "588429640.ogg"},
	{0x75AECBDC8A12988DULL, "6095375_m.fxe"},
	{0x75AEF506A902F038ULL, "6176640_m.fxe"},
	{0x75AEFCF50ED9680DULL, "240607628.ogg"},
	{0x75AF25CBBDAA0FBEULL, "6116194_m.fxe"},
	{0x75AF8BB985CF0416ULL, "62508087.ogg"},
	{0x75AFD233CCB45186ULL, "vxht_playershouse_n_1we_0_0.vlm"},
	{0x75B0981E5A56F9C7ULL, "hlac_ht_playershouse_n_855_9y.dds"},
	{0x75B0A137658ECC24ULL, "follower_banter6166455.cl"},
	{0x75B0A7F9B280631DULL, "6143704_m.fxe"},
	{0x75B159402ABCBAEFULL, "hlac_lt_0_lowtown_a2_n_c_-131073_0y.dds"},
	{0x75B1D9E97803FDE6ULL, "zz_followers_banter6158421.cl"},
	{0x75B1F83BE25C9AF5ULL, "vxw_bonepit_o0_0_0.vlm"},
	{0x75B2035DF7B2B08AULL, "6087014_m.fxe"},
	{0x75B20BC60CF170FCULL, "follower_banter6136335.cl"},
	{0x75B231E97F139235ULL, "590551828.ogg"},
	{0x75B260125712C891ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_act_c.mmh"},
	{0x75B2858CD220CD04ULL, "6065294_m.fxe"},
	{0x75B3FB19B08C5C0CULL, "6170044_m.fxe"},
	{0x75B47CFDE171A09BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_exit.evt"},
	{0x75B49A33588AA273ULL, "301007286.ogg"},
	{0x75B4D02B1966413FULL, "sb_day.msh"},
	{0x75B4D62B19664BE1ULL, "sb_day.mmh"},
	{0x75B51C096D8601B6ULL, "hlac_lt_0_lowtown_65535_8y.dds"},
	{0x75B6442B3B372482ULL, "hlac_lt_gallowstemplar_n_7198_6z.dds"},
	{0x75B67289CA95D8E9ULL, "gen00fl_aveline6132655.cl"},
	{0x75B68B3888083425ULL, "follower_banter6169981.cl"},
	{0x75B6BFDF72F7D762ULL, "6074359_m.fxe"},
	{0x75B6D3858312BC60ULL, "t3_eey_br1.tnt"},
	{0x75B711539E14D8EFULL, "6094460_m.fxe"},
	{0x75B7F104B471C280ULL, "vxlt_gallowstemplar_5v9_0_0.vlm"},
	{0x75B841C169427B0AULL, "gen00fl_aveline6121148.cl"},
	{0x75B889A4E1C9873DULL, "art\\characters\\weapons\\shields\\textures\\w_shd_seaa_0n.dds"},
	{0x75B8A5F29562BE66ULL, "6110398_m.fxe"},
	{0x75B9347A40C481F5ULL, "vxlt_0_lowtown_n_f16_0_0.vlm"},
	{0x75B95680B82F0AC7ULL, "62418772.ogg"},
	{0x75BA4E54F92414B2ULL, "340624333.ogg"},
	{0x75BB75FDE177B3DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_exit.gad"},
	{0x75BB8AC968AFCD26ULL, "vxlt_gallowscourtyard_n_6z1_0_0.vlm"},
	{0x75BBA9ADD50D537CULL, "vxca_caverns1_mines_ikp_0_0.vlm"},
	{0x75BBF4F12C599DCCULL, "879602116.ogg"},
	{0x75BC08D3510B9A47ULL, "6076231_m.fxe"},
	{0x75BC6E7A8CCC30E1ULL, "zz_qun_debug6028284.cl"},
	{0x75BD4C14F59408C5ULL, "vxht_bartrand_n_682_0_0.vlm"},
	{0x75BDDFEE3154B66CULL, "w_encounter3_106_0.msh"},
	{0x75BDE5EE3154C05EULL, "w_encounter3_106_0.mmh"},
	{0x75BDF1EE3154D4F5ULL, "w_encounter3_106_0.mao"},
	{0x75BE30A23DC36BEFULL, "511181299.ogg"},
	{0x75BF085AEB07914FULL, "6035949_f.fxe"},
	{0x75BF0B3002A7BA52ULL, "art\\characters\\playercharacter\\humanmale\\arm_masc.phy"},
	{0x75BF1D1E2FC3D390ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec02b.phy"},
	{0x75BF55DD93627AD2ULL, "hlac_lt_darktown_-1_82x.dds"},
	{0x75BF5EADD2EB8146ULL, "6072336_f.fxe"},
	{0x75BF7229E680A27FULL, "689633123.ogg"},
	{0x75BF74CD93F6555CULL, "6125939_f.fxe"},
	{0x75BFDFA251416BAFULL, "zz_applyareavfx.ncs"},
	{0x75C0D9D668CFC0D8ULL, "6038645_m.fxe"},
	{0x75C0F4CF38ACE337ULL, "39384417.ogg"},
	{0x75C166E0DE41B20DULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\til_ruins01blight_d.dds"},
	{0x75C174FDE17C1DDAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_left_exit.ani"},
	{0x75C1BE26E4D62883ULL, "gen_im_trs_gen_sta_cld_r0.uti"},
	{0x75C1CD1360B6F3EAULL, "vxlt_gallowscourtyard_n_7eq_0_0.vlm"},
	{0x75C2521A3427DE4DULL, "6115426_m.fxe"},
	{0x75C2F5EB5E2BB79BULL, "ser120st_ghyslain.stg"},
	{0x75C300BF8D7DCFE0ULL, "hlac_lt_foundry_n_320_22y.dds"},
	{0x75C34131911493B6ULL, "art\\levels\\races\\proxy\\_textures\\ter_2_d.dds"},
	{0x75C35876757827ABULL, "157721785.ogg"},
	{0x75C37BC9FB0DB9C5ULL, "w_bonepit_12_0_lc.mmh"},
	{0x75C38280BFE5BDDDULL, "6060393_m.fxe"},
	{0x75C4BD556C6E89D9ULL, "6172937_m.fxe"},
	{0x75C511C060EAFC0DULL, "dae000_arishok6081349.cl"},
	{0x75C531BC64EDEFE2ULL, "6079059_m.fxe"},
	{0x75C5BF1BD8BCD668ULL, "vxlt_gallowsprison_n_5qg_0_0.vlm"},
	{0x75C62CE669E5689CULL, "art\\cinematics\\_crit path\\followerintros\\cs_bartender_isa_intro.gad"},
	{0x75C6C0C3E8702DBCULL, "6088228_m.fxe"},
	{0x75C6CF5189F5F091ULL, "499866289.ogg"},
	{0x75C6F21D76916005ULL, "vxdr_primevaltomb_7zt_0_0.vlm"},
	{0x75C7895D53E8D9DFULL, "vxlt_0_lowtown_docks_a3_n_cyg_0_0.vlm"},
	{0x75C7E329D2B9A8DCULL, "6110558_m.fxe"},
	{0x75C82286BC3AE406ULL, "hlac_ht_keep_4519_8x.dds"},
	{0x75C82BDC78B7DCF6ULL, "113869345.ogg"},
	{0x75C851F6DA85EAAEULL, "88827263.ogg"},
	{0x75C8A50B070A9540ULL, "602193326.ogg"},
	{0x75C8EEB2DB2DD5F5ULL, "vxca_caverns2prc_9sj_0_0.vlm"},
	{0x75C98DAF78460E21ULL, "one100_ht_banter6060901.cl"},
	{0x75CA1EB1FF3C8C54ULL, "vxht_0_hightown_auo_0_0.vlm"},
	{0x75CA83572ED9D857ULL, "vxht_chantry_n_287_0_0.vlm"},
	{0x75CAC138FF438F22ULL, "zz_fol_debug6103796.cl"},
	{0x75CB288C7CC85525ULL, "6164691_m.fxe"},
	{0x75CB61586F2C1057ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medp_0s.dds"},
	{0x75CB8EC80D06D3B5ULL, "vxw_bonepitmining_a2_179_0_0.vlm"},
	{0x75CBAE350683F30FULL, "w_woundedcoast_137_0.mao"},
	{0x75CBBA35068407ACULL, "w_woundedcoast_137_0.mmh"},
	{0x75CBC0350684119EULL, "w_woundedcoast_137_0.msh"},
	{0x75CBF04EAABDDBE2ULL, "vxht_mansion2b_1b0_0_0.vlm"},
	{0x75CC0213141F07CDULL, "6137004_m.fxe"},
	{0x75CC0D6E7C0299CCULL, "436921287.ogg"},
	{0x75CC33FC68A2D2D5ULL, "1035126331.ogg"},
	{0x75CC3AAFD5D2F0E5ULL, "770262016.ogg"},
	{0x75CCA2EC62B77815ULL, "vxht_mansion_n_6zs_0_0.vlm"},
	{0x75CCD4E80C159C0CULL, "176708058.ogg"},
	{0x75CD2BE669EB859AULL, "art\\cinematics\\_crit path\\followerintros\\cs_bartender_isa_intro.ani"},
	{0x75CDD3ABB04790DCULL, "vxca_caverns2_8ok_0_0.vlm"},
	{0x75CDE2A0FB258C88ULL, "219847276.ogg"},
	{0x75CE202C3B5236A7ULL, "99819328.ogg"},
	{0x75CE52F059B5CF77ULL, "dae150_chandan6148644.cl"},
	{0x75CE72F40C013865ULL, "plt_cod_bks_dalish_history.plo"},
	{0x75CE83E2843339CAULL, "540100264.ogg"},
	{0x75CEFCA939C1FBA2ULL, "vxlt_undercityserialkiller_1kb_0_0.vlm"},
	{0x75CF437283CE7A19ULL, "6076646_m.fxe"},
	{0x75D03EF052AD0C6AULL, "vxlt_gallowscourtyard_n_7hr_0_0.vlm"},
	{0x75D04BFA07EFA91AULL, "6033596_m.fxe"},
	{0x75D08AD5DFB181C0ULL, "vxlt_0_lowtown_byd_0_0.vlm"},
	{0x75D0A23D994ED6E5ULL, "follower_barks6153411.cl"},
	{0x75D11A525533A46BULL, "mer210_coterie_contact6036765.cl"},
	{0x75D184AF79960EAAULL, "6103942_f.fxe"},
	{0x75D1D35837F12676ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_prostitute_2.evt"},
	{0x75D2806BB3C2C17BULL, "121611377.ogg"},
	{0x75D285649C7BED7DULL, "llac_w_woundedcoast_6_0.dds"},
	{0x75D2F333AFFC4D1AULL, "vxlt_gallowstemplar_n_2ol_0_0.vlm"},
	{0x75D32939E6A5BB63ULL, "fen110ip_to_alienage.utp"},
	{0x75D3837BFE2ECD68ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masf_0d.dds"},
	{0x75D3A28D66625130ULL, "qcr221_cut_diplomatic_sol6201285.cl"},
	{0x75D3A96CF2F5FD8CULL, "vxlt_0_lowtown_docks_a3_n_9zs_0_0.vlm"},
	{0x75D3C6246C2B7230ULL, "lgt200cr_gang_3_ht.utc"},
	{0x75D417CFE14A352CULL, "6048267_f.fxe"},
	{0x75D41D40904D9FE7ULL, "llac_lt_backalley_g_65537_7.dds"},
	{0x75D42B5DC506EBB9ULL, "llac_lt_gallowscourtyard_-2_0.dds"},
	{0x75D4A0BB4A5EACA8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd12_basic.mmh"},
	{0x75D4ACBB4A5EC14BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd12_basic.mao"},
	{0x75D4BEBB4A5EDFD2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0dd12_basic.msh"},
	{0x75D52436C64B6726ULL, "arm_sla.tnt"},
	{0x75D532A0DA66463CULL, "6083328_m.fxe"},
	{0x75D563FA68877885ULL, "vxdr_primevaltomb_7ws_0_0.vlm"},
	{0x75D56A0C452F31FFULL, "vxlt_0_lowtown_n_e2n_0_0.vlm"},
	{0x75D5ADD9DF109912ULL, "674776100.ogg"},
	{0x75D5CB6F2075C64DULL, "vxlt_backalley_g_1uc_0_0.vlm"},
	{0x75D60EC51239BADCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_en.ani"},
	{0x75D6516B90FBC42CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side.ani"},
	{0x75D6A938F6DD1294ULL, "vxw_woundedcoast_5f_0_0.vlm"},
	{0x75D764C6B7969F02ULL, "dae000_brothel_gamlen6157076.cl"},
	{0x75D79892BA774DECULL, "426179458.ogg"},
	{0x75D7BBCC8DD8AD26ULL, "6075424_m.fxe"},
	{0x75D7C787CEDB42E7ULL, "qun230cr_guild_contact6068879.cl"},
	{0x75D7CA5837F58201ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_prostitute_2.gad"},
	{0x75D82DB38000B324ULL, "hlac_lt_house_-1_2z.dds"},
	{0x75D834847F25F409ULL, "6187934_f.fxe"},
	{0x75D8A410AFC570FDULL, "983024478.ogg"},
	{0x75D8B3C296089C8EULL, "gen00fl_fenris6115331.cl"},
	{0x75D9CA3EB02F9455ULL, "zz_mcr_debug6080824.cl"},
	{0x75DA3D09D4401E43ULL, "457835296.ogg"},
	{0x75DA96CDD7401FD4ULL, "zz_mer_debug6045633.cl"},
	{0x75DAEE6291C4680EULL, "420678148.ogg"},
	{0x75DBC124A627C098ULL, "zz_mer_debug6039340.cl"},
	{0x75DC077538F90548ULL, "vxlt_hangedman_n_1vn_0_1.vlm"},
	{0x75DC1BC5123E3C1EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_en.gad"},
	{0x75DC3C32F5E1DCDCULL, "vxht_0_hightown_bel_0_0.vlm"},
	{0x75DCB3C56391A861ULL, "6122532_m.fxe"},
	{0x75DCD38AAE3C4393ULL, "6118625_m.fxe"},
	{0x75DD38F44272D8F4ULL, "vxlt_0_lowtown_a3_n_c_eoy_0_0.vlm"},
	{0x75DD51E6B6EDF691ULL, "6119493_m.fxe"},
	{0x75DD5520875841DAULL, "109903004.ogg"},
	{0x75DD5E6B9101F96EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side.gad"},
	{0x75DD7A24AFFE3047ULL, "6086200_f.fxe"},
	{0x75DE0085DD6F6C21ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gong_d.dds"},
	{0x75DE56C758F8493BULL, "vxht_mansion2b_1tk_0_0.vlm"},
	{0x75DE9A231D4EE465ULL, "mag311_emile_de_launcet6064726.cl"},
	{0x75DEBA96F49C3F08ULL, "452480625.ogg"},
	{0x75DECB5837FBA2AFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_prostitute_2.ani"},
	{0x75DF3EEFEFB5CA3BULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd01a.msh"},
	{0x75DF5CEFEFB5FD25ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd01a.mmh"},
	{0x75DF9EF7808E2987ULL, "552059874.ogg"},
	{0x75E00375AC221C89ULL, "dae100_gamlen_letter6146986.cl"},
	{0x75E02C8BA1CBC9E3ULL, "lt_0_lowtown_docks_a3_242_0.msh"},
	{0x75E03A8BA1CBE13DULL, "lt_0_lowtown_docks_a3_242_0.mmh"},
	{0x75E03E8BA1CBE84EULL, "lt_0_lowtown_docks_a3_242_0.mao"},
	{0x75E0B0213A6AB807ULL, "857610844.ogg"},
	{0x75E2EAE935392251ULL, "869144715.ogg"},
	{0x75E2F6C512441C3DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_manimal_en.evt"},
	{0x75E393EC5049DA89ULL, "static_static_p2.psh.fxo"},
	{0x75E3A2EC1FC8CB3DULL, "static_static_p35.psh.fxo"},
	{0x75E3D3B628EC5065ULL, "ave121_brennan6105901.cl"},
	{0x75E43D4C126BDECDULL, "vxlt_backalley_g_1rb_0_0.vlm"},
	{0x75E4596B91080F4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side.evt"},
	{0x75E4CBE57E4046D1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_exit.ani"},
	{0x75E5031D4F17A0B1ULL, "193575421.ogg"},
	{0x75E51B15E8D32B14ULL, "vxw_woundedcoast_2g_0_0.vlm"},
	{0x75E51C44FDA33D9FULL, "hlac_ht_0_hightown_a2_-196607_0x.dds"},
	{0x75E5F3E556F21AD2ULL, "376356369.ogg"},
	{0x75E66F760FBDF342ULL, "t1_tat_dpu.tnt"},
	{0x75E697D5B3561185ULL, "vxlt_undercityencounter_141_0_1.vlm"},
	{0x75E78BB10DFF28BCULL, "551444397.ogg"},
	{0x75E7A320218B2CFAULL, "257041928.ogg"},
	{0x75E7D317262FD023ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanl.ani"},
	{0x75E869805B3C280DULL, "6040540_m.fxe"},
	{0x75E8A60CF3176C0AULL, "hlac_lt_0_lowtown_docks_n_131071_3y.dds"},
	{0x75E8C06CE12A92A3ULL, "zz_followers_banter6158178.cl"},
	{0x75E8C6CB986FB2D7ULL, "mag371_elsa6064654.cl"},
	{0x75E8D4088EAA0CE2ULL, "vxht_mansion2b_1x2_0_0.vlm"},
	{0x75E922A20F7C9ACDULL, "cs_seb000cs_sebastian_intro.wbd"},
	{0x75E947884694BED2ULL, "dae000_pc_home_ambients6208600.cl"},
	{0x75E9C097B7C1C866ULL, "6136237_m.fxe"},
	{0x75EA1A01EEC551FCULL, "7213328.ogg"},
	{0x75EA69A1B6335133ULL, "894536569.ogg"},
	{0x75EA761DDA4E36DCULL, "hlac_lt_0_lowtown_a2_n_-1_8z.dds"},
	{0x75EA7BB0690EB5C9ULL, "vxca_cavern1_haunted_dr7_0_0.vlm"},
	{0x75EAD6E57E44C573ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_exit.gad"},
	{0x75EB80A6CFAEAD17ULL, "hlac_w_encounter2_camp_327678_0x.dds"},
	{0x75EBD7D107E9C8C4ULL, "447950702.ogg"},
	{0x75EC1FD5288B5F65ULL, "6034198_m.fxe"},
	{0x75EC92648FCFA45FULL, "6117804_m.fxe"},
	{0x75ED06136717F1CAULL, "vxlt_0_lowtown_n_er7_0_0.vlm"},
	{0x75ED3C2AB3C79051ULL, "dae300_alistair_king6133793.cl"},
	{0x75EF80A7E34BB2A6ULL, "art\\characters\\creatures\\dragon\\animation\\tal_rake.ani"},
	{0x75EF8A4F49643778ULL, "vxlt_0_lowtown_a3_n_c_gxr_0_0.vlm"},
	{0x75F1A250763AB580ULL, "183929732.ogg"},
	{0x75F1BC0030876FB8ULL, "vxlt_0_lowtown_docks_a3_9a1_0_0.vlm"},
	{0x75F1D3E57E4ADED8ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_exit.evt"},
	{0x75F1EBEEC0F19A51ULL, "272635583.ogg"},
	{0x75F33D51449965B5ULL, "6118195_m.fxe"},
	{0x75F39B0CB81E6036ULL, "attributes.gda"},
	{0x75F3C48BEB43270CULL, "lgt000cr_dwf_turn_in_6.utc"},
	{0x75F41984325B8C3AULL, "mrl150_merrill6085842.cl"},
	{0x75F41F6995CB59BCULL, "390393324.ogg"},
	{0x75F448827ED24CAEULL, "102791594.ogg"},
	{0x75F4F653AE2EE1F3ULL, "vxca_caverns2prc_8h0_0_0.vlm"},
	{0x75F585CF416297D7ULL, "dre000_bethany.bnk"},
	{0x75F5A1D4DCC89380ULL, "521762256.ogg"},
	{0x75F5EB17263C361AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanl.evt"},
	{0x75F61994565FDB49ULL, "vxlt_0_lowtown_f6p_0_0.vlm"},
	{0x75F625283BD092B1ULL, "vxlt_0_lowtown_docks_bgn_0_0.vlm"},
	{0x75F69C0EFB91E5E4ULL, "vxlt_0_lowtown_docks_a3_n_d53_0_0.vlm"},
	{0x75F779B4EF42E00EULL, "187488264.ogg"},
	{0x75F7D876A04CB803ULL, "vxw_encounter1_it_0_1.vlm"},
	{0x75F81085A90FAD4DULL, "w_gax_ga01a_0.msh"},
	{0x75F867CF41649AEFULL, "dre000_bethany.cnv"},
	{0x75F8C67D5603493DULL, "vxht_0_hightown_c8c_0_0.vlm"},
	{0x75F8E11A35E14F40ULL, "325723529.ogg"},
	{0x75F90F31E5624C79ULL, "art\\characters\\creatures\\shriek\\animation\\blendtree\\shriek_overlay.anb"},
	{0x75F9EC96F177F27EULL, "vxw_bonepit_1xh_0_1.vlm"},
	{0x75FA35A09ADA2791ULL, "6113293_m.fxe"},
	{0x75FA54A311FB2BD4ULL, "t3_gsp_qun.tnt"},
	{0x75FB77F0590E0A4AULL, "vxlt_0_lowtown_n_eo0_0_0.vlm"},
	{0x75FC68A7E3561417ULL, "art\\characters\\creatures\\dragon\\animation\\tal_rake.evt"},
	{0x75FCE0F8F8D4DC55ULL, "hlac_ht_0_hightown_a3_65533_2y.dds"},
	{0x75FCEA172642530DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf_injured_leanl.gad"},
	{0x75FD54A6524C5544ULL, "340310995.ogg"},
	{0x75FD9F718F015677ULL, "hf_gen_bloodmage.mor"},
	{0x75FDCF5E6BF2C418ULL, "gen00fl_dog6167638.cl"},
	{0x75FDD0EEBFAD81F2ULL, "6110831_m.fxe"},
	{0x75FDD123981F32C5ULL, "hlac_ca_caverns1_mines_24421_4x.dds"},
	{0x75FE0675962B9C3DULL, "235830611.ogg"},
	{0x75FE3356F9167060ULL, "6171094_m.fxe"},
	{0x75FE4F9C29AA4D5AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn03.mmh"},
	{0x75FE510B4D842A26ULL, "6150682_m.fxe"},
	{0x75FE54C1E6A25F9EULL, "vxht_keep_n_620_0_0.vlm"},
	{0x75FE619C29AA6BA0ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn03.msh"},
	{0x75FEBA62CD312E42ULL, "90822829.ogg"},
	{0x75FEDFE964DF17EEULL, "6075276_f.fxe"},
	{0x75FEE48C465ABF8DULL, "art\\levels\\races\\elf\\props\\_textures\\prp_vase_d.dds"},
	{0x75FEF4407F1BDBA5ULL, "seb000_install.ncs"},
	{0x75FF0823E0F017B4ULL, "dae011_wench_mob6144866.cl"},
	{0x75FF0DFAF899C096ULL, "mag100_karras6034699.cl"},
	{0x75FF97EF7757D888ULL, "mag120_macha6080674.cl"},
	{0x75FFC333BC438E68ULL, "vxw_encounter2_camp_9s_0_0.vlm"},
	{0x7600FCDC9E572FD6ULL, "1028362916.ogg"},
	{0x76018B45588D9554ULL, "199034840.ogg"},
	{0x7601CCDB385DAE65ULL, "196311240.ogg"},
	{0x760275D09F556CA5ULL, "dae000_fem_commoner2_mob6174585.cl"},
	{0x76028B427E0EA97DULL, "gen00flst_sebacttwob.stg"},
	{0x7602CCF20B57983FULL, "6169834_m.fxe"},
	{0x7602EB9F33D7AB30ULL, "vxdr_primevalentrance_189_0_0.vlm"},
	{0x7603039B1D9FFCDBULL, "gencr_spider_corrupted.utc"},
	{0x760369A7E35C34A0ULL, "art\\characters\\creatures\\dragon\\animation\\tal_rake.gad"},
	{0x7603AE67600EC0F4ULL, "lt_0_lowtown_docks_n_28_0.rml"},
	{0x7604019050AD3B3CULL, "gen00fl_merrill6107873.cl"},
	{0x76044B3439EA4036ULL, "569603124.ogg"},
	{0x76049E708622E104ULL, "168266337.ogg"},
	{0x7604B197BDF24013ULL, "dae000_templar_knight6165179.cl"},
	{0x760535273A01B97BULL, "6174784_m.fxe"},
	{0x760640085C175610ULL, "6164478_m.fxe"},
	{0x760645F455EBD85AULL, "w_bonepit_219_0.msh"},
	{0x760653F455EBEF93ULL, "w_bonepit_219_0.mao"},
	{0x760657F455EBF6A0ULL, "w_bonepit_219_0.mmh"},
	{0x7606C465B31AFB34ULL, "gen00fl_player6207814.cl"},
	{0x7607443339F725A6ULL, "6163114_m.fxe"},
	{0x76077AFA0D49D2C5ULL, "hm_comte_launc.mor"},
	{0x7607B3B7584E303FULL, "w_bonepitmining_a2_78_0.msh"},
	{0x7607B9B7584E3AE1ULL, "w_bonepitmining_a2_78_0.mmh"},
	{0x7607C5B7584E4E82ULL, "w_bonepitmining_a2_78_0.mao"},
	{0x7607F4D1A5DF9960ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzsteps.phy"},
	{0x7608390D8FF226E6ULL, "6126820_f.fxe"},
	{0x76086F2BDEF38DAEULL, "w_woundedcoast_102_0.msh"},
	{0x76087D2BDEF3A5FFULL, "w_woundedcoast_102_0.mao"},
	{0x7608892BDEF3B99CULL, "w_woundedcoast_102_0.mmh"},
	{0x7608AD9E9799209BULL, "846705029.ogg"},
	{0x7608EA3388502E72ULL, "6149907_m.fxe"},
	{0x7608F895F148E0ADULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1b.msh"},
	{0x76090695F148F8E0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1b.mao"},
	{0x76090A95F148FF33ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1b.mmh"},
	{0x76097F8E9DD0FCF8ULL, "art\\vfx\\combat\\vfx_f_fen_spirit_pulse_c.anb"},
	{0x7609AE138C3E2267ULL, "6142213_m.fxe"},
	{0x7609E15A37504402ULL, "6117522_m.fxe"},
	{0x7609FFEB208C9D6BULL, "6111543_m.fxe"},
	{0x760A6729EAF95F66ULL, "qun131_party_comment6117304.cl"},
	{0x760A728306CA2832ULL, "vxlt_0_lowtown_a2_n_fiy_0_0.vlm"},
	{0x760A8FC8C6846D74ULL, "481293515.ogg"},
	{0x760A9FB9A8EBC6BFULL, "2p_200_sitting.stg"},
	{0x760AAC128BA7CD7DULL, "gen00fl_aveline6140786.cl"},
	{0x760AB10624CE3565ULL, "vxht_keep_n_60u_0_0.vlm"},
	{0x760AEF3161DA84C4ULL, "vxht_mansion2_287_0_0.vlm"},
	{0x760AFA9C3003F56CULL, "vxlt_warehouse_n_q5_0_0.vlm"},
	{0x760AFEDD80E3E06AULL, "vxca_caverns1_iz8_0_0.vlm"},
	{0x760B05E8FCB7CD62ULL, "ser121_jethann6107102.cl"},
	{0x760B3D2D9DC24267ULL, "vxht_keep_n_55g_0_0.vlm"},
	{0x760BAD05ABB49704ULL, "6176525_m.fxe"},
	{0x760BD50CC8B74B91ULL, "680227792.ogg"},
	{0x760CA533EB0C88E0ULL, "6174394_m.fxe"},
	{0x760D72DDEE63548BULL, "6090446_m.fxe"},
	{0x760D7D867C9CDC5DULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc101a_0d.dds"},
	{0x760DAF6D38B8DEF4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_kick_atk.ani"},
	{0x760DFAAA91421013ULL, "6050427_f.fxe"},
	{0x760E3510AE39A6E8ULL, "vxw_encounter2_camp_4p_0_0.vlm"},
	{0x760EC9ABB9F70C6CULL, "171751886.ogg"},
	{0x760ED30DFF60387BULL, "vxdr_0_deeproads_ebz_0_0.vlm"},
	{0x760F5600D14C7535ULL, "151183930.ogg"},
	{0x760F829942032F9AULL, "hlac_ht_0_hightown_-196608_1y.dds"},
	{0x760F844411EDC218ULL, "follower_barks6153193.cl"},
	{0x76105FBCD0E1DA2DULL, "vxht_chantry_336_0_0.vlm"},
	{0x761086D86946783CULL, "hlac_w_bonepitmining_a2_65536_0y.dds"},
	{0x761197AB96E73AE7ULL, "vxht_0_hightown_n_alp_0_0.vlm"},
	{0x7611DF152C8FE341ULL, "cs_rom000cs_isabella.wbd"},
	{0x7611EB0483D4F1A5ULL, "53994294.ogg"},
	{0x76125802B90B06D4ULL, "vxht_0_hightown_alp_0_0.vlm"},
	{0x7612EE94A42185A0ULL, "pro000_refugee_male6052107.cl"},
	{0x7612F324954EA2BAULL, "vxlt_0_lowtown_docks_n_bvm_0_0.vlm"},
	{0x76133CAD59E85C79ULL, "lightning_stone_2.dds"},
	{0x76136341225D6826ULL, "313293061.ogg"},
	{0x76139AAC5FC15316ULL, "vxdr_0_deeproads_f4z_0_0.vlm"},
	{0x7613AB32EB76E6ADULL, "765929106.ogg"},
	{0x7613AC6D38BD4556ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_kick_atk.gad"},
	{0x7613D246683C96F1ULL, "297005691.ogg"},
	{0x76158ECAFE35AA92ULL, "919387904.ogg"},
	{0x7615AA3739EF7A3DULL, "vxht_0_hightown_c2a_0_0.vlm"},
	{0x7615CA933AE914FEULL, "dae021_chantry_initiate_mob6147590.cl"},
	{0x7615ED414714DB0FULL, "vxht_0_hightown_n_c1u_0_0.vlm"},
	{0x76165B3EE5B426D9ULL, "hm_generic_zom_a.mor"},
	{0x7616D94F6435CDCBULL, "llac_lt_0_lowtown_docks_a3_n_-65538_5.dds"},
	{0x7616EDB637AA72D5ULL, "gen_im_wep_war_one_act3_01.uti"},
	{0x761738500E1D157BULL, "hlac_lt_0_lowtown_a3_n_c_-2_3z.dds"},
	{0x761811D452F3A450ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_cass.mmh"},
	{0x76181FD452F3BC6AULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_cass.msh"},
	{0x7618619A3A927F9AULL, "417869029.ogg"},
	{0x7618A4994520E990ULL, "static_static_p18.psh.fxo.dx11.assembly"},
	{0x7618B8CAC7B3E43FULL, "hlac_ht_0_hightown_a2_1_1x.dds"},
	{0x7618DF178AE54D4AULL, "6036656_m.fxe"},
	{0x7618E45FF8C040B2ULL, "vxlt_0_lowtown_a2_n_fdz_0_0.vlm"},
	{0x761967819DA9D734ULL, "hlac_ht_bartrand_n_3365_0x.dds"},
	{0x761971E2B5CD0DAAULL, "w_blightlands_57_0.mao"},
	{0x761975E2B5CD14F9ULL, "w_blightlands_57_0.mmh"},
	{0x76197FE2B5CD25E7ULL, "w_blightlands_57_0.msh"},
	{0x761A17B83DD898C3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robc_0n.dds"},
	{0x761A2B6BA71FE4BDULL, "6088352_f.fxe"},
	{0x761A41F67657221DULL, "vxw_blightlands_10k_0_0.vlm"},
	{0x761A9BFA763B5019ULL, "pro000_friedrich6050196.cl"},
	{0x761AC76D38C391B5ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_kick_atk.evt"},
	{0x761B2621A829DD4AULL, "dae000ip_crowd_lm5_talk14.utp"},
	{0x761B89FC81480F01ULL, "vxht_0_hightown_bk3_0_0.vlm"},
	{0x761BEDFC62173EBAULL, "6142318_f.fxe"},
	{0x761C898894A714F1ULL, "gen00fl_sebastian6195805.cl"},
	{0x761CF93A9220D4E8ULL, "qun131_party_comment6117276.cl"},
	{0x761CFAC18E73AE59ULL, "6048508_m.fxe"},
	{0x761D26A9E831E079ULL, "6121087_m.fxe"},
	{0x761D8BBAA5F891EAULL, "225882184.ogg"},
	{0x761DE1E0C5CBE905ULL, "vxlt_0_lowtown_a3_n_c_gxk_0_0.vlm"},
	{0x761E7ACFF2B97AC3ULL, "vxht_bartrand_n_5va_0_0.vlm"},
	{0x761E8B63F6A30D23ULL, "hlac_lt_0_lowtown_a3_n_c_65535_9y.dds"},
	{0x761F40479309C954ULL, "816632125.ogg"},
	{0x761F8097A0E584CBULL, "mag210cr_feynriel_abominat.utc"},
	{0x7620163CE7244E10ULL, "vxlt_hangedman_25r_0_0.vlm"},
	{0x762027BF6E518F1BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_action.evt"},
	{0x7620F337FCF3AC9EULL, "167325921.ogg"},
	{0x762165018744BB3AULL, "vxlt_0_lowtown_docks_n_bsn_0_0.vlm"},
	{0x762166446D69932DULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg3a.mmh"},
	{0x762178446D69B1B3ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg3a.msh"},
	{0x7621B20982F75D8EULL, "6083582_m.fxe"},
	{0x7621B61CEB788382ULL, "dre000_sibling_has_blight6036522.cl"},
	{0x7621E32FBB49FBC1ULL, "vxlt_gallowstemplar_f_5yz_0_0.vlm"},
	{0x76225725997710EAULL, "602463892.ogg"},
	{0x7622A2216220C6CCULL, "222990000.ogg"},
	{0x7622B93B3D239D6EULL, "hlac_lt_0_lowtown_n_0_2z.dds"},
	{0x762334605BEF58EEULL, "1073159185.ogg"},
	{0x76237FD00DE89B34ULL, "6084390_m.fxe"},
	{0x7623D326912CFE05ULL, "6109430_m.fxe"},
	{0x7623EC10FC4F40E6ULL, "14109201.ogg"},
	{0x76241CEC1DCE7992ULL, "194303071.ogg"},
	{0x7624260BB17D1D03ULL, "qun221_viscount6069911.cl"},
	{0x762448B52B6AAC61ULL, "651181654.ogg"},
	{0x76244C88F9075C17ULL, "713411118.ogg"},
	{0x762457985B14F9D2ULL, "fx_blue_line.dds"},
	{0x76246E9C29CB5165ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn03.phy"},
	{0x76258723720D5EF5ULL, "vxw_bonepit_78_0_0.vlm"},
	{0x7625B3FDE0235D93ULL, "114013816.ogg"},
	{0x7625DA77451D8E0DULL, "lt_0_lowtown_a2_34_0_ll.mmh"},
	{0x7625F54109F1617CULL, "39120890.ogg"},
	{0x762635EFEFF1BDD2ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd01a.phy"},
	{0x7626649ACF5CBF4DULL, "plt_tbe100pt_mirror.plo"},
	{0x7626F78B8509ED8BULL, "vxca_caverns1_mines_ee1_0_0.vlm"},
	{0x7627102F6D8EE483ULL, "art\\characters\\playercharacter\\children\\textures\\kf_cth_grla_0n.dds"},
	{0x762720BF6E57A25CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_action.gad"},
	{0x7627D0E33AC57DE9ULL, "259373998.ogg"},
	{0x7628038B7EBC1C4FULL, "453073298.ogg"},
	{0x762844047C3482C0ULL, "6179375_m.fxe"},
	{0x7628D86F8C246ABFULL, "art\\levels\\races\\design\\trap_floorlever.msh"},
	{0x7628DE6F8C247561ULL, "art\\levels\\races\\design\\trap_floorlever.mmh"},
	{0x7628EA6F8C248902ULL, "art\\levels\\races\\design\\trap_floorlever.mao"},
	{0x762918F7145D982EULL, "l_seb321ar_keep.bnk"},
	{0x762920BA2BF7164CULL, "889693362.ogg"},
	{0x762A21009F6DC38FULL, "6049440_m.fxe"},
	{0x762A2CA20FB442F7ULL, "cs_seb000cs_sebastian_intro.bnk"},
	{0x762B6C9E8069C6D9ULL, "6043408_m.fxe"},
	{0x762C18548CFC2EABULL, "dae000_tomwise6167282.cl"},
	{0x762C40AAA848252FULL, "art\\levels\\races\\proxy\\props\\_textures\\cheese_n.dds"},
	{0x762CB028E86BFE19ULL, "mcr370_negotiation6080210.cl"},
	{0x762D1FBF6E5C0C5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\handshake_excited_action.ani"},
	{0x762D24058F2D9EDCULL, "vxlt_gallowstemplar_5sz_0_0.vlm"},
	{0x762DC6BC8F65A224ULL, "hlac_ht_0_hightown_a3_-65534_1y.dds"},
	{0x762E11D1A60099F5ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzsteps.msh"},
	{0x762E13D1A6009D6BULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzsteps.mmh"},
	{0x762E1FD1A600B108ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzsteps.mao"},
	{0x762E682BDF145167ULL, "w_woundedcoast_102_0.phy"},
	{0x762E8819D91A6690ULL, "vxlt_hangedman_22s_0_0.vlm"},
	{0x762EBD3E28024E68ULL, "art\\characters\\creatures\\dragon\\animation\\blendtree\\dragon_movement.anb"},
	{0x762F2741EA10DDACULL, "art\\characters\\creatures\\ragedemon\\animation\\wkf.ani"},
	{0x762F6C773EDD6B86ULL, "pro000_colonel_ewald6047393.cl"},
	{0x762F7839D8C9E08CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_enter.ani"},
	{0x762FBB6BCE85B44FULL, "122642118.ogg"},
	{0x762FCF67603433F4ULL, "lt_0_lowtown_docks_n_28_0.anb"},
	{0x763097A64C231B10ULL, "6043733_m.fxe"},
	{0x76317E8E9DF24C93ULL, "art\\vfx\\combat\\vfx_f_fen_spirit_pulse_c.mmh"},
	{0x7631FBDD0686E912ULL, "6201278_m.fxe"},
	{0x76320D736AEA0CFAULL, "vxca_cave_t2_0_1.vlm"},
	{0x76322DA5E3CCB939ULL, "6097355_m.fxe"},
	{0x7632450383231685ULL, "6102530_m.fxe"},
	{0x76325F446D77FAEAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg3a.phy"},
	{0x76327B9938C273C5ULL, "6122107_m.fxe"},
	{0x76328296018D245BULL, "810811143.ogg"},
	{0x76329351667789EEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bush_dp.dds"},
	{0x763297DAFF1614ACULL, "gen00fl_varric6089645.cl"},
	{0x7632C74AB70CCE9AULL, "vxlt_0_lowtown_n_exq_0_0.vlm"},
	{0x7633226DA01D2362ULL, "103221946.ogg"},
	{0x7633A3D331840C00ULL, "6104601_m.fxe"},
	{0x763406741CF0FDC1ULL, "llac_ht_keep_n_3731_1.dds"},
	{0x76346C1EB9F63207ULL, "vxht_brothel_n_5jy_0_0.vlm"},
	{0x763475B6C0D090A2ULL, "6095138_m.fxe"},
	{0x7634760666CF0F31ULL, "6133112_m.fxe"},
	{0x7634E435277850CEULL, "vxht_0_hightown_n_c18_0_0.vlm"},
	{0x7634F44D1EE7CE18ULL, "vxlt_0_lowtown_docks_a3_d3b_0_0.vlm"},
	{0x7635888AB7F54AC4ULL, "vxlt_hangedman_sb_0_0.vlm"},
	{0x7635939B774FC18BULL, "follower_banter6154033.cl"},
	{0x7635A1968107ED9AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg05a.phy"},
	{0x7635F56F35146887ULL, "hf_mother_h2s2.mor"},
	{0x76363441EA1712EEULL, "art\\characters\\creatures\\ragedemon\\animation\\wkf.gad"},
	{0x76368539D8D015CEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_enter.gad"},
	{0x76369531F61BAE11ULL, "mag360_samson6044872.cl"},
	{0x7636AFE56183C152ULL, "qun221_petrice6077941.cl"},
	{0x7636C134DC455BB4ULL, "185653903.ogg"},
	{0x763701152CAF3943ULL, "cs_rom000cs_isabella.bnk"},
	{0x7637EE262BE4A513ULL, "vxht_0_hightown_b22_0_0.vlm"},
	{0x7637F74FB73B2B4FULL, "6125618_m.fxe"},
	{0x7638D2C6DDF59421ULL, "vxlt_undercityserialkiller_1if_0_1.vlm"},
	{0x76392FEDE7C91AD8ULL, "t_table_04.dds"},
	{0x763A90F0EAE46075ULL, "and100_anders6111679.cl"},
	{0x763A96545212CC19ULL, "vxlt_gallowstemplar_4sr_0_0.vlm"},
	{0x763AAEA93B607210ULL, "hlac_lt_0_lowtown_a2_-196609_2y.dds"},
	{0x763ACDB52C2A9A97ULL, "qun211cr_elf_warrior6072190.cl"},
	{0x763B27370C2BAEDAULL, "6040958_m.fxe"},
	{0x763B3000917EBB08ULL, "zz_and_debug6134154.cl"},
	{0x763B3C72FD58F2BDULL, "vxlt_gallowstemplar_47z_0_0.vlm"},
	{0x763B4FF846753FE4ULL, "6089818_harrimanleaves.cut"},
	{0x763B8D47D58DB2C2ULL, "plt_cod_ltr_otl_elsa.plo"},
	{0x763BDC21ACD29097ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8mb.mmh"},
	{0x763BE021ACD297A4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8mb.mao"},
	{0x763BF221ACD2B609ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp8mb.msh"},
	{0x763C26985B293850ULL, "fx_blue_line.mao"},
	{0x763D302844E3FD33ULL, "zz_vrc_debug6093951.cl"},
	{0x763D38FEC219A53BULL, "art\\vfx\\combat\\vfx_f_isa_trail_shimmer_p.mmh"},
	{0x763D757DF98E0EEAULL, "vxlt_gallowstemplar_n_46i_0_0.vlm"},
	{0x763D8039D8D62BADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_enter.evt"},
	{0x763E18A5A27DCDA0ULL, "vxlt_house_d8_0_0.vlm"},
	{0x763E304F8D3BA1FBULL, "6034332_m.fxe"},
	{0x763E45E2591F854DULL, "6025354_m.fxe"},
	{0x763E62B152ECBFCEULL, "348611135.ogg"},
	{0x763F16743A07E88BULL, "vxht_0_hightown_a2_bu3_0_0.vlm"},
	{0x763FE267DA7657CDULL, "968583202.ogg"},
	{0x763FE43DC6AFAF1BULL, "hlac_ht_playershouse_n_855_2x.dds"},
	{0x76401A74FDFF2B85ULL, "gen00fl_fenris6113916.cl"},
	{0x76401EB0D571FAC8ULL, "dae000_bodhan6146888.cl"},
	{0x7640C19042ADBE44ULL, "6075398_m.fxe"},
	{0x76411C3844B18781ULL, "vxlt_gallowstemplar_5v4_0_0.vlm"},
	{0x764155238930337CULL, "287320688.ogg"},
	{0x7641C1F45F28AC0AULL, "577215070.ogg"},
	{0x7641EC408F0FFEE6ULL, "vxlt_house_gg_0_0.vlm"},
	{0x76422DA7FA571579ULL, "6134931_m.fxe"},
	{0x76424E50D5C9416AULL, "vxlt_0_lowtown_a2_e8b_0_0.vlm"},
	{0x76425676CBF284C0ULL, "dae000_worthy6166636.cl"},
	{0x764270893FC682D7ULL, "follower_banter6169882.cl"},
	{0x7642768017FDC62AULL, "mag370_meredith6046954.cl"},
	{0x7642DFD93A815B89ULL, "770720505.ogg"},
	{0x76432355CBE3BD83ULL, "vxlt_gallowscourtyard_6qb_0_0.vlm"},
	{0x7643662A10DDE698ULL, "vxlt_0_lowtown_docks_a3_d4c_0_0.vlm"},
	{0x76439085EC5EEFD1ULL, "6055596_f.fxe"},
	{0x7643FD092512B6EEULL, "495385293.ogg"},
	{0x76440588337CDB5FULL, "165908914.ogg"},
	{0x7644862E99B952C8ULL, "zz_dae_debug6029822.cl"},
	{0x76449D02BB6FFE3DULL, "207452762.ogg"},
	{0x7644C65148BC426AULL, "vxca_caverns1_in4_0_0.vlm"},
	{0x7644C74E56453881ULL, "648553369.ogg"},
	{0x7645783C372415A8ULL, "zz_mer_debug6023992.cl"},
	{0x76463E89989C72C3ULL, "6197853_m.fxe"},
	{0x7646D191DB1E2BF5ULL, "6155556_m.fxe"},
	{0x76473CE1EB0A3F08ULL, "vxlt_0_lowtown_docks_b5t_0_1.vlm"},
	{0x764745AAA4BCB26BULL, "vxlt_0_lowtown_a2_byg_0_0.vlm"},
	{0x764776B256E1D803ULL, "vxht_mansion2_3c0_0_0.vlm"},
	{0x7647AED8F495CAEEULL, "vxdr_dungeon_1mj_0_0.vlm"},
	{0x76482D64DEA7E24CULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge12.gad"},
	{0x76482FB00BCDEA5DULL, "669139292.ogg"},
	{0x764846AFC3F0246AULL, "439569541.ogg"},
	{0x7648B4D5F6956140ULL, "plt_cod_ltr_sk3_scribbles1.nss"},
	{0x7649B9B18EA14CD6ULL, "951605022.ogg"},
	{0x764A06079452C813ULL, "llac_lt_0_lowtown_a2_n_-2_0.dds"},
	{0x764A07BF7319CFDCULL, "vxlt_gallowstemplar_5yx_0_0.vlm"},
	{0x764A5185FEC6C859ULL, "619465750.ogg"},
	{0x764A59FEC22467A0ULL, "art\\vfx\\combat\\vfx_f_isa_trail_shimmer_p.anb"},
	{0x764A8BD88D61788CULL, "6136679_m.fxe"},
	{0x764A9DE27B75A8A9ULL, "hlac_lt_gallowstemplar_f_2383_0x.dds"},
	{0x764AB2D0B3A55BD7ULL, "vxht_mansion_n_3zp_0_0.vlm"},
	{0x764AC1EDB74D0515ULL, "6144309_m.fxe"},
	{0x764B51D4F7E999C2ULL, "vxw_bonepitdestroyed_a3_25e_0_0.vlm"},
	{0x764B6FE80D4AF43DULL, "607906452.ogg"},
	{0x764BECA10B28DE29ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_rnf.ani"},
	{0x764CC7F792D75267ULL, "hlac_ca_caverns1_mines_24424_16x.dds"},
	{0x764CED003E6A2D65ULL, "6116063_m.fxe"},
	{0x764CFD15CC827AACULL, "vxht_mansion2b_2ut_0_0.vlm"},
	{0x764D3242170E9820ULL, "6015767_m.fxe"},
	{0x764D7E21D6D7425CULL, "dre000_dougal6131171.cl"},
	{0x764DAA9C5F54740EULL, "mag291_torpor6027007.cl"},
	{0x764DD7958B5E6D62ULL, "197340792.ogg"},
	{0x764E25CFB758C45CULL, "6059385_m.fxe"},
	{0x764E346B7343A44AULL, "6208178_m.fxe"},
	{0x764E52CDEF3E9D30ULL, "vxca_caverns2_8kb_0_1.vlm"},
	{0x764E62B4E4CD0F69ULL, "123836769.ogg"},
	{0x764EBBB8948F6A42ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\_3_-_default.mao"},
	{0x764ECAD3851F67F2ULL, "llac_lt_0_lowtown_a2_n_c_65538_3.dds"},
	{0x764EF39A6D444797ULL, "vxht_mansion2b_34m_0_0.vlm"},
	{0x764F148D1E481109ULL, "vxht_keep_62p_0_0.vlm"},
	{0x764F2C64DEADFF4AULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge12.ani"},
	{0x764FAB049AF8FF9AULL, "vxlt_0_lowtown_n_ens_0_0.vlm"},
	{0x764FDA315FC5523EULL, "gen00fl_aveline6114827.cl"},
	{0x76501C17B66A4ECFULL, "62924282.ogg"},
	{0x76502074CADDE3A7ULL, "dae000_bodhan6129107.cl"},
	{0x76506A37689B4636ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_uppercut_action.ani"},
	{0x7650B4BBF2E92D7DULL, "197263373.ogg"},
	{0x765105912EEE2FEEULL, "383093789.ogg"},
	{0x76513FAD5EEC139CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgtp_0s.dds"},
	{0x7651541CC65C95C2ULL, "vxlt_0_lowtown_a2_f2g_0_0.vlm"},
	{0x7651EDBF426A6F44ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.mmh"},
	{0x7651F168751DBE8FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_loop.gad"},
	{0x76523854F71F63B6ULL, "6036722_m.fxe"},
	{0x76526A1A9E69E77AULL, "vxht_mansion2_25j_0_0.vlm"},
	{0x7652DD1BE28FA1B3ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpa.phy"},
	{0x765307A10B2F2AABULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_rnf.gad"},
	{0x7653252E294EEAC2ULL, "zz_nathanz.cut"},
	{0x76534F8253C15BE4ULL, "925509280.ogg"},
	{0x765359E54B23559EULL, "vxlt_gallowstemplar_3ix_0_0.vlm"},
	{0x765379EE6742C679ULL, "vxht_0_hightown_n_c39_0_0.vlm"},
	{0x76541E4AE1175689ULL, "hlac_lt_backalley_g_65537_3z.dds"},
	{0x76548BEA7DD35A51ULL, "mrl000_marethari6069607.cl"},
	{0x7654DF6F8C49B1DEULL, "art\\levels\\races\\design\\trap_floorlever.phy"},
	{0x7654F5BEC9023329ULL, "6059251_f.fxe"},
	{0x7655350C5FC64861ULL, "6173677_m.fxe"},
	{0x7655405D65734A7DULL, "6150087_m.fxe"},
	{0x765594EF3EA1F38DULL, "612997728.ogg"},
	{0x7655A1D8F0963554ULL, "vxdr_primevaltomb_77f_0_0.vlm"},
	{0x7655A2E0AA5E66A5ULL, "vxlt_gallowsprison_n_6jl_0_0.vlm"},
	{0x76560A99CA198966ULL, "mag291_vincento6027360.cl"},
	{0x76566F62B131C443ULL, "hf_ave3_coterie.mor"},
	{0x765695659B6F74D6ULL, "gen00fl_fenris6126412.cl"},
	{0x7656D3E8DCFD8322ULL, "108617269.ogg"},
	{0x7657170D0009317EULL, "6107475_m.fxe"},
	{0x7657B4A2A6AE110CULL, "llac_lt_warehouse_n_65537_0.dds"},
	{0x7658010DAAFB81C2ULL, "hlac_dr_0_deeproads_17038_3z.dds"},
	{0x765818D45329946BULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_cass.phy"},
	{0x76581CABC97E3530ULL, "dae000ip_crowd_ff7_talk15.utp"},
	{0x765902498CFEF45FULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5bluelyrium.msh"},
	{0x765904A10B339110ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180r_rnf.evt"},
	{0x765908498CFEFE01ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5bluelyrium.mmh"},
	{0x765914498CFF12A2ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5bluelyrium.mao"},
	{0x76591F963EDAC9DDULL, "6026016_m.fxe"},
	{0x765932189CDA4896ULL, "608257846.ogg"},
	{0x76597B805712356AULL, "vxht_0_hightown_n_bzd_0_0.vlm"},
	{0x7659D0FC50B5B567ULL, "vxht_0_hightown_n_auo_0_0.vlm"},
	{0x7659F18CA1A9F4F1ULL, "ht_0_hightown_a2_lc.mmh"},
	{0x765A380BDD18208DULL, "vxht_keep_n_5db_0_0.vlm"},
	{0x765A602A68669253ULL, "6089835_m.fxe"},
	{0x765AD4F9194B6A8CULL, "mag120_macha6044542.cl"},
	{0x765B030F4AD79C94ULL, "rock_large_diffuse.dds"},
	{0x765B6EF2BE78932CULL, "vxht_mansion2b_2nu_0_0.vlm"},
	{0x765B7B3A672C586EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_exit.evt"},
	{0x765BFA2E1DF4198BULL, "vxht_0_hightown_a2_bc5_0_0.vlm"},
	{0x765C0E84459049FDULL, "477829899.ogg"},
	{0x765C60DA6CDAFA71ULL, "6110032_m.fxe"},
	{0x765D2077F5B05FFAULL, "6147667_f.fxe"},
	{0x765D5772B5A3EA5CULL, "dae000_orana6149084.cl"},
	{0x765D7BBA2393EA4EULL, "6096038_m.fxe"},
	{0x765DE54CF023A95FULL, "6155305_m.fxe"},
	{0x765DFA777D87256DULL, "892019802.ogg"},
	{0x765E1CE18CEF181AULL, "vxlt_0_lowtown_n_ekl_0_0.vlm"},
	{0x765E723768A79127ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_uppercut_action.evt"},
	{0x766024B56D5E018DULL, "vxdr_slavers_1bb_0_0.vlm"},
	{0x7660D8227D9E0F2CULL, "plt_mag320pt_otl_emile.plo"},
	{0x76610C95E6CD8CC6ULL, "cs_mcr370cs_meredith_dies.bnk"},
	{0x766128FBACEA20A5ULL, "319421632.ogg"},
	{0x7661723A6730B419ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_exit.gad"},
	{0x7661D4210FE3C0FDULL, "6027774_m.fxe"},
	{0x7661EBCB5938DEF9ULL, "vxht_0_hightown_n_c88_0_0.vlm"},
	{0x7662A5DFE260D5E6ULL, "hlac_lt_0_lowtown_n_65538_1x.dds"},
	{0x7662F1CEAF72E486ULL, "vxdr_0_deeproads_ewl_0_0.vlm"},
	{0x76632581F19D126AULL, "821776148.ogg"},
	{0x76632BA135667A20ULL, "591071911.ogg"},
	{0x76635C7695361DECULL, "979121818.ogg"},
	{0x76636186D641EF78ULL, "gen00fl_player.cnv"},
	{0x7663E31C381A4B30ULL, "6104708_m.fxe"},
	{0x766413B5E28C4DD4ULL, "vxdr_primevaltomb_70g_0_0.vlm"},
	{0x7664933768AC3570ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_uppercut_action.gad"},
	{0x7664CFDC2BE857D2ULL, "6103909_m.fxe"},
	{0x7664EB148102A4E9ULL, "6034078_f.fxe"},
	{0x76654C1240D688FEULL, "lt_0_lowtown_a2_n_26_0_lc.mmh"},
	{0x76658B77ECF95C93ULL, "6166794_m.fxe"},
	{0x7665D142574E5F95ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_4y.dds"},
	{0x766606994E1EB33CULL, "dae000_arishok6078474.cl"},
	{0x76660E68752E9955ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_loop.ani"},
	{0x76661FCB7E92740AULL, "gen00fl_varric6033589.cl"},
	{0x76667692ED74646AULL, "mag321_comte_de_launcet6064924.cl"},
	{0x76669BE00614F011ULL, "hlac_w_bonepit_458754_0z.dds"},
	{0x7666F5AE522C701AULL, "6112060_m.fxe"},
	{0x76674386D645A480ULL, "gen00fl_player.bnk"},
	{0x7667461FFDABBBEFULL, "6204885_m.fxe"},
	{0x76679FAA940D385BULL, "6116375_m.fxe"},
	{0x7667BA01CFA7E887ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_feynriel.gad"},
	{0x7667BD227DA40063ULL, "plt_mag320pt_otl_emile.nss"},
	{0x7667E29BAF208C4AULL, "vxw_bonepitmining_a2_cq_0_0.vlm"},
	{0x766844293E66AFADULL, "6074341_m.fxe"},
	{0x766879DDDE0E6863ULL, "fen110_fenris6117937.cl"},
	{0x7668833A6736F067ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_exit.ani"},
	{0x766949FE2A1ADB70ULL, "fx_icestormcloud.dds"},
	{0x766A42DB564B81CCULL, "147907332.ogg"},
	{0x766AB3F95CCA1461ULL, "gen00fl_mrl_act3_pt1.stg"},
	{0x766B546C88173DA1ULL, "6166770_m.fxe"},
	{0x766B5F8249DA1A20ULL, "rdr111dg_martin6053581.cl"},
	{0x766C242E36765F69ULL, "vxlt_hangedman_20x_0_0.vlm"},
	{0x766C6EA404395482ULL, "6088250_f.fxe"},
	{0x766CC952B1F6018CULL, "6087427_m.fxe"},
	{0x766CCB9D180469FDULL, "88389669.ogg"},
	{0x766CCEC9B6C6A7E9ULL, "6217090_m.fxe"},
	{0x766D072D98D411A1ULL, "hlac_ht_0_hightown_a2_-131074_1x.dds"},
	{0x766D78DC42EC14D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_roll_shoulders.evt"},
	{0x766DBA68C914A711ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_angleout.ani"},
	{0x766DD2FB98EF8B90ULL, "hlac_lt_0_lowtown_a2_n_c_1_7x.dds"},
	{0x766DF0F76B43BB6AULL, "art\\vfx\\cinematic\\vfx_lyrium_cor_c.anb"},
	{0x766DFB523C2F54D3ULL, "vxht_0_hightown_aqh_0_0.vlm"},
	{0x766EAF01CFADF494ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_feynriel.evt"},
	{0x766F4C8318310D97ULL, "dae000_mother6121038.cl"},
	{0x766F5E71A6BBFB07ULL, "6142316_m.fxe"},
	{0x766FD39D2DEE6F45ULL, "gen00fl_fenris6127072.cl"},
	{0x76704A97D87EA0C7ULL, "507134012.ogg"},
	{0x7670940C32C66C8EULL, "vxca_caverns1_mines_ei4_0_0.vlm"},
	{0x7670C8211DF1B988ULL, "233302061.ogg"},
	{0x76712B0A09E52897ULL, "vxlt_0_lowtown_docks_a3_n_coy_0_0.vlm"},
	{0x76715063D42B28ADULL, "vxlt_backalley_n_1oi_0_1.vlm"},
	{0x76717BD0D7B36458ULL, "vxlt_gallowstemplar_n_2ja_0_0.vlm"},
	{0x7671A783F98476B3ULL, "6083385_m.fxe"},
	{0x7671F9F65D6E6DF2ULL, "follower_banter6156146.cl"},
	{0x7672869A8E4A97A5ULL, "vxlt_gallowsprison_n_6pj_0_0.vlm"},
	{0x7672C228F956112EULL, "zz_charlesanimproptest.cut"},
	{0x7673223AF1852638ULL, "w_bonepitmining_a2_157_0.phy"},
	{0x767372F0DD4EB91CULL, "6092278_m.fxe"},
	{0x7673C66D6BF4C4A5ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgtp_0d.dds"},
	{0x7673F4DF92FE11B1ULL, "72177720.ogg"},
	{0x76746119FA70F570ULL, "vxlt_0_lowtown_a3_n_c_hty_0_0.vlm"},
	{0x76746540316CFB4CULL, "160248197.ogg"},
	{0x76746FDC42F22461ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_roll_shoulders.gad"},
	{0x7674A65E5DD2A200ULL, "vxht_keep_3zt_0_0.vlm"},
	{0x7674C568C91AD8B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_angleout.gad"},
	{0x7674CC042BA45137ULL, "6107166_m.fxe"},
	{0x7674F4F5A7A0FC8EULL, "197551666.ogg"},
	{0x767575E2B61B7352ULL, "w_blightlands_57_0.rml"},
	{0x76757B8F38651DFFULL, "w_bonepit_98_0.msh"},
	{0x7675818F386527A1ULL, "w_bonepit_98_0.mmh"},
	{0x76758D8F38653C42ULL, "w_bonepit_98_0.mao"},
	{0x76761E349490A701ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_cst.ani"},
	{0x767670347DDF5F82ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_0.pwk"},
	{0x76768BF7EA90B9BAULL, "6076173_m.fxe"},
	{0x7676B606305CC82DULL, "6079041_m.fxe"},
	{0x7676D17790E950BEULL, "vxht_0_hightown_aob_0_1.vlm"},
	{0x7676E68B1634AE72ULL, "6134250_f.fxe"},
	{0x76771CA780FF2CA5ULL, "gen00fl_isabela6097285.cl"},
	{0x76777C07FD83F352ULL, "vxht_0_hightown_aqm_0_0.vlm"},
	{0x7677CC66BF9BE1CDULL, "follower_banter6167860.cl"},
	{0x767870E50AA7C974ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_sibling_ogrewarr.gad"},
	{0x7678810A10769400ULL, "gen00fl_anders6125480.cl"},
	{0x7678DDE801E04A8BULL, "vxht_0_hightown_a2_bi7_0_0.vlm"},
	{0x76795E089DB576C1ULL, "6107454_m.fxe"},
	{0x7679FB3748D28B3CULL, "pro000_opening6207373.cl"},
	{0x767A44C6B57ACEF3ULL, "plt_prm000pt_duru_amulet.nss"},
	{0x767A70DC42F6910FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_roll_shoulders.ani"},
	{0x767A93A2AB6297F8ULL, "845509175.ogg"},
	{0x767AE9AE5FB2D34EULL, "vxca_caverns2_79d_0_0.vlm"},
	{0x767AF0BF428D793BULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_nat_p.anb"},
	{0x767BC74EF45D66ACULL, "6144924_m.fxe"},
	{0x767C2934949525A3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_cst.gad"},
	{0x767C9AB63F0C172AULL, "6161404_m.fxe"},
	{0x767CA701CFBA24CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_get_down_feynriel.ani"},
	{0x767D153B41479C54ULL, "ht_brothel_n_3369.rml"},
	{0x767D583A09584E25ULL, "vxht_mansion_n_4hk_0_0.vlm"},
	{0x767D61BEA2DE76C5ULL, "vxht_mansion2_25a_0_0.vlm"},
	{0x767E06C6E373805DULL, "mer120_hubert6070242.cl"},
	{0x767E729DE7AF58B1ULL, "hlac_lt_gallowscourtyard_-65538_1z.dds"},
	{0x767EF424F6D01283ULL, "dae000cr_grand_cleric6147895.cl"},
	{0x767F04337C611322ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcblc.phy"},
	{0x767F13549C19517BULL, "331285979.ogg"},
	{0x767F7FE50AAE0152ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_sibling_ogrewarr.ani"},
	{0x767FA798486AB68BULL, "dae300_alistair_king6134985.cl"},
	{0x768058FE2A2DD30EULL, "fx_icestormcloud.mao"},
	{0x76805FC9E1A55182ULL, "893212407.ogg"},
	{0x76807CBF902C6A33ULL, "6141102_m.fxe"},
	{0x768094EA1A4C6B5BULL, "art\\vfx\\combat\\vfx_m_pri_fb_cast_c.mmh"},
	{0x7680EECF3502772BULL, "6111334_m.fxe"},
	{0x76817B9AC3B1B6A9ULL, "vxw_encounter3_77_0_0.vlm"},
	{0x7682298C3FBE2948ULL, "vxlt_0_lowtown_a2_n_h2h_0_0.vlm"},
	{0x7682D54989F50050ULL, "696009677.ogg"},
	{0x7682F12D58CF8D83ULL, "zz_tombs_debug.ncs"},
	{0x76830634949B0888ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_cst.evt"},
	{0x768315AB4AA0DD89ULL, "778207066.ogg"},
	{0x768350DBB485C487ULL, "464990119.ogg"},
	{0x7683F6C57E31FAA5ULL, "6095692_m.fxe"},
	{0x7684075CF8C5FD92ULL, "follower_barks6153890.cl"},
	{0x76849DFB87EA2B7CULL, "gen00fl_aveline6122037.cl"},
	{0x7684E09F40DCA1B0ULL, "62173173.ogg"},
	{0x7684ECDEA9784AC3ULL, "71238112.ogg"},
	{0x7684F8BC07FB7D13ULL, "772160334.ogg"},
	{0x768509498D243B7EULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5bluelyrium.phy"},
	{0x768519BF9E9CEAF4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf.ani"},
	{0x76859FFA6B335940ULL, "6105992_m.fxe"},
	{0x7685EDE4EF7A0BD2ULL, "vxht_0_hightown_atn_0_0.vlm"},
	{0x7686175A59BBCC6DULL, "l_env_wal_lowtown_act_2_3_calm_p2p.wbd"},
	{0x76863835F24F91F9ULL, "vxw_woundedcoast_dt_0_0.vlm"},
	{0x76869513FB140104ULL, "ca_caverns1_26038.rml"},
	{0x7686A3FD99C2731FULL, "122144580.ogg"},
	{0x7686ADD5F6CAB603ULL, "plt_cod_ltr_sk3_scribbles1.plo"},
	{0x7686BDF245246123ULL, "6069631_m.fxe"},
	{0x7686BF020ECDCF6DULL, "dae000_gamlen6102416.cl"},
	{0x7686C489945ADCACULL, "vxht_mansion2b_2ah_0_0.vlm"},
	{0x7686C55F4DEDA7C3ULL, "6132833_f.fxe"},
	{0x7687006B623E0773ULL, "259803240.ogg"},
	{0x7687040AD92E6640ULL, "vxw_encounter2_gi_0_0.vlm"},
	{0x7687128905EADDF9ULL, "al_rogue_t01.gda"},
	{0x768723A0E2B6D6CDULL, "hlac_lt_0_lowtown_a3_n_c_65535_2z.dds"},
	{0x76874442FA486FBCULL, "hlac_dr_primevaltomb_658_13y.dds"},
	{0x768777EDB339324CULL, "6075170_m.fxe"},
	{0x7688717C9BF8CEDBULL, "gen00fl_merrill6128664.cl"},
	{0x7688BCC7FDBA5904ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_channel_d.dds"},
	{0x768935CFB731B4A8ULL, "193151141.ogg"},
	{0x768979E6A1D7579DULL, "dae000_bartrand6031143.cl"},
	{0x7689E22F657C1D81ULL, "89253970.ogg"},
	{0x768A6F548E3706E9ULL, "6200740_m.fxe"},
	{0x768AB95C6404362BULL, "hlac_ht_0_hightown_a2_-65536_1y.dds"},
	{0x768AFAD02562C83DULL, "gen00fl_varric6099157.cl"},
	{0x768B16BF9EA15156ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf.gad"},
	{0x768B3D01D92FD6B0ULL, "w_encounter3_176_0.mao"},
	{0x768B4101D92FDDC3ULL, "w_encounter3_176_0.mmh"},
	{0x768B4F01D92FF51DULL, "w_encounter3_176_0.msh"},
	{0x768B60443E198374ULL, "6155580_m.fxe"},
	{0x768B7E22EDDEF927ULL, "vfx_chunkyparticle_p2.psh.fxo.dx11.assembly"},
	{0x768C671005B6BB4AULL, "lgt000cr_baz_qun_ds.utc"},
	{0x768C6991AE0F996BULL, "vxw_bonepitdestroyed_a3_1p0_0_0.vlm"},
	{0x768D116BCC95D036ULL, "vxlt_undercityencounter_wi_0_0.vlm"},
	{0x768D169FAAEC7BD8ULL, "vxht_0_hightown_cb0_0_0.vlm"},
	{0x768D77C616B29D8EULL, "vxca_caverns1_mines_ec6_0_0.vlm"},
	{0x768DD1883025D2C6ULL, "qun120_arishok_guard6051771.cl"},
	{0x768DE30B4EF68763ULL, "dae000_orana6149122.cl"},
	{0x768E24CA52FB9853ULL, "vxht_0_hightown_a2_am8_0_0.vlm"},
	{0x768EB5EA1A58E0C0ULL, "art\\vfx\\combat\\vfx_m_pri_fb_cast_c.anb"},
	{0x768EFF9FCE1D9A26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_1.evt"},
	{0x768F0D9FA0D1F996ULL, "vxht_chantry_2c8_0_0.vlm"},
	{0x76909F60701BC6AEULL, "qun211_party_comment6072165.cl"},
	{0x7690ABF337537A22ULL, "651849672.ogg"},
	{0x76919A96815640C3ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg05a.msh"},
	{0x7691A8968156581DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bg05a.mmh"},
	{0x7691C4376EA6EEB2ULL, "6065296_m.fxe"},
	{0x7691FF3368B13B42ULL, "follower_barks6153380.cl"},
	{0x769231BF9EA79DB5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runf.evt"},
	{0x769362806425F13FULL, "80804302.ogg"},
	{0x7693B78E70C74069ULL, "hlac_lt_backalley_n_1_4y.dds"},
	{0x7693D9150C65DC19ULL, "qun231_party_comment6077011.cl"},
	{0x7694384D22C25EA4ULL, "vxdr_0_deeproads_cqz_0_0.vlm"},
	{0x7694D39CC29520ACULL, "lgt110_orz_harrowmont6109003.cl"},
	{0x7695007B60B2B633ULL, "269334956.ogg"},
	{0x7695169FCE222D71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_1.gad"},
	{0x76953325570179BCULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.cess.ani"},
	{0x7695A27CC9201D5FULL, "party_med.mao"},
	{0x76966073001B56C8ULL, "gen_im_copper.uti"},
	{0x76967EC8E0490A9FULL, "hair2.vsh.fxo"},
	{0x76968441D88C22B5ULL, "6101827_m.fxe"},
	{0x7696F3F76B66C55DULL, "art\\vfx\\cinematic\\vfx_lyrium_cor_c.mmh"},
	{0x76980DC0EBBCE7DCULL, "vxca_caverns2_85e_0_0.vlm"},
	{0x769869E8CE6D50ADULL, "6052114_m.fxe"},
	{0x7698BF3071A06FEFULL, "524924892.ogg"},
	{0x7698FF33B841D39BULL, "lit_conj_elet.mao"},
	{0x769A1C1E9258AA0FULL, "7940177.ogg"},
	{0x769A879EDC41FCB8ULL, "6140348_m.fxe"},
	{0x769AD7C45523045CULL, "158940401.ogg"},
	{0x769ADB6EA005B1EBULL, "vxw_bonepitdestroyed_a3_1u3_0_0.vlm"},
	{0x769B8348BE8BE8B6ULL, "vxlt_undercityencounter_rj_0_0.vlm"},
	{0x769BAD3E6B3C2495ULL, "6043080_m.fxe"},
	{0x769BFB11B2225813ULL, "gen00fl_fenris6112085.cl"},
	{0x769C179FCE284D1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_1.ani"},
	{0x769C40255707AEFEULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.cess.gad"},
	{0x769C4201D93E52E8ULL, "w_encounter3_176_0.phy"},
	{0x769CA3B36D3087A2ULL, "66312293.ogg"},
	{0x769CA5FAAA0D722DULL, "vxlt_backalley_n_1pd_0_1.vlm"},
	{0x769DA9A6E0EA5D4DULL, "vxdr_0_deeproads_bd1_0_0.vlm"},
	{0x769E29EC2B79B068ULL, "6088308_f.fxe"},
	{0x769E525C6D2B76C6ULL, "vxlt_foundry_n_xj_0_0.vlm"},
	{0x769E899561A8EDC7ULL, "6172100_m.fxe"},
	{0x769ED726ADBD3BEAULL, "vxdr_primevalentrance_3bg_0_0.vlm"},
	{0x769F20C94901D7C0ULL, "ht_0_hightown_a3_16_0_ll.mmh"},
	{0x769F71453B86D95EULL, "6100297_f.fxe"},
	{0x769FBDDAFE93EE61ULL, "vxca_caverns1_mines_dpe_0_0.vlm"},
	{0x76A03F102675C972ULL, "vxca_cavern1_haunted_fmo_0_0.vlm"},
	{0x76A071B283AAA81BULL, "vxdr_primevaltomb_8of_0_0.vlm"},
	{0x76A1030F6DD1F74CULL, "6005322_m.fxe"},
	{0x76A135C255E1FC8BULL, "vxlt_0_lowtown_a2_hd6_0_0.vlm"},
	{0x76A17F853605A6B0ULL, "vxlt_undercityserialkiller_yi_0_0.vlm"},
	{0x76A18647D5E45989ULL, "plt_cod_ltr_otl_elsa.nss"},
	{0x76A19A4773FDA4B3ULL, "vxht_0_hightown_a2_c12_0_0.vlm"},
	{0x76A20E95E705276CULL, "cs_mcr370cs_meredith_dies.wbd"},
	{0x76A22D77DCA577A7ULL, "6172232_m.fxe"},
	{0x76A296B4E8E659F6ULL, "6075705_m.fxe"},
	{0x76A2CE73CFE5C988ULL, "6130995_f.fxe"},
	{0x76A2F154A6392621ULL, "gen00fl_isabela6075380.cl"},
	{0x76A309528D128337ULL, "6187306_m.fxe"},
	{0x76A31B25570D8E1DULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.cess.evt"},
	{0x76A327F7C6650547ULL, "6067803_f.fxe"},
	{0x76A376F9BEB8DF06ULL, "lowtown_fem_alienage1_sta6162557.cl"},
	{0x76A39B3A61D56CA7ULL, "vxlt_0_lowtown_a2_n_e8l_0_0.vlm"},
	{0x76A3C822EEB5CCBEULL, "gen00fl_isabela6015853.cl"},
	{0x76A401796A57E280ULL, "gen00fl_aveline_varric.stg"},
	{0x76A4C02D78D0A537ULL, "6099204_f.fxe"},
	{0x76A4DF337C81A3D6ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcblc.mao"},
	{0x76A4EB337C81B875ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcblc.mmh"},
	{0x76A4ED337C81BBEBULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcblc.msh"},
	{0x76A51B0A268D8B6BULL, "btn_triangle.dds"},
	{0x76A5F19EE01A86CCULL, "disabled_unlock.cur"},
	{0x76A68618777C9E5AULL, "vxw_encounter2_jt_0_0.vlm"},
	{0x76A7B2E25D3D5F11ULL, "31178137.ogg"},
	{0x76A897A7C7E60343ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.cess.evt"},
	{0x76A8FE53621B450CULL, "413413714.ogg"},
	{0x76A9363B416D0F54ULL, "ht_brothel_n_3369.anb"},
	{0x76A9C13622B2111BULL, "vxlt_0_lowtown_gc4_0_0.vlm"},
	{0x76AA253E2590AB1CULL, "6170930_m.fxe"},
	{0x76AA4CA59511731BULL, "hlac_lt_darktown_-1_24y.dds"},
	{0x76AA8963134F188AULL, "6081997_m.fxe"},
	{0x76AAE7A6E01C4427ULL, "vxca_caverns1_mines_j52_0_0.vlm"},
	{0x76AB6278C513CC92ULL, "6129274_f.fxe"},
	{0x76AB9AAF90A8978EULL, "vxht_mansion_n_53x_0_0.vlm"},
	{0x76ABA1BE714FBDD3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_meda_0s.dds"},
	{0x76ABBCD9B3B5DBB2ULL, "78199123.ogg"},
	{0x76ABF6E0CB97E4D1ULL, "708611297.ogg"},
	{0x76ABF764A215FFE9ULL, "1048536143.ogg"},
	{0x76AC1FAB06D1DEBBULL, "435461491.ogg"},
	{0x76AC64F2FB3DE5CAULL, "gen00fl_aveline6122008.cl"},
	{0x76AC76FB29FA531EULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.dura.evt"},
	{0x76ACAE104F7B7626ULL, "vxlt_0_lowtown_a2_gvb_0_0.vlm"},
	{0x76ACC2B1358BDE6CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_portcullis.msh"},
	{0x76ACC4395F218F46ULL, "vxlt_foundry_n_sk_0_0.vlm"},
	{0x76ACC8B1358BE85EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_portcullis.mmh"},
	{0x76ACD4B1358BFCF5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_portcullis.mao"},
	{0x76ACE5694A465A09ULL, "lgt260_eau_tmp_roderick6143701.cl"},
	{0x76AD48C4EDA17FE9ULL, "6095825_m.fxe"},
	{0x76ADB0A30ABB640AULL, "vxdr_dungeon_1sp_0_0.vlm"},
	{0x76ADB2A77864054AULL, "lowtown_male_alienage2_mob6166544.cl"},
	{0x76ADD099A26729E1ULL, "vxht_mansion2_211_0_0.vlm"},
	{0x76AE194F950F91D4ULL, "dae021_sister_lorena_sta6150087.cl"},
	{0x76AE613B7786A431ULL, "570388196.ogg"},
	{0x76AE69B5E0CF1A86ULL, "6169444_m.fxe"},
	{0x76AEB7CC00E6AF93ULL, "100256094.ogg"},
	{0x76AF28A09982569FULL, "gencr_buzz_perch.utc"},
	{0x76AF56E98DCD5578ULL, "6198702_f.fxe"},
	{0x76AF5754CDBBF985ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_1y.dds"},
	{0x76AF70A7C7EBDFC4ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.cess.gad"},
	{0x76AF8D8D6708A26DULL, "276889450.ogg"},
	{0x76AFA9BAA342909DULL, "dae000_arishok6067864.cl"},
	{0x76B01C430B3B42C6ULL, "vxlt_gallowstemplar_f_5ws_0_0.vlm"},
	{0x76B01F1189F45942ULL, "vxw_encounter2_camp_60_0_0.vlm"},
	{0x76B0D7D7106A88E3ULL, "vxlt_0_lowtown_n_eqc_0_0.vlm"},
	{0x76B103E604E9E70CULL, "6161345_m.fxe"},
	{0x76B1A475DFA1F7D1ULL, "355168467.ogg"},
	{0x76B1C1AC8FC07CF2ULL, "vxw_encounter3_fr_0_1.vlm"},
	{0x76B20C6C88CCDF2CULL, "6149871_m.fxe"},
	{0x76B257AD80F2A018ULL, "6140627_m.fxe"},
	{0x76B28DFB29FEE649ULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.dura.gad"},
	{0x76B2C198BF3D3AFDULL, "face_face_p17.psh.fxo.dx11"},
	{0x76B2E94CAB101249ULL, "319936802.ogg"},
	{0x76B3554A80431BB5ULL, "character_character_p217.psh.fxo.dx11.assembly"},
	{0x76B3DFCF79C5F2E5ULL, "llac_lt_0_lowtown_docks_a3_n_-65537_0.dds"},
	{0x76B3FEFD5D14C7DBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containerb_d.dds"},
	{0x76B4AAF89CD3E0F9ULL, "vxw_bonepit_1ws_0_0.vlm"},
	{0x76B53B61B118A176ULL, "6169255_m.fxe"},
	{0x76B55DDD6CA8A24DULL, "6025145_m.fxe"},
	{0x76B57FA7C7F064A2ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.cess.ani"},
	{0x76B5C391EA556CA3ULL, "627495943.ogg"},
	{0x76B5D61BE2E41328ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpa.mmh"},
	{0x76B5E3A560E29D68ULL, "vxca_caverns1_jdg_0_0.vlm"},
	{0x76B5EF3F57DFC580ULL, "sst_ss_carver.gda"},
	{0x76B5F41BE2E44652ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_tmpa.msh"},
	{0x76B63E9537CF838DULL, "vxw_bonepit_29r_0_0.vlm"},
	{0x76B6589EFE673493ULL, "156591987.ogg"},
	{0x76B65F3BD204EAC6ULL, "hlac_ca_caverns1_mines_24423_13y.dds"},
	{0x76B670448F2AA9A9ULL, "w_woundedcoast_63_0.rml"},
	{0x76B7594F63BBD5C8ULL, "vxht_mansion2_214_0_0.vlm"},
	{0x76B7BF2883F1E2EBULL, "vxw_bonepitdestroyed_a3_1c5_0_0.vlm"},
	{0x76B868A1E5A9FCC0ULL, "394475736.ogg"},
	{0x76B88C7FAB5633B9ULL, "751625087.ogg"},
	{0x76B8FB37A65B2680ULL, "llac_ca_caverns1_mines_24424_24.dds"},
	{0x76B95EFB2A04B497ULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.dura.ani"},
	{0x76B961D2788A2F84ULL, "dae221ar_brothel_nt.lst"},
	{0x76B96F70476773DDULL, "vxca_cavern1_haunted_egm_0_0.vlm"},
	{0x76B97BDC59A2824DULL, "hm_ave3_guardm.mor"},
	{0x76BA77D09336E901ULL, "vxca_caverns2_854_0_0.vlm"},
	{0x76BA7FA6588EFA8AULL, "and121_karl6093996.cl"},
	{0x76BB1E15D550065EULL, "vxht_chantry_26o_0_0.vlm"},
	{0x76BB2251DB555DECULL, "isa321_velasco6117615.cl"},
	{0x76BB23FC89367F49ULL, "38210162.ogg"},
	{0x76BB58BADF4FE728ULL, "vxht_keep_n_3ij_0_0.vlm"},
	{0x76BBB16756877935ULL, "6089576_m.fxe"},
	{0x76BBECF0D605DAA4ULL, "vxht_chantry_2cz_0_0.vlm"},
	{0x76BBF9460F7CB4E9ULL, "hlac_lt_0_lowtown_docks_n_-1_3y.dds"},
	{0x76BC4796CBDDFCB7ULL, "404611810.ogg"},
	{0x76BC7DE271B437EBULL, "dae000_gamlen6101411.cl"},
	{0x76BC9860CDDB768FULL, "gen00fl_anders6125179.cl"},
	{0x76BCFA98FAAB14AFULL, "hlac_ht_bartrand_n_3369_2y.dds"},
	{0x76BD21CAB000BB18ULL, "zz_ave_debug6059573.cl"},
	{0x76BDBC7EADDD9680ULL, "9029585.ogg"},
	{0x76BDC3A68A3AE43FULL, "llac_ht_0_hightown_a3_65538_0.dds"},
	{0x76BE90EE7BEA71C2ULL, "vxw_encounter2_camp_11_0_0.vlm"},
	{0x76BF0FCD59491E10ULL, "6118037_m.fxe"},
	{0x76BFB34C65EC6900ULL, "vxdr_primevaltomb_57d_0_0.vlm"},
	{0x76BFDD04BF4A9EB0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05b.mmh"},
	{0x76BFE4372CC3A835ULL, "6156786_m.fxe"},
	{0x76BFEB04BF4AB6CAULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05b.msh"},
	{0x76BFFFA9BA27F3F9ULL, "vxw_woundedcoast_px_0_0.vlm"},
	{0x76C01339B82CFE4DULL, "llac_lt_0_lowtown_docks_a3_n_131071_2.dds"},
	{0x76C04281F50C3DABULL, "355502326.ogg"},
	{0x76C061D278904EEDULL, "dae221ar_brothel_nt.ncs"},
	{0x76C064448F32E989ULL, "w_woundedcoast_63_0.mao"},
	{0x76C070448F32FE22ULL, "w_woundedcoast_63_0.mmh"},
	{0x76C072448F3301D8ULL, "w_woundedcoast_63_0.msh"},
	{0x76C1C7F0668137DAULL, "character_character_p207.psh.fxo.dx11.assembly"},
	{0x76C1EE3D6366D436ULL, "qun200_party_comment6095397.cl"},
	{0x76C21B2C5CDD6A7BULL, "457929359.ogg"},
	{0x76C24782DEB34DEAULL, "6101020_m.fxe"},
	{0x76C24DC0087507F8ULL, "640281182.ogg"},
	{0x76C2CE1307282ACAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgtb_0d.dds"},
	{0x76C333C1A91747B3ULL, "898570611.ogg"},
	{0x76C36923D863E5A9ULL, "6049477_m.fxe"},
	{0x76C369D25B68CF5AULL, "vxw_encounter2_pv_0_0.vlm"},
	{0x76C38A3410B849F1ULL, "6110088_f.fxe"},
	{0x76C3908B65932A6BULL, "hlac_lt_0_lowtown_docks_a3_65533_1y.dds"},
	{0x76C5261E7974820EULL, "character_character_p115.psh.fxo"},
	{0x76C5EE0DE336816AULL, "vxw_encounter3_2w_0_0.vlm"},
	{0x76C6215A59F1C097ULL, "l_env_wal_lowtown_act_2_3_calm_p2p.bnk"},
	{0x76C6310575E7FB6BULL, "vxw_bonepitdestroyed_a3_1d4_0_0.vlm"},
	{0x76C76F974EC2C3E6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_roundtable_d.dds"},
	{0x76C7ED5E56897A07ULL, "490525473.ogg"},
	{0x76C87E697494C88EULL, "vxht_mansion_n_59z_0_0.vlm"},
	{0x76C89A9F41837C6EULL, "vxht_0_hightown_a2_7iw_0_0.vlm"},
	{0x76C8D6B7223D65B0ULL, "vxdr_primevaltomb_6yq_0_0.vlm"},
	{0x76C915B946E57A90ULL, "101943042.ogg"},
	{0x76C942BE2FF4FCA0ULL, "vxlt_gallowstemplar_352_0_0.vlm"},
	{0x76C97D793E533882ULL, "llac_lt_0_lowtown_docks_-65538_6.dds"},
	{0x76CA290C4E4F8656ULL, "402334114.ogg"},
	{0x76CA3B2298ABFE62ULL, "840988188.ogg"},
	{0x76CA7566247E758AULL, "art\\vfx\\combat\\vfx_glb_bld_imp_min_c.impa.gad"},
	{0x76CBD191EB7ADE92ULL, "dae000_gamlen6121577.cl"},
	{0x76CC37BDC8634B83ULL, "vxca_caverns1_iny_0_0.vlm"},
	{0x76CC74D39D3F7B41ULL, "dae200_cullen6140264.cl"},
	{0x76CCA109DA71EE21ULL, "385183845.ogg"},
	{0x76CD17301EFBF7C1ULL, "vxlt_gallowstemplar_n_5qz_0_0.vlm"},
	{0x76CDABBB0F18B2FCULL, "61508389.ogg"},
	{0x76CDF351BB527AD1ULL, "6149333_m.fxe"},
	{0x76CEBB5CDECDE13AULL, "vxlt_0_lowtown_docks_n_b7b_0_0.vlm"},
	{0x76CEEB0235C3EBF4ULL, "vxdr_primevaltomb_50d_0_1.vlm"},
	{0x76CEF6B963397A63ULL, "vxdr_0_deeproads_8cd_0_0.vlm"},
	{0x76CF0F3AF1D364CDULL, "w_bonepitmining_a2_157_0.msh"},
	{0x76CF1D3AF1D37C00ULL, "w_bonepitmining_a2_157_0.mao"},
	{0x76CF213AF1D38353ULL, "w_bonepitmining_a2_157_0.mmh"},
	{0x76CFC4C0C8A74C10ULL, "81892178.ogg"},
	{0x76D091406459B269ULL, "gen00fl_bethany6101379.cl"},
	{0x76D0C510D953FDCBULL, "hm_lgt_riv_leg.mor"},
	{0x76D0CC50D8289219ULL, "vxht_0_hightown_c1t_0_0.vlm"},
	{0x76D1309D79AB3547ULL, "6136355_m.fxe"},
	{0x76D1586624846321ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_min_c.impa.evt"},
	{0x76D18AF87C7E4185ULL, "vxdr_0_deeproads_g31_0_0.vlm"},
	{0x76D1C2676D7D5193ULL, "152765982.ogg"},
	{0x76D201794D6BC653ULL, "lgt000ip_redblossom_crate.utp"},
	{0x76D237A53F468855ULL, "dae000_hate_enemies6167588.cl"},
	{0x76D26688A99CA7B7ULL, "gen00fl_aveline6114420.cl"},
	{0x76D26BB85725DBDBULL, "vxlt_0_lowtown_a2_ekl_0_2.vlm"},
	{0x76D29EAB7ACC7087ULL, "vxht_0_hightown_a2_c6k_0_0.vlm"},
	{0x76D2BE0A9584678FULL, "vxht_0_hightown_a2_cfc_0_0.vlm"},
	{0x76D30E123E9584C7ULL, "fxe_water_sp.dds"},
	{0x76D394A99021539BULL, "vxlt_backalley_g_1y0_0_1.vlm"},
	{0x76D399DF6931C7FFULL, "41284622.ogg"},
	{0x76D3BAAC0810028CULL, "skin_depth.psh.fxo"},
	{0x76D4AB5CAF6A082EULL, "hlac_w_bonepitmining_a2_0_8z.dds"},
	{0x76D53028951A292CULL, "docks_scene_dockworkers6176222.cl"},
	{0x76D5415DCBB13EEFULL, "vxdr_0_deeproads_gfx_0_0.vlm"},
	{0x76D5B168942F0616ULL, "dae020cr_ef_servant.utc"},
	{0x76D5CCE1BAF67342ULL, "vxca_caverns1_j52_0_0.vlm"},
	{0x76D5EEBE229C879AULL, "821740167.ogg"},
	{0x76D61C140AC4AD45ULL, "lgt000ip_ancient_stone_key.utp"},
	{0x76D6B6A691E7E62EULL, "6144078_m.fxe"},
	{0x76D6F9AF8588BAEAULL, "6155652_m.fxe"},
	{0x76D717507DF21CFBULL, "6032899_m.fxe"},
	{0x76D7216F6C0E4787ULL, "l_mer221ar_player_home.wbd"},
	{0x76D7640E80ADBE85ULL, "vxlt_darktown_4v1_0_0.vlm"},
	{0x76D773AE56303292ULL, "350274465.ogg"},
	{0x76D85B6BBD51DF47ULL, "495422478.ogg"},
	{0x76D86F25E9AEFDE8ULL, "vxw_encounter2_camp_nb_0_0.vlm"},
	{0x76D94791FC088802ULL, "6152171_f.fxe"},
	{0x76D9C7492A723AB7ULL, "6137566_f.fxe"},
	{0x76D9C9E0D8EF1D3AULL, "vxca_caverns1_mines_ckp_0_0.vlm"},
	{0x76DA3537D8D15C14ULL, "176434328.ogg"},
	{0x76DA41352E9C6D9DULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsharda.phy"},
	{0x76DA93F66260234CULL, "follower_banter6160633.cl"},
	{0x76DAB76777922C32ULL, "dae000_dalish_apprentices6147540.cl"},
	{0x76DAEE013CDC78B9ULL, "vxca_caverns2_9ws_0_0.vlm"},
	{0x76DB034AB66C29ABULL, "vxlt_darktown_57y_0_0.vlm"},
	{0x76DB0F41AD1DA300ULL, "art\\characters\\playercharacter\\humanmale\\arm_masa.phy"},
	{0x76DB6B55F7B606B9ULL, "6140550_m.fxe"},
	{0x76DBCF85CED5C060ULL, "one280_iwan6096904.cl"},
	{0x76DBEED07E347AB1ULL, "107873988.ogg"},
	{0x76DC87510D2B2484ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_c.impa.ani"},
	{0x76DCC01DA25640FEULL, "follower_banter6170582.cl"},
	{0x76DD149BB50E3D26ULL, "869280872.ogg"},
	{0x76DDAA388F4BE919ULL, "gen_im_trs_gen_sta_cld_r2.uti"},
	{0x76DDBAB229BACAC5ULL, "293793811.ogg"},
	{0x76DDC88D1E5B93DDULL, "gen00fl_anders6111112.cl"},
	{0x76DE37C0A3C9B8B8ULL, "vxht_chantry_36d_0_0.vlm"},
	{0x76DE5066248EE070ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_min_c.impa.ani"},
	{0x76DE59B240451869ULL, "vxlt_0_lowtown_docks_b5i_0_1.vlm"},
	{0x76DE68D278A95877ULL, "dae221ar_brothel_nt.are"},
	{0x76DEBC597DBD51F3ULL, "6065806_m.fxe"},
	{0x76DEE8213F166923ULL, "lowtown_male_drunk_mob6151176.cl"},
	{0x76DF10755BD6FA07ULL, "llac_lt_0_lowtown_a3_n_c_131071_7.dds"},
	{0x76DF9499361360A0ULL, "108251777.ogg"},
	{0x76E01FCFDF0CCD2AULL, "pet_sto.mao"},
	{0x76E03B47B15671E6ULL, "ht_0_hightown_a3_38_0.rml"},
	{0x76E03FC6B5D1795CULL, "plt_prm000pt_duru_amulet.plo"},
	{0x76E08820BC9C168FULL, "gen00fl_merrill6106866.cl"},
	{0x76E08C2EF7E09F82ULL, "vxca_caverns2_7mt_0_0.vlm"},
	{0x76E095B2123C5EDFULL, "6111143_m.fxe"},
	{0x76E0ABD390640696ULL, "800087757.ogg"},
	{0x76E132CF3B5C4EEAULL, "vxlt_0_lowtown_a2_equ_0_0.vlm"},
	{0x76E14CFE3487A9AAULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.anb"},
	{0x76E180B7477D6932ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.impa.gad"},
	{0x76E1DAC08B984C9EULL, "vxw_encounter2_m9_0_0.vlm"},
	{0x76E1DB32530C2BA2ULL, "vxlt_gallowstemplar_f_4sr_0_0.vlm"},
	{0x76E1E906C0204B64ULL, "gen00fl_isabela_init6077335.cl"},
	{0x76E2DDD6E3828304ULL, "hlac_lt_0_lowtown_a2_1_1y.dds"},
	{0x76E384510D313EE6ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_c.impa.gad"},
	{0x76E5A5E0E8657750ULL, "vxdr_primevaltomb_5bv_0_1.vlm"},
	{0x76E60C69CAFB9671ULL, "gen00fl_dog6165709.cl"},
	{0x76E6195B0A86C919ULL, "330950157.ogg"},
	{0x76E658ED43B07715ULL, "631062506.ogg"},
	{0x76E6766CBA451807ULL, "6142107_m.fxe"},
	{0x76E6A582EE6E9957ULL, "sib131_slave_a.bnk"},
	{0x76E6B2C7206ADFB3ULL, "ht_0_hightown_a2_58_0.rml"},
	{0x76E6E102DBA51668ULL, "vxw_encounter2_camp_ic_0_0.vlm"},
	{0x76E6F9C48CAA4172ULL, "vxlt_0_lowtown_a2_n_c_e1p_0_0.vlm"},
	{0x76E70CCC05D7FB90ULL, "6147701_m.fxe"},
	{0x76E73E9A10486253ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.cess.ani"},
	{0x76E78A459FC3A36BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium.ani"},
	{0x76E85695A93DD27DULL, "hlac_lt_warehouse_65537_0z.dds"},
	{0x76E870E96B4B5839ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottle_n.dds"},
	{0x76E883B747838D69ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.impa.evt"},
	{0x76E891564E1F34C7ULL, "801655846.ogg"},
	{0x76E899BCB7463CD5ULL, "324390589.ogg"},
	{0x76E89AFA273CA165ULL, "dae000_bartrand6016672.cl"},
	{0x76E8D459914A329EULL, "nat380_nathaniel6066386.cl"},
	{0x76E90AAB9FB7AF14ULL, "6097286_m.fxe"},
	{0x76E96F448F5601C9ULL, "w_woundedcoast_63_0.anb"},
	{0x76E97F510D35A1C5ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_c.impa.evt"},
	{0x76E98782EE709C6FULL, "sib131_slave_a.cnv"},
	{0x76E9A5763B27803DULL, "vxht_mansion2b_1ge_0_1.vlm"},
	{0x76E9FAEA02E828C1ULL, "vxlt_gallowstemplar_n_5ox_0_0.vlm"},
	{0x76EA0B62476C51FDULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_bdy_0d.dds"},
	{0x76EA11D52A2180E9ULL, "vxlt_foundry_n_11m_0_0.vlm"},
	{0x76EB23E7C0EF080FULL, "6116862_f.fxe"},
	{0x76EB2B115E3BBAA5ULL, "zz_vault_debug6201719.cl"},
	{0x76EB42352EAAE2A2ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsharda.mmh"},
	{0x76EB44352EAAE658ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsharda.msh"},
	{0x76EBA0F88EEE511FULL, "one140cr_charade6137632.cl"},
	{0x76EBDAE7BCA87FBBULL, "hi_glow.mao"},
	{0x76EBF5F3F46A7131ULL, "gen00fl_player6207743.cl"},
	{0x76EC63A8F6C4A8D4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed.ani"},
	{0x76EC6FF24EEEFAB9ULL, "art\\characters\\weapons\\longbows\\w_lbw_lw104a.msh"},
	{0x76EC79F24EEF0BA7ULL, "art\\characters\\weapons\\longbows\\w_lbw_lw104a.mmh"},
	{0x76EC9FB135C186F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_portcullis.phy"},
	{0x76ED08138DD49562ULL, "6174482_m.fxe"},
	{0x76ED357AA44DD9D1ULL, "vxca_cavern1_haunted_e97_0_0.vlm"},
	{0x76EE726C64AC42F9ULL, "vxw_bonepit_1kw_0_0.vlm"},
	{0x76EE791235F68833ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobd_0n.dds"},
	{0x76EEA5557474AD0AULL, "740470774.ogg"},
	{0x76EF5ACE367148BBULL, "hla_ht_0_hightown_a3_131072_0z.dds"},
	{0x76EF82655EB8A187ULL, "vxht_0_hightown_a2_c8e_0_0.vlm"},
	{0x76EF9E9D3C3636FAULL, "6075323_m.fxe"},
	{0x76F0422720F3DFEBULL, "15523393.ogg"},
	{0x76F04D0F45024422ULL, "vxlt_gallowstemplar_f_4xs_0_0.vlm"},
	{0x76F07C080BA9D8F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_up.evt"},
	{0x76F100219467FCA2ULL, "hlac_ca_cavern1_haunted_24424_10y.dds"},
	{0x76F10B72836474EBULL, "6060900_m.fxe"},
	{0x76F12EFC126F4C6BULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_wavesplash.mmh"},
	{0x76F1B68DD13789E1ULL, "vxlt_gallowscourtyard_78z_0_0.vlm"},
	{0x76F1D611B4A9AED7ULL, "478368449.ogg"},
	{0x76F2464B3685EC8EULL, "6042069_f.fxe"},
	{0x76F2CC0FB299F7F0ULL, "71289241.ogg"},
	{0x76F2FE53AA6F2E57ULL, "vxlt_hangedman_n_10v_0_0.vlm"},
	{0x76F34535C29B625EULL, "vxw_sundermount_30i_0_1.vlm"},
	{0x76F360A8F6CAC236ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed.gad"},
	{0x76F3A97DC824BF8EULL, "gen00fl_isabela6099325.cl"},
	{0x76F3C71D9CAAEE83ULL, "zz_followers_banter6159609.cl"},
	{0x76F3EC14EA61F448ULL, "6004115_f.fxe"},
	{0x76F3F8A5C4935D24ULL, "t1_mul_ro1.tnt"},
	{0x76F3FDC6A52F5F15ULL, "llac_ht_0_hightown_-196607_0.dds"},
	{0x76F4369A1052DE8AULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.cess.evt"},
	{0x76F472459FCE04E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium.evt"},
	{0x76F4A30BA4AFDEEFULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyo_0n.dds"},
	{0x76F4F0FB7C79E83FULL, "241891011.ogg"},
	{0x76F56BB7478DEED8ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.impa.ani"},
	{0x76F5D1479C6AF3E9ULL, "vxht_0_hightown_a2_7m0_0_0.vlm"},
	{0x76F5EFA0AF71AF35ULL, "mag270cr_hugh.utc"},
	{0x76F6111A14073F62ULL, "eye0.vsh.fxo"},
	{0x76F6A30A2DCD9235ULL, "6022562_m.fxe"},
	{0x76F6BA6A8CA28453ULL, "vxlt_0_lowtown_n_ep5_0_1.vlm"},
	{0x76F6EF3BEC13A42CULL, "167227694.ogg"},
	{0x76F70398843EF440ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_6y.dds"},
	{0x76F7356C7B9C771AULL, "6087326_m.fxe"},
	{0x76F76CB20F2DA35DULL, "372972177.ogg"},
	{0x76F775080BAFEB34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_up.gad"},
	{0x76F780EAA1C0AEA6ULL, "188176001.ogg"},
	{0x76F7C865746CDD39ULL, "vxht_0_hightown_apu_0_0.vlm"},
	{0x76F887069CA38810ULL, "6062797_f.fxe"},
	{0x76F8C4BB90DE1A4CULL, "551788590.ogg"},
	{0x76F97BA8F6CF5B95ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopentoclosed.evt"},
	{0x76F9CC4F768070BBULL, "hlac_lt_0_lowtown_a3_n_c_-65537_1x.dds"},
	{0x76FAB6636435BCA1ULL, "l_dae110ar_docks.wbd"},
	{0x76FABE9AE2239C39ULL, "484569362.ogg"},
	{0x76FB159A1058C5FDULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.cess.gad"},
	{0x76FB390F50674876ULL, "102429174.ogg"},
	{0x76FB71459FD421F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium.gad"},
	{0x76FBD5CB3B680B50ULL, "dae000_female_tranquil6165145.cl"},
	{0x76FC3982DED8C4B0ULL, "hlac_ht_mansion2b_818_53x.dds"},
	{0x76FCE18A4AA0C760ULL, "6020496_f.fxe"},
	{0x76FD015FF282259AULL, "vxlt_0_lowtown_n_e2g_0_0.vlm"},
	{0x76FD1329C4C614EFULL, "hlac_lt_0_lowtown_a3_n_c_-3_0x.dds"},
	{0x76FD84080BB47012ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_r_up.ani"},
	{0x76FDBC05B6960DDEULL, "371983524.ogg"},
	{0x76FDD191EE126701ULL, "vxht_keep_3zy_0_0.vlm"},
	{0x76FE460CCBD41EAFULL, "57786392.ogg"},
	{0x76FEB2F29FA7CF80ULL, "vxw_bonepit_mg_0_0.vlm"},
	{0x76FECE0B33B47D14ULL, "vxlt_undercityserialkiller_18m_0_0.vlm"},
	{0x76FFE404BF808ECBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05b.phy"},
	{0x77002C41AD3CF095ULL, "art\\characters\\playercharacter\\humanmale\\arm_masa.msh"},
	{0x77002E41AD3CF40BULL, "art\\characters\\playercharacter\\humanmale\\arm_masa.mmh"},
	{0x7700307D9CC61E58ULL, "hlac_ca_caverns1_mines_24423_3z.dds"},
	{0x77003A41AD3D08A8ULL, "art\\characters\\playercharacter\\humanmale\\arm_masa.mao"},
	{0x770075B41B2583FEULL, "223717448.ogg"},
	{0x77007AEABCB1D17AULL, "llac_ca_cavern1_haunted_24424_19.dds"},
	{0x77009EF7B0329108ULL, "hlac_w_blightlands_327681_0z.dds"},
	{0x7702629E23BB716BULL, "mag100_cullen6086360.cl"},
	{0x77032CDBB12FF831ULL, "vxw_woundedcoast_om_0_0.vlm"},
	{0x77038902B3FA4036ULL, "314874430.ogg"},
	{0x7703A0BE2B93E509ULL, "6128322_f.fxe"},
	{0x77046836530D666DULL, "198446092.ogg"},
	{0x770484F1E578A6FFULL, "317597234.ogg"},
	{0x77048552CE656F1AULL, "97776549.ogg"},
	{0x770488204A152342ULL, "vxdr_primevalentrance_22b_0_0.vlm"},
	{0x77049B040D3D5854ULL, "84070508.ogg"},
	{0x7704DE2F6E70DA70ULL, "vxht_brothel_n_6f9_0_0.vlm"},
	{0x77051AD810FCAD80ULL, "gen00fl_aveline6139666.cl"},
	{0x7705556B02C13C1BULL, "vxht_chantry_n_302_0_0.vlm"},
	{0x7705611269E2B472ULL, "hlac_ca_caverns1_mines_24426_14y.dds"},
	{0x77057BCB096CCAC7ULL, "451274992.ogg"},
	{0x7705A133369C2B37ULL, "vxlt_0_lowtown_a2_n_c_ehr_0_0.vlm"},
	{0x7705BB2E566B06FEULL, "dae000cr_grand_cleric6182020.cl"},
	{0x77063FC3861CDAC0ULL, "vxht_brothel_n_6be_0_0.vlm"},
	{0x7706439812BEC239ULL, "vxca_caverns2_9fp_0_0.vlm"},
	{0x77065C1CDF7955CBULL, "llac_ca_caverns2prc_5802_10.dds"},
	{0x77069F5B9A8CE9AEULL, "180233793.ogg"},
	{0x7706B9FBC914C718ULL, "hlac_ht_0_hightown_a3_-131071_0z.dds"},
	{0x7706F8E38E99A090ULL, "6069697_f.fxe"},
	{0x7707721100AD5FB3ULL, "gen00fl_anders6045336.cl"},
	{0x7708E885EDCB2854ULL, "zz_dae_debug6050736.cl"},
	{0x7709E1F3B9A5641FULL, "dae000_bartrand6206862.cl"},
	{0x770A14CCD4525A5CULL, "1040678592.ogg"},
	{0x770A4FFE34AAB39DULL, "art\\vfx\\combat\\vfx_r_gn2_speed_bl_loop_c.mmh"},
	{0x770A8AA74A95F35DULL, "531484035.ogg"},
	{0x770B60E399F41466ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_ltrb.anb"},
	{0x770B770863BB5109ULL, "hlac_dr_0_deeproads_17035_1x.dds"},
	{0x770BAA2C636C633CULL, "6042235_m.fxe"},
	{0x770BAF7635590119ULL, "hlac_ca_caverns2prc_5802_18z.dds"},
	{0x770C4843ACEF426BULL, "vxdr_primevaltomb_8y4_0_0.vlm"},
	{0x770D4BB63BA8DD1DULL, "follower_banter6170424.cl"},
	{0x770DE44AD455441EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_andraste_sa.dds"},
	{0x770DE7B8FCAADA46ULL, "630465084.ogg"},
	{0x770E11F109D0FC70ULL, "gen00fl_aveline6115390.cl"},
	{0x770E3F15196EBA1AULL, "6119592_f.fxe"},
	{0x770EA932B40FAE27ULL, "vxlt_0_lowtown_docks_n_apm_0_0.vlm"},
	{0x770F2068B2B3598EULL, "vxca_caverns2prc_8bw_0_0.vlm"},
	{0x770F235B6EF5861EULL, "6070217_f.fxe"},
	{0x770F4536A207C982ULL, "644475578.ogg"},
	{0x770F58A69E7225F6ULL, "796973562.ogg"},
	{0x770FBFA90627D4C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v2.gad"},
	{0x770FD89D95B92B8EULL, "lgt000cr_dwf_turn_in_4.utc"},
	{0x7710875428784713ULL, "gen00flst_sebacttwod.stg"},
	{0x7711675E9E56F9EEULL, "vxht_0_hightown_a2_bwh_0_0.vlm"},
	{0x77128AC6FBCBE7C2ULL, "792800080.ogg"},
	{0x77128EF053A01C23ULL, "vxht_0_hightown_a3_c3e_0_0.vlm"},
	{0x7712B501805724E9ULL, "vxht_0_hightown_a2_7g6_0_0.vlm"},
	{0x7712E011B6E8B7F8ULL, "zz_followers_banter6157703.cl"},
	{0x771375EC4D4E7492ULL, "vxlt_0_lowtown_a2_n_c_ghy_0_0.vlm"},
	{0x7713BA31F341B938ULL, "190653813.ogg"},
	{0x7713FE944D75CC02ULL, "6039129_m.fxe"},
	{0x7714651853D8C9C3ULL, "combat_lmm_r1_a5.gda"},
	{0x7714949BC0F9FB0EULL, "vxw_encounter2_camp_ma_0_0.vlm"},
	{0x7714B7804C128B84ULL, "257378299.ogg"},
	{0x7715168CBBBBBF80ULL, "gen_im_cns_flu_merril_04.uti"},
	{0x7715AE6A9367A383ULL, "498023238.ogg"},
	{0x771652E23824961BULL, "6104952_m.fxe"},
	{0x771689A8B30F9655ULL, "289496809.ogg"},
	{0x7716A006944220DCULL, "77430982.ogg"},
	{0x7716D60E90C28D0DULL, "6182718_m.fxe"},
	{0x7716E2A9062E2EB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v2.evt"},
	{0x771743AF5A66224FULL, "493771411.ogg"},
	{0x77175249F5FF582EULL, "isa211_isabela6083237.cl"},
	{0x771755B4A0F71C07ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_003a_0t.dds"},
	{0x77178D8DA3CB885AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgtb_0t.dds"},
	{0x7717B6C7E7E5F9C8ULL, "hlac_lt_0_lowtown_n_-3_2x.dds"},
	{0x7717CA08770F8FC0ULL, "vxw_blightlands_9h_0_0.vlm"},
	{0x77182B6F6C45EFBDULL, "l_mer221ar_player_home.bnk"},
	{0x77184388F5E0AFACULL, "373062988.ogg"},
	{0x7718B778BFF62216ULL, "lgt220_shp_mag_guard6149919.cl"},
	{0x771A0D2115F7760DULL, "80724440.ogg"},
	{0x771A27572A381E3EULL, "nat380_temmerin6067331.cl"},
	{0x771A3135015B2EB1ULL, "6165454_m.fxe"},
	{0x771A339C994A6526ULL, "2474205.ogg"},
	{0x771A90CA49DC8186ULL, "198035758.ogg"},
	{0x771B9F3BDFE70BE5ULL, "257706000.ogg"},
	{0x771BA1C4A36D23B1ULL, "6072249_f.fxe"},
	{0x771BF640737A122AULL, "vxlt_foundry_n_xf_0_0.vlm"},
	{0x771C1D8FCC98C411ULL, "357898331.ogg"},
	{0x771C26F9D4B39CEAULL, "dae000_flemeth6000888.cl"},
	{0x771D61C5E2FA7F7BULL, "hlac_lt_0_lowtown_a3_n_c_-65534_0z.dds"},
	{0x771D870FCC5A6508ULL, "vxw_bonepitdestroyed_a3_240_0_0.vlm"},
	{0x771E05A98818A71DULL, "vxlt_0_lowtown_a2_n_c_gxr_0_0.vlm"},
	{0x771E1DE25E4E7CBCULL, "452878877.ogg"},
	{0x771E760E0F301AD1ULL, "6042762_m.fxe"},
	{0x771E9C1DE5BBEFE9ULL, "339613413.ogg"},
	{0x771EAA0FC8B343B5ULL, "hlac_ht_0_hightown_a3_65535_3x.dds"},
	{0x771F64E9DFB67558ULL, "dae000_arishok6072865.cl"},
	{0x771FB6689B7A0EBAULL, "113278604.ogg"},
	{0x771FD863645512A3ULL, "l_dae110ar_docks.bnk"},
	{0x771FDBEE2C0B8C54ULL, "vxdr_primevaltomb_7qp_0_0.vlm"},
	{0x77201095951C2931ULL, "vxw_woundedcoast_ac_0_0.vlm"},
	{0x7720EEE92820661CULL, "llac_lt_gallowscourtyard_-65536_0.dds"},
	{0x772100CD459634A3ULL, "vxht_0_hightown_a3_c4d_0_0.vlm"},
	{0x7721313594281043ULL, "lgt000_chanter6147187.cl"},
	{0x77216C2700C7429BULL, "6173820_m.fxe"},
	{0x77216C2DD0FE8946ULL, "ico_bomb_fel.dds"},
	{0x7721DD2C144CD309ULL, "llac_dr_primevalentrance_1724_3.dds"},
	{0x772292CA059165B7ULL, "6202281_m.fxe"},
	{0x772297DF5465E101ULL, "925155312.ogg"},
	{0x7722B39EC664C38AULL, "llac_lt_warehouse_-65536_1.dds"},
	{0x77230678B2F0138EULL, "vxw_encounter2_camp_hb_0_0.vlm"},
	{0x772322B0582E2531ULL, "hla_ht_0_hightown_n_-65536_0z.dds"},
	{0x77237D861297520FULL, "82779767.ogg"},
	{0x7723B7F275A07A93ULL, "mag100_thrask6051764.cl"},
	{0x7723EAA90638C6A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v2.ani"},
	{0x77241EB37CE6570FULL, "vxlt_0_lowtown_docks_a3_cwv_0_0.vlm"},
	{0x7724BAD137E3ADC1ULL, "llac_lt_0_lowtown_docks_a3_n_-1_3.dds"},
	{0x77255D57FC05B467ULL, "6070831_m.fxe"},
	{0x772562EB7BA44D2EULL, "387039608.ogg"},
	{0x77261B00CAE7683DULL, "vxdr_primevaltomb_4h9_0_0.vlm"},
	{0x772631FD15887E79ULL, "824350700.ogg"},
	{0x77271F3C55B56BDBULL, "w_blightlands_141_0.msh"},
	{0x7727313C55B58A26ULL, "w_blightlands_141_0.mao"},
	{0x77273D3C55B59EC5ULL, "w_blightlands_141_0.mmh"},
	{0x77273F78CD108427ULL, "vxlt_warehouse_o5_0_0.vlm"},
	{0x7727677218DE5D89ULL, "914991569.ogg"},
	{0x77276B25691480C0ULL, "fen201ar_slavers_dungeon.are"},
	{0x772780A42ECD070FULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.cess.gad"},
	{0x7728575E2F984610ULL, "fex_ruin_spt_b.msh"},
	{0x7728655E2F985E2AULL, "fex_ruin_spt_b.mmh"},
	{0x7729304C824A2D3CULL, "gen00fl_anders6144081.cl"},
	{0x77296BF6BB94AF7BULL, "1009408740.ogg"},
	{0x77296C1FF52AC96AULL, "vxlt_0_lowtown_a2_ehr_0_0.vlm"},
	{0x7729A038B25CCD44ULL, "pro000cr_friedrich_thug_2.utc"},
	{0x772A23A20996C694ULL, "vxlt_undercityserialkiller_17p_0_0.vlm"},
	{0x772A62256916A742ULL, "fen201ar_slavers_dungeon.ncs"},
	{0x772A762E0224E733ULL, "vxca_cavern1_haunted_cn2_0_0.vlm"},
	{0x772AF72FBD8F00BFULL, "gen00fl_bethany6101207.cl"},
	{0x772B0DF2308B2CC5ULL, "1026913379.ogg"},
	{0x772BBA1A487A55D3ULL, "563160910.ogg"},
	{0x772C770E038BC280ULL, "follower_banter6151471.cl"},
	{0x772CB03D5A936F2BULL, "gen00fl_anders6144109.cl"},
	{0x772CB9BE2E82A54AULL, "bink_bink_p1.psh.fxo"},
	{0x772CEB96E2486013ULL, "966446703.ogg"},
	{0x772D0D091DB16712ULL, "440445690.ogg"},
	{0x772D95A42ED196BCULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.cess.evt"},
	{0x772E41DDA55A576AULL, "6000774_m.fxe"},
	{0x7730A55F8498E478ULL, "dre000_sibling_has_blight6037026.cl"},
	{0x7730C5A60CCC6A00ULL, "685022483.ogg"},
	{0x773104A6F7877E37ULL, "6209381_m.fxe"},
	{0x77317855A4E62C0EULL, "vxw_encounter2_camp_cc_0_0.vlm"},
	{0x77317A25691CEED3ULL, "fen201ar_slavers_dungeon.lst"},
	{0x773185A3D367F79CULL, "838055748.ogg"},
	{0x7731C47F62DF190DULL, "6171954_m.fxe"},
	{0x7731D168B8125290ULL, "gen00fl_sebastian6179978.cl"},
	{0x773280489953253FULL, "art\\characters\\playercharacter\\elfmale\\em_har_hb1a.msh"},
	{0x7732864899532FE1ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hb1a.mmh"},
	{0x773290906EDC6F8FULL, "vxlt_0_lowtown_docks_a3_czq_0_0.vlm"},
	{0x77329AE0275F1BC4ULL, "vxht_0_hightown_c3s_0_0.vlm"},
	{0x7732B73EA823A542ULL, "rot_dis.mao"},
	{0x7732BB61831AE0C2ULL, "vxht_0_hightown_n_a6o_0_0.vlm"},
	{0x7732F9D8E87914E6ULL, "6088899_m.fxe"},
	{0x7733069341BAC0F4ULL, "6173654_m.fxe"},
	{0x7733AB727CF6A21BULL, "lt_0_lowtown_docks_a3_n_225_0.mmh"},
	{0x7733B7727CF6B6B8ULL, "lt_0_lowtown_docks_a3_n_225_0.mao"},
	{0x7733C9727CF6D505ULL, "lt_0_lowtown_docks_a3_n_225_0.msh"},
	{0x7733CB515A81FA52ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge9.gad"},
	{0x77340046C72EC95DULL, "6081257_m.fxe"},
	{0x77348CDDBCDD80BDULL, "vxdr_primevaltomb_4c8_0_0.vlm"},
	{0x773505F424D656F7ULL, "vxca_caverns2_8k8_0_0.vlm"},
	{0x7735B155BF069CA7ULL, "vxlt_warehouse_p4_0_0.vlm"},
	{0x77361E1140B3E22EULL, "face_face_p12.psh.fxo.dx11"},
	{0x77367E18A64064BBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platform.msh"},
	{0x77368F419DE21492ULL, "gen00fl_varric6033732.cl"},
	{0x77369018A6408306ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platform.mao"},
	{0x77369C18A64097A5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platform.mmh"},
	{0x7736AB8B751C641AULL, "vxht_keep_n_7a4_0_0.vlm"},
	{0x7736F64E54A39F05ULL, "zz_mag_debug6054495.cl"},
	{0x77373D767E293ACDULL, "ser221_party_comments6164311.cl"},
	{0x7737430E21559B97ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf.gad"},
	{0x77376488642D746BULL, "vxht_chantry_n_32l_0_0.vlm"},
	{0x7737EB1F17D8BF61ULL, "738939156.ogg"},
	{0x773873E7B66AC43AULL, "lla_ca_cave_2194_0.dds"},
	{0x77390EDC7314F484ULL, "6122069_m.fxe"},
	{0x77394EB0319DD171ULL, "fen210_hadriana6110574.cl"},
	{0x7739B61436937268ULL, "crt_bee_group.mao"},
	{0x7739D8B590861B4CULL, "follower_banter6171136.cl"},
	{0x7739DDDA3C79077EULL, "hlac_lt_0_lowtown_docks_a3_n_196607_2y.dds"},
	{0x773A6B8070A48B07ULL, "hlac_lt_undercitydungeon_-1_45y.dds"},
	{0x773A6CA304361581ULL, "vxdr_primevaltomb_52j_0_0.vlm"},
	{0x773B4F0A9902A2FBULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardmgfront.evt"},
	{0x773B89CA0C68106FULL, "vxlt_warehouse_n_py_0_0.vlm"},
	{0x773B9DA42EDDE1D5ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.cess.ani"},
	{0x773BB52133F5CDC5ULL, "dae000_serendipity6161378.cl"},
	{0x773BC718BD2A835FULL, "vxlt_0_lowtown_docks_n_b51_0_0.vlm"},
	{0x773BCA348B457663ULL, "gen00fl_aveline6122083.cl"},
	{0x773C0DB090965ADAULL, "ser221_gascard6111727.cl"},
	{0x773C1E793362C92FULL, "vxlt_0_lowtown_h5m_0_0.vlm"},
	{0x773CBEE7B0133E12ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend3_darelief.dds"},
	{0x773CE53C7019E664ULL, "sse_ss_grace.gda"},
	{0x773D34EA5B6487F0ULL, "41477074.ogg"},
	{0x773D54815D39AE5CULL, "6110121_m.fxe"},
	{0x773E3A3B8C48D145ULL, "vxht_0_hightown_a3_ap4_0_0.vlm"},
	{0x773E580E215BDE24ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf.evt"},
	{0x773E7F2DC426DA4AULL, "6119875_f.fxe"},
	{0x773ECB5F7046C8D2ULL, "151061996.ogg"},
	{0x773ED32E25C44015ULL, "6115835_m.fxe"},
	{0x773F227DF10033FBULL, "lt_0_lowtown_docks_153_0.mmh"},
	{0x773F2DFC9365DF41ULL, "369148613.ogg"},
	{0x773F2E7DF1004798ULL, "lt_0_lowtown_docks_153_0.mao"},
	{0x773F3936C5A92574ULL, "hlac_dr_primevalentrance_1726_2x.dds"},
	{0x773F407DF10066E5ULL, "lt_0_lowtown_docks_153_0.msh"},
	{0x773F5AE8D84C7A83ULL, "pro000_merchant6051560.cl"},
	{0x773F838382BC7C76ULL, "hlac_w_encounter3_-131069_0x.dds"},
	{0x773F8B5294B49CB5ULL, "948263979.ogg"},
	{0x773F92C5EE648C67ULL, "cs_fen110cs_fenris_intro.bnk"},
	{0x773FBDD12CE449EDULL, "6199750_m.fxe"},
	{0x773FFF2108EE4D3CULL, "pro000_athenril6049471.cl"},
	{0x774001B62E3B57B9ULL, "vxht_0_hightown_aox_0_0.vlm"},
	{0x7740117B4DAA5FB1ULL, "vxw_bonepit_1wk_0_0.vlm"},
	{0x774039172A046841ULL, "6214424_m.fxe"},
	{0x7740AF6F943D777BULL, "zz_followers_banter6158703.cl"},
	{0x7740C7BA113F90C9ULL, "6105702_m.fxe"},
	{0x7740E1DB81F8739DULL, "follower_banter6143422.cl"},
	{0x7741078067AD612FULL, "vxht_chantry_n_2ze_0_0.vlm"},
	{0x77410CBD19553444ULL, "vxht_0_hightown_c6p_0_0.vlm"},
	{0x774125A8497E12D6ULL, "dae000_bodhan6146895.cl"},
	{0x7741E1A0CDF1E349ULL, "vxlt_0_lowtown_a3_n_c_glc_0_0.vlm"},
	{0x7741EA1F18818CF1ULL, "lt_shop_3567_ll.mmh"},
	{0x7741F0976D314BD2ULL, "6118501_m.fxe"},
	{0x7742480A9908B53CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardmgfront.gad"},
	{0x7742E04B478798F0ULL, "vxw_sundermount_276_0_0.vlm"},
	{0x77439708B7F502FBULL, "vxlt_gallowscourtyard_n_79x_0_0.vlm"},
	{0x7743B082F495CB7DULL, "mcr371ip_vfx_source.utp"},
	{0x7744375CCE85CD33ULL, "zz_dae_debug6029481.cl"},
	{0x7744771B169BFDFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_right.ani"},
	{0x7744F20B38F59600ULL, "6124420_f.fxe"},
	{0x77450AACBF7B493DULL, "vxca_caverns2_78p_0_0.vlm"},
	{0x77451DD6E3C39122ULL, "mcr371_harvester_resurect.cnv"},
	{0x774548604440BE53ULL, "832836100.ogg"},
	{0x7745C05C1F8D3972ULL, "124348394.ogg"},
	{0x7745F1A1D834C581ULL, "vxdr_primevaltomb_6v1_0_0.vlm"},
	{0x7745F906B807B765ULL, "dae000_flemeth6000904.cl"},
	{0x7746D5D7B2DDD7CEULL, "hlac_lt_0_lowtown_docks_n_131071_7y.dds"},
	{0x77471095FA9ADBF1ULL, "171366610.ogg"},
	{0x77473B941FF88048ULL, "slideshow.gfx"},
	{0x7747B6515A927FF8ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge9.ani"},
	{0x7747D12CE8DE1031ULL, "art\\vfx\\combat\\vfx_c_pride_cast_f2_c.anb"},
	{0x7747D56848EDD59EULL, "follower_banter6168770.cl"},
	{0x7747DAD668DF2FABULL, "act00pt_main.ncs"},
	{0x7748E96BE15ADF04ULL, "vxht_brothel_n_61x_0_0.vlm"},
	{0x7749470A990ED23AULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speardmgfront.ani"},
	{0x7749764371123592ULL, "755814910.ogg"},
	{0x7749FBA6FE5E28EFULL, "vxlt_warehouse_n_wx_0_0.vlm"},
	{0x774A428D194A4E9CULL, "6156343_m.fxe"},
	{0x774A912135FB3FB8ULL, "6202062_m.fxe"},
	{0x774B4281C2206718ULL, "zz_otl_debug6060945.cl"},
	{0x774B700E2166911DULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf.ani"},
	{0x774B841B16A2323EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_right.gad"},
	{0x774B9CE8797B880EULL, "vxdr_slavers_2hh_0_0.vlm"},
	{0x774BAE77B6134704ULL, "vxw_encounter2_gm_0_0.vlm"},
	{0x774BFA29C9B7CE10ULL, "hlac_ht_0_hightown_65534_1y.dds"},
	{0x774BFBB7F977C0EDULL, "hlac_lt_0_lowtown_docks_65534_1y.dds"},
	{0x774CA829A31BFF14ULL, "one300ar_vanard_fight.lst"},
	{0x774CCA08707E40B2ULL, "6096194_f.fxe"},
	{0x774D660CEC8CD549ULL, "6114031_f.fxe"},
	{0x774DF1C7E23723ABULL, "gen_im_wep_war_one_act3_07.uti"},
	{0x774E3DAF1610D24BULL, "6128052_f.fxe"},
	{0x774E55AEC1916CF9ULL, "zz_one_debug6075026.cl"},
	{0x774E618617DB2D67ULL, "vxw_bonepit_28g_0_0.vlm"},
	{0x774E874E6D54C8F6ULL, "87296107.ogg"},
	{0x774E8B7D4CB11CD2ULL, "static_static_p25.psh.fxo"},
	{0x774E8EA225C9ADEAULL, "art\\characters\\creatures\\pridedemon\\animation\\blendtree\\pridedemon.anb"},
	{0x774EA0F4BAB7D195ULL, "vxw_sundermount_24z_0_0.vlm"},
	{0x774EAFA502218C2BULL, "follower_banter6163888.cl"},
	{0x774EF5BECCA22E7FULL, "vxlt_0_lowtown_docks_a3_cy4_0_0.vlm"},
	{0x774F795D59A379AFULL, "vxht_chantry_n_2ud_0_0.vlm"},
	{0x774FA6FAFEBC554BULL, "6175726_f.fxe"},
	{0x774FBC5575FD9663ULL, "vxlt_0_lowtown_n_e8l_0_0.vlm"},
	{0x7750537DBFE7FBC9ULL, "vxlt_0_lowtown_a3_n_c_gcb_0_0.vlm"},
	{0x7750979612CC90EEULL, "vxlt_gallowstemplar_5j3_0_0.vlm"},
	{0x77510858F3C97B06ULL, "502629824.ogg"},
	{0x775142B55C54D29CULL, "hlac_lt_0_lowtown_a2_n_-131072_3z.dds"},
	{0x77515228397DB170ULL, "vxw_sundermount_287_0_0.vlm"},
	{0x7751AECAA59E5468ULL, "one200_gold_followers6117251.cl"},
	{0x7751E0585C66BB77ULL, "vxht_0_hightown_bj6_0_0.vlm"},
	{0x7751E854226AD00BULL, "and121_karl6093865.cl"},
	{0x7752135B7BE318ECULL, "gen00fl_anders6096629.cl"},
	{0x7752536795824876ULL, "gen00fl_anders6121222.cl"},
	{0x77525F1B16A8125DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_roll_right.evt"},
	{0x7752A1750B9FCC6AULL, "llac_ht_mansion2b_818_2.dds"},
	{0x7752B2F99DA4C8AEULL, "vxw_woundedcoast_1cj_0_1.vlm"},
	{0x77530247C30186B8ULL, "gen00fl_merrill6111221.cl"},
	{0x775332883E68BFACULL, "vxht_0_hightown_a3_a81_0_0.vlm"},
	{0x775388578C12F0CBULL, "dre000_varric6027768.cl"},
	{0x7753923C2026E5E9ULL, "6051310_f.fxe"},
	{0x7753C829A322541DULL, "one300ar_vanard_fight.ncs"},
	{0x7754F22CE8E8D2AAULL, "art\\vfx\\combat\\vfx_c_pride_cast_f2_c.mmh"},
	{0x77573EF1E6B62A3AULL, "823668605.ogg"},
	{0x7757A27EFDEBCD0CULL, "6116476_f.fxe"},
	{0x7757A509AEC9BE92ULL, "6053899_m.fxe"},
	{0x77580880BB6CA4DCULL, "dae000_madam_lusine6151630.cl"},
	{0x7758C13EEC68C193ULL, "87980210.ogg"},
	{0x7759FB144F3DABF1ULL, "darktown_male_fereldan_sta.bnk"},
	{0x775A51BBE1D89F36ULL, "l_env_wal_gallows_int_military_day.wbd"},
	{0x775AADCCD612EDB2ULL, "vxht_mansion2_281_0_0.vlm"},
	{0x775B2706B65ECAFDULL, "vxlt_0_lowtown_a2_n_fib_0_0.vlm"},
	{0x775B68D11B9B5909ULL, "vxlt_warehouse_w3_0_0.vlm"},
	{0x775BB2C1A505D919ULL, "propalpha1_ds2.psh.fxo"},
	{0x775BC7054E3C2578ULL, "114353160.ogg"},
	{0x775BD8A5E5309A17ULL, "6155895_m.fxe"},
	{0x775BF66DB83372E9ULL, "6087668_m.fxe"},
	{0x775C772785A9A7E3ULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.cess.ani"},
	{0x775CDD144F3FAE01ULL, "darktown_male_fereldan_sta.cnv"},
	{0x775CFD5A3EA73552ULL, "static_static_p54.psh.fxo"},
	{0x775D376F6FD9B4E3ULL, "llac_lt_gallowscourtyard_-2_2.dds"},
	{0x775D42EE72F4F081ULL, "zz_lgt_3_debug6140428.cl"},
	{0x775D8A16896A29DFULL, "vxca_caverns2_96f_0_0.vlm"},
	{0x775D9096E8EF123DULL, "346896304.ogg"},
	{0x775E24C719611890ULL, "6147110_m.fxe"},
	{0x775E45E1E9D65004ULL, "567739193.ogg"},
	{0x775E541B9528CEF9ULL, "6034363_m.fxe"},
	{0x775E874899786C5EULL, "art\\characters\\playercharacter\\elfmale\\em_har_hb1a.phy"},
	{0x775EC30A39EB55AFULL, "vxca_cavern1_haunted_eaq_0_0.vlm"},
	{0x775F2FC83886E5E8ULL, "llac_lt_0_lowtown_a2_-65537_6.dds"},
	{0x775F950580F10CD1ULL, "hlac_lt_0_lowtown_a3_n_c_65538_5x.dds"},
	{0x775FA95893B7E452ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm_facem1.mao"},
	{0x775FF6FAF818BBD7ULL, "6027383_m.fxe"},
	{0x776002FBB7384F1DULL, "6179785_m.fxe"},
	{0x7760486EDD1662E8ULL, "907306250.ogg"},
	{0x77609E96395B2C9BULL, "dae000ip_crowd_mf5_talk6.utp"},
	{0x7760C1012B437F88ULL, "6084471_m.fxe"},
	{0x7760F1F27C6399E1ULL, "vxw_bonepitdestroyed_a3_18z_0_0.vlm"},
	{0x7760F82036BB7CD1ULL, "cod_ltr_sk_meredith.utp"},
	{0x776115ED3D89BDC0ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava2_d.dds"},
	{0x77623DF4C18BD7F1ULL, "vxht_mansion_n_2bs_0_0.vlm"},
	{0x7762956622CCF4C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_cry.ani"},
	{0x7762B182F4B07D3CULL, "mcr371ip_vfx_source.ncs"},
	{0x7762B38FB90D3A6BULL, "345925748.ogg"},
	{0x77630C711157235EULL, "381599584.ogg"},
	{0x776356D5CBAC2B9CULL, "6140292_m.fxe"},
	{0x77641CC4FA89E323ULL, "dae000_bartrand6027443.cl"},
	{0x7764BE9DCA9C1E6FULL, "6125926_m.fxe"},
	{0x7765812C80B48804ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_meda.msh"},
	{0x7765932C80B4A76DULL, "art\\characters\\playercharacter\\dwarfmale\\glv_meda.mao"},
	{0x7765972C80B4ADA6ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_meda.mmh"},
	{0x7766A613CFAD665EULL, "hlac_lt_0_lowtown_a2_1_8x.dds"},
	{0x7766D530B004150DULL, "dae020cr_hm_noble2.utc"},
	{0x77672762D6B04FF2ULL, "ch2alphnotnt.psh.fxo"},
	{0x7767D18F7DA1F875ULL, "one110_pryce6088407.cl"},
	{0x7767D196029F16E7ULL, "6033621_m.fxe"},
	{0x7767DE8DE81C55BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_enter.gad"},
	{0x776897B5E58409A5ULL, "318496222.ogg"},
	{0x776897F428EEB5B3ULL, "vxca_caverns2_7xn_0_0.vlm"},
	{0x7768B06622D18E4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_cry.gad"},
	{0x7768E8442886588AULL, "977272330.ogg"},
	{0x7768EE7C19E24E27ULL, "64135497.ogg"},
	{0x7768F74BC2867A8AULL, "vxht_0_hightown_a2_bpu_0_0.vlm"},
	{0x7768FA890E511DF2ULL, "vxlt_0_lowtown_a2_n_c_eny_0_0.vlm"},
	{0x77698F2785B45ADAULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.cess.evt"},
	{0x776990A4952043E0ULL, "94453610.ogg"},
	{0x77699CA3A768A1ACULL, "6117469_m.fxe"},
	{0x7769C22060E35746ULL, "vxlt_gallowstemplar_n_5rt_0_0.vlm"},
	{0x7769D86E18A136A4ULL, "llac_lt_0_lowtown_n_-196608_3.dds"},
	{0x7769DAAE0D917189ULL, "vxlt_warehouse_x2_0_0.vlm"},
	{0x776A04758F26E6B3ULL, "dae000_bodhan6146878.cl"},
	{0x776A727C21C85332ULL, "989140346.ogg"},
	{0x776AB456E19D264BULL, "art\\vfx\\combat\\vfx_g_spear_whirlwind_c.impa.evt"},
	{0x776AC2360F490714ULL, "6153890_m.fxe"},
	{0x776B7FEA63218287ULL, "mcr370_assignments6176749.cl"},
	{0x776B88C5EE89B59DULL, "cs_fen110cs_fenris_intro.wbd"},
	{0x776CD56636FF4FFEULL, "hlac_lt_warehouse_n_-65536_0z.dds"},
	{0x776D2AB9454E770EULL, "vxlt_0_lowtown_docks_cs8_0_0.vlm"},
	{0x776D7B4FF6B8C1EEULL, "vxlt_gallowstemplar_5p5_0_0.vlm"},
	{0x776E55025C16CF0FULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_enter.gad"},
	{0x776EE18DE82279D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_enter.evt"},
	{0x776EEB2A304B0107ULL, "one111ip_ggt_note.ncs"},
	{0x776F163C55F31272ULL, "w_blightlands_141_0.phy"},
	{0x776F41E4E9CB659FULL, "552150409.ogg"},
	{0x776F7093DD2305D0ULL, "mag131_taraneh6088729.cl"},
	{0x776F83717ACFCBB4ULL, "hlac_ht_brothel_n_3373_0y.dds"},
	{0x776F8E2C80BCF371ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_meda.phy"},
	{0x776F96B38190F9AEULL, "vxw_woundedcoast_1il_0_1.vlm"},
	{0x776FAD6622D7A7B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_cry.evt"},
	{0x77700D1F728B624BULL, "plt_cod_hst_mag_dragonblood.plo"},
	{0x777014B0A3D38CF0ULL, "186314465.ogg"},
	{0x7770795A710A73CEULL, "vxca_caverns2_7nu_0_0.vlm"},
	{0x77708E2785BA77CDULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.cess.gad"},
	{0x7770C02DB2F1F55AULL, "vxw_encounter2_4j_0_0.vlm"},
	{0x77718947A06F4EC8ULL, "fxa_web_distortion.mao"},
	{0x7771AAA7C8F243E6ULL, "llac_lt_0_lowtown_65538_4.dds"},
	{0x7771CD56E1A36F0CULL, "art\\vfx\\combat\\vfx_g_spear_whirlwind_c.impa.gad"},
	{0x7772390733CA541EULL, "hlac_dr_slavers_3207_2x.dds"},
	{0x777241BBA802CB61ULL, "147043271.ogg"},
	{0x7772AF29A33CDAE7ULL, "one300ar_vanard_fight.are"},
	{0x7773DAD0F252A8B3ULL, "combat_smaoe_range.gda"},
	{0x7773FFF758396983ULL, "6032279_m.fxe"},
	{0x77746A025C1B5EBCULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_enter.evt"},
	{0x77751DB08C97ABB3ULL, "309252549.ogg"},
	{0x7775201A172CC6FDULL, "hlac_ht_mansion2_818_5z.dds"},
	{0x777535EDED0E2710ULL, "dae000_keep_noblewoman_sta.cnv"},
	{0x7775596B84D119F2ULL, "6164212_m.fxe"},
	{0x77759DBBA7293E28ULL, "qcr000_aveline6054365.cl"},
	{0x777640FA4D29101EULL, "6167208_m.fxe"},
	{0x777692DB83F8D6ECULL, "vxht_mansion2_1u7_0_0.vlm"},
	{0x7776CFEAFA1791D2ULL, "6033982_f.fxe"},
	{0x77771702E9540D4DULL, "436202257.ogg"},
	{0x77771735DA3C29F7ULL, "694188225.ogg"},
	{0x77778A9674AA1885ULL, "rdr111dg_assistant.bnk"},
	{0x7777BC831DE023A4ULL, "fen110_fenris6100244.cl"},
	{0x7777CC56E1A7D90AULL, "art\\vfx\\combat\\vfx_g_spear_whirlwind_c.impa.ani"},
	{0x7778127D7081F879ULL, "6117907_m.fxe"},
	{0x77787E31017DC5A2ULL, "vxht_keep_67h_0_0.vlm"},
	{0x77788038001004ADULL, "587147474.ogg"},
	{0x77788F76A64C757AULL, "hlac_lt_gallowscourtyard_n_-2_2z.dds"},
	{0x7778EF53737628AEULL, "117287462.ogg"},
	{0x77792284ED80C76AULL, "26501574.ogg"},
	{0x7779264376721DD2ULL, "6028404_m.fxe"},
	{0x777932590A461097ULL, "ai_abilities_cond.gda"},
	{0x77794A0F05AC2BBDULL, "6107682_f.fxe"},
	{0x777957EDED124A68ULL, "dae000_keep_noblewoman_sta.bnk"},
	{0x7779B4A81B7526E7ULL, "6148695_m.fxe"},
	{0x7779E7C91BFF372BULL, "vxlt_darktown_5nr_0_0.vlm"},
	{0x777A48FAE00D450DULL, "vxlt_0_lowtown_docks_a3_d33_0_0.vlm"},
	{0x777A6C9674AC1BBDULL, "rdr111dg_assistant.cnv"},
	{0x777ACEF42F85C32FULL, "vxht_chantry_n_2fa_0_0.vlm"},
	{0x777B1B14598FAA83ULL, "76680282.ogg"},
	{0x777B28ABC5D5024EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_exit.ani"},
	{0x777B3FF47A9B9DE0ULL, "vxht_bartrand_n_54l_0_0.vlm"},
	{0x777B7F3DAF2FDA53ULL, "vxdr_primevaltomb_7lh_0_0.vlm"},
	{0x777BA91495CA4549ULL, "vxlt_0_lowtown_a3_n_c_gpg_0_0.vlm"},
	{0x777BC7E203A73279ULL, "vxw_woundedcoast_1k_0_0.vlm"},
	{0x777BF98DE82D2CA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_enter.ani"},
	{0x777C1A7F195AF7B3ULL, "gen00fl_isabela6089525.cl"},
	{0x777C67B26D89AA84ULL, "w_bonepitmining_a2_189_0.msh"},
	{0x777C79B26D89C9EDULL, "w_bonepitmining_a2_189_0.mao"},
	{0x777C7DB26D89D026ULL, "w_bonepitmining_a2_189_0.mmh"},
	{0x777CF24A9BF29675ULL, "1069416761.ogg"},
	{0x777D1C8CA7686DB4ULL, "514256919.ogg"},
	{0x777D35D8FE04AD8DULL, "vxlt_0_lowtown_docks_a3_apk_0_0.vlm"},
	{0x777D48585239D6ECULL, "hlac_lt_0_lowtown_a2_n_-65536_6z.dds"},
	{0x777D7518A67C5852ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platform.phy"},
	{0x777DF2AAC22A6C0BULL, "follower_banter6167112.cl"},
	{0x777F8566CF2C425AULL, "vxw_blightlands_cp_0_0.vlm"},
	{0x777FBB1C6ADBBE29ULL, "dae000cr_grand_cleric6147888.cl"},
	{0x77801850446F0668ULL, "dae000_keep_guard_1_sta6163692.cl"},
	{0x7780AD294D252DC5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wic_0s.dds"},
	{0x7780C00FF83940D3ULL, "llac_w_bonepitmining_a2_0_10.dds"},
	{0x7780FE593EE8989DULL, "one000_ilen6037066.cl"},
	{0x77814F44AE3D158FULL, "vxlt_gamlenshouse_n_jy_0_0.vlm"},
	{0x778151F7334FF694ULL, "mag101_grace6053060.cl"},
	{0x77815F0000B51464ULL, "6037931_f.fxe"},
	{0x77818F6FC3C91BF7ULL, "dae000_pc_home_ambients6208676.cl"},
	{0x7781B59096DA43EDULL, "vxht_0_hightown_bxz_0_0.vlm"},
	{0x7781F889AC27266EULL, "vxlt_0_lowtown_a2_hgn_0_0.vlm"},
	{0x7781FF4877DA5DA4ULL, "6050077_f.fxe"},
	{0x778261FF57BCDCBBULL, "vxdr_0_deeproads_9ku_0_0.vlm"},
	{0x77826A010B87911CULL, "mag311_emile_de_launcet6043454.cl"},
	{0x77826AA0CF5AB198ULL, "hlac_lt_gallowscourtyard_n_-65535_0x.dds"},
	{0x778272025C27A9D5ULL, "art\\characters\\creatures\\stonegolem\\animation\\kd_enter.ani"},
	{0x7782AD86D4FF6E50ULL, "269589269.ogg"},
	{0x778315B0C8DB945AULL, "zz_followers_banter6157309.cl"},
	{0x77832EE825435CF1ULL, "vxht_0_hightown_apm_0_0.vlm"},
	{0x7783532C8042D7D9ULL, "one220_yevhen6120899.cl"},
	{0x77838B11D17BFA31ULL, "6042088_f.fxe"},
	{0x7783AF7246C96C8AULL, "mag160_varian_ilithis6171468.cl"},
	{0x778402B8A021B8F4ULL, "vxht_brothel_n_3fd_0_0.vlm"},
	{0x77846C416168965DULL, "genip_pp_gli_one_shot.utp"},
	{0x77847F58086EAEF6ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrypot.anb"},
	{0x7784FFF3708478D2ULL, "vxdr_primevaltomb_7lm_0_0.vlm"},
	{0x778504B875EEEF6CULL, "vxht_mansion2_1x4_0_0.vlm"},
	{0x7785AFB47C1113A5ULL, "vxw_bonepit_xu_0_0.vlm"},
	{0x7785CD5F4AF259D4ULL, "llac_lt_darktown_-1_20.dds"},
	{0x7786160F7DACE309ULL, "lgt260_way_blood_mage6148357.cl"},
	{0x7786437DF13C6E30ULL, "lt_0_lowtown_docks_153_0.phy"},
	{0x778674B26D9215F1ULL, "w_bonepitmining_a2_189_0.phy"},
	{0x778677BBE1FE1AA4ULL, "l_env_wal_gallows_int_military_day.bnk"},
	{0x7786F00DF373DE22ULL, "vxht_keep_60i_0_0.vlm"},
	{0x778770C08FC9204DULL, "6111123_m.fxe"},
	{0x7787BC9D30B55F23ULL, "vxht_keep_4hv_0_0.vlm"},
	{0x77884ACC0778DA31ULL, "vxw_bonepit_1ln_0_0.vlm"},
	{0x778852998CAA066DULL, "223095967.ogg"},
	{0x778871A43C70AC8AULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_pc_.gad"},
	{0x7788A05AAA8FA063ULL, "dae200cr_viscount6087218.cl"},
	{0x7788A7EE4CBF28E0ULL, "216263468.ogg"},
	{0x7788A89E9BB6C16EULL, "w_blightlands_254_0.msh"},
	{0x7788B69E9BB6D8BFULL, "w_blightlands_254_0.mao"},
	{0x7788BAD7D2035D8DULL, "vxlt_0_lowtown_docks_a3_d42_0_0.vlm"},
	{0x7788C29E9BB6ED5CULL, "w_blightlands_254_0.mmh"},
	{0x7788EC2C11DEC383ULL, "593114000.ogg"},
	{0x778920ABC5E1319FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_exit.evt"},
	{0x778934159E9649D0ULL, "hla_ht_0_hightown_65537_0z.dds"},
	{0x7789709DE19FB71CULL, "w_bonepitdestroyed_a3_70_0.mao"},
	{0x77897C9DE19FCBBFULL, "w_bonepitdestroyed_a3_70_0.mmh"},
	{0x7789829DE19FD661ULL, "w_bonepitdestroyed_a3_70_0.msh"},
	{0x7789F62C40DA2682ULL, "hlac_lt_0_lowtown_docks_a3_-65540_1y.dds"},
	{0x778A641276578B6DULL, "vxlt_gallowstemplar_5qu_0_0.vlm"},
	{0x778A7D3B3DD6610EULL, "207801109.ogg"},
	{0x778B5A87B1AEEF66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_reaction.ani"},
	{0x778B701A2BBAF52CULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv4.ani"},
	{0x778C46B6D776886CULL, "866221645.ogg"},
	{0x778C4EDEACCFB4C5ULL, "gen00fl_fenris6107179.cl"},
	{0x778CCEA2D1D0A142ULL, "420329673.ogg"},
	{0x778D0D53D3BB0961ULL, "hlac_dr_slavers_3215_0z.dds"},
	{0x778D1DCBBF643718ULL, "hlac_w_woundedcoast_65536_0z.dds"},
	{0x778DF743C1225ADAULL, "vxw_blightlands_dq_0_0.vlm"},
	{0x778DF9EB808AA424ULL, "6121051_m.fxe"},
	{0x778E563A12EB24C2ULL, "6116587_m.fxe"},
	{0x778F87BF370B56A7ULL, "vxlt_blackemporium_134_0_0.vlm"},
	{0x778F8CED80DC8649ULL, "191792585.ogg"},
	{0x778FC121A0332E0FULL, "vxlt_gamlenshouse_n_mx_0_0.vlm"},
	{0x778FE8B48670EE9CULL, "mag291_ambient_26218725.cl"},
	{0x779001ABC5E71B88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\bloodmage_cast_exit.gad"},
	{0x77900694528F28ECULL, "gen_im_trs_gem_ruby_02.uti"},
	{0x779030A2C18035E0ULL, "zz_followers_banter6158856.cl"},
	{0x77903A31DF13891AULL, "526294992.ogg"},
	{0x779080B2E4779B21ULL, "dae300_alistair_king6133528.cl"},
	{0x7790FBBF037CC718ULL, "vxca_cave_gx_0_0.vlm"},
	{0x77913643AF45A42CULL, "6087218_m.fxe"},
	{0x77913704342F44E7ULL, "6179061_f.fxe"},
	{0x77916E1809287F80ULL, "qun110_petrice6018637.cl"},
	{0x7791AA1AC39541F1ULL, "vxca_caverns2_9fh_0_0.vlm"},
	{0x779250D0D13333BCULL, "gen00fl_anders6119436.cl"},
	{0x7792634CD4936C72ULL, "6039275_m.fxe"},
	{0x77927D1A2BC12A6EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv4.gad"},
	{0x77928258087AF161ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_foundrypot.mmh"},
	{0x77928FFCACBBB536ULL, "6076328_bethany_intro.cut"},
	{0x779295643D80E69DULL, "llac_ht_0_hightown_a3_-131074_0.dds"},
	{0x7793769567E507ECULL, "vxht_mansion2_1c5_0_0.vlm"},
	{0x7793F54338E834FAULL, "vxht_keep_5ss_0_0.vlm"},
	{0x7793FBA55C7F43DFULL, "gen00fl_anders6125311.cl"},
	{0x7794334386D88C95ULL, "6045965_m.fxe"},
	{0x77944B0E34393B06ULL, "6077417_m.fxe"},
	{0x7794837C9B180BB1ULL, "llac_lt_0_lowtown_docks_a3_n_-196611_0.dds"},
	{0x77950F36F4F1707CULL, "gen00fl_isabela6118356.cl"},
	{0x779517B736C5A0C6ULL, "vxlt_gallowstemplar_n_5cw_0_0.vlm"},
	{0x779569AC27BBD51BULL, "308410207.ogg"},
	{0x77963EFEFBDDCC7AULL, "6163222_m.fxe"},
	{0x7796680F98F3C6A9ULL, "vxlt_0_lowtown_by1_0_0.vlm"},
	{0x7796E72D64730A1CULL, "art\\characters\\playercharacter\\elffemale\\animation\\blendtree\\merril_exp_movement.anb"},
	{0x779843788272671FULL, "6120987_m.fxe"},
	{0x7798793789EE6D06ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.dura.evt"},
	{0x779898E3216F0797ULL, "hlac_lt_0_lowtown_a3_n_c_65534_4y.dds"},
	{0x779909DC3996953EULL, "6165277_m.fxe"},
	{0x779931684A65F080ULL, "vxht_keep_n_3cp_0_0.vlm"},
	{0x77994287B1BB03D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_reaction.evt"},
	{0x7799781A2BC7404DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv4.evt"},
	{0x7799C900DA85A3B2ULL, "hf_isabela_b.mor"},
	{0x779B137E699F30EDULL, "zz_followers_banter6158727.cl"},
	{0x779B5BC2B5550343ULL, "vxlt_gallowstemplar_51k_0_0.vlm"},
	{0x779C4CA43C811770ULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_pc_.ani"},
	{0x779C8D1E153DE37FULL, "6110647_m.fxe"},
	{0x779CB52520387D03ULL, "vxht_0_hightown_n_b94_0_0.vlm"},
	{0x779D0495B7C92AA3ULL, "124878050.ogg"},
	{0x779D3810E878B855ULL, "685237899.ogg"},
	{0x779D5EF38FBA1260ULL, "6148843_m.fxe"},
	{0x779D67D0E400EBD7ULL, "vxlt_hangedman_n_252_0_0.vlm"},
	{0x779DB6122D0C9932ULL, "6096179_m.fxe"},
	{0x779DEAB2C6CA180CULL, "790697279.ogg"},
	{0x779DF99C29016F27ULL, "vxlt_blackemporium_167_0_0.vlm"},
	{0x779E038CC5048A6DULL, "203126757.ogg"},
	{0x779E29A4086D42DCULL, "781844730.ogg"},
	{0x779F0FBBA1DB4B4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_up.gad"},
	{0x779F78545BC6E8AAULL, "w_encounter3_75_0.mao"},
	{0x779F7C545BC6EFF9ULL, "w_encounter3_75_0.mmh"},
	{0x779F86545BC700E7ULL, "w_encounter3_75_0.msh"},
	{0x779F903789F4B351ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.dura.gad"},
	{0x77A0284B4F658218ULL, "534439707.ogg"},
	{0x77A04387B1C12460ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_reaction.gad"},
	{0x77A047E5EFD5100FULL, "vxdr_primevaltomb_2li_0_0.vlm"},
	{0x77A08E03A6B8D116ULL, "vxca_caverns1_mines_k2o_0_0.vlm"},
	{0x77A090FDDF57C821ULL, "639375604.ogg"},
	{0x77A0BA3885EE4776ULL, "vxca_caverns2prc_9ff_0_0.vlm"},
	{0x77A0E672840DE9F4ULL, "vxht_brothel_n_3hj_0_0.vlm"},
	{0x77A14A7F886CF93FULL, "vxht_mansion2_1u8_0_0.vlm"},
	{0x77A1E3AD5470A9D2ULL, "vxdr_primevaltomb_7bo_0_0.vlm"},
	{0x77A23B93487C4096ULL, "hlac_ht_bartrand_n_3365_0z.dds"},
	{0x77A2936E5FFD44A5ULL, "vxw_bonepit_ns_0_0.vlm"},
	{0x77A29E662BD246C2ULL, "plt_and000pt_gift_key.nss"},
	{0x77A2C1FCD6297FF2ULL, "vxlt_0_lowtown_a2_n_c_ere_0_0.vlm"},
	{0x77A325178D8BFC30ULL, "hf_templar01.mor"},
	{0x77A37357CA1ED079ULL, "llac_ht_chantry_2534_33.dds"},
	{0x77A421768CAF1F59ULL, "gen00fl_isabela6089538.cl"},
	{0x77A49982341E3BCDULL, "gen_im_arm_cht_hvy_act2_02.uti"},
	{0x77A4AF511AE5F4CBULL, "770371822.ogg"},
	{0x77A570AEF7E7F769ULL, "964529754.ogg"},
	{0x77A5913789F920FFULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.dura.ani"},
	{0x77A5C57A832EA46FULL, "ch2alphem.psh.fxo"},
	{0x77A5E870561B112CULL, "6077006_m.fxe"},
	{0x77A5F2BBA1E138E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_up.evt"},
	{0x77A6F56C68B034AAULL, "and100_anders6095963.cl"},
	{0x77A71A987B022F99ULL, "6080210_m.fxe"},
	{0x77A7F643D4A46383ULL, "hlac_dr_slavers_3214_12z.dds"},
	{0x77A7FE231B3DB6AAULL, "6081654_m.fxe"},
	{0x77A80E8A96E0CDE5ULL, "1011097969.ogg"},
	{0x77A84957AF0B2183ULL, "vxdr_primevalentrance_375_0_0.vlm"},
	{0x77A8E99B146A9B64ULL, "vxca_caverns1_mines_i2w_0_0.vlm"},
	{0x77A955BDF3923924ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstr.msh"},
	{0x77A967BDF392578DULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstr.mao"},
	{0x77A96BBDF3925EC6ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstr.mmh"},
	{0x77A9C4B26C8203FAULL, "vxca_caverns2_8ol_0_1.vlm"},
	{0x77A9F73C6876B895ULL, "t3_shd_elm.tnt"},
	{0x77A9FBDA0C2F5558ULL, "vxlt_0_lowtown_n_e2u_0_0.vlm"},
	{0x77AA324813AA8997ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1cw_basic.phy"},
	{0x77AA82014F6011EDULL, "128600471.ogg"},
	{0x77AADAFDA50E8BDAULL, "vxw_blightlands_rs_0_0.vlm"},
	{0x77AAFA10C13E3607ULL, "6106990_m.fxe"},
	{0x77AAFA2FAD54DDA3ULL, "art\\vfx\\cinematic\\vfx_bld_punch_ds_p.impa.ani"},
	{0x77AB02700BB666EEULL, "lt_0_lowtown_docks_110_0.msh"},
	{0x77AB10700BB67E3FULL, "lt_0_lowtown_docks_110_0.mao"},
	{0x77AB1C700BB692DCULL, "lt_0_lowtown_docks_110_0.mmh"},
	{0x77AC14D5C70DA39EULL, "vxw_encounter2_77_0_0.vlm"},
	{0x77AC38A4CCE5B46BULL, "dae300_alistair_king6133830.cl"},
	{0x77AC53143C9ADA81ULL, "zz_fol_debug6066139.cl"},
	{0x77ACD31571308E94ULL, "6071101_m.fxe"},
	{0x77ACEC0A915310A5ULL, "6173048_m.fxe"},
	{0x77AD72A6D15FCF29ULL, "vxht_0_hightown_a2_a4m_0_0.vlm"},
	{0x77AD82D53B018173ULL, "art\\vfx\\combat\\vfx_glb_ele_c_additive_c.anb"},
	{0x77AE143DFAEA8F47ULL, "331377995.ogg"},
	{0x77AE5FD91D1F3134ULL, "hlac_ca_cavern1_haunted_24424_19x.dds"},
	{0x77AF7B29C5369711ULL, "6024106_m.fxe"},
	{0x77AF80ACEF2B5209ULL, "6102880_m.fxe"},
	{0x77AFFA12062562C6ULL, "vxdr_0_deeproads_api_0_0.vlm"},
	{0x77B0B182851BC8DDULL, "vxlt_undercitydungeon_1a6_0_0.vlm"},
	{0x77B0EE38148B62DCULL, "gen00fl_anders6215068.cl"},
	{0x77B11DBD750B357CULL, "hlac_lt_undercityserialkiller_-1_51z.dds"},
	{0x77B2E6F0241CCB46ULL, "vxdr_0_deeproads_d31_0_0.vlm"},
	{0x77B311364210F553ULL, "vxlt_0_lowtown_i4c_0_0.vlm"},
	{0x77B320BC02F589ECULL, "6042132_m.fxe"},
	{0x77B3CF3E16CE6AB4ULL, "98600151.ogg"},
	{0x77B3EABBA1ED6830ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_up.ani"},
	{0x77B4193775D007E2ULL, "950860301.ogg"},
	{0x77B45A4D481129ACULL, "6078106_m.fxe"},
	{0x77B462BDF39C5691ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_candlclstr.phy"},
	{0x77B5790E6828157CULL, "art\\vfx\\environment\\vfx_water_splash_s_p.dura.ani"},
	{0x77B65D54616835A9ULL, "demo_mage.utc"},
	{0x77B6767DD2C37110ULL, "hlac_lt_gallowstemplar_n_2383_5z.dds"},
	{0x77B6974FFA39EBF1ULL, "qcr000_arishok6166983.cl"},
	{0x77B6B7BFE80B7D0BULL, "403099257.ogg"},
	{0x77B7C5379CF6A29AULL, "mag321cr_servant_gossip_1a.utc"},
	{0x77B7D0C6EEE45D40ULL, "vxw_sundermount_2cz_0_0.vlm"},
	{0x77B8055B27495A46ULL, "6108553_m.fxe"},
	{0x77B82A193F0EF245ULL, "llac_lt_0_lowtown_a2_n_-2_6.dds"},
	{0x77B87ECD6D9CB1FCULL, "gen00fl_merrill6108696.cl"},
	{0x77B8940054199AC6ULL, "431224709.ogg"},
	{0x77B8B6FC1D62A164ULL, "631313775.ogg"},
	{0x77B8DE1146F26457ULL, "vxht_bartrand_n_4b4_0_0.vlm"},
	{0x77B9122FAD61439AULL, "art\\vfx\\cinematic\\vfx_bld_punch_ds_p.impa.evt"},
	{0x77B94CDA9704A45AULL, "vxw_blightlands_wl_0_0.vlm"},
	{0x77B98A1ECCADFEDDULL, "6084720_m.fxe"},
	{0x77BA838B757C6725ULL, "vxlt_0_lowtown_a3_n_c_e8i_0_0.vlm"},
	{0x77BB0D4CA8DD326BULL, "vxw_bonepit_1r6_0_0.vlm"},
	{0x77BB4AAF17D464A8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medb_0n.dds"},
	{0x77BB71A90F4EEA76ULL, "6101114_m.fxe"},
	{0x77BB9FD53B0DEFDCULL, "art\\vfx\\combat\\vfx_glb_ele_c_additive_c.mmh"},
	{0x77BBE0F0BFEE255FULL, "ico_injury.dds"},
	{0x77BBF0A30CACE0E1ULL, "vxlt_gallowscourtyard_7bt_0_0.vlm"},
	{0x77BC1E5C11252FECULL, "555450503.ogg"},
	{0x77BC6D6B3E92C3AEULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.mmh"},
	{0x77BC860E682E49BEULL, "art\\vfx\\environment\\vfx_water_splash_s_p.dura.gad"},
	{0x77BC91EC036593D2ULL, "vxlt_0_lowtown_i4f_0_0.vlm"},
	{0x77BCACB6278F807BULL, "799829775.ogg"},
	{0x77BE8902F742BDE7ULL, "fen110cr_anso.utc"},
	{0x77BE92A30E1539C9ULL, "544934475.ogg"},
	{0x77BED2F3D8A05093ULL, "268253247.ogg"},
	{0x77BF235F7711E15DULL, "vxlt_undercitydungeon_1d5_0_0.vlm"},
	{0x77BF8EB540242C91ULL, "218766247.ogg"},
	{0x77BFC06CFC2CC3D9ULL, "and121ar_chantry.are"},
	{0x77C007A90D0D0C0EULL, "one000_ilen6205648.cl"},
	{0x77C0112FAD67608DULL, "art\\vfx\\cinematic\\vfx_bld_punch_ds_p.impa.gad"},
	{0x77C03B03A570D10CULL, "vxlt_0_lowtown_a2_e26_0_0.vlm"},
	{0x77C0D2CBB9D4D12DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_trapdoor_n.dds"},
	{0x77C2D6349997B43BULL, "6050067_f.fxe"},
	{0x77C2DC9A37D39EC8ULL, "llac_lt_gallowscourtyard_n_-1_3.dds"},
	{0x77C30B603DE9F886ULL, "vxlt_gallowscourtyard_n_75m_0_0.vlm"},
	{0x77C35344C37CEC21ULL, "hlac_lt_0_lowtown_docks_n_131070_3y.dds"},
	{0x77C3610E683429DDULL, "art\\vfx\\environment\\vfx_water_splash_s_p.dura.evt"},
	{0x77C3746CC36D384AULL, "641759644.ogg"},
	{0x77C3A93039ED5639ULL, "6202359_m.fxe"},
	{0x77C3D82A3093215AULL, "one111ip_ggt_note.utp"},
	{0x77C3E76A6248C984ULL, "mag170_thrask6054051.cl"},
	{0x77C45407812EB1DEULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.cess.ani"},
	{0x77C533B401A4A56EULL, "mag291_vincento6027337.cl"},
	{0x77C57958CE018CC8ULL, "vxlt_gallowscourtyard_7bq_0_0.vlm"},
	{0x77C5BC7369C817C4ULL, "6125373_f.fxe"},
	{0x77C5C63157A330F8ULL, "zz_lgt_3_debug6176138.cl"},
	{0x77C5CA4A47AAD350ULL, "6137614_m.fxe"},
	{0x77C73B2CEE3A2091ULL, "6065967_m.fxe"},
	{0x77C74698B4701910ULL, "vxht_0_hightown_n_avm_0_0.vlm"},
	{0x77C7688089DF049FULL, "6131156_m.fxe"},
	{0x77C79BFF9EF3B456ULL, "vxht_0_hightown_a6p_0_0.vlm"},
	{0x77C7B4318C38469DULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgta.phy"},
	{0x77C82A729774D11AULL, "281877199.ogg"},
	{0x77C844D2596B9833ULL, "39067176.ogg"},
	{0x77C8B6A93826E879ULL, "vxw_bonepitmining_a2_1ou_0_0.vlm"},
	{0x77C9277F55F8E2CEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperleft.phy"},
	{0x77C9772B7BF527F0ULL, "gen00fl_isabela6075224.cl"},
	{0x77C9B2BE07877884ULL, "mage_deathmagic.dds"},
	{0x77C9CF987FB95DF1ULL, "6070125_m.fxe"},
	{0x77CA627FFEA2F961ULL, "vxlt_gallowscourtyard_7gk_0_0.vlm"},
	{0x77CA8C6B3E9F35ADULL, "art\\vfx\\combat\\vfx_q_ari_appearance_c.anb"},
	{0x77CAAD64E1149869ULL, "vxdr_slavers_13y_0_0.vlm"},
	{0x77CAEA306301BB84ULL, "vxht_brothel_n_6ba_0_0.vlm"},
	{0x77CAFCDD5F85CAE5ULL, "6113891_m.fxe"},
	{0x77CB1B3949F3AFE9ULL, "vxht_0_hightown_n_aln_0_0.vlm"},
	{0x77CB3AD5083C8108ULL, "432873351.ogg"},
	{0x77CB4B91B6BB1315ULL, "hlac_dr_0_deeproads_5845_12x.dds"},
	{0x77CC1699FC86BCCEULL, "924504266.ogg"},
	{0x77CC280E5642B9A2ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_4z.dds"},
	{0x77CC528C4E7A468EULL, "dae000_orsino6081363.cl"},
	{0x77CCC41C0BE210D4ULL, "6171582_m.fxe"},
	{0x77CD09BFEF7AF0D1ULL, "vxw_woundedcoast_1g4_0_0.vlm"},
	{0x77CD1B6D63F03534ULL, "gen_im_trs_gem_cub_zirc_03.uti"},
	{0x77CD1F3E62AE92A0ULL, "darktown_fem_fer_urch_sta6174801.cl"},
	{0x77CD39442A52F352ULL, "vxdr_primevaltomb_7uj_0_0.vlm"},
	{0x77CD5BCCB146CE07ULL, "6164503_m.fxe"},
	{0x77CD5EC42F52DCC5ULL, "vxht_keep_5vg_0_0.vlm"},
	{0x77CD81F41D1CEF87ULL, "6033544_m.fxe"},
	{0x77CD953C6907F9DDULL, "vxlt_undercitydungeon_1o4_0_0.vlm"},
	{0x77CDFA5AD9EC48B5ULL, "ser221_party_comments6166199.cl"},
	{0x77CE0A96C428C2A7ULL, "6116530_f.fxe"},
	{0x77CEAC8F63EB433EULL, "vxht_mansion2b_1kg_0_0.vlm"},
	{0x77CEAFDA2F96E78DULL, "cd_vrc221_bartrand_fake.wbd"},
	{0x77CEEC6FA68B3D03ULL, "vxlt_0_lowtown_ewf_0_0.vlm"},
	{0x77CFF1372638F380ULL, "vxlt_0_lowtown_gbl_0_0.vlm"},
	{0x77D0D943EFAC59FDULL, "one110_meeran6090193.cl"},
	{0x77D124E10A59EA8CULL, "vxca_caverns1_ikm_0_0.vlm"},
	{0x77D153E8A6560E69ULL, "mi_gla_cyc.mao"},
	{0x77D16C07813964AFULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.cess.evt"},
	{0x77D1750ED46F7D0FULL, "vxlt_0_lowtown_docks_a3_c3z_0_0.vlm"},
	{0x77D281C55281B11FULL, "6168024_m.fxe"},
	{0x77D28B662BFB4A45ULL, "plt_and000pt_gift_key.plo"},
	{0x77D2A0316C351BC5ULL, "vxht_0_hightown_n_ccm_0_0.vlm"},
	{0x77D2A1EAC32578B8ULL, "dae000_sabina6162327.cl"},
	{0x77D3246FC57F6FAEULL, "vxlt_warehouse_n_8l_0_0.vlm"},
	{0x77D3EB35BFF7A548ULL, "vxlt_gallowscourtyard_7gn_0_0.vlm"},
	{0x77D423CD638BE06FULL, "dae000_gamlen6102533.cl"},
	{0x77D47DE9008DB72AULL, "6027328_m.fxe"},
	{0x77D47E91F3DD88E9ULL, "hlac_ht_0_hightown_n_-65538_1x.dds"},
	{0x77D4B480D2D08E40ULL, "vxw_sundermount_2ix_0_0.vlm"},
	{0x77D54DAC3AA53A79ULL, "llac_lt_0_lowtown_a2_65535_4.dds"},
	{0x77D560706FF3E3D1ULL, "mag131_keran6086203.cl"},
	{0x77D5A8301369DEE5ULL, "vxw_blightlands_gw_0_0.vlm"},
	{0x77D5C1CB2ADE9557ULL, "vxht_bartrand_n_4l6_0_0.vlm"},
	{0x77D6B4B05EE24B99ULL, "518978582.ogg"},
	{0x77D6C05556BA6D1DULL, "6108037_m.fxe"},
	{0x77D73DAAC332BD5BULL, "6114827_m.fxe"},
	{0x77D7D1799D409D83ULL, "6132867_m.fxe"},
	{0x77D802AC46060347ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantryb_d.dds"},
	{0x77D84678DBB54B72ULL, "zz_lgt_2_debug6206609.cl"},
	{0x77D86D07813F8558ULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.cess.gad"},
	{0x77D89B58D4C35C6EULL, "ser221_gamlen6133139.cl"},
	{0x77D8A9318C46A709ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgta.mao"},
	{0x77D8AACD6944CC91ULL, "hlac_ht_playershouse_n_855_35x.dds"},
	{0x77D8B5318C46BBA2ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgta.mmh"},
	{0x77D8B7318C46BF58ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgta.msh"},
	{0x77D8C0F537007B75ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_countthrone_sa.dds"},
	{0x77D8C4FD3C80C0A7ULL, "gen_im_cft_rec_wrn_07.uti"},
	{0x77D91F41D30AB0E9ULL, "vxdr_slavers_18x_0_0.vlm"},
	{0x77D9890E3F3E5CC7ULL, "6058864_m.fxe"},
	{0x77D9C94647285DAAULL, "6110501_m.fxe"},
	{0x77DAEFE73C0202C3ULL, "vxca_caverns2_8fg_0_0.vlm"},
	{0x77DBA173D9CC50EEULL, "vxht_0_hightown_a2_b9f_0_0.vlm"},
	{0x77DBF16EA0DA78E4ULL, "6118617_m.fxe"},
	{0x77DC747553F47084ULL, "vxw_blightlands_9l_0_0.vlm"},
	{0x77DCB04061744F02ULL, "llac_lt_0_lowtown_a2_n_-196608_3.dds"},
	{0x77DCB54043FABF13ULL, "gen_im_npc_arm_hum_glv_thga.uti"},
	{0x77DCB710E63B9E44ULL, "ser270_templar6127409.cl"},
	{0x77DCD17A4B08B6A4ULL, "44200938.ogg"},
	{0x77DD0DD9832532E6ULL, "vxht_0_hightown_a2_aua_0_0.vlm"},
	{0x77DD1EBD895D020CULL, "vxlt_0_lowtown_a2_e84_0_0.vlm"},
	{0x77DD31F84D2EB139ULL, "gen00fl_anders6119259.cl"},
	{0x77DD3855DCD8F2D1ULL, "vxlt_darktown_59u_0_0.vlm"},
	{0x77DD3E63FF29A327ULL, "vxlt_house_gz_0_0.vlm"},
	{0x77DDC55B2932E035ULL, "gen_im_trs_rdr_one_lsw_r1.uti"},
	{0x77DDC84E3AB8995FULL, "rog_sneakattack.dds"},
	{0x77DE394DEFDD6CCCULL, "art\\vfx\\combat\\arm_roba.mao"},
	{0x77DE78BCE0DB4777ULL, "109312229.ogg"},
	{0x77DE91D2C531E9B8ULL, "240569431.ogg"},
	{0x77DEB2942FF4A5DAULL, "lt_0_lowtown_docks_n_149_0.phy"},
	{0x77DF71F11227C6FBULL, "583580820.ogg"},
	{0x77DF8AFAA82D3D5FULL, "6076828_m.fxe"},
	{0x77DF96BDFC50030CULL, "vxca_caverns1_inb_0_0.vlm"},
	{0x77E076553799FFDAULL, "395427255.ogg"},
	{0x77E0B6B08D26BE3BULL, "w_blightlands_127_0.msh"},
	{0x77E0C8B08D26DC86ULL, "w_blightlands_127_0.mao"},
	{0x77E0D4B08D26F125ULL, "w_blightlands_127_0.mmh"},
	{0x77E12DD7D37CEEAAULL, "art\\characters\\weapons\\axes\\textures\\w_axe_001a_0s.dds"},
	{0x77E1979068EFACE7ULL, "6118373_m.fxe"},
	{0x77E1DAC8F1E3FB64ULL, "hla_ht_0_hightown_-131071_0z.dds"},
	{0x77E23B1FA720A351ULL, "375725095.ogg"},
	{0x77E24F45CA5072E2ULL, "vxlt_0_lowtown_n_fe1_0_0.vlm"},
	{0x77E24F84BE96E4DFULL, "mer340_cowan6083529.cl"},
	{0x77E3265DC4C6A6C0ULL, "vxw_sundermount_2ry_0_0.vlm"},
	{0x77E345F211420FF3ULL, "vxlt_undercityencounter_10z_0_0.vlm"},
	{0x77E353FDAAD94F10ULL, "6113582_m.fxe"},
	{0x77E3A2F3CD0BD4E1ULL, "zz_ave_debug6207724.cl"},
	{0x77E41A0D055FF765ULL, "vxw_blightlands_hv_0_0.vlm"},
	{0x77E451B7C7FE9B09ULL, "vxht_mansion_n_61z_0_0.vlm"},
	{0x77E47F5575A77879ULL, "792466995.ogg"},
	{0x77E481D8AF116E7DULL, "676566667.ogg"},
	{0x77E48DD8CAEA7361ULL, "vxlt_0_lowtown_docks_bal_0_0.vlm"},
	{0x77E4D16CFC4BFCAFULL, "and121ar_chantry.ncs"},
	{0x77E4D68148087812ULL, "6128348_m.fxe"},
	{0x77E4E50A55CC3670ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killbag_d.dds"},
	{0x77E4FDF606FFF379ULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation2.gad"},
	{0x77E500CA2C5A00B9ULL, "plt_genpt_app_fenris.plo"},
	{0x77E5B42F1195EBC2ULL, "hlac_dr_0_deeproads_5849_2x.dds"},
	{0x77E5F83C7CA972FAULL, "924799091.ogg"},
	{0x77E62FA5B92A3A1EULL, "vxw_blightlands_o0_0_0.vlm"},
	{0x77E675DA283DDAA1ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\jagged_d.dds"},
	{0x77E6D19B35E15386ULL, "212948510.ogg"},
	{0x77E6D6E8453E236EULL, "vxca_caverns2_835_0_0.vlm"},
	{0x77E77DB937DC455AULL, "872817724.ogg"},
	{0x77E78603C051A1E1ULL, "hlac_lt_0_lowtown_a3_n_c_65536_7x.dds"},
	{0x77E7B9E3C462EC87ULL, "hlac_w_sundermount_262143_0x.dds"},
	{0x77E7FEF52EF2C32AULL, "hlac_lt_0_lowtown_docks_a3_131071_1y.dds"},
	{0x77E827C3A8D87002ULL, "6080154_m.fxe"},
	{0x77E8AE2E25F6E7BDULL, "llac_lt_0_lowtown_n_196607_0.dds"},
	{0x77E8CE39A623C32DULL, "vxht_mansion2_3cn_0_0.vlm"},
	{0x77E9006F3E840636ULL, "zz_approval_debug6003911.cl"},
	{0x77E916FAE805185AULL, "lt_0_lowtown_docks_n_8_0.rml"},
	{0x77E9198F0A1D5751ULL, "vxlt_undercitydungeon_yo_0_0.vlm"},
	{0x77E95786D962491CULL, "6189044_m.fxe"},
	{0x77E961C42DF81B43ULL, "vxca_caverns2_8mf_0_0.vlm"},
	{0x77E9773DC3E6806AULL, "vxht_mansion_n_6jh_0_0.vlm"},
	{0x77E99113683FF832ULL, "6151102_m.fxe"},
	{0x77E9BDB14B436E82ULL, "lt_0_lowtown_docks_a3_n_29_0_ll.mmh"},
	{0x77EA1A50B579FCE2ULL, "hlac_lt_0_lowtown_a2_n_65536_5x.dds"},
	{0x77EAB2EB3D7B5EAAULL, "mrl150_merrill6070983.cl"},
	{0x77EACDFABB8CBC15ULL, "lla_ht_0_hightown_n_0_0.dds"},
	{0x77EAE389E44879B0ULL, "698890101.ogg"},
	{0x77EAED89B4CBCAB3ULL, "ch2notnt.psh.fxo"},
	{0x77EB0EF607047BC7ULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation2.ani"},
	{0x77EB562EE4AD4B7CULL, "844347435.ogg"},
	{0x77EB9ECAB4DC9590ULL, "horizontalblur.vsh.fxo"},
	{0x77EBD0298A776E03ULL, "vxlt_0_lowtown_emh_0_0.vlm"},
	{0x77EBD16CFC521BFEULL, "and121ar_chantry.lst"},
	{0x77EBE23571BCD3D6ULL, "6134801_f.fxe"},
	{0x77ECE5ADA9948318ULL, "vxdr_0_deeproads_cp5_0_0.vlm"},
	{0x77ED06196021F515ULL, "zz_vault_debug6200349.cl"},
	{0x77ED17B8D5D90919ULL, "gen00fl_fenris6107236.cl"},
	{0x77ED8B7877782AD2ULL, "133085680.ogg"},
	{0x77EDE08FE3F8E5B0ULL, "lt_gallowscourtyard_50_0.rml"},
	{0x77EDFC76D1C6C832ULL, "9363360.ogg"},
	{0x77EE168E8C3F1F48ULL, "vxlt_0_lowtown_docks_bai_0_0.vlm"},
	{0x77EE70EEC2FC704CULL, "6149911_f.fxe"},
	{0x77EF0957E234363CULL, "seb221_ruxton_harimann6187816.cl"},
	{0x77EF24A7351E57A9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_meda_0t.dds"},
	{0x77EF598A88B550CFULL, "follower_banter6155580.cl"},
	{0x77EFA7F1B0E7B95BULL, "hlac_w_bonepit_458754_0x.dds"},
	{0x77F05FB9ED3E5424ULL, "6111052_m.fxe"},
	{0x77F067186D854796ULL, "6068035_m.fxe"},
	{0x77F0F393D60A6B3DULL, "vxca_cavern1_haunted_fmt_0_0.vlm"},
	{0x77F1074813E643EFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1cw_basic.mao"},
	{0x77F1134813E6578CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1cw_basic.mmh"},
	{0x77F1194813E6627EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1cw_basic.msh"},
	{0x77F1694356F66F3AULL, "268613970.ogg"},
	{0x77F1983AB6BCBF40ULL, "vxw_sundermount_2wf_0_0.vlm"},
	{0x77F1AF9E73574F0CULL, "559155907.ogg"},
	{0x77F22224E5EF8D7AULL, "583793438.ogg"},
	{0x77F24DA3F901EDE8ULL, "rune_armor_electric.dds"},
	{0x77F27EDF7CDB7799ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_bdaa_0d.dds"},
	{0x77F281EEA47622CBULL, "and110_refugee6111614.cl"},
	{0x77F28BE9F7560FE5ULL, "vxw_blightlands_mu_0_0.vlm"},
	{0x77F2FFB5BCE08BE1ULL, "vxlt_0_lowtown_docks_bdc_0_0.vlm"},
	{0x77F3597168053943ULL, "734303755.ogg"},
	{0x77F39325BCB28569ULL, "vxdr_primevalentrance_2zm_0_0.vlm"},
	{0x77F3A40F3AA69E1DULL, "6106237_m.fxe"},
	{0x77F3B73A23090003ULL, "vxca_cave_1ad_0_0.vlm"},
	{0x77F4042190CA5F58ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_basic.phy"},
	{0x77F44BED0AA5F60CULL, "vxht_0_hightown_n_bz2_0_0.vlm"},
	{0x77F49D2F79C7B661ULL, "vxht_0_hightown_n_cb6_0_0.vlm"},
	{0x77F4A182AB20529EULL, "vxw_blightlands_p1_0_0.vlm"},
	{0x77F4DE02896FC60BULL, "follower_banter6141844.cl"},
	{0x77F5374FFF9C891DULL, "gen00fl_merrill6141128.cl"},
	{0x77F5466997CAEBD4ULL, "hlac_dr_primevaltomb_658_3x.dds"},
	{0x77F5E725B1573847ULL, "w_encounter2_36_0_lc.mmh"},
	{0x77F651CA972F8DBDULL, "mag210_arianni6021234.cl"},
	{0x77F6E57094EEFA0DULL, "941306631.ogg"},
	{0x77F740169819DBADULL, "vxht_mansion2_3fm_0_0.vlm"},
	{0x77F740B678926619ULL, "333059877.ogg"},
	{0x77F74818DA484C93ULL, "mag121_idunna6084224.cl"},
	{0x77F7A5E3A59F41A3ULL, "art\\vfx\\combat\\vfx_g_dual_att_front_c.impa.evt"},
	{0x77F7AB458F3B9955ULL, "mrl000_marethari6129240.cl"},
	{0x77F7CA4B73926743ULL, "vxht_0_hightown_bgm_0_0.vlm"},
	{0x77F841A5255A32F0ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1ww_basic.msh"},
	{0x77F84FA5255A4A0AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1ww_basic.mmh"},
	{0x77F853A5255A5141ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1ww_basic.mao"},
	{0x77F8CD2F01CD05C3ULL, "6054006_m.fxe"},
	{0x77F907E4219600ADULL, "vxlt_gallowscourtyard_n_7ha_0_0.vlm"},
	{0x77F9F193671163E6ULL, "vxht_0_hightown_a2_akc_0_0.vlm"},
	{0x77F9F382C7797091ULL, "vxca_cave_ib_0_1.vlm"},
	{0x77FA7A6F4A226184ULL, "vxw_sundermount_26q_0_0.vlm"},
	{0x77FA7D3D41F78B56ULL, "6108966_m.fxe"},
	{0x77FB578A9B8A9B98ULL, "vxdr_0_deeproads_cm2_0_0.vlm"},
	{0x77FB6D0DBF985881ULL, "vxw_bonepit_5b_0_0.vlm"},
	{0x77FBD18DE817E75CULL, "6001043_m.fxe"},
	{0x77FBE879262A899CULL, "6094684_f.fxe"},
	{0x77FBF29EAF3C2066ULL, "mcr310_argument6075075.cl"},
	{0x77FC5B8F5B5823F9ULL, "follower_banter6161237.cl"},
	{0x77FC886B7E3537C8ULL, "vxlt_0_lowtown_docks_bdf_0_0.vlm"},
	{0x77FCD9F04FADDD02ULL, "6076546_m.fxe"},
	{0x77FD3AEFB665B3DEULL, "759317514.ogg"},
	{0x77FD44B089CA747DULL, "llac_lt_0_lowtown_a2_65536_7.dds"},
	{0x77FD95F778A0DF3DULL, "dae300_cullen6135966.cl"},
	{0x77FDA2E117339BB7ULL, "77122422.ogg"},
	{0x77FDB2067A785700ULL, "13540179.ogg"},
	{0x77FE25E53B1C6248ULL, "vxht_0_hightown_n_cb3_0_0.vlm"},
	{0x77FE2874B3475133ULL, "hlac_ca_caverns2_5802_7y.dds"},
	{0x77FE488405B9E4EFULL, "vxht_0_hightown_a3_577_0_0.vlm"},
	{0x77FE7EE3A5A51E24ULL, "art\\vfx\\combat\\vfx_g_dual_att_front_c.impa.gad"},
	{0x77FEA97FD5878EDEULL, "zz_dre_debug6168315.cl"},
	{0x77FF0A15159ACB03ULL, "mer340_bonepit_comments6083143.cl"},
	{0x7800FDC6E94C2865ULL, "vxw_blightlands_vt_0_0.vlm"},
	{0x780176328ABA7A0FULL, "mcr370_meredith6083616.cl"},
	{0x78018646DB1D35D3ULL, "mag320_orsino6071154.cl"},
	{0x7801F797A4987F86ULL, "vxdr_primevalentrance_21n_0_1.vlm"},
	{0x78020200A70BE104ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_up_c.impa.ani"},
	{0x780232146621BB28ULL, "6150176_m.fxe"},
	{0x78025F11D8593716ULL, "vxht_mansion2_24y_0_0.vlm"},
	{0x7802A553EDCC3103ULL, "hlac_w_encounter2_131073_0z.dds"},
	{0x7802E86AF3F58CBAULL, "vxht_0_hightown_ajx_0_0.vlm"},
	{0x7802F0DBBC17EE02ULL, "vxca_cavern1_haunted_e98_0_0.vlm"},
	{0x7803135F9D166B1EULL, "vxw_blightlands_u2_0_0.vlm"},
	{0x78038DD0A61F18C8ULL, "487487411.ogg"},
	{0x7803BFAAC4DCECEDULL, "vxca_caverns2_82u_0_0.vlm"},
	{0x780450810F8AC140ULL, "vxht_chantry_1yo_0_0.vlm"},
	{0x78049DD390E6B487ULL, "vrc221_bartrand6065296.cl"},
	{0x7804E2AD11CC11E9ULL, "vxht_0_hightown_n_apr_0_0.vlm"},
	{0x7804F17C3629C3A7ULL, "vxht_keep_n_6f9_0_0.vlm"},
	{0x7804F2E8F9B94740ULL, "6036420_f.fxe"},
	{0x78058C3EB037065BULL, "vxw_bonepitmining_a2_ev_0_0.vlm"},
	{0x78058DE3A5AB5602ULL, "art\\vfx\\combat\\vfx_g_dual_att_front_c.impa.ani"},
	{0x7805D46DE8EE4B66ULL, "one220_yevhen6099071.cl"},
	{0x78061E3ADECB9399ULL, "114642854.ogg"},
	{0x78065142A7D7160CULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.dura.ani"},
	{0x780651CCEDAF1AEEULL, "6133066_f.fxe"},
	{0x7806DAFC5290B16CULL, "hlac_ht_brothel_n_2995_2x.dds"},
	{0x78072E64B47DB131ULL, "vxw_woundedcoast_1s_0_0.vlm"},
	{0x78078ED46209B0E1ULL, "lt_0_lowtown_34_0.anb"},
	{0x7807BDBD32E08BD4ULL, "23047907.ogg"},
	{0x7807C99AA16519EAULL, "vxlt_undercitydungeon_120_0_0.vlm"},
	{0x7807F7679AB5D281ULL, "11251768.ogg"},
	{0x780812873E5BACD9ULL, "vxw_encounter1_bz_0_0.vlm"},
	{0x7808807CDB858424ULL, "vxlt_0_lowtown_a2_n_h2d_0_0.vlm"},
	{0x7808853712A0BD7DULL, "plt_prm_cod_ivo_amulet.plo"},
	{0x7808AF3D689996A6ULL, "mag311_nyssa6042120.cl"},
	{0x7808BC30FB488F75ULL, "vxlt_0_lowtown_a2_n_c_cow_0_0.vlm"},
	{0x7808FF00A711FB66ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_up_c.impa.gad"},
	{0x78091FE122F0DF8FULL, "vxht_playershouse_n_1st_0_0.vlm"},
	{0x7809267F562EC491ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperleft.mmh"},
	{0x78092BA1500219CCULL, "vxlt_0_lowtown_a3_n_c_gxf_0_0.vlm"},
	{0x78092F03D09C417EULL, "hla_ht_0_hightown_a2_65538_0y.dds"},
	{0x780931E16AF1A925ULL, "w_woundedcoast_67_0.rml"},
	{0x7809327F562ED932ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperleft.mao"},
	{0x7809407F562EF14FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperleft.msh"},
	{0x780955B9C823BCDDULL, "754256134.ogg"},
	{0x780966B110006BBFULL, "496245230.ogg"},
	{0x78098B276BEE28ADULL, "mag210_arianni6043595.cl"},
	{0x7809DFF79AF63A58ULL, "hlac_lt_0_lowtown_a2_-3_7z.dds"},
	{0x780A76D8828D2287ULL, "hlac_lt_0_lowtown_-65536_8z.dds"},
	{0x780B216E0F3E2841ULL, "6102186_f.fxe"},
	{0x780B61F109192F08ULL, "vxlt_gallowstemplar_2bf_0_0.vlm"},
	{0x780C40B6EDF93A8FULL, "175446.ogg"},
	{0x780D1A6082FCEFF0ULL, "vxw_sundermount_2y8_0_0.vlm"},
	{0x780D5E42A7DD4B4EULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.dura.gad"},
	{0x780D9CD685963642ULL, "one100pt_herb_ironwood.ncs"},
	{0x780DD284649D8B44ULL, "230220278.ogg"},
	{0x780DEFFD7A187723ULL, "dre000_sandal6028406.cl"},
	{0x780E84AF566395E8ULL, "mcr310_argument6075107.cl"},
	{0x780EA21F582D0326ULL, "zz_arishockcu.cut"},
	{0x780EAA00E7CE8082ULL, "gen00fl_aveline6120990.cl"},
	{0x780EB9DA2FCCDCB7ULL, "cd_vrc221_bartrand_fake.bnk"},
	{0x780EF621281A0A9FULL, "vxht_0_hightown_a3_bxu_0_0.vlm"},
	{0x780EFA00A7165E45ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_up_c.impa.evt"},
	{0x780F693C4B31B9D9ULL, "rip_flip.dds"},
	{0x78100914DA3A9BBEULL, "vxlt_0_lowtown_a2_n_gwv_0_0.vlm"},
	{0x7810469BF3BDB1BEULL, "770572213.ogg"},
	{0x7810A3F15817D3A8ULL, "6157076_m.fxe"},
	{0x7810EFF9F1F562F9ULL, "vxw_bonepitmining_a2_1px_0_0.vlm"},
	{0x78114D07A4077A05ULL, "6034082_m.fxe"},
	{0x78115B7F4C3E1417ULL, "6209948_f.fxe"},
	{0x781163E3482659E4ULL, "563218421.ogg"},
	{0x781173E4A610FD6FULL, "6198684_f.fxe"},
	{0x7811CC98D06519A2ULL, "135582688.ogg"},
	{0x7811D09259E339A4ULL, "6036499_f.fxe"},
	{0x7812DB67CEB62522ULL, "tbe100_xenon_golem.cnv"},
	{0x7813548A03C22A69ULL, "vxht_0_hightown_n_auq_0_0.vlm"},
	{0x78136359281FDC27ULL, "vxht_keep_n_6a8_0_0.vlm"},
	{0x78136853112EB309ULL, "llac_lt_0_lowtown_n_-65537_0.dds"},
	{0x78139BE0A7BD9822ULL, "638922828.ogg"},
	{0x7813A43AFECA6881ULL, "810646147.ogg"},
	{0x7813A47D61A5767DULL, "hlac_dr_0_deeproads_5849_10z.dds"},
	{0x7813A79C5838280AULL, "vxlt_0_lowtown_docks_cnk_0_0.vlm"},
	{0x78140C4C8B4A06FBULL, "6128510_m.fxe"},
	{0x78145942A7E3612DULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.dura.evt"},
	{0x78150B25B322E21BULL, "237517314.ogg"},
	{0x78150D6EB95DB785ULL, "778621291.ogg"},
	{0x7815BD67CEB8285AULL, "tbe100_xenon_golem.bnk"},
	{0x78164B74E809F2AEULL, "6147224_m.fxe"},
	{0x781684643051C559ULL, "vxw_encounter1_gy_0_0.vlm"},
	{0x7816D56B89BCB54CULL, "215244323.ogg"},
	{0x7816EB78000AC4B1ULL, "570586534.ogg"},
	{0x781791BE14E6F80FULL, "vxht_playershouse_n_1vw_0_0.vlm"},
	{0x7817A2F3502E309AULL, "dae000_sabina6162336.cl"},
	{0x78183B447F76CC98ULL, "vxdr_0_deeproads_cg0_0_0.vlm"},
	{0x78187970DB55BBAFULL, "dae200_cullen6140491.cl"},
	{0x7818AFE2835B9A48ULL, "dae100_alistair_drunk6135142.cl"},
	{0x78190F5D5105873AULL, "252494105.ogg"},
	{0x7819284F60AA3B8AULL, "727650120.ogg"},
	{0x78197E0C83268420ULL, "art\\characters\\playercharacter\\elfmale\\bdy_fls.phy"},
	{0x7819D4E639B87086ULL, "6078706_m.fxe"},
	{0x7819E18FE41E2218ULL, "lt_gallowscourtyard_50_0.anb"},
	{0x7819EDA8B0F6478AULL, "396888585.ogg"},
	{0x781A766FD4EC21E8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic_1.mao"},
	{0x781BCACB775B5AD3ULL, "rdr111dg_assistant6016676.cl"},
	{0x781CA6002E710055ULL, "llac_ca_caverns1_24423_7.dds"},
	{0x781CC895EAF3511FULL, "6021226_m.fxe"},
	{0x781CF1F7A11E6D7AULL, "vxht_brothel_n_5jx_0_0.vlm"},
	{0x781D5C8736FBDFC3ULL, "vxlt_gallowstemplar_5br_0_0.vlm"},
	{0x781D67FE1A10231FULL, "vxht_0_hightown_a3_bwt_0_0.vlm"},
	{0x781D813B9A296F28ULL, "lt_0_lowtown_docks_a3_n_18_0.anb"},
	{0x781D853E4641CE5FULL, "6165478_m.fxe"},
	{0x781E237430C0B961ULL, "vxca_cavern1_haunted_i1s_0_0.vlm"},
	{0x781E7AF1CC30B43EULL, "vxlt_0_lowtown_a2_n_glw_0_0.vlm"},
	{0x781F5792EBB76342ULL, "6028181_m.fxe"},
	{0x781F8410DB2E6502ULL, "6068784_f.fxe"},
	{0x781FA183E0883F8CULL, "vxht_0_hightown_n_bk1_0_0.vlm"},
	{0x7820663EEBCC64E9ULL, "6127030_m.fxe"},
	{0x7820F99C135BEC31ULL, "hlac_lt_0_lowtown_a2_n_c_131074_0y.dds"},
	{0x78216E5BE3586BA6ULL, "306372982.ogg"},
	{0x7821D844602A3D4DULL, "vxw_bonepitmining_a2_7f_0_0.vlm"},
	{0x7822570C5B4A2545ULL, "137822169.ogg"},
	{0x78226FF89423375BULL, "vxw_bonepitmining_a2_oh_0_0.vlm"},
	{0x78228EF3903851C1ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_darren.ani"},
	{0x7822F3FAE8365566ULL, "lt_0_lowtown_docks_n_8_0.anb"},
	{0x7823B4A324430B76ULL, "vxlt_0_lowtown_a3_n_c_ews_0_0.vlm"},
	{0x78242D57CB0EB453ULL, "148014656.ogg"},
	{0x78246735BEA5230BULL, "6174076_m.fxe"},
	{0x7824BD51DA000B2EULL, "6140324_m.fxe"},
	{0x782562D657C0D676ULL, "w_sundermount_208_0.mao"},
	{0x78256ED657C0EA15ULL, "w_sundermount_208_0.mmh"},
	{0x782570D657C0ED8BULL, "w_sundermount_208_0.msh"},
	{0x78258B526B0FFFA1ULL, "96811347.ogg"},
	{0x7825C9C55E5F9388ULL, "vxht_0_hightown_a2_bnu_0_0.vlm"},
	{0x7825F9067EC0AF36ULL, "w_encounter3_23_0.msh"},
	{0x782603067EC0C034ULL, "w_encounter3_23_0.mmh"},
	{0x782607067EC0C747ULL, "w_encounter3_23_0.mao"},
	{0x78262B66DA209C8BULL, "cn_bdy_hlka_0t.dds"},
	{0x78265D0D1F8FA118ULL, "619507028.ogg"},
	{0x7826AF18418300E4ULL, "6077154_m.fxe"},
	{0x78277E8321C6AA2FULL, "vxlt_0_lowtown_a2_n_c_hlc_0_0.vlm"},
	{0x782796A80C0FAEBCULL, "6164193_m.fxe"},
	{0x78287E9DCCF7D728ULL, "and121_karl6093987.cl"},
	{0x7828F2F18ECDB58EULL, "vxlt_0_lowtown_docks_c05_0_0.vlm"},
	{0x782954DB65C7D5A8ULL, "6080286_m.fxe"},
	{0x7829639589A565A0ULL, "81455474.ogg"},
	{0x78297BAFD04AC82DULL, "hm_carver_h3s2.mor"},
	{0x782999F3903E8363ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_darren.gad"},
	{0x7829D1F9568A996DULL, "6190995_m.fxe"},
	{0x782A34474DCD86A3ULL, "624277249.ogg"},
	{0x782A38B7167EA6ECULL, "6042305_m.fxe"},
	{0x782A5F584243F475ULL, "gen00fl_anders6125164.cl"},
	{0x782A6089CD4DE570ULL, "zz_vrc_debug6200202.cl"},
	{0x782AF3BDBE41E339ULL, "967327960.ogg"},
	{0x782B2C07B7D8F0E1ULL, "vxw_bonepitdestroyed_a3_1bt_0_0.vlm"},
	{0x782B33343909C2EAULL, "zz_usespecificobject.ncs"},
	{0x782B6D590DEC1E29ULL, "6108510_m.fxe"},
	{0x782B78798275484EULL, "zz_ave_debug6063268.cl"},
	{0x782B825D3396AA99ULL, "vxca_caverns2_74h_0_0.vlm"},
	{0x782BB420AD85B9B4ULL, "281000124.ogg"},
	{0x782BCE6428F1F843ULL, "vxlt_gallowstemplar_5ou_0_0.vlm"},
	{0x782BD9DB0C063B9FULL, "vxht_0_hightown_a3_brw_0_0.vlm"},
	{0x782C3478908AE6B7ULL, "436908375.ogg"},
	{0x782C73DB9217525CULL, "28078371.ogg"},
	{0x782CFB0F2295B389ULL, "6119313_f.fxe"},
	{0x782D5A99848AE769ULL, "vxdr_primevalentrance_2nq_0_0.vlm"},
	{0x782D5B91EA5FD21DULL, "w_bonepitmining_a2_246_0.mao"},
	{0x782D5F91EA5FD956ULL, "w_bonepitmining_a2_246_0.mmh"},
	{0x782D6991EA5FEA54ULL, "w_bonepitmining_a2_246_0.msh"},
	{0x782DC930CC8A6D37ULL, "dre000_cut_thewayout6030029.cl"},
	{0x782E3545CF475C2FULL, "vxca_caverns2prc_abw_0_0.vlm"},
	{0x782E8654AD262E22ULL, "6171948_m.fxe"},
	{0x782E8783776FFA16ULL, "6026200_m.fxe"},
	{0x782ED52BEFD6B60AULL, "6080990_m.fxe"},
	{0x782EE6F46F0A1FFBULL, "371860257.ogg"},
	{0x782F11E16B12424EULL, "w_woundedcoast_67_0.mmh"},
	{0x782F1DE16B1256E5ULL, "w_woundedcoast_67_0.mao"},
	{0x782F2B29A9A14E46ULL, "vxht_0_hightown_a3_c96_0_0.vlm"},
	{0x782F2BE16B126E3CULL, "w_woundedcoast_67_0.msh"},
	{0x782F6CA672E02318ULL, "964935765.ogg"},
	{0x782F9D815BAFF524ULL, "vxca_caverns1_mines_dpl_0_0.vlm"},
	{0x7830169EFCBA12C1ULL, "vxht_keep_73j_0_0.vlm"},
	{0x78308B563C24590AULL, "vxlt_0_lowtown_docks_cte_0_0.vlm"},
	{0x7830985D2D3D5600ULL, "hlac_lt_0_lowtown_n_-65537_3x.dds"},
	{0x78309E7415513788ULL, "follower_banter6173079.cl"},
	{0x7830E1D586194FDBULL, "vxw_bonepitmining_a2_jk_0_0.vlm"},
	{0x7830E811D2E672FAULL, "223835416.ogg"},
	{0x783133C299FF1CD0ULL, "vxht_mansion2b_2w2_0_0.vlm"},
	{0x7831AFEB07092FA3ULL, "vxca_caverns1_mines_e6t_0_0.vlm"},
	{0x7832E8637DF2B8D7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bnt_dblw.gad"},
	{0x7832FE932A4DB2EEULL, "vxht_brothel_n_4fe_0_0.vlm"},
	{0x78334DA2F9928104ULL, "character_character_p241.psh.fxo.dx11.assembly"},
	{0x78336AC7441A6DEDULL, "lgt320_mis_resp_swordhunter6150345.cl"},
	{0x7833A41C8088A43DULL, "vxlt_0_lowtown_n_byz_0_0.vlm"},
	{0x7834B4BD792D9CC8ULL, "vxw_bonepitdestroyed_a3_1bq_0_0.vlm"},
	{0x7835416569A81F61ULL, "755225591.ogg"},
	{0x783561EEADD5A4D5ULL, "llac_w_bonepitdestroyed_a3_196612_0.dds"},
	{0x78358ADE0B67A539ULL, "6070482_m.fxe"},
	{0x7835B395B97F37DFULL, "llac_w_sundermount_0_6.dds"},
	{0x7835E716FC65E9C8ULL, "170101282.ogg"},
	{0x783726D148946C39ULL, "ht_0_hightownexplosion_n_26_0_lc.mmh"},
	{0x7837AE483C3EE33DULL, "vxlt_0_lowtown_a2_n_c_e1k_0_0.vlm"},
	{0x78382D798AE108A4ULL, "gen_im_trs_tmp_med_boo_r1.uti"},
	{0x7838304EDC9A1C63ULL, "6095834_m.fxe"},
	{0x783885AD05329E80ULL, "1066170816.ogg"},
	{0x7838AB3E2F50772DULL, "sse_ss_caress.gda"},
	{0x7839483EDF72CDECULL, "plt_cod_hst_primeval_thaig.plo"},
	{0x783977BEB39CE445ULL, "vxht_brothel_n_5gt_0_0.vlm"},
	{0x7839C5CA5A76E66EULL, "197188093.ogg"},
	{0x7839FD637DF8FB64ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bnt_dblw.evt"},
	{0x783A109799CF8F14ULL, "514268950.ogg"},
	{0x783A13E0CE708152ULL, "dr_primevaltomb_666.rml"},
	{0x783A3689EE217C74ULL, "6161019_m.fxe"},
	{0x783A5F24C47AF9D7ULL, "153908434.ogg"},
	{0x783A72B4A8C1B49AULL, "57157883.ogg"},
	{0x783AAB943042F803ULL, "lt_0_lowtown_docks_n_149_0.msh"},
	{0x783AB9943043105DULL, "lt_0_lowtown_docks_n_149_0.mmh"},
	{0x783ABD943043176EULL, "lt_0_lowtown_docks_n_149_0.mao"},
	{0x783AEB6FF3E0D086ULL, "6161336_m.fxe"},
	{0x783B5BE151072B68ULL, "104973178.ogg"},
	{0x783C16A4AA511747ULL, "hlac_ht_mansion2b_818_42y.dds"},
	{0x783C30E16B1D014DULL, "w_woundedcoast_67_0.anb"},
	{0x783C3F189177BB51ULL, "180360476.ogg"},
	{0x783C41CFE2B835A7ULL, "hlac_lt_undercityencounter_-1_25x.dds"},
	{0x783D16F06D8E9C82ULL, "6128713_m.fxe"},
	{0x783D7DC8E059826DULL, "vxht_0_hightown_b9m_0_0.vlm"},
	{0x783DB03FFF8330A4ULL, "vxht_mansion2_218_0_0.vlm"},
	{0x783DD225292836ECULL, "one280_iwan6120911.cl"},
	{0x783DDD9DC2AF9494ULL, "pro000_refugee_male6052114.cl"},
	{0x783E9B0C8345D1B5ULL, "art\\characters\\playercharacter\\elfmale\\bdy_fls.msh"},
	{0x783E9D0C8345D52BULL, "art\\characters\\playercharacter\\elfmale\\bdy_fls.mmh"},
	{0x783EA90C8345E9C8ULL, "art\\characters\\playercharacter\\elfmale\\bdy_fls.mao"},
	{0x783EF8EDEAED4666ULL, "hf_zmother_h1s2.mor"},
	{0x783F19AA959DA970ULL, "mcr310_argument6072674.cl"},
	{0x783F42A443B51304ULL, "hm_crazedloner.mor"},
	{0x783FA59F8BF53550ULL, "vxht_mansion2b_2p3_0_0.vlm"},
	{0x783FD18957F6A04EULL, "6121582_f.fxe"},
	{0x784077BA9F0FCFFCULL, "vxlt_gallowstemplar_3gd_0_0.vlm"},
	{0x7840C71D2A2C03D3ULL, "fire_u_06.dds"},
	{0x78412274796FBF3FULL, "251621725.ogg"},
	{0x78414D048A1DEC55ULL, "6142057_m.fxe"},
	{0x78416FD4623AF471ULL, "lt_0_lowtown_34_0.rml"},
	{0x7841DA66756DA14CULL, "gen_im_wep_war_shd_lore_02.uti"},
	{0x7841F6B642B24DB5ULL, "42737183.ogg"},
	{0x784223F732CC8BACULL, "6043672_m.fxe"},
	{0x78423DA3C40274E6ULL, "413953690.ogg"},
	{0x7842998392C65746ULL, "t1_mub_ta1.tnt"},
	{0x7842B7336B673F8AULL, "318736275.ogg"},
	{0x78431204D870757EULL, "art\\vfx\\cinematic\\vfx_anders_romance_hand_c.impa.ani"},
	{0x7843269A6B23B548ULL, "vxw_bonepitdestroyed_a3_1gn_0_0.vlm"},
	{0x78436AAA77D54924ULL, "hlac_lt_0_lowtown_docks_n_-131075_3z.dds"},
	{0x784377C1EDC02836ULL, "hlac_lt_0_lowtown_n_65535_3y.dds"},
	{0x784396DAFF229B91ULL, "vxlt_0_lowtown_a3_n_c_i16_0_0.vlm"},
	{0x7844143A9ADB7928ULL, "vxca_cavern1_haunted_jg7_0_0.vlm"},
	{0x7844B3C89A0AAE8CULL, "vxca_caverns1_i3e_0_0.vlm"},
	{0x7844CAC87D2A0B3DULL, "vxlt_0_lowtown_h80_0_0.vlm"},
	{0x7844CE9E4ABE8BE8ULL, "isa311_isabela6175476.cl"},
	{0x7844DC654661AD19ULL, "art\\levels\\races\\proxy\\props\\_textures\\corpses_sa.dds"},
	{0x7845EDAF18FDEE61ULL, "vxdr_primevaltomb_8nw_0_0.vlm"},
	{0x7846961F3B6AFEE2ULL, "vxht_keep_3c9_0_0.vlm"},
	{0x7846999E5EEB23E3ULL, "vxht_0_hightown_9ut_0_0.vlm"},
	{0x7847731E36B8BA8BULL, "zz_qun_debug6028255.cl"},
	{0x7847773B17C4E5E9ULL, "mcr370_meredith6083607.cl"},
	{0x784779B25F1ABF56ULL, "mrl310_merrill6144192.cl"},
	{0x784780AF42213BACULL, "mag210_arianni6056936.cl"},
	{0x7847AB2B4FB9C594ULL, "llac_lt_0_lowtown_n_131070_2.dds"},
	{0x7847DBF710C3E811ULL, "vxw_sundermount_30k_0_0.vlm"},
	{0x7847E99BA592FCC5ULL, "vxht_brothel_n_5js_0_0.vlm"},
	{0x7847EE78B5E3A57CULL, "6156318_m.fxe"},
	{0x7847F4C26A565CB7ULL, "vxdr_0_deeproads_ebn_0_0.vlm"},
	{0x78480FC19BC521E1ULL, "vxw_bonepitdestroyed_a3_1hj_0_0.vlm"},
	{0x784815637E05615DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bnt_dblw.ani"},
	{0x784815E7088AA16DULL, "654775610.ogg"},
	{0x7848B21E0CDE2943ULL, "vxlt_gallowstemplar_5uw_0_0.vlm"},
	{0x7848D0F9AA31CF79ULL, "483509455.ogg"},
	{0x7849528FF14FC6CEULL, "gen00fl_fenris6106630.cl"},
	{0x78496A6CBF3D4CE9ULL, "hlac_lt_gallowstemplar_n_2382_1x.dds"},
	{0x78497B4201460629ULL, "6169784_m.fxe"},
	{0x7849D088A212FDBEULL, "vxlt_0_lowtown_a2_n_gcj_0_0.vlm"},
	{0x784A92D5AB8DAA1BULL, "241412010.ogg"},
	{0x784B0D6CE36ECECBULL, "vxca_caverns1_mines_cjv_0_0.vlm"},
	{0x784B16AC517592E4ULL, "zz_dae_debug6045442.cl"},
	{0x784B1E55C617FDD8ULL, "6120801_m.fxe"},
	{0x784BAF09B05B94E0ULL, "plt_mrl350pt_newpath.plo"},
	{0x784BF11BF438D9CBULL, "dae000_merchant_access_hf6162966.cl"},
	{0x784C0EE38D8D7F46ULL, "vxht_0_hightown_a3_c78_0_0.vlm"},
	{0x784C43371D2CAB0CULL, "qun160_vashothbandits6069281.cl"},
	{0x784C7659286D0A34ULL, "6169671_m.fxe"},
	{0x784CA9C8CE32EABBULL, "gen00fl_merrill6130139.cl"},
	{0x784CEE3559FD6D35ULL, "243514657.ogg"},
	{0x784D10FE5B26CA23ULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.cess.ani"},
	{0x784D2A5D0F8370BFULL, "llac_lt_0_lowtown_n_1_2.dds"},
	{0x784D3D9B04E00ADFULL, "llac_lt_0_lowtown_1_10.dds"},
	{0x784DD4765D161893ULL, "526589514.ogg"},
	{0x784DFEEA4C2446BBULL, "665143444.ogg"},
	{0x784E1573D1397957ULL, "vxw_woundedcoast_1fa_0_0.vlm"},
	{0x784E219FB2614588ULL, "vxlt_0_lowtown_a2_her_0_0.vlm"},
	{0x784E3D707890FDD4ULL, "6063822_f.fxe"},
	{0x784E75791F56FA43ULL, "vxht_0_hightown_bys_0_0.vlm"},
	{0x784E8B1A641C28C8ULL, "hlac_lt_backalley_g_0_15x.dds"},
	{0x784F6513F934F317ULL, "vxca_caverns2_95w_0_0.vlm"},
	{0x784F6EA677573AB4ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mash.phy"},
	{0x784F7664DA529A48ULL, "vxdr_primevaltomb_8nr_0_0.vlm"},
	{0x784FD62BE3FCC978ULL, "zz_dae_debug6131006.cl"},
	{0x784FD833EE812AC9ULL, "6097856_m.fxe"},
	{0x784FEC3B44D5AC1CULL, "6108314_f.fxe"},
	{0x784FF12191189DEDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_basic.msh"},
	{0x784FFF219118B520ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_basic.mao"},
	{0x785003219118BC73ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_basic.mmh"},
	{0x785081D9719281A2ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_3x.dds"},
	{0x7850B6B2C7B66B7DULL, "vxlt_0_lowtown_a2_n_f4q_0_0.vlm"},
	{0x7850E45169440D73ULL, "zz_isa.ncs"},
	{0x7850F8A034FB96ECULL, "follower_banter6173383.cl"},
	{0x785119176186CD25ULL, "280382646.ogg"},
	{0x7851232F2C7CCEA4ULL, "vxlt_shop_ew_0_0.vlm"},
	{0x78512A04D87CDB4FULL, "art\\vfx\\cinematic\\vfx_anders_romance_hand_c.impa.evt"},
	{0x7851357E9EB50314ULL, "182245158.ogg"},
	{0x785135A407D5C7DAULL, "384414578.ogg"},
	{0x7851847A5E8FFDC7ULL, "dae000_oldmangossip6174910.cl"},
	{0x785198775D19CDC8ULL, "vxw_bonepitdestroyed_a3_1ho_0_0.vlm"},
	{0x78525F7985B46FF1ULL, "519189161.ogg"},
	{0x785287A365BACCEAULL, "gen00fl_aveline6139729.cl"},
	{0x78528C8DAC0F307FULL, "llac_lt_0_lowtown_docks_-65540_2.dds"},
	{0x7852D1A46FBC32B9ULL, "596473803.ogg"},
	{0x785349B91FFEB2D5ULL, "393452333.ogg"},
	{0x78535DCB12BD7546ULL, "clf360_evets.cnv"},
	{0x785387A7FC3E39F0ULL, "llac_lt_shop_-1_0.dds"},
	{0x78545AC294BB5BF9ULL, "589975217.ogg"},
	{0x78546EAD5AAB9567ULL, "894019961.ogg"},
	{0x7854B6E2F91FAF96ULL, "vxlt_gallowscourtyard_781_0_0.vlm"},
	{0x785522009EB6D82AULL, "67255261.ogg"},
	{0x78555EAD055B04C1ULL, "lt_0_lowtown_a3_n_c_12_0.phy"},
	{0x7855D1AFBB4FDC1EULL, "hand.cur"},
	{0x7855D4CEDBF42CF5ULL, "hlac_ht_0_hightown_65537_0z.dds"},
	{0x785645AB2A31CD26ULL, "6133803_m.fxe"},
	{0x785664BF3E8EA74BULL, "hlac_w_encounter2_131072_0y.dds"},
	{0x78568E5458912AB2ULL, "art\\cinematics\\props\\letterfront.mao"},
	{0x7856C5E141791844ULL, "vxht_0_hightown_a2_c11_0_0.vlm"},
	{0x7856CDE413A76C3BULL, "mag321_comtesse_de_launcet6064889.cl"},
	{0x7856D4D9A0F2D3F3ULL, "vxlt_undercityencounter_1hb_0_0.vlm"},
	{0x7856E764A65B9872ULL, "vxht_bartrand_n_4a8_0_0.vlm"},
	{0x7856FF9B1D165777ULL, "cyc_cloud.dds"},
	{0x78570894F59C27C3ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\prp_zzhumanhead_d.dds"},
	{0x78575FCB12C16106ULL, "clf360_evets.bnk"},
	{0x785776A206623728ULL, "hlac_lt_0_lowtown_131071_0x.dds"},
	{0x7857EB355D207C54ULL, "282879963.ogg"},
	{0x78581CC05A9B3761ULL, "vxlt_0_lowtown_docks_b9d_0_0.vlm"},
	{0x78582B04D882FBF8ULL, "art\\vfx\\cinematic\\vfx_anders_romance_hand_c.impa.gad"},
	{0x78583521440E3B8AULL, "6075451_m.fxe"},
	{0x78587F17C7EEA2F1ULL, "6041324_f.fxe"},
	{0x7858EC08E579429AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_pailempty.phy"},
	{0x78594BA6775F53B1ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mash.msh"},
	{0x785965A6775F806FULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mash.mmh"},
	{0x785AEE633C74F200ULL, "mag170_thrask6167407.cl"},
	{0x785B28FE5B33301AULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.cess.evt"},
	{0x785B3B8FA5B24FB3ULL, "vxlt_0_lowtown_a2_n_c_equ_0_0.vlm"},
	{0x785B67AD055F7FDDULL, "lt_0_lowtown_a3_n_c_12_0.rml"},
	{0x785C59AE6C9212FEULL, "6125400_m.fxe"},
	{0x785C7D3132CE1906ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.dura.ani"},
	{0x785C89596FE16650ULL, "vxht_mansion2b_2f1_0_0.vlm"},
	{0x785C937CA4575E08ULL, "vxlt_0_lowtown_a2_hbs_0_0.vlm"},
	{0x785CDF6C8004C928ULL, "llac_ht_0_hightown_a3_131074_0.dds"},
	{0x785CFDCAEF8FCDD1ULL, "dae000_gamlen6121503.cl"},
	{0x785D2899B138CB5BULL, "817611753.ogg"},
	{0x785DD82F9D51F777ULL, "mag370_meredith6092202.cl"},
	{0x785DE841CC48B2C8ULL, "vxdr_primevaltomb_8is_0_0.vlm"},
	{0x785E44603E8280ADULL, "vxw_encounter3_dn_0_0.vlm"},
	{0x785E72CA18A0A0A3ULL, "vxlt_hangedman_27o_0_0.vlm"},
	{0x785E803B9A6104E0ULL, "lt_0_lowtown_docks_a3_n_18_0.rml"},
	{0x785E94CA5C0DC66DULL, "6091708_m.fxe"},
	{0x785EDCED9D6D7A86ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur4_flemeth_real.ani"},
	{0x785F6443CF6740F7ULL, "vxdr_primevaltomb_6m6_0_0.vlm"},
	{0x785F9C0020DF4DA2ULL, "vxht_keep_6gx_0_0.vlm"},
	{0x786026D2EF892B3FULL, "6220209_m.fxe"},
	{0x786043B0EBC0D76EULL, "6148647_m.fxe"},
	{0x78605A58322F041EULL, "6067834_m.fxe"},
	{0x78607D897BC9549DULL, "llac_w_sundermount_131067_0.dds"},
	{0x78609F62C393A77CULL, "6087368_m.fxe"},
	{0x7860D77DB41AC1B8ULL, "hlac_lt_darktown_-1_35x.dds"},
	{0x78613AE5AB7F962BULL, "hlac_lt_0_lowtown_n_-65536_3x.dds"},
	{0x786145F6E99F0C2FULL, "vxlt_0_lowtown_a2_n_c_hxo_0_0.vlm"},
	{0x7861628AB2F00441ULL, "184190734.ogg"},
	{0x786182186F16AED9ULL, "vxw_bonepitdestroyed_a3_28i_0_0.vlm"},
	{0x7861A5761BEFE348ULL, "vxlt_0_lowtown_docks_b9a_0_0.vlm"},
	{0x7862098724B7701BULL, "6127220_f.fxe"},
	{0x786227FE5B394D0DULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.cess.gad"},
	{0x78624127AC6C3BBEULL, "13248846.ogg"},
	{0x7863C7222A691BC8ULL, "mag100_karras6035373.cl"},
	{0x7864F3D535468F2CULL, "llac_lt_0_lowtown_a2_65534_4.dds"},
	{0x786559419851B0F2ULL, "vxht_bartrand_n_4f9_0_0.vlm"},
	{0x786560BA420EF8ADULL, "vxht_0_hightown_n_bty_0_0.vlm"},
	{0x78660020E44DA4E4ULL, "hlac_lt_0_lowtown_docks_n_-131074_0x.dds"},
	{0x78662CD22B0106FCULL, "hlac_ht_mansion2b_818_29y.dds"},
	{0x78663E71C877032DULL, "vxlt_gallowstemplar_n_46z_0_0.vlm"},
	{0x78674621B2B9C403ULL, "vxca_cave_19l_0_0.vlm"},
	{0x7867E82FE57A525CULL, "art\\characters\\creatures\\ragedemon\\animation\\wk_p_rf.gad"},
	{0x7868A31BBF0F2D0FULL, "6116177_m.fxe"},
	{0x7868D61C55DDC7E3ULL, "336162220.ogg"},
	{0x786939020FC70B3FULL, "1007296457.ogg"},
	{0x78693ADF1ADF0DA9ULL, "vxht_0_hightown_a2_asx_0_0.vlm"},
	{0x78693B4C8730C41CULL, "llac_dr_dungeon_3207_0.dds"},
	{0x78693FA6BE866FF6ULL, "390118347.ogg"},
	{0x78698FEB0573C100ULL, "334208804.ogg"},
	{0x7869B1D5041F13C7ULL, "vxlt_gallowstemplar_2r5_0_0.vlm"},
	{0x786A4278CD6BAC98ULL, "6044655_m.fxe"},
	{0x786A47464FA30274ULL, "vxlt_0_lowtown_a2_n_fh8_0_0.vlm"},
	{0x786A653132DA2E77ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.dura.evt"},
	{0x786A859056476628ULL, "dre000_carver6037601.cl"},
	{0x786ACB8B5E882B7EULL, "6122500_m.fxe"},
	{0x786AF92DB525AA57ULL, "vxw_woundedcoast_1lc_0_0.vlm"},
	{0x786B007A16B1CFACULL, "hlac_lt_0_lowtown_a2_n_c_65536_11y.dds"},
	{0x786BBE789CB190C5ULL, "zz_mer_debug6180869.cl"},
	{0x786BCCA7ED3D81E1ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_007a_0n.dds"},
	{0x786BDEC119AC0EE2ULL, "6033501_m.fxe"},
	{0x786C0B0FBC21EC74ULL, "546986248.ogg"},
	{0x786C0D96EC67B3CFULL, "hlac_lt_0_lowtown_a2_n_-196609_4y.dds"},
	{0x786C126B9AFB7BDCULL, "cntr_wispy.mao"},
	{0x786C4C1A8899E9DDULL, "6132821_m.fxe"},
	{0x786CC606F22614EEULL, "vxht_brothel_n_4zy_0_0.vlm"},
	{0x786D27E567066FEFULL, "70577104.ogg"},
	{0x786D49051E62A5CAULL, "vxht_chantry_n_2zl_0_0.vlm"},
	{0x786DE72FE57EBC5AULL, "art\\characters\\creatures\\ragedemon\\animation\\wk_p_rf.ani"},
	{0x786E6185410A7386ULL, "399750663.ogg"},
	{0x786E67CF044D2AE8ULL, "zz_followers_banter6158373.cl"},
	{0x786E82683DF76635ULL, "6175077_m.fxe"},
	{0x786ED904891978B1ULL, "420924811.ogg"},
	{0x786F80ADAD1EDD13ULL, "184614867.ogg"},
	{0x786F92D44FB99D6AULL, "dae300_vashothlead6062797.cl"},
	{0x786FF3F5610CC759ULL, "vxw_bonepitdestroyed_a3_23h_0_0.vlm"},
	{0x787033C3A78351CBULL, "vxca_caverns2_98h_0_0.vlm"},
	{0x7870663132DE9B00ULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.dura.gad"},
	{0x7870FEE96174C237ULL, "6065287_f.fxe"},
	{0x7870FFEC92A83519ULL, "vxht_0_hightown_9b7_0_0.vlm"},
	{0x787177D0D2529BD4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared3a.phy"},
	{0x7871CDE9DE0B2A2CULL, "vxlt_0_lowtown_n_ekj_0_0.vlm"},
	{0x7872857C00EC1603ULL, "107649530.ogg"},
	{0x7872C5ED9D7DFC80ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur4_flemeth_real.gad"},
	{0x7873053EC51912A7ULL, "6127682_f.fxe"},
	{0x7873CBA0E8847738ULL, "139009932.ogg"},
	{0x787410AF2D8FDA63ULL, "6085426_m.fxe"},
	{0x787454DB8DA84D96ULL, "wm_ico_kirkwall.dds"},
	{0x787524483E817C7FULL, "6208334_f.fxe"},
	{0x787542A043E600EDULL, "vxdr_primevaltomb_544_0_0.vlm"},
	{0x7876620385E693DEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_up_to_down.evt"},
	{0x787689F16289C1AEULL, "163733917.ogg"},
	{0x7876DE848C2C41D7ULL, "load_light.gfx"},
	{0x7876FC5DA2E041A3ULL, "vxca_caverns1_mines_fn3_0_0.vlm"},
	{0x7877020926E81439ULL, "gen00fl_fenris6112098.cl"},
	{0x78772067341623C4ULL, "6115733_m.fxe"},
	{0x7877A23B2A4565C1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsarch.phy"},
	{0x7877AC899A8F79F5ULL, "vxlt_backalley_g_1vn_0_0.vlm"},
	{0x7878211129AC78C1ULL, "llac_ht_0_hightown_a3_65533_1.dds"},
	{0x787830B0EDF0C15FULL, "6140923_m.fxe"},
	{0x78789688DADB7568ULL, "6149044_m.fxe"},
	{0x7879C73C9E5A58EDULL, "dae000_arishok6082018.cl"},
	{0x787A3F2E66E831FCULL, "vxlt_gallowstemplar_3sx_0_0.vlm"},
	{0x787AEA3489C6149EULL, "6153290_m.fxe"},
	{0x787AF1CBD8B4BFB4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chantryframe.phy"},
	{0x787B03067F090037ULL, "w_encounter3_23_0.rml"},
	{0x787B2F134AC7A6E2ULL, "177748822.ogg"},
	{0x787BA951D1529B62ULL, "follower_banter6155976.cl"},
	{0x787BFF69EDD295B9ULL, "hlac_lt_gallowstemplar_2391_3y.dds"},
	{0x787C47FDB35371F7ULL, "vxdr_primevaltomb_6w0_0_0.vlm"},
	{0x787C54D0D25C67D1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared3a.msh"},
	{0x787C5E5CCAEC2485ULL, "329838792.ogg"},
	{0x787C62D0D25C7FECULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared3a.mao"},
	{0x787C6ED0D25C938FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared3a.mmh"},
	{0x787C7FBA04CB7EA2ULL, "vxht_keep_6mz_0_0.vlm"},
	{0x787D30D7198DC2F6ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_lbrt.ani"},
	{0x787D790385ECD909ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_up_to_down.gad"},
	{0x787DA170BACD101EULL, "6166593_m.fxe"},
	{0x787DC1D332821EE2ULL, "6148755_m.fxe"},
	{0x787E7562C1B35ADCULL, "hlac_ht_0_hightown_a3_-131074_0y.dds"},
	{0x787F387E22ED2BF8ULL, "vxht_bartrand_n_5vn_0_0.vlm"},
	{0x787F780DFC450C42ULL, "dae110ip_at_gamlenhouse.utp"},
	{0x787F975CC6F73FF6ULL, "35537786.ogg"},
	{0x787FACA74B3AF800ULL, "hlac_ht_0_hightown_a3_-65536_2z.dds"},
	{0x787FF48C0F3526D4ULL, "6172969_m.fxe"},
	{0x788019AC312FB431ULL, "hlac_lt_0_lowtown_docks_a3_65535_4z.dds"},
	{0x78805AD0450B78CDULL, "lgt150_nat_were_supporter6110980.cl"},
	{0x7881429315B8C882ULL, "vxlt_warehouse_15o_0_0.vlm"},
	{0x78816A4672ED0B21ULL, "616003177.ogg"},
	{0x788196342E5390A4ULL, "vxlt_gallowscourtyard_78s_0_0.vlm"},
	{0x7881E6CEC4049BF6ULL, "895497794.ogg"},
	{0x7881E99000D0DDB8ULL, "mcr310_scared_mage6072858.cl"},
	{0x788226C76F4DAAE4ULL, "hlac_w_woundedcoast_196611_0y.dds"},
	{0x78822E0838F1C956ULL, "ca_cavern1_haunted_24426.rml"},
	{0x7882447425FB29ADULL, "vxht_0_hightown_n_bzw_0_0.vlm"},
	{0x7882597B9B9784B2ULL, "vxht_0_hightown_c3u_0_0.vlm"},
	{0x78831AE0AF656ED9ULL, "6076369_m.fxe"},
	{0x788351AB0ED9EC21ULL, "hlac_dr_0_deeproads_5841_4y.dds"},
	{0x78837B43794D053DULL, "vxht_chantry_n_2s4_0_0.vlm"},
	{0x788380F8724DECB1ULL, "hlac_lt_0_lowtown_1_4y.dds"},
	{0x7883D37CF7D1B165ULL, "933820619.ogg"},
	{0x7884472C08C8801DULL, "6170012_m.fxe"},
	{0x78844A0385F2A857ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_up_to_down.ani"},
	{0x788462CB935BB2F6ULL, "769239707.ogg"},
	{0x7884931815FCF29FULL, "500741219.ogg"},
	{0x7884CECBD8BCD8B1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chantryframe.msh"},
	{0x7884DCCBD8BCF0CCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chantryframe.mao"},
	{0x7884E8CBD8BD056FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww12_chantryframe.mmh"},
	{0x78857F6C84078A60ULL, "vxlt_0_lowtown_docks_n_cw1_0_0.vlm"},
	{0x788597195F36039FULL, "zz_approval_debug6003660.cl"},
	{0x7885A6606847639FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appp_0d.dds"},
	{0x7885C5B2DAAC5A9EULL, "tow_spine.mao"},
	{0x78860595B76ED78AULL, "clf260_harley6069578.cl"},
	{0x7886226D452ACA6EULL, "6217493_m.fxe"},
	{0x788651409959EB5AULL, "hf_meredith.mor"},
	{0x78867670345E9274ULL, "lgt220_shp_mag_guard6149922.cl"},
	{0x788687CAD5330832ULL, "6202345_m.fxe"},
	{0x7886AE9C5A826073ULL, "hlac_lt_0_lowtown_a2_n_c_-1_2y.dds"},
	{0x78870D0BA9C9981AULL, "vxca_caverns2_7ju_0_0.vlm"},
	{0x7888313E45931B6FULL, "6129136_f.fxe"},
	{0x788849CDE63E6759ULL, "20040153.ogg"},
	{0x78884B91A90C3134ULL, "gen00fl_carver6097406.cl"},
	{0x7888CA6AA48A658BULL, "272870108.ogg"},
	{0x78893DD8A22AFC48ULL, "vxdr_primevaltomb_8zn_0_0.vlm"},
	{0x788987C5DEF9EE34ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_3y.dds"},
	{0x788988EA1E6AF5BEULL, "vxlt_0_lowtown_docks_n_c4v_0_0.vlm"},
	{0x788A2CBF024ED6CAULL, "vxht_chantry_n_2pn_0_0.vlm"},
	{0x788AB9DAA5498A77ULL, "vxdr_primevaltomb_6z3_0_0.vlm"},
	{0x788B58343B72060AULL, "378623669.ogg"},
	{0x788BCBC285B88C1BULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_p.impa.evt"},
	{0x788C0C5D2C613824ULL, "hlac_ht_keep_3735_1z.dds"},
	{0x788CF5B852B2C943ULL, "hlac_lt_0_lowtown_a2_n_-2_0x.dds"},
	{0x788CFA6BE0F72A71ULL, "639588071.ogg"},
	{0x788D105F4ADAD30EULL, "qcr220_grey_wardens6060472.cl"},
	{0x788D193CAA8BAA16ULL, "dae000cr_grand_cleric6147720.cl"},
	{0x788D3F527FDC9C3BULL, "pssmmesh.vsh.fxo"},
	{0x788D5918BCF40187ULL, "296049065.ogg"},
	{0x788D5A80452C5D4FULL, "6097591_m.fxe"},
	{0x788D6205FCE0BC24ULL, "zz_vault_debug6203174.cl"},
	{0x788D637D979DD237ULL, "art\\characters\\creatures\\mubari\\animation\\drop.gad"},
	{0x788DAA5B14E34478ULL, "vxht_bartrand_n_5yo_0_0.vlm"},
	{0x788DAAA512DBEAD4ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_bda.phy"},
	{0x788DFFB55636A0A2ULL, "mag370_meredith6046833.cl"},
	{0x788E2DCAF27340B1ULL, "855358919.ogg"},
	{0x788EACDE93E2F4A4ULL, "dre000_dougal6128890.cl"},
	{0x788EB1F7B50C4B44ULL, "zz_pro_debug6059738.cl"},
	{0x788EE9D13C09354CULL, "6133500_m.fxe"},
	{0x788EF5EF13D69151ULL, "vxca_caverns1_mines_j1y_0_1.vlm"},
	{0x788F359FADD12CA7ULL, "pro000_refugee_female6052091.cl"},
	{0x788FA67D680EF13AULL, "l_mrl211ar_house.bnk"},
	{0x78901C8B390FB41AULL, "llac_ht_0_hightown_a3_1_1.dds"},
	{0x789033F62E2B41F2ULL, "6085833_f.fxe"},
	{0x789080DF30CF317AULL, "vxht_brothel_n_52p_0_0.vlm"},
	{0x78909C4D68CB35F3ULL, "vxlt_undercityencounter_1tf_0_0.vlm"},
	{0x78909E0A33F025DFULL, "445400363.ogg"},
	{0x7890AB88EEAEE8C4ULL, "6149905_m.fxe"},
	{0x7890C96D354CF878ULL, "6095020_m.fxe"},
	{0x7890CB588D8D9D32ULL, "vxht_0_hightown_c6v_0_0.vlm"},
	{0x7890FEE14ECC0A95ULL, "746939915.ogg"},
	{0x78910A9ED86D271EULL, "vxdr_0_deeproads_gbq_0_0.vlm"},
	{0x789123205B9ED706ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkright.evt"},
	{0x789159D7199EB130ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_lbrt.gad"},
	{0x7891632824334391ULL, "vxlt_0_lowtown_docks_a3_d24_0_0.vlm"},
	{0x78919A6D9BEFECE1ULL, "511969237.ogg"},
	{0x7891EEDA9BFEE0C6ULL, "hla_ht_0_hightown_-131071_0x.dds"},
	{0x7892333ADE17710BULL, "vxlt_backalley_n_1wf_0_0.vlm"},
	{0x7892C4C285BE9F5CULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_p.impa.gad"},
	{0x7892D1B34BF62C16ULL, "vxht_mansion2_2bc_0_0.vlm"},
	{0x7893787D97A261C4ULL, "art\\characters\\creatures\\mubari\\animation\\drop.evt"},
	{0x7893923AF0018AC0ULL, "mcr370_cut_player_rules6171826.cl"},
	{0x7893A5CF791F8726ULL, "6122860_f.fxe"},
	{0x7893C33D55EEC2CDULL, "mag311_emile_de_launcet6064717.cl"},
	{0x7893E4459149946FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masa_0n.dds"},
	{0x7894052061392DD9ULL, "vxlt_0_lowtown_f15_0_0.vlm"},
	{0x78940EB77B7DD3D2ULL, "6076375_m.fxe"},
	{0x789490437E7BAAF5ULL, "vxlt_backalley_g_1xl_0_0.vlm"},
	{0x7895641DA9C6B8A7ULL, "vxht_keep_n_643_0_0.vlm"},
	{0x78956E6949C85AC7ULL, "6059338_f.fxe"},
	{0x7895823C80B20B6EULL, "mag370_meredith6092182.cl"},
	{0x789614EC8B398142ULL, "6099417_m.fxe"},
	{0x789635B07B40376BULL, "6149919_m.fxe"},
	{0x78968F957A02E059ULL, "vxw_sundermount_31q_0_1.vlm"},
	{0x7896B8F397EE1EFEULL, "art\\characters\\weapons\\maces\\textures\\w_mce_dsmca_0t.dds"},
	{0x7896EEBC126EF660ULL, "929172664.ogg"},
	{0x78970576E234C8F1ULL, "bad.mao"},
	{0x78971AC0630AA93FULL, "prop1.psh.fxo"},
	{0x789764F5E690CF41ULL, "vxw_encounter1_ff_0_0.vlm"},
	{0x78976576A38EBA80ULL, "hlac_lt_0_lowtown_a2_n_65534_1x.dds"},
	{0x78976DF04AF6C4ECULL, "hlac_ca_cavern1_haunted_24427_3x.dds"},
	{0x7897829D57FD26D7ULL, "gen00fl_sebastian6179773.cl"},
	{0x7897AFB5942114C8ULL, "vxdr_primevaltomb_8uo_0_0.vlm"},
	{0x7897C4F2807FCB87ULL, "vxlt_0_lowtown_a2_n_c_f4m_0_0.vlm"},
	{0x78983A205BA51D51ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkright.gad"},
	{0x789887A512E5B6D1ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_bda.msh"},
	{0x7898A1A512E5E28FULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_bda.mmh"},
	{0x7898C3C285C3095AULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_p.impa.ani"},
	{0x7898CDB615AD7AB8ULL, "dae000_mother6136930.cl"},
	{0x7898F6AF19736A3EULL, "6152312_m.fxe"},
	{0x78993EBC8645D657ULL, "6155961_m.fxe"},
	{0x7899766BE296B2CAULL, "vxca_cavern1_haunted_efy_0_0.vlm"},
	{0x7899C2AC32C31AFEULL, "453324012.ogg"},
	{0x789A1EFBB4F2D2FBULL, "vxlt_warehouse_ec_0_0.vlm"},
	{0x789A6280ED8F9C5EULL, "369353829.ogg"},
	{0x789A98967932FB9EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_loop.ani"},
	{0x789AA15AEC7852B1ULL, "character_character_p59.psh.fxo"},
	{0x789ABF6FEFE83D86ULL, "hlac_w_encounter3_0_12x.dds"},
	{0x789C31AD6137E744ULL, "197484975.ogg"},
	{0x789C8D8C565AFF0CULL, "dae000ip_crowd_ff2_talk16.utp"},
	{0x789CA02903D4CC11ULL, "hlac_ca_caverns1_24422_0z.dds"},
	{0x789CD6E57D359D73ULL, "19702318.ogg"},
	{0x789D139792E46821ULL, "mag100_cullen6140947.cl"},
	{0x789D691C0048822EULL, "hla_ht_0_hightown_a3_-65535_0z.dds"},
	{0x789DBF2144D6640BULL, "701571363.ogg"},
	{0x789E34011F5FE408ULL, "hlac_lt_gallowscourtyard_n_-131072_4x.dds"},
	{0x789E3B205BA98AFFULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkright.ani"},
	{0x789E554232F62AA8ULL, "vxlt_backalley_g_1e2_0_0.vlm"},
	{0x789F08061647B35EULL, "art\\characters\\creatures\\desiredemon\\animation\\dblw_dmg_v1.evt"},
	{0x789F2749BC70B225ULL, "vxlt_0_lowtown_gpg_0_0.vlm"},
	{0x789F85708B519F29ULL, "6120747_m.fxe"},
	{0x78A06D5F7B874B20ULL, "hlac_lt_0_lowtown_65537_7x.dds"},
	{0x78A088EE3976A5E4ULL, "ico_shield_0.dds"},
	{0x78A18BE2F687EA94ULL, "ser221_gamlen6135261.cl"},
	{0x78A1907D97AEC7BDULL, "art\\characters\\creatures\\mubari\\animation\\drop.ani"},
	{0x78A1A66992D94D00ULL, "67588974.ogg"},
	{0x78A1B7F9AD34BC7DULL, "25378042.ogg"},
	{0x78A226849C7EE17EULL, "6138802_m.fxe"},
	{0x78A2E72323FC4309ULL, "hf_elsa.mor"},
	{0x78A303ADC0BC3A23ULL, "6059334_f.fxe"},
	{0x78A35EA75DF4A104ULL, "6149497_m.fxe"},
	{0x78A35F0F96DA850BULL, "vxw_blightlands_141_0_0.vlm"},
	{0x78A3AC9F3049FF46ULL, "6105675_m.fxe"},
	{0x78A3F3B877AC4A99ULL, "llac_ca_cavern1_haunted_24422_1.dds"},
	{0x78A49C09B0A79D6FULL, "plt_mrl350pt_newpath.nss"},
	{0x78A4A413CF0A37B0ULL, "vxlt_0_lowtown_a2_n_c_gj5_0_0.vlm"},
	{0x78A4FB6ECBDAF06AULL, "729222501.ogg"},
	{0x78A50FDEDB1275EFULL, "pro000_captain_wright6047488.cl"},
	{0x78A51DF0FB532418ULL, "vxht_mansion2_1bc_0_1.vlm"},
	{0x78A51E83961CA8C6ULL, "6107205_f.fxe"},
	{0x78A558C34D135F3DULL, "hlac_lt_gallowstemplar_f_2391_1x.dds"},
	{0x78A56F442BE6CA3AULL, "gen00fl_anders6201140.cl"},
	{0x78A5741624C20373ULL, "plt_isa200pt_thief.plo"},
	{0x78A5D0698BAB4EEDULL, "vxw_bonepit_e6_0_0.vlm"},
	{0x78A5D6D2D886E7C1ULL, "vxw_encounter1_ke_0_0.vlm"},
	{0x78A61F06164DF889ULL, "art\\characters\\creatures\\desiredemon\\animation\\dblw_dmg_v1.gad"},
	{0x78A66626E9EEDE47ULL, "vxlt_gallowstemplar_5oi_0_0.vlm"},
	{0x78A7158B5DB73AFEULL, "vxlt_gallowscourtyard_6y3_0_0.vlm"},
	{0x78A71D56EA2FDA64ULL, "272696509.ogg"},
	{0x78A74ED6CF01872FULL, "vxlt_warehouse_134_0_0.vlm"},
	{0x78A7548BAAB90EBBULL, "gen00fl_fenris6128022.cl"},
	{0x78A7D09E6A228B61ULL, "sse_ss_raider_m_1.gda"},
	{0x78A84940DE08A0BFULL, "hlac_ht_0_hightown_131074_0z.dds"},
	{0x78A8EE3FBC7F6125ULL, "dae011_drunk1_sta6144379.cl"},
	{0x78AB4572361CA60FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.cess.gad"},
	{0x78AB86943841281EULL, "gen_im_npc_arm_hum_hlm_rdra.uti"},
	{0x78AC9BC52897508FULL, "lgt120cr_orz_bln_traitor.utc"},
	{0x78ACF0061653C7D7ULL, "art\\characters\\creatures\\desiredemon\\animation\\dblw_dmg_v1.ani"},
	{0x78AD0632359B8FD4ULL, "qun210ar_javaris_encounter.are"},
	{0x78AD8AE7ABD8C107ULL, "hlac_ht_0_hightown_a2_65533_2x.dds"},
	{0x78AD900F9F54DD85ULL, "vxht_bartrand_n_5vw_0_0.vlm"},
	{0x78AE358E752D7DA9ULL, "995383363.ogg"},
	{0x78AE44FE6D3AB4FCULL, "vxht_0_hightown_aq3_0_0.vlm"},
	{0x78AE5D913CE748AAULL, "38330722.ogg"},
	{0x78AE95BC1F63EE2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_point_small.evt"},
	{0x78AEB1967943CF18ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_loop.gad"},
	{0x78AEBDC399DFE072ULL, "177095182.ogg"},
	{0x78AED0DA4F2F4EBDULL, "vxht_chantry_n_2d1_0_0.vlm"},
	{0x78AFBC17024CFF3CULL, "hlac_lt_0_lowtown_a2_n_c_-65534_0x.dds"},
	{0x78B03F0DE7015216ULL, "6180816_m.fxe"},
	{0x78B0576B0F76CB27ULL, "genip_pp_lti_standing.utp"},
	{0x78B067DB22C40818ULL, "6206255_m.fxe"},
	{0x78B073D588E90101ULL, "art\\vfx\\combat\\vfx_sta_rng_atk_fin_c.anb"},
	{0x78B0C28726693329ULL, "ach_soc_approval_max.dds"},
	{0x78B0D3EC01E02340ULL, "6146965_f.fxe"},
	{0x78B0D8A6F8345142ULL, "w_bonepitmining_a2_110_0.msh"},
	{0x78B0DAA6F83454F8ULL, "w_bonepitmining_a2_110_0.mmh"},
	{0x78B0E6A6F834689BULL, "w_bonepitmining_a2_110_0.mao"},
	{0x78B14E8501AA9974ULL, "6161208_m.fxe"},
	{0x78B15A72362135BCULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.cess.evt"},
	{0x78B250242094ABBAULL, "dae110ip_store5.utp"},
	{0x78B250F6BA298BDCULL, "lowtown_male_urchin_sta6164239.cl"},
	{0x78B2AFE437DEC0B1ULL, "hlac_ca_caverns2prc_5802_16x.dds"},
	{0x78B2F8A66F25B242ULL, "6093617_m.fxe"},
	{0x78B306E308575118ULL, "233561356.ogg"},
	{0x78B345B500DEE906ULL, "916174288.ogg"},
	{0x78B3C33C2D5F7321ULL, "375236719.ogg"},
	{0x78B43A1337EE0542ULL, "hlac_ca_cave_2194_8y.dds"},
	{0x78B45AD23DB08BCBULL, "687384533.ogg"},
	{0x78B4B83E7BE4F4CDULL, "6118365_f.fxe"},
	{0x78B4C07D682E39D8ULL, "l_mrl211ar_house.wbd"},
	{0x78B4D803DBE4F6C7ULL, "vxlt_gallowstemplar_5hh_0_0.vlm"},
	{0x78B4E508E5C795C3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_pailempty.msh"},
	{0x78B4F308E5C7AD1DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_pailempty.mmh"},
	{0x78B4F708E5C7B42EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_pailempty.mao"},
	{0x78B5AEBC1F6A37ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_point_small.gad"},
	{0x78B60B3398140E59ULL, "vxdr_primevaltomb_90h_0_0.vlm"},
	{0x78B63C99FC335027ULL, "6155778_m.fxe"},
	{0x78B643EB0C0E70E6ULL, "vxw_encounter2_46_0_0.vlm"},
	{0x78B6442152B540C0ULL, "6043485_m.fxe"},
	{0x78B6534B3B8DA480ULL, "dae000_elegant6166905.cl"},
	{0x78B6696626862D0FULL, "6101292_m.fxe"},
	{0x78B702B598DF03FBULL, "vxlt_warehouse_se_0_0.vlm"},
	{0x78B762724C9BCCECULL, "vxw_encounter1_dy_0_0.vlm"},
	{0x78B763AD05ADD7FDULL, "lt_0_lowtown_a3_n_c_12_0.mao"},
	{0x78B767AD05ADDE36ULL, "lt_0_lowtown_a3_n_c_12_0.mmh"},
	{0x78B76D62B7926A5EULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowsheildweapon.evt"},
	{0x78B771AD05ADEF34ULL, "lt_0_lowtown_a3_n_c_12_0.msh"},
	{0x78B7BF6CDD5982CCULL, "lt_backalley_n_mmi.gff"},
	{0x78B7D8CA4E6BB18CULL, "6096858_m.fxe"},
	{0x78B7E4FD7C2679FCULL, "hf_cgn_005.mop"},
	{0x78B836B1E8E268D4ULL, "dae200cr_cullen.utc"},
	{0x78B863CDA3FB554EULL, "513535462.ogg"},
	{0x78BA06383981A71CULL, "vxht_0_hightown_an2_0_0.vlm"},
	{0x78BA34C43032463EULL, "hlac_ca_caverns1_mines_24427_4y.dds"},
	{0x78BA9ABC5A74AB06ULL, "ser120_ghyslain6165799.cl"},
	{0x78BACE0ED5EF7CA8ULL, "one110_meeran6089516.cl"},
	{0x78BB187E1D49FCABULL, "static_static_p37.psh.fxo"},
	{0x78BB1B99065B7509ULL, "vxht_keep_n_616_0_0.vlm"},
	{0x78BB248461F4ABFAULL, "vxht_0_hightown_a2_c6j_0_0.vlm"},
	{0x78BB38FC16E25BA8ULL, "vxlt_backalley_g_1o0_0_0.vlm"},
	{0x78BBFE6305084A5AULL, "vxlt_warehouse_n_q3_0_0.vlm"},
	{0x78BC01EC914AF605ULL, "vxht_bartrand_n_5yv_0_0.vlm"},
	{0x78BC2449E59B42BCULL, "162336089.ogg"},
	{0x78BC58F429B365B8ULL, "678176904.ogg"},
	{0x78BCADBC1F7054EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_point_small.ani"},
	{0x78BCB6DB5F30CD7CULL, "vxht_0_hightown_at0_0_0.vlm"},
	{0x78BCD673D44C3F2DULL, "797113629.ogg"},
	{0x78BCE77CB8505A61ULL, "audiogui.gfx"},
	{0x78BD2BABB04B54D6ULL, "vxht_0_hightown_aia_0_0.vlm"},
	{0x78BD7703B15B5BCDULL, "character_rendertodepth_p1.vsh.fxo.dx11"},
	{0x78BD7CFA5B1E84C0ULL, "mer340_bonepit_comments6090829.cl"},
	{0x78BD8A4BFBEA2D3DULL, "41735793.ogg"},
	{0x78BDFEE4273F3022ULL, "fex_ruin_circ_d.mmh"},
	{0x78BE00E4273F33D8ULL, "fex_ruin_circ_d.msh"},
	{0x78BE54D588F50A7AULL, "art\\vfx\\combat\\vfx_sta_rng_atk_fin_c.mmh"},
	{0x78BE6CA456A69E3CULL, "6109241_f.fxe"},
	{0x78BE6E2AD856E173ULL, "790138353.ogg"},
	{0x78BE8462B798AF89ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowsheildweapon.gad"},
	{0x78BEDCE99D273D0CULL, "vxca_cave_1zz_0_0.vlm"},
	{0x78BF6272362D80D5ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.cess.ani"},
	{0x78BF8B363132685BULL, "gui_guipixelshader_p7.psh.fxo"},
	{0x78BFC061F3A0B1F4ULL, "6166308_m.fxe"},
	{0x78C02B19720FFE0DULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleaf.msh"},
	{0x78C0391972101640ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleaf.mao"},
	{0x78C03D1972101C93ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleaf.mmh"},
	{0x78C05C52D3EB6DD1ULL, "hlac_w_sundermount_65531_0x.dds"},
	{0x78C0B684A4085671ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley_1.mao"},
	{0x78C115B3C3000E70ULL, "hlac_lt_gamlenshouse_n_-1_6x.dds"},
	{0x78C1397D3E74A917ULL, "hlac_lt_0_lowtown_n_65536_0x.dds"},
	{0x78C18100E123F45DULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_1x.dds"},
	{0x78C1AC69D951C87FULL, "hlac_lt_gallowstemplar_n_2388_5x.dds"},
	{0x78C1BB6C33AC4A01ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_3.ani"},
	{0x78C1C9F627630C6DULL, "190065987.ogg"},
	{0x78C1D130DEE3D467ULL, "vxlt_0_lowtown_a2_n_c_gc5_0_0.vlm"},
	{0x78C2096C514080B8ULL, "606099073.ogg"},
	{0x78C20DB65B638AC3ULL, "vxca_caverns2_86w_0_0.vlm"},
	{0x78C266CB1E6A88E7ULL, "498754219.ogg"},
	{0x78C2BE0DEBB429C4ULL, "vxca_cavern1_haunted_cp3_0_0.vlm"},
	{0x78C301572CA2184CULL, "532555094.ogg"},
	{0x78C3392327272AF4ULL, "843196120.ogg"},
	{0x78C349E0CDDB0F47ULL, "vxlt_gallowstemplar_5uk_0_0.vlm"},
	{0x78C3971E279C4C52ULL, "6160605_m.fxe"},
	{0x78C3F247A44BBF96ULL, "vxw_bonepitdestroyed_a3_181_0_0.vlm"},
	{0x78C405C3444922EFULL, "vxht_mansion2_1by_0_0.vlm"},
	{0x78C409C64E998F34ULL, "one220_yevhen6116321.cl"},
	{0x78C41A046032B6D5ULL, "508418197.ogg"},
	{0x78C423E7FB424B39ULL, "hlac_lt_gallowstemplar_f_7198_2y.dds"},
	{0x78C46D803A2B26C1ULL, "6138827_m.fxe"},
	{0x78C47870FC0C0FD1ULL, "vxlt_0_lowtown_docks_a3_n_d0i_0_0.vlm"},
	{0x78C4A26F7D8E973CULL, "6110617_m.fxe"},
	{0x78C4ACDE79E0A168ULL, "vxca_caverns1_gkc_0_0.vlm"},
	{0x78C51DF8726CE56EULL, "6060189_m.fxe"},
	{0x78C5328216A937BCULL, "vxht_chantry_n_27m_0_0.vlm"},
	{0x78C55562B79E7ED7ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowsheildweapon.ani"},
	{0x78C570C0696E27F7ULL, "vxlt_0_lowtown_n_f0s_0_0.vlm"},
	{0x78C574928AD51C7BULL, "vxlt_warehouse_td_0_0.vlm"},
	{0x78C5CE9351998FA6ULL, "vxca_caverns2prc_9xh_0_0.vlm"},
	{0x78C62FAE5BC63824ULL, "gen00fl_anders6103743.cl"},
	{0x78C6321835C1A871ULL, "zz_fen_debug6134355.cl"},
	{0x78C64561E76C67B0ULL, "hlac_lt_0_lowtown_a2_n_1_5x.dds"},
	{0x78C6DD366D1E959AULL, "6096033_f.fxe"},
	{0x78C70FC5B707F361ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_0x.dds"},
	{0x78C75CA79BE9415FULL, "dae000_orsino6085526.cl"},
	{0x78C771CEDCBBA678ULL, "vxht_bartrand_n_5mk_0_0.vlm"},
	{0x78C79F134AA6E92CULL, "6031278_f.fxe"},
	{0x78C7A01A7CA6302DULL, "vxdr_primevalentrance_3bt_0_0.vlm"},
	{0x78C7BB2418C93A6EULL, "w_sundermount_187_0.phy"},
	{0x78C7C66C33B0C8A3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_3.gad"},
	{0x78C80D48707B8ED6ULL, "hlac_dr_primevalentrance_1726_2z.dds"},
	{0x78C8AB94ACB8D5B7ULL, "llac_ht_0_hightown_a2_-196607_0.dds"},
	{0x78C8FAEDEC6FA204ULL, "vxht_chantry_1yk_0_0.vlm"},
	{0x78C90D3BE053E047ULL, "vxht_0_hightown_byo_0_0.vlm"},
	{0x78C92073D2B203B1ULL, "dae210_gate_guard6084316.cl"},
	{0x78C98AD06814904CULL, "390567664.ogg"},
	{0x78C9B2C793DC1AAAULL, "vxht_0_hightown_a3_c74_0_0.vlm"},
	{0x78CA06672FAFBCCEULL, "219124861.ogg"},
	{0x78CA2B547F5CDFFAULL, "dae000_bodhan6146837.cl"},
	{0x78CA7A47601C4471ULL, "vxdr_primevaltomb_7ya_0_0.vlm"},
	{0x78CB62C416A9CC9BULL, "hlac_ht_mansion2b_818_49x.dds"},
	{0x78CBB40C285B9321ULL, "gen_im_npc_arm_hum_hlm_slva.uti"},
	{0x78CBED3235B615B6ULL, "qun210ar_javaris_encounter.ncs"},
	{0x78CC1992DB29D1B1ULL, "llac_lt_0_lowtown_docks_65535_3.dds"},
	{0x78CC1CFD3F5F1FA9ULL, "vxht_brothel_n_6b0_0_0.vlm"},
	{0x78CC4B3606039D17ULL, "art\\vfx\\materials\\gui\\vfx_gui_combo.mao"},
	{0x78CC7E448E25AF99ULL, "hlac_lt_0_lowtown_a2_n_c_-1_3x.dds"},
	{0x78CCDCCE3984BC06ULL, "351822616.ogg"},
	{0x78CD5519042B9AB5ULL, "hlac_ca_caverns1_24423_5z.dds"},
	{0x78CE6E768406F87CULL, "vxht_0_hightown_a2_amc_0_0.vlm"},
	{0x78CE84993232B995ULL, "gen00fl_aveline6132673.cl"},
	{0x78CE968BCF487782ULL, "hlac_lt_0_lowtown_a2_n_131071_3z.dds"},
	{0x78CEA0B764E70E68ULL, "6100281_m.fxe"},
	{0x78CEA36C33B6AB88ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_3.evt"},
	{0x78CED7291B1671B7ULL, "gen00fl_varric6033704.cl"},
	{0x78CF555D386F6101ULL, "6095431_f.fxe"},
	{0x78D021B6EDC98E5FULL, "vxdr_primevaltomb_75k_0_0.vlm"},
	{0x78D0447DDF09FE58ULL, "dr_slavers_3218_ll.mmh"},
	{0x78D05ABE94F972BDULL, "6156236_m.fxe"},
	{0x78D063A8A6678766ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_001a_0n.dds"},
	{0x78D0BEF0C2D5C7D6ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.impa.ani"},
	{0x78D0C36DF90ADA44ULL, "13411493.ogg"},
	{0x78D124259DCB9D16ULL, "one300_zevran6076922.cl"},
	{0x78D16BC0C2893E99ULL, "6111408_m.fxe"},
	{0x78D1A7E8D1E9D839ULL, "lt_backalley_g_0_0_ll.mmh"},
	{0x78D211FDACFFFAC9ULL, "vxht_mansion2b_1pl_0_1.vlm"},
	{0x78D3053235BC5D4FULL, "qun210ar_javaris_encounter.lst"},
	{0x78D307F9C43B82DAULL, "lt_0_lowtown_a2_n_45_0_ll.mmh"},
	{0x78D3E66F7CCB34FBULL, "vxlt_warehouse_yg_0_0.vlm"},
	{0x78D3E672FCFF9B9CULL, "459998794.ogg"},
	{0x78D3E8DAF82D36FEULL, "6106379_m.fxe"},
	{0x78D41A6D85164A4BULL, "796212703.ogg"},
	{0x78D4303A519F01D6ULL, "138070665.ogg"},
	{0x78D442FCC374532EULL, "ser210_witness6168161.cl"},
	{0x78D45B1624E94910ULL, "plt_isa200pt_thief.nss"},
	{0x78D48D5A853F23D9ULL, "851527151.ogg"},
	{0x78D5068EE4BD5C20ULL, "vxdr_primevaltomb_7lo_0_0.vlm"},
	{0x78D51A1679DB8445ULL, "dae000_arishok6080492.cl"},
	{0x78D5C433C2242541ULL, "356283097.ogg"},
	{0x78D6DAA559355CFCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_ian.ani"},
	{0x78D7066BA552B90BULL, "vxca_caverns1_dq9_0_0.vlm"},
	{0x78D8083E45E0DCFAULL, "vxht_0_hightown_a2_c8d_0_0.vlm"},
	{0x78D88B78C7A8FCFBULL, "hlac_lt_0_lowtown_a2_131072_2z.dds"},
	{0x78D8BADC5F81C210ULL, "vxw_bonepit_21s_0_0.vlm"},
	{0x78D8D1411EE82B08ULL, "character_character_p256.psh.fxo.dx11.assembly"},
	{0x78D966F37F1AE695ULL, "6218701_m.fxe"},
	{0x78D9A44956BCEE46ULL, "6198189_m.fxe"},
	{0x78D9A73B2A9838FDULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsarch.mao"},
	{0x78D9AB3B2A983F36ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsarch.mmh"},
	{0x78D9B53B2A985034ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsarch.msh"},
	{0x78D9DCE0BE24540AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_female.gad"},
	{0x78DA2B34795443CAULL, "zz_spy131cr_booker_thug2.utc"},
	{0x78DA421739131F18ULL, "792659279.ogg"},
	{0x78DA625073D4CFEFULL, "cs_rom000cs_anders.wbd"},
	{0x78DA6715AD850614ULL, "art\\vfx\\creatures\\fxc_ogre_db_d_c.anb"},
	{0x78DB02DDF06B02C6ULL, "733574168.ogg"},
	{0x78DB83548B6D76FCULL, "hlac_w_woundedcoast_262150_0y.dds"},
	{0x78DBD353B75014B1ULL, "dre100cr_dwarven_scout.utc"},
	{0x78DBEBD5B748A23AULL, "gen00fl_aveline6113985.cl"},
	{0x78DC7B91F286C66AULL, "hlac_w_bonepitmining_a2_0_4z.dds"},
	{0x78DD3CDC526E1376ULL, "651279295.ogg"},
	{0x78DD5E1EE645EECBULL, "6033618_m.fxe"},
	{0x78DD8C4A536E8B30ULL, "20822686.ogg"},
	{0x78DD9CC488656B92ULL, "6118296_f.fxe"},
	{0x78DDC6F0C2E05FC7ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.impa.evt"},
	{0x78DDD2CEDDC9244CULL, "625552496.ogg"},
	{0x78DDD58F397F4746ULL, "llac_lt_0_lowtown_docks_a3_n_131070_4.dds"},
	{0x78DDE7A5593B913EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_ian.gad"},
	{0x78DDEE23509B326DULL, "6104598_m.fxe"},
	{0x78DE4A7A13CD78CFULL, "340960654.ogg"},
	{0x78DE6834002806FCULL, "hlac_lt_0_lowtown_docks_-196608_1z.dds"},
	{0x78DEE1FD72EF52F7ULL, "671760583.ogg"},
	{0x78DF45663BDE6F5BULL, "hlac_lt_0_lowtown_a2_-65537_6y.dds"},
	{0x78DF67142AB48E7BULL, "primitives.mat"},
	{0x78DF6D15F6BC6E58ULL, "hlac_lt_0_lowtown_a2_n_c_-2_3x.dds"},
	{0x78DF9188EED0204CULL, "vxlt_darktown_4x2_0_0.vlm"},
	{0x78DFA609B4166693ULL, "6150517_m.fxe"},
	{0x78E03DB8B82C49B0ULL, "6126946_m.fxe"},
	{0x78E04524C0295D2DULL, "vxw_encounter3_9e_0_0.vlm"},
	{0x78E05DAADE2E8CE9ULL, "hla_ht_0_hightown_a3_-1_0z.dds"},
	{0x78E0880B93909DB8ULL, "6121172_m.fxe"},
	{0x78E0BB2AA3F0F1F9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile1a_1.mao"},
	{0x78E0BFE0BE2A41A1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_female.evt"},
	{0x78E0EDB12F930D6BULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_fire02a.mao"},
	{0x78E0F5C806D74B60ULL, "lgt100cr_gang_2_ht.utc"},
	{0x78E11FFFEAB99A72ULL, "6085852_m.fxe"},
	{0x78E16BE8A8276D9CULL, "6080479_m.fxe"},
	{0x78E1F721505A8B12ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk_sa.dds"},
	{0x78E21A16510FEF3FULL, "vxw_encounter1_dv_0_0.vlm"},
	{0x78E25FBBE7898824ULL, "vxw_bonepitdestroyed_a3_17r_0_0.vlm"},
	{0x78E297F42F58C7DEULL, "462846088.ogg"},
	{0x78E3C6EDBC890E8FULL, "vxlt_warehouse_112_0_0.vlm"},
	{0x78E3D1CA40F7C390ULL, "mag121_idunna6084005.cl"},
	{0x78E3D291CE5BB827ULL, "6092498_f.fxe"},
	{0x78E3EF799223B978ULL, "882995739.ogg"},
	{0x78E46D27FC0124F2ULL, "vxht_bartrand_n_72z_0_0.vlm"},
	{0x78E4BAE37C72360CULL, "vxca_caverns2_3i2_0_0.vlm"},
	{0x78E4CD2F53F81AF8ULL, "6109390_m.fxe"},
	{0x78E4E7F0C2E6B610ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.impa.gad"},
	{0x78E50DC35DFBE7D4ULL, "vxlt_undercityencounter_192_0_0.vlm"},
	{0x78E51F2B88781D59ULL, "6121989_f.fxe"},
	{0x78E585A93C0E831FULL, "vxca_caverns2prc_9l1_0_0.vlm"},
	{0x78E61C2866693B45ULL, "vxht_0_hightown_a2_c6a_0_0.vlm"},
	{0x78E6B06D5F066045ULL, "11923135.ogg"},
	{0x78E71687556B49FFULL, "vxdr_0_deeproads_eu3_0_0.vlm"},
	{0x78E74ED39E042AE6ULL, "6088150_m.fxe"},
	{0x78E7F181B1BF71BDULL, "vxlt_0_lowtown_a2_n_c_h27_0_0.vlm"},
	{0x78E803D9E3594F4AULL, "llac_lt_0_lowtown_a2_-65537_4.dds"},
	{0x78E812B3DDFCA8E9ULL, "32019064.ogg"},
	{0x78E84615AD910BD7ULL, "art\\vfx\\creatures\\fxc_ogre_db_d_c.mmh"},
	{0x78E87849BE5D41EFULL, "zz_followers_banter6158664.cl"},
	{0x78E8A2D4B9BEFCDAULL, "744697454.ogg"},
	{0x78E8DB4DF96BB395ULL, "vxlt_foundry_n_zw_0_0.vlm"},
	{0x78E8E49536AFADB2ULL, "42039002.ogg"},
	{0x78E900B7234B50A9ULL, "vxht_brothel_n_6h6_0_0.vlm"},
	{0x78E91BA6829B884AULL, "gen00fl_sebastian6179802.cl"},
	{0x78EA365F20E7CA58ULL, "hlac_w_sundermount_-262149_0x.dds"},
	{0x78EA4C8423F048DCULL, "lgt000cr_ultra_turn_in.ncs"},
	{0x78EB56231D9B92E5ULL, "6119769_m.fxe"},
	{0x78EC386E3492C24EULL, "art\\characters\\weapons\\maces\\textures\\w_mce_dsmca_0d.dds"},
	{0x78EC4B41B38F1B68ULL, "6062992_m.fxe"},
	{0x78EC99B4CE7C98DCULL, "6181779_f.fxe"},
	{0x78ECD01F484F50D3ULL, "dae000_mother6127182.cl"},
	{0x78ECF6852C106E7BULL, "w_bonepitmining_a2_54_0.mmh"},
	{0x78ED02852C108218ULL, "w_bonepitmining_a2_54_0.mao"},
	{0x78ED14852C10A165ULL, "w_bonepitmining_a2_54_0.msh"},
	{0x78ED217198DB8825ULL, "vxca_cavern1_haunted_j5h_0_0.vlm"},
	{0x78EDB7E0BE34BEF0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\kiss_sweet_action_female.ani"},
	{0x78EDD20AC9D570B1ULL, "420705333.ogg"},
	{0x78EE457257E7F783ULL, "474767321.ogg"},
	{0x78EE48A0B191BFDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\right_down_rhand.ani"},
	{0x78EEB701B21F75ADULL, "vxw_encounter3_2d_0_0.vlm"},
	{0x78EEC155412609C2ULL, "86204141.ogg"},
	{0x78EEEE0A8D72DF5FULL, "6113291_m.fxe"},
	{0x78EF76A914AED9A9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galofliplo.mao"},
	{0x78EF82A914AEEE42ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galofliplo.mmh"},
	{0x78EF84A914AEF1F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galofliplo.msh"},
	{0x78F07E1B6BF133D5ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_bdy_hlk_0d.dds"},
	{0x78F08BF3430607BFULL, "vxw_encounter1_iq_0_0.vlm"},
	{0x78F0C57753713F9EULL, "665091016.ogg"},
	{0x78F0D198D97FA0A4ULL, "vxw_bonepitdestroyed_a3_18s_0_0.vlm"},
	{0x78F11B591ED6ADCBULL, "6103913_m.fxe"},
	{0x78F13477E9C0D0A2ULL, "6096028_f.fxe"},
	{0x78F1A7F86863C668ULL, "vxht_0_hightown_a2_al9_0_1.vlm"},
	{0x78F1B3E329B2437DULL, "6102673_m.fxe"},
	{0x78F1EA48C8A98D20ULL, "vxdr_primevaltomb_7bm_0_0.vlm"},
	{0x78F20BDEB206C573ULL, "6125389_f.fxe"},
	{0x78F248CF82166676ULL, "hlac_w_bonepitmining_a2_393216_0x.dds"},
	{0x78F2BA1320408F28ULL, "gen_im_trs_krk_hgh_junk2_01.uti"},
	{0x78F2F3B6F4A30702ULL, "vxlt_darktown_59z_0_0.vlm"},
	{0x78F302C670D6788CULL, "504912236.ogg"},
	{0x78F438EE77398A40ULL, "hlac_ca_cavern1_haunted_24423_12z.dds"},
	{0x78F43DFF46581023ULL, "156241547.ogg"},
	{0x78F44D16D84C0007ULL, "mer110_jansen6032866.cl"},
	{0x78F47346D000F419ULL, "zz_prmpt_debug6200955.cl"},
	{0x78F48AAA94A20109ULL, "271295058.ogg"},
	{0x78F4C398907AB5CEULL, "710722568.ogg"},
	{0x78F5669E6A9F55D9ULL, "966287968.ogg"},
	{0x78F59FA6BED4FC3CULL, "hlac_w_encounter2_65539_0z.dds"},
	{0x78F5A6E2F90DF45FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ff_basic.msh"},
	{0x78F5ACE2F90DFE01ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ff_basic.mmh"},
	{0x78F5B8E2F90E12A2ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ff_basic.mao"},
	{0x78F5C0B08FFA4366ULL, "6083050_m.fxe"},
	{0x78F5DA1BC249F622ULL, "docks_fem_sailor_mob6176302.cl"},
	{0x78F65651EE9CABBBULL, "683050924.ogg"},
	{0x78F73787BBE3F786ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_act_c.anb"},
	{0x78F74D2AEB61CC15ULL, "vxlt_foundry_n_uv_0_0.vlm"},
	{0x78F74EF12DACF608ULL, "6163646_m.fxe"},
	{0x78F770208B8AEFFCULL, "hlac_lt_gallowscourtyard_n_-131072_3y.dds"},
	{0x78F7729415416929ULL, "vxht_brothel_n_6u5_0_0.vlm"},
	{0x78F7B280CBC87297ULL, "follower_banter6165451.cl"},
	{0x78F81A29C9FF194EULL, "vxdr_0_deeproads_eq2_0_0.vlm"},
	{0x78F88AF654A8BFDCULL, "1023407755.ogg"},
	{0x78F8B2855DB89D49ULL, "6087170_m.fxe"},
	{0x78F96F25F2F1653BULL, "hlac_lt_darktown_-1_43z.dds"},
	{0x78F99F3D9AE0AA41ULL, "llac_lt_0_lowtown_n_-65535_6.dds"},
	{0x78F9A1B49D0277F2ULL, "291204761.ogg"},
	{0x78FA0069E1BF16BBULL, "dre120pt_main.ncs"},
	{0x78FA92EC2C609E18ULL, "438718359.ogg"},
	{0x78FA9E248196BE64ULL, "6202800_m.fxe"},
	{0x78FABD1EA58533E8ULL, "6069892_m.fxe"},
	{0x78FAD1F26F6CFF14ULL, "lgt000ip_ultra_elf_remains.ncs"},
	{0x78FB199BB2956A45ULL, "one280_emrys6104290.cl"},
	{0x78FB60A0B19C72AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\right_down_rhand.evt"},
	{0x78FB9BE12345A74FULL, "6085505_m.fxe"},
	{0x78FB9F084CF2D393ULL, "lt_0_lowtown_docks_a3_n_39_0.anb"},
	{0x78FBD039B291BEDAULL, "6020417_m.fxe"},
	{0x78FBFAA8D6844C7EULL, "character_character_p203.psh.fxo.dx11.assembly"},
	{0x78FCAA97F2B76A8BULL, "6112551_m.fxe"},
	{0x78FCACE283A38D89ULL, "dm_gen_3.mor"},
	{0x78FCDF6E71383FD3ULL, "959567105.ogg"},
	{0x78FCE5EFFBD5E361ULL, "6093824_f.fxe"},
	{0x78FDD79CDA472A12ULL, "6103592_m.fxe"},
	{0x78FE5B8BEC56D905ULL, "6105412_m.fxe"},
	{0x78FEB7FC2DED5FC0ULL, "vxlt_gallowstemplar_f_5yw_0_0.vlm"},
	{0x78FED097F0211FE6ULL, "vxw_blightlands_w0_0_0.vlm"},
	{0x78FEE90E943846E3ULL, "vxlt_hangedman_n_1i1_0_0.vlm"},
	{0x78FF0AE7D900E360ULL, "gen00fl_sebastian6197246.cl"},
	{0x78FF113D3B6FD61CULL, "6027375_m.fxe"},
	{0x78FF27A93087D208ULL, "vxht_brothel_n_4dm_0_0.vlm"},
	{0x78FF2F29BE64258CULL, "vxlt_0_lowtown_ems_0_0.vlm"},
	{0x79000769323FF161ULL, "vxw_woundedcoast_nb_0_0.vlm"},
	{0x79002877714885AFULL, "lt_0_lowtown_docks_a3_113_0.mao"},
	{0x79002C1C1A4ECD7FULL, "vxlt_warehouse_10q_0_0.vlm"},
	{0x7900347771489A4CULL, "lt_0_lowtown_docks_a3_113_0.mmh"},
	{0x79003A777148A43EULL, "lt_0_lowtown_docks_a3_113_0.msh"},
	{0x7900473FCDBD2F1EULL, "139676259.ogg"},
	{0x790060FFDF7A7AEEULL, "6126456_m.fxe"},
	{0x7900A3F4B5D7AD68ULL, "vxw_encounter2_qr_0_1.vlm"},
	{0x790196992632841AULL, "225545564.ogg"},
	{0x7901FAACD89586C7ULL, "51802562.ogg"},
	{0x7902050D1B21993BULL, "tbe100im_sturdy_belt.uti"},
	{0x7902082C60A55B01ULL, "960091772.ogg"},
	{0x79021D936CEDD1EAULL, "vxht_brothel_n_6du_0_0.vlm"},
	{0x79025F0D9275AEF3ULL, "dae000ip_crowd_ff7_talk7.utp"},
	{0x790261A0B1A29358ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\right_down_rhand.gad"},
	{0x7902A61ABB35EDF9ULL, "143603475.ogg"},
	{0x7902FFE24A556C45ULL, "vxht_0_hightown_a2_c8o_0_0.vlm"},
	{0x7903C6DA84C0C749ULL, "972208863.ogg"},
	{0x7903FB549D27E1FEULL, "vxw_encounter3_fo_0_0.vlm"},
	{0x7904378595677094ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_patient_andersintro.gad"},
	{0x79047FB3BBE2CD9FULL, "hlac_ht_brothel_n_3369_0y.dds"},
	{0x7904C61BDA94D090ULL, "512823704.ogg"},
	{0x7904E6561F2C5334ULL, "6116097_m.fxe"},
	{0x790525266B2BC807ULL, "vxlt_undercityserialkiller_11y_0_0.vlm"},
	{0x790534886042A668ULL, "vxlt_0_lowtown_docks_a3_d0g_0_0.vlm"},
	{0x79053A87BBF03A31ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_act_c.mmh"},
	{0x7905CE963EED73F1ULL, "6155264_m.fxe"},
	{0x790605331635C06AULL, "6127271_m.fxe"},
	{0x790692896B830A9AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_flat.phy"},
	{0x7906B1205E764C71ULL, "848784220.ogg"},
	{0x7906C8D0B3BC8113ULL, "6080667_m.fxe"},
	{0x7906C912EA6EC878ULL, "str_sh_2.mao"},
	{0x7906CCCBE4A55441ULL, "vxlt_hangedman_1xs_0_0.vlm"},
	{0x7906CE9FF13FCFFFULL, "6150979_m.fxe"},
	{0x7907383E3AED2A26ULL, "hlac_lt_undercitydungeon_-1_34x.dds"},
	{0x79073FFD9916D4BBULL, "gen00fl_merrill6106512.cl"},
	{0x79078E3B6AEDFE83ULL, "vxht_brothel_n_631_0_0.vlm"},
	{0x7907E604E009B295ULL, "685607593.ogg"},
	{0x7908192673FBD5E1ULL, "vxht_0_hightown_a2_c72_0_0.vlm"},
	{0x7908BA241900CF31ULL, "w_sundermount_187_0.mmh"},
	{0x7908C2EEB9B941F5ULL, "6150335_m.fxe"},
	{0x7908C6241900E3D2ULL, "w_sundermount_187_0.mao"},
	{0x7908D4241900FBEFULL, "w_sundermount_187_0.msh"},
	{0x7908FF8B63C7297DULL, "face_rendertoshadowmap_p0.vsh.fxo"},
	{0x79097B6C37F28E7BULL, "6096228_m.fxe"},
	{0x7909901EF3949D48ULL, "vxw_woundedcoast_ng_0_0.vlm"},
	{0x7909DF4A2883350EULL, "vxw_blightlands_126_0_0.vlm"},
	{0x7909FBC30A45432DULL, "vxlt_0_lowtown_a2_eqf_0_0.vlm"},
	{0x790A0ADA0E27F406ULL, "dr_slavers_mmap_ph.dds"},
	{0x790A2E8F9A5696F7ULL, "sst_ss_cityelf_m_1.gda"},
	{0x790A4216A487D75AULL, "6027517_m.fxe"},
	{0x790A61FE184B5C99ULL, "458795139.ogg"},
	{0x790A9F6E03AEFE9FULL, "264378690.ogg"},
	{0x790B4685956DA972ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_patient_andersintro.ani"},
	{0x790B987B21EFFDC9ULL, "348983589.ogg"},
	{0x790C3ADC00373698ULL, "lt_0_lowtown_19_0_ll.mmh"},
	{0x790D155569EA2D61ULL, "6076278_f.fxe"},
	{0x790DA106B05A3E0CULL, "vxlt_0_lowtown_ehp_0_0.vlm"},
	{0x790DF6BAAE55B12EULL, "41888148.ogg"},
	{0x790E65BE8D0A05D3ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_elna_0n.dds"},
	{0x790EC2B176E52302ULL, "6084113_f.fxe"},
	{0x790EE55F12FFA1C6ULL, "starline.dds"},
	{0x790EF3943C8B0674ULL, "138316095.ogg"},
	{0x790FE81CE5922325ULL, "crt_spider.wbd"},
	{0x79106411C820B58FULL, "hlac_lt_0_lowtown_a2_-131072_7x.dds"},
	{0x791099E279876A4DULL, "6147843_m.fxe"},
	{0x7910D2037D56CEECULL, "vxw_bonepitdestroyed_a3_26n_0_0.vlm"},
	{0x7910D91F10E3BDFBULL, "vxlt_blackemporium_10e_0_0.vlm"},
	{0x7910EB8F5E35A71AULL, "hlac_lt_darktown_-1_35z.dds"},
	{0x791121D8261FAE8EULL, "6110475_m.fxe"},
	{0x79112FBE8EE29CCEULL, "vxlt_warehouse_14p_0_0.vlm"},
	{0x79114EE982036600ULL, "892386442.ogg"},
	{0x791168580495608DULL, "6114928_m.fxe"},
	{0x7911800837D46F75ULL, "332281387.ogg"},
	{0x7911A1DC355081C8ULL, "vxht_0_hightown_a2_c77_0_0.vlm"},
	{0x7911D696D2DC2F38ULL, "hlac_lt_gallowscourtyard_n_65536_0z.dds"},
	{0x7911F1A37481C76DULL, "vxlt_0_lowtown_g94_0_0.vlm"},
	{0x7911F21875E9E77BULL, "vxw_encounter3_8t_0_0.vlm"},
	{0x7912B0F26F815945ULL, "lgt000ip_ultra_elf_remains.utp"},
	{0x7912BBA16BB5C9FFULL, "gen00fl_isabela6096778.cl"},
	{0x7913293A972D94C3ULL, "3603807.ogg"},
	{0x791397035D21E087ULL, "vxlt_undercityserialkiller_16x_0_0.vlm"},
	{0x7913D3504F850646ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_lft.ani"},
	{0x791436C4FD10299EULL, "vxlt_warehouse_n_vr_0_0.vlm"},
	{0x79154FF39009642BULL, "vxlt_warehouse_e3_0_0.vlm"},
	{0x7915A6B973DBB890ULL, "llac_lt_0_lowtown_65538_2.dds"},
	{0x7915F9FBDAC4DC90ULL, "396843638.ogg"},
	{0x7915FFBD276F45FEULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowshieldweapon.evt"},
	{0x791614328E688A46ULL, "hlac_lt_0_lowtown_docks_n_-131074_0z.dds"},
	{0x7916D44B4BC9CAECULL, "vxlt_0_lowtown_a2_f5l_0_0.vlm"},
	{0x7917B0BDAE2C7F5EULL, "vxht_0_hightown_a3_bx8_0_0.vlm"},
	{0x7917D278FE268DB0ULL, "hightown_fem_noble2_mob6169469.cl"},
	{0x7917FD6E4EE776C1ULL, "vxht_mansion_n_5zp_0_0.vlm"},
	{0x7918132697AF4ADAULL, "dae300_bran6143467.cl"},
	{0x7918B4D78D3C4F47ULL, "art\\vfx\\combat\\vfx_g_spider_leap_act_c.anb"},
	{0x7918C8F9E7FABA07ULL, "art\\vfx\\creatures\\vfx_c_hd_roar_c.anb"},
	{0x79191BBF4B3B1600ULL, "6102694_m.fxe"},
	{0x7919265129DC55A2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexb.phy"},
	{0x79199B675FC38098ULL, "vxlt_gallowsprison_n_6ju_0_0.vlm"},
	{0x7919F4C79F235155ULL, "dae000ip_crowd_lm2_drunk1.utp"},
	{0x7919F958C884D400ULL, "gen00fl_isabela_init6077955.cl"},
	{0x7919FF11661CDCDBULL, "qcr000_arishok6060144.cl"},
	{0x791A7C50740ADFC5ULL, "cs_rom000cs_anders.bnk"},
	{0x791AA10949E4B854ULL, "6108951_m.fxe"},
	{0x791AF8BF227AA08FULL, "42423754.ogg"},
	{0x791B594768B156F6ULL, "qun221_viscount6069376.cl"},
	{0x791B7175E8535E57ULL, "llac_ht_0_hightown_a3_-131074_2.dds"},
	{0x791B74A9C2CAEDB4ULL, "hlac_ht_0_hightown_a2_131071_0x.dds"},
	{0x791BC912F0E29862ULL, "cs_qcr000cs_thechampion.bnk"},
	{0x791BCCC8782477E3ULL, "vxlt_hangedman_n_1c3_0_0.vlm"},
	{0x791C0F3DCAC73CBEULL, "109121408.ogg"},
	{0x791C7B8B91792655ULL, "vxht_bartrand_n_64n_0_0.vlm"},
	{0x791CACD293A392E0ULL, "6048808_m.fxe"},
	{0x791D3FDF9E8BD6A0ULL, "vxdr_primevaltomb_7uh_0_0.vlm"},
	{0x791DC38DF66A2944ULL, "vxca_caverns2_7xm_0_0.vlm"},
	{0x791DD55166E7DBB2ULL, "717792469.ogg"},
	{0x791DD7886C2E1906ULL, "vxw_encounter3_dq_0_0.vlm"},
	{0x791E03102514D53FULL, "hlac_w_bonepit_262149_0y.dds"},
	{0x791E13D1AE1060A4ULL, "6108615_f.fxe"},
	{0x791E2D491A37AC95ULL, "dae000_orana6149058.cl"},
	{0x791E9886E6B2CF62ULL, "gen00fl_isabela6099336.cl"},
	{0x791EB090535AB686ULL, "vxca_caverns1_ik7_0_0.vlm"},
	{0x791ED46FADE13983ULL, "pro000_athenril6048774.cl"},
	{0x791F014D50DA02EAULL, "vxht_brothel_n_6jw_0_0.vlm"},
	{0x791F43E06F4CE76CULL, "vxw_bonepitdestroyed_a3_21o_0_0.vlm"},
	{0x791F5E25446DC6E7ULL, "6057094_m.fxe"},
	{0x791F94658836D95EULL, "6053327_m.fxe"},
	{0x791FDA34F68B790DULL, "6113828_m.fxe"},
	{0x792013B927469A48ULL, "vxht_0_hightown_a2_c24_0_0.vlm"},
	{0x79205C76FB79B44AULL, "w_bonepit_47_0.mao"},
	{0x79206076FB79BA99ULL, "w_bonepit_47_0.mmh"},
	{0x792063F567DFFFFBULL, "vxw_encounter3_5w_0_0.vlm"},
	{0x79206A76FB79CB87ULL, "w_bonepit_47_0.msh"},
	{0x7920EB66A03B8ABAULL, "gen00fl_fenris6126508.cl"},
	{0x792122157AFD5FC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_faint_loop.gad"},
	{0x792123615EA04F0DULL, "395716565.ogg"},
	{0x792173764C220F64ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage3_.ani"},
	{0x7921ADE2F9333B7EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ff_basic.phy"},
	{0x7921BB504F911AB7ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_lft.evt"},
	{0x792226420CC51AE1ULL, "hlac_lt_undercityserialkiller_-1_23z.dds"},
	{0x792270287E9CCA07ULL, "165304387.ogg"},
	{0x7922EAF28E832855ULL, "227268626.ogg"},
	{0x7923084C6AC25B1FULL, "vxht_keep_5xk_0_0.vlm"},
	{0x79231CD855923B33ULL, "dae000_arishok6081310.cl"},
	{0x792348221250F56AULL, "gen00fl_fenris6126720.cl"},
	{0x792378750A7CCE36ULL, "lt_0_lowtown_a2_mmap.dds"},
	{0x79239559829ACE90ULL, "46613172.ogg"},
	{0x7923B085C8918541ULL, "vxlt_hangedman_1bq_0_0.vlm"},
	{0x79245802AD3F7F9EULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredidol.ani"},
	{0x7924D0B893F957D9ULL, "vxlt_hangedman_10s_0_0.vlm"},
	{0x7924F8071048B124ULL, "84935452.ogg"},
	{0x7924FCE057E806E1ULL, "vxht_0_hightown_a2_c90_0_0.vlm"},
	{0x7925415CB0CAB5B0ULL, "945741565.ogg"},
	{0x792595A8D180AF58ULL, "vxca_caverns2_7uf_0_0.vlm"},
	{0x792599458D87A4B9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medf_0n.dds"},
	{0x7925C542B23D677FULL, "296545585.ogg"},
	{0x7925D199F6229B86ULL, "vxlt_gallowstemplar_n_46e_0_0.vlm"},
	{0x7925EC5F13128C14ULL, "starline.mao"},
	{0x79260F320C110E18ULL, "lt_0_lowtown_a2_n_c_26_0_lc.mmh"},
	{0x792611F586EFD75CULL, "38079032.ogg"},
	{0x792632122AB30FB3ULL, "vxlt_warehouse_18m_0_0.vlm"},
	{0x79265F71483E5CB7ULL, "t3_hey_bl1.tnt"},
	{0x7926C13B6F8E77D1ULL, "w_encounter3_198_0.msh"},
	{0x7926CF3B6F8E8FECULL, "w_encounter3_198_0.mao"},
	{0x7926DB3B6F8EA38FULL, "w_encounter3_198_0.mmh"},
	{0x79270C644A6D3E7FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scraplight_sa.dds"},
	{0x7927B4554D574DB2ULL, "hlac_ht_keep_n_3732_14x.dds"},
	{0x7927BC084D18404FULL, "lt_0_lowtown_docks_a3_n_39_0.rml"},
	{0x792870764C2828C6ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage3_.gad"},
	{0x7928A1642DA93BDAULL, "6169044_m.fxe"},
	{0x7928BC504F973B40ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_hvy_dmg_lft.gad"},
	{0x7928E900393DE2AFULL, "hlac_lt_gallowstemplar_f_2391_8y.dds"},
	{0x792A0D9303CFA940ULL, "vxlt_gallowstemplar_f_5jr_0_0.vlm"},
	{0x792A32D4D5613FEFULL, "hlac_lt_0_lowtown_docks_-65538_7x.dds"},
	{0x792AC871EA9ADA5FULL, "715911407.ogg"},
	{0x792AFE59E4DEFE28ULL, "209339136.ogg"},
	{0x792B4160FF0B4F06ULL, "32189735.ogg"},
	{0x792B751D267A8DD9ULL, "95153415.ogg"},
	{0x792B9AD3DDB2136EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ef_transition.ani"},
	{0x792C09DE2A38E025ULL, "546152042.ogg"},
	{0x792C49655E243186ULL, "vxw_encounter3_ar_0_0.vlm"},
	{0x792C7F6974F198A3ULL, "llac_ht_chantry_2534_31.dds"},
	{0x792CF56BF896FCB8ULL, "vxca_caverns1_inv_0_0.vlm"},
	{0x792CF7DC489CC4DDULL, "dae000_orsino6085465.cl"},
	{0x792D226D4550CF06ULL, "vxca_caverns1_in8_0_0.vlm"},
	{0x792D306FE53BB85CULL, "1068967740.ogg"},
	{0x792D7230B722FF7DULL, "1061270099.ogg"},
	{0x792DF44774429B39ULL, "vxlt_warehouse_n_u0_0_0.vlm"},
	{0x792E8596193CB2C8ULL, "vxht_0_hightown_a2_c95_0_0.vlm"},
	{0x792F14183B4537F8ULL, "hlac_ht_keep_n_3732_6y.dds"},
	{0x7930787480DFD504ULL, "art\\vfx\\combat\\vfx_glb_ele_c_extreme_c.mmh"},
	{0x7930A3CDBC753643ULL, "vxdr_primevaltomb_6qn_0_0.vlm"},
	{0x79312266C4C45F84ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrda_0d.dds"},
	{0x793183C2E24FA5F4ULL, "6127112_m.fxe"},
	{0x7931B6F39D852527ULL, "339808840.ogg"},
	{0x79321CAE61EE5D71ULL, "ave200_aveline_bar6116598.cl"},
	{0x79323A5B6AB9FA0DULL, "vxdr_0_deeproads_cpd_0_0.vlm"},
	{0x79323FDE609EE5FAULL, "gen00fl_anders6125389.cl"},
	{0x793244253DC6BE4EULL, "zz_dae_debug6049794.cl"},
	{0x7932F0E7B95989BDULL, "and100_anders6173572.cl"},
	{0x7933918B4E1AAC7FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkoc.phy"},
	{0x7933FDC7B87CD77CULL, "59639323.ogg"},
	{0x7934970021CDAF91ULL, "vxlt_undercityencounter_1s9_0_0.vlm"},
	{0x7934D99BCC1DCE47ULL, "mag360_thrask6043918.cl"},
	{0x7934DFA16836992DULL, "lt_0_lowtown_docks_9_0_lc.mmh"},
	{0x79350F157B0DE80DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneeling_faint_loop.ani"},
	{0x79353B58B0B48F8AULL, "hlac_lt_0_lowtown_a2_n_c_131070_5y.dds"},
	{0x79355F1CB1355016ULL, "vxdr_primevaltomb_864_0_0.vlm"},
	{0x7935CA63802EC90DULL, "6060267_m.fxe"},
	{0x7935D15EFB0EDF02ULL, "llac_lt_gallowstemplar_f_7198_2.dds"},
	{0x7935E2CB8C653C6EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\turn_left_180.evt"},
	{0x7936015AFE3B7A59ULL, "vxht_mansion_n_52p_0_0.vlm"},
	{0x79367F6875F67875ULL, "vxht_mansion2b_2p2_0_0.vlm"},
	{0x7936CF7A9E62DCC0ULL, "mag110_samson6018598.cl"},
	{0x793728BBC4339413ULL, "vxdr_0_deeproads_f0g_0_0.vlm"},
	{0x793733EAC7F19A96ULL, "hlac_ca_cavern1_haunted_24424_19z.dds"},
	{0x7937AB284332AAF6ULL, "zz_one_debug6079320.cl"},
	{0x7937D59B35650410ULL, "9977952.ogg"},
	{0x79383ABF5A36ECF1ULL, "follower_banter6170437.cl"},
	{0x79387102AD505318ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredidol.gad"},
	{0x793878C220999D4DULL, "dae000_keep_male_servant_mo.bnk"},
	{0x793892D3DDBC8FBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ef_transition.evt"},
	{0x79389BF1B84BC0A5ULL, "6121151_m.fxe"},
	{0x7938EF1CF8603408ULL, "vxht_brothel_n_4xi_0_0.vlm"},
	{0x7938F0575A33754FULL, "hlac_lt_0_lowtown_docks_a3_-131073_0x.dds"},
	{0x79399054BDFD098CULL, "6150222_m.fxe"},
	{0x793997EE936375E4ULL, "zz_followers_banter6157237.cl"},
	{0x7939BDDB3062BE83ULL, "730380710.ogg"},
	{0x793A133369899FD8ULL, "523435246.ogg"},
	{0x793A2983C2DE9069ULL, "dae000cr_grand_cleric6187952.cl"},
	{0x793A46C4C907182DULL, "6039719_m.fxe"},
	{0x793AA0468191A217ULL, "mag210_arianni6030472.cl"},
	{0x793ACB31C85FCB13ULL, "6046957_m.fxe"},
	{0x793AE42B722CA87EULL, "llac_ht_0_hightown_1_2.dds"},
	{0x793AF6852C528F08ULL, "w_bonepitmining_a2_54_0.rml"},
	{0x793B116C79049FBDULL, "gen00fl_fenris6099838.cl"},
	{0x793B851BF56DB3CFULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic_4.mao"},
	{0x793B89A781CAF13FULL, "vxw_bonepitdestroyed_a3_26a_0_0.vlm"},
	{0x793B9AC2209C0CB5ULL, "dae000_keep_male_servant_mo.cnv"},
	{0x793BA8B6C005CD7BULL, "vxlt_0_lowtown_a2_eru_0_0.vlm"},
	{0x793BD9CB8C699819ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\turn_left_180.gad"},
	{0x793C340BE7497ABEULL, "bw_dis.mao"},
	{0x793C66246638B3B9ULL, "vxlt_warehouse_n_x7_0_0.vlm"},
	{0x793CEC08CEBA58E1ULL, "vxw_bonepit_1vh_0_0.vlm"},
	{0x793CF80DDE6C8C3EULL, "92468977.ogg"},
	{0x793DA6A45221FE21ULL, "ts_beasts_dragons_02.gda"},
	{0x793E0845F1EE2666ULL, "6042127_m.fxe"},
	{0x793E12CD1C78DEA7ULL, "vxca_caverns1_co8_0_0.vlm"},
	{0x793E4D0D38688EC8ULL, "tal_sha_01.mao"},
	{0x793F015129FCE656ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexb.mao"},
	{0x793F0D5129FCFAF5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexb.mmh"},
	{0x793F0F5129FCFE6BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexb.msh"},
	{0x793F59584491AC5EULL, "vxht_mansion2b_50_0_0.vlm"},
	{0x793F73D3DDC279A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ef_transition.gad"},
	{0x793F92220FA1BEADULL, "6104950_m.fxe"},
	{0x793FBB9E2E3CDAA3ULL, "qun120_arishok6051486.cl"},
	{0x79401FAF4604906AULL, "vxw_sundermount_25f_0_0.vlm"},
	{0x794020C680974B39ULL, "1029043824.ogg"},
	{0x7940296502E3829AULL, "t3_shd_rst.tnt"},
	{0x794061B5B27DB83AULL, "763475291.ogg"},
	{0x7940A58A65EBADABULL, "vxlt_warehouse_t4_0_0.vlm"},
	{0x7940B312F1018E80ULL, "cs_qcr000cs_thechampion.wbd"},
	{0x7940B352C4ED9C97ULL, "vxlt_0_lowtown_f04_0_0.vlm"},
	{0x7940C532AB539657ULL, "6207376_m.fxe"},
	{0x794159FE378DA38EULL, "vxlt_0_lowtown_docks_a3_d1n_0_0.vlm"},
	{0x79416033042FEB23ULL, "follower_banter6161284.cl"},
	{0x794186E01BB918EEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxa.msh"},
	{0x79418BEF503DED3FULL, "vxlt_0_lowtown_a2_f5c_0_0.vlm"},
	{0x794194E01BB9303FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxa.mao"},
	{0x7941A0E01BB944DCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxa.mmh"},
	{0x7941A29AA6031716ULL, "6075198_m.fxe"},
	{0x7941B1D78D5F4F48ULL, "art\\vfx\\combat\\vfx_g_spider_leap_act_c.mmh"},
	{0x7941C5F9E81DBA08ULL, "art\\vfx\\creatures\\vfx_c_hd_roar_c.mmh"},
	{0x7941D79C53A6C206ULL, "vxht_0_hightown_n_bzh_0_0.vlm"},
	{0x79425353A3CCE846ULL, "llac_lt_0_lowtown_docks_a3_-131073_5.dds"},
	{0x79427962B56CE058ULL, "vxca_caverns2_7od_0_0.vlm"},
	{0x7942A2B2A71DBAF9ULL, "hlac_lt_0_lowtown_docks_a3_-131074_2y.dds"},
	{0x7942EACB8C6FD467ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\turn_left_180.ani"},
	{0x794315CC0E9F40B3ULL, "vxlt_warehouse_12o_0_0.vlm"},
	{0x794385472513724DULL, "749334647.ogg"},
	{0x7943A51FE29119D7ULL, "vxw_woundedcoast_11p_0_0.vlm"},
	{0x7943B14BC3B8E608ULL, "vxlt_0_lowtown_a2_h2c_0_0.vlm"},
	{0x7943B50546B24FEFULL, "follower_barks6153790.cl"},
	{0x7943C51DC0CB5A92ULL, "6065946_m.fxe"},
	{0x794406D41A85AC59ULL, "6095397_m.fxe"},
	{0x7944198B3B38FC0EULL, "vxlt_darktown_2sg_0_0.vlm"},
	{0x7944928B4E292084ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkoc.mmh"},
	{0x7944968B4E2927D7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkoc.mao"},
	{0x7944A88B4E294626ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkoc.msh"},
	{0x7944ABE0F89AE9FDULL, "vxlt_undercityencounter_13l_0_1.vlm"},
	{0x7944BC485FDB43C0ULL, "6118389_m.fxe"},
	{0x7944C55D5FB69987ULL, "hlac_lt_0_lowtown_docks_a3_131072_1y.dds"},
	{0x7945241840340714ULL, "lt_undercitydungeon_mmi.gff"},
	{0x7945A34F650B5373ULL, "vxht_mansion2b_1cv_0_0.vlm"},
	{0x794662300B26F21EULL, "llac_lt_0_lowtown_docks_a3_65535_5.dds"},
	{0x794674BE900F04C8ULL, "vxw_bonepit_1vm_0_0.vlm"},
	{0x794699ACDD315505ULL, "402596960.ogg"},
	{0x794840B9EC0E6BE1ULL, "vxw_woundedcoast_we_0_0.vlm"},
	{0x7948C6AC48255F90ULL, "613425102.ogg"},
	{0x7949702C39018067ULL, "8143988.ogg"},
	{0x7949C5419A2B29B5ULL, "18534448.ogg"},
	{0x7949FBD86CB838B0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic_3.mao"},
	{0x794A3ECD30709C7CULL, "zz_followers_banter6158580.cl"},
	{0x794A510551F86797ULL, "llac_ht_0_hightown_a2_131070_0.dds"},
	{0x794A5C93C638762CULL, "vxht_keep_n_7a2_0_0.vlm"},
	{0x794ADBFEA9C15B22ULL, "mag291_ntbanter6166384.cl"},
	{0x794B7D222C075D8DULL, "vxht_chantry_2ie_0_0.vlm"},
	{0x794B96BB842E52FFULL, "6027505_m.fxe"},
	{0x794B9C7D501F60E6ULL, "hlac_dr_primevaltomb_659_16y.dds"},
	{0x794BDDD44C29EFE0ULL, "323503934.ogg"},
	{0x794C258AF9D9054BULL, "hlac_lt_gallowsprison_n_-1_3y.dds"},
	{0x794C914164B6CB50ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_fire02.dds"},
	{0x794D009789D8C677ULL, "hm_shw_secret_thug.mor"},
	{0x794D13B88EEE12F5ULL, "ca_caverns1_24424_ll.mmh"},
	{0x794D8761758426B1ULL, "hla_ht_0_hightown_a3_65538_0y.dds"},
	{0x794DB36DE4ACD6AFULL, "vxlt_hangedman_n_2gl_0_0.vlm"},
	{0x794DD6A013153A8EULL, "hlac_ht_keep_4519_7y.dds"},
	{0x794E43B99D6CCD4AULL, "520660269.ogg"},
	{0x794EC5F80EA70B52ULL, "679131267.ogg"},
	{0x794EFF145EA73C76ULL, "hlac_ht_0_hightown_a3_-65534_0x.dds"},
	{0x794F176757E1C62BULL, "vxlt_warehouse_y7_0_0.vlm"},
	{0x794F1E154EA62B0DULL, "vxdr_0_deeproads_cjb_0_0.vlm"},
	{0x794F2212781BDC8AULL, "vxw_bonepit_25m_0_0.vlm"},
	{0x794F252FB6E3B517ULL, "vxlt_0_lowtown_f57_0_0.vlm"},
	{0x794FAF21128AEBC4ULL, "llac_lt_0_lowtown_docks_65532_1.dds"},
	{0x795012C3B89E1857ULL, "zz_dre_debug6025554.cl"},
	{0x795020552C14D111ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottlecin_n.dds"},
	{0x79509CFFE3918B46ULL, "ico_shield_2.dds"},
	{0x7950C21CE5C979CFULL, "crt_spider.bnk"},
	{0x7951325AF795ECBDULL, "vxlt_0_lowtown_docks_a3_9do_0_0.vlm"},
	{0x7951BD32B2187F68ULL, "vxlt_undercityencounter_1ox_0_0.vlm"},
	{0x7951E6B45318DAB4ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basket.phy"},
	{0x79523513C413BDADULL, "vxlt_0_lowtown_a2_eha_0_0.vlm"},
	{0x79525C1764CD9D16ULL, "vxht_keep_5ws_0_0.vlm"},
	{0x7952E3ECE4A19924ULL, "vxca_caverns1_hli_0_0.vlm"},
	{0x7953C007D87BEECAULL, "vxlt_0_lowtown_a2_n_c_hlj_0_0.vlm"},
	{0x7953C93A7F0B5252ULL, "gen00fl_aveline6114838.cl"},
	{0x795433E7608E0F29ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appc_0s.dds"},
	{0x79546C291DA44741ULL, "zz_isa_debug6175466.cl"},
	{0x7954B6CDE6CC5118ULL, "727580229.ogg"},
	{0x7954BC3F8FCAFB37ULL, "vxlt_darktown_5t0_0_0.vlm"},
	{0x795503935F0DB212ULL, "889146363.ogg"},
	{0x79550F7DEA2596DBULL, "dae000_bodhan6129136.cl"},
	{0x79551BAE0FB60503ULL, "vxw_sundermount_31c_0_1.vlm"},
	{0x79553599423FAA63ULL, "6179054_f.fxe"},
	{0x7955A1C3A95690C1ULL, "vxw_encounter2_camp_5z_0_0.vlm"},
	{0x79578EDBE41CDFB3ULL, "910850303.ogg"},
	{0x7957C20C1C318722ULL, "follower_banter6143250.cl"},
	{0x79580B01B73CCBDFULL, "6150247_m.fxe"},
	{0x795810218FE751D6ULL, "vxlt_0_lowtown_n_gf5_0_0.vlm"},
	{0x795812BBF043252DULL, "vxdr_primevalentrance_30j_0_0.vlm"},
	{0x79583A90873E4A6FULL, "vxht_keep_4bl_0_0.vlm"},
	{0x79586D06E60FA833ULL, "6042051_f.fxe"},
	{0x79587B7481012CFBULL, "art\\vfx\\combat\\vfx_glb_ele_c_extreme_c.anb"},
	{0x7958B360337DD9A6ULL, "plt_dre120pt_main.nss"},
	{0x7958F3C06ED0279EULL, "hlac_ht_keep_n_3732_1z.dds"},
	{0x79591490CBBF471FULL, "vxca_caverns2prc_949_0_0.vlm"},
	{0x79591D43CA3A83ADULL, "6176722_m.fxe"},
	{0x79596F53DFA4C6D6ULL, "l_dre200ar_dougal_encounter.wbd"},
	{0x7959EEFF1DFD760DULL, "vxht_chantry_2bd_0_0.vlm"},
	{0x795A7E33EB0CC71EULL, "gen00fl_aveline6122817.cl"},
	{0x795B1291FBAAD8B4ULL, "dae000_gamlen6144343.cl"},
	{0x795B42CAB5AD1A41ULL, "845556017.ogg"},
	{0x795B784B52E302E6ULL, "qcr220_grey_wardens6135903.cl"},
	{0x795BC3B45320F3B1ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basket.msh"},
	{0x795BC3F1D7A78564ULL, "46683032.ogg"},
	{0x795BCD7F5FB19771ULL, "rdr110dg_guard_warehouse6174110.cl"},
	{0x795BD1B453210BCCULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basket.mao"},
	{0x795BDDB45321206FULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basket.mmh"},
	{0x795BFA54CF8A7EEFULL, "658080947.ogg"},
	{0x795CD7AF23C64E39ULL, "6101193_m.fxe"},
	{0x795CFBFB0B1A0111ULL, "lowtown_argument6169441.cl"},
	{0x795D0C2606C0C874ULL, "hlac_lt_0_lowtown_docks_-2_5x.dds"},
	{0x795D3ECC227BB8F3ULL, "qun120_arishok6051749.cl"},
	{0x795D4CEA54649878ULL, "6051613_m.fxe"},
	{0x795D51CD3DC86178ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredlong_sa.dds"},
	{0x795D7E26AFC35835ULL, "vxlt_0_lowtown_hac_0_0.vlm"},
	{0x795D8FF2409C438DULL, "vxdr_0_deeproads_cga_0_0.vlm"},
	{0x795D93EF6A11F50AULL, "vxw_bonepit_20n_0_0.vlm"},
	{0x795D9758288DFD42ULL, "vxw_sundermount_30d_0_0.vlm"},
	{0x795E15BCD3F7ACB2ULL, "llac_w_bonepitdestroyed_a3_262147_0.dds"},
	{0x795E59BDE57802A3ULL, "llac_lt_0_lowtown_a2_65535_6.dds"},
	{0x795E8AE180C12118ULL, "llac_w_sundermount_-327679_0.dds"},
	{0x795EEA911B714ABCULL, "hlac_lt_0_lowtown_n_-3_5y.dds"},
	{0x796003E54E8133A4ULL, "391932375.ogg"},
	{0x796071EF03E21B9CULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.dura.ani"},
	{0x7960AEE94B3F609DULL, "129997718.ogg"},
	{0x7960B3F27D5778DCULL, "fx_flamefb02.dds"},
	{0x7960C0CB323274FAULL, "551367065.ogg"},
	{0x7960C988C0598DC1ULL, "art\\vfx\\combat\\vfx_m_blood_change_c.anb"},
	{0x7960CDF456C3B596ULL, "vxht_keep_5zt_0_0.vlm"},
	{0x7961B6DF141795DBULL, "hlac_ht_playershouse_n_855_35z.dds"},
	{0x796280567A2A1E12ULL, "vxht_0_hightown_a3_b8t_0_0.vlm"},
	{0x7962870948F78473ULL, "vxht_mansion2b_1yx_0_0.vlm"},
	{0x79628B896BD15DC3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_flat.msh"},
	{0x796299896BD1751DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_flat.mmh"},
	{0x79629D896BD17C2EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_flat.mao"},
	{0x7962C78CFAA49720ULL, "195318487.ogg"},
	{0x796312FF829E4648ULL, "515579167.ogg"},
	{0x79636A729AB70DA2ULL, "6101396_m.fxe"},
	{0x796434F0BF4187BFULL, "llac_lt_0_lowtown_a2_n_-65535_2.dds"},
	{0x79648BF7C64F18FEULL, "6170356_m.fxe"},
	{0x7964ED312E6625ADULL, "gen00fl_carver6108943.cl"},
	{0x7964F495A2A8EAA0ULL, "narrative_flemeth6002811.cl"},
	{0x7965407EAA647E2FULL, "dae000_party_night_comment6201600.cl"},
	{0x79658A0F03DCC415ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beambrace.msh"},
	{0x79658C0F03DCC78BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beambrace.mmh"},
	{0x7965980F03DCDC28ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beambrace.mao"},
	{0x79667FE01BD828A7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxa.phy"},
	{0x7966EA69FBA27A4AULL, "6115555_m.fxe"},
	{0x7966F9F6DD71FF8BULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.mmh"},
	{0x79677EEF03E850DEULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.dura.gad"},
	{0x7967866DBDB55F9FULL, "vxca_caverns2prc_938_0_0.vlm"},
	{0x7967905F00C46D5EULL, "qaphys.arl"},
	{0x7967AE8DFDA2D00AULL, "50898918.ogg"},
	{0x79683464ABA123F1ULL, "6101247_m.fxe"},
	{0x79683AB63E8E0B02ULL, "w_sundermount_15_0.rml"},
	{0x79684E8E142F212AULL, "6169246_m.fxe"},
	{0x79686A9395BE4DACULL, "art\\characters\\creatures\\mubari\\animation\\attack2.ani"},
	{0x7968838E26745C71ULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.anb"},
	{0x7968A4EB28C19684ULL, "isa321cr_sabina.utc"},
	{0x7968D35178A8ADF0ULL, "6164496_m.fxe"},
	{0x7968EE25E92598A9ULL, "llac_ca_caverns1_24426_6.dds"},
	{0x7969089D248890E0ULL, "lt_0_lowtown_a2_n_c_36_0.anb"},
	{0x7969185113D723EAULL, "6001183_f.fxe"},
	{0x7969A80E0F1A2F4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_loop.ani"},
	{0x7969DF8D8F02A0C5ULL, "6054017_m.fxe"},
	{0x7969E4250874F008ULL, "6170057_m.fxe"},
	{0x7969F90915EF2347ULL, "ts_kirk_darktown_01.gda"},
	{0x796A549D8B5EAAFAULL, "chu_02_f.mao"},
	{0x796B872442743354ULL, "6144461_m.fxe"},
	{0x796C1A0C09E74AFCULL, "6097329_m.fxe"},
	{0x796CC860338EA6E1ULL, "plt_dre120pt_main.plo"},
	{0x796D006D91D9B6BCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ingots_sa.dds"},
	{0x796DC69BFB3B4381ULL, "11832138.ogg"},
	{0x796DD1DF03BA62BFULL, "ave121_jeven6105686.cl"},
	{0x796E46E030F50EB7ULL, "vxlt_gallowscourtyard_n_76m_0_0.vlm"},
	{0x796E4977B449790FULL, "follower_banter6160984.cl"},
	{0x796E59EF03EE30FDULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.dura.evt"},
	{0x796EA9E99F9118B3ULL, "gen00fl_fenris6116579.cl"},
	{0x796EAA88C065963AULL, "art\\vfx\\combat\\vfx_m_blood_change_c.mmh"},
	{0x796EC38863E43C4CULL, "6103306_f.fxe"},
	{0x796F6BD70F746476ULL, "darktown_male_fer_urch_sta6174967.cl"},
	{0x796F779395C482EEULL, "art\\characters\\creatures\\mubari\\animation\\attack2.gad"},
	{0x796F796C2766E7D7ULL, "gen_im_arm_cht_nobc_t4.uti"},
	{0x796FC113508B2882ULL, "rog_stealth.dds"},
	{0x796FDC44ADB17372ULL, "vxlt_darktown_5a2_0_0.vlm"},
	{0x796FEB882AF792A2ULL, "vxht_mansion_n_3v5_0_0.vlm"},
	{0x796FF0A11CB33D5AULL, "lt_0_lowtown_docks_a3_n_21_0_ll.mmh"},
	{0x797029E58785A81CULL, "vxca_caverns2prc_7qk_0_0.vlm"},
	{0x7970B50E0F20638EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_loop.gad"},
	{0x7970C8C5EC3E0ABCULL, "dae000_tranquil_16165104.cl"},
	{0x797120F36BF7C16EULL, "1010062750.ogg"},
	{0x79716CB0BA6E1A8FULL, "art\\characters\\creatures\\corpse\\textures\\cn_bdy_cor_0s.dds"},
	{0x797171578B2E05DFULL, "mcr370_assignments6177319.cl"},
	{0x79719DFEC7613FF7ULL, "6039246_f.fxe"},
	{0x79720D40FACB2749ULL, "gen00fl_fenris6105601.cl"},
	{0x79723D8D0B20319FULL, "vxht_brothel_n_5cz_0_0.vlm"},
	{0x79727991480D5258ULL, "318487250.ogg"},
	{0x79731B2CB8008013ULL, "zz_dre_debug6075959.cl"},
	{0x797324672FFCDD3DULL, "war_cleansearea.dds"},
	{0x7973B6228F309ED7ULL, "915209922.ogg"},
	{0x7973D208F5441007ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_loop.gad"},
	{0x79746ECC62E1AA85ULL, "6041451_m.fxe"},
	{0x7974B8CF2A6CAB1BULL, "art\\vfx\\creatures\\vfx_c_drag_bfft_p.mmh"},
	{0x7974DAF6DD7E08F0ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.anb"},
	{0x7974FCA4D4AD5B5CULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fr.ani"},
	{0x7975702A87DE2F7BULL, "vxlt_0_lowtown_a2_efi_0_0.vlm"},
	{0x79757F77D00DD12AULL, "283912304.ogg"},
	{0x7975EC711750E6E1ULL, "645952536.ogg"},
	{0x797641F38CB22132ULL, "250461886.ogg"},
	{0x7976729395CA98CDULL, "art\\characters\\creatures\\mubari\\animation\\attack2.evt"},
	{0x7976A48E2680D1EAULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.mmh"},
	{0x7976AE914C5DFDCDULL, "6053896_m.fxe"},
	{0x79770305E0B5AA6AULL, "6083159_m.fxe"},
	{0x7977733DF39872DFULL, "face_rendertoshadowmap_p0.vsh.fxo.dx11.assembly"},
	{0x79779A0ADC37FDB4ULL, "175285592.ogg"},
	{0x7977A846E089439CULL, "follower_banter6172924.cl"},
	{0x7977BA6137E2C568ULL, "375676310.ogg"},
	{0x7978853B0DC1A9A2ULL, "193486395.ogg"},
	{0x7978B1E5C1C3C948ULL, "lt_0_lowtown_docks_n_201_0.msh"},
	{0x7978B40DFADC8C07ULL, "vxlt_undercityserialkiller_1iq_0_0.vlm"},
	{0x7978BEC1FE5E2157ULL, "6102415_m.fxe"},
	{0x7978C3E5C1C3E7D9ULL, "lt_0_lowtown_docks_n_201_0.mao"},
	{0x7978CFE5C1C3FC72ULL, "lt_0_lowtown_docks_n_201_0.mmh"},
	{0x7978E3F93F7863ABULL, "6082158_f.fxe"},
	{0x797905B440453805ULL, "llac_ht_brothel_n_3364_0.dds"},
	{0x7979124B0887FF36ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback.evt"},
	{0x797924B63E9C5957ULL, "w_sundermount_15_0.msh"},
	{0x797936B63E9C77BAULL, "w_sundermount_15_0.mao"},
	{0x79793AB63E9C7EC9ULL, "w_sundermount_15_0.mmh"},
	{0x79793EBAD41EAB8FULL, "352138416.ogg"},
	{0x797987C2F28BF8A2ULL, "140783601.ogg"},
	{0x7979F901346A4BD4ULL, "6141761_m.fxe"},
	{0x797ACE6FC4CBC737ULL, "425121954.ogg"},
	{0x797AE0EAC8F67446ULL, "vxht_mansion_n_5az_0_0.vlm"},
	{0x797B09A4D4B1DC9EULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fr.gad"},
	{0x797B15D258A4F9EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_stand_up.gad"},
	{0x797B6F2D9BA7E022ULL, "800633712.ogg"},
	{0x797B7E8F8F669243ULL, "hlac_lt_0_lowtown_docks_a3_-4_7z.dds"},
	{0x797C4AD9B28CE32AULL, "llac_ht_0_hightown_n_-65537_1.dds"},
	{0x797C6076FBC819F2ULL, "w_bonepit_47_0.rml"},
	{0x797C7C4772FEB47FULL, "932646984.ogg"},
	{0x797E9BC2797BC09CULL, "vxca_caverns2prc_7th_0_0.vlm"},
	{0x797EB5D3E7C340ADULL, "vxlt_0_lowtown_er7_0_0.vlm"},
	{0x797F6AC32CE3B573ULL, "vxht_mansion2b_1wz_0_0.vlm"},
	{0x797F97E3110D1BD4ULL, "vxlt_0_lowtown_docks_a3_n_btj_0_0.vlm"},
	{0x797FACF87238680AULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_01skelpile.mao"},
	{0x797FADADB3741150ULL, "865871663.ogg"},
	{0x797FB0F872386F59ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_01skelpile.mmh"},
	{0x797FB17DDC667A21ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_large.phy"},
	{0x797FBAF872388047ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_01skelpile.msh"},
	{0x797FC596B594AB4AULL, "70104921.ogg"},
	{0x7980094B088E0EC1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback.gad"},
	{0x7980755F4C66C15EULL, "vxw_bonepitdestroyed_a3_15b_0_0.vlm"},
	{0x7980926D478C7583ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose7.anb"},
	{0x7980CF7E98F60753ULL, "art\\levels\\races\\design\\animation\\gasbarrel.active.evt"},
	{0x79812E3E4634BD37ULL, "572834446.ogg"},
	{0x7981426DEF88A4CEULL, "art\\vfx\\models\\vfx_trap_bolt.mmh"},
	{0x79814E6DEF88B965ULL, "art\\vfx\\models\\vfx_trap_bolt.mao"},
	{0x79815C6DEF88D0BCULL, "art\\vfx\\models\\vfx_trap_bolt.msh"},
	{0x7981694B96110D36ULL, "mag121_idunna6084249.cl"},
	{0x798194BC1E14184CULL, "923417435.ogg"},
	{0x7981CEC2BCB2F980ULL, "6100057_f.fxe"},
	{0x7981D9CF2A776D80ULL, "art\\vfx\\creatures\\vfx_c_drag_bfft_p.anb"},
	{0x7981E4A4D4B7BCBDULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fr.evt"},
	{0x7981F8D258AAE681ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_stand_up.evt"},
	{0x7982048EF92BB034ULL, "art\\cinematics\\additional animation\\cs_sail_cycle.gad"},
	{0x7982100EA886767BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_spare.ani"},
	{0x798243C494412571ULL, "llac_lt_0_lowtown_docks_a3_131072_0.dds"},
	{0x798278D0F74359ECULL, "vxca_caverns2_5x7_0_0.vlm"},
	{0x7982899B11BFC56BULL, "784030542.ogg"},
	{0x79829D51FA7AC2B2ULL, "6149495_m.fxe"},
	{0x7982CE2909E653BEULL, "vxht_mansion_n_4x4_0_0.vlm"},
	{0x7982D5124FAF9F1DULL, "hlac_ht_0_hightown_a3_65533_1x.dds"},
	{0x7982E60661273836ULL, "vxlt_darktown_4a3_0_0.vlm"},
	{0x7982EDE2DA143EDDULL, "mag360_alain6044636.cl"},
	{0x79834D5750005544ULL, "hlac_lt_gallowscourtyard_-131071_2z.dds"},
	{0x7984D848111CCA37ULL, "zz_isa_debug6133202.cl"},
	{0x79850F6E3930FD58ULL, "hlac_w_sundermount_65541_0z.dds"},
	{0x798525598888D361ULL, "vxw_bonepit_1oc_0_0.vlm"},
	{0x798533BAA2781960ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_1z.dds"},
	{0x79859553DFCA4244ULL, "l_dre200ar_dougal_encounter.bnk"},
	{0x7985D841775D9DEAULL, "6141778_m.fxe"},
	{0x7985F34A9F3E98BFULL, "6014275_m.fxe"},
	{0x798642DC10022AFBULL, "vxht_keep_n_6gn_0_0.vlm"},
	{0x798664C5D57E4D79ULL, "355693110.ogg"},
	{0x7986E1E161697014ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrda_0t.dds"},
	{0x79870A4B08942F6FULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkback.ani"},
	{0x798725EAECD2A487ULL, "vxlt_undercityserialkiller_1np_0_0.vlm"},
	{0x798767AF774197E6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_ands_0n.dds"},
	{0x7987C87E98FC1994ULL, "art\\levels\\races\\design\\animation\\gasbarrel.active.gad"},
	{0x7988138EF9303512ULL, "art\\cinematics\\additional animation\\cs_sail_cycle.ani"},
	{0x7988244722F45D7BULL, "6076262_m.fxe"},
	{0x798831C8868309F5ULL, "art\\vfx\\combat\\vfx_c_buff_act_c.mmh"},
	{0x79888FBD9698963BULL, "574247686.ogg"},
	{0x7988BF08F5564C4DULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_loop.ani"},
	{0x798923E8882AC6C4ULL, "w_bonepit_4_0_lc.mmh"},
	{0x79892D53D9856587ULL, "943452976.ogg"},
	{0x79894EF3B7873D06ULL, "zz_dae_debug6029169.cl"},
	{0x79898722CC196465ULL, "hlac_lt_0_lowtown_65534_5z.dds"},
	{0x79898EFFEC94DB03ULL, "vxht_brothel_n_6d8_0_0.vlm"},
	{0x79899CBEF6953CFEULL, "seb321_leliana6197954.cl"},
	{0x7989C004D0CA98B1ULL, "304435518.ogg"},
	{0x798A2527A9493782ULL, "gen00fl_merrill6130082.cl"},
	{0x798A67E99FA7D1E4ULL, "gen00fl_anders6125989.cl"},
	{0x798AAE0B2E13CE81ULL, "dae000_gamlen6102409.cl"},
	{0x798ABF136248ACD5ULL, "6144939_f.fxe"},
	{0x798B0C35BC7B87FAULL, "6139302_m.fxe"},
	{0x798B7AE70EACBB11ULL, "mcr310_sibling6076733.cl"},
	{0x798B923653D6E546ULL, "dre000_bodahn6028181.cl"},
	{0x798BA77996E5F676ULL, "mag320_orsino6071139.cl"},
	{0x798C11A51C3B9112ULL, "vxca_caverns1_mines_dnt_0_0.vlm"},
	{0x798C4DEAE883A026ULL, "vxht_brothel_n_5k7_0_0.vlm"},
	{0x798C9E27DB2EC0A5ULL, "6152805_m.fxe"},
	{0x798CD4C3A1058E6AULL, "vxca_caverns2_73s_0_1.vlm"},
	{0x798D6570B5E780EEULL, "mrl210_merrill6116332.cl"},
	{0x798E6C5106E19090ULL, "llac_ht_0_hightown_65538_0.dds"},
	{0x798EAE0F49DD7F48ULL, "vxw_bonepit_1of_0_0.vlm"},
	{0x798ED77E99025272ULL, "art\\levels\\races\\design\\animation\\gasbarrel.active.ani"},
	{0x798EF78277EDE4D6ULL, "lgt000ip_store_epic1_2.utp"},
	{0x798EF8D529DA4F43ULL, "llac_lt_gallowstemplar_2383_7.dds"},
	{0x798FC990C7CE6D1FULL, "363892543.ogg"},
	{0x798FF0D258B716D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\zombie_mom_stand_up.ani"},
	{0x7990180EA892C172ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_spare.evt"},
	{0x79901DC9AE4E32F9ULL, "399608837.ogg"},
	{0x79909C1A47499A94ULL, "tel_06.dds"},
	{0x7990CF98D4D29CF5ULL, "290797175.ogg"},
	{0x799165AB28799D2BULL, "621477714.ogg"},
	{0x7991D307EA72D79AULL, "qcr000_arishok6059349.cl"},
	{0x7992854CCEBC5F99ULL, "gen00fl_anders6121216.cl"},
	{0x7992A1DC0974B866ULL, "333542352.ogg"},
	{0x799300171110C0E9ULL, "vxlt_hangedman_n_9_0_0.vlm"},
	{0x799324B030489B9BULL, "6107056_m.fxe"},
	{0x79932E08FF4A83A1ULL, "vxlt_0_lowtown_docks_a3_n_b9d_0_0.vlm"},
	{0x79932E2A4342AAEEULL, "6019084_m.fxe"},
	{0x799388BE0DBE59A9ULL, "6122802_m.fxe"},
	{0x7993B5016099BDF7ULL, "vxlt_0_lowtown_a2_n_c_i0v_0_0.vlm"},
	{0x79948B0A54DCCFD5ULL, "plt_zz_fampt_debug.nss"},
	{0x79949B1A4693DAD8ULL, "6113642_f.fxe"},
	{0x7994B4B901F8437BULL, "vxht_keep_n_6ba_0_0.vlm"},
	{0x7994F2A226BB1C7CULL, "lgt250_way_dalish_conflict6147131.cl"},
	{0x79952EC8868D8E82ULL, "art\\vfx\\combat\\vfx_c_buff_act_c.anb"},
	{0x7996370EA8976105ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_spare.gad"},
	{0x799643330CC9DC8CULL, "hlac_w_woundedcoast_131077_0z.dds"},
	{0x79966BF104111A83ULL, "524674342.ogg"},
	{0x7997832DD6766EE1ULL, "zz_mrl_debug6086515.cl"},
	{0x7997B84B0EDED264ULL, "hlac_w_encounter2_0_0z.dds"},
	{0x7997BB12A870CEBFULL, "hlac_ht_keep_3739_1y.dds"},
	{0x79983760D6EFCD26ULL, "vxht_mansion2_2r2_0_1.vlm"},
	{0x7998A713704480B0ULL, "dae000_gamlen6048476.cl"},
	{0x7998AEFB62EA82A1ULL, "548456999.ogg"},
	{0x79992B23CFE30BDAULL, "hlac_lt_0_lowtown_n_-131072_3z.dds"},
	{0x799977FF88B9ECB9ULL, "6116569_f.fxe"},
	{0x799A21CFEBC1C920ULL, "hlac_lt_0_lowtown_-131072_2z.dds"},
	{0x799AAF53B335A144ULL, "37533449.ogg"},
	{0x799ABE7B29E5E81BULL, "6040663_m.fxe"},
	{0x799B66069C8790B9ULL, "lgt200_eau_tome6180568.cl"},
	{0x799BA9FB888D0C9EULL, "vxlt_0_lowtown_n_ekj_0_2.vlm"},
	{0x799BAB23A1BFCC0EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic.phy"},
	{0x799BE8A49946F83FULL, "814474652.ogg"},
	{0x799BF958924A694AULL, "vxlt_0_lowtown_a2_n_c_hsg_0_0.vlm"},
	{0x799C3FB148705B44ULL, "267850147.ogg"},
	{0x799C4D2790775865ULL, "gen_im_npc_dwf_arm_hlm_lgtb.uti"},
	{0x799CA2F71C392F4CULL, "dae000cr_paivel.utc"},
	{0x799D0DA7DB118787ULL, "dre000_dougal6129782.cl"},
	{0x799D3E1D74375F91ULL, "342362197.ogg"},
	{0x799D5DFC535B5022ULL, "gen00fl_aveline6114287.cl"},
	{0x799E9C738AE39408ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic_1.mao"},
	{0x799ED89A2CDC60E3ULL, "w_sundermount_72_0.mao"},
	{0x799EDC9A2CDC6730ULL, "w_sundermount_72_0.mmh"},
	{0x799EEA9A2CDC7F4AULL, "w_sundermount_72_0.msh"},
	{0x799F506710602F6FULL, "gen00fl_bethany6112229.cl"},
	{0x799F716B9445A26BULL, "6105996_f.fxe"},
	{0x799FA78CC1816FC3ULL, "6135695_m.fxe"},
	{0x799FAB4BA11B2338ULL, "vxw_encounter3_94_0_0.vlm"},
	{0x799FC9C045136936ULL, "vxlt_darktown_4k5_0_0.vlm"},
	{0x799FD84CEDD4F586ULL, "vxw_encounter3_9z_0_0.vlm"},
	{0x799FEA8FD0B21523ULL, "vxw_encounter2_camp_j1_0_0.vlm"},
	{0x79A026BFBF18004DULL, "w_bonepitmining_a2_216_0.mmh"},
	{0x79A02ABFBF18069EULL, "w_bonepitmining_a2_216_0.mao"},
	{0x79A038BFBF181ED3ULL, "w_bonepitmining_a2_216_0.msh"},
	{0x79A08FF36AEB1D44ULL, "gen00fl_merrill6106320.cl"},
	{0x79A0B49F7F93144FULL, "hlac_lt_darktown_-1_20y.dds"},
	{0x79A0DEF41ADD2A9CULL, "lt_0_lowtown_a2_n_43_0_ll.mmh"},
	{0x79A15BD1B56C1323ULL, "gen00fl_anders6128316.cl"},
	{0x79A288A792D8301BULL, "6172508_m.fxe"},
	{0x79A29A0754F8DA8CULL, "vxlt_undercitydungeon_12f_0_0.vlm"},
	{0x79A2D5062C628C24ULL, "172093965.ogg"},
	{0x79A3E90A5C8C824AULL, "l_qcr220ar_hightown_nt.wbd"},
	{0x79A43FB3249EEF27ULL, "vxw_encounter2_p7_0_0.vlm"},
	{0x79A4595B56AF8637ULL, "zz_vis_debug.cnv"},
	{0x79A4975777144CFAULL, "6165130_m.fxe"},
	{0x79A4E64D58F0F5A4ULL, "one280_merin6105459.cl"},
	{0x79A553D364ED22C5ULL, "hlac_lt_gallowstemplar_2391_4x.dds"},
	{0x79A57B88B8C5AA7CULL, "follower_barks6153654.cl"},
	{0x79A5B8C8634B71BAULL, "6151393_m.fxe"},
	{0x79A6847CB5B18132ULL, "vxht_0_hightown_a2_c17_0_0.vlm"},
	{0x79A6B61E94A419B7ULL, "mag160_feynriel6033381.cl"},
	{0x79A769AD22F01023ULL, "mag161ip_at_to_coast.utp"},
	{0x79A7DD88572663D5ULL, "dae000_flemeth6013427.cl"},
	{0x79A8058B0BCAFBB3ULL, "6045606_m.fxe"},
	{0x79A833DF79DBE0A3ULL, "follower_banter6172842.cl"},
	{0x79A85529F4C15087ULL, "gen00fl_fenris6126910.cl"},
	{0x79A8DBCB005210C1ULL, "t3_hlk_b02.tnt"},
	{0x79A8E79D24BE3C08ULL, "lt_0_lowtown_a2_n_c_36_0.rml"},
	{0x79A93BB8D957D584ULL, "hlac_ca_cavern1_haunted_24425_5x.dds"},
	{0x79A99282DB6B9137ULL, "vxca_caverns1_mines_jkx_0_0.vlm"},
	{0x79A9A759E86876EAULL, "238152979.ogg"},
	{0x79A9B6D132EBB793ULL, "qun100_mage_party_comment6019095.cl"},
	{0x79A9C965DF7DA58EULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man2_.ani"},
	{0x79A9D08FB34EF318ULL, "vxlt_shop_f6_0_0.vlm"},
	{0x79A9EBFC919F5BC0ULL, "vxlt_gallowstemplar_n_5qw_0_0.vlm"},
	{0x79A9EF0AA27CC685ULL, "follower_banter6148746.cl"},
	{0x79AA5382338FBC03ULL, "dre000_hireling16038595.cl"},
	{0x79AAC215BF523048ULL, "gen00fl_anders6125893.cl"},
	{0x79ABF052C5D1DC48ULL, "666192877.ogg"},
	{0x79ACEAC66ED9834DULL, "254214241.ogg"},
	{0x79ACF82895629559ULL, "one300_nuncio6083400.cl"},
	{0x79ADAF6389D15E98ULL, "544236433.ogg"},
	{0x79AE031A8DDCC024ULL, "vxca_cavern1_haunted_i1z_0_0.vlm"},
	{0x79AE445B56B7B764ULL, "zz_vis_debug.ncs"},
	{0x79AEB58459D2A936ULL, "6110981_m.fxe"},
	{0x79AEDF14F1CEA71CULL, "96979376.ogg"},
	{0x79AEF6BBAEC0A06EULL, "gen00fl_aveline6053289.cl"},
	{0x79AF6659D0E92A18ULL, "hlac_w_bonepit_196609_0y.dds"},
	{0x79AF6DF7213C3254ULL, "gen00fl_merrill6106892.cl"},
	{0x79AFC9846D6202EBULL, "6119716_m.fxe"},
	{0x79AFF27F1F5965AAULL, "hairalpha_hair_p1.vsh.fxo.dx11"},
	{0x79B01F8374C4065CULL, "659419572.ogg"},
	{0x79B079A4255FCB8BULL, "172655669.ogg"},
	{0x79B08AF90556B7F3ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.dura.evt"},
	{0x79B1EFE6F0EB28B5ULL, "6027468_m.fxe"},
	{0x79B2AAB4F747F22EULL, "vxca_caverns1_mines_e87_0_0.vlm"},
	{0x79B2E201471FE9E8ULL, "el_lr_outside_day.mmh"},
	{0x79B2F1B4E835B330ULL, "6034230_m.fxe"},
	{0x79B465DB05A4656FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appp_0t.dds"},
	{0x79B478D1FA4895DEULL, "vxca_caverns2prc_9px_0_0.vlm"},
	{0x79B5321FD026B776ULL, "vxlt_0_lowtown_a2_n_c_cif_0_0.vlm"},
	{0x79B5406114FCE093ULL, "llac_ht_playershouse_n_855_2.dds"},
	{0x79B608A9E8007C59ULL, "658502496.ogg"},
	{0x79B68A0A54F9B436ULL, "plt_zz_fampt_debug.plo"},
	{0x79B6DA26D8711358ULL, "mag170_thrask6054103.cl"},
	{0x79B720135028BEF7ULL, "6043063_f.fxe"},
	{0x79B775746FFF0429ULL, "vxw_sundermount_2w3_0_0.vlm"},
	{0x79B783F9055CCA34ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.dura.gad"},
	{0x79B7841BB0A10C5FULL, "vxlt_0_lowtown_a3_n_c_gc4_0_0.vlm"},
	{0x79B8B5D1BA223B09ULL, "vxlt_0_lowtown_a3_n_c_hjf_0_0.vlm"},
	{0x79B91E264D395F4BULL, "6096533_f.fxe"},
	{0x79B97D66D3C1A826ULL, "hlac_ca_caverns1_mines_24424_3y.dds"},
	{0x79B98F31D1C5CC2DULL, "vxca_caverns1_mines_k2q_0_1.vlm"},
	{0x79B99073A87D9229ULL, "6144958_m.fxe"},
	{0x79BA120E3EB94678ULL, "317108167.ogg"},
	{0x79BB85518F3D9309ULL, "vxw_encounter2_r3_0_0.vlm"},
	{0x79BB8C1FA48F5360ULL, "6121640_m.fxe"},
	{0x79BC8108CAFC6A5FULL, "narrative_varric6172467.cl"},
	{0x79BD0A7715ED5219ULL, "vxlt_0_lowtown_a2_n_c_etx_0_0.vlm"},
	{0x79BD92F905614F12ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.dura.ani"},
	{0x79BDA265DF8E0CC8ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man2_.gad"},
	{0x79BDB9823EDA708BULL, "482686773.ogg"},
	{0x79BDD3BF4E231B12ULL, "venge_fury.dds"},
	{0x79BE28189ED4898FULL, "6064522_m.fxe"},
	{0x79BE52F30EDAF32FULL, "lt_0_lowtown_n_42_0.anb"},
	{0x79BE688AD28D18C5ULL, "361957046.ogg"},
	{0x79BECDFDB7276106ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_2_sylvesta.ani"},
	{0x79BF7DC138E50B8CULL, "vxlt_undercitydungeon_18d_0_0.vlm"},
	{0x79BF85E6407039EBULL, "cth_nb2.tnt"},
	{0x79C00C4F23597B23ULL, "6071211_f.fxe"},
	{0x79C0338833EC7500ULL, "808217936.ogg"},
	{0x79C04C8F9D2AE588ULL, "6077415_m.fxe"},
	{0x79C0CA99272F7BC0ULL, "hlac_lt_0_lowtown_a2_-196609_3z.dds"},
	{0x79C0F44498C66773ULL, "mcr371_hostile_merrill6082160.cl"},
	{0x79C17F797F5DA526ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_dbl_c.anb"},
	{0x79C1ACF401495728ULL, "zz_followers_banter6159657.cl"},
	{0x79C1C3810BBFACF4ULL, "6072294_m.fxe"},
	{0x79C1EAA87CF7149AULL, "hlac_lt_gallowscourtyard_n_65536_0x.dds"},
	{0x79C28405D7A3B5F9ULL, "lgt260_eau_tmp_flowergirl6165093.cl"},
	{0x79C29497C9486247ULL, "vxht_brothel_n_4w7_0_0.vlm"},
	{0x79C2B7FA924A02CDULL, "gen00fl_varric6078707.cl"},
	{0x79C2C7A3DA872726ULL, "mcr310_party_banter6087780.cl"},
	{0x79C2CC531EECD27AULL, "hlac_w_bonepitdestroyed_a3_4_0z.dds"},
	{0x79C373A20A233BBBULL, "316215386.ogg"},
	{0x79C3E11607455C48ULL, "hlac_lt_0_lowtown_docks_-3_8x.dds"},
	{0x79C3EB1FC5E5B73BULL, "6111613_m.fxe"},
	{0x79C41720E1B4CA76ULL, "hlac_lt_0_lowtown_docks_-4_2y.dds"},
	{0x79C43364AF1E2D30ULL, "llac_ht_chantry_2534_25.dds"},
	{0x79C472FD65533B76ULL, "hlac_dr_primevaltomb_663_0y.dds"},
	{0x79C47ABD3CC6A874ULL, "6025993_m.fxe"},
	{0x79C49E2B7E0DA92AULL, "zz_store_dre180pt_sandal.utm"},
	{0x79C571A7716C8549ULL, "sse_ss_player_male.gda"},
	{0x79C6223598A43FA5ULL, "docks_male_human_worker_sta6176065.cl"},
	{0x79C6300B2909A9A0ULL, "vrc321_housekeeper6085993.cl"},
	{0x79C643CA6ED27A5FULL, "mag100_thrask6034831.cl"},
	{0x79C6CFB6758B8CC0ULL, "vxlt_gallowstemplar_n_5ou_0_0.vlm"},
	{0x79C6D6D835320E6DULL, "gen00fl_aveline6122074.cl"},
	{0x79C74710AFA5A3CEULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.impa.evt"},
	{0x79C853DA7867E4B5ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_4y.dds"},
	{0x79C85E4D872527F3ULL, "gen00fl_aveline6139238.cl"},
	{0x79C97D8B06EB1024ULL, "vxw_bonepitdestroyed_a3_1gb_0_0.vlm"},
	{0x79CA084ECF82586CULL, "vxdr_primevalentrance_23c_0_0.vlm"},
	{0x79CA17237D813DACULL, "731429617.ogg"},
	{0x79CA253335701D64ULL, "281280333.ogg"},
	{0x79CA3E595EC83971ULL, "gencr_spider_poison.utc"},
	{0x79CA5EB1F5D14A1CULL, "plt_mer000pt_mine_payments.nss"},
	{0x79CAAF97234BDA56ULL, "vxdr_primevaltomb_7ta_0_0.vlm"},
	{0x79CABBA4F009A169ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_enter.ani"},
	{0x79CB139030113D10ULL, "llac_lt_0_lowtown_docks_196608_0.dds"},
	{0x79CBE1FC5DF458AAULL, "ht_0_hightown_a2_24_0.rml"},
	{0x79CC2B4C76582002ULL, "6047726_m.fxe"},
	{0x79CC63AE153B1743ULL, "454907086.ogg"},
	{0x79CCAD3E514E33EBULL, "6113116_m.fxe"},
	{0x79CD2B167F2447A3ULL, "vxlt_0_lowtown_docks_cou_0_0.vlm"},
	{0x79CDA032AF54D22EULL, "w_sundermount_247_0.mmh"},
	{0x79CDAC32AF54E6C5ULL, "w_sundermount_247_0.mao"},
	{0x79CDBA32AF54FE1CULL, "w_sundermount_247_0.msh"},
	{0x79CDDEF774B2FF09ULL, "885047711.ogg"},
	{0x79CE3E10AFABB379ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.impa.gad"},
	{0x79CE633CE705B4AAULL, "6031693_m.fxe"},
	{0x79CE72C23EEAC507ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.gad"},
	{0x79CE7FA5C0EDF3A2ULL, "lgt110_orz_harrowmont6108987.cl"},
	{0x79CEDA110A6478ADULL, "792742439.ogg"},
	{0x79CF16C743A8B13EULL, "229466533.ogg"},
	{0x79CF82797F69E7D1ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_dbl_c.mmh"},
	{0x79CF9574099125F3ULL, "6149267_m.fxe"},
	{0x79D022B5E6E1ACCEULL, "6042475_m.fxe"},
	{0x79D039D6A3D5167BULL, "6018529_m.fxe"},
	{0x79D03D62F2B92C7CULL, "6081656_m.fxe"},
	{0x79D089BDA4A614C5ULL, "vxdr_primevaltomb_26i_0_0.vlm"},
	{0x79D0AFA36BFF3F86ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_chac_sa.dds"},
	{0x79D10D0E3F741C35ULL, "vxlt_0_lowtown_h9k_0_0.vlm"},
	{0x79D1B2BFA0493716ULL, "w_sundermount_231_0.mao"},
	{0x79D1B6FDC3B26225ULL, "marksman_rapidfire.dds"},
	{0x79D1BEBFA0494BB5ULL, "w_sundermount_231_0.mmh"},
	{0x79D1C0BFA0494F2BULL, "w_sundermount_231_0.msh"},
	{0x79D1D6A4F00FEDEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_enter.gad"},
	{0x79D1D85E5B30B989ULL, "hlac_lt_0_lowtown_196607_1x.dds"},
	{0x79D1DB35C1F5C9D3ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_act_c.impa.evt"},
	{0x79D2B2582C60C5A1ULL, "ico_shortbow_0.dds"},
	{0x79D2B6FDB737E300ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_2_sylvesta.gad"},
	{0x79D2BF13045ACF6DULL, "216607557.ogg"},
	{0x79D30DFF4B2C277DULL, "vxht_0_hightown_c3n_0_0.vlm"},
	{0x79D391596801D2F8ULL, "37604240.ogg"},
	{0x79D44F10AFB03BC7ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.impa.ani"},
	{0x79D46BDA4066EFD7ULL, "815411560.ogg"},
	{0x79D4ED6552F0B15BULL, "gen00fl_merrill6162082.cl"},
	{0x79D4F330964D73F9ULL, "vxlt_0_lowtown_docks_a3_n_9rb_0_0.vlm"},
	{0x79D541936781A540ULL, "vxlt_gallowstemplar_n_5br_0_0.vlm"},
	{0x79D567C23EF0D114ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.evt"},
	{0x79D5D4CC20155DB8ULL, "6206866_m.fxe"},
	{0x79D6072152184683ULL, "follower_barks6153592.cl"},
	{0x79D69FB5AB556144ULL, "vxlt_0_lowtown_a2_n_c_emr_0_0.vlm"},
	{0x79D6B2FA55746984ULL, "34263877.ogg"},
	{0x79D6EEF6D4A14547ULL, "vxdr_primevaltomb_2wn_0_2.vlm"},
	{0x79D7DA31F052F9E5ULL, "hlac_ht_bartrand_n_3371_2z.dds"},
	{0x79D7EF67F8E128A4ULL, "vxw_bonepitdestroyed_a3_1hc_0_0.vlm"},
	{0x79D8B1C6DC2EE921ULL, "409612416.ogg"},
	{0x79D8D435C1FBDC14ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_act_c.impa.gad"},
	{0x79D8DC9A2D0DDFF3ULL, "w_sundermount_72_0.rml"},
	{0x79D8EB36C27CAF1DULL, "758379747.ogg"},
	{0x79D921741541F2D6ULL, "vxdr_primevaltomb_7qb_0_0.vlm"},
	{0x79D92FEEDAC91754ULL, "vxlt_0_lowtown_docks_a3_n_as4_0_0.vlm"},
	{0x79D981E30A908EB7ULL, "264438541.ogg"},
	{0x79D9D8B50F4749D1ULL, "vxlt_0_lowtown_docks_a3_n_cp6_0_0.vlm"},
	{0x79DA33D213D62402ULL, "vxlt_0_lowtown_docks_a3_n_d0f_0_0.vlm"},
	{0x79DA566822AFEAD6ULL, "mag100_karras6034538.cl"},
	{0x79DB24497EECC054ULL, "vxht_bartrand_n_5rn_0_0.vlm"},
	{0x79DB50FE83CD57D2ULL, "6050189_f.fxe"},
	{0x79DC2F26A1EC8DE8ULL, "6170934_m.fxe"},
	{0x79DC778F07A37D4AULL, "vxca_caverns2prc_a1w_0_0.vlm"},
	{0x79DCAA23A1F761D1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic.mmh"},
	{0x79DCB623A1F77672ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic.mao"},
	{0x79DCC423A1F78D8FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic.msh"},
	{0x79DD39F35436C652ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand_3.mao"},
	{0x79DDB50EEC88D8FAULL, "6076937_m.fxe"},
	{0x79DDC3ED522846B8ULL, "qcr000_arishok6059668.cl"},
	{0x79DE0750FB873E73ULL, "6142367_m.fxe"},
	{0x79DEBC41DC7412DEULL, "ser221_party_comments6163607.cl"},
	{0x79DED51E7AB36CB1ULL, "skin_depthlowlod.vsh.fxo"},
	{0x79DEE39A2D12574BULL, "w_sundermount_72_0.phy"},
	{0x79DF042DAB979DEAULL, "vxw_sundermount_2ly_0_0.vlm"},
	{0x79DF296BB96F5544ULL, "hlac_ca_caverns1_mines_24426_8x.dds"},
	{0x79DF36E2F43A182CULL, "6067376_f.fxe"},
	{0x79DFAD9155D0743BULL, "6094194_m.fxe"},
	{0x79DFB358FA7F7971ULL, "6133035_m.fxe"},
	{0x79DFCF093662F199ULL, "zz_vault_debug6178930.cl"},
	{0x79DFE335C20214F2ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_act_c.impa.ani"},
	{0x79E04AB3B134A08AULL, "hlac_lt_0_lowtown_131071_0z.dds"},
	{0x79E097E0A516A773ULL, "vxlt_0_lowtown_a2_n_bz5_0_0.vlm"},
	{0x79E0F853B99B6E65ULL, "vxht_0_hightown_a3_cay_0_0.vlm"},
	{0x79E15E2EAAA7A67CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_enter.ani"},
	{0x79E17FDC3D223FFDULL, "vxht_0_hightown_c6m_0_0.vlm"},
	{0x79E1B67DDCB94D5DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_large.mao"},
	{0x79E1BA7DDCB95396ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_large.mmh"},
	{0x79E1C47DDCB96494ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_large.msh"},
	{0x79E2F7A77D727BA2ULL, "vfx_chunkyparticle_p1.vsh.fxo.dx11.assembly"},
	{0x79E330FA30A0DAECULL, "6110727_m.fxe"},
	{0x79E35FC23EFD014DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_1.ani"},
	{0x79E3B5A5F3A7B984ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_8.pwk"},
	{0x79E3D1DCC31ED146ULL, "12029901.ogg"},
	{0x79E3D33490213EC8ULL, "hf_gen_dalish_archer.mor"},
	{0x79E4B6D4D8FC413FULL, "895045961.ogg"},
	{0x79E4BD9F1D7CD0E6ULL, "hlac_lt_backalley_n_0_7x.dds"},
	{0x79E4F8E91CE6320CULL, "6208169_m.fxe"},
	{0x79E659143C7E2E38ULL, "439144335.ogg"},
	{0x79E6D58876F5182DULL, "llac_lt_undercityencounter_-1_34.dds"},
	{0x79E732CADDF2425BULL, "vxlt_hangedman_n_1me_0_0.vlm"},
	{0x79E763D776068BCBULL, "6089088_m.fxe"},
	{0x79E7B5B65EFFC98FULL, "hlac_lt_0_lowtown_a2_-65537_2y.dds"},
	{0x79E7FC66B7B73E24ULL, "vxlt_0_lowtown_docks_b9m_0_0.vlm"},
	{0x79E7FF34C26FA3E1ULL, "517807221.ogg"},
	{0x79E86B2EAAADDABEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_enter.gad"},
	{0x79E87AAFE86A43F7ULL, "lgt210_eau_senestra6144936.cl"},
	{0x79E91BBC6F694A1AULL, "l_dae121ar_brothel_nt.bnk"},
	{0x79E92FD66090D9B8ULL, "llac_ca_caverns1_mines_24424_15.dds"},
	{0x79E94B8390BF30B4ULL, "mer231ar_arlan_hideout.are"},
	{0x79E9962670E2D8D4ULL, "vxht_bartrand_n_5uo_0_0.vlm"},
	{0x79E9C34FD42EDED0ULL, "vxw_bonepitmining_a2_8n_0_0.vlm"},
	{0x79E9ED62EDEAD384ULL, "aip_follower_anders.gda"},
	{0x79EA1263564DB597ULL, "6133583_m.fxe"},
	{0x79EA16F80DAEBE57ULL, "lt_0_lowtown_a3_n_c_mmap_ph.dds"},
	{0x79EA2FF30EFFF2D3ULL, "lt_0_lowtown_n_42_0.rml"},
	{0x79EA8A158FA80D7BULL, "6080241_m.fxe"},
	{0x79EB365CAF732E8AULL, "gen00fl_fenris6126975.cl"},
	{0x79EB654103AF573DULL, "1066027010.ogg"},
	{0x79EB7898113B3C23ULL, "gen00fl_aveline6114402.cl"},
	{0x79EB8EE014E390ECULL, "plt_cod_itm_sundering.plo"},
	{0x79EBC70A5CC9FE60ULL, "l_qcr220ar_hightown_nt.bnk"},
	{0x79ECD5C5BAB759CBULL, "6079066_f.fxe"},
	{0x79ECFBCC44624A41ULL, "l_mrl311ar_house.wbd"},
	{0x79ECFE764054B99FULL, "w_woundedcoast_51_0.msh"},
	{0x79ED04764054C441ULL, "w_woundedcoast_51_0.mmh"},
	{0x79ED10764054D8E2ULL, "w_woundedcoast_51_0.mao"},
	{0x79ED41C52CFD0E86ULL, "hlac_lt_house_-1_2x.dds"},
	{0x79ED760A9D8DB66AULL, "vxw_sundermount_2qz_0_0.vlm"},
	{0x79ED7851725CD845ULL, "hlac_lt_gallowstemplar_f_7198_5x.dds"},
	{0x79EDBF223979B3ECULL, "vxht_0_hightown_a2_by4_0_0.vlm"},
	{0x79EE603FC549623BULL, "565734182.ogg"},
	{0x79EEECE503B06613ULL, "407446037.ogg"},
	{0x79EF3A1B280EF55EULL, "6167626_m.fxe"},
	{0x79EF462EAAB3BADDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_enter.evt"},
	{0x79EFC9E74E23F850ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_4x.dds"},
	{0x79F026F94BBD2FE0ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\gibgolem_s2_movement.anb"},
	{0x79F032899E08A494ULL, "6155395_m.fxe"},
	{0x79F073E058D79EB4ULL, "plt_genpt_follower_approval.plo"},
	{0x79F17432E5274887ULL, "6092057_m.fxe"},
	{0x79F184E942D87494ULL, "hlac_dr_slavers_3217_12z.dds"},
	{0x79F18DDEB153A23FULL, "vxca_caverns2prc_86v_0_0.vlm"},
	{0x79F1C50F82A20E3BULL, "6070192_m.fxe"},
	{0x79F1D869879A0676ULL, "llac_lt_0_lowtown_a2_65538_2.dds"},
	{0x79F222766DE1AAFDULL, "hlac_lt_0_lowtown_a2_n_-262144_2y.dds"},
	{0x79F24CB3EB3DF046ULL, "vxw_encounter2_camp_6t_0_0.vlm"},
	{0x79F27EA1A6E90F22ULL, "and100_anders6095822.cl"},
	{0x79F2C012F0E320A2ULL, "15599627.ogg"},
	{0x79F43C58554D7D96ULL, "ch1multnt_vl.psh.fxo"},
	{0x79F45FDBCFBB5E74ULL, "mag110_sailor6107413.cl"},
	{0x79F4BFF2D3F67ABFULL, "vxdr_primevalentrance_23l_0_0.vlm"},
	{0x79F4DBCC12147FEDULL, "gen00fl_fenris6127106.cl"},
	{0x79F5DAFF774A95A5ULL, "1006345963.ogg"},
	{0x79F60462D3206ED8ULL, "gen_im_trs_gen_statue_01.uti"},
	{0x79F629DEBB15C370ULL, "vxlt_0_lowtown_docks_n_bda_0_0.vlm"},
	{0x79F63D547060496BULL, "460545276.ogg"},
	{0x79F727162BEE4C79ULL, "176720228.ogg"},
	{0x79F77C4963494391ULL, "one280cr_uncontrolled_golem.utc"},
	{0x79F7CD20CD819422ULL, "6139180_m.fxe"},
	{0x79F8080362D8F154ULL, "vxht_bartrand_n_5xl_0_0.vlm"},
	{0x79F808BA490D6D57ULL, "vxlt_0_lowtown_a2_n_e31_0_0.vlm"},
	{0x79F884CECF450FAAULL, "vxht_mansion_n_5av_0_0.vlm"},
	{0x79F8A924E4A72F66ULL, "45516522.ogg"},
	{0x79FA2E76B054DDDAULL, "ave331_jeven6151135.cl"},
	{0x79FA41986D3A9800ULL, "6051406_m.fxe"},
	{0x79FA84A2172FD7DBULL, "vxlt_0_lowtown_eu8_0_0.vlm"},
	{0x79FAEE91030F727CULL, "art\\characters\\creatures\\highdragon\\cn_bdy_hdr.mao"},
	{0x79FAFA91030F861FULL, "art\\characters\\creatures\\highdragon\\cn_bdy_hdr.mmh"},
	{0x79FB0091030F90C1ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_hdr.msh"},
	{0x79FB930C5A01A53FULL, "vxht_0_hightown_a2_bs9_0_0.vlm"},
	{0x79FBC64E5F803A33ULL, "6077059_f.fxe"},
	{0x79FC6B196AF62A32ULL, "narrative_varric6124283.cl"},
	{0x79FC9712DE6BAA71ULL, "6139635_m.fxe"},
	{0x79FCB3A260EB18C2ULL, "217186698.ogg"},
	{0x79FD7F86E19008A5ULL, "lgt210_mag_escp_rescued6156647.cl"},
	{0x79FDEA2C855F72A7ULL, "character_character_p264.psh.fxo.dx11.assembly"},
	{0x79FDFECDF3D4997DULL, "lgt260_eau_tmp_roderick.bnk"},
	{0x79FEE99CDD689164ULL, "6140201_m.fxe"},
	{0x79FF12FF54EAD92BULL, "vxlt_0_lowtown_a2_ei8_0_0.vlm"},
	{0x79FF67F19B4E9418ULL, "121492184.ogg"},
	{0x79FFB1D513214BD4ULL, "17907772.ogg"},
	{0x79FFF80B667932F3ULL, "llac_lt_0_lowtown_a2_n_-262143_0.dds"},
	{0x7A000E57964E8EDCULL, "lowtown_guard_chase6169500.cl"},
	{0x7A0054114B466A5EULL, "vxdr_primevaltomb_84f_0_0.vlm"},
	{0x7A0074E3E54A0C88ULL, "6136680_m.fxe"},
	{0x7A00C211B35ED844ULL, "lowtown_fem_alienage1_sta6162568.cl"},
	{0x7A0109996597665CULL, "pro000_merchant6051617.cl"},
	{0x7A0112F56223FBFEULL, "art\\vfx\\combat\\vfx_glb_lit_imp_maj_c.impa.evt"},
	{0x7A013F3351ACB563ULL, "79571693.ogg"},
	{0x7A0182205E24C5E6ULL, "vxlt_warehouse_n_qo_0_0.vlm"},
	{0x7A01E0CDF3D84E85ULL, "lgt260_eau_tmp_roderick.cnv"},
	{0x7A01F62044E36A59ULL, "dre000_cut2d_ragstoriches6025094.cl"},
	{0x7A02C57B82C8F7A3ULL, "vxdr_primevalentrance_2eb_0_0.vlm"},
	{0x7A02D2E0EDEF956EULL, "vxw_bonepitdestroyed_a3_1wy_0_1.vlm"},
	{0x7A02DF761614736FULL, "hlac_lt_darktown_-1_47z.dds"},
	{0x7A0311EC6C48DBFEULL, "726751190.ogg"},
	{0x7A03EF16FB4F81AAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowalcove.mao"},
	{0x7A03F316FB4F88F9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowalcove.mmh"},
	{0x7A03FD16FB4F99E7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowalcove.msh"},
	{0x7A0438DE6E7F8561ULL, "90728362.ogg"},
	{0x7A049BBBAD0BDBF0ULL, "vxlt_0_lowtown_docks_n_ban_0_0.vlm"},
	{0x7A04BDA179E62C68ULL, "6112075_m.fxe"},
	{0x7A056C818B987CAAULL, "305224004.ogg"},
	{0x7A058625A00DB436ULL, "299398994.ogg"},
	{0x7A05CEA425F26488ULL, "6089616_m.fxe"},
	{0x7A05EAAC516D8981ULL, "6126980_f.fxe"},
	{0x7A05F962DE9BBAD7ULL, "follower_banter6155430.cl"},
	{0x7A061D43B7B80C5BULL, "325653755.ogg"},
	{0x7A06C2A2D10A59BFULL, "895716375.ogg"},
	{0x7A06C83B04A802B0ULL, "vxdr_0_deeproads_fig_0_0.vlm"},
	{0x7A06E4131A7E4C9BULL, "6179475_m.fxe"},
	{0x7A07328390D80396ULL, "mer231ar_arlan_hideout.ncs"},
	{0x7A07A7794C5AEE0AULL, "vxlt_0_lowtown_docks_a3_aqk_0_0.vlm"},
	{0x7A081FC5945DE79FULL, "335057101.ogg"},
	{0x7A0829F5622A4129ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_maj_c.impa.gad"},
	{0x7A08F1B86156450AULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower1_varric_.gad"},
	{0x7A091D54E182F0CAULL, "6151624_f.fxe"},
	{0x7A0A60798743F09BULL, "vxht_mansion2b_257_0_0.vlm"},
	{0x7A0B8D0CBFB4ABE4ULL, "373632103.ogg"},
	{0x7A0BC55CA82B5F23ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_02tent.msh"},
	{0x7A0BD35CA82B777DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_02tent.mmh"},
	{0x7A0BD75CA82B7D8EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_02tent.mao"},
	{0x7A0BDC8CE21C98DBULL, "6176302_m.fxe"},
	{0x7A0BEE51645049B6ULL, "gen00fl_isabela6085852.cl"},
	{0x7A0C0ABB215CA9C4ULL, "lla_ht_0_hightown_a3_-131074_0.dds"},
	{0x7A0C3239085BC0FFULL, "llac_lt_gallowscourtyard_n_131071_1.dds"},
	{0x7A0C455098D9932BULL, "vxdr_dungeon_23b_0_0.vlm"},
	{0x7A0DD6C53DCF1670ULL, "871255181.ogg"},
	{0x7A0DE3A7C65DD97FULL, "lgt311_shw_mer_mettin6156456.cl"},
	{0x7A0DE3CF45388A38ULL, "llac_lt_0_lowtown_a2_n_65536_9.dds"},
	{0x7A0E0AD907037A92ULL, "dae100_viscount6118496.cl"},
	{0x7A0E35BC6F8892B8ULL, "l_dae121ar_brothel_nt.wbd"},
	{0x7A0E4A8390DE4B2FULL, "mer231ar_arlan_hideout.lst"},
	{0x7A0EFAF562301077ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_maj_c.impa.ani"},
	{0x7A0FE5B0DD0ED4ABULL, "vxht_keep_n_6b1_0_0.vlm"},
	{0x7A0FF3FD501ADE66ULL, "vxlt_warehouse_n_tp_0_0.vlm"},
	{0x7A105E41B556C388ULL, "llac_lt_0_lowtown_docks_a3_65535_3.dds"},
	{0x7A106729732DC344ULL, "vxlt_0_lowtown_a2_n_c_eqv_0_0.vlm"},
	{0x7A121DCC4481A043ULL, "l_mrl311ar_house.bnk"},
	{0x7A12A0FF942DCDDAULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\1024texel.dds"},
	{0x7A12B08113383455ULL, "lt_0_lowtown_a2_n_c_11_0.phy"},
	{0x7A12E47640755D6AULL, "w_woundedcoast_51_0.rml"},
	{0x7A130E58FAB135DEULL, "mag200_feynriel6030445.cl"},
	{0x7A13B1394E868B47ULL, "follower_banter6155940.cl"},
	{0x7A13C15EB9E14DB1ULL, "w_bonepitmining_a2_126_0.msh"},
	{0x7A13CF5EB9E165CCULL, "w_bonepitmining_a2_126_0.mao"},
	{0x7A13DB5EB9E17A6FULL, "w_bonepitmining_a2_126_0.mmh"},
	{0x7A140FC70C8FF76EULL, "644832294.ogg"},
	{0x7A148481087EEC3BULL, "vxlt_hangedman_n_31o_0_0.vlm"},
	{0x7A159581237E8DE4ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.dura.ani"},
	{0x7A17CAD592C2220FULL, "285250281.ogg"},
	{0x7A17FA841C400A46ULL, "vxca_caverns1_j5v_0_0.vlm"},
	{0x7A18B6CF913E02B6ULL, "vxlt_undercityserialkiller_1im_0_1.vlm"},
	{0x7A1924C56707135AULL, "vxht_mansion2_1r0_0_0.vlm"},
	{0x7A19368166CF76D5ULL, "6118577_m.fxe"},
	{0x7A1937411E1EA486ULL, "659224820.ogg"},
	{0x7A199CBE61478296ULL, "6177296_m.fxe"},
	{0x7A19D181133E8A81ULL, "lt_0_lowtown_a2_n_c_11_0.rml"},
	{0x7A1A24A23AF226C4ULL, "6039277_m.fxe"},
	{0x7A1AB0F72443E1F2ULL, "6095039_m.fxe"},
	{0x7A1B3F531AB487E6ULL, "zz_seb_debug6215748.cl"},
	{0x7A1B6D17CED08E18ULL, "hlac_lt_0_lowtown_a2_-2_4x.dds"},
	{0x7A1C66278D596DC2ULL, "6103642_m.fxe"},
	{0x7A1C92812384A746ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.dura.gad"},
	{0x7A1C9C0A40D28E22ULL, "6113834_m.fxe"},
	{0x7A1CA9FC04F9DDDDULL, "zz_andpt_debug.ncs"},
	{0x7A1CC25AC8D3AA51ULL, "535495437.ogg"},
	{0x7A1CCCB86166AFF0ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower1_varric_.ani"},
	{0x7A1D114D4B0797E3ULL, "vxdr_primevaltomb_7zs_0_1.vlm"},
	{0x7A1D15881CAA85A5ULL, "vxht_brothel_n_3va_0_0.vlm"},
	{0x7A1D3A789F14D0E6ULL, "vxdr_0_deeproads_g2y_0_0.vlm"},
	{0x7A1D3E9AF20A9449ULL, "6156878_m.fxe"},
	{0x7A1D447676B57A26ULL, "6180799_m.fxe"},
	{0x7A1DF66276363E9EULL, "55674380.ogg"},
	{0x7A1E66BD235685BEULL, "hlac_lt_0_lowtown_a2_n_c_1_3z.dds"},
	{0x7A1E8DBEFAB3DDC2ULL, "hlac_lt_undercityserialkiller_-1_37x.dds"},
	{0x7A1EDE1EAEF12894ULL, "vxca_caverns1_idh_0_0.vlm"},
	{0x7A1F93ED49753789ULL, "vxw_sundermount_2ze_0_1.vlm"},
	{0x7A20B20E84C58D2DULL, "vxht_mansion_n_6mz_0_0.vlm"},
	{0x7A20F25A0875F494ULL, "93069292.ogg"},
	{0x7A20FB63E147984FULL, "vxht_bartrand_n_67y_0_0.vlm"},
	{0x7A215AB60B08B008ULL, "vxw_sundermount_342_0_0.vlm"},
	{0x7A2175BD87FD9586ULL, "6068668_m.fxe"},
	{0x7A2200A2CD8235F0ULL, "507729015.ogg"},
	{0x7A22266B16B79664ULL, "vxlt_0_lowtown_a2_n_c_db6_0_0.vlm"},
	{0x7A226FB1F61BDD47ULL, "plt_mer000pt_mine_payments.plo"},
	{0x7A228C2088744D6BULL, "6093744_m.fxe"},
	{0x7A22C1DCD011F039ULL, "mer230tr_preload_coterie.utt"},
	{0x7A2373F8330734E8ULL, "ave200_donnic6112125.cl"},
	{0x7A238D81238ABD25ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.dura.evt"},
	{0x7A24808FEA45202AULL, "952118581.ogg"},
	{0x7A24D70433CE42E2ULL, "6121967_m.fxe"},
	{0x7A2549A127D59DCFULL, "hlac_ca_caverns2_5802_18y.dds"},
	{0x7A2564BFD11D96D2ULL, "712279688.ogg"},
	{0x7A25ABFE68477677ULL, "hlac_lt_backalley_n_0_11z.dds"},
	{0x7A25AC3AAEA7955BULL, "hlac_ca_caverns1_24422_0x.dds"},
	{0x7A25C7D7D562CD2FULL, "hlac_lt_gallowstemplar_n_2392_0x.dds"},
	{0x7A25F80386278F10ULL, "hlac_ca_caverns2_5802_21z.dds"},
	{0x7A2678502A31D82BULL, "zz_followers_banter6158040.cl"},
	{0x7A26E000CFE50015ULL, "llac_ht_0_hightown_n_-65536_3.dds"},
	{0x7A2746F50F236BB2ULL, "hlac_ca_cavern1_haunted_24426_4x.dds"},
	{0x7A2796A258FD2BDAULL, "vxht_mansion2_1u1_0_0.vlm"},
	{0x7A27DAC6767A21D3ULL, "668661479.ogg"},
	{0x7A280C1C095B99C2ULL, "lt_0_lowtown_n_43_0.rml"},
	{0x7A28854137585DB2ULL, "vxht_0_hightown_a2_cd8_0_0.vlm"},
	{0x7A2954A4906DDE8CULL, "vxlt_0_lowtown_a3_n_c_hxa_0_0.vlm"},
	{0x7A296DB4D6262D7CULL, "llac_lt_0_lowtown_docks_n_-196610_5.dds"},
	{0x7A29AF63B0CC0A51ULL, "zz_rogue1.cut"},
	{0x7A29B4AB84485DD2ULL, "hlac_lt_gallowstemplar_2388_3x.dds"},
	{0x7A29C90DFDE62EFCULL, "follower_banter6170089.cl"},
	{0x7A2A0390A0244543ULL, "113062608.ogg"},
	{0x7A2A325477BFD159ULL, "zz_approval_debug6003947.cl"},
	{0x7A2A456AD74A3AEEULL, "art\\vfx\\combat\\vfx_gib_hm_masc_rblt.anb"},
	{0x7A2A4C11E30283ADULL, "84885273.ogg"},
	{0x7A2A784DC7090E2EULL, "limp_distor.mao"},
	{0x7A2A80E173D0D730ULL, "6051309_m.fxe"},
	{0x7A2AD7363C9B14F5ULL, "dae200_bran6094824.cl"},
	{0x7A2B026F50512F5DULL, "zz_followers_banter6159081.cl"},
	{0x7A2B1C1F57F2EA67ULL, "1021215049.ogg"},
	{0x7A2B1EBAB388B8F0ULL, "957336298.ogg"},
	{0x7A2B694EEBB6440AULL, "921153065.ogg"},
	{0x7A2B9F2D9A8047BBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedflurrish.evt"},
	{0x7A2C078D8702172BULL, "fer_ter_dirt_001_diff.dds"},
	{0x7A2CFD8E3ED8A3D1ULL, "hlac_ht_keep_3732_5z.dds"},
	{0x7A2D3987FDBB3C2CULL, "169418414.ogg"},
	{0x7A2D4FFBA0E74114ULL, "vxca_caverns1_ioi_0_0.vlm"},
	{0x7A2D57C88260ADB1ULL, "121269844.ogg"},
	{0x7A2DAFD456845AE6ULL, "em_shadow.mor"},
	{0x7A2DE8B244D080A5ULL, "6206845_m.fxe"},
	{0x7A2DE8F736E4C15EULL, "6015920_f.fxe"},
	{0x7A2DFCF6E3537DE9ULL, "769200050.ogg"},
	{0x7A2E4DB5D6C20041ULL, "vxw_encounter2_camp_jk_0_0.vlm"},
	{0x7A2E990846CA33DCULL, "6166173_m.fxe"},
	{0x7A2EAA4735823E1EULL, "gui_reticle_p0.vsh.fxo.dx11.assembly"},
	{0x7A2F1EA8C9281E3AULL, "lgt120_orz_bln_official6140215.cl"},
	{0x7A2F1FD7DFD3D9A8ULL, "632173893.ogg"},
	{0x7A2F50246308893AULL, "vxlt_darktown_4x4_0_0.vlm"},
	{0x7A2F5C85C2F52BF8ULL, "qun120cr_javaris6048962.cl"},
	{0x7A2F8E8136179C32ULL, "188406768.ogg"},
	{0x7A2F9054644BED9EULL, "652114850.ogg"},
	{0x7A2F93D91A3C9EFDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_singledoor.phy"},
	{0x7A2FAA1DD6BEEE38ULL, "vxlt_0_lowtown_ewi_0_0.vlm"},
	{0x7A2FCC92FCFEC888ULL, "vxw_sundermount_333_0_0.vlm"},
	{0x7A304976CF1B4B96ULL, "vxht_0_hightown_n_c5l_0_0.vlm"},
	{0x7A30F3BB3C875762ULL, "846538646.ogg"},
	{0x7A3154D9CFD07552ULL, "ser221_gamlen6131919.cl"},
	{0x7A315AC8417D271CULL, "dae000_unhappy_mage6165320.cl"},
	{0x7A31982D9A84A7FCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedflurrish.gad"},
	{0x7A31B2397A0F6F8DULL, "vxca_caverns2prc_4x3_0_1.vlm"},
	{0x7A32680FEDF48A6DULL, "gen00fl_varric6105379.cl"},
	{0x7A33ED8DC23075E0ULL, "6204802_m.fxe"},
	{0x7A342FE53F7177C7ULL, "qun131_taskmaster6039400.cl"},
	{0x7A36087F4AF3445AULL, "vxht_mansion2_1x2_0_0.vlm"},
	{0x7A360D36ADF9C3C0ULL, "llac_lt_undercityencounter_-1_48.dds"},
	{0x7A3620D9317D21DCULL, "6089744_m.fxe"},
	{0x7A362D85CA81CC49ULL, "gen00fl_isabela_init6175661.cl"},
	{0x7A370076FE093204ULL, "6171461_m.fxe"},
	{0x7A371B9E8615D25FULL, "gen00fl_merrill6145139.cl"},
	{0x7A374A28F069A79DULL, "vxht_0_hightown_a3_bfq_0_0.vlm"},
	{0x7A375CFA851573FAULL, "6139253_f.fxe"},
	{0x7A37ECE88E634D98ULL, "6120045_m.fxe"},
	{0x7A38972D9A8AC4FAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedflurrish.ani"},
	{0x7A3B0170F2EE5B81ULL, "vxht_bartrand_n_6k3_0_0.vlm"},
	{0x7A3B132011795FB5ULL, "vxht_chantry_n_2cl_0_0.vlm"},
	{0x7A3B50017489FC4EULL, "vxlt_gallowscourtyard_n_711_0_0.vlm"},
	{0x7A3B605EFFB2FFF5ULL, "6096204_m.fxe"},
	{0x7A3B61609645145EULL, "llac_ht_0_hightown_n_-131073_0.dds"},
	{0x7A3B7F83D9A5F811ULL, "dae000_keep_male_servant_mo6161482.cl"},
	{0x7A3BA204B5009762ULL, "408210089.ogg"},
	{0x7A3BC1BC321F40B4ULL, "vxht_0_hightown_a3_c5p_0_0.vlm"},
	{0x7A3BD00529CF65BBULL, "ico_crossbow_4.dds"},
	{0x7A3BE2BF1DEAD07FULL, "627985549.ogg"},
	{0x7A3C1353E1934B3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_dead_loop.gad"},
	{0x7A3CBF92C8B818C1ULL, "vxw_encounter2_camp_ej_0_0.vlm"},
	{0x7A3D396A783C1C6AULL, "mcr310_sibling6076268.cl"},
	{0x7A3D629A611FA46FULL, "vxlt_hangedman_26d_0_0.vlm"},
	{0x7A3DB26388478636ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic_4.mao"},
	{0x7A3DC79B7354E51CULL, "lt_0_lowtown_docks_a3_n_27_0_ll.mmh"},
	{0x7A3E1BFAC8B506B8ULL, "vxlt_0_lowtown_erf_0_0.vlm"},
	{0x7A3FAC1114DFC44CULL, "gen_im_trs_gen_med_hlm_r1.uti"},
	{0x7A401332AB7D3806ULL, "6167243_f.fxe"},
	{0x7A4088D91A4AFF69ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_singledoor.mao"},
	{0x7A4094D91A4B1302ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_singledoor.mmh"},
	{0x7A4096D91A4B16B8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_singledoor.msh"},
	{0x7A40C2DCD02AEFFCULL, "mer230tr_preload_coterie.ncs"},
	{0x7A40C9246B63212BULL, "mi_chun_cl.mao"},
	{0x7A415378CB4E71CDULL, "168696331.ogg"},
	{0x7A415D456D4C1119ULL, "dae000_bartrand6206626.cl"},
	{0x7A41D0C718B6D8EEULL, "353785331.ogg"},
	{0x7A41D1A157EA8647ULL, "mag210_arianni6024121.cl"},
	{0x7A41F1FE88BE0A95ULL, "38314374.ogg"},
	{0x7A41F33915DD7F08ULL, "469567212.ogg"},
	{0x7A41FB2A8AC43C33ULL, "rdr111dg_martin6058282.cl"},
	{0x7A431253E199683AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_dead_loop.ani"},
	{0x7A432D20E0900B95ULL, "216904746.ogg"},
	{0x7A436CA891CAB058ULL, "hlac_lt_gallowscourtyard_n_-65537_4z.dds"},
	{0x7A4398F20FCC7AD0ULL, "6004120_f.fxe"},
	{0x7A447A5C3CE95CDAULL, "vxht_mansion2_1c3_0_0.vlm"},
	{0x7A454BB102177FDAULL, "mrl250cr_spiderling_cv.utc"},
	{0x7A4560C4F0CD1FC5ULL, "arm_pl1.tnt"},
	{0x7A456F49638B4F5FULL, "one280cr_uncontrolled_golem.ncs"},
	{0x7A45D667A2AA1B08ULL, "6078704_m.fxe"},
	{0x7A462D28217FFC89ULL, "hightown_male_noble2_mob6170908.cl"},
	{0x7A46AD6F1D1A42F3ULL, "6149014_m.fxe"},
	{0x7A4792B9B991B3D0ULL, "vxlt_0_lowtown_n_ep3_0_0.vlm"},
	{0x7A47CC5CB8F9B7C6ULL, "hlac_w_encounter2_0_0x.dds"},
	{0x7A47E86EA99D61CBULL, "vxdr_primevaltomb_7jz_0_0.vlm"},
	{0x7A486582D1AF1F42ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_thin.msh"},
	{0x7A486782D1AF22F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_thin.mmh"},
	{0x7A487382D1AF369BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_thin.mao"},
	{0x7A48E31DA946EC26ULL, "hlac_w_encounter3_-262147_0x.dds"},
	{0x7A4991E047527603ULL, "vxw_sundermount_2f2_0_0.vlm"},
	{0x7A49B282053851BCULL, "dae000ip_crowd_mf2_talk13.utp"},
	{0x7A49C1DE668014CEULL, "vxlt_gallowscourtyard_n_742_0_0.vlm"},
	{0x7A4A1285C9EEFECBULL, "llac_ht_keep_n_3731_3.dds"},
	{0x7A4A1F725E2BCC78ULL, "vxca_caverns2_8mi_0_0.vlm"},
	{0x7A4B1B636A90AABAULL, "zz_followers_banter6157587.cl"},
	{0x7A4BEFE8903C31A9ULL, "gen00fl_anders6125559.cl"},
	{0x7A4C1B432CCE4EFEULL, "w_encounter2_camp_36_0.rml"},
	{0x7A4C728BFF0FB9D7ULL, "vxlt_0_lowtown_docks_a3_n_ctd_0_0.vlm"},
	{0x7A4C8DD7BAAB1F38ULL, "vxlt_0_lowtown_emg_0_0.vlm"},
	{0x7A4CB0811369AC11ULL, "lt_0_lowtown_a2_n_c_11_0.anb"},
	{0x7A4DE0F2CF7A9336ULL, "dae000cr_grand_cleric6147874.cl"},
	{0x7A4DE1C581CFDC7BULL, "hlac_lt_0_lowtown_docks_n_-131071_0x.dds"},
	{0x7A4EB0E590F96951ULL, "zz_vault_debug6177768.cl"},
	{0x7A4F3FEC0E78FB71ULL, "6076408_m.fxe"},
	{0x7A4F4ABC56D315EEULL, "hlac_ca_caverns2prc_5802_25y.dds"},
	{0x7A4F9370C78D2C9AULL, "gen00fl_isabela6106289.cl"},
	{0x7A5020691D476F0FULL, "art\\vfx\\combat\\vfx_glb_spi_imp_maj_c.mmh"},
	{0x7A50E07684DAA833ULL, "mag101_grace6035854.cl"},
	{0x7A50F7B96A0D6061ULL, "vxw_bonepitmining_a2_1nb_0_0.vlm"},
	{0x7A51833CA4A1B2EBULL, "vxw_woundedcoast_z_0_0.vlm"},
	{0x7A5252F64656DBDBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1rockpile.mmh"},
	{0x7A525EF64656F078ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1rockpile.mao"},
	{0x7A526E82D1B78283ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_thin.phy"},
	{0x7A5270F646570EC5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1rockpile.msh"},
	{0x7A527416D524E79FULL, "6106998_m.fxe"},
	{0x7A530D4F7F0F23E2ULL, "zz_followers_banter6159516.cl"},
	{0x7A5334432CD497C2ULL, "w_encounter2_camp_36_0.phy"},
	{0x7A5357477D0418CCULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_bethany.gad"},
	{0x7A5381AAE80A8443ULL, "383293571.ogg"},
	{0x7A5444B2F9674370ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_016a_0n.dds"},
	{0x7A5490428EADDD5CULL, "llac_lt_0_lowtown_docks_a3_-196610_3.dds"},
	{0x7A5492D1B47D2D65ULL, "1031983550.ogg"},
	{0x7A5499859124DB12ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar2_.gad"},
	{0x7A54A22743E87AA6ULL, "dre000_bodahn6031312.cl"},
	{0x7A54AA3B6650280CULL, "vxlt_0_lowtown_a3_n_c_hod_0_0.vlm"},
	{0x7A558CA1C401121EULL, "1065470209.ogg"},
	{0x7A558E69A6BDF6E7ULL, "mrl150_hunter6071692.cl"},
	{0x7A55CC0548CEEBAFULL, "6042562_m.fxe"},
	{0x7A568A0A1FB482EAULL, "244830968.ogg"},
	{0x7A569BE1B0D526F6ULL, "zz_mageslate_baddy.cut"},
	{0x7A57659A9D8AF193ULL, "vxht_bartrand_n_622_0_0.vlm"},
	{0x7A57926C093A1C4AULL, "follower_banter6155925.cl"},
	{0x7A57E52AD6DA8C81ULL, "vxht_bartrand_n_6q1_0_0.vlm"},
	{0x7A5803BD39488E83ULL, "vxw_sundermount_2k5_0_0.vlm"},
	{0x7A5810028B81BE92ULL, "vxht_0_hightown_a3_bge_0_0.vlm"},
	{0x7A58179E7E3EDC1DULL, "6200713_m.fxe"},
	{0x7A584446491C2FA4ULL, "art\\vfx\\cinematic\\vfx_m_fire_pillar_imp_p.impa.ani"},
	{0x7A5887F9A3C61878ULL, "vxht_0_hightown_bgb_0_0.vlm"},
	{0x7A58EC539811C0A4ULL, "6119957_m.fxe"},
	{0x7A59048405D88804ULL, "art\\levels\\races\\vfx_models\\prp_brazlogmed.mmh"},
	{0x7A59088405D88F57ULL, "art\\levels\\races\\vfx_models\\prp_brazlogmed.mao"},
	{0x7A591A8405D8ADA6ULL, "art\\levels\\races\\vfx_models\\prp_brazlogmed.msh"},
	{0x7A59263BC86FF076ULL, "336558970.ogg"},
	{0x7A59934CE7E5D251ULL, "559724108.ogg"},
	{0x7A59AA35B605F559ULL, "760409894.ogg"},
	{0x7A59C38113745770ULL, "lt_0_lowtown_a2_n_c_11_0.msh"},
	{0x7A59D18113746E8AULL, "lt_0_lowtown_a2_n_c_11_0.mmh"},
	{0x7A59D581137475C1ULL, "lt_0_lowtown_a2_n_c_11_0.mao"},
	{0x7A5A5599182721F3ULL, "534020117.ogg"},
	{0x7A5A56477D0A35CAULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_bethany.ani"},
	{0x7A5A806F2B620C48ULL, "vxw_bonepitmining_a2_1ng_0_0.vlm"},
	{0x7A5AA5BB38EAD2BAULL, "vxlt_darktown_4i7_0_0.vlm"},
	{0x7A5B06CF83115986ULL, "vxw_sundermount_37h_0_1.vlm"},
	{0x7A5B7FA876DD6213ULL, "874882174.ogg"},
	{0x7A5BAF209B09DFCEULL, "vxlt_0_lowtown_docks_a3_n_d4f_0_0.vlm"},
	{0x7A5BB95A5894BDBDULL, "6075091_f.fxe"},
	{0x7A5BC352045BC2B2ULL, "6094087_m.fxe"},
	{0x7A5CE2328AD68EF8ULL, "hlac_lt_0_lowtown_docks_n_65534_0z.dds"},
	{0x7A5CF56363091FECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_action.ani"},
	{0x7A5D053DE619A09FULL, "415985625.ogg"},
	{0x7A5D59904EAAB216ULL, "vxht_brothel_n_59q_0_0.vlm"},
	{0x7A5E039103641364ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_hdr.phy"},
	{0x7A5E414649209606ULL, "art\\vfx\\cinematic\\vfx_m_fire_pillar_imp_p.impa.gad"},
	{0x7A5E42F744BA7428ULL, "m_dae210ar_lowtown_nt.bnk"},
	{0x7A5E493E52AFF07FULL, "6078985_m.fxe"},
	{0x7A5EF42E05E41450ULL, "412022597.ogg"},
	{0x7A5F69965C0378E1ULL, "vxw_bonepitmining_a2_1ka_0_0.vlm"},
	{0x7A5F8A8FA47A3DD9ULL, "lgt220_shp_mag_guard6149510.cl"},
	{0x7A60448370F3DB0CULL, "llac_lt_gallowsprison_n_-1_7.dds"},
	{0x7A605B4798626227ULL, "vxw_blightlands_c1_0_0.vlm"},
	{0x7A60CC3ED66E97F4ULL, "fen210_orana6133032.cl"},
	{0x7A6107DD366D1D15ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgtp_0t.dds"},
	{0x7A61362CDAC14B98ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2_tool.phy"},
	{0x7A6169BD1070C3ABULL, "gen00fl_anders6114438.cl"},
	{0x7A6194C41AACA3FEULL, "6161237_m.fxe"},
	{0x7A62091C098D0CBEULL, "lt_0_lowtown_n_43_0.anb"},
	{0x7A620EBEB8D08A37ULL, "vxht_mansion2_3gh_0_0.vlm"},
	{0x7A62C1F1223D254DULL, "vxw_encounter2_pd_0_1.vlm"},
	{0x7A630263630DA12EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_action.gad"},
	{0x7A6382C49E54AB6EULL, "6113067_m.fxe"},
	{0x7A64289278CD28F8ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mere.phy"},
	{0x7A647EBD7A1C5706ULL, "6086256_f.fxe"},
	{0x7A6488C85DE36932ULL, "zz_mcr_debug6091908.cl"},
	{0x7A6496C39BAFCFAFULL, "vxht_0_hightown_a3_7g1_0_0.vlm"},
	{0x7A650785467C4C1DULL, "vxht_mansion_n_73y_0_0.vlm"},
	{0x7A653A2FFA88EE3CULL, "6020459_m.fxe"},
	{0x7A653C464926ACE5ULL, "art\\vfx\\cinematic\\vfx_m_fire_pillar_imp_p.impa.evt"},
	{0x7A65BAB912728D6CULL, "vxca_caverns2prc_8it_0_0.vlm"},
	{0x7A66142F71E444C3ULL, "lt_0_lowtown_docks_a3_38_0.rml"},
	{0x7A66759A2B3EA703ULL, "vxw_sundermount_2l4_0_0.vlm"},
	{0x7A66DCF01B2FC7B1ULL, "73626924.ogg"},
	{0x7A66F416FBA407E6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowalcove.phy"},
	{0x7A67620B52CC16C4ULL, "185586975.ogg"},
	{0x7A67907A36CEC0A7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_2arm_shrug.gad"},
	{0x7A67F30ACDF6DC3DULL, "mag110_samson6049739.cl"},
	{0x7A6820A4CADC5612ULL, "vxht_keep_55i_0_0.vlm"},
	{0x7A68F24C1D5824C8ULL, "vxw_bonepitmining_a2_1kd_0_0.vlm"},
	{0x7A690078B16B74C0ULL, "6113674_m.fxe"},
	{0x7A695645E2FBEAD7ULL, "vxlt_0_lowtown_docks_a3_n_czf_0_0.vlm"},
	{0x7A6978AC75077206ULL, "vxw_sundermount_32k_0_1.vlm"},
	{0x7A698485913713B8ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar2_.ani"},
	{0x7A69BF19849582AEULL, "mcr370_negotiation6080233.cl"},
	{0x7A69E05806C5927AULL, "hlac_lt_0_lowtown_65535_1z.dds"},
	{0x7A69EA5C8E14EB33ULL, "hlac_lt_backalley_g_65537_3x.dds"},
	{0x7A69FD636313B70DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_mage_action.evt"},
	{0x7A6A272F3738F00FULL, "hlac_ht_playershouse_n_855_31z.dds"},
	{0x7A6A696488247928ULL, "hlac_ca_caverns1_24421_6x.dds"},
	{0x7A6B197B9BE3FB96ULL, "vxht_0_hightown_a2_c7y_0_0.vlm"},
	{0x7A6BDFF275392E5CULL, "235141837.ogg"},
	{0x7A6D3BC5E4BB6D24ULL, "atl_loginconne_dxt5_dat.xml"},
	{0x7A6D4A47E54242E9ULL, "gen_im_npc_arm_hum_cht_krkb.uti"},
	{0x7A6D6AE9B3CE0DD8ULL, "dm_anso.mor"},
	{0x7A6D82296AA29232ULL, "vxca_caverns2_7xk_0_0.vlm"},
	{0x7A6D99FB5024E071ULL, "isa111_isabela6068986.cl"},
	{0x7A6DE81A2E9CF775ULL, "mer221_sabin6047450.cl"},
	{0x7A6E857A36D4CCB4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_2arm_shrug.evt"},
	{0x7A6E96696E843C84ULL, "vxlt_gallowstemplar_n_5qs_0_0.vlm"},
	{0x7A6ECD248A587AA7ULL, "vxw_blightlands_d0_0_0.vlm"},
	{0x7A6FCD7C4BC0B4EFULL, "mrl310_merrill6202134.cl"},
	{0x7A70BF4905B5FBF6ULL, "230872857.ogg"},
	{0x7A711459914091B6ULL, "hlac_ht_0_hightown_a3_-65537_0y.dds"},
	{0x7A718058DFC3DA46ULL, "hlac_dr_primevaltomb_659_7y.dds"},
	{0x7A722C553D0FCFD4ULL, "lt_0_lowtown_n_43_0_ll.mmh"},
	{0x7A7300E77D039C1BULL, "lt_0_lowtown_docks_n_1_0_ll.mmh"},
	{0x7A7318367E531A88ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_007a_0d.dds"},
	{0x7A733B6E17A62B9FULL, "hlac_lt_hangedman_n_-1_11y.dds"},
	{0x7A73613E85C5A4F2ULL, "6167860_m.fxe"},
	{0x7A7441990FDDD5A2ULL, "gen00fl_fenris6106609.cl"},
	{0x7A74495481772293ULL, "vxht_bartrand_n_684_0_0.vlm"},
	{0x7A7467534F07CA94ULL, "vxht_mansion_n_345_0_0.vlm"},
	{0x7A7476BE08111E80ULL, "mage_cleansingaura.dds"},
	{0x7A74E3BBCE319792ULL, "223048175.ogg"},
	{0x7A75387AFC045E81ULL, "6090238_f.fxe"},
	{0x7A757509340E15F8ULL, "vxca_caverns2_8rd_0_0.vlm"},
	{0x7A75A7472006E865ULL, "6081746_f.fxe"},
	{0x7A76A95F7C0ABBB1ULL, "546764063.ogg"},
	{0x7A76E0B820FB9F4DULL, "519915094.ogg"},
	{0x7A77663B046593D6ULL, "6020651_m.fxe"},
	{0x7A780DDCD338B38AULL, "lt_blackemporium_mmi.gff"},
	{0x7A780F432CF37576ULL, "w_encounter2_camp_36_0.mao"},
	{0x7A781B432CF38915ULL, "w_encounter2_camp_36_0.mmh"},
	{0x7A781D432CF38C8BULL, "w_encounter2_camp_36_0.msh"},
	{0x7A7821691D68C34CULL, "art\\vfx\\combat\\vfx_glb_spi_imp_maj_c.anb"},
	{0x7A798B588DDA1416ULL, "vxht_0_hightown_a2_c2z_0_0.vlm"},
	{0x7A7A0F31CC8FF793ULL, "fx_stormcloud.mao"},
	{0x7A7A2DDCBED6A19EULL, "6127597_f.fxe"},
	{0x7A7AC377ADF44575ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden4.msh"},
	{0x7A7AC577ADF448EBULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden4.mmh"},
	{0x7A7C3B75121FF7AEULL, "zz_mcr_debug6071551.cl"},
	{0x7A7C7D7A36E0FCEDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_2arm_shrug.ani"},
	{0x7A7D0F16465C973CULL, "ht_0_hightown_mmi.gff"},
	{0x7A7D32499962A454ULL, "535524329.ogg"},
	{0x7A7D3F017C4E9327ULL, "vxw_blightlands_i3_0_0.vlm"},
	{0x7A7D7AA0D219EA95ULL, "6075229_m.fxe"},
	{0x7A7D84C6ABCFC2D1ULL, "6076358_m.fxe"},
	{0x7A7DA8DC1F246FEFULL, "login.gfx"},
	{0x7A7DBE535F710714ULL, "lgt210cr_shp_kill_qunari2.utc"},
	{0x7A7DC6E2CC536B40ULL, "vxht_0_hightown_a3_c32_0_0.vlm"},
	{0x7A7DE579926DE619ULL, "93410679.ogg"},
	{0x7A7E6274E382A42AULL, "vxw_encounter2_camp_1y_0_0.vlm"},
	{0x7A7E8DD993166B67ULL, "vxlt_0_lowtown_a2_df0_0_0.vlm"},
	{0x7A7EE4ADBB4B3A04ULL, "6051059_m.fxe"},
	{0x7A7F5FA87C18AF24ULL, "939811201.ogg"},
	{0x7A7F963C0C54F6AAULL, "llac_ht_mansion2_818_55.dds"},
	{0x7A7FA1BBD8158319ULL, "vxht_bartrand_n_4h4_0_0.vlm"},
	{0x7A7FFA6C7F33DFD7ULL, "vxlt_0_lowtown_docks_n_as0_0_0.vlm"},
	{0x7A805AC31A65FFBEULL, "gen00fl_aveline6140071.cl"},
	{0x7A80D57F71A0C1FFULL, "one000_ilen6205694.cl"},
	{0x7A80EA21C59DE90CULL, "mag360_thrask6043645.cl"},
	{0x7A81BE2C17E2350FULL, "llac_lt_hangedman_-1_19.dds"},
	{0x7A81ED9C35B31B38ULL, "hlac_ht_mansion2b_818_50y.dds"},
	{0x7A81FE1AB2519A9BULL, "534654761.ogg"},
	{0x7A82A77FC467B32EULL, "6155472_m.fxe"},
	{0x7A82EB39B7CC8707ULL, "mag360_thrask6044296.cl"},
	{0x7A833702193AD72AULL, "6155577_m.fxe"},
	{0x7A833A1F3341AD00ULL, "vxht_mansion2b_2ny_0_0.vlm"},
	{0x7A8359540F2AD803ULL, "vxw_sundermount_2z6_0_0.vlm"},
	{0x7A83BC2A5DD25DF0ULL, "gen_im_arm_hlm_hvy_act3_02.uti"},
	{0x7A83D55C73A2B583ULL, "hlac_w_bonepit_262146_0x.dds"},
	{0x7A842960592CAB54ULL, "788745900.ogg"},
	{0x7A8434F744DB2BD2ULL, "m_dae210ar_lowtown_nt.wbd"},
	{0x7A84475BDE101377ULL, "9675970.ogg"},
	{0x7A845138B0D13BE5ULL, "6095488_f.fxe"},
	{0x7A84654A51FC761BULL, "575567370.ogg"},
	{0x7A848F80DFBDD6DAULL, "440037695.ogg"},
	{0x7A8525457DA11A76ULL, "vxw_bonepitmining_a2_21s_0_0.vlm"},
	{0x7A859DE8C6860021ULL, "vxht_0_hightown_a3_c82_0_0.vlm"},
	{0x7A85BBE0D448BB78ULL, "19032078.ogg"},
	{0x7A869F75A7687D42ULL, "6071122_f.fxe"},
	{0x7A87060E8D6CDCE3ULL, "319198682.ogg"},
	{0x7A87242538474993ULL, "6143180_m.fxe"},
	{0x7A87D24F87AA8CF4ULL, "mcr310_argument6072300.cl"},
	{0x7A88282B9DA645C3ULL, "qcr000_arishok6060175.cl"},
	{0x7A895568CEA9CAEDULL, "805347.ogg"},
	{0x7A89775EEAB0BBFBULL, "728193997.ogg"},
	{0x7A898660AFBAF870ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuary.mao"},
	{0x7A898A60AFBAFE83ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuary.mmh"},
	{0x7A899860AFBB16DDULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuary.msh"},
	{0x7A89D447D31ACE41ULL, "345997585.ogg"},
	{0x7A8A82D4D766CE23ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.impa.evt"},
	{0x7A8B1AA9D6416439ULL, "vxht_mansion2_1v3_0_0.vlm"},
	{0x7A8B2FCC7BC14781ULL, "genip_cin_moneypursea.utp"},
	{0x7A8B7A2352706D84ULL, "vxlt_gallowstemplar_n_5oq_0_0.vlm"},
	{0x7A8B94A32C8DD72BULL, "one300_zevran6089769.cl"},
	{0x7A8BB0DE6E44ABA7ULL, "vxw_blightlands_r2_0_0.vlm"},
	{0x7A8BBE98ECF41B48ULL, "lt_0_lowtown_docks_231_0.mmh"},
	{0x7A8BC7FAEA9C638CULL, "w_sundermount_9_0.msh"},
	{0x7A8BCA98ECF42FEBULL, "lt_0_lowtown_docks_231_0.mao"},
	{0x7A8BCDFAEA9C6E7EULL, "w_sundermount_9_0.mmh"},
	{0x7A8BD9FAEA9C8215ULL, "w_sundermount_9_0.mao"},
	{0x7A8BDC98ECF44E72ULL, "lt_0_lowtown_docks_231_0.msh"},
	{0x7A8C38BFBE4983C0ULL, "vxht_0_hightown_a3_c43_0_0.vlm"},
	{0x7A8C8C125925A978ULL, "vxlt_gallowstemplar_5oz_0_0.vlm"},
	{0x7A8D72C6ADD5A153ULL, "vxht_bartrand_n_5qh_0_0.vlm"},
	{0x7A8D8548E0B3C3DBULL, "dae000_oldmangossip6174774.cl"},
	{0x7A8DBF580351242FULL, "llac_lt_0_lowtown_65536_3.dds"},
	{0x7A8E00A564C46A0FULL, "dae000_quintus6154733.cl"},
	{0x7A8E40087FC5F4C2ULL, "art\\characters\\weapons\\staff\\w_stf_sf01a.msh"},
	{0x7A8E40E142FB239EULL, "vxht_mansion2_1aq_0_0.vlm"},
	{0x7A8E42087FC5F878ULL, "art\\characters\\weapons\\staff\\w_stf_sf01a.mmh"},
	{0x7A8E4E087FC60C1BULL, "art\\characters\\weapons\\staff\\w_stf_sf01a.mao"},
	{0x7A8EAA1F6D6F18BDULL, "502155177.ogg"},
	{0x7A8ECF416967097AULL, "164303557.ogg"},
	{0x7A8ED83B95CF5E0CULL, "191286432.ogg"},
	{0x7A8FC4C28F1C458EULL, "338289012.ogg"},
	{0x7A8FE70F8F5BF824ULL, "vxw_woundedcoast_nk_0_0.vlm"},
	{0x7A9021BF636BDC4BULL, "vxdr_primevaltomb_7sq_0_0.vlm"},
	{0x7A903271BD334055ULL, "lt_0_lowtown_27_0_ll.mmh"},
	{0x7A90789D914D16E2ULL, "283192876.ogg"},
	{0x7A9084FCFC841D42ULL, "art\\vfx\\combat\\vfx_a_disengage_act_c.impa.gad"},
	{0x7A90950EDF2C1C64ULL, "6121910_m.fxe"},
	{0x7A91104FE854D6ECULL, "vxca_caverns2prc_8xk_0_0.vlm"},
	{0x7A9128A188D8E2C1ULL, "vxdr_primevalentrance_2e8_0_0.vlm"},
	{0x7A91545E1F675F5AULL, "character_character_p124.psh.fxo"},
	{0x7A915BD4D76CAAA4ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.impa.gad"},
	{0x7A917B4E6027CF55ULL, "qun221cr_petrice_mob.utc"},
	{0x7A92153294BF6D76ULL, "6150307_m.fxe"},
	{0x7A9243993EA9C063ULL, "612414009.ogg"},
	{0x7A92596D74ABBC43ULL, "seb221ip_door_block_chair.utp"},
	{0x7A9369591AA62BEBULL, "352610929.ogg"},
	{0x7A936B23707AFB4EULL, "hlac_lt_0_lowtown_a2_n_c_131070_1y.dds"},
	{0x7A939F2E20DA41F9ULL, "hlac_lt_0_lowtown_a2_n_-65537_0z.dds"},
	{0x7A93BC0219EF736DULL, "182964292.ogg"},
	{0x7A9429C23EB9A9FCULL, "lt_0_lowtown_13_0.rml"},
	{0x7A94690768A1A2E1ULL, "lowtown_fem_urchin_mob6165931.cl"},
	{0x7A948E5BE71F68FFULL, "6075810_m.fxe"},
	{0x7A95BEE397A779BCULL, "hlac_lt_backalley_g_0_12y.dds"},
	{0x7A96125CEBC75A8DULL, "6099502_m.fxe"},
	{0x7A96486A86CD2972ULL, "6051605_f.fxe"},
	{0x7A9648EC028F9D3EULL, "vxht_bartrand_n_5ob_0_1.vlm"},
	{0x7A9697D519CDBC15ULL, "vxht_0_hightown_a3_c19_0_0.vlm"},
	{0x7A96B1DE5F72B161ULL, "gui_guipixelshader_p9.psh.fxo"},
	{0x7A96F37C6F2A3FD2ULL, "vxht_bartrand_n_5qm_0_0.vlm"},
	{0x7A97993A085BC897ULL, "6165386_m.fxe"},
	{0x7A97A7FCFC8A7739ULL, "art\\vfx\\combat\\vfx_a_disengage_act_c.impa.evt"},
	{0x7A97C0BEF75CBE7DULL, "6174193_m.fxe"},
	{0x7A980A3E8FAB3C2CULL, "6113831_f.fxe"},
	{0x7A986AD4D772E282ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.impa.ani"},
	{0x7A98A1421E52BD48ULL, "6087225_m.fxe"},
	{0x7A98D9D0298B0E4CULL, "vxht_bartrand_n_5vz_0_0.vlm"},
	{0x7A991ECC33BB5073ULL, "gen00fl_fenris6126891.cl"},
	{0x7A9943E80AABBF1AULL, "llac_ca_caverns1_mines_24424_17.dds"},
	{0x7A99EC0044668604ULL, "vxlt_gallowstemplar_n_5bv_0_0.vlm"},
	{0x7A9A5B46AE7A5006ULL, "6085419_f.fxe"},
	{0x7A9A73F64694C910ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1rockpile.phy"},
	{0x7A9BE4A39FCBB9D3ULL, "vxht_bartrand_n_5tk_0_0.vlm"},
	{0x7A9C070B0CBC9C33ULL, "vxlt_0_lowtown_docks_a3_n_d1d_0_0.vlm"},
	{0x7A9C3FB28FB33F87ULL, "hlac_lt_0_lowtown_a3_n_c_65535_2x.dds"},
	{0x7A9C87399C3873EAULL, "dae200cr_viscount6087257.cl"},
	{0x7A9CB2BE34F13C1EULL, "vxht_mansion2_1dr_0_0.vlm"},
	{0x7A9D16B64646AEE2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruinbrick.msh"},
	{0x7A9D18B64646B198ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruinbrick.mmh"},
	{0x7A9D24B64646C63BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruinbrick.mao"},
	{0x7A9D3AB9BE65D224ULL, "6117018_m.fxe"},
	{0x7A9D7A87C2804865ULL, "6089953_f.fxe"},
	{0x7A9D8D6D22A29483ULL, "231328929.ogg"},
	{0x7A9DC751C6F5E2EAULL, "hm_ruxton_harimann.mor"},
	{0x7A9E164E5D89BE54ULL, "vxht_mansion2b_1t0_0_0.vlm"},
	{0x7A9E83B0119ADE5EULL, "6016500_m.fxe"},
	{0x7A9E88BD1B2671B7ULL, "t3_arm_thg.tnt"},
	{0x7A9EAA6B39ED881CULL, "hlac_lt_0_lowtown_131071_7y.dds"},
	{0x7A9F4732AAA9133EULL, "369605029.ogg"},
	{0x7AA05C8957904CF1ULL, "one100_party_comments6106775.cl"},
	{0x7AA0677028AD994BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robc_n.dds"},
	{0x7AA163AA838C723BULL, "1053002541.ogg"},
	{0x7AA2128D08A8C2EBULL, "zz_mag_debug6039845.cl"},
	{0x7AA251183F7D1C76ULL, "vxlt_0_lowtown_a2_n_fe1_0_0.vlm"},
	{0x7AA281A2AA723121ULL, "vxht_0_hightown_a3_c60_0_0.vlm"},
	{0x7AA2AC0DA2A8EEDBULL, "pro000_colonel_ewald6040548.cl"},
	{0x7AA2DBE1D88DCE8DULL, "897794988.ogg"},
	{0x7AA2F67CD8E34F89ULL, "vxw_blightlands_l6_0_0.vlm"},
	{0x7AA33D0522889ED8ULL, "mag210_arianni6024559.cl"},
	{0x7AA3DC78DC7048D7ULL, "12927790.ogg"},
	{0x7AA448BB6D9D5716ULL, "hlac_ht_0_hightown_a2_131071_0z.dds"},
	{0x7AA4AFFCFC950F28ULL, "art\\vfx\\combat\\vfx_a_disengage_act_c.impa.ani"},
	{0x7AA4C1CDA87D6028ULL, "696229044.ogg"},
	{0x7AA52520CAE78743ULL, "vxlt_hangedman_n_21c_0_1.vlm"},
	{0x7AA52F21AE902791ULL, "hlac_w_encounter2_camp_196612_0x.dds"},
	{0x7AA5655961205852ULL, "vxht_bartrand_n_5tn_0_0.vlm"},
	{0x7AA5DE8B4E36B6C5ULL, "6171136_m.fxe"},
	{0x7AA623CAD8594B4DULL, "6172263_m.fxe"},
	{0x7AA6569FDABDDA5AULL, "6133735_m.fxe"},
	{0x7AA69E234A657212ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.cess.gad"},
	{0x7AA6DC122494C8DFULL, "gen00fl_anders6113493.cl"},
	{0x7AA6F72F721BAA3FULL, "lt_0_lowtown_docks_a3_38_0.anb"},
	{0x7AA78580E37DF714ULL, "llac_lt_gallowstemplar_f_2390_1.dds"},
	{0x7AA846E228AFE244ULL, "dae000cr_grand_cleric6147902.cl"},
	{0x7AA8617CCDE20AD0ULL, "61424007.ogg"},
	{0x7AA920A5CF46EFD8ULL, "6208575_f.fxe"},
	{0x7AA96D6D39E2BDF0ULL, "90067680.ogg"},
	{0x7AA96FCC3D11DA78ULL, "vxlt_gallowstemplar_5ux_0_0.vlm"},
	{0x7AAA7A956B57D685ULL, "6102884_m.fxe"},
	{0x7AAB1A79A79B67A9ULL, "dre000_hireling16038588.cl"},
	{0x7AAC7332412973A0ULL, "901569644.ogg"},
	{0x7AAC8B8D6D0AC7C1ULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.anb"},
	{0x7AAC9431D39E7C01ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb103.mao"},
	{0x7AACE06CA3960E9FULL, "6137543_m.fxe"},
	{0x7AAD03077E386315ULL, "6163759_m.fxe"},
	{0x7AAD4196863D6CB9ULL, "vxlt_0_lowtown_docks_a3_n_bsr_0_0.vlm"},
	{0x7AADA1234A6B9649ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.cess.evt"},
	{0x7AADA4BF0D04FDE6ULL, "984950492.ogg"},
	{0x7AADEF9D005F2A94ULL, "6152762_f.fxe"},
	{0x7AAE6C12D9CD9128ULL, "init.wbd"},
	{0x7AAE80020F2A4C2CULL, "6075444_m.fxe"},
	{0x7AAE8FB60923F680ULL, "vxht_mansion2b_2ad_0_0.vlm"},
	{0x7AAF0FB983BEA88CULL, "655081072.ogg"},
	{0x7AAF3D3B078003A7ULL, "ts_screenshot.ncs"},
	{0x7AAFB1AC04AAA04FULL, "719268569.ogg"},
	{0x7AAFF643957E10C4ULL, "6102662_f.fxe"},
	{0x7AB094C03EB9B5E4ULL, "framebuffereffects_downsample_p0.psh.fxo.dx11"},
	{0x7AB0BF81B5CEF386ULL, "419786230.ogg"},
	{0x7AB16859CAD96809ULL, "vxw_blightlands_q5_0_0.vlm"},
	{0x7AB1F7B17B2603D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_7.ani"},
	{0x7AB21AE93E1EB7F0ULL, "6098315_m.fxe"},
	{0x7AB2CF15A8B1D0D4ULL, "hlac_ht_brothel_n_2995_1y.dds"},
	{0x7AB3D55AA5325D26ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_enter.evt"},
	{0x7AB3FD650FA435D0ULL, "vxw_bonepitmining_a2_fp_0_0.vlm"},
	{0x7AB4DC86B5040DA4ULL, "vxht_0_hightown_a2_c99_0_0.vlm"},
	{0x7AB640383800DFD5ULL, "vxlt_gallowstemplar_4wr_0_0.vlm"},
	{0x7AB6AC5C0B9D2CB2ULL, "combat_smaoe_r2_a2.gda"},
	{0x7AB6C545AF054764ULL, "397579857.ogg"},
	{0x7AB703F83AEA9C40ULL, "hlac_ca_caverns1_24424_13y.dds"},
	{0x7AB739648E04194DULL, "528599847.ogg"},
	{0x7AB743B49D3AA5A6ULL, "849018240.ogg"},
	{0x7AB77B5BB06EEBF7ULL, "395687997.ogg"},
	{0x7AB802B17B2A8273ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_7.gad"},
	{0x7AB861B1BB2869DBULL, "130118386.ogg"},
	{0x7AB9064C04C87C6FULL, "6051816_m.fxe"},
	{0x7AB9C34AE9774334ULL, "6027850_m.fxe"},
	{0x7AB9EC5AA536F071ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_enter.gad"},
	{0x7AB9F68BD0A5CC82ULL, "108851434.ogg"},
	{0x7AB9FE7B829C26C6ULL, "6156626_m.fxe"},
	{0x7ABA13DCCA6DE51FULL, "hlac_lt_darktown_-1_0y.dds"},
	{0x7ABA4D6C0C1A1FD5ULL, "vxht_0_hightown_a3_avn_0_0.vlm"},
	{0x7ABA6C8D6D16D03AULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.mmh"},
	{0x7ABB14EC5A1A8745ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_d.dds"},
	{0x7ABB89234A77AAB8ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.cess.ani"},
	{0x7ABB8C1352539901ULL, "gen00fl_dog6107572.cl"},
	{0x7ABBD47B31C9D7E0ULL, "llac_lt_0_lowtown_a2_65538_4.dds"},
	{0x7ABC1CA9E1B9A036ULL, "6078328_m.fxe"},
	{0x7ABC68FB987D73FDULL, "vxht_0_hightown_n_c79_0_0.vlm"},
	{0x7ABD01CDE4766257ULL, "t3_wpn_rlr.tnt"},
	{0x7ABD232CDB0F8A2DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2_tool.msh"},
	{0x7ABD312CDB0FA260ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2_tool.mao"},
	{0x7ABD352CDB0FA8B3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2_tool.mmh"},
	{0x7ABE8DA63F70179DULL, "vxlt_0_lowtown_docks_bqh_0_0.vlm"},
	{0x7ABE905D55C16B47ULL, "qun200_party_comment6090443.cl"},
	{0x7ABEB27CBAE3B378ULL, "hlac_w_bonepitmining_a2_65542_0z.dds"},
	{0x7ABEFFB17B309BD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_7.evt"},
	{0x7ABF15927919B38DULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mere.msh"},
	{0x7ABF27927919D213ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mere.mmh"},
	{0x7AC011AD43B56772ULL, "6181918_m.fxe"},
	{0x7AC01F99C77EE18FULL, "hlac_lt_undercitydungeon_-1_42x.dds"},
	{0x7AC0D4F2F77BBEA4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masa_0n.dds"},
	{0x7AC0ED5AA53D101FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_enter.ani"},
	{0x7AC11332E724800CULL, "6018679_f.fxe"},
	{0x7AC12690A53FF672ULL, "ave310_donnic6134767.cl"},
	{0x7AC17BEB2435A95AULL, "art\\characters\\weapons\\dagger\\w_dgr_103.mao"},
	{0x7AC1ECCC54C66DE7ULL, "dae000ip_crowd_ff5_talk14.utp"},
	{0x7AC1F51541D8442FULL, "lt_gallowstemplar_f_8165.anb"},
	{0x7AC25298E9F38DD0ULL, "6135300_m.fxe"},
	{0x7AC26F42019A4E50ULL, "vxw_bonepitmining_a2_mq_0_0.vlm"},
	{0x7AC335A37DA2D69DULL, "llac_lt_undercityencounter_-1_27.dds"},
	{0x7AC347C3236015DBULL, "hlac_w_encounter3_-327680_0y.dds"},
	{0x7AC3A163F6437C37ULL, "zz_followers_banter6158655.cl"},
	{0x7AC3B95EDA43F0F5ULL, "vxw_bonepitmining_a2_6a_0_0.vlm"},
	{0x7AC42D83E0A1535EULL, "722473546.ogg"},
	{0x7AC47F1077649B88ULL, "mag211_marethari6025332.cl"},
	{0x7AC491D076AC8B9DULL, "w_sundermount_162_0.phy"},
	{0x7AC4E3B76B00C843ULL, "6053924_m.fxe"},
	{0x7AC4F5EE301C692BULL, "549994979.ogg"},
	{0x7AC510166389F967ULL, "288068090.ogg"},
	{0x7AC57969F6B7150FULL, "vxlt_gallowstemplar_f_45u_0_0.vlm"},
	{0x7AC659F9742BAED5ULL, "6087031_m.fxe"},
	{0x7AC66902E36AEE7EULL, "vxht_mansion_n_6os_0_0.vlm"},
	{0x7AC6A0870E0F2C94ULL, "6083712_f.fxe"},
	{0x7AC6F2EF8BA60452ULL, "llac_ca_caverns1_24423_11.dds"},
	{0x7AC75806D4936162ULL, "gen_im_npc_arm_hum_hlm_mrca.uti"},
	{0x7AC7C343E8ED2799ULL, "979517003.ogg"},
	{0x7AC803CFE2ED7538ULL, "vxca_cave_19c_0_0.vlm"},
	{0x7AC81622D8FCF665ULL, "vxht_0_hightown_a3_c1i_0_0.vlm"},
	{0x7AC87F2926A96290ULL, "gen00fl_isabela6135601.cl"},
	{0x7AC88B32B7CC758BULL, "art\\characters\\playercharacter\\humanfemale\\hf_arm_flmb.mao"},
	{0x7AC8AB312C33996DULL, "hlac_lt_hangedman_n_-1_18x.dds"},
	{0x7AC8F87AC40135A3ULL, "l_dae110ar_lowtown.wbd"},
	{0x7AC96784CD398E69ULL, "396758506.ogg"},
	{0x7AC97D5240719378ULL, "zz_orsino.ncs"},
	{0x7AC985BD1156D828ULL, "hlac_lt_0_lowtown_-65537_0z.dds"},
	{0x7AC9AE8357345A24ULL, "vxw_woundedcoast_zo_0_0.vlm"},
	{0x7ACA1E16E6948901ULL, "clf260_harley6069676.cl"},
	{0x7ACA89A814152AD6ULL, "432069350.ogg"},
	{0x7ACA8AB462DB458BULL, "6140624_m.fxe"},
	{0x7ACB4185CDAA465FULL, "llac_lt_0_lowtown_a2_-65536_4.dds"},
	{0x7ACB51E2A6A76B51ULL, "644874329.ogg"},
	{0x7ACB58414D4E68D6ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.impa.evt"},
	{0x7ACB5BAD8E6E659BULL, "vxlt_gallowstemplar_n_2p8_0_0.vlm"},
	{0x7ACB7807BBEAAAA8ULL, "6088407_f.fxe"},
	{0x7ACB9234B727A5DBULL, "qun220cr_javaris6068902.cl"},
	{0x7ACBBB42A8EDC003ULL, "6133598_m.fxe"},
	{0x7ACC14D25A1B82BEULL, "608996682.ogg"},
	{0x7ACC7E3F07A2388CULL, "vxw_sundermount_203_0_0.vlm"},
	{0x7ACCCBAF2BD65FA4ULL, "vxw_bonepit_1va_0_0.vlm"},
	{0x7ACCE3C162553548ULL, "864225889.ogg"},
	{0x7ACCFF833166301DULL, "vxlt_0_lowtown_docks_btg_0_0.vlm"},
	{0x7ACD2228DA19E8B8ULL, "vxca_caverns2prc_61j_0_0.vlm"},
	{0x7ACD3F6D79AF30E1ULL, "vxlt_0_lowtown_a2_n_c_f58_0_0.vlm"},
	{0x7ACD4051F70EE238ULL, "6079268_m.fxe"},
	{0x7ACDC38650788808ULL, "one350_vinell6087503.cl"},
	{0x7ACE143D35F01CBFULL, "hlac_w_woundedcoast_327684_0z.dds"},
	{0x7ACE27804E2FECD9ULL, "6115013_m.fxe"},
	{0x7ACED0C88FC711FAULL, "gen00fl_fenris6116536.cl"},
	{0x7ACF14EAEE268E90ULL, "vxht_chantry_26q_0_0.vlm"},
	{0x7ACF3C54853C8295ULL, "mag330_walter6045863.cl"},
	{0x7ACF49D7A8D16553ULL, "vxht_keep_4i5_0_0.vlm"},
	{0x7ACF9B4DBE296806ULL, "19863889.ogg"},
	{0x7ACFB7E4490AE71DULL, "vxlt_gallowscourtyard_7fl_0_0.vlm"},
	{0x7ACFEE6452C6E9DBULL, "gen_im_acc_amu_lore_01.uti"},
	{0x7AD0352E31767E64ULL, "293679440.ogg"},
	{0x7AD0C475DBE9A650ULL, "6130000_m.fxe"},
	{0x7AD0FB867F4A1B11ULL, "hlac_ht_keep_3734_1x.dds"},
	{0x7AD1075FBB84C91DULL, "806991378.ogg"},
	{0x7AD195B7CD4EC7B5ULL, "vxca_caverns2prc_aep_0_0.vlm"},
	{0x7AD21619B586D7C0ULL, "hlac_ht_0_hightown_65534_0z.dds"},
	{0x7AD24F414D547861ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.impa.gad"},
	{0x7AD2925B1B1C6A43ULL, "6091882_f.fxe"},
	{0x7AD2D1EAF2658C18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\bow.anb"},
	{0x7AD34E12A5FC497FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_standing.anb"},
	{0x7AD3ACE91349DF11ULL, "6015058_m.fxe"},
	{0x7AD3ADFAEAD9ED35ULL, "w_sundermount_9_0.rml"},
	{0x7AD43F3F1B8B213CULL, "cd_qun221_braggart.wbd"},
	{0x7AD441793629F54FULL, "lt_0_lowtown_docks_165_0.mao"},
	{0x7AD44A684226630DULL, "llac_lt_undercitydungeon_-1_17.dds"},
	{0x7AD44D79362A09ECULL, "lt_0_lowtown_docks_165_0.mmh"},
	{0x7AD45379362A13DEULL, "lt_0_lowtown_docks_165_0.msh"},
	{0x7AD468913E0915A2ULL, "hf_amb_chanter_taletha.mor"},
	{0x7AD49E617C786787ULL, "142720061.ogg"},
	{0x7AD4BDE9E14EF0D8ULL, "vxlt_warehouse_10j_0_0.vlm"},
	{0x7AD52F4B65296707ULL, "llac_lt_0_lowtown_docks_a3_n_131071_0.dds"},
	{0x7AD55EF2B1A7B1A9ULL, "hlac_ca_caverns2_5802_66z.dds"},
	{0x7AD586D076BAEC09ULL, "w_sundermount_162_0.mao"},
	{0x7AD592D076BB00A2ULL, "w_sundermount_162_0.mmh"},
	{0x7AD594D076BB0458ULL, "w_sundermount_162_0.msh"},
	{0x7AD5E297C6A3EBCFULL, "distortionvfx_cs.vsh.fxo"},
	{0x7AD69E83A9C96D18ULL, "6074447_m.fxe"},
	{0x7AD75E20920AEC1EULL, "mer220_hubert6041208.cl"},
	{0x7AD76381C7C1558BULL, "vxht_chantry_2t8_0_0.vlm"},
	{0x7AD7D0C6D2CE7718ULL, "357938573.ogg"},
	{0x7AD7E2157CF5B17CULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.dura.ani"},
	{0x7AD7E5CB5D654BC5ULL, "hlac_lt_0_lowtown_n_131070_5z.dds"},
	{0x7AD80DAA31441251ULL, "6028810_m.fxe"},
	{0x7AD8108E78E540B4ULL, "83334191.ogg"},
	{0x7AD82060492A72A4ULL, "vxw_woundedcoast_wl_0_0.vlm"},
	{0x7AD820DE3DA16CBAULL, "mer120_hubert6059398.cl"},
	{0x7AD850414D58E50FULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.impa.ani"},
	{0x7AD8CA8D6A2603D2ULL, "vxht_keep_4i0_0_0.vlm"},
	{0x7AD94CB57C69A4FDULL, "vxht_0_hightown_n_c97_0_0.vlm"},
	{0x7AD950D211C42E35ULL, "vxlt_0_lowtown_a2_n_c_hnk_0_0.vlm"},
	{0x7ADA04B591E2BE8DULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_workstool.mmh"},
	{0x7ADA08B591E2C5DEULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_workstool.mao"},
	{0x7ADA16B591E2DD13ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_workstool.msh"},
	{0x7ADA3279BAB06358ULL, "258671843.ogg"},
	{0x7ADA37F5DAB8F8C7ULL, "202631812.ogg"},
	{0x7ADA5C1059851386ULL, "one100dg_kelder6060536.cl"},
	{0x7ADA60212B91D45BULL, "6098547_m.fxe"},
	{0x7ADA92B66FF073FBULL, "vxht_brothel_n_62z_0_0.vlm"},
	{0x7ADAF01BF998510CULL, "vxw_sundermount_250_0_0.vlm"},
	{0x7ADB07F03121D3D3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cartflat.phy"},
	{0x7ADB179097A6F27CULL, "hlac_ht_0_hightown_-196607_0y.dds"},
	{0x7ADB5DABF82E8344ULL, "vxlt_warehouse_18n_0_0.vlm"},
	{0x7ADC9CE801963F62ULL, "6041834_m.fxe"},
	{0x7ADCAE94654EB64BULL, "gen_im_trs_gen_shd_wdn_r1.uti"},
	{0x7ADD02D34EB8E66AULL, "mag170_thrask6016961.cl"},
	{0x7ADD03261E9FF56EULL, "vxlt_gallowscourtyard_n_6qt_0_0.vlm"},
	{0x7ADD1C680D2C0D97ULL, "mrl251im_amu_radha.uti"},
	{0x7ADD5D451AF716E3ULL, "406072268.ogg"},
	{0x7ADD8AB8CBF25D05ULL, "vxht_0_hightown_bye_0_0.vlm"},
	{0x7ADDBBB49AC77DD3ULL, "vxht_keep_4n4_0_0.vlm"},
	{0x7ADDE8DA2C020E5CULL, "933041358.ogg"},
	{0x7ADE2CCB70343273ULL, "hlac_ca_caverns1_24424_23y.dds"},
	{0x7ADE3E4FF6BADDB2ULL, "llac_w_bonepit_327684_0.dds"},
	{0x7ADED02B383D6094ULL, "isa311_isabela6092292.cl"},
	{0x7ADEEA27FE23E93EULL, "hlac_lt_gallowscourtyard_-196608_0y.dds"},
	{0x7ADEEF157CFBE5BEULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.dura.gad"},
	{0x7ADF2CCD28F8BA52ULL, "vxht_bartrand_n_5hj_0_0.vlm"},
	{0x7ADF4CE7C5ADB819ULL, "6134874_m.fxe"},
	{0x7ADF58D2B86ADB21ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_roba_0n.dds"},
	{0x7ADF6220E2FC1186ULL, "qun110_petrice6059148.cl"},
	{0x7ADF7A6AD0527377ULL, "6169672_m.fxe"},
	{0x7ADFAF1FA847B94EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangb.mmh"},
	{0x7ADFBB1FA847CDE5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangb.mao"},
	{0x7ADFC91FA847E53CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarphangb.msh"},
	{0x7AE04536B2712586ULL, "vxht_mansion_n_6me_0_0.vlm"},
	{0x7AE09D18BE3021F5ULL, "vxw_bonepitmining_a2_8o_0_0.vlm"},
	{0x7AE0B360827E518DULL, "vxlt_shop_ch_0_0.vlm"},
	{0x7AE106A182BA8E17ULL, "101265004.ogg"},
	{0x7AE145FDA8D6A34CULL, "vxdr_primevaltomb_910_0_2.vlm"},
	{0x7AE14FCE9F2A1D62ULL, "29083200.ogg"},
	{0x7AE1ACDA25A79E6DULL, "525218822.ogg"},
	{0x7AE1C4357D51A1F8ULL, "jubjub.mao"},
	{0x7AE203EE9B574497ULL, "666072633.ogg"},
	{0x7AE204D1B9F20358ULL, "mcr370_meredith6136157.cl"},
	{0x7AE209E130FA9EF4ULL, "gen00fl_aveline6139657.cl"},
	{0x7AE2670526C48D9DULL, "dae000_gen_turn_in_56198675.cl"},
	{0x7AE281D4CD8F2177ULL, "w_bonepitmining_a2_144_0.mmh"},
	{0x7AE285D4CD8F2784ULL, "w_bonepitmining_a2_144_0.mao"},
	{0x7AE297D4CD8F46E9ULL, "w_bonepitmining_a2_144_0.msh"},
	{0x7AE2A71FD7AB8403ULL, "81437495.ogg"},
	{0x7AE2C19C6933B4C3ULL, "qun120cr_javaris6043975.cl"},
	{0x7AE2E6E6ECB66A7FULL, "vxca_caverns2prc_8ol_0_1.vlm"},
	{0x7AE30A04E2713003ULL, "llac_w_bonepitmining_a2_458758_0.dds"},
	{0x7AE30CA679752192ULL, "313080636.ogg"},
	{0x7AE3243ADAEC28E6ULL, "6180096_m.fxe"},
	{0x7AE362A55226927EULL, "6037075_m.fxe"},
	{0x7AE3644DECA99F4EULL, "gen00fl_anders6111359.cl"},
	{0x7AE39AD22838DFA4ULL, "hlac_ca_caverns2prc_5802_38y.dds"},
	{0x7AE39AD45350E987ULL, "6107298_m.fxe"},
	{0x7AE4355E05890F44ULL, "hlac_dr_primevaltomb_657_0x.dds"},
	{0x7AE44EA648A9B3B9ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_orsino_alt_.ani"},
	{0x7AE4705DD71A8ACEULL, "vxlt_gallowstemplar_f_458_0_0.vlm"},
	{0x7AE4837D79DB01C1ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_8z.dds"},
	{0x7AE4E57B045323A7ULL, "365944273.ogg"},
	{0x7AE55835507D1025ULL, "hlac_w_encounter2_131070_0y.dds"},
	{0x7AE5CA157D01C5DDULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.dura.evt"},
	{0x7AE5E037B19331D6ULL, "hlac_lt_0_lowtown_docks_-2_5z.dds"},
	{0x7AE6BB5F5B3D6FF6ULL, "art\\vfx\\combat\\vfx_glb_charm_c.dura.ani"},
	{0x7AE6E7EA06DE85A8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_basket_d.dds"},
	{0x7AE6F14F599017E1ULL, "400006511.ogg"},
	{0x7AE73474435B81F2ULL, "6051628_m.fxe"},
	{0x7AE751DF6380B14EULL, "mag120_macha6087848.cl"},
	{0x7AE76EABFC67F2ACULL, "vxlt_gamlenshouse_n_l0_0_0.vlm"},
	{0x7AE80D5896C4CCF6ULL, "llac_lt_0_lowtown_a2_n_131071_3.dds"},
	{0x7AE877EB74153BC6ULL, "vxht_0_hightown_n_an3_0_0.vlm"},
	{0x7AE8AB4BFC81C411ULL, "mrl000_marethari6129464.cl"},
	{0x7AE8BB4BF7CF8984ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_calf_stretch.ani"},
	{0x7AE8D34C87BD199EULL, "6119195_m.fxe"},
	{0x7AE8ED8DA8E3F951ULL, "6172656_m.fxe"},
	{0x7AE918A595028AA1ULL, "797988492.ogg"},
	{0x7AE9788662A1466EULL, "1039588729.ogg"},
	{0x7AE984F20D53ED98ULL, "zz_one000_barber.cnv"},
	{0x7AE9CFB60BD7E8D4ULL, "llac_ca_caverns2prc_5802_26.dds"},
	{0x7AEAFDB591F1264AULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_workstool.phy"},
	{0x7AEB234767AD13A0ULL, "6102657_f.fxe"},
	{0x7AEB29A648AF93FBULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_orsino_alt_.gad"},
	{0x7AEB7A27AF1F9623ULL, "lt_0_lowtown_docks_n_131_0.mao"},
	{0x7AEB7E27AF1F9D70ULL, "lt_0_lowtown_docks_n_131_0.mmh"},
	{0x7AEB8C27AF1FB48AULL, "lt_0_lowtown_docks_n_131_0.msh"},
	{0x7AEBA5108DD818B6ULL, "6166618_m.fxe"},
	{0x7AEC093584FA6A36ULL, "925868929.ogg"},
	{0x7AEC23EF22C83E6AULL, "vxlt_gallowstemplar_f_5pj_0_0.vlm"},
	{0x7AEC265AE7433F11ULL, "vxlt_0_lowtown_a2_gna_0_0.vlm"},
	{0x7AEC796B1C6BBE67ULL, "vxlt_gallowstemplar_f_6sm_0_0.vlm"},
	{0x7AED0C2A1ABEBDA2ULL, "129198806.ogg"},
	{0x7AED718AAE628549ULL, "dae000_gamlen6082925.cl"},
	{0x7AED938CFD96DC25ULL, "ach_soc_lover.dds"},
	{0x7AEDAAFF5254037AULL, "vxlt_undercityserialkiller_11x_0_0.vlm"},
	{0x7AEDD21541FD43D3ULL, "lt_gallowstemplar_f_8165.rml"},
	{0x7AEDDF1F22A694EDULL, "6052752_m.fxe"},
	{0x7AEDFFD1E035F439ULL, "weaponhighlight.vsh.fxo"},
	{0x7AEEEA7AC421ED51ULL, "l_dae110ar_lowtown.bnk"},
	{0x7AEF0709F43BE41EULL, "ht_0_hightown_a3_30_0.rml"},
	{0x7AEF253D74746A0DULL, "vxlt_shop_fg_0_0.vlm"},
	{0x7AEFB84BF7D5A3E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_calf_stretch.gad"},
	{0x7AEFBFEC9A08F125ULL, "6105434_m.fxe"},
	{0x7AEFC79A6BB66B50ULL, "481480135.ogg"},
	{0x7AEFD053DF1280B4ULL, "6051907_m.fxe"},
	{0x7AEFD1975477F294ULL, "gen00fl_aveline6139983.cl"},
	{0x7AF007B18B7145C0ULL, "vxht_keep_79y_0_0.vlm"},
	{0x7AF03579CB4E2637ULL, "625424695.ogg"},
	{0x7AF10ECB709A9B9EULL, "12456288.ogg"},
	{0x7AF11A8F79707C7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_enter.ani"},
	{0x7AF1F3D7D132EAE9ULL, "571095503.ogg"},
	{0x7AF2087D034DBF85ULL, "ht_0_hightown_a3_20_0.rml"},
	{0x7AF219D8C8AD9813ULL, "art\\effects\\cloudshadows\\w_encounter2_camp\\cloud2.dds"},
	{0x7AF263B54306333AULL, "mrl000_marethari6069707.cl"},
	{0x7AF2A8C1BEE8E48AULL, "isa111_isabela6067069.cl"},
	{0x7AF322184F864C73ULL, "lgt220_mag_tmplr_templar6140247.cl"},
	{0x7AF39BDA1177ED8DULL, "dae000_quintus6154616.cl"},
	{0x7AF3C35F5B4807E7ULL, "art\\vfx\\combat\\vfx_glb_charm_c.dura.evt"},
	{0x7AF43953DAD888AAULL, "hlac_lt_0_lowtown_docks_a3_n_-2_5x.dds"},
	{0x7AF467F5CB400D3DULL, "w_sundermount_193_0.mao"},
	{0x7AF46BF5CB401476ULL, "w_sundermount_193_0.mmh"},
	{0x7AF475319DE42FD7ULL, "vxht_mansion2b_3en_0_0.vlm"},
	{0x7AF475F5CB402574ULL, "w_sundermount_193_0.msh"},
	{0x7AF51DF3D4AC7C23ULL, "hlac_ca_cave_2193_0y.dds"},
	{0x7AF523AD7234F6B1ULL, "6149409_m.fxe"},
	{0x7AF535718DE74BADULL, "371293057.ogg"},
	{0x7AF58E2E50B1E623ULL, "mag131_keran6086281.cl"},
	{0x7AF5B34BF7DA06C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_calf_stretch.evt"},
	{0x7AF5FCA8CED45704ULL, "6051248_m.fxe"},
	{0x7AF63EA625E65B13ULL, "684722930.ogg"},
	{0x7AF685F7EAFCE3ADULL, "cth_nb4.tnt"},
	{0x7AF69212DA0B87EAULL, "init.bnk"},
	{0x7AF72A347FBE30A6ULL, "vxlt_0_lowtown_a2_n_c_cmj_0_0.vlm"},
	{0x7AF744F9FD6918EBULL, "zz_mcr_debug6088439.cl"},
	{0x7AF75F6A9ADA11D1ULL, "6175756_m.fxe"},
	{0x7AF79071C039E27CULL, "vxlt_darktown_4uu_0_0.vlm"},
	{0x7AF8100D313290E3ULL, "dae000_gamlen6131108.cl"},
	{0x7AF8278F7976B0BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_enter.gad"},
	{0x7AF84165DC1AB444ULL, "vxlt_warehouse_12l_0_0.vlm"},
	{0x7AF85A2350A5FF2EULL, "hlac_w_encounter2_camp_-65534_0x.dds"},
	{0x7AF869993190100EULL, "482409829.ogg"},
	{0x7AF8852EA6E2C942ULL, "gen00fl_anders6125190.cl"},
	{0x7AF886C886E9A569ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_gesture.gad"},
	{0x7AF8B5CB01A339D6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_x4broken.msh"},
	{0x7AF8BFCB01A34AD4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_x4broken.mmh"},
	{0x7AF8C3CB01A351E7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_x4broken.mao"},
	{0x7AF8C65D30BCB15DULL, "llac_w_blightlands_-3_0.dds"},
	{0x7AF8F08715A8FA0EULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantrya.msh"},
	{0x7AF8FE8715A9125FULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantrya.mao"},
	{0x7AF90A8715A926FCULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantrya.mmh"},
	{0x7AFA3194595C6B31ULL, "785737912.ogg"},
	{0x7AFA33AFAD11E3A3ULL, "vxw_bonepit_29l_0_0.vlm"},
	{0x7AFA350A8125A0A2ULL, "426271116.ogg"},
	{0x7AFA65D4C7D28EB6ULL, "vxht_0_hightown_a3_c5b_0_0.vlm"},
	{0x7AFADE32D272014BULL, "art\\characters\\weapons\\promo\\longswords\\textures\\w_lsw_025a_0n.dds"},
	{0x7AFAE45F5B4E5E30ULL, "art\\vfx\\combat\\vfx_glb_charm_c.dura.gad"},
	{0x7AFB6BABA8BE7FDDULL, "vxht_mansion2b_3cf_0_0.vlm"},
	{0x7AFB6DDC6B795636ULL, "hlac_ca_cave_2194_1z.dds"},
	{0x7AFB8C1D0F80F096ULL, "vxht_0_hightown_a2_cac_0_0.vlm"},
	{0x7AFB9110BABE9B28ULL, "hlac_w_bonepitmining_a2_3_0y.dds"},
	{0x7AFBA2BF96D7CA3DULL, "317696924.ogg"},
	{0x7AFBC4F2B53F7D50ULL, "6098218_m.fxe"},
	{0x7AFC00070C08ABEFULL, "683847177.ogg"},
	{0x7AFC5E24B43EA477ULL, "6163472_m.fxe"},
	{0x7AFC74C6856285AAULL, "21158834.ogg"},
	{0x7AFCF465E82DF357ULL, "art\\characters\\playercharacter\\humanmale\\arm_wic.phy"},
	{0x7AFD2C8AC4CF611AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45win.mmh"},
	{0x7AFD308AC4CF6851ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45win.mao"},
	{0x7AFD3E8AC4CF8060ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45win.msh"},
	{0x7AFDA8110653B40EULL, "583302542.ogg"},
	{0x7AFE3F1DA9E963B6ULL, "413478650.ogg"},
	{0x7AFE665639410122ULL, "hlac_ht_0_hightown_65533_3x.dds"},
	{0x7AFE6A924335ECB8ULL, "vxlt_undercitydungeon_15s_0_0.vlm"},
	{0x7AFE8384ABB5ACD6ULL, "mrl310_party_comments6142631.cl"},
	{0x7AFE9646F5CB40B7ULL, "vxlt_warehouse_tp_0_0.vlm"},
	{0x7AFE9D7174C9B89DULL, "1028840671.ogg"},
	{0x7AFEBBFE8CF6390AULL, "6160946_m.fxe"},
	{0x7AFEDFE0368FE1ECULL, "dae000ip_crowd_ff6_talk9.utp"},
	{0x7AFEF3DC9C7A9C9FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow1win.msh"},
	{0x7AFEF9DC9C7AA741ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow1win.mmh"},
	{0x7AFF028F797C90DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_enter.evt"},
	{0x7AFF05DC9C7ABBE2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galow1win.mao"},
	{0x7AFF1366001FA171ULL, "725089129.ogg"},
	{0x7AFF57C886EF73B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_gesture.ani"},
	{0x7AFF5F2A68D19826ULL, "6161479_m.fxe"},
	{0x7AFFA30EADBFA2B3ULL, "vxht_keep_631_0_0.vlm"},
	{0x7B0012C48129B557ULL, "155456814.ogg"},
	{0x7B003D3F1BB058B6ULL, "cd_qun221_braggart.bnk"},
	{0x7B006CDE191EED99ULL, "lit_up_01_l3.mao"},
	{0x7B0095743A5A49A4ULL, "hlac_dr_dungeon_3217_3z.dds"},
	{0x7B009BC52FE35326ULL, "gen00fl_fenris6106717.cl"},
	{0x7B00C7480DBDD9B1ULL, "vxca_caverns1_mines_geu_0_0.vlm"},
	{0x7B01E9642CF7E19AULL, "primitives_primitive_p0.vsh.fxo.dx11"},
	{0x7B01FB6FF1A8133CULL, "6124419_f.fxe"},
	{0x7B020B9D78583F8AULL, "w_encounter3_13_0.mao"},
	{0x7B020F9D785846D9ULL, "w_encounter3_13_0.mmh"},
	{0x7B02199D785857C7ULL, "w_encounter3_13_0.msh"},
	{0x7B021EC96B9587F5ULL, "6018545_m.fxe"},
	{0x7B0298C0421C1045ULL, "29113612.ogg"},
	{0x7B02D9A9B59FD8FAULL, "distor_cyl04_alp.dds"},
	{0x7B03061716E50079ULL, "vxlt_gallowstemplar_f_479_0_0.vlm"},
	{0x7B03E62607EA8C52ULL, "sse_ss_raider_f_1.gda"},
	{0x7B03EE01C979A9BAULL, "mage_elementalwep.dds"},
	{0x7B04D5BB4446521CULL, "follower_barks6153922.cl"},
	{0x7B04E7F978C9190EULL, "6089914_f.fxe"},
	{0x7B053DD7C61B8EF8ULL, "vxdr_primevalentrance_145_0_0.vlm"},
	{0x7B054603D0FA96E0ULL, "671059848.ogg"},
	{0x7B0615AA87EEDD90ULL, "1020696709.ogg"},
	{0x7B063FB1A714E73BULL, "590506215.ogg"},
	{0x7B0640B3A2CCEAC4ULL, "6135252_m.fxe"},
	{0x7B069C63046C99CBULL, "zz_vault_debug6203290.cl"},
	{0x7B06A53F0A6193D2ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_breaker_c.impa.gad"},
	{0x7B06D996A703EDC5ULL, "547415052.ogg"},
	{0x7B0705270A487524ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplinterb.msh"},
	{0x7B071B270A489AC6ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplinterb.mmh"},
	{0x7B075D3BD9F5FD3CULL, "llac_lt_0_lowtown_a3_n_c_65535_5.dds"},
	{0x7B076F266EAA989CULL, "864235233.ogg"},
	{0x7B07B06B553FD95CULL, "428967635.ogg"},
	{0x7B07C07AA7D9D774ULL, "vxlt_gallowstemplar_477_0_0.vlm"},
	{0x7B07CD9E90728103ULL, "39968372.ogg"},
	{0x7B07DB98AC8B2AD3ULL, "448259340.ogg"},
	{0x7B08601FB51109BCULL, "6132464_m.fxe"},
	{0x7B086F1B7999BF5EULL, "mcr371_arthuris6087139.cl"},
	{0x7B088CDCB02CB4BBULL, "vrc221_bartrand6065285.cl"},
	{0x7B08AD591A8B257BULL, "vxw_encounter2_kb_0_0.vlm"},
	{0x7B08BB0A92EED77FULL, "ht_0_hightownexplosion_n_36_0_ll.mmh"},
	{0x7B08F65B8F939A1BULL, "hlac_lt_0_lowtown_a3_n_c_65536_6y.dds"},
	{0x7B0954B888A20127ULL, "mag360_alain6062087.cl"},
	{0x7B09894461E52FBBULL, "170656491.ogg"},
	{0x7B0A0B67474BB5F6ULL, "9849541.ogg"},
	{0x7B0AADE197953A79ULL, "6080236_m.fxe"},
	{0x7B0B5DD9309ABC6BULL, "6201558_m.fxe"},
	{0x7B0BB123FA8216D7ULL, "hlac_ht_0_hightown_a3_65533_1z.dds"},
	{0x7B0C5FE58D88449FULL, "vxw_woundedcoast_8v_0_0.vlm"},
	{0x7B0C6552C3817201ULL, "vxht_mansion2b_2nt_0_0.vlm"},
	{0x7B0CA83F0A660409ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_breaker_c.impa.evt"},
	{0x7B0CDC6F352C0538ULL, "vxlt_undercitydungeon_18p_0_0.vlm"},
	{0x7B0CEB6B6F5D76C0ULL, "vxht_keep_73g_0_0.vlm"},
	{0x7B0D037B363CC39AULL, "vxlt_darktown_5bc_0_0.vlm"},
	{0x7B0D0823E7C15937ULL, "vxlt_warehouse_ys_0_0.vlm"},
	{0x7B0D477E2E7A312BULL, "isa211_isabela6070838.cl"},
	{0x7B0D69B3DA963BB3ULL, "6070233_m.fxe"},
	{0x7B0D8EF190FF741DULL, "vxht_keep_n_4mh_0_0.vlm"},
	{0x7B0E11EB04D54485ULL, "783374793.ogg"},
	{0x7B0E8CA759F2F59CULL, "218718553.ogg"},
	{0x7B0EF4A83A0EB6CBULL, "1003006487.ogg"},
	{0x7B0F38E61833550DULL, "hlac_lt_0_lowtown_a3_n_c_65538_2y.dds"},
	{0x7B0F819121D784ACULL, "art\\characters\\creatures\\highdragon\\cn_eye_hdr.mao"},
	{0x7B0F839FDEFB52EDULL, "hlac_lt_darktown_-1_9x.dds"},
	{0x7B0F89796FB7CAB7ULL, "6030466_m.fxe"},
	{0x7B0FE890811A0FA1ULL, "hlac_lt_0_lowtown_a2_n_c_131071_5y.dds"},
	{0x7B10184C79C64B4DULL, "vxw_encounter2_4x_0_1.vlm"},
	{0x7B1020F00B9D3733ULL, "6027534_m.fxe"},
	{0x7B107BD52B2FB271ULL, "120134257.ogg"},
	{0x7B10C3974A1A9B3DULL, "gen00fl_varric6087473.cl"},
	{0x7B111E98E975F091ULL, "hlac_dr_primevaltomb_9621_1z.dds"},
	{0x7B1126658455138DULL, "vxlt_gallowscourtyard_76l_0_0.vlm"},
	{0x7B1127B7C8B28CD7ULL, "static1_punch_s2.psh.fxo"},
	{0x7B116CDD56483DBAULL, "hlac_ht_keep_3732_14y.dds"},
	{0x7B118DB0462EE3FAULL, "960388530.ogg"},
	{0x7B11E853016E5264ULL, "zz_tanner_quanari.cut"},
	{0x7B1212270A529291ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashsplinterb.phy"},
	{0x7B12AE230FF911F8ULL, "hightown_fem_courtesan_mob6169018.cl"},
	{0x7B13016D2A95CC29ULL, "hlac_dr_slavers_3215_5y.dds"},
	{0x7B13484B29A750D4ULL, "dae220st_merch_armor.utm"},
	{0x7B136BE9C508195FULL, "hlac_lt_0_lowtown_a3_n_c_-196607_0z.dds"},
	{0x7B139DB21E54AD9DULL, "vxw_bonepitdestroyed_a3_1wo_0_0.vlm"},
	{0x7B140B099BA93E0EULL, "art\\vfx\\combat\\vfx_w_wns_front_close3_c.impa.ani"},
	{0x7B145931484181E0ULL, "vxht_0_hightown_b83_0_0.vlm"},
	{0x7B151820CA2A9C2AULL, "6002523_m.fxe"},
	{0x7B1550D9F1E9BA5CULL, "vxht_mansion2b_2a8_0_0.vlm"},
	{0x7B158B42E1D9832CULL, "plt_cod_crt_ash_wraith.plo"},
	{0x7B1596BB78BF2649ULL, "vxlt_0_lowtown_docks_n_c1v_0_0.vlm"},
	{0x7B15C795E227BF22ULL, "vxlt_darktown_38c_0_0.vlm"},
	{0x7B1636EEF3B4B89FULL, "mcr370_moira.bnk"},
	{0x7B163F219A8F53B2ULL, "6171463_m.fxe"},
	{0x7B16A2B0D7413202ULL, "6070262_f.fxe"},
	{0x7B17436C4CCC5C87ULL, "vxca_cave_1ai_0_1.vlm"},
	{0x7B1762E0392CE748ULL, "gen00fl_anders6119410.cl"},
	{0x7B177E0CF89B230BULL, "854609243.ogg"},
	{0x7B179F2CA435822AULL, "vxca_caverns1_jfs_0_0.vlm"},
	{0x7B18A2A356C892C5ULL, "vxlt_undercityserialkiller_11s_0_0.vlm"},
	{0x7B18ABAB6B7D8569ULL, "vxlt_gallowscourtyard_n_6z2_0_0.vlm"},
	{0x7B18C42A1695193BULL, "hlac_lt_0_lowtown_docks_n_-131072_0y.dds"},
	{0x7B18EB07E0DD706DULL, "rock_jagged.msh"},
	{0x7B18F907E0DD87A0ULL, "rock_jagged.mao"},
	{0x7B18FD07E0DD8EF3ULL, "rock_jagged.mmh"},
	{0x7B1945CF94825135ULL, "627574095.ogg"},
	{0x7B19B64C13B1B8F2ULL, "vxlt_warehouse_ar_0_0.vlm"},
	{0x7B19C57F20A4E110ULL, "6074461_m.fxe"},
	{0x7B19CD28F3322971ULL, "vxca_caverns1_mines_hi6_0_0.vlm"},
	{0x7B19CE870EC7469AULL, "5810099.ogg"},
	{0x7B1A18EEF3B86EA7ULL, "mcr370_moira.cnv"},
	{0x7B1A63FF8BAAECB1ULL, "qun110_petrice6037394.cl"},
	{0x7B1A809DF75E9513ULL, "735878429.ogg"},
	{0x7B1A903F0A721978ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_breaker_c.impa.ani"},
	{0x7B1AD1C27F7E5D1FULL, "vxw_woundedcoast_5q_0_0.vlm"},
	{0x7B1B16B31ED57CD0ULL, "6107356_m.fxe"},
	{0x7B1BDB90CC8C5433ULL, "6075133_m.fxe"},
	{0x7B1C00CE82F58C9DULL, "vxht_keep_n_4hg_0_0.vlm"},
	{0x7B1C0B6D5BA80128ULL, "dae000_quintus6153135.cl"},
	{0x7B1CCE59E0C3F74DULL, "vxht_chantry_n_2yr_0_0.vlm"},
	{0x7B1DE0AFFA7C78DBULL, "376782358.ogg"},
	{0x7B1E4EE6BCD6E63EULL, "vxlt_warehouse_n_11t_0_0.vlm"},
	{0x7B1E94E63F00DB20ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright.phy"},
	{0x7B1E953E2331F534ULL, "lt_gallowscourtyard_36_0.rml"},
	{0x7B1EAD2989D929E5ULL, "dae000_hate_follower_react6103245.cl"},
	{0x7B1ED9FF322C4768ULL, "430406087.ogg"},
	{0x7B1EE98715C9BDC7ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantrya.phy"},
	{0x7B1F4412FDC987B6ULL, "6023889_m.fxe"},
	{0x7B204ACC2F91DFBFULL, "lt_0_lowtown_a2_43_0.rml"},
	{0x7B2064D4C9B3BE42ULL, "391828287.ogg"},
	{0x7B20703D708A16F5ULL, "mag130cr_danzig.utc"},
	{0x7B209AAC5D6F74D1ULL, "ico_gloves_light_5.dds"},
	{0x7B20B4CCB07909B1ULL, "item_variations.gda"},
	{0x7B2103099BB3BB5FULL, "art\\vfx\\combat\\vfx_w_wns_front_close3_c.impa.evt"},
	{0x7B21431A10B8E5B3ULL, "hlac_dr_0_deeproads_17035_1z.dds"},
	{0x7B218A22CB92A8B5ULL, "vxlt_0_lowtown_a2_n_c_huv_0_0.vlm"},
	{0x7B21BB07180AF464ULL, "mcr310st_confused_guard.stg"},
	{0x7B220868B5CACFDEULL, "886020051.ogg"},
	{0x7B221E9D98D4BA80ULL, "vxht_mansion2b_29l_0_0.vlm"},
	{0x7B224B8AC4EEB225ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45win.phy"},
	{0x7B22B568F1D05D69ULL, "1040058669.ogg"},
	{0x7B22EE5C39F34DD4ULL, "368620162.ogg"},
	{0x7B23269F2144EB62ULL, "hm_feynriel_boy.mor"},
	{0x7B232AEAA6D2A564ULL, "hlac_dr_dungeon_3216_0x.dds"},
	{0x7B236472BFB90AC5ULL, "llac_ht_playershouse_n_855_4.dds"},
	{0x7B23FC73A3D78199ULL, "hlac_ca_caverns1_24424_22x.dds"},
	{0x7B241A3FE4380F78ULL, "llac_lt_undercitydungeon_-1_49.dds"},
	{0x7B24B6D1734F12F9ULL, "592707359.ogg"},
	{0x7B25077D037917ADULL, "ht_0_hightown_a3_20_0.anb"},
	{0x7B258B4AAA7B858DULL, "6080171_m.fxe"},
	{0x7B265E511F8DCA32ULL, "vxlt_0_lowtown_a2_n_c_emt_0_0.vlm"},
	{0x7B2687449014B931ULL, "386011738.ogg"},
	{0x7B26D9CAA73EEB35ULL, "vxlt_hangedman_n_20g_0_0.vlm"},
	{0x7B26F0F4BE29FB77ULL, "zz_dae_debug6045832.cl"},
	{0x7B26FA90F0EFB14AULL, "qun221_petrice6083487.cl"},
	{0x7B272E4B21B3F1B5ULL, "one110_meeran6089490.cl"},
	{0x7B27D6D774A937D8ULL, "6033648_m.fxe"},
	{0x7B27E4099BB9A548ULL, "art\\vfx\\combat\\vfx_w_wns_front_close3_c.impa.gad"},
	{0x7B28852C775B203AULL, "gen00fl_aveline6139277.cl"},
	{0x7B28CBA6F905ED14ULL, "fen110_anso6094985.cl"},
	{0x7B28FFD9A6A9477DULL, "34654719.ogg"},
	{0x7B290409F46D57C2ULL, "ht_0_hightown_a3_30_0.anb"},
	{0x7B29439F7174759FULL, "vxw_woundedcoast_2p_0_0.vlm"},
	{0x7B29B2AFEDE81B09ULL, "990132272.ogg"},
	{0x7B29B3AF474BB100ULL, "vxlt_0_lowtown_a2_n_f1s_0_0.vlm"},
	{0x7B2A02164CED231DULL, "6140994_m.fxe"},
	{0x7B2A117D170B13A2ULL, "683806485.ogg"},
	{0x7B2AE6873BD6C62DULL, "vxht_brothel_n_6df_0_0.vlm"},
	{0x7B2B9B0831EB62CEULL, "w_woundedcoast_169_0.phy"},
	{0x7B2BA01A9BEC9EECULL, "vxlt_hangedman_21c_0_1.vlm"},
	{0x7B2BD564B523BA96ULL, "6084355_f.fxe"},
	{0x7B2BEA7E01B06632ULL, "and121st_karl.stg"},
	{0x7B2BF75C46EFEC13ULL, "ht_bartrand_n_3367_lc.mmh"},
	{0x7B2C055A845559A3ULL, "and111_lirene6097133.cl"},
	{0x7B2C5EE6B4561915ULL, "6139222_m.fxe"},
	{0x7B2C8E3ACB37FA55ULL, "vxht_keep_5j3_0_0.vlm"},
	{0x7B2D44E2E8F8AB39ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_009a_0t.dds"},
	{0x7B2D970EDFA273E3ULL, "pro00fl_bethany_stunt.utc"},
	{0x7B2E32BD4BA7CD83ULL, "vxlt_0_lowtown_a2_n_c_bxq_0_0.vlm"},
	{0x7B2E46FCA220B932ULL, "llac_lt_gallowstemplar_2394_6.dds"},
	{0x7B2E7AB5FCEC1998ULL, "vxdr_primevaltomb_8o1_0_0.vlm"},
	{0x7B2EA14E828BB4D0ULL, "318028077.ogg"},
	{0x7B2F007858DFA031ULL, "483319607.ogg"},
	{0x7B2F9F6CA5A7CF13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel_loop.evt"},
	{0x7B30A772B48E6E87ULL, "vxht_keep_62n_0_0.vlm"},
	{0x7B313D15B0B933CEULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.ani"},
	{0x7B31E7D81D9E6173ULL, "vxdr_0_deeproads_9iw_0_0.vlm"},
	{0x7B32AC20E86E0163ULL, "vxw_woundedcoast_1r1_0_0.vlm"},
	{0x7B32FDB0DA5A8912ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masa_0n.dds"},
	{0x7B3325DB296CDA51ULL, "da2_prc_one_pl-pl.cre"},
	{0x7B33C67C2C283D87ULL, "6218923_m.fxe"},
	{0x7B3448D4E42D7C21ULL, "957434933.ogg"},
	{0x7B34BD4D24D55A51ULL, "623444626.ogg"},
	{0x7B352FCB1A6B3BE4ULL, "qun221_viscount6074075.cl"},
	{0x7B356153DD98EF23ULL, "follower_barks6153888.cl"},
	{0x7B358C77E58D3AA5ULL, "vxht_mansion2b_2jt_0_0.vlm"},
	{0x7B3593328DEB5436ULL, "552261131.ogg"},
	{0x7B35986CA5AC2F54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel_loop.gad"},
	{0x7B3640962C6EAE00ULL, "6174707_m.fxe"},
	{0x7B36598D756757CCULL, "hlac_lt_blackemporium_1055_2z.dds"},
	{0x7B368F0B81FBDB02ULL, "6106686_f.fxe"},
	{0x7B36E478B426E309ULL, "976932676.ogg"},
	{0x7B379D73911D87E3ULL, "hlac_lt_0_lowtown_n_131070_6y.dds"},
	{0x7B37A93DC14A67B3ULL, "llac_dr_primevalentrance_1724_1.dds"},
	{0x7B37BAE99963BB81ULL, "vxht_mansion2b_2ai_0_0.vlm"},
	{0x7B37CA399BF38298ULL, "vxlt_0_lowtown_gcc_0_0.vlm"},
	{0x7B380F833257C918ULL, "dae000_arishok6072403.cl"},
	{0x7B381A92F45418B5ULL, "555843227.ogg"},
	{0x7B3893541BD2116FULL, "dae011_corff_sta6164464.cl"},
	{0x7B389375CD277594ULL, "6200734_m.fxe"},
	{0x7B38B142230E678CULL, "6052982_m.fxe"},
	{0x7B38E605CB123398ULL, "follower_banter6160655.cl"},
	{0x7B38F7C058D3442EULL, "6163688_m.fxe"},
	{0x7B3909393F5C4CF1ULL, "10387075.ogg"},
	{0x7B399A46B802788DULL, "vxlt_0_lowtown_a3_n_c_eou_0_1.vlm"},
	{0x7B39AEF7E7454814ULL, "hlac_lt_0_lowtown_docks_n_-2_4x.dds"},
	{0x7B3A3AC031BDD39CULL, "6160838_m.fxe"},
	{0x7B3A9A2925482E5AULL, "follower_banter6169962.cl"},
	{0x7B3AC6E2E4E1AECBULL, "llac_lt_0_lowtown_docks_a3_n_-1_1.dds"},
	{0x7B3B3209F9C9FFEFULL, "vxht_0_hightown_a2_bdx_0_0.vlm"},
	{0x7B3B3D1041D5F03FULL, "gen00fl_anders6119280.cl"},
	{0x7B3B689900225122ULL, "6175966_m.fxe"},
	{0x7B3BA4E3C4D96FE9ULL, "hlac_ca_caverns1_24423_1z.dds"},
	{0x7B3C45539835207EULL, "hlac_w_sundermount_-196612_0y.dds"},
	{0x7B3C50584D55BC6EULL, "40220592.ogg"},
	{0x7B3CA76CA5B26732ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_feynriel_loop.ani"},
	{0x7B3CD04DA2EDF2A4ULL, "80835907.ogg"},
	{0x7B3D36A81A96DB5DULL, "vxdr_primevalentrance_16k_0_1.vlm"},
	{0x7B3D4246D35E2B54ULL, "lgt300cr_gang_4_ht.utc"},
	{0x7B3E00F031764548ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cartflat.mmh"},
	{0x7B3E0CF0317659EBULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cartflat.mao"},
	{0x7B3E1EF031767872ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cartflat.msh"},
	{0x7B3E466FE7CCADE2ULL, "zz_vault_debug6203334.cl"},
	{0x7B3E6C82A392F6F5ULL, "vxlt_0_lowtown_n_ep2_0_0.vlm"},
	{0x7B3EF348F436F71DULL, "vxw_bonepitdestroyed_a3_1fl_0_0.vlm"},
	{0x7B3F319EE186F112ULL, "6134954_m.fxe"},
	{0x7B3F3515B0C5631FULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.evt"},
	{0x7B3F7BD1D1A60BC4ULL, "6190987_m.fxe"},
	{0x7B3F8DC73EBC8CD2ULL, "186273467.ogg"},
	{0x7B3FD1D83CF5AB94ULL, "6005771_m.fxe"},
	{0x7B3FFD3503F6DC81ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.impa.ani"},
	{0x7B4025625FE7B280ULL, "smok.mao"},
	{0x7B4061BAAFC1057DULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isab.phy"},
	{0x7B407421D27A985CULL, "6179062_f.fxe"},
	{0x7B4191A8447A0350ULL, "character_rendertodepth_p8.psh.fxo.dx11"},
	{0x7B41F5F8CD484A1DULL, "396221761.ogg"},
	{0x7B4265E0F60A79B7ULL, "143267664.ogg"},
	{0x7B42714FA9384C04ULL, "vxht_0_hightown_a3_c36_0_0.vlm"},
	{0x7B428DF5AE1364CAULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys10x10.mmh"},
	{0x7B42D9828CA32533ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_cara_n.dds"},
	{0x7B430105BCD135C2ULL, "w_blightlands_214_0.msh"},
	{0x7B430305BCD13978ULL, "w_blightlands_214_0.mmh"},
	{0x7B430F05BCD14D1BULL, "w_blightlands_214_0.mao"},
	{0x7B435E938185F9BEULL, "336921246.ogg"},
	{0x7B436209A117E454ULL, "hlac_ca_cavern1_haunted_24427_0y.dds"},
	{0x7B43B1E63F2028B5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright.msh"},
	{0x7B43B3E63F202C2BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright.mmh"},
	{0x7B43BFE63F2040C8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright.mao"},
	{0x7B43FD2A10579EFCULL, "static_static_p9.psh.fxo.dx11.assembly"},
	{0x7B442BE98C81FE59ULL, "dae000cr_tranquil.utc"},
	{0x7B44C965E86B5FAFULL, "art\\characters\\playercharacter\\humanmale\\arm_wic.mao"},
	{0x7B44D08D883185C3ULL, "6211924_m.fxe"},
	{0x7B44D565E86B744CULL, "art\\characters\\playercharacter\\humanmale\\arm_wic.mmh"},
	{0x7B44DB65E86B7E3EULL, "art\\characters\\playercharacter\\humanmale\\arm_wic.msh"},
	{0x7B45CA45F2458104ULL, "lt_0_lowtown_docks_n_106_0.msh"},
	{0x7B45DC45F245A06DULL, "lt_0_lowtown_docks_n_106_0.mao"},
	{0x7B45E045F245A6A6ULL, "lt_0_lowtown_docks_n_106_0.mmh"},
	{0x7B46083503FB5B23ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.impa.gad"},
	{0x7B461615B0CB4D08ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.gad"},
	{0x7B466A12113C0E22ULL, "vxdr_primevaltomb_7zv_0_1.vlm"},
	{0x7B46A37380D60600ULL, "gen00fl_merrill6128694.cl"},
	{0x7B4705085C48685DULL, "hlac_ca_caverns1_24424_28z.dds"},
	{0x7B473740DC2C91B9ULL, "283607766.ogg"},
	{0x7B474BAECFC31861ULL, "dre000_varric6208037.cl"},
	{0x7B4784A3CADA1B31ULL, "art\\vfx\\textures\\vfx_vein_tile.dds"},
	{0x7B47BEC5217DB651ULL, "524346953.ogg"},
	{0x7B482AAFA75568D3ULL, "6112893_f.fxe"},
	{0x7B485AF2ACF418E4ULL, "6041825_m.fxe"},
	{0x7B48F0F5A4F890A9ULL, "gen_im_npc_wep_two_qund.uti"},
	{0x7B49149EAAECAEF1ULL, "mag131tr_eau_undead1.utt"},
	{0x7B494190E9BB93FBULL, "494830656.ogg"},
	{0x7B49475E1A57AFE9ULL, "hlac_lt_0_lowtown_docks_n_131070_0x.dds"},
	{0x7B49EAB1EA163079ULL, "239547485.ogg"},
	{0x7B4A2147F94E9BE4ULL, "6160733_m.fxe"},
	{0x7B4A4AC218E2CE2CULL, "123365718.ogg"},
	{0x7B4A4C5461291E48ULL, "gen_im_wep_mag_sta_act3_09.uti"},
	{0x7B4A8BEC8EEC14B2ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_rnf.gad"},
	{0x7B4AB19801CCD3EFULL, "28066058.ogg"},
	{0x7B4AB63E23576834ULL, "lt_gallowscourtyard_36_0.anb"},
	{0x7B4AC17A51AA3081ULL, "643012098.ogg"},
	{0x7B4B10DE700AC7C8ULL, "556888295.ogg"},
	{0x7B4B1BD5ABB8A741ULL, "6097593_m.fxe"},
	{0x7B4B33A75F20176DULL, "one3510cr_dead_fighter_2.utc"},
	{0x7B4B38F28DF65FA4ULL, "dae000ip_crowd_lm1_talk5.utp"},
	{0x7B4B4DDF212E9EA3ULL, "vrc321_varric6082563.cl"},
	{0x7B4BCB85C95F2824ULL, "llac_ca_caverns2_5802_3.dds"},
	{0x7B4C0BBD306697CBULL, "ave200_donnic6110724.cl"},
	{0x7B4C55994B601E0EULL, "vxht_chantry_2s0_0_0.vlm"},
	{0x7B4CE53504013E08ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.impa.evt"},
	{0x7B4D09E7AC9147ECULL, "6154385_m.fxe"},
	{0x7B4EB55D56D56622ULL, "hlac_lt_0_lowtown_65538_1y.dds"},
	{0x7B4ECB92018A9273ULL, "vxdr_0_deeproads_9gy_0_0.vlm"},
	{0x7B4F6D53E772A682ULL, "346111684.ogg"},
	{0x7B4FD745F24DEC71ULL, "lt_0_lowtown_docks_n_106_0.phy"},
	{0x7B4FDEB4EF4A60A5ULL, "6026957_m.fxe"},
	{0x7B501D9A78BC40BEULL, "vxht_0_hightown_a2_9uq_0_0.vlm"},
	{0x7B507FEB474D87F9ULL, "t3_rob_sla.tnt"},
	{0x7B50BDEC68915008ULL, "hlac_lt_0_lowtown_docks_n_65534_1y.dds"},
	{0x7B50E32C9B2E6484ULL, "vxht_0_hightown_a3_c47_0_0.vlm"},
	{0x7B5156BAAFCF65E9ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isab.mao"},
	{0x7B5162BAAFCF7982ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isab.mmh"},
	{0x7B5164BAAFCF7D38ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isab.msh"},
	{0x7B518EEC8EF238E9ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_rnf.evt"},
	{0x7B519B40495FCF30ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga106a_0t.dds"},
	{0x7B51BBD126768FD7ULL, "186731548.ogg"},
	{0x7B52143F47B9CA1CULL, "802987145.ogg"},
	{0x7B521B1AC7FD3B87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_loop.gad"},
	{0x7B5262427C4AEBB7ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_010a_0n.dds"},
	{0x7B528841549A2C19ULL, "6110920_m.fxe"},
	{0x7B5290F6BA983B71ULL, "and111_lirene6084539.cl"},
	{0x7B52E198D804098CULL, "325854838.ogg"},
	{0x7B52ECA0F843FE34ULL, "6179044_f.fxe"},
	{0x7B530DE60A5A96C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v2.ani"},
	{0x7B53DA82AFFE63C6ULL, "407480290.ogg"},
	{0x7B53E7B344C942C3ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgte_0s.dds"},
	{0x7B53EDFDF8B6C18AULL, "vxht_chantry_27h_0_0.vlm"},
	{0x7B5416AF4D74FA62ULL, "s_smoke_02.dds"},
	{0x7B5452D4703C0C6FULL, "da2_prc_seb_es-es.cre"},
	{0x7B54E481DBC0A2E8ULL, "hm_wright.mor"},
	{0x7B556AF994D3EB59ULL, "342565622.ogg"},
	{0x7B55B4BAD418CD9BULL, "mag370_meredith6037409.cl"},
	{0x7B55BA0466E1125CULL, "184564512.ogg"},
	{0x7B55DBE8BA3DB8E4ULL, "6150734_m.fxe"},
	{0x7B5695CDA89C594DULL, "vxht_chantry_n_2en_0_0.vlm"},
	{0x7B56A94EFC121AA7ULL, "6144556_m.fxe"},
	{0x7B56C7E930879CE2ULL, "mcr370_negotiation6095525.cl"},
	{0x7B56E4DCB56821F7ULL, "173888394.ogg"},
	{0x7B57136B60E5CFBAULL, "6042516_m.fxe"},
	{0x7B575E091C89BC72ULL, "311088186.ogg"},
	{0x7B577F73CE8CADD9ULL, "vxw_bonepitdestroyed_a3_81_0_0.vlm"},
	{0x7B57A8B02D0F02A6ULL, "vxlt_backalley_g_1kb_0_0.vlm"},
	{0x7B57CC19A327EB79ULL, "vxlt_gallowscourtyard_6yx_0_0.vlm"},
	{0x7B57D112EF16F27EULL, "hlac_lt_0_lowtown_a2_65537_5z.dds"},
	{0x7B58053C418848D4ULL, "vxlt_gallowstemplar_55t_0_0.vlm"},
	{0x7B58E1F4892825EAULL, "441978186.ogg"},
	{0x7B590E8D55D463F3ULL, "6039775_m.fxe"},
	{0x7B59355423DC0B28ULL, "gen00fl_fenris6126946.cl"},
	{0x7B595C7A33C36464ULL, "hlac_ht_bartrand_n_2995_2z.dds"},
	{0x7B5962098E176E2BULL, "6126766_f.fxe"},
	{0x7B5985A9D2D62437ULL, "vxw_encounter3_5c_0_0.vlm"},
	{0x7B59FC99F5827986ULL, "vxlt_0_lowtown_a3_n_c_hv9_0_0.vlm"},
	{0x7B5A287CE59010A3ULL, "6054756_m.fxe"},
	{0x7B5A61B2DEFD5ABFULL, "vxht_keep_n_5x4_0_0.vlm"},
	{0x7B5A76C622F76C2DULL, "vxw_sundermount_20v_0_0.vlm"},
	{0x7B5A7A0D736B09A9ULL, "1064670991.ogg"},
	{0x7B5BAFE939E0143FULL, "mrl000_marethari6129291.cl"},
	{0x7B5BD4D805975175ULL, "644709180.ogg"},
	{0x7B5C937AF197507BULL, "vxht_brothel_n_6gu_0_0.vlm"},
	{0x7B5CA3903694B010ULL, "lgt000cr_baz_riv_leg.utc"},
	{0x7B5CFD0F4B2E653DULL, "106159432.ogg"},
	{0x7B5D16541DF82986ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comb_0s.dds"},
	{0x7B5D6D70AD3BBFDAULL, "hlac_dr_slavers_3207_1y.dds"},
	{0x7B5D7A9C6ACB98F2ULL, "6067294_m.fxe"},
	{0x7B5DB453F02AB216ULL, "6163420_m.fxe"},
	{0x7B5E0F9D78A6A532ULL, "w_encounter3_13_0.rml"},
	{0x7B5E2CF2D1333429ULL, "vrc221_varric6044070.cl"},
	{0x7B5E76EC8EFC9A58ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90l_rnf.ani"},
	{0x7B5EA882C0238EE6ULL, "vxht_mansion2_1rl_0_0.vlm"},
	{0x7B5ECA9D1C51CED9ULL, "6149636_m.fxe"},
	{0x7B5EDCB8714544F6ULL, "6153108_m.fxe"},
	{0x7B5F3A7145FF79FFULL, "6112612_m.fxe"},
	{0x7B6025C4E7662C32ULL, "vxlt_0_lowtown_a2_n_c_eqp_0_0.vlm"},
	{0x7B606DCC2FC7FF23ULL, "lt_0_lowtown_a2_43_0.anb"},
	{0x7B60CB340B48D6D4ULL, "6154420_m.fxe"},
	{0x7B6125E60A66FCBAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v2.evt"},
	{0x7B6139E6E204C77AULL, "vxlt_undercityserialkiller_1ip_0_0.vlm"},
	{0x7B615272BD3C52D3ULL, "6087522_f.fxe"},
	{0x7B620E51EE63FF9AULL, "6137242_f.fxe"},
	{0x7B625FDAEAACDA0AULL, "vxht_chantry_20i_0_0.vlm"},
	{0x7B628AD9FEE00457ULL, "fen311cr_patron_3.utc"},
	{0x7B62AD90E87BA3B6ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast02.evt"},
	{0x7B62CEA55E5F99CDULL, "6065949_m.fxe"},
	{0x7B62D16CC330ED45ULL, "531784269.ogg"},
	{0x7B633D67BB4DE9F1ULL, "dre180_db_rock_rog.cut"},
	{0x7B639A17E1F14303ULL, "6042572_f.fxe"},
	{0x7B63AB73A7BD4DFAULL, "zz_followers_banter6159138.cl"},
	{0x7B645B587B97AE47ULL, "zz_prmpt_debug6200979.cl"},
	{0x7B648B1300FF0CFEULL, "art\\vfx\\combat\\vfx_m_bld_mgc_act_c.impa.evt"},
	{0x7B6513547684BC1DULL, "154296491.ogg"},
	{0x7B655B990EEC0C71ULL, "278927943.ogg"},
	{0x7B656F9F6A7F41BBULL, "dae000_flemeth6013371.cl"},
	{0x7B65A6D47F0ECAC2ULL, "48317669.ogg"},
	{0x7B65C7E5673E1AC0ULL, "vxw_encounter1_aj_0_0.vlm"},
	{0x7B65E45E469CF766ULL, "qun221_petrice6076667.cl"},
	{0x7B6600E77CF44501ULL, "169586688.ogg"},
	{0x7B661A8D1F051B26ULL, "vxlt_backalley_g_1pc_0_0.vlm"},
	{0x7B66788580571D30ULL, "6208496_f.fxe"},
	{0x7B66B1DB5225BBFEULL, "85651434.ogg"},
	{0x7B67081AC80F77CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_loop.ani"},
	{0x7B670A3F5D2A6B3EULL, "hlac_lt_gallowstemplar_n_2383_1x.dds"},
	{0x7B6724E60A6B66ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v2.gad"},
	{0x7B6785AD91982395ULL, "713316831.ogg"},
	{0x7B67AE792BB877B4ULL, "hm_iconicmale_nomark.mor"},
	{0x7B67AE792BB877B6ULL, "hm_iconicmale_nomark.mop"},
	{0x7B6826CD47520D1DULL, "vxdr_primevaltomb_8en_0_0.vlm"},
	{0x7B68B37C344F882DULL, "gen_im_arm_glv_lgt_set_high.uti"},
	{0x7B68E8A314ED84ADULL, "vxw_sundermount_25u_0_0.vlm"},
	{0x7B692DADF5415C01ULL, "358636859.ogg"},
	{0x7B694C61895FE5AEULL, "vxdr_slavers_132_0_0.vlm"},
	{0x7B69A490E881B341ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast02.gad"},
	{0x7B6A05ED78E83D92ULL, "llac_lt_gallowstemplar_f_2383_3.dds"},
	{0x7B6A1C1FE1D97336ULL, "498427947.ogg"},
	{0x7B6A93D26573B305ULL, "llac_w_bonepitmining_a2_393217_0.dds"},
	{0x7B6B0A7B8BBA17BFULL, "95913426.ogg"},
	{0x7B6B9A0832214491ULL, "w_woundedcoast_169_0.mmh"},
	{0x7B6BA21301055229ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_act_c.impa.gad"},
	{0x7B6BA60832215932ULL, "w_woundedcoast_169_0.mao"},
	{0x7B6BB4083221714FULL, "w_woundedcoast_169_0.msh"},
	{0x7B6CD02B28BDF301ULL, "780156680.ogg"},
	{0x7B6D80E37316143BULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztorch_0.pwk"},
	{0x7B6DA90F68DEAE8EULL, "6101770_m.fxe"},
	{0x7B6DFD5A3F8F5205ULL, "vxht_0_hightown_b7k_0_0.vlm"},
	{0x7B6E5D21292670F4ULL, "qun110_petrice6018624.cl"},
	{0x7B6E8CE381E0F62FULL, "vxlt_0_lowtown_a2_n_emg_0_0.vlm"},
	{0x7B6EC1C52B880727ULL, "dae200cr_viscount6087171.cl"},
	{0x7B6EE0730BDD1B86ULL, "6208192_m.fxe"},
	{0x7B6F0B8A573EAA72ULL, "703196697.ogg"},
	{0x7B6FABC3D3FADFFAULL, "vxlt_undercityserialkiller_1nq_0_0.vlm"},
	{0x7B709EB2CCBCD72DULL, "vxht_0_hightown_n_cbc_0_1.vlm"},
	{0x7B70A590E887D3EFULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast02.ani"},
	{0x7B70AD8D5B77B567ULL, "w_bonepitdestroyed_a3_8_0.rml"},
	{0x7B7103DDD4A72D62ULL, "149231835.ogg"},
	{0x7B71FBC13FC1CA6FULL, "vxlt_0_lowtown_a2_hdz_0_0.vlm"},
	{0x7B720CE145B24C56ULL, "art\\characters\\creatures\\mubari\\animation\\com_cocked.evt"},
	{0x7B726F8CFEEC4E71ULL, "vxlt_0_lowtown_docks_a3_n_bgn_0_0.vlm"},
	{0x7B727313010B2177ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_act_c.impa.ani"},
	{0x7B73570D9F4CAFB8ULL, "857034106.ogg"},
	{0x7B73A90735ECEA50ULL, "sse_ss_crazy_f_1.gda"},
	{0x7B73BB88EE275EACULL, "mag110_arianni6016061.cl"},
	{0x7B73C79B9FBB1259ULL, "vxw_sundermount_37j_0_0.vlm"},
	{0x7B73CE42478C4533ULL, "csp_clo.mao"},
	{0x7B73FEBE403F686EULL, "dae370ip_mage_store.utp"},
	{0x7B74101F7ED77E8AULL, "957313820.ogg"},
	{0x7B7439C259343340ULL, "vxw_encounter1_fk_0_0.vlm"},
	{0x7B744D434B4DF951ULL, "6096749_m.fxe"},
	{0x7B747A4B4621DCB5ULL, "llac_lt_0_lowtown_a2_1_10.dds"},
	{0x7B74ADED25F8C2E9ULL, "6053320_m.fxe"},
	{0x7B74CA6B2AFBF1B7ULL, "vxlt_0_lowtown_a2_era_0_0.vlm"},
	{0x7B758BC44FB547AAULL, "art\\vfx\\creatures\\fxc_suc_att_s_c.anb"},
	{0x7B768295E82D6663ULL, "6161648_f.fxe"},
	{0x7B76B1BA309C06ACULL, "ser270_emeric6149379.cl"},
	{0x7B76D81613963A5AULL, "vxw_encounter1_ix_0_0.vlm"},
	{0x7B77616E99271AFBULL, "lt_0_lowtown_docks_a3_253_0.mmh"},
	{0x7B776D6E99272E98ULL, "lt_0_lowtown_docks_a3_253_0.mao"},
	{0x7B777F6E99274DE5ULL, "lt_0_lowtown_docks_a3_253_0.msh"},
	{0x7B7903E145B85BE1ULL, "art\\characters\\creatures\\mubari\\animation\\com_cocked.gad"},
	{0x7B7908F15E113E1FULL, "vxca_caverns1_e8w_0_0.vlm"},
	{0x7B7933B90529ACB4ULL, "vxlt_hangedman_20y_0_0.vlm"},
	{0x7B797C1E18F8D8D4ULL, "6081270_m.fxe"},
	{0x7B79884C3FF65135ULL, "mcr370_negotiation6080258.cl"},
	{0x7B79FC74ABB9D626ULL, "vxw_bonepit_1wq_0_1.vlm"},
	{0x7B7A71D1C32B07DFULL, "971581756.ogg"},
	{0x7B7AFD997107EFC6ULL, "262976239.ogg"},
	{0x7B7B17BDDC1FC439ULL, "vis121_saemus6099488.cl"},
	{0x7B7B6BB2E41C7642ULL, "935358780.ogg"},
	{0x7B7B6F2F2F749DDBULL, "hlac_ht_0_hightown_a2_-131074_0y.dds"},
	{0x7B7C9220BE22EC6CULL, "ave260tr_long_road_smuggler.ncs"},
	{0x7B7D2B88D04DD023ULL, "dae000_gamlen6121581.cl"},
	{0x7B7DA1894B9FB21AULL, "vxht_mansion2_61_0_0.vlm"},
	{0x7B7DB0237F1D888CULL, "255800880.ogg"},
	{0x7B7EB9C4EB451BABULL, "follower_banter6169767.cl"},
	{0x7B7EE88B51AD16E5ULL, "gen00fl_aveline6139122.cl"},
	{0x7B7EEC7FC59D9206ULL, "llac_lt_0_lowtown_n_-196608_1.dds"},
	{0x7B7EFFE13D6808E0ULL, "602793247.ogg"},
	{0x7B7F04E145BCC88FULL, "art\\characters\\creatures\\mubari\\animation\\com_cocked.ani"},
	{0x7B7F4FBDE2D265EDULL, "6137228_m.fxe"},
	{0x7B7FE14339A821EBULL, "prx_stairs.dds"},
	{0x7B80442050C6ADBCULL, "dae000pt_amb_actions.ncs"},
	{0x7B808BC26919E4EDULL, "follower_banter6164685.cl"},
	{0x7B809E4FD463E5A6ULL, "6164784_m.fxe"},
	{0x7B80F3CF41C2DA84ULL, "gen00fl_anders6111195.cl"},
	{0x7B81045B7A146166ULL, "6087854_m.fxe"},
	{0x7B81751C43739E4FULL, "character_rendertodepth_p3.vsh.fxo.dx11"},
	{0x7B819396DA4F2B68ULL, "abi_w_aoe_knockback.wbd"},
	{0x7B81B693AFC6B5A2ULL, "dae000ip_crowd_mf2_talk15.utp"},
	{0x7B81BA1BE6E5300FULL, "30242303.ogg"},
	{0x7B82296949FE9C4BULL, "884673238.ogg"},
	{0x7B82397891B12AD9ULL, "vxw_sundermount_32i_0_0.vlm"},
	{0x7B828F8319E2954BULL, "6095654_m.fxe"},
	{0x7B82AB9F4B2A4BC0ULL, "vxw_encounter1_kh_0_0.vlm"},
	{0x7B82E86E6303B84FULL, "271150036.ogg"},
	{0x7B8307F217DE016FULL, "6035212_m.fxe"},
	{0x7B839D705C0BDE41ULL, "w_woundedcoast_122_0.phy"},
	{0x7B83BED30805FEC3ULL, "6000989_f.fxe"},
	{0x7B83C8BAAD9B31ECULL, "6020529_m.fxe"},
	{0x7B83DE50F5A1B6ABULL, "vxw_encounter2_k2_0_0.vlm"},
	{0x7B84443DC4BA9CFDULL, "blood_conj_06a_l3.mao"},
	{0x7B846679EC4156C7ULL, "llac_lt_undercitydungeon_-1_15.dds"},
	{0x7B85238F907A5402ULL, "6115344_m.fxe"},
	{0x7B85253A8FB27E22ULL, "llac_lt_hangedman_-1_3.dds"},
	{0x7B8540771383B6C5ULL, "6068555_m.fxe"},
	{0x7B8552B7B812F8CDULL, "hlac_lt_gallowstemplar_2383_5x.dds"},
	{0x7B85668DFDF76137ULL, "gen_im_arm_cht_med_start.uti"},
	{0x7B85A5B99E6C618CULL, "6071228_m.fxe"},
	{0x7B85B749B4DFA43FULL, "vxht_keep_n_5k1_0_0.vlm"},
	{0x7B85C21E189A79B8ULL, "6177671_f.fxe"},
	{0x7B8601A84A4459E4ULL, "6172521_m.fxe"},
	{0x7B8607032E7241A4ULL, "6139840_f.fxe"},
	{0x7B8637BAADD2E9ACULL, "78535800.ogg"},
	{0x7B865CF14933475EULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_throwguts.evt"},
	{0x7B867BDF5956022EULL, "strcave01_n.dds"},
	{0x7B869B0465038BD0ULL, "6042508_f.fxe"},
	{0x7B86B93A0BA60CFCULL, "dae000_bodhan6146905.cl"},
	{0x7B86EE23BCA79281ULL, "845140447.ogg"},
	{0x7B86EE6EE1AF30CCULL, "92127160.ogg"},
	{0x7B8730A9A6531CCDULL, "bl_pol_l.mao"},
	{0x7B8797F5B9A8709EULL, "m_dae360ar_wounded_coast.bnk"},
	{0x7B87B3C0D9AEC270ULL, "hlac_lt_gallowstemplar_n_2394_7x.dds"},
	{0x7B881370603192A4ULL, "gen_im_trs_krk_one_lsw_r2.uti"},
	{0x7B887C41FAA6AD8FULL, "zz_approval_debug6003509.cl"},
	{0x7B88930EFAB00FDBULL, "vxca_caverns2prc_a6w_0_0.vlm"},
	{0x7B8899B43E40E673ULL, "zz_followers_banter6158325.cl"},
	{0x7B889D433344A6E4ULL, "13487270.ogg"},
	{0x7B88A45DFA2F937EULL, "6082422_m.fxe"},
	{0x7B88AD8D5B8C4684ULL, "w_bonepitdestroyed_a3_8_0.mmh"},
	{0x7B88B18D5B8C4DD7ULL, "w_bonepitdestroyed_a3_8_0.mao"},
	{0x7B88C38D5B8C6C26ULL, "w_bonepitdestroyed_a3_8_0.msh"},
	{0x7B88C7BB7AE56E9DULL, "vxlt_0_lowtown_docks_b7z_0_0.vlm"},
	{0x7B89B230CD201DEEULL, "496828311.ogg"},
	{0x7B89FE199605D866ULL, "vxht_mansion2_1co_0_0.vlm"},
	{0x7B8ADB483432B51CULL, "211985703.ogg"},
	{0x7B8B4B8CC41C4DF4ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev2.ani"},
	{0x7B8BC2FB6B80604BULL, "6172629_m.fxe"},
	{0x7B8BE8106B347C5AULL, "6136952_m.fxe"},
	{0x7B8C301BDDB3B92AULL, "6002772_m.fxe"},
	{0x7B8CB36CB7ED378FULL, "vxht_0_hightown_n_bsv_0_0.vlm"},
	{0x7B8CFB2E0240795AULL, "6076919_m.fxe"},
	{0x7B8CFF591B8BA8B5ULL, "l_dae360ar_wounded_coast.bnk"},
	{0x7B8D73F149398C89ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_throwguts.gad"},
	{0x7B8D775938805A53ULL, "zz_vault_debug6203193.cl"},
	{0x7B8E1A9446A35A4DULL, "736062456.ogg"},
	{0x7B8E71D973C58644ULL, "6048840_m.fxe"},
	{0x7B8EC2276E438F09ULL, "6079796_f.fxe"},
	{0x7B8EC94A7912BDC3ULL, "358547470.ogg"},
	{0x7B8EE8AE1359C6EBULL, "291280795.ogg"},
	{0x7B8F915F12621F26ULL, "gen00fl_anders6107712.cl"},
	{0x7B9020EF2AF317FBULL, "6036467_f.fxe"},
	{0x7B905AA4D4F490E0ULL, "6069056_m.fxe"},
	{0x7B906AD211DBF128ULL, "vxlt_0_lowtown_n_ere_0_0.vlm"},
	{0x7B90AB5583A74359ULL, "vxw_sundermount_39h_0_0.vlm"},
	{0x7B90F0E754AB24A4ULL, "gen00fl_aveline6120765.cl"},
	{0x7B91488CC420B456ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev2.gad"},
	{0x7B91617BECA8E778ULL, "vxdr_primevaltomb_6qa_0_0.vlm"},
	{0x7B91A425D276CC56ULL, "6132893_m.fxe"},
	{0x7B925A1F7C27722DULL, "isa211_isabela6202192.cl"},
	{0x7B92AF5529E01EACULL, "643532933.ogg"},
	{0x7B9323ECFC339C92ULL, "580237326.ogg"},
	{0x7B933E1021DDB625ULL, "hlac_lt_0_lowtown_a2_n_-2_1y.dds"},
	{0x7B934328F593F56AULL, "197985418.ogg"},
	{0x7B937C641D34569DULL, "vxdr_primevaltomb_8vk_0_0.vlm"},
	{0x7B9404327BDB9FD7ULL, "art\\characters\\weapons\\shields\\w_shd_010a.mmh"},
	{0x7B941A327BDBC549ULL, "art\\characters\\weapons\\shields\\w_shd_010a.msh"},
	{0x7B9444F1493F5BD7ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_throwguts.ani"},
	{0x7B94741E3577D353ULL, "6076077_m.fxe"},
	{0x7B947D311707B94EULL, "vxlt_0_lowtown_a2_e11_0_0.vlm"},
	{0x7B95569AB68C53A9ULL, "fadesky.mao"},
	{0x7B95DCCA68C8D7B9ULL, "291416110.ogg"},
	{0x7B95DDABCEE61615ULL, "hlac_dr_0_deeproads_5838_0y.dds"},
	{0x7B95E3DC6D01110FULL, "vxht_keep_6ao_0_0.vlm"},
	{0x7B9625EDB232A552ULL, "6093841_f.fxe"},
	{0x7B9628F1B101D61EULL, "76330409.ogg"},
	{0x7B963CB9A8E31A7CULL, "gen00fl_aveline6137655.cl"},
	{0x7B96C38889C3ECBEULL, "205561256.ogg"},
	{0x7B96E40082D3D8F3ULL, "805937201.ogg"},
	{0x7B9708B2A517E741ULL, "vxw_bonepit_245_0_0.vlm"},
	{0x7B971B759CB2BA68ULL, "6218010_m.fxe"},
	{0x7B9757EF867A3A6AULL, "vxlt_gallowstemplar_n_5xj_0_0.vlm"},
	{0x7B979C09D81F3436ULL, "ring_of_resilience.dds"},
	{0x7B97E5A6760DCE7DULL, "6152276_m.fxe"},
	{0x7B982AC4F9382B98ULL, "follower_banter6171110.cl"},
	{0x7B98638CC42700B5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev2.evt"},
	{0x7B98A8ED6A939821ULL, "532570731.ogg"},
	{0x7B98FC0C1F03AC3FULL, "gen_im_acc_amu_act1_06.uti"},
	{0x7B99053BEB08FD2EULL, "vxht_0_hightown_a2_b8i_0_0.vlm"},
	{0x7B9985E481909663ULL, "vxlt_darktown_5lt_0_0.vlm"},
	{0x7B99FBB6128C6B37ULL, "hlac_w_encounter2_camp_1_0x.dds"},
	{0x7B9A184BBC9B1F66ULL, "6036708_m.fxe"},
	{0x7B9A762FC8063C89ULL, "vxca_caverns1_j2b_0_2.vlm"},
	{0x7B9A918597F279F5ULL, "dae000_gen_player_turn_in6177841.cl"},
	{0x7B9AA367D86F6F45ULL, "vxlt_undercityserialkiller_1nz_0_0.vlm"},
	{0x7B9BFF8635076E53ULL, "6133279_m.fxe"},
	{0x7B9C53553A0F9EC1ULL, "6116742_m.fxe"},
	{0x7B9C5AE30A363A41ULL, "lt_0_lowtown_docks_n_10_0.anb"},
	{0x7B9CD35A8D9B5791ULL, "620352418.ogg"},
	{0x7B9CD443CE3E6A27ULL, "vxlt_warehouse_n_qr_0_0.vlm"},
	{0x7B9D47786EE5C004ULL, "gen00fl_merrill6105660.cl"},
	{0x7B9D4AA0ED0E7D05ULL, "vxht_brothel_n_4c1_0_0.vlm"},
	{0x7B9D698AEBF22CECULL, "llac_lt_0_lowtown_docks_-65538_8.dds"},
	{0x7B9DABEC0849879AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamgrid.mmh"},
	{0x7B9DAFEC08498ED1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamgrid.mao"},
	{0x7B9DBDEC0849A6E0ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamgrid.msh"},
	{0x7B9E159EAB34F044ULL, "mag131tr_eau_undead1.ncs"},
	{0x7B9E73B48A9BB492ULL, "vxht_brothel_n_5hj_0_0.vlm"},
	{0x7B9E8EC44FD8519DULL, "art\\vfx\\creatures\\fxc_suc_att_s_c.mmh"},
	{0x7B9F108FCA5FC0F9ULL, "6096604_m.fxe"},
	{0x7B9F9A354366D906ULL, "gen_im_npc_wep_one_quna.uti"},
	{0x7B9FD36E8DB02C38ULL, "hlac_w_woundedcoast_196615_0y.dds"},
	{0x7B9FEFB345D398E5ULL, "dr_primevaltomb_mmi.gff"},
	{0x7BA02526F7BC606DULL, "follower_banter6171948.cl"},
	{0x7BA07808E9A9D6E8ULL, "6097314_m.fxe"},
	{0x7BA08743B348B87BULL, "mcr370_negotiation6080263.cl"},
	{0x7BA0A315C7086A94ULL, "llac_lt_gamlenshouse_n_-1_4.dds"},
	{0x7BA0D3E06358F980ULL, "llac_ca_cavern1_haunted_24422_4.dds"},
	{0x7BA1080C684ADC2FULL, "cs_pro000cs_act1_crim_mage.wbd"},
	{0x7BA1521C3215B0E2ULL, "vxdr_0_deeproads_fbh_0_0.vlm"},
	{0x7BA1EE410F2A6F1DULL, "vxdr_primevaltomb_8qj_0_0.vlm"},
	{0x7BA222BD21A295C2ULL, "mag291_vincento6027328.cl"},
	{0x7BA29490EA621CBEULL, "prm000im_deep_blade.uti"},
	{0x7BA297B11BB16218ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_007a_0t.dds"},
	{0x7BA2A65C96E9A971ULL, "vxht_bartrand_n_5pm_0_0.vlm"},
	{0x7BA381A9BA175F2EULL, "vxht_0_hightown_n_ap9_0_0.vlm"},
	{0x7BA3AB4BC3F61EADULL, "gen_im_trs_gen_bracelet_04.uti"},
	{0x7BA42076F0BEB5E2ULL, "42322069.ogg"},
	{0x7BA431B1341CA9D2ULL, "6131171_m.fxe"},
	{0x7BA455B95EF7298FULL, "vxht_keep_6jn_0_0.vlm"},
	{0x7BA461EDD5242DC4ULL, "618155354.ogg"},
	{0x7BA471046D6E14EFULL, "vxht_keep_n_58d_0_0.vlm"},
	{0x7BA48E0694C4354EULL, "vxw_encounter3_e6_0_0.vlm"},
	{0x7BA520DA3ACA51A9ULL, "art\\vfx\\textures\\vfx_shield.dds"},
	{0x7BA550A1445C9B7AULL, "montage_subs_demohack6217132.cl"},
	{0x7BA64D48941055D4ULL, "6108645_m.fxe"},
	{0x7BA65D47FBA693C0ULL, "658678637.ogg"},
	{0x7BA6C25108423AC1ULL, "6060233_f.fxe"},
	{0x7BA6E1D379F20966ULL, "vxht_mansion2_1iq_0_0.vlm"},
	{0x7BA70878D1617699ULL, "vxlt_darktown_5m1_0_0.vlm"},
	{0x7BA71ABAE6047280ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga106a_0d.dds"},
	{0x7BA7BB198D677C57ULL, "gen00fl_sebastian6189015.cl"},
	{0x7BA7BEEBA8D36EB0ULL, "6060510_m.fxe"},
	{0x7BA7F641EBAF0576ULL, "art\\characters\\creatures\\desiredemon\\animation\\attackv1.ani"},
	{0x7BA7F7C17386AEE3ULL, "vxlt_darktown_5qw_0_0.vlm"},
	{0x7BA81FDD82C143D0ULL, "150981653.ogg"},
	{0x7BA87E07BE986CF1ULL, "6103289_m.fxe"},
	{0x7BA997269BD9688FULL, "vxht_0_hightown_n_byp_0_0.vlm"},
	{0x7BA99DD64E5890E5ULL, "vxlt_0_lowtown_docks_cvi_0_1.vlm"},
	{0x7BA9B136ECB879DDULL, "558104534.ogg"},
	{0x7BAA148D10DF13C5ULL, "follower_barks6153547.cl"},
	{0x7BAA4BD9401F6E0DULL, "408529285.ogg"},
	{0x7BAAD72DF9A578F5ULL, "vxw_bonepitmining_a2_fq_0_0.vlm"},
	{0x7BAADF8DA3C7A11EULL, "759913398.ogg"},
	{0x7BAB182626832752ULL, "6079077_m.fxe"},
	{0x7BAB4620C03482A7ULL, "vxlt_warehouse_n_tm_0_0.vlm"},
	{0x7BAC05D27EA02C5AULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordb.mmh"},
	{0x7BAC09D27EA03291ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordb.mao"},
	{0x7BAC17D27EA04AA0ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordb.msh"},
	{0x7BAC2E7295C4AC3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow.ani"},
	{0x7BAC70CCDEB12C16ULL, "519882906.ogg"},
	{0x7BACB27174AA353CULL, "378958808.ogg"},
	{0x7BACC86B7768F26BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_destroyed_1.mao"},
	{0x7BAD20AA70414E0DULL, "vxca_cavern1_haunted_iio_0_0.vlm"},
	{0x7BAD4E8BF5C82228ULL, "vxlt_0_lowtown_n_ehc_0_0.vlm"},
	{0x7BAD979CFD316152ULL, "8819729.ogg"},
	{0x7BAEF6A1571A1036ULL, "6144430_m.fxe"},
	{0x7BAF30673C9D3D42ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzchartest.msh"},
	{0x7BAF32673C9D40F8ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzchartest.mmh"},
	{0x7BAF3E673C9D549BULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzchartest.mao"},
	{0x7BAF76B645651FB0ULL, "vxht_0_hightown_a3_by5_0_0.vlm"},
	{0x7BAFA8DD9A085A11ULL, "mag131_keran6086180.cl"},
	{0x7BB0175DB1CA065DULL, "vxlt_gallowstemplar_f_3mh_0_0.vlm"},
	{0x7BB05A49C8836EF9ULL, "hlac_w_encounter2_327678_0x.dds"},
	{0x7BB0969083BDD07DULL, "6083426_m.fxe"},
	{0x7BB0B4371EFC32DBULL, "6075388_m.fxe"},
	{0x7BB119F84E40A218ULL, "vxca_cavern1_haunted_j5q_0_0.vlm"},
	{0x7BB1420449CB8EF1ULL, "sw_p_gr.mao"},
	{0x7BB1C0689281DFCBULL, "644973729.ogg"},
	{0x7BB2096A40F49E60ULL, "llac_lt_0_lowtown_a2_n_131071_5.dds"},
	{0x7BB231CDC8782596ULL, "6117791_m.fxe"},
	{0x7BB33B7295CAE17EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow.gad"},
	{0x7BB34E0AA211F6B4ULL, "6129921_m.fxe"},
	{0x7BB3ABB3208961C2ULL, "gen00fl_anders6125859.cl"},
	{0x7BB3EB7344659958ULL, "160290769.ogg"},
	{0x7BB475A1AE92A62DULL, "6134925_f.fxe"},
	{0x7BB47D1AAC217954ULL, "vxca_caverns2_8lv_0_0.vlm"},
	{0x7BB4D77BDF30227DULL, "hlac_lt_0_lowtown_docks_131071_3y.dds"},
	{0x7BB4FE41EBB99D67ULL, "art\\characters\\creatures\\desiredemon\\animation\\attackv1.evt"},
	{0x7BB5198099596041ULL, "102030799.ogg"},
	{0x7BB5EFEDA53C7ED4ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_cupwood_01_0_n.dds"},
	{0x7BB61C7802E86F96ULL, "40188869.ogg"},
	{0x7BB62F718ECD87C3ULL, "ts_darkspawn_wpn_table.gda"},
	{0x7BB6AF869A9CD5D4ULL, "art\\characters\\creatures\\stonegolem\\animation\\blendtree\\stonegolem_movement.anb"},
	{0x7BB7A81648CF427CULL, "lt_0_lowtown_docks_a3_n_29_0.anb"},
	{0x7BB7BCE77AFFA591ULL, "dae000cr_grand_cleric6148126.cl"},
	{0x7BB7FC7FF2894F67ULL, "dae000_orsino6085496.cl"},
	{0x7BB8013998FF4FAEULL, "one110_gustav6089794.cl"},
	{0x7BB809038DCF810FULL, "vxht_0_hightown_n_bds_0_0.vlm"},
	{0x7BB86E5C766423E5ULL, "265966898.ogg"},
	{0x7BB8DD50CBBA5D22ULL, "648305.ogg"},
	{0x7BB924D6E429ADC8ULL, "llac_ht_keep_n_3735_0.dds"},
	{0x7BB925591BB1243FULL, "l_dae360ar_wounded_coast.wbd"},
	{0x7BB9490AEB9B9175ULL, "vxw_bonepitmining_a2_mp_0_0.vlm"},
	{0x7BB95D3689D1F141ULL, "6085507_m.fxe"},
	{0x7BB9A932E2067870ULL, "vxht_mansion_n_5b5_0_0.vlm"},
	{0x7BB9D81EA49ED419ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_exit.ani"},
	{0x7BB9FDF54FE9F358ULL, "vxlt_darktown_5bq_0_0.vlm"},
	{0x7BBA167295D0C09DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow.evt"},
	{0x7BBC0BFA0995EA19ULL, "hlac_lt_0_lowtown_docks_-131074_6z.dds"},
	{0x7BBC1F41EBBFF3B0ULL, "art\\characters\\creatures\\desiredemon\\animation\\attackv1.gad"},
	{0x7BBC63F007C5C42DULL, "6107054_m.fxe"},
	{0x7BBCABFBB9103A8AULL, "vxdr_0_deeproads_cwb_0_0.vlm"},
	{0x7BBD6BE07C5DE72CULL, "6131119_m.fxe"},
	{0x7BBD6FDD53812579ULL, "6136357_m.fxe"},
	{0x7BBE008A9B14FB10ULL, "ser191_emeric6104758.cl"},
	{0x7BBE35D61601E1E2ULL, "vxdr_0_deeproads_fhj_0_0.vlm"},
	{0x7BBE65B2135DF227ULL, "vxlt_darktown_4wd_0_0.vlm"},
	{0x7BBEB8E55D3715F2ULL, "art\\vfx\\materials\\vfx_leaf_01.mao"},
	{0x7BBECD445BE891CFULL, "vxht_0_hightown_a3_auy_0_0.vlm"},
	{0x7BBED0C1B6EFE77CULL, "llac_lt_0_lowtown_0_3.dds"},
	{0x7BBF3FFF62856944ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_basic.mmh"},
	{0x7BBF43FF62856F97ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_basic.mao"},
	{0x7BBF55FF62858EE6ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_basic.msh"},
	{0x7BBF64907AF84F37ULL, "vxht_keep_n_6gz_0_0.vlm"},
	{0x7BBFD58AF10610C9ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge6.gad"},
	{0x7BC08DF6F8603474ULL, "follower_banter6164453.cl"},
	{0x7BC094C9D8F579DDULL, "570436651.ogg"},
	{0x7BC0A9D96DF8CCAFULL, "875825868.ogg"},
	{0x7BC0B31EA4A4B45BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_exit.gad"},
	{0x7BC1205582CD0E42ULL, "dae000_bodhan6166372.cl"},
	{0x7BC12D68CE9182B3ULL, "dre000_mother.bnk"},
	{0x7BC1458FC2376489ULL, "381258835.ogg"},
	{0x7BC14FE2F751D818ULL, "vxdr_primevalentrance_2su_0_2.vlm"},
	{0x7BC17579430E8684ULL, "354288644.ogg"},
	{0x7BC1CABE012B6DCFULL, "hlac_ht_mansion2b_818_45x.dds"},
	{0x7BC1D5CA525B23DAULL, "vxw_bonepitdestroyed_a3_26h_0_0.vlm"},
	{0x7BC2C128882CC8D8ULL, "226802817.ogg"},
	{0x7BC2CF1C17E2F33AULL, "137351020.ogg"},
	{0x7BC2CF90C32885E9ULL, "842196397.ogg"},
	{0x7BC2DA3E0C680DC3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_roba_0t.dds"},
	{0x7BC2EEF79E1791D4ULL, "vxca_caverns2_8kw_0_0.vlm"},
	{0x7BC2F08400BC96BEULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.mmh"},
	{0x7BC3847CF43AFDD3ULL, "157702593.ogg"},
	{0x7BC3FDBB016310ADULL, "6040661_m.fxe"},
	{0x7BC44F68CE93F203ULL, "dre000_mother.cnv"},
	{0x7BC4DB1EB80B6C5CULL, "vxht_0_hightown_a3_a6c_0_0.vlm"},
	{0x7BC4FEAC9B38E3F9ULL, "vxht_keep_6b0_0_0.vlm"},
	{0x7BC591637DA2F8E8ULL, "follower_banter6155916.cl"},
	{0x7BC59A41BF7DEE85ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\til_hightown_main_d.dds"},
	{0x7BC5FC54F99F277FULL, "vxht_0_hightown_n_bx7_0_0.vlm"},
	{0x7BC5FF19592A2FA5ULL, "vxht_mansion2b_28z_0_0.vlm"},
	{0x7BC62CAD8B3856F0ULL, "nat380_sibling6207671.cl"},
	{0x7BC63863DEE71A64ULL, "856785560.ogg"},
	{0x7BC64EE270123931ULL, "vxlt_gallowscourtyard_6wb_0_0.vlm"},
	{0x7BC6A68AF10BDF17ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge6.ani"},
	{0x7BC6BDDE735C8740ULL, "vxlt_hangedman_1vp_0_0.vlm"},
	{0x7BC6DB642E13DAC4ULL, "701068260.ogg"},
	{0x7BC738A6D188B767ULL, "t1_mub_or2.tnt"},
	{0x7BC7A428259D74F0ULL, "mag160_feynriel6020414.cl"},
	{0x7BC7D01EA4AB0440ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_exit.evt"},
	{0x7BC7D81220CE1FDAULL, "vxlt_0_lowtown_a2_f5j_0_0.vlm"},
	{0x7BC81A483734F242ULL, "vxlt_0_lowtown_a2_gzb_0_0.vlm"},
	{0x7BC869F5B9DFB9A4ULL, "m_dae360ar_wounded_coast.wbd"},
	{0x7BC893CB71B6B813ULL, "vxht_keep_57r_0_0.vlm"},
	{0x7BC8B996DA8B6E2AULL, "abi_w_aoe_knockback.bnk"},
	{0x7BC8DD78F8C29150ULL, "6202057_m.fxe"},
	{0x7BC8FEA0AB8F7C6FULL, "vxlt_0_lowtown_docks_ctl_0_0.vlm"},
	{0x7BCA0464542D7F0DULL, "vxca_cavern1_haunted_icm_0_0.vlm"},
	{0x7BCA604DB78EB46FULL, "vxlt_0_lowtown_docks_b5o_0_1.vlm"},
	{0x7BCA7806A81DB168ULL, "vxht_0_hightown_bk6_0_0.vlm"},
	{0x7BCA9A6DCE4EEFFDULL, "vxlt_0_lowtown_docks_a3_cq6_0_0.vlm"},
	{0x7BCAB120BE654269ULL, "ave260tr_long_road_smuggler.utt"},
	{0x7BCAF8B72D9CD3A6ULL, "gen00fl_anders6142500.cl"},
	{0x7BCB09C3CF5FB426ULL, "6054750_f.fxe"},
	{0x7BCB2511EA28C0B9ULL, "6099461_m.fxe"},
	{0x7BCB48B1A766FE21ULL, "410215244.ogg"},
	{0x7BCC5A70295150B0ULL, "vxht_0_hightown_a3_bs3_0_0.vlm"},
	{0x7BCC6C680B257A67ULL, "atl_debug_dxt5_dat.xml"},
	{0x7BCCA3D3553EBC67ULL, "vxht_0_hightown_a3_bp5_0_0.vlm"},
	{0x7BCCABB915429F43ULL, "183883655.ogg"},
	{0x7BCCD78F05540AA7ULL, "vxlt_darktown_4rg_0_0.vlm"},
	{0x7BCD1352635CF1C9ULL, "6035302_m.fxe"},
	{0x7BCD3F214DDEAA4FULL, "vxht_0_hightown_a3_arx_0_0.vlm"},
	{0x7BCD7E0679968D91ULL, "6141665_m.fxe"},
	{0x7BCD89FED5C18890ULL, "one280_iwan6096746.cl"},
	{0x7BCDD66D6CEE67B7ULL, "vxht_keep_n_6bu_0_0.vlm"},
	{0x7BCDEE7274116542ULL, "359355255.ogg"},
	{0x7BCE0F27F3DEB975ULL, "hlac_ca_caverns2_5802_62z.dds"},
	{0x7BCE96D84DF0D220ULL, "41549331.ogg"},
	{0x7BCED738932EB871ULL, "plt_lgt321pt_shw_ors_cache.plo"},
	{0x7BD00931702214CEULL, "6045863_m.fxe"},
	{0x7BD019FA7E7AF7E9ULL, "758141971.ogg"},
	{0x7BD047A744513C5AULL, "vxw_bonepitdestroyed_a3_21i_0_0.vlm"},
	{0x7BD07D42F82F0D5BULL, "448713925.ogg"},
	{0x7BD09D7B9A7C277EULL, "vxht_brothel_n_6fg_0_0.vlm"},
	{0x7BD0CF8400C89C7DULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.anb"},
	{0x7BD0F0387967ACFEULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.mmh"},
	{0x7BD1577EE3B1DDD4ULL, "vxht_0_hightown_bax_0_0.vlm"},
	{0x7BD195D84C425784ULL, "vxht_keep_73c_0_0.vlm"},
	{0x7BD1E303FD986103ULL, "gen00fl_merrill6128776.cl"},
	{0x7BD2B77FFA2CC889ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_32bb.mao"},
	{0x7BD2C261972842CAULL, "gen00fl_anders6119849.cl"},
	{0x7BD2C37FFA2CDD22ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_32bb.mmh"},
	{0x7BD2C57FFA2CE0D8ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_32bb.msh"},
	{0x7BD2D487BC24F9F3ULL, "6112047_m.fxe"},
	{0x7BD33EFC50ED8AC6ULL, "hlac_dr_dungeon_3216_0z.dds"},
	{0x7BD3797F70643D18ULL, "dae000cr_brat.utc"},
	{0x7BD3E124D7E47776ULL, "w_woundedcoast_238_0.msh"},
	{0x7BD3EB24D7E48874ULL, "w_woundedcoast_238_0.mmh"},
	{0x7BD3EF24D7E48E87ULL, "w_woundedcoast_238_0.mao"},
	{0x7BD4253D2DD253FFULL, "zz_vault_debug6203391.cl"},
	{0x7BD46E31EB953FFFULL, "vxht_0_hightown_n_bc6_0_0.vlm"},
	{0x7BD470F64B204825ULL, "vxht_mansion2b_25y_0_0.vlm"},
	{0x7BD49C896FAA21F7ULL, "hlac_ht_chantry_n_2534_41z.dds"},
	{0x7BD525437CFE8CFFULL, "gen00fl_aveline6120805.cl"},
	{0x7BD5EE14F7B862C5ULL, "dae000pt_location.ncs"},
	{0x7BD5F6852F9D6970ULL, "6095158_m.fxe"},
	{0x7BD60512A1BA9218ULL, "1068426025.ogg"},
	{0x7BD63BE30A677DD1ULL, "lt_0_lowtown_docks_n_10_0.rml"},
	{0x7BD63F2B8E894243ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_hed_hlk_0n.dds"},
	{0x7BD6DE230511177EULL, "vxdr_primevalentrance_393_0_1.vlm"},
	{0x7BD742DEF33A86B5ULL, "38877501.ogg"},
	{0x7BD75DBCF463E9D7ULL, "802950465.ogg"},
	{0x7BD7E30377DB4622ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v6.gad"},
	{0x7BD7E43BC1E2E1E0ULL, "scrollingtexture.mao"},
	{0x7BD7E566C70D5CF7ULL, "ht_0_hightown_a3_27_0_lc.mmh"},
	{0x7BD8DB59AE555886ULL, "gen_im_trs_chn_andraste_02.uti"},
	{0x7BD9141BED066C31ULL, "vxht_mansion2_3d3_0_0.vlm"},
	{0x7BD9232FAB8A6C55ULL, "w_bonepit_200_0.phy"},
	{0x7BD97BED3BE1F40FULL, "qun221_petrice6098710.cl"},
	{0x7BD9841FC9EE3AE1ULL, "vxlt_undercityserialkiller_1g9_0_0.vlm"},
	{0x7BD9B8E4995506A5ULL, "415928671.ogg"},
	{0x7BD9D1D4E36CF301ULL, "lgt311cr_shw_mer_agatha.utc"},
	{0x7BDA40F890D4736FULL, "122487014.ogg"},
	{0x7BDA7A3EE38E741EULL, "hightown_fem_elfservant_mob6169096.cl"},
	{0x7BDA827966E53191ULL, "hlac_ca_caverns2_5802_57z.dds"},
	{0x7BDA849598DEC772ULL, "vxht_keep_n_6zt_0_0.vlm"},
	{0x7BDADBCADC8D2CEAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3lrg.mao"},
	{0x7BDADFCADC8D3339ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3lrg.mmh"},
	{0x7BDAE9CADC8D4427ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3lrg.msh"},
	{0x7BDB1A899BF0839CULL, "m_mag291ar_gallows.wbd"},
	{0x7BDB6CF487AC4FDDULL, "vxlt_gallowstemplar_f_3re_0_0.vlm"},
	{0x7BDBC6ACCA171961ULL, "hlac_ht_mansion2b_818_16z.dds"},
	{0x7BDBD090DA5CB52AULL, "pro000_merchant6051628.cl"},
	{0x7BDC2E7D7C00AD8BULL, "8318638.ogg"},
	{0x7BDC332701915027ULL, "6141447_m.fxe"},
	{0x7BDC45AEDDBCBE73ULL, "6102676_f.fxe"},
	{0x7BDC4DF3FC465339ULL, "143573910.ogg"},
	{0x7BDC5B895D88D69CULL, "fxc_spider_webline.mao"},
	{0x7BDCA77391ABFF12ULL, "lt_0_lowtown_a2_n_34_0_ll.mmh"},
	{0x7BDD0C79C3953353ULL, "44595134.ogg"},
	{0x7BDD5E646B643F3FULL, "6069792_m.fxe"},
	{0x7BDD9BAC1FF6823CULL, "lt_0_lowtown_docks_a3_n_17_0_lc.mmh"},
	{0x7BDE13A2C407E733ULL, "gen00fl_aveline6132716.cl"},
	{0x7BDE838605A8F7D5ULL, "qun231_varnell6078104.cl"},
	{0x7BDECF387973B1BDULL, "art\\vfx\\combat\\vfx_str_g_pound_lp_c.anb"},
	{0x7BDF060377E1A019ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v6.evt"},
	{0x7BDF92FD9055FC02ULL, "llac_ht_chantry_n_2534_46.dds"},
	{0x7BDFAC2AD8D890CAULL, "1035322262.ogg"},
	{0x7BDFBF5BB3634036ULL, "vxca_caverns1_mines_e8o_0_0.vlm"},
	{0x7BDFDE945F3B98FDULL, "darktown_male_sickly_sta6175319.cl"},
	{0x7BE06FE42C7D330DULL, "vxdr_primevaltomb_5n2_0_0.vlm"},
	{0x7BE106A30899F63FULL, "humanfemale_c.fxa"},
	{0x7BE19D59768976A1ULL, "6099343_f.fxe"},
	{0x7BE1ACA7723AC97BULL, "vxw_blightlands_bf_0_0.vlm"},
	{0x7BE1DCC5976B1859ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_ichf.msh"},
	{0x7BE1E2667F2514F9ULL, "vxht_keep_6h6_0_0.vlm"},
	{0x7BE1E6C5976B2947ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_ichf.mmh"},
	{0x7BE21F1AED916FC9ULL, "fx_dust06.dds"},
	{0x7BE2220C68829F05ULL, "cs_pro000cs_act1_crim_mage.bnk"},
	{0x7BE240338BEE6A2EULL, "hlac_w_sundermount_0_2y.dds"},
	{0x7BE27B651E541313ULL, "6025091_m.fxe"},
	{0x7BE27E749F7B6F3EULL, "qun120cr_javaris6042438.cl"},
	{0x7BE2EEB020AB46AFULL, "ico_gloves_medium_2.dds"},
	{0x7BE34160B3DD4DEEULL, "lt_0_lowtown_docks_a3_210_0.msh"},
	{0x7BE34F60B3DD653FULL, "lt_0_lowtown_docks_a3_210_0.mao"},
	{0x7BE353BB7F58008EULL, "6041489_m.fxe"},
	{0x7BE35B60B3DD79DCULL, "lt_0_lowtown_docks_a3_210_0.mmh"},
	{0x7BE35FAEC92A193DULL, "237254943.ogg"},
	{0x7BE4609F6C53B68AULL, "vxht_chantry_2eb_0_0.vlm"},
	{0x7BE47363EA15B108ULL, "llac_w_encounter1_0_3.dds"},
	{0x7BE4749CAE1C4947ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_3.pwk"},
	{0x7BE4E3182F2BF4AAULL, "880579878.ogg"},
	{0x7BE5A2705C5EB17DULL, "w_woundedcoast_122_0.mao"},
	{0x7BE5A3C056277B72ULL, "blood_imp_02.mao"},
	{0x7BE5A6705C5EB7B6ULL, "w_woundedcoast_122_0.mmh"},
	{0x7BE5B0705C5EC8B4ULL, "w_woundedcoast_122_0.msh"},
	{0x7BE5C212324EC9B8ULL, "sum_conj_05.mao"},
	{0x7BE5F5DDEF096E82ULL, "6217303_m.fxe"},
	{0x7BE61515CBD02E99ULL, "hlac_lt_0_lowtown_docks_n_-196611_2y.dds"},
	{0x7BE6B8B14B7F1165ULL, "pro000_colonel_ewald6040599.cl"},
	{0x7BE6B8CFFA03C168ULL, "gencr_dalish_warrior_2h.utc"},
	{0x7BE6FA17E3632320ULL, "vxht_bartrand_n_5qo_0_0.vlm"},
	{0x7BE70DC6EE6CFD8CULL, "393374049.ogg"},
	{0x7BE72C940851FC7CULL, "139232837.ogg"},
	{0x7BE737CBD7ADE0C0ULL, "hlac_lt_0_lowtown_docks_a3_-65539_6z.dds"},
	{0x7BE7BEA50D4D5789ULL, "follower_banter6155991.cl"},
	{0x7BE7F142B607D704ULL, "945021302.ogg"},
	{0x7BE86F485587C1C0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maga_0n.dds"},
	{0x7BE8801D8649EB9BULL, "hlac_lt_0_lowtown_docks_a3_n_65534_1y.dds"},
	{0x7BE8C7BA8733E18AULL, "241019124.ogg"},
	{0x7BE93E2A0D3D81B0ULL, "vxht_0_hightown_a3_be1_0_0.vlm"},
	{0x7BE965907294DD88ULL, "dae000ip_crowd_lm4_talk15.utp"},
	{0x7BE9A773D155D0BDULL, "gen00fl_merrill6121157.cl"},
	{0x7BEA362AB428D4CAULL, "hlac_dr_dungeon_3217_8y.dds"},
	{0x7BEA4CAE1CC88C38ULL, "vxht_brothel_n_6d7_0_0.vlm"},
	{0x7BEA6B38E243565FULL, "6099701_f.fxe"},
	{0x7BEB170F82C9A17BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj01.phy"},
	{0x7BEB60EE15830BD1ULL, "gen00fl_fenris6107157.cl"},
	{0x7BEB6FD81135F723ULL, "1012665299.ogg"},
	{0x7BEBB0F3BCC21476ULL, "gen_im_trs_dwv_junk2_04.uti"},
	{0x7BEC0E0377EC3808ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v6.ani"},
	{0x7BEC98F129B708F2ULL, "vxlt_house_g3_0_0.vlm"},
	{0x7BECABF7FE903A21ULL, "91701601.ogg"},
	{0x7BECF4270350546AULL, "vrc000pt_main.ncs"},
	{0x7BED74DDFDCCDEEEULL, "279956596.ogg"},
	{0x7BEDFA89199EA47BULL, "mag370_meredith6172377.cl"},
	{0x7BEE06B939CB8D52ULL, "470320628.ogg"},
	{0x7BEE0E88DBD43A1DULL, "6127063_m.fxe"},
	{0x7BEE59F98825576FULL, "732132131.ogg"},
	{0x7BEE725419E60C73ULL, "vxdr_primevaltomb_53o_0_0.vlm"},
	{0x7BEE7A91D47CDD51ULL, "961412646.ogg"},
	{0x7BEE886841DB9DB4ULL, "art\\vfx\\combat\\vfx_m_cre_heal_imp_c.impa.ani"},
	{0x7BEEE1C11E734B8DULL, "vxdr_primevaltomb_5c1_0_0.vlm"},
	{0x7BEEE2BA1D2EA471ULL, "6080733_m.fxe"},
	{0x7BEF225929DED85CULL, "6027946_m.fxe"},
	{0x7BEF493CB91F7FA4ULL, "vxw_bonepitmining_a2_1kh_0_0.vlm"},
	{0x7BEF5AD97054DE6AULL, "6110615_m.fxe"},
	{0x7BEF64158B018619ULL, "vxlt_0_lowtown_docks_a3_n_d3h_0_0.vlm"},
	{0x7BF05344F9901894ULL, "vxlt_undercitydungeon_13m_0_0.vlm"},
	{0x7BF075A1BDB7135DULL, "6064050_m.fxe"},
	{0x7BF078A45A7EC627ULL, "856861021.ogg"},
	{0x7BF087D8573F87A3ULL, "follower_barks6153167.cl"},
	{0x7BF0A60E641366BCULL, "6033792_m.fxe"},
	{0x7BF0F65CDC83E63FULL, "688087765.ogg"},
	{0x7BF15C6A0281459CULL, "hlac_lt_0_lowtown_a2_n_c_65536_10z.dds"},
	{0x7BF1B19BB4686D76ULL, "241219089.ogg"},
	{0x7BF204710CC15C32ULL, "static_static_p50.psh.fxo.dx11"},
	{0x7BF23D3C93CD3D43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\charge.evt"},
	{0x7BF33D7632F6E28AULL, "hlac_ca_caverns1_24421_6z.dds"},
	{0x7BF34DB4C53307EBULL, "gen00fl_anders6125373.cl"},
	{0x7BF35FE8B411A81AULL, "chargen.gfx"},
	{0x7BF386EAEA2B4A06ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_enter_elf_male.ani"},
	{0x7BF3A592A15281F7ULL, "vxlt_0_lowtown_a2_n_fgj_0_0.vlm"},
	{0x7BF440920AF40837ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_camp_fire.gad"},
	{0x7BF493E23227B81FULL, "397654968.ogg"},
	{0x7BF49E124F6B5C7AULL, "rdr300ar_encounter.lst"},
	{0x7BF4B29F1FF74BF3ULL, "lgt160pt_cod_avern_lett_0.utp"},
	{0x7BF4B6707007B2C3ULL, "vxht_0_hightown_a3_4ap_0_0.vlm"},
	{0x7BF4CB6908440509ULL, "vxlt_gallowstemplar_n_5lx_0_1.vlm"},
	{0x7BF4ECCCA8898423ULL, "llac_dr_primevaltomb_657_5.dds"},
	{0x7BF4F7E2C57E8681ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_exit.ani"},
	{0x7BF5377B182BEB13ULL, "vxht_0_hightown_a3_ap2_0_0.vlm"},
	{0x7BF56BF4D5593BA0ULL, "vxht_bartrand_n_5tl_0_0.vlm"},
	{0x7BF5856841E1B716ULL, "art\\vfx\\combat\\vfx_m_cre_heal_imp_c.impa.gad"},
	{0x7BF593EF0F1FDAD6ULL, "art\\vfx\\combat\\vfx_g_appearance_c.cess.evt"},
	{0x7BF5CE8D4DF2AA23ULL, "771614050.ogg"},
	{0x7BF64B3600B3D3FFULL, "vxht_0_hightown_n_9as_0_0.vlm"},
	{0x7BF6CBFFA4EF1187ULL, "739493866.ogg"},
	{0x7BF6DEEA6E6D89EDULL, "6116377_m.fxe"},
	{0x7BF7009CDE7954E1ULL, "creaturetypes.gda"},
	{0x7BF7A7164905240CULL, "lt_0_lowtown_docks_a3_n_29_0.rml"},
	{0x7BF7AEC5177CDF68ULL, "qun211_party_comment6072242.cl"},
	{0x7BF8B1E10C998AE9ULL, "zz_approval_debug6003103.cl"},
	{0x7BF9163C93D319C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\charge.gad"},
	{0x7BF930B50B55D3DDULL, "6042388_m.fxe"},
	{0x7BF95B44888A79ADULL, "vxw_sundermount_2kk_0_0.vlm"},
	{0x7BFA38C5B35276A8ULL, "766188000.ogg"},
	{0x7BFA55920AF897C4ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_camp_fire.evt"},
	{0x7BFB02E2C5830523ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_exit.gad"},
	{0x7BFB2C9FD989B5D0ULL, "vxlt_0_lowtown_a3_n_c_fit_0_0.vlm"},
	{0x7BFB4568CEC32380ULL, "dre000_mother.stg"},
	{0x7BFB7085ABA3DCFCULL, "docks_scene_dockworkers6176495.cl"},
	{0x7BFB943EB0B9AFFCULL, "vxdr_primevaltomb_5ao_0_1.vlm"},
	{0x7BFBB636124AC0EDULL, "165147351.ogg"},
	{0x7BFBB967259639B7ULL, "6121046_f.fxe"},
	{0x7BFBD4DD237CBC9BULL, "qun111_ketojan6040852.cl"},
	{0x7BFC31E5A6CB4400ULL, "464884550.ogg"},
	{0x7BFC8AEF0F25EA61ULL, "art\\vfx\\combat\\vfx_g_appearance_c.cess.gad"},
	{0x7BFC9262CCCFB510ULL, "qun221_petrice6075963.cl"},
	{0x7BFC932BB91207A2ULL, "art\\vfx\\textures\\vfx_icecrystal01.dds"},
	{0x7BFC9DA99CDB9EC1ULL, "774867831.ogg"},
	{0x7BFCA06841E80475ULL, "art\\vfx\\combat\\vfx_m_cre_heal_imp_c.impa.evt"},
	{0x7BFD4A83693E9AF6ULL, "43005286.ogg"},
	{0x7BFD83BA20FAC10EULL, "hlac_lt_0_lowtown_a2_n_131072_5y.dds"},
	{0x7BFDB0324E583397ULL, "character_character_p84.psh.fxo"},
	{0x7BFE60A002117A97ULL, "6171057_m.fxe"},
	{0x7BFED4EE364EE841ULL, "vxlt_0_lowtown_a3_n_c_i0q_0_0.vlm"},
	{0x7BFEF0FB0F34B549ULL, "484791662.ogg"},
	{0x7BFF05F61E780E54ULL, "6084703_m.fxe"},
	{0x7BFF076C0AE0ED38ULL, "189794938.ogg"},
	{0x7BFF253C93D79EA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\charge.ani"},
	{0x7BFF3079DC33704EULL, "687632347.ogg"},
	{0x7BFF6799A300B6CBULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.dura.evt"},
	{0x7BFFBCA70CC491D8ULL, "hlac_ca_cave_2194_14z.dds"},
	{0x7C0007DB0F19A13FULL, "zz_followers_banter6159072.cl"},
	{0x7C01BBC463B35C5EULL, "257236249.ogg"},
	{0x7C01D77822685EEEULL, "gen_im_arm_cht_med_set_high.uti"},
	{0x7C01DAE5227F02D5ULL, "dae000_dalish_apprentices6147611.cl"},
	{0x7C01DFE2C588E808ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_exit.evt"},
	{0x7C01E1BBFF0D5442ULL, "vxlt_0_lowtown_a2_gnn_0_0.vlm"},
	{0x7C0209C97A1EF87EULL, "ht_0_hightown_n_29_0.rml"},
	{0x7C028BEF0F2A570FULL, "art\\vfx\\combat\\vfx_g_appearance_c.cess.ani"},
	{0x7C02AA052E80DE9DULL, "hlac_lt_0_lowtown_docks_-131075_3z.dds"},
	{0x7C02B38BC994A261ULL, "135384157.ogg"},
	{0x7C034347FEB84D07ULL, "6070480_m.fxe"},
	{0x7C03455C9ADD404DULL, "vxlt_0_lowtown_a2_n_c_hkl_0_0.vlm"},
	{0x7C03DBB6C2EA569BULL, "vxlt_0_lowtown_n_e84_0_0.vlm"},
	{0x7C03E780111B7523ULL, "6098330_m.fxe"},
	{0x7C03FBAC891A5072ULL, "dae000_bodahn.stg"},
	{0x7C0428E843719E81ULL, "vxht_0_hightown_a3_a80_0_0.vlm"},
	{0x7C042DA632411C97ULL, "vxw_bonepitdestroyed_a3_98_0_0.vlm"},
	{0x7C045A871E06AD03ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.impa.ani"},
	{0x7C049EE584EB23C0ULL, "vxht_0_hightown_4an_0_0.vlm"},
	{0x7C04A7A39F5F1541ULL, "gen00fl_fenris6128059.cl"},
	{0x7C04BF75C7FC4983ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gtowera.mao"},
	{0x7C04C375C7FC50D0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gtowera.mmh"},
	{0x7C04D175C7FC68EAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_gtowera.msh"},
	{0x7C04FFEB7763F899ULL, "6018844_m.fxe"},
	{0x7C050BA131CCD10EULL, "6111128_m.fxe"},
	{0x7C05104D21C3AA4EULL, "6147141_m.fxe"},
	{0x7C05CD3D30D0A880ULL, "combat_lc_r1_a3.gda"},
	{0x7C05DA2C6EC110F2ULL, "vxht_keep_n_6kw_0_0.vlm"},
	{0x7C05DFB4889DEAE5ULL, "23981389.ogg"},
	{0x7C067CB6D9A77926ULL, "gen00fl_anders6125428.cl"},
	{0x7C068099A306FF8CULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.dura.gad"},
	{0x7C068952945655F3ULL, "478280570.ogg"},
	{0x7C07041041CACED6ULL, "6075327_m.fxe"},
	{0x7C0718899C15BB16ULL, "m_mag291ar_gallows.bnk"},
	{0x7C07306800B4BD38ULL, "vxht_brothel_n_6j5_0_0.vlm"},
	{0x7C0736DF52EA180AULL, "dae000_bartrand6201389.cl"},
	{0x7C076FEAEA3BCC00ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_enter_elf_male.gad"},
	{0x7C07A8CFFD6829BBULL, "6133082_m.fxe"},
	{0x7C07CD217A80922DULL, "vxw_sundermount_2lj_0_0.vlm"},
	{0x7C07F6CF363AE891ULL, "ave331_jeven6162381.cl"},
	{0x7C0839E0C1164F53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v5.ani"},
	{0x7C084DB4FAEF963EULL, "ht_0_hightown_n_27_0_ll.mmh"},
	{0x7C086D920B04FDBDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\dg_tk_camp_fire.ani"},
	{0x7C094FA079A149F7ULL, "6020550_f.fxe"},
	{0x7C09708BDDDE49C6ULL, "hlac_ht_bartrand_n_2995_2x.dds"},
	{0x7C0986B6A4A2A7E5ULL, "vxw_bonepitdestroyed_a3_27p_0_0.vlm"},
	{0x7C09EF9A161F945BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_loop.ani"},
	{0x7C0A128D5FC77B8DULL, "lgt110_orz_harrowmont6109022.cl"},
	{0x7C0A554356456C69ULL, "zz_isa_debug6175457.cl"},
	{0x7C0B8C194900F053ULL, "6076604_m.fxe"},
	{0x7C0B9C4F586CEA9EULL, "vxw_bonepitdestroyed_a3_246_0_0.vlm"},
	{0x7C0C199174FD8B41ULL, "vxca_caverns1_ghk_0_0.vlm"},
	{0x7C0C220F404E4C17ULL, "character_character_p75.psh.fxo"},
	{0x7C0C7F99A30B698AULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.dura.ani"},
	{0x7C0C84004031E45BULL, "vxw_woundedcoast_1kb_0_0.vlm"},
	{0x7C0CFEA073A0AA04ULL, "root_lyriumghost.dds"},
	{0x7C0D1A10C499F278ULL, "vxlt_undercityserialkiller_1c1_0_1.vlm"},
	{0x7C0D44E85F32C9FEULL, "cd_seb221_brett_harimann.bnk"},
	{0x7C0DA3866404BB4EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_maga_0n.dds"},
	{0x7C0E9F977392E5FEULL, "vxw_sundermount_27h_0_0.vlm"},
	{0x7C0EBCBD6019649CULL, "vxlt_undercityencounter_1lu_0_0.vlm"},
	{0x7C0ECFC59791F07CULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_ichf.phy"},
	{0x7C0F055C44B6585FULL, "pro000_captain_wright6047475.cl"},
	{0x7C0F0D8060593964ULL, "6120896_m.fxe"},
	{0x7C0F3078B06A3C28ULL, "hlac_lt_0_lowtown_-65535_7y.dds"},
	{0x7C0F595965050C64ULL, "isa311_isabela6095640.cl"},
	{0x7C1004D232FBD6EEULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv1.ani"},
	{0x7C100DE799B0B85FULL, "llac_w_bonepitdestroyed_a3_0_1.dds"},
	{0x7C1023EFA09FDCCEULL, "gen_im_npc_wep_shd_genb.uti"},
	{0x7C108553E7B5D896ULL, "211142241.ogg"},
	{0x7C11615DECAE59E4ULL, "6096432_m.fxe"},
	{0x7C1172871E1160FAULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.impa.evt"},
	{0x7C11969EF617E531ULL, "6096522_f.fxe"},
	{0x7C119ADBAB50B7B4ULL, "follower_barks6153724.cl"},
	{0x7C11BAE30B6D9B3BULL, "llac_dr_0_deeproads_5845_10.dds"},
	{0x7C11CC9CF9E1247DULL, "gen00fl_aveline6106380.cl"},
	{0x7C122A8100ABAABBULL, "llac_lt_backalley_g_0_9.dds"},
	{0x7C126CE8B42C6F7BULL, "chargen.ncs"},
	{0x7C12705A612C17B0ULL, "vxht_mansion2_2id_0_0.vlm"},
	{0x7C12BCF8E2790A59ULL, "plt_ave100pt_start.plo"},
	{0x7C12EEB2728376B8ULL, "cod_bks_lite_tombs_last_4.utp"},
	{0x7C131B7B1B9B3221ULL, "mcr371_scared_mage_f6086920.cl"},
	{0x7C13822A13B9C2CEULL, "6140041_m.fxe"},
	{0x7C13B55D87693E47ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyn_0d.dds"},
	{0x7C13DA41F28A6F32ULL, "mi_gla_flare.mao"},
	{0x7C1406FEA6C2CE68ULL, "llac_w_bonepit_2_0.dds"},
	{0x7C1413662BAD3D60ULL, "dae000_bodhan6146817.cl"},
	{0x7C1416E81C46E8DEULL, "vxlt_0_lowtown_a2_n_c_fg5_0_0.vlm"},
	{0x7C143C7CC8B71414ULL, "hlac_ht_brothel_n_3365_0z.dds"},
	{0x7C144C0960B72972ULL, "vxht_keep_n_6fp_0_0.vlm"},
	{0x7C14A889FF0599D5ULL, "6153198_m.fxe"},
	{0x7C1531E0C120CB8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v5.evt"},
	{0x7C16288ADDE8DDF9ULL, "6096015_m.fxe"},
	{0x7C163EFE6C76AAADULL, "vxw_sundermount_2qi_0_0.vlm"},
	{0x7C16AEF3F052E33CULL, "28614348.ogg"},
	{0x7C16C61CE47ECAE8ULL, "164911640.ogg"},
	{0x7C16D4E657F69751ULL, "zz_ser_debug6091989.cl"},
	{0x7C16FE0F82EEB11AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj01.msh"},
	{0x7C17100F82EED060ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj01.mmh"},
	{0x7C178A6D1AC11F3EULL, "t3_wdg_gw2.tnt"},
	{0x7C1791D019574B0AULL, "6099233_m.fxe"},
	{0x7C17C389C523A3A7ULL, "pro000_opening_pc6001213.cl"},
	{0x7C17F79A162BDF52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_loop.evt"},
	{0x7C180A14DBC60C4EULL, "one100dg_vanard6047400.cl"},
	{0x7C1832D12C47C412ULL, "6122752_f.fxe"},
	{0x7C183EBF3AD327DFULL, "mcr370_desten6087321.cl"},
	{0x7C184B6D527530A9ULL, "w_sundermount_166_0.phy"},
	{0x7C186845966461D2ULL, "6148218_m.fxe"},
	{0x7C18709BBAECB562ULL, "6094678_m.fxe"},
	{0x7C1871871E177DEDULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.impa.gad"},
	{0x7C18902682F9CEF4ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_ice_c.impa.ani"},
	{0x7C189503BF08391BULL, "6105595_m.fxe"},
	{0x7C18EED2A1592756ULL, "10213311.ogg"},
	{0x7C190A3C9C9C1B66ULL, "vxlt_gallowscourtyard_3bs_0_0.vlm"},
	{0x7C194E0698763194ULL, "6023898_m.fxe"},
	{0x7C196E8286FD1969ULL, "qcr210ip_to_hightown.utp"},
	{0x7C199D124F8A76FDULL, "rdr300ar_encounter.are"},
	{0x7C19AA47E51ECB1CULL, "dae110ip_at_docks.utp"},
	{0x7C19F1498BACF7ECULL, "m_mag311ar_warehouse_2.bnk"},
	{0x7C1A09EE50A30486ULL, "vxw_sundermount_20i_0_0.vlm"},
	{0x7C1A93EC32446497ULL, "character_character_p26.psh.fxo"},
	{0x7C1A9ED38F01C924ULL, "vxw_bonepitmining_a2_1zo_0_0.vlm"},
	{0x7C1B00055A4377B2ULL, "gen00fl_fenris6126618.cl"},
	{0x7C1B574DE81A3BEBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\shuffle.evt"},
	{0x7C1B6088A144CED8ULL, "mag370_playerarriveearly.cut"},
	{0x7C1BE0A8A2462B77ULL, "ef_gen_dalish_archer.mor"},
	{0x7C1C10E0C126B2FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v5.gad"},
	{0x7C1CAB3FCBAD0A23ULL, "hlac_lt_0_lowtown_a2_n_-65537_0x.dds"},
	{0x7C1CE7F16FCCE0B4ULL, "dae011_gossips6143748.cl"},
	{0x7C1CFCD23306533FULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv1.evt"},
	{0x7C1D11746588FE7EULL, "vxw_sundermount_28i_0_0.vlm"},
	{0x7C1D1EBDDF5E0A7AULL, "hlac_w_encounter2_camp_327681_0y.dds"},
	{0x7C1D39EAA680C61AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03tree.msh"},
	{0x7C1D47EAA680DE53ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03tree.mao"},
	{0x7C1D4BEAA680E560ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03tree.mmh"},
	{0x7C1E060123490D96ULL, "6161602_m.fxe"},
	{0x7C1E8D2682FE3556ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_ice_c.impa.gad"},
	{0x7C1EBE23BAE2AA79ULL, "vxca_cave_1ag_0_1.vlm"},
	{0x7C1F169A163232E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_loop.gad"},
	{0x7C1FB10A3C7DFBE3ULL, "mcr310_confused_guard6077111.cl"},
	{0x7C2017FCCDF58587ULL, "vxlt_warehouse_i3_0_0.vlm"},
	{0x7C20362FABC68F70ULL, "w_bonepit_200_0.msh"},
	{0x7C20442FABC6A68AULL, "w_bonepit_200_0.mmh"},
	{0x7C20482FABC6ADC1ULL, "w_bonepit_200_0.mao"},
	{0x7C2062C3F444354DULL, "dae000_gamlen6149466.cl"},
	{0x7C20E2037E08C003ULL, "6127568_m.fxe"},
	{0x7C210EC11ED8AA89ULL, "dae010_commoner_m2.utc"},
	{0x7C21704DE81ED1ACULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\shuffle.gad"},
	{0x7C21777211CB5C83ULL, "mag131_keran6094512.cl"},
	{0x7C220D30372CA87EULL, "and121_karl6094005.cl"},
	{0x7C2240876094101DULL, "vxw_woundedcoast_ga_0_0.vlm"},
	{0x7C224D1FF4C29AA2ULL, "6080183_m.fxe"},
	{0x7C22A82A9B09BCF7ULL, "12555650.ogg"},
	{0x7C22B0D64214A11FULL, "hlac_lt_darktown_-1_90z.dds"},
	{0x7C22BDE652AD41F2ULL, "vxht_keep_n_6aq_0_0.vlm"},
	{0x7C22D5DFE8114F69ULL, "hlac_w_encounter1_-65531_0y.dds"},
	{0x7C231465DBA99680ULL, "578404255.ogg"},
	{0x7C231786ADFB326CULL, "vxw_sundermount_31i_0_0.vlm"},
	{0x7C2345B7644521DAULL, "vis121_saemus6099368.cl"},
	{0x7C23DA9FC3A14808ULL, "gen00fl_merrill6128971.cl"},
	{0x7C23DDD2330C3D28ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\attackv1.gad"},
	{0x7C248C89E1545ABBULL, "6139682_m.fxe"},
	{0x7C24B0DB5E6CC32DULL, "vxw_sundermount_2zh_0_0.vlm"},
	{0x7C24BE22713967E7ULL, "6031708_m.fxe"},
	{0x7C2528A9B53E9504ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medd.msh"},
	{0x7C253AA9B53EB46DULL, "art\\characters\\playercharacter\\humanfemale\\arm_medd.mao"},
	{0x7C253EA9B53EBAA6ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medd.mmh"},
	{0x7C25660411A20BB7ULL, "618424332.ogg"},
	{0x7C258F426F8D40E5ULL, "169935031.ogg"},
	{0x7C25A826830481B5ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_ice_c.impa.evt"},
	{0x7C26DA22885A7A52ULL, "6145118_m.fxe"},
	{0x7C26E238937940D6ULL, "plt_lgt321pt_shw_ors_cache.nss"},
	{0x7C27CCFB34502C79ULL, "vxlt_0_lowtown_e0q_0_0.vlm"},
	{0x7C2841CD5ADC831DULL, "938216794.ogg"},
	{0x7C286F4DE824EEAAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\shuffle.ani"},
	{0x7C287BCB42991D06ULL, "vxw_sundermount_25j_0_0.vlm"},
	{0x7C28E2B3B7DBCA5DULL, "405299493.ogg"},
	{0x7C2910B080F7E1A4ULL, "vxw_bonepitmining_a2_1wl_0_0.vlm"},
	{0x7C292EB2F9237006ULL, "210866111.ogg"},
	{0x7C293464BE2C47B3ULL, "llac_lt_0_lowtown_n_-65537_2.dds"},
	{0x7C2967BA241E155BULL, "vxw_woundedcoast_1qd_0_0.vlm"},
	{0x7C2A1C8DF00AF9B9ULL, "vxdr_dungeon_1m1_0_0.vlm"},
	{0x7C2A57961F452C6DULL, "vxht_mansion2b_2v8_0_0.vlm"},
	{0x7C2A625C0CF3A4DAULL, "scaleform_sf_main_p0.vsh.fxo.dx11.assembly"},
	{0x7C2A6B8C084D69D6ULL, "vxlt_0_lowtown_docks_a3_n_bqw_0_0.vlm"},
	{0x7C2A72524422FB84ULL, "vxw_encounter1_an_0_0.vlm"},
	{0x7C2B0982AAAF42E9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_oldmanwalkf_1_monique.ani"},
	{0x7C2B217126FA37F6ULL, "6146872_m.fxe"},
	{0x7C2B268F6AACAE44ULL, "mag170_recruit016080836.cl"},
	{0x7C2B532259CDAA00ULL, "vxlt_gallowstemplar_n_2ed_0_0.vlm"},
	{0x7C2B64A2871A8DD8ULL, "qun120_arishok6051717.cl"},
	{0x7C2BAE364AB30DE1ULL, "mag121_eau_banter6121085.cl"},
	{0x7C2BB7F27F76BA58ULL, "vxca_caverns1_mines_jg7_0_0.vlm"},
	{0x7C2C147F708D4B16ULL, "6069883_m.fxe"},
	{0x7C2C724C421BB478ULL, "llac_lt_0_lowtown_a2_n_65536_12.dds"},
	{0x7C2D20C293D76DE3ULL, "104354741.ogg"},
	{0x7C2D7A1DC9AA2C0CULL, "6151391_m.fxe"},
	{0x7C2D994F01764BD6ULL, "6150985_m.fxe"},
	{0x7C2E16F1001282B6ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_2y.dds"},
	{0x7C2E3B335962F0DBULL, "hlac_w_encounter2_camp_196612_0z.dds"},
	{0x7C2E89D9BFEB9E07ULL, "vxlt_warehouse_r2_0_0.vlm"},
	{0x7C2F33689D319DA0ULL, "vxht_bartrand_n_5hh_0_0.vlm"},
	{0x7C2F35A9B5470071ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medd.phy"},
	{0x7C2FBEE20764C944ULL, "zz_and_debug6134055.cl"},
	{0x7C3009F074C752EFULL, "gen00fl_aveline6114037.cl"},
	{0x7C30567F47F6E53DULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax103a_0n.dds"},
	{0x7C30E43A6EB816DDULL, "6205888_m.fxe"},
	{0x7C311B25B7CBB8E8ULL, "6027371_m.fxe"},
	{0x7C319FA67C1D856DULL, "vxht_bartrand_n_66p_0_0.vlm"},
	{0x7C31C172D3A3DF3CULL, "vxlt_0_lowtown_docks_a3_n_d31_0_0.vlm"},
	{0x7C31DF504FCD3D90ULL, "gen00fl_isabela6079017.cl"},
	{0x7C322482AAB58F6BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_oldmanwalkf_1_monique.gad"},
	{0x7C32594C3DD2FA80ULL, "mrl310_merrill6144402.cl"},
	{0x7C3276901BF9F4C0ULL, "vxw_encounter2_camp_5w_0_0.vlm"},
	{0x7C32C1A67F64E325ULL, "6126697_m.fxe"},
	{0x7C331086BA50A6D5ULL, "612727206.ogg"},
	{0x7C331D4D702289F1ULL, "6160611_m.fxe"},
	{0x7C335687A2A6A922ULL, "hlac_lt_0_lowtown_-2_6z.dds"},
	{0x7C337EEA06DCD0B2ULL, "qun111_ketojan6040932.cl"},
	{0x7C33F53AF715D619ULL, "follower_banter6151263.cl"},
	{0x7C34716132FF805BULL, "732030264.ogg"},
	{0x7C34948409B60E92ULL, "vxw_woundedcoast_it_0_0.vlm"},
	{0x7C349ACA8F23CDE6ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig2_sa.dds"},
	{0x7C34B7F8E295E86EULL, "plt_ave100pt_start.nss"},
	{0x7C34C1F957EF4589ULL, "764082046.ogg"},
	{0x7C35355D10F62922ULL, "hlac_ca_caverns2_5802_44y.dds"},
	{0x7C35662D1ED1B4A2ULL, "ht_bartrand_n_9156.rml"},
	{0x7C357226E626B452ULL, "6055233_m.fxe"},
	{0x7C35F615F8366345ULL, "6053931_m.fxe"},
	{0x7C36509168F6EE4AULL, "106361315.ogg"},
	{0x7C36D1BE5E56DF67ULL, "vxca_cavern1_haunted_ima_0_0.vlm"},
	{0x7C374305D054FFA0ULL, "vxht_keep_4n3_0_0.vlm"},
	{0x7C3765BF5D5B6704ULL, "6126775_m.fxe"},
	{0x7C37912E243D9CDCULL, "isa120_hayder6075451.cl"},
	{0x7C37C028847289A4ULL, "vxca_cavern1_haunted_j29_0_1.vlm"},
	{0x7C37D7D19ADCFAB6ULL, "vxlt_hangedman_20k_0_0.vlm"},
	{0x7C38361FCDDAE8D6ULL, "vxw_bonepitdestroyed_a3_n4_0_0.vlm"},
	{0x7C3887EBC48FB0FEULL, "llac_w_encounter2_camp_4_0.dds"},
	{0x7C38C8A4EB63CE7BULL, "mag360_thrask6043907.cl"},
	{0x7C38E42F36191404ULL, "vxw_encounter1_fo_0_0.vlm"},
	{0x7C38F2CE8770CF0CULL, "ss_isaachawke1.cut"},
	{0x7C392D16CD65C22EULL, "mag210_arianni6024575.cl"},
	{0x7C397A1530411C7BULL, "vxw_sundermount_2oq_0_0.vlm"},
	{0x7C39921AB3D21293ULL, "mag291_vincento6027712.cl"},
	{0x7C39A0519ED3FE00ULL, "495422824.ogg"},
	{0x7C3A4E9027EC50A5ULL, "132460957.ogg"},
	{0x7C3B1BC8D147C2CEULL, "6127584_f.fxe"},
	{0x7C3B39813AE3AB15ULL, "follower_banter6141778.cl"},
	{0x7C3B44C3A9F88C42ULL, "gen00fl_aveline6115279.cl"},
	{0x7C3C06C97A506B22ULL, "ht_0_hightown_n_29_0.anb"},
	{0x7C3D51CA38886D34ULL, "107165112.ogg"},
	{0x7C3DEF81EF9FDB02ULL, "6101333_f.fxe"},
	{0x7C3F31297EAD4A31ULL, "6106387_m.fxe"},
	{0x7C4011836E139DEDULL, "vxht_bartrand_n_63o_0_0.vlm"},
	{0x7C403C1EB2DE8BDFULL, "fen110_fenris6118091.cl"},
	{0x7C40746968C1F0D8ULL, "10543692.ogg"},
	{0x7C407B63F7D723D3ULL, "clf221_jalen6070339.cl"},
	{0x7C40A48AC6D9A780ULL, "6088398_m.fxe"},
	{0x7C40C407D17858B1ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffb.mmh"},
	{0x7C40D007D1786D52ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffb.mao"},
	{0x7C40DE07D178856FULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffb.msh"},
	{0x7C4134D54122F7D1ULL, "hlac_w_woundedcoast_0_12y.dds"},
	{0x7C414A0D3F08048CULL, "6094091_f.fxe"},
	{0x7C419FC0DDA0EFD2ULL, "lt_0_lowtown_docks_a3_21_0_lc.mmh"},
	{0x7C41C060AD6EF0A3ULL, "93688102.ogg"},
	{0x7C41C720C999592EULL, "86304054.ogg"},
	{0x7C422CF3EBC2B92EULL, "w_woundedcoast_113_0.mmh"},
	{0x7C4238F3EBC2CDC5ULL, "w_woundedcoast_113_0.mao"},
	{0x7C42416D1922469DULL, "vxw_bonepit_1qi_0_0.vlm"},
	{0x7C4246F3EBC2E51CULL, "w_woundedcoast_113_0.msh"},
	{0x7C424938EB215795ULL, "174850466.ogg"},
	{0x7C425C783E61E4CDULL, "vxlt_hangedman_n_26y_0_0.vlm"},
	{0x7C42F694BC55C79FULL, "vxlt_undercityserialkiller_18z_0_0.vlm"},
	{0x7C42FCC2992C1D2BULL, "gen00fl_isabela6101467.cl"},
	{0x7C430660FBAC2712ULL, "vxw_woundedcoast_fu_0_0.vlm"},
	{0x7C43E2AE0BBA04D7ULL, "992310793.ogg"},
	{0x7C44346D529A44CEULL, "w_sundermount_166_0.mmh"},
	{0x7C44406D529A5965ULL, "w_sundermount_166_0.mao"},
	{0x7C444E6D529A70BCULL, "w_sundermount_166_0.msh"},
	{0x7C4453A2C5D0DEE9ULL, "6167621_m.fxe"},
	{0x7C445CB1164CA76BULL, "vxlt_0_lowtown_n_e3a_0_0.vlm"},
	{0x7C4539F40172EB4DULL, "ico_2h_axe_1.dds"},
	{0x7C4541579E0F4C91ULL, "qun100cr_warrior1.utc"},
	{0x7C4548CF5AA57AACULL, "856427226.ogg"},
	{0x7C455787C4EA0937ULL, "llac_lt_foundry_n_320_19.dds"},
	{0x7C46A7FCBFD10156ULL, "vxw_bonepitdestroyed_a3_i5_0_0.vlm"},
	{0x7C46C9CD19DA64F8ULL, "hlac_lt_gamlenshouse_n_-1_5y.dds"},
	{0x7C473B5003315D6DULL, "vxht_mansion2b_2l6_0_0.vlm"},
	{0x7C47560C280F2C84ULL, "vxw_encounter1_kl_0_0.vlm"},
	{0x7C47986384FF1140ULL, "vxlt_foundry_n_107_0_0.vlm"},
	{0x7C47C3C0290186C2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medc_t.dds"},
	{0x7C47EBF2223734FBULL, "vxw_sundermount_2pp_0_0.vlm"},
	{0x7C48000150531D32ULL, "vxlt_warehouse_12j_0_0.vlm"},
	{0x7C482A76D0F143FFULL, "6175909_m.fxe"},
	{0x7C498456C13C57F2ULL, "art\\vfx\\environment\\vfx_blt_fire_burn_wood_p.mmh"},
	{0x7C4A09251815DBB7ULL, "llac_lt_0_lowtown_a2_196607_1.dds"},
	{0x7C4A98DC82FE463BULL, "147276437.ogg"},
	{0x7C4B0A12214A0FB3ULL, "6191288_m.fxe"},
	{0x7C4B62339BA59735ULL, "art\\toolset\\tools_bulb.msh"},
	{0x7C4B64339BA59AABULL, "art\\toolset\\tools_bulb.mmh"},
	{0x7C4B70339BA5AF48ULL, "art\\toolset\\tools_bulb.mao"},
	{0x7C4BED1FECF55FA1ULL, "dae000_sabina6161439.cl"},
	{0x7C4BF97B3E46BE38ULL, "hlac_lt_0_lowtown_a2_n_1_6y.dds"},
	{0x7C4C11B528754E57ULL, "llac_lt_undercityencounter_-1_25.dds"},
	{0x7C4C4A0D03BBA8DBULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.impa.evt"},
	{0x7C4C665B2A691B31ULL, "ser270_emeric6111395.cl"},
	{0x7C4CA13B5B5F9041ULL, "vxlt_0_lowtown_docks_a3_d3w_0_0.vlm"},
	{0x7C4CD9812385E9DDULL, "vxlt_gallowstemplar_n_3ni_0_0.vlm"},
	{0x7C4D2E893C20C797ULL, "mag330_walter6044682.cl"},
	{0x7C4D566BDDF41B00ULL, "follower_barks6153315.cl"},
	{0x7C4DCF2AB26F54BFULL, "vxw_sundermount_31f_0_0.vlm"},
	{0x7C4E16E85F6A1204ULL, "cd_seb221_brett_harimann.wbd"},
	{0x7C4E814E01A93ACFULL, "framebuffereffects_twopassblurshadows_p0.psh.fxo.dx11"},
	{0x7C4E8A42856529B1ULL, "vxw_encounter3_c3_0_0.vlm"},
	{0x7C4E8C5478FEEA6CULL, "514331380.ogg"},
	{0x7C4EA1172EA6FE1BULL, "hlac_lt_0_lowtown_a3_n_c_65538_5z.dds"},
	{0x7C4EB2A9C954D36EULL, "dae300_alistair_king6133510.cl"},
	{0x7C4F634A6C91C731ULL, "vxca_caverns1_ihu_0_0.vlm"},
	{0x7C4FAB8A43D5CAFDULL, "6050321_m.fxe"},
	{0x7C502A3A5BCA0AA2ULL, "llac_ht_mansion2_818_19.dds"},
	{0x7C509C693242A03AULL, "vxlt_gallowscourtyard_3x9_0_0.vlm"},
	{0x7C50B34A0B185F1DULL, "vxw_bonepit_1zh_0_0.vlm"},
	{0x7C50B8A81E6F2DB5ULL, "hlac_w_sundermount_-262141_0y.dds"},
	{0x7C5132A4DA380FC4ULL, "vxlt_0_lowtown_docks_a3_n_d1g_0_0.vlm"},
	{0x7C51D964A3BD2224ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medp_0n.dds"},
	{0x7C522883AAA349F8ULL, "zz_lgt_1_debug.cnv"},
	{0x7C5259CEBC29418AULL, "hlac_lt_0_lowtown_-65537_0x.dds"},
	{0x7C5270C5DAD18D8BULL, "133119720.ogg"},
	{0x7C53430D03C1BB1CULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.impa.gad"},
	{0x7C5349325B065E7DULL, "hlac_ca_caverns1_24423_8y.dds"},
	{0x7C536893B872C7D2ULL, "vxca_caverns2_8ek_0_0.vlm"},
	{0x7C541D6E271AB49AULL, "vxht_mansion2_3e7_0_0.vlm"},
	{0x7C541DECD49A775DULL, "6080524.ogg"},
	{0x7C54D77ADEE9FCEAULL, "613011815.ogg"},
	{0x7C554F812F5A5CB1ULL, "hlac_lt_0_lowtown_docks_65534_6x.dds"},
	{0x7C55572EFE310019ULL, "llac_lt_0_lowtown_docks_a3_n_65535_5.dds"},
	{0x7C55930975A14A4EULL, "6040932_m.fxe"},
	{0x7C55A20551922A0DULL, "hlac_w_bonepitdestroyed_a3_393221_0y.dds"},
	{0x7C55B2EC40D0AFEBULL, "gen00fl_aveline6139528.cl"},
	{0x7C55BF45872758E4ULL, "art\\vfx\\environment\\vfx_water_splash_s_p.mmh"},
	{0x7C56060603AFE2D3ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.anb"},
	{0x7C560D13D77CE78EULL, "185443850.ogg"},
	{0x7C5611FE0493E4DAULL, "917723386.ogg"},
	{0x7C5686F2837D6B70ULL, "6166606_f.fxe"},
	{0x7C573E26E5E8F94EULL, "gen00fl_varric6075334.cl"},
	{0x7C576F164BB8F7B6ULL, "demo_weapon_swap6207146.cl"},
	{0x7C5829D894D840CCULL, "llac_ca_cavern1_haunted_24423_12.dds"},
	{0x7C587EBC24D1A486ULL, "hlac_lt_0_lowtown_docks_n_-131075_3x.dds"},
	{0x7C59C26B87030D2CULL, "1044149557.ogg"},
	{0x7C59EC53107FDBA1ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\bsh_wideleaf_n.dds"},
	{0x7C59FB498BE2EC1EULL, "m_mag311ar_warehouse_2.wbd"},
	{0x7C5A07982A1CE45BULL, "hlac_ht_keep_3734_1z.dds"},
	{0x7C5A1B0C9BCCA115ULL, "gen_im_cft_reg_g.uti"},
	{0x7C5A420D03C7D81AULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.impa.ani"},
	{0x7C5A6BFA67ED3822ULL, "vxw_woundedcoast_124_0_0.vlm"},
	{0x7C5B107D3AFBB9D2ULL, "vxlt_0_lowtown_a2_n_eom_0_1.vlm"},
	{0x7C5B13184D55A8C1ULL, "vxlt_0_lowtown_docks_a3_d4v_0_0.vlm"},
	{0x7C5B4B5E157C025DULL, "vxlt_gallowstemplar_n_3eh_0_0.vlm"},
	{0x7C5B8518C21F14EBULL, "6060121_f.fxe"},
	{0x7C5BADEAA115BEBDULL, "1039263096.ogg"},
	{0x7C5CDD9F4D515AABULL, "vxw_blightlands_b6_0_0.vlm"},
	{0x7C5CF53D51FFCEEDULL, "vxht_bartrand_n_69m_0_0.vlm"},
	{0x7C5CFBCF0454C7E7ULL, "die.ncs"},
	{0x7C5D8E4814D02C4EULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_enter.evt"},
	{0x7C5D9BCEA6CC5D5CULL, "6022570_m.fxe"},
	{0x7C5DCE126B180496ULL, "gen00fl_aveline6116476.cl"},
	{0x7C5DDFD218FCD9FFULL, "6168525_m.fxe"},
	{0x7C5EF56A4A06B45FULL, "art\\vfx\\combat\\vfx_gib_hm_bld.mao"},
	{0x7C5F9C7BCA055BE1ULL, "hlac_dr_slavers_3216_7z.dds"},
	{0x7C5FDF86894A7ADFULL, "zz_approval_debug6003721.cl"},
	{0x7C5FEA1ADF985812ULL, "vxw_woundedcoast_xw_0_0.vlm"},
	{0x7C60D425EA9076F9ULL, "one100im_athenril_goods.uti"},
	{0x7C60ED946DAF4427ULL, "dae000_gen_player_turn_in6177861.cl"},
	{0x7C610D8C595C1E1CULL, "vxlt_0_lowtown_docks_a3_n_bf8_0_0.vlm"},
	{0x7C6130E4441B0724ULL, "6018526_m.fxe"},
	{0x7C6161AD4657E203ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_wkf.evt"},
	{0x7C619ABECBDA3E24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_db.ani"},
	{0x7C61F7163771640CULL, "hlac_w_sundermount_196613_0y.dds"},
	{0x7C632CBB95C1E16DULL, "6103390_f.fxe"},
	{0x7C637A34B4349693ULL, "303631030.ogg"},
	{0x7C63FB4D8C019A04ULL, "hlac_ht_0_hightown_65536_0x.dds"},
	{0x7C64230603BC513CULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.mmh"},
	{0x7C64854814D63BF9ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_enter.gad"},
	{0x7C65540E9E014AD1ULL, "tbe100ip_shop_codex_2.utp"},
	{0x7C655B15D461CBE3ULL, "zz_qcr_debug6074960.cl"},
	{0x7C6593FF189E1845ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\blendtree\\floortrap.anb"},
	{0x7C662548F121C74CULL, "llac_lt_0_lowtown_docks_-3_2.dds"},
	{0x7C666E0B7B604953ULL, "256430245.ogg"},
	{0x7C674E55F137D062ULL, "655902094.ogg"},
	{0x7C6797BECBDEA486ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_db.gad"},
	{0x7C67DF47110BF37CULL, "zz_fol_debug6096539.cl"},
	{0x7C683650575D3DD6ULL, "art\\vfx\\combat\\vfx_w_wns_front_clear_c.impa.ani"},
	{0x7C683AAD465DBE84ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_wkf.gad"},
	{0x7C685B487E04A03BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastwood_n.dds"},
	{0x7C68809893032CE8ULL, "gen00fl_isabela6099356.cl"},
	{0x7C688F7BF425F35CULL, "seb201_lady_harimann6177589.cl"},
	{0x7C68F4F252B6CE45ULL, "vxht_keep_67j_0_1.vlm"},
	{0x7C694E5537A54FF8ULL, "487678143.ogg"},
	{0x7C69E813D971C8ACULL, "follower_banter6143186.cl"},
	{0x7C6A2ADDCE88F837ULL, "zz_ave_debug6207678.cl"},
	{0x7C6A6ACFC3465ABDULL, "vxlt_warehouse_ai_0_0.vlm"},
	{0x7C6A743FD4BD802AULL, "vxca_cave_2bc_0_0.vlm"},
	{0x7C6A964814DAC447ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_enter.ani"},
	{0x7C6B0F336042084DULL, "6016064_m.fxe"},
	{0x7C6B78CF8BAF4EA6ULL, "vxlt_undercityserialkiller_158_0_0.vlm"},
	{0x7C6C00BEEA93DFC2ULL, "6015685_f.fxe"},
	{0x7C6D1DA5DCE8D726ULL, "222106041.ogg"},
	{0x7C6E0012F16AD434ULL, "hlac_lt_darktown_-1_75x.dds"},
	{0x7C6E4C2B9238111FULL, "vxlt_undercityserialkiller_17g_0_0.vlm"},
	{0x7C6E5BF7D18E7092ULL, "vxw_woundedcoast_up_0_0.vlm"},
	{0x7C6E92BECBE4BB65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_db.evt"},
	{0x7C6E980CDBA10281ULL, "235440652.ogg"},
	{0x7C6EEAA5EBF38B40ULL, "508856860.ogg"},
	{0x7C6F49AD4663F762ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_wkf.ani"},
	{0x7C6F8597033E6069ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_groundivy_e.dds"},
	{0x7C6FCC30414967C9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainspikexa.mao"},
	{0x7C6FD83041497C62ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainspikexa.mmh"},
	{0x7C6FDA3041497F18ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainspikexa.msh"},
	{0x7C702D7526E2368DULL, "vxca_caverns2prc_96y_0_0.vlm"},
	{0x7C70A9754D65549EULL, "hlac_lt_gamlenshouse_n_-1_2z.dds"},
	{0x7C70D4D6FBB19BADULL, "100879527.ogg"},
	{0x7C70E07811C783C0ULL, "6125745_m.fxe"},
	{0x7C71389DEE8DE357ULL, "vxlt_gallowstemplar_n_3yn_0_0.vlm"},
	{0x7C71814410012E34ULL, "mcr370_cut_player_rules6171817.cl"},
	{0x7C718356C15DA799ULL, "art\\vfx\\environment\\vfx_blt_fire_burn_wood_p.anb"},
	{0x7C7217E477084952ULL, "mrl150_merrill_pt5.stg"},
	{0x7C72374E794868CBULL, "vxlt_undercityencounter_111_0_0.vlm"},
	{0x7C72955BC361F75AULL, "6127594_m.fxe"},
	{0x7C72AB4E76F72177ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_alienagetree.phy"},
	{0x7C73014B9B639636ULL, "pro000_meeran6050117.cl"},
	{0x7C73043A943C7766ULL, "254961940.ogg"},
	{0x7C7358568F64C0CBULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.cess.ani"},
	{0x7C740382672FE8E2ULL, "146899353.ogg"},
	{0x7C7404697AF9ABE7ULL, "zz_ser_debug6086990.cl"},
	{0x7C74A531321D863BULL, "39704967.ogg"},
	{0x7C74BEDF41DFE814ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_aoe_p.impa.ani"},
	{0x7C74FFC4E5C5A0C7ULL, "zz_rockwraith3.cut"},
	{0x7C753E505767D5C7ULL, "art\\vfx\\combat\\vfx_w_wns_front_clear_c.impa.evt"},
	{0x7C7559F7A90FFDF7ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.dura.gad"},
	{0x7C7584BA22172F71ULL, "662022667.ogg"},
	{0x7C75CBDFA305B79EULL, "6155586_m.fxe"},
	{0x7C75E42197904DE0ULL, "6016898_m.fxe"},
	{0x7C768B57D35D8057ULL, "56659233.ogg"},
	{0x7C768DCC1CAF40A5ULL, "1051790631.ogg"},
	{0x7C76921BE592C7B1ULL, "vxlt_gallowscourtyard_5z6_0_0.vlm"},
	{0x7C76BB72FED8A62CULL, "cor_dark_clo.mao"},
	{0x7C7712D4CF226CFDULL, "vxlt_0_lowtown_a2_n_c_emo_0_0.vlm"},
	{0x7C7715A15106AFFEULL, "972356442.ogg"},
	{0x7C773D3CB2CE6BECULL, "vxca_caverns1_dp9_0_0.vlm"},
	{0x7C774FB44BD96922ULL, "vxw_woundedcoast_186_0_0.vlm"},
	{0x7C782AA2571BF74EULL, "65099953.ogg"},
	{0x7C7850134B9A98FEULL, "200072830.ogg"},
	{0x7C7917438A572CF8ULL, "gen00fl_aveline6139929.cl"},
	{0x7C793EE3F7FB5E09ULL, "6072601_f.fxe"},
	{0x7C794C7702BB2C9BULL, "mag291ip_wryme_blocker.ncs"},
	{0x7C799CCA36E5B45FULL, "910517734.ogg"},
	{0x7C7A2EE1DCA4EE7DULL, "hlac_lt_gallowstemplar_2384_5y.dds"},
	{0x7C7ABB98AC7A4920ULL, "264649085.ogg"},
	{0x7C7B0C15A486AD5AULL, "vxdr_primevalentrance_23e_0_0.vlm"},
	{0x7C7BBBDF41E60276ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_aoe_p.impa.gad"},
	{0x7C7BDC0FBC00B2D4ULL, "224092065.ogg"},
	{0x7C7C5F50576E2C10ULL, "art\\vfx\\combat\\vfx_w_wns_front_clear_c.impa.gad"},
	{0x7C7C6EF7A9163F84ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.dura.evt"},
	{0x7C7CB55C6EE81300ULL, "vxlt_0_lowtown_n_eux_0_0.vlm"},
	{0x7C7D1CF69AD9846FULL, "vxht_keep_n_5gu_0_0.vlm"},
	{0x7C7D1EE75D18F98DULL, "vxlt_0_lowtown_gc2_0_0.vlm"},
	{0x7C7D2B172D770C91ULL, "follower_banter6143307.cl"},
	{0x7C7D3190F851185AULL, "w_bonepitmining_a2_138_0.mmh"},
	{0x7C7D3590F8511E91ULL, "w_bonepitmining_a2_138_0.mao"},
	{0x7C7D4390F85136A0ULL, "w_bonepitmining_a2_138_0.msh"},
	{0x7C7E04B1019A2E74ULL, "26584324.ogg"},
	{0x7C7E5893DAD39F39ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_templar2.gad"},
	{0x7C7E74FACED25A82ULL, "820019868.ogg"},
	{0x7C7E8066400576E6ULL, "6189510_m.fxe"},
	{0x7C7E89B4BAFB3D0FULL, "gen_im_arm_cht_lgt_act3_01.uti"},
	{0x7C7E9E33865907AFULL, "vxlt_gallowstemplar_n_38u_0_0.vlm"},
	{0x7C7EBBE1F73AFDD1ULL, "6075481_m.fxe"},
	{0x7C7EC0FD26A92CD2ULL, "6095897_m.fxe"},
	{0x7C7EC245874A62DBULL, "art\\vfx\\environment\\vfx_water_splash_s_p.anb"},
	{0x7C7EF2BB7913A332ULL, "79065122.ogg"},
	{0x7C7F806CEA48A11BULL, "isa311_isabela6091964.cl"},
	{0x7C7F9CCAC674C553ULL, "llac_w_blightlands_262141_0.dds"},
	{0x7C80288B94B385A8ULL, "707054908.ogg"},
	{0x7C8067434875EE47ULL, "6134528_m.fxe"},
	{0x7C80871D7F0D6300ULL, "dae000_tomwise6167351.cl"},
	{0x7C809122F39350E6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkr.ani"},
	{0x7C8097C3693118CEULL, "593043671.ogg"},
	{0x7C81045F630C1B6EULL, "qun120_arishok6051683.cl"},
	{0x7C8140568F70D542ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.cess.evt"},
	{0x7C816D623733F7DCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chab_s.dds"},
	{0x7C81D638CBEA096CULL, "zz_vault_debug6177781.cl"},
	{0x7C821DF4CF7FFA1DULL, "205749769.ogg"},
	{0x7C82757FAF54520BULL, "6044656_m.fxe"},
	{0x7C82D6DF41EC4ED5ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_aoe_p.impa.evt"},
	{0x7C8309A02B1F656FULL, "vxdr_slavers_1au_0_0.vlm"},
	{0x7C83769476A21E14ULL, "vxht_0_hightown_a2_cbi_0_1.vlm"},
	{0x7C83C3AAA1BEAFDAULL, "6096908_m.fxe"},
	{0x7C842244CF67819BULL, "zz_one_debug6025818.cl"},
	{0x7C844C677A1224CEULL, "gen_im_cns_tom_mag_frcmage.uti"},
	{0x7C84A08C2E23CD3AULL, "6144777_m.fxe"},
	{0x7C856993DAD9DA87ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_templar2.ani"},
	{0x7C857FBF26FB3514ULL, "vxlt_0_lowtown_a3_n_c_hdm_0_0.vlm"},
	{0x7C85827716D4FBFFULL, "lt_hangedman_n_3567_ll.mmh"},
	{0x7C86C5AE271467BBULL, "clf260_harley6088756.cl"},
	{0x7C873F568F753F55ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.cess.gad"},
	{0x7C8783D990933CA1ULL, "hlac_lt_0_lowtown_a2_-65535_2y.dds"},
	{0x7C8816B2A6112700ULL, "6148160_m.fxe"},
	{0x7C889EF77F963B3DULL, "155222840.ogg"},
	{0x7C88EEE0941C4470ULL, "6085959_m.fxe"},
	{0x7C894C8C2B2FDC83ULL, "vxlt_0_lowtown_docks_n_bcc_0_0.vlm"},
	{0x7C897DF2967CC5DAULL, "vxdr_primevalentrance_26f_0_0.vlm"},
	{0x7C8986F7A920F37DULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.dura.ani"},
	{0x7C89FCB2116C46BAULL, "6169934_m.fxe"},
	{0x7C8A7ABDD2F0C11DULL, "vxw_bonepit_1nd_0_0.vlm"},
	{0x7C8A97E69B7BA612ULL, "gen00fl_varric6076050.cl"},
	{0x7C8A9B036DE173ADULL, "hlac_lt_0_lowtown_docks_-131073_1z.dds"},
	{0x7C8B273960DE2B80ULL, "vxlt_0_lowtown_n_eny_0_0.vlm"},
	{0x7C8B684B50416804ULL, "vxlt_hangedman_1vt_0_0.vlm"},
	{0x7C8BC3B6C02345D1ULL, "mer210cr_coterie_lieutenant.utc"},
	{0x7C8C5C8B46A0224AULL, "vxlt_0_lowtown_a2_n_ehm_0_0.vlm"},
	{0x7C8C90FE8A442F04ULL, "art\\vfx\\combat\\vfx_gui_target_ret_20_p.dura.ani"},
	{0x7C8CB3D74BB6A36EULL, "demoral_cl.mao"},
	{0x7C8CB50B2237C08BULL, "vxlt_0_lowtown_docks_a3_cvx_0_1.vlm"},
	{0x7C8D669DBA5864D0ULL, "vxca_caverns2_7m6_0_0.vlm"},
	{0x7C8D994DB708ADE1ULL, "6083839_m.fxe"},
	{0x7C8DC3EEBE74A954ULL, "736737216.ogg"},
	{0x7C8E0A95FD72FBBCULL, "plt_mrl200pt_n_nterrors.nss"},
	{0x7C8E2D22D6AA79CBULL, "llac_ht_0_hightown_a3_65533_3.dds"},
	{0x7C8E7922F39F6557ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkr.evt"},
	{0x7C8ED1E47D97790FULL, "cr_center_sp.mao"},
	{0x7C8EFCF0FB0C237DULL, "vxdr_primevaltomb_2u8_0_0.vlm"},
	{0x7C8F165B3AABF84CULL, "vxca_caverns2_8rp_0_0.vlm"},
	{0x7C8F326B6C264A2FULL, "vxlt_gallowstemplar_f_3sx_0_0.vlm"},
	{0x7C8F4C3CE27119EAULL, "mag110_vincento6171498.cl"},
	{0x7C8FD0FDB777F231ULL, "hlac_ht_mansion2_818_29z.dds"},
	{0x7C9006644A07EE0BULL, "mer320_hubert6079340.cl"},
	{0x7C9056028FDD633BULL, "vfx_distortionmask_p3.vsh.fxo"},
	{0x7C91445BF159791CULL, "vxht_bartrand_n_5w9_0_0.vlm"},
	{0x7C914CE070DC365BULL, "6174821_m.fxe"},
	{0x7C91AC9CFD9115C9ULL, "character_character_p44.psh.fxo.dx11"},
	{0x7C924D0B4A936E79ULL, "vxca_cave_19o_0_1.vlm"},
	{0x7C929AE56705A7E6ULL, "838707414.ogg"},
	{0x7C931684DA2D92A6ULL, "gen00fl_aveline6139682.cl"},
	{0x7C93505A72511917ULL, "vxca_caverns1_mines_hul_0_0.vlm"},
	{0x7C938DFE8A4A4966ULL, "art\\vfx\\combat\\vfx_gui_target_ret_20_p.dura.gad"},
	{0x7C93997E2E54422AULL, "129198006.ogg"},
	{0x7C93EEDB0E727D48ULL, "hair0_pass1_depth.psh.fxo"},
	{0x7C941367177AC9D6ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.cess.ani"},
	{0x7C9448D45AF1F088ULL, "888468712.ogg"},
	{0x7C94E624405C4D10ULL, "574469236.ogg"},
	{0x7C950C0AF7FD9B4FULL, "66856198.ogg"},
	{0x7C957A22F3A585E0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkr.gad"},
	{0x7C95C0669928A43DULL, "vxlt_warehouse_pf_0_0.vlm"},
	{0x7C964CCDE25E05F2ULL, "vxht_keep_n_69y_0_0.vlm"},
	{0x7C968F26067D6E81ULL, "6129317_m.fxe"},
	{0x7C9722A6D3BAA1F3ULL, "62410714.ogg"},
	{0x7C975BDDC93070BAULL, "hlac_w_encounter2_camp_-65538_0x.dds"},
	{0x7C98E9FF3974BEDEULL, "6079346_f.fxe"},
	{0x7C995A2B81D2F434ULL, "dae000_bartrand6107542.cl"},
	{0x7C9988FE8A4EAC45ULL, "art\\vfx\\combat\\vfx_gui_target_ret_20_p.dura.evt"},
	{0x7C99991652D44400ULL, "vxlt_0_lowtown_n_ekf_0_0.vlm"},
	{0x7C99FF75B78C4E83ULL, "657979279.ogg"},
	{0x7C9A00B07EC5B56FULL, "vxht_keep_n_5mw_0_0.vlm"},
	{0x7C9A111CF2831430ULL, "715050480.ogg"},
	{0x7C9A1A6623568A87ULL, "zz_approval_debug6003927.cl"},
	{0x7C9A2AAA9448B9DBULL, "hlac_dr_primevaltomb_9621_1x.dds"},
	{0x7C9A31CDC982A240ULL, "gen00fl_anders6125527.cl"},
	{0x7C9AC3BC096708E0ULL, "6116321_m.fxe"},
	{0x7C9ACE6838963ACAULL, "vxlt_0_lowtown_a2_n_emn_0_0.vlm"},
	{0x7C9AD6F2F680D5FDULL, "282650802.ogg"},
	{0x7C9B8684D3CB0943ULL, "6069451_m.fxe"},
	{0x7C9C96684C04FC3CULL, "pro000_meeran6048802.cl"},
	{0x7C9CC3794CDAEB3BULL, "403805797.ogg"},
	{0x7C9D445619B202A4ULL, "art\\characters\\playercharacter\\humanfemale\\arm_urva.msh"},
	{0x7C9D565619B2210DULL, "art\\characters\\playercharacter\\humanfemale\\arm_urva.mao"},
	{0x7C9D5A5619B22846ULL, "art\\characters\\playercharacter\\humanfemale\\arm_urva.mmh"},
	{0x7C9DA9A051193E94ULL, "vxlt_undercitydungeon_1oy_0_0.vlm"},
	{0x7C9DC34CA83936CFULL, "gen_im_npc_dwf_arm_cht_cxt4.uti"},
	{0x7C9E6730C4255959ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish_additive.ani"},
	{0x7C9EC2E90E7E08DAULL, "vxht_0_hightown_a2_by2_0_0.vlm"},
	{0x7C9F1238E72FDAFEULL, "vxw_sundermount_2eb_0_0.vlm"},
	{0x7C9F15CD46141F2DULL, "779362421.ogg"},
	{0x7C9F93853308F9A8ULL, "778277594.ogg"},
	{0x7CA043B4883F0F7FULL, "vxw_sundermount_33i_0_1.vlm"},
	{0x7CA05A6CF7B6D678ULL, "6117281_m.fxe"},
	{0x7CA0BA362BC03C0AULL, "mrl000_marethari6129263.cl"},
	{0x7CA11B67178561C7ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.cess.evt"},
	{0x7CA1206ED95D9386ULL, "hlac_ht_keep_3735_1x.dds"},
	{0x7CA1ED6FF3667C8DULL, "6104020_m.fxe"},
	{0x7CA263790AE76614ULL, "vxlt_0_lowtown_a3_n_c_hns_0_0.vlm"},
	{0x7CA289B0575552FBULL, "qcr270_fake_meredith6066331.cl"},
	{0x7CA2C4508EF8FCB1ULL, "zz_lgt_1_debug6114360.cl"},
	{0x7CA35090F8721C65ULL, "w_bonepitmining_a2_138_0.phy"},
	{0x7CA3606AEB8DD195ULL, "lgt220_mag_apost_apost2.bnk"},
	{0x7CA3E8537F795A06ULL, "gen_im_cns_cft_bmb_02.uti"},
	{0x7CA42ABEA59D6874ULL, "6179797_m.fxe"},
	{0x7CA432438B1EBCBDULL, "vxlt_warehouse_ue_0_0.vlm"},
	{0x7CA4E389EB86C63AULL, "238300501.ogg"},
	{0x7CA4FCE7817173B6ULL, "6036699_f.fxe"},
	{0x7CA54230C42B389BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish_additive.gad"},
	{0x7CA5636DEFF4FEF7ULL, "hlac_ht_bartrand_n_3382_0x.dds"},
	{0x7CA5A6D4F514A98CULL, "vis110_commoner16069054.cl"},
	{0x7CA5BAEE7577522FULL, "one120_athenril6088918.cl"},
	{0x7CA5D7172CA1E772ULL, "hlac_lt_0_lowtown_a2_n_-131072_7x.dds"},
	{0x7CA5E9BD7E7E3231ULL, "w_blightlands_222_0.msh"},
	{0x7CA5F7BD7E7E4A4CULL, "w_blightlands_222_0.mao"},
	{0x7CA603BD7E7E5EEFULL, "w_blightlands_222_0.mmh"},
	{0x7CA698C0D4C59AF0ULL, "dre000_varric6033315.cl"},
	{0x7CA73EBFB7F455AFULL, "gen_im_wep_war_one_act0_02.uti"},
	{0x7CA77A2DE439BB94ULL, "mag311_nyssa6171516.cl"},
	{0x7CA7826AEB91F4EDULL, "lgt220_mag_apost_apost2.cnv"},
	{0x7CA7D87F574DD6E8ULL, "6096483_m.fxe"},
	{0x7CA7FC741733EAEBULL, "6095023_m.fxe"},
	{0x7CA81A5B639FDDA9ULL, "hlac_lt_undercityserialkiller_-1_26y.dds"},
	{0x7CA826F9331F078AULL, "97293459.ogg"},
	{0x7CA83C67178BB810ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.cess.gad"},
	{0x7CA8515619BC2011ULL, "art\\characters\\playercharacter\\humanfemale\\arm_urva.phy"},
	{0x7CA8CE13D0E8C97BULL, "w_encounter1_4_0.mmh"},
	{0x7CA8DA13D0E8DD18ULL, "w_encounter1_4_0.mao"},
	{0x7CA8EC13D0E8FC65ULL, "w_encounter1_4_0.msh"},
	{0x7CA9756D3221B8CCULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tool.msh"},
	{0x7CA97B6D3221C2BEULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tool.mmh"},
	{0x7CA984607A996E5DULL, "ht_0_hightown_35_0_lc.mmh"},
	{0x7CA9876D3221D755ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tool.mao"},
	{0x7CAA0A65D7ADFFF8ULL, "6135667_m.fxe"},
	{0x7CAA0EC6CE880179ULL, "gen00fl_fenris6127021.cl"},
	{0x7CAA7C8FC43FF986ULL, "vxw_sundermount_2fs_0_0.vlm"},
	{0x7CAA8FC6809641CAULL, "6156789_m.fxe"},
	{0x7CAB2E7F1AAAE341ULL, "dae000_arishok6078446.cl"},
	{0x7CAC11739EA4EEACULL, "62553760.ogg"},
	{0x7CAC165A6FF16D2AULL, "hlac_lt_0_lowtown_a2_n_c_-65537_3z.dds"},
	{0x7CAC22AA081E8751ULL, "dae000cr_merchant_weapons.utc"},
	{0x7CAC357941D33A33ULL, "6133752_m.fxe"},
	{0x7CAC5F30C4318880ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\flourish_additive.evt"},
	{0x7CAC9E9DB3F65558ULL, "67469679.ogg"},
	{0x7CACE17E3AABB011ULL, "6117028_m.fxe"},
	{0x7CAD8415D925F37EULL, "vxw_sundermount_2nc_0_0.vlm"},
	{0x7CAD94A38C296D49ULL, "hlac_lt_0_lowtown_n_-2_4x.dds"},
	{0x7CADA47201726456ULL, "6163226_m.fxe"},
	{0x7CAE2961A5853EF0ULL, "vxlt_0_lowtown_docks_n_b5s_0_1.vlm"},
	{0x7CAE7CB78BB57CE8ULL, "415806493.ogg"},
	{0x7CAE8B54F38108CDULL, "345811618.ogg"},
	{0x7CAF70F850FC343FULL, "hlac_w_woundedcoast_-65532_0z.dds"},
	{0x7CAF9B9B7459787DULL, "vxht_playershouse_n_1q8_0_0.vlm"},
	{0x7CAFF386F727FC00ULL, "primitives_primitive_p4.vsh.fxo.dx11.assembly"},
	{0x7CB072A482CBED35ULL, "ht_0_hightownexplosion_n_30_0_ll.mmh"},
	{0x7CB0D555FCDD7E94ULL, "vxlt_0_lowtown_a3_n_c_hkp_0_0.vlm"},
	{0x7CB0DA4896FACEFDULL, "vxlt_0_lowtown_a2_n_c_eqk_0_0.vlm"},
	{0x7CB19D2DBBAFCA1CULL, "hlac_ca_caverns1_24421_1y.dds"},
	{0x7CB27D55D8C44494ULL, "390877781.ogg"},
	{0x7CB2C55CFC07FAE6ULL, "primitives_primitive_p0.vsh.fxo"},
	{0x7CB2D6F13410ED90ULL, "6035278_f.fxe"},
	{0x7CB302B694C8F77EULL, "6129844_m.fxe"},
	{0x7CB34E32450C52B7ULL, "vxlt_gallowstemplar_3k7_0_0.vlm"},
	{0x7CB3653D43D48CE7ULL, "6074376_m.fxe"},
	{0x7CB37896C94AA7FDULL, "201219110.ogg"},
	{0x7CB47AC74A9BB750ULL, "vxlt_0_lowtown_h2g_0_0.vlm"},
	{0x7CB4B9C12406757FULL, "follower_banter6156048.cl"},
	{0x7CB4CEA87B3B1644ULL, "vxht_keep_632_0_0.vlm"},
	{0x7CB4F5FAF1687346ULL, "6056811_f.fxe"},
	{0x7CB5657E30179893ULL, "hlac_lt_0_lowtown_docks_196609_0z.dds"},
	{0x7CB56B7EDF98B763ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_appr.ani"},
	{0x7CB5B1E199FC504BULL, "6104937_f.fxe"},
	{0x7CB5CA5DEC8DFB7CULL, "mag360cr_hostage_bethany.utc"},
	{0x7CB5CEFAE2C87801ULL, "gen00fl_fenris_base_act1_1.stg"},
	{0x7CB6DCD5529BBC38ULL, "680692771.ogg"},
	{0x7CB78DC0D848F0E2ULL, "w_blightlands_100_0.msh"},
	{0x7CB78FC0D848F398ULL, "w_blightlands_100_0.mmh"},
	{0x7CB79BC0D849083BULL, "w_blightlands_100_0.mao"},
	{0x7CB8265A485D2313ULL, "llac_w_sundermount_-196605_0.dds"},
	{0x7CB8EE6CB6361206ULL, "vxw_sundermount_2kt_0_0.vlm"},
	{0x7CB9145F2BB2683AULL, "dae011_corff_sta6164571.cl"},
	{0x7CBA0AC7833CA4CFULL, "vxht_brothel_n_5b9_0_0.vlm"},
	{0x7CBA66FE7434F331ULL, "123097944.ogg"},
	{0x7CBA804E77348DCFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_alienagetree.mao"},
	{0x7CBA8C4E7734A26CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_alienagetree.mmh"},
	{0x7CBA90F1CBBA5ACAULL, "663041114.ogg"},
	{0x7CBA924E7734AC5EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_alienagetree.msh"},
	{0x7CBAB6E51621596BULL, "hlac_w_encounter2_327679_0y.dds"},
	{0x7CBAF0A5436427E7ULL, "6131170_m.fxe"},
	{0x7CBB479D5609A58AULL, "dae000_arishok6075196.cl"},
	{0x7CBB4D44868D76B3ULL, "93658402.ogg"},
	{0x7CBB7069FD98D4B5ULL, "248272646.ogg"},
	{0x7CBBF5F2CB1C0BFEULL, "vxw_sundermount_2sd_0_0.vlm"},
	{0x7CBC62F391B36E52ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_aurora.dds"},
	{0x7CBC738E7E319789ULL, "6163449_m.fxe"},
	{0x7CBC8FEF786E7BE7ULL, "240278112.ogg"},
	{0x7CBCDE9506DBD048ULL, "466683270.ogg"},
	{0x7CBD4C7CC22AE9F6ULL, "lt_0_lowtown_docks_235_0.msh"},
	{0x7CBD567CC22AFAF4ULL, "lt_0_lowtown_docks_235_0.mmh"},
	{0x7CBD5A7CC22B0107ULL, "lt_0_lowtown_docks_235_0.mao"},
	{0x7CBD74D2880EA078ULL, "484292434.ogg"},
	{0x7CBD779B808B4FF4ULL, "vxca_cavern1_haunted_fm9_0_0.vlm"},
	{0x7CBDB378D951D906ULL, "ser121_jethann6104576.cl"},
	{0x7CBE43DB8EF9570BULL, "vxht_mansion2_2as_0_0.vlm"},
	{0x7CBE781E5B9BF9ADULL, "art\\levels\\races\\proxy\\props\\_textures\\foodstorage_d.dds"},
	{0x7CBEB57144D49BB5ULL, "859404591.ogg"},
	{0x7CBF35E29F5932BAULL, "6111237_f.fxe"},
	{0x7CBF72838576B2BDULL, "vxlt_0_lowtown_a2_n_ey1_0_0.vlm"},
	{0x7CBFC7E73D060D01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_stand_loop.ani"},
	{0x7CC00C7F234B1C63ULL, "dae000_friend_turn_in6179069.cl"},
	{0x7CC03F1781DF28B3ULL, "startmenu2.gfx"},
	{0x7CC07F4704FED7F9ULL, "vrc221_varric6040841.cl"},
	{0x7CC0B7E00FF7B1F9ULL, "357357252.ogg"},
	{0x7CC154624B6B07AEULL, "mrl000_marethari6129339.cl"},
	{0x7CC15E184DDC90CCULL, "pou_sh.mao"},
	{0x7CC1D71A35CAA567ULL, "6079076_m.fxe"},
	{0x7CC207F3B175D286ULL, "596418050.ogg"},
	{0x7CC2633AE362E759ULL, "6170786_m.fxe"},
	{0x7CC298ABAAEF5602ULL, "vxht_0_hightown_a2_a6p_0_0.vlm"},
	{0x7CC2ECA43C91CFD0ULL, "vxlt_0_lowtown_h7d_0_0.vlm"},
	{0x7CC3111E43AFBCCCULL, "vxlt_gallowstemplar_5ul_0_0.vlm"},
	{0x7CC3278D91751CC5ULL, "6128282_m.fxe"},
	{0x7CC340B08E4643CAULL, "6087779_m.fxe"},
	{0x7CC3C4B2359D7805ULL, "gen00fl_sebastian6195870.cl"},
	{0x7CC52FBF1819CA04ULL, "vxlt_hangedman_1bx_0_0.vlm"},
	{0x7CC59971252F144AULL, "llac_lt_0_lowtown_a2_n_-131072_9.dds"},
	{0x7CC5C84E19144CF9ULL, "6058866_m.fxe"},
	{0x7CC5D2E73D0A8BA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_stand_loop.gad"},
	{0x7CC5E49D5E44F887ULL, "vxht_mansion_n_61d_0_0.vlm"},
	{0x7CC64A7893711079ULL, "mag160_feynriel6107439.cl"},
	{0x7CC69EB8582C6C96ULL, "vxlt_blackemporium_s4_0_0.vlm"},
	{0x7CC6E3A761563C36ULL, "gen_im_trs_dem_lgt_hlm_r2.uti"},
	{0x7CC76049A82C2A86ULL, "vxw_sundermount_2lu_0_0.vlm"},
	{0x7CC82FF1B43A7F65ULL, "6179301_m.fxe"},
	{0x7CC8524C191726CDULL, "gen00fl_anders6096644.cl"},
	{0x7CC858656BE12310ULL, "mag320_orsino6071054.cl"},
	{0x7CC87CA47532BD4FULL, "vxht_brothel_n_5e8_0_0.vlm"},
	{0x7CC8E6B6B9E6CB5DULL, "dae000_arishok6096338.cl"},
	{0x7CC9827EDFA9874DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_appr.gad"},
	{0x7CC99394BF79721AULL, "vxw_encounter3_b7_0_0.vlm"},
	{0x7CC9E2F6C9014170ULL, "6037584_f.fxe"},
	{0x7CCA526B3F0A95C3ULL, "1022305578.ogg"},
	{0x7CCA67CFBD12247EULL, "vxw_sundermount_2te_0_0.vlm"},
	{0x7CCB22B4B9C1B740ULL, "899725532.ogg"},
	{0x7CCC399F1186B9C2ULL, "6137032_m.fxe"},
	{0x7CCC76EF0419F074ULL, "darktown_male_fer_urch_sta6174958.cl"},
	{0x7CCCDFBC9E46D92BULL, "558528516.ogg"},
	{0x7CCD0AE0F6FD9BCCULL, "567135883.ogg"},
	{0x7CCD3A5D094C8076ULL, "follower_barks6153375.cl"},
	{0x7CCDBC72DEE1017CULL, "vxht_bartrand_n_5q3_0_0.vlm"},
	{0x7CCE72604CF38996ULL, "gen00fl_sebastian6180451.cl"},
	{0x7CCEED1933B567C1ULL, "26780723.ogg"},
	{0x7CCF0BBCC13217B0ULL, "6113223_m.fxe"},
	{0x7CCF57123D8CA36CULL, "513190612.ogg"},
	{0x7CCFCB1DF766476BULL, "hlac_lt_hangedman_n_-1_15y.dds"},
	{0x7CCFFE6943590952ULL, "344947678.ogg"},
	{0x7CD031B8727EB1D0ULL, "408309734.ogg"},
	{0x7CD0748008E32681ULL, "6118772_f.fxe"},
	{0x7CD12444E73D4CCFULL, "ht_0_hightown_a2_10_0.rml"},
	{0x7CD1609312BFAD0AULL, "vxca_cave_1cv_0_1.vlm"},
	{0x7CD19B98F738C33CULL, "vxca_caverns1_mines_imf_0_0.vlm"},
	{0x7CD2C6DB815D723AULL, "gen00fl_merrill6119638.cl"},
	{0x7CD2C8A5289FBBA9ULL, "vxht_keep_732_0_0.vlm"},
	{0x7CD2FC743B467588ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_paintingsmle.mmh"},
	{0x7CD308743B468A2BULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_paintingsmle.mao"},
	{0x7CD31A743B46A8B2ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_paintingsmle.msh"},
	{0x7CD3A71A190D8DE8ULL, "lt_gallowscourtyard_n_35_0_ll.mmh"},
	{0x7CD416BC7762E29DULL, "286398294.ogg"},
	{0x7CD44528BF69BC8EULL, "art\\characters\\weapons\\mauls\\w_mal_ml106a.phy"},
	{0x7CD45B374A104B77ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_aoe_imp_p.impa.gad"},
	{0x7CD50D787E28136EULL, "qun221_petrice6076656.cl"},
	{0x7CD510954A228516ULL, "vxlt_blackemporium_t5_0_0.vlm"},
	{0x7CD5857770482B7AULL, "mcr370_guardsman6086988.cl"},
	{0x7CD5A1974BA6EF90ULL, "lgt260_eau_tmp_angelique6143966.cl"},
	{0x7CD5A5254D6870B8ULL, "vxw_sundermount_2q8_0_0.vlm"},
	{0x7CD5C146A3B0E93BULL, "517339281.ogg"},
	{0x7CD5D4B8DCBD8A73ULL, "vxlt_0_lowtown_a3_n_c_ei0_0_0.vlm"},
	{0x7CD5D844354220D7ULL, "art\\vfx\\materials\\chunky\\vfx_c_rw_lightplane_vt.mao"},
	{0x7CD6D43A24B2B316ULL, "6092261_m.fxe"},
	{0x7CD74D7F20EEA7BFULL, "vxw_bonepit_2a0_0_0.vlm"},
	{0x7CD7F99B998CDDD7ULL, "vxht_0_hightown_asr_0_0.vlm"},
	{0x7CD82AC418A6FD34ULL, "vxht_0_hightown_n_alo_0_0.vlm"},
	{0x7CD887FC3B094ABAULL, "444396440.ogg"},
	{0x7CD88B9017F605B0ULL, "vxlt_0_lowtown_docks_api_0_0.vlm"},
	{0x7CD8D9ACAF083CFEULL, "vxw_sundermount_2yf_0_0.vlm"},
	{0x7CD97857DE693308ULL, "ht_keep_n_4519_lc.mmh"},
	{0x7CD9FF8412C531B8ULL, "hlac_lt_0_lowtown_a2_65538_2x.dds"},
	{0x7CDA1A7F7852C174ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_act_c.mmh"},
	{0x7CDA366B0183C6C9ULL, "vxht_0_hightown_b8f_0_0.vlm"},
	{0x7CDA9D2ADB630EFBULL, "mag110_vincento6032994.cl"},
	{0x7CDAAF5D6E82CFA3ULL, "gen_im_trs_gen_bracelet_02.uti"},
	{0x7CDB29B52777B0ACULL, "6182470_m.fxe"},
	{0x7CDB70374A168D04ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_aoe_imp_p.impa.evt"},
	{0x7CDB834374CF2AE3ULL, "w_bonepitdestroyed_a3_210_0.msh"},
	{0x7CDB914374CF423DULL, "w_bonepitdestroyed_a3_210_0.mmh"},
	{0x7CDB954374CF494EULL, "w_bonepitdestroyed_a3_210_0.mao"},
	{0x7CDBB44166A50C54ULL, "gen00fl_merrill6129613.cl"},
	{0x7CDBDED93B3D12B3ULL, "zz_followers_banter6159681.cl"},
	{0x7CDBE0F20EF8777FULL, "6111225_f.fxe"},
	{0x7CDC2E4FD0D719FCULL, "vxht_bartrand_n_5t0_0_0.vlm"},
	{0x7CDC9CFE801F781EULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.cess.evt"},
	{0x7CDCCC91420099CAULL, "479756963.ogg"},
	{0x7CDCE98CD06E5E23ULL, "6101479_f.fxe"},
	{0x7CDD722CAB2CA078ULL, "6196699_m.fxe"},
	{0x7CDD863B60B18CFCULL, "qun110_petrice6018677.cl"},
	{0x7CDDFFEB3616A966ULL, "hlac_ht_0_hightown_a3_65534_0y.dds"},
	{0x7CDE615CD9BB6745ULL, "6108074_f.fxe"},
	{0x7CDEA0D415CD1133ULL, "105521364.ogg"},
	{0x7CDEA59C910A43CDULL, "dae130ip_at_undercity.utp"},
	{0x7CDEB3218F764127ULL, "hf_isa_tevint_highmage.mor"},
	{0x7CDF195F2601CEDAULL, "485631041.ogg"},
	{0x7CDF1DCD6F935E92ULL, "vxlt_0_lowtown_n_ep0_0_1.vlm"},
	{0x7CDF36F7C8CA395EULL, "art\\cinematics\\deathblows\\highdragon\\cs_db_pc_dragwarr.ani"},
	{0x7CDF6001EFEA1F6DULL, "sys_chargen.ncs"},
	{0x7CDFABF7C5738571ULL, "405288323.ogg"},
	{0x7CE0097F54B18503ULL, "6171044_m.fxe"},
	{0x7CE00D75E92EDBBCULL, "vxca_caverns1_mines_ijg_0_0.vlm"},
	{0x7CE2578024E8142EULL, "232975145.ogg"},
	{0x7CE2B3FE80240B49ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.cess.gad"},
	{0x7CE372D0F1CD9E66ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\varric_combat_face_additive.anb"},
	{0x7CE417023F5E8938ULL, "vxw_sundermount_2z9_0_0.vlm"},
	{0x7CE444038C185B86ULL, "vxw_sundermount_2zw_0_0.vlm"},
	{0x7CE4BFBC907AA5FCULL, "113605527.ogg"},
	{0x7CE4D429E24BC4F3ULL, "vxht_keep_n_3ie_0_0.vlm"},
	{0x7CE61B95FDBD8EE7ULL, "plt_mrl200pt_n_nterrors.plo"},
	{0x7CE620B03F597DA1ULL, "character_rendertodepth_p3.psh.fxo.dx11.assembly"},
	{0x7CE632FAE84231E7ULL, "hlac_lt_0_lowtown_docks_n_-196610_4z.dds"},
	{0x7CE66DCF0D144F68ULL, "vxdr_primevaltomb_6s7_0_0.vlm"},
	{0x7CE6792412E51A46ULL, "gen00fl_anders6103679.cl"},
	{0x7CE75ED3B01907AAULL, "dae000_bartrand6206973.cl"},
	{0x7CE7955CE786FFEFULL, "vxw_bonepit_25d_0_0.vlm"},
	{0x7CE7A4073C422951ULL, "285301813.ogg"},
	{0x7CE7B3213C8BC932ULL, "hlac_lt_0_lowtown_a2_-262144_3x.dds"},
	{0x7CE7B465660D8D6DULL, "follower_barks6153697.cl"},
	{0x7CE7C01A8983E3C3ULL, "6083828_m.fxe"},
	{0x7CE84A7C42E32BFBULL, "ser121_jethann6175441.cl"},
	{0x7CE875929E3FB69CULL, "6042068_m.fxe"},
	{0x7CE888374A2140FDULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_aoe_imp_p.impa.ani"},
	{0x7CE895F6BB8F8DA3ULL, "mag121_idunna6084862.cl"},
	{0x7CE8AA071EE8144FULL, "dae000_bodhan6129091.cl"},
	{0x7CE8C28C4621641DULL, "w_sundermount_144_0.mao"},
	{0x7CE8C68C46216B56ULL, "w_sundermount_144_0.mmh"},
	{0x7CE8C793356AB395ULL, "zz_prmpt_debug6203644.cl"},
	{0x7CE8D08C46217C54ULL, "w_sundermount_144_0.msh"},
	{0x7CE984FE8029D997ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.cess.ani"},
	{0x7CE9997264DBA08BULL, "vxht_mansion2_2pt_0_0.vlm"},
	{0x7CE9E48435E3D74AULL, "vxht_mansion2b_3gq_0_0.vlm"},
	{0x7CE9F0E432D2B231ULL, "6117969_m.fxe"},
	{0x7CEA526D32591459ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tool.phy"},
	{0x7CEB0553A3359C1EULL, "vxlt_gallowstemplar_2go_0_0.vlm"},
	{0x7CEB5F6C691504F9ULL, "374287497.ogg"},
	{0x7CEBCFC85F31D3B1ULL, "vxlt_0_lowtown_a2_es4_0_0.vlm"},
	{0x7CEC6F9DFE4CA267ULL, "hlac_w_bonepitdestroyed_a3_0_12z.dds"},
	{0x7CECEC08DBAFD2BAULL, "6118185_m.fxe"},
	{0x7CED953BD8743A72ULL, "840771111.ogg"},
	{0x7CEDBA58E5EB11DAULL, "6080110_m.fxe"},
	{0x7CEE2CD06A8E6FD6ULL, "w_woundedcoast_18_0.msh"},
	{0x7CEE36D06A8E80D4ULL, "w_woundedcoast_18_0.mmh"},
	{0x7CEE380959818A8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small.gad"},
	{0x7CEE3AD06A8E87E7ULL, "w_woundedcoast_18_0.mao"},
	{0x7CEEDE84F4AB34DEULL, "6156756_m.fxe"},
	{0x7CEEF880ECBEEE53ULL, "846331732.ogg"},
	{0x7CEF58234518AC87ULL, "vxht_keep_n_6m4_0_0.vlm"},
	{0x7CF060F077E5F1A3ULL, "follower_banter6171898.cl"},
	{0x7CF074FE3F7FFE13ULL, "vxht_brothel_n_5gr_0_0.vlm"},
	{0x7CF090668A1CD1A6ULL, "qun120cr_javaris6043226.cl"},
	{0x7CF0B63F2B5DDDB4ULL, "hlac_lt_0_lowtown_-2_1y.dds"},
	{0x7CF0F04A199FDCB8ULL, "309227850.ogg"},
	{0x7CF100CD1DA6CA5CULL, "art\\characters\\creatures\\ragedemon\\animation\\dblw_lp_v1.ani"},
	{0x7CF11803AD69988EULL, "llac_w_encounter2_196613_0.dds"},
	{0x7CF1295519DE737DULL, "305010118.ogg"},
	{0x7CF17EE7424AC372ULL, "face0_s1y.psh.fxo"},
	{0x7CF1A7E58EC50E3CULL, "hlac_lt_0_lowtown_docks_n_-131074_8x.dds"},
	{0x7CF207106FC58379ULL, "vxlt_0_lowtown_ejc_0_0.vlm"},
	{0x7CF2CD9A4AB843B3ULL, "vxlt_0_lowtown_n_goj_0_1.vlm"},
	{0x7CF2E534C4D74D83ULL, "hlac_ht_chantry_2534_40x.dds"},
	{0x7CF336E994FE5D52ULL, "vrc221_bartrand6065307.cl"},
	{0x7CF34FF7C8DB0CD8ULL, "art\\cinematics\\deathblows\\highdragon\\cs_db_pc_dragwarr.gad"},
	{0x7CF395DAF4E87794ULL, "gen00fl_anders6119511.cl"},
	{0x7CF44E0B1246D38FULL, "6059665_m.fxe"},
	{0x7CF50CB3E5A3C9B6ULL, "vxdr_primevalentrance_24p_0_0.vlm"},
	{0x7CF51B0959877821ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small.evt"},
	{0x7CF51DB1B76722B4ULL, "vxdr_primevalentrance_167_0_0.vlm"},
	{0x7CF5238E81081399ULL, "gen00fl_sebastian6180832.cl"},
	{0x7CF555E37C80A832ULL, "hlac_lt_gallowsprison_n_-1_13y.dds"},
	{0x7CF56F49FBE236B0ULL, "vxlt_0_lowtown_docks_azw_0_0.vlm"},
	{0x7CF589D8727FDA2DULL, "960456658.ogg"},
	{0x7CF60739D97D186FULL, "vxw_bonepit_20g_0_0.vlm"},
	{0x7CF643396F0F2B75ULL, "zz_approval_debug6003514.cl"},
	{0x7CF6CE13D12AEA08ULL, "w_encounter1_4_0.rml"},
	{0x7CF70DCD1DAB4B9EULL, "art\\characters\\creatures\\ragedemon\\animation\\dblw_lp_v1.gad"},
	{0x7CF720FCF8DED584ULL, "vxw_encounter2_camp_5s_0_0.vlm"},
	{0x7CF76857AED4743EULL, "6169651_m.fxe"},
	{0x7CF76EB30D31BF81ULL, "vxlt_0_lowtown_n_exv_0_0.vlm"},
	{0x7CF7CC8A1D01D068ULL, "vxdr_primevaltomb_57l_0_0.vlm"},
	{0x7CF824F179560F46ULL, "36182374.ogg"},
	{0x7CF880DAE883028FULL, "6188655_m.fxe"},
	{0x7CF89F1CE7FCF1ACULL, "123301132.ogg"},
	{0x7CF994B0745BEEBCULL, "vxht_mansion2b_31u_0_0.vlm"},
	{0x7CF9B0C84DB850D5ULL, "vxht_keep_n_70l_0_0.vlm"},
	{0x7CF9FAA6303D3327ULL, "vxht_mansion2_1rq_0_0.vlm"},
	{0x7CFA42EA1BC5E3B8ULL, "355051627.ogg"},
	{0x7CFA68CAEDC6FD9BULL, "6027586_f.fxe"},
	{0x7CFAF94377D069B5ULL, "hlac_lt_0_lowtown_a2_n_c_65534_3y.dds"},
	{0x7CFB0269A5F2E721ULL, "lgt251cr_bancroft.utc"},
	{0x7CFB4775DD51EBE0ULL, "gen_im_trs_bst_spd_fang_04.uti"},
	{0x7CFBA5F946122D79ULL, "art\\vfx\\materials\\vfx_web_head_blend.mao"},
	{0x7CFC003635DF0D6BULL, "6109393_m.fxe"},
	{0x7CFCA7BCB078EFC9ULL, "hf_guard.mor"},
	{0x7CFD92C3E53732A8ULL, "vxca_caverns2_70y_0_0.vlm"},
	{0x7CFDAEE7A545080CULL, "6115562_m.fxe"},
	{0x7CFDC1F0A4870788ULL, "954202330.ogg"},
	{0x7CFDCA00370EC507ULL, "vxht_keep_n_6h7_0_0.vlm"},
	{0x7CFE3C24A1E3AB67ULL, "character_character_p8.psh.fxo.dx11.assembly"},
	{0x7CFEBECD05D5061FULL, "vxlt_undercityserialkiller_1yi_0_0.vlm"},
	{0x7CFEE6DB31761693ULL, "vxht_brothel_n_5ju_0_0.vlm"},
	{0x7CFFD31193557A3BULL, "vxlt_0_lowtown_n_f6p_0_1.vlm"},
	{0x7D00145EA3614B3EULL, "vxca_caverns2_83t_0_1.vlm"},
	{0x7D0022DF5178AF2DULL, "150429402.ogg"},
	{0x7D0028242B0988F4ULL, "vxlt_0_lowtown_n_by7_0_0.vlm"},
	{0x7D00655854DCFD13ULL, "6103922_m.fxe"},
	{0x7D008F9B60625106ULL, "mag160_varian_ilithis6016503.cl"},
	{0x7D00A2B08B65D2B4ULL, "vxdr_primevalentrance_2zl_0_0.vlm"},
	{0x7D0172863761D2EBULL, "vxht_0_hightown_9ze_0_0.vlm"},
	{0x7D01AFDD2839148DULL, "ht_chantry_n.arl"},
	{0x7D01EB13DAD0EFD7ULL, "vxca_cavern1_haunted_dpc_0_0.vlm"},
	{0x7D01EFE67F566624ULL, "vxlt_blackemporium_tw_0_0.vlm"},
	{0x7D0213095991F570ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small.ani"},
	{0x7D021D7F7874182BULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_act_c.anb"},
	{0x7D0246F08BC24139ULL, "vxht_0_hightown_a2_bx2_0_0.vlm"},
	{0x7D026C858CC07EDAULL, "6076502_m.fxe"},
	{0x7D029FAF6F53FAA2ULL, "vxlt_foundry_n_yy_0_0.vlm"},
	{0x7D02BE598846EEB2ULL, "279867591.ogg"},
	{0x7D03236BDE25C4EFULL, "character_character_p219.psh.fxo.dx11.assembly"},
	{0x7D035AC9C792AE67ULL, "6105898_m.fxe"},
	{0x7D039CD82D9E5155ULL, "dae000_arishok6075351.cl"},
	{0x7D03C36C2EEFC9CFULL, "follower_banter6165626.cl"},
	{0x7D04F084F85B1648ULL, "qun211_braggart6075478.cl"},
	{0x7D0567ACCAA64C69ULL, "6049815_m.fxe"},
	{0x7D056D27F87C009EULL, "vxht_0_hightown_a2_ber_0_0.vlm"},
	{0x7D05A5151B95F2F0ULL, "zz_qun_debug6028247.cl"},
	{0x7D05E08FFF27D801ULL, "vxlt_0_lowtown_n_euu_0_0.vlm"},
	{0x7D061070E0F7E876ULL, "610729564.ogg"},
	{0x7D06788FBF05F8ACULL, "438485352.ogg"},
	{0x7D06A20AF0F24AAAULL, "vxht_mansion_n_53m_0_1.vlm"},
	{0x7D06EB2A1169D97BULL, "24711035.ogg"},
	{0x7D073C30F0A1573FULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medj.phy"},
	{0x7D075987AB779F66ULL, "vxw_bonepitdestroyed_a3_264_0_0.vlm"},
	{0x7D07CB15BAFAB1B3ULL, "zz_one000_barker6179351.cl"},
	{0x7D086C8322334BA7ULL, "vxht_mansion2_1up_0_0.vlm"},
	{0x7D08EBEE81910203ULL, "one200_gold_followers6120128.cl"},
	{0x7D09078F3F60783CULL, "clf260_fell_orden6070492.cl"},
	{0x7D091A211EC8BBBCULL, "vxlt_gallowstemplar_6su_0_0.vlm"},
	{0x7D095E567FF5981DULL, "vxw_woundedcoast_7q_0_0.vlm"},
	{0x7D096D124E2AB318ULL, "hlac_ht_0_hightown_a2_-131071_0y.dds"},
	{0x7D09E5578AC857AAULL, "6076043_m.fxe"},
	{0x7D09F4203C0A2DCEULL, "vxlt_gallowscourtyard_n_7ap_0_1.vlm"},
	{0x7D0ABCECEA5B349BULL, "vxht_mansion_n_33k_0_0.vlm"},
	{0x7D0AE54E766957BCULL, "vfx0_vt_addv.psh.fxo"},
	{0x7D0B0EC353BC0A82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_loop.gad"},
	{0x7D0B228249656440ULL, "vxw_encounter2_camp_jf_0_0.vlm"},
	{0x7D0B2A58186F219BULL, "6099359_m.fxe"},
	{0x7D0BE59FCB644A60ULL, "vxlt_0_lowtown_docks_a3_bvk_0_0.vlm"},
	{0x7D0C3BDD2904DD87ULL, "vxht_keep_n_6c6_0_0.vlm"},
	{0x7D0C42D061E3F204ULL, "vxlt_foundry_n_103_0_0.vlm"},
	{0x7D0CAE4C2D109A07ULL, "874678058.ogg"},
	{0x7D0D03E3F3B14C22ULL, "169438375.ogg"},
	{0x7D0D5BCF79EA9B66ULL, "vxlt_0_lowtown_a2_f56_0_0.vlm"},
	{0x7D0D7C64F73413B9ULL, "vxca_caverns2_8pc_0_0.vlm"},
	{0x7D0E3D40AFD7A00AULL, "470734157.ogg"},
	{0x7D0E97E35F6737ACULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.dura.ani"},
	{0x7D0F07CE28980234ULL, "qun211_party_comment6182714.cl"},
	{0x7D0F7FCF986F210FULL, "one220_yevhen6102831.cl"},
	{0x7D0F8D438D634B06ULL, "6071231_m.fxe"},
	{0x7D0FAB4F47DEAB4BULL, "llac_lt_0_lowtown_n_-65535_4.dds"},
	{0x7D1007AEB47ABF12ULL, "vxlt_0_lowtown_docks_a3_cp2_0_0.vlm"},
	{0x7D1029C0B82E0E73ULL, "vxht_keep_n_3xf_0_0.vlm"},
	{0x7D111572678F1971ULL, "vxlt_gallowstemplar_56d_0_0.vlm"},
	{0x7D116ED4D270E031ULL, "vxht_0_hightown_n_b01_0_0.vlm"},
	{0x7D11D76AF79CCEA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_90_walk.ani"},
	{0x7D11F237E07F5F34ULL, "vxht_0_hightown_n_aps_0_0.vlm"},
	{0x7D11F9552A6C2C72ULL, "6038834_m.fxe"},
	{0x7D122F266BEBE807ULL, "6136737_m.fxe"},
	{0x7D1280618A130DFEULL, "542884265.ogg"},
	{0x7D1289F6265D5F17ULL, "zz_mcr_debug6072386.cl"},
	{0x7D1385ADDB9C771FULL, "6126369_m.fxe"},
	{0x7D13DF04EA72191EULL, "vxht_0_hightown_a2_bns_0_0.vlm"},
	{0x7D14526CF11DF081ULL, "vxlt_0_lowtown_n_ent_0_0.vlm"},
	{0x7D14C0AEB0A41C10ULL, "6079949_m.fxe"},
	{0x7D151766032FF2F8ULL, "one300_zevran6075572.cl"},
	{0x7D154428BFA15251ULL, "art\\characters\\weapons\\mauls\\w_mal_ml106a.mmh"},
	{0x7D155E28BFA17E0FULL, "art\\characters\\weapons\\mauls\\w_mal_ml106a.msh"},
	{0x7D15A4E35F6D6CEEULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.dura.gad"},
	{0x7D15CB649D6DB7E6ULL, "vxw_bonepitdestroyed_a3_215_0_0.vlm"},
	{0x7D15D6AEF8E847CBULL, "6027691_m.fxe"},
	{0x7D1673F0044EBD0FULL, "129240919.ogg"},
	{0x7D16A38CAA2E00EBULL, "hlac_lt_darktown_-1_4y.dds"},
	{0x7D16DE6014296427ULL, "vxht_mansion2_1xs_0_0.vlm"},
	{0x7D16EEA9C4C2FB3FULL, "darktown_male_fereldan_sta6175140.cl"},
	{0x7D179923694243FAULL, "gen00fl_isabela6106407.cl"},
	{0x7D183D30F0AFCC44ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medj.mmh"},
	{0x7D185330F0AFF1E6ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medj.msh"},
	{0x7D186878FAB98C3BULL, "680664845.ogg"},
	{0x7D187CCA1D66514BULL, "6167595_m.fxe"},
	{0x7D188F4FA191464BULL, "mag210_arianni6030469.cl"},
	{0x7D18E93F7CD4649AULL, "vxht_0_hightown_n_bi9_0_0.vlm"},
	{0x7D192D4B9BB6A9FAULL, "vxht_keep_62o_0_0.vlm"},
	{0x7D1964648BF0378AULL, "52147407.ogg"},
	{0x7D19945F3B5B7CC0ULL, "vxw_encounter2_camp_eg_0_0.vlm"},
	{0x7D1A041FD17B5F06ULL, "hla_ht_0_hightown_-131072_0y.dds"},
	{0x7D1A04B18EB4CF08ULL, "lgt250_way_dalish_conflict6147113.cl"},
	{0x7D1A2D5E64A0479FULL, "6095500_m.fxe"},
	{0x7D1A2F8A1FB42ECAULL, "zz_vault_debug6203363.cl"},
	{0x7D1A84F0F095474EULL, "art\\levels\\races\\proxy\\props\\_textures\\breadoven_d.dds"},
	{0x7D1A95C56EB06891ULL, "ts_dalish_02.gda"},
	{0x7D1B04B0CF954273ULL, "211582150.ogg"},
	{0x7D1B3B531CA48997ULL, "gen_im_trs_krk_junk1_03.uti"},
	{0x7D1B52D18EA95697ULL, "llac_lt_0_lowtown_a2_-262144_4.dds"},
	{0x7D1B96AA4D4BCA9FULL, "6040634_m.fxe"},
	{0x7D1BA286E9C1371FULL, "vxlt_undercityserialkiller_1sk_0_0.vlm"},
	{0x7D1BC6C2E5914988ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comde_0s.dds"},
	{0x7D1C285E3E34A34AULL, "qun220cr_javaris6071907.cl"},
	{0x7D1C4841755DAF4BULL, "6312317.ogg"},
	{0x7D1C9FE35F7382CDULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.dura.evt"},
	{0x7D1D32B2134BD6DDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medd_0n.dds"},
	{0x7D1D866A6F5203B4ULL, "vxdr_primevalentrance_2dr_0_0.vlm"},
	{0x7D1DE289AE06C79AULL, "vxht_0_hightown_n_9az_0_0.vlm"},
	{0x7D1E0560CBD2DDFCULL, "vxht_keep_4no_0_0.vlm"},
	{0x7D1E45F1AA5B4859ULL, "459630543.ogg"},
	{0x7D1FA0BB0DEA2F96ULL, "vxlt_undercityserialkiller_1mp_0_0.vlm"},
	{0x7D1FB11FCA48A977ULL, "hla_ht_0_hightown_a3_-65537_0y.dds"},
	{0x7D1FEF6AF7A9349AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_90_walk.evt"},
	{0x7D1FF59EACCCB9AEULL, "w_woundedcoast_150_0.phy"},
	{0x7D2039C353CEB068ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_loop.ani"},
	{0x7D206414D27577B4ULL, "vxht_0_hightown_n_aup_0_0.vlm"},
	{0x7D2073CDF909C780ULL, "493887683.ogg"},
	{0x7D209C8B091EED41ULL, "weapon_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0x7D20A1035DE20087ULL, "6131637_m.fxe"},
	{0x7D20EBCB3597BFB8ULL, "6119441_m.fxe"},
	{0x7D216B0C12250CC6ULL, "329467063.ogg"},
	{0x7D23A01D8B991452ULL, "isa211_isabela6202056.cl"},
	{0x7D25E527AE5C5AB3ULL, "68201550.ogg"},
	{0x7D260607533760BFULL, "976210293.ogg"},
	{0x7D2645E3F0F4EDFFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd02a.msh"},
	{0x7D264BE3F0F4F7A1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd02a.mmh"},
	{0x7D26A4A815A594F0ULL, "vxw_encounter2_camp_d8_0_0.vlm"},
	{0x7D26EE6AF7AF518DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_90_walk.gad"},
	{0x7D273432BFCC60B1ULL, "438223544.ogg"},
	{0x7D273C98E21F2D6DULL, "vxlt_gallowstemplar_f_2e5_0_0.vlm"},
	{0x7D274B14E4BE2BF6ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_exit.evt"},
	{0x7D27877850487E07ULL, "6079335_m.fxe"},
	{0x7D27B0A9C9FF9BAFULL, "lgt210_shp_kill_qunari1.bnk"},
	{0x7D27E549B5578B40ULL, "qun221ar_chantry_nt.are"},
	{0x7D28340199CACA80ULL, "6126793_m.fxe"},
	{0x7D2884B2C311E37DULL, "hlac_lt_undercityencounter_-1_36z.dds"},
	{0x7D28DD45C3E7C259ULL, "lowtown_male_lowlife1_sta6164988.cl"},
	{0x7D28E1E6F46173B5ULL, "542233828.ogg"},
	{0x7D29ADD69D11DBF4ULL, "767000608.ogg"},
	{0x7D2AD2A9CA020B17ULL, "lgt210_shp_kill_qunari1.cnv"},
	{0x7D2ADC49B559B1C2ULL, "qun221ar_chantry_nt.ncs"},
	{0x7D2AE7E2A4E5B074ULL, "82699022.ogg"},
	{0x7D2B39132D1B91C3ULL, "6106046_m.fxe"},
	{0x7D2B7B3B8EB7E29AULL, "mcr370_meredith6084405.cl"},
	{0x7D2C4036A2D6F654ULL, "hlac_lt_undercitydungeon_-1_52z.dds"},
	{0x7D2D4214E4C28781ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_exit.gad"},
	{0x7D2D89364B078E9BULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.impa.ani"},
	{0x7D2ED503FB5E7208ULL, "plt_dre120pt_reqs_maps.nss"},
	{0x7D2EE6406F0D2479ULL, "vxht_mansion_n_6p5_0_0.vlm"},
	{0x7D2F07B52133A0A5ULL, "vxlt_0_lowtown_n_equ_0_0.vlm"},
	{0x7D2F60D85E57BCCCULL, "lt_hangedman_mmap.dds"},
	{0x7D3075DBA88B241CULL, "vxlt_0_lowtown_a2_n_gvk_0_0.vlm"},
	{0x7D30EA56AA1243F5ULL, "mag200ip_corpse_1.utp"},
	{0x7D30EE28476F0770ULL, "lt_0_lowtown_docks_a3_n_157_0.mao"},
	{0x7D30F228476F0D83ULL, "lt_0_lowtown_docks_a3_n_157_0.mmh"},
	{0x7D310028476F25DDULL, "lt_0_lowtown_docks_a3_n_157_0.msh"},
	{0x7D312406811AF1B6ULL, "w_blightlands_66_0.mao"},
	{0x7D313006811B0655ULL, "w_blightlands_66_0.mmh"},
	{0x7D313206811B09CBULL, "w_blightlands_66_0.msh"},
	{0x7D316A1BE83A09C6ULL, "6037605_f.fxe"},
	{0x7D3190B02D575711ULL, "741583214.ogg"},
	{0x7D3191FC75808B27ULL, "6164239_m.fxe"},
	{0x7D31F240D4F32BADULL, "llac_ht_mansion2_818_41.dds"},
	{0x7D31F449B55FF953ULL, "qun221ar_chantry_nt.lst"},
	{0x7D327D71036CDEC0ULL, "6042139_f.fxe"},
	{0x7D32908CC1885E8DULL, "6176223_m.fxe"},
	{0x7D32D311380D62B7ULL, "vxht_0_hightown_a2_av9_0_0.vlm"},
	{0x7D33054B0E41BFB9ULL, "vxlt_gallowstemplar_5jh_0_0.vlm"},
	{0x7D330D5D35FAB86DULL, "418649670.ogg"},
	{0x7D333C6B2673645CULL, "vxlt_0_lowtown_docks_aoh_0_0.vlm"},
	{0x7D33922EB61B2694ULL, "vxw_woundedcoast_1c1_0_0.vlm"},
	{0x7D339995780CE63CULL, "6071951_f.fxe"},
	{0x7D33C09DA4F9AF84ULL, "6122497_f.fxe"},
	{0x7D33C219F8159527ULL, "vxht_mansion2_1fm_0_0.vlm"},
	{0x7D344314E4C8A82FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_exit.ani"},
	{0x7D3464AE38F7A29AULL, "gen_im_trs_dem_lgt_cht_r0.uti"},
	{0x7D34EAE3178DB784ULL, "99808170.ogg"},
	{0x7D35DB7E017C6129ULL, "qun211cr_elf_warrior6072336.cl"},
	{0x7D35DE03FB64A01BULL, "plt_dre120pt_reqs_maps.plo"},
	{0x7D362230BFDA68F2ULL, "gen00fl_merrill6129541.cl"},
	{0x7D3689B7AA5E9811ULL, "vxw_bonepit_74_0_0.vlm"},
	{0x7D36CC3056BD36E8ULL, "vxw_bonepit_mo_0_0.vlm"},
	{0x7D36FA890B9A5022ULL, "1048194377.ogg"},
	{0x7D37B057567FBBC9ULL, "802603206.ogg"},
	{0x7D37B078AFE98BBEULL, "hlac_lt_0_lowtown_docks_a3_-196609_1x.dds"},
	{0x7D37C23EFC0482FDULL, "873922628.ogg"},
	{0x7D382B288A6A8554ULL, "6125194_m.fxe"},
	{0x7D3846985A760CF6ULL, "dae000_meredith6092563.cl"},
	{0x7D387D94B8675180ULL, "vxlt_0_lowtown_n_e2m_0_0.vlm"},
	{0x7D38960D0D03C792ULL, "vxw_woundedcoast_3f_0_0.vlm"},
	{0x7D391535F35E0090ULL, "w_bonepit_234_0.mao"},
	{0x7D391935F35E07A3ULL, "w_bonepit_234_0.mmh"},
	{0x7D392735F35E1FFDULL, "w_bonepit_234_0.msh"},
	{0x7D39665B73563723ULL, "hlac_w_encounter2_327678_0z.dds"},
	{0x7D396B53FEE901E2ULL, "vxht_mansion2_28a_0_0.vlm"},
	{0x7D398694F3A69937ULL, "llac_lt_undercitydungeon_-1_20.dds"},
	{0x7D3992E408ADB38FULL, "vxht_mansion_n_6it_0_0.vlm"},
	{0x7D39E6C1452FB5B1ULL, "vxdr_primevalentrance_320_0_0.vlm"},
	{0x7D39FFB44856DDDCULL, "zz_qcr_debug6052129.cl"},
	{0x7D3A1EED14B0650DULL, "art\\levels\\races\\proxy\\props\\door\\prp_blackdoor.mmh"},
	{0x7D3A22ED14B06C5EULL, "art\\levels\\races\\proxy\\props\\door\\prp_blackdoor.mao"},
	{0x7D3A30ED14B08393ULL, "art\\levels\\races\\proxy\\props\\door\\prp_blackdoor.msh"},
	{0x7D3A51095B36644FULL, "vxht_0_hightown_n_c0f_0_0.vlm"},
	{0x7D3A6A24533E34B4ULL, "vxdr_primevalentrance_2np_0_0.vlm"},
	{0x7D3A91364B122692ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.impa.evt"},
	{0x7D3A919E03CC8CF2ULL, "dae000ip_crowd_ff2_talk10.utp"},
	{0x7D3AE13548A7CD23ULL, "gen00fl_anders6119688.cl"},
	{0x7D3B419897880CD0ULL, "llac_ht_0_hightown_a3_0_2.dds"},
	{0x7D3B961A5A31AF12ULL, "6144580_m.fxe"},
	{0x7D3BA797F3301925ULL, "llac_w_sundermount_65531_0.dds"},
	{0x7D3BCF380FB0F7DBULL, "w_bonepit_11_0.rml"},
	{0x7D3BF89809160943ULL, "50577292.ogg"},
	{0x7D3C0E64539AA339ULL, "vxht_0_hightown_a2_bl6_0_0.vlm"},
	{0x7D3C36DF3E73CB1EULL, "vxlt_0_lowtown_a2_het_0_0.vlm"},
	{0x7D3C5FE284424C5FULL, "fex_bd_a_rf_icn.msh"},
	{0x7D3C65E284425601ULL, "fex_bd_a_rf_icn.mmh"},
	{0x7D3C6B093D7162F1ULL, "vxlt_gallowstemplar_59y_0_0.vlm"},
	{0x7D3CCB1F20538DA7ULL, "368031110.ogg"},
	{0x7D3D8EF5BDEA304DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandleft.msh"},
	{0x7D3D99E00D1F8CC7ULL, "981595338.ogg"},
	{0x7D3D9CF5BDEA4780ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandleft.mao"},
	{0x7D3DA0F5BDEA4ED3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandleft.mmh"},
	{0x7D3DCF851983F0B8ULL, "6113241_m.fxe"},
	{0x7D3E12AB5C38B988ULL, "vxlt_0_lowtown_ens_0_1.vlm"},
	{0x7D3E702A7EF1858CULL, "gen00fl_merrill6150965.cl"},
	{0x7D3E8522B86CCB91ULL, "follower_banter6149208.cl"},
	{0x7D3EF9DE4BCDE221ULL, "vxdr_primevalentrance_19w_0_0.vlm"},
	{0x7D3F31802A9BFEAEULL, "hlac_ca_caverns1_mines_24424_6z.dds"},
	{0x7D3F6696D3685D5FULL, "220964451.ogg"},
	{0x7D3FBD8BCDC6CCA4ULL, "vxlt_0_lowtown_docks_a3_n_95j_0_0.vlm"},
	{0x7D404279C7937C2DULL, "vxlt_gallowstemplar_f_5av_0_0.vlm"},
	{0x7D405E161C77A3E4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magc_0d.dds"},
	{0x7D413CC96239066EULL, "6079030_f.fxe"},
	{0x7D4177280037D839ULL, "vxlt_gallowstemplar_5wo_0_0.vlm"},
	{0x7D41B0364B187A25ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.impa.gad"},
	{0x7D41C9F18ADF641EULL, "zz_bet_debug6113395.cl"},
	{0x7D41CF0D858149B7ULL, "vxw_encounter2_kv_0_0.vlm"},
	{0x7D42040BA8113F14ULL, "vxw_woundedcoast_1n6_0_0.vlm"},
	{0x7D4233043A61F471ULL, "zz_dae_debug6142959.cl"},
	{0x7D4233F6EA0BADA7ULL, "vxht_mansion2_1il_0_0.vlm"},
	{0x7D424B491DFD3102ULL, "631720160.ogg"},
	{0x7D4336D06AD6C0D7ULL, "w_woundedcoast_18_0.rml"},
	{0x7D434F0EC52D5B93ULL, "vxdr_0_deeproads_cvr_0_2.vlm"},
	{0x7D4394E9A59AAAE8ULL, "static_static_p5.psh.fxo.dx11.assembly"},
	{0x7D4411677F014018ULL, "vxlt_backalley_n_1w0_0_1.vlm"},
	{0x7D4424EFA02B3945ULL, "vxht_keep_62d_0_0.vlm"},
	{0x7D44EB251AAAA4A9ULL, "hlac_w_encounter1_-65536_0x.dds"},
	{0x7D4509BCB5ACE449ULL, "gen00fl_merrill6129845.cl"},
	{0x7D4591D820747B02ULL, "hlac_w_encounter1_65536_0x.dds"},
	{0x7D45C483DE005E9CULL, "vxht_mansion2b_1vr_0_0.vlm"},
	{0x7D46385A047081E5ULL, "ss_t_four_v5.cut"},
	{0x7D4733AF1CDB2D1FULL, "468725351.ogg"},
	{0x7D4743C277B8087CULL, "58486450.ogg"},
	{0x7D475613BC5BF853ULL, "vxlt_gallowstemplar_6dj_0_1.vlm"},
	{0x7D47CA92F945ECB7ULL, "art\\vfx\\combat\\vfx_str_appearance_c.dura.gad"},
	{0x7D4804C0FAA3CC0FULL, "vxht_mansion_n_6lw_0_0.vlm"},
	{0x7D48BD8C4BE6CC37ULL, "vxht_0_hightown_ccm_0_1.vlm"},
	{0x7D4901400FA1346BULL, "llac_dr_primevaltomb_663_0.dds"},
	{0x7D4908A0340F001DULL, "vxlt_backalley_g_1w9_0_0.vlm"},
	{0x7D49898130827C9AULL, "w_encounter3_61_0.mmh"},
	{0x7D498D81308283D1ULL, "w_encounter3_61_0.mao"},
	{0x7D499B8130829BE0ULL, "w_encounter3_61_0.msh"},
	{0x7D4A07F74C27C792ULL, "6141680_m.fxe"},
	{0x7D4A4414637582C5ULL, "vxlt_gallowstemplar_n_4yk_0_0.vlm"},
	{0x7D4AA8BC3069E39EULL, "vxlt_0_lowtown_a2_hbu_0_0.vlm"},
	{0x7D4ABC5311BE11FDULL, "weapon2_dy.psh.fxo"},
	{0x7D4AFBEEE104A644ULL, "llac_lt_0_lowtown_docks_a3_-2_0.dds"},
	{0x7D4B7D8A0893B6B7ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dsla_0d.dds"},
	{0x7D4B9185AD9B79E9ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_001a_0n.dds"},
	{0x7D4BB13142683C5EULL, "vxlt_undercityserialkiller_187_0_0.vlm"},
	{0x7D4BDD44DED788D2ULL, "hlac_lt_darktown_-1_66z.dds"},
	{0x7D4BF455641DE1EFULL, "196123951.ogg"},
	{0x7D4C391A1C12BC44ULL, "dre000_vampire6034359.cl"},
	{0x7D4D87F504BB23AEULL, "672067101.ogg"},
	{0x7D4DDF92F94A7C44ULL, "art\\vfx\\combat\\vfx_str_appearance_c.dura.evt"},
	{0x7D4E47F3CF42A87DULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_7x.dds"},
	{0x7D4E5775E070D3DDULL, "6050318_m.fxe"},
	{0x7D4E5BAF57B0B8AAULL, "vxlt_undercityencounter_12i_0_0.vlm"},
	{0x7D4E7A395E61FB04ULL, "942862854.ogg"},
	{0x7D4ECB4CE11686FFULL, "hairalpha.mat"},
	{0x7D4F2D003DAA8299ULL, "473351044.ogg"},
	{0x7D4F5D588D1BEC61ULL, "vxlt_gallowstemplar_n_4n6_0_0.vlm"},
	{0x7D4FE904F22DF0B9ULL, "vxlt_gallowstemplar_5pn_0_0.vlm"},
	{0x7D4FFF4B23CDE9F4ULL, "vxht_chantry_n_2sy_0_0.vlm"},
	{0x7D5075E89A075794ULL, "vxw_woundedcoast_1i7_0_0.vlm"},
	{0x7D5083D066CEA59BULL, "386808950.ogg"},
	{0x7D50E2DDCAA2260FULL, "204250042.ogg"},
	{0x7D511DF17ECEDCB4ULL, "qun221_viscount6069921.cl"},
	{0x7D513F7A734394A3ULL, "275738825.ogg"},
	{0x7D51697019538ECAULL, "hm_gencr_merc_archer.mor"},
	{0x7D51968BDC79DDD5ULL, "vxlt_0_lowtown_a2_n_c_e8h_0_0.vlm"},
	{0x7D51DE48A09DD08BULL, "58544632.ogg"},
	{0x7D51E45E8FDB9444ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop.ani"},
	{0x7D5245BF17165EA4ULL, "6129441_f.fxe"},
	{0x7D527F12AE6E7F83ULL, "gen00fl_fenris6126745.cl"},
	{0x7D52B88FD308F9C6ULL, "vxlt_undercityencounter_1gn_0_0.vlm"},
	{0x7D52F69BEE00DEB0ULL, "393027347.ogg"},
	{0x7D5317B08FA7F316ULL, "30514928.ogg"},
	{0x7D533EDD10811DF2ULL, "hla_ht_0_hightown_65538_1z.dds"},
	{0x7D534CE3F11BE71EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd02a.phy"},
	{0x7D535F7EC4704B6BULL, "555329404.ogg"},
	{0x7D5386A166A53383ULL, "vxlt_0_lowtown_docks_n_b9q_0_0.vlm"},
	{0x7D544A954D6EA17DULL, "ser221_gamlen6135073.cl"},
	{0x7D55304518B10877ULL, "hlac_ca_caverns1_24425_7x.dds"},
	{0x7D5584D1B8A12678ULL, "28963871.ogg"},
	{0x7D55C0DC45AD51F0ULL, "dae111st_gamlens_intro.stg"},
	{0x7D55E3B73DF0D4F9ULL, "160885926.ogg"},
	{0x7D55F99C4AB35A66ULL, "hlac_ca_caverns1_mines_24424_24x.dds"},
	{0x7D56362C5DC56278ULL, "6108592_f.fxe"},
	{0x7D56B22A3E18CB2FULL, "vxht_chantry_n_1yk_0_0.vlm"},
	{0x7D56CA302D984CCCULL, "mag330_walter6064740.cl"},
	{0x7D56EE66A169AB16ULL, "one280_golem6115976.cl"},
	{0x7D574D1CCE83A0F6ULL, "vxw_bonepitmining_a2_dl_0_0.vlm"},
	{0x7D57645EFF90AC9CULL, "hlac_dr_primevaltomb_659_9y.dds"},
	{0x7D5788B9B670B873ULL, "65522193.ogg"},
	{0x7D5791E3D08361A6ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_rtlb.anb"},
	{0x7D57A148D9C4E017ULL, "hlac_lt_0_lowtown_n_-196609_2y.dds"},
	{0x7D57A9FD75DF299AULL, "vxht_0_hightown_n_9uv_0_0.vlm"},
	{0x7D57FCB1EE05EA6DULL, "dae000_gamlen6147217.cl"},
	{0x7D587BAB00D3067BULL, "llac_dr_dungeon_3214_3.dds"},
	{0x7D58E15E8FE1ADA6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop.gad"},
	{0x7D593E20B8C75E54ULL, "54172348.ogg"},
	{0x7D59471AC9C582CCULL, "vxht_0_hightown_b7f_0_0.vlm"},
	{0x7D59C5E6CA433E16ULL, "dae200_bran6094717.cl"},
	{0x7D59CAEFBC4C685BULL, "vxlt_0_lowtown_a2_n_c_bz3_0_0.vlm"},
	{0x7D59DBF7A84328CBULL, "vxht_bartrand_n_5wz_0_0.vlm"},
	{0x7D59DF17E37656F2ULL, "6055440_m.fxe"},
	{0x7D5A326206E7D10EULL, "6101958_m.fxe"},
	{0x7D5A8EC4711968A4ULL, "vrc221_varric6040701.cl"},
	{0x7D5B2AA1BF1BDD9DULL, "41990619.ogg"},
	{0x7D5BF792F956E23DULL, "art\\vfx\\combat\\vfx_str_appearance_c.dura.ani"},
	{0x7D5BFFDE5E18E466ULL, "hlac_w_encounter3_-393214_0x.dds"},
	{0x7D5C52EAE510ED3EULL, "vxht_brothel_n_3r2_0_0.vlm"},
	{0x7D5C60B5364225F3ULL, "hlac_lt_0_lowtown_n_-2_4z.dds"},
	{0x7D5CC38B8A751BA4ULL, "character_character_p108.psh.fxo"},
	{0x7D5D17F78D66FBB9ULL, "hlac_lt_undercityserialkiller_-1_12z.dds"},
	{0x7D5D2633AB7FAD2DULL, "vxlt_gallowstemplar_f_5ot_0_0.vlm"},
	{0x7D5D37EC048E18AFULL, "6127649_m.fxe"},
	{0x7D5D4286243367CCULL, "456410848.ogg"},
	{0x7D5D5D516E4A7A42ULL, "fen110_fenris6118034.cl"},
	{0x7D5D6053775BCC50ULL, "29233474.ogg"},
	{0x7D5D8847846C4128ULL, "6083377_m.fxe"},
	{0x7D5D9745D577BB94ULL, "mer120_hubert6017817.cl"},
	{0x7D5E10B9031F3057ULL, "ht_0_hightownexplosion_n_42_0_lc.mmh"},
	{0x7D5E3A7423EF3315ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend03_n.dds"},
	{0x7D5EB2C7696D7AB7ULL, "vxw_encounter2_qp_0_0.vlm"},
	{0x7D5EBCFDDB71DB47ULL, "6127123_f.fxe"},
	{0x7D5EE7C58BFD7014ULL, "vxw_woundedcoast_1t4_0_0.vlm"},
	{0x7D5F0222D8047937ULL, "6081259_m.fxe"},
	{0x7D5F545D2957D499ULL, "vxlt_warehouse_vp_0_0.vlm"},
	{0x7D5FC1D95EE2F18FULL, "197686334.ogg"},
	{0x7D5FDC5E8FE7C385ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_stop.evt"},
	{0x7D5FF49EAD029C71ULL, "w_woundedcoast_150_0.mmh"},
	{0x7D5FF67CD2ED9DB8ULL, "hlac_ht_keep_n_3731_2x.dds"},
	{0x7D60009EAD02B012ULL, "w_woundedcoast_150_0.mao"},
	{0x7D600E9EAD02C82FULL, "w_woundedcoast_150_0.msh"},
	{0x7D60366A56E98A78ULL, "art\\levels\\ztools\\default\\generic\\prp_joe_3.mao"},
	{0x7D605E67D2A76335ULL, "vxca_caverns1_mines_iig_0_0.vlm"},
	{0x7D608504ED88A91EULL, "vxca_caverns2prc_8o0_0_0.vlm"},
	{0x7D60D9E45DAA104DULL, "91082128.ogg"},
	{0x7D60EED3A0B5C37BULL, "gen00fl_fenris6107205.cl"},
	{0x7D610016A08F9826ULL, "hlac_lt_gallowstemplar_2394_3y.dds"},
	{0x7D6116603EE46503ULL, "llac_lt_warehouse_n_65536_2.dds"},
	{0x7D61516EFD3919F2ULL, "524730371.ogg"},
	{0x7D618D9C2248A465ULL, "zz_dre_debug6168291.cl"},
	{0x7D61AE6C546878EAULL, "vxlt_hangedman_n_vz_0_0.vlm"},
	{0x7D61F87E589B4C03ULL, "vxlt_0_lowtown_docks_n_b6p_0_0.vlm"},
	{0x7D620E46D86B306CULL, "155673317.ogg"},
	{0x7D6226D272B38D8BULL, "430704142.ogg"},
	{0x7D623609BBBB5067ULL, "278711454.ogg"},
	{0x7D624EA6393C0FD3ULL, "vxca_caverns1_mines_i1k_0_2.vlm"},
	{0x7D62693B6CA9A240ULL, "vxht_0_hightown_n_as2_0_0.vlm"},
	{0x7D630C680C1BA9B1ULL, "hlac_lt_0_lowtown_docks_-131074_8x.dds"},
	{0x7D638E8B11B7FADBULL, "hlac_ca_caverns2_5802_57x.dds"},
	{0x7D63D7FCB5EA38D8ULL, "vxlt_warehouse_b1_0_0.vlm"},
	{0x7D640112F6637A71ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_hammer_sa.dds"},
	{0x7D6436F47F99B2B1ULL, "6166198_m.fxe"},
	{0x7D64663EA79C5539ULL, "vxlt_0_lowtown_a2_hg5_0_0.vlm"},
	{0x7D64C0C65DB0E8C4ULL, "pro000st_varric_intro.stg"},
	{0x7D64E51145C81E7CULL, "dae000_merchant_access_hf6151503.cl"},
	{0x7D654C19BD1BB380ULL, "6140186_m.fxe"},
	{0x7D65896AC7B46764ULL, "794532767.ogg"},
	{0x7D65B3318EC43A6FULL, "vxw_blightlands_11r_0_0.vlm"},
	{0x7D65BEF9C079B976ULL, "vxw_bonepitmining_a2_cm_0_0.vlm"},
	{0x7D65EC5A17FB311DULL, "vxlt_backalley_g_1y7_0_0.vlm"},
	{0x7D66B61824BFF790ULL, "219749114.ogg"},
	{0x7D66B61F782635CCULL, "6189909_m.fxe"},
	{0x7D66BA6D83B09DC0ULL, "118799588.ogg"},
	{0x7D66F00FA3958B2AULL, "vxlt_0_lowtown_docks_a3_d2s_0_0.vlm"},
	{0x7D66FB8CC41AE3EAULL, "vrc221_varric6044440.cl"},
	{0x7D6727CE4761B3C5ULL, "vxlt_gallowstemplar_n_4si_0_0.vlm"},
	{0x7D672A2EBA3A1DFBULL, "llac_lt_0_lowtown_a3_n_c_-131072_3.dds"},
	{0x7D675DE13A19D4F6ULL, "409162167.ogg"},
	{0x7D67AEBBAC60664FULL, "364730354.ogg"},
	{0x7D67D46990AFF56BULL, "6197862_m.fxe"},
	{0x7D67E1169DF0931EULL, "zz_meredith.ncs"},
	{0x7D68066D45123B87ULL, "6041769_m.fxe"},
	{0x7D695D9DF6ABFFC7ULL, "174059908.ogg"},
	{0x7D6986E576275956ULL, "dm_lgt_harrow_carta.mor"},
	{0x7D69C9F13EC44898ULL, "432222089.ogg"},
	{0x7D69D07CD7F2505EULL, "hm_ave3_insurgent2.mor"},
	{0x7D69FD5CD1C9B0D6ULL, "218982073.ogg"},
	{0x7D6A3D4F7063861CULL, "vxlt_0_lowtown_a2_n_gbo_0_0.vlm"},
	{0x7D6B047CD8FB1E65ULL, "6162037_m.fxe"},
	{0x7D6B18D3C84988FBULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8_tool.phy"},
	{0x7D6B5B5957F68F5CULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravea_n.dds"},
	{0x7D6B81069AD99283ULL, "mag110_vincento6031821.cl"},
	{0x7D6BAA1A95D306B3ULL, "vxht_keep_n_60s_0_0.vlm"},
	{0x7D6BF8539213AEEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv2.evt"},
	{0x7D6C0E4A40EB06AFULL, "art\\characters\\weapons\\quivers\\w_qui_qu03a.mao"},
	{0x7D6C1A4A40EB1B4CULL, "art\\characters\\weapons\\quivers\\w_qui_qu03a.mmh"},
	{0x7D6C204A40EB253EULL, "art\\characters\\weapons\\quivers\\w_qui_qu03a.msh"},
	{0x7D6C5D831B1434A8ULL, "6141156_m.fxe"},
	{0x7D6C933DC0EA1327ULL, "782683799.ogg"},
	{0x7D6D293C8C7B30ABULL, "dae200_cullen6187960.cl"},
	{0x7D6D7F12D5538447ULL, "one220cr_dera.utc"},
	{0x7D6DFD6DFB7C46C0ULL, "vxdr_primevalentrance_2e5_0_0.vlm"},
	{0x7D6E0104AD4B192BULL, "vxlt_warehouse_n_av_0_0.vlm"},
	{0x7D6E7A45C0660ED5ULL, "vxlt_0_lowtown_a2_n_c_e2v_0_0.vlm"},
	{0x7D6E8EEF2BE8A473ULL, "6097344_m.fxe"},
	{0x7D6ED2AA6CD29113ULL, "317218160.ogg"},
	{0x7D6EE40F9B457D76ULL, "hlac_lt_0_lowtown_a2_n_c_0_2z.dds"},
	{0x7D6EFBB62EA5E9B1ULL, "vxlt_0_lowtown_eod_0_0.vlm"},
	{0x7D6F3EB878C8556EULL, "106047967.ogg"},
	{0x7D6FE277B3DABDD7ULL, "295754630.ogg"},
	{0x7D701E876B2FE6F7ULL, "252520456.ogg"},
	{0x7D706989C5270ED1ULL, "art\\vfx\\combat\\vfx_m_spi_conj_c.anb"},
	{0x7D706E8EA6090DDCULL, "aflip_good.dds"},
	{0x7D70B4A665EEBF88ULL, "322373703.ogg"},
	{0x7D70D68096D9B960ULL, "770920937.ogg"},
	{0x7D712D13490E23FDULL, "vxdr_0_deeproads_ev0_0_0.vlm"},
	{0x7D714FDCA1907BB2ULL, "6109828_m.fxe"},
	{0x7D715092C99A511AULL, "qun120_arishok6052912.cl"},
	{0x7D71598ECADD4F98ULL, "art\\characters\\heraldry\\hld_dalish_1.dds"},
	{0x7D716B16B9ED576FULL, "vxlt_backalley_n_1r5_0_0.vlm"},
	{0x7D71A71B9271BBDDULL, "185303470.ogg"},
	{0x7D71CD76CE5556F9ULL, "zz_fen_debug6165696.cl"},
	{0x7D71EE9F7CDC9F57ULL, "vxdr_primevaltomb_76z_0_0.vlm"},
	{0x7D721153921844ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv2.gad"},
	{0x7D7222AC3F0DE1FDULL, "735302045.ogg"},
	{0x7D7287ECA49BA0EBULL, "ach_oth_templars.dds"},
	{0x7D72F15D19B161CCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_basic.msh"},
	{0x7D72F75D19B16BBEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_basic.mmh"},
	{0x7D73035D19B18055ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_basic.mao"},
	{0x7D73183A7EA7CF6FULL, "darktown_male_lowlife_mob6174395.cl"},
	{0x7D7324AD8F139620ULL, "6137052_m.fxe"},
	{0x7D7333159A385C45ULL, "6053009_m.fxe"},
	{0x7D7410DEDD54BE73ULL, "372677807.ogg"},
	{0x7D7430D6B26FD1F6ULL, "vxw_bonepitmining_a2_nn_0_0.vlm"},
	{0x7D74399D9E31BA34ULL, "6161650_f.fxe"},
	{0x7D74708B1F4B687AULL, "6110075_m.fxe"},
	{0x7D7499C364E04806ULL, "6177431_f.fxe"},
	{0x7D76775F6794FE54ULL, "274975747.ogg"},
	{0x7D76F59980FD247FULL, "6172751_m.fxe"},
	{0x7D771F9128DB9EF9ULL, "vxht_mansion_n_6o8_0_0.vlm"},
	{0x7D77E0AE07409A68ULL, "6087231_f.fxe"},
	{0x7D77E684A58812F0ULL, "gen00fl_anders6189370.cl"},
	{0x7D7819A920CAA4D0ULL, "splash_additive.mao"},
	{0x7D784C4E098982CBULL, "6172401_m.fxe"},
	{0x7D78F751C59721C7ULL, "vxht_bartrand_n_67a_0_0.vlm"},
	{0x7D78FFB15A52E2A5ULL, "6119466_m.fxe"},
	{0x7D79005525642A66ULL, "vxca_caverns2_5xm_0_0.vlm"},
	{0x7D791053921E61AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv2.ani"},
	{0x7D791FCCC4AF1242ULL, "6174965_m.fxe"},
	{0x7D7962683415254FULL, "one140cr_charade6128513.cl"},
	{0x7D799D18FD87A917ULL, "6143526_m.fxe"},
	{0x7D7A1BA5E87A49AFULL, "6129849_m.fxe"},
	{0x7D7A582E94F84628ULL, "6161118_m.fxe"},
	{0x7D7BC6DE21454923ULL, "6109025_m.fxe"},
	{0x7D7BD0CB1BFB8EFDULL, "vxlt_warehouse_18s_0_0.vlm"},
	{0x7D7C38410DD60A04ULL, "hlac_ht_keep_n_3732_12x.dds"},
	{0x7D7D00B944B66942ULL, "zz_approval_debug6003784.cl"},
	{0x7D7D2782AD79FECAULL, "237586821.ogg"},
	{0x7D7D5F70C11D6C62ULL, "168936393.ogg"},
	{0x7D7D751B3FEF4730ULL, "gen_im_acc_blt_act2_01.uti"},
	{0x7D7E3D51156669B0ULL, "6071704_f.fxe"},
	{0x7D7E8A89C533844AULL, "art\\vfx\\combat\\vfx_m_spi_conj_c.mmh"},
	{0x7D80DBE10661756BULL, "281858140.ogg"},
	{0x7D80E261DC3F746DULL, "lgt160_wk_soph_sophia6119829.cl"},
	{0x7D80ECE6605D58D2ULL, "403076163.ogg"},
	{0x7D817E401899F9AEULL, "one350tr_spiders.ncs"},
	{0x7D81BA37FEC0DCCCULL, "zz_ave_debug6065035.cl"},
	{0x7D822EB4E977BC36ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkb.evt"},
	{0x7D8278952047717DULL, "442664612.ogg"},
	{0x7D827B982F787A87ULL, "22643903.ogg"},
	{0x7D82914B805BA772ULL, "6209897_f.fxe"},
	{0x7D837C1667BB68AFULL, "vxw_woundedcoast_1bt_0_1.vlm"},
	{0x7D83F95E9311BC56ULL, "w_bonepit_76_0.mao"},
	{0x7D84055E9311D0F5ULL, "w_bonepit_76_0.mmh"},
	{0x7D84075E9311D46BULL, "w_bonepit_76_0.msh"},
	{0x7D844066A459A7F5ULL, "vxca_caverns2_7m7_0_0.vlm"},
	{0x7D8469BE6563A893ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_reaction.evt"},
	{0x7D849E1A10E7C93BULL, "dae200_saemus.bnk"},
	{0x7D84CAF471DFD6BCULL, "6101505_m.fxe"},
	{0x7D855D01771471E6ULL, "hlac_lt_darktown_-1_12y.dds"},
	{0x7D8568D1C60F9A11ULL, "6043233_m.fxe"},
	{0x7D8584D441837D92ULL, "vxw_bonepitmining_a2_1it_0_0.vlm"},
	{0x7D85AC155C144427ULL, "6048151_m.fxe"},
	{0x7D86A53EA0EC6806ULL, "6071060_m.fxe"},
	{0x7D8723FB4AC86821ULL, "vxht_0_hightown_n_av0_0_0.vlm"},
	{0x7D8742FB4C6AAE7FULL, "vxlt_undercityencounter_ty_0_0.vlm"},
	{0x7D8756F679AD07DDULL, "vxdr_primevaltomb_75q_0_0.vlm"},
	{0x7D87A01A10EA02FBULL, "dae200_saemus.cnv"},
	{0x7D87A9DEA735B7E9ULL, "249097990.ogg"},
	{0x7D87BA8A1DCC3CF2ULL, "vxht_0_hightown_a2_au5_0_0.vlm"},
	{0x7D87CBBC55D0CF19ULL, "193535803.ogg"},
	{0x7D883A2DBF8CBE8BULL, "vxht_keep_n_4b3_0_0.vlm"},
	{0x7D8844A9794A0B72ULL, "vxw_encounter2_os_0_0.vlm"},
	{0x7D888DFD6C495A6AULL, "hlac_lt_0_lowtown_a2_-131073_3z.dds"},
	{0x7D88C5DEADCEB670ULL, "911972023.ogg"},
	{0x7D88EA883EC93FDBULL, "4641653.ogg"},
	{0x7D8905AC756409E9ULL, "vxca_caverns2prc_62p_0_2.vlm"},
	{0x7D8925B4E97DCBC1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkb.gad"},
	{0x7D893C1A097C723AULL, "hlac_lt_0_lowtown_131072_2y.dds"},
	{0x7D89470400A71B6EULL, "w_blightlands_85_0.msh"},
	{0x7D89550400A732BFULL, "w_blightlands_85_0.mao"},
	{0x7D89578FA1DC8427ULL, "29297530.ogg"},
	{0x7D89610400A7475CULL, "w_blightlands_85_0.mmh"},
	{0x7D89D4C0CF08654EULL, "vxlt_undercityencounter_uw_0_0.vlm"},
	{0x7D89DCE6D6C144F5ULL, "llac_lt_gallowstemplar_2383_1.dds"},
	{0x7D89EC417DE4CD21ULL, "6185539_m.fxe"},
	{0x7D8A1395BCE0171AULL, "hlac_lt_0_lowtown_a2_65538_2z.dds"},
	{0x7D8A14E3C943DA74ULL, "6128984_m.fxe"},
	{0x7D8A38BB133B61A3ULL, "6102125_m.fxe"},
	{0x7D8A62BE656808D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_reaction.gad"},
	{0x7D8ACED8ED45D4C6ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_3y.dds"},
	{0x7D8AD56999C53FDAULL, "42165365.ogg"},
	{0x7D8B03DA56F1A34CULL, "977514351.ogg"},
	{0x7D8B44AE06C66A42ULL, "vxdr_primevaltomb_503_0_0.vlm"},
	{0x7D8B542DD1B6DE68ULL, "llac_lt_foundry_n_320_1.dds"},
	{0x7D8B5486EBEA3C09ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\blendtree\\trooper.anb"},
	{0x7D8B611AE82F7EC2ULL, "6149796_m.fxe"},
	{0x7D8B67C3626CE4FFULL, "w_sundermount_24_0.msh"},
	{0x7D8B6DC3626CEEA1ULL, "w_sundermount_24_0.mmh"},
	{0x7D8B79C3626D0342ULL, "w_sundermount_24_0.mao"},
	{0x7D8B8616D9535657ULL, "hlac_lt_0_lowtown_docks_-131075_3x.dds"},
	{0x7D8BE5C4CEBE90C2ULL, "gen00fl_aveline6053292.cl"},
	{0x7D8BFF45C251D9C8ULL, "mag311_keran6041397.cl"},
	{0x7D8C701782D8AD88ULL, "6205690_m.fxe"},
	{0x7D8CC66A234A806DULL, "sky_lt_darktown.mao"},
	{0x7D8CE25F61C4F0B5ULL, "6047475_m.fxe"},
	{0x7D8D6268953326A1ULL, "hlac_ht_0_hightown_a2_65537_2z.dds"},
	{0x7D8D96566A56BC5DULL, "fen110_anso6095031.cl"},
	{0x7D8DA9F3F92DD9BDULL, "gen00fl_anders6119321.cl"},
	{0x7D8E8CAEFE31F620ULL, "vxca_caverns1_mines_eh9_0_0.vlm"},
	{0x7D8E90A06DAC4531ULL, "6060707_f.fxe"},
	{0x7D8EE5827D38DAE6ULL, "6066428_m.fxe"},
	{0x7D8F40B9B4DDC86DULL, "6029022_m.fxe"},
	{0x7D8F87F64A9634A9ULL, "vxlt_hangedman_27m_0_0.vlm"},
	{0x7D8FC182F38AA16BULL, "lgt110pt_har_escape_rew.ncs"},
	{0x7D8FF59A3A45C6FDULL, "246906493.ogg"},
	{0x7D9026B4E983EC6FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkb.ani"},
	{0x7D90461C7CF972DEULL, "zz_ave_debug6134010.cl"},
	{0x7D90745EB53552B6ULL, "vxht_keep_n_55j_0_0.vlm"},
	{0x7D90D224AA405365ULL, "zz_mer_debug6039343.cl"},
	{0x7D91218454B2E6DEULL, "985538011.ogg"},
	{0x7D9171BE656E40B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_reaction.ani"},
	{0x7D91C6E9D249DC33ULL, "6136939_m.fxe"},
	{0x7D91C77A806DD47DULL, "gen_im_trs_dem_junk1_01.uti"},
	{0x7D9216182ACD29E4ULL, "vxht_0_hightown_a3_c69_0_0.vlm"},
	{0x7D93046400650630ULL, "vxht_0_hightown_n_c6o_0_0.vlm"},
	{0x7D9378876434A054ULL, "gen00fl_fenris6110789.cl"},
	{0x7D93D4B1A7708673ULL, "6075551_m.fxe"},
	{0x7D93DCD08750152BULL, "47823492.ogg"},
	{0x7D940B1D66F1E9BBULL, "vxht_mansion2_2k1_0_0.vlm"},
	{0x7D941DF24E0A5CA7ULL, "6043051_m.fxe"},
	{0x7D948BC80AD22627ULL, "vxlt_gallowscourtyard_4m4_0_0.vlm"},
	{0x7D948E3556D51651ULL, "6079016_m.fxe"},
	{0x7D94F6B1BCCB5F17ULL, "hlac_lt_0_lowtown_docks_a3_n_131072_0y.dds"},
	{0x7D9535C69763D83EULL, "97560232.ogg"},
	{0x7D954B2736538785ULL, "216443270.ogg"},
	{0x7D957FF135F0AAE7ULL, "6095896_m.fxe"},
	{0x7D95B52F5C8D74B7ULL, "vxht_brothel_n_6ga_0_0.vlm"},
	{0x7D95F7BE694009FBULL, "6110971_m.fxe"},
	{0x7D962C670FC25572ULL, "vxht_0_hightown_a2_ap6_0_0.vlm"},
	{0x7D96FFD3C86E989AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8_tool.msh"},
	{0x7D970DD3C86EB0D3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8_tool.mao"},
	{0x7D9711D3C86EB7E0ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8_tool.mmh"},
	{0x7D97F82665ED9C69ULL, "lt_0_lowtown_docks_a3_n_8_0_ll.mmh"},
	{0x7D97F8662EC9A724ULL, "hlac_lt_0_lowtown_a2_n_-131073_3y.dds"},
	{0x7D98B484FFE7BFFDULL, "vxlt_warehouse_12q_0_0.vlm"},
	{0x7D99A6B8757D75B2ULL, "376451660.ogg"},
	{0x7D99A94296775E75ULL, "vxca_caverns2prc_8kb_0_1.vlm"},
	{0x7D99DE6F743DCD16ULL, "357320392.ogg"},
	{0x7D99E887D2FE7A40ULL, "qun230cr_guild_contact6068878.cl"},
	{0x7D99EC889EBB8B0BULL, "lgt310ip_body_4.ncs"},
	{0x7D9A40C42658B404ULL, "mrl150_merrill6070831.cl"},
	{0x7D9A5900E4C2952FULL, "art\\characters\\playercharacter\\humanmale\\hlm_medj.mao"},
	{0x7D9A6C11BE66CA54ULL, "zz_seb_debug6183083.cl"},
	{0x7D9B5B24BCE67D34ULL, "sib100_father_contact6127654.cl"},
	{0x7D9B7260774622A5ULL, "vxw_bonepitmining_a2_a4_0_0.vlm"},
	{0x7D9B85C66572CC75ULL, "gen_im_arm_cht_lgt_act3_03.uti"},
	{0x7D9B9805D2AB540AULL, "vxlt_gallowscourtyard_71j_0_0.vlm"},
	{0x7D9BA875AC84556FULL, "147486186.ogg"},
	{0x7D9C0A913CDE9A4EULL, "407793978.ogg"},
	{0x7D9C8128E29FA976ULL, "gen00fl_varric6118970.cl"},
	{0x7D9CAB8C5F1C940CULL, "vxlt_0_lowtown_docks_n_bcx_0_0.vlm"},
	{0x7D9CF5612B5F8168ULL, "zz_followers_banter6159576.cl"},
	{0x7D9D1967F24C5513ULL, "vxht_0_hightown_a2_c6g_0_0.vlm"},
	{0x7D9D81ECDDF90231ULL, "art\\vfx\\combat\\vfx_sg_app_c.dura.ani"},
	{0x7D9DBD5A10D6CB0CULL, "382749177.ogg"},
	{0x7D9DE1E4999055E5ULL, "6086362_m.fxe"},
	{0x7D9DE419933FD623ULL, "vxht_chantry_2bz_0_0.vlm"},
	{0x7D9DED28F0AE5BCBULL, "song_cloud.mao"},
	{0x7D9E3FD75D4118D4ULL, "6075141_m.fxe"},
	{0x7D9E557D7D981259ULL, "6078055_m.fxe"},
	{0x7D9F1780C8AF413DULL, "11671649.ogg"},
	{0x7D9F3EB8A12D155CULL, "vxlt_0_lowtown_n_e21_0_0.vlm"},
	{0x7D9FE440FA75ED35ULL, "hlac_ca_caverns1_24426_0z.dds"},
	{0x7D9FF5EA34E82AF8ULL, "dre000_sibling_has_blight6036542.cl"},
	{0x7DA01AC3BBC6116FULL, "vxlt_0_lowtown_docks_a3_aqb_0_0.vlm"},
	{0x7DA02D6D30F9A812ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_2y.dds"},
	{0x7DA036DDAE3D49F3ULL, "mcr370_meredith6085904.cl"},
	{0x7DA05FD04BA799AFULL, "vxw_woundedcoast_1hv_0_1.vlm"},
	{0x7DA067E09A68C2A3ULL, "w_bonepitdestroyed_a3_167_0.mao"},
	{0x7DA06BE09A68C9F0ULL, "w_bonepitdestroyed_a3_167_0.mmh"},
	{0x7DA079E09A68E10AULL, "w_bonepitdestroyed_a3_167_0.msh"},
	{0x7DA0C467868E3AB2ULL, "static_static_p16.psh.fxo.dx11.assembly"},
	{0x7DA0C66950D4BFF9ULL, "282186857.ogg"},
	{0x7DA0F1404E7078B2ULL, "vxlt_0_lowtown_docks_a3_n_d1a_0_0.vlm"},
	{0x7DA1369F22154FD1ULL, "zz_followers_banter6157326.cl"},
	{0x7DA14C53A3A66343ULL, "6044869_m.fxe"},
	{0x7DA14D3A5B80548CULL, "mrl150_merrill6094091.cl"},
	{0x7DA15D85150C5001ULL, "llac_lt_0_lowtown_n_-262143_0.dds"},
	{0x7DA167996F9DBAC3ULL, "t1_mul_bk2.tnt"},
	{0x7DA17640F25B1EB0ULL, "vxht_0_hightown_n_c3l_0_0.vlm"},
	{0x7DA1FED5C4AAB04BULL, "dae100_bran6142991.cl"},
	{0x7DA202D965AC451CULL, "121474669.ogg"},
	{0x7DA215486211A34BULL, "vxht_bartrand_n_5nq_0_0.vlm"},
	{0x7DA23AD0BBC55E48ULL, "gen00fl_sebastian6174032.cl"},
	{0x7DA25EF6D0756827ULL, "hlac_lt_0_lowtown_a2_n_c_65535_7x.dds"},
	{0x7DA29E0683796811ULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_p.anb"},
	{0x7DA3123F5B26BB9BULL, "one280_iwan6097159.cl"},
	{0x7DA37CD9077111AAULL, "w_encounter2_38_0_ll.mmh"},
	{0x7DA43224513D543FULL, "gen00fl_anders6111323.cl"},
	{0x7DA45338712AB931ULL, "206356577.ogg"},
	{0x7DA4578731BD8CB6ULL, "gen_im_trs_gen_med_boo_r0.uti"},
	{0x7DA48CECDDFF33D3ULL, "art\\vfx\\combat\\vfx_sg_app_c.dura.gad"},
	{0x7DA4CF381009E0A8ULL, "w_bonepit_11_0.mmh"},
	{0x7DA4DB381009F54BULL, "w_bonepit_11_0.mao"},
	{0x7DA4ED38100A13D2ULL, "w_bonepit_11_0.msh"},
	{0x7DA50CA57BBD81B5ULL, "293817528.ogg"},
	{0x7DA5C97B73EDC215ULL, "and230_mage_girl6135808.cl"},
	{0x7DA6B8D1782F1DC5ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m.mmh"},
	{0x7DA6E17154F27FA5ULL, "hm_amb_keep_servant.mor"},
	{0x7DA7F82F81EB2F49ULL, "103384785.ogg"},
	{0x7DA7FDDC27B4E3A9ULL, "hlac_w_bonepitdestroyed_a3_458758_0z.dds"},
	{0x7DA81189D4D5D5A9ULL, "vxca_caverns1_mines_fn1_0_0.vlm"},
	{0x7DA833DB5AE1C52BULL, "framebuffereffects_screentexrect_p0.psh.fxo"},
	{0x7DA909FC6E6D0D7EULL, "qun221_viscount6070543.cl"},
	{0x7DA9D2CDC594C964ULL, "hlac_lt_0_lowtown_docks_n_-3_6y.dds"},
	{0x7DA9F0FBB148E79BULL, "cst_dots.mao"},
	{0x7DA9F5B0ECBB7E1CULL, "vxlt_0_lowtown_docks_n_c5k_0_0.vlm"},
	{0x7DA9FB029B64AE20ULL, "114628019.ogg"},
	{0x7DAA16080F7CD949ULL, "561558193.ogg"},
	{0x7DAA24A104A1DF08ULL, "art\\characters\\playercharacter\\humanmale\\glv_medd.msh"},
	{0x7DAA36A104A1FD99ULL, "art\\characters\\playercharacter\\humanmale\\glv_medd.mao"},
	{0x7DAA42A104A21232ULL, "art\\characters\\playercharacter\\humanmale\\glv_medd.mmh"},
	{0x7DAA4CA1CDA85CDFULL, "zz_spawn.ncs"},
	{0x7DAA5CB8172B16C9ULL, "vxlt_warehouse_n_14p_0_0.vlm"},
	{0x7DAA80A717EC9D3EULL, "fen210_hadriana6110544.cl"},
	{0x7DAA89ECDE039A38ULL, "art\\vfx\\combat\\vfx_sg_app_c.dura.evt"},
	{0x7DAB14D116593C59ULL, "77802285.ogg"},
	{0x7DAB4ABC67F5D4A7ULL, "hlac_w_blightlands_0_2z.dds"},
	{0x7DAB5490349CFA75ULL, "vxlt_0_lowtown_h2f_0_0.vlm"},
	{0x7DABD21B42A146EFULL, "151870151.ogg"},
	{0x7DAC437A8875A83BULL, "1018182130.ogg"},
	{0x7DAC57F01A26305AULL, "123212469.ogg"},
	{0x7DACA1FC2F62D931ULL, "273572083.ogg"},
	{0x7DACB87B79532D65ULL, "vxht_0_hightown_n_aqi_0_0.vlm"},
	{0x7DACBFED6CE1B403ULL, "hlac_lt_0_lowtown_a2_n_c_65536_4x.dds"},
	{0x7DACE80BB3CEB9D4ULL, "754261326.ogg"},
	{0x7DACEFD9459EB77DULL, "lgt110cr_urn_guard.utc"},
	{0x7DAE4CAE2AF5B0D0ULL, "vxlt_0_lowtown_a2_n_c_hjf_0_0.vlm"},
	{0x7DAED1AD3D9DB22FULL, "vxw_woundedcoast_1sw_0_1.vlm"},
	{0x7DAED58948CE6360ULL, "zz_facefx_test6088487.cl"},
	{0x7DAF686481A15F1BULL, "hlac_w_sundermount_65531_0z.dds"},
	{0x7DAF747201648563ULL, "6057347_f.fxe"},
	{0x7DAFA541D537C10EULL, "vxlt_0_lowtown_a2_byf_0_0.vlm"},
	{0x7DAFBF0683842A8AULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_p.mmh"},
	{0x7DAFD9E99AE39A61ULL, "6001334_m.fxe"},
	{0x7DAFE81DE4513730ULL, "vxht_0_hightown_n_c8m_0_0.vlm"},
	{0x7DAFF60BEAC12C35ULL, "121309470.ogg"},
	{0x7DAFF9F0CD2226F1ULL, "vxlt_gallowstemplar_5aq_0_0.vlm"},
	{0x7DB023AF8EC18C68ULL, "zz_dae_debug6050743.cl"},
	{0x7DB0585F7813926EULL, "gen00fl_carver6102328.cl"},
	{0x7DB05C58E4A863FBULL, "vxlt_0_lowtown_docks_n_cs3_0_1.vlm"},
	{0x7DB05D128ED99417ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_1z.dds"},
	{0x7DB0B3D9EFF61F33ULL, "zz_isa_debug.ncs"},
	{0x7DB0E35B0940E18EULL, "vxlt_gallowscourtyard_7c4_0_0.vlm"},
	{0x7DB0EED74ADE1ABBULL, "vxht_mansion2_2q3_0_0.vlm"},
	{0x7DB11E6F49E1C62DULL, "192665561.ogg"},
	{0x7DB14DC3628D87CAULL, "w_sundermount_24_0.rml"},
	{0x7DB14F8CF8A3059AULL, "224575210.ogg"},
	{0x7DB19306A8209AE8ULL, "69844795.ogg"},
	{0x7DB1B345607EB818ULL, "dae000_keep_bureaucrat_2_st.cnv"},
	{0x7DB1E1790FEC0B07ULL, "6055171_m.fxe"},
	{0x7DB24E3B26BEAEF0ULL, "vxht_0_hightown_a2_8pu_0_0.vlm"},
	{0x7DB33CCBEE348BE1ULL, "mer221_sabin6080561.cl"},
	{0x7DB37C0347DFBC29ULL, "llac_ca_caverns1_24425_1.dds"},
	{0x7DB3CE5D19E8BD59ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_basic.phy"},
	{0x7DB3E9F416B62964ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_lowtown_gate_n.dds"},
	{0x7DB4835EB33DB5CCULL, "ht_0_hightown_a2_43_0_lc.mmh"},
	{0x7DB495456080BB20ULL, "dae000_keep_bureaucrat_2_st.bnk"},
	{0x7DB4B6B604AB3811ULL, "372700383.ogg"},
	{0x7DB4D11381E55B70ULL, "w_bonepit_177_0.mao"},
	{0x7DB4D51381E56183ULL, "w_bonepit_177_0.mmh"},
	{0x7DB4E31381E579DDULL, "w_bonepit_177_0.msh"},
	{0x7DB69FFC05A678DEULL, "item_variations_seb.gda"},
	{0x7DB7043F180174B9ULL, "qun120_arishok_guard6054852.cl"},
	{0x7DB70B0B1F74682BULL, "6049652_f.fxe"},
	{0x7DB73AFF16936CA0ULL, "6102791_m.fxe"},
	{0x7DB7B4D57C54733AULL, "6026187_m.fxe"},
	{0x7DB7BF1E533AA6E0ULL, "hlac_w_bonepitmining_a2_327684_0x.dds"},
	{0x7DB7D97AB8824110ULL, "hlac_ht_0_hightown_a3_-131070_0y.dds"},
	{0x7DB80B78D015CAA9ULL, "isa211tr_night_terrors.ncs"},
	{0x7DB8134841E45A9BULL, "vxht_mansion_n_3wg_0_0.vlm"},
	{0x7DB8561A5B3253A5ULL, "vxw_bonepitmining_a2_k2_0_0.vlm"},
	{0x7DB88B3BAC38531AULL, "6093653_m.fxe"},
	{0x7DB8F1889ED644F6ULL, "lgt310ip_body_4.utp"},
	{0x7DB921EC5DE6A04FULL, "22307719.ogg"},
	{0x7DB92AC6A15C8E99ULL, "vxlt_blackemporium_15p_0_0.vlm"},
	{0x7DB93591256837A8ULL, "85527375.ogg"},
	{0x7DB935DCE0372714ULL, "art\\characters\\creatures\\dragon\\animation\\blendtree\\dragon_pcoverride.anb"},
	{0x7DB9598EE87451C5ULL, "6015682_f.fxe"},
	{0x7DB95AFC2EB67128ULL, "vxca_caverns2_7qj_0_0.vlm"},
	{0x7DB986D1B03095FCULL, "vxca_caverns1_mines_e7h_0_0.vlm"},
	{0x7DB9BF5E52AF1235ULL, "gen_im_npc_dwf_arm_cht_cxt6.uti"},
	{0x7DB9FD21D6388613ULL, "vxht_0_hightown_a2_c8i_0_0.vlm"},
	{0x7DBAA3D7505693FEULL, "vxlt_gallowstemplar_f_5rt_0_0.vlm"},
	{0x7DBAC988C267B48AULL, "385568669.ogg"},
	{0x7DBAEE26013738CFULL, "zz_fol_debug6096578.cl"},
	{0x7DBAEF9E3B503AA5ULL, "hlac_lt_warehouse_0_0x.dds"},
	{0x7DBB73F022DB8ECAULL, "vxlt_gallowstemplar_f_3sq_0_0.vlm"},
	{0x7DBBD2774E8BB842ULL, "vxw_bonepit_y1_0_0.vlm"},
	{0x7DBD6B7E27853620ULL, "6090210_m.fxe"},
	{0x7DBDB5E0C0A670E6ULL, "vxca_caverns1_mines_edk_0_0.vlm"},
	{0x7DBDCB90C05DFA4BULL, "gen00fl_varric6078006.cl"},
	{0x7DBE6BCDBF183F71ULL, "vxlt_gallowstemplar_5jp_0_0.vlm"},
	{0x7DBEBF15EBFF8C8DULL, "397617488.ogg"},
	{0x7DBF3C5B8DFA2913ULL, "63720829.ogg"},
	{0x7DBF3DD2A2D78010ULL, "gen00fl_aveline6115408.cl"},
	{0x7DBF5537FB36FA0EULL, "vxlt_gallowscourtyard_7d5_0_0.vlm"},
	{0x7DBF60B2E0ECD7C0ULL, "isa311_castillon6092122.cl"},
	{0x7DBF60B43CD4333BULL, "vxht_mansion2_2z2_0_0.vlm"},
	{0x7DBFFA88303F0674ULL, "vxlt_0_lowtown_a2_n_c_er3_0_0.vlm"},
	{0x7DC0327C86F098E8ULL, "mer141cr_drake_boss.utc"},
	{0x7DC08B67B029AAB0ULL, "6165733_f.fxe"},
	{0x7DC08FD2FDA328E6ULL, "vxdr_primevalentrance_239_0_0.vlm"},
	{0x7DC0A309EB914D30ULL, "127463494.ogg"},
	{0x7DC1208E67270444ULL, "6202668_m.fxe"},
	{0x7DC1E7E71E9C63BAULL, "vxht_0_hightown_bgp_0_0.vlm"},
	{0x7DC1EA96E7617CAAULL, "vxlt_undercityencounter_1ja_0_0.vlm"},
	{0x7DC20DDC3DD1A55DULL, "vxlt_0_lowtown_n_gc1_0_0.vlm"},
	{0x7DC25BD53F1D0FCCULL, "vxht_brothel_n_4w0_0_0.vlm"},
	{0x7DC25EA968C51B82ULL, "73512284.ogg"},
	{0x7DC26666A885ECEEULL, "w_blightlands_241_0.mmh"},
	{0x7DC27266A8860085ULL, "w_blightlands_241_0.mao"},
	{0x7DC28066A88618DCULL, "w_blightlands_241_0.msh"},
	{0x7DC371D69E32872BULL, "6082157_m.fxe"},
	{0x7DC396197EB001C0ULL, "llac_w_blightlands_-2_0.dds"},
	{0x7DC563FCD3C56B4EULL, "71810728.ogg"},
	{0x7DC56E56F98D239DULL, "llac_w_encounter2_131073_0.dds"},
	{0x7DC62FFBDEE5B4A4ULL, "237402253.ogg"},
	{0x7DC64DFFCC27B119ULL, "zz_dae_debug6029304.cl"},
	{0x7DC74442913F9FC7ULL, "gen00fl_dog6117184.cl"},
	{0x7DC7481EDB385286ULL, "117886795.ogg"},
	{0x7DC8816C99394905ULL, "370393990.ogg"},
	{0x7DC8819FE513432AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killervase_n.dds"},
	{0x7DC8F478D024168CULL, "isa211tr_night_terrors.utt"},
	{0x7DC913005461AED3ULL, "6018532_m.fxe"},
	{0x7DC982A0DEEA8D0AULL, "p_armoire_01n.dds"},
	{0x7DCA4E28531D742DULL, "6076253-brotherunderattack.cut"},
	{0x7DCADD03C7B3F758ULL, "vxht_0_hightown_n_9ah_0_0.vlm"},
	{0x7DCD4DD8D3457450ULL, "6126919_m.fxe"},
	{0x7DCD83E897DE193BULL, "follower_banter6151311.cl"},
	{0x7DCDC714ED2D128EULL, "vxlt_gallowscourtyard_7i6_0_0.vlm"},
	{0x7DCDDDE429B68AD9ULL, "6166785_m.fxe"},
	{0x7DCE007C86D42080ULL, "6187828_m.fxe"},
	{0x7DCE23CD5E9ADE66ULL, "344013485.ogg"},
	{0x7DCEBE395B82F710ULL, "6164272_m.fxe"},
	{0x7DCF39C6D23522F4ULL, "hlac_lt_0_lowtown_a2_1_0z.dds"},
	{0x7DCF3B2CB4C97B63ULL, "6167638_m.fxe"},
	{0x7DCF7186F6BE7EEDULL, "vxht_0_hightown_a3_atp_0_0.vlm"},
	{0x7DCFC8CF53D03382ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho04.msh"},
	{0x7DCFCACF53D03738ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho04.mmh"},
	{0x7DCFCCEF264A4504ULL, "vxw_encounter2_camp_jb_0_0.vlm"},
	{0x7DCFFB9258A40F21ULL, "250731526.ogg"},
	{0x7DD05C73D957952AULL, "vxlt_undercityencounter_1gb_0_0.vlm"},
	{0x7DD0A0ABFEE46EABULL, "54697534.ogg"},
	{0x7DD0DFC57F18C2C4ULL, "dae000_mother6136952.cl"},
	{0x7DD119A4A62FC2DBULL, "vxlt_0_lowtown_docks_a3_n_com_0_0.vlm"},
	{0x7DD127AE351E0FEDULL, "vxht_0_hightown_cah_0_0.vlm"},
	{0x7DD1DA28F0F1906BULL, "mag100_karras6034731.cl"},
	{0x7DD1FF11FB6DE426ULL, "6110127_m.fxe"},
	{0x7DD22F57C16FE65FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww21_galocelltop.msh"},
	{0x7DD23557C16FF001ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww21_galocelltop.mmh"},
	{0x7DD24157C17004A2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww21_galocelltop.mao"},
	{0x7DD2E5739B5230A4ULL, "lt_0_lowtown_docks_240_0.mmh"},
	{0x7DD2E9739B5237F7ULL, "lt_0_lowtown_docks_240_0.mao"},
	{0x7DD2FB739B525646ULL, "lt_0_lowtown_docks_240_0.msh"},
	{0x7DD460D29F4BD54FULL, "mcr310_argument6074387.cl"},
	{0x7DD47330E713D7D5ULL, "lgt260_eau_tmp_ending.bnk"},
	{0x7DD4800B1F5CCB06ULL, "vxlt_gallowstemplar_f_5pf_0_0.vlm"},
	{0x7DD49DBEC5DF00BEULL, "6161633_m.fxe"},
	{0x7DD50A1FB2229C1FULL, "901234727.ogg"},
	{0x7DD543600B58DBE7ULL, "dae000_arishok6070888.cl"},
	{0x7DD5C36E383F150DULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb103a_0n.dds"},
	{0x7DD5DED9F01584F8ULL, "zz_isa_debug.cnv"},
	{0x7DD6052294896805ULL, "399576316.ogg"},
	{0x7DD612E3D8254FBDULL, "art\\vfx\\combat\\vfx_gib_hlk_p_3.mao"},
	{0x7DD64D3581098DA4ULL, "879420118.ogg"},
	{0x7DD6823F06B0728BULL, "vxw_bonepitdestroyed_a3_1o1_0_0.vlm"},
	{0x7DD68BC5108E893EULL, "hlac_ca_caverns2_5802_25x.dds"},
	{0x7DD74C6E5DD3B4FEULL, "6036528_m.fxe"},
	{0x7DD7662AF87410F0ULL, "hlac_ht_0_hightown_n_-196609_0z.dds"},
	{0x7DD774E3A39C4279ULL, "6151453_m.fxe"},
	{0x7DD79530E716472DULL, "lgt260_eau_tmp_ending.cnv"},
	{0x7DD7C4EBE4ABB969ULL, "dae380st_temmerin.stg"},
	{0x7DD7DF6ADAA9D93BULL, "lt_0_lowtown_docks_a3_n_96_0.msh"},
	{0x7DD7F16ADAA9F786ULL, "lt_0_lowtown_docks_a3_n_96_0.mao"},
	{0x7DD7FD6ADAAA0C25ULL, "lt_0_lowtown_docks_a3_n_96_0.mmh"},
	{0x7DD82EDE70309414ULL, "426068851.ogg"},
	{0x7DD8337D37D839FDULL, "21207674.ogg"},
	{0x7DD84FE7103547C0ULL, "hlac_ht_mansion_n_3368_0z.dds"},
	{0x7DD8DE795F433AEFULL, "74180058.ogg"},
	{0x7DD8E8E423C753DCULL, "vxht_keep_n_7ab_0_0.vlm"},
	{0x7DD9D1CF53D897C3ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho04.phy"},
	{0x7DD9D765AEA81929ULL, "548056717.ogg"},
	{0x7DDA4F9753B5382DULL, "6031650_m.fxe"},
	{0x7DDAB4871C8108F3ULL, "6126376_m.fxe"},
	{0x7DDABB77671901A3ULL, "6085974_m.fxe"},
	{0x7DDB510A53E0FBDCULL, "6096641_m.fxe"},
	{0x7DDC0D693C6631DAULL, "t3_wdg_tn2.tnt"},
	{0x7DDC1D7FBD77724FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_extreme_v1.gad"},
	{0x7DDC5CA151DDA254ULL, "151261471.ogg"},
	{0x7DDC6236A30BB91FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_thin.phy"},
	{0x7DDD087D85D3B1F1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mc.mmh"},
	{0x7DDD147D85D3C592ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mc.mao"},
	{0x7DDD227D85D3DDAFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mc.msh"},
	{0x7DDD894018E86B4FULL, "one350tr_spiders.utt"},
	{0x7DDDA17FE61A3637ULL, "hlac_lt_0_lowtown_docks_a3_-262144_1y.dds"},
	{0x7DDDD5C0E81011A1ULL, "gen_im_cft_rec_bmb_01.uti"},
	{0x7DDE3ECC18405D84ULL, "vxw_encounter2_camp_ec_0_0.vlm"},
	{0x7DDE42B3670A785CULL, "vxca_caverns1_mines_gkk_0_1.vlm"},
	{0x7DDE4A8BAE620759ULL, "vxlt_0_lowtown_a2_gyl_0_0.vlm"},
	{0x7DDEAA4B2456FBF0ULL, "718190091.ogg"},
	{0x7DDECE352AB8DAC8ULL, "552861452.ogg"},
	{0x7DDFF26F90DFD042ULL, "vxlt_0_lowtown_n_f15_0_0.vlm"},
	{0x7DE0482519C74F11ULL, "358369070.ogg"},
	{0x7DE0693735BBF6C3ULL, "vxht_0_hightown_a3_a4n_0_0.vlm"},
	{0x7DE0CCB4264439BBULL, "6119657_m.fxe"},
	{0x7DE155D1D478445CULL, "pro000cs_avaline_intro.cut"},
	{0x7DE172E09AA06C0BULL, "w_bonepitdestroyed_a3_167_0.phy"},
	{0x7DE19163425497E8ULL, "cd_qcr221_221_cut_qunari_defeated.wbd"},
	{0x7DE29FA64734DDC8ULL, "658742473.ogg"},
	{0x7DE2D4A59F668335ULL, "305375042.ogg"},
	{0x7DE305E60A1690F2ULL, "6135679_f.fxe"},
	{0x7DE307111A051E8DULL, "6120875_m.fxe"},
	{0x7DE3327FBD7DB4FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_extreme_v1.evt"},
	{0x7DE34F10446645ABULL, "hlac_ca_cavern1_haunted_24424_21x.dds"},
	{0x7DE35502EB473944ULL, "lgt260_eau_tmp_asst_harbor.cnv"},
	{0x7DE3A845A376F89DULL, "dae000_keep_maid_sta6162409.cl"},
	{0x7DE446A6679A8466ULL, "vxht_0_hightown_a2_byn_0_0.vlm"},
	{0x7DE44CC5C6EAC498ULL, "hlac_ca_caverns1_mines_24427_5x.dds"},
	{0x7DE46E5999CFCC2CULL, "190565116.ogg"},
	{0x7DE5F96E2638DD7EULL, "vxlt_gallowstemplar_f_5cw_0_0.vlm"},
	{0x7DE64601C5F6580CULL, "hlac_lt_0_lowtown_docks_n_-65538_8y.dds"},
	{0x7DE694489047E1FCULL, "w_bonepit_52_0.rml"},
	{0x7DE6D3366C0D49DAULL, "390045683.ogg"},
	{0x7DE75C17D8842C33ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masf.phy"},
	{0x7DE77702EB4B5BA4ULL, "lgt260_eau_tmp_asst_harbor.bnk"},
	{0x7DE77A22E205DC3AULL, "vxlt_gallowstemplar_5bo_0_0.vlm"},
	{0x7DE7A470082E4E33ULL, "hlac_lt_0_lowtown_196607_1z.dds"},
	{0x7DE7AAAEC2B9EDF6ULL, "dae000_mother6103068.cl"},
	{0x7DE7CD9D43B52E76ULL, "static_static_p18.psh.fxo.dx11"},
	{0x7DE81421F2CE12D0ULL, "vxlt_0_lowtown_a2_n_c_hvb_0_0.vlm"},
	{0x7DE89068E850D05DULL, "666166372.ogg"},
	{0x7DE96E4614B59D0AULL, "723151669.ogg"},
	{0x7DE9B12DB128C81FULL, "lt_0_lowtown_a2_n_43_0_lc.mmh"},
	{0x7DE9FF48F2FFC5EBULL, "6105999_m.fxe"},
	{0x7DEA1068CF0B286AULL, "hlac_w_sundermount_0_6y.dds"},
	{0x7DEA5EDF48596509ULL, "llac_w_woundedcoast_131072_0.dds"},
	{0x7DEA74350EB13DA7ULL, "hlac_lt_0_lowtown_a3_n_c_196606_0x.dds"},
	{0x7DEA7A48904B9E51ULL, "w_bonepit_52_0.msh"},
	{0x7DEA8848904BB66CULL, "w_bonepit_52_0.mao"},
	{0x7DEA9448904BCA0FULL, "w_bonepit_52_0.mmh"},
	{0x7DEB5722CD797D05ULL, "zz_isa_debug6133165.cl"},
	{0x7DEB6ED54AC164ABULL, "6051045_m.fxe"},
	{0x7DEB767A2B45782DULL, "vxlt_gallowstemplar_n_5iv_0_0.vlm"},
	{0x7DEC4B96337C59F2ULL, "lt_0_lowtown_docks_a3_n_33_0_lc.mmh"},
	{0x7DED3CE8BC9A579EULL, "vxht_brothel_n_4dk_0_0.vlm"},
	{0x7DED6336A31A2E24ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_thin.mmh"},
	{0x7DED6736A31A3577ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_thin.mao"},
	{0x7DED7936A31A53C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_thin.msh"},
	{0x7DED984E70EC92EAULL, "vxlt_0_lowtown_f23_0_0.vlm"},
	{0x7DEDB5104148EB72ULL, "vxw_blightlands_6w_0_0.vlm"},
	{0x7DEDFD5E8A1BF3DBULL, "vxlt_0_lowtown_docks_a3_n_cuo_0_0.vlm"},
	{0x7DEE8EEE4B3E843BULL, "6064037_f.fxe"},
	{0x7DEE91D1786C9172ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m.phy"},
	{0x7DEF1CE2973CDF10ULL, "hlac_lt_backalley_g_0_3y.dds"},
	{0x7DEF81153A28A9C3ULL, "gen_im_trs_gem_diamond_03.uti"},
	{0x7DEF86326A802626ULL, "hlac_w_blightlands_327678_0x.dds"},
	{0x7DEF99AB6CFCDF91ULL, "563817355.ogg"},
	{0x7DEFACA82B492442ULL, "vxca_caverns2prc_9si_0_0.vlm"},
	{0x7DF03A7FBD884C15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_lightning_extreme_v1.ani"},
	{0x7DF043CAF0BC6BFEULL, "384170856.ogg"},
	{0x7DF1BDCA350557F5ULL, "atl_partypicke_dxt5_00.dds"},
	{0x7DF1D33F8EC70CBBULL, "571279480.ogg"},
	{0x7DF28573BCD273FDULL, "vxht_bartrand_n_73n_0_0.vlm"},
	{0x7DF38C528ADB3593ULL, "one300_murder_followers6086732.cl"},
	{0x7DF3C349D3F15DD0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_rob_chaa_0d.dds"},
	{0x7DF42CF88AAAB944ULL, "art\\characters\\creatures\\spider\\c_giaspidr.msh"},
	{0x7DF43EF88AAAD7ADULL, "art\\characters\\creatures\\spider\\c_giaspidr.mao"},
	{0x7DF442F88AAADEE6ULL, "art\\characters\\creatures\\spider\\c_giaspidr.mmh"},
	{0x7DF4B6956F606689ULL, "6195839_m.fxe"},
	{0x7DF4C928DE8AA80EULL, "vxw_bonepitdestroyed_a3_1p1_0_0.vlm"},
	{0x7DF522819512A617ULL, "vxlt_0_lowtown_a2_n_gm1_0_0.vlm"},
	{0x7DF5732E83E8C03DULL, "vxht_keep_n_64a_0_0.vlm"},
	{0x7DF5B37ADE04D1FAULL, "635815442.ogg"},
	{0x7DF69690096E018AULL, "vxht_mansion2_2a6_0_0.vlm"},
	{0x7DF6C229678B9469ULL, "249512200.ogg"},
	{0x7DF6D3EF4B3310D7ULL, "576226564.ogg"},
	{0x7DF6D4249C3D3D96ULL, "hlac_lt_darktown_-1_75z.dds"},
	{0x7DF6FAF88A4C8BD3ULL, "6129815_f.fxe"},
	{0x7DF73472369E30CEULL, "hlac_dr_primevaltomb_659_2z.dds"},
	{0x7DF7846F03DB77A0ULL, "dirty.mao"},
	{0x7DF7C3290D14B90FULL, "537323091.ogg"},
	{0x7DF86F8E7784C299ULL, "vxlt_house_j3_0_0.vlm"},
	{0x7DF8EECCF2B24AB6ULL, "ave200_aveline_bar6116410.cl"},
	{0x7DF901B72900E23EULL, "lt_warehouse_n_4_0_lc.mmh"},
	{0x7DF90EE7D4EC9A27ULL, "and111_lirene6084757.cl"},
	{0x7DF93038A7D14284ULL, "109122623.ogg"},
	{0x7DF94BDE787AF420ULL, "6082226_m.fxe"},
	{0x7DF95ED719A9FAE3ULL, "vxlt_house_fu_0_0.vlm"},
	{0x7DF984FE9AF25498ULL, "vxht_mansion2b_2jm_0_0.vlm"},
	{0x7DF9E8571D3B90ADULL, "vxlt_gallowstemplar_n_5lu_0_0.vlm"},
	{0x7DFA0C4CEB716AE8ULL, "hlac_dr_0_deeproads_5845_4x.dds"},
	{0x7DFAB7FEDD8EAD84ULL, "w_bonepitdestroyed_a3_190_0.msh"},
	{0x7DFAC9FEDD8ECCEDULL, "w_bonepitdestroyed_a3_190_0.mao"},
	{0x7DFACDFEDD8ED326ULL, "w_bonepitdestroyed_a3_190_0.mmh"},
	{0x7DFAF3F01D254D11ULL, "art\\characters\\creatures\\corpse\\animation\\blendtree\\corpse_overlay.anb"},
	{0x7DFB52F3CB415501ULL, "vxht_keep_n_7a3_0_0.vlm"},
	{0x7DFB5FED13D29B61ULL, "lgt110_arl_bann_teagan6115561.cl"},
	{0x7DFB92AB344C76A4ULL, "114809711.ogg"},
	{0x7DFC4C2DE4945F1EULL, "plt_cod_cha_carver.plo"},
	{0x7DFC538787B11ADEULL, "gen_im_trs_bst_spd_fang_02.uti"},
	{0x7DFCFA7991F4FF89ULL, "6127571_m.fxe"},
	{0x7DFD3F73F0E528C8ULL, "llac_ht_0_hightown_a3_65537_0.dds"},
	{0x7DFDC5A391307786ULL, "851080682.ogg"},
	{0x7DFE3657C1952D7EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww21_galocelltop.phy"},
	{0x7DFE39F88AB323B1ULL, "art\\characters\\creatures\\spider\\c_giaspidr.phy"},
	{0x7DFE42CE666A0FB3ULL, "6172830_m.fxe"},
	{0x7DFF5E557A153B4FULL, "hlac_lt_0_lowtown_a3_n_c_65536_2y.dds"},
	{0x7DFF68A0E64D2D9AULL, "gen00fl_aveline6139880.cl"},
	{0x7DFF88278B64D4DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_enter.ani"},
	{0x7DFF9FE003E3CB5DULL, "1435916.ogg"},
	{0x7E00158A403BB349ULL, "rdr111dg_martin6054514.cl"},
	{0x7E007BCAF8030C7EULL, "hlac_w_bonepitmining_a2_196610_0y.dds"},
	{0x7E007DB030483CCAULL, "6087821_m.fxe"},
	{0x7E012A604B86B566ULL, "vxht_0_hightown_a2_bgp_0_0.vlm"},
	{0x7E01BC232AB652D6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\blendtree\\trooper_movement.anb"},
	{0x7E01E77A15B94068ULL, "qcr220_grey_wardens6060423.cl"},
	{0x7E025F90EE371DF3ULL, "and230_mage_girl6109113.cl"},
	{0x7E02A09E95A3AA4CULL, "6070168_m.fxe"},
	{0x7E031E45AB3C3B2CULL, "w_sundermount_43_0_lc.mmh"},
	{0x7E033B05D080C08EULL, "vxw_bonepitdestroyed_a3_1u2_0_0.vlm"},
	{0x7E037AB94C7B6012ULL, "vxht_playershouse_n_1w1_0_0.vlm"},
	{0x7E03945E8708BE97ULL, "vxlt_0_lowtown_a2_n_gj0_0_0.vlm"},
	{0x7E039CA75313C46DULL, "art\\levels\\races\\vfx_models\\prp_coalsmelt.mmh"},
	{0x7E03A0A75313CABEULL, "art\\levels\\races\\vfx_models\\prp_coalsmelt.mao"},
	{0x7E03AEA75313E2F3ULL, "art\\levels\\races\\vfx_models\\prp_coalsmelt.msh"},
	{0x7E03FE833CA93DA4ULL, "307599449.ogg"},
	{0x7E0416A531BA2AD2ULL, "vxdr_primevalentrance_1ap_0_0.vlm"},
	{0x7E04233163DC90E3ULL, "gen00fl_aveline6114820.cl"},
	{0x7E0442981E45BD6FULL, "follower_banter6155751.cl"},
	{0x7E0444453C0F54C2ULL, "675589057.ogg"},
	{0x7E0446D6FC077E32ULL, "6043249_m.fxe"},
	{0x7E045DDCC5F20D3AULL, "vxlt_gallowstemplar_5hi_0_0.vlm"},
	{0x7E047949F565C1A4ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_dwpa.msh"},
	{0x7E048B49F565E00DULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_dwpa.mao"},
	{0x7E048D43EF737F32ULL, "vxht_keep_634_0_0.vlm"},
	{0x7E048F49F565E746ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_dwpa.mmh"},
	{0x7E04983B56B8C1DEULL, "6092467_m.fxe"},
	{0x7E04D329F77A79F6ULL, "hlac_ca_caverns2prc_5802_39x.dds"},
	{0x7E05086CFB641A0AULL, "vxht_mansion2_2j7_0_0.vlm"},
	{0x7E062DB6F25C9B3FULL, "llac_w_bonepitmining_a2_0_1.dds"},
	{0x7E0641A104F06F2DULL, "art\\characters\\playercharacter\\humanmale\\glv_medd.phy"},
	{0x7E074774BDC7998FULL, "vxw_sundermount_37m_0_1.vlm"},
	{0x7E077AC97F69EC69ULL, "ht_0_hightown_30_0_ll.mmh"},
	{0x7E07808B233D3EDFULL, "509500530.ogg"},
	{0x7E082234CEF878DCULL, "6131212_m.fxe"},
	{0x7E088E7720881CF6ULL, "gen00fl_dog6117770.cl"},
	{0x7E089DD5B51DC0FAULL, "vxlt_warehouse_i2_0_0.vlm"},
	{0x7E0A097F36679F8AULL, "vxlt_shop_5y_0_0.vlm"},
	{0x7E0A7530AB24BDCDULL, "77383179.ogg"},
	{0x7E0AD4553BEE8E2AULL, "hlac_lt_0_lowtown_docks_a3_n_-131071_0z.dds"},
	{0x7E0B1B35E48CA339ULL, "685750327.ogg"},
	{0x7E0B723A7104D22BULL, "395630788.ogg"},
	{0x7E0BBC49EAFAB1FFULL, "vxht_0_hightown_a2_asn_0_0.vlm"},
	{0x7E0BC760573741C1ULL, "lt_0_lowtown_docks_11_0_ll.mmh"},
	{0x7E0BC7FC562BDB71ULL, "vxdr_primevalentrance_184_0_0.vlm"},
	{0x7E0C036C856FD6DAULL, "vxht_keep_n_5k8_0_0.vlm"},
	{0x7E0C69FAEF672B23ULL, "6179260_m.fxe"},
	{0x7E0C8A6F4F7C1D68ULL, "gen_im_cns_pot_avernus.uti"},
	{0x7E0CE3AAC06D1A7AULL, "473271653.ogg"},
	{0x7E0D312E56CC7CD6ULL, "1049781439.ogg"},
	{0x7E0D759D3E415629ULL, "vxlt_foundry_n_10b_0_0.vlm"},
	{0x7E0D90278B711FD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_enter.evt"},
	{0x7E0DA0B3C5798661ULL, "gen00fl_fenris6115312.cl"},
	{0x7E0DEDEAAFA19844ULL, "6050334_f.fxe"},
	{0x7E0E4CF062DFC0ABULL, "651337884.ogg"},
	{0x7E0E5C1F94EA0E4BULL, "vxht_mansion2b_1v1_0_0.vlm"},
	{0x7E0E7CD99F5329D9ULL, "204803129.ogg"},
	{0x7E0F2D64DA939B83ULL, "spout.mao"},
	{0x7E0F8649F56FDF11ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_dwpa.phy"},
	{0x7E0F91AD1ACB87D6ULL, "art\\characters\\playercharacter\\humanmale\\arm_vael.phy"},
	{0x7E0FCFDD6DE9DE50ULL, "gen00fl_merrill6127255.cl"},
	{0x7E100A8E7D08831AULL, "hlac_ht_keep_n_3731_2z.dds"},
	{0x7E10905179A7B542ULL, "6111304_m.fxe"},
	{0x7E10B81DAFF6F32EULL, "562565782.ogg"},
	{0x7E119749D235C326ULL, "vxca_caverns1_mines_gk0_0_0.vlm"},
	{0x7E11D764467044D8ULL, "180430459.ogg"},
	{0x7E12AB0BE2116098ULL, "gen_im_arm_cht_hvy_act3_04.uti"},
	{0x7E134174CF02CE00ULL, "dae000_bodhan6129106.cl"},
	{0x7E13485A08B47D4EULL, "hlac_dr_dungeon_3214_4x.dds"},
	{0x7E134FA615E3DB44ULL, "zz_deathblow.ncs"},
	{0x7E13DD18AC4D7F40ULL, "lgt110_gxa_vgk_silver_order6117009.cl"},
	{0x7E141328D855D59FULL, "follower_banter6143367.cl"},
	{0x7E1453F7E21B7274ULL, "6116797_f.fxe"},
	{0x7E148C5D9F1EBAB3ULL, "vxca_caverns2prc_1xa_0_0.vlm"},
	{0x7E14AF278B777365ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_on_face_enter.gad"},
	{0x7E14C238A8E7E45FULL, "vxht_0_hightown_a2_7fw_0_0.vlm"},
	{0x7E14E6741AED594AULL, "vxht_0_hightown_a2_a86_0_0.vlm"},
	{0x7E14EAE283079F88ULL, "46219611.ogg"},
	{0x7E1509EAD7DFAB7BULL, "vxlt_gallowscourtyard_n_4we_0_0.vlm"},
	{0x7E1559021AF86BE4ULL, "mer221_sabin6047219.cl"},
	{0x7E1572D56F3AD59CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_appr.ani"},
	{0x7E159375C9EE4ADDULL, "523402061.ogg"},
	{0x7E16291F5B0A051DULL, "21160117.ogg"},
	{0x7E166DACDD953172ULL, "6196683_f.fxe"},
	{0x7E170C025A6977DDULL, "qcr220_orsino6061799.cl"},
	{0x7E170FB2A713D97AULL, "vxlt_warehouse_r3_0_0.vlm"},
	{0x7E181B5ACFDE7EF1ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerb.msh"},
	{0x7E18295ACFDE960CULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerb.mao"},
	{0x7E18355ACFDEAAAFULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerb.mmh"},
	{0x7E189314CA60F6BAULL, "vxht_0_hightown_byn_0_0.vlm"},
	{0x7E18A71B6E75C800ULL, "161051343.ogg"},
	{0x7E19924B8D7ED2B2ULL, "384516650.ogg"},
	{0x7E19B59D877F96FEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\leftarm_gesture_behind.ani"},
	{0x7E19FBA9C4428BEEULL, "vxlt_0_lowtown_docks_bsi_0_0.vlm"},
	{0x7E1A00D249C60675ULL, "root_clu_black.mao"},
	{0x7E1A2E26DCF0CA7FULL, "vxht_0_hightown_a2_ani_0_0.vlm"},
	{0x7E1A7AAD1AD56877ULL, "art\\characters\\playercharacter\\humanmale\\arm_vael.msh"},
	{0x7E1A8CAD1AD586DAULL, "art\\characters\\playercharacter\\humanmale\\arm_vael.mao"},
	{0x7E1A90AD1AD58DE9ULL, "art\\characters\\playercharacter\\humanmale\\arm_vael.mmh"},
	{0x7E1ACE5BDD30EDB7ULL, "vxw_blightlands_br_0_0.vlm"},
	{0x7E1B2EE3F67114ABULL, "429412083.ogg"},
	{0x7E1B3BB41AF71D5BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_basic.msh"},
	{0x7E1B4DB41AF73BA6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_basic.mao"},
	{0x7E1B59B41AF75045ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_basic.mmh"},
	{0x7E1C1BE9A510CA9FULL, "llac_ht_0_hightown_a2_-65536_2.dds"},
	{0x7E1C4F01280BBF9BULL, "art\\levels\\races\\natural\\highlands\\props\\prp_sundermt.msh"},
	{0x7E1C6101280BDEE6ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_sundermt.mao"},
	{0x7E1C6D01280BF285ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_sundermt.mmh"},
	{0x7E1C7FD56F410ADEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_appr.gad"},
	{0x7E1CB95BD92B4586ULL, "vxlt_gallowstemplar_f_5ok_0_0.vlm"},
	{0x7E1CD66220DADD2FULL, "plt_cod_crt_abomination.plo"},
	{0x7E1D17460FB5233FULL, "llac_lt_0_lowtown_docks_n_131070_2.dds"},
	{0x7E1D79CAED2E29B6ULL, "532605830.ogg"},
	{0x7E1E2589E32F9049ULL, "vxlt_backalley_n_1i6_0_0.vlm"},
	{0x7E1EEE0247F67A2BULL, "274680895.ogg"},
	{0x7E1F0616605B55EAULL, "hlac_w_bonepit_196610_0y.dds"},
	{0x7E1F1A2C73F33193ULL, "vxht_0_hightown_a2_cen_0_0.vlm"},
	{0x7E1F444A191DEDBDULL, "llac_lt_0_lowtown_n_0_1.dds"},
	{0x7E201EBFB46CF18EULL, "vxw_bonepitdestroyed_a3_1c4_0_0.vlm"},
	{0x7E205759CCB833EFULL, "57708040.ogg"},
	{0x7E205F3271DE7A35ULL, "vxlt_0_lowtown_a2_n_eoo_0_0.vlm"},
	{0x7E20B59FB19C60DBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_pad_6.pwk"},
	{0x7E20B70CF89C0544ULL, "gen00fl_sebastian6188615.cl"},
	{0x7E2162C76B156BE6ULL, "6077952_m.fxe"},
	{0x7E216A9AE1FCEFFDULL, "43200727.ogg"},
	{0x7E2182B8F073EE88ULL, "mag210_arianni6021219.cl"},
	{0x7E219E04D9DAF8D4ULL, "vxlt_0_lowtown_a2_n_c_i09_0_0.vlm"},
	{0x7E21EC26DF504B0AULL, "vxht_mansion2_2p1_0_0.vlm"},
	{0x7E22612819BBB78DULL, "hlac_lt_0_lowtown_docks_a3_-262144_0x.dds"},
	{0x7E22DFCE18B00337ULL, "gen00fl_varric6076026.cl"},
	{0x7E22EF75546747E9ULL, "vxca_caverns2prc_9xk_0_0.vlm"},
	{0x7E22F0E2EC0BC073ULL, "art\\vfx\\combat\\vfx_gib_corm_2.mao"},
	{0x7E231392400275D2ULL, "563315917.ogg"},
	{0x7E233281DB43E143ULL, "891164530.ogg"},
	{0x7E235AD56F46EAFDULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_appr.evt"},
	{0x7E235C1D31AFA09AULL, "vxlt_gallowstemplar_34k_0_0.vlm"},
	{0x7E23667B8B910834ULL, "vxw_bonepit_23y_0_0.vlm"},
	{0x7E2389C06487A291ULL, "llac_w_sundermount_-131068_0.dds"},
	{0x7E238ABB2C9E615DULL, "gen_im_wep_war_one_act2_06.uti"},
	{0x7E23973A73192DE0ULL, "fxe_water_tiny.dds"},
	{0x7E23B7284FCC0A4CULL, "6169518_m.fxe"},
	{0x7E23E343032686D9ULL, "w_blightlands_110_0.msh"},
	{0x7E23ED43032697C7ULL, "w_blightlands_110_0.mmh"},
	{0x7E23F14303269E14ULL, "w_blightlands_110_0.mao"},
	{0x7E24BB800B1CF108ULL, "hlac_lt_backalley_g_1_3z.dds"},
	{0x7E2506A461E106ABULL, "llac_lt_0_lowtown_a2_n_-3_1.dds"},
	{0x7E2519CEB39DD908ULL, "vxht_0_hightown_a3_bzd_0_0.vlm"},
	{0x7E25699EA384DE8EULL, "dae000_meredith6092578.cl"},
	{0x7E25DE1BEC8327CBULL, "664289389.ogg"},
	{0x7E261CEBFE97F532ULL, "15015509.ogg"},
	{0x7E262D76B001F015ULL, "llac_lt_darktown_-1_25.dds"},
	{0x7E265189A3E21D4CULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\mov_p.gad"},
	{0x7E26EB974276039DULL, "730594566.ogg"},
	{0x7E27045C8472B99EULL, "vxht_brothel_n_4xo_0_0.vlm"},
	{0x7E2719D6387ABD48ULL, "w_bonepitmining_a2_192_0.msh"},
	{0x7E272BD6387ADBD9ULL, "w_bonepitmining_a2_192_0.mao"},
	{0x7E2737D6387AF072ULL, "w_bonepitmining_a2_192_0.mmh"},
	{0x7E27414300BB8709ULL, "zz_vault_debug6200334.cl"},
	{0x7E27C946E088B7BEULL, "qcr000_arishok6166999.cl"},
	{0x7E281066292C3854ULL, "419586933.ogg"},
	{0x7E2813A055CA36A1ULL, "mag311_nyssa6051208.cl"},
	{0x7E282566132605B0ULL, "war_silence.dds"},
	{0x7E287C16EBD979BFULL, "w_bonepitmining_a2_39_0.rml"},
	{0x7E28B068CD8AF2ECULL, "w_sundermount_211_0.phy"},
	{0x7E28B7634290DAAAULL, "cd_qcr221_221_cut_qunari_defeated.bnk"},
	{0x7E28BEB19923502AULL, "245250367.ogg"},
	{0x7E28E1E0D2CCA02EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge02dirt_sa.dds"},
	{0x7E29741BF3218642ULL, "vxca_caverns2prc_9ou_0_0.vlm"},
	{0x7E297AD5243D0C83ULL, "6056620_m.fxe"},
	{0x7E298E7264CD2B99ULL, "vxlt_warehouse_8b_0_0.vlm"},
	{0x7E29DC64E0C4259AULL, "6205700_m.fxe"},
	{0x7E2A2B014279A560ULL, "fen110_fenris6096222.cl"},
	{0x7E2AD9BAAB379C1EULL, "271300522.ogg"},
	{0x7E2B59ADAE8702D6ULL, "6139117_m.fxe"},
	{0x7E2BBC9CAAC77B01ULL, "art\\vfx\\combat\\vfx_glb_ele_n_minor_c.anb"},
	{0x7E2BC26F2DC878ADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerb.msh"},
	{0x7E2BD06F2DC890E0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerb.mao"},
	{0x7E2BD46F2DC89733ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerb.mmh"},
	{0x7E2BF7D605C49420ULL, "w_woundedcoast_185_0.phy"},
	{0x7E2C1F2768537459ULL, "6043501_m.fxe"},
	{0x7E2C220A44E8AB47ULL, "823379999.ogg"},
	{0x7E2CB0BA11994A53ULL, "merchant_core.ncs"},
	{0x7E2CB6E18D2FE98BULL, "dae011_wench_mob6149851.cl"},
	{0x7E2D065AF779A5E5ULL, "fxe_splash_blend.mao"},
	{0x7E2D2142EF820BA3ULL, "82055672.ogg"},
	{0x7E2D34AE19D9BE15ULL, "865301646.ogg"},
	{0x7E2D5089A3E83A4AULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\mov_p.ani"},
	{0x7E2DCA91D86FC81DULL, "vxdr_0_deeproads_bxe_0_0.vlm"},
	{0x7E2DEBE15880C807ULL, "w_bonepit_18_0_lc.mmh"},
	{0x7E2DF19E61BBB02FULL, "88601101.ogg"},
	{0x7E2DF824D216B613ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft90.ani"},
	{0x7E2E546002CAEBEAULL, "vxw_encounter2_camp_7k_0_0.vlm"},
	{0x7E2E909CA6630A0EULL, "vxw_bonepitdestroyed_a3_1d5_0_0.vlm"},
	{0x7E2ECE9D87921D78ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\leftarm_gesture_behind.gad"},
	{0x7E2EFD3DF51765E7ULL, "232488741.ogg"},
	{0x7E2F77DE28C18E6BULL, "6162080_m.fxe"},
	{0x7E2FB547336759DBULL, "6112663_m.fxe"},
	{0x7E31AB3ACAC3B4D8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic_4.mao"},
	{0x7E31C68B4FCB6B9EULL, "vxlt_0_lowtown_a2_h2e_0_0.vlm"},
	{0x7E31D351E9BAE3BFULL, "w_sundermount_54_0_ll.mmh"},
	{0x7E322E22D0CB1E35ULL, "follower_barks6154982.cl"},
	{0x7E3244B04CE1AAEDULL, "vxlt_0_lowtown_n_ev8_0_0.vlm"},
	{0x7E327325D196EC94ULL, "283098488.ogg"},
	{0x7E3292FFB94AE64CULL, "llac_ht_mansion2_818_2.dds"},
	{0x7E32D6F4BBD19AA2ULL, "dae000_bodhan_letter6144320.cl"},
	{0x7E331528A52B60C6ULL, "vxht_mansion2b_1aa_0_0.vlm"},
	{0x7E338BABA593F188ULL, "vxht_0_hightown_a3_bqe_0_0.vlm"},
	{0x7E341E66492EE751ULL, "hlac_lt_0_lowtown_docks_n_65535_2z.dds"},
	{0x7E3474F6FA68118BULL, "t3_ogr_h01.tnt"},
	{0x7E35761787A5AA53ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgte.phy"},
	{0x7E3585853B849B84ULL, "vxht_0_hightown_n_at7_0_0.vlm"},
	{0x7E35E321F8EEB385ULL, "1034061677.ogg"},
	{0x7E35F26FE22AA745ULL, "vxht_mansion2_336_0_0.vlm"},
	{0x7E3685831C5E3CE6ULL, "qun110_petrice6018510.cl"},
	{0x7E368D73192F0833ULL, "plt_prm_cod_isaac_armor.plo"},
	{0x7E3760EC8EF4DA5BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha02.phy"},
	{0x7E379916F552933CULL, "lowtown_fem_cityguard_mob6164813.cl"},
	{0x7E37DDEC39FEC6CAULL, "915839193.ogg"},
	{0x7E37E14AD65C7FE3ULL, "6162769_m.fxe"},
	{0x7E385F5D23A09A2FULL, "347058002.ogg"},
	{0x7E38DBA04124AF9FULL, "6202353_m.fxe"},
	{0x7E38FC138A0383F9ULL, "198747951.ogg"},
	{0x7E3947CE117E8DE3ULL, "w_encounter3_141_0.msh"},
	{0x7E3955CE117EA53DULL, "w_encounter3_141_0.mmh"},
	{0x7E3959CE117EAC4EULL, "w_encounter3_141_0.mao"},
	{0x7E399D9CAAD3847AULL, "art\\vfx\\combat\\vfx_glb_ele_n_minor_c.mmh"},
	{0x7E39D61CE1663939ULL, "hlac_ht_mansion2_818_60y.dds"},
	{0x7E39D91A5810CE2CULL, "6024584_f.fxe"},
	{0x7E39E5A34D9D7CC5ULL, "6154054_m.fxe"},
	{0x7E39F6CA63F40161ULL, "vxht_mansion_n_6ew_0_1.vlm"},
	{0x7E3A2B85DEF6C2A3ULL, "w_bonepitdestroyed_a3_21_0_lc.mmh"},
	{0x7E3A59109698252AULL, "mrl310_party_comments6142837.cl"},
	{0x7E3B2BBCBCBC78A2ULL, "88887584.ogg"},
	{0x7E3BDEE38F7BEEF7ULL, "205103187.ogg"},
	{0x7E3BF024D222E64AULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft90.evt"},
	{0x7E3BF334D8BF68BBULL, "vxht_0_hightown_n_ca9_0_0.vlm"},
	{0x7E3C32BD77B2E8B2ULL, "mag370_meredith6039228.cl"},
	{0x7E3C5A84FBDCBF99ULL, "296122131.ogg"},
	{0x7E3C913C457123A3ULL, "nat320_delilah6065930.cl"},
	{0x7E3CB5A4087BB1C6ULL, "follower_banter6154009.cl"},
	{0x7E3DF0C4816BC06CULL, "hlac_w_bonepitdestroyed_a3_131072_0x.dds"},
	{0x7E3E1544C836AD38ULL, "hlac_lt_shop_-1_6x.dds"},
	{0x7E3E3D4AAED1B110ULL, "hlac_w_woundedcoast_65537_0y.dds"},
	{0x7E3E458118AA2D1AULL, "vxdr_dungeon_2hf_0_0.vlm"},
	{0x7E3E4A99685047B0ULL, "34511024.ogg"},
	{0x7E3F026F4A3B4C54ULL, "dae000_bodhan6146804.cl"},
	{0x7E3F1E6FB63D5C13ULL, "1011075.ogg"},
	{0x7E3F51892828A94FULL, "6094778_m.fxe"},
	{0x7E3F609F4B7E326DULL, "620969313.ogg"},
	{0x7E3F872F8AB0EAA1ULL, "6100305_m.fxe"},
	{0x7E3FE5684614F528ULL, "hlac_dr_primevaltomb_664_0z.dds"},
	{0x7E401BFA548D74CEULL, "spt_billboards1.vsh.fxo"},
	{0x7E40B68D3ED7C36DULL, "vxlt_0_lowtown_n_es7_0_0.vlm"},
	{0x7E41A88C815DF0F8ULL, "225327595.ogg"},
	{0x7E41E4BF15DF1535ULL, "tanktacticspreset.gda"},
	{0x7E420756AB8868C5ULL, "vxlt_warehouse_r8_0_0.vlm"},
	{0x7E42334EF4F21E9CULL, "w_encounter3_59_0.rml"},
	{0x7E42CF24D228CCBDULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft90.gad"},
	{0x7E434304621B51A7ULL, "793465816.ogg"},
	{0x7E434A2D4CCC3816ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.impa.ani"},
	{0x7E43BCB6792A4F36ULL, "mag120_macha6080603.cl"},
	{0x7E44553D2C4B28DEULL, "6064974_m.fxe"},
	{0x7E44A66496FC7B96ULL, "hm_velasco.mor"},
	{0x7E45194EF4F428F1ULL, "w_encounter3_59_0.msh"},
	{0x7E45274EF4F4400CULL, "w_encounter3_59_0.mao"},
	{0x7E45334EF4F454AFULL, "w_encounter3_59_0.mmh"},
	{0x7E4587F6063DC136ULL, "516229737.ogg"},
	{0x7E45FCFA2397C053ULL, "dae000_light_tombs_boss6179458.cl"},
	{0x7E45FDD30FE05702ULL, "705500821.ogg"},
	{0x7E4686BF380BEEBBULL, "6047671_m.fxe"},
	{0x7E46DD14623F5D33ULL, "611111925.ogg"},
	{0x7E47928AE2572933ULL, "lt_0_lowtown_36_0_ll.mmh"},
	{0x7E47984ABEAF6155ULL, "zz_mag_debug6051359.cl"},
	{0x7E47B3002BFD6303ULL, "866624339.ogg"},
	{0x7E47E1609172C9F5ULL, "lt_0_lowtown_docks_a3_n_11_0_ll.mmh"},
	{0x7E48F2E6C5F215A7ULL, "6108991_f.fxe"},
	{0x7E490DF89E5064DFULL, "hlac_lt_0_lowtown_a3_n_c_-65539_0y.dds"},
	{0x7E493EC8B21D343CULL, "vxlt_house_ek_0_0.vlm"},
	{0x7E4942C470960120ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_rob_chaa_0t.dds"},
	{0x7E49A38E1821F780ULL, "562422535.ogg"},
	{0x7E49ABE6AA58E430ULL, "vxca_caverns2_4oc_0_0.vlm"},
	{0x7E49DAC9F80239E7ULL, "dae000cr_grand_cleric6173654.cl"},
	{0x7E4A5517D8D89DA8ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masf.mmh"},
	{0x7E4A7317D8D8D0D2ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masf.msh"},
	{0x7E4A87C1DD3D3EA9ULL, "art\\vfx\\cinematic\\vfx_chest_pierce_blood_p.impa.ani"},
	{0x7E4A931807BF3D27ULL, "6147643_m.fxe"},
	{0x7E4D7D384259BEFCULL, "vxca_caverns2_5zy_0_0.vlm"},
	{0x7E4DB9733AA2011CULL, "vxca_caverns2_8g3_0_0.vlm"},
	{0x7E4E0DB785B0E4C3ULL, "hlac_lt_0_lowtown_docks_a3_-1_2z.dds"},
	{0x7E4E1D68DEBDADE1ULL, "6041697_m.fxe"},
	{0x7E4E7DC86AC5E7A2ULL, "w_woundedcoast_36_0.msh"},
	{0x7E4E7FC86AC5EB58ULL, "w_woundedcoast_36_0.mmh"},
	{0x7E4E898FCA1952D5ULL, "vxlt_gamlenshouse_n_6o_0_0.vlm"},
	{0x7E4E8BC86AC5FFFBULL, "w_woundedcoast_36_0.mao"},
	{0x7E4F15CEDD9D689EULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.impa.ani"},
	{0x7E4F4355C2CA5DC5ULL, "311538181.ogg"},
	{0x7E4F47BC508B653FULL, "vxdr_0_deeproads_cv4_0_0.vlm"},
	{0x7E4FB9B3724E3ED1ULL, "vxht_0_hightown_a3_cat_0_1.vlm"},
	{0x7E4FF864978F3ED7ULL, "hlac_lt_0_lowtown_docks_n_-4_0y.dds"},
	{0x7E4FFBDD10F990D8ULL, "vxht_keep_n_6n2_0_0.vlm"},
	{0x7E5088EDC09C11AEULL, "909523529.ogg"},
	{0x7E50C7C7930BDA73ULL, "6028939_m.fxe"},
	{0x7E50D6667C43043BULL, "vxlt_gallowstemplar_43x_0_0.vlm"},
	{0x7E50EC78F3AB068AULL, "dae000_flemeth6013400.cl"},
	{0x7E5114D605E3E1B5ULL, "w_woundedcoast_185_0.msh"},
	{0x7E5116D605E3E52BULL, "w_woundedcoast_185_0.mmh"},
	{0x7E5122D605E3F9C8ULL, "w_woundedcoast_185_0.mao"},
	{0x7E513C4415351542ULL, "llac_lt_0_lowtown_a2_-131071_1.dds"},
	{0x7E51522D4CD88307ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.impa.evt"},
	{0x7E515C46095A52ECULL, "vxlt_0_lowtown_a2_n_byk_0_0.vlm"},
	{0x7E517542D76572B9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\hf_exp_movement.anb"},
	{0x7E51917F2B5EE94BULL, "dae000_merchant_generic_hm6172583.cl"},
	{0x7E519F036F6B24D2ULL, "6164564_m.fxe"},
	{0x7E51A2C1DD438B2BULL, "art\\vfx\\cinematic\\vfx_chest_pierce_blood_p.impa.gad"},
	{0x7E51ECEE46B8149BULL, "vxlt_hangedman_n_10b_0_0.vlm"},
	{0x7E52490257E51992ULL, "vxlt_undercityserialkiller_19h_0_0.vlm"},
	{0x7E526905676A18DCULL, "6096050_m.fxe"},
	{0x7E527FC86AC9D46BULL, "w_woundedcoast_36_0.rml"},
	{0x7E53460A227A2B29ULL, "gen00fl_fenris6126758.cl"},
	{0x7E53BCE2C9243CE6ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.impa.evt"},
	{0x7E53C685BC4251AFULL, "415284612.ogg"},
	{0x7E54023033CD05EAULL, "vxca_caverns1_j4e_0_0.vlm"},
	{0x7E543051D799ADA6ULL, "6149424_m.fxe"},
	{0x7E54A157A2140775ULL, "6134849_m.fxe"},
	{0x7E5546DA9FA5C797ULL, "675046308.ogg"},
	{0x7E5589B1E4C85781ULL, "564248242.ogg"},
	{0x7E55B80FB74B0FB7ULL, "vxw_sundermount_2ak_0_0.vlm"},
	{0x7E55CA9442CBBBDFULL, "llac_lt_0_lowtown_a2_n_1_8.dds"},
	{0x7E55E9E8A1C78E34ULL, "472761344.ogg"},
	{0x7E55EE13E3FF03B5ULL, "vxca_caverns1_mines_i6v_0_0.vlm"},
	{0x7E56061D78205BD2ULL, "362231744.ogg"},
	{0x7E561547D7C3D9B5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_hole.phy"},
	{0x7E562577DBBFFF32ULL, "dr_0_deeproads_17039.rml"},
	{0x7E573E75BF7E0F36ULL, "vxht_brothel_n_4y4_0_0.vlm"},
	{0x7E579FC1DD47F190ULL, "art\\vfx\\cinematic\\vfx_chest_pierce_blood_p.impa.evt"},
	{0x7E57ADF6DCCC5A27ULL, "vxw_encounter1_i9_0_0.vlm"},
	{0x7E57E14457123143ULL, "966842637.ogg"},
	{0x7E57FA6039C9D900ULL, "527789154.ogg"},
	{0x7E580E322EC79E69ULL, "mrl310_merrill6115551.cl"},
	{0x7E5859384E01B9FAULL, "gen00fl_bethany6101235.cl"},
	{0x7E5863643B6AE27AULL, "llac_lt_0_lowtown_a3_n_c_65537_5.dds"},
	{0x7E58732D4CDEDA50ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.impa.gad"},
	{0x7E588FCACE386FD7ULL, "401455307.ogg"},
	{0x7E58D27916B97EC5ULL, "6083835_m.fxe"},
	{0x7E5904F4F9B555A7ULL, "6140743_m.fxe"},
	{0x7E59966E7EA3069EULL, "366107717.ogg"},
	{0x7E5A29D4BF859319ULL, "6147146_m.fxe"},
	{0x7E5AD3E2C92A8231ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.impa.gad"},
	{0x7E5B00F9F6B8DCC7ULL, "vxht_0_hightown_a3_a4r_0_0.vlm"},
	{0x7E5B50D47E927D16ULL, "6019130_m.fxe"},
	{0x7E5BBBCA6F1C7F6FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\revenant.anb"},
	{0x7E5CA94E99AB3FC5ULL, "6134828_f.fxe"},
	{0x7E5CB77E091BDA04ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn6plank.msh"},
	{0x7E5CC97E091BF96DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn6plank.mao"},
	{0x7E5CCD7E091BFFA6ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn6plank.mmh"},
	{0x7E5CE274D4CF8832ULL, "6119520_m.fxe"},
	{0x7E5CF94B129F6011ULL, "20327872.ogg"},
	{0x7E5D0787165C26CDULL, "mrl310_merrill6140728.cl"},
	{0x7E5D2DCEDDA9CF6FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.impa.evt"},
	{0x7E5D9A4722C3F46DULL, "vxlt_0_lowtown_n_ei5_0_0.vlm"},
	{0x7E5DD2B22F121F98ULL, "170487463.ogg"},
	{0x7E5E2E012BBFA818ULL, "ht_0_hightown_n_21_0_ll.mmh"},
	{0x7E5E6DBA02EFA958ULL, "vxht_keep_n_6i3_0_0.vlm"},
	{0x7E5F461F410C36B0ULL, "dre000_sandal6033762.cl"},
	{0x7E5FD3C10E0509B0ULL, "6125121_m.fxe"},
	{0x7E6030A9A1A45ADFULL, "609491895.ogg"},
	{0x7E60F09BA6B57AF2ULL, "vxw_encounter2_4b_0_0.vlm"},
	{0x7E612B63E57709E8ULL, "6139942_m.fxe"},
	{0x7E61D4E2C930A2DFULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.impa.ani"},
	{0x7E62024D2B4699E6ULL, "6111667_m.fxe"},
	{0x7E631C230C3BCF2CULL, "481709090.ogg"},
	{0x7E632ECEDDAE3C18ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.impa.gad"},
	{0x7E6332B41B34C3F2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_basic.phy"},
	{0x7E6347EC8F19EAFAULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha02.msh"},
	{0x7E6359EC8F1A0940ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha02.mmh"},
	{0x7E63D2ACCBB5B759ULL, "6202801_m.fxe"},
	{0x7E641448898EB494ULL, "6152624_m.fxe"},
	{0x7E64AEAF7D2136C2ULL, "one100_ht_banter6060900.cl"},
	{0x7E64D858FC0A3D6BULL, "454780485.ogg"},
	{0x7E65207B6123B80FULL, "lgt320_shw_ors_secret_meet6156354.cl"},
	{0x7E65536412C2D626ULL, "mrl000_marethari6142376.cl"},
	{0x7E66E5A19A91EB0CULL, "vxlt_0_lowtown_docks_n_b9j_0_0.vlm"},
	{0x7E671B8C8F75BEF4ULL, "vxlt_hangedman_n_21q_0_0.vlm"},
	{0x7E67CE767836C6ACULL, "vxht_mansion2_3e1_0_0.vlm"},
	{0x7E67D2073E6814FDULL, "6110054_m.fxe"},
	{0x7E680017EEFDC05DULL, "sw_bl_cl.mao"},
	{0x7E68324812EFC7B9ULL, "115433941.ogg"},
	{0x7E6855FEC0C00333ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyn.phy"},
	{0x7E689DF25CCD4BD4ULL, "hlac_lt_gallowstemplar_n_2383_9z.dds"},
	{0x7E68DD9BCB5E3F63ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_020a_n.dds"},
	{0x7E698768CDC243B7ULL, "w_sundermount_211_0.mmh"},
	{0x7E698B68CDC24AC4ULL, "w_sundermount_211_0.mao"},
	{0x7E699D68CDC26929ULL, "w_sundermount_211_0.msh"},
	{0x7E69CDA4897BF92DULL, "seb120pt_duty.ncs"},
	{0x7E69D8A65FFE1C75ULL, "6169850_m.fxe"},
	{0x7E6A14ABAD95010FULL, "1014886582.ogg"},
	{0x7E6A51DD05F0E28AULL, "hlac_lt_0_lowtown_0_4y.dds"},
	{0x7E6B5AFBB3CDD1C3ULL, "art\\levels\\4cinematics\\dwarf\\_textures\\dwarfchair_d.dds"},
	{0x7E6BC6B0B91C45DFULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_bdy_suca_0t.dds"},
	{0x7E6BEB33697B0C2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\pp_sebastian.anb"},
	{0x7E6C642025A0DDD6ULL, "46062226.ogg"},
	{0x7E6D8FA99DD756F3ULL, "vxca_cavern1_haunted_jg8_0_0.vlm"},
	{0x7E6DC2A09557DFD6ULL, "vxca_caverns2_8ld_0_0.vlm"},
	{0x7E6DED9D271C8EAAULL, "6116496_f.fxe"},
	{0x7E6E162F27BDB715ULL, "293891526.ogg"},
	{0x7E6E19B002526559ULL, "6173808_m.fxe"},
	{0x7E6E4C190FFDB130ULL, "follower_banter6166972.cl"},
	{0x7E6E720C74B8CA00ULL, "6119540_m.fxe"},
	{0x7E6F2DAE92E955EAULL, "mcr370_assignments6081686.cl"},
	{0x7E6FDE75750F9DFAULL, "ht_0_hightown_60_0.rml"},
	{0x7E702B73A5C83870ULL, "504083501.ogg"},
	{0x7E707AA135098634ULL, "rdr110dg_party_comments6094673.cl"},
	{0x7E70D4C835D10794ULL, "vxlt_0_lowtown_docks_a3_n_8pl_0_0.vlm"},
	{0x7E7110D9E990E8CFULL, "vxw_woundedcoast_37_0_0.vlm"},
	{0x7E716D3E23A9CAA0ULL, "gen00fl_aveline6139257.cl"},
	{0x7E718479875077B7ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax105a_0n.dds"},
	{0x7E71C2A04FA1A707ULL, "vxw_encounter2_a0_0_0.vlm"},
	{0x7E71F64437D2EA5AULL, "hlac_lt_0_lowtown_docks_n_65534_0x.dds"},
	{0x7E723A06F4246A4FULL, "darktown_fem_lowlife_mob6176158.cl"},
	{0x7E724489BDABCFD9ULL, "6209901_m.fxe"},
	{0x7E726354CFE54311ULL, "hla_ht_0_hightown_a2_-131072_0z.dds"},
	{0x7E728DD68CBF1349ULL, "hlac_w_encounter2_-1_0y.dds"},
	{0x7E729BC99B3740B7ULL, "vxw_sundermount_2oe_0_0.vlm"},
	{0x7E7367395DCFC6B3ULL, "gen00fl_fenris6126932.cl"},
	{0x7E7378AE145F4CF6ULL, "dre000_varric6035148.cl"},
	{0x7E7380193AC50502ULL, "492332542.ogg"},
	{0x7E73C7CB73B697FBULL, "134272945.ogg"},
	{0x7E73DD7F17BB4B40ULL, "vxw_bonepit_248_0_0.vlm"},
	{0x7E747BF593932F50ULL, "mag110_samson6049808.cl"},
	{0x7E74A7DADD6A510EULL, "vxdr_primevaltomb_8y5_0_0.vlm"},
	{0x7E74D9765EC272B5ULL, "655431049.ogg"},
	{0x7E75577E8C88038CULL, "vxlt_0_lowtown_docks_n_b6k_0_0.vlm"},
	{0x7E7573188C2AA229ULL, "6109254_m.fxe"},
	{0x7E767016EC1B86AFULL, "w_bonepitmining_a2_39_0.mao"},
	{0x7E767C16EC1B9B4CULL, "w_bonepitmining_a2_39_0.mmh"},
	{0x7E768216EC1BA53EULL, "w_bonepitmining_a2_39_0.msh"},
	{0x7E775E28FB498725ULL, "seb321_leliana6180951.cl"},
	{0x7E77885026C494FBULL, "ch2alphemarmtnt.psh.fxo"},
	{0x7E778B654ECE671AULL, "vxht_0_hightown_a2_a4r_0_0.vlm"},
	{0x7E77DF3C11FF25F0ULL, "49580152.ogg"},
	{0x7E793F587C7371ADULL, "vxht_bartrand_n_34c_0_0.vlm"},
	{0x7E797D00CF09B1A7ULL, "zz_temp_start_area.ncs"},
	{0x7E79BDD1F6F49705ULL, "774657364.ogg"},
	{0x7E7AC31473706719ULL, "214462769.ogg"},
	{0x7E7B9EB8D91E75B1ULL, "118157062.ogg"},
	{0x7E7BCAE0CFA22804ULL, "hlac_ca_cavern1_haunted_24426_2x.dds"},
	{0x7E7C13EA0DCE753FULL, "vxlt_0_lowtown_a2_n_byd_0_0.vlm"},
	{0x7E7C2163D6A8082BULL, "816781509.ogg"},
	{0x7E7C347D874DF856ULL, "vxca_caverns2_8ke_0_0.vlm"},
	{0x7E7CE2E32EB885B2ULL, "dae000_meredith6085260.cl"},
	{0x7E7CEB24C20D9CC4ULL, "vxlt_warehouse_5x_0_0.vlm"},
	{0x7E7D170E1F09D18DULL, "dae011cr_bartender_friend.utc"},
	{0x7E7D765760FB43E2ULL, "983314291.ogg"},
	{0x7E7E20DBD3B6F16EULL, "625338342.ogg"},
	{0x7E7E27B1B4E5E072ULL, "vxw_blightlands_dy_0_0.vlm"},
	{0x7E7E812CEA4E76EEULL, "gen_im_acc_blt_act2_07.uti"},
	{0x7E7ED0264140DD65ULL, "follower_banter6172332.cl"},
	{0x7E7F67B7D664FBC5ULL, "6108257_m.fxe"},
	{0x7E7FA14CBF2519D2ULL, "388584118.ogg"},
	{0x7E7FB40B830EC706ULL, "vxlt_gallowstemplar_n_5xf_0_0.vlm"},
	{0x7E7FFD73DA68250AULL, "vxlt_0_lowtown_docks_buw_0_0.vlm"},
	{0x7E80347D4197BF87ULL, "vxw_encounter2_l3_0_0.vlm"},
	{0x7E809B8DAD4372ECULL, "gen00fl_fenris6110792.cl"},
	{0x7E80C6E789A8E28FULL, "da2_prc_one_ru-ru.cre"},
	{0x7E810DA68D2D5937ULL, "vxw_sundermount_2pd_0_0.vlm"},
	{0x7E81989DB47CEFF6ULL, "vxlt_0_lowtown_a3_n_c_fef_0_0.vlm"},
	{0x7E82261098CEFAE1ULL, "art\\vfx\\materials\\vfx_half_dot_blend.mao"},
	{0x7E824E07518691B8ULL, "6060401_f.fxe"},
	{0x7E82609086F8FDD7ULL, "vxlt_darktown_427_0_0.vlm"},
	{0x7E8266E18497841CULL, "zz_one_debug6099755.cl"},
	{0x7E82B0778E36F9DDULL, "ico_armor_heavy_1.dds"},
	{0x7E82B55DAAD31D4FULL, "6171567_m.fxe"},
	{0x7E82E54E247D71E2ULL, "6099347_m.fxe"},
	{0x7E8381E30F9A61BFULL, "6093300_m.fxe"},
	{0x7E83A014BCD66308ULL, "hlac_lt_gallowscourtyard_65535_0z.dds"},
	{0x7E83C99550A8494BULL, "vxca_caverns2prc_4ti_0_1.vlm"},
	{0x7E83FBA63497D0C8ULL, "662054281.ogg"},
	{0x7E845CD8EAA37681ULL, "l_lgt351ar_shw_finale_cave.wbd"},
	{0x7E845D9133439A25ULL, "hlac_w_sundermount_196608_0x.dds"},
	{0x7E84BEBBA693CC25ULL, "vxdr_primevaltomb_747_0_1.vlm"},
	{0x7E8531E10F7EDB36ULL, "vxdr_primevaltomb_396_0_0.vlm"},
	{0x7E8555E16D0EC760ULL, "dra_dust.mao"},
	{0x7E856E3CF5C4F9BFULL, "vxht_chantry_2a7_0_0.vlm"},
	{0x7E858DEEDE23EED6ULL, "6133506_m.fxe"},
	{0x7E85AAB0B3087089ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_inquisitionsoldier1_.gad"},
	{0x7E8617D90DDC6EB2ULL, "vxca_caverns1_mines_j1w_0_0.vlm"},
	{0x7E865614FD04497EULL, "hlac_lt_0_lowtown_docks_n_-65538_3z.dds"},
	{0x7E866C61048188C5ULL, "hlac_lt_undercityserialkiller_-1_13y.dds"},
	{0x7E8699D509A0829FULL, "lt_0_lowtown_docks_a3_n_27_0_lc.mmh"},
	{0x7E86EDA2C8150BFBULL, "one000_ilen6205654.cl"},
	{0x7E874EFBB3693E80ULL, "104586655.ogg"},
	{0x7E875609A78783A6ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_activ_c.impa.evt"},
	{0x7E879671943BF186ULL, "6115577_m.fxe"},
	{0x7E87B1356E698A2DULL, "vxht_bartrand_n_33b_0_0.vlm"},
	{0x7E8819C6ECF10E0BULL, "vxht_0_hightown_a2_c5f_0_0.vlm"},
	{0x7E88F84F192F32A0ULL, "qun231_varnell6077241.cl"},
	{0x7E89666BF287ED08ULL, "crt_harvester.bnk"},
	{0x7E89B1CC0542930FULL, "6177240_m.fxe"},
	{0x7E8ABD78BC90A570ULL, "gen00fl_anders6113016.cl"},
	{0x7E8B5BEB57BC931EULL, "dae000_keep_noble_sta6161525.cl"},
	{0x7E8B89AD1DAB75ADULL, "vxht_0_hightown_n_5eq_0_0.vlm"},
	{0x7E8C117F39DD1143ULL, "80975684.ogg"},
	{0x7E8C7BB0B30E3FD7ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_inquisitionsoldier1_.ani"},
	{0x7E8C998EA6DBF8F2ULL, "vxw_blightlands_iz_0_0.vlm"},
	{0x7E8CAA6C409C48C9ULL, "lgt210_qun_blk_qunari6142310.cl"},
	{0x7E8CB984A47DB0FAULL, "dae000_flemeth6013343.cl"},
	{0x7E8CD477E97115CCULL, "6149321_m.fxe"},
	{0x7E8CF1BC9AE0F34BULL, "119860990.ogg"},
	{0x7E8D5D7C59F91E91ULL, "67981341.ogg"},
	{0x7E8D6D09A78C16F1ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_activ_c.impa.gad"},
	{0x7E8DD183E576AC7EULL, "946914898.ogg"},
	{0x7E8DE009216E49E8ULL, "dae000_worthy6166580.cl"},
	{0x7E8E27501E8AB105ULL, "85071046.ogg"},
	{0x7E8EC3BA0E2D3919ULL, "gen00fl_fenris6115309.cl"},
	{0x7E8ECD4349BE6F32ULL, "hlac_w_woundedcoast_-131070_0x.dds"},
	{0x7E8F710916007E42ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.impa.gad"},
	{0x7E8F7745C9176033ULL, "7772831.ogg"},
	{0x7E8F94890DEEDF6AULL, "vxht_mansion_n_57p_0_0.vlm"},
	{0x7E8F9F8FC9BCCA50ULL, "vxw_woundedcoast_1gy_0_0.vlm"},
	{0x7E8FBC160EB5F16CULL, "hlac_lt_gallowstemplar_f_2394_1z.dds"},
	{0x7E8FF92B294FE4E2ULL, "vxlt_0_lowtown_docks_aqx_0_0.vlm"},
	{0x7E905CEB64AC9A3EULL, "723828891.ogg"},
	{0x7E90BBCC4AD65ACFULL, "vxlt_0_lowtown_docks_a3_con_0_0.vlm"},
	{0x7E90D26D78EF1657ULL, "vxlt_darktown_496_0_0.vlm"},
	{0x7E9103A9E45D4FF4ULL, "vxht_bartrand_n_4kl_0_0.vlm"},
	{0x7E914AEBD9D8DB74ULL, "589192556.ogg"},
	{0x7E919D01BF23AA66ULL, "vxht_0_hightown_a2_b5z_0_0.vlm"},
	{0x7E92091DA66F31A3ULL, "follower_banter6170585.cl"},
	{0x7E92466B93A30203ULL, "vxlt_hangedman_1o0_0_0.vlm"},
	{0x7E92B1CF58C7FC22ULL, "genip_blocker_wide.utp"},
	{0x7E92B91A39E7053BULL, "hlac_lt_undercityserialkiller_-1_45x.dds"},
	{0x7E931A63DCEFD2A3ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkr.ani"},
	{0x7E93CC08DC0AA3D3ULL, "llac_lt_blackemporium_1055_25.dds"},
	{0x7E93E019E7BB123FULL, "vxht_chantry_2z6_0_0.vlm"},
	{0x7E94361B7B291737ULL, "llac_ht_keep_n_4519_6.dds"},
	{0x7E9446D72ED7BD10ULL, "zz_follower_army.cut"},
	{0x7E944B5991D12A51ULL, "hlac_lt_foundry_n_320_18y.dds"},
	{0x7E94502E9D811C98ULL, "6201522_m.fxe"},
	{0x7E946E09A792369FULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_activ_c.impa.ani"},
	{0x7E947B0A5FDF241EULL, "gen00fl_varric6104939.cl"},
	{0x7E948A06BF58121CULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_rf.ani"},
	{0x7E94B80CCB30C02DULL, "hlac_lt_0_lowtown_docks_n_-4_1x.dds"},
	{0x7E95495168AF834CULL, "139601772.ogg"},
	{0x7E95650AB5B07D1FULL, "831227770.ogg"},
	{0x7E95D013AA5EA802ULL, "t3_grl_g02.tnt"},
	{0x7E95DCA9D7F4D369ULL, "hlac_lt_0_lowtown_a2_n_c_131071_6x.dds"},
	{0x7E96000428F069DBULL, "vxht_mansion2b_3di_0_0.vlm"},
	{0x7E968BA3DEE7268BULL, "vxht_0_hightown_a2_c0y_0_0.vlm"},
	{0x7E9694091606D839ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.impa.evt"},
	{0x7E96B3144FBEAE76ULL, "654624829.ogg"},
	{0x7E96BB5A36E716BCULL, "6084520_f.fxe"},
	{0x7E96EF83FBD6EA70ULL, "vxw_bonepitdestroyed_a3_7y_0_0.vlm"},
	{0x7E96F77B9A168088ULL, "6143640_m.fxe"},
	{0x7E97F0456A5AB693ULL, "zz_followers_banter6158754.cl"},
	{0x7E9800FE8AFA35EAULL, "mag131_keran6091387.cl"},
	{0x7E981910F36D9853ULL, "285826509.ogg"},
	{0x7E986F1787FA1BC8ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgte.mmh"},
	{0x7E988D1787FA4EF2ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgte.msh"},
	{0x7E98B029EFCA416BULL, "qun131_party_comment6117303.cl"},
	{0x7E991EE7ED4C85E6ULL, "1005349105.ogg"},
	{0x7E994CF1A8A4BA19ULL, "llac_lt_undercityserialkiller_-1_21.dds"},
	{0x7E9979C7E4B66250ULL, "clf260_harley6069590.cl"},
	{0x7E99AB970F067FA2ULL, "gen00fl_varric6105789.cl"},
	{0x7E9A2669B8FC9DEEULL, "hlac_lt_0_lowtown_docks_a3_131071_5y.dds"},
	{0x7E9A4D6EC70D4066ULL, "hlac_lt_0_lowtown_docks_n_131073_1y.dds"},
	{0x7E9AAF1D7C67BA7BULL, "hlac_w_blightlands_-131070_0y.dds"},
	{0x7E9AC788CF7F73E7ULL, "vxlt_0_lowtown_a2_n_c_cmw_0_0.vlm"},
	{0x7E9B792C9771E95DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keep.phy"},
	{0x7E9B9706BF5E475EULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_rf.gad"},
	{0x7E9BAB701B95EEE1ULL, "vfx_rendertoshadowmap_p1.vsh.fxo.dx11.assembly"},
	{0x7E9BD2F5E1136700ULL, "one140cr_charade6137066.cl"},
	{0x7E9C46677B9C5C4EULL, "llac_w_sundermount_196610_0.dds"},
	{0x7E9CAD9CA4105EB0ULL, "mcr310_sibling6076280.cl"},
	{0x7E9CBBEE1B66F66EULL, "528438349.ogg"},
	{0x7E9CBE54D50C1E67ULL, "477204599.ogg"},
	{0x7E9CC2B7DE18E67DULL, "114425428.ogg"},
	{0x7E9CFF461C866BFFULL, "6103197_m.fxe"},
	{0x7E9D60142C2ADC60ULL, "hlac_lt_0_lowtown_a2_131071_2x.dds"},
	{0x7E9D6B5CD0818D75ULL, "vxca_caverns2prc_aal_0_0.vlm"},
	{0x7E9E21DDF69AE63DULL, "vxca_caverns1_coj_0_0.vlm"},
	{0x7E9E2847D801AFD0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_hole.msh"},
	{0x7E9E2EA6F69B69FFULL, "vxw_sundermount_266_0_0.vlm"},
	{0x7E9E327AC440E559ULL, "460972553.ogg"},
	{0x7E9E3647D801C7EAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_hole.mmh"},
	{0x7E9E3A47D801CE21ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_hole.mao"},
	{0x7E9EDEF4CA450C51ULL, "dae000_brothel_orana6160795.cl"},
	{0x7E9F2FD74D754B9DULL, "6110388_m.fxe"},
	{0x7E9F9D0A7A6359B4ULL, "6081800_f.fxe"},
	{0x7E9FAEF3B2FB901AULL, "zz_and_debug6134079.cl"},
	{0x7E9FC97B4764DF81ULL, "227984049.ogg"},
	{0x7EA0410D21B1F7D3ULL, "vxlt_backalley_n_1dw_0_0.vlm"},
	{0x7EA0A4A0CF921DFCULL, "vxca_caverns1_mines_egx_0_0.vlm"},
	{0x7EA0B7928A723DC2ULL, "lt_gallowstemplar_f_2395.rml"},
	{0x7EA13263DCFC389AULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkr.evt"},
	{0x7EA13F94A8F2750DULL, "pro000_refugee_male6052104.cl"},
	{0x7EA14EA48F27C63FULL, "w_bonepitdestroyed_a3_87_0.msh"},
	{0x7EA154A48F27D0E1ULL, "w_bonepitdestroyed_a3_87_0.mmh"},
	{0x7EA160A48F27E482ULL, "w_bonepitdestroyed_a3_87_0.mao"},
	{0x7EA1C5D7E2987721ULL, "58487823.ogg"},
	{0x7EA1F5E6F4062059ULL, "vxlt_gallowsprison_n_6nt_0_0.vlm"},
	{0x7EA26F5B80F380F3ULL, "6218938_m.fxe"},
	{0x7EA27206BF64277DULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_rf.evt"},
	{0x7EA28A9F9DE4C36AULL, "w_bonepit_16_0_ll.mmh"},
	{0x7EA321E2CF2CEBD7ULL, "gen_im_trs_krk_junk2_02.uti"},
	{0x7EA32F0D8E3AB218ULL, "6130112_m.fxe"},
	{0x7EA338495B9B6A53ULL, "13401217.ogg"},
	{0x7EA3809DE0AE7984ULL, "6044541_m.fxe"},
	{0x7EA39C0916117028ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.impa.ani"},
	{0x7EA3E4862D4B7207ULL, "226594920.ogg"},
	{0x7EA48DD3497D82B4ULL, "362997031.ogg"},
	{0x7EA4917E27A2FE7CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldc.mao"},
	{0x7EA49D7E27A3121FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldc.mmh"},
	{0x7EA4A37E27A31CC1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldc.msh"},
	{0x7EA5118FCA880591ULL, "44180383.ogg"},
	{0x7EA5267714F6F3F5ULL, "vxlt_0_lowtown_a2_n_c_hjg_0_0.vlm"},
	{0x7EA5AD7745529BB9ULL, "6095403_m.fxe"},
	{0x7EA62F8013DA0EC3ULL, "llac_lt_0_lowtown_131072_6.dds"},
	{0x7EA73638D219119CULL, "577093670.ogg"},
	{0x7EA7A01E75781032ULL, "6076241_m.fxe"},
	{0x7EA7B70E701132A6ULL, "vxw_bonepitmining_a2_oq_0_0.vlm"},
	{0x7EA7DB059B9238D2ULL, "gentr_autosave.ncs"},
	{0x7EA83163DD02558DULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkr.gad"},
	{0x7EA845B925E376D0ULL, "6060723_m.fxe"},
	{0x7EA90B868F1E2783ULL, "898710912.ogg"},
	{0x7EA93965C1758C67ULL, "vxlt_0_lowtown_a2_n_c_cfv_0_0.vlm"},
	{0x7EA957B11971374BULL, "553853907.ogg"},
	{0x7EA96A95FC0B1172ULL, "6032533_m.fxe"},
	{0x7EA9781141957CF3ULL, "6121258_m.fxe"},
	{0x7EA97D488AC829F2ULL, "vxw_blightlands_wd_0_0.vlm"},
	{0x7EA9BB671A456DE8ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_reva_0n.dds"},
	{0x7EA9C1C2E3069652ULL, "vxlt_backalley_n_1dr_0_0.vlm"},
	{0x7EAA7ED8EAC47F83ULL, "l_lgt351ar_shw_finale_cave.bnk"},
	{0x7EAA93A9455D2527ULL, "mrl310_merrill6144186.cl"},
	{0x7EAAFB6ABAF02962ULL, "zz_spy121cr_jalen.utc"},
	{0x7EAB0A3A96F15735ULL, "qun131_party_comment6117275.cl"},
	{0x7EAB4FFBF3CDD4AEULL, "vxht_chantry_285_0_0.vlm"},
	{0x7EAB530AB04A6E8AULL, "vxlt_0_lowtown_docks_bbr_0_0.vlm"},
	{0x7EAB68967A8964C5ULL, "vxw_encounter3_46_0_0.vlm"},
	{0x7EABB1E23388C54BULL, "94718792.ogg"},
	{0x7EABC737CC68E2DBULL, "vxlt_gallowstemplar_f_31z_0_0.vlm"},
	{0x7EABD2B88DF55F90ULL, "vxht_brothel_n_5i7_0_0.vlm"},
	{0x7EABF3FD08AC0376ULL, "cli400cr_city_gate_defender.utc"},
	{0x7EAC6E2C978049C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keep.mao"},
	{0x7EAC7A2C97805E62ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keep.mmh"},
	{0x7EAC7C2C97806118ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keep.msh"},
	{0x7EACC7866F13D6D9ULL, "vxht_mansion_n_4zu_0_0.vlm"},
	{0x7EAD4C1ABDC31D10ULL, "6208697_m.fxe"},
	{0x7EADA098321D7D18ULL, "6121005_f.fxe"},
	{0x7EAE603DA7048279ULL, "vxw_sundermount_2z5_0_1.vlm"},
	{0x7EAE6A76456D33FAULL, "vxht_mansion_n_61e_0_0.vlm"},
	{0x7EAE9E73CEF6EF44ULL, "ser270_emeric6113747.cl"},
	{0x7EAF3D819F615181ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_medj_0s.dds"},
	{0x7EAF586BF2A8A4B2ULL, "crt_harvester.wbd"},
	{0x7EAF7B2AC51BC461ULL, "6039301_f.fxe"},
	{0x7EAFA7B3282C9AB0ULL, "vxlt_0_lowtown_docks_a3_cuk_0_0.vlm"},
	{0x7EAFA992D2EC5E5CULL, "266238155.ogg"},
	{0x7EAFFFE5BA6F501DULL, "6124022_m.fxe"},
	{0x7EB00E0B6636A461ULL, "vxlt_0_lowtown_e1n_0_0.vlm"},
	{0x7EB01F9BAA139936ULL, "738138479.ogg"},
	{0x7EB080C112AC60AFULL, "255475403.ogg"},
	{0x7EB095482221D819ULL, "vxht_keep_n_67t_0_0.vlm"},
	{0x7EB0B90CAB79FC2AULL, "vxht_mansion2b_1am_0_0.vlm"},
	{0x7EB0CE2F86D48766ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkr.evt"},
	{0x7EB0E6CB3F809D3DULL, "dae000_female_tranquil6165146.cl"},
	{0x7EB171462801183FULL, "follower_banter6155508.cl"},
	{0x7EB1D16E7BA2F0F0ULL, "hlac_ht_0_hightown_a3_131070_0z.dds"},
	{0x7EB1F445989E33B9ULL, "1014531127.ogg"},
	{0x7EB209473FDA4E00ULL, "vxlt_0_lowtown_docks_a3_cy7_0_0.vlm"},
	{0x7EB2784CCB1BCFB6ULL, "427041255.ogg"},
	{0x7EB2AE6A41DFE876ULL, "qun120cr_javaris6043994.cl"},
	{0x7EB3259F7B48D04BULL, "6145120_m.fxe"},
	{0x7EB3408EC7E04809ULL, "vxht_mansion_n_57n_0_1.vlm"},
	{0x7EB35F9D915A554DULL, "vxlt_gallowstemplar_6qf_0_0.vlm"},
	{0x7EB39DAC3E02E812ULL, "lt_0_lowtown_docks_18_0_ll.mmh"},
	{0x7EB3C4937A392CDFULL, "6144567_m.fxe"},
	{0x7EB43E4D9872FE4FULL, "art\\characters\\creatures\\dragon\\animation\\kd_lp.gad"},
	{0x7EB5562D39ABB0A3ULL, "tools_fx_collisionplane.mao"},
	{0x7EB5A07063B5E544ULL, "851915656.ogg"},
	{0x7EB6020F0903286DULL, "zz_followers_banter6157915.cl"},
	{0x7EB60D672189DB13ULL, "6048731_m.fxe"},
	{0x7EB628EB62074B26ULL, "vxw_bonepitmining_a2_jr_0_0.vlm"},
	{0x7EB6794701F814CFULL, "w_bonepit_251_0.mao"},
	{0x7EB6854701F8296CULL, "w_bonepit_251_0.mmh"},
	{0x7EB68B4701F8335EULL, "w_bonepit_251_0.msh"},
	{0x7EB69A542F39F312ULL, "vxdr_0_deeproads_ew8_0_0.vlm"},
	{0x7EB6BCBA18387198ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgta.phy"},
	{0x7EB74A9FC3DD449DULL, "lt_0_lowtown_docks_a3_n_189_0.mao"},
	{0x7EB74E9FC3DD4BD6ULL, "lt_0_lowtown_docks_a3_n_189_0.mmh"},
	{0x7EB7589FC3DD5CD4ULL, "lt_0_lowtown_docks_a3_n_189_0.msh"},
	{0x7EB7E52F86DACCB1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkr.gad"},
	{0x7EB84B781744295BULL, "dr_primevaltomb.arl"},
	{0x7EB84DB06801C4B7ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_4z.dds"},
	{0x7EB8BE1DED70C8B3ULL, "244518717.ogg"},
	{0x7EB8D07DD3F75B89ULL, "79757680.ogg"},
	{0x7EB8D734DF133AF7ULL, "vfx_vfx_p4.vsh.fxo"},
	{0x7EB907D507890458ULL, "1069045456.ogg"},
	{0x7EB996C1278B5048ULL, "vxlt_0_lowtown_e1k_0_0.vlm"},
	{0x7EB9AF9FB7F26DB5ULL, "vxlt_0_lowtown_docks_n_bs5_0_0.vlm"},
	{0x7EB9DA736C7F7D45ULL, "vxw_encounter3_15_0_0.vlm"},
	{0x7EBA44957FEB7810ULL, "vxht_brothel_n_5l4_0_0.vlm"},
	{0x7EBAEF43599695CFULL, "118471207.ogg"},
	{0x7EBAFFD8A557644DULL, "hla_ht_0_hightown_-196608_0x.dds"},
	{0x7EBB2E4F19519517ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3.gad"},
	{0x7EBB52579CC97230ULL, "vxw_woundedcoast_41_0_0.vlm"},
	{0x7EBBEB8911CF4E9BULL, "dae000_orsino6085485.cl"},
	{0x7EBCECA947AF910BULL, "vxw_bonepit_1p0_0_0.vlm"},
	{0x7EBD24C7059E28D3ULL, "vxlt_backalley_n_1jy_0_0.vlm"},
	{0x7EBD3554EB7952D0ULL, "l_and231ar_gallowsbasement.wbd"},
	{0x7EBD36158C6EBEC4ULL, "6149398_m.fxe"},
	{0x7EBD890E4A7EEAC9ULL, "6054062_m.fxe"},
	{0x7EBE7399782C46D4ULL, "vxlt_0_lowtown_docks_a3_bta_0_0.vlm"},
	{0x7EBEE62F86E0ED5FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkr.ani"},
	{0x7EBFFB28EC97E395ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medd_0n.dds"},
	{0x7EC0470BB659F8BCULL, "qun221_viscount6069910.cl"},
	{0x7EC04CC4D193E0B9ULL, "22475797.ogg"},
	{0x7EC08B4528E9EEC9ULL, "575318394.ogg"},
	{0x7EC098272B2328C9ULL, "410128663.ogg"},
	{0x7EC2434F1957D7A4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3.evt"},
	{0x7EC28CEDD6EE39CBULL, "608318953.ogg"},
	{0x7EC3945D5BEEE7C0ULL, "mag110_arianni6020499.cl"},
	{0x7EC3A0EEAE328B53ULL, "6151257_m.fxe"},
	{0x7EC3CB730A81E471ULL, "1050065046.ogg"},
	{0x7EC3FD89951B0248ULL, "lt_0_lowtown_docks_a3_131_0.mmh"},
	{0x7EC40989951B16EBULL, "lt_0_lowtown_docks_a3_131_0.mao"},
	{0x7EC41B89951B3572ULL, "lt_0_lowtown_docks_a3_131_0.msh"},
	{0x7EC41D142DCE2963ULL, "vxca_caverns2_6hg_0_0.vlm"},
	{0x7EC4C0B937014B25ULL, "582610755.ogg"},
	{0x7EC4E6556193B396ULL, "719225668.ogg"},
	{0x7EC4EEA40B965849ULL, "6068574_f.fxe"},
	{0x7EC5579471595DC2ULL, "llac_lt_0_lowtown_a2_-2_0.dds"},
	{0x7EC58E64370C9944ULL, "llac_lt_gallowscourtyard_65535_0.dds"},
	{0x7EC6582ADF3986D3ULL, "6164958_m.fxe"},
	{0x7EC65BF27CB0BBDEULL, "gen00fl_isabela6075253.cl"},
	{0x7EC6ADDF2B8C7B87ULL, "6165597_m.fxe"},
	{0x7EC73FBC5FB5CC06ULL, "hlac_lt_0_lowtown_a2_131071_3y.dds"},
	{0x7EC7CDF4402DFACEULL, "art\\characters\\weapons\\staff\\w_stf_sf03a.mmh"},
	{0x7EC7D2342CA8AD38ULL, "329494654.ogg"},
	{0x7EC7D9F4402E0F65ULL, "art\\characters\\weapons\\staff\\w_stf_sf03a.mao"},
	{0x7EC7E7F4402E26BCULL, "art\\characters\\weapons\\staff\\w_stf_sf03a.msh"},
	{0x7EC85B4D9883D815ULL, "art\\characters\\creatures\\dragon\\animation\\kd_lp.ani"},
	{0x7EC8632678260009ULL, "vxw_bonepitdestroyed_a3_238_0_0.vlm"},
	{0x7EC8705EDBE78B99ULL, "96479972.ogg"},
	{0x7EC879BA5643D674ULL, "dae000cr_arishok.utc"},
	{0x7EC89146974E59D2ULL, "848224107.ogg"},
	{0x7EC8AF0101CDEDDEULL, "dae200_bran6094726.cl"},
	{0x7EC8BF7FA7C68F1BULL, "302609146.ogg"},
	{0x7EC8D128ED48D9BAULL, "mcr370_negotiation6080211.cl"},
	{0x7EC969E204936BA9ULL, "6066030_m.fxe"},
	{0x7ECA41009650F155ULL, "zz_and_debug6134157.cl"},
	{0x7ECB3CA7A42D6813ULL, "1002497884.ogg"},
	{0x7ECB4EFEC11474A8ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyn.mmh"},
	{0x7ECB5E8639A5A98BULL, "vxw_bonepit_1u3_0_0.vlm"},
	{0x7ECB6CFEC114A7D2ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_hvyn.msh"},
	{0x7ECBF18658DC900EULL, "vxca_cavern1_haunted_j2r_0_0.vlm"},
	{0x7ECC89D165C0F132ULL, "423823516.ogg"},
	{0x7ECCCE059BB13EB3ULL, "gentr_autosave.utt"},
	{0x7ECCD68F89CFDEC5ULL, "fxa_circle_out.mao"},
	{0x7ECCFDB25464060BULL, "mrl150_merrill6071943.cl"},
	{0x7ECDD0E125FF6EADULL, "w_woundedcoast_239_0.mmh"},
	{0x7ECDD4E125FF75FEULL, "w_woundedcoast_239_0.mao"},
	{0x7ECDE2E125FF8D33ULL, "w_woundedcoast_239_0.msh"},
	{0x7ECE09FAE15212A8ULL, "830983061.ogg"},
	{0x7ECE97896656E2B8ULL, "6125107_m.fxe"},
	{0x7ECEB43BA5D75DE6ULL, "vxdr_primevaltomb_3iv_0_0.vlm"},
	{0x7ECEFB9F343C2F01ULL, "art\\characters\\creatures\\strider\\animation\\turnright90.ani"},
	{0x7ECF5B4F19628A9DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3.ani"},
	{0x7ECF6E3FA276F2BDULL, "hlac_lt_darktown_-1_57y.dds"},
	{0x7ECFE1A1EE763606ULL, "hlac_w_encounter2_3_0y.dds"},
	{0x7ED0066711E08757ULL, "6097285_m.fxe"},
	{0x7ED056974B5CD058ULL, "vxlt_gallowstemplar_34y_0_0.vlm"},
	{0x7ED0684488AD66E7ULL, "hlac_ht_chantry_2534_2z.dds"},
	{0x7ED0A776D0C3742DULL, "dae000_worthy6166635.cl"},
	{0x7ED0FFBC02A7A929ULL, "vxw_encounter2_camp_j3_0_0.vlm"},
	{0x7ED1579D8518A2E4ULL, "909126551.ogg"},
	{0x7ED15AF1F00C55A7ULL, "6043969_m.fxe"},
	{0x7ED1E43BD6848939ULL, "vxht_0_hightown_a2_caq_0_1.vlm"},
	{0x7ED22DF75C3E5535ULL, "6065659_m.fxe"},
	{0x7ED2310EC961037EULL, "hlac_w_sundermount_-393213_0y.dds"},
	{0x7ED27D39C3D6AB47ULL, "hlac_lt_0_lowtown_docks_a3_-262144_0z.dds"},
	{0x7ED3B88E9439CB35ULL, "fx_markofdeath.dds"},
	{0x7ED3C68E67E00EDAULL, "hlac_w_bonepitmining_a2_65542_0x.dds"},
	{0x7ED3CAAC790A6EF6ULL, "894197176.ogg"},
	{0x7ED41B4D82FF875AULL, "vxw_sundermount_31o_0_0.vlm"},
	{0x7ED42D459C311225ULL, "255319561.ogg"},
	{0x7ED492A37A1F058EULL, "168212952.ogg"},
	{0x7ED4F859FA6D0ADEULL, "6036551_m.fxe"},
	{0x7ED5069F3440ADA3ULL, "art\\characters\\creatures\\strider\\animation\\turnright90.gad"},
	{0x7ED51F34C4A54AC5ULL, "vxlt_0_lowtown_a2_ev7_0_0.vlm"},
	{0x7ED56CE337C415AEULL, "125207281.ogg"},
	{0x7ED59FF033556CC1ULL, "qun200_party_comment6202022.cl"},
	{0x7ED6055A24C1AC5CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs102a_0t.dds"},
	{0x7ED6460A07F45271ULL, "llac_ht_mansion2_818_30.dds"},
	{0x7ED7DDFC2C40E7FAULL, "vxht_keep_n_6m5_0_0.vlm"},
	{0x7ED8C44BF942DF0FULL, "gen00fl_carver6108912.cl"},
	{0x7ED96A1ACDF24F84ULL, "6079296_f.fxe"},
	{0x7EDA3CF57AD4E7C0ULL, "6167282_m.fxe"},
	{0x7EDA7B63B185D5BAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5_0.pwk"},
	{0x7EDAAB1ECC3A6FFAULL, "hlac_dr_0_deeproads_5840_4y.dds"},
	{0x7EDAF0D1659B993AULL, "vxdr_primevaltomb_3d9_0_0.vlm"},
	{0x7EDBD5305102028CULL, "fex_spt_sd.msh"},
	{0x7EDBDB3051020D7EULL, "fex_spt_sd.mmh"},
	{0x7EDBE39F34469088ULL, "art\\characters\\creatures\\strider\\animation\\turnright90.evt"},
	{0x7EDC0EA3815C45AAULL, "vxht_mansion2b_1rh_0_0.vlm"},
	{0x7EDCAFFD7CC283A5ULL, "220776706.ogg"},
	{0x7EDD449D1095842CULL, "vxw_encounter3_b1_0_0.vlm"},
	{0x7EDE61F04FE1A762ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_torturerack_d.dds"},
	{0x7EDE6CAF398890C6ULL, "695219644.ogg"},
	{0x7EDE72B240F87E38ULL, "6173333_m.fxe"},
	{0x7EDEBD11F4082122ULL, "static_static_p47.psh.fxo.dx11"},
	{0x7EDEC8743D52E8D8ULL, "vxlt_gallowstemplar_31v_0_0.vlm"},
	{0x7EDEE0917E6136C2ULL, "llac_lt_0_lowtown_docks_131072_0.dds"},
	{0x7EDEE81770CB8CCFULL, "hlac_lt_0_lowtown_docks_a3_n_65534_5y.dds"},
	{0x7EDEEDEADCCF55F5ULL, "vxlt_0_lowtown_a2_n_c_hvc_0_0.vlm"},
	{0x7EDF3D3435F1AFB2ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_1_ian.gad"},
	{0x7EDF5D0F91A13AA7ULL, "756167636.ogg"},
	{0x7EDF7198F49DC1A9ULL, "vxw_encounter2_camp_e2_0_0.vlm"},
	{0x7EDFE5CD7B0EEFCEULL, "6150905_m.fxe"},
	{0x7EE0E203E0B84B06ULL, "754791113.ogg"},
	{0x7EE10939C49472BCULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_2x.dds"},
	{0x7EE140CC2784E2ACULL, "85808519.ogg"},
	{0x7EE18F67E7E2CA4FULL, "vxdr_primevaltomb_2gd_0_0.vlm"},
	{0x7EE25B00133A82DDULL, "265616508.ogg"},
	{0x7EE266AFF48181F4ULL, "6068539_f.fxe"},
	{0x7EE4BFA5A2AA49E4ULL, "lgt100cr_captain_qerth.utc"},
	{0x7EE4D9ED16941BC5ULL, "one000_ilen6057349.cl"},
	{0x7EE5BECCF9666368ULL, "hm_amb_keep_guard3.mor"},
	{0x7EE624093839C4E3ULL, "hlac_lt_undercityserialkiller_-1_12x.dds"},
	{0x7EE64FD91E37007AULL, "vxht_keep_n_6h6_0_0.vlm"},
	{0x7EE6BB232286DE26ULL, "mag311_emile_de_launcet6064727.cl"},
	{0x7EE732EAD65A48DDULL, "59829501.ogg"},
	{0x7EE76B0A1AD4C5BDULL, "vxw_encounter2_jg_0_0.vlm"},
	{0x7EE78420F681A6CEULL, "vxw_encounter2_3k_0_0.vlm"},
	{0x7EE79B018FAC9A30ULL, "6086184_f.fxe"},
	{0x7EE7DCF51065E014ULL, "mcr370_meredith6136117.cl"},
	{0x7EE7F8246BD63D18ULL, "vxlt_gallowstemplar_f_2lk_0_0.vlm"},
	{0x7EE8A34F3E43B796ULL, "6070614_m.fxe"},
	{0x7EE8D43A0A04953CULL, "314962278.ogg"},
	{0x7EE8EE5DE7A0EF40ULL, "vxca_caverns1_ghf_0_0.vlm"},
	{0x7EE96ECCEB515869ULL, "vxht_brothel_n_5k4_0_0.vlm"},
	{0x7EE9BA2FC13B2FC3ULL, "art\\vfx\\creatures\\fxc_bm_vom_c.anb"},
	{0x7EEA0A3A5B638DB8ULL, "vxlt_0_lowtown_docks_n_bcl_0_0.vlm"},
	{0x7EEA13830641799BULL, "vxca_cavern1_haunted_e4r_0_0.vlm"},
	{0x7EEA3E625115B38AULL, "311876464.ogg"},
	{0x7EEA808073525E2AULL, "vxht_mansion2b_1ui_0_0.vlm"},
	{0x7EEB0ECFDAF6B86EULL, "136026321.ogg"},
	{0x7EEBF4B62E244A26ULL, "ave121_brennan6105900.cl"},
	{0x7EEC8B139FB3BF29ULL, "6109045_m.fxe"},
	{0x7EECAB5E24479977ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lrgchair_d.dds"},
	{0x7EED298C659393CCULL, "gen00fl_merrill_act1_pt2.stg"},
	{0x7EEDBC55C8772D0BULL, "ser221_gamlen6134818.cl"},
	{0x7EEDC12F292751E4ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumtile_d.dds"},
	{0x7EEE29567251929AULL, "hlac_lt_shop_-1_6z.dds"},
	{0x7EEE5D28994D208CULL, "635957975.ogg"},
	{0x7EEE625A9EF94EA5ULL, "qcr000_aveline6041089.cl"},
	{0x7EEE911027C8E608ULL, "gen_im_trs_gem_jade_04.uti"},
	{0x7EEED18A8BB2128BULL, "6074444_m.fxe"},
	{0x7EEEFA9321C457CBULL, "6041022_m.fxe"},
	{0x7EEF2B4D859AFB0FULL, "596271405.ogg"},
	{0x7EEF334D9271BA97ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlarge.msh"},
	{0x7EEF454D9271D9FAULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlarge.mao"},
	{0x7EEF494D9271E009ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlarge.mmh"},
	{0x7EEFC0993B80FFDDULL, "6143511_m.fxe"},
	{0x7EF16451445FDAA3ULL, "ser121st_jethann.stg"},
	{0x7EF1A5C3FE051B7DULL, "vxlt_0_lowtown_docks_a3_n_d1z_0_0.vlm"},
	{0x7EF1BDFB6B4E226BULL, "6060809_m.fxe"},
	{0x7EF1F1399D38EC0AULL, "vxw_bonepitdestroyed_a3_1az_0_0.vlm"},
	{0x7EF29CC1E8500BBFULL, "498713207.ogg"},
	{0x7EF2B0B9819C2F36ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.cess.evt"},
	{0x7EF2D18E9454A63FULL, "fx_markofdeath.mao"},
	{0x7EF3283436023558ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_1_ian.ani"},
	{0x7EF3523150BC720EULL, "vxht_brothel_n_75b_0_1.vlm"},
	{0x7EF3826E8EFF1625ULL, "vxht_chantry_n_28m_0_0.vlm"},
	{0x7EF39595F52F62B2ULL, "dae000_orsino6085507.cl"},
	{0x7EF399B10F378394ULL, "162264980.ogg"},
	{0x7EF3BC356F3A06CFULL, "vxht_playershouse_n_1le_0_0.vlm"},
	{0x7EF3D5BFD874DA31ULL, "dae000_arishok6096323.cl"},
	{0x7EF3EFEC5EF77C39ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabexit.ani"},
	{0x7EF401C61BC9062CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge03_dp.dds"},
	{0x7EF45D3DCCD967F1ULL, "vfx_addcloud02.dds"},
	{0x7EF468CB8363560FULL, "gen00fl_varric6033586.cl"},
	{0x7EF4969CBC72C388ULL, "vxlt_0_lowtown_docks_btv_0_0.vlm"},
	{0x7EF4C1B6102D18FAULL, "vxht_keep_n_6c7_0_0.vlm"},
	{0x7EF52AF18CDA1841ULL, "hlac_ht_playershouse_n_855_33z.dds"},
	{0x7EF531B6180A879FULL, "558283565.ogg"},
	{0x7EF56CD8BD26ADA8ULL, "mcr310_argument6072285.cl"},
	{0x7EF5816D706569F6ULL, "vxca_cavern1_haunted_fn2_0_0.vlm"},
	{0x7EF5DF1BECAF5C37ULL, "6126704_f.fxe"},
	{0x7EF67CFDE52C2BABULL, "vxht_mansion2_1sv_0_0.vlm"},
	{0x7EF6A9D4302B2088ULL, "6139909_f.fxe"},
	{0x7EF6D51BFE97FB7AULL, "hlac_lt_0_lowtown_docks_196607_0y.dds"},
	{0x7EF6FF252285D5F1ULL, "hlac_w_encounter3_-131068_0x.dds"},
	{0x7EF724C6833FEE28ULL, "vxlt_0_lowtown_a2_n_c_hl8_0_0.vlm"},
	{0x7EF72DA5A705A861ULL, "vxlt_undercitydungeon_1fj_0_1.vlm"},
	{0x7EF7972FC14731ECULL, "art\\vfx\\creatures\\fxc_bm_vom_c.mmh"},
	{0x7EF7D094AE0F750EULL, "182436413.ogg"},
	{0x7EF81BF985C194B3ULL, "6093555_m.fxe"},
	{0x7EF875F8294531B6ULL, "hlac_dr_0_deeproads_5843_8x.dds"},
	{0x7EF88E753E4042EAULL, "vxw_encounter2_camp_qy_0_0.vlm"},
	{0x7EF8D43BF4525646ULL, "6122367_f.fxe"},
	{0x7EF8F96E00BE5C7BULL, "vxlt_gallowscourtyard_4xc_0_0.vlm"},
	{0x7EF95E76824C722EULL, "ser221_party_comments6164310.cl"},
	{0x7EF970EF35A7B4CEULL, "6182713_m.fxe"},
	{0x7EF9A7B981A23EC1ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.cess.gad"},
	{0x7EF9B8A37BC4F2A8ULL, "195530043.ogg"},
	{0x7EFA712CCA79D392ULL, "fen210_fenris6108713.cl"},
	{0x7EFA8F8F26D902CBULL, "hlac_lt_0_lowtown_a3_n_c_-65536_8x.dds"},
	{0x7EFACAEC5EFD5C7BULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabexit.gad"},
	{0x7EFAFCF091A9D7A9ULL, "6102145_m.fxe"},
	{0x7EFB5F1857745DE1ULL, "vxht_mansion_n_627_0_0.vlm"},
	{0x7EFC06D2FC64EFEDULL, "6103205_m.fxe"},
	{0x7EFC0CFB02EE55C0ULL, "vxdr_primevaltomb_34r_0_1.vlm"},
	{0x7EFC19E09D4C75EFULL, "499950982.ogg"},
	{0x7EFC3100EA49CC7EULL, "779586718.ogg"},
	{0x7EFC7D6923E39817ULL, "vxdr_primevaltomb_81y_0_1.vlm"},
	{0x7EFC8858717B4F37ULL, "6118169_f.fxe"},
	{0x7EFCD81EF893A7AFULL, "framebuffereffects_cocmask_p0.psh.fxo"},
	{0x7EFD2DF1290AF5C7ULL, "dae000_osric.bnk"},
	{0x7EFDD45B57736A75ULL, "6163841_m.fxe"},
	{0x7EFE4042B119DBA9ULL, "vxlt_0_lowtown_docks_cow_0_0.vlm"},
	{0x7EFE49B6486F90A2ULL, "art\\levels\\ztools\\default\\generic\\9ff_bed.phy"},
	{0x7F004FF1290D653FULL, "dae000_osric.cnv"},
	{0x7F00A8B981A85F6FULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.cess.ani"},
	{0x7F00E7EC5F01F960ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabexit.evt"},
	{0x7F0144B0F80E2D2CULL, "vxht_0_hightown_n_bx8_0_0.vlm"},
	{0x7F0160F69AAE336FULL, "vxdr_0_deeproads_f65_0_0.vlm"},
	{0x7F01725A2647D2BAULL, "6147673_m.fxe"},
	{0x7F017516DDB61866ULL, "mag200pt_nt_ambient.ncs"},
	{0x7F019CBA4237E328ULL, "6108040_f.fxe"},
	{0x7F01C3028A048B63ULL, "gen00fl_varric6063837.cl"},
	{0x7F01EBF75A4F4690ULL, "vxdr_primevalentrance_24w_0_1.vlm"},
	{0x7F01F44B80F52EA5ULL, "vxht_chantry_n_23l_0_0.vlm"},
	{0x7F0287E2D9C574B1ULL, "lt_0_lowtown_docks_146_0.mmh"},
	{0x7F0293E2D9C58952ULL, "lt_0_lowtown_docks_146_0.mao"},
	{0x7F02A1E2D9C5A16FULL, "lt_0_lowtown_docks_146_0.msh"},
	{0x7F02D590E33C6A1DULL, "vxht_chantry_32y_0_0.vlm"},
	{0x7F03475701590476ULL, "and111_lirene6084372.cl"},
	{0x7F038F28A4700535ULL, "hlac_ca_caverns2prc_5802_9z.dds"},
	{0x7F03CA3D1A7C878FULL, "vxw_sundermount_2bz_0_0.vlm"},
	{0x7F03D297F917EE2BULL, "6148138_m.fxe"},
	{0x7F040B54EBB50D92ULL, "l_and231ar_gallowsbasement.bnk"},
	{0x7F04299B1A560A3EULL, "244415073.ogg"},
	{0x7F044EC3FEC0F6BDULL, "vxw_encounter2_pe_0_0.vlm"},
	{0x7F0467DADA6DD7CEULL, "vxw_encounter2_9m_0_0.vlm"},
	{0x7F04E7CE18C909C8ULL, "vxht_mansion_n_622_0_0.vlm"},
	{0x7F05497D142E42F0ULL, "6067317_m.fxe"},
	{0x7F0558FF1529969CULL, "follower_banter6136340.cl"},
	{0x7F057F891F1BF52FULL, "1023227521.ogg"},
	{0x7F059F8298FBC0E1ULL, "vxlt_undercitydungeon_1ik_0_1.vlm"},
	{0x7F05AD5236DD9D65ULL, "6148482_m.fxe"},
	{0x7F05B6BBEB8403E9ULL, "rdr111dg_martin6054969.cl"},
	{0x7F05F7EFB92AC7F6ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn4plank.msh"},
	{0x7F05F8D799236CD1ULL, "w_bonepitdestroyed_a3_94_0.mmh"},
	{0x7F0601EFB92AD8F4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn4plank.mmh"},
	{0x7F0604D799238172ULL, "w_bonepitdestroyed_a3_94_0.mao"},
	{0x7F0605EFB92ADF07ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn4plank.mao"},
	{0x7F0612D79923988FULL, "w_bonepitdestroyed_a3_94_0.msh"},
	{0x7F065C67999919FBULL, "gen00fl_anders6121225.cl"},
	{0x7F06D20A1547846DULL, "gen00fl_anders6125483.cl"},
	{0x7F07005230365B6AULL, "vxw_encounter2_camp_lz_0_0.vlm"},
	{0x7F07A67E27F79F64ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldc.phy"},
	{0x7F084B5EF3D5DEB3ULL, "6076569_m.fxe"},
	{0x7F08990985149F5AULL, "128390131.ogg"},
	{0x7F08FCB2E3AB8CA4ULL, "llac_ht_0_hightown_a2_65536_0.dds"},
	{0x7F0929DE5DE363CCULL, "mrl150_merrill6070800.cl"},
	{0x7F09479BB4963D21ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_hole_1.mao"},
	{0x7F0A05ECE22BC032ULL, "6080266_m.fxe"},
	{0x7F0A078833617286ULL, "vxht_keep_7ag_0_0.vlm"},
	{0x7F0A11F600DF419AULL, "mag311_emile_de_launcet6043760.cl"},
	{0x7F0A239929A96A19ULL, "art\\vfx\\creatures\\fxc_drg_brth_e_c.impa.ani"},
	{0x7F0AD9C9CFD31A75ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_walltypexa.phy"},
	{0x7F0B6D26E49E6A59ULL, "gen_im_trs_gem_diamond_01.uti"},
	{0x7F0B96D0A48F1A66ULL, "dae000_fem_commoner2_mob6174584.cl"},
	{0x7F0BCA96570AFE41ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_visa_d.dds"},
	{0x7F0C0D1E0105A5FCULL, "hlac_ca_caverns1_mines_24422_6z.dds"},
	{0x7F0CBA8C90C356B1ULL, "vxht_brothel_n_6f4_0_0.vlm"},
	{0x7F0CBD2E677D28FAULL, "199778946.ogg"},
	{0x7F0D67DA308918F8ULL, "hlac_lt_gallowstemplar_n_2394_4y.dds"},
	{0x7F0D8B767B43CC49ULL, "llac_w_sundermount_65540_0.dds"},
	{0x7F0E1515BB0FBF05ULL, "mag311_keran6041505.cl"},
	{0x7F0E1C39C1C7215BULL, "323512557.ogg"},
	{0x7F0E257D99629BD7ULL, "vxlt_warehouse_n_p1_0_0.vlm"},
	{0x7F0ED4F381251D0AULL, "vxw_bonepitdestroyed_a3_1ot_0_0.vlm"},
	{0x7F0EEF6B8CFE7135ULL, "689293813.ogg"},
	{0x7F0F6FF5971B359EULL, "vxw_sundermount_344_0_0.vlm"},
	{0x7F0F788EB66E0B40ULL, "6072171_m.fxe"},
	{0x7F0FB68DEA0445ACULL, "vxht_0_hightown_n_bc9_0_0.vlm"},
	{0x7F104250518AD10FULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.mmh"},
	{0x7F1066BA56816B02ULL, "dae000cr_arishok.ncs"},
	{0x7F10AAB8F8C26B30ULL, "art\\characters\\playercharacter\\humanfemale\\mt_hed_flma.mao"},
	{0x7F10C708EA7F7D4DULL, "zz_dae_debug6049971.cl"},
	{0x7F10FE9929AF4A5BULL, "art\\vfx\\creatures\\fxc_drg_brth_e_c.impa.gad"},
	{0x7F1167E08483F372ULL, "sky_ca_cave.mao"},
	{0x7F117EAF4AC160B2ULL, "hlac_lt_gallowscourtyard_65534_1x.dds"},
	{0x7F11C68DBB603684ULL, "hlac_dr_primevaltomb_658_9z.dds"},
	{0x7F120505EA7B279EULL, "t1_mul_wh3.tnt"},
	{0x7F121A70B35578BAULL, "vxht_brothel_n_4w6_0_0.vlm"},
	{0x7F12712FC9863E2EULL, "216773323.ogg"},
	{0x7F12A9BA1886B02DULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgta.msh"},
	{0x7F12AE98A5149C25ULL, "follower_banter6160771.cl"},
	{0x7F12B7BA1886C860ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgta.mao"},
	{0x7F12BBBA1886CEB3ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgta.mmh"},
	{0x7F12C2D5D09B8D37ULL, "6128504_f.fxe"},
	{0x7F12D87FCC3EF3CEULL, "vxht_keep_n_63b_0_1.vlm"},
	{0x7F12E55D7B79EE41ULL, "6104012_f.fxe"},
	{0x7F136779DFC22D3EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_hv_dmg.ani"},
	{0x7F13BAA2F1885A76ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3_up.evt"},
	{0x7F13EECFAB2FD891ULL, "133570752.ogg"},
	{0x7F14283F7C8947CAULL, "llac_lt_foundry_n_320_3.dds"},
	{0x7F1434F73B26DB1AULL, "vxht_brothel_n_616_0_0.vlm"},
	{0x7F144D0C6E4CE385ULL, "art\\vfx\\materials\\vfx_blood_whispy.mao"},
	{0x7F149F27062AD83BULL, "vxht_0_hightown_n_c46_0_0.vlm"},
	{0x7F14CBC8B5E3B317ULL, "serev_item_acquired.ncs"},
	{0x7F14EBA99F74FD13ULL, "6176618_m.fxe"},
	{0x7F155FD13145D738ULL, "vxlt_0_lowtown_docks_n_brk_0_0.vlm"},
	{0x7F1599940F6EFC7BULL, "llac_ht_0_hightown_n_65538_0.dds"},
	{0x7F15A985E756A506ULL, "hlac_dr_dungeon_3217_3x.dds"},
	{0x7F16B4CD4A938EB7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_sadasdasdasd.msh"},
	{0x7F16C6CD4A93AD1AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_sadasdasdasd.mao"},
	{0x7F16C93240381B47ULL, "vxht_0_hightown_a3_aua_0_0.vlm"},
	{0x7F16CACD4A93B429ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_sadasdasdasd.mmh"},
	{0x7F16ECEBD5FB9D60ULL, "vxht_bartrand_n_625_0_0.vlm"},
	{0x7F173CDAB20C080BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_rfoot.evt"},
	{0x7F175065C4BE2AE5ULL, "vrc321_spooky_scenes.bnk"},
	{0x7F17E33FD49811B2ULL, "llac_lt_0_lowtown_docks_a3_131070_1.dds"},
	{0x7F17EA7CDA3B9266ULL, "plt_zz_lgtpt_3_debug.nss"},
	{0x7F181B9929B59A40ULL, "art\\vfx\\creatures\\fxc_drg_brth_e_c.impa.evt"},
	{0x7F18CA6B28A011DCULL, "vxlt_0_lowtown_docks_a3_cyk_0_0.vlm"},
	{0x7F19038B661CA18CULL, "99769783.ogg"},
	{0x7F1989867B5C5330ULL, "vxca_caverns1_e93_0_0.vlm"},
	{0x7F19B1A2F18CB601ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3_up.gad"},
	{0x7F1AE2D7A06A2CFCULL, "vxlt_0_lowtown_a2_hfs_0_0.vlm"},
	{0x7F1B1690C6419C8FULL, "dae000_arishok6078466.cl"},
	{0x7F1B23FC95060CA9ULL, "vxlt_0_lowtown_docks_cuu_0_0.vlm"},
	{0x7F1B3265C4C1E01DULL, "vrc321_spooky_scenes.cnv"},
	{0x7F1B465B32DC5D08ULL, "vxca_caverns1_hzo_0_0.vlm"},
	{0x7F1B53BAE8553017ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_enter.gad"},
	{0x7F1B96DE32E3D654ULL, "755438262.ogg"},
	{0x7F1BCC39D5470B65ULL, "vxw_sundermount_36q_0_0.vlm"},
	{0x7F1BD1C525C0D9DDULL, "hlac_dr_0_deeproads_5838_3x.dds"},
	{0x7F1C099FEC8E951BULL, "hlac_ht_keep_3732_5x.dds"},
	{0x7F1C0B385EE47387ULL, "llac_ca_caverns1_mines_24421_11.dds"},
	{0x7F1C975A8B58B457ULL, "vxlt_warehouse_n_w0_0_0.vlm"},
	{0x7F1D25425ED0FA5EULL, "vxca_caverns1_mines_gi2_0_0.vlm"},
	{0x7F1DE1D289114E1EULL, "vxw_sundermount_335_0_0.vlm"},
	{0x7F1E4BA4E4B9EB3FULL, "hlac_lt_0_lowtown_docks_a3_262143_0z.dds"},
	{0x7F1E55DAB21251CCULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_rfoot.gad"},
	{0x7F1EFF5FB20783A4ULL, "6025157_m.fxe"},
	{0x7F1F01FF83312828ULL, "w_bonepit_28_0_lc.mmh"},
	{0x7F1F0B3A32A75A76ULL, "6149705_m.fxe"},
	{0x7F1F56999E4B07F8ULL, "916233084.ogg"},
	{0x7F1F9852C329AD8BULL, "zz_gallows.ncs"},
	{0x7F1FEC3392550D08ULL, "vxlt_0_lowtown_docks_beu_0_0.vlm"},
	{0x7F205CA0751EB240ULL, "563601683.ogg"},
	{0x7F207F79DFCCE00FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_hv_dmg.evt"},
	{0x7F20B2A2F192D6AFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v3_up.ani"},
	{0x7F20D5B4634E7A44ULL, "vxht_keep_n_60p_0_0.vlm"},
	{0x7F20F670243E8286ULL, "zz_followers_banter6159369.cl"},
	{0x7F221EAFE8B0F01BULL, "6173969_m.fxe"},
	{0x7F22625868D25422ULL, "283675210.ogg"},
	{0x7F2268BAE85B72A4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_enter.evt"},
	{0x7F22C130FADFFD4EULL, "vxw_sundermount_2b7_0_0.vlm"},
	{0x7F22F9EF3E38B529ULL, "6177277_m.fxe"},
	{0x7F2350D61D46C201ULL, "257803965.ogg"},
	{0x7F2361894C23FD3DULL, "6100059_m.fxe"},
	{0x7F2362C17CDDC2F6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobospikexa.mao"},
	{0x7F236EC17CDDD695ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobospikexa.mmh"},
	{0x7F2370C17CDDDA0BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobospikexa.msh"},
	{0x7F23E40C14228C6AULL, "vxw_encounter2_camp_bt_0_0.vlm"},
	{0x7F2424B648902156ULL, "art\\levels\\ztools\\default\\generic\\9ff_bed.mao"},
	{0x7F2430B6489035F5ULL, "art\\levels\\ztools\\default\\generic\\9ff_bed.mmh"},
	{0x7F2432B64890396BULL, "art\\levels\\ztools\\default\\generic\\9ff_bed.msh"},
	{0x7F24834F0E50A65FULL, "vxht_0_hightown_b26_0_0.vlm"},
	{0x7F24973F545CC9E0ULL, "mag320_nobleman_2.cnv"},
	{0x7F24A87AF7C0B1A3ULL, "vxlt_shop_cv_0_0.vlm"},
	{0x7F2528754A36D655ULL, "vxlt_undercityserialkiller_1a3_0_0.vlm"},
	{0x7F2554DAB2186ECAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_rfoot.ani"},
	{0x7F25B2C04B342EAEULL, "257116134.ogg"},
	{0x7F25C1CE3D7FA410ULL, "framebuffereffects_downsample_p0.psh.fxo"},
	{0x7F25DF1B7417FBCAULL, "8600127.ogg"},
	{0x7F26319308AA08E6ULL, "vxht_keep_n_5xa_0_0.vlm"},
	{0x7F267E44CBB12C8AULL, "391983467.ogg"},
	{0x7F2682957ED4639FULL, "gen00fl_aveline6128531.cl"},
	{0x7F26BC9BD8E432DFULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em_eyesm1.mao"},
	{0x7F2727415FDD1A2FULL, "isa211_isabela6202293.cl"},
	{0x7F278079DFD300B8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_hv_dmg.gad"},
	{0x7F27A2C0397718D6ULL, "dae000_gamlen6048492.cl"},
	{0x7F27BA6B899ADF7EULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_lf.ani"},
	{0x7F2868F0E4BAC3E8ULL, "gen_im_fla_carver_03.uti"},
	{0x7F28793F54607F18ULL, "mag320_nobleman_2.bnk"},
	{0x7F28B374F31FC2FAULL, "qcr220_grey_wardens6135950.cl"},
	{0x7F28D3A983D551E0ULL, "6103024_m.fxe"},
	{0x7F2954B49260457CULL, "vxlt_0_lowtown_a2_hcp_0_0.vlm"},
	{0x7F297607CE02F440ULL, "vxlt_0_lowtown_docks_a3_d3z_0_0.vlm"},
	{0x7F298B580A6BAF14ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamsml.dura.ani"},
	{0x7F2A040B53D3621CULL, "dae000_orana6149123.cl"},
	{0x7F2AC396F7680C0EULL, "vxw_encounter2_camp_ov_0_1.vlm"},
	{0x7F2B09377D4ECCD7ULL, "vxlt_warehouse_n_z3_0_0.vlm"},
	{0x7F2B42D0171A27D0ULL, "follower_banter6172861.cl"},
	{0x7F2BD6178207122BULL, "hlac_lt_0_lowtown_docks_-65539_4z.dds"},
	{0x7F2BF1C13D2AAC37ULL, "vxlt_warehouse_n_a2_0_0.vlm"},
	{0x7F2C2575934A1EECULL, "dae000_bodhan6146879.cl"},
	{0x7F2C26A0D0BAD32FULL, "vrc000pt_n_rival_mid.ncs"},
	{0x7F2C355C1EF7D5D6ULL, "lt_0_lowtown_a3_n_c_27_0_lc.mmh"},
	{0x7F2CFF7CDA4E11A1ULL, "plt_zz_lgtpt_3_debug.plo"},
	{0x7F2D4F67544B5C5AULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown4.msh"},
	{0x7F2D6167544B7AA0ULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown4.mmh"},
	{0x7F2D83A9371268CFULL, "vxht_playershouse_n_1pa_0_0.vlm"},
	{0x7F2DE8DB0443D659ULL, "gen00fl_varric6089646.cl"},
	{0x7F2DF44F11DAC887ULL, "gen00fl_aveline6121954.cl"},
	{0x7F2E031E79F20C91ULL, "llac_ca_cavern1_haunted_24425_2.dds"},
	{0x7F2E6588E1B60A6BULL, "vxlt_hangedman_n_27p_0_0.vlm"},
	{0x7F2EA1486A385EFFULL, "vxw_sundermount_2d8_0_0.vlm"},
	{0x7F2F80BAE866259DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_enter.ani"},
	{0x7F2FFA151142B7BAULL, "qun231_party_comment6077010.cl"},
	{0x7F302CE4D1AB4AD2ULL, "6208361_m.fxe"},
	{0x7F3059495AEF6171ULL, "ave331_jeven6138751.cl"},
	{0x7F3088580A71C976ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamsml.dura.gad"},
	{0x7F30AF92D7DAB7B6ULL, "fen110_fenris6096196.cl"},
	{0x7F30D815CDF2774DULL, "gen00fl_carver6104932.cl"},
	{0x7F30EBC6BEBC42CBULL, "dae200_cullen6140327.cl"},
	{0x7F310FF2FE84C2D0ULL, "lt_0_lowtown_a2_36_0_lc.mmh"},
	{0x7F312AE436A48B4CULL, "follower_banter6163197.cl"},
	{0x7F31BEE5810DABB3ULL, "fen121cr_shade.utc"},
	{0x7F31DFBCC23AD4CEULL, "mag120_macha6080618.cl"},
	{0x7F31EF46E3F16714ULL, "vxlt_0_lowtown_docks_a3_n_cs2_0_0.vlm"},
	{0x7F320B5B4A5A7ACEULL, "184622441.ogg"},
	{0x7F32278DC62F69B4ULL, "hlac_w_bonepitdestroyed_a3_196609_0y.dds"},
	{0x7F32635C0C43ACDDULL, "gen00fl_anders6125678.cl"},
	{0x7F32B6F593A9FB1DULL, "474123431.ogg"},
	{0x7F33009B3ADC973CULL, "llac_w_bonepit_196613_0.dds"},
	{0x7F331A57E9B6CA23ULL, "vxlt_shop_fy_0_0.vlm"},
	{0x7F332B73291C154BULL, "vxht_0_hightown_bfv_0_0.vlm"},
	{0x7F332DDCB6218B9DULL, "vxlt_blackemporium_4j_0_0.vlm"},
	{0x7F3334C8D577E1CBULL, "vxca_caverns2_8ir_0_0.vlm"},
	{0x7F3341BA66E74AFBULL, "art\\levels\\ztools\\default\\generic\\3wi_angle.mmh"},
	{0x7F334DBA66E75E98ULL, "art\\levels\\ztools\\default\\generic\\3wi_angle.mao"},
	{0x7F335FBA66E77DE5ULL, "art\\levels\\ztools\\default\\generic\\3wi_angle.msh"},
	{0x7F33D0A5B9E7CE60ULL, "vxht_bartrand_n_683_0_0.vlm"},
	{0x7F34D35F63A32E35ULL, "mcr371_hostile_merrill6082157.cl"},
	{0x7F34F372B50DC62FULL, "hlac_lt_0_lowtown_a2_131072_6z.dds"},
	{0x7F351B75371E4183ULL, "6036768_f.fxe"},
	{0x7F352A2DAE40136FULL, "464343197.ogg"},
	{0x7F353E1C840FB6E7ULL, "one300_zevran6076936.cl"},
	{0x7F35D26B89A7454FULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_lf.evt"},
	{0x7F36B85217E1FFA9ULL, "vxht_0_hightown_n_atf_0_0.vlm"},
	{0x7F36C9B6D2ABD851ULL, "252815040.ogg"},
	{0x7F36DE3B7BC57D68ULL, "hlac_w_sundermount_65532_0x.dds"},
	{0x7F371EB0CCF265DFULL, "zz_followers_banter6157306.cl"},
	{0x7F3747826414089EULL, "hlac_lt_gallowstemplar_n_2394_3z.dds"},
	{0x7F379C0B8B3D1F02ULL, "191682896.ogg"},
	{0x7F37A3580A7815D5ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamsml.dura.evt"},
	{0x7F37C69184565DFCULL, "vxlt_0_lowtown_a2_hhq_0_0.vlm"},
	{0x7F38435051AC254CULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.anb"},
	{0x7F3934466C7D8BEDULL, "mcr310_argument6072318.cl"},
	{0x7F3A129BD4C9341AULL, "art\\characters\\playercharacter\\humanfemale\\bdy_rob_maga.mao"},
	{0x7F3AA4B715812A48ULL, "l_lgt320ar_gang_hideout_ht.wbd"},
	{0x7F3B5CB8451174E3ULL, "vxlt_undercityencounter_zc_0_0.vlm"},
	{0x7F3BC08AF18744B7ULL, "hlac_ht_keep_3731_0z.dds"},
	{0x7F3BE525C1B1DD4DULL, "648810262.ogg"},
	{0x7F3C6F7FFD0FE4FAULL, "hlac_lt_0_lowtown_65536_6z.dds"},
	{0x7F3CD36B89AD65F8ULL, "art\\characters\\creatures\\mubari\\animation\\cwk_rdy_lf.gad"},
	{0x7F3CD765D3AC22EBULL, "vxlt_hangedman_n_22s_0_0.vlm"},
	{0x7F3D6684248FA1B9ULL, "424699263.ogg"},
	{0x7F3DC83B42AB57E9ULL, "472058375.ogg"},
	{0x7F3E2093026DA0A4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl_3.mao"},
	{0x7F3E6993F0DD8D3DULL, "vxw_blightlands_6l_0_0.vlm"},
	{0x7F4086F3F0130AA5ULL, "vxdr_primevaltomb_7u8_0_1.vlm"},
	{0x7F4143AF026646FFULL, "qun220cr_javaris6071706.cl"},
	{0x7F420DBC90B64A16ULL, "6076516_f.fxe"},
	{0x7F424C84A6B4A429ULL, "vxdr_primevalentrance_2hg_0_0.vlm"},
	{0x7F424DAB0CAEB8ADULL, "llac_lt_0_lowtown_docks_a3_-3_4.dds"},
	{0x7F429B9AECA88157ULL, "6084414_m.fxe"},
	{0x7F433428A309885CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_ctn180l.ani"},
	{0x7F433A534F74ADB0ULL, "gen00fl_sebastian6188116.cl"},
	{0x7F436181306F5692ULL, "361740094.ogg"},
	{0x7F436D448F2342DCULL, "gen00fl_aveline6139873.cl"},
	{0x7F4397E80E022008ULL, "vxht_keep_n_4hv_0_0.vlm"},
	{0x7F43B8DE29037DA3ULL, "fo_glow.mao"},
	{0x7F43B9730718F4B7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_3windows.phy"},
	{0x7F43F1F0F01E7A76ULL, "and100_anders6111678.cl"},
	{0x7F43F80C926D1E69ULL, "6156751_m.fxe"},
	{0x7F443058BD3A8AE5ULL, "vxdr_0_deeproads_crs_0_1.vlm"},
	{0x7F44853706E0C11BULL, "w_bonepitdestroyed_a3_1_0.msh"},
	{0x7F44973706E0E066ULL, "w_bonepitdestroyed_a3_1_0.mao"},
	{0x7F44A33706E0F405ULL, "w_bonepitdestroyed_a3_1_0.mmh"},
	{0x7F44CD31EF8B6796ULL, "6080482_m.fxe"},
	{0x7F45091119289F4FULL, "hlac_lt_0_lowtown_a3_n_c_65538_1z.dds"},
	{0x7F4513217664C8B3ULL, "vxw_blightlands_hp_0_0.vlm"},
	{0x7F458C63BF50B338ULL, "269267905.ogg"},
	{0x7F45A8D231E0C219ULL, "322586704.ogg"},
	{0x7F4655430E6E58ADULL, "734643893.ogg"},
	{0x7F46773D6BA09220ULL, "hlac_ca_caverns1_mines_24423_12x.dds"},
	{0x7F47543583C530F0ULL, "clf260_harley6173873.cl"},
	{0x7F478BA6C83A42F7ULL, "6052555_f.fxe"},
	{0x7F487EBEBA0F0418ULL, "hlac_lt_0_lowtown_a2_n_c_1_4y.dds"},
	{0x7F48C8A65D5FDC8FULL, "and121_karl6093999.cl"},
	{0x7F48D48F983B6BC4ULL, "6074506_m.fxe"},
	{0x7F48FF36B6D09DA9ULL, "452861742.ogg"},
	{0x7F490CCFD7A8F181ULL, "255136656.ogg"},
	{0x7F4916307F48F22BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magc_0n.dds"},
	{0x7F494128A30E099EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_ctn180l.gad"},
	{0x7F495C81B9566BF2ULL, "6119611_f.fxe"},
	{0x7F498D0459FD4D7DULL, "lt_0_lowtown_docks_32_0_lc.mmh"},
	{0x7F498F193F96CEC3ULL, "269051556.ogg"},
	{0x7F499AAEE42CAF5DULL, "289713995.ogg"},
	{0x7F4A71FF11BEE562ULL, "vxht_0_hightown_n_c7r_0_0.vlm"},
	{0x7F4ABBDB9A8843E9ULL, "gen_im_npc_arm_hum_cht_cirb.uti"},
	{0x7F4ABF2F46CB7066ULL, "vxlt_0_lowtown_docks_c4i_0_0.vlm"},
	{0x7F4AF15FB76BD21CULL, "gen00fl_isabela6079079.cl"},
	{0x7F4B43BA5DD742C1ULL, "789358195.ogg"},
	{0x7F4B45F6E9182536ULL, "gen00fl_aveline6122783.cl"},
	{0x7F4B65B6846F0E07ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredidol.gad"},
	{0x7F4BCA376353381FULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava_e.dds"},
	{0x7F4BFE760136730AULL, "art\\vfx\\environment\\vfx_trap_glow_p.cess.gad"},
	{0x7F4C4ECA65AAB8AEULL, "ser221_party_comments6209952.cl"},
	{0x7F4C7B39C33C2B5AULL, "art\\levels\\4cinematics\\ferelden\\letterfront.mao"},
	{0x7F4CB7D9662D759FULL, "llac_lt_0_lowtown_docks_a3_-65539_1.dds"},
	{0x7F4CD195754A4478ULL, "585766926.ogg"},
	{0x7F4CD229559CE1C0ULL, "6174413_m.fxe"},
	{0x7F4D498D2DF6955AULL, "genip_cin_kill_dagger.utp"},
	{0x7F4D4A54DA811074ULL, "6129812_f.fxe"},
	{0x7F4D75BBF182B764ULL, "vxlt_foundry_n_125_0_0.vlm"},
	{0x7F4DB397855CB347ULL, "lla_ht_0_hightown_a3_-65534_0.dds"},
	{0x7F4DD34747DA1157ULL, "285889407.ogg"},
	{0x7F4E2C0F802773C1ULL, "vxlt_gallowsprison_n_6pf_0_0.vlm"},
	{0x7F4EA12B465B9669ULL, "l_one101ar_ruin_int.wbd"},
	{0x7F4EBB4BB526D07AULL, "gen_im_npc_wep_one_swdc.uti"},
	{0x7F4EC304E29680E3ULL, "vxlt_gallowstemplar_3qy_0_0.vlm"},
	{0x7F5037577C39D69AULL, "vxht_0_hightown_a2_aia_0_0.vlm"},
	{0x7F50956763A52E56ULL, "hlac_ht_chantry_2534_13y.dds"},
	{0x7F50B374433D9700ULL, "240317902.ogg"},
	{0x7F51010546F62CC5ULL, "gen00fl_merrill6150706.cl"},
	{0x7F510258AEBD8722ULL, "hlac_lt_0_lowtown_a2_n_c_2_0z.dds"},
	{0x7F514F14F086CB50ULL, "hlac_lt_gallowscourtyard_-2_3x.dds"},
	{0x7F51A9F2A0465742ULL, "6135777_m.fxe"},
	{0x7F51DDC165B0246DULL, "6013276_m.fxe"},
	{0x7F51DEFBB6385991ULL, "9931296.ogg"},
	{0x7F51ECC9D00F3C90ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_walltypexa.msh"},
	{0x7F51FAC9D00F54AAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_walltypexa.mmh"},
	{0x7F51FEC9D00F5BE1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_walltypexa.mao"},
	{0x7F523756E9DDAF00ULL, "llac_lt_0_lowtown_docks_a3_n_65535_0.dds"},
	{0x7F527894C33BCEA0ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_manystrip_d.dds"},
	{0x7F5297788DD52DBAULL, "hlac_ca_caverns2prc_5802_32y.dds"},
	{0x7F52E176013C60A1ULL, "art\\vfx\\environment\\vfx_trap_glow_p.cess.evt"},
	{0x7F53A624013FABFBULL, "zz_vault_debug6178943.cl"},
	{0x7F53AAA1FEF70464ULL, "279509367.ogg"},
	{0x7F5485F7AF75591FULL, "45163585.ogg"},
	{0x7F551F46287B5DEFULL, "vxdr_0_deeproads_cwk_0_0.vlm"},
	{0x7F5625F91E792C39ULL, "mag120_macha6044541.cl"},
	{0x7F567531DE033BCEULL, "6161510-mrl_movedinwpc.cut"},
	{0x7F56D99A645B43E1ULL, "vxca_cave_1ng_0_0.vlm"},
	{0x7F56FE0EF1422373ULL, "6080836_m.fxe"},
	{0x7F57061032000375ULL, "vxw_sundermount_37o_0_1.vlm"},
	{0x7F57CE5EAB4C8472ULL, "6114111_f.fxe"},
	{0x7F57F0107AF821A2ULL, "zz_start_qcr.cnv"},
	{0x7F580DD87D078C56ULL, "hlac_lt_0_lowtown_a2_1_0x.dds"},
	{0x7F585FFDB07A3686ULL, "vxlt_gallowstemplar_n_55u_0_0.vlm"},
	{0x7F58C599ECEDA4CDULL, "vxlt_0_lowtown_a2_n_ewo_0_0.vlm"},
	{0x7F59080430C01527ULL, "vxca_caverns1_mines_edv_0_0.vlm"},
	{0x7F593C4A026C4921ULL, "llac_lt_undercitydungeon_-1_0.dds"},
	{0x7F59CB78F8E28A6FULL, "pro000_colonel_ewald6040618.cl"},
	{0x7F59DF8CC55C0B49ULL, "522232076.ogg"},
	{0x7F5A4244A1DC85F4ULL, "llac_lt_gallowscourtyard_-65538_2.dds"},
	{0x7F5A454281479537ULL, "m2da_base.gda"},
	{0x7F5A599F0E3FC619ULL, "236959377.ogg"},
	{0x7F5BDE8EFB9264A0ULL, "6129345_m.fxe"},
	{0x7F5BE47A91E8E19AULL, "227803831.ogg"},
	{0x7F5C4430CD45C4AFULL, "pro000_cut2d_thefirstyear6005322.cl"},
	{0x7F5C459932676119ULL, "vxlt_gallowstemplar_3r6_0_0.vlm"},
	{0x7F5C5890C234B6FBULL, "1006482477.ogg"},
	{0x7F5D35230140A749ULL, "336346271.ogg"},
	{0x7F5D47DAE43DE7A0ULL, "6174031_m.fxe"},
	{0x7F5D6C1B0DC13F0CULL, "vxlt_hangedman_1t6_0_0.vlm"},
	{0x7F5D99A1F17EDA5AULL, "vxw_bonepit_2a9_0_0.vlm"},
	{0x7F5D9A6B4A0451CBULL, "gen00fl_isabela6135684.cl"},
	{0x7F5E35418738290BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_legs_up_pose_loop.ani"},
	{0x7F5EE1A7C0B49403ULL, "oip_rock_04.mao"},
	{0x7F5F1D3D34CDECF8ULL, "vxht_bartrand_n_67r_0_0.vlm"},
	{0x7F5FD771516E644CULL, "qun131cr_warrens_thug3.utc"},
	{0x7F5FD9760146DDF0ULL, "art\\vfx\\environment\\vfx_trap_glow_p.cess.ani"},
	{0x7F6009B7C36DF34CULL, "synclargecreatures3.gda"},
	{0x7F6025B24244653FULL, "hlac_ht_mansion2_818_39y.dds"},
	{0x7F6052B684814A4DULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredidol.ani"},
	{0x7F605F35D9C96853ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower2_fenrintro.ani"},
	{0x7F60625025AFEFC8ULL, "vxca_cave_1nb_0_0.vlm"},
	{0x7F60628DCD9FA6A4ULL, "lt_0_lowtown_docks_a3_8_0_ll.mmh"},
	{0x7F60967C1CB71F16ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_deda_0n.dds"},
	{0x7F60BA849CCD0F9CULL, "llac_dr_0_deeproads_5840_3.dds"},
	{0x7F60DCDCC891692FULL, "llac_ht_keep_3732_1.dds"},
	{0x7F60DF97765B8386ULL, "llac_ca_caverns2_5802_1.dds"},
	{0x7F60E95F8690C7EBULL, "6114518_f.fxe"},
	{0x7F6160E378D1907AULL, "57452348.ogg"},
	{0x7F618A10F3B00ED1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_visa_t.dds"},
	{0x7F619948E3E053BDULL, "ip_restrict_player.ncs"},
	{0x7F61BAC620F1DC29ULL, "hlac_ht_mansion2b_818_13y.dds"},
	{0x7F61F6DB5A50F9B3ULL, "vxw_blightlands_vr_0_0.vlm"},
	{0x7F626D520F2A7E7BULL, "hf_amb_access.mor"},
	{0x7F6279B3031A2140ULL, "740267679.ogg"},
	{0x7F62B2676BA4A046ULL, "dre000_carver6037580.cl"},
	{0x7F632A865880C645ULL, "hlac_ht_mansion2_818_67x.dds"},
	{0x7F6361B795A3796CULL, "ca_caverns1_24423_ll.mmh"},
	{0x7F63D29C05951696ULL, "770963901.ogg"},
	{0x7F64766D482C887BULL, "676334097.ogg"},
	{0x7F64A1036B51F8AEULL, "vxlt_gallowscourtyard_n_66u_0_0.vlm"},
	{0x7F654862E0D259ABULL, "and100_anders6173581.cl"},
	{0x7F6577ED23F61BF5ULL, "vxw_sundermount_32l_0_1.vlm"},
	{0x7F657DF0C6BE5C43ULL, "6070379_f.fxe"},
	{0x7F657E5776B474B5ULL, "zz_mrl_debug6086560.cl"},
	{0x7F65E5B060C84BE6ULL, "vxdr_slavers_1je_0_0.vlm"},
	{0x7F65F2D834350A2BULL, "910056525.ogg"},
	{0x7F662A80FF759F8DULL, "vxlt_blackemporium_ck_0_0.vlm"},
	{0x7F66791DB2D27E05ULL, "593182254.ogg"},
	{0x7F66EF1C3C07BC80ULL, "112489300.ogg"},
	{0x7F6705FB0F7B93B5ULL, "vxlt_0_lowtown_docks_n_b79_0_0.vlm"},
	{0x7F673776DEE3BD4DULL, "vxlt_0_lowtown_a2_n_exn_0_0.vlm"},
	{0x7F6738AD16DEDECBULL, "198557244.ogg"},
	{0x7F673AAA56F1F9ECULL, "ht_0_hightown_57_0.rml"},
	{0x7F67D1731E24A624ULL, "hlac_lt_darktown_-1_84x.dds"},
	{0x7F68277C426B2426ULL, "isa311_isabela6116486.cl"},
	{0x7F68E5107B068225ULL, "zz_start_qcr.ncs"},
	{0x7F690AE33870D5BDULL, "plt_cod_hst_merchantguild.plo"},
	{0x7F692406FEE5224EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_ent.ani"},
	{0x7F6A01C952017F24ULL, "25391419.ogg"},
	{0x7F6A0FDC699F4D1DULL, "142233914.ogg"},
	{0x7F6A9092D32E682DULL, "vxht_0_hightown_7g6_0_0.vlm"},
	{0x7F6A97595E5DD8B6ULL, "6075301_m.fxe"},
	{0x7F6A98E0BA33D892ULL, "6048121_m.fxe"},
	{0x7F6AEC7BA39C73CAULL, "dae000_mother6127170.cl"},
	{0x7F6B5B05C6FB3BACULL, "vxht_0_hightown_n_9up_0_0.vlm"},
	{0x7F6BCB63EEF48211ULL, "vxlt_backalley_g_1ll_0_0.vlm"},
	{0x7F6BF524481FADFDULL, "vxlt_gallowstemplar_f_439_0_0.vlm"},
	{0x7F6D052D457CA360ULL, "592551841.ogg"},
	{0x7F6D38651152314CULL, "hlac_ht_0_hightown_a3_65536_1z.dds"},
	{0x7F6D63395AAF978BULL, "318540655.ogg"},
	{0x7F6D77CDDC61B74EULL, "zz_and_debug6134138.cl"},
	{0x7F6DB06761165EC4ULL, "gen00fl_anders6096478.cl"},
	{0x7F6E52B7D3E88E7DULL, "6132753_m.fxe"},
	{0x7F6E8ED7794A5FC2ULL, "hlac_w_sundermount_196607_0z.dds"},
	{0x7F6E9CBD7803618CULL, "vxht_mansion_n_5v5_0_0.vlm"},
	{0x7F6EA52922FFC926ULL, "vxca_caverns1_ii1_0_0.vlm"},
	{0x7F6EF5B531404BACULL, "hlac_ht_0_hightown_-65538_1z.dds"},
	{0x7F6F3CADF3706BEEULL, "vxw_woundedcoast_do_0_0.vlm"},
	{0x7F6F3CF5650E38ACULL, "719777486.ogg"},
	{0x7F6F5A72D4F54829ULL, "dae011_drunk2_sta6144330.cl"},
	{0x7F6F6849B9308C19ULL, "6095128_m.fxe"},
	{0x7F706F54A2823E85ULL, "6207565_m.fxe"},
	{0x7F70BB78497A0A45ULL, "128403829.ogg"},
	{0x7F70C216A71527D3ULL, "vxlt_0_lowtown_docks_a3_bua_0_0.vlm"},
	{0x7F70FFDF2FEB27C3ULL, "hlac_lt_warehouse_n_0_9x.dds"},
	{0x7F7119213F07BB05ULL, "1065656176.ogg"},
	{0x7F7154E7DBE3CC73ULL, "vxlt_0_lowtown_a2_n_fep_0_0.vlm"},
	{0x7F71669244442D17ULL, "qun160_vashothlead6053929.cl"},
	{0x7F71BC74B6D721C2ULL, "6118776_m.fxe"},
	{0x7F7203000C678EEFULL, "vxdr_0_deeproads_ciu_0_0.vlm"},
	{0x7F730CE1E9E4593EULL, "hlac_ht_mansion2b_818_30y.dds"},
	{0x7F731C41874A5A95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_legs_up_pose_loop.gad"},
	{0x7F73BC3DD0E41837ULL, "hlac_ht_0_hightown_65535_1y.dds"},
	{0x7F740FA2FC66BDF7ULL, "gen00fl_fenris6126352.cl"},
	{0x7F7412D62FF5076DULL, "character_character_p40.psh.fxo.dx11.assembly"},
	{0x7F743635D9D9CBFDULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_follower2_fenrintro.gad"},
	{0x7F74732B467C174BULL, "l_one101ar_ruin_int.bnk"},
	{0x7F74B5A3A2087B78ULL, "944186127.ogg"},
	{0x7F74D9203A8C424FULL, "609073068.ogg"},
	{0x7F7511C2CEEDFD55ULL, "ico_armor_light_4.dds"},
	{0x7F76205AAD822189ULL, "535800585.ogg"},
	{0x7F76F51A307ACB28ULL, "vxlt_warehouse_13o_0_0.vlm"},
	{0x7F771C06FEF1519FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_ent.evt"},
	{0x7F784C00DDFF74CCULL, "gen00fl_anders6119267.cl"},
	{0x7F787778CB24EC36ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.impa.ani"},
	{0x7F78B86817AB698BULL, "127805596.ogg"},
	{0x7F78CC9F0289FC97ULL, "511893897.ogg"},
	{0x7F79974021B1EB76ULL, "vxht_chantry_n_2uq_0_0.vlm"},
	{0x7F79D6107B14DC7CULL, "zz_start_qcr.utp"},
	{0x7F7A2E4D376EAF78ULL, "hlac_lt_0_lowtown_a2_n_c_131069_0z.dds"},
	{0x7F7A42CC6869C652ULL, "vxlt_0_lowtown_docks_a3_bud_0_0.vlm"},
	{0x7F7A545B53CA47F5ULL, "6018879_m.fxe"},
	{0x7F7A62BA67238530ULL, "art\\levels\\ztools\\default\\generic\\3wi_angle.phy"},
	{0x7F7BFD7606BA62BFULL, "vxlt_gallowsprison_n_6lt_0_0.vlm"},
	{0x7F7C14502648D1F5ULL, "6118717_m.fxe"},
	{0x7F7C42CDB0917F6AULL, "598737339.ogg"},
	{0x7F7CA997FE4D43DEULL, "art\\vfx\\combat\\vfx_m_spi_conj_c.impa.evt"},
	{0x7F7CD8643954219EULL, "ht_0_hightown_a2_53_0_ll.mmh"},
	{0x7F7D1F01FF3B7589ULL, "vxlt_0_lowtown_n_ghl_0_0.vlm"},
	{0x7F7DD251D15C1EC5ULL, "clf260_harley6088783.cl"},
	{0x7F7DFD06FEF73B88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_ent.gad"},
	{0x7F7E16BD8E595603ULL, "lt_0_lowtown_33_0_lc.mmh"},
	{0x7F7E3DBADA7CF0BBULL, "1033627243.ogg"},
	{0x7F7F37B6BC2D9F51ULL, "hlac_lt_0_lowtown_a2_n_0_8y.dds"},
	{0x7F7F98273D7165CAULL, "art\\vfx\\combat\\vfx_c_shadow_creature_c.anb"},
	{0x7F7F98C9D7B428A7ULL, "vxht_0_hightown_a2_bys_0_0.vlm"},
	{0x7F7FBCD8B4E62C5BULL, "1032336892.ogg"},
	{0x7F7FD5FC63F64063ULL, "follower_banter6166726.cl"},
	{0x7F7FE29E6D1815CDULL, "vxca_caverns1_mines_i1g_0_0.vlm"},
	{0x7F7FF128BEBB177AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmexit.gad"},
	{0x7F8074DCFE5DA76FULL, "vxdr_0_deeproads_cft_0_0.vlm"},
	{0x7F80D674C74E8239ULL, "209465339.ogg"},
	{0x7F815C83D2D92E15ULL, "394511116.ogg"},
	{0x7F816F688624316EULL, "hlac_w_encounter2_camp_-65537_0y.dds"},
	{0x7F8182F00C990BCFULL, "vxlt_0_lowtown_n_epa_0_1.vlm"},
	{0x7F81ADF1888D21B0ULL, "m_dae220ar_hightown.bnk"},
	{0x7F81CAB715BD6D0AULL, "l_lgt320ar_gang_hideout_ht.bnk"},
	{0x7F81E8E7BE58AB0EULL, "mer120_hubert6077286.cl"},
	{0x7F8251DDD20B9471ULL, "t3_ogr_b01.tnt"},
	{0x7F82D2EF6D6DF7EAULL, "lgt000cr_mage_turn_in_1.utc"},
	{0x7F82E05E9643D44AULL, "hlac_dr_0_deeproads_5845_4z.dds"},
	{0x7F830F2EECE08989ULL, "t3_wdg_gb2.tnt"},
	{0x7F8375665B094661ULL, "6118022_f.fxe"},
	{0x7F83C097FE538909ULL, "art\\vfx\\combat\\vfx_m_spi_conj_c.impa.gad"},
	{0x7F83E90EBB25DC58ULL, "1003824220.ogg"},
	{0x7F840C02676B65F1ULL, "narrative_varric_demo6170816.cl"},
	{0x7F8479F6D10A2D90ULL, "zz_followers_banter6158535.cl"},
	{0x7F84831F8895081FULL, "mage_horror.dds"},
	{0x7F84A3370716D80EULL, "w_bonepitdestroyed_a3_1_0.rml"},
	{0x7F84A62D929E8823ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.dura.ani"},
	{0x7F84E8D25C103B04ULL, "ser221_gamlen6121606.cl"},
	{0x7F84EF14F10AEC95ULL, "951744069.ogg"},
	{0x7F854203DE879562ULL, "vxht_0_hightown_a2_c1g_0_0.vlm"},
	{0x7F855303C4E0EBBDULL, "153539634.ogg"},
	{0x7F8566F72270E3A8ULL, "vxlt_warehouse_16l_0_0.vlm"},
	{0x7F8573789C7F4659ULL, "vxht_mansion_n_47f_0_0.vlm"},
	{0x7F85A7F1349D1E19ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_1.pwk"},
	{0x7F85DD032B27C480ULL, "6049627_m.fxe"},
	{0x7F85F428BEBF8791ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmexit.evt"},
	{0x7F867958B3BE0D75ULL, "vxw_woundedcoast_1gx_0_0.vlm"},
	{0x7F867F78CB313727ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.impa.evt"},
	{0x7F86A960EB774A90ULL, "hla_ht_0_hightown_a3_65534_0x.dds"},
	{0x7F872BCDE93A4525ULL, "hf_ruvena.mor"},
	{0x7F88091D13A803F6ULL, "vxht_chantry_n_2pr_0_0.vlm"},
	{0x7F88CD3D80D026D3ULL, "hlac_lt_0_lowtown_docks_-196609_1z.dds"},
	{0x7F88E0ED01F9C776ULL, "vxca_cavern1_haunted_ead_0_0.vlm"},
	{0x7F8A862D183EEA75ULL, "6117617_m.fxe"},
	{0x7F8A8E730754AE0FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_3windows.mao"},
	{0x7F8A9197FE595857ULL, "art\\vfx\\combat\\vfx_m_spi_conj_c.impa.ani"},
	{0x7F8A9A730754C2ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_3windows.mmh"},
	{0x7F8AA0730754CC9EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_3windows.msh"},
	{0x7F8ACB444B792084ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney2_d.dds"},
	{0x7F8BD261D08EF2D4ULL, "vxca_cavern1_haunted_gk3_0_1.vlm"},
	{0x7F8CA078CB35DB70ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.impa.gad"},
	{0x7F8CB3A8E5CBB49EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masc_0t.dds"},
	{0x7F8D33EDAFEC4FE5ULL, "162431079.ogg"},
	{0x7F8D45A95084E4BDULL, "6088773_m.fxe"},
	{0x7F8D5A730137DCB3ULL, "qun221_petrice6077348.cl"},
	{0x7F8D74CEB1359E85ULL, "vxht_bartrand_n_67k_0_0.vlm"},
	{0x7F8D90B0C93FE6DBULL, "vxlt_0_lowtown_docks_n_b02_0_0.vlm"},
	{0x7F8E299952C7A9CAULL, "6020567_m.fxe"},
	{0x7F8E3C38427B0F26ULL, "t3_bow_ash.tnt"},
	{0x7F8E71BEAFDD6483ULL, "vxw_encounter2_camp_l7_0_0.vlm"},
	{0x7F8EBE707E86289CULL, "531029578.ogg"},
	{0x7F8F2258A8B5604CULL, "6078656_m.fxe"},
	{0x7F8F8EFEC609CC05ULL, "40648562.ogg"},
	{0x7F8FA0273026B40CULL, "gen00fl_sebastian6188624.cl"},
	{0x7F8FBC17C7CD5057ULL, "qun110_petrice6059150.cl"},
	{0x7F8FE1BE3FAEDFBBULL, "6178465_m.fxe"},
	{0x7F903C3E6F469DB1ULL, "hlac_ht_chantry_n_2534_15y.dds"},
	{0x7F9044997D670D35ULL, "vxlt_0_lowtown_n_d5t_0_0.vlm"},
	{0x7F917993FFB63C4BULL, "1020782490.ogg"},
	{0x7F91B7526BAB222CULL, "vxht_0_hightown_n_b66_0_0.vlm"},
	{0x7F92BE2D92AAEE1AULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.dura.evt"},
	{0x7F92C181B6093CA1ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_deeproads_n.dds"},
	{0x7F9313192978B429ULL, "218610029.ogg"},
	{0x7F931C88EBBC4C7FULL, "437160787.ogg"},
	{0x7F93288D49383A60ULL, "gen00fl_isabela_init6076602.cl"},
	{0x7F934FD6F739E709ULL, "dae000_gen_turn_in_5.bnk"},
	{0x7F93BDAB66701E96ULL, "6036973_m.fxe"},
	{0x7F93FDA42534DA4DULL, "mag101_grace6052833.cl"},
	{0x7F940C28BECBEE60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmexit.ani"},
	{0x7F9416E92B87F170ULL, "1039099818.ogg"},
	{0x7F941B46C717C519ULL, "vxw_encounter2_bs_0_0.vlm"},
	{0x7F94332D5165F2FAULL, "static_static_p3.vsh.fxo"},
	{0x7F944EE01D1DDD00ULL, "6044527_m.fxe"},
	{0x7F947DB843675FA5ULL, "198828140.ogg"},
	{0x7F9480700DBF2EE3ULL, "vxht_0_hightown_a2_ato_0_0.vlm"},
	{0x7F949B06848C8A43ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01bargewreck.mao"},
	{0x7F949F06848C9090ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01bargewreck.mmh"},
	{0x7F94AD06848CA8AAULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01bargewreck.msh"},
	{0x7F95C98FCC7379FCULL, "159705531.ogg"},
	{0x7F96657102F820B5ULL, "138603055.ogg"},
	{0x7F969491120BF3F5ULL, "qun110_petrice6041360.cl"},
	{0x7F96A7EEA11C0F75ULL, "vxlt_darktown_599_0_0.vlm"},
	{0x7F970D1BE0125BE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v1.ani"},
	{0x7F9731D6F73D9D49ULL, "dae000_gen_turn_in_5.cnv"},
	{0x7F975ABB3D2BB170ULL, "art\\characters\\weapons\\dagger\\w_dgr_001a.mao"},
	{0x7F975EBB3D2BB783ULL, "art\\characters\\weapons\\dagger\\w_dgr_001a.mmh"},
	{0x7F976CBB3D2BCFDDULL, "art\\characters\\weapons\\dagger\\w_dgr_001a.msh"},
	{0x7F97817074171B32ULL, "mag131_keran6086202.cl"},
	{0x7F97CB8D6214DAEEULL, "hlac_lt_0_lowtown_docks_a3_-65539_2x.dds"},
	{0x7F97E9428F6D1BC6ULL, "llac_w_sundermount_-131078_0.dds"},
	{0x7F9822BACB2BAF00ULL, "hlac_w_woundedcoast_131075_0x.dds"},
	{0x7F985DF472F34FD4ULL, "mer341_jansen6054856.cl"},
	{0x7F9861DCE67D7DDEULL, "818123391.ogg"},
	{0x7F98EBCD9B677BE6ULL, "hlac_lt_hangedman_n_-1_3y.dds"},
	{0x7F99BD2D92B10B0DULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.dura.gad"},
	{0x7F9A02BBE327A689ULL, "vxlt_0_lowtown_n_gbr_0_0.vlm"},
	{0x7F9A2F03E7E89322ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_act_c.impa.gad"},
	{0x7F9A6EC9650F6187ULL, "hlac_lt_gallowstemplar_2383_5z.dds"},
	{0x7F9A862B567946AFULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_bdy_suca_0d.dds"},
	{0x7F9AA446B0E7F844ULL, "dus_rock_s.mao"},
	{0x7F9B86ACF9BD1274ULL, "vxlt_darktown_5je_0_0.vlm"},
	{0x7F9BB786427AFD3DULL, "6083673_m.fxe"},
	{0x7F9BC435F84514C9ULL, "llac_lt_0_lowtown_a2_n_c_-196608_2.dds"},
	{0x7F9BD66422765DB4ULL, "follower_banter6164772.cl"},
	{0x7F9BDD6EE0F0FE15ULL, "327186990.ogg"},
	{0x7F9BF1AAC12BCF1FULL, "02_impact.mao"},
	{0x7F9C296A5D0A136FULL, "hlac_lt_undercityserialkiller_-1_49x.dds"},
	{0x7F9C7C83BBA059A7ULL, "vxht_0_hightown_a2_bgm_0_0.vlm"},
	{0x7F9C8466FB42E87CULL, "hf_lgt_apostate03.mor"},
	{0x7F9CE39BA1D37D03ULL, "vxw_encounter2_camp_g6_0_0.vlm"},
	{0x7F9CFC93A6DFC15EULL, "llac_lt_0_lowtown_a3_n_c_-2_4.dds"},
	{0x7F9DAF709DF4BAEEULL, "vxw_bonepit_1wh_0_0.vlm"},
	{0x7F9DB6ED231FA743ULL, "452228521.ogg"},
	{0x7F9E096F429C8ABBULL, "zz_approval_debug6003916.cl"},
	{0x7F9E5C11C5454F35ULL, "qun110_petrice6018609.cl"},
	{0x7F9EA2F67B09BA3CULL, "vxlt_gallowstemplar_3ro_0_0.vlm"},
	{0x7F9EA6C7A08517A6ULL, "hlac_w_bonepitdestroyed_a3_196610_0y.dds"},
	{0x7F9EDC91A6AF8633ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_011a_0t.dds"},
	{0x7F9F5BFDBF520121ULL, "185494303.ogg"},
	{0x7F9FC7F77510485FULL, "6115377_f.fxe"},
	{0x7FA02A45E567D67EULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comc_0s.dds"},
	{0x7FA04EEA76F4309BULL, "vxlt_darktown_500_0_0.vlm"},
	{0x7FA0542785D72D04ULL, "vxlt_0_lowtown_a3_n_c_i0x_0_0.vlm"},
	{0x7FA0F1AA5967B4EDULL, "6095102_m.fxe"},
	{0x7FA11B4E20C90EBBULL, "genip_node_c1.utp"},
	{0x7FA15203E7EEED19ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_act_c.impa.evt"},
	{0x7FA1ECF813FBE988ULL, "vxlt_0_lowtown_docks_b0j_0_0.vlm"},
	{0x7FA28D23B90DDD99ULL, "vxw_encounter2_mr_0_0.vlm"},
	{0x7FA2918B1CAFEB1DULL, "pro000_cut2d_real_story6013662.cl"},
	{0x7FA336AF22D93C21ULL, "8408117.ogg"},
	{0x7FA343FF12A48CA1ULL, "tablewood_05.mao"},
	{0x7FA3B2CF65B44A3FULL, "vxht_mansion_n_4vw_0_0.vlm"},
	{0x7FA3D0CF653D2984ULL, "6178347_m.fxe"},
	{0x7FA3EFE6ECD191ABULL, "6126769_m.fxe"},
	{0x7FA42C5576883C1FULL, "vxlt_gallowscourtyard_61v_0_0.vlm"},
	{0x7FA4D32A0AB6515EULL, "vxw_bonepitmining_a2_br_0_0.vlm"},
	{0x7FA4EE1D3306F031ULL, "gen_im_arm_cht_lgt_champion.uti"},
	{0x7FA4F51BE01E7057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v1.evt"},
	{0x7FA50788725CCB64ULL, "vxdr_dungeon_1lv_0_1.vlm"},
	{0x7FA542F3BE80F9F8ULL, "dae000_bartrand6206863.cl"},
	{0x7FA6D0535D178101ULL, "6187757_m.fxe"},
	{0x7FA6EE4C452C4C97ULL, "gen00fl_isabela6099506.cl"},
	{0x7FA731DD0A660976ULL, "1015044162.ogg"},
	{0x7FA7363204ABDD0BULL, "zz_vis_debug6072526.cl"},
	{0x7FA78FF188ADBE7AULL, "m_dae220ar_hightown.wbd"},
	{0x7FA79A61EC391F68ULL, "char1.vsh.fxo"},
	{0x7FA79FD0B0CA8134ULL, "zz_vrc_debug6064251.cl"},
	{0x7FA8561705ACDF4AULL, "6075644_m.fxe"},
	{0x7FA865EC5F3CEE4DULL, "6171082_m.fxe"},
	{0x7FA88D84E73DCF7FULL, "hlac_lt_gallowsprison_n_-1_7y.dds"},
	{0x7FA89B273D946FBDULL, "art\\vfx\\combat\\vfx_c_shadow_creature_c.mmh"},
	{0x7FA8F5A4CBECFE77ULL, "419592988.ogg"},
	{0x7FA90421BB48CDEEULL, "vxw_woundedcoast_pc_0_0.vlm"},
	{0x7FA9F889EBB32AF4ULL, "vxlt_darktown_5oj_0_0.vlm"},
	{0x7FA9FCE41D9E4913ULL, "6135365_m.fxe"},
	{0x7FAA2963347115BDULL, "6086973_m.fxe"},
	{0x7FAA6FF9D9847EEFULL, "dae000_flemeth6000887.cl"},
	{0x7FAA84B7CC1AD812ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sword_d.dds"},
	{0x7FAB4DA005F07B55ULL, "sib100_father_contact6110088.cl"},
	{0x7FAB557893C99583ULL, "vxw_encounter2_camp_b9_0_0.vlm"},
	{0x7FAB9775230958D2ULL, "6129357_f.fxe"},
	{0x7FABB35E513F4F15ULL, "vxlt_undercityencounter_1br_0_0.vlm"},
	{0x7FABF61BE02490E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v1.gad"},
	{0x7FAC11BC2BD9AA42ULL, "676703753.ogg"},
	{0x7FAC214D8FEAD36EULL, "vxw_bonepit_1xi_0_0.vlm"},
	{0x7FAC5279DE48DD77ULL, "hlac_dr_slavers_3216_12x.dds"},
	{0x7FAD040E6D825F9BULL, "vxw_woundedcoast_1fu_0_0.vlm"},
	{0x7FAD5B322DE598E2ULL, "vxca_caverns2_710_0_0.vlm"},
	{0x7FADE80919ABF352ULL, "nat320_delilah6065883.cl"},
	{0x7FAE46833593C4FAULL, "471355805.ogg"},
	{0x7FAE492598295391ULL, "ave331_jeven6137282.cl"},
	{0x7FAE5A03E7F98508ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_act_c.impa.ani"},
	{0x7FAEB20956EB90D5ULL, "hf_ave3_insurgent_guard4.mor"},
	{0x7FAED2AF514A0EBCULL, "hlac_dr_dungeon_3216_8x.dds"},
	{0x7FAEDE379B382A83ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04b.mmh"},
	{0x7FAEEC379B3842DDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04b.msh"},
	{0x7FAF3FFFCE101946ULL, "6115292_m.fxe"},
	{0x7FAF53AE957F8D5BULL, "vxca_caverns2_95c_0_0.vlm"},
	{0x7FAFF2BFDDE006C5ULL, "gen00fl_anders6111178.cl"},
	{0x7FB008D62EFE7A2AULL, "6066271_m.fxe"},
	{0x7FB05ED505F20208ULL, "vxlt_0_lowtown_docks_b7k_0_0.vlm"},
	{0x7FB0C4D435CB2CB9ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potc.msh"},
	{0x7FB0CED435CB3DA7ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potc.mmh"},
	{0x7FB0D2D435CB44F4ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potc.mao"},
	{0x7FB0FF00AB03F619ULL, "vxw_encounter2_hq_0_0.vlm"},
	{0x7FB16429F1AB5FE3ULL, "vxht_0_hightown_a2_ajq_0_0.vlm"},
	{0x7FB190FB29046DEAULL, "6133528_m.fxe"},
	{0x7FB1A2FBB4C30AFCULL, "vxdr_primevaltomb_3lw_0_0.vlm"},
	{0x7FB2262CB89BD7A0ULL, "802982366.ogg"},
	{0x7FB347C3FB05B987ULL, "6000703_f.fxe"},
	{0x7FB38F4D2F683F4FULL, "vxw_bonepitmining_a2_164_0_0.vlm"},
	{0x7FB3AC772E1D8B7CULL, "mag371_elsa6041517.cl"},
	{0x7FB47150E392B706ULL, "vxlt_0_lowtown_docks_n_b90_0_0.vlm"},
	{0x7FB4A2EEA8995A24ULL, "and110_refugee6111615.cl"},
	{0x7FB4B53ECFAAC795ULL, "117619619.ogg"},
	{0x7FB4EF58816126ECULL, "vrc321ip_haunted_chair.utp"},
	{0x7FB542304F0D9981ULL, "6188657_m.fxe"},
	{0x7FB55F3206DCC1E2ULL, "llac_lt_0_lowtown_docks_a3_n_65532_1.dds"},
	{0x7FB5C4458821F02AULL, "6217970_m.fxe"},
	{0x7FB630209B8A941AULL, "802320313.ogg"},
	{0x7FB6884641675D91ULL, "299520660.ogg"},
	{0x7FB691B6E680D8FBULL, "mcr370_messenger6071838.cl"},
	{0x7FB73C391045E12DULL, "lt_0_lowtown_docks_234_0.mmh"},
	{0x7FB740391045E87EULL, "lt_0_lowtown_docks_234_0.mao"},
	{0x7FB74E391045FFB3ULL, "lt_0_lowtown_docks_234_0.msh"},
	{0x7FB74F1DFC70175DULL, "199471455.ogg"},
	{0x7FB8069F35B8BDC2ULL, "6063950_f.fxe"},
	{0x7FB8167AE7187ED0ULL, "550797256.ogg"},
	{0x7FB84032BEEEC5F9ULL, "6059347_f.fxe"},
	{0x7FB87BE60485998DULL, "hlac_ht_0_hightown_65535_0x.dds"},
	{0x7FB884A4CF75A765ULL, "6092579_f.fxe"},
	{0x7FB94CEAE02210A8ULL, "face_rendertodepth_p0.psh.fxo.dx11"},
	{0x7FB96F88983D34B2ULL, "92626164.ogg"},
	{0x7FB97128DC0DC588ULL, "lgt160_ntb_elf_conflict6163352.cl"},
	{0x7FB9802D9DAEFA5BULL, "320748586.ogg"},
	{0x7FB9E8850AF1F1F3ULL, "qcr000_cut2d_thechampion6025130.cl"},
	{0x7FB9F996C78ABD28ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillard.msh"},
	{0x7FBA1796C78AF052ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillard.mmh"},
	{0x7FBA205DF7E4A09FULL, "6111231_m.fxe"},
	{0x7FBA6D30361D2E62ULL, "llac_w_woundedcoast_-65531_0.dds"},
	{0x7FBAE713CBF15263ULL, "w_woundedcoast_25_0_ll.mmh"},
	{0x7FBB2F8192357CE4ULL, "qcr220_grey_wardens6207448.cl"},
	{0x7FBB75EB5F78781BULL, "vxw_woundedcoast_1at_0_0.vlm"},
	{0x7FBC26E72AF53578ULL, "vxca_caverns2_7n1_0_2.vlm"},
	{0x7FBC6614B4CC5C14ULL, "6076899_f.fxe"},
	{0x7FBC86A7EEB82039ULL, "ts_coterie_acc_table.gda"},
	{0x7FBC8D46B5E684C9ULL, "vxdr_0_deeproads_f05_0_0.vlm"},
	{0x7FBD15E1FA362D44ULL, "hlac_lt_0_lowtown_a3_n_c_1_10y.dds"},
	{0x7FBD40F07AFEAD42ULL, "834770247.ogg"},
	{0x7FBE329E7DF23465ULL, "zz_approval_debug6003778.cl"},
	{0x7FBE5A20221190B6ULL, "hm_carver_inf.mor"},
	{0x7FBEE24BCC072BEAULL, "222750794.ogg"},
	{0x7FBF0B324AF51346ULL, "mag130_danzig6018848.cl"},
	{0x7FBF160E5190678FULL, "hlac_w_bonepit_196612_0z.dds"},
	{0x7FBF48EBDC647EA2ULL, "6115393_m.fxe"},
	{0x7FBF760BA05880F8ULL, "6178439_m.fxe"},
	{0x7FBFDF379B469FA8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04b.phy"},
	{0x7FC030DF667D9939ULL, "6178359_m.fxe"},
	{0x7FC06B03D93168F3ULL, "vxht_mansion2b_2om_0_0.vlm"},
	{0x7FC0B8807324F21EULL, "follower_banter6136719.cl"},
	{0x7FC22A0FC08F97C4ULL, "w_bonepitmining_a2_80_0.mmh"},
	{0x7FC22E0FC08F9E17ULL, "w_bonepitmining_a2_80_0.mao"},
	{0x7FC23786D8FE7060ULL, "lt_0_lowtown_docks_a3_24_0.anb"},
	{0x7FC2400FC08FBD66ULL, "w_bonepitmining_a2_80_0.msh"},
	{0x7FC25B0D05D08C0CULL, "vxht_mansion_n_29w_0_0.vlm"},
	{0x7FC288A55C341E2EULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyn.phy"},
	{0x7FC2E32DD588CF86ULL, "vxlt_0_lowtown_docks_n_b61_0_0.vlm"},
	{0x7FC2F22EC69A2CB3ULL, "602446229.ogg"},
	{0x7FC3223C892ECA33ULL, "vxlt_0_lowtown_a2_n_ci5_0_0.vlm"},
	{0x7FC356DA0ED1BF98ULL, "306000759.ogg"},
	{0x7FC3768715CB1356ULL, "vxdr_primevaltomb_6wm_0_0.vlm"},
	{0x7FC3A3551B60A856ULL, "pou_cyc.mao"},
	{0x7FC3F47D298C72D8ULL, "vxw_encounter2_g1_0_0.vlm"},
	{0x7FC3FA54043AB463ULL, "zz_spy121cr_cordy.utc"},
	{0x7FC409CF5AF9AF63ULL, "art\\characters\\playercharacter\\humanmale\\hlf_bda.mao"},
	{0x7FC40DCF5AF9B5B0ULL, "art\\characters\\playercharacter\\humanmale\\hlf_bda.mmh"},
	{0x7FC41BCF5AF9CDCAULL, "art\\characters\\playercharacter\\humanmale\\hlf_bda.msh"},
	{0x7FC421C5BC9ED5B2ULL, "fonts_es-es.gfx"},
	{0x7FC4AFACF16813E6ULL, "hlac_lt_0_lowtown_a2_n_c_65538_3z.dds"},
	{0x7FC4B4E8F9AF26A8ULL, "gen_im_delete_me_02.uti"},
	{0x7FC4E429AF719612ULL, "qun211cr_elf_warrior6087669.cl"},
	{0x7FC66EFC51B1C544ULL, "6125298_f.fxe"},
	{0x7FC6B21561E425BFULL, "6031273_f.fxe"},
	{0x7FC6BBBFC3131F66ULL, "hlac_ht_mansion2_818_9y.dds"},
	{0x7FC6C18816B4539EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardright.ani"},
	{0x7FC75B37EDCA0C90ULL, "default_rogue_female2.utc"},
	{0x7FC7E8A343ED4C60ULL, "6176161_m.fxe"},
	{0x7FC881A7C5B74E4EULL, "hlac_lt_0_lowtown_0_0y.dds"},
	{0x7FC8B979F0E75E8AULL, "hlac_dr_primevaltomb_664_0x.dds"},
	{0x7FC8C8B276D280FAULL, "227824674.ogg"},
	{0x7FC9E7C8516E909BULL, "vxw_woundedcoast_1lw_0_0.vlm"},
	{0x7FC9F5DBD614AED8ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_3x.dds"},
	{0x7FCA2C606A99CF4AULL, "6128000_f.fxe"},
	{0x7FCA3B06DB2BCC2AULL, "208783008.ogg"},
	{0x7FCA86B0FF9D020CULL, "hlac_ht_keep_4519_0z.dds"},
	{0x7FCA8D79C4AEA8FBULL, "vxw_encounter1_co_0_0.vlm"},
	{0x7FCA9630CB080FFBULL, "vxlt_hangedman_n_rs_0_0.vlm"},
	{0x7FCA9E121C87B43AULL, "art\\characters\\creatures\\corpse\\textures\\cn_hed_cor_0n.dds"},
	{0x7FCB70543FC90041ULL, "vxw_sundermount_319_0_1.vlm"},
	{0x7FCCABF35446B51FULL, "dae000_sabina6162339.cl"},
	{0x7FCCC9A6FDAC1ABAULL, "144749416.ogg"},
	{0x7FCDA8DB7A0B3231ULL, "6017430_m.fxe"},
	{0x7FCE74B50CF09EEAULL, "6139328_m.fxe"},
	{0x7FCEDC35647A823DULL, "vxw_blightlands_db_0_0.vlm"},
	{0x7FCF45EA6BBC5E66ULL, "854144785.ogg"},
	{0x7FCF4C4DCFB4C08DULL, "vxht_mansion2b_282_0_0.vlm"},
	{0x7FCF9FE2E2ABAC93ULL, "vxlt_undercityserialkiller_11u_0_0.vlm"},
	{0x7FCFFD3AF844024DULL, "narrative_varric6172483.cl"},
	{0x7FD08D7AFBBDC7C7ULL, "follower_banter6163242.cl"},
	{0x7FD0F964390E881DULL, "119361510.ogg"},
	{0x7FD0FEFC15C1F13EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_1.ani"},
	{0x7FD11438C1D2E8F8ULL, "isa211_isabela6081967.cl"},
	{0x7FD13EA6C50FE2F6ULL, "gen00fl_aveline6107243.cl"},
	{0x7FD1DD0634C3BDA6ULL, "mag291_demon_marethari6027963.cl"},
	{0x7FD1E86407C12BD6ULL, "vxdr_primevaltomb_6zn_0_0.vlm"},
	{0x7FD20A1F2AFDA79FULL, "6101199_m.fxe"},
	{0x7FD210EB5955B2CBULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v1.ani"},
	{0x7FD2449F97D7B52CULL, "dae000_templar_recruit_16164426.cl"},
	{0x7FD32321D4CE1A8FULL, "6050424_m.fxe"},
	{0x7FD38019A0444A4FULL, "800017595.ogg"},
	{0x7FD42F87079FCE61ULL, "6161124_m.fxe"},
	{0x7FD4D98816C0BA6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardright.evt"},
	{0x7FD4E4F6678C31F9ULL, "follower_banter6160630.cl"},
	{0x7FD4EA197698FED0ULL, "vxlt_hangedman_n_21d_0_1.vlm"},
	{0x7FD4EA28F5B5287FULL, "art\\vfx\\combat\\vfx_glb_ele_e_deathblow_c.impa.gad"},
	{0x7FD540C55146787DULL, "6070147_m.fxe"},
	{0x7FD59D72B43A5659ULL, "127943380.ogg"},
	{0x7FD6353AE67F6F87ULL, "art\\vfx\\materials\\chunky\\vfx_c_sphere_deform.mao"},
	{0x7FD670AFDAF083BEULL, "llac_w_bonepit_262147_0.dds"},
	{0x7FD6D2D41920E72EULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.cess.ani"},
	{0x7FD776E465CD1CEEULL, "vxw_bonepit_1kl_0_0.vlm"},
	{0x7FD78AF466A98BCDULL, "6142831_m.fxe"},
	{0x7FD796A9A854C50CULL, "vxw_encounter2_camp_22_0_0.vlm"},
	{0x7FD7D4235716ED47ULL, "vxlt_0_lowtown_docks_art_0_0.vlm"},
	{0x7FD84DABE078B5B1ULL, "922015073.ogg"},
	{0x7FD8CE134F6C7B5FULL, "gen_im_cns_pot_injury_01.uti"},
	{0x7FD8FF56B6A4C17BULL, "vxw_encounter1_hn_0_0.vlm"},
	{0x7FD90C0289673CBFULL, "6164231_m.fxe"},
	{0x7FD9121654E1B218ULL, "464534266.ogg"},
	{0x7FD9225354FF048BULL, "34433781.ogg"},
	{0x7FD928909AE75DF5ULL, "706417905.ogg"},
	{0x7FD9672BD9F0F75CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr_2.mao"},
	{0x7FDA744272763880ULL, "6206127_m.fxe"},
	{0x7FDA98A5447EEF56ULL, "dae000_hate_enemies6167589.cl"},
	{0x7FDA9A70DF78F308ULL, "dae200_cullen6140490.cl"},
	{0x7FDADA8816C52718ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkforwardright.gad"},
	{0x7FDADF28F5B9816CULL, "art\\vfx\\combat\\vfx_glb_ele_e_deathblow_c.impa.evt"},
	{0x7FDAE0960F4BDFF1ULL, "256380784.ogg"},
	{0x7FDB2F49DD770A31ULL, "vxw_sundermount_2e1_0_0.vlm"},
	{0x7FDBEC5F80C970AFULL, "687196338.ogg"},
	{0x7FDC42C65E50C439ULL, "6126994_m.fxe"},
	{0x7FDC51289F8A8E44ULL, "qapath3.arl"},
	{0x7FDCB36BA3C0AE36ULL, "lgt250ip_at_tomecave.utp"},
	{0x7FDCB7D435F051DCULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potc.phy"},
	{0x7FDCB8652CC2DD35ULL, "vxht_0_hightown_7m0_0_0.vlm"},
	{0x7FDCEAA9DF93E225ULL, "gen00fl_sebastian6177889.cl"},
	{0x7FDCF88B597CA2A5ULL, "llac_lt_0_lowtown_a3_n_c_65536_7.dds"},
	{0x7FDCF9CC432D38D4ULL, "616165805.ogg"},
	{0x7FDD40F61CC439F0ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_04bonetrack.msh"},
	{0x7FDD4E1256709ABDULL, "vxw_blightlands_ia_0_0.vlm"},
	{0x7FDD4EBDBD1D99F3ULL, "vxht_mansion2b_2eo_0_0.vlm"},
	{0x7FDD4EF61CC4510AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_04bonetrack.mmh"},
	{0x7FDD52F61CC45841ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_04bonetrack.mao"},
	{0x7FDDB5814C11D503ULL, "vxlt_0_lowtown_a3_n_c_gbo_0_0.vlm"},
	{0x7FDDBAADE94E26CDULL, "645017716.ogg"},
	{0x7FDDBE2AC1AAD90DULL, "vxht_mansion2b_251_0_0.vlm"},
	{0x7FDE16FC15CCA40FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_1.evt"},
	{0x7FDE8E08D79E5A42ULL, "ward_ex.mao"},
	{0x7FDEC33F66C95EB4ULL, "6086165_m.fxe"},
	{0x7FDF22EDE55721F5ULL, "688515914.ogg"},
	{0x7FDF75E8FB2D5005ULL, "follower_banter6170482.cl"},
	{0x7FDFAEA6C66E9E34ULL, "lgt311_shw_mer_mettin6156545.cl"},
	{0x7FDFBFCAAA6ED6B5ULL, "one200_gold_followers6117252.cl"},
	{0x7FDFF8EB5961C742ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v1.evt"},
	{0x7FE004F292FFDF3DULL, "vxlt_0_lowtown_a3_n_c_by1_0_0.vlm"},
	{0x7FE04E17B1FAD558ULL, "vxht_0_hightown_a2_asu_0_0.vlm"},
	{0x7FE06AB109CA52FAULL, "vxca_cavern1_haunted_hlz_0_0.vlm"},
	{0x7FE110AC90AC705AULL, "lt_0_lowtown_docks_n_236_0.msh"},
	{0x7FE11EAC90AC8793ULL, "lt_0_lowtown_docks_n_236_0.mao"},
	{0x7FE122AC90AC8EA0ULL, "lt_0_lowtown_docks_n_236_0.mmh"},
	{0x7FE19CDDDC3A456AULL, "art\\vfx\\cinematic\\vfx_merrill_reflect_c.anb"},
	{0x7FE1AF5471E6F5B6ULL, "6075110_m.fxe"},
	{0x7FE1F255AE955EC6ULL, "gen00fl_isabela6015839.cl"},
	{0x7FE1F6437298A15FULL, "follower_banter6155841.cl"},
	{0x7FE211FC9713B2F7ULL, "6113906_f.fxe"},
	{0x7FE23323827058EEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masc_0d.dds"},
	{0x7FE23410B2B08D26ULL, "vxca_caverns1_i19_0_0.vlm"},
	{0x7FE32B1D9980547AULL, "6149172_m.fxe"},
	{0x7FE4052216D77D89ULL, "193322016.ogg"},
	{0x7FE4975D4E053403ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_ric_0s.dds"},
	{0x7FE4CAD4192D177FULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.cess.evt"},
	{0x7FE517FC15D2C4B8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k_gesture_both_hands_1.gad"},
	{0x7FE5909520E01EA9ULL, "hlac_dr_slavers_3216_2y.dds"},
	{0x7FE5B6A61589D276ULL, "dae021_sister_alema_mob6144771.cl"},
	{0x7FE5F7EB59663155ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v1.gad"},
	{0x7FE6CB82355AC19BULL, "vxw_woundedcoast_1ri_0_0.vlm"},
	{0x7FE7DAD2CDEC1B47ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_orsino_orsinosdefiance.gad"},
	{0x7FE8A1BA04C6A2FBULL, "gen00fl_isabela6036659.cl"},
	{0x7FE8C728F5C595C5ULL, "art\\vfx\\combat\\vfx_glb_ele_e_deathblow_c.impa.ani"},
	{0x7FE8F997E34C6D33ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_window_e.dds"},
	{0x7FE92283BEB6E9A7ULL, "vxca_caverns2prc_8d0_0_0.vlm"},
	{0x7FE9D5F41B0325CCULL, "379065226.ogg"},
	{0x7FE9D7AD917C6132ULL, "lt_gallowstemplar_f_2386.rml"},
	{0x7FE9EF7B19C3FDE8ULL, "vxht_bartrand_n_6k6_0_0.vlm"},
	{0x7FEA298CB6069310ULL, "lt_0_lowtown_docks_n_19_0.rml"},
	{0x7FEA426A254E3D67ULL, "708086576.ogg"},
	{0x7FEA4EEA77AE70F6ULL, "vxlt_warehouse_1c1_0_0.vlm"},
	{0x7FEA62EED5B376E0ULL, "44594097.ogg"},
	{0x7FEA70FC409D570AULL, "art\\characters\\creatures\\mubari\\animation\\bnt_dblw.gad"},
	{0x7FEB8260F909AD7AULL, "lt_0_lowtown_docks_246_0.mmh"},
	{0x7FEB8660F909B3B1ULL, "lt_0_lowtown_docks_246_0.mao"},
	{0x7FEB8BCC7CC288C6ULL, "6071613_m.fxe"},
	{0x7FEB9460F909CBC0ULL, "lt_0_lowtown_docks_246_0.msh"},
	{0x7FEB9E7B84AB5A15ULL, "isa111_isabela6164592.cl"},
	{0x7FEBABD419330168ULL, "art\\vfx\\combat\\vfx_glb_dispel_crust_c.cess.gad"},
	{0x7FEC2C825834D0E7ULL, "lgt210_magu_selby6142052.cl"},
	{0x7FEC9689FD402A9EULL, "306699388.ogg"},
	{0x7FEDEF1C23431261ULL, "zz_mer_debug6062732.cl"},
	{0x7FEECF6D5A1CAC7CULL, "gen_im_cns_cft_psn_04.uti"},
	{0x7FEED29CE760C68CULL, "vxca_caverns1_j43_0_0.vlm"},
	{0x7FEF15210BC7144FULL, "vxdr_0_deeproads_ewi_0_0.vlm"},
	{0x7FEF7C0C53D2A956ULL, "art\\characters\\creatures\\stonegolem\\animation\\turnr90.evt"},
	{0x7FEF8E33B12879A6ULL, "plt_fen300pt_alone.nss"},
	{0x7FEFCA8678D4F21FULL, "911787930.ogg"},
	{0x7FF01229D6266841ULL, "833418918.ogg"},
	{0x7FF03361AEFD69B2ULL, "atl_debugoptio_dxt5_dat.xml"},
	{0x7FF04EBBD6F43B81ULL, "gencr_mercenary_warrior_ws.utc"},
	{0x7FF06FEA9EC04FC6ULL, "art\\levels\\races\\design\\_textures\\prp_gasbarrel_n.dds"},
	{0x7FF0976F1D95DF1FULL, "vxht_0_hightown_n_c3n_0_1.vlm"},
	{0x7FF110413CADF2FEULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.impa.evt"},
	{0x7FF153FC40A344A1ULL, "art\\characters\\creatures\\mubari\\animation\\bnt_dblw.evt"},
	{0x7FF20C64E0BDF113ULL, "6147210_m.fxe"},
	{0x7FF22A33A2FD0521ULL, "222043819.ogg"},
	{0x7FF23059485E147EULL, "680368344.ogg"},
	{0x7FF24C2F054C40C6ULL, "mag131_keran6088222.cl"},
	{0x7FF27F2E76835895ULL, "vxlt_0_lowtown_docks_a3_n_aph_0_0.vlm"},
	{0x7FF2D29178574662ULL, "vxca_cave_1fq_0_1.vlm"},
	{0x7FF2DC4056F5D966ULL, "978779410.ogg"},
	{0x7FF2F131206A5F8CULL, "art\\vfx\\materials\\vfx_smoke_soft_add.mao"},
	{0x7FF3003BD698BA98ULL, "vxlt_0_lowtown_n_eql_0_0.vlm"},
	{0x7FF308A6C99FB219ULL, "265117645.ogg"},
	{0x7FF346B02A4F367FULL, "llac_ca_cavern1_haunted_24424_25.dds"},
	{0x7FF35577B1C2721CULL, "dae000cr_fenarel.utc"},
	{0x7FF3C67ED37DD737ULL, "6139853_m.fxe"},
	{0x7FF3C9CDE6B82255ULL, "vxlt_0_lowtown_docks_b6t_0_1.vlm"},
	{0x7FF40479247D5ABEULL, "6087128_m.fxe"},
	{0x7FF44C1163756866ULL, "zz_vault_debug6201718.cl"},
	{0x7FF487773CA96E28ULL, "vxlt_undercitydungeon_wt_0_0.vlm"},
	{0x7FF4D95EA4F5472EULL, "mcr370_cut_marriage6189916.cl"},
	{0x7FF4F56C7812D76DULL, "6109092_m.fxe"},
	{0x7FF53F5DC2243B8AULL, "vxw_bonepit_1pu_0_0.vlm"},
	{0x7FF55D96A8946C57ULL, "6132975_m.fxe"},
	{0x7FF59F6017D87333ULL, "67087575.ogg"},
	{0x7FF630794C76E80FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medf_0n.dds"},
	{0x7FF6730C53D8B8E1ULL, "art\\characters\\creatures\\stonegolem\\animation\\turnr90.gad"},
	{0x7FF706229B6FAD7BULL, "vxca_caverns2_808_0_1.vlm"},
	{0x7FF707FF5AD56DE9ULL, "ef_gen_act3.mor"},
	{0x7FF7280FA6EB2666ULL, "944735334.ogg"},
	{0x7FF72A04871926F7ULL, "545757950.ogg"},
	{0x7FF78F2DAB775296ULL, "6136141_m.fxe"},
	{0x7FF7CD10797401C6ULL, "239196457.ogg"},
	{0x7FF81303C1633B31ULL, "vxw_sundermount_2s7_0_0.vlm"},
	{0x7FF827413CB43829ULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.impa.gad"},
	{0x7FF864B4CD2E0859ULL, "vxlt_0_lowtown_a3_n_c_e1z_0_0.vlm"},
	{0x7FF8B49959DDAF68ULL, "lt_0_lowtown_docks_39_0_lc.mmh"},
	{0x7FF93186AB89A4EEULL, "ser221tr_chase_jump2.ncs"},
	{0x7FF98D684D5DB7FAULL, "6020429_f.fxe"},
	{0x7FF9C85E573F79A0ULL, "vxlt_backalley_n_1dp_0_0.vlm"},
	{0x7FF9CEA5EDB53EE1ULL, "llac_lt_0_lowtown_a2_n_1_6.dds"},
	{0x7FFA2E257F5D8B34ULL, "llac_lt_backalley_n_0_10.dds"},
	{0x7FFA4FC00B00110CULL, "ave331cr_insurgent_f2.utc"},
	{0x7FFAC83978826B9AULL, "6095375_f.fxe"},
	{0x7FFAFCF51638D348ULL, "169448169.ogg"},
	{0x7FFB65ED2D94B1F6ULL, "llac_lt_0_lowtown_docks_a3_65534_4.dds"},
	{0x7FFB6F667AB80C5BULL, "hla_ht_0_hightown_a2_-131072_0x.dds"},
	{0x7FFBC7D2CDFCA38DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_orsino_orsinosdefiance.ani"},
	{0x7FFC1102DC5F2088ULL, "31168982.ogg"},
	{0x7FFC740C53DD258FULL, "art\\characters\\creatures\\stonegolem\\animation\\turnr90.ani"},
	{0x7FFC93448108ADD5ULL, "character_character_p239.psh.fxo.dx11.assembly"},
	{0x7FFCD1D650C9127EULL, "llac_w_bonepit_131072_0.dds"},
	{0x7FFD0055F7DBCA04ULL, "6018724_m.fxe"},
	{0x7FFD019A7AB6E0B0ULL, "gen_im_trs_gen_two_2hs_r1.uti"},
	{0x7FFD86FDFDBD2CCFULL, "vxdr_0_deeproads_eph_0_0.vlm"},
	{0x7FFE4BFC40ADC1F0ULL, "art\\characters\\creatures\\mubari\\animation\\bnt_dblw.ani"},
	{0x7FFE777B36EB5147ULL, "6031263_m.fxe"},
	{0x7FFEE1724559106AULL, "268613206.ogg"},
	{0x7FFEF8413CBA0777ULL, "art\\vfx\\creatures\\fxc_wyn_gui_c.impa.ani"},
	{0x7FFF40D0474E42D4ULL, "vxht_bartrand_n_6jy_0_0.vlm"},
	{0x7FFFC238B3AB7891ULL, "470707291.ogg"},
	{0x8000642F67163C55ULL, "elffemale.fxa"},
	{0x80007E41D2B40993ULL, "6140310_m.fxe"},
	{0x8000BB0DF16520C7ULL, "ca_caverns2_mmi.gff"},
	{0x8000F01A86C51A05ULL, "llac_lt_blackemporium_1055_23.dds"},
	{0x8001DC2BC0386177ULL, "lowtown_fem_lowlife_mob6168953.cl"},
	{0x80021686D9341B88ULL, "lt_0_lowtown_docks_a3_24_0.rml"},
	{0x800287A55C6A00F1ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyn.mmh"},
	{0x800293A55C6A1492ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyn.mao"},
	{0x8002A1A55C6A2CAFULL, "art\\characters\\playercharacter\\dwarfmale\\boo_hvyn.msh"},
	{0x8002D00E56E6C21DULL, "6042493_f.fxe"},
	{0x8003508DC35C363EULL, "6189006_m.fxe"},
	{0x8003A333B13946E1ULL, "plt_fen300pt_alone.plo"},
	{0x8003CC1BC56B6167ULL, "venge_groupheal.dds"},
	{0x800414CF5B2FA5CBULL, "art\\characters\\playercharacter\\humanmale\\hlf_bda.phy"},
	{0x80057DA5B4A04FAEULL, "dre000_varric6208025.cl"},
	{0x80058546A5FDC55CULL, "452436123.ogg"},
	{0x8005AAEF5DA8DBC6ULL, "mcr370_meredith6083053.cl"},
	{0x80061F1CE7806BB5ULL, "hlac_ca_caverns2prc_5802_58z.dds"},
	{0x8006310F9D702510ULL, "gen00fl_merrill6133254.cl"},
	{0x8006678613EA5309ULL, "6129919_m.fxe"},
	{0x8006B4EE342C008BULL, "vxht_playershouse_n_1iv_0_0.vlm"},
	{0x8006CD8322B259F1ULL, "fen110_fenris6100247.cl"},
	{0x8006D334FDB02EE8ULL, "vxht_bartrand_n_6q4_0_0.vlm"},
	{0x8007748E26A54585ULL, "vxht_0_hightown_a3_b87_0_0.vlm"},
	{0x800775C171C11E3AULL, "and100_anders6096030.cl"},
	{0x800855BEEBCE622AULL, "6071169_m.fxe"},
	{0x8008641D2F0362CFULL, "6152133_m.fxe"},
	{0x800897813210CFEEULL, "37604602.ogg"},
	{0x8009696A211F7D8DULL, "ts_gems_05.gda"},
	{0x80098CB1D799033CULL, "vxca_caverns2_62n_0_0.vlm"},
	{0x800A37BA3596D831ULL, "fen200cr_tev_hunter_2.utc"},
	{0x800A74299DEA74FFULL, "162432054.ogg"},
	{0x800A93EAB7B1D7B0ULL, "w_blightlands_55_0.rml"},
	{0x800A9FDDDC5D4F5DULL, "art\\vfx\\cinematic\\vfx_merrill_reflect_c.mmh"},
	{0x800AA6175F51F176ULL, "lt_0_lowtown_a2_n_c_52_0.rml"},
	{0x800B7BCA4A56C54EULL, "qalit1.arl"},
	{0x800BE3C9B0026AC0ULL, "201171435.ogg"},
	{0x800C4CF790F6DFDDULL, "hlac_w_encounter2_393215_0x.dds"},
	{0x800C540E4FB8D8F5ULL, "6089491_m.fxe"},
	{0x800C6AED2C768964ULL, "455486376.ogg"},
	{0x800CA991317A930DULL, "6082181_f.fxe"},
	{0x800CF5DD5D95BD8EULL, "gen00fl_isabela6098201.cl"},
	{0x800D0D3686769D2BULL, "6102910_m.fxe"},
	{0x800D334B2159B5D5ULL, "hlac_w_encounter2_-131072_0x.dds"},
	{0x800D58D391A0D7A0ULL, "narrative_varric6172527.cl"},
	{0x800D77F2E76E0069ULL, "hlac_lt_0_lowtown_a2_-65535_1x.dds"},
	{0x800DA323600BC41BULL, "408135062.ogg"},
	{0x800DF97EC96FDF23ULL, "vxca_cavern1_haunted_ics_0_0.vlm"},
	{0x800E16A6F6203AF9ULL, "em_mrl150cr_dalish_hunter.mor"},
	{0x800E340BBB23C269ULL, "hlac_ca_caverns1_24426_4z.dds"},
	{0x800EAAF05468E7F9ULL, "hm_qfq_mob1.mor"},
	{0x800F19490CB42D45ULL, "llac_ht_0_hightown_n_65535_3.dds"},
	{0x800F4E772299645AULL, "6115426_f.fxe"},
	{0x800FA8AE91D7EFB5ULL, "t1_mub_br1.tnt"},
	{0x800FE9A20FFE7B5FULL, "mag291_demon_marethari6029041.cl"},
	{0x80115E7E652CEE7BULL, "vxht_mansion_n_5cy_0_0.vlm"},
	{0x80118BD10F4E46A8ULL, "llac_lt_0_lowtown_a2_n_c_-65535_3.dds"},
	{0x8011FDB838181598ULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_spiderweb1_da.dds"},
	{0x801234341928108CULL, "gen00fl_aveline6120822.cl"},
	{0x8012AF3534EA58ECULL, "plt_ser120pt_sacrifice.plo"},
	{0x8012CA84C9CF62C6ULL, "gen00fl_carver6097387.cl"},
	{0x80131BF319859141ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookmisc_sa.dds"},
	{0x80133E341458ED6EULL, "6197650_m.fxe"},
	{0x8014A691899F76D0ULL, "6096062_f.fxe"},
	{0x8014F6BDA54F6C31ULL, "vxw_sundermount_2y5_0_0.vlm"},
	{0x8015E66B189B5E05ULL, "vxht_0_hightown_a3_b76_0_0.vlm"},
	{0x80162B6DA7E39642ULL, "plt_vrc000pt_n_nt_completed.plo"},
	{0x801713B346AF5A03ULL, "llac_ht_0_hightown_a2_131074_0.dds"},
	{0x801715E5546DAA7CULL, "399006821.ogg"},
	{0x80171C1FED656582ULL, "mag210_arianni6024566.cl"},
	{0x80172A8CB62D8378ULL, "lt_0_lowtown_docks_n_19_0.anb"},
	{0x8017557D8248D442ULL, "906968857.ogg"},
	{0x80177CF513EA3703ULL, "vxlt_0_lowtown_a3_n_c_gvs_0_0.vlm"},
	{0x80185D120A1F6236ULL, "343694179.ogg"},
	{0x80195358EF7882C0ULL, "post_game_save.gda"},
	{0x8019943534F04923ULL, "plt_ser120pt_sacrifice.nss"},
	{0x80199F0ADC1C02E6ULL, "vxw_sundermount_313_0_0.vlm"},
	{0x801A74AD73D82B7DULL, "zz_fol_debug6092327.cl"},
	{0x801A9EB4235205F7ULL, "gen00fl_aveline6115863.cl"},
	{0x801AB585FFEF75D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v10.evt"},
	{0x801B3DF82296A721ULL, "gen_im_wep_war_one_lore_04.uti"},
	{0x801B5EBBD5406F73ULL, "729896295.ogg"},
	{0x801BD963AB4748ADULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mattock_sa.dds"},
	{0x801BFD3832D66AC2ULL, "430183434.ogg"},
	{0x801C2D3C25E04ADEULL, "6163349_m.fxe"},
	{0x801C491836822B76ULL, "one351ar_zevran_cave.are"},
	{0x801CBDD8CC38C563ULL, "vxdr_primevaltomb_766_0_0.vlm"},
	{0x801CFC3579948864ULL, "325984164.ogg"},
	{0x801D193BEE2D282DULL, "vxlt_gallowsprison_n_5wl_0_0.vlm"},
	{0x801D46C02A52C86DULL, "vxht_keep_n_73m_0_0.vlm"},
	{0x801DB7FA679B205BULL, "prty_varric_armor_1.dds"},
	{0x801E0F2949CA9F0CULL, "lt_0_lowtown_a3_n_c_1_0.msh"},
	{0x801E152949CAA9FEULL, "lt_0_lowtown_a3_n_c_1_0.mmh"},
	{0x801E20DABD7B01BDULL, "6029394_m.fxe"},
	{0x801E212949CABD95ULL, "lt_0_lowtown_a3_n_c_1_0.mao"},
	{0x801EB27BF6FE4FC6ULL, "gen00fl_aveline6137460.cl"},
	{0x801F1ECAF3B30532ULL, "dae000_gamlen6121502.cl"},
	{0x801F5B43F9CF23B5ULL, "6095344_m.fxe"},
	{0x801F7A874FA76B6DULL, "317131544.ogg"},
	{0x801FDB92315D13D6ULL, "character_character_p139.psh.fxo.dx11.assembly"},
	{0x80201801551E8141ULL, "458556059.ogg"},
	{0x80206EF4B6FC069DULL, "649523468.ogg"},
	{0x80207B8128F9D861ULL, "872678595.ogg"},
	{0x802093D481558BC9ULL, "649835752.ogg"},
	{0x802093E45D5FA2BBULL, "gen_im_arm_glv_hvy_act2_02.uti"},
	{0x8020E6656973307CULL, "al_warrior_s03.gda"},
	{0x802115073CB364CFULL, "6083103_m.fxe"},
	{0x80217865E1653FCFULL, "6144480_m.fxe"},
	{0x8021A15CD4CB894BULL, "378931693.ogg"},
	{0x8021A877B3FEAFC8ULL, "pro000_meeran6048824.cl"},
	{0x8021AC85FFF58561ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v10.gad"},
	{0x802202B1456577C4ULL, "rdr111dg_assistant6016624.cl"},
	{0x8022117826237736ULL, "zz_followers_banter6158739.cl"},
	{0x80226AA82000F05AULL, "6143192_m.fxe"},
	{0x8022A3EFF638834AULL, "one280_emrys6120037.cl"},
	{0x8022C8701BA21053ULL, "gen00fl_dog6167653.cl"},
	{0x802313688B424737ULL, "qcr221_cut_diplomatic_sol6042759.cl"},
	{0x8023ACE2532BEED3ULL, "6082298_m.fxe"},
	{0x8023E4E587DC85CBULL, "vxht_keep_3r3_0_0.vlm"},
	{0x80241613984E7B01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v1.ani"},
	{0x8024414C87778269ULL, "gen00fl_anders6144082.cl"},
	{0x802491B216816F8EULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.cess.ani"},
	{0x8024C68371CD2AE8ULL, "154803887.ogg"},
	{0x80253E2A95F6AB8EULL, "198013422.ogg"},
	{0x8025566A31F939D5ULL, "279387740.ogg"},
	{0x80258354CE1F4E38ULL, "829439999.ogg"},
	{0x80258BC6F6541748ULL, "659121954.ogg"},
	{0x8025A15CECEFB4F4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_window.phy"},
	{0x8025A564F876B396ULL, "665186709.ogg"},
	{0x80261694422219B6ULL, "105268899.ogg"},
	{0x802618540BDA0ACCULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_ext_c.impa.ani"},
	{0x8027A3F8CB797DC2ULL, "812538528.ogg"},
	{0x8027AD85FFF9F20FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_v10.ani"},
	{0x8029042A17A2B445ULL, "gen00fl_varric6201266.cl"},
	{0x802937C833A509F5ULL, "6087691_m.fxe"},
	{0x80299B26D856E4F5ULL, "6140932_m.fxe"},
	{0x802A21139852F9A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v1.gad"},
	{0x802A75AEC97FA58FULL, "vxlt_foundry_n_xo_0_0.vlm"},
	{0x802B48DBDF0777A9ULL, "vxw_bonepit_29n_0_0.vlm"},
	{0x802BEF1E2208EF6EULL, "6048460_m.fxe"},
	{0x802BF8F380198880ULL, "6173412_m.fxe"},
	{0x802C127CF6A926B0ULL, "6151360_m.fxe"},
	{0x802C876151E89767ULL, "lt_0_lowtown_43_0.rml"},
	{0x802CFD9BE27FA734ULL, "180145931.ogg"},
	{0x802D25540BE03F0EULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_ext_c.impa.gad"},
	{0x802DADEF55B71D7CULL, "vxlt_0_lowtown_docks_c3z_0_0.vlm"},
	{0x802E5257AD4E7FC6ULL, "844683702.ogg"},
	{0x802E6A2D68631DDAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxc.mmh"},
	{0x802E6C2E6DA79074ULL, "qun230cr_guild_contact.utc"},
	{0x802E6E2D68632411ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxc.mao"},
	{0x802E7050EEC47DFBULL, "vxlt_hangedman_211_0_0.vlm"},
	{0x802E7C2D68633C20ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxc.msh"},
	{0x802EF133A892DD09ULL, "llac_dr_slavers_3216_5.dds"},
	{0x802F7513EB185536ULL, "737914830.ogg"},
	{0x802F7E5CECF7CDF1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_window.msh"},
	{0x802F8C5CECF7E50CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_window.mao"},
	{0x802F985CECF7F9AFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww12_window.mmh"},
	{0x802FA72A5D864E5BULL, "vxlt_gallowstemplar_n_3rk_0_0.vlm"},
	{0x802FD22DC4168B68ULL, "185516422.ogg"},
	{0x802FE180D1986574ULL, "269810909.ogg"},
	{0x802FFB596AE5262BULL, "gen00fl_aveline6139469.cl"},
	{0x8030143603FC723CULL, "859307477.ogg"},
	{0x8030F16EB325088BULL, "gen00fl_isabela6123229.cl"},
	{0x8030FE139858DC88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v1.evt"},
	{0x8030FEEEEA5130ABULL, "dre120tr_spider_appear.ncs"},
	{0x80310FC989E6924AULL, "ch2multnt.psh.fxo"},
	{0x803138590B6DBD9FULL, "6111006_m.fxe"},
	{0x80313CC5DD8E6A94ULL, "qun120_arishok6051757.cl"},
	{0x80314E7462E45732ULL, "6069652_m.fxe"},
	{0x803155929AF23C05ULL, "art\\vfx\\combat\\vfx_r_ibl_activation_c.anb"},
	{0x80316FAFABB35CB4ULL, "6167343_m.fxe"},
	{0x803189B2168BECDFULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.cess.evt"},
	{0x803256C279D29E4BULL, "vxht_keep_3y2_0_0.vlm"},
	{0x8032A390B10C4F39ULL, "hlac_lt_warehouse_n_65537_3z.dds"},
	{0x8032BDB7650114AEULL, "aip_arcanehorror_base.gda"},
	{0x80330C8EBBA1B8F7ULL, "sst_ss_crazy_f_1.gda"},
	{0x803320540BE4A2EDULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_ext_c.impa.evt"},
	{0x80332F270955D07DULL, "vxht_0_hightown_n_bu6_0_0.vlm"},
	{0x80338AFDA8B411F9ULL, "6188706_f.fxe"},
	{0x8033B2029F37C01EULL, "vxlt_0_lowtown_a2_n_c_gxe_0_0.vlm"},
	{0x8033E684907A36AEULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_4y.dds"},
	{0x80340CDE8FBDDAFDULL, "vxlt_gallowstemplar_n_457_0_0.vlm"},
	{0x80349CE355650A51ULL, "vxht_0_hightown_a3_c8b_0_0.vlm"},
	{0x8034E32E4849CCD7ULL, "lt_0_lowtown_54_0.rml"},
	{0x80351C4A01961EB8ULL, "mag100_cullen6171207.cl"},
	{0x8035DAB38DB5E062ULL, "471879653.ogg"},
	{0x8035E11A09FE0817ULL, "hlac_ca_caverns1_24424_28x.dds"},
	{0x8036155C6F19F633ULL, "44404254.ogg"},
	{0x8036259093179BFAULL, "llac_ht_chantry_n_2534_0.dds"},
	{0x803696CEF6507888ULL, "188839812.ogg"},
	{0x8036A46D2DC88543ULL, "hlac_w_encounter2_camp_196613_0y.dds"},
	{0x803708D261055FF1ULL, "6167293_f.fxe"},
	{0x80370D2582CBB2ACULL, "vxlt_gallowstemplar_34m_0_0.vlm"},
	{0x803757733EAA12DCULL, "lgt220_mag_apost_apost26156744.cl"},
	{0x80376467BAEB0CE3ULL, "6060496_m.fxe"},
	{0x803780BB4DB79D0CULL, "llac_dr_primevaltomb_665_1.dds"},
	{0x8037954693C368DBULL, "6126552_m.fxe"},
	{0x8037E875439A14D9ULL, "qun120_arishok6051303.cl"},
	{0x80386AB21691D6C8ULL, "art\\vfx\\combat\\vfx_c_mer_grab_c.cess.gad"},
	{0x803873BFE011D427ULL, "718303533.ogg"},
	{0x80388807870E067BULL, "vxca_caverns1_mines_fmr_0_0.vlm"},
	{0x8038E78BBB75BE0FULL, "vxlt_foundry_n_sn_0_0.vlm"},
	{0x8038EADD1DE9D4C6ULL, "zz_and_debug6178545.cl"},
	{0x803907595B4B12E2ULL, "gen00fl_merrill6129668.cl"},
	{0x8039122803830FAAULL, "6151263_m.fxe"},
	{0x80393E3EF9CA61C6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_ex_charge.evt"},
	{0x803A0BED42CAA3D9ULL, "vxlt_0_lowtown_a2_bxr_0_0.vlm"},
	{0x803A3D6B2A124BA8ULL, "vxlt_hangedman_n_2i5_0_0.vlm"},
	{0x803A60FB13FF07EEULL, "6043560_m.fxe"},
	{0x803AB7202B3D2CACULL, "6115396_f.fxe"},
	{0x803AE4C69940C79FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.gad"},
	{0x803AE91F91067212ULL, "6038597_m.fxe"},
	{0x803B571BCE419D97ULL, "dre000_varric6036951.cl"},
	{0x803B58B052600103ULL, "dae000_unhappy_mage6165318.cl"},
	{0x803BBCCBB5C14444ULL, "6180469_m.fxe"},
	{0x803C3F95678F718FULL, "vxlt_gallowstemplar_f_5wv_0_0.vlm"},
	{0x803C9C91AD237D41ULL, "gen00fl_carver6097405.cl"},
	{0x803D705F08F250A7ULL, "mag100_karras6034520.cl"},
	{0x803D94D2867C0849ULL, "6071194_f.fxe"},
	{0x803DFA1D1B5784EAULL, "fx_glow02.dds"},
	{0x803E36D238B8C655ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\tre_stump01_sa.dds"},
	{0x803EA03173ECE6FEULL, "vxlt_hangedman_n_se_0_0.vlm"},
	{0x803F40AF83C842ACULL, "6141206_m.fxe"},
	{0x803F7BAAF810A13CULL, "6046961_m.fxe"},
	{0x804040161EFB08DCULL, "hlac_dr_0_deeproads_5845_3y.dds"},
	{0x8040553EF9D0A711ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_ex_charge.gad"},
	{0x8040763101593027ULL, "gen_im_wep_war_two_act0_04.uti"},
	{0x804136A8F6A3D019ULL, "zz_dae_debug6036582.cl"},
	{0x8041865DEF180FDAULL, "llac_lt_0_lowtown_a2_n_65536_10.dds"},
	{0x8041A103FB4BE8FDULL, "vxht_0_hightown_n_bp5_0_0.vlm"},
	{0x8041D9C69946D38CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.evt"},
	{0x8042501836A30641ULL, "one351ar_zevran_cave.lst"},
	{0x80427BD0492EB02EULL, "lgt150_nat_were_supporter6110981.cl"},
	{0x80427DBAA330232AULL, "404850793.ogg"},
	{0x804365E646C7F3FFULL, "264528653.ogg"},
	{0x8043A3175F81B12AULL, "lt_0_lowtown_a2_n_c_52_0.anb"},
	{0x8043A7E35AA57E54ULL, "gen00fl_fenris6116567.cl"},
	{0x804476447482DA66ULL, "88820002.ogg"},
	{0x8045734AF89F2277ULL, "zz_ave_debug6207519.cl"},
	{0x8045D08DD01A241BULL, "6103603_m.fxe"},
	{0x8046A2601819D2A4ULL, "vxdr_primevaltomb_50g_0_2.vlm"},
	{0x8046A3B396FA46F4ULL, "6153775_m.fxe"},
	{0x80470861F656E4B6ULL, "vxlt_0_lowtown_a2_n_c_gc8_0_0.vlm"},
	{0x8047563EF9D6C7BFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_ch_ex_charge.ani"},
	{0x80477047273E029DULL, "ht_0_hightown_12_0.rml"},
	{0x80477DDDB5A4C405ULL, "6177923_m.fxe"},
	{0x80479E39816948B3ULL, "art\\characters\\creatures\\desiredemon\\animation\\crn_rdy.evt"},
	{0x8047A89C90ADC97CULL, "74445448.ogg"},
	{0x8047B8149FA6CCB6ULL, "6066709_m.fxe"},
	{0x8047C6BF25953A1AULL, "vxw_sundermount_2c0_0_0.vlm"},
	{0x8048481836A76438ULL, "one351ar_zevran_cave.ncs"},
	{0x8048484422046959ULL, "vxlt_warehouse_n_135_0_0.vlm"},
	{0x8048FD8898376926ULL, "6117944_m.fxe"},
	{0x8049437DF555204EULL, "6101385_m.fxe"},
	{0x8049BC61D9DAE7ABULL, "6187870_m.fxe"},
	{0x804A4360C428D946ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1.evt"},
	{0x804A9047A25DCD73ULL, "llac_lt_0_lowtown_a2_n_c_-196608_0.dds"},
	{0x804B05CBF57DE0EEULL, "vxw_bonepit_13d_0_0.vlm"},
	{0x804B840C00BCD4B8ULL, "941667175.ogg"},
	{0x804BD0DEAE01AB58ULL, "6120908_m.fxe"},
	{0x804C30AFED79DEAEULL, "298350483.ogg"},
	{0x804CA12F60DF5D96ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wmda_0t.dds"},
	{0x804D2D8A7D1D871CULL, "vxlt_warehouse_n_15j_0_0.vlm"},
	{0x804D8DA4635570E3ULL, "hlac_lt_0_lowtown_a2_n_-65534_0y.dds"},
	{0x804DDCED0FA1E108ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottle_n.dds"},
	{0x804DED87EA06B9BCULL, "6041861_m.fxe"},
	{0x804E0FEDF206C137ULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_swd01.phy"},
	{0x804E1A06BD3FB126ULL, "dae000_flemeth6000905.cl"},
	{0x804E22AEBA3B6682ULL, "6130011_m.fxe"},
	{0x804E9739816F5BF4ULL, "art\\characters\\creatures\\desiredemon\\animation\\crn_rdy.gad"},
	{0x804F03EEEA6A3692ULL, "dre120tr_spider_appear.utt"},
	{0x804F9B6792FEF1F7ULL, "6180941_m.fxe"},
	{0x804FC149CC952CB7ULL, "gen_im_trs_dal_blt_r1.uti"},
	{0x804FC1C69952E8E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_enter.ani"},
	{0x804FF08AD93A0299ULL, "hlac_dr_0_deeproads_5842_15z.dds"},
	{0x80501EF006D9E14AULL, "740284810.ogg"},
	{0x80505828FFCE6D8EULL, "55726744.ogg"},
	{0x80511458ECF2AA1EULL, "zz_mcr_debug6080854.cl"},
	{0x80515A60C42F1E91ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1.gad"},
	{0x8051672589DC967AULL, "w_bonepit_220_0.msh"},
	{0x8051752589DCADB3ULL, "w_bonepit_220_0.mao"},
	{0x8051792589DCB4C0ULL, "w_bonepit_220_0.mmh"},
	{0x80519E4B306236FCULL, "hlac_w_encounter1_0_1x.dds"},
	{0x80524EFC82710D96ULL, "vxca_caverns2prc_91r_0_1.vlm"},
	{0x8052569FB1F61700ULL, "pro000_refugee_female6052090.cl"},
	{0x80533DDA914B616DULL, "982615767.ogg"},
	{0x80543C86ABD6628FULL, "ser221tr_chase_jump2.utt"},
	{0x80545ECC5A07456BULL, "vxlt_darktown_5ce_0_0.vlm"},
	{0x805499D5CEFAED72ULL, "6208556_m.fxe"},
	{0x8054BA7AC76D9C58ULL, "vxw_blightlands_90_0_0.vlm"},
	{0x80550C54F1A07A2AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.down.gad"},
	{0x8055A639817593D2ULL, "art\\characters\\creatures\\desiredemon\\animation\\crn_rdy.ani"},
	{0x805602ACBBE5575FULL, "campaign_base_es-es.cre"},
	{0x8056389C178B529AULL, "vxw_sundermount_2d1_0_0.vlm"},
	{0x8056BA2113FA81D9ULL, "vxlt_warehouse_n_184_0_0.vlm"},
	{0x8057F97AD8A4BA69ULL, "mag160_varian_ilithis6171454.cl"},
	{0x80585B60C4353F3FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v1.ani"},
	{0x80587F584C96A686ULL, "434476231.ogg"},
	{0x8058A4C3C555307FULL, "vxlt_gallowstemplar_f_5v5_0_0.vlm"},
	{0x8058D959D2818BB1ULL, "hlac_dr_primevaltomb_661_0x.dds"},
	{0x805956929B139046ULL, "art\\vfx\\combat\\vfx_r_ibl_activation_c.mmh"},
	{0x80597CE2DAA1184EULL, "6055596_m.fxe"},
	{0x8059BA68A52542F7ULL, "799509723.ogg"},
	{0x805A052E37E8D18CULL, "6087746_m.fxe"},
	{0x805A08760192CF67ULL, "vxca_cavern1_haunted_icm_0_2.vlm"},
	{0x805A2B60CCED58FDULL, "vfx_vfx_p2.psh.fxo.dx11"},
	{0x805A2DF22DB04804ULL, "350202390.ogg"},
	{0x805A487936C9E27AULL, "vxw_encounter2_a1_0_0.vlm"},
	{0x805A5605AE6F5407ULL, "ico_2h_axe_3.dds"},
	{0x805A63D867DBF963ULL, "387328862.ogg"},
	{0x805A782BEA0C849BULL, "6195863_f.fxe"},
	{0x805AF5EA91E36189ULL, "vxca_cave_2fq_0_1.vlm"},
	{0x805B948223FAE9A7ULL, "vxdr_primevaltomb_3th_0_0.vlm"},
	{0x805C4772CB8FBB20ULL, "6108794_f.fxe"},
	{0x805C9A9D3AA93DAEULL, "gen_im_acc_blt_act1_06.uti"},
	{0x805D1A5C7CA21899ULL, "397879781.ogg"},
	{0x805E329948BA845AULL, "435128779.ogg"},
	{0x805E33AD68C493BAULL, "lt_backalley_n_1_0_lc.mmh"},
	{0x805E9AF093081D08ULL, "plt_lgt160pt_wk_sophia.nss"},
	{0x805EEC294A01FA99ULL, "lt_0_lowtown_a3_n_c_1_0.phy"},
	{0x805F107CCE53287EULL, "6021686_f.fxe"},
	{0x805F2A194BAE7EC2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\til_hightown_main_drelief.dds"},
	{0x805F50F9CCF8C60AULL, "vxht_0_hightown_a2_cb6_0_0.vlm"},
	{0x805F81EAB7F9F92DULL, "w_blightlands_55_0.msh"},
	{0x805F8FEAB7FA1160ULL, "w_blightlands_55_0.mao"},
	{0x805F93EAB7FA17B3ULL, "w_blightlands_55_0.mmh"},
	{0x805FF67BB035A77BULL, "t3_rob_mer.tnt"},
	{0x806051F23733223CULL, "lgt310_mis_burn_murder16152754.cl"},
	{0x8060817888F0CDA0ULL, "hlac_lt_0_lowtown_docks_n_-131076_0z.dds"},
	{0x806122D0F322F76DULL, "one200_gold_followers6117249.cl"},
	{0x806155296DEDEFC1ULL, "vxw_bonepit_xy_0_0.vlm"},
	{0x806164414BAEC8A5ULL, "mag291_ntbanter6116203.cl"},
	{0x8061917D69B3564BULL, "vxlt_undercityserialkiller_xu_0_0.vlm"},
	{0x806234FB31D84350ULL, "hlac_lt_0_lowtown_a2_65537_1x.dds"},
	{0x80629AA7771A15B5ULL, "zz_mrl_debug6184187.cl"},
	{0x80630DBBF1377FA8ULL, "hlac_lt_0_lowtown_a2_n_c_131072_4z.dds"},
	{0x806327D7AD668F71ULL, "6142325_m.fxe"},
	{0x806335799D0446B2ULL, "28932354.ogg"},
	{0x80633FA48D86ED00ULL, "vxlt_warehouse_10r_0_0.vlm"},
	{0x8063D3437D9B003EULL, "vxht_0_hightown_c8t_0_0.vlm"},
	{0x8064A5F8E9015BD4ULL, "6007719_m.fxe"},
	{0x8064AA7909816B1AULL, "vxw_sundermount_2i2_0_0.vlm"},
	{0x8064CC5A04C80C6EULL, "vxw_woundedcoast_1p_0_0.vlm"},
	{0x806507E39DD73DD3ULL, "249527624.ogg"},
	{0x8065597903334401ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.cess.ani"},
	{0x806578D3CF4FAF92ULL, "837476427.ogg"},
	{0x80658CE38236E8C5ULL, "vxlt_0_lowtown_a3_n_c_hqw_0_0.vlm"},
	{0x80658F30C3A9F8ADULL, "vxw_encounter2_camp_2w_0_0.vlm"},
	{0x8065A3F0930E4B1BULL, "plt_lgt160pt_wk_sophia.plo"},
	{0x8065A9D1927969E9ULL, "vxlt_0_lowtown_a2_n_c_fe0_0_0.vlm"},
	{0x80666CCC7AA663A5ULL, "6082322_m.fxe"},
	{0x8067BC59AA57D0DCULL, "hlac_ht_mansion2_818_49y.dds"},
	{0x8068025BE9AD90F9ULL, "vxlt_foundry_n_v7_0_0.vlm"},
	{0x806815906E19900EULL, "mrl311ip_mirror_shattered.utp"},
	{0x8068742E6DD9159DULL, "qun230cr_guild_contact.bnk"},
	{0x806897F7AEA08D47ULL, "hlac_ht_0_hightown_65535_0z.dds"},
	{0x8068BA5628BFFAFAULL, "vxw_encounter2_l2_0_0.vlm"},
	{0x8068E754F1B0E410ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.down.ani"},
	{0x806A60A3145D15BBULL, "485704964.ogg"},
	{0x806ACF5D5F9CD4F8ULL, "gen_im_npc_wep_one_swda.uti"},
	{0x806B18904FFE21F2ULL, "art\\cinematics\\act1\\cs_fp_closedsneer.gad"},
	{0x806B30CD2FD687E4ULL, "770258694.ogg"},
	{0x806B64790337C2A3ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.cess.gad"},
	{0x806BA51AEBF3AB59ULL, "vxht_0_hightown_a2_9h1_0_0.vlm"},
	{0x806BC144BDA4E7DEULL, "vxca_caverns1_id2_0_0.vlm"},
	{0x806C562E6DDCCBA5ULL, "qun230cr_guild_contact.cnv"},
	{0x806C5EA5D981E1CEULL, "fighting60b.mao"},
	{0x806C8A61521E807BULL, "lt_0_lowtown_43_0.anb"},
	{0x806CB557383FAC65ULL, "131374507.ogg"},
	{0x806CFA43F489349EULL, "one110_meeran6090192.cl"},
	{0x806CFFA5623D2FE6ULL, "hlac_lt_gallowstemplar_n_2384_5y.dds"},
	{0x806D5735A7F3A331ULL, "277722978.ogg"},
	{0x806DC4C0E0EF3C66ULL, "vxca_caverns1_dpc_0_0.vlm"},
	{0x806DC67A96A06174ULL, "6087625_m.fxe"},
	{0x806EA7AB7A975DC0ULL, "426397229.ogg"},
	{0x806F12226BF6C519ULL, "zz_mer_debug6062729.cl"},
	{0x806F1B3CF079AF59ULL, "6080614_m.fxe"},
	{0x806F6A481AE2A77BULL, "6107540_m.fxe"},
	{0x806F7EA870F40853ULL, "horizontalblurshadows.vsh.fxo"},
	{0x8070044437D81ED3ULL, "hlac_lt_0_lowtown_n_-65535_2z.dds"},
	{0x8070205C4905D7BFULL, "gen_im_flw_two_carver_01.uti"},
	{0x80703985F35408E1ULL, "zz_dae_debug6050735.cl"},
	{0x8070944FD786E332ULL, "vxht_keep_n_60v_0_0.vlm"},
	{0x8070E9617264BB66ULL, "6088248_m.fxe"},
	{0x807119ACAC4CA075ULL, "363454864.ogg"},
	{0x80718AD6AB38BF77ULL, "vxlt_0_lowtown_docks_a3_cz9_0_0.vlm"},
	{0x80724179033DA588ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.cess.evt"},
	{0x8072440FD2831BDDULL, "llac_lt_0_lowtown_docks_a3_131071_3.dds"},
	{0x80724930D24B949DULL, "hlac_ht_keep_n_4519_10y.dds"},
	{0x8072D7938429FB11ULL, "vxht_mansion2b_2c3_0_0.vlm"},
	{0x807372A477F7A9E1ULL, "455857605.ogg"},
	{0x807398F2E9171CF7ULL, "vxca_caverns2_7tv_0_0.vlm"},
	{0x8073A09241858581ULL, "676725017.ogg"},
	{0x80745AB583EC6E79ULL, "mer341_jansen6065741.cl"},
	{0x807522F10F596FDDULL, "gen00fl_aveline6115393.cl"},
	{0x80752D37C182ED08ULL, "vxw_woundedcoast_vo_0_0.vlm"},
	{0x80759A138922EA8DULL, "193296904.ogg"},
	{0x8075E713E8D918E0ULL, "vxht_mansion2_2fi_0_0.vlm"},
	{0x807639A8B8C3B7C2ULL, "art\\characters\\playercharacter\\humanfemale\\cth_zbm.msh"},
	{0x80763BA8B8C3BB78ULL, "art\\characters\\playercharacter\\humanfemale\\cth_zbm.mmh"},
	{0x807647A8B8C3CF1BULL, "art\\characters\\playercharacter\\humanfemale\\cth_zbm.mao"},
	{0x8076B870CB2360E4ULL, "vxht_0_hightown_n_av9_0_0.vlm"},
	{0x8076DE43A5E2F2DDULL, "vxw_bonepitdestroyed_a3_p5_0_0.vlm"},
	{0x8076E202AEA9D341ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_down.ani"},
	{0x8076E61D49C5AFC2ULL, "6053860_m.fxe"},
	{0x80784A639624F288ULL, "vxw_bonepit_1zy_0_0.vlm"},
	{0x8079232B93913D82ULL, "6098209_m.fxe"},
	{0x8079BDBCED577D57ULL, "vxht_mansion2_1t7_0_0.vlm"},
	{0x807A6F4727695AB5ULL, "ht_0_hightown_12_0.anb"},
	{0x807AF111BC72DE45ULL, "zz_followers_banter6157700.cl"},
	{0x807B3915E6794D19ULL, "vxw_encounter2_2c_0_0.vlm"},
	{0x807B5CD5A2BD91ECULL, "w_encounter2_camp_23_0.phy"},
	{0x807B6877C3244AFDULL, "vxht_0_hightown_n_bl1_0_0.vlm"},
	{0x807C246DA83A3D09ULL, "plt_vrc000pt_n_nt_completed.nss"},
	{0x807C26C318AE8144ULL, "6114837_m.fxe"},
	{0x807C320A682B5DE7ULL, "293266784.ogg"},
	{0x807C8C4859E5F431ULL, "6080218_m.fxe"},
	{0x807CA3F69C8508A8ULL, "6077676_f.fxe"},
	{0x807CCB2402B5FA2DULL, "vxca_caverns1_j4v_0_0.vlm"},
	{0x807CF5DB4BD0CB2DULL, "6080208_f.fxe"},
	{0x807DED02AEB004E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_down.gad"},
	{0x807F49C99D282FE2ULL, "6047393_m.fxe"},
	{0x807F5B3E645AD3E6ULL, "6083348_m.fxe"},
	{0x807FB4632FE98EEBULL, "vxlt_darktown_5rf_0_0.vlm"},
	{0x807FC3F3D1E8AF82ULL, "zz_ave_debug6207725.cl"},
	{0x8080039050105998ULL, "art\\cinematics\\act1\\cs_fp_closedsneer.ani"},
	{0x808042A8B8CC1B03ULL, "art\\characters\\playercharacter\\humanfemale\\cth_zbm.phy"},
	{0x8080A4109EA79DCBULL, "fen110_fenris6117959.cl"},
	{0x8080B6FD6187313EULL, "vxht_0_hightown_c2v_0_0.vlm"},
	{0x8080D805C0B72B73ULL, "6136724_m.fxe"},
	{0x8080ED5CCEAB2D7DULL, "hlac_lt_0_lowtown_a2_n_c_65534_0x.dds"},
	{0x8081C0B5EE50BB02ULL, "zz_mcr_debug6088339.cl"},
	{0x80821316547859C3ULL, "lgt110_gxa_coa_meeting6117509.cl"},
	{0x808233D5A2C36AF4ULL, "w_encounter2_camp_23_0.rml"},
	{0x80829144F6F6743CULL, "6080440_m.fxe"},
	{0x8082CA2955E68165ULL, "161859885.ogg"},
	{0x80839F14B3790588ULL, "vxw_woundedcoast_sl_0_0.vlm"},
	{0x8083F4898FF78855ULL, "hlac_ht_playershouse_n_855_10y.dds"},
	{0x80840C0015F4040BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge04_n.dds"},
	{0x80841F37D3A23612ULL, "hla_ht_0_hightown_65534_0y.dds"},
	{0x80844C447FBEB0A7ULL, "vxht_brothel_n_4bd_0_0.vlm"},
	{0x8084CA02AEB5E7C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_right_down.evt"},
	{0x8084D4334663182EULL, "hlac_ht_0_hightown_n_-131073_1y.dds"},
	{0x8084DD163C68409FULL, "865016018.ogg"},
	{0x8085132249007548ULL, "qcr000pt_general.ncs"},
	{0x8085495546E9D8A6ULL, "6077325_m.fxe"},
	{0x8085687F0E11C57EULL, "vxht_mansion_n_5bk_0_0.vlm"},
	{0x8085D03CCA23037DULL, "6170166_m.fxe"},
	{0x8087269F327995F6ULL, "6163873_m.fxe"},
	{0x80873C6D66107464ULL, "hlac_lt_undercitydungeon_-1_22y.dds"},
	{0x808783A410C97859ULL, "6125255_f.fxe"},
	{0x808859FF63C33370ULL, "dae000_quintus6155249.cl"},
	{0x808986707CB90877ULL, "vxlt_0_lowtown_n_bxo_0_0.vlm"},
	{0x8089FE4F02B39A19ULL, "vxw_blightlands_au_0_0.vlm"},
	{0x808A530402E4267DULL, "6163500_m.fxe"},
	{0x808A70D99ACE3A67ULL, "447683317.ogg"},
	{0x808B3C6D9FEA792CULL, "vxht_0_hightown_a2_a4t_0_0.vlm"},
	{0x808B423F47910B5AULL, "6025354_f.fxe"},
	{0x808B67C902C927E2ULL, "6178877_m.fxe"},
	{0x808C051433C80B2AULL, "dae000_epic_merchant_26179371.cl"},
	{0x808C0FF72361B615ULL, "vxht_mansion2b_1fo_0_0.vlm"},
	{0x808CA1B01ECF6F3AULL, "llac_dr_slavers_3217_3.dds"},
	{0x808DFCA7F7930AD0ULL, "vxlt_0_lowtown_docks_a3_n_cwz_0_0.vlm"},
	{0x808E7A6EB9B6D857ULL, "one280_emrys6116800.cl"},
	{0x808EA4A9FBCD0977ULL, "lgt210_magu_selby6169028.cl"},
	{0x808F4AA84A9C1BEEULL, "zz_one_debug6215795.cl"},
	{0x808F54A53B78CEFEULL, "hlac_lt_foundry_n_320_20y.dds"},
	{0x808F59363525099AULL, "6089904_m.fxe"},
	{0x808F6BB6B55C551EULL, "art\\cinematics\\additional animation\\cs_femnobleiv_cinematicanim.ani"},
	{0x8090138B9C04E9CCULL, "6112141_m.fxe"},
	{0x809018B58C6AF3C2ULL, "w_woundedcoast_216_0.msh"},
	{0x80901AB58C6AF778ULL, "w_woundedcoast_216_0.mmh"},
	{0x809026B58C6B0B1BULL, "w_woundedcoast_216_0.mao"},
	{0x80902AF1C2ACF928ULL, "359279911.ogg"},
	{0x80903AB3EC743F2BULL, "plt_cod_bks_lite_vials.plo"},
	{0x80903DF3DA49AF6BULL, "dae011_corff_sta6164626.cl"},
	{0x80905DE48C908161ULL, "19086689.ogg"},
	{0x809085D17E4BAF60ULL, "55658968.ogg"},
	{0x8090C74E02FD8D97ULL, "vxlt_0_lowtown_a3_n_c_cif_0_0.vlm"},
	{0x8090D73F51EC52C6ULL, "vxht_keep_3xh_0_0.vlm"},
	{0x809131B15E2CF9AEULL, "one101_orders_followers.cnv"},
	{0x8091E4BA3C2BC818ULL, "vxlt_0_lowtown_n_e8c_0_0.vlm"},
	{0x8092038101AB713CULL, "6013369_m.fxe"},
	{0x80926C378D2D927FULL, "vxlt_gallowstemplar_f_5j1_0_0.vlm"},
	{0x80929E59B2252D29ULL, "art\\characters\\playercharacter\\humanmale\\textures\\arm_bdaa_0d.dds"},
	{0x8092B91447AE30B7ULL, "fen210_fenris6108743.cl"},
	{0x8092BB96D85F137CULL, "gen00fl_merrill6128654.cl"},
	{0x8093B1FA2D348A45ULL, "vxw_encounter2_l9_0_0.vlm"},
	{0x809421DD8CE3E79BULL, "vxw_woundedcoast_16e_0_0.vlm"},
	{0x809453B15E2F691EULL, "one101_orders_followers.bnk"},
	{0x8094778942789F4EULL, "108234972.ogg"},
	{0x8094E4EDF2427A8FULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_swd01.mao"},
	{0x8094F0EDF2428F2CULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_swd01.mmh"},
	{0x8094F6EDF242991EULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_swd01.msh"},
	{0x80953186C77078AFULL, "274668448.ogg"},
	{0x80956FDEB2096186ULL, "gen00fl_fenris6107178.cl"},
	{0x8095B7DA0E0FA2FEULL, "gen00fl_anders6125231.cl"},
	{0x8095BFB6F5FFE13CULL, "qun110_petrice6037696.cl"},
	{0x80960B810AF02319ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\_textures\\vst_gallows_n.dds"},
	{0x8096290067594341ULL, "vxlt_0_lowtown_a3_n_c_e8e_0_0.vlm"},
	{0x8096549355A5A9ABULL, "mag330_walter6045627.cl"},
	{0x80965F234CC8C2B4ULL, "lt_0_lowtown_a2.rml"},
	{0x8096A176D143AE57ULL, "vxht_mansion2_1b1_0_0.vlm"},
	{0x8096F83D6D6A78ABULL, "mag311_nyssa6042127.cl"},
	{0x80970B6C67D484E3ULL, "600655399.ogg"},
	{0x8097493332C115D9ULL, "vxlt_0_lowtown_a3_n_c_ejg_0_0.vlm"},
	{0x80976CEC8C8F59D2ULL, "isa211_isabela6070198.cl"},
	{0x809799985B674583ULL, "gen_im_arm_glv_lgt_act3_03.uti"},
	{0x8097BE4F803478D4ULL, "hlac_ht_mansion2_818_11x.dds"},
	{0x809816C8F07B11A2ULL, "vxlt_0_lowtown_n_gpc_0_0.vlm"},
	{0x809818E2DC566645ULL, "6207662_m.fxe"},
	{0x80981CCFCA657E19ULL, "vxw_encounter2_8a_0_0.vlm"},
	{0x8098702BF4A9B299ULL, "vxw_blightlands_jt_0_0.vlm"},
	{0x8098B6F77D7C07DEULL, "dae300_cullen6135967.cl"},
	{0x809947056A37E8D8ULL, "hla_ht_0_hightown_a2_65538_1x.dds"},
	{0x80996146AF807530ULL, "hlac_ht_0_hightown_n_0_0x.dds"},
	{0x8099829A128BCF13ULL, "fen110_fenris6100308.cl"},
	{0x8099D9F81E210393ULL, "vxlt_undercityserialkiller_1wc_0_0.vlm"},
	{0x8099EFADFD57DB58ULL, "337128934.ogg"},
	{0x809A09DF8424A657ULL, "t3_dog_d07.tnt"},
	{0x809A21B58C735703ULL, "w_woundedcoast_216_0.phy"},
	{0x809A63277B4FF073ULL, "6168172_m.fxe"},
	{0x809B2B151A77A6BCULL, "mer340_bonepit_comments6083142.cl"},
	{0x809C298D14885B8FULL, "6069837_m.fxe"},
	{0x809C8CEBB93A368FULL, "6166754_m.fxe"},
	{0x809CA0D042B549B7ULL, "6101379_m.fxe"},
	{0x809CEA5474988EE2ULL, "6072443_m.fxe"},
	{0x809D266FB3E0B4D2ULL, "307579579.ogg"},
	{0x809DC3CD3B253426ULL, "77781655.ogg"},
	{0x809E293FE75963DFULL, "llac_lt_0_lowtown_-262144_2.dds"},
	{0x809E2C78ABA6FB61ULL, "hlac_lt_0_lowtown_a3_n_c_65534_2y.dds"},
	{0x809E6A7EEAAAC842ULL, "vxlt_backalley_g_1vm_0_0.vlm"},
	{0x809E6F803145FE19ULL, "307290103.ogg"},
	{0x809F436B6B6AAB02ULL, "mrl000_marethari6129322.cl"},
	{0x809F8D7379B1EC7EULL, "vxlt_hangedman_1qk_0_0.vlm"},
	{0x809FE51A54451819ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16a.msh"},
	{0x809FEF1A54452907ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16a.mmh"},
	{0x80A04C56D154A061ULL, "56592921.ogg"},
	{0x80A0909155B1FDFBULL, "444356272.ogg"},
	{0x80A0AA8AFEA4BF65ULL, "vxht_mansion2b_1h1_0_0.vlm"},
	{0x80A187EF5BC09BC0ULL, "gen00fl_fenris6126784.cl"},
	{0x80A18B1555C60B73ULL, "443969389.ogg"},
	{0x80A18D60B62255AFULL, "vxlt_warehouse_n_8a_0_0.vlm"},
	{0x80A1903C4AA33BD1ULL, "6113041_f.fxe"},
	{0x80A1F123CEADE4F5ULL, "6043714_m.fxe"},
	{0x80A24C38F1FDF67EULL, "vxht_mansion_n_5hm_0_0.vlm"},
	{0x80A256F9C57311B6ULL, "follower_banter6161157.cl"},
	{0x80A293BA7EDA001BULL, "vxw_woundedcoast_11d_0_0.vlm"},
	{0x80A2D13D6029EE04ULL, "gen00fl_anders6144108.cl"},
	{0x80A2DC5615F96259ULL, "genip_bones_pile_2.utp"},
	{0x80A31D5F4A9B55E6ULL, "vxw_bonepit_2ae_0_0.vlm"},
	{0x80A36F26D710E8F9ULL, "650890672.ogg"},
	{0x80A384B6B56D2898ULL, "art\\cinematics\\additional animation\\cs_femnobleiv_cinematicanim.gad"},
	{0x80A38DF3D149B0FCULL, "combat_ws_range.gda"},
	{0x80A4403B4CBFE6BDULL, "eye_eyelash_p0.psh.fxo"},
	{0x80A4427FEE1EADD2ULL, "gen00fl_varric6076432.cl"},
	{0x80A5359A7A852D24ULL, "vxht_mansion_n_3ot_0_0.vlm"},
	{0x80A53712F3ADD632ULL, "zz_mag_debug6039785.cl"},
	{0x80A5738B34B223D3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo12_basic_1.mao"},
	{0x80A601723AEA4237ULL, "bld_mac_03a.mao"},
	{0x80A688074B9D8BE3ULL, "fx_c_twister.mao"},
	{0x80A6F8B00BE50979ULL, "6113929_m.fxe"},
	{0x80A733E28E0B0A26ULL, "vxht_keep_6h4_0_1.vlm"},
	{0x80A73F2DF6480E1BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairright.msh"},
	{0x80A7512DF6482D66ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairright.mao"},
	{0x80A75D2DF6484105ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairright.mmh"},
	{0x80A81C7F757D0791ULL, "base_light.msh"},
	{0x80A8367F757D344FULL, "base_light.mmh"},
	{0x80A866780BE47275ULL, "724440619.ogg"},
	{0x80A86783C3F1D135ULL, "178652750.ogg"},
	{0x80A895C0843CA484ULL, "lowtown_male_alienage1_sta6162802.cl"},
	{0x80A8B1F2AC33238BULL, "275392511.ogg"},
	{0x80A91B96E7A7B06BULL, "6118382_m.fxe"},
	{0x80A92EA84F082BDBULL, "dae000_bodhan6146892.cl"},
	{0x80A9C47121DFBFF1ULL, "hlac_w_bonepitmining_a2_393217_0x.dds"},
	{0x80A9CA90327A4D51ULL, "vxht_keep_4aw_0_0.vlm"},
	{0x80A9E11386D506D9ULL, "828803583.ogg"},
	{0x80AB0BB53C3BFBB2ULL, "hlac_ca_caverns1_24424_18z.dds"},
	{0x80ABDD5D613A7579ULL, "hlac_lt_0_lowtown_-196607_0z.dds"},
	{0x80ABDEA458B9CE7BULL, "ht_0_hightown_n_33_0_ll.mmh"},
	{0x80AD2A7D0680618EULL, "zz_docks_oceans11.cut"},
	{0x80AD6C447EA3F2D3ULL, "art\\vfx\\combat\\vfx_gib_qunw_lbrt.anb"},
	{0x80ADC6343C056345ULL, "vxlt_0_lowtown_a3_n_c_hjr_0_0.vlm"},
	{0x80AE2FF2239E51B9ULL, "6174953_m.fxe"},
	{0x80AE8B91BAC010C3ULL, "6127025_m.fxe"},
	{0x80AEC048D671C33DULL, "zz_dae_debug6029876.cl"},
	{0x80AECB97A0E33429ULL, "vxca_caverns1_gh3_0_0.vlm"},
	{0x80AEF6C28E5A2DB9ULL, "6034266_m.fxe"},
	{0x80AF1C67F09AD7E5ULL, "vxht_mansion2b_1c0_0_0.vlm"},
	{0x80AFCDF1A84B45AAULL, "537743621.ogg"},
	{0x80B0596DB3F858D7ULL, "and000st_100.stg"},
	{0x80B0A5B76DBB54DDULL, "vxw_bonepitdestroyed_a3_l1_0_0.vlm"},
	{0x80B0EB33B1692D5CULL, "vxdr_0_deeproads_eul_0_0.vlm"},
	{0x80B0F1B751EC114BULL, "gen00fl_anders6111143.cl"},
	{0x80B1580B8E0764C2ULL, "dae100_viscount6103412.cl"},
	{0x80B17946A60DBB1EULL, "6152734_m.fxe"},
	{0x80B1957A7D3A19E4ULL, "6144054_m.fxe"},
	{0x80B1BB213947C26EULL, "vxw_bonepitmining_a2_1on_0_0.vlm"},
	{0x80B211D75DFD5488ULL, "vxw_bonepit_1nu_0_0.vlm"},
	{0x80B213FE61475D32ULL, "481789399.ogg"},
	{0x80B2267C9B57CED6ULL, "dae210ip_store_weapons.utp"},
	{0x80B242D3FABC95EAULL, "924630431.ogg"},
	{0x80B2860D2B1AA9CFULL, "w_blightlands_96_0.mao"},
	{0x80B28E59A9F0E149ULL, "6091388_m.fxe"},
	{0x80B2920D2B1ABE6CULL, "w_blightlands_96_0.mmh"},
	{0x80B2980D2B1AC85EULL, "w_blightlands_96_0.msh"},
	{0x80B2C04601523EB8ULL, "follower_banter6170702.cl"},
	{0x80B33458FC501647ULL, "27089854.ogg"},
	{0x80B3475AC996D929ULL, "83271671.ogg"},
	{0x80B34A99028C27F4ULL, "vxlt_0_lowtown_a2_n_e2m_0_0.vlm"},
	{0x80B37F0F3DF4F06CULL, "llac_ht_chantry_n_2534_48.dds"},
	{0x80B46DBA750BBEB7ULL, "vxht_mansion2_1s2_0_0.vlm"},
	{0x80B49931EF5C5F6FULL, "65875420.ogg"},
	{0x80B4BECC509C96E0ULL, "vxht_0_hightown_a3_ap5_0_0.vlm"},
	{0x80B4C9922FDCBAE9ULL, "vxlt_backalley_g_1ka_0_0.vlm"},
	{0x80B4FA82D46742A2ULL, "vxlt_0_lowtown_n_gjm_0_0.vlm"},
	{0x80B553E5D895E399ULL, "vxw_blightlands_pz_0_0.vlm"},
	{0x80B6C19519FFDE90ULL, "w_bonepitdestroyed_a3_4_0_lc.mmh"},
	{0x80B70F3024C4D5DBULL, "llac_ca_cavern1_haunted_24425_0.dds"},
	{0x80B7658DF943FF95ULL, "vxht_mansion2b_1wl_0_0.vlm"},
	{0x80B7A858CC3E1865ULL, "258526942.ogg"},
	{0x80B82792AEA3042BULL, "6148831_m.fxe"},
	{0x80B85D33B4C46E4DULL, "gen00fl_aveline6106911.cl"},
	{0x80B8728FC760F0C4ULL, "6076474_m.fxe"},
	{0x80B89F17EF0F8670ULL, "art\\characters\\playercharacter\\elfmale\\glv_lgtp.mao"},
	{0x80B8A317EF0F8C83ULL, "art\\characters\\playercharacter\\elfmale\\glv_lgtp.mmh"},
	{0x80B8B117EF0FA4DDULL, "art\\characters\\playercharacter\\elfmale\\glv_lgtp.msh"},
	{0x80B930BDFE0B4176ULL, "hlac_lt_0_lowtown_a2_n_-131073_2x.dds"},
	{0x80B9DDE4F3D11079ULL, "697739550.ogg"},
	{0x80BA1137D5F8A61DULL, "qun110_petrice6042310.cl"},
	{0x80BA1452E5D82489ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout_2.mao"},
	{0x80BA4675ADFF8857ULL, "dae000ip_crowd_lm4_drunk1.utp"},
	{0x80BB0D6F5AAD4105ULL, "100420482.ogg"},
	{0x80BB4E38CE96F942ULL, "vxlt_backalley_g_1xo_0_0.vlm"},
	{0x80BB535CEA571468ULL, "vxht_mansion2b_2nq_0_0.vlm"},
	{0x80BBFF87040A4704ULL, "vxdr_primevalentrance_1k5_0_1.vlm"},
	{0x80BC2CD627CE9C46ULL, "vxht_keep_3ik_0_0.vlm"},
	{0x80BC33D5A2F4E2B7ULL, "w_encounter2_camp_23_0.mmh"},
	{0x80BC37D5A2F4E9C4ULL, "w_encounter2_camp_23_0.mao"},
	{0x80BC49D5A2F50829ULL, "w_encounter2_camp_23_0.msh"},
	{0x80BC6462FA95CB1EULL, "711996928.ogg"},
	{0x80BCE6A915C4881FULL, "846838969.ogg"},
	{0x80BD07B11885B32BULL, "223736112.ogg"},
	{0x80BD780885DB9BA9ULL, "910278990.ogg"},
	{0x80BEEF99B0A0D9CFULL, "6060395_m.fxe"},
	{0x80BF3195F1FE3EEBULL, "zz_approval_debug6003743.cl"},
	{0x80BFB24D2697E6CAULL, "hlac_w_sundermount_65532_0z.dds"},
	{0x80C0498AEA996577ULL, "vxlt_gamlenshouse_n_j1_0_0.vlm"},
	{0x80C09D738FBACFC6ULL, "993277393.ogg"},
	{0x80C109062316D0C5ULL, "t3_shd_rlr.tnt"},
	{0x80C13D6CB994F26BULL, "6135664_m.fxe"},
	{0x80C1571D3AB7C6B3ULL, "follower_banter6163123.cl"},
	{0x80C18D57100B4912ULL, "art\\vfx\\creatures\\fxc_bm_vomr_c.mmh"},
	{0x80C1A6394C50EC1FULL, "hlac_ca_caverns1_mines_24425_0z.dds"},
	{0x80C1FCD2902203ADULL, "follower_barks6153734.cl"},
	{0x80C27C234CEE2F66ULL, "lt_0_lowtown_a2.arl"},
	{0x80C280234CEE35B4ULL, "lt_0_lowtown_a2.anb"},
	{0x80C2D0304CD2897EULL, "vxlt_gallowsprison_n_5qa_0_0.vlm"},
	{0x80C310BC311F4194ULL, "vxlt_0_lowtown_docks_n_b5f_0_0.vlm"},
	{0x80C339E628C32924ULL, "786412774.ogg"},
	{0x80C36C5FC65D5B22ULL, "vxlt_0_lowtown_n_gon_0_0.vlm"},
	{0x80C3911093406497ULL, "6105656_m.fxe"},
	{0x80C3C5C2CA8BFC19ULL, "vxw_blightlands_uy_0_0.vlm"},
	{0x80C410945681E22DULL, "vxlt_hangedman_19d_0_0.vlm"},
	{0x80C4897297542890ULL, "283884023.ogg"},
	{0x80C561FED7C48360ULL, "llac_lt_0_lowtown_docks_a3_65534_2.dds"},
	{0x80C56785025FF037ULL, "weapon1_d.psh.fxo"},
	{0x80C58A665D33A49FULL, "llac_lt_0_lowtown_-131071_2.dds"},
	{0x80C5BC265DA41D6FULL, "narrative_flemeth6002784.cl"},
	{0x80C5DA34BC6A9839ULL, "28700581.ogg"},
	{0x80C6337F96431174ULL, "545623188.ogg"},
	{0x80C666E8442629A5ULL, "vxlt_hangedman_n_25h_0_0.vlm"},
	{0x80C66E3BFEF91486ULL, "vxca_caverns1_mines_flx_0_0.vlm"},
	{0x80C6CF59CECDEC2DULL, "vxht_mansion_n_50b_0_0.vlm"},
	{0x80C6FA9273282CCBULL, "gen00fl_isabela6079092.cl"},
	{0x80C72729E68450B8ULL, "one140cr_charade6139396.cl"},
	{0x80C78ED318B6F758ULL, "hlac_w_encounter2_camp_327684_0x.dds"},
	{0x80C9011D6AF20404ULL, "mcr370_cut_epilogue6086871.cl"},
	{0x80C98528705B084BULL, "dae300_bran6143679.cl"},
	{0x80C9A417EF1E01A8ULL, "art\\characters\\playercharacter\\elfmale\\glv_lgtp.phy"},
	{0x80CA32EBF79BE3F6ULL, "88110176.ogg"},
	{0x80CA668D092EFA18ULL, "qcr000_arishok6059291.cl"},
	{0x80CA6721B78FCF96ULL, "277890322.ogg"},
	{0x80CA6902B619AF35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkenda.phy"},
	{0x80CABAA3F8145203ULL, "bkl_dot.mao"},
	{0x80CBB388C87EF93AULL, "hlac_lt_backalley_n_65537_7z.dds"},
	{0x80CBD81A546A3D3CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16a.phy"},
	{0x80CC087F934468BEULL, "llac_w_woundedcoast_-262143_0.dds"},
	{0x80CD725068100146ULL, "vxdr_primevaltomb_50w_0_0.vlm"},
	{0x80CD95024B9D20A3ULL, "470204695.ogg"},
	{0x80CDAFDD64995AA5ULL, "465129853.ogg"},
	{0x80CDE0AA6668D6E3ULL, "art\\characters\\weapons\\staff\\w_stf_018a.msh"},
	{0x80CDEEAA6668EE3DULL, "art\\characters\\weapons\\staff\\w_stf_018a.mmh"},
	{0x80CDF2AA6668F54EULL, "art\\characters\\weapons\\staff\\w_stf_018a.mao"},
	{0x80CE3270349559F5ULL, "6076123_m.fxe"},
	{0x80CE687A860BEB78ULL, "135468487.ogg"},
	{0x80CE9BB97AB67705ULL, "lt_shop_3567.anb"},
	{0x80CEECCD57425F2FULL, "6101032_m.fxe"},
	{0x80CF1660A486BCF5ULL, "plt_gen00pt_attributes.plo"},
	{0x80CF3EF2DBC66979ULL, "w_bonepitmining_a2_214_0.msh"},
	{0x80CF48F2DBC67A67ULL, "w_bonepitmining_a2_214_0.mmh"},
	{0x80CF4CF2DBC680B4ULL, "w_bonepitmining_a2_214_0.mao"},
	{0x80CF548F377F9189ULL, "lt_0_lowtown_docks_163_0.mao"},
	{0x80CF608F377FA622ULL, "lt_0_lowtown_docks_163_0.mmh"},
	{0x80CF628F377FA9D8ULL, "lt_0_lowtown_docks_163_0.msh"},
	{0x80CF814FEE85A951ULL, "336524673.ogg"},
	{0x80CFA23F3069BDA4ULL, "671080770.ogg"},
	{0x80CFE6592E9826A8ULL, "vxlt_0_lowtown_docks_a3_n_d2v_0_0.vlm"},
	{0x80CFE9132144CAFCULL, "mag100_karras6034932.cl"},
	{0x80D03D4BAB7857DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_rf.evt"},
	{0x80D04B908ED669D9ULL, "art\\vfx\\models\\vfx_cin_rock_01.msh"},
	{0x80D04DA4FC836BC7ULL, "vxw_woundedcoast_1i9_0_1.vlm"},
	{0x80D055908ED67AC7ULL, "art\\vfx\\models\\vfx_cin_rock_01.mmh"},
	{0x80D0E23252DC9199ULL, "hlac_w_encounter2_131074_0y.dds"},
	{0x80D1E88A0F865379ULL, "mrl150_merrill6085717.cl"},
	{0x80D3037735C33A31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v1.ani"},
	{0x80D36466A015CE08ULL, "vxca_caverns1_egq_0_0.vlm"},
	{0x80D430C79C125B8CULL, "vxlt_0_lowtown_a3_n_c_ghv_0_0.vlm"},
	{0x80D484BD767814A0ULL, "6152365_m.fxe"},
	{0x80D4F814B8FF46E4ULL, "combat_smcs_range.gda"},
	{0x80D542E002AACE2FULL, "6136095_f.fxe"},
	{0x80D5795D1BA2FBAEULL, "6202341_m.fxe"},
	{0x80D5AC4E19D8C35AULL, "730629919.ogg"},
	{0x80D5BDD83424F4F5ULL, "6133321_m.fxe"},
	{0x80D65B5ADF24AFB6ULL, "ser221_party_comments6166198.cl"},
	{0x80D664211BF1E91DULL, "19004302.ogg"},
	{0x80D6CB7F960D83FEULL, "75789213.ogg"},
	{0x80D7364BAB7E6A1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_rf.gad"},
	{0x80D736BB748CFF92ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmle_1.mao"},
	{0x80D75B8FF6203BB3ULL, "gen00fl_fenris6106637.cl"},
	{0x80D780158A0B27F5ULL, "vxlt_0_lowtown_docks_a3_n_c3w_0_0.vlm"},
	{0x80D7A094D039B203ULL, "6169791_m.fxe"},
	{0x80D7ADFE8A4A7EE0ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_2z.dds"},
	{0x80D864A2AFE922B6ULL, "vxht_chantry_28m_0_0.vlm"},
	{0x80D8D0DC16727B54ULL, "mag170_thrask6016000.cl"},
	{0x80D8D2C61657891BULL, "ser270_templar6149400.cl"},
	{0x80D944C5CD9F6C7FULL, "llac_lt_0_lowtown_a2_n_c_1_2.dds"},
	{0x80D954E7418A4E7EULL, "vxlt_hangedman_1mo_0_0.vlm"},
	{0x80D9576FF0035F67ULL, "754667507.ogg"},
	{0x80D95B1C8935FA0DULL, "192670084.ogg"},
	{0x80D966762C4FC724ULL, "6018892_m.fxe"},
	{0x80D969B8F7D1D5CEULL, "6217596_m.fxe"},
	{0x80D9A0A566594B26ULL, "6147941_m.fxe"},
	{0x80DA0E7735C96BD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v1.gad"},
	{0x80DA339A4ED8FED8ULL, "264841212.ogg"},
	{0x80DA35598D835DB6ULL, "6077893_m.fxe"},
	{0x80DA96BCCDD1FFE4ULL, "rdr111dg_martin6014881.cl"},
	{0x80DA993D96B27314ULL, "6092505_f.fxe"},
	{0x80DAEA040EF1EB61ULL, "184012281.ogg"},
	{0x80DB72053A910D5CULL, "628958701.ogg"},
	{0x80DC1E8870156120ULL, "6033150_m.fxe"},
	{0x80DC5CE494142BE0ULL, "987254632.ogg"},
	{0x80DC7C300C24933BULL, "vxht_playershouse_n_1o4_0_0.vlm"},
	{0x80DD87D94E64002EULL, "vfx_vfx_p7.psh.fxo.dx11"},
	{0x80DDB7AE0F513F73ULL, "vxlt_hangedman_rr_0_0.vlm"},
	{0x80DDEBA900CCF83FULL, "6101398_f.fxe"},
	{0x80DE354BAB84871AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_rf.ani"},
	{0x80DE415FC949BF36ULL, "llac_ht_bartrand_n_3373_1.dds"},
	{0x80DE4AEE0CD0400BULL, "mag160_varian_ilithis6018911.cl"},
	{0x80DE612037EC0206ULL, "6120462_f.fxe"},
	{0x80DE9B5438DCA529ULL, "dae000_bartrand6204156.cl"},
	{0x80DEDD80102A511CULL, "gen_im_trs_rdr_junk1_02.uti"},
	{0x80DFD85687AA5248ULL, "hlac_ht_keep_n_4519_0y.dds"},
	{0x80E0D0A36A8BAEEFULL, "gen00fl_aveline6139726.cl"},
	{0x80E1C4F34240D63FULL, "6201587_m.fxe"},
	{0x80E1D22E5F0C1F16ULL, "hlac_w_woundedcoast_262145_0y.dds"},
	{0x80E23A014A1ADE45ULL, "ico_mm_exit.dds"},
	{0x80E2CDEBECAE0EC1ULL, "dae300_alistair_king6133595.cl"},
	{0x80E30221D009912EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_basic.phy"},
	{0x80E316564CAEEF56ULL, "llac_lt_gallowscourtyard_-65538_0.dds"},
	{0x80E380F72B72D33FULL, "llac_w_bonepitmining_a2_458754_0.dds"},
	{0x80E3F8A52E69B0E7ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_chestb_d.dds"},
	{0x80E40EB8ABDC64ECULL, "989731166.ogg"},
	{0x80E46137F8539D73ULL, "241495068.ogg"},
	{0x80E470B48F348D4FULL, "art\\characters\\creatures\\dragonling\\textures\\c_dragling_0t.dds"},
	{0x80E4CC74A9463C2FULL, "zz_dae_debug6049982.cl"},
	{0x80E4D3B9FD5EE814ULL, "vxca_caverns2_961_0_0.vlm"},
	{0x80E571540B5BDC53ULL, "vxht_chantry_2c9_0_0.vlm"},
	{0x80E5873E83BD2181ULL, "and111cr_refugee_girl.utc"},
	{0x80E5BDA97684B7B1ULL, "743199879.ogg"},
	{0x80E60E980C763A33ULL, "6042809_m.fxe"},
	{0x80E61152B6E514A7ULL, "vxlt_0_lowtown_docks_c4t_0_0.vlm"},
	{0x80E64994F2E46C7AULL, "vxlt_0_lowtown_docks_a3_cvm_0_1.vlm"},
	{0x80E679F220B3075DULL, "hlac_ht_0_hightown_a3_-3_0y.dds"},
	{0x80E692786EED49A9ULL, "840326646.ogg"},
	{0x80E6B42FF769EEAEULL, "hlac_lt_gallowstemplar_2394_6z.dds"},
	{0x80E70E611CA1B655ULL, "232394817.ogg"},
	{0x80E74BAB1A4C53FBULL, "vxlt_hangedman_13t_0_0.vlm"},
	{0x80E772A0BA063766ULL, "6113984_m.fxe"},
	{0x80E78FC49B4D2154ULL, "86572.ogg"},
	{0x80E854997C586E49ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_bird02.mao"},
	{0x80E9061C347E87DBULL, "vxht_mansion_n_221_0_0.vlm"},
	{0x80E93F634145E26DULL, "mag170_thrask6167404.cl"},
	{0x80E976536851BCA4ULL, "art\\vfx\\combat\\vfx_str_scream_act_c.impa.ani"},
	{0x80E9A2F738DA8A21ULL, "mag101_grace6053063.cl"},
	{0x80E9AC266E2EF1D8ULL, "vxlt_0_lowtown_a2_n_c_d6f_0_0.vlm"},
	{0x80EA1A57E7618B69ULL, "seb221_ruxton_harimann6187815.cl"},
	{0x80EA8C57102E4CB9ULL, "art\\vfx\\creatures\\fxc_bm_vomr_c.anb"},
	{0x80EB702E7A2AFE58ULL, "19281665.ogg"},
	{0x80EBE13063091EAFULL, "214914358.ogg"},
	{0x80ECE0F6738F444DULL, "77714096.ogg"},
	{0x80ED2F27ECE5E478ULL, "vxht_keep_580_0_0.vlm"},
	{0x80ED75676F8CC91FULL, "6125176_f.fxe"},
	{0x80EDEBFD51F7A8C3ULL, "1020598055.ogg"},
	{0x80EE362DF68401B2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairright.phy"},
	{0x80EE62E4EDD9A332ULL, "llac_lt_0_lowtown_65534_4.dds"},
	{0x80EF735368562306ULL, "art\\vfx\\combat\\vfx_str_scream_act_c.impa.gad"},
	{0x80EFE6B237BA820EULL, "atl_load_contr_dxt5_00.dds"},
	{0x80F01DD0ABB631CAULL, "mag100cr_templar.utc"},
	{0x80F03247E1CAFDD4ULL, "101531090.ogg"},
	{0x80F0D9CA23DBBF4FULL, "target_creaturelock.cur"},
	{0x80F1BD246BC587C6ULL, "vxht_0_hightown_n_4c9_0_0.vlm"},
	{0x80F1C7CAE18AE4C1ULL, "vxw_bonepit_6w_0_0.vlm"},
	{0x80F1FF4DE478D72DULL, "lt_hangedman_mmap_ph.dds"},
	{0x80F205A783E15C6CULL, "t1_stb_red.tnt"},
	{0x80F2401FA53CA1A9ULL, "vxlt_0_lowtown_docks_a3_apw_0_0.vlm"},
	{0x80F2581A6BAB7815ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand_2.mao"},
	{0x80F2C9440A0EE17BULL, "w_sundermount_136_0.msh"},
	{0x80F2DB440A0EFFC6ULL, "w_sundermount_136_0.mao"},
	{0x80F2E7440A0F1465ULL, "w_sundermount_136_0.mmh"},
	{0x80F2EEE4188447D4ULL, "mag321_comtesse_de_launcet6064888.cl"},
	{0x80F2F81E4AC504F8ULL, "203581030.ogg"},
	{0x80F4271E0B7E4C50ULL, "vfx_chunkyparticle_p1.psh.fxo.dx11.assembly"},
	{0x80F46FE34DB49EDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_fall_look_back.ani"},
	{0x80F4832FA8DB2D27ULL, "vxlt_0_lowtown_docks_c1w_0_0.vlm"},
	{0x80F4BB71E4DA84FAULL, "vxlt_0_lowtown_docks_a3_cyl_0_1.vlm"},
	{0x80F4CDCDD1DC887AULL, "follower_banter6151323.cl"},
	{0x80F5180A1FE0EB38ULL, "w_bonepitdestroyed_a3_18_0_ll.mmh"},
	{0x80F529A0B18F3437ULL, "vxlt_gallowstemplar_n_53j_0_0.vlm"},
	{0x80F56719655A5C16ULL, "zz_vault_debug6200348.cl"},
	{0x80F58B6D6A51D0F6ULL, "lt_0_lowtown_docks_a3_135_0.msh"},
	{0x80F5956D6A51E1F4ULL, "lt_0_lowtown_docks_a3_135_0.mmh"},
	{0x80F5996D6A51E807ULL, "lt_0_lowtown_docks_a3_135_0.mao"},
	{0x80F5E572BD37860FULL, "vxlt_gallowstemplar_n_5rq_0_0.vlm"},
	{0x80F61E1BA2DB82A0ULL, "vxw_sundermount_2zp_0_1.vlm"},
	{0x80F651B3ECCB19F8ULL, "plt_cod_bks_lite_vials.nss"},
	{0x80F66E53685C39E5ULL, "art\\vfx\\combat\\vfx_str_scream_act_c.impa.evt"},
	{0x80F6917D891926A8ULL, "902042100.ogg"},
	{0x80F7048292B18183ULL, "701520376.ogg"},
	{0x80F70CF91D60B153ULL, "hlac_ht_0_hightown_a3_1_2z.dds"},
	{0x80F72B2F0A74674FULL, "llac_ht_chantry_n_2534_21.dds"},
	{0x80F771100B332A57ULL, "vxht_bartrand_n_5sr_0_0.vlm"},
	{0x80F7CED758E00EE0ULL, "6085933_m.fxe"},
	{0x80F81F1E493E2EEBULL, "gen_im_cft_reg_a.uti"},
	{0x80F869DDDD9114E0ULL, "zz_vrc_debug6200181.cl"},
	{0x80F87C4D848E767BULL, "vxht_mansion_n_53i_0_0.vlm"},
	{0x80F890DF8805195EULL, "9671896.ogg"},
	{0x80F8B9A9BD97EE4CULL, "6117450_m.fxe"},
	{0x80F8E8AD1BBB15DDULL, "llac_lt_0_lowtown_docks_-196609_1.dds"},
	{0x80F9ADDF732A1349ULL, "vxht_bartrand_n_68f_0_0.vlm"},
	{0x80F9F471F3163CEEULL, "vxw_bonepitmining_a2_1pc_0_0.vlm"},
	{0x80F9FEF6CDE608AEULL, "art\\characters\\heraldry\\hld_chantry.dds"},
	{0x80FA3EBB43E03C44ULL, "vxw_blightlands_hs_0_0.vlm"},
	{0x80FA7BFF3991BCD6ULL, "904766921.ogg"},
	{0x80FA95D34DA1471BULL, "6112079_f.fxe"},
	{0x80FC78CF091DAD35ULL, "zz_followers_banter6158370.cl"},
	{0x80FCEF1DAF6EE678ULL, "vxht_0_hightown_a3_aur_0_0.vlm"},
	{0x80FD35E204633872ULL, "gen_im_trs_bst_drg_scale_02.uti"},
	{0x80FD8E7712F07B6BULL, "hla_ht_0_hightown_a3_65535_0x.dds"},
	{0x80FDA80D86596B59ULL, "6175420_m.fxe"},
	{0x80FDE6D2E9034064ULL, "837792417.ogg"},
	{0x80FE0019AFD54F46ULL, "vxw_bonepit_yu_0_0.vlm"},
	{0x80FE41D2CD6B952EULL, "vxlt_gallowstemplar_n_6t9_0_0.vlm"},
	{0x80FE4B9F14C76F49ULL, "g_p_rock_sm.mao"},
	{0x80FE9739053C2FEFULL, "vxht_mansion2_235_0_0.vlm"},
	{0x80FEB9FD75F81788ULL, "vxca_caverns1_ehl_0_0.vlm"},
	{0x80FFBCBE2A56CF62ULL, "dae000_mother6103046.cl"},
	{0x81005175547522B9ULL, "hlac_ca_cavern1_haunted_24426_13y.dds"},
	{0x8100C6F2BCF799F1ULL, "w_sundermount_233_0.msh"},
	{0x8100D4F2BCF7B10CULL, "w_sundermount_233_0.mao"},
	{0x8100E0F2BCF7C5AFULL, "w_sundermount_233_0.mmh"},
	{0x8101269C1041DE03ULL, "298278887.ogg"},
	{0x81014BA89E0F7472ULL, "art\\characters\\creatures\\mubari\\animation\\kd_enter.gad"},
	{0x8101997DB8ED834EULL, "vxht_0_hightown_a3_cbo_0_0.vlm"},
	{0x8101A2F802869E06ULL, "vxlt_hangedman_1jr_0_0.vlm"},
	{0x8102600CDE5EA7DAULL, "vxlt_0_lowtown_a2_n_bym_0_0.vlm"},
	{0x810277E34DC0E9D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_fall_look_back.evt"},
	{0x8102B94EDE92A1C2ULL, "combat_lc_r1_a1.gda"},
	{0x81039B7DA3854CB7ULL, "vxlt_gallowstemplar_n_56e_0_0.vlm"},
	{0x81040851B8F1D8B0ULL, "481702965.ogg"},
	{0x8104574FAF2D9E8FULL, "vxlt_gallowstemplar_n_5up_0_0.vlm"},
	{0x81052381DCC8A6C3ULL, "vrc221cr_bartrand_fake.utc"},
	{0x81057063BDA37E79ULL, "zz_isa_debug6133972.cl"},
	{0x810597D8FEB702D3ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.anb"},
	{0x8105CC30D8A3D941ULL, "hlac_ca_caverns1_24425_5x.dds"},
	{0x8105F73FB23B2C30ULL, "6137193_m.fxe"},
	{0x810673DD618C2D7CULL, "6146889_m.fxe"},
	{0x81067810E7B45AEEULL, "vxlt_0_lowtown_docks_a3_c42_0_0.vlm"},
	{0x81067AB5FE7095F1ULL, "hlac_ht_0_hightown_a3_-65535_0x.dds"},
	{0x8106BDD03E1EAE31ULL, "dae000cr_mage_unhappy.utc"},
	{0x8106CDF32AF40EE9ULL, "dae010_lowtown_guard.utc"},
	{0x8106FA50EEE14D81ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_0y.dds"},
	{0x8107485EDBFE8A3BULL, "6181810_m.fxe"},
	{0x81079CB97AE63E0DULL, "lt_shop_3567.rml"},
	{0x8107B8DB213123BFULL, "424056330.ogg"},
	{0x8107C172D4A8EB01ULL, "dae021_keep_noble.utc"},
	{0x8108028DA22DD248ULL, "str_cra_pro.mao"},
	{0x81084EA89E1597A9ULL, "art\\characters\\creatures\\mubari\\animation\\kd_enter.evt"},
	{0x8108B09835D654C4ULL, "vxw_blightlands_mp_0_0.vlm"},
	{0x8108B83F17F932BCULL, "one300_nuncio6077192.cl"},
	{0x81091549478687AAULL, "vxlt_gallowstemplar_n_6s4_0_0.vlm"},
	{0x810915EC78D44FF6ULL, "924874626.ogg"},
	{0x81094B0E5D77430AULL, "vxw_bonepitmining_a2_1ls_0_0.vlm"},
	{0x810996E34DC73D65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_fall_look_back.gad"},
	{0x810A2A601DCFAA97ULL, "6107345_f.fxe"},
	{0x810A63040523B940ULL, "579757066.ogg"},
	{0x810A717EDF78C960ULL, "957668944.ogg"},
	{0x810AE5FA04C34300ULL, "310325772.ogg"},
	{0x810B3563C17D65F6ULL, "461000618.ogg"},
	{0x810B5419EA427393ULL, "hlac_ht_chantry_n_2534_5x.dds"},
	{0x810B714ABDA32886ULL, "dae000ip_crowd_mf6_talk15.utp"},
	{0x810B78F5E13ADD04ULL, "ser270_templar6112544.cl"},
	{0x810CE4322DE1648EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual"},
	{0x810CE8B88E0F9A0FULL, "vfx_rendertoshadowmap_p1.psh.fxo.dx11"},
	{0x810D448F846133C7ULL, "vxw_woundedcoast_r8_0_0.vlm"},
	{0x810D8432A0EBA4E0ULL, "6098517_m.fxe"},
	{0x810D927BC5DD4BD1ULL, "924405562.ogg"},
	{0x810DEC9E551B92D2ULL, "mag101_grace6034076.cl"},
	{0x810E5A6E439207BEULL, "75049803.ogg"},
	{0x810E88C7181A9555ULL, "dre100cr_warden_f.utc"},
	{0x810EFCDF09F4D7D3ULL, "331764551.ogg"},
	{0x810F204EF9A364F0ULL, "dae000_dalish_apprentices6147537.cl"},
	{0x810F6391FFC224C1ULL, "dae000_gamlen6144340.cl"},
	{0x810F923B71F5B2D5ULL, "gen_im_wep_war_shd_act3_04"},
	{0x810FBEAAEC82E896ULL, "vrc321_varric6087264.cl"},
	{0x81100FBF48229583ULL, "hlac_lt_0_lowtown_docks_a3_-3_2x.dds"},
	{0x811030BE571F2A2CULL, "6095547_m.fxe"},
	{0x811057F2BF613452ULL, "6098898_m.fxe"},
	{0x811082F244202B4EULL, "hlac_lt_0_lowtown_a2_n_c_-65536_6z.dds"},
	{0x81109EC005888F05ULL, "1040968618.ogg"},
	{0x8110FF534B9FA6EFULL, "gen00fl_dog6114941.cl"},
	{0x81127623DFA0F48FULL, "460190511.ogg"},
	{0x81127C02B6578550ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkenda.msh"},
	{0x81128A02B6579D6AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkenda.mmh"},
	{0x81128E02B657A3A1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ff_sidewlkenda.mao"},
	{0x8112A7847F2349AEULL, "vxht_mansion2b_9d_0_0.vlm"},
	{0x8113A2F5FAA0CCB0ULL, "hlac_w_bonepit_0_3y.dds"},
	{0x8113B4D8FEC3713CULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.mmh"},
	{0x811456AB73AD8EFBULL, "6114489_m.fxe"},
	{0x8114805723E863CDULL, "llac_dr_dungeon_3217_11.dds"},
	{0x8114C54006D73D23ULL, "312297461.ogg"},
	{0x8114E5BA538245FCULL, "6143789_m.fxe"},
	{0x81155E248C30B238ULL, "zz_t_trailer_merril.cut"},
	{0x81156007687AA77BULL, "vxht_mansion_n_59k_0_0.vlm"},
	{0x811568484BF15AF8ULL, "leap_link_so.mao"},
	{0x811575A63FBE1E53ULL, "hlac_w_bonepitmining_a2_458753_0y.dds"},
	{0x811636A89E21AC18ULL, "art\\characters\\creatures\\mubari\\animation\\kd_enter.ani"},
	{0x81166C4BB8D410CBULL, "6116656_m.fxe"},
	{0x8116B07BFF5F7886ULL, "vxw_encounter2_camp_7g_0_0.vlm"},
	{0x8116B18AB06E55B1ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredidies.ani"},
	{0x8116C8EAE8559D3CULL, "zz_vrc_debug6064260.cl"},
	{0x8117227527CC6D44ULL, "vxw_blightlands_vq_0_0.vlm"},
	{0x81172A1BE53F84E1ULL, "lt_0_lowtown_a2_n_c_46_0_ll.mmh"},
	{0x8117B8BF5F5CEAC2ULL, "zz_vault_debug6203222.cl"},
	{0x8117C7104CC0B172ULL, "vxht_mansion2_1u9_0_0.vlm"},
	{0x8117F536CFD1823DULL, "121801736.ogg"},
	{0x81183EBE4029B4BBULL, "vxlt_warehouse_1dq_0_0.vlm"},
	{0x8118B5FE44D2C642ULL, "566949616.ogg"},
	{0x8119004D32729A7BULL, "hlac_lt_warehouse_n_0_7z.dds"},
	{0x81190A442569C33EULL, "6027337_m.fxe"},
	{0x81196FF2CA667E17ULL, "6069284_m.fxe"},
	{0x811989203ED70B7FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookstand_sa.dds"},
	{0x811A0C86F3E810CDULL, "6208354_f.fxe"},
	{0x811A7FEE3E71E070ULL, "character_character_p140.psh.fxo"},
	{0x811A922A653C5FF6ULL, "prop1_ds1.psh.fxo"},
	{0x811AE3D393C18046ULL, "vxw_bonepit_ow_0_0.vlm"},
	{0x811B02E1F2A9E67AULL, "follower_banter6143207.cl"},
	{0x811B1E5D83395274ULL, "6141758_m.fxe"},
	{0x811B446ED496FF28ULL, "6132882_m.fxe"},
	{0x811BD2975283188FULL, "hlac_lt_0_lowtown_a2_n_c_-196609_2y.dds"},
	{0x811C195F4FACFBF3ULL, "qcr220_grey_wardens6060475.cl"},
	{0x811C3884E0B9281FULL, "dae000ip_crowd_mf5_drunk2.utp"},
	{0x811C730601B2F271ULL, "zz_vault_debug6203177.cl"},
	{0x811CC6DEDA52F062ULL, "cd_qun221_petrice.bnk"},
	{0x811CCC4C886CE4A8ULL, "165301329.ogg"},
	{0x811D05F4744D94E2ULL, "vxdr_primevaltomb_2us_0_0.vlm"},
	{0x811D141C91BA7DFBULL, "hlac_ca_caverns2prc_5802_45z.dds"},
	{0x811D317569D9C767ULL, "character_character_p64.psh.fxo.dx11.assembly"},
	{0x811D50EA977ECD7EULL, "gen00fl_aveline6115451.cl"},
	{0x811DBC8AB0748753ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredidies.gad"},
	{0x811DBDDE98B52AF1ULL, "dre000_dougal6128893.cl"},
	{0x811E093EAAEA7FA8ULL, "mag311_keran6041685.cl"},
	{0x811E3078426A385DULL, "936905937.ogg"},
	{0x811E556D9D29D57BULL, "966012306.ogg"},
	{0x811E5CDDC20C29BDULL, "zz_test_suhwan.utc"},
	{0x811E74C2ED7FF9BDULL, "vxlt_darktown_4w6_0_0.vlm"},
	{0x811ED7CAC103A4F7ULL, "6134310_f.fxe"},
	{0x811F781249AAE464ULL, "886294603.ogg"},
	{0x81203A13D5A33221ULL, "db_dust.mao"},
	{0x8120BC7E0CF35D7FULL, "vxlt_gallowstemplar_n_5t7_0_0.vlm"},
	{0x81211BA456B3DC02ULL, "isa211_isabela6069439.cl"},
	{0x81213B099319CF8FULL, "vxlt_gallowstemplar_n_5cr_0_0.vlm"},
	{0x81214C463A64936FULL, "6085472_f.fxe"},
	{0x812157690D29372DULL, "vxw_encounter2_camp_qj_0_0.vlm"},
	{0x81217212CC63D7C4ULL, "art\\vfx\\combat\\vfx_m_arc_element_act_c.impa.ani"},
	{0x8121F3D12E9CDE68ULL, "hlac_ht_chantry_n_2534_34y.dds"},
	{0x81220D5AB6EB57C1ULL, "hlac_lt_0_lowtown_docks_n_65532_1z.dds"},
	{0x81222011F2296746ULL, "vxlt_0_lowtown_n_f1q_0_0.vlm"},
	{0x8122C88865A3A77BULL, "6113589_m.fxe"},
	{0x8122EF832165CC30ULL, "vxlt_0_lowtown_docks_a3_n_9r7_0_0.vlm"},
	{0x81230121D03F73F1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_basic.mmh"},
	{0x81230B2C0351FAFEULL, "hlac_lt_0_lowtown_-131071_1x.dds"},
	{0x81230D21D03F8792ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_basic.mao"},
	{0x81231B21D03F9FAFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_basic.msh"},
	{0x8123B25B18446F5BULL, "607054735.ogg"},
	{0x812424C8A6055596ULL, "6041851_f.fxe"},
	{0x8124787B4974D462ULL, "lt_0_lowtown_a2_n_c_29_0_ll.mmh"},
	{0x81248127D27498C7ULL, "6062061_m.fxe"},
	{0x81252258F1559106ULL, "vxw_encounter2_camp_2h_0_0.vlm"},
	{0x81255C4F313426C0ULL, "6135061_m.fxe"},
	{0x81259B0580F41A4CULL, "6113677_f.fxe"},
	{0x8125A0BEEDAE1888ULL, "vxw_bonepit_16m_0_0.vlm"},
	{0x8125B8DCA59271A2ULL, "gen_im_trs_gen_one_axe_r2.uti"},
	{0x8126347E2AFC2AE3ULL, "one300_zevran6070905.cl"},
	{0x81272CC6220BA4DCULL, "373910793.ogg"},
	{0x812746E0C761596BULL, "6166289_f.fxe"},
	{0x8127617EAE87B588ULL, "dae000_party_night_comment6201601.cl"},
	{0x81277446155B3BEEULL, "hlac_dr_dungeon_3216_10y.dds"},
	{0x81279B9491A35CFBULL, "1023018015.ogg"},
	{0x8127A3E978ECF922ULL, "226014386.ogg"},
	{0x8127C52D37090DDAULL, "crt_hurlock.bnk"},
	{0x8127E5FF8C42ED2CULL, "vxht_brothel_n_610_0_0.vlm"},
	{0x81281B0870234AB6ULL, "character_character_p146.psh.fxo.dx11"},
	{0x81281E3CA74AFE10ULL, "hlac_w_woundedcoast_-262143_0x.dds"},
	{0x8128355A8A70E5D5ULL, "224963827.ogg"},
	{0x81286F12CC69F126ULL, "art\\vfx\\combat\\vfx_m_arc_element_act_c.impa.gad"},
	{0x8128A57E61D1A3BFULL, "cd_sib100_mother.bnk"},
	{0x812985E8A248C84CULL, "hlac_lt_0_lowtown_docks_n_131071_0z.dds"},
	{0x8129913C7E0E199CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_reaction.ani"},
	{0x8129CB1E8DA4F05AULL, "6142205_m.fxe"},
	{0x8129D5CE97A9C5EFULL, "mer210_seldrek6078857.cl"},
	{0x812AA34DB6C6B8A8ULL, "lla_ht_0_hightown_a2_-131071_0.dds"},
	{0x812ADBF547D6EE8CULL, "vxlt_0_lowtown_a3_n_c_gvh_0_0.vlm"},
	{0x812AE9EFFD308650ULL, "vxca_caverns1_mines_i2c_0_0.vlm"},
	{0x812B11EEC5B967AAULL, "421377990.ogg"},
	{0x812C33FA61CE95B3ULL, "6208409_m.fxe"},
	{0x812C71635AB0FC0FULL, "22627754.ogg"},
	{0x812C874333E9658DULL, "seb321_leliana6180906.cl"},
	{0x812C89E6F1C08E99ULL, "gen00fl_anders6125397.cl"},
	{0x812C928C3ACA1AD9ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v1.ani"},
	{0x812D1C98173CC9B2ULL, "follower_banter6163649.cl"},
	{0x812D43429DAC1068ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medh.msh"},
	{0x812D61429DAC4292ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medh.mmh"},
	{0x812D66F6A92B0552ULL, "6140575_m.fxe"},
	{0x812E59C4B9BAD218ULL, "fx_c_cloud.mao"},
	{0x812E5F63E1E865ADULL, "857299203.ogg"},
	{0x812E786420C9A1B3ULL, "6013304_m.fxe"},
	{0x812EF8533577B9E8ULL, "6152592_m.fxe"},
	{0x812F2E5AFEE975FFULL, "vxlt_gallowstemplar_n_5o6_0_0.vlm"},
	{0x812F32C1FBA4D162ULL, "vxht_playershouse_n_1zr_0_0.vlm"},
	{0x812F4E4415D1DF31ULL, "6136339_m.fxe"},
	{0x812F53D588C003FEULL, "807585816.ogg"},
	{0x812F6A12CC700705ULL, "art\\vfx\\combat\\vfx_m_arc_element_act_c.impa.evt"},
	{0x812FAA4AF7F9D783ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_fire_e.dds"},
	{0x812FB5A520578108ULL, "246648756.ogg"},
	{0x812FC945FF1F4FADULL, "vxw_encounter2_camp_li_0_0.vlm"},
	{0x812FD12879B0F390ULL, "hlac_ht_chantry_2534_32z.dds"},
	{0x813020BE92F86120ULL, "6000804_m.fxe"},
	{0x8130267E71EB7315ULL, "6071496_m.fxe"},
	{0x8130273F72713F84ULL, "6062178_m.fxe"},
	{0x813091EEE41F7FC6ULL, "vxlt_0_lowtown_n_f4r_0_0.vlm"},
	{0x81309E3C7E144EDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_reaction.gad"},
	{0x8131031B09C5E8E0ULL, "653932022.ogg"},
	{0x813158CFC813D351ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_c.impa.ani"},
	{0x8131CCA969D13E55ULL, "vxlt_gallowscourtyard_n_7dm_0_0.vlm"},
	{0x8131F82A044E7665ULL, "hlac_lt_0_lowtown_131070_4z.dds"},
	{0x81320300C64C2512ULL, "6111447_m.fxe"},
	{0x81326D0CA9D3E5D8ULL, "dae000ip_crowd_lm1_drunk1.utp"},
	{0x8132B62E467012DEULL, "cod_hst_merchantguild.utp"},
	{0x8132BB0DF8297D45ULL, "follower_banter6171526.cl"},
	{0x81336D8C3ACFFA1BULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v1.gad"},
	{0x81339F3CBE949389ULL, "and231ip_tomb.utp"},
	{0x8133BD13045C6A53ULL, "hlac_lt_hangedman_-1_8z.dds"},
	{0x8133E7440A46AB6EULL, "w_sundermount_136_0.rml"},
	{0x813472EE5FD2FECAULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_nrma_0d.dds"},
	{0x81354703970C88DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_point_up_simple.evt"},
	{0x8135E913942C5E1CULL, "235170859.ogg"},
	{0x8135FF4D966780D0ULL, "vxht_mansion2b_3bp_0_0.vlm"},
	{0x81364556A73198FBULL, "6214790_m.fxe"},
	{0x81368541AA755C6AULL, "vxlt_darktown_2rp_0_0.vlm"},
	{0x8136E2DBBEFE89ECULL, "vxht_keep_n_4n4_0_0.vlm"},
	{0x813763CFC81851F3ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_c.impa.gad"},
	{0x8137793C7E1A2EFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_reaction.evt"},
	{0x8137C993E8662C36ULL, "t3_dey_te1.tnt"},
	{0x8138F377755F9FDFULL, "vxlt_0_lowtown_docks_a3_bvh_0_0.vlm"},
	{0x81397B8A9DCF984CULL, "1031922370.ogg"},
	{0x8139D0BA2E1D37AAULL, "vxht_keep_3ig_0_0.vlm"},
	{0x8139E72316D4A6ABULL, "623050830.ogg"},
	{0x813A88656B428C01ULL, "605454595.ogg"},
	{0x813A8A8C3AD64A00ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v1.evt"},
	{0x813AB2EAC8AF9F05ULL, "dae000_sabina6162324.cl"},
	{0x813AC0440A4C8712ULL, "w_sundermount_136_0.phy"},
	{0x813B27386AD52F45ULL, "one110_meeran6089865.cl"},
	{0x813B4CF806A027F6ULL, "art\\characters\\creatures\\mubari\\animation\\kd_lp.ani"},
	{0x813B500DCCE355CEULL, "clf260_harley6069686.cl"},
	{0x813B587CD16C2ABDULL, "vxlt_darktown_4y4_0_0.vlm"},
	{0x813B7B716E233FB6ULL, "w_bonepit_3_0_ll.mmh"},
	{0x813B9276EFCAE225ULL, "seb201_lady_harimann6187817.cl"},
	{0x813BD0EE94F276E7ULL, "one100dg_lia6055152.cl"},
	{0x813BDC4B5B7A3575ULL, "1005365948.ogg"},
	{0x813C400397129B1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_point_up_simple.gad"},
	{0x813C5E4C30619286ULL, "vxca_caverns1_j4i_0_0.vlm"},
	{0x813DA49EED9AE9E2ULL, "vxht_playershouse_n_1qs_0_0.vlm"},
	{0x813E09EEF81E7B1FULL, "art\\vfx\\combat\\vfx_r_var_activation2_c.impa.gad"},
	{0x813E372B5729EE8DULL, "hlac_w_encounter2_camp_262146_0z.dds"},
	{0x813E3B22F115682DULL, "vxw_encounter2_camp_gh_0_0.vlm"},
	{0x813E60CFC81E6B58ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_c.impa.evt"},
	{0x813E758CDB8507EAULL, "159734461.ogg"},
	{0x813E8D28A155E95BULL, "6121997_m.fxe"},
	{0x813EBEA8316C5BB3ULL, "gen00fl_isabela6090848.cl"},
	{0x813EC35B222CA8D6ULL, "782564680.ogg"},
	{0x813EE1A103359BE9ULL, "hf_mother_h3s1.mor"},
	{0x813F0A428070FA0FULL, "vxdr_0_deeproads_d34_0_0.vlm"},
	{0x813F9F7E20985676ULL, "hlac_lt_0_lowtown_docks_a3_-65538_4y.dds"},
	{0x814003281AED73C3ULL, "vxw_woundedcoast_60_0_0.vlm"},
	{0x8140231BAED34F8BULL, "mer000pt_mine_payments.ncs"},
	{0x814099421A0D5F10ULL, "6051604_m.fxe"},
	{0x8140EAD293DD4B67ULL, "127712953.ogg"},
	{0x814106206C4ED9AEULL, "mag291_demon_marethari6027952.cl"},
	{0x8141B0DEDA71E680ULL, "cd_qun221_petrice.wbd"},
	{0x8141C2083A537694ULL, "6066718_m.fxe"},
	{0x8141F6F37C08BC32ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_threaten.gad"},
	{0x814226462411372CULL, "104297201.ogg"},
	{0x814230C4748FD34DULL, "zz_openwhite.cut"},
	{0x8142BC2515801278ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\1024texel.dds"},
	{0x81433F039718B81AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_point_up_simple.ani"},
	{0x8143B0A71B8A8E60ULL, "24515687.ogg"},
	{0x8143BDD8F3D17CDEULL, "980716944.ogg"},
	{0x814407C1A755AEF6ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_c.anb"},
	{0x81441FD4ED2414DFULL, "vxca_caverns1_gk9_0_1.vlm"},
	{0x81447A72E5768B66ULL, "dae000_gamlen6102714.cl"},
	{0x8144D296EF4FBD3FULL, "6085871_m.fxe"},
	{0x8144F71E9C6B74EAULL, "vxlt_darktown_2uq_0_0.vlm"},
	{0x8144FEEEF824870CULL, "art\\vfx\\combat\\vfx_r_var_activation2_c.impa.evt"},
	{0x81457D3C9EBE0B96ULL, "vxht_mansion_n_6bl_0_0.vlm"},
	{0x814618371CAB35EDULL, "zz_one_debug6079303.cl"},
	{0x8146594CB2950122ULL, "gen00fl_varric6033628.cl"},
	{0x814689B9B1322F6DULL, "zz_spy130cr_darktown_res_3.utc"},
	{0x8146991E410135ECULL, "vxlt_0_lowtown_a3_n_c_eht_0_0.vlm"},
	{0x8146B52D6D219212ULL, "6121651_m.fxe"},
	{0x8146E037B390ED14ULL, "hlac_lt_0_lowtown_docks_n_-131073_1z.dds"},
	{0x8146EA53FF04B3E5ULL, "925686663.ogg"},
	{0x8147B7CF05EE3AAEULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef_facem1.mao"},
	{0x814812A46D77BCFDULL, "zz_dre_debug6069206.cl"},
	{0x814859D7156DDF81ULL, "hlac_w_woundedcoast_-262137_0z.dds"},
	{0x81487FAFFC21EF80ULL, "vxlt_darktown_5bi_0_0.vlm"},
	{0x8148F9F37C0EE069ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_threaten.evt"},
	{0x8149029582F95604ULL, "fen110_fenris6118137.cl"},
	{0x81491DBB2990AA46ULL, "6103897_m.fxe"},
	{0x814924C23088E160ULL, "vxht_0_hightown_n_ccl_0_0.vlm"},
	{0x814957C5AE31BD43ULL, "24869011.ogg"},
	{0x8149B51C37867072ULL, "456187733.ogg"},
	{0x8149D9CF3CFF6484ULL, "hlac_lt_gallowsprison_n_-1_11y.dds"},
	{0x8149F2BAE7C9911DULL, "491728439.ogg"},
	{0x814A5103B31E9E17ULL, "lt_0_lowtown_docks_a3_n.arl"},
	{0x814A58446D2F1E82ULL, "vxht_0_hightown_a3_c8m_0_0.vlm"},
	{0x814AE2BBB92640C5ULL, "vxw_woundedcoast_126_0_1.vlm"},
	{0x814B3916EB732B35ULL, "ach_exp_deep_roads.dds"},
	{0x814B4A818C8E3D58ULL, "hlac_lt_undercityencounter_-1_17z.dds"},
	{0x814BDB8E6DD12097ULL, "vxlt_backalley_n_1et_0_0.vlm"},
	{0x814BFD5BD9125E1CULL, "6045425_m.fxe"},
	{0x814C2FDA523E71F8ULL, "6161385_m.fxe"},
	{0x814C49A03A295799ULL, "follower_banter6173380.cl"},
	{0x814C6C0C6A05CD4BULL, "769073052.ogg"},
	{0x814C7789A0C63050ULL, "mag111im_thrask_letter.uti"},
	{0x814C89A63756CC83ULL, "6044964_m.fxe"},
	{0x814CB24270992B5EULL, "vxlt_hangedman_n_1xk_0_0.vlm"},
	{0x814CCCC2A927BCECULL, "vxca_cavern1_haunted_i38_0_0.vlm"},
	{0x814D4DE9649F59F8ULL, "279564223.ogg"},
	{0x814D7A55D3B2583BULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured.ani"},
	{0x814DDF2D372A0978ULL, "crt_hurlock.wbd"},
	{0x814DE6BD8474C340ULL, "gen00fl_anders6110123.cl"},
	{0x814E385B2EE52670ULL, "mer320_hubert6079338.cl"},
	{0x814E837A4B1802FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_ex.evt"},
	{0x814E87296E4CDE2DULL, "546345475.ogg"},
	{0x814E8A76D3699D99ULL, "6182714_m.fxe"},
	{0x814F1827EAB2E6F2ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_fereldencoin_d.dds"},
	{0x814F2D22A29B30BAULL, "6131575_m.fxe"},
	{0x814F3CBE6EE6677FULL, "vxca_caverns2prc_9dx_0_1.vlm"},
	{0x814F6274F132D8DCULL, "hlac_ht_0_hightown_-196608_2z.dds"},
	{0x814F75F806B11630ULL, "art\\characters\\creatures\\mubari\\animation\\kd_lp.gad"},
	{0x814F7727D86B44C1ULL, "vxlt_0_lowtown_gcv_0_0.vlm"},
	{0x814F8CB1EDF0567EULL, "mag370_meredith6154385.cl"},
	{0x814FC6A4512F85A6ULL, "vxca_cavern1_haunted_hm9_0_0.vlm"},
	{0x81503492305B8FD5ULL, "vxca_caverns1_mines_j3z_0_0.vlm"},
	{0x81504E0F0913A0C4ULL, "gen00fl_fenris6105848.cl"},
	{0x8150946D6A9E8BEFULL, "lt_0_lowtown_docks_a3_135_0.phy"},
	{0x815159C637048143ULL, "sp_star_02.mao"},
	{0x81516D654D416900ULL, "555325551.ogg"},
	{0x8151A76300D4308DULL, "6135622_m.fxe"},
	{0x8151C92B10251B3FULL, "mer120_hubert6017849.cl"},
	{0x8152046C65A564B1ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_kida_0n.dds"},
	{0x81520AC1A761F161ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_c.mmh"},
	{0x815236BD01B50F82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_up_big.gad"},
	{0x8152E6EEF8309C65ULL, "art\\vfx\\combat\\vfx_r_var_activation2_c.impa.ani"},
	{0x8153045CC2E9C17BULL, "vxw_encounter2_camp_rq_0_0.vlm"},
	{0x815310261D3A687FULL, "hslmatrixfade.psh.fxo"},
	{0x81531C591C93472FULL, "ave310_brennan6136767.cl"},
	{0x815346D414D71F66ULL, "6165035_m.fxe"},
	{0x815350A2726215FDULL, "6077735_m.fxe"},
	{0x81539D0FD1A1AD01ULL, "vfx0_addv_alt.psh.fxo"},
	{0x8153AD402F804053ULL, "qcr220_orsino6060743.cl"},
	{0x8153BA72FFD4E939ULL, "atl_chanters_dxt5_00.dds"},
	{0x8153BF0D47FDC4D4ULL, "dre000_varric6034673.cl"},
	{0x8154214641F0CEF1ULL, "genip_pp_gle_one_shot.utp"},
	{0x8154611F6E80D668ULL, "zz_mer_debug6019364.cl"},
	{0x815472B39D18080DULL, "210604543.ogg"},
	{0x815488E0C471930FULL, "gen00fl_varric6104943.cl"},
	{0x81548B7E61F6B2F5ULL, "cd_sib100_mother.wbd"},
	{0x8154C787B7D2026BULL, "dae011_drunk2_sta6149704.cl"},
	{0x8154FD189E1B6F48ULL, "6065016_m.fxe"},
	{0x8154FF42F1A63A4FULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.dura.gad"},
	{0x815540AE60986E71ULL, "gen00fl_anders6103740.cl"},
	{0x81557682DB5FD02CULL, "vxht_0_hightown_a2_anf_0_0.vlm"},
	{0x81557C7A4B1E153CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_ex.gad"},
	{0x8155BA6EA7D55ECEULL, "67225819.ogg"},
	{0x8155E1F37C1941D8ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_threaten.ani"},
	{0x8155F25CF613DB5EULL, "268514879.ogg"},
	{0x81579F2AD0F4ACAEULL, "zz_ave_debug6064647.cl"},
	{0x815841F80BAE1D76ULL, "6165439_m.fxe"},
	{0x815882C9CB854F99ULL, "576784468.ogg"},
	{0x81588321A58B3052ULL, "294761313.ogg"},
	{0x8158840828F0710AULL, "vxlt_0_lowtown_docks_n_bdz_0_1.vlm"},
	{0x8158CE357500E221ULL, "938431931.ogg"},
	{0x815937635ED1CD3FULL, "vxw_bonepitdestroyed_a3_1q3_0_0.vlm"},
	{0x81595559C31163BCULL, "9516823.ogg"},
	{0x815959BD01BB6A79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_up_big.evt"},
	{0x8159DD11A9642346ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheada_n.dds"},
	{0x815A5026AA911696ULL, "fex_bd_a_bs_icn.msh"},
	{0x815A5A26AA912794ULL, "fex_bd_a_bs_icn.mmh"},
	{0x815A8255D3BCF032ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured.evt"},
	{0x815AC5A99CFD533AULL, "signatureeditionitems.dds"},
	{0x815B1A048B252190ULL, "dre000_vampire6167460.cl"},
	{0x815B241F628F43DEULL, "vxlt_hangedman_n_1ul_0_0.vlm"},
	{0x815B40E854DD8CDEULL, "248171322.ogg"},
	{0x815B8DD23142FFB3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffc_d.dds"},
	{0x815BAC0EC2F28076ULL, "l_lgt331ar_shw_bloodmage.wbd"},
	{0x815BF7FA2826DC45ULL, "hlac_lt_warehouse_n_65537_2y.dds"},
	{0x815C0206A699387BULL, "vxca_caverns2prc_8bf_0_0.vlm"},
	{0x815C1442F1AC7CFCULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.dura.evt"},
	{0x815C37E050DAF8CFULL, "vxca_caverns1_mines_gf2_0_0.vlm"},
	{0x815C53140E967D5FULL, "172305577.ogg"},
	{0x815C7B7A4B24323AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_ex.ani"},
	{0x815CCFFBBEE83869ULL, "94588645.ogg"},
	{0x815CD0EDEA9592A1ULL, "6154731_m.fxe"},
	{0x815D3D5DC611C6C2ULL, "vxht_mansion2b_2p1_0_0.vlm"},
	{0x815EEDB23E5A2D4AULL, "vxlt_darktown_4lh_0_0.vlm"},
	{0x815F3767A8B6F02BULL, "vxht_0_hightown_a2_brv_0_0.vlm"},
	{0x815F6B52F49E9637ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_avea_n.dds"},
	{0x81602AC0C61257F9ULL, "46246621.ogg"},
	{0x8160A0F45A154DF4ULL, "vxlt_0_lowtown_a2_n_eny_0_0.vlm"},
	{0x8160BCCA178FB20BULL, "zz_and_debug6134083.cl"},
	{0x8160EF874771DED5ULL, "6137574_m.fxe"},
	{0x81616F386E38AB92ULL, "vxlt_0_lowtown_a2_e28_0_0.vlm"},
	{0x81617639B4DFD9FBULL, "vxw_encounter2_camp_mp_0_0.vlm"},
	{0x816189CDACD8081CULL, "476376195.ogg"},
	{0x8161A155D3C343C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured.gad"},
	{0x8161D0AC5A415DEDULL, "mag311_keran6041718.cl"},
	{0x8161F94FDE157E00ULL, "6120557_m.fxe"},
	{0x81624BFCC844C813ULL, "ser210_witness6168166.cl"},
	{0x816260F682AA3C96ULL, "vxht_mansion_n_6hn_0_0.vlm"},
	{0x81627689EC39C8AAULL, "hlac_lt_warehouse_65537_5z.dds"},
	{0x816293113966CD94ULL, "vxlt_0_lowtown_feh_0_0.vlm"},
	{0x8162AE3562327740ULL, "457441339.ogg"},
	{0x8162DC527AF5BFE2ULL, "hlac_w_woundedcoast_-196607_0y.dds"},
	{0x8163609373D4C842ULL, "hlac_ht_keep_n_4519_9x.dds"},
	{0x81636E29325DE99BULL, "vxlt_0_lowtown_docks_n_b7v_0_1.vlm"},
	{0x8163A0EB091C276EULL, "31976190.ogg"},
	{0x8163DDF628A456D3ULL, "674226653.ogg"},
	{0x8163E2722BAFFD10ULL, "351466917.ogg"},
	{0x8163E85FCD55E8ACULL, "vxht_0_hightown_a2_aig_0_0.vlm"},
	{0x81645D5094267CC8ULL, "dae000_arishok6068516.cl"},
	{0x81654523FE846FBCULL, "855031792.ogg"},
	{0x816652CA547EE49FULL, "dre000_cut2d_ragstoriches.bnk"},
	{0x8166C276F587409CULL, "vxca_cavern1_haunted_gkj_0_0.vlm"},
	{0x81672FA1D1E7C296ULL, "460026062.ogg"},
	{0x81673BFE511B4F82ULL, "vxht_0_hightown_a3_c6o_0_0.vlm"},
	{0x816761BD01C7B568ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_up_big.ani"},
	{0x8167DE0B11754052ULL, "6107273_m.fxe"},
	{0x8167EA260D1E826BULL, "vxca_caverns1_mines_inp_0_0.vlm"},
	{0x816845F0304D96D3ULL, "gen00fl_anders6144057.cl"},
	{0x81685C6A34DE2E87ULL, "sse_ss_justice.gda"},
	{0x8168AA3B1DCC2AD0ULL, "vxlt_gallowstemplar_f_39m_0_0.vlm"},
	{0x8168BF4851BD5197ULL, "vxlt_backalley_n_1kv_0_0.vlm"},
	{0x81691C42F1B71415ULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.dura.ani"},
	{0x816928466E5955A0ULL, "465553246.ogg"},
	{0x8169A241779B44EDULL, "850681168.ogg"},
	{0x816A34CA54829AA7ULL, "dre000_cut2d_ragstoriches.cnv"},
	{0x816A367F4C5AA12DULL, "842736915.ogg"},
	{0x816A7EEE09339267ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.cess.gad"},
	{0x816AEB26DC87F225ULL, "mag170_thrask6054100.cl"},
	{0x816AF8291FF34D50ULL, "gen00fl_varric6033705.cl"},
	{0x816B073BD9C65D63ULL, "rog_stunparalyze.dds"},
	{0x816C066C62081B4AULL, "6041950_f.fxe"},
	{0x816CAFD9542187B4ULL, "vxw_sundermount_30s_0_1.vlm"},
	{0x816D8E1EC8F733A4ULL, "461490969.ogg"},
	{0x816DE7E3B5B4E4B5ULL, "lgt000ip_ragged_manifest.utp"},
	{0x816E1FF76E28FE43ULL, "vxlt_0_lowtown_docks_a3_coj_0_0.vlm"},
	{0x816E2D6AE013C8C7ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.cess.gad"},
	{0x816E33C5071A4E9DULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em04.phy"},
	{0x816EF5A6710AFDA6ULL, "vxca_caverns2prc_aic_0_0.vlm"},
	{0x816F0C5D21AAA105ULL, "6127306_m.fxe"},
	{0x816FE816A6D5F27BULL, "vxw_encounter2_camp_hs_0_0.vlm"},
	{0x817011E68012C2BAULL, "35541808.ogg"},
	{0x8170541312A85D88ULL, "123570912.ogg"},
	{0x817073EE0937EB74ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.cess.evt"},
	{0x81708CAE82A85159ULL, "mag160_feynriel6033400.cl"},
	{0x81714DCA9847BE4FULL, "6075044_f.fxe"},
	{0x81725915D321C512ULL, "vxca_caverns1_ikc_0_0.vlm"},
	{0x8173138C77737814ULL, "art\\characters\\creatures\\corpse\\animation\\blendtree\\corpse.anb"},
	{0x8173AD455FA507ABULL, "vxht_mansion_n_539_0_0.vlm"},
	{0x8173D433B3CB8041ULL, "239825604.ogg"},
	{0x817452EE597E2DDEULL, "564030208.ogg"},
	{0x817482C7D41C4BE7ULL, "177982939.ogg"},
	{0x8175226AE019D4D4ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.cess.evt"},
	{0x817538D2830BE8C0ULL, "hlac_lt_backalley_g_0_2z.dds"},
	{0x817581D7F84594A1ULL, "6081701_m.fxe"},
	{0x8175B0D53A2ACF72ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_6.dds"},
	{0x81760F6E07979139ULL, "vxht_0_hightown_n_7g1_0_0.vlm"},
	{0x81765FF8D3331DEEULL, "6140221_m.fxe"},
	{0x8177312543B36A17ULL, "vxlt_backalley_n_1vq_0_0.vlm"},
	{0x81778466AD9BDF3FULL, "vxca_cavern1_haunted_i37_0_0.vlm"},
	{0x81779ABC7E57BDBCULL, "212121579.ogg"},
	{0x817842B22AA64FD7ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.cess.gad"},
	{0x8178819988663F16ULL, "zz_vault_debug6203341.cl"},
	{0x8178ABA03297EDFEULL, "atl_worldmap_dxt1_01.dds"},
	{0x8178BE8F3B6BCF4CULL, "vxht_bartrand_n_67f_0_0.vlm"},
	{0x8179297024510E44ULL, "97075309.ogg"},
	{0x8179D18E77F7216BULL, "ave200_donnic6111073.cl"},
	{0x817ACCBEAE4D8E41ULL, "vxlt_0_lowtown_gpk_0_0.vlm"},
	{0x817B0678F77A4302ULL, "6128942_m.fxe"},
	{0x817B2EAF37DC167EULL, "6001896_m.fxe"},
	{0x817B82C741AD57FBULL, "zz_fen_debug6215252.cl"},
	{0x817BA3FDB623EAAAULL, "vxw_bonepit_yy_0_0.vlm"},
	{0x817BC5B386A5115DULL, "hm_icarver_h1s1.mor"},
	{0x817C1B29CFBFDD3CULL, "840583086.ogg"},
	{0x817CCC7138D6C5BAULL, "sw_imp_dis.mao"},
	{0x817CEE42F3EDD6CDULL, "vxht_0_hightown_ap9_0_0.vlm"},
	{0x817D151DB194D538ULL, "w_blightlands_121_0.phy"},
	{0x817D866C3DD65030ULL, "6116048_m.fxe"},
	{0x817D8C4D14813233ULL, "6095250_m.fxe"},
	{0x817E178834A804C3ULL, "748050171.ogg"},
	{0x817E59F398CC0AFBULL, "vxw_encounter2_camp_cr_0_0.vlm"},
	{0x817E6BEE09441AADULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.cess.ani"},
	{0x817E937CABBD955BULL, "qun131_taskmaster6059002.cl"},
	{0x817F34C50728C3A2ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em04.mmh"},
	{0x817F36C50728C758ULL, "art\\characters\\playercharacter\\elfmale\\mt_art_em04.msh"},
	{0x817F55DF8676B8C9ULL, "6092187_m.fxe"},
	{0x817F55F2B0A6FDDAULL, "240973319.ogg"},
	{0x817F57B22AAC9264ULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.cess.evt"},
	{0x817FB40B32F3D42EULL, "dae011_badpoet_sta6146925.cl"},
	{0x8180ACF4D349DB2DULL, "926187464.ogg"},
	{0x8180D78D64D67521ULL, "fx_whirlwind.dds"},
	{0x8180E7A8CFEADE2EULL, "rdr111dg_martin6054494.cl"},
	{0x818154DE89C5E310ULL, "hlac_lt_0_lowtown_docks_65533_0z.dds"},
	{0x81815CD531F09D25ULL, "34614643.ogg"},
	{0x818175C8DDB04EE8ULL, "659400259.ogg"},
	{0x81818865755F80A4ULL, "575554050.ogg"},
	{0x8181AD804AC828A6ULL, "hlac_lt_0_lowtown_n_131073_1x.dds"},
	{0x81821A6AE024510DULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.cess.ani"},
	{0x81825FDC7D7CEC31ULL, "hlac_w_woundedcoast_-131064_0z.dds"},
	{0x8182692BC410309BULL, "533753838.ogg"},
	{0x818302372A2FF496ULL, "llac_lt_backalley_n_0_12.dds"},
	{0x818315449CE99EACULL, "mcr371_hostile_merrill6082161.cl"},
	{0x8183F3B4EA3BAD21ULL, "6084401_m.fxe"},
	{0x81841DC859020CECULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_cleric_chan_ex3.gad"},
	{0x81842E5DD6E9CB97ULL, "dae000_merchant_dm_mob6174304.cl"},
	{0x81843E03B34FF615ULL, "lt_0_lowtown_docks_a3_n.rml"},
	{0x81843E157625B339ULL, "vxdr_0_deeproads_fir_0_0.vlm"},
	{0x81847DCE915AE595ULL, "6149632_m.fxe"},
	{0x8184D8FA966D3A2EULL, "gen00fl_varric6078706.cl"},
	{0x81860154E1BCE0E1ULL, "93745939.ogg"},
	{0x81860B54D933C4A8ULL, "isa311_isabela6092267.cl"},
	{0x81863580AFBFB238ULL, "vxht_mansion_n_503_0_0.vlm"},
	{0x81865F19E088BD9FULL, "6101696_m.fxe"},
	{0x818674CAF935E68AULL, "6155559_m.fxe"},
	{0x8186861C90FE4357ULL, "dae000_friend_turn_in.bnk"},
	{0x81872EC30FF3F929ULL, "ts_kirk_guards_prog_junk2.gda"},
	{0x81878FEA11CCE2FEULL, "mag100_thrask6031985.cl"},
	{0x8187964D3F8FD213ULL, "6172470_m.fxe"},
	{0x8187D20EC317FBE4ULL, "l_lgt331ar_shw_bloodmage.bnk"},
	{0x8188387C2E5B2B73ULL, "6137491_m.fxe"},
	{0x81885F7846E045C7ULL, "vxca_caverns1_ei8_0_0.vlm"},
	{0x81893268FC765C5AULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_nrma_0t.dds"},
	{0x818960429DFA9F8DULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medh.phy"},
	{0x8189681C9100466FULL, "dae000_friend_turn_in.cnv"},
	{0x818A0EC91CD607C9ULL, "6088914_f.fxe"},
	{0x818A15DAA81A032AULL, "vxw_bonepit_dz_0_0.vlm"},
	{0x818A2FCD70E77F15ULL, "llac_ca_caverns2prc_5802_23.dds"},
	{0x818A51B2C0D9E065ULL, "6081734_m.fxe"},
	{0x818A6F24F4DE7A41ULL, "28537769.ogg"},
	{0x818AFD9D31210E10ULL, "vxlt_undercityserialkiller_174_0_0.vlm"},
	{0x818B0B196BE12287ULL, "6127224_f.fxe"},
	{0x818B1CC8590829EAULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_cleric_chan_ex3.ani"},
	{0x818B5C2A39534730ULL, "32302068.ogg"},
	{0x818B819860BE75BCULL, "vxlt_darktown_48b_0_0.vlm"},
	{0x818B860599A44A9BULL, "dae000ip_fake_talk_1_d.utp"},
	{0x818B8BD11484A637ULL, "hlac_ca_cavern1_haunted_24421_7y.dds"},
	{0x818BD49E67F81FBCULL, "6143154_m.fxe"},
	{0x818BD96054F72EA6ULL, "vxca_caverns2prc_ace_0_0.vlm"},
	{0x818C1E709BA28A30ULL, "6082107_m.fxe"},
	{0x818C56143177FA08ULL, "vxw_bonepitmining_a2_1js_0_0.vlm"},
	{0x818D51109CFCE4B9ULL, "243722056.ogg"},
	{0x818D6FB22AB8F85DULL, "art\\vfx\\combat\\vfx_ha_bd_appear_c.cess.ani"},
	{0x818D9F2012DC7863ULL, "vxlt_warehouse_n_cx_0_0.vlm"},
	{0x818DC7BC786CD149ULL, "6097287_m.fxe"},
	{0x818DF2090FEBE737ULL, "dae100cr_qunari_warrior4.utc"},
	{0x818E26EE669589F3ULL, "ht_mansion_n_3381.anb"},
	{0x818E4549714244AFULL, "61252146.ogg"},
	{0x818E874AEBE255E6ULL, "gen00fl_aveline6112296.cl"},
	{0x818EAC81DBD6A564ULL, "gen00fl_anders6113027.cl"},
	{0x818EC03BEA95F56AULL, "vxlt_undercityencounter_1fm_0_0.vlm"},
	{0x818EF448B9AEA1AEULL, "vxca_cavern1_haunted_im4_0_0.vlm"},
	{0x818F56A36AB3A8B2ULL, "lt_0_lowtown_docks_45_0.rml"},
	{0x818FABDD0D33BDD3ULL, "310531277.ogg"},
	{0x818FDFE47C2DFF7CULL, "vfx_chunkyparticle_p1.vsh.fxo.dx11"},
	{0x818FE09BC306EBA5ULL, "580864886.ogg"},
	{0x819056DE5C970AB0ULL, "402422334.ogg"},
	{0x81908E621DB7B7DAULL, "vxlt_house_el_0_1.vlm"},
	{0x8190AFB3BE1CD363ULL, "6129881_m.fxe"},
	{0x81912EE68BCA2C16ULL, "707289090.ogg"},
	{0x81918914D7A75AC3ULL, "vxlt_gamlenshouse_n_nq_0_0.vlm"},
	{0x81918A204CDABB37ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postb.phy"},
	{0x8192154B702B9A7FULL, "330506602.ogg"},
	{0x8192776C3FD75D1FULL, "dae150_chandan6148811.cl"},
	{0x8192B8E976FF771DULL, "vxlt_gallowscourtyard_63r_0_0.vlm"},
	{0x8192D34D88F01748ULL, "458482485.ogg"},
	{0x8193811A642EFA62ULL, "85464124.ogg"},
	{0x8193F317582AD040ULL, "vrc221_bartrand6043280.cl"},
	{0x8194E5F220280819ULL, "vxht_0_hightown_a3_c3f_0_1.vlm"},
	{0x8194ECF7D9D4EFE4ULL, "ico_gloves_heavy_5.dds"},
	{0x81958EF52B15BA93ULL, "vxca_caverns1_mines_jf7_0_0.vlm"},
	{0x819598B52D19E27FULL, "98995245.ogg"},
	{0x8195C0F907E488F5ULL, "273830411.ogg"},
	{0x8195FA3C1BBE65F6ULL, "607923036.ogg"},
	{0x81962D1994516900ULL, "llac_lt_undercityserialkiller_-1_24.dds"},
	{0x81963221805091FEULL, "6139244_f.fxe"},
	{0x8196ADE7AC730211ULL, "t3_bow_com.tnt"},
	{0x8196F75F733168C2ULL, "6107509_m.fxe"},
	{0x81970794CBF2B2E1ULL, "t3_boy_b03.tnt"},
	{0x819715106D6F3699ULL, "dae000_arishok6078169.cl"},
	{0x81979889D729F4CBULL, "root_mage_creation.dds"},
	{0x819866D6FAEFC622ULL, "zz_and_debug6134101.cl"},
	{0x8198A0E79E60F8C7ULL, "83945187.ogg"},
	{0x81998964FD1981FBULL, "m_dae110ar_docks_nt.bnk"},
	{0x819A5FE2BA20B00AULL, "llac_lt_0_lowtown_a2_n_c_-65535_1.dds"},
	{0x819ADDFEFB4566DEULL, "and110_refugee6090301.cl"},
	{0x819B7BAD51B3F730ULL, "lgt000cr_hate_demon.utc"},
	{0x819BBA1EFC5A26AFULL, "vxht_0_hightown_a3_c57_0_0.vlm"},
	{0x819BD1C465246500ULL, "vxlt_gallowscourtyard_n_700_0_0.vlm"},
	{0x819C597B818DEE28ULL, "one100pt_loose_end_athe.ncs"},
	{0x819C8B92BEE10BB5ULL, "19537321.ogg"},
	{0x819D0A2CDE8D4095ULL, "vxlt_0_lowtown_a2_gc6_0_0.vlm"},
	{0x819D2124B7D8C862ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.cess.gad"},
	{0x819D3218DC8C0DEAULL, "vxlt_undercityencounter_1cn_0_0.vlm"},
	{0x819D52003298133DULL, "6143685_m.fxe"},
	{0x819D6625ABA4BA2EULL, "vxca_cavern1_haunted_ij5_0_0.vlm"},
	{0x819DA857E42E46C7ULL, "105668735.ogg"},
	{0x819DAE91FA456279ULL, "art\\levels\\ztest\\chad\\textures\\prp_orlaischair_n.dds"},
	{0x819E27508AFBFE71ULL, "6048409_m.fxe"},
	{0x819E3B2A8CA5B361ULL, "plt_cod_crt_profane.plo"},
	{0x819E9D15FFC7DC6CULL, "48313195.ogg"},
	{0x819EE88D64EF8FABULL, "fx_whirlwind.mao"},
	{0x819F6528D02B6541ULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v1.ani"},
	{0x819F79069E9A02C6ULL, "147375795.ogg"},
	{0x81A02C6F525BAC17ULL, "qcr220_meredith6185149.cl"},
	{0x81A0627A5DC8A6B7ULL, "gen_im_trs_gen_amu_r1.uti"},
	{0x81A0630DB3A11891ULL, "6052756_m.fxe"},
	{0x81A073D4FE54B370ULL, "hlac_lt_0_lowtown_a2_n_c_-3_2x.dds"},
	{0x81A16F8CB1658A7EULL, "llac_lt_warehouse_1_2.dds"},
	{0x81A17B7D61F1206FULL, "cloud.mao"},
	{0x81A2CD697FD59BE7ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.impa.gad"},
	{0x81A34424B7DD6F59ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.cess.evt"},
	{0x81A3654F3DA04C83ULL, "6016163_m.fxe"},
	{0x81A3939E477B5853ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.cess.evt"},
	{0x81A3BFDBA2FC2B40ULL, "fen110_fenris6100385.cl"},
	{0x81A43092E6DEBDB5ULL, "6020421_m.fxe"},
	{0x81A434C7B9F4CC7DULL, "62267161.ogg"},
	{0x81A43C29301316F6ULL, "lt_0_lowtown_n_32_0.anb"},
	{0x81A57D94AA05AEFEULL, "art\\vfx\\cinematic\\vfx_blood_b_splatters_p.impa.ani"},
	{0x81A583BB8611C939ULL, "308516309.ogg"},
	{0x81A62C34280431A0ULL, "hlac_lt_0_lowtown_docks_n_-131074_7y.dds"},
	{0x81A638189481B326ULL, "hlac_lt_gallowstemplar_n_7198_7y.dds"},
	{0x81A67028D03196E3ULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v1.gad"},
	{0x81A696A3A5CAC82DULL, "static_static_p20.psh.fxo.dx11.assembly"},
	{0x81A6DD70A16D0331ULL, "6111400_m.fxe"},
	{0x81A72974F7E14A4AULL, "6136813_m.fxe"},
	{0x81A7CF3B6C0800A2ULL, "6043454_m.fxe"},
	{0x81A8C0D18EDE1B9AULL, "14551275.ogg"},
	{0x81A8C2697FD9F4F4ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.impa.evt"},
	{0x81A8EC779F40D8CCULL, "621631883.ogg"},
	{0x81A91AC8103C7E64ULL, "6135357_m.fxe"},
	{0x81A955B0979A9EADULL, "644366932.ogg"},
	{0x81A95E59212F34AAULL, "dae000_arishok6068525.cl"},
	{0x81A9939154A37119ULL, "vxlt_warehouse_n_i4_0_0.vlm"},
	{0x81A9F132C5278C26ULL, "vxlt_0_lowtown_a3_n_c_hp3_0_0.vlm"},
	{0x81AA8C9E47816A94ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.cess.gad"},
	{0x81AAAC68516436ECULL, "hlac_lt_0_lowtown_196606_0y.dds"},
	{0x81AB0B5CB87A6FC1ULL, "vxht_chantry_n_30p_0_0.vlm"},
	{0x81AB61269B58A01BULL, "vxlt_0_lowtown_docks_a3_n_ctp_0_0.vlm"},
	{0x81AB7C09D0835915ULL, "vxlt_0_lowtown_a2_gf5_0_0.vlm"},
	{0x81AC552BE84A50CBULL, "211520015.ogg"},
	{0x81AC67B4B8567BCEULL, "llac_ht_chantry_2534_38.dds"},
	{0x81AD1660E1B64CC6ULL, "127113690.ogg"},
	{0x81AD4D28D03779C8ULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v1.evt"},
	{0x81AD64F1BB9E516DULL, "dre000_hireling16038636.cl"},
	{0x81ADCFE09A5985B2ULL, "hlac_ht_0_hightown_-2_0z.dds"},
	{0x81AE37D91E88B03BULL, "mag120_macha6087854.cl"},
	{0x81AF40D3E4E7B35CULL, "vxlt_darktown_5b5_0_0.vlm"},
	{0x81AF5581B3E67F8AULL, "s_blood.mao"},
	{0x81AFA78BF86C6E3FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_roba_t.dds"},
	{0x81AFEEA4E5A5E689ULL, "vxw_bonepitdestroyed_a3_8a_0_0.vlm"},
	{0x81B0160FF75A40BFULL, "6118477_f.fxe"},
	{0x81B04CE1F1A2161DULL, "732573097.ogg"},
	{0x81B096730A6256A7ULL, "hlac_ht_keep_n_3739_1z.dds"},
	{0x81B11B1842F581A7ULL, "dae000ip_crowd_lm4_talk3.utp"},
	{0x81B11E66C692A279ULL, "qun110_petrice6036861.cl"},
	{0x81B144D2BE8264B7ULL, "hlac_lt_0_lowtown_docks_n_-65540_1z.dds"},
	{0x81B14C24B7E9BA48ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.cess.ani"},
	{0x81B17ABBB96C8AB5ULL, "6101455_f.fxe"},
	{0x81B19B9E4787A372ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.cess.ani"},
	{0x81B1AC9A7AB38C41ULL, "6042483_m.fxe"},
	{0x81B1D5D5EF37C9AFULL, "lt_0_lowtown_docks_a3_n_12_0_lc.mmh"},
	{0x81B20F3FC1B90EE6ULL, "dae011_drunk1_sta6144378.cl"},
	{0x81B24D87AF890295ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaritent.phy"},
	{0x81B316F26A4740E8ULL, "vxht_0_hightown_a3_a85_0_0.vlm"},
	{0x81B33BE1A40986C7ULL, "stoney2_n.dds"},
	{0x81B3730931AE50E2ULL, "hlac_w_bonepit_0_8z.dds"},
	{0x81B385AA05DD0619ULL, "gen_im_arm_glv_lgt_act3_01.uti"},
	{0x81B39594AA1214CFULL, "art\\vfx\\cinematic\\vfx_blood_b_splatters_p.impa.evt"},
	{0x81B47E914BE585F1ULL, "bink_bink_p0.psh.fxo.dx11"},
	{0x81B4BC9D65992E82ULL, "static_static_p83.psh.fxo.dx11"},
	{0x81B4C20E70B43828ULL, "dae300_alistair_king6133611.cl"},
	{0x81B5213B85FBEE7EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_floor_n.dds"},
	{0x81B5636FB09E921DULL, "vxlt_0_lowtown_e2g_0_0.vlm"},
	{0x81B5D37E4B3C5A0BULL, "6148848_m.fxe"},
	{0x81B647ED971237CFULL, "one100dg_nabil6033150.cl"},
	{0x81B6BA697FE6242DULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.impa.ani"},
	{0x81B6C995CF52DA7AULL, "435846000.ogg"},
	{0x81B6D27714504CBEULL, "al_warrior_s01.gda"},
	{0x81B6F880DD69922BULL, "dae000_arishok6067776.cl"},
	{0x81B6F8C06537B4F9ULL, "6103018_m.fxe"},
	{0x81B71BDDE4CA4905ULL, "6171051_m.fxe"},
	{0x81B77AF544643018ULL, "vxht_chantry_n_28t_0_0.vlm"},
	{0x81B7E36FECF0F3A5ULL, "vxht_mansion2_1kq_0_1.vlm"},
	{0x81B84DC5AEBFFF0FULL, "qun110_petrice6018746.cl"},
	{0x81B890757D378945ULL, "6030459_m.fxe"},
	{0x81B8C682088C4BB3ULL, "827391629.ogg"},
	{0x81B99CAD159D3881ULL, "59770379.ogg"},
	{0x81B9CF7473A2C83EULL, "63604317.ogg"},
	{0x81B9D3038D4EB89BULL, "vxlt_0_lowtown_docks_a3_n_cqs_0_0.vlm"},
	{0x81B9E64114D7CEF3ULL, "vxlt_0_lowtown_n_erj_0_0.vlm"},
	{0x81B9E82C017283EAULL, "6126593_m.fxe"},
	{0x81BA038B3A80861AULL, "vxlt_gallowstemplar_f_5j8_0_0.vlm"},
	{0x81BA148EEE8C71CCULL, "387945109.ogg"},
	{0x81BA27D49C86EE69ULL, "vxlt_undercityserialkiller_104_0_0.vlm"},
	{0x81BA43EE66BAF5AFULL, "ht_mansion_n_3381.rml"},
	{0x81BA4F9E80D001E5ULL, "6143695_f.fxe"},
	{0x81BA8EC16E9BB56CULL, "hlac_w_bonepit_131078_0z.dds"},
	{0x81BA9694AA183578ULL, "art\\vfx\\cinematic\\vfx_blood_b_splatters_p.impa.gad"},
	{0x81BA97A837399E07ULL, "6164020_m.fxe"},
	{0x81BAD4574076FDEFULL, "6097384_m.fxe"},
	{0x81BAE1E016CBB5F3ULL, "mcr370_negotiation6095513.cl"},
	{0x81BB372126710D03ULL, "vxlt_0_lowtown_docks_a3_d62_0_0.vlm"},
	{0x81BBAFA25BDF1863ULL, "hlac_lt_warehouse_n_65537_3x.dds"},
	{0x81BBC7FA0A2BAF8FULL, "art\\vfx\\materials\\vfx_crater_full.mao"},
	{0x81BBF6EBF4733EA5ULL, "6171213_m.fxe"},
	{0x81BC4D9CA931C569ULL, "74082071.ogg"},
	{0x81BC816812DD6DA2ULL, "6033646_m.fxe"},
	{0x81BCDAB3E9D1D9D7ULL, "29971799.ogg"},
	{0x81BD4320F4CACD2EULL, "dae000_flemeth6000974.cl"},
	{0x81BD53F0A17898EFULL, "gen_im_trs_dal_junk2_01.uti"},
	{0x81BD54CCD2BA2223ULL, "6156182_m.fxe"},
	{0x81BDC65A0EDDB45EULL, "537563232.ogg"},
	{0x81BE299B818E826FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grindstone_d.dds"},
	{0x81BE6C244C742D4CULL, "dae000cr_grand_cleric6124718.cl"},
	{0x81BE6E2C69E6C260ULL, "vxht_brothel_n_5jr_0_0.vlm"},
	{0x81BEE32DBB19C132ULL, "6125989_m.fxe"},
	{0x81C05B2015FFC071ULL, "llac_lt_gallowstemplar_7198_4.dds"},
	{0x81C0777D53F20E55ULL, "vxht_mansion2_34o_0_0.vlm"},
	{0x81C0E48C00F80413ULL, "vxca_caverns1_mines_jw8_0_0.vlm"},
	{0x81C0F7516868CE3BULL, "gen00fl_isabela6085855.cl"},
	{0x81C12532B7F16271ULL, "6164552_m.fxe"},
	{0x81C12D9E2297857CULL, "224298128.ogg"},
	{0x81C148D3871B85FDULL, "vxht_keep_n_60m_0_0.vlm"},
	{0x81C1A34ED1BCE48DULL, "vxw_bonepitmining_a2_kz_0_0.vlm"},
	{0x81C288C1DA8211F5ULL, "gen00fl_isabela6079270.cl"},
	{0x81C2B982B1043C53ULL, "764205599.ogg"},
	{0x81C330A9DB58F388ULL, "vxw_bonepitmining_a2_277_0_0.vlm"},
	{0x81C34EA1E80371D4ULL, "gen00fl_varric6033572.cl"},
	{0x81C3577E3239F35EULL, "hlac_lt_0_lowtown_a2_n_c_-65535_4y.dds"},
	{0x81C3C2EDA0E7073EULL, "and111_lirene6084527.cl"},
	{0x81C3DF9C13A3F903ULL, "hlac_ht_store_187_2x.dds"},
	{0x81C3F05CB7737B1BULL, "dae000ip_crowd_em1_talk7.utp"},
	{0x81C4EE6FF5CE581FULL, "vxlt_undercitydungeon_1al_0_0.vlm"},
	{0x81C58612DE38B55AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_loop.gad"},
	{0x81C5AD9614EEF09CULL, "6152164_f.fxe"},
	{0x81C5ECD2365A4898ULL, "vxht_chantry_n_23u_0_0.vlm"},
	{0x81C61D87F9505C08ULL, "vxw_bonepitmining_a2_1vo_0_0.vlm"},
	{0x81C61F835CED218BULL, "6104237_m.fxe"},
	{0x81C6A57A6424C9A0ULL, "dae000_oldmangossip6174911.cl"},
	{0x81C6CD12C8AC9CE1ULL, "hlac_lt_0_lowtown_n_65536_2x.dds"},
	{0x81C7A8EE9833F831ULL, "zz_followers_banter6157234.cl"},
	{0x81C7BB5A041B6957ULL, "zz_dre_debug6168213.cl"},
	{0x81C7F51D4C365F5EULL, "vxca_cavern1_haunted_eby_0_0.vlm"},
	{0x81C8581E06CDE773ULL, "vxlt_0_lowtown_n_eom_0_0.vlm"},
	{0x81C85A22DFAA7A89ULL, "6053232_m.fxe"},
	{0x81C85B7FB295D869ULL, "6072843_f.fxe"},
	{0x81C8C6E68AB86441ULL, "vxca_cavern1_haunted_j5d_0_0.vlm"},
	{0x81C937644496FA07ULL, "fonts_cs-cz.gfx"},
	{0x81C9547DDA3E52ABULL, "577497758.ogg"},
	{0x81C95D9EC15F6292ULL, "6171838_m.fxe"},
	{0x81C9A658C2DB31EFULL, "868491667.ogg"},
	{0x81C9BAC279CBE357ULL, "vxht_keep_499_0_0.vlm"},
	{0x81CA68C8E6695725ULL, "6174113_m.fxe"},
	{0x81CB6AAD1634CD12ULL, "307796699.ogg"},
	{0x81CBE29E53731E7BULL, "dae000_gamlen6026200.cl"},
	{0x81CC2E7E185FED3DULL, "and100_anders6095948.cl"},
	{0x81CC8912DE3ED971ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_loop.evt"},
	{0x81CC8C641D4FCFD5ULL, "vxlt_warehouse_qq_0_0.vlm"},
	{0x81CCF2E1741F827BULL, "338318337.ogg"},
	{0x81CD029FBFDF4182ULL, "t3_arm_dsk.tnt"},
	{0x81CD07C79341EB5BULL, "382892449.ogg"},
	{0x81CD550AAD0FD9B2ULL, "6122889_m.fxe"},
	{0x81CDA395DBB45804ULL, "6129385_m.fxe"},
	{0x81CDCEE56CD79A4AULL, "vxlt_warehouse_n_8h_0_0.vlm"},
	{0x81CDF7CE3C3A7A02ULL, "vxht_keep_4fy_0_0.vlm"},
	{0x81CE887A9DBC7AE5ULL, "vxlt_0_lowtown_a2_gcf_0_0.vlm"},
	{0x81CE88DF3D42EB80ULL, "59148065.ogg"},
	{0x81CEAE00E5195A22ULL, "656754571.ogg"},
	{0x81CEFC79651FB9BCULL, "191090662.ogg"},
	{0x81CF1077484DE77AULL, "212158573.ogg"},
	{0x81CF3714CE7A48BBULL, "295279830.ogg"},
	{0x81D01C3BBED57594ULL, "892623497.ogg"},
	{0x81D04B794826278DULL, "hlac_ca_cavern1_haunted_24421_6x.dds"},
	{0x81D0611AEAD9073DULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyo_d.dds"},
	{0x81D07D8DC6A19A70ULL, "vxca_caverns2_8og_0_0.vlm"},
	{0x81D0BB2F8FFADB29ULL, "723057313.ogg"},
	{0x81D0DDCC8138AA41ULL, "qun220cr_javaris6071890.cl"},
	{0x81D0EAB0F5301D72ULL, "vxlt_0_lowtown_a2_fer_0_0.vlm"},
	{0x81D12060928CAA43ULL, "vxht_playershouse_n_1la_0_0.vlm"},
	{0x81D1A286CD4F0C08ULL, "vxw_bonepitmining_a2_224_0_0.vlm"},
	{0x81D1AAE38C61B481ULL, "vxlt_darktown_5bd_0_0.vlm"},
	{0x81D2689466E47B55ULL, "412561309.ogg"},
	{0x81D26B354C969DE0ULL, "6135825_m.fxe"},
	{0x81D2F72AE2C44531ULL, "hlac_lt_0_lowtown_docks_n_131068_0y.dds"},
	{0x81D3A8094D439F6DULL, "lgt150_nat_were_supporter6110935.cl"},
	{0x81D3C5E18EDE707FULL, "6149128_m.fxe"},
	{0x81D46D1D7E3AA65AULL, "6111123_f.fxe"},
	{0x81D4791B74C8C997ULL, "character_character_p15.psh.fxo.dx11"},
	{0x81D4E9282A85BA99ULL, "vxlt_warehouse_n_v9_0_0.vlm"},
	{0x81D4EE23E848A654ULL, "pro000_friedrich6051066.cl"},
	{0x81D4F23D5265DBFDULL, "dae000_gamlen6124148.cl"},
	{0x81D52DCCDD39827CULL, "art\\characters\\playercharacter\\dwarfmale\\bianca.mao"},
	{0x81D5803A5DC4B0FEULL, "vxw_encounter2_camp_nd_0_0.vlm"},
	{0x81D634EB1F3D9B04ULL, "6110954_m.fxe"},
	{0x81D666FA3E2C77DEULL, "vxca_cavern1_haunted_egz_0_0.vlm"},
	{0x81D6C9FAF8C3FFF3ULL, "vxlt_0_lowtown_n_ehl_0_0.vlm"},
	{0x81D71F29303E3E8AULL, "lt_0_lowtown_n_32_0.rml"},
	{0x81D73967FD513F0DULL, "vxlt_0_lowtown_n_e82_0_0.vlm"},
	{0x81D75E51105541AAULL, "6018615_f.fxe"},
	{0x81D7D9DA39FEB11CULL, "art\\vfx\\combat\\vfx_gib_boo_lgta.mao"},
	{0x81D85DE161306DBCULL, "492318823.ogg"},
	{0x81D85F204D16748FULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postb.mao"},
	{0x81D86B204D16892CULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postb.mmh"},
	{0x81D871204D16931EULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postb.msh"},
	{0x81D8C853B77021E4ULL, "436432899.ogg"},
	{0x81D8D98BCDCD3F6CULL, "43343041.ogg"},
	{0x81D8F5F1CDBC8494ULL, "vxca_cavern1_haunted_dr6_0_0.vlm"},
	{0x81D9021DB1E313CDULL, "w_blightlands_121_0.msh"},
	{0x81D9101DB1E32B00ULL, "w_blightlands_121_0.mao"},
	{0x81D9141DB1E33253ULL, "w_blightlands_121_0.mmh"},
	{0x81D93A29538EB0C9ULL, "6154726_m.fxe"},
	{0x81D97702F50A6FACULL, "vxw_sundermount_269_0_0.vlm"},
	{0x81D9A112DE498C40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\mage_combat_loop.ani"},
	{0x81DA272A6ED4AD75ULL, "6143471_f.fxe"},
	{0x81DA683C8563B89DULL, "391756558.ogg"},
	{0x81DA948E2E044C79ULL, "dae300_alistair_king6134886.cl"},
	{0x81DAA8A9A8CA6389ULL, "vxw_sundermount_329_0_0.vlm"},
	{0x81DADC9E3318035CULL, "qun120_arishok6051487.cl"},
	{0x81DAFE410F45E855ULL, "vxlt_warehouse_zp_0_0.vlm"},
	{0x81DBF14E27FCA9C4ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_mother_victim.ani"},
	{0x81DC98CB8738EF16ULL, "llac_lt_0_lowtown_-131072_8.dds"},
	{0x81DCBADFF064B683ULL, "6101022_f.fxe"},
	{0x81DF30A5E41BE146ULL, "vxca_caverns2prc_979_0_0.vlm"},
	{0x81DF6F5F0303468BULL, "toolset_event_tracks.gda"},
	{0x81E00A87DEA01E0AULL, "375358283.ogg"},
	{0x81E00E9E5A164C7DULL, "vxht_mansion_n_6k8_0_0.vlm"},
	{0x81E03B1E463651BAULL, "147228310.ogg"},
	{0x81E0EA913AD4CF86ULL, "vxw_encounter2_camp_qu_0_0.vlm"},
	{0x81E17764FD5718F9ULL, "m_dae110ar_docks_nt.wbd"},
	{0x81E1D229D9BA891FULL, "vxlt_undercitydungeon_1on_0_0.vlm"},
	{0x81E22A445BD5C78AULL, "vxlt_blackemporium_ij_0_0.vlm"},
	{0x81E237BE80D488FFULL, "6142028_m.fxe"},
	{0x81E24FFABE30F45BULL, "w_sundermount_79_0.mmh"},
	{0x81E25BFABE3108F8ULL, "w_sundermount_79_0.mao"},
	{0x81E26DFABE312745ULL, "w_sundermount_79_0.msh"},
	{0x81E2A892E8C585FFULL, "449633622.ogg"},
	{0x81E2EE4E2802C326ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_mother_victim.gad"},
	{0x81E2F0655736A4C3ULL, "mcr000pt_main.ncs"},
	{0x81E3B8A68CFFEE26ULL, "vxlt_0_lowtown_a3_n_c_hl7_0_0.vlm"},
	{0x81E3F2174FBAC97EULL, "vxw_encounter2_camp_ie_0_0.vlm"},
	{0x81E43165A89CCDB8ULL, "heal.mao"},
	{0x81E442FA418D977BULL, "794628106.ogg"},
	{0x81E4B9FCCD20A492ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyejitter_qunari.gad"},
	{0x81E5885C5DF3D42DULL, "w_encounter3_77_0.msh"},
	{0x81E5965C5DF3EC60ULL, "w_encounter3_77_0.mao"},
	{0x81E59A5C5DF3F2B3ULL, "w_encounter3_77_0.mmh"},
	{0x81E60B4EACE3CFB1ULL, "369758720.ogg"},
	{0x81E625365B60CE65ULL, "hlac_ht_chantry_2534_41y.dds"},
	{0x81E6258755B1547FULL, "vxht_0_hightown_a3_c3q_0_0.vlm"},
	{0x81E64A937A839FBAULL, "598010438.ogg"},
	{0x81E66A314C14287AULL, "6112752_m.fxe"},
	{0x81E74969D05FE3CEULL, "mind_custic.mao"},
	{0x81E749F1210D61BCULL, "gen00fl_anders6125567.cl"},
	{0x81E8B7BA2F25E447ULL, "vxlt_0_lowtown_docks_a3_cov_0_0.vlm"},
	{0x81E8D4671C82ED1DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_rough.mao"},
	{0x81E8D8671C82F456ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_rough.mmh"},
	{0x81E8E2671C830554ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_rough.msh"},
	{0x81E91A869AC07C09ULL, "vxw_sundermount_398_0_0.vlm"},
	{0x81E9844EEDA5C84CULL, "hlac_dr_0_deeproads_5842_9z.dds"},
	{0x81E9E99232CC73EDULL, "vxdr_primevaltomb_6zp_0_1.vlm"},
	{0x81EA48ABE3F126EEULL, "vxca_cave_18c_0_1.vlm"},
	{0x81EAF84EB80C1FA0ULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.mmh"},
	{0x81EB2485CC0C93FDULL, "6133057_m.fxe"},
	{0x81EB52BD4ADCCFF1ULL, "hlac_lt_0_lowtown_docks_0_0y.dds"},
	{0x81EBE1CD9C09E90AULL, "hlac_lt_0_lowtown_a2_n_c_131072_4x.dds"},
	{0x81EBE657942A8281ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comc_0s.dds"},
	{0x81EC2EE0D3B5F144ULL, "gen00fl_sebastian6180394.cl"},
	{0x81EC33C3BD7AF3F0ULL, "zz_dre_debug6025555.cl"},
	{0x81ECC5B44E30EE15ULL, "vxca_caverns2_815_0_1.vlm"},
	{0x81ECC718E2BF35A8ULL, "6054389_f.fxe"},
	{0x81ECFEFEF3E51CF5ULL, "6014251_m.fxe"},
	{0x81ED15A422538F73ULL, "static_passthrough_p0.vsh.fxo.dx11"},
	{0x81ED6147AA8DCB70ULL, "vxca_caverns2_8qe_0_0.vlm"},
	{0x81ED86EA61A13446ULL, "477366849.ogg"},
	{0x81ED8FD6A48EE928ULL, "6020464_m.fxe"},
	{0x81EDF49DC1605C22ULL, "6209854_m.fxe"},
	{0x81EE345731986125ULL, "zz_warriorslate_white.cut"},
	{0x81EE533D0144E247ULL, "hlac_w_encounter2_camp_262146_0x.dds"},
	{0x81EE54180458A599ULL, "llac_w_encounter3_-262142_0.dds"},
	{0x81EE807B4C0C64FDULL, "vxht_mansion_n_6n7_0_0.vlm"},
	{0x81EEA833B0016C85ULL, "zz_qun_debug6103424.cl"},
	{0x81EEAA1ED1CB32F4ULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_bdy_ggg_0n.dds"},
	{0x81EEB83BE89D89B7ULL, "973038559.ogg"},
	{0x81EF2F6CE01115B8ULL, "vxw_encounter2_camp_l8_0_0.vlm"},
	{0x81EF400E5A28BD2AULL, "hlac_ht_mansion2_818_0z.dds"},
	{0x81EF5C6E2CCAE806ULL, "vxw_encounter2_camp_lv_0_0.vlm"},
	{0x81EFBB3F6C8BECB7ULL, "818536381.ogg"},
	{0x81EFF2753B6C8D8CULL, "6175069_m.fxe"},
	{0x81F023E6D1886307ULL, "hlac_ht_chantry_2534_29y.dds"},
	{0x81F0362A3EF2387CULL, "285510463.ogg"},
	{0x81F04406CBB0A19FULL, "vxlt_undercitydungeon_1ri_0_0.vlm"},
	{0x81F070FABE3D69C0ULL, "w_sundermount_79_0.anb"},
	{0x81F09B686FF8D97FULL, "fen310_party_comment6116865.cl"},
	{0x81F0EAB06D95243BULL, "vxlt_warehouse_11n_0_0.vlm"},
	{0x81F19FACF6815B11ULL, "vxw_bonepitmining_a2_8c_0_0.vlm"},
	{0x81F1EB13613EB2C9ULL, "6081716_m.fxe"},
	{0x81F215C963DFBB1CULL, "gen00fl_merrill6128922.cl"},
	{0x81F243F2BD197A98ULL, "and231ip_at_darktown.ncs"},
	{0x81F263F441B0E1FEULL, "vxw_encounter2_camp_df_0_0.vlm"},
	{0x81F264CD692ED09BULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowtwohandedweapon.evt"},
	{0x81F271647DA0376EULL, "w_bonepitdestroyed_a3_64_0.msh"},
	{0x81F27F647DA04EBFULL, "w_bonepitdestroyed_a3_64_0.mao"},
	{0x81F28B647DA0635CULL, "w_bonepitdestroyed_a3_64_0.mmh"},
	{0x81F2ED92FE47E0C4ULL, "t3_abm_b02.tnt"},
	{0x81F2FDADF9F6C5C2ULL, "hlac_lt_0_lowtown_n_0_3y.dds"},
	{0x81F32C6A14673EA4ULL, "60703156.ogg"},
	{0x81F367C28FF7AB57ULL, "522830544.ogg"},
	{0x81F3D0DE04E5908DULL, "6118138_f.fxe"},
	{0x81F42483435655AFULL, "6127970_f.fxe"},
	{0x81F44B285786A26EULL, "982963904.ogg"},
	{0x81F46700065FA026ULL, "hlac_lt_warehouse_1_4x.dds"},
	{0x81F4976447A76CFFULL, "vxht_0_hightown_a3_c4p_0_0.vlm"},
	{0x81F4FDDF857EC91FULL, "vxlt_0_lowtown_a3_n_c_hkg_0_0.vlm"},
	{0x81F55C1976C828E6ULL, "dae000_arishok6067634.cl"},
	{0x81F55F8985B9DC36ULL, "zz_t_trailer_aveline_v2.cut"},
	{0x81F5950246248ABEULL, "vxht_0_hightown_9v0_0_0.vlm"},
	{0x81F5CB410624258CULL, "153872556.ogg"},
	{0x81F6F1B81142B3D3ULL, "vxca_caverns1_mines_k2n_0_0.vlm"},
	{0x81F72929CA4523CEULL, "vxht_0_hightown_a3_c7p_0_0.vlm"},
	{0x81F75268E1DD6DCBULL, "dae000ip_crowd_mf8_talk7.utp"},
	{0x81F765E8BF87048BULL, "hlac_w_woundedcoast_-262137_0x.dds"},
	{0x81F86F3CB4F1CE33ULL, "301193190.ogg"},
	{0x81F88BE2590DA497ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_hv_dmg.gad"},
	{0x81F8DB4EB8182BBFULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.anb"},
	{0x81F935920701E797ULL, "169209530.ogg"},
	{0x81F95DCD6934E3DCULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowtwohandedweapon.gad"},
	{0x81F96087AFC525B0ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaritent.msh"},
	{0x81F96E87AFC53DCAULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaritent.mmh"},
	{0x81F97287AFC54401ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunaritent.mao"},
	{0x81F973F9C966A24BULL, "875261000.ogg"},
	{0x81F9A4FCCD32DD38ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\mh_o.eyejitter_qunari.ani"},
	{0x81FA6986BF56BE72ULL, "llac_ca_caverns2_5802_45.dds"},
	{0x81FA95E788E4FBE1ULL, "268799019.ogg"},
	{0x81FAB95A54F7FACCULL, "711308516.ogg"},
	{0x81FB222F2C0E5313ULL, "vxht_keep_62b_0_0.vlm"},
	{0x81FB912F0C836290ULL, "6162094_m.fxe"},
	{0x81FCC72612F91C59ULL, "6021116_m.fxe"},
	{0x81FD77D9E1788006ULL, "6057121_m.fxe"},
	{0x81FD9B99F84E1BC6ULL, "vxht_keep_n_3zq_0_0.vlm"},
	{0x81FDA149D2072E38ULL, "vxw_encounter2_camp_g9_0_0.vlm"},
	{0x81FDCE4B1EC10086ULL, "vxw_encounter2_camp_gw_0_0.vlm"},
	{0x81FE063BDD683FB1ULL, "vxlt_0_lowtown_docks_n_bdl_0_0.vlm"},
	{0x81FEA970F268581AULL, "l_dae110ar_docks_nt.bnk"},
	{0x81FEAE926B63163BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_basic.mmh"},
	{0x81FEB5E3BDA6BA1FULL, "vxlt_undercitydungeon_1uh_0_0.vlm"},
	{0x81FEBA926B632AD8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_basic.mao"},
	{0x81FECC926B634925ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_basic.msh"},
	{0x81FF2C6B800FCC9BULL, "402171262.ogg"},
	{0x81FF5C8D5F8B3CBBULL, "vxlt_warehouse_14a_0_0.vlm"},
	{0x81FF5CCD69394DDAULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowtwohandedweapon.ani"},
	{0x81FFA0E25913E724ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_hv_dmg.evt"},
	{0x82003D03FD562634ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_action.ani"},
	{0x82005CF05334CB49ULL, "6084416_m.fxe"},
	{0x8200DEB61B37A105ULL, "dae000_mother6136933.cl"},
	{0x8201033EA7336424ULL, "6043737_m.fxe"},
	{0x8201DEE2451E6D64ULL, "6107385_m.fxe"},
	{0x8201F82A93D1B296ULL, "w_encounter2_camp_9_0.rml"},
	{0x82029B7DBF1F076CULL, "761507910.ogg"},
	{0x8202CE07AA6ED8ADULL, "555862495.ogg"},
	{0x8202CF83AD07086BULL, "598084364.ogg"},
	{0x8203C62ACE30323DULL, "llac_w_encounter3_0_0.dds"},
	{0x8203DDEA91D1B800ULL, "155551514.ogg"},
	{0x82040C1C4FA372ABULL, "405346647.ogg"},
	{0x82043A9BB7CF1706ULL, "one280_emrys6104291.cl"},
	{0x8204C7CE2FF019F4ULL, "dae000_sketch6179028.cl"},
	{0x820512B9C32501DEULL, "static_static_p10.psh.fxo.dx11"},
	{0x820674C1BB068706ULL, "6137686_f.fxe"},
	{0x82073A03FD5C3F96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_action.gad"},
	{0x820772CECB4975D5ULL, "6208317_m.fxe"},
	{0x8208083FAFF8C4FDULL, "6139657_m.fxe"},
	{0x820809F12EE092B9ULL, "t3_dog_d09.tnt"},
	{0x820817ECCCCF27C0ULL, "hlac_ht_0_hightown_a3_65534_1x.dds"},
	{0x82083EDA703887B9ULL, "227137861.ogg"},
	{0x82088FAA750C96F6ULL, "gen00fl_varric6033549.cl"},
	{0x8208D4A88131FB3CULL, "vxlt_warehouse_n_sc_0_0.vlm"},
	{0x8208F12A93D7C5BAULL, "w_encounter2_camp_9_0.phy"},
	{0x82095AF2BD2C8091ULL, "and231ip_at_darktown.utp"},
	{0x8209D17C83CC8F74ULL, "vxlt_house_ge_0_0.vlm"},
	{0x8209D9CE953D0062ULL, "vxlt_warehouse_188_0_0.vlm"},
	{0x820A380E72F2BB13ULL, "130680799.ogg"},
	{0x820A5035C3108441ULL, "tools_emitter.mmh"},
	{0x820AE7D45A650C43ULL, "vxht_playershouse_n_1pe_0_0.vlm"},
	{0x820B0F7B642983BEULL, "lt_gallowstemplar_2387.rml"},
	{0x820B2464437917A4ULL, "lt_0_lowtown_docks_a3_140_0.mmh"},
	{0x820B286443791EF7ULL, "lt_0_lowtown_docks_a3_140_0.mao"},
	{0x820B2F81DD48E5FFULL, "hlac_ht_bartrand_n_3374_1z.dds"},
	{0x820B3A6443793D46ULL, "lt_0_lowtown_docks_a3_140_0.msh"},
	{0x820BDC0CDF017E6EULL, "6092077_f.fxe"},
	{0x820C1326C3FD46B8ULL, "vxw_encounter2_camp_b6_0_0.vlm"},
	{0x820C402810B71906ULL, "vxw_encounter2_camp_bx_0_0.vlm"},
	{0x820C7818CF5E5831ULL, "vxlt_0_lowtown_docks_n_bac_0_0.vlm"},
	{0x820CB8E2591E9A1DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_hv_dmg.ani"},
	{0x820DA88772420B1BULL, "mer341cr_earl.utc"},
	{0x820E5503FD628CF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_action.evt"},
	{0x820E6A33447049AAULL, "6001197_m.fxe"},
	{0x820E6AD4B12D7FCCULL, "character_character_p225.psh.fxo.dx11"},
	{0x820E8AA51F0D5D88ULL, "art\\characters\\weapons\\axes\\w_axe_ax104a.msh"},
	{0x820EA8A51F0D90B2ULL, "art\\characters\\weapons\\axes\\w_axe_ax104a.mmh"},
	{0x820F3EC95780AD5EULL, "hlac_lt_gallowstemplar_2382_1x.dds"},
	{0x820F5C8C3C53E924ULL, "dae000_mother6121026.cl"},
	{0x820F62448C2BE628ULL, "318864675.ogg"},
	{0x820FA49826FC32F1ULL, "vxdr_0_deeproads_fij_0_0.vlm"},
	{0x820FBCA05CCA1584ULL, "621558058.ogg"},
	{0x82105012189A9819ULL, "dae000_light_tombs_boss6179463.cl"},
	{0x8210D1E5951CA0B2ULL, "zz_vault_debug6177769.cl"},
	{0x8210DAE944D41413ULL, "21790428.ogg"},
	{0x8211202C3C5205CFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_coma_0s.dds"},
	{0x821125F4D3C5CB16ULL, "vxlt_undercitydungeon_1fp_0_0.vlm"},
	{0x8211824E0BF20655ULL, "971767947.ogg"},
	{0x8211B0AB77806244ULL, "6103217_m.fxe"},
	{0x8212210B219AE5B7ULL, "vxw_encounter1_hz_0_0.vlm"},
	{0x821264F366F110DBULL, "6095009_m.fxe"},
	{0x82126DCE7E687CBFULL, "mag291_torpor6021688.cl"},
	{0x82127199D710364FULL, "hlac_ht_keep_3732_9x.dds"},
	{0x8212B40C62917075ULL, "hlac_lt_0_lowtown_65533_3y.dds"},
	{0x8212D66853FDD967ULL, "mer220_hubert6049096.cl"},
	{0x82130AA95EAB767BULL, "6147562_m.fxe"},
	{0x82138CF99995DAC1ULL, "lgt120_orz_bln_official6136237.cl"},
	{0x8213A2E614F1FB5EULL, "follower_banter6155679.cl"},
	{0x8214AFBE2F8206D6ULL, "869172384.ogg"},
	{0x8215C10DF904074CULL, "gen_im_arm_cht_med_champion.uti"},
	{0x8215F7A75258A959ULL, "hlac_lt_0_lowtown_a2_-131072_5x.dds"},
	{0x821615FA6D907945ULL, "vxlt_0_lowtown_a2_n_cmo_0_0.vlm"},
	{0x8216907F92AEB866ULL, "hlac_ht_0_hightown_a2_-131072_0z.dds"},
	{0x8216AD0E8FF9BE46ULL, "6082245_f.fxe"},
	{0x82176072B954FC1FULL, "6037690_m.fxe"},
	{0x8217B5C29259AE25ULL, "vxlt_darktown_5lb_0_0.vlm"},
	{0x8217CF80E2853E67ULL, "mag100_cullen6082298.cl"},
	{0x8218E5F755410665ULL, "6182018_f.fxe"},
	{0x821959B14C5B24C3ULL, "vxht_playershouse_n_1wd_0_0.vlm"},
	{0x821966D9FB4D663AULL, "813649289.ogg"},
	{0x8219DECF710CDBB4ULL, "6170831_m.fxe"},
	{0x821A4322CF6DC15EULL, "179096185.ogg"},
	{0x821AD3212A4AEF4CULL, "bl_ar_sh.mao"},
	{0x821B291F6177FF6FULL, "vxlt_gallowscourtyard_7as_0_0.vlm"},
	{0x821C4725ACE7A34BULL, "lgt220ip_shp_mag_bofort.utp"},
	{0x821DB403B7450A38ULL, "112501202.ogg"},
	{0x821E250DD344CD93ULL, "vxlt_gallowstemplar_n_4so_0_0.vlm"},
	{0x821E3A5464A4330FULL, "dae000ip_ambient_sitdrunk_s.utp"},
	{0x821E66921B9DECD8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolema_overlay.anb"},
	{0x821E6E8FBABABF11ULL, "6182831_m.fxe"},
	{0x821E77D7C6EDF1B4ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medg.phy"},
	{0x821F06D9A4935262ULL, "llac_ht_0_hightown_a3_131072_0.dds"},
	{0x821F3873E9A10F65ULL, "hlac_lt_0_lowtown_docks_a3_-2_1x.dds"},
	{0x821F97D1C5BBE396ULL, "vxlt_undercitydungeon_1iq_0_0.vlm"},
	{0x821F9DF047F5ABC5ULL, "6127012_m.fxe"},
	{0x821FD1A58745809BULL, "6019103_f.fxe"},
	{0x8220228869767AC4ULL, "6104317_m.fxe"},
	{0x822141A45AC063CEULL, "6175975_m.fxe"},
	{0x82214776120CE3A8ULL, "6005313_m.fxe"},
	{0x8221C3B8E731C975ULL, "vxca_caverns1_mines_i2b_0_0.vlm"},
	{0x8221D12821BDE7AEULL, "plt_mag320pt_otl_huon.plo"},
	{0x8221D1F2A4AB7AB1ULL, "character_character_p132.psh.fxo.dx11"},
	{0x8222A8E16A46B926ULL, "plt_lgt220pt_shp_kill_qun.plo"},
	{0x8222C133F8E55E77ULL, "vxlt_warehouse_11z_0_0.vlm"},
	{0x82234E9E0DF0AB96ULL, "w_encounter2_58_0.rml"},
	{0x8223C370F287A0B8ULL, "l_dae110ar_docks_nt.wbd"},
	{0x822438ED2EA32BE9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edge.phy"},
	{0x82244A42988DA3D6ULL, "253272300.ogg"},
	{0x8224CF9A97245127ULL, "6139277_m.fxe"},
	{0x82250528248A6EFFULL, "518670040.ogg"},
	{0x82250D18AC21B6A0ULL, "t3_wdg_bw1.tnt"},
	{0x8225576DEEC9CD07ULL, "ave150tr_enter_ambush_area.ncs"},
	{0x8225D8BD6460F40FULL, "dae000_tranquil_16165116.cl"},
	{0x8225D8C42FC0DDC3ULL, "675026478.ogg"},
	{0x8226279F844FC6A5ULL, "vxlt_darktown_5qa_0_0.vlm"},
	{0x82264840713ED96EULL, "hlac_dr_primevaltomb_659_13z.dds"},
	{0x822662CD955489D5ULL, "vxlt_blackemporium_12q_0_0.vlm"},
	{0x8226697B2CEC60EEULL, "6202107_m.fxe"},
	{0x8226DED22FD47767ULL, "6089857_m.fxe"},
	{0x82279E653DC63651ULL, "6164721_m.fxe"},
	{0x8227A982D808BEA0ULL, "hlac_dr_primevalentrance_1725_3z.dds"},
	{0x8227FE6FB86FCD7AULL, "llac_lt_0_lowtown_a3_n_c_-65536_5.dds"},
	{0x822852EE1E024F48ULL, "31711382.ogg"},
	{0x822854D7C6F60AB1ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medg.msh"},
	{0x82286ED7C6F6376FULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medg.mmh"},
	{0x82290F374A01C7B3ULL, "vxht_0_hightown_a3_c1o_0_0.vlm"},
	{0x82292B8E3AF860FCULL, "313582648.ogg"},
	{0x8229854BCD592961ULL, "747048957.ogg"},
	{0x8229A5DECB8FAFC5ULL, "hlac_ca_cavern1_haunted_24426_14x.dds"},
	{0x8229B5AF423E6D50ULL, "vxlt_gallowstemplar_378_0_0.vlm"},
	{0x822A479E0DF6BEBAULL, "w_encounter2_58_0.phy"},
	{0x822A5DF8952D5476ULL, "vxlt_0_lowtown_docks_a3_n_d28_0_0.vlm"},
	{0x822A639515A4420AULL, "vxlt_blackemporium_vg_0_0.vlm"},
	{0x822A720B49243BB5ULL, "663917034.ogg"},
	{0x822ABB5285D5E516ULL, "480214632.ogg"},
	{0x822AD55FE4CA8F78ULL, "t1_mue_ye1.tnt"},
	{0x822AEFAD73C81D1BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_back_look_up.ani"},
	{0x822B00DBF2CAD7C0ULL, "vxlt_gallowsprison_n_6pk_0_0.vlm"},
	{0x822B24C45418A5EBULL, "340678661.ogg"},
	{0x822BBC17DC2CEF82ULL, "6175526_m.fxe"},
	{0x822C1CFE2F47C16DULL, "character_character_p107.psh.fxo.dx11"},
	{0x822CBD59BD6A8CC1ULL, "462630376.ogg"},
	{0x822CCC499375D732ULL, "dae000cr_grand_cleric6147665.cl"},
	{0x822CD9168068C3F5ULL, "vxht_mansion2b_3bq_0_0.vlm"},
	{0x822CE5337DFD9C47ULL, "zz_ave_debug6066167.cl"},
	{0x822D1088FBA9E43BULL, "6072125_m.fxe"},
	{0x822D609F9775BB51ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_imp_c.impa.ani"},
	{0x822D838163A6024EULL, "cod_ltr_pro_athenril.ncs"},
	{0x822D9272CA314A1AULL, "vxlt_gallowstemplar_f_520_0_0.vlm"},
	{0x822E09AEB7B1FC16ULL, "vxlt_undercitydungeon_1lr_0_0.vlm"},
	{0x822E10BB78C0D1F9ULL, "6103209_m.fxe"},
	{0x822E14FAFCE579CFULL, "162905604.ogg"},
	{0x822E70A8DBE644E7ULL, "vxw_bonepitmining_a2_er_0_0.vlm"},
	{0x822EBEB70FBCAFA2ULL, "hlac_w_woundedcoast_-196604_0x.dds"},
	{0x822F5D68E90B7F36ULL, "43981151.ogg"},
	{0x82303E53F328DF0BULL, "751310631.ogg"},
	{0x823086C84EF345A0ULL, "vxht_chantry_2f1_0_0.vlm"},
	{0x82308A02765EFFC4ULL, "6154789_m.fxe"},
	{0x8230D6B4B945BE67ULL, "6110597_m.fxe"},
	{0x8230DFFD4397A5E4ULL, "hlac_ca_cavern1_haunted_24424_13z.dds"},
	{0x82313310EADB76F7ULL, "vxlt_warehouse_14u_0_0.vlm"},
	{0x82314FFABE74C8E8ULL, "w_sundermount_79_0.rml"},
	{0x823169A102F61533ULL, "298749794.ogg"},
	{0x82317E82F997E6EDULL, "rune_weapon_striking.dds"},
	{0x8231834B73BDE060ULL, "755687675.ogg"},
	{0x823186BA3BA2CA4AULL, "81783698.ogg"},
	{0x8231E78FCD037D40ULL, "6188726_m.fxe"},
	{0x8232345484E6F8FFULL, "dae000_bodhan6146838.cl"},
	{0x823291B7CCC4FEF7ULL, "ach_soc_friend.dds"},
	{0x823354D8B18F02D0ULL, "sst_ss_raider_m_1.gda"},
	{0x82336B9F977A39F3ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_imp_c.impa.gad"},
	{0x8233A42DDB534982ULL, "zz_mrl_debug6086514.cl"},
	{0x823454CED2CCDE4DULL, "dae011_corff_sta6142439.cl"},
	{0x823484B430238B4FULL, "vxht_mansion2_3ct_0_0.vlm"},
	{0x82348CA378C11FD4ULL, "lgt110_arl_bann_teagan6115944.cl"},
	{0x8234E96F0C0D3DA3ULL, "hlac_lt_0_lowtown_-196607_0x.dds"},
	{0x82357AF86BD1DE8EULL, "ter_sand01_d.dds"},
	{0x823586110C72AC0AULL, "713191934.ogg"},
	{0x8235D27B6DD88E40ULL, "vxht_mansion_n_4hj_0_0.vlm"},
	{0x8235EDA3EC50CBD5ULL, "vxw_encounter3_3o_0_0.vlm"},
	{0x823626F6F2D5F73BULL, "773245731.ogg"},
	{0x82364C81B99FB04FULL, "vxlt_gallowscourtyard_n_7ex_0_0.vlm"},
	{0x8236C22821D02A3DULL, "plt_mag320pt_otl_huon.nss"},
	{0x8237080A6D02D2B8ULL, "dae200_bran6092476.cl"},
	{0x823744BB03929728ULL, "6163251_m.fxe"},
	{0x82374D6C094B465CULL, "lgt210_eau_symp_templar6173993.cl"},
	{0x8237F7AD73D2B512ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_back_look_up.evt"},
	{0x8237F9F69829D71CULL, "121118299.ogg"},
	{0x82380EA1ECC8A834ULL, "gen00fl_fenris6126599.cl"},
	{0x823847DCB02EE694ULL, "narrative_varric6172534.cl"},
	{0x82387ADD4F7FA376ULL, "6092575_m.fxe"},
	{0x8238E23E6356DC25ULL, "6177945_m.fxe"},
	{0x8239087F8BDB6432ULL, "zz_followers_banter6159387.cl"},
	{0x82392D25A355B61BULL, "one300_zevran6076925.cl"},
	{0x8239AD5560D08FADULL, "528109147.ogg"},
	{0x823A689F97805358ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_imp_c.impa.evt"},
	{0x823B11548A1819F1ULL, "6118205_m.fxe"},
	{0x823B53183B567452ULL, "zz_fen_debug6134354.cl"},
	{0x823BA49DDED222CAULL, "21941137.ogg"},
	{0x823BB6B9D396F4EAULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_0y.dds"},
	{0x823CA3A2613285C5ULL, "hf_bethany_h1s1.mor"},
	{0x823CD5191E3E1492ULL, "dae000_junar6147843.cl"},
	{0x823D10CBF5E2E1B4ULL, "6154976_m.fxe"},
	{0x823D234F8BC38B01ULL, "mcr310_argument6072303.cl"},
	{0x823D64DB6D7446C0ULL, "496046087.ogg"},
	{0x823E022C135EEA4EULL, "lt_0_lowtown_docks_167_0.mmh"},
	{0x823E0E2C135EFEE5ULL, "lt_0_lowtown_docks_167_0.mao"},
	{0x823E1C2C135F163CULL, "lt_0_lowtown_docks_167_0.msh"},
	{0x823EEC0ABC25926CULL, "6031124_f.fxe"},
	{0x823F03F902DDE69BULL, "10892104.ogg"},
	{0x823F16AD73D908A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_back_look_up.gad"},
	{0x823F189BA4E0AC9FULL, "dae000ip_craft_potions2.utp"},
	{0x823F36853400DADDULL, "458786405.ogg"},
	{0x823FA1D74851F90AULL, "vxlt_0_lowtown_docks_n_b4j_0_1.vlm"},
	{0x82400FEFD0829485ULL, "vxlt_0_lowtown_docks_b10_0_0.vlm"},
	{0x82405ECAE3988970ULL, "ser221_gamlen6123411.cl"},
	{0x824163E4F51F5A84ULL, "art\\levels\\races\\design\\_textures\\prp_chestb_sa.dds"},
	{0x8241B10F880465E3ULL, "sib131_slave_a6165279.cl"},
	{0x8241B83938317630ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_006a_0n.dds"},
	{0x824231087C8B6852ULL, "llac_dr_0_deeproads_17038_7.dds"},
	{0x8242A04A23E13CF5ULL, "vxlt_gallowstemplar_f_37n_0_0.vlm"},
	{0x8242BC0B329970C5ULL, "428158140.ogg"},
	{0x82430369601A9F44ULL, "461469176.ogg"},
	{0x8243202CB460D2B7ULL, "715666909.ogg"},
	{0x8243432C1FE2DF1FULL, "zz_lgt_2_debug6174314.cl"},
	{0x82435F94B2A8FE7FULL, "148843959.ogg"},
	{0x824374915132A3E0ULL, "hlac_lt_0_lowtown_n_65533_2z.dds"},
	{0x8243886F21187EBDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hanginglight_da.dds"},
	{0x824410A9C6CFBFECULL, "fen110_party_comment6095240.cl"},
	{0x824425EAB8728367ULL, "6140574_m.fxe"},
	{0x82450AFFDA04FC44ULL, "995995929.ogg"},
	{0x82454FCD359DF0A9ULL, "zz_followers_banter6158583.cl"},
	{0x82457765F20A4C7EULL, "6083339_m.fxe"},
	{0x8245A85056CB1939ULL, "dae000ip_crowd_ff1_talk16.utp"},
	{0x8245CF926B9F5170ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_basic.phy"},
	{0x82467482386C97BCULL, "dre000_hireling16038594.cl"},
	{0x824692B329DA7ADEULL, "433294241.ogg"},
	{0x8247E3CA377B2366ULL, "vxlt_0_lowtown_a2_n_by1_0_0.vlm"},
	{0x8247F4FB04461066ULL, "hlac_dr_0_deeproads_17038_2y.dds"},
	{0x82480FB7142D4661ULL, "vxw_woundedcoast_10q_0_0.vlm"},
	{0x82483BA9134FD44CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v2.ani"},
	{0x824854AC16DA622DULL, "mag110_arianni6016776.cl"},
	{0x824899D3194E3ECFULL, "isa211_isabela6072601.cl"},
	{0x8248B7BDCC0C750CULL, "vrc221_hugin6042684.cl"},
	{0x82490404A94F5D87ULL, "vxht_mansion_n_501_0_1.vlm"},
	{0x824995686D5D4F80ULL, "hlac_ca_caverns1_mines_24424_2x.dds"},
	{0x8249FD56B818A532ULL, "vxw_blightlands_hu_0_0.vlm"},
	{0x824A8032D02312B7ULL, "vxht_mansion_n_5cm_0_0.vlm"},
	{0x824A8B54F94D5313ULL, "zz_isa_debug6133185.cl"},
	{0x824AADC27BC2322DULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_nicerchair.mmh"},
	{0x824AB1C27BC2397EULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_nicerchair.mao"},
	{0x824ABFC27BC250B3ULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_nicerchair.msh"},
	{0x824AED689B9E2D16ULL, "vxlt_undercitydungeon_1zt_0_0.vlm"},
	{0x824B433EA30B30E4ULL, "clf260_harley6173868.cl"},
	{0x824B5462BFD275E7ULL, "vxw_bonepitmining_a2_ol_0_0.vlm"},
	{0x824B64A1B6FE3BC1ULL, "vxht_brothel_n_32q_0_0.vlm"},
	{0x824B79074758325CULL, "111697210.ogg"},
	{0x824CF03161922B37ULL, "683177115.ogg"},
	{0x824CF3D5B690228EULL, "hlac_lt_0_lowtown_a2_131071_6z.dds"},
	{0x824D15E8E37687DCULL, "gen_im_cns_cft_arn_06.uti"},
	{0x824D2441B6149A8AULL, "271358307.ogg"},
	{0x824D2D347DF52ABFULL, "343578678.ogg"},
	{0x824DA05F5788E6E0ULL, "382643165.ogg"},
	{0x824E69384B1B45A2ULL, "6135632_m.fxe"},
	{0x824E8958C991C1CFULL, "117021069.ogg"},
	{0x824E9FDA80553D4DULL, "264519097.ogg"},
	{0x824ECC27E6FEE8CDULL, "591180751.ogg"},
	{0x824F48A91356088EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v2.gad"},
	{0x824FBDF40619DA75ULL, "zz_followers_banter6159654.cl"},
	{0x824FC302CA388A47ULL, "vxw_bonepitmining_a2_1u9_0_0.vlm"},
	{0x824FDAEDA3DF3325ULL, "w_bonepitmining_a2_37_0_ll.mmh"},
	{0x824FEE74CB3C4977ULL, "dae200_saemus6142746.cl"},
	{0x825021ED2EC83F0EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edge.mmh"},
	{0x82502DED2EC853A5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edge.mao"},
	{0x82503BED2EC86BFCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edge.msh"},
	{0x8251003EA908ACEFULL, "gen00fl_isabela6079038.cl"},
	{0x825110A3DF58092BULL, "mcr310_party_banter6087787.cl"},
	{0x8251326544769955ULL, "vxlt_0_lowtown_a2_epa_0_0.vlm"},
	{0x82517A64010B1153ULL, "6117620_m.fxe"},
	{0x8251902649E5B0C9ULL, "6071100_m.fxe"},
	{0x8251986CD581F248ULL, "vxw_woundedcoast_10t_0_0.vlm"},
	{0x8252AB41013281FEULL, "6105442_m.fxe"},
	{0x8252B63551C4BF40ULL, "vxht_mansion_n_4bh_0_0.vlm"},
	{0x825347694BD2DE68ULL, "follower_banter6141720.cl"},
	{0x825363CC3A2D161BULL, "vxht_mansion2b_3ez_0_0.vlm"},
	{0x82536E9FF7FC3F41ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc103a_0n.dds"},
	{0x8253B677EA498B96ULL, "vxlt_backalley_g_1pb_0_1.vlm"},
	{0x8253C9B4F85C5EBCULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.impa.ani"},
	{0x8254DD58706F3975ULL, "vxht_0_hightown_n_9x6_0_0.vlm"},
	{0x825543A9135A6C6DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v2.evt"},
	{0x8255754AC9B311D5ULL, "vxht_mansion_n_638_0_0.vlm"},
	{0x8256AAC04B12BDFEULL, "223095894.ogg"},
	{0x82579AAD77A47DB6ULL, "6136956_m.fxe"},
	{0x82579BA5FC616F50ULL, "437495084.ogg"},
	{0x82579BFB88DE32DBULL, "lt_0_lowtown_a2_n_c_5_0.mmh"},
	{0x8257A7FB88DE4778ULL, "lt_0_lowtown_a2_n_c_5_0.mao"},
	{0x8257B9FB88DE65C5ULL, "lt_0_lowtown_a2_n_c_5_0.msh"},
	{0x8257F9CAC459DA96ULL, "vxlt_darktown_35p_0_0.vlm"},
	{0x82586F33AA0EBDB2ULL, "vxw_blightlands_mv_0_0.vlm"},
	{0x825896A71563A772ULL, "vxw_encounter1_dw_0_0.vlm"},
	{0x8258BE57F9FDF3D8ULL, "hlac_w_encounter3_-65539_0z.dds"},
	{0x8259BD2347A6332AULL, "hm_friedrich_thug2.mor"},
	{0x8259BD884BF348F9ULL, "vxlt_0_lowtown_docks_n_b4d_0_0.vlm"},
	{0x8259C63FB1C88E67ULL, "vxw_bonepitmining_a2_jo_0_0.vlm"},
	{0x8259F2A72E9AB14DULL, "6038598_m.fxe"},
	{0x825A2ECF7BAF347BULL, "lt_0_lowtown_docks_a3_n_35_0.rml"},
	{0x825A33FDB669B76CULL, "bw_wav.mao"},
	{0x825A5B7CBBE03C5FULL, "w_woundedcoast_143_0.msh"},
	{0x825A617CBBE04601ULL, "w_woundedcoast_143_0.mmh"},
	{0x825A6D7CBBE05AA2ULL, "w_woundedcoast_143_0.mao"},
	{0x825A7381ACFB94DCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodramp.mao"},
	{0x825A7F81ACFBA97FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodramp.mmh"},
	{0x825A8581ACFBB321ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodramp.msh"},
	{0x825AD6B4F86293FEULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.impa.gad"},
	{0x825B22554A3C7EFDULL, "294923291.ogg"},
	{0x825B3890721C8B1AULL, "388133967.ogg"},
	{0x825B3FE5F3DC6E3AULL, "ch2notnt_s1.psh.fxo"},
	{0x825B4E339BE318B6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mcn.mao"},
	{0x825B533432C6977FULL, "vxht_keep_4hb_0_0.vlm"},
	{0x825B5A339BE32D55ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mcn.mmh"},
	{0x825B5C339BE330CBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mcn.msh"},
	{0x825B77C776B14C2CULL, "vxw_sundermount_2c6_0_0.vlm"},
	{0x825BA6C27BD099EAULL, "art\\levels\\races\\ferelden\\props\\animhack\\prp_nicerchair.phy"},
	{0x825BB05730A07BA3ULL, "hlac_lt_gallowstemplar_n_2388_6y.dds"},
	{0x825C040057BB5BE2ULL, "and100_anders6111656.cl"},
	{0x825C511E3DB912DCULL, "gen00fl_isabela6100061.cl"},
	{0x825C9989C512DE88ULL, "lowtown_fem_prostitute_sta6161604.cl"},
	{0x825CD309431754D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_down.ani"},
	{0x825CD489EA6A7CAAULL, "vxca_caverns2prc_975_0_0.vlm"},
	{0x825D237A552B43ECULL, "vxw_bonepit_1s6_0_0.vlm"},
	{0x825D51F79912518AULL, "971691380.ogg"},
	{0x825DB84C25F2AAC9ULL, "vxht_0_hightown_a3_anb_0_0.vlm"},
	{0x825DDCE3554CE70DULL, "mag311_emile_de_launcet6043619.cl"},
	{0x825E4AAA4D8B4526ULL, "6028527_m.fxe"},
	{0x825ED89761A1BEC1ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_meda_0n.dds"},
	{0x825F632985D4EF44ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masd.msh"},
	{0x825F752985D50DADULL, "art\\characters\\playercharacter\\humanfemale\\glv_masd.mao"},
	{0x825F792985D514E6ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masd.mmh"},
	{0x826090F2FA1310BFULL, "hlac_ht_chantry_n_2534_21x.dds"},
	{0x826135574C820B73ULL, "art\\levels\\races\\design\\prp_bodybaga_2.mao"},
	{0x826137306A1BD277ULL, "llac_ca_caverns1_24424_15.dds"},
	{0x82619531DB678E58ULL, "6030425_m.fxe"},
	{0x8261B1B4F868731DULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.impa.evt"},
	{0x8261F918A2394F33ULL, "llac_lt_0_lowtown_docks_0_5.dds"},
	{0x82620516CC9A94D5ULL, "vxht_keep_n_6yj_0_0.vlm"},
	{0x8262092AEBA37DD8ULL, "pro000_mother_blightlands_2.stg"},
	{0x8262286848AEC52CULL, "544830820.ogg"},
	{0x82625819E0AB1F5EULL, "6109635_m.fxe"},
	{0x82629014757D6756ULL, "store_dae300cr_hubert.utm"},
	{0x8262D97569732623ULL, "122222624.ogg"},
	{0x8262F8F1D1720A33ULL, "ser221_gamlen6135246.cl"},
	{0x8263A054524DEA22ULL, "w_sundermount_151_0.msh"},
	{0x8263A254524DEDD8ULL, "w_sundermount_151_0.mmh"},
	{0x8263AE54524E027BULL, "w_sundermount_151_0.mao"},
	{0x8263C0B31F8C9A86ULL, "mag210_arianni6031663.cl"},
	{0x8264766B0C91DBC4ULL, "art\\characters\\creatures\\shade\\textures\\c_shade_0d.dds"},
	{0x8264A5273BC339B7ULL, "vxlt_hangedman_1pm_0_0.vlm"},
	{0x8264EA2A942618E3ULL, "w_encounter2_camp_9_0.msh"},
	{0x8264EBC457C1269AULL, "mag370_meredith6064859.cl"},
	{0x8264F82A9426303DULL, "w_encounter2_camp_9_0.mmh"},
	{0x8264FC2A9426374EULL, "w_encounter2_camp_9_0.mao"},
	{0x826543B82033D5C2ULL, "6121761_m.fxe"},
	{0x826549A6CAABF3B2ULL, "102701403.ogg"},
	{0x82666F9779B0CA4BULL, "art\\characters\\creatures\\mubari\\animation\\com_play.evt"},
	{0x8266E1109C04D632ULL, "vxw_blightlands_vw_0_0.vlm"},
	{0x826708840759BFF2ULL, "vxw_encounter1_ip_0_0.vlm"},
	{0x82671605AB6F4905ULL, "isa311_castillon6092091.cl"},
	{0x82675E47F6AE449EULL, "6043602_f.fxe"},
	{0x826763ECB40F43B7ULL, "vxht_mansion_n_5io_0_0.vlm"},
	{0x8267BFFEA3719382ULL, "457768658.ogg"},
	{0x8268C527D8DBCCA5ULL, "llac_ht_mansion2b_818_65.dds"},
	{0x8268FAA94BE4CDFBULL, "gen00fl_sebastian6180801.cl"},
	{0x826962D210390A2CULL, "6209851_f.fxe"},
	{0x8269702985DD59B1ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masd.phy"},
	{0x8269C9D9F42E9CF4ULL, "253529853.ogg"},
	{0x8269D64FA06A7B33ULL, "gen00fl_aveline6139784.cl"},
	{0x8269DB094321ECC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_down.evt"},
	{0x8269E9A468A764ACULL, "vxw_sundermount_2d7_0_0.vlm"},
	{0x826A57D7087776A3ULL, "vxht_brothel_n_3v7_0_0.vlm"},
	{0x826A74C10CBABF82ULL, "450016459.ogg"},
	{0x826B34986C967E26ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masb_0n.dds"},
	{0x826B818EC764EF71ULL, "6090813_m.fxe"},
	{0x826B94E85B340FB9ULL, "16363093.ogg"},
	{0x826C1FBC522C7720ULL, "6187705_m.fxe"},
	{0x826C2A2917E8C349ULL, "vxht_0_hightown_a3_aka_0_0.vlm"},
	{0x826C57A6C1A84A33ULL, "353636836.ogg"},
	{0x826CF73B7BE4293DULL, "440836080.ogg"},
	{0x826CFAFC85F3709BULL, "dae000_gamlen6048518.cl"},
	{0x826D32A3AE0A6BB5ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrdb_0d.dds"},
	{0x826D3B79ABBE9F0AULL, "dre000_hireling16038589.cl"},
	{0x826D6AA14360951FULL, "6076037_m.fxe"},
	{0x826D889779B7130CULL, "art\\characters\\creatures\\mubari\\animation\\com_play.gad"},
	{0x826D88D451BF8286ULL, "gen00fl_aveline6139570.cl"},
	{0x826D89E16A867065ULL, "plt_lgt220pt_shp_kill_qun.nss"},
	{0x826DC3B2FCB2F6BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ba_transition.gad"},
	{0x826E4F6B9FC7A28CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2_tool.msh"},
	{0x826E556B9FC7AD7EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2_tool.mmh"},
	{0x826E616B9FC7C115ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2_tool.mao"},
	{0x826EA95452580163ULL, "w_sundermount_151_0.phy"},
	{0x826ED835073C3983ULL, "6037290_m.fxe"},
	{0x826FCE728C5DB626ULL, "art\\characters\\creatures\\mubari\\animation\\dblw_dmg_v1.ani"},
	{0x827011D79AFA046CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext.msh"},
	{0x827017D79AFA0E5EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext.mmh"},
	{0x827023D79AFA22F5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext.mao"},
	{0x82705ABB751E6334ULL, "tbe100_xenon6190995.cl"},
	{0x827085532EE94A83ULL, "336029024.ogg"},
	{0x8270EFFF0E657DEDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobc_0d.dds"},
	{0x8270FC0943284310ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_left_down.gad"},
	{0x8271A9BBF2042AB7ULL, "vxca_caverns1_mines_fmf_0_0.vlm"},
	{0x8271E81A9108D776ULL, "llac_w_encounter1_-65535_0.dds"},
	{0x827254E35F938B6FULL, "hightown_male_elfservnt_mob6170129.cl"},
	{0x82726E8510382669ULL, "298166623.ogg"},
	{0x8272B8266C2A7BFBULL, "w_bonepitdestroyed_a3_205_0.mmh"},
	{0x8272B946E5B698C9ULL, "follower_banter6172927.cl"},
	{0x8272C4266C2A8F98ULL, "w_bonepitdestroyed_a3_205_0.mao"},
	{0x8272D6266C2AAEE5ULL, "w_bonepitdestroyed_a3_205_0.msh"},
	{0x8273879779BB7D0AULL, "art\\characters\\creatures\\mubari\\animation\\com_play.ani"},
	{0x82738BE4EC50D12BULL, "hlac_ca_caverns1_mines_24424_15y.dds"},
	{0x82738C2452B149EBULL, "vxlt_0_lowtown_a3_n_c_ekl_0_2.vlm"},
	{0x8273CCC39AB0E726ULL, "399692019.ogg"},
	{0x82748FCE2F35054BULL, "6061684_m.fxe"},
	{0x8274C6B2FCB91AD1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ba_transition.evt"},
	{0x8275746D99FB800DULL, "53641738.ogg"},
	{0x8275D5C9A6055C37ULL, "vxht_mansion_n_5rn_0_0.vlm"},
	{0x827623DBE5E4DA43ULL, "89681500.ogg"},
	{0x827685A432BB063FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firesml.mmh"},
	{0x82769092F8F36DDDULL, "969414200.ogg"},
	{0x8276AB6C7C987D46ULL, "2807377.ogg"},
	{0x8276EA6562CDE761ULL, "511525224.ogg"},
	{0x827719576D120395ULL, "vxlt_0_lowtown_n_fha_0_0.vlm"},
	{0x82774836FF1614A5ULL, "hlac_lt_warehouse_n_0_8y.dds"},
	{0x827837DF25120A76ULL, "6064862_m.fxe"},
	{0x82785B815A9D7D2CULL, "vxw_sundermount_2i4_0_0.vlm"},
	{0x82786D0FFF362F18ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_arh_0s.dds"},
	{0x827872E2AEBCB8F8ULL, "vxht_0_hightown_n_c0m_0_0.vlm"},
	{0x8278E5990D14B1A2ULL, "hlac_lt_blackemporium_1055_6x.dds"},
	{0x827900ED8DF4F4F1ULL, "seb321_leliana6181831.cl"},
	{0x827927520DEFC215ULL, "dae000ip_crowd_lm5_talk4.utp"},
	{0x827984730EA931E0ULL, "vxht_0_hightown_a2_c8n_0_0.vlm"},
	{0x8279BB5DA6BFA0D6ULL, "vxlt_darktown_615_0_0.vlm"},
	{0x8279F36BB95B07F1ULL, "6095713_m.fxe"},
	{0x8279F60B2B928837ULL, "hlac_ht_mansion_n_3369_1y.dds"},
	{0x827A33CB0D82888CULL, "hlac_lt_0_lowtown_a2_n_131072_2z.dds"},
	{0x827A446DEF11ED5EULL, "ave150tr_enter_ambush_area.utt"},
	{0x827AA79261D2D053ULL, "vxlt_shop_e4_0_0.vlm"},
	{0x827AFA5EDC9CE8BCULL, "cd_qcr220_qunari_hunter.wbd"},
	{0x827B6443F57C9AFDULL, "6064551_m.fxe"},
	{0x827B872BD546DCDDULL, "301175383.ogg"},
	{0x827BE31A450585EDULL, "gen00fl_anders6214958.cl"},
	{0x827CB6728C681797ULL, "art\\characters\\creatures\\mubari\\animation\\dblw_dmg_v1.evt"},
	{0x827CCF1718961B01ULL, "llac_w_sundermount_-393215_0.dds"},
	{0x827D37022C3DA50BULL, "dre000_vampire6034307.cl"},
	{0x827D5478DBFC8EC9ULL, "follower_banter6163459.cl"},
	{0x827E71F9206B9052ULL, "6162782_m.fxe"},
	{0x827EE5F240AF3AE7ULL, "hlac_lt_0_lowtown_a2_n_-1_5z.dds"},
	{0x827F8449466143E8ULL, "406238797.ogg"},
	{0x827F8B4B06918461ULL, "lt_0_lowtown_docks_105_0.phy"},
	{0x827FE53B9EE89579ULL, "zz_qcr_debug6106186.cl"},
	{0x828054B6C310036EULL, "98277304.ogg"},
	{0x8280678AF2411B47ULL, "hlac_ca_cavern1_haunted_24421_6z.dds"},
	{0x8280CE585F0568A5ULL, "lgt110_orz_harr_carta6109051.cl"},
	{0x8281A1B220420BCBULL, "gen00fl_merrill6106088.cl"},
	{0x8281DEB2FCC3CDA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ba_transition.ani"},
	{0x828254EF70E659E9ULL, "565420742.ogg"},
	{0x8282E9FCC0140ABDULL, "dae100_alistair_drunk6135116.cl"},
	{0x8282ED0C7927A916ULL, "772689317.ogg"},
	{0x828356D97BBA1566ULL, "223025798.ogg"},
	{0x82837FA4A33FCF12ULL, "vxlt_0_lowtown_ews_0_0.vlm"},
	{0x8283B192DECE9BECULL, "507615903.ogg"},
	{0x8283B6BE9511A413ULL, "art\\vfx\\textures\\vfx_blood_splatter_n.dds"},
	{0x8283B7728C6E3820ULL, "art\\characters\\creatures\\mubari\\animation\\dblw_dmg_v1.gad"},
	{0x8283BA5BF9F2655CULL, "llac_lt_0_lowtown_docks_65533_0.dds"},
	{0x8283CCE9C8130E26ULL, "6027940_m.fxe"},
	{0x8283EF4C9A84AB7FULL, "611996883.ogg"},
	{0x828421B4C665F206ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wmda_0d.dds"},
	{0x8284284823276ED2ULL, "vxlt_shop_e1_0_0.vlm"},
	{0x8284B1D199B06E0CULL, "art\\characters\\weapons\\greatswords\\w_gsw_011a.phy"},
	{0x8284E48F6DB017A1ULL, "6125555_m.fxe"},
	{0x82850B19F623EF3BULL, "vxlt_0_lowtown_n_eot_0_1.vlm"},
	{0x8285434EA9ABBE55ULL, "vxht_0_hightown_n_b8e_0_0.vlm"},
	{0x82857F971B61F247ULL, "vxht_playershouse_n_1py_0_0.vlm"},
	{0x8285850B9D2F8E48ULL, "229835773.ogg"},
	{0x8285F0E2063E811EULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand_2.mao"},
	{0x828603A3E8E91AFDULL, "hlac_lt_undercityserialkiller_-1_2y.dds"},
	{0x8286409E0E4511E3ULL, "w_encounter2_58_0.msh"},
	{0x82864E9E0E45293DULL, "w_encounter2_58_0.mmh"},
	{0x8286529E0E45304EULL, "w_encounter2_58_0.mao"},
	{0x82865A952D995984ULL, "w_blightlands_173_0.msh"},
	{0x82866C71E29C8BAEULL, "hlac_ht_mansion_n_2995_5x.dds"},
	{0x82866C952D9978EDULL, "w_blightlands_173_0.mao"},
	{0x828670952D997F26ULL, "w_blightlands_173_0.mmh"},
	{0x82867381C18FFB6EULL, "6126385_f.fxe"},
	{0x8286BA17FF3596B9ULL, "6135701_m.fxe"},
	{0x8286C65CCDEADC95ULL, "vxdr_0_deeproads_9go_0_0.vlm"},
	{0x82876D23112EEAD0ULL, "vxlt_gallowstemplar_n_3iu_0_0.vlm"},
	{0x8287A41C4CD7F697ULL, "gen00fl_aveline6137694.cl"},
	{0x8287B73C35AD0E26ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.cess.ani"},
	{0x8287C3EBEE0B33C4ULL, "6047654_m.fxe"},
	{0x8287C9E5EC6CE9CDULL, "vxlt_0_lowtown_a3_n_c_bxn_0_0.vlm"},
	{0x8287DB1E599F663FULL, "vxw_bonepit_1s9_0_0.vlm"},
	{0x82889A9713115609ULL, "vxw_bonepitdestroyed_a3_mr_0_0.vlm"},
	{0x8289452B8BF593D9ULL, "vxlt_backalley_g_1dl_0_0.vlm"},
	{0x828974153A1ED148ULL, "49374131.ogg"},
	{0x82898341FB4AC8FDULL, "hlac_ca_caverns2prc_5802_4y.dds"},
	{0x8289C198C24B61AFULL, "6165635_f.fxe"},
	{0x828A138E91593A64ULL, "163032009.ogg"},
	{0x828A64840CA40F6FULL, "vxw_bonepitdestroyed_a3_1as_0_0.vlm"},
	{0x828A6CED9C1647C3ULL, "gen00fl_anders6183816.cl"},
	{0x828A8AB6FAB05981ULL, "6163478_m.fxe"},
	{0x828B1D4EB04F3FBBULL, "6142392_m.fxe"},
	{0x828B2754E5A34E3DULL, "6110300_m.fxe"},
	{0x828B8233FC8AE3DBULL, "6028529_f.fxe"},
	{0x828B8A23C2A67AE2ULL, "476714791.ogg"},
	{0x828BB7C3F882C157ULL, "isa300pt_wicked.ncs"},
	{0x828BD5FE24E00056ULL, "6110071_m.fxe"},
	{0x828C5591655BB97EULL, "71852010.ogg"},
	{0x828C977A9C4E1851ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v5.gad"},
	{0x828CE3D61261A8D2ULL, "6167482_m.fxe"},
	{0x828DAB9569E9D26FULL, "mer140_bonepit_comments6062368.cl"},
	{0x828E5BD488521328ULL, "plt_lgt160pt_ntb_elves.plo"},
	{0x828F610C2B954F04ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrubble_n.dds"},
	{0x829003A22BE870A9ULL, "lgt250_way_dalish_conflict6147132.cl"},
	{0x829005FB0158C7F5ULL, "vxht_mansion_n_4zp_0_1.vlm"},
	{0x82902E08868EDE17ULL, "6077052_m.fxe"},
	{0x829067952DA1C4F1ULL, "w_blightlands_173_0.phy"},
	{0x829077458B5D8EAEULL, "6134987_m.fxe"},
	{0x8290F9EA9A83B727ULL, "vxlt_0_lowtown_a2_n_c_gnb_0_0.vlm"},
	{0x829106AC9AC4FDB2ULL, "gen00fl_fenris6127051.cl"},
	{0x829130068BCB75E5ULL, "146807928.ogg"},
	{0x82919E78489B6E5EULL, "vxlt_ruinedhouse_n_5e_0_0.vlm"},
	{0x829201CC1F1839F2ULL, "621683875.ogg"},
	{0x82920C99589E0B3BULL, "setss.ncs"},
	{0x82922F3287ABB65EULL, "vxca_cavern1_haunted_e8w_0_0.vlm"},
	{0x8292987A9C5285FFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v5.ani"},
	{0x82929DB4A680582EULL, "901777153.ogg"},
	{0x8292BD183E248606ULL, "arm_kwl.tnt"},
	{0x8292CB6291F34512ULL, "392272106.ogg"},
	{0x8293566C5FA63021ULL, "6122455_m.fxe"},
	{0x8293FD1150FE3495ULL, "vxlt_0_lowtown_n_ffo_0_0.vlm"},
	{0x82948978AE076B7DULL, "hla_ht_0_hightown_a2_65534_0z.dds"},
	{0x82949E99922A3728ULL, "vxlt_gallowstemplar_f_351_0_0.vlm"},
	{0x82949F3C35B76F97ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.cess.evt"},
	{0x8294D0DF1832760FULL, "6156173_m.fxe"},
	{0x8295DF0003250350ULL, "vxlt_gallowstemplar_n_3dr_0_0.vlm"},
	{0x82961300A914D55EULL, "dae000_flemeth6000798.cl"},
	{0x829615D1BE08E225ULL, "mcr370_meredith6136154.cl"},
	{0x82964099AE45FABDULL, "6149412_m.fxe"},
	{0x829668FC80BE8275ULL, "mag100_thrask6031746.cl"},
	{0x82969EF32BD80725ULL, "vxca_caverns1_j26_0_2.vlm"},
	{0x8297363488AD2CAEULL, "5121247.ogg"},
	{0x82987916CAB78A2FULL, "6172962_m.fxe"},
	{0x82993820975E0893ULL, "853029172.ogg"},
	{0x8299ADC7CB741FDBULL, "34491341.ogg"},
	{0x829A40467C0F3476ULL, "ev_wispy_smoke.mao"},
	{0x829A48DE6628FEFFULL, "gen00fl_anders6125386.cl"},
	{0x829AA4EAB1F9D14DULL, "qun000pt_dialogs.ncs"},
	{0x829B053FE300AB18ULL, "6116329_m.fxe"},
	{0x829B367D2F5DB72BULL, "188914080.ogg"},
	{0x829B8F626DAD818DULL, "crt_living_armor.wbd"},
	{0x829BA03C35BD9020ULL, "art\\vfx\\combat\\vfx_glb_arw_trail_c.cess.gad"},
	{0x829BCC8A9472F6D5ULL, "vxht_keep_n_6ev_0_0.vlm"},
	{0x829C28B561A3BCFEULL, "gen00fl_fenris6112033.cl"},
	{0x829C9B940A7DA32BULL, "6104578_m.fxe"},
	{0x829CAA3FAE314F95ULL, "hlac_ht_0_hightown_-1_2z.dds"},
	{0x829D2B3540997F84ULL, "54376328.ogg"},
	{0x829E4936DC5295A6ULL, "873278412.ogg"},
	{0x829E5D0593170919ULL, "6101382_m.fxe"},
	{0x829E6C9B039B9BB7ULL, "vxlt_hangedman_1li_0_0.vlm"},
	{0x829EB0E2A8205B3EULL, "vxlt_0_lowtown_n_bym_0_0.vlm"},
	{0x829F01FAA8ED996CULL, "6065830_m.fxe"},
	{0x829F0352CEC42EEBULL, "hlac_w_woundedcoast_-262142_0x.dds"},
	{0x829F86A432DE0D7CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firesml.anb"},
	{0x829F927345EA9AE1ULL, "195455655.ogg"},
	{0x829FBCFB891C2010ULL, "lt_0_lowtown_a2_n_c_5_0.phy"},
	{0x829FFFEF036B34E5ULL, "hlac_lt_0_lowtown_a2_n_c_65536_5y.dds"},
	{0x82A08C88BDF2FEA9ULL, "follower_barks6153657.cl"},
	{0x82A0F413F214E030ULL, "133304507.ogg"},
	{0x82A11391C0C10475ULL, "llac_lt_0_lowtown_131072_0.dds"},
	{0x82A1998710F48F2EULL, "vxlt_0_lowtown_docks_b6j_0_0.vlm"},
	{0x82A1F7F62AB74C71ULL, "406950298.ogg"},
	{0x82A21FBFDC14C6B4ULL, "1038551218.ogg"},
	{0x82A2404B38137AD5ULL, "mag320_orsino6071066.cl"},
	{0x82A24521A996357CULL, "vxdr_primevaltomb_1a8_0_2.vlm"},
	{0x82A2EF4A18B1E8ABULL, "hlac_lt_darktown_-1_45x.dds"},
	{0x82A3038F853EDDF2ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.impa.gad"},
	{0x82A310EA6F6F9EA0ULL, "230252507.ogg"},
	{0x82A3C04F9ADB4219ULL, "vxht_0_hightown_a3_cdf_0_0.vlm"},
	{0x82A3DE76F7E3677EULL, "6152158_m.fxe"},
	{0x82A3FEAFD7A5A41BULL, "hla_ht_0_hightown_131070_0y.dds"},
	{0x82A4ADFB2425BB82ULL, "vxlt_hangedman_25p_0_0.vlm"},
	{0x82A4EF0AED5F56F3ULL, "llac_lt_gallowscourtyard_n_65539_0.dds"},
	{0x82A55B86CE268F1FULL, "vxdr_0_deeproads_erw_0_1.vlm"},
	{0x82A55EB18C88ACAFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masd_0s.dds"},
	{0x82A5C9011710569EULL, "hlac_lt_0_lowtown_-4_0z.dds"},
	{0x82A6297033081852ULL, "vxlt_gallowscourtyard_774_0_0.vlm"},
	{0x82A65CBD340761E8ULL, "vxlt_0_lowtown_n_exs_0_0.vlm"},
	{0x82A670B6FEF5BBECULL, "6105487_f.fxe"},
	{0x82A69BFB89220768ULL, "lt_0_lowtown_a2_n_c_5_0.rml"},
	{0x82A6DBA7DEB833A8ULL, "vxht_0_hightown_n_9zi_0_0.vlm"},
	{0x82A6F7579A1C454AULL, "dae000_porfiria6162646.cl"},
	{0x82A6F85EDCC22036ULL, "cd_qcr220_qunari_hunter.bnk"},
	{0x82A7CEB3D702944CULL, "6172662_m.fxe"},
	{0x82A87CB5C2A52685ULL, "vxdr_primevaltomb_2uq_0_1.vlm"},
	{0x82A8EA99AE20EC57ULL, "293019982.ogg"},
	{0x82A962D8FF680D91ULL, "6116373_m.fxe"},
	{0x82A9B99497806E54ULL, "vxlt_hangedman_54_0_0.vlm"},
	{0x82A9FADAC88248CFULL, "vxw_encounter3_dt_0_0.vlm"},
	{0x82AA02CD3CCB5E5CULL, "mag320pt_otl_huon.ncs"},
	{0x82AA068F85450129ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.impa.evt"},
	{0x82AA40EDB5D4C606ULL, "854387049.ogg"},
	{0x82AAD3E2AFAEE64FULL, "vxca_caverns1_ik2_0_0.vlm"},
	{0x82ABF05521437CE0ULL, "hlac_ht_mansion2_818_46x.dds"},
	{0x82AC3CA4014DAD1BULL, "ht_keep_n_4519.rml"},
	{0x82AD385F52E96765ULL, "262124086.ogg"},
	{0x82AD3C97453A501AULL, "plt_cod_hst_tev_slavery.plo"},
	{0x82ADB7710F65B1E1ULL, "mag320_orsino6071101.cl"},
	{0x82AE1D84139D50FEULL, "6088068_m.fxe"},
	{0x82AE631DE588EF92ULL, "711729537.ogg"},
	{0x82AEF663A996B746ULL, "6166500_m.fxe"},
	{0x82AF13BE50DF2CDDULL, "vxht_0_hightown_a2_7lo_0_0.vlm"},
	{0x82AF2774CD50CDFDULL, "mer120_hubert6020715.cl"},
	{0x82AF2C6B9FFEFE19ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_1x2_tool.phy"},
	{0x82AF673D629A7636ULL, "6038984_m.fxe"},
	{0x82AF75030DCCF040ULL, "hlac_lt_darktown_-1_13x.dds"},
	{0x82AFA057D50F900CULL, "843037171.ogg"},
	{0x82AFA5A81F0B81E4ULL, "6202022_m.fxe"},
	{0x82AFEED79B2FACF9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext.phy"},
	{0x82B058D4886EF4E7ULL, "plt_lgt160pt_ntb_elves.nss"},
	{0x82B07E239BA3C065ULL, "zz_blightlandvistas.cut"},
	{0x82B0A5B9B9A51BEAULL, "313647802.ogg"},
	{0x82B0D0FDD5F72B12ULL, "qun131tr_party_ambient2.ncs"},
	{0x82B0EDE0056ECF34ULL, "847939471.ogg"},
	{0x82B184F837D7B735ULL, "ave200_donnic6112126.cl"},
	{0x82B2B9848CBC809AULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_cwk.gad"},
	{0x82B2E38FFEC1AC4CULL, "6091694_f.fxe"},
	{0x82B31FD8161BD402ULL, "vxlt_hangedman_22q_0_0.vlm"},
	{0x82B35C8276AA704FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls104a_0t.dds"},
	{0x82B3755DF684BBD1ULL, "hlac_w_bonepit_327681_0z.dds"},
	{0x82B39759E69A055EULL, "166668955.ogg"},
	{0x82B3B6F9598FE74CULL, "one220_yevhen6098719.cl"},
	{0x82B3D6452E711424ULL, "mrl150_merrill6071843.cl"},
	{0x82B467148BE76E5BULL, "6125456_m.fxe"},
	{0x82B49B4D24FE30D2ULL, "vxlt_gallowscourtyard_785_0_0.vlm"},
	{0x82B4CE9A25FD7A68ULL, "vxlt_0_lowtown_n_eup_0_0.vlm"},
	{0x82B504113C9B6BD0ULL, "vxdr_0_deeproads_crq_0_2.vlm"},
	{0x82B57F2B51C2C35EULL, "vxlt_0_lowtown_a3_n_c_hr5_0_0.vlm"},
	{0x82B5C10CAB3178ADULL, "hlac_w_bonepitdestroyed_a3_0_9z.dds"},
	{0x82B6787CAF240291ULL, "hlac_ca_cavern1_haunted_24421_11x.dds"},
	{0x82B6CDBF2EB1FDCFULL, "vxlt_0_lowtown_a2_e8k_0_0.vlm"},
	{0x82B7CEFBE1F365A0ULL, "plt_cod_bks_demons_gift.plo"},
	{0x82B7D5B84600744DULL, "763722257.ogg"},
	{0x82B7EE8F85511598ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.impa.ani"},
	{0x82B808801491A572ULL, "vxlt_0_lowtown_a2_f5b_0_0.vlm"},
	{0x82B83B324ABDF007ULL, "33928752.ogg"},
	{0x82B85DBC1B7B70F9ULL, "168764867.ogg"},
	{0x82B86A54871CA0A8ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgte_0s.dds"},
	{0x82B86CB7BA78614FULL, "vxw_encounter3_aw_0_0.vlm"},
	{0x82B89BBC4648F34DULL, "follower_banter6173186.cl"},
	{0x82B8BD74431658A0ULL, "gen00fl_aveline6101979.cl"},
	{0x82B902CF5C36A8B6ULL, "hlac_lt_0_lowtown_docks_n_-3_7x.dds"},
	{0x82B935E5A9353954ULL, "art\\characters\\creatures\\shade\\textures\\c_shade_0t.dds"},
	{0x82B9B055653038DDULL, "717425313.ogg"},
	{0x82B9B845D01497B6ULL, "683475573.ogg"},
	{0x82B9BC848CC2A4B1ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_cwk.evt"},
	{0x82B9CAB3B4CB3C7DULL, "571978273.ogg"},
	{0x82B9D9266C66B630ULL, "w_bonepitdestroyed_a3_205_0.phy"},
	{0x82B9ED84B3C8B7D6ULL, "gen00fl_bethany6102906.cl"},
	{0x82B9FC1E64C67F34ULL, "6156260_m.fxe"},
	{0x82BB4CBFEE857C13ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masb.phy"},
	{0x82BC6FB6EA8A60EAULL, "gen00fl_anders6125925.cl"},
	{0x82BCF2BFAA45447EULL, "6144467_m.fxe"},
	{0x82BD258BC3C9AFCDULL, "441818870.ogg"},
	{0x82BD4357F86EA4B5ULL, "hm_corff.mor"},
	{0x82BD84C95D8DFFDEULL, "vxca_cavern1_haunted_e7j_0_0.vlm"},
	{0x82BD8881AD5035C4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodramp.phy"},
	{0x82BEB5B35F34098DULL, "hlac_ht_mansion_n_3369_0x.dds"},
	{0x82BEE1540F75F321ULL, "gen_im_trs_rdr_med_boo_r1.uti"},
	{0x82BF69623E9C496EULL, "hlac_lt_backalley_n_0_8y.dds"},
	{0x82BFB66F32AD5697ULL, "6161522_m.fxe"},
	{0x82C022E1F13B9979ULL, "hlac_lt_gallowscourtyard_0_1x.dds"},
	{0x82C078153814C1F2ULL, "vxw_bonepitdestroyed_a3_21a_0_0.vlm"},
	{0x82C082F8D0880019ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage1.gad"},
	{0x82C24B5C66EE4ECCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge03_n.dds"},
	{0x82C26C339493209FULL, "qun230cr_guild_contact6068695.cl"},
	{0x82C2838095C42EFEULL, "gen00fl_merrill6124047.cl"},
	{0x82C2B46AAACA8E7CULL, "hlac_lt_0_lowtown_docks_a3_0_3z.dds"},
	{0x82C30526153BA5C9ULL, "6077691_f.fxe"},
	{0x82C38B60E557FCDAULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb4.mmh"},
	{0x82C39D60E5581B20ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb4.msh"},
	{0x82C44F0F6C1213D4ULL, "mcr370_meredith6083103.cl"},
	{0x82C554011BF644A2ULL, "hlac_lt_0_lowtown_docks_-65537_3x.dds"},
	{0x82C57DE18BDDA371ULL, "hlac_ht_mansion2_818_68y.dds"},
	{0x82C588D199E7BFD7ULL, "art\\characters\\weapons\\greatswords\\w_gsw_011a.mmh"},
	{0x82C58CD199E7C6E4ULL, "art\\characters\\weapons\\greatswords\\w_gsw_011a.mao"},
	{0x82C59ED199E7E549ULL, "art\\characters\\weapons\\greatswords\\w_gsw_011a.msh"},
	{0x82C5AC17977A5363ULL, "203922940.ogg"},
	{0x82C5EA4152DE5D24ULL, "6140605_m.fxe"},
	{0x82C62572958FACC1ULL, "6152131_m.fxe"},
	{0x82C70CEAE07B3D0FULL, "6057371_m.fxe"},
	{0x82C72717762BA056ULL, "w_sundermount_42_0.rml"},
	{0x82C75318E69917EEULL, "ser221_quentin6129877.cl"},
	{0x82C793F8D08E3C67ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_mage1.ani"},
	{0x82C7D4848CCF0A80ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90l_cwk.ani"},
	{0x82C8349C83F67B0BULL, "6038589_m.fxe"},
	{0x82C847FCAD9D8FEEULL, "565297433.ogg"},
	{0x82C84F4B13B021B3ULL, "mag291_demon_marethari6056843.cl"},
	{0x82C8DD92B49A3E39ULL, "qun221_viscount6069755.cl"},
	{0x82C901328734E9FEULL, "307523675.ogg"},
	{0x82C96ECB7A0E8483ULL, "267650436.ogg"},
	{0x82CA452482404341ULL, "6010217.ogg"},
	{0x82CB233FBB54E156ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masd.msh"},
	{0x82CB2D3FBB54F254ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masd.mmh"},
	{0x82CB313FBB54F967ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masd.mao"},
	{0x82CB5B82CC2EAA07ULL, "325994993.ogg"},
	{0x82CBF63ED2DA836CULL, "lgt320_mis_resp_swordhunter6150265.cl"},
	{0x82CC794C2F19B7C3ULL, "vxdr_0_deeproads_ewm_0_0.vlm"},
	{0x82CD08E6718A04D8ULL, "36604726.ogg"},
	{0x82CD209D1F2FA892ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.gad"},
	{0x82CD49D67DC9EA5BULL, "929935046.ogg"},
	{0x82CD732563A8A9E7ULL, "mag101_alain6032261.cl"},
	{0x82CD870F254931CDULL, "hlac_w_bonepit_458752_0x.dds"},
	{0x82CE0A98A5532AFFULL, "art\\characters\\playercharacter\\humanfemale\\glv_wrba.msh"},
	{0x82CE0D0555214EB8ULL, "mag170_thrask6020777.cl"},
	{0x82CE1098A55334A1ULL, "art\\characters\\playercharacter\\humanfemale\\glv_wrba.mmh"},
	{0x82CE1C98A5534942ULL, "art\\characters\\playercharacter\\humanfemale\\glv_wrba.mao"},
	{0x82CE20177631B37AULL, "w_sundermount_42_0.phy"},
	{0x82CEA00BE162097BULL, "vxca_caverns1_d60_0_0.vlm"},
	{0x82CEF945C5356A29ULL, "vxht_chantry_2bx_0_0.vlm"},
	{0x82CF210756F7155BULL, "lgt260_eau_tmp_lyrium_mess6143646.cl"},
	{0x82CF5319A41F423CULL, "hlac_w_encounter3_0_9z.dds"},
	{0x82CF9F0DCC421020ULL, "follower_banter6148024.cl"},
	{0x82CFDA9E7ACC0D04ULL, "hlac_ht_0_hightown_a3_-131072_1x.dds"},
	{0x82D01BEA5253CD07ULL, "hla_ht_0_hightown_-196608_0z.dds"},
	{0x82D174F49CD9604EULL, "fence_n.dds"},
	{0x82D1CF839B9A77F9ULL, "dr_primevalentrance_mmi.gff"},
	{0x82D24BF2E3AD8B72ULL, "vxca_cave_tz_0_1.vlm"},
	{0x82D262E535AEF45EULL, "vxlt_0_lowtown_a3_n_c_hd7_0_0.vlm"},
	{0x82D29DAEC734A9F0ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls109a_0n.dds"},
	{0x82D2A104B3C697CBULL, "997244211.ogg"},
	{0x82D2F0FB7FFE7770ULL, "follower_barks6153492.cl"},
	{0x82D30CF5BF370CC8ULL, "6066884_m.fxe"},
	{0x82D37C9E8AF31A38ULL, "gen00fl_merrill6145138.cl"},
	{0x82D491113E289573ULL, "vxlt_0_lowtown_docks_a3_cya_0_1.vlm"},
	{0x82D4A8D4E496ABADULL, "hlac_lt_0_lowtown_docks_131073_0z.dds"},
	{0x82D55144EBC0294DULL, "6043916_m.fxe"},
	{0x82D5602D00953CB7ULL, "hlac_w_encounter3_-327678_0z.dds"},
	{0x82D56A101470DA56ULL, "vxht_bartrand_n_6pm_0_0.vlm"},
	{0x82D5CFFD54CE6C86ULL, "hlac_ht_mansion2_818_47y.dds"},
	{0x82D5D15CBA360E67ULL, "6121225_m.fxe"},
	{0x82D6AE094E7C69ADULL, "w_blightlands_158_0.msh"},
	{0x82D6BC094E7C81E0ULL, "w_blightlands_158_0.mao"},
	{0x82D6C0094E7C8833ULL, "w_blightlands_158_0.mmh"},
	{0x82D6C3FDD617E4F3ULL, "qun131tr_party_ambient2.utt"},
	{0x82D6DCABA425237DULL, "6036740_f.fxe"},
	{0x82D7032E476121D8ULL, "hlac_lt_gallowscourtyard_-2_0y.dds"},
	{0x82D7B71F0107FCBDULL, "vxlt_gallowscourtyard_4wh_0_0.vlm"},
	{0x82D7C9F173F238C0ULL, "6128801_m.fxe"},
	{0x82D8A5CB29F69BC4ULL, "6177819_f.fxe"},
	{0x82D8AD2CEEF0AB58ULL, "6030674_m.fxe"},
	{0x82D8CEF89CBC3DA2ULL, "407448566.ogg"},
	{0x82D8DF66AD8EA3A3ULL, "art\\levels\\races\\design\\prp_deepmushroom.mao"},
	{0x82D8E366AD8EAAF0ULL, "art\\levels\\races\\design\\prp_deepmushroom.mmh"},
	{0x82D8F166AD8EC20AULL, "art\\levels\\races\\design\\prp_deepmushroom.msh"},
	{0x82D8F40027D9950AULL, "mcr310_argument6074450.cl"},
	{0x82D912FD70B79197ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothe_1.mao"},
	{0x82D93D31EECCF6C1ULL, "art\\vfx\\combat\\vfx_sg_awake_act_c.anb"},
	{0x82D988BBB6313605ULL, "vxw_woundedcoast_i7_0_0.vlm"},
	{0x82D9AD7AC9F1C0C4ULL, "21060313.ogg"},
	{0x82DAEB29210FD043ULL, "vxdr_0_deeproads_epl_0_0.vlm"},
	{0x82DB601C808CDA3CULL, "6160863_m.fxe"},
	{0x82DB99626DE376B7ULL, "crt_living_armor.bnk"},
	{0x82DBAEDCB5ACE6BAULL, "6139247_m.fxe"},
	{0x82DC6CCBA1F10A4FULL, "vxht_0_hightown_n_bel_0_0.vlm"},
	{0x82DCE436701AAB21ULL, "616497099.ogg"},
	{0x82DD00801C3D1F7FULL, "6103900_m.fxe"},
	{0x82DD459C2DCC1B3BULL, "193484529.ogg"},
	{0x82DD4996E828D6AEULL, "48212373.ogg"},
	{0x82DDA5263DD4F842ULL, "6204967_m.fxe"},
	{0x82DDB0686756C6B8ULL, "pro000_opening6178536.cl"},
	{0x82DDC459BCBE0177ULL, "ave200_party_comment6185170.cl"},
	{0x82DE540F021B026CULL, "vxca_cavern1_haunted_jfl_0_0.vlm"},
	{0x82DE670695C96F0FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_roba_d.dds"},
	{0x82DED65CA33FAEC1ULL, "6083101_m.fxe"},
	{0x82DEF96C2E3134BCULL, "llac_lt_0_lowtown_docks_n_-65538_5.dds"},
	{0x82DF063407C82E7FULL, "9593041.ogg"},
	{0x82DF8D55C802F36BULL, "vxlt_house_dg_0_0.vlm"},
	{0x82DF97BFAEDE6432ULL, "art\\characters\\weapons\\mauls\\w_mal_ml107.mao"},
	{0x82DFC5E5D2829ADAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablered.mmh"},
	{0x82DFC9E5D282A111ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablered.mao"},
	{0x82DFD7E5D282B920ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablered.msh"},
	{0x82E07D3C9EE36A6DULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga104a_0s.dds"},
	{0x82E0AD0527CDEB5EULL, "art\\vfx\\combat\\vfx_m_ent_conj_c.impa.ani"},
	{0x82E0D631311AB2CDULL, "6095516_f.fxe"},
	{0x82E0D8C281BAF3DDULL, "vxdr_0_deeproads_cu9_0_2.vlm"},
	{0x82E0F40EEDB28B46ULL, "hlac_ca_cavern1_haunted_24424_13x.dds"},
	{0x82E13EF0DF5C7AE3ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumf.mao"},
	{0x82E142F0DF5C8130ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumf.mmh"},
	{0x82E150F0DF5C994AULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumf.msh"},
	{0x82E186683A9F02DAULL, "t1_skn_h02.tnt"},
	{0x82E1E235CE443995ULL, "one100dg_vanard6054575.cl"},
	{0x82E20B9D1F41E138ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_p_rfoot.ani"},
	{0x82E24A27C6CB9C63ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_basic.msh"},
	{0x82E25827C6CBB3BDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_basic.mmh"},
	{0x82E25C27C6CBBACEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0cc_basic.mao"},
	{0x82E27826B0E241D7ULL, "vxht_0_hightown_a2_bu7_0_0.vlm"},
	{0x82E2893AD54201BEULL, "t3_wdg_w07.tnt"},
	{0x82E2904B06E6099DULL, "lt_0_lowtown_docks_105_0.mao"},
	{0x82E2944B06E610D6ULL, "lt_0_lowtown_docks_105_0.mmh"},
	{0x82E29E4B06E621D4ULL, "lt_0_lowtown_docks_105_0.msh"},
	{0x82E2A31C1D8DF37DULL, "559346364.ogg"},
	{0x82E2EB2748B95CABULL, "dre000_bodahn6031315.cl"},
	{0x82E367013D74F49BULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabloop.ani"},
	{0x82E388DEFEAEDCD9ULL, "follower_banter6161105.cl"},
	{0x82E3B8512AB22AC9ULL, "6067122_f.fxe"},
	{0x82E3BAF1E34D8527ULL, "349949539.ogg"},
	{0x82E4036D37082B2AULL, "701986661.ogg"},
	{0x82E404C37FF5E8A1ULL, "50756543.ogg"},
	{0x82E45B9E5A6E9611ULL, "hlac_ht_chantry_2534_22z.dds"},
	{0x82E4E644543DF1B7ULL, "1004669034.ogg"},
	{0x82E51114155A93CBULL, "fire_f.dds"},
	{0x82E56D249289EDE8ULL, "vxlt_gallowstemplar_n_26e_0_0.vlm"},
	{0x82E6228E26B30C0BULL, "vxht_keep_67e_0_0.vlm"},
	{0x82E63C94B5D7563CULL, "face_face_p7.psh.fxo.dx11.assembly"},
	{0x82E64A36CA6C6E34ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_f_c.impa.ani"},
	{0x82E71E31EED8FF3AULL, "art\\vfx\\combat\\vfx_sg_awake_act_c.mmh"},
	{0x82E73086BB69C103ULL, "vxlt_0_lowtown_docks_n_cqt_0_0.vlm"},
	{0x82E7C825D38ADE04ULL, "vxlt_0_lowtown_a2_g60_0_0.vlm"},
	{0x82E7F328E11E4851ULL, "dae000cr_grand_cleric6181918.cl"},
	{0x82E82266A37CE565ULL, "hlac_w_blightlands_262145_0z.dds"},
	{0x82E834CC977451F1ULL, "6087387_m.fxe"},
	{0x82E8E08D2DD4E0D3ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_min_c.impa.ani"},
	{0x82E8E56A58CF7178ULL, "dae000_fem_commoner1_mob6174491.cl"},
	{0x82E8FB6D1B725D32ULL, "57529846.ogg"},
	{0x82EA6D4C89E09D2AULL, "gen00fl_merrill6115757.cl"},
	{0x82EA77542A5F3F42ULL, "vxw_bonepitmining_a2_6b_0_0.vlm"},
	{0x82EA97D572F75429ULL, "6088323_f.fxe"},
	{0x82EB2887A2B4B419ULL, "gen00fl_fenris6105555.cl"},
	{0x82EB3A78A9A4C47FULL, "art\\characters\\weapons\\longswords\\w_lsw_010a.phy"},
	{0x82EB6D75BBA5AAA9ULL, "dae000ip_crowd_mf7_talk11.utp"},
	{0x82EB725D0E3337FFULL, "6104800_m.fxe"},
	{0x82EC819F008F55EDULL, "plt_vrc300pt_n_hntd_done.nss"},
	{0x82ECEFAF6E0DC113ULL, "vxht_mansion2_330_0_0.vlm"},
	{0x82ED1C02D9A17D73ULL, "cs_rom000cs_fenris_alt.bnk"},
	{0x82ED208336705BE5ULL, "dae000_arishok6072400.cl"},
	{0x82ED3FA4018549E7ULL, "ht_keep_n_4519.anb"},
	{0x82ED4736CA728796ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_f_c.impa.gad"},
	{0x82ED494BE937BD59ULL, "hlac_lt_0_lowtown_a2_-65537_4y.dds"},
	{0x82ED50B55A3EF073ULL, "fen110_cut_fenris_intro.bnk"},
	{0x82ED8827DF849EFBULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.cess.ani"},
	{0x82EDC50527D89E2FULL, "art\\vfx\\combat\\vfx_m_ent_conj_c.impa.evt"},
	{0x82EE09A79C8E48DCULL, "596843346.ogg"},
	{0x82EE47F6774E4D5FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chandelier_sa.dds"},
	{0x82EF11E4B926158EULL, "hm_sabin.mor"},
	{0x82EF5226426256ACULL, "75919717.ogg"},
	{0x82EF63187C54D4FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_wall_lean_left.anb"},
	{0x82EFAB48CFAFB884ULL, "vxlt_gallowsprison_n_6po_0_0.vlm"},
	{0x82EFD937275C0D02ULL, "87615069.ogg"},
	{0x82F06F013D7F8C92ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabloop.evt"},
	{0x82F0C02F8B0C4D73ULL, "wispy_smo_bright.dds"},
	{0x82F172B55A4312C3ULL, "fen110_cut_fenris_intro.cnv"},
	{0x82F24BAABC3381BBULL, "313580869.ogg"},
	{0x82F28B1E927A8BFDULL, "619248519.ogg"},
	{0x82F293D6F3428E17ULL, "6087051_m.fxe"},
	{0x82F2B61D77D4AD52ULL, "vxht_mansion2_2bg_0_0.vlm"},
	{0x82F2E01A7D47F431ULL, "qun211_braggart6075712.cl"},
	{0x82F30997E1F90BFEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordc_sa.dds"},
	{0x82F3B3F5FC580CC9ULL, "hlac_lt_hangedman_-1_28z.dds"},
	{0x82F40285DEAFB8B4ULL, "art\\vfx\\environment\\vfx_trap_glow_p.mmh"},
	{0x82F46236CA78D4F5ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_f_c.impa.evt"},
	{0x82F4946B18A9248BULL, "vxht_keep_60d_0_0.vlm"},
	{0x82F4C60527DEBED8ULL, "art\\vfx\\combat\\vfx_m_ent_conj_c.impa.gad"},
	{0x82F4E541002A693EULL, "692727630.ogg"},
	{0x82F50F74232A48B6ULL, "gen00fl_aveline6150247.cl"},
	{0x82F5D3A465D1746BULL, "l_env_wal_lowtowndocks_whistle.wbd"},
	{0x82F5D88D2DDF5D0AULL, "art\\vfx\\combat\\vfx_glb_lit_imp_min_c.impa.evt"},
	{0x82F609DA76490D5AULL, "6170684_m.fxe"},
	{0x82F610008872EE28ULL, "324292141.ogg"},
	{0x82F65AC6D5C5A45CULL, "vxht_mansion2_3ea_0_0.vlm"},
	{0x82F66C759A1D6705ULL, "vxw_woundedcoast_c5_0_0.vlm"},
	{0x82F730F83B02529FULL, "mag230cr_evelina.utc"},
	{0x82F78E013D85E025ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabloop.gad"},
	{0x82F7BBD5AD783B0AULL, "lt_gallowstemplar_n_2386.rml"},
	{0x82F7F7A50D633460ULL, "clf260_harley6088766.cl"},
	{0x82F86ADC0392D691ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_mlw.dds"},
	{0x82F8BAFD0E8761C1ULL, "vxht_brothel_n_3vm_0_0.vlm"},
	{0x82F99C5056AF32FFULL, "6072262_m.fxe"},
	{0x82F9EF261BDAEB15ULL, "6149229_m.fxe"},
	{0x82F9F6D24F516F21ULL, "vxht_0_hightown_a3_bwz_0_0.vlm"},
	{0x82FA0D0FDE16DDCBULL, "85738136.ogg"},
	{0x82FAC217405F1DDDULL, "gen_im_fla_bethany_02.uti"},
	{0x82FB1198A57A241EULL, "art\\characters\\playercharacter\\humanfemale\\glv_wrba.phy"},
	{0x82FB1F021286E8E8ULL, "363071046.ogg"},
	{0x82FB1F579B7BE309ULL, "lowtown_guard_chase6169503.cl"},
	{0x82FB9027DF90E9F2ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.cess.evt"},
	{0x82FB9B79BC019619ULL, "cod_hst_calendar.utp"},
	{0x82FBB2BBDD50DDD4ULL, "vxht_0_hightown_n_cab_0_0.vlm"},
	{0x82FBFBD9AE716B9DULL, "393918844.ogg"},
	{0x82FC3B78A9B33884ULL, "art\\characters\\weapons\\longswords\\w_lsw_010a.mmh"},
	{0x82FC3F78A9B33FD7ULL, "art\\characters\\weapons\\longswords\\w_lsw_010a.mao"},
	{0x82FC5178A9B35E26ULL, "art\\characters\\weapons\\longswords\\w_lsw_010a.msh"},
	{0x82FC78B07B5EA4FCULL, "643836879.ogg"},
	{0x82FC8350C09625C9ULL, "252533586.ogg"},
	{0x82FCB78D2DE5447DULL, "art\\vfx\\combat\\vfx_glb_lit_imp_min_c.impa.gad"},
	{0x82FCBE7D963795C4ULL, "602147631.ogg"},
	{0x82FCEE1AB8371F3AULL, "seb321st_rebel_mage.stg"},
	{0x82FCF87B5218EA9CULL, "vxht_0_hightown_a2_c3f_0_1.vlm"},
	{0x82FD2FA9A8F151BFULL, "gen_im_trs_crc_junk3_01.uti"},
	{0x82FE1E7853C5862FULL, "6088234_m.fxe"},
	{0x82FE220311E2F597ULL, "vxlt_gallowscourtyard_n_6yi_0_0.vlm"},
	{0x82FE52F1C9E54949ULL, "vxht_0_hightown_c0w_0_0.vlm"},
	{0x82FE90C05A56457DULL, "hlac_w_woundedcoast_5_0y.dds"},
	{0x82FF02CDD4F6478BULL, "dae000cr_grand_cleric6199346.cl"},
	{0x82FF06C27BEF9284ULL, "6053182_f.fxe"},
	{0x82FF5BE094CE72D7ULL, "vxht_0_hightown_a2_bc1_0_0.vlm"},
	{0x83006CD01BAAEBB7ULL, "gen00fl_aveline6132873.cl"},
	{0x8300D2C53A089F9CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs105a_0s.dds"},
	{0x83012417765D130AULL, "w_sundermount_42_0.anb"},
	{0x830153668CE46E61ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firetarp_b.mmh"},
	{0x8301985A795F81A0ULL, "hlac_dr_primevaltomb_659_6x.dds"},
	{0x8301AF27DF958985ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.cess.gad"},
	{0x8301B49B6A567ADAULL, "6118365_m.fxe"},
	{0x83024156EF9392E2ULL, "mag311_emile_de_launcet6198486.cl"},
	{0x8302AE356DB0EAB0ULL, "vxw_encounter3_fq_0_0.vlm"},
	{0x8302FE5BBCED68BBULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv2.evt"},
	{0x830405FAE8D42715ULL, "172396287.ogg"},
	{0x830417CD3BC7F9C4ULL, "234857537.ogg"},
	{0x83048922345B23B0ULL, "vxht_keep_6yd_0_0.vlm"},
	{0x8305AA5DAFBB5292ULL, "gui_guipixelshader_p6.psh.fxo"},
	{0x8305E4E5D2A39EE5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablered.phy"},
	{0x83060F66A7732179ULL, "llac_w_woundedcoast_196615_0.dds"},
	{0x830619419570BEFEULL, "hlac_w_encounter1_131071_0z.dds"},
	{0x8306345CA56167FDULL, "llac_ht_0_hightown_a2_-131074_1.dds"},
	{0x83067F8C4AC65BD3ULL, "6033126_m.fxe"},
	{0x8306C7A43B482B18ULL, "vfx_blood_f.dds"},
	{0x83075B0E0E4B7042ULL, "vxw_bonepitmining_a2_8l_0_0.vlm"},
	{0x83076EF05E938100ULL, "140164736.ogg"},
	{0x83079B70BDDD2A8FULL, "204516544.ogg"},
	{0x8307B1BA4630ED79ULL, "vxlt_0_lowtown_fim_0_0.vlm"},
	{0x8307D37FE4C6F0F3ULL, "mag320_orsino6071121.cl"},
	{0x8307D60D02B519E4ULL, "hlac_lt_gallowstemplar_f_2395_0z.dds"},
	{0x8307FD3A7B9379FFULL, "630375320.ogg"},
	{0x8308B2A671857CB3ULL, "6029391_f.fxe"},
	{0x8308F75BBCF1C8FCULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv2.gad"},
	{0x83090BB3068F24BFULL, "vxca_cavern1_haunted_jfc_0_0.vlm"},
	{0x8309CD4BEDBBFF33ULL, "305556768.ogg"},
	{0x830A0E3C2F33C964ULL, "vxlt_0_lowtown_a2_n_c_gwv_0_0.vlm"},
	{0x830A809549597BEEULL, "6128115_m.fxe"},
	{0x830AE811B810D018ULL, "6113376_f.fxe"},
	{0x830B370339D8194BULL, "hlac_ht_playershouse_n_855_33x.dds"},
	{0x830B8FF387445E54ULL, "mcr370_assignments6176731.cl"},
	{0x830C905545BB9EAFULL, "6083334_m.fxe"},
	{0x830CC4CEBBDB61C9ULL, "vxht_0_hightown_c7v_0_0.vlm"},
	{0x830D911865BB6B30ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_paleempty.mao"},
	{0x830D951865BB7243ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_paleempty.mmh"},
	{0x830D9D04CBD95CFDULL, "mcr310_argument6074363.cl"},
	{0x830DA31865BB899DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_paleempty.msh"},
	{0x830DBCDB0304481AULL, "vxht_0_hightown_a3_c3x_0_0.vlm"},
	{0x830E28215456878CULL, "w_encounter3_64_0.phy"},
	{0x830EE68548E10EB0ULL, "832873030.ogg"},
	{0x830F7E16781FAAEEULL, "w_woundedcoast_247_0.mmh"},
	{0x830F8A16781FBE85ULL, "w_woundedcoast_247_0.mao"},
	{0x830F9816781FD6DCULL, "w_woundedcoast_247_0.msh"},
	{0x830FC320F758A94EULL, "135950364.ogg"},
	{0x830FF65BBCF7E5FAULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\takeoffflylandv2.ani"},
	{0x831026785C4C935AULL, "6117265_m.fxe"},
	{0x831034D16D0A51CFULL, "375663337.ogg"},
	{0x83104E34BFB76391ULL, "903468650.ogg"},
	{0x83109D65FC42F610ULL, "489421909.ogg"},
	{0x8310B936CBA6649BULL, "one140cr_charade6139432.cl"},
	{0x831114F693B21459ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_enter.ani"},
	{0x83111C203B4DC43AULL, "llac_lt_gallowscourtyard_n_-131073_1.dds"},
	{0x8311E39A034D26D0ULL, "613273420.ogg"},
	{0x8311F91A46D3A0B0ULL, "vxca_caverns1_ih8_0_0.vlm"},
	{0x83122EA7E0A65460ULL, "dre000_dougal6129783.cl"},
	{0x83128645380B812FULL, "one100dg_kelder6063996.cl"},
	{0x8312FAFF26513C30ULL, "vxht_keep_6be_0_0.vlm"},
	{0x831364D0DF7F5915ULL, "dre000_bethany6037336.cl"},
	{0x83137B18758EA0B7ULL, "vxca_caverns1_mines_ec3_0_0.vlm"},
	{0x8313BF0C766D01DBULL, "6166866_m.fxe"},
	{0x8313F6477547C97EULL, "zz_lgt_1_debug6114358.cl"},
	{0x831411876775B989ULL, "art\\effects\\cloudshadows\\w_encounter2\\cloud2.dds"},
	{0x831415F8791A10B5ULL, "mcr370_meredith6085933.cl"},
	{0x83143B1EDF40E734ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tray_n.dds"},
	{0x83144DCA5169F6F0ULL, "326057102.ogg"},
	{0x83148AFE8C625853ULL, "vxht_mansion2_2ug_0_0.vlm"},
	{0x8314E860B143C837ULL, "vxlt_hangedman_2u9_0_0.vlm"},
	{0x8314F8368C7A43FFULL, "character_character_p82.psh.fxo.dx11.assembly"},
	{0x83154013B9F74F8BULL, "727644591.ogg"},
	{0x831598E2166A5E97ULL, "vxdr_0_deeproads_f5p_0_0.vlm"},
	{0x83159D8F700B55D9ULL, "vxw_bonepitmining_a2_26e_0_0.vlm"},
	{0x8315A78C5257F84FULL, "391133616.ogg"},
	{0x8315DBF9E1F55E99ULL, "6133790_m.fxe"},
	{0x831622E6B6BA561AULL, "6111340_m.fxe"},
	{0x8316400706009713ULL, "hlac_lt_0_lowtown_n_65537_5x.dds"},
	{0x83169A44018A04FBULL, "lt_0_lowtown_a2_n_51_0.rml"},
	{0x8316BEF726C512E1ULL, "gen00fl_merrill6106891.cl"},
	{0x831713C5B4959183ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em02.mmh"},
	{0x831721C5B495A9DDULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em02.msh"},
	{0x83179749BFDF9F0BULL, "vxlt_gallowstemplar_n_4uq_0_0.vlm"},
	{0x8317EFF693B7F39BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_enter.gad"},
	{0x8318326B9EDEFFC9ULL, "vxht_keep_5y2_0_0.vlm"},
	{0x8318B1A6DD5C5FA9ULL, "54878818.ogg"},
	{0x8318C4D67D3FA581ULL, "vxht_mansion2_3e0_0_0.vlm"},
	{0x8318D2A8B0EA182CULL, "249681819.ogg"},
	{0x8319681D6339425EULL, "vxca_caverns2prc_a0p_0_0.vlm"},
	{0x831A0AB27B2146FEULL, "71649005.ogg"},
	{0x831A40A6DCABD0F9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\electr.gad"},
	{0x831A4E036967A5F3ULL, "loadscreens_seb.gda"},
	{0x831A55EA478DDB2AULL, "564773064.ogg"},
	{0x831A55FC0ED5A253ULL, "vxw_bonepitdestroyed_a3_180_0_0.vlm"},
	{0x831AAC7573161DF0ULL, "vxdr_primevaltomb_527_0_0.vlm"},
	{0x831B70A3F5CE76DAULL, "516729598.ogg"},
	{0x831B83B29D86CF92ULL, "6069043_m.fxe"},
	{0x831BB139F4BAB57FULL, "art\\characters\\creatures\\dragonling\\textures\\c_dragling_0d.dds"},
	{0x831BB5A465F21039ULL, "l_env_wal_lowtowndocks_whistle.bnk"},
	{0x831C2EB7F4FA609AULL, "vxht_0_hightown_a3_c4y_0_0.vlm"},
	{0x831CD698773C7034ULL, "lgt110tr_noble_attack.ncs"},
	{0x831D0585DED2C36BULL, "art\\vfx\\environment\\vfx_trap_glow_p.anb"},
	{0x831D1DE03EE22725ULL, "vxdr_dungeon_1ls_0_1.vlm"},
	{0x831D45BFEED83A88ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masb.mmh"},
	{0x831D51BFEED84F2BULL, "art\\characters\\playercharacter\\humanmale\\hlf_masb.mao"},
	{0x831D63BFEED86DB2ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masb.msh"},
	{0x831DAC424B75CE20ULL, "hlac_dr_dungeon_3214_0z.dds"},
	{0x831DBB1C13FBA5BDULL, "w_bonepit_22_0_ll.mmh"},
	{0x831E10C490A7E806ULL, "llac_ht_0_hightown_a2_65536_2.dds"},
	{0x831E35E92C1107EBULL, "hlac_ca_caverns1_mines_24425_4z.dds"},
	{0x831E961865C9E768ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_paleempty.phy"},
	{0x831EBCFEC5477072ULL, "mcr370_meredith6083071.cl"},
	{0x831EF9DC06CBD896ULL, "288684253.ogg"},
	{0x831F0CF693BE4380ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_enter.evt"},
	{0x831F75C032B8F1F8ULL, "42974839.ogg"},
	{0x831F8F313F504DD7ULL, "6200737_m.fxe"},
	{0x831F932789840136ULL, "isa311_isabela6094461.cl"},
	{0x831FD7D13882374CULL, "qun100_mage_party_comment6019094.cl"},
	{0x831FF2DBB6EBFC78ULL, "ach_seb_avenged.dds"},
	{0x83200EAA89DDB536ULL, "vxlt_gallowscourtyard_72o_0_0.vlm"},
	{0x8320154D8BE0846EULL, "25441138.ogg"},
	{0x8320432BF10B8443ULL, "99412163.ogg"},
	{0x832051A6DCB05947ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\electr.ani"},
	{0x832058D8E3BEEFCEULL, "461381087.ogg"},
	{0x83207AB339F3DF3BULL, "vxht_bartrand_n_288_0_0.vlm"},
	{0x83208F782C3FB075ULL, "vxlt_gallowstemplar_379_0_0.vlm"},
	{0x8320B4154C91E90EULL, "852077052.ogg"},
	{0x83215E37D2222FE6ULL, "llac_lt_0_lowtown_docks_a3_n_0_2.dds"},
	{0x8321A973A13C59BDULL, "zz_mag_debug6054590.cl"},
	{0x8321D34A98885999ULL, "w_woundedcoast_22_0.msh"},
	{0x8321DD4A98886A87ULL, "w_woundedcoast_22_0.mmh"},
	{0x8321E14A988871D4ULL, "w_woundedcoast_22_0.mao"},
	{0x832203D4B8BDB522ULL, "llac_w_bonepitmining_a2_131072_0.dds"},
	{0x83225E6190397B29ULL, "6166543_m.fxe"},
	{0x83226A6CE33DAF9AULL, "664627241.ogg"},
	{0x83229BA217D0DD02ULL, "vxlt_foundry_n_12c_0_0.vlm"},
	{0x8322A23C865983B4ULL, "6153313_m.fxe"},
	{0x8322D7171E3DE093ULL, "hlac_lt_0_lowtown_a3_n_c_-3_3y.dds"},
	{0x8322F8CE13589E1BULL, "69767475.ogg"},
	{0x83232A77F6E4D89BULL, "hlac_lt_0_lowtown_docks_n_65535_2x.dds"},
	{0x8323AE212D3F6F01ULL, "qun110_petrice6018627.cl"},
	{0x8323BE6EF1CA4A18ULL, "and100_anders6095842.cl"},
	{0x8323D52B6E6F834BULL, "109338452.ogg"},
	{0x8323F0FA8D80B944ULL, "sse_ss_ragedemon_1.gda"},
	{0x8323FE65119F3274ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_rnf.ani"},
	{0x83241D30A51809BFULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv3.gad"},
	{0x83246A151630DCF3ULL, "6093678_m.fxe"},
	{0x8324EEC5EB5D2C05ULL, "hlac_lt_0_lowtown_docks_a3_-131074_5x.dds"},
	{0x8324F3382724CD9BULL, "vxlt_gallowstemplar_f_3qz_0_0.vlm"},
	{0x83251F71932BFFCEULL, "hlac_w_bonepit_196611_0z.dds"},
	{0x83255BDFC4F68CDFULL, "dae000ip_crowd_lm3_talk4.utp"},
	{0x83257BDF40421B92ULL, "vxht_mansion_n_6j0_0_0.vlm"},
	{0x8325BB003C673C7CULL, "vxht_chantry_2cb_0_0.vlm"},
	{0x8325FB3105AE18CEULL, "vxht_mansion_n_5a2_0_0.vlm"},
	{0x8326A4972CAFF350ULL, "464245678.ogg"},
	{0x8326EF3CB01851DAULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grassmed.mmh"},
	{0x8327013CB0187020ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grassmed.msh"},
	{0x83272C3FBBA34F4FULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masd.phy"},
	{0x8328035BB09EDBECULL, "hf_hadriana.mor"},
	{0x832814C5B4A406A8ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em02.phy"},
	{0x83284AD2C38AE46DULL, "6033753_m.fxe"},
	{0x8328970A0C6E8DC5ULL, "6036457_m.fxe"},
	{0x8328DED7A3026D09ULL, "sky_ca_caverns2.mao"},
	{0x83291917767E52A3ULL, "w_sundermount_42_0.msh"},
	{0x83292717767E6AFDULL, "w_sundermount_42_0.mmh"},
	{0x83292B17767E710EULL, "w_sundermount_42_0.mao"},
	{0x83293CCB9B0A5D1CULL, "vxlt_0_lowtown_a2_n_fig_0_0.vlm"},
	{0x83294F1F00D1B8A1ULL, "water_pan_blur.mao"},
	{0x8329F6E7D45E4CC8ULL, "6057329_m.fxe"},
	{0x8329FB6511A398D6ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_rnf.gad"},
	{0x832A144DB1827EF8ULL, "169563598.ogg"},
	{0x832A785E5D6E49B4ULL, "17166603.ogg"},
	{0x832AB228B313CA54ULL, "zz_jp_linktoactor_bug.cut"},
	{0x832AD2E810CF0A13ULL, "gen_im_trs_dem_lgt_boo_r0.uti"},
	{0x832B1230A51E15ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv3.evt"},
	{0x832BABDA374559F9ULL, "6167997_m.fxe"},
	{0x832C5D47E9C41EF9ULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_c.mmh"},
	{0x832C67434AF0D946ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon_n.dds"},
	{0x832C874FEA8BB763ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_r_5p.ani"},
	{0x832D558E3D14D710ULL, "950357748.ogg"},
	{0x832DA9127CEE5155ULL, "568492706.ogg"},
	{0x832E3118F2B74173ULL, "dae_mage_aoe_duration.ncs"},
	{0x832EFDE457D5333BULL, "154673364.ogg"},
	{0x832F3358D7A73CD0ULL, "hlac_lt_0_lowtown_a3_n_c_2_1x.dds"},
	{0x832F9C7150693635ULL, "6033704_m.fxe"},
	{0x832FDEB90A3D6D30ULL, "vxht_keep_6hc_0_0.vlm"},
	{0x832FFF00231A9534ULL, "vxca_caverns2prc_9xj_0_0.vlm"},
	{0x83308A076D7E7D82ULL, "lowtown_fem_urchin_mob6165930.cl"},
	{0x8331166511A9E535ULL, "art\\characters\\creatures\\spider\\animation\\p_tn90r_rnf.evt"},
	{0x8331AE487F58274EULL, "vxlt_hangedman_1w1_0_0.vlm"},
	{0x8331E2C52FACF180ULL, "dae200cr_viscount6087170.cl"},
	{0x833219A4753DC12DULL, "dae000ip_crowd_lm1_beg.utp"},
	{0x833243B4998FB3D6ULL, "w_bonepit_15_0.msh"},
	{0x83324CCDE16A0221ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_arrow_01_n.dds"},
	{0x83324DB4998FC4D4ULL, "w_bonepit_15_0.mmh"},
	{0x833251B4998FCBE7ULL, "w_bonepit_15_0.mao"},
	{0x83325856E0407785ULL, "6143520_m.fxe"},
	{0x8332814953F786D9ULL, "vxw_bonepitmining_a2_28k_0_0.vlm"},
	{0x8332B49E81274DCEULL, "mag370_meredith6064958.cl"},
	{0x83332E756698A379ULL, "6147533_m.fxe"},
	{0x8333EDBC32383412ULL, "vxht_mansion_n_6m1_0_0.vlm"},
	{0x83340B8D0EA9E988ULL, "lt_gallowstemplar_2385.rml"},
	{0x83343EFB1DB0DFC1ULL, "hightown_male_noble_gossip6171409.cl"},
	{0x83347198F6E50E2CULL, "w_blightlands_73_0.phy"},
	{0x8334A19762880C28ULL, "121552411.ogg"},
	{0x8334B5987750CA61ULL, "lgt110tr_noble_attack.utt"},
	{0x83353A02D9DF65D1ULL, "cs_rom000cs_fenris_alt.wbd"},
	{0x8335AAEE3558D831ULL, "439169238.ogg"},
	{0x833691334FF00384ULL, "zz_firestaff.cut"},
	{0x8336C32E5A481A83ULL, "15754458.ogg"},
	{0x8336EA8C18DA1073ULL, "6070335_f.fxe"},
	{0x833737057A0C5FB3ULL, "6047250_f.fxe"},
	{0x833781BA5ECD5F80ULL, "vxht_0_hightown_n_bc5_0_0.vlm"},
	{0x8337BAE3FB01CEF5ULL, "vxca_caverns2_92x_0_0.vlm"},
	{0x8337E92F633CD81AULL, "6087498_m.fxe"},
	{0x8338454A71E042F1ULL, "hlac_w_woundedcoast_196613_0z.dds"},
	{0x8338D1231C49225DULL, "6170636_m.fxe"},
	{0x8338FA30A52A2A05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\attackv3.ani"},
	{0x833966644AE1652CULL, "6142544_m.fxe"},
	{0x83399F4FEA966A5AULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_r_5p.evt"},
	{0x8339C8567B7CFBB6ULL, "vxlt_gallowstemplar_n_6qb_0_0.vlm"},
	{0x8339F7D00DAB952FULL, "gen00fl_aveline6129037.cl"},
	{0x833A23833F83F0ACULL, "hlac_ca_caverns2_5802_36z.dds"},
	{0x833A39CB959015AAULL, "6163864_m.fxe"},
	{0x833A432BC37679E7ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_father_andersintro.gad"},
	{0x833A5330DF640D7BULL, "6162376_m.fxe"},
	{0x833AC6D381EC5BB1ULL, "lt_0_lowtown_docks_a3_246_0.mmh"},
	{0x833AD2D381EC7052ULL, "lt_0_lowtown_docks_a3_246_0.mao"},
	{0x833AE0D381EC886FULL, "lt_0_lowtown_docks_a3_246_0.msh"},
	{0x833B22334DB1F7B1ULL, "158130436.ogg"},
	{0x833CE332C5B6FC17ULL, "hlac_lt_hangedman_n_-1_19y.dds"},
	{0x833D023C66677A24ULL, "6140014_m.fxe"},
	{0x833D52F43D0A8439ULL, "ser210_witness6168159.cl"},
	{0x833D8F98E2F9CC94ULL, "975002050.ogg"},
	{0x833E47F8CFD1F8B6ULL, "6127590_m.fxe"},
	{0x833ECDBF2495CA94ULL, "vxlt_0_lowtown_a3_n_c_gcc_0_0.vlm"},
	{0x833EFC4EDB1DBC22ULL, "llac_w_encounter2_camp_327679_0.dds"},
	{0x833FAF40FA257DCAULL, "co_bl_clo.mao"},
	{0x83409E4FEA9C874DULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_r_5p.gad"},
	{0x8340B5EBEED1248DULL, "6165959_m.fxe"},
	{0x8341CFCA32AFBEE9ULL, "6132739_m.fxe"},
	{0x8342149DEBBB4307ULL, "vxw_encounter1_b4_0_0.vlm"},
	{0x834228F56681A7F9ULL, "llac_w_sundermount_262149_0.dds"},
	{0x834271D1E603F034ULL, "hlac_lt_0_lowtown_a3_n_c_1_11z.dds"},
	{0x83429CEA330AD0BAULL, "6079653_m.fxe"},
	{0x8342B6D7EE4EC8D8ULL, "gen00fl_anders6125466.cl"},
	{0x8342EA7652C18DB7ULL, "6088300_m.fxe"},
	{0x83430B5BBED53F84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_loop.gad"},
	{0x834369277ABBB436ULL, "vxht_0_hightown_a3_bu7_0_0.vlm"},
	{0x8343702CA3D2978DULL, "6048114_f.fxe"},
	{0x8343D2456D5EBD03ULL, "qun221_viscount6070197.cl"},
	{0x83445EB2D4550D7DULL, "llac_lt_0_lowtown_a2_n_c_0_1.dds"},
	{0x8344634B36C99B51ULL, "hlac_dr_dungeon_3207_5x.dds"},
	{0x8344709F00D9AEFEULL, "plt_vrc300pt_n_hntd_done.plo"},
	{0x834496D95E64D38FULL, "6087513_f.fxe"},
	{0x83449774CCBC55D7ULL, "vxlt_0_lowtown_a2_hd2_0_0.vlm"},
	{0x834528ECCE2A74FDULL, "vxw_sundermount_291_0_0.vlm"},
	{0x8345413F2EF0B0C5ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_1x.dds"},
	{0x8345E193D5DE5BB9ULL, "6147865_m.fxe"},
	{0x8346B462012A4807ULL, "dae000ip_crowd_ff1_talk10.utp"},
	{0x8347230FFD4BD1C3ULL, "fen210_fenris6109418.cl"},
	{0x8348E0AC44629DA9ULL, "106330233.ogg"},
	{0x8348FCD96F5BF502ULL, "hlac_dr_primevalentrance_1724_0z.dds"},
	{0x8349DBEF6BB27FBBULL, "vxlt_0_lowtown_docks_a3_n_d4w_0_0.vlm"},
	{0x834A1A5BBEDB7862ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_3_loop.ani"},
	{0x834A5990E3EE420FULL, "22266991.ogg"},
	{0x834B5907BF16B7CCULL, "hlac_lt_0_lowtown_docks_196607_2y.dds"},
	{0x834C72B25EA4F898ULL, "gen00fl_merrill6116314.cl"},
	{0x834CB985C9353021ULL, "6075998_f.fxe"},
	{0x834CE3E1291E3450ULL, "6125118_m.fxe"},
	{0x834D3BC4FD50BE65ULL, "follower_banter6171113.cl"},
	{0x834DDBDA6C40E4C0ULL, "vxca_cavern1_haunted_il9_0_0.vlm"},
	{0x834E302BC387022DULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_father_andersintro.ani"},
	{0x834EDDB449CC25A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side_left.ani"},
	{0x834EFF21548DD957ULL, "w_encounter3_64_0.mmh"},
	{0x834F0321548DE064ULL, "w_encounter3_64_0.mao"},
	{0x834F1521548DFEC9ULL, "w_encounter3_64_0.msh"},
	{0x834F559B6443D14BULL, "qun160_vashothlead6053950.cl"},
	{0x834F9C89818BDF0AULL, "756508713.ogg"},
	{0x834FD9565A4DAB21ULL, "6026983_f.fxe"},
	{0x834FF1FA50672441ULL, "mh_nuncio_guard.mor"},
	{0x835072542A772D75ULL, "framebuffereffects_cocdownsample_p0.vsh.fxo.dx11.assembly"},
	{0x83507B57A2EC6122ULL, "vxlt_0_lowtown_docks_a3_n_a0k_0_0.vlm"},
	{0x8350867ADDB15B87ULL, "vxw_encounter1_g7_0_0.vlm"},
	{0x835095AE8B76DB0FULL, "420677588.ogg"},
	{0x83511097124985FCULL, "vxht_chantry_2ra_0_0.vlm"},
	{0x835156C963337526ULL, "dae000_madam_lusine6154903.cl"},
	{0x8352536D8AD42B08ULL, "hlac_lt_0_lowtown_a3_n_c_-65538_0y.dds"},
	{0x835264AF4DE69A57ULL, "6117699_m.fxe"},
	{0x8352CFDDEC049035ULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_act_c.mmh"},
	{0x8352E6977B30A1C1ULL, "6104885_f.fxe"},
	{0x835305DC979611E0ULL, "hlac_lt_darktown_-1_83y.dds"},
	{0x8354119DFA659336ULL, "spt_leafcards.vsh.fxo"},
	{0x835446575687D34BULL, "42607934.ogg"},
	{0x83544F4E8E98CEB6ULL, "vxlt_0_lowtown_a2_n_hd0_0_0.vlm"},
	{0x83545370C7D47439ULL, "6140965_m.fxe"},
	{0x835485D25559ADA7ULL, "986695035.ogg"},
	{0x8354B49BFF224D9BULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.impa.evt"},
	{0x8354DAB449D08C06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side_left.gad"},
	{0x8354DD4A98B3C4E4ULL, "w_woundedcoast_22_0.rml"},
	{0x835503864D0F9205ULL, "14455446.ogg"},
	{0x835528FAF8B8D33DULL, "6134904_m.fxe"},
	{0x835550E02542E19CULL, "hlac_lt_0_lowtown_131073_2x.dds"},
	{0x83555A47E9E71EEEULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_c.anb"},
	{0x8355705B5EC0CA4EULL, "471263357.ogg"},
	{0x835585DA9FB2993AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryleftb_1.mao"},
	{0x8355BF186EB6FE4BULL, "plt_cod_itm_cummerbund.plo"},
	{0x8355CA71C1E2BC2DULL, "dae221ip_home_amb_chest.utp"},
	{0x8355D88958DB90EEULL, "gen_im_cft_reg_i3.uti"},
	{0x83566CF576598E8FULL, "405766559.ogg"},
	{0x83571C48509D2610ULL, "443984343.ogg"},
	{0x8357227A10749420ULL, "hlac_ca_cavern1_haunted_24423_9z.dds"},
	{0x8357BEC093C495DAULL, "vxca_caverns2prc_9uc_0_0.vlm"},
	{0x83585936BD44BB35ULL, "dae000ip_crowd_ff4_talk2.utp"},
	{0x8358D3F2C83B10D7ULL, "6076895_m.fxe"},
	{0x8358E430B31EE63CULL, "llac_lt_0_lowtown_a2_n_c_65537_6.dds"},
	{0x8359607DAFA387BFULL, "25166489.ogg"},
	{0x8359D14E68753A16ULL, "fen210_fenris6109390.cl"},
	{0x835B5390A089D727ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_cast_c.impa.gad"},
	{0x835BAD9BFF2860DCULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.impa.gad"},
	{0x835BD5B449D6A2E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_spit_side_left.evt"},
	{0x835BF93D7E5828C9ULL, "6108011_m.fxe"},
	{0x835C0659AD8FFE3EULL, "6068844_m.fxe"},
	{0x835CEE659E3BAE41ULL, "hlac_dr_slavers_3216_10x.dds"},
	{0x835D55D76EC74E5DULL, "2973068.ogg"},
	{0x835DD3528AFECD49ULL, "lt_0_lowtown_docks_a3_98_0.mao"},
	{0x835DDF528AFEE1E2ULL, "lt_0_lowtown_docks_a3_98_0.mmh"},
	{0x835DE1528AFEE498ULL, "lt_0_lowtown_docks_a3_98_0.msh"},
	{0x835E4599B4C6211BULL, "vxca_caverns2_93u_0_0.vlm"},
	{0x835E877204693BAEULL, "hm_mag_controlled_2.mor"},
	{0x835EC818EAC78B90ULL, "mag121_idunna6083981.cl"},
	{0x835EC9723866AD37ULL, "6131599_f.fxe"},
	{0x835EEAA94AF6D302ULL, "3365315.ogg"},
	{0x835F2AE1B4825442ULL, "hlac_w_sundermount_-196604_0z.dds"},
	{0x835F4E35795E50ADULL, "vxlt_darktown_2rc_0_0.vlm"},
	{0x835F769936DC68E8ULL, "l_dre180ar_keep_fake.wbd"},
	{0x835FAE0D30AAB399ULL, "618821562.ogg"},
	{0x83605596780AFB1DULL, "vxca_caverns2prc_ae8_0_0.vlm"},
	{0x836070BDCF5CB925ULL, "6127328_m.fxe"},
	{0x8360AB0E2F111E9AULL, "435549529.ogg"},
	{0x8360CCDDEC10C8C2ULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_act_c.anb"},
	{0x8360DE7D63E323DFULL, "vxht_bartrand_n_680_0_0.vlm"},
	{0x8360E7D7302B4F6AULL, "110308917.ogg"},
	{0x83610BAFFA03A2F6ULL, "563122724.ogg"},
	{0x8361282C9A498383ULL, "al_mage_talents.gda"},
	{0x8361AC9BFF2CCADAULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.impa.ani"},
	{0x8361FCED2EEF7F17ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_002a_0n.dds"},
	{0x8361FE885517B6C7ULL, "6093766_m.fxe"},
	{0x83624890A08FE334ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_cast_c.impa.evt"},
	{0x836265D6066C7E93ULL, "vxw_encounter3_40_0_0.vlm"},
	{0x8363EAAD7017FD6CULL, "vxlt_backalley_g_1uf_0_0.vlm"},
	{0x83648949FE9D8F81ULL, "gen00fl_aveline6002141.cl"},
	{0x836513602D9F0D1AULL, "6169871_m.fxe"},
	{0x83652A77F3AE4542ULL, "vxlt_0_lowtown_n_ep1_0_0.vlm"},
	{0x83653329CF3FDE22ULL, "496272277.ogg"},
	{0x8365F9D768CF3E87ULL, "art\\levels\\4cinematics\\dwarf\\_textures\\dwarfchair_sa.dds"},
	{0x8365FB0AE330C43CULL, "vxht_mansion2_1pl_0_0.vlm"},
	{0x83661C27D4A6E909ULL, "qcr000_arishok6173714.cl"},
	{0x83663F16EB7164A2ULL, "clf260_harley6069677.cl"},
	{0x836753CAA467354CULL, "797268375.ogg"},
	{0x836773D7B3E3AF4AULL, "dre000_varric6035135.cl"},
	{0x83679D7C24621BD0ULL, "vxca_caverns2prc_2uj_0_0.vlm"},
	{0x8367A6BF51DF7DB9ULL, "hlac_lt_0_lowtown_a3_n_c_-3_2x.dds"},
	{0x8367B933795505CFULL, "6132827_m.fxe"},
	{0x8367E7D49947ED9BULL, "6025323_m.fxe"},
	{0x8367ED4B7ABD7EABULL, "499410909.ogg"},
	{0x83680A6F75A75A72ULL, "vxca_caverns2prc_86w_0_0.vlm"},
	{0x83682686DD3801D8ULL, "6067714_f.fxe"},
	{0x8368402485E19B69ULL, "llac_ht_keep_4519_14.dds"},
	{0x83688BD70E84F20AULL, "6140816_m.fxe"},
	{0x8369C530C278EEF7ULL, "vxlt_gallowscourtyard_n_747_0_0.vlm"},
	{0x8369CDC0B01EDF36ULL, "llac_lt_0_lowtown_docks_n_-131074_4.dds"},
	{0x8369FCCC17A94BCEULL, "gen00fl_fenris6127107.cl"},
	{0x836A2355FA781414ULL, "vxlt_0_lowtown_a3_n_c_gpf_0_0.vlm"},
	{0x836A51088F492C21ULL, "dae021_fem_worshipper_sta.bnk"},
	{0x836B2F1C04848119ULL, "llac_w_encounter3_-393211_0.dds"},
	{0x836B3C83984A32F2ULL, "vxdr_primevalentrance_23m_0_0.vlm"},
	{0x836B8FF4362D2606ULL, "129222907.ogg"},
	{0x836B9AB2F2F70A35ULL, "6139143_m.fxe"},
	{0x836B9EEE318EE046ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgte_0d.dds"},
	{0x836BAB4B9757D1EAULL, "art\\vfx\\combat\\vfx_r_dwp_front_med1_c.impa.gad"},
	{0x836BC6ADF632BB93ULL, "static_static_p48.psh.fxo.dx11"},
	{0x836BD0ED6E2461DCULL, "vxw_encounter3_ba_0_0.vlm"},
	{0x836BFB3FC8027A53ULL, "gen00fl_fenris6113837.cl"},
	{0x836C4B65783DFDB1ULL, "vxlt_hangedman_n_t7_0_0.vlm"},
	{0x836CF5239F4BF13EULL, "vxlt_0_lowtown_a2_n_c_ewj_0_0.vlm"},
	{0x836D328BAC61005FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_chantrybasic_1.mao"},
	{0x836DA2ADF4B4CFB7ULL, "llac_w_sundermount_65537_0.dds"},
	{0x836DC0126B54692DULL, "vxlt_darktown_2ub_0_0.vlm"},
	{0x836DE86B3033B0A1ULL, "vxw_sundermount_373_0_1.vlm"},
	{0x836DF8D7C41E3009ULL, "vxht_mansion2b_1gx_0_0.vlm"},
	{0x836DFE20F140993AULL, "606343864.ogg"},
	{0x836E3498A1407AA9ULL, "6083619_m.fxe"},
	{0x836E3FA0501AA2F8ULL, "864785332.ogg"},
	{0x836E73088F4D4E91ULL, "dae021_fem_worshipper_sta.cnv"},
	{0x836EB82BD5037B69ULL, "vxca_caverns1_mines_gk3_0_0.vlm"},
	{0x836ED1C5094EFD47ULL, "hlac_ht_mansion_n_3369_0z.dds"},
	{0x836EE066F507FA5BULL, "6152754_f.fxe"},
	{0x836F08E76786C9FDULL, "vxw_encounter2_5f_0_0.vlm"},
	{0x836FDC513C6950EDULL, "hlac_lt_0_lowtown_a2_n_c_65533_3x.dds"},
	{0x83704090A09C136DULL, "art\\vfx\\combat\\vfx_d_qun_aoe_cast_c.impa.ani"},
	{0x8370693453C70B5DULL, "gen_im_wep_mag_sta_act2_03.uti"},
	{0x83707499A0605EB8ULL, "isa311_isabela6095370.cl"},
	{0x83708212E41A7533ULL, "vxlt_0_lowtown_a2_f4x_0_0.vlm"},
	{0x837087652C928BC4ULL, "6151045_m.fxe"},
	{0x837096D26F7CB8E8ULL, "vxlt_0_lowtown_n_e2e_0_0.vlm"},
	{0x8370C7AC63BFA1CBULL, "409989688.ogg"},
	{0x8370D7B2F8629713ULL, "vxw_encounter3_13_0_0.vlm"},
	{0x8371602F65ACB6A3ULL, "794191158.ogg"},
	{0x83718C893BEC9997ULL, "ico_armor_heavy_3.dds"},
	{0x8371F8E22086B4BCULL, "174698238.ogg"},
	{0x83720F9D1E555D72ULL, "vxht_0_hightown_a2_bs8_0_0.vlm"},
	{0x837242171FAFC4A6ULL, "character_character_p18.psh.fxo.dx11.assembly"},
	{0x8372442B6C26A0E7ULL, "x_narrativeroom_probe_2q.dds"},
	{0x83725C8A620E15ECULL, "vxlt_backalley_g_1rg_0_0.vlm"},
	{0x83728E4B975DBE81ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med1_c.impa.evt"},
	{0x8372ED4771B0D9DCULL, "vxht_0_hightown_n_brj_0_0.vlm"},
	{0x8372F278FA8BFB5AULL, "art\\vfx\\combat\\vfx_f_fen_spirit_pulse_c.impa.gad"},
	{0x8372FDCCD9FCA740ULL, "vxlt_0_lowtown_a3_n_c_e8h_0_0.vlm"},
	{0x83734FF132D5940AULL, "zz_and_debug6134170.cl"},
	{0x83735BA3B53D6E29ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_follower2_andersintro.ani"},
	{0x8373800605511C0AULL, "hlac_ht_bartrand_n_3365_1y.dds"},
	{0x8373EC8FB325B62AULL, "6033968_f.fxe"},
	{0x837457327E279D30ULL, "lt_0_lowtown_docks_a3_n_255_0.msh"},
	{0x837465327E27B54AULL, "lt_0_lowtown_docks_a3_n_255_0.mmh"},
	{0x837469327E27BB81ULL, "lt_0_lowtown_docks_a3_n_255_0.mao"},
	{0x8374A56A61914656ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww12_basic.phy"},
	{0x8374AB4162157AECULL, "650404845.ogg"},
	{0x8374F371C8C48369ULL, "art\\characters\\creatures\\textures\\cn_gut_0s.dds"},
	{0x83754898F71C5FF7ULL, "w_blightlands_73_0.mmh"},
	{0x83754C98F71C6604ULL, "w_blightlands_73_0.mao"},
	{0x83755E98F71C8569ULL, "w_blightlands_73_0.msh"},
	{0x8375858FE68B0BF8ULL, "350287623.ogg"},
	{0x8375DEF6FC793281ULL, "follower_banter6164450.cl"},
	{0x8376118CB923DA6BULL, "6151059_m.fxe"},
	{0x8376D11F55775836ULL, "lt_0_lowtown_docks_n_2_0_ll.mmh"},
	{0x8377CFFDBFBCDE7FULL, "ico_boots_heavy_5.dds"},
	{0x8378CBBA42502CABULL, "mag110_samson6018696.cl"},
	{0x8378F1A8079D91B3ULL, "vxw_bonepitdestroyed_a3_22y_0_0.vlm"},
	{0x83793FBF13112EC6ULL, "6129322_f.fxe"},
	{0x83798B437C695BC7ULL, "gen00fl_isabela_init6136107.cl"},
	{0x8379B1495E7BD04EULL, "154861122.ogg"},
	{0x8379D9FF2A637706ULL, "6175092_m.fxe"},
	{0x8379EA199A24B526ULL, "hlac_lt_0_lowtown_docks_a3_-65540_2x.dds"},
	{0x8379F578FA921F71ULL, "art\\vfx\\combat\\vfx_f_fen_spirit_pulse_c.impa.evt"},
	{0x837A3D25652CE4BCULL, "6101429_f.fxe"},
	{0x837A439D3DE9B7F1ULL, "6121919_m.fxe"},
	{0x837A461D6FDD2DADULL, "345884159.ogg"},
	{0x837A55E676EB05F9ULL, "121885992.ogg"},
	{0x837A76A3B543BAABULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_follower2_andersintro.gad"},
	{0x837ABD426A341631ULL, "vxlt_hangedman_n_s6_0_0.vlm"},
	{0x837AF61FA5B3C0C8ULL, "clf260_harley6093560.cl"},
	{0x837B2426811148A2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_lfoot.gad"},
	{0x837B97AA7F5ED895ULL, "161152630.ogg"},
	{0x837BB18FFC345560ULL, "fen221ar_base.are"},
	{0x837BB272F702B9F7ULL, "vxlt_0_lowtown_docks_a3_n_d4c_0_0.vlm"},
	{0x837BC256FACBFD92ULL, "llac_lt_hangedman_n_-1_17.dds"},
	{0x837C053B733765C6ULL, "mrl311ar_house.lst"},
	{0x837C31EF5D4A81ADULL, "vxlt_darktown_2xa_0_0.vlm"},
	{0x837C3AB1AA4BEC5DULL, "vxht_0_hightown_a2_4c9_0_0.vlm"},
	{0x837C633C18D2F815ULL, "vxca_caverns2prc_9q4_0_0.vlm"},
	{0x837CE584CB210186ULL, "hlac_lt_darktown_-1_84z.dds"},
	{0x837CFF1C8A552A8DULL, "vxht_mansion_n_3zb_0_0.vlm"},
	{0x837D1DCAC9DA7C70ULL, "477983497.ogg"},
	{0x837E36DF3F32AA72ULL, "999906286.ogg"},
	{0x837EA88FFC367BE2ULL, "fen221ar_base.ncs"},
	{0x837F6A1FB4D7B534ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_loop.ani"},
	{0x837F8E6A619B26F7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww12_basic.msh"},
	{0x837FA06A619B455AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww12_basic.mao"},
	{0x837FA46A619B4C69ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww12_basic.mmh"},
	{0x837FAEC777468CE2ULL, "vxlt_0_lowtown_docks_a3_n_d11_0_0.vlm"},
	{0x837FF93B4E929185ULL, "vxca_caverns1_ec4_0_0.vlm"},
	{0x83807225032C989CULL, "gen_im_trs_krk_low_junk2_04.uti"},
	{0x83807554C4CEE5DEULL, "vxw_woundedcoast_1ck_0_0.vlm"},
	{0x8380864B9769EED0ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med1_c.impa.ani"},
	{0x838109AB0923B75DULL, "ach_kill_dragon.dds"},
	{0x8381DE196AD54386ULL, "82253978.ogg"},
	{0x83821265EF42EB48ULL, "404849999.ogg"},
	{0x838247268117A299ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_lfoot.evt"},
	{0x83827E49CF12DBC1ULL, "dae000ip_crowd_ff3_talk1.utp"},
	{0x83829DA673E74942ULL, "qun110_petrice6018558.cl"},
	{0x8382BFC835389F47ULL, "383150449.ogg"},
	{0x8382F5C97076484EULL, "synclargecreatures1.gda"},
	{0x8383AEA258FA1B40ULL, "157994124.ogg"},
	{0x8383B131BA764379ULL, "6040562_m.fxe"},
	{0x8383BBA1787EB597ULL, "hlac_lt_darktown_-1_8y.dds"},
	{0x838456F2518FD190ULL, "6033495_m.fxe"},
	{0x83848C712542D7CFULL, "hlac_lt_undercityserialkiller_-1_18z.dds"},
	{0x8384D5ED897BD7ECULL, "402431309.ogg"},
	{0x83856CF70DE2DAB0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge.mao"},
	{0x838570F70DE2E1C3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge.mmh"},
	{0x83857EF70DE2F91DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge.msh"},
	{0x8385C08FFC3CC373ULL, "fen221ar_base.lst"},
	{0x8385E779B5294572ULL, "one200dg_solivitus6066529.cl"},
	{0x8385F5A2E32986B3ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyn_s.dds"},
	{0x83860D506D5CAD8AULL, "599919930.ogg"},
	{0x8386671FB4DDCE96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_loop.gad"},
	{0x83868F0D43C135AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_action_loop.gad"},
	{0x83869D5D9246C8AAULL, "gen00fl_isabela6090317.cl"},
	{0x838707BDDEF1AD9BULL, "art\\characters\\creatures\\ragedemon\\animation\\hv_dmg.ani"},
	{0x83870D78FA9CD240ULL, "art\\vfx\\combat\\vfx_f_fen_spirit_pulse_c.impa.ani"},
	{0x8387143B734186D1ULL, "mrl311ar_house.are"},
	{0x83871DA743BBF8DAULL, "mer240_jansen6081427.cl"},
	{0x83874DB499D804D7ULL, "w_bonepit_15_0.rml"},
	{0x838838D139AE0168ULL, "dre000pt_extra_funds.ncs"},
	{0x83891CE002C38CBDULL, "hlac_lt_backalley_g_65536_6z.dds"},
	{0x838955E848DBAD9BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_carv.phy"},
	{0x838991DADE4C9E1BULL, "vxht_0_hightown_a7j_0_0.vlm"},
	{0x8389C326F97A304FULL, "vxht_0_hightown_n_b9p_0_0.vlm"},
	{0x8389C4E3DFA8A090ULL, "960352203.ogg"},
	{0x8389E26E8A14BBBEULL, "one220_yevhen6103134.cl"},
	{0x838A38E957F94F76ULL, "vxlt_0_lowtown_a2_n_ehq_0_0.vlm"},
	{0x838A3D132FA4E69BULL, "6085307_m.fxe"},
	{0x838A5F6EF98B4398ULL, "vxlt_hangedman_n_25q_0_0.vlm"},
	{0x838A7F76F8B3C79CULL, "6112916_m.fxe"},
	{0x838A829ACB9C0205ULL, "dae000_sandal6148190.cl"},
	{0x838ACD2F7DF912CDULL, "w_woundedcoast_2_0.mmh"},
	{0x838AD12F7DF9191EULL, "w_woundedcoast_2_0.mao"},
	{0x838ADF2F7DF93153ULL, "w_woundedcoast_2_0.msh"},
	{0x838B195D6F62540FULL, "6072868_m.fxe"},
	{0x838B70F97C4B430DULL, "vxht_mansion_n_3wa_0_0.vlm"},
	{0x838BC2D11BBF32D3ULL, "vxdr_primevaltomb_865_0_0.vlm"},
	{0x838BDC11EB6B4ED3ULL, "vxht_keep_5ip_0_0.vlm"},
	{0x838BE45268DFBEBFULL, "zz_approval_debug6003615.cl"},
	{0x838C13363E5B2F7DULL, "hlac_ca_caverns2prc_5802_55y.dds"},
	{0x838C26112DDFE5B7ULL, "vxw_encounter2_camp_re_0_0.vlm"},
	{0x838CA4D36FAF0B5EULL, "586793574.ogg"},
	{0x838CACBAED947F6DULL, "vxca_caverns1_mines_jlk_0_0.vlm"},
	{0x838CE1E33988378FULL, "vxw_encounter2_camp_1p_0_0.vlm"},
	{0x838D0320DA2D39B9ULL, "668282922.ogg"},
	{0x838D3AE1CA43731AULL, "6111137_m.fxe"},
	{0x838E3AFD159E6301ULL, "vxw_encounter3_b0_0_0.vlm"},
	{0x838EABD158F79694ULL, "art\\characters\\weapons\\mauls\\w_mal_ml105.mao"},
	{0x838EB1DA6FE0A710ULL, "scaleform_sf_main_p0.vsh.fxo"},
	{0x838ECBB8CD13628AULL, "pro000st_gamlen.stg"},
	{0x838ED3C7B482C488ULL, "82807531.ogg"},
	{0x838EE731B6C4FE5EULL, "vxw_woundedcoast_1nl_0_0.vlm"},
	{0x838F425EE46B0ADAULL, "hlac_lt_0_lowtown_a2_n_c_131069_0x.dds"},
	{0x838F4F2681223A88ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_p_lfoot.ani"},
	{0x838FD101559D0ADCULL, "vxht_0_hightown_n_bxh_0_0.vlm"},
	{0x83901CB90E44EC14ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13b.phy"},
	{0x83905F59D5C4DD13ULL, "vxht_mansion_n_61h_0_0.vlm"},
	{0x8390813DAD29C09CULL, "one300_zevran6075734.cl"},
	{0x8390AF2E558D0EDCULL, "mag131_keran6086280.cl"},
	{0x8390B9FD3FC59A50ULL, "hlac_lt_0_lowtown_docks_n_-65538_7x.dds"},
	{0x839109362BC1E02CULL, "6051491_m.fxe"},
	{0x839150A1B9130DBCULL, "vxht_mansion2_1ac_0_0.vlm"},
	{0x8391C76E615330B1ULL, "6052770_m.fxe"},
	{0x83925F5496C47E28ULL, "scaleform_sf_main_p2.psh.fxo.dx11"},
	{0x83928B64CC68E8FCULL, "d15_b_shot.dds"},
	{0x83932E9551F6A38DULL, "133014739.ogg"},
	{0x83940FBDDEFC4592ULL, "art\\characters\\creatures\\ragedemon\\animation\\hv_dmg.evt"},
	{0x839431CB51025076ULL, "89774519.ogg"},
	{0x839433418AB51C7CULL, "479245580.ogg"},
	{0x8394C5E52CF669CAULL, "6076432_m.fxe"},
	{0x8394CD2F7E016746ULL, "w_woundedcoast_2_0.rml"},
	{0x83954386DD13D152ULL, "vxdr_primevaltomb_860_0_0.vlm"},
	{0x8395475923BF64D2ULL, "gen00fl_merrill6141132.cl"},
	{0x83955CC7ACBFED52ULL, "vxht_keep_5iu_0_0.vlm"},
	{0x83957F3B657ECE7AULL, "6175760_m.fxe"},
	{0x8395B58FA214762CULL, "llac_w_woundedcoast_196614_0.dds"},
	{0x83963A046A65E190ULL, "6148461_m.fxe"},
	{0x839671F70DF156E8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge.phy"},
	{0x8396AE2A8C34A99FULL, "vrc321_varric6154765.cl"},
	{0x8396E98EA22245D2ULL, "6111212_m.fxe"},
	{0x8396F1E7A38B6084ULL, "w_bonepit_60_0.mmh"},
	{0x8396F5E7A38B67D7ULL, "w_bonepit_60_0.mao"},
	{0x839707E7A38B8626ULL, "w_bonepit_60_0.msh"},
	{0x8397629852A98530ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03a.mmh"},
	{0x83976D85C87F5245ULL, "qun120cr_javaris6048961.cl"},
	{0x8397709852A99D4AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03a.msh"},
	{0x8397CA3FCE12D23AULL, "dae000_mother6099204.cl"},
	{0x83980A65C3554258ULL, "537637537.ogg"},
	{0x83986C52FFB0E327ULL, "dr_dungeon_3217.rml"},
	{0x8398AAC649EF67F6ULL, "vxlt_0_lowtown_a2_n_emr_0_0.vlm"},
	{0x8398E5574273825AULL, "vxlt_0_lowtown_docks_br2_0_0.vlm"},
	{0x83991FA91F961EC9ULL, "222606414.ogg"},
	{0x839985955A293D90ULL, "dae100_alistair_drunk6135098.cl"},
	{0x8399BA72E90E4ED9ULL, "6177854_f.fxe"},
	{0x8399CB9F457AA688ULL, "849077788.ogg"},
	{0x8399F9B90E4D0511ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13b.msh"},
	{0x839A13B90E4D31CFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13b.mmh"},
	{0x839A35D9F5B508F7ULL, "1006017930.ogg"},
	{0x839A83398FDB4BB0ULL, "character_character_p126.psh.fxo.dx11"},
	{0x839A8E51427E7D13ULL, "vxlt_0_lowtown_a2_ey0_0_0.vlm"},
	{0x839AB1DA67AD47FDULL, "vxw_blightlands_hn_0_0.vlm"},
	{0x839AD534A557DDFFULL, "vxw_encounter2_camp_54_0_0.vlm"},
	{0x839ADB01CF93FC6CULL, "lowtown_male_urchin_sta6164183.cl"},
	{0x839B2EBDDF029925ULL, "art\\characters\\creatures\\ragedemon\\animation\\hv_dmg.gad"},
	{0x839B5481B7937F2AULL, "6005719_m.fxe"},
	{0x839BAC0D43D3C375ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\death_cradle_action_loop.ani"},
	{0x839C4BD1151E4D77ULL, "6078185_m.fxe"},
	{0x839C9E2F4A1F5396ULL, "vxw_woundedcoast_159_0_0.vlm"},
	{0x839CADE64586EF4CULL, "6065809_m.fxe"},
	{0x839CDCF5327EC285ULL, "vxca_caverns1_ei2_0_0.vlm"},
	{0x839D142E6682383FULL, "vxlt_backalley_g_1rh_0_0.vlm"},
	{0x839D590EA8BB16DEULL, "vxw_woundedcoast_1im_0_0.vlm"},
	{0x839D9723DB5BC195ULL, "864605147.ogg"},
	{0x839DA14CEF6FAD4AULL, "zz_aveline_working_doc.cnv"},
	{0x839EF40083348DA5ULL, "482206515.ogg"},
	{0x839F6D062114F37BULL, "vxlt_darktown_2vi_0_0.vlm"},
	{0x839F90F90034619CULL, "103980881.ogg"},
	{0x839FA7C22574208EULL, "652236240.ogg"},
	{0x83A00E8BEA891045ULL, "vxlt_gallowscourtyard_n_4v8_0_0.vlm"},
	{0x83A06C3C58A91942ULL, "nat380_sibling6068012.cl"},
	{0x83A0E38048777009ULL, "vxlt_0_lowtown_a2_n_gnj_0_0.vlm"},
	{0x83A28007A670C573ULL, "hlac_lt_hangedman_-1_28x.dds"},
	{0x83A2F5CF0374417AULL, "6209956_f.fxe"},
	{0x83A364019FD0C4EFULL, "6059309_f.fxe"},
	{0x83A39700B38C141EULL, "art\\vfx\\creatures\\fxc_arch_smite_p.impa.evt"},
	{0x83A3A2CA21A926E7ULL, "6167483_m.fxe"},
	{0x83A3BA05A879DBA7ULL, "6125824_f.fxe"},
	{0x83A3CEE96AD1225CULL, "character_character_p43.psh.fxo.dx11"},
	{0x83A41DD67894390FULL, "vxca_caverns1_j5s_0_0.vlm"},
	{0x83A4B6052C3BA11AULL, "511370605.ogg"},
	{0x83A4FFDE92F19079ULL, "ht_0_hightownexplosion_n_16_0_ll.mmh"},
	{0x83A54DB913F37FA4ULL, "6114486_m.fxe"},
	{0x83A62B60787F1AE2ULL, "370746281.ogg"},
	{0x83A69C993718ABAAULL, "l_dre180ar_keep_fake.bnk"},
	{0x83A706B052197161ULL, "970748192.ogg"},
	{0x83A787A6D6607E76ULL, "mcr310_argument6072702.cl"},
	{0x83A7E91C9F82B93FULL, "1045675157.ogg"},
	{0x83A83259D4C486BAULL, "6175816_m.fxe"},
	{0x83A8BFCBCF577FD3ULL, "vxht_keep_5wr_0_0.vlm"},
	{0x83A902FEBFC7D031ULL, "95852918.ogg"},
	{0x83A909CB11CC16B7ULL, "vxw_encounter2_camp_hg_0_0.vlm"},
	{0x83A923B759A3607DULL, "vxw_blightlands_mm_0_0.vlm"},
	{0x83A9324661BD5955ULL, "gen00fl_isabela6118373.cl"},
	{0x83A94B2AC4F8493DULL, "vxw_encounter1_dl_0_0.vlm"},
	{0x83A950E5450643A0ULL, "qun131_taskmaster6039401.cl"},
	{0x83A9AE00B390A749ULL, "art\\vfx\\creatures\\fxc_arch_smite_p.impa.gad"},
	{0x83A9D4BCECBE984FULL, "6189443_m.fxe"},
	{0x83AA3C39BA940E44ULL, "44615266.ogg"},
	{0x83AA58223196C110ULL, "vxw_bonepitmining_a2_go_0_0.vlm"},
	{0x83AA60940BB83F16ULL, "150632487.ogg"},
	{0x83AA86D2E0C5F33AULL, "vxht_0_hightown_a3_a4s_0_0.vlm"},
	{0x83AA93D9754B89ECULL, "vxlt_undercityserialkiller_20f_0_0.vlm"},
	{0x83AADFE80A906C1FULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm05.phy"},
	{0x83AB3E73707D02C5ULL, "fx_rocka.mao"},
	{0x83AB60D9CE5E0FBAULL, "ser221_party_comments6209930.cl"},
	{0x83ABBB2F44533C64ULL, "6149592_m.fxe"},
	{0x83ABECE0804298EEULL, "mag311_keran6041405.cl"},
	{0x83AC3D0A54C6CC2FULL, "156721591.ogg"},
	{0x83AC51E23FD6AE43ULL, "6066366_m.fxe"},
	{0x83AC847AE68B4E30ULL, "931198098.ogg"},
	{0x83AC8AD1A4018EE7ULL, "dae000ip_crowd_ff2_talk2.utp"},
	{0x83AD3A46C284C27AULL, "hlac_dr_0_deeproads_5843_3y.dds"},
	{0x83AD552A9EA01CCAULL, "vxca_caverns1_mines_e83_0_0.vlm"},
	{0x83AD81EE19A64332ULL, "gen00fl_fenris6107156.cl"},
	{0x83AD9543AA54C915ULL, "194424416.ogg"},
	{0x83ADDEE3130B0BFBULL, "vxlt_darktown_2yh_0_0.vlm"},
	{0x83AE88626B4D31DBULL, "6164085_f.fxe"},
	{0x83AEF0CA1213A701ULL, "1044281658.ogg"},
	{0x83AF0AA623F538F8ULL, "36705731.ogg"},
	{0x83AF1A3F16649E8EULL, "602343147.ogg"},
	{0x83AFB104B381DF24ULL, "hlac_dr_primevalentrance_1726_8z.dds"},
	{0x83AFE491BB88434AULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.impa.gad"},
	{0x83B07F00B3967597ULL, "art\\vfx\\creatures\\fxc_arch_smite_p.impa.ani"},
	{0x83B0A0BD19323AF4ULL, "6036801_m.fxe"},
	{0x83B12EE4B31736D1ULL, "ser191_emeric6104784.cl"},
	{0x83B23B6FD38A969CULL, "294999417.ogg"},
	{0x83B26FE5FBDA86E2ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkc.msh"},
	{0x83B271E5FBDA8998ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkc.mmh"},
	{0x83B27DE5FBDA9E3BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkc.mao"},
	{0x83B2A33EC8D7E47DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldb_sa.dds"},
	{0x83B35F6EB4474D71ULL, "hf_lgt_selby.mor"},
	{0x83B3614DE629D5C7ULL, "and111_lirene6084366.cl"},
	{0x83B364FB20B0C03EULL, "zz_dae_debug6142947.cl"},
	{0x83B365446EC70D26ULL, "719493573.ogg"},
	{0x83B3705D2BB07621ULL, "528542230.ogg"},
	{0x83B374A9A5E3D211ULL, "hlac_lt_warehouse_65536_0y.dds"},
	{0x83B3DD32182A5A18ULL, "llac_lt_gallowstemplar_2383_8.dds"},
	{0x83B412253F312BC7ULL, "vxht_chantry_334_0_0.vlm"},
	{0x83B4478C4E72D498ULL, "vxht_bartrand_n_4gt_0_0.vlm"},
	{0x83B47FE33ADFEE67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium.gad"},
	{0x83B4B90655CFFA03ULL, "6175872_f.fxe"},
	{0x83B600424CA3E59AULL, "abi_seb.gda"},
	{0x83B637C3CB862116ULL, "ht_0_hightown_a3_25_0_ll.mmh"},
	{0x83B63CE84902713AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_carv.msh"},
	{0x83B64EE849028F80ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_carv.mmh"},
	{0x83B6C597F1813120ULL, "mag291_demon_marethari6029141.cl"},
	{0x83B6C791BB8E30E1ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.impa.evt"},
	{0x83B6F4BD34C9711EULL, "6167572_m.fxe"},
	{0x83B731A8C14D9853ULL, "vxht_keep_5zu_0_0.vlm"},
	{0x83B77BA803C22F37ULL, "vxw_encounter2_camp_cf_0_0.vlm"},
	{0x83B795944B9978FDULL, "vxw_blightlands_vl_0_0.vlm"},
	{0x83B7BD07B6EE61BDULL, "vxw_encounter1_ik_0_0.vlm"},
	{0x83B7EA1EACCFFF46ULL, "hlac_lt_gallowstemplar_f_2395_0x.dds"},
	{0x83B7FB7B45EFEAD9ULL, "gen00fl_varric6118805.cl"},
	{0x83B8531367B94FF1ULL, "vxht_0_hightown_a3_by8_0_0.vlm"},
	{0x83B862B365B0AAB1ULL, "gen00fl_varric6081619.cl"},
	{0x83B8A51DEFDB443AULL, "art\\characters\\playercharacter\\humanmale\\hlf_medj.phy"},
	{0x83B905B66741A26CULL, "vxlt_undercityserialkiller_27g_0_0.vlm"},
	{0x83B999FF56E60E1EULL, "6110277_m.fxe"},
	{0x83B9A5FF2DE52CB6ULL, "6048179_f.fxe"},
	{0x83BA3B4CA22CBF5EULL, "6083111_m.fxe"},
	{0x83BA3DB296041044ULL, "dae000ip_crowd_em4_talk5.utp"},
	{0x83BA74E33AE44774ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium.evt"},
	{0x83BAB7551CB177D8ULL, "vxht_0_hightown_a3_c3j_0_0.vlm"},
	{0x83BB1066DD5559C0ULL, "92159860.ogg"},
	{0x83BB8E1669055DF0ULL, "313839120.ogg"},
	{0x83BB93DF23DFACDDULL, "6118586_m.fxe"},
	{0x83BBE0E80A9EE124ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm05.mmh"},
	{0x83BBE5F4F7004AEEULL, "436162374.ogg"},
	{0x83BBF6E80A9F06C6ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm05.msh"},
	{0x83BC08EF75C9EFB6ULL, "vxca_cavern1_haunted_iml_0_0.vlm"},
	{0x83BC0CBEEBDD05A5ULL, "6035849_m.fxe"},
	{0x83BC1BFBAF8D4846ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_5y.dds"},
	{0x83BC53B80A2DBDA2ULL, "585502656.ogg"},
	{0x83BC54545148E02FULL, "art\\vfx\\creatures\\fxc_ft_dirt_l_c.mmh"},
	{0x83BC57A48F5E7AB3ULL, "352534801.ogg"},
	{0x83BC87AD8E38A333ULL, "437288638.ogg"},
	{0x83BD55CCA9EA7ABBULL, "mrl210_merrill6102143.cl"},
	{0x83BD66E6746D82C4ULL, "248864259.ogg"},
	{0x83BDDFDDB08757E7ULL, "dae200_bran6094769.cl"},
	{0x83BE08A3DA1E4312ULL, "396090629.ogg"},
	{0x83BE2AC79C589F39ULL, "878965604.ogg"},
	{0x83BE31E504D90888ULL, "847413533.ogg"},
	{0x83BEE0B37C3F35E9ULL, "32360331.ogg"},
	{0x83BF26773C6C56F1ULL, "plt_lgt310pt_fixer.plo"},
	{0x83BFFDF893D873E2ULL, "mag100_karras6035304.cl"},
	{0x83C007C28AE727A5ULL, "4890400.ogg"},
	{0x83C07617E83D2322ULL, "vxw_bonepitdestroyed_a3_ko_0_0.vlm"},
	{0x83C08304D659F7FFULL, "vxca_caverns1_j26_0_0.vlm"},
	{0x83C0A9418BA329E7ULL, "749391200.ogg"},
	{0x83C0ACEDEF0D2481ULL, "vxht_0_hightown_n_bc8_0_0.vlm"},
	{0x83C15C04CEB5CDE4ULL, "vxlt_gallowscourtyard_n_515_0_0.vlm"},
	{0x83C185900458F4B1ULL, "vxht_mansion_n_5b8_0_0.vlm"},
	{0x83C1BDF2EDF6547EULL, "vxca_caverns1_gke_0_0.vlm"},
	{0x83C1CA151F759472ULL, "one000_ilen6059261.cl"},
	{0x83C1D6CF5BCDF474ULL, "vxht_0_hightown_a2_aka_0_0.vlm"},
	{0x83C289CE02E0957BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_dksa_0d.dds"},
	{0x83C2B94C08DF8DB0ULL, "vfx_vfx_p1.vsh.fxo.dx11.assembly"},
	{0x83C2BB3035BD4E52ULL, "vxw_bonepitdestroyed_a3_1s8_0_0.vlm"},
	{0x83C2CDBDEBDF2881ULL, "6195808_m.fxe"},
	{0x83C3C3B6AA34C385ULL, "vxw_bonepit_111_0_0.vlm"},
	{0x83C465705DC5BEC0ULL, "hlac_lt_0_lowtown_docks_n_131073_0x.dds"},
	{0x83C4BF91BB9A6030ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.impa.ani"},
	{0x83C4F5A562AC9D6CULL, "613205861.ogg"},
	{0x83C546D6EDD57B0EULL, "plt_mer210pt_cave_crawling.plo"},
	{0x83C55B5EFBD32BBEULL, "279128629.ogg"},
	{0x83C5887ED62AD329ULL, "vxlt_gallowsprison_n_6a9_0_0.vlm"},
	{0x83C5B32F51B6D559ULL, "hlac_lt_undercitydungeon_-1_40x.dds"},
	{0x83C5CFACD82B4521ULL, "vxht_keep_5da_0_0.vlm"},
	{0x83C5EBDBAF6803C1ULL, "follower_barks6153727.cl"},
	{0x83C639B646ED044DULL, "vxlt_darktown_4ts_0_0.vlm"},
	{0x83C66EE7C00EA91EULL, "vrc321_party_barks6096311.cl"},
	{0x83C6C62E8B5408CDULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_005a_0t.dds"},
	{0x83C6F2CA50C43269ULL, "isa111_isabela6067077.cl"},
	{0x83C700460612B373ULL, "378883457.ogg"},
	{0x83C7033035536728ULL, "6115799_m.fxe"},
	{0x83C74D1595EE1260ULL, "6167255_m.fxe"},
	{0x83C7A1B6EB55442CULL, "82860154.ogg"},
	{0x83C81458368BD797ULL, "417634176.ogg"},
	{0x83C816E119A8BDBDULL, "6166899_m.fxe"},
	{0x83C86CE33AF076ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium.ani"},
	{0x83C87C9B1703D28CULL, "6136477_m.fxe"},
	{0x83C8AA2B2A0C4B1DULL, "gencr_coterie_thug_2h_f.utc"},
	{0x83C9154A83E040CBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedenter.evt"},
	{0x83C9A7E215EB25A9ULL, "hlac_dr_slavers_3217_6z.dds"},
	{0x83CA05BC15D5C55DULL, "6117486_m.fxe"},
	{0x83CA88D95935B912ULL, "vxlt_0_lowtown_a3_n_c_hxo_0_0.vlm"},
	{0x83CA9EA56DBC8202ULL, "zz_followers_banter6158304.cl"},
	{0x83CC3EF2F10C0ED9ULL, "mag110_samson6063960.cl"},
	{0x83CC76435A68A3CAULL, "zz_isa_debug6175456.cl"},
	{0x83CCA325BB5F366BULL, "gencr_street_thug_2h_f.utc"},
	{0x83CD1F53F0E608FEULL, "6219125_m.fxe"},
	{0x83CD405D4614940BULL, "vxht_keep_6gu_0_0.vlm"},
	{0x83CD52CE52E28415ULL, "153388210.ogg"},
	{0x83CDF8A553F46EC8ULL, "llac_lt_0_lowtown_a3_n_c_-2_2.dds"},
	{0x83CEC546FF347C74ULL, "384949029.ogg"},
	{0x83CEDDA59E5CAF78ULL, "vxlt_0_lowtown_docks_a3_coe_0_0.vlm"},
	{0x83CEF9D3A0FE3EDBULL, "qcr000_aveline6054359.cl"},
	{0x83CF72F4AFED5458ULL, "hlac_w_bonepitmining_a2_458757_0z.dds"},
	{0x83CFAD31E0CDBA80ULL, "plt_dae000pt_talkedto.plo"},
	{0x83CFB5D4C6D40C34ULL, "6092151_m.fxe"},
	{0x83CFC5C36B8C31D6ULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_hed_gib_0n.dds"},
	{0x83D01ABBDFDBB305ULL, "264972645.ogg"},
	{0x83D01FDE67DC73FBULL, "ts_apostates_arm_props.gda"},
	{0x83D02E4A83E6898CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedenter.gad"},
	{0x83D0407CEA7F3DECULL, "222774475.ogg"},
	{0x83D07CB108D9DEFBULL, "w_encounter3_132_0.mmh"},
	{0x83D08253E27417DCULL, "follower_barks6153889.cl"},
	{0x83D088B108D9F298ULL, "w_encounter3_132_0.mao"},
	{0x83D09AB108DA11E5ULL, "w_encounter3_132_0.msh"},
	{0x83D107CB0470674DULL, "ave100_donnic6105954.cl"},
	{0x83D119B3B9B42BB1ULL, "mcr310_argument6072686.cl"},
	{0x83D11EEAF5AE24C4ULL, "6178440_m.fxe"},
	{0x83D12B46325F0598ULL, "vxht_bartrand_n_4qr_0_0.vlm"},
	{0x83D17E0CFA5EAC2CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushc.mmh"},
	{0x83D17F5CC12E6F82ULL, "6206868_m.fxe"},
	{0x83D1840CFA5EB61EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushc.msh"},
	{0x83D1CF1E6122FCACULL, "llac_w_sundermount_262148_0.dds"},
	{0x83D1E09177C606E3ULL, "6144098_f.fxe"},
	{0x83D220CA87048F20ULL, "gen00fl_fenris6107217.cl"},
	{0x83D230410F0A8924ULL, "hlac_lt_gallowstemplar_2388_5x.dds"},
	{0x83D27A5BA142AEE7ULL, "w_bonepitdestroyed_a3_37_0_ll.mmh"},
	{0x83D3490CEDFCFA19ULL, "bld_mac_lopa.mao"},
	{0x83D397503550F259ULL, "202050530.ogg"},
	{0x83D3C048269CF1CCULL, "dae000_gamlen6022564.cl"},
	{0x83D3D5A6D2B1DB25ULL, "294607483.ogg"},
	{0x83D42EE397605220ULL, "vxlt_shop_e3_0_0.vlm"},
	{0x83D45E50B45E9BC4ULL, "366550658.ogg"},
	{0x83D4AEC75E4E4561ULL, "271933679.ogg"},
	{0x83D4B78569ED214DULL, "497501139.ogg"},
	{0x83D4BD45EC4BF6A0ULL, "vxlt_0_lowtown_docks_n_b66_0_1.vlm"},
	{0x83D4D63FA35283B6ULL, "855585075.ogg"},
	{0x83D4F47385A6F66DULL, "gen00fl_merrill6128697.cl"},
	{0x83D51A99E5761F1FULL, "6104754_m.fxe"},
	{0x83D5C4DF1AB06A3BULL, "hlac_lt_0_lowtown_a2_n_131070_5z.dds"},
	{0x83D62D4A83EAF38AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedenter.ani"},
	{0x83D65E286BB84E2EULL, "6083398_f.fxe"},
	{0x83D6E79A796D1059ULL, "hlac_w_blightlands_262143_0x.dds"},
	{0x83D6EE7808F9EB0CULL, "6131577_m.fxe"},
	{0x83D7115BF8BF9316ULL, "hlac_w_bonepit_1_0y.dds"},
	{0x83D756303B459923ULL, "and121_karl6094002.cl"},
	{0x83D7699852DF754BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03a.phy"},
	{0x83D7877DC332C1D3ULL, "vxca_cavern1_haunted_i19_0_0.vlm"},
	{0x83D78ABD16074284ULL, "gen00fl_anders6114439.cl"},
	{0x83D79B60CCEB9ACCULL, "vxlt_0_lowtown_docks_ars_0_0.vlm"},
	{0x83D7FB567A35D981ULL, "53135541.ogg"},
	{0x83D848699554B20BULL, "6114873_f.fxe"},
	{0x83D84BED2D1FE87CULL, "follower_banter6173314.cl"},
	{0x83D8D08B061103C7ULL, "scrub_config.gda"},
	{0x83D8E94AF7B1517AULL, "hlac_lt_0_lowtown_docks_-196610_4z.dds"},
	{0x83D8ECA959B620B6ULL, "vxca_cavern1_haunted_ign_0_0.vlm"},
	{0x83D93720ADE47BE5ULL, "vxca_caverns2prc_9fa_0_0.vlm"},
	{0x83D96F52FFE87F3BULL, "dr_dungeon_3217.anb"},
	{0x83D98E0BBCA0DF49ULL, "dre000_vampire6034402.cl"},
	{0x83D9D3B9E9898454ULL, "6071505_m.fxe"},
	{0x83DA37D6EDE7BD9DULL, "plt_mer210pt_cave_crawling.nss"},
	{0x83DA9A9C7A5B7FA7ULL, "w_blightlands_16_0.rml"},
	{0x83DACBD2E1CE8E61ULL, "6155871_m.fxe"},
	{0x83DBA4DE355F35FBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_ceiling.mmh"},
	{0x83DBB0DE355F4998ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_ceiling.mao"},
	{0x83DBC2DE355F68E5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_ceiling.msh"},
	{0x83DBC6B1C4CC068FULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_0z.dds"},
	{0x83DCA668F3B4CCC5ULL, "gen00fl_aveline6106876.cl"},
	{0x83DCAFE9CB553842ULL, "32592164.ogg"},
	{0x83DCCA938BD2157BULL, "152333242.ogg"},
	{0x83DCF238850BD932ULL, "534759232.ogg"},
	{0x83DD2C16670D5891ULL, "27492544.ogg"},
	{0x83DDA298AFA07945ULL, "87115395.ogg"},
	{0x83DE3269D38140B7ULL, "731734301.ogg"},
	{0x83DE3AD1177D3B44ULL, "vxht_0_hightown_a3_c1l_0_0.vlm"},
	{0x83DED6B624A202F6ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb106a.mmh"},
	{0x83DEE0B624A213F4ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb106a.msh"},
	{0x83DF316B62BB4C50ULL, "tables_n.dds"},
	{0x83DF52D7FCBCE405ULL, "676818467.ogg"},
	{0x83DFCDCBC3F2FBEEULL, "ave331_jeven6151386.cl"},
	{0x83E007003BB6DFC2ULL, "vxw_bonepitmining_a2_ms_0_0.vlm"},
	{0x83E00920C593A6F7ULL, "vxht_0_hightown_a3_7iw_0_0.vlm"},
	{0x83E00E4DE912831EULL, "gen00fl_varric6107282.cl"},
	{0x83E024A33E515B95ULL, "412310112.ogg"},
	{0x83E121C3C9E3957FULL, "llac_w_bonepitmining_a2_65538_0.dds"},
	{0x83E1334190C97FBFULL, "vxw_bonepitmining_a2_25z_0_0.vlm"},
	{0x83E1B6C682F039F0ULL, "mag110tr_arianni_reset.utt"},
	{0x83E1C3AED4B5933BULL, "6142085_f.fxe"},
	{0x83E1D050E3D9C957ULL, "6179375_f.fxe"},
	{0x83E265FC6DA660EBULL, "altar_sa.dds"},
	{0x83E26679E4FDE76CULL, "t1_skn_h04.tnt"},
	{0x83E29530C7A8EDBDULL, "mag110_ws_banter6079655.cl"},
	{0x83E2ABDC4885491EULL, "vxlt_0_lowtown_docks_btp_0_0.vlm"},
	{0x83E2C3B805734C0AULL, "40750246.ogg"},
	{0x83E2D08AFCB265D4ULL, "vxht_0_hightown_n_c1r_0_0.vlm"},
	{0x83E32599FE050389ULL, "243097628.ogg"},
	{0x83E33C862BA34E03ULL, "92989949.ogg"},
	{0x83E34ACB18499362ULL, "vxht_keep_63d_0_0.vlm"},
	{0x83E3D4E615E4F378ULL, "vxlt_gamlenshouse_n_ka_0_0.vlm"},
	{0x83E3EF129310737DULL, "115305686.ogg"},
	{0x83E407C1421DC4DEULL, "hlac_lt_0_lowtown_docks_a3_1_0z.dds"},
	{0x83E4306F4A9A2C29ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_player_meredithsintro.ani"},
	{0x83E471BAC74A1339ULL, "llac_lt_0_lowtown_a3_n_c_-1_8.dds"},
	{0x83E4B93BFEDC3D1DULL, "272609931.ogg"},
	{0x83E51E6F790AA8BAULL, "384171455.ogg"},
	{0x83E55554516BE76CULL, "art\\vfx\\creatures\\fxc_ft_dirt_l_c.anb"},
	{0x83E65D20B9324513ULL, "mag100_cullen6043932.cl"},
	{0x83E68C604B4A6EF2ULL, "6072102_m.fxe"},
	{0x83E7C0E2473D0476ULL, "vxht_mansion2b_3gm_0_0.vlm"},
	{0x83E7E91488B299D8ULL, "hlac_lt_0_lowtown_a2_65537_2y.dds"},
	{0x83E7F4368EF2B9E9ULL, "vxw_bonepitmining_a2_el_0_0.vlm"},
	{0x83E801AB4499C977ULL, "dae000_arishok6067840.cl"},
	{0x83E8C02354C13E16ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapbase.mao"},
	{0x83E8CC2354C152B5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapbase.mmh"},
	{0x83E8CE2354C1562BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapbase.msh"},
	{0x83E91B4595FB9CE6ULL, "gen00fl_fenris6126548.cl"},
	{0x83E9E179DD328E72ULL, "one110_gustav6089804.cl"},
	{0x83EA0B33F74E36F2ULL, "vxca_caverns2prc_8cx_0_0.vlm"},
	{0x83EA24172A00C50BULL, "vxht_keep_6mw_0_0.vlm"},
	{0x83EB4B6F4AA078ABULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_player_meredithsintro.gad"},
	{0x83EBEFA82B9AE525ULL, "hlac_lt_foundry_n_320_17x.dds"},
	{0x83EC79876604DA77ULL, "dae000ip_crowd_mf7_talk17.utp"},
	{0x83ECB6D89DD5731CULL, "6081202_f.fxe"},
	{0x83ECB832D720FEEBULL, "vxlt_gallowstemplar_3jd_0_0.vlm"},
	{0x83ECFDEB676B12F0ULL, "6042364_f.fxe"},
	{0x83EDC31FCC295B34ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appp_0s.dds"},
	{0x83EE8E7E55727A61ULL, "44049747.ogg"},
	{0x83EEA44A4AAC692FULL, "986116930.ogg"},
	{0x83EF1DBDA3997A30ULL, "dae200_cullen6140339.cl"},
	{0x83EF3133194FA03DULL, "llac_lt_0_lowtown_docks_n_131071_1.dds"},
	{0x83EF333AE8730F62ULL, "ser100_party_comments6108248.cl"},
	{0x83EF66C252B75281ULL, "795096490.ogg"},
	{0x83EF7B29B86CC82DULL, "lt_0_lowtown_docks_a3_134_0.mmh"},
	{0x83EF7F29B86CCF7EULL, "lt_0_lowtown_docks_a3_134_0.mao"},
	{0x83EF8608C994643FULL, "fx_fl_smoke_blend.dds"},
	{0x83EF8D29B86CE6B3ULL, "lt_0_lowtown_docks_a3_134_0.msh"},
	{0x83EFF1F0A866F970ULL, "lowtown_fem_alienage1_sta6162528.cl"},
	{0x83F0714D04CFA0B7ULL, "6050107_m.fxe"},
	{0x83F105D59BA46C40ULL, "6206633_m.fxe"},
	{0x83F150A381574BADULL, "vxlt_undercityencounter_1rz_0_0.vlm"},
	{0x83F19A8041BB1F38ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.anb"},
	{0x83F1E3FC3C09EC04ULL, "hlac_w_bonepitdestroyed_a3_196613_0z.dds"},
	{0x83F24220CCBEE0E4ULL, "103808789.ogg"},
	{0x83F29A9C7A7011C4ULL, "w_blightlands_16_0.mmh"},
	{0x83F29E9C7A701817ULL, "w_blightlands_16_0.mao"},
	{0x83F2B09C7A703766ULL, "w_blightlands_16_0.msh"},
	{0x83F2DCD5360B7522ULL, "71144151.ogg"},
	{0x83F2DE41E2720566ULL, "vxht_0_hightown_577_0_0.vlm"},
	{0x83F3BBCA3F6B47FBULL, "hlac_lt_0_lowtown_docks_a3_n_131071_4x.dds"},
	{0x83F3F4BA72AB767CULL, "45410033.ogg"},
	{0x83F4554096861750ULL, "868979320.ogg"},
	{0x83F4592885C2BA36ULL, "6076220_f.fxe"},
	{0x83F4EC9E6E01F1A4ULL, "hm_samson.mor"},
	{0x83F5056CF476AAB2ULL, "vxca_caverns2_7en_0_2.vlm"},
	{0x83F5DE702F180292ULL, "vxlt_0_lowtown_a3_n_c_hoj_0_0.vlm"},
	{0x83F60D1BE1EC5D75ULL, "lightni_5.dds"},
	{0x83F61ADA91D0ACE5ULL, "vxca_caverns2prc_9ho_0_0.vlm"},
	{0x83F67B58577C4819ULL, "671748789.ogg"},
	{0x83F67E57BA04A5B4ULL, "vxht_brothel_n_5k5_0_0.vlm"},
	{0x83F710E93EBBA918ULL, "mrl000_marethari6129290.cl"},
	{0x83F79CE516058ADAULL, "vxlt_0_lowtown_a3_n_c_ehr_0_0.vlm"},
	{0x83F8AB2A9055CF64ULL, "231797937.ogg"},
	{0x83F96B8F466B80A5ULL, "6050152_m.fxe"},
	{0x83F9B8AEE3BE3203ULL, "llac_lt_gallowstemplar_2392_1.dds"},
	{0x83F9FD986599D8DCULL, "6147169_m.fxe"},
	{0x83F9FE27C1611F62ULL, "vxlt_gallowstemplar_f_43r_0_0.vlm"},
	{0x83FA6FBD565EF726ULL, "gen00fl_anders6125349.cl"},
	{0x83FA81C7E45CB341ULL, "196148280.ogg"},
	{0x83FA8AB951980F77ULL, "hlac_lt_gallowstemplar_2383_4y.dds"},
	{0x83FB07E8F0E10000ULL, "267203512.ogg"},
	{0x83FB0A412645947BULL, "hm_comt_butler.mor"},
	{0x83FB7D195B9FBB0EULL, "dae150_radha6148943.cl"},
	{0x83FBD63B1400C7A0ULL, "bink.mao"},
	{0x83FBD63B1400C7BBULL, "bink.mat"},
	{0x83FBEEC0E8A1619FULL, "lt_0_lowtown_docks_151_0.msh"},
	{0x83FBF4C0E8A16C41ULL, "lt_0_lowtown_docks_151_0.mmh"},
	{0x83FC00C0E8A180E2ULL, "lt_0_lowtown_docks_151_0.mao"},
	{0x83FC12D327FB4058ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potc_n.dds"},
	{0x83FC261F2D46B114ULL, "qun221_petrice6082679.cl"},
	{0x83FC460F31495836ULL, "hlac_lt_0_lowtown_a2_65536_4z.dds"},
	{0x83FC695EBFB7BAB9ULL, "vxlt_0_lowtown_docks_bs2_0_0.vlm"},
	{0x83FCAA7510EE1A8BULL, "6060118_f.fxe"},
	{0x83FCC90AA944B450ULL, "669697476.ogg"},
	{0x83FCD5472A1E5D65ULL, "220493117.ogg"},
	{0x83FD57EDF92A6BDBULL, "6217342_m.fxe"},
	{0x83FD6309F8C19B15ULL, "sky_motionvectorssky_p0.vsh.fxo.dx11.assembly"},
	{0x83FDE1E176B000DEULL, "zz_testspiders2.ncs"},
	{0x83FE0BF96296AC28ULL, "vxlt_undercityencounter_wg_0_0.vlm"},
	{0x83FE6076AB12A1C5ULL, "664081943.ogg"},
	{0x83FEC8BAA66AA1F1ULL, "6020098_m.fxe"},
	{0x83FF0F4810810B72ULL, "473556080.ogg"},
	{0x83FF65DE6BDC6A3AULL, "6088288_m.fxe"},
	{0x83FFA80D756000C3ULL, "w_encounter1_14_0.mao"},
	{0x83FFAC0D75600710ULL, "w_encounter1_14_0.mmh"},
	{0x83FFB881B753EDACULL, "zz_one_debug6051416.cl"},
	{0x83FFBA0D75601F2AULL, "w_encounter1_14_0.msh"},
	{0x83FFC19ECC995BD3ULL, "967381409.ogg"},
	{0x83FFE6BB4973A749ULL, "follower_barks6153921.cl"},
	{0x840010CB594FB866ULL, "6106651_f.fxe"},
	{0x8400422173DF3450ULL, "6135973_m.fxe"},
	{0x840044A59F8C7CE0ULL, "6129239_m.fxe"},
	{0x84007429B87B2FEAULL, "lt_0_lowtown_docks_a3_134_0.phy"},
	{0x8400C240EC4E34A3ULL, "lt_0_lowtown_25_0_ll.mmh"},
	{0x8400DA508750E679ULL, "25744416.ogg"},
	{0x84010D2F60228C92ULL, "6016889_m.fxe"},
	{0x84011AB645D48262ULL, "496656631.ogg"},
	{0x8401534DBB0FC055ULL, "vxw_encounter2_ct_0_0.vlm"},
	{0x84015391ADC6AA36ULL, "6172747_m.fxe"},
	{0x84015DEA4F5B0DCDULL, "llac_lt_darktown_-1_14.dds"},
	{0x8402459CD458597EULL, "vxlt_0_lowtown_docks_a3_d0a_0_0.vlm"},
	{0x840308FCFBDC3211ULL, "vxlt_0_lowtown_a3_n_c_fiy_0_0.vlm"},
	{0x8403A50637A30C02ULL, "vxht_0_hightown_n_aib_0_0.vlm"},
	{0x840421A56A5AE9E8ULL, "dae000_quintus6154732.cl"},
	{0x840456575EB2ECFFULL, "676402527.ogg"},
	{0x8404690C746BDCCCULL, "257741808.ogg"},
	{0x8404A1E899371420ULL, "vxlt_gallowscourtyard_787_0_0.vlm"},
	{0x8404ACE0A2C93055ULL, "6140583_m.fxe"},
	{0x84051677EC0D8477ULL, "hlac_w_bonepitdestroyed_a3_393219_0x.dds"},
	{0x84056F4E8A939998ULL, "6049673_f.fxe"},
	{0x8405AEA516D76CDFULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar4_.gad"},
	{0x8405CF5EB6962753ULL, "6117109_m.fxe"},
	{0x8405D6DCB7BE4D9BULL, "591032315.ogg"},
	{0x8406320F76F2F226ULL, "ave310_brennan6137222.cl"},
	{0x84065188C48834DDULL, "6077065_m.fxe"},
	{0x84068F3B1A0504A9ULL, "vxlt_backalley_n_1vn_0_1.vlm"},
	{0x8406A5C6830F38ADULL, "mag110tr_arianni_reset.ncs"},
	{0x8407B0D2A264964BULL, "6164824_m.fxe"},
	{0x8407B802E1F6454FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_3.pwk"},
	{0x84081CA60977919DULL, "vxht_0_hightown_a3_9zb_0_0.vlm"},
	{0x8408B7407969896CULL, "304145278.ogg"},
	{0x8408D26E6CC49552ULL, "vxdr_primevaltomb_8n8_0_0.vlm"},
	{0x8408F1CB3561EFC8ULL, "6081652_m.fxe"},
	{0x8409E8A2E6D24B1DULL, "hlac_ht_playershouse_n_855_13x.dds"},
	{0x840AADD6D3767997ULL, "hlac_dr_0_deeproads_5838_3z.dds"},
	{0x840AB46A398ED3D0ULL, "hlac_ht_mansion_n_2995_0y.dds"},
	{0x840B58449284FB5DULL, "hlac_lt_hangedman_-1_21y.dds"},
	{0x840BD40C021BEBFFULL, "1062950354.ogg"},
	{0x840BEC8AECE7DD83ULL, "6088228_f.fxe"},
	{0x840BF1789AB1B189ULL, "vxlt_0_lowtown_docks_n_c2k_0_0.vlm"},
	{0x840C398181D5F918ULL, "hlac_lt_gallowscourtyard_n_-131071_1z.dds"},
	{0x840C438FEC6B9FA2ULL, "vxlt_0_lowtown_docks_a3_n_aqx_0_0.vlm"},
	{0x840C5582305C5834ULL, "946125056.ogg"},
	{0x840C78164FD17C3BULL, "demo_weapon_swap6207141.cl"},
	{0x840CB6FF6089363AULL, "vxht_0_hightown_a3_bv2_0_0.vlm"},
	{0x840D6C40E69CFBCFULL, "zz_mcr_debug6098522.cl"},
	{0x840DD7BB5DD282BAULL, "6083388_m.fxe"},
	{0x840E01731E67929EULL, "vxlt_0_lowtown_docks_bes_0_0.vlm"},
	{0x840E9708C9AF31D1ULL, "fx_fl_smoke_blend.mao"},
	{0x840EE4C6297025A4ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgta.msh"},
	{0x840EF6C62970440DULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgta.mao"},
	{0x840EFAC629704B46ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgta.mmh"},
	{0x840F03C1ECCAAEF8ULL, "dae300_alistair_king6134931.cl"},
	{0x840F2E37E04421A8ULL, "6133571_m.fxe"},
	{0x840F7273E043C233ULL, "6059396_f.fxe"},
	{0x84107E9134401CD5ULL, "6081676_f.fxe"},
	{0x841131ECD76A16C1ULL, "vxht_mansion2b_2jx_0_0.vlm"},
	{0x8411A74F54D05B5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_damage_reactions.anb"},
	{0x8411A77429BAC92FULL, "dae000_bartrand6030674.cl"},
	{0x841259C267B2A071ULL, "art\\vfx\\creatures\\fxc_arch_dblt_p.anb"},
	{0x841268FBD74AEAB0ULL, "6187802_m.fxe"},
	{0x84131B3A13E8B8B7ULL, "zz_dae_debug6029081.cl"},
	{0x841349CD64D50369ULL, "vxw_bonepitmining_a2_jq_0_0.vlm"},
	{0x84144107091F534BULL, "l_env_wal_hightown_int_brothel_laugh.wbd"},
	{0x84144CBB9B673684ULL, "237365075.ogg"},
	{0x84148A6BED6CED78ULL, "6035916_m.fxe"},
	{0x84149E1DF0299763ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medj.msh"},
	{0x8414AC1DF029AEBDULL, "art\\characters\\playercharacter\\humanmale\\hlf_medj.mmh"},
	{0x8414BC34AA9503AAULL, "hlac_lt_0_lowtown_n_65535_0x.dds"},
	{0x84154163F6773094ULL, "229755064.ogg"},
	{0x84157D895D2518D2ULL, "one100_party_comments6106774.cl"},
	{0x8415FC9EA365C888ULL, "hlac_ht_0_hightown_a3_-65536_0x.dds"},
	{0x841602EDF9EA8FECULL, "vxca_cavern1_haunted_j21_0_1.vlm"},
	{0x84162CFEA7D034E3ULL, "rdr111dg_assistant6055112.cl"},
	{0x841665364B60DA08ULL, "scaleform_sf_depth_p0.vsh.fxo"},
	{0x8416ADAF101271D5ULL, "vxlt_undercitydungeon_1j3_0_0.vlm"},
	{0x8417088337120A8AULL, "22824082.ogg"},
	{0x841731773CB6DF56ULL, "plt_lgt310pt_fixer.nss"},
	{0x8417338D0E3D8EC4ULL, "zz_mag_debug6039844.cl"},
	{0x841738BDE0FC9FDAULL, "vis121_saemus6099489.cl"},
	{0x8417878F697EB44EULL, "vrc221_hugin6042798.cl"},
	{0x8417885039E7245CULL, "vxlt_0_lowtown_a2_ewy_0_0.vlm"},
	{0x84179DB109161930ULL, "w_encounter3_132_0.phy"},
	{0x8417E5D7656D95A7ULL, "vxht_chantry_2px_0_0.vlm"},
	{0x8418A4D610561513ULL, "hlac_lt_undercitydungeon_-1_2x.dds"},
	{0x8418A74C9F56B6AEULL, "art\\characters\\playercharacter\\humanmale\\bdy_masd.phy"},
	{0x8418EBF7E3CC55F6ULL, "752724616.ogg"},
	{0x841948B321A1E3CFULL, "203077299.ogg"},
	{0x84197D613BB9DE51ULL, "135746238.ogg"},
	{0x84198BA516E7DA25ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar4_.ani"},
	{0x8419B213F21AE833ULL, "149481448.ogg"},
	{0x8419F1C6297A4311ULL, "art\\characters\\playercharacter\\dwarfmale\\hoo_lgta.phy"},
	{0x841A5E67DEDDF603ULL, "6083328_f.fxe"},
	{0x841A998041DE22D3ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.mmh"},
	{0x841A9DFDFC2B84ABULL, "vxlt_darktown_2v9_0_0.vlm"},
	{0x841B4CB7838047F1ULL, "vxht_mansion2_2ii_0_0.vlm"},
	{0x841B9D9C7A931BBBULL, "w_blightlands_16_0.anb"},
	{0x841BFDD46E3FFE38ULL, "143262800.ogg"},
	{0x841C5165CC041A81ULL, "vxht_0_hightown_n_9uq_0_0.vlm"},
	{0x841CA9C782856122ULL, "vxht_bartrand_n_5w7_0_0.vlm"},
	{0x841D2B529CCD8468ULL, "6172499_m.fxe"},
	{0x841D54952A2EEE4DULL, "ca_cave_2194_ll.mmh"},
	{0x841D655AC8CED43FULL, "6075296_m.fxe"},
	{0x841D7448FE096DA5ULL, "vxht_mansion2b_3e4_0_0.vlm"},
	{0x841E37079EFBF155ULL, "vxw_encounter2_ir_0_0.vlm"},
	{0x841E4F20C054936AULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.impa.gad"},
	{0x841E8273254212E5ULL, "zz_isa_debug6133950.cl"},
	{0x841ECC77AC61FD9CULL, "vxlt_0_lowtown_a2_n_f4t_0_0.vlm"},
	{0x841F1560065ACBB8ULL, "hlac_ht_0_hightown_n_0_3y.dds"},
	{0x841FD26DA6982DE6ULL, "vxlt_0_lowtown_a2_n_hcy_0_0.vlm"},
	{0x8420034A211C1C4EULL, "vxht_0_hightown_n_a6c_0_0.vlm"},
	{0x84207AC267BF15EAULL, "art\\vfx\\creatures\\fxc_arch_dblt_p.mmh"},
	{0x84209E548442BD85ULL, "vxlt_gamlenshouse_n_ng_0_0.vlm"},
	{0x8420F427112D91DFULL, "vxht_mansion2_25c_0_0.vlm"},
	{0x84219CCEED35C40BULL, "vxlt_0_lowtown_e83_0_0.vlm"},
	{0x8421AEFB541BB024ULL, "llac_lt_0_lowtown_131070_3.dds"},
	{0x8422C5DE359B7030ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_ceiling.phy"},
	{0x8422CAB2BF6EF4A5ULL, "547101577.ogg"},
	{0x8423A797A4423A21ULL, "6136692_m.fxe"},
	{0x8423C151A130947AULL, "lt_0_lowtown_docks_a3_146_0.mmh"},
	{0x8423C551A1309AB1ULL, "lt_0_lowtown_docks_a3_146_0.mao"},
	{0x8423D351A130B2C0ULL, "lt_0_lowtown_docks_a3_146_0.msh"},
	{0x8423ECD7C05C946EULL, "vxlt_gallowscourtyard_6wa_0_0.vlm"},
	{0x84240A88157B7271ULL, "dae000ip_crowd_lm2_talk16.utp"},
	{0x84240E9471D42E80ULL, "dae000_gen_player_turn_in6177860.cl"},
	{0x84241B274CAB86BDULL, "ser221ip_barrier_mansion.ncs"},
	{0x842420920281BE29ULL, "hlac_lt_0_lowtown_a3_n_c_65534_1x.dds"},
	{0x84247DA34D314C68ULL, "48556091.ogg"},
	{0x8424FD9498BE1742ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.cess.gad"},
	{0x84253220C05A8001ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.impa.evt"},
	{0x842546F6A733461CULL, "dae021_keep_noble3.utc"},
	{0x84259C3A8B0E81FCULL, "675918255.ogg"},
	{0x8425B55F7C452CC5ULL, "qcr000_arishok6042483.cl"},
	{0x8425D24414187A41ULL, "mcr370_cut_player_rules6171814.cl"},
	{0x8425DB60F6119002ULL, "follower_barks6153458.cl"},
	{0x8425FF151294058AULL, "1021999302.ogg"},
	{0x84261E9AFA3B158FULL, "vxca_caverns1_jfz_0_0.vlm"},
	{0x84267FFF16C42F71ULL, "16268770.ogg"},
	{0x8426B04FBD2AD095ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggera_sa.dds"},
	{0x8426CE298F1123A6ULL, "dae000_hate_follower_react6103244.cl"},
	{0x84270D5C502B5BE9ULL, "521750779.ogg"},
	{0x84273BB8FDECA02CULL, "6111049_m.fxe"},
	{0x842772B62083AB11ULL, "6113921_m.fxe"},
	{0x84279A31E118100FULL, "plt_dae000pt_talkedto.nss"},
	{0x8427F5C0E8C6A8BEULL, "lt_0_lowtown_docks_151_0.phy"},
	{0x84280A4B9F7C1ABBULL, "pro000_meeran6050110.cl"},
	{0x84282F7C8C2458A5ULL, "ch1.psh.fxo"},
	{0x842831F3EF20621FULL, "6085054_f.fxe"},
	{0x842868C2C1991D03ULL, "hlac_w_woundedcoast_65542_0z.dds"},
	{0x84294A0B3573C69EULL, "vxw_bonepitdestroyed_a3_1wx_0_0.vlm"},
	{0x84298B9D60DA028BULL, "zz_kneelssister_v2.cut"},
	{0x8429EC969D9385C3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_clenchfists.ani"},
	{0x842A16601C2BDF3CULL, "379050958.ogg"},
	{0x842A47171F2B45A8ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_min_c.anb"},
	{0x842A74BA9B7051CCULL, "plotactionstates.gda"},
	{0x842AC5DBCEDBF4B1ULL, "llac_ca_caverns2_5802_64.dds"},
	{0x842B1BA4747B79A2ULL, "vxht_bartrand_n_5z0_0_0.vlm"},
	{0x842B7BF6033A6674ULL, "lt_0_lowtown_a3_n_c_32_0_lc.mmh"},
	{0x842C209498C47139ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.cess.evt"},
	{0x842C92DDBD90CBC3ULL, "vxlt_gallowscourtyard_n_75l_0_0.vlm"},
	{0x842D26394C462F38ULL, "vxlt_darktown_2u3_0_0.vlm"},
	{0x842D28BC68BD335BULL, "6149388_m.fxe"},
	{0x842D2A66FD28D396ULL, "6166409_m.fxe"},
	{0x842D59702B740E3FULL, "808822490.ogg"},
	{0x842D5C040BB392AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_7_steps.gad"},
	{0x842E91F34CC959F2ULL, "qcr000_arishok6196970.cl"},
	{0x842E9E82B0BAC411ULL, "325581764.ogg"},
	{0x842EED776FFA2650ULL, "6088292_f.fxe"},
	{0x842F054B8DBF9898ULL, "6060114_m.fxe"},
	{0x842F08AA83B35449ULL, "hlac_ht_chantry_n_2534_28y.dds"},
	{0x842F10317638D605ULL, "vxlt_gamlenshouse_n_af_0_0.vlm"},
	{0x842F2C5C312491A6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.down.ani"},
	{0x842F4533B05A6F1FULL, "art\\vfx\\combat\\vfx_gib_insect_bld_p.impa.gad"},
	{0x842F8C686174223EULL, "vxht_bartrand_n_4k6_0_0.vlm"},
	{0x842F8F4B26EC58B6ULL, "one110_meeran6089491.cl"},
	{0x842FE501F9C6BAD8ULL, "vxht_keep_4hy_0_0.vlm"},
	{0x84300E57D00E6FE8ULL, "w_encounter2_35_0.msh"},
	{0x84302057D00E8E79ULL, "w_encounter2_35_0.mao"},
	{0x84302C57D00EA212ULL, "w_encounter2_35_0.mmh"},
	{0x84308D62DDE6E514ULL, "w_encounter2_7_0.phy"},
	{0x843155E437D8473CULL, "ico_1h_sword_2.dds"},
	{0x8431AD279A14A59EULL, "vxht_keep_n_4hp_0_0.vlm"},
	{0x8431DE0EC2C05B78ULL, "vxht_playershouse_n_1ln_0_0.vlm"},
	{0x84321CBF2ECDA3CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_twitch.ani"},
	{0x84332A20C066B050ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.impa.ani"},
	{0x84332EB7C50AA8F5ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_p.mmh"},
	{0x8433524BF103D598ULL, "zz_approval_debug6003957.cl"},
	{0x843462D3EC577C29ULL, "989524105.ogg"},
	{0x8434685241A7ECFDULL, "w_bonepitmining_a2_20_0.rml"},
	{0x843471040BB9D55CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_7_steps.evt"},
	{0x84347223C0C4CDDBULL, "mcr371_harvester_resurect6217085.cl"},
	{0x8435A36501009321ULL, "vxca_caverns2prc_947_0_0.vlm"},
	{0x8435C49860B1F4A9ULL, "hf_zmother_h3s3.mor"},
	{0x8435F9538A1BF0ECULL, "ser221_party_comments6122871.cl"},
	{0x843606112D3BBECFULL, "544886466.ogg"},
	{0x84363A33B0607B0CULL, "art\\vfx\\combat\\vfx_gib_insect_bld_p.impa.evt"},
	{0x843658D6A2FC97D6ULL, "127883876.ogg"},
	{0x8436A387A8ECE647ULL, "ht_brothel_n_3366.rml"},
	{0x8436A9731BEBBF7CULL, "6001195_m.fxe"},
	{0x8436B2113D50C76EULL, "vxht_mansion2_3d0_0_0.vlm"},
	{0x8437087BA8F4F55DULL, "str_slime_n.dds"},
	{0x84372BF905F3422AULL, "gen00fl_aveline6114005.cl"},
	{0x843769639DBE3360ULL, "qun221_petrice6077327.cl"},
	{0x84378462DDECF4BCULL, "w_encounter2_7_0.rml"},
	{0x8437A839B6E18804ULL, "vxlt_0_lowtown_a3_n_c_e8l_0_0.vlm"},
	{0x8437CDA6F8804D69ULL, "lt_0_lowtown_docks_n_3_0.rml"},
	{0x843804969D9FEBBAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_clenchfists.evt"},
	{0x843851D1D6B97CD5ULL, "6089823_f.fxe"},
	{0x84387451F8A6717DULL, "6042982_m.fxe"},
	{0x843898FFF0C8DEADULL, "vxlt_undercitydungeon_1io_0_1.vlm"},
	{0x8438B97325C77CA8ULL, "gst_female_neutral.gda"},
	{0x8439183299311D3CULL, "6149862_m.fxe"},
	{0x8439289498CF0928ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.cess.ani"},
	{0x84395B91E583E1B4ULL, "gen_im_trs_dem_junk1_04.uti"},
	{0x84399408E8B9277AULL, "61593875.ogg"},
	{0x8439AC0D75917FD3ULL, "w_encounter1_14_0.rml"},
	{0x843A2307093FEF19ULL, "l_env_wal_hightown_int_brothel_laugh.bnk"},
	{0x843A2932CC2282F3ULL, "vxht_chantry_n_2z1_0_0.vlm"},
	{0x843A45D700D9FDF6ULL, "6143699_m.fxe"},
	{0x843A6646CBB0B616ULL, "46965105.ogg"},
	{0x843A6A62DDEEFE11ULL, "w_encounter2_7_0.msh"},
	{0x843A7862DDEF162CULL, "w_encounter2_7_0.mao"},
	{0x843A8462DDEF2ACFULL, "w_encounter2_7_0.mmh"},
	{0x843AC06A5C04500CULL, "6024987_m.fxe"},
	{0x843B17EDCE3A596EULL, "art\\vfx\\combat\\vfx_glb_charm_c.mmh"},
	{0x843B7E911C821206ULL, "210512784.ogg"},
	{0x843B98163E3C47B8ULL, "vxlt_darktown_2x0_0_0.vlm"},
	{0x843C5CA3239011EEULL, "w_encounter3_171_0.msh"},
	{0x843C5EFE44F55ACCULL, "589912277.ogg"},
	{0x843C6AA32390293FULL, "w_encounter3_171_0.mao"},
	{0x843C76A323903DDCULL, "w_encounter3_171_0.mmh"},
	{0x843C8BE5480F9BC0ULL, "6067762_m.fxe"},
	{0x843D81393B14B394ULL, "hlac_ca_cavern1_haunted_24424_6z.dds"},
	{0x843DFE45536A3ABEULL, "vxht_bartrand_n_4p7_0_0.vlm"},
	{0x843E03969DA455ADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\2arm_clenchfists.gad"},
	{0x843E330E47401B1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_sw.anb"},
	{0x843E463270AA380DULL, "972467989.ogg"},
	{0x843E4A2CDB6128DCULL, "hlac_lt_gallowscourtyard_n_-131074_0z.dds"},
	{0x843E9609CB78473FULL, "jobboardplots_one.gda"},
	{0x843F180B14FE7DC4ULL, "146595235.ogg"},
	{0x843F848E59F6CBDBULL, "hlac_ca_cavern1_haunted_24421_11z.dds"},
	{0x843FC0D6F8A5DD19ULL, "vxht_0_hightown_n_aob_0_1.vlm"},
	{0x84400362FADDDBB0ULL, "vxlt_0_lowtown_ej6_0_0.vlm"},
	{0x844014BF2ED9D31FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_twitch.evt"},
	{0x84402BB7C5152D82ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_p.anb"},
	{0x84404209CAC2A34CULL, "275989297.ogg"},
	{0x844131ADA322962AULL, "6166687_m.fxe"},
	{0x84413AB7F330C38AULL, "6136876_f.fxe"},
	{0x84414852DB4C656BULL, "follower_banter6155823.cl"},
	{0x8441636A1C17519EULL, "vxlt_0_lowtown_a3_n_c_en0_0_0.vlm"},
	{0x8441C4A845D84C11ULL, "6029026_m.fxe"},
	{0x84424795F9704F1BULL, "t3_cth_nb6.tnt"},
	{0x844279040BC62075ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_forward_7_steps.ani"},
	{0x844283C95800D47FULL, "vxca_caverns1_jg9_0_0.vlm"},
	{0x8442ADB270B40101ULL, "vxht_0_hightown_n_b67_0_0.vlm"},
	{0x8442CA2A59B0722BULL, "126347712.ogg"},
	{0x8442E878939B55C7ULL, "vxlt_0_lowtown_docks_n_b5h_0_1.vlm"},
	{0x844307E0CEC1BBAFULL, "6042243_m.fxe"},
	{0x8443155C313513A0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.down.gad"},
	{0x844334D8F9FECBF9ULL, "221705442.ogg"},
	{0x8443C63028C9C646ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magiclecturn_n.dds"},
	{0x84441541F2F6ABA1ULL, "vxca_caverns2prc_936_0_0.vlm"},
	{0x84442233B06C9065ULL, "art\\vfx\\combat\\vfx_gib_insect_bld_p.impa.ani"},
	{0x844484AB4A1D9A7CULL, "vxht_0_hightown_a3_b1q_0_0.vlm"},
	{0x8444BD45559071B3ULL, "llac_dr_slavers_3216_7.dds"},
	{0x8444EBF979509F4EULL, "6065788_m.fxe"},
	{0x8445158EDA9E8B4EULL, "vxca_caverns1_jf7_0_0.vlm"},
	{0x844531F0EBAFD469ULL, "lt_0_lowtown_a3_n_c_52_0.rml"},
	{0x84457F08E7C54021ULL, "gen00fl_isabela6088241.cl"},
	{0x84461861EFB3E1D4ULL, "ser211ip_to_lowtown.utp"},
	{0x8446D3E4CBBE778BULL, "qcr000_arishok6059672.cl"},
	{0x8446DB9D6C541C32ULL, "vxht_0_hightown_a3_caz_0_0.vlm"},
	{0x8446F0C454759D6DULL, "vxlt_gallowscourtyard_n_7fo_0_0.vlm"},
	{0x8446F5BF2EDFBD08ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_twitch.gad"},
	{0x8447212B3DC7F421ULL, "isa311_isabela6092291.cl"},
	{0x844804CB7A8ED117ULL, "bad_bs.mao"},
	{0x844836EDCE45186DULL, "art\\vfx\\combat\\vfx_glb_charm_c.anb"},
	{0x844893545FA49401ULL, "dae000_elegant6166919.cl"},
	{0x84489CD32F00B0DDULL, "982525765.ogg"},
	{0x84489FD15AABEF1BULL, "vxht_0_hightown_a2_at7_0_0.vlm"},
	{0x84492EF39C0E61A3ULL, "hlac_lt_gallowscourtyard_0_1z.dds"},
	{0x844978A089086A94ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.cess.ani"},
	{0x8449DAFF16CC3DD3ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bld_flma.mao"},
	{0x844A2F5F978F02AFULL, "301368543.ogg"},
	{0x844A680EB165180DULL, "6045997_f.fxe"},
	{0x844AA909952B9DF2ULL, "character_character_p30.psh.fxo.dx11"},
	{0x844AE6F7DD7D6B71ULL, "633304321.ogg"},
	{0x844B009FF21437D9ULL, "173047670.ogg"},
	{0x844B4538F16CA195ULL, "framebuffereffects_cocmask_p0.psh.fxo.dx11.assembly"},
	{0x844B593C6481E233ULL, "6070800_m.fxe"},
	{0x844B630F5C5E40A0ULL, "48058762.ogg"},
	{0x844B709171944F82ULL, "hlac_lt_0_lowtown_n_-65537_6y.dds"},
	{0x844BCD10FA7171B0ULL, "6048383_m.fxe"},
	{0x844BDE56F1718682ULL, "vxht_0_hightown_n_avo_0_0.vlm"},
	{0x844C2EAAAE09CA2BULL, "6125993_m.fxe"},
	{0x844C44AE63AD4BCCULL, "hlac_w_encounter2_-131071_0x.dds"},
	{0x844C53A2467A8312ULL, "area_core.ncs"},
	{0x844CA753AAD39BDFULL, "6151459_m.fxe"},
	{0x844CD24FD99BB498ULL, "lgt110cr_urn_shill.utc"},
	{0x844DB4B3EB579FF7ULL, "vxlt_gallowscourtyard_515_0_0.vlm"},
	{0x844DDEC800199FF6ULL, "658862753.ogg"},
	{0x844E08D2B7D0D085ULL, "vxdr_primevaltomb_2gv_0_0.vlm"},
	{0x844E3F3CE83DD50BULL, "vxca_caverns1_i6i_0_0.vlm"},
	{0x844F48AEEF090B83ULL, "777840275.ogg"},
	{0x844FB5580D4374ACULL, "vxlt_gallowstemplar_f_557_0_0.vlm"},
	{0x844FBD037F357441ULL, "mer340_bonepit_comments6090835.cl"},
	{0x8450066D238BE2DDULL, "6156650_m.fxe"},
	{0x845075A0890E84F6ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.cess.gad"},
	{0x845105E66CABEF06ULL, "1055055032.ogg"},
	{0x845122467D78CCA7ULL, "l_env_wal_lowtown_int_cough.wbd"},
	{0x845177D784B1A545ULL, "one100_solivitus6066036.cl"},
	{0x845188664BFEF607ULL, "vxca_caverns2prc_804_0_0.vlm"},
	{0x8451F7F8C6C3BCBEULL, "857483861.ogg"},
	{0x8452062472FD37D6ULL, "127373269.ogg"},
	{0x845258540D67E0D9ULL, "663655204.ogg"},
	{0x8452A2A016D69CAFULL, "llac_ca_cavern1_haunted_24424_10.dds"},
	{0x845334FB5B0B61F6ULL, "dae000_gamlen6026184.cl"},
	{0x845343C1437C2E96ULL, "vxlt_foundry_n_wg_0_0.vlm"},
	{0x845346171F4E4943ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_min_c.mmh"},
	{0x845346DE30C19CAFULL, "686786112.ogg"},
	{0x845382FA1A101504ULL, "6149425_f.fxe"},
	{0x8453B09897ED695CULL, "vxht_keep_n_6iw_0_0.vlm"},
	{0x84544933EC80D084ULL, "552720216.ogg"},
	{0x8454BE908D0BBEE8ULL, "gen_im_trs_slv_one_axe_r1.uti"},
	{0x845511BA1EB5502EULL, "vxca_caverns1_e4n_0_0.vlm"},
	{0x845562A1466BB5EDULL, "vxlt_gallowscourtyard_n_7in_0_0.vlm"},
	{0x84556E0A01D418E6ULL, "vxdr_dungeon_1m3_0_1.vlm"},
	{0x8455BDA729B645A9ULL, "vxlt_shop_ct_0_0.vlm"},
	{0x8455CEA2C3E0EE85ULL, "vxw_bonepitmining_a2_1x4_0_0.vlm"},
	{0x8455CFD8CF5C82E3ULL, "6109421_m.fxe"},
	{0x8455E9DD1EEBDADEULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospear.evt"},
	{0x8456225BDBCCFCD7ULL, "vxdr_primevalentrance_1jd_0_0.vlm"},
	{0x84563321C24AC4C7ULL, "and100_anders6095999.cl"},
	{0x8456351A7E0F044FULL, "qun221_viscount6069160.cl"},
	{0x8456397EE6890785ULL, "zz_followers_banter6157681.cl"},
	{0x845643898CBBEA93ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_grill.phy"},
	{0x845660BA821DBE26ULL, "vxlt_0_lowtown_docks_a3_n_cot_0_0.vlm"},
	{0x84566D10F4BD31E0ULL, "171515615.ogg"},
	{0x8456CF85DE957B1AULL, "6201263_m.fxe"},
	{0x84570CECB00EB3F3ULL, "vxht_chantry_n_2p3_0_0.vlm"},
	{0x84573C87DD58F81AULL, "6135733_m.fxe"},
	{0x845784DA810D4E5AULL, "dae000_arishok6072801.cl"},
	{0x845790A08914D155ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.cess.evt"},
	{0x8457A18B20F0F240ULL, "vxlt_0_lowtown_gpf_0_0.vlm"},
	{0x845822975A028621ULL, "gen00fl_aveline6139980.cl"},
	{0x8458A64C9F8C9971ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masd.mmh"},
	{0x8458B24C9F8CAD12ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masd.mao"},
	{0x8458BCB209F1B471ULL, "pro000cr_hurlock_aveline.utc"},
	{0x8458C04C9F8CC52FULL, "art\\characters\\playercharacter\\humanmale\\bdy_masd.msh"},
	{0x8459A0B4C804CC9AULL, "vxlt_hangedman_27r_0_0.vlm"},
	{0x8459BD871B214FC9ULL, "6179988_f.fxe"},
	{0x8459E53F817D55C8ULL, "414949058.ogg"},
	{0x845A15C515FADE7FULL, "gen00fl_dog6166823.cl"},
	{0x845C04DA0B62632CULL, "vxlt_gallowstemplar_n_5t8_0_0.vlm"},
	{0x845C4E3E547F0820ULL, "330357748.ogg"},
	{0x845D00DD1EF22009ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospear.gad"},
	{0x845D17ED2B4E1969ULL, "gen_im_wep_rog_dua_act3_09.uti"},
	{0x845D3E3E32579621ULL, "vxht_brothel_n_5fk_0_0.vlm"},
	{0x845D496BEB3E395EULL, "art\\characters\\creatures\\pridedemon\\animation\\a_strt.ani"},
	{0x845DF6AC1CCF3CCFULL, "vxht_mansion_n_6ja_0_0.vlm"},
	{0x845EB1E79581D307ULL, "vxlt_0_lowtown_n_ghr_0_0.vlm"},
	{0x845F5988588ACF63ULL, "341952179.ogg"},
	{0x845F876F53125BD5ULL, "vxca_caverns1_mines_gkf_0_1.vlm"},
	{0x845FC5C6FB399A5AULL, "6043601_m.fxe"},
	{0x84600BAAA5A8A66AULL, "6136101_m.fxe"},
	{0x846059B59734C8E6ULL, "hlac_lt_0_lowtown_a2_n_1_9z.dds"},
	{0x8460D48733632EB8ULL, "art\\vfx\\creatures\\fxc_bm_spit_p.anb"},
	{0x8460D9AE7719D61BULL, "vxca_cavern1_haunted_dpw_0_0.vlm"},
	{0x84618E57622B1775ULL, "1051549028.ogg"},
	{0x8461A4A8C2A207A4ULL, "6027784_m.fxe"},
	{0x84623AB901F19F87ULL, "lt_0_lowtown_a2_10_0.rml"},
	{0x8462CBB48DB9E784ULL, "dae000_gamlen6005017.cl"},
	{0x846356FDEA893220ULL, "hlac_dr_dungeon_3216_7y.dds"},
	{0x84639106CDE7F78CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4.phy"},
	{0x8463919D073E53BBULL, "w_bonepitmining_a2_187_0.phy"},
	{0x846396E58A5FE703ULL, "vxht_0_hightown_bca_0_0.vlm"},
	{0x8463D1DD1EF7EF57ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospear.ani"},
	{0x8463ECB60CEABF81ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_ctn180r.gad"},
	{0x84640FC25EE5E28AULL, "lgt110_urn_guard6133019.cl"},
	{0x8464407FB5D70705ULL, "vxw_bonepitmining_a2_1u3_0_0.vlm"},
	{0x84649438CDC31557ULL, "vxdr_primevalentrance_1og_0_0.vlm"},
	{0x8464EAF39EFA43E0ULL, "follower_banter6143227.cl"},
	{0x84650CF2D3C05101ULL, "t3_wdg_g00.tnt"},
	{0x84652672BD9E5C3DULL, "6185482_f.fxe"},
	{0x8465306A69656132ULL, "6164490_m.fxe"},
	{0x84656B70526D0BA0ULL, "vxw_bonepitdestroyed_a3_176_0_0.vlm"},
	{0x8465BFB579875D01ULL, "mag311_keran6041703.cl"},
	{0x8465F305F2DAFE36ULL, "6134464_m.fxe"},
	{0x84660B5EB0A934CCULL, "hm_carver_h1s1.mor"},
	{0x8466446C1D23D57DULL, "gen00fl_bethany6101450.cl"},
	{0x8466640CED15841FULL, "art\\levels\\races\\proxy\\props\\_textures\\storage_n.dds"},
	{0x84667E2EFBA9B0F2ULL, "instancing1.psh.fxo"},
	{0x84669327A2FEC5D0ULL, "character_character_p267.psh.fxo.dx11.assembly"},
	{0x8466F227E6D4D657ULL, "character_character_p123.psh.fxo.dx11"},
	{0x846702DA6AA1637BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_rack_d.dds"},
	{0x8467330F5EA7C0C9ULL, "face_face_p13.psh.fxo.dx11"},
	{0x8467675241D34415ULL, "w_bonepitmining_a2_20_0.anb"},
	{0x84676E2CDBABE32FULL, "vxht_0_hightown_a2_b8d_0_0.vlm"},
	{0x8467EE3ADBFAE9FAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_caveent.phy"},
	{0x846826079996DD91ULL, "vxlt_0_lowtown_a3_n_c_f4r_0_0.vlm"},
	{0x84689F6478AD1F03ULL, "6097058_m.fxe"},
	{0x8468B757313BE4DDULL, "6087000_m.fxe"},
	{0x846928E9191EDC1FULL, "hlac_lt_gallowstemplar_2395_0z.dds"},
	{0x846985DCF1FA0F65ULL, "873633692.ogg"},
	{0x846998721F01A6A1ULL, "ave121_brennan6105822.cl"},
	{0x8469D21D67DCE18EULL, "vxht_0_hightown_9zd_0_0.vlm"},
	{0x846A1990DF2D972FULL, "pro000_merchant6051627.cl"},
	{0x846A76B6FD587BACULL, "vxlt_gallowstemplar_n_5o9_0_0.vlm"},
	{0x846AA1BB3294C9D2ULL, "340006102.ogg"},
	{0x846ACCA6F8ABA489ULL, "lt_0_lowtown_docks_n_3_0.anb"},
	{0x846AEDB60CF0E02FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_ctn180r.ani"},
	{0x846B0C274CE76F84ULL, "ser221ip_barrier_mansion.utp"},
	{0x846B5B0797B83355ULL, "mcr370ar_gallows_nt.are"},
	{0x846BA5828A98BD78ULL, "vxht_playershouse_n_1pj_0_0.vlm"},
	{0x846BA5A5C722DE2BULL, "t3_arm_hid.tnt"},
	{0x846BB0F3B7C942A4ULL, "ht_0_hightown_a3_53_0_lc.mmh"},
	{0x846C516D69D6FF0DULL, "6094527_m.fxe"},
	{0x846D25178675975EULL, "llac_w_sundermount_131073_0.dds"},
	{0x846D67B005415F5BULL, "hlac_ht_chantry_2534_22x.dds"},
	{0x846E040402753CBCULL, "gen00fl_merrill6128777.cl"},
	{0x846F37301B8DB4A9ULL, "210719700.ogg"},
	{0x846F5F50197F0E3BULL, "vxlt_gallowstemplar_n_486_0_0.vlm"},
	{0x846F7C949FF16AA5ULL, "132652428.ogg"},
	{0x846FFB9AF21531A6ULL, "6096647_m.fxe"},
	{0x8470863D32AF9BD8ULL, "zz_vault_debug6203390.cl"},
	{0x847096A4A1B2D935ULL, "dre180ip_crypt_exit.utp"},
	{0x8470BA124CAB2C81ULL, "1071016692.ogg"},
	{0x8471626BEB4F0CD8ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_strt.gad"},
	{0x8471864381DBD4D8ULL, "gen00fl_aveline6120804.cl"},
	{0x8471EA9AAABA452FULL, "vxht_0_hightown_n_ap4_0_0.vlm"},
	{0x84721D2976081C55ULL, "vxht_brothel_n_6ea_0_0.vlm"},
	{0x847262127CFA2DB8ULL, "924825774.ogg"},
	{0x8472B25CA7CD1F85ULL, "vxw_bonepitmining_a2_1r2_0_0.vlm"},
	{0x8472C55E131D3839ULL, "6175154_m.fxe"},
	{0x84733BF08F11D3D6ULL, "vxht_0_hightown_a3_a2m_0_0.vlm"},
	{0x8473DD4D44632420ULL, "vxw_bonepitdestroyed_a3_187_0_0.vlm"},
	{0x847482651B6EF897ULL, "234912673.ogg"},
	{0x84750B45DB17A5CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_enter.evt"},
	{0x847535917504B8D5ULL, "vxw_blightlands_7w_0_0.vlm"},
	{0x84753679189663B2ULL, "103546100.ogg"},
	{0x84757C0797C0BFBBULL, "mcr370ar_gallows_nt.ncs"},
	{0x8475EB416519B305ULL, "hlac_lt_0_lowtown_a2_n_-65537_1y.dds"},
	{0x8476056DD330C34BULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_180_c.mmh"},
	{0x84761B345FE5EBCCULL, "hlac_lt_0_lowtown_a2_n_c_131070_6z.dds"},
	{0x84765B2BF8824C99ULL, "6175972_m.fxe"},
	{0x84766855A37CC3ADULL, "mag320cr_crowd_2.utc"},
	{0x8476942AC12377C7ULL, "6166601_f.fxe"},
	{0x84774F09F3E9C29AULL, "6118732_m.fxe"},
	{0x847788366FED26C2ULL, "w_encounter1_15_0.rml"},
	{0x8477A687A924825BULL, "ht_brothel_n_3366.anb"},
	{0x847830F0EBDB2B89ULL, "lt_0_lowtown_a3_n_c_52_0.anb"},
	{0x8479DF78E374950BULL, "gen_im_trs_dal_med_hlm_r1.uti"},
	{0x8479ECB072BCF9B3ULL, "fire01.mao"},
	{0x847A811EF335D1B2ULL, "ef_orana.mor"},
	{0x847A9CE6F1FFDD3EULL, "llac_w_blightlands_1_0.dds"},
	{0x847B4C72CDD905C3ULL, "dae000ip_crowd_ff7_talk12.utp"},
	{0x847B95A1796E0407ULL, "vxlt_0_lowtown_n_gbl_0_0.vlm"},
	{0x847BDA0641D09BF8ULL, "zz_mer_debug6051474.cl"},
	{0x847BE85C16EEF987ULL, "6174951_m.fxe"},
	{0x847C178C9F756CFBULL, "follower_banter6161077.cl"},
	{0x847C2445DB1DEE8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_enter.gad"},
	{0x847C282E125B934BULL, "hlac_w_woundedcoast_65543_0y.dds"},
	{0x847C5C0797C6A862ULL, "mcr370ar_gallows_nt.lst"},
	{0x847CB8C7B24C8AF1ULL, "6128134_m.fxe"},
	{0x847CD05C85D6C500ULL, "gen00fl_fenris6126407.cl"},
	{0x847D195A23143E05ULL, "vxht_playershouse_n_1fu_0_0.vlm"},
	{0x847D69656FF9B5FDULL, "mag320_orsino6071057.cl"},
	{0x847D87992AE2B266ULL, "972758844.ogg"},
	{0x847DBD685B06071BULL, "vxca_cavern1_haunted_dni_0_0.vlm"},
	{0x847E1239DB83C3A9ULL, "410391654.ogg"},
	{0x847E285FB3E8DA1DULL, "348007256.ogg"},
	{0x847E292700CB0174ULL, "vxw_encounter2_44_0_0.vlm"},
	{0x847E81FAB106A9C5ULL, "dae000cr_grand_cleric6124744.cl"},
	{0x847EBACAF43DA3A8ULL, "6076922_m.fxe"},
	{0x847EEF7AB7908FB6ULL, "cod_bks_tranquil.utp"},
	{0x847F76F87BC03F34ULL, "gen00fl_fenris6126775.cl"},
	{0x847FA5DBD844F18AULL, "688238373.ogg"},
	{0x847FD95AF5F4E638ULL, "74167151.ogg"},
	{0x847FDFCBF30141FFULL, "hlac_w_encounter2_-65535_0x.dds"},
	{0x84805C779CB05DAFULL, "vxht_0_hightown_n_au7_0_0.vlm"},
	{0x84806BBF66D2ACB6ULL, "64898883.ogg"},
	{0x84807A9F6E4C1803ULL, "vxht_0_hightown_bic_0_0.vlm"},
	{0x8480FBD040CA4189ULL, "ts_templars_arm_table.gda"},
	{0x848179026529C4A8ULL, "336720850.ogg"},
	{0x8481D9CCE4B14EF5ULL, "6543337.ogg"},
	{0x84822345DB22588AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_enter.ani"},
	{0x84825EF47992DF86ULL, "follower_banter6155781.cl"},
	{0x84826A447AB5AEEEULL, "qun120cr_javaris6043338.cl"},
	{0x848378958B8BCF56ULL, "6149327_m.fxe"},
	{0x8483A76E66FAD155ULL, "vxw_blightlands_8v_0_0.vlm"},
	{0x8483E66DD33CCBB0ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_180_c.anb"},
	{0x84841A51A80A5704ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hed_orsa_0n.dds"},
	{0x8484607A49EF457AULL, "hlac_lt_0_lowtown_-65535_6x.dds"},
	{0x8484624B3DDAA7A6ULL, "follower_banter6172896.cl"},
	{0x8484B9BD6559D148ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grand.msh"},
	{0x8484CBBD6559EFD9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grand.mao"},
	{0x8484D7BD655A0472ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grand.mmh"},
	{0x84854AB82E6B5C27ULL, "193924959.ogg"},
	{0x8485BFC51C4D61B5ULL, "qun211_party_comment6072241.cl"},
	{0x848623827A4B349CULL, "198157232.ogg"},
	{0x84862E9631E969E4ULL, "598356255.ogg"},
	{0x8486AD2B39A52E53ULL, "vxht_0_hightown_n_c5m_0_0.vlm"},
	{0x8486F09C3A03DBF7ULL, "vxht_mansion2b_1yu_0_1.vlm"},
	{0x848872366FFB7417ULL, "w_encounter1_15_0.msh"},
	{0x848884366FFB937AULL, "w_encounter1_15_0.mao"},
	{0x848888366FFB9989ULL, "w_encounter1_15_0.mmh"},
	{0x8488BDB4C461029FULL, "art\\vfx\\environment\\vfx_dark_fog.dura.gad"},
	{0x848949F90CAAB1C1ULL, "101526921.ogg"},
	{0x8489755C3F8BC711ULL, "art\\characters\\playercharacter\\humanfemale\\hf_glv_bda.mmh"},
	{0x84898F5C3F8BF3CFULL, "art\\characters\\playercharacter\\humanfemale\\hf_glv_bda.msh"},
	{0x8489D38733863253ULL, "art\\vfx\\creatures\\fxc_bm_spit_p.mmh"},
	{0x848A7310A2BA2910ULL, "w_bonepitdestroyed_a3_252_0.mao"},
	{0x848A7710A2BA3023ULL, "w_bonepitdestroyed_a3_252_0.mmh"},
	{0x848A8510A2BA487DULL, "w_bonepitdestroyed_a3_252_0.msh"},
	{0x848ABAE5EC43EECFULL, "vxlt_undercitydungeon_158_0_0.vlm"},
	{0x848AC65BA56DB53FULL, "dae000_arishok6068441.cl"},
	{0x848AD71DD9131743ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_peak.mao"},
	{0x848ADB1DD9131D90ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_peak.mmh"},
	{0x848AE91DD91335AAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3co_peak.msh"},
	{0x848B7EF33112E212ULL, "mag100_thrask6031996.cl"},
	{0x848BA8D85C1AB05CULL, "follower_barks6153166.cl"},
	{0x848BCBB0C36B8739ULL, "art\\vfx\\combat\\vfx_d_arc_direct_cast_c.impa.ani"},
	{0x848C2B57D05CFF0DULL, "w_encounter2_35_0.phy"},
	{0x848C8290E4E79F4BULL, "zz_mage2_whiteout3.cut"},
	{0x848C95A3EE6EB536ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\2arm_clasphands.evt"},
	{0x848CBD1CA3B7F3FFULL, "lt_0_lowtown_docks_a3_196_0.msh"},
	{0x848CC31CA3B7FDA1ULL, "lt_0_lowtown_docks_a3_196_0.mmh"},
	{0x848CCF1CA3B81242ULL, "lt_0_lowtown_docks_a3_196_0.mao"},
	{0x848CE72B8F874187ULL, "vxht_mansion_n_588_0_0.vlm"},
	{0x848CEE6A45BC1149ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_3x.dds"},
	{0x848D30950A82D0FEULL, "gen00fl_sebastian6174690.cl"},
	{0x848D3778F0BD81D8ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.anb"},
	{0x848DB9581FD51AF5ULL, "773589453.ogg"},
	{0x848E80FEF1824C27ULL, "505102381.ogg"},
	{0x848EF21215BA18BDULL, "vxlt_0_lowtown_docks_a3_n_d3e_0_1.vlm"},
	{0x848F27341B1C5860ULL, "vxlt_undercityserialkiller_11r_0_0.vlm"},
	{0x848F789D0763645AULL, "w_bonepitmining_a2_187_0.msh"},
	{0x848F827386576CBDULL, "qun131_taskmaster6039756.cl"},
	{0x848F869D07637B93ULL, "w_bonepitmining_a2_187_0.mao"},
	{0x848F8A9D076382A0ULL, "w_bonepitmining_a2_187_0.mmh"},
	{0x848FB2B4C4670E8CULL, "art\\vfx\\environment\\vfx_dark_fog.dura.evt"},
	{0x848FBBD1947B3247ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.anb"},
	{0x848FCD576DCE4345ULL, "419388028.ogg"},
	{0x8490118DC1FB2D48ULL, "follower_banter6164787.cl"},
	{0x84902DE0FAF9CCD2ULL, "vxht_0_hightown_n_c5h_0_0.vlm"},
	{0x8490645241F6441DULL, "w_bonepitmining_a2_20_0.mao"},
	{0x8490685241F64B56ULL, "w_bonepitmining_a2_20_0.mmh"},
	{0x8490725241F65C54ULL, "w_bonepitmining_a2_20_0.msh"},
	{0x8490A6ED3098D84BULL, "vxht_0_hightown_c6z_0_1.vlm"},
	{0x8492043470D292FCULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluviancrack_sa.dds"},
	{0x84920E6E72760072ULL, "38014255.ogg"},
	{0x84920F2EEE65A7C7ULL, "413225992.ogg"},
	{0x84921434B55A9B4CULL, "vxht_0_hightown_n_c0y_0_0.vlm"},
	{0x84922C467DB074DDULL, "l_env_wal_lowtown_int_cough.bnk"},
	{0x84924324F09926F6ULL, "qcr220_cut_meredith_arrival.cnv"},
	{0x84925A21ADC5E496ULL, "follower_banter6172786.cl"},
	{0x84927BA62F25FFB6ULL, "408401214.ogg"},
	{0x84929B18EDAA6B78ULL, "6118288_m.fxe"},
	{0x8492A6B0C371677BULL, "art\\vfx\\combat\\vfx_d_arc_direct_cast_c.impa.gad"},
	{0x8492F4E86301DB0BULL, "6015463_m.fxe"},
	{0x84932C57D0631FB9ULL, "w_encounter2_35_0.rml"},
	{0x84937705CA4D9FBBULL, "6090864_m.fxe"},
	{0x84938CA3EE74C4C1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\2arm_clasphands.gad"},
	{0x84939F543EC23DA9ULL, "302483445.ogg"},
	{0x8494CAFB75C7440AULL, "gen00fl_merrill6128748.cl"},
	{0x8494D2E568245C69ULL, "462398221.ogg"},
	{0x84951B77A639D05FULL, "6052754_m.fxe"},
	{0x8495B0C1CC89B348ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_ornate.msh"},
	{0x8495C2C1CC89D1D9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_ornate.mao"},
	{0x8495CEC1CC89E672ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_ornate.mmh"},
	{0x84964524F09D12D6ULL, "qcr220_cut_meredith_arrival.bnk"},
	{0x8496470F24083DF5ULL, "lt_0_lowtown_docks_n_43_0.rml"},
	{0x8496810391032170ULL, "6197850_m.fxe"},
	{0x8496E23559C9E5FBULL, "800987773.ogg"},
	{0x8497596624DDA497ULL, "6060716_f.fxe"},
	{0x8497627BA753E4A2ULL, "vxht_mansion2b_1bp_0_0.vlm"},
	{0x8498C3250198D6A4ULL, "6034630_m.fxe"},
	{0x8498C3B0C3760460ULL, "art\\vfx\\combat\\vfx_d_arc_direct_cast_c.impa.evt"},
	{0x849951BB2D506A4AULL, "hlac_dr_primevalentrance_1726_3y.dds"},
	{0x84996B90C21156FCULL, "vxca_caverns1_mines_dpt_0_0.vlm"},
	{0x8499D302C81A6BD5ULL, "hlac_lt_0_lowtown_-65536_0x.dds"},
	{0x8499FD1407006F05ULL, "vxht_playershouse_n_1ps_0_0.vlm"},
	{0x849A073624C66997ULL, "fen110_fenris6101941.cl"},
	{0x849A15EDB2A5FC85ULL, "art\\vfx\\combat\\vfx_gib_corm_ltrb.anb"},
	{0x849A15EDB2A5FC8EULL, "art\\vfx\\combat\\vfx_gib_corm_ltrb.ani"},
	{0x849A1E68CEEB74D6ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgte_0t.dds"},
	{0x849A8DA3EE7AE56FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\2arm_clasphands.ani"},
	{0x849AD5C413BEDAE1ULL, "hlac_w_sundermount_-2_1z.dds"},
	{0x849B5D10A71D8289ULL, "6180680_m.fxe"},
	{0x849B9BAF69AAFB32ULL, "6034940_m.fxe"},
	{0x849BDAFA504F50C8ULL, "zz_vrc_debug6064282.cl"},
	{0x849C2B251C267AB6ULL, "llac_lt_0_lowtown_-3_7.dds"},
	{0x849C71EC4CAC1CDDULL, "hlac_ca_caverns1_mines_24425_8x.dds"},
	{0x849CD1EAB96DD0B1ULL, "ave100_ambush_16106198.cl"},
	{0x849D24E3809CD798ULL, "hlac_ht_mansion2b_818_31x.dds"},
	{0x849D72C04BEA65D5ULL, "vxht_brothel_n_6tn_0_0.vlm"},
	{0x849D9AB4C47323E5ULL, "art\\vfx\\environment\\vfx_dark_fog.dura.ani"},
	{0x849E450E93263B58ULL, "kbfield.ncs"},
	{0x849E74149204879EULL, "vxlt_0_lowtown_docks_b7m_0_0.vlm"},
	{0x849ED3E9C85C497AULL, "hlac_w_sundermount_-4_0x.dds"},
	{0x849EED2763DBDE9AULL, "vxlt_hangedman_1i7_0_0.vlm"},
	{0x849F0F1908FEAF39ULL, "753325381.ogg"},
	{0x849F2EDDCBD1EA71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_end.ani"},
	{0x849F6B00275B5DC6ULL, "vxht_mansion2b_2pu_0_0.vlm"},
	{0x849FF0D4A64BB40FULL, "vxht_0_hightown_n_ay3_0_0.vlm"},
	{0x84A02246F11BE2D5ULL, "llac_lt_gallowscourtyard_n_-65536_0.dds"},
	{0x84A02D1BDFD381C0ULL, "18419825.ogg"},
	{0x84A02E71D32D7DBEULL, "vxht_bartrand_n_72v_0_0.vlm"},
	{0x84A08DC599BCA77BULL, "sg7_water_cyc.mao"},
	{0x84A0C1AD81357FFFULL, "vxlt_0_lowtown_docks_coa_0_0.vlm"},
	{0x84A0EB2A83F9297CULL, "hlac_w_encounter2_camp_131075_0y.dds"},
	{0x84A116C56AE851FCULL, "633363200.ogg"},
	{0x84A13B9A3C37377AULL, "gen00fl_aveline6132728.cl"},
	{0x84A1A48E4B070BC1ULL, "vxht_mansion2b_28v_0_0.vlm"},
	{0x84A1BE58CCE2666FULL, "vxw_bonepitmining_a2_1lz_0_0.vlm"},
	{0x84A1E8E2BC43B83BULL, "6095764_m.fxe"},
	{0x84A20037333D1499ULL, "vxdr_primevalentrance_2og_0_1.vlm"},
	{0x84A22DC61E4211E0ULL, "6100094_m.fxe"},
	{0x84A25AE0E4B98FA7ULL, "dre000_dougal6129737.cl"},
	{0x84A26466307E2CCDULL, "dae000_bodhan6146814.cl"},
	{0x84A2981A49101454ULL, "593778313.ogg"},
	{0x84A2EE80B2AFF47AULL, "vxlt_0_lowtown_a2_fg0_0_0.vlm"},
	{0x84A3079F1E0D26BBULL, "vxht_keep_6a3_0_0.vlm"},
	{0x84A3A5CB55435B13ULL, "513117944.ogg"},
	{0x84A3AC774D0CCD2EULL, "art\\vfx\\materials\\vfx_drain_blend.mao"},
	{0x84A3EEC846101B5CULL, "con_ico_yes.dds"},
	{0x84A3F2859D2C93F6ULL, "dae000_gen_player_turn_in6177840.cl"},
	{0x84A4116EC06DE89CULL, "vxlt_hangedman_sz_0_0.vlm"},
	{0x84A4232AE9AC32E9ULL, "minimap.gfx"},
	{0x84A46806CE1F4957ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball4.mmh"},
	{0x84A46C1DA074FB02ULL, "character_character_p32.psh.fxo.dx11.assembly"},
	{0x84A4990057CC0B1CULL, "537720336.ogg"},
	{0x84A5023C64B2E2AFULL, "ca_cavern1_haunted.arl"},
	{0x84A5360053D98687ULL, "941516555.ogg"},
	{0x84A59BB2FD5BC840ULL, "vxca_cavern1_haunted_j5i_0_0.vlm"},
	{0x84A5EFDC11C8E1F7ULL, "1054878525.ogg"},
	{0x84A639DDCBD81B13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_end.gad"},
	{0x84A6705177EACE05ULL, "6059036_m.fxe"},
	{0x84A702FFEB047C70ULL, "follower_banter6141575.cl"},
	{0x84A72AA07E8DB1C7ULL, "vxlt_gallowscourtyard_n_75p_0_0.vlm"},
	{0x84A736E237379656ULL, "vxlt_hangedman_t9_0_0.vlm"},
	{0x84A76D45FE53BC53ULL, "gen_im_wep_mag_sta_act2_05.uti"},
	{0x84A85070D9C6F12EULL, "item_property_types.gda"},
	{0x84A8A1D5DA37FF6EULL, "335639245.ogg"},
	{0x84A8B9EDFEE92900ULL, "dre000_sibling_has_blight6036476.cl"},
	{0x84A9268A1BE8F749ULL, "6112984_m.fxe"},
	{0x84A9F9FCBC47ABFDULL, "t3_eey_am2.tnt"},
	{0x84AA134A0F340A25ULL, "vxdr_dungeon_21r_0_0.vlm"},
	{0x84AB1C5B0DE1833BULL, "6172719_m.fxe"},
	{0x84AB5139F6769DF0ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_6y.dds"},
	{0x84AB5C50F308C0EFULL, "407621366.ogg"},
	{0x84ABD036442E6183ULL, "vxht_0_hightown_bxd_0_0.vlm"},
	{0x84AC01211076007DULL, "803791904.ogg"},
	{0x84AC2837D5AF84EEULL, "vxw_encounter3_c0_0_0.vlm"},
	{0x84AC375114ED58CBULL, "6173107_m.fxe"},
	{0x84AC66B1491848DCULL, "dae300_cullen6140657.cl"},
	{0x84ACFDFE31A331E5ULL, "6069398_m.fxe"},
	{0x84AD013FBCDC226EULL, "vxca_caverns1_ihv_0_0.vlm"},
	{0x84AD2B4709943D00ULL, "6034920_m.fxe"},
	{0x84AD36DDCBDE3578ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_end.evt"},
	{0x84AD4655F7CFFD34ULL, "498727038.ogg"},
	{0x84ADA882B169C3E7ULL, "6111213_m.fxe"},
	{0x84ADEEEDB2B663C8ULL, "art\\vfx\\combat\\vfx_gib_corm_ltrb.gad"},
	{0x84AE3C7B20765AC2ULL, "mcr371_scared_mage_f6086921.cl"},
	{0x84AF2F68FF865CAFULL, "nat380_nathaniel6066363.cl"},
	{0x84AF7ED2D1969C78ULL, "6112088_m.fxe"},
	{0x84AFB404EF8035C5ULL, "gen00fl_varric6093601.cl"},
	{0x84B0166B3CFD2441ULL, "vxht_mansion2b_25u_0_0.vlm"},
	{0x84B079EE93D2F625ULL, "254871884.ogg"},
	{0x84B11BE460D96481ULL, "vxca_caverns2_3p8_0_0.vlm"},
	{0x84B124C460E4ED4AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_roba_0n.dds"},
	{0x84B12A34A9923FD8ULL, "vxw_encounter1_f3_0_0.vlm"},
	{0x84B14E627B0C025EULL, "6205977_m.fxe"},
	{0x84B1797C10033F3BULL, "vxht_keep_6j2_0_0.vlm"},
	{0x84B20B6E21AB5C0EULL, "vxht_0_hightown_9ay_0_0.vlm"},
	{0x84B23DAAE7E5B402ULL, "hlac_lt_0_lowtown_a2_n_65534_4y.dds"},
	{0x84B265DB5CA6F062ULL, "6031312_m.fxe"},
	{0x84B2821B9B55254AULL, "784186651.ogg"},
	{0x84B2EB9FC8737E55ULL, "gen00fl_merrill6128972.cl"},
	{0x84B37007484A2E8DULL, "6150483_m.fxe"},
	{0x84B3BD44B93494D3ULL, "vxht_brothel_n_59p_0_0.vlm"},
	{0x84B3CCF546C1AE32ULL, "6054754_f.fxe"},
	{0x84B3E2535E7AB84CULL, "220544182.ogg"},
	{0x84B446358B4015A2ULL, "vxht_mansion2b_1xr_0_0.vlm"},
	{0x84B47EB823ABA903ULL, "vxlt_gallowstemplar_5qk_0_0.vlm"},
	{0x84B49FBF3FB06FB8ULL, "mcr370_desten6087320.cl"},
	{0x84B4E8E190C1B588ULL, "6076283_m.fxe"},
	{0x84B528F45422B95FULL, "vxht_keep_57v_0_0.vlm"},
	{0x84B58ED6C89543A7ULL, "llac_lt_0_lowtown_a2_n_0_9.dds"},
	{0x84B5B831C4B6133EULL, "vxca_caverns2_4ti_0_1.vlm"},
	{0x84B6145B5635F042ULL, "128856073.ogg"},
	{0x84B63678F0E08573ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.mmh"},
	{0x84B64A6ED4E24FBFULL, "182313078.ogg"},
	{0x84B6EF7F616F5A27ULL, "hlac_ht_keep_3736_3z.dds"},
	{0x84B72DF3405D91F6ULL, "vxlt_gallowstemplar_n_37a_0_0.vlm"},
	{0x84B79DFA5ADEC18EULL, "6118101_m.fxe"},
	{0x84B83C898D0EA908ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_grill.mmh"},
	{0x84B848898D0EBDABULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_grill.mao"},
	{0x84B85A898D0EDC32ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_grill.msh"},
	{0x84B8B8D1949E3248ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.mmh"},
	{0x84B9378F6F7D3091ULL, "mag170_recruit016080835.cl"},
	{0x84B9504E3434BBF6ULL, "vxlt_gallowstemplar_f_3sm_0_0.vlm"},
	{0x84B9E872E10A690CULL, "hlac_ht_keep_3732_16y.dds"},
	{0x84BA1221858C8047ULL, "gen_im_trs_cot_lgt_boo_r0.uti"},
	{0x84BAF13E92DE307CULL, "vxlt_shop_eo_0_0.vlm"},
	{0x84BB4593BFDE44A2ULL, "hlac_lt_0_lowtown_65537_2y.dds"},
	{0x84BB75E01217E293ULL, "hlac_lt_0_lowtown_n_65536_10y.dds"},
	{0x84BBDB4739B70C22ULL, "vxlt_0_lowtown_a2_n_gve_0_0.vlm"},
	{0x84BC57246CEE3391ULL, "vxlt_0_lowtown_h2j_0_0.vlm"},
	{0x84BCB50E925234A8ULL, "6165312_m.fxe"},
	{0x84BCD986B010EECAULL, "6065656_m.fxe"},
	{0x84BD3DFA7A893352ULL, "vxht_brothel_n_59u_0_0.vlm"},
	{0x84BD83F15D8B1E9FULL, "6029406_m.fxe"},
	{0x84BE6B3A8038E8A1ULL, "hlac_w_sundermount_-1_0x.dds"},
	{0x84BE8A7928930626ULL, "hlac_lt_0_lowtown_a2_n_131071_2y.dds"},
	{0x84BF9C119B885858ULL, "vxw_encounter1_k0_0_0.vlm"},
	{0x84BF9EBB3A2E9CCBULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic_2.mao"},
	{0x84BFA24522FF35CAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\posev1.gad"},
	{0x84BFE86B8742E03AULL, "6111318_f.fxe"},
	{0x84BFF9C692598E9DULL, "llac_ht_mansion2b_818_54.dds"},
	{0x84C0603C0DDE2A3BULL, "hlac_lt_0_lowtown_a2_n_-65536_1x.dds"},
	{0x84C1384442E3DF3EULL, "llac_lt_0_lowtown_docks_-65539_3.dds"},
	{0x84C14DE76EB9EE4AULL, "good_med.mao"},
	{0x84C15C9C6ADDEAF7ULL, "6017844_f.fxe"},
	{0x84C168CD1D359C81ULL, "6097827_m.fxe"},
	{0x84C17881AEB74660ULL, "6027673_f.fxe"},
	{0x84C17D30066DDE53ULL, "vxht_0_hightown_a2_c7x_0_0.vlm"},
	{0x84C18988BBE25C02ULL, "lt_backalley_n_0_0_ll.mmh"},
	{0x84C2581758222D75ULL, "hlac_dr_slavers_3217_2z.dds"},
	{0x84C2AB1DF9AC10D4ULL, "199235782.ogg"},
	{0x84C2B8127D362E22ULL, "vxht_mansion2b_1ss_0_0.vlm"},
	{0x84C2E73ADC478923ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_caveent.msh"},
	{0x84C2F53ADC47A17DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_caveent.mmh"},
	{0x84C2F74E54B6D2ECULL, "lowtown_male_lowlife1_sta6164994.cl"},
	{0x84C2F93ADC47A78EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_caveent.mao"},
	{0x84C303CF321FFA40ULL, "hlac_lt_hangedman_n_-1_2x.dds"},
	{0x84C406E27AF9F9B1ULL, "follower_banter6169174.cl"},
	{0x84C458841A552799ULL, "225990837.ogg"},
	{0x84C49158F0BD0F77ULL, "vxw_encounter3_e3_0_0.vlm"},
	{0x84C4E0C72A3CF5CCULL, "ht_0_hightownexplosion_n_5_0.msh"},
	{0x84C4E6C72A3CFFBEULL, "ht_0_hightownexplosion_n_5_0.mmh"},
	{0x84C4E8EA16ABBB5FULL, "vxht_0_hightown_a3_an4_0_0.vlm"},
	{0x84C4F2C72A3D1455ULL, "ht_0_hightownexplosion_n_5_0.mao"},
	{0x84C51BE3B3F3F730ULL, "vxht_mansion_n_6om_0_0.vlm"},
	{0x84C53BD8F48B042EULL, "hlac_lt_gallowscourtyard_n_-65536_1x.dds"},
	{0x84C55A497D0BA490ULL, "llac_lt_0_lowtown_docks_a3_n_0_4.dds"},
	{0x84C578AD4C97D8AEULL, "6148104_m.fxe"},
	{0x84C597A4A1FB1A84ULL, "dre180ip_crypt_exit.ncs"},
	{0x84C5E37E70632169ULL, "6102732_m.fxe"},
	{0x84C661B84E871C38ULL, "llac_lt_undercityserialkiller_-1_15.dds"},
	{0x84C6854523052361ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\posev1.evt"},
	{0x84C6B5D3E4DDB074ULL, "vxw_blightlands_w2_0_0.vlm"},
	{0x84C747FEFFC1990AULL, "377501928.ogg"},
	{0x84C7599D1D9625BDULL, "6077331_m.fxe"},
	{0x84C769F002733805ULL, "339044833.ogg"},
	{0x84C7A677305FC4E9ULL, "6177322_f.fxe"},
	{0x84C8160F97387AD4ULL, "mag291ip_book.utp"},
	{0x84C838D9FD61E08EULL, "mag111tr_abomination.utt"},
	{0x84C85B9718D7609AULL, "lgt220_mag_apost_apost26170153.cl"},
	{0x84C8E5A887F88F78ULL, "vxlt_undercityserialkiller_192_0_0.vlm"},
	{0x84C9C5E91F438DFCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sider.mao"},
	{0x84C9D1E91F43A19FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sider.mmh"},
	{0x84C9D4E66A492F8DULL, "lt_0_lowtown_a2_45_0_ll.mmh"},
	{0x84C9D7E91F43AC41ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sider.msh"},
	{0x84C9EC15BF1DE9ECULL, "zz_one000_barker6179350.cl"},
	{0x84CA02E4E2C589F0ULL, "vxlt_darktown_6oy_0_0.vlm"},
	{0x84CA62E439D4ED3DULL, "vxdr_primevalentrance_26u_0_0.vlm"},
	{0x84CA79CD5D3BD163ULL, "6098954_m.fxe"},
	{0x84CAB24B73D99619ULL, "360220475.ogg"},
	{0x84CAC9015EE44C11ULL, "vxlt_0_lowtown_h7i_0_0.vlm"},
	{0x84CAE7B980C9E9FEULL, "6081276_m.fxe"},
	{0x84CAEAA7220506B4ULL, "744276314.ogg"},
	{0x84CB2CB0A68B46B1ULL, "hlac_lt_gallowstemplar_f_2391_6y.dds"},
	{0x84CB45E390D65996ULL, "hlac_lt_0_lowtown_a3_n_c_1_11x.dds"},
	{0x84CBA0BA5ABC2161ULL, "202232415.ogg"},
	{0x84CBC60B25CD8169ULL, "vxca_caverns1_ii2_0_0.vlm"},
	{0x84CC88FDE0B12E81ULL, "llac_lt_0_lowtown_a2_n_c_-196609_3.dds"},
	{0x84CCC6BBFCFC2DA1ULL, "follower_banter6160739.cl"},
	{0x84CCE9185211D501ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_hed_suca_0n.dds"},
	{0x84CD3C6C55C8AF98ULL, "vxlt_darktown_5kz_0_0.vlm"},
	{0x84CD8D4B11EAB248ULL, "166956434.ogg"},
	{0x84CDA9690CF664DBULL, "gen00fl_isabela6122925.cl"},
	{0x84CDDA5D9CB95033ULL, "vxca_caverns2prc_8ms_0_0.vlm"},
	{0x84CDF1261010321DULL, "crack3.dds"},
	{0x84CE7D5675963D7CULL, "6169837_m.fxe"},
	{0x84CEAE2DCD0FDD36ULL, "vxlt_0_lowtown_docks_b62_0_0.vlm"},
	{0x84CEDE727A34BE4FULL, "472763468.ogg"},
	{0x84CF0C72DBCF8D97ULL, "gen00fl_aveline6137478.cl"},
	{0x84CFBEEFF1AED771ULL, "mag100_karras6034746.cl"},
	{0x84D06D8257D2D23DULL, "w_bonepit_248_0.mao"},
	{0x84D0718257D2D976ULL, "w_bonepit_248_0.mmh"},
	{0x84D07B8257D2EA74ULL, "w_bonepit_248_0.msh"},
	{0x84D0DE43A35112F2ULL, "vxht_keep_n_5x5_0_0.vlm"},
	{0x84D0E06C3B659DB2ULL, "363821290.ogg"},
	{0x84D0EEF1F4AB4143ULL, "1067374135.ogg"},
	{0x84D17545E3C71E09ULL, "w_bonepitmining_a2_21_0_lc.mmh"},
	{0x84D44ED5055240CBULL, "hlac_ht_0_hightown_a2_65533_1y.dds"},
	{0x84D47D45231152B0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\posev1.ani"},
	{0x84D4DA207937F547ULL, "409584452.ogg"},
	{0x84D51FFBD2740433ULL, "plt_and000pt_gift_amulet.nss"},
	{0x84D52258FF44954EULL, "vxw_encounter1_jh_0_1.vlm"},
	{0x84D527B0D6D3C8F4ULL, "vxw_blightlands_x3_0_0.vlm"},
	{0x84D59B983E966D6DULL, "vxht_mansion2_32q_0_0.vlm"},
	{0x84D63FAF4AC2C1E4ULL, "6191517_m.fxe"},
	{0x84D67A74A7992A57ULL, "gen00fl_aveline6139384.cl"},
	{0x84D6DC0555F75A25ULL, "hlac_lt_0_lowtown_a2_n_0_7x.dds"},
	{0x84D6EC21CC3E5431ULL, "6174169_f.fxe"},
	{0x84D7561D35AE9051ULL, "hlac_w_woundedcoast_262147_0x.dds"},
	{0x84D7A12C7982CF99ULL, "hlac_lt_0_lowtown_a2_n_-196609_2y.dds"},
	{0x84D80D90D4244233ULL, "143097597.ogg"},
	{0x84D84CF7A6588261ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junka.phy"},
	{0x84D8B7D1AF4BB590ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tray_sa.dds"},
	{0x84D8DA7729A133E2ULL, "vxht_keep_6lu_0_0.vlm"},
	{0x84D9554D7A7B41DAULL, "223645541.ogg"},
	{0x84DA5D939F66817DULL, "llac_w_sundermount_-393211_0.dds"},
	{0x84DAEC5D55D76F01ULL, "vxlt_undercitydungeon_1c5_0_1.vlm"},
	{0x84DB312E82CACEC4ULL, "605665336.ogg"},
	{0x84DB99A1BFDA87AAULL, "llac_w_woundedcoast_131079_0.dds"},
	{0x84DBA76BE2C50B6DULL, "follower_barks6153316.cl"},
	{0x84DBCB826A9B5ACEULL, "vxht_0_hightown_n_awp_0_0.vlm"},
	{0x84DC13B3E7405CE8ULL, "6121183_m.fxe"},
	{0x84DC544AD0942EA3ULL, "llac_w_sundermount_65533_0.dds"},
	{0x84DD09C3EB9C85F5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pausetwitch_v2.evt"},
	{0x84DD8FE5502CD90BULL, "vxht_mansion_n_3v8_0_0.vlm"},
	{0x84DE60E9EA5A0F53ULL, "vxht_0_hightown_a2_c9z_0_0.vlm"},
	{0x84DE69149B8FFDE6ULL, "vxlt_0_lowtown_docks_brn_0_0.vlm"},
	{0x84DEB1C3B4E5D988ULL, "6084677_m.fxe"},
	{0x84DF352F3DD3786FULL, "hlac_lt_0_lowtown_a2_n_131070_1z.dds"},
	{0x84DF7E4EB1524133ULL, "vxdr_primevalentrance_2sh_0_0.vlm"},
	{0x84DFA2CAA5563DE2ULL, "6025339_m.fxe"},
	{0x84DFE9BD74E1CC30ULL, "48606716.ogg"},
	{0x84E0051E76102198ULL, "445309018.ogg"},
	{0x84E00E84A8A7D695ULL, "624090224.ogg"},
	{0x84E02BD9FD765CB7ULL, "mag111tr_abomination.ncs"},
	{0x84E063B064ABFE62ULL, "3003699.ogg"},
	{0x84E064EA0417EDBEULL, "hlac_ht_chantry_2534_36x.dds"},
	{0x84E0896890373B84ULL, "gencr_strider.utc"},
	{0x84E091D7F90B0CA5ULL, "vxlt_gallowstemplar_f_3j5_0_0.vlm"},
	{0x84E0C7C4E080D455ULL, "plt_ave200pt_opening.nss"},
	{0x84E11ECF9E45A337ULL, "zz_followers_banter6157782.cl"},
	{0x84E1878516D72D12ULL, "338949992.ogg"},
	{0x84E1B07F07F42EEDULL, "vxlt_warehouse_sc_0_0.vlm"},
	{0x84E1B636245397CCULL, "vxw_woundedcoast_c8_0_0.vlm"},
	{0x84E1CCA3FA97EC5FULL, "vxht_0_hightown_a3_ap6_0_0.vlm"},
	{0x84E1F07BB98D5B53ULL, "105464913.ogg"},
	{0x84E20EB9DA372D9AULL, "6099502_f.fxe"},
	{0x84E22259C7DC9609ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_lgtc_0d.dds"},
	{0x84E2A5C56704F395ULL, "w_encounter3_235_0.msh"},
	{0x84E2A7C56704F70BULL, "w_encounter3_235_0.mmh"},
	{0x84E2B3C567050BA8ULL, "w_encounter3_235_0.mao"},
	{0x84E2C240C9E5D441ULL, "gen00fl_bethany6097890.cl"},
	{0x84E3F5B27A5838FBULL, "hlac_lt_0_lowtown_a2_n_-65535_0z.dds"},
	{0x84E4382FA986E695ULL, "gen_im_npc_arm_hum_boo_mrca.uti"},
	{0x84E4C4F2EC8FB623ULL, "zz_dae_debug6029287.cl"},
	{0x84E54A22514CB4A0ULL, "vxdr_primevaltomb_862_0_0.vlm"},
	{0x84E57256CA9B66C9ULL, "6078723_m.fxe"},
	{0x84E5C9626BE4C078ULL, "vxlt_undercityserialkiller_130_0_0.vlm"},
	{0x84E6EC900DA0A8BEULL, "art\\vfx\\combat\\vfx_m_sph_revival_cast_c.impa.evt"},
	{0x84E71396059E4EA6ULL, "6096056_m.fxe"},
	{0x84E7C6E173A29A5EULL, "6068736_m.fxe"},
	{0x84E803337CAE8C3DULL, "gen00fl_anders6111268.cl"},
	{0x84E8038587D49181ULL, "vxlt_gallowstemplar_34l_0_0.vlm"},
	{0x84E83D6A97EE951BULL, "hlac_lt_0_lowtown_n_65536_3y.dds"},
	{0x84E84C92FBD23A63ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgte_0n.dds"},
	{0x84E854EE3A9D8ED1ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_6y.dds"},
	{0x84E872A3716D9143ULL, "6083961_m.fxe"},
	{0x84E8916CD835B877ULL, "vxht_0_hightown_n_by8_0_0.vlm"},
	{0x84E906B578C95A73ULL, "vxlt_darktown_516_0_1.vlm"},
	{0x84E90E28B5E23BDEULL, "gen00fl_anders6215044.cl"},
	{0x84E936FA2B646CD6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_flowergroupa.msh"},
	{0x84E940FA2B647DD4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_flowergroupa.mmh"},
	{0x84E94F8940FDA330ULL, "mag330_walter6044683.cl"},
	{0x84E9AF6964302203ULL, "character_character_p11.psh.fxo.dx11"},
	{0x84E9FDB7D56B4163ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.dura.ani"},
	{0x84EA202639B4E098ULL, "vxlt_darktown_5qx_0_0.vlm"},
	{0x84EA5683BFA2CE00ULL, "vxlt_undercityencounter_tz_0_0.vlm"},
	{0x84EA8FA25A63F327ULL, "349190502.ogg"},
	{0x84EA9D07DC30987FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls104a_0d.dds"},
	{0x84EAB6DE66FB9584ULL, "sst_ss_cityelf_f_1.gda"},
	{0x84EABB7867B55A59ULL, "ave121_brennan6105839.cl"},
	{0x84EADE54E595AD2EULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga103a_0s.dds"},
	{0x84EC3748A5766D07ULL, "hm_karras.mor"},
	{0x84EC44340F558F73ULL, "gen00fl_sebastian6180373.cl"},
	{0x84EC8DC158798B62ULL, "vxht_0_hightown_n_bfx_0_0.vlm"},
	{0x84ECFFF9BDEB00E9ULL, "cd_dre000_anders.bnk"},
	{0x84ED050F9757798DULL, "mag291ip_book.ncs"},
	{0x84ED3690CD840492ULL, "dae000ip_crowd_mf8_talk2.utp"},
	{0x84ED67D437A5A735ULL, "41496412.ogg"},
	{0x84ED6EC73DC3FDC2ULL, "vxlt_0_lowtown_a3_n_c_ffk_0_0.vlm"},
	{0x84EDA06CB1188542ULL, "zz_pro_debug6206781.cl"},
	{0x84EE03900DA6EEE9ULL, "art\\vfx\\combat\\vfx_m_sph_revival_cast_c.impa.gad"},
	{0x84EE0A6D007C0E8FULL, "one300_nuncio6083486.cl"},
	{0x84EE49DA9700AA29ULL, "94906614.ogg"},
	{0x84EE7631B9557927ULL, "vxht_0_hightown_a3_b9n_0_0.vlm"},
	{0x84EEC7AE88209240ULL, "hlac_lt_0_lowtown_a2_n_c_65538_4y.dds"},
	{0x84EF2365980240F0ULL, "gen00fl_isabela6123233.cl"},
	{0x84EF33F1C06B65D5ULL, "mcr370_varric6086807.cl"},
	{0x84EF875E905CE24DULL, "646416590.ogg"},
	{0x84EF9352A52ACD66ULL, "vxlt_gallowscourtyard_7ej_0_0.vlm"},
	{0x84F0115C33EE245CULL, "976183398.ogg"},
	{0x84F015EDC696B388ULL, "vxlt_gallowscourtyard_6uq_0_0.vlm"},
	{0x84F02C7070CD9456ULL, "vxht_mansion2_2ev_0_0.vlm"},
	{0x84F08D773E5E9D0FULL, "795325121.ogg"},
	{0x84F0B2D0FCB246E3ULL, "hlac_lt_0_lowtown_a3_n_c_-3_2z.dds"},
	{0x84F0CDC1CCD6906DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wo_ornate.phy"},
	{0x84F0D3C159CB9DC0ULL, "hlac_ht_mansion2_818_8x.dds"},
	{0x84F0F40B86ADE954ULL, "vxlt_0_lowtown_docks_8pl_0_0.vlm"},
	{0x84F12D2CAAEE22FDULL, "vxlt_gallowstemplar_f_5t5_0_0.vlm"},
	{0x84F13507743CB8FFULL, "50331564.ogg"},
	{0x84F141C44090C14AULL, "dae000ip_crowd_em4_talk3.utp"},
	{0x84F1A1E63AB96C1CULL, "752983761.ogg"},
	{0x84F25E45235602F9ULL, "vxlt_0_lowtown_a3_n_c_gze_0_0.vlm"},
	{0x84F350F5CD4D65AAULL, "464092249.ogg"},
	{0x84F37598F06B0A14ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_c.impa.ani"},
	{0x84F3B553D046D1D4ULL, "hlac_lt_warehouse_-65536_1x.dds"},
	{0x84F3BE1B164ED020ULL, "dae000_meredith6170975.cl"},
	{0x84F3CE7715F3CAE5ULL, "sib100_father_contact6108953.cl"},
	{0x84F4D4900DACBD37ULL, "art\\vfx\\combat\\vfx_m_sph_revival_cast_c.impa.ani"},
	{0x84F53063F2506550ULL, "352579975.ogg"},
	{0x84F558E91E8B7D73ULL, "lt_0_lowtown_n_19_0.anb"},
	{0x84F56C10D6B917A9ULL, "vxw_bonepitdestroyed_a3_16p_0_0.vlm"},
	{0x84F59A559F1B7EF4ULL, "tbe100cr_urchin.utc"},
	{0x84F5D191EB6417CEULL, "vxdr_primevalentrance_2lp_0_0.vlm"},
	{0x84F5FCD20380A9D1ULL, "6085496_f.fxe"},
	{0x84F6E203C9B49E3DULL, "6161061_m.fxe"},
	{0x84F715B7D575F45AULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.dura.evt"},
	{0x84F79F74B736B3B9ULL, "862508806.ogg"},
	{0x84F860106EA77F9FULL, "704711973.ogg"},
	{0x84F8D913112CF358ULL, "vis121_saemus6086708.cl"},
	{0x84F9F854A9225F5AULL, "mer110_jansen6032915.cl"},
	{0x84FA59A6F475ABF2ULL, "624539526.ogg"},
	{0x84FA7298F0712476ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_c.impa.gad"},
	{0x84FAA27E26C42B68ULL, "vxca_caverns2_7jw_0_0.vlm"},
	{0x84FB29093EAC7F97ULL, "hlac_w_encounter2_393215_0z.dds"},
	{0x84FB43ABEE4CCB2DULL, "llac_w_encounter3_-65534_0.dds"},
	{0x84FB5A0072B2ED95ULL, "6119887_f.fxe"},
	{0x84FB9B71D6EE9510ULL, "fx_smoke_vol_add.mao"},
	{0x84FBA604F498BEC4ULL, "625533093.ogg"},
	{0x84FBD510C9C78C20ULL, "6014723_f.fxe"},
	{0x84FBDEC64B36B201ULL, "art\\vfx\\combat\\vfx_gib_qun_rtlb.ani"},
	{0x84FC40868E27C2B8ULL, "zz_approval_debug6003720.cl"},
	{0x84FC89CF1F8F68EFULL, "fx_0006_diff.dds"},
	{0x84FCE0FB9AC7723BULL, "and000pt_gift_key.ncs"},
	{0x84FD52A81EDAFE15ULL, "gen00fl_anders6128369.cl"},
	{0x84FE14B7D57C114DULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_aoe_p.dura.gad"},
	{0x84FE4BA71E3666F1ULL, "vxlt_0_lowtown_docks_azz_0_0.vlm"},
	{0x84FE67AABCBAAD56ULL, "6151628_f.fxe"},
	{0x84FE9438EBE05FEDULL, "vxlt_warehouse_ya_0_0.vlm"},
	{0x84FE9D42003D2D68ULL, "zz_approval_debug6003508.cl"},
	{0x84FF9E532AA82CF0ULL, "6173669_m.fxe"},
	{0x84FFAEDBC6ACCEBEULL, "486524816.ogg"},
	{0x8500458845B97FB9ULL, "hlac_lt_0_lowtown_n_65536_11x.dds"},
	{0x8500754CE5904004ULL, "llac_lt_0_lowtown_n_-1_6.dds"},
	{0x8500921BCE7A7E5CULL, "6081285_m.fxe"},
	{0x8500A797A18C5459ULL, "vxw_bonepitmining_a2_5l_0_0.vlm"},
	{0x85017C15D93EA69CULL, "zz_qcr_debug6074961.cl"},
	{0x85018D98F07770D5ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_c.impa.evt"},
	{0x8501A54ADF99CBD5ULL, "814902797.ogg"},
	{0x8501C98A5FDEF361ULL, "312134678.ogg"},
	{0x8501E9C64B3B30A3ULL, "art\\vfx\\combat\\vfx_gib_qun_rtlb.gad"},
	{0x85021FC57B02A999ULL, "gen00fl_anders6111295.cl"},
	{0x85023F3841928563ULL, "549212529.ogg"},
	{0x8502471D04E501A0ULL, "vxht_keep_5wu_0_0.vlm"},
	{0x8502BE26C50D4E5CULL, "vxlt_0_lowtown_byx_0_0.vlm"},
	{0x8502C6C4E09DB8B6ULL, "plt_ave200pt_opening.plo"},
	{0x8502EC577A91BDEBULL, "vxdr_primevaltomb_748_0_0.vlm"},
	{0x85032B6E6215C7B1ULL, "vxw_encounter2_camp_s0_0_0.vlm"},
	{0x85038B5823090DCDULL, "hlac_ca_caverns1_24424_0x.dds"},
	{0x85040ED2310B211BULL, "89166946.ogg"},
	{0x85045B9D4866F777ULL, "gen_im_wep_rog_arc_act3_05.uti"},
	{0x85045D478F024297ULL, "vxht_mansion2_27q_0_0.vlm"},
	{0x8504628D6400AA93ULL, "hlac_lt_0_lowtown_a2_n_65537_7x.dds"},
	{0x85047D43EB805625ULL, "hlac_lt_foundry_n_320_4x.dds"},
	{0x8504CA01A6C7E28AULL, "vxht_0_hightown_65f_0_0.vlm"},
	{0x850575BA982C4120ULL, "671502942.ogg"},
	{0x8505BBDE02DD2948ULL, "vxlt_backalley_n_1yj_0_1.vlm"},
	{0x8505BDC72A745159ULL, "ht_0_hightownexplosion_n_5_0.phy"},
	{0x8505E90637FDEB2CULL, "hlac_dr_slavers_3207_3y.dds"},
	{0x8506035C3E2FDEACULL, "6034340_m.fxe"},
	{0x850688E084C3C6C9ULL, "6161313_m.fxe"},
	{0x8506D4E85B8A7DE1ULL, "ht_0_hightown_n_44_0_ll.mmh"},
	{0x8507121521DD1BEEULL, "vxdr_primevalentrance_3d9_0_0.vlm"},
	{0x850800E436155B36ULL, "mag291_ntbanter6056726.cl"},
	{0x850871E690CADF12ULL, "564055443.ogg"},
	{0x8508785D84C6760AULL, "hlac_ca_cavern1_haunted_24424_18y.dds"},
	{0x85088EA647DD7FC1ULL, "6043692_f.fxe"},
	{0x8508AFF072C66211ULL, "6085904_f.fxe"},
	{0x85097DF9E5A70925ULL, "hlac_w_blightlands_0_12y.dds"},
	{0x8509BEAB71724766ULL, "vxlt_0_lowtown_docks_bcq_0_0.vlm"},
	{0x8509E7070390382FULL, "vxlt_0_lowtown_a3_n_c_cmr_0_0.vlm"},
	{0x850A8585A90385CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxa.msh"},
	{0x850A8B85A9038FBEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxa.mmh"},
	{0x850A9785A903A455ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxa.mao"},
	{0x850AA191CD481580ULL, "231599810.ogg"},
	{0x850AA5B76B2974F2ULL, "vxht_keep_n_5l1_0_0.vlm"},
	{0x850AAAB9D91E6985ULL, "6180918_m.fxe"},
	{0x850B23B6BB64D5A5ULL, "730298264.ogg"},
	{0x850B826A675306CCULL, "zz_followers_banter6158070.cl"},
	{0x850C6807A8624C0EULL, "art\\characters\\playercharacter\\humanfemale\\hf_hat_zbm.mao"},
	{0x850C86796A8FE59FULL, "726819819.ogg"},
	{0x850C89FBD10B86E0ULL, "834566015.ogg"},
	{0x850CA36E0B8F7185ULL, "902935423.ogg"},
	{0x850CF0DA538114D7ULL, "mag170_recruit016094520.cl"},
	{0x850D2FD5243A2753ULL, "6114492_m.fxe"},
	{0x850D6D90D811485DULL, "407016269.ogg"},
	{0x850E27412A2F98DAULL, "vxw_bonepit_1s0_0_0.vlm"},
	{0x850E7967D975DC7AULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_splash.dura.gad"},
	{0x850E8C04A8588A4EULL, "391971563.ogg"},
	{0x850E952F5C93930DULL, "6046843_m.fxe"},
	{0x850EADB7A24DE7D1ULL, "llac_ht_brothel_n_3368_0.dds"},
	{0x850EC073ACE094CBULL, "mcr310_argument6182459.cl"},
	{0x850F9DFA5E9326DFULL, "43549653.ogg"},
	{0x85107A3E71E2B6B9ULL, "vxht_0_hightown_n_c6z_0_0.vlm"},
	{0x85109D4AB74B9BD3ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcmale_fenrisrom_alt.ani"},
	{0x8510B8F9F6DB1A20ULL, "vxht_keep_5zr_0_0.vlm"},
	{0x8510FA9649A610E3ULL, "341663934.ogg"},
	{0x851100366DEB099FULL, "llac_w_blightlands_0_4.dds"},
	{0x85112E12381A251EULL, "hightown_male_noble_gossip6174058.cl"},
	{0x851147D761463A51ULL, "framebuffereffects_verticalblur_p0.vsh.fxo.dx11.assembly"},
	{0x85119D4B540BE031ULL, "vxw_encounter2_camp_n7_0_0.vlm"},
	{0x8511E8DF53762EC3ULL, "vxht_mansion2_3cp_0_0.vlm"},
	{0x8511FB4DABE5190AULL, "gen_im_trs_qun_hvy_glv_r2.uti"},
	{0x851223487241A02AULL, "6133304_m.fxe"},
	{0x8512CF2480F85B17ULL, "vxht_mansion2_20p_0_0.vlm"},
	{0x851358E0CBE79C87ULL, "6111316_m.fxe"},
	{0x85138C83C1DCA34DULL, "vxca_caverns1_mines_hib_0_0.vlm"},
	{0x8513B01DFE1F54EBULL, "hlac_lt_0_lowtown_a3_n_c_-2_5x.dds"},
	{0x8513D74179E8E3B4ULL, "w_blightlands_118_0.phy"},
	{0x85147C67D97A4C91ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_splash.dura.evt"},
	{0x8514CF8FE0F95897ULL, "w_blightlands_239_0.mmh"},
	{0x8514D38FE0F95FA4ULL, "w_blightlands_239_0.mao"},
	{0x8514E58FE0F97E09ULL, "w_blightlands_239_0.msh"},
	{0x85156200234F62D2ULL, "6116594_m.fxe"},
	{0x851567D8BD3288F8ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_dirt_d.dds"},
	{0x8516D7EA627C2374ULL, "987806350.ogg"},
	{0x85175FE064CF5D26ULL, "hd_sp_dot.mao"},
	{0x851767F44A8AB697ULL, "qun160_party_comment_26061243.cl"},
	{0x85179FECC49F76F9ULL, "841866435.ogg"},
	{0x8517BDD02053344DULL, "vxlt_0_lowtown_a2_gos_0_1.vlm"},
	{0x8517D2BC88CF8264ULL, "vxlt_gallowstemplar_f_6q8_0_0.vlm"},
	{0x851807805B66EC27ULL, "vxca_caverns1_mines_don_0_0.vlm"},
	{0x851822325C85171EULL, "73942335.ogg"},
	{0x8518C8DC81A5B72EULL, "zz_followers_banter6157402.cl"},
	{0x8519507635D92811ULL, "vxlt_darktown_5kn_0_1.vlm"},
	{0x8519DBE1FB60E383ULL, "mag100_karras6206105.cl"},
	{0x8519FFDF998DAC41ULL, "6108626_m.fxe"},
	{0x851A43337B4F865CULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st105a_0s.dds"},
	{0x851A4E55B93DC434ULL, "snow_02_l3.mao"},
	{0x851A6761A121C931ULL, "vxca_caverns1_jw7_0_0.vlm"},
	{0x851AC93A5786445DULL, "llac_lt_backalley_g_1_6.dds"},
	{0x851B0BF97E39F4E2ULL, "6034221_m.fxe"},
	{0x851B471A04DE6526ULL, "vxht_mansion2b_3dp_0_0.vlm"},
	{0x851BB06DA813C5CFULL, "w_bonepitdestroyed_a3_99_0.mao"},
	{0x851BBC6DA813DA6CULL, "w_bonepitdestroyed_a3_99_0.mmh"},
	{0x851BC26DA813E45EULL, "w_bonepitdestroyed_a3_99_0.msh"},
	{0x851BD84285A1DA4BULL, "hlac_ca_caverns1_24425_5z.dds"},
	{0x851C4BF7FDD5D304ULL, "vxlt_0_lowtown_gpb_0_0.vlm"},
	{0x851D005A3CE45DD1ULL, "vxlt_0_lowtown_docks_a3_aov_0_0.vlm"},
	{0x851D070C4E89AB8DULL, "6041943_m.fxe"},
	{0x851D0EE42DA9F92AULL, "vxht_mansion2b_2py_0_0.vlm"},
	{0x851D2A02C707A720ULL, "643569807.ogg"},
	{0x851DB44179F0FCB1ULL, "w_blightlands_118_0.msh"},
	{0x851DC24179F114CCULL, "w_blightlands_118_0.mao"},
	{0x851DC853F6FCB773ULL, "lt_hangedman_3567_lc.mmh"},
	{0x851DC9B3516206ABULL, "lt_0_lowtown_docks_a3_n_31_0.anb"},
	{0x851DCE4179F1296FULL, "w_blightlands_118_0.mmh"},
	{0x851E99BA279E33CFULL, "503883417.ogg"},
	{0x851EA719FA9B1130ULL, "1055958931.ogg"},
	{0x851EE3452BC498ADULL, "6137071_m.fxe"},
	{0x851F15FB7683C8CBULL, "mag160_feynriel6020060.cl"},
	{0x851F1B12CFEA16E2ULL, "hlac_ht_keep_4519_4x.dds"},
	{0x8520138ECBCD92E5ULL, "vxdr_0_deeproads_boy_0_1.vlm"},
	{0x8520434E30DB9374ULL, "vrc221ip_servant_corpse_1.utp"},
	{0x85207E2003D69AFAULL, "34339295.ogg"},
	{0x8520A035863BCBF3ULL, "vxlt_0_lowtown_a2_n_c_hl7_0_0.vlm"},
	{0x852104D7B2C25967ULL, "vxca_caverns2prc_ady_0_0.vlm"},
	{0x85215F84DEFE74ABULL, "gen00fl_aveline6139685.cl"},
	{0x852175E91EB0E92FULL, "lt_0_lowtown_n_19_0.rml"},
	{0x85219E541D816D38ULL, "6119548_m.fxe"},
	{0x8522292164468987ULL, "vxca_cave_20d_0_0.vlm"},
	{0x85222AE106255918ULL, "vxw_woundedcoast_162_0_0.vlm"},
	{0x85222AF117794229ULL, "vxca_cavern1_haunted_iis_0_0.vlm"},
	{0x85229467D986B360ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_splash.dura.ani"},
	{0x852327B225B77B61ULL, "hlac_ca_caverns2prc_5802_10z.dds"},
	{0x85233DFB2F779FC5ULL, "807626480.ogg"},
	{0x8524744AB75BFF7DULL, "art\\cinematics\\act3\\romances\\fenris\\cs_pcmale_fenrisrom_alt.gad"},
	{0x8524772EB5580BB3ULL, "6112102_m.fxe"},
	{0x85248244D86165A9ULL, "367424764.ogg"},
	{0x8524BA73952F6100ULL, "6051819_m.fxe"},
	{0x852535121F9EAFC0ULL, "59246530.ogg"},
	{0x8525739A11A06F5AULL, "6028800_m.fxe"},
	{0x85257FE1E09CCC76ULL, "and271_lyrium_smuggler6108514.cl"},
	{0x8525CC940D72E2EEULL, "hlac_lt_backalley_g_0_6x.dds"},
	{0x852625AEFFB22F37ULL, "vxlt_0_lowtown_a2_eow_0_1.vlm"},
	{0x852627EE552F09F8ULL, "hlac_lt_0_lowtown_a2_n_c_-3_1y.dds"},
	{0x8526862B4406732FULL, "vxdr_primevaltomb_39c_0_0.vlm"},
	{0x85269DD5B14BBE89ULL, "654959166.ogg"},
	{0x8526F89BA5D9CBC2ULL, "222826130.ogg"},
	{0x85272EF116FC2943ULL, "231864115.ogg"},
	{0x85273476DB929FE6ULL, "hf_qmq_patronf4.mor"},
	{0x852759FC412403BAULL, "vxlt_0_lowtown_docks_aru_0_0.vlm"},
	{0x85277730C7BC0C12ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_ext_c.mmh"},
	{0x85278FE232D2D1F1ULL, "6218972_m.fxe"},
	{0x852882CDCE5391ADULL, "gen00fl_anders6125524.cl"},
	{0x85288B60E293868DULL, "622220780.ogg"},
	{0x8528ACD139445E1FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shortbed_n.dds"},
	{0x852906330CC4753EULL, "523894564.ogg"},
	{0x85297DD670300D62ULL, "6031121_m.fxe"},
	{0x8529999C0630D517ULL, "gen00fl_anders6114478.cl"},
	{0x8529CE0CCC9A91A0ULL, "546452817.ogg"},
	{0x852A2F874A143195ULL, "follower_banter6164065.cl"},
	{0x852A830B4D1F2D95ULL, "mcr370_negotiation6078745.cl"},
	{0x852A9A76D2E37E7AULL, "vxw_encounter1_b5_0_0.vlm"},
	{0x852AC85C24E86E67ULL, "6071116_f.fxe"},
	{0x852B1C64DF124C7CULL, "mer320_hubert6089953.cl"},
	{0x852BC7D562730CFFULL, "ser191_emeric6108131.cl"},
	{0x852BEC31EB5C09AEULL, "hlac_lt_gallowstemplar_n_7198_2z.dds"},
	{0x852C38E8D1A882C4ULL, "lt_0_lowtown_n_water_21475.mmh"},
	{0x852C3CE8D1A88917ULL, "lt_0_lowtown_n_water_21475.mao"},
	{0x852C4EE8D1A8A866ULL, "lt_0_lowtown_n_water_21475.msh"},
	{0x852CD783EE63F4D9ULL, "vxht_chantry_n_2tl_0_1.vlm"},
	{0x852CDAE91F982EE4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sider.phy"},
	{0x852CFB1092239948ULL, "951989494.ogg"},
	{0x852D36FA5F4D68F8ULL, "vxdr_0_deeproads_ewb_0_0.vlm"},
	{0x852D5DF855CEE7B9ULL, "vxht_0_hightown_n_c8x_0_0.vlm"},
	{0x852D6A5FED1F5A66ULL, "vxht_keep_n_4nn_0_0.vlm"},
	{0x852D84A931D0794DULL, "6055638_m.fxe"},
	{0x852DAADF38C00A78ULL, "ser221_gascard6127889.cl"},
	{0x852DF27B01A2D9ADULL, "6082014_m.fxe"},
	{0x852DF96C8BB5A432ULL, "vxht_0_hightown_a3_c1j_0_0.vlm"},
	{0x852E08D6074868B5ULL, "6148738_m.fxe"},
	{0x852E1D41D5190544ULL, "vxw_bonepitdestroyed_a3_22z_0_0.vlm"},
	{0x852EB5282B25E85DULL, "mag160_feynriel6020417.cl"},
	{0x852EFE63D2049818ULL, "vxdr_0_deeproads_bhe_0_1.vlm"},
	{0x852F0702562DD12AULL, "6139504_m.fxe"},
	{0x852F30924E819FF4ULL, "vxlt_darktown_44n_0_0.vlm"},
	{0x852F9A19DA1A126CULL, "01_bloodimpact_1.mao"},
	{0x852FA3A17A2C1694ULL, "462017260.ogg"},
	{0x85309CBDF81B7198ULL, "vxw_woundedcoast_113_0_0.vlm"},
	{0x8530BE3F80394DA2ULL, "one280_iwan6096869.cl"},
	{0x8530CC4D0D564928ULL, "6094988_m.fxe"},
	{0x8530E4A90DC097D4ULL, "264458867.ogg"},
	{0x8530FDC9BE83F755ULL, "vxw_blightlands_tz_0_0.vlm"},
	{0x8531A092EE854D4AULL, "6104943_f.fxe"},
	{0x8531E8A0142CDD3EULL, "hlac_lt_0_lowtown_docks_65533_4x.dds"},
	{0x85322009D0C5B755ULL, "vxw_sundermount_2gg_0_0.vlm"},
	{0x853244855D6A38A7ULL, "6164316_m.fxe"},
	{0x8533A32E904FEEEEULL, "692850736.ogg"},
	{0x8533B8D3FA614E98ULL, "follower_banter6173277.cl"},
	{0x8533D584A9900BDAULL, "mag121_idunna6083950.cl"},
	{0x8533DDF9BE26C93BULL, "cd_dre000_anders.wbd"},
	{0x85343E8A31EC530AULL, "art\\characters\\weapons\\dagger\\w_dgr_102a.mao"},
	{0x8534ED8243202B43ULL, "392671883.ogg"},
	{0x8535393A82CB06C8ULL, "darktown_male_lowlife_mob6174394.cl"},
	{0x85353D941811408FULL, "6143359_m.fxe"},
	{0x853594ED54F1CA57ULL, "vxlt_backalley_n_1wb_0_0.vlm"},
	{0x8535B50339C0EE32ULL, "6201465_m.fxe"},
	{0x8535FE2F2A8DECFCULL, "follower_banter6156092.cl"},
	{0x85364CBD7591DD46ULL, "384262560.ogg"},
	{0x853651872AC50097ULL, "6061240_m.fxe"},
	{0x853665C1FB639898ULL, "6061283_m.fxe"},
	{0x853691AB11E4D381ULL, "355927282.ogg"},
	{0x8536E3997D79DD41ULL, "6106426_m.fxe"},
	{0x8537490BF94DA00EULL, "zz_dog_debug6204100.cl"},
	{0x853772C679D65691ULL, "6103010_m.fxe"},
	{0x8537ECE077B3291EULL, "vxw_woundedcoast_gv_0_0.vlm"},
	{0x8537F9160BC825CCULL, "454586208.ogg"},
	{0x8538196C63E958CBULL, "hlac_lt_0_lowtown_docks_n_65532_1x.dds"},
	{0x85381DE0C86E1733ULL, "art\\vfx\\combat\\vfx_c_rw_attack1_c.impa.ani"},
	{0x85386403E2F89EC6ULL, "zz_approval_debug6133731.cl"},
	{0x853864054187DAD7ULL, "one300dg_lia6064529.cl"},
	{0x85390C53C4D996FAULL, "vxw_encounter1_g6_0_0.vlm"},
	{0x8539A95946612C06ULL, "vxlt_0_lowtown_docks_n_cv4_0_0.vlm"},
	{0x853A07C5C7624AEAULL, "vxw_bonepit_zd_0_0.vlm"},
	{0x853A77E67D580945ULL, "ser270_templar6112569.cl"},
	{0x853B126A0838E224ULL, "6206062_m.fxe"},
	{0x853B1AFBD2CAAE9CULL, "plt_and000pt_gift_amulet.plo"},
	{0x853B51F7A6AD079DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junka.mao"},
	{0x853B55F7A6AD0ED6ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junka.mmh"},
	{0x853B5FF7A6AD1FD4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junka.msh"},
	{0x853B86E87FF8E8EDULL, "vxlt_blackemporium_10c_0_0.vlm"},
	{0x853BBA7BEFD4C9F3ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_3z.dds"},
	{0x853BC530D50D083FULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_imp_p.mmh"},
	{0x853BFE2C2A9E1C97ULL, "gen00fl_varric6063842.cl"},
	{0x853C326C20A07048ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_fire_sa.dds"},
	{0x853C85206B7D0DA8ULL, "vxht_0_hightown_a3_c9y_0_1.vlm"},
	{0x853C9FE1E4FF126DULL, "vxw_encounter3_8r_0_0.vlm"},
	{0x853E7522EC1A00B0ULL, "llac_lt_0_lowtown_docks_n_-4_4.dds"},
	{0x853E820E016D9E38ULL, "6113348_m.fxe"},
	{0x853E8E531AB5CFF7ULL, "6153163_m.fxe"},
	{0x853EA6779D3488ACULL, "357408133.ogg"},
	{0x853ED346F7054B30ULL, "vxlt_0_lowtown_a3_n_c_ht9_0_0.vlm"},
	{0x853ED6078384D530ULL, "fen110_fenris6117943.cl"},
	{0x853EF44674314C38ULL, "hlac_ht_keep_n_4519_1z.dds"},
	{0x853F0EAAFB657329ULL, "vxca_cavern1_haunted_icq_0_0.vlm"},
	{0x85400724031F0B60ULL, "574961749.ogg"},
	{0x85404C498D875848ULL, "517547757.ogg"},
	{0x8540EB4CE4B32816ULL, "862381714.ogg"},
	{0x854128312357328AULL, "clf260_harley6069646.cl"},
	{0x85418405BB6A5D47ULL, "art\\vfx\\creatures\\fxc_arch_fb_p.anb"},
	{0x8541938673444980ULL, "zz_dae_debug6029435.cl"},
	{0x8541EF524A22B3EEULL, "6079359_f.fxe"},
	{0x8541FFB4945306FEULL, "6081405_m.fxe"},
	{0x85420FB1213CD24BULL, "vxht_mansion2_28l_0_0.vlm"},
	{0x8542840486D48F47ULL, "6103100_m.fxe"},
	{0x85428BA53C61A9F5ULL, "lt_0_lowtown_docks_a3_n_210_0.msh"},
	{0x85428DA53C61AD6BULL, "lt_0_lowtown_docks_a3_n_210_0.mmh"},
	{0x854296F429FF049BULL, "gen00fl_isabela6080452.cl"},
	{0x854299A53C61C108ULL, "lt_0_lowtown_docks_a3_n_210_0.mao"},
	{0x8542CD38B14D5611ULL, "740554651.ogg"},
	{0x8542F1D95179C6EAULL, "6085842_f.fxe"},
	{0x85431B386E2D7940ULL, "ht_0_hightown_a3_45_0_ll.mmh"},
	{0x85433885FCEF8239ULL, "vxw_woundedcoast_o5_0_0.vlm"},
	{0x8543637844DCD6C1ULL, "480284357.ogg"},
	{0x8543CCF96FDD7EA2ULL, "vxlt_0_lowtown_docks_n_c0f_0_0.vlm"},
	{0x8543DB0C0DE64102ULL, "6137704_m.fxe"},
	{0x8545C9D596DAA057ULL, "gen00fl_fenris6107160.cl"},
	{0x854615E0C87A476AULL, "art\\vfx\\combat\\vfx_c_rw_attack1_c.impa.evt"},
	{0x8546C0C20DB4DE57ULL, "vxw_blightlands_145_0_0.vlm"},
	{0x854703C1B23E24B7ULL, "6154306_m.fxe"},
	{0x85471647CAA11203ULL, "gen00fl_fenris6110667.cl"},
	{0x85474F7F1F87BEE2ULL, "dae000_arishok6078447.cl"},
	{0x85478B22D1203903ULL, "675350540.ogg"},
	{0x85478E0F45CB9C78ULL, "855190776.ogg"},
	{0x8547EC940EDD0C8CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc2a.msh"},
	{0x8547F2940EDD177EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc2a.mmh"},
	{0x85480EE9E8273489ULL, "dae120cr_hm_guard_1.utc"},
	{0x854815B06B799B19ULL, "hlac_lt_gallowstemplar_f_2383_6z.dds"},
	{0x85481B3638574486ULL, "vxlt_0_lowtown_docks_n_cq5_0_0.vlm"},
	{0x85488863278C261AULL, "34989063.ogg"},
	{0x8548C6D85FB93AA5ULL, "gen_im_trs_dal_hvy_glv_r0.uti"},
	{0x8548CA670D11F959ULL, "902053103.ogg"},
	{0x8548CAA3CDF24654ULL, "mag291st_vincento.stg"},
	{0x85495FC71D2C824EULL, "hlac_w_blightlands_-2_0x.dds"},
	{0x85496DBDAF7C2EEEULL, "vxlt_0_lowtown_a2_es7_0_0.vlm"},
	{0x8549753D898AEA1DULL, "95527791.ogg"},
	{0x854990F0A84DA7C3ULL, "6167443_m.fxe"},
	{0x8549DC28E00D29B9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\tn90r.gad"},
	{0x854A02415ADBE932ULL, "276627918.ogg"},
	{0x854AC6B35188EFF7ULL, "lt_0_lowtown_docks_a3_n_31_0.rml"},
	{0x854AF15BFC1262C6ULL, "6218603_m.fxe"},
	{0x854B11BED6F52AEDULL, "vxw_encounter3_5q_0_0.vlm"},
	{0x854B3EEBF6A8C08DULL, "mcr370_meredith6136125.cl"},
	{0x854B6285A93AE159ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxa.phy"},
	{0x854BC1DB86C04044ULL, "vxht_0_hightown_a2_2xc_0_0.vlm"},
	{0x854C1BD3F20A2CE7ULL, "qun211_party_comment6072261.cl"},
	{0x854C22F93A277DE8ULL, "6218832_m.fxe"},
	{0x854C2DCB4DEAE9FCULL, "vxw_bonepitdestroyed_a3_17j_0_0.vlm"},
	{0x854C631C3D8047A0ULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.mmh"},
	{0x854CF4E0C8802DDDULL, "art\\vfx\\combat\\vfx_c_rw_attack1_c.impa.gad"},
	{0x854D00300CABE877ULL, "6156263_m.fxe"},
	{0x854D1BDD20492521ULL, "l_dae321ar_player_home_nt.bnk"},
	{0x854D293439D62518ULL, "542108303.ogg"},
	{0x854D8B869E32248BULL, "follower_barks6153862.cl"},
	{0x854E2D3DD5B69C21ULL, "hlac_ht_bartrand_n_3364_1y.dds"},
	{0x854EE0BDAA8D1702ULL, "vxw_bonepit_26p_0_0.vlm"},
	{0x854EFFBC4B713221ULL, "284260623.ogg"},
	{0x854F03C3B4B1E855ULL, "vxw_sundermount_2me_0_0.vlm"},
	{0x854FEC5C28BC2262ULL, "hlac_w_bonepitdestroyed_a3_458754_0y.dds"},
	{0x854FF6C22ED90B93ULL, "hlac_lt_0_lowtown_docks_-65538_4y.dds"},
	{0x8550079688B9F99EULL, "hlac_lt_0_lowtown_a2_n_c_-3_6z.dds"},
	{0x855066C6AB514879ULL, "6016684_m.fxe"},
	{0x85507630C7DF0FB9ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_ext_c.anb"},
	{0x85508350FAC0ADF2ULL, "gen00fl_merrill6150682.cl"},
	{0x8550ED28E0136507ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\tn90r.ani"},
	{0x8550F5E178CAA7C7ULL, "6104000_m.fxe"},
	{0x85517B35F38BF263ULL, "6000726_m.fxe"},
	{0x8551CF6398EBAB1FULL, "vxw_encounter3_fs_0_1.vlm"},
	{0x8551DCBE93E69B45ULL, "defaultstate.sta.dx11"},
	{0x8551EFD17A0AD353ULL, "vxht_0_hightown_a2_cab_0_0.vlm"},
	{0x855209E0744266F4ULL, "6169028_m.fxe"},
	{0x8552642505FABC06ULL, "6136444_m.fxe"},
	{0x855306A2AFD08DABULL, "ht_brothel_n_3371.anb"},
	{0x85531358E2D7C542ULL, "6082145_m.fxe"},
	{0x8553268CBB9C4AABULL, "6120114_m.fxe"},
	{0x855391C5474F4A53ULL, "vrc321_party_barks6095731.cl"},
	{0x8553C482224DE897ULL, "one300_zevran6089729.cl"},
	{0x8554A12FDAF9CCEDULL, "gui_guipixelshader_p3.psh.fxo.dx11.assembly"},
	{0x8554D09A5B9F5A1EULL, "vxw_woundedcoast_yh_0_0.vlm"},
	{0x8555336B778A5668ULL, "1020082104.ogg"},
	{0x85557848361FB02DULL, "6034068_m.fxe"},
	{0x85565FF33B3E673AULL, "mag101_alain6032288.cl"},
	{0x8556C7F361CF78B3ULL, "436709327.ogg"},
	{0x8556F789143DCCFFULL, "hlac_lt_0_lowtown_a3_n_c_-65536_4x.dds"},
	{0x85572EAD3ACE61B8ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_floortile_sa.dds"},
	{0x85573A5DEFFD5D74ULL, "6144457_f.fxe"},
	{0x855766FBC373CEE0ULL, "6142419_f.fxe"},
	{0x8557B71B631E81D1ULL, "gen00fl_anders6215193.cl"},
	{0x85588C91352FB278ULL, "vxdr_0_deeproads_efa_0_0.vlm"},
	{0x8559BF72B947F9FFULL, "qun200_party_comment6095400.cl"},
	{0x855A461C3D8C53BFULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.anb"},
	{0x855A9FA83FE1027CULL, "vxw_bonepitdestroyed_a3_18k_0_0.vlm"},
	{0x855ADA7F850D587FULL, "gen00fl_dog6167675.cl"},
	{0x855B461005B281F4ULL, "1002767755.ogg"},
	{0x855B70873B5F71D2ULL, "vxht_0_hightown_a2_cag_0_0.vlm"},
	{0x855B7FAAD3333072ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv2.gad"},
	{0x855BA712C699D0B0ULL, "265681538.ogg"},
	{0x855BF8A53A04AC5BULL, "face0_ds1.psh.fxo"},
	{0x855C772A0B398EFAULL, "466284545.ogg"},
	{0x855C9EFBAC185CFBULL, "627361277.ogg"},
	{0x855CA04709DBB29AULL, "vrc221_varric6040840.cl"},
	{0x855CEF41673C2141ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\heavydamage.ani"},
	{0x855D529A9C832F82ULL, "vxw_bonepit_21q_0_0.vlm"},
	{0x855D75A0A6A800D5ULL, "vxw_sundermount_2vd_0_0.vlm"},
	{0x855EB444B3AB41F3ULL, "hlac_lt_0_lowtown_docks_a3_-131073_3y.dds"},
	{0x855FC66BCE1307CEULL, "gen00fl_anders6119179.cl"},
	{0x85604E47FFD0FBFEULL, "mer240_jansen6050380.cl"},
	{0x8560A10E20555AA7ULL, "w_encounter3_34_0.rml"},
	{0x856120218038BB97ULL, "llac_lt_0_lowtown_docks_a3_131071_1.dds"},
	{0x856239A038CA0DB6ULL, "6096002_f.fxe"},
	{0x85626B78984DEB1AULL, "mag160_feynriel6107438.cl"},
	{0x856282AAD33953A9ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv2.evt"},
	{0x8562A6907383232FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\prp_guardianlegs.mao"},
	{0x8562B290738337CCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\prp_guardianlegs.mmh"},
	{0x8562B890738341BEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\prp_guardianlegs.msh"},
	{0x8562C1FE12E95C03ULL, "6036305_m.fxe"},
	{0x8562EE23322D8580ULL, "160236574.ogg"},
	{0x856342774D95729EULL, "vxw_woundedcoast_vi_0_0.vlm"},
	{0x8563BF43BC3071B8ULL, "lt_0_lowtown_20_0_ll.mmh"},
	{0x8563CE6512A0BD09ULL, "gen00fl_anders6061156.cl"},
	{0x8563DB75B70B1FF7ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_5z.dds"},
	{0x8563F6D730B8104AULL, "vxdr_slavers_136_0_0.vlm"},
	{0x8563FA41674252E3ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\heavydamage.gad"},
	{0x85648EDBF9BE7FCFULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_f_c.mmh"},
	{0x8564C373C7E87D13ULL, "vxlt_0_lowtown_a3_n_c_hjt_0_0.vlm"},
	{0x8564C630D5300F7CULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_imp_p.anb"},
	{0x8564EE0B00FBBE50ULL, "801500815.ogg"},
	{0x8564F27AC1519E9FULL, "192804832.ogg"},
	{0x856507B6BEC3A6FEULL, "dae000_arishok6096339.cl"},
	{0x856569C5452D430EULL, "w_woundedcoast_199_0.msh"},
	{0x856577C5452D5B5FULL, "w_woundedcoast_199_0.mao"},
	{0x85657DD93C5DC83EULL, "gen00flst_carver.stg"},
	{0x856583C5452D6FFCULL, "w_woundedcoast_199_0.mmh"},
	{0x8565B2157E083015ULL, "lgt200ip_eau_tome03.utp"},
	{0x856738AE36C8542EULL, "vxlt_undercityserialkiller_15p_0_0.vlm"},
	{0x8567A00E205B77BFULL, "w_encounter3_34_0.phy"},
	{0x8567B2FAE007482DULL, "851604544.ogg"},
	{0x856830741FEF583BULL, "6053784_m.fxe"},
	{0x85686EB4CAC7D3C4ULL, "gen00fl_anders6125372.cl"},
	{0x8568763945044FE6ULL, "vxlt_0_lowtown_a3_n_c_ey1_0_0.vlm"},
	{0x85688C3E73185D8AULL, "vxht_mansion_n_32i_0_0.vlm"},
	{0x856959F53888D066ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft_1.mao"},
	{0x8569BDEA8FFE41B3ULL, "follower_banter6172994.cl"},
	{0x856A0068CDAB3305ULL, "vxdr_0_deeproads_epz_0_0.vlm"},
	{0x856A1B1D0553C81AULL, "vxca_caverns1_jg0_0_0.vlm"},
	{0x856A461FDA40A904ULL, "vxca_cavern1_haunted_j5m_0_0.vlm"},
	{0x856A76334F9C3548ULL, "115259003.ogg"},
	{0x856A7F67BA31902CULL, "6078285_m.fxe"},
	{0x856A8105BB8D5D48ULL, "art\\vfx\\creatures\\fxc_arch_fb_p.mmh"},
	{0x856AD08B7B61CC07ULL, "ht_0_hightown_37_0_lc.mmh"},
	{0x856AD741674835C8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\heavydamage.evt"},
	{0x856B42FBD8C8C24CULL, "vxca_caverns1_ec9_0_0.vlm"},
	{0x856B82B7E5713887ULL, "lgt320_mis_resp_swordhunter6150321.cl"},
	{0x856C28D6C7E09210ULL, "mcr310_argument6127886.cl"},
	{0x856C5C0089071548ULL, "gen00fl_anders6125581.cl"},
	{0x856C999AFBAC7B6AULL, "vxlt_0_lowtown_n_f5i_0_0.vlm"},
	{0x856CA90A7757984BULL, "vxlt_hangedman_s9_0_0.vlm"},
	{0x856CE81ED850DC5BULL, "118084711.ogg"},
	{0x856D4288909005C7ULL, "1068951566.ogg"},
	{0x856D794E85C699F3ULL, "follower_banter6141714.cl"},
	{0x856D9DD08AF8B209ULL, "gen00fl_carver6104460.cl"},
	{0x856E2B908B6DDBADULL, "llac_w_bonepitdestroyed_a3_196609_0.dds"},
	{0x856F0CCA22BB3B6EULL, "vxht_0_hightown_n_b02_0_0.vlm"},
	{0x856F135F994CCCACULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziertame.msh"},
	{0x856F195F994CD69EULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziertame.mmh"},
	{0x856F255F994CEB35ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziertame.mao"},
	{0x856F46B9D0BCF883ULL, "llac_lt_0_lowtown_65535_8.dds"},
	{0x856F85B29FDFAAE2ULL, "vxlt_0_lowtown_docks_n_b6d_0_1.vlm"},
	{0x856FCC7E43354EB4ULL, "vxlt_0_lowtown_a2_n_c_cfi_0_0.vlm"},
	{0x856FCD3C0FF62A61ULL, "llac_ht_mansion2b_818_5.dds"},
	{0x856FECA6A749776AULL, "lt_0_lowtown_docks_a3_n_1_0_ll.mmh"},
	{0x85700C528E3F3DCEULL, "mcr370_negotiation6080246.cl"},
	{0x85706AAAD3456818ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontrightlegattackv2.ani"},
	{0x85710C43CD5F4F43ULL, "6206911_m.fxe"},
	{0x857187A798EB589AULL, "6080171_f.fxe"},
	{0x8571AED973757297ULL, "24280566.ogg"},
	{0x8571B4543F8B8B1EULL, "vxw_woundedcoast_sj_0_0.vlm"},
	{0x857249D603EC43B6ULL, "926751479.ogg"},
	{0x857261822727561AULL, "gen_im_arm_cht_noba_t3.uti"},
	{0x8572B95C612CB734ULL, "vxlt_0_lowtown_a2_n_c_fe1_0_0.vlm"},
	{0x8573C79C7BF1F2EBULL, "1024355186.ogg"},
	{0x857411C9B0130F2BULL, "927027538.ogg"},
	{0x8574392EA91B3499ULL, "vxca_caverns1_mines_ea7_0_0.vlm"},
	{0x8574ED93D0FE2ABAULL, "6031852_f.fxe"},
	{0x8575A38175ACE84CULL, "407775789.ogg"},
	{0x8575BCC3C4E9A9A0ULL, "vxdr_primevaltomb_8d8_0_0.vlm"},
	{0x8575F14C8D73AB73ULL, "445624558.ogg"},
	{0x85763299DA8436DFULL, "vxht_0_hightown_n_cco_0_0.vlm"},
	{0x857675C9F8A2019AULL, "llac_lt_undercityserialkiller_-1_17.dds"},
	{0x857698C73281568BULL, "gen_im_trs_krk_grd_junk1_03.uti"},
	{0x8576BE2AE03FE994ULL, "mag110_vincento6032995.cl"},
	{0x8576F5E5BE4C9E8CULL, "vxht_0_hightown_bcz_0_0.vlm"},
	{0x8577166BC976E918ULL, "244058887.ogg"},
	{0x85772B4B0CD07BDDULL, "and121_karl6093874.cl"},
	{0x85774E306E187043ULL, "vxlt_0_lowtown_a2_e2m_0_0.vlm"},
	{0x8577B066E4B199FFULL, "vxdr_slavers_1bh_0_0.vlm"},
	{0x857887F5C2001D62ULL, "vxht_mansion_n_60b_0_0.vlm"},
	{0x85788AE187CBF897ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_mainwaterfal.msh"},
	{0x85789CE187CC17FAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_mainwaterfal.mao"},
	{0x8578A0E187CC1E09ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_mainwaterfal.mmh"},
	{0x8578A10E2069ECC4ULL, "w_encounter3_34_0.mmh"},
	{0x8578A50E2069F317ULL, "w_encounter3_34_0.mao"},
	{0x8578B70E206A1266ULL, "w_encounter3_34_0.msh"},
	{0x8578F3AE2AC3138FULL, "vxca_cave_2bj_0_0.vlm"},
	{0x8578FA52C5562D21ULL, "6119414_f.fxe"},
	{0x85798F332B4BEA81ULL, "6189023_m.fxe"},
	{0x857993B1C4916146ULL, "6075712_m.fxe"},
	{0x8579A42E15C9EF10ULL, "hlac_dr_0_deeproads_5849_3y.dds"},
	{0x857A0C6D01AAC536ULL, "886926836.ogg"},
	{0x857A4383EC1756F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180_alt.ani"},
	{0x857ACAB2D9503702ULL, "6076780_m.fxe"},
	{0x857ADB30E4EDFF81ULL, "1007197769.ogg"},
	{0x857B515B98B7C84EULL, "595706204.ogg"},
	{0x857B950F8ACA538BULL, "llac_lt_0_lowtown_a2_n_c_-196609_1.dds"},
	{0x857C1393173D769AULL, "mage_bloodwound.dds"},
	{0x857CA94F8FB94F67ULL, "vxca_caverns2prc_6t6_0_0.vlm"},
	{0x857CE8CFEEF6131CULL, "mag291_demon_marethari6028865.cl"},
	{0x857CEF8B8F0F389AULL, "6017503_m.fxe"},
	{0x857D0158096EBA6AULL, "gen00fl_dog6167016.cl"},
	{0x857D1FA8AB1C1913ULL, "w_sundermount_195_0.phy"},
	{0x857D41AF5028EA00ULL, "vxht_0_hightown_a2_anj_0_0.vlm"},
	{0x857DE6FAD757B6F1ULL, "cod_bks_lite_vials_2.utp"},
	{0x857E2041B50842F3ULL, "684471012.ogg"},
	{0x857E604C5A36C735ULL, "l_mrl111ar_house.bnk"},
	{0x857EAB6FE4F4F7B3ULL, "6118513_m.fxe"},
	{0x857ED6F78E335F6FULL, "gen_im_arm_boo_med_act3_01.uti"},
	{0x857EF4ED473EFEDFULL, "757232454.ogg"},
	{0x857F989EDD853972ULL, "6048495_m.fxe"},
	{0x858003A2AFF776F7ULL, "ht_brothel_n_3371.rml"},
	{0x85800882DAE93433ULL, "765195464.ogg"},
	{0x8580191BF0DE10DAULL, "qun221_petrice6083649.cl"},
	{0x85804083EC1BBD56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180_alt.gad"},
	{0x858052A247D2C19EULL, "564256332.ogg"},
	{0x85805F1943A17D7CULL, "vxlt_darktown_52t_0_1.vlm"},
	{0x858061BFA7E9480DULL, "vxw_woundedcoast_1fs_0_0.vlm"},
	{0x8580920D278684AFULL, "dre000_dougal6129812.cl"},
	{0x8580BF5959DCCCFAULL, "789959621.ogg"},
	{0x8582B38E6E5EC6A0ULL, "ht_0_hightown_4_0.rml"},
	{0x858323300887FC9DULL, "vxlt_0_lowtown_eoq_0_1.vlm"},
	{0x85833A3A1307296DULL, "vxlt_0_lowtown_a2_exq_0_0.vlm"},
	{0x858357DDE8E59B5EULL, "hlac_ca_caverns1_24424_11y.dds"},
	{0x8583980B2C42135BULL, "248391002.ogg"},
	{0x8583FA0A33BD1D6BULL, "284797848.ogg"},
	{0x85846B7C47C00694ULL, "ser121_jethann6175440.cl"},
	{0x8584D3446D04C804ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\mwoverlay.anb"},
	{0x85853DD4A06B0F4AULL, "vxht_0_hightown_a2_b7l_0_0.vlm"},
	{0x8585B02DE6A51C52ULL, "6032865_m.fxe"},
	{0x85868192AAE75B1AULL, "follower_banter6173132.cl"},
	{0x85870EE997F8FA7CULL, "hlac_lt_0_lowtown_a3_n_c_131072_3z.dds"},
	{0x85873F2D444ED1EAULL, "llac_w_encounter2_camp_65540_0.dds"},
	{0x85875B83EC2209B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_180_alt.evt"},
	{0x85875F05EBD4EC43ULL, "vxw_encounter3_dp_0_0.vlm"},
	{0x8587FB339DA21448ULL, "6218601_m.fxe"},
	{0x8588273789182481ULL, "6019168_m.fxe"},
	{0x8588380DD30189C3ULL, "vxca_caverns1_ik6_0_0.vlm"},
	{0x85883E116F6BDC5FULL, "hlac_lt_0_lowtown_docks_-65539_0z.dds"},
	{0x8588782428B8F581ULL, "359836827.ogg"},
	{0x8588C9940F146819ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc2a.phy"},
	{0x858912D9A383F86EULL, "vxdr_primevalentrance_392_0_0.vlm"},
	{0x858932655A28BD89ULL, "6153503_m.fxe"},
	{0x85897314DCCDEBD9ULL, "hlac_ht_0_hightown_a2_-131074_2y.dds"},
	{0x858974A71D977D48ULL, "6070215_f.fxe"},
	{0x858A2A62B9F31D3DULL, "6148187_m.fxe"},
	{0x858ABCB71FB97EA7ULL, "dae000_gamlen6048486.cl"},
	{0x858ADBB3156792C4ULL, "624807140.ogg"},
	{0x858B1A23A76C62C7ULL, "vxlt_gallowscourtyard_50v_0_0.vlm"},
	{0x858B56CEEE57550AULL, "mrl150_merrill6070767.cl"},
	{0x858B62C5454E06C7ULL, "w_woundedcoast_199_0.phy"},
	{0x858B71D6B6BDFCB9ULL, "vxw_woundedcoast_p8_0_0.vlm"},
	{0x858B85BDC6B33550ULL, "gen00fl_aveline6122107.cl"},
	{0x858C4876CFAB0689ULL, "729381331.ogg"},
	{0x858C5FA53FB1A25EULL, "160727371.ogg"},
	{0x858C80A2147314C7ULL, "vxht_mansion2_3cl_0_0.vlm"},
	{0x858C8B574AD3D4CBULL, "6034538_m.fxe"},
	{0x858C8C99606750A7ULL, "6080748_m.fxe"},
	{0x858CA69D8E20289AULL, "vxw_encounter2_camp_o3_0_0.vlm"},
	{0x858CD1E97E3C845CULL, "vxca_caverns1_cjt_0_0.vlm"},
	{0x858D0AF97EAEFC17ULL, "6027659_m.fxe"},
	{0x858D1D4CD6EB1033ULL, "6117413_m.fxe"},
	{0x858D2F14EB98AD86ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bzrk.ani"},
	{0x858D52E7C87797DAULL, "6147202_m.fxe"},
	{0x858D6D5C6AFC2DDCULL, "dae000ip_crowd_mf6_talk17.utp"},
	{0x858D8FDBF9E1860CULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_f_c.anb"},
	{0x858DD1569C90A115ULL, "6140204_m.fxe"},
	{0x858DDB7EE015E930ULL, "hf_hayder_henchman.mor"},
	{0x858E0A7BCF7B51F2ULL, "6043595_m.fxe"},
	{0x858E8F028B600239ULL, "vxw_bonepit_1x2_0_0.vlm"},
	{0x858EB3BB90648F9DULL, "hlac_lt_gallowstemplar_f_2382_3z.dds"},
	{0x858ED39C99DF608DULL, "vxw_woundedcoast_1ar_0_0.vlm"},
	{0x858F02E04C1BF03EULL, "399815474.ogg"},
	{0x858F57D5D71F575AULL, "vxca_cavern1_haunted_jfj_0_0.vlm"},
	{0x858F591775C96EC3ULL, "vxlt_gallowstemplar_n_46d_0_0.vlm"},
	{0x858F60FC8CF218E1ULL, "313944500.ogg"},
	{0x858FE29C875F3837ULL, "vxw_blightlands_123_0_0.vlm"},
	{0x858FF91743B7908EULL, "161145075.ogg"},
	{0x85900237EABC509DULL, "prj_base.gda"},
	{0x859060EF3C574E08ULL, "mcr370cs_epilogue.cut"},
	{0x8590B260624986C8ULL, "945880901.ogg"},
	{0x85910282B30036E2ULL, "6105482_m.fxe"},
	{0x85913D4C52602AA8ULL, "dae000_gamlen6165966.cl"},
	{0x8591D459AA8F4092ULL, "hlac_ht_0_hightown_-131074_1y.dds"},
	{0x8591F9D9E5719530ULL, "588905430.ogg"},
	{0x8591FE6863C59900ULL, "492160141.ogg"},
	{0x859200C3EBA12AADULL, "mrl150_merrill.bnk"},
	{0x859261B7E5BCB73AULL, "1053929926.ogg"},
	{0x859269CC67287B96ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_whoresign_d.dds"},
	{0x8592B46C64025D3EULL, "one110_tree_minder6133765.cl"},
	{0x8592C4B491A4C2D7ULL, "hlac_ht_playershouse_n_855_13z.dds"},
	{0x85933510C8AB5586ULL, "322210222.ogg"},
	{0x85939227167B2879ULL, "718885521.ogg"},
	{0x8593AFB1926127CAULL, "vxht_0_hightown_a2_b8m_0_0.vlm"},
	{0x8593DE4E8ABDCC76ULL, "541666893.ogg"},
	{0x85940164B1DF658DULL, "one300_zevran6071485.cl"},
	{0x8594C66A627AA8B9ULL, "hlac_lt_0_lowtown_docks_-65538_5x.dds"},
	{0x8594CB0D8316C3FFULL, "792995989.ogg"},
	{0x8594CBB9C67F70AAULL, "6147743_m.fxe"},
	{0x859508F2B6CAB8A8ULL, "mer221_sabin6065659.cl"},
	{0x859509E8D6F483FFULL, "6130043_m.fxe"},
	{0x859511CA7949A1D1ULL, "6121061_m.fxe"},
	{0x85951332D1806263ULL, "949767294.ogg"},
	{0x859522C3EBA39A15ULL, "mrl150_merrill.cnv"},
	{0x859529DD2086F2D3ULL, "l_dae321ar_player_home_nt.wbd"},
	{0x859555FFA38D7C85ULL, "vxdr_0_deeproads_ecg_0_0.vlm"},
	{0x85963DEBAB9BA030ULL, "6140780_m.fxe"},
	{0x85966C6EFCB253B6ULL, "hlac_lt_undercitydungeon_-1_23x.dds"},
	{0x85969B9031359D2CULL, "vxht_keep_4hm_0_0.vlm"},
	{0x8596A9EAC4F7A243ULL, "vxca_caverns1_in9_0_0.vlm"},
	{0x859717A6F8F5AE40ULL, "hlac_lt_gallowstemplar_n_2384_4x.dds"},
	{0x859790211BE8BEFDULL, "vxht_0_hightown_a3_caa_0_0.vlm"},
	{0x8598A825B96C333DULL, "hlac_lt_0_lowtown_65533_0x.dds"},
	{0x8598FE412DDBD952ULL, "402820521.ogg"},
	{0x8598FEFCBCA8FDF3ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsblock.phy"},
	{0x8599C242C0836944ULL, "llac_lt_0_lowtown_n_-65536_8.dds"},
	{0x859A01D68C6DB73BULL, "mcr310_argument6074404.cl"},
	{0x859A3A7227FD3CBEULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv1.ani"},
	{0x859A9A5CA6042BF2ULL, "78867688.ogg"},
	{0x859A9D7952F933B5ULL, "6093562_m.fxe"},
	{0x859AC9F4500F4819ULL, "vxht_0_hightown_a2_bz4_0_0.vlm"},
	{0x859AF27F06692D47ULL, "vxht_mansion2_3fo_0_0.vlm"},
	{0x859B1714EBA4C2F7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bzrk.evt"},
	{0x859B61EA688EFE64ULL, "lgt260cr_way_merc_lead.utc"},
	{0x859B8F29C8E128B3ULL, "6112713_m.fxe"},
	{0x859BB626D691575EULL, "vxlt_0_lowtown_docks_a3_n_cyj_0_0.vlm"},
	{0x859BE0C122B93671ULL, "vxht_0_hightown_n_c6b_0_0.vlm"},
	{0x859C3E7DD4BE9BD9ULL, "vxht_0_hightown_c14_0_0.vlm"},
	{0x859C65B76594234DULL, "vxht_bartrand_n_5p9_0_0.vlm"},
	{0x859C9D26609D18EBULL, "6055074_m.fxe"},
	{0x859CBA4C8FE55247ULL, "vxlt_gallowscourtyard_n_7hc_0_0.vlm"},
	{0x859CCE4602637D13ULL, "6116476_m.fxe"},
	{0x859CD7B96EC72E7FULL, "gen00fl_anders6125841.cl"},
	{0x859D45798BD5790DULL, "vxw_woundedcoast_1lq_0_0.vlm"},
	{0x859DEE5C2B1FE827ULL, "148678708.ogg"},
	{0x859E1B42F200C55FULL, "6116371_m.fxe"},
	{0x859FE2DFF331FEDDULL, "113595643.ogg"},
	{0x85A02716EA0FDA1EULL, "vxw_bonepit_1ns_0_0.vlm"},
	{0x85A087CD984A273FULL, "6148422_m.fxe"},
	{0x85A0C9AF7043412FULL, "lgt110tr_orz_harrow_betray.utt"},
	{0x85A11814EBA92F80ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bzrk.gad"},
	{0x85A123C42FD7B99CULL, "vxht_0_hightown_c3k_0_0.vlm"},
	{0x85A1A40E208CF6BBULL, "w_encounter3_34_0.anb"},
	{0x85A1DB7231B7CE6AULL, "vxlt_warehouse_n_103_0_0.vlm"},
	{0x85A1EB565CD7DDE3ULL, "vxlt_hangedman_2fp_0_0.vlm"},
	{0x85A33D96B875892AULL, "6140643_m.fxe"},
	{0x85A381065328EE9AULL, "w_sundermount_67_0_lc.mmh"},
	{0x85A3D7F4DA6188A3ULL, "vxlt_undercityencounter_1c2_0_1.vlm"},
	{0x85A4270CC947990FULL, "hightown_male_noble_gossip6171425.cl"},
	{0x85A483ECE74E9119ULL, "hlac_lt_0_lowtown_docks_a3_-131073_2x.dds"},
	{0x85A498EC0BC7123DULL, "222387128.ogg"},
	{0x85A4C3775DD0C112ULL, "6099071_m.fxe"},
	{0x85A4E5F5875E2466ULL, "vxdr_primevaltomb_8xj_0_0.vlm"},
	{0x85A4E7D561E32417ULL, "6030645_f.fxe"},
	{0x85A51CA0BBAEE9EDULL, "6032057_m.fxe"},
	{0x85A56DF3D9CFF674ULL, "6066721_m.fxe"},
	{0x85A5F325A123346FULL, "vxlt_blackemporium_xl_0_0.vlm"},
	{0x85A7032D2EFB2AD5ULL, "llac_lt_gamlenshouse_n_-1_1.dds"},
	{0x85A71085B3F84954ULL, "hlac_lt_0_lowtown_-3_3x.dds"},
	{0x85A72CBBBEF77DCBULL, "vxht_mansion2_2ge_0_0.vlm"},
	{0x85A736CC2C2F7F7DULL, "sho_clo.mao"},
	{0x85A752722807EF8FULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv1.evt"},
	{0x85A8ACEE52ED1B80ULL, "gen_im_trs_gen_one_axe_r0.uti"},
	{0x85A8C87BBAF830F3ULL, "llac_w_bonepitmining_a2_196610_0.dds"},
	{0x85A915B193870555ULL, "lgt250_way_dalish_conflict6147110.cl"},
	{0x85A944429614FDFEULL, "vxca_caverns1_gk4_0_1.vlm"},
	{0x85A9B2D1585AB4F2ULL, "213678236.ogg"},
	{0x85A9C2C823B19C1FULL, "w_bonepitmining_a2_56_0.msh"},
	{0x85A9C49AFA03ACA3ULL, "6155598_m.fxe"},
	{0x85A9C8C823B1A6C1ULL, "w_bonepitmining_a2_56_0.mmh"},
	{0x85A9D4C823B1BB62ULL, "w_bonepitmining_a2_56_0.mao"},
	{0x85AA0F703F8E9722ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicechair_d.dds"},
	{0x85AA529E14AF4EF1ULL, "vxht_0_hightown_n_c3a_0_0.vlm"},
	{0x85AA715E4305854FULL, "qun220cr_javaris6071908.cl"},
	{0x85AA864C5A5C42BFULL, "l_mrl111ar_house.wbd"},
	{0x85AAE244C538CE83ULL, "vxdr_primevaltomb_6xe_0_0.vlm"},
	{0x85AB8417098231F2ULL, "6135073_f.fxe"},
	{0x85ACC61CDE2BADEEULL, "vxlt_0_lowtown_a2_n_ep0_0_1.vlm"},
	{0x85ACC9CB9EBAABC2ULL, "w_bonepitdestroyed_a3_24_0.msh"},
	{0x85ACCBCB9EBAAF78ULL, "w_bonepitdestroyed_a3_24_0.mmh"},
	{0x85ACD7CB9EBAC31BULL, "w_bonepitdestroyed_a3_24_0.mao"},
	{0x85ACEA0A31C2D80CULL, "921191384.ogg"},
	{0x85AD07388219E793ULL, "vxht_brothel_n_6c1_0_0.vlm"},
	{0x85AD81B920B8101CULL, "vrc221_varric6043973.cl"},
	{0x85AD9B855AA6C6B7ULL, "532648964.ogg"},
	{0x85AE5372280E1038ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\rangedattackv1.gad"},
	{0x85AE7827352BA96FULL, "hlac_lt_0_lowtown_docks_a3_-3_8z.dds"},
	{0x85AE8AC18D5F7391ULL, "6103201_m.fxe"},
	{0x85AEE6AE666BC9A2ULL, "l_dae340ar_bone_pit.wbd"},
	{0x85AF6B7FE91B2268ULL, "hlac_lt_0_lowtown_a2_n_-131071_1y.dds"},
	{0x85AF95A121CDD21CULL, "vxht_0_hightown_c6h_0_0.vlm"},
	{0x85AFF782AFCC746CULL, "vxw_bonepitmining_a2_7c_0_0.vlm"},
	{0x85B0552C59CED809ULL, "6087053_m.fxe"},
	{0x85B0CBCB9EBE978BULL, "w_bonepitdestroyed_a3_24_0.rml"},
	{0x85B0F0A57F1755EDULL, "vxht_keep_n_6gh_0_0.vlm"},
	{0x85B17779265F4E67ULL, "vxca_caverns2prc_a6s_0_0.vlm"},
	{0x85B19DAC7104019FULL, "652699534.ogg"},
	{0x85B24FA9C99E9018ULL, "darktown_male_fereldan_sta6175141.cl"},
	{0x85B2B5AFD86B8E07ULL, "follower_banter6170206.cl"},
	{0x85B2BAF2F4F800FBULL, "825879534.ogg"},
	{0x85B31744567E34AFULL, "189714959.ogg"},
	{0x85B323EC475B4469ULL, "dae000cr_grand_cleric6148089.cl"},
	{0x85B353FFE61B543FULL, "611938832.ogg"},
	{0x85B3B83F1113C299ULL, "vxht_0_hightown_n_a7k_0_0.vlm"},
	{0x85B3C4669D8E8F2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_enter.gad"},
	{0x85B417A113A62670ULL, "45957718.ogg"},
	{0x85B49F1D9EED4CCFULL, "vxw_sundermount_25o_0_0.vlm"},
	{0x85B4B6ECDEC812B3ULL, "6087280_m.fxe"},
	{0x85B4C0ACB6884973ULL, "184978518.ogg"},
	{0x85B4EE28DDCD6AFEULL, "hlac_ht_0_hightown_n_65535_2x.dds"},
	{0x85B4F841C6874456ULL, "vxht_0_hightown_n_cap_0_0.vlm"},
	{0x85B50B3367730403ULL, "dae011_conspiracy_mob6149581.cl"},
	{0x85B564805F411554ULL, "vxdr_primevaltomb_2ws_0_0.vlm"},
	{0x85B58C5B284FFAABULL, "6116313_m.fxe"},
	{0x85B59E58D02A497CULL, "vxlt_gallowscourtyard_7be_0_0.vlm"},
	{0x85B5A64E2E3EECF7ULL, "6153310_m.fxe"},
	{0x85B6290B4D482D11ULL, "vxlt_0_lowtown_a2_n_c_hjk_0_0.vlm"},
	{0x85B6402A832627ACULL, "180007466.ogg"},
	{0x85B6E156C6042F8AULL, "174520007.ogg"},
	{0x85B71B6F329F5225ULL, "vxlt_gallowstemplar_n_3s2_0_0.vlm"},
	{0x85B748B1FEF71303ULL, "gen_im_trs_crc_lgt_boo_r0.uti"},
	{0x85B829EB72A3059AULL, "vxdr_slavers_12q_0_0.vlm"},
	{0x85B944E6F28C4FFDULL, "vxlt_gallowstemplar_n_5j3_0_0.vlm"},
	{0x85B9A7B0679BE14AULL, "6090848_f.fxe"},
	{0x85B9C3ECB6BC6F39ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_4.mao"},
	{0x85B9E49961424BB9ULL, "vxw_bonepit_1k7_0_0.vlm"},
	{0x85BA29336FC1AA0DULL, "vxw_woundedcoast_1ro_0_0.vlm"},
	{0x85BA2A9849112B31ULL, "vxht_0_hightown_a2_8px_0_0.vlm"},
	{0x85BA788BD81C3A01ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtc.phy"},
	{0x85BAA7669D947CC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_enter.evt"},
	{0x85BABDD55B8792FAULL, "llac_w_blightlands_327676_0.dds"},
	{0x85BB9B37E22F8DD2ULL, "llac_lt_0_lowtown_docks_n_-65537_0.dds"},
	{0x85BCBBBB634F19DCULL, "6083521_f.fxe"},
	{0x85BD5F188D30CA23ULL, "vxw_bonepitmining_a2_pa_0_0.vlm"},
	{0x85BDCBC5A0F7C769ULL, "50899907.ogg"},
	{0x85BE469539ECD163ULL, "mag101_grace6034062.cl"},
	{0x85BE4B545A2BD59FULL, "6178542_m.fxe"},
	{0x85BE7F35615E8DD7ULL, "6036420_m.fxe"},
	{0x85BEB48FC410FC6BULL, "vxdr_primevaltomb_7u7_0_0.vlm"},
	{0x85BED217EC1108E9ULL, "6108987_m.fxe"},
	{0x85BF6282710D6E6DULL, "vxht_keep_n_6bg_0_0.vlm"},
	{0x85C10F6F7D0C4FA9ULL, "vxca_cavern1_haunted_i0x_0_0.vlm"},
	{0x85C179A62C28539EULL, "vxca_caverns1_egw_0_0.vlm"},
	{0x85C2005A9F9B1AEDULL, "6038657_m.fxe"},
	{0x85C25D6E82F19D5FULL, "325123612.ogg"},
	{0x85C2D6DF850F656FULL, "vxlt_blackemporium_vn_0_0.vlm"},
	{0x85C3FE45C8C2EAFAULL, "lowtown_male_lowlife1_sta6164989.cl"},
	{0x85C41035C22061FCULL, "vxlt_gallowscourtyard_7gz_0_0.vlm"},
	{0x85C44BBB76C30CF2ULL, "vxw_sundermount_31g_0_0.vlm"},
	{0x85C4E12AE5760539ULL, "504607821.ogg"},
	{0x85C4E861F6BD52DDULL, "hlac_ht_playershouse_n_855_6z.dds"},
	{0x85C4F2383CE71779ULL, "vxca_cave_2e9_0_0.vlm"},
	{0x85C5669625F2E41BULL, "6102820_m.fxe"},
	{0x85C5BB4C8232470FULL, "997637565.ogg"},
	{0x85C5FA22F422AB0CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9pp_cap.msh"},
	{0x85C60022F422B5FEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9pp_cap.mmh"},
	{0x85C60C22F422C995ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9pp_cap.mao"},
	{0x85C6B8DC9D5A23A8ULL, "character_character_p87.psh.fxo"},
	{0x85C6F85A6CCF0EB4ULL, "hlac_ht_bartrand_n_3370_2x.dds"},
	{0x85C70CB25C09C5A1ULL, "art\\levels\\ztools\\default\\generic\\0rw_basic.phy"},
	{0x85C728760F60FDC3ULL, "vxca_caverns2_5ab_0_1.vlm"},
	{0x85C75D509669A6B7ULL, "vxlt_0_lowtown_docks_a3_d1k_0_0.vlm"},
	{0x85C764D941DAB4ADULL, "character_character_p75.psh.fxo.dx11"},
	{0x85C79F669D9EF910ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_enter.ani"},
	{0x85C82A271D86EE98ULL, "6176169_m.fxe"},
	{0x85C83E85A48DBFADULL, "p_dust.mao"},
	{0x85C859012E88739AULL, "vxlt_0_lowtown_docks_coh_0_0.vlm"},
	{0x85C8B3EBF5DDB3F2ULL, "375279388.ogg"},
	{0x85C94480C2CB15D2ULL, "dae170ip_weapon_store.utp"},
	{0x85C9675C52E15674ULL, "vxlt_warehouse_n_qm_0_0.vlm"},
	{0x85C9966E08A7EAA7ULL, "vxlt_0_lowtown_a2_n_hfg_0_0.vlm"},
	{0x85C9A566F3FBBF7FULL, "vxw_encounter2_camp_u_0_0.vlm"},
	{0x85C9DB736A0ADFD6ULL, "vxlt_gallowstemplar_431_0_0.vlm"},
	{0x85CA156FF01309BDULL, "vxdr_primevaltomb_3i9_0_0.vlm"},
	{0x85CA1CDD9A14FF37ULL, "vxlt_darktown_2sb_0_0.vlm"},
	{0x85CA5E3F99698B0CULL, "6079651_m.fxe"},
	{0x85CABA60B5E65D21ULL, "6054365_m.fxe"},
	{0x85CAD3EC46657BC4ULL, "gen00fl_aveline6139529.cl"},
	{0x85CAD6E4F7008310ULL, "601411593.ogg"},
	{0x85CB4A3E7FA9FE24ULL, "6162379_m.fxe"},
	{0x85CB64DE1A6E5153ULL, "6157163_m.fxe"},
	{0x85CB8CC6A1D6427DULL, "face_rendertodepth_p1.psh.fxo"},
	{0x85CBD22F18AB918CULL, "211255736.ogg"},
	{0x85CBDCF0757BC08BULL, "520297572.ogg"},
	{0x85CC131B85F6B0CCULL, "6072172_f.fxe"},
	{0x85CC780A5EE6A406ULL, "mag370_meredith6092164.cl"},
	{0x85CC832B0DAE3A53ULL, "hlac_ca_caverns2_5802_39x.dds"},
	{0x85CC904349A8CD2DULL, "6034239_m.fxe"},
	{0x85CCD316F006A483ULL, "zz_dae_debug6029795.cl"},
	{0x85CCDB3C93B8A56CULL, "vxw_bonepitmining_a2_9a_0_0.vlm"},
	{0x85CCF2E2F35F98A5ULL, "vxlt_gallowstemplar_3qo_0_0.vlm"},
	{0x85CD0AD900D9D352ULL, "398532839.ogg"},
	{0x85CD52EFDCB3ED4FULL, "zz_mag_debug6053401.cl"},
	{0x85CDCA46AF231BFFULL, "vxw_bonepit_29x_0_0.vlm"},
	{0x85CDEFF4209E2A49ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_act_c.impa.ani"},
	{0x85CE54FF287A21D0ULL, "13497495.ogg"},
	{0x85CF37A195D37D0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_en.evt"},
	{0x85CF505C6B053030ULL, "vxht_brothel_n_6fy_0_0.vlm"},
	{0x85CF693AFBF54D2DULL, "125630161.ogg"},
	{0x85CFCF1253FBF7CCULL, "6129265_f.fxe"},
	{0x85D005EC94BA059CULL, "vxlt_0_lowtown_docks_aq6_0_0.vlm"},
	{0x85D03E642C1EE4E9ULL, "framebuffereffects_twopassblurshadows_p0.psh.fxo"},
	{0x85D04A81E131D090ULL, "119100501.ogg"},
	{0x85D04AF9F08620B4ULL, "vxlt_backalley_g_1pa_0_0.vlm"},
	{0x85D0D5441D275358ULL, "vxlt_hangedman_2mr_0_0.vlm"},
	{0x85D190DE1A284BC1ULL, "ser191_emeric.bnk"},
	{0x85D2128584B848C2ULL, "vxlt_0_lowtown_h2e_0_0.vlm"},
	{0x85D23F2B0FE2BB92ULL, "6072555_m.fxe"},
	{0x85D31684F622C31DULL, "6139845_f.fxe"},
	{0x85D3438386CF7BC7ULL, "l_env_wal_hightown_calm_p2p.bnk"},
	{0x85D3AA25ED9D28EAULL, "vxht_0_hightown_a2_9v0_0_0.vlm"},
	{0x85D3CBBA8520C9D1ULL, "mag100_cullen6095128.cl"},
	{0x85D3CEE34AE56AFAULL, "6046911_f.fxe"},
	{0x85D40AF420A2C3CBULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_act_c.impa.gad"},
	{0x85D47EAE90691501ULL, "6129928_m.fxe"},
	{0x85D4B2DE1A2ABB31ULL, "ser191_emeric.cnv"},
	{0x85D5041DD9DBEB6FULL, "zz_lgt_1_debug6114384.cl"},
	{0x85D50AC7837FAE56ULL, "141662453.ogg"},
	{0x85D52AB98F503C28ULL, "character_character_p76.psh.fxo"},
	{0x85D546636019D2EDULL, "6123650_m.fxe"},
	{0x85D5BFB333888151ULL, "6041763_m.fxe"},
	{0x85D5F0C7A8145496ULL, "661700460.ogg"},
	{0x85D6032AD88E6A54ULL, "vxht_playershouse_n_1ss_0_0.vlm"},
	{0x85D61B5B1F443B2DULL, "hlac_lt_hangedman_-1_11z.dds"},
	{0x85D62EA195D98CB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_en.gad"},
	{0x85D6B563EC86F14BULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.cess.evt"},
	{0x85D793FEAECEA117ULL, "follower_banter6143330.cl"},
	{0x85D7B6C5E11F9531ULL, "6164426_m.fxe"},
	{0x85D7BF1195F4D2E7ULL, "hlac_lt_0_lowtown_a2_n_-196608_4z.dds"},
	{0x85D7D93944D76EF4ULL, "vxlt_warehouse_n_tr_0_0.vlm"},
	{0x85D88B952BF39204ULL, "679346346.ogg"},
	{0x85D8DC5F914BCC01ULL, "vxht_brothel_n_611_0_0.vlm"},
	{0x85D9378375AABB26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwielddeselectedexit.evt"},
	{0x85D997DF41975CE3ULL, "31534401.ogg"},
	{0x85DA151EB8B3828BULL, "6191451_m.fxe"},
	{0x85DAAF26B44096BFULL, "vxw_bonepitmining_a2_7l_0_0.vlm"},
	{0x85DABCA4364327B2ULL, "lgt310_mis_burn_dump6180595.cl"},
	{0x85DB02203B9EE5ADULL, "6031139_m.fxe"},
	{0x85DB07F420A8DD30ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_act_c.impa.evt"},
	{0x85DB196ADDB64F64ULL, "vxht_0_hightown_a3_bxr_0_0.vlm"},
	{0x85DBAD9349ED5AF2ULL, "hlac_dr_primevaltomb_664_7y.dds"},
	{0x85DBDD256C0A24A6ULL, "688098647.ogg"},
	{0x85DC424E0C08DA26ULL, "vxlt_gallowstemplar_n_39c_0_0.vlm"},
	{0x85DC962005BAF725ULL, "hf_gen_refugee1.mor"},
	{0x85DCEA65A7020B3FULL, "hla_ht_0_hightown_a3_-196608_0y.dds"},
	{0x85DCFA893BAB2588ULL, "vxlt_0_lowtown_e2v_0_0.vlm"},
	{0x85DD3FA195DFC807ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_en.ani"},
	{0x85DDCE63EC8D3A0CULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.cess.gad"},
	{0x85DE664E756BEA9AULL, "ht_0_hightownexplosion_n_19_0.rml"},
	{0x85DE89D2B93E0CA9ULL, "6070451_m.fxe"},
	{0x85DEFAABAA8A30B6ULL, "dae000_gamlen6147209.cl"},
	{0x85DF18A8AB6ED788ULL, "w_sundermount_195_0.mmh"},
	{0x85DF24A8AB6EEC2BULL, "w_sundermount_195_0.mao"},
	{0x85DF36A8AB6F0AB2ULL, "w_sundermount_195_0.msh"},
	{0x85DF4E8375AF4E71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwielddeselectedexit.gad"},
	{0x85DFE295C3A6A96AULL, "1017089581.ogg"},
	{0x85E0347C6F32B020ULL, "vxht_0_hightown_n_c5j_0_0.vlm"},
	{0x85E069EC2DB23CE9ULL, "gen00fl_merrill6119569.cl"},
	{0x85E184B3FA358EC5ULL, "vxlt_0_lowtown_f1q_0_0.vlm"},
	{0x85E198FE46EC89C7ULL, "6176207_m.fxe"},
	{0x85E21F3F7ACFFEC0ULL, "vxw_bonepit_1wg_0_1.vlm"},
	{0x85E235B8A9B11C62ULL, "6152198_m.fxe"},
	{0x85E2908302D9D6CDULL, "6120152_m.fxe"},
	{0x85E39C96814654A8ULL, "character_character_p25.psh.fxo"},
	{0x85E3CD63EC91A40AULL, "art\\vfx\\combat\\vfx_sg_hurl_proj_p.cess.ani"},
	{0x85E47507CA8482D4ULL, "vxht_playershouse_n_1vp_0_0.vlm"},
	{0x85E49F081F629DCDULL, "vxht_bartrand_n_5o4_0_0.vlm"},
	{0x85E5822A93FFE68AULL, "vxlt_hangedman_n_1yl_0_0.vlm"},
	{0x85E60C66098A5B2BULL, "6153596_m.fxe"},
	{0x85E630598196694FULL, "dre000_stroud6036722.cl"},
	{0x85E64F8375B56E1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwielddeselectedexit.ani"},
	{0x85E65E3C8C9294A0ULL, "151981098.ogg"},
	{0x85E6B30B71B4391DULL, "1038190191.ogg"},
	{0x85E8379B3F698C10ULL, "w_encounter3_103_0.mao"},
	{0x85E83B9B3F699323ULL, "w_encounter3_103_0.mmh"},
	{0x85E8499B3F69AB7DULL, "w_encounter3_103_0.msh"},
	{0x85E87EF7A4FA91CCULL, "vxlt_0_lowtown_docks_a3_coq_0_0.vlm"},
	{0x85E8893D5AB14310ULL, "limp_dot.mao"},
	{0x85E8B4AF631313A3ULL, "vxw_bonepitmining_a2_ab_0_0.vlm"},
	{0x85E9DCA5253D7F68ULL, "gen_im_cns_flu_anders_04.uti"},
	{0x85EA05F92ABA0478ULL, "dae000cr_grand_cleric6148107.cl"},
	{0x85EA169E00332768ULL, "700481916.ogg"},
	{0x85EA2B910B74178CULL, "998722586.ogg"},
	{0x85EA4F93451F1377ULL, "247262260.ogg"},
	{0x85EA89BFCE888F29ULL, "cloud_f_bl.dds"},
	{0x85EAB035F6BA82FCULL, "art\\levels\\races\\proxy\\props\\_textures\\altar_sa.dds"},
	{0x85EAB42AFDFEF2A6ULL, "vxlt_gallowstemplar_n_34d_0_0.vlm"},
	{0x85EAD60FBD1F22E1ULL, "vxht_0_hightown_auv_0_0.vlm"},
	{0x85EAFCB6A7538EC4ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_7.dds"},
	{0x85EB7C0413E217DFULL, "vxht_brothel_n_5jq_0_0.vlm"},
	{0x85EC967C906E7648ULL, "722094921.ogg"},
	{0x85ECAD6B805DF47CULL, "vxlt_0_lowtown_docks_bdr_0_0.vlm"},
	{0x85ECCF3D020A9D1EULL, "vxca_caverns1_ehj_0_0.vlm"},
	{0x85ECD42ED443E30AULL, "712914640.ogg"},
	{0x85ECE1D36F69C8B6ULL, "ave200_aveline_bar6116191.cl"},
	{0x85ED0CB606D4E6A5ULL, "hlac_w_bonepit_393217_0x.dds"},
	{0x85ED19559BBE36CFULL, "cd_dre000_carver.wbd"},
	{0x85ED2EB3D47251B4ULL, "vxlt_backalley_g_1zg_0_0.vlm"},
	{0x85ED85BDC566F8B1ULL, "art\\vfx\\cinematic\\vfx_ship_wake_p.anb"},
	{0x85EDB25310E8E37BULL, "312411069.ogg"},
	{0x85EDD08E61A2B777ULL, "dae000_bartrand6030645.cl"},
	{0x85EDE0BD8C7C1B30ULL, "6053255_m.fxe"},
	{0x85EE1DD959376C6EULL, "560196834.ogg"},
	{0x85EE2DA5601CE697ULL, "setflag.ncs"},
	{0x85EED7E1915321E4ULL, "dae011_wench_mob6149850.cl"},
	{0x85EF036AAED3602BULL, "hlac_lt_0_lowtown_a2_n_-3_7z.dds"},
	{0x85EF08581E4C319AULL, "hlac_ht_keep_n_4519_1x.dds"},
	{0x85EF657B6FA3AC50ULL, "w_sundermount_71_0.mao"},
	{0x85EF697B6FA3B363ULL, "w_sundermount_71_0.mmh"},
	{0x85EF777B6FA3CABDULL, "w_sundermount_71_0.msh"},
	{0x85EFA11165BC318BULL, "static_lava_p0.psh.fxo"},
	{0x85F05D6ED61342C8ULL, "421452790.ogg"},
	{0x85F077EE9D46F277ULL, "eye_eye_p0.psh.fxo.dx11.assembly"},
	{0x85F0911C6CC61740ULL, "vxw_bonepit_1xf_0_1.vlm"},
	{0x85F095A2BA95EBAFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_critical_shot.gad"},
	{0x85F0F9028363A9F3ULL, "6153853_m.fxe"},
	{0x85F118183A2B3F80ULL, "hlac_lt_gallowstemplar_2394_2x.dds"},
	{0x85F1372CA7A48DB9ULL, "131639846.ogg"},
	{0x85F174B237233A72ULL, "hlac_dr_dungeon_3214_13z.dds"},
	{0x85F2428DF23ACB13ULL, "tal_sha_clu.mao"},
	{0x85F25F33A35A7F1AULL, "vxht_0_hightown_n_asx_0_0.vlm"},
	{0x85F274B02AE4C65CULL, "vxht_0_hightown_a2_ai7_0_0.vlm"},
	{0x85F2C4A95197F6A0ULL, "vxw_encounter2_camp_qr_0_1.vlm"},
	{0x85F2C8827A3DBD71ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffold_1.mao"},
	{0x85F34090FFCB9026ULL, "mer320_hubert6053174.cl"},
	{0x85F37566181ED67CULL, "vxht_keep_n_4ay_0_0.vlm"},
	{0x85F3F40785F5FF0AULL, "vxlt_hangedman_n_1vm_0_0.vlm"},
	{0x85F45EC57E73CEC8ULL, "vxht_0_hightown_aus_0_0.vlm"},
	{0x85F4E6D0ED41CC91ULL, "vxdr_primevalentrance_2q6_0_2.vlm"},
	{0x85F551DAF13CCB50ULL, "dae021_brothel_onenter6196705.cl"},
	{0x85F599444E952F8BULL, "vxca_cavern1_haunted_hvm_0_0.vlm"},
	{0x85F5EFF08CBF7C39ULL, "vxlt_undercitydungeon_1ic_0_1.vlm"},
	{0x85F5F4AE66A7E458ULL, "l_dae340ar_bone_pit.bnk"},
	{0x85F6CCF49DA0BC03ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_ogre_victim_brother.ani"},
	{0x85F6F9A7DA8E14B1ULL, "dae000ip_crowd_lm4_talk18.utp"},
	{0x85F7268C55092C23ULL, "vxw_bonepitmining_a2_le_0_0.vlm"},
	{0x85F76DF1F6F3DC95ULL, "6036296_f.fxe"},
	{0x85F792E0982CC7BFULL, "vxw_bonepitmining_a2_9n_0_0.vlm"},
	{0x85F7AAA2BA9C2E5CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_critical_shot.evt"},
	{0x85F895CCA3D0EA9AULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_lp.gad"},
	{0x85F8C768C7CA739FULL, "vxlt_0_lowtown_docks_n_b8v_0_0.vlm"},
	{0x85F932BAD63D4A3EULL, "static_static_p32.psh.fxo"},
	{0x85F968F73A962E1BULL, "6148930_m.fxe"},
	{0x85F9969B895F3849ULL, "llac_lt_0_lowtown_docks_-4_0.dds"},
	{0x85F999F61AAE4611ULL, "lgt000cr_epic_merchant_2.utc"},
	{0x85F9EDE105D8305FULL, "vxht_brothel_n_5mp_0_0.vlm"},
	{0x85FAA6BDC571BB2AULL, "art\\vfx\\cinematic\\vfx_ship_wake_p.mmh"},
	{0x85FAF7FCBCFBBC68ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsblock.mmh"},
	{0x85FB00E7E6872770ULL, "lt_0_lowtown_docks_n_209_0.msh"},
	{0x85FB03FCBCFBD00BULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsblock.mao"},
	{0x85FB0EE7E6873E8AULL, "lt_0_lowtown_docks_n_209_0.mmh"},
	{0x85FB12E7E68745C1ULL, "lt_0_lowtown_docks_n_209_0.mao"},
	{0x85FB15FCBCFBEE92ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsblock.msh"},
	{0x85FB3CA4674E4B3DULL, "ave200_donnic6128945.cl"},
	{0x85FB8970032B092BULL, "sibev_item_acquired.ncs"},
	{0x85FB9F78D3F56012ULL, "6084899_m.fxe"},
	{0x85FBDCB49D06FDE3ULL, "6171541_m.fxe"},
	{0x85FC0023E8E3EFAFULL, "6110621_m.fxe"},
	{0x85FC4ADBF3370507ULL, "vxlt_0_lowtown_docks_n_cvx_0_1.vlm"},
	{0x85FC92594827B672ULL, "mrl150_merrill6072369.cl"},
	{0x85FEB8383EF21A14ULL, "528321960.ogg"},
	{0x85FEE09B5225D630ULL, "vxw_sundermount_22i_0_0.vlm"},
	{0x85FF29497E593670ULL, "848624596.ogg"},
	{0x85FF398386F4A5FDULL, "l_env_wal_hightown_calm_p2p.wbd"},
	{0x85FFA05C73666CA9ULL, "6169201_m.fxe"},
	{0x85FFB61B9442A540ULL, "6161284_m.fxe"},
	{0x85FFDBD24802B56AULL, "vxht_keep_n_3el_0_0.vlm"},
	{0x86002BD49398A31CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyo_0n.dds"},
	{0x86006A7B6FB22788ULL, "w_sundermount_71_0.phy"},
	{0x8600A65113CB7549ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.impa.ani"},
	{0x8601422F69D38980ULL, "vxw_sundermount_265_0_0.vlm"},
	{0x8601F6C8ACD1D247ULL, "vxw_encounter3_dl_0_0.vlm"},
	{0x86021F363E8CBA10ULL, "6114110_m.fxe"},
	{0x86029E31FF0023EBULL, "llac_ca_caverns1_24422_0.dds"},
	{0x8602A34667F6DEEFULL, "art\\characters\\creatures\\strider\\animation\\blendtree\\strider_overlay.anb"},
	{0x86034FC0DEAC1665ULL, "hlac_lt_0_lowtown_docks_a3_-3_3y.dds"},
	{0x86036BB365793A89ULL, "llac_w_encounter2_-131069_0.dds"},
	{0x8603FD6ADC63F09DULL, "6108785_m.fxe"},
	{0x86045D103C66EE52ULL, "259475460.ogg"},
	{0x86045ECF4ABAD587ULL, "hlac_lt_backalley_g_65537_0y.dds"},
	{0x86047475A1B032A2ULL, "vxca_caverns2_8t8_0_0.vlm"},
	{0x8604BF40007FDE86ULL, "art\\vfx\\materials\\vfx_drain_add.mao"},
	{0x8605902BAC428130ULL, "563294560.ogg"},
	{0x8605979D4658BAD5ULL, "lgt110_gxa_coa_meeting6117486.cl"},
	{0x8605B2A2BAA87975ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_critical_shot.ani"},
	{0x8605CE842C027F7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_twitch.ani"},
	{0x8605E6A360E06C6FULL, "454118015.ogg"},
	{0x8605F44E01B539EBULL, "140647702.ogg"},
	{0x86066CE2E068AF01ULL, "vxht_0_hightown_a2_ang_0_0.vlm"},
	{0x86067C2157EF5F5DULL, "171928337.ogg"},
	{0x8606C15113D00ECBULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.impa.gad"},
	{0x8606D722F45A0699ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9pp_cap.phy"},
	{0x86073945B9C08C1FULL, "vxlt_0_lowtown_docks_n_b5q_0_0.vlm"},
	{0x8607697B6FB84480ULL, "w_sundermount_71_0.rml"},
	{0x8607E05B98C3BAB6ULL, "lgt160_wk_soph_sophia6119833.cl"},
	{0x8607E4873B5B0F26ULL, "406913201.ogg"},
	{0x8607E7F0545553FEULL, "gen00fl_carver6108418.cl"},
	{0x8607FCA978200F2EULL, "44327752.ogg"},
	{0x860834C0F07CC2C1ULL, "6059259_f.fxe"},
	{0x860858344A1F8D5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_horrored.evt"},
	{0x86098527449ABFA3ULL, "6198690_m.fxe"},
	{0x8609AA2A20170A1AULL, "6128353_m.fxe"},
	{0x8609F0DA36C654C7ULL, "vxlt_gallowstemplar_n_46x_0_0.vlm"},
	{0x860A4EF312DE99F1ULL, "vxca_caverns2_4tk_0_0.vlm"},
	{0x860AB6E01C12C456ULL, "lt_0_lowtown_a2_n_c_34_0.anb"},
	{0x860AE3F49DB18CEDULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_ogre_victim_brother.gad"},
	{0x860AEDF244DCC4E9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galosmallhi.mao"},
	{0x860AF9F244DCD882ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galosmallhi.mmh"},
	{0x860AFBF244DCDC38ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galosmallhi.msh"},
	{0x860B00A47ADCE374ULL, "nat380_nathaniel6198827.cl"},
	{0x860B21F3C421220EULL, "fxe_water_form_a.mao"},
	{0x860B43EA73CA42F8ULL, "lt_blackemporium_mmap.dds"},
	{0x860B83A3DAD4EF31ULL, "llac_lt_gallowscourtyard_-131072_0.dds"},
	{0x860B9A322FD96386ULL, "l_dae221ar_player_home_nt.bnk"},
	{0x860BE3B4DA088DAAULL, "6147152_m.fxe"},
	{0x860BFDD832D6B856ULL, "dae000_arishok6075350.cl"},
	{0x860C1DF97A694AFBULL, "qcr000_arishok6060193.cl"},
	{0x860C48350CAE3019ULL, "hlac_lt_darktown_-1_26y.dds"},
	{0x860D5278441BEEB0ULL, "vxw_sundermount_27v_0_0.vlm"},
	{0x860D890848615BF9ULL, "t1_mue_bk1.tnt"},
	{0x860DA9C0DA4E0E5DULL, "dre000_varric6033316.cl"},
	{0x860DB0CCA3E37480ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_lp.ani"},
	{0x860DBE5113D62830ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.impa.evt"},
	{0x860DEAD5EA9444F8ULL, "hlac_w_bonepit_131072_0x.dds"},
	{0x860DEE76D332319AULL, "zz_fen_debug6165697.cl"},
	{0x860DF61CF8FD2549ULL, "cd_mag210_arianni.bnk"},
	{0x860E27F88638BDC0ULL, "6166384_m.fxe"},
	{0x860E3918CB3279EDULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumc.mmh"},
	{0x860E3D18CB32803EULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumc.mao"},
	{0x860E4632BBF0A2FCULL, "qun160_party_comment_1.cnv"},
	{0x860E4B18CB329873ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumc.msh"},
	{0x860E572A2D523C15ULL, "w_woundedcoast_27_0_ll.mmh"},
	{0x860F4B8501BD4340ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi05.phy"},
	{0x860F54D43FEEA8E9ULL, "6083116_m.fxe"},
	{0x860F66FE8C0BA5F1ULL, "219107700.ogg"},
	{0x860F6F344A25D289ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_horrored.gad"},
	{0x860FF81605043411ULL, "vxca_caverns1_mines_ile_0_1.vlm"},
	{0x86102F236FEFCCD1ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue_r.dds"},
	{0x86104936F13186F3ULL, "6149551_m.fxe"},
	{0x86111BFBE7927583ULL, "84757880.ogg"},
	{0x86116832BBF3124CULL, "qun160_party_comment_1.bnk"},
	{0x861173A982A3F839ULL, "and100_anders6096048.cl"},
	{0x861195C6855ED8E5ULL, "v_c_blood.mao"},
	{0x8611B58B0A46744DULL, "6131961_m.fxe"},
	{0x8611BE2885FA47ECULL, "496961722.ogg"},
	{0x8611F9C606348729ULL, "vxlt_0_lowtown_d4t_0_0.vlm"},
	{0x8612994907A2BDD1ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_0y.dds"},
	{0x8613B27F2F8220A4ULL, "dae000_merchant_generic_hm6172582.cl"},
	{0x8613B5FE79C09E1EULL, "vxw_bonepit_16k_0_0.vlm"},
	{0x8614073397BBBA07ULL, "vxdr_slavers_238_0_0.vlm"},
	{0x8614DEBFD25EC781ULL, "vxht_0_hightown_a2_aif_0_0.vlm"},
	{0x8614E0DEA58EB164ULL, "vxht_0_hightown_a3_bdv_0_0.vlm"},
	{0x8614E44B4A042DADULL, "wm_ico_backalley_roll.dds"},
	{0x861594A4C4174C60ULL, "vxlt_gallowstemplar_f_4qj_0_0.vlm"},
	{0x8615C7A4BD3C8075ULL, "265118709.ogg"},
	{0x861640344A2BA1D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_horrored.ani"},
	{0x8616A4B9CFC524ADULL, "hlac_lt_gallowstemplar_f_2384_1z.dds"},
	{0x8616DF0C55CAF2EBULL, "llac_w_sundermount_-196612_0.dds"},
	{0x86175F408AD1366AULL, "171212046.ogg"},
	{0x8617ACE48995A403ULL, "494813938.ogg"},
	{0x8617D9DBC8FBF30FULL, "653574351.ogg"},
	{0x8617E550948DC892ULL, "zz_lgt_1_debug6114361.cl"},
	{0x861871717B964672ULL, "vxca_caverns1_dp7_0_0.vlm"},
	{0x861884B087F4F1E7ULL, "vxw_woundedcoast_1ed_0_0.vlm"},
	{0x8618C57B01667FE2ULL, "780668290.ogg"},
	{0x8618D549AECE6E19ULL, "vxht_0_hightown_n_ajl_0_0.vlm"},
	{0x8618F7FF462AAB62ULL, "126005747.ogg"},
	{0x861922B6ECFA08FEULL, "712497804.ogg"},
	{0x8619E7A6794D3AC3ULL, "t3_boy_b01.tnt"},
	{0x8619F3E6F083FA2AULL, "vxlt_gallowstemplar_f_2ot_0_0.vlm"},
	{0x8619F5842C136A05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_twitch.gad"},
	{0x861AB4DB8A01F837ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_5x.dds"},
	{0x861AE545CD7F6919ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic.msh"},
	{0x861AE6DA86E17C20ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_2x.dds"},
	{0x861AEF45CD7F7A07ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic.mmh"},
	{0x861AF345CD7F8154ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic.mao"},
	{0x861B79E2A8005F85ULL, "994627158.ogg"},
	{0x861B87D66BD790BAULL, "28995241.ogg"},
	{0x861BC45536120730ULL, "vxw_sundermount_28w_0_0.vlm"},
	{0x861BCEC0BD831222ULL, "hlac_w_bonepitdestroyed_a3_458757_0x.dds"},
	{0x861D7D8BD870C03DULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtc.mao"},
	{0x861D818BD870C776ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtc.mmh"},
	{0x861D8B8BD870D874ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgtc.msh"},
	{0x861DC1EF3E756F80ULL, "hlac_w_encounter3_-131072_0z.dds"},
	{0x861DE46E5E286243ULL, "gen00fl_aveline6116707.cl"},
	{0x861E066CA3939BE1ULL, "y_crate_01.dds"},
	{0x861EBB13E3279F73ULL, "6142451_m.fxe"},
	{0x861EFB34E422C313ULL, "235698873.ogg"},
	{0x861F0AAF9714B11EULL, "vxw_woundedcoast_7f_0_0.vlm"},
	{0x861F32449E40A179ULL, "6069163_m.fxe"},
	{0x861F3FFC71DB606CULL, "vxdr_0_deeproads_bot_0_1.vlm"},
	{0x861F499E5BD8488AULL, "vxlt_hangedman_n_1eh_0_0.vlm"},
	{0x861F8F2F13A63E42ULL, "hlac_ca_caverns2prc_5802_26x.dds"},
	{0x861FBFDFCADD22A8ULL, "art\\characters\\weapons\\quivers\\w_qui_qu01a.msh"},
	{0x861FD1DFCADD4139ULL, "art\\characters\\weapons\\quivers\\w_qui_qu01a.mao"},
	{0x861FDDDFCADD55D2ULL, "art\\characters\\weapons\\quivers\\w_qui_qu01a.mmh"},
	{0x862016B6FDB6CBCDULL, "6016886_m.fxe"},
	{0x862039F9DF2D09CEULL, "w_bonepit_253_0.mmh"},
	{0x862045F9DF2D1E65ULL, "w_bonepit_253_0.mao"},
	{0x862053F9DF2D35BCULL, "w_bonepit_253_0.msh"},
	{0x8620995A52C9001CULL, "weapon2_nolit.psh.fxo"},
	{0x86209BB11D333659ULL, "qun211_braggart6075521.cl"},
	{0x8620EBF829E8D680ULL, "454568042.ogg"},
	{0x8621222B107E5F25ULL, "vxht_0_hightown_a2_aox_0_0.vlm"},
	{0x8621580E950033BFULL, "pro000_colonel_ewald6039940.cl"},
	{0x862159BD0515FA81ULL, "601770260.ogg"},
	{0x86215E0A72E0FB03ULL, "vxdr_primevaltomb_5a5_0_0.vlm"},
	{0x8621FCD898838A91ULL, "vxca_caverns2prc_a9x_0_0.vlm"},
	{0x86227ED656EDBA46ULL, "483059573.ogg"},
	{0x86228DD2C2FF9507ULL, "680022535.ogg"},
	{0x8622994EAE89454FULL, "872503730.ogg"},
	{0x8622C94E0C606A58ULL, "48284893.ogg"},
	{0x8622FB42DBEF62D4ULL, "mag360_thrask6044284.cl"},
	{0x862352BB9784C9E4ULL, "vxht_0_hightown_a3_bcw_0_0.vlm"},
	{0x8623605137126ADAULL, "vxlt_backalley_g_1ra_0_0.vlm"},
	{0x862362547DFE14E6ULL, "mcr310_loyalist_mages.cnv"},
	{0x8623ACC73B3ED7B7ULL, "ico_helm_light_1.dds"},
	{0x8623E0513B933DD1ULL, "vxlt_0_lowtown_docks_n_d2y_0_1.vlm"},
	{0x86243C292F83BA7AULL, "6074320_m.fxe"},
	{0x86248CF654D6853AULL, "6197258_m.fxe"},
	{0x86249D8384F784E1ULL, "vxht_0_hightown_aij_0_0.vlm"},
	{0x8624D0C8FD99A83DULL, "mag101_alain6170899.cl"},
	{0x8624D9BD7B56597CULL, "vxw_bonepitdestroyed_a3_1be_0_0.vlm"},
	{0x8625383BD8C7C566ULL, "vxw_woundedcoast_hs_0_0.vlm"},
	{0x8625396F86BC7C57ULL, "zz_mer_debug6180870.cl"},
	{0x8625890E0F1C6A45ULL, "185748538.ogg"},
	{0x86266C69CD871C84ULL, "20551045.ogg"},
	{0x86272E2A8DBC815FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ent.gad"},
	{0x862764547E0200A6ULL, "mcr310_loyalist_mages.bnk"},
	{0x86276BE8B20E9192ULL, "6049218_m.fxe"},
	{0x862794B6BB288F39ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keepthin.msh"},
	{0x86279EB6BB28A027ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keepthin.mmh"},
	{0x8627A2B6BB28A774ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keepthin.mao"},
	{0x862828623DB407B3ULL, "335118019.ogg"},
	{0x862851C42B293751ULL, "mrl150_merrill6070832.cl"},
	{0x86286AF936C7B7C6ULL, "6267702.ogg"},
	{0x862877F7A6CD8FC6ULL, "6139939_m.fxe"},
	{0x8628997619EFBC23ULL, "art\\vfx\\cinematic\\vfx_ground_pound_p.impa.evt"},
	{0x8628B5D77704A0C0ULL, "zz_ave_debug6064098.cl"},
	{0x8628F80AA4AC95B6ULL, "6018743_f.fxe"},
	{0x862911B25C5C98DDULL, "art\\levels\\ztools\\default\\generic\\0rw_basic.mao"},
	{0x862915B25C5C9F16ULL, "art\\levels\\ztools\\default\\generic\\0rw_basic.mmh"},
	{0x86291FB25C5CB014ULL, "art\\levels\\ztools\\default\\generic\\0rw_basic.msh"},
	{0x8629263998F831D4ULL, "6142041_m.fxe"},
	{0x8629B7E42754E9C3ULL, "gen00fl_sebastian6177906.cl"},
	{0x8629DDC29EFDFCACULL, "gen00fl_fenris6115337.cl"},
	{0x862A717B5FB8AB9DULL, "vxht_store_bv_0_0.vlm"},
	{0x862AE6CA1D781C66ULL, "mag200_feynriel6023957.cl"},
	{0x862AEA2B3AF62DA5ULL, "vxca_cavern1_haunted_dp9_0_0.vlm"},
	{0x862B0B69022ED2D5ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtbzz.phy"},
	{0x862BC7912F13557AULL, "6131122_m.fxe"},
	{0x862BD20C264A68FFULL, "weapon_vfx_p0.vsh.fxo.dx11.assembly"},
	{0x862BDEF6AD0B4C12ULL, "6148091_f.fxe"},
	{0x862BF11E0C613C80ULL, "253285418.ogg"},
	{0x862BFCF07E2ADA97ULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gsw_002a_s.dds"},
	{0x862BFD25ACB47D4BULL, "6017488_f.fxe"},
	{0x862C38730425C217ULL, "152728772.ogg"},
	{0x862C79C1EADC463FULL, "narrative_varric6172508.cl"},
	{0x862CEEA20EA013B2ULL, "67556841.ogg"},
	{0x862D1D253C326FEFULL, "368789408.ogg"},
	{0x862D32B6634B69C6ULL, "zz_prmpt_debug6198231.cl"},
	{0x862D61C008914484ULL, "6124709_f.fxe"},
	{0x862DBB7B4DCE610AULL, "vxlt_hangedman_n_1bi_0_0.vlm"},
	{0x862E232A8DC28D4CULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ent.evt"},
	{0x862E33559BF5F9A5ULL, "cd_dre000_carver.bnk"},
	{0x862E4463E505B0C5ULL, "vxlt_undercitydungeon_22d_0_0.vlm"},
	{0x862E8CFDB948D170ULL, "ser221_party_comments6166202.cl"},
	{0x862F5E1A8E8F4CD7ULL, "mcr370_negotiation6080146.cl"},
	{0x862F64EB6B0E8C3AULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidermode.gad"},
	{0x862F727619F598A4ULL, "art\\vfx\\cinematic\\vfx_ground_pound_p.impa.gad"},
	{0x862F92F1A194674FULL, "hlac_w_encounter3_-262139_0x.dds"},
	{0x862FCFE764D71383ULL, "vxdr_primevaltomb_5f4_0_0.vlm"},
	{0x862FD0E962613F9AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_n.dds"},
	{0x86300701B7C46F5AULL, "dae000cr_grand_cleric6148116.cl"},
	{0x86300C3C032AD7A7ULL, "66294860.ogg"},
	{0x863023D2C53D0ABAULL, "hlac_ht_chantry_n_2534_35x.dds"},
	{0x86304BD0C095E095ULL, "gen00fl_sebastian6174031.cl"},
	{0x863095BDCD68E13AULL, "hlac_w_woundedcoast_131078_0y.dds"},
	{0x8630AB3639F72ABAULL, "one100dg_lia6060524.cl"},
	{0x8630B91D9ED70BBEULL, "hlac_w_sundermount_-6_0z.dds"},
	{0x8631232362B75774ULL, "265221143.ogg"},
	{0x86328EAB6506773CULL, "llac_ca_cavern1_haunted_24423_1.dds"},
	{0x8632C773B427A98CULL, "vxca_caverns2_87s_0_0.vlm"},
	{0x86331668B797AEC3ULL, "vxlt_0_lowtown_a2_eqx_0_0.vlm"},
	{0x863407D3204BFB0EULL, "hlac_dr_0_deeproads_5845_8z.dds"},
	{0x86341F39F551D1CBULL, "art\\vfx\\creatures\\vfx_c_corpse_c.impa.evt"},
	{0x863439F28BA1B95AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_singletile.phy"},
	{0x86345DC79769B63BULL, "zz_and_debug6134124.cl"},
	{0x8634688501DC90D5ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi05.msh"},
	{0x86346A8501DC944BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi05.mmh"},
	{0x863493D72EE1B5BBULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightpath_n.dds"},
	{0x86349AC4716D88A6ULL, "vxca_caverns2_95z_0_0.vlm"},
	{0x86349BC0CEFBE302ULL, "937206180.ogg"},
	{0x8635329A2CD5344AULL, "6045618_m.fxe"},
	{0x8635634F6DB4F29FULL, "6178353_m.fxe"},
	{0x8635686A6BE122E7ULL, "vxw_woundedcoast_1kf_0_0.vlm"},
	{0x86363B767757EF19ULL, "6072829_f.fxe"},
	{0x863645FD9AA34609ULL, "6092341_m.fxe"},
	{0x863667EB6B14B051ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidermode.evt"},
	{0x86367B60527DA29BULL, "gen00fl_sebastian6180456.cl"},
	{0x8636817619FBD082ULL, "art\\vfx\\cinematic\\vfx_ground_pound_p.impa.ani"},
	{0x8636F30955AA18FFULL, "6119284_m.fxe"},
	{0x8636FE51A6FB7A2FULL, "conversation_icons.gda"},
	{0x863824FD45BA94F0ULL, "vxht_bartrand_n_6ql_0_0.vlm"},
	{0x86386343572BDFF3ULL, "vxlt_hangedman_n_2x9_0_0.vlm"},
	{0x863873146B78A75CULL, "vxht_brothel_n_6by_0_0.vlm"},
	{0x8639121C761005C4ULL, "6167295_m.fxe"},
	{0x86395C66777BBDE3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_06bonetrack.mao"},
	{0x86396066777BC430ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_06bonetrack.mmh"},
	{0x8639665CC08D85F2ULL, "6112091_m.fxe"},
	{0x86396E66777BDC4AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_06bonetrack.msh"},
	{0x863A0E3F3327317AULL, "vxca_caverns2prc_805_0_0.vlm"},
	{0x863A139D8C3489DCULL, "lgt261ip_treasure_pile.utp"},
	{0x863A18DEDA0F5959ULL, "6187952_f.fxe"},
	{0x863A8756B8E3C13CULL, "493901332.ogg"},
	{0x863ACFDB86E78B3FULL, "gen00fl_merrill6119637.cl"},
	{0x863B0B2A8DCCEEA5ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gptrl_ent.ani"},
	{0x863B3839F5581A8CULL, "art\\vfx\\creatures\\vfx_c_corpse_c.impa.gad"},
	{0x863B5189C36F26C1ULL, "mag170_thrask6020558.cl"},
	{0x863B5FC15B9D7ED3ULL, "6027757_f.fxe"},
	{0x863B63CF014879C6ULL, "6068879_m.fxe"},
	{0x863BB99AC9E44479ULL, "633946804.ogg"},
	{0x863C2377E9BFC3EDULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\til_ruins01_d.dds"},
	{0x863C981638424948ULL, "vxht_0_hightown_all_0_0.vlm"},
	{0x863D26D13FC2954BULL, "358384600.ogg"},
	{0x863D8E7775D2447FULL, "mcr370_guardsman6086987.cl"},
	{0x863D99E01C3DECEAULL, "lt_0_lowtown_a2_n_c_34_0.rml"},
	{0x863E18AB9F93EF7CULL, "6019104_f.fxe"},
	{0x863E9C4402A1546BULL, "6167289_m.fxe"},
	{0x863F22BAA979BDDAULL, "39810702.ogg"},
	{0x863F469FB48BB53CULL, "vxlt_0_lowtown_a2_hef_0_0.vlm"},
	{0x863F57B626C9DAA3ULL, "355411040.ogg"},
	{0x863F812BEDDC939EULL, "6156473_m.fxe"},
	{0x863F9B64DC7B56FCULL, "vxdr_primevaltomb_8nf_0_0.vlm"},
	{0x863FBAC6ACC34E33ULL, "6134901_f.fxe"},
	{0x863FD073AF138228ULL, "190143660.ogg"},
	{0x86402C76705C72BAULL, "dm_iwan.mor"},
	{0x8640341C70B3A54CULL, "321943778.ogg"},
	{0x8640505A2EE973CCULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc03.phy"},
	{0x8640790229C64BD7ULL, "zz_followers_banter6157878.cl"},
	{0x8640C0A0DEDE10F8ULL, "342913648.ogg"},
	{0x864118634E171F4CULL, "vxw_bonepitmining_a2_1x9_0_0.vlm"},
	{0x86413739F55C848AULL, "art\\vfx\\creatures\\vfx_c_corpse_c.impa.ani"},
	{0x86421BF5BCB3F666ULL, "vxw_woundedcoast_bu_0_0.vlm"},
	{0x864258CC4E156D2AULL, "hlac_ht_0_hightown_a3_-65537_3x.dds"},
	{0x864305416C2DECE1ULL, "gen00fl_merrill6129610.cl"},
	{0x86437FEB6B1F6320ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidermode.ani"},
	{0x8643D2035C416F93ULL, "art\\characters\\playercharacter\\humanmale\\hlf_wmda.phy"},
	{0x8644600486B36712ULL, "1019555538.ogg"},
	{0x86447D8044BA4FAEULL, "96156971.ogg"},
	{0x8644B7DA8C995E89ULL, "6097041_m.fxe"},
	{0x8644E310325F771CULL, "637363657.ogg"},
	{0x86450D7E15CBD0EAULL, "vxlt_0_lowtown_gky_0_0.vlm"},
	{0x864511BF128A41CFULL, "vxw_sundermount_2kq_0_0.vlm"},
	{0x86454F16B59E9CFEULL, "6189428_m.fxe"},
	{0x8645E87029809017ULL, "vxht_keep_5mx_0_0.vlm"},
	{0x8646698D87E62DE3ULL, "vxlt_0_lowtown_docks_cth_0_0.vlm"},
	{0x8646D845CDA48E3CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic.phy"},
	{0x86471F201631A36BULL, "llac_w_bonepit_458754_0.dds"},
	{0x864737C07CAA0E7AULL, "vxlt_0_lowtown_n_ghs_0_0.vlm"},
	{0x8647646280F77BADULL, "vxlt_gallowstemplar_5t6_0_0.vlm"},
	{0x8647753FCB3A21DCULL, "hlac_lt_gallowscourtyard_n_65534_0x.dds"},
	{0x8647ADBBF199E966ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.dura.ani"},
	{0x8647CCBB6122A13EULL, "295922290.ogg"},
	{0x8647D1417BD3F9C6ULL, "gen00fl_aveline6139457.cl"},
	{0x8647FB84CA4C0EAFULL, "729461947.ogg"},
	{0x8648D5D83D2C0657ULL, "llac_ht_mansion2b_818_56.dds"},
	{0x86490CC78C47F7F4ULL, "v_cage_02.dds"},
	{0x8649FD595C6B5C5CULL, "vxw_blightlands_9t_0_0.vlm"},
	{0x864A64541F21122AULL, "mag320_orsino6070415.cl"},
	{0x864AB2911F973958ULL, "zz_followers_banter6158838.cl"},
	{0x864ADF40C68F38CEULL, "vxw_sundermount_30h_0_0.vlm"},
	{0x864B1003FBB78CD7ULL, "6152627_m.fxe"},
	{0x864B3AAAC0FD0375ULL, "gen00fl_fenris6106666.cl"},
	{0x864B503F88532F95ULL, "gen_im_wep_rog_dua_act0_04.uti"},
	{0x864B866215ABD2FAULL, "41989001.ogg"},
	{0x864B9C32300F4B2CULL, "l_dae221ar_player_home_nt.wbd"},
	{0x864C03FF9352F369ULL, "6006076_m.fxe"},
	{0x864C632505540051ULL, "vxht_chantry_23s_0_0.vlm"},
	{0x864C8881D86D2774ULL, "combat_ar_r1_a5.gda"},
	{0x864C90BF597E6DB0ULL, "str_bigrock02_d.dds"},
	{0x864CDABA922804CAULL, "6151003_m.fxe"},
	{0x864D9FBB4D776E11ULL, "6176304_m.fxe"},
	{0x864DB87CA681CDBCULL, "vxlt_0_lowtown_a2_hbg_0_0.vlm"},
	{0x864DCA0FE5425FF2ULL, "vxw_bonepit_2a1_0_0.vlm"},
	{0x864DD07C68B6964EULL, "737418279.ogg"},
	{0x864E52ACB76412C4ULL, "hlac_lt_0_lowtown_a2_n_-131072_1x.dds"},
	{0x864E8B3307395105ULL, "mcr370_meredith6084435.cl"},
	{0x864F56CA5B14297DULL, "sib111_mother6098836.cl"},
	{0x864F7999AE3229FBULL, "563637542.ogg"},
	{0x86501584D61E2FAAULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.anb"},
	{0x8650E4B834458789ULL, "pro000cr_soldier3.utc"},
	{0x8651893EFD89EC42ULL, "gen00fl_merrill6108438.cl"},
	{0x8651D3577FA2FA3EULL, "vxlt_0_lowtown_a2_n_e27_0_0.vlm"},
	{0x865271FC525A9021ULL, "zz_fol_debug6066140.cl"},
	{0x8652B086D5A008ADULL, "cd_rdr111dg_assistant.wbd"},
	{0x8652B847B29C3B16ULL, "dae000_tomwise6167304.cl"},
	{0x8652C7AFFABBB7CAULL, "vxdr_primevaltomb_39j_0_0.vlm"},
	{0x865347D614A90784ULL, "6147543_m.fxe"},
	{0x86534C21FE617067ULL, "mag121ip_corthwaite_store.utp"},
	{0x8653839C04805A4FULL, "vxw_sundermount_2lp_0_0.vlm"},
	{0x865387B6BB4DB45CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keepthin.phy"},
	{0x8653B5A36BD6A822ULL, "w_woundedcoast_231_0.phy"},
	{0x8654092A881FB471ULL, "97845109.ogg"},
	{0x86543AD755199C83ULL, "459023828.ogg"},
	{0x8654B7F3B714149FULL, "zz_and_debug6134076.cl"},
	{0x8654C5C8417C9BBEULL, "39560597.ogg"},
	{0x8654C8FF8F95AB09ULL, "volumesmoke.mao"},
	{0x8654D41CF938ED9BULL, "cd_mag210_arianni.wbd"},
	{0x86558D71F166519BULL, "6106816_f.fxe"},
	{0x865595BBF1A5FDD7ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.dura.evt"},
	{0x865597C593528A21ULL, "vxca_caverns1_e8y_0_0.vlm"},
	{0x86561D3137F4236FULL, "490551027.ogg"},
	{0x86575E8D2A692317ULL, "vxht_0_hightown_c56_0_0.vlm"},
	{0x8657CF22F3CDA781ULL, "387959440.ogg"},
	{0x8658049DE9B4703CULL, "one280cr_merin.utc"},
	{0x865835691CC18FAFULL, "mcr310_confused_guard6077065.cl"},
	{0x865854E974090C5BULL, "art\\characters\\playercharacter\\humanfemale\\hlh_meda.msh"},
	{0x865866E974092AA6ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_meda.mao"},
	{0x865872E974093F45ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_meda.mmh"},
	{0x86588D920FB57E72ULL, "889901335.ogg"},
	{0x8658AFD8554B3B5AULL, "925080365.ogg"},
	{0x8659BD3439A3355BULL, "m_one100ar_ruin_ext.bnk"},
	{0x8659BFBA53412654ULL, "280792800.ogg"},
	{0x8659CA17B3ECF748ULL, "vxlt_undercitydungeon_258_0_0.vlm"},
	{0x865A0A28CD7E9384ULL, "zz_approval_debug6003643.cl"},
	{0x865A5C547E2D4CA9ULL, "mcr310_loyalist_mages.stg"},
	{0x865AEBF044C3F812ULL, "lt_0_lowtown_a3_n_c_20_0.anb"},
	{0x865B48C68D89F00DULL, "6105845_m.fxe"},
	{0x865B64BA0BE442F2ULL, "6095783_m.fxe"},
	{0x865B8416ABA8C30EULL, "hlac_ht_mansion2_818_42z.dds"},
	{0x865B9C8EE6A3114AULL, "6098591_m.fxe"},
	{0x865BBF05A856707DULL, "16871886.ogg"},
	{0x865BE7E51960C393ULL, "hlac_ht_0_hightown_0_2x.dds"},
	{0x865C3107F0194ADFULL, "mag330cr_walter.utc"},
	{0x865C3692BF85B0A0ULL, "121728944.ogg"},
	{0x865C7F1EC06787FCULL, "vxdr_primevaltomb_8dd_0_0.vlm"},
	{0x865C96BBF1AC1E60ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.dura.gad"},
	{0x865CCA686E6E4A3AULL, "vxw_woundedcoast_r9_0_0.vlm"},
	{0x865CD5656BA2594EULL, "follower_barks6153696.cl"},
	{0x865D018882FF4C63ULL, "character_rendertodepth_p9.psh.fxo.dx11"},
	{0x865D34DE63FCD3E1ULL, "hlac_lt_0_lowtown_n_-196609_4y.dds"},
	{0x865D48F9660BCE62ULL, "ip_restrict_attribute.ncs"},
	{0x865D8FDD552D6CE0ULL, "hlac_lt_undercitydungeon_-1_12z.dds"},
	{0x865DE1C4B3677380ULL, "t3_dog_ob1.tnt"},
	{0x865EFFAFA0A02766ULL, "vxw_woundedcoast_tw_0_0.vlm"},
	{0x865FF42910CFF535ULL, "vxlt_gallowstemplar_2mn_0_0.vlm"},
	{0x866008C49833EA6EULL, "hlac_lt_darktown_-1_17z.dds"},
	{0x866059C8ECBA6B6BULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.mmh"},
	{0x8661356CC6C8D4FBULL, "797655893.ogg"},
	{0x86617600C949FC18ULL, "gen00fl_sebastian6174663.cl"},
	{0x86618E419BD76A76ULL, "6058946_m.fxe"},
	{0x8661F578F67672CFULL, "vxw_sundermount_2qs_0_0.vlm"},
	{0x86622AB9725CFB27ULL, "6095364_m.fxe"},
	{0x8662640C05110179ULL, "48000047.ogg"},
	{0x8662B9E66EB97577ULL, "6092152_m.fxe"},
	{0x8662EC89972E804BULL, "gen_im_trs_cot_junk1_03.uti"},
	{0x86635B76E56E5F9FULL, "6102913_f.fxe"},
	{0x8663FAFB929CF513ULL, "dae120cr_hm_guard_3.utc"},
	{0x86641B7A60963F7AULL, "vxlt_0_lowtown_n_gbm_0_0.vlm"},
	{0x86641F9B91566648ULL, "423230022.ogg"},
	{0x8664423AA0C05A6AULL, "hlac_w_encounter1_-131070_0x.dds"},
	{0x8664490BAFDF09F7ULL, "244794148.ogg"},
	{0x86646094AD17606EULL, "pro000_refugee_male6052105.cl"},
	{0x866470427A96C4DDULL, "pro000_meeran6048966.cl"},
	{0x866533B0B5C21F6FULL, "one100_party_comments6112615.cl"},
	{0x86655B3FC1E5B36AULL, "pro000cr_merchants_guard_2.utc"},
	{0x866584FD50846849ULL, "960711399.ogg"},
	{0x8665D5E406B69297ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_extreme_v1.gad"},
	{0x8666223D864892C4ULL, "vxca_caverns2prc_8gr_0_0.vlm"},
	{0x86667CC1E6620FCEULL, "t3_abm_h02.tnt"},
	{0x86672ED9E483A6F8ULL, "foodstorage_d.dds"},
	{0x866750C5999E8D80ULL, "6214713_m.fxe"},
	{0x8667B2FD8E29F653ULL, "vxdr_primevaltomb_2vo_0_0.vlm"},
	{0x8668035352994D5CULL, "vxw_sundermount_26i_0_0.vlm"},
	{0x8668254AE6962DB9ULL, "6167979_m.fxe"},
	{0x866896EDF2AD2AEDULL, "vxht_0_hightown_a2_a78_0_0.vlm"},
	{0x8668E3EED364F4C7ULL, "104401728.ogg"},
	{0x8669DB526E70B570ULL, "zz_t_sir_issac_two_v1.cut"},
	{0x866A6038698569D7ULL, "6146965_m.fxe"},
	{0x866A8B71C9D04C2FULL, "hlac_ht_bartrand_n_3374_0y.dds"},
	{0x866AA0A45B732447ULL, "ts_beasts_03.gda"},
	{0x866AA1F2A817E714ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock_d.dds"},
	{0x866AB6519918BB48ULL, "20250528.ogg"},
	{0x866AB958091D3960ULL, "6126560_m.fxe"},
	{0x866AEC8D38B02B6DULL, "532863880.ogg"},
	{0x866B5282EAF581C7ULL, "vxdr_primevalentrance_3bv_0_0.vlm"},
	{0x866B5B272D6DC260ULL, "78628256.ogg"},
	{0x866B90F2C26286A8ULL, "one100st_ilen.stg"},
	{0x866BD52C5E69E0C0ULL, "27622553.ogg"},
	{0x866BF05A3E0D9C69ULL, "6174092_m.fxe"},
	{0x866BF976CB94671EULL, "vxw_bonepitmining_a2_1mt_0_0.vlm"},
	{0x866C3FDAD9571A21ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_c.impa.ani"},
	{0x866CAB8C9A17691AULL, "ht_0_hightown_a2_46_0.rml"},
	{0x866CEAE406BCD524ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_extreme_v1.evt"},
	{0x866D3AC8ECC4C0D0ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.anb"},
	{0x866D51BD1920DEACULL, "vxlt_hangedman_27t_0_0.vlm"},
	{0x866DAD6DB5506A14ULL, "vxlt_0_lowtown_a2_n_gxq_0_0.vlm"},
	{0x866DB34E133A5CFAULL, "ht_0_hightown_n_35_0_lc.mmh"},
	{0x866DC33E201B58DAULL, "hm_thrask.mor"},
	{0x866E44398B603C21ULL, "224586372.ogg"},
	{0x866E9C752C70E1A5ULL, "6206373_m.fxe"},
	{0x866F707A857F49DAULL, "da2terrain_terrain_tools_p0.psh.fxo.dx11"},
	{0x866F884B862E919FULL, "vxlt_gallowstemplar_n_3e2_0_0.vlm"},
	{0x86702B96F64EDCCBULL, "70548708.ogg"},
	{0x86706755E86C8B4FULL, "vxw_sundermount_2zr_0_0.vlm"},
	{0x86708EB867384566ULL, "vxw_bonepit_1yn_0_0.vlm"},
	{0x867128D80F68E023ULL, "ave221_brennan6110485.cl"},
	{0x8671291F3441A55BULL, "140427768.ogg"},
	{0x8671E1898EFDA9DCULL, "hlac_ht_mansion2b_818_57x.dds"},
	{0x86721E69026AF5F0ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtbzz.msh"},
	{0x86722C69026B0D0AULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtbzz.mmh"},
	{0x86722F64811E9DC5ULL, "vxlt_0_lowtown_n_ghx_0_0.vlm"},
	{0x86723412DEA0379DULL, "dae000_merchant_access_hf6162955.cl"},
	{0x86725D5B588425A5ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_app_p.mmh"},
	{0x867292A472204124ULL, "art\\characters\\weapons\\longswords\\w_lsw_016a.msh"},
	{0x8672A4A472205F8DULL, "art\\characters\\weapons\\longswords\\w_lsw_016a.mao"},
	{0x8672A8A4722066C6ULL, "art\\characters\\weapons\\longswords\\w_lsw_016a.mmh"},
	{0x8672BB84AB9F7843ULL, "vxdr_primevalentrance_37u_0_0.vlm"},
	{0x86734ADAD95D4BC3ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_c.impa.gad"},
	{0x8673D031E67BF142ULL, "ser270_templar6122497.cl"},
	{0x8673DC24E7E51467ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_0y.dds"},
	{0x86742C9B0FB9B007ULL, "rdr110dg_dockworker6015941.cl"},
	{0x867475E2505AEBCDULL, "qun110tr_isabela_leaves.utt"},
	{0x8674C7599078320AULL, "ht_0_hightown_a2_51_0.rml"},
	{0x8675155FAC30F227ULL, "vxca_caverns1_hii_0_0.vlm"},
	{0x867538B67839843CULL, "vxdr_primevalentrance_2q5_0_0.vlm"},
	{0x8675537B4835A358ULL, "vxlt_0_lowtown_docks_coz_0_0.vlm"},
	{0x86756D0476AF7CFAULL, "vxht_mansion_n_589_0_0.vlm"},
	{0x86756F7DF99B1E8FULL, "vxlt_0_lowtown_docks_a3_d2z_0_0.vlm"},
	{0x86757DF8D15063EFULL, "324269346.ogg"},
	{0x86758D2E2A72328AULL, "hlac_ht_0_hightown_a2_131071_1y.dds"},
	{0x8675F141534FCADAULL, "vxw_bonepitmining_a2_20l_0_0.vlm"},
	{0x8675F16BB940E88DULL, "vxw_woundedcoast_11b_0_0.vlm"},
	{0x8676037DE1D602BEULL, "vxca_caverns1_mines_j5w_0_0.vlm"},
	{0x8677011FA5E04D5BULL, "6110213_f.fxe"},
	{0x867717D56922FEC1ULL, "vxw_bonepitmining_a2_a8_0_0.vlm"},
	{0x86772FAE8260EF71ULL, "977258971.ogg"},
	{0x8677F832495A1057ULL, "150103559.ogg"},
	{0x8677FF7A090125BEULL, "fen110_fenris6100255.cl"},
	{0x867845CC4C298AD6ULL, "6117247_m.fxe"},
	{0x8678CE3DA59B0D8CULL, "302451850.ogg"},
	{0x86791884D641399DULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.mmh"},
	{0x867965D42B111187ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ci_basic_2.mao"},
	{0x867990A36BF738D6ULL, "w_woundedcoast_231_0.mao"},
	{0x86799CA36BF74D75ULL, "w_woundedcoast_231_0.mmh"},
	{0x86799EA36BF750EBULL, "w_woundedcoast_231_0.msh"},
	{0x867A02E406C7881DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_extreme_v1.ani"},
	{0x867A04E42BA4A6BAULL, "gen00fl_varric6099541.cl"},
	{0x867A27DAD9632EA8ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_c.impa.evt"},
	{0x867A50B54C38B411ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartospiderdeadlp.ani"},
	{0x867A666E2E6A6326ULL, "6198696_f.fxe"},
	{0x867A6B53BD8A7F9EULL, "vxw_bonepitmining_a2_1ju_0_0.vlm"},
	{0x867AF76EC19A578CULL, "vxht_0_hightown_b9h_0_0.vlm"},
	{0x867B0F0F8146998DULL, "gen_im_trs_bst_liver_03.uti"},
	{0x867B8E57FDDAFFD1ULL, "gen_im_cns_cft_wrn_01.uti"},
	{0x867C4C4E75F93305ULL, "6105650_m.fxe"},
	{0x867C5654FC41EF13ULL, "51375741.ogg"},
	{0x867CC54E464BDCB3ULL, "vxw_sundermount_36w_0_0.vlm"},
	{0x867D9FA4722A5E91ULL, "art\\characters\\weapons\\longswords\\w_lsw_016a.phy"},
	{0x867DFFFDE80E81F2ULL, "383153089.ogg"},
	{0x867E1A6BE5701BC5ULL, "dae330ar_undercity_darktown.are"},
	{0x867E1DE3583E7196ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3_tool.phy"},
	{0x867E24A54B5F6540ULL, "265195431.ogg"},
	{0x867E2BE36DEDE767ULL, "hlac_ht_0_hightown_a2_0_0x.dds"},
	{0x867E51AE0D187B90ULL, "318432553.ogg"},
	{0x867E5AA533D0B8AAULL, "540760615.ogg"},
	{0x867EADF03B4A46FDULL, "vxht_0_hightown_a3_c1q_0_0.vlm"},
	{0x867EF6859A0A8FA7ULL, "185901735.ogg"},
	{0x867F262A106D88CEULL, "hlac_ht_0_hightown_a2_-131069_0z.dds"},
	{0x867F5A5B588EAB72ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_app_p.anb"},
	{0x867FC41C51F0C736ULL, "mer220_hubert.stg"},
	{0x867FCDF9FBF8ED7CULL, "6147819_m.fxe"},
	{0x8680275A2F1F1197ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc03.mmh"},
	{0x86803C1BCF5E4115ULL, "vxlt_undercityserialkiller_247_0_0.vlm"},
	{0x86803D5A2F1F3709ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc03.msh"},
	{0x86807D3594435636ULL, "vxlt_backalley_g_1kr_0_0.vlm"},
	{0x868091F3A09CDEE6ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_abilities_v2.evt"},
	{0x86815BB54C3EE5B3ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartospiderdeadlp.gad"},
	{0x868173A2EDD933B1ULL, "hlac_dr_primevaltomb_11985_0y.dds"},
	{0x8681A57122178AE7ULL, "193880851.ogg"},
	{0x8682434D7FB8D0E8ULL, "vxdr_0_deeproads_eu8_0_0.vlm"},
	{0x8683C61839CFC508ULL, "vxca_cave_1oq_0_0.vlm"},
	{0x8683F78D2E78AEC7ULL, "vxw_encounter3_9g_0_0.vlm"},
	{0x868458FE699DD3FFULL, "121559228.ogg"},
	{0x86849803C56A17A8ULL, "art\\levels\\ztools\\default\\generic\\1oi_basic.msh"},
	{0x8684AA03C56A3639ULL, "art\\levels\\ztools\\default\\generic\\1oi_basic.mao"},
	{0x8684B603C56A4AD2ULL, "art\\levels\\ztools\\default\\generic\\1oi_basic.mmh"},
	{0x8684D898D4892072ULL, "gen_im_trs_cot_med_cht_r1.uti"},
	{0x86851142F110056CULL, "hlac_lt_0_lowtown_n_-131071_0x.dds"},
	{0x868558181FD6BEBBULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack1_c.impa.ani"},
	{0x8686B65835C2BFB7ULL, "29598581.ogg"},
	{0x8686C538F5B3562CULL, "6046874_f.fxe"},
	{0x8686D58B259A5338ULL, "distortiongeom.psh.fxo"},
	{0x868743462C452191ULL, "726009283.ogg"},
	{0x86876406753CE308ULL, "vxlt_0_lowtown_h92_0_0.vlm"},
	{0x8687974A10C69365ULL, "852865456.ogg"},
	{0x86879D11ED096A60ULL, "6128760_m.fxe"},
	{0x8687A8F3A0A32431ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_abilities_v2.gad"},
	{0x868806E358469F37ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3_tool.msh"},
	{0x868818E35846BD9AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3_tool.mao"},
	{0x86881B6BE57871ABULL, "dae330ar_undercity_darktown.ncs"},
	{0x86881CE35846C4A9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3_tool.mmh"},
	{0x868879A29EB2FA6CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_004a_0t.dds"},
	{0x86892726C7A6C220ULL, "dre000_stroud6036749.cl"},
	{0x86895199F08C48E3ULL, "hlac_lt_0_lowtown_n_65536_11z.dds"},
	{0x868A6D62FA134E81ULL, "vxw_sundermount_268_0_0.vlm"},
	{0x868A9C99DF06071CULL, "340323272.ogg"},
	{0x868AE576154471A7ULL, "vxlt_gallowscourtyard_7ew_0_0.vlm"},
	{0x868B9537772711F9ULL, "224471311.ogg"},
	{0x868B96E4BB3CB71DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3b.mao"},
	{0x868B9AE4BB3CBE56ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3b.mmh"},
	{0x868BA4E4BB3CCF54ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight3b.msh"},
	{0x868C1DCDC63661F0ULL, "705366821.ogg"},
	{0x868DEEF044EF574EULL, "lt_0_lowtown_a3_n_c_20_0.rml"},
	{0x868E027A2C2056B5ULL, "147667778.ogg"},
	{0x868E3263684D652FULL, "495790360.ogg"},
	{0x868E3D007F93F441ULL, "517244042.ogg"},
	{0x868E3FD6EDEF05FCULL, "6140919_m.fxe"},
	{0x868E7C206B8890DBULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumplrg.phy"},
	{0x868E953906050F64ULL, "hlac_lt_0_lowtown_65538_2z.dds"},
	{0x868EA9F3A0A944DFULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_abilities_v2.ani"},
	{0x868EEF1286396EB6ULL, "vxlt_backalley_g_1ps_0_0.vlm"},
	{0x868F1B6BE57E90F2ULL, "dae330ar_undercity_darktown.lst"},
	{0x868F46B091860053ULL, "hlac_w_sundermount_-3_0x.dds"},
	{0x868FECDC1B36D00FULL, "follower_banter6170501.cl"},
	{0x868FF87A1A89C2B5ULL, "qcr220_grey_wardens6060420.cl"},
	{0x869032F28BF00B83ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_singletile.msh"},
	{0x869040F28BF023DDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_singletile.mmh"},
	{0x869044F28BF02AEEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_singletile.mao"},
	{0x86910DB39B090FE0ULL, "112225407.ogg"},
	{0x8691AF272F3DA25FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_loop.gad"},
	{0x8691D25DC8CBE9D2ULL, "vxht_playershouse_n_1rj_0_0.vlm"},
	{0x8691D4AC5760DD7FULL, "vxlt_0_lowtown_docks_a3_d39_0_0.vlm"},
	{0x8691F7C6FE0ACA81ULL, "one300_zevran6093492.cl"},
	{0x86923B226DBBD6D8ULL, "439115193.ogg"},
	{0x86925C48DB1DF2A4ULL, "654662461.ogg"},
	{0x86925EE25073C240ULL, "qun110tr_isabela_leaves.ncs"},
	{0x869260181FE156B2ULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack1_c.impa.evt"},
	{0x8692696A206EC747ULL, "vxw_encounter3_2f_0_0.vlm"},
	{0x86928C57B1757900ULL, "801665204.ogg"},
	{0x86930BD2B54C0B35ULL, "5457185.ogg"},
	{0x8693BA86D5D7B0D7ULL, "cd_rdr111dg_assistant.bnk"},
	{0x8693EDF9047C4228ULL, "vxlt_hangedman_1w6_0_1.vlm"},
	{0x8693FC1BA06DA4A0ULL, "406034104.ogg"},
	{0x86944B5FE3CF4D23ULL, "986602123.ogg"},
	{0x86946671D9FE944EULL, "vxlt_0_lowtown_docks_a3_d27_0_0.vlm"},
	{0x8694A83EE57D264EULL, "6075481_f.fxe"},
	{0x8694A9E87ECDB4C4ULL, "vxdr_primevalentrance_2sk_0_0.vlm"},
	{0x8694B994E06FCA42ULL, "hm_anders.mor"},
	{0x8694C64E05D1ABE6ULL, "785685908.ogg"},
	{0x8694CDA28FE7FEA3ULL, "art\\vfx\\combat\\vfx_c_pride_s_p.impa.ani"},
	{0x8695500AF5C6D22FULL, "6189674_m.fxe"},
	{0x869553EBAC7139CBULL, "lowtown_argument6169425.cl"},
	{0x8695FA9635889CEDULL, "6119116_m.fxe"},
	{0x869602715E46A85CULL, "ser221_quentin6123367.cl"},
	{0x86960CE9E4EE0CBBULL, "lla_ht_0_hightown_a2_131073_0.dds"},
	{0x86963B4E69F59261ULL, "vxca_cave_2dw_0_1.vlm"},
	{0x86964692883FE9FCULL, "vxdr_primevaltomb_8px_0_0.vlm"},
	{0x869689FB7F99BABCULL, "6095631_m.fxe"},
	{0x8696C5201231190EULL, "vxlt_hangedman_n_27q_0_0.vlm"},
	{0x869723706F7BDE10ULL, "vxlt_ruinedhouse_n_2z_0_0.vlm"},
	{0x86976BFC5E76B0A6ULL, "hlac_ht_keep_n_3732_9z.dds"},
	{0x86977BECEC0F4691ULL, "vxw_woundedcoast_1gt_0_0.vlm"},
	{0x8697A5ADABEAA5BBULL, "6172571_m.fxe"},
	{0x8697B75A1B91868AULL, "plt_tbe100pt_free_tome.plo"},
	{0x8697C56AC156E421ULL, "421026165.ogg"},
	{0x8697D122FB8E4308ULL, "135790382.ogg"},
	{0x869808A0E431AEFEULL, "6122092_m.fxe"},
	{0x8698377F89F94C21ULL, "134352843.ogg"},
	{0x86986112D6CE449BULL, "307210475.ogg"},
	{0x8698C1352B4E9362ULL, "6060372_m.fxe"},
	{0x8698C2C1852A6408ULL, "vxlt_0_lowtown_eqk_0_0.vlm"},
	{0x8698E63D6301C3C6ULL, "vxlt_0_lowtown_n_ewt_0_0.vlm"},
	{0x8698F324AF784D26ULL, "zz_mer_debug6039342.cl"},
	{0x86991A231AE0267FULL, "364376110.ogg"},
	{0x86995B1101BEAC9AULL, "vxht_keep_n_5da_0_0.vlm"},
	{0x86997F181FE7AA45ULL, "art\\vfx\\combat\\vfx_c_rw_s2_attack1_c.impa.gad"},
	{0x8699F4EEA60EC2D3ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack02.ani"},
	{0x869AD6AAA40EDE43ULL, "556428306.ogg"},
	{0x869B58E2449C5468ULL, "vxlt_warehouse_10z_0_0.vlm"},
	{0x869B958D157D9EB5ULL, "qun221_viscount6084725.cl"},
	{0x869BC807EB7E098AULL, "zz_dae_debug6029068.cl"},
	{0x869BCBFD4C234AF4ULL, "dae000_arishok6095677.cl"},
	{0x869BE0689F6AEFC5ULL, "plt_prm_cod_adder_longbow.plo"},
	{0x869BF8DDC350DBCCULL, "6142368_m.fxe"},
	{0x869C3CD9F2ACE77EULL, "vxca_caverns2_8o9_0_0.vlm"},
	{0x869C3E6BB2EA747AULL, "vxca_caverns2prc_7vt_0_0.vlm"},
	{0x869C6CA94E807B1EULL, "6161105_m.fxe"},
	{0x869C7D94A119067DULL, "vxlt_hangedman_n_43_0_0.vlm"},
	{0x869CA4232C61E59DULL, "w_blightlands_90_0.phy"},
	{0x869D351DB36BBC87ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_small_1.mao"},
	{0x869D3B31706C7169ULL, "vxlt_undercitydungeon_105_0_0.vlm"},
	{0x869DF1C536476BF6ULL, "mag101st_decimus.stg"},
	{0x869E135FF0A1F956ULL, "qcr220cr_templar.utc"},
	{0x869ED49C6849D42CULL, "mag291_torpor6027001.cl"},
	{0x869F4632328E2909ULL, "sib130ip_at_family_estate.utp"},
	{0x869F47BA92048D9AULL, "vxdr_slavers_1ba_0_0.vlm"},
	{0x869F59AC113B8F75ULL, "tbe100_xenon6190978.cl"},
	{0x869F9B68D03A7719ULL, "107246649.ogg"},
	{0x869FB3A81AC7B1ABULL, "515085641.ogg"},
	{0x86A026ADCF669700ULL, "vxw_sundermount_2i8_0_0.vlm"},
	{0x86A046894956F5FFULL, "vxlt_0_lowtown_docks_a3_d48_0_0.vlm"},
	{0x86A04BE97446B2F2ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_meda.phy"},
	{0x86A05AF22437923DULL, "6028537_m.fxe"},
	{0x86A0B775391DA209ULL, "vxlt_warehouse_n_13e_0_0.vlm"},
	{0x86A1AB3439E0CC59ULL, "m_one100ar_ruin_ext.wbd"},
	{0x86A1E97697718449ULL, "dae000_bartrand6033985.cl"},
	{0x86A29E3F80A750E1ULL, "vxlt_0_lowtown_n_gl8_0_0.vlm"},
	{0x86A2E5A28FF4649AULL, "art\\vfx\\combat\\vfx_c_pride_s_p.impa.evt"},
	{0x86A394EDE8191A2BULL, "vxht_0_hightown_5h3_0_0.vlm"},
	{0x86A3A632B343B41CULL, "dae210_gate_guard6085143.cl"},
	{0x86A3E83EBB7F55A6ULL, "vxlt_0_lowtown_fg0_0_0.vlm"},
	{0x86A42712DE8E7767ULL, "vxlt_0_lowtown_docks_b57_0_1.vlm"},
	{0x86A510CEE18BEBA7ULL, "vxlt_0_lowtown_docks_bcl_0_0.vlm"},
	{0x86A536FD0427318EULL, "vxlt_hangedman_n_22r_0_0.vlm"},
	{0x86A5519824756565ULL, "131878865.ogg"},
	{0x86A55C3AB26C4B0DULL, "411197967.ogg"},
	{0x86A58C272F4E0FA5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_loop.ani"},
	{0x86A5CB035C942E08ULL, "art\\characters\\playercharacter\\humanmale\\hlf_wmda.mmh"},
	{0x86A5D7035C9442ABULL, "art\\characters\\playercharacter\\humanmale\\hlf_wmda.mao"},
	{0x86A5E9035C946132ULL, "art\\characters\\playercharacter\\humanmale\\hlf_wmda.msh"},
	{0x86A5FA396B4C31F3ULL, "433528862.ogg"},
	{0x86A62607A0823E90ULL, "vxca_cavern1_haunted_e8i_0_0.vlm"},
	{0x86A64F76CC77E1DBULL, "vxlt_gallowstemplar_5c9_0_0.vlm"},
	{0x86A6924AB8A4875FULL, "vxht_0_hightown_a2_c8m_0_0.vlm"},
	{0x86A6B060B52CF674ULL, "vxht_0_hightown_a3_cf9_0_0.vlm"},
	{0x86A6ECEEA6193F0AULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack02.evt"},
	{0x86A70D2A597AC222ULL, "t3_wdg_bw3.tnt"},
	{0x86A730921F20B309ULL, "vxht_0_hightown_a2_9bc_0_0.vlm"},
	{0x86A7627A6115371CULL, "vxht_0_hightown_n_c92_0_0.vlm"},
	{0x86A7D9CACE067195ULL, "575361004.ogg"},
	{0x86A862402F4C30DEULL, "hlac_lt_0_lowtown_a3_n_c_131073_0z.dds"},
	{0x86A862AD3377F0F3ULL, "llac_lt_0_lowtown_docks_-4_2.dds"},
	{0x86A94FB2A8441695ULL, "vxlt_0_lowtown_a2_n_fef_0_0.vlm"},
	{0x86A9D3C9EA536010ULL, "6172697_m.fxe"},
	{0x86A9D8132B013F77ULL, "vxlt_undercityencounter_ur_0_0.vlm"},
	{0x86A9E4A28FFA818DULL, "art\\vfx\\combat\\vfx_c_pride_s_p.impa.gad"},
	{0x86AA16A364DBEE42ULL, "151337024.ogg"},
	{0x86AA1DD876FB7FA1ULL, "6088728_m.fxe"},
	{0x86AA41C79B5CA35FULL, "366214829.ogg"},
	{0x86AA917BA8C97D2AULL, "1034555146.ogg"},
	{0x86AACC08E3CD2236ULL, "gen_im_trs_drk_shd_met_r0.uti"},
	{0x86AADE864076939EULL, "6166005_m.fxe"},
	{0x86AB48DAE9F10297ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillarwindow.phy"},
	{0x86AB7722AF985520ULL, "vxht_0_hightown_a2_cae_0_0.vlm"},
	{0x86ABA75050D7FFC4ULL, "6174422_m.fxe"},
	{0x86ABD4071B3F7BAEULL, "6119524_m.fxe"},
	{0x86ABDB02381C1745ULL, "dae020cr_em_servant.utc"},
	{0x86AC23F596FF14F2ULL, "hlac_dr_dungeon_3215_9x.dds"},
	{0x86AC8C3B68B16C8CULL, "93822103.ogg"},
	{0x86AC8CFB9575366EULL, "286509640.ogg"},
	{0x86AC99A05A51EA52ULL, "6080095_m.fxe"},
	{0x86ACB8E8F86B5EC7ULL, "dae021cr_male_worshipper.utc"},
	{0x86AD8E01AFEC3AAEULL, "vxca_cavern1_haunted_j20_0_2.vlm"},
	{0x86AD99232C704609ULL, "w_blightlands_90_0.mao"},
	{0x86ADA5232C705AA2ULL, "w_blightlands_90_0.mmh"},
	{0x86ADA7232C705E58ULL, "w_blightlands_90_0.msh"},
	{0x86ADAE080F56CAD7ULL, "ts_profane_wpn_table.gda"},
	{0x86ADAE2B789494C7ULL, "vxlt_0_lowtown_a2_eqd_0_0.vlm"},
	{0x86ADCBEEA61F267DULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\attack02.gad"},
	{0x86AE25DFE73CE26DULL, "hlac_dr_slavers_3215_4z.dds"},
	{0x86AE988AC15CAF80ULL, "vxw_sundermount_2r9_0_0.vlm"},
	{0x86AF17537ACA83EEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\choke.ani"},
	{0x86AF29522B13BA89ULL, "vxlt_warehouse_n_18d_0_0.vlm"},
	{0x86AFCF68072F6D8CULL, "vxdr_primevaltomb_6uy_0_0.vlm"},
	{0x86AFDD55CC9A6564ULL, "ser221_gamlen6134819.cl"},
	{0x86B0A1B2571109EFULL, "774346473.ogg"},
	{0x86B0AA3772269954ULL, "vxlt_warehouse_115_0_0.vlm"},
	{0x86B0C11EB02FD707ULL, "lt_0_lowtown_a2_n_20_0.rml"},
	{0x86B0DDB2FB0EA935ULL, "vxht_0_hightown_by5_0_0.vlm"},
	{0x86B10E2D3E838292ULL, "6116603_f.fxe"},
	{0x86B145E9676B791EULL, "vxw_bonepitmining_a2_271_0_0.vlm"},
	{0x86B1A8F2F1AAC7E4ULL, "vxca_caverns2prc_9l6_0_0.vlm"},
	{0x86B1EF0EACB2618CULL, "gen_im_arm_hlm_med_act1_01.uti"},
	{0x86B2334B86F40244ULL, "character_character_p261.psh.fxo.dx11"},
	{0x86B242C6EA4129EBULL, "vxlt_darktown_4uv_0_0.vlm"},
	{0x86B29E2FB4F7F0ACULL, "vxlt_hangedman_1i1_0_0.vlm"},
	{0x86B2C81C521C27DDULL, "mer220_hubert.bnk"},
	{0x86B322B66B38F62AULL, "273261023.ogg"},
	{0x86B35F82CAD44DEEULL, "6217105_m.fxe"},
	{0x86B3CF543DB5C9E6ULL, "wm_ico_ruins.dds"},
	{0x86B3EAD9DE741268ULL, "6111355_m.fxe"},
	{0x86B432C78562E19EULL, "vxw_bonepitmining_a2_1vi_0_0.vlm"},
	{0x86B4CFB605535F5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_90_walk.ani"},
	{0x86B4EE40397449A6ULL, "vxlt_gallowstemplar_n_3rr_0_0.vlm"},
	{0x86B508A2F439595AULL, "gen00fl_varric6105518.cl"},
	{0x86B5AA1C521E2AE5ULL, "mer220_hubert.cnv"},
	{0x86B5C9F746EDF4C6ULL, "vxlt_0_lowtown_n_emv_0_0.vlm"},
	{0x86B6F0FCC73862CAULL, "ts_beasts_drag_prog_scale.gda"},
	{0x86B6FE3B532264BAULL, "sebcr_merc_warrior_f.utc"},
	{0x86B71F9DAA2D79E5ULL, "hlac_ht_store_187_3y.dds"},
	{0x86B724AA6E680D4AULL, "hlac_lt_0_lowtown_docks_-2_4y.dds"},
	{0x86B79253003543A2ULL, "6151704_m.fxe"},
	{0x86B7BD179F8CE7BFULL, "6126400_m.fxe"},
	{0x86B87B660B3B4E4FULL, "dae000_elegant6166935.cl"},
	{0x86B884104A80CAF1ULL, "vxlt_0_lowtown_docks_a3_cuf_0_0.vlm"},
	{0x86B88FB568F19821ULL, "6083628_m.fxe"},
	{0x86B8AA5749FDF2E1ULL, "fen210_tevinter_ambusher26108642.cl"},
	{0x86B8FA0277F10C6FULL, "au_ripple.mao"},
	{0x86B948FEE07569B0ULL, "443613324.ogg"},
	{0x86B9E09A6207FBC8ULL, "w_blightlands_47_0.msh"},
	{0x86B9EFC7F1DAAC0BULL, "6100376_m.fxe"},
	{0x86B9F29A62081A59ULL, "w_blightlands_47_0.mao"},
	{0x86B9FE9A62082EF2ULL, "w_blightlands_47_0.mmh"},
	{0x86BA0014EDF1001CULL, "dae021_keep_noblewoman3.utc"},
	{0x86BA2ABCBB41B02AULL, "gen00fl_merrill6129844.cl"},
	{0x86BA63206BADA17AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumplrg.msh"},
	{0x86BA75206BADBFC0ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumplrg.mmh"},
	{0x86BAA980647EFEE0ULL, "vxht_keep_62e_0_0.vlm"},
	{0x86BAAB1A63A4B28CULL, "vxlt_warehouse_n_1cl_0_0.vlm"},
	{0x86BADCB60557E09EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_90_walk.gad"},
	{0x86BAE04DBEAADDD8ULL, "6116573_f.fxe"},
	{0x86BB3AB1CD170176ULL, "gen_im_trs_tmp_one_mac_r2.uti"},
	{0x86BB57600BEE2F7BULL, "6119705_m.fxe"},
	{0x86BBC9FC2D6D2D01ULL, "953301437.ogg"},
	{0x86BC0F537AD5003FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\choke.evt"},
	{0x86BCD0FF3D71D885ULL, "hlac_lt_gallowstemplar_f_2391_9x.dds"},
	{0x86BDEE2AB57A5994ULL, "lgt260cr_way_merc_2.utc"},
	{0x86BE713023A2C075ULL, "6114919_m.fxe"},
	{0x86BEA7BFADF8CC40ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.mmh"},
	{0x86BEED7DD0BBEB24ULL, "vxht_0_hightown_a6r_0_0.vlm"},
	{0x86BFB6A08690AD0CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha08a.msh"},
	{0x86BFBCA08690B7FEULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha08a.mmh"},
	{0x86BFD0AFC09F55EDULL, "1050626184.ogg"},
	{0x86C0196D6A3185E0ULL, "6032610_m.fxe"},
	{0x86C021A17FDDE1D3ULL, "hlac_lt_0_lowtown_a2_-1_4z.dds"},
	{0x86C073479FC0B017ULL, "vxht_brothel_n_5ad_0_0.vlm"},
	{0x86C0832C94D4751AULL, "vxlt_undercityserialkiller_1j4_0_1.vlm"},
	{0x86C0916789970E82ULL, "isa311_isabela6092315.cl"},
	{0x86C0B04E00B62312ULL, "hlac_lt_0_lowtown_n_65535_7y.dds"},
	{0x86C0C8A527C94860ULL, "vxlt_gallowstemplar_n_4yj_0_0.vlm"},
	{0x86C1B7B6055DC0BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_90_walk.evt"},
	{0x86C1B82CFE699BD2ULL, "hlac_lt_0_lowtown_1_10y.dds"},
	{0x86C1CEE6CFCD571BULL, "dae200_bran6094710.cl"},
	{0x86C1EDB31433607FULL, "6128972_m.fxe"},
	{0x86C225541E8B48C5ULL, "6115753_f.fxe"},
	{0x86C2268C091E04EEULL, "llac_ht_keep_n_4519_10.dds"},
	{0x86C2479A55948594ULL, "446348111.ogg"},
	{0x86C281CB6681CBAAULL, "6059784_f.fxe"},
	{0x86C2A4A47758FA1EULL, "vxw_bonepitmining_a2_1sj_0_0.vlm"},
	{0x86C2F0537ADAEA28ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\choke.gad"},
	{0x86C3039FE92CBF4BULL, "hlac_lt_0_lowtown_a3_n_c_65535_1y.dds"},
	{0x86C349135442F027ULL, "hlac_lt_0_lowtown_docks_-262144_1z.dds"},
	{0x86C3520210E304C2ULL, "6099181_m.fxe"},
	{0x86C3CDDB28EF8B8EULL, "6117028_f.fxe"},
	{0x86C3E0AD06E6815DULL, "vxca_caverns1_mines_gkn_0_1.vlm"},
	{0x86C41015AAA58316ULL, "6092100_m.fxe"},
	{0x86C42952209CDBC2ULL, "vxca_caverns2prc_aao_0_0.vlm"},
	{0x86C43DC27320DB71ULL, "llac_ht_mansion_n_3382_1.dds"},
	{0x86C4D496A82AAEA9ULL, "mag320cr_mage_1.utc"},
	{0x86C5E845DB024F21ULL, "mer120_hubert6017814.cl"},
	{0x86C61481715FDD08ULL, "vxht_chantry_n_2r6_0_0.vlm"},
	{0x86C6DA410663A4DEULL, "llac_w_sundermount_-131070_0.dds"},
	{0x86C6F5ED3C76E371ULL, "vxlt_0_lowtown_docks_a3_cxe_0_0.vlm"},
	{0x86C6F6D76AB08E7CULL, "vxht_0_hightown_n_c56_0_0.vlm"},
	{0x86C7375F6B3A3BB4ULL, "vxlt_undercityserialkiller_105_0_0.vlm"},
	{0x86C76387343C0518ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_meda_0n.dds"},
	{0x86C7D19A2FC68458ULL, "vxdr_primevalentrance_2ok_0_0.vlm"},
	{0x86C7D7BBEFA73B4AULL, "rdr111dg_martin6054968.cl"},
	{0x86C7F666C88F8E13ULL, "follower_banter6167866.cl"},
	{0x86C8154682C47CC7ULL, "character_character_p181.psh.fxo.dx11.assembly"},
	{0x86C83E466648F65DULL, "lt_0_lowtown_docks_a3_n_18_0_lc.mmh"},
	{0x86C8791616789B85ULL, "w_bonepitmining_a2_4_0_ll.mmh"},
	{0x86C88F3A18EA28F5ULL, "6110836_m.fxe"},
	{0x86C891F520607F15ULL, "character_character_p150.psh.fxo.dx11"},
	{0x86C8983231390DC6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v2.ani"},
	{0x86C8B7123EBD8A92ULL, "hlac_ht_0_hightown_a2_-65534_1z.dds"},
	{0x86C8D0CF7A261015ULL, "6046920_m.fxe"},
	{0x86C8DAD84C42A1D1ULL, "6176743_f.fxe"},
	{0x86C9146E204E4152ULL, "ht_keep_n_4023_lc.mmh"},
	{0x86C96C9616907EE5ULL, "37331747.ogg"},
	{0x86C9B7EDEB9A963FULL, "851182677.ogg"},
	{0x86C9C93CFDE447FBULL, "6114005_m.fxe"},
	{0x86CA09798EAF67F2ULL, "6140276_m.fxe"},
	{0x86CA0D011412BD29ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierslim.mao"},
	{0x86CA19011412D1C2ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierslim.mmh"},
	{0x86CA1B011412D578ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierslim.msh"},
	{0x86CA3D315AD3E17BULL, "w_blightlands_219_0.msh"},
	{0x86CA44A6B499B46FULL, "character_character_p149.psh.fxo.dx11"},
	{0x86CA4F315AD3FFC6ULL, "w_blightlands_219_0.mao"},
	{0x86CA5B315AD41465ULL, "w_blightlands_219_0.mmh"},
	{0x86CA68DD95E8DF66ULL, "art\\characters\\creatures\\strider\\animation\\attkfl_leg.ani"},
	{0x86CA823A24968F05ULL, "vxht_mansion2_4d_0_0.vlm"},
	{0x86CB5283553AE988ULL, "17049724.ogg"},
	{0x86CBA8BB712920E6ULL, "vxca_caverns1_edv_0_0.vlm"},
	{0x86CBC5B9C8AC5116ULL, "rock_small_normals.dds"},
	{0x86CBE46C65124242ULL, "vxlt_0_lowtown_a2_n_c_hjd_0_0.vlm"},
	{0x86CC1FF6708B843AULL, "gen_im_acc_amu_act3_09.uti"},
	{0x86CC4C360E7B9672ULL, "6107413_m.fxe"},
	{0x86CC5E2AB93E3D03ULL, "mcr371_angry_mage_m6087727.cl"},
	{0x86CC8ABFAE04D85FULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.anb"},
	{0x86CC8C994DEE5601ULL, "535619668.ogg"},
	{0x86CCB24F284E31AAULL, "hlac_ca_cave_2194_0y.dds"},
	{0x86CCE30D1598D8F5ULL, "6113819_m.fxe"},
	{0x86CE1DB38A211E18ULL, "dae150ip_loot_a.utp"},
	{0x86CE836A21C72B1EULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_aoe_p.mmh"},
	{0x86CEB464B56BA8E9ULL, "6042270_m.fxe"},
	{0x86CECB43EA8E3827ULL, "vxw_bonepitdestroyed_a3_1zz_0_0.vlm"},
	{0x86CEEA730BC1C439ULL, "one110_meeran.bnk"},
	{0x86CEEF45D2C75916ULL, "gencr_raider_archer.utc"},
	{0x86CFD998515E7131ULL, "dae300_alistair_king6134987.cl"},
	{0x86CFDB9CEE2530A1ULL, "286921120.ogg"},
	{0x86D09B0F55519C27ULL, "472512318.ogg"},
	{0x86D0E61DD16EC72FULL, "903444485.ogg"},
	{0x86D0F92C07463B92ULL, "748816091.ogg"},
	{0x86D17B9E76648810ULL, "vxca_cavern1_haunted_e7t_0_0.vlm"},
	{0x86D20C730BC43399ULL, "one110_meeran.cnv"},
	{0x86D25E1503ACEA3FULL, "6165085_m.fxe"},
	{0x86D261166C3534BBULL, "ts_raiders_arm_props.gda"},
	{0x86D281F29C9B9B96ULL, "6097200_m.fxe"},
	{0x86D2A091B113BAF1ULL, "vxw_woundedcoast_13_0_0.vlm"},
	{0x86D2BEB7F0BA9340ULL, "zz_dae_debug6050000.cl"},
	{0x86D2E808EEA2B4AEULL, "zz_dae_debug6049970.cl"},
	{0x86D334C3A96CD399ULL, "gen_im_trs_crc_lgt_boo_r2.uti"},
	{0x86D382ADA2F0725BULL, "art\\vfx\\textures\\vfx_crater_half.dds"},
	{0x86D3A87EE18CD575ULL, "6149900_m.fxe"},
	{0x86D3F87D827A6353ULL, "695359552.ogg"},
	{0x86D4A76B2DE75320ULL, "248434005.ogg"},
	{0x86D4D8327AF60B8CULL, "vxw_bonepit_z2_0_0.vlm"},
	{0x86D4DB93303BD2E6ULL, "vxca_cavern1_haunted_jf6_0_0.vlm"},
	{0x86D4DE44657C5F9FULL, "hlac_lt_0_lowtown_a2_n_-2_4x.dds"},
	{0x86D5158D487048C2ULL, "one100dg_kelder6173944.cl"},
	{0x86D52D0FE73E26A0ULL, "959276974.ogg"},
	{0x86D5EC114E6EDCA8ULL, "116101721.ogg"},
	{0x86D606C3C8E4933CULL, "vxdr_primevaltomb_3oz_0_0.vlm"},
	{0x86D6803231452237ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v2.evt"},
	{0x86D73DFC4CDBAC42ULL, "hlac_w_bonepit_393218_0z.dds"},
	{0x86D78327F0B46084ULL, "724609354.ogg"},
	{0x86D78C8D7F9145B0ULL, "vxw_sundermount_2yx_0_0.vlm"},
	{0x86D78D87D09E620EULL, "6038593_m.fxe"},
	{0x86D7E55130ED5C3AULL, "vxca_caverns1_ei9_0_0.vlm"},
	{0x86D8127AFC4207BEULL, "gen00fl_varric6079150.cl"},
	{0x86D83B19EFDA607BULL, "6113105_m.fxe"},
	{0x86D8454D574BC53BULL, "6015997_m.fxe"},
	{0x86D850DD95F4F3D7ULL, "art\\characters\\creatures\\strider\\animation\\attkfl_leg.evt"},
	{0x86D90901B876AB95ULL, "6160736_m.fxe"},
	{0x86D9DB6C6D0DCCEEULL, "vxlt_0_lowtown_a3_n_c_htw_0_0.vlm"},
	{0x86D9DD03513B427CULL, "6075528_f.fxe"},
	{0x86DA1F7DA0A5D26CULL, "12695533.ogg"},
	{0x86DACE00677350E5ULL, "vxlt_0_lowtown_a2_n_fe6_0_0.vlm"},
	{0x86DB4FCB94FEF0FFULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25.mmh"},
	{0x86DB8D7141BB5A10ULL, "vxlt_0_lowtown_docks_b65_0_1.vlm"},
	{0x86DB8EFC5F4B29C8ULL, "127980685.ogg"},
	{0x86DBFAB1D3F436C4ULL, "6068023_f.fxe"},
	{0x86DC0888B1D6F250ULL, "6094944_m.fxe"},
	{0x86DC48C2BAA71EE1ULL, "842128275.ogg"},
	{0x86DC6226EE8B6BB9ULL, "vxw_bonepitmining_a2_1p8_0_0.vlm"},
	{0x86DC626A21D330DDULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_aoe_p.anb"},
	{0x86DCA23893754C49ULL, "gen00fl_isabela6135620.cl"},
	{0x86DCCCAC0590E95AULL, "vxdr_0_deeproads_bd2_0_0.vlm"},
	{0x86DCCECFEDCD735FULL, "fen210_orana6185201.cl"},
	{0x86DD006E81979C0EULL, "377709856.ogg"},
	{0x86DD09E14326BED9ULL, "gen00fl_anders6119509.cl"},
	{0x86DD0E69A750E76FULL, "6028162_m.fxe"},
	{0x86DD8132314B42C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v2.gad"},
	{0x86DDA8955E0CEA3CULL, "6153195_m.fxe"},
	{0x86DDAC5F0BB57960ULL, "vxlt_gallowstemplar_n_4sh_0_0.vlm"},
	{0x86DDC8994BD68F1FULL, "llac_w_encounter3_-65532_0.dds"},
	{0x86DED9B9D6A0A70EULL, "1046852598.ogg"},
	{0x86DEF784AB10865DULL, "gen00fl_anders6189373.cl"},
	{0x86DF453697E829B8ULL, "79604372.ogg"},
	{0x86DF51DD95FB1460ULL, "art\\characters\\creatures\\strider\\animation\\attkfl_leg.gad"},
	{0x86DFBEE659B5B2BEULL, "6180492_m.fxe"},
	{0x86E02806E14BB8EDULL, "592912171.ogg"},
	{0x86E053D01B32B9BDULL, "follower_banter6172862.cl"},
	{0x86E099715F848041ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetal_n.dds"},
	{0x86E0A3B67A0EE478ULL, "41399307.ogg"},
	{0x86E0B2517A617CA0ULL, "dae000_mother6099228.cl"},
	{0x86E0B503C5B8A7CDULL, "art\\levels\\ztools\\default\\generic\\1oi_basic.phy"},
	{0x86E170A1647E1214ULL, "6070502_m.fxe"},
	{0x86E1BF14DDC9B9C7ULL, "vxht_0_hightown_n_b9h_0_0.vlm"},
	{0x86E1D400092FED14ULL, "6191885_m.fxe"},
	{0x86E27AAB7B65C102ULL, "6104607_m.fxe"},
	{0x86E29A11EECFFF7EULL, "926364251.ogg"},
	{0x86E2A320D2459A87ULL, "hlac_w_bonepit_458752_0z.dds"},
	{0x86E2D9883361BE52ULL, "6140179_m.fxe"},
	{0x86E33AF8FADDED53ULL, "vxlt_blackemporium_o1_0_0.vlm"},
	{0x86E388C9B8C953C9ULL, "zz_isabella_kife_v1.cut"},
	{0x86E42602C52EB26AULL, "vxlt_darktown_6c7_0_0.vlm"},
	{0x86E42AAA335F1048ULL, "6155571_m.fxe"},
	{0x86E4AC3CB5443CA1ULL, "608039205.ogg"},
	{0x86E4C3A27965C895ULL, "6150731_m.fxe"},
	{0x86E4C66A7D3E04D8ULL, "151645433.ogg"},
	{0x86E556C7441E0698ULL, "llac_w_blightlands_131068_0.dds"},
	{0x86E58925295FEA5DULL, "6080984_m.fxe"},
	{0x86E5B892DBF33C3BULL, "fen110_fenris6096191.cl"},
	{0x86E5FA7372AB1523ULL, "1039400297.ogg"},
	{0x86E626F4B12D4422ULL, "ht_0_hightown_n_12_0_ll.mmh"},
	{0x86E6624B4DD3B9AEULL, "6093811_m.fxe"},
	{0x86E6E4A2D131C327ULL, "vxlt_0_lowtown_a3_n_c_es2_0_0.vlm"},
	{0x86E6F58555B5F115ULL, "battlemenu.gfx"},
	{0x86E76E964A802116ULL, "853465032.ogg"},
	{0x86E772B9DDE77BB9ULL, "151076384.ogg"},
	{0x86E7BE47F6430D6CULL, "6043517_m.fxe"},
	{0x86E80E0DDBB3F1FFULL, "93785137.ogg"},
	{0x86E85CAC2ADE7CAFULL, "vxlt_0_lowtown_a2_n_c_gcm_0_0.vlm"},
	{0x86E892EF70212586ULL, "126037849.ogg"},
	{0x86E8FA4257C4E6E9ULL, "vxca_caverns2prc_acf_0_0.vlm"},
	{0x86E96EB8539F531EULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_reaction.ani"},
	{0x86EA0417D44029DCULL, "vxw_sundermount_2cf_0_0.vlm"},
	{0x86EA452D4881A3EDULL, "hlac_lt_0_lowtown_docks_a3_196609_0y.dds"},
	{0x86EA71776F5BC044ULL, "qun221_petrice6083724.cl"},
	{0x86EACFFD263AF88DULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxabunch.mmh"},
	{0x86EAD3FD263AFFDEULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxabunch.mao"},
	{0x86EAE1FD263B1713ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxabunch.msh"},
	{0x86EAF4327DF187B3ULL, "w_blightlands_89_0.phy"},
	{0x86EB4831FA72C4EFULL, "6054068_m.fxe"},
	{0x86EB58D02CF6F22FULL, "follower_banter6154110.cl"},
	{0x86EB7D4B5476BB18ULL, "1024116616.ogg"},
	{0x86EC4314EBD23F84ULL, "vrc221im_fake_bianca.uti"},
	{0x86EC70BFAF0F6C6EULL, "ser221_gascard6134458.cl"},
	{0x86ED53476C9C5E47ULL, "vxdr_primevalentrance_37i_0_0.vlm"},
	{0x86ED85F60AC6CC5AULL, "hlac_w_sundermount_2_0y.dds"},
	{0x86EE06B94A0D7AC0ULL, "vxht_mansion2b_2ju_0_0.vlm"},
	{0x86EE3663F847FBE6ULL, "art\\vfx\\combat\\vfx_glb_ele_e_minor_c.anb"},
	{0x86EEBFDFC0F3D33CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\enter.ani"},
	{0x86EF161E6DFF5AB1ULL, "hlac_lt_gallowstemplar_f_2383_8x.dds"},
	{0x86EF2B96205E9FE4ULL, "vxht_keep_5xl_0_0.vlm"},
	{0x86EF49B842CD5B92ULL, "353539378.ogg"},
	{0x86EF5DC75FB2A7D6ULL, "6117456_m.fxe"},
	{0x86EFE27E56742A94ULL, "6075602_m.fxe"},
	{0x86F01AB4DE5A890FULL, "hla_ht_0_hightown_-65537_0x.dds"},
	{0x86F070C84D98ABB4ULL, "6136671_m.fxe"},
	{0x86F0C41EB065C01BULL, "lt_0_lowtown_a2_n_20_0.anb"},
	{0x86F14EBDF7726E06ULL, "vxht_chantry_n_2ve_0_1.vlm"},
	{0x86F1627005ECCEE7ULL, "zz_followers_banter6159555.cl"},
	{0x86F16A1847422688ULL, "vxht_chantry_n_2c5_0_0.vlm"},
	{0x86F21C72740FEE9EULL, "dae000_bartrand6015463.cl"},
	{0x86F21DDAEA2CBCEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillarwindow.mao"},
	{0x86F229DAEA2CD08CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillarwindow.mmh"},
	{0x86F22FDAEA2CDB7EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillarwindow.msh"},
	{0x86F25B4FCECA42BEULL, "vxlt_warehouse_op_0_0.vlm"},
	{0x86F295C26037AECCULL, "6165680_m.fxe"},
	{0x86F2DEB1F215ADFEULL, "vxht_0_hightown_bz3_0_0.vlm"},
	{0x86F349F90458A0FFULL, "6020786_m.fxe"},
	{0x86F36F979CFE5EB9ULL, "265025247.ogg"},
	{0x86F38381C032E240ULL, "85359127.ogg"},
	{0x86F3D51F2F51A0F3ULL, "279869044.ogg"},
	{0x86F3E0F9C7ACF964ULL, "lowtown_fem_alienage1_sta6162551.cl"},
	{0x86F4072BCE0BE38DULL, "hlac_dr_dungeon_3207_2y.dds"},
	{0x86F51A254FF30F50ULL, "825533961.ogg"},
	{0x86F531756EC57297ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_maj_c.impa.gad"},
	{0x86F5619A5C760F9AULL, "vxw_bonepit_29q_0_0.vlm"},
	{0x86F59F1FEB76DC94ULL, "vxlt_0_lowtown_docks_n_c6v_0_0.vlm"},
	{0x86F5B5DFBC5E82E3ULL, "6094802_m.fxe"},
	{0x86F5CCDFC0FA087EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\enter.gad"},
	{0x86F5E77C32630F47ULL, "vxlt_0_lowtown_a2_ehc_0_0.vlm"},
	{0x86F5F01428D6AA7FULL, "6162682_m.fxe"},
	{0x86F75F1C8832EE40ULL, "one300_zevran6076937.cl"},
	{0x86F786B853ABB9EFULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_reaction.evt"},
	{0x86F814425286DC02ULL, "282925321.ogg"},
	{0x86F8647E7E7153DEULL, "vxlt_gallowstemplar_f_3rr_0_0.vlm"},
	{0x86F875F4C636425CULL, "vxw_sundermount_2dg_0_0.vlm"},
	{0x86F902D85365B65FULL, "seb221_flora_harimann6197724.cl"},
	{0x86F9B51EF5BD59FCULL, "vxw_woundedcoast_ns_0_0.vlm"},
	{0x86FA20DAC07081A7ULL, "vxw_bonepitdestroyed_a3_1ew_0_0.vlm"},
	{0x86FA2BEE8CA612D9ULL, "6095674_m.fxe"},
	{0x86FA76B074223D8CULL, "follower_banter6161278.cl"},
	{0x86FA87E3773035E3ULL, "vxht_keep_n_5gq_0_0.vlm"},
	{0x86FAA2284E34F619ULL, "zz_vrc_debug6093953.cl"},
	{0x86FB3963F8528A91ULL, "art\\vfx\\combat\\vfx_glb_ele_e_minor_c.mmh"},
	{0x86FBC5245E9276C7ULL, "vxdr_primevalentrance_30h_0_0.vlm"},
	{0x86FBC8FD2649604AULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxabunch.phy"},
	{0x86FBCB60D4C8FDACULL, "vxca_caverns1_j29_0_0.vlm"},
	{0x86FC46756ECBB524ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_maj_c.impa.evt"},
	{0x86FC8AD41490F94AULL, "gen00fl_varric6033596.cl"},
	{0x86FCA7DFC0FFE79DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\enter.evt"},
	{0x86FD049152E36193ULL, "6096858_f.fxe"},
	{0x86FD1D1B12DB17D6ULL, "lt_0_lowtown_docks_118_0.msh"},
	{0x86FD1F829A0F0060ULL, "430108136.ogg"},
	{0x86FD271B12DB28D4ULL, "lt_0_lowtown_docks_118_0.mmh"},
	{0x86FD2B1B12DB2FE7ULL, "lt_0_lowtown_docks_118_0.mao"},
	{0x86FD76AE97BA37EFULL, "mcr370_assignments6081689.cl"},
	{0x86FD87B853B02698ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_reaction.gad"},
	{0x86FDC266BCD71CE2ULL, "801443091.ogg"},
	{0x86FDDB03E43D83B1ULL, "vxca_caverns1_gk6_0_0.vlm"},
	{0x86FF01E0B849DB67ULL, "ht_0_hightown_29_0.rml"},
	{0x86FF37958C1108B6ULL, "174927572.ogg"},
	{0x8700748D077616E3ULL, "vxdr_slavers_1aq_0_0.vlm"},
	{0x870093A086C80899ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha08a.phy"},
	{0x8702501F8D44717EULL, "6169017_m.fxe"},
	{0x8702C5537BE466F4ULL, "dre000_varric6207956.cl"},
	{0x8702F4009DA5F024ULL, "w_encounter1_41_0.msh"},
	{0x870306009DA60E8DULL, "w_encounter1_41_0.mao"},
	{0x87030A009DA615C6ULL, "w_encounter1_41_0.mmh"},
	{0x87032224F1BD1267ULL, "699586411.ogg"},
	{0x8703986B5B1E4E03ULL, "6109241_m.fxe"},
	{0x870410FCDD6CF514ULL, "vxlt_0_lowtown_docks_n_c1w_0_0.vlm"},
	{0x8704A71C86B7C57BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_to_face.ani"},
	{0x87050E392ECACE15ULL, "6126338_m.fxe"},
	{0x870629045FAE893FULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfr.gad"},
	{0x8706419946DFF154ULL, "llac_lt_hangedman_-1_23.dds"},
	{0x87064D4EC50B3287ULL, "6125842_f.fxe"},
	{0x870667EA18CDEBE9ULL, "6205970_m.fxe"},
	{0x870682F25EFA7FDBULL, "w_sundermount_244_0.msh"},
	{0x870694F25EFA9E26ULL, "w_sundermount_244_0.mao"},
	{0x8706A0F25EFAB2C5ULL, "w_sundermount_244_0.mmh"},
	{0x8707D664B1A3B951ULL, "275378188.ogg"},
	{0x8708150F5412E3A2ULL, "6074395_m.fxe"},
	{0x870826FBE7B3727CULL, "vxw_woundedcoast_kp_0_0.vlm"},
	{0x87088B23387F270BULL, "follower_banter6171872.cl"},
	{0x870892B7B2669A27ULL, "vxw_bonepitdestroyed_a3_1nv_0_0.vlm"},
	{0x8708A3CB04C33364ULL, "6139784_m.fxe"},
	{0x8708AB42ED97351FULL, "clf221_jalen6070379.cl"},
	{0x8708AFDA972524C0ULL, "w_encounter2_camp_36_0_ll.mmh"},
	{0x8709067280EEEA2DULL, "6174016_m.fxe"},
	{0x87095E756ED6681DULL, "art\\vfx\\combat\\vfx_glb_nat_imp_maj_c.impa.ani"},
	{0x8709690045464F57ULL, "one280_emrys6187705.cl"},
	{0x8709C7BDE9FD2C6FULL, "161591431.ogg"},
	{0x8709CE3682064703ULL, "6110617_f.fxe"},
	{0x8709E1423E391FFFULL, "837847927.ogg"},
	{0x8709F7B4746C0482ULL, "gen_im_npc_arm_hum_cht_gwda.uti"},
	{0x870A4B7EAE93B3B3ULL, "6039949_f.fxe"},
	{0x870ACE9ED9B34EB5ULL, "vxht_0_hightown_a2_b1a_0_0.vlm"},
	{0x870B1BA42449FC25ULL, "hlac_w_woundedcoast_65541_0y.dds"},
	{0x870B2F659195D403ULL, "vxht_chantry_n_360_0_0.vlm"},
	{0x870B6E1F95C248ABULL, "449610373.ogg"},
	{0x870BCC836063EBDCULL, "w_bonepitdestroyed_a3_12_0.rml"},
	{0x870C6806094FEF2FULL, "6189425_m.fxe"},
	{0x870C6E277BBA2B01ULL, "vxw_sundermount_2c7_0_0.vlm"},
	{0x870C8DE4746212B8ULL, "vxht_0_hightown_bxk_0_0.vlm"},
	{0x870C9C1E2DC67F89ULL, "vxlt_0_lowtown_a3_n_c_ep5_0_0.vlm"},
	{0x870CB94C51BDF094ULL, "323035352.ogg"},
	{0x870CDA6C47FE1F3FULL, "vxlt_0_lowtown_a2_cmr_0_0.vlm"},
	{0x870D1E045FB4952CULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfr.evt"},
	{0x870D2A5E029DD332ULL, "prty_aveline_armor_1.dds"},
	{0x870DB05E56C7C637ULL, "hlac_dr_slavers_3214_2y.dds"},
	{0x870DB94321264A5FULL, "6114823_m.fxe"},
	{0x870DDCA0965F3497ULL, "102707668.ogg"},
	{0x870E01009DB00D91ULL, "w_encounter1_41_0.phy"},
	{0x870E32B9DB727981ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_masf_0d.dds"},
	{0x870E526C5F3A180CULL, "6047467_m.fxe"},
	{0x870E77498A3D5ED4ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kita_0n.dds"},
	{0x870EA80A9AD63F1DULL, "vxlt_0_lowtown_docks_n_bga_0_0.vlm"},
	{0x870F7336CCCB52E6ULL, "vxw_bonepit_10y_0_0.vlm"},
	{0x870FB2836067A831ULL, "w_bonepitdestroyed_a3_12_0.msh"},
	{0x870FC0836067C04CULL, "w_bonepitdestroyed_a3_12_0.mao"},
	{0x870FCC836067D4EFULL, "w_bonepitdestroyed_a3_12_0.mmh"},
	{0x871057ACEE56B41BULL, "6042060_f.fxe"},
	{0x87106B86D41C96AFULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.cess.gad"},
	{0x87107C43C6548CFEULL, "gen00fl_carver6104898.cl"},
	{0x871196B996B9CDA7ULL, "gen00fl_fenris6106642.cl"},
	{0x8711EF4E65F4CA05ULL, "mag360_samson6042096.cl"},
	{0x87120BCB5B812CA3ULL, "vxht_0_hightown_9u4_0_0.vlm"},
	{0x87121EBB3FB4D3B9ULL, "vxht_playershouse_n_1ux_0_0.vlm"},
	{0x8712AF1C86C41072ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_to_face.evt"},
	{0x8713183D6C5C3454ULL, "qun200_party_comment6095391.cl"},
	{0x8714080371F4F427ULL, "6173569_m.fxe"},
	{0x871498BEA2762ABFULL, "gen_im_wep_mag_sta_act0_03.uti"},
	{0x8714E962096C6995ULL, "6172524_m.fxe"},
	{0x871559AEAA22735CULL, "vxw_sundermount_2re_0_0.vlm"},
	{0x87159E32875EAA93ULL, "6176531_m.fxe"},
	{0x8715D51A063C2E42ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medd_0d.dds"},
	{0x871610F8EDD2AD01ULL, "mag100_karras6034940.cl"},
	{0x87166E16969FF0D8ULL, "hlac_lt_0_lowtown_docks_n_-65538_4y.dds"},
	{0x87168A2169505F2AULL, "vxlt_0_lowtown_n_ewx_0_0.vlm"},
	{0x87169DC0526DB165ULL, "audioevent.ncs"},
	{0x8716BA6969F2E333ULL, "static1_vbl.vsh.fxo"},
	{0x87178086D422D95CULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.cess.evt"},
	{0x87178FCD3B370757ULL, "hlac_lt_gallowstemplar_f_2382_3x.dds"},
	{0x8717EBBD7561C755ULL, "vxlt_gallowstemplar_f_56g_0_0.vlm"},
	{0x8717FABC0843C3F0ULL, "vxht_mansion2b_2a4_0_0.vlm"},
	{0x8718219927147598ULL, "ser270_emeric6149339.cl"},
	{0x8718CE1C86C8B005ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_hand_to_face.gad"},
	{0x8719500603E63094ULL, "hlac_lt_0_lowtown_docks_-65540_1z.dds"},
	{0x8719542E8104AB6CULL, "6039725_m.fxe"},
	{0x8719A142838BEC83ULL, "vxht_chantry_n_333_0_0.vlm"},
	{0x8719C56E7C134D06ULL, "679303534.ogg"},
	{0x8719F50F7077A514ULL, "ht_0_hightown_a3_29_0.rml"},
	{0x8719FC5A4A3A73E8ULL, "6110517_f.fxe"},
	{0x871A7FD461ABC213ULL, "hlac_lt_0_lowtown_docks_262142_0x.dds"},
	{0x871ABA8A2CA7A35BULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3.phy"},
	{0x871ACF3BAF0CF09DULL, "lt_0_lowtown_docks_a3_205_0.mao"},
	{0x871AD33BAF0CF7D6ULL, "lt_0_lowtown_docks_a3_205_0.mmh"},
	{0x871ADD3BAF0D08D4ULL, "lt_0_lowtown_docks_a3_205_0.msh"},
	{0x871AE0046DB04381ULL, "vxw_sundermount_2d6_0_0.vlm"},
	{0x871B06045FC0A985ULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfr.ani"},
	{0x871B93BA5FE92625ULL, "900874371.ogg"},
	{0x871BCA59DC6F56E1ULL, "210867883.ogg"},
	{0x871BEBDC68FC3C50ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_maso_0n.dds"},
	{0x871BF109263ECC35ULL, "dae000_worthy6166583.cl"},
	{0x871BFE9A625AF899ULL, "w_blightlands_47_0.rml"},
	{0x871C2B20131C62DFULL, "6113923_m.fxe"},
	{0x871C4AF883BDF80EULL, "1048185010.ogg"},
	{0x871CB6BA44524021ULL, "759025715.ogg"},
	{0x871CC01DCCCD29B8ULL, "48035987.ogg"},
	{0x871D02145E9CB239ULL, "526709208.ogg"},
	{0x871D9C8D929BA74BULL, "knights01.mao"},
	{0x871DB3E7327614A9ULL, "6070382_f.fxe"},
	{0x871DC6AC4DC633F7ULL, "w_encounter3_162_0.phy"},
	{0x871DD27C0D4D71E3ULL, "hlac_lt_0_lowtown_docks_-65538_5z.dds"},
	{0x871DE513BEC16B66ULL, "vxw_bonepit_15z_0_0.vlm"},
	{0x871E2992C229BB75ULL, "zz_followers_banter6158235.cl"},
	{0x871E62B2E3F7D1B3ULL, "605725000.ogg"},
	{0x871EED158A19B210ULL, "vxca_caverns2prc_a5a_0_0.vlm"},
	{0x871F0416BA1B68E7ULL, "6016483_f.fxe"},
	{0x871F074A539D62E7ULL, "plt_gen00pt_random.plo"},
	{0x871F0F045AD93E26ULL, "vxlt_0_lowtown_docks_n_b4f_0_1.vlm"},
	{0x871F1A5A099E45C5ULL, "lt_0_lowtown_docks_a3_n_10_0_lc.mmh"},
	{0x871FBF807913E891ULL, "169273490.ogg"},
	{0x8720140A162A71BBULL, "801936206.ogg"},
	{0x8720BC87F9448CA9ULL, "6108526_f.fxe"},
	{0x8720E51D13F118F6ULL, "ser270_emeric6113514.cl"},
	{0x8720F37552FBA042ULL, "6054692_m.fxe"},
	{0x87211147BE67C406ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward_mirror.evt"},
	{0x87213CD50269FACDULL, "art\\characters\\weapons\\staff\\textures\\w_stf_005a_0t.dds"},
	{0x87220FB89A047D23ULL, "vxht_mansion_n_33s_0_0.vlm"},
	{0x8722230EC4E2CABAULL, "art\\vfx\\cinematic\\vfx_flash_p.anb"},
	{0x8722683C56129247ULL, "vxlt_0_lowtown_er5_0_0.vlm"},
	{0x8722AA969F62D288ULL, "vxca_cave_18d_0_0.vlm"},
	{0x8722B710145784DFULL, "hlac_lt_gallowstemplar_2383_12x.dds"},
	{0x872362622AD5AB1FULL, "6174959_m.fxe"},
	{0x87237058B4836604ULL, "var_placeable_react.gda"},
	{0x8723AD3AC7009225ULL, "637116351.ogg"},
	{0x87241F6C87F58BA8ULL, "hlac_lt_0_lowtown_docks_a3_-196610_1x.dds"},
	{0x8724AC3CA41E2CA1ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_hed_hlkb_0d.dds"},
	{0x8724B65A90DD7AEDULL, "6176018_m.fxe"},
	{0x87251906CFBB50FBULL, "vxlt_gallowscourtyard_744_0_0.vlm"},
	{0x872587A089C89558ULL, "6129613_m.fxe"},
	{0x87258886D42F2475ULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.cess.ani"},
	{0x872603E906EF78EFULL, "plt_rdr100pt_keepers_note.plo"},
	{0x8726415945CB93C7ULL, "dae000_bartrand6027490.cl"},
	{0x872649FE8FCB17EFULL, "mag131_keran6091388.cl"},
	{0x87268DBEB4E01F74ULL, "vxht_mansion2_1rn_0_0.vlm"},
	{0x8726C05F3F3B3C40ULL, "da2terrain_terrain_p2.psh.fxo.dx11"},
	{0x8726C5AF70F9EE6CULL, "hf_lgt_spy.mor"},
	{0x8727C8EF968F4608ULL, "vxca_caverns2prc_62o_0_0.vlm"},
	{0x87282847BE6E0A51ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward_mirror.gad"},
	{0x87284835C396B82AULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.dura.gad"},
	{0x872951E15FA65C01ULL, "vxw_sundermount_2i5_0_0.vlm"},
	{0x872A6E606C02298AULL, "vxlt_ruinedhouse_n_4f_0_0.vlm"},
	{0x872BC8D76D15A587ULL, "368013903.ogg"},
	{0x872BE7A95D511877ULL, "253355558.ogg"},
	{0x872C0A6274E129B5ULL, "579012069.ogg"},
	{0x872C4185D7C9E72EULL, "one280_iwan6096906.cl"},
	{0x872C83BF41940274ULL, "llac_lt_0_lowtown_a3_n_c_-262144_3.dds"},
	{0x872CD6A354CE3F85ULL, "6086159_m.fxe"},
	{0x872D4745202F8D80ULL, "349547970.ogg"},
	{0x872D5EF27C0FCA90ULL, "vxca_caverns2prc_a0n_0_0.vlm"},
	{0x872DE85CBE7A8474ULL, "seb201_lady_harimann6177318.cl"},
	{0x872E05A8FBFF83A5ULL, "6094484_m.fxe"},
	{0x872E2947BE7277FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_forward_mirror.ani"},
	{0x872F2B35C39CA5C1ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.dura.evt"},
	{0x872F6DE435EE5612ULL, "vxlt_0_lowtown_docks_n_bfu_0_0.vlm"},
	{0x872F8AC9285D9FF3ULL, "887399652.ogg"},
	{0x87307324F2005E3DULL, "hlac_w_woundedcoast_0_0z.dds"},
	{0x8730A67F8EF68F46ULL, "vxw_bonepitmining_a2_h8_0_0.vlm"},
	{0x873104056D859F6FULL, "ss_meredith.bnk"},
	{0x87316090114F5A5AULL, "380951206.ogg"},
	{0x87316461D1EF759DULL, "llac_dr_primevalentrance_1726_10.dds"},
	{0x87318C13640DA255ULL, "t1_mub_pu2.tnt"},
	{0x8731B6E00C12A75EULL, "vxca_caverns2_96k_0_0.vlm"},
	{0x8731B7AF943E9D8CULL, "hlac_lt_0_lowtown_131072_0y.dds"},
	{0x8732216EF7F80A61ULL, "vxdr_primevaltomb_7tj_0_0.vlm"},
	{0x87327DCACA8C4A0DULL, "6210363_m.fxe"},
	{0x87328461D9A8D718ULL, "1030582108.ogg"},
	{0x8732A25714EFFFEEULL, "tut00pt_tutorials.ncs"},
	{0x8732ACC18A4BDEABULL, "vxht_mansion_n_4d1_0_0.vlm"},
	{0x8732FF9E14F3E8D9ULL, "sebcr_merch_warrior.utc"},
	{0x87335B2289A0506DULL, "vxw_encounter2_kd_0_0.vlm"},
	{0x87336DDB4D3C902AULL, "vxlt_0_lowtown_n_emz_0_0.vlm"},
	{0x87337C92BD95BBFCULL, "vxw_woundedcoast_zw_0_0.vlm"},
	{0x87338AE3C1B1697BULL, "vxlt_gallowscourtyard_797_0_0.vlm"},
	{0x8733AD716B7B914DULL, "llac_lt_undercityencounter_-1_41.dds"},
	{0x8733C388795F3AA6ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_waves.dds"},
	{0x8733DB5803DCF5EFULL, "w_bonepit_86_0.mao"},
	{0x8733E75803DD098CULL, "w_bonepit_86_0.mmh"},
	{0x8733ED5803DD147EULL, "w_bonepit_86_0.msh"},
	{0x8734C2DB5C46CAFCULL, "6208492_f.fxe"},
	{0x8734C53C44559B5FULL, "qun120_arishok6051378.cl"},
	{0x8734FF9BA6D637F4ULL, "vxht_mansion2_1uo_0_0.vlm"},
	{0x8735337F8CBD6010ULL, "6096780_m.fxe"},
	{0x8735379265BBAD40ULL, "6085573_m.fxe"},
	{0x873583705528E0A7ULL, "vxlt_0_lowtown_docks_b5f_0_0.vlm"},
	{0x8735C994DD829741ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_lp.ani"},
	{0x8735E6D7B98DBC9CULL, "qcr000_arishok6084983.cl"},
	{0x8735E90D1AE78686ULL, "1072698388.ogg"},
	{0x87360CD371ACEF81ULL, "hlac_lt_0_lowtown_a3_n_c_65538_3z.dds"},
	{0x873626056FEDE01CULL, "hlac_w_bonepitmining_a2_458752_0y.dds"},
	{0x873683D1F4E68DDCULL, "vxht_chantry_1ys_0_0.vlm"},
	{0x87369856960FE0CFULL, "6172881_m.fxe"},
	{0x8736F5D680105C19ULL, "w_sundermount_123_0.msh"},
	{0x8736FFD680106D07ULL, "w_sundermount_123_0.mmh"},
	{0x873703D680107454ULL, "w_sundermount_123_0.mao"},
	{0x8737325FD0D76D6EULL, "868104933.ogg"},
	{0x8737B839B04D9FC2ULL, "vxca_caverns2prc_a9w_0_0.vlm"},
	{0x8737E3AE2975C3D8ULL, "vxht_keep_n_5cr_0_0.vlm"},
	{0x87383D931932FABAULL, "vxlt_0_lowtown_docks_a3_cow_0_0.vlm"},
	{0x8738C129E09A3CC7ULL, "vxca_cave_1a8_0_0.vlm"},
	{0x8739D5AA61C46FDFULL, "6093256_m.fxe"},
	{0x873A9055ED40E19BULL, "llac_ht_chantry_2534_8.dds"},
	{0x873B1876A12622FAULL, "443224430.ogg"},
	{0x873B5AD74021E043ULL, "vxht_mansion_n_6je_0_0.vlm"},
	{0x873B873BDFD82C98ULL, "vxlt_house_fz_0_0.vlm"},
	{0x873BAA24B94CB648ULL, "vxdr_primevaltomb_7to_0_0.vlm"},
	{0x873C2335C3A72210ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.dura.ani"},
	{0x873C4FF0E37E1F0AULL, "6151245_m.fxe"},
	{0x873C6922D58344A1ULL, "vxlt_undercityserialkiller_12j_0_0.vlm"},
	{0x873CABDEE54B8F0CULL, "vxlt_warehouse_n_10e_0_0.vlm"},
	{0x873CD494DD88C8E3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_lp.gad"},
	{0x873DDFC127E46E92ULL, "vxlt_0_lowtown_docks_n_bcv_0_0.vlm"},
	{0x873E58CB95537D44ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_25x_25.phy"},
	{0x873E62ADCA9F88B1ULL, "vxlt_0_lowtown_docks_a3_d02_0_0.vlm"},
	{0x873EA94AB01FF4C6ULL, "hlac_lt_0_lowtown_65538_2x.dds"},
	{0x873F04E0B87FC47BULL, "ht_0_hightown_29_0.anb"},
	{0x873F4BF639FEC347ULL, "vxlt_0_lowtown_eh7_0_0.vlm"},
	{0x873F73077F8D77EFULL, "lt_0_lowtown_a2_n_c_34_0_lc.mmh"},
	{0x874012A1EEBD61E6ULL, "214773141.ogg"},
	{0x87403818F777C8A7ULL, "vxdr_primevaltomb_8xw_0_0.vlm"},
	{0x8740934BE9EE22E1ULL, "vxdr_primevaltomb_7qi_0_0.vlm"},
	{0x8740D9888FCDA669ULL, "vxlt_0_lowtown_a3_n_c_hl4_0_0.vlm"},
	{0x8741343BB0CB8A94ULL, "504433308.ogg"},
	{0x8741A44B3449523EULL, "gut_glow.mao"},
	{0x8741D01506679ECEULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_0z.dds"},
	{0x87431BCFA8645B0CULL, "art\\characters\\heraldry\\hld_winters_mercs.dds"},
	{0x8743717898CC5074ULL, "vxht_mansion2_1xl_0_0.vlm"},
	{0x8743B194DD8EABC8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_lp.evt"},
	{0x8744CFEF492E031EULL, "50887451.ogg"},
	{0x87450A338805F94FULL, "6175981_m.fxe"},
	{0x874607F03DD27B06ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\flemeth_idle_loop.evt"},
	{0x8746160F709D1814ULL, "ht_0_hightown_a3_29_0.anb"},
	{0x8746558B1B6BDC58ULL, "vxht_keep_n_5ds_0_0.vlm"},
	{0x8746692E804F08D1ULL, "86941030.ogg"},
	{0x8746B38A2CCCD240ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3.mmh"},
	{0x87470930061266A1ULL, "382384746.ogg"},
	{0x8747625151EB53DEULL, "llac_dr_slavers_3215_1.dds"},
	{0x87476583E1637595ULL, "art\\characters\\creatures\\gallowsslavestatue\\textures\\cn_bdy_css_0d.dds"},
	{0x87476707C419AA2AULL, "1036525932.ogg"},
	{0x8748BB4C029ABF32ULL, "zz_qcr_debug6058842.cl"},
	{0x8749CCB43217F8C3ULL, "vxht_mansion_n_6md_0_0.vlm"},
	{0x874A1C01AB42CEC8ULL, "vxdr_primevaltomb_7ql_0_0.vlm"},
	{0x874A457EF6BD6BADULL, "6060308_m.fxe"},
	{0x874AA0A21F144AF7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamc.msh"},
	{0x874AB2A21F14695AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamc.mao"},
	{0x874AB6A21F147069ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeamc.mmh"},
	{0x874AB701BA87224DULL, "738833935.ogg"},
	{0x874ACFF1196CED04ULL, "gen00fl_anders6212032.cl"},
	{0x874ADAFFC7795D21ULL, "vxlt_undercityserialkiller_17i_0_0.vlm"},
	{0x874B260EC505D56DULL, "art\\vfx\\cinematic\\vfx_flash_p.mmh"},
	{0x874B2B6BD950A44DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc05.mmh"},
	{0x874B3D6BD950C2D3ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc05.msh"},
	{0x874BEF5C3861822AULL, "6182025_m.fxe"},
	{0x874C07CB445BC5DEULL, "dae000_female_tranquil6165147.cl"},
	{0x874C28076DC36BD5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_medj_0n.dds"},
	{0x874C34CEA3CF9725ULL, "gen_im_arm_glv_hvy_champion.uti"},
	{0x874C3DA194D2382AULL, "554079978.ogg"},
	{0x874C459336B6BBF6ULL, "930231066.ogg"},
	{0x874C62FF5B251BC7ULL, "690761042.ogg"},
	{0x874CBB30E42E66C6ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.cess.ani"},
	{0x874CD2B3B2B28EE4ULL, "vxht_brothel_n_5fb_0_0.vlm"},
	{0x874CD48ABC95A131ULL, "vxlt_0_lowtown_docks_a3_d51_0_0.vlm"},
	{0x874D159DEE6374CDULL, "751441986.ogg"},
	{0x874D1EF03DD8C151ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\flemeth_idle_loop.gad"},
	{0x874D86DC3B023ADBULL, "art\\vfx\\materials\\vfx_distortion_edge.mao"},
	{0x874D99856A95E39FULL, "47653167.ogg"},
	{0x874DCC73E37C1F36ULL, "vxw_sundermount_378_0_1.vlm"},
	{0x874DED327E45F928ULL, "w_blightlands_89_0.mmh"},
	{0x874DF5389212C3EBULL, "follower_banter6169987.cl"},
	{0x874DF9327E460DCBULL, "w_blightlands_89_0.mao"},
	{0x874E0B327E462C52ULL, "w_blightlands_89_0.msh"},
	{0x874FCBFEA741650FULL, "137142711.ogg"},
	{0x874FCC6C17A17816ULL, "hlac_ht_bartrand_n_3370_2z.dds"},
	{0x87505734C377DB21ULL, "gen00fl_varric6033650.cl"},
	{0x8750A065C0AB3A03ULL, "318293030.ogg"},
	{0x8750B64629F2A806ULL, "gen00fl_merrillact2_gift.stg"},
	{0x8751E3558AC268F4ULL, "vxht_mansion2_1cm_0_0.vlm"},
	{0x875270068A69478DULL, "hlac_dr_slavers_3214_3x.dds"},
	{0x87528E2C302BBDBDULL, "hlac_lt_0_lowtown_docks_a3_n_1_0x.dds"},
	{0x8752C9914EDFDB1BULL, "w_bonepit_69_0.msh"},
	{0x8752DB914EDFFA66ULL, "w_bonepit_69_0.mao"},
	{0x8752E7914EE00E05ULL, "w_bonepit_69_0.mmh"},
	{0x87531FF03DDD2EFFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\flemeth\\flemeth_idle_loop.ani"},
	{0x87538A4B5FA3FFFCULL, "6064899_m.fxe"},
	{0x8754C7680D61F4D8ULL, "vxht_keep_n_5ip_0_0.vlm"},
	{0x8755CAF375165B04ULL, "6111597_m.fxe"},
	{0x875650DAFEC7284DULL, "mag321_comtesse_de_launcet6064890.cl"},
	{0x8756E2B1C5598F9DULL, "darktown_male_elf_sta6174550.cl"},
	{0x87578397BFBC7062ULL, "268317537.ogg"},
	{0x8757A2BB11AC487AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drink_to_drunk.gad"},
	{0x8757A39C5D75EEE1ULL, "vxca_cave_237_0_0.vlm"},
	{0x8757BE932995318FULL, "6129157_m.fxe"},
	{0x87587D41C0AAC826ULL, "hm_amb_darktown_fereldan.mor"},
	{0x8758E247EAD78819ULL, "521654017.ogg"},
	{0x8759087E7B5495F9ULL, "51268360.ogg"},
	{0x8759261B132985CFULL, "lt_0_lowtown_docks_118_0.phy"},
	{0x87596DB726B3559FULL, "hlac_lt_undercityencounter_-1_22y.dds"},
	{0x875A1C0B7C1C2DB8ULL, "vfx_vfx_p9.vsh.fxo.dx11.assembly"},
	{0x875A7DCCA0F0F179ULL, "dae000_sandal6147779.cl"},
	{0x875A85CFF0B194DCULL, "dae100_viscount6118483.cl"},
	{0x875AA330E43A7B37ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.cess.evt"},
	{0x875AC37B0BD09F92ULL, "vxlt_0_lowtown_docks_n_bup_0_0.vlm"},
	{0x875AF782979F0CF1ULL, "203875772.ogg"},
	{0x875B798CDC6F6561ULL, "561209208.ogg"},
	{0x875BF0D0B4E1CD41ULL, "zz_vrc_debug6064252.cl"},
	{0x875BFF3273F6A9ECULL, "vxdr_primevaltomb_4is_0_0.vlm"},
	{0x875C13318C96D6FDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pot_n.dds"},
	{0x875C1D0C49D57177ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheese.phy"},
	{0x875C246BD95F0B0AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc05.phy"},
	{0x875C7A52D68062F7ULL, "6040788_m.fxe"},
	{0x875DB00DB3288112ULL, "vxht_keep_7a3_0_0.vlm"},
	{0x875E88E933957F46ULL, "303835976.ogg"},
	{0x875F73114FDF2442ULL, "610706138.ogg"},
	{0x875F903F8B39D775ULL, "character_character_p39.psh.fxo.dx11"},
	{0x875FA22843DD9EDDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecart.phy"},
	{0x875FB4371FEF4A1FULL, "216129046.ogg"},
	{0x875FB79308A7DE7FULL, "dae200cr_viscount6087197.cl"},
	{0x875FFDE610D60E9BULL, "llac_ht_brothel_n_3367_1.dds"},
	{0x8760271913585454ULL, "288669913.ogg"},
	{0x876055327CB88174ULL, "vxht_mansion2_1fr_0_0.vlm"},
	{0x876096A85719E7BAULL, "hlac_lt_0_lowtown_n_-196608_4y.dds"},
	{0x8760A35D7BB0FACFULL, "vxlt_gallowstemplar_f_5pc_0_0.vlm"},
	{0x8760B8C06CD3B1F2ULL, "77499761.ogg"},
	{0x8760BC7DA46943D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_large.ani"},
	{0x8761A430E4409BC0ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.cess.gad"},
	{0x876233DD93647969ULL, "833519749.ogg"},
	{0x87626A018C85FF17ULL, "follower_banter6148739.cl"},
	{0x87628C9C473EE5F9ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.impa.ani"},
	{0x8762A41475E10F46ULL, "6106637_f.fxe"},
	{0x8762D0B9964B3F2DULL, "vxw_bonepit_zw_0_0.vlm"},
	{0x8762E8D68035813CULL, "w_sundermount_123_0.phy"},
	{0x8762F2347815D711ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_masf_0t.dds"},
	{0x8763ACBF41E0391EULL, "6097424_m.fxe"},
	{0x8763BAD118CC5374ULL, "6139308_m.fxe"},
	{0x8763D09C8BF7646AULL, "6149860_m.fxe"},
	{0x87640B192F9867FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_in.evt"},
	{0x87648774F200A495ULL, "crt_swarm_bee.mao"},
	{0x8764EE1CEB2DE93EULL, "6156200_m.fxe"},
	{0x8765261947125214ULL, "gen00fl_fenris6099891.cl"},
	{0x87653052A1D7E0FBULL, "llac_lt_0_lowtown_a3_n_c_131071_3.dds"},
	{0x8765604110409CCDULL, "llac_ca_caverns2prc_5802_52.dds"},
	{0x87658518867D4A59ULL, "6128380_f.fxe"},
	{0x87659BAC4E03A04FULL, "w_encounter3_162_0.mao"},
	{0x8765A7AC4E03B4ECULL, "w_encounter3_162_0.mmh"},
	{0x8765ADAC4E03BEDEULL, "w_encounter3_162_0.msh"},
	{0x876600A079327BBAULL, "hlac_w_encounter3_-3_0y.dds"},
	{0x87660E735098C8B2ULL, "6179262_f.fxe"},
	{0x876615794F6C0761ULL, "vxca_cave_246_0_0.vlm"},
	{0x8766CBD710C3A788ULL, "w_encounter2_camp_17_0.msh"},
	{0x8766D13C3211FDE8ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls101a_0n.dds"},
	{0x8766D8F4837A7917ULL, "6143207_m.fxe"},
	{0x8766DDD710C3C619ULL, "w_encounter2_camp_17_0.mao"},
	{0x8766E9D710C3DAB2ULL, "w_encounter2_camp_17_0.mmh"},
	{0x8766ED4D43A343B5ULL, "6101059_f.fxe"},
	{0x87670C3DA357B2ABULL, "follower_barks6153417.cl"},
	{0x8767EF0CF5E05EC5ULL, "395050531.ogg"},
	{0x8768196B9083D8B0ULL, "gui_passthrough_p0.psh.fxo.dx11"},
	{0x8768DF76EEEACBCEULL, "668081467.ogg"},
	{0x87693557FDC6B812ULL, "vxlt_0_lowtown_docks_n_brq_0_0.vlm"},
	{0x8769679C4744C53BULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.impa.gad"},
	{0x8769764023B40520ULL, "6113890_m.fxe"},
	{0x8769AD529071B54CULL, "one300_murder_followers6086733.cl"},
	{0x8769FFD6803BC764ULL, "w_sundermount_123_0.rml"},
	{0x876A06679C064DF1ULL, "zz_followers_banter6159351.cl"},
	{0x876A5881E10DB1EEULL, "vxht_brothel_n_6f7_0_0.vlm"},
	{0x876AC560C96CC5D8ULL, "follower_banter6163210.cl"},
	{0x876AD0C1C0E23515ULL, "lgt211ar_rand_sewer_1_nt.are"},
	{0x876AF18FDF3CAC40ULL, "819409889.ogg"},
	{0x876AFB06D96BC369ULL, "vxca_caverns2prc_a6m_0_0.vlm"},
	{0x876B04192F9E7A3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_in.gad"},
	{0x876B136A8F3D6659ULL, "6107384_m.fxe"},
	{0x876B49E246478636ULL, "6075208_m.fxe"},
	{0x876B9FB2DFA3291AULL, "vxlt_0_lowtown_a3_n_c_i06_0_0.vlm"},
	{0x876BBDBB11BD1F60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drink_to_drunk.ani"},
	{0x876BD2429B13B394ULL, "256562686.ogg"},
	{0x876C51FBCBB42F10ULL, "6067634_m.fxe"},
	{0x876C6FB4425A8D5DULL, "w_woundedcoast_162_0.phy"},
	{0x876CC8F1ED50B284ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closerock.mmh"},
	{0x876CCCF1ED50B9D7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closerock.mao"},
	{0x876CD0C9DEBB4813ULL, "pro000cr_soldier1.utc"},
	{0x876CDB25283EABBEULL, "877706238.ogg"},
	{0x876CDEF1ED50D826ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closerock.msh"},
	{0x876D17B1528D0717ULL, "vxht_0_hightown_a2_c4q_0_0.vlm"},
	{0x876D917678943907ULL, "751362397.ogg"},
	{0x876DAE52EC5E0858ULL, "gen00fl_isabela6101075.cl"},
	{0x876DB47DA473C00AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_large.evt"},
	{0x876DEA9A6AD72314ULL, "206794790.ogg"},
	{0x876E7BB46A4B6721ULL, "198764857.ogg"},
	{0x876E7E94F893B0B7ULL, "54872353.ogg"},
	{0x876EB8FC832BBB8EULL, "107277699.ogg"},
	{0x876EC70F6EAE99F4ULL, "vxht_mansion2_1is_0_0.vlm"},
	{0x876F43E14AF2BC69ULL, "llac_ht_0_hightown_n_-65535_0.dds"},
	{0x876F4459552A5B11ULL, "494732327.ogg"},
	{0x876F9E2F10C0B348ULL, "vxca_cave_243_0_0.vlm"},
	{0x877022C2FF361250ULL, "vxlt_0_lowtown_a2_n_dow_0_0.vlm"},
	{0x8770849C474B1520ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.impa.evt"},
	{0x8770972843EBFF49ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecart.mao"},
	{0x8770A32843EC13E2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecart.mmh"},
	{0x8770A52843EC1698ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecart.msh"},
	{0x8770C3A2CD20013FULL, "209851334.ogg"},
	{0x8770DCD42F204337ULL, "vxdr_0_deeproads_cu9_0_0.vlm"},
	{0x87711E40BF5B0AE1ULL, "69232205.ogg"},
	{0x877122E9072F98D4ULL, "plt_rdr100pt_keepers_note.nss"},
	{0x87716A259C4E3EF2ULL, "ave331_jeven6137283.cl"},
	{0x87718F88178194C7ULL, "6115381_m.fxe"},
	{0x87719270BF1148D9ULL, "hlac_ca_caverns2_5802_59x.dds"},
	{0x8771C5F3277D286BULL, "art\\levels\\races\\proxy\\props\\_textures\\bedsingle_d.dds"},
	{0x8771F268609D2529ULL, "mag371ip_at_courtyard.utp"},
	{0x877203192FA4973AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_twist_in.ani"},
	{0x87720CFC9AFE8EC0ULL, "hlac_dr_0_deeproads_17038_3x.dds"},
	{0x8772FA774B39AF4BULL, "hlac_dr_slavers_3216_10z.dds"},
	{0x877347710371F5C9ULL, "6127045_m.fxe"},
	{0x87737226B5D3C9E3ULL, "vxw_bonepit_20c_0_0.vlm"},
	{0x87739135246960F8ULL, "al_mage_s01.gda"},
	{0x87746E5B0D2604F8ULL, "vxlt_gallowscourtyard_n_7es_0_0.vlm"},
	{0x8774937DA479A77DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_large.gad"},
	{0x8774EB422E77A0BCULL, "vxdr_primevaltomb_36q_0_0.vlm"},
	{0x87751538C025D264ULL, "hlac_ca_caverns2_5802_47z.dds"},
	{0x8775368A45D07F2AULL, "rdr111dg_martin6054515.cl"},
	{0x877568471FD763AFULL, "hlac_lt_warehouse_n_0_3z.dds"},
	{0x877599356E239B83ULL, "follower_banter6172253.cl"},
	{0x87759F9C22EE3133ULL, "6117977_f.fxe"},
	{0x8775E258DE7CC0D4ULL, "117701735.ogg"},
	{0x8776046F78741253ULL, "vxlt_undercityserialkiller_1mq_0_0.vlm"},
	{0x87760BC8E44ED548ULL, "6002525_m.fxe"},
	{0x87766C26184CADE3ULL, "aip_qunarithrallmage_grd.gda"},
	{0x87767C9BDBE299FAULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock3.gad"},
	{0x8776FB5B80B0495AULL, "121949794.ogg"},
	{0x87771E9F7B9005AFULL, "lgt100cr_gang_3_ht.utc"},
	{0x877731E2FFD02847ULL, "mag100_karras6206237.cl"},
	{0x877778D45BB1CB9BULL, "6147782_m.fxe"},
	{0x8777AF2366EEECC1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww12_chantryframe.phy"},
	{0x87781398670F7151ULL, "hlac_dr_0_deeproads_5845_16x.dds"},
	{0x87797C167F1F472FULL, "hlac_ca_caverns2prc_5802_41z.dds"},
	{0x8779B8ECE470BAAFULL, "vxht_chantry_288_0_0.vlm"},
	{0x877A1D0E7AAB5F28ULL, "static1_punch_s2y.psh.fxo"},
	{0x877A3DB632F52C2BULL, "ave121_brennan6105907.cl"},
	{0x877B01025F4F8C6CULL, "one300dg_vanard6063950.cl"},
	{0x877B46D80E4BFACAULL, "517959958.ogg"},
	{0x877B898E44831F97ULL, "vxht_0_hightown_a2_c3p_0_0.vlm"},
	{0x877BD1C1C0F0AA22ULL, "lgt211ar_rand_sewer_1_nt.lst"},
	{0x877BEC8DB2713399ULL, "gen00fl_fenris6110791.cl"},
	{0x877BF0E02CD25CA6ULL, "t3_dey_bl4.tnt"},
	{0x877C2A1FECB2290BULL, "332802144.ogg"},
	{0x877C3437EF886E8AULL, "art\\vfx\\combat\\vfx_glb_ele_e_major_c.anb"},
	{0x877C9B1A546C41BBULL, "209931692.ogg"},
	{0x877D33E3513EE441ULL, "gen00fl_aveline6002523.cl"},
	{0x877D64B44268EDC9ULL, "w_woundedcoast_162_0.mao"},
	{0x877D70B442690262ULL, "w_woundedcoast_162_0.mmh"},
	{0x877D72B442690518ULL, "w_woundedcoast_162_0.msh"},
	{0x877DCC25566FDA2CULL, "vxca_caverns1_jg6_0_0.vlm"},
	{0x877E476A1DD8D9C2ULL, "661453919.ogg"},
	{0x877ED30F4C063BFCULL, "vxw_bonepit_1or_0_0.vlm"},
	{0x877F02A9A8EC9A40ULL, "6095386_f.fxe"},
	{0x877F4EB121165BB7ULL, "vxdr_0_deeproads_cr8_0_0.vlm"},
	{0x877F852539C8B0D2ULL, "vxlt_undercityserialkiller_1mt_0_0.vlm"},
	{0x877FC51466F31078ULL, "hlac_lt_hangedman_n_-1_20z.dds"},
	{0x8780894C82B38870ULL, "t3_wdg_w09.tnt"},
	{0x87809FC02DC3AD4CULL, "vxdr_0_deeproads_ecj_0_0.vlm"},
	{0x8780E06D43AED07DULL, "560518606.ogg"},
	{0x8780EBA0C93BE212ULL, "6013293_m.fxe"},
	{0x8781D3AAF7B011ADULL, "hlac_lt_backalley_n_1_5z.dds"},
	{0x8781F31119EF8D41ULL, "6139859_m.fxe"},
	{0x8781FB8E6BF0E916ULL, "6142414_m.fxe"},
	{0x8782096B1AAAA625ULL, "6108514_m.fxe"},
	{0x8782AC1FB61FFF63ULL, "138123054.ogg"},
	{0x8782AF59E7FBA6E7ULL, "vxlt_0_lowtown_a2_n_c_bz7_0_0.vlm"},
	{0x8782F36A2EC3881CULL, "214578816.ogg"},
	{0x8782FA739702F3EDULL, "608282555.ogg"},
	{0x8783162765EB1104ULL, "6129709_m.fxe"},
	{0x87840D24B94B0265ULL, "lowtown_male_lowlife1_sta6164948.cl"},
	{0x87843A430213FD18ULL, "33436702.ogg"},
	{0x87847865043666B7ULL, "6075061_m.fxe"},
	{0x87849B795E4FDD63ULL, "docks_fem_sailor_mob.bnk"},
	{0x8784D176E36156B5ULL, "310460991.ogg"},
	{0x8784D903FE2E8350ULL, "character_character_p89.psh.fxo.dx11.assembly"},
	{0x87854D5A2660CC62ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07a.msh"},
	{0x87854F5A2660CF18ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07a.mmh"},
	{0x8785E3C63551CBEEULL, "495506760.ogg"},
	{0x87872473B860C75AULL, "w_blightlands_102_0.mmh"},
	{0x87872873B860CD91ULL, "w_blightlands_102_0.mao"},
	{0x87873673B860E5A0ULL, "w_blightlands_102_0.msh"},
	{0x8787414460AC7B6DULL, "vxht_brothel_n_6gs_0_0.vlm"},
	{0x87876776871CE713ULL, "ser221_party_comments6164317.cl"},
	{0x87877D795E51DF93ULL, "docks_fem_sailor_mob.cnv"},
	{0x8787BB1997239A4BULL, "759964043.ogg"},
	{0x8787DAA28457D0DBULL, "hf_qun_petrice.mor"},
	{0x8787E234D90EDE79ULL, "436638326.ogg"},
	{0x8787EC4A2C2D6185ULL, "322844653.ogg"},
	{0x87881665F049A7B4ULL, "354932569.ogg"},
	{0x878A5BEBF4325768ULL, "820464853.ogg"},
	{0x878A5EC5580F4158ULL, "follower_banter6172735.cl"},
	{0x878A979BDBF370E0ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock3.ani"},
	{0x878BBEF4AA9B0B42ULL, "825858758.ogg"},
	{0x878C14C0A4F4E9BDULL, "844223163.ogg"},
	{0x878C9A000DDBB72EULL, "120603312.ogg"},
	{0x878CEB37AFA08840ULL, "vxht_mansion2b_25x_0_0.vlm"},
	{0x878CFBD8647A772FULL, "6084313_m.fxe"},
	{0x878D30B4C7B5A10DULL, "vxlt_0_lowtown_docks_a3_n_cqu_0_0.vlm"},
	{0x878D370626BE608DULL, "dae121ar_brothel_nt.lst"},
	{0x878DA4FCE801BE3EULL, "zz_vault_debug6203165.cl"},
	{0x878DE5CD7E846A39ULL, "zz_followers_banter6157999.cl"},
	{0x878DE66100DB65CBULL, "vxht_0_hightown_a2_c1j_0_0.vlm"},
	{0x878E26506C2D88ADULL, "vxw_bonepit_kt_0_0.vlm"},
	{0x878E60A400A70219ULL, "751669818.ogg"},
	{0x878EEFD57F03F6BEULL, "dre000_dougal6128889.cl"},
	{0x878F01A0097BE1E3ULL, "255745881.ogg"},
	{0x878F16771B551C07ULL, "fen200ip_slaver_chest_2.utp"},
	{0x878F26B88B63E127ULL, "hlac_w_encounter2_196609_0x.dds"},
	{0x878F565A26692FA3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye07a.phy"},
	{0x8790096CED5D301AULL, "124799972.ogg"},
	{0x8790587BA7221A79ULL, "mag330_walter6045997.cl"},
	{0x879058F57FE4EEC9ULL, "234348238.ogg"},
	{0x879075C76C78FFBFULL, "follower_banter6172288.cl"},
	{0x879081AA6F9A20C2ULL, "vxlt_0_lowtown_docks_b67_0_1.vlm"},
	{0x8790B5CE001F8A44ULL, "510719489.ogg"},
	{0x8790D05A255DDC80ULL, "hlac_ht_mansion_n_2995_1z.dds"},
	{0x8792397DAB16337CULL, "vxht_0_hightown_a2_ca9_0_0.vlm"},
	{0x87923BD79D78BB6BULL, "6014994_m.fxe"},
	{0x8792E8295C604353ULL, "vxlt_undercityserialkiller_1gs_0_0.vlm"},
	{0x8792FE2BCF29C9DEULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.impa.ani"},
	{0x87931A28E755416CULL, "6112013_m.fxe"},
	{0x87932F320150A9E0ULL, "zz_one_debug6099939.cl"},
	{0x8794143881C32DAEULL, "static_billboard_p0.vsh.fxo.dx11.assembly"},
	{0x87944F0626C4A81CULL, "dae121ar_brothel_nt.ncs"},
	{0x87946C29DDA56E10ULL, "vxw_woundedcoast_1cu_0_0.vlm"},
	{0x879500F4CAB33F23ULL, "vxlt_0_lowtown_a3_n_c_f5g_0_0.vlm"},
	{0x879541463EF4B049ULL, "933877488.ogg"},
	{0x87956280270C39ACULL, "55041160.ogg"},
	{0x879563C945C72F48ULL, "lt_0_lowtown_a3_n_c_28_0_ll.mmh"},
	{0x8795C05D961A8DB6ULL, "pro000cr_hurlock_fake.ncs"},
	{0x8795FD253F079F9DULL, "vxlt_gallowstemplar_f_3e0_0_0.vlm"},
	{0x879652F404C5C894ULL, "275946520.ogg"},
	{0x87965516F5CC2C54ULL, "follower_banter6163139.cl"},
	{0x87967ABE5DC1CCE8ULL, "42183796.ogg"},
	{0x8796E43E708E2C25ULL, "mer140_bonepit_comments6077809.cl"},
	{0x879709EF816275D6ULL, "mag120_macha6080676.cl"},
	{0x87984DFF4F1487FBULL, "42851377.ogg"},
	{0x8798B032B2FA821CULL, "vxlt_0_lowtown_docks_a3_cq3_0_0.vlm"},
	{0x8799034EB6B23FCBULL, "art\\vfx\\combat\\vfx_str_att_back_act_c.impa.evt"},
	{0x87991650864A0067ULL, "zz_ser_debug6078559.cl"},
	{0x8799180A1E00A92CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstall.msh"},
	{0x87991E0A1E00B31EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstall.mmh"},
	{0x87992A0A1E00C7B5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstall.mao"},
	{0x8799DFD0A95FFC6BULL, "dae000_fem_commoner2_mob6174583.cl"},
	{0x879A0FE45503386EULL, "mag200_feynriel6023966.cl"},
	{0x879A3C5E2EE74563ULL, "6083125_m.fxe"},
	{0x879A8B098FF59299ULL, "zz_one_debug6058979.cl"},
	{0x879AC9D25D338CA0ULL, "llac_lt_0_lowtown_docks_n_-131074_2.dds"},
	{0x879B6D033A6D0AA9ULL, "vxht_brothel_n_3v5_0_0.vlm"},
	{0x879C8502E745AD0CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_basic.phy"},
	{0x879D2E423B90E6D8ULL, "vxdr_primevalentrance_16j_0_1.vlm"},
	{0x879E2E8CFF9A1AFEULL, "600736501.ogg"},
	{0x879E3734BAFBE2D8ULL, "vxw_encounter2_camp_n_0_0.vlm"},
	{0x879ECB7717C133E8ULL, "517252336.ogg"},
	{0x879F2367FCE7019FULL, "mcr310_sibling6076380.cl"},
	{0x879F48DBB445A0BAULL, "vxw_bonepitmining_a2_1u8_0_0.vlm"},
	{0x879F5540DEC6E01DULL, "6129444_f.fxe"},
	{0x879F7FAECA86CC1FULL, "zz_one_debug6075028.cl"},
	{0x879FBD9252D3C659ULL, "hlac_dr_0_deeproads_5841_1z.dds"},
	{0x87A0162BCF347CAFULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.impa.evt"},
	{0x87A01C4EB6B8888CULL, "art\\vfx\\combat\\vfx_str_att_back_act_c.impa.gad"},
	{0x87A0259E8E1EB57CULL, "vxca_caverns1_efv_0_0.vlm"},
	{0x87A049FEAAD1E22BULL, "vxlt_gallowscourtyard_74d_0_0.vlm"},
	{0x87A057A5DF3C3AACULL, "vxw_encounter2_camp_o5_0_0.vlm"},
	{0x87A09AA006AECFB2ULL, "lgt310ip_body_1.ncs"},
	{0x87A0EBC982FE0A74ULL, "vxdr_primevalentrance_2kq_0_2.vlm"},
	{0x87A0F5F8CAA70099ULL, "132633285.ogg"},
	{0x87A0FE1F00119185ULL, "vxht_mansion2_3cf_0_0.vlm"},
	{0x87A159177D55C490ULL, "6125756_m.fxe"},
	{0x87A1E7AF15C40484ULL, "914260925.ogg"},
	{0x87A2206E7423732FULL, "293864065.ogg"},
	{0x87A25C1476233F58ULL, "vxw_bonepit_29c_0_0.vlm"},
	{0x87A2816FA43AAD1BULL, "hlac_w_bonepit_327681_0x.dds"},
	{0x87A2A4A7060DD9AAULL, "hlac_lt_undercityencounter_-1_33z.dds"},
	{0x87A2C5003938BEFAULL, "w_bonepitdestroyed_a3_242_0.msh"},
	{0x87A2C5EC898DB685ULL, "vxlt_gallowscourtyard_n_7ej_0_0.vlm"},
	{0x87A2C6486CDD797BULL, "vxw_bonepitdestroyed_a3_197_0_0.vlm"},
	{0x87A2D3003938D633ULL, "w_bonepitdestroyed_a3_242_0.mao"},
	{0x87A2D7003938DD40ULL, "w_bonepitdestroyed_a3_242_0.mmh"},
	{0x87A2DE06CF9B8690ULL, "vxw_woundedcoast_1nr_0_0.vlm"},
	{0x87A376ED1B2A382EULL, "vxw_bonepitmining_a2_o9_0_0.vlm"},
	{0x87A3C6E14ADDFB3FULL, "6113642_m.fxe"},
	{0x87A3F20C4A12DDCFULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheese.mao"},
	{0x87A3FE0C4A12F26CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheese.mmh"},
	{0x87A4040C4A12FC5EULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheese.msh"},
	{0x87A43737EFA9C67DULL, "art\\vfx\\combat\\vfx_glb_ele_e_major_c.mmh"},
	{0x87A4B7BC620138BEULL, "vxht_keep_n_6uy_0_0.vlm"},
	{0x87A55FFDF149F91CULL, "558573924.ogg"},
	{0x87A61B4EB6BCF28AULL, "art\\vfx\\combat\\vfx_str_att_back_act_c.impa.ani"},
	{0x87A6578434C73BBCULL, "906961203.ogg"},
	{0x87A6918181A8BC9FULL, "plt_cod_cha_sebastian2.plo"},
	{0x87A6BEB4899B1F7DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofne.mao"},
	{0x87A6C2B4899B25B6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofne.mmh"},
	{0x87A6CCB4899B36B4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofne.msh"},
	{0x87A6D78295D8CACDULL, "gen00fl_fenris6128019.cl"},
	{0x87A7172BCF3A9D58ULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.impa.gad"},
	{0x87A76FF1724FBA5DULL, "dae000ip_crowd_lm3_talk6.utp"},
	{0x87A85A9DBA24706EULL, "vxw_sundermount_278_0_0.vlm"},
	{0x87A8BE11E18E62D4ULL, "ser120_ghyslain6104187.cl"},
	{0x87A8DF96DD5F0DE4ULL, "6102809_m.fxe"},
	{0x87A9734398D76733ULL, "gen00fl_varric6033630.cl"},
	{0x87A9E9C36B0C28F9ULL, "577128577.ogg"},
	{0x87A9FFC4FFC61CFEULL, "527343046.ogg"},
	{0x87AA15629665862DULL, "hlac_lt_0_lowtown_docks_n_-1_2z.dds"},
	{0x87AA8A8B1A299857ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn01.phy"},
	{0x87AAA8F9C3753C37ULL, "hlac_lt_0_lowtown_docks_-3_3z.dds"},
	{0x87AAEF48F8AEE178ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.anb"},
	{0x87AAF7F079C8944DULL, "304037765.ogg"},
	{0x87AB246D19D674EBULL, "ts_prog_coral.gda"},
	{0x87AB26A5118AE041ULL, "hlac_lt_gallowstemplar_2383_2y.dds"},
	{0x87AB5FA7E71E3BA5ULL, "vxlt_darktown_42m_0_0.vlm"},
	{0x87ABE00E944DA3B0ULL, "710704768.ogg"},
	{0x87AC602B8BC201A8ULL, "hlac_lt_0_lowtown_docks_131070_1y.dds"},
	{0x87AC8B842B792608ULL, "vxlt_0_lowtown_a2_n_c_hd0_0_0.vlm"},
	{0x87ACB553778CBC07ULL, "vxlt_gallowstemplar_5nk_0_1.vlm"},
	{0x87ACDC1D1792008CULL, "hlac_ht_0_hightown_a3_-196609_0z.dds"},
	{0x87ACDEC9BD6BD7CEULL, "66229995.ogg"},
	{0x87AD1B1EC2048DA2ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastplants_sa.dds"},
	{0x87AD4373B881CB65ULL, "w_blightlands_102_0.phy"},
	{0x87ADBAB8A63BB93AULL, "vxw_bonepitmining_a2_1r9_0_0.vlm"},
	{0x87ADC7B8392BFD1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ed_transition.ani"},
	{0x87AE115D66EF873FULL, "t3_grndoak.tnt"},
	{0x87AE4A6395452AAAULL, "vxw_bonepitmining_a2_h4_0_0.vlm"},
	{0x87AE708EC4511B8AULL, "art\\characters\\creatures\\strider\\animation\\blendtree\\strider.anb"},
	{0x87AEBBDB9CC7FAABULL, "vxlt_gallowscourtyard_79g_0_0.vlm"},
	{0x87AEDDC308CE9C72ULL, "6128872_m.fxe"},
	{0x87AF7637481A8CDEULL, "vxlt_0_lowtown_fe2_0_0.vlm"},
	{0x87AFA72314F032C4ULL, "hlac_w_encounter3_-131070_0x.dds"},
	{0x87AFF80A653408A1ULL, "vxlt_undercityserialkiller_1jb_0_0.vlm"},
	{0x87B045665C8B71E9ULL, "90348496.ogg"},
	{0x87B0C69E0F89770FULL, "vxht_mansion2b_1ne_0_0.vlm"},
	{0x87B14FE3C1919F10ULL, "vxw_woundedcoast_1is_0_0.vlm"},
	{0x87B20A5BD419685EULL, "vxlt_gallowstemplar_n_3qs_0_0.vlm"},
	{0x87B2510673CF4025ULL, "71143221.ogg"},
	{0x87B27DBDFEE7E110ULL, "hlac_lt_undercityserialkiller_-1_36y.dds"},
	{0x87B2916C1DB10BDBULL, "gen00fl_fenris6126994.cl"},
	{0x87B2A16EC9B172FBULL, "vxca_cavern1_haunted_in0_0_0.vlm"},
	{0x87B36B3A9C29BE36ULL, "qun131_party_comment6117274.cl"},
	{0x87B3D962B505D94CULL, "vxht_chantry_333_0_0.vlm"},
	{0x87B468D4AF7B5E4BULL, "6094976_m.fxe"},
	{0x87B47D29E2138BA4ULL, "vxht_0_hightown_aic_0_0.vlm"},
	{0x87B4D4B83932325EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ed_transition.gad"},
	{0x87B4DBA790AE4A67ULL, "gen00fl_anders6201323.cl"},
	{0x87B5273DAF5BDD35ULL, "vxlt_0_lowtown_a2_h1w_0_0.vlm"},
	{0x87B543E98269ABC4ULL, "zz_followers_banter6158427.cl"},
	{0x87B54FDD9AA36B98ULL, "gen00fl_anders6121642.cl"},
	{0x87B56A6F943BFA0AULL, "449409376.ogg"},
	{0x87B5F29A011EC647ULL, "vxht_mansion_n_6jy_0_0.vlm"},
	{0x87B5FEAE74C5E930ULL, "vxlt_foundry_n_vb_0_0.vlm"},
	{0x87B64E36637FB562ULL, "vxw_sundermount_23x_0_0.vlm"},
	{0x87B6E73DD285A3A4ULL, "ip_restrict_follower.ncs"},
	{0x87B72D5F6494ABBFULL, "llac_ht_0_hightown_-131074_1.dds"},
	{0x87B73D01EE8A557FULL, "mer210_seldrek6034152.cl"},
	{0x87B75173CF552664ULL, "6107709_m.fxe"},
	{0x87B763425F4A6640ULL, "bodyimp.dds"},
	{0x87B8035AF5E6E120ULL, "6064666_m.fxe"},
	{0x87B809AEE15029B2ULL, "cry_val_b01.mao"},
	{0x87B813F087AFEF21ULL, "gen00fl_aveline6113975.cl"},
	{0x87B81449A0A58DD2ULL, "6096276_m.fxe"},
	{0x87B8B9600DB060D6ULL, "vxlt_0_lowtown_docks_a3_czd_0_0.vlm"},
	{0x87B9400626E3AB4AULL, "dae121ar_brothel_nt.are"},
	{0x87B956B820DF9CB0ULL, "vxlt_gallowstemplar_f_5rj_0_0.vlm"},
	{0x87B96C0004B3D11AULL, "vxlt_0_lowtown_docks_a3_d30_0_0.vlm"},
	{0x87B986ECEF688C0EULL, "hlac_w_woundedcoast_131080_0z.dds"},
	{0x87B9D184D9145425ULL, "vxlt_darktown_49l_0_0.vlm"},
	{0x87BA3C7757F017DEULL, "191262982.ogg"},
	{0x87BABCC21B923BFFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba01a.mmh"},
	{0x87BAC2C21B9245A1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba01a.msh"},
	{0x87BACB5D9639BCD0ULL, "pro000cr_hurlock_fake.utc"},
	{0x87BB3546808BFC7FULL, "gen_im_trs_krk_low_junk1_02.uti"},
	{0x87BB4E0A738C7C64ULL, "355061893.ogg"},
	{0x87BB5201460B1D0AULL, "crt_pro_cycle2.mao"},
	{0x87BB74FEAC6C4666ULL, "vxw_bonepitmining_a2_200_0_0.vlm"},
	{0x87BBAFB83938127DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ed_transition.evt"},
	{0x87BBD4B83DD70833ULL, "vxca_caverns1_j1y_0_2.vlm"},
	{0x87BBF7B74357607CULL, "6096488_f.fxe"},
	{0x87BBFD19732179CFULL, "hla_ht_0_hightown_-65534_0y.dds"},
	{0x87BC2A1BEAE0A38CULL, "115939710.ogg"},
	{0x87BC58DE22AB399AULL, "vxlt_0_lowtown_docks_a3_aph_0_0.vlm"},
	{0x87BCABC173A9DCE8ULL, "gen00fl_aveline6121146.cl"},
	{0x87BCBE638A82540EULL, "vxlt_darktown_5cd_0_0.vlm"},
	{0x87BCEC7F506685B8ULL, "6044934_m.fxe"},
	{0x87BCFB0DF1F0BF25ULL, "6163580_m.fxe"},
	{0x87BD64AA887BF4E0ULL, "hlac_lt_gallowstemplar_2388_2y.dds"},
	{0x87BDA2DEF86AFA67ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementaldamage\\physical\\extremekncokdown.gad"},
	{0x87BDB8B51756A51DULL, "hlac_ca_caverns1_24422_4z.dds"},
	{0x87BDF30DACE411EEULL, "547971201.ogg"},
	{0x87BE69E7572A2121ULL, "vxlt_undercityserialkiller_1oa_0_0.vlm"},
	{0x87BE6E3778C316C1ULL, "art\\levels\\ztools\\shaders\\blank_e.dds"},
	{0x87BEA83E32FF392EULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_c.impa.evt"},
	{0x87BF387B017F8F8FULL, "vxht_mansion2b_1ad_0_0.vlm"},
	{0x87C01E8406EFF87DULL, "vxw_bonepitdestroyed_a3_27h_0_0.vlm"},
	{0x87C065D82559A440ULL, "hlac_ht_mansion2b_818_58y.dds"},
	{0x87C08A5F48E169A7ULL, "vxw_woundedcoast_hn_0_0.vlm"},
	{0x87C0C8AC470DBB66ULL, "14662863.ogg"},
	{0x87C0D6279CE6096AULL, "76424184.ogg"},
	{0x87C1134BBBA78B7BULL, "vxca_cavern1_haunted_ik3_0_0.vlm"},
	{0x87C15B4138857566ULL, "703241296.ogg"},
	{0x87C15D7A0A4688ADULL, "6060519_m.fxe"},
	{0x87C1EE8A2FC0C97AULL, "llac_lt_0_lowtown_-65535_0.dds"},
	{0x87C24B3FA6FBF1CCULL, "vxht_chantry_360_0_0.vlm"},
	{0x87C288E6471A164DULL, "142491941.ogg"},
	{0x87C2B591C222E62DULL, "vxdr_primevaltomb_5ak_0_0.vlm"},
	{0x87C2E8D7111237ADULL, "w_encounter2_camp_17_0.phy"},
	{0x87C30B9FA5E70064ULL, "zz_followers_banter6158193.cl"},
	{0x87C3125957022461ULL, "544965792.ogg"},
	{0x87C350DD29FB6C81ULL, "6017662_f.fxe"},
	{0x87C37A1E332FFF50ULL, "6000934_m.fxe"},
	{0x87C397DEF86F5374ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementaldamage\\physical\\extremekncokdown.evt"},
	{0x87C3B8EE2DE3FC2EULL, "718025133.ogg"},
	{0x87C46476F314DEC7ULL, "vxht_mansion_n_6mx_0_0.vlm"},
	{0x87C4C0002542AC10ULL, "sib131_slave_a6165256.cl"},
	{0x87C4C0135575CDE2ULL, "vxw_sundermount_24y_0_0.vlm"},
	{0x87C4C516607E3A86ULL, "hlac_dr_primevalentrance_1726_8x.dds"},
	{0x87C51BCFC9A0E2A7ULL, "vxht_0_hightown_a2_3ub_0_0.vlm"},
	{0x87C59F3E330548D9ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_c.impa.gad"},
	{0x87C5C44D14DF1D72ULL, "vxdr_0_deeproads_cv5_0_0.vlm"},
	{0x87C61E4BE8A85509ULL, "vxlt_darktown_6jc_0_0.vlm"},
	{0x87C67F1CE49CB236ULL, "389998675.ogg"},
	{0x87C68DA006CF8897ULL, "lgt310ip_body_1.utp"},
	{0x87C694057AC13573ULL, "vxca_caverns2_7a7_0_2.vlm"},
	{0x87C6A96F9F388C7AULL, "lt_0_lowtown_docks_a3_n_30_0.anb"},
	{0x87C74D78FCD55DA4ULL, "dae000_bartrand6064047.cl"},
	{0x87C78500CE6575F0ULL, "vxht_0_hightown_a3_a6o_0_0.vlm"},
	{0x87C7B8C7A26F21F4ULL, "llac_w_woundedcoast_0_7.dds"},
	{0x87C7C89512D5B530ULL, "vxlt_gallowstemplar_f_5uk_0_0.vlm"},
	{0x87C7DDDCF6A9E99AULL, "vxlt_0_lowtown_docks_a3_d41_0_0.vlm"},
	{0x87C810578D227E8BULL, "vxw_woundedcoast_129_0_0.vlm"},
	{0x87C8434D54CBDC74ULL, "w_bonepit_38_0.rml"},
	{0x87C8450E04A1EF58ULL, "869769119.ogg"},
	{0x87C8829238ACC7F6ULL, "gen00fl_carver6102346.cl"},
	{0x87C8831E0A9B2386ULL, "vxht_0_hightown_b92_0_0.vlm"},
	{0x87C890AB1B869201ULL, "6155934_m.fxe"},
	{0x87C895B8E154E137ULL, "qcr000_arishok6059988.cl"},
	{0x87C8CB64A55D7B51ULL, "vxlt_gallowstemplar_f_6tb_0_0.vlm"},
	{0x87C8E34987FFB091ULL, "gen00fl_carver6103410.cl"},
	{0x87C953795B0D0D7DULL, "6105459_f.fxe"},
	{0x87C97CE553778D51ULL, "lt_0_lowtown_docks_254_0.msh"},
	{0x87C98AE55377A56CULL, "lt_0_lowtown_docks_254_0.mao"},
	{0x87C996E55377B90FULL, "lt_0_lowtown_docks_254_0.mmh"},
	{0x87C99D4485421D30ULL, "one300_nuncio6084495.cl"},
	{0x87C9E9D711185859ULL, "w_encounter2_camp_17_0.rml"},
	{0x87C9EA4CF5CD45E7ULL, "da2terrain_terrain_p3.psh.fxo.dx11.assembly"},
	{0x87CA377EB0D5621BULL, "6087339-toomanyshemlen.cut"},
	{0x87CA501DFE51E2D6ULL, "6052587_f.fxe"},
	{0x87CC48C82C9E2F2CULL, "vxht_0_hightown_n_aoz_0_0.vlm"},
	{0x87CC954ACF620451ULL, "hlac_w_bonepit_3_0x.dds"},
	{0x87CCB03E330B8427ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_c.impa.ani"},
	{0x87CD5EBDFE081409ULL, "471997565.ogg"},
	{0x87CDA23E1308D79DULL, "vxlt_gallowstemplar_n_358_0_0.vlm"},
	{0x87CE28A7BFCE7E59ULL, "988804585.ogg"},
	{0x87CE34BE6E1EB6F0ULL, "one300_zevran6089085.cl"},
	{0x87CE9F0767B6FEF0ULL, "dae000_madam_lusine6151506.cl"},
	{0x87CEAEDB7D027AB0ULL, "hlac_lt_0_lowtown_a3_n_c_131070_0y.dds"},
	{0x87CEC82CDDCF9D62ULL, "hlac_lt_0_lowtown_a2_n_131072_6x.dds"},
	{0x87CEE7D7A89AE055ULL, "vxht_bartrand_n_32m_0_0.vlm"},
	{0x87CEFC3C3AD78227ULL, "vxw_woundedcoast_ei_0_0.vlm"},
	{0x87CF0389D06082AEULL, "qun110_petrice6042148.cl"},
	{0x87CF16B4ADF84A09ULL, "hlac_ht_playershouse_n_855_36y.dds"},
	{0x87CF27E380DBF241ULL, "vxht_chantry_n_28a_0_0.vlm"},
	{0x87CF7D7C1F6A8E7FULL, "923678183.ogg"},
	{0x87CFFF89BCF37292ULL, "6095167_m.fxe"},
	{0x87D00DE9E8C4A907ULL, "615798771.ogg"},
	{0x87D0290FBB3FCF6EULL, "6093643_f.fxe"},
	{0x87D03F4619DB2843ULL, "6026177_m.fxe"},
	{0x87D0F9E3EE0140B5ULL, "6013311_m.fxe"},
	{0x87D18FDEF87B82ADULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementaldamage\\physical\\extremekncokdown.ani"},
	{0x87D1919ACB09DBD5ULL, "vxlt_gallowstemplar_n_51f_0_0.vlm"},
	{0x87D1BDD6DBB846C4ULL, "zz_runcombatdemo.ncs"},
	{0x87D1C16FCCBCD67DULL, "dae000_pc_home_ambients6208674.cl"},
	{0x87D1F0214CC53DDEULL, "106530856.ogg"},
	{0x87D2083E3B0166E9ULL, "vxw_woundedcoast_1ez_0_0.vlm"},
	{0x87D21D5F6160CA43ULL, "6156722_m.fxe"},
	{0x87D2C492DE6D2CB2ULL, "vxca_caverns2prc_8mv_0_0.vlm"},
	{0x87D2EE48F8D03113ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.mmh"},
	{0x87D3B7DD9394FC07ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_h01_0t.dds"},
	{0x87D50346B7AB3C6BULL, "hlac_lt_0_lowtown_a2_131072_4x.dds"},
	{0x87D50570056F08BAULL, "vxht_playershouse_n_1px_0_0.vlm"},
	{0x87D51BCE9517C48AULL, "6178381_m.fxe"},
	{0x87D5B0EFC72A46BBULL, "vxw_encounter2_3z_0_0.vlm"},
	{0x87D614269867049EULL, "dae021_keep_noblewoman1.utc"},
	{0x87D63A7204CBCDB0ULL, "vxlt_gallowstemplar_f_5xh_0_0.vlm"},
	{0x87D64AC9AE84349FULL, "vxlt_0_lowtown_n_gck_0_0.vlm"},
	{0x87D66A4B088A2F66ULL, "6077170_m.fxe"},
	{0x87D6711C39C186D2ULL, "6097867_f.fxe"},
	{0x87D6A09A6A629BBEULL, "lgt120_orz_bln_off_messenge6140224.cl"},
	{0x87D73D41975393D1ULL, "vxlt_gallowstemplar_f_6qa_0_0.vlm"},
	{0x87D78CB18AB05BDAULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_pot_n.dds"},
	{0x87D851E72E98E6F1ULL, "6099339_m.fxe"},
	{0x87D892F60E496F22ULL, "vxlt_warehouse_n_15d_0_0.vlm"},
	{0x87D8E77AC69AD453ULL, "603100162.ogg"},
	{0x87D8F31D65C250ACULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2_up.ani"},
	{0x87D9B4236741BFFDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww12_chantryframe.mao"},
	{0x87D9B8236741C636ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww12_chantryframe.mmh"},
	{0x87D9C2236741D734ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww12_chantryframe.msh"},
	{0x87D9F50A1E3804B9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstall.phy"},
	{0x87DA3DBEE9926343ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallgreen_0.pwk"},
	{0x87DA4BF57CF7F79DULL, "ave331_post_fight6138811.cl"},
	{0x87DAF82EA3C23C16ULL, "zz_dae_debug6029820.cl"},
	{0x87DB03DD3148F05EULL, "vxlt_0_lowtown_n_g94_0_0.vlm"},
	{0x87DB0DB044CAF5DAULL, "6131144_m.fxe"},
	{0x87DBCEB109BAD20FULL, "247144302.ogg"},
	{0x87DC0A097FA485ACULL, "vxlt_0_lowtown_docks_con_0_0.vlm"},
	{0x87DC2CA2AB2E36B0ULL, "ip_restrict_style_onehanded.ncs"},
	{0x87DC3F8B7DC21BE4ULL, "vxca_cavern1_haunted_i1k_0_1.vlm"},
	{0x87DCEA3C412EB2F6ULL, "zz_mer_debug6023990.cl"},
	{0x87DD2A8D227EA850ULL, "6174369_m.fxe"},
	{0x87DD4555264F83E4ULL, "vxht_brothel_n_54h_0_0.vlm"},
	{0x87DD5820B5FDC54DULL, "6029048_m.fxe"},
	{0x87DD5C02E77CFED7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_basic.mmh"},
	{0x87DD6002E77D05E4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_basic.mao"},
	{0x87DD667D3969D41CULL, "654161267.ogg"},
	{0x87DD6E192CCD9AA7ULL, "vxw_woundedcoast_bh_0_0.vlm"},
	{0x87DD7202E77D2449ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_basic.msh"},
	{0x87DD97B5DC925C38ULL, "dae000_mother6121053.cl"},
	{0x87DDCC0A1A692D17ULL, "vrc221_bartrand6065344.cl"},
	{0x87DDD8FAAB8B39A6ULL, "zz_lgt_2_debug6140448.cl"},
	{0x87DDE853F47B86A8ULL, "au_dot.mao"},
	{0x87DDF6A2F5DF22F0ULL, "hlac_lt_0_lowtown_a2_n_c_65537_7y.dds"},
	{0x87DEA5648270364FULL, "hlac_lt_0_lowtown_n_65536_7y.dds"},
	{0x87DF313BC3D00700ULL, "hlac_lt_warehouse_n_65536_0y.dds"},
	{0x87DF332FFA5A6173ULL, "tbe300im_battle_helm.uti"},
	{0x87E0001D65C885EEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2_up.gad"},
	{0x87E0023A04FED792ULL, "354466521.ogg"},
	{0x87E052E8A762A646ULL, "613102676.ogg"},
	{0x87E0A42266BD0147ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion2.gad"},
	{0x87E167394D802176ULL, "334358088.ogg"},
	{0x87E1D8F93D04CD20ULL, "vxht_playershouse_n_1rh_0_0.vlm"},
	{0x87E1E09A235A0C2EULL, "6121568_f.fxe"},
	{0x87E29C0D1C4239FFULL, "22427172.ogg"},
	{0x87E2B993517A440EULL, "555404467.ogg"},
	{0x87E2E0241D71E672ULL, "6162962_m.fxe"},
	{0x87E31E1E6B1AEBE3ULL, "vxw_blightlands_11v_0_0.vlm"},
	{0x87E32FCEC47C81B4ULL, "549698308.ogg"},
	{0x87E35CAD90B79749ULL, "l_fen200ar_ambush_exterior.bnk"},
	{0x87E3B0FEB2E97E0AULL, "art\\levels\\races\\vfx_models\\_textures\\prp_lava1_e.dds"},
	{0x87E3B23166798A2DULL, "plt_cod_itm_gem_keroshek.plo"},
	{0x87E4AC4EF6C1E630ULL, "vxlt_gallowstemplar_f_5ci_0_0.vlm"},
	{0x87E518E51BC6148BULL, "hlac_lt_0_lowtown_a3_n_c_65538_3x.dds"},
	{0x87E547459CEF7219ULL, "lgt220ar_gang_hideout_ht.are"},
	{0x87E5B4A286C3562DULL, "433291843.ogg"},
	{0x87E66BC19F216AFEULL, "vxdr_primevaltomb_52i_0_0.vlm"},
	{0x87E6B51E4595B508ULL, "llac_lt_0_lowtown_a2_0_9.dds"},
	{0x87E6FB1D65CE9BCDULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2_up.evt"},
	{0x87E74452BC06E486ULL, "hlac_lt_gallowstemplar_2388_5z.dds"},
	{0x87E78EAB5AA33A73ULL, "hlac_ht_mansion2b_818_46y.dds"},
	{0x87E7992266C30D54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion2.evt"},
	{0x87E7AFBF848EDE78ULL, "fxc_puke_ani2.mao"},
	{0x87E7CB68C8570A16ULL, "vxdr_primevaltomb_705_0_0.vlm"},
	{0x87E7F7948F303787ULL, "store_dae310_tailor.utm"},
	{0x87E813FA60649D8EULL, "vxlt_darktown_5rg_0_0.vlm"},
	{0x87E81B2683E6EC12ULL, "hlac_lt_ruinedhouse_n_-1_0z.dds"},
	{0x87E8698A1EA4133EULL, "art\\characters\\creatures\\corpse\\animation\\combat\\runforward.ani"},
	{0x87E871C116146E05ULL, "100218195.ogg"},
	{0x87E920619C92C35DULL, "693503948.ogg"},
	{0x87E93AF13EF9382BULL, "lowtown_male_urchin_sta6161589.cl"},
	{0x87E988C976C59FFDULL, "22904448.ogg"},
	{0x87E98C1C271D4ECDULL, "vxlt_0_lowtown_a2_hgy_0_0.vlm"},
	{0x87E9C1F04AEEBB5BULL, "llac_lt_darktown_-1_89.dds"},
	{0x87E9C96DD4BB0C4CULL, "230944307.ogg"},
	{0x87EA2FC3462F01A2ULL, "6095875_m.fxe"},
	{0x87EA56F47CD1BC8EULL, "425801705.ogg"},
	{0x87EA913B6BBA3CF3ULL, "6174946_m.fxe"},
	{0x87EADCA36BE66964ULL, "6168953_m.fxe"},
	{0x87EAFEE3BC8FB8E9ULL, "789374221.ogg"},
	{0x87EB3879541B6862ULL, "gen00fl_merrill6111407.cl"},
	{0x87EB62D340DC069CULL, "w_woundedcoast_189_0.mao"},
	{0x87EB63C7026B6446ULL, "839086299.ogg"},
	{0x87EB688A6311FD66ULL, "662259189.ogg"},
	{0x87EB6ED340DC1B3FULL, "w_woundedcoast_189_0.mmh"},
	{0x87EB74D340DC25E1ULL, "w_woundedcoast_189_0.msh"},
	{0x87EBF8CBA2D501DDULL, "mag371_elsa6064656.cl"},
	{0x87EC2192E37EE9D6ULL, "952808275.ogg"},
	{0x87EC6F33D9BECDEAULL, "lt_0_lowtown_docks_216_0.mao"},
	{0x87EC7333D9BED439ULL, "lt_0_lowtown_docks_216_0.mmh"},
	{0x87EC7D33D9BEE527ULL, "lt_0_lowtown_docks_216_0.msh"},
	{0x87ECB797077E0983ULL, "447955368.ogg"},
	{0x87ECC7CD5224AA9BULL, "491189508.ogg"},
	{0x87ED676DA0B7D679ULL, "vxlt_0_lowtown_docks_n_cno_0_0.vlm"},
	{0x87ED9C536B525B4CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs107a_0e.dds"},
	{0x87ED9D531ADF109BULL, "gen00fl_anders6104017.cl"},
	{0x87EE115C3B4C4342ULL, "646183606.ogg"},
	{0x87EE8F64ACE4EF39ULL, "dr_dungeon_3215.anb"},
	{0x87EEB17A8E4D8217ULL, "96215581.ogg"},
	{0x87EEEBF81EED97E9ULL, "vxw_woundedcoast_1kx_0_0.vlm"},
	{0x87EF0F667028FA30ULL, "llac_lt_0_lowtown_a2_n_c_65536_3.dds"},
	{0x87EF79EC13AA31FFULL, "vxht_chantry_21g_0_0.vlm"},
	{0x87F016C5A5FD1FF8ULL, "vxlt_gallowstemplar_n_46k_0_0.vlm"},
	{0x87F0B456D7120A41ULL, "213585047.ogg"},
	{0x87F0E6677BF15248ULL, "art\\levels\\races\\design\\prp_lyrium.mmh"},
	{0x87F0F2677BF166EBULL, "art\\levels\\races\\design\\prp_lyrium.mao"},
	{0x87F104677BF18572ULL, "art\\levels\\races\\design\\prp_lyrium.msh"},
	{0x87F108FB74A5BEBDULL, "vxlt_0_lowtown_a2_n_c_gn0_0_0.vlm"},
	{0x87F188E5D24936C8ULL, "hlac_ca_cavern1_haunted_24423_10x.dds"},
	{0x87F1F0C073201AFEULL, "vxdr_primevaltomb_6v2_0_0.vlm"},
	{0x87F1F2CF9FF54DC2ULL, "407359288.ogg"},
	{0x87F225343D0A8335ULL, "504228831.ogg"},
	{0x87F26B8B1A67194CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn01.mmh"},
	{0x87F2718B1A67233EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn01.msh"},
	{0x87F28E4DE9261142ULL, "mer210_smith6051384.cl"},
	{0x87F2907AD2222D72ULL, "vxlt_0_lowtown_a2_n_bye_0_0.vlm"},
	{0x87F29AE8557F45D9ULL, "60710221.ogg"},
	{0x87F2A4EB5D456643ULL, "dae000_keep_noble_sta6161522.cl"},
	{0x87F2FC2506F4FA63ULL, "vxca_caverns1_ei5_0_1.vlm"},
	{0x87F405AE52C716CDULL, "vxlt_0_lowtown_docks_b55_0_1.vlm"},
	{0x87F412DC91514A89ULL, "6061242_m.fxe"},
	{0x87F43CC0747E532BULL, "vxht_0_hightown_a2_c71_0_1.vlm"},
	{0x87F45D70C84DBA05ULL, "hairalpha_rendertodepth_p0.psh.fxo.dx11"},
	{0x87F4912266CD898DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion2.ani"},
	{0x87F4DD9E9117837EULL, "vxdr_primevaltomb_57j_0_0.vlm"},
	{0x87F50A75B3254F26ULL, "zz_ave_debug6206685.cl"},
	{0x87F57DBB9A14F1B6ULL, "6142201_m.fxe"},
	{0x87F57FC00065B200ULL, "vxlt_gallowstemplar_f_5j2_0_0.vlm"},
	{0x87F5818A1EAEC60FULL, "art\\characters\\creatures\\corpse\\animation\\combat\\runforward.evt"},
	{0x87F59AA235A493E4ULL, "595241385.ogg"},
	{0x87F5C5D9BA37C71DULL, "vxlt_0_lowtown_docks_n_b7q_0_0.vlm"},
	{0x87F5F397E0B44997ULL, "vxca_cavern1_haunted_ckq_0_0.vlm"},
	{0x87F6218298102725ULL, "6120743_m.fxe"},
	{0x87F7184D536D7C68ULL, "hlac_lt_0_lowtown_-1_2y.dds"},
	{0x87F759E0DA78B76FULL, "zz_followers_banter6157575.cl"},
	{0x87F76798F52B5378ULL, "vxca_caverns1_in6_0_0.vlm"},
	{0x87F7AD2E66D936DBULL, "vxdr_primevalentrance_1ko_0_0.vlm"},
	{0x87F7AE343BBA29D4ULL, "vxlt_gallowscourtyard_n_74x_0_0.vlm"},
	{0x87F7E9F2428D9AEAULL, "256594999.ogg"},
	{0x87F81D523DDA70DFULL, "6082583_f.fxe"},
	{0x87F8315B1F028321ULL, "vxlt_undercityserialkiller_1se_0_0.vlm"},
	{0x87F93008676B4BE3ULL, "6108054_f.fxe"},
	{0x87F9987C42A93C67ULL, "6140178_m.fxe"},
	{0x87F9AC6F9F63EAA6ULL, "lt_0_lowtown_docks_a3_n_30_0.rml"},
	{0x87F9E77936825282ULL, "hlac_w_bonepitmining_a2_0_9y.dds"},
	{0x87F9FE0ED1DF0167ULL, "vxlt_0_lowtown_docks_a3_n_coi_0_0.vlm"},
	{0x87FB4E8D82AD2FC8ULL, "530273152.ogg"},
	{0x87FB54D69A73466EULL, "6047336_f.fxe"},
	{0x87FB7D3DDD18FD33ULL, "6119297_m.fxe"},
	{0x87FC4997F916836BULL, "vxht_mansion2_1wz_0_0.vlm"},
	{0x87FC828A1EB4E6B8ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\runforward.gad"},
	{0x87FC86DFC91D9FFBULL, "vxlt_0_lowtown_a2_byw_0_0.vlm"},
	{0x87FC8D0C7EE3F57AULL, "vxdr_slavers_239_0_0.vlm"},
	{0x87FD1E3FCB18D314ULL, "vxca_caverns2_982_0_1.vlm"},
	{0x87FD2E1603766738ULL, "vxdr_primevaltomb_6um_0_0.vlm"},
	{0x87FD7E9C060E4487ULL, "sst_ss_hurlock_1.gda"},
	{0x87FDFB227A7AE73FULL, "6119653_m.fxe"},
	{0x87FE5892CAB72995ULL, "491229967.ogg"},
	{0x87FE8DB0D9788170ULL, "llac_lt_backalley_n_1_6.dds"},
	{0x87FECB8CC837636CULL, "war_berserkmodal.dds"},
	{0x87FED5E6F50FE572ULL, "hlac_lt_0_lowtown_a2_n_c_131072_3y.dds"},
	{0x87FF5B63BCE75E2AULL, "hlac_lt_0_lowtown_a2_n_c_-131072_1x.dds"},
	{0x87FFABDDE8189849ULL, "fen110_fenris6117935.cl"},
	{0x8800D45DB459031DULL, "vxca_caverns1_mines_ckc_0_0.vlm"},
	{0x88016674A8CF2B8EULL, "hlac_ht_0_hightown_a3_0_2x.dds"},
	{0x88019008DAAE1730ULL, "vxlt_gallowstemplar_f_5iw_0_0.vlm"},
	{0x8801A2E29AC0BAB4ULL, "gen00flst_isa_act3.stg"},
	{0x8801A95CD7D6C619ULL, "zz_dae_debug6029483.cl"},
	{0x8801BE3283037ECCULL, "hlac_dr_0_deeproads_5843_1y.dds"},
	{0x8801CE4531E0946DULL, "369492288.ogg"},
	{0x880209E69B9F5AE5ULL, "6084680_f.fxe"},
	{0x8802434D54FD5437ULL, "w_bonepit_38_0.mmh"},
	{0x8802474D54FD5B44ULL, "w_bonepit_38_0.mao"},
	{0x88024AD4A6B66238ULL, "6053825_f.fxe"},
	{0x8802594D54FD79A9ULL, "w_bonepit_38_0.msh"},
	{0x88028B94BD04CE0FULL, "w_bonepitmining_a2_17_0_ll.mmh"},
	{0x88028C1834843B47ULL, "hlac_dr_slavers_3214_3z.dds"},
	{0x88037ADE630F7279ULL, "mrl150_merrill6070803.cl"},
	{0x88040AD273C4565DULL, "vxca_caverns1_icq_0_1.vlm"},
	{0x880435516F6956B7ULL, "follower_barks6153264.cl"},
	{0x88044E2CFFB39827ULL, "157451813.ogg"},
	{0x880623418EB63424ULL, "321244095.ogg"},
	{0x88064578F3FC5B18ULL, "vxca_caverns2_6hh_0_0.vlm"},
	{0x88066FB71A7E8DEFULL, "vxht_0_hightown_n_621_0_0.vlm"},
	{0x88069C43DA885118ULL, "378061090.ogg"},
	{0x8806E02435445087ULL, "6128363_m.fxe"},
	{0x8806F316D6C7AD50ULL, "6185156_m.fxe"},
	{0x8807BC411D5DD552ULL, "6064652_f.fxe"},
	{0x880819D66359C592ULL, "vxlt_0_lowtown_docks_n_b9d_0_0.vlm"},
	{0x880836C9ACA57E86ULL, "vxlt_warehouse_n_1c6_0_0.vlm"},
	{0x88087481CB70F3A6ULL, "zz_otl_debug6060947.cl"},
	{0x88088666EB7ABA7EULL, "6101356_m.fxe"},
	{0x8808BD727CDCBB14ULL, "6101075_m.fxe"},
	{0x880921597CAC2894ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkback.gad"},
	{0x8809D7C04BB5FF0CULL, "hlac_lt_0_lowtown_docks_-196611_1x.dds"},
	{0x880A230A8CA2C09DULL, "dae000ip_crowd_mf5_talk8.utp"},
	{0x880A5CAAAC0ECA72ULL, "vxht_chantry_2z7_0_0.vlm"},
	{0x880B3C108BFE7853ULL, "art\\characters\\weapons\\promo\\shields\\textures\\w_shd_011a_0d.dds"},
	{0x880B547C7DA145E4ULL, "dae200st_mother.stg"},
	{0x880B9FF2F56C7FB8ULL, "vxdr_primevaltomb_6xj_0_0.vlm"},
	{0x880BB9BD7A3039E3ULL, "vxlt_backalley_n_1x3_0_0.vlm"},
	{0x880BCEFE23A59025ULL, "vxlt_warehouse_n_cn_0_0.vlm"},
	{0x880BD75DDF62F6CFULL, "vxlt_gallowstemplar_n_5xc_0_0.vlm"},
	{0x880BD865188258BAULL, "6140631_f.fxe"},
	{0x880BD92F5AC75E27ULL, "lt_0_lowtown_docks_30_0_lc.mmh"},
	{0x880CEEF6B89D0E6AULL, "6118578_m.fxe"},
	{0x880CFCB9B870F445ULL, "hlac_ht_0_hightown_a2_1_1z.dds"},
	{0x880D1752693C4680ULL, "llac_ht_mansion_n_3365_1.dds"},
	{0x880D472F2E5C702BULL, "501900185.ogg"},
	{0x880D7CB425DF5DE9ULL, "hlac_w_sundermount_196604_0x.dds"},
	{0x880D8792492108B0ULL, "qun160_vashothlead6053928.cl"},
	{0x880DC0AFFD6B40B8ULL, "zz_perfect_circle.cut"},
	{0x880DD129F55F0D44ULL, "qun131_party_comment6117302.cl"},
	{0x880DDE942DA09188ULL, "6101057_m.fxe"},
	{0x880E1612960D7B43ULL, "gen00fl_aveline6140780.cl"},
	{0x880E9F93E498D788ULL, "775791893.ogg"},
	{0x880EC09196737A23ULL, "art\\characters\\creatures\\stonegolem\\animation\\exit.ani"},
	{0x880EDC07771C911BULL, "590782532.ogg"},
	{0x880F2D1DC5E89487ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffy_d.dds"},
	{0x880F8193679FD4D4ULL, "6117461_m.fxe"},
	{0x881001E5CCA42FB0ULL, "vxlt_gallowstemplar_f_5lt_0_0.vlm"},
	{0x8810157C619FA933ULL, "vxht_mansion2b_1c6_0_0.vlm"},
	{0x88101F9649783BCFULL, "llac_lt_0_lowtown_a2_n_65536_2.dds"},
	{0x881030597CB26172ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkback.ani"},
	{0x881032ADFCD888D5ULL, "hlac_dr_primevaltomb_665_0x.dds"},
	{0x8810A481D7998F47ULL, "6142284_m.fxe"},
	{0x881158459D14CA3EULL, "lgt220ar_gang_hideout_ht.lst"},
	{0x88115BA83C75A5A4ULL, "6210421_m.fxe"},
	{0x88115D113AAAC405ULL, "one100_party_comments6133506.cl"},
	{0x8812E6F563BF8D13ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.cess.ani"},
	{0x8812F37938EFF0FCULL, "hlac_lt_gallowscourtyard_n_-1_1x.dds"},
	{0x881350F11B273524ULL, "mer120_hubert6070212.cl"},
	{0x881355B332A6F9CFULL, "6088308_m.fxe"},
	{0x88138688617D0127ULL, "965355987.ogg"},
	{0x8813FBC129FFAD5CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robb_d.dds"},
	{0x88142B1677DA2C75ULL, "w_bonepitmining_a2_mmap.dds"},
	{0x88144395A763EA13ULL, "6151172_m.fxe"},
	{0x88149E30E193C14AULL, "vxlt_0_lowtown_a2_n_c_gcd_0_0.vlm"},
	{0x8814A24D82B26620ULL, "169297475.ogg"},
	{0x8814DC07CF867C14ULL, "hlac_lt_gallowscourtyard_n_-131072_0x.dds"},
	{0x8814E41345CD8FB1ULL, "zz_dre_debug6201310.cl"},
	{0x8814FD3B0D5BF3CBULL, "31685189.ogg"},
	{0x8815CE97BB3AFFD3ULL, "xx_extboards.dds"},
	{0x8815DF69170E15FBULL, "static_static_p31.psh.fxo.dx11.assembly"},
	{0x8815E22097ADAFE9ULL, "llac_lt_0_lowtown_docks_n_65533_2.dds"},
	{0x8815E9F13F071F56ULL, "vxlt_gamlenshouse_n_mm_0_0.vlm"},
	{0x8816515FA320A808ULL, "vxht_bartrand_n_5od_0_1.vlm"},
	{0x8816744598704F05ULL, "vxw_encounter3_df_0_0.vlm"},
	{0x881679FBFC577687ULL, "llac_lt_darktown_-1_16.dds"},
	{0x88168F83B4F79869ULL, "179056591.ogg"},
	{0x8816C2478F02A84BULL, "vxca_cavern1_haunted_cnj_0_0.vlm"},
	{0x8816CED19D4A85F4ULL, "44579759.ogg"},
	{0x8816E1C8B5CB3267ULL, "vxlt_0_lowtown_docks_a3_n_cuk_0_0.vlm"},
	{0x88172281A2C2C9B9ULL, "781668223.ogg"},
	{0x8817599DBA87A7D9ULL, "art\\vfx\\combat\\vfx_gib_corm.msh"},
	{0x8817639DBA87B8C7ULL, "art\\vfx\\combat\\vfx_gib_corm.mmh"},
	{0x8817691E59AF6CCAULL, "vfx_vfx_p7.vsh.fxo.dx11.assembly"},
	{0x88177013B423D026ULL, "dae000_hate_follower_react.cnv"},
	{0x8818274B82AD922AULL, "one100dg_kelder.cnv"},
	{0x88183EB54BB09595ULL, "hlac_lt_0_lowtown_docks_a3_-4_0y.dds"},
	{0x881880B95ED8659AULL, "zz_followers_banter6157316.cl"},
	{0x881881E9FA5B831EULL, "vxht_chantry_32n_0_0.vlm"},
	{0x8819453CE835A10AULL, "vxw_woundedcoast_1bl_0_0.vlm"},
	{0x8819AABCE6FBA654ULL, "ts_prog_diamond.gda"},
	{0x881A4247734A2862ULL, "6119410_m.fxe"},
	{0x881A59C26836A2F7ULL, "152288496.ogg"},
	{0x881A71D9C29515ACULL, "hlac_lt_0_lowtown_docks_n_65534_6x.dds"},
	{0x881A7213B42609E6ULL, "dae000_hate_follower_react.bnk"},
	{0x881AA90071899F4DULL, "6084896_m.fxe"},
	{0x881ABADA2B50DB2AULL, "6131346_m.fxe"},
	{0x881AF3801CC8AB49ULL, "vxca_caverns1_mines_jfe_0_0.vlm"},
	{0x881B0D61BD137343ULL, "llac_lt_gallowstemplar_2394_12.dds"},
	{0x881B60D3AA9326EAULL, "6117478_m.fxe"},
	{0x881B8847093E2273ULL, "vxlt_0_lowtown_a2_enx_0_1.vlm"},
	{0x881BAF7B38B9D64CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk2_n.dds"},
	{0x881BDE53351EA4BDULL, "hlac_lt_0_lowtown_docks_a3_n_-3_3z.dds"},
	{0x881C094B82B14852ULL, "one100dg_kelder.bnk"},
	{0x881C46C0084F9ECEULL, "w_blightlands_94_0.mmh"},
	{0x881C52C0084FB365ULL, "w_blightlands_94_0.mao"},
	{0x881C5A86A1ACCEA2ULL, "dae000_mother6156398.cl"},
	{0x881C60C0084FCABCULL, "w_blightlands_94_0.msh"},
	{0x881C71217C37DB1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_1.ani"},
	{0x881CA991E143056CULL, "gen_im_acc_rng_act3_07.uti"},
	{0x881CD891967FE01AULL, "art\\characters\\creatures\\stonegolem\\animation\\exit.evt"},
	{0x881D6B6752E55CB0ULL, "llac_w_woundedcoast_0_10.dds"},
	{0x881D8338B4DA5278ULL, "mrl000_marethari6129342.cl"},
	{0x881DC5C21BE6C844ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba01a.phy"},
	{0x881DF74047F40AABULL, "vxw_bonepitdestroyed_a3_19g_0_0.vlm"},
	{0x881E6E572264D185ULL, "vxlt_gallowstemplar_n_46r_0_0.vlm"},
	{0x881E913E6BA46D6EULL, "darktown_fem_fer_urch_sta6174803.cl"},
	{0x881EC5FA80A244BFULL, "pro000_friedrich6050198.cl"},
	{0x881EE787D93CCDC2ULL, "hlac_w_sundermount_262148_0x.dds"},
	{0x881EED17FB39C7CCULL, "6112083_f.fxe"},
	{0x881F057646448739ULL, "541497012.ogg"},
	{0x881F41A553BE9DC4ULL, "520715563.ogg"},
	{0x882040D5EC641E29ULL, "916407134.ogg"},
	{0x8820DEF563CBBD4AULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.cess.evt"},
	{0x8821798D4E0929CDULL, "gen00fl_isabela_init6076601.cl"},
	{0x88217B934B82E944ULL, "6027589_f.fxe"},
	{0x882256F5F4572EF4ULL, "llac_w_encounter2_camp_-131071_0.dds"},
	{0x88225758A9A52985ULL, "gen00fl_varric6199024.cl"},
	{0x8822CF5038AC6889ULL, "6092169_m.fxe"},
	{0x8822DFFF7BC0D1F1ULL, "633592588.ogg"},
	{0x88233847B6989CD6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_filler.msh"},
	{0x88234247B698ADD4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_filler.mmh"},
	{0x88234647B698B4E7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_filler.mao"},
	{0x8823BA0445246379ULL, "6048489_m.fxe"},
	{0x8823BB44D5AEBF57ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv2.gad"},
	{0x8823D7919685FD0DULL, "art\\characters\\creatures\\stonegolem\\animation\\exit.gad"},
	{0x8823F3B37CB7C9CEULL, "vxw_sundermount_35f_0_1.vlm"},
	{0x882413A37375FDC3ULL, "821074771.ogg"},
	{0x88246B2D3666B1F7ULL, "vxw_encounter2_9h_0_0.vlm"},
	{0x8824CA9C4F6DDFCAULL, "6149644_m.fxe"},
	{0x8824DBC9457F7EE7ULL, "hlac_w_encounter2_camp_131070_0y.dds"},
	{0x882503EF4A153544ULL, "hlac_lt_backalley_n_0_14z.dds"},
	{0x882527256DDCCA5FULL, "594551457.ogg"},
	{0x8825294A6A40B7C6ULL, "hlac_ca_caverns2_5802_47x.dds"},
	{0x8825A7727F4AE01EULL, "vxca_caverns1_j29_0_2.vlm"},
	{0x8825B06ABE624C0AULL, "574659187.ogg"},
	{0x8826398CBAB914D5ULL, "473286515.ogg"},
	{0x8826578E3BE7A57FULL, "15394234.ogg"},
	{0x8827395700E7CC77ULL, "349435342.ogg"},
	{0x8827B719DA2BB98AULL, "vxw_woundedcoast_1mm_0_0.vlm"},
	{0x8827BDF563D1A3BDULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.cess.gad"},
	{0x8827BF687B8EC13AULL, "hlac_w_encounter2_camp_262143_0z.dds"},
	{0x882842E899ABABB0ULL, "zz_seb_debug6183124.cl"},
	{0x882866AF8ADB0746ULL, "6295750.ogg"},
	{0x88289064AD166939ULL, "dr_dungeon_3215.rml"},
	{0x8829A24FBC2ED579ULL, "hlac_w_blightlands_-65538_0z.dds"},
	{0x8829B8C55D9CD609ULL, "321091959.ogg"},
	{0x8829D1275E9F6D3CULL, "6027634_m.fxe"},
	{0x882A24B0215DE213ULL, "6208169_f.fxe"},
	{0x882A3AAD90F35F9BULL, "l_fen200ar_ambush_exterior.wbd"},
	{0x882A89217C4441EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_1.evt"},
	{0x882AD044D5B501E4ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv2.evt"},
	{0x882B7BE43BD099F9ULL, "follower_banter6163194.cl"},
	{0x882BDD17CCAA2BF0ULL, "qun110_petrice6059151.cl"},
	{0x882C2B306C6F0E5FULL, "6103012_m.fxe"},
	{0x882C2F2C8A6F62EEULL, "w_bonepitdestroyed_a3_71_0.mmh"},
	{0x882C3B2C8A6F7685ULL, "w_bonepitdestroyed_a3_71_0.mao"},
	{0x882C492C8A6F8EDCULL, "w_bonepitdestroyed_a3_71_0.msh"},
	{0x882C620E19CAB02CULL, "61423037.ogg"},
	{0x882C6D808DAECFDAULL, "mcr310_argument6075822.cl"},
	{0x882D023C943D56DEULL, "qun221_petrice6083814.cl"},
	{0x882D1B27EF029A72ULL, "lowtown_male_alienage1_sta6162765.cl"},
	{0x882D5BF856225E5FULL, "gen00fl_anders6119257.cl"},
	{0x882D5E538E468999ULL, "hlac_ht_mansion2b_818_47x.dds"},
	{0x882DE4C63CB38BB8ULL, "6077796_m.fxe"},
	{0x882E114CAB90DA7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_around.gad"},
	{0x882E2153CF582CA2ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge10.gad"},
	{0x882E5394B340963CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbucket_0.pwk"},
	{0x882FB3BF94842508ULL, "vxht_0_hightown_a3_9zs_0_0.vlm"},
	{0x88300C1441673292ULL, "vxca_caverns1_mines_cot_0_0.vlm"},
	{0x88305F0430D342E0ULL, "vxw_encounter3_14_0_0.vlm"},
	{0x88306DC5A5DF425AULL, "6021234_m.fxe"},
	{0x88308A217C48AE98ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_forward_1.gad"},
	{0x8830EAF14A9852FBULL, "269482615.ogg"},
	{0x8830F80107587303ULL, "dae200_bran6094721.cl"},
	{0x883167D6738143A5ULL, "363632992.ogg"},
	{0x883170C073AE8685ULL, "vxht_mansion2_35l_0_0.vlm"},
	{0x883181EAC57FF24AULL, "vxlt_0_lowtown_a2_n_c_gyf_0_0.vlm"},
	{0x8831F0E813C75044ULL, "vxw_sundermount_36t_0_0.vlm"},
	{0x8831F7E9CD4D3837ULL, "985315628.ogg"},
	{0x8832F98F2442C009ULL, "hlac_lt_gallowstemplar_n_2383_10y.dds"},
	{0x88333C794163F84AULL, "6142744_m.fxe"},
	{0x88335D9D8AF0D07DULL, "vxw_encounter2_camp_34_0_0.vlm"},
	{0x88336F5CE47F8C9BULL, "hlac_dr_dungeon_3207_5z.dds"},
	{0x8833A9E11DF02951ULL, "6076035_m.fxe"},
	{0x8833DD44CAD2E9FEULL, "vxw_bonepit_mi_0_0.vlm"},
	{0x8834144CAB954A91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_around.evt"},
	{0x88342DB190C8489FULL, "lt_0_lowtown_docks_a3_251_0.msh"},
	{0x883433B190C85341ULL, "lt_0_lowtown_docks_a3_251_0.mmh"},
	{0x88343FB190C867E2ULL, "lt_0_lowtown_docks_a3_251_0.mao"},
	{0x8835D02F4ED9E71CULL, "6139325_f.fxe"},
	{0x883628F6CC21D20AULL, "vxw_woundedcoast_1hn_0_0.vlm"},
	{0x883635BF2BB20474ULL, "hlac_lt_gallowsprison_n_-1_10z.dds"},
	{0x8836C30D01180B86ULL, "llac_lt_0_lowtown_131070_1.dds"},
	{0x88371B336F7CD754ULL, "6087210_m.fxe"},
	{0x88372494F987D9A5ULL, "vxlt_warehouse_n_pk_0_0.vlm"},
	{0x88372DB15DF17481ULL, "gen00fl_merrill6116069.cl"},
	{0x883862A70E180B11ULL, "vxca_caverns2_936_0_1.vlm"},
	{0x8838773B59DB8194ULL, "6089538_f.fxe"},
	{0x883898C9F8DC1D31ULL, "dre000_varric6033309.cl"},
	{0x8838E844D5C167DDULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv2.ani"},
	{0x88398C7F22ABF099ULL, "gen00fl_isabela6089527.cl"},
	{0x8839D1EBF4A65BC7ULL, "mag370_meredith6037852.cl"},
	{0x883A8B15DF2D1625ULL, "vxdr_0_deeproads_cqn_0_1.vlm"},
	{0x883AA824527D1AC6ULL, "art\\vfx\\creatures\\fxc_rage_slam_c.impa.ani"},
	{0x883BED6897914CDCULL, "hlac_lt_0_lowtown_a2_n_1_1x.dds"},
	{0x883C444396BE1CABULL, "vxca_cavern1_haunted_ikc_0_0.vlm"},
	{0x883C64CB22DD900FULL, "dre000_vampire6029972.cl"},
	{0x883D345F68DB9536ULL, "mcr371_hostile_merrill6082156.cl"},
	{0x883D3670878661C2ULL, "ser221_gascard6122439.cl"},
	{0x883D946912012622ULL, "322522121.ogg"},
	{0x883E1152DD29C628ULL, "character_character_p102.psh.fxo.dx11.assembly"},
	{0x883E98C6B558CF9FULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpb.phy"},
	{0x883EA135116460D5ULL, "dae000cr_grand_cleric6156767.cl"},
	{0x883F5D44669AE577ULL, "6108048_m.fxe"},
	{0x883FBDD42066622FULL, "6132470_m.fxe"},
	{0x883FED5D1818FF62ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v3.gad"},
	{0x88400822A39BD441ULL, "mag311_keran6041480.cl"},
	{0x884030621A97A5E8ULL, "static1_punch_s1.psh.fxo"},
	{0x88407DC8066DE007ULL, "6175156_m.fxe"},
	{0x8840B6B3C119F8F8ULL, "vrc221_cassandra6040889.cl"},
	{0x8840EB75469084CCULL, "6095795_f.fxe"},
	{0x8841CED800183C39ULL, "gen_im_npc_wep_two_2haa.uti"},
	{0x8841D771F3872257ULL, "hlac_lt_0_lowtown_65534_2y.dds"},
	{0x88422C4CABA1B160ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sneaking_peek_around.ani"},
	{0x88424C53CF691E88ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge10.ani"},
	{0x8842C48356D8D777ULL, "vxw_sundermount_2b2_0_0.vlm"},
	{0x8843168CF4A76850ULL, "vxlt_hangedman_285_0_0.vlm"},
	{0x8843350654C2794DULL, "vxht_0_hightown_a2_b9q_0_0.vlm"},
	{0x88433ECE2046F3EDULL, "hlac_lt_undercityserialkiller_-1_27z.dds"},
	{0x88439C48A2BE43ABULL, "6103074_m.fxe"},
	{0x88443B371054DB30ULL, "gen_im_trs_crc_amu_r2.uti"},
	{0x88443E927F6D5CAEULL, "hlac_lt_0_lowtown_a2_n_131071_7z.dds"},
	{0x8845438C89FF27ADULL, "6095130_f.fxe"},
	{0x88459671EB7DF225ULL, "vxlt_warehouse_n_wj_0_0.vlm"},
	{0x8845B01EA84B509FULL, "gen00fl_sebastian6188636.cl"},
	{0x8845C876D658400EULL, "dae000_worthy6166634.cl"},
	{0x8845EBA728EC1121ULL, "6188598_m.fxe"},
	{0x8846105D181DA659ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v3.evt"},
	{0x884679AE6E343CA3ULL, "96578844.ogg"},
	{0x8846D74A465F398EULL, "275282844.ogg"},
	{0x884715C05CAD6668ULL, "186300972.ogg"},
	{0x884738526FA83E0BULL, "6060117_m.fxe"},
	{0x88473F1C1FB77DEEULL, "vxw_sundermount_2n3_0_0.vlm"},
	{0x88474424BB6A166DULL, "vxw_sundermount_2ay_0_0.vlm"},
	{0x884771A0CA67ADDEULL, "mag311cr_keran.utc"},
	{0x88479D4033E4C422ULL, "dae300_bran6143801.cl"},
	{0x8847A8C2CD6C3913ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_armripoff.evt"},
	{0x8847DCF4E88DA629ULL, "ser221ip_chest_clothes.utp"},
	{0x8848902452892F37ULL, "art\\vfx\\creatures\\fxc_rage_slam_c.impa.evt"},
	{0x8848AD15B0DA469FULL, "159185813.ogg"},
	{0x8848B271F3B690C7ULL, "fen110_fenris6118077.cl"},
	{0x8848BC033566D8B6ULL, "vxht_chantry_353_0_0.vlm"},
	{0x88499C2C9200A184ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_013a_0d.dds"},
	{0x8849F1FCE66BB8DBULL, "art\\toolset\\characters\\playercharacter\\elffemale\\mt_art_ef00.phy"},
	{0x884A18D2054B2475ULL, "vxlt_0_lowtown_a2_n_deh_0_0.vlm"},
	{0x884A1F5765B503EAULL, "art\\levels\\races\\proxy\\props\\_textures\\altar_d.dds"},
	{0x884AA56F2D8AC8FCULL, "vxw_bonepitmining_a2_1ns_0_0.vlm"},
	{0x884BA3C588F91060ULL, "vxlt_0_lowtown_a2_ev6_0_0.vlm"},
	{0x884BA605A760CC1DULL, "6129599_m.fxe"},
	{0x884BFBA7F73962CAULL, "6051563_m.fxe"},
	{0x884C44D9FCE357EDULL, "6120813_m.fxe"},
	{0x884C44E5C7C851D8ULL, "fen110_fenris6100284.cl"},
	{0x884CA940458CBF99ULL, "llac_lt_0_lowtown_docks_-65538_1.dds"},
	{0x884D3929A801108BULL, "cs_qcr000cs_arishoks_address.wbd"},
	{0x884DA1C2CD709954ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_armripoff.gad"},
	{0x884DA55D8044D1EAULL, "hlac_ht_0_hightown_a2_-65537_4x.dds"},
	{0x884DA600348FAE66ULL, "vxdr_0_deeproads_buq_0_0.vlm"},
	{0x884E7EC9ADCAD598ULL, "vxca_caverns2_6we_0_0.vlm"},
	{0x884EADE6CCC2A98DULL, "30680224.ogg"},
	{0x884ED4A1F7B38144ULL, "vxw_sundermount_38z_0_0.vlm"},
	{0x884F8BBD5A648D87ULL, "vxht_mansion2b_1gf_0_0.vlm"},
	{0x884F9124528F4FC0ULL, "art\\vfx\\creatures\\fxc_rage_slam_c.impa.gad"},
	{0x884F99C6B56744A4ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpb.mmh"},
	{0x884F9DC6B5674BF7ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpb.mao"},
	{0x884FAE68A86AF1FCULL, "llac_lt_hangedman_n_-1_19.dds"},
	{0x884FAFC6B5676A46ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpb.msh"},
	{0x884FE8199E39CF23ULL, "6127106_m.fxe"},
	{0x884FF753D3E89A2CULL, "6076210_m.fxe"},
	{0x8850B81501C754C5ULL, "662399390.ogg"},
	{0x8850DB75599A76F7ULL, "6128733_f.fxe"},
	{0x88512184FB8B3072ULL, "6172556_m.fxe"},
	{0x88516E4B31E6F413ULL, "gen_im_arm_hlm_hvy_set_med.uti"},
	{0x88518869E69D80D0ULL, "vxlt_hangedman_252_0_0.vlm"},
	{0x8851E17DF3F05476ULL, "6110866_m.fxe"},
	{0x88520E17B98A5EBEULL, "vxht_keep_n_69u_0_0.vlm"},
	{0x885217CEF9EFF730ULL, "350374631.ogg"},
	{0x885253E8202F57B8ULL, "6138754_m.fxe"},
	{0x88529451E55FFA3AULL, "isa321_velasco6117617.cl"},
	{0x8852B2B6EB5DB394ULL, "mcr370_messenger6071839.cl"},
	{0x88532398A2AD5E76ULL, "hlac_w_bonepitmining_a2_0_0z.dds"},
	{0x8853967889C1A9B5ULL, "6120844_m.fxe"},
	{0x8854084EDD740AA5ULL, "vxlt_warehouse_n_zi_0_0.vlm"},
	{0x8854185D1829F148ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v3.ani"},
	{0x885440FBE72BEF7AULL, "465492293.ogg"},
	{0x88549F68B6C3DC78ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_act_c.anb"},
	{0x8854B0C2CD76D132ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_armripoff.ani"},
	{0x8854D113DCB529E2ULL, "837350044.ogg"},
	{0x8854F5B6F8410806ULL, "vxlt_0_lowtown_n_f5e_0_0.vlm"},
	{0x8855B0F911AD966EULL, "vxw_sundermount_2s4_0_0.vlm"},
	{0x8855B601AD602EEDULL, "vxw_sundermount_2jx_0_0.vlm"},
	{0x885610C9E36AD27BULL, "art\\levels\\races\\vfx_models\\prp_brazloglrg.mmh"},
	{0x88561CC9E36AE618ULL, "art\\levels\\races\\vfx_models\\prp_brazloglrg.mao"},
	{0x88562EC9E36B0565ULL, "art\\levels\\races\\vfx_models\\prp_brazloglrg.msh"},
	{0x88565FAAA77C109BULL, "vxht_bartrand_n_5sf_0_0.vlm"},
	{0x8856C51A2B2FF49FULL, "dae000_hate_follower_react6103221.cl"},
	{0x8857651CE73D0930ULL, "567747884.ogg"},
	{0x8857A19DBB314FC7ULL, "mcr310_argument6072716.cl"},
	{0x8858A0CF114A9CE8ULL, "vxht_0_hightown_a3_7fy_0_0.vlm"},
	{0x8858DB308C1F3ADAULL, "lt_0_lowtown_docks_n_108_0.msh"},
	{0x8858E9308C1F5213ULL, "lt_0_lowtown_docks_n_108_0.mao"},
	{0x8858ED308C1F5920ULL, "lt_0_lowtown_docks_n_108_0.mmh"},
	{0x8858FB36FF78D9CFULL, "qun120_arishok_guard6049221.cl"},
	{0x88595FFB00B6A8C9ULL, "hlac_w_encounter2_196607_0x.dds"},
	{0x88597499D47F3744ULL, "mag291_vincento6027366.cl"},
	{0x885992062AD4F297ULL, "vxlt_0_lowtown_a2_n_c_fij_0_0.vlm"},
	{0x8859BF65A5D4B674ULL, "gen00fl_fenris6126414.cl"},
	{0x8859C6F4DC037997ULL, "dae000_gen_turn_in_46198678.cl"},
	{0x885A24425B4C4B72ULL, "vxlt_0_lowtown_a2_ghh_0_0.vlm"},
	{0x885A2B104C2CBF71ULL, "718647532.ogg"},
	{0x885A49E4AB6CFEFBULL, "10313894.ogg"},
	{0x885AC049C8760E4DULL, "vxlt_0_lowtown_docks_97m_0_0.vlm"},
	{0x885AD7084FB24BC4ULL, "6072328_m.fxe"},
	{0x885AE7EA18E2A965ULL, "228893782.ogg"},
	{0x885BFCBDE99AC513ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.impa.evt"},
	{0x885C40E3E92B75DCULL, "vxlt_gallowstemplar_f_5jn_0_0.vlm"},
	{0x885C84A392E5FEF8ULL, "vxca_caverns1_i31_0_0.vlm"},
	{0x885C911633D362BBULL, "6016556_m.fxe"},
	{0x885CE5165E414972ULL, "llac_ca_cave_2194_9.dds"},
	{0x885D0666E588F81DULL, "vxht_0_hightown_7fy_0_0.vlm"},
	{0x885D6E1313128FA2ULL, "lt_0_lowtown_docks_136_0.msh"},
	{0x885D701313129358ULL, "lt_0_lowtown_docks_136_0.mmh"},
	{0x885D7437A6F1274CULL, "763623670.ogg"},
	{0x885D7C131312A7FBULL, "lt_0_lowtown_docks_136_0.mao"},
	{0x885DB6219908C65AULL, "6145103_m.fxe"},
	{0x885E88CE75894B5EULL, "hlac_lt_0_lowtown_docks_a3_n_0_4x.dds"},
	{0x885E89974DF664A3ULL, "405884819.ogg"},
	{0x885E8E1C63BB1BC6ULL, "lt_0_lowtown_n_38_0_ll.mmh"},
	{0x885EE6D153178D02ULL, "434149384.ogg"},
	{0x885EFFFDACEA4922ULL, "dae000_mother.cnv"},
	{0x885F16B5C53FFEF1ULL, "vxdr_0_deeproads_erx_0_0.vlm"},
	{0x886027278CFD917DULL, "6149247_m.fxe"},
	{0x88605864167D43D5ULL, "one280cr_emrys.utc"},
	{0x88608B67FA12D18DULL, "6127092_m.fxe"},
	{0x886104AFA444C69AULL, "46816276.ogg"},
	{0x88612EF01D3745A1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockfxxl.phy"},
	{0x8861E1FDACEC4C5AULL, "dae000_mother.bnk"},
	{0x8861F5BDE99F2554ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.impa.gad"},
	{0x8862A0D5CEB6E1BAULL, "hlac_ca_cavern1_haunted_24422_2z.dds"},
	{0x8862E3B8DB51D944ULL, "vxlt_undercityencounter_109_0_0.vlm"},
	{0x886308D3A35AF5FFULL, "vxht_chantry_2io_0_0.vlm"},
	{0x8863C259980AE118ULL, "882362414.ogg"},
	{0x886422D603A3AEEEULL, "vxw_sundermount_2t5_0_0.vlm"},
	{0x886427DE9F56476DULL, "vxw_sundermount_2ow_0_0.vlm"},
	{0x88642D2C3D4F19A5ULL, "gen00fl_isabela_init6110251.cl"},
	{0x88659DF79507AFACULL, "6095870_f.fxe"},
	{0x8865E596323A3D11ULL, "character_character_p222.psh.fxo.dx11.assembly"},
	{0x8866229DB4077998ULL, "llac_ht_keep_n_4519_16.dds"},
	{0x88664B01105E63A8ULL, "ht_bartrand_n_3370_ll.mmh"},
	{0x8866C5F4E8A82FA0ULL, "ser221ip_chest_clothes.ncs"},
	{0x8866D371928AC5EAULL, "413546670.ogg"},
	{0x8866FADEE142C527ULL, "vxca_caverns1_edk_0_0.vlm"},
	{0x88675CD82FFDE204ULL, "vxlt_gallowstemplar_f_2xs_0_0.vlm"},
	{0x88680CD0906F6B7CULL, "823546581.ogg"},
	{0x886826239CB1A294ULL, "llac_ca_cavern1_haunted_24426_4.dds"},
	{0x88687713131CA6E3ULL, "lt_0_lowtown_docks_136_0.phy"},
	{0x8868AA54856C7CA2ULL, "550233732.ogg"},
	{0x886904BDE9A55D32ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.impa.ani"},
	{0x88693FB9F6615D23ULL, "vxlt_hangedman_n_10z_0_0.vlm"},
	{0x886A1D789DAF79F8ULL, "vxw_encounter3_9t_0_0.vlm"},
	{0x886AA50BCF149725ULL, "496322650.ogg"},
	{0x886B30C1255FE311ULL, "6031269_m.fxe"},
	{0x886B3BED41A972F6ULL, "vxlt_darktown_4ku_0_0.vlm"},
	{0x886B59982B8CCAA2ULL, "193188581.ogg"},
	{0x886B60625DD36AA8ULL, "vxht_0_hightown_a3_c5l_0_0.vlm"},
	{0x886B69C0FC7E3984ULL, "dae000ip_crowd_mf7_talk18.utp"},
	{0x886C11F439F62222ULL, "hlac_lt_0_lowtown_docks_n_131071_4x.dds"},
	{0x886C3632B2B9FF67ULL, "313241254.ogg"},
	{0x886C8750A5A13F25ULL, "53532441.ogg"},
	{0x886C9D10DC4B4628ULL, "250188019.ogg"},
	{0x886CB461EC293063ULL, "hlac_ht_0_hightown_a2_1_2y.dds"},
	{0x886CC1BBB6702BF4ULL, "287462885.ogg"},
	{0x886D68E524AEB380ULL, "demo_rockwraith_rogue.cut"},
	{0x886DCA8BB037C900ULL, "654439395.ogg"},
	{0x886E4F7B3713E0C8ULL, "gen00fl_dog6141102.cl"},
	{0x886E764F94CFE7A9ULL, "sst_ss_harvester_1.gda"},
	{0x886EB6E69A444C3FULL, "vxht_mansion_n_3i5_0_0.vlm"},
	{0x886ECA261FC8A82BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_kill.evt"},
	{0x886EE53CB76D530CULL, "zz_mrl_debug6185309.cl"},
	{0x886F0A7EF70C8AE1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.open2.ani"},
	{0x886FFA3DDB44D92CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic_1.mao"},
	{0x8870440C9489410BULL, "vxlt_foundry_n_yt_0_0.vlm"},
	{0x88706E802D24E548ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_7whitelyrium.msh"},
	{0x887080275961AC84ULL, "6016652_m.fxe"},
	{0x887080802D2503D9ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_7whitelyrium.mao"},
	{0x88708C802D251872ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_7whitelyrium.mmh"},
	{0x8870E4E6CD1562C3ULL, "vxht_brothel_n_6jz_0_0.vlm"},
	{0x8871603589879555ULL, "152197918.ogg"},
	{0x8871874A67FC3AF2ULL, "640864644.ogg"},
	{0x8871DA634980144DULL, "vxdr_primevaltomb_3hv_0_0.vlm"},
	{0x8871E36996ADAE00ULL, "mer140ip_item_rubble_2.utp"},
	{0x88721B29A81FFA59ULL, "cs_qcr000cs_arishoks_address.bnk"},
	{0x8872270A7E38B4BFULL, "6042975_m.fxe"},
	{0x887242C546D33654ULL, "790377924.ogg"},
	{0x887294B2F599C76EULL, "vxw_sundermount_2y6_0_0.vlm"},
	{0x887299BB914C5FEDULL, "vxw_sundermount_2pv_0_0.vlm"},
	{0x88736C6BFF305EE5ULL, "6053113_m.fxe"},
	{0x88739790BC86F46EULL, "hlac_lt_hangedman_n_-1_6z.dds"},
	{0x8873A0CE44ADD188ULL, "hlac_lt_hangedman_n_-1_21y.dds"},
	{0x8873CFDDF14D85B0ULL, "6069030_m.fxe"},
	{0x88741B7A5742B40CULL, "llac_lt_foundry_n_320_23.dds"},
	{0x88744966DE215751ULL, "6116980_m.fxe"},
	{0x8874DB5C25EE9BF0ULL, "vxht_0_hightown_a3_ajs_0_0.vlm"},
	{0x88751A21491EE609ULL, "lowtown_male_drunk_mob6151174.cl"},
	{0x88753C58EF4DB0B5ULL, "plt_prm000pt_ivo_amulet.nss"},
	{0x88754EEC0CD81BD5ULL, "vxlt_0_lowtown_a2_fhr_0_0.vlm"},
	{0x88758C763500E34FULL, "lt_0_lowtown_n_32_0_lc.mmh"},
	{0x8875D8FCE690C97AULL, "art\\toolset\\characters\\playercharacter\\elffemale\\mt_art_ef00.msh"},
	{0x8875E3261FCEF1ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_kill.gad"},
	{0x8875EAFCE690E7C0ULL, "art\\toolset\\characters\\playercharacter\\elffemale\\mt_art_ef00.mmh"},
	{0x88760CBCDE228FEBULL, "one100dg_nabil6176024.cl"},
	{0x8876157EF712BB83ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.open2.gad"},
	{0x88766BEC48A59414ULL, "art\\vfx\\textures\\vfx_entropic.dds"},
	{0x8876B23A62001F05ULL, "clf221_jalen6070364.cl"},
	{0x8876DED21517F651ULL, "479302640.ogg"},
	{0x88771997ABDE85F0ULL, "llac_ca_caverns2_5802_36.dds"},
	{0x88773B87F825DA22ULL, "lgt250_way_dalish_conflict6147164.cl"},
	{0x8877DB1FC1A53087ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_strt.gad"},
	{0x8877DF311CA42F70ULL, "hlac_lt_0_lowtown_n_131071_3z.dds"},
	{0x88780AB15D507750ULL, "6201377_m.fxe"},
	{0x88788F558FA59278ULL, "vxw_encounter3_2u_0_0.vlm"},
	{0x8878C8D15F48B855ULL, "vxht_keep_70c_0_0.vlm"},
	{0x8878D4C64A2B0815ULL, "37193851.ogg"},
	{0x8878F5F0D83D51ABULL, "vxlt_gallowscourtyard_7cu_0_0.vlm"},
	{0x88791673530B5701ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_kneeling.anb"},
	{0x887959A69427CDA6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_crnf.ani"},
	{0x8879A0DE429BA917ULL, "504175261.ogg"},
	{0x8879C664586AB315ULL, "931613275.ogg"},
	{0x8879ECDFB7016A29ULL, "hlac_lt_0_lowtown_1_1z.dds"},
	{0x887A06E58BD59906ULL, "character_rendertodepth_p0.psh.fxo.dx11.assembly"},
	{0x887A4321394BD087ULL, "110226652.ogg"},
	{0x887A6F84D927EC13ULL, "vxw_bonepitmining_a2_1xc_0_1.vlm"},
	{0x887A732E51412F41ULL, "339090465.ogg"},
	{0x887A875DFD77BBAEULL, "577718522.ogg"},
	{0x887AD6B13C2CD3B2ULL, "vxw_encounter2_po_0_1.vlm"},
	{0x887AE2BB7BA8A364ULL, "259781424.ogg"},
	{0x887B1457D87441E6ULL, "follower_banter6171273.cl"},
	{0x887B68E5ACF786EDULL, "art\\vfx\\textures\\vfx_soulloop.dds"},
	{0x887B8CB25B08D350ULL, "follower_banter6155703.cl"},
	{0x887BC7B6A3B6C8B4ULL, "vxca_caverns1_mines_gk2_0_0.vlm"},
	{0x887BF82E2A317526ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_exit.evt"},
	{0x887C408AC7EAD7C9ULL, "6036422_m.fxe"},
	{0x887C9E68B6E52C13ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_act_c.mmh"},
	{0x887CE2261FD50EEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_exit_kill.ani"},
	{0x887D0E0C95E7A66EULL, "6173960_m.fxe"},
	{0x887D1C2E65D5E009ULL, "6150670_m.fxe"},
	{0x887D28C38C3A64BFULL, "vxht_mansion_n_3f4_0_0.vlm"},
	{0x887D39EDC43E01FBULL, "568445670.ogg"},
	{0x887D741D743BFBFFULL, "867069159.ogg"},
	{0x887D7A5A052F9516ULL, "6218823_m.fxe"},
	{0x887DD81110F65532ULL, "219737218.ogg"},
	{0x887EEA9C16F8BA90ULL, "vxlt_backalley_g_1nv_0_1.vlm"},
	{0x887EF00D7BA6611CULL, "vxlt_0_lowtown_a2_n_c_eqn_0_0.vlm"},
	{0x887EF66F78C2730CULL, "vxlt_darktown_6ca_0_0.vlm"},
	{0x887F4147B6E70ACFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_filler.phy"},
	{0x887F6C94D6288DFCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1.ani"},
	{0x887F778D0D478CE5ULL, "qcr000_arishok6059292.cl"},
	{0x8880062CF71CF33EULL, "hlac_w_bonepitdestroyed_a3_393216_0x.dds"},
	{0x88803D0ADEFC35FDULL, "hlac_ca_caverns1_mines_24421_7y.dds"},
	{0x88805B13489309ACULL, "608859219.ogg"},
	{0x888086BA1DC143C9ULL, "vxht_keep_620_0_0.vlm"},
	{0x8880952F0A1D22CBULL, "mag131_keran6088225.cl"},
	{0x88813E930F89568DULL, "6137721_m.fxe"},
	{0x88816E72687DC9B0ULL, "vxlt_gallowstemplar_n_5ph_0_0.vlm"},
	{0x88820F2E2A360871ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_exit.gad"},
	{0x88827665DEBDCD87ULL, "vxlt_0_lowtown_a2_n_gnt_0_0.vlm"},
	{0x88828DAD4DB2BF51ULL, "zz_dre_debug6056972.cl"},
	{0x88833FC7E9B48FA4ULL, "vxca_caverns2_7gb_0_2.vlm"},
	{0x88834D3917E4B470ULL, "vxht_0_hightown_a3_awp_0_0.vlm"},
	{0x88838A099233490EULL, "w_bonepitmining_a2_10_0_ll.mmh"},
	{0x88839F0D03CBA638ULL, "6125921_m.fxe"},
	{0x8883D006802B7D00ULL, "vxlt_gallowstemplar_n_5t4_0_0.vlm"},
	{0x8883F1F77F687C14ULL, "80314738.ogg"},
	{0x88842C4DD6554205ULL, "6153507_m.fxe"},
	{0x88846CE2F5632AFCULL, "vxw_bonepitmining_a2_1zw_0_0.vlm"},
	{0x88847C69370D77BBULL, "hlac_ht_0_hightown_-65537_3z.dds"},
	{0x88849ADAE1CC6AACULL, "ts_prog_rose_diamond.gda"},
	{0x8884E72B7D013A53ULL, "16825441.ogg"},
	{0x8884F9AB911A0862ULL, "mag160_feynriel6016589.cl"},
	{0x88850C1C2D570CEBULL, "lowtown_male_lowlife1_sta6164973.cl"},
	{0x888641A694322F17ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_crnf.evt"},
	{0x88867994D62EC23EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1.gad"},
	{0x88867E57E5C25899ULL, "1036773242.ogg"},
	{0x8886971A2728A3ADULL, "hlac_lt_0_lowtown_65534_3x.dds"},
	{0x888752D705DE41A4ULL, "891150664.ogg"},
	{0x888767CDCA336A2BULL, "vxlt_gallowscourtyard_7dt_0_0.vlm"},
	{0x8887ADDF7EA35F40ULL, "vxlt_gallowstemplar_n_2ds_0_0.vlm"},
	{0x8887CA4CB8A2AF35ULL, "638396611.ogg"},
	{0x888805CFBB274B2BULL, "6199288_m.fxe"},
	{0x888821F1411493B5ULL, "llac_w_bonepit_327683_0.dds"},
	{0x88882EBD5338B9DDULL, "art\\characters\\playercharacter\\children\\km_bdy_boy.phy"},
	{0x8888671F78CCA95CULL, "dae000_sabina6162262.cl"},
	{0x888878643296A2FEULL, "160305861.ogg"},
	{0x8888886A335D5288ULL, "dae000_orana6149097.cl"},
	{0x8888E2F891BE4253ULL, "vxlt_gallowstemplar_f_4ua_0_0.vlm"},
	{0x8889102E2A3C281FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_exit.ani"},
	{0x88896F356F731E29ULL, "6085785_m.fxe"},
	{0x8889C4DEB081BC1BULL, "vxht_keep_4b0_0_0.vlm"},
	{0x888A0A8CE19288B8ULL, "vxlt_warehouse_n_10q_0_0.vlm"},
	{0x888A318D5924962AULL, "187225322.ogg"},
	{0x888B3B55A5C6E7C4ULL, "dae240ar_bone_pit.lst"},
	{0x888B704711C8E737ULL, "6078602_m.fxe"},
	{0x888B7FE987DDBEEEULL, "6176660_m.fxe"},
	{0x888C44F623CD611CULL, "hlac_dr_slavers_3207_2z.dds"},
	{0x888C63EF9AE55E10ULL, "738976417.ogg"},
	{0x888C92B87CC04336ULL, "gen_im_npc_arm_hum_cht_rdrc.uti"},
	{0x888CC81FC1B76CCDULL, "art\\characters\\creatures\\pridedemon\\animation\\d_strt.ani"},
	{0x888D0F77B4D1A301ULL, "fonttest.gfx"},
	{0x888D42A694384FA0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_crnf.gad"},
	{0x888D5494D634A25DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv1.evt"},
	{0x888D75B2B40D0D32ULL, "llac_lt_gallowstemplar_f_2393_0.dds"},
	{0x888E6379D65C07D9ULL, "268457870.ogg"},
	{0x888E7A27AB95C11AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_xenon_n.dds"},
	{0x888EBBDD501736C6ULL, "464171432.ogg"},
	{0x888EE76A674A219AULL, "214443217.ogg"},
	{0x888F641554DD9B2BULL, "vxdr_primevaltomb_85x_0_0.vlm"},
	{0x888F872D0CBCF394ULL, "306970845.ogg"},
	{0x888FA112324D9B4DULL, "6126751_f.fxe"},
	{0x888FABC03F0131DBULL, "dae000_gamlen6048495.cl"},
	{0x888FCB7D76315E93ULL, "ser200_aveline6201551.cl"},
	{0x88903A58B87D75D4ULL, "vxlt_gallowstemplar_n_4qf_0_0.vlm"},
	{0x8890C1075849DB90ULL, "hlac_lt_gallowscourtyard_131069_0z.dds"},
	{0x8890EB69A8C4B298ULL, "vxlt_gallowstemplar_3qv_0_0.vlm"},
	{0x889143C6B0AF435EULL, "6103586_f.fxe"},
	{0x889173CA074D7706ULL, "527936505.ogg"},
	{0x8891C3BF96D9A1AEULL, "plt_cod_cha_aveline1.plo"},
	{0x8891EF9E0F4587CBULL, "6148141_m.fxe"},
	{0x88920348C23FF043ULL, "vxw_sundermount_25k_0_0.vlm"},
	{0x88921211EB548EF2ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_1.dds"},
	{0x889223F02EA1B92AULL, "vxca_caverns2_72m_0_0.vlm"},
	{0x88923B55A5CD062DULL, "dae240ar_bone_pit.ncs"},
	{0x88925CB3AC549FD4ULL, "vxlt_gallowstemplar_f_45r_0_0.vlm"},
	{0x8892DEBFE759437CULL, "vxw_bonepitmining_a2_1wt_0_0.vlm"},
	{0x8892ED2DC42F5CACULL, "716288862.ogg"},
	{0x8893B96DEFEC71FCULL, "6180923_f.fxe"},
	{0x8893D416E7CB5FF8ULL, "vxlt_0_lowtown_a2_eqw_0_0.vlm"},
	{0x88940D2CDF4892D1ULL, "vxlt_gallowscourtyard_75q_0_0.vlm"},
	{0x8894BC71DCB9D2BFULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_act_c.impa.gad"},
	{0x88954B1EFDAB98D4ULL, "zz_dae_debug6049789.cl"},
	{0x8895642349873A31ULL, "8077249.ogg"},
	{0x88957AF5D87819C0ULL, "gen00fl_fenris6126825.cl"},
	{0x889595044EEB0050ULL, "vfx_distortionmask_p0.vsh.fxo.dx11"},
	{0x8895A6E3F75A276AULL, "6166449_m.fxe"},
	{0x88962ADA8FB5A24FULL, "vxht_keep_7ab_0_0.vlm"},
	{0x8896470E65A3CA74ULL, "6144340_m.fxe"},
	{0x8896C9C492730C8EULL, "dae000_bartrand6201356.cl"},
	{0x8896D744EE39D816ULL, "703088257.ogg"},
	{0x8896E0A4FADE616BULL, "vxht_bartrand_n_5t3_0_0.vlm"},
	{0x8896F88CA8CE4438ULL, "zz_vault_debug6203242.cl"},
	{0x88973B58EF6A9516ULL, "plt_prm000pt_ivo_amulet.plo"},
	{0x889783915F3CCEB4ULL, "6054382_m.fxe"},
	{0x88978F7A1230B283ULL, "hlac_lt_0_lowtown_a2_n_65538_2z.dds"},
	{0x889796D5333B07C2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_mountain_d.dds"},
	{0x8897A51E3FAC7531ULL, "zz_qun_debug6028257.cl"},
	{0x889836BBA277D49BULL, "vxht_keep_4g3_0_0.vlm"},
	{0x8898750A1A172B85ULL, "vxw_encounter3_9m_0_0.vlm"},
	{0x8898BE7A970F6458ULL, "189254083.ogg"},
	{0x88992FBD53472EE2ULL, "art\\characters\\playercharacter\\children\\km_bdy_boy.mmh"},
	{0x889931BD53473198ULL, "art\\characters\\playercharacter\\children\\km_bdy_boy.msh"},
	{0x88997ED2739393BFULL, "hlac_lt_0_lowtown_docks_n_196608_1y.dds"},
	{0x889AD19E2466622BULL, "llac_ca_caverns2_5802_48.dds"},
	{0x889B4AC2F24C4234ULL, "6084461_f.fxe"},
	{0x889BA9D39671C70DULL, "narrative_varric6172524.cl"},
	{0x889BB171DCBFDEACULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_act_c.impa.evt"},
	{0x889BF1487611FDD6ULL, "fxa_pow_mist.mao"},
	{0x889BFEDD6267E673ULL, "gen00fl_isabela6098206.cl"},
	{0x889C100B04C41166ULL, "art\\levels\\races\\proxy\\props\\_textures\\cotbedroll_sa.dds"},
	{0x889C39C6BC38B245ULL, "6048194_f.fxe"},
	{0x889C77EBCABE6F9EULL, "309828244.ogg"},
	{0x889CD392BC340AC9ULL, "one280_iwan6096882.cl"},
	{0x889D89AACD3CA576ULL, "969358417.ogg"},
	{0x889DBA8DB3BAE956ULL, "vxht_playershouse_n_1vb_0_0.vlm"},
	{0x889DDF4A7DF52ABAULL, "6020439_m.fxe"},
	{0x889DEFD8AFC68878ULL, "6049832_m.fxe"},
	{0x889E065C2FF461EFULL, "vxht_keep_6g9_0_0.vlm"},
	{0x889EAC35AA738E54ULL, "vxlt_gallowstemplar_n_4vg_0_0.vlm"},
	{0x889EC84494B56076ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edgebase.phy"},
	{0x889EC967C693E36BULL, "54029007.ogg"},
	{0x889F3FE554DF11B7ULL, "w_bonepitdestroyed_a3_28_0_ll.mmh"},
	{0x889F5BA72EA5B214ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_013a_0t.dds"},
	{0x889F83659A754F7BULL, "vxlt_0_lowtown_ghq_0_0.vlm"},
	{0x88A095CD2097D1AAULL, "vxca_caverns2_77n_0_0.vlm"},
	{0x88A0BC543CFFDC8AULL, "dae000_bartrand6204157.cl"},
	{0x88A0E3744623731DULL, "6152235_f.fxe"},
	{0x88A0EFD5246B8E0CULL, "691738446.ogg"},
	{0x88A18C026761EDC9ULL, "art\\characters\\creatures\\dragon\\c_dragon.phy"},
	{0x88A1CCD1140447EEULL, "art\\vfx\\combat\\vfx_str_g_pound_r_leg_c.impa.ani"},
	{0x88A1F469E6123B29ULL, "6165524_m.fxe"},
	{0x88A2428C27EF9E78ULL, "qun210cr_guard6072078.cl"},
	{0x88A245F3D9C17878ULL, "vxlt_0_lowtown_a2_ett_0_0.vlm"},
	{0x88A25A736D7B804EULL, "6033655_m.fxe"},
	{0x88A2B0CF44EE9653ULL, "142968816.ogg"},
	{0x88A31F5C45F0DA12ULL, "hlac_dr_0_deeproads_5840_3x.dds"},
	{0x88A3432B7FF82B5FULL, "6103203_m.fxe"},
	{0x88A3EFB7EC606728ULL, "mcr310_argument6075118.cl"},
	{0x88A43BB86FFD47E6ULL, "6039756_m.fxe"},
	{0x88A44E8070720BDEULL, "vxdr_primevaltomb_75f_0_0.vlm"},
	{0x88A54672C6722984ULL, "556864734.ogg"},
	{0x88A6437BC88389F9ULL, "83404479.ogg"},
	{0x88A700D0737548BBULL, "202032297.ogg"},
	{0x88A70F292710A4E1ULL, "follower_banter6160696.cl"},
	{0x88A79C2B6BA15192ULL, "vxlt_0_lowtown_fab_0_0.vlm"},
	{0x88A7C7DDB841F0A0ULL, "gen00fl_aveline6115772.cl"},
	{0x88A872095E8A92F9ULL, "6099735_m.fxe"},
	{0x88A8B14494BD8D17ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edgebase.msh"},
	{0x88A8C34494BDAC7AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edgebase.mao"},
	{0x88A8C74494BDB289ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww_edgebase.mmh"},
	{0x88A912C677F5FF55ULL, "775170730.ogg"},
	{0x88A91BCB7C923E29ULL, "hlac_ca_caverns2prc_5802_15y.dds"},
	{0x88A923E62800E487ULL, "6172247_m.fxe"},
	{0x88A988938BF8FC3AULL, "mrl150_dalish_guard6089413.cl"},
	{0x88A99971DCCBF305ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_act_c.impa.ani"},
	{0x88A9A3F2F1DE4EE6ULL, "11508056.ogg"},
	{0x88AA65F6FF009E03ULL, "6020459_f.fxe"},
	{0x88AC046AD7455937ULL, "vxlt_gallowsprison_n_6hx_0_0.vlm"},
	{0x88AC3B57EB84C2CAULL, "seb221_ruxton_harimann6187814.cl"},
	{0x88AC56863698C473ULL, "art\\characters\\playercharacter\\humanmale\\hlm_tmpa.mao"},
	{0x88AC86ECE0D7BF2DULL, "vxht_chantry_n_36n_0_0.vlm"},
	{0x88ACC4093E601330ULL, "vxlt_gallowstemplar_n_5aw_0_0.vlm"},
	{0x88ACF37E32C7F50AULL, "hlac_lt_0_lowtown_docks_a3_-196610_1z.dds"},
	{0x88AD0C1952DABEACULL, "vxht_keep_n_5dg_0_0.vlm"},
	{0x88ADC6C69C7824C0ULL, "982813344.ogg"},
	{0x88AE323D4296BBA8ULL, "440094126.ogg"},
	{0x88AE6535894DA45DULL, "clf260_harley6173870.cl"},
	{0x88AE75E917DD6B75ULL, "gen00fl_fenris6126798.cl"},
	{0x88AE8D750511D7ACULL, "87648919.ogg"},
	{0x88AEC4D1140EC43FULL, "art\\vfx\\combat\\vfx_str_g_pound_r_leg_c.impa.evt"},
	{0x88AF4A67345FE42FULL, "6077827_m.fxe"},
	{0x88AF9F49A3F74CC4ULL, "llac_lt_0_lowtown_a2_n_-65538_0.dds"},
	{0x88AFD6FC07483A5BULL, "6107302_m.fxe"},
	{0x88AFF6B75F66F781ULL, "647937922.ogg"},
	{0x88B0275038A9E1E2ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun.msh"},
	{0x88B0295038A9E498ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun.mmh"},
	{0x88B0355038A9F93BULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun.mao"},
	{0x88B047337FC263E7ULL, "llac_lt_0_lowtown_docks_n_65537_0.dds"},
	{0x88B14255A5E7C2B7ULL, "dae240ar_bone_pit.are"},
	{0x88B17C75E84AC746ULL, "65735680.ogg"},
	{0x88B18A28126BD1D2ULL, "scaleform_sf_depth_p0.vsh.fxo.dx11.assembly"},
	{0x88B2374D1F859C63ULL, "962209304.ogg"},
	{0x88B25D22409C4A2CULL, "6080255_m.fxe"},
	{0x88B2C33C380EBF81ULL, "rdr110dg_guard_warehouse6016698.cl"},
	{0x88B359ABB8131AB0ULL, "126364073.ogg"},
	{0x88B37597E7485F7EULL, "741468983.ogg"},
	{0x88B3CBD1E0E3CEB5ULL, "gen_im_npc_dwf_arm_glv_lega.uti"},
	{0x88B3FE3DA4350FC4ULL, "hlac_lt_0_lowtown_docks_-65535_0y.dds"},
	{0x88B4E16B0DC1EDBEULL, "hlac_w_blightlands_393215_0x.dds"},
	{0x88B557D0EC666BD6ULL, "llac_lt_0_lowtown_a3_n_c_-262144_1.dds"},
	{0x88B568B7B73AAE6CULL, "6121907_m.fxe"},
	{0x88B5A5D11414AE28ULL, "art\\vfx\\combat\\vfx_str_g_pound_r_leg_c.impa.gad"},
	{0x88B617E4253EBB26ULL, "lt_0_lowtown_a2_n_54_0.rml"},
	{0x88B67AE734078459ULL, "868011643.ogg"},
	{0x88B6A1A75A7D63C5ULL, "450170503.ogg"},
	{0x88B6D615D9B2E4BBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_edge_1.mao"},
	{0x88B725C407D971BAULL, "vxlt_0_lowtown_a2_n_c_eow_0_1.vlm"},
	{0x88B795C319F6FD07ULL, "6175347_m.fxe"},
	{0x88B7C6D002AB6E97ULL, "352479294.ogg"},
	{0x88B7E06A3C6E18A5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_water_d.dds"},
	{0x88B84C6CD391103FULL, "6208575_m.fxe"},
	{0x88B8D70D2DE57DEDULL, "hlac_w_encounter1_65541_0y.dds"},
	{0x88B9E954B93C8874ULL, "964813300.ogg"},
	{0x88BA40737CC1ED57ULL, "llac_dr_primevalentrance_1726_12.dds"},
	{0x88BABE4F79D0DCEDULL, "6125645_m.fxe"},
	{0x88BB7DF644D0D72CULL, "vxht_keep_n_5id_0_0.vlm"},
	{0x88BB81504AF3A813ULL, "6173136_m.fxe"},
	{0x88BBFB05E8D953B7ULL, "lt_0_lowtown_a2_n_c.arl"},
	{0x88BC146BFEA04B03ULL, "ts_kirkwall_acc_table.gda"},
	{0x88BC350BC517ADDFULL, "vxlt_undercityserialkiller_11q_0_0.vlm"},
	{0x88BC540B077658CDULL, "390147256.ogg"},
	{0x88BC673CC85BCD77ULL, "zz_followers_banter6157603.cl"},
	{0x88BC96F7B23E156FULL, "gen00fl_aveline6002882.cl"},
	{0x88BC9AACA51AD7B1ULL, "584514666.ogg"},
	{0x88BCDBBEAD87182BULL, "vxw_bonepitdestroyed_a3_1pp_0_0.vlm"},
	{0x88BD39668C3AAEF6ULL, "vxlt_warehouse_n_5s_0_0.vlm"},
	{0x88BE1A90823F2848ULL, "hlac_lt_0_lowtown_a2_-196609_1x.dds"},
	{0x88BEB021E19B166AULL, "dre000_dougal6131173.cl"},
	{0x88BEE46E549FA867ULL, "vxht_mansion2b_3dm_0_0.vlm"},
	{0x88BF29511E217F69ULL, "w_woundedcoast_166_0.phy"},
	{0x88BF70E868A5F5CCULL, "100744307.ogg"},
	{0x88BF7DCEEE02A07CULL, "hlac_w_sundermount_262145_0y.dds"},
	{0x88BFCB140DD9C756ULL, "153443680.ogg"},
	{0x88C0882D2FACE165ULL, "6146905_m.fxe"},
	{0x88C143FD86F3941EULL, "vxlt_undercityserialkiller_1d2_0_1.vlm"},
	{0x88C20F525C5C365AULL, "290528374.ogg"},
	{0x88C22BA864331185ULL, "vxlt_0_lowtown_a2_eqn_0_0.vlm"},
	{0x88C29DD3CDB191E8ULL, "6140924_m.fxe"},
	{0x88C2C10FA040A2D7ULL, "vxht_mansion2_26b_0_0.vlm"},
	{0x88C30D5152B40064ULL, "132921826.ogg"},
	{0x88C333F01D8A18DDULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockfxxl.mao"},
	{0x88C337F01D8A1F16ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockfxxl.mmh"},
	{0x88C341F01D8A3014ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockfxxl.msh"},
	{0x88C35D5C691FFCDFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh02.msh"},
	{0x88C3635C69200681ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh02.mmh"},
	{0x88C3F2C17BB516ADULL, "6214972_m.fxe"},
	{0x88C43DD7A0C2D84CULL, "vxht_mansion2_9n_0_0.vlm"},
	{0x88C47FAF30D6E55EULL, "6125853_m.fxe"},
	{0x88C4AC591802096DULL, "ss_qunari_m_1.bnk"},
	{0x88C4FB1DCA1B046CULL, "645402232.ogg"},
	{0x88C515A9692D7FCFULL, "640731664.ogg"},
	{0x88C547E744492837ULL, "vxht_mansion_n_4wx_0_0.vlm"},
	{0x88C562EB605F6E5AULL, "560794399.ogg"},
	{0x88C5C74F1D331761ULL, "hlac_lt_0_lowtown_docks_65534_0z.dds"},
	{0x88C637FD441D57E2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowleft.msh"},
	{0x88C639FD441D5A98ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowleft.mmh"},
	{0x88C645FD441D6F3BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowleft.mao"},
	{0x88C6460A7DBD9DB3ULL, "6027699_m.fxe"},
	{0x88C6539E832A2E26ULL, "zz_approval_debug6003779.cl"},
	{0x88C77B9E10435494ULL, "def_rock2_d.dds"},
	{0x88C7A413AAB41C35ULL, "6202683_m.fxe"},
	{0x88C7A6640DBFD530ULL, "vxdr_0_deeproads_ave_0_0.vlm"},
	{0x88C7B2E0A41547E8ULL, "vxht_mansion2_3e5_0_0.vlm"},
	{0x88C7D15214525F6CULL, "vxht_chantry_n_2ou_0_0.vlm"},
	{0x88C7D54C419739CCULL, "dae000_arishok6068466.cl"},
	{0x88C89B543882AB7CULL, "vxht_playershouse_n_1r4_0_0.vlm"},
	{0x88C8ADE2EE07EBDBULL, "vxlt_backalley_n_1kb_0_0.vlm"},
	{0x88C8B0AC66319594ULL, "vxht_0_hightown_b8g_0_0.vlm"},
	{0x88C94EDC83808D66ULL, "mag311_nyssa6042524.cl"},
	{0x88C9A978361141B7ULL, "129500411.ogg"},
	{0x88CB022CB293D0AFULL, "100123976.ogg"},
	{0x88CB03F9F8643A11ULL, "461360870.ogg"},
	{0x88CB402A2C8708DFULL, "6176131_m.fxe"},
	{0x88CB41D63B5E8FA6ULL, "gen00fl_fenris6127006.cl"},
	{0x88CB83BA5AD322C3ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibgolem_bloodfrenzy.ani"},
	{0x88CBAB437E30C776ULL, "vxlt_warehouse_n_8t_0_0.vlm"},
	{0x88CBEA5D9D94839CULL, "dae000_bartrand6027529.cl"},
	{0x88CC40B4159D61C1ULL, "6105815_f.fxe"},
	{0x88CC4D97579F89C0ULL, "884148414.ogg"},
	{0x88CCC5603650F66AULL, "6197715_m.fxe"},
	{0x88CD2736C93B2820ULL, "llac_lt_0_lowtown_-3_1.dds"},
	{0x88CD959F9D0576D9ULL, "dae000_templar_recruit_16164425.cl"},
	{0x88CE302C575AAEC8ULL, "dre000_cut_bartrandbetrayal6029367.cl"},
	{0x88CE75026788B4EEULL, "art\\characters\\creatures\\dragon\\c_dragon.mmh"},
	{0x88CE81026788C885ULL, "art\\characters\\creatures\\dragon\\c_dragon.mao"},
	{0x88CE8972C067BE6AULL, "dae000_orana6149086.cl"},
	{0x88CE8F026788E0DCULL, "art\\characters\\creatures\\dragon\\c_dragon.msh"},
	{0x88CE98EFAA2617C7ULL, "vxht_brothel_n_6dd_0_0.vlm"},
	{0x88CED33DAFED0BE7ULL, "6096277_m.fxe"},
	{0x88CEFA0A21A2F9E5ULL, "6143605_m.fxe"},
	{0x88CF3EC3A0342FF4ULL, "vxw_bonepitdestroyed_a3_266_0_0.vlm"},
	{0x88CF73B762E90FCFULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.dura.gad"},
	{0x88CF8BC0AE019420ULL, "vxlt_undercityserialkiller_1mv_0_0.vlm"},
	{0x88CFA22E123C28D5ULL, "lt_0_lowtown_a2_n_c_25_0_lc.mmh"},
	{0x88CFF3DED9B59B5FULL, "hlac_ca_caverns1_mines_24424_11y.dds"},
	{0x88D040FD4425BB23ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gatelowleft.phy"},
	{0x88D049BD3D99910EULL, "vxca_caverns1_mines_inq_0_0.vlm"},
	{0x88D05716E1B41173ULL, "vxht_0_hightown_n_9zf_0_0.vlm"},
	{0x88D05E5BA1FC163AULL, "vxdr_0_deeproads_ecl_0_0.vlm"},
	{0x88D082FE2D84B2FDULL, "gen_im_trs_prf_junk2_02.uti"},
	{0x88D084D04DFF2513ULL, "lgt150_nat_were_supporter6110986.cl"},
	{0x88D0963BA6E4974EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basic.phy"},
	{0x88D163A143392CA3ULL, "571475459.ogg"},
	{0x88D1ECECAB2673EBULL, "1017411505.ogg"},
	{0x88D2010C48333EAEULL, "6178341_m.fxe"},
	{0x88D27D0C0DEE2981ULL, "lt_0_lowtown_a2_n_26_0.anb"},
	{0x88D2E47952EBE81DULL, "limp_flare.mao"},
	{0x88D2FC34954FA649ULL, "gen00fl_aveline6122081.cl"},
	{0x88D3100EDC370B91ULL, "6174827_m.fxe"},
	{0x88D31ABD3D580B22ULL, "391039678.ogg"},
	{0x88D34A3A55845693ULL, "llac_lt_0_lowtown_n_1_6.dds"},
	{0x88D37932DBD32978ULL, "vxdr_primevalentrance_37z_0_0.vlm"},
	{0x88D4090AF0656F25ULL, "mag311_huon.bnk"},
	{0x88D459996D1C3D4DULL, "vxht_brothel_n_4zn_0_0.vlm"},
	{0x88D46C187B32510DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2.mmh"},
	{0x88D4D33EB27E21C8ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden3.mmh"},
	{0x88D4F13EB27E54F2ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden3.msh"},
	{0x88D4F367F9A44423ULL, "follower_banter6172969.cl"},
	{0x88D5410B6EA72BF4ULL, "vxlt_0_lowtown_a2_f54_0_0.vlm"},
	{0x88D57CC75E2EF780ULL, "character_character_p212.psh.fxo.dx11.assembly"},
	{0x88D5F557DAE26193ULL, "follower_barks6153440.cl"},
	{0x88D67D0C232C143FULL, "vxca_cavern1_haunted_ehj_0_0.vlm"},
	{0x88D688B762EF527CULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.dura.evt"},
	{0x88D6EB0AF067725DULL, "mag311_huon.cnv"},
	{0x88D73F22A086AE7DULL, "6149030_m.fxe"},
	{0x88D75013F0B9742BULL, "6072907_m.fxe"},
	{0x88D7781BD31E7930ULL, "dre000_varric6036950.cl"},
	{0x88D779B0573CDCBCULL, "dae000_unhappy_mage6165319.cl"},
	{0x88D829669D110B5DULL, "vxht_mansion2_25y_0_0.vlm"},
	{0x88D84BED995181F5ULL, "hlac_ht_0_hightown_-196609_0z.dds"},
	{0x88D85B8F187F0938ULL, "one100dg_elren6014275.cl"},
	{0x88D85F18BAA49AA9ULL, "hlac_w_sundermount_196603_0y.dds"},
	{0x88D8E4660A6E01CCULL, "hlac_lt_0_lowtown_-65535_0x.dds"},
	{0x88D8FF88C60DC6FCULL, "940273197.ogg"},
	{0x88D919563D9F00CBULL, "6102795_f.fxe"},
	{0x88D91C2096AA120BULL, "mag131tr_eau_taraneh.ncs"},
	{0x88D93EF88493A89AULL, "seb300im_ether_amulet.uti"},
	{0x88D99BBA5ADF88BAULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibgolem_bloodfrenzy.evt"},
	{0x88DAA3CEC034CB0FULL, "6088407_m.fxe"},
	{0x88DB257A0B57D938ULL, "1017966095.ogg"},
	{0x88DB4768FBF9C358ULL, "800864110.ogg"},
	{0x88DBB2F32886C925ULL, "6125906_m.fxe"},
	{0x88DC0D67A199DA78ULL, "vxlt_0_lowtown_a2_ehp_0_0.vlm"},
	{0x88DC1F4D76FB285CULL, "vxlt_gallowstemplar_n_5qk_0_0.vlm"},
	{0x88DC59023ADA1F73ULL, "6143446_f.fxe"},
	{0x88DC7747A9875B63ULL, "6102895_m.fxe"},
	{0x88DD376B144F6C5AULL, "2506298.ogg"},
	{0x88DDB0A0922A4874ULL, "vxw_bonepitdestroyed_a3_217_0_0.vlm"},
	{0x88DDB5BA1A1AC60AULL, "6040579_f.fxe"},
	{0x88DDBB0122C06CB3ULL, "6139429_f.fxe"},
	{0x88DDBE8FABB90E9AULL, "ico_itemsetbonus.dds"},
	{0x88DDD3E6E2860528ULL, "gen00fl_fenris6127134.cl"},
	{0x88DDDC049AF38639ULL, "z_cloud03.dds"},
	{0x88DDF3EF5DC992AEULL, "vxca_caverns1_i1q_0_0.vlm"},
	{0x88DE4A4E01442E0EULL, "dre000_dougal6131209.cl"},
	{0x88DE88429ED87FEDULL, "528950471.ogg"},
	{0x88DF2353EFAE4765ULL, "55858440.ogg"},
	{0x88DF9ABA5AE3F2ADULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibgolem_bloodfrenzy.gad"},
	{0x88E01728491F74D5ULL, "vxw_encounter1_ew_0_0.vlm"},
	{0x88E0A3F9BAA186D7ULL, "hlac_ca_caverns1_mines_24424_19y.dds"},
	{0x88E0B0105560F8FFULL, "6019158_warren_qunari_mad.cut"},
	{0x88E3300EF6D66926ULL, "227081385.ogg"},
	{0x88E390B762F9E995ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.dura.ani"},
	{0x88E401C75716E8ADULL, "hlac_w_blightlands_-65540_0y.dds"},
	{0x88E40829D9180272ULL, "vxlt_gallowsprison_n_6ct_0_0.vlm"},
	{0x88E483500CCE664FULL, "vxlt_gallowstemplar_n_55p_0_0.vlm"},
	{0x88E4E15C5E9F0FBAULL, "gen00fl_dog6117785.cl"},
	{0x88E51B06CA859992ULL, "gen00fl_isabela_init6077337.cl"},
	{0x88E565187B40B8CAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2.phy"},
	{0x88E5DD637D47E7DEULL, "6077813_m.fxe"},
	{0x88E609D1E063CA54ULL, "69170845.ogg"},
	{0x88E60BA9E4CD8FE6ULL, "gen00fl_sebastian6177888.cl"},
	{0x88E623DFF38F3C96ULL, "846244936.ogg"},
	{0x88E68A507359FB06ULL, "lgt110_gxa_vgk_conspirator16117068.cl"},
	{0x88E69A89EEE8924BULL, "nat380_temmerin6067398.cl"},
	{0x88E6F50C351AA1E4ULL, "vxw_woundedcoast_5v_0_0.vlm"},
	{0x88E781F685B3E0C0ULL, "vxlt_0_lowtown_n_eqt_0_0.vlm"},
	{0x88E7C9584B94DFFBULL, "gen00fl_anders6125162.cl"},
	{0x88E7EC6CA789EF93ULL, "6135992_m.fxe"},
	{0x88E7F2ABFBB79E2AULL, "gen00fl_aveline6139736.cl"},
	{0x88E7FA48FB36D9D2ULL, "6114094_f.fxe"},
	{0x88E809CE8CB29F70ULL, "art\\levels\\races\\proxy\\props\\_textures\\breadoven_sa.dds"},
	{0x88E81F2B5D541D24ULL, "890027848.ogg"},
	{0x88E83155836961ABULL, "vxw_bonepitdestroyed_a3_1cu_0_0.vlm"},
	{0x88E8AEE773FA15E7ULL, "one000_ilen.bnk"},
	{0x88E8B836D31AEBC2ULL, "6176471_m.fxe"},
	{0x88E9F39903907925ULL, "gen00fl_fenris6115349.cl"},
	{0x88EAEEED0A8C7E39ULL, "6097833_f.fxe"},
	{0x88EB12511E46928EULL, "w_woundedcoast_166_0.mmh"},
	{0x88EB1DC1942DBC52ULL, "1001240260.ogg"},
	{0x88EB1E511E46A725ULL, "w_woundedcoast_166_0.mao"},
	{0x88EB2C511E46BF7CULL, "w_woundedcoast_166_0.msh"},
	{0x88EB6AEC9191EA75ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_board_1.mao"},
	{0x88EB7CA98E1248C7ULL, "vxht_brothel_n_6jf_0_0.vlm"},
	{0x88EBD0E773FC855FULL, "one000_ilen.cnv"},
	{0x88EC83E088A3CFB9ULL, "ser221_quentin6129054.cl"},
	{0x88ED2BFE55BE8D46ULL, "vxca_caverns2_7mp_0_0.vlm"},
	{0x88ED68CB7CD782DEULL, "6132655_m.fxe"},
	{0x88EDBA77A3A98A0DULL, "6173848_m.fxe"},
	{0x88EE48BC28EA842CULL, "49665044.ogg"},
	{0x88EE4CE5A4368C92ULL, "hlac_ht_0_hightown_a3_-65537_4y.dds"},
	{0x88EE89053B158D55ULL, "vxw_encounter1_jv_0_0.vlm"},
	{0x88EF29A49D0E8FCEULL, "vxlt_0_lowtown_a2_gnb_0_0.vlm"},
	{0x88EF2A8312FD9245ULL, "721121331.ogg"},
	{0x88EF645C694543FEULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh02.phy"},
	{0x88EFD38A2D02B98BULL, "476487265.ogg"},
	{0x88EFD7032E20B878ULL, "qun110_petrice6042270.cl"},
	{0x88F02B740787109DULL, "vxht_0_hightown_avo_0_0.vlm"},
	{0x88F08939920F1762ULL, "840593243.ogg"},
	{0x88F08E4B908FCDA0ULL, "6103370_m.fxe"},
	{0x88F0A25918273297ULL, "ss_qunari_m_1.wbd"},
	{0x88F0F97A0DFFD1A1ULL, "vxca_caverns1_mines_giq_0_0.vlm"},
	{0x88F11DAD3D168C8CULL, "llac_ca_caverns1_24424_1.dds"},
	{0x88F174D5D92BCF29ULL, "gen00fl_merrill6150676.cl"},
	{0x88F1BBE0910C0CD0ULL, "dae000_hate_enemies6167536.cl"},
	{0x88F1F12F15A1108AULL, "vxw_woundedcoast_17c_0_0.vlm"},
	{0x88F227E286B328A2ULL, "vxca_cavern1_haunted_gkd_0_0.vlm"},
	{0x88F239DB20563B01ULL, "vxlt_gallowstemplar_f_527_0_0.vlm"},
	{0x88F27388940317BFULL, "268269040.ogg"},
	{0x88F27A06CB0E1AF2ULL, "vxlt_gallowsprison_n_6lu_0_0.vlm"},
	{0x88F3A0DC30E3B3ACULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dsla_0s.dds"},
	{0x88F3CA2E72C975BAULL, "w_encounter1_33_0.phy"},
	{0x88F3D28343A9170AULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxa.mao"},
	{0x88F3D68343A91E59ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxa.mmh"},
	{0x88F3E08343A92F47ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxa.msh"},
	{0x88F3E82735F710EAULL, "6043645_m.fxe"},
	{0x88F42EEFEBADFC77ULL, "ht_0_hightown_a2_18_0.rml"},
	{0x88F45947882D6AE1ULL, "splatter_01.mao"},
	{0x88F57A4E44E7F338ULL, "234204272.ogg"},
	{0x88F582E956D205ADULL, "w_woundedcoast_water_111.mmh"},
	{0x88F586E956D20CFEULL, "w_woundedcoast_water_111.mao"},
	{0x88F58F987CC6FD31ULL, "6040544_m.fxe"},
	{0x88F594E956D22433ULL, "w_woundedcoast_water_111.msh"},
	{0x88F5E805E90AABB5ULL, "lt_0_lowtown_a2_n_c.rml"},
	{0x88F61BF759D6AEC8ULL, "129640563.ogg"},
	{0x88F63EC4BA2422FAULL, "6039778_m.fxe"},
	{0x88F6F7F131367439ULL, "201665746.ogg"},
	{0x88F71D564DE22CFEULL, "6133537_m.fxe"},
	{0x88F76260DF46CF5CULL, "6113847_m.fxe"},
	{0x88F8044CE21C1B4BULL, "6050194_m.fxe"},
	{0x88F8070D757FE937ULL, "useable_states.gda"},
	{0x88F8212096C4CBF2ULL, "mag131tr_eau_taraneh.utt"},
	{0x88F8290B30D9FA82ULL, "hlac_lt_hangedman_n_-1_28x.dds"},
	{0x88F903075AE7595CULL, "vxlt_gallowstemplar_n_5oi_0_0.vlm"},
	{0x88F9038E88AF95C2ULL, "243981652.ogg"},
	{0x88F91A19E9435FDFULL, "6013270_m.fxe"},
	{0x88F9EDD4CFF97C08ULL, "vxht_0_hightown_a3_90e_0_0.vlm"},
	{0x88FA7761782E20F7ULL, "6197199_f.fxe"},
	{0x88FB130E1897590CULL, "hlac_ca_caverns2prc_5802_30y.dds"},
	{0x88FB1590D7797D3EULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.mmh"},
	{0x88FB5E3CD6FC9CA3ULL, "dialoguepop.gfx"},
	{0x88FBAEE419F988BEULL, "6084442_m.fxe"},
	{0x88FC4259B58B2FB0ULL, "34184927.ogg"},
	{0x88FCD23E3162CF16ULL, "775183120.ogg"},
	{0x88FD31C3524CF017ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hgrass2_da.dds"},
	{0x88FDB542352B517DULL, "mcr371_hostile_aveline6082185.cl"},
	{0x88FE3D301823EFA2ULL, "342573327.ogg"},
	{0x88FF57440616B2DFULL, "hlac_lt_darktown_-1_49x.dds"},
	{0x88FF5C1620DB5156ULL, "art\\vfx\\textures\\vfx_d_sphere.dds"},
	{0x88FFC01E0198F8C0ULL, "hlac_dr_0_deeproads_5849_2z.dds"},
	{0x88FFCE424D36B6E2ULL, "dae000cr_templar_knight.utc"},
	{0x8900DB6590B00B52ULL, "275390602.ogg"},
	{0x89012247E953C3EDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_httree.msh"},
	{0x89013047E953DB20ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_httree.mao"},
	{0x89013447E953E273ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_httree.mmh"},
	{0x8901651B66358ADFULL, "fen110_fenris6067195.cl"},
	{0x890218B1E0DF36AAULL, "160337860.ogg"},
	{0x89027951167FBDEFULL, "hlac_w_bonepitmining_a2_0_12z.dds"},
	{0x89028583EE1A7C8BULL, "331354008.ogg"},
	{0x8902B96F91C2EC31ULL, "llac_dr_primevalentrance_1725_2.dds"},
	{0x8903638A5D4D8327ULL, "vxht_brothel_n_757_0_1.vlm"},
	{0x890397967EC32DECULL, "6150518_m.fxe"},
	{0x890397FE293E423AULL, "vxht_chantry_335_0_0.vlm"},
	{0x8903A99DAC00FFE6ULL, "43232875.ogg"},
	{0x8904263FA57D09E0ULL, "hlac_lt_0_lowtown_a2_n_65536_5z.dds"},
	{0x8904B0254D13E792ULL, "520235999.ogg"},
	{0x8904D5B1F2094723ULL, "mag370_meredith6154382.cl"},
	{0x8904D964959BD90FULL, "mag100st_karras.stg"},
	{0x8904EDC2F26AC52CULL, "vxw_encounter2_camp_z_0_0.vlm"},
	{0x8905949E2E20BF91ULL, "mag100_cullen6086362.cl"},
	{0x89059BDD88650026ULL, "hlac_lt_0_lowtown_a2_n_c_0_8x.dds"},
	{0x89063EF9EB9C3013ULL, "6018679_m.fxe"},
	{0x89067CFD3C1E5C75ULL, "mer210_coterie_contact6072495.cl"},
	{0x89068A5E7C8FDD29ULL, "vxlt_blackemporium_rt_0_0.vlm"},
	{0x8906C84314CAA2EAULL, "vxht_mansion2b_2bg_0_0.vlm"},
	{0x8906D431BF4582C3ULL, "523094907.ogg"},
	{0x890724DAC1F853E0ULL, "hf_cgn_001.mop"},
	{0x890724DAC1F853E2ULL, "hf_cgn_001.mor"},
	{0x89081932306545D7ULL, "ht_mansion_n_3371_lc.mmh"},
	{0x89082C0CAACF6173ULL, "hlac_w_encounter2_196607_0z.dds"},
	{0x89084547D6500464ULL, "vxca_caverns2_4x7_0_1.vlm"},
	{0x8908803469FF7C5CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bedroll_sa.dds"},
	{0x8908F490D78582FDULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.anb"},
	{0x8909013A2EF3854EULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_poundground.evt"},
	{0x890912A2AD9221ACULL, "vxw_bonepit_29w_0_0.vlm"},
	{0x8909B02A412572D4ULL, "hlac_dr_primevaltomb_658_10x.dds"},
	{0x890B5F1A628A9742ULL, "414393857.ogg"},
	{0x890BAB2B8121429BULL, "811178823.ogg"},
	{0x890C30BC75317AC1ULL, "mrl150_dalish_guard6068422.cl"},
	{0x890C3A059C14675BULL, "121763047.ogg"},
	{0x890C52548D2EEF83ULL, "ss_spellcast.cut"},
	{0x890C5E0C0E1F6D11ULL, "lt_0_lowtown_a2_n_26_0.rml"},
	{0x890C68AD86BFB604ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock1.gad"},
	{0x890C9B0B833CD647ULL, "vxw_sundermount_25w_0_0.vlm"},
	{0x890CAF3574AA1CBAULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb102a.mmh"},
	{0x890CC13574AA3B00ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb102a.msh"},
	{0x890D02670B1EE0BBULL, "6075400_f.fxe"},
	{0x890D53F3C40B2045ULL, "dae000_bartrand6206860.cl"},
	{0x890EB784D9C48ACFULL, "mag291ip_barrel_small.utp"},
	{0x890EC4C5C5244A15ULL, "vxlt_0_lowtown_n_ehz_0_0.vlm"},
	{0x890F7579DEF4C543ULL, "597453420.ogg"},
	{0x890FF83A2EF994F9ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_poundground.gad"},
	{0x89102B048F3DB47DULL, "dre000_vampire6167463.cl"},
	{0x8910953BA71A7911ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basic.mmh"},
	{0x8910A13BA71A8DB2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basic.mao"},
	{0x8910AF3BA71AA5CFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wo_basic.msh"},
	{0x8911E77B41923834ULL, "vxw_bonepitmining_a2_os_0_0.vlm"},
	{0x8911EDA9CAFB1DE5ULL, "260890295.ogg"},
	{0x8912C2C9D38B03AAULL, "ht_0_hightownexplosion_n_35_0_ll.mmh"},
	{0x891304E6686666AFULL, "vxlt_gallowstemplar_f_6s5_0_0.vlm"},
	{0x891377AD86C5EEE2ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_hurlock1.ani"},
	{0x89138FA06F147F62ULL, "88132958.ogg"},
	{0x8914040D43E6CCC3ULL, "vxw_sundermount_2fr_0_0.vlm"},
	{0x8914A16DFEEDEA2AULL, "6058937_m.fxe"},
	{0x8914EFB78E723559ULL, "6149593_m.fxe"},
	{0x8914FB6478908FF7ULL, "6133864_m.fxe"},
	{0x89152631636776F0ULL, "6053927_m.fxe"},
	{0x8915B0B7CE00DC47ULL, "lgt160cr_ntb_elf_werewolf.utc"},
	{0x8915E1C2D152FAB7ULL, "gen00fl_varric6093683.cl"},
	{0x8916093A2EFE1D47ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_poundground.ani"},
	{0x8916423DC25681B9ULL, "6195888_m.fxe"},
	{0x8916D467ED25C27FULL, "static_static_p53.psh.fxo.dx11"},
	{0x8916FB887B1ADB0FULL, "219089174.ogg"},
	{0x89171ECD0D94E0CFULL, "6068104_m.fxe"},
	{0x891745F11D3F5555ULL, "hlac_w_sundermount_-327686_0y.dds"},
	{0x89181D7038249F29ULL, "character_character_p213.psh.fxo.dx11.assembly"},
	{0x8918471C924B9997ULL, "vxlt_0_lowtown_a2_n_c_ey1_0_0.vlm"},
	{0x89188CEA0149C028ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2lrg.msh"},
	{0x89188D4CF9BF13A6ULL, "lgt000ip_show_world_map.utp"},
	{0x89189EEA0149DEB9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2lrg.mao"},
	{0x8918A769D0057687ULL, "hlac_ca_caverns1_24424_0z.dds"},
	{0x8918AAEA0149F352ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2lrg.mmh"},
	{0x8918D3244CC26EFBULL, "vxca_cavern1_haunted_j2c_0_0.vlm"},
	{0x89190A4F73C69D1FULL, "290305215.ogg"},
	{0x8919DE093163793CULL, "vxlt_0_lowtown_a2_n_c_gbq_0_0.vlm"},
	{0x891A2F602A080272ULL, "vxht_mansion_n_4vv_0_0.vlm"},
	{0x891A6C80D45CBE96ULL, "art\\vfx\\environment\\vfx_hnd_r_pickaxe_c.anb"},
	{0x891B3B3975720402ULL, "art\\characters\\creatures\\dragon\\animation\\dance.gad"},
	{0x891B7F40AA1BBE29ULL, "262854002.ogg"},
	{0x891C777389477EE8ULL, "6119496_m.fxe"},
	{0x891EC78B4C98A7BBULL, "one111ip_ggt_crate_3.utp"},
	{0x891EDB17B4B40FB7ULL, "fxa_deacti_blink.mao"},
	{0x891EEF8D4DA66B82ULL, "779476974.ogg"},
	{0x891EF7BD206A4506ULL, "vxlt_0_lowtown_glt_0_0.vlm"},
	{0x891F328E2AC8BBCBULL, "776321201.ogg"},
	{0x891F6B09BE013732ULL, "qun120_arishok6053168.cl"},
	{0x891F6CA6D5147658ULL, "hlac_lt_foundry_n_320_21x.dds"},
	{0x891FA455E7DB1304ULL, "vxlt_0_lowtown_a2_n_c_eoy_0_1.vlm"},
	{0x891FC173B7141C24ULL, "vxht_brothel_n_272_0_0.vlm"},
	{0x89205958338850B4ULL, "vxw_bonepitmining_a2_jp_0_0.vlm"},
	{0x8920E809369380D2ULL, "hlac_ht_0_hightown_a2_65536_2z.dds"},
	{0x8920F556DB262103ULL, "6160918_m.fxe"},
	{0x892179BBE73C6555ULL, "6076000_m.fxe"},
	{0x89225E3975785EF9ULL, "art\\characters\\creatures\\dragon\\animation\\dance.evt"},
	{0x892275EA35DCE543ULL, "vxw_sundermount_2ku_0_0.vlm"},
	{0x89233E93CD27B0E3ULL, "qun131_taskmaster6177188.cl"},
	{0x8923619480684DD9ULL, "6142493_m.fxe"},
	{0x892379168E01CF7DULL, "vxca_caverns2prc_8mm_0_0.vlm"},
	{0x8923CFA6E7D1F138ULL, "llac_lt_0_lowtown_a2_n_-65536_6.dds"},
	{0x8923FF1A165D309AULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_coneweb.mmh"},
	{0x8924031A165D37D1ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_coneweb.mao"},
	{0x8924111A165D4FE0ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_coneweb.msh"},
	{0x89254BFB32EBDDF8ULL, "vxw_sundermount_37f_0_1.vlm"},
	{0x892563A1EE43082DULL, "hlac_ca_caverns1_mines_24424_18x.dds"},
	{0x89263752C2C7E423ULL, "art\\vfx\\combat\\vfx_w_2nd_charge_c.anb"},
	{0x89263A121E30CA62ULL, "601684335.ogg"},
	{0x8926DCB67B6088CDULL, "gen_im_trs_crc_sta_phy_r0.uti"},
	{0x892794179900BE43ULL, "6064940_m.fxe"},
	{0x8927B3B43C5F2866ULL, "vxlt_blackemporium_8i_0_0.vlm"},
	{0x89282CF197286EACULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stackshelf_sa.dds"},
	{0x89286F80D4690101ULL, "art\\vfx\\environment\\vfx_hnd_r_pickaxe_c.mmh"},
	{0x8928970B05E41940ULL, "vxw_bonepitmining_a2_p6_0_0.vlm"},
	{0x89289E266AA08E00ULL, "dae000_arishok6068541.cl"},
	{0x8928AA0466D806C3ULL, "hlac_ca_cavern1_haunted_24423_4z.dds"},
	{0x89297D014336AE04ULL, "art\\vfx\\combat\\vfx_c_rw_aoeburst_c.mmh"},
	{0x89298634E1B72B5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\blendtree\\trooper.anb"},
	{0x892A708BC4AA0AB5ULL, "mcr310_argument6182291.cl"},
	{0x892A71A733BE5845ULL, "vxlt_darktown_2w4_0_0.vlm"},
	{0x892B2B3B8C57B212ULL, "vxw_encounter3_u_0_0.vlm"},
	{0x892BD29610CCA43BULL, "gen00fl_varric6118850.cl"},
	{0x892BDF606D8BC021ULL, "713426433.ogg"},
	{0x892C191B92E20ADDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowflat.phy"},
	{0x892C2A1D0B3B18C6ULL, "503686611.ogg"},
	{0x892C51630BD83B71ULL, "hlac_ht_0_hightown_131071_2y.dds"},
	{0x892CB8D9AC714A65ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_mera_0s.dds"},
	{0x892CC83C035A2BC8ULL, "w_bonepitmining_a2_30_0_ll.mmh"},
	{0x892DA65F8E799712ULL, "fen110_anso6095009.cl"},
	{0x892E639F563966DBULL, "55177486.ogg"},
	{0x892E8461107E97CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_scared.evt"},
	{0x892F184064F5EF59ULL, "gen00fl_isabela6097605.cl"},
	{0x892F947BAD03D447ULL, "6137648_m.fxe"},
	{0x8930002E788AF77AULL, "hlac_ht_keep_n_3736_1z.dds"},
	{0x893066397584A9E8ULL, "art\\characters\\creatures\\dragon\\animation\\dance.ani"},
	{0x8930E7C727D2FDC3ULL, "vxw_sundermount_2lt_0_0.vlm"},
	{0x8930F7D728D5B52AULL, "mer210_seldrek6078865.cl"},
	{0x89310C1FB60E00D8ULL, "gen00fl_aveline6110362.cl"},
	{0x8931386CFE6576F1ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse4cloth_orsinosdefiance.gad"},
	{0x89318CDBB204141EULL, "6027759_m.fxe"},
	{0x89318ED1F9A526D5ULL, "lgt150pt_ntb_wr_ingrdts.ncs"},
	{0x893191FF808548EFULL, "llac_lt_gallowstemplar_f_2384_1.dds"},
	{0x8931B38B424D8CCBULL, "one300_zevran6089750.cl"},
	{0x8932BD0EF8179717ULL, "6106925_m.fxe"},
	{0x8932C64D1F830291ULL, "vxca_caverns1_mines_i2n_0_0.vlm"},
	{0x89332C5BC8426018ULL, "mer120_hubert6083917.cl"},
	{0x8933752DD0069396ULL, "gencr_coterie_thug_ws.utc"},
	{0x89339D4EC82D0776ULL, "static_static_p41.psh.fxo.dx11.assembly"},
	{0x8933BDD824E1F678ULL, "vxw_sundermount_32e_0_1.vlm"},
	{0x89340C87C465E16CULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\twitchv1.ani"},
	{0x89341452C2D3E64CULL, "art\\vfx\\combat\\vfx_w_2nd_charge_c.mmh"},
	{0x89341C2E88977DCBULL, "vxdr_primevaltomb_2pq_0_0.vlm"},
	{0x89344243A5B9584EULL, "6135629_m.fxe"},
	{0x8934474053930366ULL, "554716459.ogg"},
	{0x893459E0F84492F5ULL, "132727162.ogg"},
	{0x89348944BA518007ULL, "6101556_f.fxe"},
	{0x893547B5DCDA0395ULL, "vxht_0_hightown_a2_56w_0_0.vlm"},
	{0x89357B611084A779ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_scared.gad"},
	{0x8935DDF42A4E036FULL, "6152766_m.fxe"},
	{0x8935E637A6F502B1ULL, "597745693.ogg"},
	{0x8936132A676B27B9ULL, "hlac_dr_0_deeproads_5845_18z.dds"},
	{0x8936C1C3154FAA3CULL, "vxlt_0_lowtown_a2_n_c_gxw_0_0.vlm"},
	{0x8936C36A46844C26ULL, "6164407_m.fxe"},
	{0x8937029E78F3583FULL, "6033648_f.fxe"},
	{0x89374C0438567751ULL, "vxca_caverns2prc_9pk_0_0.vlm"},
	{0x89376B1210A4B9AFULL, "zz_one_debug6069465.cl"},
	{0x893799717A8BBB3BULL, "vxlt_foundry_n_x3_0_0.vlm"},
	{0x8937B618D0D93C87ULL, "cage_04.mao"},
	{0x8937D46621174CD5ULL, "mag110cr_samson.utc"},
	{0x8938396CFE6B969FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse4cloth_orsinosdefiance.ani"},
	{0x8938724ECCEE5A33ULL, "251030057.ogg"},
	{0x8939229C4BF485B1ULL, "579257222.ogg"},
	{0x893940F11A1DF0B2ULL, "245501470.ogg"},
	{0x89395B2D6D33AFE8ULL, "6113296_m.fxe"},
	{0x89399771CFD7D178ULL, "175033924.ogg"},
	{0x8939F055DEA8AB75ULL, "vxlt_hangedman_284_0_0.vlm"},
	{0x8939F26F85D9A056ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs05a.msh"},
	{0x8939FC6F85D9B154ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs05a.mmh"},
	{0x893A006F85D9B867ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs05a.mao"},
	{0x893A1987C46A62AEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\twitchv1.gad"},
	{0x893B5F466424C024ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_sitting_male_loop.gad"},
	{0x893B5F6710043F56ULL, "hlac_ht_0_hightown_a3_-2_0y.dds"},
	{0x893B8C6110892FC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_scared.ani"},
	{0x893B9F16C0AEAA14ULL, "vxdr_primevalentrance_1m2_0_0.vlm"},
	{0x893BD4D18D7FFECAULL, "80423296.ogg"},
	{0x893BF30D22FD9984ULL, "6037125_m.fxe"},
	{0x893CB63C9BE5DE1FULL, "6162245_f.fxe"},
	{0x893CC4F65F2A5638ULL, "799903115.ogg"},
	{0x893CD8EA15E5D3A3ULL, "mag100_thrask6031982.cl"},
	{0x893CFC02A6A32DDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackleft.ani"},
	{0x893D0E1B92F06B49ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowflat.mao"},
	{0x893D1A1B92F07FE2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowflat.mmh"},
	{0x893D1C1B92F08298ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowflat.msh"},
	{0x893D29073C740568ULL, "vxw_encounter3_b5_0_0.vlm"},
	{0x893DC1267E9A3C3FULL, "llac_lt_0_lowtown_a2_n_-131072_0.dds"},
	{0x893DD076DA9D18B7ULL, "189083406.ogg"},
	{0x893E35D046BE8A5FULL, "vxlt_undercityserialkiller_1nx_0_0.vlm"},
	{0x893EA55C0786C46FULL, "51194988.ogg"},
	{0x893EE5CE26392742ULL, "6135236_m.fxe"},
	{0x893F0250C7EC74F1ULL, "art\\characters\\weapons\\dagger\\w_dgr_003a.mmh"},
	{0x893F0E50C7EC8892ULL, "art\\characters\\weapons\\dagger\\w_dgr_003a.mao"},
	{0x893F19C7E38BD643ULL, "one100_solivitus6059911.cl"},
	{0x893F1C50C7ECA0AFULL, "art\\characters\\weapons\\dagger\\w_dgr_003a.msh"},
	{0x893F42D82DB9B61CULL, "vxdr_0_deeproads_ev5_0_0.vlm"},
	{0x893FAD444E404650ULL, "15157032.ogg"},
	{0x893FFAD3F8B2F60EULL, "hlac_ca_cavern1_haunted_24425_13y.dds"},
	{0x89403D6985AC7F53ULL, "character_character_p145.psh.fxo.dx11"},
	{0x8940D5BD346E392CULL, "mcr370_messenger6071822.cl"},
	{0x89410B7297C88659ULL, "dae200_saemus6142803.cl"},
	{0x89411487C470788DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\twitchv1.evt"},
	{0x8941F6A33CDA1442ULL, "llac_w_bonepit_196611_0.dds"},
	{0x89420EA9856082E7ULL, "mag320_orsino6071194.cl"},
	{0x89422A0978CBA51BULL, "vxca_caverns1_mines_e51_0_0.vlm"},
	{0x89422D6720F6996DULL, "gen00fl_anders6125106.cl"},
	{0x89426E46642AF802ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_sitting_male_loop.ani"},
	{0x894295BF8B23C451ULL, "6134728_m.fxe"},
	{0x8942A12B616F9690ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_wmda_0n.dds"},
	{0x8942A2AF4CE3385DULL, "6162389_m.fxe"},
	{0x8942E7F874F975C1ULL, "gen_im_trs_tmp_hvy_cht_r2.uti"},
	{0x89430661E1B8321BULL, "6208208_f.fxe"},
	{0x89432FBDEE966545ULL, "6156733_m.fxe"},
	{0x89437BD3F4F895BEULL, "vxht_0_hightown_a2_au9_0_0.vlm"},
	{0x8943D1C24A12BC72ULL, "505399441.ogg"},
	{0x89444CDCE3676B3BULL, "1005878163.ogg"},
	{0x89452C3B13AD82E4ULL, "vxca_caverns1_e8p_0_0.vlm"},
	{0x89456FE7A5E59E1CULL, "static1_relief.psh.fxo"},
	{0x8945A6BA23D905C7ULL, "646134335.ogg"},
	{0x894694C6C2291CD7ULL, "hlac_ca_caverns1_24422_4x.dds"},
	{0x8946DDFBC9BD67BFULL, "lt_0_lowtown_a2_3_0.rml"},
	{0x8946E31DEBC6D260ULL, "6166854_m.fxe"},
	{0x89475AE3AEBBEB21ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfireout.anb"},
	{0x8947AB68835CA86EULL, "hlac_lt_gallowstemplar_n_2384_0z.dds"},
	{0x89482AEFEC9164E5ULL, "490349822.ogg"},
	{0x894855452A8FD227ULL, "vxw_bonepitmining_a2_252_0_0.vlm"},
	{0x89486232D09EC3F5ULL, "vxlt_hangedman_253_0_0.vlm"},
	{0x8948BACD07975DF6ULL, "one350_terath6087118.cl"},
	{0x8949390FC8E97400ULL, "vxht_0_hightown_a3_c3r_0_0.vlm"},
	{0x894ABA08B9F6863BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galofloor.phy"},
	{0x894AC4D4CC856467ULL, "hlac_lt_0_lowtown_docks_n_0_6x.dds"},
	{0x894ACF54640C51FDULL, "884739909.ogg"},
	{0x894B0402A6AF78D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackleft.evt"},
	{0x894B13E8D4AB8CD9ULL, "vxht_bartrand_n_4ht_0_0.vlm"},
	{0x894B3F685353847CULL, "follower_banter6168776.cl"},
	{0x894B9A26B521D20DULL, "vxca_caverns2_6h9_0_0.vlm"},
	{0x894C78A6CAC9F43FULL, "6151024_m.fxe"},
	{0x894CA372B9D65601ULL, "gen_im_trs_crc_sta_nat_r0.uti"},
	{0x894CF368381D97F9ULL, "vxca_caverns2_8or_0_0.vlm"},
	{0x894D3EA6CC38B985ULL, "vxca_cave_1a2_0_0.vlm"},
	{0x894DCB810BBF2EC3ULL, "vxw_sundermount_2zv_0_0.vlm"},
	{0x894DE529CDDFBA29ULL, "474190068.ogg"},
	{0x894E11B60B7ED255ULL, "vxca_caverns1_mines_fob_0_0.vlm"},
	{0x894E2BE69C921F52ULL, "vxlt_gallowstemplar_f_46i_0_0.vlm"},
	{0x894E4AF028B9A798ULL, "framebuffereffects_hslmatrixfade_p0.psh.fxo"},
	{0x894E98C8DD4C3E0BULL, "6061280_m.fxe"},
	{0x894F068031DCFF4FULL, "vxlt_0_lowtown_docks_n_b64_0_0.vlm"},
	{0x894F4E9DE36DB8EEULL, "fen110_anso6094976.cl"},
	{0x894FB1CE92080478ULL, "316247870.ogg"},
	{0x894FBED1F5E7CE32ULL, "6016650_m.fxe"},
	{0x894FC32E7317C8E3ULL, "w_encounter1_33_0.msh"},
	{0x894FD12E7317E03DULL, "w_encounter1_33_0.mmh"},
	{0x894FD52E7317E74EULL, "w_encounter1_33_0.mao"},
	{0x89508CD94D418B56ULL, "llac_w_woundedcoast_0_5.dds"},
	{0x8950B9E949641C28ULL, "vxca_caverns1_mines_e8a_0_0.vlm"},
	{0x8950C9DD58E680BFULL, "vxlt_hangedman_n_137_0_0.vlm"},
	{0x8951148C3ED950DDULL, "6167089_m.fxe"},
	{0x89518001435805FBULL, "art\\vfx\\combat\\vfx_c_rw_aoeburst_c.anb"},
	{0x89519C7A2DA5D617ULL, "fen210_hadriana6110392.cl"},
	{0x8951CACEF50E2244ULL, "6140274_m.fxe"},
	{0x89522302A6B5CC65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkbackleft.gad"},
	{0x895277D0426C7CBEULL, "vxw_encounter2_jp_0_0.vlm"},
	{0x8952B355907FEEFAULL, "vxw_bonepitdestroyed_a3_mm_0_0.vlm"},
	{0x8953138A533C29C9ULL, "vxlt_warehouse_co_0_0.vlm"},
	{0x89532DD683F64454ULL, "6076395_m.fxe"},
	{0x895333B113DD97D7ULL, "vxht_mansion2_2dt_0_0.vlm"},
	{0x89535422B3BD572CULL, "vxdr_primevaltomb_3lv_0_1.vlm"},
	{0x8953A2DE5C327611ULL, "vxlt_gallowstemplar_f_37b_0_0.vlm"},
	{0x8953A60478D88DA7ULL, "354511631.ogg"},
	{0x8953C9F8B21C1D4AULL, "fen210_dying_magister6108547.cl"},
	{0x895408D3AFA12E4DULL, "dae340_db_drag_mage.cut"},
	{0x89540D8B74DA50C4ULL, "6107411_m.fxe"},
	{0x89542FBE1C42A851ULL, "vxca_caverns2prc_9zi_0_0.vlm"},
	{0x8954B94B864BBC58ULL, "6060459_m.fxe"},
	{0x89553BE3AEC7F39AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfireout.mmh"},
	{0x89566DAD3A6503EEULL, "6134778_m.fxe"},
	{0x8956985CC1566609ULL, "6132999_m.fxe"},
	{0x8956A26701447254ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_2arm_shrug.ani"},
	{0x8956D40FC294DC75ULL, "vxlt_hangedman_222_0_0.vlm"},
	{0x8956D64F485EF991ULL, "338697416.ogg"},
	{0x8956D78343FD9D46ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxa.phy"},
	{0x89570B1B92857782ULL, "6090322_m.fxe"},
	{0x8957125F7306DB7EULL, "llac_lt_0_lowtown_a2_-3_5.dds"},
	{0x8957306EE78732E9ULL, "gen00fl_fenris6103360.cl"},
	{0x89577DCE66B5A0A7ULL, "6041585_m.fxe"},
	{0x8957AAECBADF8C80ULL, "vxht_0_hightown_a3_c4s_0_0.vlm"},
	{0x8958E86F726597E7ULL, "dae200_bran_st.stg"},
	{0x8959509BE7168215ULL, "935557850.ogg"},
	{0x8959AE0EE5E83B4EULL, "134551780.ogg"},
	{0x8959BE614EECE188ULL, "545182193.ogg"},
	{0x895A037DD152FA6EULL, "6098223_m.fxe"},
	{0x895A26850D6F3135ULL, "873075735.ogg"},
	{0x895A61891AF2871AULL, "qun221_petrice6083748.cl"},
	{0x895AB17763B04755ULL, "lightning_bolt3.dds"},
	{0x895AB8967DEAB7F7ULL, "hlac_ht_mansion2b_818_25z.dds"},
	{0x895B2F3DA500ADF4ULL, "art\\levels\\races\\design\\trap_floorlever_0.pwk"},
	{0x895BBA4C2971A303ULL, "vxht_bartrand_n_66n_0_0.vlm"},
	{0x895BC86DB370E6E2ULL, "6096180_m.fxe"},
	{0x895C08A1C95341DEULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.cess.evt"},
	{0x895C6F1BC837007EULL, "vxca_caverns2_80y_0_1.vlm"},
	{0x895CECB33879ABDBULL, "vxlt_undercityserialkiller_1id_0_1.vlm"},
	{0x895D9F67014A8BB6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\leaning_back_low_2arm_shrug.gad"},
	{0x895DBDBA0EA2E4B4ULL, "6096850_m.fxe"},
	{0x895E1E507F289782ULL, "qcr220_grey_wardens6060454.cl"},
	{0x895EB6869BA0FB1AULL, "art\\vfx\\combat\\vfx_c_single_2nd_stun_c.anb"},
	{0x895EBCBE930C926CULL, "6046958_m.fxe"},
	{0x895ED55FA6BA695AULL, "w_sundermount_68_0.rml"},
	{0x895EFA6C9429AFE2ULL, "vxlt_foundry_n_zt_0_0.vlm"},
	{0x895F118EF48F33F0ULL, "aip_gateguardian_base.gda"},
	{0x895F4B7912AA234AULL, "vxlt_gallowscourtyard_n_66q_0_0.vlm"},
	{0x895FB0CEF0F20C58ULL, "mag360cr_hostage_merrill.utc"},
	{0x895FDC0768ADD619ULL, "llac_lt_0_lowtown_a2_n_-196608_4.dds"},
	{0x895FF579B1CA9D48ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1b.mmh"},
	{0x895FFBA1F8F584ACULL, "76893621.ogg"},
	{0x89600179B1CAB1EBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1b.mao"},
	{0x89601379B1CAD072ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1b.msh"},
	{0x8960E1213AAD9DC2ULL, "llac_lt_0_lowtown_a3_n_c_65534_2.dds"},
	{0x8960E65F9EB8BB79ULL, "685655503.ogg"},
	{0x8960EFF255A887A4ULL, "220654344.ogg"},
	{0x89616DBF667CE18AULL, "591884467.ogg"},
	{0x89616DC38F511BB3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\ready.evt"},
	{0x8961A58E05D3B057ULL, "vxht_mansion2_2aw_0_0.vlm"},
	{0x896212D4DC7D1505ULL, "279607659.ogg"},
	{0x89625E7ECDBC7B40ULL, "vxw_bonepitmining_a2_l2_0_0.vlm"},
	{0x89625F51C71A8329ULL, "315358109.ogg"},
	{0x8962C726229E9BD4ULL, "vxht_keep_6ah_0_0.vlm"},
	{0x89631FA1C9598709ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.cess.gad"},
	{0x896336943A17B75CULL, "6165154_m.fxe"},
	{0x896352DC61451706ULL, "164583115.ogg"},
	{0x8963A0058A11F8EFULL, "6143724_m.fxe"},
	{0x8964599EDAB675A1ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_001a_0t.dds"},
	{0x8964AF2DAF7A8F80ULL, "753358559.ogg"},
	{0x89650FFDADCE7BC8ULL, "llac_ca_caverns2prc_5802_39.dds"},
	{0x89651F1DFC88352CULL, "6136374_m.fxe"},
	{0x896571EDFE48DD7CULL, "hlac_lt_0_lowtown_docks_a3_n_65533_0x.dds"},
	{0x89659A56C889F71FULL, "w_encounter1_6_0.msh"},
	{0x8965A056C88A01C1ULL, "w_encounter1_6_0.mmh"},
	{0x8965AC56C88A1662ULL, "w_encounter1_6_0.mao"},
	{0x8965B65FA6C0534EULL, "w_sundermount_68_0.phy"},
	{0x8965D146ED87B726ULL, "one120_athenril6173938.cl"},
	{0x8965E5231C6D611DULL, "6083358_m.fxe"},
	{0x89661FEC635DE0BCULL, "168065723.ogg"},
	{0x8966835D9DE4D71FULL, "751269719.ogg"},
	{0x89675855F081D686ULL, "365391742.ogg"},
	{0x896866C38F572EF4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\ready.gad"},
	{0x8968B677445EED10ULL, "6055068_m.fxe"},
	{0x8968DB2591BA2BBDULL, "dae210ip_compound_gate.ncs"},
	{0x8969257065134A99ULL, "6054999_f.fxe"},
	{0x896974DF5BC3F378ULL, "combat_lmm_r1_a2.gda"},
	{0x8969BA19EE1E9974ULL, "waterflowtoolset.psh.fxo"},
	{0x8969CC3544789663ULL, "vxht_chantry_27e_0_0.vlm"},
	{0x8969CDCC6FADFC79ULL, "vxht_0_hightown_bnt_0_0.vlm"},
	{0x8969D7221C09C8F9ULL, "vxca_caverns2_8qp_0_0.vlm"},
	{0x8969F0A1C95F5657ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.cess.ani"},
	{0x896A2C291B67BB83ULL, "vxht_bartrand_n_63q_0_0.vlm"},
	{0x896A59017F268712ULL, "vxw_encounter2_camp_73_0_0.vlm"},
	{0x896B52B9480F4C7CULL, "6066609_f.fxe"},
	{0x896B5335BB80DC9CULL, "hlac_lt_0_lowtown_docks_a3_-196610_6y.dds"},
	{0x896B646F9F243A8CULL, "zz_qun_debug6088087.cl"},
	{0x896BBDF5E5B7C7FCULL, "vxlt_blackemporium_to_0_0.vlm"},
	{0x896C1A150E835786ULL, "vxht_chantry_n_36q_0_0.vlm"},
	{0x896C20BCA953615AULL, "gen_im_npc_wep_two_qunc.uti"},
	{0x896D005CA7C7ADBEULL, "character_character_p209.psh.fxo.dx11"},
	{0x896D420344F013A6ULL, "hlac_ht_0_hightown_-65536_4y.dds"},
	{0x896DD8AC7AFE1946ULL, "character_character_p191.psh.fxo.dx11"},
	{0x896E098A14632E1AULL, "mrl150_merrill6085716.cl"},
	{0x896E0CC087248E89ULL, "vxw_bonepitmining_a2_265_0_0.vlm"},
	{0x896E6B36FA58B5F7ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_008a_0n.dds"},
	{0x896E79125B371723ULL, "6027560_f.fxe"},
	{0x896EECE88A63161BULL, "vxlt_hangedman_268_0_0.vlm"},
	{0x896F5B8A2658ADBEULL, "vxw_encounter2_pr_0_0.vlm"},
	{0x896F75C38F5D66D2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\ready.ani"},
	{0x896FC434D782F8D5ULL, "443882822.ogg"},
	{0x8970176AF7C9C8D7ULL, "vxht_mansion2_2jv_0_0.vlm"},
	{0x897029E1154DF347ULL, "tal_tau_clu.mao"},
	{0x89706CBAD737C93CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bnt_dblw_lp.ani"},
	{0x8970D05BBFB293C0ULL, "vxw_bonepitmining_a2_k3_0_0.vlm"},
	{0x897139031494B454ULL, "vxht_keep_6ji_0_0.vlm"},
	{0x89722811F334143BULL, "402855366.ogg"},
	{0x89726BE6C8F70DA7ULL, "6033724_m.fxe"},
	{0x89729D14F93871F5ULL, "hlac_lt_0_lowtown_1_5z.dds"},
	{0x897318FD090098B0ULL, "vxlt_hangedman_n_5i_0_0.vlm"},
	{0x89738EFD696DB514ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transitiontostage1.ani"},
	{0x8973B1EA9578F7BDULL, "421909293.ogg"},
	{0x89743926DA37DD6BULL, "vxca_caverns1_mines_e7k_0_0.vlm"},
	{0x89743C26A76F76ACULL, "follower_banter6170382.cl"},
	{0x8974683FB50263BEULL, "6040767_m.fxe"},
	{0x89748E70723444EFULL, "vxht_mansion_n_3jh_0_0.vlm"},
	{0x8974B6A0905312DEULL, "975266474.ogg"},
	{0x89750467E02BBE33ULL, "vxlt_gallowstemplar_f_6dh_0_0.vlm"},
	{0x89752A936F0B9B5AULL, "406478511.ogg"},
	{0x8975A69518EDF392ULL, "502273346.ogg"},
	{0x89764DA3210002CEULL, "46028493.ogg"},
	{0x89767A705176F354ULL, "vxht_0_hightown_n_byw_0_0.vlm"},
	{0x8976A108BA1B96DAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galofloor.msh"},
	{0x8976AF08BA1BAE13ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galofloor.mao"},
	{0x8976B308BA1BB520ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff_galofloor.mmh"},
	{0x8976E9117DF0D9CEULL, "hlac_lt_0_lowtown_65536_11x.dds"},
	{0x8976EFBD8B041B8DULL, "vxca_caverns2_6y6_0_0.vlm"},
	{0x8977285436550590ULL, "6052168_m.fxe"},
	{0x897779BAD73DFE7EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_bnt_dblw_lp.gad"},
	{0x8977840C62608D46ULL, "6157214_m.fxe"},
	{0x89778A35282010B8ULL, "826750116.ogg"},
	{0x8977A8228FA7D683ULL, "dae000_ineria6136963.cl"},
	{0x8977A87CF0E4ECB4ULL, "character_character_p122.psh.fxo.dx11"},
	{0x89780BEDA4FE45E3ULL, "and111_lirene6084520.cl"},
	{0x897811934884D6E9ULL, "vxdr_dungeon_1m0_0_1.vlm"},
	{0x89781BA42EFFB03BULL, "rdr110dg_dockworker6015938.cl"},
	{0x897888819A614BCDULL, "155290801.ogg"},
	{0x8978908BB41D22ABULL, "character_character_p33.psh.fxo.dx11.assembly"},
	{0x89789687FF221074ULL, "vxlt_hangedman_2yb_0_0.vlm"},
	{0x8978A761B9E528A1ULL, "vxw_bonepitmining_a2_b1_0_0.vlm"},
	{0x8978D3EFDECE5E95ULL, "art\\characters\\playercharacter\\shared\\textures\\gore_default.dds"},
	{0x897982A60D538F4CULL, "vxlt_gallowscourtyard_6j8_0_0.vlm"},
	{0x8979CF0D81AD458BULL, "vxht_brothel_n_75c_0_1.vlm"},
	{0x897A33B31C263423ULL, "6075105_f.fxe"},
	{0x897A8BFD6973CF76ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transitiontostage1.gad"},
	{0x897B0FB39B6866DBULL, "w_woundedcoast_110_0.msh"},
	{0x897B21B39B688526ULL, "w_woundedcoast_110_0.mao"},
	{0x897B2DB39B6899C5ULL, "w_woundedcoast_110_0.mmh"},
	{0x897B2F711B36CC15ULL, "vxw_bonepitmining_a2_n9_0_0.vlm"},
	{0x897B34327229417BULL, "hlac_ca_caverns1_24425_4y.dds"},
	{0x897B9D28BA7D3F05ULL, "395591423.ogg"},
	{0x897C01D5FA3B3C71ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackr_1.mao"},
	{0x897C2FBF6C5BC5D6ULL, "vxdr_0_deeproads_egl_0_0.vlm"},
	{0x897C75C1BF8A2028ULL, "6115315_m.fxe"},
	{0x897D3D373E1F1B15ULL, "vxht_0_hightown_a2_c3k_0_1.vlm"},
	{0x897D3F084153715EULL, "vxlt_0_lowtown_a3_n_c_gl8_0_0.vlm"},
	{0x897DA05A5230F482ULL, "6135940_m.fxe"},
	{0x897DAAF994F47E45ULL, "vxw_bonepitdestroyed_a3_mf_0_0.vlm"},
	{0x897DF575036B09EFULL, "vxlt_hangedman_n_1ye_0_0.vlm"},
	{0x897E6921291E7349ULL, "vxlt_warehouse_rl_0_0.vlm"},
	{0x897ECBE37704E4E3ULL, "vxdr_0_deeproads_f61_0_0.vlm"},
	{0x897FD69B6E1BDC11ULL, "992772904.ogg"},
	{0x89802CCCDFC59DE3ULL, "art\\characters\\weapons\\arrows\\textures\\w_arw_001_d.dds"},
	{0x8980495C1BEF48F0ULL, "hightown_male_cityguard_mob6170071.cl"},
	{0x898080D80D697DF8ULL, "213786988.ogg"},
	{0x8980D1FAC12FE319ULL, "mer341_jansen6054868.cl"},
	{0x8980F397092310A6ULL, "ave200_donnic6111049.cl"},
	{0x8981A6FD697A1BD5ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transitiontostage1.evt"},
	{0x8981AF8387FFB932ULL, "vxw_sundermount_36r_0_0.vlm"},
	{0x8981BD353CB1E70FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext.gad"},
	{0x89826E2546BA6B6CULL, "zz_dae_debug6049792.cl"},
	{0x8982E50539DE96A8ULL, "6070119_m.fxe"},
	{0x8983D78A662367A4ULL, "art\\characters\\creatures\\mubari\\animation\\die_v1.ani"},
	{0x8983EC45AE593C5FULL, "hlac_ca_caverns2_5802_71x.dds"},
	{0x8984208D8C0B58F9ULL, "llac_lt_gallowstemplar_2384_6.dds"},
	{0x8984576519859610ULL, "vxlt_0_lowtown_docks_a3_n_cvi_0_0.vlm"},
	{0x8984C8F645082955ULL, "hlac_lt_0_lowtown_docks_n_65533_4z.dds"},
	{0x8984E8E569719377ULL, "936537174.ogg"},
	{0x8984EC4D436D0BD4ULL, "vxht_0_hightown_n_bdt_0_0.vlm"},
	{0x89854635712B1A8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\neck_shot_from_behind_dead_loop.ani"},
	{0x898558E6124809B4ULL, "6083607_m.fxe"},
	{0x89859B1CB5630174ULL, "vxca_caverns1_mines_edi_0_0.vlm"},
	{0x8985BC44B6F9E4B4ULL, "6144186_m.fxe"},
	{0x8985C0469A9895B3ULL, "isa111_isabela6094554.cl"},
	{0x89862E594A43ABD8ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend3_n.dds"},
	{0x898655F3BF797E8EULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st104a_0d.dds"},
	{0x8986B186539A2D79ULL, "vxht_0_hightown_btz_0_0.vlm"},
	{0x89870FE2FF53EC83ULL, "vxht_bartrand_n_69s_0_0.vlm"},
	{0x89879E6E96787B12ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.mmh"},
	{0x8987B9869BC405CDULL, "art\\vfx\\combat\\vfx_c_single_2nd_stun_c.mmh"},
	{0x8987D2353CB676BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext.evt"},
	{0x8987D5284C265714ULL, "zz_one_debug6079326.cl"},
	{0x898939E028309E3EULL, "hlac_w_blightlands_-65533_0x.dds"},
	{0x89896E2E659D0218ULL, "6129046_m.fxe"},
	{0x898979055BB426BBULL, "character_character_p110.psh.fxo"},
	{0x8989B8442EB8A27DULL, "790466070.ogg"},
	{0x8989C3C51841A8FFULL, "194738771.ogg"},
	{0x8989D1076F11D8CEULL, "art\\levels\\races\\design\\_textures\\prp_gasbarrel_sa.dds"},
	{0x8989D48A6627CE06ULL, "art\\characters\\creatures\\mubari\\animation\\die_v1.gad"},
	{0x898A12E8A1952C0FULL, "hlac_lt_backalley_g_65537_7x.dds"},
	{0x898A242CB1C64802ULL, "205781388.ogg"},
	{0x898B4DFDFA61BC97ULL, "hlac_ca_caverns1_mines_24425_8z.dds"},
	{0x898B8056C8AA9AEAULL, "w_encounter1_6_0.rml"},
	{0x898BB0E5334989DEULL, "vxlt_0_lowtown_a3_n_c_gc9_0_0.vlm"},
	{0x898BD0EAED2851E8ULL, "112780409.ogg"},
	{0x898BEA488D0669B7ULL, "6133671_m.fxe"},
	{0x898C47DF8A47C24CULL, "vxht_mansion2_3ck_0_0.vlm"},
	{0x898C6751F561226FULL, "vxlt_hangedman_n_1vd_0_0.vlm"},
	{0x898CA48A5B4FEE6CULL, "494826094.ogg"},
	{0x898CDAFE1B148BC9ULL, "vxlt_warehouse_ws_0_0.vlm"},
	{0x898CFB24DBB5F9D7ULL, "vxht_mansion2_2pp_0_0.vlm"},
	{0x898D15D17DEDFE42ULL, "463623172.ogg"},
	{0x898DCC96DD8C67A9ULL, "gen00fl_merrill6128657.cl"},
	{0x898DDAFDE959245FULL, "959642493.ogg"},
	{0x898DF31BDA7E6207ULL, "dae200st_solivitus.utm"},
	{0x898E411A734CB71AULL, "vxdr_0_deeproads_cja_0_0.vlm"},
	{0x898E517F4DC035E7ULL, "vxlt_gallowstemplar_n_34y_0_0.vlm"},
	{0x898E5849A7C2732AULL, "64426503.ogg"},
	{0x898E7618B040B589ULL, "hlac_w_sundermount_-327685_0x.dds"},
	{0x898E94CD776A48EBULL, "540692443.ogg"},
	{0x898E9BD004E3B2C7ULL, "vxw_sundermount_2fn_0_0.vlm"},
	{0x898ED5BD8C19587CULL, "6165737_f.fxe"},
	{0x898F7CF3E7593B60ULL, "mrl150_cut_amulet_ritual6147029.cl"},
	{0x898F837101B64FD3ULL, "6110823_m.fxe"},
	{0x898FA07011CDB2DDULL, "w_woundedcoast_144_0.mao"},
	{0x898FA47011CDB916ULL, "w_woundedcoast_144_0.mmh"},
	{0x898FAE7011CDCA14ULL, "w_woundedcoast_144_0.msh"},
	{0x898FC974BBFAAD41ULL, "vxca_cavern1_haunted_e40_0_0.vlm"},
	{0x898FCA04F679D565ULL, "vxw_bonepitmining_a2_dk_0_0.vlm"},
	{0x898FCFAB09FDC4D0ULL, "mer231_brekker6075164.cl"},
	{0x89901A8729B06325ULL, "mag210_arianni6024112.cl"},
	{0x899082EC5C1B244DULL, "6080795_m.fxe"},
	{0x8990CF8A662DE4E5ULL, "art\\characters\\creatures\\mubari\\animation\\die_v1.evt"},
	{0x89914A2CE84A9569ULL, "vxht_mansion2b_34c_0_0.vlm"},
	{0x899156E22BD4AF28ULL, "6075219_m.fxe"},
	{0x8991649AEA7F776CULL, "vxlt_0_lowtown_a2_n_c_hki_0_0.vlm"},
	{0x899177730A0C00AAULL, "126732256.ogg"},
	{0x8991B699B9398330ULL, "hlac_ht_brothel_n_3381_0x.dds"},
	{0x899206604A63DEDDULL, "zz_dae_debug6047599.cl"},
	{0x89922F121082D7FFULL, "157053712.ogg"},
	{0x8992A97910C5AA1FULL, "gen00fl_anders6096493.cl"},
	{0x89930565BC3C7591ULL, "llac_lt_undercityencounter_-1_30.dds"},
	{0x89935E2A35632454ULL, "vxht_0_hightown_n_bou_0_0.vlm"},
	{0x8993B3963D4D7A48ULL, "hlac_lt_0_lowtown_a3_n_c_131071_1y.dds"},
	{0x8993CAC3043E2234ULL, "6086907_m.fxe"},
	{0x8993E7054FA60995ULL, "follower_barks6153792.cl"},
	{0x8994041C27D9002BULL, "vxca_cavern1_haunted_j23_0_0.vlm"},
	{0x89940F5F593E4688ULL, "gen_im_trs_qun_hvy_glv_r0.uti"},
	{0x899410885C793504ULL, "zz_ave_debug6056778.cl"},
	{0x8994B125A06A709BULL, "6016013_m.fxe"},
	{0x89952FA1492A4D38ULL, "733516904.ogg"},
	{0x8995520DDFA12B2CULL, "6164957_m.fxe"},
	{0x8995DA353CC2C1D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext.ani"},
	{0x89963C151FA4FBE9ULL, "mer340_bonepit_comments6083141.cl"},
	{0x8996C89366291F82ULL, "qun131ip_treasure.utp"},
	{0x8996FB6A9EC6364EULL, "516561552.ogg"},
	{0x89973AE4886797EAULL, "vxht_mansion2b_20y_0_0.vlm"},
	{0x899865ED12884F8FULL, "lt_gallowstemplar_n_2383.rml"},
	{0x8998B25FA6EBA666ULL, "w_sundermount_68_0.anb"},
	{0x8998BDE1F29DE091ULL, "6191228_m.fxe"},
	{0x8998D6F97BD8E594ULL, "vxlt_warehouse_n_14q_0_0.vlm"},
	{0x899905C875A7A068ULL, "hlac_ht_chantry_2534_10z.dds"},
	{0x8999120442A4021AULL, "vxht_chantry_2bg_0_0.vlm"},
	{0x899913795047F6D6ULL, "vxdr_0_deeproads_emn_0_0.vlm"},
	{0x89991F35713B81C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\neck_shot_from_behind_dead_loop.gad"},
	{0x8999C3BE75FC445FULL, "6098752_m.fxe"},
	{0x899A9CB5CA58FC53ULL, "6149911_m.fxe"},
	{0x899AB9BC7C3DDACCULL, "vxht_mansion2_3fh_0_0.vlm"},
	{0x899AD92EE7573AEFULL, "vxlt_hangedman_n_1sg_0_0.vlm"},
	{0x899B0580065C754AULL, "ser120_ghyslain6118423.cl"},
	{0x899B4CDB0D0AA449ULL, "vxlt_warehouse_xr_0_0.vlm"},
	{0x899B59D9C005B351ULL, "6104756_m.fxe"},
	{0x899BC7562DAAC927ULL, "625278986.ogg"},
	{0x899C188570559178ULL, "vxht_mansion_n_6jj_0_0.vlm"},
	{0x899C19EA081BC562ULL, "vxlt_gallowstemplar_f_5bx_0_0.vlm"},
	{0x899C58D67C95F4E8ULL, "llac_ca_caverns1_mines_24422_6.dds"},
	{0x899C813847D3630CULL, "6179255_f.fxe"},
	{0x899C818A05B9FFCCULL, "vxlt_gallowscourtyard_n_7hd_0_0.vlm"},
	{0x899D0DACF6D9CB47ULL, "vxw_sundermount_2ki_0_0.vlm"},
	{0x899D2D4A07204505ULL, "mag100_cullen6171204.cl"},
	{0x899DBA8670A2A904ULL, "vxlt_0_lowtown_n_ets_0_0.vlm"},
	{0x899E3D7E740471B3ULL, "zz_followers_banter6158721.cl"},
	{0x899E4E92610B7B16ULL, "6160968_m.fxe"},
	{0x899F29D1B6934024ULL, "6164276_m.fxe"},
	{0x899F501661DBF694ULL, "vxht_0_hightown_a2_9ao_0_0.vlm"},
	{0x899FE0E774039351ULL, "6179456_f.fxe"},
	{0x89A0C4C419FE4E4BULL, "lt_hangedman_n_3567.anb"},
	{0x89A0D45FD70ADF91ULL, "water_water_p2.psh.fxo.dx11"},
	{0x89A13CC23AD1C095ULL, "109588697.ogg"},
	{0x89A172865BB69E2DULL, "ft_dust.mao"},
	{0x89A18A58D0F77CD1ULL, "6096257_m.fxe"},
	{0x89A1D75F8F5CB786ULL, "dre000_sibling_has_blight6037024.cl"},
	{0x89A20C5D360305C1ULL, "vxlt_hangedman_n_25d_0_0.vlm"},
	{0x89A2BCD44D7CBB25ULL, "vxht_mansion2_1tm_0_0.vlm"},
	{0x89A3397F9D56FF51ULL, "6107228_f.fxe"},
	{0x89A360DB47ADB133ULL, "music_projects.gda"},
	{0x89A369C5497E4DBCULL, "573293456.ogg"},
	{0x89A3954037865E79ULL, "vxht_0_hightown_bzx_0_0.vlm"},
	{0x89A3D1FF70AFB770ULL, "6201849_m.fxe"},
	{0x89A421FD8420F409ULL, "dre000_sandal6028404.cl"},
	{0x89A42AA00630DA9CULL, "333673184.ogg"},
	{0x89A4A42CFFF03FFEULL, "6165354_m.fxe"},
	{0x89A4BCF6DC791E45ULL, "vxht_mansion2b_1xp_0_1.vlm"},
	{0x89A4E72316E5FC46ULL, "6076656_m.fxe"},
	{0x89A4F6AF606C7F36ULL, "mcr310_argument6075105.cl"},
	{0x89A58CE7D2CBB813ULL, "71624409.ogg"},
	{0x89A5B55FA6F63511ULL, "w_sundermount_68_0.mmh"},
	{0x89A5C15FA6F649B2ULL, "w_sundermount_68_0.mao"},
	{0x89A5CF5FA6F661CFULL, "w_sundermount_68_0.msh"},
	{0x89A61D21D0646EC1ULL, "lgt110_gxa_vgk_silver_order6117015.cl"},
	{0x89A67F06DDED44E6ULL, "vxw_bonepitmining_a2_79_0_0.vlm"},
	{0x89A6E6FD0D1C9C6CULL, "zz_qun_debug6028885.cl"},
	{0x89A87B4CE5AA779EULL, "753706460.ogg"},
	{0x89A896C10EFDEB8AULL, "vxlt_foundry_n_y1_0_0.vlm"},
	{0x89A924FE91F97D4EULL, "art\\characters\\creatures\\desiredemon\\animation\\die_v1.evt"},
	{0x89A98A9BC8102DF1ULL, "6093632_m.fxe"},
	{0x89A99C7EE6CE553FULL, "10489126.ogg"},
	{0x89A9D7F690445E48ULL, "vxlt_0_lowtown_docks_a3_n_btv_0_0.vlm"},
	{0x89AA8A62624BA9F8ULL, "vxht_mansion_n_6mk_0_0.vlm"},
	{0x89AAD8C60AE0ED3DULL, "vxlt_0_lowtown_a2_ghs_0_0.vlm"},
	{0x89AAF8E3602E5EE1ULL, "gen00fl_fenris6116564.cl"},
	{0x89AB31B8B5223ECCULL, "hlac_lt_0_lowtown_0_7z.dds"},
	{0x89AB7F89E8CFE3C7ULL, "vxw_sundermount_2lh_0_0.vlm"},
	{0x89AB8492049F0062ULL, "dae000_gamlen6144341.cl"},
	{0x89AB946631AF9C42ULL, "vxlt_0_lowtown_docks_a3_n_cwx_0_0.vlm"},
	{0x89ABDE4B43F64175ULL, "6085855_m.fxe"},
	{0x89AC2BE25C5714F4ULL, "vxht_0_hightown_a2_byl_0_0.vlm"},
	{0x89AC2C636298C184ULL, "vxlt_0_lowtown_n_eqp_0_0.vlm"},
	{0x89AC8547EB071E53ULL, "548339726.ogg"},
	{0x89ACADBEDA660665ULL, "vxw_bonepitmining_a2_ni_0_0.vlm"},
	{0x89AD1EF14BB353A8ULL, "vxw_sundermount_30o_0_1.vlm"},
	{0x89AD7635E4911843ULL, "vxca_caverns1_foa_0_0.vlm"},
	{0x89ADB05FC4B7ECD4ULL, "dre000_dougal6130990.cl"},
	{0x89AE100934E27717ULL, "vxlt_undercityserialkiller_14f_0_0.vlm"},
	{0x89AE55E43A0CA6EFULL, "vxht_mansion_n_3vt_0_0.vlm"},
	{0x89AE70DFD5E573BEULL, "6132981_m.fxe"},
	{0x89AE961B2EE2B151ULL, "vxht_keep_5x5_0_0.vlm"},
	{0x89AFACFBEF67DF90ULL, "vxlt_0_lowtown_docks_a3_n_cyt_0_0.vlm"},
	{0x89AFB12D63C47678ULL, "lt_0_lowtown_a2_n_c_54_0.rml"},
	{0x89AFCC2591F61484ULL, "dae210ip_compound_gate.utp"},
	{0x89B01BFE91FF8CF9ULL, "art\\characters\\creatures\\desiredemon\\animation\\die_v1.gad"},
	{0x89B08242CA727064ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_femalestatue_n.dds"},
	{0x89B09D6E969B7EB9ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.anb"},
	{0x89B10D59406BA469ULL, "6180819_m.fxe"},
	{0x89B140732858DF8BULL, "vxw_encounter2_camp_ow_0_1.vlm"},
	{0x89B2AE6167019DA3ULL, "hlac_w_blightlands_-65538_0x.dds"},
	{0x89B2F671D3630EC5ULL, "limp_ele.mao"},
	{0x89B3819C84EB7669ULL, "vxlt_0_lowtown_docks_a3_n_cow_0_0.vlm"},
	{0x89B3A0E73ECE003FULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzcot.msh"},
	{0x89B3A6E73ECE0AE1ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzcot.mmh"},
	{0x89B3B2E73ECE1E82ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzcot.mao"},
	{0x89B44448FD9C6C43ULL, "796407930.ogg"},
	{0x89B493FBE598AE85ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_amllcrstold_sa.dds"},
	{0x89B50F2A1885C448ULL, "llac_ca_cavern1_haunted_24426_11.dds"},
	{0x89B55EC1DF40E9B9ULL, "ser221_party_comments6209948.cl"},
	{0x89B56FF7C67F6D5DULL, "6160956_m.fxe"},
	{0x89B5A0499F5AC259ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\blendtree\\meredith_movement.anb"},
	{0x89B5FD670673716BULL, "hlac_lt_undercityencounter_-1_26y.dds"},
	{0x89B62CFE92041547ULL, "art\\characters\\creatures\\desiredemon\\animation\\die_v1.ani"},
	{0x89B63D2951B187A3ULL, "6068522_m.fxe"},
	{0x89B6751AF4EF2567ULL, "vxlt_0_lowtown_docks_n_b06_0_0.vlm"},
	{0x89B6E56AF533E642ULL, "llac_lt_undercityencounter_-1_19.dds"},
	{0x89B87C2CB7557FCBULL, "prty_bethany_armor_3_warden.dds"},
	{0x89B8B2F1F2B356C4ULL, "gen00fl_anders6107475.cl"},
	{0x89B919197759D331ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_001a_0d.dds"},
	{0x89B91F6C33A09BF4ULL, "wm_ico_mansion_roll.dds"},
	{0x89B980BAA326B282ULL, "6109792_m.fxe"},
	{0x89B9F166DAC5FC47ULL, "vxw_sundermount_2qk_0_0.vlm"},
	{0x89BA2336B533EF71ULL, "287186698.ogg"},
	{0x89BA7349CCD26EFCULL, "vfx_crack01.dds"},
	{0x89BBDA5E076FB61CULL, "284870984.ogg"},
	{0x89BBFB0D47CFE92AULL, "88902489.ogg"},
	{0x89BC3EC4E772036BULL, "t3_fat_f02.tnt"},
	{0x89BC815CA3E423A6ULL, "6063948_m.fxe"},
	{0x89BCABB92216F4ADULL, "6121509_f.fxe"},
	{0x89BCD323CBE38DD5ULL, "framebuffereffects_range_dof_p0.psh.fxo.dx11"},
	{0x89BD2CBF5CB2C3D3ULL, "w_sundermount_55_0_lc.mmh"},
	{0x89BDE2F802040A42ULL, "mcr310_sibling6095300.cl"},
	{0x89BDE7F417339236ULL, "14849174.ogg"},
	{0x89BE0A0222A4549CULL, "llac_lt_undercitydungeon_-1_38.dds"},
	{0x89BE2E3DFB80786FULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyo_0s.dds"},
	{0x89BE66BCA5FEB700ULL, "vxht_keep_4ha_0_0.vlm"},
	{0x89BF09D0D6846DD6ULL, "hlac_lt_gallowsprison_n_-1_10x.dds"},
	{0x89BF95D641DE0625ULL, "lt_0_lowtown_26_0_lc.mmh"},
	{0x89BF9CC530FC0D51ULL, "366709659.ogg"},
	{0x89BFA08E3168EC25ULL, "vxht_mansion2_1bk_0_0.vlm"},
	{0x89BFA82F9D3619CEULL, "sib131_letter_react6127316.cl"},
	{0x89BFE0A8F243FB11ULL, "6041558_m.fxe"},
	{0x89C006672184F0B9ULL, "vxlt_gallowstemplar_f_42z_0_0.vlm"},
	{0x89C068A3A929E30DULL, "vxlt_gallowstemplar_n_3sz_0_0.vlm"},
	{0x89C06B7A7913A3D2ULL, "6069609_m.fxe"},
	{0x89C082F469F4D8FBULL, "766344067.ogg"},
	{0x89C0CED91DC57ED2ULL, "6166988_m.fxe"},
	{0x89C2A5822C2ABA0FULL, "117240866.ogg"},
	{0x89C2AE5F488E04F3ULL, "6201248_m.fxe"},
	{0x89C33447CDC3A7C6ULL, "gen00fl_merrill6111223.cl"},
	{0x89C34893DBFCA2FDULL, "6195879_m.fxe"},
	{0x89C3704D6AA3D789ULL, "character_character_p176.psh.fxo.dx11"},
	{0x89C37C8A2FE11677ULL, "fx_lava_blend.mao"},
	{0x89C3C1C7D0E549DDULL, "w_sundermount_23_0.rml"},
	{0x89C3C9B70CDFB41CULL, "vxlt_0_lowtown_docks_a3_95i_0_0.vlm"},
	{0x89C4569D748F6907ULL, "6092274_f.fxe"},
	{0x89C456DA1D2BA823ULL, "vxw_woundedcoast_1fm_0_0.vlm"},
	{0x89C4630F6C354397ULL, "lowtown_male_alienage1_sta6162731.cl"},
	{0x89C4A7F2BE710112ULL, "6104451_f.fxe"},
	{0x89C541162F1B1CC4ULL, "vxht_mansion2b_1c5_0_0.vlm"},
	{0x89C63A52D5FD3C4BULL, "714245334.ogg"},
	{0x89C6622EE364819BULL, "hlac_w_woundedcoast_262147_0z.dds"},
	{0x89C6AFDD6C076C32ULL, "6191251_f.fxe"},
	{0x89C786785ECEC064ULL, "6096191_m.fxe"},
	{0x89C7BC7FEECD1E3DULL, "vxlt_0_lowtown_a2_gbq_0_0.vlm"},
	{0x89C7D46BF2478D4FULL, "6053568_m.fxe"},
	{0x89C80C80911BC84FULL, "6024562_m.fxe"},
	{0x89C818975912190AULL, "hlac_lt_darktown_-1_74y.dds"},
	{0x89C85DB40C09E60CULL, "mag291_vincento6027331.cl"},
	{0x89C86343CCBC14C7ULL, "vxw_sundermount_2zj_0_0.vlm"},
	{0x89C8E87FDD5FD27CULL, "gen00fl_fenris6126447.cl"},
	{0x89C905728BF24927ULL, "vxdr_primevalentrance_2oi_0_1.vlm"},
	{0x89C9277BFAEFD6F9ULL, "vxlt_gallowstemplar_5ya_0_0.vlm"},
	{0x89C93A80C63078EAULL, "dae000_madam_lusine6151632.cl"},
	{0x89C969FAB913EC76ULL, "6164227_m.fxe"},
	{0x89C9C63BCF38B367ULL, "vxca_caverns2_90d_0_0.vlm"},
	{0x89CA1027A9491012ULL, "6171917_m.fxe"},
	{0x89CA7016ECBD4305ULL, "vxht_mansion_n_6js_0_0.vlm"},
	{0x89CAB088BB7AFD1AULL, "mage_arcaneshield.dds"},
	{0x89CAC5A5A661B953ULL, "vxlt_0_lowtown_n_h78_0_0.vlm"},
	{0x89CACCB00F442C1FULL, "634660347.ogg"},
	{0x89CAFA1326721F29ULL, "mag100_karras6034931.cl"},
	{0x89CB3F6FF2D9BCE8ULL, "vxht_0_hightown_n_9ut_0_0.vlm"},
	{0x89CBF717D677C20CULL, "6045609_m.fxe"},
	{0x89CC939F58FC0333ULL, "art\\characters\\playercharacter\\humanfemale\\hat.mao"},
	{0x89CCA92B865A76FEULL, "gen00fl_isabela6075226.cl"},
	{0x89CCC1C41A238397ULL, "lt_hangedman_n_3567.rml"},
	{0x89CCE4518F415074ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementaldamage\\physical\\extremeknockdownpose.gad"},
	{0x89CCF26A91BDCC30ULL, "mcr310_argument6182443.cl"},
	{0x89CD1D0855CFE32CULL, "vxlt_0_lowtown_docks_a3_d36_0_0.vlm"},
	{0x89CD673876C67C36ULL, "6118361_m.fxe"},
	{0x89CE7844137B0939ULL, "vxlt_gallowstemplar_f_47y_0_0.vlm"},
	{0x89CEE0247FFA2302ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb3a.msh"},
	{0x89CEE2247FFA26B8ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb3a.mmh"},
	{0x89CEF5CADE2A8AE3ULL, "gen00fl_sebastian6188694.cl"},
	{0x89D00643FDAD29D5ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_c.anb"},
	{0x89D0082F84ECD419ULL, "514670666.ogg"},
	{0x89D009E673C74BACULL, "vxlt_0_lowtown_docks_a3_apn_0_0.vlm"},
	{0x89D0655668D7A769ULL, "vxlt_0_lowtown_docks_a3_n_cuu_0_0.vlm"},
	{0x89D08B93D7458636ULL, "vxw_bonepitmining_a2_oa_0_0.vlm"},
	{0x89D153AF9185496EULL, "ca_cavern1_haunted_26037.rml"},
	{0x89D28A3304D20F63ULL, "vxdr_0_deeproads_cwo_0_0.vlm"},
	{0x89D2C8B70F21C0A3ULL, "vxw_woundedcoast_1al_0_0.vlm"},
	{0x89D3ADE88F321DC3ULL, "sf_strip.vsh.fxo"},
	{0x89D3E0806C4150A0ULL, "843138292.ogg"},
	{0x89D3E3B891ECD69AULL, "llac_lt_0_lowtown_a2_n_-65536_4.dds"},
	{0x89D3F3518F478852ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementaldamage\\physical\\extremeknockdownpose.ani"},
	{0x89D3F47A5177E0C4ULL, "hlac_ht_chantry_n_2534_44z.dds"},
	{0x89D44343FA1207C3ULL, "one110_meeran6090195.cl"},
	{0x89D457E2E30A0B23ULL, "mag360_alain6044630.cl"},
	{0x89D50CE5D6E98649ULL, "202664980.ogg"},
	{0x89D5D21E4BAB50DCULL, "genip_dagger.utp"},
	{0x89D5DF7DF8A07D90ULL, "qun120_arishok6051924.cl"},
	{0x89D643D0F8B7C34EULL, "one200_gold_followers6117248.cl"},
	{0x89D712DD8529C05CULL, "gen00fl_varric6076049.cl"},
	{0x89D74E9C31C05BF8ULL, "vxlt_undercitydungeon_180_0_0.vlm"},
	{0x89D754A84CE9152DULL, "644496181.ogg"},
	{0x89D807C9F113F822ULL, "601331978.ogg"},
	{0x89D8579848D00983ULL, "vrc321ip_spooky1_obj_1.utp"},
	{0x89D859EAE8F416B1ULL, "vxca_caverns2_8oj_0_0.vlm"},
	{0x89D87B2C1E917147ULL, "254977326.ogg"},
	{0x89D8E1F3DEB35B85ULL, "vxht_mansion_n_6mr_0_0.vlm"},
	{0x89D8E3483F902102ULL, "hlac_ht_0_hightown_n_-1_0x.dds"},
	{0x89D8E92480028743ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb3a.phy"},
	{0x89D95B8DBEECBBD7ULL, "302748151.ogg"},
	{0x89D977101B3EC47EULL, "vxdr_primevaltomb_6vc_0_1.vlm"},
	{0x89D9A5A9C154E12CULL, "6069661_m.fxe"},
	{0x89DA3B05519A4F80ULL, "6105707_m.fxe"},
	{0x89DA5E14B101A859ULL, "fen210_orana6133066.cl"},
	{0x89DB1CA0B00493AFULL, "vxlt_gallowstemplar_n_6q7_0_0.vlm"},
	{0x89DB8EE547C5FBACULL, "vxlt_0_lowtown_docks_a3_d47_0_0.vlm"},
	{0x89DBD91568BC94B6ULL, "6117261_m.fxe"},
	{0x89DC588BBE0BBB24ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grand.mmh"},
	{0x89DC5C8BBE0BC277ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grand.mao"},
	{0x89DC6E8BBE0BE0C6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grand.msh"},
	{0x89DCBF2C16B10C4CULL, "art\\characters\\weapons\\longswords\\w_lsw_ls09.msh"},
	{0x89DCC52C16B1163EULL, "art\\characters\\weapons\\longswords\\w_lsw_ls09.mmh"},
	{0x89DCD4A375E2349BULL, "w_encounter3_200_0.mmh"},
	{0x89DCE0A375E24938ULL, "w_encounter3_200_0.mao"},
	{0x89DCE7641BCEB6F9ULL, "qun111_ketojan6059229.cl"},
	{0x89DCF2A375E26785ULL, "w_encounter3_200_0.msh"},
	{0x89DD1D5EAFDA92CEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut4flat.msh"},
	{0x89DD2B5EAFDAAA1FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut4flat.mao"},
	{0x89DD375EAFDABEBCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut4flat.mmh"},
	{0x89DE0065AFD79F1BULL, "lt_0_lowtown_docks_a3_38_0_ll.mmh"},
	{0x89DEA0295FB616B2ULL, "415087319.ogg"},
	{0x89DF31B60B3F1A76ULL, "face_face_p9.psh.fxo.dx11.assembly"},
	{0x89DF4156C7DC4B3AULL, "qcr220_grey_wardens6060469.cl"},
	{0x89DF51A5E6E22AD8ULL, "vxca_cave_i6_0_0.vlm"},
	{0x89DF760F58AF1D9FULL, "6033166_m.fxe"},
	{0x89DFA7E73EF3475EULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzcot.phy"},
	{0x89DFC1370D78AD33ULL, "145935097.ogg"},
	{0x89DFE0AC58998B75ULL, "6043945_m.fxe"},
	{0x89E04A81B92624F8ULL, "dae000_dalish_apprentices6147591.cl"},
	{0x89E0FC0FF6C827E3ULL, "vxdr_0_deeproads_cln_0_0.vlm"},
	{0x89E13A940117D923ULL, "vxw_woundedcoast_1lo_0_0.vlm"},
	{0x89E18F5824A6F59CULL, "93600848.ogg"},
	{0x89E232E7E6057A82ULL, "hairalpha_rendertoshadowmap_p0.vsh.fxo.dx11"},
	{0x89E23D07F366F1F8ULL, "qcr000_arishok6059347.cl"},
	{0x89E24D17ED0357E0ULL, "872370841.ogg"},
	{0x89E26A49C74BC420ULL, "vxlt_gallowstemplar_f_4uf_0_0.vlm"},
	{0x89E2F04915C8EC6CULL, "art\\characters\\weapons\\staff\\textures\\w_stf_004a_0t.dds"},
	{0x89E32DE35DE24C73ULL, "6118759_m.fxe"},
	{0x89E35A9E86AD0409ULL, "159367563.ogg"},
	{0x89E39DBF569C6DD0ULL, "vxw_encounter1_d4_0_1.vlm"},
	{0x89E3F5F096814564ULL, "llac_ht_0_hightown_0_1.dds"},
	{0x89E46E90F7A639DCULL, "6108062_f.fxe"},
	{0x89E4BD71AF434C2FULL, "vxht_0_hightown_a3_bub_0_0.vlm"},
	{0x89E5296050B3E009ULL, "genip_station_potions.utp"},
	{0x89E57201FD0BB023ULL, "6152726_m.fxe"},
	{0x89E5DCBA930BD508ULL, "6175807_m.fxe"},
	{0x89E6C2722E8E62B1ULL, "vxht_0_hightown_bem_0_0.vlm"},
	{0x89E7187DB12CAFAAULL, "361083740.ogg"},
	{0x89E726F4D7AF69BBULL, "plt_cod_crt_wisp.plo"},
	{0x89E73A4985981BBAULL, "6191285_m.fxe"},
	{0x89E7CA60F81F6F49ULL, "vxw_encounter3_4b_0_0.vlm"},
	{0x89E7FBB1668B7069ULL, "vxlt_gallowscourtyard_n_3v7_0_0.vlm"},
	{0x89E8310CE826557FULL, "fex_ruin_wl_a_cn.msh"},
	{0x89E8370CE8265F21ULL, "fex_ruin_wl_a_cn.mmh"},
	{0x89E8BC92FB37FC4EULL, "139038329.ogg"},
	{0x89E8EA297BE378C7ULL, "vrc221_hugin6041110.cl"},
	{0x89E96C7D1BA02672ULL, "292788398.ogg"},
	{0x89EA2586A88B4C02ULL, "6131132_f.fxe"},
	{0x89EA54CE05426D93ULL, "llac_lt_0_lowtown_a2_n_-65535_6.dds"},
	{0x89EA69DA935BCF07ULL, "vxlt_gallowsprison_n_6k5_0_0.vlm"},
	{0x89EB1DAD4C4DB78FULL, "370880809.ogg"},
	{0x89EB2DB22C3C1BCFULL, "6143517_m.fxe"},
	{0x89EB7F79A854B12BULL, "vxw_bonepit_1rv_0_0.vlm"},
	{0x89EBF1F6F0F5946CULL, "vxht_bartrand_n_4ez_0_0.vlm"},
	{0x89EC1008A2FBB720ULL, "vxht_bartrand_n_29x_0_0.vlm"},
	{0x89EC5B6654510EE2ULL, "llac_lt_0_lowtown_docks_a3_-65537_2.dds"},
	{0x89EC5CE9B22D862AULL, "83274312.ogg"},
	{0x89EC77D2964F0274ULL, "296594625.ogg"},
	{0x89EC855DC7E6E9DEULL, "6169628_m.fxe"},
	{0x89EC871E53CDA77DULL, "232806824.ogg"},
	{0x89ED2499882DCCC5ULL, "854157139.ogg"},
	{0x89ED2A2B1500B018ULL, "mer120_hubert6017848.cl"},
	{0x89ED4177E2C8FA04ULL, "vxw_bonepitmining_a2_p2_0_0.vlm"},
	{0x89ED4E70A56F5050ULL, "6154576_m.fxe"},
	{0x89ED8CFB6DC9EDA2ULL, "6072300_m.fxe"},
	{0x89EEFA80855929DDULL, "115667713.ogg"},
	{0x89EEFC5E55CD4CA6ULL, "867814118.ogg"},
	{0x89EFACFC7E42C14DULL, "814697831.ogg"},
	{0x89EFEEA2D53F7200ULL, "vrc221_hugin6042859.cl"},
	{0x89F09C636B32A2BAULL, "706835807.ogg"},
	{0x89F0AF7750B38CCAULL, "gen00fl_fenris6126478.cl"},
	{0x89F0FAB2D6FCC9FBULL, "vxw_bonepitmining_a2_2aj_0_0.vlm"},
	{0x89F15459E36ECAF6ULL, "mag321_comtesse_de_launcet6043265.cl"},
	{0x89F1AD30A4EA03CCULL, "vxlt_0_lowtown_a2_e2v_0_0.vlm"},
	{0x89F1C1F373E08AF2ULL, "gen00fl_merrill6106322.cl"},
	{0x89F29C011649B22AULL, "mag311_emile_de_launcet6043456.cl"},
	{0x89F2BDAAB81CDFECULL, "17295107.ogg"},
	{0x89F30300AE4D4A95ULL, "gen00flst_fenris_gift2.stg"},
	{0x89F353FE00E3D5EAULL, "w_sundermount_158_0.mao"},
	{0x89F357FE00E3DC39ULL, "w_sundermount_158_0.mmh"},
	{0x89F35ACD48ADC7F0ULL, "100093355.ogg"},
	{0x89F361FE00E3ED27ULL, "w_sundermount_158_0.msh"},
	{0x89F548C7D2DC593AULL, "6075806_m.fxe"},
	{0x89F5A398C32066AFULL, "vxlt_0_lowtown_a2_n_hde_0_0.vlm"},
	{0x89F5F09B558F73D6ULL, "dae000_bartrand6206159.cl"},
	{0x89F6531FC2CE6FD7ULL, "534981836.ogg"},
	{0x89F6B079689AC83AULL, "vxlt_gallowscourtyard_n_75q_0_0.vlm"},
	{0x89F6FFA2AD677069ULL, "388974644.ogg"},
	{0x89F719AC666E7221ULL, "hlac_lt_0_lowtown_docks_-131075_1x.dds"},
	{0x89F7D46826568437ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he02.phy"},
	{0x89F7E81E9D99873DULL, "mag160_feynriel6033383.cl"},
	{0x89F80743FDCE7D16ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_c.mmh"},
	{0x89F83F11F9D1E9FFULL, "vxlt_shop_cb_0_0.vlm"},
	{0x89F8816CCD1988E2ULL, "905003788.ogg"},
	{0x89F952022EBD2B6BULL, "dre000_sibling_has_blight6036518.cl"},
	{0x89F9B1521B1A537AULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlookout.phy"},
	{0x89FA289423593335ULL, "genip_candle.utp"},
	{0x89FA2EE42150A351ULL, "828633454.ogg"},
	{0x89FA3ED6B8158B63ULL, "vxht_chantry_2eo_0_0.vlm"},
	{0x89FA833BB8E69F03ULL, "310116638.ogg"},
	{0x89FA903E3EC11DEFULL, "hlac_lt_0_lowtown_a3_n_c_-131072_7x.dds"},
	{0x89FACCA6D1487206ULL, "6164357_m.fxe"},
	{0x89FAD8985556D49BULL, "372901955.ogg"},
	{0x89FBDBA73CCF56E1ULL, "vxht_bartrand_n_5rw_0_0.vlm"},
	{0x89FBE6A9066970CBULL, "106786568.ogg"},
	{0x89FCD0D553A327DDULL, "vxht_keep_5eb_0_0.vlm"},
	{0x89FDD67275651DA1ULL, "6004306_m.fxe"},
	{0x89FDDFC9DAB458E3ULL, "vxdr_0_deeproads_cfp_0_0.vlm"},
	{0x89FE1E4DE5040A23ULL, "vxw_woundedcoast_1rq_0_0.vlm"},
	{0x89FF083A4E7E45A2ULL, "art\\vfx\\models\\vfx_vol_light_wide_6.mmh"},
	{0x89FF0889F6F37EC4ULL, "vxht_mansion2b_1w9_0_0.vlm"},
	{0x89FF0A3A4E7E4958ULL, "art\\vfx\\models\\vfx_vol_light_wide_6.msh"},
	{0x89FF941F170929D5ULL, "vxht_mansion_n_54n_0_0.vlm"},
	{0x8A001F0D96E01C4CULL, "vxlt_0_lowtown_a2_e5w_0_0.vlm"},
	{0x8A00862979261F18ULL, "6129231_m.fxe"},
	{0x8A0088EBE4F5DB02ULL, "hlac_ht_mansion2_818_20z.dds"},
	{0x8A0098736812E8BCULL, "153458336.ogg"},
	{0x8A00AFD486385BC7ULL, "6135646_m.fxe"},
	{0x8A01A69604EAF632ULL, "6067311_m.fxe"},
	{0x8A01BE0622A67FCCULL, "vxw_encounter3_dk_0_0.vlm"},
	{0x8A01F79060520376ULL, "dre000_carver6037603.cl"},
	{0x8A02165EAFF9A287ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_galojut4flat.phy"},
	{0x8A023BB875F86520ULL, "76852622.ogg"},
	{0x8A02CB88EE924341ULL, "6175983_m.fxe"},
	{0x8A02F50B877B7D3CULL, "6041083_m.fxe"},
	{0x8A032B20DF40471FULL, "371643959.ogg"},
	{0x8A03797518A18DB7ULL, "6112788_m.fxe"},
	{0x8A0390E5D29E1FA9ULL, "586247139.ogg"},
	{0x8A03C93F1D2687E9ULL, "one300_nuncio6077191.cl"},
	{0x8A03E17DA2B9E6F1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1b_1.mao"},
	{0x8A03FA7DF7D9B8D5ULL, "6172317_m.fxe"},
	{0x8A04156454E34FF7ULL, "docks_male_elven_worker_sta6175782.cl"},
	{0x8A0425731F55D4FDULL, "6092208_m.fxe"},
	{0x8A044AE48B7D8E6BULL, "422641419.ogg"},
	{0x8A04AD4747C26D21ULL, "gen00fl_aveline6139247.cl"},
	{0x8A051750606C072DULL, "mrl210_merrill6166785.cl"},
	{0x8A055870EB047BF6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_short.msh"},
	{0x8A0559AF5FE9BB5CULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernrock_d.dds"},
	{0x8A056270EB048CF4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_short.mmh"},
	{0x8A05645CFE2402C8ULL, "vxht_bartrand_n_5rr_0_0.vlm"},
	{0x8A056670EB049307ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_short.mao"},
	{0x8A05D9130AF4BC09ULL, "dae000_male_commoner2_mob6174427.cl"},
	{0x8A060CCB1CAA7A90ULL, "merrillhouseact2.stg"},
	{0x8A0683B6D6C729DDULL, "480772265.ogg"},
	{0x8A06F40C6BD22952ULL, "vxht_0_hightown_n_bwk_0_0.vlm"},
	{0x8A0706F7F5BB302CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edgebase.msh"},
	{0x8A070CF7F5BB3A1EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edgebase.mmh"},
	{0x8A0718F7F5BB4EB5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edgebase.mao"},
	{0x8A0752D4F26BFEBAULL, "gen_im_trs_chn_junk2_02.uti"},
	{0x8A07BD12ACA079DCULL, "6172855_m.fxe"},
	{0x8A07D1C3F23A33D2ULL, "vxlt_gallowstemplar_n_43j_0_0.vlm"},
	{0x8A07DA73E4A365CBULL, "vxdr_0_deeproads_evv_0_0.vlm"},
	{0x8A080FD39B4C70CDULL, "vrc221_bartrand6065294.cl"},
	{0x8A09159B1333D4A7ULL, "6112656_m.fxe"},
	{0x8A093E6DF353F944ULL, "one220_yevhen6099077.cl"},
	{0x8A097460E823C0E5ULL, "6066328_m.fxe"},
	{0x8A09A54C053BE417ULL, "llac_lt_backalley_g_1_4.dds"},
	{0x8A0A368AF1884FE0ULL, "vxw_encounter2_l8_0_0.vlm"},
	{0x8A0AE7A41AD4BA29ULL, "vxht_0_hightown_a2_c3o_0_1.vlm"},
	{0x8A0AEC58EBC06DF4ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_f_c.anb"},
	{0x8A0B25F61E1582CAULL, "6118664_m.fxe"},
	{0x8A0B421A4E22CA22ULL, "w_bonepitmining_a2_35_0.msh"},
	{0x8A0B441A4E22CDD8ULL, "w_bonepitmining_a2_35_0.mmh"},
	{0x8A0B501A4E22E27BULL, "w_bonepitmining_a2_35_0.mao"},
	{0x8A0B77BE8930E36AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic0col.mao"},
	{0x8A0B7BBE8930E9B9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic0col.mmh"},
	{0x8A0B85BE8930FAA7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic0col.msh"},
	{0x8A0BBC4A4E083329ULL, "art\\levels\\races\\design\\_textures\\prp_bodybagcoins_sa.dds"},
	{0x8A0BD9D31CD9D4BAULL, "dae021_sister_lorena_sta6142657.cl"},
	{0x8A0C213C6C7DAE9DULL, "vxdr_primevaltomb_7dn_0_0.vlm"},
	{0x8A0C6F96DF8B0907ULL, "672946375.ogg"},
	{0x8A0C8F5A633B2793ULL, "vxw_encounter1_b8_0_0.vlm"},
	{0x8A0CC952EDB98641ULL, "258094980.ogg"},
	{0x8A0CCB29CAF03343ULL, "6087368_f.fxe"},
	{0x8A0D024743C14250ULL, "979352829.ogg"},
	{0x8A0D18DE8DAC0228ULL, "6149336_m.fxe"},
	{0x8A0D239EDF92B30EULL, "hlac_lt_undercitydungeon_-1_16x.dds"},
	{0x8A0D8BA5A31CBF0DULL, "vxht_0_hightown_a2_ay2_0_0.vlm"},
	{0x8A0D9B97731E7312ULL, "dae000_denier6162676.cl"},
	{0x8A0DBD0B37C44913ULL, "dae011_badpoet_sta6146922.cl"},
	{0x8A0E4FFE219562E9ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_006a_s.dds"},
	{0x8A0EE7B45D5ADC89ULL, "ht_brothel_n_3373.rml"},
	{0x8A0EF0A8D4BB5313ULL, "rdr111dg_martin6054493.cl"},
	{0x8A0F441A4E26B6EBULL, "w_bonepitmining_a2_35_0.rml"},
	{0x8A0FE65FD6BA2D0FULL, "fire_shot.mao"},
	{0x8A1003CE4DE501D1ULL, "6093801_m.fxe"},
	{0x8A102FE3149C984CULL, "vxw_encounter3_ah_0_0.vlm"},
	{0x8A104F402C665D5CULL, "vxlt_0_lowtown_a2_ffu_0_0.vlm"},
	{0x8A1060B7F038B204ULL, "docks_male_elven_worker_sta6175839.cl"},
	{0x8A10D1135170D810ULL, "fen310_fenris6118225.cl"},
	{0x8A10FA49C93981E5ULL, "lt_0_lowtown_docks_n_29_0.anb"},
	{0x8A11234F159241F3ULL, "mag110_arianni6016895.cl"},
	{0x8A1170F3D945CEA5ULL, "158760224.ogg"},
	{0x8A11EDFA7878B57EULL, "501715355.ogg"},
	{0x8A122DB5B74A529DULL, "vxlt_0_lowtown_docks_n_bdy_0_1.vlm"},
	{0x8A127E2BB790D0CBULL, "vxlt_warehouse_185_0_0.vlm"},
	{0x8A128E16693DD2ACULL, "dae000_orana6149035.cl"},
	{0x8A1290A1C79AC23AULL, "gencr_raider_2w.utc"},
	{0x8A1295E3D5B73AC4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4_0.pwk"},
	{0x8A12976840D5766CULL, "356505428.ogg"},
	{0x8A12977677C900ABULL, "6085941_m.fxe"},
	{0x8A12E1FA9B3DAF13ULL, "gen00fl_varric6078701.cl"},
	{0x8A13213EAD2BE448ULL, "gen00fl_isabela6079039.cl"},
	{0x8A1361799A021F64ULL, "6151166_m.fxe"},
	{0x8A1398C6CF7BA1FFULL, "352645501.ogg"},
	{0x8A13ED1226FEB124ULL, "1047140193.ogg"},
	{0x8A141C54DE0447F5ULL, "isa311_isabela6092264.cl"},
	{0x8A14485FBFF81FABULL, "476396778.ogg"},
	{0x8A147FE8901FC881ULL, "6129155_m.fxe"},
	{0x8A1498C78E8A57ABULL, "vxdr_0_deeproads_bwo_0_1.vlm"},
	{0x8A15170C386B9EDFULL, "mcr371_hostile_aveline6132473.cl"},
	{0x8A15523BF338A977ULL, "vxht_bartrand_n_3i6_0_0.vlm"},
	{0x8A15913238DB4E76ULL, "hlac_lt_darktown_-1_67y.dds"},
	{0x8A15A8E34A280E15ULL, "6168159_f.fxe"},
	{0x8A166208ACB01D57ULL, "vxlt_0_lowtown_docks_bb3_0_0.vlm"},
	{0x8A16A99AF569B6BFULL, "vxht_bartrand_n_4eu_0_0.vlm"},
	{0x8A17412CC92108B9ULL, "144849128.ogg"},
	{0x8A178F65AEDAC0EAULL, "243938010.ogg"},
	{0x8A17A3C7468A3294ULL, "zz_fen_debug6215253.cl"},
	{0x8A181CFC34783549ULL, "art\\vfx\\materials\\chunky\\vfx_c_flat.mao"},
	{0x8A18970EB8AB4384ULL, "vxw_bonepitmining_a2_a1_0_0.vlm"},
	{0x8A18BF6120BB87E1ULL, "vxht_bartrand_n_5xu_0_0.vlm"},
	{0x8A18CB58EBCC72B7ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_f_c.mmh"},
	{0x8A191FC638C5F639ULL, "vxht_mansion2_3dk_0_0.vlm"},
	{0x8A193DBEFE1EE833ULL, "vxdr_0_deeproads_9g9_0_0.vlm"},
	{0x8A1965883876FB76ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mc.mao"},
	{0x8A19718838770F15ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mc.mmh"},
	{0x8A1973883877128BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mc.msh"},
	{0x8A1997D3100B9B4AULL, "6117764_m.fxe"},
	{0x8A19B47CB098BDF4ULL, "qun131_taskmaster6059003.cl"},
	{0x8A19F4C45BD9AB1FULL, "mag370_meredith6064856.cl"},
	{0x8A1B138A079013A8ULL, "6169738_m.fxe"},
	{0x8A1C32ABC6925B77ULL, "dae210_gate_guard6085047.cl"},
	{0x8A1C7B930C84EE3AULL, "329611514.ogg"},
	{0x8A1C95A0B1847569ULL, "store_dae110_armor.utm"},
	{0x8A1CA1382E8877B7ULL, "6087239_m.fxe"},
	{0x8A1CB1E57DE1D3BEULL, "vxw_encounter2_4n_0_0.vlm"},
	{0x8A1CBD81E0A727B1ULL, "gen00fl_anders6113024.cl"},
	{0x8A1CFA91AE40818FULL, "vxht_keep_3xm_0_0.vlm"},
	{0x8A1D824B287E528EULL, "lla_ht_0_hightown_a3_1_0.dds"},
	{0x8A1D9C2C16E867D9ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls09.phy"},
	{0x8A1E66A91128BA8BULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v2.impa.ani"},
	{0x8A1E73D2F9ADCA81ULL, "follower_banter6173326.cl"},
	{0x8A1E84E5771BAE2FULL, "vxht_0_hightown_a3_ban_0_0.vlm"},
	{0x8A1EF3F8F35414CAULL, "6136909_m.fxe"},
	{0x8A1FBDC7D133A1FDULL, "w_sundermount_23_0.mao"},
	{0x8A1FC1C7D133A836ULL, "w_sundermount_23_0.mmh"},
	{0x8A1FCBC7D133B934ULL, "w_sundermount_23_0.msh"},
	{0x8A20217DCAEFF9BFULL, "6135884_m.fxe"},
	{0x8A2037E50A276691ULL, "dr_primevaltomb_mmap.dds"},
	{0x8A204F5DDBC6A730ULL, "dae000_merchant_dm_mob6174305.cl"},
	{0x8A20B17F2F773ED0ULL, "zz_mag_debug6089450.cl"},
	{0x8A21066CAA666C43ULL, "vxdr_0_deeproads_cvq_0_1.vlm"},
	{0x8A2130F0637899F8ULL, "promoitem_col.gda"},
	{0x8A21662CDCBB0C59ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_4y.dds"},
	{0x8A219CD014E65D77ULL, "vxht_0_hightown_a3_cbj_0_0.vlm"},
	{0x8A21EA5688609353ULL, "211060441.ogg"},
	{0x8A224816E21033C8ULL, "vxht_bartrand_n_5xp_0_0.vlm"},
	{0x8A22FEE79790A05AULL, "hlac_w_bonepit_131072_0z.dds"},
	{0x8A23C50E922A734FULL, "ser221_party_comments6116129.cl"},
	{0x8A23DE0EF742178BULL, "ave121ip_duty_roster.ncs"},
	{0x8A246BB135109AA1ULL, "6183968_m.fxe"},
	{0x8A2531BF25DA6DE5ULL, "602451520.ogg"},
	{0x8A2538B218BD54EFULL, "one220_yevhen6103021.cl"},
	{0x8A255F655BE4D801ULL, "gen00fl_merrill6162080.cl"},
	{0x8A25946D8DF7D4E3ULL, "vxht_chantry_2tp_0_0.vlm"},
	{0x8A259F5A65890779ULL, "vxca_caverns2_80c_0_0.vlm"},
	{0x8A26F42AEBB051C5ULL, "qun231cr_varnell_minion3.utc"},
	{0x8A2708EBAAA15C04ULL, "vxw_bonepitmining_a2_l6_0_0.vlm"},
	{0x8A278768659289F3ULL, "vxlt_undercityencounter_wh_0_0.vlm"},
	{0x8A279847F9442822ULL, "33526452.ogg"},
	{0x8A27D093E48976FFULL, "169142321.ogg"},
	{0x8A2886487D9D6F89ULL, "art\\characters\\playercharacter\\dwarfmale\\hlm_meda.mao"},
	{0x8A28F784AA3A94E5ULL, "llac_w_sundermount_0_4.dds"},
	{0x8A291D48986A8290ULL, "6095837_m.fxe"},
	{0x8A2920F2A522D151ULL, "vxht_brothel_n_5l9_0_0.vlm"},
	{0x8A293384737E207FULL, "160707736.ogg"},
	{0x8A2AF6C5A5292046ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue2_da.dds"},
	{0x8A2B66AF6033D52CULL, "6178394_m.fxe"},
	{0x8A2B672D97BFFAD5ULL, "nat380_temmerin6067349.cl"},
	{0x8A2BE380ECA53B7BULL, "vxht_brothel_n_3uv_0_0.vlm"},
	{0x8A2C02B3E49DE02BULL, "6114055_f.fxe"},
	{0x8A2C3F4BB1E9AB32ULL, "hlac_ht_keep_3733_0y.dds"},
	{0x8A2C4EA91134CF02ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v2.impa.evt"},
	{0x8A2C5477C513E9D9ULL, "hlac_lt_0_lowtown_a2_n_c_131071_0z.dds"},
	{0x8A2C5A0AB85A0C10ULL, "vxht_mansion2b_3co_0_0.vlm"},
	{0x8A2CC9B79B03CE58ULL, "vxdr_0_deeproads_cmr_0_0.vlm"},
	{0x8A2D9289D7B3347EULL, "zz_vrc_debug6200200.cl"},
	{0x8A2DC2D8B58D012FULL, "6076969_m.fxe"},
	{0x8A2DCB4C11A4C15CULL, "llac_w_encounter3_-65539_0.dds"},
	{0x8A2E12FF052A5681ULL, "one140cr_charade6137054.cl"},
	{0x8A2EFA3A1EC60A44ULL, "character_character_p96.psh.fxo.dx11.assembly"},
	{0x8A2F61E59B7D01CBULL, "vxlt_warehouse_127_0_0.vlm"},
	{0x8A2FB894B0687F5BULL, "483033059.ogg"},
	{0x8A30015F23B71BEBULL, "vxca_caverns1_mines_ehf_0_0.vlm"},
	{0x8A307BFF6E843C60ULL, "6217370_m.fxe"},
	{0x8A3118FECDB4207FULL, "352119387.ogg"},
	{0x8A315D70B47DD6ABULL, "hlac_ht_chantry_2534_48x.dds"},
	{0x8A3163074B72BC6EULL, "6173007_m.fxe"},
	{0x8A317EB8448FB27DULL, "542846646.ogg"},
	{0x8A31920C6490E991ULL, "vxlt_gallowstemplar_375_0_0.vlm"},
	{0x8A31A6382CCE2225ULL, "147251371.ogg"},
	{0x8A31EF980DA1288DULL, "llac_ca_caverns2prc_5802_3.dds"},
	{0x8A3221BE0BE18E87ULL, "6034668_m.fxe"},
	{0x8A323724FCF7C99DULL, "hlac_dr_primevaltomb_9621_5z.dds"},
	{0x8A325244DF13322EULL, "6024978_m.fxe"},
	{0x8A330A3007E1B1ADULL, "narrative_banter6165536.cl"},
	{0x8A331FF538210B89ULL, "6129035_f.fxe"},
	{0x8A333191AC2D5994ULL, "932238458.ogg"},
	{0x8A334DA9113AEC15ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v2.impa.gad"},
	{0x8A3385A026CF2381ULL, "150033100.ogg"},
	{0x8A342FF6FC15414CULL, "vxht_0_hightown_n_bes_0_0.vlm"},
	{0x8A34375FB180BE8DULL, "vxca_cavern1_haunted_dpq_0_0.vlm"},
	{0x8A3479D0C6197D81ULL, "gen_im_trs_qun_hvy_hlm_r1.uti"},
	{0x8A34A727D2D9B93BULL, "w_woundedcoast_136_0.msh"},
	{0x8A34B141DEDE27FDULL, "gen00fl_carver6108873.cl"},
	{0x8A34B927D2D9D786ULL, "w_woundedcoast_136_0.mao"},
	{0x8A34C527D2D9EC25ULL, "w_woundedcoast_136_0.mmh"},
	{0x8A34D66F3DC8CC4DULL, "lit_01.mao"},
	{0x8A34FA71A54F8F69ULL, "404874477.ogg"},
	{0x8A350182DC4EC1F3ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.anb"},
	{0x8A35D1D5688B1683ULL, "6033792_f.fxe"},
	{0x8A35F9455788A273ULL, "vxlt_undercityencounter_rk_0_0.vlm"},
	{0x8A361D8B54EFBCB1ULL, "w_bonepitdestroyed_a3_214_0.mmh"},
	{0x8A36298B54EFD152ULL, "w_bonepitdestroyed_a3_214_0.mao"},
	{0x8A36378B54EFE96FULL, "w_bonepitdestroyed_a3_214_0.msh"},
	{0x8A364C9F0CF4A233ULL, "6021213_m.fxe"},
	{0x8A36501A2E1D5A80ULL, "hlac_lt_gallowstemplar_n_7198_6x.dds"},
	{0x8A367C702F33AA2FULL, "hlac_lt_0_lowtown_a2_n_-4_0z.dds"},
	{0x8A3694C74E7FAEB3ULL, "lgt320_mis_resp_swordhunter6150343.cl"},
	{0x8A36B59AC81A0E26ULL, "dre000_sibling_has_blight6036492.cl"},
	{0x8A36C45D4B24266CULL, "482109910.ogg"},
	{0x8A36D606453D911BULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallowsb.mmh"},
	{0x8A36E206453DA5B8ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallowsb.mao"},
	{0x8A36E6AD681CE4B7ULL, "vxlt_gallowscourtyard_7c1_0_0.vlm"},
	{0x8A36F406453DC405ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallowsb.msh"},
	{0x8A36F807A0A8098CULL, "334690593.ogg"},
	{0x8A371B1CB3163ACFULL, "vxlt_hangedman_1ti_0_0.vlm"},
	{0x8A3785E7F4693CB5ULL, "vxlt_gallowscourtyard_n_6xp_0_0.vlm"},
	{0x8A379AA16A5B1958ULL, "vxht_chantry_2it_0_0.vlm"},
	{0x8A381196418CC8FAULL, "vxht_mansion2b_1gg_0_0.vlm"},
	{0x8A38DA548B7CE7B3ULL, "347712412.ogg"},
	{0x8A39711F3F147CB7ULL, "6165254_m.fxe"},
	{0x8A39A80BB5C53897ULL, "6102694_f.fxe"},
	{0x8A39D45CA3022005ULL, "6036332_m.fxe"},
	{0x8A39D9E22B6244CFULL, "vxht_mansion_n_5mk_0_0.vlm"},
	{0x8A3A0DBD285D3057ULL, "vxht_keep_n_4i6_0_0.vlm"},
	{0x8A3A23CD4A4FAE98ULL, "6174462_m.fxe"},
	{0x8A3A69C06BD3A645ULL, "ca_caverns2_5802_ll.mmh"},
	{0x8A3A951DA414DB07ULL, "vxlt_0_lowtown_n_fid_0_0.vlm"},
	{0x8A3AAF8433FA0740ULL, "gen00fl_anders6125941.cl"},
	{0x8A3ACBE7AA502490ULL, "vxht_mansion2b_3fl_0_0.vlm"},
	{0x8A3AD4EA992B88D6ULL, "mcr310_argument6074353.cl"},
	{0x8A3B3B948CF9E6D8ULL, "vxdr_0_deeproads_cjs_0_0.vlm"},
	{0x8A3B97427A65A4AFULL, "831325425.ogg"},
	{0x8A3C38DEDA6EB1F2ULL, "6016701_f.fxe"},
	{0x8A3C45F3C529F131ULL, "135197249.ogg"},
	{0x8A3CA37DFC6EABF5ULL, "hla_ht_0_hightown_a3_65538_1z.dds"},
	{0x8A3D6BC6488847B9ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_8y.dds"},
	{0x8A3D739D6632D40FULL, "783834444.ogg"},
	{0x8A3D863F692E4ECBULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_180.evt"},
	{0x8A3D959C66D5C570ULL, "80878805.ogg"},
	{0x8A3DE91B036DFFDEULL, "294851500.ogg"},
	{0x8A3E07889F039E43ULL, "vxlt_gallowstemplar_f_5pg_0_0.vlm"},
	{0x8A3E501AF6D53B86ULL, "192891596.ogg"},
	{0x8A3EB5682692522CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he02.mmh"},
	{0x8A3EBB6826925C1EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he02.msh"},
	{0x8A3F703246C3286EULL, "692626883.ogg"},
	{0x8A3FB7B446F4D66DULL, "vxlt_gallowscourtyard_n_4wc_0_0.vlm"},
	{0x8A403A52F894AA51ULL, "gen00fl_aveline6139908.cl"},
	{0x8A403C088EFC8A92ULL, "llac_ht_bartrand_n_2995_0.dds"},
	{0x8A4101173B757F19ULL, "6167309_f.fxe"},
	{0x8A41257F89492EF8ULL, "vxlt_0_lowtown_a3_n_c_hx5_0_0.vlm"},
	{0x8A414861FC838BBAULL, "251409540.ogg"},
	{0x8A416E4C004080C5ULL, "llac_lt_0_lowtown_n_1_0.dds"},
	{0x8A419348E9F5CD3DULL, "follower_banter6141925.cl"},
	{0x8A41E6B45D8634A9ULL, "ht_brothel_n_3373.anb"},
	{0x8A424B0B09EB4BDDULL, "vxlt_gallowstemplar_454_0_0.vlm"},
	{0x8A424BAAA002F87EULL, "mcr310_argument6072676.cl"},
	{0x8A42A27C6E12A1DFULL, "vxlt_gallowstemplar_f_4qi_0_0.vlm"},
	{0x8A42E30EF75CD176ULL, "ave121ip_duty_roster.utp"},
	{0x8A431E82DC5B305CULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.mmh"},
	{0x8A432E8A7AA2BCBDULL, "vxlt_gallowsprison_n_6ln_0_0.vlm"},
	{0x8A43C2B224654324ULL, "gen00fl_merrill6106089.cl"},
	{0x8A44757036D9AFFBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_leftrailing.msh"},
	{0x8A44877036D9CE46ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_leftrailing.mao"},
	{0x8A44937036D9E2E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_leftrailing.mmh"},
	{0x8A449F3F6934978CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_180.gad"},
	{0x8A44CBA3A980B8DCULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic.mao"},
	{0x8A44D7A3A980CD7FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic.mmh"},
	{0x8A44DDA3A980D721ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic.msh"},
	{0x8A45588A5A12FD37ULL, "vxlt_gallowscourtyard_7d0_0_0.vlm"},
	{0x8A458CF9A50C534FULL, "vxlt_hangedman_1sh_0_0.vlm"},
	{0x8A45A5F96852B51BULL, "vxlt_0_lowtown_n_cow_0_0.vlm"},
	{0x8A45D07C9CCA8C85ULL, "llac_lt_0_lowtown_docks_-65540_0.dds"},
	{0x8A45D1DBD8B13822ULL, "mag110_sailor6107411.cl"},
	{0x8A4604DC2CAD6441ULL, "vxlt_darktown_50r_0_0.vlm"},
	{0x8A4682815BA5347DULL, "6118267_m.fxe"},
	{0x8A46FBC2F514C31FULL, "rdr111dg_assistant6016647.cl"},
	{0x8A4791F0363E7C01ULL, "vxht_keep_4hl_0_0.vlm"},
	{0x8A47CC2EB46894EEULL, "vxlt_gallowstemplar_n_5a9_0_0.vlm"},
	{0x8A47E4A3565F6045ULL, "dre000_vampire6034279.cl"},
	{0x8A47ECECF87E020EULL, "vxlt_house_df_0_0.vlm"},
	{0x8A4850288422CB0FULL, "vxht_keep_3in_0_0.vlm"},
	{0x8A485AD87E0851BEULL, "vxw_blightlands_iv_0_0.vlm"},
	{0x8A4881AE374D17C2ULL, "vxca_caverns1_mines_i2a_0_0.vlm"},
	{0x8A48B8D41B7AEF94ULL, "art\\cinematics\\act3\\romances\\anders\\cs_anders_andersrom.gad"},
	{0x8A49B4BC80C37D84ULL, "6116202_m.fxe"},
	{0x8A4A1A9A0C4AD903ULL, "920439923.ogg"},
	{0x8A4A622C00F47FC9ULL, "6155718_m.fxe"},
	{0x8A4A71F1D357FA3FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\til_lowtown_main_n.dds"},
	{0x8A4A90E831F65E17ULL, "prp_str_slime_s.dds"},
	{0x8A4A9E3F6939018AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_180.ani"},
	{0x8A4AFB49C96AFBEDULL, "lt_0_lowtown_docks_n_29_0.rml"},
	{0x8A4B30BD95F1E47AULL, "hlac_lt_undercitydungeon_-1_28y.dds"},
	{0x8A4B48730A13B749ULL, "vxlt_0_lowtown_a2_ejg_0_0.vlm"},
	{0x8A4BC80EE672BEB9ULL, "465782236.ogg"},
	{0x8A4BF3A9794227DAULL, "vxca_cavern1_haunted_k2q_0_2.vlm"},
	{0x8A4BF4595FB32CE1ULL, "6084005_m.fxe"},
	{0x8A4C2E05317D0FBBULL, "vxdr_0_deeproads_d3h_0_0.vlm"},
	{0x8A4CE7FA93D3F34CULL, "453762148.ogg"},
	{0x8A4D9DADB7F27D48ULL, "vxht_bartrand_n_5io_0_0.vlm"},
	{0x8A4DA947A0419A49ULL, "vxht_keep_n_6hi_0_0.vlm"},
	{0x8A4E07081B75DF2CULL, "mag121_viveka6083380.cl"},
	{0x8A4E407C0956FCF3ULL, "6101964_m.fxe"},
	{0x8A4E718146F573FFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v2.gad"},
	{0x8A4EF3F6E2EFE951ULL, "376673171.ogg"},
	{0x8A4F2C3AC48A022CULL, "666342863.ogg"},
	{0x8A4F40BB565FC8ADULL, "799119823.ogg"},
	{0x8A4F74B4922A75DEULL, "6118699_m.fxe"},
	{0x8A4FC00069F03D41ULL, "6152229_m.fxe"},
	{0x8A4FC7D41B812872ULL, "art\\cinematics\\act3\\romances\\anders\\cs_anders_andersrom.ani"},
	{0x8A503DAD0F6F0AB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v5.ani"},
	{0x8A50D40908678679ULL, "52317814.ogg"},
	{0x8A511B19956CEF8DULL, "vxca_cavern1_haunted_dno_0_0.vlm"},
	{0x8A514908FFA34743ULL, "608284313.ogg"},
	{0x8A5180E6D712237DULL, "plt_genpt_party_rand_banter.plo"},
	{0x8A51A0676C98D53DULL, "vxlt_gallowsprison_n_6im_0_0.vlm"},
	{0x8A52923409F4BD9FULL, "181518838.ogg"},
	{0x8A52AFF33C20816FULL, "68527403.ogg"},
	{0x8A52BA00233B9341ULL, "vxw_woundedcoast_1f7_0_0.vlm"},
	{0x8A536652C2A4634FULL, "123737457.ogg"},
	{0x8A536F7FCBD59AB8ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_5x.dds"},
	{0x8A53970BD0841242ULL, "vxht_mansion2b_3br_0_0.vlm"},
	{0x8A54485ACA25BAF7ULL, "sst_ss_qunari_arishok.gda"},
	{0x8A5461BEB17E9B65ULL, "llac_w_bonepitdestroyed_a3_65536_0.dds"},
	{0x8A54668146F9CCECULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v2.evt"},
	{0x8A5476B91EA37CC1ULL, "vxlt_darktown_55q_0_0.vlm"},
	{0x8A548DFEABC655B1ULL, "vxlt_gallowstemplar_5yy_0_0.vlm"},
	{0x8A5494BBBAD5FE11ULL, "6118465_m.fxe"},
	{0x8A54AA521B66F2A3ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlookout.msh"},
	{0x8A54B8521B670AFDULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlookout.mmh"},
	{0x8A54BC521B67110EULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlookout.mao"},
	{0x8A54F67A68F5B90DULL, "dae000_oldmangossip6174912.cl"},
	{0x8A55231E0E280DB3ULL, "6148228_m.fxe"},
	{0x8A5540449836D2FBULL, "art\\characters\\playercharacter\\humanfemale\\boo_masd.msh"},
	{0x8A5552449836F146ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masd.mao"},
	{0x8A555E44983705E5ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masd.mmh"},
	{0x8A5596EBEA0B4F59ULL, "vxlt_darktown_5gs_0_0.vlm"},
	{0x8A559B46AA125590ULL, "vxht_chantry_354_0_1.vlm"},
	{0x8A55D4668005D741ULL, "mag110_arianni6020485.cl"},
	{0x8A55F47F12CC02B7ULL, "qun120_arishok_intro.stg"},
	{0x8A5612182D260322ULL, "zz_followers_banter6157925.cl"},
	{0x8A56CCB56FFE6A3EULL, "vxw_blightlands_rw_0_0.vlm"},
	{0x8A573AAD0F752416ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v5.gad"},
	{0x8A578902B723EC99ULL, "vxlt_0_lowtown_a2_gpg_0_0.vlm"},
	{0x8A5817FC9747DC68ULL, "zz_followers_banter6159648.cl"},
	{0x8A588B9489358CE7ULL, "vxlt_gallowstemplar_n_3ro_0_0.vlm"},
	{0x8A58A5EE1FEA661AULL, "6082181_m.fxe"},
	{0x8A58ABCE32DCC726ULL, "549526926.ogg"},
	{0x8A58E5840FABF734ULL, "583545060.ogg"},
	{0x8A590170F781F5FEULL, "w_bonepitmining_a2_11_0_lc.mmh"},
	{0x8A59042F88563983ULL, "plt_cod_prc_basketlostsocks.plo"},
	{0x8A591D73433AADF1ULL, "gen00fl_dog6145219.cl"},
	{0x8A59823DB3E03AF4ULL, "vxht_0_hightown_a2_b5x_0_0.vlm"},
	{0x8A59B05C2A6C84DDULL, "seb201_lady_harimann6197258.cl"},
	{0x8A59D0E1B3881EA7ULL, "93566886.ogg"},
	{0x8A59DEC6C45188F1ULL, "vxlt_gallowstemplar_f_47a_0_0.vlm"},
	{0x8A5A6DAEDBA45BD0ULL, "sse_ss_slavers_m_1.gda"},
	{0x8A5B2FE911D74FC2ULL, "583130804.ogg"},
	{0x8A5B935B698CF321ULL, "204144660.ogg"},
	{0x8A5BA4312DB29AEEULL, "vxlt_0_lowtown_docks_n_bdo_0_0.vlm"},
	{0x8A5BEC5BD621E8DFULL, "479127173.ogg"},
	{0x8A5C1B249237B2C9ULL, "vxht_keep_n_6ch_0_0.vlm"},
	{0x8A5C9BC88D543513ULL, "qun160_vashothlead6064659.cl"},
	{0x8A5D4B06AFE37B8BULL, "6095179_m.fxe"},
	{0x8A5D5D39AA5187AFULL, "vxca_cavern1_haunted_im9_0_0.vlm"},
	{0x8A5D936CE9534751ULL, "w_bonepit_30_0_lc.mmh"},
	{0x8A5DE69184208E5EULL, "6117799_m.fxe"},
	{0x8A5DFB26DA8DF737ULL, "456956172.ogg"},
	{0x8A5E55AD0F7B7175ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\attk_v5.evt"},
	{0x8A5F064955B83BB3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_action.ani"},
	{0x8A5F2FC494922300ULL, "40240417.ogg"},
	{0x8A5F3C271C8B10AEULL, "974860369.ogg"},
	{0x8A5F6793DE85B4C5ULL, "hlac_dr_0_deeproads_5845_19y.dds"},
	{0x8A5FD911FBC7FAEBULL, "6073891_f.fxe"},
	{0x8A60240B6A065D9BULL, "6037320_m.fxe"},
	{0x8A604A30556D0BCEULL, "177387188.ogg"},
	{0x8A606170EB5136EFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wi_short.phy"},
	{0x8A60737847DCA771ULL, "dae011_badpoet_sta6145172.cl"},
	{0x8A6090C9E6B5A536ULL, "hlac_lt_gallowstemplar_n_2391_6z.dds"},
	{0x8A60A4CA15CFA84BULL, "6055607_fenriskills.cut"},
	{0x8A60B22A249110DBULL, "850516761.ogg"},
	{0x8A612BDD1531ABC1ULL, "vxw_woundedcoast_1a6_0_0.vlm"},
	{0x8A6162ED6E3BCA36ULL, "6112730_m.fxe"},
	{0x8A624E814705E145ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v2.ani"},
	{0x8A62AD88C216360AULL, "follower_barks6153656.cl"},
	{0x8A63093A46015845ULL, "vxht_mansion2b_1gl_0_0.vlm"},
	{0x8A6310FFBBB8379DULL, "vxlt_0_lowtown_a3_n_c_gnu_0_0.vlm"},
	{0x8A63663B3F91657DULL, "6112067_m.fxe"},
	{0x8A6375CDA7CD21EAULL, "gen00fl_aveline6004160.cl"},
	{0x8A637CA39359C99BULL, "6069752_m.fxe"},
	{0x8A6394FB001E2633ULL, "6145128_m.fxe"},
	{0x8A63DC5A08F844F0ULL, "zz_dre_debug6168212.cl"},
	{0x8A645C6E5AD9AB50ULL, "hlac_ht_keep_3732_19x.dds"},
	{0x8A64EF872D57E4D9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medc_s.dds"},
	{0x8A64FDA8E9604090ULL, "6085425_m.fxe"},
	{0x8A6529FBCEBA5E3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal.ani"},
	{0x8A653E9261F482BEULL, "vxw_blightlands_wh_0_0.vlm"},
	{0x8A65D3BEF9EDC173ULL, "zz_amb_test.are"},
	{0x8A65EA813CC94D2CULL, "6173851_f.fxe"},
	{0x8A65FF15C95EFA79ULL, "394846443.ogg"},
	{0x8A6613248564DA2CULL, "25345577.ogg"},
	{0x8A66210C6E10CECDULL, "6095339_m.fxe"},
	{0x8A66CFF52CD70CBFULL, "173723565.ogg"},
	{0x8A670B69D338B1C9ULL, "art\\characters\\playercharacter\\elfmale\\em_har_fls.phy"},
	{0x8A6717CB11E07E9AULL, "6085081_m.fxe"},
	{0x8A67343BE04956D1ULL, "599688387.ogg"},
	{0x8A6771A2006DA982ULL, "ave200_donnic6061399.cl"},
	{0x8A67C76AC85BAA1EULL, "one350ip_to_zevran_cave.utp"},
	{0x8A68086545B2BB42ULL, "mer221_sabin6080102.cl"},
	{0x8A684F624A36EEE0ULL, "1066960071.ogg"},
	{0x8A685086CA3D47DCULL, "93266787.ogg"},
	{0x8A685DD18693CA4EULL, "180413072.ogg"},
	{0x8A688E0B475A8D55ULL, "vxw_encounter2_camp_pe_0_0.vlm"},
	{0x8A689165FA1F64F2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgta_0d.dds"},
	{0x8A69EB1F340CBC82ULL, "vxca_caverns2prc_9zf_0_0.vlm"},
	{0x8A6A0242530D14E1ULL, "6027499_m.fxe"},
	{0x8A6A32E4C100859CULL, "794172169.ogg"},
	{0x8A6A33319E372AE2ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masf.phy"},
	{0x8A6A6ADC75CFC642ULL, "hlac_lt_0_lowtown_docks_a3_131071_6x.dds"},
	{0x8A6AA9E82BD48CECULL, "mag110_samson6016174.cl"},
	{0x8A6AF1EB07305719ULL, "hlac_lt_0_lowtown_a3_n_c_65536_4y.dds"},
	{0x8A6AFC3EC5E608FAULL, "vxlt_0_lowtown_a2_n_gnu_0_0.vlm"},
	{0x8A6B2E67DB9110A0ULL, "620418157.ogg"},
	{0x8A6BB140BD395E20ULL, "lgt110ar_gang_hideout_ht.lst"},
	{0x8A6BCF169C47A02FULL, "vxca_cavern1_haunted_ij8_0_0.vlm"},
	{0x8A6BDE0098F8B1A0ULL, "633274023.ogg"},
	{0x8A6C3BF9D5C66870ULL, "art\\levels\\atmosphere\\sky\\sb03_perlin.dds"},
	{0x8A6C841683E6130BULL, "dae000_arishok6080494.cl"},
	{0x8A6C8E89BA8A9360ULL, "vxht_brothel_n_6c6_0_0.vlm"},
	{0x8A6CCFBEF9F3D88EULL, "zz_amb_test.cnv"},
	{0x8A6CEAEEBCA763CCULL, "follower_banter6154098.cl"},
	{0x8A6CF04011E65F67ULL, "735227430.ogg"},
	{0x8A6CFE4955C46BEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_action.evt"},
	{0x8A6D493E6185D764ULL, "mcr310_sibling6076334.cl"},
	{0x8A6DACCAAF2DD25BULL, "534926729.ogg"},
	{0x8A6DFBEDA1C9420AULL, "zz_prmpt_debug6198147.cl"},
	{0x8A6EC367887E5ABCULL, "script.ldf"},
	{0x8A6EE03CBF4634F8ULL, "elf_male_heads.gda"},
	{0x8A6F0DA893DD5E00ULL, "20640572.ogg"},
	{0x8A6F6358C0EC37D9ULL, "lgt210_magu_selby6142025.cl"},
	{0x8A6F7923EA96617BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperrigh.phy"},
	{0x8A6F9DBA0727C441ULL, "vxw_woundedcoast_1l5_0_0.vlm"},
	{0x8A7060DB874C3E27ULL, "vxlt_0_lowtown_docks_a3_n_97o_0_0.vlm"},
	{0x8A70904629EB94A5ULL, "247947842.ogg"},
	{0x8A709EF6ED22693FULL, "fx_firetwister.dds"},
	{0x8A70BB98AE19BCAFULL, "129654738.ogg"},
	{0x8A70F5D416F6A1FFULL, "lt_0_lowtown_a2_n_c_12_0.msh"},
	{0x8A70FBD416F6ABA1ULL, "lt_0_lowtown_a2_n_c_12_0.mmh"},
	{0x8A710265EEEC2622ULL, "547148858.ogg"},
	{0x8A7107D416F6C042ULL, "lt_0_lowtown_a2_n_c_12_0.mao"},
	{0x8A7160D2C343CCB9ULL, "6072850_f.fxe"},
	{0x8A7241FBCEC5110FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal.evt"},
	{0x8A72A5E4CE21DC13ULL, "6151391_f.fxe"},
	{0x8A72FC36C46DAB4FULL, "ser210_witness.bnk"},
	{0x8A72FFAF5ED3349CULL, "hlac_ht_keep_n_3732_1x.dds"},
	{0x8A7362452F3C21FBULL, "6197871_m.fxe"},
	{0x8A73A462A0A33984ULL, "lt_0_lowtown_mmi.gff"},
	{0x8A73DD4955CA525DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rushing_guard_action.gad"},
	{0x8A73EA9FF99A0A57ULL, "gen_im_trs_krk_hvy_glv_r1.uti"},
	{0x8A74EE6D88EE3AA1ULL, "6004117_m.fxe"},
	{0x8A7564FC64DA3612ULL, "vxlt_0_lowtown_a3_n_c_ghx_0_0.vlm"},
	{0x8A75877CB2B4F8DBULL, "vxdr_0_deeproads_f5d_0_0.vlm"},
	{0x8A75A840BD41A3EBULL, "lgt110ar_gang_hideout_ht.are"},
	{0x8A75FAD12229EEFFULL, "vxdr_primevaltomb_511_0_0.vlm"},
	{0x8A75FFB5A03944B6ULL, "569842891.ogg"},
	{0x8A761E36C4701AB7ULL, "ser210_witness.cnv"},
	{0x8A762AE79F9AD669ULL, "gen_im_trs_dwv_med_boo_r0.uti"},
	{0x8A76314EFF2BD85DULL, "dae000_dalish_apprentices6147534.cl"},
	{0x8A76491067B1B780ULL, "420531562.ogg"},
	{0x8A768B030B6EA0CBULL, "6017487_m.fxe"},
	{0x8A76C3F995758AE5ULL, "73171451.ogg"},
	{0x8A76D6019C0F6F97ULL, "vxht_bartrand_n_756_0_0.vlm"},
	{0x8A76FFE83950A5D5ULL, "vxw_encounter2_camp_kd_0_0.vlm"},
	{0x8A7751B7E2AE81F9ULL, "vxdr_primevaltomb_6it_0_0.vlm"},
	{0x8A77DCF8B8A5F139ULL, "ogre_abilities.ncs"},
	{0x8A77E94DF58626A6ULL, "gen_im_trs_gen_one_mac_r1.uti"},
	{0x8A77F20AE231AF84ULL, "6196022_m.fxe"},
	{0x8A782C79A4211F94ULL, "479436038.ogg"},
	{0x8A7867FC07AC1D69ULL, "vxht_mansion2b_3ds_0_0.vlm"},
	{0x8A78860AEA89AEA6ULL, "vxca_caverns1_mines_hua_0_0.vlm"},
	{0x8A78A52DD3EFD3E6ULL, "545462279.ogg"},
	{0x8A7942FBCECB31B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_formal.gad"},
	{0x8A7945778A12DACAULL, "tbe100pt_emporium.ncs"},
	{0x8A79C6B4F32E8E48ULL, "56514023.ogg"},
	{0x8A79E3F17D8ACBD7ULL, "vxw_bonepitdestroyed_a3_1o5_0_0.vlm"},
	{0x8A79ECBCDB195EBEULL, "307016791.ogg"},
	{0x8A79F2C60F5DF085ULL, "anim_combat.gda"},
	{0x8A7A3FE0D8867391ULL, "gen00fl_sebastian6180397.cl"},
	{0x8A7ACC7B3D66AB37ULL, "vxw_bonepitdestroyed_a3_1p4_0_0.vlm"},
	{0x8A7B0955559CF635ULL, "hlac_ht_0_hightown_a2_-65536_4x.dds"},
	{0x8A7B15F71049EA2DULL, "hlac_ht_0_hightown_a3_65535_0y.dds"},
	{0x8A7B4C6B5F8C6DE5ULL, "6001262_m.fxe"},
	{0x8A7B9E27D315ACD2ULL, "w_woundedcoast_136_0.phy"},
	{0x8A7BAA8F23714E9DULL, "696283373.ogg"},
	{0x8A7C6C6F5946FBA3ULL, "zz_followers_banter6159087.cl"},
	{0x8A7C8DC0363B9C7CULL, "w_bonepitdestroyed_a3_223_0.mao"},
	{0x8A7C99C0363BB01FULL, "w_bonepitdestroyed_a3_223_0.mmh"},
	{0x8A7C9FC0363BBAC1ULL, "w_bonepitdestroyed_a3_223_0.msh"},
	{0x8A7CD5FCABACAB36ULL, "548694093.ogg"},
	{0x8A7CF67DC6DBB92FULL, "gen00fl_aveline6116721.cl"},
	{0x8A7D3DBAB54F9C2FULL, "gen00fl_fenris6126684.cl"},
	{0x8A7D3FE633EA9917ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev2_loop.gad"},
	{0x8A7D675B3982E668ULL, "follower_barks6154163.cl"},
	{0x8A7D8B128F035B15ULL, "6099335_m.fxe"},
	{0x8A7D8B294F54730CULL, "vxht_bartrand_n_63j_0_0.vlm"},
	{0x8A7D9532353BB384ULL, "718937122.ogg"},
	{0x8A7DB7B318D36624ULL, "6208693_m.fxe"},
	{0x8A7DCA0B8E796E18ULL, "vxlt_darktown_49u_0_0.vlm"},
	{0x8A7DF7064579CC50ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_flaggallowsb.phy"},
	{0x8A7E6644A2176059ULL, "mcr371_hostile_merrill6082162.cl"},
	{0x8A7F2C6D3907A7A0ULL, "vxlt_gallowstemplar_n_4xy_0_0.vlm"},
	{0x8A8024075881C82BULL, "103110043.ogg"},
	{0x8A80240FDF2E9B39ULL, "vxw_encounter3_fd_0_0.vlm"},
	{0x8A80A51C6009A426ULL, "llac_lt_0_lowtown_a2_n_c_-1_4.dds"},
	{0x8A80FB568BBC55E1ULL, "rdr111dg_assistant6175872.cl"},
	{0x8A813046BFFA5116ULL, "6207962_m.fxe"},
	{0x8A8242C454C13CEAULL, "vxht_0_hightown_cd4_0_0.vlm"},
	{0x8A8242DD1DFA46F3ULL, "6147874_m.fxe"},
	{0x8A825BCEDA0DCAB2ULL, "335299755.ogg"},
	{0x8A83456F0529BD40ULL, "434440904.ogg"},
	{0x8A83A5197B990AEBULL, "dae000_arishok6067633.cl"},
	{0x8A841B5FDE9ACE30ULL, "vxht_mansion_n_5bu_0_0.vlm"},
	{0x8A8429057F613A2CULL, "art\\characters\\playercharacter\\humanfemale\\hlf_lgtd.phy"},
	{0x8A84AAEB5D14615CULL, "6091846_m.fxe"},
	{0x8A84B4F5E472328CULL, "vxht_mansion2b_2l3_0_0.vlm"},
	{0x8A8571C52B46BE55ULL, "vxw_encounter2_camp_fc_0_0.vlm"},
	{0x8A857962CB9AE547ULL, "6101318_f.fxe"},
	{0x8A85A93027B2D516ULL, "vxlt_undercityencounter_13b_0_0.vlm"},
	{0x8A85B7C4444BE4AEULL, "898223390.ogg"},
	{0x8A85F51EE4E747C3ULL, "mcr310tr_bring_shades.ncs"},
	{0x8A8633BBECB9B0D7ULL, "vxdr_primevaltomb_8f9_0_0.vlm"},
	{0x8A8653263403E16AULL, "133441950.ogg"},
	{0x8A8675D4437FB493ULL, "6094091_m.fxe"},
	{0x8A87676958AC4068ULL, "220441803.ogg"},
	{0x8A87A64336A960E2ULL, "656732694.ogg"},
	{0x8A87E42C616E3C2DULL, "6122028_m.fxe"},
	{0x8A883570C8A2359AULL, "fen110_party_comment6095237.cl"},
	{0x8A8897381D43EA15ULL, "vxlt_0_lowtown_a3_n_c_fe1_0_0.vlm"},
	{0x8A88F4348B5644DDULL, "6070998_m.fxe"},
	{0x8A89014F77F81853ULL, "hlac_ca_caverns1_mines_24426_3z.dds"},
	{0x8A890919699BBFC5ULL, "zz_approval_debug6003662.cl"},
	{0x8A892F05113EB2AEULL, "llac_lt_0_lowtown_65535_1.dds"},
	{0x8A893E582F5CC3B7ULL, "vxw_bonepitdestroyed_a3_1u7_0_0.vlm"},
	{0x8A895C9C7B99A0CDULL, "one000_ilen6037668.cl"},
	{0x8A8A0DD8D42900C8ULL, "hlac_lt_0_lowtown_n_65538_4y.dds"},
	{0x8A8A236C0C297E0FULL, "vxw_bonepit_yp_0_0.vlm"},
	{0x8A8A3479786463A1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v2.ani"},
	{0x8A8A5E290B573EF8ULL, "382890758.ogg"},
	{0x8A8AF146343E846AULL, "vxlt_gallowstemplar_4o1_0_0.vlm"},
	{0x8A8B1FA8EBDDA433ULL, "hlac_lt_0_lowtown_a2_-131072_4y.dds"},
	{0x8A8C6C7037175592ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ss_leftrailing.phy"},
	{0x8A8D0A0392A2C11BULL, "sib131_slaver_lord6165204.cl"},
	{0x8A8D348F2044772AULL, "vxca_cavern1_haunted_e90_0_0.vlm"},
	{0x8A8D36087B691813ULL, "art\\characters\\creatures\\strider\\animation\\attackbackright.evt"},
	{0x8A8DBAA8BF15E193ULL, "lgt110cr_coa_kirk_guard.utc"},
	{0x8A8DC64A5EDACDCFULL, "vxlt_hangedman_1jw_0_0.vlm"},
	{0x8A8DE2627396B646ULL, "vxlt_warehouse_15k_0_0.vlm"},
	{0x8A8E0114F223567BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ready.ani"},
	{0x8A8E7F206DBBCA66ULL, "6150501_m.fxe"},
	{0x8A8E95ECD124B3B9ULL, "vxw_encounter3_ck_0_0.vlm"},
	{0x8A8EACBA1F3F77BFULL, "lt_0_lowtown_docks_n_138_0.msh"},
	{0x8A8EB2BA1F3F8261ULL, "lt_0_lowtown_docks_n_138_0.mmh"},
	{0x8A8EBEBA1F3F9602ULL, "lt_0_lowtown_docks_n_138_0.mao"},
	{0x8A8F0E319E560796ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masf.mao"},
	{0x8A8F1A319E561C35ULL, "art\\characters\\playercharacter\\humanmale\\hlm_masf.mmh"},
	{0x8A8F1A4C7E385509ULL, "528495769.ogg"},
	{0x8A8F1C319E561FABULL, "art\\characters\\playercharacter\\humanmale\\hlm_masf.msh"},
	{0x8A8F22877FD55E8FULL, "6053667_m.fxe"},
	{0x8A8F498033816711ULL, "vxlt_gallowstemplar_n_3ix_0_0.vlm"},
	{0x8A8F5A339A43B8D5ULL, "6096222_m.fxe"},
	{0x8A8F923275B4AA6FULL, "gen00fl_anders6125603.cl"},
	{0x8A8FB350E77E65E8ULL, "darktown_fem_fer_urch_sta6174795.cl"},
	{0x8A8FE02D3C3469E1ULL, "954623213.ogg"},
	{0x8A8FE20F7BCCBB68ULL, "hla_ht_0_hightown_n_65538_1y.dds"},
	{0x8A9009B4321014B6ULL, "war_penetration.dds"},
	{0x8A90266E16CDB988ULL, "133641194.ogg"},
	{0x8A9069D11B6B64EDULL, "6103822_f.fxe"},
	{0x8A9070D997CEECB9ULL, "vxht_keep_6bp_0_0.vlm"},
	{0x8A90B4BA0FF05F73ULL, "6140974_m.fxe"},
	{0x8A90B5320CBF3ED0ULL, "vxlt_0_lowtown_a3_n_c_cow_0_0.vlm"},
	{0x8A913F79786A9543ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v2.gad"},
	{0x8A91433CB4F29EB4ULL, "dae000cr_grand_cleric6147726.cl"},
	{0x8A916CE633FB8E9DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev2_loop.ani"},
	{0x8A91D4E3F04800EFULL, "6092421_f.fxe"},
	{0x8A91DCBEFA130B38ULL, "zz_amb_test.lst"},
	{0x8A9230A7964374CCULL, "vxw_encounter3_2a_0_0.vlm"},
	{0x8A932F087B6D7854ULL, "art\\characters\\creatures\\strider\\animation\\attackbackright.gad"},
	{0x8A93CCD9D0231FD6ULL, "377532157.ogg"},
	{0x8A93F469D35F78EEULL, "art\\characters\\playercharacter\\elfmale\\em_har_fls.mmh"},
	{0x8A940A02409C2AC6ULL, "llac_ht_0_hightown_0_3.dds"},
	{0x8A940E69D35FA4DCULL, "art\\characters\\playercharacter\\elfmale\\em_har_fls.msh"},
	{0x8A9466C217AD2A97ULL, "71242953.ogg"},
	{0x8A9480469AB48EA7ULL, "207226710.ogg"},
	{0x8A949B6F5DF0016CULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_pcfem_.gad"},
	{0x8A94D4E6CA7D1050ULL, "6206361_m.fxe"},
	{0x8A94E78413721494ULL, "vxlt_0_lowtown_docks_a3_bg8_0_0.vlm"},
	{0x8A951CE98DE9F880ULL, "vxlt_hangedman_27x_0_0.vlm"},
	{0x8A9581830C527BC4ULL, "6149047_f.fxe"},
	{0x8A958368001C238FULL, "6156018_m.fxe"},
	{0x8A961B42E3464298ULL, "6043396_f.fxe"},
	{0x8A967539979C5A55ULL, "lt_0_lowtown_a2_n_c_42_0_ll.mmh"},
	{0x8A9699779A93EE74ULL, "vxdr_primevaltomb_3it_0_0.vlm"},
	{0x8A96DBD4171744CAULL, "lt_0_lowtown_a2_n_c_12_0.rml"},
	{0x8A9707D645061B1EULL, "692260632.ogg"},
	{0x8A978C8526C40F95ULL, "781830185.ogg"},
	{0x8A979BD6D29711EEULL, "lit_aoe_02.mao"},
	{0x8A97A947D42D0C41ULL, "one300_nuncio6075162.cl"},
	{0x8A97B58F85351CE6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_amllcrstnew_n.dds"},
	{0x8A97ED7BA6D4BBFFULL, "vxw_bonepitdestroyed_a3_16f_0_0.vlm"},
	{0x8A97FB364C99E75CULL, "vxw_encounter1_av_0_0.vlm"},
	{0x8A98AC3C8B174C0CULL, "mag370_meredith6092184.cl"},
	{0x8A99197E8C30BD37ULL, "vxca_cavern1_haunted_i28_0_0.vlm"},
	{0x8A993960514A4C37ULL, "926983404.ogg"},
	{0x8A9A3E087B73B032ULL, "art\\characters\\creatures\\strider\\animation\\attackbackright.ani"},
	{0x8A9A6EE33340A40CULL, "vxht_bartrand_n_69h_0_0.vlm"},
	{0x8A9B6023EABB711AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperrigh.msh"},
	{0x8A9B604213EC62F0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpoff.mao"},
	{0x8A9B644213EC6903ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpoff.mmh"},
	{0x8A9B6E23EABB8953ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperrigh.mao"},
	{0x8A9B7223EABB9060ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_wo2upperrigh.mmh"},
	{0x8A9B724213EC815DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpoff.msh"},
	{0x8A9B9A6F5DF61E6AULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_pcfem_.ani"},
	{0x8A9C0914F22FA172ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ready.evt"},
	{0x8A9C59E899304B0FULL, "gen00fl_anders6125557.cl"},
	{0x8A9C792E556B2DA2ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.impa.gad"},
	{0x8A9C8D2FCEC5B0C5ULL, "hlac_ht_0_hightown_131074_0x.dds"},
	{0x8A9CAFDA34D39551ULL, "vxht_brothel_n_541_0_0.vlm"},
	{0x8A9CF597307B9F8BULL, "cs_dre180_db_rock_warr.wbd"},
	{0x8A9D374498747892ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masd.phy"},
	{0x8A9D496BA1B98417ULL, "679510739.ogg"},
	{0x8A9D6AB2CC4C5466ULL, "6024538_m.fxe"},
	{0x8A9D7F72AD582BB4ULL, "zz_followers_banter6159204.cl"},
	{0x8A9DC6D479C9D7CAULL, "6083084_f.fxe"},
	{0x8A9DCFB042FE9EFBULL, "223596646.ogg"},
	{0x8A9DE3A7CB7DD86FULL, "hlac_lt_gallowstemplar_f_2383_11x.dds"},
	{0x8A9DFCD4171D9B1EULL, "lt_0_lowtown_a2_n_c_12_0.phy"},
	{0x8A9E328210CB02A0ULL, "vxlt_gallowstemplar_f_46k_0_0.vlm"},
	{0x8A9E40544A3AF530ULL, "vxlt_hangedman_1qu_0_0.vlm"},
	{0x8A9EADC85DC8D194ULL, "80116796.ogg"},
	{0x8A9F44DB87525245ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_avgdesk_sa.dds"},
	{0x8AA02018CD114068ULL, "art\\vfx\\materials\\chunky\\vfx_c_lit_arc_add.mao"},
	{0x8AA052ACB9C63D30ULL, "601538465.ogg"},
	{0x8AA0B71055F9F9CCULL, "6040602_m.fxe"},
	{0x8AA0FF19C286FF30ULL, "vxht_mansion_n_5hs_0_0.vlm"},
	{0x8AA10004297E9D22ULL, "clf360_evets6072214.cl"},
	{0x8AA171DA60E40F3BULL, "dre100cr_hireling_1.utc"},
	{0x8AA198AFC85E638CULL, "vxht_mansion2b_2b1_0_0.vlm"},
	{0x8AA1AE66B35EA0BDULL, "937491790.ogg"},
	{0x8AA22814F2344105ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ready.gad"},
	{0x8AA25B0F273D1C17ULL, "6100057_m.fxe"},
	{0x8AA2940024696DB1ULL, "vxca_caverns2_85d_0_0.vlm"},
	{0x8AA29EC5C6EAE4D3ULL, "fxc_puke_sp.dds"},
	{0x8AA2A74EB890CB79ULL, "vxdr_primevaltomb_6vy_0_0.vlm"},
	{0x8AA2B1439FC5915DULL, "vxht_brothel_n_5jk_0_0.vlm"},
	{0x8AA2D3CE5BB3A84FULL, "vxw_encounter2_camp_7b_0_0.vlm"},
	{0x8AA39C2E55718799ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.impa.evt"},
	{0x8AA3C44F913C6A1AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01b.phy"},
	{0x8AA43F817480D434ULL, "dae000_madam_lusine6156789.cl"},
	{0x8AA44CD2B95DF40FULL, "lt_0_lowtown_docks_a3_n_125_0.mao"},
	{0x8AA458D2B95E08ACULL, "lt_0_lowtown_docks_a3_n_125_0.mmh"},
	{0x8AA45ED2B95E129EULL, "lt_0_lowtown_docks_a3_n_125_0.msh"},
	{0x8AA48648E99C75F1ULL, "vxht_mansion2_3ds_0_0.vlm"},
	{0x8AA4F59DFDC02432ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\blendtree\\circlehighmage_override.anb"},
	{0x8AA512A4AD3CF0BBULL, "546207968.ogg"},
	{0x8AA5321406B265E9ULL, "atl_debugoptio_dxt5_00.dds"},
	{0x8AA535A2D1E88E17ULL, "ts_beasts_drag_prog_glands.gda"},
	{0x8AA540C7310CFE76ULL, "387052650.ogg"},
	{0x8AA60069AFB21DADULL, "mrl150_hunter6071690.cl"},
	{0x8AA60C77B8E9A65BULL, "debug_fight.ncs"},
	{0x8AA66773621FC357ULL, "mrl150_cut_merrill_intro.bnk"},
	{0x8AA66D133E8FFFDCULL, "vxw_encounter1_fw_0_0.vlm"},
	{0x8AA6A5B3A23B55F9ULL, "vxlt_undercityserialkiller_wv_0_0.vlm"},
	{0x8AA73FCBC008E1CDULL, "vxlt_0_lowtown_a2_n_c_gjl_0_0.vlm"},
	{0x8AA7E0A3A9D559C4ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0sc_basic.phy"},
	{0x8AA8F4F3B297164DULL, "6151071_m.fxe"},
	{0x8AA949736221C66FULL, "mrl150_cut_merrill_intro.cnv"},
	{0x8AA95E4D91400EF3ULL, "hm_templar1.mor"},
	{0x8AAA9FBCACCA6FE9ULL, "vxdr_primevaltomb_6qn_0_2.vlm"},
	{0x8AAAF77CA24EEB30ULL, "31951164.ogg"},
	{0x8AAB04EC6CE1993AULL, "hf_cgn_007.mop"},
	{0x8AAB2AD725E05EB8ULL, "vxht_brothel_n_3i5_0_0.vlm"},
	{0x8AAB4F7D13087D92ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_6y.dds"},
	{0x8AABFCAFF37F5640ULL, "vxht_chantry_n_28l_0_0.vlm"},
	{0x8AAD3D52766FC26BULL, "vxht_0_hightown_a2_boq_0_0.vlm"},
	{0x8AAD54937BBB1DB9ULL, "vxht_keep_6hv_0_0.vlm"},
	{0x8AAD7568EE69424CULL, "vxlt_0_lowtown_a2_eqc_0_0.vlm"},
	{0x8AADB6AF7C0BCE2EULL, "vxca_cavern1_haunted_e50_0_0.vlm"},
	{0x8AAE43644E7D960FULL, "vxlt_0_lowtown_n_f1t_0_0.vlm"},
	{0x8AAEA902BE9EFD64ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_0y.dds"},
	{0x8AAF4A729CE16A8AULL, "art\\vfx\\textures\\vfx_z.dds"},
	{0x8AB060EBB740E4D8ULL, "mcr310_confused_guard6077099.cl"},
	{0x8AB0A42E557C1F88ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.impa.ani"},
	{0x8AB0D7120EA25802ULL, "zz_dae_debug6049949.cl"},
	{0x8AB0F3BA24F09056ULL, "vxlt_gallowscourtyard_n_74j_0_0.vlm"},
	{0x8AB1232091BBA9DDULL, "vxht_brothel_n_5mj_0_0.vlm"},
	{0x8AB145AB4DA9C0CFULL, "vxw_encounter2_camp_2m_0_0.vlm"},
	{0x8AB183003FC651BAULL, "6117500_m.fxe"},
	{0x8AB2406AF98D13FCULL, "438250503.ogg"},
	{0x8AB32BE9CF016CC1ULL, "gen00fl_merrill6128727.cl"},
	{0x8AB3C473A472F297ULL, "hlac_ht_playershouse_n_855_6x.dds"},
	{0x8AB493EF053F0D37ULL, "vxw_bonepitdestroyed_a3_1d0_0_0.vlm"},
	{0x8AB49D6582D7FCC5ULL, "82601574.ogg"},
	{0x8AB4CA316F92E547ULL, "260740143.ogg"},
	{0x8AB4DEF03086185CULL, "vxw_encounter1_kt_0_0.vlm"},
	{0x8AB4E084572D9D75ULL, "hlac_lt_0_lowtown_docks_a3_65533_4x.dds"},
	{0x8AB4F058CA0D2485ULL, "hlac_lt_gallowstemplar_n_2388_5z.dds"},
	{0x8AB51AC34984F5EDULL, "lt_0_lowtown_docks_a3_165_0.msh"},
	{0x8AB528C349850D20ULL, "lt_0_lowtown_docks_a3_165_0.mao"},
	{0x8AB52CC349851473ULL, "lt_0_lowtown_docks_a3_165_0.mmh"},
	{0x8AB58655B9B1E52BULL, "vxht_brothel_n_3p9_0_0.vlm"},
	{0x8AB58C1F56A9C6C6ULL, "plt_fen200pt_opening.nss"},
	{0x8AB5A1E76D61E4E6ULL, "6080671_m.fxe"},
	{0x8AB5EFFFC13EF7ECULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobb.msh"},
	{0x8AB5F5FFC13F01DEULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobb.mmh"},
	{0x8AB601FFC13F1675ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobb.mao"},
	{0x8AB6FEEB8DC258CCULL, "dae300_alistair_king6134962.cl"},
	{0x8AB77C9938E5644EULL, "6113041_m.fxe"},
	{0x8AB7C3033E2C2AE9ULL, "89636636.ogg"},
	{0x8AB877137EFD834CULL, "6089865_m.fxe"},
	{0x8AB8BF59E32AFE7AULL, "886986445.ogg"},
	{0x8AB8FD2602486619ULL, "llac_lt_gallowstemplar_f_7198_5.dds"},
	{0x8AB9E5629AA77BF9ULL, "w_woundedcoast_0_0.msh"},
	{0x8AB9EF629AA78CE7ULL, "w_woundedcoast_0_0.mmh"},
	{0x8AB9F3629AA79334ULL, "w_woundedcoast_0_0.mao"},
	{0x8AB9F79471E2028AULL, "6018876_m.fxe"},
	{0x8ABA1585DC732465ULL, "vxlt_0_lowtown_a3_n_c_fea_0_0.vlm"},
	{0x8ABA6E8CE5756EC0ULL, "vxht_chantry_n_23m_0_0.vlm"},
	{0x8ABAB3BA1F64BEDEULL, "lt_0_lowtown_docks_n_138_0.phy"},
	{0x8ABADFF3BAC18072ULL, "546063828.ogg"},
	{0x8ABB7BD59E6CD8B3ULL, "6095881_m.fxe"},
	{0x8ABB8BFA24AA5AA1ULL, "hlac_lt_0_lowtown_docks_n_-4_6y.dds"},
	{0x8ABB91AB215540C8ULL, "mag291_vincento.cnv"},
	{0x8ABBCC93C1D4AF8CULL, "ener_smo.mao"},
	{0x8ABCA977DA06BCECULL, "hlac_w_sundermount_65536_0x.dds"},
	{0x8ABCB5414073AE8FULL, "vxlt_0_lowtown_n_f4w_0_0.vlm"},
	{0x8ABD9B817FBB4EBAULL, "hlac_lt_0_lowtown_a2_n_-131071_0x.dds"},
	{0x8ABE3DD02B639894ULL, "vxlt_darktown_6io_0_0.vlm"},
	{0x8ABE43D7AB3B04D2ULL, "vxht_mansion_n_62x_0_0.vlm"},
	{0x8ABEB3AB2157B030ULL, "mag291_vincento.bnk"},
	{0x8ABF1F3C123DF0E3ULL, "hlac_dr_0_deeproads_5845_18x.dds"},
	{0x8ABF5139F657B574ULL, "vxht_brothel_n_4gx_0_0.vlm"},
	{0x8ABFC3D3CA3DBCC7ULL, "venge_healingaura.dds"},
	{0x8ABFFB2FB49212D7ULL, "vxdr_primevaltomb_8r5_0_0.vlm"},
	{0x8AC01F839E74431DULL, "486413072.ogg"},
	{0x8AC071A0B4689DE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_down_small.gad"},
	{0x8AC077604AF07B9BULL, "348659466.ogg"},
	{0x8AC0859C5E81FFDAULL, "6137537_m.fxe"},
	{0x8AC0D26544EE1648ULL, "vxht_0_hightown_n_c9w_0_1.vlm"},
	{0x8AC1731EBF603BF3ULL, "art\\characters\\creatures\\strider\\animation\\appear.evt"},
	{0x8AC1D797309A8959ULL, "cs_dre180_db_rock_warr.bnk"},
	{0x8AC1EBED57F57BFEULL, "hlac_ht_0_hightown_a2_131074_0y.dds"},
	{0x8AC1EEFB7E6D4445ULL, "41544997.ogg"},
	{0x8AC26D571B468C15ULL, "6149823_m.fxe"},
	{0x8AC275ADF8B134EFULL, "gen_im_trs_rdr_med_glv_r0.uti"},
	{0x8AC2DAC9DF83D785ULL, "aip_templarwarrior_base.gda"},
	{0x8AC2E8C5144113FDULL, "356065855.ogg"},
	{0x8AC36C544DDB6DFFULL, "vxw_bonepit_n1_0_0.vlm"},
	{0x8AC37A999F59D759ULL, "299825956.ogg"},
	{0x8AC43B77B722AC08ULL, "6047456_m.fxe"},
	{0x8AC4AC5BE794E67FULL, "vxca_cavern1_haunted_iie_0_0.vlm"},
	{0x8AC500057F988BF7ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_lgtd.mmh"},
	{0x8AC50CAC144CDC41ULL, "w_sundermount_101_0_lc.mmh"},
	{0x8AC516057F98B169ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_lgtd.msh"},
	{0x8AC5160DF7F6409BULL, "609734639.ogg"},
	{0x8AC52C8FB5B2C4FCULL, "mag110_ws_banter6214707.cl"},
	{0x8AC53FB73F171F34ULL, "zz_followers_banter6157523.cl"},
	{0x8AC5893466EB7552ULL, "6025317_m.fxe"},
	{0x8AC5C77F88ADD1D0ULL, "6044530_m.fxe"},
	{0x8AC60B629F5BBC6CULL, "vxlt_hangedman_n_2mm_0_0.vlm"},
	{0x8AC62F66E6FAED2AULL, "467854002.ogg"},
	{0x8AC69362A491CBBDULL, "vrc321ip_spooky_6_owner.utp"},
	{0x8AC694A0B46D44D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_down_small.evt"},
	{0x8AC765F318B7E814ULL, "hlac_lt_0_lowtown_docks_n_65534_5y.dds"},
	{0x8AC779D14704688FULL, "lgt110cr_gxa_vgk_cnspr_2.utc"},
	{0x8AC7E80F052E1940ULL, "gen00fl_anders6116169.cl"},
	{0x8AC7FE3E1A902FFBULL, "324252003.ogg"},
	{0x8AC86C1EBF664E34ULL, "art\\characters\\creatures\\strider\\animation\\appear.gad"},
	{0x8AC8819E8CB5C24FULL, "vxca_caverns1_j4l_0_0.vlm"},
	{0x8AC885951330DA3FULL, "dae000_porfiria6155290.cl"},
	{0x8AC8CE4035150C0CULL, "qcr220_orsino6060742.cl"},
	{0x8AC916A7F5075D6DULL, "6127197_m.fxe"},
	{0x8AC94284A3DC629AULL, "mag170_thrask6054062.cl"},
	{0x8AC952A58DE2B306ULL, "326660544.ogg"},
	{0x8AC9A11F56BA9301ULL, "plt_fen200pt_opening.plo"},
	{0x8AC9F4656D63EE80ULL, "903233200.ogg"},
	{0x8ACA38103121AAD0ULL, "vxdr_primevaltomb_35d_0_0.vlm"},
	{0x8ACA3C0C73D3AEAFULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_wkf.gad"},
	{0x8ACA61AE662D3A52ULL, "gen00fl_anders6103741.cl"},
	{0x8ACAE216BB714AE4ULL, "vxlt_undercityserialkiller_1yn_0_0.vlm"},
	{0x8ACB0FC8C7BDAB1EULL, "vxlt_0_lowtown_e2p_0_0.vlm"},
	{0x8ACBD3A404442F9EULL, "250379266.ogg"},
	{0x8ACC102A109EC912ULL, "vxlt_0_lowtown_a3_n_c_gvf_0_0.vlm"},
	{0x8ACC67BE8D2A672AULL, "plt_genpt_app_sebastian.plo"},
	{0x8ACCE600E27FE125ULL, "dae000_meredith6170921.cl"},
	{0x8ACE0E5B859741F8ULL, "6153476_m.fxe"},
	{0x8ACE1D4965E96B3FULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_004a_0d.dds"},
	{0x8ACE738555886A49ULL, "171151236.ogg"},
	{0x8ACE7B1EBF6AD312ULL, "art\\characters\\creatures\\strider\\animation\\appear.ani"},
	{0x8ACEBD782839D033ULL, "vxht_0_hightown_n_4an_0_0.vlm"},
	{0x8ACEE544E84586F0ULL, "6067652_m.fxe"},
	{0x8ACF1BE04C25D976ULL, "6115872_m.fxe"},
	{0x8ACF862E9C0116BAULL, "677622196.ogg"},
	{0x8ACFB3F423258572ULL, "vxw_bonepitdestroyed_a3_1q2_0_0.vlm"},
	{0x8ACFC784D90AAA18ULL, "vxlt_warehouse_n_cw_0_0.vlm"},
	{0x8AD05985E3B23C06ULL, "rdr110dg_guard_warehouse6016710.cl"},
	{0x8AD0C4C3235A191CULL, "gen00fl_aveline6140077.cl"},
	{0x8AD11D3D8E036672ULL, "vxca_cavern1_haunted_i94_0_0.vlm"},
	{0x8AD11DD4FEE213BAULL, "181044675.ogg"},
	{0x8AD1510C73D9F15CULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_wkf.evt"},
	{0x8AD1594359C4B0CFULL, "hla_ht_0_hightown_a2_-196608_0z.dds"},
	{0x8AD170B42737C486ULL, "6027728_m.fxe"},
	{0x8AD1A294FD4913F8ULL, "vxht_brothel_n_6ju_0_0.vlm"},
	{0x8AD1DB54468E55E2ULL, "dae000_meredith6166310.cl"},
	{0x8AD2CB6162942C7AULL, "414548932.ogg"},
	{0x8AD3124904003B37ULL, "dae100_viscount6118507.cl"},
	{0x8AD47D3F9151D4ECULL, "vxlt_hangedman_n_2hb_0_0.vlm"},
	{0x8AD49CA0B4798FC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod_down_small.ani"},
	{0x8AD52A0B205DEA4DULL, "981940547.ogg"},
	{0x8AD540D80992A56EULL, "gen00fl_carver6097329.cl"},
	{0x8AD5952CCA71746BULL, "vxlt_0_lowtown_a2_hfp_0_0.vlm"},
	{0x8AD5E85937D4FBD7ULL, "99381852.ogg"},
	{0x8AD5FB5296B1869BULL, "vxht_keep_n_588_0_0.vlm"},
	{0x8AD68C5FC11016A6ULL, "6043357_m.fxe"},
	{0x8AD6C23F85153B7DULL, "6089121_m.fxe"},
	{0x8AD6D9A8D77C7453ULL, "dae000ip_crowd_mf4_drunk1.utp"},
	{0x8AD6E41BA8F7AB3BULL, "vxht_0_hightown_a3_c2n_0_0.vlm"},
	{0x8AD74744D08E969BULL, "397330356.ogg"},
	{0x8AD78DCD2D0CFCE7ULL, "6166989_m.fxe"},
	{0x8AD80405AFA6EFAAULL, "art\\vfx\\cinematic\\vfx_orge_fire_imp.impa.gad"},
	{0x8AD807C812135730ULL, "vxlt_hangedman_1mq_0_0.vlm"},
	{0x8AD84D68A1512B78ULL, "6089586_m.fxe"},
	{0x8AD94E83EF1984CDULL, "hla_ht_0_hightown_a2_-131074_0x.dds"},
	{0x8AD95C0770B0FC4DULL, "503366060.ogg"},
	{0x8ADA2F969E5E42D4ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst_lp.ani"},
	{0x8ADAB9CAC488A5EFULL, "134461569.ogg"},
	{0x8ADAC3295E7BCE14ULL, "632912193.ogg"},
	{0x8ADAFA1EE52F905AULL, "mcr310tr_bring_shades.utt"},
	{0x8ADB0925D306D53EULL, "6182490_m.fxe"},
	{0x8ADB125188D4CCF9ULL, "fx_firetorchfb.dds"},
	{0x8ADB50B5F8ACEDC5ULL, "310983278.ogg"},
	{0x8ADBB104C72F43A0ULL, "579759239.ogg"},
	{0x8ADC067AFE802B3AULL, "vxht_mansion2_3cm_0_0.vlm"},
	{0x8ADC3978B147015DULL, "al_warrior_talents.gda"},
	{0x8ADC6A0F8ADF40C9ULL, "928337795.ogg"},
	{0x8ADC6DF545B9EDD4ULL, "w_blightlands_28_0.rml"},
	{0x8ADCE76FA602EDEBULL, "115641368.ogg"},
	{0x8ADD0B5D208D4E02ULL, "lt_gallowsprison_n_probe_6lh.dds"},
	{0x8ADDCE4887FC503AULL, "vxlt_gallowscourtyard_n_7ea_0_0.vlm"},
	{0x8ADDD20F8CE1409CULL, "sib131_slave_a6165278.cl"},
	{0x8ADDE75CF832D42BULL, "mag200_feynriel6023877.cl"},
	{0x8ADE25D1151B9DF2ULL, "vxw_bonepitdestroyed_a3_1z3_0_0.vlm"},
	{0x8ADEE705AFACDD41ULL, "art\\vfx\\cinematic\\vfx_orge_fire_imp.impa.evt"},
	{0x8ADF01DCC1996A66ULL, "6075572_m.fxe"},
	{0x8ADF17D1DD680ECEULL, "art\\characters\\playercharacter\\elffemale\\ef_har_ha1a.msh"},
	{0x8ADF25ACCA80665DULL, "6103887_m.fxe"},
	{0x8ADF2AE557ABAEA6ULL, "6040910_m.fxe"},
	{0x8ADF31D1DD683ABCULL, "art\\characters\\playercharacter\\elffemale\\ef_har_ha1a.mmh"},
	{0x8ADF590C73E63C75ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn180l_wkf.ani"},
	{0x8ADFBD0138F0C6BCULL, "gen00fl_aveline6116687.cl"},
	{0x8ADFF48A63BD9FD2ULL, "ht_0_hightown_20_0.anb"},
	{0x8AE03D6F1D63206DULL, "mag320_orsino6038227.cl"},
	{0x8AE07F43FF4EAA9DULL, "ave331_jeven6151152.cl"},
	{0x8AE08A998DF0581BULL, "zz_vault_debug6203346.cl"},
	{0x8AE0C88D6947849CULL, "6015059_m.fxe"},
	{0x8AE10DC991CD8EAEULL, "lt_0_lowtown_docks_a3_n_150_0.msh"},
	{0x8AE11BC991CDA6FFULL, "lt_0_lowtown_docks_a3_n_150_0.mao"},
	{0x8AE127C991CDBA9CULL, "lt_0_lowtown_docks_a3_n_150_0.mmh"},
	{0x8AE12C969E645C36ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst_lp.gad"},
	{0x8AE1586950A360B5ULL, "follower_banter6141723.cl"},
	{0x8AE15ABDF3988EB7ULL, "w_bonepit_105_0.phy"},
	{0x8AE18652539E674CULL, "vxht_0_hightown_n_b9o_0_0.vlm"},
	{0x8AE1D897B9422210ULL, "fen210_hadriana6110520.cl"},
	{0x8AE1F9B7C9E8AC00ULL, "6060297_m.fxe"},
	{0x8AE20BA9B7BC1C47ULL, "llac_ca_caverns2prc_5802_1.dds"},
	{0x8AE247CFF69E5491ULL, "art\\levels\\races\\proxy\\props\\door\\animation\\doorclosed.gad"},
	{0x8AE2885CAEA9AF43ULL, "6169157_m.fxe"},
	{0x8AE2A65CADB548B0ULL, "181458232.ogg"},
	{0x8AE30DACC2169831ULL, "hla_ht_0_hightown_n_2_0z.dds"},
	{0x8AE3166C95C49485ULL, "vxht_brothel_n_6dn_0_0.vlm"},
	{0x8AE3AD49A9796CCAULL, "6048451_m.fxe"},
	{0x8AE3AE349F47A6B6ULL, "dae000_dalish_apprentices6147746.cl"},
	{0x8AE3F4E235E2CDC7ULL, "vxlt_gallowstemplar_f_5at_0_0.vlm"},
	{0x8AE40709BC678CEBULL, "vxlt_0_lowtown_a2_hcs_0_0.vlm"},
	{0x8AE444FFD718E173ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.anb"},
	{0x8AE44F1A007D88F8ULL, "vxw_sundermount_20g_0_0.vlm"},
	{0x8AE487C332BD8D48ULL, "hlac_lt_0_lowtown_docks_a3_-65539_4x.dds"},
	{0x8AE4AD28DC3CBFC4ULL, "vxlt_0_lowtown_a2_gcc_0_0.vlm"},
	{0x8AE4B9EA7FC1E962ULL, "vrc221_bartrand6047714.cl"},
	{0x8AE4E95A924071C1ULL, "vxlt_darktown_5ko_0_0.vlm"},
	{0x8AE51F9451E4022FULL, "6140865_f.fxe"},
	{0x8AE550FDC2E4043AULL, "vxdr_0_deeproads_cuj_0_2.vlm"},
	{0x8AE555F89AEDC3BBULL, "vxht_0_hightown_a3_c7a_0_0.vlm"},
	{0x8AE55688F5A1BB4EULL, "6076322_m.fxe"},
	{0x8AE583FD7C16B514ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm05.phy"},
	{0x8AE646186041A577ULL, "llac_w_bonepit_393220_0.dds"},
	{0x8AE6FCC7C73F7F54ULL, "195764772.ogg"},
	{0x8AE7008F8DD9B0FEULL, "vxht_bartrand_n_6k0_0_0.vlm"},
	{0x8AE747969E68F595ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_cst_lp.evt"},
	{0x8AE7B5ACFC2A4567ULL, "1016922798.ogg"},
	{0x8AE7B7580E7A545FULL, "vxht_keep_62f_0_0.vlm"},
	{0x8AE7C4BD1CE99DB0ULL, "highlords_belt.dds"},
	{0x8AE7C5D09F5D7BE4ULL, "vxlt_undercityserialkiller_1sl_0_0.vlm"},
	{0x8AE7CD5309C91072ULL, "vxlt_0_lowtown_a3_n_c_ehz_0_0.vlm"},
	{0x8AE8533D68CB61EAULL, "hlac_lt_0_lowtown_docks_-4_1x.dds"},
	{0x8AE8AF19EC69D2EAULL, "hlac_dr_primevaltomb_663_3x.dds"},
	{0x8AE93B8902B59A43ULL, "vxlt_gallowstemplar_n_5xg_0_0.vlm"},
	{0x8AE948CFF6A4753FULL, "art\\levels\\races\\proxy\\props\\door\\animation\\doorclosed.ani"},
	{0x8AE94B7F2EDBE980ULL, "6097853_m.fxe"},
	{0x8AE9545135C4BD6BULL, "gen_im_wep_rog_dua_act0_02.uti"},
	{0x8AE99273DA5F0D54ULL, "vxht_mansion_n_6cm_0_0.vlm"},
	{0x8AE9A44882052EE6ULL, "6046941_m.fxe"},
	{0x8AE9F90CE8DB7B5FULL, "seb321cr_leliana.utc"},
	{0x8AEA07C9C339FBB3ULL, "498349806.ogg"},
	{0x8AEA0B4BA3BD49A5ULL, "6127430_m.fxe"},
	{0x8AEA19FC9A3650C6ULL, "llac_ht_mansion2_818_28.dds"},
	{0x8AEA34D73DBE3BCEULL, "w_bonepit_233_0.msh"},
	{0x8AEA42D73DBE531FULL, "w_bonepit_233_0.mao"},
	{0x8AEA4ED73DBE67BCULL, "w_bonepit_233_0.mmh"},
	{0x8AEA6DBAB0C9FC01ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_freize_sa.dds"},
	{0x8AEA7857F07643BAULL, "vxht_mansion2_3fn_0_0.vlm"},
	{0x8AEB20D7CBC8FA68ULL, "vxht_0_hightown_a2_a4p_0_0.vlm"},
	{0x8AEB2F3B8DA08080ULL, "isa111_isabela6066763.cl"},
	{0x8AEB8BB1965E14C4ULL, "6100282_f.fxe"},
	{0x8AEBAA89C9E514D5ULL, "lowtown_fem_prostitute_sta6161607.cl"},
	{0x8AEBC8FECC31B7F0ULL, "6069452_m.fxe"},
	{0x8AEBDF05AFB75990ULL, "art\\vfx\\cinematic\\vfx_orge_fire_imp.impa.ani"},
	{0x8AEBF0E5732BA93FULL, "zz_followers_banter6157940.cl"},
	{0x8AEC09BD6B122557ULL, "6042139_m.fxe"},
	{0x8AEC1A6047F55DA8ULL, "plt_qcr000pt_general.plo"},
	{0x8AEC272D2951CA71ULL, "mag291ip_folly.utp"},
	{0x8AEC402579F268BAULL, "vxlt_gallowscourtyard_n_7hb_0_0.vlm"},
	{0x8AEC79ADB3395F96ULL, "78439557.ogg"},
	{0x8AECEF629AD2E644ULL, "w_woundedcoast_0_0.rml"},
	{0x8AED33D46B7F8FABULL, "6135817_m.fxe"},
	{0x8AED741F2F583943ULL, "472875501.ogg"},
	{0x8AEDA479751A91CAULL, "vxlt_gallowscourtyard_723_0_0.vlm"},
	{0x8AEE00F771EF9772ULL, "vxca_cavern1_haunted_i36_0_0.vlm"},
	{0x8AEE6C615D9085C4ULL, "hlac_lt_0_lowtown_a2_n_c_65536_12x.dds"},
	{0x8AEEA01261C0D4D0ULL, "art\\characters\\weapons\\dagger\\w_ssw_dssa.mmh"},
	{0x8AEEA0B697901F68ULL, "6024103_m.fxe"},
	{0x8AEEAE1261C0ECEAULL, "art\\characters\\weapons\\dagger\\w_ssw_dssa.msh"},
	{0x8AEEE14C2C57901AULL, "hlac_w_bonepit_131077_0y.dds"},
	{0x8AEF0CC50D7DCF05ULL, "6061246_m.fxe"},
	{0x8AEF1A96A960D50EULL, "dre000_carver6037618.cl"},
	{0x8AEF60FD7C1ECE11ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm05.msh"},
	{0x8AEF7AFD7C1EFACFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm05.mmh"},
	{0x8AEFA04EF5613BABULL, "6129715_m.fxe"},
	{0x8AEFEC1BD95AEA0FULL, "310815496.ogg"},
	{0x8AF008DAFFE62F5EULL, "402988164.ogg"},
	{0x8AF029A392E35CF1ULL, "278056056.ogg"},
	{0x8AF0A4115DBB95D0ULL, "ch1em_vl.psh.fxo"},
	{0x8AF132F76675F09BULL, "mag170_hugh6080715.cl"},
	{0x8AF1853E0C924C81ULL, "vxca_caverns1_mines_e6n_0_0.vlm"},
	{0x8AF1884987BAAD05ULL, "vxht_brothel_n_6am_0_0.vlm"},
	{0x8AF18A16397812C6ULL, "585736240.ogg"},
	{0x8AF2010D173CDFF8ULL, "1025606150.ogg"},
	{0x8AF21EFFA0400084ULL, "1038552408.ogg"},
	{0x8AF21F269B6F854AULL, "6043551_m.fxe"},
	{0x8AF261FFD7254FDCULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.mmh"},
	{0x8AF28D74B8FD5AECULL, "pro000_colonel_ewald.cnv"},
	{0x8AF2B9D9D9F12125ULL, "vxht_keep_49c_0_0.vlm"},
	{0x8AF2BBA5261D32DEULL, "6101547_m.fxe"},
	{0x8AF2C0F6F273A178ULL, "vxw_sundermount_25d_0_0.vlm"},
	{0x8AF2EC360B00BA70ULL, "nat380_sibling6207576.cl"},
	{0x8AF337EB9CC9099EULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.mmh"},
	{0x8AF35B3784368A41ULL, "vxlt_darktown_5ln_0_0.vlm"},
	{0x8AF399E48B937208ULL, "vxlt_0_lowtown_a2_n_eu4_0_0.vlm"},
	{0x8AF3A5F969016C7DULL, "6087321_m.fxe"},
	{0x8AF52445484BABB8ULL, "728381331.ogg"},
	{0x8AF53122853D5C78ULL, "6087786_m.fxe"},
	{0x8AF5422CFBEEDA31ULL, "hlac_lt_darktown_-1_56z.dds"},
	{0x8AF5520C7E199CDBULL, "6048259_f.fxe"},
	{0x8AF56F74B8FF5D1CULL, "pro000_colonel_ewald.bnk"},
	{0x8AF579407D120D61ULL, "855042654.ogg"},
	{0x8AF57EF8CFF95C83ULL, "vxlt_gallowstemplar_4uz_0_0.vlm"},
	{0x8AF5CCFFC174A079ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobb.phy"},
	{0x8AF6BF0663CA1683ULL, "21908328.ogg"},
	{0x8AF6C7D7F265A7A5ULL, "gen00fl_anders6125465.cl"},
	{0x8AF74E4283B3FCB6ULL, "dae000_flemeth6001094.cl"},
	{0x8AF773FB0B3D6919ULL, "6033652_m.fxe"},
	{0x8AF8287DE75E9B4EULL, "mag370_meredith6040637.cl"},
	{0x8AF896483F1BA6FDULL, "924068400.ogg"},
	{0x8AF8F12F58B853DCULL, "vxlt_hangedman_1vl_0_0.vlm"},
	{0x8AF93817B7639916ULL, "vxlt_undercityencounter_1kj_0_0.vlm"},
	{0x8AF94B84D6ABDA9DULL, "vxlt_0_lowtown_docks_n_b4i_0_1.vlm"},
	{0x8AF9751A2B6919DFULL, "338110847.ogg"},
	{0x8AF9887CC85ECC69ULL, "761519224.ogg"},
	{0x8AF98C2CA4129C83ULL, "vxht_0_hightown_a3_atn_0_0.vlm"},
	{0x8AF9E8BBE316E6EFULL, "zz_followers_banter6157250.cl"},
	{0x8AFA4FC13C05AF78ULL, "ca_cavern1_haunted_26035.rml"},
	{0x8AFA9976D667A452ULL, "hlac_ht_0_hightown_a2_-65537_3y.dds"},
	{0x8AFAAB2D3C465437ULL, "vxdr_primevaltomb_8y0_0_0.vlm"},
	{0x8AFB4926582B999BULL, "w_blightlands_242_0.msh"},
	{0x8AFB5B26582BB8E6ULL, "w_blightlands_242_0.mao"},
	{0x8AFB6726582BCC85ULL, "w_blightlands_242_0.mmh"},
	{0x8AFB872900E24A30ULL, "vxw_encounter2_camp_9k_0_0.vlm"},
	{0x8AFB9B057A670A9DULL, "hlac_ca_caverns2_5802_53z.dds"},
	{0x8AFBBB2C6D04DC50ULL, "vxw_bonepit_262_0_0.vlm"},
	{0x8AFBEC8AABFABB16ULL, "130567806.ogg"},
	{0x8AFC306A2BB9390CULL, "712834434.ogg"},
	{0x8AFCB3F657CB0FAEULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf04.phy"},
	{0x8AFCDF38DD98F26DULL, "hlac_ht_mansion2b_818_12z.dds"},
	{0x8AFD11D378B38E2DULL, "vxht_bartrand_n_660_0_0.vlm"},
	{0x8AFD74F165C3FA79ULL, "gen00fl_sebastian6174642.cl"},
	{0x8AFDE8BD188FFD80ULL, "vxw_encounter2_camp_57_0_0.vlm"},
	{0x8AFF2C4A3B4C664BULL, "area_transition_pressed.cur"},
	{0x8AFF63D4AFDD5231ULL, "24407259.ogg"},
	{0x8AFF77C2D7E23813ULL, "mer221_sabin6080578.cl"},
	{0x8AFFAB92B7B24394ULL, "6126469_m.fxe"},
	{0x8AFFBD4F918ABD43ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01b.msh"},
	{0x8AFFCB4F918AD49DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01b.mmh"},
	{0x8AFFFA2679B0C585ULL, "vxht_brothel_n_6jl_0_0.vlm"},
	{0x8B006F6D559AB462ULL, "421257050.ogg"},
	{0x8B00AFC2102615B8ULL, "vxht_bartrand_n_6il_0_0.vlm"},
	{0x8B00D89C19CEFEC7ULL, "vxlt_gallowstemplar_f_5ov_0_0.vlm"},
	{0x8B00E8177C24CE7DULL, "460991576.ogg"},
	{0x8B0116EB9CD50F5DULL, "art\\vfx\\combat\\vfx_q_ari_buf_lp_c.anb"},
	{0x8B014A132A162F92ULL, "plt_cod_itm_icon_armor.plo"},
	{0x8B0183B262BD8B65ULL, "gen00fl_merrill6116317.cl"},
	{0x8B01963DCBF7FEDCULL, "6093570_m.fxe"},
	{0x8B01A3A9C5AD32ABULL, "dae000_gamlen6149437.cl"},
	{0x8B01DA413E287B7FULL, "6137525_m.fxe"},
	{0x8B01E2454CE44249ULL, "169081394.ogg"},
	{0x8B01FCDF7752A6DDULL, "880824678.ogg"},
	{0x8B022460318D2586ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_reaction.ani"},
	{0x8B0255CDA5E2E12FULL, "vxca_cavern1_haunted_gk4_0_1.vlm"},
	{0x8B02714072D17C07ULL, "mag210_arianni6024525.cl"},
	{0x8B02A84CCA6F3E96ULL, "vxw_bonepitdestroyed_a3_nt_0_0.vlm"},
	{0x8B0338FB8E72296DULL, "472762465.ogg"},
	{0x8B03839175F0801AULL, "hlac_lt_0_lowtown_a2_n_c_-65535_5z.dds"},
	{0x8B03A04FD7EC7556ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basicnb.mao"},
	{0x8B03AC4FD7EC89F5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basicnb.mmh"},
	{0x8B03AE4FD7EC8D6BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basicnb.msh"},
	{0x8B03D03689842CC7ULL, "708078667.ogg"},
	{0x8B03E44971C5E1FEULL, "vxht_bartrand_n_6q2_0_0.vlm"},
	{0x8B0410D1DD871E87ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_ha1a.phy"},
	{0x8B04F4AA5706D42AULL, "6110817_m.fxe"},
	{0x8B0527CCD88F85F1ULL, "sky_w_encounter2.mao"},
	{0x8B053DFC7B963766ULL, "vxlt_gallowscourtyard_6xx_0_0.vlm"},
	{0x8B055CDDDE3A6564ULL, "6018643_m.fxe"},
	{0x8B057FAE62522B55ULL, "gen00fl_anders6125915.cl"},
	{0x8B058C550A28AE10ULL, "hlac_w_sundermount_131073_0x.dds"},
	{0x8B05DC01F4E7DD3AULL, "633828075.ogg"},
	{0x8B06762DBE4B3E54ULL, "vxht_mansion_n_6is_0_0.vlm"},
	{0x8B06C2E0710DC69FULL, "6140637_m.fxe"},
	{0x8B0706C991EE5267ULL, "lt_0_lowtown_docks_a3_n_150_0.phy"},
	{0x8B074906D98956D9ULL, "hlac_w_encounter2_camp_65540_0y.dds"},
	{0x8B077AA2D1095240ULL, "934138347.ogg"},
	{0x8B07A41BF45E3E96ULL, "312554590.ogg"},
	{0x8B089B70F2C97032ULL, "blood_conj_08_l3.mao"},
	{0x8B08BA9F3447E577ULL, "vxht_0_hightown_a3_c2z_0_0.vlm"},
	{0x8B08FACBC6C9069BULL, "vxlt_0_lowtown_docks_cq3_0_0.vlm"},
	{0x8B09013794E4FE80ULL, "647117600.ogg"},
	{0x8B09148C34F29A77ULL, "water_water_p2.vsh.fxo.dx11"},
	{0x8B0939571FD23DE2ULL, "dre000_dougal6131212.cl"},
	{0x8B093A44638BF4E5ULL, "398820652.ogg"},
	{0x8B093F5F794086CCULL, "gen_im_trs_slv_lgt_boo_r2.uti"},
	{0x8B09570D24137919ULL, "6148666_m.fxe"},
	{0x8B095ED70812CDC7ULL, "307597725.ogg"},
	{0x8B098EF8A0E30C18ULL, "vxlt_warehouse_n_ws_0_0.vlm"},
	{0x8B099FA1575EE751ULL, "683923178.ogg"},
	{0x8B09EB9A043EB771ULL, "vxlt_0_lowtown_a2_e7x_0_0.vlm"},
	{0x8B09F905F2D862B0ULL, "vxw_encounter2_camp_4h_0_0.vlm"},
	{0x8B0A0669B3E1D904ULL, "6114020_m.fxe"},
	{0x8B0A2D095EFAF4D0ULL, "vxw_bonepit_213_0_0.vlm"},
	{0x8B0A78B81A7CEBFEULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_dwpa_0n.dds"},
	{0x8B0A8F7FD7F11C42ULL, "vxht_chantry_356_0_1.vlm"},
	{0x8B0ABA36B5B0CEDFULL, "vxlt_gallowstemplar_n_4sk_0_0.vlm"},
	{0x8B0B864679E551AAULL, "vxlt_warehouse_15g_0_0.vlm"},
	{0x8B0B9648CEC2516DULL, "w_sundermount_190_0.mmh"},
	{0x8B0B9A48CEC257BEULL, "w_sundermount_190_0.mao"},
	{0x8B0BA848CEC26FF3ULL, "w_sundermount_190_0.msh"},
	{0x8B0C5B539747DC10ULL, "vxlt_gallowstemplar_f_3rl_0_0.vlm"},
	{0x8B0CE18A08825DA8ULL, "dae110cr_refugee_m_01.utc"},
	{0x8B0D2C9F36DE2123ULL, "llac_lt_gallowstemplar_2384_4.dds"},
	{0x8B0E030E854587D5ULL, "6076268_m.fxe"},
	{0x8B0E176048123F67ULL, "plt_qcr000pt_general.nss"},
	{0x8B0E52768E2BA119ULL, "mag101_grace6035856.cl"},
	{0x8B0EB9545AD577CCULL, "6028936_m.fxe"},
	{0x8B0EF06956507786ULL, "6110642_m.fxe"},
	{0x8B0F04DFA45684B0ULL, "hightown_market_guard_sta6169671.cl"},
	{0x8B0F219F021C2E38ULL, "vxht_bartrand_n_6rm_0_0.vlm"},
	{0x8B0F3F7F2A28159EULL, "w_woundedcoast_77_0.rml"},
	{0x8B0F44C34CA4096EULL, "plt_pro000pt_main.plo"},
	{0x8B0F652A8148A3DDULL, "423769127.ogg"},
	{0x8B100C6031993AF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_reaction.evt"},
	{0x8B1043621462568BULL, "6085938_m.fxe"},
	{0x8B110E59A1482E7EULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyo_0n.dds"},
	{0x8B1240F23939F105ULL, "w_woundedcoast_49_0.rml"},
	{0x8B12A0793845ED59ULL, "6139391_f.fxe"},
	{0x8B12A46B3B8800F4ULL, "vxlt_0_lowtown_docks_c4k_0_0.vlm"},
	{0x8B12E51ACF429316ULL, "ico_bomb_combustion.dds"},
	{0x8B12F78A63E8FE0EULL, "ht_0_hightown_20_0.rml"},
	{0x8B132615BEFC5AC3ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_nuda.mao"},
	{0x8B132A15BEFC6110ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_nuda.mmh"},
	{0x8B133815BEFC792AULL, "art\\characters\\playercharacter\\dwarfmale\\boo_nuda.msh"},
	{0x8B1359241B26B873ULL, "228243387.ogg"},
	{0x8B13773182156993ULL, "defaultsun.dds"},
	{0x8B14A81CB991FAE3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_shield_bash.ani"},
	{0x8B14DEEFA1C033CEULL, "6160652_m.fxe"},
	{0x8B14E80AB04156D4ULL, "vxht_mansion_n_6lp_0_0.vlm"},
	{0x8B14FFB1A35385B2ULL, "vxht_mansion2b_1c3_0_0.vlm"},
	{0x8B1557DA98032F74ULL, "hlac_w_woundedcoast_-131072_0x.dds"},
	{0x8B156DF545E9B297ULL, "w_blightlands_28_0.mmh"},
	{0x8B1571F545E9B9A4ULL, "w_blightlands_28_0.mao"},
	{0x8B1583F545E9D809ULL, "w_blightlands_28_0.msh"},
	{0x8B15D4F5287F847BULL, "39436783.ogg"},
	{0x8B160D60319DA780ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_reaction.gad"},
	{0x8B1610080AD94A85ULL, "vxht_0_hightown_a3_9b2_0_0.vlm"},
	{0x8B1675018166FAD4ULL, "llac_lt_darktown_-1_8.dds"},
	{0x8B16A7FF457ED0A2ULL, "lgt160_ntb_elf_conflict6163383.cl"},
	{0x8B16B63F3B7A01DBULL, "art\\characters\\playercharacter\\elffemale\\cth_come.mmh"},
	{0x8B16C23F3B7A1678ULL, "art\\characters\\playercharacter\\elffemale\\cth_come.mao"},
	{0x8B16D43F3B7A34C5ULL, "art\\characters\\playercharacter\\elffemale\\cth_come.msh"},
	{0x8B172C7C263DFDF7ULL, "vxht_0_hightown_a3_c7u_0_0.vlm"},
	{0x8B1800D592D92498ULL, "vxlt_warehouse_n_zp_0_0.vlm"},
	{0x8B18040404F76043ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_bridge8x.mao"},
	{0x8B18080404F76690ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_bridge8x.mmh"},
	{0x8B18160404F77EAAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_bridge8x.msh"},
	{0x8B18B4655B1B024AULL, "6148737_m.fxe"},
	{0x8B19FC6F81FB3371ULL, "vxw_encounter3_cs_0_0.vlm"},
	{0x8B1A291E6159563AULL, "6118677_m.fxe"},
	{0x8B1A38857466C88CULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower2_mcr370_playerrules.gad"},
	{0x8B1A4970563547EFULL, "gen_im_cft_rec_pot_05.uti"},
	{0x8B1A86946E73902EULL, "vxdr_primevaltomb_26g_0_1.vlm"},
	{0x8B1A960D7FC6730FULL, "vxw_bonepit_7z_0_0.vlm"},
	{0x8B1AD135FE129915ULL, "vxca_caverns2_7qs_0_0.vlm"},
	{0x8B1AD5776927D0F1ULL, "vxca_caverns1_ih5_0_0.vlm"},
	{0x8B1AEF4F6B40C853ULL, "6103306_m.fxe"},
	{0x8B1B230EB0E7A8D0ULL, "dae311ar_gamlens_house.ncs"},
	{0x8B1B4C1DCFC6B2F3ULL, "art\\vfx\\combat\\vfx_c_rw_s2_s1_trans_c.impa.ani"},
	{0x8B1B68A6F2248983ULL, "801211947.ogg"},
	{0x8B1B744D3EFA1B70ULL, "vxlt_0_lowtown_docks_a3_cyg_0_0.vlm"},
	{0x8B1BC71832E291A9ULL, "static_static_p2.vsh.fxo.dx11"},
	{0x8B1BD75C48A56C71ULL, "vxht_keep_6bh_0_0.vlm"},
	{0x8B1C063BA3C5701AULL, "zz_qcr_debug6106187.cl"},
	{0x8B1C1177670F3EDBULL, "llac_lt_undercityencounter_-1_32.dds"},
	{0x8B1C842DD87C79E8ULL, "mcr371_orsino6083342.cl"},
	{0x8B1C8F4C9B9E7494ULL, "6128669_m.fxe"},
	{0x8B1CA553FBFF176EULL, "338290964.ogg"},
	{0x8B1E481D1E29BD81ULL, "vxlt_hangedman_27u_0_0.vlm"},
	{0x8B1E7FCC119F992AULL, "6142865_m.fxe"},
	{0x8B1EAA0019C3489AULL, "ser270_cut_aftermath6162301.cl"},
	{0x8B1ED42111A60306ULL, "vxlt_0_lowtown_a2_n_eqg_0_1.vlm"},
	{0x8B1FD665A724DD9AULL, "vxlt_shop_ck_0_0.vlm"},
	{0x8B200559C6109B1FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsbuild.phy"},
	{0x8B203785746B328AULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower2_mcr370_playerrules.ani"},
	{0x8B204F087C6359A3ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_exit.evt"},
	{0x8B209B9619A73FCCULL, "llac_ht_0_hightown_a3_2_0.dds"},
	{0x8B2109E675DD341AULL, "art\\vfx\\combat\\vfx_ha_ar_armoff_act_c.anb"},
	{0x8B2115C10DAD0724ULL, "vxw_bonepitdestroyed_a3_c1_0_0.vlm"},
	{0x8B2116482D7E1974ULL, "vxlt_0_lowtown_docks_c1h_0_0.vlm"},
	{0x8B2118C98587AB71ULL, "vxht_0_hightown_n_bqf_0_0.vlm"},
	{0x8B211F59F1BD7EADULL, "vxht_mansion2b_2iu_0_0.vlm"},
	{0x8B2129974B1B84E5ULL, "6162594_m.fxe"},
	{0x8B21C01CB99CADDAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_shield_bash.evt"},
	{0x8B21D9DA207A09CAULL, "hlac_ht_chantry_2534_10x.dds"},
	{0x8B21E51D79D5AB4AULL, "vxlt_0_lowtown_a2_n_hdl_0_0.vlm"},
	{0x8B21FDE66D0B6B3CULL, "192372505.ogg"},
	{0x8B2220713CF0A8F1ULL, "275684799.ogg"},
	{0x8B2294F9E643B14CULL, "llac_lt_gallowsprison_n_-1_14.dds"},
	{0x8B22A0BE256588EEULL, "hlac_dr_0_deeproads_17038_7z.dds"},
	{0x8B232045C084A8A7ULL, "6197682_m.fxe"},
	{0x8B2335C34CB498FDULL, "plt_pro000pt_main.nss"},
	{0x8B238CBFA30658F0ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangsign.mao"},
	{0x8B2390BFA3065F03ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangsign.mmh"},
	{0x8B2393869E082E43ULL, "6036314-takecareofmotherb.cut"},
	{0x8B23968D49921780ULL, "vxca_caverns1_j25_0_0.vlm"},
	{0x8B239EBFA306775DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangsign.msh"},
	{0x8B23A7376FD3876DULL, "vxlt_gallowscourtyard_4xe_0_0.vlm"},
	{0x8B23C51C51B4D230ULL, "gen00fl_aveline6137695.cl"},
	{0x8B23DFEF7142BA44ULL, "80655525.ogg"},
	{0x8B243ECFCDA220D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_concrete_d.dds"},
	{0x8B24557E4445917EULL, "971700671.ogg"},
	{0x8B247BF5AD3A25F6ULL, "6045922_m.fxe"},
	{0x8B24D17C9FEFB0ACULL, "llac_lt_undercityencounter_-1_17.dds"},
	{0x8B252CE827685E4AULL, "llac_ca_caverns1_mines_24422_4.dds"},
	{0x8B26486218D884CEULL, "hlac_lt_darktown_-1_70y.dds"},
	{0x8B2683A1EC2DB2F3ULL, "static_static_p62.psh.fxo.dx11"},
	{0x8B27231F82BF9DC2ULL, "630356403.ogg"},
	{0x8B2728087C693624ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_exit.gad"},
	{0x8B2750CBAB0F6BDEULL, "hlac_w_bonepitdestroyed_a3_0_0z.dds"},
	{0x8B27B778FE51B98DULL, "hlac_dr_0_deeproads_5845_11y.dds"},
	{0x8B282FBDF3D4480FULL, "w_bonepit_105_0.mao"},
	{0x8B283BB886FEC029ULL, "vxw_bonepitmining_a2_lg_0_0.vlm"},
	{0x8B283BBDF3D45CACULL, "w_bonepit_105_0.mmh"},
	{0x8B2841BDF3D4669EULL, "w_bonepit_105_0.msh"},
	{0x8B28441DCFD12F2AULL, "art\\vfx\\combat\\vfx_c_rw_s2_s1_trans_c.impa.evt"},
	{0x8B28BF1CB9A2CACDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_shield_bash.gad"},
	{0x8B28E730B9C05799ULL, "6209844_m.fxe"},
	{0x8B28F128FFE23E6AULL, "one280_emrys6120042.cl"},
	{0x8B28F3671D499F32ULL, "6116200_m.fxe"},
	{0x8B294312F008B195ULL, "vxca_caverns2_7vr_0_0.vlm"},
	{0x8B29D756CF012DCFULL, "hlac_dr_dungeon_3207_1z.dds"},
	{0x8B29D8791D9DB76CULL, "473250593.ogg"},
	{0x8B2A171B2DC557CEULL, "hairalpha_hair_p3.psh.fxo.dx11"},
	{0x8B2A1B6B6D54EDE6ULL, "123275710.ogg"},
	{0x8B2A77E03DE2A803ULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsqr.mao"},
	{0x8B2A7BE03DE2AF50ULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsqr.mmh"},
	{0x8B2A86981A5A77E2ULL, "vxlt_hangedman_n_4x_0_0.vlm"},
	{0x8B2A89E03DE2C76AULL, "art\\levels\\races\\vfx_models\\prp_lightbeamsqr.msh"},
	{0x8B2A8D8576DDBEAAULL, "mcr370_cut_meredith_dies.cnv"},
	{0x8B2AA168B31BD1EAULL, "hlac_lt_gallowscourtyard_-131073_0x.dds"},
	{0x8B2AA3E6268D7002ULL, "vrc221_hugin6065256.cl"},
	{0x8B2B2123AE5D422FULL, "vxht_brothel_n_5zq_0_0.vlm"},
	{0x8B2B48F281E0CE88ULL, "6118099_m.fxe"},
	{0x8B2B521A37EDE2FBULL, "mer221_sabin6047456.cl"},
	{0x8B2B786407A99BB6ULL, "6043605_m.fxe"},
	{0x8B2B7B7963CB6757ULL, "vxht_0_hightown_a2_c5b_0_0.vlm"},
	{0x8B2BC6834B1B6839ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_showcase_sa.dds"},
	{0x8B2C412F1D17C9F3ULL, "1030923623.ogg"},
	{0x8B2C9A8F348A06C4ULL, "65996687.ogg"},
	{0x8B2CA9C72E16E592ULL, "hlac_lt_0_lowtown_docks_a3_-131075_7z.dds"},
	{0x8B2DF8BCD2DF1C92ULL, "6046475_m.fxe"},
	{0x8B2DFE55985B9E65ULL, "vxca_cavern1_haunted_fn5_0_0.vlm"},
	{0x8B2E07285AE90812ULL, "hlac_lt_undercityencounter_-1_0z.dds"},
	{0x8B2E0DD169674A31ULL, "vxdr_primevaltomb_6va_0_0.vlm"},
	{0x8B2E11F87233CCD9ULL, "303418563.ogg"},
	{0x8B2E37087C6F6E02ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_exit.ani"},
	{0x8B2E4842991AF61AULL, "vxlt_shop_fd_0_0.vlm"},
	{0x8B2E6F8576E174D2ULL, "mcr370_cut_meredith_dies.bnk"},
	{0x8B2E720A8A81E48CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04b.msh"},
	{0x8B2E780A8A81EF7EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04b.mmh"},
	{0x8B2EA71261F6C4EBULL, "art\\characters\\weapons\\dagger\\w_ssw_dssa.phy"},
	{0x8B2F15DF7AA60460ULL, "6014881_m.fxe"},
	{0x8B2F231DCFD7159DULL, "art\\vfx\\combat\\vfx_c_rw_s2_s1_trans_c.impa.gad"},
	{0x8B2F879DFFA31FA4ULL, "vxw_bonepitdestroyed_a3_n6_0_0.vlm"},
	{0x8B2F9B743D119D65ULL, "6167694_m.fxe"},
	{0x8B2FB89D285AFA60ULL, "dae000_madam_lusine6152143.cl"},
	{0x8B2FC7F1A08BF0F6ULL, "6080607_f.fxe"},
	{0x8B300BB7A18B54C2ULL, "6034795_m.fxe"},
	{0x8B302EE72C0BB924ULL, "vxlt_0_lowtown_docks_a3_n_btz_0_0.vlm"},
	{0x8B306A76CAF6CD02ULL, "vxht_brothel_n_5i5_0_0.vlm"},
	{0x8B3080EA3F28F224ULL, "6048826_m.fxe"},
	{0x8B30C5205A8A2665ULL, "947416721.ogg"},
	{0x8B30DA6A8BB03382ULL, "6124682_f.fxe"},
	{0x8B310659C61F1024ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsbuild.mmh"},
	{0x8B310A59C61F1777ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsbuild.mao"},
	{0x8B311C59C61F35C6ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsbuild.msh"},
	{0x8B31D0C4381C50BDULL, "art\\characters\\weapons\\shields\\w_shd_srdb.mao"},
	{0x8B31D4C4381C57F6ULL, "art\\characters\\weapons\\shields\\w_shd_srdb.mmh"},
	{0x8B31DEC4381C68F4ULL, "art\\characters\\weapons\\shields\\w_shd_srdb.msh"},
	{0x8B31DFAF0F1F2539ULL, "346585985.ogg"},
	{0x8B320724A706131BULL, "fen210_tevinter_ambusher6108323.cl"},
	{0x8B320C365311D4B1ULL, "vxlt_undercityserialkiller_wn_0_0.vlm"},
	{0x8B3323E51362865DULL, "vxht_brothel_n_58a_0_0.vlm"},
	{0x8B333C3F596FDCA6ULL, "gen00fl_anders6125781.cl"},
	{0x8B337BE139F078A2ULL, "gen00fl_aveline6139655.cl"},
	{0x8B3491BFA314D428ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangsign.phy"},
	{0x8B3500E350332ED1ULL, "764065863.ogg"},
	{0x8B352E13202FA590ULL, "zz_pro_debug6013509.cl"},
	{0x8B3540579EED274FULL, "dae000_porfiria6162649.cl"},
	{0x8B354AE38AC09453ULL, "hlac_lt_gallowscourtyard_0_0y.dds"},
	{0x8B355C0BBB01FED6ULL, "lt_0_lowtown_docks_a3_218_0.msh"},
	{0x8B35660BBB020FD4ULL, "lt_0_lowtown_docks_a3_218_0.mmh"},
	{0x8B356A0BBB0216E7ULL, "lt_0_lowtown_docks_a3_218_0.mao"},
	{0x8B35A4A8FDEE833DULL, "w_bonepitdestroyed_a3_80_0.mao"},
	{0x8B35A8A8FDEE8A76ULL, "w_bonepitdestroyed_a3_80_0.mmh"},
	{0x8B35B2A8FDEE9B74ULL, "w_bonepitdestroyed_a3_80_0.msh"},
	{0x8B35CDF051FCA206ULL, "dae260ar_wounded_coast.are"},
	{0x8B363E8A3AE92A4CULL, "6041152_f.fxe"},
	{0x8B365361735180F7ULL, "6122766_f.fxe"},
	{0x8B3720F23958D72EULL, "w_woundedcoast_49_0.mmh"},
	{0x8B372CF23958EBC5ULL, "w_woundedcoast_49_0.mao"},
	{0x8B373AF23959031CULL, "w_woundedcoast_49_0.msh"},
	{0x8B37B4EFE1FECA15ULL, "vxca_caverns2_7kq_0_0.vlm"},
	{0x8B37B9800BC2216AULL, "954286620.ogg"},
	{0x8B37F392B51EA877ULL, "6165412_m.fxe"},
	{0x8B3812AA269D40FEULL, "hlac_lt_0_lowtown_n_-262144_4z.dds"},
	{0x8B3872AC135992BCULL, "hlac_lt_0_lowtown_a2_n_c_1_3x.dds"},
	{0x8B3899ADEAB6EAC0ULL, "hlac_lt_undercityserialkiller_-1_37z.dds"},
	{0x8B38BB162C919D71ULL, "vxht_keep_6hn_0_0.vlm"},
	{0x8B38FD654E3F6DA5ULL, "vxlt_hangedman_26v_0_0.vlm"},
	{0x8B397145FADA5D18ULL, "489548378.ogg"},
	{0x8B3980E2A204D998ULL, "668306973.ogg"},
	{0x8B398D016029E8CEULL, "vxw_bonepit_77_0_0.vlm"},
	{0x8B3A82FEBD32E100ULL, "6116260_m.fxe"},
	{0x8B3AFDBE5C6A9001ULL, "vxht_mansion_n_5q1_0_0.vlm"},
	{0x8B3B337F2A4D3C16ULL, "w_woundedcoast_77_0.mao"},
	{0x8B3B3F7F2A4D50B5ULL, "w_woundedcoast_77_0.mmh"},
	{0x8B3B417F2A4D542BULL, "w_woundedcoast_77_0.msh"},
	{0x8B3C2C1E8B9FFEE6ULL, "6013349_m.fxe"},
	{0x8B3C548D3BDADA4AULL, "vxht_0_hightown_a3_bjf_0_0.vlm"},
	{0x8B3C6288BF9C9669ULL, "follower_banter6155775.cl"},
	{0x8B3C6A99C4D53512ULL, "6041575_m.fxe"},
	{0x8B3C9BB76D469AD8ULL, "vxlt_gallowsprison_n_6a4_0_0.vlm"},
	{0x8B3CA12F7A749252ULL, "6025106_m.fxe"},
	{0x8B3CA7525BA764A9ULL, "6000899_f.fxe"},
	{0x8B3CAEE49E2260CCULL, "mrl000_marethari6117699.cl"},
	{0x8B3CB2F65800F271ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf04.mmh"},
	{0x8B3CCCF658011E2FULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf04.msh"},
	{0x8B3D71F8C3EAECD1ULL, "561018705.ogg"},
	{0x8B3DF97AF1993824ULL, "vxw_bonepitdestroyed_a3_i7_0_0.vlm"},
	{0x8B3DFC836973DC71ULL, "vxht_0_hightown_n_bgd_0_0.vlm"},
	{0x8B3E3A3F6C4628CAULL, "6092294_m.fxe"},
	{0x8B3E6033F141C3A5ULL, "hlac_ht_0_hightown_a2_-196607_0z.dds"},
	{0x8B3E9DFBD352F201ULL, "vxlt_gallowsprison_n_5r7_0_0.vlm"},
	{0x8B3EA7EED28B826FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_action.gad"},
	{0x8B3EDC53BCECE582ULL, "vxht_brothel_n_5l6_0_0.vlm"},
	{0x8B3EFB6941DE40F6ULL, "6020083_f.fxe"},
	{0x8B3F06FF7E21BED9ULL, "244462623.ogg"},
	{0x8B3F50F904E9C23EULL, "6074433_f.fxe"},
	{0x8B3FC8F276FBA4A1ULL, "303429444.ogg"},
	{0x8B40532443D004A9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmwht_0.pwk"},
	{0x8B4056224CDE3132ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\gibgolem_s2.anb"},
	{0x8B40590AAC85A34BULL, "follower_banter6148740.cl"},
	{0x8B405DAC67BD022FULL, "hlac_lt_0_lowtown_a2_n_-65535_4z.dds"},
	{0x8B40E3D0609BCFDDULL, "vxht_0_hightown_a2_c6y_0_0.vlm"},
	{0x8B40F646080815C6ULL, "hlac_lt_0_lowtown_a2_n_c_2_1y.dds"},
	{0x8B414ED8CE2F69D0ULL, "llac_ht_keep_n_3734_1.dds"},
	{0x8B415585D72B1790ULL, "mag101_grace6034041.cl"},
	{0x8B4182D747CB9E06ULL, "gen00fl_anders6119218.cl"},
	{0x8B4185823D99ECE9ULL, "dre000_hireling16038597.cl"},
	{0x8B41AC1E9D01E9DEULL, "6195858_m.fxe"},
	{0x8B41E27FC29E0B18ULL, "6033392_m.fxe"},
	{0x8B426E0955EB685DULL, "mcr370_meredith.bnk"},
	{0x8B4282531A996C69ULL, "6132652_m.fxe"},
	{0x8B42DAAEA9BCB813ULL, "6115562_f.fxe"},
	{0x8B42F03C96262D3FULL, "qun231cr_varnell_minion1.utc"},
	{0x8B42F401B75DABD8ULL, "925175081.ogg"},
	{0x8B438A43BF710809ULL, "6136649_m.fxe"},
	{0x8B4408BDD13A36B9ULL, "vrc221_hugin6042687.cl"},
	{0x8B442CD3551C1E9EULL, "hlac_w_sundermount_-262150_0x.dds"},
	{0x8B448DF3A36F2E65ULL, "follower_banter6163172.cl"},
	{0x8B449810C69E87D2ULL, "clf221_jalen6070390.cl"},
	{0x8B44CED23819641BULL, "698368889.ogg"},
	{0x8B453FF23965492DULL, "w_woundedcoast_49_0.anb"},
	{0x8B4543D6FE5DE1FFULL, "vxw_bonepitdestroyed_a3_1rj_0_0.vlm"},
	{0x8B45500955ED6B65ULL, "mcr370_meredith.cnv"},
	{0x8B4552732D835A0AULL, "mag210_arianni6024580.cl"},
	{0x8B45BCEED291C41CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_action.evt"},
	{0x8B46888680A63724ULL, "6016949_f.fxe"},
	{0x8B46AA30F7ADC382ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha15a.msh"},
	{0x8B46AC30F7ADC738ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha15a.mmh"},
	{0x8B4753BCEFA1C5CFULL, "271636402.ogg"},
	{0x8B47D50D00CCAAE9ULL, "vxlt_gallowstemplar_n_34g_0_0.vlm"},
	{0x8B47F6ED1C13851AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_magb_0n.dds"},
	{0x8B482CAC65CCFF88ULL, "6112299_m.fxe"},
	{0x8B48C9831877FA07ULL, "llac_lt_undercityencounter_-1_43.dds"},
	{0x8B49896831396E80ULL, "llac_lt_0_lowtown_docks_-65538_4.dds"},
	{0x8B499E4097F18201ULL, "dae000_orana6149061.cl"},
	{0x8B49D87114428C82ULL, "mag320_orsino6071100.cl"},
	{0x8B4A0CE676003ECDULL, "art\\vfx\\combat\\vfx_ha_ar_armoff_act_c.mmh"},
	{0x8B4A1379270E6AD6ULL, "6079024_m.fxe"},
	{0x8B4A157FCF65FA2BULL, "dae000cr_grand_cleric6124694.cl"},
	{0x8B4A792FB758910EULL, "hlac_dr_primevalentrance_1726_7y.dds"},
	{0x8B4A7BACFDE6A899ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_3y.dds"},
	{0x8B4B6A814FDA6038ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wi_basic.phy"},
	{0x8B4CA8CB21FEC5EDULL, "combat_ar_nmv_default.gda"},
	{0x8B4D76B439E8AD30ULL, "6045336_f.fxe"},
	{0x8B4DA67F29AB4BA7ULL, "mag360tr_bsc_outskirts.ncs"},
	{0x8B4DC7872EE81FF5ULL, "vxlt_0_lowtown_a3_n_c_csj_0_0.vlm"},
	{0x8B4DD805DD889FE8ULL, "hlac_w_woundedcoast_262144_0x.dds"},
	{0x8B4E66E60D29AB36ULL, "645674307.ogg"},
	{0x8B4F63F2FBC0AFB7ULL, "qcr220_orsino6061775.cl"},
	{0x8B4F6ADC9B373A1AULL, "hlac_lt_gallowstemplar_2382_2y.dds"},
	{0x8B4FDAFC96DF8A44ULL, "6206602_m.fxe"},
	{0x8B5002BAFE29A1A8ULL, "vxht_0_hightown_n_aul_0_0.vlm"},
	{0x8B50B330F7B627C3ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha15a.phy"},
	{0x8B51169F2D242C4DULL, "follower_banter6167046.cl"},
	{0x8B517CA196DEE8BAULL, "vxw_encounter3_dm_0_0.vlm"},
	{0x8B51814DC9BF1865ULL, "6111609_f.fxe"},
	{0x8B51B03DFF9508DFULL, "ts_chantry_acc_table.gda"},
	{0x8B51F1F5654B9997ULL, "mcr000cr_leliana.utc"},
	{0x8B52C4EED29C5C35ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_kill_master_action.ani"},
	{0x8B52CDE4D38C2FA3ULL, "904747860.ogg"},
	{0x8B52E315F1678AEEULL, "w_encounter3_145_0.phy"},
	{0x8B52FCDD3A16E360ULL, "5657058.ogg"},
	{0x8B5322DE09CF0388ULL, "vxlt_0_lowtown_docks_n_byn_0_0.vlm"},
	{0x8B532FF6219BC03AULL, "w_sundermount_37_0.rml"},
	{0x8B543115BF34042BULL, "art\\characters\\playercharacter\\dwarfmale\\boo_nuda.phy"},
	{0x8B5441C4D2C622F5ULL, "vxdr_slavers_1t8_0_0.vlm"},
	{0x8B548E3D92895781ULL, "557603296.ogg"},
	{0x8B561866842F706BULL, "429353896.ogg"},
	{0x8B5636377B7A2629ULL, "6152742_m.fxe"},
	{0x8B5643FB928C04B9ULL, "6043277_m.fxe"},
	{0x8B565C29F22DC5F4ULL, "face2_s1y.psh.fxo"},
	{0x8B56760615A98CCEULL, "synccombatanims.gda"},
	{0x8B56FAA7E00B0CA0ULL, "vxht_0_hightown_n_bwi_0_0.vlm"},
	{0x8B57357F836D79E6ULL, "gen_im_acc_blt_act3_04.uti"},
	{0x8B5799AE4F687760ULL, "76266902.ogg"},
	{0x8B57D85F66731720ULL, "vxlt_gallowstemplar_n_43h_0_0.vlm"},
	{0x8B581ABDA74EC4D0ULL, "gen00fl_aveline6139756.cl"},
	{0x8B5878DFAFFE97C5ULL, "llac_lt_0_lowtown_a2_n_-65535_0.dds"},
	{0x8B5898FB9A6B60BAULL, "vxht_0_hightown_n_btz_0_0.vlm"},
	{0x8B58CAC53E0654C4ULL, "vrc321ip_shade_proj_target.utp"},
	{0x8B590F04052F09ABULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_bridge8x.phy"},
	{0x8B5956661E77646DULL, "6072211_m.fxe"},
	{0x8B5976B320D36B3AULL, "vxlt_gallowstemplar_n_46y_0_0.vlm"},
	{0x8B59D4C2464F5FA8ULL, "llac_lt_0_lowtown_docks_a3_n_-65536_0.dds"},
	{0x8B59F4CD664849B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\disengage.ani"},
	{0x8B5A10F621A1AA2EULL, "w_sundermount_37_0.phy"},
	{0x8B5A321D6828E2A2ULL, "clf260_harley6069938.cl"},
	{0x8B5A73FBB15D4650ULL, "lgt000cr_baz_ant_ass1.utc"},
	{0x8B5A909924417126ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls105a_0d.dds"},
	{0x8B5ADD6E2774C4AAULL, "llac_lt_0_lowtown_docks_-131074_3.dds"},
	{0x8B5B4F28AB2A4DFBULL, "llac_lt_0_lowtown_docks_a3_n_131071_4.dds"},
	{0x8B5B4F7E52F62516ULL, "6076810_m.fxe"},
	{0x8B5B879E0E07029EULL, "dae150_chandan6205599.cl"},
	{0x8B5BD4F0521D7DF1ULL, "dae260ar_wounded_coast.lst"},
	{0x8B5BDDB0B95B58D5ULL, "6052833_m.fxe"},
	{0x8B5BEC709D197960ULL, "88804143.ogg"},
	{0x8B5BF15F056A835AULL, "vxht_chantry_n_2yq_0_0.vlm"},
	{0x8B5C4F023F7FEFBAULL, "art\\characters\\creatures\\mubari\\animation\\dblw_loop_v1.gad"},
	{0x8B5D395B58582CBFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.gad"},
	{0x8B5D6F83A650D9EAULL, "182757624.ogg"},
	{0x8B5D7E1723BB746CULL, "gen00fl_isabela6099701.cl"},
	{0x8B5DC78A448800DDULL, "vxht_0_hightown_a2_c8w_0_0.vlm"},
	{0x8B5DCD33996EB678ULL, "qun230cr_guild_contact6068694.cl"},
	{0x8B5DE833F57275FCULL, "gen00fl_aveline6122811.cl"},
	{0x8B5EC375306E1C4FULL, "793603429.ogg"},
	{0x8B5EC67B711AF763ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredround.mao"},
	{0x8B5ECA7B711AFDB0ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredround.mmh"},
	{0x8B5ECF8EB46ED5E5ULL, "68774061.ogg"},
	{0x8B5ED73F3BB7EF10ULL, "art\\characters\\playercharacter\\elffemale\\cth_come.phy"},
	{0x8B5ED87B711B15CAULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_rugredround.msh"},
	{0x8B5EFEDB22C795FEULL, "6076567_f.fxe"},
	{0x8B5F24BBDBC1E745ULL, "dae000ip_crowd_em1_talk14.utp"},
	{0x8B5F56FEC95F4F80ULL, "vxlt_hangedman_2mj_0_0.vlm"},
	{0x8B5F7300D32AE30DULL, "6107072_m.fxe"},
	{0x8B5FBA8D05491C41ULL, "gen_im_arm_hlm_lgt_set_high.uti"},
	{0x8B604E5BBE89EA57ULL, "vxw_bonepit_1p4_0_0.vlm"},
	{0x8B60578EB25CE038ULL, "6034928_m.fxe"},
	{0x8B6060E848BD1E4CULL, "6218937_m.fxe"},
	{0x8B608AFC9C8AE550ULL, "524454543.ogg"},
	{0x8B6090AB738720AAULL, "6042156_m.fxe"},
	{0x8B60957824968CDDULL, "character_character_p101.psh.fxo"},
	{0x8B60A2AA97FA30ACULL, "113005910.ogg"},
	{0x8B61ECF052221188ULL, "dae260ar_wounded_coast.ncs"},
	{0x8B622F0FB71D42A2ULL, "626109574.ogg"},
	{0x8B624C3F690E5F58ULL, "462838890.ogg"},
	{0x8B626E0A2246BAC6ULL, "6018564_m.fxe"},
	{0x8B62A374133EE470ULL, "tal_tau_cluc.mao"},
	{0x8B62D6B196AAE151ULL, "6151215_m.fxe"},
	{0x8B63948FBA00CF81ULL, "vxlt_hangedman_1i0_0_0.vlm"},
	{0x8B63AF92C9F7E334ULL, "dae000_bartrand6206162.cl"},
	{0x8B63EE11606118BFULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_drag_flemeth_real.gad"},
	{0x8B642E5B585E38ACULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.evt"},
	{0x8B64A9E10155C15CULL, "vxw_encounter2_camp_5k_0_0.vlm"},
	{0x8B64E2DB76ACEE00ULL, "zz_isa_debug6134422.cl"},
	{0x8B64F0B4CC43E2AEULL, "w_bonepit_170_0.msh"},
	{0x8B64FEB4CC43FAFFULL, "w_bonepit_170_0.mao"},
	{0x8B650AB4CC440E9CULL, "w_bonepit_170_0.mmh"},
	{0x8B65433F2E33B8D8ULL, "6121148_f.fxe"},
	{0x8B654C67CB6DD71DULL, "one110_tree_minder6140865.cl"},
	{0x8B65E32E208DC5D3ULL, "256293020.ogg"},
	{0x8B66933B3730F286ULL, "6085997_m.fxe"},
	{0x8B66C9840F1B140BULL, "gen00fl_varric6118726.cl"},
	{0x8B6708E3E25996DAULL, "6208354_m.fxe"},
	{0x8B6714BC719E192DULL, "6044636_m.fxe"},
	{0x8B67DF4A6B4B2D7BULL, "6095917_m.fxe"},
	{0x8B67F01ACFF828BEULL, "vxlt_warehouse_n_q7_0_0.vlm"},
	{0x8B67FCCD665494A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\disengage.evt"},
	{0x8B67FED089416CD0ULL, "1032949172.ogg"},
	{0x8B6847289A9E316CULL, "vxw_woundedcoast_139_0_0.vlm"},
	{0x8B693999B58CBF2BULL, "dae000_keep_servant_1_sta6160887.cl"},
	{0x8B694C15095F2FA2ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_dsmla_0s.dds"},
	{0x8B69521A3F5625F1ULL, "vxht_0_hightown_n_bhy_0_0.vlm"},
	{0x8B6967BE19C41801ULL, "m_mag100ar_encounter.bnk"},
	{0x8B69DC876A489F0DULL, "782229164.ogg"},
	{0x8B6A1137CD34A5C3ULL, "mcr370_assignments6081704.cl"},
	{0x8B6AA78E7BC26A90ULL, "w_woundedcoast_62_0.rml"},
	{0x8B6ACB6317E1A03DULL, "ave260ip_signal_fire1.ncs"},
	{0x8B6B0FD023A25A67ULL, "hlac_lt_0_lowtown_a3_n_c_65537_2z.dds"},
	{0x8B6B5F20B2B06D2FULL, "character_character_p4.psh.fxo.dx11"},
	{0x8B6B855F84B143ADULL, "qcr220_meredith6054722.cl"},
	{0x8B6C0E8FC98D9D39ULL, "633369009.ogg"},
	{0x8B6C2A60D96620E5ULL, "6108301_m.fxe"},
	{0x8B6CEE0D919EB471ULL, "hlac_lt_0_lowtown_a2_n_-262144_3z.dds"},
	{0x8B6D08D46B718E2AULL, "vxw_woundedcoast_1g3_0_0.vlm"},
	{0x8B6D0DE39728FF4FULL, "vxw_encounter2_camp_qp_0_0.vlm"},
	{0x8B6D369DDC0AA2B1ULL, "240125728.ogg"},
	{0x8B6D39215C227769ULL, "follower_barks6153590.cl"},
	{0x8B6D55DC0DF9CFA7ULL, "dae000_dalish_apprentices6147608.cl"},
	{0x8B6DDC0F31C37CF0ULL, "mag311_nyssa6041981.cl"},
	{0x8B6DE5BAC7CC419BULL, "6116043_m.fxe"},
	{0x8B6E1DCD665938F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\disengage.gad"},
	{0x8B6EC038B08002D7ULL, "vxw_bonepit_1u7_0_0.vlm"},
	{0x8B6EC08A282EB291ULL, "770441088.ogg"},
	{0x8B6EC58D66D977AAULL, "vxlt_darktown_54q_0_0.vlm"},
	{0x8B6EF379C8836B06ULL, "art\\characters\\creatures\\mubari\\animation\\com_pant.evt"},
	{0x8B6F4F0A8AB94019ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye04b.phy"},
	{0x8B7002780197752FULL, "vxlt_0_lowtown_docks_b6g_0_0.vlm"},
	{0x8B706A023F90C6A0ULL, "art\\characters\\creatures\\mubari\\animation\\dblw_loop_v1.ani"},
	{0x8B70A71CD0643704ULL, "vxht_chantry_n_28h_0_0.vlm"},
	{0x8B70EE98A4C8350CULL, "462707783.ogg"},
	{0x8B7139D48DE49E0AULL, "586672752.ogg"},
	{0x8B71EE4C77BCC534ULL, "6077672_f.fxe"},
	{0x8B72165B586A4D05ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.ani"},
	{0x8B721B32CBFC39B9ULL, "w_sundermount_178_0.msh"},
	{0x8B722532CBFC4AA7ULL, "w_sundermount_178_0.mmh"},
	{0x8B722932CBFC51F4ULL, "w_sundermount_178_0.mao"},
	{0x8B724528B5A898C7ULL, "vxlt_gallowstemplar_n_5sz_0_0.vlm"},
	{0x8B72572FF4B6F890ULL, "822444444.ogg"},
	{0x8B727FE0DD2551E3ULL, "vrc221_bartrand6043177.cl"},
	{0x8B72A702E6317FD9ULL, "616058090.ogg"},
	{0x8B72E1D1C0AC8547ULL, "dae000_pc_home_ambients6208548.cl"},
	{0x8B72E2B084052446ULL, "art\\vfx\\combat\\vfx_str_leap_act_c.impa.evt"},
	{0x8B72FA80AF24C021ULL, "208762459.ogg"},
	{0x8B731189F377BD92ULL, "sarah_test_stg.stg"},
	{0x8B738F05B30845C3ULL, "vxht_keep_7af_0_0.vlm"},
	{0x8B742CF17CA52A31ULL, "6067385_m.fxe"},
	{0x8B757946014F689EULL, "vxlt_0_lowtown_h94_0_0.vlm"},
	{0x8B757CB57E5791BAULL, "vxht_0_hightown_n_bzt_0_0.vlm"},
	{0x8B75866BE9480329ULL, "6146844_m.fxe"},
	{0x8B7596081C6464BCULL, "vxca_caverns1_mines_djf_0_0.vlm"},
	{0x8B75B0856CFD40E0ULL, "llac_w_blightlands_393215_0.dds"},
	{0x8B75E9346F82E109ULL, "220958955.ogg"},
	{0x8B760A79C889B151ULL, "art\\characters\\creatures\\mubari\\animation\\com_pant.gad"},
	{0x8B762E8F09839D5BULL, "942912440.ogg"},
	{0x8B7662C3BA4D1E3EULL, "vxdr_primevaltomb_7e4_0_2.vlm"},
	{0x8B7683001A52C87CULL, "vxlt_gallowstemplar_f_416_0_0.vlm"},
	{0x8B76E1AE2D1F7012ULL, "6086273_m.fxe"},
	{0x8B7722F4C71DB5FDULL, "zz_dae_debug6045830.cl"},
	{0x8B77B2561F3A9A72ULL, "6137674_m.fxe"},
	{0x8B78403C7306A47FULL, "hlac_ca_caverns1_24425_1z.dds"},
	{0x8B785E62FF000148ULL, "llac_dr_slavers_3215_7.dds"},
	{0x8B78971A692151A9ULL, "6065310_m.fxe"},
	{0x8B78C55C7C849603ULL, "6071951_m.fxe"},
	{0x8B78CB1160733905ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_drag_flemeth_real.ani"},
	{0x8B78D518E956B45AULL, "vxht_chantry_n_2os_0_0.vlm"},
	{0x8B79555BCBE7CD5CULL, "face_face_p14.psh.fxo.dx11"},
	{0x8B7994F52A5309DDULL, "85750231.ogg"},
	{0x8B79F9B0840B6991ULL, "art\\vfx\\combat\\vfx_str_leap_act_c.impa.gad"},
	{0x8B7A3D6C59291D73ULL, "91736898.ogg"},
	{0x8B7A7C52BD3D0587ULL, "llac_ca_caverns2prc_5802_50.dds"},
	{0x8B7A86A9E9388888ULL, "825651201.ogg"},
	{0x8B7A951F2717D6D8ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyo.phy"},
	{0x8B7B0F79972F83EFULL, "vxlt_darktown_4t8_0_1.vlm"},
	{0x8B7B7FC0891F17CFULL, "vxw_encounter2_camp_ls_0_0.vlm"},
	{0x8B7C0B79C88E1EFFULL, "art\\characters\\creatures\\mubari\\animation\\com_pant.ani"},
	{0x8B7C2F171237A968ULL, "ser270pt_suspect.ncs"},
	{0x8B7C6929FA15B843ULL, "6081656_f.fxe"},
	{0x8B7D376A58CF902AULL, "vxlt_darktown_59r_0_0.vlm"},
	{0x8B7DC4D95178511EULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_idle.ani"},
	{0x8B7DEB797912A0CBULL, "6016310_m.fxe"},
	{0x8B7E0089E37B3D72ULL, "6060494_f.fxe"},
	{0x8B803BC3995B7F44ULL, "lt_backalley_g_2_0_lc.mmh"},
	{0x8B805F3B352D1A71ULL, "lt_gallowscourtyard_n_43_0_ll.mmh"},
	{0x8B806DCC65F086C9ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v2.ani"},
	{0x8B80B33FDD26F17BULL, "100469753.ogg"},
	{0x8B80B507DD8734EDULL, "hlac_dr_slavers_3216_3z.dds"},
	{0x8B80FAB084118A3FULL, "art\\vfx\\combat\\vfx_str_leap_act_c.impa.ani"},
	{0x8B81445959502010ULL, "6110401_f.fxe"},
	{0x8B816C9BD05841F8ULL, "6127699_m.fxe"},
	{0x8B818F1A89B6618AULL, "dae320ip_at_player_home.utp"},
	{0x8B81920FD25335CEULL, "6044356_m.fxe"},
	{0x8B81C0634643AC92ULL, "vxca_cave_2k0_0_0.vlm"},
	{0x8B81E1B41DE19EC4ULL, "hlac_w_bonepitmining_a2_65539_0x.dds"},
	{0x8B825AADF8F749E4ULL, "27047617.ogg"},
	{0x8B8288FDE05B1BDEULL, "6092269_m.fxe"},
	{0x8B828E26299E3600ULL, "mag101_decimus.cnv"},
	{0x8B82A46F8D1E23BCULL, "179293380.ogg"},
	{0x8B82A980284F9F59ULL, "vxht_mansion2b_31l_0_0.vlm"},
	{0x8B8359C566EF18A2ULL, "dae000_flemeth6000729.cl"},
	{0x8B83A2726E92F61EULL, "ca_cave_probe_2nj.dds"},
	{0x8B8546C3FBB90216ULL, "llac_lt_undercityencounter_-1_8.dds"},
	{0x8B85894954F1E52CULL, "hlac_w_encounter2_196606_0z.dds"},
	{0x8B86702629A1EC38ULL, "mag101_decimus.bnk"},
	{0x8B8688CC65F5204BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v2.gad"},
	{0x8B86AD73307B5807ULL, "mcr310_cut_chantry_blows_up6087575.cl"},
	{0x8B86D801113CE99EULL, "vxlt_0_lowtown_eqm_0_0.vlm"},
	{0x8B87279A13322E13ULL, "ser211ip_to_foundry2.utp"},
	{0x8B874987A7918FBAULL, "gen00fl_fenris6105554.cl"},
	{0x8B87527E9A163DDBULL, "vxht_bartrand_n_640_0_0.vlm"},
	{0x8B87592B4F0F22B2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masd_0t.dds"},
	{0x8B877BAA5600C49EULL, "setcombatresult.ncs"},
	{0x8B87C56596A24253ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_orsa_0s.dds"},
	{0x8B87CA69B8216E68ULL, "vxht_brothel_n_614_0_0.vlm"},
	{0x8B888EFF3CD1F57BULL, "vxlt_blackemporium_z6_0_0.vlm"},
	{0x8B8918A38613F4AAULL, "vxlt_undercitydungeon_yh_0_0.vlm"},
	{0x8B893FA767C2724CULL, "and121_karl6093882.cl"},
	{0x8B89BEEC54A24A8FULL, "vxht_keep_n_3zt_0_0.vlm"},
	{0x8B89F19D7B15304FULL, "vxw_encounter2_camp_gr_0_0.vlm"},
	{0x8B8A17588A39F7CEULL, "vxca_caverns1_k6j_0_0.vlm"},
	{0x8B8AAF04C6D49070ULL, "6162649_m.fxe"},
	{0x8B8AE9B4CC64A667ULL, "w_bonepit_170_0.phy"},
	{0x8B8B5B15356528FFULL, "vxdr_primevaltomb_8oj_0_0.vlm"},
	{0x8B8B93BF0F32F0B9ULL, "dae300_alistair_king6133803.cl"},
	{0x8B8BCB4ECC0473D7ULL, "179702401.ogg"},
	{0x8B8C3C8B7C57ABB4ULL, "lgt210_magu_selby6142048.cl"},
	{0x8B8C7F1F7725CE47ULL, "vxht_bartrand_n_3zo_0_0.vlm"},
	{0x8B8C9A8B31E43EE0ULL, "vxlt_0_lowtown_a2_n_cmn_0_0.vlm"},
	{0x8B8D26170B057881ULL, "6152535_m.fxe"},
	{0x8B8D85CC65FB39B0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v2.evt"},
	{0x8B8D8AF4C738E52EULL, "vxw_woundedcoast_1k3_0_0.vlm"},
	{0x8B8DCCEEE7E8DC47ULL, "vrc321_party_barks6095784.cl"},
	{0x8B8E1DA55E7BB3ACULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he3a.phy"},
	{0x8B8E2146BCA98C09ULL, "llac_lt_0_lowtown_a2_n_c_131072_5.dds"},
	{0x8B8ED7787255A5B9ULL, "llac_dr_slavers_3214_3.dds"},
	{0x8B8F28E29994C9C7ULL, "vxlt_gallowstemplar_n_5it_0_0.vlm"},
	{0x8B8F8DBDC81128E5ULL, "457961609.ogg"},
	{0x8B9030F148F5409AULL, "gen_im_cft_reg_g1.uti"},
	{0x8B9032403839C512ULL, "vxca_cave_2l1_0_0.vlm"},
	{0x8B90B621C40DFDEAULL, "vxlt_0_lowtown_n_e7x_0_0.vlm"},
	{0x8B90BC37C15A1807ULL, "vxlt_warehouse_n_18z_0_0.vlm"},
	{0x8B9110AB60915B31ULL, "6069185_m.fxe"},
	{0x8B918C060692F035ULL, "ser221_party_comments6116134.cl"},
	{0x8B919B98EB4026FDULL, "dae000_party_night_comment6201633.cl"},
	{0x8B91BBA6743B40E9ULL, "vxca_caverns2_95y_0_0.vlm"},
	{0x8B91DDD951892498ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_idle.gad"},
	{0x8B91F8146B1EB9A9ULL, "854225733.ogg"},
	{0x8B92252446232604ULL, "463816903.ogg"},
	{0x8B922EF0D2BC6C59ULL, "vxlt_gallowstemplar_6sl_0_0.vlm"},
	{0x8B92FECC9F1253BFULL, "vxw_woundedcoast_136_0_0.vlm"},
	{0x8B9328CB08939EAEULL, "ave100_donnic6105955.cl"},
	{0x8B933965C470D543ULL, "xx_extplaster.dds"},
	{0x8B935776A512AA32ULL, "hlac_lt_darktown_-1_38y.dds"},
	{0x8B938045D272B0ACULL, "vxdr_0_deeproads_cmf_0_0.vlm"},
	{0x8B93C96A7E4B5B2BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_enter.ani"},
	{0x8B93E215F19F1FB1ULL, "w_encounter3_145_0.mmh"},
	{0x8B93EE15F19F3452ULL, "w_encounter3_145_0.mao"},
	{0x8B93FC15F19F4C6FULL, "w_encounter3_145_0.msh"},
	{0x8B94AA98D46C91FFULL, "vxlt_0_lowtown_gc8_0_0.vlm"},
	{0x8B94D92613EF6BDAULL, "hlac_lt_hangedman_n_-1_20x.dds"},
	{0x8B954E6621149136ULL, "vxlt_hangedman_n_1ml_0_0.vlm"},
	{0x8B9551DF947A9624ULL, "gen00fl_fenris6112062.cl"},
	{0x8B95B8D2CD42E55AULL, "vxht_chantry_n_2em_0_0.vlm"},
	{0x8B9649BBBAD5F356ULL, "vxht_mansion2_3bn_0_0.vlm"},
	{0x8B965B32C332C5F3ULL, "pro000_friedrich6051045.cl"},
	{0x8B96624E7380ED6CULL, "llac_w_blightlands_196611_1.dds"},
	{0x8B9699F75EC7353AULL, "971388635.ogg"},
	{0x8B969B4010959C4BULL, "180244178.ogg"},
	{0x8B96E93A3C4C11A3ULL, "gen00fl_aveline6116633.cl"},
	{0x8B97A88E7BE95AF8ULL, "w_woundedcoast_62_0.anb"},
	{0x8B97B5BE1B83A4FAULL, "6089867_m.fxe"},
	{0x8B980946990620F4ULL, "hlac_lt_gallowscourtyard_65536_1x.dds"},
	{0x8B982D91099305C4ULL, "134738482.ogg"},
	{0x8B982DCA6E07268CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelwht.phy"},
	{0x8B98590A93D9A232ULL, "6114043_f.fxe"},
	{0x8B98E6F3A7BA7812ULL, "char_ext_test_day.arl"},
	{0x8B99B333FF23641AULL, "mag370_meredith6092199.cl"},
	{0x8B99E3FDB9CFD9FEULL, "llac_lt_gallowstemplar_n_2394_7.dds"},
	{0x8B9A05AEEEE06EFAULL, "148484858.ogg"},
	{0x8B9A0FF621D78CF1ULL, "w_sundermount_37_0.mmh"},
	{0x8B9A1BF621D7A092ULL, "w_sundermount_37_0.mao"},
	{0x8B9A29F621D7B8AFULL, "w_sundermount_37_0.msh"},
	{0x8B9A6FC02C54F64DULL, "1005824986.ogg"},
	{0x8B9A8F1CE76AB610ULL, "51792514.ogg"},
	{0x8B9AA0F816E56D26ULL, "500232053.ogg"},
	{0x8B9AD10C565DDA6AULL, "917915033.ogg"},
	{0x8B9B62C6D40015BEULL, "6136651_m.fxe"},
	{0x8B9BE0F10947EE53ULL, "vxht_mansion_n_6bm_0_0.vlm"},
	{0x8B9BF9BD69F773E8ULL, "dae000_tranquil_16165117.cl"},
	{0x8B9CF81DB047A975ULL, "vxdr_primevalentrance_1nz_0_0.vlm"},
	{0x8B9D2157A227D5F6ULL, "lt_0_lowtown_docks_205_0.mao"},
	{0x8B9D2D57A227E995ULL, "lt_0_lowtown_docks_205_0.mmh"},
	{0x8B9D2F57A227ED0BULL, "lt_0_lowtown_docks_205_0.msh"},
	{0x8B9D511AC16A880EULL, "vxlt_gallowstemplar_3sz_0_0.vlm"},
	{0x8B9DD5FE45B679A2ULL, "ter_brickround01_s.dds"},
	{0x8B9E0E32CC215EDCULL, "w_sundermount_178_0.phy"},
	{0x8B9E32CEA28B595DULL, "hlac_lt_0_lowtown_docks_a3_-4_3x.dds"},
	{0x8B9E73C955C42720ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urvb_0n.dds"},
	{0x8B9ECFF12189931BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09a.mmh"},
	{0x8B9EEDF12189C605ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09a.msh"},
	{0x8B9FC9FC85F89C76ULL, "mag100_thrask6031747.cl"},
	{0x8BA098DCB5B97A21ULL, "narrative_varric6172537.cl"},
	{0x8BA0A5272953D87EULL, "6175666_f.fxe"},
	{0x8BA112282935937AULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.cess.gad"},
	{0x8BA1B16A7E576FA2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_enter.evt"},
	{0x8BA1F222C468C92CULL, "vxdr_0_deeproads_cjg_0_0.vlm"},
	{0x8BA20CA742750E0DULL, "qcr270_fake_meredith6066328.cl"},
	{0x8BA22468C37420E0ULL, "514323916.ogg"},
	{0x8BA25320E2EE95FFULL, "343777116.ogg"},
	{0x8BA27F39A7BE35D4ULL, "6149876_f.fxe"},
	{0x8BA2937F29F36BFEULL, "mag360tr_bsc_outskirts.utt"},
	{0x8BA2D93E5E42ACC8ULL, "follower_banter6172937.cl"},
	{0x8BA314AC3988F726ULL, "ht_0_hightown_a2_19_0.rml"},
	{0x8BA35181BAC88D9AULL, "w_blightlands.rml"},
	{0x8BA35978263DA346ULL, "630271726.ogg"},
	{0x8BA3D06D58CDC8B3ULL, "gen_im_trs_tmp_hvy_boo_r1.uti"},
	{0x8BA3EFC0D48D16D9ULL, "hlac_lt_0_lowtown_a2_-65535_7z.dds"},
	{0x8BA412DA676C8A7DULL, "6202184_m.fxe"},
	{0x8BA4196D47A24A3CULL, "951555569.ogg"},
	{0x8BA4340A59C30511ULL, "6048481_m.fxe"},
	{0x8BA491223114F11DULL, "531681607.ogg"},
	{0x8BA561A6CA9C8CD2ULL, "vxht_mansion_n_6bh_0_0.vlm"},
	{0x8BA561FFEF3A387AULL, "723358325.ogg"},
	{0x8BA677960CB732C5ULL, "115662008.ogg"},
	{0x8BA71528293A0391ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.cess.evt"},
	{0x8BA7578150289ECDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wi_basic.msh"},
	{0x8BA765815028B600ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wi_basic.mao"},
	{0x8BA769815028BD53ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wi_basic.mmh"},
	{0x8BA76FCAE920FCDDULL, "ser221_gamlen6123412.cl"},
	{0x8BA887AC785864D7ULL, "vxw_bonepit_1i3_0_0.vlm"},
	{0x8BA8A03D5A5226B6ULL, "vxlt_0_lowtown_eu1_0_0.vlm"},
	{0x8BA8B06A7E5D8CB5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_enter.gad"},
	{0x8BA8FE8E52CB8E4EULL, "hlac_lt_0_lowtown_-2_2z.dds"},
	{0x8BA9703638344437ULL, "vxw_bonepit_1r2_0_0.vlm"},
	{0x8BA9745C17CD1C99ULL, "6126437_m.fxe"},
	{0x8BA988E619246713ULL, "vxca_caverns2prc_7gz_0_0.vlm"},
	{0x8BA9A6AC1FECF0AFULL, "gen00fl_aveline6122128.cl"},
	{0x8BA9BE04D0B6379EULL, "mcr310_argument6074362.cl"},
	{0x8BAA0C717C2BACB9ULL, "hlac_lt_0_lowtown_n_-2_2z.dds"},
	{0x8BAA14F49EA379B5ULL, "6118313_m.fxe"},
	{0x8BAA31E33D8DE5A2ULL, "vxlt_0_lowtown_a2_n_f4z_0_0.vlm"},
	{0x8BAA4ABC3F1FBEB4ULL, "vxlt_undercitydungeon_104_0_0.vlm"},
	{0x8BAA6EAEAB25162EULL, "vxw_woundedcoast_1q5_0_0.vlm"},
	{0x8BAA835763CAE348ULL, "6013326_f.fxe"},
	{0x8BAADD63C11B0E4EULL, "hlac_ca_caverns2_5802_48y.dds"},
	{0x8BAB3873B191DB14ULL, "441584460.ogg"},
	{0x8BAB8BFB83CED560ULL, "515586929.ogg"},
	{0x8BABAAFA1AC93877ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scl_d.mao"},
	{0x8BAC0401C172FB1AULL, "vxdr_primevaltomb_569_0_0.vlm"},
	{0x8BAC21E942D43BE5ULL, "mrl000_marethari6129293.cl"},
	{0x8BACC1C0D9D1DC81ULL, "vxca_caverns1_j28_0_0.vlm"},
	{0x8BACC30C93C0142CULL, "vxht_chantry_2ba_0_0.vlm"},
	{0x8BACDC03417B07B8ULL, "dae000cr_grand_cleric6124758.cl"},
	{0x8BAD0D903545DDB1ULL, "arc_mit_01.mao"},
	{0x8BAD701DB7D2C893ULL, "hf_thug01.mor"},
	{0x8BAE055F0737AF73ULL, "6032288_m.fxe"},
	{0x8BAE3F41BB08E9BCULL, "224246556.ogg"},
	{0x8BAE6EB838F99E59ULL, "vxca_caverns1_mines_j29_0_0.vlm"},
	{0x8BAE892F3BF2DFF8ULL, "vxw_sundermount_2zy_0_0.vlm"},
	{0x8BAEA2F3E56B1CB5ULL, "pro000ip_corpse_3.utp"},
	{0x8BAF1E49160C7710ULL, "6071160_f.fxe"},
	{0x8BAFD52678C645D4ULL, "443142370.ogg"},
	{0x8BAFD5327F5B0EF2ULL, "3784398.ogg"},
	{0x8BAFD92B99F8FCA2ULL, "6118339_f.fxe"},
	{0x8BAFE28682FE84BFULL, "vxw_woundedcoast_190_0_0.vlm"},
	{0x8BB02EF423AAFDE8ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_5craggyrock.msh"},
	{0x8BB040F423AB1C79ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_5craggyrock.mao"},
	{0x8BB04CF423AB3012ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_5craggyrock.mmh"},
	{0x8BB075BE1A0032B3ULL, "m_mag100ar_encounter.wbd"},
	{0x8BB0D646428B5554ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x2_0.pwk"},
	{0x8BB0F9ADFA6A23B8ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_gravec_d.dds"},
	{0x8BB153274897D093ULL, "vxht_brothel_n_2lt_0_0.vlm"},
	{0x8BB1763D1E6D28E3ULL, "vxlt_gallowscourtyard_71g_0_0.vlm"},
	{0x8BB18B29E2784805ULL, "vxw_sundermount_2kc_0_0.vlm"},
	{0x8BB1B88757A0C46FULL, "vxw_woundedcoast_1be_0_0.vlm"},
	{0x8BB1BC63181D8904ULL, "ave260ip_signal_fire1.utp"},
	{0x8BB1D5F73DF1B49FULL, "vxlt_0_lowtown_a2_n_gxf_0_0.vlm"},
	{0x8BB2658DB088637FULL, "hlac_w_sundermount_65540_0z.dds"},
	{0x8BB27EE2FBA39FCAULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.impa.gad"},
	{0x8BB293792D257BDBULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.dura.ani"},
	{0x8BB31000EFE2D58BULL, "vxht_0_hightown_an1_0_0.vlm"},
	{0x8BB310EF0AA2ECA8ULL, "617170518.ogg"},
	{0x8BB3300AC6DE63C9ULL, "vxw_encounter2_fo_0_0.vlm"},
	{0x8BB391D8C99722E7ULL, "dae000cr_grand_cleric6144098.cl"},
	{0x8BB3D24E5D5AF80AULL, "vxca_caverns1_mines_cjs_0_0.vlm"},
	{0x8BB433031A93452FULL, "583997723.ogg"},
	{0x8BB4D316343F2F8DULL, "6053382_m.fxe"},
	{0x8BB52D2829466A60ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_mod_c.cess.ani"},
	{0x8BB5313E53580F55ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_scalpm1.mao"},
	{0x8BB55AED073E5168ULL, "vxht_0_hightown_a2_anb_0_0.vlm"},
	{0x8BB6EE7112F705B9ULL, "vxca_cave_1c8_0_0.vlm"},
	{0x8BB7121A4C483F36ULL, "vxlt_0_lowtown_ep2_0_0.vlm"},
	{0x8BB71792FB3F1528ULL, "w_bonepitdestroyed_a3_68_0.mmh"},
	{0x8BB72392FB3F29CBULL, "w_bonepitdestroyed_a3_68_0.mao"},
	{0x8BB73592FB3F4852ULL, "w_bonepitdestroyed_a3_68_0.msh"},
	{0x8BB78B27BB67653BULL, "vxlt_gallowscourtyard_45x_0_0.vlm"},
	{0x8BB7E63909C33519ULL, "6123737_m.fxe"},
	{0x8BB8A73FF79345A2ULL, "hlac_lt_0_lowtown_a2_n_c_131070_2x.dds"},
	{0x8BB8A8B1B6BFFBA7ULL, "6135039_m.fxe"},
	{0x8BB8B11932D09C6CULL, "hlac_lt_undercityserialkiller_-1_39x.dds"},
	{0x8BB906C30D26A49BULL, "6198023_m.fxe"},
	{0x8BB93B5642EA6240ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masf.phy"},
	{0x8BB95EE7AD181621ULL, "6206984_m.fxe"},
	{0x8BB961E2FBA98D61ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.impa.evt"},
	{0x8BB9B46DA787A898ULL, "ht_0_hightown_n_51_0_lc.mmh"},
	{0x8BBA6D5868091C74ULL, "895549587.ogg"},
	{0x8BBAA363742EC8CCULL, "mag321_comtesse_de_launcet6043322.cl"},
	{0x8BBAC2EECFCC97ADULL, "vxca_cave_2cr_0_1.vlm"},
	{0x8BBAD85BD502983AULL, "vxdr_primevalentrance_3bw_0_0.vlm"},
	{0x8BBAE51D6890766BULL, "vrc321_varric6087058.cl"},
	{0x8BBB1336A7CA4157ULL, "hlac_dr_primevaltomb_9621_5x.dds"},
	{0x8BBB8F19F9106D8CULL, "6149655_m.fxe"},
	{0x8BBC015D875C945EULL, "zz_followers_banter6158115.cl"},
	{0x8BBC77FA59A55438ULL, "vxht_bartrand_n_66a_0_0.vlm"},
	{0x8BBCBD51EAA6F3E3ULL, "plt_cod_hst_vehnadahl.plo"},
	{0x8BBCD728441468F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_reaction.ani"},
	{0x8BBD12B76C3ABDC7ULL, "gen_im_trs_crc_lgt_cht_r1.uti"},
	{0x8BBD2A9FBD3E1133ULL, "vxlt_0_lowtown_a3_n_c_ewr_0_0.vlm"},
	{0x8BBD30DABB59430BULL, "humanchild.fxa"},
	{0x8BBD424B3E8A15A4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_zbm_0s.dds"},
	{0x8BBD59557186346AULL, "vxht_0_hightown_90b_0_0.vlm"},
	{0x8BBD97E9590663D5ULL, "ave200_donnic6110674.cl"},
	{0x8BBDF29B95368F01ULL, "246596298.ogg"},
	{0x8BBE91ABEA740002ULL, "zz_followers_banter6158865.cl"},
	{0x8BBF84CF5427AF9AULL, "vxw_bonepitdestroyed_a3_16o_0_0.vlm"},
	{0x8BBF958E7C0A8C0DULL, "w_woundedcoast_62_0.msh"},
	{0x8BBFA38E7C0AA440ULL, "w_woundedcoast_62_0.mao"},
	{0x8BBFA78E7C0AAA93ULL, "w_woundedcoast_62_0.mmh"},
	{0x8BBFFD06D46E6085ULL, "vxw_sundermount_2lb_0_0.vlm"},
	{0x8BC003D857508C2DULL, "vxht_mansion2b_20j_0_0.vlm"},
	{0x8BC02A644996DCEFULL, "vxw_woundedcoast_1md_0_0.vlm"},
	{0x8BC0317A7C3041DDULL, "mer221cr_sabin.utc"},
	{0x8BC033C5347DE1EDULL, "dae200cr_viscount6087173.cl"},
	{0x8BC09B792D31C6D2ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.dura.evt"},
	{0x8BC0B6B68DB4E4E2ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_011a_0t.dds"},
	{0x8BC17BBAA00A033AULL, "mer320_hubert6053189.cl"},
	{0x8BC1A1E7B8D47C49ULL, "vxw_encounter2_an_0_0.vlm"},
	{0x8BC1A6ECC33B3B76ULL, "994064749.ogg"},
	{0x8BC1AE4629DADDC2ULL, "558843896.ogg"},
	{0x8BC1E77B302109C3ULL, "vxlt_gallowstemplar_n_55t_0_0.vlm"},
	{0x8BC2080A36A07297ULL, "ht_0_hightown_21_0_ll.mmh"},
	{0x8BC27E3419048344ULL, "1072744533.ogg"},
	{0x8BC2B3F24B01E8A3ULL, "vxlt_0_lowtown_docks_a3_n_cth_0_0.vlm"},
	{0x8BC2D4284418CF56ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_reaction.gad"},
	{0x8BC32C7B4E5D8CA8ULL, "cd_ser221_quentin.bnk"},
	{0x8BC3CCC9F93469E8ULL, "vxht_0_hightown_a2_aic_0_0.vlm"},
	{0x8BC3E40E1A2BF914ULL, "470531680.ogg"},
	{0x8BC3E9DEB1A0CFC6ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.dura.ani"},
	{0x8BC4E354204B6F04ULL, "vxlt_0_lowtown_a3_n_c_f54_0_0.vlm"},
	{0x8BC55DDCB055DA63ULL, "gen00fl_isabela6096842.cl"},
	{0x8BC6D64E5062DFFEULL, "vxht_mansion2b_2kh_0_0.vlm"},
	{0x8BC6F8AE828FAAB5ULL, "6112113_m.fxe"},
	{0x8BC7316126E5B346ULL, "303007209.ogg"},
	{0x8BC731A3E4EE8804ULL, "mcr310_party_banter6087786.cl"},
	{0x8BC759E2FBB5BCB0ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.impa.ani"},
	{0x8BC7BA792D381A65ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.dura.gad"},
	{0x8BC822D254055480ULL, "vxw_encounter2_camp_o9_0_0.vlm"},
	{0x8BC8386BF0ED0612ULL, "41808269.ogg"},
	{0x8BC8FEEB84A3CB5BULL, "follower_banter6170563.cl"},
	{0x8BC95AAE7FF61EBFULL, "gen00fl_aveline6139630.cl"},
	{0x8BC9EF28441F1BB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\receive_and_inspect_reaction.evt"},
	{0x8BC9F6BB76578683ULL, "6101505_f.fxe"},
	{0x8BCA05A213F04FAFULL, "vxht_brothel_n_53z_0_0.vlm"},
	{0x8BCA442A5F4EA39AULL, "llac_ca_caverns2prc_5802_28.dds"},
	{0x8BCA5FF7C95E74D7ULL, "vxht_0_hightown_a2_cbw_0_0.vlm"},
	{0x8BCA88F02D8EC4FCULL, "fen310_fenris6118711.cl"},
	{0x8BCAF0080971DC8EULL, "gen_im_trs_dal_junk2_04.uti"},
	{0x8BCB41BDE513715FULL, "vis121_saemus6099486.cl"},
	{0x8BCB781A88A987ABULL, "hlac_lt_0_lowtown_-196608_4y.dds"},
	{0x8BCBB9F273291589ULL, "lgt320_shw_ors_marlein.bnk"},
	{0x8BCBD963DC496ABCULL, "6092552_f.fxe"},
	{0x8BCBE89A8C9ACBFAULL, "6135172_f.fxe"},
	{0x8BCC27CDDD9E1A9CULL, "6144564_m.fxe"},
	{0x8BCC299BA0FD7001ULL, "829781400.ogg"},
	{0x8BCCBCE57DE52DA2ULL, "6112139_f.fxe"},
	{0x8BCCCC855B5834E9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_descend_male.ani"},
	{0x8BCCD0DFC0D20D58ULL, "vxlt_shop_cy_0_0.vlm"},
	{0x8BCD2B9E696B18ECULL, "llac_lt_0_lowtown_docks_a3_n_131072_0.dds"},
	{0x8BCE65F3FCDC4E4AULL, "vxht_keep_5cs_0_0.vlm"},
	{0x8BCE6F24DD10D171ULL, "vxht_0_hightown_n_b5r_0_0.vlm"},
	{0x8BCE9C413B8CF56FULL, "vxw_woundedcoast_1hg_0_0.vlm"},
	{0x8BCEC4BFDA5845BAULL, "lt_0_lowtown_docks_n_28_0_ll.mmh"},
	{0x8BCEF4A55EB30577ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he3a.mmh"},
	{0x8BCF0AA55EB32AE9ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he3a.msh"},
	{0x8BCF9BF2732CCBC9ULL, "lgt320_shw_ors_marlein.cnv"},
	{0x8BCFE3D6AA947206ULL, "atl_partypicke_dxt5_dat.xml"},
	{0x8BD013C4AACA94C9ULL, "vxw_encounter2_lm_0_0.vlm"},
	{0x8BD0275894F28C0FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.impa.gad"},
	{0x8BD0AF06E00878C5ULL, "ts_darkspawn_acc_table.gda"},
	{0x8BD0B51F453CF028ULL, "lgt110_arl_bann_teagan6115588.cl"},
	{0x8BD10378706440F8ULL, "llac_lt_gamlenshouse_n_-1_8.dds"},
	{0x8BD128B788898029ULL, "63266897.ogg"},
	{0x8BD1B2D5A463E4BDULL, "lt_0_lowtown_n_26_0_ll.mmh"},
	{0x8BD1D1DEB1ACE437ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.dura.evt"},
	{0x8BD2566737C1167BULL, "hlac_dr_slavers_3216_11y.dds"},
	{0x8BD2D3F9CA5A44BBULL, "hlac_lt_gallowstemplar_f_2394_6x.dds"},
	{0x8BD2EFB37A840A7AULL, "vxlt_gallowsprison_n_6k4_0_0.vlm"},
	{0x8BD3494240C2BE13ULL, "lt_0_lowtown_docks_n_45_0.rml"},
	{0x8BD37D661885C53AULL, "vxw_encounter3_9f_0_0.vlm"},
	{0x8BD3E7855B5E816BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_descend_male.gad"},
	{0x8BD408D9EDFF0016ULL, "6032995_m.fxe"},
	{0x8BD4331B2E4DB9B0ULL, "vxw_encounter2_camp_nz_0_0.vlm"},
	{0x8BD43517D58406B6ULL, "llac_lt_hangedman_-1_7.dds"},
	{0x8BD4A37578931D59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf.ani"},
	{0x8BD4AB44D6D6F36FULL, "follower_banter6172275.cl"},
	{0x8BD50834CD293579ULL, "6166868_m.fxe"},
	{0x8BD534786A040FE1ULL, "vxca_caverns2_8lo_0_0.vlm"},
	{0x8BD554779FC7D1D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjutted.phy"},
	{0x8BD56E2B21B0445DULL, "6089118_m.fxe"},
	{0x8BD5B5BADC1C35E2ULL, "660303547.ogg"},
	{0x8BD60AD7AC13301BULL, "132952603.ogg"},
	{0x8BD63C5894F71BBCULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.impa.evt"},
	{0x8BD6821F2766156DULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyo.msh"},
	{0x8BD6901F27662CA0ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyo.mao"},
	{0x8BD6941F276633F3ULL, "art\\characters\\playercharacter\\elffemale\\glv_hvyo.mmh"},
	{0x8BD73A8149882C5CULL, "dae000_pc_home_ambients6208365.cl"},
	{0x8BD74F0B706323C3ULL, "203922399.ogg"},
	{0x8BD7F141C08F54EFULL, "gen00fl_anders6116102.cl"},
	{0x8BD8069E52D73E5AULL, "character_character_p184.psh.fxo.dx11.assembly"},
	{0x8BD818A369A25DACULL, "vxht_chantry_2ud_0_0.vlm"},
	{0x8BD85B37BA7E716EULL, "vxlt_hangedman_n_s5_0_0.vlm"},
	{0x8BD8AE7EE91A8D5BULL, "zz_dae_debug6030599.cl"},
	{0x8BD8CC8606373ECEULL, "hlac_lt_blackemporium_1055_22x.dds"},
	{0x8BD8D2DEB1B304C0ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_lar_aoe_p.dura.gad"},
	{0x8BD8EE5F14EEE4B9ULL, "character_character_p79.psh.fxo.dx11.assembly"},
	{0x8BD904CA6E3E7857ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelwht.mmh"},
	{0x8BD908CA6E3E7F64ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelwht.mao"},
	{0x8BD91ACA6E3E9DC9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelwht.msh"},
	{0x8BD984D85079C25FULL, "vxht_mansion2_334_0_0.vlm"},
	{0x8BD992557E451D97ULL, "vxht_keep_61w_0_0.vlm"},
	{0x8BD99B63C6824A5AULL, "6103743_m.fxe"},
	{0x8BDA524AD2EAFF7CULL, "w_encounter3_110_0.mao"},
	{0x8BDA5E4AD2EB131FULL, "w_encounter3_110_0.mmh"},
	{0x8BDA644AD2EB1DC1ULL, "w_encounter3_110_0.msh"},
	{0x8BDAC78855C191B5ULL, "vxht_0_hightown_a3_b8g_0_0.vlm"},
	{0x8BDAE4855B649AD0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_descend_male.evt"},
	{0x8BDB3F3157088953ULL, "hlac_lt_0_lowtown_a2_131069_0x.dds"},
	{0x8BDB42BCB2C825D8ULL, "vxlt_shop_fv_0_0.vlm"},
	{0x8BDB52E821850106ULL, "gen00fl_fenris6112059.cl"},
	{0x8BDB7E757898FC9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf.gad"},
	{0x8BDB90FA616B7DC1ULL, "vxca_caverns1_mines_ild_0_0.vlm"},
	{0x8BDBA3004272BB23ULL, "597258562.ogg"},
	{0x8BDD0A4EE8D7D72BULL, "6016668_m.fxe"},
	{0x8BDD0E1E2D830DEFULL, "vxw_woundedcoast_1sf_0_0.vlm"},
	{0x8BDD219DB28D0C28ULL, "w_bonepit_192_0.mmh"},
	{0x8BDD2D9DB28D20CBULL, "w_bonepit_192_0.mao"},
	{0x8BDD2E81BAF9CAA6ULL, "w_blightlands.anb"},
	{0x8BDD3281BAF9D1FCULL, "w_blightlands.arl"},
	{0x8BDD3F9DB28D3F52ULL, "w_bonepit_192_0.msh"},
	{0x8BDD492AE86528ECULL, "166094273.ogg"},
	{0x8BDDDCAD97902D05ULL, "6039965_f.fxe"},
	{0x8BDE58564309AFD5ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masf.msh"},
	{0x8BDE5A564309B34BULL, "art\\characters\\playercharacter\\humanfemale\\boo_masf.mmh"},
	{0x8BDEDA04E6E267C4ULL, "6122757_m.fxe"},
	{0x8BDF7F57D354AC1EULL, "vxht_chantry_n_2c3_0_0.vlm"},
	{0x8BDF97AC2EEE19A3ULL, "vxlt_0_lowtown_docks_a3_n_czj_0_0.vlm"},
	{0x8BE08DEF8E5ACA64ULL, "vxca_caverns1_mines_gix_0_0.vlm"},
	{0x8BE1039D9B0C68C5ULL, "vxlt_gallowsprison_n_6e1_0_0.vlm"},
	{0x8BE149848E03E3B8ULL, "dre000st_varric_framingb.stg"},
	{0x8BE15678624E4DD4ULL, "hlac_lt_0_lowtown_n_-3_6x.dds"},
	{0x8BE18446D0D89374ULL, "vxw_sundermount_311_0_0.vlm"},
	{0x8BE19385BCCD99E7ULL, "6116119_m.fxe"},
	{0x8BE1A62B0BBFA954ULL, "vxht_mansion_n_6nk_0_1.vlm"},
	{0x8BE1EDA8CFC093B8ULL, "dae000_merchant_generic_hm6172577.cl"},
	{0x8BE1EF430A7BDDBAULL, "vxw_encounter3_2g_0_0.vlm"},
	{0x8BE23087BBBFC57DULL, "172758094.ogg"},
	{0x8BE239CC79B415C7ULL, "360781673.ogg"},
	{0x8BE29B75789F4C80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf.evt"},
	{0x8BE2DC642F705B95ULL, "527360943.ogg"},
	{0x8BE32641E6DB73BCULL, "ser221_party_comments6163601.cl"},
	{0x8BE338626529FB21ULL, "hlac_w_encounter2_196612_0z.dds"},
	{0x8BE3BA08344F10FEULL, "vxht_mansion2b_2aj_0_0.vlm"},
	{0x8BE41452D5ECB397ULL, "hlac_lt_0_lowtown_a2_n_c_65533_2y.dds"},
	{0x8BE4241889498059ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vent_n.dds"},
	{0x8BE44458950366D5ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.impa.ani"},
	{0x8BE4984B2B04DD3BULL, "one110_meeran6089496.cl"},
	{0x8BE4DE156A97C287ULL, "6120059_m.fxe"},
	{0x8BE5BEF92A36BD2AULL, "vrc321tr_spooky_scenes_2.utt"},
	{0x8BE5F0F121C5CE50ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos09a.phy"},
	{0x8BE6401667DCE8C6ULL, "592415149.ogg"},
	{0x8BE68D47BA5530E6ULL, "vxht_mansion2_2xt_0_1.vlm"},
	{0x8BE6B202C2939C5CULL, "hlac_lt_0_lowtown_a2_1_8z.dds"},
	{0x8BE6D3A171B411C6ULL, "vxlt_0_lowtown_docks_a3_n_d0b_0_0.vlm"},
	{0x8BE7FA927C417006ULL, "vxht_bartrand_n_692_0_0.vlm"},
	{0x8BE81294DD86C626ULL, "fxa_lf_flower.mao"},
	{0x8BE91DD158C94E44ULL, "10983403.ogg"},
	{0x8BE91E7B4E7E4452ULL, "cd_ser221_quentin.wbd"},
	{0x8BE9396547B7AA35ULL, "vxht_0_hightown_a3_b7f_0_0.vlm"},
	{0x8BE9564FBB4E46A5ULL, "152215210.ogg"},
	{0x8BE979CD7E680931ULL, "13773220.ogg"},
	{0x8BE98B5363D56B1CULL, "gaussianblur.mao"},
	{0x8BE9B05B09C0A5EEULL, "follower_banter6143382.cl"},
	{0x8BEAA4A21C56698CULL, "llac_dr_0_deeproads_5849_2.dds"},
	{0x8BEB03A7FB2E619AULL, "vxw_bonepit_n8_0_0.vlm"},
	{0x8BEB097472C19626ULL, "26474058.ogg"},
	{0x8BEB19514E7715DEULL, "llac_ht_mansion_n_3373_2.dds"},
	{0x8BEB7A7E993BB378ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containera_d.dds"},
	{0x8BEB99D2461E8947ULL, "6032136_m.fxe"},
	{0x8BEBB73CB2C3FA2DULL, "6133788_f.fxe"},
	{0x8BEBE9AE715EE893ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf_v2.evt"},
	{0x8BEC1F0032DF3BAFULL, "797584546.ogg"},
	{0x8BEC327396A6AE32ULL, "vxlt_0_lowtown_a2_gxx_0_0.vlm"},
	{0x8BECE15529E585FCULL, "hlac_lt_warehouse_n_-65536_0x.dds"},
	{0x8BECF77E8EB6C5C9ULL, "vxw_encounter2_rs_0_0.vlm"},
	{0x8BED09E48456035FULL, "ht_0_hightownexplosion_n_14_0.rml"},
	{0x8BED386A3E764753ULL, "612375966.ogg"},
	{0x8BEE19FFAA59F307ULL, "zz_followers_banter6157971.cl"},
	{0x8BEE84AE0046BEC4ULL, "6167619_m.fxe"},
	{0x8BEECBD29291328AULL, "6092588_m.fxe"},
	{0x8BEF6A304000A2B2ULL, "mag320pt_otl_evelina.ncs"},
	{0x8BEF757A8D028145ULL, "vxlt_gallowsprison_n_6n0_0_0.vlm"},
	{0x8BEFD36D5E703B7AULL, "vxlt_gallowsprison_n_6q6_0_0.vlm"},
	{0x8BF0164E0D3F69B2ULL, "gen00fl_merrillact2_pt2.stg"},
	{0x8BF02B8A1A99D3E7ULL, "6082232_m.fxe"},
	{0x8BF0FA1AF1ABFE9CULL, "821455184.ogg"},
	{0x8BF10912B9DC76E3ULL, "6076162_f.fxe"},
	{0x8BF116D51239EAB0ULL, "vxw_encounter2_camp_dx_0_0.vlm"},
	{0x8BF126371E1070B6ULL, "6028804_m.fxe"},
	{0x8BF18062DE894E8EULL, "vxht_0_hightown_a2_ama_0_0.vlm"},
	{0x8BF1C8D351366757ULL, "6087734_m.fxe"},
	{0x8BF1E2AE716348D4ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf_v2.gad"},
	{0x8BF24FA51C0AD45FULL, "mag360_samson6047196.cl"},
	{0x8BF294F557061F75ULL, "vxw_bonepit_263_0_0.vlm"},
	{0x8BF2B07CB2D2213BULL, "art\\characters\\weapons\\greatswords\\w_gsw_013a.msh"},
	{0x8BF2C27CB2D23F86ULL, "art\\characters\\weapons\\greatswords\\w_gsw_013a.mao"},
	{0x8BF2C93A334E8BECULL, "75604949.ogg"},
	{0x8BF2CE7CB2D25425ULL, "art\\characters\\weapons\\greatswords\\w_gsw_013a.mmh"},
	{0x8BF374385C53A528ULL, "6118356_m.fxe"},
	{0x8BF39286504EFD02ULL, "w_blightlands_51_0_ll.mmh"},
	{0x8BF3A1DF6FED9364ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight2a.msh"},
	{0x8BF3B3DF6FEDB1CDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight2a.mao"},
	{0x8BF3B551CD6754BBULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb6.msh"},
	{0x8BF3B7DF6FEDB806ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight2a.mmh"},
	{0x8BF3D351CD6787A5ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb6.mmh"},
	{0x8BF3EC093A0BB84EULL, "514828851.ogg"},
	{0x8BF428E4845C56F7ULL, "ht_0_hightownexplosion_n_14_0.phy"},
	{0x8BF4898791376A8AULL, "233906393.ogg"},
	{0x8BF4ADAEEBD7D489ULL, "vxht_0_hightown_c1d_0_0.vlm"},
	{0x8BF53B4333C3D3EDULL, "vxw_encounter1_ci_0_0.vlm"},
	{0x8BF541833BA855A6ULL, "dae000_arishok6072401.cl"},
	{0x8BF543FA3A1D3AEDULL, "vxlt_hangedman_n_ru_0_0.vlm"},
	{0x8BF62C6EA57FB728ULL, "vxdr_primevaltomb_398_0_0.vlm"},
	{0x8BF6A2319B545B4AULL, "art\\vfx\\combat\\vfx_r_ibl_fli_app_imp_p.anb"},
	{0x8BF6A42A7CCB193DULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_eye_hdra_0d.dds"},
	{0x8BF6D6991B639837ULL, "6191269_f.fxe"},
	{0x8BF6DB1FBD32E181ULL, "qun231_party_comment.bnk"},
	{0x8BF6DED5FB94A854ULL, "lt_0_lowtown_docks_a3_154_0.phy"},
	{0x8BF6E19945DBAD5EULL, "mag311_nyssa6064832.cl"},
	{0x8BF6EDEC5340D387ULL, "zz_one000_merchant6179347.cl"},
	{0x8BF7B6F1425E1332ULL, "6113122_m.fxe"},
	{0x8BF7C20AF4BBDFF5ULL, "and100_anders6149708.cl"},
	{0x8BF80495E943D17FULL, "genip_door_lowxt_gate.utp"},
	{0x8BF881BEDA866297ULL, "mag101_grace6034035.cl"},
	{0x8BF886D4108C5279ULL, "lgt251ar_tome_cave.are"},
	{0x8BF8F1AE716980B2ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf_v2.ani"},
	{0x8BF8FF2635E1E440ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtc_0s.dds"},
	{0x8BF903F1AEE5241DULL, "static1_rl.psh.fxo"},
	{0x8BF96262F44B6BE9ULL, "hlac_lt_0_lowtown_a2_n_c_-131071_0y.dds"},
	{0x8BF968F3B9A69148ULL, "604684549.ogg"},
	{0x8BF96ADFBAD34716ULL, "541343143.ogg"},
	{0x8BF97584ED247A1AULL, "vxw_bonepit_i9_0_0.vlm"},
	{0x8BF9AC12BEA362C7ULL, "vxlt_0_lowtown_docks_n_br8_0_0.vlm"},
	{0x8BF9EB0DBBC9ED8EULL, "art\\characters\\playercharacter\\elffemale\\ef_glv_lgte.phy"},
	{0x8BF9FD1FBD3551F1ULL, "qun231_party_comment.cnv"},
	{0x8BFA167C3DBBD23AULL, "zz_followers_banter6159105.cl"},
	{0x8BFA471A081037C5ULL, "vxht_mansion_n_4z1_0_0.vlm"},
	{0x8BFA9F9D29250029ULL, "gen_im_trs_rdr_two_2ha_r0.uti"},
	{0x8BFB3A14E822D779ULL, "6122749_m.fxe"},
	{0x8BFB7AEFC24082ACULL, "374507695.ogg"},
	{0x8BFB979262E92048ULL, "vxht_0_hightown_bad_0_0.vlm"},
	{0x8BFBAF0CD57834D6ULL, "vxlt_0_lowtown_a2_e1i_0_0.vlm"},
	{0x8BFBE76B1F817B17ULL, "and121_karl6045422.cl"},
	{0x8BFBFF61BA2AC7C0ULL, "6177733_m.fxe"},
	{0x8BFC710C2305BDFEULL, "919447808.ogg"},
	{0x8BFCDCC403466C0FULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_004a_0t.dds"},
	{0x8BFD340462A1AB3AULL, "vxlt_0_lowtown_a2_eqe_0_0.vlm"},
	{0x8BFD99F92A4B1013ULL, "vrc321tr_spooky_scenes_2.ncs"},
	{0x8BFDABFF17F6245AULL, "qun210cr_random_thug_leader.utc"},
	{0x8BFDBBFC269591EBULL, "863979838.ogg"},
	{0x8BFE07BE65B2BE27ULL, "hlac_dr_dungeon_3215_0z.dds"},
	{0x8BFE1411EBE1439FULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar4_mcr370_playerrules.gad"},
	{0x8BFE18BE61175269ULL, "hlac_w_sundermount_-196607_0x.dds"},
	{0x8BFE57AF1DF31E72ULL, "vxw_bonepit_1s8_0_0.vlm"},
	{0x8BFE9D670C8FEC67ULL, "6085332_m.fxe"},
	{0x8BFF1A51D45BC9F8ULL, "dae000_bodhan6146790.cl"},
	{0x8BFF20A2AF550FFDULL, "6097360_m.fxe"},
	{0x8BFF362DD56EB93EULL, "rdr111dg_martin6036902.cl"},
	{0x8C006030A0BEB5FFULL, "t1_mul_pi2.tnt"},
	{0x8C0105A4765FEE6DULL, "vxca_caverns1_mines_e93_0_0.vlm"},
	{0x8C0106D248FC37F5ULL, "vxw_bonepit_212_0_0.vlm"},
	{0x8C01087631FC893FULL, "follower_banter6167880.cl"},
	{0x8C0144EBEE01629CULL, "vxht_0_hightown_n_bu3_0_0.vlm"},
	{0x8C0190CF4A78530CULL, "art\\characters\\playercharacter\\elfmale\\rob_appd.phy"},
	{0x8C01A6F630B03E54ULL, "dae100_alistair_drunk.cnv"},
	{0x8C01B54396B0DBBFULL, "w_encounter1_24_0.msh"},
	{0x8C01BB4396B0E661ULL, "w_encounter1_24_0.mmh"},
	{0x8C01BBD5FB9E7451ULL, "lt_0_lowtown_docks_a3_154_0.msh"},
	{0x8C01C74396B0FA02ULL, "w_encounter1_24_0.mao"},
	{0x8C01C938AD611B19ULL, "ser111ip_to_lowtown.utp"},
	{0x8C01C9D5FB9E8C6CULL, "lt_0_lowtown_docks_a3_154_0.mao"},
	{0x8C01D5D5FB9EA00FULL, "lt_0_lowtown_docks_a3_154_0.mmh"},
	{0x8C0251DFE3B97749ULL, "vxlt_undercitydungeon_1y2_0_0.vlm"},
	{0x8C0265FF3F06BF0EULL, "422432906.ogg"},
	{0x8C02C203138E7DA1ULL, "llac_ht_mansion2_818_52.dds"},
	{0x8C02F669F91E2E1CULL, "dre000_cut2d_ragstoriches6025112.cl"},
	{0x8C0329C71259E244ULL, "385310468.ogg"},
	{0x8C03AD2025B9EC6DULL, "vxw_encounter1_hh_0_0.vlm"},
	{0x8C03BE671AF6EFD9ULL, "vxlt_0_lowtown_docks_n_ary_0_0.vlm"},
	{0x8C03BFF717E886ABULL, "vxlt_blackemporium_zf_0_0.vlm"},
	{0x8C03D535F46D9F80ULL, "vxlt_0_lowtown_docks_cob_0_0.vlm"},
	{0x8C042EE26F76397EULL, "6065779_m.fxe"},
	{0x8C0488F630B24094ULL, "dae100_alistair_drunk.bnk"},
	{0x8C048AF6ED2BEA4FULL, "vxlt_0_lowtown_a3_n_c_gv7_0_0.vlm"},
	{0x8C052B5687619723ULL, "vxw_bonepitmining_a2_k_0_0.vlm"},
	{0x8C0558EE41C5BF71ULL, "vxdr_primevalentrance_32p_0_0.vlm"},
	{0x8C05A88B7370A992ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_0y.dds"},
	{0x8C05BB2615B0580BULL, "vxlt_0_lowtown_n_gpn_0_0.vlm"},
	{0x8C065F6A120F8B8FULL, "6071133_m.fxe"},
	{0x8C06C45165E6C339ULL, "follower_banter6163235.cl"},
	{0x8C072394DDA193C8ULL, "fxa_lf_flower.dds"},
	{0x8C080C04C644478FULL, "153631277.ogg"},
	{0x8C082424AE27B2EBULL, "song_smo.mao"},
	{0x8C083163A4C55431ULL, "6139198_m.fxe"},
	{0x8C085277774B0685ULL, "152314276.ogg"},
	{0x8C08D2E21E8994CBULL, "vxht_keep_69k_0_0.vlm"},
	{0x8C08DDF903E6140CULL, "vxlt_gallowstemplar_4ua_0_0.vlm"},
	{0x8C09276978D40B1AULL, "vxca_cavern1_haunted_icn_0_0.vlm"},
	{0x8C098657324089BBULL, "llac_lt_undercitydungeon_-1_11.dds"},
	{0x8C09EA7431E37596ULL, "hla_ht_0_hightown_65538_0y.dds"},
	{0x8C0A0659DFAF9A36ULL, "hm_ave1_raider_assassin.mor"},
	{0x8C0ABD4B5EBA9217ULL, "llac_ht_keep_3733_0.dds"},
	{0x8C0B5ECFC1497013ULL, "632494094.ogg"},
	{0x8C0B773E11077FA5ULL, "vxht_0_hightown_a2_bum_0_0.vlm"},
	{0x8C0B79A4F15B3945ULL, "lt_0_lowtown_docks_a3_n_28_0.rml"},
	{0x8C0B8E2D8D9352DDULL, "hlac_w_sundermount_-65536_0z.dds"},
	{0x8C0BBFE25334B0D6ULL, "vxw_encounter3_en_0_0.vlm"},
	{0x8C0BC2A7EDDBAF8BULL, "gen00fl_varric6093584.cl"},
	{0x8C0C36B45EB36B37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_burst_shot.gad"},
	{0x8C0C4CFD5B950E47ULL, "6028881_m.fxe"},
	{0x8C0C5972A51898C9ULL, "gen00fl_dog6167718.cl"},
	{0x8C0C6305A6C45F2CULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright90.ani"},
	{0x8C0CE5A5AD1F8B20ULL, "dae000_mother6103071.cl"},
	{0x8C0D5900CF37A04BULL, "lt_gallowstemplar_f_8165_ll.mmh"},
	{0x8C0D84E7208010DAULL, "298703511.ogg"},
	{0x8C0D875869E19BE3ULL, "mag291_demon_marethari6028943.cl"},
	{0x8C0DC8E970A560C7ULL, "6107139_m.fxe"},
	{0x8C0DC97D9FF16944ULL, "hlac_lt_0_lowtown_-65537_5y.dds"},
	{0x8C0E342AC7E434EDULL, "6067998_m.fxe"},
	{0x8C0E4A731F73E820ULL, "vxht_mansion_n_62z_0_0.vlm"},
	{0x8C0E892BA855136EULL, "gen00fl_varric6033668.cl"},
	{0x8C0EA154165EE50CULL, "vxht_0_hightown_cam_0_0.vlm"},
	{0x8C0EF2968580A561ULL, "vxht_0_hightown_bwg_0_0.vlm"},
	{0x8C0EFBEC31DC71E1ULL, "vxca_caverns2_8xc_0_0.vlm"},
	{0x8C0F6EC3D961001CULL, "729633026.ogg"},
	{0x8C0FA1E535AE1E5FULL, "hlac_lt_0_lowtown_a2_n_-65538_0y.dds"},
	{0x8C0FC4FA47038083ULL, "ht_0_hightownexplosion_n_3_0.mao"},
	{0x8C0FC8FA470387D0ULL, "ht_0_hightownexplosion_n_3_0.mmh"},
	{0x8C0FD6FA47039FEAULL, "ht_0_hightownexplosion_n_3_0.msh"},
	{0x8C0FEA49D72569D1ULL, "6090453_m.fxe"},
	{0x8C1000437CA38060ULL, "242218058.ogg"},
	{0x8C100CF829734B6EULL, "364763203.ogg"},
	{0x8C10BFD62B75581EULL, "vxca_cave_18b_0_0.vlm"},
	{0x8C10FC7E4922B348ULL, "6066277_m.fxe"},
	{0x8C1175D990335BE3ULL, "vxlt_shop_2u_0_0.vlm"},
	{0x8C11B0B93B84B402ULL, "w_encounter3.rml"},
	{0x8C11F1116BEABFD7ULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobd.phy"},
	{0x8C120F223C4201E9ULL, "vxlt_gallowstemplar_n_3rq_0_0.vlm"},
	{0x8C124BB45EB7FAC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_burst_shot.evt"},
	{0x8C12F111EBF364E5ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar4_mcr370_playerrules.ani"},
	{0x8C1323015E484CCFULL, "79411241.ogg"},
	{0x8C1361776EB338D8ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_reva_0n.dds"},
	{0x8C137005A6CA946EULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright90.gad"},
	{0x8C1394DD8846B784ULL, "sst_ss_crazy_m_1.gda"},
	{0x8C13B9F3DD293356ULL, "vxlt_gallowstemplar_n_47b_0_0.vlm"},
	{0x8C1414CDAA119E1DULL, "lowtown_male_beggar_sta6151059.cl"},
	{0x8C14219F96A891B4ULL, "vxlt_0_lowtown_a3_n_c_hp1_0_0.vlm"},
	{0x8C143A6D98AB50CDULL, "21259242.ogg"},
	{0x8C147721268030D7ULL, "342723329.ogg"},
	{0x8C14E2B53DA8241BULL, "isa321_velasco6091888.cl"},
	{0x8C14E65D9717AAAFULL, "gen00fl_isabela6090318.cl"},
	{0x8C14EA4F2D3BC332ULL, "gen_im_arm_cht_med_act3_03.uti"},
	{0x8C157B0F52C0F77BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_upper_d.dds"},
	{0x8C158264DEC9C2EAULL, "vxht_playershouse_n_203_0_0.vlm"},
	{0x8C15D9A4BC94BC3AULL, "hlac_lt_blackemporium_1055_14y.dds"},
	{0x8C16DD5DE065A0EBULL, "222340048.ogg"},
	{0x8C16EEA3230CC6C2ULL, "402800707.ogg"},
	{0x8C17895CD5E3E770ULL, "vxlt_gamlenshouse_n_5s_0_0.vlm"},
	{0x8C179D4AFC93B427ULL, "6095768_m.fxe"},
	{0x8C17A9577A6852A3ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.cess.ani"},
	{0x8C17F92C692BD56BULL, "follower_banter6171098.cl"},
	{0x8C18478861E8E3DBULL, "dae000_flemeth6013421.cl"},
	{0x8C187B4C46D55148ULL, "vxht_0_hightown_bwb_0_0.vlm"},
	{0x8C1884A1F3311DC8ULL, "vxca_caverns2_8xf_0_0.vlm"},
	{0x8C18ADF8A6AC274EULL, "llac_lt_darktown_-1_66.dds"},
	{0x8C18C8568D8DD88AULL, "qcr000_arishok6087513.cl"},
	{0x8C191302206AAC55ULL, "66459513.ogg"},
	{0x8C191FAB10CA719BULL, "vxw_bonepit_258_0_0.vlm"},
	{0x8C1A28D13DAE45F9ULL, "qun100_mage_party_comment6019097.cl"},
	{0x8C1A4C64819CB43BULL, "6072402_m.fxe"},
	{0x8C1A6B05A6D0AA4DULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright90.evt"},
	{0x8C1AB1334A638AD9ULL, "vxlt_0_lowtown_a2_n_ekg_0_0.vlm"},
	{0x8C1AF008A628C5FFULL, "6103751_m.fxe"},
	{0x8C1BB746B5E68E50ULL, "dre000_hireling26038689.cl"},
	{0x8C1C273DA273CE34ULL, "vxca_cavern1_haunted_e5z_0_0.vlm"},
	{0x8C1CB04277B7D941ULL, "hlac_lt_0_lowtown_docks_n_196606_0z.dds"},
	{0x8C1D03D564D6CB85ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_esm_ice_0d.dds"},
	{0x8C1D64737776BDE1ULL, "vxht_0_hightown_bpf_0_0.vlm"},
	{0x8C1D978822B34A97ULL, "6110526_m.fxe"},
	{0x8C1D97D410AB8B4FULL, "lgt251ar_tome_cave.ncs"},
	{0x8C1DC86570B5AB3EULL, "271757391.ogg"},
	{0x8C1E89C8419C5282ULL, "6051039_m.fxe"},
	{0x8C1F29C3DD199BE0ULL, "ave200_aveline_bar6116409.cl"},
	{0x8C1F5A2BA7B075A9ULL, "qcr000_arishok6060177.cl"},
	{0x8C1F92E95123BF1AULL, "vxdr_primevaltomb_5n1_0_0.vlm"},
	{0x8C1FA5319B77653DULL, "art\\vfx\\combat\\vfx_r_ibl_fli_app_imp_p.mmh"},
	{0x8C1FCDA676C18565ULL, "6133461_f.fxe"},
	{0x8C1FE7B682297463ULL, "vxlt_shop_5t_0_0.vlm"},
	{0x8C2063B45EC460BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_burst_shot.ani"},
	{0x8C20F97FC6D2F88AULL, "514943594.ogg"},
	{0x8C219D1383B88350ULL, "projdecal_projdecal_p1.psh.fxo.dx11"},
	{0x8C21A2107D42CBF0ULL, "ave221_brennan6059772.cl"},
	{0x8C21CE0143547A96ULL, "vxht_mansion2b_1tb_0_0.vlm"},
	{0x8C22CBD29FA4A565ULL, "vxlt_warehouse_1c6_0_0.vlm"},
	{0x8C2497D410B1A99EULL, "lgt251ar_tome_cave.lst"},
	{0x8C24AE2481E5B4EAULL, "lt_0_lowtown_docks_a3_116_0.mao"},
	{0x8C24B22481E5BB39ULL, "lt_0_lowtown_docks_a3_116_0.mmh"},
	{0x8C24BC2481E5CC27ULL, "lt_0_lowtown_docks_a3_116_0.msh"},
	{0x8C24C461F8B37CB4ULL, "hf_ibethany_h3s3.mor"},
	{0x8C24F409ACE09A49ULL, "vrc221_bartrand_fake.bnk"},
	{0x8C25238EA1ADBB38ULL, "llac_ht_brothel_n_3382_1.dds"},
	{0x8C25974385F2B3A5ULL, "gen00fl_aveline6120807.cl"},
	{0x8C25C1577A74B89AULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.cess.evt"},
	{0x8C269B4396CFCB8AULL, "w_encounter1_24_0.rml"},
	{0x8C26A56AC753DA66ULL, "512990866.ogg"},
	{0x8C26C68C6693753BULL, "vxw_bonepitdestroyed_a3_25x_0_0.vlm"},
	{0x8C26CE283A3A1C88ULL, "6164355_m.fxe"},
	{0x8C2724F5C9A5633CULL, "vxlt_0_lowtown_a2_n_hey_0_0.vlm"},
	{0x8C274DFA70BE7FBBULL, "w_sundermount_53_0.mmh"},
	{0x8C2759FA70BE9458ULL, "w_sundermount_53_0.mao"},
	{0x8C276BFA70BEB2A5ULL, "w_sundermount_53_0.msh"},
	{0x8C2772F2A709759EULL, "hlac_lt_0_lowtown_n_131071_7x.dds"},
	{0x8C27BFBAABCA25FDULL, "hlac_lt_gallowstemplar_2391_7y.dds"},
	{0x8C27D609ACE29C89ULL, "vrc221_bartrand_fake.cnv"},
	{0x8C27E6B27129ADA7ULL, "26185940.ogg"},
	{0x8C28531D8F9D224BULL, "464593419.ogg"},
	{0x8C286817EBF225CCULL, "framebuffereffects_horizontalblurshadows_p0.vsh.fxo.dx11"},
	{0x8C286B773F659688ULL, "vxca_caverns1_mines_ckr_0_0.vlm"},
	{0x8C28AEEA416010BDULL, "280974599.ogg"},
	{0x8C29687F843ED163ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackr.mao"},
	{0x8C296C7F843ED7B0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackr.mmh"},
	{0x8C297A7F843EEFCAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackr.msh"},
	{0x8C29EF92A6792F3CULL, "6137579_m.fxe"},
	{0x8C2A44679F7B5E3AULL, "character_character_p186.psh.fxo.dx11"},
	{0x8C2AAC76C1ABA3BEULL, "lt_0_lowtown_docks_n_10_0_lc.mmh"},
	{0x8C2B38BD50B43825ULL, "6036314_m.fxe"},
	{0x8C2B46199E770736ULL, "art\\vfx\\materials\\weapontrail\\vfx_wt_talent.mao"},
	{0x8C2B8BE20497930EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_galoflipalco.msh"},
	{0x8C2B99E20497AB5FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_galoflipalco.mao"},
	{0x8C2BA5E20497BFFCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_galoflipalco.mmh"},
	{0x8C2BEA1AFAA3163CULL, "345616511.ogg"},
	{0x8C2C6AAB552FA2C3ULL, "vxlt_0_lowtown_docks_n_b60_0_0.vlm"},
	{0x8C2CC0577A7AD58DULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.cess.gad"},
	{0x8C2CD1EF2BE38E03ULL, "mag160_feynriel6020464.cl"},
	{0x8C2D5C54822512FFULL, "zz_approval_debug6003949.cl"},
	{0x8C2DBC4396D622DEULL, "w_encounter1_24_0.phy"},
	{0x8C2DDB68693FAB00ULL, "vxht_0_hightown_n_asr_0_0.vlm"},
	{0x8C2DE6B0A3106EA1ULL, "6156539_m.fxe"},
	{0x8C2DEE5C9B5EE021ULL, "vxlt_undercitydungeon_19j_0_0.vlm"},
	{0x8C2E04C64319D79AULL, "vxdr_primevaltomb_5c2_0_0.vlm"},
	{0x8C2E6270C49A97C9ULL, "hlac_lt_0_lowtown_a2_n_65533_3z.dds"},
	{0x8C2E6B79DB401C71ULL, "vxca_caverns2_3td_0_0.vlm"},
	{0x8C2EC2855B78B901ULL, "vxca_caverns1_jg7_0_0.vlm"},
	{0x8C2F58B5C7A18A58ULL, "6034057_f.fxe"},
	{0x8C2F76DA5199D0A2ULL, "vxlt_hangedman_st_0_0.vlm"},
	{0x8C2F97098EB4A350ULL, "cnvprev_human_male_archer1.utc"},
	{0x8C2FBFF63AAD59D6ULL, "vxht_0_hightown_n_bye_0_0.vlm"},
	{0x8C2FD12E50471B20ULL, "532770259.ogg"},
	{0x8C30E7531C128A18ULL, "301268734.ogg"},
	{0x8C314177A016106DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjutted.msh"},
	{0x8C3144EDC7D6D03AULL, "vxht_0_hightown_n_b9i_0_0.vlm"},
	{0x8C314F77A01627A0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjutted.mao"},
	{0x8C315377A0162EF3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjutted.mmh"},
	{0x8C322F3240B6BC8EULL, "6114848_m.fxe"},
	{0x8C326BC373CCD849ULL, "vxht_brothel_n_6cc_0_0.vlm"},
	{0x8C340A03C0FD119EULL, "art\\vfx\\combat\\vfx_glb_levelup_c.mmh"},
	{0x8C3459BB471E69C9ULL, "6042524_m.fxe"},
	{0x8C3468D176A23B3CULL, "zz_followers_banter6157552.cl"},
	{0x8C346BC4507528B2ULL, "vxlt_0_lowtown_a2_gce_0_0.vlm"},
	{0x8C34E09D2AE2CD64ULL, "318835282.ogg"},
	{0x8C34FE6273AE6E31ULL, "931666918.ogg"},
	{0x8C353B4A9F7F80F3ULL, "523891588.ogg"},
	{0x8C35A9FCD5ABCCF4ULL, "vxca_caverns1_dpa_0_0.vlm"},
	{0x8C361E055AAB7505ULL, "mag170_thrask6020774.cl"},
	{0x8C365E509C632D1CULL, "46807962.ogg"},
	{0x8C36F558C200721FULL, "535524249.ogg"},
	{0x8C376882AF48D9FFULL, "w_bonepit_203_0.msh"},
	{0x8C376E82AF48E3A1ULL, "w_bonepit_203_0.mmh"},
	{0x8C377A82AF48F842ULL, "w_bonepit_203_0.mao"},
	{0x8C377DD0805570C9ULL, "zz_mag_debug6102187.cl"},
	{0x8C3794ED2E4FBBD9ULL, "vxlt_0_lowtown_a2_n_eqe_0_0.vlm"},
	{0x8C3799DAF1D51E31ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbr.ani"},
	{0x8C37B3FA82074787ULL, "330097382.ogg"},
	{0x8C37C8046290443EULL, "6207953_m.fxe"},
	{0x8C37DBAF32223890ULL, "6176796_m.fxe"},
	{0x8C3881D6B95DB23DULL, "vxlt_0_lowtown_a3_n_c_eha_0_0.vlm"},
	{0x8C399D43CEE31DB0ULL, "llac_lt_0_lowtown_a2_n_c_65535_4.dds"},
	{0x8C39A77CB30E14D2ULL, "art\\characters\\weapons\\greatswords\\w_gsw_013a.phy"},
	{0x8C39B6B3912D31B3ULL, "vxlt_0_lowtown_a2_n_c_eqd_0_1.vlm"},
	{0x8C39B8B9BE730EEDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheadb_n.dds"},
	{0x8C39C47D4EF8E0F8ULL, "hlac_lt_undercitydungeon_-1_30z.dds"},
	{0x8C39F5E085130DD3ULL, "mag311_keran6041402.cl"},
	{0x8C3A0C3A7FA52FC7ULL, "682691560.ogg"},
	{0x8C3ADB30EB3BD033ULL, "hlac_lt_undercitydungeon_-1_41y.dds"},
	{0x8C3AEA0DBC018351ULL, "art\\characters\\playercharacter\\elffemale\\ef_glv_lgte.mmh"},
	{0x8C3AEC107C3DAAF6ULL, "353703632.ogg"},
	{0x8C3AF93F263E569FULL, "em_orsino_harvester.mor"},
	{0x8C3B040DBC01AF0FULL, "art\\characters\\playercharacter\\elffemale\\ef_glv_lgte.msh"},
	{0x8C3BFDE48499C34FULL, "ht_0_hightownexplosion_n_14_0.mao"},
	{0x8C3C09E48499D7ECULL, "ht_0_hightownexplosion_n_14_0.mmh"},
	{0x8C3C0FE48499E1DEULL, "ht_0_hightownexplosion_n_14_0.msh"},
	{0x8C3CBB74523CE6D1ULL, "6137026_m.fxe"},
	{0x8C3D4E69BBE00C5BULL, "4249446.ogg"},
	{0x8C3D55546C204379ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic.msh"},
	{0x8C3D55D32EC130DCULL, "vxw_encounter2_camp_jz_0_0.vlm"},
	{0x8C3D5F546C205467ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic.mmh"},
	{0x8C3D63546C205AB4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic.mao"},
	{0x8C3DCA92B997A2D8ULL, "6031357_f.fxe"},
	{0x8C3E204ABFAB66B5ULL, "vxca_caverns1_mines_drc_0_0.vlm"},
	{0x8C3E78A4F186916DULL, "lt_0_lowtown_docks_a3_n_28_0.anb"},
	{0x8C3E9ABA61F0F57CULL, "vxlt_gallowstemplar_n_434_0_0.vlm"},
	{0x8C3EA4DAF1DB4FD3ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbr.gad"},
	{0x8C3F6D6BB418EA28ULL, "mag130tr_predanzig.utt"},
	{0x8C3FCD3AF355639AULL, "6118138_m.fxe"},
	{0x8C4054CD6522EAAAULL, "zz_mrl_debug6109299.cl"},
	{0x8C405A97AE6E5B30ULL, "vxlt_0_lowtown_a3_n_c_gz0_0_0.vlm"},
	{0x8C406C5720FCC136ULL, "766836268.ogg"},
	{0x8C417EDBE0FC8C6CULL, "vxlt_gallowstemplar_6qc_0_0.vlm"},
	{0x8C41E903C109175DULL, "art\\vfx\\combat\\vfx_glb_levelup_c.anb"},
	{0x8C424AF7F82DFE38ULL, "mcr310_loyalist_mages6076865.cl"},
	{0x8C4267CF4AAFA4D7ULL, "art\\characters\\playercharacter\\elfmale\\rob_appd.mmh"},
	{0x8C426861BF83CC5AULL, "vxlt_0_lowtown_a2_n_c_hko_0_0.vlm"},
	{0x8C426BCF4AAFABE4ULL, "art\\characters\\playercharacter\\elfmale\\rob_appd.mao"},
	{0x8C427DCF4AAFCA49ULL, "art\\characters\\playercharacter\\elfmale\\rob_appd.msh"},
	{0x8C4291BE2571A7B9ULL, "vxht_mansion_n_6ox_0_0.vlm"},
	{0x8C42BE99F1A6DE47ULL, "32447700.ogg"},
	{0x8C42DDA1426B4132ULL, "vxlt_0_lowtown_a2_gff_0_0.vlm"},
	{0x8C436E6DAF88F5E7ULL, "hlac_w_sundermount_262140_0y.dds"},
	{0x8C43FF7934F71307ULL, "331284312.ogg"},
	{0x8C4401A77B6EAC97ULL, "ht_mansion2b_mmap_ph.dds"},
	{0x8C447D086408F850ULL, "art\\characters\\weapons\\spears\\w_qui_qspr02a.msh"},
	{0x8C448B086409106AULL, "art\\characters\\weapons\\spears\\w_qui_qspr02a.mmh"},
	{0x8C44A1DAF1DFB638ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwbr.evt"},
	{0x8C44BA55870E7E90ULL, "rdr110cr_dockworker.utc"},
	{0x8C45391F3243C8BEULL, "vxht_0_hightown_a3_cfc_0_0.vlm"},
	{0x8C455BF38D4FF846ULL, "644286435.ogg"},
	{0x8C457852E07FA56BULL, "mag110_samson6018713.cl"},
	{0x8C45897F45A7D5DBULL, "126931290.ogg"},
	{0x8C45B10B581C40ADULL, "6151369_m.fxe"},
	{0x8C45FE8361A6F2F7ULL, "vxht_mansion_n_5a7_0_0.vlm"},
	{0x8C463A31284C6EEEULL, "362492305.ogg"},
	{0x8C4662AF5A72ED2DULL, "6118471_f.fxe"},
	{0x8C4673761FB35187ULL, "zz_qun_debug6028901.cl"},
	{0x8C46776353DC13AAULL, "485507607.ogg"},
	{0x8C46979AB0C070A1ULL, "6087509_m.fxe"},
	{0x8C46E3692186FCEEULL, "6115498_m.fxe"},
	{0x8C4774043BC84EFFULL, "dae000st_varric_cassandra.stg"},
	{0x8C48F8A3B5E712E6ULL, "vxht_mansion_n_3fa_0_0.vlm"},
	{0x8C4961B46932A938ULL, "6216711_m.fxe"},
	{0x8C497E3CA7CFDD1CULL, "vxht_0_hightown_n_bl4_0_0.vlm"},
	{0x8C4A265626BA4B8DULL, "showstats.ncs"},
	{0x8C4A46265C986187ULL, "fen210_orana6133046.cl"},
	{0x8C4A800287DEE119ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masb_0d.dds"},
	{0x8C4AAC3D6BE3906DULL, "vxlt_hangedman_248_0_0.vlm"},
	{0x8C4AB035F5E8D44DULL, "hf_elthina.mor"},
	{0x8C4B2D3B8F755B16ULL, "fen110_hunter6066862.cl"},
	{0x8C4B4AFD489F8D3AULL, "vxdr_primevalentrance_30i_0_0.vlm"},
	{0x8C4B6B649C382A5FULL, "hlac_lt_0_lowtown_a2_n_-3_3z.dds"},
	{0x8C4BA1B93BB61354ULL, "w_encounter3.arl"},
	{0x8C4BA2AE7F60F008ULL, "6080573_f.fxe"},
	{0x8C4BAD2EDB387F4CULL, "mag311st_mage_conspirator.stg"},
	{0x8C4BBB68740DDA29ULL, "hlac_lt_0_lowtown_docks_65534_5y.dds"},
	{0x8C4C146F118D1274ULL, "6077976_m.fxe"},
	{0x8C4C33FD4928EFA0ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_4x.dds"},
	{0x8C4C46DD2C70E441ULL, "qun111_ketojan6040850.cl"},
	{0x8C4C47B2B630F9C8ULL, "dae000_orsino6085431.cl"},
	{0x8C4CA54060B150A1ULL, "mrl351cr_phantomelf.ncs"},
	{0x8C4CAB0DAAB7F497ULL, "vxca_caverns1_mines_j41_0_0.vlm"},
	{0x8C4CC9EC6C429E41ULL, "mag110_vincento6031812.cl"},
	{0x8C4CEF665CA3E735ULL, "75487486.ogg"},
	{0x8C4D3D67AF89FC3AULL, "6092287_m.fxe"},
	{0x8C4D9D66F8C5B5B0ULL, "hlac_w_bonepitdestroyed_a3_3_0y.dds"},
	{0x8C4DCED4544F9616ULL, "6119487_m.fxe"},
	{0x8C4DE9135E80F3B4ULL, "vxlt_0_lowtown_a3_n_c_hl5_0_0.vlm"},
	{0x8C4DFAEE1DE5D8BFULL, "6129290_m.fxe"},
	{0x8C4E3F17E54B7C1AULL, "6117038_m.fxe"},
	{0x8C4E4E771BDBBD18ULL, "vxht_0_hightown_a3_5y2_0_0.vlm"},
	{0x8C4EACC24E7335D4ULL, "vxw_sundermount_21t_0_0.vlm"},
	{0x8C4EFAFEE8D87FDFULL, "vxw_encounter3_44_0_0.vlm"},
	{0x8C4F1F1338CA2968ULL, "1042274428.ogg"},
	{0x8C4F80B5563611FEULL, "hlac_lt_0_lowtown_a2_n_-131073_0z.dds"},
	{0x8C4F99AED068DBAFULL, "953524230.ogg"},
	{0x8C4FCFFA473977EBULL, "ht_0_hightownexplosion_n_3_0.phy"},
	{0x8C5006427DA0AABCULL, "vxht_0_hightown_a3_bft_0_0.vlm"},
	{0x8C5034773EFD62CFULL, "291087154.ogg"},
	{0x8C503480DB9B9D8EULL, "819791465.ogg"},
	{0x8C505BDB8CBACA60ULL, "mag101_alain6032134.cl"},
	{0x8C506282B67C1056ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_nud_0d.dds"},
	{0x8C50960A317163A2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_basic_1.mao"},
	{0x8C50DFE60E26AF95ULL, "vxca_caverns2prc_ag6_0_0.vlm"},
	{0x8C50F66A5E5997C5ULL, "dae000_fem_commoner1_mob6174492.cl"},
	{0x8C514E05E4451981ULL, "vxw_encounter2_camp_o4_0_0.vlm"},
	{0x8C5184E204B856C7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo13_galoflipalco.phy"},
	{0x8C51B6FAC5F8BD2FULL, "dae000_keep_noble_sta6161544.cl"},
	{0x8C51FF15BA03F54BULL, "6000800_f.fxe"},
	{0x8C5201448301C9E7ULL, "259336876.ogg"},
	{0x8C520FFDA89208E5ULL, "l_dae240ar_bone_pit.wbd"},
	{0x8C53527B4E56846EULL, "637502955.ogg"},
	{0x8C535DC890FE18D5ULL, "6164945_m.fxe"},
	{0x8C5367C27A51A62FULL, "569154454.ogg"},
	{0x8C536EB9B233A60AULL, "gen00fl_aveline6137657.cl"},
	{0x8C539FD7111CAD74ULL, "6132774_m.fxe"},
	{0x8C53FFF1DD4B8535ULL, "dae000ip_crowd_ff6_talk17.utp"},
	{0x8C54443ADF02215FULL, "art\\vfx\\textures\\vfx_link_smo_wispy.dds"},
	{0x8C5447B264E46172ULL, "art\\vfx\\creatures\\fxc_succubus_c.impa.gad"},
	{0x8C5469C586365E74ULL, "994882710.ogg"},
	{0x8C547060539D0B77ULL, "vxht_mansion_n_5j6_0_0.vlm"},
	{0x8C54B2DF3D40F144ULL, "wm_ico_camp_roll.dds"},
	{0x8C54B8FED38701C3ULL, "6000700_m.fxe"},
	{0x8C54DEF34A9F01D4ULL, "6136011_m.fxe"},
	{0x8C559350AA0CAE9CULL, "character_character_p232.psh.fxo.dx11"},
	{0x8C55DBE83517AC87ULL, "6082171_m.fxe"},
	{0x8C569007A67E0FECULL, "94762963.ogg"},
	{0x8C57A08F4CA96D91ULL, "6032053_m.fxe"},
	{0x8C580811CC947652ULL, "715132243.ogg"},
	{0x8C582EB5A7052F0AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicerchair_d.dds"},
	{0x8C58437E10EBBC98ULL, "vxht_mansion2_1qw_0_0.vlm"},
	{0x8C589D0FF1E3AF77ULL, "vxw_bonepitdestroyed_a3_25l_0_0.vlm"},
	{0x8C58AF223036A058ULL, "246146618.ogg"},
	{0x8C590584D4E79936ULL, "6147206_m.fxe"},
	{0x8C591E1A5DD9A8EDULL, "vxlt_hangedman_217_0_0.vlm"},
	{0x8C59C6116C282C2FULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobd.mao"},
	{0x8C59CD74F3035BF2ULL, "gen_im_npc_arm_hum_boo_rdrb.uti"},
	{0x8C59D2116C2840CCULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobd.mmh"},
	{0x8C59D8116C284ABEULL, "art\\characters\\playercharacter\\humanfemale\\cth_nobd.msh"},
	{0x8C5A451C3A30EC4EULL, "qun120_arishok_guard6049213.cl"},
	{0x8C5A8CB0FD940A78ULL, "ts_dwarven_01.gda"},
	{0x8C5ABF435F3985CFULL, "zz_isa_debug6175459.cl"},
	{0x8C5AE04065C15568ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallotopleft.msh"},
	{0x8C5AF24065C173F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallotopleft.mao"},
	{0x8C5AFE4065C18792ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallotopleft.mmh"},
	{0x8C5B4AB264EA84A9ULL, "art\\vfx\\creatures\\fxc_succubus_c.impa.evt"},
	{0x8C5C2AABE276554AULL, "fxe_day_beam_orz.mao"},
	{0x8C5C34D61D71F822ULL, "6107062_m.fxe"},
	{0x8C5C65F3A3B5CC46ULL, "6014673_f.fxe"},
	{0x8C5CBC2F545A8500ULL, "llac_lt_0_lowtown_a2_n_-196608_1.dds"},
	{0x8C5CE3219406A413ULL, "art\\characters\\creatures\\dragon\\animation\\rst_p.ani"},
	{0x8C5CE67FC30ECDBDULL, "513609258.ogg"},
	{0x8C5D57B3716BC60BULL, "gen00fl_sebastian6179988.cl"},
	{0x8C5D6CDBDACE985FULL, "vxw_encounter3_17_0_0.vlm"},
	{0x8C5DAB2FED96491BULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_p.mmh"},
	{0x8C5DAB95918BB727ULL, "dae000_unhappy_mage6165343.cl"},
	{0x8C5DCACB1A3446ACULL, "6110965_m.fxe"},
	{0x8C5DDFE242970CE7ULL, "500871491.ogg"},
	{0x8C5EBEEE910234C6ULL, "hla_ht_0_hightown_a2_1_0y.dds"},
	{0x8C5EEBD16E75877BULL, "166403689.ogg"},
	{0x8C5F3101EBC50B65ULL, "841229959.ogg"},
	{0x8C60649952498756ULL, "583635943.ogg"},
	{0x8C6071CA8BD57E8DULL, "gen00fl_fenris6107214.cl"},
	{0x8C60BE338D9253EAULL, "6136008_m.fxe"},
	{0x8C60F609F99E17FEULL, "1016179391.ogg"},
	{0x8C611BA9C130529CULL, "mag370_meredith6046927.cl"},
	{0x8C619BF9A258B70AULL, "gen_im_npc_elf_arm_hlm_dalc.uti"},
	{0x8C620CDBB444DF62ULL, "follower_barks6153726.cl"},
	{0x8C6223FBF52A3B1EULL, "dae000_orana6149103.cl"},
	{0x8C62FE8FB856B24CULL, "cod_bks_lite_tombs_last_0.utp"},
	{0x8C630497DBA46B5FULL, "hlac_lt_gallowstemplar_2394_11y.dds"},
	{0x8C6328FED1B9AD47ULL, "6053069_m.fxe"},
	{0x8C6335D0F623316DULL, "vxca_caverns1_mines_fmt_0_0.vlm"},
	{0x8C63E5C9D088A3D6ULL, "lt_0_lowtown_docks_a3_26_0.anb"},
	{0x8C6460A53ACC4FD3ULL, "6082922_m.fxe"},
	{0x8C646F82AF6FD31EULL, "w_bonepit_203_0.phy"},
	{0x8C6477857802AD2AULL, "vxlt_0_lowtown_docks_a3_n_d0n_0_0.vlm"},
	{0x8C649F007D057FC7ULL, "44439500.ogg"},
	{0x8C64BFAB18C7E6AFULL, "seal_hilight.mao"},
	{0x8C64C1A8012CFEB7ULL, "art\\levels\\races\\proxy\\props\\_textures\\p_armoire_01_s.dds"},
	{0x8C65171AE0854EFBULL, "330997033.ogg"},
	{0x8C65418AD0C995A9ULL, "zz_followers_banter6159588.cl"},
	{0x8C6599B22B8A4BA5ULL, "art\\characters\\weapons\\dagger\\w_dgr_108.mao"},
	{0x8C65BACCB5F5ED7EULL, "vxht_mansion2b_22c_0_0.vlm"},
	{0x8C65D393DE5601AAULL, "dae000_mother6099187.cl"},
	{0x8C66036A36DA8DA1ULL, "vxca_cavern1_haunted_e84_0_0.vlm"},
	{0x8C66132D7A9984A0ULL, "w_sundermount_44_0.phy"},
	{0x8C66B55B02E1D518ULL, "vxht_mansion2_1tt_0_0.vlm"},
	{0x8C66C8AC23D779A7ULL, "vxlt_gallowscourtyard_6li_0_0.vlm"},
	{0x8C66F5D93FE70C45ULL, "504971120.ogg"},
	{0x8C671FAA14C5629BULL, "hlac_dr_0_deeproads_5845_16z.dds"},
	{0x8C677761C6DDB1B6ULL, "6140306_m.fxe"},
	{0x8C683BD73A402164ULL, "vxw_bonepitmining_a2_b8_0_0.vlm"},
	{0x8C687911F07C0ED8ULL, "hla_ht_0_hightown_-131073_0x.dds"},
	{0x8C68E6E53EC13EECULL, "hlac_w_encounter2_camp_-1_0z.dds"},
	{0x8C6932B264F69918ULL, "art\\vfx\\creatures\\fxc_succubus_c.impa.ani"},
	{0x8C693589CE197E6DULL, "pro000_opening_pc6001211.cl"},
	{0x8C6948546C45679CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0dd_basic.phy"},
	{0x8C6A20C17DF8F52CULL, "649201666.ogg"},
	{0x8C6A2650F91489BEULL, "6133133_m.fxe"},
	{0x8C6AB697E9D65EA3ULL, "vxht_mansion2b_3eb_0_0.vlm"},
	{0x8C6ACC2FEDA10B80ULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_f_p.anb"},
	{0x8C6ADB219412D44AULL, "art\\characters\\creatures\\dragon\\animation\\rst_p.evt"},
	{0x8C6ADEC6BDB84223ULL, "llac_ht_mansion2_818_9.dds"},
	{0x8C6AE2558BCB4DA3ULL, "6154045_m.fxe"},
	{0x8C6AEF7DD4CA8D7AULL, "qun160_vashothlead6061101.cl"},
	{0x8C6B029B3F57E674ULL, "vxw_bonepit_2ag_0_0.vlm"},
	{0x8C6B26D0A0612A6BULL, "6167472_m.fxe"},
	{0x8C6B3E37530FB50AULL, "one300_zevran6076948.cl"},
	{0x8C6B438E3997BA4FULL, "llac_w_encounter2_327683_0.dds"},
	{0x8C6B4BEEAB0CCC8EULL, "vxdr_primevaltomb_749_0_0.vlm"},
	{0x8C6C47E024C3B83BULL, "art\\characters\\weapons\\mauls\\w_mal_dsmla.mmh"},
	{0x8C6C65E024C3EB25ULL, "art\\characters\\weapons\\mauls\\w_mal_dsmla.msh"},
	{0x8C6CDB2AABB0D59AULL, "vxw_bonepitdestroyed_a3_1rc_0_0.vlm"},
	{0x8C6D322D7A9FD858ULL, "w_sundermount_44_0.rml"},
	{0x8C6D968C2CE193A5ULL, "vxlt_0_lowtown_a2_hdh_0_0.vlm"},
	{0x8C6DFAD234FE04C1ULL, "vxlt_0_lowtown_docks_a3_n_cz0_0_0.vlm"},
	{0x8C6E4236159BF306ULL, "690533154.ogg"},
	{0x8C6E6EFA70FABAF0ULL, "w_sundermount_53_0.phy"},
	{0x8C6EE3934F7AAB8EULL, "6072551_m.fxe"},
	{0x8C6F17CF19191E27ULL, "dae000_arishok6075348.cl"},
	{0x8C6F30107F886C6AULL, "6131108_m.fxe"},
	{0x8C6FACA7BF6DEC8CULL, "zz_tbe_debug6189392.cl"},
	{0x8C6FBA09212AE7BFULL, "gen00fl_sebastian6179782.cl"},
	{0x8C6FE438B939DB7DULL, "hlac_lt_gallowstemplar_f_7198_6y.dds"},
	{0x8C7039BD2FC596DAULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchwall_1.mao"},
	{0x8C70A985B9420619ULL, "vxca_caverns2prc_9ri_0_0.vlm"},
	{0x8C717A83B975FE7DULL, "gen00fl_aveline6053309.cl"},
	{0x8C71BA219418BABDULL, "art\\characters\\creatures\\dragon\\animation\\rst_p.gad"},
	{0x8C71D9E9ACC19B29ULL, "vxht_chantry_n_289_0_0.vlm"},
	{0x8C71EE15401FA8BAULL, "vxlt_0_lowtown_er4_0_0.vlm"},
	{0x8C723B99705B3A6AULL, "pro000_merchant6051615.cl"},
	{0x8C7279562619A5F7ULL, "6150229_m.fxe"},
	{0x8C728019D5FF30BEULL, "hlac_lt_0_lowtown_docks_n_-131074_5y.dds"},
	{0x8C72BC4C6855929CULL, "art\\characters\\weapons\\shields\\w_shd_007a.mao"},
	{0x8C72C84C6855A73FULL, "art\\characters\\weapons\\shields\\w_shd_007a.mmh"},
	{0x8C72CE4C6855B1E1ULL, "art\\characters\\weapons\\shields\\w_shd_007a.msh"},
	{0x8C73140C12612AA0ULL, "ss_justice.bnk"},
	{0x8C731FBBA607B164ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.cess.ani"},
	{0x8C7328FBD6CDC081ULL, "6126485_f.fxe"},
	{0x8C734220EDC36214ULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.mmh"},
	{0x8C73885BE24C8788ULL, "dae300cr_gamlen.utc"},
	{0x8C745A0C59DCF1D7ULL, "vxw_bonepitmining_a2_1l2_0_0.vlm"},
	{0x8C74EAE27A379BE1ULL, "6169704_m.fxe"},
	{0x8C7504551E22935AULL, "6164757_m.fxe"},
	{0x8C754D84BE47017CULL, "6079570_m.fxe"},
	{0x8C754DFA7100A148ULL, "w_sundermount_53_0.rml"},
	{0x8C758F2B3FC14071ULL, "llac_lt_0_lowtown_a3_n_c_65537_2.dds"},
	{0x8C75D960919A9A71ULL, "906972574.ogg"},
	{0x8C75F6C858D43B7EULL, "art\\vfx\\environment\\vfx_fog_ground_p.dura.evt"},
	{0x8C75FB863045C653ULL, "vxlt_undercitydungeon_1ip_0_0.vlm"},
	{0x8C76183B32F69294ULL, "540235425.ogg"},
	{0x8C765DBA09DE2EA6ULL, "mag291_demon_marethari6029019.cl"},
	{0x8C76ADB42C3639E4ULL, "vxw_bonepitmining_a2_i9_0_0.vlm"},
	{0x8C76DECB5AB9EA86ULL, "mcr371_emile_de_launcet6086944.cl"},
	{0x8C7780FAC18BC9DFULL, "6067369_m.fxe"},
	{0x8C779A93293B5B4BULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.mmh"},
	{0x8C780A91A81F10B4ULL, "vxca_caverns2prc_a6l_0_0.vlm"},
	{0x8C780FE97189DAC2ULL, "plt_lgt110pt_pinpricks.plo"},
	{0x8C784D37BAE9129AULL, "333980479.ogg"},
	{0x8C78896B66B4EC0DULL, "mcr370cr_moira.utc"},
	{0x8C78D8BC1F8D91A0ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_barge_anim.mmh"},
	{0x8C791C3E19C68232ULL, "6081698_m.fxe"},
	{0x8C795874FF1E3E11ULL, "6126554_m.fxe"},
	{0x8C79C4727287AAF1ULL, "vrc221_bartrand6043063.cl"},
	{0x8C7A1CBBA60DCAC6ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.cess.gad"},
	{0x8C7A4C77D4610D12ULL, "llac_w_blightlands_262144_0.dds"},
	{0x8C7A8D417BAE3F24ULL, "hla_ht_0_hightown_a2_2_0z.dds"},
	{0x8C7B18069A6AA2BCULL, "825030916.ogg"},
	{0x8C7B21637A8F0FFDULL, "6172582_m.fxe"},
	{0x8C7B8C496F2FA475ULL, "dae000ip_crowd_mf8_talk14.utp"},
	{0x8C7B8D35307A1BDEULL, "gen00fl_aveline6139853.cl"},
	{0x8C7BED999619B2FFULL, "852771616.ogg"},
	{0x8C7BF51A8C735E14ULL, "6080729_m.fxe"},
	{0x8C7C588F73A21BF2ULL, "mag170_recruit016080834.cl"},
	{0x8C7C716B31A1DEEBULL, "hlac_w_sundermount_131072_0x.dds"},
	{0x8C7D0DC858DA80A9ULL, "art\\vfx\\environment\\vfx_fog_ground_p.dura.gad"},
	{0x8C7D9CB426039BD0ULL, "vxlt_0_lowtown_a2_n_c_gze_0_0.vlm"},
	{0x8C7DB1DF9116256FULL, "gen00fl_sebastian6179892.cl"},
	{0x8C7DB630CC84165EULL, "mag110_ws_banter6079654.cl"},
	{0x8C7DDC3F5B1C4080ULL, "qcr220_orsino6061349.cl"},
	{0x8C7DE859A430629BULL, "gen00fl_merrill6121160.cl"},
	{0x8C7E30332018988CULL, "vxlt_0_lowtown_n_es2_0_0.vlm"},
	{0x8C7EE129A8A669C0ULL, "vxlt_hangedman_n_25i_0_0.vlm"},
	{0x8C7F7C3BF19A64D2ULL, "vxlt_undercitydungeon_1iu_0_0.vlm"},
	{0x8C7F8843CAE2DCF2ULL, "vxca_cavern1_haunted_jfb_0_0.vlm"},
	{0x8C7F970084FC2486ULL, "vxht_mansion2b_20u_0_0.vlm"},
	{0x8C7FF2B9DBF9CF9AULL, "6107723_carver_sister.cut"},
	{0x8C802099BB562978ULL, "6196714_m.fxe"},
	{0x8C809F6CC85C193CULL, "qun131_party_comment6117294.cl"},
	{0x8C80D732F3E0019BULL, "323632253.ogg"},
	{0x8C80F0E2105A36F2ULL, "zz_and_debug6134057.cl"},
	{0x8C81162E4D2D783CULL, "hlac_lt_gallowstemplar_n_2383_1z.dds"},
	{0x8C8117BBA613E0A5ULL, "art\\vfx\\combat\\vfx_r_asn_pin_strike_loop_c.cess.evt"},
	{0x8C817F7D630870DAULL, "vxlt_hangedman_n_26z_0_0.vlm"},
	{0x8C8181ECF7C538F1ULL, "6102546_f.fxe"},
	{0x8C81E56ABA12D2AAULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv3.gad"},
	{0x8C81E86B81D8F087ULL, "vxca_cavern1_haunted_j41_0_0.vlm"},
	{0x8C82406454859466ULL, "mer240_jansen6059691.cl"},
	{0x8C82C5BAB3960946ULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.cess.evt"},
	{0x8C831A32B1C43C76ULL, "vxht_0_hightown_a2_b7p_0_0.vlm"},
	{0x8C839914E6CE0618ULL, "vxht_mansion2_1br_0_0.vlm"},
	{0x8C83DEC858E04FF7ULL, "art\\vfx\\environment\\vfx_fog_ground_p.dura.ani"},
	{0x8C84771725398257ULL, "hlac_ca_caverns2_5802_53x.dds"},
	{0x8C8547314CB19246ULL, "680419345.ogg"},
	{0x8C857B93294763B0ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.anb"},
	{0x8C85E94680DF28A9ULL, "gen00fl_sebastian6195885.cl"},
	{0x8C8601443C9D8F4EULL, "dae011_badpoet_sta6146998.cl"},
	{0x8C867A85920EA919ULL, "vxlt_0_lowtown_docks_bz5_0_0.vlm"},
	{0x8C869866F20543F0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a.mao"},
	{0x8C869C66F2054A03ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a.mmh"},
	{0x8C86AA66F205625DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a.msh"},
	{0x8C86B32482388726ULL, "lt_0_lowtown_docks_a3_116_0.phy"},
	{0x8C86BBBC1F999DBFULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_barge_anim.anb"},
	{0x8C870A0ADD00BABEULL, "6139733_m.fxe"},
	{0x8C87421DC173C224ULL, "345411574.ogg"},
	{0x8C876BD7DC995DCDULL, "6090234_f.fxe"},
	{0x8C878BA51604C487ULL, "642841466.ogg"},
	{0x8C879CAF3088EBB0ULL, "624379287.ogg"},
	{0x8C88031673385A79ULL, "6095920_m.fxe"},
	{0x8C884702CAA7533AULL, "vxca_cave_1a9_0_0.vlm"},
	{0x8C88C86ABA18C041ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv3.evt"},
	{0x8C88DC47B2DDDD98ULL, "vxlt_0_lowtown_docks_a3_n_bu5_0_0.vlm"},
	{0x8C891013C19E05B6ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls105a_0t.dds"},
	{0x8C89536300F19AA1ULL, "vxlt_0_lowtown_a3_n_c_hrv_0_0.vlm"},
	{0x8C89634B4183862DULL, "6037322_m.fxe"},
	{0x8C89DCBAB39C4E91ULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.cess.gad"},
	{0x8C8A6479850C33EFULL, "vxw_woundedcoast_17j_0_0.vlm"},
	{0x8C8ACC4CBC9EF158ULL, "vxca_cavern1_haunted_ila_0_0.vlm"},
	{0x8C8B302D7AB8D235ULL, "w_sundermount_44_0.msh"},
	{0x8C8B322D7AB8D5ABULL, "w_sundermount_44_0.mmh"},
	{0x8C8B3E2D7AB8EA48ULL, "w_sundermount_44_0.mao"},
	{0x8C8B4FF6DF3F56AEULL, "hlac_lt_0_lowtown_a2_n_c_0_7y.dds"},
	{0x8C8B5606EB59360FULL, "llac_lt_undercitydungeon_-1_42.dds"},
	{0x8C8B923DB25715C9ULL, "one300_zevran6075737.cl"},
	{0x8C8BC02E5ABA6309ULL, "mag131_keran6086283.cl"},
	{0x8C8CF03DDDD7D38AULL, "6112217_m.fxe"},
	{0x8C8D26F42195B85FULL, "153248220.ogg"},
	{0x8C8D641439690CF0ULL, "hlac_w_sundermount_-196613_0x.dds"},
	{0x8C8DF39EEE867B4DULL, "284223644.ogg"},
	{0x8C8F5670587937CEULL, "528390944.ogg"},
	{0x8C8F69446FED062BULL, "6086045_m.fxe"},
	{0x8C8F99C19E248F2BULL, "611271009.ogg"},
	{0x8C8FC3BF7D76F1FDULL, "6027363_m.fxe"},
	{0x8C900242D2FEC15AULL, "llac_w_encounter2_-131071_0.dds"},
	{0x8C90C5FA1815765DULL, "character_character_p114.psh.fxo.dx11"},
	{0x8C90D010BF4C0265ULL, "vxca_caverns2prc_abe_0_0.vlm"},
	{0x8C90DDBAB3A26F3FULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.cess.ani"},
	{0x8C9160088AF9CFF2ULL, "130992576.ogg"},
	{0x8C917BC231EA02E9ULL, "vxlt_0_lowtown_f59_0_0.vlm"},
	{0x8C918C0FA3BA54F6ULL, "vxht_0_hightown_a2_b8q_0_0.vlm"},
	{0x8C9196683458D99AULL, "mrl210_merrill6116329.cl"},
	{0x8C919778CB432182ULL, "one100_solivitus6048129.cl"},
	{0x8C9199F26897C5A5ULL, "vxht_keep_n_57w_0_0.vlm"},
	{0x8C91E9FDA8C7AB8FULL, "l_dae240ar_bone_pit.bnk"},
	{0x8C9232A56F2B6C35ULL, "dae000_quintus6154731.cl"},
	{0x8C92AB0198960713ULL, "hlac_lt_0_lowtown_docks_n_-196610_0z.dds"},
	{0x8C92BC0417658D6EULL, "llac_lt_undercityserialkiller_-1_2.dds"},
	{0x8C930A85466E82D8ULL, "vxlt_0_lowtown_a3_n_c_e22_0_0.vlm"},
	{0x8C93319EEC8E1482ULL, "cs_dre180cs_varric_intro.bnk"},
	{0x8C93AF8AE462BB87ULL, "81128885.ogg"},
	{0x8C93C0212D4EA3EDULL, "hlac_ht_mansion_n_3373_1x.dds"},
	{0x8C93C941E27A2E57ULL, "6167282_f.fxe"},
	{0x8C93F8F00934D0D0ULL, "vrc321_spooky_scenes6219388.cl"},
	{0x8C947F6A5F452995ULL, "6173905_m.fxe"},
	{0x8C94971EA3D2AC67ULL, "536844337.ogg"},
	{0x8C9592812F4AE83DULL, "vxlt_0_lowtown_docks_b54_0_0.vlm"},
	{0x8C95AD03BB3976A2ULL, "lt_0_lowtown_docks_a3_236_0.msh"},
	{0x8C95AF03BB397A58ULL, "lt_0_lowtown_docks_a3_236_0.mmh"},
	{0x8C95B54562EBE0EEULL, "art\\vfx\\combat\\vfx_m_bld_conj_c.impa.evt"},
	{0x8C95BB03BB398EFBULL, "lt_0_lowtown_docks_a3_236_0.mao"},
	{0x8C95C06ABA233C90ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv3.ani"},
	{0x8C962C01D4C1BD19ULL, "lowtown_male_urchin_sta6164180.cl"},
	{0x8C963C2EE30A6F11ULL, "6128754_m.fxe"},
	{0x8C965408B8AAB4A4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_pull.ani"},
	{0x8C96AF83001FFCCEULL, "mag100_cullen6148678.cl"},
	{0x8C96C8C9D0B3CC6AULL, "lt_0_lowtown_docks_a3_26_0.rml"},
	{0x8C96D194DB876A00ULL, "dae000_brothel_orana6161221.cl"},
	{0x8C97869BF8DDA6A5ULL, "vxdr_slavers_11w_0_0.vlm"},
	{0x8C979D66F213BF28ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile4a.phy"},
	{0x8C980982A67A800AULL, "6147765_m.fxe"},
	{0x8C9863064F6F99D3ULL, "6048832_m.fxe"},
	{0x8C98940C3DCEB0E3ULL, "vxlt_gallowscourtyard_7aw_0_0.vlm"},
	{0x8C98B3A4FEC4EA20ULL, "dre000_varric6035151.cl"},
	{0x8C98D65677024C6FULL, "vxw_woundedcoast_12u_0_0.vlm"},
	{0x8C99440AF5AA6EA1ULL, "6166367_m.fxe"},
	{0x8C9A806E09EFCCD0ULL, "vxlt_0_lowtown_a2_n_c_gpc_0_0.vlm"},
	{0x8C9A8F806C455B78ULL, "161456913.ogg"},
	{0x8C9AA0C3B78BAB9EULL, "vxlt_0_lowtown_a2_n_c_hd6_0_0.vlm"},
	{0x8C9AE3B3F35487ADULL, "vxht_0_hightown_a3_at0_0_0.vlm"},
	{0x8C9B13ED0404C98CULL, "vxlt_0_lowtown_n_ei0_0_0.vlm"},
	{0x8C9B5C6BB4672C55ULL, "mag130tr_predanzig.ncs"},
	{0x8C9BAC4562F03C99ULL, "art\\vfx\\combat\\vfx_m_bld_conj_c.impa.gad"},
	{0x8C9BC87FC9C65C94ULL, "6099648_m.fxe"},
	{0x8C9C13B92F075189ULL, "281034009.ogg"},
	{0x8C9C43FCB84CB9CAULL, "vxlt_0_lowtown_docks_b6n_0_0.vlm"},
	{0x8C9C4520EDE66CCBULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.anb"},
	{0x8C9C5108B8AF1B06ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_pull.gad"},
	{0x8C9C52E56EC9EF0EULL, "follower_banter6160819.cl"},
	{0x8C9C5FF5D58695D2ULL, "vxlt_undercitydungeon_1ww_0_0.vlm"},
	{0x8C9DE2C64267322DULL, "vxlt_gallowstemplar_n_2mu_0_0.vlm"},
	{0x8C9E02F5C409C39BULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernrock_drelief.dds"},
	{0x8C9E0353E57D725AULL, "586826207.ogg"},
	{0x8C9E2BEC42D598D6ULL, "hlac_w_woundedcoast_-131072_0z.dds"},
	{0x8C9E31E75E249170ULL, "mag330st_walter_2.stg"},
	{0x8C9E61D74C855920ULL, "100068919.ogg"},
	{0x8C9EC536CF374F78ULL, "vxlt_gallowstemplar_f_5ph_0_0.vlm"},
	{0x8C9F4F2D2C6EBA2BULL, "qcr220_grey_wardens6060493.cl"},
	{0x8C9FF17552E83D39ULL, "441603322.ogg"},
	{0x8CA08E85CDB8FF06ULL, "qun120cr_javaris6048960.cl"},
	{0x8CA0901FEBAFDBA7ULL, "vxlt_gallowscourtyard_6xe_0_0.vlm"},
	{0x8CA097B069AFDD28ULL, "6087250_m.fxe"},
	{0x8CA0EE6644A3E03FULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.cess.gad"},
	{0x8CA120581DB9E731ULL, "dae000_dalish_apprentices6147527.cl"},
	{0x8CA1511D06280FD3ULL, "vxlt_undercitydungeon_1zu_0_0.vlm"},
	{0x8CA158338E5A101AULL, "980415430.ogg"},
	{0x8CA1AE5E24C420D4ULL, "dae000cr_grand_cleric6173596.cl"},
	{0x8CA1C9746FDA547CULL, "882147192.ogg"},
	{0x8CA1E861510D8938ULL, "6035215_m.fxe"},
	{0x8CA21FE386EB3F52ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_001a_0d.dds"},
	{0x8CA2BD4562F678E7ULL, "art\\vfx\\combat\\vfx_m_bld_conj_c.impa.ani"},
	{0x8CA2CA85DA3CD917ULL, "vxca_caverns1_ign_0_0.vlm"},
	{0x8CA3074F1355A435ULL, "dae000_porfiria6162654.cl"},
	{0x8CA34C08B8B531E5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tal_pull.evt"},
	{0x8CA36CDEA0127651ULL, "fxe_day_dust_shine.mao"},
	{0x8CA370EB81D49115ULL, "6043627_f.fxe"},
	{0x8CA3C9A88E5E4DF2ULL, "hlac_w_encounter3_-327679_0x.dds"},
	{0x8CA43EBB5A11AB2DULL, "qun120_arishok_guard.bnk"},
	{0x8CA47B965823EABAULL, "6076564_m.fxe"},
	{0x8CA4BD8A656805F1ULL, "633837384.ogg"},
	{0x8CA5448D130E1111ULL, "zz_mag_debug6039847.cl"},
	{0x8CA58CC933DD6CAFULL, "gen00fl_anders6128327.cl"},
	{0x8CA616A08D9F9119ULL, "vxlt_gallowscourtyard_7f0_0_0.vlm"},
	{0x8CA6378EEAD4C412ULL, "121238337.ogg"},
	{0x8CA63C08CA3569B6ULL, "vxca_caverns2prc_adt_0_0.vlm"},
	{0x8CA65368037D8731ULL, "art\\vfx\\combat\\vfx_w_off_penetra_activ_c.anb"},
	{0x8CA6D314192DB314ULL, "788891939.ogg"},
	{0x8CA6D4E34165B253ULL, "6043932_m.fxe"},
	{0x8CA760BB5A141A95ULL, "qun120_arishok_guard.cnv"},
	{0x8CA7E36644A9EC2CULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.cess.evt"},
	{0x8CA836AD09232BB9ULL, "dae000ip_crowd_mf2_drunk1.utp"},
	{0x8CA87D34B273C42BULL, "6037139_m.fxe"},
	{0x8CA8A5C0725704D9ULL, "159496492.ogg"},
	{0x8CA8EF0792C865C1ULL, "6143515_m.fxe"},
	{0x8CA8F2221579370AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglehead_sa.dds"},
	{0x8CA978FB3CF36B63ULL, "art\\vfx\\combat\\vfx_w_tmp_holy_smt_aoe_p.impa.evt"},
	{0x8CA9B2310CD4F38DULL, "293572496.ogg"},
	{0x8CA9E42368999581ULL, "vxlt_0_lowtown_docks_cum_0_0.vlm"},
	{0x8CAA1DC9CC10CA53ULL, "hlac_w_woundedcoast_131076_0z.dds"},
	{0x8CAA348D4A8F7EF7ULL, "128812748.ogg"},
	{0x8CAA966D734B541FULL, "ser221_quentin6128914.cl"},
	{0x8CAA9E41C6FDBCAFULL, "974566628.ogg"},
	{0x8CAAACF4B9273A49ULL, "m_dae121ar_keep.bnk"},
	{0x8CAAB72E1E82FF3FULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_reva_0e.dds"},
	{0x8CAABD4F0B99D87CULL, "mag130_danzig6018796.cl"},
	{0x8CAB265F16623CC4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangb.mmh"},
	{0x8CAB2A5F16624317ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangb.mao"},
	{0x8CAB3C5F16626266ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhangb.msh"},
	{0x8CAB8B459EB33D1CULL, "6118129_m.fxe"},
	{0x8CABCF49FFBC48C0ULL, "gen00fl_anders6104009.cl"},
	{0x8CAC178E85992675ULL, "mag160_varian_ilithis6016483.cl"},
	{0x8CAC8380A9BDF28EULL, "mag170_thrask6020560.cl"},
	{0x8CADCF87F400E19DULL, "hlac_w_bonepitdestroyed_a3_327681_0z.dds"},
	{0x8CADD4E04335D096ULL, "gen00fl_anders6119412.cl"},
	{0x8CAE5D7F2EE3BD53ULL, "vxlt_darktown_35q_0_0.vlm"},
	{0x8CAF2D5E5552B9B0ULL, "dre000_carver6037599.cl"},
	{0x8CAF507A4B88C29AULL, "vxca_caverns2prc_96z_0_0.vlm"},
	{0x8CAFEE3F2A5AB3D8ULL, "vxlt_0_lowtown_a3_n_c_e80_0_0.vlm"},
	{0x8CB02806DE5CF3FEULL, "vxht_0_hightown_a3_a83_0_0.vlm"},
	{0x8CB051FB3CF947E4ULL, "art\\vfx\\combat\\vfx_w_tmp_holy_smt_aoe_p.impa.gad"},
	{0x8CB053DF8BFFF2EEULL, "hlac_dr_0_deeproads_5849_6x.dds"},
	{0x8CB0E4B56F18515AULL, "6176501_m.fxe"},
	{0x8CB0F6335EEE4E35ULL, "51957891.ogg"},
	{0x8CB1017B7EA1E09CULL, "558112060.ogg"},
	{0x8CB1388AE3D57F55ULL, "dre000_bodahn6033409.cl"},
	{0x8CB13C62CC32F197ULL, "vxca_caverns1_iji_0_0.vlm"},
	{0x8CB1C0642B9DEA58ULL, "6081375_m.fxe"},
	{0x8CB24DFEACAD0F9CULL, "rdr111dg_assistant6055113.cl"},
	{0x8CB2705F57C8E8BFULL, "vxlt_0_lowtown_docks_n_aot_0_0.vlm"},
	{0x8CB2F268E2B81C9AULL, "vxdr_primevaltomb_8oc_0_0.vlm"},
	{0x8CB368E024FFF370ULL, "art\\characters\\weapons\\mauls\\w_mal_dsmla.phy"},
	{0x8CB37468038849AAULL, "art\\vfx\\combat\\vfx_w_off_penetra_activ_c.mmh"},
	{0x8CB37D6D65B29E76ULL, "dae000_quintus6153137.cl"},
	{0x8CB38E51B00E27ACULL, "879374327.ogg"},
	{0x8CB39205D0A3D8C3ULL, "vxlt_hangedman_n_vw_0_0.vlm"},
	{0x8CB3BA012FE2500EULL, "hlac_lt_0_lowtown_a2_n_65536_1x.dds"},
	{0x8CB4DFC0C1DD1269ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgta_0n.dds"},
	{0x8CB5172993E205ABULL, "dae000_hate_follower_react6103243.cl"},
	{0x8CB57A4F19F2FBF9ULL, "vxca_caverns2prc_3te_0_0.vlm"},
	{0x8CB5CB6644B60085ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.cess.ani"},
	{0x8CB660BBA9FB8D35ULL, "166471615.ogg"},
	{0x8CB67E9F4CB85FFAULL, "497542274.ogg"},
	{0x8CB6B975F26A36D4ULL, "shatter_02.dds"},
	{0x8CB70BA3F11F2222ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_enter.gad"},
	{0x8CB760FB3CFF7FC2ULL, "art\\vfx\\combat\\vfx_w_tmp_holy_smt_aoe_p.impa.ani"},
	{0x8CB784CF00DC27BAULL, "663377392.ogg"},
	{0x8CB784EFF4705149ULL, "vxw_blightlands_fp_0_0.vlm"},
	{0x8CB7866764F48BE1ULL, "129636699.ogg"},
	{0x8CB7A5A149CF8384ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_1z.dds"},
	{0x8CB7ADEB1646C005ULL, "811828661.ogg"},
	{0x8CB81B9EECAD0BA0ULL, "cs_dre180cs_varric_intro.wbd"},
	{0x8CB8455589489410ULL, "vxdr_primevaltomb_75x_0_0.vlm"},
	{0x8CB8813D7C11E636ULL, "420418621.ogg"},
	{0x8CB8C3012BB9EA07ULL, "vxca_cavern1_haunted_gcu_0_0.vlm"},
	{0x8CB8EFF88A2DA986ULL, "ht_0_hightown_a2_17_0_ll.mmh"},
	{0x8CB9A5A38EB2D11AULL, "gen_im_acc_rng_act3_09.uti"},
	{0x8CB9E0003467B0C9ULL, "238581262.ogg"},
	{0x8CBA29F7657DC3EDULL, "6020088_f.fxe"},
	{0x8CBB5C31A28FCD68ULL, "vxw_sundermount_2c2_0_0.vlm"},
	{0x8CBB879A1BEAC215ULL, "vxlt_warehouse_19h_0_0.vlm"},
	{0x8CBB8FAF4A927547ULL, "6202278_m.fxe"},
	{0x8CBBA15C7D17F59BULL, "hlac_w_bonepit_3_0z.dds"},
	{0x8CBC0BB12949AA7EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masd_0s.dds"},
	{0x8CBC41EC81BF859AULL, "vxlt_0_lowtown_gc1_0_0.vlm"},
	{0x8CBC510EED492340ULL, "6171551_m.fxe"},
	{0x8CBC8003EA6A3045ULL, "138689916.ogg"},
	{0x8CBCAC0D9147B9D7ULL, "rune_weapon_sandal.dds"},
	{0x8CBD6A4C83EE6EF7ULL, "vxw_encounter2_camp_69_0_0.vlm"},
	{0x8CBD705BDDB5B60DULL, "6129627_m.fxe"},
	{0x8CBDFA06C540E706ULL, "6027500_m.fxe"},
	{0x8CBE166A900B66BEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_cth_homa_0n.dds"},
	{0x8CBE2EA3F1257C19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_enter.evt"},
	{0x8CBE385B35A5FD35ULL, "vxht_mansion2b_6r_0_0.vlm"},
	{0x8CBE8574BE65A365ULL, "llac_lt_0_lowtown_a2_-65536_6.dds"},
	{0x8CBF0BC897D8C0F6ULL, "74363858.ogg"},
	{0x8CBF35DBC546CD03ULL, "face_face_p20.psh.fxo.dx11"},
	{0x8CBF8ADE1470881FULL, "hlac_lt_0_lowtown_a2_n_c_65535_0y.dds"},
	{0x8CBFAE3FBE290A17ULL, "vxca_caverns1_imh_0_0.vlm"},
	{0x8CBFD91EC71CFA1BULL, "vxlt_backalley_n_1xk_0_0.vlm"},
	{0x8CC0160C0E8523E8ULL, "580313242.ogg"},
	{0x8CC104941D129889ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d03_0t.dds"},
	{0x8CC1945B8F41F0F2ULL, "966522808.ogg"},
	{0x8CC1B34C72EFEADAULL, "zz_isa_debug6133199.cl"},
	{0x8CC21DB3F53E0BA3ULL, "81881960.ogg"},
	{0x8CC232B1DFF3B5A8ULL, "139125924.ogg"},
	{0x8CC247EBA6709A75ULL, "t3_arm_rgb.tnt"},
	{0x8CC2582C26AD2CC5ULL, "hlac_w_woundedcoast_327684_0x.dds"},
	{0x8CC31AD6C8C9FCA1ULL, "vxlt_0_lowtown_a3_n_c_hnj_0_0.vlm"},
	{0x8CC4063816827744ULL, "vxca_caverns2prc_9qa_0_0.vlm"},
	{0x8CC4766983EDD2ADULL, "zz_ser_debug6086992.cl"},
	{0x8CC4C1367E19D538ULL, "lt_0_lowtown_docks_1_0.rml"},
	{0x8CC5599E2E3026D7ULL, "vxlt_undercityserialkiller_1je_0_1.vlm"},
	{0x8CC5BA63A28F22CDULL, "qun221_petrice6077324.cl"},
	{0x8CC615929802D95EULL, "vxlt_0_lowtown_a2_n_d4t_0_0.vlm"},
	{0x8CC61620FCA8D329ULL, "vxlt_0_lowtown_a3_n_c_f5e_0_0.vlm"},
	{0x8CC6E7675C9A71F5ULL, "llac_ht_0_hightown_-65537_3.dds"},
	{0x8CC771CDBA087776ULL, "6167895_m.fxe"},
	{0x8CC7A4D64A823DA9ULL, "254421782.ogg"},
	{0x8CC8FFC18070AFF4ULL, "6075260_m.fxe"},
	{0x8CC94AAB3DB45A77ULL, "6047262_m.fxe"},
	{0x8CC94C9D1973E46BULL, "vxw_blightlands_13x_0_0.vlm"},
	{0x8CC9CE0E9485E5E8ULL, "vxw_sundermount_2d3_0_0.vlm"},
	{0x8CC9DC87A4EB124BULL, "vxht_0_hightown_n_bup_0_0.vlm"},
	{0x8CCA011E58A3AAECULL, "vxlt_0_lowtown_a2_n_eit_0_0.vlm"},
	{0x8CCA018C3C888DC7ULL, "6209378_m.fxe"},
	{0x8CCABA3F2B531CCBULL, "vxlt_gallowstemplar_n_45q_0_0.vlm"},
	{0x8CCAC2EBDF3F3BC0ULL, "6174651_m.fxe"},
	{0x8CCAD534C64CB377ULL, "dae000_merchant_generic_hm6172653.cl"},
	{0x8CCB1144EEFDCE79ULL, "192019018.ogg"},
	{0x8CCB2C29837F29F6ULL, "ht_0_hightown_n_8_0.msh"},
	{0x8CCB3629837F3AF4ULL, "ht_0_hightown_n_8_0.mmh"},
	{0x8CCB36A3F1301408ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_enter.ani"},
	{0x8CCB3A29837F4107ULL, "ht_0_hightown_n_8_0.mao"},
	{0x8CCB58561E849F1EULL, "6167496_f.fxe"},
	{0x8CCB59D817707481ULL, "static_static_p73.psh.fxo.dx11"},
	{0x8CCB83C8B7246887ULL, "gen00fl_aveline6139601.cl"},
	{0x8CCC093D3E2D361BULL, "6061811_m.fxe"},
	{0x8CCC38796BDBEA16ULL, "bink_bink_p1.psh.fxo.dx11"},
	{0x8CCC974C580C4142ULL, "seb321_leliana6180918.cl"},
	{0x8CCD06DF82FDD54DULL, "follower_banter6151308.cl"},
	{0x8CCDCB6594D35F8CULL, "qcr220_grey_wardens6135934.cl"},
	{0x8CCE11482BC90079ULL, "dae000_gamlen6022567.cl"},
	{0x8CCE431807C83B70ULL, "lt_0_lowtown_a2_35_0.rml"},
	{0x8CCE602D1FFAC54CULL, "vxht_brothel_n_6dc_0_0.vlm"},
	{0x8CD01F9613A61DE4ULL, "6115593_m.fxe"},
	{0x8CD077170D611E4EULL, "276186183.ogg"},
	{0x8CD08DC86478B7A0ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_3z.dds"},
	{0x8CD0A3930D7F9A09ULL, "521331781.ogg"},
	{0x8CD0E173B103CC24ULL, "mcr310_argument6182458.cl"},
	{0x8CD18CB3BAC01521ULL, "vxlt_0_lowtown_a3_n_c_hki_0_0.vlm"},
	{0x8CD1DF4205180631ULL, "vxlt_darktown_6ot_0_0.vlm"},
	{0x8CD1FF86EB1A1713ULL, "art\\vfx\\combat\\vfx_glb_ele_f_major_c.impa.ani"},
	{0x8CD253986610E4D9ULL, "vxlt_house_er_0_0.vlm"},
	{0x8CD34CD130FB1522ULL, "6107211_m.fxe"},
	{0x8CD35CED324D3CA9ULL, "follower_banter6173317.cl"},
	{0x8CD35F0D89E15F34ULL, "6151399_m.fxe"},
	{0x8CD360DAF34639C5ULL, "hlac_lt_0_lowtown_n_-2_3y.dds"},
	{0x8CD37C7637D296DBULL, "6030435_m.fxe"},
	{0x8CD3885A23ED0A3EULL, "vxht_mansion2_1u5_0_0.vlm"},
	{0x8CD45051FBF88663ULL, "6078967_m.fxe"},
	{0x8CD465E881847948ULL, "292967601.ogg"},
	{0x8CD468A9D85C8249ULL, "vxw_blightlands_lv_0_0.vlm"},
	{0x8CD4BE836018EBE9ULL, "dae121ar_chantry.lst"},
	{0x8CD51A71A1B06540ULL, "590018511.ogg"},
	{0x8CD55B1B894346E7ULL, "hlac_lt_0_lowtown_n_-65536_7x.dds"},
	{0x8CD5E5D28452869BULL, "vxca_caverns1_j28_0_2.vlm"},
	{0x8CD62E61B046D22EULL, "vxlt_undercityserialkiller_1iu_0_1.vlm"},
	{0x8CD6AC17885B094AULL, "hlac_w_woundedcoast_262144_0z.dds"},
	{0x8CD72FFEA0E99DA5ULL, "6028808_f.fxe"},
	{0x8CD747D854DCB59AULL, "1012762818.ogg"},
	{0x8CD75B857F4A14DDULL, "hlac_lt_warehouse_0_3y.dds"},
	{0x8CD7803C908ED671ULL, "42314809.ogg"},
	{0x8CD7B12C1934F27EULL, "140600096.ogg"},
	{0x8CD7DBAF5241E249ULL, "372088430.ogg"},
	{0x8CD80A8CE2579DF9ULL, "1028565400.ogg"},
	{0x8CD83FEB867BFE68ULL, "vxw_sundermount_2i0_0_0.vlm"},
	{0x8CD84E6496E12ACBULL, "vxht_0_hightown_n_bps_0_0.vlm"},
	{0x8CD86B53FFD6F315ULL, "vxlt_warehouse_13v_0_0.vlm"},
	{0x8CD872FB4A99C36CULL, "vxlt_0_lowtown_a2_n_eru_0_0.vlm"},
	{0x8CD8F619999F46DDULL, "qun160_vashothlead6055569.cl"},
	{0x8CD91298406AF8EBULL, "gen_im_arm_boo_lgt_start.uti"},
	{0x8CD935F8D3822D16ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_dwhammer.mao"},
	{0x8CD941F8D38241B5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_dwhammer.mmh"},
	{0x8CD943F8D382452BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_dwhammer.msh"},
	{0x8CD96AE04BF26C7AULL, "mer240_jansen6081452.cl"},
	{0x8CDA1A620862078AULL, "6045856_m.fxe"},
	{0x8CDA27BBF5B9F3E4ULL, "gen00fl_anders6196688.cl"},
	{0x8CDAAB34BA447BDFULL, "59177785.ogg"},
	{0x8CDABD2C9C25293FULL, "llac_w_blightlands_131074_0.dds"},
	{0x8CDAC3EF57E9B980ULL, "vxdr_slavers_1bk_0_0.vlm"},
	{0x8CDB2CC2531D9D81ULL, "6156269_m.fxe"},
	{0x8CDB44B7F2E80621ULL, "vxlt_undercitydungeon_1uf_0_0.vlm"},
	{0x8CDB8ADE48071A98ULL, "mer120_hubert6059396.cl"},
	{0x8CDB9683601EC620ULL, "dae121ar_chantry.ncs"},
	{0x8CDBD2766ADAF1D0ULL, "201021356.ogg"},
	{0x8CDBD8C2A24C26B4ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_seaa_0s.dds"},
	{0x8CDBE7213B5EC683ULL, "6084520_m.fxe"},
	{0x8CDC12FC71BD863AULL, "500528280.ogg"},
	{0x8CDC1F725C537DC0ULL, "follower_banter6155934.cl"},
	{0x8CDC79DA13187763ULL, "vxw_bonepitdestroyed_a3_1vr_0_0.vlm"},
	{0x8CDC7B9B99D14DA3ULL, "6162664_m.fxe"},
	{0x8CDCFE98988237B6ULL, "6081270-speech.cut"},
	{0x8CDD2E826EB35073ULL, "hlac_lt_0_lowtown_a2_n_65533_3x.dds"},
	{0x8CDE08E971E08189ULL, "plt_lgt110pt_pinpricks.nss"},
	{0x8CDE0F34230F2E1CULL, "808982540.ogg"},
	{0x8CDE221E1A3A8911ULL, "llac_ht_mansion2_818_45.dds"},
	{0x8CDE5ABB5A435346ULL, "qun120_arishok_guard.stg"},
	{0x8CDEB06494CC7EE3ULL, "aip_templarcaptain_base.gda"},
	{0x8CDF1AC5A03DB89EULL, "gen00fl_merrill6115832.cl"},
	{0x8CDF4B6A6C82F107ULL, "lgt210_magu_selby6142009.cl"},
	{0x8CDFBC1225AF92C3ULL, "hlac_lt_0_lowtown_n_-65535_4x.dds"},
	{0x8CDFC7D2B2183177ULL, "zz_dae_debug6029126.cl"},
	{0x8CDFE6AC25CE5D03ULL, "art\\vfx\\combat\\vfx_v_fir_imm_act_c.anb"},
	{0x8CDFF786EB26474AULL, "art\\vfx\\combat\\vfx_glb_ele_f_major_c.impa.evt"},
	{0x8CE01BAA3AF4C627ULL, "6119143_m.fxe"},
	{0x8CE0386452FBA8B1ULL, "mer320_hubert6079342.cl"},
	{0x8CE0A6B4DC89CBFFULL, "art\\characters\\playercharacter\\humanmale\\hlh_lgta.msh"},
	{0x8CE0ACB4DC89D5A1ULL, "art\\characters\\playercharacter\\humanmale\\hlh_lgta.mmh"},
	{0x8CE0B8B4DC89EA42ULL, "art\\characters\\playercharacter\\humanmale\\hlh_lgta.mao"},
	{0x8CE1FA3715E322BEULL, "vxht_mansion2_1x6_0_0.vlm"},
	{0x8CE206BBCAF5FCD9ULL, "178002238.ogg"},
	{0x8CE215B4B4D7042BULL, "6167267_m.fxe"},
	{0x8CE2DA86CA529AC9ULL, "vxw_blightlands_qu_0_0.vlm"},
	{0x8CE2DCBEDB26CEA7ULL, "lt_0_lowtown_a2_n_37_0_ll.mmh"},
	{0x8CE34922B999F100ULL, "vxht_keep_n_5ih_0_0.vlm"},
	{0x8CE37295FC79DBDAULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar4_.gad"},
	{0x8CE3771AEC23FDA4ULL, "738121192.ogg"},
	{0x8CE40C228E10B3D5ULL, "mag100_karras6034917.cl"},
	{0x8CE468AE20BD8218ULL, "vxht_bartrand_n_5s1_0_0.vlm"},
	{0x8CE4B869EDEA0A93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_shake_no.evt"},
	{0x8CE4FB7DA8B3B821ULL, "follower_banter6163518.cl"},
	{0x8CE52C9122CAE21AULL, "art\\vfx\\combat\\vfx_c_rw_teleport_out_c.anb"},
	{0x8CE5AB7378D98F69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_barter.ani"},
	{0x8CE5B7836027528EULL, "dae121ar_chantry.are"},
	{0x8CE5ED77C01E7F8CULL, "plt_cod_ltr_sibling_act2.plo"},
	{0x8CE6726C495EEF04ULL, "hlac_ht_0_hightown_-196608_0x.dds"},
	{0x8CE6B1C8787216E8ULL, "vxw_sundermount_2r1_0_0.vlm"},
	{0x8CE6D686EB2C2DBDULL, "art\\vfx\\combat\\vfx_glb_ele_f_major_c.impa.gad"},
	{0x8CE6DF57BDA3F231ULL, "isa311_party_comments6117824.cl"},
	{0x8CE6E4D83C8FDBECULL, "vxlt_0_lowtown_a2_n_ewj_0_0.vlm"},
	{0x8CE72E82BF467C22ULL, "568894178.ogg"},
	{0x8CE73F6396FF06E1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbed_n.dds"},
	{0x8CE74D026AEFD2E0ULL, "472012646.ogg"},
	{0x8CE77581A9811DE0ULL, "vfx_protect02.dds"},
	{0x8CE7798CB4D4E89BULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluelyrium.msh"},
	{0x8CE78B8CB4D507E6ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluelyrium.mao"},
	{0x8CE7978CB4D51B85ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluelyrium.mmh"},
	{0x8CE7984984001FEFULL, "lt_0_lowtown_docks_a3_n_200_0.mao"},
	{0x8CE79E1904E61F49ULL, "911796624.ogg"},
	{0x8CE7A4498400338CULL, "lt_0_lowtown_docks_a3_n_200_0.mmh"},
	{0x8CE7AA4984003E7EULL, "lt_0_lowtown_docks_a3_n_200_0.msh"},
	{0x8CE7AA4F3BD2019DULL, "ss_wryme.bnk"},
	{0x8CE7BDEBD4C9B570ULL, "703155504.ogg"},
	{0x8CE82820D7D7E214ULL, "vxht_bartrand_n_68g_0_0.vlm"},
	{0x8CE873FDD641EA3CULL, "vxlt_gallowscourtyard_7fi_0_0.vlm"},
	{0x8CE94DE2C4A03B7DULL, "gen_im_trs_gem_coral_04.uti"},
	{0x8CE9AB252BBC667EULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.dura.ani"},
	{0x8CE9B2279606A2EBULL, "100423531.ogg"},
	{0x8CE9CC2B8AC71BE2ULL, "dae000_bartrand6107540.cl"},
	{0x8CE9FC6E62E95799ULL, "vxw_bonepitdestroyed_a3_1w3_0_0.vlm"},
	{0x8CEAB169EDEE6AD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_shake_no.gad"},
	{0x8CEAED788BC76623ULL, "6167564_m.fxe"},
	{0x8CEB43E703E6F64CULL, "vxht_brothel_n_6ja_0_0.vlm"},
	{0x8CEBBC5DC6E870A2ULL, "m_one101ar_ggt_cavern.wbd"},
	{0x8CEC55006FEC70E3ULL, "vxca_cavern1_haunted_i7o_0_0.vlm"},
	{0x8CECC67378DFDBEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_barter.gad"},
	{0x8CECD277C02470C3ULL, "plt_cod_ltr_sibling_act2.nss"},
	{0x8CECD620EE9EBF82ULL, "6033291_m.fxe"},
	{0x8CED05E7DB19FC65ULL, "vxlt_warehouse_198_0_0.vlm"},
	{0x8CED33ECB0E69A0EULL, "407863679.ogg"},
	{0x8CEDC3AC25DA5F2CULL, "art\\vfx\\combat\\vfx_v_fir_imm_act_c.mmh"},
	{0x8CEDD099A74E4455ULL, "fxe_lava_explo_blend.mao"},
	{0x8CEDDFA24CE11DD8ULL, "644879671.ogg"},
	{0x8CEE0DBF2C5754DAULL, "lla_ht_0_hightown_a3_65538_0.dds"},
	{0x8CEEAC4381FE27A0ULL, "gen00fl_isabela_init6136106.cl"},
	{0x8CEEB05D208EB0EDULL, "follower_banter6164032.cl"},
	{0x8CEEECBA47E6AB84ULL, "mag110_samson6018697.cl"},
	{0x8CEF5BCEEC64C0C4ULL, "vxca_caverns2prc_9fd_0_0.vlm"},
	{0x8CF06C1407D93B3EULL, "vxht_mansion2_1c7_0_0.vlm"},
	{0x8CF0A2367E3EDBF0ULL, "lt_0_lowtown_docks_1_0.anb"},
	{0x8CF0D05AE24D8EABULL, "149699612.ogg"},
	{0x8CF146D6EBB8CB7BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbarout.msh"},
	{0x8CF14C63BC48B349ULL, "vxw_blightlands_zt_0_0.vlm"},
	{0x8CF158D6EBB8E9C6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbarout.mao"},
	{0x8CF164D6EBB8FE65ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpbarout.mmh"},
	{0x8CF18AF4B963029BULL, "m_dae121ar_keep.wbd"},
	{0x8CF18C08DE86BFB9ULL, "599669773.ogg"},
	{0x8CF1C069EDF4A2B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_shake_no.ani"},
	{0x8CF2066C2441A873ULL, "vxlt_darktown_4p1_0_0.vlm"},
	{0x8CF2F61C9BCA6A2EULL, "hlac_ht_0_hightown_-65536_1z.dds"},
	{0x8CF2FC3B4D6D3190ULL, "vxlt_0_lowtown_n_gc8_0_0.vlm"},
	{0x8CF31DAAB1405BA1ULL, "6133299_m.fxe"},
	{0x8CF33BF3A3CB334DULL, "follower_banter6143224.cl"},
	{0x8CF3C37378E5F550ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_barter.evt"},
	{0x8CF4F0E1F52FA707ULL, "6051254_m.fxe"},
	{0x8CF529E20EF4FC3DULL, "hlac_lt_0_lowtown_docks_a3_65535_0z.dds"},
	{0x8CF5321E7ACD5BCBULL, "vxht_0_hightown_n_bfu_0_0.vlm"},
	{0x8CF54884EBE20E03ULL, "6089897_m.fxe"},
	{0x8CF556B52E85F46CULL, "vxlt_0_lowtown_a2_n_exk_0_0.vlm"},
	{0x8CF568423ED57020ULL, "vxht_mansion_n_6bj_0_0.vlm"},
	{0x8CF58BE7BEB91F6EULL, "t3_wpn_lst.tnt"},
	{0x8CF62AEEA8B1A5DAULL, "z_dust8_si.dds"},
	{0x8CF6AD29B9F714E1ULL, "1050058028.ogg"},
	{0x8CF70ECC3C291AD4ULL, "230276361.ogg"},
	{0x8CF76FC09E304F5CULL, "745715962.ogg"},
	{0x8CF77084D4407A8DULL, "153975281.ogg"},
	{0x8CF78AC6340F612DULL, "581165209.ogg"},
	{0x8CF78D95FC8AB2C0ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar4_.ani"},
	{0x8CF7C3252BC8CC4FULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.dura.evt"},
	{0x8CF8104372E87229ULL, "6059292_m.fxe"},
	{0x8CF865006C2ED0B6ULL, "mag131cr_keran.utc"},
	{0x8CF86E4B54DF7019ULL, "vxw_bonepitdestroyed_a3_1x2_0_0.vlm"},
	{0x8CF8912E501334BDULL, "hlac_lt_0_lowtown_1_8y.dds"},
	{0x8CF919BDEFE649C1ULL, "vxht_mansion2b_3e8_0_0.vlm"},
	{0x8CF92386D0A11B10ULL, "81986907.ogg"},
	{0x8CF9BE2A14C5CCD5ULL, "vxw_bonepitdestroyed_a3_o0_0_0.vlm"},
	{0x8CF9D9FA56586751ULL, "vxlt_undercityserialkiller_179_0_0.vlm"},
	{0x8CFA441807ED77D8ULL, "lt_0_lowtown_a2_35_0.anb"},
	{0x8CFAF5A4ED1CF384ULL, "gen00fl_aveline6128552.cl"},
	{0x8CFB75488AA19858ULL, "6106526_m.fxe"},
	{0x8CFBA2A1703D147AULL, "hlac_ht_brothel_n_3371_1z.dds"},
	{0x8CFBE268B214CCA0ULL, "vxlt_hangedman_n_141_0_0.vlm"},
	{0x8CFBFB82EDA0E534ULL, "vxw_woundedcoast_1ky_0_0.vlm"},
	{0x8CFC593CE47A62F0ULL, "dae021_sister_samea_sta6150186.cl"},
	{0x8CFCB60797D67833ULL, "479375907.ogg"},
	{0x8CFD798FB4A67EEEULL, "6174925_m.fxe"},
	{0x8CFE8A495E18AC82ULL, "453692876.ogg"},
	{0x8CFEB3B809C2D1AEULL, "415049443.ogg"},
	{0x8CFEC4252BCEECF8ULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.dura.gad"},
	{0x8D007D6A7DE4A77BULL, "940017385.ogg"},
	{0x8D008717ED9FC150ULL, "vxht_chantry_261_0_0.vlm"},
	{0x8D00F9238BC33F0EULL, "art\\vfx\\models\\vfx_ogre_fist_left_short.msh"},
	{0x8D0113238BC36BFCULL, "art\\vfx\\models\\vfx_ogre_fist_left_short.mmh"},
	{0x8D0149A84159C117ULL, "ch1notnt_d.psh.fxo"},
	{0x8D0191247A892D29ULL, "179629977.ogg"},
	{0x8D02656C2200B01EULL, "gen00fl_bethany6101451.cl"},
	{0x8D0292FD83278FA4ULL, "6054900_m.fxe"},
	{0x8D02AE95C0D9B723ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rdy.evt"},
	{0x8D032A9F4F0DE5BFULL, "vxlt_0_lowtown_a2_n_erz_0_0.vlm"},
	{0x8D038FD751AE8FB6ULL, "19478852.ogg"},
	{0x8D03BE1432792440ULL, "6034083_m.fxe"},
	{0x8D03C2401098E9AEULL, "928467932.ogg"},
	{0x8D04474942F7A17BULL, "art\\characters\\playercharacter\\elffemale\\rob_encn.mmh"},
	{0x8D04534942F7B518ULL, "art\\characters\\playercharacter\\elffemale\\rob_encn.mao"},
	{0x8D04654942F7D465ULL, "art\\characters\\playercharacter\\elffemale\\rob_encn.msh"},
	{0x8D04BC12DEDAA3CDULL, "vxht_0_hightown_a3_b5x_0_0.vlm"},
	{0x8D04F1E60676C480ULL, "6038689_m.fxe"},
	{0x8D05484EEA0DA8BAULL, "hlac_lt_0_lowtown_-1_3x.dds"},
	{0x8D057872EB2BDCBAULL, "vxht_mansion_n_6jx_0_0.vlm"},
	{0x8D05C15FF458B570ULL, "follower_banter6173152.cl"},
	{0x8D05E26EFECFA8BEULL, "llac_w_woundedcoast_262144_0.dds"},
	{0x8D0636D92D0F68D2ULL, "vxlt_0_lowtown_a3_n_c_h1t_0_0.vlm"},
	{0x8D06AFD8B5C37665ULL, "hlac_lt_0_lowtown_a3_n_c_-196607_0x.dds"},
	{0x8D073E0E8F152CB7ULL, "6180804_m.fxe"},
	{0x8D077C501306B482ULL, "hlac_w_sundermount_0_1x.dds"},
	{0x8D07A40D85AFA618ULL, "hlac_dr_0_deeproads_5842_5z.dds"},
	{0x8D07CF70E8FAC0E3ULL, "vxw_bonepitdestroyed_a3_1aw_0_0.vlm"},
	{0x8D0831D7225EA62CULL, "vxlt_shop_rn_0_0.vlm"},
	{0x8D08600ADD958C08ULL, "mud_cl.dds"},
	{0x8D086A9411B9555FULL, "llac_ca_caverns1_mines_24423_4.dds"},
	{0x8D09419B6C74DD4AULL, "dae000_fenarel6128131.cl"},
	{0x8D0972596792FDD6ULL, "vrc321ip_spooky7_obj_6.utp"},
	{0x8D098795C0DF93A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rdy.gad"},
	{0x8D0A501F45988809ULL, "docks_male_overseer_mob6176604.cl"},
	{0x8D0AD76F75C94CDBULL, "6026977_m.fxe"},
	{0x8D0B1C5DFB258C2BULL, "follower_banter6170758.cl"},
	{0x8D0B62782C313B50ULL, "6208636_f.fxe"},
	{0x8D0BFEDCBA9270E1ULL, "vrc221_bartrand6065287.cl"},
	{0x8D0C6248F91183CCULL, "vxw_sundermount_25p_0_0.vlm"},
	{0x8D0C9620D6A74D19ULL, "6113046_m.fxe"},
	{0x8D0CC308E0764D3DULL, "vis121_saemus6084921.cl"},
	{0x8D0CC46F0A8CA6E0ULL, "6088691_m.fxe"},
	{0x8D0CD6872D42F388ULL, "6131350_f.fxe"},
	{0x8D0D50BE2DC6EA72ULL, "llac_w_encounter2_262145_0.dds"},
	{0x8D0D95EE31CD9C27ULL, "cd_fen310_fenris.bnk"},
	{0x8D0DADB4DCB0C51EULL, "art\\characters\\playercharacter\\humanmale\\hlh_lgta.phy"},
	{0x8D0DE07B70BBE3DCULL, "hlac_w_bonepit_0_12y.dds"},
	{0x8D0E107B34342E69ULL, "611959435.ogg"},
	{0x8D0E1D78E245F364ULL, "ser121_jethann6104570.cl"},
	{0x8D0E2F9122EDECCDULL, "art\\vfx\\combat\\vfx_c_rw_teleport_out_c.mmh"},
	{0x8D0E3018F1C4BC3BULL, "vxht_0_hightown_n_a6r_0_0.vlm"},
	{0x8D0E3575FE90767BULL, "75103829.ogg"},
	{0x8D0E3CD3F62D6440ULL, "qun211_party_comment6072260.cl"},
	{0x8D0E42FB417DB9C7ULL, "vxht_0_hightown_7g4_0_0.vlm"},
	{0x8D0E7D9BC6EE18B1ULL, "1019752358.ogg"},
	{0x8D0F217DBA039D0AULL, "6147134_m.fxe"},
	{0x8D0F2FC26229058AULL, "6187783_m.fxe"},
	{0x8D0F491204E2126CULL, "925449157.ogg"},
	{0x8D0F6F00BC0586B8ULL, "gen00fl_merrill6106971.cl"},
	{0x8D0F7165E5FA53D0ULL, "522785199.ogg"},
	{0x8D100D89E4B86E5CULL, "art\\cinematics\\act3\\romances\\fenris\\cs_fenris_fenrisrom.ani"},
	{0x8D1042D2DA8752DCULL, "vxht_keep_n_5u0_0_0.vlm"},
	{0x8D109695C0E5CB82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rdy.ani"},
	{0x8D10BCDAFEF5BB9CULL, "and130tr_anders_intro.utt"},
	{0x8D10C007BA2FE793ULL, "201940199.ogg"},
	{0x8D10C8814682038FULL, "6108040_m.fxe"},
	{0x8D113A88EE412D8FULL, "qun110_petrice6036229.cl"},
	{0x8D119C7C4103FE3FULL, "vxlt_0_lowtown_a2_n_ewe_0_0.vlm"},
	{0x8D1215D85EB98CCBULL, "vxht_0_hightown_n_blw_0_0.vlm"},
	{0x8D12BC7F7A4281FFULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.impa.gad"},
	{0x8D13317A7A8F40EFULL, "6078729_m.fxe"},
	{0x8D1345A8683C1231ULL, "mag320_bsc_banter6100769.cl"},
	{0x8D13EA4FDD21F53AULL, "vxht_mansion_n_6my_0_0.vlm"},
	{0x8D142D9754804903ULL, "gen00fl_varric6087475.cl"},
	{0x8D1470D2F876277DULL, "dae011_gossips6149802.cl"},
	{0x8D14FAA128FDF5EDULL, "hlac_lt_0_lowtown_a2_n_c_65535_9x.dds"},
	{0x8D15520538CBA119ULL, "vxw_bonepitdestroyed_a3_1f0_0_0.vlm"},
	{0x8D15974803E9ECA3ULL, "mer240_jansen6050387.cl"},
	{0x8D161A89E4BCEF9EULL, "art\\cinematics\\act3\\romances\\fenris\\cs_fenris_fenrisrom.gad"},
	{0x8D16931CAEDCD5E0ULL, "vxlt_gallowscourtyard_n_4v9_0_0.vlm"},
	{0x8D173E0D4EB4E843ULL, "vxht_mansion_n_5mo_0_0.vlm"},
	{0x8D1774789C666F9FULL, "mag160_feynriel6107437.cl"},
	{0x8D17CDE229C32D98ULL, "vxht_keep_4bg_0_0.vlm"},
	{0x8D17E38A1D286EE3ULL, "llac_dr_slavers_3214_1.dds"},
	{0x8D17EFBC8F1F7120ULL, "6167336_m.fxe"},
	{0x8D1820A474D0A9FAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexc.phy"},
	{0x8D182ACADE847CD4ULL, "hlac_lt_backalley_g_0_11x.dds"},
	{0x8D18B17F7A46DAECULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.impa.evt"},
	{0x8D18FAD0476B4E96ULL, "mrl310_merrill6120073.cl"},
	{0x8D194582CDA2CEE1ULL, "540951412.ogg"},
	{0x8D1971F75E3A17F8ULL, "dae000_flemeth6000921.cl"},
	{0x8D1A099FC7EEE4E2ULL, "dre000_carver6037621.cl"},
	{0x8D1AD7A8AB34DF77ULL, "qun150cr_merc_2.utc"},
	{0x8D1AECA461B5D9FAULL, "vxlt_0_lowtown_a3_n_c_e15_0_0.vlm"},
	{0x8D1B542F7C45F503ULL, "hlac_lt_gallowscourtyard_-131074_0x.dds"},
	{0x8D1B59DEEF3B2A47ULL, "6094503_m.fxe"},
	{0x8D1BAD6490656844ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_exit.ani"},
	{0x8D1C10C7D55F6208ULL, "1025226541.ogg"},
	{0x8D1C2433B242BA15ULL, "398509711.ogg"},
	{0x8D1CC20EE965194EULL, "vxlt_0_lowtown_a2_n_c_cox_0_0.vlm"},
	{0x8D1D28041034DFEDULL, "240115688.ogg"},
	{0x8D1D5C02FA23F1F3ULL, "vxlt_darktown_4a2_0_0.vlm"},
	{0x8D1DB05704C1BCBDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bedroll_0.pwk"},
	{0x8D1DC8FF2096AA9EULL, "vxht_0_hightown_a3_9zu_0_0.vlm"},
	{0x8D1E228DC6CBAF95ULL, "follower_banter6164784.cl"},
	{0x8D1EF9872D791399ULL, "vxlt_0_lowtown_a3_n_c_ety_0_0.vlm"},
	{0x8D1F17CBBFF625C9ULL, "art\\vfx\\combat\\vfx_glb_ele_n_extreme_c.mmh"},
	{0x8D1F6B99D691897FULL, "hlac_lt_0_lowtown_a3_n_c_65535_5y.dds"},
	{0x8D1F943266334785ULL, "6143201_m.fxe"},
	{0x8D1FCE044108C558ULL, "hlac_lt_house_-1_6z.dds"},
	{0x8D20231845757D88ULL, "377741780.ogg"},
	{0x8D20DD5843D88A56ULL, "hlac_lt_gallowscourtyard_65536_1z.dds"},
	{0x8D21165280CA8B34ULL, "hlac_lt_0_lowtown_131071_4x.dds"},
	{0x8D21344AB4BE558DULL, "war_chargeattack.dds"},
	{0x8D220537396B0224ULL, "zz_test_general.cnv"},
	{0x8D224DA1DA31AF5AULL, "6043916_f.fxe"},
	{0x8D22621709DD9724ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_lelianna_epilogue.ani"},
	{0x8D22AA64906B81A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_exit.gad"},
	{0x8D22B0A2BC1418F6ULL, "vxht_brothel_n_4yt_0_0.vlm"},
	{0x8D23C3E22AC1B999ULL, "vxw_bonepitdestroyed_a3_1k7_0_0.vlm"},
	{0x8D23CA80A89A9E46ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_imp_p.impa.evt"},
	{0x8D245E16768F4D00ULL, "471842103.ogg"},
	{0x8D24B32ACCE6F1E3ULL, "vxw_bonepitdestroyed_a3_1oy_0_0.vlm"},
	{0x8D24CB4A5BE3B597ULL, "vxca_caverns1_i2a_0_0.vlm"},
	{0x8D25245F14B7375BULL, "vxht_0_hightown_a3_av0_0_0.vlm"},
	{0x8D25B103E52D7390ULL, "6107433_m.fxe"},
	{0x8D25D33C69019185ULL, "109174944.ogg"},
	{0x8D25FF37B9D5D6C8ULL, "vxdr_primevaltomb_6pb_0_0.vlm"},
	{0x8D26173EAAC68151ULL, "6025127_m.fxe"},
	{0x8D26352983CBE4EFULL, "ht_0_hightown_n_8_0.phy"},
	{0x8D26997F7A52EF45ULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.impa.ani"},
	{0x8D26AE04BB51384EULL, "6182285_m.fxe"},
	{0x8D273D6646C64701ULL, "426034648.ogg"},
	{0x8D274BA3991D9B67ULL, "841752839.ogg"},
	{0x8D2753C9DF19D79FULL, "6049736_m.fxe"},
	{0x8D284677941783B0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_noba_0n.dds"},
	{0x8D285D5481CF37CFULL, "and100_anders6095933.cl"},
	{0x8D285F1709E1FD86ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_lelianna_epilogue.gad"},
	{0x8D2900D64B760F63ULL, "fxc_bm_leak_add.mao"},
	{0x8D290C741F59FD2DULL, "vxlt_0_lowtown_n_eiu_0_0.vlm"},
	{0x8D29A56490719785ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_exit.evt"},
	{0x8D29B7A83188614DULL, "253909353.ogg"},
	{0x8D29B8ADA7C875CDULL, "gen00fl_aveline6131614.cl"},
	{0x8D2AA98105E5B4BBULL, "659853887.ogg"},
	{0x8D2AD278B42EF38CULL, "6118402_m.fxe"},
	{0x8D2AE180A8A0E391ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_imp_p.impa.gad"},
	{0x8D2AE85EEA774361ULL, "vxdr_primevaltomb_6wf_0_0.vlm"},
	{0x8D2B3F55515A957EULL, "6071503_m.fxe"},
	{0x8D2B52A3CF582C84ULL, "98335758.ogg"},
	{0x8D2B7BC7B4DA76FEULL, "hlac_w_encounter2_camp_-131069_0y.dds"},
	{0x8D2BB3DDD37535EBULL, "66562301.ogg"},
	{0x8D2BBB8C6648DE3CULL, "hlac_lt_0_lowtown_docks_a3_-65539_3y.dds"},
	{0x8D2BEF2C81C0CE18ULL, "gen00fl_aveline6139279.cl"},
	{0x8D2C41E40FBD0D8EULL, "472525740.ogg"},
	{0x8D2C59CBAB8BB0BCULL, "vxw_bonepitdestroyed_a3_1wj_0_0.vlm"},
	{0x8D2C6E811FB98F1FULL, "63636726.ogg"},
	{0x8D2C892FC103A1A2ULL, "vxht_0_hightown_c3e_0_0.vlm"},
	{0x8D2CB8EF6733DF00ULL, "vxw_encounter2_camp_v_0_0.vlm"},
	{0x8D2D0C954B01A83FULL, "hlac_lt_undercityserialkiller_-1_16x.dds"},
	{0x8D2D63CE7AABA2C9ULL, "18703747.ogg"},
	{0x8D2DA0372CB3A208ULL, "hlac_lt_undercitydungeon_-1_31y.dds"},
	{0x8D2DB7E7C5977C43ULL, "vrc321_party_barks6096316.cl"},
	{0x8D2DD28E1F8C38DAULL, "6095516_m.fxe"},
	{0x8D2DD426E4D7FDA3ULL, "vxlt_gallowstemplar_f_3jc_0_0.vlm"},
	{0x8D2E708CB510DC32ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_1bluelyrium.phy"},
	{0x8D2EA30FD4AAD0ECULL, "6017490_f.fxe"},
	{0x8D2EDCFF86D0B4C5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hed_visa_0d.dds"},
	{0x8D2FA31302ED2BF1ULL, "hlac_lt_0_lowtown_docks_131071_2z.dds"},
	{0x8D2FC7C356186137ULL, "vrc321_varric6087083.cl"},
	{0x8D2FDC32812EEE23ULL, "fxa_evl_spout.mao"},
	{0x8D2FE9F1799AE070ULL, "vxlt_darktown_510_0_0.vlm"},
	{0x8D2FF5B97E9FE7C3ULL, "vxca_caverns1_mines_fly_0_0.vlm"},
	{0x8D307C2CB7B288C2ULL, "llac_lt_0_lowtown_a3_n_c_-65535_2.dds"},
	{0x8D30C39540D50199ULL, "lt_0_lowtown_docks_122_0.msh"},
	{0x8D30CD9540D51287ULL, "lt_0_lowtown_docks_122_0.mmh"},
	{0x8D30D19540D519D4ULL, "lt_0_lowtown_docks_122_0.mao"},
	{0x8D30DE60D7B19110ULL, "gen_im_arm_cht_med_act3_01.uti"},
	{0x8D318192F6F93392ULL, "zz_vrc_debug6200239.cl"},
	{0x8D31CDDA8841EEF4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd04.phy"},
	{0x8D31E280A8A7043FULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_imp_p.impa.ani"},
	{0x8D320F8F879C738FULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.mmh"},
	{0x8D324047B36A8B1AULL, "vxca_caverns2_1yj_0_0.vlm"},
	{0x8D3261809FC92B77ULL, "ht_0_hightown_a2_36_0.anb"},
	{0x8D32CA5DC7248B58ULL, "m_one101ar_ggt_cavern.bnk"},
	{0x8D33188326FE75E3ULL, "hlac_lt_0_lowtown_n_-2_2x.dds"},
	{0x8D333555C486EA86ULL, "6174895_m.fxe"},
	{0x8D335F3FF229DDCDULL, "6129054_m.fxe"},
	{0x8D33876DF840EFACULL, "vxlt_shop_cm_0_0.vlm"},
	{0x8D33963C06AD4FDBULL, "vxht_0_hightown_a3_as3_0_0.vlm"},
	{0x8D339642A379433BULL, "art\\characters\\creatures\\ragedemon\\animation\\dazed.evt"},
	{0x8D33A6FBF16C36E9ULL, "mer120_hubert6172113.cl"},
	{0x8D34480434B7A0EFULL, "6161319_m.fxe"},
	{0x8D347114ABCBEF48ULL, "vxdr_primevaltomb_6wc_0_0.vlm"},
	{0x8D34BC75A5F8374BULL, "94987152.ogg"},
	{0x8D34BCE8102C8FBCULL, "vxht_keep_n_4hb_0_0.vlm"},
	{0x8D351B3C2395C04AULL, "6080626_m.fxe"},
	{0x8D3591B428DC9E3FULL, "mer320_hubert6053667.cl"},
	{0x8D3691E0065B34E1ULL, "6060101_f.fxe"},
	{0x8D36F7D82FB011F8ULL, "vxht_mansion_n_47j_0_1.vlm"},
	{0x8D3715265E7EB6E8ULL, "follower_banter6143281.cl"},
	{0x8D37341B26D8BE29ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ci_basic.phy"},
	{0x8D381AB9C366F99FULL, "t3_eye_tea.tnt"},
	{0x8D385DF2593A2CFAULL, "vxht_keep_n_36a_0_0.vlm"},
	{0x8D38B7641C190FE4ULL, "290344695.ogg"},
	{0x8D38FBBD1D7FEA9BULL, "608539073.ogg"},
	{0x8D396C7882787492ULL, "character_character_p128.psh.fxo.dx11"},
	{0x8D398F42A37DA37CULL, "art\\characters\\creatures\\ragedemon\\animation\\dazed.gad"},
	{0x8D3A3FBCDE1022F3ULL, "vxlt_darktown_4k4_0_0.vlm"},
	{0x8D3A764CA8A6881CULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_shapes.dds"},
	{0x8D3A8BEE31F4795DULL, "cd_fen310_fenris.wbd"},
	{0x8D3AFB0CB2F9BA22ULL, "vxht_0_hightown_c6f_0_0.vlm"},
	{0x8D3B31613901824DULL, "follower_banter6171588.cl"},
	{0x8D3B5AE464F527B6ULL, "atl_prccontent_dxt5_dat.xml"},
	{0x8D3BAADA884A07F1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd04.msh"},
	{0x8D3BC4DA884A33AFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd04.mmh"},
	{0x8D3BDBEE18937969ULL, "711069874.ogg"},
	{0x8D3C351BFE9533B0ULL, "lt_0_lowtown_docks_27_0_ll.mmh"},
	{0x8D3CC2143B67F17EULL, "6042807_f.fxe"},
	{0x8D3CEA6027BA4893ULL, "142533579.ogg"},
	{0x8D3D53C418836B29ULL, "6130990_f.fxe"},
	{0x8D3DF24F50BCFE1EULL, "mag110_vincento6116303.cl"},
	{0x8D3EAA39D653D511ULL, "gen_im_trs_chn_robe.uti"},
	{0x8D3EB45A6408751EULL, "6034518_f.fxe"},
	{0x8D3F40C40092CFCCULL, "6111221_m.fxe"},
	{0x8D40069C7D14F677ULL, "vxht_0_hightown_n_a6f_0_0.vlm"},
	{0x8D400DD0D78FD26DULL, "vxht_mansion_n_59m_0_0.vlm"},
	{0x8D408E42A383C07AULL, "art\\characters\\creatures\\ragedemon\\animation\\dazed.ani"},
	{0x8D4096BDCACBC73DULL, "gen00fl_aveline6122104.cl"},
	{0x8D410A0B7E690EEFULL, "vxlt_foundry_n_y8_0_0.vlm"},
	{0x8D4156EC39057A93ULL, "6126761_m.fxe"},
	{0x8D415E331CBD9AC3ULL, "gen00fl_aveline6139308.cl"},
	{0x8D417E8D1AE9A28BULL, "follower_barks6153541.cl"},
	{0x8D41851340AF7D5CULL, "vxdr_slavers_1b7_0_0.vlm"},
	{0x8D41E97AF08A493AULL, "mag311_nyssa6041950.cl"},
	{0x8D41F94AEA37082CULL, "vxlt_shop_fb_0_0.vlm"},
	{0x8D41FB716D85D7EDULL, "331257394.ogg"},
	{0x8D42E86A9D6DD726ULL, "6075345_m.fxe"},
	{0x8D431DE93B73D293ULL, "248980569.ogg"},
	{0x8D438B96858B4A84ULL, "vxlt_hangedman_n_25m_0_0.vlm"},
	{0x8D44A2D893148C34ULL, "24630023.ogg"},
	{0x8D44F8031FD06790ULL, "hlac_w_encounter1_5_0x.dds"},
	{0x8D45670CD422F9ABULL, "dae000_meredith6085292.cl"},
	{0x8D46745EC15AA695ULL, "6130052_m.fxe"},
	{0x8D4713A1A4625BFDULL, "734428782.ogg"},
	{0x8D476165D7A6E603ULL, "vxht_keep_n_73s_0_0.vlm"},
	{0x8D478C7CCE62456FULL, "llac_lt_0_lowtown_1_5.dds"},
	{0x8D47AFA419597B71ULL, "vxlt_0_lowtown_a3_n_c_ht4_0_0.vlm"},
	{0x8D47CC61AF4C379DULL, "trail_detection.ncs"},
	{0x8D47D52CC3DF1C69ULL, "6101540_m.fxe"},
	{0x8D47DA992F2B990DULL, "mrl351_pridedemon6165551.cl"},
	{0x8D47EDF0813F8345ULL, "6078815_m.fxe"},
	{0x8D4814CBC01925FEULL, "art\\vfx\\combat\\vfx_glb_ele_n_extreme_c.anb"},
	{0x8D4854CF01BD3B4FULL, "282660524.ogg"},
	{0x8D48CAA00A155382ULL, "229025477.ogg"},
	{0x8D48E311BF3F7109ULL, "409504007.ogg"},
	{0x8D4994F6F228A859ULL, "lt_0_lowtown_a3_n_c_13_0.msh"},
	{0x8D499EF6F228B947ULL, "lt_0_lowtown_a3_n_c_13_0.mmh"},
	{0x8D49A2F6F228BF94ULL, "lt_0_lowtown_a3_n_c_13_0.mao"},
	{0x8D49F93732E94B78ULL, "vxlt_gallowstemplar_n_5xh_0_0.vlm"},
	{0x8D4A3E2B6F62319BULL, "354291803.ogg"},
	{0x8D4A6E5912EE81F0ULL, "hlac_lt_0_lowtown_n_131072_4z.dds"},
	{0x8D4AD7734894BEE1ULL, "578284797.ogg"},
	{0x8D4B68494333DBB0ULL, "art\\characters\\playercharacter\\elffemale\\rob_encn.phy"},
	{0x8D4C09A937A155AEULL, "6205098_m.fxe"},
	{0x8D4C4C94B700CB68ULL, "gen_im_arm_cht_comc_t1.uti"},
	{0x8D4CC313DC36854CULL, "w_blightlands_107_0.phy"},
	{0x8D4CD4CEE0DEBF7BULL, "6180440_m.fxe"},
	{0x8D4D166C64EBC56EULL, "lgt220pt_mag_meeting.ncs"},
	{0x8D4D1DDCFDFA745BULL, "gamlens_intro.cut"},
	{0x8D4D2FF2A75A1C5FULL, "644801363.ogg"},
	{0x8D4D472D8769CB82ULL, "l_lgt111ar_rand_sewer_1_nt.wbd"},
	{0x8D4D622AB2170425ULL, "vxht_keep_n_5tz_0_0.vlm"},
	{0x8D4DE4B81C3F0C26ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.up.ani"},
	{0x8D4EA27AC22EAD80ULL, "hlac_lt_0_lowtown_a2_n_131071_3x.dds"},
	{0x8D4EF104698766C8ULL, "494950416.ogg"},
	{0x8D4F324779D906FCULL, "6093303_m.fxe"},
	{0x8D4FB9C24F4ADC1DULL, "w_encounter3_164_0.phy"},
	{0x8D4FBDB081D2BDB6ULL, "34421916.ogg"},
	{0x8D502FF83ECCDF2AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent.gad"},
	{0x8D50C9F7C8147231ULL, "17456875.ogg"},
	{0x8D50EFF659D87F97ULL, "art\\vfx\\combat\\vfx_str_att_front_r_leg_c.anb"},
	{0x8D50F99B69566B64ULL, "mag160_varian_ilithis6016505.cl"},
	{0x8D51632346981F75ULL, "794060494.ogg"},
	{0x8D518A015A6C0608ULL, "vxht_0_hightown_ajz_0_0.vlm"},
	{0x8D521EAD5479FD12ULL, "ca_cave.arl"},
	{0x8D5349B51C88AEC9ULL, "6099168_m.fxe"},
	{0x8D53C64878A24D0CULL, "llac_dr_primevaltomb_659_15.dds"},
	{0x8D542786865CC767ULL, "6174043_m.fxe"},
	{0x8D54B68F8C2F49C6ULL, "6169408_m.fxe"},
	{0x8D554811A26D3E86ULL, "plt_cod_cha_carver2.plo"},
	{0x8D5549BE4034B70BULL, "6019885_m.fxe"},
	{0x8D55838251FAA274ULL, "654670996.ogg"},
	{0x8D55D44045556BE3ULL, "ave331_jeven6138748.cl"},
	{0x8D563F9423CF5BB8ULL, "602850492.ogg"},
	{0x8D56B6EB7CD5FBE8ULL, "gen00fl_isabela_init6076480.cl"},
	{0x8D56C1F4F274F6F0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_beta_d.dds"},
	{0x8D5712F83ED2CCC1ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent.evt"},
	{0x8D575A52284A6D9FULL, "gen_im_arm_cht_nobe_t6.uti"},
	{0x8D57AF62816DFA3CULL, "vxw_bonepitdestroyed_a3_1fi_0_0.vlm"},
	{0x8D57DC3932FBD79DULL, "vxht_bartrand_n_5mj_0_0.vlm"},
	{0x8D583FDE0929AC04ULL, "dae110ip_at_tavern.utp"},
	{0x8D588EB444E83767ULL, "dae000_quintus6154711.cl"},
	{0x8D58CB3EA05B3FF1ULL, "6177704_f.fxe"},
	{0x8D5903A176DA5612ULL, "vxht_0_hightown_a3_8q1_0_0.vlm"},
	{0x8D5915CBDE5F2D42ULL, "460055204.ogg"},
	{0x8D5955BA0E25AA35ULL, "vrc321st_varric.stg"},
	{0x8D59E9D106A2EB90ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_3z.dds"},
	{0x8D5A108F87BDC7CCULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.anb"},
	{0x8D5A39EB30097CF2ULL, "vxlt_undercityserialkiller_27i_0_0.vlm"},
	{0x8D5A60E3EADBD33DULL, "vxw_woundedcoast_t8_0_0.vlm"},
	{0x8D5AA7A646E90954ULL, "dae000_madam_lusine6152158.cl"},
	{0x8D5B9E041318F807ULL, "zz_one_debug6073985.cl"},
	{0x8D5C000212A5DEB8ULL, "vxht_0_hightown_cay_0_0.vlm"},
	{0x8D5C2D035F5FB106ULL, "vxht_0_hightown_ca7_0_0.vlm"},
	{0x8D5CB69540FA26BCULL, "lt_0_lowtown_docks_122_0.phy"},
	{0x8D5E1A9F39732FBEULL, "430179581.ogg"},
	{0x8D5E7BDA2D9FF34EULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyn.msh"},
	{0x8D5E89DA2DA00A9FULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyn.mao"},
	{0x8D5E95DA2DA01F3CULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyn.mmh"},
	{0x8D5EF704F41312D1ULL, "vxlt_undercityserialkiller_1j2_0_0.vlm"},
	{0x8D5F5E809FF0138BULL, "ht_0_hightown_a2_36_0.rml"},
	{0x8D5FD6B97854FD2FULL, "dae140st_dragon_reveal.stg"},
	{0x8D5FECE2FC654C58ULL, "vxdr_primevaltomb_8oq_0_0.vlm"},
	{0x8D602315CE49D335ULL, "327632448.ogg"},
	{0x8D604E2B1A0B80FBULL, "6079039_f.fxe"},
	{0x8D60AA77A01FC55EULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.cess.evt"},
	{0x8D60AEC24F593C89ULL, "w_encounter3_164_0.mao"},
	{0x8D60BAC24F595122ULL, "w_encounter3_164_0.mmh"},
	{0x8D60BCC24F5954D8ULL, "w_encounter3_164_0.msh"},
	{0x8D60BCCFAD18536CULL, "6217150_m.fxe"},
	{0x8D60C963B26022EEULL, "vxw_encounter2_camp_s3_0_0.vlm"},
	{0x8D6161B0F86B4030ULL, "vxw_bonepitdestroyed_a3_79_0_0.vlm"},
	{0x8D61837D6FAA8C49ULL, "6083062_m.fxe"},
	{0x8D61CDB81C4F8E20ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.up.gad"},
	{0x8D621E16D1C92519ULL, "art\\characters\\creatures\\dragon\\c_dragon_2x.msh"},
	{0x8D622816D1C93607ULL, "art\\characters\\creatures\\dragon\\c_dragon_2x.mmh"},
	{0x8D62BEC4978F8319ULL, "llac_ht_chantry_n_2534_41.dds"},
	{0x8D62C2E94B4C3EE0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vault_sa.dds"},
	{0x8D62D5231465115CULL, "58667448.ogg"},
	{0x8D63061F5733CFCFULL, "pro000_colonel_ewald6040562.cl"},
	{0x8D6308BFA18D2414ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala3a.phy"},
	{0x8D6316AC0F365158ULL, "6049447_m.fxe"},
	{0x8D631D1B26FDD24EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ci_basic.mmh"},
	{0x8D63291B26FDE6E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ci_basic.mao"},
	{0x8D63371B26FDFE3CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ci_basic.msh"},
	{0x8D63C8908C3AD4B0ULL, "6034361_f.fxe"},
	{0x8D63EF22E8297E81ULL, "vxdr_slavers_1bf_0_0.vlm"},
	{0x8D63FA72FA17B1ACULL, "vxlt_gallowscourtyard_76i_0_0.vlm"},
	{0x8D640AF83EDD4910ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ent.ani"},
	{0x8D6471A5FCD5A784ULL, "gen_im_npc_wep_shd_drka.uti"},
	{0x8D6507819D48CF28ULL, "6127488_m.fxe"},
	{0x8D65684CC52632DDULL, "6004158_f.fxe"},
	{0x8D66213F736412BCULL, "vxw_bonepitdestroyed_a3_1kn_0_0.vlm"},
	{0x8D6696B77E459AE8ULL, "6115500_m.fxe"},
	{0x8D66A1CF21DD6A51ULL, "10083768.ogg"},
	{0x8D6760A9CDB722E5ULL, "darktown_male_fereldan_sta6175142.cl"},
	{0x8D67C177A0260A89ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.cess.gad"},
	{0x8D67E94E3A0E81FDULL, "dae000_bartrand6027544.cl"},
	{0x8D6802721ED28F9AULL, "mrl310_merrill6144361.cl"},
	{0x8D684C1EF870340EULL, "hlac_lt_0_lowtown_docks_-2_0y.dds"},
	{0x8D68ECB6B84A9C30ULL, "hlac_lt_0_lowtown_docks_a3_n_0_0z.dds"},
	{0x8D69150407A291E9ULL, "gen00fl_merrill6128774.cl"},
	{0x8D692CE7DA8417A0ULL, "and100_anders6111661.cl"},
	{0x8D6942E34423F0BEULL, "6168986_m.fxe"},
	{0x8D697DA95D195426ULL, "6150923_m.fxe"},
	{0x8D69DB2061C90099ULL, "vxdr_primevaltomb_8v7_0_0.vlm"},
	{0x8D69F8773F821C99ULL, "6160655_m.fxe"},
	{0x8D6A1DB29AF5F9F3ULL, "character_character_p132.psh.fxo"},
	{0x8D6A9EE05155C986ULL, "vxht_0_hightown_cd8_0_0.vlm"},
	{0x8D6AA333B2450BF4ULL, "mag170_hugh6080699.cl"},
	{0x8D6AD033B68C3AFAULL, "vxlt_warehouse_n_13z_0_0.vlm"},
	{0x8D6B0AAD04F9EB1EULL, "354155898.ogg"},
	{0x8D6B1653F97E0DA7ULL, "6128859_m.fxe"},
	{0x8D6B2A75822D7143ULL, "vxlt_hangedman_1js_0_0.vlm"},
	{0x8D6B2D72E0AC6930ULL, "gen00fl_aveline6137479.cl"},
	{0x8D6BA48B9E7C6551ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_1.pwk"},
	{0x8D6BB3DAFF425839ULL, "and130tr_anders_intro.ncs"},
	{0x8D6C745649D9B601ULL, "vxht_keep_n_5ie_0_0.vlm"},
	{0x8D6CC1979DE60723ULL, "t1_mub_re2.tnt"},
	{0x8D6CE5BFA1953D11ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala3a.msh"},
	{0x8D6CE7BD496EF6B3ULL, "6208373_m.fxe"},
	{0x8D6CF3BFA195552CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala3a.mao"},
	{0x8D6CFFBFA19569CFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala3a.mmh"},
	{0x8D6D1CB1FAD19B16ULL, "vxht_keep_36v_0_0.vlm"},
	{0x8D6D5516BAFCFEF8ULL, "llac_lt_0_lowtown_a3_n_c_65538_4.dds"},
	{0x8D6D68E1E6092B51ULL, "vxlt_undercityserialkiller_1o1_0_0.vlm"},
	{0x8D6D6B0A0BC4DD7CULL, "vxca_cave_1dt_0_0.vlm"},
	{0x8D6DF272D24430DBULL, "692671648.ogg"},
	{0x8D6E9277A02BD9D7ULL, "art\\vfx\\combat\\vfx_w_2nd_mblow_c.cess.ani"},
	{0x8D6EFD1387E98779ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masa_0d.dds"},
	{0x8D6F2A5EEC4D8FEDULL, "art\\characters\\weapons\\mauls\\w_mal_dsml.mao"},
	{0x8D6F2DE5396B4C5CULL, "w_bonepit_227_0.mao"},
	{0x8D6F39E5396B60FFULL, "w_bonepit_227_0.mmh"},
	{0x8D6F3A26993D7FEDULL, "6135121_m.fxe"},
	{0x8D6F3B40A4563B6EULL, "vxw_encounter2_camp_n4_0_0.vlm"},
	{0x8D6F3FE5396B6AA1ULL, "w_bonepit_227_0.msh"},
	{0x8D6F843A2EB0AF11ULL, "6125628_f.fxe"},
	{0x8D6F9100CB498A51ULL, "98523105.ogg"},
	{0x8D6FA261A39078C4ULL, "6111298_m.fxe"},
	{0x8D6FAFF83E4DC7F3ULL, "6052930_m.fxe"},
	{0x8D6FBBF858CCF29FULL, "579297923.ogg"},
	{0x8D70FAB68795BC00ULL, "zz_otl_debug6060585.cl"},
	{0x8D7180EDD37F7660ULL, "gen00fl_aveline6122774.cl"},
	{0x8D7256398AB984C4ULL, "vxlt_0_lowtown_a3_n_c_ewq_0_0.vlm"},
	{0x8D72808C7779E8BEULL, "235518929.ogg"},
	{0x8D72DF4F26A0503BULL, "vxdr_slavers_14f_0_0.vlm"},
	{0x8D72F39D0A424B74ULL, "dm_temmerin.mor"},
	{0x8D73108D8577BC83ULL, "339701697.ogg"},
	{0x8D7319A4751D4923ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexc.msh"},
	{0x8D7327A4751D617DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexc.mmh"},
	{0x8D732BA4751D678EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexc.mao"},
	{0x8D733BA72D8072CAULL, "6042839_m.fxe"},
	{0x8D73A1D25BD75E39ULL, "zz_dre_debug6092020.cl"},
	{0x8D73EF87F301B5FBULL, "lt_0_lowtown_docks_n_114_0.mmh"},
	{0x8D73FB44742E0738ULL, "6109233_m.fxe"},
	{0x8D73FB87F301C998ULL, "lt_0_lowtown_docks_n_114_0.mao"},
	{0x8D740D87F301E8E5ULL, "lt_0_lowtown_docks_n_114_0.msh"},
	{0x8D742BF00BFE841BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgb.msh"},
	{0x8D7431DA29CD6478ULL, "6125240_m.fxe"},
	{0x8D743DF00BFEA366ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgb.mao"},
	{0x8D7449F00BFEB705ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgb.mmh"},
	{0x8D74767D1004DEF5ULL, "vxlt_0_lowtown_a2_n_c_gzd_0_0.vlm"},
	{0x8D74931C655A2B3CULL, "vxw_bonepitdestroyed_a3_1lo_0_0.vlm"},
	{0x8D749E04A7018193ULL, "754855346.ogg"},
	{0x8D74BFA2963FC156ULL, "hlac_ca_caverns1_24424_19y.dds"},
	{0x8D7507BB4F08732AULL, "follower_barks6153920.cl"},
	{0x8D754405CBDF7F8EULL, "419552534.ogg"},
	{0x8D7544EC4B7E7BCAULL, "dae000cr_grand_cleric6148088.cl"},
	{0x8D756EEFEEDDB959ULL, "hlac_w_blightlands_196606_0x.dds"},
	{0x8D759865646F69A3ULL, "vxca_caverns2_93m_0_0.vlm"},
	{0x8D75AE843C3B2C92ULL, "hlac_dr_0_deeproads_5843_4x.dds"},
	{0x8D75CD5BD235C5C9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medb_0d.dds"},
	{0x8D75F48466B74F84ULL, "lt_0_lowtown_docks_226_0.mmh"},
	{0x8D75F88466B756D7ULL, "lt_0_lowtown_docks_226_0.mao"},
	{0x8D760A8466B77526ULL, "lt_0_lowtown_docks_226_0.msh"},
	{0x8D7687F6F24F807CULL, "lt_0_lowtown_a3_n_c_13_0.phy"},
	{0x8D770F297460115AULL, "vxlt_hangedman_n_2mk_0_0.vlm"},
	{0x8D773043E5B4F503ULL, "hlac_ca_caverns1_mines_24423_6z.dds"},
	{0x8D7748FE478779F0ULL, "6207656_m.fxe"},
	{0x8D7768530CAF2B5FULL, "238884365.ogg"},
	{0x8D784A9A3F501197ULL, "llac_ca_caverns1_24426_8.dds"},
	{0x8D78C359BC8E2477ULL, "hlac_ca_caverns1_24424_1y.dds"},
	{0x8D78CD4440964F82ULL, "sse_ss_carta_thug_m.gda"},
	{0x8D78D3B21D30993DULL, "6164417_m.fxe"},
	{0x8D79B00E6E263769ULL, "425619701.ogg"},
	{0x8D79D7C98B3DE72AULL, "vxlt_0_lowtown_docks_a3_n_cp1_0_0.vlm"},
	{0x8D79ECF659FB7F98ULL, "art\\vfx\\combat\\vfx_str_att_front_r_leg_c.mmh"},
	{0x8D79F0EAAF199D79ULL, "dae000_denier6162691.cl"},
	{0x8D7A0B5708B161B7ULL, "6182279_m.fxe"},
	{0x8D7A56897D5CAF5DULL, "724132697.ogg"},
	{0x8D7AF48824F1EE42ULL, "939497313.ogg"},
	{0x8D7B4789BC6580D7ULL, "72886075.ogg"},
	{0x8D7B7E7CFB9A54BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\groundstab.evt"},
	{0x8D7BCD6378387E95ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb4.msh"},
	{0x8D7BCF637838820BULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb4.mmh"},
	{0x8D7C05CFEBE5AAC3ULL, "6177811_m.fxe"},
	{0x8D7C07751EB7E1D9ULL, "vxht_keep_n_674_0_0.vlm"},
	{0x8D7C3DE37BA262D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_local_ambient_tavern.anb"},
	{0x8D7C9EF6F25412A4ULL, "lt_0_lowtown_a3_n_c_13_0.rml"},
	{0x8D7D86F26793FE30ULL, "6117905_f.fxe"},
	{0x8D7DB22631FEEC6DULL, "vxw_encounter2_camp_rw_0_0.vlm"},
	{0x8D7EC8044F153D1FULL, "vxca_caverns1_j3t_0_1.vlm"},
	{0x8D7ECCCDB1938534ULL, "291767381.ogg"},
	{0x8D7ED9C3B2580C6EULL, "hlac_ca_cave_2194_4y.dds"},
	{0x8D7F4161290076BBULL, "art\\characters\\creatures\\strider\\animation\\scream.evt"},
	{0x8D7F45034070EDABULL, "hlac_lt_0_lowtown_a2_n_131070_3x.dds"},
	{0x8D7FE694A9051C34ULL, "539073322.ogg"},
	{0x8D802807E210EFC3ULL, "699828470.ogg"},
	{0x8D8037F97F064BC3ULL, "vxw_encounter2_camp_7f_0_0.vlm"},
	{0x8D8061A3E22E8425ULL, "vxlt_0_lowtown_docks_c5a_0_0.vlm"},
	{0x8D809C797BA935FBULL, "gen00fl_aveline6137508.cl"},
	{0x8D80E2E6D489103CULL, "vxdr_primevaltomb_8ek_0_0.vlm"},
	{0x8D80E8D451F15CEEULL, "291138859.ogg"},
	{0x8D8124AA3DCD2916ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v2.evt"},
	{0x8D812C52E453EDCBULL, "vxht_mansion_n_60o_0_0.vlm"},
	{0x8D816C00B07CB769ULL, "15324734.ogg"},
	{0x8D817BE31EC405BAULL, "enchant.ncs"},
	{0x8D81BF008175803BULL, "vxht_0_hightown_n_anj_0_0.vlm"},
	{0x8D8205D0F1E4D629ULL, "nat380_sibling6103592.cl"},
	{0x8D822DF6B5B6AC02ULL, "mrl150_hunter.cnv"},
	{0x8D828D40EC31C7A8ULL, "zz_mcr_debug6098523.cl"},
	{0x8D82957CFBA09AE9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\groundstab.gad"},
	{0x8D83652E4BA02D18ULL, "720274552.ogg"},
	{0x8D8374DA2DBF0307ULL, "art\\characters\\playercharacter\\elffemale\\bdy_hvyn.phy"},
	{0x8D83BEC012026B08ULL, "one100dg_nabil6054974.cl"},
	{0x8D8441D521015571ULL, "452011073.ogg"},
	{0x8D84C8DF98ED619FULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb104a_0s.dds"},
	{0x8D850984DF4809EEULL, "81603341.ogg"},
	{0x8D853A612904D6FCULL, "art\\characters\\creatures\\strider\\animation\\scream.gad"},
	{0x8D854EBBA359CB1AULL, "16215243.ogg"},
	{0x8D856975DAEAE4D0ULL, "llac_ht_0_hightown_a3_65536_1.dds"},
	{0x8D85773007823235ULL, "hlac_lt_0_lowtown_docks_0_1z.dds"},
	{0x8D857C2D22F17493ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.activetoinactive.ani"},
	{0x8D860FF6B5BA623AULL, "mrl150_hunter.bnk"},
	{0x8D86B9D861480489ULL, "follower_barks6153165.cl"},
	{0x8D86BA6EA743B788ULL, "w_bonepitmining_a2_79_0.mmh"},
	{0x8D86C66EA743CC2BULL, "w_bonepitmining_a2_79_0.mao"},
	{0x8D86D86EA743EAB2ULL, "w_bonepitmining_a2_79_0.msh"},
	{0x8D87299E79EF6625ULL, "vxht_keep_n_5hf_0_0.vlm"},
	{0x8D881009313DF3AAULL, "473049281.ogg"},
	{0x8D881BAA3DD338A1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v2.gad"},
	{0x8D884D45ED3C672FULL, "llac_ht_chantry_n_2534_9.dds"},
	{0x8D8870986369B914ULL, "dae000_meredith6092565.cl"},
	{0x8D88EE8B8BF2E695ULL, "6135373_f.fxe"},
	{0x8D88FEA7C263A14CULL, "t3_shl_ice.tnt"},
	{0x8D89667CFBA66937ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\groundstab.ani"},
	{0x8D8A426FC12A8A7BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv3.evt"},
	{0x8D8A84F069E76B20ULL, "6174567_m.fxe"},
	{0x8D8AB5643D68619CULL, "hlac_lt_gamlenshouse_n_-1_2x.dds"},
	{0x8D8ABFB72E9A8F61ULL, "gen00fl_varric6104992.cl"},
	{0x8D8AC852B5999F39ULL, "vxw_sundermount_27c_0_0.vlm"},
	{0x8D8ADD9A33D233E9ULL, "577561585.ogg"},
	{0x8D8B06463591AE17ULL, "6070435_m.fxe"},
	{0x8D8B6F8A76D0FB76ULL, "673612814.ogg"},
	{0x8D8BB760B4ECD77AULL, "gen_im_trs_dwv_junk5_01.uti"},
	{0x8D8BD1858B195875ULL, "6074503_m.fxe"},
	{0x8D8C1EFA88426C6EULL, "vxw_encounter2_camp_d6_0_0.vlm"},
	{0x8D8C240323F504EDULL, "vxw_encounter2_camp_mv_0_0.vlm"},
	{0x8D8C3961290AF3FAULL, "art\\characters\\creatures\\strider\\animation\\scream.ani"},
	{0x8D8C51DA51167421ULL, "fxe_cracked_glass1.mao"},
	{0x8D8C9A13DC6C2317ULL, "w_blightlands_107_0.mmh"},
	{0x8D8C9E13DC6C2A24ULL, "w_blightlands_107_0.mao"},
	{0x8D8CB013DC6C4889ULL, "w_blightlands_107_0.msh"},
	{0x8D8CB9D67AFB3237ULL, "902657110.ogg"},
	{0x8D8CD09256331B43ULL, "llac_w_encounter2_196612_0.dds"},
	{0x8D8D43AAED6B7AA6ULL, "509858119.ogg"},
	{0x8D8D8E589C79E806ULL, "6125182_f.fxe"},
	{0x8D8DF8892E82BB4DULL, "and100_anders6095816.cl"},
	{0x8D8E1116D1EE4A3CULL, "art\\characters\\creatures\\dragon\\c_dragon_2x.phy"},
	{0x8D8EAB59066773B6ULL, "gen_im_cft_reg_h.uti"},
	{0x8D8EB1CAF0D7FDF5ULL, "6083586_f.fxe"},
	{0x8D8F1CAA3DD9594FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_atk_v2.ani"},
	{0x8D8F481AE897A07CULL, "936431770.ogg"},
	{0x8D8F5E986BC7181CULL, "vxht_keep_n_60h_0_0.vlm"},
	{0x8D8F6C8992426A9AULL, "6048114_m.fxe"},
	{0x8D90101B11D1CF48ULL, "702689923.ogg"},
	{0x8D903A079E882B66ULL, "art\\vfx\\combat\\vfx_gib_corm_rtlb.ani"},
	{0x8D903A079E882B6DULL, "art\\vfx\\combat\\vfx_gib_corm_rtlb.anb"},
	{0x8D9078C64A92153FULL, "plt_lgt210pt_search_rescue.nss"},
	{0x8D907E6AAC2F2665ULL, "l_env_wal_lowtowndocks_military_nt_end.wbd"},
	{0x8D90C69C04467D15ULL, "6107380_m.fxe"},
	{0x8D912213A3F8C941ULL, "lt_0_lowtown_a3_n_c_38_0_ll.mmh"},
	{0x8D913B6FC1309CBCULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv3.gad"},
	{0x8D915A36F3F10FF5ULL, "vxlt_0_lowtown_a2_n_c_gpb_0_0.vlm"},
	{0x8D9169A7C85448C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v2.ani"},
	{0x8D9254BE9DD60D74ULL, "253759719.ogg"},
	{0x8D9294BCA0229D08ULL, "6085995_m.fxe"},
	{0x8D930058222CC3B7ULL, "529784871.ogg"},
	{0x8D93620C8146E8C6ULL, "lit_aoe_03_l2.mao"},
	{0x8D93742D22FDA4CAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.activetoinactive.evt"},
	{0x8D939BC3881068CAULL, "vfx_vfx_p1.vsh.fxo"},
	{0x8D93FFF8A8DA2907ULL, "296691772.ogg"},
	{0x8D94D8C49D26F3D2ULL, "hlac_ht_keep_3732_13x.dds"},
	{0x8D954756BEEE3452ULL, "hlac_lt_0_lowtown_docks_-4_6y.dds"},
	{0x8D95552D87A79938ULL, "l_lgt111ar_rand_sewer_1_nt.bnk"},
	{0x8D95600ED08733B5ULL, "67539049.ogg"},
	{0x8D95C7D7BB00CA45ULL, "vxlt_warehouse_n_13q_0_0.vlm"},
	{0x8D95D7B19100E909ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_boiler_d.dds"},
	{0x8D96841F868E66F3ULL, "isa211_isabela6202194.cl"},
	{0x8D96BC563A1D997CULL, "1054342184.ogg"},
	{0x8D96FAEC597D5BBAULL, "6119173_m.fxe"},
	{0x8D97652DD3D511DBULL, "vxht_keep_5jm_0_0.vlm"},
	{0x8D980E4032B48F15ULL, "69730770.ogg"},
	{0x8D983A6FC136B9BAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv3.ani"},
	{0x8D9866A7C85A6226ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v2.gad"},
	{0x8D98D9BE8A9180B7ULL, "960107958.ogg"},
	{0x8D993A2FA78FB7B9ULL, "vxw_sundermount_28b_0_0.vlm"},
	{0x8D9A532D23038B3DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.activetoinactive.gad"},
	{0x8D9A54E62856F129ULL, "vxlt_hangedman_n_10x_0_0.vlm"},
	{0x8D9A95E015EB1D6DULL, "vxw_encounter2_camp_hu_0_0.vlm"},
	{0x8D9B5AB0B6141E72ULL, "static_lightblend_p0.vsh.fxo.dx11.assembly"},
	{0x8D9B6386BB438914ULL, "hlac_lt_backalley_g_1_7z.dds"},
	{0x8D9B9DDF03244D04ULL, "vxca_caverns2prc_a20_0_0.vlm"},
	{0x8D9C3200BC569991ULL, "ttx_window_arch_a_1.mmh"},
	{0x8D9C4C00BC56C64FULL, "ttx_window_arch_a_1.msh"},
	{0x8D9C63A32601C31BULL, "static_static_p17.psh.fxo.dx11.assembly"},
	{0x8D9C6FF06B2E23CBULL, "6188601_f.fxe"},
	{0x8D9C853C5F728BABULL, "t1_skn_d07.tnt"},
	{0x8D9CD4EA6CAE78CCULL, "vxw_sundermount_2kn_0_0.vlm"},
	{0x8D9D40DD90465307ULL, "vxlt_darktown_65l_0_0.vlm"},
	{0x8D9DBBE19F680236ULL, "vxw_woundedcoast_1ei_0_0.vlm"},
	{0x8D9E36D69CFE3BD7ULL, "752717105.ogg"},
	{0x8D9E48A419DD763BULL, "gen00fl_fenris6110919.cl"},
	{0x8D9E4D0D2E71A8FEULL, "qcr210cr_looter_thug.utc"},
	{0x8D9E635C3C56C005ULL, "zz_mageogre.cut"},
	{0x8D9EBA0A22238C5AULL, "6107286_m.fxe"},
	{0x8D9F098A5D586C69ULL, "6127517_m.fxe"},
	{0x8D9F486898FD30DFULL, "vxlt_undercitydungeon_17v_0_0.vlm"},
	{0x8D9F61A7C8607805ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v2.evt"},
	{0x8DA0201D153D9543ULL, "vxlt_0_lowtown_docks_b5k_0_1.vlm"},
	{0x8DA03E776C5A9C42ULL, "w_sundermount_41_0_lc.mmh"},
	{0x8DA0C2C278F05300ULL, "460095725.ogg"},
	{0x8DA10F8B13757B39ULL, "w_bonepit_16_0.msh"},
	{0x8DA1198B13758C27ULL, "w_bonepit_16_0.mmh"},
	{0x8DA11D8B13759374ULL, "w_bonepit_16_0.mao"},
	{0x8DA197E361CE2EAFULL, "hlac_w_bonepit_458753_0y.dds"},
	{0x8DA19C57A5595E6BULL, "226483592.ogg"},
	{0x8DA1BB9CF9F35199ULL, "one111ip_ggt_crate_1.utp"},
	{0x8DA24116E5CED4CBULL, "6198100_m.fxe"},
	{0x8DA252188C34E7CBULL, "lowtown_male_alienage1_sta6162748.cl"},
	{0x8DA338750D522FB1ULL, "6099506_m.fxe"},
	{0x8DA345DB24DF77DCULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb104a_0n.dds"},
	{0x8DA439B4ACF6E2C5ULL, "vxlt_warehouse_n_18p_0_0.vlm"},
	{0x8DA4422A5D3E4A33ULL, "hlac_w_sundermount_-327685_0z.dds"},
	{0x8DA4575A5AD9171EULL, "vxlt_undercitydungeon_1d3_0_1.vlm"},
	{0x8DA46640704E10CEULL, "follower_banter6154210.cl"},
	{0x8DA479EA3AFD441FULL, "mrl150_merrill6093914.cl"},
	{0x8DA4B5D2B1F08A77ULL, "vxdr_slavers_14r_0_0.vlm"},
	{0x8DA4F943BDB026A1ULL, "mcr370_negotiation6080261.cl"},
	{0x8DA523079E9A6060ULL, "art\\vfx\\combat\\vfx_gib_corm_rtlb.gad"},
	{0x8DA550BA35BB7A1BULL, "combat_int_test.arl"},
	{0x8DA56BC64AA45B50ULL, "plt_lgt210pt_search_rescue.plo"},
	{0x8DA5CF29A398976DULL, "6032138_m.fxe"},
	{0x8DA5E5C9B00865C3ULL, "vxca_caverns1_j4h_0_0.vlm"},
	{0x8DA60FD939BC38DBULL, "sum_conj_02.mao"},
	{0x8DA65A23967882E4ULL, "1011477535.ogg"},
	{0x8DA6A1B0ABEF8795ULL, "llac_lt_hangedman_-1_26.dds"},
	{0x8DA6A9615BDEF085ULL, "870081339.ogg"},
	{0x8DA6AEFD86E80D4BULL, "6056731_f.fxe"},
	{0x8DA6B5455A8144B2ULL, "and100_anders6095893.cl"},
	{0x8DA6D1DC86762C13ULL, "zz_followers_banter6157405.cl"},
	{0x8DA73055ADE18D51ULL, "vxlt_undercityserialkiller_1s5_0_0.vlm"},
	{0x8DA7502102ABFCFEULL, "698304518.ogg"},
	{0x8DA777B14E459D09ULL, "dae300_cullen6140654.cl"},
	{0x8DA7A860BA3E7B9CULL, "gen00fl_dog6165711.cl"},
	{0x8DA7C3882DAC3D22ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_basic.phy"},
	{0x8DA896216C9834CEULL, "vxlt_backalley_g_1xc_0_0.vlm"},
	{0x8DA907BD07E135EDULL, "vxw_encounter2_camp_ct_0_0.vlm"},
	{0x8DA90EE2A4C96070ULL, "hlac_w_bonepitmining_a2_262145_0z.dds"},
	{0x8DA911DA585DF070ULL, "mag170_recruit016094521.cl"},
	{0x8DA99AC38CD88EA9ULL, "92985734.ogg"},
	{0x8DA9C30BB22F2D1BULL, "6111359_f.fxe"},
	{0x8DA9DB3ED48359CFULL, "295622434.ogg"},
	{0x8DAA773B05B2FACDULL, "6129245_m.fxe"},
	{0x8DAB46C75EA4914CULL, "vxw_sundermount_2lo_0_0.vlm"},
	{0x8DAB6642732DCCD0ULL, "vxlt_gallowscourtyard_n_63p_0_0.vlm"},
	{0x8DABB9FFC57D2A5DULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure_6.pwk"},
	{0x8DABCB9D819B9047ULL, "dae000_madam_lusine6151705.cl"},
	{0x8DAC06A956599079ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hat_medc_0n.dds"},
	{0x8DAC387DAA6B59BCULL, "vxdr_primevaltomb_8vn_0_0.vlm"},
	{0x8DAC6C217543F75AULL, "hlac_w_encounter3_-262140_0y.dds"},
	{0x8DAD0E4ABF0B8C76ULL, "dae000cr_epi_inquisition.utc"},
	{0x8DAE046E025F9606ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.cess.ani"},
	{0x8DAE842B3EC58FA0ULL, "55361307.ogg"},
	{0x8DAEAEE2FA41ACB2ULL, "817005306.ogg"},
	{0x8DAEDC5A612BD4CBULL, "dae300_bran6143514.cl"},
	{0x8DAFBEF196650FA5ULL, "plt_cod_aow_follower_bases.plo"},
	{0x8DB01346E9BE4135ULL, "6121563_f.fxe"},
	{0x8DB080C78AD98BF6ULL, "gen00fl_anders6125518.cl"},
	{0x8DB0E857CAC02BFFULL, "plt_fen000pt_endstage.nss"},
	{0x8DB1064ACD5D07D0ULL, "705238511.ogg"},
	{0x8DB12A0F71420AC8ULL, "and341tr_spider_ambush2.utt"},
	{0x8DB14C300BBCE605ULL, "6139405_f.fxe"},
	{0x8DB1D610DBBFADAAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_rotateback.gad"},
	{0x8DB1DFAE9A132D7EULL, "vxlt_0_lowtown_n_euv_0_0.vlm"},
	{0x8DB2126E3663C9CDULL, "vxht_0_hightown_a3_byl_0_0.vlm"},
	{0x8DB2287418DFCB02ULL, "hlac_lt_undercitydungeon_-1_38x.dds"},
	{0x8DB2E51F4267CAD7ULL, "375904461.ogg"},
	{0x8DB388F44F679230ULL, "qun160_party_comment_26061242.cl"},
	{0x8DB391E654EED21FULL, "hlac_lt_0_lowtown_-131072_3z.dds"},
	{0x8DB3A1A64D688CB7ULL, "hlac_lt_undercityserialkiller_-1_41x.dds"},
	{0x8DB3AD39786016FBULL, "zz_approval_debug6003512.cl"},
	{0x8DB46D9526AB03B2ULL, "hlac_ht_mansion2b_818_53z.dds"},
	{0x8DB485ECEB7DBBB5ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_roba.phy"},
	{0x8DB49AD7B7369279ULL, "vxht_bartrand_n_6qy_0_0.vlm"},
	{0x8DB4A9728633DCFBULL, "6118391_m.fxe"},
	{0x8DB520DA78DE25CBULL, "6053431_f.fxe"},
	{0x8DB54DBCA109486CULL, "hlac_lt_0_lowtown_a2_1_9y.dds"},
	{0x8DB58D8B0DDA9D9FULL, "vxdr_primevalentrance_177_0_0.vlm"},
	{0x8DB5DD4DEDF51500ULL, "llac_lt_gallowstemplar_f_7198_0.dds"},
	{0x8DB61A17B0082A68ULL, "533940718.ogg"},
	{0x8DB62A3A0CA0A840ULL, "6094874_m.fxe"},
	{0x8DB63EE41039A82BULL, "6116238_m.fxe"},
	{0x8DB658BF141BECCCULL, "6111070_m.fxe"},
	{0x8DB6F70CF576139DULL, "763596865.ogg"},
	{0x8DB7DF8A06A81934ULL, "mag100_cullen6082286.cl"},
	{0x8DB802EF8C57C0A8ULL, "vxlt_0_lowtown_a2_n_c_gc6_0_0.vlm"},
	{0x8DB807D69CC6C1C3ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.anb"},
	{0x8DB8199E4DFF5765ULL, "6116782_m.fxe"},
	{0x8DB82F6660527C9AULL, "vxht_brothel_n_326_0_0.vlm"},
	{0x8DB8B910DBC59B41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_rotateback.evt"},
	{0x8DB8D504F4B9E286ULL, "gen00fl_varric6093600.cl"},
	{0x8DB917F6B5E5C9E5ULL, "mrl150_hunter.stg"},
	{0x8DB91B69F46CC14DULL, "gen00fl_varric6076569.cl"},
	{0x8DB95D0C04A416C0ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_109a_0t.dds"},
	{0x8DB9610982D8CE4AULL, "6208439_m.fxe"},
	{0x8DB9768CF554314FULL, "vxlt_darktown_2ux_0_0.vlm"},
	{0x8DB9B8A4509AA9CCULL, "vxw_sundermount_2ql_0_0.vlm"},
	{0x8DB9E07E9D292166ULL, "vxw_bonepit_1x0_0_1.vlm"},
	{0x8DB9E59EABA3796DULL, "hlac_ca_caverns1_24427_4y.dds"},
	{0x8DBA0902B29B0081ULL, "vxw_bonepit_29v_0_0.vlm"},
	{0x8DBA3284FB0FA619ULL, "143406900.ogg"},
	{0x8DBA3C341B0D2675ULL, "plt_mag200pt_night_terrors.plo"},
	{0x8DBA9F9B83543336ULL, "vxw_woundedcoast_1kk_0_0.vlm"},
	{0x8DBAF01A6512DE7CULL, "llac_ht_0_hightown_1_0.dds"},
	{0x8DBB22F00C3A77B2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgb.phy"},
	{0x8DBB4C9FCDABE556ULL, "gen00fl_merrill6128973.cl"},
	{0x8DBBD9B8F5D11FF4ULL, "6149178_m.fxe"},
	{0x8DBBDE7A9E2437F2ULL, "6064989_m.fxe"},
	{0x8DBBEC6E026BAB77ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.cess.evt"},
	{0x8DBC3F4002ECC2D2ULL, "6033141_m.fxe"},
	{0x8DBC964030B7F71CULL, "vxht_mansion_n_6n2_0_0.vlm"},
	{0x8DBCCDE4C7021DCBULL, "static_static_p57.psh.fxo.dx11"},
	{0x8DBCFA7258788666ULL, "qun110_petrice6037557.cl"},
	{0x8DBD298EE3554778ULL, "mer_merchant_events.ncs"},
	{0x8DBD420C949728C1ULL, "294688424.ogg"},
	{0x8DBDE2983ACE18F9ULL, "ser120pt_sacricfice.ncs"},
	{0x8DBE484E59E49399ULL, "lowtown_male_lowlife1_sta6164997.cl"},
	{0x8DBE4F57F890168FULL, "6170047_m.fxe"},
	{0x8DBE87A8F4DB39FAULL, "custics002.dds"},
	{0x8DBEB1359157BA09ULL, "vxlt_0_lowtown_hb2_0_0.vlm"},
	{0x8DBEC2F970621D82ULL, "662004440.ogg"},
	{0x8DBF72B0EA350200ULL, "ht_keep_3736.anb"},
	{0x8DC023A1F310983AULL, "gen00fl_sebastian6174935.cl"},
	{0x8DC028C600D908CEULL, "467815153.ogg"},
	{0x8DC0518B8C0945FEULL, "vxlt_0_lowtown_n_enw_0_0.vlm"},
	{0x8DC06221D6AF1FAEULL, "vxlt_gallowscourtyard_n_3bp_0_0.vlm"},
	{0x8DC0CBA45C7C79C7ULL, "vxlt_gallowsprison_n_5wn_0_0.vlm"},
	{0x8DC1BE2704E3D5C0ULL, "vxlt_darktown_5kb_0_0.vlm"},
	{0x8DC1D3CDDC7C283BULL, "hlac_lt_0_lowtown_n_-1_3z.dds"},
	{0x8DC1ED6E02701800ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.cess.gad"},
	{0x8DC23AACA1CC6C06ULL, "6077880_m.fxe"},
	{0x8DC31B4F7829F57BULL, "6117291_m.fxe"},
	{0x8DC3F15DC0F3BC10ULL, "vxlt_0_lowtown_ep5_0_1.vlm"},
	{0x8DC46065C8F360ECULL, "lowtown_fem_alien_elf_2_mob6166345.cl"},
	{0x8DC4DB57CAD0BE10ULL, "plt_fen000pt_endstage.plo"},
	{0x8DC4E0522C21FA2BULL, "carpetold_sa.dds"},
	{0x8DC504E24C43CE38ULL, "mer340cr_high_dragon.ncs"},
	{0x8DC51B9568A50598ULL, "6128830_m.fxe"},
	{0x8DC5B110DBD01790ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_rotateback.ani"},
	{0x8DC5BA9C0B0DB0B0ULL, "gen00fl_anders6114479.cl"},
	{0x8DC5E4D69CD2C3ECULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.mmh"},
	{0x8DC66B6CB03CB394ULL, "isa321_velasco6117624.cl"},
	{0x8DC6C692C41EA40CULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobd.phy"},
	{0x8DC731B480DB68E9ULL, "llac_lt_gallowstemplar_f_2383_4.dds"},
	{0x8DC7A447334A7407ULL, "dae000_keep_guard_1_sta6163688.cl"},
	{0x8DC828D5675054D8ULL, "ser191_emeric6108130.cl"},
	{0x8DC82A814290C24CULL, "vxw_sundermount_2zm_0_0.vlm"},
	{0x8DC910F80BE7E378ULL, "5739677.ogg"},
	{0x8DC9CA573ACDDD2AULL, "follower_banter6165485.cl"},
	{0x8DCA089F9441C3E1ULL, "156486960.ogg"},
	{0x8DCA4FBBCD894FCCULL, "one350ip_zevran_note.ncs"},
	{0x8DCA9F0822939817ULL, "133678430.ogg"},
	{0x8DCADA2AB8F9E79EULL, "mag211_marethari6025367.cl"},
	{0x8DCB7DB7A9F278F4ULL, "738177267.ogg"},
	{0x8DCB89545EB1AF1BULL, "6087481_f.fxe"},
	{0x8DCC0B4847A430DEULL, "vxht_0_hightown_a2_7lx_0_0.vlm"},
	{0x8DCC6A5FA0837B39ULL, "274885557.ogg"},
	{0x8DCCCB4858211EFBULL, "6135673_m.fxe"},
	{0x8DCD8A15AFB49C78ULL, "hlac_w_encounter3_-393216_0x.dds"},
	{0x8DCD9E882DCCCDD6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_basic.mao"},
	{0x8DCDAA882DCCE275ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_basic.mmh"},
	{0x8DCDAC882DCCE5EBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_basic.msh"},
	{0x8DCDD3F345ED6E35ULL, "6028850_m.fxe"},
	{0x8DCE14C97EC367F1ULL, "hlac_ca_caverns1_24423_9z.dds"},
	{0x8DCE2D909977BD56ULL, "vxht_mansion_n_6lb_0_0.vlm"},
	{0x8DCE30F57403548DULL, "vxlt_0_lowtown_docks_a3_n_bud_0_0.vlm"},
	{0x8DCE96E5F035A543ULL, "hlac_ht_mansion2_818_58x.dds"},
	{0x8DCF0CE77C86CE43ULL, "297741634.ogg"},
	{0x8DCF3CDF177406D7ULL, "one280_merin6104369.cl"},
	{0x8DCF63793D629925ULL, "vxw_bonepitmining_a2_1wx_0_1.vlm"},
	{0x8DCF71312828148FULL, "clf260_harley6069649.cl"},
	{0x8DD03CC77A777EBDULL, "vxca_cavern1_haunted_jfy_0_0.vlm"},
	{0x8DD04A2215B40834ULL, "6144189_f.fxe"},
	{0x8DD0586AAC64C90FULL, "l_env_wal_lowtowndocks_military_nt_end.bnk"},
	{0x8DD06AEE5B3E54C1ULL, "6107000_f.fxe"},
	{0x8DD086FA99251211ULL, "470515315.ogg"},
	{0x8DD16A723C36FCCDULL, "vxlt_0_lowtown_a2_n_hgr_0_0.vlm"},
	{0x8DD28092280C2145ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\dm_exp_idle.anb"},
	{0x8DD2D634938268E5ULL, "gen_im_wep_rog_arc_act1_05.uti"},
	{0x8DD4198B13A0E584ULL, "w_bonepit_16_0.rml"},
	{0x8DD44378AB86EF18ULL, "6100242_f.fxe"},
	{0x8DD44CB3E33BF6F8ULL, "vxht_keep_7ai_0_0.vlm"},
	{0x8DD474D0366E7D13ULL, "457777589.ogg"},
	{0x8DD485054664B670ULL, "one300dg_lia6064528.cl"},
	{0x8DD48AFBFFD2AA1BULL, "6107669_m.fxe"},
	{0x8DD48E53E1FE8A3FULL, "qun110cr_arishok_guard_nt.utc"},
	{0x8DD4951B4B16B7E9ULL, "605576084.ogg"},
	{0x8DD55E0EAE453C1EULL, "6111206_f.fxe"},
	{0x8DD62C61A18DF0A8ULL, "gen00fl_anders6119847.cl"},
	{0x8DD64843B653686DULL, "vxlt_0_lowtown_docks_cv0_0_0.vlm"},
	{0x8DD67FDB1DD1A2FBULL, "gen00fl_bethany6098062.cl"},
	{0x8DD6A0BFFEAADA28ULL, "amb_scene_domesticabuse.bnk"},
	{0x8DD7835567406436ULL, "vxw_woundedcoast_1qm_0_0.vlm"},
	{0x8DD7E3BA59FBCFD3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.ani"},
	{0x8DD88033E2F6745FULL, "mrl150_merrill6170442.cl"},
	{0x8DD9BB08CCC8F6CDULL, "vxht_mansion2b_4d_0_0.vlm"},
	{0x8DD9F7181C518417ULL, "6045369_m.fxe"},
	{0x8DDB348CBC827D99ULL, "vxlt_0_lowtown_df2_0_0.vlm"},
	{0x8DDB45FE77E11915ULL, "vxht_keep_61l_0_1.vlm"},
	{0x8DDBD8C81D6B582FULL, "and100_anders6114230.cl"},
	{0x8DDC18119217DBF6ULL, "797029277.ogg"},
	{0x8DDC2A9D3A534E7CULL, "vxht_mansion_n_6b6_0_0.vlm"},
	{0x8DDC844BBB4DB379ULL, "hlac_lt_0_lowtown_n_-65539_0z.dds"},
	{0x8DDCD7E5C5E0F48CULL, "w_bonepitmining_a2_178_0.msh"},
	{0x8DDCDDE5C5E0FF7EULL, "w_bonepitmining_a2_178_0.mmh"},
	{0x8DDCE9E5C5E11315ULL, "w_bonepitmining_a2_178_0.mao"},
	{0x8DDD1BE24C588722ULL, "mer340cr_high_dragon.utc"},
	{0x8DDD4D8E85F831C9ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_fdr_arm.mao"},
	{0x8DDD598E85F84662ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_fdr_arm.mmh"},
	{0x8DDD5B8E85F84918ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_fdr_arm.msh"},
	{0x8DDE1B095889846EULL, "6083199_m.fxe"},
	{0x8DDE73C66A99A3D6ULL, "wm_ico_gallows_roll.dds"},
	{0x8DDEACDB393C9864ULL, "826641603.ogg"},
	{0x8DDFACAB237DC230ULL, "vxca_cave_2gg_0_1.vlm"},
	{0x8DDFD99B42BDB22DULL, "160641766.ogg"},
	{0x8DE0106ACD149089ULL, "hlac_lt_0_lowtown_docks_n_65533_0z.dds"},
	{0x8DE05923C6F7B1C4ULL, "6037011_m.fxe"},
	{0x8DE0D957F41A75F2ULL, "hlac_dr_slavers_3207_6x.dds"},
	{0x8DE25B3E38526150ULL, "lt_backalley_g_3_0_ll.mmh"},
	{0x8DE28C254DB9D4D8ULL, "vxca_cavern1_haunted_j51_0_0.vlm"},
	{0x8DE29EA3D7EC3145ULL, "vxca_cavern1_haunted_gkf_0_1.vlm"},
	{0x8DE2B93A63D7524AULL, "zz_followers_banter6158601.cl"},
	{0x8DE2F19B53415466ULL, "hlac_lt_0_lowtown_n_0_0x.dds"},
	{0x8DE2FCD86EF4A3F4ULL, "vxdr_0_deeproads_cvc_0_0.vlm"},
	{0x8DE335BA2A2929ABULL, "vxdr_primevalentrance_2q4_0_2.vlm"},
	{0x8DE453D1F2C48FE1ULL, "691693168.ogg"},
	{0x8DE45CB17E470B44ULL, "mag291_demon_marethari6029022.cl"},
	{0x8DE56B9823B03A5BULL, "216021496.ogg"},
	{0x8DE66CBD39D72A7AULL, "705898315.ogg"},
	{0x8DE6F932E0ABEB5DULL, "vxlt_undercityserialkiller_1ic_0_0.vlm"},
	{0x8DE803145C0C019EULL, "vxht_0_hightown_a3_90c_0_0.vlm"},
	{0x8DE82E4CEC143EFCULL, "zz_qun_debug6036828.cl"},
	{0x8DE843BB8212ECAAULL, "6085870_m.fxe"},
	{0x8DE859045FF6448EULL, "373014275.ogg"},
	{0x8DE8EA2427533990ULL, "w_encounter3_208_0.mao"},
	{0x8DE8EE24275340A3ULL, "w_encounter3_208_0.mmh"},
	{0x8DE8EF022B9061DEULL, "vxht_0_hightown_a2_7fz_0_0.vlm"},
	{0x8DE8FC24275358FDULL, "w_encounter3_208_0.msh"},
	{0x8DE93D972DB43935ULL, "follower_banter6154189.cl"},
	{0x8DE9E0BF3FF16522ULL, "6162720_m.fxe"},
	{0x8DEA31853D42CE36ULL, "50554258.ogg"},
	{0x8DEA3F84CE14D232ULL, "450692655.ogg"},
	{0x8DEA7CD37DD5D40CULL, "374803249.ogg"},
	{0x8DEA932D9E97B39DULL, "lt_0_lowtown_n_35_0.rml"},
	{0x8DEAC818F521220EULL, "848598743.ogg"},
	{0x8DEB2D373BA653C8ULL, "follower_barks6153232.cl"},
	{0x8DEBBABA5A0C337DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.gad"},
	{0x8DEBBBE1C328C416ULL, "122583351.ogg"},
	{0x8DEBBC67C6D7EAB5ULL, "6128062_m.fxe"},
	{0x8DEC2FF48964FC7DULL, "gen_im_trs_gen_coinbag_01.uti"},
	{0x8DECEDC21E77044CULL, "6052069_m.fxe"},
	{0x8DED1173DBBB7E83ULL, "gen00fl_merrill6121151.cl"},
	{0x8DED5CF7D49E1C77ULL, "vxht_0_hightown_n_arz_0_0.vlm"},
	{0x8DED6617B581BE8EULL, "llac_lt_0_lowtown_a3_n_c_-1_3.dds"},
	{0x8DED664E96F8A1FFULL, "133219149.ogg"},
	{0x8DED905C01208148ULL, "511403374.ogg"},
	{0x8DEDBB3EADD0764CULL, "hlac_w_sundermount_196605_0z.dds"},
	{0x8DEE873706A8CF0DULL, "zz_followers_banter6159522.cl"},
	{0x8DEEF8E11EBEF501ULL, "111372235.ogg"},
	{0x8DEF74ACBFF21367ULL, "6135765_m.fxe"},
	{0x8DEFB6194C177B0EULL, "6176116_m.fxe"},
	{0x8DEFE1DAB276C5BAULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_br.gad"},
	{0x8DF03A5AE37515FBULL, "186898698.ogg"},
	{0x8DF10E3A0326D892ULL, "6038672_m.fxe"},
	{0x8DF12EAC213695D0ULL, "follower_banter6161300.cl"},
	{0x8DF1392CDCB23B9BULL, "506876427.ogg"},
	{0x8DF14710FD5C7FD4ULL, "6081981_f.fxe"},
	{0x8DF15A0C7754E1FBULL, "367532007.ogg"},
	{0x8DF196F7D676024EULL, "fonts_pl-pl.gfx"},
	{0x8DF1A204C69CE1A5ULL, "vxw_bonepit_20u_0_0.vlm"},
	{0x8DF1E11EFBCBEE02ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_0z.dds"},
	{0x8DF2500BA717B4BDULL, "6092148_f.fxe"},
	{0x8DF2CF8629D47438ULL, "hlac_lt_0_lowtown_a3_n_c_131071_0z.dds"},
	{0x8DF301E3B4A93653ULL, "vxw_woundedcoast_158_0_0.vlm"},
	{0x8DF330D5CDA01DF1ULL, "dae100_bran6142993.cl"},
	{0x8DF48EA73A9A962AULL, "qcr000_arishok6059965.cl"},
	{0x8DF4904A2EC4CC2FULL, "690900975.ogg"},
	{0x8DF4A4245A317BE5ULL, "gen00fl_anders6111321.cl"},
	{0x8DF4FA790061F7F1ULL, "hlac_lt_gallowstemplar_2384_4z.dds"},
	{0x8DF506D3CAA8F4C7ULL, "l_mcr372ar_gallows_templar_nt.wbd"},
	{0x8DF52008BCB103D0ULL, "447385838.ogg"},
	{0x8DF56922E3F82799ULL, "vfx_vfx_p0.psh.fxo"},
	{0x8DF68CB32E1201C2ULL, "hlac_dr_primevaltomb_659_12y.dds"},
	{0x8DF6E4DAB27CE9D1ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_br.evt"},
	{0x8DF71F66A19FC201ULL, "hlac_ca_cavern1_haunted_24421_1y.dds"},
	{0x8DF72F341B40BE52ULL, "plt_mag200pt_night_terrors.nss"},
	{0x8DF73B7F89EAA058ULL, "gen00fl_dog6167674.cl"},
	{0x8DF758F87F1D3054ULL, "vxlt_foundry_n_xh_0_0.vlm"},
	{0x8DF760E0D7A10475ULL, "vxht_chantry_260_0_0.vlm"},
	{0x8DF88CB8D23F8289ULL, "639312449.ogg"},
	{0x8DF8E6A0DD58058AULL, "vxdr_primevaltomb_7st_0_0.vlm"},
	{0x8DF97CAA0D8A150AULL, "gen00fl_sebastian6174878.cl"},
	{0x8DF9851C79AD3CB8ULL, "6102162_m.fxe"},
	{0x8DF9868C49E59E0DULL, "vxlt_0_lowtown_docks_a3_n_bba_0_0.vlm"},
	{0x8DFA09C46E73BC55ULL, "348770242.ogg"},
	{0x8DFA5D6722AD41A6ULL, "953686551.ogg"},
	{0x8DFA62F2B3C7A7C0ULL, "263981375.ogg"},
	{0x8DFB17DD764BE959ULL, "ave200_donnic6112175.cl"},
	{0x8DFB2D52EEE70102ULL, "vxdr_primevaltomb_3sa_0_1.vlm"},
	{0x8DFB98C5D365B243ULL, "17335192.ogg"},
	{0x8DFC8EDC9DCFF732ULL, "412318159.ogg"},
	{0x8DFC92BFFECB91D2ULL, "amb_scene_domesticabuse.wbd"},
	{0x8DFC98ECEBBB91D0ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_roba.msh"},
	{0x8DFCA5E5D2333FB1ULL, "6119804_m.fxe"},
	{0x8DFCA6ECEBBBA9EAULL, "art\\characters\\playercharacter\\humanfemale\\hoo_roba.mmh"},
	{0x8DFCAAECEBBBB021ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_roba.mao"},
	{0x8DFCF5EC03A48798ULL, "vxlt_hangedman_26o_0_0.vlm"},
	{0x8DFD7F089C970BAFULL, "mag370_meredith6046892.cl"},
	{0x8DFD8BF1F66F1F5CULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml107a_0n.dds"},
	{0x8DFE1552930FB2B3ULL, "mcr370_negotiation6080241.cl"},
	{0x8DFEE302DC60AEF5ULL, "zz_qun_debug.cnv"},
	{0x8DFF105699E2762EULL, "hla_ht_0_hightown_n_65534_0x.dds"},
	{0x8DFF478951C28A2EULL, "zz_facefx_test6088485.cl"},
	{0x8DFFC679A0B291D8ULL, "623936777.ogg"},
	{0x8E0051B0EA6C6128ULL, "ht_keep_3736.rml"},
	{0x8E007A9C4B478A6DULL, "6059267_m.fxe"},
	{0x8E011D06695130BFULL, "lt_gallowscourtyard_n_16_0.rml"},
	{0x8E01A3FEE0F78910ULL, "103740511.ogg"},
	{0x8E023CD445BDD0EAULL, "hlac_lt_backalley_g_65536_3z.dds"},
	{0x8E02AB8A5328D38AULL, "31376699.ogg"},
	{0x8E0316095056F7F3ULL, "174413919.ogg"},
	{0x8E0317AC59F98E78ULL, "vxht_bartrand_n_34r_0_0.vlm"},
	{0x8E033FAA437EB1CFULL, "vxlt_undercityencounter_1fd_0_0.vlm"},
	{0x8E0351D68A6C1565ULL, "llac_w_bonepitdestroyed_a3_0_3.dds"},
	{0x8E03E33B5AA9160EULL, "zz_marketing.cut"},
	{0x8E03FB8B6A7F54C1ULL, "vxca_caverns2_937_0_0.vlm"},
	{0x8E03FCDAB2879CA0ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_br.ani"},
	{0x8E045DBE1E6C4323ULL, "vxlt_gallowstemplar_n_3sd_0_0.vlm"},
	{0x8E0554879B9F4E6CULL, "6147694_f.fxe"},
	{0x8E05A3C04E984765ULL, "dae000_flemeth6002115.cl"},
	{0x8E05BF052CDF5B10ULL, "hlac_lt_0_lowtown_docks_n_-3_1x.dds"},
	{0x8E05CAD5711348D4ULL, "vxlt_foundry_n_si_0_0.vlm"},
	{0x8E05D36A6C7F1579ULL, "zz_followers_banter6158073.cl"},
	{0x8E05EE529A5A6948ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyo_0n.dds"},
	{0x8E0649180F10C03FULL, "6164583_m.fxe"},
	{0x8E0679925D34FD85ULL, "6165605_m.fxe"},
	{0x8E06AAEC327B929BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd02.msh"},
	{0x8E06B4ACE80FBD90ULL, "6112964_f.fxe"},
	{0x8E06C8EC327BC585ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd02.mmh"},
	{0x8E0751DA4AD37124ULL, "bl_ar_dot.mao"},
	{0x8E078B717D5D5976ULL, "hlac_w_woundedcoast_4_0y.dds"},
	{0x8E079D92C455F5D7ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobd.mmh"},
	{0x8E07A192C455FCE4ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobd.mao"},
	{0x8E07B392C4561B49ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobd.msh"},
	{0x8E07B479E09A2A89ULL, "736842520.ogg"},
	{0x8E07D27466C26571ULL, "6031306_m.fxe"},
	{0x8E08697FFE6F3F15ULL, "hlac_lt_0_lowtown_-131072_8y.dds"},
	{0x8E088D25B87B70F6ULL, "vxw_encounter2_rq_0_1.vlm"},
	{0x8E08E29025B75112ULL, "424957744.ogg"},
	{0x8E08E602DC690852ULL, "zz_qun_debug.ncs"},
	{0x8E0922DF1D2CFC54ULL, "vxlt_gallowstemplar_f_5wq_0_0.vlm"},
	{0x8E093A46C75EE1F4ULL, "one220_yevhen6103370.cl"},
	{0x8E09419931F94A47ULL, "vxw_encounter2_camp_2u_0_0.vlm"},
	{0x8E0966472D4FFA4FULL, "vxlt_foundry_n_12s_0_1.vlm"},
	{0x8E0999C928F5AFBBULL, "966553728.ogg"},
	{0x8E09CEEFA8C27078ULL, "6163365_m.fxe"},
	{0x8E09E191A79C161AULL, "780812604.ogg"},
	{0x8E0AA562CDA922DEULL, "vxlt_0_lowtown_n_cml_0_0.vlm"},
	{0x8E0AD46B1706FEBDULL, "vxw_bonepit_106_0_0.vlm"},
	{0x8E0B1D950DE61063ULL, "vxw_bonepit_1px_0_0.vlm"},
	{0x8E0B4A580E3F9C6FULL, "gen00fl_dog6167019.cl"},
	{0x8E0BE2BD774E4219ULL, "gen00fl_varric6104989.cl"},
	{0x8E0BEB1028A3EC48ULL, "dae300_cullen6140583.cl"},
	{0x8E0BF7D422A3F372ULL, "264485427.ogg"},
	{0x8E0C25980F44FBC8ULL, "hlac_lt_backalley_n_0_2y.dds"},
	{0x8E0CCAEB05704368ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medc_0s.dds"},
	{0x8E0CD935EED2E01FULL, "vxht_0_hightown_b8p_0_0.vlm"},
	{0x8E0D190F71904CF5ULL, "and341tr_spider_ambush2.ncs"},
	{0x8E0D6520D412FA4AULL, "vxlt_0_lowtown_a2_n_c_fhx_0_0.vlm"},
	{0x8E0D879FC7D5EB2CULL, "6164463_m.fxe"},
	{0x8E0E08CD94B03AB2ULL, "6122113_m.fxe"},
	{0x8E0E298FAA41DF85ULL, "vxlt_0_lowtown_docks_n_br2_0_0.vlm"},
	{0x8E0E2E8F00E33CA8ULL, "lt_0_lowtown_docks_n_17_0_ll.mmh"},
	{0x8E0E9CD76C57DEC9ULL, "vfx_vfx_p6.psh.fxo.dx11"},
	{0x8E0EA681DBDFE087ULL, "vxht_0_hightown_a3_aob_0_1.vlm"},
	{0x8E0F02F946F707D2ULL, "llac_w_encounter3_-262141_0.dds"},
	{0x8E0F65DAA17FAE8CULL, "60029669.ogg"},
	{0x8E0F757FE26B2A27ULL, "vxlt_0_lowtown_exg_0_0.vlm"},
	{0x8E0F7D3CA6C395F5ULL, "gen_im_trs_bst_pelt_02.uti"},
	{0x8E0F83D80B386B61ULL, "6139653_m.fxe"},
	{0x8E0F955B6E227B82ULL, "vxlt_undercityserialkiller_176_0_0.vlm"},
	{0x8E0FC9062C37B3F4ULL, "vxlt_0_lowtown_a2_n_by3_0_0.vlm"},
	{0x8E1018C62AEFFCB6ULL, "lt_0_lowtown_docks_8_0_lc.mmh"},
	{0x8E103E4F7C4134ABULL, "6107726_carver_brother.cut"},
	{0x8E103FBC9DB7DA61ULL, "329559057.ogg"},
	{0x8E107DD315311610ULL, "567746577.ogg"},
	{0x8E10A3E394717A40ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2_0.pwk"},
	{0x8E10AA545B25199AULL, "632259788.ogg"},
	{0x8E10BE47CB0F9964ULL, "174762845.ogg"},
	{0x8E10FF221F2EA911ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.dura.ani"},
	{0x8E1189894BEFA6F8ULL, "vxht_bartrand_n_33s_0_0.vlm"},
	{0x8E119555C63336CBULL, "132334124.ogg"},
	{0x8E11F2EABF029C92ULL, "ave100_ambush_16106199.cl"},
	{0x8E123956C2DA52FEULL, "character_character_p180.psh.fxo.dx11.assembly"},
	{0x8E126D685C756D41ULL, "vxca_caverns2_946_0_0.vlm"},
	{0x8E128862D0EA1005ULL, "one100dg_vanard.bnk"},
	{0x8E133E56EBA83627ULL, "217859829.ogg"},
	{0x8E136FDF828142B8ULL, "hlac_ca_caverns1_24424_10x.dds"},
	{0x8E137F603B8186D5ULL, "t3_hey_bl3.tnt"},
	{0x8E13A41CAE79B725ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_act_c.anb"},
	{0x8E13E7197412BD55ULL, "6025152_m.fxe"},
	{0x8E13EB94FF71313BULL, "1048319548.ogg"},
	{0x8E13F889567848A7ULL, "687346614.ogg"},
	{0x8E1422F206FD6882ULL, "6140180_m.fxe"},
	{0x8E1521AA62A8947AULL, "101552262.ogg"},
	{0x8E152AA057C1F134ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredacti.gad"},
	{0x8E152BC032F54706ULL, "228441988.ogg"},
	{0x8E156A62D0EC133DULL, "one100dg_vanard.cnv"},
	{0x8E157907B434C620ULL, "vxlt_ruinedhouse_n_h8_0_0.vlm"},
	{0x8E15E42C0E54EE22ULL, "vxht_0_hightown_n_c72_0_0.vlm"},
	{0x8E1609A565F8C045ULL, "6207460_m.fxe"},
	{0x8E1646FFF30D44C4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.up.ani"},
	{0x8E16BAE666332835ULL, "vxht_brothel_n_5hy_0_0.vlm"},
	{0x8E16EF2B0B67CD3CULL, "629017539.ogg"},
	{0x8E1706F8AFF76AAEULL, "vxht_mansion2b_3d8_0_0.vlm"},
	{0x8E180A221F34DAB3ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.dura.gad"},
	{0x8E1830F22D16296BULL, "410747743.ogg"},
	{0x8E186EBBCDCBA5CDULL, "one350ip_zevran_note.utp"},
	{0x8E189FCDFFF7A7F7ULL, "mer300cs_high_dragon_reveal.cut"},
	{0x8E18A0295DB6F099ULL, "vxw_bonepit_1q5_0_0.vlm"},
	{0x8E1933BBCB17E6D6ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_mag_0n.dds"},
	{0x8E194838D5F4A5BAULL, "zz_vault_debug6177783.cl"},
	{0x8E198401F43AE063ULL, "6207703_m.fxe"},
	{0x8E19934D1748C28EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxd.mmh"},
	{0x8E199EB21574A649ULL, "6084910_f.fxe"},
	{0x8E199F4D1748D725ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxd.mao"},
	{0x8E19AD4D1748EF7CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxd.msh"},
	{0x8E1A4C02B94E1451ULL, "6102919_carver_titles.cut"},
	{0x8E1A4EABB70F80BCULL, "6176547_m.fxe"},
	{0x8E1A566920E4C708ULL, "mcr310_confused_guard6077064.cl"},
	{0x8E1A60A8B797745FULL, "vxw_encounter2_f9_0_0.vlm"},
	{0x8E1AA25E1ABC388EULL, "vxlt_darktown_4uw_0_0.vlm"},
	{0x8E1AB0716F31F6E5ULL, "vrc220ip_at_mansion.ncs"},
	{0x8E1B39A057C67612ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredacti.ani"},
	{0x8E1B3AF429369439ULL, "vxw_sundermount_2ei_0_0.vlm"},
	{0x8E1B638958BE5988ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edgebase.mmh"},
	{0x8E1B6F8958BE6E2BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edgebase.mao"},
	{0x8E1B818958BE8CB2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edgebase.msh"},
	{0x8E1BEDA22D71FD2BULL, "dae000_gen_turn_in_46178434.cl"},
	{0x8E1CB0BF453A9605ULL, "mcr370_desten6087323.cl"},
	{0x8E1CBDFA7AC0805EULL, "gen00fl_anders6119441.cl"},
	{0x8E1CCBD9A4BBFB74ULL, "hlac_lt_0_lowtown_docks_-196611_2y.dds"},
	{0x8E1D04465E7B204FULL, "6096483_f.fxe"},
	{0x8E1D43FFF3135E26ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.up.gad"},
	{0x8E1D5FAE56D5B594ULL, "mrl150_dalish_guard6089420.cl"},
	{0x8E1D922D9EC30BB5ULL, "lt_0_lowtown_n_35_0.anb"},
	{0x8E1D9245EE82BA65ULL, "lowtown_male_lowlife1_sta6165015.cl"},
	{0x8E1DAED7BB4A9339ULL, "6171811_m.fxe"},
	{0x8E1E083D2506EA7AULL, "6136729_m.fxe"},
	{0x8E1E123AEDD91927ULL, "vxht_mansion_n_3rx_0_0.vlm"},
	{0x8E1E1EC894332E3AULL, "vxht_brothel_n_6de_0_0.vlm"},
	{0x8E1E38487364FEE2ULL, "hlac_lt_0_lowtown_a2_131071_7y.dds"},
	{0x8E1E40D9F2A10147ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magb_0d.dds"},
	{0x8E1E56940D76E5D8ULL, "hlac_w_encounter2_2_0x.dds"},
	{0x8E1E7F6E9D18F17AULL, "6111223_m.fxe"},
	{0x8E1E9B42169C2033ULL, "vxw_bonepitmining_a2_21r_0_0.vlm"},
	{0x8E1F07221F3AF418ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.dura.evt"},
	{0x8E1F0D841FC00CC5ULL, "6070338_m.fxe"},
	{0x8E1F59C2459110FEULL, "hlac_lt_darktown_-1_81y.dds"},
	{0x8E1F6D54FDC41A38ULL, "hlac_w_encounter3_-262145_0x.dds"},
	{0x8E1FC7757E4A46C0ULL, "creeper_grasp.dds"},
	{0x8E1FE4CF6841AB52ULL, "336542890.ogg"},
	{0x8E2033A1E11CCF36ULL, "hlac_w_encounter2_-65534_0z.dds"},
	{0x8E207D8C71054BC7ULL, "vxca_caverns1_j4t_0_0.vlm"},
	{0x8E20A46D96060200ULL, "442303738.ogg"},
	{0x8E2119884ECF5C56ULL, "6072397_m.fxe"},
	{0x8E22112CB384F783ULL, "lt_0_lowtown_42_0_lc.mmh"},
	{0x8E228848121CC8E9ULL, "lt_0_lowtown_n_44_0_ll.mmh"},
	{0x8E22D5546E2A46F8ULL, "bind_clu.mao"},
	{0x8E231EC340749948ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_killdagger_n.dds"},
	{0x8E232645D722F80EULL, "507467083.ogg"},
	{0x8E23946491185BFDULL, "6169285_m.fxe"},
	{0x8E239DFC350592C0ULL, "fen220ip_to_base.utp"},
	{0x8E23C88A7E63386FULL, "6170256_m.fxe"},
	{0x8E23D7163FED99FCULL, "49043086.ogg"},
	{0x8E24108AA55FE5B1ULL, "hlac_w_bonepitdestroyed_a3_262147_0y.dds"},
	{0x8E249D633CB2FDF5ULL, "6055447_m.fxe"},
	{0x8E24A0B258663771ULL, "vxw_sundermount_28z_0_0.vlm"},
	{0x8E256E6B1208D92CULL, "vrc321ip_spooky7_obj_4.utp"},
	{0x8E25AC06C0633CE8ULL, "504980312.ogg"},
	{0x8E261024CC5ECD13ULL, "hlac_ht_keep_n_3732_19z.dds"},
	{0x8E262D64E43FA0A9ULL, "mer320_hubert6089950.cl"},
	{0x8E268A2A7E8ECFE9ULL, "cage_06.mao"},
	{0x8E26E541438C2AB8ULL, "388740473.ogg"},
	{0x8E26E5CB56ACDC1CULL, "866628981.ogg"},
	{0x8E2712064FAD0919ULL, "vxw_bonepit_1z4_0_0.vlm"},
	{0x8E2730C1B56067B7ULL, "hlac_ht_0_hightown_n_-196608_0z.dds"},
	{0x8E2733B7E5455A70ULL, "344026047.ogg"},
	{0x8E2754B0B9E556C8ULL, "one100_party_comments6112614.cl"},
	{0x8E27B7C62EFD6AD9ULL, "vxlt_gallowstemplar_3rv_0_0.vlm"},
	{0x8E27F814659777E6ULL, "vxht_0_hightown_a3_9ut_0_0.vlm"},
	{0x8E27FDE84409FF1EULL, "sky_lt_0_lowtown.mao"},
	{0x8E28131A7B3D80A8ULL, "6172653_m.fxe"},
	{0x8E286042E27244B0ULL, "dae000ip_crowd_ff6_talk5.utp"},
	{0x8E28B5A7826C5097ULL, "vxht_keep_n_687_0_0.vlm"},
	{0x8E28D285A98D8CDFULL, "vxw_encounter2_a8_0_0.vlm"},
	{0x8E295EA7BBECCA33ULL, "606873948.ogg"},
	{0x8E296607C233D129ULL, "635501793.ogg"},
	{0x8E2967B487F101F6ULL, "art\\characters\\creatures\\mubari\\animation\\cwkb.evt"},
	{0x8E29ACD11B2CACB9ULL, "vxw_sundermount_2nh_0_0.vlm"},
	{0x8E29C31E9B5B38D9ULL, "6093913_m.fxe"},
	{0x8E2AD28078073AB7ULL, "vxht_mansion2b_23e_0_0.vlm"},
	{0x8E2AE00991A35753ULL, "vxca_caverns1_fmk_0_0.vlm"},
	{0x8E2B8E5283AF8C8FULL, "vxht_mansion2b_2pp_0_0.vlm"},
	{0x8E2BA1716F40503CULL, "vrc220ip_at_mansion.utp"},
	{0x8E2BF48FDB34C9A3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04b.mmh"},
	{0x8E2C028FDB34E1FDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04b.msh"},
	{0x8E2C487A0D186463ULL, "fen110_fenris6100252.cl"},
	{0x8E2C7A1A16FD02FAULL, "6131629_m.fxe"},
	{0x8E2C8924F3A5776EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rain_of_arrows.evt"},
	{0x8E2C8AC3FBD4FABCULL, "vxw_woundedcoast_jg_0_0.vlm"},
	{0x8E2CF2BCB38368B7ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.cess.gad"},
	{0x8E2CFCDB36F69411ULL, "459456361.ogg"},
	{0x8E2D4A5B1DF6115BULL, "art\\characters\\playercharacter\\humanmale\\hed_viscount.msh"},
	{0x8E2D5537E68209B3ULL, "vxlt_0_lowtown_a2_n_fj1_0_0.vlm"},
	{0x8E2D5C5B1DF62FA6ULL, "art\\characters\\playercharacter\\humanmale\\hed_viscount.mao"},
	{0x8E2D685B1DF64445ULL, "art\\characters\\playercharacter\\humanmale\\hed_viscount.mmh"},
	{0x8E2E038E585D3238ULL, "6061815_f.fxe"},
	{0x8E2E54AF4807706FULL, "388289812.ogg"},
	{0x8E2EEF6962FB6447ULL, "vxca_caverns1_j1w_0_0.vlm"},
	{0x8E2F5EB487F55D81ULL, "art\\characters\\creatures\\mubari\\animation\\cwkb.gad"},
	{0x8E2F6D99E574F53BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16ma.mmh"},
	{0x8E2F734B7270A5EFULL, "6118791_m.fxe"},
	{0x8E2F7999E57509D8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16ma.mao"},
	{0x8E2F8B99E5752825ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16ma.msh"},
	{0x8E2F8EDE4BE2DD74ULL, "dae300_petrice6097901.cl"},
	{0x8E301E85E001CEB4ULL, "vxlt_0_lowtown_docks_a3_n_c1h_0_0.vlm"},
	{0x8E30438D0F705852ULL, "llac_lt_gallowscourtyard_n_0_1.dds"},
	{0x8E315C92FDF44E95ULL, "vxlt_0_lowtown_docks_a3_n_d2o_0_0.vlm"},
	{0x8E325403BD04D146ULL, "6001193_m.fxe"},
	{0x8E328024F3A9D319ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rain_of_arrows.gad"},
	{0x8E32885FDAD939EDULL, "llac_ht_0_hightown_n_131071_0.dds"},
	{0x8E32CC06972B0AEDULL, "101107476.ogg"},
	{0x8E3307BCB387F844ULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.cess.evt"},
	{0x8E33923824769E81ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarflean_2.mao"},
	{0x8E33AC52A4D04DA8ULL, "llac_ht_0_hightown_-1_1.dds"},
	{0x8E343E3C0D043E4FULL, "6170630_m.fxe"},
	{0x8E3460BF52F7F5D2ULL, "vxca_caverns1_fmn_0_0.vlm"},
	{0x8E3510D3CADEE9FDULL, "l_mcr372ar_gallows_templar_nt.bnk"},
	{0x8E3587310D0F9B0DULL, "648304239.ogg"},
	{0x8E35C7CCEC011EA6ULL, "t1_skn_t02.tnt"},
	{0x8E35E8E19FBDEA28ULL, "6024984_f.fxe"},
	{0x8E365FB487FB7E2FULL, "art\\characters\\creatures\\mubari\\animation\\cwkb.ani"},
	{0x8E3684F76D339928ULL, "6175753_m.fxe"},
	{0x8E36AB999ED35957ULL, "hlac_w_bonepitdestroyed_a3_327681_0x.dds"},
	{0x8E36B4B34BC6A2E0ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_dis_imp_p.mmh"},
	{0x8E36B815C24F0218ULL, "llac_lt_undercityserialkiller_-1_4.dds"},
	{0x8E36D8433334209AULL, "16695760.ogg"},
	{0x8E37278474626917ULL, "vxht_keep_n_636_0_0.vlm"},
	{0x8E3834EB3DDC9903ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\selected.ani"},
	{0x8E3860E7F134730CULL, "6148551_m.fxe"},
	{0x8E38665D185B8084ULL, "566446700.ogg"},
	{0x8E3898D61A179102ULL, "94807059.ogg"},
	{0x8E38C60A5EFDFAF8ULL, "gen00fl_varric6105493.cl"},
	{0x8E38D0E7FE22DB84ULL, "6018820_f.fxe"},
	{0x8E3952C27D3A6AA6ULL, "hlac_ht_chantry_2534_52z.dds"},
	{0x8E399124F3B00F67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_rain_of_arrows.ani"},
	{0x8E3A2C8F022AE7A2ULL, "373940435.ogg"},
	{0x8E3A50117F02E444ULL, "llac_w_bonepitdestroyed_a3_131078_0.dds"},
	{0x8E3B0282781F5F3AULL, "vxht_brothel_n_6jg_0_0.vlm"},
	{0x8E3BC714D8782233ULL, "vxlt_0_lowtown_a2_n_fe0_0_0.vlm"},
	{0x8E3C7B60E87054F6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_cone.msh"},
	{0x8E3C8560E87065F4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_cone.mmh"},
	{0x8E3C8960E8706C07ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_cone.mao"},
	{0x8E3CA51CAE9CBE66ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_act_c.mmh"},
	{0x8E3CF58FDB433EC8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos04b.phy"},
	{0x8E3D3C5345900484ULL, "522201066.ogg"},
	{0x8E3D4BB9A499FC76ULL, "6163796_m.fxe"},
	{0x8E3D5716D7CCB6CCULL, "mag210_arianni6024573.cl"},
	{0x8E3D5CAA200E3055ULL, "vxlt_0_lowtown_a2_n_gc5_0_0.vlm"},
	{0x8E3D992A7EA1C797ULL, "cage_06.dds"},
	{0x8E3DC98B95EEEB38ULL, "69737756.ogg"},
	{0x8E3DF5F5EC2C604AULL, "follower_barks6153775.cl"},
	{0x8E3DF94E55BB0208ULL, "966690510.ogg"},
	{0x8E3E06087D76AA39ULL, "6129145_f.fxe"},
	{0x8E3F163BE1A88BEDULL, "6107300_m.fxe"},
	{0x8E3F1A6A609D29B2ULL, "vxca_caverns2prc_9fb_0_0.vlm"},
	{0x8E3FEDFA8E314B7FULL, "one101_orders_followers6115785.cl"},
	{0x8E40015A680D1131ULL, "vxht_bartrand_n_6qa_0_0.vlm"},
	{0x8E40240DD9E86CC5ULL, "6169708_m.fxe"},
	{0x8E4085F944DCD2EDULL, "407056126.ogg"},
	{0x8E4095E569E963CAULL, "6096467_m.fxe"},
	{0x8E411FBCB3945E3DULL, "art\\vfx\\combat\\vfx_glb_stat_sleep_c.cess.ani"},
	{0x8E412ADA2D67E629ULL, "403919384.ogg"},
	{0x8E4151D41CEDCAFBULL, "6068566_m.fxe"},
	{0x8E4180137B851E69ULL, "hlac_lt_0_lowtown_docks_n_-4_5x.dds"},
	{0x8E41B5CE9435E9ACULL, "vxdr_primevaltomb_5c4_0_0.vlm"},
	{0x8E43090EC647F138ULL, "191864618.ogg"},
	{0x8E431733773C13E8ULL, "hlac_w_bonepit_0_0x.dds"},
	{0x8E43789038106776ULL, "isa311_hawke_comments.cnv"},
	{0x8E43E8132471895FULL, "mer120_hubert6017870.cl"},
	{0x8E440B6164884B1AULL, "6201542_m.fxe"},
	{0x8E441499CEEAE5AFULL, "lit_conj_dust.mao"},
	{0x8E4497B34BD2AEFFULL, "art\\vfx\\combat\\vfx_r_ibl_fli_dis_imp_p.anb"},
	{0x8E44A23976A940E6ULL, "vxlt_ruinedhouse_n_tj_0_0.vlm"},
	{0x8E44AE1EBD45FA05ULL, "fen110_fenris6118093.cl"},
	{0x8E44D39431AFF711ULL, "mer340_cowan6083549.cl"},
	{0x8E454891D7715852ULL, "6180373_m.fxe"},
	{0x8E45EED95E68B760ULL, "6111095_f.fxe"},
	{0x8E46664AB2B71CC0ULL, "24565824.ogg"},
	{0x8E46908AFF18DDB9ULL, "vxw_sundermount_2tn_0_0.vlm"},
	{0x8E46E82FB9FA8485ULL, "gen00fl_varric6079577.cl"},
	{0x8E470816300AB93BULL, "6068697_f.fxe"},
	{0x8E47168D22B12327ULL, "s_ballista_dust.mao"},
	{0x8E472EC2A3931F51ULL, "gen00fl_isabela6101465.cl"},
	{0x8E477A9038145356ULL, "isa311_hawke_comments.bnk"},
	{0x8E47AB1A1F222E5EULL, "hlac_lt_hangedman_-1_18z.dds"},
	{0x8E47C83FCCCD51FBULL, "cd_6099574_isabelainlove.bnk"},
	{0x8E47F0C8175DDEE5ULL, "6087714_m.fxe"},
	{0x8E47F380E1754B7FULL, "vxht_mansion2b_2q7_0_0.vlm"},
	{0x8E4842D9AD6CE6F4ULL, "354875848.ogg"},
	{0x8E488C056B2DA524ULL, "lt_0_lowtown_docks_a3_96_0.mmh"},
	{0x8E4890056B2DAC77ULL, "lt_0_lowtown_docks_a3_96_0.mao"},
	{0x8E48A2056B2DCAC6ULL, "lt_0_lowtown_docks_a3_96_0.msh"},
	{0x8E491E4562A8B4E9ULL, "gen00fl_varric6107298.cl"},
	{0x8E495F3045E9FEA8ULL, "vxht_chantry_28c_0_0.vlm"},
	{0x8E496721ECB5A317ULL, "437156575.ogg"},
	{0x8E4A067156BD8E49ULL, "follower_banter6154066.cl"},
	{0x8E4A720EBA7BA2C3ULL, "vxw_encounter2_camp_qt_0_0.vlm"},
	{0x8E4A9E3C765D7B9BULL, "6043288_m.fxe"},
	{0x8E4B1B069B057088ULL, "666072212.ogg"},
	{0x8E4B391CC096BBA7ULL, "vxht_chantry_n_2t3_0_0.vlm"},
	{0x8E4B9C9B7A6BF24EULL, "follower_banter6161052.cl"},
	{0x8E4C2695902D8C26ULL, "llac_lt_0_lowtown_a3_n_c_-3_5.dds"},
	{0x8E4C328D8AA07994ULL, "vxht_0_hightown_a3_anc_0_0.vlm"},
	{0x8E4C4BEB3DED69EDULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\selected.gad"},
	{0x8E4C9313C2A46AE2ULL, "tutorials.gda"},
	{0x8E4DA1EC32B78632ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd02.phy"},
	{0x8E4DAF077B53D43AULL, "vxw_sundermount_20u_0_0.vlm"},
	{0x8E4EA64C3CF8A364ULL, "6152080_m.fxe"},
	{0x8E4EC4E2052460B5ULL, "6075121_f.fxe"},
	{0x8E4EE39F1C390F68ULL, "232238930.ogg"},
	{0x8E4EED86106B77DBULL, "qun231_varnell6078102.cl"},
	{0x8E4F193490584740ULL, "vxca_caverns1_j27_0_2.vlm"},
	{0x8E4F243B978E113EULL, "6076159_m.fxe"},
	{0x8E4F2BD8539FB1E4ULL, "931542725.ogg"},
	{0x8E502227C320EFA5ULL, "vxlt_backalley_n_1l9_0_0.vlm"},
	{0x8E50667EC60BBEBFULL, "52422871.ogg"},
	{0x8E50C86BE859D74EULL, "follower_barks6153317.cl"},
	{0x8E5118640947E43AULL, "vxht_0_hightown_a3_9b9_0_0.vlm"},
	{0x8E511CFD503C4901ULL, "dae000_arishok6095674.cl"},
	{0x8E5199029E685FA5ULL, "6102141_m.fxe"},
	{0x8E51B8DE8D8FCE69ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_hed_0n.dds"},
	{0x8E51BC679FEC98BCULL, "hlac_lt_0_lowtown_docks_a3_-196609_1z.dds"},
	{0x8E51C42F85BE3553ULL, "6181928_m.fxe"},
	{0x8E51DCA9D3BA140CULL, "dae300_alistair_king6133516.cl"},
	{0x8E524A842520B0B1ULL, "sse_ss_coterie_mage_f.gda"},
	{0x8E5283159BC87BAAULL, "916098865.ogg"},
	{0x8E52E0C7C6DE7E1EULL, "6094995_m.fxe"},
	{0x8E5315EA1D474833ULL, "warrior_1.cut"},
	{0x8E535157976DF9F1ULL, "llac_ht_chantry_n_2534_7.dds"},
	{0x8E538D2D369BF4FFULL, "402378234.ogg"},
	{0x8E53A9CB59B3ECEEULL, "isa111_isabela6067124.cl"},
	{0x8E54F70603CE6EB2ULL, "6035363_m.fxe"},
	{0x8E550267F10EF639ULL, "vxw_sundermount_2ym_0_0.vlm"},
	{0x8E558E97FAD6ED13ULL, "6148839_m.fxe"},
	{0x8E560DF2B5966B98ULL, "353891377.ogg"},
	{0x8E5634659D8AB45DULL, "gen00fl_isabela6123230.cl"},
	{0x8E563D74A1484C20ULL, "vxlt_0_lowtown_a3_n_c_h1v_0_0.vlm"},
	{0x8E564237175F59E5ULL, "hlac_lt_0_lowtown_-65539_0y.dds"},
	{0x8E56425270F10591ULL, "335318121.ogg"},
	{0x8E566144CEE219D6ULL, "vxht_brothel_n_60m_0_0.vlm"},
	{0x8E5774EBB0947124ULL, "lowtown_argument6169424.cl"},
	{0x8E58025896E28B50ULL, "follower_banter6172946.cl"},
	{0x8E582B627C4A9126ULL, "vxlt_warehouse_n_1e0_0_0.vlm"},
	{0x8E5834B19D69BF62ULL, "6174323_m.fxe"},
	{0x8E58B105341C278EULL, "6029026_f.fxe"},
	{0x8E58B877297BC914ULL, "gen00fl_dog6117776.cl"},
	{0x8E58E3EBAC71BB43ULL, "vxw_encounter2_camp_lw_0_0.vlm"},
	{0x8E590958A9CFA09BULL, "w_woundedcoast_50_0.rml"},
	{0x8E591073BFF4C7A5ULL, "234064228.ogg"},
	{0x8E5A19F604A42216ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_edge_1.mao"},
	{0x8E5A2EB23FA654CBULL, "gen00fl_sebastian6195876.cl"},
	{0x8E5A9F7AEAC50686ULL, "710499354.ogg"},
	{0x8E5AC0660B939D8CULL, "vxht_keep_n_73h_0_0.vlm"},
	{0x8E5AFD86A65949BCULL, "vxw_bonepit_1ql_0_0.vlm"},
	{0x8E5B500DB9F31353ULL, "211827453.ogg"},
	{0x8E5B9F056D8ABCA4ULL, "ht_0_hightownexplosion_n_42_0_ll.mmh"},
	{0x8E5BA958B35EC0DCULL, "vxlt_0_lowtown_docks_a3_d3f_0_0.vlm"},
	{0x8E5BBE3292CA35C1ULL, "349533983.ogg"},
	{0x8E5C014EEEB66DB9ULL, "vxlt_undercityencounter_13p_0_1.vlm"},
	{0x8E5C1E92B07DDC0DULL, "6017811_f.fxe"},
	{0x8E5C1EE707C6F781ULL, "art\\characters\\creatures\\dragon\\animation\\cwbr.ani"},
	{0x8E5C20E46D49ECBAULL, "vxw_sundermount_25v_0_0.vlm"},
	{0x8E5C20F24E23EA95ULL, "str_crater_halff.mao"},
	{0x8E5C284325BA0119ULL, "428261879.ogg"},
	{0x8E5C3EBE38F39315ULL, "633368363.ogg"},
	{0x8E5C8B9D4338C748ULL, "hlac_lt_0_lowtown_a2_65534_4z.dds"},
	{0x8E5CC4CA5D69E70AULL, "6080529_m.fxe"},
	{0x8E5D228FEAE8A48DULL, "vxht_keep_4b6_0_0.vlm"},
	{0x8E5D9842382370CAULL, "15953610.ogg"},
	{0x8E5DA8DDB78E42D9ULL, "mcr370_meredith6085906.cl"},
	{0x8E5E54EC4C8612CDULL, "201332961.ogg"},
	{0x8E5E9404B5170825ULL, "vxlt_backalley_n_1w8_0_0.vlm"},
	{0x8E5E97AFF6743890ULL, "290540895.ogg"},
	{0x8E5F425045B2632CULL, "gen00fl_merrill6129671.cl"},
	{0x8E6001DD76DD98DEULL, "gen00fl_merrill6127257.cl"},
	{0x8E607DDB54BAF612ULL, "nat380_nathaniel6071023.cl"},
	{0x8E6132D0612405F2ULL, "860687124.ogg"},
	{0x8E6176C5DB2214D9ULL, "ca_caverns2_probe_9ep.dds"},
	{0x8E617A6467A181F4ULL, "hlac_w_blightlands_3_0z.dds"},
	{0x8E61F1330EBFB879ULL, "vxht_bartrand_n_65e_0_0.vlm"},
	{0x8E6229E707CB7623ULL, "art\\characters\\creatures\\dragon\\animation\\cwbr.gad"},
	{0x8E636115F3B4B40DULL, "ht_0_hightown_26_0_lc.mmh"},
	{0x8E63D2E7BC1A7A52ULL, "vxw_bonepitmining_a2_1xf_0_1.vlm"},
	{0x8E64518693B1E905ULL, "zz_approval_debug6003723.cl"},
	{0x8E646D9CB1B44902ULL, "hlac_ca_caverns1_mines_24424_7y.dds"},
	{0x8E6478D3E4E98C73ULL, "6151305_m.fxe"},
	{0x8E648034C79FA91DULL, "vxca_cave_2e4_0_1.vlm"},
	{0x8E6531F4CFCC7E2EULL, "art\\characters\\playercharacter\\elffemale\\boo_lgtp.phy"},
	{0x8E6647DDA4C29EDCULL, "vxca_caverns1_dr7_0_0.vlm"},
	{0x8E66869DA3B7B080ULL, "6035307_m.fxe"},
	{0x8E66AA78B89D8081ULL, "follower_banter6141741.cl"},
	{0x8E672E2FC50D525BULL, "llac_ht_mansion2_818_47.dds"},
	{0x8E6755C89E67D3C3ULL, "vxw_encounter2_camp_gv_0_0.vlm"},
	{0x8E678CBD1F53BD62ULL, "gen00fl_anders6119926.cl"},
	{0x8E67AE1BF163ED15ULL, "hlac_w_encounter3_-65532_0z.dds"},
	{0x8E67E64DC8068836ULL, "614583508.ogg"},
	{0x8E68E41DBC67BCB4ULL, "lt_gallowscourtyard_n_ll.mmh"},
	{0x8E6906E707D15908ULL, "art\\characters\\creatures\\dragon\\animation\\cwbr.evt"},
	{0x8E692ADA450B962EULL, "gen00fl_fenris6116554.cl"},
	{0x8E692E397F9194BFULL, "502221638.ogg"},
	{0x8E696F63984F623CULL, "vxw_bonepit_1zm_0_0.vlm"},
	{0x8E697034F4F525A9ULL, "6149751_m.fxe"},
	{0x8E69827DD47287DFULL, "vxlt_undercitydungeon_1yd_0_0.vlm"},
	{0x8E6A1B35A554D95CULL, "vxlt_0_lowtown_docks_a3_d4g_0_0.vlm"},
	{0x8E6A6E4469012BFAULL, "vxca_cavern1_haunted_j40_0_0.vlm"},
	{0x8E6B0F863D04725CULL, "art\\vfx\\creatures\\fxc_shade_c.impa.ani"},
	{0x8E6B7D45E42D61B6ULL, "6091413_m.fxe"},
	{0x8E6CA155F0E6785BULL, "mcr370_meredith6092214.cl"},
	{0x8E6DA7E6E3926BFEULL, "vxlt_0_lowtown_n_e2c_0_0.vlm"},
	{0x8E6DC397094EC793ULL, "542820727.ogg"},
	{0x8E6DC7821D4B023EULL, "vxw_bonepitdestroyed_a3_26l_0_0.vlm"},
	{0x8E6DE9A472AF25ACULL, "mer210cr_seldrek.utc"},
	{0x8E6DFF50B5A3E759ULL, "vxw_encounter2_m2_0_0.vlm"},
	{0x8E6ECD5C3B290F09ULL, "6128883_f.fxe"},
	{0x8E6EE7342CC3D137ULL, "hlac_w_encounter3_0_5y.dds"},
	{0x8E6F05FF4FCADAC2ULL, "438981832.ogg"},
	{0x8E6F5DDB5E9482C8ULL, "6001191_m.fxe"},
	{0x8E702CCD0EEBA65CULL, "gen00fl_aveline6139778.cl"},
	{0x8E711C863D08F39EULL, "art\\vfx\\creatures\\fxc_shade_c.impa.gad"},
	{0x8E71388520477ABAULL, "390645340.ogg"},
	{0x8E714DA4691BDF5CULL, "vfx0_cs_alt.vsh.fxo"},
	{0x8E715BBB3A0CAC9EULL, "hlac_w_sundermount_-131072_0x.dds"},
	{0x8E716748BFEBAE7BULL, "vxht_0_hightown_626_0_0.vlm"},
	{0x8E726DD805E2DF2DULL, "plt_cod_aow_tactics.plo"},
	{0x8E729914716EB94BULL, "gen_im_trs_apo_sta_ele_r1.uti"},
	{0x8E7354FB19CA071BULL, "lgt160_wk_soph_cultist.bnk"},
	{0x8E736E863237818DULL, "dae100_bran.bnk"},
	{0x8E73EEDF71D563CCULL, "gen00fl_fenris6106720.cl"},
	{0x8E7411CE21A570BDULL, "gen00fl_varric6076024.cl"},
	{0x8E74C414B5A3753CULL, "vxw_woundedcoast_sx_0_0.vlm"},
	{0x8E74E8EE3F68490EULL, "dae000_keep_bureaucrat_2_st6165383.cl"},
	{0x8E74EB604CFCB23AULL, "fen110_fenris6118016.cl"},
	{0x8E75219EFAD15C0EULL, "llac_w_woundedcoast_131075_0.dds"},
	{0x8E752833D5F4CDDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow_death.ani"},
	{0x8E7574E4311A3801ULL, "rock_01.dds"},
	{0x8E75D27145FE6BF5ULL, "vxdr_primevalentrance_wr_0_0.vlm"},
	{0x8E762445261C2868ULL, "gen_im_trs_gem_onyx_02.uti"},
	{0x8E7656FB19CC40DBULL, "lgt160_wk_soph_cultist.cnv"},
	{0x8E76D0DADFF2857BULL, "dae000_pc_home_ambients6208557.cl"},
	{0x8E7723E9D3F5CF6CULL, "gen00fl_anders6119537.cl"},
	{0x8E77703877BCC163ULL, "866852471.ogg"},
	{0x8E779086323BA4F5ULL, "dae100_bran.cnv"},
	{0x8E77B461A8E032CAULL, "6067733_m.fxe"},
	{0x8E77F7863D0ED3BDULL, "art\\vfx\\creatures\\fxc_shade_c.impa.evt"},
	{0x8E78E0215AF7447AULL, "vxca_cavern1_haunted_j31_0_0.vlm"},
	{0x8E797ADB240BD4C7ULL, "6004098_m.fxe"},
	{0x8E79E5D201C86C97ULL, "104083645.ogg"},
	{0x8E79E80B1A7F515CULL, "6680798.ogg"},
	{0x8E79F030C0AEE162ULL, "dae000_gamlen6145100.cl"},
	{0x8E7A3D6D3D2C0B0DULL, "67068512.ogg"},
	{0x8E7A6F81566665A8ULL, "616725274.ogg"},
	{0x8E7A7ADA72C5A48CULL, "zz_followers_banter6157358.cl"},
	{0x8E7A97FE8771F1B1ULL, "6172722_m.fxe"},
	{0x8E7A9D014B6DCC2EULL, "fen110_fenris6096220.cl"},
	{0x8E7AAE122E03E374ULL, "cd_mer300cs_high_dragon_reveal.wbd"},
	{0x8E7AC2834EDC247DULL, "201044226.ogg"},
	{0x8E7BF1B75F993C34ULL, "cth_cm1.tnt"},
	{0x8E7BF59BCC76F96EULL, "6060303_m.fxe"},
	{0x8E7C395F0F411ABEULL, "vxw_bonepitdestroyed_a3_21m_0_0.vlm"},
	{0x8E7C3D34E3A0A665ULL, "gen00fl_aveline6121901.cl"},
	{0x8E7C712DA799FFD9ULL, "vxw_encounter2_h1_0_0.vlm"},
	{0x8E7CA95BC1CA819DULL, "static_static_p6.psh.fxo"},
	{0x8E7CC5260BF1316EULL, "359189326.ogg"},
	{0x8E7CE04B39013C13ULL, "hlac_ht_chantry_2534_6z.dds"},
	{0x8E7CF44442C73226ULL, "vxdr_primevalentrance_1kv_0_0.vlm"},
	{0x8E7D5C29FC57DE61ULL, "gen00flst_sebactthreed.stg"},
	{0x8E7D70931BA4331AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_tablecloth.mmh"},
	{0x8E7D74931BA43A51ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_tablecloth.mao"},
	{0x8E7D82931BA45260ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_tablecloth.msh"},
	{0x8E7E08940C7FEE6EULL, "lgt210cr_mag_escp_rescued.utc"},
	{0x8E7E13685AD8C201ULL, "vxlt_0_lowtown_docks_a3_d37_0_0.vlm"},
	{0x8E7E160BA6AEBE9FULL, "444739253.ogg"},
	{0x8E7E87E22D52568AULL, "dae000_merchant_rare_em6177943.cl"},
	{0x8E7EC7F6C5980A89ULL, "mag121_idunna6084860.cl"},
	{0x8E7ED50A4B4E426CULL, "120802391.ogg"},
	{0x8E7F34B39E7061ABULL, "art\\vfx\\combat\\vfx_glb_ele_n_additive_c.mmh"},
	{0x8E7F57D12308263CULL, "lowtown_male_alienage2_mob6166537.cl"},
	{0x8E7F62750F219089ULL, "vxht_chantry_n_2l5_0_0.vlm"},
	{0x8E8064250C792660ULL, "2da_base.gda"},
	{0x8E80DFBDE3E51103ULL, "hlac_lt_0_lowtown_a2_-3_2z.dds"},
	{0x8E80FF03E8AF2AF7ULL, "follower_banter6170783.cl"},
	{0x8E819022D6124357ULL, "909646627.ogg"},
	{0x8E819EB1100487ABULL, "zz_amb_lever.ncs"},
	{0x8E81B2AF3EA7B23BULL, "art\\vfx\\models\\vfx_vol_light_wide.msh"},
	{0x8E81D0AF3EA7E525ULL, "art\\vfx\\models\\vfx_vol_light_wide.mmh"},
	{0x8E823ECF70C4463AULL, "6042804_m.fxe"},
	{0x8E82BFFB8BDEBF94ULL, "tal_tau_in.mao"},
	{0x8E833033D60118D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow_death.evt"},
	{0x8E837FD339B2D729ULL, "677028515.ogg"},
	{0x8E842F6CECF05930ULL, "vxw_sundermount_307_0_1.vlm"},
	{0x8E84B5C69DD3EE34ULL, "221200623.ogg"},
	{0x8E852DDBE7A1A6CBULL, "vxht_mansion_n_6nq_0_0.vlm"},
	{0x8E85A97A8DDB6205ULL, "llac_lt_0_lowtown_a2_1_9.dds"},
	{0x8E85D4EA754A5246ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.dura.evt"},
	{0x8E861770DBCF293BULL, "vxca_cave_2b6_0_0.vlm"},
	{0x8E86534C4E2135A4ULL, "78545546.ogg"},
	{0x8E865D1756F880B0ULL, "198990170.ogg"},
	{0x8E866893E1C8B684ULL, "vxlt_darktown_5kf_0_0.vlm"},
	{0x8E869EACDF8B1D57ULL, "152778175.ogg"},
	{0x8E86BF1DC45E6DE9ULL, "hlac_lt_0_lowtown_n_196607_0x.dds"},
	{0x8E87336875BDA496ULL, "472426908.ogg"},
	{0x8E87841AB298A306ULL, "vxlt_0_lowtown_n_e0u_0_0.vlm"},
	{0x8E87901ED9AF593AULL, "mag110cr_sailor_1.utc"},
	{0x8E87EE6239E03195ULL, "523775822.ogg"},
	{0x8E88CDC53B864FC7ULL, "972868641.ogg"},
	{0x8E89319D3AF82EC3ULL, "hlac_lt_0_lowtown_a2_2_0x.dds"},
	{0x8E89601507A74D89ULL, "6047259_m.fxe"},
	{0x8E89FC94CF3A07B9ULL, "238289717.ogg"},
	{0x8E8A4C48FB72A27EULL, "qcr000_arishok6057084.cl"},
	{0x8E8A4F33D6076C65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\mightyblow_death.gad"},
	{0x8E8AA81946C4CB08ULL, "vxlt_0_lowtown_a3_n_c_gnd_0_0.vlm"},
	{0x8E8AE30A99901859ULL, "vxw_encounter2_s0_0_0.vlm"},
	{0x8E8B3E31F3F0FC87ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_maso_0t.dds"},
	{0x8E8BCE4EFD63C2FCULL, "571504855.ogg"},
	{0x8E8BCFBDA0CF290FULL, "6077676_m.fxe"},
	{0x8E8C41BD904781AAULL, "dae000_keep_noblewoman_sta6162074.cl"},
	{0x8E8C4D94385D8113ULL, "llac_w_bonepitmining_a2_0_5.dds"},
	{0x8E8C83DB3BC8E59DULL, "vxlt_gallowstemplar_51e_0_0.vlm"},
	{0x8E8C85454CCEDA81ULL, "vxlt_0_lowtown_docks_a3_d46_0_0.vlm"},
	{0x8E8CB52577A4E1E4ULL, "llac_ca_caverns1_24421_4.dds"},
	{0x8E8CEBEA75509791ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.dura.gad"},
	{0x8E8CF6EAB3997F3FULL, "gen00fl_merrill6129986.cl"},
	{0x8E8D15B39E7C6A10ULL, "art\\vfx\\combat\\vfx_glb_ele_n_additive_c.anb"},
	{0x8E8D3DC2B0B42BF5ULL, "w_bonepitdestroyed_a3_249_0.msh"},
	{0x8E8D3FC2B0B42F6BULL, "w_bonepitdestroyed_a3_249_0.mmh"},
	{0x8E8D4BC2B0B44308ULL, "w_bonepitdestroyed_a3_249_0.mao"},
	{0x8E8D5529CBE71C35ULL, "350108187.ogg"},
	{0x8E8DB9CC271086B5ULL, "vxht_keep_n_42m_0_0.vlm"},
	{0x8E8E36B7A7503B0DULL, "16906421.ogg"},
	{0x8E8F7B992D7CC2EDULL, "221131407.ogg"},
	{0x8E8FB63FCD0AE8F9ULL, "cd_6099574_isabelainlove.wbd"},
	{0x8E8FBB0E6835878AULL, "zz_dae_debug6029789.cl"},
	{0x8E8FFB22FC747D4CULL, "llac_ca_caverns1_24424_27.dds"},
	{0x8E905A145659EA05ULL, "206345106.ogg"},
	{0x8E90E0FD351B88B6ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masa.phy"},
	{0x8E91020D5DDA40A8ULL, "art\\characters\\creatures\\spider\\textures\\c_poispidra_0n.dds"},
	{0x8E911A5C4C430696ULL, "gen_im_arm_cht_med_act1_01.uti"},
	{0x8E912C549DB2F6CDULL, "6054573_f.fxe"},
	{0x8E93202C225C14EAULL, "hlac_lt_0_lowtown_a2_n_c_0_3y.dds"},
	{0x8E9385E43133528BULL, "rock_01.mao"},
	{0x8E938E8C4E02849CULL, "312600066.ogg"},
	{0x8E93ECEA7556B83FULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.dura.ani"},
	{0x8E94B097465FD3ECULL, "87728442.ogg"},
	{0x8E94DBD160CF4873ULL, "vxht_0_hightown_a3_c5c_0_0.vlm"},
	{0x8E94FCD63B4AAAF7ULL, "247369491.ogg"},
	{0x8E95320B47B768C1ULL, "vxlt_undercityencounter_zy_0_0.vlm"},
	{0x8E9570A44B514C26ULL, "984015405.ogg"},
	{0x8E95BA78B085B5B0ULL, "vxlt_undercityencounter_1pm_0_0.vlm"},
	{0x8E963C7068F1DD5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext.evt"},
	{0x8E96646B9FD4EB1AULL, "6045997_m.fxe"},
	{0x8E9687E96EEFBEE0ULL, "gen00fl_sebastian6188714.cl"},
	{0x8E969B0C94D7FEFFULL, "lt_hangedman_n.arl"},
	{0x8E9705C98155F05AULL, "llac_w_sundermount_327674_0.dds"},
	{0x8E978460E8BD0FEFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_cone.phy"},
	{0x8E97CC674420E5EAULL, "203298237.ogg"},
	{0x8E9800B4F0A896AFULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.mmh"},
	{0x8E98C49BFB525503ULL, "702903422.ogg"},
	{0x8E98EC744C5BBCDFULL, "6048483_m.fxe"},
	{0x8E99207F2665B78AULL, "525965349.ogg"},
	{0x8E999C616485E8DFULL, "vxht_brothel_n_6c5_0_0.vlm"},
	{0x8E99C640808197B0ULL, "hlac_ht_mansion2b_818_5y.dds"},
	{0x8E99CFA76DB647E4ULL, "vxca_cavern1_haunted_eg0_0_0.vlm"},
	{0x8E99E89CFB982887ULL, "mag311_keran6041734.cl"},
	{0x8E9A0C58AA073D67ULL, "w_woundedcoast_50_0.anb"},
	{0x8E9A4CCA00F660ADULL, "dae000cr_grand_cleric6173656.cl"},
	{0x8E9A6E8DB2F21F52ULL, "46980324.ogg"},
	{0x8E9AC9FD3523B657ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masa.msh"},
	{0x8E9ADBFD3523D4BAULL, "art\\characters\\playercharacter\\humanmale\\hlh_masa.mao"},
	{0x8E9ADFFD3523DBC9ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masa.mmh"},
	{0x8E9AF3EE43511D1BULL, "849147101.ogg"},
	{0x8E9AFBF1847485E8ULL, "character_character_p139.psh.fxo.dx11"},
	{0x8E9B1493E78AE72AULL, "1020034312.ogg"},
	{0x8E9BBE0B3DFE165AULL, "mcr310_argument6074506.cl"},
	{0x8E9BEADF24E38BABULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\additivedamage.ani"},
	{0x8E9C586C4A4D7C96ULL, "6093900_f.fxe"},
	{0x8E9C692DA6E408B7ULL, "llac_ca_cavern1_haunted_24427_2.dds"},
	{0x8E9C7D03D54AB738ULL, "6064643_f.fxe"},
	{0x8E9D357068F7EF9CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext.gad"},
	{0x8E9D5E25237A3F19ULL, "6125784_m.fxe"},
	{0x8E9D8C662DBC46CDULL, "zz_approval_debug6003925.cl"},
	{0x8E9D8D1552C4504FULL, "art\\vfx\\combat\\vfx_m_amg_estorm_lit_p.impa.gad"},
	{0x8E9D9B9961E440EFULL, "838541712.ogg"},
	{0x8E9E59B7B6FDE2B4ULL, "vxht_mansion2b_34b_0_0.vlm"},
	{0x8E9F30F53C608193ULL, "6087746_f.fxe"},
	{0x8E9F5DB72655466FULL, "355111407.ogg"},
	{0x8E9F86013B3E5167ULL, "845701703.ogg"},
	{0x8E9F9D69E0FD997DULL, "350323731.ogg"},
	{0x8E9FA3B1101D8D96ULL, "zz_amb_lever.utp"},
	{0x8EA07CAE9F8B190AULL, "atl_purchaseco_dxt5_00.dds"},
	{0x8EA0B79AC0499140ULL, "300096754.ogg"},
	{0x8EA0C81C99F182D2ULL, "dae000_gamlen6048445.cl"},
	{0x8EA0E5382AFD6531ULL, "gen_im_cns_flu_sebastian_04.uti"},
	{0x8EA10DC41C9C38DDULL, "hlac_ht_keep_3732_1z.dds"},
	{0x8EA10EAF5F6B194BULL, "mrl150_merrill6077659.cl"},
	{0x8EA116E9353AA93FULL, "hlac_ht_0_hightown_n_-131074_1y.dds"},
	{0x8EA127E5472BDD68ULL, "vxlt_gallowstemplar_f_522_0_0.vlm"},
	{0x8EA185B47D27A56EULL, "llac_dr_dungeon_3217_5.dds"},
	{0x8EA2400B5D2F0FF5ULL, "vxlt_gallowscourtyard_n_63q_0_0.vlm"},
	{0x8EA29B420E6573B4ULL, "899497824.ogg"},
	{0x8EA2A97B78BFE78BULL, "vxlt_undercityencounter_11q_0_0.vlm"},
	{0x8EA2E397D3EF599AULL, "hlac_lt_0_lowtown_a3_n_c_131071_0x.dds"},
	{0x8EA336D76027C43CULL, "vxw_bonepit_1na_0_0.vlm"},
	{0x8EA3735F2B138129ULL, "w_encounter3_160_0.phy"},
	{0x8EA3B4E741B94445ULL, "hlac_w_woundedcoast_-65532_0x.dds"},
	{0x8EA3C8894C6D8E8AULL, "180321915.ogg"},
	{0x8EA3D31057474E27ULL, "6021356_m.fxe"},
	{0x8EA3FD3C89673C52ULL, "146008171.ogg"},
	{0x8EA4347068FE0C9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext.ani"},
	{0x8EA45F6DE6410960ULL, "6059231_m.fxe"},
	{0x8EA4A21552CA92FCULL, "art\\vfx\\combat\\vfx_m_amg_estorm_lit_p.impa.evt"},
	{0x8EA4BC66F3A5E1FCULL, "6164232_m.fxe"},
	{0x8EA530F4D00260F1ULL, "art\\characters\\playercharacter\\elffemale\\boo_lgtp.mmh"},
	{0x8EA53CF4D0027492ULL, "art\\characters\\playercharacter\\elffemale\\boo_lgtp.mao"},
	{0x8EA54AF4D0028CAFULL, "art\\characters\\playercharacter\\elffemale\\boo_lgtp.msh"},
	{0x8EA5F9A3FA03241DULL, "mer340_cowan6054181.cl"},
	{0x8EA69C122E28FF0EULL, "cd_mer300cs_high_dragon_reveal.bnk"},
	{0x8EA6A6CDEDB37812ULL, "318352637.ogg"},
	{0x8EA6C026229DB224ULL, "ttx_window_large_a__2.msh"},
	{0x8EA6D626229DD7C6ULL, "ttx_window_large_a__2.mmh"},
	{0x8EA6E707890EFB9DULL, "fen110_fenris6117940.cl"},
	{0x8EA75212662CAA7EULL, "vxht_keep_n_7a0_0_0.vlm"},
	{0x8EA7AEDEF059D6AFULL, "ht_0_hightown_n_ll.mmh"},
	{0x8EA7B2D076CDE763ULL, "693263668.ogg"},
	{0x8EA7BB38568BE222ULL, "900245899.ogg"},
	{0x8EA7CB725F62DBF8ULL, "gen_im_trs_dwv_junk5_03.uti"},
	{0x8EA7EA8823C6CDCCULL, "187971566.ogg"},
	{0x8EA83EF97F462594ULL, "qcr000_arishok6060192.cl"},
	{0x8EA94BE14A816780ULL, "fxa_cen_aoe_volt2.mao"},
	{0x8EA9A775AE061153ULL, "vxlt_foundry_n_wf_0_0.vlm"},
	{0x8EA9D2DF24EFA022ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\additivedamage.evt"},
	{0x8EA9D30B7FE15892ULL, "585373566.ogg"},
	{0x8EA9EA642B9CF496ULL, "hlac_lt_0_lowtown_131071_4z.dds"},
	{0x8EAA041ABE94A0F9ULL, "mag291_vincento6027710.cl"},
	{0x8EAA4C475B2DAD98ULL, "vxlt_hangedman_2zc_0_0.vlm"},
	{0x8EAAB39DF1665C78ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_glowbugs.mmh"},
	{0x8EAB3B20BF9D82D4ULL, "hlac_ca_caverns1_mines_24421_10x.dds"},
	{0x8EAC7664550DCC5DULL, "1039252812.ogg"},
	{0x8EAC85329F03FFDDULL, "97294884.ogg"},
	{0x8EACBB7D5A81C84AULL, "6125691_m.fxe"},
	{0x8EACD8D7D9C2269EULL, "dae110st_jansen.stg"},
	{0x8EAD239035C2C5B2ULL, "sst_ss_lowclass_thugs_f_1.gda"},
	{0x8EAE5E3805024B9BULL, "vxht_chantry_2e7_0_0.vlm"},
	{0x8EAE6442D2C52B4DULL, "fx_blood04.mao"},
	{0x8EAE693BD8F22AF0ULL, "640222585.ogg"},
	{0x8EAECCF83EA72432ULL, "pro000_merchant6051804.cl"},
	{0x8EAED00655ECECFDULL, "qcr000_arishok6076624.cl"},
	{0x8EAF565363B341C5ULL, "mag130_danzig6018805.cl"},
	{0x8EAF62BE5B726133ULL, "59390256.ogg"},
	{0x8EAF78E304BD900CULL, "6148740_m.fxe"},
	{0x8EAF8A28B7C2B149ULL, "6087736_m.fxe"},
	{0x8EB00258E20E59A1ULL, "vxlt_0_lowtown_a2_n_c_byx_0_0.vlm"},
	{0x8EB03D09BCC97B15ULL, "33883361.ogg"},
	{0x8EB0553A6AD22958ULL, "642743729.ogg"},
	{0x8EB0B06D2232FA16ULL, "423322903.ogg"},
	{0x8EB0D1DF24F5BD35ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\additivedamage.gad"},
	{0x8EB0FAFCED633234ULL, "6077305_m.fxe"},
	{0x8EB13B711EAADB9FULL, "hla_ht_0_hightown_a2_-65535_0z.dds"},
	{0x8EB1AA1552D52A15ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_lit_p.impa.ani"},
	{0x8EB299F9DB45E46CULL, "hlac_lt_0_lowtown_docks_n_-65539_0z.dds"},
	{0x8EB3282B6F5AAFD2ULL, "vxlt_foundry_n_wc_0_0.vlm"},
	{0x8EB366A2747A1AF0ULL, "gen00fl_aveline6152566.cl"},
	{0x8EB3A6DC6065528DULL, "hlac_w_encounter3_0_4x.dds"},
	{0x8EB42FAE449D0640ULL, "237254652.ogg"},
	{0x8EB4494E5DFFBCE4ULL, "hlac_w_woundedcoast_-196606_0x.dds"},
	{0x8EB4B7D56E40D773ULL, "vxlt_hangedman_1vf_0_1.vlm"},
	{0x8EB4BA02A325DD06ULL, "6118417_m.fxe"},
	{0x8EB5099DDC88B498ULL, "rdr111ip_at_warehouse.utp"},
	{0x8EB51353CC032C71ULL, "vxw_sundermount_2np_0_0.vlm"},
	{0x8EB5923A2EE9310CULL, "vxca_caverns2_7w8_0_0.vlm"},
	{0x8EB59301D8C06B87ULL, "83698768.ogg"},
	{0x8EB609ED375A9FC1ULL, "lgt210pt_eau_debt.ncs"},
	{0x8EB6B3BB5BC2C07BULL, "vxht_0_hightown_9zu_0_0.vlm"},
	{0x8EB700F8149AC8CFULL, "6118580_m.fxe"},
	{0x8EB70481D2750B40ULL, "vxw_sundermount_2y9_0_1.vlm"},
	{0x8EB70A1C4030631DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoorbird_d.dds"},
	{0x8EB72EFD73898D2DULL, "vxdr_0_deeproads_6nv_0_1.vlm"},
	{0x8EB780E1E8DD221BULL, "dae000cr_grand_cleric6144087.cl"},
	{0x8EB798C04953297BULL, "26057500.ogg"},
	{0x8EB7FE5582517791ULL, "vxca_caverns1_mines_hmb_0_0.vlm"},
	{0x8EB87A8F48EE31A0ULL, "mag101_grace6053103.cl"},
	{0x8EB8E4ACEF2A3CB8ULL, "6143244_m.fxe"},
	{0x8EBA91FDA29EBD9AULL, "49256527.ogg"},
	{0x8EBAA5F06786649EULL, "dae150_chandan6148647.cl"},
	{0x8EBAC57940926F2EULL, "lt_0_lowtown_a2_46_0.anb"},
	{0x8EBAC80547934A24ULL, "vxdr_primevaltomb_6wo_0_0.vlm"},
	{0x8EBB2D5A4C77E0CAULL, "6122791_m.fxe"},
	{0x8EBB4D92D7439AB5ULL, "vxw_woundedcoast_ff_0_0.vlm"},
	{0x8EBB93E8B696F99CULL, "1061182524.ogg"},
	{0x8EBBDD194749986BULL, "hlac_w_encounter1_4_0x.dds"},
	{0x8EBBE5FDDF8A15BCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_45basic.mao"},
	{0x8EBBF1FDDF8A2A5FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_45basic.mmh"},
	{0x8EBBF7FDDF8A3401ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_45basic.msh"},
	{0x8EBC96A6352363CAULL, "6173685_m.fxe"},
	{0x8EBDFD8C73E53ED3ULL, "w_woundedcoast_195_0.phy"},
	{0x8EBE3335B5F8F586ULL, "boulder_mac_lig.mao"},
	{0x8EBEC72C9C4C90BAULL, "vis121_saemus6101159.cl"},
	{0x8EBF09F6210F5E70ULL, "vxlt_undercityserialkiller_1c9_0_1.vlm"},
	{0x8EBFB19D60710768ULL, "dae000_arishok6075198.cl"},
	{0x8EBFB3099BE0ACEFULL, "6089417_m.fxe"},
	{0x8EBFD190F78816D9ULL, "and230_mage_girl6109111.cl"},
	{0x8EC0650A2648F566ULL, "525988902.ogg"},
	{0x8EC09127539EC3B4ULL, "vxlt_0_lowtown_docks_a3_n_cov_0_0.vlm"},
	{0x8EC0FDE7D4768118ULL, "vxht_0_hightown_a2_bpw_0_0.vlm"},
	{0x8EC101B4F0CB9DECULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.anb"},
	{0x8EC15E1AD2AC064BULL, "isa211_party_reaction6083428.cl"},
	{0x8EC1C41F5BCD6B33ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.cess.ani"},
	{0x8EC20958AA288A68ULL, "w_woundedcoast_50_0.mmh"},
	{0x8EC21558AA289E0BULL, "w_woundedcoast_50_0.mao"},
	{0x8EC22758AA28BC92ULL, "w_woundedcoast_50_0.msh"},
	{0x8EC22BC7E0FC8451ULL, "6135709_m.fxe"},
	{0x8EC27A93C24EEF25ULL, "gen00fl_isabela6110175.cl"},
	{0x8EC320BF789A2281ULL, "6071131_m.fxe"},
	{0x8EC33252D34D782FULL, "6143140_m.fxe"},
	{0x8EC356589CA304E7ULL, "283544863.ogg"},
	{0x8EC37D867774FAF6ULL, "593367962.ogg"},
	{0x8EC383F12C9C281AULL, "hlac_ca_caverns1_24424_10z.dds"},
	{0x8EC38530BDF944F1ULL, "vxw_sundermount_2sw_0_0.vlm"},
	{0x8EC3B39995C27E6BULL, "lla_ht_0_hightown_-131072_0.dds"},
	{0x8EC3B3EAB15CCAF8ULL, "llac_lt_0_lowtown_docks_n_-1_1.dds"},
	{0x8EC3F925137F3353ULL, "919356172.ogg"},
	{0x8EC4041720DF498CULL, "vxca_caverns2_7l9_0_0.vlm"},
	{0x8EC46120D0733850ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeproofsml.mao"},
	{0x8EC46520D0733F63ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeproofsml.mmh"},
	{0x8EC47320D07356BDULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeproofsml.msh"},
	{0x8EC48966A2403DBCULL, "vxca_caverns1_ege_0_0.vlm"},
	{0x8EC57B2FDA319CD5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robb_0s.dds"},
	{0x8EC58AF609BC1EA7ULL, "6131097_m.fxe"},
	{0x8EC5CCABA2B55BCFULL, "vxlt_0_lowtown_docks_n_cv1_0_0.vlm"},
	{0x8EC5D8E42C31C57CULL, "gen00fl_sebastian6177907.cl"},
	{0x8EC6A37133D42EACULL, "vxdr_primevaltomb_8oe_0_0.vlm"},
	{0x8EC6C1DDE1729BE4ULL, "707577092.ogg"},
	{0x8EC6FBD20848F9B7ULL, "497099719.ogg"},
	{0x8EC70543F4C5C417ULL, "6155514_m.fxe"},
	{0x8EC7DC825A5D4671ULL, "1048375720.ogg"},
	{0x8EC84402DEE9E771ULL, "llac_lt_0_lowtown_a3_n_c_-196609_2.dds"},
	{0x8EC8B140D3BAB2EDULL, "ave200_donnic6111206.cl"},
	{0x8EC8CFBB7B58C557ULL, "6191886_m.fxe"},
	{0x8EC8F69C86B304E0ULL, "594279729.ogg"},
	{0x8EC9138AF2FFB1F1ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_0y.dds"},
	{0x8EC9BF6FC939B335ULL, "vxw_woundedcoast_ce_0_0.vlm"},
	{0x8EC9E4750108BAD6ULL, "gen_im_trs_tmp_insignia_02.uti"},
	{0x8ECA27E716AF4F46ULL, "6169831_m.fxe"},
	{0x8ECA4C831792AEE8ULL, "689095904.ogg"},
	{0x8ECAA3C65D3CB095ULL, "dae000_hate_enemies6167544.cl"},
	{0x8ECAAC233CA25F64ULL, "follower_banter6171873.cl"},
	{0x8ECAEACD252DABE2ULL, "118274922.ogg"},
	{0x8ECB1673956BBCEDULL, "qun131_taskmaster6059039.cl"},
	{0x8ECB1BBF2C6B91BFULL, "rdr111dg_martin6057074.cl"},
	{0x8ECB441DFF9F366FULL, "hlac_lt_0_lowtown_n_-1_7z.dds"},
	{0x8ECB7B944DC28623ULL, "387163270.ogg"},
	{0x8ECC38B378992CCAULL, "567596918.ogg"},
	{0x8ECD034A3CCDF68CULL, "489115574.ogg"},
	{0x8ECD064BEED7A15BULL, "vxht_0_hightown_a2_c4e_0_0.vlm"},
	{0x8ECD9595FD828675ULL, "6113932_m.fxe"},
	{0x8ECE3C4E436394F4ULL, "ef_comt_servant_2a.mor"},
	{0x8ECE6EA30C313F8CULL, "mer320_hubert6081997.cl"},
	{0x8ECECEC174DDCE04ULL, "sst_ss_danzig.gda"},
	{0x8ECEF1F779689EA7ULL, "mag210_arianni6024972.cl"},
	{0x8ECF2837D2368E56ULL, "6208336_m.fxe"},
	{0x8ECF387B08CD43C6ULL, "vxht_0_hightown_a2_a6t_0_0.vlm"},
	{0x8ECF5C5F2B38954EULL, "w_encounter3_160_0.mmh"},
	{0x8ECF685F2B38A9E5ULL, "w_encounter3_160_0.mao"},
	{0x8ECF765F2B38C13CULL, "w_encounter3_160_0.msh"},
	{0x8ECF78C86A03065DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloorb.mao"},
	{0x8ECF7CC86A030C96ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloorb.mmh"},
	{0x8ECF8166A7DEFF9AULL, "hlac_w_encounter3_-262145_0z.dds"},
	{0x8ECF86C86A031D94ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainfloorb.msh"},
	{0x8ECFBC1F5BD99B6AULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.cess.evt"},
	{0x8ECFF2F4D2DB97ACULL, "vxlt_0_lowtown_gc7_0_0.vlm"},
	{0x8ECFF362A5452A00ULL, "6102412_m.fxe"},
	{0x8ED07EC79C4691D4ULL, "zz_and_debug6134125.cl"},
	{0x8ED134E9B5543EEDULL, "dae000_pc_home_ambients6208573.cl"},
	{0x8ED13A9964CB2954ULL, "vxlt_gallowstemplar_n_5uw_0_0.vlm"},
	{0x8ED1F70DAFEF5D71ULL, "vxw_sundermount_2tv_0_0.vlm"},
	{0x8ED219820B8A3BC1ULL, "390906703.ogg"},
	{0x8ED22CF58D0938E6ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredacti.ani"},
	{0x8ED252A0B49C0213ULL, "6076253_m.fxe"},
	{0x8ED26DD612229636ULL, "hlac_w_woundedcoast_7_0x.dds"},
	{0x8ED2FB439436563CULL, "vxca_caverns1_ebz_0_0.vlm"},
	{0x8ED2FE34123440E7ULL, "6033140_m.fxe"},
	{0x8ED37BAE6D6EA147ULL, "vxw_encounter2_camp_lk_0_0.vlm"},
	{0x8ED43C0B703DEFBAULL, "mag320_orsino6060238.cl"},
	{0x8ED473DA6D13540DULL, "vxlt_0_lowtown_a2_n_d6y_0_0.vlm"},
	{0x8ED4807F80319709ULL, "vxdr_0_deeproads_es2_0_1.vlm"},
	{0x8ED525B197B05323ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.anb"},
	{0x8ED5407F3BE9149BULL, "119146438.ogg"},
	{0x8ED56620D081B388ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeproofsml.phy"},
	{0x8ED59C6B971738EBULL, "vxdr_0_deeproads_ckc_0_0.vlm"},
	{0x8ED5CB4886A6E3FDULL, "static1_s2y.psh.fxo"},
	{0x8ED5FF0B72C726C0ULL, "lt_0_lowtown_a2_n_c_43_0.anb"},
	{0x8ED61B212B1CACB9ULL, "884599897.ogg"},
	{0x8ED630DB8BC32018ULL, "gen00fl_merrill6119636.cl"},
	{0x8ED69B1F5BDF81DDULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.cess.gad"},
	{0x8ED6DE39C89D02D7ULL, "6075601_m.fxe"},
	{0x8ED71176C87A1F16ULL, "vxlt_0_lowtown_n_f0n_0_0.vlm"},
	{0x8ED72F4806232103ULL, "6013369_f.fxe"},
	{0x8ED7408F1FF8FA34ULL, "hlac_w_blightlands_262144_0x.dds"},
	{0x8ED7461D7F574EF5ULL, "llac_lt_backalley_n_65537_8.dds"},
	{0x8ED79EC171E6C992ULL, "fx_glyph_ward.mao"},
	{0x8ED81A6F536370F8ULL, "6098962_m.fxe"},
	{0x8ED8314CBB2FCBB5ULL, "vxw_woundedcoast_xd_0_0.vlm"},
	{0x8ED87CEB3A92269BULL, "203908827.ogg"},
	{0x8ED893EB88DDE45EULL, "660346548.ogg"},
	{0x8ED8FE7A15B1F493ULL, "427175926.ogg"},
	{0x8ED994A65BBF2235ULL, "w_sundermount_52_0.phy"},
	{0x8ED9EF777AAF8C58ULL, "mcr370_guardsman6086986.cl"},
	{0x8EDA2C26FDF16DA6ULL, "vxlt_warehouse_n_129_0_0.vlm"},
	{0x8EDB24B98DF0D850ULL, "6216522_m.fxe"},
	{0x8EDB5659DBAA7EC7ULL, "physics1.vsh.fxo"},
	{0x8EDB87EFA02C21A6ULL, "808339410.ogg"},
	{0x8EDBC402F4F3C302ULL, "gen00fl_fenris6110862.cl"},
	{0x8EDBDC816D5BD400ULL, "593363883.ogg"},
	{0x8EDBE243E7FE57D3ULL, "438344894.ogg"},
	{0x8EDC9AEEFB45EB00ULL, "fx_smoke_vol2_blend.mao"},
	{0x8EDCC9FF2828DCD9ULL, "250655431.ogg"},
	{0x8EDCFF83047A301FULL, "vxlt_warehouse_n_14f_0_0.vlm"},
	{0x8EDD395F953B1A25ULL, "365884522.ogg"},
	{0x8EDE4F36DFC2D9BEULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkback.evt"},
	{0x8EDEE89B98D96D57ULL, "6135061_f.fxe"},
	{0x8EDEF7AF34ED880CULL, "548833155.ogg"},
	{0x8EDF30B53519C0A4ULL, "plt_cod_cha_fenris2.plo"},
	{0x8EDFE53C809105ACULL, "vxca_caverns2prc_98r_0_0.vlm"},
	{0x8EE06D55666E83EBULL, "vxht_chantry_2ca_0_0.vlm"},
	{0x8EE0802F2D4C03EEULL, "310001434.ogg"},
	{0x8EE083639981AAB4ULL, "6161055_m.fxe"},
	{0x8EE0B5A65BC57961ULL, "w_sundermount_52_0.rml"},
	{0x8EE138981C365B3EULL, "6047241_m.fxe"},
	{0x8EE17603E4015337ULL, "dre000_varric6036968.cl"},
	{0x8EE17F421DDA1D0CULL, "6119849_m.fxe"},
	{0x8EE17F78FA3935E8ULL, "gst_standing_left.gda"},
	{0x8EE1ED8B5F64B9C7ULL, "vxw_encounter2_camp_gj_0_0.vlm"},
	{0x8EE2AA6898A439CDULL, "zz_followers_banter6158262.cl"},
	{0x8EE302B197BC554CULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.mmh"},
	{0x8EE3709A07DECF0CULL, "6097185_m.fxe"},
	{0x8EE3ADD9F12DD7F1ULL, "mcr370_templar26087651.cl"},
	{0x8EE3BC1FABB77B4CULL, "vxlt_gallowstemplar_f_5as_0_0.vlm"},
	{0x8EE45EC3CE75CAC4ULL, "dae100_bran6144426.cl"},
	{0x8EE47EF7724713F1ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_mushroom_sa.dds"},
	{0x8EE4D8B21B131424ULL, "hlac_lt_gallowscourtyard_n_131071_0x.dds"},
	{0x8EE4F4BF81D436FEULL, "6108562_m.fxe"},
	{0x8EE533414F604D42ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax106a_0t.dds"},
	{0x8EE54162677E346BULL, "6153475_m.fxe"},
	{0x8EE56636DFC91FE9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkback.gad"},
	{0x8EE66F1BBD1BABF1ULL, "6105496_m.fxe"},
	{0x8EE6D7FF470858FFULL, "vxht_mansion2b_288_0_0.vlm"},
	{0x8EE6EFFC52B0FFC4ULL, "vxlt_0_lowtown_a2_n_c_csf_0_0.vlm"},
	{0x8EE715F58D1B6DE0ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredacti.gad"},
	{0x8EE74F998235084FULL, "mag291ip_barrel_red.utp"},
	{0x8EE7508A523DCCC8ULL, "6106886_f.fxe"},
	{0x8EE7E8516A04D51CULL, "201253801.ogg"},
	{0x8EE845DD7C7D977FULL, "582179436.ogg"},
	{0x8EE867E7305F91C1ULL, "vxlt_hangedman_b_0_0.vlm"},
	{0x8EE8F263FA06499DULL, "vxht_0_hightown_bgc_0_0.vlm"},
	{0x8EE929C8C0AFFAB8ULL, "hlac_lt_0_lowtown_n_65538_5z.dds"},
	{0x8EE92D6F751DE4EEULL, "hlac_lt_undercityserialkiller_-1_33x.dds"},
	{0x8EEA54BCEAA355BAULL, "llac_w_bonepit_0_11.dds"},
	{0x8EEA8F029C72595CULL, "304852026.ogg"},
	{0x8EEAA6E51CF21801ULL, "6040782_m.fxe"},
	{0x8EEADE657072CE33ULL, "follower_barks6153691.cl"},
	{0x8EEAF0D6746EFCA6ULL, "vxca_caverns2_80a_0_1.vlm"},
	{0x8EEB10951181D790ULL, "qcr000_arishok6109244.cl"},
	{0x8EEB3EAF32A22744ULL, "906971192.ogg"},
	{0x8EEB715FF670489FULL, "vxlt_warehouse_n_19a_0_0.vlm"},
	{0x8EEB784CDAD06585ULL, "6106923_m.fxe"},
	{0x8EEBA3D01B0B2FC3ULL, "6171383_m.fxe"},
	{0x8EEC3736DFCEEE37ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkback.ani"},
	{0x8EEC3E045A70C7B4ULL, "6153656_m.fxe"},
	{0x8EEC464BA1694772ULL, "355319923.ogg"},
	{0x8EEC8B508579FFC0ULL, "6041853_f.fxe"},
	{0x8EEC9385E0E6E1BAULL, "vxw_sundermount_2kh_0_0.vlm"},
	{0x8EECCD332E931814ULL, "vxht_0_hightown_c0v_0_0.vlm"},
	{0x8EED0C750E6DE3CFULL, "6087231_m.fxe"},
	{0x8EED555EED2CE3F4ULL, "6139670_m.fxe"},
	{0x8EEDB6D0FF06892BULL, "6154870_m.fxe"},
	{0x8EEDC87940BDCD92ULL, "lt_0_lowtown_a2_46_0.rml"},
	{0x8EEE1E5348B75A54ULL, "vxlt_gallowstemplar_n_5cu_0_0.vlm"},
	{0x8EEE6A959C15ED87ULL, "hla_ht_0_hightown_a2_-131074_0z.dds"},
	{0x8EEE7D8851F61611ULL, "llac_lt_0_lowtown_docks_a3_-3_0.dds"},
	{0x8EEEB391BBA4C5C0ULL, "478661158.ogg"},
	{0x8EEEF5408B77C334ULL, "6166155_m.fxe"},
	{0x8EEFDEFD7822873CULL, "vxca_caverns1_ehx_0_0.vlm"},
	{0x8EF018F3BBEFAA78ULL, "zz_and_debug6134077.cl"},
	{0x8EF097B40091963CULL, "hlac_ca_caverns2_5802_25z.dds"},
	{0x8EF0FB766B8C8023ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_7y.dds"},
	{0x8EF2085A3F0CEBF5ULL, "257053850.ogg"},
	{0x8EF26994B1E7D553ULL, "pro000_refugee_male6052102.cl"},
	{0x8EF282A2B600E15AULL, "vxlt_gallowstemplar_6qe_0_0.vlm"},
	{0x8EF315E4FBDC9670ULL, "1030174383.ogg"},
	{0x8EF37709EF458954ULL, "6065045_m.fxe"},
	{0x8EF3B33F59744CEBULL, "6156775_m.fxe"},
	{0x8EF4026E836B7A4CULL, "plt_prm000pt_fade_lion.plo"},
	{0x8EF41936EC1F2047ULL, "gen00fl_carver6122534.cl"},
	{0x8EF48D11F00CB948ULL, "6075963_m.fxe"},
	{0x8EF4D1490E1BFD0DULL, "87387365.ogg"},
	{0x8EF515069F1BFCB5ULL, "vxw_woundedcoast_rb_0_0.vlm"},
	{0x8EF5487D15AA5EADULL, "6061813_m.fxe"},
	{0x8EF552EE8E38ABB2ULL, "rdr110cr_dockworker2.utc"},
	{0x8EF569E74A2046BDULL, "gen_im_trs_krk_drk_junk2_04.uti"},
	{0x8EF56C064F213C29ULL, "vxw_woundedcoast_1fo_0_0.vlm"},
	{0x8EF62A42F7424391ULL, "6125450_m.fxe"},
	{0x8EF65AF7CEE9D61CULL, "vxw_sundermount_236_0_0.vlm"},
	{0x8EF661FF9BB1FA6FULL, "seb221cr_lady_harimann.utc"},
	{0x8EF699038AF956F4ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.cess.ani"},
	{0x8EF6A1E2BAC4E6A3ULL, "vxdr_primevaltomb_73y_0_0.vlm"},
	{0x8EF750927808CB90ULL, "hlac_ht_keep_n_4519_8y.dds"},
	{0x8EF76BFA1CEB58F8ULL, "vxlt_hangedman_n_lz_0_0.vlm"},
	{0x8EF7E6B6443E3A14ULL, "dae000_sketch6179011.cl"},
	{0x8EF822945E3A17CBULL, "6034692_f.fxe"},
	{0x8EF8777F5C506B23ULL, "vxlt_0_lowtown_n_el9_0_1.vlm"},
	{0x8EF904A8EC1FADF0ULL, "vxht_mansion2_3em_0_0.vlm"},
	{0x8EF9C54F1C357BDEULL, "6143116_m.fxe"},
	{0x8EFA3B8F44AB24E4ULL, "312232052.ogg"},
	{0x8EFAC55B9C4EE318ULL, "vxht_0_hightown_a2_bds_0_0.vlm"},
	{0x8EFAD147FEC92441ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.ani"},
	{0x8EFAE76E83716A83ULL, "plt_prm000pt_fade_lion.nss"},
	{0x8EFAFC7997EDB928ULL, "vxca_caverns2prc_92u_0_1.vlm"},
	{0x8EFB0562D2DCFA3AULL, "vxw_sundermount_2li_0_0.vlm"},
	{0x8EFB06F56335F304ULL, "dae000_bartrand6197640.cl"},
	{0x8EFB3F1020893094ULL, "vxht_0_hightown_c7w_0_0.vlm"},
	{0x8EFB6F85BA3EDD75ULL, "6076471_m.fxe"},
	{0x8EFBBB24DDB8597CULL, "w_woundedcoast_8_0.rml"},
	{0x8EFBDA9745F24B01ULL, "ave310_brennan6137010.cl"},
	{0x8EFC96038AFDBD56ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.cess.gad"},
	{0x8EFC966DD393464FULL, "258214467.ogg"},
	{0x8EFD0593E4251C8EULL, "llac_lt_0_lowtown_docks_-131073_4.dds"},
	{0x8EFD2B9786A85148ULL, "6078748_m.fxe"},
	{0x8EFD4D2DC350ED2BULL, "llac_ca_cave_2194_4.dds"},
	{0x8EFD510F4A5AB4EBULL, "vxht_chantry_2yc_0_0.vlm"},
	{0x8EFD55EE48FE5E1FULL, "6082179_m.fxe"},
	{0x8EFD6122A07B4AFDULL, "134743944.ogg"},
	{0x8EFEA22A62E99C78ULL, "32383636.ogg"},
	{0x8EFEC8C88DDDDB91ULL, "6120817_m.fxe"},
	{0x8EFEED0C15913AFBULL, "vxht_0_hightown_9ah_0_0.vlm"},
	{0x8EFF5B8F23AD5916ULL, "6090191_m.fxe"},
	{0x8EFF6F2AD9A2F069ULL, "6096874_f.fxe"},
	{0x8EFFA124DDBC15D1ULL, "w_woundedcoast_8_0.msh"},
	{0x8EFFAF24DDBC2DECULL, "w_woundedcoast_8_0.mao"},
	{0x8EFFBB24DDBC418FULL, "w_woundedcoast_8_0.mmh"},
	{0x8EFFD5C1EAAEEB86ULL, "dae010pt_amb_docks.ncs"},
	{0x8F005C59CE69721FULL, "vxht_0_hightown_a2_9wz_0_0.vlm"},
	{0x8F0079EE499CBF32ULL, "23790470.ogg"},
	{0x8F009FD98FA3AC4CULL, "vxlt_gallowstemplar_f_5oq_0_0.vlm"},
	{0x8F01531EE111E3D7ULL, "vxlt_gallowstemplar_f_4xz_0_0.vlm"},
	{0x8F019C05388F47B7ULL, "hlac_ht_0_hightown_a3_131071_0z.dds"},
	{0x8F01DBB809002D2AULL, "vxlt_gallowstemplar_f_3x4_0_0.vlm"},
	{0x8F01DC47FECF55E3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.gad"},
	{0x8F025D93C1ADF6FFULL, "smoke_ciga_f2_a.dds"},
	{0x8F0350DCE441046DULL, "vxw_bonepit_227_0_0.vlm"},
	{0x8F0377E24186119FULL, "529102358.ogg"},
	{0x8F03B1038B0409B5ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.cess.evt"},
	{0x8F044818465DCC58ULL, "llac_w_bonepitdestroyed_a3_196613_0.dds"},
	{0x8F04CA9F9482AF9EULL, "vxca_caverns1_mines_ii8_0_0.vlm"},
	{0x8F04CCD4C0DFEE9CULL, "vxw_sundermount_247_0_0.vlm"},
	{0x8F04CE692A683DBCULL, "w_woundedcoast_13_0.rml"},
	{0x8F04CFCBE6DFFE80ULL, "345906072.ogg"},
	{0x8F053006E2BF6708ULL, "mer120_hubert6059411.cl"},
	{0x8F0581ADFD0EFE5BULL, "hlac_lt_0_lowtown_docks_-131075_0y.dds"},
	{0x8F05D4A58FE274E8ULL, "299707992.ogg"},
	{0x8F06DA01DA930044ULL, "6087799_m.fxe"},
	{0x8F073BBB1F807C41ULL, "6108547_m.fxe"},
	{0x8F07B4692A6A4711ULL, "w_woundedcoast_13_0.msh"},
	{0x8F07C2692A6A5F2CULL, "w_woundedcoast_13_0.mao"},
	{0x8F07C686167A930BULL, "vxlt_undercityencounter_1lv_0_0.vlm"},
	{0x8F07CE692A6A73CFULL, "w_woundedcoast_13_0.mmh"},
	{0x8F07F24837A99C56ULL, "vxw_sundermount_21f_0_0.vlm"},
	{0x8F0882A5D8D1292BULL, "204038107.ogg"},
	{0x8F08A25C391A7CCEULL, "hairalpha_rendertoshadowmap_p0.psh.fxo.dx11.assembly"},
	{0x8F08B732DE6AF528ULL, "qun110_petrice6018646.cl"},
	{0x8F08B8E50A19DC1CULL, "832006878.ogg"},
	{0x8F08B947FED538C8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_exit.evt"},
	{0x8F08D71525FB41FEULL, "zz_qun_debug6028245.cl"},
	{0x8F08EBCCFCEC4A06ULL, "s_ballista_04_l2.mao"},
	{0x8F09773FC4D312BAULL, "vxw_sundermount_2qj_0_0.vlm"},
	{0x8F09AADDD8B08B82ULL, "hlac_lt_0_lowtown_docks_a3_-262143_0z.dds"},
	{0x8F0A2F8A7BB513C1ULL, "86557701.ogg"},
	{0x8F0B4A25F69D3A04ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_v1.ani"},
	{0x8F0B78EA6C244024ULL, "408592273.ogg"},
	{0x8F0B7E381DFBEBE2ULL, "w_woundedcoast_151_0.msh"},
	{0x8F0B80381DFBEE98ULL, "w_woundedcoast_151_0.mmh"},
	{0x8F0B8C381DFC033BULL, "w_woundedcoast_151_0.mao"},
	{0x8F0BA94C62574655ULL, "496906258.ogg"},
	{0x8F0BBFF9EE9A6A07ULL, "lgt150_nat_were_supporter6110951.cl"},
	{0x8F0BE92EF1C2D24EULL, "6085496_m.fxe"},
	{0x8F0C1A0511C12BFDULL, "isa211_isabela6202062.cl"},
	{0x8F0C49D8144408DCULL, "ave221_brennan6110484.cl"},
	{0x8F0C68729FB84C15ULL, "6095458_m.fxe"},
	{0x8F0C9183A12BCABCULL, "621736908.ogg"},
	{0x8F0CA3934BF015E8ULL, "6141660_m.fxe"},
	{0x8F0D84B7C208A416ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_004a_n.dds"},
	{0x8F0DD29447726D7AULL, "ts_kirk_guards_02.gda"},
	{0x8F0DF89D4B9121D6ULL, "lgt110_gxa_coa_meeting6117487.cl"},
	{0x8F0E4AF26E097EE0ULL, "mage_hexmisdirect.dds"},
	{0x8F0E4CF6FEBEC37CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_mash_0t.dds"},
	{0x8F0E833A07C2E17DULL, "vxdr_primevaltomb_531_0_0.vlm"},
	{0x8F0E97A55046E7AEULL, "vxlt_0_lowtown_docks_n_b0c_0_0.vlm"},
	{0x8F0F787F35BE0B5EULL, "758250764.ogg"},
	{0x8F0F8421AA6F7164ULL, "612921298.ogg"},
	{0x8F0F948FB7E77BF9ULL, "vxlt_0_lowtown_a2_n_f16_0_0.vlm"},
	{0x8F0FAB260736EA44ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobc_0t.dds"},
	{0x8F0FDB138834B93DULL, "llac_lt_0_lowtown_a2_n_-131073_3.dds"},
	{0x8F101856565ACB3DULL, "zz_openwhite_prep.cut"},
	{0x8F10307EA739694FULL, "6078163_m.fxe"},
	{0x8F107C6A9C3F840CULL, "one110_meeran6089884.cl"},
	{0x8F10C24450EFE47BULL, "vxht_chantry_n_26m_0_0.vlm"},
	{0x8F11055C4670CB1EULL, "hlac_w_bonepitmining_a2_131074_0x.dds"},
	{0x8F1169B9442396FDULL, "6136700_m.fxe"},
	{0x8F124725F6A35466ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_v1.gad"},
	{0x8F126334E94A519FULL, "zz_followers_banter6157812.cl"},
	{0x8F12803BD188D258ULL, "vxht_brothel_n_3v8_0_0.vlm"},
	{0x8F133EB1B2D6071CULL, "vxw_sundermount_294_0_0.vlm"},
	{0x8F1360CB5A16E234ULL, "6087566_m.fxe"},
	{0x8F1470F084C1BBC2ULL, "6126481_m.fxe"},
	{0x8F14B0D769716BE7ULL, "6053854_m.fxe"},
	{0x8F1587381E044F23ULL, "w_woundedcoast_151_0.phy"},
	{0x8F1587BAB3328D8DULL, "vxca_caverns1_mines_e57_0_0.vlm"},
	{0x8F164F3D6A10676CULL, "ser191_emeric6120464.cl"},
	{0x8F16775C3990E7FDULL, "398323007.ogg"},
	{0x8F16BBF630C4611DULL, "zz_mcr_debug6072384.cl"},
	{0x8F16DE0B72FE85E8ULL, "lt_0_lowtown_a2_n_c_43_0.rml"},
	{0x8F16FEB4DD16E5F3ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_action.ani"},
	{0x8F1706358644B368ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock2_n.dds"},
	{0x8F1720745D5B5D2DULL, "209539243.ogg"},
	{0x8F17308DB8E79588ULL, "art\\vfx\\materials\\vfx_shapes_add.mao"},
	{0x8F17329CF3F4433CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_sitting_female_loop.gad"},
	{0x8F178ABF17AD380CULL, "6080138_m.fxe"},
	{0x8F17DE03443CFE17ULL, "vxlt_0_lowtown_a3_n_c_hfd_0_0.vlm"},
	{0x8F17E91CB6C92B3AULL, "vxw_sundermount_2zk_0_0.vlm"},
	{0x8F181EDF9EF532C9ULL, "llac_ht_chantry_n_2534_34.dds"},
	{0x8F182C3808C9C61AULL, "zz_ave_debug6065037.cl"},
	{0x8F183997EDA9547CULL, "429538582.ogg"},
	{0x8F184225F6A7B745ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\attack_v1.evt"},
	{0x8F18502E368A147DULL, "6074042_m.fxe"},
	{0x8F1871DB02EA5FAEULL, "mag321_comtesse_de_launcet6064891.cl"},
	{0x8F193659910683F0ULL, "lt_0_lowtown_docks_n_226_0.mao"},
	{0x8F193A5991068A03ULL, "lt_0_lowtown_docks_n_226_0.mmh"},
	{0x8F1948599106A25DULL, "lt_0_lowtown_docks_n_226_0.msh"},
	{0x8F1988B3E1E1F2ECULL, "vxca_caverns2_9f9_0_0.vlm"},
	{0x8F19CCE0E14A4C58ULL, "zz_test_general6203151.cl"},
	{0x8F1A0679D1E47213ULL, "vxht_mansion2b_1gj_0_0.vlm"},
	{0x8F1A4AFA850ACA90ULL, "6004181_m.fxe"},
	{0x8F1AAD8319585617ULL, "vxw_encounter2_n9_0_0.vlm"},
	{0x8F1AE08C6EF15142ULL, "6084860_m.fxe"},
	{0x8F1B094A3DF86574ULL, "vxht_0_hightown_a3_b9q_0_0.vlm"},
	{0x8F1B3D1DECA41B03ULL, "423439240.ogg"},
	{0x8F1BA2104FC1AD40ULL, "zz_vault_debug6200351.cl"},
	{0x8F1BAB4B081BBBBCULL, "6164841_m.fxe"},
	{0x8F1C1A8DF430A3EBULL, "hlac_lt_0_lowtown_a2_n_c_65535_4y.dds"},
	{0x8F1C48ADEDAE7806ULL, "clf260_harley6093816.cl"},
	{0x8F1C708EBDB67B22ULL, "hlac_dr_primevaltomb_659_3y.dds"},
	{0x8F1D7762DC2690E7ULL, "lt_foundry_n_mmap.dds"},
	{0x8F1DBEB9994AB60EULL, "1003571731.ogg"},
	{0x8F1DFAFDDFDD03A4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_45basic.phy"},
	{0x8F1E10F169B9FD7DULL, "hlac_w_encounter3_-196607_0x.dds"},
	{0x8F1E319CF3FA603AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cuddle_sitting_female_loop.ani"},
	{0x8F1E4034CD8E9FBEULL, "213535447.ogg"},
	{0x8F1E422140D0B148ULL, "6161378_m.fxe"},
	{0x8F1EBBE2D7D24BC3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsblock.mao"},
	{0x8F1EBFE2D7D25210ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsblock.mmh"},
	{0x8F1ECDE2D7D26A2AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsblock.msh"},
	{0x8F1ED660CD83A4A5ULL, "follower_banter6163213.cl"},
	{0x8F1F675C78A3AA19ULL, "vxlt_backalley_g_1we_0_0.vlm"},
	{0x8F1F9C4D610A3D8EULL, "6085904_m.fxe"},
	{0x8F1FD7192A964031ULL, "vxht_mansion2b_2a9_0_0.vlm"},
	{0x8F2007B435630D1DULL, "lgt160_wk_soph_hunter6119759.cl"},
	{0x8F2055F2D3299A6EULL, "6102829_m.fxe"},
	{0x8F205A7E20BBC589ULL, "w_sundermount_188_0.phy"},
	{0x8F20F68C7439B048ULL, "w_woundedcoast_195_0.mmh"},
	{0x8F21028C7439C4EBULL, "w_woundedcoast_195_0.mao"},
	{0x8F21148C7439E372ULL, "w_woundedcoast_195_0.msh"},
	{0x8F2126EF4F90F76AULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.cess.gad"},
	{0x8F21A7A65BFCF850ULL, "w_sundermount_52_0.msh"},
	{0x8F21AD3526D553F6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_aoe.ani"},
	{0x8F21B1CF1F706BFAULL, "p_leather_stretcher_01s.dds"},
	{0x8F21B5A65BFD106AULL, "w_sundermount_52_0.mmh"},
	{0x8F21B9A65BFD16A1ULL, "w_sundermount_52_0.mao"},
	{0x8F21D54E6E34C4ACULL, "llac_lt_0_lowtown_docks_n_196608_1.dds"},
	{0x8F21D9F451672E40ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexcity_0.pwk"},
	{0x8F221A700D2C9B3CULL, "vxw_woundedcoast_7t_0_0.vlm"},
	{0x8F222D496B9BFE7DULL, "vxw_bonepit_p5_0_0.vlm"},
	{0x8F223DD1000312ADULL, "w_blightlands_mmap_ph.dds"},
	{0x8F2279956AC07134ULL, "822772127.ogg"},
	{0x8F23023173A228AEULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxbbunch.msh"},
	{0x8F23103173A240FFULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxbbunch.mao"},
	{0x8F231C3173A2549CULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxbbunch.mmh"},
	{0x8F234590123FF68DULL, "20931965.ogg"},
	{0x8F234809AC629AB0ULL, "175625402.ogg"},
	{0x8F237F4F589A1F04ULL, "233259066.ogg"},
	{0x8F23F6B4DD21622AULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_action.evt"},
	{0x8F23FC6585DCAC87ULL, "6043275_m.fxe"},
	{0x8F245C3119824F7DULL, "vxht_0_hightown_n_4ap_0_0.vlm"},
	{0x8F2480F5DB335D0BULL, "rock_03.dds"},
	{0x8F24C63C5D0D753BULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medd.msh"},
	{0x8F24D4D4F4AD6E7AULL, "art\\levels\\ztools\\default\\generic\\0ww_basic.phy"},
	{0x8F24D83C5D0D9386ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medd.mao"},
	{0x8F24E43C5D0DA825ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medd.mmh"},
	{0x8F254C3DCC793200ULL, "qun120_arishok_guard_nt.cnv"},
	{0x8F2573A63C76CF9BULL, "939203889.ogg"},
	{0x8F25A2C48E2AEB13ULL, "hlac_ca_cavern1_haunted_24426_9z.dds"},
	{0x8F26473C7EDB5080ULL, "mrl000_marethari6129276.cl"},
	{0x8F26C20B28802CFDULL, "6079342_m.fxe"},
	{0x8F2710B14B2317B8ULL, "6129767_m.fxe"},
	{0x8F271CDD823F0F94ULL, "ht_0_hightown_a3_12_0_ll.mmh"},
	{0x8F271D7BC157CDB2ULL, "634581513.ogg"},
	{0x8F274E1FA005100EULL, "art\\vfx\\combat\\vfx_r_ibl_appear_imp_p.impa.ani"},
	{0x8F27A65C48C432D1ULL, "vxw_bonepitmining_a2_ba_0_0.vlm"},
	{0x8F27BCA4DCEF6032ULL, "lgt210_mag_escp_rescued.cnv"},
	{0x8F27C8762E51CB76ULL, "6168601_m.fxe"},
	{0x8F27DD8E7066F01DULL, "hm_amb_conspiracy.mor"},
	{0x8F2809EF4F96E401ULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.cess.evt"},
	{0x8F28798B4DFDAA7EULL, "ave310_brennan6136782.cl"},
	{0x8F288686618BC5C1ULL, "w_bonepitmining_a2_20_0_ll.mmh"},
	{0x8F291F600B4E6E97ULL, "vxw_encounter2_i8_0_0.vlm"},
	{0x8F292E3DCC7CE838ULL, "qun120_arishok_guard_nt.bnk"},
	{0x8F29F1144C5108E3ULL, "6127502_m.fxe"},
	{0x8F2A6B59782BABECULL, "6173917_m.fxe"},
	{0x8F2AB25F48FAA1DCULL, "6149343_m.fxe"},
	{0x8F2AD5B4DD27489DULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_action.gad"},
	{0x8F2ADEA4DCF1CF8AULL, "lgt210_mag_escp_rescued.bnk"},
	{0x8F2AF4DAFEBC5BFDULL, "6168112_m.fxe"},
	{0x8F2B09B699E4939FULL, "6074457_m.fxe"},
	{0x8F2B7B0B11B661E2ULL, "w_sundermount_198_0.phy"},
	{0x8F2B8729B2EAC2ACULL, "vxht_chantry_n_301_0_0.vlm"},
	{0x8F2BB1F0A44BBB9FULL, "vxht_0_hightown_a2_9xg_0_0.vlm"},
	{0x8F2BBA69A8B74042ULL, "231001937.ogg"},
	{0x8F2BF1033247253CULL, "demo_montage.cut"},
	{0x8F2CAAAE8DAADFF8ULL, "vxlt_0_lowtown_fj9_0_0.vlm"},
	{0x8F2CB3FE32C6C9C8ULL, "6166278_m.fxe"},
	{0x8F2D141931C739D7ULL, "428899065.ogg"},
	{0x8F2D19182F90FAB2ULL, "873248037.ogg"},
	{0x8F2D301F3A972642ULL, "6071828_m.fxe"},
	{0x8F2D3B4665C781E7ULL, "vxlt_0_lowtown_docks_n_ct3_0_0.vlm"},
	{0x8F2D3E1959DEBEE1ULL, "hlac_w_bonepitdestroyed_a3_65539_0y.dds"},
	{0x8F2D473D727EAC09ULL, "zz_followers_banter6159306.cl"},
	{0x8F2D69FD3AE78370ULL, "gen00fl_anders6125703.cl"},
	{0x8F2DBC1093458ABFULL, "6140255_m.fxe"},
	{0x8F2DFAD54D796AD6ULL, "8006387.ogg"},
	{0x8F2E65E7B95B7C50ULL, "ht_bartrand_n_3365.rml"},
	{0x8F2EA8B725A1B485ULL, "6119582_f.fxe"},
	{0x8F2EB53526DFEBE7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_aoe.evt"},
	{0x8F2EC7CFC51FB2EEULL, "6107729_m.fxe"},
	{0x8F2F20275C00FD9DULL, "816520500.ogg"},
	{0x8F2F337A16F99E29ULL, "vxw_woundedcoast_1rs_0_0.vlm"},
	{0x8F2F40D375DE1924ULL, "fen310_fenris6118669.cl"},
	{0x8F2F5E48D6D91F76ULL, "vxlt_0_lowtown_a2_n_c_gyz_0_0.vlm"},
	{0x8F2F846963CBA99FULL, "w_bonepitdestroyed_a3_207_0.msh"},
	{0x8F2F8A6963CBB441ULL, "w_bonepitdestroyed_a3_207_0.mmh"},
	{0x8F2F966963CBC8E2ULL, "w_bonepitdestroyed_a3_207_0.mao"},
	{0x8F2FAAE39884313EULL, "hlac_dr_dungeon_3216_5y.dds"},
	{0x8F3007E46E5B0797ULL, "vxlt_undercitydungeon_1mx_0_0.vlm"},
	{0x8F305AA31C868E2EULL, "vxca_caverns2_952_0_0.vlm"},
	{0x8F31314DFB992A47ULL, "vxlt_0_lowtown_a3_n_c_gy2_0_0.vlm"},
	{0x8F3138AA1E2CC8AAULL, "6020549_m.fxe"},
	{0x8F318F12E246AD98ULL, "vxlt_0_lowtown_a2_hdq_0_0.vlm"},
	{0x8F329E093E630479ULL, "6054259_m.fxe"},
	{0x8F32AABB992B7FBBULL, "vis121_saemus6071798.cl"},
	{0x8F32CC7A41E2930BULL, "art\\vfx\\combat\\vfx_h_cb_tal_threaten.impa.evt"},
	{0x8F3334F6BF3FD5C6ULL, "dae000cr_grand_cleric6179581.cl"},
	{0x8F337488E64399B9ULL, "vxlt_foundry_n_vw_0_0.vlm"},
	{0x8F33DEE377DA5353ULL, "6162088_m.fxe"},
	{0x8F340915028CCD34ULL, "hlac_ca_caverns1_24421_2x.dds"},
	{0x8F34102502422ACBULL, "6016036_f.fxe"},
	{0x8F34461FA00F8D5FULL, "art\\vfx\\combat\\vfx_r_ibl_appear_imp_p.impa.evt"},
	{0x8F344FAFE6569CF9ULL, "art\\characters\\playercharacter\\qunarimale\\uni_mag.msh"},
	{0x8F3459AFE656ADE7ULL, "art\\characters\\playercharacter\\qunarimale\\uni_mag.mmh"},
	{0x8F349F3532544C36ULL, "519250459.ogg"},
	{0x8F34E5D5700AA957ULL, "1067787713.ogg"},
	{0x8F356FAA616CA43EULL, "vxdr_primevalentrance_26b_0_0.vlm"},
	{0x8F358907E5F10C22ULL, "mcr370_meredith6085950.cl"},
	{0x8F35C5254804303FULL, "vxht_0_hightown_ap3_0_0.vlm"},
	{0x8F35D63526E64230ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_aoe.gad"},
	{0x8F3601EF4FA31450ULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.cess.ani"},
	{0x8F3619FC48FA155AULL, "zz_and_debug6134066.cl"},
	{0x8F3684D4F2E267B4ULL, "6078414_m.fxe"},
	{0x8F373D54CBD6E42CULL, "6107572_m.fxe"},
	{0x8F37A251B3CC9171ULL, "6152180_m.fxe"},
	{0x8F3848FC2A6E3BFEULL, "vxht_0_hightown_cbw_0_1.vlm"},
	{0x8F388D276F544266ULL, "361744762.ogg"},
	{0x8F38F0C6C8A683A6ULL, "hlac_dr_primevalentrance_1724_1y.dds"},
	{0x8F390C0F9737D8BBULL, "vxw_sundermount_30y_0_0.vlm"},
	{0x8F3964CEFC309B4DULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_strip_n.dds"},
	{0x8F398EF4A2E5BE74ULL, "6117249_m.fxe"},
	{0x8F39E57A41E8DCCCULL, "art\\vfx\\combat\\vfx_h_cb_tal_threaten.impa.gad"},
	{0x8F3AAF285423D05AULL, "lt_0_lowtown_docks_n_98_0.mmh"},
	{0x8F3AB3285423D691ULL, "lt_0_lowtown_docks_n_98_0.mao"},
	{0x8F3AC1285423EEA0ULL, "lt_0_lowtown_docks_n_98_0.msh"},
	{0x8F3B04216CEB868DULL, "llac_lt_0_lowtown_a2_n_65537_1.dds"},
	{0x8F3B271FA0157748ULL, "art\\vfx\\combat\\vfx_r_ibl_appear_imp_p.impa.gad"},
	{0x8F3BEEB9F8D540DFULL, "dae000_bartrand6027595.cl"},
	{0x8F3C2647CE5C0FB3ULL, "430571302.ogg"},
	{0x8F3C4B165C8FDB19ULL, "vxlt_backalley_g_1yk_0_0.vlm"},
	{0x8F3CBF51165B1DE9ULL, "805464161.ogg"},
	{0x8F3CC93721BFC746ULL, "llac_ca_caverns1_24421_6.dds"},
	{0x8F3D1EAC4C325C1CULL, "vxw_encounter2_5c_0_0.vlm"},
	{0x8F3DDA70BB85350BULL, "526784829.ogg"},
	{0x8F3E55C583AE57DFULL, "793944393.ogg"},
	{0x8F3E74DE7A051F60ULL, "6117916_f.fxe"},
	{0x8F3EE45E5E693D5AULL, "gen_im_trs_cot_dua_dag_r0.uti"},
	{0x8F3F23A9DA2E2256ULL, "art\\vfx\\materials\\vfx_sleep_z_add.mao"},
	{0x8F3F32F8C2240AD8ULL, "6120777_m.fxe"},
	{0x8F3FCE424B58F5EFULL, "545134911.ogg"},
	{0x8F407B13816DA493ULL, "llac_lt_0_lowtown_a2_n_c_-4_0.dds"},
	{0x8F407B4B811968C5ULL, "zz_mcr_debug6071529.cl"},
	{0x8F40E47A41EEF9CAULL, "art\\vfx\\combat\\vfx_h_cb_tal_threaten.impa.ani"},
	{0x8F413BF07612C83DULL, "hlac_ht_0_hightown_n_-65536_4y.dds"},
	{0x8F41E56A6F305AF4ULL, "232019223.ogg"},
	{0x8F41EDFCEC18FCE3ULL, "zz_vault_debug6203162.cl"},
	{0x8F4273EFE248B928ULL, "zz_mag_debug6053400.cl"},
	{0x8F428F2E5C509074ULL, "408681171.ogg"},
	{0x8F42A300E5786890ULL, "hlac_lt_gallowscourtyard_n_-65535_2z.dds"},
	{0x8F42A7DE9921D79BULL, "705168938.ogg"},
	{0x8F436E138144A071ULL, "6075984_m.fxe"},
	{0x8F4374AF2DC35D55ULL, "474394549.ogg"},
	{0x8F437FB15CF6C292ULL, "vxht_0_hightown_brw_0_0.vlm"},
	{0x8F4399F5DB4E38EDULL, "rock_03.mao"},
	{0x8F43D0972ACE826DULL, "6080979_m.fxe"},
	{0x8F4434C7AFDDE521ULL, "vxlt_0_lowtown_docks_a3_n_cv4_0_0.vlm"},
	{0x8F443DFB1449C112ULL, "6107692_f.fxe"},
	{0x8F444057E656153FULL, "vxca_caverns2_9f6_0_0.vlm"},
	{0x8F44E8BE5D7B4D7BULL, "50944027.ogg"},
	{0x8F44F192FB390015ULL, "mag321_comte_de_launcet6064929.cl"},
	{0x8F4533F1954512FCULL, "zz_bet_debug6113393.cl"},
	{0x8F455C10A7C6BB93ULL, "vxht_mansion2b_1po_0_0.vlm"},
	{0x8F458AB94F27AD7EULL, "6042729_m.fxe"},
	{0x8F45D6FAB5E09984ULL, "vxlt_undercitydungeon_1j7_0_1.vlm"},
	{0x8F461E5747E05EF3ULL, "vxca_caverns1_j2c_0_1.vlm"},
	{0x8F4695AF98CA75B6ULL, "zz_dae_debug6050741.cl"},
	{0x8F4789A01CC9CFF4ULL, "6133603_m.fxe"},
	{0x8F47FDC9B9754315ULL, "zz_fight_qunari.ncs"},
	{0x8F48532744E39FA9ULL, "1047330566.ogg"},
	{0x8F48BF7FC5735EF3ULL, "gen00fl_aveline6137513.cl"},
	{0x8F48D060C6706332ULL, "112219802.ogg"},
	{0x8F48FB3173C2EC67ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxbbunch.phy"},
	{0x8F4914C5F5716495ULL, "6142497_m.fxe"},
	{0x8F498AD0668AA63DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masf_0d.dds"},
	{0x8F4A03A547D048E9ULL, "595355845.ogg"},
	{0x8F4A24CD08ECF21AULL, "185066950.ogg"},
	{0x8F4AB4A0D96DE73EULL, "vxht_0_hightown_a2_by6_0_0.vlm"},
	{0x8F4ACF515E9BA387ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\right_down.gad"},
	{0x8F4B2235B7ACD435ULL, "6080922_m.fxe"},
	{0x8F4B73861446FECFULL, "6156656_m.fxe"},
	{0x8F4B754EF4BB50FBULL, "hlac_lt_hangedman_-1_0z.dds"},
	{0x8F4BA61C255DC393ULL, "38301473.ogg"},
	{0x8F4C437E20E0D9AEULL, "w_sundermount_188_0.mmh"},
	{0x8F4C4598CCC61B6FULL, "vxca_caverns1_mines_cjz_0_0.vlm"},
	{0x8F4C4F7E20E0EE45ULL, "w_sundermount_188_0.mao"},
	{0x8F4C5D7E20E1059CULL, "w_sundermount_188_0.msh"},
	{0x8F4CF122563DB213ULL, "vxlt_0_lowtown_a2_n_gnx_0_0.vlm"},
	{0x8F4CF951DA330436ULL, "86019901.ogg"},
	{0x8F4D3FEEBE095BAFULL, "llac_w_encounter2_-131070_0.dds"},
	{0x8F4E06C9B1861BF9ULL, "dae000_meredith6085306.cl"},
	{0x8F4E1D23FA2E57EBULL, "626616927.ogg"},
	{0x8F4E7ECCD1335091ULL, "633997305.ogg"},
	{0x8F4E991C628AE855ULL, "hlac_lt_warehouse_65536_1z.dds"},
	{0x8F4EACB1663C559DULL, "hlac_lt_0_lowtown_docks_a3_-2_2y.dds"},
	{0x8F4EB5FCB4F22256ULL, "vxw_encounter2_33_0_0.vlm"},
	{0x8F4EDCFED1FB6457ULL, "mag321_comtesse_de_launcet6043402.cl"},
	{0x8F4F35B4C9610A1CULL, "97548540.ogg"},
	{0x8F4F5D6F628C6242ULL, "89557107.ogg"},
	{0x8F4F6FA0365CB3EAULL, "6074385_f.fxe"},
	{0x8F4FC0892BB511A3ULL, "w_encounter3_89_0.mao"},
	{0x8F4FC4892BB518F0ULL, "w_encounter3_89_0.mmh"},
	{0x8F4FD2892BB5300AULL, "w_encounter3_89_0.msh"},
	{0x8F50560B11D53E96ULL, "w_sundermount_198_0.mao"},
	{0x8F50620B11D55335ULL, "w_sundermount_198_0.mmh"},
	{0x8F50640B11D556ABULL, "w_sundermount_198_0.msh"},
	{0x8F512BB778944EF2ULL, "vxw_bonepitmining_a2_7m_0_0.vlm"},
	{0x8F519A8B9936F8F0ULL, "hlac_w_bonepitmining_a2_196611_0x.dds"},
	{0x8F51C4515EA1AF94ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\right_down.evt"},
	{0x8F5216F930442AC5ULL, "dae000cr_grand_cleric6148104.cl"},
	{0x8F522A6148C9F339ULL, "6105953_m.fxe"},
	{0x8F525F5C8D3D5C75ULL, "hlac_ht_brothel_n_3369_1x.dds"},
	{0x8F52639636877E04ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\idle.gad"},
	{0x8F527A09B53ED093ULL, "llac_lt_0_lowtown_n_-131073_3.dds"},
	{0x8F532D7BF05C6E12ULL, "222475035.ogg"},
	{0x8F5361BB86186DEFULL, "951448912.ogg"},
	{0x8F540BEF7363873DULL, "hlac_ca_caverns1_24425_9x.dds"},
	{0x8F555C617F8955C9ULL, "117330280.ogg"},
	{0x8F55707BFF49A3D9ULL, "vxw_blightlands_a5_0_0.vlm"},
	{0x8F5575291D6D80CFULL, "vrc221st_bartrand_fake.stg"},
	{0x8F55823AE1C7D711ULL, "6166860_m.fxe"},
	{0x8F55930FCF0D3D11ULL, "hlac_w_sundermount_-65542_0y.dds"},
	{0x8F55E2327CFEE3DBULL, "vxca_cavern1_haunted_cke_0_0.vlm"},
	{0x8F5683666A28139EULL, "180502972.ogg"},
	{0x8F57B49557D44F43ULL, "ser221_gamlen6135075.cl"},
	{0x8F580944BE895460ULL, "vxlt_0_lowtown_f1p_0_0.vlm"},
	{0x8F597296368DB6E2ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\idle.ani"},
	{0x8F5A49E164253EEBULL, "535702966.ogg"},
	{0x8F5A55C25A62C19CULL, "vxw_blightlands_gj_0_0.vlm"},
	{0x8F5A66E7B980B8B8ULL, "ht_bartrand_n_3365.anb"},
	{0x8F5A6A0CCFA70BF0ULL, "6035941_m.fxe"},
	{0x8F5A727EF8FD9964ULL, "84067917.ogg"},
	{0x8F5A809E43292FD3ULL, "477687693.ogg"},
	{0x8F5AFF44200F30ABULL, "follower_barks6153196.cl"},
	{0x8F5B1B663D9C5D91ULL, "vxlt_0_lowtown_a2_n_ded_0_0.vlm"},
	{0x8F5B4523F33F2BA6ULL, "1022235009.ogg"},
	{0x8F5B78F93487DEEEULL, "vxca_caverns1_gk5_0_0.vlm"},
	{0x8F5B9C2132434617ULL, "783159059.ogg"},
	{0x8F5BE186483A8021ULL, "one100dg_kelder6198472.cl"},
	{0x8F5BF4BD3374607DULL, "vxw_bonepit_l1_0_0.vlm"},
	{0x8F5CFDD04CB130B6ULL, "vxht_mansion2b_34p_0_0.vlm"},
	{0x8F5D20BBD35DA4FFULL, "follower_banter6160863.cl"},
	{0x8F5D87A3FF4703CFULL, "hlac_dr_0_deeproads_5845_12z.dds"},
	{0x8F5DBEA3F292720AULL, "zz_followers_banter6157288.cl"},
	{0x8F5DCE66AA651215ULL, "vxlt_0_lowtown_a2_ep0_0_1.vlm"},
	{0x8F5DF0FD2C02C1C5ULL, "ht_keep_n_3736.anb"},
	{0x8F5E6F8E8226E10CULL, "mag100_cullen6140953.cl"},
	{0x8F5EAA944B30B6D4ULL, "6006284_m.fxe"},
	{0x8F5EBF51AB002558ULL, "361903767.ogg"},
	{0x8F5ECEACC8BF01E6ULL, "vxw_woundedcoast_13c_0_0.vlm"},
	{0x8F5EEEC20636E3C0ULL, "ch2emarmtnt.psh.fxo"},
	{0x8F5F473823FF5909ULL, "vxlt_darktown_58c_0_1.vlm"},
	{0x8F5F715AD318CF0CULL, "6172767_m.fxe"},
	{0x8F5FA78CAAAD6B11ULL, "hla_ht_0_hightown_a2_65535_0y.dds"},
	{0x8F5FBC515EADDFCDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\right_down.ani"},
	{0x8F5FC6E2D809F52BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsblock.phy"},
	{0x8F6042AFE67BC11CULL, "art\\characters\\playercharacter\\qunarimale\\uni_mag.phy"},
	{0x8F606047EFD35DF1ULL, "framebuffereffects_screentexrect_p0.psh.fxo.dx11.assembly"},
	{0x8F60636B40E2F392ULL, "vxht_0_hightown_bxq_0_0.vlm"},
	{0x8F609659962672B0ULL, "235750275.ogg"},
	{0x8F60D743586B05D7ULL, "260658398.ogg"},
	{0x8F60F0C61718324CULL, "vxht_bartrand_n_3uu_0_0.vlm"},
	{0x8F60F108F3732993ULL, "zz_dae_debug6049977.cl"},
	{0x8F610945DFDD77C2ULL, "mer120_hubert6017815.cl"},
	{0x8F61BB38D5476B16ULL, "1039738416.ogg"},
	{0x8F61E323DE207ECAULL, "vxlt_undercityserialkiller_15t_0_0.vlm"},
	{0x8F620F89639F35BAULL, "sst_ss_fenris.gda"},
	{0x8F6242D3FC0653BAULL, "plt_lgt000pt_civ_hounds.nss"},
	{0x8F628794C0A14BEFULL, "6092501_m.fxe"},
	{0x8F62F6BF13A2CF69ULL, "llac_lt_0_lowtown_-196608_1.dds"},
	{0x8F630165A3C1D4F5ULL, "sst_ss_karras.gda"},
	{0x8F631D4673CDBC19ULL, "hlac_w_encounter2_65538_0x.dds"},
	{0x8F6381BCCA929D1AULL, "countcreatures.ncs"},
	{0x8F639B7D1BC98103ULL, "vxdr_0_deeproads_ebb_0_0.vlm"},
	{0x8F63C2EE0A804647ULL, "hlac_w_encounter3_0_4z.dds"},
	{0x8F63CA1175915AABULL, "6032033_m.fxe"},
	{0x8F63E258F13FBC59ULL, "vxw_blightlands_j4_0_0.vlm"},
	{0x8F645D60081AA246ULL, "hlac_w_woundedcoast_-196606_0z.dds"},
	{0x8F646AA0A3DA47D4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobc_0d.dds"},
	{0x8F6487BB5655685BULL, "vxht_keep_n_702_0_0.vlm"},
	{0x8F64E806BD794D84ULL, "cod_tomes_tome_4.utp"},
	{0x8F64E9AFD86FFE2BULL, "gen_im_cft_rec_bmb_03.uti"},
	{0x8F65243A34CD0FE4ULL, "dae000cr_grand_cleric6147649.cl"},
	{0x8F65629748FBC56BULL, "lgt200_eau_tome6208291.cl"},
	{0x8F65905D66207BA3ULL, "hlac_lt_0_lowtown_n_-65539_0x.dds"},
	{0x8F65A594E24AFF79ULL, "zz_and_debug6134180.cl"},
	{0x8F66678030B93646ULL, "6139728_m.fxe"},
	{0x8F66833000AA4D59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_reaction.ani"},
	{0x8F6767C8107DA622ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneya_0.pwk"},
	{0x8F6860636679688AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action.gad"},
	{0x8F687F2ABE1B18BCULL, "mcr371_angry_mage_m6087726.cl"},
	{0x8F68C79F4C58DA1CULL, "vxw_blightlands_hk_0_0.vlm"},
	{0x8F690DD9BD4AC333ULL, "ave200_aveline_bar6116189.cl"},
	{0x8F693CB3232E565FULL, "6038651_m.fxe"},
	{0x8F696B7931F748A6ULL, "gen00fl_aveline6132768.cl"},
	{0x8F6A3494AD347B18ULL, "bee_cyc.dds"},
	{0x8F6A75F53C793AE7ULL, "fen210_fenris6109443.cl"},
	{0x8F6AE293228812F7ULL, "vxw_sundermount_30m_0_0.vlm"},
	{0x8F6AF19368186887ULL, "zz_one_debug6051439.cl"},
	{0x8F6B5A96D8AEFA9AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothc.mmh"},
	{0x8F6B5E96D8AF01D1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothc.mao"},
	{0x8F6B6C96D8AF19E0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothc.msh"},
	{0x8F6B8B9CB87252D0ULL, "vxht_0_hightown_a3_bwg_0_0.vlm"},
	{0x8F6BB9B9EB23C01FULL, "669150332.ogg"},
	{0x8F6BBD3C5D4968D2ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medd.phy"},
	{0x8F6C6A429ED98C40ULL, "dae000_arishok6078180.cl"},
	{0x8F6CEB8E04638138ULL, "4427043.ogg"},
	{0x8F6D5E3000B02C9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_reaction.gad"},
	{0x8F6D78646FDD986BULL, "191829777.ogg"},
	{0x8F6D983D6912FD10ULL, "542145487.ogg"},
	{0x8F6DCFCC2C77BF2BULL, "mcr370_cullen6087426.cl"},
	{0x8F6E0F715C807FF2ULL, "vxw_bonepitmining_a2_9o_0_0.vlm"},
	{0x8F6E4CF6CC772C44ULL, "lgt210pt_mistress_selby.ncs"},
	{0x8F6E7831E52C86DEULL, "lt_0_lowtown_docks_a3_n_38_0.rml"},
	{0x8F6E7ACF847E6B70ULL, "vxw_encounter3_bm_0_0.vlm"},
	{0x8F6E912FE1B5E7E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v2.gad"},
	{0x8F6F4363667F5621ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action.evt"},
	{0x8F702FEF619B0EB5ULL, "vxca_caverns1_eib_0_0.vlm"},
	{0x8F7057EDB1E31A36ULL, "vxht_store_c6_0_0.vlm"},
	{0x8F7094381FADD625ULL, "871103245.ogg"},
	{0x8F70DD4A2FBE98DEULL, "6143325_m.fxe"},
	{0x8F7134183BF41287ULL, "863997290.ogg"},
	{0x8F716FCD60365FA8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_nocollision.mmh"},
	{0x8F717BCD6036744BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_nocollision.mao"},
	{0x8F718DCD603692D2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_nocollision.msh"},
	{0x8F71A644908C4BA4ULL, "6039691_m.fxe"},
	{0x8F728E354708BE76ULL, "lt_0_lowtown_docks_a3_n_182_0.mao"},
	{0x8F729A354708D215ULL, "lt_0_lowtown_docks_a3_n_182_0.mmh"},
	{0x8F729C354708D58BULL, "lt_0_lowtown_docks_a3_n_182_0.msh"},
	{0x8F72DD8C5E407839ULL, "6111682_m.fxe"},
	{0x8F7318D4F7AC49A6ULL, "art\\characters\\creatures\\spider\\animation\\dblow_blunt.evt"},
	{0x8F747B3000B67C80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\console_reaction.evt"},
	{0x8F74BF98408A57A8ULL, "vxlt_0_lowtown_a2_df3_0_0.vlm"},
	{0x8F75047CA9C9D122ULL, "gen00fl_merrill6128667.cl"},
	{0x8F753CF6069C8F78ULL, "vxlt_hangedman_1tb_0_0.vlm"},
	{0x8F75564E98D433BAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\blendtree\\doorsimple.anb"},
	{0x8F75C712A0AAFD53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_chest_nervous.ani"},
	{0x8F76597AF845123DULL, "650816335.ogg"},
	{0x8F765A4A87E18655ULL, "vxht_mansion_n_4xk_0_0.vlm"},
	{0x8F769753F2BD17FBULL, "hlac_ht_mansion2_818_56z.dds"},
	{0x8F76C9034F19B02AULL, "6034715_m.fxe"},
	{0x8F77397C3E4EF29CULL, "vxw_blightlands_mh_0_0.vlm"},
	{0x8F7810D807769905ULL, "vxht_mansion_n_2k4_0_0.vlm"},
	{0x8F782AE14801E77AULL, "gen00fl_anders6119508.cl"},
	{0x8F78707E483A467FULL, "6034509_m.fxe"},
	{0x8F78AFE9DEC095C9ULL, "6176498_m.fxe"},
	{0x8F78E5D4E73F4DA1ULL, "origin_screenshots.gda"},
	{0x8F792FD4F7B0DCF1ULL, "art\\characters\\creatures\\spider\\animation\\dblow_blunt.gad"},
	{0x8F7967ADCD0E17A3ULL, "x_narrativeroom_0_ll.mmh"},
	{0x8F79CB1A634603F8ULL, "w_sundermount_163_0.phy"},
	{0x8F79D100295B3EFDULL, "sib131_slave_a6165255.cl"},
	{0x8F79F577C84BD32EULL, "6167961_m.fxe"},
	{0x8F79FA87FEC9E855ULL, "vxlt_gallowsprison_n_6mi_0_0.vlm"},
	{0x8F7A35584C5FF9F3ULL, "vxca_caverns1_mines_e8n_0_0.vlm"},
	{0x8F7A83406D4F2C62ULL, "6187817_m.fxe"},
	{0x8F7AC451EBCF5D21ULL, "plt_cod_hst_orz_deeproads.plo"},
	{0x8F7B04E52DA7FFDAULL, "vxlt_0_lowtown_a2_n_eir_0_0.vlm"},
	{0x8F7B0AA380A9875CULL, "6061099_f.fxe"},
	{0x8F7B187CE8EE9327ULL, "526048378.ogg"},
	{0x8F7B1884AFED61FEULL, "gen00fl_anders6189372.cl"},
	{0x8F7B70A88132029CULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_lp.ani"},
	{0x8F7B79FE104DDFE2ULL, "llac_lt_0_lowtown_docks_a3_-4_4.dds"},
	{0x8F7B7AC563962972ULL, "one121_trap.cnv"},
	{0x8F7BAC4CF13E919CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.gad"},
	{0x8F7BB266ACAB32E6ULL, "vxw_woundedcoast_19e_0_0.vlm"},
	{0x8F7BC78C59132119ULL, "6182282_m.fxe"},
	{0x8F7C16684FAF00B7ULL, "6167087_m.fxe"},
	{0x8F7C3B636689D370ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action.ani"},
	{0x8F7C8203E3C081FBULL, "seb221_ruxton_harimann6197715.cl"},
	{0x8F7C82BBE3AD0E5FULL, "6123656_m.fxe"},
	{0x8F7D58222DAB376FULL, "qun211ip_gas_effect1.ncs"},
	{0x8F7D75088305FBFBULL, "vxlt_hangedman_n_1q1_0_0.vlm"},
	{0x8F7D7B1433A269BCULL, "vxw_bonepitmining_a2_1jg_0_0.vlm"},
	{0x8F7D9B148B84ECA6ULL, "6114941_m.fxe"},
	{0x8F7DA2BB3059802AULL, "6130957_f.fxe"},
	{0x8F7DCCF5AC1D0A50ULL, "249874387.ogg"},
	{0x8F7DF87BBB4EC5F4ULL, "6165015_m.fxe"},
	{0x8F7DFD3305F293D6ULL, "206605058.ogg"},
	{0x8F7E72452C273720ULL, "hlac_ht_0_hightown_65533_3z.dds"},
	{0x8F7E79742F3EA6B2ULL, "hlac_lt_0_lowtown_a2_n_1_5z.dds"},
	{0x8F7E9CC5639898CAULL, "one121_trap.bnk"},
	{0x8F7EA873A532343CULL, "vxlt_backalley_g_1y2_0_0.vlm"},
	{0x8F7F353FE95211B6ULL, "vxlt_0_lowtown_docks_a3_n_cod_0_0.vlm"},
	{0x8F7F3918102AC1A7ULL, "vxlt_hangedman_n_285_0_0.vlm"},
	{0x8F7F74C51B6A7CFAULL, "hlac_lt_gallowstemplar_f_2388_5y.dds"},
	{0x8F7FCDD4F4FA0DA3ULL, "art\\levels\\ztools\\default\\generic\\0ww_basic.msh"},
	{0x8F7FDBD4F4FA25FDULL, "art\\levels\\ztools\\default\\generic\\0ww_basic.mmh"},
	{0x8F7FDFD4F4FA2C0EULL, "art\\levels\\ztools\\default\\generic\\0ww_basic.mao"},
	{0x8F7FEAAF2B560208ULL, "156433266.ogg"},
	{0x8F8018F7954102A0ULL, "6060150_m.fxe"},
	{0x8F8030D4F7B6FC9FULL, "art\\characters\\creatures\\spider\\animation\\dblow_blunt.ani"},
	{0x8F807F36FFB5B203ULL, "vxdr_0_deeproads_ehd_0_0.vlm"},
	{0x8F80ADCB597D8BABULL, "6038633_m.fxe"},
	{0x8F81AAB9946DBA81ULL, "hlac_w_bonepitmining_a2_393220_0z.dds"},
	{0x8F81C122D78259CBULL, "zz_isa_debug6133163.cl"},
	{0x8F82677618B9FB95ULL, "hlac_lt_foundry_n_320_19z.dds"},
	{0x8F827DA8813837DEULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_lp.gad"},
	{0x8F82AB4CF144AE9AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.ani"},
	{0x8F82B61CC077304CULL, "393203646.ogg"},
	{0x8F82BC2FE1C6D9C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_lp_v2.ani"},
	{0x8F82BF12A0B5798AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_chest_nervous.evt"},
	{0x8F82CAFF3DF2BCA7ULL, "lt_gallowscourtyard_n_41_0.rml"},
	{0x8F8326BA6978FFA8ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flmb_0d.dds"},
	{0x8F83584FA67BEF19ULL, "714586844.ogg"},
	{0x8F842928D4AF8644ULL, "hlac_dr_primevaltomb_664_5y.dds"},
	{0x8F843A5E6FA45E36ULL, "gen00fl_isabela6135702.cl"},
	{0x8F84675AC730D920ULL, "vxca_caverns1_fml_0_0.vlm"},
	{0x8F847EF749844825ULL, "zz_dae_debug6029335.cl"},
	{0x8F857786D65F6AB4ULL, "vxht_mansion2b_3dr_0_0.vlm"},
	{0x8F858C4DF3B936FEULL, "421178012.ogg"},
	{0x8F85AB5930450B1CULL, "vxw_blightlands_vi_0_0.vlm"},
	{0x8F85B01C8D03DDADULL, "one300_zevran6076934.cl"},
	{0x8F85D354A4DB3010ULL, "341825120.ogg"},
	{0x8F85E755D3EF08BAULL, "mer120_hubert6070980.cl"},
	{0x8F8644D53167E4EDULL, "6041969_m.fxe"},
	{0x8F8649660DF3C606ULL, "one300_zevran6075570.cl"},
	{0x8F86875E8B1A0840ULL, "face_rendertoshadowmap_p1.vsh.fxo.dx11"},
	{0x8F86A825B3FA1B29ULL, "hlac_ht_chantry_n_2534_10z.dds"},
	{0x8F87A0EBCD12C2CCULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v1.ani"},
	{0x8F87C1DD2ADF25C0ULL, "lgt310_mag_melson6152200.cl"},
	{0x8F87E7D7FC570CC4ULL, "llac_lt_0_lowtown_-1_3.dds"},
	{0x8F87F491987554E4ULL, "hlac_lt_0_lowtown_docks_a3_65532_1y.dds"},
	{0x8F8831C5BB9307D4ULL, "vxht_0_hightown_a3_c94_0_0.vlm"},
	{0x8F8833D9B4923412ULL, "art\\vfx\\materials\\chunky\\vfs_grenade_bottle.mao"},
	{0x8F886F569C5E83D0ULL, "vxht_0_hightown_a3_bie_0_0.vlm"},
	{0x8F8891F24E145F98ULL, "vxlt_0_lowtown_docks_ctg_0_0.vlm"},
	{0x8F894CBA82F9E309ULL, "576699753.ogg"},
	{0x8F8958A8813E17FDULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_lp.evt"},
	{0x8F895B73C3F2E39CULL, "art\\characters\\playercharacter\\elffemale\\arm_lgtp.mao"},
	{0x8F896773C3F2F83FULL, "art\\characters\\playercharacter\\elffemale\\arm_lgtp.mmh"},
	{0x8F896D73C3F302E1ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgtp.msh"},
	{0x8F8984409D8D6557ULL, "vxht_keep_67a_0_0.vlm"},
	{0x8F899E12A0BB60FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_chest_nervous.gad"},
	{0x8F89F30CB95078D6ULL, "vxw_sundermount_2bo_0_0.vlm"},
	{0x8F8A098D9DF3E83AULL, "mag330_walter6044519.cl"},
	{0x8F8A9BA9717028A4ULL, "small_sa.dds"},
	{0x8F8AA12E0DD642D0ULL, "llac_lt_0_lowtown_a2_n_c_-1_2.dds"},
	{0x8F8B0AFB38DD5B58ULL, "hlac_lt_0_lowtown_docks_n_-131075_7z.dds"},
	{0x8F8B1BCD018C1909ULL, "124772288.ogg"},
	{0x8F8B2666E2F3A714ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_004a_0s.dds"},
	{0x8F8B5233D90FCA8EULL, "6202734_m.fxe"},
	{0x8F8B74ADC98DDB1FULL, "6195882_m.fxe"},
	{0x8F8BDDFD4722E352ULL, "llac_w_encounter3_-393212_0.dds"},
	{0x8F8BF282962C1344ULL, "mer110_jansen6032593.cl"},
	{0x8F8CC113E6FEDF1BULL, "plt_cod_itm_staff_violation.plo"},
	{0x8F8D05124F5DF8D9ULL, "215220193.ogg"},
	{0x8F8DCD508E9DB9C4ULL, "zz_followers_banter6159402.cl"},
	{0x8F8EADEBCD18F70EULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v1.gad"},
	{0x8F8EBDCCA69C9251ULL, "lt_gallowscourtyard_mmap.dds"},
	{0x8F8EDC0678CD3B19ULL, "vxw_bonepitmining_a2_29x_0_0.vlm"},
	{0x8F8EF0200F69774CULL, "vxlt_gallowstemplar_n_5is_0_0.vlm"},
	{0x8F8EFE58DF42EC97ULL, "6119540_f.fxe"},
	{0x8F8F119749F5E833ULL, "749367825.ogg"},
	{0x8F8F37EFC72205D9ULL, "vxw_blightlands_u9_0_0.vlm"},
	{0x8F8FBE9306CCBDDAULL, "724873005.ogg"},
	{0x8F8FCEEE1031CC7DULL, "vxca_caverns2prc_9fy_0_0.vlm"},
	{0x8F9003C19785BE51ULL, "785097894.ogg"},
	{0x8F907996D8CE4BA5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothc.phy"},
	{0x8F90B367AED398F8ULL, "vxca_caverns1_mines_flv_0_0.vlm"},
	{0x8F90F33C4545847BULL, "zz_mer_debug6023997.cl"},
	{0x8F9111B9AB6A9CE4ULL, "vxlt_undercitydungeon_1ak_0_0.vlm"},
	{0x8F913D6CD3A38897ULL, "911747927.ogg"},
	{0x8F91B5AF0B5A6DA5ULL, "qun220cr_javaris6071704.cl"},
	{0x8F9255CF4A746547ULL, "6170289_m.fxe"},
	{0x8F927AE36BF71470ULL, "vxht_0_hightown_n_bxd_0_0.vlm"},
	{0x8F92BDAF5CEA54CCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_goodbye_aveline.ani"},
	{0x8F92DE1C81BFD4B2ULL, "6197856_m.fxe"},
	{0x8F92FB718FC9FCA2ULL, "6145095_m.fxe"},
	{0x8F93256D8BBF2C34ULL, "hlac_lt_0_lowtown_-65537_4z.dds"},
	{0x8F932B16BF0B5158ULL, "llac_lt_0_lowtown_65535_7.dds"},
	{0x8F93911BDCA227FAULL, "hlac_ht_mansion2b_818_62z.dds"},
	{0x8F9408E8988FFF33ULL, "15901890.ogg"},
	{0x8F9428449EE14DD6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_middle.phy"},
	{0x8F94B01C8760D3BCULL, "zz_ave_debug6134016.cl"},
	{0x8F9564F65C590DADULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgtp.mmh"},
	{0x8F9568F65C5914FEULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgtp.mao"},
	{0x8F9576F65C592C33ULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgtp.msh"},
	{0x8F959DFDBED144DDULL, "ser221_party_comments6166201.cl"},
	{0x8F95F3C9075FA25CULL, "627977482.ogg"},
	{0x8F96928FEFE52924ULL, "547117885.ogg"},
	{0x8F96A3A2AD892054ULL, "vxht_0_hightown_a3_c25_0_0.vlm"},
	{0x8F96FCC285295BC0ULL, "hlac_w_encounter1_-131071_0y.dds"},
	{0x8F9716B0D9510FB5ULL, "vxw_bonepit_11a_0_0.vlm"},
	{0x8F97E89F119430DAULL, "vxlt_0_lowtown_a2_n_ewl_0_0.vlm"},
	{0x8F97F1FD2C343BCDULL, "ht_keep_n_3736.rml"},
	{0x8F980B64A91FF1FBULL, "775437073.ogg"},
	{0x8F98A5750C2B3FA4ULL, "gen00fl_fenris6113911.cl"},
	{0x8F990E679AEF8D7EULL, "art\\effects\\cloudshadows\\w_encounter3\\cloud2.dds"},
	{0x8F993DDA6ACAE01AULL, "hlac_lt_0_lowtown_n_65538_5x.dds"},
	{0x8F99CAAF5CF0890EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_goodbye_aveline.gad"},
	{0x8F9A58C266F22CFBULL, "vxlt_hangedman_n_1k3_0_0.vlm"},
	{0x8F9B45222DC41412ULL, "qun211ip_gas_effect1.utp"},
	{0x8F9C00A31AE9E3EEULL, "vxlt_0_lowtown_docks_a3_d01_0_0.vlm"},
	{0x8F9C18F9CD3E42B6ULL, "vxlt_0_lowtown_docks_a3_n_cuf_0_0.vlm"},
	{0x8F9C57FDEF82BFB7ULL, "823550563.ogg"},
	{0x8F9C8A686D124A48ULL, "mag200_feynriel6030425.cl"},
	{0x8F9D8EB2D19805C5ULL, "6127250_m.fxe"},
	{0x8F9DB1C85F6A27C6ULL, "vxlt_0_lowtown_n_db6_0_0.vlm"},
	{0x8F9DE53FE21C8B06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_inward_grasp.ani"},
	{0x8F9E1211126CE7BAULL, "hla_ht_0_hightown_n_65538_0x.dds"},
	{0x8F9E25DC64D1C490ULL, "gen_im_wep_rog_arc_start.uti"},
	{0x8F9E8A4DAFB4F647ULL, "z_cloud02_offsetup.dds"},
	{0x8F9EB732B8710949ULL, "dae210_gate_guard6085140.cl"},
	{0x8F9EE62AB1F30FE5ULL, "framebuffereffects_vsmblur_p0.psh.fxo.dx11.assembly"},
	{0x8F9F11449EEB2E77ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_middle.msh"},
	{0x8F9F217BE2BCAA0EULL, "hlac_w_blightlands_131070_0y.dds"},
	{0x8F9F23449EEB4CDAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_middle.mao"},
	{0x8F9F27449EEB53E9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_middle.mmh"},
	{0x8F9FC5AF5CF4ECEDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_goodbye_aveline.evt"},
	{0x8FA0A1AB22FB3931ULL, "6103124_m.fxe"},
	{0x8FA1090F5FDA25E3ULL, "6127313_m.fxe"},
	{0x8FA10F8AEB228A56ULL, "18830071.ogg"},
	{0x8FA1C2E3F1DDE42EULL, "hlac_w_woundedcoast_0_9z.dds"},
	{0x8FA1E24F02B02567ULL, "6128305_m.fxe"},
	{0x8FA26B61F6A522B5ULL, "vxw_woundedcoast_6v_0_0.vlm"},
	{0x8FA2C2739FE52273ULL, "250323350.ogg"},
	{0x8FA2F16974289A24ULL, "6002882_m.fxe"},
	{0x8FA3005B3636843AULL, "664147181.ogg"},
	{0x8FA3DD0E7443D9B5ULL, "528286514.ogg"},
	{0x8FA40C42F272CAF8ULL, "clf221_jalen6070378.cl"},
	{0x8FA40D12C2880317ULL, "dae000_bodhan_letter6144284.cl"},
	{0x8FA411F8E359C8CDULL, "vxht_0_hightown_n_c35_0_0.vlm"},
	{0x8FA4235CABBDF064ULL, "256269624.ogg"},
	{0x8FA47F1AA2124BA9ULL, "vxlt_darktown_5q5_0_0.vlm"},
	{0x8FA4A2AA376DD741ULL, "608057223.ogg"},
	{0x8FA4DC9CDD4E2288ULL, "445301726.ogg"},
	{0x8FA50EEF906BA98EULL, "561178260.ogg"},
	{0x8FA55310804AB4D0ULL, "vxht_0_hightown_a3_bcc_0_0.vlm"},
	{0x8FA555B7BFA5B516ULL, "vxlt_0_lowtown_a2_n_c_i0k_0_0.vlm"},
	{0x8FA65A7C038A495AULL, "vxlt_0_lowtown_a2_n_exm_0_0.vlm"},
	{0x8FA65DF65C67756AULL, "art\\characters\\playercharacter\\elfmale\\bdy_lgtp.phy"},
	{0x8FA68E75BAACB5C8ULL, "dae100_gamlen_letter6146989.cl"},
	{0x8FA7A6A19C0B7000ULL, "zz_followers_banter6158214.cl"},
	{0x8FA7DA30F36B0EEDULL, "108250353.ogg"},
	{0x8FA84C1554946986ULL, "vxht_keep_n_732_0_0.vlm"},
	{0x8FA87531E55DF982ULL, "lt_0_lowtown_docks_a3_n_38_0.anb"},
	{0x8FA8BAAC1675A976ULL, "tbe100_xenon6190979.cl"},
	{0x8FA9C1AE0EB63BF7ULL, "lt_undercityserialkiller_3567.anb"},
	{0x8FAA06B6651CA435ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\prp_barkburn_s.dds"},
	{0x8FAA72800CDFFC6EULL, "vxlt_0_lowtown_docks_a3_d52_0_0.vlm"},
	{0x8FAB42CE012A2BBBULL, "6152810_m.fxe"},
	{0x8FABCD3FE228A077ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_inward_grasp.evt"},
	{0x8FABF8B42DF8C208ULL, "face_face_p9.psh.fxo.dx11"},
	{0x8FACD9955545B200ULL, "vxdr_0_deeproads_cpm_0_0.vlm"},
	{0x8FAD3BC4C3661C47ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\disengage.gad"},
	{0x8FAD3F1107E19A50ULL, "129080248.ogg"},
	{0x8FAD7557AC7F7B7CULL, "6201584_m.fxe"},
	{0x8FAD7D5A32470B80ULL, "29499469.ogg"},
	{0x8FADE21276FB3D0FULL, "llac_ca_caverns1_24423_5.dds"},
	{0x8FAE0AB9EF0F040BULL, "31204186.ogg"},
	{0x8FAF8A6EC0BF0CD8ULL, "579261738.ogg"},
	{0x8FAFD9024654A17FULL, "hf_qfq_mob2.mor"},
	{0x8FB085484A2FDF42ULL, "lt_0_lowtown_docks_a3_105_0.phy"},
	{0x8FB0DD3EE89B3B35ULL, "vxw_woundedcoast_3u_0_0.vlm"},
	{0x8FB139F7352AC305ULL, "6082421_intro.cut"},
	{0x8FB1CE3FE22D0D00ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_merril_inward_grasp.gad"},
	{0x8FB1FBB210482959ULL, "6114469_m.fxe"},
	{0x8FB21E16154ACB10ULL, "535492910.ogg"},
	{0x8FB239AEEAA3CB2FULL, "gen00fl_anders6116083.cl"},
	{0x8FB24D2983F1ECFFULL, "mcr370_varric6098330.cl"},
	{0x8FB283D5D54FE14DULL, "vxht_0_hightown_n_c84_0_0.vlm"},
	{0x8FB31C677498B9A4ULL, "dre000_carver6037586.cl"},
	{0x8FB398576CC2D524ULL, "973706365.ogg"},
	{0x8FB3DD80B4C00DCBULL, "vxlt_0_lowtown_a2_n_c_df1_0_0.vlm"},
	{0x8FB3EA0E064C6DF7ULL, "glow4_a.dds"},
	{0x8FB430C4C36C2854ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\disengage.evt"},
	{0x8FB45082A7D31DD3ULL, "art\\characters\\creatures\\spider\\animation\\cwkfr.evt"},
	{0x8FB469C435F76E77ULL, "vxlt_0_lowtown_docks_a3_d22_0_0.vlm"},
	{0x8FB49EC610E09BA7ULL, "vxlt_ruinedhouse_n_kz_0_0.vlm"},
	{0x8FB4EFE261AD4D0AULL, "6070092_m.fxe"},
	{0x8FB5910DEE154E4FULL, "gen00fl_bethany6102884.cl"},
	{0x8FB5CB0F1A95ABECULL, "6095608_m.fxe"},
	{0x8FB5EACC74395B80ULL, "ss_deeproadcombat.cut"},
	{0x8FB6355C04342C8BULL, "character_character_p37.psh.fxo.dx11"},
	{0x8FB637883E544B5FULL, "787090907.ogg"},
	{0x8FB693538AE2E9A6ULL, "hlac_lt_0_lowtown_a2_65535_3z.dds"},
	{0x8FB6D523329F6BB4ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_open.gad"},
	{0x8FB71C299A5A6D09ULL, "6057566_m.fxe"},
	{0x8FB7645F041C82CAULL, "art\\vfx\\textures\\vfx_burnt_tile.dds"},
	{0x8FB7BAE30A0BCFB1ULL, "t3_gsp_b03.tnt"},
	{0x8FB813BA44480BBBULL, "966520348.ogg"},
	{0x8FB819CE16B851D8ULL, "34938814.ogg"},
	{0x8FB85DC2940A0FDAULL, "60421726.ogg"},
	{0x8FB890061CBE1E00ULL, "hlac_lt_gallowscourtyard_-65536_1y.dds"},
	{0x8FB8E02978ADE77CULL, "6110789_m.fxe"},
	{0x8FB90F4CA6E0C730ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masb_0n.dds"},
	{0x8FB9A862DBDF945FULL, "vxlt_0_lowtown_a2_n_cmm_0_0.vlm"},
	{0x8FBA4A85DC9A5C13ULL, "one280_iwan6096901.cl"},
	{0x8FBAB85534715C11ULL, "lt_shop.arl"},
	{0x8FBB4982A7D93014ULL, "art\\characters\\creatures\\spider\\animation\\cwkfr.gad"},
	{0x8FBB94E5EAC812A9ULL, "t3_ogr_h03.tnt"},
	{0x8FBC6D7EE516F55BULL, "vxlt_0_lowtown_a2_n_c_e86_0_0.vlm"},
	{0x8FBCE42332A3F092ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_open.ani"},
	{0x8FBDD07A41D95DF0ULL, "vxht_0_hightown_n_bic_0_0.vlm"},
	{0x8FBDD9045540DCB6ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_1x.dds"},
	{0x8FBDE16EC9948528ULL, "6075484_m.fxe"},
	{0x8FBEB988B76D3DDEULL, "gen00fl_aveline6114423.cl"},
	{0x8FBEDB0B971A1971ULL, "vxlt_foundry_n_vo_0_0.vlm"},
	{0x8FBEF29385D385CBULL, "vxw_sundermount_23u_0_0.vlm"},
	{0x8FBF285CC0F2C89EULL, "125641027.ogg"},
	{0x8FBF84240E70D8C7ULL, "817102950.ogg"},
	{0x8FBF8DE8DA19D27DULL, "6052086_m.fxe"},
	{0x8FC09EA60FE98756ULL, "mcr370_assignments6081695.cl"},
	{0x8FC128C4C376A48DULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\disengage.ani"},
	{0x8FC14D061D0474BFULL, "897462388.ogg"},
	{0x8FC1AA9EC762EC77ULL, "lt_0_lowtown_a2_n_28_0_ll.mmh"},
	{0x8FC1EF764194485DULL, "vxlt_gallowstemplar_n_4sq_0_0.vlm"},
	{0x8FC233154333CCF1ULL, "vxlt_gallowstemplar_f_5rg_0_0.vlm"},
	{0x8FC25882A7DF68F2ULL, "art\\characters\\creatures\\spider\\animation\\cwkfr.ani"},
	{0x8FC26A6F0C113D4DULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_htb.mmh"},
	{0x8FC2EF88DC608E52ULL, "6114475_m.fxe"},
	{0x8FC30C9EE9B4F983ULL, "vxlt_darktown_580_0_0.vlm"},
	{0x8FC361BF53A3658AULL, "6075392_m.fxe"},
	{0x8FC375A6D6AE0629ULL, "859817788.ogg"},
	{0x8FC41D8E48773343ULL, "ser221_gascard6122238.cl"},
	{0x8FC51BE5D19A212DULL, "ht_0_hightown_n_9_0.mmh"},
	{0x8FC51FE5D19A287EULL, "ht_0_hightown_n_9_0.mao"},
	{0x8FC52DE5D19A3FB3ULL, "ht_0_hightown_n_9_0.msh"},
	{0x8FC53F2E432314B1ULL, "qun160_vashothlead6053875.cl"},
	{0x8FC5E9ED67C90427ULL, "932743711.ogg"},
	{0x8FC5F3EC059312C3ULL, "554725409.ogg"},
	{0x8FC61C5C16931580ULL, "vxht_chantry_2il_0_0.vlm"},
	{0x8FC666E2907BFF1AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16ma.mmh"},
	{0x8FC66AE2907C0651ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16ma.mao"},
	{0x8FC678E2907C1E60ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16ma.msh"},
	{0x8FC6D0C872B0F241ULL, "6085258_f.fxe"},
	{0x8FC6EAF1490DEB73ULL, "llac_ht_chantry_n_2534_36.dds"},
	{0x8FC794112F204972ULL, "hlac_ht_keep_n_3732_5z.dds"},
	{0x8FC85E637C707BCDULL, "vxw_woundedcoast_oy_0_0.vlm"},
	{0x8FC85FD3FC5D382DULL, "plt_lgt000pt_civ_hounds.plo"},
	{0x8FC89EA9CF8A16B1ULL, "6176076_m.fxe"},
	{0x8FC89F7B9827D1F8ULL, "61805068.ogg"},
	{0x8FC8ECCCCFECBC79ULL, "1071411679.ogg"},
	{0x8FC95E9D3C1C3664ULL, "gen00fl_fenris6127075.cl"},
	{0x8FC9D03F6E628D8FULL, "mcr310_party_banter6087822.cl"},
	{0x8FC9DCF3534D7D90ULL, "zz_fen_debug6166702.cl"},
	{0x8FC9E2E7BD2BBBDFULL, "6060433_f.fxe"},
	{0x8FCA47A1BAA3DEA7ULL, "196855852.ogg"},
	{0x8FCA991D792F6ACBULL, "llac_lt_gallowstemplar_2393_0.dds"},
	{0x8FCAB5556CDBF06CULL, "667944963.ogg"},
	{0x8FCAB6429BA55B0DULL, "gen00fl_dog6117186.cl"},
	{0x8FCAD79993107F37ULL, "llac_ca_cave_2194_13.dds"},
	{0x8FCB9AFE82476389ULL, "6146989_m.fxe"},
	{0x8FCBA8727A84B589ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_spire.phy"},
	{0x8FCBEB45220E7D4AULL, "hlac_w_bonepit_0_0z.dds"},
	{0x8FCD057833A61955ULL, "vxht_mansion_n_4fu_0_0.vlm"},
	{0x8FCD647077C99E4BULL, "vxw_sundermount_24t_0_0.vlm"},
	{0x8FCD89B962F0EA1FULL, "zz_followers_banter6157319.cl"},
	{0x8FCD8A81EE1E33AAULL, "hlac_lt_0_lowtown_a2_65536_5y.dds"},
	{0x8FCE1DB95E95C6CCULL, "vxlt_0_lowtown_docks_n_bc8_0_0.vlm"},
	{0x8FCE4E2FEE5B00D5ULL, "6018897_m.fxe"},
	{0x8FCEC9134EDB96EAULL, "636196300.ogg"},
	{0x8FCF1BE5D1A275A6ULL, "ht_0_hightown_n_9_0.rml"},
	{0x8FCF615DF0B7F231ULL, "vxht_0_hightown_a3_a6b_0_0.vlm"},
	{0x8FCFC59BAE1DA9A3ULL, "gen00fl_aveline6053350.cl"},
	{0x8FCFEB8991FD1EE6ULL, "189477769.ogg"},
	{0x8FD05CD248576C12ULL, "w_bonepitmining_a2_29_0_ll.mmh"},
	{0x8FD06048DF652879ULL, "vxlt_gallowstemplar_4yl_0_0.vlm"},
	{0x8FD06564A57E7023ULL, "6161139_m.fxe"},
	{0x8FD0818A5B4A61AEULL, "53154789.ogg"},
	{0x8FD1098CF60B0FA0ULL, "qun200_party_comment6095430.cl"},
	{0x8FD127AE71E59C65ULL, "vxlt_0_lowtown_a2_n_emu_0_0.vlm"},
	{0x8FD13A72F1A79781ULL, "and230_mage_girl6135831.cl"},
	{0x8FD15586A4940E36ULL, "579191529.ogg"},
	{0x8FD1FFC9E4C9B6B2ULL, "dae000_pc_home_ambients6208688.cl"},
	{0x8FD213076D308F15ULL, "el_lp_middle_dusk.mmh"},
	{0x8FD24DCE33D9B473ULL, "vxlt_hangedman_20j_0_0.vlm"},
	{0x8FD2C33C4F79A089ULL, "nat320_delilah6065932.cl"},
	{0x8FD2CB984D86F164ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_board.msh"},
	{0x8FD2DD984D870FCDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_board.mao"},
	{0x8FD2E1984D871606ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_board.mmh"},
	{0x8FD33D4725AFDF1EULL, "vxlt_0_lowtown_a2_n_ep1_0_0.vlm"},
	{0x8FD38B2078192D49ULL, "389571310.ogg"},
	{0x8FD3C6DBE41793C4ULL, "vxw_bonepitmining_a2_21q_0_0.vlm"},
	{0x8FD48E3908892E00ULL, "vxht_chantry_2bm_0_0.vlm"},
	{0x8FD49F8DB2918ABCULL, "209549137.ogg"},
	{0x8FD4AF11BB552DACULL, "dae111ip_at_tavern_ltown.utp"},
	{0x8FD4B81A63928E8DULL, "w_sundermount_163_0.msh"},
	{0x8FD4C61A6392A6C0ULL, "w_sundermount_163_0.mao"},
	{0x8FD4CA1A6392AD13ULL, "w_sundermount_163_0.mmh"},
	{0x8FD4E536F76EB26BULL, "495427279.ogg"},
	{0x8FD54921481AF685ULL, "vxlt_gallowsprison_n_5wd_0_0.vlm"},
	{0x8FD560484A4EBCF6ULL, "lt_0_lowtown_docks_a3_105_0.mao"},
	{0x8FD56C484A4ED095ULL, "lt_0_lowtown_docks_a3_105_0.mmh"},
	{0x8FD56E484A4ED40BULL, "lt_0_lowtown_docks_a3_105_0.msh"},
	{0x8FD56EB63587222AULL, "dae000ip_crowd_mf1_talk7.utp"},
	{0x8FD5EE8A6289ADC0ULL, "vxht_mansion2b_3e5_0_0.vlm"},
	{0x8FD614E5D1A888EAULL, "ht_0_hightown_n_9_0.phy"},
	{0x8FD67951E97C3D5BULL, "mag310ar_alienage_nt.ncs"},
	{0x8FD693715FD5AA81ULL, "art\\vfx\\combat\\vfx_m_cre_haste_act_c.anb"},
	{0x8FD6BEAE0EDD240BULL, "lt_undercityserialkiller_3567.rml"},
	{0x8FD6DED2A733BA06ULL, "llac_lt_0_lowtown_docks_a3_n_196608_0.dds"},
	{0x8FD7123A13801C1AULL, "mrl310_merrill6198100.cl"},
	{0x8FD792D2A9B12475ULL, "mcr310_argument6074385.cl"},
	{0x8FD79C2C5F8E0AAFULL, "hlac_ca_caverns1_24425_0y.dds"},
	{0x8FD7AEB79864CB3FULL, "qun220cr_javaris6071719.cl"},
	{0x8FD7DD01E1FD287CULL, "hlac_lt_0_lowtown_a2_65537_5x.dds"},
	{0x8FD88CAC822036EDULL, "vxlt_darktown_2yn_0_0.vlm"},
	{0x8FD89DDF3FD64B5EULL, "357677470.ogg"},
	{0x8FD89F684FFF360CULL, "hlac_ht_chantry_n_2534_33x.dds"},
	{0x8FD937867E71AA9DULL, "23895123.ogg"},
	{0x8FD94B0A65391764ULL, "vxlt_undercitydungeon_1xl_0_0.vlm"},
	{0x8FD9787E468439B0ULL, "plt_cod_mgc_chant_schism.plo"},
	{0x8FD9788136B2CCF4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_enter.ani"},
	{0x8FD9A8E6023621CDULL, "6060003_m.fxe"},
	{0x8FD9C2FFE94760E3ULL, "vxdr_0_deeproads_bcz_0_0.vlm"},
	{0x8FD9D22382918562ULL, "llac_ht_brothel_n_2995_2.dds"},
	{0x8FDA1E89BD7454C9ULL, "hlac_w_encounter2_262142_0x.dds"},
	{0x8FDAFC2FF334920AULL, "vxlt_undercityencounter_114_0_0.vlm"},
	{0x8FDB37971A9913DFULL, "art\\characters\\playercharacter\\humanmale\\hlh_medh.msh"},
	{0x8FDB3D971A991D81ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medh.mmh"},
	{0x8FDB49971A993222ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medh.mao"},
	{0x8FDBCC2F32FDEC85ULL, "6080200_f.fxe"},
	{0x8FDBEF75EC337E84ULL, "04_tendrils.mao"},
	{0x8FDC521BB91D12EFULL, "280530636.ogg"},
	{0x8FDCD8984D8F5BD1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_board.phy"},
	{0x8FDD8DC7EE7325BBULL, "vxlt_undercityencounter_1j4_0_0.vlm"},
	{0x8FDF16CF271FFDF1ULL, "vxlt_gallowstemplar_f_5xe_0_0.vlm"},
	{0x8FDF56D093EB97AEULL, "6054829_m.fxe"},
	{0x8FDF58BE9447F8A5ULL, "6085688_m.fxe"},
	{0x8FDF6383615C6AC0ULL, "llac_lt_0_lowtown_a2_-2_2.dds"},
	{0x8FDF758136B73356ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_enter.gad"},
	{0x8FDF9E9F34E710E1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_8.pwk"},
	{0x8FE0403F14076A23ULL, "984548324.ogg"},
	{0x8FE0F5A7AF39FCF8ULL, "vxw_encounter2_camp_7i_0_0.vlm"},
	{0x8FE15884CDDAEB30ULL, "6051587_m.fxe"},
	{0x8FE1849E4A066B28ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_3y.dds"},
	{0x8FE194167D56E110ULL, "vxw_bonepitdestroyed_a3_l8_0_0.vlm"},
	{0x8FE1A984D7451F1BULL, "follower_banter6170101.cl"},
	{0x8FE1AFA7D9599A42ULL, "6104144_m.fxe"},
	{0x8FE1DE5433044F1AULL, "350451251.ogg"},
	{0x8FE25BA1DDFA7A4CULL, "6093677_m.fxe"},
	{0x8FE277BA59A5FEDAULL, "6121017_m.fxe"},
	{0x8FE2AB29C9DFAAA8ULL, "vxlt_gallowstemplar_f_6sn_0_0.vlm"},
	{0x8FE2C1300315BF8DULL, "dae000_madam_lusine6152243.cl"},
	{0x8FE2C4534F219942ULL, "hlac_ht_bartrand_n_3372_0z.dds"},
	{0x8FE42B64B3808E50ULL, "vxca_cavern1_haunted_io0_0_0.vlm"},
	{0x8FE474715FE1B3FAULL, "art\\vfx\\combat\\vfx_m_cre_haste_act_c.mmh"},
	{0x8FE50A9525866FC9ULL, "165656271.ogg"},
	{0x8FE5421D605CACCDULL, "vxw_woundedcoast_aw_0_0.vlm"},
	{0x8FE5BBFFDB3E660CULL, "108039742.ogg"},
	{0x8FE5E7A6CCACBD8CULL, "fen110_fenris6098901.cl"},
	{0x8FE60469F4DA7FA5ULL, "hlac_lt_0_lowtown_n_-65535_5y.dds"},
	{0x8FE61281768F41C3ULL, "vxw_bonepitdestroyed_a3_58_0_0.vlm"},
	{0x8FE620C3C6DCB660ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinswall.phy"},
	{0x8FE68158F6F18F49ULL, "hlac_w_bonepitdestroyed_a3_393220_0z.dds"},
	{0x8FE6908136BD7FB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\weapon_draw_ready_enter.evt"},
	{0x8FE6BA72589643B3ULL, "hlac_ca_caverns2_5802_58y.dds"},
	{0x8FE733996B05687EULL, "vxlt_0_lowtown_docks_a3_cy9_0_0.vlm"},
	{0x8FE746DB18E6B101ULL, "601613354.ogg"},
	{0x8FE75D89C41EF30AULL, "vxht_0_hightown_cbc_0_1.vlm"},
	{0x8FE767F28A7BDCD1ULL, "6053878_m.fxe"},
	{0x8FE7B2CBC900DAF4ULL, "6014720_m.fxe"},
	{0x8FE7BDF5CC69F2B7ULL, "389787306.ogg"},
	{0x8FE7BE67900886B6ULL, "6149208_m.fxe"},
	{0x8FE7FC135541AE25ULL, "vxht_chantry_27s_0_0.vlm"},
	{0x8FE86AD25C8A02A6ULL, "38284542.ogg"},
	{0x8FE87D5B09969451ULL, "6125206_f.fxe"},
	{0x8FE8A4093F6DE44BULL, "6127476_m.fxe"},
	{0x8FE8B3F545726360ULL, "vxlt_0_lowtown_n_ghy_0_0.vlm"},
	{0x8FE9112F538F4775ULL, "670808392.ogg"},
	{0x8FE963DC247FF222ULL, "6060457_m.fxe"},
	{0x8FE9AA4D83A9EB4BULL, "tbe100cr_golem.utc"},
	{0x8FEA0B90559D29D4ULL, "m_dae210ar_docks_nt.wbd"},
	{0x8FEA154162CB6ECFULL, "6092310_m.fxe"},
	{0x8FEA16CB219A40E9ULL, "zz_isa_debug6175879.cl"},
	{0x8FEA48792BBC3457ULL, "hlac_lt_0_lowtown_131070_3y.dds"},
	{0x8FEA4E761273EB56ULL, "hlac_w_blightlands_3_0x.dds"},
	{0x8FEA6D9090F8CB81ULL, "6106242_m.fxe"},
	{0x8FEA71AF6C58BCC6ULL, "26847005.ogg"},
	{0x8FEA75A0542D41D5ULL, "plt_zz_spy100pt_among_shad.plo"},
	{0x8FEB203317067A47ULL, "llac_lt_0_lowtown_a2_n_65537_3.dds"},
	{0x8FEB6B087F494A42ULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\grass_dp.dds"},
	{0x8FEB7073C445D184ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgtp.phy"},
	{0x8FEB78D3871AB129ULL, "vxlt_0_lowtown_a2_ep4_0_1.vlm"},
	{0x8FEB81AB0249EE53ULL, "lt_0_lowtown_docks_n_12_0_ll.mmh"},
	{0x8FEBA3CD52BF20F4ULL, "6095294_f.fxe"},
	{0x8FEBFFA4E0693E3BULL, "vxlt_undercityencounter_1g7_0_0.vlm"},
	{0x8FEC33120644C1A2ULL, "vxht_playershouse_n_1g1_0_0.vlm"},
	{0x8FEC376D4EB25B21ULL, "vxht_mansion2b_3c2_0_0.vlm"},
	{0x8FEC638841345513ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_reaver.ani"},
	{0x8FEC7AFA331B8F1BULL, "mer320_hubert6089691.cl"},
	{0x8FED88AC19161671ULL, "vxlt_gallowstemplar_f_5cd_0_0.vlm"},
	{0x8FED991AD5DA197FULL, "82583218.ogg"},
	{0x8FEE2DCE356A0871ULL, "6060366_m.fxe"},
	{0x8FEE3147996DF49AULL, "6171823_m.fxe"},
	{0x8FEE4730025AC1C8ULL, "hlac_lt_gallowstemplar_n_2390_1z.dds"},
	{0x8FEE4FD82E31B33EULL, "6065924_m.fxe"},
	{0x8FEF447C1437900EULL, "art\\characters\\playercharacter\\humanfemale\\hf_rob_appa.msh"},
	{0x8FEF4F883151C6A8ULL, "31162092.ogg"},
	{0x8FEF527C1437A85FULL, "art\\characters\\playercharacter\\humanfemale\\hf_rob_appa.mao"},
	{0x8FEF5E7C1437BCFCULL, "art\\characters\\playercharacter\\humanfemale\\hf_rob_appa.mmh"},
	{0x8FEF6784A1301578ULL, "vxw_encounter2_camp_2f_0_0.vlm"},
	{0x8FEF8CA7A1628916ULL, "253652023.ogg"},
	{0x8FEFA8DB23F75F8EULL, "460938343.ogg"},
	{0x8FEFAFB715CC2716ULL, "dae021_keep_maid.utc"},
	{0x8FEFD118A4DC960EULL, "vxlt_darktown_5km_0_1.vlm"},
	{0x8FEFFEF33497E718ULL, "6170376_m.fxe"},
	{0x8FF0159200BBF91FULL, "lgt000cr_kirk_turn_in_3.utc"},
	{0x8FF01E3F1402E281ULL, "704459769.ogg"},
	{0x8FF07B94062547F0ULL, "vxw_encounter3_7r_0_0.vlm"},
	{0x8FF08D55FE714409ULL, "gen00fl_aveline6140157.cl"},
	{0x8FF0CE5F3963578EULL, "vxht_bartrand_n_5q1_0_0.vlm"},
	{0x8FF0F40F76AF99D3ULL, "genip_node_k1.utp"},
	{0x8FF13174C4F0063BULL, "441534688.ogg"},
	{0x8FF1DCCE2C9CDCD7ULL, "gen00fl_anders6119200.cl"},
	{0x8FF21313C3AA35B3ULL, "428310562.ogg"},
	{0x8FF21782BFB38BBDULL, "static_static_p51.psh.fxo.dx11"},
	{0x8FF294CADBDC8BB9ULL, "vxlt_gallowstemplar_f_51j_0_0.vlm"},
	{0x8FF29D41A576A6D0ULL, "vxca_cavern1_haunted_id1_0_0.vlm"},
	{0x8FF2E6F5F7E6BBA4ULL, "6171168_m.fxe"},
	{0x8FF2F9682440D9F8ULL, "347287029.ogg"},
	{0x8FF303B1CA366B3EULL, "darktown_male_elf_sta6174551.cl"},
	{0x8FF38F88CF5C7C8CULL, "6122105_f.fxe"},
	{0x8FF3C6FAC8300769ULL, "hlac_lt_0_lowtown_docks_0_5z.dds"},
	{0x8FF4F207AB6FC663ULL, "clf221_jalen6070388.cl"},
	{0x8FF51076CED7E2F4ULL, "6165204_m.fxe"},
	{0x8FF54E5ABBDE51F8ULL, "6126928_m.fxe"},
	{0x8FF67FFA5D141E68ULL, "vxht_keep_4hi_0_0.vlm"},
	{0x8FF6946F7CE3E99FULL, "vxca_cavern1_haunted_gjf_0_0.vlm"},
	{0x8FF6BB43B3DB6054ULL, "147940603.ogg"},
	{0x8FF6E1188CBE3AB8ULL, "566566419.ogg"},
	{0x8FF768320D9F97B1ULL, "zz_vis_debug6072524.cl"},
	{0x8FF78DE84689B802ULL, "fxa_bodyimp_lining.mao"},
	{0x8FF791727AA9C9AEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_spire.mmh"},
	{0x8FF79D727AA9DE45ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_spire.mao"},
	{0x8FF7AB727AA9F59CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_spire.msh"},
	{0x8FF7F60B40480374ULL, "hlac_ht_mansion2b_818_9y.dds"},
	{0x8FF8087BA7979C63ULL, "vxlt_blackemporium_ig_0_0.vlm"},
	{0x8FF83A1C1048C70FULL, "323544910.ogg"},
	{0x8FFA430DF4BAF320ULL, "782207415.ogg"},
	{0x8FFA5B884140854AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_reaver.evt"},
	{0x8FFA71A5E319AB45ULL, "llac_w_bonepitmining_a2_0_3.dds"},
	{0x8FFAE69B85D84812ULL, "follower_banter6154030.cl"},
	{0x8FFB256F6F0DEDE4ULL, "hlac_lt_0_lowtown_65537_1z.dds"},
	{0x8FFB2B441D170D6AULL, "6208318_m.fxe"},
	{0x8FFBF2DFE709560FULL, "6054389_m.fxe"},
	{0x8FFC02845E5777B5ULL, "6139122_f.fxe"},
	{0x8FFC49F74A095E84ULL, "6174527_m.fxe"},
	{0x8FFCAE0AFCB42F08ULL, "6078982_m.fxe"},
	{0x8FFCBB9DA8D9098FULL, "6020464_f.fxe"},
	{0x8FFD209021D271C0ULL, "gen00fl_aveline6132667.cl"},
	{0x8FFD2EC136980785ULL, "lt_0_lowtown_a2_n_2_0.rml"},
	{0x8FFD4DBB6EAC2624ULL, "6209892_m.fxe"},
	{0x8FFD76AE14728E6EULL, "6119421_m.fxe"},
	{0x8FFE49B70F4953AAULL, "qun211cr_elf_warrior6072342.cl"},
	{0x8FFE8B019162DAB0ULL, "follower_banter6148738.cl"},
	{0x8FFF27762FC69331ULL, "6056724_f.fxe"},
	{0x8FFF403C2B59700EULL, "vxht_bartrand_n_5t2_0_0.vlm"},
	{0x8FFFB68B1EE7A261ULL, "103364534.ogg"},
	{0x900007F3519F6EAEULL, "zz_dog_debug6127779.cl"},
	{0x90003A093F821F1EULL, "474220555.ogg"},
	{0x9000534714916EAEULL, "character_rendertodepth_p8.psh.fxo.dx11.assembly"},
	{0x900055E5CBDF30A5ULL, "fen110_fenris6100287.cl"},
	{0x900079DA9B5112B2ULL, "hlac_ht_0_hightown_a2_-65538_0y.dds"},
	{0x9000BA33E15A495DULL, "ts_dwarven_prog_junk3.gda"},
	{0x9000ECEA4DF76737ULL, "6149061_m.fxe"},
	{0x9000FAEB73A9C0E1ULL, "716578865.ogg"},
	{0x90013A8841466BBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_reaver.gad"},
	{0x9001793AC7C042F9ULL, "vxht_0_hightown_cbm_0_0.vlm"},
	{0x900181A2CDA18F13ULL, "vxw_sundermount_2z7_0_1.vlm"},
	{0x9001909CB2DA182AULL, "823167832.ogg"},
	{0x9001A17EC9BB819AULL, "lowtown_male_cityguard_mob6150480.cl"},
	{0x90025BDDF88CF616ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh1a.msh"},
	{0x900265DDF88D0714ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh1a.mmh"},
	{0x9002D413CB92B16EULL, "hlac_lt_gallowstemplar_n_2394_10z.dds"},
	{0x90031DE977E9FD06ULL, "vxca_caverns2_7wb_0_0.vlm"},
	{0x900323A97DF96390ULL, "gen00fl_fenris6126490.cl"},
	{0x90032EC6E3939320ULL, "vxlt_foundry_n_wa_0_0.vlm"},
	{0x90038EB9C55FF022ULL, "aip_qunariarishock_base.gda"},
	{0x9004294E1D047219ULL, "vxdr_0_deeproads_bx9_0_0.vlm"},
	{0x9004A335034EA18FULL, "vxlt_hangedman_4l_0_0.vlm"},
	{0x9004BA67EE17AAC6ULL, "dae200_cullen6140298.cl"},
	{0x9004C88E76EAD07DULL, "531139674.ogg"},
	{0x9004F2B7A33587D4ULL, "mag100_karras6034428.cl"},
	{0x900531298298B420ULL, "llac_lt_hangedman_-1_1.dds"},
	{0x9006A377E03C16F8ULL, "vxca_caverns1_j4g_0_0.vlm"},
	{0x90072FDE2BDF392EULL, "368222540.ogg"},
	{0x900732FBB0A6C436ULL, "6113837_f.fxe"},
	{0x900733D1FC0C99A6ULL, "1063871529.ogg"},
	{0x90073E971ABE5AFEULL, "art\\characters\\playercharacter\\humanmale\\hlh_medh.phy"},
	{0x900760CAB396FFECULL, "6098272_m.fxe"},
	{0x90078522D9545959ULL, "263777610.ogg"},
	{0x9007EED5BB86489BULL, "731490218.ogg"},
	{0x900851E1AE145B95ULL, "vxht_0_hightown_n_9zp_0_0.vlm"},
	{0x90087D1C87D6FC11ULL, "6095928_m.fxe"},
	{0x90088C28D0A0EB25ULL, "136577763.ogg"},
	{0x900908475F1337D1ULL, "fx_1018_diff_a.dds"},
	{0x9009217C76CC6FF4ULL, "qun211_braggart6075463.cl"},
	{0x90095790073B18F5ULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetnew_n.dds"},
	{0x900A0CCE45DAB77FULL, "dae110cr_elf_f_01.utc"},
	{0x900A6C65FD024771ULL, "vxlt_gallowstemplar_f_5iz_0_0.vlm"},
	{0x900A7CA559A2B11EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_legs_up.ani"},
	{0x900C3DC3C6FDB6F5ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinswall.msh"},
	{0x900C3FC3C6FDBA6BULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinswall.mmh"},
	{0x900C4BC3C6FDCE08ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinswall.mao"},
	{0x900CA53E421226BFULL, "art\\characters\\creatures\\desiredemon\\animation\\enter.gad"},
	{0x900CA801505B2852ULL, "hlac_lt_0_lowtown_131072_5x.dds"},
	{0x900D40AF79F13C8DULL, "vxht_brothel_n_4bw_0_1.vlm"},
	{0x900DE401A111D4B6ULL, "lgt310pt_nt_melson.ncs"},
	{0x900DF1FE2030588FULL, "hlac_lt_0_lowtown_docks_-131075_4y.dds"},
	{0x900E6F52F45B4C9AULL, "vxht_mansion2b_281_0_0.vlm"},
	{0x900EB8485D211BEBULL, "dae000_hate_follower_react6187867.cl"},
	{0x900EC829055416EBULL, "seb321_leliana6180957.cl"},
	{0x900EE2B56F8B826EULL, "218368862.ogg"},
	{0x900F1065A9EDB481ULL, "gen00fl_fenris6126417.cl"},
	{0x900F4D392BA1AE85ULL, "vxw_encounter2_camp_7p_0_0.vlm"},
	{0x900F5EC73FED7FB7ULL, "6142161_m.fxe"},
	{0x900F7A0B84864984ULL, "vxlt_0_lowtown_docks_a3_n_cz9_0_0.vlm"},
	{0x900F8E15BDE6710DULL, "gen_im_trs_gen_junk3_03.uti"},
	{0x901003C4A46A98D7ULL, "6111679_m.fxe"},
	{0x9010219634AEEDEBULL, "hlac_lt_0_lowtown_-131072_7z.dds"},
	{0x9010908B54F42B17ULL, "vxw_blightlands_e7_0_0.vlm"},
	{0x901097B4363EF64DULL, "vxw_woundedcoast_pt_0_0.vlm"},
	{0x9010FE3954C7995DULL, "mag291_torpor6026966.cl"},
	{0x9010FF5A20971D74ULL, "static1_punch_d.psh.fxo"},
	{0x90111256AAFACD4DULL, "dre000_dougal6130988.cl"},
	{0x90115B3500723345ULL, "llac_lt_0_lowtown_docks_n_131070_0.dds"},
	{0x90118FC669E01586ULL, "vxca_caverns2_7lc_0_0.vlm"},
	{0x90122B3F098DA479ULL, "6053953_m.fxe"},
	{0x9012A75C207AE15AULL, "hlac_lt_0_lowtown_n_131071_4y.dds"},
	{0x9012C958AE86907DULL, "vxlt_0_lowtown_a3_n_c_ewl_0_0.vlm"},
	{0x90131FFDCD0D5C47ULL, "331116673.ogg"},
	{0x90139A3E421832ACULL, "art\\characters\\creatures\\desiredemon\\animation\\enter.evt"},
	{0x90139AB93E0B75DCULL, "vxdr_0_deeproads_cp1_0_0.vlm"},
	{0x90147AB13E99B338ULL, "6016638_m.fxe"},
	{0x901518516E4BF2AEULL, "759575105.ogg"},
	{0x901535DAC7F88FE5ULL, "blood_mud.dds"},
	{0x901559B30F6FB2E5ULL, "6036501_m.fxe"},
	{0x9015781B2274D9EBULL, "hlac_lt_warehouse_n_0_1x.dds"},
	{0x9015E9243C0F20ECULL, "vxlt_0_lowtown_a3_n_c_bxk_0_0.vlm"},
	{0x90163E5D2DC970F9ULL, "llac_w_encounter2_camp_-65531_0.dds"},
	{0x9016A02853C42BBBULL, "6162422_m.fxe"},
	{0x9016F99055C3F96EULL, "m_dae210ar_docks_nt.bnk"},
	{0x901729D9EB72F335ULL, "one350_variel6085536.cl"},
	{0x90175634DE1C7F7AULL, "ser191_emeric6120479.cl"},
	{0x901794265BBBB0A0ULL, "6156066_m.fxe"},
	{0x9017D8345F30C839ULL, "6120134_m.fxe"},
	{0x9017DA6A3063B563ULL, "framebuffereffects_composedof_p0.psh.fxo"},
	{0x90189200C3AE3C33ULL, "155039307.ogg"},
	{0x901894A559AF17EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_legs_up.evt"},
	{0x9018C755FEC861A2ULL, "vxlt_0_lowtown_a2_n_c_eva_0_0.vlm"},
	{0x9018D71E91D341D2ULL, "79141673.ogg"},
	{0x9019285F26EE7F10ULL, "zz_mrl_debug6109301.cl"},
	{0x901934DB88B54BE2ULL, "6017495_m.fxe"},
	{0x90198BEA81BCBD6BULL, "hlac_ca_caverns2prc_5802_43x.dds"},
	{0x9019B5DB073C19C3ULL, "603701703.ogg"},
	{0x901A104E6B2CC4C6ULL, "mag360_samson6042097.cl"},
	{0x901A266523563791ULL, "art\\cinematics\\props\\bookinquisb_dagger.msh"},
	{0x901A323250412DF0ULL, "vxlt_0_lowtown_a2_ewu_0_0.vlm"},
	{0x901A40652356644FULL, "art\\cinematics\\props\\bookinquisb_dagger.mmh"},
	{0x901AB0347C5C889AULL, "6044990_m.fxe"},
	{0x901B187C37A08EB9ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdgliding01.dura.ani"},
	{0x901B4A7096F6433DULL, "ser200_aveline6201474.cl"},
	{0x901B7071A21B5BA2ULL, "hlac_lt_0_lowtown_-65535_4z.dds"},
	{0x901B9851D72882FFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scroll_n.dds"},
	{0x901C2BA38A623212ULL, "vxht_0_hightown_b9f_0_0.vlm"},
	{0x901CB8FE609770CCULL, "vxht_bartrand_n_34f_0_0.vlm"},
	{0x901CE12FE651651AULL, "vxht_mansion2b_252_0_0.vlm"},
	{0x901DA253E7A7CB8CULL, "363031802.ogg"},
	{0x901E95A559B38498ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_legs_up.gad"},
	{0x901F026846EA4397ULL, "vxw_blightlands_n6_0_0.vlm"},
	{0x901F291E5D07B7FEULL, "6070896_m.fxe"},
	{0x901F6C020290BFA3ULL, "ser270_emeric6111447.cl"},
	{0x901FD9D2C8D0956CULL, "hlac_w_blightlands_65539_1x.dds"},
	{0x90202FC39DC891DCULL, "6033402_m.fxe"},
	{0x90203DCA6F3338E5ULL, "vxht_mansion2b_3gj_0_0.vlm"},
	{0x90203FC1E331169CULL, "vxca_caverns1_fn9_0_0.vlm"},
	{0x9020772CB8C6BD5DULL, "vxca_caverns1_mines_j69_0_0.vlm"},
	{0x902095A3CD0D3CC4ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.dura.ani"},
	{0x902148FB0CD45952ULL, "424800167.ogg"},
	{0x902158EDACF9C8C1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_table_0.pwk"},
	{0x9021823E42244705ULL, "art\\characters\\creatures\\desiredemon\\animation\\enter.ani"},
	{0x9021902D25470AFAULL, "hlac_lt_0_lowtown_docks_131070_0x.dds"},
	{0x9021D27CEBFD4B62ULL, "vxlt_undercityencounter_1bq_0_0.vlm"},
	{0x9021F37C37A66EFBULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdgliding01.dura.gad"},
	{0x90231D04B6C321F7ULL, "gen00fl_isabela_init6136305.cl"},
	{0x90234FCF6A94EFF0ULL, "hlac_w_sundermount_-393212_0x.dds"},
	{0x9023937BCB31EF3DULL, "364918019.ogg"},
	{0x9023CC5F394C7A7FULL, "6121005_m.fxe"},
	{0x9023D3BC87B0D659ULL, "6173123_m.fxe"},
	{0x9024435AF68CB7AEULL, "591508851.ogg"},
	{0x902468BB8E4CEA93ULL, "364492052.ogg"},
	{0x9025098FBF55F28FULL, "vxlt_0_lowtown_n_ewq_0_0.vlm"},
	{0x902515F05798C648ULL, "888981109.ogg"},
	{0x90255585A975FBEDULL, "748286851.ogg"},
	{0x902582E1AD886D11ULL, "6149882_f.fxe"},
	{0x902768A05460D9B2ULL, "plt_zz_spy100pt_among_shad.nss"},
	{0x902792A3CD135626ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.dura.gad"},
	{0x9028107C37AB0BE0ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdgliding01.dura.evt"},
	{0x9028269628008169ULL, "w_bonepitmining_a2_60_0.mao"},
	{0x9028329628009502ULL, "w_bonepitmining_a2_60_0.mmh"},
	{0x90283496280098B8ULL, "w_bonepitmining_a2_60_0.msh"},
	{0x9028889D8BB922C7ULL, "305487404.ogg"},
	{0x9028B9464761E703ULL, "6166884_m.fxe"},
	{0x9028DD40DE4AE27BULL, "w_bonepitdestroyed_a3_186_0.mmh"},
	{0x9028E940DE4AF618ULL, "w_bonepitdestroyed_a3_186_0.mao"},
	{0x9028FB40DE4B1565ULL, "w_bonepitdestroyed_a3_186_0.msh"},
	{0x902909BDD081A706ULL, "6091839_m.fxe"},
	{0x9029219CF2FD3AB5ULL, "vxw_bonepitmining_a2_1xd_0_0.vlm"},
	{0x90293CA219ADC90DULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6whitelyrium.mmh"},
	{0x902940A219ADD05EULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6whitelyrium.mao"},
	{0x90294EA219ADE793ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_6whitelyrium.msh"},
	{0x902970D5A2779DF1ULL, "6126639_m.fxe"},
	{0x90297FA45EEC528BULL, "vxlt_0_lowtown_a2_n_gvh_0_0.vlm"},
	{0x90297FEE27185EAAULL, "791322886.ogg"},
	{0x902A0A95334913C8ULL, "follower_banter6173035.cl"},
	{0x902A6263472207B7ULL, "ser120_ghyslain6135797.cl"},
	{0x902AA19A0301CABBULL, "ach_soc_bigwigs.dds"},
	{0x902B055A7B2E28EEULL, "framebuffereffects_ps3downsamplemsaa_p0.psh.fxo.dx11"},
	{0x902B22BCD023CC2EULL, "art\\vfx\\combat\\vfx_glb_ele_n_major_c.mmh"},
	{0x902B6C5A55B3307FULL, "564302348.ogg"},
	{0x902BAE0998C628A3ULL, "art\\vfx\\creatures\\vfx_c_drag_roar_c.impa.ani"},
	{0x902BCB81A60B4DBEULL, "hightown_male_noble_gossip6171390.cl"},
	{0x902C3D5BB649C3BEULL, "da2terrain_terrain_p0.hsh.fxo.dx11"},
	{0x902CA1E7F7588C11ULL, "vxca_caverns1_mines_e40_0_0.vlm"},
	{0x902CBFA90889CE2DULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtc.mmh"},
	{0x902CC3A90889D57EULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtc.mao"},
	{0x902CD1A90889ECB3ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtc.msh"},
	{0x902D6CDB8FC9689CULL, "follower_banter6143425.cl"},
	{0x902D6DBC95712B7BULL, "74322089.ogg"},
	{0x902D744538E05C17ULL, "vxw_blightlands_s1_0_0.vlm"},
	{0x902D9C6610D01A28ULL, "dae000_elegant6166934.cl"},
	{0x902DCE0EB91C59CCULL, "6150374_m.fxe"},
	{0x902E8DA3CD196C05ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.dura.evt"},
	{0x902F08215F5DB5ADULL, "hlac_lt_0_lowtown_131070_2x.dds"},
	{0x902F24B073ECF798ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_101a_n.dds"},
	{0x902F2FE7E8FD1FA8ULL, "1055499045.ogg"},
	{0x902F55F41B4193EAULL, "vxca_caverns1_mines_jl8_0_0.vlm"},
	{0x902F6EC825CBC2F1ULL, "6180517_f.fxe"},
	{0x903021872E7FA19BULL, "vxht_mansion2_23i_0_0.vlm"},
	{0x90307E7321F7A6DCULL, "vxdr_0_deeproads_cj7_0_0.vlm"},
	{0x90309DB25F21653DULL, "follower_banter6155700.cl"},
	{0x9030B6CB3E86DF8BULL, "hlac_w_bonepitmining_a2_393220_0x.dds"},
	{0x90311B852E4F9287ULL, "vfx0_contactsheet.vsh.fxo"},
	{0x9031B4B3EDD0BCB0ULL, "hlac_w_bonepitdestroyed_a3_0_4x.dds"},
	{0x9031FD5D9146CD1EULL, "149103501.ogg"},
	{0x90320493F65143ABULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2_0.pwk"},
	{0x903226CCB3971D62ULL, "mag100cr_thrask.utc"},
	{0x90322C31EBFE95D1ULL, "hlac_lt_hangedman_-1_20z.dds"},
	{0x90329E236FB5979EULL, "6037685_f.fxe"},
	{0x9032B24303D05B0CULL, "mer240_jansen6045098.cl"},
	{0x9032BC5424B5DA4BULL, "hlac_lt_0_lowtown_docks_n_196606_0x.dds"},
	{0x9032BF9AA5A3B937ULL, "vxlt_0_lowtown_n_e3e_0_0.vlm"},
	{0x9032DE6A1F042979ULL, "6043682_m.fxe"},
	{0x903339661BD48C4BULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar5_.ani"},
	{0x90334B8FE5B4967CULL, "vxw_woundedcoast_15c_0_0.vlm"},
	{0x90334FAB4F0F372FULL, "545642173.ogg"},
	{0x90337ABB1FEFE43AULL, "vxlt_gallowstemplar_f_5au_0_0.vlm"},
	{0x90338D4E3ACE469CULL, "928119107.ogg"},
	{0x9033901B3FFE606EULL, "600775364.ogg"},
	{0x9033AD69D021B6D4ULL, "392923653.ogg"},
	{0x9033B22440E92BA9ULL, "vxlt_0_lowtown_a2_eo9_0_1.vlm"},
	{0x9033D151FB59FB80ULL, "824512734.ogg"},
	{0x9033E76FA68D3A0BULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masc_0t.dds"},
	{0x9033F5A9FF920B21ULL, "451958597.ogg"},
	{0x90340374883BA96BULL, "llac_lt_0_lowtown_n_131071_0.dds"},
	{0x90344C539F94F673ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_017a_0d.dds"},
	{0x9034B5D61C1F8042ULL, "477692184.ogg"},
	{0x9034BF0BA966D3B6ULL, "1020011142.ogg"},
	{0x9035BDCCE2380446ULL, "mrl200pt_hunteramulets.ncs"},
	{0x90364C93376772E9ULL, "m_mag371ar_gallows_templar.wbd"},
	{0x90364F812C15B1C3ULL, "ser270_emeric6113610.cl"},
	{0x90366F18014087B2ULL, "6095951_f.fxe"},
	{0x90368702C02E0AC8ULL, "ht_bartrand_n_3372.anb"},
	{0x903697E90D61A477ULL, "6115569_m.fxe"},
	{0x9037301D6CB90CD1ULL, "6201620_m.fxe"},
	{0x903758A22699930FULL, "967589256.ogg"},
	{0x90377B9AC277BF86ULL, "6096939_m.fxe"},
	{0x90379379E4F35335ULL, "vxw_bonepitmining_a2_1uc_0_0.vlm"},
	{0x9037D66B95D7E078ULL, "6085177_m.fxe"},
	{0x9037F7B9873A362CULL, "vxdr_primevaltomb_8py_0_1.vlm"},
	{0x9037FD2665B2380BULL, "106286642.ogg"},
	{0x903806600F1D5CE6ULL, "vxlt_gallowstemplar_6q9_0_0.vlm"},
	{0x903903090F1AFEB1ULL, "follower_banter6141569.cl"},
	{0x903941BCD0303E2DULL, "art\\vfx\\combat\\vfx_glb_ele_n_major_c.anb"},
	{0x9039C60998D28E9AULL, "art\\vfx\\creatures\\vfx_c_drag_roar_c.impa.evt"},
	{0x9039F268E66503D0ULL, "follower_banter6165469.cl"},
	{0x903A029956B18C13ULL, "6104570_m.fxe"},
	{0x903A2338DF6AF134ULL, "vxlt_gallowstemplar_f_3nl_0_0.vlm"},
	{0x903A38F2FFE57711ULL, "321121546.ogg"},
	{0x903A6422E92D7FA1ULL, "vxlt_0_lowtown_a2_n_c_ekl_0_1.vlm"},
	{0x903B4F5CF14EF1DCULL, "vxht_chantry_2b1_0_0.vlm"},
	{0x903B647ECADEDE26ULL, "hlac_lt_0_lowtown_n_-65537_5x.dds"},
	{0x903BE6222AD67497ULL, "vxw_blightlands_t0_0_0.vlm"},
	{0x903C467B66F29D4CULL, "vxlt_gallowstemplar_n_55o_0_0.vlm"},
	{0x903C4BD441CD4433ULL, "s_aoe_distor.mao"},
	{0x903CC09814F8C94CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_major_v1.ani"},
	{0x903CDBE7E91D0A61ULL, "vxht_bartrand_n_6pf_0_0.vlm"},
	{0x903DB8A9089835EAULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgtc.phy"},
	{0x903DBD5C56FFBA06ULL, "986937718.ogg"},
	{0x903DCB9276E64289ULL, "140197498.ogg"},
	{0x903E0FFEA11E3216ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrain.ani"},
	{0x903E2031BC6912F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_enter.gad"},
	{0x903E869E6B3A94E1ULL, "6093492_m.fxe"},
	{0x903EA85F3BACCF3DULL, "crusts.gda"},
	{0x903ECE4C94618D0BULL, "vxht_mansion2b_1vq_0_0.vlm"},
	{0x903ED37466C98439ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_lgtc_0s.dds"},
	{0x903EF05013EDBF5CULL, "vxdr_0_deeproads_cg4_0_0.vlm"},
	{0x903F7305DD100A8CULL, "llac_ht_keep_4519_4.dds"},
	{0x903FEFEA53E617E1ULL, "6042986_m.fxe"},
	{0x9040247DD647E159ULL, "gen00fl_isabela6099320.cl"},
	{0x904041CC616616E3ULL, "vxlt_blackemporium_vj_0_0.vlm"},
	{0x9040760A8455CED6ULL, "6176731_f.fxe"},
	{0x9040C50998D8AB8DULL, "art\\vfx\\creatures\\vfx_c_drag_roar_c.impa.gad"},
	{0x90410F12C74595FEULL, "985152576.ogg"},
	{0x9041174D2655340EULL, "vxht_keep_4nm_0_0.vlm"},
	{0x904127625380DE4FULL, "character_character_p46.psh.fxo.dx11.assembly"},
	{0x90416DCA7CD284ACULL, "569895291.ogg"},
	{0x904184B36703F442ULL, "llac_ca_cavern1_haunted_24421_1.dds"},
	{0x9041916CB7C1E83EULL, "art\\vfx\\creatures\\fxc_golem_eburst_p.impa.ani"},
	{0x9041ED49A342238FULL, "vxlt_0_lowtown_n_ems_0_0.vlm"},
	{0x904235D0BA5DF078ULL, "vxht_mansion_n_57r_0_0.vlm"},
	{0x9042746F54A10782ULL, "dae000_bodhan6146806.cl"},
	{0x90427692695E91D0ULL, "hlac_lt_0_lowtown_n_-262144_0x.dds"},
	{0x904297DF5FAB55A4ULL, "6060132_m.fxe"},
	{0x904326489EF7F94EULL, "6127099_m.fxe"},
	{0x90434910BB04B5BFULL, "vxdr_primevaltomb_7m3_0_0.vlm"},
	{0x904354E9873A2E11ULL, "zz_followers_banter6158424.cl"},
	{0x904394765B793F9BULL, "vxlt_shop_g2_0_0.vlm"},
	{0x9043CD9814FEFD8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_major_v1.gad"},
	{0x904418A1287A281EULL, "6148121_f.fxe"},
	{0x904476A5D77B8F81ULL, "vxdr_0_deeproads_cmg_0_0.vlm"},
	{0x904491B1E859B486ULL, "89232514.ogg"},
	{0x9044A48914826F22ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_flsa_0n.dds"},
	{0x9044AEAD51D5F6B2ULL, "zz_dre_debug6056973.cl"},
	{0x9044FA9856F33D12ULL, "dae300_alistair_king6134986.cl"},
	{0x90452A36BCC62B71ULL, "llac_lt_0_lowtown_a3_n_c_-65536_2.dds"},
	{0x90453531BC6F5484ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_enter.evt"},
	{0x904581C628111A18ULL, "llac_dr_dungeon_3217_3.dds"},
	{0x90458C827FFDFE31ULL, "6095069_m.fxe"},
	{0x904621E26F9A41F7ULL, "plt_genpt_beware_of_dog.plo"},
	{0x90462C5CE55F5CD2ULL, "dae121cr_sister_alema.utc"},
	{0x9046649DAA71B648ULL, "vxht_bartrand_n_6pc_0_0.vlm"},
	{0x904672BADD89DFB4ULL, "vxlt_gallowstemplar_n_39a_0_0.vlm"},
	{0x904720661BE50AD5ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar5_.gad"},
	{0x904860D627C5579BULL, "w_woundedcoast_244_0.msh"},
	{0x904872D627C576E6ULL, "w_woundedcoast_244_0.mao"},
	{0x90487ED627C58A85ULL, "w_woundedcoast_244_0.mmh"},
	{0x904943639BE1761DULL, "hlac_lt_0_lowtown_docks_65532_0x.dds"},
	{0x904950CD34CA365EULL, "6160829_m.fxe"},
	{0x9049C8981503616DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_cold_major_v1.evt"},
	{0x9049FEF537AC9952ULL, "812733538.ogg"},
	{0x904A54218EE4F270ULL, "hlac_lt_0_lowtown_a2_-65536_1x.dds"},
	{0x904A782172F6390DULL, "vxht_keep_n_5xo_0_1.vlm"},
	{0x904B43A7FF4DD233ULL, "6141134_f.fxe"},
	{0x904C17FEA12A7D07ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrain.evt"},
	{0x904C1D44B22FFFF0ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_dssba_0d.dds"},
	{0x904C716CCCCF5472ULL, "131280489.ogg"},
	{0x904C8717528C7065ULL, "gen00fl_anders6136021.cl"},
	{0x904C94DDE4F5641DULL, "208703753.ogg"},
	{0x904CAF27A1C26484ULL, "137732454.ogg"},
	{0x904D0541126BD29BULL, "vxht_mansion2_29k_0_0.vlm"},
	{0x904D7C53D2B95DCDULL, "vxw_bonepit_1lz_0_0.vlm"},
	{0x904DADEFB13C3FE9ULL, "639596797.ogg"},
	{0x904DE8ED3DD684F7ULL, "llac_ht_mansion_n_3370_0.dds"},
	{0x904E918CB245157EULL, "119124299.ogg"},
	{0x904EA96CB7CC9B0FULL, "art\\vfx\\creatures\\fxc_golem_eburst_p.impa.evt"},
	{0x904ED88EFBF53C5AULL, "hlac_lt_undercitydungeon_-1_30x.dds"},
	{0x904F454117350929ULL, "hlac_dr_primevaltomb_661_3y.dds"},
	{0x904F478FA6D2DA81ULL, "vxht_chantry_2ia_0_0.vlm"},
	{0x904FD830B4058AE5ULL, "6115995_f.fxe"},
	{0x9050522336EA63C2ULL, "character_character_p52.psh.fxo.dx11"},
	{0x90505E7503DC153AULL, "vxlt_gallowstemplar_f_5ow_0_0.vlm"},
	{0x90506C4EBD38CC0FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2lightchain.mao"},
	{0x9050784EBD38E0ACULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2lightchain.mmh"},
	{0x90507E4EBD38EA9EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2lightchain.msh"},
	{0x9051039CCFF74812ULL, "vxdr_primevaltomb_6uw_0_0.vlm"},
	{0x9051362E648C7489ULL, "dae000cr_grand_cleric6182025.cl"},
	{0x905153715417159CULL, "hlac_lt_gallowstemplar_n_2394_3x.dds"},
	{0x905179E1B0BB9852ULL, "129614595.ogg"},
	{0x90522E81C72BD96BULL, "270483221.ogg"},
	{0x90524D31BC7A087DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\scared_enter.ani"},
	{0x905258E07E68F56CULL, "vxdr_primevaltomb_7y0_0_0.vlm"},
	{0x90525FA9632600B3ULL, "6202304_m.fxe"},
	{0x9052699D146E7812ULL, "6151674_m.fxe"},
	{0x9052A67A2C3557D5ULL, "vxlt_gallowsprison_n_60z_0_0.vlm"},
	{0x9052B721B71A0BD2ULL, "qun221_petrice6082574.cl"},
	{0x9052E882C971A801ULL, "vxdr_0_deeproads_cjf_0_0.vlm"},
	{0x905338FEA130D450ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_blooddrain.gad"},
	{0x9053F388B07D7D75ULL, "vxht_0_hightown_a3_bxg_0_0.vlm"},
	{0x90541DE5EC189E6BULL, "vxca_caverns1_mines_dpw_0_0.vlm"},
	{0x9054A652A663E578ULL, "6167718_m.fxe"},
	{0x9054ADA16A5E4EF5ULL, "6103212_m.fxe"},
	{0x9055A87632937E76ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medp_0n.dds"},
	{0x9055AA6CB7D2BBB8ULL, "art\\vfx\\creatures\\fxc_golem_eburst_p.impa.gad"},
	{0x9055AFFC2F7E3CB9ULL, "llac_ht_0_hightown_a3_-65538_1.dds"},
	{0x9056F9841AD1373AULL, "6095985_m.fxe"},
	{0x905702AF8DFC7A67ULL, "68973411.ogg"},
	{0x90573460243E9E23ULL, "gen_im_npc_dwf_arm_hlm_meda.uti"},
	{0x905739788E141089ULL, "6075737_m.fxe"},
	{0x9057ACABC6DBB62FULL, "279224474.ogg"},
	{0x9058429B6545A31CULL, "758660167.ogg"},
	{0x90588B47DF6DF66FULL, "setskillchecks.ncs"},
	{0x9059677322A9567AULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm_scalpm1.mao"},
	{0x905AAE4574A392B4ULL, "259792477.ogg"},
	{0x905AD9287747215AULL, "6204255_m.fxe"},
	{0x905B30E26FACB76CULL, "plt_genpt_beware_of_dog.nss"},
	{0x905B4453EC790B2EULL, "vxht_keep_n_6h3_0_1.vlm"},
	{0x905BDF4F2C6C8D12ULL, "vxlt_warehouse_n_1cb_0_0.vlm"},
	{0x905C1E933787F3CBULL, "m_mag371ar_gallows_templar.bnk"},
	{0x905C891D16F12D03ULL, "vxht_bartrand_n_5p6_0_1.vlm"},
	{0x905D074D9793DEC1ULL, "hlac_lt_0_lowtown_docks_n_-262144_1y.dds"},
	{0x905DAA84FB7C050BULL, "551512668.ogg"},
	{0x905E15B30A2028FBULL, "hlac_w_woundedcoast_65539_0x.dds"},
	{0x905E52780107E27FULL, "vxlt_0_lowtown_n_ej6_0_0.vlm"},
	{0x905E64DDF8DB640FULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh1a.phy"},
	{0x905E6BAA55835D78ULL, "6102711_m.fxe"},
	{0x905E71330DD936A8ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.anb"},
	{0x905EB1E7F8B7FA5AULL, "6131961_f.fxe"},
	{0x905EBB74A934A610ULL, "6206974_m.fxe"},
	{0x905F7579C1ED6092ULL, "vxdr_primevaltomb_6xp_0_0.vlm"},
	{0x905F82E01A405CD2ULL, "6068519_f.fxe"},
	{0x905F9EEC833C3200ULL, "llac_ht_chantry_n_2534_44.dds"},
	{0x905FF15B199EE025ULL, "6100213_m.fxe"},
	{0x906007EA5CC80E59ULL, "484877483.ogg"},
	{0x906014A0CACB6396ULL, "hlac_w_blightlands_262144_0z.dds"},
	{0x906063BA20B39A51ULL, "vxlt_0_lowtown_a3_n_c_fhn_0_0.vlm"},
	{0x9060CABD705F0DECULL, "vxdr_primevaltomb_7r1_0_0.vlm"},
	{0x906105110EDB343AULL, "gen00fl_anders6045335.cl"},
	{0x906120356A9EFF6FULL, "677335055.ogg"},
	{0x90619D5C63A11A7EULL, "6033686_m.fxe"},
	{0x9061C27186D7F465ULL, "525235319.ogg"},
	{0x9061FB5CCCED25ECULL, "882529053.ogg"},
	{0x90620572DBBE5E92ULL, "hlac_ht_mansion2_818_33z.dds"},
	{0x90621EA8047C654CULL, "6208028_m.fxe"},
	{0x90623032CAF8D4D7ULL, "vxht_keep_6zv_0_0.vlm"},
	{0x90625C025737497BULL, "758392698.ogg"},
	{0x90626565A27395F5ULL, "vxht_0_hightown_a3_bwf_0_0.vlm"},
	{0x90628D8B73B44A40ULL, "vxdr_dungeon_21s_0_0.vlm"},
	{0x9062F2343DA48660ULL, "dae300cr_bodahn.utc"},
	{0x90633484D2C37C24ULL, "gen00fl_carver6097381.cl"},
	{0x9063471826C4B48BULL, "vxlt_0_lowtown_a2_n_gbl_0_0.vlm"},
	{0x9063FE52959DC3F9ULL, "one300_murder_followers6086730.cl"},
	{0x906402D0DDE708C4ULL, "467962161.ogg"},
	{0x9064122D915A816FULL, "6172294_m.fxe"},
	{0x9064180AAAF2F354ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc03a.phy"},
	{0x9064BFE4A2CAC3DCULL, "209628559.ogg"},
	{0x9064D2107358A02CULL, "6060452_f.fxe"},
	{0x9064D602708B689BULL, "339128908.ogg"},
	{0x9064DAC3B2AFE7AAULL, "140741401.ogg"},
	{0x906509A74BACAFCAULL, "6168182_m.fxe"},
	{0x906545C948F32571ULL, "6060537_m.fxe"},
	{0x90654AE50D03FD08ULL, "470198623.ogg"},
	{0x9065545267D4B358ULL, "character_character_p111.psh.fxo.dx11.assembly"},
	{0x90655964C0D51CE8ULL, "zz_one_debug6099953.cl"},
	{0x9065D95B6D5199F0ULL, "917746594.ogg"},
	{0x9065DC86FBF67E98ULL, "6120452_m.fxe"},
	{0x90666715C70D01F9ULL, "6178889_m.fxe"},
	{0x9067EA2E97681E56ULL, "vxlt_gallowstemplar_f_6s1_0_1.vlm"},
	{0x906841AA350F8F1EULL, "249343117.ogg"},
	{0x90689D712C3C56F7ULL, "993437404.ogg"},
	{0x90689E4D114154E2ULL, "vxlt_0_lowtown_docks_a3_n_apk_0_0.vlm"},
	{0x9068B7953027F475ULL, "llac_ht_0_hightown_a2_0_2.dds"},
	{0x9068F28C9828CBF0ULL, "6098193_m.fxe"},
	{0x90690FFDA7FE0911ULL, "llac_lt_gallowscourtyard_65534_0.dds"},
	{0x90698908674DCAA4ULL, "qun231_varnell6027806.cl"},
	{0x9069B5D12C61B978ULL, "566202526.ogg"},
	{0x9069CBE3AB6C805EULL, "6149190_m.fxe"},
	{0x906A3DB9E3DB6776ULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.cess.evt"},
	{0x906AA8F057743806ULL, "394775825.ogg"},
	{0x906B75A38D917AE7ULL, "51199213.ogg"},
	{0x906BA39F51BC49DEULL, "hlac_lt_darktown_-1_33x.dds"},
	{0x906C1F8F34303476ULL, "mag311_keran6054227.cl"},
	{0x906C43AAB3008D3BULL, "361807792.ogg"},
	{0x906C549204D61C09ULL, "6084232_m.fxe"},
	{0x906C783881BDBE10ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexint_sa.dds"},
	{0x906C932F44530DC5ULL, "42183195.ogg"},
	{0x906C9D4AE747B1BBULL, "208330344.ogg"},
	{0x906CC437433A2E08ULL, "follower_banter6141948.cl"},
	{0x906D3CA3CE801074ULL, "6019253_m.fxe"},
	{0x906D53D343167095ULL, "hlac_lt_0_lowtown_a2_n_0_9z.dds"},
	{0x906D776C3CDF4BBAULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelarge.mmh"},
	{0x906D7B6C3CDF52F1ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelarge.mao"},
	{0x906D896C3CDF6A00ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelarge.msh"},
	{0x906DA2A9A9DBB1D1ULL, "34653769.ogg"},
	{0x906DA8E360CA5A6FULL, "llac_lt_0_lowtown_docks_-131072_0.dds"},
	{0x906DE26FD199B11EULL, "dae000_pc_home_ambients6208675.cl"},
	{0x906E47EBFE5DDBA9ULL, "6042429_m.fxe"},
	{0x906E509786C7CBE4ULL, "248313704.ogg"},
	{0x906E5821D296198BULL, "mag170_thrask6015989.cl"},
	{0x906E6E83B0B37EEBULL, "fen110_anso6094940.cl"},
	{0x906E94D36E468FEAULL, "gen00fl_carver6103648.cl"},
	{0x906EABA2CF44D193ULL, "zz_followers_banter6158853.cl"},
	{0x906EF50AAAFCBF51ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc03a.msh"},
	{0x906F0F0AAAFCEB0FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc03a.mmh"},
	{0x906F1C7BFFF26924ULL, "gen00fl_aveline6137466.cl"},
	{0x906F71BB5A25CA90ULL, "6016511_m.fxe"},
	{0x906FB0554F49593DULL, "vxw_sundermount_34r_0_1.vlm"},
	{0x906FFE40DE871CB0ULL, "w_bonepitdestroyed_a3_186_0.phy"},
	{0x907034B9E3DFC301ULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.cess.gad"},
	{0x90705A22A4E9669EULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_c.impa.evt"},
	{0x9070A20FBCEEED57ULL, "vxht_keep_6gq_0_0.vlm"},
	{0x9070D4448274BA36ULL, "fen210_hadriana6110588.cl"},
	{0x907174FD66C07618ULL, "follower_banter6171932.cl"},
	{0x9071CDBCB93F5B1BULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_disp_p.impa.ani"},
	{0x907211783137E4EAULL, "art\\characters\\weapons\\shields\\w_shd_001a.mao"},
	{0x907215783137EB39ULL, "art\\characters\\weapons\\shields\\w_shd_001a.mmh"},
	{0x90721F783137FC27ULL, "art\\characters\\weapons\\shields\\w_shd_001a.msh"},
	{0x90726BE0F7C2AC8DULL, "art\\characters\\playercharacter\\elfmale\\rob_appb.mmh"},
	{0x90727DE0F7C2CB13ULL, "art\\characters\\playercharacter\\elfmale\\rob_appb.msh"},
	{0x907288C9DE1B9548ULL, "hlac_lt_backalley_g_0_0x.dds"},
	{0x9072B95D48580A33ULL, "art\\characters\\creatures\\desiredemon\\animation\\p_wkf.evt"},
	{0x9072BE97367F54B0ULL, "219396459.ogg"},
	{0x9073A9651726345AULL, "vxw_bonepitdestroyed_a3_1wm_0_1.vlm"},
	{0x9074853A1C1B555AULL, "275470591.ogg"},
	{0x90748D95883FE6CAULL, "cod_ltr_sk3_scribbles1.utp"},
	{0x9074996A23BA9651ULL, "6128369_m.fxe"},
	{0x9074DC75C36C91CBULL, "vxdr_0_deeproads_ad5_0_1.vlm"},
	{0x9075A393BCE577EBULL, "vxlt_shop_el_0_0.vlm"},
	{0x9075E8B4F163790EULL, "llac_lt_warehouse_n_65537_2.dds"},
	{0x90762AA04717EABDULL, "hm_carver_h2s3.mor"},
	{0x90767122A4EDF9C9ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_c.impa.gad"},
	{0x907671A9E732ADDAULL, "1039212657.ogg"},
	{0x9076F071B8B31A40ULL, "lgt000im_ultra_redblossom.uti"},
	{0x9076FA7D4FB6388CULL, "vxdr_0_deeproads_eby_0_0.vlm"},
	{0x907735B9E3E5E3AFULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.cess.ani"},
	{0x90778602C0659F80ULL, "ht_bartrand_n_3372.rml"},
	{0x90778999FCC8DF5BULL, "llac_lt_0_lowtown_docks_a3_-3_2.dds"},
	{0x9077D3BA48D981D9ULL, "w_woundedcoast_123_0.msh"},
	{0x9077DDBA48D992C7ULL, "w_woundedcoast_123_0.mmh"},
	{0x9077E1BA48D99914ULL, "w_woundedcoast_123_0.mao"},
	{0x9078064DABD2FE6DULL, "390376027.ogg"},
	{0x90786993543472EDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglebase.mmh"},
	{0x90786D935434793EULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglebase.mao"},
	{0x90787B9354349173ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglebase.msh"},
	{0x9078C1B5555CD0B5ULL, "gen00fl_fenris6128077.cl"},
	{0x9079668D70501C53ULL, "llac_ca_cavern1_haunted_24424_21.dds"},
	{0x9079B25D485E1D74ULL, "art\\characters\\creatures\\desiredemon\\animation\\p_wkf.gad"},
	{0x9079D2EFFC6CE6EAULL, "gen00fl_sebastian6189041.cl"},
	{0x907A0C7F35E66065ULL, "6095350_f.fxe"},
	{0x907A3E02CBE47839ULL, "6116337_m.fxe"},
	{0x907A66F74FBD1198ULL, "6114070_f.fxe"},
	{0x907AC16E37EF8340ULL, "495103927.ogg"},
	{0x907B3C84457B1FAEULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.cess.ani"},
	{0x907B8FCA671C6429ULL, "6106915_f.fxe"},
	{0x907BF3AC017B3718ULL, "6129984_f.fxe"},
	{0x907C0FA7EB7D70A5ULL, "vxdr_0_deeproads_cff_0_0.vlm"},
	{0x907C2F0C19303509ULL, "vxht_bartrand_n_435_0_0.vlm"},
	{0x907C6776F6F06C32ULL, "llac_ht_chantry_2534_27.dds"},
	{0x907D4222A4F3C817ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_c.impa.ani"},
	{0x907D9D9713ABA4D8ULL, "vxlt_blackemporium_ry_0_0.vlm"},
	{0x907DC9240E02EFF5ULL, "955849100.ogg"},
	{0x907DD79823A93540ULL, "47016356.ogg"},
	{0x907E2232413F71BDULL, "vxw_sundermount_33s_0_1.vlm"},
	{0x907E333CAD59F340ULL, "fxa_evl_dirty.mao"},
	{0x907ED5BCB949F312ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_disp_p.impa.evt"},
	{0x907F585AF63122F5ULL, "hlac_ht_chantry_n_2534_14z.dds"},
	{0x907F917F075F1B62ULL, "33987479.ogg"},
	{0x908013270FFDA08CULL, "14617388.ogg"},
	{0x90804529083D6AD1ULL, "366880011.ogg"},
	{0x9080C15D48645552ULL, "art\\characters\\creatures\\desiredemon\\animation\\p_wkf.ani"},
	{0x9082059C184300B8ULL, "vxlt_0_lowtown_a2_n_c_ehq_0_0.vlm"},
	{0x90834A850034FDD2ULL, "vxdr_0_deeproads_7tw_0_0.vlm"},
	{0x908356FD0C51D772ULL, "vxlt_0_lowtown_a2_cms_0_0.vlm"},
	{0x908364E0F7D1144AULL, "art\\characters\\playercharacter\\elfmale\\rob_appb.phy"},
	{0x90838A2FCBBB2226ULL, "gen00fl_bethany6101204.cl"},
	{0x90838EFC4CC85C29ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrockjagged.phy"},
	{0x9083C31DB677ECBEULL, "6018696_m.fxe"},
	{0x9084C1C9954508F6ULL, "6076443_m.fxe"},
	{0x9084C374A5FDA269ULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkb.ani"},
	{0x9085126832A91209ULL, "hlac_lt_gallowstemplar_2383_1x.dds"},
	{0x90852F22EE380DBCULL, "6040004_m.fxe"},
	{0x9085F4BCB95046A5ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_disp_p.impa.gad"},
	{0x9086519DFBA93D71ULL, "ts_demonic_wpn_props.gda"},
	{0x9086E4E2AFBF55B8ULL, "gen_im_trs_dwv_junk2_02.uti"},
	{0x908770330DFC3A43ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.mmh"},
	{0x908774E62DB2FCD9ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_basic_1.mao"},
	{0x90882831D32A860DULL, "497298785.ogg"},
	{0x90885E1FD298F964ULL, "rdr111st_harbormaster.stg"},
	{0x90889AC6694D6CC0ULL, "hlac_w_sundermount_196607_0x.dds"},
	{0x9088E0BFDBE9B604ULL, "lav_crust01d.dds"},
	{0x9088EA9B678D0D73ULL, "hlac_w_encounter2_262142_0z.dds"},
	{0x90890B36EDC4AB35ULL, "dre000_varric6034627.cl"},
	{0x9089348445874FFFULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.cess.evt"},
	{0x90894AFCAAC5B1DCULL, "hlac_w_bonepit_0_7y.dds"},
	{0x9089582F6FEF47C7ULL, "item_materials.gda"},
	{0x908962935442D9AAULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglebase.phy"},
	{0x908985F1B20D48EEULL, "639290285.ogg"},
	{0x908A4ACF8AD1F9F1ULL, "vxlt_gallowstemplar_n_5pe_0_0.vlm"},
	{0x908A660AAB602E9EULL, "terrain2.psh.fxo"},
	{0x908B012EEDB35A04ULL, "art\\levels\\races\\natural\\highlands\\vista\\vista_fogoccluder.msh"},
	{0x908B132EEDB3796DULL, "art\\levels\\races\\natural\\highlands\\vista\\vista_fogoccluder.mao"},
	{0x908B172EEDB37FA6ULL, "art\\levels\\races\\natural\\highlands\\vista\\vista_fogoccluder.mmh"},
	{0x908B2B3110DB3644ULL, "381690562.ogg"},
	{0x908B2F35746A6B10ULL, "and330ip_sal_petrae.ncs"},
	{0x908B7F09947E1941ULL, "35771262.ogg"},
	{0x908BDE74A603EEEBULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkb.gad"},
	{0x908BE4B07AE61482ULL, "642953892.ogg"},
	{0x908C2E24BE82FC26ULL, "lowtown_male_lowlife1_sta6164949.cl"},
	{0x908C9F8D7A7D6BBEULL, "vxht_keep_n_5x9_0_0.vlm"},
	{0x908D0EE05040F917ULL, "hlac_lt_0_lowtown_docks_a3_-4_3z.dds"},
	{0x908D48F4007B0C9FULL, "w_encounter3_130_0.msh"},
	{0x908D4EF4007B1741ULL, "w_encounter3_130_0.mmh"},
	{0x908D5AF4007B2BE2ULL, "w_encounter3_130_0.mao"},
	{0x908D8442A4DE2EDBULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_resolved.mmh"},
	{0x908D9042A4DE4378ULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_resolved.mao"},
	{0x908DA242A4DE61C5ULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_resolved.msh"},
	{0x908DA7EE427CD04BULL, "isa311_party_comments6092031.cl"},
	{0x908DC56AA1E2A0DFULL, "hlac_ht_chantry_2534_44x.dds"},
	{0x908DEF78C6F5F47EULL, "gen00fl_anders6113014.cl"},
	{0x908EBAD2BCD9D145ULL, "vxlt_0_lowtown_a2_n_fab_0_0.vlm"},
	{0x908FA15328AA9A86ULL, "hla_ht_0_hightown_a2_2_0x.dds"},
	{0x908FC3865BF223DAULL, "fx_chest.mao"},
	{0x908FCF4826660697ULL, "seb221_flora_harimann6177700.cl"},
	{0x9090080757B2F297ULL, "6156582_m.fxe"},
	{0x90901584458D39E8ULL, "art\\vfx\\combat\\vfx_glb_obscured_crust_c.cess.gad"},
	{0x9090C18807EAD728ULL, "6083389_f.fxe"},
	{0x909109381FDE9128ULL, "vxlt_gallowscourtyard_72a_0_0.vlm"},
	{0x9091E9783525ABD9ULL, "vxlt_warehouse_n_sz_0_0.vlm"},
	{0x9092181526BE61F9ULL, "6211953_m.fxe"},
	{0x9092966C3CFE9CC5ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzfirelarge.phy"},
	{0x9092DB74A60A0850ULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkb.evt"},
	{0x90932904A84DB83FULL, "hm_informant.mor"},
	{0x90938529FC8AEDD5ULL, "6033137_m.fxe"},
	{0x9093E3DCB145EF95ULL, "6162499_m.fxe"},
	{0x909457F411F8A070ULL, "6152646_m.fxe"},
	{0x9094957E6F95CFBDULL, "gen_im_arm_cht_nobc_t6.uti"},
	{0x9094A3BAA9A32229ULL, "6195829_f.fxe"},
	{0x9094C2B5B2F1BC0BULL, "ht_keep_3740.anb"},
	{0x90954D6AA10A47F3ULL, "hlac_w_bonepitdestroyed_a3_393220_0x.dds"},
	{0x9095EE6E08176E29ULL, "w_bonepit_80_0.mao"},
	{0x9095FA6E081782C2ULL, "w_bonepit_80_0.mmh"},
	{0x9095FC6E08178678ULL, "w_bonepit_80_0.msh"},
	{0x9095FC7C9785DADAULL, "lgt210ip_ingredient3.utp"},
	{0x909644E58ABA906DULL, "6055058_m.fxe"},
	{0x90965D0C8C0FD6A8ULL, "dre000_hireling16038604.cl"},
	{0x909684858CE4DE4AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeb.mao"},
	{0x909688858CE4E499ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeb.mmh"},
	{0x909692858CE4F587ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeb.msh"},
	{0x9096BD81D041D5ABULL, "zz_otl_debug6060940.cl"},
	{0x9096CEB32EF41886ULL, "541233323.ogg"},
	{0x90972D13743A3AB2ULL, "vxlt_backalley_g_1ex_0_0.vlm"},
	{0x909753A09D3D1BEBULL, "6196717_m.fxe"},
	{0x9097757FC9118642ULL, "353761247.ogg"},
	{0x9097B4F2DE862E2FULL, "6105689_m.fxe"},
	{0x90987CBA52012589ULL, "423063591.ogg"},
	{0x909896BF1A347224ULL, "mcr371ip_to_templar_hall.utp"},
	{0x9098A228D1C34A60ULL, "pro000_colonel_ewald6039970.cl"},
	{0x9098ADE8E679E2CAULL, "pro000_merchant6051563.cl"},
	{0x9098D1AF93517A73ULL, "vxlt_gallowstemplar_2p3_0_0.vlm"},
	{0x9098F3C4B8454DA0ULL, "art\\characters\\playercharacter\\humanmale\\arm_nuda.phy"},
	{0x909979CE0E398FE8ULL, "llac_w_sundermount_-327684_0.dds"},
	{0x9099C4E49C55B857ULL, "606328088.ogg"},
	{0x909A46342A11C4C9ULL, "default_stage.stg"},
	{0x909A98F73F6E8E84ULL, "vxht_mansion2b_3e1_0_0.vlm"},
	{0x909AFDF02D32F4BFULL, "hlac_w_sundermount_-131076_0y.dds"},
	{0x909B7088A304792CULL, "l_lgt351ar_rand_1.bnk"},
	{0x909B76DC67C47852ULL, "6044662_m.fxe"},
	{0x909B90A5B36F4E4BULL, "514119664.ogg"},
	{0x909BCFEAAC472E58ULL, "vxdr_0_deeproads_cwb_0_2.vlm"},
	{0x909BE229FA2F8F91ULL, "qun131_party_comment6117301.cl"},
	{0x909C2AF2FB81F47BULL, "hlac_w_woundedcoast_-196603_0x.dds"},
	{0x909C2CD96A4BF7F5ULL, "vxht_0_hightown_a3_bcb_0_0.vlm"},
	{0x909CB060F1E4CFF4ULL, "6127493_m.fxe"},
	{0x909CB9DE389D86DCULL, "loginviewtos.gfx"},
	{0x909CC104EB6AD7C1ULL, "621050966.ogg"},
	{0x909D9FE0C35DE591ULL, "vxdr_primevalentrance_1o0_0_1.vlm"},
	{0x909E1EAA0B454C70ULL, "lgt261tr_spider_ambush.ncs"},
	{0x909E3E4ACAE1CAF6ULL, "vxlt_gallowstemplar_f_685_0_0.vlm"},
	{0x909E7BE4785AFF63ULL, "seb321_leliana6181828.cl"},
	{0x909ECB3142EE8E34ULL, "vxca_caverns2_95x_0_0.vlm"},
	{0x909EE22E4E2F2A56ULL, "llac_w_encounter2_camp_327684_0.dds"},
	{0x909F9BDE67EC4D1AULL, "mrl150_merrill6070802.cl"},
	{0x909FC59BC37A6D82ULL, "hlac_ht_0_hightown_65532_0z.dds"},
	{0x90A0053E75C7D9E6ULL, "mer140_bonepit_comments6077808.cl"},
	{0x90A01FE642ED19A6ULL, "hlac_ht_keep_4519_13z.dds"},
	{0x90A05B55271BC459ULL, "vxlt_warehouse_n_vy_0_0.vlm"},
	{0x90A065C246E430C9ULL, "vxw_encounter1_gi_0_0.vlm"},
	{0x90A07FC4EB54153EULL, "83544992.ogg"},
	{0x90A09CBF4ED2F040ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_dssba_0t.dds"},
	{0x90A0B1EDADD2B33BULL, "6060728_m.fxe"},
	{0x90A0EC62CA66C9B7ULL, "824285394.ogg"},
	{0x90A14D1ACCC5D225ULL, "vxlt_gallowscourtyard_n_693_0_0.vlm"},
	{0x90A19FEB33D02FB1ULL, "qun211cr_henchman2.utc"},
	{0x90A22555BD46CF48ULL, "6166681_m.fxe"},
	{0x90A261F11FF96B71ULL, "mer120_hubert6070211.cl"},
	{0x90A2821D73458242ULL, "603454569.ogg"},
	{0x90A2FD8539CC0DB8ULL, "425176311.ogg"},
	{0x90A2FF79A31CAC8BULL, "vxlt_0_lowtown_docks_n_c5h_0_0.vlm"},
	{0x90A30A4B8685394AULL, "art\\vfx\\combat\\vfx_glb_ele_f_additive_c.impa.gad"},
	{0x90A313D950F6D8C6ULL, "67904639.ogg"},
	{0x90A36BC3F4E0A169ULL, "mag170_thrask6016946.cl"},
	{0x90A3966A6EA724ABULL, "lla_ht_0_hightown_a3_-1_0.dds"},
	{0x90A3C44B9EC0ECA5ULL, "vxht_chantry_2tf_0_0.vlm"},
	{0x90A4181295448F3EULL, "art\\characters\\weapons\\unique\\textures\\w_vael_0n.dds"},
	{0x90A47F15C5189BCBULL, "mag311_keran6041503.cl"},
	{0x90A4C6BA490059FCULL, "w_woundedcoast_123_0.phy"},
	{0x90A5047841146A03ULL, "38426963.ogg"},
	{0x90A58122641F6A71ULL, "mcr371_scared_mage_m6087698.cl"},
	{0x90A795EAFD58F75EULL, "gen00fl_sebastian6174067.cl"},
	{0x90A7ED0C35139C9CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse1cloth_orsinosdefiance.ani"},
	{0x90A8933B91E0FF7BULL, "6119262_m.fxe"},
	{0x90A89E3C78081692ULL, "vxlt_0_lowtown_n_f51_0_0.vlm"},
	{0x90A8E50F44C84D00ULL, "6135295_f.fxe"},
	{0x90A93B60534BE5B7ULL, "vxlt_0_lowtown_n_h85_0_0.vlm"},
	{0x90A9D54AC2240EC2ULL, "701402077.ogg"},
	{0x90A9ED4B868B26E1ULL, "art\\vfx\\combat\\vfx_glb_ele_f_additive_c.impa.evt"},
	{0x90AA1898AF1F2BEBULL, "follower_banter6160777.cl"},
	{0x90AAD55C508137B8ULL, "6066191_f.fxe"},
	{0x90AAD9803CCAC826ULL, "6175953_m.fxe"},
	{0x90AAEB572B728B69ULL, "character_rendertodepth_p3.vsh.fxo"},
	{0x90AB22E2171307C3ULL, "lgt311_shw_mer_mettin6156539.cl"},
	{0x90AB32049A0FF462ULL, "cod_ltr_olivia_wayward.utp"},
	{0x90AB39811928D346ULL, "hlac_lt_0_lowtown_65537_1x.dds"},
	{0x90AC2DA73269724CULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzcrossbones.msh"},
	{0x90AC33A732697C3EULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzcrossbones.mmh"},
	{0x90AC3FA7326990D5ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzcrossbones.mao"},
	{0x90AC4D6A2F91850BULL, "946502500.ogg"},
	{0x90AC4DF6C7B56E39ULL, "vfx_fluff01.dds"},
	{0x90AC9A3E7074E253ULL, "darktown_fem_fer_urch_sta6174804.cl"},
	{0x90ACB4795DA70271ULL, "82104378.ogg"},
	{0x90ACF40FB3290E72ULL, "396781781.ogg"},
	{0x90AD30ECC5DE7BDBULL, "6179439_m.fxe"},
	{0x90AD9C865C0ADE8CULL, "fx_chest.dds"},
	{0x90ADA3A2B955B9EEULL, "art\\levels\\races\\design\\prp_moneypouchb.mao"},
	{0x90ADF154792CE79FULL, "al_merril_s01.gda"},
	{0x90AE4818804229D1ULL, "6201471_m.fxe"},
	{0x90AEBBC48AEBBA7AULL, "nat300pt_find_nathaniel.ncs"},
	{0x90AEFA0C3519D1DEULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse1cloth_orsinosdefiance.gad"},
	{0x90AF1DF98B22A9E3ULL, "657910551.ogg"},
	{0x90AF77FC4CED704EULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrockjagged.mmh"},
	{0x90AF83FC4CED84E5ULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrockjagged.mao"},
	{0x90AF91FC4CED9C3CULL, "art\\levels\\races\\natural\\proxy\\props\\prp_zzrockjagged.msh"},
	{0x90AFC716BEFAD06AULL, "hlac_w_bonepitmining_a2_131078_0x.dds"},
	{0x90B037A472470E80ULL, "mer210_seldrek6078801.cl"},
	{0x90B09E7E685396D6ULL, "vxlt_foundry_n_xz_0_0.vlm"},
	{0x90B0EE3D7268F0C0ULL, "one110_pryce6088866.cl"},
	{0x90B17AA653C82C69ULL, "211490813.ogg"},
	{0x90B1AE613AD05C50ULL, "art\\vfx\\combat\\vfx_m_arc_conj_c.mmh"},
	{0x90B1F095E5C1EB78ULL, "store_dae110cr_store4.utm"},
	{0x90B1FFB21B5FA516ULL, "zz_firestaff2_forest.cut"},
	{0x90B264A2699BEB2AULL, "6070197_m.fxe"},
	{0x90B286EC8F903191ULL, "729032321.ogg"},
	{0x90B36276958CDF20ULL, "hlac_lt_0_lowtown_-2_6x.dds"},
	{0x90B3BE803BFE0534ULL, "vxlt_0_lowtown_docks_a3_n_d01_0_1.vlm"},
	{0x90B463B510BC9028ULL, "vxht_0_hightown_a3_bu9_0_0.vlm"},
	{0x90B46E61EC3972B5ULL, "gen_im_trs_prf_amu_r1.uti"},
	{0x90B4E6AD7F80637FULL, "6174142_m.fxe"},
	{0x90B58A848F4AE6D7ULL, "197768799.ogg"},
	{0x90B59A8B7C01FB98ULL, "vxw_bonepit_20l_0_0.vlm"},
	{0x90B5A06660B44371ULL, "vxlt_gallowstemplar_n_5az_0_0.vlm"},
	{0x90B5BD5C9F328B43ULL, "593568111.ogg"},
	{0x90B62475AD798E86ULL, "6076573_m.fxe"},
	{0x90B664F4CA05E552ULL, "6034683_m.fxe"},
	{0x90B67EA61C80CF73ULL, "follower_banter6155889.cl"},
	{0x90B686289B8E121CULL, "gen00fl_fenris6108090.cl"},
	{0x90B69346912AAB11ULL, "llac_ht_0_hightown_n_65537_1.dds"},
	{0x90B7A33F371DCDA0ULL, "572568753.ogg"},
	{0x90B7E54B86975630ULL, "art\\vfx\\combat\\vfx_glb_ele_f_additive_c.impa.ani"},
	{0x90B8650610CE68A3ULL, "isa211_isabela6081988.cl"},
	{0x90B88558D035C2A3ULL, "238356114.ogg"},
	{0x90B8E20D1F3D742DULL, "hlac_ca_caverns1_24421_12z.dds"},
	{0x90B92585860FFBE5ULL, "114999925.ogg"},
	{0x90B92D542ECE6D34ULL, "928802481.ogg"},
	{0x90B93EAC47B4C546ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemb_bdy_0s.dds"},
	{0x90B94FF400A053BEULL, "w_encounter3_130_0.phy"},
	{0x90B96082D2B66A5BULL, "vxlt_gamlenshouse_n_61_0_0.vlm"},
	{0x90B9A199DCA56C78ULL, "vxw_encounter2_camp_lx_0_0.vlm"},
	{0x90BA6C6EDAB98703ULL, "hlac_lt_gallowscourtyard_n_-65537_1z.dds"},
	{0x90BA72906864567CULL, "lt_0_lowtown_a2_n_34_0.rml"},
	{0x90BB35E2F6E5EBF6ULL, "20626039.ogg"},
	{0x90BB846CDB4A4614ULL, "700358551.ogg"},
	{0x90BBE57CB856E5E7ULL, "hlac_w_encounter2_camp_-65531_0z.dds"},
	{0x90BC3D6EA96E67C3ULL, "w_bonepitmining_a2_159_0.msh"},
	{0x90BC4B6EA96E7F1DULL, "w_bonepitmining_a2_159_0.mmh"},
	{0x90BC4F6EA96E862EULL, "w_bonepitmining_a2_159_0.mao"},
	{0x90BCDD26AD5F3696ULL, "hlac_ca_caverns1_24421_2z.dds"},
	{0x90BD5B06A465F0E6ULL, "zz_dre_debug6054661.cl"},
	{0x90BDA11139B7BCE9ULL, "160778916.ogg"},
	{0x90BDB5F1B7A640F3ULL, "vxht_0_hightown_n_a4t_0_0.vlm"},
	{0x90BDE2504BB6BD7AULL, "671821041.ogg"},
	{0x90BDEBCD2CFF9D49ULL, "vxdr_primevaltomb_9cq_0_0.vlm"},
	{0x90BE10C4B8649B35ULL, "art\\characters\\playercharacter\\humanmale\\arm_nuda.msh"},
	{0x90BE12C4B8649EABULL, "art\\characters\\playercharacter\\humanmale\\arm_nuda.mmh"},
	{0x90BE1EC4B864B348ULL, "art\\characters\\playercharacter\\humanmale\\arm_nuda.mao"},
	{0x90BE2C8DA9248C15ULL, "vxlt_gallowstemplar_n_376_0_0.vlm"},
	{0x90BE904359A74251ULL, "6147735_m.fxe"},
	{0x90BEC170AC1E3CABULL, "848117336.ogg"},
	{0x90BEC3F50C1521BDULL, "vxw_bonepitmining_a2_9t_0_0.vlm"},
	{0x90BED1613ADB226FULL, "art\\vfx\\combat\\vfx_m_arc_conj_c.anb"},
	{0x90BF42F18F99B51CULL, "vxlt_gallowstemplar_f_5t0_0_0.vlm"},
	{0x90BF6A7E2D5BDBDEULL, "21074729.ogg"},
	{0x90C01AABA23D38C0ULL, "599881945.ogg"},
	{0x90C0612BA9826150ULL, "vxht_brothel_n_5ci_0_0.vlm"},
	{0x90C07B062CF82C60ULL, "285387393.ogg"},
	{0x90C09428A21C969BULL, "vxht_mansion2_2ac_0_0.vlm"},
	{0x90C0B5480B215D4DULL, "vxlt_0_lowtown_docks_a3_cx1_0_0.vlm"},
	{0x90C0BFB5B316F257ULL, "ht_keep_3740.rml"},
	{0x90C1045FA138FF12ULL, "lgt260_eau_tmp_asst_harbor6143624.cl"},
	{0x90C11B16E900BA75ULL, "hlac_lt_hangedman_n_-1_10x.dds"},
	{0x90C2BC0D335999F6ULL, "6165093_m.fxe"},
	{0x90C3C4C0591A60E6ULL, "6176624_m.fxe"},
	{0x90C48E5E98692268ULL, "6043236_m.fxe"},
	{0x90C49A9A2B16F366ULL, "ht_0_hightown_11_0.rml"},
	{0x90C4AB26830B5651ULL, "gen00fl_fenris6107975.cl"},
	{0x90C4FE8164101D57ULL, "vxlt_0_lowtown_e87_0_0.vlm"},
	{0x90C57FF31801C351ULL, "art\\characters\\playercharacter\\elffemale\\animation\\blendtree\\ef_exp_idle.anb"},
	{0x90C5BCFA40914896ULL, "mag100_cullen6140986.cl"},
	{0x90C5F838899AAB49ULL, "647124430.ogg"},
	{0x90C69565EB9A8F4CULL, "445175020.ogg"},
	{0x90C718ADF8C39B68ULL, "w_blightlands_133_0.msh"},
	{0x90C72AADF8C3B9F9ULL, "w_blightlands_133_0.mao"},
	{0x90C736ADF8C3CD92ULL, "w_blightlands_133_0.mmh"},
	{0x90C76D2B11906584ULL, "6048256_f.fxe"},
	{0x90C84F6194DE90FBULL, "77578151.ogg"},
	{0x90C91EDCB4D3E70FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_med.dura.gad"},
	{0x90C9A9A52A53B0A4ULL, "gen00fl_anders6114466.cl"},
	{0x90CA3DA821B37574ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantrya_d.dds"},
	{0x90CABD23571A647FULL, "vxw_bonepitmining_a2_lq_0_0.vlm"},
	{0x90CAC0E1E0E92075ULL, "6047651_m.fxe"},
	{0x90CACDF28D808763ULL, "gen_im_cns_flu_fenris_01.uti"},
	{0x90CB4E54260AA5E3ULL, "190539477.ogg"},
	{0x90CB72E46326A040ULL, "400265520.ogg"},
	{0x90CBB6B5C1E3F661ULL, "dae000_bartrand6201371.cl"},
	{0x90CBBBC8B515CCE6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_filecabinet_sa.dds"},
	{0x90CC38B62DA02B21ULL, "vxlt_gallowstemplar_f_3r1_0_0.vlm"},
	{0x90CC5C3DB28BA53BULL, "vxht_mansion2b_1a8_0_0.vlm"},
	{0x90CC5FEC7FBE34CDULL, "vxw_woundedcoast_1g_0_0.vlm"},
	{0x90CD362744FF5FE8ULL, "vxlt_hangedman_27p_0_0.vlm"},
	{0x90CD99D9DF8909ECULL, "llac_w_encounter3_-327679_0.dds"},
	{0x90CDEC161194EFC5ULL, "6042252_m.fxe"},
	{0x90CF06059412AF1BULL, "vxht_mansion2_2jb_0_0.vlm"},
	{0x90CF2DD980E62630ULL, "6086200_m.fxe"},
	{0x90CF33DCB4D876BCULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_med.dura.evt"},
	{0x90CF8574AEA70CD1ULL, "vxlt_hangedman_n_1xx_0_0.vlm"},
	{0x90CF9F24124722ACULL, "6034627_m.fxe"},
	{0x90CFAC19AFEF6444ULL, "6136777_f.fxe"},
	{0x90D03CDE91ABC7BAULL, "vxlt_gallowstemplar_n_5vt_0_0.vlm"},
	{0x90D09181656070E5ULL, "831745674.ogg"},
	{0x90D09AF12C9A7B2EULL, "173944837.ogg"},
	{0x90D179CA4F54C9A1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_s1_movement.anb"},
	{0x90D17A8428B433A9ULL, "224224104.ogg"},
	{0x90D19C184774141DULL, "gen00fl_merrill6129597.cl"},
	{0x90D1A3E7662ED7F4ULL, "vxlt_0_lowtown_docks_bcs_0_0.vlm"},
	{0x90D1C9B1E894E5FAULL, "6076021_m.fxe"},
	{0x90D1DDD066DC3AF2ULL, "vxdr_primevaltomb_4xr_0_0.vlm"},
	{0x90D248E1BD33BBA4ULL, "6068569_m.fxe"},
	{0x90D2B3E00C8CF998ULL, "6103368_f.fxe"},
	{0x90D313334F072B72ULL, "6177700_m.fxe"},
	{0x90D358CD302FBB5EULL, "vxlt_0_lowtown_docks_a3_bba_0_0.vlm"},
	{0x90D381AEF5149B75ULL, "6040701_f.fxe"},
	{0x90D3B3C2292E9DA8ULL, "6079013_m.fxe"},
	{0x90D3E786DAE82DD4ULL, "vxht_mansion_n_5ax_0_0.vlm"},
	{0x90D4487C6D26CC69ULL, "mer320_hubert6096165.cl"},
	{0x90D45DDE6DC9A260ULL, "285854165.ogg"},
	{0x90D47338EE7F01ACULL, "221874172.ogg"},
	{0x90D49F60BFC2F6ACULL, "qun111_ketojan6039181.cl"},
	{0x90D4A6BC4DF5BEACULL, "mud_sh.dds"},
	{0x90D4D176DB2A68F3ULL, "dae000_worthy6166633.cl"},
	{0x90D4D82DE9F8C2B2ULL, "508555406.ogg"},
	{0x90D4DF3DBF609225ULL, "hlac_ht_mansion2_818_59y.dds"},
	{0x90D50432CAED1669ULL, "6066113_m.fxe"},
	{0x90D510C3180A1A08ULL, "6170794_m.fxe"},
	{0x90D5563CFD0BBFA1ULL, "vxlt_0_lowtown_docks_n_b8x_0_0.vlm"},
	{0x90D5A542A51C1C10ULL, "art\\levels\\races\\proxy\\props\\4tools\\prp_resolved.phy"},
	{0x90D6404A5ADE7F73ULL, "crt_aoe_01.mao"},
	{0x90D6641BFB08436BULL, "lt_0_lowtown_docks_a3_45_0.rml"},
	{0x90D67CA7E50AFEBEULL, "6071240_m.fxe"},
	{0x90D7D298F7ADECD4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grill_sa.dds"},
	{0x90D7D9121FF47979ULL, "6032134_m.fxe"},
	{0x90D81DA1446C95FBULL, "795692883.ogg"},
	{0x90D8D0292AD29034ULL, "hlac_lt_0_lowtown_-65535_3y.dds"},
	{0x90D92F0049107CFFULL, "vxw_bonepitmining_a2_kp_0_0.vlm"},
	{0x90DA263574AE3019ULL, "and330ip_sal_petrae.utp"},
	{0x90DA991DA1803AFCULL, "vxlt_0_lowtown_docks_cvt_0_1.vlm"},
	{0x90DACFCBC9FB3D82ULL, "6173414_m.fxe"},
	{0x90DB11C8B0B0B4FCULL, "240127473.ogg"},
	{0x90DB20B53B6E783EULL, "qun211cr_elf_warrior6072193.cl"},
	{0x90DB714E41512199ULL, "hm_aden.mor"},
	{0x90DBC0C5CBAE33CEULL, "vxw_bonepitmining_a2_hp_0_0.vlm"},
	{0x90DC587D012FEF0DULL, "vxlt_0_lowtown_docks_cq5_0_0.vlm"},
	{0x90DC7A88A33C215EULL, "l_lgt351ar_rand_1.wbd"},
	{0x90DC98BE0AB2EC0EULL, "vxca_caverns1_mines_e7j_0_0.vlm"},
	{0x90DCA108A4786B53ULL, "1015964935.ogg"},
	{0x90DD010664C3545EULL, "hlac_w_sundermount_327674_0z.dds"},
	{0x90DD1C80E4C26D18ULL, "zz_qun_debug6028294.cl"},
	{0x90DD3BDCB4E4C1D5ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_med.dura.ani"},
	{0x90DDE013AE136992ULL, "vxlt_warehouse_n_10k_0_0.vlm"},
	{0x90DE03CE91E38202ULL, "gen00fl_isabela6098223.cl"},
	{0x90DE62491B6ECDE2ULL, "gen_im_wep_war_two_act2_03.uti"},
	{0x90DE6B9217C9A53FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadcth.phy"},
	{0x90DEAEBB83A1E03AULL, "vxlt_gallowstemplar_n_5iu_0_0.vlm"},
	{0x90DEC7353A84CA4EULL, "6176743_m.fxe"},
	{0x90DF38C258C0DAADULL, "973849931.ogg"},
	{0x90DF826E787FB9FBULL, "139968317.ogg"},
	{0x90DFF01867BD455EULL, "12549070.ogg"},
	{0x90E10306A8088B57ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_104a_n.dds"},
	{0x90E1159DCED14937ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_lgta_0n.dds"},
	{0x90E236EE20CB9FFBULL, "6178429_m.fxe"},
	{0x90E250234A2D97F7ULL, "490096623.ogg"},
	{0x90E2DD1A37045CD1ULL, "877499481.ogg"},
	{0x90E2E679E15133ECULL, "6062124_f.fxe"},
	{0x90E328A4CD5B0610ULL, "hlac_lt_undercitydungeon_-1_25x.dds"},
	{0x90E329A78B0A2F45ULL, "llac_w_bonepitdestroyed_a3_131075_0.dds"},
	{0x90E348A61AA57D13ULL, "9614629.ogg"},
	{0x90E3B00B4FB1E488ULL, "gen_im_npc_elf_arm_hlm_dala.uti"},
	{0x90E51A56384A5B90ULL, "351841297.ogg"},
	{0x90E51D1FBA24DFA5ULL, "gen00fl_aveline6110361.cl"},
	{0x90E52242E83741E5ULL, "mag310ip_store_arthuris.utp"},
	{0x90E54B5BDA338162ULL, "vxht_mansion2b_1hb_0_0.vlm"},
	{0x90E5553EBFCD3CF4ULL, "zz_mcr_debug6080823.cl"},
	{0x90E5E9EFB0E57A49ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05a.phy"},
	{0x90E621DA1703BC5CULL, "gen00fl_anders6125237.cl"},
	{0x90E67390688992CCULL, "lt_0_lowtown_a2_n_34_0.anb"},
	{0x90E70DB7B4EB641CULL, "dae000cr_commoner_f2.utc"},
	{0x90E773EE1768BE35ULL, "vxlt_undercityserialkiller_13y_0_0.vlm"},
	{0x90E77552C123ED7BULL, "6094554_m.fxe"},
	{0x90E7D8F3F33ACBBFULL, "6030674_f.fxe"},
	{0x90E7F92B590D1E05ULL, "vxw_encounter2_camp_la_0_0.vlm"},
	{0x90E803A621C59EF1ULL, "vxlt_undercityserialkiller_1c4_0_1.vlm"},
	{0x90E859CBAF6BE330ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_walllcap.mao"},
	{0x90E85DCBAF6BEA43ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_walllcap.mmh"},
	{0x90E86B055100BB8BULL, "gen00fl_merrill6150700.cl"},
	{0x90E86BCBAF6C019DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_walllcap.msh"},
	{0x90E8D36E73B6F55AULL, "vxlt_0_lowtown_docks_b7x_0_1.vlm"},
	{0x90E8F6F96EAE21E3ULL, "136368781.ogg"},
	{0x90E958E4C20FB886ULL, "6172754_m.fxe"},
	{0x90E97DD530CB3396ULL, "w_woundedcoast_54_0.msh"},
	{0x90E987D530CB4494ULL, "w_woundedcoast_54_0.mmh"},
	{0x90E98BD530CB4BA7ULL, "w_woundedcoast_54_0.mao"},
	{0x90E9C8FA3DE6B63FULL, "648766325.ogg"},
	{0x90E9E62733378CE6ULL, "702489527.ogg"},
	{0x90EA3937945DF67EULL, "hlac_lt_0_lowtown_a2_n_c_-2_5x.dds"},
	{0x90EA5A66C678105FULL, "6080696_m.fxe"},
	{0x90EA9CF3178DA63FULL, "mrl150_merrill6078333.cl"},
	{0x90EAC86E19AE6E9CULL, "183464871.ogg"},
	{0x90EBCBF2B826CDEEULL, "607038765.ogg"},
	{0x90EBDE395CA6C4E6ULL, "vxlt_0_lowtown_a2_n_ex1_0_0.vlm"},
	{0x90EBE9BF77FEE01BULL, "vxht_mansion2_2pd_0_0.vlm"},
	{0x90EC0B5D6E8F16A3ULL, "179350103.ogg"},
	{0x90EC15AA0B875E7DULL, "lgt261tr_spider_ambush.utt"},
	{0x90ED0605DF287975ULL, "522210348.ogg"},
	{0x90ED20987597F8BAULL, "vxlt_gallowstemplar_n_5lv_0_0.vlm"},
	{0x90ED41DC6810E77EULL, "622356213.ogg"},
	{0x90ED8C05FA289468ULL, "vxlt_gallowsprison_n_5r2_0_0.vlm"},
	{0x90ED93DECD2D11EBULL, "561545816.ogg"},
	{0x90EDB246FDFC9FFFULL, "ser221_party_comments6209890.cl"},
	{0x90EDBDC44CCF98A2ULL, "hlac_lt_0_lowtown_0_3x.dds"},
	{0x90EDF25301707B24ULL, "10648528.ogg"},
	{0x90EE00BEBE19F3E4ULL, "gen00fl_fenris6114518.cl"},
	{0x90EE958F72B71C3EULL, "6034709_f.fxe"},
	{0x90EEC51974A4D77FULL, "vxlt_0_lowtown_n_e88_0_0.vlm"},
	{0x90EF6C9217D81A44ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadcth.mmh"},
	{0x90EF709217D82097ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadcth.mao"},
	{0x90EF829217D83FE6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadcth.msh"},
	{0x90EFD56FC6B8E4B7ULL, "vxht_0_hightown_9za_0_0.vlm"},
	{0x90EFE6534FA9C962ULL, "6147691_m.fxe"},
	{0x90F0722CC475E6C3ULL, "hlac_w_encounter2_camp_65536_0z.dds"},
	{0x90F0AABDC14B5CDCULL, "llac_ht_0_hightown_a2_196607_0.dds"},
	{0x90F0C319677EA6B3ULL, "vxw_bonepitmining_a2_dm_0_0.vlm"},
	{0x90F142F67599E026ULL, "29826397.ogg"},
	{0x90F1475915E57542ULL, "233961589.ogg"},
	{0x90F17F52EC328D5BULL, "6173374_m.fxe"},
	{0x90F2261A300B8A78ULL, "dae000_hate_follower_react6103220.cl"},
	{0x90F22FB575BF3A22ULL, "vxlt_warehouse_12z_0_0.vlm"},
	{0x90F2739FAFE68690ULL, "308476055.ogg"},
	{0x90F298C47ECD5F00ULL, "419973384.ogg"},
	{0x90F298E18F59A52AULL, "zz_one_debug6099757.cl"},
	{0x90F2DB3D02AF58E0ULL, "vxht_brothel_n_29x_0_0.vlm"},
	{0x90F380F9CE6871D4ULL, "follower_banter6161151.cl"},
	{0x90F3BD38CC2999E2ULL, "vxht_mansion2b_1cc_0_0.vlm"},
	{0x90F3C91E7CF77CA9ULL, "vxlt_0_lowtown_docks_a3_n_ctj_0_0.vlm"},
	{0x90F4E550C142D59CULL, "vxw_encounter2_camp_p9_0_1.vlm"},
	{0x90F4EA059712C7B2ULL, "6179892_m.fxe"},
	{0x90F4ECE4BB309CEDULL, "gen00fl_isabela6103962.cl"},
	{0x90F50BAA26BD2731ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsc.mmh"},
	{0x90F517AA26BD3BD2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsc.mao"},
	{0x90F525AA26BD53EFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindowsc.msh"},
	{0x90F573BBAB8272BFULL, "6015993_m.fxe"},
	{0x90F5FA94280501D2ULL, "158744706.ogg"},
	{0x90F71DC40E9D1758ULL, "6064525_m.fxe"},
	{0x90F7A49DDB5FACB5ULL, "6148235_m.fxe"},
	{0x90F7ED8DD02C9C50ULL, "gen00fl_merrill6116145.cl"},
	{0x90F83E30D29DBAFEULL, "6103915_f.fxe"},
	{0x90F841D8E5ABD22CULL, "805916221.ogg"},
	{0x90F889858D37B086ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeb.phy"},
	{0x90F8976969F27E5DULL, "vxca_caverns1_i0i_0_0.vlm"},
	{0x90F95C292B26AB5AULL, "vxlt_0_lowtown_n_d6f_0_0.vlm"},
	{0x90F9DCFFB2BCFB10ULL, "vxlt_0_lowtown_a2_n_go0_0_0.vlm"},
	{0x90F9E4FB743884B8ULL, "qcr220_orsino6041800.cl"},
	{0x90FA021A9D53C81EULL, "hlac_lt_0_lowtown_n_131072_0x.dds"},
	{0x90FA6A3F3B48F297ULL, "hlac_w_sundermount_-65536_0x.dds"},
	{0x90FAA7FCF7E9BFDCULL, "mag170_thrask6016930.cl"},
	{0x90FABD0F503AA50BULL, "6107452_m.fxe"},
	{0x90FAD761FF842AABULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas_n.dds"},
	{0x90FBA531B28E027AULL, "llac_lt_0_lowtown_a2_131073_0.dds"},
	{0x90FBAECED0AE98BFULL, "411543591.ogg"},
	{0x90FC607B3BE46315ULL, "gen00fl_dog6141101.cl"},
	{0x90FC844662A56AEDULL, "hm_karl.mor"},
	{0x90FC8EC0CE166978ULL, "177497645.ogg"},
	{0x90FCAA61FDFECD5BULL, "235082858.ogg"},
	{0x90FD88768CB366CCULL, "ser221_party_comments6164316.cl"},
	{0x90FD9F3DF003B888ULL, "6112615_m.fxe"},
	{0x90FE75A9AE68446BULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.cess.evt"},
	{0x90FEBEB7EFB0C56DULL, "gen00fl_merrill6121096.cl"},
	{0x90FEF0AF6CBD2E45ULL, "1041860861.ogg"},
	{0x90FF12EF86EC8EDBULL, "mag120_macha6080671.cl"},
	{0x90FF34F65974BF33ULL, "vxw_bonepitmining_a2_cl_0_0.vlm"},
	{0x90FF873529AE677CULL, "576479961.ogg"},
	{0x90FFF8FADB60559CULL, "516122249.ogg"},
	{0x910012CC510E46D7ULL, "vxca_caverns1_i9l_0_0.vlm"},
	{0x9100258EF032FE0EULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_exit.evt"},
	{0x91002B8CCBA40AE0ULL, "zz_fol_debug6057006.cl"},
	{0x9100D1132505AC1DULL, "ser221_quentin6129685.cl"},
	{0x91022F15BE1FB262ULL, "vxht_mansion2b_1fl_0_0.vlm"},
	{0x9102D7F66A24F4DAULL, "c_dog_test.tnt"},
	{0x9103009DE7ADA247ULL, "vxlt_darktown_2ra_0_0.vlm"},
	{0x91048EA9AE6CDA2CULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.cess.gad"},
	{0x9104DCE53CF94F05ULL, "vxw_encounter2_camp_bo_0_0.vlm"},
	{0x91055D71FCF03156ULL, "6028930_m.fxe"},
	{0x91056808DB6757F0ULL, "vxlt_0_lowtown_byt_0_0.vlm"},
	{0x91056D826E3466D4ULL, "hlac_ca_caverns1_24423_13z.dds"},
	{0x91064097C5FD8F38ULL, "36432808.ogg"},
	{0x910661FF27F207FDULL, "pro000_cut2d_thefirstyear6015291.cl"},
	{0x91066ACD467BAC8AULL, "light02.mao"},
	{0x9106D78946148476ULL, "79354527.ogg"},
	{0x91071C8EF0390DB9ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_exit.gad"},
	{0x91079DAA3DFDF4A3ULL, "266916676.ogg"},
	{0x9107E4CC5B89E7B7ULL, "6144602_m.fxe"},
	{0x9107F3859BC873F5ULL, "hlac_lt_darktown_-1_1x.dds"},
	{0x91088733BDB81A93ULL, "gen00fl_aveline6106917.cl"},
	{0x91090E83382B774FULL, "lt_0_lowtown_docks_n_140_0.mao"},
	{0x91091A83382B8BECULL, "lt_0_lowtown_docks_n_140_0.mmh"},
	{0x91092083382B95DEULL, "lt_0_lowtown_docks_n_140_0.msh"},
	{0x91093C5EA36BC3F6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grand.msh"},
	{0x9109465EA36BD4F4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grand.mmh"},
	{0x91094A5EA36BDB07ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grand.mao"},
	{0x9109607D9B30F263ULL, "592303836.ogg"},
	{0x9109EC33FC543C55ULL, "vxdr_primevaltomb_5bf_0_0.vlm"},
	{0x910A26EFD898D4DCULL, "one121_trap6088459.cl"},
	{0x910AD385846F41D3ULL, "6002896_m.fxe"},
	{0x910ADA29DBFDCC8AULL, "vxht_bartrand_n_5nt_0_0.vlm"},
	{0x910B7232D156E01EULL, "vxlt_0_lowtown_docks_n_b7f_0_0.vlm"},
	{0x910B8DA9AE72F72AULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.cess.ani"},
	{0x910BBCC030371441ULL, "load_fade.gfx"},
	{0x910C8F7F5951A1A7ULL, "125938889.ogg"},
	{0x910CB46D946EF819ULL, "lt_0_lowtown_docks_n_32_0.anb"},
	{0x910CC93BDE945313ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basic.phy"},
	{0x910D3D68CC8811DEULL, "art\\vfx\\creatures\\fxc_suc_att_s_c.impa.ani"},
	{0x910D4A28904F08E1ULL, "6096770_m.fxe"},
	{0x910D5EDD73CA141FULL, "819060388.ogg"},
	{0x910DA6D34B6AD7B3ULL, "vxw_bonepitmining_a2_no_0_0.vlm"},
	{0x910DEB0EBAEFE3BAULL, "6042798_m.fxe"},
	{0x910E0A38B752674CULL, "vxht_0_hightown_7g3_0_0.vlm"},
	{0x910E2D8EF03F4907ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_exit.ani"},
	{0x910E32011C6C8837ULL, "400516836.ogg"},
	{0x910EBE00E8085E1AULL, "gen00fl_anders6119265.cl"},
	{0x910EC2267768B28AULL, "6198678_f.fxe"},
	{0x910F00D0AEF4C844ULL, "dae000_fem_commoner2_mob6174582.cl"},
	{0x910F932DFA2B0D8AULL, "501933960.ogg"},
	{0x91100F11E7174361ULL, "ser120_ghyslain6104184.cl"},
	{0x91101A2B775818A0ULL, "6154705_m.fxe"},
	{0x9110ACD860E3ADA9ULL, "vxlt_0_lowtown_docks_a3_n_czh_0_0.vlm"},
	{0x91110DB917A244DCULL, "6118323_m.fxe"},
	{0x91113123D0066FC6ULL, "plt_lgt111pt_gxa_coa.plo"},
	{0x9111A16C3BEFCA01ULL, "6136678_m.fxe"},
	{0x91126E8776A64741ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_l_5p.ani"},
	{0x91128299E0D671E8ULL, "vxlt_hangedman_1i5_0_0.vlm"},
	{0x91128F93D576FB45ULL, "hlac_lt_0_lowtown_docks_a3_262143_0x.dds"},
	{0x9112A56F7B2DC40DULL, "follower_banter6169347.cl"},
	{0x9112C37C0FBB6558ULL, "6129312_m.fxe"},
	{0x9112CA09EB64B5C1ULL, "vxlt_warehouse_n_pg_0_0.vlm"},
	{0x9112D2EFB10C416EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05a.mmh"},
	{0x9112ECEFB10C6D5CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye05a.msh"},
	{0x911345739EB0C889ULL, "vxlt_0_lowtown_docks_a3_n_d3x_0_0.vlm"},
	{0x91138C11FCE04D73ULL, "83272993.ogg"},
	{0x9113D6FF4ED6AC53ULL, "615764615.ogg"},
	{0x9113FA8B53A11952ULL, "6132682_m.fxe"},
	{0x911514E63F69C083ULL, "127612594.ogg"},
	{0x91152CF3135786EDULL, "450648747.ogg"},
	{0x9115BEF1C88785BDULL, "genip_door_lowin_residence.utp"},
	{0x9115E3CC6D83826CULL, "vxlt_warehouse_n_dz_0_0.vlm"},
	{0x9116201CE95FDBC4ULL, "gen00fl_varric6093659.cl"},
	{0x91162ACCE3BE9872ULL, "872708075.ogg"},
	{0x9117048A1CC4D727ULL, "vxw_woundedcoast_19x_0_0.vlm"},
	{0x91172681E9140073ULL, "gen00fl_varric6076530.cl"},
	{0x9117338FE554F33DULL, "t3_wdg_gn2.tnt"},
	{0x91173B5430F35692ULL, "and121_karl6093866.cl"},
	{0x91176D31B1F59AE7ULL, "vxlt_gallowstemplar_5x8_0_0.vlm"},
	{0x9117996A382F88D5ULL, "dae000_orana6149094.cl"},
	{0x9117BF8D6935E28EULL, "art\\characters\\weapons\\longbows\\w_lbw_lb102a.msh"},
	{0x9117D98D69360F7CULL, "art\\characters\\weapons\\longbows\\w_lbw_lb102a.mmh"},
	{0x91189965CD4C2EB7ULL, "6104989_m.fxe"},
	{0x9118BCB3C61E003CULL, "vxlt_backalley_g_1wm_0_1.vlm"},
	{0x9118C67510E04EBCULL, "127766027.ogg"},
	{0x9118E75A2231E488ULL, "hlac_lt_0_lowtown_a2_n_-65539_0y.dds"},
	{0x91194EB3B43DEACDULL, "vxw_bonepitmining_a2_1oy_0_0.vlm"},
	{0x9119798776AC78E3ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_l_5p.gad"},
	{0x9119A32FF658069DULL, "197952105.ogg"},
	{0x9119F82AD98A58D0ULL, "gen00fl_aveline6116650.cl"},
	{0x911A32F3DB6F12E9ULL, "6146955_f.fxe"},
	{0x911A3AC480541984ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_rdy.ani"},
	{0x911A5568CC92C4AFULL, "art\\vfx\\creatures\\fxc_suc_att_s_c.impa.evt"},
	{0x911B14029B186726ULL, "766041987.ogg"},
	{0x911BEB35A18A3A76ULL, "art\\vfx\\textures\\vfx_dotblur.dds"},
	{0x911BEC035200A6ACULL, "mer341_jansen6054874.cl"},
	{0x911BF97F36919596ULL, "hlac_lt_0_lowtown_-65537_4x.dds"},
	{0x911C0C2BC23AB90BULL, "216220992.ogg"},
	{0x911C58254D2F0D74ULL, "6093471_m.fxe"},
	{0x911D4B188FCAFFCBULL, "gen_im_npc_arm_hum_glv_cotb.uti"},
	{0x911D83F24083AECAULL, "lgt310_mis_burn_murder16152756.cl"},
	{0x911DE7CA8DD80474ULL, "327047577.ogg"},
	{0x911EE579D41DA9E5ULL, "zz_followers_banter6158091.cl"},
	{0x911F117675AB8954ULL, "345915418.ogg"},
	{0x911F1EAD0AF54F34ULL, "vxlt_gallowstemplar_n_3rp_0_0.vlm"},
	{0x911F7C0F51EF7D95ULL, "atl_quickbar_dxt5_00.dds"},
	{0x911FE457CB99D347ULL, "vxlt_darktown_2xc_0_0.vlm"},
	{0x912004A7806B013BULL, "zz_mrl_debug6184181.cl"},
	{0x9120119C05000CE6ULL, "vxlt_0_lowtown_docks_n_bet_0_0.vlm"},
	{0x9120568776B25BC8ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_180_l_5p.evt"},
	{0x9120E4A043DA20B2ULL, "gen_im_cft_reg_e3.uti"},
	{0x91211321127EB0FDULL, "lt_0_lowtown_docks_a3_191_0.mao"},
	{0x91211721127EB736ULL, "lt_0_lowtown_docks_a3_191_0.mmh"},
	{0x91212121127EC834ULL, "lt_0_lowtown_docks_a3_191_0.msh"},
	{0x912137C4805A33E6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_rdy.gad"},
	{0x91213BE6DD5ACE41ULL, "vxlt_warehouse_n_wf_0_0.vlm"},
	{0x91215668CC98E558ULL, "art\\vfx\\creatures\\fxc_suc_att_s_c.impa.gad"},
	{0x91216DE907E28E9CULL, "6087033_f.fxe"},
	{0x9121F393FF80F093ULL, "vxlt_0_lowtown_n_fgj_0_0.vlm"},
	{0x912296499C54ED76ULL, "515959277.ogg"},
	{0x912332E9D682A3ABULL, "character_rendertoshadowmap_p0.vsh.fxo.dx11"},
	{0x912342A8C53443ACULL, "art\\characters\\playercharacter\\humanmale\\hlh_medj.phy"},
	{0x912349B07600EA95ULL, "vxlt_undercitydungeon_1jb_0_1.vlm"},
	{0x912355B31219AB7BULL, "6086706_m.fxe"},
	{0x91236120467D0233ULL, "302429447.ogg"},
	{0x91245DE0B550388EULL, "gen00fl_merrill6128735.cl"},
	{0x9124EABC4BB4C1B6ULL, "269594504.ogg"},
	{0x9124F87F484F0496ULL, "807984626.ogg"},
	{0x912522F6C62896ABULL, "hlac_lt_0_lowtown_docks_a3_n_-65535_0x.dds"},
	{0x9125468331491D98ULL, "vxw_blightlands_11y_0_0.vlm"},
	{0x9126215AA9993B0FULL, "vxca_caverns2_62p_0_1.vlm"},
	{0x912643F664FD8006ULL, "864098311.ogg"},
	{0x91270F3182390913ULL, "vrc321_varric6084345.cl"},
	{0x912712273A31515AULL, "gen00fl_sebastian6188626.cl"},
	{0x91275F29DE7DBF66ULL, "vxht_0_hightown_n_2xr_0_0.vlm"},
	{0x912859B4777ED71AULL, "458806043.ogg"},
	{0x91287223D5220E16ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcsbl.mao"},
	{0x91287E23D52222B5ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcsbl.mmh"},
	{0x91288023D522262BULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcsbl.msh"},
	{0x912907371E22DB69ULL, "vxlt_0_lowtown_a2_n_gp8_0_0.vlm"},
	{0x91290913849F0629ULL, "81919803.ogg"},
	{0x91292F8EE98F86DCULL, "6135601_m.fxe"},
	{0x91295F457C80F541ULL, "hlac_ca_caverns1_24424_7y.dds"},
	{0x9129895F55F39687ULL, "ser221_gascard.bnk"},
	{0x912989FAE907C0C8ULL, "hlac_lt_0_lowtown_a3_n_c_1_4z.dds"},
	{0x9129A829825FB139ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf04.msh"},
	{0x9129B229825FC227ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf04.mmh"},
	{0x9129F6725FA4888DULL, "lt_0_lowtown_docks_a3_10_0_ll.mmh"},
	{0x912AEA5B4E320C73ULL, "vxht_keep_n_55k_0_0.vlm"},
	{0x912B54730FB964A7ULL, "ave200_donnic6061280.cl"},
	{0x912BA3E5BCD3BB71ULL, "566435376.ogg"},
	{0x912CA562F2DBD778ULL, "hlac_lt_warehouse_n_1_3z.dds"},
	{0x912CAB5F55F606FFULL, "ser221_gascard.cnv"},
	{0x912CB6CD71B80116ULL, "dae000_gamlen6102530.cl"},
	{0x912D037D0BCACBA8ULL, "vxca_cavern1_haunted_imb_0_0.vlm"},
	{0x912D33C9D6654BA0ULL, "74995161.ogg"},
	{0x912D7927D7077288ULL, "gen00flst_isa_mage_crisis1.stg"},
	{0x912D84AC9401FBE2ULL, "vxht_mansion2b_1wo_0_0.vlm"},
	{0x912D98E24BE7807BULL, "243586357.ogg"},
	{0x912EF62A5630EFC0ULL, "gui_guipixelshader_p8.psh.fxo"},
	{0x912F055441D0BE8FULL, "dae000_bartrand6204158.cl"},
	{0x912FADC3CF50E6C1ULL, "vxlt_warehouse_n_ze_0_0.vlm"},
	{0x913071948032494DULL, "llac_ht_mansion2b_818_36.dds"},
	{0x9130C018CCC882E2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04b.phy"},
	{0x9130FA607130BB26ULL, "174318451.ogg"},
	{0x913102A3737F6895ULL, "gen00fl_aveline6139724.cl"},
	{0x91323F6D2835E632ULL, "vxlt_0_lowtown_a2_n_fj4_0_0.vlm"},
	{0x91324B177E92D094ULL, "6042688_m.fxe"},
	{0x91327BFB70650701ULL, "577096319.ogg"},
	{0x913298C851815343ULL, "286146990.ogg"},
	{0x913381DC62A54150ULL, "58476329.ogg"},
	{0x913396E81676A788ULL, "6165169_m.fxe"},
	{0x9133DEE660845301ULL, "6087426_m.fxe"},
	{0x913405F73CBD577BULL, "w_bonepitdestroyed_a3_29_0.msh"},
	{0x913417F73CBD75C6ULL, "w_bonepitdestroyed_a3_29_0.mao"},
	{0x913423F73CBD8A65ULL, "w_bonepitdestroyed_a3_29_0.mmh"},
	{0x913429C2441457DDULL, "vxht_mansion_n_61v_0_0.vlm"},
	{0x91346EFC01EF42C2ULL, "641830644.ogg"},
	{0x9134CDE0B3339AEBULL, "lgt210cr_magu_selby.utc"},
	{0x9135A61D2F46E1A7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medb_d.dds"},
	{0x9135F54B2CD26EF3ULL, "dae000_meredith6166308.cl"},
	{0x913718B841A149D2ULL, "zz_qun_debug6045378.cl"},
	{0x91373CF276FE3320ULL, "6084435_m.fxe"},
	{0x91375F2DF74BF625ULL, "616832252.ogg"},
	{0x91378AC1E088681EULL, "6110079_m.fxe"},
	{0x9137C6E9D9B505D9ULL, "character_character_p200.psh.fxo.dx11.assembly"},
	{0x91385CCF8199B6A9ULL, "llac_dr_0_deeproads_17038_0.dds"},
	{0x9138661C3E3E2E73ULL, "266654297.ogg"},
	{0x91387D518EE8D542ULL, "llac_lt_0_lowtown_docks_n_-65539_2.dds"},
	{0x91389EE6C846FB4EULL, "771083202.ogg"},
	{0x9138CDF8BEBBB4AFULL, "vxca_caverns1_iit_0_0.vlm"},
	{0x913901679D5F2BC0ULL, "821423637.ogg"},
	{0x913A12270A8BE8BEULL, "art\\vfx\\materials\\chunky\\vfx_c_d_slowzone_blend.mao"},
	{0x913A30C97016FD2FULL, "6144927_m.fxe"},
	{0x913AAB00A113715BULL, "zz_and_debug6134151.cl"},
	{0x913B0C6864493284ULL, "llac_ca_cavern1_haunted_24424_2.dds"},
	{0x913B1C0E7C7AD314ULL, "art\\vfx\\creatures\\fxc_reven_pull_p.anb"},
	{0x913B47230A8C9679ULL, "51444092.ogg"},
	{0x913B7559FDC0E428ULL, "vxca_cavern1_haunted_ijc_0_0.vlm"},
	{0x913BBFA5876903D4ULL, "art\\characters\\playercharacter\\humanfemale\\glv_maga.phy"},
	{0x913BED1C34EA2884ULL, "6137698_m.fxe"},
	{0x913C1A50890D9AC0ULL, "gen00fl_merrill6121118.cl"},
	{0x913C3EBED17B2162ULL, "lowtown_fem_urchin_mob6166320.cl"},
	{0x913C7385EFE32BAFULL, "6103962_m.fxe"},
	{0x913DB9396134A21DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05b.phy"},
	{0x913DC2AA5F1FEA31ULL, "vxlt_0_lowtown_docks_aoi_0_0.vlm"},
	{0x913E87D531138497ULL, "w_woundedcoast_54_0.rml"},
	{0x913E96F4918E3A28ULL, "ach_kill_knockout.dds"},
	{0x913ECBD0172840D0ULL, "550409386.ogg"},
	{0x9140B9C0E29B42E1ULL, "rdr111dg_assistant6050917.cl"},
	{0x9140D658F0F0B7C2ULL, "6098704_m.fxe"},
	{0x9141AA9BD70728A0ULL, "ca_caverns1_mines_probe_jze.dds"},
	{0x9142B9EC6AE28E28ULL, "vxca_caverns2prc_9vx_0_0.vlm"},
	{0x9142BAB69653843FULL, "mcr371_scared_mage_m6086912.cl"},
	{0x9143012EA94C551BULL, "zz_dae_debug6029827.cl"},
	{0x91438214CA77521FULL, "6135814_f.fxe"},
	{0x9143B94C8B71407DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appb_0d.dds"},
	{0x9143D51FD7BCFEEDULL, "504321211.ogg"},
	{0x9143FFB3416ACE68ULL, "hlac_w_sundermount_-65540_0z.dds"},
	{0x9144D4E21F24A99BULL, "32748418.ogg"},
	{0x91459462494DA898ULL, "mrl000_marethari6142161.cl"},
	{0x91469CA58772CFD1ULL, "art\\characters\\playercharacter\\humanfemale\\glv_maga.msh"},
	{0x9146AAA58772E7ECULL, "art\\characters\\playercharacter\\humanfemale\\glv_maga.mao"},
	{0x9146B56D94A07219ULL, "lt_0_lowtown_docks_n_32_0.rml"},
	{0x9146B6A58772FB8FULL, "art\\characters\\playercharacter\\humanfemale\\glv_maga.mmh"},
	{0x9146C33BF1032F8BULL, "vxca_caverns2_8g0_0_0.vlm"},
	{0x91472FC3328221B1ULL, "ss_types.gda"},
	{0x9147449A0D75B0AAULL, "ser191_emeric6104738.cl"},
	{0x91474B4DD93DA0F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_left.ani"},
	{0x91478E47F397B343ULL, "458807904.ogg"},
	{0x9147A02349ED075FULL, "hlac_lt_0_lowtown_a2_n_c_65536_0x.dds"},
	{0x914891130759C001ULL, "6077674_m.fxe"},
	{0x9148DBDDE6853BAEULL, "zz_vrc_debug6200183.cl"},
	{0x9148FB0E7C86D8D7ULL, "art\\vfx\\creatures\\fxc_reven_pull_p.mmh"},
	{0x91493F9CF8CA1C76ULL, "962036464.ogg"},
	{0x914994A029376D11ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firecoal_sa.dds"},
	{0x9149A2C42102D458ULL, "6103575_f.fxe"},
	{0x9149C764CB11A229ULL, "dae000_bartrand6030610.cl"},
	{0x9149E3F8BBE608B7ULL, "vxht_chantry_n_26y_0_0.vlm"},
	{0x914A32043957DBF0ULL, "qcr270cr_fake_meredith.utc"},
	{0x914AC84F8A4440D1ULL, "vxca_cavern1_haunted_ebj_0_0.vlm"},
	{0x914B534FFD2FA9D4ULL, "vxw_encounter2_camp_6v_0_0.vlm"},
	{0x914BC922FC2E35CCULL, "6103662_m.fxe"},
	{0x914BE1B0A19D2E3CULL, "6121059_m.fxe"},
	{0x914C44E079C9B53AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgtb_n.dds"},
	{0x914D00F6941DCDA4ULL, "6119005_m.fxe"},
	{0x914DAB9416589D15ULL, "465701057.ogg"},
	{0x914E14301D0AFC68ULL, "hla_ht_0_hightown_a2_65537_0z.dds"},
	{0x914E264DD943803BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_left.gad"},
	{0x914E31CE5FBD37E9ULL, "582314363.ogg"},
	{0x914EBA3961431722ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05b.mmh"},
	{0x914EBC3961431AD8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw05b.msh"},
	{0x914F0C435F8BC6A7ULL, "llac_w_bonepitmining_a2_262146_0.dds"},
	{0x914F44A9487A1504ULL, "vxlt_gallowstemplar_f_3o3_0_0.vlm"},
	{0x91503CE7F7CC5C4CULL, "w_encounter3_199_0.msh"},
	{0x915042E7F7CC663EULL, "w_encounter3_199_0.mmh"},
	{0x91504EE7F7CC7AD5ULL, "w_encounter3_199_0.mao"},
	{0x9150CBDE246D9C6AULL, "dae000_bartrand6031159.cl"},
	{0x9150E328E61862D6ULL, "lgt160_ntb_elf_conflict6163350.cl"},
	{0x9152139F13B5C232ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_bdaa_0s.dds"},
	{0x91521B4B04EF97D7ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.cess.gad"},
	{0x9152469A8C8BFEDDULL, "145808009.ogg"},
	{0x9152847049707EF6ULL, "w_woundedcoast_191_0.msh"},
	{0x91528E7049708FF4ULL, "w_woundedcoast_191_0.mmh"},
	{0x915290D99D8EA299ULL, "6082591_m.fxe"},
	{0x9152927049709607ULL, "w_woundedcoast_191_0.mao"},
	{0x9152976F242F9440ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01a.phy"},
	{0x9152CFAEB34E8428ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_short.msh"},
	{0x9152E1AEB34EA2B9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_short.mao"},
	{0x9152EDAEB34EB752ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_short.mmh"},
	{0x91537D5F66856206ULL, "vxlt_0_lowtown_docks_a3_9r6_0_0.vlm"},
	{0x915443756C608AAFULL, "llac_w_sundermount_327675_0.dds"},
	{0x915458F4968F3B6CULL, "vxlt_darktown_4tv_0_0.vlm"},
	{0x9154A67066EF534FULL, "vxht_0_hightown_b97_0_0.vlm"},
	{0x9154B405E00971F4ULL, "m_pro000ar_mansion.bnk"},
	{0x91553E9D47C75A0AULL, "6122730_f.fxe"},
	{0x9155434DD949D020ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_left.evt"},
	{0x91559B298284D65CULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf04.phy"},
	{0x9155A718CCE77435ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04b.mmh"},
	{0x9155A918CCE777ABULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04b.msh"},
	{0x9155F2494700CC2DULL, "vxlt_darktown_57o_0_0.vlm"},
	{0x9156C4FE57E5CECCULL, "lowtown_fem_urchin_mob6165929.cl"},
	{0x915787881212F8DDULL, "lt_house_3567.anb"},
	{0x915818240BA71A21ULL, "zz_vault_debug6178941.cl"},
	{0x9158298D76237A78ULL, "263985281.ogg"},
	{0x915922A117928102ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_7x.dds"},
	{0x9159304B04F5DA64ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.cess.evt"},
	{0x91598AD6402F71ABULL, "gen00fl_fenris6127001.cl"},
	{0x9159FFAF648503BBULL, "mcr310_argument6075102.cl"},
	{0x915AF639BCEE89EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\dr1a_basicdoor.mao"},
	{0x915B0239BCEE9D8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\dr1a_basicdoor.mmh"},
	{0x915B0839BCEEA87EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\dr1a_basicdoor.msh"},
	{0x915B529A6D0A0980ULL, "vxlt_shop_ca_0_0.vlm"},
	{0x915C1223D0462605ULL, "plt_lgt111pt_gxa_coa.nss"},
	{0x915C412C5C2B3115ULL, "dre000_cut_bartrandbetrayal6029364.cl"},
	{0x915CBF801CA63A46ULL, "buffertacticpreset.gda"},
	{0x915CD272C538A06FULL, "dae000_orana6149089.cl"},
	{0x915E01FB461A9F25ULL, "gen00fl_varric6076208.cl"},
	{0x915E7F3A98F3D2D8ULL, "927099275.ogg"},
	{0x915ED7B5E12D892BULL, "vxlt_hangedman_n_270_0_0.vlm"},
	{0x915F64A313A05253ULL, "llac_lt_0_lowtown_a2_n_c_1_6.dds"},
	{0x915FAE155E4AF498ULL, "vxlt_0_lowtown_docks_a3_aqi_0_0.vlm"},
	{0x9160AE4B2EC3F9C2ULL, "vxlt_hangedman_287_0_0.vlm"},
	{0x9161122677A3F1F0ULL, "6142316_f.fxe"},
	{0x916123D3A52979C0ULL, "vxlt_gallowstemplar_n_3t2_0_0.vlm"},
	{0x916188046E0C654DULL, "vxw_bonepitmining_a2_1pt_0_0.vlm"},
	{0x91621688D9CBCCF4ULL, "vxlt_0_lowtown_docks_b5y_0_0.vlm"},
	{0x9162A6EA43EA3BDBULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masc_0d.dds"},
	{0x9162CAD1888553ECULL, "vxlt_darktown_4sw_0_0.vlm"},
	{0x916339C2F8048A65ULL, "vxlt_gallowstemplar_n_3on_0_0.vlm"},
	{0x916419A8C56B9577ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medj.mmh"},
	{0x91641DA8C56B9B84ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medj.mao"},
	{0x91642FA8C56BBAE9ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medj.msh"},
	{0x9164455EA3B87EEFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grand.phy"},
	{0x9164642638F6E4ADULL, "vxlt_darktown_58n_0_0.vlm"},
	{0x9164663B60D9A27DULL, "822012396.ogg"},
	{0x9164858815087FB7ULL, "6170911_m.fxe"},
	{0x9165213C1A97AD8DULL, "vxht_keep_n_5y1_0_0.vlm"},
	{0x9165A5E60094511CULL, "vxht_0_hightown_a3_cad_0_0.vlm"},
	{0x9165C6D571F67EAAULL, "245975895.ogg"},
	{0x9165C8789AFECA78ULL, "538738684.ogg"},
	{0x9165CF9419FBD3D5ULL, "vxw_bonepit_8e_0_0.vlm"},
	{0x9165D73AC15967F9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\parypicker\\idle.gad"},
	{0x9166051B97C552A3ULL, "6115100_f.fxe"},
	{0x916618781F850F47ULL, "620190298.ogg"},
	{0x916624BBFE52403EULL, "6051630_f.fxe"},
	{0x9166447C90994C98ULL, "689195868.ogg"},
	{0x9166AEB0AA08D617ULL, "gen00fl_anders6110063.cl"},
	{0x91670D579047313BULL, "hlac_w_woundedcoast_-196600_0y.dds"},
	{0x91672AF281B8782CULL, "48681292.ogg"},
	{0x9167484B0502405DULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.cess.ani"},
	{0x916757AA871D89C5ULL, "62694512.ogg"},
	{0x916761612A2F4F07ULL, "6017670_m.fxe"},
	{0x916775FE51B5126AULL, "vxlt_warehouse_e6_0_0.vlm"},
	{0x91679EC9B67D804BULL, "vxlt_foundry_n_zy_0_0.vlm"},
	{0x916805B3F325CDF5ULL, "vxht_keep_5xy_0_0.vlm"},
	{0x916805CC53345ADCULL, "vxht_mansion_n_4hv_0_0.vlm"},
	{0x91688F7FAD72A947ULL, "290470272.ogg"},
	{0x9168A8F9D2A9C8B5ULL, "vxdr_0_deeproads_ec7_0_0.vlm"},
	{0x9168B69FA1E09F7AULL, "dae000_templar_recruit_16164424.cl"},
	{0x9168BD78600FC495ULL, "772893306.ogg"},
	{0x9168CB59775DFED6ULL, "vxht_0_hightown_a3_cb7_0_0.vlm"},
	{0x91690285E8FBE899ULL, "lt_0_lowtown_docks_a3_222_0.msh"},
	{0x91690C61BE700ECEULL, "vxdr_0_deeproads_cg4_0_2.vlm"},
	{0x91690C85E8FBF987ULL, "lt_0_lowtown_docks_a3_222_0.mmh"},
	{0x91691085E8FC00D4ULL, "lt_0_lowtown_docks_a3_222_0.mao"},
	{0x9169C4775F002200ULL, "vxlt_shop_fn_0_0.vlm"},
	{0x916A0DD221A9A5B3ULL, "6015764_m.fxe"},
	{0x916A258C0DB5C2B0ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgta_0d.dds"},
	{0x916A961A3C12DF56ULL, "w_bonepit_196_0.msh"},
	{0x916AA01A3C12F054ULL, "w_bonepit_196_0.mmh"},
	{0x916AA41A3C12F767ULL, "w_bonepit_196_0.mao"},
	{0x916ACBC6784AA930ULL, "6089979.ogg"},
	{0x916B6E580BC18E24ULL, "art\\characters\\heraldry\\hld_qunari_tide.dds"},
	{0x916BA181F4DF3688ULL, "6136921_m.fxe"},
	{0x916BE4E6E7568875ULL, "gen00fl_fenris6127137.cl"},
	{0x916BE83AC15DF047ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\parypicker\\idle.ani"},
	{0x916D64B57C2CE2C0ULL, "360196311.ogg"},
	{0x916DBAEAA072E6DCULL, "gen00fl_aveline6115457.cl"},
	{0x916E473994978F55ULL, "21352015.ogg"},
	{0x916EB7E48289F29AULL, "w_woundedcoast_45_0.mmh"},
	{0x916EBBE48289F9D1ULL, "w_woundedcoast_45_0.mao"},
	{0x916EC23BDEE71188ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basic.mmh"},
	{0x916EC9E4828A11E0ULL, "w_woundedcoast_45_0.msh"},
	{0x916ECE3BDEE7262BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basic.mao"},
	{0x916EE03BDEE744B2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_basic.msh"},
	{0x916F0BEAFC1FE4B2ULL, "mag320_orsino6071225.cl"},
	{0x916F202820BA1242ULL, "vxlt_hangedman_250_0_0.vlm"},
	{0x916F20ADF40993CAULL, "hlac_w_blightlands_327682_0y.dds"},
	{0x916F2E37989ADE33ULL, "art\\vfx\\combat\\vfx_g_spider_mode_c.impa.ani"},
	{0x916F5AEDE8B06690ULL, "523114666.ogg"},
	{0x916F95B0971F9240ULL, "vxlt_gallowstemplar_n_3s3_0_0.vlm"},
	{0x916FAEE07D659D71ULL, "6119408_m.fxe"},
	{0x916FC21281EDDD22ULL, "vxlt_undercitydungeon_xs_0_0.vlm"},
	{0x91700A6638984D54ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtc.phy"},
	{0x917050ADFA3190EBULL, "vxlt_gallowscourtyard_7bf_0_0.vlm"},
	{0x917101202E31C3D4ULL, "vxca_caverns1_j5t_0_0.vlm"},
	{0x91710B6A01C9BB03ULL, "vxlt_gallowscourtyard_n_7fq_0_0.vlm"},
	{0x91710F4E451B8461ULL, "character_character_p44.psh.fxo.dx11.assembly"},
	{0x9171B89DABB8DD6CULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv2.ani"},
	{0x9171E71C41A61A8AULL, "659613480.ogg"},
	{0x9171EE6918E72E2CULL, "6206079_m.fxe"},
	{0x91721A4B0DCE642DULL, "6066197_m.fxe"},
	{0x917222BB49CF1097ULL, "6139422_m.fxe"},
	{0x91728BA4DB6FEDF2ULL, "dae300_alistair_king6133833.cl"},
	{0x91739BACF0A15FB0ULL, "vxca_caverns2_806_0_0.vlm"},
	{0x91743F8CFE8A8346ULL, "140759057.ogg"},
	{0x9174755A5745CD43ULL, "resmetricsconfig.xml"},
	{0x9174A8F417EA53C1ULL, "llac_lt_0_lowtown_65537_4.dds"},
	{0x917523F73CF5216EULL, "w_bonepitdestroyed_a3_29_0.rml"},
	{0x917576790AB1CFC3ULL, "495001925.ogg"},
	{0x91757AF0D3266CDDULL, "vxlt_warehouse_91_0_0.vlm"},
	{0x91760556CA8A5020ULL, "mcr370_meredith6085880.cl"},
	{0x917610A6A87398CBULL, "vxlt_foundry_n_ux_0_0.vlm"},
	{0x91761F1B387DAE82ULL, "vxlt_0_lowtown_a3_n_c_fha_0_0.vlm"},
	{0x91763BAC0088802FULL, "gen00fl_aveline6139739.cl"},
	{0x91767E320B22A8B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_exit.ani"},
	{0x9176D0B165E6D8EDULL, "219147964.ogg"},
	{0x917721010ED6378AULL, "vxht_mansion2b_1v4_0_0.vlm"},
	{0x91772F88BAB47D7CULL, "dr_primevaltomb_659.rml"},
	{0x9177524DE182223FULL, "6067714_m.fxe"},
	{0x91775EE1B8C90595ULL, "159052922.ogg"},
	{0x9177B46F244EE1D5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01a.msh"},
	{0x9177B66F244EE54BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk01a.mmh"},
	{0x9177C59DABBD5EAEULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv2.gad"},
	{0x9178183AAE1A022CULL, "one100dg_elren6058877.cl"},
	{0x9178E290479A290EULL, "vxlt_0_lowtown_docks_a3_n_cz2_0_1.vlm"},
	{0x917981A03396269DULL, "hlac_ht_0_hightown_a2_65535_0z.dds"},
	{0x917A26A43B5D9452ULL, "655696454.ogg"},
	{0x917A5565529286CBULL, "1052613607.ogg"},
	{0x917A62E439D006BFULL, "vxlt_warehouse_n_xq_0_0.vlm"},
	{0x917AE76638A21951ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtc.msh"},
	{0x917AF1F6AE4178BEULL, "242310064.ogg"},
	{0x917AF56638A2316CULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtc.mao"},
	{0x917B016638A2450FULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtc.mmh"},
	{0x917BF4C535B32725ULL, "hlac_lt_darktown_-1_90x.dds"},
	{0x917BF55A2B54C8F1ULL, "vxlt_hangedman_n_5d_0_0.vlm"},
	{0x917C0E425C0B7756ULL, "vxlt_gallowstemplar_n_5rd_0_0.vlm"},
	{0x917C71FF35F36ADEULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackback.ani"},
	{0x917D137FF87A85C9ULL, "hlac_ca_cavern1_haunted_24421_2x.dds"},
	{0x917D263798A70E6AULL, "art\\vfx\\combat\\vfx_g_spider_mode_c.impa.evt"},
	{0x917D552B53872DC8ULL, "vxca_caverns1_mines_k6m_0_0.vlm"},
	{0x917D61953C7B8548ULL, "hlac_ca_caverns1_mines_24424_21y.dds"},
	{0x917DB1433CDEC5D3ULL, "seb321_leliana6180900.cl"},
	{0x917DC2CCAEA96729ULL, "gen_im_trs_qun_biscuit_01.uti"},
	{0x917E73D541E21F9BULL, "art\\levels\\races\\elf\\props\\_textures\\prp_evilchya_d.dds"},
	{0x917EB832C19C927FULL, "6117985_m.fxe"},
	{0x917EC09DABC3748DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\meleeattackv2.evt"},
	{0x917EFD307D347234ULL, "19409160.ogg"},
	{0x917F35E4AAD53B72ULL, "6126931_m.fxe"},
	{0x917F4216025630F7ULL, "6153655_m.fxe"},
	{0x917FC4F55CF9B41DULL, "6148796_m.fxe"},
	{0x917FFE2D2BB255ACULL, "hlac_lt_0_lowtown_a2_131070_5z.dds"},
	{0x91800D054DEF65BEULL, "vxw_sundermount_31k_0_0.vlm"},
	{0x9180F45C4AF6B309ULL, "vxw_encounter3_g6_0_1.vlm"},
	{0x9181F22B8A7367A3ULL, "gen00fl_isabela6075221.cl"},
	{0x9181F952335D2C77ULL, "6166156_m.fxe"},
	{0x9184053798ACF4DDULL, "art\\vfx\\combat\\vfx_g_spider_mode_c.impa.gad"},
	{0x9184504BA89E7547ULL, "follower_banter6164706.cl"},
	{0x918476320B2ED8EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_exit.evt"},
	{0x918492017E2F2BAFULL, "hlac_ht_0_hightown_a2_-131073_0x.dds"},
	{0x9184AFA72AACBB3DULL, "mag330_walter6045016.cl"},
	{0x9184E785688ED29FULL, "lgt260cr_eau_tmp_flowergirl.utc"},
	{0x9184F8B2741A243DULL, "6075351_m.fxe"},
	{0x9185070D5A683612ULL, "6095905_m.fxe"},
	{0x91852CA3AF77FB88ULL, "6168773_m.fxe"},
	{0x918576CBFAA3989DULL, "spark_01.mao"},
	{0x9186509C9523BA02ULL, "vxca_caverns1_i77_0_0.vlm"},
	{0x91875AABD737E21CULL, "vxlt_gallowstemplar_n_5j6_0_0.vlm"},
	{0x9187B99C407EFC02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\buildup_lp.gad"},
	{0x91880EF55805C4DAULL, "475109305.ogg"},
	{0x918886CED74DE87EULL, "6087216_m.fxe"},
	{0x918968993E99E83AULL, "hla_ht_0_hightown_n_131073_1y.dds"},
	{0x918989FF35FE1DAFULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackback.evt"},
	{0x918A3141EEF8CA52ULL, "756173953.ogg"},
	{0x918A9DBD83166447ULL, "lt_warehouse_n_2_0.rml"},
	{0x918AF331377DC133ULL, "693781537.ogg"},
	{0x918B55320B34BF5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_chair_exit.gad"},
	{0x918B64E10BB456EBULL, "6125243_m.fxe"},
	{0x918BB0BCA33C58F3ULL, "319997203.ogg"},
	{0x918BB5C8B2E599CCULL, "hlac_ht_playershouse_n_855_29x.dds"},
	{0x918C318E21710ED5ULL, "vxlt_0_lowtown_a3_n_c_euz_0_0.vlm"},
	{0x918D1B132A95568AULL, "mag100_karras6034930.cl"},
	{0x918D3467DE1DC1EBULL, "vxlt_gallowscourtyard_7hx_0_0.vlm"},
	{0x918D719C8925DF43ULL, "ht_0_hightown_a2_25_0_ll.mmh"},
	{0x918D82758CF9763FULL, "vxlt_darktown_4sx_0_0.vlm"},
	{0x918DB3F2F44C4977ULL, "6156555_m.fxe"},
	{0x918E67836539F0D4ULL, "rune_armor_nature.dds"},
	{0x918EBC1F1742757DULL, "zz_start_dre6080518.cl"},
	{0x918EBFA8E93CB518ULL, "hlac_w_encounter2_-131069_0y.dds"},
	{0x918ECAFDE3861D77ULL, "6128922_f.fxe"},
	{0x918EDC9C408556F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\buildup_lp.evt"},
	{0x918EDE480D3A71CCULL, "vxlt_hangedman_1tv_0_0.vlm"},
	{0x918F75ED976ABFABULL, "isa211_isabela6072555.cl"},
	{0x918F79F50A231BA5ULL, "hlac_lt_0_lowtown_a2_2_1y.dds"},
	{0x918F7BCB94968DF0ULL, "vxca_caverns1_cjx_0_0.vlm"},
	{0x918F8CC80ED5FBF8ULL, "607533905.ogg"},
	{0x91903D4A3290E3B0ULL, "6172100_f.fxe"},
	{0x91908AFF36043E58ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackback.gad"},
	{0x9190BF74541F2FE7ULL, "vxht_0_hightown_b44_0_0.vlm"},
	{0x9190F8C017A045A0ULL, "and231cr_smuggler.utc"},
	{0x91918888124472D5ULL, "lt_house_3567.rml"},
	{0x91918DE8230A497BULL, "12155368.ogg"},
	{0x91919CBD831C815FULL, "lt_warehouse_n_2_0.phy"},
	{0x9191ADB9D1DA69ACULL, "vxht_mansion2b_2j9_0_0.vlm"},
	{0x9191B97CBAFDF9B1ULL, "487016557.ogg"},
	{0x9191E81CE0AAEE01ULL, "mag100_cullen6082304.cl"},
	{0x919259D74C2853F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_af_transition.evt"},
	{0x9192CB9527975BEAULL, "vxlt_warehouse_t1_0_0.vlm"},
	{0x9192EE17D2329F5DULL, "qun110_petrice6059152.cl"},
	{0x91935C222775A153ULL, "isa311_isabela6092288.cl"},
	{0x91943AA03C8773E7ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammerb_1.mao"},
	{0x9194CD8FFF713499ULL, "gen00fl_fenris6106635.cl"},
	{0x9194CE05E03F81A6ULL, "m_pro000ar_mansion.wbd"},
	{0x9194DBAC974CFB30ULL, "6067803_m.fxe"},
	{0x919563AE88541E1BULL, "6016138_m.fxe"},
	{0x91956E3A5A7D06F7ULL, "285872438.ogg"},
	{0x9195EADCAC05E867ULL, "129752692.ogg"},
	{0x9195F7D5C99DFFA9ULL, "991911115.ogg"},
	{0x91963E72A89A1638ULL, "fen310_varania6118631.cl"},
	{0x919640CE6A2C34A6ULL, "6093576_m.fxe"},
	{0x919695D82032DEB3ULL, "gen00fl_aveline6139663.cl"},
	{0x9196C7879E805380ULL, "1064605103.ogg"},
	{0x919716D3E578108FULL, "539314657.ogg"},
	{0x919718E6F40F854AULL, "6082270_f.fxe"},
	{0x91974498BF35CD4DULL, "vxht_playershouse_n_1u4_0_0.vlm"},
	{0x91978D9C6300D5FBULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_arh_0d.dds"},
	{0x919860F654322A8EULL, "vxca_caverns1_mines_ehg_0_0.vlm"},
	{0x919934958BED10B2ULL, "fen110_fenris6118135.cl"},
	{0x919952D74C2E6634ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_af_transition.gad"},
	{0x919AA36B13672755ULL, "vxlt_0_lowtown_a3_n_c_exy_0_0.vlm"},
	{0x919AB7E482AF2DB1ULL, "w_woundedcoast_45_0.rml"},
	{0x919ABBE1487F5D19ULL, "6060739_m.fxe"},
	{0x919AFD7D989E84D4ULL, "dae000ip_crowd_ff8_talk10.utp"},
	{0x919BFA0DD9A53024ULL, "745447645.ogg"},
	{0x919C4252E2A6392EULL, "465714799.ogg"},
	{0x919C54C49F1EC5FFULL, "art\\vfx\\environment\\vfx_smoke_med_p.mmh"},
	{0x919C981B6CB7AF18ULL, "hlac_lt_0_lowtown_docks_n_-131074_4x.dds"},
	{0x919CE49C4091A1E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\buildup_lp.ani"},
	{0x919D38FB7E23557EULL, "6077008_m.fxe"},
	{0x919D4C9B1654EA8CULL, "vxlt_gallowscourtyard_n_76z_0_0.vlm"},
	{0x919D70070F855614ULL, "6077143_m.fxe"},
	{0x919DCA2D3B5B7801ULL, "781581339.ogg"},
	{0x919E038C508F545FULL, "llac_lt_0_lowtown_a2_-1_6.dds"},
	{0x919E580EA39C42A9ULL, "6020938_f.fxe"},
	{0x919E6EBB155AB8ABULL, "llac_lt_0_lowtown_docks_a3_n_-131075_5.dds"},
	{0x919E8B4D009D5034ULL, "vxlt_0_lowtown_f58_0_0.vlm"},
	{0x919EDCD1E7AFAC76ULL, "354223717.ogg"},
	{0x919EE0CE4E1BEF37ULL, "llac_lt_blackemporium_1055_1.dds"},
	{0x919F0843269525FCULL, "6034717_m.fxe"},
	{0x919F5E400E494225ULL, "468169250.ogg"},
	{0x919F61D74C32EB12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_af_transition.ani"},
	{0x919F96CE62695264ULL, "vxlt_0_lowtown_a2_n_c_byq_0_0.vlm"},
	{0x91A07461BCD9311BULL, "vxca_cavern1_haunted_i77_0_0.vlm"},
	{0x91A0BC88F0756FBBULL, "hlac_w_encounter2_camp_196608_0y.dds"},
	{0x91A10DA340896747ULL, "pro000ip_corpse_weapons2.utp"},
	{0x91A13D72198D746AULL, "vxlt_warehouse_y2_0_0.vlm"},
	{0x91A16A9D1B18A1ADULL, "6147689_m.fxe"},
	{0x91A17A34E93448AFULL, "dre000_cut_bartrandbetrayal6029379.cl"},
	{0x91A17B6F4D0246E1ULL, "zz_approval_debug6003914.cl"},
	{0x91A1985A2DE49421ULL, "6136450_m.fxe"},
	{0x91A1FE206839E16AULL, "587228772.ogg"},
	{0x91A21752973A6CE4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_couldron_n.dds"},
	{0x91A28925191D332BULL, "hlac_w_bonepitmining_a2_0_10x.dds"},
	{0x91A29DBD832AF664ULL, "lt_warehouse_n_2_0.mmh"},
	{0x91A2A1BD832AFCB7ULL, "lt_warehouse_n_2_0.mao"},
	{0x91A2B3BD832B1B06ULL, "lt_warehouse_n_2_0.msh"},
	{0x91A3951C2FE8458AULL, "808222307.ogg"},
	{0x91A3B0E82DF4EEB0ULL, "6161482_m.fxe"},
	{0x91A3BD426268C6A3ULL, "120946197.ogg"},
	{0x91A3F00F2269A891ULL, "934858871.ogg"},
	{0x91A40088F12D8464ULL, "dae170ip_mage_store.utp"},
	{0x91A40C4A0A30F06FULL, "234567084.ogg"},
	{0x91A43EBFAD1E1344ULL, "vxht_0_hightown_a2_5i3_0_0.vlm"},
	{0x91A4BE472A6F95CBULL, "vxlt_gallowstemplar_f_5bu_0_0.vlm"},
	{0x91A4E4C5F91054B4ULL, "6076290_m.fxe"},
	{0x91A4F31824AD2F5FULL, "503376434.ogg"},
	{0x91A52B634D6F9409ULL, "6166306_f.fxe"},
	{0x91A62F7B55CE3EA3ULL, "657751297.ogg"},
	{0x91A64542383DAC89ULL, "435608285.ogg"},
	{0x91A69B372CF002D5ULL, "vxlt_0_lowtown_gof_0_0.vlm"},
	{0x91A7517C7FD210C4ULL, "hlac_lt_0_lowtown_a2_n_c_131072_1y.dds"},
	{0x91A75FC0AA653BEBULL, "vxlt_0_lowtown_docks_bdq_0_0.vlm"},
	{0x91A763D931EDC0D6ULL, "vxlt_gallowstemplar_n_5cg_0_0.vlm"},
	{0x91A7B83B7344817CULL, "272557200.ogg"},
	{0x91A7E94C3FC1FAC3ULL, "vxca_cave_vx_0_0.vlm"},
	{0x91A9931934FAA26AULL, "hlac_w_woundedcoast_0_5z.dds"},
	{0x91A9C4F7D9D99B31ULL, "vxlt_0_lowtown_n_e20_0_0.vlm"},
	{0x91AA54794BE20D35ULL, "vxht_chantry_36m_0_0.vlm"},
	{0x91AA79FB6FF1FD53ULL, "6060126_m.fxe"},
	{0x91AA8B70261E8A20ULL, "6055162_f.fxe"},
	{0x91AA92220349B2D7ULL, "6206633_f.fxe"},
	{0x91AACDCA74FA2DC1ULL, "6087233_m.fxe"},
	{0x91AAF7DE4DAF0DBCULL, "33922308.ogg"},
	{0x91AB204C4F934E9DULL, "gen00fl_isabela6099504.cl"},
	{0x91AB99E8E5F759F6ULL, "sf_cxform.psh.fxo"},
	{0x91ABAAD870196DFEULL, "6078108_m.fxe"},
	{0x91ABB28EA694B337ULL, "14769187.ogg"},
	{0x91ABED8FD35CC1C1ULL, "40689998.ogg"},
	{0x91AC3758F2B23536ULL, "fx_energy_blast.mao"},
	{0x91AC601E22F37E55ULL, "6147700_m.fxe"},
	{0x91ACD1946B918059ULL, "6037308_m.fxe"},
	{0x91AD3D4E939AFB72ULL, "vxht_chantry_n_20v_0_0.vlm"},
	{0x91AD493E3A54DD05ULL, "vxlt_undercityencounter_1bs_0_1.vlm"},
	{0x91AD7EB2099EBE1CULL, "6052757_m.fxe"},
	{0x91AD9FE1A0E2CEACULL, "467035320.ogg"},
	{0x91ADBACF10E5C9D2ULL, "vxht_0_hightown_au9_0_0.vlm"},
	{0x91AE0542E5341805ULL, "6018740_f.fxe"},
	{0x91AE33750EDE3684ULL, "lt_0_lowtown_docks_a3_126_0.mmh"},
	{0x91AE35B3BD69BA6FULL, "ave200_party_comment6116365.cl"},
	{0x91AE37750EDE3DD7ULL, "lt_0_lowtown_docks_a3_126_0.mao"},
	{0x91AE49750EDE5C26ULL, "lt_0_lowtown_docks_a3_126_0.msh"},
	{0x91AEECAEB39D144DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_short.phy"},
	{0x91AFDC7A0D0EBA2DULL, "6140789_m.fxe"},
	{0x91B0EECA20836CB1ULL, "zz_and_debug6134081.cl"},
	{0x91B1381AD73BB67BULL, "vxw_woundedcoast_1j1_0_0.vlm"},
	{0x91B1394B9128EDCCULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb02a.phy"},
	{0x91B165BC56DBA30DULL, "6052147_2.cut"},
	{0x91B1A2B3B7FE17C3ULL, "vxlt_0_lowtown_docks_n_cq4_0_0.vlm"},
	{0x91B1BFE2285551E9ULL, "6180932_m.fxe"},
	{0x91B1CBA8C9FDA46EULL, "6119308_m.fxe"},
	{0x91B1E88E41FB9B40ULL, "lgt210ip_ingredient1.utp"},
	{0x91B217C93627DD1EULL, "501825252.ogg"},
	{0x91B235D0770BF667ULL, "6044663_m.fxe"},
	{0x91B23696CBC38A67ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\flurry.gad"},
	{0x91B242904997A212ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaritrooper.anb"},
	{0x91B33CE12D37264AULL, "6042084_m.fxe"},
	{0x91B36E51F5379A72ULL, "71526418.ogg"},
	{0x91B37560FBC1CD3EULL, "hlac_lt_0_lowtown_docks_131072_0z.dds"},
	{0x91B3845B1F707875ULL, "6151040_m.fxe"},
	{0x91B3D35FC9695771ULL, "820414859.ogg"},
	{0x91B44A73F932397AULL, "6172332_m.fxe"},
	{0x91B49F4499E716EAULL, "gen00fl_aveline6139871.cl"},
	{0x91B4CF01AC16AF2AULL, "fx_letter_z.mao"},
	{0x91B4FBC8999D487DULL, "6093654_m.fxe"},
	{0x91B5305DAC9998D2ULL, "6044615-onlyremembertemplar.cut"},
	{0x91B53CCC73E798CCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wic_0n.dds"},
	{0x91B544B0381E5B8EULL, "vxca_caverns1_mines_eby_0_0.vlm"},
	{0x91B652C6B746553EULL, "hlac_lt_0_lowtown_docks_n_-3_5z.dds"},
	{0x91B66D09D9A38ECCULL, "hlac_ht_0_hightown_a2_131071_2z.dds"},
	{0x91B6C08C376EB39AULL, "661050961.ogg"},
	{0x91B6DC18B2410D6FULL, "vxca_cave_266_0_0.vlm"},
	{0x91B78C895D972D5FULL, "llac_lt_0_lowtown_docks_131071_6.dds"},
	{0x91B7FB6908311028ULL, "mag371_elsa6064699.cl"},
	{0x91B81A77BE079916ULL, "pro000_meeran6048826.cl"},
	{0x91B82B96CBC7E374ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\flurry.evt"},
	{0x91B8C1154AC2F392ULL, "344473274.ogg"},
	{0x91B8C4471708B5A8ULL, "w_bonepitmining_a2_206_0.msh"},
	{0x91B8D6471708D439ULL, "w_bonepitmining_a2_206_0.mao"},
	{0x91B8E2471708E8D2ULL, "w_bonepitmining_a2_206_0.mmh"},
	{0x91B8EEF5E7834990ULL, "925142029.ogg"},
	{0x91B91C131870F0CFULL, "llac_ht_0_hightown_n_-65536_1.dds"},
	{0x91B9A5341941F9DDULL, "amulet_durasblueflame.dds"},
	{0x91B9E2CF126E98BBULL, "zz_followers_banter6158376.cl"},
	{0x91BA310297F84C92ULL, "follower_banter6141847.cl"},
	{0x91BA59211B5F4590ULL, "mer120_hubert6059420.cl"},
	{0x91BAA39DCE927D14ULL, "251637282.ogg"},
	{0x91BAC8ECB5F3A039ULL, "vxlt_hangedman_n_sn_0_0.vlm"},
	{0x91BB0C5CED3A1172ULL, "1000405354.ogg"},
	{0x91BB2A2928E707DEULL, "gen00fl_varric6033707.cl"},
	{0x91BB576034EB3796ULL, "19165237.ogg"},
	{0x91BB733E6E9ABCFFULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_reddust_d.dds"},
	{0x91BB9A8A19B430E3ULL, "llac_ht_mansion_n_3374_0.dds"},
	{0x91BBFFBDE905AEF8ULL, "412364640.ogg"},
	{0x91BC2B5899F170BBULL, "6204162_m.fxe"},
	{0x91BC689ACF2E083AULL, "prop1_d.psh.fxo"},
	{0x91BC8E3E1D1D2BBBULL, "231298910.ogg"},
	{0x91BD6F3C32212960ULL, "322695443.ogg"},
	{0x91BD7B219FCDF843ULL, "6026234_m.fxe"},
	{0x91BDF20E12C1F3B3ULL, "follower_banter6151474.cl"},
	{0x91BE434D460017FFULL, "w_bonepit_185_0.phy"},
	{0x91BEB9FC715FFFFAULL, "6116647_f.fxe"},
	{0x91BF27EFF4AA781AULL, "5411647.ogg"},
	{0x91BF68EC69A49483ULL, "6101429_m.fxe"},
	{0x91BF71A07FFC6204ULL, "vxca_caverns2_6ht_0_0.vlm"},
	{0x91BF7FF71BCF9AC1ULL, "6077061_m.fxe"},
	{0x91BFCAD4FAD6ECEBULL, "llac_lt_0_lowtown_docks_-262144_3.dds"},
	{0x91C02DF8BE81F718ULL, "580724015.ogg"},
	{0x91C0480C212DF97EULL, "vxlt_0_lowtown_docks_a3_cvi_0_1.vlm"},
	{0x91C092E1D251274CULL, "vxlt_gallowsprison_n_5wi_0_0.vlm"},
	{0x91C0A1510CE3A0CFULL, "195784759.ogg"},
	{0x91C121E1A0727F37ULL, "mag291_demon_marethari6028846.cl"},
	{0x91C1425BEC663822ULL, "hlac_w_woundedcoast_196612_0z.dds"},
	{0x91C1E17ADD7EF3E9ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.dura.ani"},
	{0x91C1EA846872C01AULL, "lt_0_lowtown_docks_12_0.anb"},
	{0x91C216FA542DB3D7ULL, "dre000_dougal6130959.cl"},
	{0x91C22DDA544B4985ULL, "112486735.ogg"},
	{0x91C2576072AE209BULL, "dae100_bran6118310.cl"},
	{0x91C29CD9E75FD5A9ULL, "6064854_m.fxe"},
	{0x91C2B5AD3F34665AULL, "399535699.ogg"},
	{0x91C2BC50EB2851FAULL, "6175232_m.fxe"},
	{0x91C3543C8575F7C0ULL, "dae120cr_amb_dwarf_6.utc"},
	{0x91C3B68D2A14740EULL, "vxca_caverns1_mines_egz_0_0.vlm"},
	{0x91C4592B4BFD3238ULL, "vxdr_0_deeproads_ebm_0_0.vlm"},
	{0x91C4B9393EDCF6A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_malepc_loop.ani"},
	{0x91C4CB9479DC6197ULL, "6076671_m.fxe"},
	{0x91C4DB2DD7240B37ULL, "llac_lt_0_lowtown_docks_-1_3.dds"},
	{0x91C4E8414547F17FULL, "1041528778.ogg"},
	{0x91C599A3C9F060BDULL, "vxlt_0_lowtown_a2_cfi_0_0.vlm"},
	{0x91C59DC7135AB100ULL, "hlac_w_encounter1_65536_0z.dds"},
	{0x91C5D0E683D3A3DFULL, "180899220.ogg"},
	{0x91C5D16A5B07247AULL, "237224895.ogg"},
	{0x91C5F6D1CD808269ULL, "6143322_m.fxe"},
	{0x91C61519D44366C0ULL, "392878391.ogg"},
	{0x91C62396CBD412ADULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\flurry.ani"},
	{0x91C632AFE053DD4DULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_smoke_soft_fog.mao"},
	{0x91C658D0B5651CA9ULL, "6139543_m.fxe"},
	{0x91C6EC0ACA2C1213ULL, "cs_mcr371cs_orsino_defiance.bnk"},
	{0x91C72767310A10C6ULL, "6140647_m.fxe"},
	{0x91C72940A3E0588CULL, "hlac_dr_primevalentrance_1726_0z.dds"},
	{0x91C75D3D9E108AEEULL, "457153966.ogg"},
	{0x91C7B7FE19FD0898ULL, "vxlt_0_lowtown_docks_a3_bs8_0_0.vlm"},
	{0x91C81D5D65EB2902ULL, "art\\vfx\\combat\\vfx_c_mer_burst_c.mmh"},
	{0x91C8FC7ADD85406BULL, "art\\vfx\\creatures\\fxc_rage_crust_c.dura.gad"},
	{0x91C9334FF04C9D2FULL, "323535965.ogg"},
	{0x91C97447720EACC4ULL, "ms_rock.mao"},
	{0x91C9E7ABAB37DAABULL, "gen_im_wep_war_two_default.uti"},
	{0x91CA0118F84AA3CFULL, "373431420.ogg"},
	{0x91CA4858F2CB4FB8ULL, "fx_energy_blast.dds"},
	{0x91CA5E513BF7312AULL, "character_character_p214.psh.fxo.dx11.assembly"},
	{0x91CBA0BD834E005BULL, "lt_warehouse_n_2_0.anb"},
	{0x91CBD4393EE3432BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_malepc_loop.gad"},
	{0x91CCBB8E46391124ULL, "vxht_bartrand_n_6po_0_0.vlm"},
	{0x91CCEA4D0DBA8458ULL, "726705179.ogg"},
	{0x91CCECA09313ECFCULL, "6089686_f.fxe"},
	{0x91CCF0D9A03A9D2FULL, "312906031.ogg"},
	{0x91CE19C15EB029B6ULL, "6118228_m.fxe"},
	{0x91CE88318615D1CEULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.dura.ani"},
	{0x91CEA3815784D805ULL, "weapon2_ds1y.psh.fxo"},
	{0x91CEAF62ADE9D56EULL, "6113508_m.fxe"},
	{0x91CEE4967BC1173BULL, "vxlt_0_lowtown_a2_gkp_0_0.vlm"},
	{0x91CF444D460E8C04ULL, "w_bonepit_185_0.mmh"},
	{0x91CF484D460E9357ULL, "w_bonepit_185_0.mao"},
	{0x91CF4BC812213D97ULL, "500315056.ogg"},
	{0x91CF5A4D460EB1A6ULL, "w_bonepit_185_0.msh"},
	{0x91CF6E23B2B22864ULL, "mcr370_negotiation6080154.cl"},
	{0x91CFAD2EB3A83EB1ULL, "6115365_m.fxe"},
	{0x91CFC20478399412ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_6z.dds"},
	{0x91CFF97ADD8B59D0ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.dura.evt"},
	{0x91D08159A646A650ULL, "one280_emrys.stg"},
	{0x91D08ED81C61CE6CULL, "6149589_m.fxe"},
	{0x91D1924100D3B385ULL, "one110_meeran6089858.cl"},
	{0x91D21F7546B3EDD7ULL, "hlac_lt_0_lowtown_docks_65532_0z.dds"},
	{0x91D2837F1E45696BULL, "vxca_cavern1_haunted_i1a_0_0.vlm"},
	{0x91D2BB05797F4906ULL, "vrc321_varric6087000.cl"},
	{0x91D2DA2EFF2D1D77ULL, "675788963.ogg"},
	{0x91D30F574D4166F0ULL, "859004558.ogg"},
	{0x91D3297679AEBA4EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v3.ani"},
	{0x91D36758C97CC1D1ULL, "6000796_f.fxe"},
	{0x91D3B44981FBFED4ULL, "pro000_varric6179433.cl"},
	{0x91D3FAEAF1A629CAULL, "zz_vrc_debug6064262.cl"},
	{0x91D40EE655CEDB53ULL, "341886511.ogg"},
	{0x91D41EAE776FD575ULL, "vxht_brothel_n_5iv_0_0.vlm"},
	{0x91D537633DF0B08CULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopening.ani"},
	{0x91D58DDD48F66DECULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basicarch.phy"},
	{0x91D5C4506717D308ULL, "vxht_0_hightown_n_c3e_0_1.vlm"},
	{0x91D5C58D60EA02D1ULL, "6121757_m.fxe"},
	{0x91D5D90062275DC1ULL, "qun111_ketojan6041138.cl"},
	{0x91D6035FD4B3E3D2ULL, "hlac_ca_caverns2prc_5802_35x.dds"},
	{0x91D626D15B082C51ULL, "dae300_alistair_king6134910.cl"},
	{0x91D652C6056C1905ULL, "6110520_m.fxe"},
	{0x91D728E48BC444F9ULL, "6114533_m.fxe"},
	{0x91D7D0285AAFD992ULL, "852239333.ogg"},
	{0x91D7D68829C4A556ULL, "319528172.ogg"},
	{0x91D8FD0FE6E87F77ULL, "434555973.ogg"},
	{0x91D90B596C57D814ULL, "vxlt_0_lowtown_a2_n_c_euz_0_0.vlm"},
	{0x91D90E5995620E80ULL, "vxdr_primevaltomb_512_0_0.vlm"},
	{0x91D95663D5A372C5ULL, "gen_im_arm_cht_nobe_t4.uti"},
	{0x91D99C6C74FE9F40ULL, "vxht_0_hightown_a2_aoy_0_0.vlm"},
	{0x91D9DCCAF62D50CFULL, "vxw_bonepit_kn_0_0.vlm"},
	{0x91DA215D68A081A0ULL, "6049087_m.fxe"},
	{0x91DA8625697D37D0ULL, "sound_geometry_mat_types.gda"},
	{0x91DAB6AFDBC5394EULL, "vxlt_0_lowtown_a3_n_c_ffg_0_0.vlm"},
	{0x91DACCC121E02B37ULL, "770648498.ogg"},
	{0x91DBA577E4EE4510ULL, "hlac_lt_0_lowtown_a2_-2_6z.dds"},
	{0x91DBFC34871F9DB9ULL, "355042231.ogg"},
	{0x91DC44633DF6E5CEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_pcopening.gad"},
	{0x91DC80318622011FULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.dura.evt"},
	{0x91DD5150FB48CDC6ULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.cess.evt"},
	{0x91DD56736DB72FBBULL, "vxlt_0_lowtown_a2_gns_0_0.vlm"},
	{0x91DD579DE83E2DD3ULL, "fen110_anso6094971.cl"},
	{0x91DDB08F824E5B68ULL, "vxca_caverns1_jg2_0_0.vlm"},
	{0x91DDC2C5A0A89D36ULL, "vxlt_gallowstemplar_n_3rb_0_0.vlm"},
	{0x91DDCF7CA8301326ULL, "6208602_m.fxe"},
	{0x91DE5748E9086ACCULL, "hlac_w_sundermount_-196606_0z.dds"},
	{0x91DEDB0B0C951C6FULL, "gen00fl_varric6099397.cl"},
	{0x91DF2F3BE85543E4ULL, "694116041.ogg"},
	{0x91DF8C12A55DA0EBULL, "vxw_bonepitdestroyed_a3_1bf_0_0.vlm"},
	{0x91E0426FA3800E42ULL, "zz_followers_banter6158700.cl"},
	{0x91E0890B4DE92764ULL, "zz_followers_banter6157866.cl"},
	{0x91E08BCFE5E5FF8DULL, "6061152_m.fxe"},
	{0x91E0F63E1A1B064FULL, "6080982_m.fxe"},
	{0x91E0FCAB8B931BA4ULL, "6113014_m.fxe"},
	{0x91E1217679BAE99FULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v3.evt"},
	{0x91E1EC706915F91DULL, "hlac_lt_0_lowtown_docks_n_-196609_0z.dds"},
	{0x91E29E275CB0F7DAULL, "hlac_w_encounter3_-393216_0z.dds"},
	{0x91E2DB89DBCC2523ULL, "zz_vrc_debug6200207.cl"},
	{0x91E2DF1C0FE92AFAULL, "120551277.ogg"},
	{0x91E361318627EB08ULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.dura.gad"},
	{0x91E3646C9F766DC0ULL, "vxca_cavern1_haunted_dp8_0_0.vlm"},
	{0x91E36F235C73D9B1ULL, "vxht_keep_n_7ac_0_0.vlm"},
	{0x91E46850FB4F1311ULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.cess.gad"},
	{0x91E4734E24CD3DDAULL, "da2terrain_terrain_p4.psh.fxo.dx11"},
	{0x91E47DCDFD49CE81ULL, "vxlt_shop_cl_0_0.vlm"},
	{0x91E4B9C310AC5383ULL, "vxw_encounter3_ge_0_0.vlm"},
	{0x91E533775E980472ULL, "vxht_mansion_n_3i4_0_0.vlm"},
	{0x91E5966D9F5DC97DULL, "hlac_ht_mansion2_818_64y.dds"},
	{0x91E59AC17DBA5D79ULL, "6113433_m.fxe"},
	{0x91E5A10924DAC1DDULL, "follower_banter6172817.cl"},
	{0x91E5B22222EA4845ULL, "llac_ht_0_hightown_n_2_1.dds"},
	{0x91E5E74FFC16EC6DULL, "759343570.ogg"},
	{0x91E673BC4D317EB3ULL, "dae000_flemeth6000717.cl"},
	{0x91E68A9F1B0A9285ULL, "llac_ca_cavern1_haunted_24424_27.dds"},
	{0x91E6BAF12647B694ULL, "679889328.ogg"},
	{0x91E71552D1440D0FULL, "vxlt_0_lowtown_a3_n_c_fi7_0_0.vlm"},
	{0x91E7B876EBF65BE4ULL, "zz_joshua.ncs"},
	{0x91E7E19723CEDD7DULL, "vxlt_backalley_g_1ec_0_0.vlm"},
	{0x91E8027679C0D388ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v3.gad"},
	{0x91E83FB48E3DDC46ULL, "835428581.ogg"},
	{0x91E864AC28E52194ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_up.ani"},
	{0x91E8764864C7FBF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\blendtree\\trooper_movement.anb"},
	{0x91E9283595944F38ULL, "6071923_m.fxe"},
	{0x91E946E3C7FAB720ULL, "303964351.ogg"},
	{0x91E964C40F287DFAULL, "6129941-mrl-mirror-intro.cut"},
	{0x91E993D77D68CBBCULL, "6113848_f.fxe"},
	{0x91E9DC02FB09FD39ULL, "6114865_m.fxe"},
	{0x91EA5AA7E824A7BCULL, "6099157_f.fxe"},
	{0x91EACA292E72A13AULL, "6137623_m.fxe"},
	{0x91EB3B214EB565EAULL, "lowtown_male_drunk_mob6151175.cl"},
	{0x91EB4FC3F680D13BULL, "675283954.ogg"},
	{0x91EB6950FB5533BFULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.cess.ani"},
	{0x91EB9D6B9F3A680DULL, "6128744_f.fxe"},
	{0x91EBBFA5FA9B11F8ULL, "6204158_m.fxe"},
	{0x91EBC0270065E0A5ULL, "vxht_mansion2_268_0_0.vlm"},
	{0x91EC14C7F606EB11ULL, "hlac_ht_keep_n_4519_11z.dds"},
	{0x91ECBED6E82AD2F0ULL, "isa211_isabela6072774.cl"},
	{0x91ECE70D8A1E0E89ULL, "vxca_caverns1_mines_iy3_0_0.vlm"},
	{0x91ED1D77CD8455F4ULL, "vxw_woundedcoast_hq_0_0.vlm"},
	{0x91ED3FD39F799289ULL, "gen00fl_anders6125775.cl"},
	{0x91EDC1BC09E0A2E2ULL, "isa311_castillon6092119.cl"},
	{0x91EE8D9FBD3996CEULL, "hlac_lt_blackemporium_1055_2x.dds"},
	{0x91EEB555BB027941ULL, "t1_mue_re3.tnt"},
	{0x91EED88E04D7E123ULL, "gen00fl_merrill6127306.cl"},
	{0x91EF30DDDEB7F416ULL, "dae000cr_grand_cleric6147698.cl"},
	{0x91EF61AC28EB3BF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_up.gad"},
	{0x91EFDDBC9282928DULL, "llac_ht_0_hightown_131071_1.dds"},
	{0x91F01C5D660C79A9ULL, "art\\vfx\\combat\\vfx_c_mer_burst_c.anb"},
	{0x91F056DFB46F21C0ULL, "vxca_caverns2_63a_0_0.vlm"},
	{0x91F1104B915E8B97ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb02a.mmh"},
	{0x91F1155F75D91E0EULL, "art\\vfx\\combat\\vfx_q_ari_buf_act_c.mmh"},
	{0x91F1264B915EB109ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb02a.msh"},
	{0x91F18F1B18385CA3ULL, "vxht_0_hightown_a3_627_0_0.vlm"},
	{0x91F1B472587AC8F3ULL, "vxht_brothel_n_4y5_0_0.vlm"},
	{0x91F21C0F52D0D25BULL, "llac_lt_gallowscourtyard_65534_2.dds"},
	{0x91F2479C36F558D0ULL, "6109224_m.fxe"},
	{0x91F286C357A68F5CULL, "vxlt_gallowscourtyard_769_0_0.vlm"},
	{0x91F2C6023D109FACULL, "vxlt_blackemporium_wb_0_0.vlm"},
	{0x91F2EFAAEF3FE701ULL, "vxlt_shop_fc_0_0.vlm"},
	{0x91F302A41E40149CULL, "215400721.ogg"},
	{0x91F3719EBA97BA4DULL, "llac_ht_0_hightown_a2_-1_2.dds"},
	{0x91F372AA4555ED91ULL, "vxlt_0_lowtown_docks_cp6_0_0.vlm"},
	{0x91F3A554508E1CF2ULL, "vxht_mansion_n_3f5_0_0.vlm"},
	{0x91F3E2F7D4F66E7EULL, "zz_followers_banter6159729.cl"},
	{0x91F3E5D279298A1FULL, "6103346_m.fxe"},
	{0x91F3E779280EA9A7ULL, "dae011_drunk2_sta6144329.cl"},
	{0x91F4426FF9D0B587ULL, "573150327.ogg"},
	{0x91F466C792ECDFB8ULL, "6100014_m.fxe"},
	{0x91F4A91135BAE520ULL, "llac_ht_0_hightown_-131073_1.dds"},
	{0x91F4D6C7AC93D9ADULL, "vxlt_darktown_5nd_0_0.vlm"},
	{0x91F4ED84689E1F46ULL, "lt_0_lowtown_docks_12_0.rml"},
	{0x91F51A769AF5EC91ULL, "gen_im_acc_amu_lore_03.uti"},
	{0x91F57A6FB47D2236ULL, "character_character_p72.psh.fxo.dx11.assembly"},
	{0x91F5DC7C7A122533ULL, "hlac_lt_0_lowtown_docks_n_65533_0x.dds"},
	{0x91F5F94E02CF8B14ULL, "6101534_m.fxe"},
	{0x91F61831AFB4E257ULL, "6067762_f.fxe"},
	{0x91F62D14EFF13F26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_forward.evt"},
	{0x91F67CAC28F18855ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\right_up.evt"},
	{0x91F69CB142A9DC2EULL, "hlac_lt_0_lowtown_a2_65533_3x.dds"},
	{0x91F72388C95745E3ULL, "vrc321_varric6086884.cl"},
	{0x91F78B9FA14C195EULL, "6042079_m.fxe"},
	{0x91F794AAA41BE923ULL, "mcr310_argument6072671.cl"},
	{0x91F7EAAE294812DEULL, "vxdr_0_deeproads_f0n_0_0.vlm"},
	{0x91F833F89E8DBA06ULL, "one140cr_charade6137631.cl"},
	{0x91F84DDFED0015B9ULL, "6113965_m.fxe"},
	{0x91F859D7462980B5ULL, "aip_coteriethug2h_base.gda"},
	{0x91F8943C5FA6BAAEULL, "lt_0_lowtown_docks_a3_n_106_0.phy"},
	{0x91F8C8FC222C81CDULL, "dae011_wench_mob.bnk"},
	{0x91F9111ED2DE7A68ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_101a_0t.dds"},
	{0x91F9AC8DF0A41B96ULL, "6147938_m.fxe"},
	{0x91F9C036712BCDE7ULL, "hlac_ht_chantry_n_2534_6y.dds"},
	{0x91FA0BB7A4CC9472ULL, "dae150_chandan6148799.cl"},
	{0x91FA127785B75FD1ULL, "670014490.ogg"},
	{0x91FA4DBA06829E6BULL, "vxdr_primevaltomb_8ne_0_0.vlm"},
	{0x91FA565AC5E49BC0ULL, "gen_im_wep_war_two_act2_01.uti"},
	{0x91FA835E07990E09ULL, "511976268.ogg"},
	{0x91FAA98A3BEEE4F1ULL, "6128332_m.fxe"},
	{0x91FB8F54BF7A6E74ULL, "vxw_woundedcoast_ev_0_0.vlm"},
	{0x91FBEAFC222EF135ULL, "dae011_wench_mob.cnv"},
	{0x91FC4414EFF5D271ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_forward.gad"},
	{0x91FC6FCC8949D1EBULL, "vxw_bonepitdestroyed_a3_1hx_0_0.vlm"},
	{0x91FCE187F20D8313ULL, "vxlt_0_lowtown_a3_n_c_e18_0_0.vlm"},
	{0x91FD2EE5315AAD60ULL, "gen00fl_sebastian6177819.cl"},
	{0x91FD48B044F06F01ULL, "6139756_m.fxe"},
	{0x91FDC1799E29C873ULL, "dre000_mother6030241.cl"},
	{0x91FDE3FB5FE7DEF2ULL, "874355130.ogg"},
	{0x91FE255F1FC85197ULL, "153483087.ogg"},
	{0x91FF098CAE586A85ULL, "zz_vault_debug6203241.cl"},
	{0x91FF345F75E5900DULL, "art\\vfx\\combat\\vfx_q_ari_buf_act_c.anb"},
	{0x91FF8870C66A79B1ULL, "dae000_gamlen6149488.cl"},
	{0x91FF9EC9EAD7057EULL, "6087025_m.fxe"},
	{0x91FFB63DEBEA8B03ULL, "vxht_mansion2_32o_0_0.vlm"},
	{0x920063B7179E6E81ULL, "229847229.ogg"},
	{0x920070073285F437ULL, "6089116_m.fxe"},
	{0x92009F30C5D1E33DULL, "vxlt_gallowscourtyard_6x7_0_0.vlm"},
	{0x9200B3C5FC5CB7BEULL, "6026084-mother-thinking.cut"},
	{0x9200C653F663709AULL, "vxw_bonepitmining_a2_ky_0_0.vlm"},
	{0x9200E9D228A5FB42ULL, "hlac_ca_caverns1_mines_24424_28x.dds"},
	{0x9201C4F705F617E8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowalcove.msh"},
	{0x9201D6F705F63679ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowalcove.mao"},
	{0x9201E2F705F64A12ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowalcove.mmh"},
	{0x92025B52FCB7E1B2ULL, "gen00fl_aveline6139909.cl"},
	{0x920296646E44F221ULL, "vxlt_0_lowtown_n_df1_0_0.vlm"},
	{0x9202A6577663C4B3ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_meda_s.dds"},
	{0x92033A50FE452FB7ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_106a_d.dds"},
	{0x92034514EFFBF21FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_jolt_forward.ani"},
	{0x92034DA1B2799E30ULL, "hlac_w_sundermount_327677_0y.dds"},
	{0x92038959A6720D03ULL, "one280_emrys.bnk"},
	{0x92039F74EF9ED438ULL, "6067861_m.fxe"},
	{0x9203ACBCFCFE02C8ULL, "532857147.ogg"},
	{0x9203ADDEAE9C4467ULL, "vxdr_primevalentrance_1qi_0_0.vlm"},
	{0x9204303C650ADE2BULL, "107742947.ogg"},
	{0x9204A2D708C22337ULL, "6156366_m.fxe"},
	{0x92053A93356FBE9BULL, "6125948_m.fxe"},
	{0x92054649005E7916ULL, "vxlt_0_lowtown_a2_n_fgw_0_0.vlm"},
	{0x92054C45DD20D762ULL, "hlac_dr_0_deeproads_5845_9y.dds"},
	{0x9205742A6DF06F1DULL, "hlac_ht_chantry_2534_26z.dds"},
	{0x9205D7CD1EA6299EULL, "vxht_0_hightown_avx_0_0.vlm"},
	{0x92060F6475425AB2ULL, "151219434.ogg"},
	{0x92066B59A6741033ULL, "one280_emrys.cnv"},
	{0x9206D3386DF80D84ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.impa.ani"},
	{0x920704F0565B12C0ULL, "6155892_m.fxe"},
	{0x92072CF441A8A9CFULL, "zz_fen_debug6165665.cl"},
	{0x920800A4FC064D56ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_2pennants.msh"},
	{0x92080AA4FC065E54ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_2pennants.mmh"},
	{0x92080EA4FC066567ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_2pennants.mao"},
	{0x920832733701E38AULL, "6110839_m.fxe"},
	{0x920908D6A7CDF7CCULL, "vxw_encounter2_camp_2r_0_0.vlm"},
	{0x920911797D557667ULL, "6016026_m.fxe"},
	{0x9209CB929C6D5D39ULL, "769800045.ogg"},
	{0x9209DD716C52BB00ULL, "dae220ip_at_player_home_nt.utp"},
	{0x9209E3FEF291FAA0ULL, "dae021_torelin_mob6147941.cl"},
	{0x920A0131B17086F4ULL, "vxw_woundedcoast_bw_0_0.vlm"},
	{0x920A3B2E3913D212ULL, "dae020pt_amb_keep.ncs"},
	{0x920B267A90BF1BE2ULL, "hlac_w_bonepitmining_a2_327681_0y.dds"},
	{0x920B2B88F71A92A3ULL, "6125395_m.fxe"},
	{0x920B2CF4F4BB08AFULL, "309291980.ogg"},
	{0x920B9511AEB7026EULL, "vrc321_housekeeper6085652.cl"},
	{0x920BE0EC2DCA44ECULL, "284481449.ogg"},
	{0x920BF04C7E823902ULL, "hlac_lt_gallowstemplar_2394_7y.dds"},
	{0x920C09C1074A46CAULL, "lgt311_shw_mer_mettin6156576.cl"},
	{0x920C9AB50CA3DA96ULL, "dre000_varric6029394.cl"},
	{0x920CCE124F51A4A7ULL, "mer210_coterie_contact6079359.cl"},
	{0x920CEEAFCF720370ULL, "fx_dis_noise02.mao"},
	{0x920D08750E1EB004ULL, "6135692_m.fxe"},
	{0x920DD0386DFE27E6ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.impa.gad"},
	{0x920DD679BA98C229ULL, "226350582.ogg"},
	{0x920E0A0ACA684871ULL, "cs_mcr371cs_orsino_defiance.wbd"},
	{0x920E6C0C4C22AE4FULL, "vxdr_primevaltomb_5aq_0_0.vlm"},
	{0x920EA8BA0BE3C661ULL, "vxlt_hangedman_t2_0_0.vlm"},
	{0x920ED48F9221E618ULL, "hla_ht_0_hightown_n_-131073_0y.dds"},
	{0x920F101E2AA480B4ULL, "llac_lt_0_lowtown_a2_n_-65537_3.dds"},
	{0x920F2CF6F90769F7ULL, "vxlt_0_lowtown_a2_gng_0_0.vlm"},
	{0x920F8558CE6616ABULL, "vxw_woundedcoast_1do_0_0.vlm"},
	{0x920FCD772608F631ULL, "llac_dr_slavers_3217_4.dds"},
	{0x9210C5A21220492BULL, "6122534_m.fxe"},
	{0x9210D163E7BE4E81ULL, "21244334.ogg"},
	{0x9211BA8190800AADULL, "vxlt_darktown_5tb_0_0.vlm"},
	{0x92125B11C67D6D0BULL, "zz_followers_banter6157706.cl"},
	{0x9212884D8F5D0B81ULL, "lt_gallowscourtyard_n_59_0_lc.mmh"},
	{0x9213294B794B3CE2ULL, "vxlt_0_lowtown_a2_gc5_0_0.vlm"},
	{0x9213CB386E028AC5ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.impa.evt"},
	{0x9214A949158375FBULL, "222204472.ogg"},
	{0x9214B1318F6F5500ULL, "dre000_cut_sibling_dies6036323.cl"},
	{0x9214D82CCA04FAD8ULL, "269608317.ogg"},
	{0x9214F89F2EBE504FULL, "6046892_m.fxe"},
	{0x92150264C5B997D1ULL, "vxca_cavern1_haunted_e8d_0_0.vlm"},
	{0x9215857174AC6486ULL, "vxlt_darktown_57p_0_0.vlm"},
	{0x92165C6D21F7CB1AULL, "vxlt_0_lowtown_n_e81_0_0.vlm"},
	{0x921662CFC1BD56D5ULL, "224486450.ogg"},
	{0x921664DD492DBEB7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basicarch.mmh"},
	{0x921668DD492DC5C4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basicarch.mao"},
	{0x921674F3C944CD06ULL, "dae000_bartrand6206861.cl"},
	{0x921675ED8F715D32ULL, "character_character_p173.psh.fxo.dx11"},
	{0x92167ADD492DE429ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_basicarch.msh"},
	{0x92173173EA6ECF6BULL, "vxdr_primevaltomb_8dg_0_0.vlm"},
	{0x92179CFFF821012EULL, "6105767_m.fxe"},
	{0x9218316FCD387248ULL, "vxlt_hangedman_t7_0_0.vlm"},
	{0x92188A2BBBFFC488ULL, "hlac_lt_0_lowtown_a3_n_c_-1_1z.dds"},
	{0x92191853924B49C3ULL, "hlac_lt_0_lowtown_a2_n_65535_9x.dds"},
	{0x921982DC566DB9DCULL, "dae000_orsino6085462.cl"},
	{0x9219B31FD952147FULL, "smoke_add.mao"},
	{0x921A65716D06BE59ULL, "vxht_mansion2_1az_0_0.vlm"},
	{0x921AC88F919DF78EULL, "277744120.ogg"},
	{0x921B0E9A4D77F5B3ULL, "vxlt_0_lowtown_a3_n_c_feg_0_0.vlm"},
	{0x921B3FBF43F3004CULL, "follower_banter6155484.cl"},
	{0x921B5DE429EFB8C5ULL, "one220_yevhen6104156.cl"},
	{0x921C1A3F0D719B13ULL, "6131577_f.fxe"},
	{0x921C1C1F85A3A981ULL, "427407001.ogg"},
	{0x921C24D680D017B4ULL, "741480266.ogg"},
	{0x921C425F4B1520A9ULL, "6135598_m.fxe"},
	{0x921CBB917831D987ULL, "795846578.ogg"},
	{0x921D1945D06123B5ULL, "1043981573.ogg"},
	{0x921D25D0F6FF03C4ULL, "lla_ht_0_hightown_n_-131072_0.dds"},
	{0x921DDEAD7E3DD943ULL, "zz_fol_debug6092321.cl"},
	{0x921DF735C05C2F2BULL, "vxw_woundedcoast_1on_0_0.vlm"},
	{0x921E599B77CEED5EULL, "289952830.ogg"},
	{0x921E74F43404AFB9ULL, "vxht_mansion_n_5hf_0_0.vlm"},
	{0x921E9C37E67FAB10ULL, "vxlt_0_lowtown_n_cmr_0_0.vlm"},
	{0x921EAA94122AE060ULL, "dae150_chandan6148692.cl"},
	{0x921EDC71A35C4B5DULL, "6087585_m.fxe"},
	{0x921EE06E5404571CULL, "combat_smaoe_r2_a4.gda"},
	{0x921FC17B8B2ED876ULL, "vxlt_0_lowtown_a2_n_hdp_0_0.vlm"},
	{0x921FDB51A68EA446ULL, "198366012.ogg"},
	{0x921FE24450B3579DULL, "6126587_m.fxe"},
	{0x922061E2ED0ECE4BULL, "6043381_m.fxe"},
	{0x9220811CD0969888ULL, "hlac_lt_warehouse_n_1_2y.dds"},
	{0x92208912F1F329ECULL, "gen00fl_anders6116824.cl"},
	{0x92209553FB901403ULL, "art\\levels\\races\\design\\_textures\\trap_generic_d.dds"},
	{0x922129CE7484A775ULL, "t3_dog_d05.tnt"},
	{0x9221302746A1E320ULL, "t3_hey_gy4.tnt"},
	{0x92215A9B749C5FE8ULL, "mag170_thrask6020535.cl"},
	{0x9221658701254FDBULL, "6082342_m.fxe"},
	{0x92216A12E225439FULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medazz.phy"},
	{0x9222205B45775EACULL, "vxht_mansion2b_287_0_0.vlm"},
	{0x92227A1C591FD52FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxa.mao"},
	{0x9222861C591FE9CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxa.mmh"},
	{0x92228C1C591FF3BEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxa.msh"},
	{0x922292541670DCBDULL, "vxdr_primevaltomb_4xi_0_0.vlm"},
	{0x9222BAD22E2B51EFULL, "ave121_jeven6105707.cl"},
	{0x9222D6EC9A102648ULL, "958374567.ogg"},
	{0x9222DE2C8F856158ULL, "one220_yevhen6120896.cl"},
	{0x922307FE118CD7CBULL, "6143485_m.fxe"},
	{0x9223379E92EBFBB6ULL, "6032048_f.fxe"},
	{0x922346A5A2A7717BULL, "mer220_hubert6081405.cl"},
	{0x9223F74E66A27D06ULL, "vxlt_darktown_58q_0_0.vlm"},
	{0x92242DB680144962ULL, "llac_dr_dungeon_3215_7.dds"},
	{0x92251BD65FE41FD3ULL, "444719369.ogg"},
	{0x9225813E04DD5E54ULL, "6028947_m.fxe"},
	{0x9225A411B478EDE3ULL, "lt_0_lowtown_docks_n_250_0.mao"},
	{0x9225A811B478F430ULL, "lt_0_lowtown_docks_n_250_0.mmh"},
	{0x9225B611B4790C4AULL, "lt_0_lowtown_docks_n_250_0.msh"},
	{0x9225D5511B225BF1ULL, "6172065_m.fxe"},
	{0x92261BFA85E0FF1DULL, "ht_0_hightownexplosion_n_12_0.phy"},
	{0x922636AF4181D84FULL, "dae000_gen_player_turn_in6177832.cl"},
	{0x922651352572DCA6ULL, "6087818_m.fxe"},
	{0x92267FDCC1138D80ULL, "vxw_bonepit_n2_0_0.vlm"},
	{0x922683BB625463FBULL, "vxlt_darktown_5iw_0_0.vlm"},
	{0x9226E4EB955CB7F4ULL, "vxw_woundedcoast_tu_0_0.vlm"},
	{0x9227139659C2352EULL, "dae000_flemeth6013369.cl"},
	{0x9227BFE45ACD0600ULL, "vxca_cavern1_haunted_iif_0_0.vlm"},
	{0x9227E014760E51DFULL, "hlac_lt_0_lowtown_-196608_0y.dds"},
	{0x922809E6C771076FULL, "445267018.ogg"},
	{0x922A269803A28146ULL, "156472274.ogg"},
	{0x922AE367B4FF8D41ULL, "6111995_f.fxe"},
	{0x922B5DF8F1031013ULL, "hlac_lt_0_lowtown_a2_n_-1_1z.dds"},
	{0x922B84114859F695ULL, "6030680_m.fxe"},
	{0x922BBF5DEF9D98E8ULL, "vxdr_slavers_12s_0_0.vlm"},
	{0x922BDF91A2933E73ULL, "hlac_ca_cavern1_haunted_24421_2z.dds"},
	{0x922C50AA60792C9FULL, "vxht_bartrand_n_68p_0_0.vlm"},
	{0x922C6912B25247ABULL, "vxw_woundedcoast_1ja_0_0.vlm"},
	{0x922C8BA0A30C02A6ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.anb"},
	{0x922CC889687D9F62ULL, "hlac_ht_mansion2_818_43y.dds"},
	{0x922CE5198E064039ULL, "plt_lgt000pt_lite_key_04.nss"},
	{0x922D1CFA85E71FA9ULL, "ht_0_hightownexplosion_n_12_0.rml"},
	{0x922D5378FD442415ULL, "ser191_emeric6104779.cl"},
	{0x922F14FABE98AC2EULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyo.msh"},
	{0x922F22FABE98C47FULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyo.mao"},
	{0x922F2EFABE98D81CULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyo.mmh"},
	{0x922F49F42A357DC7ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.cess.gad"},
	{0x922F732327219F5DULL, "art\\characters\\weapons\\greataxes\\w_gax_001a.phy"},
	{0x922FCD85529D1546ULL, "6147156_m.fxe"},
	{0x922FD763F31B685BULL, "6085242_m.fxe"},
	{0x92304AF2853FE39AULL, "mag100_thrask6051767.cl"},
	{0x92304D37550712C2ULL, "hlac_lt_darktown_-1_16y.dds"},
	{0x9230859D99C31DF4ULL, "762191907.ogg"},
	{0x923098D479B1DD42ULL, "isa111_isabela6067115.cl"},
	{0x9230F689084F1622ULL, "ch2armtnt.psh.fxo"},
	{0x923159065A2D8364ULL, "gen00fl_anders6125718.cl"},
	{0x9232227EB161F6FEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_pc_flemeth_real.ani"},
	{0x9232259A35925A24ULL, "vxlt_0_lowtown_a2_n_fg5_0_0.vlm"},
	{0x92322DB69AE7341CULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestwood.mao"},
	{0x923239B69AE748BFULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestwood.mmh"},
	{0x92323FB69AE75361ULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestwood.msh"},
	{0x92326B12E233B8A4ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medazz.mmh"},
	{0x92326DD4F800F0E0ULL, "240257080.ogg"},
	{0x92326F12E233BFF7ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medazz.mao"},
	{0x923280CDBD75DDB5ULL, "452632410.ogg"},
	{0x92328112E233DE46ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medazz.msh"},
	{0x9232F2346461148DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appb_0t.dds"},
	{0x92339506D20E3645ULL, "vxht_0_hightown_a2_5jc_0_0.vlm"},
	{0x9233B14BD595A0CDULL, "6049150_m.fxe"},
	{0x9233CC198E0C3412ULL, "plt_lgt000pt_lite_key_04.plo"},
	{0x92340732351ABA98ULL, "958002264.ogg"},
	{0x9234401D74047304ULL, "vxlt_gallowstemplar_n_3s7_0_0.vlm"},
	{0x92348FE6E6538FACULL, "vxw_bonepitmining_a2_1wu_0_1.vlm"},
	{0x9234A4B029391E86ULL, "vxdr_primevaltomb_9e5_0_1.vlm"},
	{0x9234BECEC01F337DULL, "6208007_m.fxe"},
	{0x9234F1B9B309A600ULL, "vxw_bonepit_i3_0_0.vlm"},
	{0x9234F598544A7C7BULL, "vxlt_darktown_5rv_0_0.vlm"},
	{0x923530863CDD6F03ULL, "zz_approval_debug6003839.cl"},
	{0x92355F92183D20DCULL, "765057101.ogg"},
	{0x923597BF7A63361BULL, "mag360_thrask6043536.cl"},
	{0x9235BF0ECB3875A2ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_7y.dds"},
	{0x9235FB7C02CD2748ULL, "plt_cod_itm_handsofthecarta.plo"},
	{0x92363EF42A3B89D4ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.cess.evt"},
	{0x9236800A7679469EULL, "6077970_m.fxe"},
	{0x92368B51B54112AFULL, "6018805_f.fxe"},
	{0x9236CAA6471437D3ULL, "wm_ico_woundedcoast.dds"},
	{0x923710FA85EF5F89ULL, "ht_0_hightownexplosion_n_12_0.mao"},
	{0x92371CFA85EF7422ULL, "ht_0_hightownexplosion_n_12_0.mmh"},
	{0x92371EFA85EF77D8ULL, "ht_0_hightownexplosion_n_12_0.msh"},
	{0x923786EE3C4BE5DDULL, "tbe200im_dull_amulet.uti"},
	{0x92378A31E30477DDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfallr.mao"},
	{0x92378E31E3047E16ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfallr.mmh"},
	{0x92379831E3048F14ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfallr.msh"},
	{0x92379BC205C1A95EULL, "47775723.ogg"},
	{0x9237A1A47168FD8DULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobb.mmh"},
	{0x9237A5A4716904DEULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobb.mao"},
	{0x9237B3A471691C13ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobb.msh"},
	{0x9237BC657FCA5E02ULL, "413257606.ogg"},
	{0x9237FE2E521C5B76ULL, "l_ave100ar_patrol.wbd"},
	{0x9238168DE6C80078ULL, "31340766.ogg"},
	{0x92382DCCFA729DA0ULL, "6125561_m.fxe"},
	{0x9238933C5FDC9D71ULL, "lt_0_lowtown_docks_a3_n_106_0.mmh"},
	{0x92389F3C5FDCB112ULL, "lt_0_lowtown_docks_a3_n_106_0.mao"},
	{0x9238AD3C5FDCC92FULL, "lt_0_lowtown_docks_a3_n_106_0.msh"},
	{0x9238BF2F41E94B56ULL, "12974746.ogg"},
	{0x9238E76D49CB8046ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_left.ani"},
	{0x923900F9B8B181E0ULL, "rdr361ar_cave.are"},
	{0x923979AA8AAA6FFAULL, "6104760_m.fxe"},
	{0x9239854E0CE25F6DULL, "t1_skn_d01.tnt"},
	{0x9239DB26860C9CC0ULL, "127070014.ogg"},
	{0x923A1DFFCB37750AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastplants_n.dds"},
	{0x923A86D69F17FB70ULL, "w_encounter2_45_0.mao"},
	{0x923A8AD69F180183ULL, "w_encounter2_45_0.mmh"},
	{0x923A8BDDF788CD24ULL, "206115966.ogg"},
	{0x923A8EA0A3184551ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.mmh"},
	{0x923A98D69F1819DDULL, "w_encounter2_45_0.msh"},
	{0x923B16D9EFBD9F09ULL, "6163900_m.fxe"},
	{0x923B1CF681D2B0F0ULL, "vxht_brothel_n_6bu_0_0.vlm"},
	{0x923B92B1ED1A644EULL, "w_bonepitdestroyed_a3_255_0.msh"},
	{0x923BA0B1ED1A7B9FULL, "w_bonepitdestroyed_a3_255_0.mao"},
	{0x923BACB1ED1A903CULL, "w_bonepitdestroyed_a3_255_0.mmh"},
	{0x923BACBC0E43CEC7ULL, "gen_im_npc_arm_hum_cht_cotd.uti"},
	{0x923BED427355A831ULL, "mag291_torpor6026979.cl"},
	{0x923BEE73AB8306E8ULL, "vxlt_0_lowtown_docks_coj_0_0.vlm"},
	{0x923C1615F4B93AADULL, "store_dae270cr_weapons.utm"},
	{0x923C191B4A3108A9ULL, "6087120_f.fxe"},
	{0x923CADC9FEC55783ULL, "hlac_lt_0_lowtown_a2_n_65536_8z.dds"},
	{0x923CD8B1EE4FEE90ULL, "gen_im_cft_reg_e1.uti"},
	{0x923D425F05F4018EULL, "art\\vfx\\combat\\vfx_m_blood_change_c.cess.ani"},
	{0x923D72DAB787823DULL, "348743385.ogg"},
	{0x923DF4C3FAEC830EULL, "vxlt_0_lowtown_a2_hfq_0_0.vlm"},
	{0x923F078BE30D3A2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_openarms.gad"},
	{0x923F10DCCA185314ULL, "gen00fl_anders6125808.cl"},
	{0x923F3ABC06584D07ULL, "qun211_party_comment6072258.cl"},
	{0x923F891BD8A89F9DULL, "dre000_varric6036953.cl"},
	{0x923FBE5FDD2B96E6ULL, "mag121cr_viveka.utc"},
	{0x923FCF3D7F748EE8ULL, "gen00fl_dog6146972.cl"},
	{0x923FFFB794845F83ULL, "211329766.ogg"},
	{0x92404C2C1AC4D8C9ULL, "vxca_caverns2prc_8kw_0_0.vlm"},
	{0x924057FB9B9BE151ULL, "vxca_cavern1_haunted_e7y_0_0.vlm"},
	{0x92406823272FFFC9ULL, "art\\characters\\weapons\\greataxes\\w_gax_001a.mao"},
	{0x9240742327301462ULL, "art\\characters\\weapons\\greataxes\\w_gax_001a.mmh"},
	{0x9240762327301718ULL, "art\\characters\\weapons\\greataxes\\w_gax_001a.msh"},
	{0x9240DAE6A1A93749ULL, "ser191cr_emeric.utc"},
	{0x92410B49CC896296ULL, "zz_followers_banter6158667.cl"},
	{0x92413C1BBA4F7225ULL, "6071940_m.fxe"},
	{0x92417E878606F554ULL, "vxht_0_hightown_n_an1_0_0.vlm"},
	{0x9241806A9C2B794BULL, "vxlt_gallowstemplar_n_5wr_0_0.vlm"},
	{0x9241A2AC2C6624A6ULL, "6190978_m.fxe"},
	{0x9241D44F14DC5630ULL, "967699440.ogg"},
	{0x92424638CC96BC06ULL, "isa211_isabela6081965.cl"},
	{0x9242571526825EF3ULL, "6004373-kirkwall-intro.cut"},
	{0x92425BF117C925CFULL, "19566725.ogg"},
	{0x924280439B21D41BULL, "6216987_m.fxe"},
	{0x92430FF9B8B9EFF3ULL, "rdr361ar_cave.lst"},
	{0x924336F42A46060DULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.cess.ani"},
	{0x9243754C6A9D7CDDULL, "468763676.ogg"},
	{0x9243F06E5EFE0506ULL, "324936566.ogg"},
	{0x92440D995DF8B315ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryc.phy"},
	{0x9244449962564EA3ULL, "one300_zevran6086513.cl"},
	{0x924460677ED8BDEEULL, "vxht_0_hightown_ben_0_0.vlm"},
	{0x924476D8E580420EULL, "259761558.ogg"},
	{0x9244E5748681B265ULL, "440336971.ogg"},
	{0x92461C8BE3137CDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_openarms.evt"},
	{0x92461FF89B04344DULL, "740781620.ogg"},
	{0x9246A5D05395A4CCULL, "lgt150_nat_were_supporter6110987.cl"},
	{0x9246AAC9D8D4C333ULL, "dae000_bartrand_b.stg"},
	{0x9246CF6D49D794B7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_left.evt"},
	{0x9246D02FA6C2F2F8ULL, "vxw_bonepitdestroyed_a3_57_0_0.vlm"},
	{0x924711143E9FD42EULL, "art\\vfx\\models\\vfx_rockarmor_lshin.mmh"},
	{0x92472B143EA0001CULL, "art\\vfx\\models\\vfx_rockarmor_lshin.msh"},
	{0x92473B7EB1747D78ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_pc_flemeth_real.gad"},
	{0x924784D0835B4129ULL, "393957867.ogg"},
	{0x9247CB1180DD27CCULL, "vxca_cavern1_haunted_gkk_0_1.vlm"},
	{0x92481D349AE4722AULL, "gen00fl_aveline6122080.cl"},
	{0x924835C4FB625BD8ULL, "hlac_dr_dungeon_3217_7z.dds"},
	{0x92489AA47177654AULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_nobb.phy"},
	{0x9248A65081879B90ULL, "6116761_m.fxe"},
	{0x9248E883D81D23C1ULL, "6059963_f.fxe"},
	{0x92494CCC963E78ABULL, "vxw_woundedcoast_1pc_0_0.vlm"},
	{0x924993F2FC4BFE6DULL, "6161186_m.fxe"},
	{0x9249963E0C3251CEULL, "6076035_f.fxe"},
	{0x9249F4878ED9BEAFULL, "vxlt_0_lowtown_eud_0_0.vlm"},
	{0x924A3A5F05FE7EDFULL, "art\\vfx\\combat\\vfx_m_blood_change_c.cess.evt"},
	{0x924A675C665D2DFDULL, "dae000_worthy6166565.cl"},
	{0x924A8151D5BA2764ULL, "llac_lt_warehouse_0_1.dds"},
	{0x924AB311501C7ED1ULL, "220687396.ogg"},
	{0x924B20F473DC95A2ULL, "vxlt_0_lowtown_docks_a3_n_d1q_0_0.vlm"},
	{0x924B2F244CC9A701ULL, "hlac_ht_0_hightown_a3_-131073_0y.dds"},
	{0x924B784F78BD0093ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_dead_loop.ani"},
	{0x924B8BD69F2676A8ULL, "w_encounter2_45_0.phy"},
	{0x924C1657E078E14CULL, "follower_barks6153441.cl"},
	{0x924C61A119A0999DULL, "llac_ca_cave_2193_1.dds"},
	{0x924C66A0ECE29B8EULL, "vxlt_0_lowtown_a2_hcr_0_0.vlm"},
	{0x924CA73B72C565F0ULL, "vxw_blightlands_9x_0_0.vlm"},
	{0x924CC3B51FF5D91CULL, "vxht_0_hightown_a3_c1t_0_0.vlm"},
	{0x924CC51746E3E413ULL, "vxlt_warehouse_n_13w_0_0.vlm"},
	{0x924CF99CE77FCC72ULL, "vxca_cavern1_haunted_ehk_0_0.vlm"},
	{0x924DA44973F06A06ULL, "687801329.ogg"},
	{0x924DD06D49DDB540ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_gesture_left.gad"},
	{0x924EBE090CBAF149ULL, "vxca_caverns2prc_8lv_0_0.vlm"},
	{0x924ED1A1EB9BF103ULL, "6127173_m.fxe"},
	{0x924F3763161DAB92ULL, "face_face_p1.vsh.fxo.dx11"},
	{0x92507C3EB9EEAE51ULL, "vxlt_undercitydungeon_1a2_0_0.vlm"},
	{0x9250C2E7AD1F0A2AULL, "625403806.ogg"},
	{0x92511B5F060468C8ULL, "art\\vfx\\combat\\vfx_m_blood_change_c.cess.gad"},
	{0x92518AD69F2AE0A0ULL, "w_encounter2_45_0.rml"},
	{0x92518E5B8B77CE3FULL, "gen00fl_anders6096624.cl"},
	{0x9251D072B16F77C1ULL, "zz_followers_banter6159207.cl"},
	{0x9252261B9E506425ULL, "496791159.ogg"},
	{0x9252C3F091E30361ULL, "llac_w_bonepit_393217_0.dds"},
	{0x9252FD47185A8732ULL, "954278580.ogg"},
	{0x92533E15EB503B47ULL, "xx_extwood.dds"},
	{0x925363519E14A725ULL, "lgt320_shw_ors_marlein6154749.cl"},
	{0x92538A286AC6A1CEULL, "mer241ip_tunnel_blocker.utp"},
	{0x925413F9FB5A343CULL, "6174783_m.fxe"},
	{0x9254248BE31FC7F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_openarms.ani"},
	{0x925440E9E85A0FABULL, "6053934_f.fxe"},
	{0x9254A49BC0C33164ULL, "one280_emrys6104297.cl"},
	{0x92550DFABEB96FE7ULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyo.phy"},
	{0x92556FD21439EE42ULL, "static_static_p23.psh.fxo.dx11.assembly"},
	{0x92557187EE1E1479ULL, "sse_ss_qunari_m_2.gda"},
	{0x9255BD411B451D5CULL, "347611127.ogg"},
	{0x9255C30D2F437E9FULL, "cd_seb321_leliana.bnk"},
	{0x9255FAA57DE2A8D8ULL, "142609309.ogg"},
	{0x925690771DAFCA3BULL, "cod_ltr_ggt_note4.utp"},
	{0x925691BB33B0EC62ULL, "lt_gallowscourtyard_n.rml"},
	{0x9256965C2F4A7A8EULL, "97112196.ogg"},
	{0x9257172F32A7FDFAULL, "sound_weapon_hits.gda"},
	{0x92576F462229F667ULL, "hlac_ca_cavern1_haunted_24424_25x.dds"},
	{0x9257AC7BF45B2AF8ULL, "795966569.ogg"},
	{0x9257E80333AADEC5ULL, "qun110_petrice6042273.cl"},
	{0x925800B065BEE1F0ULL, "vxht_brothel_n_6hs_0_0.vlm"},
	{0x9258737E76E03F48ULL, "20545923.ogg"},
	{0x92588C0A6B6D6F82ULL, "fen110_fenris6096219.cl"},
	{0x92589AA30B863F97ULL, "1058913921.ogg"},
	{0x9258D22D8F6F37E8ULL, "vxlt_0_lowtown_docks_cuh_0_0.vlm"},
	{0x9259704F78C930CAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_dead_loop.evt"},
	{0x925975EAE84EDC88ULL, "hla_ht_0_hightown_a2_-1_0z.dds"},
	{0x9259854AD3870CD0ULL, "6033621_f.fxe"},
	{0x9259BC342FB277DEULL, "gencr_arcane_horror.utc"},
	{0x9259E0D31B0E5BFCULL, "gen_im_wep_rog_arc_act2_01.uti"},
	{0x925A56BC426C233AULL, "6072270_m.fxe"},
	{0x925A934F5C0277E5ULL, "6155628_m.fxe"},
	{0x925AC35A39D060FFULL, "vxht_mansion_n_3nl_0_0.vlm"},
	{0x925B30875312D252ULL, "390475925.ogg"},
	{0x925B701B853CB9F3ULL, "147683756.ogg"},
	{0x925B8173A06F1A19ULL, "w_woundedcoast_224_0.msh"},
	{0x925B8B73A06F2B07ULL, "w_woundedcoast_224_0.mmh"},
	{0x925B8F73A06F3254ULL, "w_woundedcoast_224_0.mao"},
	{0x925BB35B1459F725ULL, "6067846_m.fxe"},
	{0x925BB44254779DA1ULL, "74763800.ogg"},
	{0x925BD2186FAA9407ULL, "vxdr_0_deeproads_esl_0_1.vlm"},
	{0x925BDAC5F6DE1FA7ULL, "vxlt_0_lowtown_a2_n_ei3_0_0.vlm"},
	{0x925BEFF37348608FULL, "hlac_ht_bartrand_n_3373_3y.dds"},
	{0x925C0746D2F024E9ULL, "qun100ar_exterior.lst"},
	{0x925C748E3C2EA189ULL, "6039940_m.fxe"},
	{0x925DC8D42B8AD03AULL, "vxht_0_hightown_7g5_0_0.vlm"},
	{0x925DDFC5ED8E1E39ULL, "llac_lt_gallowscourtyard_n_-131071_0.dds"},
	{0x925DE1F70644A70DULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowalcove.phy"},
	{0x925E8A69F39D75F1ULL, "vxlt_0_lowtown_fi5_0_0.vlm"},
	{0x925ECA1B67395911ULL, "vxlt_0_lowtown_a3_n_c_csf_0_0.vlm"},
	{0x925EE95316CD91A9ULL, "vxlt_gallowstemplar_f_3ja_0_0.vlm"},
	{0x925EEE1BABE4C6D1ULL, "vxlt_undercitydungeon_1d1_0_0.vlm"},
	{0x925F3947C812D196ULL, "ser221_quentin6123318.cl"},
	{0x925F5C23E2021CEBULL, "mrl150cr_shadow_warrior.utc"},
	{0x925F7BB85D4C5E70ULL, "fen210_dying_magister6108606.cl"},
	{0x925FED9DDFECC3E2ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_act_c.mmh"},
	{0x92600989357639FDULL, "359375540.ogg"},
	{0x92604F4F78CF173DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_dead_loop.gad"},
	{0x926083B0A20E5502ULL, "vxca_cavern1_haunted_ebe_0_0.vlm"},
	{0x926126FE06CE7048ULL, "vxht_mansion2_2ex_0_0.vlm"},
	{0x926141E6833DB01EULL, "872857557.ogg"},
	{0x92619D428CCEEEA5ULL, "163666928.ogg"},
	{0x9261C22802D29F5CULL, "vxw_bonepitdestroyed_a3_169_0_0.vlm"},
	{0x9261E22441FBBE3AULL, "6137472_m.fxe"},
	{0x92620BBB9EBA82EAULL, "vxht_keep_n_6b4_0_0.vlm"},
	{0x9262361B1D40822AULL, "ht_0_hightown_a3_33_0_lc.mmh"},
	{0x92624044998665B9ULL, "6071838_m.fxe"},
	{0x926249C76B94F425ULL, "720352680.ogg"},
	{0x926277A41C8BAFDFULL, "57061668.ogg"},
	{0x9262DF46D2F5FF20ULL, "qun100ar_exterior.ncs"},
	{0x92632B0CD2C1C90BULL, "gen00fl_anders6125298.cl"},
	{0x9263DB9AA7AEEDEAULL, "hlac_lt_0_lowtown_docks_a3_-65538_7x.dds"},
	{0x9263FDBEA113577AULL, "679102500.ogg"},
	{0x92641E37B535B370ULL, "ch1armtnt_nolit.psh.fxo"},
	{0x9264242E5241D6E4ULL, "l_ave100ar_patrol.bnk"},
	{0x9264F3899C939F45ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_act_c.mmh"},
	{0x9265023F29B6091EULL, "6093874_f.fxe"},
	{0x926582FED46DF7D2ULL, "vxlt_0_lowtown_n_glw_0_0.vlm"},
	{0x92659B15C2935D19ULL, "73700800.ogg"},
	{0x9265A703FA5C1668ULL, "gen00fl_fenris6127970.cl"},
	{0x9265BE4B36DE6D84ULL, "w_sundermount_210_0.msh"},
	{0x9265C83586161E63ULL, "vxlt_gallowstemplar_n_4u9_0_0.vlm"},
	{0x9265D04B36DE8CEDULL, "w_sundermount_210_0.mao"},
	{0x9265D44B36DE9326ULL, "w_sundermount_210_0.mmh"},
	{0x9265EA4CE50B2F66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedflurrish.ani"},
	{0x926629FD087176DDULL, "vxca_cavern1_haunted_hli_0_0.vlm"},
	{0x9266728D57B4FA70ULL, "vxht_brothel_n_6up_0_0.vlm"},
	{0x9266B56FA451FC39ULL, "zz_qun_debug6088084.cl"},
	{0x9266BA8792533003ULL, "vxw_encounter3_8l_0_0.vlm"},
	{0x9266D84172C5EFAFULL, "vxlt_0_lowtown_ekf_0_0.vlm"},
	{0x926761B246F9949CULL, "w_woundedcoast_75_0.rml"},
	{0x9268679E13761C56ULL, "387313195.ogg"},
	{0x9268F5611DE08CD0ULL, "vxw_woundedcoast_5b_0_0.vlm"},
	{0x92690CC1A11A44F0ULL, "hlac_lt_backalley_n_65536_6x.dds"},
	{0x926A47B246FB9EF1ULL, "w_woundedcoast_75_0.msh"},
	{0x926A4F6C69CA45CFULL, "vxw_bonepit_9p_0_0.vlm"},
	{0x926A50F09BFAC50CULL, "242352969.ogg"},
	{0x926A55B246FBB60CULL, "w_woundedcoast_75_0.mao"},
	{0x926A61B246FBCAAFULL, "w_woundedcoast_75_0.mmh"},
	{0x926A703B54AF71ECULL, "6121129_m.fxe"},
	{0x926AAD3568F356F0ULL, "vxw_sundermount_26e_0_0.vlm"},
	{0x926AB97949C53B17ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupright.phy"},
	{0x926B29514F622E4EULL, "vxlt_0_lowtown_a3_n_c_f4q_0_0.vlm"},
	{0x926B2AB8FDA7CF3DULL, "one130_thug6133100.cl"},
	{0x926B4187FBCBF5B8ULL, "hlac_lt_backalley_n_0_3z.dds"},
	{0x926BAAD6E7217939ULL, "212283144.ogg"},
	{0x926BCD73D6729DEEULL, "gen00fl_sebastian6179860.cl"},
	{0x926C11503ED45F3DULL, "6139899_m.fxe"},
	{0x926C34A67E58D388ULL, "plc_base.gda"},
	{0x926CBB9C25DBE517ULL, "200190173.ogg"},
	{0x926CE120BC52AC69ULL, "vxht_0_hightown_n_c3x_0_1.vlm"},
	{0x926D0046D2FE8B8EULL, "qun100ar_exterior.are"},
	{0x926D068EFF317B99ULL, "lla_ht_0_hightown_n_-1_0.dds"},
	{0x926D5FF89DDADF51ULL, "vxlt_undercitydungeon_1o0_0_0.vlm"},
	{0x926D74D10ABB0F1DULL, "dae000ip_crowd_em1_talk9.utp"},
	{0x926DC314CCFF280EULL, "6118484_m.fxe"},
	{0x926E0EB3C84AF773ULL, "vxw_bonepit_28k_0_0.vlm"},
	{0x926E217F2BD50975ULL, "vxw_encounter2_s5_0_1.vlm"},
	{0x926E774EEC5EE1CBULL, "118870617.ogg"},
	{0x926E863B04A462D5ULL, "llac_ca_cavern1_haunted_24426_1.dds"},
	{0x926EF58D94046D82ULL, "vxca_cavern1_haunted_egf_0_0.vlm"},
	{0x926F3CE2FE786155ULL, "127725621.ogg"},
	{0x926F7BB3575B9B27ULL, "gen00fl_isabela6090435.cl"},
	{0x926F98DAF8C488C8ULL, "vxht_mansion2_2ny_0_0.vlm"},
	{0x926FCB4B36E6D8F1ULL, "w_sundermount_210_0.phy"},
	{0x926FDA4640F242FBULL, "gen_im_wep_rog_arc_act1_03.uti"},
	{0x927000315B5C2337ULL, "6166530_m.fxe"},
	{0x9270396D76863DAEULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartospider.evt"},
	{0x92706443FEEEE37CULL, "one110_meeran6090194.cl"},
	{0x92706C84EF5A30DFULL, "6122743_m.fxe"},
	{0x92712B481EED605EULL, "zz_isa_debug6133201.cl"},
	{0x9271BBBA46F3A1F2ULL, "llac_lt_shop_-1_2.dds"},
	{0x9271BE11A597AF38ULL, "hlac_w_bonepitdestroyed_a3_2_0z.dds"},
	{0x92720016E3583D5AULL, "383571935.ogg"},
	{0x927291456BD3C90FULL, "569014790.ogg"},
	{0x927293F8FC5D7FB5ULL, "6208637_f.fxe"},
	{0x9272F0899C9FD712ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_act_c.anb"},
	{0x927318A3931D440DULL, "406777769.ogg"},
	{0x92733C22765DB9BFULL, "zz_mer_debug6062727.cl"},
	{0x9273D24CE51743D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedflurrish.evt"},
	{0x92740B3989E29609ULL, "995415718.ogg"},
	{0x927439FB07E49F7DULL, "vxw_bonepitmining_a2_21l_0_0.vlm"},
	{0x9274F0DFE333DA9EULL, "combat_deathblow_14.gda"},
	{0x9275195A1AD9C644ULL, "6109440_m.fxe"},
	{0x92754A1E64BC082FULL, "vxlt_0_lowtown_efa_0_0.vlm"},
	{0x92755736FD539FA8ULL, "gen00flst_varric_actthreea.stg"},
	{0x9275DD92B696021CULL, "weapon_weapon_p1.psh.fxo.dx11.assembly"},
	{0x9275F32D82399E84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_enter.ani"},
	{0x927643376BF107A5ULL, "6153262_m.fxe"},
	{0x9276637CF591DDB8ULL, "vxlt_0_lowtown_a2_f6p_0_1.vlm"},
	{0x927674303C9DD507ULL, "vxht_0_hightown_9b9_0_0.vlm"},
	{0x9276762681948CB9ULL, "367042621.ogg"},
	{0x927680403E43F290ULL, "6073886_m.fxe"},
	{0x9276BC6D55B36883ULL, "hlac_ca_cave_2194_11z.dds"},
	{0x9276DDD1098794B5ULL, "6079276_m.fxe"},
	{0x9277306D768C4D59ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartospider.gad"},
	{0x9277673E0FD6A550ULL, "vxw_woundedcoast_2c_0_0.vlm"},
	{0x9277F7799284F362ULL, "6091713_f.fxe"},
	{0x9278091519CE2347ULL, "mrl000_marethari6142296.cl"},
	{0x9278214FB0E859EFULL, "6155712_m.fxe"},
	{0x92788A39F818089FULL, "framebuffereffects_fadedistortion_p0.psh.fxo"},
	{0x9279086420AB919AULL, "qun111_ketojan6059228.cl"},
	{0x92796B6914E3CBAEULL, "dae021_keep_guard_1.utc"},
	{0x927990B7432E6B41ULL, "zz_followers_banter6157520.cl"},
	{0x9279CAB1A970DEDDULL, "162850290.ogg"},
	{0x927A0504170CEC63ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgta_0s.dds"},
	{0x927A3EAC5C5A2520ULL, "6067696_m.fxe"},
	{0x927A634FC77D19CEULL, "6202121_m.fxe"},
	{0x927AD34CE51D6460ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedflurrish.gad"},
	{0x927AF84F2C35EBAAULL, "596880864.ogg"},
	{0x927BBD21ABF9C239ULL, "6049231_m.fxe"},
	{0x927CDAFC32FA2DB5ULL, "6175658_f.fxe"},
	{0x927CF02D823FB8E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_enter.gad"},
	{0x927D038BBAE345ACULL, "32665424.ogg"},
	{0x927D6C8263D7D540ULL, "hlac_ht_keep_n_4519_9z.dds"},
	{0x927E07763496C4CFULL, "hf_gencr_raider.mor"},
	{0x927E416D769288A7ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartospider.ani"},
	{0x927E62493F599741ULL, "vxht_mansion2_1ta_0_0.vlm"},
	{0x927ECF23B7E338DCULL, "vxw_bonepit_ci_0_0.vlm"},
	{0x927F9B489644E4F2ULL, "ser120_ghyslain6119716.cl"},
	{0x927FEC31607C0013ULL, "6024987_f.fxe"},
	{0x927FF2C1E3D2FD7BULL, "gen00fl_isabela6079276.cl"},
	{0x927FFAB5A11DCF68ULL, "vxdr_primevalentrance_2hz_0_0.vlm"},
	{0x92800F06BD0E8447ULL, "dae000_mother6156403.cl"},
	{0x928025FCCBB3497FULL, "gen_im_trs_bst_spd_silk_04.uti"},
	{0x92818809CA9267B2ULL, "dae321st_isabela.stg"},
	{0x9281A5793E401678ULL, "hlac_lt_warehouse_n_65536_3x.dds"},
	{0x9281A90D2F688DD5ULL, "cd_seb321_leliana.wbd"},
	{0x9281A92C94B7D5EBULL, "vxht_0_hightown_n_c55_0_0.vlm"},
	{0x9281B089F25417CCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabs.phy"},
	{0x9281EDE9B8168B3CULL, "6015060_m.fxe"},
	{0x9282638E13808AC0ULL, "6108418_m.fxe"},
	{0x9282859C489FA097ULL, "6120557_f.fxe"},
	{0x9282EAB905173AD0ULL, "6179061_m.fxe"},
	{0x9282EB2D82441BC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_enter.evt"},
	{0x9282FDDC5E1F1328ULL, "hlac_w_woundedcoast_3_0z.dds"},
	{0x928360B12889D979ULL, "631792711.ogg"},
	{0x92838B370CCFDEC4ULL, "6102540_m.fxe"},
	{0x9283911DB6E2EBBDULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statuea.mao"},
	{0x928391B130B1E326ULL, "hlac_lt_0_lowtown_a3_n_c_0_3z.dds"},
	{0x9283951DB6E2F2F6ULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statuea.mmh"},
	{0x92839F1DB6E303F4ULL, "art\\levels\\races\\ferelden\\props\\chantry\\vista_statuea.msh"},
	{0x92842E5FEC708D9FULL, "501582389.ogg"},
	{0x9284774D6F57BDB0ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_sepulcher.mao"},
	{0x92847B4D6F57C4C3ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_sepulcher.mmh"},
	{0x9284894D6F57DC1DULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_sepulcher.msh"},
	{0x9284E362891ECCA8ULL, "6149008_m.fxe"},
	{0x928506F3E2694CEFULL, "follower_banter6166717.cl"},
	{0x928574433A7EEB96ULL, "69136698.ogg"},
	{0x9285D5830093A23FULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolem_overlay.anb"},
	{0x9285DFEE3667FF4FULL, "mag100_cullen6096062.cl"},
	{0x92877366D8429C35ULL, "llac_dr_primevaltomb_658_13.dds"},
	{0x9287773C43ABEDDBULL, "zz_mag_debug6025831.cl"},
	{0x928799EFB05B3688ULL, "6147654_f.fxe"},
	{0x9288259BCF60A86EULL, "hla_ht_0_hightown_n_65537_0y.dds"},
	{0x92884D63DB6C42DCULL, "vxlt_0_lowtown_a3_n_c_e2v_0_0.vlm"},
	{0x9288C0872E56739DULL, "llac_ht_keep_4519_18.dds"},
	{0x9288EC9DE00FC689ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_act_c.anb"},
	{0x9288F97878DB3015ULL, "llac_ht_0_hightown_a3_-65535_0.dds"},
	{0x92890BA4FF3320D1ULL, "935176644.ogg"},
	{0x928918708400EAB0ULL, "w_bonepitdestroyed_a3_mmap_ph.dds"},
	{0x92896850A4538E39ULL, "art\\characters\\creatures\\mubari\\animation\\p_rnf.ani"},
	{0x92898F744F6856BDULL, "6076858_f.fxe"},
	{0x9289B8903676F3FEULL, "hlac_lt_gallowstemplar_2388_0y.dds"},
	{0x9289F3D1C97CA95BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_enter.ani"},
	{0x928AA53A5F64ED4BULL, "vxdr_0_deeproads_d41_0_0.vlm"},
	{0x928AA6CEB0EB590EULL, "6112684_m.fxe"},
	{0x928AFF5468CF3239ULL, "art\\levels\\races\\proxy\\props\\_textures\\fence_d.dds"},
	{0x928B20995E34D630ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryc.msh"},
	{0x928B2E995E34EE4AULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryc.mmh"},
	{0x928B32995E34F481ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_flagchantryc.mao"},
	{0x928B69AA8AA11558ULL, "6152332_m.fxe"},
	{0x928BEC6CC6A4ABD4ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic.phy"},
	{0x928C120D284CB78DULL, "257393828.ogg"},
	{0x928C99B838A7386EULL, "vxht_0_hightown_bzc_0_0.vlm"},
	{0x928CA2B8E51B9298ULL, "gen00fl_fenris6107239.cl"},
	{0x928D30381CF9F146ULL, "97930170.ogg"},
	{0x928D4100A9D9515CULL, "vxw_bonepit_nn_0_0.vlm"},
	{0x928DA5D70C774405ULL, "vxht_mansion2b_1fn_0_1.vlm"},
	{0x928DB0406EABD2AEULL, "6041508_m.fxe"},
	{0x928EF76A09606FB3ULL, "vxca_cavern1_haunted_fn3_0_0.vlm"},
	{0x928F7FF5A0E96F7FULL, "vxlt_hangedman_n_1lz_0_0.vlm"},
	{0x928FA67E2BCD1FD9ULL, "hlac_lt_backalley_g_65537_5x.dds"},
	{0x92901B0986ADEE6BULL, "vxht_0_hightown_n_c24_0_0.vlm"},
	{0x92904350A4596E7BULL, "art\\characters\\creatures\\mubari\\animation\\p_rnf.gad"},
	{0x92904E48A286C83BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_212windows.mmh"},
	{0x92905A48A286DCD8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_212windows.mao"},
	{0x92906C48A286FB25ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_212windows.msh"},
	{0x929082BB33E24AB4ULL, "lt_gallowscourtyard_n.arl"},
	{0x9290AC014BC3BF08ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bench.msh"},
	{0x9290BE014BC3DD99ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bench.mao"},
	{0x9290CA014BC3F232ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bench.mmh"},
	{0x9290DD142C246F58ULL, "vxw_bonepitmining_a2_am_0_0.vlm"},
	{0x9290DE01E5EB9FA8ULL, "240119371.ogg"},
	{0x929155CB0527195FULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.mmh"},
	{0x92916D2ED814C0E5ULL, "llac_lt_0_lowtown_-262144_0.dds"},
	{0x92920F168E2E97C7ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_megatiki_sa.dds"},
	{0x92924A2DE8BF9406ULL, "255534998.ogg"},
	{0x929274A855015C8CULL, "zz_one_debug6215793.cl"},
	{0x9292A544DAA7EFF3ULL, "6043262_m.fxe"},
	{0x929362B2471ED1ECULL, "w_woundedcoast_75_0.anb"},
	{0x9293DA517DF3202FULL, "6161575_m.fxe"},
	{0x92948916277AE830ULL, "103112891.ogg"},
	{0x9294FE477818E8EAULL, "hlac_lt_0_lowtown_docks_a3_n_-3_6z.dds"},
	{0x92954C0ACEA81454ULL, "vxdr_primevalentrance_2i5_0_0.vlm"},
	{0x92957C4D6F6639E8ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_sepulcher.phy"},
	{0x92962E69FEE0581AULL, "905156693.ogg"},
	{0x92966050A45E0B60ULL, "art\\characters\\creatures\\mubari\\animation\\p_rnf.evt"},
	{0x929691E9E1154C93ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\gibexplode.ani"},
	{0x9296BA55B735A78DULL, "18075134.ogg"},
	{0x9296BF40CD625B5CULL, "vxlt_0_lowtown_a3_n_c_e5w_0_0.vlm"},
	{0x9296C96CC6AE77D1ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic.msh"},
	{0x9296D76CC6AE8FECULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic.mao"},
	{0x9296E36CC6AEA38FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic.mmh"},
	{0x9296E997BD5AB4FDULL, "fen210_hadriana6110523.cl"},
	{0x92974F6DF8247B91ULL, "one220_yevhen6099074.cl"},
	{0x92978B84C3EB4315ULL, "6139781_m.fxe"},
	{0x9297EA27B07618F9ULL, "vxlt_0_lowtown_docks_a3_cyr_0_0.vlm"},
	{0x9297FBD1C988F452ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_enter.evt"},
	{0x9298DBF0F20A661BULL, "vxht_0_hightown_a2_c0i_0_0.vlm"},
	{0x9299471C03498970ULL, "hlac_lt_0_lowtown_n_-131072_1z.dds"},
	{0x929A6F1454534B1BULL, "6167317_m.fxe"},
	{0x929A81F0B2EC7B7CULL, "art\\vfx\\combat\\vfx_m_pri_petrify_conj_c.impa.ani"},
	{0x929B58F4F95ABEC4ULL, "202083359.ogg"},
	{0x929B654F63229BF1ULL, "lt_0_lowtown_docks_a3_181_0.mmh"},
	{0x929B714F6322AF92ULL, "lt_0_lowtown_docks_a3_181_0.mao"},
	{0x929B73A96F401EB9ULL, "6084473_f.fxe"},
	{0x929B7F4F6322C7AFULL, "lt_0_lowtown_docks_a3_181_0.msh"},
	{0x929BE11CABB7D13AULL, "dae000_oldmangossip6174892.cl"},
	{0x929BF9AA4F640B0AULL, "vxca_caverns2_8iw_0_0.vlm"},
	{0x929C198FC96AC79CULL, "120205402.ogg"},
	{0x929C480554DE7096ULL, "follower_barks6153793.cl"},
	{0x929CBFCE712CBB3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_enter.ani"},
	{0x929D907FDB645FEDULL, "vxlt_gallowsprison_n_6hj_0_0.vlm"},
	{0x929DB00F83AAF7D1ULL, "6082039_f.fxe"},
	{0x929E179492F28C25ULL, "vxlt_shop_5b_0_0.vlm"},
	{0x929E5FC5A56494ECULL, "vxht_mansion2_1vz_0_0.vlm"},
	{0x929F1AD1C98F47E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_enter.gad"},
	{0x929F2AB094055198ULL, "6122123_f.fxe"},
	{0x929F4747F76CBBC0ULL, "6106618_m.fxe"},
	{0x929F4EF11E1A87D8ULL, "vxw_bonepitmining_a2_lj_0_0.vlm"},
	{0x929F9DE79E947CD8ULL, "ave121_jeven6105698.cl"},
	{0x929FA6D909A3CF85ULL, "391907622.ogg"},
	{0x92A0D1370D5CFFBFULL, "6122216_f.fxe"},
	{0x92A159580BC61F64ULL, "hlac_w_bonepitdestroyed_a3_458755_0x.dds"},
	{0x92A183FFFBCF37C9ULL, "131188120.ogg"},
	{0x92A18EF0B2F2AFBEULL, "art\\vfx\\combat\\vfx_m_pri_petrify_conj_c.impa.gad"},
	{0x92A2199BD98B19F7ULL, "hlac_ca_caverns2prc_5802_63z.dds"},
	{0x92A2C5331AE0EE6FULL, "fake_im_arm_boo_lgt_champ.uti"},
	{0x92A3088B9E5BDEE2ULL, "art\\vfx\\models\\vfx_crater_half.msh"},
	{0x92A30A8B9E5BE198ULL, "art\\vfx\\models\\vfx_crater_half.mmh"},
	{0x92A31CBC78187DB4ULL, "13606033.ogg"},
	{0x92A3C245C529E47FULL, "one100dg_nabil6060510.cl"},
	{0x92A3CCCE7132F07EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_enter.gad"},
	{0x92A41688CB8E2A9EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\combo_finish.ani"},
	{0x92A442CC6DCBE9C6ULL, "ico_longbow_5.dds"},
	{0x92A489E9E1217CCAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\gibexplode.evt"},
	{0x92A4DD0784FC6022ULL, "w_bonepitdestroyed_a3_42_0.msh"},
	{0x92A4DF0784FC63D8ULL, "w_bonepitdestroyed_a3_42_0.mmh"},
	{0x92A4EB0784FC787BULL, "w_bonepitdestroyed_a3_42_0.mao"},
	{0x92A4EFD0C767662BULL, "hlac_ht_0_hightown_a2_0_3y.dds"},
	{0x92A4F97FF366CB9EULL, "6070166_m.fxe"},
	{0x92A54E00D2F24BB6ULL, "6103747_m.fxe"},
	{0x92A5DF526CD70FF8ULL, "vxca_caverns1_gey_0_0.vlm"},
	{0x92A5FDD01EB1F5DBULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cast.ani"},
	{0x92A6283FE2576FBCULL, "302692769.ogg"},
	{0x92A69264660202DEULL, "llac_lt_0_lowtown_a2_n_c_131071_1.dds"},
	{0x92A6F513138D7208ULL, "654235472.ogg"},
	{0x92A71CB16C8F2571ULL, "vxlt_hangedman_1qx_0_0.vlm"},
	{0x92A7F714D054A79FULL, "sw_bl_dot.mao"},
	{0x92A84C21AEAE4273ULL, "101279081.ogg"},
	{0x92A8575F71960180ULL, "vxw_bonepitdestroyed_a3_1ri_0_0.vlm"},
	{0x92A869F0B2F88FDDULL, "art\\vfx\\combat\\vfx_m_pri_petrify_conj_c.impa.evt"},
	{0x92A8DF0785004CEBULL, "w_bonepitdestroyed_a3_42_0.rml"},
	{0x92A94410167073D9ULL, "6171446_m.fxe"},
	{0x92A994DDE9E4F4BEULL, "follower_banter6156173.cl"},
	{0x92AA13CF642BFC47ULL, "lgt000cr_medan.utc"},
	{0x92AA90785E478382ULL, "gui.mat"},
	{0x92AAA7CE7138CF9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_enter.evt"},
	{0x92AACE1F113EA713ULL, "character_character_p64.psh.fxo.dx11"},
	{0x92AB65E241667DC7ULL, "one140cr_charade.bnk"},
	{0x92AB68E9E127633DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\gibexplode.gad"},
	{0x92AB758BF0F7BA66ULL, "vxw_encounter3_ff_0_1.vlm"},
	{0x92AB852171CB061EULL, "hlac_lt_warehouse_n_65536_2y.dds"},
	{0x92ABBA6CE28D19D3ULL, "hlac_w_blightlands_0_6z.dds"},
	{0x92ABD6C9D3A8507DULL, "6208270_m.fxe"},
	{0x92ABEF86B4791196ULL, "269639276.ogg"},
	{0x92ABF49A0AA6C818ULL, "vxht_chantry_27j_0_0.vlm"},
	{0x92AC69E47CBB92DCULL, "6113069_m.fxe"},
	{0x92AC6D10B2AF37FFULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.cess.gad"},
	{0x92AC87C4E882A49AULL, "6127092_f.fxe"},
	{0x92ACC9F2845B09FCULL, "6111614_m.fxe"},
	{0x92ACCCDD91F036D2ULL, "vxdr_primevaltomb_7o8_0_0.vlm"},
	{0x92ACE463F3D78009ULL, "hlac_ca_cavern1_haunted_24421_12y.dds"},
	{0x92AD1CB095582D39ULL, "vxca_caverns1_j3c_0_0.vlm"},
	{0x92AD22EC3950BF41ULL, "6180009_f.fxe"},
	{0x92AD4E50BB0145E6ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic_1.mao"},
	{0x92AD54DFC05B4703ULL, "vxlt_0_lowtown_a2_exo_0_0.vlm"},
	{0x92AE22B468C3D3CFULL, "vxht_keep_n_3ya_0_0.vlm"},
	{0x92AE3B049681B512ULL, "6044966_f.fxe"},
	{0x92AE808BF177E79AULL, "hlac_w_encounter3_-393213_0y.dds"},
	{0x92AE87E24168ED3FULL, "one140cr_charade.cnv"},
	{0x92AEB0F9775B2026ULL, "plt_cod_hst_wardens.plo"},
	{0x92AF45C80DFBBAA0ULL, "6125730_m.fxe"},
	{0x92AFA0BBD3F47250ULL, "ave100_donnic6106123.cl"},
	{0x92B06938242018EFULL, "gen00fl_anders6215065.cl"},
	{0x92B168D9395EB710ULL, "1057880210.ogg"},
	{0x92B18E794A00F56FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupright.mao"},
	{0x92B19A794A01090CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupright.mmh"},
	{0x92B1A0794A0113FEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairupright.msh"},
	{0x92B1A6C2BADCA974ULL, "gen00fl_fenris6106658.cl"},
	{0x92B1A844031B587DULL, "502605555.ogg"},
	{0x92B22A48E0D77103ULL, "zz_dae_debug6029870.cl"},
	{0x92B22BF3FC10B0EEULL, "vxlt_gallowstemplar_5yz_0_0.vlm"},
	{0x92B22E88CB9A916FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\combo_finish.evt"},
	{0x92B2309466CDF817ULL, "308297118.ogg"},
	{0x92B2347C2F397FA9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\isabelle_dualweild_idle.ani"},
	{0x92B26210B2B390ECULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.cess.evt"},
	{0x92B2857ADD635AA2ULL, "gen_im_arm_hlm_med_act3_01.uti"},
	{0x92B2D99DE3C580A1ULL, "vxlt_0_lowtown_n_gce_0_0.vlm"},
	{0x92B312B5C4B89D33ULL, "507146310.ogg"},
	{0x92B386621A3737C4ULL, "dae000cr_grand_cleric6185496.cl"},
	{0x92B405D01EBE40D2ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cast.evt"},
	{0x92B41D1E17B385DBULL, "vxca_caverns1_mines_gjz_0_0.vlm"},
	{0x92B4609026F032C4ULL, "vxw_bonepitmining_a2_co_0_0.vlm"},
	{0x92B467741FC4A16BULL, "vxw_woundedcoast_np_0_0.vlm"},
	{0x92B4868AFF4F1C0CULL, "6115328_m.fxe"},
	{0x92B57B3EB4F51CF6ULL, "mag311_keran6041687.cl"},
	{0x92B6267B3E7053A9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_basic.phy"},
	{0x92B642FB8E18EE62ULL, "6044903_m.fxe"},
	{0x92B64873BB4DD141ULL, "6097012_m.fxe"},
	{0x92B67E7D5565B31CULL, "6116372_m.fxe"},
	{0x92B6D3FC48682EEDULL, "vxht_mansion_n_4wj_0_0.vlm"},
	{0x92B6DBFDD240A779ULL, "vxca_caverns2_7o9_0_0.vlm"},
	{0x92B6FAF92A9FB434ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml.mmh"},
	{0x92B74E67268B654EULL, "gen00fl_anders6125107.cl"},
	{0x92B7A035675AE158ULL, "755237786.ogg"},
	{0x92B7BDEF19A95379ULL, "459480924.ogg"},
	{0x92B7CC85F0E24A3DULL, "417390896.ogg"},
	{0x92B7D12E2B59300EULL, "vxdr_primevaltomb_3vs_0_0.vlm"},
	{0x92B7E4D89C41EDCCULL, "dae000ip_crowd_lm3_talk18.utp"},
	{0x92B829BD0731F941ULL, "vxht_mansion2_1he_0_0.vlm"},
	{0x92B82F88CB9EFE18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\combo_finish.gad"},
	{0x92B84349A2B44C17ULL, "hf_clf_harley.mor"},
	{0x92B887B306681F87ULL, "265013231.ogg"},
	{0x92B88D6011ECAA14ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v1.ani"},
	{0x92B94F7C2F3FCC2BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\isabelle_dualweild_idle.gad"},
	{0x92B956CB05486C9CULL, "art\\vfx\\combat\\vfx_v_arc_single_proj_c.anb"},
	{0x92B98CEC6EB48D4FULL, "vxlt_0_lowtown_gkp_0_0.vlm"},
	{0x92B9BC9311A6E63CULL, "hlac_lt_backalley_g_0_7y.dds"},
	{0x92B9E016E6D4F3AEULL, "mer110_jansen6032865.cl"},
	{0x92BA6676FC9CE098ULL, "vxht_chantry_20k_0_0.vlm"},
	{0x92BA9C3DAC31E0EDULL, "one220_yevhen6103368.cl"},
	{0x92BB10ED3E9673AAULL, "6148672_m.fxe"},
	{0x92BB24D01EC49465ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cast.gad"},
	{0x92BBD9246575298CULL, "zz_car_debug.ncs"},
	{0x92BC0C8F08DA12C2ULL, "299576813.ogg"},
	{0x92BC43C5FA5F794FULL, "dae000_tranquil_16165109.cl"},
	{0x92BC76279A52F171ULL, "w_bonepit_126_0.mmh"},
	{0x92BC82279A530512ULL, "w_bonepit_126_0.mao"},
	{0x92BC90279A531D2FULL, "w_bonepit_126_0.msh"},
	{0x92BDE6856F71E571ULL, "vis121_saemus6167899.cl"},
	{0x92BE97DF3CEFDC59ULL, "vxdr_primevaltomb_6r4_0_1.vlm"},
	{0x92BEA4E01F80DA2EULL, "hlac_dr_primevalentrance_1724_4z.dds"},
	{0x92BF35A8EE04E670ULL, "6033544_f.fxe"},
	{0x92BF5A6E22B9844EULL, "vxlt_0_lowtown_h2i_0_0.vlm"},
	{0x92BF6FF0BCEBCEE4ULL, "dre180cr_stone_golem_statue.utc"},
	{0x92BF8A6011F2C476ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v1.gad"},
	{0x92BF91086EE85643ULL, "da2_prc_seb_pl-pl.cre"},
	{0x92BFBE4B9510B890ULL, "6116530_m.fxe"},
	{0x92BFF0D0A11E77C7ULL, "182996835.ogg"},
	{0x92C04A10B2BFA545ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.cess.ani"},
	{0x92C07FB52F84354AULL, "bu_dark.mao"},
	{0x92C0E8D18BC90B2AULL, "gen00fl_anders6125208.cl"},
	{0x92C12E673A5358B5ULL, "static_static_p59.psh.fxo.dx11"},
	{0x92C14DF00679419CULL, "gen_im_acc_rng_lore_02.uti"},
	{0x92C18789F289B597ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabs.mmh"},
	{0x92C18B89F289BCA4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabs.mao"},
	{0x92C19D89F289DB09ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_slabs.msh"},
	{0x92C1B7C20A26A1B4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha04a.phy"},
	{0x92C24009707AF9BAULL, "6042097_m.fxe"},
	{0x92C26395FC638D11ULL, "zz_approval_debug6003741.cl"},
	{0x92C2D95111BAB9EBULL, "vxw_woundedcoast_ks_0_0.vlm"},
	{0x92C33967E1861531ULL, "hlac_ca_caverns1_24427_3x.dds"},
	{0x92C349ECD1EAFD68ULL, "llac_lt_0_lowtown_n_65537_2.dds"},
	{0x92C3ABFD65D9BEF4ULL, "hlac_dr_0_deeproads_5845_0x.dds"},
	{0x92C3F97FE28D26A9ULL, "gen00fl_fenris6126444.cl"},
	{0x92C44798BF96F48AULL, "448800499.ogg"},
	{0x92C46C5AD2E4DE2CULL, "6060907_m.fxe"},
	{0x92C4D10EF14D364CULL, "vxw_encounter2_camp_qo_0_0.vlm"},
	{0x92C4E6D355785873ULL, "one350_terath6087122.cl"},
	{0x92C4E7C030992CD6ULL, "vxlt_0_lowtown_docks_b6c_0_1.vlm"},
	{0x92C4FE9ACCEAF02BULL, "dre000_sibling_has_blight6036495.cl"},
	{0x92C57054BDF72822ULL, "vxlt_gallowstemplar_f_432_0_0.vlm"},
	{0x92C61009AD8F2B9EULL, "6169516_m.fxe"},
	{0x92C670BEF054A07AULL, "6164014_m.fxe"},
	{0x92C6A56011F910D5ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v1.evt"},
	{0x92C6B7B93E8857AFULL, "6087028_m.fxe"},
	{0x92C6CE4E65D060C0ULL, "ser120_ghyslain6105526.cl"},
	{0x92C6ECEB111375DAULL, "vxlt_0_lowtown_a3_n_c_bxm_0_0.vlm"},
	{0x92C7052634E85B13ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.cess.evt"},
	{0x92C72C8FFC736DD0ULL, "vxca_caverns1_e8d_0_0.vlm"},
	{0x92C7466E1D3A38F2ULL, "vxlt_hangedman_n_136_0_0.vlm"},
	{0x92C74B153B1E0D4DULL, "169255393.ogg"},
	{0x92C7D6C3DF32DBA7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgta_0s.dds"},
	{0x92C8092A77DE3338ULL, "89706354.ogg"},
	{0x92C84D13C029E634ULL, "1065043234.ogg"},
	{0x92C88722A75A0576ULL, "vxca_caverns2prc_9g9_0_0.vlm"},
	{0x92C91769A9D8B73FULL, "vxht_mansion2_1vu_0_0.vlm"},
	{0x92C92D7C87923120ULL, "dae000_gen_player_turn_in6177859.cl"},
	{0x92C9350C5DEBD8EEULL, "811879220.ogg"},
	{0x92C9A64132EEC8E8ULL, "hlac_lt_gallowstemplar_7198_4y.dds"},
	{0x92C9B55C7B46DE6CULL, "vxht_mansion2_1gy_0_0.vlm"},
	{0x92CA09B4E23B009CULL, "727741251.ogg"},
	{0x92CAA7E83499C4CBULL, "6109216_m.fxe"},
	{0x92CAB191FD26F788ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st03a_0d.dds"},
	{0x92CB38BA673EB5ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.ani"},
	{0x92CB3CC3102652B0ULL, "6058864_f.fxe"},
	{0x92CB6B3A80DEAAA4ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_s2_movement.anb"},
	{0x92CB843785C5D6E6ULL, "370848710.ogg"},
	{0x92CB94C20A2EBAB1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha04a.msh"},
	{0x92CBA0601B227B25ULL, "vxca_cavern1_haunted_iht_0_0.vlm"},
	{0x92CBAEC20A2EE76FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha04a.mmh"},
	{0x92CBB122A2B93D24ULL, "reticletint.psh.fxo"},
	{0x92CC636615F2FCBFULL, "vxca_caverns1_mines_hih_0_0.vlm"},
	{0x92CC88B4BE5AC885ULL, "llac_lt_0_lowtown_a2_n_c_1_0.dds"},
	{0x92CCD3C4EC3D37CAULL, "hlac_w_sundermount_-65540_0x.dds"},
	{0x92CCF20F37613D9AULL, "lgt200_eau_tome6180576.cl"},
	{0x92CCFE2634ECBB54ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.cess.gad"},
	{0x92CD205F7A7B48B2ULL, "docks_male_sailor_mob6176639.cl"},
	{0x92CD384A5B3C85FAULL, "vxht_0_hightown_c4t_0_0.vlm"},
	{0x92CD441876623BB1ULL, "hlac_lt_gallowscourtyard_n_-65538_1y.dds"},
	{0x92CD690235315FB1ULL, "dre000_vampire6034305.cl"},
	{0x92CD7EF821B15DE0ULL, "hlac_w_bonepit_0_8x.dds"},
	{0x92CDCC4B14AF9CCEULL, "vxlt_0_lowtown_h7j_0_0.vlm"},
	{0x92CE657B0E580923ULL, "705445499.ogg"},
	{0x92CE9DBFF59B676CULL, "135895032.ogg"},
	{0x92CF7665286607F4ULL, "lt_0_lowtown_a3_n_c_43_0_lc.mmh"},
	{0x92CF979EB3EF3E68ULL, "6104599_m.fxe"},
	{0x92D06E5969740976ULL, "249914397.ogg"},
	{0x92D07C2AFFA7FAA6ULL, "fx_dis_blank.mao"},
	{0x92D081E24185C91EULL, "one140cr_charade.utc"},
	{0x92D08A378195CC59ULL, "122056977.ogg"},
	{0x92D1094714202ABFULL, "ts_kirk_hightown_01.gda"},
	{0x92D14288B78CFF66ULL, "6070799_m.fxe"},
	{0x92D145BA6743372EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.gad"},
	{0x92D15AA8163075D7ULL, "qun221_viscount6070157.cl"},
	{0x92D1A6D3EE852F30ULL, "art\\characters\\playercharacter\\shared\\textures\\cn_stn_0n.dds"},
	{0x92D1AEF0251BEA93ULL, "vxlt_0_lowtown_docks_n_b9p_0_1.vlm"},
	{0x92D1C6CC8850A653ULL, "6113677_m.fxe"},
	{0x92D222F47DEA2071ULL, "325307889.ogg"},
	{0x92D235C62975BAA0ULL, "gen_im_cns_cft_arn_02.uti"},
	{0x92D23A9DA9574C49ULL, "170841536.ogg"},
	{0x92D2F6704029A008ULL, "qun220cr_javaris6071505.cl"},
	{0x92D3B7B62C545FEDULL, "vxht_mansion_n_4ah_0_0.vlm"},
	{0x92D3E231AFED40A2ULL, "vxlt_gallowstemplar_f_483_0_0.vlm"},
	{0x92D40D2634F2F332ULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.cess.ani"},
	{0x92D44B468DC1913AULL, "mag110_arianni6016889.cl"},
	{0x92D4AE9335CAF967ULL, "vxlt_backalley_n_1ed_0_0.vlm"},
	{0x92D4B374CA6E9DA0ULL, "gen00fl_dog6167160.cl"},
	{0x92D51D9BCD0F5C4CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s1_c.impa.ani"},
	{0x92D5C81DC4E107E8ULL, "6096018_m.fxe"},
	{0x92D679785E6C97A7ULL, "gui.bnk"},
	{0x92D690114F5F658AULL, "235767374.ogg"},
	{0x92D6A2F9B7054AB9ULL, "6216787_m.fxe"},
	{0x92D6E841C7DD65CAULL, "hla_ht_0_hightown_a2_65537_0x.dds"},
	{0x92D708DF25860BA2ULL, "hlac_w_bonepitmining_a2_327682_0x.dds"},
	{0x92D75DA127AD2BF0ULL, "llac_dr_0_deeproads_5843_0.dds"},
	{0x92D76E9A55BC1886ULL, "6110008_m.fxe"},
	{0x92D76F48A2C30370ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_212windows.phy"},
	{0x92D7814B3DFC3F3FULL, "dae000_bodhan6129150.cl"},
	{0x92D79F90D2122174ULL, "hlac_w_sundermount_-65531_0z.dds"},
	{0x92D7A8BB657F14AEULL, "66199463.ogg"},
	{0x92D8199BCA0093BCULL, "357774719.ogg"},
	{0x92D840BA67494D0DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.evt"},
	{0x92D8716371668C2AULL, "76908660.ogg"},
	{0x92D87A1DF78F3574ULL, "903484679.ogg"},
	{0x92D8C7FC5E59265AULL, "llac_lt_0_lowtown_docks_n_-1_3.dds"},
	{0x92D8E330E293B210ULL, "ave331_jeven6138727.cl"},
	{0x92D9464358F05267ULL, "6052615_m.fxe"},
	{0x92D96666AEB9CB65ULL, "gen00fl_fenris6126505.cl"},
	{0x92D96AAAAEE007AFULL, "6197721_m.fxe"},
	{0x92D96B40F6C685F9ULL, "hlac_ht_bartrand_n_3382_1y.dds"},
	{0x92D9EE08C6EDC00BULL, "702472861.ogg"},
	{0x92DA503FD9366B18ULL, "733529021.ogg"},
	{0x92DA7A1749420170ULL, "hlac_ht_playershouse_n_855_26y.dds"},
	{0x92DA9169A5E3C875ULL, "380248192.ogg"},
	{0x92DB0B4B2C6B7383ULL, "vxht_keep_n_6gv_0_0.vlm"},
	{0x92DB1376B0CF39D9ULL, "ch1alphemnotnt.psh.fxo"},
	{0x92DB1AC84077AD1DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01plank.mao"},
	{0x92DB1EC84077B456ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01plank.mmh"},
	{0x92DB28C84077C554ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_01plank.msh"},
	{0x92DB3B830FC4C777ULL, "ser221_quentin6123388.cl"},
	{0x92DBBD85F45BB5B3ULL, "craft.ncs"},
	{0x92DC2A9BCD15908EULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s1_c.impa.gad"},
	{0x92DC76F9376521FFULL, "zz_nat000ar_temp.are"},
	{0x92DC9D6F20BF215AULL, "6126751_m.fxe"},
	{0x92DD25140554B2DFULL, "ser121_jethann6120003.cl"},
	{0x92DD7DABEFA050D0ULL, "fen310_fenris6118188.cl"},
	{0x92DE1AF966460D05ULL, "317619936.ogg"},
	{0x92DEBAD9812FF18CULL, "6095987_m.fxe"},
	{0x92DF49C60DB57672ULL, "hlac_lt_gamlenshouse_n_-1_6z.dds"},
	{0x92DF684E2FEBFC42ULL, "dae120cr_amb_dwarf_4.utc"},
	{0x92DFBB7E7E615A63ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_1.ani"},
	{0x92DFFDF92AC2BEEBULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml.anb"},
	{0x92E020CD17120313ULL, "vxlt_0_lowtown_docks_n_b6q_0_1.vlm"},
	{0x92E07BE9C70E3DFDULL, "vxht_mansion2b_2o3_0_0.vlm"},
	{0x92E1811A95C824C4ULL, "dae310ar_docks.lst"},
	{0x92E18F25D5C526ACULL, "c_dog_tatb.dds"},
	{0x92E1B4C8D539674CULL, "vxw_encounter2_camp_gm_0_0.vlm"},
	{0x92E1E8655AA7C4C6ULL, "llac_lt_0_lowtown_a2_65536_8.dds"},
	{0x92E20F7B3E9567CEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_basic.mmh"},
	{0x92E21B7B3E957C65ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_basic.mao"},
	{0x92E2259BCD19F46DULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s1_c.impa.evt"},
	{0x92E225F7166B3908ULL, "6076571_m.fxe"},
	{0x92E2297B3E9593BCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0pp_basic.msh"},
	{0x92E229931E4A786DULL, "vxht_mansion_n_4dg_0_0.vlm"},
	{0x92E26AF2C5736B57ULL, "gen00fl_aveline6121983.cl"},
	{0x92E29086446C0518ULL, "llac_ht_keep_3741_0.dds"},
	{0x92E2E93391491BFDULL, "dre000_dougal6131158.cl"},
	{0x92E3690EFD44B7E9ULL, "6004110_f.fxe"},
	{0x92E40440C7F4407AULL, "hlac_lt_gallowstemplar_f_2392_0y.dds"},
	{0x92E4CA0CA273726AULL, "544527245.ogg"},
	{0x92E5CA4FB37DFBF4ULL, "vxw_bonepit_15x_0_0.vlm"},
	{0x92E5EF0C96058FA1ULL, "6110986_m.fxe"},
	{0x92E652CD74C8BDBEULL, "372912499.ogg"},
	{0x92E6B9564AFBFAA7ULL, "vxlt_0_lowtown_docks_n_btj_0_0.vlm"},
	{0x92E73CCCC036B536ULL, "llac_w_blightlands_131071_0.dds"},
	{0x92E758BC1363E025ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisb_1.mao"},
	{0x92E7FE24659AA3BFULL, "zz_car_debug.cnv"},
	{0x92E819BFBACFF8BFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_n.dds"},
	{0x92E83C317F042ECAULL, "plt_cod_itm_shield_resolute.plo"},
	{0x92E8811A95CE432DULL, "dae310ar_docks.ncs"},
	{0x92E9268956E46D78ULL, "vxlt_darktown_2rr_0_0.vlm"},
	{0x92E928638C5FC01FULL, "6135228_m.fxe"},
	{0x92E9A39FFF872FEFULL, "950164794.ogg"},
	{0x92E9BE116DBAFCC5ULL, "vxht_0_hightown_c1p_0_0.vlm"},
	{0x92EA18E784825381ULL, "149707197.ogg"},
	{0x92EAAB49129BE438ULL, "vxlt_gallowscourtyard_n_76n_0_0.vlm"},
	{0x92EACE2F040A1394ULL, "185452425.ogg"},
	{0x92EAFC8F3619B5A1ULL, "vxca_caverns2prc_7un_0_0.vlm"},
	{0x92EB09617C2BF2F9ULL, "static_static_p42.psh.fxo.dx11"},
	{0x92EC471FAAA4D1B2ULL, "6135690_m.fxe"},
	{0x92ECADF9EA9A3370ULL, "vxw_sundermount_2cj_0_0.vlm"},
	{0x92ECC9014C124F2DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bench.phy"},
	{0x92ECD37E7E6C0D5AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_1.evt"},
	{0x92ED6FF93773889CULL, "zz_nat000ar_temp.lst"},
	{0x92ED9186D252A720ULL, "vxlt_0_lowtown_a3_n_c_d9b_0_0.vlm"},
	{0x92ED9309B9A96BD1ULL, "6049734_m.fxe"},
	{0x92EE2EE7E79D036BULL, "vxw_woundedcoast_zt_0_0.vlm"},
	{0x92EE6BAC7F9967D7ULL, "vxlt_undercitydungeon_1ne_0_0.vlm"},
	{0x92EE6E9DA89D4880ULL, "vxdr_primevaltomb_8oi_0_0.vlm"},
	{0x92EEC1329510EEEDULL, "878172608.ogg"},
	{0x92EEFD01078EA2FEULL, "59708836.ogg"},
	{0x92EF5602265D9FF1ULL, "vxlt_hangedman_1hs_0_0.vlm"},
	{0x92F0316B172D305DULL, "6119981_m.fxe"},
	{0x92F0F688C6E7180FULL, "follower_barks6153659.cl"},
	{0x92F1838EF910CFF8ULL, "art\\vfx\\combat\\vfx_glb_ele_c_minor_c.anb"},
	{0x92F18B2DCA95CE3CULL, "6078296_m.fxe"},
	{0x92F1924D19B72A67ULL, "vxlt_backalley_n_1kf_0_0.vlm"},
	{0x92F1CDFA7F45E1B5ULL, "6137710_m.fxe"},
	{0x92F1F56684E2B2E2ULL, "mag110_arianni6020484.cl"},
	{0x92F210B3A815C11BULL, "291535853.ogg"},
	{0x92F24EE28D8A4BADULL, "812940894.ogg"},
	{0x92F2A8B84207FE82ULL, "vxlt_undercitydungeon_195_0_0.vlm"},
	{0x92F3D27E7E722A4DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_1.gad"},
	{0x92F3D28CCCFF6365ULL, "zz_mer_debug6017181.cl"},
	{0x92F42F8A594843E0ULL, "vxw_bonepitdestroyed_a3_mg_0_0.vlm"},
	{0x92F46D007FBB00BFULL, "vxht_mansion2_1gv_0_0.vlm"},
	{0x92F56E203175F1EFULL, "497659906.ogg"},
	{0x92F668D435921341ULL, "6160686_m.fxe"},
	{0x92F6CF48FA536CACULL, "bookfront.gfx"},
	{0x92F70BB73761FE0CULL, "161686819.ogg"},
	{0x92F72D3AF5D790A4ULL, "6091684_m.fxe"},
	{0x92F746FEE31D5A5BULL, "art\\characters\\weapons\\mauls\\w_mal_ml102a.msh"},
	{0x92F75EA0553C4D86ULL, "hlac_w_bonepit_0_9y.dds"},
	{0x92F764FEE31D8D45ULL, "art\\characters\\weapons\\mauls\\w_mal_ml102a.mmh"},
	{0x92F7F6B71588A8FAULL, "6013406_m.fxe"},
	{0x92F824CF06736D7EULL, "6081988_m.fxe"},
	{0x92F827B1DA2A4ED7ULL, "hf_vis_hightown2.mor"},
	{0x92F8725E14315D07ULL, "552573176.ogg"},
	{0x92F893EF36A9A914ULL, "6090809_m.fxe"},
	{0x92F91D260491FCB8ULL, "vxlt_gallowscourtyard_n_79o_0_0.vlm"},
	{0x92F9B0334BCAD83FULL, "vxlt_0_lowtown_docks_a3_n_bv7_0_0.vlm"},
	{0x92F9ECC3C80F6F86ULL, "hlac_lt_gallowscourtyard_n_131071_0z.dds"},
	{0x92FA95637FD50CC6ULL, "llac_lt_warehouse_0_3.dds"},
	{0x92FABDB3C1D5414EULL, "810119321.ogg"},
	{0x92FB1FD6DC904BF0ULL, "vxw_sundermount_2dk_0_0.vlm"},
	{0x92FB4B62FC3AF473ULL, "6048133_m.fxe"},
	{0x92FB5DF8A31BB4E7ULL, "dae000ip_crowd_mf8_drunk1.utp"},
	{0x92FB7D899F817BB4ULL, "weapon_distortionmask_p0.psh.fxo"},
	{0x92FBAB7688897529ULL, "vxlt_gallowstemplar_n_3gb_0_0.vlm"},
	{0x92FBEC8A9854B9F6ULL, "follower_banter6155583.cl"},
	{0x92FC44EDA69A240FULL, "zz_prmpt_debug6198148.cl"},
	{0x92FC515A3E4FCF3CULL, "43856468.ogg"},
	{0x92FCE57B68A8D8E2ULL, "vxdr_0_deeproads_9gl_0_0.vlm"},
	{0x92FD30D6BE790D5FULL, "zz_arishok_shoulder_axe.cut"},
	{0x92FD5FA3AAFA6EFDULL, "vxht_mansion2b_2e1_0_0.vlm"},
	{0x92FD6919B30CCE2FULL, "lt_0_lowtown_docks_n_179_0.mao"},
	{0x92FD7519B30CE2CCULL, "lt_0_lowtown_docks_n_179_0.mmh"},
	{0x92FD7B19B30CECBEULL, "lt_0_lowtown_docks_n_179_0.msh"},
	{0x92FD83EDABF186ECULL, "hlac_ht_keep_n_4519_7x.dds"},
	{0x92FDA41A447415CFULL, "art\\toolset\\tools_emitter.mao"},
	{0x92FDB01A44742A6CULL, "art\\toolset\\tools_emitter.mmh"},
	{0x92FDB61A4474345EULL, "art\\toolset\\tools_emitter.msh"},
	{0x92FDF1D2432F9D3DULL, "vxht_chantry_n_20m_0_0.vlm"},
	{0x92FE11C90CF27E56ULL, "cth_cm3.tnt"},
	{0x92FE5B0302FF4A52ULL, "6069858_m.fxe"},
	{0x92FF1A21115A30DDULL, "6150322_m.fxe"},
	{0x92FF531A049A7D97ULL, "vxca_caverns2_8ki_0_1.vlm"},
	{0x92FFA54EB9D267C4ULL, "324861057.ogg"},
	{0x9300204B85876B5CULL, "weapon2_ds2y.psh.fxo"},
	{0x9300488C970AC7B6ULL, "ht_0_hightown_n_18_0.rml"},
	{0x930080B86F6BE4FBULL, "vxht_0_hightown_a3_c3l_0_1.vlm"},
	{0x9301BAE0966FB0C1ULL, "6154991_m.fxe"},
	{0x93025DB1DE689E57ULL, "hlac_ht_0_hightown_a2_65535_0x.dds"},
	{0x930316B67517F98DULL, "l_dae221ar_keep.bnk"},
	{0x93033581227E1DDCULL, "store_one100cr_ilen.utm"},
	{0x9303488702DCC751ULL, "vxw_woundedcoast_1cx_0_0.vlm"},
	{0x93036F785E9374DDULL, "gui.wbd"},
	{0x9303A68241BD244AULL, "dre000_hireling16038596.cl"},
	{0x9303D7F57C26E4B2ULL, "follower_banter6163762.cl"},
	{0x9303DB18E9E7B59AULL, "mag121_idunna6084227.cl"},
	{0x930430C0E51CF695ULL, "6137555_f.fxe"},
	{0x9304638623367A71ULL, "vxw_encounter2_camp_9f_0_0.vlm"},
	{0x9304DAB127882BC1ULL, "6165986_m.fxe"},
	{0x9304DFA44953769FULL, "ss_wilmod.bnk"},
	{0x930516AD9CCB3512ULL, "6016161_f.fxe"},
	{0x93052571D4341A7EULL, "129070847.ogg"},
	{0x930526BCE206BB5FULL, "760260183.ogg"},
	{0x9305C6A9B065FD1CULL, "26993834.ogg"},
	{0x93060A433AD09E78ULL, "vxlt_darktown_2xp_0_0.vlm"},
	{0x930647D811AF912CULL, "hlac_w_encounter2_262143_0z.dds"},
	{0x93065A7804B5A6D6ULL, "172331607.ogg"},
	{0x93065AED5F8DF6C5ULL, "vxht_0_hightown_n_c70_0_1.vlm"},
	{0x93066203EFEA7F7CULL, "hlac_lt_0_lowtown_docks_n_-65538_3x.dds"},
	{0x93074D43BDDB24FBULL, "hlac_lt_gallowscourtyard_-1_3z.dds"},
	{0x93077FF5ACAA2619ULL, "87055194.ogg"},
	{0x9307881A95E8FFB7ULL, "dae310ar_docks.are"},
	{0x9307A0C8115998C9ULL, "llac_ht_0_hightown_-65534_0.dds"},
	{0x9307D94454A2A388ULL, "vxlt_gallowstemplar_2r6_0_0.vlm"},
	{0x93084DF04652C139ULL, "6103054_m.fxe"},
	{0x93085B8F6F94BF5DULL, "hlac_lt_foundry_n_320_14y.dds"},
	{0x9308716EC7372175ULL, "lt_0_lowtown_12_0.rml"},
	{0x9308C0B26446A5A5ULL, "and110cr_street_archer.utc"},
	{0x9308C5BC4F3C9F93ULL, "follower_banter6173180.cl"},
	{0x9308DB1BB0668DC4ULL, "gen00fl_merrill6108476.cl"},
	{0x9308DCC0044522FCULL, "834235241.ogg"},
	{0x9308E77C546DF79EULL, "mer300cr_high_dragon.utc"},
	{0x93098518C07D4934ULL, "hlac_dr_primevaltomb_664_4z.dds"},
	{0x930991B3CE866470ULL, "vxw_sundermount_2ih_0_0.vlm"},
	{0x9309BD3FEE310CF3ULL, "6043033_m.fxe"},
	{0x930A6771534811F9ULL, "44289671.ogg"},
	{0x930AE8D23A4FE57AULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_h02_0t.dds"},
	{0x930B4027205A3F6AULL, "vxht_0_hightown_a2_a7k_0_0.vlm"},
	{0x930C5AEACCFEEC79ULL, "dae000cr_pickpocket_victim.utc"},
	{0x930C90849EA36CBCULL, "vxlt_gallowscourtyard_6zx_0_0.vlm"},
	{0x930D0CABDB227F8DULL, "vxlt_0_lowtown_docks_a3_bsi_0_0.vlm"},
	{0x930D19B78E5B8358ULL, "dae000_pc_home_ambients6208596.cl"},
	{0x930D4442DC1F9622ULL, "and211ip_store_lirene.utp"},
	{0x930D8B3CE1B2C597ULL, "178820401.ogg"},
	{0x930DDF166E6B9459ULL, "dae000_orana6149036.cl"},
	{0x930E8F7F89D0FB32ULL, "vxw_bonepitmining_a2_4x_0_0.vlm"},
	{0x930F37E0E7645E45ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_sitting_table"},
	{0x930F9690B8FB5E4CULL, "302775677.ogg"},
	{0x930F9FCAD2DC6D50ULL, "841027851.ogg"},
	{0x9310064B95CCBB35ULL, "hlac_w_bonepitdestroyed_a3_65541_0z.dds"},
	{0x931010381002C366ULL, "vxca_caverns2_9fc_0_0.vlm"},
	{0x93105DEADCAA7FCEULL, "llac_ca_cavern1_haunted_24423_10.dds"},
	{0x9310993E766EA067ULL, "qcr220cr_nobleman2.utc"},
	{0x9310BF4EF371B487ULL, "1025939868.ogg"},
	{0x931150BA63A731BFULL, "vxht_mansion2_1mp_0_0.vlm"},
	{0x931167132F89E444ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track.msh"},
	{0x931179132F8A02ADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track.mao"},
	{0x93117D132F8A09E6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track.mmh"},
	{0x931241C179904761ULL, "1068628194.ogg"},
	{0x93127CC69B226201ULL, "vxdr_primevaltomb_82c_0_0.vlm"},
	{0x9312D563152C92F1ULL, "vxw_encounter2_camp_4e_0_0.vlm"},
	{0x9312DD2D0EFF20E8ULL, "vxdr_slavers_1bc_0_0.vlm"},
	{0x9312F9F511EA3094ULL, "w_sundermount_170_0.phy"},
	{0x931345FA8E3B901BULL, "vxca_cavern1_haunted_cjv_0_0.vlm"},
	{0x93135BB891A6AF1EULL, "vxdr_primevaltomb_7kx_0_0.vlm"},
	{0x931449929B3520F0ULL, "6162551_m.fxe"},
	{0x9314BE8DCFA0C685ULL, "vxlt_0_lowtown_docks_a3_d6d_0_0.vlm"},
	{0x93154317565275E0ULL, "mag110_vincento6167356.cl"},
	{0x93155B50D6DF88CBULL, "vxht_0_hightown_a3_c17_0_0.vlm"},
	{0x9315C01B00D1A8D9ULL, "vxw_bonepitdestroyed_a3_2a6_0_0.vlm"},
	{0x9315E974626B8E2DULL, "363784099.ogg"},
	{0x931617989B33BDD5ULL, "vxht_brothel_n_3fa_0_0.vlm"},
	{0x9316B329A161B13DULL, "6058282_m.fxe"},
	{0x93172AE292EE9C9BULL, "dae100_alistair_drunk6135147.cl"},
	{0x93177E1AD34C1F05ULL, "vxlt_darktown_2rk_0_0.vlm"},
	{0x9317C51531C73251ULL, "gen_im_trs_tmp_shd_met_r1.uti"},
	{0x9317E06E7F51935EULL, "ht_0_hightown_a2_11_0.rml"},
	{0x93180390C07C7CF0ULL, "vxw_sundermount_2ri_0_0.vlm"},
	{0x93196B74F181A597ULL, "vxlt_gallowscourtyard_4v9_0_0.vlm"},
	{0x9319828EF9321F93ULL, "art\\vfx\\combat\\vfx_glb_ele_c_minor_c.mmh"},
	{0x9319853FE65ACD88ULL, "6208639_m.fxe"},
	{0x93199F8318C90115ULL, "1056593998.ogg"},
	{0x9319F7C5D8EBCAFAULL, "dae000_gen_turn_in_36178447.cl"},
	{0x931A2A1A840E2E20ULL, "ca_cavern1_haunted_24424.rml"},
	{0x931A5C8D6ACFB037ULL, "6092031_m.fxe"},
	{0x931A5E138D82EA1EULL, "673257963.ogg"},
	{0x931A7AA088A686ECULL, "ht_0_hightownexplosion_n_33_0_ll.mmh"},
	{0x931A885BB13C95D2ULL, "vxca_cave_23x_0_0.vlm"},
	{0x931AC28A52B7CD99ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.impa.ani"},
	{0x931ADB1927857E01ULL, "hlac_dr_dungeon_3207_3z.dds"},
	{0x931B5599A5E6DB1AULL, "hlac_lt_backalley_n_0_3x.dds"},
	{0x931B74132F924EB1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track.phy"},
	{0x931B77CFD61C6198ULL, "6083799_m.fxe"},
	{0x931C46B6B7FF185AULL, "follower_banter6155499.cl"},
	{0x931CC872710D1C37ULL, "vxlt_hangedman_n_27t_0_0.vlm"},
	{0x931CD6F511F24991ULL, "w_sundermount_170_0.msh"},
	{0x931CE4F511F261ACULL, "w_sundermount_170_0.mao"},
	{0x931CF0F511F2764FULL, "w_sundermount_170_0.mmh"},
	{0x931D7D54E33E61F6ULL, "isa311_isabela6092265.cl"},
	{0x931D8B75AB9B867FULL, "hlac_lt_0_lowtown_a2_n_-65537_4x.dds"},
	{0x931DAE01EE8FBFD1ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2_lp.gad"},
	{0x931DC450EC4EE835ULL, "darktown_fem_fer_urch_sta6174796.cl"},
	{0x931DE5256C9CD0ADULL, "mag101_alain6032263.cl"},
	{0x931E0D2A85DEABB5ULL, "361267130.ogg"},
	{0x931E595B38F4064EULL, "llac_lt_0_lowtown_docks_-3_0.dds"},
	{0x931E5ED4A7C7F481ULL, "zz_charge_hightown.cut"},
	{0x931EC7075795554AULL, "430376106.ogg"},
	{0x931F2A24E935040CULL, "vxlt_warehouse_sf_0_0.vlm"},
	{0x931F93075FEB3C01ULL, "lgt260_eau_tmp_lyrium_mess6143644.cl"},
	{0x931FBBDB30B9FA21ULL, "530426364.ogg"},
	{0x931FE2CB3C1A9B07ULL, "6105951_m.fxe"},
	{0x931FE558FE28A318ULL, "mcr310_cut_chantry_blows_up6087545.cl"},
	{0x93201314E14CA8BCULL, "ht_0_hightown_a2_37_0_ll.mmh"},
	{0x9321357525FA5287ULL, "206501793.ogg"},
	{0x93216A85AAE45A6AULL, "661366961.ogg"},
	{0x93219D8A52BDADDBULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.impa.gad"},
	{0x9321B7D78031A89BULL, "vxca_cavern1_haunted_coi_0_0.vlm"},
	{0x9321C59D40DE9832ULL, "hlac_ht_chantry_2534_17y.dds"},
	{0x9321D2234FB2949AULL, "hlac_w_bonepitdestroyed_a3_2_0x.dds"},
	{0x9321DA5E59DD09A0ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_a_c.mmh"},
	{0x932256E99AF4CD33ULL, "vxca_cavern1_haunted_eae_0_0.vlm"},
	{0x9322C6F672220637ULL, "hm_templar01.mor"},
	{0x932350AA614C97B1ULL, "vxlt_0_lowtown_docks_a3_cyj_0_0.vlm"},
	{0x9323910B41C19685ULL, "6218662_m.fxe"},
	{0x9323AF01EE942D7FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_die_v2_lp.ani"},
	{0x9323D84727E08CD8ULL, "vxca_caverns1_mines_gko_0_1.vlm"},
	{0x93240120EFB4F085ULL, "333916831.ogg"},
	{0x932525069357C9BDULL, "6055582_m.fxe"},
	{0x9325EFF7C5423785ULL, "vxlt_darktown_2uj_0_0.vlm"},
	{0x93260D3AFF396EF3ULL, "mcr370_cut_player_rules6171823.cl"},
	{0x9326192D24ABE116ULL, "mer340_bonepit_comments6090841.cl"},
	{0x93263A71DABD5D96ULL, "mag100_thrask6032070.cl"},
	{0x93267524EECC2E54ULL, "6218634_m.fxe"},
	{0x9326E0AD2879BFF5ULL, "dae000_oldmangossip6174978.cl"},
	{0x9326FA6BA10659B4ULL, "art\\cinematics\\follower plots\\sebastian\\cs_letter2_sebastianintro.gad"},
	{0x932716297192043EULL, "445795093.ogg"},
	{0x932720517D88AD08ULL, "hlac_ht_mansion_n_2995_3x.dds"},
	{0x932721189EA9289EULL, "126322653.ogg"},
	{0x9327696BE6B2C3DEULL, "lt_0_lowtown_docks_28_0_lc.mmh"},
	{0x932791EF8625952FULL, "456558022.ogg"},
	{0x93282AF153B7B555ULL, "wm_ico_road.dds"},
	{0x932837B23A587CFCULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medc_0t.dds"},
	{0x93287F8E6629E919ULL, "6072310_m.fxe"},
	{0x9328BA8A52C3FDC0ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.impa.evt"},
	{0x9328BB8D371D4CFCULL, "vxht_0_hightown_a6j_0_0.vlm"},
	{0x93290D5A93D6232CULL, "845691694.ogg"},
	{0x93294E9265A75B1BULL, "vxht_0_hightown_a2_cbc_0_0.vlm"},
	{0x93296341EB2068E1ULL, "6085434_f.fxe"},
	{0x932A115654389471ULL, "vxca_caverns1_j6d_0_0.vlm"},
	{0x932A1E545F4C973BULL, "vxht_keep_3x1_0_0.vlm"},
	{0x932A2D0B9E736402ULL, "vxlt_ruinedhouse_n_5y_0_0.vlm"},
	{0x932A434077DE2770ULL, "6088862_m.fxe"},
	{0x932ABA83A81C3B88ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookinquisb_n.dds"},
	{0x932ABDC5DD83AC02ULL, "vxca_cavern1_haunted_e8k_0_0.vlm"},
	{0x932ABE62BF76EA81ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_wep_sw.mao"},
	{0x932ABF2D03B87F18ULL, "plt_cod_itm_girdle_elders.plo"},
	{0x932B38313CF051ACULL, "gen00fl_carver6108944.cl"},
	{0x932B9D382375887BULL, "vxca_caverns1_mines_ikc_0_0.vlm"},
	{0x932BCC7BA01AD8F1ULL, "6122046_m.fxe"},
	{0x932CDED3BE752A99ULL, "716532885.ogg"},
	{0x932D096BA10ADE92ULL, "art\\cinematics\\follower plots\\sebastian\\cs_letter2_sebastianintro.ani"},
	{0x932D6586A925D75CULL, "vxca_cavern1_haunted_i01_0_0.vlm"},
	{0x932DAAA3FB7B1482ULL, "vxca_cavern1_haunted_hw3_0_0.vlm"},
	{0x932DB0C8ECD04E15ULL, "merev_item_acquired.ncs"},
	{0x932E41AC23FA9F33ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_hunt_cap_fenrintro.ani"},
	{0x932EE6E594F03927ULL, "water_water_p0.psh.fxo.dx11.assembly"},
	{0x932F47EAEE8D2822ULL, "skel_si.dds"},
	{0x932F5A216045AECAULL, "follower_barks6153591.cl"},
	{0x932F7CBA0A7ED1E9ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb1a.phy"},
	{0x932F81CD94399FF2ULL, "6046541_m.fxe"},
	{0x932FBD5E59E915BFULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_a_c.anb"},
	{0x932FD96F6DB555C3ULL, "hlac_lt_0_lowtown_a2_-1_2x.dds"},
	{0x93300CB6753ED6B7ULL, "l_dae221ar_keep.wbd"},
	{0x933038836BBDDEA7ULL, "vfx_vfx_p2.psh.fxo"},
	{0x933078F8CB9D3E3DULL, "785604625.ogg"},
	{0x9331829301D5C681ULL, "vxw_bonepitdestroyed_a3_1rd_0_0.vlm"},
	{0x9331C7E99046E633ULL, "6060610_m.fxe"},
	{0x9332449FB79E0884ULL, "6053923_m.fxe"},
	{0x93327D79FFD09C5CULL, "art\\characters\\playercharacter\\humanmale\\arm_urva_1.mao"},
	{0x933298CFB34751F2ULL, "vxlt_0_lowtown_a2_n_c_hkg_0_0.vlm"},
	{0x9332B171F635E67EULL, "240399111.ogg"},
	{0x9332FB527F1FEED5ULL, "vxht_brothel_n_3ho_0_0.vlm"},
	{0x93332AF53CAE484BULL, "vxlt_hangedman_n_4u_0_0.vlm"},
	{0x93337BB02CCC59DFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3open.msh"},
	{0x933381B02CCC6381ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3open.mmh"},
	{0x93338DB02CCC7822ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt3open.mao"},
	{0x9333CA47D909E7E2ULL, "one300_nuncio6075163.cl"},
	{0x9334327C1FBD97B9ULL, "989152221.ogg"},
	{0x933461D4B7385005ULL, "vxlt_darktown_2xi_0_0.vlm"},
	{0x933485B1EE1D436CULL, "6034032_m.fxe"},
	{0x933500A4CD35CB52ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_die_lp_v1.gad"},
	{0x933572EA5061D729ULL, "vxlt_gallowstemplar_n_3sf_0_0.vlm"},
	{0x933589C8867A8F49ULL, "6152129_f.fxe"},
	{0x9335E4E1A2AD8956ULL, "6111227_m.fxe"},
	{0x9336645031BF6BC4ULL, "llac_lt_0_lowtown_a2_n_c_-131073_0.dds"},
	{0x9336AE76D4B55980ULL, "6071862_m.fxe"},
	{0x9336D59D673F517EULL, "gen00fl_sebastian6179770.cl"},
	{0x933750D607DFBA77ULL, "lgt110_gxa_coa_spy6117519.cl"},
	{0x9337D4F5F85B88AAULL, "6076454_m.fxe"},
	{0x93380FF2F186FC0AULL, "vxdr_0_deeproads_9c8_0_0.vlm"},
	{0x933863F513CEC69FULL, "6167264_f.fxe"},
	{0x933890315142AFBBULL, "vxht_keep_3g0_0_0.vlm"},
	{0x933911A384E2556AULL, "vxlt_gallowstemplar_3nu_0_0.vlm"},
	{0x93395D4C57A88239ULL, "follower_banter6169453.cl"},
	{0x933980F01A8FB7DBULL, "vxht_keep_61c_0_0.vlm"},
	{0x9339B6D492195D1AULL, "6173848_f.fxe"},
	{0x933A458C973C3B6AULL, "ht_0_hightown_n_18_0.anb"},
	{0x933A85FE1EA8B6ACULL, "229884881.ogg"},
	{0x933AB567AAC20EF7ULL, "gen00fl_aveline6140137.cl"},
	{0x933AB58ED994877AULL, "hlac_lt_0_lowtown_a2_0_7z.dds"},
	{0x933BA7CE85A38981ULL, "299594649.ogg"},
	{0x933C0DDECD21350CULL, "vxlt_warehouse_yd_0_0.vlm"},
	{0x933C248D04599C68ULL, "mer240_jansen.cnv"},
	{0x933CA76C174DEE63ULL, "6180909_m.fxe"},
	{0x933CBD9426C47824ULL, "6067630_m.fxe"},
	{0x933CD650F872957FULL, "vxlt_hangedman_n_10n_0_0.vlm"},
	{0x933D49B2DBFC675DULL, "gen00fl_anders6196690.cl"},
	{0x933DF0CED179B4C4ULL, "6179502_m.fxe"},
	{0x933EA6E466573C26ULL, "627186501.ogg"},
	{0x933F3DFEE35B00F2ULL, "art\\characters\\weapons\\mauls\\w_mal_ml102a.phy"},
	{0x933F468D045C0BD0ULL, "mer240_jansen.bnk"},
	{0x93405B9DBFBE55EDULL, "lgt330ip_shw_mer_to_sewer.utp"},
	{0x93408F1855FCF82EULL, "162030844.ogg"},
	{0x9340E7157D52BCBBULL, "267635653.ogg"},
	{0x9341618397050678ULL, "vxlt_undercityencounter_1fo_0_0.vlm"},
	{0x9341BFE34822E1CBULL, "6109007_m.fxe"},
	{0x9342079DFB78D761ULL, "llac_lt_0_lowtown_a2_-1_8.dds"},
	{0x9342A0B9E85D7795ULL, "46266551.ogg"},
	{0x934310FD57C38929ULL, "6084543_f.fxe"},
	{0x934318AC240CB5DDULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_hunt_cap_fenrintro.gad"},
	{0x934372357396A7C1ULL, "zz_fol_debug6096558.cl"},
	{0x93453D13313C1557ULL, "gen_im_trs_gem_rosediam_03.uti"},
	{0x9345CC86ECB0550BULL, "6138821_m.fxe"},
	{0x9345D6F7E0609E54ULL, "vxca_cave_2bm_0_0.vlm"},
	{0x93463753337EB58BULL, "vxlt_warehouse_n_15i_0_0.vlm"},
	{0x934681CFE37D148AULL, "vxdr_0_deeproads_9d9_0_0.vlm"},
	{0x9346A7067E5B0BE8ULL, "zz_one_debug6051984.cl"},
	{0x9347937142DF400EULL, "llac_lt_0_lowtown_docks_-131076_1.dds"},
	{0x9347CE4B4AC3D5B3ULL, "dae000_elegant6166900.cl"},
	{0x934835E324F5C695ULL, "gencr_gate_slave_hanging.utc"},
	{0x934842019AB3A790ULL, "gen00fl_varric6081201.cl"},
	{0x934869A516575B2EULL, "clf260_harley6088764.cl"},
	{0x9348EBA4CD4650F8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_die_lp_v1.ani"},
	{0x9349576594A92101ULL, "hlac_w_encounter2_camp_196606_0y.dds"},
	{0x9349CE562B7F58E6ULL, "vxht_0_hightown_aok_0_0.vlm"},
	{0x934A28AF31EFD3AAULL, "weapon1_s2.psh.fxo"},
	{0x934AA8D39FB08765ULL, "gen00fl_aveline6122746.cl"},
	{0x934AB130C9F93637ULL, "321900505.ogg"},
	{0x934C43F817D90472ULL, "hf_tbe_barber.mor"},
	{0x934CBC19A805367CULL, "237286853.ogg"},
	{0x934CD71E7414351DULL, "792432424.ogg"},
	{0x934CEB9FA85B6ADDULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypouchb.mao"},
	{0x934CEF9FA85B7116ULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypouchb.mmh"},
	{0x934CF99FA85B8214ULL, "art\\levels\\4cinematics\\ferelden\\prp_moneypouchb.msh"},
	{0x934D657C45E2D759ULL, "vxdr_primevaltomb_9e7_0_0.vlm"},
	{0x934D6FD0787C6E09ULL, "vxlt_warehouse_8r_0_0.vlm"},
	{0x934E2D372F0003F1ULL, "524173708.ogg"},
	{0x934E6D882235DD88ULL, "281128815.ogg"},
	{0x934EFD4F2B8A166FULL, "gen00fl_anders6113040.cl"},
	{0x934F2EA7EA3960AAULL, "227039056.ogg"},
	{0x934F66E7D618B56EULL, "gen00fl_aveline6141090.cl"},
	{0x934FC6D03ED07BEDULL, "vxlt_gallowscourtyard_742_0_0.vlm"},
	{0x935059B31E97E6F5ULL, "vxlt_0_lowtown_a2_n_c_fgj_0_0.vlm"},
	{0x9350DE5BBC791EE3ULL, "6143319_m.fxe"},
	{0x935122137EC2C163ULL, "dae000_gamlen6048473.cl"},
	{0x935158BD8EDA7E8EULL, "gen00fl_anders6110121.cl"},
	{0x93516D69B5E104C6ULL, "hlac_w_bonepitdestroyed_a3_458755_0z.dds"},
	{0x93519ED024A0593DULL, "gen00fl_aveline6132871.cl"},
	{0x9351A8F59FCB6A39ULL, "vxw_encounter2_camp_no_0_0.vlm"},
	{0x93521016CA8E531AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lowdoor_n.dds"},
	{0x9352A84B44F282B1ULL, "tow_ex01.mao"},
	{0x9353A59D087984B4ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax103a_0s.dds"},
	{0x9353B88466906C8AULL, "vxlt_0_lowtown_n_fe9_0_0.vlm"},
	{0x9353F27C119E61FFULL, "hf_amb_templar_recruit.mor"},
	{0x93541D966D11F4AAULL, "anim_base.gda"},
	{0x9354730CC4184B71ULL, "vxlt_hangedman_15t_0_0.vlm"},
	{0x93547AB742E67E47ULL, "1071897666.ogg"},
	{0x9355226CB7EAC5B7ULL, "vxht_bartrand_n_4nx_0_0.vlm"},
	{0x93552FC191630C5CULL, "vxdr_primevaltomb_8o5_0_0.vlm"},
	{0x93553C99B9AA7CBFULL, "vxht_bartrand_n_5p3_0_0.vlm"},
	{0x9355A74F3D184451ULL, "one300_zevran6076237.cl"},
	{0x9355EA16EA562CB4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03a.phy"},
	{0x9356135CB365F582ULL, "vxca_cavern1_haunted_e7v_0_0.vlm"},
	{0x935622A73DFA2AD0ULL, "llac_lt_0_lowtown_a3_n_c_-3_3.dds"},
	{0x93568C6284AD4155ULL, "6027781_m.fxe"},
	{0x93568EFDD9E86AF7ULL, "vxlt_gallowstemplar_n_6su_0_0.vlm"},
	{0x9356F813CB0D0C46ULL, "vxht_0_hightown_a3_c8i_0_0.vlm"},
	{0x9357310D52638002ULL, "dre000_varric6034671.cl"},
	{0x935782FAE9D57444ULL, "hlac_w_encounter1_65534_0x.dds"},
	{0x9358114EA3F2BEFEULL, "isa311_party_comments6117814.cl"},
	{0x935818D7C719CC3AULL, "hlac_lt_undercityencounter_-1_46z.dds"},
	{0x9358BAE0CED89535ULL, "gen00fl_varric6104941.cl"},
	{0x9358C7721C0660BAULL, "vxw_encounter2_camp_2t_0_0.vlm"},
	{0x935A576DBEDCC0AEULL, "rdr111dg_harbormaster6020388.cl"},
	{0x935AA9B823E18F25ULL, "vxw_bonepitdestroyed_a3_1vd_0_0.vlm"},
	{0x935AC7F690B80BB6ULL, "616382815.ogg"},
	{0x935AD7F3B825FCEBULL, "vxca_caverns1_efu_0_0.vlm"},
	{0x935B16B63CC3D6A0ULL, "gen00fl_bethany6102764.cl"},
	{0x935B532EE8F90B9AULL, "art\\characters\\creatures\\hurlock\\textures\\w_qui_hlk_0s.dds"},
	{0x935B65BA0AA3E50EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb1a.mmh"},
	{0x935B76620B5A6AE6ULL, "6149572_m.fxe"},
	{0x935B7FBA0AA411FCULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb1a.msh"},
	{0x935B8238D86B24F6ULL, "vxdr_primevaltomb_90z_0_1.vlm"},
	{0x935BC86408D513D2ULL, "285264870.ogg"},
	{0x935CE78884BF94EEULL, "hlac_lt_gallowstemplar_f_2394_12x.dds"},
	{0x935CF00ED76C1E41ULL, "gen_im_trs_prf_rng_r1.uti"},
	{0x935D4A5715D6EF6EULL, "llac_lt_backalley_n_65537_7.dds"},
	{0x935DF072DA9548E8ULL, "1043039129.ogg"},
	{0x935E38AD30C6946DULL, "vxlt_gallowscourtyard_791_0_0.vlm"},
	{0x935E7F45306A755CULL, "vxlt_0_lowtown_gcg_0_0.vlm"},
	{0x935EADB98B5A61E0ULL, "vxlt_0_lowtown_docks_a3_n_d0e_0_1.vlm"},
	{0x935EF8A1641D34DEULL, "148483334.ogg"},
	{0x935FC716EA5E45B1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03a.msh"},
	{0x935FDFC73E86A6F5ULL, "843876797.ogg"},
	{0x935FE116EA5E726FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec03a.mmh"},
	{0x93601AD291C182B9ULL, "vxw_encounter2_camp_in_0_0.vlm"},
	{0x936034C792353931ULL, "one300_zevran6089099.cl"},
	{0x9360517879263547ULL, "qun230cr_guild_contact6068853.cl"},
	{0x9360562FE3CF5BEEULL, "vxdr_primevaltomb_6io_0_0.vlm"},
	{0x9361473821769F85ULL, "vxlt_undercityencounter_1iw_0_0.vlm"},
	{0x9361BE994AEFA680ULL, "vxlt_gallowstemplar_f_4sq_0_1.vlm"},
	{0x93620B887E02B3EDULL, "6116636_m.fxe"},
	{0x9362828EED05137AULL, "art\\levels\\4cinematics\\ferelden\\bookinquisb.mmh"},
	{0x9362948EED0531C0ULL, "art\\levels\\4cinematics\\ferelden\\bookinquisb.msh"},
	{0x9362B1A4F7C93AC7ULL, "415366707.ogg"},
	{0x9362E4350C516DCAULL, "vxlt_hangedman_n_1m0_0_0.vlm"},
	{0x9362F725CE9D0F12ULL, "vxht_bartrand_n_6rw_0_0.vlm"},
	{0x9363336698D8DEC4ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_0z.dds"},
	{0x936346630A92B896ULL, "6034305_f.fxe"},
	{0x93635A322D168588ULL, "442006769.ogg"},
	{0x9363D9E94B9E7BC0ULL, "vxlt_darktown_42l_0_0.vlm"},
	{0x9363E695180C6F18ULL, "dae000_porfiria6155291.cl"},
	{0x9364B8F396010890ULL, "6127682_m.fxe"},
	{0x9364BC9FE5817D4CULL, "6126923_f.fxe"},
	{0x9364FD9D70A99CEFULL, "vxlt_undercityencounter_142_0_0.vlm"},
	{0x93669D10C0736CECULL, "face_rendertodepth_p4.psh.fxo.dx11"},
	{0x936734AB30D77008ULL, "vxht_0_hightown_byl_0_0.vlm"},
	{0x936767286AB938D2ULL, "dre000_sandal6033759.cl"},
	{0x936776A7D48BADE6ULL, "lgt311_shw_mer_mettin6156455.cl"},
	{0x93678C02E4CC7F6DULL, "zz_followers_banter6159636.cl"},
	{0x9367BC277F7D7454ULL, "6169297_m.fxe"},
	{0x936826A623084420ULL, "351588312.ogg"},
	{0x9368366B1770A51DULL, "hlac_lt_0_lowtown_a3_n_c_65533_0z.dds"},
	{0x9368F70CD970A2E8ULL, "vxw_bonepit_29s_0_0.vlm"},
	{0x936A8D6EC0DA0AA4ULL, "vxht_0_hightown_a3_a2o_0_0.vlm"},
	{0x936AC4D501473572ULL, "6148096_m.fxe"},
	{0x936AC88213E870D7ULL, "hlac_lt_0_lowtown_docks_n_-196609_0x.dds"},
	{0x936B112CD20AE978ULL, "775813447.ogg"},
	{0x936B4D3C34A959CEULL, "vxdr_primevaltomb_4ho_0_1.vlm"},
	{0x936B5D6FB53E5032ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_bld_flma_0n.dds"},
	{0x936BA0E6558B6669ULL, "6106322_f.fxe"},
	{0x936BC5D30DD5653DULL, "seb221_ruxton_harimann.bnk"},
	{0x936C4D44BCDF3131ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_loop.ani"},
	{0x936CC45C9D45D75EULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc106a_0d.dds"},
	{0x936D48B062057082ULL, "6129365_m.fxe"},
	{0x936D94AC4FCBBB2BULL, "fxa_unholy_dusts.dds"},
	{0x936DE82A034A442EULL, "gen00fl_fenris6126913.cl"},
	{0x936E02F3765A89EBULL, "art\\vfx\\combat\\vfx_q_ari_buf_act_c.impa.ani"},
	{0x936E4106F9D7E875ULL, "vxw_woundedcoast_2b_0_0.vlm"},
	{0x936E5DBB082C4A84ULL, "322106240.ogg"},
	{0x936EA7D30DD76845ULL, "seb221_ruxton_harimann.cnv"},
	{0x936EF810944E714AULL, "906839328.ogg"},
	{0x936F6192D664357BULL, "448333358.ogg"},
	{0x936F8490BC8D31BDULL, "vxlt_0_lowtown_ex5_0_0.vlm"},
	{0x936FA0AD3FC2BDEAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\runforward.gad"},
	{0x937003655B12623AULL, "vxca_caverns1_j4u_0_0.vlm"},
	{0x9370AE942AC9C86CULL, "vrc321ip_spooky_7_owner.utp"},
	{0x9371085738F1E123ULL, "6144374.ogg"},
	{0x9371492E9A649E59ULL, "zz_one_debug6079318.cl"},
	{0x93724BC63D949440ULL, "vxlt_darktown_49m_0_0.vlm"},
	{0x9372BE466F04D4ECULL, "vxht_bartrand_n_5o1_0_0.vlm"},
	{0x9372CC737AEC63B3ULL, "6045630_f.fxe"},
	{0x93732DFBC11190A4ULL, "6178426_m.fxe"},
	{0x93735844BCE562D3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_loop.gad"},
	{0x9373B2E5EB702595ULL, "911229496.ogg"},
	{0x9373DBCDAEF93D46ULL, "vxht_0_hightown_a3_c6k_0_0.vlm"},
	{0x93741A0260E377AFULL, "vxlt_0_lowtown_a3_n_c_hpv_0_0.vlm"},
	{0x937451192F2DDDB5ULL, "vxht_mansion2_3c6_0_0.vlm"},
	{0x9374575C91BEDEC5ULL, "907732616.ogg"},
	{0x93745D7412FBD323ULL, "6117252_f.fxe"},
	{0x9374857C7F036E42ULL, "vxlt_0_lowtown_a3_n_c_csi_0_0.vlm"},
	{0x937520D9A0D9B45BULL, "hlac_ht_keep_n_4519_11x.dds"},
	{0x937536BE07933ABCULL, "170486943.ogg"},
	{0x93754FC4DCA74A9BULL, "6100314_m.fxe"},
	{0x93755FD898AF50A5ULL, "llac_ht_0_hightown_a2_-65536_0.dds"},
	{0x9375823AA65387A1ULL, "100439750.ogg"},
	{0x9376698B1E71C0C6ULL, "gen00fl_dog6122020.cl"},
	{0x93767F5B1629645CULL, "6044372_m.fxe"},
	{0x937683AD3FC8AA81ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\runforward.evt"},
	{0x937715B26F647914ULL, "pro000tr_aveline_fight.ncs"},
	{0x93775DC797CA2F0CULL, "llac_dr_primevaltomb_658_5.dds"},
	{0x937785EF59EC5B4CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordc_d.dds"},
	{0x93778EDD485EAA8BULL, "6106356_m.fxe"},
	{0x937799D138DD0D81ULL, "vxdr_primevaltomb_8od_0_0.vlm"},
	{0x9377A650512435DEULL, "6136872_f.fxe"},
	{0x9377ADA9B08BBCD0ULL, "133988394.ogg"},
	{0x9378B04C135EB029ULL, "ser270cr_emeric.utc"},
	{0x9378BD58D2D7509EULL, "vxlt_0_lowtown_a3_n_c_gnb_0_0.vlm"},
	{0x9379008E6139BE0CULL, "vxlt_blackemporium_10f_0_0.vlm"},
	{0x937943F3E3E2E201ULL, "hlac_dr_slavers_3214_4y.dds"},
	{0x93794A5B597B33FFULL, "259879115.ogg"},
	{0x93795544BCE9C938ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\custom\\arm_around_shoulder_action_loop.evt"},
	{0x937A08B87BB32368ULL, "lgt311_shw_mer_mettin6156567.cl"},
	{0x937AAD06C157D2F9ULL, "502605498.ogg"},
	{0x937AEAF37664EB62ULL, "art\\vfx\\combat\\vfx_q_ari_buf_act_c.impa.evt"},
	{0x937B1001B005A671ULL, "fxc_puke_blend.dds"},
	{0x937B28F75EDD6878ULL, "vxlt_undercityencounter_1rk_0_0.vlm"},
	{0x937BF5E41F482396ULL, "sf_d_glyph.vsh.fxo"},
	{0x937BFA13108B3BEAULL, "dae000_male_commoner2_mob6174426.cl"},
	{0x937C1CC3C5E8BBDDULL, "vxlt_0_lowtown_a3_n_c_h2b_0_0.vlm"},
	{0x937CA044042B863EULL, "ave331_jeven6151153.cl"},
	{0x937CAB9992CD33B4ULL, "zz_vault_debug6203347.cl"},
	{0x937CF589417E0EE2ULL, "6125970_m.fxe"},
	{0x937D2A541B595694ULL, "805741678.ogg"},
	{0x937D39E9C7BB8CEEULL, "vxdr_primevaltomb_6sa_0_0.vlm"},
	{0x937DB882C0B58460ULL, "6030630_m.fxe"},
	{0x937DD1B2343F56FCULL, "fx_iceboof.dds"},
	{0x937DDC064D326B9FULL, "vxlt_gallowstemplar_4xy_0_0.vlm"},
	{0x937DF66DAE834A3DULL, "vxlt_0_lowtown_es4_0_0.vlm"},
	{0x937E3D3E7DBD9192ULL, "594672865.ogg"},
	{0x937E9E0260FC4E4BULL, "6016039_m.fxe"},
	{0x937F65EB59DBDE39ULL, "6103245_m.fxe"},
	{0x937F9990F895A854ULL, "173966596.ogg"},
	{0x937FDE9761AE1F25ULL, "ht_0_hightownexplosion_n_16_0.rml"},
	{0x937FE04038039767ULL, "6120073_m.fxe"},
	{0x937FF4C74C14C621ULL, "zz_fen_debug6215250.cl"},
	{0x93800CF7BF68D958ULL, "932511186.ogg"},
	{0x938026C7E5B2F56EULL, "vxdr_primevaltomb_529_0_0.vlm"},
	{0x93803744792A9917ULL, "vxlt_0_lowtown_a2_fh8_0_0.vlm"},
	{0x93805C70E9B3B8FCULL, "hlac_ht_mansion_n_2995_4y.dds"},
	{0x9380949D4C4DFD76ULL, "vxht_brothel_n_5kg_0_0.vlm"},
	{0x9380C653F8830D0DULL, "follower_barks6153349.cl"},
	{0x9380E954D7E47681ULL, "vxlt_0_lowtown_gc6_0_0.vlm"},
	{0x9380E9A28CB7F77CULL, "6131937_m.fxe"},
	{0x9380F3A65B809126ULL, "hlac_w_blightlands_-4_0z.dds"},
	{0x938102E5F3523033ULL, "darktown_male_fer_urch_sta6174946.cl"},
	{0x9381389608B32D17ULL, "one110_tree_minder6133735.cl"},
	{0x93819CAD79ED3990ULL, "6129882_m.fxe"},
	{0x9381CA5E83723B4EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheada_sa.dds"},
	{0x9381E9F3766B0875ULL, "art\\vfx\\combat\\vfx_q_ari_buf_act_c.impa.gad"},
	{0x9382C2F62123F635ULL, "vxht_mansion2_3f5_0_0.vlm"},
	{0x93832362C8522321ULL, "hlac_lt_gallowstemplar_f_2382_1x.dds"},
	{0x93832EB0FA110B3DULL, "84919727.ogg"},
	{0x93835A3AA128D6F4ULL, "vxw_bonepitmining_a2_202_0_0.vlm"},
	{0x9383F497845EE19DULL, "6024528_m.fxe"},
	{0x9384071C626B34EDULL, "6134767_f.fxe"},
	{0x93847BAD3FD4DAD0ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\runforward.ani"},
	{0x93848AC3AAA01B35ULL, "vxlt_gallowscourtyard_n_7h9_0_0.vlm"},
	{0x938511A8DA51D2CCULL, "rdr111dg_martin6054492.cl"},
	{0x9385171C6884A9DFULL, "vxlt_0_lowtown_f1s_0_0.vlm"},
	{0x938567862B4F0AC6ULL, "gencr_gate_guardian.utc"},
	{0x93859E1FB8DB1B38ULL, "327008.ogg"},
	{0x9385A2B8684ABD02ULL, "347592442.ogg"},
	{0x9385F852948A960EULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.cess.ani"},
	{0x9386B5BEDE23ABFDULL, "vxlt_gallowstemplar_f_2p6_0_0.vlm"},
	{0x9387221E1AB050B0ULL, "6059338_m.fxe"},
	{0x9387546D126D691BULL, "follower_banter6163909.cl"},
	{0x9388605DE150CD9DULL, "dae000_merchant_dm_mob6174306.cl"},
	{0x93888B64B836000CULL, "vxw_encounter3_5t_0_0.vlm"},
	{0x938938181F828F79ULL, "hlac_ca_caverns1_24425_6y.dds"},
	{0x9389566A1D514840ULL, "vxdr_0_deeproads_gfk_0_0.vlm"},
	{0x938964582C98FE67ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_brkbrgmst_1.mao"},
	{0x938AC96532A418A8ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumh.msh"},
	{0x938ADB6532A43739ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumh.mao"},
	{0x938AE76532A44BD2ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumh.mmh"},
	{0x938B28F70D80CFACULL, "vxlt_0_lowtown_n_eq8_0_0.vlm"},
	{0x938B29D016A102D5ULL, "gen00fl_aveline6129035.cl"},
	{0x938B342E69B88441ULL, "llac_ca_caverns2prc_5802_63.dds"},
	{0x938B75855DCAA9F9ULL, "403275844.ogg"},
	{0x938B96357EA57A10ULL, "647390801.ogg"},
	{0x938BABC6B9B1A56EULL, "vxdr_primevaltomb_6vb_0_0.vlm"},
	{0x938C2A4471920A9CULL, "209375815.ogg"},
	{0x938C684AA07962BDULL, "vxlt_0_lowtown_en3_0_0.vlm"},
	{0x938CA4A5538C6094ULL, "hlac_ht_0_hightown_a3_-65536_4x.dds"},
	{0x938CF1C14597AA37ULL, "6081660_m.fxe"},
	{0x938D1BB192881494ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp_basic.phy"},
	{0x938D262BB9BD6EF2ULL, "vxht_bartrand_n_4et_0_0.vlm"},
	{0x938D2D0CC41E9456ULL, "vxht_mansion_n_5aj_0_0.vlm"},
	{0x938D72F6B8FC4A35ULL, "hlac_lt_0_lowtown_a3_n_c_-3_5y.dds"},
	{0x938DD3D879C829A1ULL, "art\\characters\\creatures\\spider\\animation\\enter.ani"},
	{0x938DD3FE692C79AFULL, "162414421.ogg"},
	{0x938DDE3759D42E40ULL, "6132679_m.fxe"},
	{0x938E1E2E310538C4ULL, "llac_w_sundermount_-262147_0.dds"},
	{0x938E31BF80783D5BULL, "vxlt_gallowscourtyard_n_7a0_0_0.vlm"},
	{0x938E3448EB9B8A9CULL, "vxw_bonepitdestroyed_a3_27m_0_0.vlm"},
	{0x938E503C18C2E6D7ULL, "hlac_ht_chantry_2534_26x.dds"},
	{0x938E624369E2F539ULL, "t3_glk_h03.tnt"},
	{0x938E774E2AB4C2D6ULL, "w_encounter3_179_0.msh"},
	{0x938E7B02C61C4ED8ULL, "6207552_f.fxe"},
	{0x938E814E2AB4D3D4ULL, "w_encounter3_179_0.mmh"},
	{0x938E854E2AB4DAE7ULL, "w_encounter3_179_0.mao"},
	{0x938EF4B26F78D341ULL, "pro000tr_aveline_fight.utt"},
	{0x938F96359EA4B2FFULL, "358802728.ogg"},
	{0x938FA8DAC6DAF5BCULL, "gen00fl_aveline6137619.cl"},
	{0x938FDDF48D951485ULL, "891766099.ogg"},
	{0x938FE9D5CA51D897ULL, "hlac_ht_keep_3732_1x.dds"},
	{0x9390E1FA29B648B4ULL, "gen_im_trs_gem_coral_01.uti"},
	{0x9390F9477CF36028ULL, "vxlt_0_lowtown_a2_n_c_cmt_0_0.vlm"},
	{0x93911A335AB7F750ULL, "vxlt_undercityserialkiller_18i_0_0.vlm"},
	{0x93911BC16334664DULL, "dae020tr_arishok_load.utt"},
	{0x9391DE81E63BF392ULL, "gen00fl_anders6113025.cl"},
	{0x9392F0529495135FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.cess.evt"},
	{0x9392FA4F2B4F579BULL, "6062803_m.fxe"},
	{0x9393FD3C90450DB9ULL, "mag370_meredith6092187.cl"},
	{0x939440CC3DC91F11ULL, "hlac_lt_0_lowtown_a2_n_c_1_10y.dds"},
	{0x9394DED879CE5B43ULL, "art\\characters\\creatures\\spider\\animation\\enter.gad"},
	{0x939541DB0BDC10DEULL, "lgt300cr_gang_2_docks.utc"},
	{0x9395DAC734412422ULL, "vxht_keep_6c4_0_0.vlm"},
	{0x9396D20AB0C24666ULL, "vxdr_primevalentrance_39k_0_1.vlm"},
	{0x9396F8B192902D91ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp_basic.msh"},
	{0x939706B1929045ACULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp_basic.mao"},
	{0x939712B192905A4FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1pp_basic.mmh"},
	{0x93972BB8058BBA5BULL, "dae000_arishok6068496.cl"},
	{0x939748B625407DCBULL, "dae000_mother6136935.cl"},
	{0x9397651762908B8EULL, "vrc221_bartrand6043282.cl"},
	{0x939771D86CC0DBB8ULL, "sky_lt_0_lowtown_docks_n.mao"},
	{0x9397BE18B90CF047ULL, "6176580_m.fxe"},
	{0x9397E42FD576EA16ULL, "llac_lt_0_lowtown_a2_n_-65537_1.dds"},
	{0x93983A94C8639C1FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_5.pwk"},
	{0x93985C530C96012BULL, "vxca_caverns1_mines_in0_0_0.vlm"},
	{0x9398E60E97BF3F28ULL, "ser221_party_comments6116128.cl"},
	{0x9398E970C0FCA0D5ULL, "6147568_m.fxe"},
	{0x939904B0D6FB5C53ULL, "907298590.ogg"},
	{0x93992101B01EC01BULL, "fxc_puke_blend.mao"},
	{0x9399573C89C9C74EULL, "w_encounter3_104_0.msh"},
	{0x9399653C89C9DE9FULL, "w_encounter3_104_0.mao"},
	{0x9399713C89C9F33CULL, "w_encounter3_104_0.mmh"},
	{0x939977AAC32B1FE6ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.dura.ani"},
	{0x9399D152949AFD48ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_mo_c.cess.gad"},
	{0x9399D36AA0B45670ULL, "6167440_m.fxe"},
	{0x939A1241C4B28C48ULL, "gen00fl_anders6116103.cl"},
	{0x939A205AC9A5E6F6ULL, "vxlt_0_lowtown_docks_b6r_0_0.vlm"},
	{0x939A4B464D0D0A68ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblepile.msh"},
	{0x939A5D464D0D28F9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblepile.mao"},
	{0x939A69464D0D3C92ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblepile.mmh"},
	{0x939ADA27926F7B3DULL, "vxlt_0_lowtown_ei2_0_0.vlm"},
	{0x939B1BA3F3BEC78AULL, "vxht_keep_3y7_0_0.vlm"},
	{0x939B5E4010DD3F77ULL, "dae000cr_grand_cleric6182009.cl"},
	{0x939BBBD879D43E28ULL, "art\\characters\\creatures\\spider\\animation\\enter.evt"},
	{0x939BC87636C3F1EEULL, "102483084.ogg"},
	{0x939BD47E1956E807ULL, "597022932.ogg"},
	{0x939C6C20F8454256ULL, "346280023.ogg"},
	{0x939CD650A8C2AABDULL, "vxht_chantry_n_2yb_0_0.vlm"},
	{0x939CE03AFB9FAEE4ULL, "dre000_stroud6036688.cl"},
	{0x939D01FE45267105ULL, "702400897.ogg"},
	{0x939D20F77B129A6DULL, "hlac_ht_0_hightown_-131070_0y.dds"},
	{0x939F1EC550159962ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm01.msh"},
	{0x939F20C550159C18ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm01.mmh"},
	{0x939F47FF05ACC7BCULL, "and110_refugee6090307.cl"},
	{0x939F851CBA02338EULL, "w_sundermount_76_0.msh"},
	{0x939F931CBA024BDFULL, "w_sundermount_76_0.mao"},
	{0x939F9F1CBA02607CULL, "w_sundermount_76_0.mmh"},
	{0x939FB9432320ADD5ULL, "468218475.ogg"},
	{0x939FC4FBD4830112ULL, "vis121_saemus6089668.cl"},
	{0x939FCB99545B50D6ULL, "vxw_bonepitdestroyed_a3_251_0_0.vlm"},
	{0x93A0C03883514097ULL, "vxlt_0_lowtown_a3_n_c_daq_0_0.vlm"},
	{0x93A0CE20A9004661ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchhang_n.dds"},
	{0x93A10DC53B2C762DULL, "vxht_mansion2_321_0_0.vlm"},
	{0x93A15F5747B72BA0ULL, "hlac_w_encounter2_camp_3_0y.dds"},
	{0x93A18BA31B714CCAULL, "vxlt_0_lowtown_efh_0_0.vlm"},
	{0x93A1CF8AC1F74B1BULL, "6166480_m.fxe"},
	{0x93A25CB896C48219ULL, "6090857_m.fxe"},
	{0x93A27974C0A6C448ULL, "52330575.ogg"},
	{0x93A2F0D7EE0D47A0ULL, "448207036.ogg"},
	{0x93A32C17F18E2AE2ULL, "117698215.ogg"},
	{0x93A40827BBED33B5ULL, "729411739.ogg"},
	{0x93A50E05C2759E37ULL, "zz_qun_debug6028260.cl"},
	{0x93A55BB7EDCFCF54ULL, "345850177.ogg"},
	{0x93A58879F055407CULL, "841099506.ogg"},
	{0x93A6B81DA20B78F6ULL, "805347249.ogg"},
	{0x93A6CE2FFE8C19ABULL, "vxca_caverns1_mines_ik3_0_0.vlm"},
	{0x93A6D2FDA8596B82ULL, "gen00fl_merrill6106511.cl"},
	{0x93A6D38EADD08FB3ULL, "857761645.ogg"},
	{0x93A6F00296FDA63EULL, "202966531.ogg"},
	{0x93A6F7B07B977E95ULL, "gen00fl_anders6125249.cl"},
	{0x93A701C5F2B69877ULL, "vxdr_primevaltomb_555_0_0.vlm"},
	{0x93A71DA33929C7A2ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hd1a.phy"},
	{0x93A752012FBEBB56ULL, "vxlt_hangedman_1rj_0_0.vlm"},
	{0x93A75FAAC3373457ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.dura.evt"},
	{0x93A762EA066ABBB9ULL, "6176177_m.fxe"},
	{0x93A7BE899D8C7AC3ULL, "6125841_m.fxe"},
	{0x93A7FCAB675FD717ULL, "follower_banter6162094.cl"},
	{0x93A8B22FDF1DFA79ULL, "llac_lt_0_lowtown_-65537_1.dds"},
	{0x93A927C5501DFCA3ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm01.phy"},
	{0x93A93A3A9F4EF23EULL, "vxht_chantry_2d9_0_0.vlm"},
	{0x93A941A658B8B2A9ULL, "vxca_cavern1_haunted_dpm_0_0.vlm"},
	{0x93A982BFBB53DE9FULL, "284923841.ogg"},
	{0x93A9B4EE4D8B47AEULL, "461974881.ogg"},
	{0x93A9B6BCD477AE72ULL, "6180513_m.fxe"},
	{0x93A9FDCB635B802CULL, "6067340_m.fxe"},
	{0x93AB2680AE9A75B9ULL, "644523542.ogg"},
	{0x93AB482D9AB8C33DULL, "vxht_chantry_n_2ta_0_0.vlm"},
	{0x93AC1BE1E67E1DC1ULL, "56694939.ogg"},
	{0x93AC9CABCC1794B6ULL, "575781760.ogg"},
	{0x93AD3ED0F72DF51FULL, "vxht_0_hightown_a3_apv_0_0.vlm"},
	{0x93AD9CB3581DB851ULL, "6111581_m.fxe"},
	{0x93ADBC47B2175A61ULL, "gen00fl_anders6111369.cl"},
	{0x93ADE8951E878420ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01wisptree_1.mao"},
	{0x93AE3D7646516956ULL, "vxw_bonepitdestroyed_a3_202_0_0.vlm"},
	{0x93AE51C4A94E863EULL, "vxlt_0_lowtown_a2_n_byi_0_0.vlm"},
	{0x93AE60AAC33D54E0ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.dura.gad"},
	{0x93AEA528FE4A0B05ULL, "6153918_m.fxe"},
	{0x93AF04C1634D3CC0ULL, "dae020tr_arishok_load.ncs"},
	{0x93AF2F0C65FF3163ULL, "330872691.ogg"},
	{0x93AF329C4B9BB84BULL, "art\\vfx\\creatures\\fxc_wisp_c.cess.ani"},
	{0x93AF32D34CDEEEA5ULL, "vxdr_primevaltomb_8fc_0_0.vlm"},
	{0x93AF63E44DC48AC3ULL, "character_character_p125.psh.fxo"},
	{0x93AF71A956396918ULL, "llac_ht_0_hightown_n_65533_0.dds"},
	{0x93AF88EEC21EE5EAULL, "703963846.ogg"},
	{0x93B00FC15384E718ULL, "214341344.ogg"},
	{0x93B03A21DAFE6F8AULL, "6072890_m.fxe"},
	{0x93B06B0D558A0E41ULL, "6036347_m.fxe"},
	{0x93B08EF1C6039233ULL, "dre000_hireling16038630.cl"},
	{0x93B14636AE47DBCEULL, "6104756_f.fxe"},
	{0x93B1D3C62145CF42ULL, "9109296.ogg"},
	{0x93B1D5FA4818095CULL, "716709264.ogg"},
	{0x93B1F6B0B9113802ULL, "vxca_caverns2prc_a0l_0_0.vlm"},
	{0x93B22A4960CA836AULL, "6085283_m.fxe"},
	{0x93B23175D4CEC604ULL, "6211923_m.fxe"},
	{0x93B242030251E60CULL, "vxlt_0_lowtown_a2_euw_0_0.vlm"},
	{0x93B2653AA1F64DF8ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyn_0s.dds"},
	{0x93B2BE81182D5522ULL, "vxht_keep_6i6_0_0.vlm"},
	{0x93B312A04EB142CBULL, "126073048.ogg"},
	{0x93B31EF382CF2917ULL, "art\\levels\\ztools\\default\\generic\\prp_joe.msh"},
	{0x93B330F382CF487AULL, "art\\levels\\ztools\\default\\generic\\prp_joe.mao"},
	{0x93B334F382CF4E89ULL, "art\\levels\\ztools\\default\\generic\\prp_joe.mmh"},
	{0x93B37F2E17DA4190ULL, "hlac_ca_caverns2prc_5802_27y.dds"},
	{0x93B3C741CB7D3B9BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartoidle.evt"},
	{0x93B476FA682A2B27ULL, "616951889.ogg"},
	{0x93B47745577F448EULL, "hlac_w_woundedcoast_131077_0x.dds"},
	{0x93B4A995D510BFB0ULL, "258764815.ogg"},
	{0x93B53EB6260F8475ULL, "hlac_lt_0_lowtown_docks_a3_-196611_0y.dds"},
	{0x93B5CD446245BBCEULL, "6179456_m.fxe"},
	{0x93B768CD44AB1338ULL, "hlac_w_bonepitdestroyed_a3_0_7y.dds"},
	{0x93B811D4048B5AE0ULL, "664863393.ogg"},
	{0x93B81E2DAF3F67B8ULL, "6038681_m.fxe"},
	{0x93B8B33F2C921F27ULL, "hlac_lt_0_lowtown_docks_n_196607_3z.dds"},
	{0x93B8B3B762DC75DDULL, "vxw_encounter2_63_0_0.vlm"},
	{0x93B920A1D36133EAULL, "vxw_encounter2_e5_0_0.vlm"},
	{0x93B925DC8B9927CEULL, "6107228_m.fxe"},
	{0x93B92C5503137DCEULL, "vxlt_0_lowtown_a2_n_c_hjh_0_0.vlm"},
	{0x93B958092EAAE661ULL, "hlac_ht_playershouse_n_855_4x.dds"},
	{0x93B95B3A9EA9A927ULL, "6135116_m.fxe"},
	{0x93B9761DF8B7A9E4ULL, "853037503.ogg"},
	{0x93B979EDA17786F5ULL, "one100dg_nabil6033152.cl"},
	{0x93B9E5B155ADFBD0ULL, "vxw_bonepitmining_a2_15b_0_0.vlm"},
	{0x93B9EDD41086BB5EULL, "6170868_m.fxe"},
	{0x93B9F4C442429031ULL, "ser270_templar6116091.cl"},
	{0x93BA3F9E488F5F4BULL, "6051764_m.fxe"},
	{0x93BA937920FADC26ULL, "vxht_0_hightown_cf2_0_0.vlm"},
	{0x93BA9E9D6D61EA16ULL, "vxht_mansion2_2e6_0_0.vlm"},
	{0x93BAC041CB834EDCULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartoidle.gad"},
	{0x93BACE60BF58FB29ULL, "llac_ca_cavern1_haunted_24423_8.dds"},
	{0x93BB2A80E7D09451ULL, "dae000_arishok6067774.cl"},
	{0x93BB2FB125CCD823ULL, "combat_ws_r1_a5.gda"},
	{0x93BBCD2BADFB23E4ULL, "vxlt_gallowstemplar_f_3r8_0_0.vlm"},
	{0x93BBD553952E0A99ULL, "plt_mcr000pt_main.plo"},
	{0x93BC2221AEFBB69AULL, "vxht_mansion_n_3za_0_0.vlm"},
	{0x93BCAA112668ECE0ULL, "vxht_mansion2b_1io_0_0.vlm"},
	{0x93BD1A9C4BA7CCC2ULL, "art\\vfx\\creatures\\fxc_wisp_c.cess.evt"},
	{0x93BD90CC5EC59A62ULL, "923419066.ogg"},
	{0x93BE0658E674B0F5ULL, "vxca_caverns1_e8e_0_0.vlm"},
	{0x93BE2349B2EEED31ULL, "one140cr_charade6128498.cl"},
	{0x93BE3C9C2366F3ACULL, "646724937.ogg"},
	{0x93BE9F1CBA1D10CFULL, "w_sundermount_76_0.rml"},
	{0x93BF01524600C2B1ULL, "904205870.ogg"},
	{0x93BF3DBE5A24BFAFULL, "6081985_m.fxe"},
	{0x93BF6A9FDBB5EA25ULL, "llac_w_blightlands_65537_0.dds"},
	{0x93BF73B64548ABA8ULL, "follower_banter6161003.cl"},
	{0x93BFD3D5E94445FBULL, "fxc_water_drop.dds"},
	{0x93BFD4F00713639CULL, "hlac_lt_0_lowtown_-4_0x.dds"},
	{0x93C058A10201C9CCULL, "6034080_f.fxe"},
	{0x93C0831004895E83ULL, "vxlt_gallowstemplar_3n8_0_0.vlm"},
	{0x93C0B3DFF447FE8CULL, "vxlt_0_lowtown_a2_ext_0_0.vlm"},
	{0x93C0BF41CB87B8DAULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\speartoidle.ani"},
	{0x93C0C0F167B77AB7ULL, "vxlt_darktown_56n_0_0.vlm"},
	{0x93C118C194D2BC7BULL, "hlac_lt_0_lowtown_a2_n_65535_7z.dds"},
	{0x93C14EBDC5C5FDF0ULL, "6206912_m.fxe"},
	{0x93C208E26AFAFAB7ULL, "ser221_gamlen6131909.cl"},
	{0x93C21E79B56297C6ULL, "vxlt_0_lowtown_docks_b9n_0_1.vlm"},
	{0x93C243D739EA7AAEULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc106a_0t.dds"},
	{0x93C25E40ECE2DBE3ULL, "art\\characters\\weapons\\dagger\\w_dger_103a.mmh"},
	{0x93C26C40ECE2F33DULL, "art\\characters\\weapons\\dagger\\w_dger_103a.msh"},
	{0x93C29CC19F143599ULL, "vxlt_0_lowtown_haw_0_0.vlm"},
	{0x93C3199C4BAC36D5ULL, "art\\vfx\\creatures\\fxc_wisp_c.cess.gad"},
	{0x93C33A0226A9ED93ULL, "638377656.ogg"},
	{0x93C40450D12C886CULL, "126576195.ogg"},
	{0x93C41B88AFDE180DULL, "449198006.ogg"},
	{0x93C4493F8825D040ULL, "vxlt_warehouse_n_co_0_0.vlm"},
	{0x93C4C3F5182AFAFEULL, "vxht_mansion2_3e3_0_0.vlm"},
	{0x93C57E1CBA22F747ULL, "w_sundermount_76_0.phy"},
	{0x93C5892640327B9AULL, "vxlt_hangedman_n_1il_0_0.vlm"},
	{0x93C59926981208ECULL, "w_encounter3_58_0.msh"},
	{0x93C59F26981212DEULL, "w_encounter3_58_0.mmh"},
	{0x93C5A9D33A8FEFDEULL, "mag320cr_noblewoman_2.utc"},
	{0x93C5AB2698122775ULL, "w_encounter3_58_0.mao"},
	{0x93C60CFB310FA25FULL, "6025028_m.fxe"},
	{0x93C671CB4705ACA2ULL, "vxht_0_hightown_n_ba9_0_0.vlm"},
	{0x93C69973D9BB7BB1ULL, "109369489.ogg"},
	{0x93C6D4304F85F461ULL, "845333713.ogg"},
	{0x93C6E6ECFD3B2D55ULL, "fx_upperlag.mao"},
	{0x93C7085F69D97DEAULL, "vxht_0_hightown_a2_avx_0_0.vlm"},
	{0x93C7259454D28E5DULL, "vxw_encounter2_12_0_0.vlm"},
	{0x93C750D11ABF907CULL, "686377797.ogg"},
	{0x93C79776119EFF1BULL, "375466677.ogg"},
	{0x93C82AABC038E858ULL, "hlac_lt_0_lowtown_n_-262144_3y.dds"},
	{0x93C8BAF374703F82ULL, "6064743_m.fxe"},
	{0x93C8E7831B616523ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf02a.mmh"},
	{0x93C8F5831B617D7DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf02a.msh"},
	{0x93C91DB24885CDF7ULL, "vxw_bonepit_y8_0_0.vlm"},
	{0x93C9854509332C6EULL, "6030229_m.fxe"},
	{0x93C9E72AD19EA30BULL, "hlac_dr_dungeon_3207_3x.dds"},
	{0x93CA1691ED673995ULL, "follower_banter6161332.cl"},
	{0x93CA384A00E2C688ULL, "lt_gallowstemplar_n_2388.rml"},
	{0x93CA56ADB5D46D41ULL, "809367921.ogg"},
	{0x93CA93FEA0F1CF1AULL, "vxht_mansion_n_3wb_0_0.vlm"},
	{0x93CAF3FFE46D61D7ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit.mmh"},
	{0x93CAF7FFE46D68E4ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit.mao"},
	{0x93CB09FFE46D8749ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit.msh"},
	{0x93CB215F99F2FFBBULL, "vxca_caverns2prc_9st_0_0.vlm"},
	{0x93CC3BF98DDC6DAAULL, "6060201_f.fxe"},
	{0x93CC6521D3BF57DFULL, "mag360_thrask6043648.cl"},
	{0x93CC9304F3808990ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_dsgaa_0d.dds"},
	{0x93CCCAFD23926229ULL, "408610925.ogg"},
	{0x93CD04A3394A6CF5ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hd1a.mmh"},
	{0x93CD06A3394A706BULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hd1a.msh"},
	{0x93CD3DCA218A8DF9ULL, "6152231_f.fxe"},
	{0x93CD3FF551FBB82FULL, "zz_per_debug.ncs"},
	{0x93CD7411DBF2A46DULL, "vxw_bonepitdestroyed_a3_191_0_0.vlm"},
	{0x93CD7845955BAEE2ULL, "6062121_m.fxe"},
	{0x93CD7D3373BCEDD6ULL, "328180273.ogg"},
	{0x93CDE3A7F1FEE7E4ULL, "gen00fl_varric6093585.cl"},
	{0x93CE12DF9B928359ULL, "133186090.ogg"},
	{0x93CE1E01E0E9D64DULL, "hlac_lt_gallowstemplar_n_2383_11z.dds"},
	{0x93CE1F893DD5953CULL, "vxlt_0_lowtown_e2b_0_0.vlm"},
	{0x93CE3E7348CF79D2ULL, "gen00fl_dog6145218.cl"},
	{0x93CED5E612667F45ULL, "6079059_f.fxe"},
	{0x93CEF2053495829CULL, "dae000_gen_turn_in_56198672.cl"},
	{0x93CF2BE38E5B7864ULL, "hlac_lt_0_lowtown_n_65535_5y.dds"},
	{0x93CFD6794E2CB5CDULL, "llac_ht_0_hightown_-196610_0.dds"},
	{0x93CFDFAD86FAF0B6ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\prp_ccliff02_drelief.dds"},
	{0x93D0083AE5BF48F8ULL, "hlac_lt_0_lowtown_a2_-65536_2y.dds"},
	{0x93D00B6473D04F6CULL, "sib100pt_birthright.ncs"},
	{0x93D03A341AF36944ULL, "vxlt_0_lowtown_a3_n_c_fed_0_0.vlm"},
	{0x93D0549E5DDA8CA1ULL, "dae000_gamlen6026202.cl"},
	{0x93D0964406BBD0F8ULL, "vfx_vfx_p7.psh.fxo.dx11.assembly"},
	{0x93D0A7A5E98D1843ULL, "6132643_m.fxe"},
	{0x93D10594F747D038ULL, "w_woundedcoast_130_0.phy"},
	{0x93D10E47A1A3BD20ULL, "hlac_lt_0_lowtown_a2_n_c_2_0x.dds"},
	{0x93D1F6197EB73E86ULL, "vxlt_gallowscourtyard_n_7f0_0_0.vlm"},
	{0x93D27BE7C89C133CULL, "and100_anders6173575.cl"},
	{0x93D2BDFCD3075679ULL, "ser210_witness6168164.cl"},
	{0x93D2D610BBFF81C7ULL, "vxht_mansion_n_4yb_0_0.vlm"},
	{0x93D341D7D3D4E99EULL, "gen_im_cns_cft_arn_04.uti"},
	{0x93D35C8F6F5CD0E4ULL, "vxlt_0_lowtown_docks_n_b5v_0_0.vlm"},
	{0x93D35F40ECF15008ULL, "art\\characters\\weapons\\dagger\\w_dger_103a.phy"},
	{0x93D4171AA53AFBE2ULL, "347169664.ogg"},
	{0x93D476A635D11A89ULL, "vxht_keep_n_67d_0_0.vlm"},
	{0x93D49BF7D15BA7AAULL, "vxht_0_hightown_a3_c8e_0_0.vlm"},
	{0x93D5161B4CAA8F39ULL, "hightown_male_noble2_mob6170889.cl"},
	{0x93D586469F43AE10ULL, "6092498_m.fxe"},
	{0x93D6045BB74D64EAULL, "vxw_encounter2_k7_0_0.vlm"},
	{0x93D6642DD7BD2B7CULL, "llac_lt_0_lowtown_a3_n_c_-262143_0.dds"},
	{0x93D7075D9B3AE208ULL, "gen00fl_isabela6090319.cl"},
	{0x93D72CD066665802ULL, "6063643_m.fxe"},
	{0x93D75169D1E851F8ULL, "6000884_m.fxe"},
	{0x93D78257514E1B16ULL, "vxht_mansion2_2k8_0_0.vlm"},
	{0x93D7BF93C2BB4D6EULL, "218802759.ogg"},
	{0x93D7C26165D9CD44ULL, "6094737_m.fxe"},
	{0x93D7D950C1B093F6ULL, "974046862.ogg"},
	{0x93D8028188962DE3ULL, "vxht_mansion_n_672_0_0.vlm"},
	{0x93D82E5DB72990A4ULL, "6143812_f.fxe"},
	{0x93D923D105EA9925ULL, "700204093.ogg"},
	{0x93D96DC830F26C21ULL, "182128995.ogg"},
	{0x93D98DCB0A551DE0ULL, "vxht_mansion2b_1gm_0_0.vlm"},
	{0x93D9E8831B6FDA48ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf02a.phy"},
	{0x93DA15264594EA69ULL, "vxdr_primevalentrance_1ku_0_0.vlm"},
	{0x93DA3F308D80AC75ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblelarge.phy"},
	{0x93DA44395D1131EEULL, "follower_banter6155943.cl"},
	{0x93DA66AE2450D05FULL, "plt_mrl250pt_mirrorimg.nss"},
	{0x93DA687F7FCE9666ULL, "one000_ilen6205697.cl"},
	{0x93DACE013E1E1BE9ULL, "gen00fl_aveline6116684.cl"},
	{0x93DADB5FC806B869ULL, "703348552.ogg"},
	{0x93DB04D3BAD4ACFCULL, "w_encounter3_239_0.mao"},
	{0x93DB10D3BAD4C09FULL, "w_encounter3_239_0.mmh"},
	{0x93DB16D3BAD4CB41ULL, "w_encounter3_239_0.msh"},
	{0x93DB25E8B6AF1CCEULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr180runf.ani"},
	{0x93DB45A6720B768EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_basic_1.mao"},
	{0x93DC565271D3E665ULL, "zz_approval_debug6003617.cl"},
	{0x93DC79DD1AF9A03BULL, "gen_im_trs_apo_sta_fir_r0.uti"},
	{0x93DC7D89B7F3A181ULL, "vxht_chantry_n_300_0_0.vlm"},
	{0x93DC9DE51917BFCFULL, "gen00fl_anders6183827.cl"},
	{0x93DCA4EDFD50B4C7ULL, "vxlt_gallowstemplar_5k5_0_0.vlm"},
	{0x93DD0392348ED7A3ULL, "6118525_f.fxe"},
	{0x93DD0F7850A1E9F1ULL, "vxw_encounter2_camp_nw_0_0.vlm"},
	{0x93DD489E2457A39AULL, "lt_0_lowtown_a2_probe_gp1.dds"},
	{0x93DD4F3838C69DEDULL, "vxca_cavern1_haunted_in6_0_0.vlm"},
	{0x93DD5C8C018C7519ULL, "mrl150_merrill6117762.cl"},
	{0x93DDD053954AE8AEULL, "plt_mcr000pt_main.nss"},
	{0x93DE73279A775FB8ULL, "vxht_mansion2_3gk_0_0.vlm"},
	{0x93DEADECFD4F5E33ULL, "fx_upperlag.dds"},
	{0x93DFE6E794C98D37ULL, "art\\levels\\races\\proxy\\props\\_textures\\fires_n.dds"},
	{0x93DFF0F548F2334DULL, "6085738_m.fxe"},
	{0x93DFF86E6E62128CULL, "vxht_keep_n_6ll_0_0.vlm"},
	{0x93E0401A591F2F42ULL, "850424894.ogg"},
	{0x93E067F670AD5706ULL, "vxlt_gallowscourtyard_n_7i1_0_0.vlm"},
	{0x93E06D806A5A786BULL, "hlac_lt_0_lowtown_a2_n_-65535_6x.dds"},
	{0x93E07E6BBF5DC15AULL, "6091369_m.fxe"},
	{0x93E085234DF5DC58ULL, "9849617.ogg"},
	{0x93E0AF82D62DE5EAULL, "art\\vfx\\combat\\vfx_w_gn2_stun_c.anb"},
	{0x93E0F06EFB1AD6A6ULL, "and100_anders6095840.cl"},
	{0x93E148C8215FB1B7ULL, "hlac_lt_0_lowtown_a2_n_-65536_5x.dds"},
	{0x93E17E5E8F539E73ULL, "vxlt_0_lowtown_a2_n_c_gc9_0_0.vlm"},
	{0x93E24C624A353232ULL, "ss_ogrerip.cut"},
	{0x93E2BD641498EA1CULL, "vxdr_primevaltomb_91b_0_0.vlm"},
	{0x93E2E6F13F9BD7A2ULL, "127545778.ogg"},
	{0x93E2E88327C73309ULL, "vxht_keep_n_62k_0_0.vlm"},
	{0x93E345F35097C183ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforwardleft.ani"},
	{0x93E43B66FE0D20FBULL, "hlac_ca_caverns2_5802_31x.dds"},
	{0x93E45B8A4A6DD203ULL, "6152753_m.fxe"},
	{0x93E45DCBEDF76292ULL, "37495343.ogg"},
	{0x93E478F19E82AD14ULL, "85824874.ogg"},
	{0x93E47FFD17B4A7DAULL, "genip_cleaning_brush.utp"},
	{0x93E5C5891F22120EULL, "art\\vfx\\combat\\vfx_c_pride_s_p.dura.ani"},
	{0x93E5D566929B425BULL, "vxca_caverns1_iyx_0_0.vlm"},
	{0x93E6EBD79860E763ULL, "llac_lt_gallowscourtyard_n_-131071_2.dds"},
	{0x93E71DD680560921ULL, "330162156.ogg"},
	{0x93E7413C90632DBEULL, "6052558_f.fxe"},
	{0x93E799A8D7A65EE5ULL, "lgt120_orz_bln_official6140218.cl"},
	{0x93E832CFE8CD757AULL, "zz_setheadmorph.ncs"},
	{0x93E89C3D7A32E8EAULL, "401845664.ogg"},
	{0x93E8BC96B2F7489EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_carver_loop.ani"},
	{0x93E90F90E1638B8EULL, "follower_banter6155598.cl"},
	{0x93E9C73FC78C9B35ULL, "vxw_encounter3_d6_0_0.vlm"},
	{0x93EA804E2B0330CFULL, "w_encounter3_179_0.phy"},
	{0x93EAA047AEB938B5ULL, "vxca_caverns1_ikp_0_0.vlm"},
	{0x93EB16CAEF46CD47ULL, "vxlt_gallowstemplar_5t4_0_0.vlm"},
	{0x93EB66200EB83C2CULL, "564555129.ogg"},
	{0x93EB815542980271ULL, "vxw_encounter2_camp_iv_0_0.vlm"},
	{0x93EB9A9CE1F3842EULL, "mag370_meredith6046860.cl"},
	{0x93EBC1152ABCB66DULL, "vxca_cavern1_haunted_ik5_0_0.vlm"},
	{0x93EBD28F49AC998DULL, "dae000_meredith6092597.cl"},
	{0x93EC21003957749DULL, "6150700_m.fxe"},
	{0x93EC2B0D157D0F8CULL, "static_static_p3.psh.fxo.dx11"},
	{0x93EC3930121267F9ULL, "505701924.ogg"},
	{0x93EC576B8CDD585FULL, "516193677.ogg"},
	{0x93EC67AF53D76D39ULL, "482245536.ogg"},
	{0x93ECCB5038BC11B2ULL, "zz_followers_banter6158043.cl"},
	{0x93ECF34B62193311ULL, "qcr220_grey_wardens6135906.cl"},
	{0x93EDA2307ACF1DD5ULL, "vxw_bonepitmining_a2_23z_0_0.vlm"},
	{0x93EDDF6475BB32F9ULL, "hlac_ht_keep_3743_1z.dds"},
	{0x93EE1388818EA3F9ULL, "vxlt_darktown_583_0_1.vlm"},
	{0x93EE50B65507807DULL, "ht_0_hightownexplosion_n_34_0.rml"},
	{0x93EE60403B786B53ULL, "art\\characters\\weapons\\arrows\\textures\\w_arw_fire_e.dds"},
	{0x93EEC33B26783227ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic_1.mao"},
	{0x93EF59AE24631670ULL, "plt_mrl250pt_mirrorimg.plo"},
	{0x93EF65C3C7E042FEULL, "zz_dre_debug6025557.cl"},
	{0x93EF87A78B0E1E1DULL, "vxht_bartrand_n_67s_0_0.vlm"},
	{0x93EF9ED65E0819C0ULL, "vxlt_warehouse_n_pj_0_0.vlm"},
	{0x93EFFEE8B6C13608ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\pauseturnr180runf.gad"},
	{0x93F0DA2601BF9E47ULL, "vxht_0_hightown_bx8_0_0.vlm"},
	{0x93F115E031C4B86FULL, "vxlt_0_lowtown_docks_a3_n_brk_0_0.vlm"},
	{0x93F15489607F4622ULL, "llac_lt_gallowscourtyard_n_-65538_1.dds"},
	{0x93F17FB1B547D8AAULL, "vxht_0_hightown_a3_c6g_0_0.vlm"},
	{0x93F1DFF5794C625FULL, "16812119.ogg"},
	{0x93F270CBF2496938ULL, "staticdebug.psh.fxo"},
	{0x93F2825AB390F462ULL, "vxca_caverns2prc_9q7_0_0.vlm"},
	{0x93F2BD891F2C8F5FULL, "art\\vfx\\combat\\vfx_c_pride_s_p.dura.evt"},
	{0x93F2DD7AA2F869B6ULL, "6126359_m.fxe"},
	{0x93F2E8159B3995EAULL, "vxw_encounter2_q1_0_0.vlm"},
	{0x93F31793D09EB370ULL, "1066473519.ogg"},
	{0x93F347EC99F1345DULL, "hlac_lt_0_lowtown_a2_n_0_4y.dds"},
	{0x93F3B6004851D2A4ULL, "668154907.ogg"},
	{0x93F4F6CCEBF82E4BULL, "423394639.ogg"},
	{0x93F50E4311862DD1ULL, "gen_im_npc_wep_one_axeb.uti"},
	{0x93F5160BC89EE3E1ULL, "vxht_0_hightown_n_c4t_0_0.vlm"},
	{0x93F520F093D60F5EULL, "vxht_mansion2_25n_0_0.vlm"},
	{0x93F595AF1C84FA06ULL, "plt_vrc000pt_n_friend_mid.nss"},
	{0x93F5AD4EA9328332ULL, "vxw_bonepitmining_a2_dh_0_0.vlm"},
	{0x93F5EFCB948BE3A4ULL, "vxht_mansion2_2nu_0_0.vlm"},
	{0x93F5F21082DB661FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxa.msh"},
	{0x93F5F81082DB70C1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxa.mmh"},
	{0x93F6041082DB8562ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxa.mao"},
	{0x93F668464D5B998DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_2rubblepile.phy"},
	{0x93F6F83B4B8CB467ULL, "6120600_m.fxe"},
	{0x93F75CF350A8926DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkforwardleft.gad"},
	{0x93F7A74CA4746508ULL, "vxht_0_hightown_b7b_0_0.vlm"},
	{0x93F7B1A6DF542A94ULL, "mcr310_argument6072704.cl"},
	{0x93F7C13CAD6A8722ULL, "6049449_m.fxe"},
	{0x93F7F5A7FCE9B5A6ULL, "453133739.ogg"},
	{0x93F8391CB982B3B5ULL, "vxw_encounter3_a5_0_0.vlm"},
	{0x93F972B75B4B51ADULL, "6067068_m.fxe"},
	{0x93F99E891F327948ULL, "art\\vfx\\combat\\vfx_c_pride_s_p.dura.gad"},
	{0x93F9B0F4A3632D10ULL, "vxlt_warehouse_n_145_0_0.vlm"},
	{0x93F9E7E3194A389EULL, "hlac_lt_0_lowtown_a2_-65536_5z.dds"},
	{0x93F9F332348E1AF1ULL, "vxw_encounter2_camp_du_0_0.vlm"},
	{0x93FA310C9A853F18ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st03a_0t.dds"},
	{0x93FAC2E54DFC1E6EULL, "qun131_taskmaster6039403.cl"},
	{0x93FB42CD5083C745ULL, "dae000_loyal_mage6165233.cl"},
	{0x93FB6061DFC1BB63ULL, "vxlt_hangedman_n_1ya_0_0.vlm"},
	{0x93FB6E405135903EULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_aoe_p.impa.evt"},
	{0x93FC39757FF96052ULL, "137356696.ogg"},
	{0x93FCC88B784197F6ULL, "555726080.ogg"},
	{0x93FCD596B3081C18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_carver_loop.gad"},
	{0x93FCDC28524E438CULL, "vxht_keep_n_6bb_0_0.vlm"},
	{0x93FCDCE8814E15CDULL, "llac_ca_caverns1_mines_24427_2.dds"},
	{0x93FD52F7ADDD7E92ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb106a_0n.dds"},
	{0x93FDEFDB05BDC453ULL, "ico_boots_heavy_1.dds"},
	{0x93FE10B34FFE3240ULL, "vxlt_warehouse_n_wk_0_0.vlm"},
	{0x93FE7370CE044205ULL, "hlac_ht_bartrand_n_3374_1x.dds"},
	{0x93FE9EC189F38FC8ULL, "vxht_0_hightown_n_c4q_0_0.vlm"},
	{0x93FEE81292A791FCULL, "6079668_m.fxe"},
	{0x93FF18F6216F02DCULL, "134916080.ogg"},
	{0x93FF33D0305D5C41ULL, "abi_m_attwave.wbd"},
	{0x93FFC7F3189DECABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_begging.ani"},
	{0x94002AABAD66897FULL, "fen310_varania6118645.cl"},
	{0x94005A3E24DE398BULL, "769632783.ogg"},
	{0x94009B93619AC5BCULL, "zz_followers_banter6159594.cl"},
	{0x9400AF070A60395AULL, "vxlt_warehouse_n_xx_0_0.vlm"},
	{0x9400BF9F702563ECULL, "435583789.ogg"},
	{0x9400CF3B0CA62471ULL, "hlac_lt_undercityserialkiller_-1_3z.dds"},
	{0x9401406446EF24DDULL, "vxw_blightlands_q9_0_0.vlm"},
	{0x94016D67D04AB2BEULL, "one110_tree_minder6140864.cl"},
	{0x940173626F9C629CULL, "vxw_encounter2_camp_31_0_0.vlm"},
	{0x9401BD7FD32D2D7DULL, "mer240ip_sabins_stash.ncs"},
	{0x9401EC186AE9278AULL, "cod_tomes_tome_2.utp"},
	{0x94028540513BD669ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_aoe_p.impa.gad"},
	{0x9402B9207687735BULL, "vxca_caverns1_igz_0_0.vlm"},
	{0x940338D9EC560512ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\large_sa.dds"},
	{0x940387E8BA94FC61ULL, "vxht_0_hightown_n_c1k_0_0.vlm"},
	{0x94041F2B9B289BB2ULL, "vxw_bonepitmining_a2_ci_0_0.vlm"},
	{0x94044ED91F07D271ULL, "hlac_ca_caverns2prc_5802_5z.dds"},
	{0x9404762C8B6420F4ULL, "6131128_m.fxe"},
	{0x940498D5E6661056ULL, "vxw_encounter2_camp_6d_0_0.vlm"},
	{0x9404AC238CE9F167ULL, "hlac_lt_0_lowtown_n_65537_6y.dds"},
	{0x9404B8600CFD81E3ULL, "t3_cth_cm2.tnt"},
	{0x9404D3AE3F550482ULL, "llac_ca_caverns2prc_5802_46.dds"},
	{0x9404E8D361CB61BBULL, "vxca_caverns2prc_9oh_0_0.vlm"},
	{0x940500CAC89DBA47ULL, "l_dre180ar_ruins_entrance.wbd"},
	{0x94054C744D0BE5BDULL, "hlac_ht_chantry_n_2534_19y.dds"},
	{0x940590AAEAEA4E8EULL, "hm_seralrik.mor"},
	{0x940592FDD9A56BD8ULL, "6207743_f.fxe"},
	{0x9405CB832D44993CULL, "qun221_petrice6088228.cl"},
	{0x940643C967F7F507ULL, "6136881_m.fxe"},
	{0x9406E4EF7DF04F70ULL, "w_blightlands_19_0.rml"},
	{0x94076090E4DAF40BULL, "8159792.ogg"},
	{0x9407E4944D416A2DULL, "6062368_m.fxe"},
	{0x94080E70D60B6D45ULL, "gen00fl_isabela6106284.cl"},
	{0x9408120EE1951B2DULL, "vxw_encounter3_82_0_0.vlm"},
	{0x940822D195594590ULL, "vxlt_warehouse_n_192_0_0.vlm"},
	{0x9408A6FA6FF5ABF8ULL, "vxca_cave_vw_0_0.vlm"},
	{0x9408ADBAF73BF319ULL, "dae280ar_deep_roads.are"},
	{0x9408CEA00791F7BAULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic_3.mao"},
	{0x94090570C94DDC12ULL, "vxht_mansion2b_2v3_0_0.vlm"},
	{0x940915A6B6772E22ULL, "synclargecreatures5.gda"},
	{0x940948AE2F63C216ULL, "1045260028.ogg"},
	{0x940956405141A4B7ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_aoe_p.impa.ani"},
	{0x9409567E490BD4ADULL, "llac_lt_0_lowtown_docks_a3_-131075_0.dds"},
	{0x94096605EE30F79EULL, "157175040.ogg"},
	{0x94096DC181767A14ULL, "vxca_caverns1_mines_jfd_0_0.vlm"},
	{0x9409A4B267F58526ULL, "gen00fl_merrill6116316.cl"},
	{0x9409AAAF1C95C741ULL, "plt_vrc000pt_n_friend_mid.plo"},
	{0x9409B282D650EFDDULL, "art\\vfx\\combat\\vfx_w_gn2_stun_c.mmh"},
	{0x940B9129312AC785ULL, "73068344.ogg"},
	{0x940BD1EE08F17C8AULL, "hlac_w_woundedcoast_3_0x.dds"},
	{0x940C1AD6A8439649ULL, "vxw_sundermount_32y_0_0.vlm"},
	{0x940C52B4E291A9BEULL, "6108845_m.fxe"},
	{0x940C7F26984DDE95ULL, "w_encounter3_58_0.rml"},
	{0x940C829041F44AC0ULL, "vxlt_warehouse_n_zh_0_0.vlm"},
	{0x940CA7ABD897BA79ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a.msh"},
	{0x940CB1ABD897CB67ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a.mmh"},
	{0x940CB5ABD897D1B4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a.mao"},
	{0x940CBAA095A27828ULL, "fx_rock_chips.dds"},
	{0x940CD638D1F3680DULL, "409811851.ogg"},
	{0x940D109E7BE9A848ULL, "vxht_0_hightown_n_c1n_0_0.vlm"},
	{0x940DAFF318AA0122ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_begging.evt"},
	{0x940E23D586599DA5ULL, "ht_0_hightown_a2_49_0.rml"},
	{0x940E36A0613438E9ULL, "ser120_ghyslain6104320.cl"},
	{0x940E3A66BDCCF4A6ULL, "vxht_keep_5gs_0_0.vlm"},
	{0x940E672CD2C26E28ULL, "525822356.ogg"},
	{0x940FF17E698D611CULL, "art\\characters\\playercharacter\\humanmale\\hn_boo_lgtb.mao"},
	{0x941075737DC4FD86ULL, "579245406.ogg"},
	{0x9410F113F9DF44B5ULL, "ach_soc_recruit.dds"},
	{0x94110DC2C455B71FULL, "vxht_keep_5e8_0_0.vlm"},
	{0x941111DA1555BCD0ULL, "6057094_f.fxe"},
	{0x9412EBD4A0321548ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_018a_0d.dds"},
	{0x94130A046DFC438EULL, "265076064.ogg"},
	{0x94130AB2D85C28D6ULL, "vxw_encounter2_camp_1e_0_0.vlm"},
	{0x94138E8C46BAEB52ULL, "dae000_mother6121024.cl"},
	{0x9413E48A9E30B77FULL, "art\\characters\\creatures\\mubari\\animation\\cwkfl.gad"},
	{0x94143E3CC2FC72EFULL, "6066713_m.fxe"},
	{0x94148FAA65859F3AULL, "vxw_encounter2_camp_qi_0_0.vlm"},
	{0x9414A2DA401C49C1ULL, "vxlt_hangedman_26z_0_0.vlm"},
	{0x9414ABF16980D7C7ULL, "6095597_m.fxe"},
	{0x9414AEF318B01E35ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_begging.gad"},
	{0x9414BAD2F04EA60DULL, "vxdr_0_deeproads_f9n_0_0.vlm"},
	{0x9414E80BDDD01195ULL, "6119404_m.fxe"},
	{0x941538134B204B70ULL, "dae000_elegant6166884.cl"},
	{0x94158B3C294C3EF4ULL, "vxdr_0_deeproads_bus_0_0.vlm"},
	{0x9415E58C3278A43EULL, "vxlt_0_lowtown_a2_ghd_0_0.vlm"},
	{0x941683EBD38B33ADULL, "vxw_encounter3_51_0_0.vlm"},
	{0x9416A6602BA692BEULL, "lgt110_gxa_vgk_silver_order6116989.cl"},
	{0x941735A128CB61F5ULL, "941682427.ogg"},
	{0x9417FF0B33BF0F5CULL, "638900426.ogg"},
	{0x9418764381F3D0B4ULL, "364820151.ogg"},
	{0x9418AF55AEA8641CULL, "vxca_caverns1_j76_0_0.vlm"},
	{0x9418C92AC6D17780ULL, "6092182_m.fxe"},
	{0x941903E81AA6DB82ULL, "vxw_woundedcoast_1cw_0_0.vlm"},
	{0x9419334BDA387C74ULL, "sky_ht_playershouse_n.mao"},
	{0x941952EAA7216DC2ULL, "vxw_bonepit_260_0_0.vlm"},
	{0x94197C79C4B485FAULL, "vxca_caverns2_74g_0_0.vlm"},
	{0x9419B732EA154B20ULL, "vxdr_primevaltomb_7e8_0_0.vlm"},
	{0x9419D98A9E35106CULL, "art\\characters\\creatures\\mubari\\animation\\cwkfl.evt"},
	{0x941A05B4CFEAE50FULL, "vxlt_warehouse_15n_0_0.vlm"},
	{0x941A43A5E64C1D19ULL, "6097406_f.fxe"},
	{0x941A482B9C75EE36ULL, "vxlt_0_lowtown_a2_n_c_ekg_0_1.vlm"},
	{0x941B6E78B1DE44BDULL, "vxca_caverns2prc_657_0_0.vlm"},
	{0x941C30F5310AAA41ULL, "ht_0_hightown_a3_ll.mmh"},
	{0x941C36C30F1CFAA8ULL, "gen00fl_merrill6116044.cl"},
	{0x941CFBF317B65ACAULL, "537955004.ogg"},
	{0x941D088299C5F75CULL, "llac_lt_0_lowtown_a3_n_c_-2_6.dds"},
	{0x941D44C053E2AEEAULL, "686702328.ogg"},
	{0x941D4EF4246D3FE3ULL, "vxht_mansion_n_3yq_0_0.vlm"},
	{0x941DF87E09AF453EULL, "441251959.ogg"},
	{0x941E89E4B48A4ED2ULL, "vxca_cave_2ij_0_0.vlm"},
	{0x941EBF5E181C4D92ULL, "vxht_bartrand_n_63d_0_0.vlm"},
	{0x941FB95631AC3EB0ULL, "545978354.ogg"},
	{0x9420CC9595B0A180ULL, "dae000_unhappy_mage6165342.cl"},
	{0x9420E92B2F7A305DULL, "hlac_lt_foundry_n_320_7y.dds"},
	{0x942131CB0D641393ULL, "ave100_donnic6105952.cl"},
	{0x942152308DBCCE90ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblelarge.msh"},
	{0x942160308DBCE6AAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblelarge.mmh"},
	{0x942164308DBCEDE1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_1rubblelarge.mao"},
	{0x9421683205572A8AULL, "isa211_isabela6202272.cl"},
	{0x9421F9108300AD3EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxa.phy"},
	{0x9422127F90252DE0ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_dsgaa_0t.dds"},
	{0x942244E18F73D1E8ULL, "gen_im_trs_crc_junk2_01.uti"},
	{0x942274A967FFDBC9ULL, "6124365_m.fxe"},
	{0x94227E09103F8269ULL, "6170607_m.fxe"},
	{0x9422BAFD6E896103ULL, "hlac_lt_gallowstemplar_2383_11y.dds"},
	{0x9422DB79FBAD20DAULL, "6119502_m.fxe"},
	{0x9422E65C2382C8DBULL, "dae000_pickpocket_scene6175069.cl"},
	{0x94230187577BB7BAULL, "vxw_encounter2_camp_lj_0_0.vlm"},
	{0x942304A9665577D8ULL, "6187752_m.fxe"},
	{0x94231369D2178BF2ULL, "6102897_f.fxe"},
	{0x94240385C0B30983ULL, "6164082_m.fxe"},
	{0x942455D0307CB243ULL, "abi_m_attwave.bnk"},
	{0x942456C77BE6E87AULL, "376641300.ogg"},
	{0x9424FDFDDCF0C60BULL, "clouds.psh.fxo"},
	{0x942511951B71B8C6ULL, "6155775_m.fxe"},
	{0x942537F5F1FA3C08ULL, "plt_zz_mcrpt_harvester.nss"},
	{0x94253AE7BBF0D19DULL, "93645194.ogg"},
	{0x94256DC022F61FB5ULL, "571029306.ogg"},
	{0x94261C7D848B381CULL, "6114820_m.fxe"},
	{0x942621581C9039D1ULL, "plt_zz_vault_debug.nss"},
	{0x94262BC0D22E831DULL, "plt_zz_vrcpt_debug.plo"},
	{0x94272B69C7D5B20CULL, "989335988.ogg"},
	{0x94272E9DFEC83B3DULL, "art\\vfx\\combat\\vfx_gib_qunw_ltr.anb"},
	{0x942734A93832CAEDULL, "vxlt_0_lowtown_a2_byq_0_0.vlm"},
	{0x942775C50C9CF402ULL, "vxw_woundedcoast_1np_0_0.vlm"},
	{0x9427B4BF140FCC5AULL, "dae300_alistair_king6133802.cl"},
	{0x9427C18A9E4124C5ULL, "art\\characters\\creatures\\mubari\\animation\\cwkfl.ani"},
	{0x9427C4C799178642ULL, "vxw_bonepit_211_0_0.vlm"},
	{0x9427C8F167792B96ULL, "237391733.ogg"},
	{0x9428845628835C84ULL, "vxlt_darktown_42h_0_0.vlm"},
	{0x9428FDEFE9F364E2ULL, "6180618_m.fxe"},
	{0x9429BE676353D6D4ULL, "vxlt_undercityencounter_12g_0_0.vlm"},
	{0x9429CCB08AC14EC1ULL, "609061630.ogg"},
	{0x9429CD63967F940DULL, "isa321_velasco6117634.cl"},
	{0x942A0ABD6EC7F635ULL, "dae000_tranquil_16165114.cl"},
	{0x942A0D757313DD2CULL, "vxlt_0_lowtown_n_e87_0_0.vlm"},
	{0x942A198729DC2670ULL, "hlac_dr_slavers_3215_9x.dds"},
	{0x942A2F4FF3928DA3ULL, "plt_cod_cha_meredith.plo"},
	{0x942A46A617682A56ULL, "vxca_caverns1_j5f_0_0.vlm"},
	{0x942A7B87730C2B2BULL, "dae000cr_merchant_tailor.utc"},
	{0x942ABA7B29B9ABA7ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_hcliff2_drelief.dds"},
	{0x942AC62AE028882BULL, "vxlt_0_lowtown_docks_a3_n_cxl_0_0.vlm"},
	{0x942ADD53F99D1A6FULL, "vxlt_0_lowtown_docks_a3_n_bfw_0_0.vlm"},
	{0x942C2AB389B40017ULL, "dae300_bran.bnk"},
	{0x942C3A86503757C6ULL, "hightown_male_cityguard_mob6170024.cl"},
	{0x942C40F5F2006A1BULL, "plt_zz_mcrpt_harvester.plo"},
	{0x942C4677B104AD72ULL, "400726964.ogg"},
	{0x942C740799C6DFB9ULL, "steam_clo.mao"},
	{0x942CF294F7960ECDULL, "w_woundedcoast_130_0.msh"},
	{0x942D0094F7962600ULL, "w_woundedcoast_130_0.mao"},
	{0x942D017281AE6468ULL, "vxlt_0_lowtown_docks_a3_d32_0_0.vlm"},
	{0x942D0494F7962D53ULL, "w_woundedcoast_130_0.mmh"},
	{0x942D2EC0D234A74AULL, "plt_zz_vrcpt_debug.nss"},
	{0x942D43B6820F6158ULL, "6059662_m.fxe"},
	{0x942D9EF6BDE0989CULL, "83359730.ogg"},
	{0x942DBC98F01D019EULL, "dae000_party_night_comment6201632.cl"},
	{0x942E48E1DEA92704ULL, "1046749795.ogg"},
	{0x942EBEBAF75CDFEFULL, "dae280ar_deep_roads.ncs"},
	{0x942EF0C1A311EEAAULL, "sky_lt_gallowstemplar_n.mao"},
	{0x942F0EF7E10DFE08ULL, "6070578_f.fxe"},
	{0x942F851BB76853AAULL, "llac_lt_gallowscourtyard_-65535_1.dds"},
	{0x942FCD51C8BF7F5CULL, "174477253.ogg"},
	{0x942FDDCBB94C005BULL, "lowtown_fem_lowlife_mob.bnk"},
	{0x942FEE509FA5CCE8ULL, "vxlt_0_lowtown_docks_a3_apj_0_0.vlm"},
	{0x94300CB389B7B62FULL, "dae300_bran.cnv"},
	{0x94300D7B24637D00ULL, "zz_arishock_angry_v1.cut"},
	{0x9430265273CA1A48ULL, "6096516_f.fxe"},
	{0x943057338863578DULL, "zz_ave_debug6066165.cl"},
	{0x9430A048F348E5D1ULL, "580433972.ogg"},
	{0x9430E49F898D2CAFULL, "632118530.ogg"},
	{0x9430EA3E63132F15ULL, "follower_banter6172934.cl"},
	{0x943109808BEB8DE0ULL, "934554420.ogg"},
	{0x943148A196A4ED7EULL, "6202210_f.fxe"},
	{0x943173644971D03AULL, "vxw_encounter2_camp_gk_0_0.vlm"},
	{0x94317686584B9058ULL, "6188652_m.fxe"},
	{0x9431AC3449EEF30EULL, "vxw_blightlands_13y_0_0.vlm"},
	{0x9431C8AD2BB1012DULL, "vxlt_0_lowtown_a2_er3_0_0.vlm"},
	{0x94320A9B8A41F6C3ULL, "vxht_0_hightown_b93_0_0.vlm"},
	{0x9432104CDDFEE918ULL, "gen00fl_anders6121219.cl"},
	{0x943238CF1D3E0880ULL, "dae000_arishok6075349.cl"},
	{0x94328C17BB561C36ULL, "lgt110_urn_guard6170681.cl"},
	{0x9432DE1023050BF8ULL, "vxdr_primevaltomb_4zs_0_1.vlm"},
	{0x943396CBEF918FC2ULL, "6129062_m.fxe"},
	{0x9433A3D3F446384DULL, "gen_im_cns_flu_isabela_03.uti"},
	{0x9433CE3F38E34E8BULL, "vxht_0_hightown_a2_90c_0_0.vlm"},
	{0x9433DFCBB94FEC1BULL, "lowtown_fem_lowlife_mob.cnv"},
	{0x94349C62153B6AE6ULL, "6094784_m.fxe"},
	{0x9434BEBAF7614B3EULL, "dae280ar_deep_roads.lst"},
	{0x9434D8AA11A148A7ULL, "hlac_lt_0_lowtown_a3_n_c_-2_1x.dds"},
	{0x943555583B8684E1ULL, "gen00fl_aveline6121949.cl"},
	{0x9435667043E9A888ULL, "vxht_brothel_n_4h9_0_0.vlm"},
	{0x9435E7A1FE930C82ULL, "vxw_woundedcoast_1iq_0_0.vlm"},
	{0x9436A44582C37EF6ULL, "365843627.ogg"},
	{0x9436F6331A797504ULL, "vxlt_darktown_49i_0_0.vlm"},
	{0x9436F8F3D3B29562ULL, "6141899_m.fxe"},
	{0x94379FF773124268ULL, "vxca_caverns1_drc_0_0.vlm"},
	{0x9437A9173DDCFA40ULL, "6037385_m.fxe"},
	{0x9437E8D7A0F5BDF7ULL, "6076293_m.fxe"},
	{0x94381B0276F08E95ULL, "104332875.ogg"},
	{0x943836A696D7EE2EULL, "6114026_m.fxe"},
	{0x943848C884C14325ULL, "vxht_brothel_n_501_0_0.vlm"},
	{0x94389AABD8BCDE9CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a.phy"},
	{0x9438C79C316B44A2ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appc_0d.dds"},
	{0x9439206D12EB4E35ULL, "qun211_braggart6075442.cl"},
	{0x943928161A0B07F6ULL, "mag311_emile_de_launcet6043676.cl"},
	{0x9439564E33ACCB50ULL, "217565398.ogg"},
	{0x94395714C3CDA70EULL, "dae300_alistair_king6133608.cl"},
	{0x943966BBD221F5A9ULL, "qun100_mage_party_comment6019109.cl"},
	{0x943A32AE085970E3ULL, "vxht_mansion_n_3ss_0_0.vlm"},
	{0x943A3A1BB089B87EULL, "vxw_encounter3_b3_0_0.vlm"},
	{0x943B1458A8BB786BULL, "vxht_playershouse_n_1k0_0_0.vlm"},
	{0x943B734F73A47CE8ULL, "vxlt_0_lowtown_docks_a3_d43_0_0.vlm"},
	{0x943B970CC8EF9543ULL, "art\\vfx\\combat\\vfx_sg_slam_act_c.anb"},
	{0x943BA317FC087E92ULL, "vxht_bartrand_n_69f_0_0.vlm"},
	{0x943C5C82DB6B094EULL, "66955913.ogg"},
	{0x943D39F1617B6541ULL, "58655686.ogg"},
	{0x943DC4D2E457943FULL, "hlac_lt_0_lowtown_a3_n_c_-65537_3z.dds"},
	{0x943E6E14EE460586ULL, "6115331_m.fxe"},
	{0x943F1F381F6A726BULL, "vxw_bonepitmining_a2_1zt_0_0.vlm"},
	{0x943F5216AF9D557FULL, "vxca_caverns2prc_a2w_0_0.vlm"},
	{0x943FC23801E9FDFBULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_visa_0d.dds"},
	{0x943FE5413B67E8BAULL, "vxw_encounter2_camp_bd_0_0.vlm"},
	{0x9440B0397175EDC8ULL, "874970445.ogg"},
	{0x9441464C7C125921ULL, "6094897_m.fxe"},
	{0x9441624F3DAAFF16ULL, "gen00fl_aveline6152582.cl"},
	{0x944166AC78E04A61ULL, "mrl351cr_fakeharshal.utc"},
	{0x9442C1B5E95E5737ULL, "vxlt_darktown_5ca_0_0.vlm"},
	{0x9443381E976767E7ULL, "186154514.ogg"},
	{0x94435E4C1F4E919EULL, "6101204_m.fxe"},
	{0x9443903238320B38ULL, "vxht_mansion2_38b_0_0.vlm"},
	{0x9443DC03795A1CC2ULL, "hlac_lt_hangedman_n_-1_7y.dds"},
	{0x94442303D6DBE344ULL, "vxca_cavern1_haunted_fn0_0_0.vlm"},
	{0x9444403E3939546DULL, "hlac_ca_caverns2prc_5802_14z.dds"},
	{0x9444741DC9291545ULL, "vxca_caverns2_74l_0_0.vlm"},
	{0x944498F754DFC1E5ULL, "341510467.ogg"},
	{0x94450ACAC8D3AF7DULL, "l_dre180ar_ruins_entrance.bnk"},
	{0x944563EE2E40FE27ULL, "mrl351cr_fakemarethari.utc"},
	{0x9445777140E004B8ULL, "vxw_encounter3_gj_0_0.vlm"},
	{0x9445E36B18CD21CCULL, "6090250_m.fxe"},
	{0x9446076568B650BCULL, "6126717_m.fxe"},
	{0x9446A88388CE06AEULL, "6113126_m.fxe"},
	{0x9446B52C2A489B45ULL, "zz_lgt_2_debug6174316.cl"},
	{0x9447525019B17CA1ULL, "gen00fl_aveline6131229.cl"},
	{0x94478F7D03414CEDULL, "llac_w_bonepitdestroyed_a3_458752_0.dds"},
	{0x9447FD476A9439C5ULL, "669748526.ogg"},
	{0x9448683D056DAA9AULL, "934800745.ogg"},
	{0x9448E064AC89C5E6ULL, "vxdr_primevaltomb_7yj_0_0.vlm"},
	{0x94492084FE22B7F9ULL, "vxlt_warehouse_107_0_0.vlm"},
	{0x944973710494AE07ULL, "qcr220_grey_wardens6060417.cl"},
	{0x9449740CC8FB976CULL, "art\\vfx\\combat\\vfx_sg_slam_act_c.mmh"},
	{0x944996F7B68B66D2ULL, "6113881_m.fxe"},
	{0x9449AE7FD36AC944ULL, "mer240ip_sabins_stash.utp"},
	{0x944A2757F93BCC2BULL, "26945179.ogg"},
	{0x944A38FA0F8B5219ULL, "hlac_lt_0_lowtown_n_65536_1y.dds"},
	{0x944A4548A0F48374ULL, "343255484.ogg"},
	{0x944A65834116B7A2ULL, "gencr_talvashoth_warrior.utc"},
	{0x944AADEBFB80B3EFULL, "vxlt_0_lowtown_docks_a3_n_aq9_0_0.vlm"},
	{0x944AE592210ABDA6ULL, "4609285.ogg"},
	{0x944B7621C516D849ULL, "275988080.ogg"},
	{0x944B7760F5D16C20ULL, "6069399_m.fxe"},
	{0x944B8B58263045B7ULL, "6180454_f.fxe"},
	{0x944BB5185DB26DFAULL, "lgt220_mag_tmplr_templar6140244.cl"},
	{0x944BCBD323B38DF5ULL, "isa211_isabela6072603.cl"},
	{0x944C1DFE7CBD66CAULL, "llac_lt_0_lowtown_n_65537_0.dds"},
	{0x944C2C316DF712C5ULL, "372074864.ogg"},
	{0x944C742C978B762DULL, "w_woundedcoast_190_0.mmh"},
	{0x944C782C978B7D7EULL, "w_woundedcoast_190_0.mao"},
	{0x944C7EAC214156F3ULL, "mag110_arianni6016770.cl"},
	{0x944C800F10AC2856ULL, "hlac_dr_0_deeproads_5845_0z.dds"},
	{0x944C862C978B94B3ULL, "w_woundedcoast_190_0.msh"},
	{0x944CB0A1ED8243BCULL, "6118110_m.fxe"},
	{0x944CC16869BAB996ULL, "6017424_m.fxe"},
	{0x944D41C525D10DADULL, "633357435.ogg"},
	{0x944D911511608AEBULL, "vxw_bonepitmining_a2_1ww_0_0.vlm"},
	{0x944DB33313CEE7EAULL, "follower_banter6161287.cl"},
	{0x944E1D3897A756D9ULL, "vxca_caverns1_mines_efv_0_0.vlm"},
	{0x944E55E14A5DCD82ULL, "294740837.ogg"},
	{0x944E79CBE9808703ULL, "vxht_0_hightown_a2_av5_0_0.vlm"},
	{0x944EAC670F9D322DULL, "vxlt_0_lowtown_a2_ex1_0_0.vlm"},
	{0x944FD3C43B3BB709ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.cess.ani"},
	{0x9450BD397C4E203EULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.dura.evt"},
	{0x9451211C81EA1858ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_shd_reva_0n.dds"},
	{0x9451763F5982AFD1ULL, "llac_lt_0_lowtown_docks_a3_n_1_0.dds"},
	{0x9451C96E4552A57EULL, "w_bonepit_12_0_ll.mmh"},
	{0x9452975B590E0851ULL, "6120053_m.fxe"},
	{0x9452D0B6ADB0F819ULL, "6125308_m.fxe"},
	{0x94539E2192FC02CDULL, "vxlt_gallowstemplar_f_475_0_0.vlm"},
	{0x9453AB475746D4E7ULL, "6020536_f.fxe"},
	{0x94540BA7C352F2E8ULL, "default_gray50.dds"},
	{0x94547A90C3F0A525ULL, "vxht_0_hightown_a2_c58_0_0.vlm"},
	{0x9454AFBA44AD3E2DULL, "vxlt_house_he_0_0.vlm"},
	{0x9454D4E1EAD020C5ULL, "hlac_ht_chantry_n_2534_21z.dds"},
	{0x9454E9D34AA0E627ULL, "271350171.ogg"},
	{0x94556B05CEEE08FDULL, "561196807.ogg"},
	{0x9455DC712FBD2C9CULL, "vxdr_primevalentrance_27g_0_0.vlm"},
	{0x9455FBE61112A0DAULL, "237239563.ogg"},
	{0x9456228551D827CFULL, "862035429.ogg"},
	{0x94564B00964AB3CBULL, "360089621.ogg"},
	{0x945687F87A9DDB1FULL, "vxlt_gallowstemplar_45n_0_0.vlm"},
	{0x94568D576C01C090ULL, "6084909_m.fxe"},
	{0x9456EEC43B42038BULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.cess.gad"},
	{0x94575AC7E2E00B90ULL, "dae000ip_crowd_mf1_talk5.utp"},
	{0x9457990E7F6D2612ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml105a_0n.dds"},
	{0x9457D4397C546669ULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.dura.gad"},
	{0x9458C880D456C128ULL, "6208341_m.fxe"},
	{0x945901E4A686DA56ULL, "vxdr_primevalentrance_226_0_0.vlm"},
	{0x94594C223C4165FEULL, "gen_im_npc_dwf_arm_boo_lgtb.uti"},
	{0x9459832A3B62153AULL, "pro000_friedrich6051059.cl"},
	{0x94599053BC22C2F7ULL, "vxw_bonepit_72_0_0.vlm"},
	{0x9459C2E52F3CAF98ULL, "hlac_dr_dungeon_3216_4x.dds"},
	{0x945A53B36F70CDA5ULL, "vxlt_0_lowtown_dbb_0_0.vlm"},
	{0x945AFD664373607CULL, "744144952.ogg"},
	{0x945B1423919A756FULL, "vxht_bartrand_n_5wv_0_0.vlm"},
	{0x945B2DE22D239207ULL, "vxlt_hangedman_n_241_0_0.vlm"},
	{0x945B7EAF03B670FCULL, "777709667.ogg"},
	{0x945BAD0414C77932ULL, "560110432.ogg"},
	{0x945BB5984CC8A4B3ULL, "llac_w_encounter2_camp_5_0.dds"},
	{0x945BD2EF7E3870EDULL, "w_blightlands_19_0.msh"},
	{0x945BE0EF7E388820ULL, "w_blightlands_19_0.mao"},
	{0x945BE4EF7E388F73ULL, "w_blightlands_19_0.mmh"},
	{0x945C101D250AA4A7ULL, "hlac_w_encounter2_camp_262145_0y.dds"},
	{0x945CC1124F6DA565ULL, "975965976.ogg"},
	{0x945CE79BDCF23CFEULL, "6035284_m.fxe"},
	{0x945CEBA8DB769F83ULL, "vxht_0_hightown_a2_aq4_0_0.vlm"},
	{0x945CF77CEEE19748ULL, "6131145_f.fxe"},
	{0x945D5BF09784D340ULL, "hlac_lt_gallowscourtyard_-131072_0y.dds"},
	{0x945D62AD1DDD1F6EULL, "6207777_m.fxe"},
	{0x945D75C836F46E03ULL, "vxw_encounter2_kz_0_0.vlm"},
	{0x945DEBC43B481DF0ULL, "art\\vfx\\combat\\vfx_v_fir_cast_aoe_c.cess.evt"},
	{0x945E0FF520B401EDULL, "vxw_bonepit_4y_0_0.vlm"},
	{0x945E1A99C8EB21A3ULL, "w_encounter2_camp_15_0.mao"},
	{0x945E1E99C8EB28F0ULL, "w_encounter2_camp_15_0.mmh"},
	{0x945E2C99C8EB400AULL, "w_encounter2_camp_15_0.msh"},
	{0x945EA5397C5A34B7ULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.dura.ani"},
	{0x945EEAC5FF857F70ULL, "vxlt_gallowstemplar_f_5jb_0_0.vlm"},
	{0x945F0F57061E3853ULL, "6091694_m.fxe"},
	{0x945F53193BC2EBEDULL, "6152698_m.fxe"},
	{0x945F7C20762E3A1AULL, "33240808.ogg"},
	{0x945F88A0F2BF52D6ULL, "lt_0_lowtown_a2_34_0_lc.mmh"},
	{0x945FCC3FABE86F0BULL, "vxca_cavern1_haunted_gki_0_0.vlm"},
	{0x945FDA9711504A39ULL, "ts_circleofmagi_wpn_table.gda"},
	{0x94601D0346544401ULL, "414244170.ogg"},
	{0x946073A27CE48AD6ULL, "hlac_w_sundermount_-65531_0x.dds"},
	{0x9461054F13BD03E8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galloalcovlo.mmh"},
	{0x9461114F13BD178BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galloalcovlo.mao"},
	{0x9461234F13BD3612ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galloalcovlo.msh"},
	{0x94612D6C9741B6E6ULL, "672649868.ogg"},
	{0x946142939FA71099ULL, "6122208_m.fxe"},
	{0x946184746BF9B016ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medc_0t.dds"},
	{0x9461F4D3A15CAF88ULL, "795408197.ogg"},
	{0x946247C96EA1DC2AULL, "gen00fl_merrill6128920.cl"},
	{0x94626B68872DD278ULL, "171968088.ogg"},
	{0x946294F5BE699E3EULL, "6169912_m.fxe"},
	{0x9462A65F02E9B397ULL, "ene_dust.mao"},
	{0x9462D97432B036D5ULL, "dae000_bartrand6030676.cl"},
	{0x9462FF1814BF07AEULL, "rdr110cr_guard_warehouse.utc"},
	{0x946338581CC40EBAULL, "plt_zz_vault_debug.plo"},
	{0x94636741BFA5A545ULL, "400226308.ogg"},
	{0x9464499EDC03DC09ULL, "zz_followers_banter6158499.cl"},
	{0x94644D3A1CDE263DULL, "zz_dae_debug6029083.cl"},
	{0x94646D875CA9AB03ULL, "w_woundedcoast_83_0.rml"},
	{0x94646DA4A5C5891AULL, "vxlt_hangedman_n_10g_0_0.vlm"},
	{0x9465EE73FD7ACA3CULL, "6135282_f.fxe"},
	{0x94663C214CA551D9ULL, "vxht_0_hightown_a2_bzt_0_0.vlm"},
	{0x9466B9C31E6C1C7CULL, "vxdr_0_deeproads_cuo_0_1.vlm"},
	{0x9466FDA84C4F864CULL, "hlac_lt_0_lowtown_n_196606_0z.dds"},
	{0x946708F5C495E3ADULL, "721951808.ogg"},
	{0x9467866D69C96FFBULL, "dae000_quintus6153130.cl"},
	{0x9467A2BDE9F0B938ULL, "vis121_saemus6099487.cl"},
	{0x9467C23488BD6ECAULL, "6202200_m.fxe"},
	{0x9467CD85DEA40FC0ULL, "6043635_f.fxe"},
	{0x946822D141B6BE6FULL, "plt_act00pt_main.plo"},
	{0x94685A00D720FB99ULL, "hlac_w_encounter3_-131067_0y.dds"},
	{0x94690EA73D20A281ULL, "llac_lt_0_lowtown_docks_a3_-262144_0.dds"},
	{0x9469563DCE1CF890ULL, "genip_pp_lti_one_shot.utp"},
	{0x94699FBF1F19AA87ULL, "vxlt_hangedman_n_230_0_0.vlm"},
	{0x9469C625882366D0ULL, "vxlt_undercityserialkiller_1sx_0_0.vlm"},
	{0x9469D44E81BAA476ULL, "683812509.ogg"},
	{0x9469D70191E3475AULL, "270657335.ogg"},
	{0x9469EFEE78E156E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_turn_away_right.gad"},
	{0x946A223D7E6AF21EULL, "marksman_selfsong.dds"},
	{0x946A516BFBA066D6ULL, "vxht_0_hightown_n_a89_0_0.vlm"},
	{0x946A6CA72E938DB6ULL, "6107282_m.fxe"},
	{0x946ADB6C21B68C37ULL, "hlac_ht_brothel_n_2995_5z.dds"},
	{0x946B287EA78A2FB8ULL, "mag200_feynriel.cnv"},
	{0x946BBF83729EFCADULL, "6216985_m.fxe"},
	{0x946C249642B29018ULL, "6107169_m.fxe"},
	{0x946C5C360E79FA6BULL, "dae000_denier6162518.cl"},
	{0x946C88D2E57BFDB8ULL, "vxlt_warehouse_sr_0_0.vlm"},
	{0x946D093B7F228F08ULL, "6163388_f.fxe"},
	{0x946D58E4C7D808F1ULL, "zz_vis_debug6050339.cl"},
	{0x946D901175C9124BULL, "319540104.ogg"},
	{0x946E0A7EA78C32C0ULL, "mag200_feynriel.bnk"},
	{0x946E174CBF40A0B7ULL, "vxlt_darktown_5rb_0_0.vlm"},
	{0x946E89D344C5DC60ULL, "368100006.ogg"},
	{0x946EB4CFDB11FEA1ULL, "6083365_m.fxe"},
	{0x946F00FB826F1294ULL, "vxht_keep_621_0_0.vlm"},
	{0x946F182E7FC3D9A1ULL, "6144266_m.fxe"},
	{0x946F38B91DA145B8ULL, "ser121_jethann6104620.cl"},
	{0x946FB2D322223D42ULL, "273485122.ogg"},
	{0x946FD636AF13B415ULL, "724071343.ogg"},
	{0x946FE4EE78E5AFF4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_turn_away_right.evt"},
	{0x94701489AAD90144ULL, "6102768_m.fxe"},
	{0x94705300789A98DFULL, "6127191_m.fxe"},
	{0x94707423D23D069FULL, "art\\vfx\\materials\\vfx_crack_decal_add_02.mao"},
	{0x9470961DA6009474ULL, "195589928.ogg"},
	{0x9470CEE85ED2978EULL, "6032076_m.fxe"},
	{0x9470DDA267C30DDDULL, "vxlt_0_lowtown_a2_n_cmw_0_0.vlm"},
	{0x9470F13CB309C767ULL, "mcr371_orsino6083361.cl"},
	{0x9471A579B62600E8ULL, "dre000_hireling16038587.cl"},
	{0x9471B6E35028A71FULL, "zz_mer_debug6073695.cl"},
	{0x94728364A84F0838ULL, "gen00fl_aveline6139965.cl"},
	{0x94728B7E94D15B8CULL, "6137621_m.fxe"},
	{0x9472DA9F48E5DC73ULL, "vxht_chantry_28l_0_0.vlm"},
	{0x9472DEE477A45CE4ULL, "vxlt_0_lowtown_fex_0_0.vlm"},
	{0x947307335EA392B3ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.cess.evt"},
	{0x94732B398D4FDC39ULL, "988479448.ogg"},
	{0x94733B21E269A5B2ULL, "qcr220_meredith6041549.cl"},
	{0x9474B6EA8D96E645ULL, "ev_cl_blend.mao"},
	{0x947512CDB1B031E5ULL, "6088250_m.fxe"},
	{0x9475932AF8705618ULL, "613015039.ogg"},
	{0x94767381689474D7ULL, "546417088.ogg"},
	{0x94767F2359B8DD33ULL, "art\\vfx\\combat\\vfx_w_tmp_silence_activ_c.impa.ani"},
	{0x94768F3FAC3BA96BULL, "6031388_m.fxe"},
	{0x9477024DE09DF73EULL, "mcr310ip_to_docks.utp"},
	{0x947706CCC9070849ULL, "6172924_m.fxe"},
	{0x947715DFA9E0AB1DULL, "hightown_market_guard_sta6169672.cl"},
	{0x947721B5157CEBDEULL, "vxht_mansion2_2fg_0_0.vlm"},
	{0x94778E78BA9EFED8ULL, "vxht_mansion_n_3wp_0_0.vlm"},
	{0x947822BBD502B6BDULL, "6036656_f.fxe"},
	{0x94782405144C8139ULL, "6132928_m.fxe"},
	{0x9478633E93BD5DE5ULL, "vxlt_darktown_4zu_0_0.vlm"},
	{0x947871B7125A2EC0ULL, "118352871.ogg"},
	{0x94789FB0C193F1C9ULL, "hla_ht_0_hightown_-65536_0y.dds"},
	{0x9478D79F9BC7FC2FULL, "vxw_woundedcoast_1q8_0_0.vlm"},
	{0x947936D0A4856140ULL, "hlac_w_sundermount_-196604_0x.dds"},
	{0x9479F18EF21DD69FULL, "hlac_lt_0_lowtown_docks_n_-131076_1z.dds"},
	{0x947A00335EA9A5F4ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.cess.gad"},
	{0x947ADC77C3E8E236ULL, "follower_banter6160987.cl"},
	{0x947B07FD1F3E00EBULL, "564199948.ogg"},
	{0x947C215730FB4916ULL, "6148136_f.fxe"},
	{0x947DDCEE78F1DF2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_turn_away_right.ani"},
	{0x947DFF0F845ECD3CULL, "zz_suhwan_test6202888.cl"},
	{0x947E7895BBA9430EULL, "llac_lt_gallowsprison_n_-1_5.dds"},
	{0x947E839425DF8B8BULL, "lt_0_lowtown_n_11_0.anb"},
	{0x947E899FBC780B9EULL, "846166173.ogg"},
	{0x947EB626A4DF2A48ULL, "one100dg_nabil6033124.cl"},
	{0x947F0170E4371F68ULL, "art\\characters\\weapons\\staff\\w_stf_st05a.msh"},
	{0x947F1370E4373DF9ULL, "art\\characters\\weapons\\staff\\w_stf_st05a.mao"},
	{0x947F1F70E4375192ULL, "art\\characters\\weapons\\staff\\w_stf_st05a.mmh"},
	{0x947F6A6EC31B9A16ULL, "vxht_0_hightown_n_ca0_0_0.vlm"},
	{0x947FC7D524019C81ULL, "543539145.ogg"},
	{0x947FDFA611AEE32DULL, "vxdr_primevalentrance_337_0_0.vlm"},
	{0x9480B726E5A640BAULL, "vxlt_0_lowtown_a3_n_c_gy3_0_0.vlm"},
	{0x9480B94B2FE1B8D4ULL, "one110_meeran6089497.cl"},
	{0x94810F335EAFDDD2ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.cess.ani"},
	{0x948113D8BF3FCD7FULL, "6113376_m.fxe"},
	{0x94818AB606D0157DULL, "6153873_m.fxe"},
	{0x9481DA87CBC35935ULL, "916006142.ogg"},
	{0x948205D9ED409A25ULL, "character_character_p165.psh.fxo.dx11.assembly"},
	{0x94828C277F1F695BULL, "450547231.ogg"},
	{0x9482A49B84892A54ULL, "6155544_m.fxe"},
	{0x9482D18ABD30439BULL, "framebuffereffects_motionblur_p0.psh.fxo.dx11"},
	{0x948352BA11649EDBULL, "667127725.ogg"},
	{0x9483EE96ED512E29ULL, "qun131_taskmaster6059071.cl"},
	{0x948425D10B2FCF99ULL, "vxca_cavern1_haunted_edh_0_0.vlm"},
	{0x94845A174DE820EDULL, "6125322_m.fxe"},
	{0x9484772359C50D6AULL, "art\\vfx\\combat\\vfx_w_tmp_silence_activ_c.impa.evt"},
	{0x9484B9E4F82E4FDBULL, "vxca_caverns1_i2u_0_0.vlm"},
	{0x94855DD0C11F442EULL, "vxlt_0_lowtown_docks_n_ctf_0_0.vlm"},
	{0x948578A9BAFD20C9ULL, "6175824_m.fxe"},
	{0x948608E327F93C65ULL, "742290241.ogg"},
	{0x94861477715559FFULL, "hlac_w_encounter3_-65538_0z.dds"},
	{0x94862DC5D49F3180ULL, "469664443.ogg"},
	{0x948641B5551B2E40ULL, "501895965.ogg"},
	{0x948653FCCA79B883ULL, "dae100_alistair_drunk6135110.cl"},
	{0x9486F507D9DCCE67ULL, "vxlt_gallowscourtyard_n_6yy_0_0.vlm"},
	{0x948724DE9FB9785EULL, "gen_im_cft_reg_f2.uti"},
	{0x948772B10961AA20ULL, "vxca_cave_29x_0_0.vlm"},
	{0x9487CC5BAB1B1D79ULL, "855961364.ogg"},
	{0x9487DB63A6B25A2EULL, "qun221_petrice6077325.cl"},
	{0x948861153F4DFF46ULL, "gen_im_cft_rec_wrn_02.uti"},
	{0x948873E0B4FF150AULL, "6201568_m.fxe"},
	{0x948888A55DDAF4D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1_reverse.ani"},
	{0x94889216C8159DD1ULL, "ser221_gamlen6162349.cl"},
	{0x9488F3AC650AB104ULL, "vxlt_warehouse_n_ck_0_0.vlm"},
	{0x94891C0F5136005CULL, "6018907_f.fxe"},
	{0x94894D731EAF2825ULL, "vxw_bonepit_1pn_0_0.vlm"},
	{0x94896C8CC9682EB8ULL, "vxlt_warehouse_yp_0_0.vlm"},
	{0x948A5B2BE54BB680ULL, "28563732.ogg"},
	{0x948AD85F6A99BFD2ULL, "llac_lt_0_lowtown_a2_n_65534_5.dds"},
	{0x948B562359CAF3DDULL, "art\\vfx\\combat\\vfx_w_tmp_silence_activ_c.impa.gad"},
	{0x948B5E5EA1997180ULL, "6081713_m.fxe"},
	{0x948B806A3B52BCCDULL, "vxlt_gallowscourtyard_6wv_0_0.vlm"},
	{0x948BC9C32289FEFDULL, "6089903_m.fxe"},
	{0x948BCBD0BE8D9FECULL, "lt_0_lowtown_a3_n_c_29_0_lc.mmh"},
	{0x948BE070C57A8339ULL, "mrl210_merrill6116337.cl"},
	{0x948BE4B5665B4394ULL, "vxht_keep_687_0_0.vlm"},
	{0x948C07E1E035B70AULL, "103761880.ogg"},
	{0x948C57987E5803D8ULL, "586238657.ogg"},
	{0x948D097B9A63F15DULL, "ca_cave_2240.rml"},
	{0x948DCA12AB4C684EULL, "vxlt_darktown_534_0_0.vlm"},
	{0x948E0E2F1A91D067ULL, "llac_lt_0_lowtown_a3_n_c_-1_6.dds"},
	{0x948E29E94F2B737FULL, "ser221_gamlen6135277.cl"},
	{0x948F3832210BF040ULL, "6047378_f.fxe"},
	{0x948F85A55DE10E36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1_reverse.gad"},
	{0x948FAC896089A21BULL, "86584866.ogg"},
	{0x948FF474BC2D6E98ULL, "hlac_ht_bartrand_n_2995_6z.dds"},
	{0x949027A7620CDC82ULL, "6076013_f.fxe"},
	{0x94903EEA0CCDE94DULL, "hlac_lt_0_lowtown_-65536_3y.dds"},
	{0x94907926F365CD0DULL, "zz_ave_debug6064149.cl"},
	{0x94908B98696CFD44ULL, "plt_rumor_blackpowder.nss"},
	{0x94909BDC85DA3023ULL, "6127890_m.fxe"},
	{0x94909BDE4C1FAD65ULL, "mer120_hubert6059395.cl"},
	{0x949170207154E7B2ULL, "6211921_m.fxe"},
	{0x9491DB2CB1A7BCB0ULL, "6084255_m.fxe"},
	{0x9492592A6B4FB296ULL, "hlac_dr_primevaltomb_664_4x.dds"},
	{0x949278610F3F409EULL, "526124536.ogg"},
	{0x949286CE1147680EULL, "6126575_m.fxe"},
	{0x9492EE13496A0B62ULL, "vxht_keep_n_3ip_0_0.vlm"},
	{0x9493BC6657C2003EULL, "hlac_lt_undercitydungeon_-1_21z.dds"},
	{0x9493C40BEE7B09B4ULL, "gen_im_trs_gen_coinbag_04.uti"},
	{0x9493CFADB3155CAEULL, "vxlt_0_lowtown_docks_n_cog_0_0.vlm"},
	{0x9493D010500211A6ULL, "gen00fl_anders6119283.cl"},
	{0x94944DA0669F3596ULL, "vxlt_gallowstemplar_43q_0_0.vlm"},
	{0x949499883B1FAB76ULL, "952586267.ogg"},
	{0x9494CE8AEC63E0FBULL, "vxlt_house_dw_0_0.vlm"},
	{0x94951B7ECE7F70D9ULL, "254482785.ogg"},
	{0x9495619094C908AFULL, "ser270_emeric6113632.cl"},
	{0x9495A0A55DE5A795ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1_reverse.evt"},
	{0x949632B664DF336DULL, "748817842.ogg"},
	{0x949724932F173436ULL, "hlac_lt_0_lowtown_docks_a3_65532_0x.dds"},
	{0x94975B875CD4E61AULL, "w_woundedcoast_83_0.msh"},
	{0x949769875CD4FE53ULL, "w_woundedcoast_83_0.mao"},
	{0x94976D875CD50560ULL, "w_woundedcoast_83_0.mmh"},
	{0x949797DC0F203F6AULL, "one100dg_kelder6055433.cl"},
	{0x9497BF5010A540A5ULL, "vxw_bonepit_1um_0_0.vlm"},
	{0x9497BFD97038838EULL, "591789025.ogg"},
	{0x9497ED711F3CE07FULL, "vxw_woundedcoast_1fy_0_0.vlm"},
	{0x9497FC37EC87C511ULL, "dae011_gossips6149745.cl"},
	{0x94981E99C91CA0B3ULL, "w_encounter2_camp_15_0.rml"},
	{0x94998C16231018E3ULL, "vxca_caverns1_j1z_0_1.vlm"},
	{0x9499D03B938F8A00ULL, "6084413_m.fxe"},
	{0x9499F2472D48D54DULL, "vxlt_gallowscourtyard_6ru_0_0.vlm"},
	{0x949A03185CE024BFULL, "zz_dae_debug6049951.cl"},
	{0x949A392E32C06A7AULL, "llac_lt_0_lowtown_-131072_4.dds"},
	{0x949A7F36A1DA974EULL, "vxw_woundedcoast_1gw_0_0.vlm"},
	{0x949ADE79DFA6A6B3ULL, "hlac_lt_gallowstemplar_2383_1z.dds"},
	{0x949B68611C1F7E15ULL, "6065296_f.fxe"},
	{0x949BC6E563994945ULL, "vxht_chantry_n_270_0_0.vlm"},
	{0x949BF1CDEA8BA9B1ULL, "hlac_w_woundedcoast_6_0x.dds"},
	{0x949C4F16512A2015ULL, "hlac_lt_0_lowtown_a2_-196608_0y.dds"},
	{0x949C875BCB81B5F0ULL, "6108894_m.fxe"},
	{0x949CE3C44524B40BULL, "llac_ht_0_hightown_-65536_4.dds"},
	{0x949D340C58A72A49ULL, "173188863.ogg"},
	{0x949D6FF010617721ULL, "vxw_bonepitdestroyed_a3_6w_0_0.vlm"},
	{0x949D9BAFE3F92D6BULL, "6177860_m.fxe"},
	{0x949EB2908E7D3327ULL, "288314324.ogg"},
	{0x949EBE630DE7DA7FULL, "6093865_m.fxe"},
	{0x949EC3CF215EE5FAULL, "art\\levels\\races\\design\\prp_bodybagb.msh"},
	{0x949ED1CF215EFD33ULL, "art\\levels\\races\\design\\prp_bodybagb.mao"},
	{0x949ED5CF215F0440ULL, "art\\levels\\races\\design\\prp_bodybagb.mmh"},
	{0x949ED68F97F5986EULL, "vxlt_0_lowtown_gb2_0_0.vlm"},
	{0x949EDBEDA513F811ULL, "10496029.ogg"},
	{0x949F2599C922CB0BULL, "w_encounter2_camp_15_0.phy"},
	{0x949FEE24BF90DCEFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_dismissive.gad"},
	{0x94A00395147DB3D9ULL, "vxht_0_hightown_a2_bnx_0_0.vlm"},
	{0x94A031BB00CA3D19ULL, "vxht_chantry_n_2cx_0_0.vlm"},
	{0x94A03D80B25A866EULL, "942985629.ogg"},
	{0x94A06A2D1EEB9A9FULL, "hlac_w_blightlands_393212_0x.dds"},
	{0x94A0AA15E334A7CCULL, "art\\effects\\cloudshadows\\w_blightlands\\cloud2.dds"},
	{0x94A0C496F4EDE35BULL, "vxw_encounter3_31_0_0.vlm"},
	{0x94A1155784E74A24ULL, "gen_im_arm_boo_lgt_set_med.uti"},
	{0x94A13BDA2E70F129ULL, "296931320.ogg"},
	{0x94A25CCB2C424911ULL, "vxlt_0_lowtown_a2_n_c_gpn_0_0.vlm"},
	{0x94A2BB3D4CF43F44ULL, "bld_drian_tu5.dds"},
	{0x94A2BF7D58954E16ULL, "vxlt_gallowstemplar_46r_0_0.vlm"},
	{0x94A3150C6129CFA4ULL, "hlac_ht_0_hightown_a2_-65537_1y.dds"},
	{0x94A34F872C09B5D8ULL, "gen00fl_isabela6090809.cl"},
	{0x94A357F1B54B3BA9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward.ani"},
	{0x94A386A881103D3EULL, "76220528.ogg"},
	{0x94A3B8D5699FA765ULL, "vxlt_darktown_4kr_0_0.vlm"},
	{0x94A4B05431FD79FBULL, "gen00fl_fenris6126943.cl"},
	{0x94A52D7463EB9B0DULL, "vxdr_0_deeproads_fod_0_0.vlm"},
	{0x94A549964CDC0250ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3bluelyrium.mao"},
	{0x94A54C62EDDDE57EULL, "follower_banter6155433.cl"},
	{0x94A54D964CDC0963ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3bluelyrium.mmh"},
	{0x94A55B964CDC20BDULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3bluelyrium.msh"},
	{0x94A5A8F817F945A6ULL, "vxht_brothel_n_5ai_0_0.vlm"},
	{0x94A5FBC859CBFBA5ULL, "llac_lt_0_lowtown_docks_a3_-65539_3.dds"},
	{0x94A6B1798D9E0C1AULL, "664958619.ogg"},
	{0x94A70324BF971E9CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_dismissive.evt"},
	{0x94A784E92D5A55AFULL, "70141509.ogg"},
	{0x94A7EC0F188A4C48ULL, "718501844.ogg"},
	{0x94A8204958859216ULL, "hlac_dr_dungeon_3214_12y.dds"},
	{0x94A861252363F8CAULL, "vxca_cavern1_haunted_e54_0_0.vlm"},
	{0x94A864241F3EEDCDULL, "vxlt_gallowscourtyard_6yt_0_0.vlm"},
	{0x94A86902F7AFF4DAULL, "418638808.ogg"},
	{0x94A8B23C9027B42EULL, "qun231_varnell6099654.cl"},
	{0x94A8F64B65D79C7FULL, "68015720.ogg"},
	{0x94A908FB883F3FF2ULL, "hlac_lt_gallowstemplar_f_2393_1y.dds"},
	{0x94A9285080D560F9ULL, "lgt220_mag_tmplr_templar6140559.cl"},
	{0x94A96776F5F371ECULL, "971136847.ogg"},
	{0x94AA72F1B551882BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward.gad"},
	{0x94AA80942604C1D7ULL, "lt_0_lowtown_n_11_0.rml"},
	{0x94AAADCC8F38994EULL, "vxlt_darktown_596_0_0.vlm"},
	{0x94AAEFF5252DE8DFULL, "hlac_lt_0_lowtown_docks_a3_-3_6x.dds"},
	{0x94AB0B17AD60C721ULL, "202331963.ogg"},
	{0x94AB6474DA8A1C28ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyn_0t.dds"},
	{0x94AC18F3DE54773CULL, "36234857.ogg"},
	{0x94AC3325ADFF3A7AULL, "vxdr_primevaltomb_5cc_0_1.vlm"},
	{0x94AC476A990F06D6ULL, "vxca_caverns1_jfo_0_0.vlm"},
	{0x94AC49BFF84F7B24ULL, "vxht_0_hightown_n_car_0_0.vlm"},
	{0x94AC5A412CF1C4F7ULL, "815706217.ogg"},
	{0x94ACC298A62D1DD9ULL, "skin_tints.gda"},
	{0x94ACD12C2CA5DE18ULL, "w_encounter2_camp_49_0.phy"},
	{0x94AD4D21E4576358ULL, "614171954.ogg"},
	{0x94AD6B0780A23E21ULL, "hlac_lt_0_lowtown_65534_4y.dds"},
	{0x94AD780FBD610DBDULL, "6208178_f.fxe"},
	{0x94ADAED1C50B0FEAULL, "gen00fl_anders6128315.cl"},
	{0x94AE0A1AA1CA267AULL, "6090443_m.fxe"},
	{0x94AE132C3CE3ABF7ULL, "386065519.ogg"},
	{0x94AE31D95C9DE544ULL, "125788047.ogg"},
	{0x94AE4900B428B7E3ULL, "6105641_m.fxe"},
	{0x94AF03AA0EFD08EAULL, "vxw_blightlands_b3_0_0.vlm"},
	{0x94AF5C58FD5F09DDULL, "166936899.ogg"},
	{0x94B035CDAEEE79BEULL, "lowtown_male_beggar_sta6151058.cl"},
	{0x94B0575B04B6EB37ULL, "6060103_f.fxe"},
	{0x94B06FF1B555EE90ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkforward.evt"},
	{0x94B0BF7A3FAF3EF5ULL, "vxw_bonepitmining_a2_15c_0_0.vlm"},
	{0x94B116E2EAE4D813ULL, "6061929_m.fxe"},
	{0x94B14DCA8203A946ULL, "vxlt_hangedman_25t_0_0.vlm"},
	{0x94B22267AB645F40ULL, "6139384_m.fxe"},
	{0x94B28C986989E53FULL, "plt_rumor_blackpowder.plo"},
	{0x94B2FDB496ADA232ULL, "dae000_gamlen6005015.cl"},
	{0x94B341D141F6DE54ULL, "plt_act00pt_main.nss"},
	{0x94B3D02C2CABFB30ULL, "w_encounter2_camp_49_0.rml"},
	{0x94B3EF4A2EA9761BULL, "hlac_w_encounter3_-131076_0y.dds"},
	{0x94B40B24BFA1B6B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_dismissive.ani"},
	{0x94B413C64050BDA8ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_6y.dds"},
	{0x94B449433AECFA84ULL, "vxlt_warehouse_n_pn_0_0.vlm"},
	{0x94B48C8A1E885CE3ULL, "hlac_ht_chantry_n_2534_22y.dds"},
	{0x94B506A2728C5CF5ULL, "6129274_m.fxe"},
	{0x94B5383B6A1C76D7ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.dura.gad"},
	{0x94B55149381E724CULL, "llac_dr_dungeon_3214_11.dds"},
	{0x94B6132C4B8A4CDAULL, "216830235.ogg"},
	{0x94B649D1428B209AULL, "qun100_mage_party_comment6019096.cl"},
	{0x94B64E964CEA7D88ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_3bluelyrium.phy"},
	{0x94B6846E3D172B9BULL, "m_dae330ar_undercity.bnk"},
	{0x94B6B5D13BD5BDBEULL, "hlac_ht_0_hightown_n_2_0x.dds"},
	{0x94B6EE311078209FULL, "vxht_brothel_n_5j1_0_0.vlm"},
	{0x94B6FE84BA33EC81ULL, "mrl310_party_comments6142634.cl"},
	{0x94B70BAA7AD5079AULL, "295313906.ogg"},
	{0x94B740F6197FF443ULL, "vxw_bonepit_kj_0_0.vlm"},
	{0x94B764AD35FFE81BULL, "dae200_cullen6140518.cl"},
	{0x94B772EDC7F6B406ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumtile_sa.dds"},
	{0x94B7A74FD614C8CCULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb6.mmh"},
	{0x94B7AD4FD614D2BEULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb6.msh"},
	{0x94B7F0A465AFA947ULL, "llac_lt_0_lowtown_a3_n_c_196606_0.dds"},
	{0x94B7FCA4DAC70789ULL, "vxw_encounter2_bc_0_0.vlm"},
	{0x94B84B8764090588ULL, "character_character_p221.psh.fxo.dx11"},
	{0x94B86AB80C110C84ULL, "w_encounter2_camp_22_0.msh"},
	{0x94B87CB80C112BEDULL, "w_encounter2_camp_22_0.mao"},
	{0x94B880B80C113226ULL, "w_encounter2_camp_22_0.mmh"},
	{0x94B8D55E55F49CD0ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_whippingpost_sa.dds"},
	{0x94B9105539D9D9ADULL, "835281361.ogg"},
	{0x94B9453D03599185ULL, "qun120cr_javaris6044002.cl"},
	{0x94B99177CB9EEF2FULL, "vxht_0_hightown_a3_91m_0_0.vlm"},
	{0x94B9AAAA6181A99BULL, "gen_im_trs_gen_hvy_boo_r2.uti"},
	{0x94B9AC01380F6B8FULL, "411223776.ogg"},
	{0x94B9C5E7DE5138BCULL, "ser270_emeric6113233.cl"},
	{0x94BB0E4C51EF3407ULL, "503942350.ogg"},
	{0x94BB7172EA15B331ULL, "vxdr_0_deeproads_eum_0_0.vlm"},
	{0x94BB9921348FF9CBULL, "vxw_encounter2_camp_8s_0_0.vlm"},
	{0x94BBF4E60131EA98ULL, "6129131_f.fxe"},
	{0x94BC4D3B6A22B964ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.dura.evt"},
	{0x94BCFFB0E4241ED7ULL, "demo_weapon_swap6207189.cl"},
	{0x94BD4E9549963CEEULL, "161116798.ogg"},
	{0x94BD8C10EFC44E98ULL, "zz_carpt_debug.ncs"},
	{0x94BE2A6F6A5C17A5ULL, "298046947.ogg"},
	{0x94BE5000F12A47AAULL, "6113052_m.fxe"},
	{0x94BE60C197791FE3ULL, "403738082.ogg"},
	{0x94BE6DEB95C896DDULL, "297911514.ogg"},
	{0x94BE7D3477E7DB43ULL, "vxlt_0_lowtown_n_f5d_0_0.vlm"},
	{0x94BEC471EC84B459ULL, "6117055_m.fxe"},
	{0x94BEE83B7929F75EULL, "llac_ca_caverns1_24427_4.dds"},
	{0x94BF01B2E3C28D88ULL, "fx_kungfu_dust.mao"},
	{0x94BF351F90FDF40BULL, "art\\vfx\\combat\\vfx_m_ent_horror_act_c.impa.ani"},
	{0x94BF4D04AC15C7D2ULL, "lgt100cr_gang_4_docks.utc"},
	{0x94BF6D3BDD15DFE6ULL, "6034027_m.fxe"},
	{0x94BFBFA773F9C1C6ULL, "vxlt_hangedman_22u_0_0.vlm"},
	{0x94BFE6C3C9648911ULL, "hightown_male_noble_gossip6171315.cl"},
	{0x94C08BBFE8C774DCULL, "w_bonepitdestroyed_a3_245_0.mao"},
	{0x94C097BFE8C7897FULL, "w_bonepitdestroyed_a3_245_0.mmh"},
	{0x94C09DBFE8C79321ULL, "w_bonepitdestroyed_a3_245_0.msh"},
	{0x94C12CF19BD316A4ULL, "vxlt_gallowstemplar_433_0_0.vlm"},
	{0x94C19897EB3EC429ULL, "hlac_lt_gallowstemplar_f_2391_3z.dds"},
	{0x94C1B8845D06310FULL, "zz_followers_banter6159267.cl"},
	{0x94C25C43EF88C472ULL, "lt_0_lowtown_a3_n_c_51_0.rml"},
	{0x94C277B80C1977F1ULL, "w_encounter2_camp_22_0.phy"},
	{0x94C27EB5F9661A5AULL, "follower_banter6172751.cl"},
	{0x94C2BB202CE31304ULL, "vxlt_warehouse_n_wo_0_0.vlm"},
	{0x94C335385C4D9DC0ULL, "180681259.ogg"},
	{0x94C3AA926B9ADE02ULL, "515877609.ogg"},
	{0x94C4FBD6FA9394E1ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm03.phy"},
	{0x94C5303B0F86030FULL, "isa120_hayder6170308.cl"},
	{0x94C5DE339EF8DCC5ULL, "qun230cr_guild_contact6068697.cl"},
	{0x94C66E81CCBD2009ULL, "vxw_encounter2_mb_0_0.vlm"},
	{0x94C6AA7F032E6F9BULL, "adjustapproval.ncs"},
	{0x94C6AC920B3BD9E3ULL, "vxht_mansion_n_57l_0_1.vlm"},
	{0x94C73D66EA7A6B42ULL, "6064109_m.fxe"},
	{0x94C77835C286E138ULL, "vxw_encounter3_8c_0_0.vlm"},
	{0x94C7C10F7677CF02ULL, "mcr370_meredith6083101.cl"},
	{0x94C8508466EF6FFEULL, "vxht_0_hightown_n_9ur_0_0.vlm"},
	{0x94C86F30B404648CULL, "276726431.ogg"},
	{0x94C8BB7A2E53E0CDULL, "6016672_f.fxe"},
	{0x94C8C6CC64F05A77ULL, "aip_gateguardianslave_base.gda"},
	{0x94C8E60D6EC8F93AULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magiclecturn_r.dds"},
	{0x94C980B80C1FA5EDULL, "w_encounter2_camp_22_0.rml"},
	{0x94C98F8BF88482B5ULL, "vxht_0_hightown_n_b92_0_0.vlm"},
	{0x94C990F10C693CAFULL, "pro000_cut2d_real_story.bnk"},
	{0x94C9D66AEC2B1DC8ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyo_0t.dds"},
	{0x94CA0AFE2686124BULL, "vxw_encounter2_camp_3r_0_0.vlm"},
	{0x94CA0C14C2F4FE3FULL, "6116573_m.fxe"},
	{0x94CA312F098A7A6DULL, "vxlt_0_lowtown_ghw_0_0.vlm"},
	{0x94CA389A203D29D0ULL, "gencr_spider_giant.utc"},
	{0x94CA653B6A2F1F5DULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_c.dura.ani"},
	{0x94CB2E81FD6CEF5BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_3.evt"},
	{0x94CC0C0D84A2B98BULL, "atl_fonttest_dxt5_01.dds"},
	{0x94CC6BB1566CA711ULL, "llac_w_encounter2_camp_-131067_0.dds"},
	{0x94CC8036AFE3F0B9ULL, "lt_0_lowtown_a2_n_c_28_0_ll.mmh"},
	{0x94CCB2F10C6BAC17ULL, "pro000_cut2d_real_story.cnv"},
	{0x94CCB441482580BCULL, "500073577.ogg"},
	{0x94CCDC09D537AC5BULL, "vxht_keep_n_6y4_0_0.vlm"},
	{0x94CCEF6312F6D352ULL, "zz_vault_debug6203293.cl"},
	{0x94CD1D1F910A0882ULL, "art\\vfx\\combat\\vfx_m_ent_horror_act_c.impa.evt"},
	{0x94CD49EAE1F87529ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chandelier_n.dds"},
	{0x94CD6447CBFEAFE2ULL, "246276063.ogg"},
	{0x94CE9DCD69F3CCAFULL, "zz_mrl_debug6109296.cl"},
	{0x94CEF0354BA1FCC3ULL, "zz_hawke_cu_ve.cut"},
	{0x94CF20EA295609ADULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_012a_0s.dds"},
	{0x94CF2F160305F73AULL, "6095651_m.fxe"},
	{0x94CF658D91077F3AULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart_tool.msh"},
	{0x94CF738D91079773ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart_tool.mao"},
	{0x94CF778D91079D80ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart_tool.mmh"},
	{0x94CFE244C02E3F11ULL, "6053063_m.fxe"},
	{0x94D0759890BBD35FULL, "611577774.ogg"},
	{0x94D080D73BDF37E7ULL, "arm_pl3.tnt"},
	{0x94D0CF5A2DD9344BULL, "lt_0_lowtown_docks_a3_n_38_0_lc.mmh"},
	{0x94D115C1C6589801ULL, "472303297.ogg"},
	{0x94D12CFD1ED92B84ULL, "vxlt_warehouse_n_zl_0_0.vlm"},
	{0x94D130A8FC942F80ULL, "90825057.ogg"},
	{0x94D14FB816589CD5ULL, "645617398.ogg"},
	{0x94D186C3D87DA2A5ULL, "vxw_bonepit_1ii_0_0.vlm"},
	{0x94D1B36AD083FBE8ULL, "dre000_sandal6171994.cl"},
	{0x94D1B4E4E715427FULL, "vxw_woundedcoast_1re_0_0.vlm"},
	{0x94D22781FD73019CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_3.gad"},
	{0x94D2343A19C77A8FULL, "dae000_bodhan6146908.cl"},
	{0x94D24B814EB58089ULL, "dae000_pc_home_ambients6208366.cl"},
	{0x94D2C8F48286F9E4ULL, "follower_banter6155787.cl"},
	{0x94D3665C6DFC659EULL, "ht_0_hightown_19_0.rml"},
	{0x94D3794AFBCC0645ULL, "vxca_cavern1_haunted_fob_0_0.vlm"},
	{0x94D393B6C23A0BDCULL, "653958714.ogg"},
	{0x94D3B9DFA56ABDFFULL, "mag110_samson6016166.cl"},
	{0x94D41C1F91102595ULL, "art\\vfx\\combat\\vfx_m_ent_horror_act_c.impa.gad"},
	{0x94D4D1194F76D038ULL, "vxlt_blackemporium_17s_0_0.vlm"},
	{0x94D4E05EBEB33889ULL, "vxw_encounter2_ha_0_0.vlm"},
	{0x94D56C001049B787ULL, "dae000_worthy6166596.cl"},
	{0x94D56F1F960BA88AULL, "vxlt_backalley_n_1bl_0_0.vlm"},
	{0x94D5CCD7B668F8EFULL, "vxht_mansion2b_1b1_0_0.vlm"},
	{0x94D5F0AA01345B3EULL, "vxht_mansion_n_4vz_0_0.vlm"},
	{0x94D5FCE2BEA43CCFULL, "672224554.ogg"},
	{0x94D61810B864412EULL, "vxlt_0_lowtown_a2_n_gmt_0_0.vlm"},
	{0x94D622FB8A63C67EULL, "follower_barks6153490.cl"},
	{0x94D6AAB37A1E7D44ULL, "vxht_chantry_n_2pp_0_0.vlm"},
	{0x94D6D3B0C597EDFDULL, "pro000_cut_act_of_mercy6001019.cl"},
	{0x94D76D1BD2CF31EEULL, "6082105_m.fxe"},
	{0x94D7C0FA5219181EULL, "6139676_m.fxe"},
	{0x94D7FDFFA1FF79C8ULL, "423141359.ogg"},
	{0x94D889A33518C7F8ULL, "367793692.ogg"},
	{0x94D88D1010FA91C0ULL, "gen00fl_merrill6108700.cl"},
	{0x94D8D9EF434A1E43ULL, "6141919_m.fxe"},
	{0x94D9143FE92D18F0ULL, "hlac_lt_0_lowtown_a2_n_2_0z.dds"},
	{0x94D92681FD791E9AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left_3.ani"},
	{0x94D95262635A6019ULL, "vxlt_0_lowtown_a2_n_c_h1v_0_0.vlm"},
	{0x94D980596D82E895ULL, "6175169_m.fxe"},
	{0x94D9AF882C91B611ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallowbird.msh"},
	{0x94D9BD882C91CE2CULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallowbird.mao"},
	{0x94D9C9882C91E2CFULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallowbird.mmh"},
	{0x94D9F72702B4A3FDULL, "vxlt_0_lowtown_n_er4_0_0.vlm"},
	{0x94DADB451CE1A37AULL, "6137010_m.fxe"},
	{0x94DAED01B06FCC9AULL, "386656882.ogg"},
	{0x94DB233D80D6A30FULL, "696832395.ogg"},
	{0x94DB274DDB7BC5A6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_tw_v1.evt"},
	{0x94DB9C790D76E6C2ULL, "942044391.ogg"},
	{0x94DC0CE4923CB7D3ULL, "vxlt_undercityencounter_13c_0_0.vlm"},
	{0x94DCD2B2E3DB3BEAULL, "fx_kungfu_dust.dds"},
	{0x94DCFF09F1927576ULL, "vxlt_0_lowtown_a2_n_c_ch9_0_0.vlm"},
	{0x94DD2B63CA93FBFEULL, "6128963_f.fxe"},
	{0x94DD3C5DE4832D53ULL, "649842576.ogg"},
	{0x94DD88336894939FULL, "1006814706.ogg"},
	{0x94DD9A951376EA5CULL, "gen00fl_sebastian6174696.cl"},
	{0x94DE0A27A6F2FA73ULL, "dae000_gamlen6145118.cl"},
	{0x94DE2BA78C85904EULL, "6167437_m.fxe"},
	{0x94DE3CA50EEE07B9ULL, "6151093_m.fxe"},
	{0x94DE983B2D693D70ULL, "6088105_m.fxe"},
	{0x94DEC92F8D5531F0ULL, "vxlt_gallowstemplar_n_5qg_0_0.vlm"},
	{0x94DED16679F585C9ULL, "346808497.ogg"},
	{0x94DEDA2F9C468BCAULL, "6202077_m.fxe"},
	{0x94DF4828EA4CC93DULL, "vxlt_hangedman_n_28g_0_0.vlm"},
	{0x94DF62AC56127CEAULL, "llac_lt_0_lowtown_-65536_7.dds"},
	{0x94DF6FFD6EEAA41CULL, "art\\characters\\weapons\\dagger\\w_dgr_101.mao"},
	{0x94DF889A85FAA0E8ULL, "6052753_m.fxe"},
	{0x94E0773A2A673909ULL, "179953708.ogg"},
	{0x94E0FBC276075A78ULL, "gen00fl_anders6112990.cl"},
	{0x94E13E4DDB8058F1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_tw_v1.gad"},
	{0x94E18D0F7A34A5A1ULL, "6104197_f.fxe"},
	{0x94E1B813C7CBD0D8ULL, "42453303.ogg"},
	{0x94E1EBBBA97F6565ULL, "6152171_m.fxe"},
	{0x94E21811A8CADD8CULL, "6072272_m.fxe"},
	{0x94E23D7045197422ULL, "gen00fl_anders6125118.cl"},
	{0x94E24ADE968C93A9ULL, "6075705_f.fxe"},
	{0x94E33C91563BAD18ULL, "6107338_m.fxe"},
	{0x94E3615EDF55FA3CULL, "dae020cr_hf_noble2.utc"},
	{0x94E38E1E120B1082ULL, "vxlt_0_lowtown_docks_b5n_0_1.vlm"},
	{0x94E3DDF8D139A277ULL, "dr_primevaltomb_663.rml"},
	{0x94E41071DD24BB58ULL, "683804706.ogg"},
	{0x94E43C2DDAE6EE17ULL, "560750893.ogg"},
	{0x94E4783774DA69E4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\runpausefr.evt"},
	{0x94E4CBA5B066B9DEULL, "hlac_lt_0_lowtown_a2_65538_4x.dds"},
	{0x94E4CF6885DEC722ULL, "6151296_m.fxe"},
	{0x94E5368416C533DAULL, "hlac_lt_0_lowtown_n_131072_3y.dds"},
	{0x94E58D9A53915652ULL, "vxlt_undercityencounter_13f_0_0.vlm"},
	{0x94E65F88154FD83BULL, "6139432_f.fxe"},
	{0x94E6C170991391D9ULL, "6140147_m.fxe"},
	{0x94E6E84A57E14D0EULL, "proj_ele.mao"},
	{0x94E6F37C30DA7323ULL, "6109744_m.fxe"},
	{0x94E754D151123847ULL, "gen00fl_merrill6128710.cl"},
	{0x94E7E16CBE55EA1FULL, "dae000_flemeth6013313.cl"},
	{0x94E83F4DDB86789FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_tw_v1.ani"},
	{0x94E84E1C3F016133ULL, "qun120_arishok_guard6049214.cl"},
	{0x94E87D3D9F216DADULL, "6141881_m.fxe"},
	{0x94E8A128180ADBA1ULL, "388324953.ogg"},
	{0x94E8F9CB3A9AC42AULL, "80883023.ogg"},
	{0x94E981F4C34E50D3ULL, "6173532_m.fxe"},
	{0x94E9995A96C8DD71ULL, "6151347_m.fxe"},
	{0x94E9E187D326C0B6ULL, "925528496.ogg"},
	{0x94EAB9107B4C5117ULL, "vxlt_0_lowtown_n_gqy_0_0.vlm"},
	{0x94EAC712011D254EULL, "llac_ht_mansion2_818_0.dds"},
	{0x94EB08C7CAD1ADD8ULL, "482894159.ogg"},
	{0x94EB210F240727F2ULL, "59101617.ogg"},
	{0x94EB3037521D57FFULL, "combat_smm_range.gda"},
	{0x94EB6D58F61D3F02ULL, "977361544.ogg"},
	{0x94EB8860A01235AEULL, "vfx_distortionmask_p0.vsh.fxo.dx11.assembly"},
	{0x94EC03B33C7C1884ULL, "6067656_m.fxe"},
	{0x94EC552722C8A4C2ULL, "6197640_f.fxe"},
	{0x94ECA7F21888B043ULL, "6163414_m.fxe"},
	{0x94ECE76E3DA51394ULL, "6126348_m.fxe"},
	{0x94ECE9C81377153DULL, "6175943_m.fxe"},
	{0x94ECF56DBCA68EC9ULL, "vxca_cavern1_haunted_ics_0_2.vlm"},
	{0x94ED0A181F5F55F0ULL, "6083005_m.fxe"},
	{0x94ED6D76C41130F0ULL, "6144784_m.fxe"},
	{0x94EE38B64BCAB3A2ULL, "239302072.ogg"},
	{0x94EEA44643B5EC75ULL, "6174427_m.fxe"},
	{0x94EEC78C9CDF7A82ULL, "22505758.ogg"},
	{0x94EF54A4A2A191ABULL, "zz_vault_debug6203218.cl"},
	{0x94EF98F0311BB8F4ULL, "vxlt_blackemporium_8k_0_0.vlm"},
	{0x94EFB815F1D47B94ULL, "772183734.ogg"},
	{0x94EFEC3482CA0B47ULL, "vxdr_primevaltomb_6x9_0_0.vlm"},
	{0x94EFF1B0E836AB34ULL, "llac_dr_primevalentrance_1726_4.dds"},
	{0x94F062A5692607A8ULL, "llac_lt_0_lowtown_n_-65536_4.dds"},
	{0x94F0844BFA546226ULL, "zz_approval_debug6003955.cl"},
	{0x94F087A2BBDD7C7CULL, "6133085_m.fxe"},
	{0x94F0ABE7F05E1068ULL, "plt_ave300pt_opening.nss"},
	{0x94F1127CC2431CD7ULL, "hlac_lt_0_lowtown_a3_n_c_65533_0x.dds"},
	{0x94F19A2A72A30F91ULL, "6107304_m.fxe"},
	{0x94F1AC8351C8B1A4ULL, "fen210_hadriana6110388.cl"},
	{0x94F23D71768275F7ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv1.gad"},
	{0x94F34607C24CB0B5ULL, "758770159.ogg"},
	{0x94F39EA1E49B5A98ULL, "vxlt_gallowscourtyard_71h_0_0.vlm"},
	{0x94F41C93E326E3AFULL, "dae000_mother6099188.cl"},
	{0x94F44C9D5271BC40ULL, "683446292.ogg"},
	{0x94F4612DAE52D71BULL, "vxlt_blackemporium_v2_0_0.vlm"},
	{0x94F49F41C397A076ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesm.phy"},
	{0x94F5B64EB430E365ULL, "vxlt_warehouse_n_5t_0_0.vlm"},
	{0x94F658729EB82E1EULL, "gen00fl_bethany6101392.cl"},
	{0x94F6B4E7F0628B7BULL, "plt_ave300pt_opening.plo"},
	{0x94F6F0BBAEB5C74EULL, "one220_yevhen6103721.cl"},
	{0x94F7056C727F514BULL, "6034192_m.fxe"},
	{0x94F79CB5926DD788ULL, "6205311_m.fxe"},
	{0x94F79E99F873AE1CULL, "vxht_0_hightown_n_amp_0_0.vlm"},
	{0x94F7BAEAA252F4D6ULL, "genip_candles.utp"},
	{0x94F7D200F43C6D92ULL, "zz_followers_banter6159732.cl"},
	{0x94F884D0F6290C3CULL, "1028748824.ogg"},
	{0x94F8A3D308A1F9C9ULL, "188883100.ogg"},
	{0x94F8CFE4FA980406ULL, "664336617.ogg"},
	{0x94F952717688B784ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv1.evt"},
	{0x94F9639B6F4785F3ULL, "686795127.ogg"},
	{0x94F9B865FE3F5887ULL, "125419065.ogg"},
	{0x94F9D5E174443141ULL, "714348660.ogg"},
	{0x94F9DF0FBF551291ULL, "genip_vase.utp"},
	{0x94FA2EA35DE2B5B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkbackleft.ani"},
	{0x94FA3C2D8CE8259CULL, "6134523_m.fxe"},
	{0x94FA9CA5B72F6765ULL, "6083141_m.fxe"},
	{0x94FB1C50779BB3DBULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he1a.msh"},
	{0x94FB3A50779BE6C5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he1a.mmh"},
	{0x94FB6E601A1C9FB7ULL, "hlac_ca_cavern1_haunted_24427_4z.dds"},
	{0x94FB737568560C83ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_act_c.anb"},
	{0x94FBAE74411B7A5BULL, "vxdr_primevalentrance_348_0_0.vlm"},
	{0x94FCCFA1F3760EF6ULL, "vxlt_0_lowtown_a2_n_c_fec_0_0.vlm"},
	{0x94FCE9E8E8722023ULL, "vxht_mansion2_3gu_0_1.vlm"},
	{0x94FD726E3D530F99ULL, "m_dae330ar_undercity.wbd"},
	{0x94FDADCE84C783C2ULL, "6179953_m.fxe"},
	{0x94FE2318F99DC1E8ULL, "hlac_lt_0_lowtown_a3_n_c_131070_3x.dds"},
	{0x94FE36CD965AF50EULL, "897891448.ogg"},
	{0x94FE8841C39FCD17ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesm.msh"},
	{0x94FE9A41C39FEC7AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesm.mao"},
	{0x94FE9E41C39FF289ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesm.mmh"},
	{0x94FEAA68ADD6BD63ULL, "411109892.ogg"},
	{0x94FF0F63C5B9957EULL, "qcr000_varric6208255.cl"},
	{0x94FFA0B0FE98DFB8ULL, "vxlt_0_lowtown_a2_euc_0_0.vlm"},
	{0x94FFBCB62F5884C6ULL, "zz_qun_debug6219188.cl"},
	{0x950092BEA20A9FDFULL, "6072217_m.fxe"},
	{0x9500ED70D3A01994ULL, "6208477_m.fxe"},
	{0x950109A35DE895FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkbackleft.gad"},
	{0x95010F05E234ACB1ULL, "6097864_f.fxe"},
	{0x9501226E8C4B426FULL, "vxht_mansion2b_1s2_0_0.vlm"},
	{0x95014668C0C06418ULL, "llac_lt_backalley_n_65537_1.dds"},
	{0x95018ACF89D8CAB8ULL, "1001071523.ogg"},
	{0x9501955C35AB47D6ULL, "one100ip_chest_1.utp"},
	{0x9502980CE064C409ULL, "6027972_m.fxe"},
	{0x9502D14F2E0DB6D0ULL, "art\\vfx\\materials\\chunky\\vfx_c_lightbeam_falloff.mao"},
	{0x9502EDEEED7DA820ULL, "vrc321_party_barks6095785.cl"},
	{0x95032E8E76B82222ULL, "character_character_p106.psh.fxo.dx11"},
	{0x95039F8D1F0CDAE4ULL, "follower_barks6153540.cl"},
	{0x950415CF9346AAA3ULL, "vfx_blood_dot.dds"},
	{0x9504282BA626FBE5ULL, "vxlt_warehouse_n_8s_0_0.vlm"},
	{0x950521FD249AF89FULL, "300021625.ogg"},
	{0x9505881F0A55016FULL, "645841234.ogg"},
	{0x9506639A8A773319ULL, "98700610.ogg"},
	{0x95066A7176936B7DULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv1.ani"},
	{0x95068322049A90C7ULL, "vxw_bonepit_zu_0_0.vlm"},
	{0x9506DECD621838F1ULL, "hlac_ca_caverns2prc_5802_54z.dds"},
	{0x950710F2973F6F67ULL, "hlac_lt_undercitydungeon_-1_1y.dds"},
	{0x95071A2F68EF596CULL, "vxht_0_hightown_a2_bwz_0_0.vlm"},
	{0x950726521F05CBAAULL, "6047674_m.fxe"},
	{0x9507F07F74F22C8CULL, "6142904_m.fxe"},
	{0x9507F822C6D20697ULL, "hlac_ht_mansion_n_3373_0y.dds"},
	{0x95080325ED48C28FULL, "6115561_m.fxe"},
	{0x950837E94F2317C1ULL, "6182273_m.fxe"},
	{0x950845F53A6FB007ULL, "vxw_encounter3_gh_0_1.vlm"},
	{0x950852FC435EFC08ULL, "6013332_m.fxe"},
	{0x9508B51454BF848CULL, "vxlt_undercityencounter_x2_0_0.vlm"},
	{0x9508BBEBD9BE7F84ULL, "626879006.ogg"},
	{0x9508BE2C2CF41CADULL, "w_encounter2_camp_49_0.msh"},
	{0x9508CC2C2CF434E0ULL, "w_encounter2_camp_49_0.mao"},
	{0x9508D02C2CF43B33ULL, "w_encounter2_camp_49_0.mmh"},
	{0x9508F8730843754CULL, "hlac_lt_darktown_-1_77z.dds"},
	{0x9509507568620EACULL, "art\\vfx\\combat\\vfx_w_gn2_threat_act_c.mmh"},
	{0x95097B5D63660BF2ULL, "vxw_woundedcoast_137_0_0.vlm"},
	{0x950A967C7290BB15ULL, "gen00fl_dog6117432.cl"},
	{0x950AFB993350846EULL, "mrl351_pridedemon6165550.cl"},
	{0x950BB3A4BD2A4450ULL, "6105710_m.fxe"},
	{0x950BBCF7F01EF119ULL, "gen00fl_merrill6127502.cl"},
	{0x950C5AAA038A06A1ULL, "460377233.ogg"},
	{0x950D4FCD61CC6501ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f.ani"},
	{0x950D6418F6C54AC1ULL, "329866078.ogg"},
	{0x950D67E323FA046EULL, "49520211.ogg"},
	{0x950D6B5C9FC9AD7CULL, "6139685_m.fxe"},
	{0x950D6EF23D47082AULL, "dre000_varric6036985.cl"},
	{0x950EB3ED380F9F4DULL, "llac_lt_0_lowtown_a2_65534_1.dds"},
	{0x950F049042CA53B1ULL, "100658160.ogg"},
	{0x950F58BBEF09DCEEULL, "6051278_f.fxe"},
	{0x950FCC9A31658A1FULL, "6177341_m.fxe"},
	{0x950FE4B8519D36F2ULL, "gen00fl_merrill6115788.cl"},
	{0x951056AAD43C0F82ULL, "w_encounter2_36_0.rml"},
	{0x9510CFB3FDD1C052ULL, "hlac_lt_0_lowtown_docks_a3_-65538_0y.dds"},
	{0x95110704D63F0AC3ULL, "mcr310_argument6074365.cl"},
	{0x95110C638BC466BBULL, "follower_banter6155913.cl"},
	{0x951139B84F78684CULL, "vxlt_darktown_2ug_0_0.vlm"},
	{0x95113F46B587D262ULL, "hlac_ca_cavern1_haunted_24425_10x.dds"},
	{0x9511681EBF72B8D8ULL, "vxlt_hangedman_n_10u_0_0.vlm"},
	{0x9511681FAFBCAA16ULL, "clf260_harley6093562.cl"},
	{0x951186C687E4F77BULL, "vxlt_gallowstemplar_38z_0_0.vlm"},
	{0x9511C75C45DD3E7EULL, "613335957.ogg"},
	{0x95125EBD3EF0D152ULL, "ave200_donnic6110727.cl"},
	{0x9512733DAB6A3749ULL, "mrl150_cut_merrill_intro6117759.cl"},
	{0x9512A5391EE88D10ULL, "vxlt_0_lowtown_docks_a3_bqz_0_0.vlm"},
	{0x9512AA4B42D5FADBULL, "mag320_orsino6071060.cl"},
	{0x9512D1F26D5EB455ULL, "character_character_p194.psh.fxo.dx11.assembly"},
	{0x95135ACD61D0E3A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f.gad"},
	{0x9513A5BF0B639A57ULL, "vxlt_foundry_n_yp_0_0.vlm"},
	{0x9513D6ABFAD4B6B3ULL, "92817196.ogg"},
	{0x9514220FE46FE478ULL, "llac_lt_ruinedhouse_n_-1_6.dds"},
	{0x95159511F65E94D3ULL, "713108942.ogg"},
	{0x951736ECB5D6E5CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_right.gad"},
	{0x95175386F7A38CEDULL, "vxht_chantry_n_32z_0_0.vlm"},
	{0x951757AAD4423036ULL, "w_encounter2_36_0.phy"},
	{0x95179CE70ABB897AULL, "133777330.ogg"},
	{0x9517EFDB402FD0C2ULL, "hlac_lt_gallowstemplar_n_2384_1y.dds"},
	{0x9517FF7E341675C9ULL, "llac_lt_backalley_n_65536_5.dds"},
	{0x95181AB8E739C78BULL, "llac_lt_0_lowtown_docks_a3_-262144_2.dds"},
	{0x95187CCBCEE32808ULL, "and111ip_at_lowtown.utp"},
	{0x9518AA9062C373B5ULL, "vxlt_0_lowtown_emn_0_0.vlm"},
	{0x951A283CE1806B09ULL, "lt_0_lowtown_docks_149_0.phy"},
	{0x951A37CD61D6C688ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f.evt"},
	{0x951AAFB4490B6EC0ULL, "dae000_quintus6154710.cl"},
	{0x951ABA5D6A2C8EF2ULL, "one300_nuncio6073781.cl"},
	{0x951AC403FCF30022ULL, "vxht_keep_n_60f_0_0.vlm"},
	{0x951AE798E137E65AULL, "art\\vfx\\combat\\vfx_f_avl_bdygrd_imp_c.anb"},
	{0x951B29F8242D3229ULL, "6111538_m.fxe"},
	{0x951BCAE7EA539FF1ULL, "vxca_cave_2ea_0_0.vlm"},
	{0x951BD006F3222A46ULL, "cod_ltr_flora_diary_1.utp"},
	{0x951C2673B63B8DA5ULL, "zz_followers_banter6159135.cl"},
	{0x951C54E2131F94EFULL, "qun111_ketojan6039812.cl"},
	{0x951CA2BDFAECA965ULL, "6039129_f.fxe"},
	{0x951CFE83B6D72B2DULL, "vxlt_0_lowtown_bz4_0_0.vlm"},
	{0x951D0E171F813507ULL, "6055430_m.fxe"},
	{0x951D24C70A9C258DULL, "hlac_ht_chantry_n_2534_29z.dds"},
	{0x951D64AE526F9235ULL, "6164761_m.fxe"},
	{0x951DA8704FDA230AULL, "vxlt_backalley_n_1yi_0_0.vlm"},
	{0x951E07800B34C73FULL, "vxlt_0_lowtown_docks_9bt_0_0.vlm"},
	{0x951E14CB0F80DB98ULL, "ser270_templar6122503.cl"},
	{0x951E4AB9C0CF0DF5ULL, "seb321_leliana6197850.cl"},
	{0x951E4BECB5DD287CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_right.evt"},
	{0x951E6F51E13EF135ULL, "llac_w_bonepitdestroyed_a3_393222_0.dds"},
	{0x951E86AB357B4908ULL, "407593178.ogg"},
	{0x951E8959149F38D1ULL, "vxlt_0_lowtown_docks_a3_n_cyy_0_0.vlm"},
	{0x951EF8AD1CF7C5D3ULL, "art\\characters\\playercharacter\\children\\kf_har_ha01a.phy"},
	{0x951F60AB617636A5ULL, "hlac_lt_0_lowtown_a2_n_65535_8y.dds"},
	{0x951F86D5649D0D69ULL, "6161506_m.fxe"},
	{0x951FAB95416E80CCULL, "vxlt_darktown_2xd_0_0.vlm"},
	{0x951FADDFDF537FEFULL, "547868955.ogg"},
	{0x951FE99C40FBB2D8ULL, "417068860.ogg"},
	{0x951FF8CF8F2BD232ULL, "vxlt_0_lowtown_a3_n_c_feb_0_0.vlm"},
	{0x95202D25B64D3FC4ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev1_loop.ani"},
	{0x95206E705736EE34ULL, "835884592.ogg"},
	{0x952091A0FB5505B5ULL, "vxca_cave_1ab_0_0.vlm"},
	{0x9520BF23D997EB5CULL, "vfx_vfx_p1.psh.fxo.dx11"},
	{0x9520E426CABCBDC1ULL, "vxht_0_hightown_a2_bie_0_0.vlm"},
	{0x9520EC9AD5A4DECBULL, "dae000_sandal6148196.cl"},
	{0x9521384399B4871BULL, "hlac_lt_hangedman_-1_20x.dds"},
	{0x952140AAD44A5DD7ULL, "w_encounter2_36_0.msh"},
	{0x952152AAD44A7C3AULL, "w_encounter2_36_0.mao"},
	{0x952156AAD44A8349ULL, "w_encounter2_36_0.mmh"},
	{0x95218B2728D0B75DULL, "6101020_f.fxe"},
	{0x9521959FB4A668ACULL, "llac_ht_mansion2b_818_42.dds"},
	{0x952233B95511F0AEULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_fin_c.impa.ani"},
	{0x9522D6F802E756A1ULL, "278097020.ogg"},
	{0x9523C3595931FE9BULL, "qun131_taskmaster6039707.cl"},
	{0x95244A5D8CE71983ULL, "zz_followers_banter6158112.cl"},
	{0x9524BC5062710B15ULL, "1000139555.ogg"},
	{0x95251373ABA20783ULL, "zz_mag_debug6054596.cl"},
	{0x9525198F110C6027ULL, "follower_barks6153624.cl"},
	{0x9525A43BE367181FULL, "gen00fl_aveline6132882.cl"},
	{0x9525B6D58BA4AB29ULL, "zz_area_signage17.utp"},
	{0x9525E66303982EE5ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_deeproadsem_d.dds"},
	{0x9526177569E3AB55ULL, "123269495.ogg"},
	{0x95265F1747523CF2ULL, "vxw_woundedcoast_191_0_0.vlm"},
	{0x9526704B0FBF0F6BULL, "vxlt_blackemporium_tl_0_0.vlm"},
	{0x9526F33F1A6B56F8ULL, "hlac_lt_warehouse_1_0z.dds"},
	{0x95271C6D54B98C35ULL, "vxlt_0_lowtown_ehm_0_0.vlm"},
	{0x95271F7C4D0DBAFBULL, "28756786.ogg"},
	{0x95272A25B6535926ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev1_loop.gad"},
	{0x95275A68C8D94386ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_tochest.ani"},
	{0x9527B32443AE9874ULL, "876313594.ogg"},
	{0x9527B82C9F7FC653ULL, "266985513.ogg"},
	{0x9527ED00FA06C9A4ULL, "vxht_mansion2_24j_0_1.vlm"},
	{0x952804D6FAE82156ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm03.mmh"},
	{0x95280ED6FAE83254ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm03.msh"},
	{0x952848F539070711ULL, "6202272_m.fxe"},
	{0x95287BFB96D296DAULL, "6016333_m.fxe"},
	{0x9528F6F3D786E5DEULL, "6114854_f.fxe"},
	{0x952920B2481F38E3ULL, "48106181.ogg"},
	{0x952984BAA5941C3FULL, "mer320_hubert6053186.cl"},
	{0x9529B3EDA3B13CC8ULL, "vxlt_0_lowtown_docks_a3_czj_0_0.vlm"},
	{0x9529B72304E0C0A9ULL, "vxlt_darktown_429_0_0.vlm"},
	{0x952A8979D222F38BULL, "27596268.ogg"},
	{0x952B0460C6841C5DULL, "art\\characters\\weapons\\longswords\\w_lsw_ls102.mao"},
	{0x952B17F5E4DCF882ULL, "1042267585.ogg"},
	{0x952B53ECB5E7BF95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_right.ani"},
	{0x952BE03975A1CFD4ULL, "hlac_lt_0_lowtown_a2_n_c_196606_0z.dds"},
	{0x952BEBA02286011CULL, "6021275_f.fxe"},
	{0x952C2BB59CD44596ULL, "51004514.ogg"},
	{0x952C8A18B77351F0ULL, "mag170_thrask6015993.cl"},
	{0x952CAD6C6B9FD3B4ULL, "hlac_lt_0_lowtown_a2_n_c_131072_0z.dds"},
	{0x952CC1C3C626D091ULL, "613297395.ogg"},
	{0x952D33CA204C4AE2ULL, "hlac_ht_chantry_n_2534_37z.dds"},
	{0x952D8A422B353D1BULL, "6117906_f.fxe"},
	{0x952DB60C5D697E2DULL, "llac_lt_0_lowtown_a2_n_c_-131072_4.dds"},
	{0x952DF8B2569325E9ULL, "6166406_m.fxe"},
	{0x952EB5BE7A45417BULL, "106073796.ogg"},
	{0x952EF1AE885244AAULL, "vxlt_hangedman_25x_0_0.vlm"},
	{0x952EF8BE42BB92C8ULL, "4890812.ogg"},
	{0x952F21B242DB5FA3ULL, "6076359_f.fxe"},
	{0x952F38C50D2713BAULL, "vxw_encounter2_camp_po_0_1.vlm"},
	{0x95302BB9551E20FFULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_fin_c.impa.evt"},
	{0x9530EEAB9A84BB2CULL, "vxht_mansion_n_33h_0_0.vlm"},
	{0x95316A0D611EDE1FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxb.msh"},
	{0x9531700D611EE8C1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxb.mmh"},
	{0x95317C0D611EFD62ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxb.mao"},
	{0x9531BE4189F0C2A3ULL, "llac_lt_0_lowtown_-65537_3.dds"},
	{0x9531D1D36D637BBFULL, "vxht_0_hightown_a2_bwu_0_0.vlm"},
	{0x9531D28401899982ULL, "12034561.ogg"},
	{0x9531D8B8DA7CDA47ULL, "vxw_bonepit_kv_0_0.vlm"},
	{0x95326FC63ED1A2ECULL, "vxht_0_hightown_a2_bfy_0_0.vlm"},
	{0x95330B3130A187C3ULL, "107929407.ogg"},
	{0x9534091AEF4333ADULL, "vxht_0_hightown_a2_a7x_0_0.vlm"},
	{0x95344DB8FC09483FULL, "vxlt_gallowstemplar_2oz_0_0.vlm"},
	{0x9534C4E985D1EDE1ULL, "6100297_m.fxe"},
	{0x953527C41B0D00C0ULL, "6095603_f.fxe"},
	{0x95354268C8E558F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_tochest.evt"},
	{0x95361B998B963AA6ULL, "vxht_brothel_n_57s_0_0.vlm"},
	{0x95363279AC52E9F6ULL, "6017864_m.fxe"},
	{0x95363B923FA9A393ULL, "passive_abilities.gda"},
	{0x9536A3BDD00CCF63ULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.impa.ani"},
	{0x9536B452BC9FA519ULL, "vxca_caverns2prc_aed_0_0.vlm"},
	{0x95370CB955240AE8ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_fin_c.impa.gad"},
	{0x953767E54BCA5D77ULL, "gen00fl_anders6125759.cl"},
	{0x9537A4736E3B1378ULL, "gen00fl_merrill6115980.cl"},
	{0x9537EB9F5517086FULL, "753258757.ogg"},
	{0x95382747F2DF46A8ULL, "llac_lt_gallowstemplar_n_2384_2.dds"},
	{0x9538276C743CF61BULL, "326839916.ogg"},
	{0x953828FFF6D6D929ULL, "vxlt_darktown_498_0_0.vlm"},
	{0x95383B81513FE05BULL, "208553666.ogg"},
	{0x953884902EC524A9ULL, "one300_nuncio6083398.cl"},
	{0x95388EF61091E0D7ULL, "6095386_m.fxe"},
	{0x9539818282C4DB8DULL, "6116166_m.fxe"},
	{0x953989CECFEDC6A2ULL, "vxht_mansion_n_6k3_0_0.vlm"},
	{0x95398C79AC8A6C68ULL, "258152962.ogg"},
	{0x9539F4147116B883ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_1.ani"},
	{0x953A4101AD6B8B7CULL, "vxdr_primevaltomb_7qx_0_0.vlm"},
	{0x953B4368C8E9C580ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_tochest.gad"},
	{0x953BC8F95C9F24B8ULL, "vxht_keep_n_6gy_0_0.vlm"},
	{0x953BF5FAA958F706ULL, "vxht_keep_n_6g7_0_0.vlm"},
	{0x953C392FC5724971ULL, "519175733.ogg"},
	{0x953C5B404D5D4C73ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_2.evt"},
	{0x953D54BB9EB1972CULL, "hlac_lt_hangedman_n_-1_26x.dds"},
	{0x953D638B7A485D2AULL, "vxlt_hangedman_22y_0_0.vlm"},
	{0x953D7368B22A993CULL, "follower_banter6167193.cl"},
	{0x953D7D859EE34776ULL, "vxlt_backalley_g_1n1_0_0.vlm"},
	{0x953DB85B1F1C4F35ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloorb.phy"},
	{0x953DE6AE2FFD0DB9ULL, "vxca_caverns1_gkn_0_0.vlm"},
	{0x953E1B7918FD4115ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\w_var\\bianca_s.dds"},
	{0x953E5B2C12C82770ULL, "cd_mrl351_pridedemon.bnk"},
	{0x953E72CC2239FC1CULL, "gen00fl_isabela_init6076000.cl"},
	{0x953EBFE4CBC0C66FULL, "6108839_m.fxe"},
	{0x953FEE7056105166ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_loop.ani"},
	{0x953FF38F340283B1ULL, "227140409.ogg"},
	{0x954043B05F59943FULL, "vxht_0_hightown_a2_bxt_0_0.vlm"},
	{0x95405CE30BB27FD5ULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.anb"},
	{0x95416B4F3D8EA9D8ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_018a_0t.dds"},
	{0x95418ADFF96B6D02ULL, "6000864_m.fxe"},
	{0x95418DCBB7E9F894ULL, "vxlt_warehouse_cn_0_0.vlm"},
	{0x9541B9749FD832DAULL, "hlac_lt_warehouse_n_1_3x.dds"},
	{0x9542BA25C046740CULL, "vxht_brothel_n_62y_0_0.vlm"},
	{0x9542BF95EDFF60BFULL, "vxlt_gallowstemplar_2je_0_0.vlm"},
	{0x9542EA98E1593D0DULL, "art\\vfx\\combat\\vfx_f_avl_bdygrd_imp_c.mmh"},
	{0x9543135077D95A72ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_he1a.phy"},
	{0x954329C5259E4D3BULL, "qun211_party_comment6072247.cl"},
	{0x954354404D635EB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_2.gad"},
	{0x95436E7614F42AE0ULL, "art\\characters\\playercharacter\\humanmale\\bdy_rob_magb.mao"},
	{0x954380EDBBD799EDULL, "vxca_cavern1_haunted_j2e_0_0.vlm"},
	{0x9543A75FD4AD3564ULL, "390074627.ogg"},
	{0x9543BBBDD017825AULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.impa.evt"},
	{0x9543BDA8FA3C8FC6ULL, "712139674.ogg"},
	{0x9543BE3A01FC944BULL, "142744218.ogg"},
	{0x95442B6B44FA2AA1ULL, "hlac_lt_undercitydungeon_-1_48x.dds"},
	{0x95443A3712D04B9EULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hallastatue_n.dds"},
	{0x9544D41C5BE80D6AULL, "6163969_m.fxe"},
	{0x95450CC4F9CF5532ULL, "follower_banter6169764.cl"},
	{0x95451DF6204A755EULL, "vxht_keep_5vh_0_0.vlm"},
	{0x95457446CA31B074ULL, "6149260_m.fxe"},
	{0x954587E44943A2D5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urva_0n.dds"},
	{0x95458DA62D2EB207ULL, "llac_ht_mansion2b_818_34.dds"},
	{0x9545F37C63BEED9AULL, "948678833.ogg"},
	{0x9546113CE1A57F2EULL, "lt_0_lowtown_docks_149_0.mmh"},
	{0x95461D3CE1A593C5ULL, "lt_0_lowtown_docks_149_0.mao"},
	{0x95462B3CE1A5AB1CULL, "lt_0_lowtown_docks_149_0.msh"},
	{0x954646728CBEE714ULL, "gen00fl_fenris6113858.cl"},
	{0x954662572F08968BULL, "gen00fl_dog6165637.cl"},
	{0x9546AF4B8421167BULL, "267698555.ogg"},
	{0x9546EB85B9C54A8FULL, "docks_scene_dockworkers6176498.cl"},
	{0x9546F41B84643792ULL, "330431185.ogg"},
	{0x95470C1471216C7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_1.evt"},
	{0x9547B3978D1CE943ULL, "vxw_bonepit_9t_0_0.vlm"},
	{0x9547F0992D14DF46ULL, "3562590.ogg"},
	{0x9547F9B7EB41ABEDULL, "577720157.ogg"},
	{0x95481986E091158DULL, "6082263_m.fxe"},
	{0x95481D5CBD5AE9E3ULL, "781079153.ogg"},
	{0x95485C15A852695FULL, "474633679.ogg"},
	{0x9548E55C46F356C2ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4_0.pwk"},
	{0x95493C011EBE6FC1ULL, "vxlt_0_lowtown_a2_hdd_0_0.vlm"},
	{0x95495E782F43EA2FULL, "gen00fl_sebastian6179917.cl"},
	{0x954963404D67E392ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_order_2.ani"},
	{0x9549A59E1D7E4970ULL, "6174189_m.fxe"},
	{0x954A23C9D299E416ULL, "lt_0_lowtown_docks_177_0.mao"},
	{0x954A2FC9D299F8B5ULL, "lt_0_lowtown_docks_177_0.mmh"},
	{0x954A31C9D299FC2BULL, "lt_0_lowtown_docks_177_0.msh"},
	{0x954A3AD64E953D38ULL, "vxht_keep_n_6bv_0_0.vlm"},
	{0x954A5F635D2CB45EULL, "vxht_0_hightown_n_ccb_0_0.vlm"},
	{0x954A67D79B4F0F86ULL, "vxht_keep_n_6b8_0_0.vlm"},
	{0x954ABABDD01D9F4DULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.impa.gad"},
	{0x954AEB33CC64A8C3ULL, "gen00fl_fenris6126709.cl"},
	{0x954B2B106E890874ULL, "6124698_m.fxe"},
	{0x954B5465B7DB2E67ULL, "follower_banter6136741.cl"},
	{0x954BEF6290D95FF6ULL, "vxlt_backalley_g_1k2_0_0.vlm"},
	{0x954C0975FE05C01CULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.cess.ani"},
	{0x954C90F8ABACD04CULL, "vxlt_undercityencounter_1iz_0_0.vlm"},
	{0x954CA11D02CE4D36ULL, "571157523.ogg"},
	{0x954E0B147127896DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_1.gad"},
	{0x954E51D9880BCA23ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv3.ani"},
	{0x954E59535F347262ULL, "495240778.ogg"},
	{0x954E5A8E1A777C84ULL, "648254953.ogg"},
	{0x954E8884B6FB5A9AULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.impa.gad"},
	{0x954E9D2C1B5E0818ULL, "vxht_mansion2_2a4_0_0.vlm"},
	{0x954EA4D8E2A7C6DFULL, "mrl351ip_cavern_chest_a.utp"},
	{0x954ED19D39B0DD5DULL, "static_static_p46.psh.fxo.dx11.assembly"},
	{0x954F99D3A27A711DULL, "6115296_f.fxe"},
	{0x9550198932A5F2AEULL, "and100_anders6095817.cl"},
	{0x95506C5D033C372BULL, "rdr110dg_guard_warehouse.bnk"},
	{0x95513172DFF5793FULL, "vxlt_gallowstemplar_2ed_0_0.vlm"},
	{0x955131F80429707EULL, "967004635.ogg"},
	{0x9551DD79934D119FULL, "6051650_m.fxe"},
	{0x95520D9413FA57DEULL, "841440887.ogg"},
	{0x955249673DFD7795ULL, "gen00fl_aveline_training.stg"},
	{0x95524C35D906B99BULL, "one100dg_vanard6054573.cl"},
	{0x95531527421C9F4BULL, "377818380.ogg"},
	{0x95531675FE0BF55EULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.cess.gad"},
	{0x95536E5D033E700BULL, "rdr110dg_guard_warehouse.cnv"},
	{0x95539E3E56618D98ULL, "vxlt_shop_2z_0_0.vlm"},
	{0x95539F50083E5B50ULL, "m_dae110ar_docks.bnk"},
	{0x9553F10E86F4F48CULL, "vxw_encounter2_ec_0_0.vlm"},
	{0x95540CFE72AC910CULL, "487145986.ogg"},
	{0x95541ADA0457A752ULL, "vxw_bonepitdestroyed_a3_cw_0_0.vlm"},
	{0x95541D613E0D93BBULL, "hlac_w_bonepitdestroyed_a3_458753_0y.dds"},
	{0x9554765C196E379DULL, "793913276.ogg"},
	{0x9554D3AF4ABD3E7DULL, "vxw_bonepitmining_a2_cr_0_0.vlm"},
	{0x9554D77056228660ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_loop.gad"},
	{0x95552BE70294EFEBULL, "6105418_m.fxe"},
	{0x95558B84B7017EB1ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.impa.evt"},
	{0x95564C160F6A5189ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_enter.ani"},
	{0x95564D0F9C759437ULL, "vxca_caverns1_mines_int_0_0.vlm"},
	{0x95566E321DF6D015ULL, "6096194_m.fxe"},
	{0x955703355627B12EULL, "6172798_m.fxe"},
	{0x9557D297885E1F4BULL, "6116069_m.fxe"},
	{0x9558B2DB2EB4B8A5ULL, "6084218_m.fxe"},
	{0x95590C0134F92153ULL, "vxw_bonepitdestroyed_a3_nu_0_0.vlm"},
	{0x95592AE43F0ABB54ULL, "mag291_ntbanter6056720.cl"},
	{0x9559B87AAC7C77AAULL, "qun111_ketojan6039998.cl"},
	{0x9559F175FE11D57DULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.cess.evt"},
	{0x955A102103667E73ULL, "6171808_m.fxe"},
	{0x955A1E3073AFCA1BULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.dura.evt"},
	{0x955A369BDCA02E84ULL, "bigrock_02.mao"},
	{0x955AFC2925C8130CULL, "461547805.ogg"},
	{0x955B3715B1FCFB40ULL, "vxht_mansion2_1tl_0_0.vlm"},
	{0x955BB9E8489B960EULL, "972382829.ogg"},
	{0x955BBD9166543FAAULL, "6208574_m.fxe"},
	{0x955C69D98818301AULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv3.evt"},
	{0x955C6A90E2C5EC2AULL, "769206323.ogg"},
	{0x955CB3DFB049D8E9ULL, "vxlt_darktown_6k0_0_0.vlm"},
	{0x955CC732C4BB3484ULL, "ser221_party_comments6163788.cl"},
	{0x955D0F090D542098ULL, "vxht_mansion2_2j5_0_0.vlm"},
	{0x955D276A4345C53FULL, "vxht_0_hightown_a2_bfv_0_0.vlm"},
	{0x955D67160F709E0BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_enter.gad"},
	{0x955D710D6144253EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxb.phy"},
	{0x955DFB4E72DC2EB4ULL, "fen210_fenris6109396.cl"},
	{0x955E63F789C72E4FULL, "vxht_mansion2b_2ll_0_0.vlm"},
	{0x955E8A18B54F7ACBULL, "6100255_m.fxe"},
	{0x955F2A994B663283ULL, "mag311_nyssa6064835.cl"},
	{0x955F394764C05F61ULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation.gad"},
	{0x955F53D26B1EF7E9ULL, "vxca_cavern1_haunted_gi9_0_0.vlm"},
	{0x955FAF1E760924BFULL, "392308228.ogg"},
	{0x95609FEE7224A9F5ULL, "vxlt_undercityserialkiller_1sy_0_0.vlm"},
	{0x956100B401F52E7BULL, "6079350_f.fxe"},
	{0x9561173073B5DD5CULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.dura.gad"},
	{0x95611CBC308EC0E7ULL, "6099197_m.fxe"},
	{0x95628CB6F64DBFD2ULL, "vxw_bonepitdestroyed_a3_np_0_0.vlm"},
	{0x95628E447720B550ULL, "vxca_caverns1_mines_gil_0_0.vlm"},
	{0x95633D2C12E7103AULL, "cd_mrl351_pridedemon.wbd"},
	{0x9563458C3CB356FDULL, "vxw_bonepitmining_a2_nq_0_0.vlm"},
	{0x956368D9881E4D0DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv3.gad"},
	{0x9563A384B70DE480ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.impa.ani"},
	{0x9564434FFAA54B98ULL, "vxca_cavern1_haunted_iyn_0_0.vlm"},
	{0x956464160F76B870ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_enter.evt"},
	{0x9564BB082015553FULL, "hlac_lt_0_lowtown_docks_-65538_9z.dds"},
	{0x956511EC3A150F9BULL, "w_bonepitdestroyed_a3_72_0.msh"},
	{0x95651CA7771B31ABULL, "735460200.ogg"},
	{0x956523EC3A152EE6ULL, "w_bonepitdestroyed_a3_72_0.mao"},
	{0x95652FEC3A154285ULL, "w_bonepitdestroyed_a3_72_0.mmh"},
	{0x95653A4764C4CC0FULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation.ani"},
	{0x95653BC1FA276AD6ULL, "633839617.ogg"},
	{0x956587BD55DCF61BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_saint_n.dds"},
	{0x9565FA0B515646E0ULL, "vxlt_gallowsprison_n_6n1_0_0.vlm"},
	{0x9566C3F2101FBDA8ULL, "6164724_m.fxe"},
	{0x9567163073BA475AULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_imp_p.dura.ani"},
	{0x95672B51D9E5F045ULL, "dae000_bodhan6146793.cl"},
	{0x956737717C380041ULL, "mer220_hubert6049087.cl"},
	{0x9567AA054BD6CE21ULL, "vxw_bonepitdestroyed_a3_pa_0_0.vlm"},
	{0x9568054529476CCFULL, "452776936.ogg"},
	{0x95683A19942885F9ULL, "mcr370_negotiation6080236.cl"},
	{0x95685DE30BD3D316ULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.mmh"},
	{0x9568C99723866CF5ULL, "plt_cod_itm_celebrant.plo"},
	{0x95693589DC1F4EFBULL, "332696370.ogg"},
	{0x95694881C6288106ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_act_c.impa.evt"},
	{0x9569CE43F3BE975EULL, "525495860.ogg"},
	{0x956A00DB6102761FULL, "6179336_m.fxe"},
	{0x956A020B96998BB8ULL, "lt_0_lowtown_a2_n_c_45_0_ll.mmh"},
	{0x956A1811F1E69471ULL, "161912443.ogg"},
	{0x956A2F1AFB2F44B1ULL, "character_character_p200.psh.fxo.dx11"},
	{0x956A3E0E1ABF7071ULL, "vxw_bonepitdestroyed_a3_74_0_0.vlm"},
	{0x956A707261246D2DULL, "follower_banter6155937.cl"},
	{0x956A8B2C58857964ULL, "llac_ca_caverns1_mines_24426_6.dds"},
	{0x956A8F967FE20196ULL, "302887820.ogg"},
	{0x956B4A50004EB2FCULL, "104522345.ogg"},
	{0x956B51B08AA4DECAULL, "vxw_woundedcoast_1e5_0_0.vlm"},
	{0x956B80E5FF4A3918ULL, "vxht_mansion2_2w2_0_0.vlm"},
	{0x956BC8DCE167BB27ULL, "6128936_m.fxe"},
	{0x956C78CD6319E147ULL, "6129603_m.fxe"},
	{0x956CA2C8086EE4EAULL, "hlac_lt_gallowscourtyard_-65538_0z.dds"},
	{0x956CB81CE5B68FDAULL, "74744299.ogg"},
	{0x956CE3628DCC4214ULL, "vxlt_warehouse_rm_0_0.vlm"},
	{0x956DEDD73B35A587ULL, "dre000_dougal6131189.cl"},
	{0x956E01E5130BB355ULL, "105316061.ogg"},
	{0x956E41D2D695B63DULL, "vxlt_0_lowtown_docks_be5_0_1.vlm"},
	{0x956FF8924FAC8373ULL, "vxlt_0_lowtown_a2_gvs_0_0.vlm"},
	{0x95702D863A23C649ULL, "385026348.ogg"},
	{0x95705F81C62EC751ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_act_c.impa.gad"},
	{0x9570CBAACC63FD60ULL, "vxht_mansion_n_4z0_0_0.vlm"},
	{0x9570D0A5A0760859ULL, "964872429.ogg"},
	{0x9570D4C86AE1258CULL, "vxw_encounter2_ka_0_0.vlm"},
	{0x9570FD5A654F0C24ULL, "dae300_bran6143515.cl"},
	{0x957188C225FEE32BULL, "hlac_lt_0_lowtown_docks_n_0_5y.dds"},
	{0x9571EC4F31DEBF96ULL, "framebuffereffects_smallblur_p0.vsh.fxo.dx11.assembly"},
	{0x95721332A0C26FE1ULL, "199747815.ogg"},
	{0x95721F68AC011BE7ULL, "302779300.ogg"},
	{0x95725EEBB2E5C361ULL, "mag121_viveka6092280.cl"},
	{0x9572C387EBF4728BULL, "lt_0_lowtown_docks_a3_n_17_0.anb"},
	{0x9572F6F3CE08408FULL, "vxlt_0_lowtown_docks_a3_n_d0g_0_0.vlm"},
	{0x9572F7E9155C5CADULL, "6016833_m.fxe"},
	{0x957347ED8D319D9EULL, "ts_demonic_arm_props.gda"},
	{0x95747A2685CB8B6DULL, "character_character_p275.psh.fxo.dx11"},
	{0x9574C2A2395A4B1FULL, "hlac_w_blightlands_-65534_0x.dds"},
	{0x957517199E7C227EULL, "qun160_vashothlead6055568.cl"},
	{0x957536DC0CCD7483ULL, "llac_lt_hangedman_n_-1_3.dds"},
	{0x95755E3B6803CA77ULL, "gen00fl_merrill6124422.cl"},
	{0x95760787947D9AB0ULL, "hlac_lt_darktown_-1_37z.dds"},
	{0x95766081C63334FFULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_act_c.impa.ani"},
	{0x9576DF7F8BE604B2ULL, "lla_ht_0_hightown_a3_-65535_0.dds"},
	{0x9576EB76208DFB23ULL, "hlac_ht_keep_3743_1x.dds"},
	{0x9577E2720D3DDF26ULL, "832326552.ogg"},
	{0x9578621A66974752ULL, "6114079_m.fxe"},
	{0x95788150085D441AULL, "m_dae110ar_docks.wbd"},
	{0x9578B8B2D0B68B15ULL, "1013804783.ogg"},
	{0x957929F83A970B55ULL, "zz_store_dae310_store5.utm"},
	{0x9579EAA63C73CD9EULL, "684359289.ogg"},
	{0x9579F2C2F1405198ULL, "vxht_mansion2_2p3_0_0.vlm"},
	{0x957AA0A3D6C9FB63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\blendtree\\trooper.anb"},
	{0x957AB67521B2FAF9ULL, "zz_mcr_debug6071554.cl"},
	{0x957B553F7FC25A94ULL, "vxlt_warehouse_wr_0_0.vlm"},
	{0x957B6DB63E274DBBULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.cess.ani"},
	{0x957B9B1591C296FFULL, "6049673_m.fxe"},
	{0x957BB06CCD896E69ULL, "qun131_party_comment6117297.cl"},
	{0x957BB64195CCA104ULL, "6108241_m.fxe"},
	{0x957D99924326159DULL, "vxht_keep_581_0_0.vlm"},
	{0x957DA9E7C52701D4ULL, "vxht_mansion2b_1nb_0_0.vlm"},
	{0x957E36B65844045EULL, "6097208_m.fxe"},
	{0x957E6A6F41A29BF3ULL, "vxlt_0_lowtown_a2_gmr_0_0.vlm"},
	{0x957E6E58165F12B9ULL, "vxlt_0_lowtown_docks_a3_d0j_0_0.vlm"},
	{0x957EA7AD99B696A3ULL, "nat380_sibling6207674.cl"},
	{0x957F0B403E9E687FULL, "vxw_woundedcoast_16i_0_0.vlm"},
	{0x957F40E9B23A5B07ULL, "3889199.ogg"},
	{0x957F540A560FDC7BULL, "vxht_0_hightown_a87_0_0.vlm"},
	{0x957F8E2953162A2FULL, "6169377_m.fxe"},
	{0x957FD0A75AB44C83ULL, "vxw_sundermount_2jf_0_0.vlm"},
	{0x9580744FC807D76FULL, "hlac_ht_chantry_n_2534_1x.dds"},
	{0x958100953D89DDB6ULL, "6162791_m.fxe"},
	{0x9581B2A4D5C02FF9ULL, "vxlt_gallowscourtyard_n_70u_0_0.vlm"},
	{0x9581C38736886089ULL, "780754820.ogg"},
	{0x9581F1AD1D4C3748ULL, "art\\characters\\playercharacter\\children\\kf_har_ha01a.mmh"},
	{0x95820FAD1D4C6A72ULL, "art\\characters\\playercharacter\\children\\kf_har_ha01a.msh"},
	{0x95821B00C8307FC9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2d.mao"},
	{0x95822700C8309462ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2d.mmh"},
	{0x95822900C8309718ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron2d.msh"},
	{0x958272835A1E0964ULL, "526259566.ogg"},
	{0x9582A6B857C34538ULL, "vxw_sundermount_34s_0_1.vlm"},
	{0x95836D7C09CB8CFCULL, "6126403_m.fxe"},
	{0x9583B43A0A8DA191ULL, "hlac_ht_playershouse_n_855_14y.dds"},
	{0x9583B9557EE1543EULL, "6153492_m.fxe"},
	{0x9583CED0EA41D91BULL, "dre000_bethany6037330.cl"},
	{0x9583DA706A11F51AULL, "w_sundermount_184_0.mmh"},
	{0x9583DE706A11FC51ULL, "w_sundermount_184_0.mao"},
	{0x9583EC706A121460ULL, "w_sundermount_184_0.msh"},
	{0x958437184648A813ULL, "hlac_lt_0_lowtown_a2_n_-196608_0z.dds"},
	{0x9585CB5B1F5A2550ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloorb.msh"},
	{0x9585D95B1F5A3D6AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloorb.mmh"},
	{0x9585DD5B1F5A43A1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff22_galofloorb.mao"},
	{0x9585DE0A0E14A887ULL, "lt_0_lowtown_n_43_0_lc.mmh"},
	{0x958604FEE43FC943ULL, "one280_iwan6096743.cl"},
	{0x9586A8BDBAE38619ULL, "vxlt_0_lowtown_a2_n_c_hub_0_0.vlm"},
	{0x9586B2B930949AE7ULL, "6101062_f.fxe"},
	{0x9586E49544EEA71FULL, "6177136_m.fxe"},
	{0x95875DA73C47F79FULL, "llac_lt_0_lowtown_docks_a3_n_131070_1.dds"},
	{0x9587F1E6A4A26A4FULL, "357318509.ogg"},
	{0x95881324DDAFBC11ULL, "6094509_m.fxe"},
	{0x9588356A6E910FCAULL, "vxw_woundedcoast_1k7_0_0.vlm"},
	{0x958875B63E31E5B2ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.cess.evt"},
	{0x95889CF082F37396ULL, "gen00fl_aveline6114034.cl"},
	{0x9588A4E4C1E45FADULL, "21125855.ogg"},
	{0x95891323CB85FD32ULL, "14418811.ogg"},
	{0x95892CC3611866EFULL, "213889723.ogg"},
	{0x958932A2C2815897ULL, "hlac_lt_0_lowtown_a2_n_c_65535_8y.dds"},
	{0x958968AB1283D354ULL, "hlac_ht_chantry_n_2534_53z.dds"},
	{0x9589B2775611FA76ULL, "hlac_w_encounter3_-327675_0x.dds"},
	{0x9589C71C71B87314ULL, "vxlt_warehouse_xs_0_0.vlm"},
	{0x9589E950DDB24AF0ULL, "195955248.ogg"},
	{0x958A1BB1B4558045ULL, "gen00fl_merrill6129913.cl"},
	{0x958AB505C4E854CBULL, "llac_lt_0_lowtown_65537_6.dds"},
	{0x958AD54B87E17640ULL, "840074520.ogg"},
	{0x958B096F221CCB8EULL, "52984675.ogg"},
	{0x958B195508FADD39ULL, "459973284.ogg"},
	{0x958B47FCB6960674ULL, "vxw_bonepitmining_a2_99_0_0.vlm"},
	{0x958B86DF8C167743ULL, "vxht_keep_n_3ye_0_0.vlm"},
	{0x958BA1A12EF9721BULL, "6153347_m.fxe"},
	{0x958BEF4C6EA44F69ULL, "424264678.ogg"},
	{0x958BF865968FC899ULL, "vxlt_hangedman_n_20s_0_0.vlm"},
	{0x958C1BC4B71D1A54ULL, "vxht_mansion2b_1ac_0_0.vlm"},
	{0x958C39396F553F3EULL, "hlac_lt_0_lowtown_a2_-2_2x.dds"},
	{0x958CF51101A47C37ULL, "lowtown_male_urchin_sta6164205.cl"},
	{0x958D598805AF179DULL, "vxht_0_hightown_a3_aus_0_0.vlm"},
	{0x958DB8824ECD568CULL, "vxw_encounter2_qg_0_0.vlm"},
	{0x958DF8629A9B3DBEULL, "6089562_m.fxe"},
	{0x958E0E61CCC97653ULL, "hlac_lt_0_lowtown_a2_n_131073_2z.dds"},
	{0x958E42844CAA6503ULL, "vxw_sundermount_2oi_0_0.vlm"},
	{0x958E568188D51660ULL, "793771684.ogg"},
	{0x958EE3C65F61344CULL, "6169899_m.fxe"},
	{0x958F2E1E6A099F31ULL, "70086071.ogg"},
	{0x958F42A11F9E666FULL, "6083068_m.fxe"},
	{0x958F72722F7FF636ULL, "6167491_m.fxe"},
	{0x958F815E866C5863ULL, "zz_followers_banter6159348.cl"},
	{0x958F94B63E383945ULL, "art\\vfx\\environment\\vfx_trap_fire_cone_p.cess.gad"},
	{0x958FDE3A62228A38ULL, "468678667.ogg"},
	{0x95900F80F9E89341ULL, "art\\levels\\races\\proxy\\props\\_textures\\tent_sa.dds"},
	{0x95901075AA65BEE3ULL, "vxht_keep_n_8an_0_0.vlm"},
	{0x959024F9DA5DCD3AULL, "mcr370_assignments6176728.cl"},
	{0x9590A89708327709ULL, "vxw_encounter2_7l_0_0.vlm"},
	{0x9591189549B95DB8ULL, "vxw_sundermount_33r_0_1.vlm"},
	{0x95914A6ED559AC9CULL, "466970897.ogg"},
	{0x95917AE4E3805680ULL, "ht_0_hightown_a2_28_0_ll.mmh"},
	{0x9591803129817A72ULL, "213999597.ogg"},
	{0x95919C49AED884ACULL, "6114011_m.fxe"},
	{0x95921F56472BFBB6ULL, "gen00fl_fenris6126438.cl"},
	{0x9592E8B35FC7E9B6ULL, "vxca_caverns1_mines_hzp_0_0.vlm"},
	{0x959308F432F32C41ULL, "gen00fl_isabela6080450.cl"},
	{0x9593112A55019456ULL, "6083088_f.fxe"},
	{0x959316C53F1B3751ULL, "gen00fl_fenris6106712.cl"},
	{0x9593678F17B10324ULL, "vxht_0_hightown_n_c1b_0_0.vlm"},
	{0x9593DE4B9994DA1EULL, "isa311st_isabella_end.stg"},
	{0x9593E19F4B144C32ULL, "vxca_cavern1_haunted_fn6_0_0.vlm"},
	{0x959402A422BFE6FEULL, "vxlt_gallowstemplar_34o_0_0.vlm"},
	{0x95941E6437108ECBULL, "6074488_m.fxe"},
	{0x959420888B05C9B5ULL, "vxca_cave_19j_0_0.vlm"},
	{0x9594FE8979D55D40ULL, "vxht_mansion2_1hh_0_0.vlm"},
	{0x95951A9AACD99E99ULL, "vxlt_0_lowtown_a2_n_c_hpa_0_0.vlm"},
	{0x959560601EE7AB98ULL, "678657782.ogg"},
	{0x9595D156B6E2437DULL, "6070885_f.fxe"},
	{0x95962A81C15CD7FAULL, "zz_one_debug6051414.cl"},
	{0x959664BAF5B87367ULL, "lt_0_lowtown_docks_a3_n_22_0.rml"},
	{0x95966AC66E828F3CULL, "abi_w_mighty_blow.wbd"},
	{0x9596AD9B7014B82DULL, "vxlt_undercityencounter_xw_0_0.vlm"},
	{0x95977E4C2978C00AULL, "vxlt_0_lowtown_docks_a3_bre_0_0.vlm"},
	{0x95979CDD3437645AULL, "vxht_mansion2_1my_0_0.vlm"},
	{0x95979D903610D98CULL, "429917808.ogg"},
	{0x9597B756ED28B504ULL, "vxw_encounter2_camp_pa_0_1.vlm"},
	{0x9597CBD5608D510EULL, "llac_dr_slavers_3214_8.dds"},
	{0x959800CA55E9ED48ULL, "315371394.ogg"},
	{0x959863D0FE8A472DULL, "mag291_demon_marethari6029137.cl"},
	{0x959899FB79B8D204ULL, "6106041_m.fxe"},
	{0x959A93A57463D336ULL, "dae000_quintus6154730.cl"},
	{0x959B0B85CBFF22A4ULL, "429923753.ogg"},
	{0x959B3667DD916551ULL, "vxht_mansion2b_3cb_0_0.vlm"},
	{0x959B4AB6AA548A7DULL, "llac_lt_0_lowtown_a2_n_-3_7.dds"},
	{0x959BBAE28D974333ULL, "877150191.ogg"},
	{0x959BC254871611C1ULL, "939286152.ogg"},
	{0x959C275B7B0E7B47ULL, "6162714_m.fxe"},
	{0x959CB4613EA07D83ULL, "vxw_sundermount_2ph_0_0.vlm"},
	{0x959CC6C7AC126167ULL, "gen00fl_sebastian6174028.cl"},
	{0x959D5923BA022FA5ULL, "vxw_bonepitmining_a2_1lh_0_0.vlm"},
	{0x959DADAB49352E67ULL, "llac_lt_0_lowtown_docks_-131073_1.dds"},
	{0x959DC2D20E4443BDULL, "6119875_m.fxe"},
	{0x959E0E796DEB8F7CULL, "6165086_m.fxe"},
	{0x959EC087EC19A8D7ULL, "lt_0_lowtown_docks_a3_n_17_0.rml"},
	{0x959EF854ABE30C49ULL, "819612482.ogg"},
	{0x959F6B04C39762C9ULL, "vxht_mansion2b_1g8_0_0.vlm"},
	{0x959F79C2DDCC86FDULL, "hlac_lt_0_lowtown_-262144_2z.dds"},
	{0x959FAD84AED11612ULL, "vxht_0_hightown_a3_aoz_0_0.vlm"},
	{0x959FD3F7C8278429ULL, "366374989.ogg"},
	{0x95A01AFC0045C13DULL, "gen00fl_varric6105748.cl"},
	{0x95A03802CD1B9F61ULL, "hm_icarver_h3s3.mor"},
	{0x95A03A5F25E0BCA1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagviscount_n.dds"},
	{0x95A0A2FE604918E6ULL, "hlac_lt_0_lowtown_docks_-131073_0z.dds"},
	{0x95A15A9051BE0236ULL, "vxca_caverns1_mines_huq_0_0.vlm"},
	{0x95A1717913AF3021ULL, "vxw_bonepitdestroyed_a3_le_0_0.vlm"},
	{0x95A1BE6EB92C99F7ULL, "zz_followers_banter6157386.cl"},
	{0x95A24933592A2F15ULL, "6151163_m.fxe"},
	{0x95A26086D1BD41CEULL, "art\\characters\\playercharacter\\elffemale\\boo_lgte.msh"},
	{0x95A262FE7B33A83FULL, "vxlt_0_lowtown_a2_hgs_0_0.vlm"},
	{0x95A26E1364207964ULL, "vxlt_0_lowtown_n_gcl_0_0.vlm"},
	{0x95A27A86D1BD6DBCULL, "art\\characters\\playercharacter\\elffemale\\boo_lgte.mmh"},
	{0x95A2E4075DAA379CULL, "259692481.ogg"},
	{0x95A3171AF63830FCULL, "hlac_lt_0_lowtown_-131071_1z.dds"},
	{0x95A375425464A983ULL, "llac_lt_gallowstemplar_f_2394_6.dds"},
	{0x95A478742711C58CULL, "lgt311cr_shw_mer_mettin.utc"},
	{0x95A498DD02546020ULL, "6179973_m.fxe"},
	{0x95A4EFE99565D76FULL, "57391961.ogg"},
	{0x95A5006F8A312A95ULL, "6164800_m.fxe"},
	{0x95A51924527D40CAULL, "vxw_woundedcoast_1q1_0_0.vlm"},
	{0x95A58E40AF1DE686ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_medium.evt"},
	{0x95A5C3BD911EEFD8ULL, "6179047_m.fxe"},
	{0x95A5E7D324642A72ULL, "hlac_lt_gallowstemplar_n_2394_7z.dds"},
	{0x95A6373571BC2DD2ULL, "796191621.ogg"},
	{0x95A63FAC254F27BDULL, "follower_banter6161303.cl"},
	{0x95A644CEDD653BF0ULL, "598127334.ogg"},
	{0x95A64A8EB27012FEULL, "6136017_m.fxe"},
	{0x95A6600594AEE87FULL, "framebuffereffects_illuminationmask_p0.psh.fxo.dx11"},
	{0x95A6D9584E7E74E6ULL, "llac_lt_undercityencounter_-1_28.dds"},
	{0x95A7C9A46D3ECB17ULL, "vxca_caverns2prc_8s1_0_0.vlm"},
	{0x95A7DD48E2201A2FULL, "hlac_lt_hangedman_-1_4z.dds"},
	{0x95A803941592BFD5ULL, "hlac_ca_caverns1_24424_8x.dds"},
	{0x95A85AAD323BE23DULL, "hlac_lt_0_lowtown_docks_a3_-131074_6y.dds"},
	{0x95A8A5753CFFB7B0ULL, "vxht_mansion2b_2uh_0_0.vlm"},
	{0x95A8F9706A314625ULL, "w_sundermount_184_0.phy"},
	{0x95A978070212A654ULL, "hlac_ht_keep_n_4519_4y.dds"},
	{0x95A987676B5FE280ULL, "356551168.ogg"},
	{0x95AA00EAA542FB04ULL, "hlac_lt_0_lowtown_docks_-3_6z.dds"},
	{0x95AA122DCCADEE92ULL, "vxlt_gallowstemplar_4uo_0_0.vlm"},
	{0x95AA61F57D53FAD1ULL, "647332651.ogg"},
	{0x95AA8308590E9BF4ULL, "6165688_m.fxe"},
	{0x95AAB7CBDC7E7CACULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_chantryfloor_r.dds"},
	{0x95AADC1C7E876EBEULL, "6087762_m.fxe"},
	{0x95AB263E30969603ULL, "vxw_sundermount_2uk_0_0.vlm"},
	{0x95AB4C7F8E017F25ULL, "gen00fl_dog6167677.cl"},
	{0x95AB684F188E0B36ULL, "dae000_porfiria6162655.cl"},
	{0x95AB6AAC58517C9AULL, "760338979.ogg"},
	{0x95ABAA1FF73D9E75ULL, "345545238.ogg"},
	{0x95AC963F5B46F414ULL, "vxdr_primevaltomb_50w_0_2.vlm"},
	{0x95ACA540AF242CD1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_medium.gad"},
	{0x95ACA653AADCD26AULL, "6015062_m.fxe"},
	{0x95ACC4E297ED2E28ULL, "6097358_m.fxe"},
	{0x95AD1497F12A8742ULL, "dae000ip_crowd_em2_talk12.utp"},
	{0x95AD35A704D90240ULL, "756517953.ogg"},
	{0x95AD3AFA867267E0ULL, "dm_gen_6.mor"},
	{0x95AD3B14987A6AC8ULL, "vxdr_0_deeproads_adb_0_1.vlm"},
	{0x95AD53F8CFEEA2B8ULL, "6068899_m.fxe"},
	{0x95AD5EFEB1DA64C9ULL, "rdr111dg_assistant6055110.cl"},
	{0x95AD8C0BB7359E69ULL, "w_sundermount_63_0.rml"},
	{0x95ADDFAD1662DC89ULL, "vxw_blightlands_ae_0_0.vlm"},
	{0x95ADE497BB6C442CULL, "vxlt_hangedman_n_1zw_0_0.vlm"},
	{0x95AEE86D12F4183DULL, "vxca_caverns1_k2p_0_1.vlm"},
	{0x95AEFFF2A9120D86ULL, "346028535.ogg"},
	{0x95AF16E08AA7D98CULL, "mag311_keran6041403.cl"},
	{0x95AF99A6FB6B7944ULL, "plt_and200pt_night_terrors.nss"},
	{0x95AFD588DF26BF92ULL, "vxht_0_hightown_cac_0_0.vlm"},
	{0x95AFED8599D810B0ULL, "vxht_bartrand_n_650_0_0.vlm"},
	{0x95B036E8C5C57C67ULL, "vxht_bartrand_n_644_0_0.vlm"},
	{0x95B0B84808C5155CULL, "mer240_jansen6050386.cl"},
	{0x95B0C8C08346DC3BULL, "fen110cr_hunter_captain.utc"},
	{0x95B119D83F647F23ULL, "136458527.ogg"},
	{0x95B18049F69E83E1ULL, "ca_caverns1_24424.rml"},
	{0x95B1A29B02E08589ULL, "482342945.ogg"},
	{0x95B2285C4F4438A1ULL, "hlac_ca_caverns1_24422_6x.dds"},
	{0x95B256D29E55FAAFULL, "w_bonepitdestroyed_a3_127_0.mao"},
	{0x95B262D29E560F4CULL, "w_bonepitdestroyed_a3_127_0.mmh"},
	{0x95B268D29E56193EULL, "w_bonepitdestroyed_a3_127_0.msh"},
	{0x95B289A866FD14EAULL, "6041132_f.fxe"},
	{0x95B2A640AF289A7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_medium.ani"},
	{0x95B2A9855AB69F3BULL, "232764001.ogg"},
	{0x95B2CB13FEAA4D33ULL, "mer221_sabin6048343.cl"},
	{0x95B2D578A1420578ULL, "mag160_feynriel6107436.cl"},
	{0x95B315018B2D6B5EULL, "74428788.ogg"},
	{0x95B35BBB1D4DCA70ULL, "w_blightlands_164_0.msh"},
	{0x95B369BB1D4DE18AULL, "w_blightlands_164_0.mmh"},
	{0x95B36DBB1D4DE8C1ULL, "w_blightlands_164_0.mao"},
	{0x95B4080C43A89087ULL, "328358352.ogg"},
	{0x95B4EB1A44681E02ULL, "art\\vfx\\combat\\vfx_w_wns_front_close5_c.impa.gad"},
	{0x95B4EC732F4AEEABULL, "zz_isa_debug6133956.cl"},
	{0x95B5788E8AD34076ULL, "mag160_varian_ilithis6016482.cl"},
	{0x95B586082A2EB566ULL, "14627091.ogg"},
	{0x95B7715B8CFF63D4ULL, "vxht_mansion2b_1rf_0_0.vlm"},
	{0x95B7800BB73DDE49ULL, "w_sundermount_63_0.mao"},
	{0x95B78C0BB73DF2E2ULL, "w_sundermount_63_0.mmh"},
	{0x95B78E0BB73DF598ULL, "w_sundermount_63_0.msh"},
	{0x95B7E8FC4FDD7B76ULL, "184011522.ogg"},
	{0x95B867D997AEA7EFULL, "hm_ave3_insurgent_guard1.mor"},
	{0x95B86988F88DE0CFULL, "llac_lt_darktown_-1_27.dds"},
	{0x95B8FBFC2B364CF8ULL, "vxht_mansion_n_4yq_0_0.vlm"},
	{0x95B978E646A38380ULL, "vxht_mansion2b_2j5_0_0.vlm"},
	{0x95B998A1AC8A4C13ULL, "6107531_m.fxe"},
	{0x95B9C7131D943931ULL, "vxlt_darktown_5b4_0_0.vlm"},
	{0x95B9CA7C85F5D7C1ULL, "279953893.ogg"},
	{0x95BA48CD2D75CBF5ULL, "hlac_lt_gallowstemplar_f_2391_7z.dds"},
	{0x95BA92A1C9977535ULL, "hlac_lt_0_lowtown_docks_-65538_2y.dds"},
	{0x95BB2FC1BC0F8637ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_action.gad"},
	{0x95BBD19BF9047658ULL, "vxlt_hangedman_n_251_0_0.vlm"},
	{0x95BBD80DE032442EULL, "follower_banter6163142.cl"},
	{0x95BBFE2DDE14C089ULL, "vxw_encounter2_8q_0_0.vlm"},
	{0x95BC0E1A446E78F9ULL, "art\\vfx\\combat\\vfx_w_wns_front_close5_c.impa.evt"},
	{0x95BC398F02CC802DULL, "6062801_m.fxe"},
	{0x95BC9B9F06D0D658ULL, "769419781.ogg"},
	{0x95BD1E52338FA085ULL, "vxht_0_hightown_bc7_0_0.vlm"},
	{0x95BD1F5F864FBB8BULL, "qcr000_arishok6042485.cl"},
	{0x95BD6C9CCB712A57ULL, "5361297.ogg"},
	{0x95BD785577A688A6ULL, "705140850.ogg"},
	{0x95BDA69FD9896E31ULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnl.ani"},
	{0x95BDDB82EBFDA26BULL, "character_rendertodepth_p5.psh.fxo.dx11.assembly"},
	{0x95BDE10437941B59ULL, "6148802_m.fxe"},
	{0x95BE09714C164BBBULL, "6169096_m.fxe"},
	{0x95BE236924C80EACULL, "179829732.ogg"},
	{0x95BE553A4756E78CULL, "hlac_lt_blackemporium_1055_16y.dds"},
	{0x95BE64BF3B397A15ULL, "vxht_0_hightown_a3_c0v_0_0.vlm"},
	{0x95BE8B46CB77E001ULL, "one220_yevhen6103373.cl"},
	{0x95BE9BBB64231DDAULL, "static_static_p38.psh.fxo.dx11.assembly"},
	{0x95BEA8E10A90EA4BULL, "6023970_m.fxe"},
	{0x95BEC0669095FDD2ULL, "vxht_bartrand_n_5r8_0_0.vlm"},
	{0x95BED45DF9C897CDULL, "6149889_m.fxe"},
	{0x95BEF46790CEAC3AULL, "270236891.ogg"},
	{0x95BF086A44E83989ULL, "vxlt_0_lowtown_a3_n_c_bz4_0_0.vlm"},
	{0x95BF0D1955C80465ULL, "llac_ht_mansion2b_818_9.dds"},
	{0x95BF2BD7FBA2317CULL, "hlac_lt_gallowscourtyard_n_-65536_0y.dds"},
	{0x95BF5CEFE72ED7A6ULL, "6149345_m.fxe"},
	{0x95BF6BA5172ADA41ULL, "85689227.ogg"},
	{0x95BF84C4CC8FD41AULL, "vxw_woundedcoast_1fp_0_0.vlm"},
	{0x95BF90AC6866690BULL, "403019288.ogg"},
	{0x95C06FD980C99EA3ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.dur.ani"},
	{0x95C0C7CCB45036E1ULL, "mrl210_merrill6102141.cl"},
	{0x95C0EE13E31DB2D5ULL, "mcr370_negotiation6078778.cl"},
	{0x95C12991A843E502ULL, "679105686.ogg"},
	{0x95C144C1BC1415C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_action.evt"},
	{0x95C24D5C396ADAE3ULL, "vxlt_ruinedhouse_n_gr_0_0.vlm"},
	{0x95C268C66EA7C6B6ULL, "abi_w_mighty_blow.bnk"},
	{0x95C2E653E557C3BBULL, "291356351.ogg"},
	{0x95C3309785B3E104ULL, "blood_wave_l3.mao"},
	{0x95C4B19FD98F9FD3ULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnl.gad"},
	{0x95C4BAEA41ED508CULL, "vxht_brothel_n_6gv_0_0.vlm"},
	{0x95C51043739FDB0CULL, "t3_eye_ice.tnt"},
	{0x95C581E29CAC65F0ULL, "l_env_wal_hightown_day_dist_oneshots.wbd"},
	{0x95C5A6BA94BB8472ULL, "6084763_m.fxe"},
	{0x95C5DFE0797E2833ULL, "489517913.ogg"},
	{0x95C5E3387EF57C54ULL, "vxht_mansion2b_1ug_0_0.vlm"},
	{0x95C60649E050BD87ULL, "6113967_m.fxe"},
	{0x95C649947D3ABD69ULL, "art\\vfx\\creatures\\fxc_withfn_c.impa.ani"},
	{0x95C6577882BA9163ULL, "271098927.ogg"},
	{0x95C665F809AE513DULL, "vxlt_hangedman_n_2hw_0_0.vlm"},
	{0x95C75986D1DC5187ULL, "art\\characters\\playercharacter\\elffemale\\boo_lgte.phy"},
	{0x95C763624A604662ULL, "393867433.ogg"},
	{0x95C78D6D115566D8ULL, "101840754.ogg"},
	{0x95C7AA012B2D9361ULL, "vxdr_primevaltomb_2xk_0_0.vlm"},
	{0x95C7AD2675CDB3FDULL, "hlac_lt_0_lowtown_a2_n_c_-131073_2y.dds"},
	{0x95C7F69E4355505DULL, "381674767.ogg"},
	{0x95C80F12E464B232ULL, "6119241_m.fxe"},
	{0x95C8A8F5DA6105FAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforward.gad"},
	{0x95C8F0FFA40BA9CCULL, "art\\characters\\playercharacter\\elfmale\\hlf_meda.msh"},
	{0x95C8F6FFA40BB3BEULL, "art\\characters\\playercharacter\\elfmale\\hlf_meda.mmh"},
	{0x95C902FFA40BC855ULL, "art\\characters\\playercharacter\\elfmale\\hlf_meda.mao"},
	{0x95C91C659A012139ULL, "qcr220_grey_wardens6135937.cl"},
	{0x95C942048A7D3D25ULL, "714563467.ogg"},
	{0x95C96B035FE33AC5ULL, "894867126.ogg"},
	{0x95C9DE0C5B10D93DULL, "lgt300cr_gang_3_lt.utc"},
	{0x95CA161A447AC3E8ULL, "art\\vfx\\combat\\vfx_w_wns_front_close5_c.impa.ani"},
	{0x95CAAB6BF4C298ADULL, "6069901_m.fxe"},
	{0x95CAAE9FD9940638ULL, "art\\characters\\creatures\\mubari\\animation\\rn_lnl.evt"},
	{0x95CAB606EB28A3D6ULL, "835081676.ogg"},
	{0x95CAC366FA4F0D89ULL, "vxw_blightlands_ok_0_0.vlm"},
	{0x95CAFE1ECFB0E5F4ULL, "6142113_m.fxe"},
	{0x95CB494C1F9A3E46ULL, "188737178.ogg"},
	{0x95CB902F2585B905ULL, "vxht_0_hightown_bl6_0_0.vlm"},
	{0x95CBC97C5285B189ULL, "28045789.ogg"},
	{0x95CC2097CC858079ULL, "6026979_m.fxe"},
	{0x95CC7A47E6FB55CDULL, "6085510_m.fxe"},
	{0x95CC7ADC5E3DD58BULL, "6097595_m.fxe"},
	{0x95CCCF560C87294CULL, "llac_lt_0_lowtown_docks_a3_n_-3_4.dds"},
	{0x95CCD69C2D2F9295ULL, "vxht_0_hightown_a3_c5u_0_0.vlm"},
	{0x95CD3E86600D3E5BULL, "one350_vinell6087506.cl"},
	{0x95CD40D8B4B71947ULL, "hlac_ht_chantry_n_2534_29x.dds"},
	{0x95CD64947D4109EBULL, "art\\vfx\\creatures\\fxc_withfn_c.impa.gad"},
	{0x95CD92BA6D7AF78FULL, "one111_gamlen6078534.cl"},
	{0x95CDA3F2CDFB7523ULL, "vxlt_0_lowtown_docks_c57_0_0.vlm"},
	{0x95CDB4DBB950D502ULL, "ca_caverns1_mines_mmap_ph.dds"},
	{0x95CDB793741E5F97ULL, "6034920_f.fxe"},
	{0x95CDC4EF05F3BAE4ULL, "hlac_lt_0_lowtown_docks_196607_1x.dds"},
	{0x95CDE8CEDE7F04E2ULL, "782260670.ogg"},
	{0x95CE01BA6EBD1A35ULL, "vxht_keep_n_5zq_0_1.vlm"},
	{0x95CE85D92457340DULL, "lt_0_lowtown_docks_162_0.msh"},
	{0x95CE93D924574C40ULL, "lt_0_lowtown_docks_162_0.mao"},
	{0x95CE97D924575293ULL, "lt_0_lowtown_docks_162_0.mmh"},
	{0x95CEABF5DA657611ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforward.evt"},
	{0x95CEBC70BDD0CBBAULL, "594565136.ogg"},
	{0x95CEBDF4BFF49C3BULL, "802958959.ogg"},
	{0x95CEE31A6962DB3EULL, "vxlt_shop_60_0_0.vlm"},
	{0x95CF5CC1BC207BBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_action.ani"},
	{0x95CFE5146E509863ULL, "mrl210_merrill6116384.cl"},
	{0x95D05D8D6E51A78CULL, "lgt110_orz_harrowmont6109025.cl"},
	{0x95D0831BF9D22A38ULL, "qun221_petrice6083647.cl"},
	{0x95D0D192062E2524ULL, "553653788.ogg"},
	{0x95D132B6EC823F48ULL, "vxdr_primevaltomb_2xn_0_0.vlm"},
	{0x95D18627A4C95D64ULL, "6080447_f.fxe"},
	{0x95D19AA6FB88613FULL, "plt_and200pt_night_terrors.plo"},
	{0x95D1FC5736573FF7ULL, "gen00fl_fenris6127997.cl"},
	{0x95D20D52526747C0ULL, "qun210cr_guard6206325.cl"},
	{0x95D2FF4DE021E12FULL, "hlac_ht_mansion2_818_52z.dds"},
	{0x95D4381F509444E1ULL, "6033595_m.fxe"},
	{0x95D461947D472350ULL, "art\\vfx\\creatures\\fxc_withfn_c.impa.evt"},
	{0x95D4D73FA2B5F767ULL, "6070128_f.fxe"},
	{0x95D52BA8F7C71494ULL, "vxlt_blackemporium_14r_0_0.vlm"},
	{0x95D586D980DC218DULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justfeet.dur.gad"},
	{0x95D5A1928E2A2AC3ULL, "vxht_chantry_n_36p_0_0.vlm"},
	{0x95D5BCC6B23A025EULL, "10305782.ogg"},
	{0x95D644A25CCD3E14ULL, "vxw_encounter3_4c_0_0.vlm"},
	{0x95D64D9FF82181D3ULL, "llac_lt_0_lowtown_a2_n_c_-262144_1.dds"},
	{0x95D66FE451AD17B0ULL, "dae000_arishok6067834.cl"},
	{0x95D6E87A905CE598ULL, "t3_wdg_br2.tnt"},
	{0x95D747DC5BF5EED8ULL, "vxht_bartrand_n_4b7_0_0.vlm"},
	{0x95D777F7F57BF7BCULL, "vxlt_0_lowtown_a2_n_c_hp8_0_0.vlm"},
	{0x95D7AF3B2FDE03DAULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf02.phy"},
	{0x95D7B590E69B8A27ULL, "534038916.ogg"},
	{0x95D8EE1E02BF28E9ULL, "llac_lt_hangedman_n_-1_10.dds"},
	{0x95D8F927DD4459B5ULL, "vxlt_undercitydungeon_18y_0_0.vlm"},
	{0x95D8FFBF801533D6ULL, "6093671_m.fxe"},
	{0x95D9094EF40885ACULL, "6099543_m.fxe"},
	{0x95D93543EC452609ULL, "vxw_blightlands_pj_0_0.vlm"},
	{0x95D968DF5A2B0439ULL, "vxlt_undercityencounter_1dt_0_0.vlm"},
	{0x95D990685F520B88ULL, "zz_mcr_debug6080874.cl"},
	{0x95D9ACCEF3E65796ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_altar_d.dds"},
	{0x95D9C9813E03EC9CULL, "one100dg_vanard6173969.cl"},
	{0x95D9E3EF77E6B0B8ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyn_0d.dds"},
	{0x95DABC7103FCE298ULL, "vxlt_gallowscourtyard_7ax_0_0.vlm"},
	{0x95DAD59F19B54423ULL, "vxdr_primevalentrance_1ok_0_0.vlm"},
	{0x95DAEAAF5A849D80ULL, "one100dg_nabil6059784.cl"},
	{0x95DAED2B30027CBBULL, "6163296_m.fxe"},
	{0x95DB443332E6EC3CULL, "llac_lt_0_lowtown_docks_-65539_1.dds"},
	{0x95DBB7BDCFA6EFDEULL, "gen00fl_aveline6122105.cl"},
	{0x95DC353E09452622ULL, "6085122_m.fxe"},
	{0x95DC687EB07C051AULL, "vxw_woundedcoast_1lr_0_0.vlm"},
	{0x95DC7D8529027F38ULL, "hlac_ca_caverns1_mines_24424_20z.dds"},
	{0x95DCC3F5DA71DCE0ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walkforward.ani"},
	{0x95DD6B04A8E72ED0ULL, "art\\vfx\\combat\\vfx_gib_arm_tmpa.mao"},
	{0x95DD809757622D4FULL, "819092068.ogg"},
	{0x95DDBFF1DF0AAB5CULL, "isa311cr_castillon.utc"},
	{0x95DE41CD7B31D399ULL, "vxlt_ruinedhouse_n_i2_0_0.vlm"},
	{0x95DE4506F20DE12CULL, "vxht_mansion_n_3wd_0_0.vlm"},
	{0x95DE5689D3AE4A4EULL, "pro000_opening_pc6001210.cl"},
	{0x95DEAC55DE607D91ULL, "gen_im_trs_qun_lgt_cht_r1.uti"},
	{0x95DEACE13916F550ULL, "110147167.ogg"},
	{0x95DF0A9B6E26EDB1ULL, "mag160_varian_ilithis6016506.cl"},
	{0x95DF1A5134172771ULL, "llac_lt_0_lowtown_-65535_7.dds"},
	{0x95DF20E0155D47D2ULL, "6104261_m.fxe"},
	{0x95DF282EC4ECA1BFULL, "vxht_0_hightown_a2_b3y_0_0.vlm"},
	{0x95DFA493DE7857C8ULL, "vxdr_primevaltomb_2wo_0_0.vlm"},
	{0x95E034ECFF34A22EULL, "593213005.ogg"},
	{0x95E1ECB5E26B0C15ULL, "282150142.ogg"},
	{0x95E26CE904D819B0ULL, "vxht_mansion2b_2at_0_0.vlm"},
	{0x95E2A023B31C7BDBULL, "vxlt_undercitydungeon_11p_0_0.vlm"},
	{0x95E2D53A23484B43ULL, "6001195_f.fxe"},
	{0x95E2DB65F28724DCULL, "6074344_m.fxe"},
	{0x95E432881CA36A27ULL, "vxdr_primevaltomb_7yw_0_0.vlm"},
	{0x95E4367378E6FBECULL, "hlac_dr_primevaltomb_658_1z.dds"},
	{0x95E47B025C91FDF3ULL, "924971869.ogg"},
	{0x95E4C7EB81A67E35ULL, "gen00fl_isabela_init6076483.cl"},
	{0x95E4DFDE4FFBDB81ULL, "dae300_petrice6097902.cl"},
	{0x95E4F2D30C5B1968ULL, "vxdr_0_deeproads_cpe_0_0.vlm"},
	{0x95E4F361F1FB03BAULL, "511057603.ogg"},
	{0x95E5A4EBB61A5A9DULL, "6217397_m.fxe"},
	{0x95E5B9B94DEC0758ULL, "vxht_bartrand_n_4g4_0_0.vlm"},
	{0x95E5BFC7D585D34CULL, "6174071_m.fxe"},
	{0x95E5D06F942E7DC6ULL, "lt_gallowscourtyard_n_53_0.rml"},
	{0x95E5FF30D20E9B83ULL, "mag110_ws_banter6079653.cl"},
	{0x95E638D44BCE73DDULL, "hlac_w_sundermount_393212_0x.dds"},
	{0x95E6ADC402CE614CULL, "dae000_gamlen6149461.cl"},
	{0x95E7538DA79DFE85ULL, "vxht_mansion_n_4yh_0_0.vlm"},
	{0x95E78BCDBF35EA3CULL, "gen00fl_aveline6112719.cl"},
	{0x95E80FFC6577D19FULL, "6111688_m.fxe"},
	{0x95E8774D0A1A0FC4ULL, "vxlt_0_lowtown_a2_n_c_hsy_0_0.vlm"},
	{0x95E88865D1C195E8ULL, "6116376_m.fxe"},
	{0x95E8E1D78EBC340BULL, "one100_solivitus6066030.cl"},
	{0x95EA2E6F461BBF0CULL, "background_names.gda"},
	{0x95EA656E9C932E40ULL, "6127574_m.fxe"},
	{0x95EA9BF690974676ULL, "hlac_ht_0_hightown_a2_65536_3y.dds"},
	{0x95EAE505B0941D52ULL, "vxw_sundermount_2z2_0_1.vlm"},
	{0x95EAF8311F33EE2FULL, "sf_d_xy16ic32.vsh.fxo"},
	{0x95EB1C6E805470BFULL, "ss_varian_ilithis.bnk"},
	{0x95EB21FB593A46BBULL, "plt_cod_hst_starkhaven.plo"},
	{0x95EB2E8C0DD61C2EULL, "6128970_f.fxe"},
	{0x95EB6E3D59E0DDC6ULL, "6074427_m.fxe"},
	{0x95EB8AC67EB8C31BULL, "dae011_wench_mob6142278.cl"},
	{0x95EB9381A9053274ULL, "hlac_ht_chantry_n_2534_30y.dds"},
	{0x95EBD1D52FC84833ULL, "gen_im_arm_cht_elf_coma_t2.uti"},
	{0x95EBDB6BA695989FULL, "llac_lt_0_lowtown_docks_a3_-131073_0.dds"},
	{0x95EC2F6143569E16ULL, "546101539.ogg"},
	{0x95EC3CF69D87A11AULL, "669879995.ogg"},
	{0x95ECA37EF091E44BULL, "zz_followers_banter6157687.cl"},
	{0x95ED4AD35DD942D2ULL, "6072803_m.fxe"},
	{0x95ED4DFB58ED99ECULL, "849049550.ogg"},
	{0x95ED637F723CC852ULL, "t3_wpn_spb.tnt"},
	{0x95EDEA7368C67D80ULL, "art\\characters\\playercharacter\\elfmale\\animation\\blendtree\\pp_fenris.anb"},
	{0x95EE7AAAF5810875ULL, "6128348_f.fxe"},
	{0x95EE99C289CC9B6CULL, "6073936_m.fxe"},
	{0x95EEC0DAF7F49B78ULL, "hlac_lt_backalley_n_65536_5y.dds"},
	{0x95EEFD54F8173A1AULL, "6176314_m.fxe"},
	{0x95EFC3BCDBB26468ULL, "vxht_chantry_2lg_0_0.vlm"},
	{0x95EFD15338C2722DULL, "vxht_0_hightown_a2_avk_0_0.vlm"},
	{0x95EFD4BC366E2268ULL, "399558050.ogg"},
	{0x95F01AE958F845B2ULL, "w_bonepitdestroyed_a3_12_0_ll.mmh"},
	{0x95F04DC97C7A706AULL, "77942115.ogg"},
	{0x95F06132125EB678ULL, "hlac_w_sundermount_196609_0z.dds"},
	{0x95F07CD93E4069A6ULL, "hlac_ht_mansion2_818_21y.dds"},
	{0x95F0E74A81C25485ULL, "additiveclouds01.mao"},
	{0x95F14D42E47D3D5CULL, "6079044_m.fxe"},
	{0x95F1A3615C2A1B21ULL, "566041263.ogg"},
	{0x95F206AE1C9B06DCULL, "art\\characters\\playercharacter\\humanfemale\\bdy_rob_magc.mao"},
	{0x95F20A8AADA8A671ULL, "6133599_f.fxe"},
	{0x95F24C62B51B0002ULL, "fxa_cen_aoe_smoke_add.mao"},
	{0x95F2862A67A5EFC4ULL, "6148094_m.fxe"},
	{0x95F2ACF99538D6A7ULL, "qun110_petrice6037701.cl"},
	{0x95F2CEA9EAEA04E7ULL, "hlac_lt_0_lowtown_a2_n_65537_4y.dds"},
	{0x95F34A2010BE5113ULL, "follower_banter6160661.cl"},
	{0x95F39436563264E4ULL, "lt_0_lowtown_docks_a3_n_231_0.mmh"},
	{0x95F3983656326B37ULL, "lt_0_lowtown_docks_a3_n_231_0.mao"},
	{0x95F39C431CEECC06ULL, "vxht_0_hightown_a2_a7g_0_0.vlm"},
	{0x95F3AA3656328986ULL, "lt_0_lowtown_docks_a3_n_231_0.msh"},
	{0x95F3D9BA6DE2BEAFULL, "follower_banter6026237.cl"},
	{0x95F42B963FE21FD8ULL, "vxht_bartrand_n_4l5_0_0.vlm"},
	{0x95F4996DCE43E5EAULL, "397647385.ogg"},
	{0x95F49D1FD71F4750ULL, "hlac_lt_0_lowtown_a2_n_c_-2_1z.dds"},
	{0x95F4B26AB4D7BB13ULL, "gen_im_trs_drk_two_2ha_r1.uti"},
	{0x95F4CDAFADB5353DULL, "541582080.ogg"},
	{0x95F4D85934F72607ULL, "vxlt_0_lowtown_docks_a3_n_d3f_0_0.vlm"},
	{0x95F50BE2B57B014EULL, "274463341.ogg"},
	{0x95F51F10F11889BCULL, "377683226.ogg"},
	{0x95F522D532FE25C6ULL, "10430541.ogg"},
	{0x95F5D408DFC8DFEFULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgtb_0n.dds"},
	{0x95F681FEBC1B99B1ULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_bdy_ogra_0n.dds"},
	{0x95F6A5DC0A56EA94ULL, "vxht_0_hightown_c3r_0_1.vlm"},
	{0x95F6BA88967100E1ULL, "vxht_playershouse_n_1sj_0_0.vlm"},
	{0x95F6CA66AA1F546EULL, "750529493.ogg"},
	{0x95F7AAD82E73CADAULL, "6166993_m.fxe"},
	{0x95F7C11CCE36325FULL, "811060238.ogg"},
	{0x95F7DAFDD16D7DF4ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.dura.ani"},
	{0x95F855E58987B258ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyo_0d.dds"},
	{0x95F929ADB7FDF657ULL, "vxlt_0_lowtown_docks_n_b5x_0_1.vlm"},
	{0x95F92AA00997DD67ULL, "907127852.ogg"},
	{0x95F94C389468361AULL, "vxw_woundedcoast_1rl_0_0.vlm"},
	{0x95F9D7F0115F11ECULL, "6109010_f.fxe"},
	{0x95F9FDA97735BFCBULL, "vxht_0_hightown_n_aq7_0_0.vlm"},
	{0x95FA4FF975F1031AULL, "478887073.ogg"},
	{0x95FA70D8F41AFABCULL, "hlac_ht_chantry_2534_36z.dds"},
	{0x95FB25D2BCF52429ULL, "6101114_f.fxe"},
	{0x95FB5E09961717CCULL, "llac_lt_darktown_-1_37.dds"},
	{0x95FC2AA48D63C67DULL, "weapon2.psh.fxo"},
	{0x95FC2C3C8A30A3FAULL, "hlac_w_encounter1_-1_0y.dds"},
	{0x95FC32167FB768D8ULL, "525721231.ogg"},
	{0x95FCB07C81FC0998ULL, "hlac_w_bonepit_131074_0z.dds"},
	{0x95FCC9142744BA7FULL, "hlac_w_bonepitmining_a2_458756_0z.dds"},
	{0x95FD3CA4C5A1CB40ULL, "427798052.ogg"},
	{0x95FD4875EF5A3BCFULL, "vxht_mansion2b_27q_0_0.vlm"},
	{0x95FD6B53272C86A8ULL, "lgt000ip_tombs_scroll.utp"},
	{0x95FDD7FDD171E456ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.dura.gad"},
	{0x95FDD859E21FFD92ULL, "hlac_ht_0_hightown_n_-131072_1z.dds"},
	{0x95FE3599CDA87CE8ULL, "vxht_chantry_2id_0_0.vlm"},
	{0x95FE5F6578AEE0EAULL, "hlac_w_encounter2_-131070_0y.dds"},
	{0x95FE607EBE359A5BULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.impa.ani"},
	{0x95FEAB3347272CFFULL, "299877188.ogg"},
	{0x95FF4BB68C66AC6DULL, "zz_otl_debug6060586.cl"},
	{0x95FFB92575E8A7FAULL, "459426048.ogg"},
	{0x95FFD1EDD85065CDULL, "gen00fl_aveline6122777.cl"},
	{0x95FFDBA5C1DD3FB1ULL, "hlac_w_bonepit_393221_0y.dds"},
	{0x95FFF9A199224B6CULL, "vxdr_primevaltomb_3hs_0_0.vlm"},
	{0x960003636899C993ULL, "731429489.ogg"},
	{0x9600433E57C5ACC8ULL, "vxht_playershouse_n_1so_0_0.vlm"},
	{0x96009BBC43C13D27ULL, "875536302.ogg"},
	{0x960133258E298D40ULL, "zz_dae_debug6029839.cl"},
	{0x96016EFFA5EB3A6DULL, "tbe100st_magic_goods.utm"},
	{0x9601E12E6050E2EAULL, "mag131_keran6086282.cl"},
	{0x960204A3BEC20DACULL, "255130233.ogg"},
	{0x9602F9DB55D0A52CULL, "llac_lt_0_lowtown_a2_n_-1_4.dds"},
	{0x960332C37EA33B46ULL, "vxw_bonepitmining_a2_1wo_0_1.vlm"},
	{0x960359A6C23A7518ULL, "vxlt_0_lowtown_docks_bbp_0_0.vlm"},
	{0x96038DEC504F5DCFULL, "dae000cr_grand_cleric6148087.cl"},
	{0x9603B4ABFCDF9A85ULL, "47960728.ogg"},
	{0x96040A4E3EEB5C9EULL, "dae000_bartrand6027545.cl"},
	{0x9604A2412964DFA7ULL, "6083647_f.fxe"},
	{0x9604BB26E74993BBULL, "595392103.ogg"},
	{0x9604F2FDD17830B5ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.dura.evt"},
	{0x96050E99EDE518CBULL, "710604922.ogg"},
	{0x96052C6588671961ULL, "vxht_playershouse_n_1vi_0_0.vlm"},
	{0x960557A2ED5E1B05ULL, "vxht_0_hightown_bx2_0_0.vlm"},
	{0x96056330CE20CF05ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_imp_p.anb"},
	{0x96057E19FD24BA65ULL, "6076546_f.fxe"},
	{0x9606321CB82C9A2CULL, "580316973.ogg"},
	{0x9606B121C510E169ULL, "vxlt_gallowscourtyard_n_6ec_0_0.vlm"},
	{0x9606B4FBDDF608D1ULL, "t3_wpn_dsk.tnt"},
	{0x9606D6DD38082ACDULL, "dae000ip_crowd_mf8_talk9.utp"},
	{0x9606FBA0F6FA6DBFULL, "6164688_m.fxe"},
	{0x9606FCDF568B3641ULL, "337328928.ogg"},
	{0x9607DB45EB0F45C1ULL, "980413488.ogg"},
	{0x96088DB067942307ULL, "llac_ht_0_hightown_a2_-1_0.dds"},
	{0x9608FF5E2A4D0161ULL, "dae000cr_grand_cleric6173595.cl"},
	{0x96090E1705E85E4DULL, "840986884.ogg"},
	{0x96091AEDB6CF8E35ULL, "llac_lt_hangedman_n_-1_5.dds"},
	{0x9609237AC65CE526ULL, "549284501.ogg"},
	{0x9609CDFFA4430559ULL, "art\\characters\\playercharacter\\elfmale\\hlf_meda.phy"},
	{0x9609D4DAC7359271ULL, "vxlt_0_lowtown_docks_a3_d0r_0_0.vlm"},
	{0x9609EEDE77D0830CULL, "161644945.ogg"},
	{0x960A44A79E74B440ULL, "52434624.ogg"},
	{0x960A51F7E41B4093ULL, "179925497.ogg"},
	{0x960A67EBCB0201BEULL, "l_one100ar_ruin_ext.bnk"},
	{0x960AD4203CE613F8ULL, "6089471_f.fxe"},
	{0x960B3F4566CBEC4AULL, "gen00fl_varric6107299.cl"},
	{0x960BBA52E150544FULL, "vxht_mansion2b_20p_0_0.vlm"},
	{0x960BE183421E9B10ULL, "vxdr_0_deeproads_f0p_0_0.vlm"},
	{0x960BFF3834E65F15ULL, "vxlt_0_lowtown_a2_n_df1_0_0.vlm"},
	{0x960C5CF7AB509E46ULL, "map.gfx"},
	{0x960C687EBE41E552ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.impa.evt"},
	{0x960C79FC868A4BB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_1.gad"},
	{0x960D494D811568FCULL, "397103409.ogg"},
	{0x960D57E29CE9D3B2ULL, "l_env_wal_hightown_day_dist_oneshots.bnk"},
	{0x960D5D61B3706AD6ULL, "406652531.ogg"},
	{0x960E9D3FACE69116ULL, "art\\vfx\\creatures\\fxc_lotf_c.dura.evt"},
	{0x960EB51B49BBC548ULL, "vxht_playershouse_n_1vl_0_0.vlm"},
	{0x960EEE449194D771ULL, "6139399_f.fxe"},
	{0x960FC0ACA885FDF9ULL, "vxw_encounter3_gf_0_1.vlm"},
	{0x9610132DC3BCE9BAULL, "478830596.ogg"},
	{0x96102138DA35BD23ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.impa.evt"},
	{0x961044D59580F9C6ULL, "vxht_0_hightown_n_aif_0_0.vlm"},
	{0x96105CC67D0C9524ULL, "6072166_m.fxe"},
	{0x9610810A9AF7CAA2ULL, "957764880.ogg"},
	{0x96108C10B535047CULL, "587162584.ogg"},
	{0x96109E40F1024AF5ULL, "zz_mcr_debug6098520.cl"},
	{0x96117D544CA84D9CULL, "6013273_m.fxe"},
	{0x9611AC37E66AB22FULL, "vxht_store_cc_0_0.vlm"},
	{0x9611D32281B3BEDCULL, "300897245.ogg"},
	{0x9611D9CCB78CF7A1ULL, "zz_spy130cr_morton.utc"},
	{0x96128EFC868EDB44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_1.evt"},
	{0x96129EFF64ED0BB1ULL, "dae000_ineria6136316.cl"},
	{0x9612BF09F4B6551EULL, "810576919.ogg"},
	{0x9612CDFC4A4C6969ULL, "mcr310_argument6074378.cl"},
	{0x96131402AE1390A4ULL, "6025805_m.fxe"},
	{0x96135F53046BA464ULL, "hlac_lt_0_lowtown_65536_9y.dds"},
	{0x9613877EBE4838E5ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.impa.gad"},
	{0x9613D5C9FC9802DEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_0.pwk"},
	{0x9613F5F5188982CCULL, "w_bonepit_53_0.msh"},
	{0x9613FBF518898CBEULL, "w_bonepit_53_0.mmh"},
	{0x961407F51889A155ULL, "w_bonepit_53_0.mao"},
	{0x961418BD4B11DC24ULL, "6129257_m.fxe"},
	{0x96145917C3F129A0ULL, "art\\characters\\weapons\\unique\\meredith\\textures\\w_gsw_014a_0d.dds"},
	{0x96158976610A12D6ULL, "6119933_m.fxe"},
	{0x9615943FACECA0A1ULL, "art\\vfx\\creatures\\fxc_lotf_c.dura.gad"},
	{0x9616326FF081FD1EULL, "mag121_viveka6082594.cl"},
	{0x9616491279964541ULL, "gen00fl_aveline6116473.cl"},
	{0x9616C5B4332FC7E8ULL, "6196196_m.fxe"},
	{0x9616F22795E0280DULL, "vxlt_shop_g4_0_0.vlm"},
	{0x9616FA38DA3B99A4ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.impa.gad"},
	{0x9618A0832B7F8B1EULL, "hlac_lt_backalley_n_65536_2z.dds"},
	{0x9619444E607F7704ULL, "402157852.ogg"},
	{0x961A7FA60493870EULL, "art\\characters\\creatures\\rubblegolem\\c_profane_s.mao"},
	{0x961A9F3E02A05EC6ULL, "llac_ca_caverns1_mines_24426_4.dds"},
	{0x961ACEF8F8261808ULL, "hlac_lt_warehouse_1_1y.dds"},
	{0x961AEA283FEF7225ULL, "hlac_ca_caverns1_mines_24425_0x.dds"},
	{0x961B1F4CD67AE599ULL, "hlac_w_bonepitmining_a2_393218_0y.dds"},
	{0x961B693FF6B2FE28ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masd_0t.dds"},
	{0x961B8D9FE6249662ULL, "6072704_m.fxe"},
	{0x961BA87C4611F4EDULL, "vxw_bonepitmining_a2_2al_0_0.vlm"},
	{0x961BC03C99A20986ULL, "6071719_f.fxe"},
	{0x961C54E4F98F255CULL, "hlac_dr_dungeon_3217_11y.dds"},
	{0x961C8AD1FDFB5CAEULL, "6020784_f.fxe"},
	{0x961C953FACF2C14FULL, "art\\vfx\\creatures\\fxc_lotf_c.dura.ani"},
	{0x961CEEAEB0EDDB2EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_right.ani"},
	{0x961D806377132FB0ULL, "w_encounter3_252_0.msh"},
	{0x961D8E63771347CAULL, "w_encounter3_252_0.mmh"},
	{0x961D926377134E01ULL, "w_encounter3_252_0.mao"},
	{0x961E0938DA41D182ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.impa.ani"},
	{0x961E3A3A0124DFE3ULL, "weapon_rendertocubeshadows_p0.gsh.fxo.dx11"},
	{0x961ECE5550D22AA5ULL, "llac_lt_0_lowtown_-131071_0.dds"},
	{0x961F56100E9F6127ULL, "llac_lt_darktown_-1_63.dds"},
	{0x961FE1828EE1DC04ULL, "vxht_mansion2_1th_0_0.vlm"},
	{0x96203A0A2F69475CULL, "vxht_mansion2b_2ji_0_0.vlm"},
	{0x9620542B9838E993ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_4y.dds"},
	{0x9620555BDE1D4D88ULL, "w_bonepitdestroyed_a3_86_0.mmh"},
	{0x9620615BDE1D622BULL, "w_bonepitdestroyed_a3_86_0.mao"},
	{0x9620735BDE1D80B2ULL, "w_bonepitdestroyed_a3_86_0.msh"},
	{0x962081C01241D6F0ULL, "6026174_m.fxe"},
	{0x9620A6FC869B413DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_1.ani"},
	{0x9620D8BF96AE94BCULL, "vxht_0_hightown_a3_9zg_0_0.vlm"},
	{0x96217B7BA44BC07CULL, "llac_lt_warehouse_1_0.dds"},
	{0x962186833F2C5B93ULL, "and111_donation_box.bnk"},
	{0x96219596CD1139C9ULL, "6042205_m.fxe"},
	{0x96224694516B8B8BULL, "vxlt_0_lowtown_a2_n_fid_0_0.vlm"},
	{0x9622B147B6451010ULL, "mag370_meredith6172404.cl"},
	{0x9622FD7183FEAD96ULL, "489350818.ogg"},
	{0x96232DE0911472B5ULL, "pro000_opening_pc6001189.cl"},
	{0x962330437C209C6CULL, "w_bonepitmining_a2_31_0_lc.mmh"},
	{0x9623625619659013ULL, "clf221_jalen.bnk"},
	{0x9623A03AC58A6EA3ULL, "395967796.ogg"},
	{0x9623C20D440929DAULL, "vxht_bartrand_n_5o7_0_0.vlm"},
	{0x9624A8833F2ECBE3ULL, "and111_donation_box.cnv"},
	{0x9624FB7DAB2BFB5FULL, "qcr270_fake_meredith6066353.cl"},
	{0x9625127B9B98438AULL, "6095031_m.fxe"},
	{0x9625361DE2CFA515ULL, "zz_lgt_1_debug6114386.cl"},
	{0x96254A3B2A10846BULL, "10809346.ogg"},
	{0x96254F41904C761EULL, "vxht_bartrand_n_5vx_0_0.vlm"},
	{0x96256C775924704EULL, "hlac_ht_0_hightown_a3_65536_1x.dds"},
	{0x96257E2AB5B0F71CULL, "w_sundermount_48_0.mao"},
	{0x96258A2AB5B10BBFULL, "w_sundermount_48_0.mmh"},
	{0x9625902AB5B11661ULL, "w_sundermount_48_0.msh"},
	{0x9625D4C740614808ULL, "llac_lt_0_lowtown_a3_n_c_1_8.dds"},
	{0x9625E81A6FBFB723ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs107a_0n.dds"},
	{0x9626845619680063ULL, "clf221_jalen.cnv"},
	{0x9626B35E68EE48F6ULL, "gen00fl_varric6033606.cl"},
	{0x9626C2E73603DAF7ULL, "zz_area_signage11.utp"},
	{0x9627566B537EE2AEULL, "vxlt_gallowstemplar_5ch_0_0.vlm"},
	{0x9627B7DA20D244F8ULL, "6087671_f.fxe"},
	{0x9628283A88A54265ULL, "6068784_m.fxe"},
	{0x9628B163F4E6126CULL, "vxht_brothel_n_4fw_0_0.vlm"},
	{0x9628ECF01C1CA0F2ULL, "vxlt_0_lowtown_docks_n_aou_0_0.vlm"},
	{0x962987DBDEC3CD65ULL, "17182950.ogg"},
	{0x9629D31451A68230ULL, "zz_lgt_2_debug6206579.cl"},
	{0x9629DEF3F0C99690ULL, "weapon1.vsh.fxo"},
	{0x962A666754F723D8ULL, "vxca_cave_17x_0_1.vlm"},
	{0x962AE6AEB0FA0B7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_right.evt"},
	{0x962B382999788484ULL, "dae000_hate_follower_react6103242.cl"},
	{0x962B81104D210031ULL, "vxca_caverns1_mines_jgr_0_0.vlm"},
	{0x962C163A66A21CB6ULL, "hlac_lt_warehouse_65537_1z.dds"},
	{0x962CEFAD9FE7B710ULL, "lgt120_orz_bln_official6140198.cl"},
	{0x962CF1029AF7C05CULL, "263611443.ogg"},
	{0x962D6430CE422346ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_imp_p.mmh"},
	{0x962D8BCE8BE6DC28ULL, "239359180.ogg"},
	{0x962DB1AC12F95094ULL, "vxht_keep_n_6mk_0_0.vlm"},
	{0x962DB9AE486DF472ULL, "dragging.cur"},
	{0x962DEE9F1514A2A0ULL, "gen00fl_isabela6101549.cl"},
	{0x962F237433F775E3ULL, "vxw_woundedcoast_1ba_0_0.vlm"},
	{0x962FB29B26B32838ULL, "vxlt_gallowstemplar_3k4_0_0.vlm"},
	{0x962FB81A4966FD8AULL, "art\\vfx\\creatures\\fxc_golem_eburst_p.anb"},
	{0x963149E19E7DB2F8ULL, "619687234.ogg"},
	{0x9631521F3EEF9BB1ULL, "mag370_meredith6064388.cl"},
	{0x96317E80FA7500F9ULL, "ch2alphmultnt.psh.fxo"},
	{0x9631845205E6B5F1ULL, "mcr370_meredithcrazy_st.stg"},
	{0x96318C23A1FD4E67ULL, "ser270_emeric6112662.cl"},
	{0x9631C7AEB0FFF568ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walk_front_right.gad"},
	{0x9631D53CA38A80AAULL, "hlac_lt_gallowstemplar_n_2391_7y.dds"},
	{0x96324CFA5A583A16ULL, "zz_vrc_debug6064280.cl"},
	{0x963337B5263B595EULL, "vxht_brothel_n_52t_0_0.vlm"},
	{0x9633A83B302C5603ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf02.msh"},
	{0x9633B63B302C6E5DULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_hf02.mmh"},
	{0x9633C11E82428E9EULL, "vxht_bartrand_n_5yy_0_0.vlm"},
	{0x96342D3DF9E66D4DULL, "vxlt_0_lowtown_eop_0_0.vlm"},
	{0x963444BA2C694D02ULL, "vxlt_0_lowtown_docks_a3_n_cyv_0_0.vlm"},
	{0x9634639F127361EFULL, "vxht_0_hightown_bfz_0_0.vlm"},
	{0x963468A2433A0435ULL, "vxht_brothel_n_637_0_0.vlm"},
	{0x9634DB2BE04CFEB2ULL, "vxht_chantry_n_2kw_0_0.vlm"},
	{0x9634F0C4F4EDEBCFULL, "6170919_m.fxe"},
	{0x96356C300F3C7002ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ci_basic.msh"},
	{0x96356E300F3C73B8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ci_basic.mmh"},
	{0x96357A300F3C885BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ci_basic.mao"},
	{0x96359435C7CA39A0ULL, "vxlt_undercityencounter_13d_0_0.vlm"},
	{0x9635A98CD0FDA4F4ULL, "6096750_f.fxe"},
	{0x9635CCD2BC701328ULL, "vxca_cavern1_haunted_eic_0_1.vlm"},
	{0x963616C1ECD628B4ULL, "vxlt_0_lowtown_a2_n_gp9_0_0.vlm"},
	{0x963621C73CE042A5ULL, "hlac_lt_0_lowtown_a2_-1_3y.dds"},
	{0x963742A2B36947ABULL, "art\\vfx\\combat\\vfx_glb_sleep_c.cess.evt"},
	{0x96375953FE66C335ULL, "llac_lt_gallowstemplar_f_2394_0.dds"},
	{0x9638585A93F4566AULL, "and111_lirene6097130.cl"},
	{0x96389221AC576A8BULL, "821107990.ogg"},
	{0x9638C4EFBE551B7EULL, "dae011_conspiracy_mob6141281.cl"},
	{0x96392701BFED23E0ULL, "hlac_ht_keep_4519_4z.dds"},
	{0x9639520CD3637BB1ULL, "6061414_f.fxe"},
	{0x963968E85100883CULL, "6080246_m.fxe"},
	{0x9639791AB5648152ULL, "vxht_mansion2b_1my_0_0.vlm"},
	{0x9639C2B46EC18614ULL, "vxlt_0_lowtown_a2_ejf_0_0.vlm"},
	{0x963A491198801A88ULL, "hlac_ht_0_hightown_-131069_0z.dds"},
	{0x963AA82A37F1F796ULL, "6121606_m.fxe"},
	{0x963ABC3D3C571F8BULL, "6085501_m.fxe"},
	{0x963B00DEB5E0C46BULL, "ht_0_hightown_33_0_lc.mmh"},
	{0x963B1780F1944FB7ULL, "6136751_m.fxe"},
	{0x963B69183E92256EULL, "6174334_m.fxe"},
	{0x963C238904EF6914ULL, "vxht_keep_n_6hh_0_0.vlm"},
	{0x963C30224D8F10F0ULL, "494097119.ogg"},
	{0x963CC53C72CE0D04ULL, "vxht_mansion2_1bn_0_0.vlm"},
	{0x963D08A4797CA405ULL, "6166169_m.fxe"},
	{0x963D49CF84D756F8ULL, "hlac_lt_0_lowtown_-196609_1z.dds"},
	{0x963D4D5262554DE3ULL, "ef_amb_darktown_sta.mor"},
	{0x963D7011C68C1E1FULL, "6041358_m.fxe"},
	{0x963D99C47AFF127FULL, "81383219.ogg"},
	{0x963DA72C5B6A7F50ULL, "699177306.ogg"},
	{0x963DFC02C792478EULL, "vxdr_0_deeproads_ckb_0_0.vlm"},
	{0x963E028E6A9671FDULL, "dae000_bartrand6030647.cl"},
	{0x963E0BD3785F28D4ULL, "ave200_aveline_bar6116197.cl"},
	{0x963E5BA2B36F916CULL, "art\\vfx\\combat\\vfx_glb_sleep_c.cess.gad"},
	{0x963E615805BB8661ULL, "gen_im_trs_cot_med_glv_r0.uti"},
	{0x963E6200D26F2D7BULL, "vxht_0_hightown_a2_amp_0_0.vlm"},
	{0x963EC7B6A30D8826ULL, "vxlt_0_lowtown_a2_n_c_fig_0_0.vlm"},
	{0x963EF59598F02856ULL, "vxht_chantry_35b_0_1.vlm"},
	{0x963FDAEA3FD8D9F8ULL, "mrl150_merrill6093915.cl"},
	{0x964078958CE94369ULL, "vxlt_gallowscourtyard_n_6yg_0_0.vlm"},
	{0x9640D696C1D21A45ULL, "6169601_m.fxe"},
	{0x9640ED2D4D5DD958ULL, "hlac_w_encounter3_-131073_0y.dds"},
	{0x964119994135C79FULL, "hlac_dr_primevalentrance_1725_2z.dds"},
	{0x96412F0B925830AAULL, "bkl_link_smoke.mao"},
	{0x9641589AE7CF0126ULL, "197694116.ogg"},
	{0x96419943272515BFULL, "gen00fl_aveline6106931.cl"},
	{0x9641E11F5C288BBEULL, "811283335.ogg"},
	{0x9641FB087347FE97ULL, "fen110_fenris6097075.cl"},
	{0x9642A419D6E34881ULL, "vxht_mansion2b_2j8_0_0.vlm"},
	{0x9642B711F7CCC312ULL, "art\\vfx\\combat\\vfx_g_dual_att_left_c.impa.gad"},
	{0x96432D2A36434C00ULL, "6060193_f.fxe"},
	{0x964536EBDE154C4AULL, "llac_lt_0_lowtown_n_-196609_4.dds"},
	{0x96455AA2B375AE6AULL, "art\\vfx\\combat\\vfx_glb_sleep_c.cess.ani"},
	{0x96455F8AD1E70CDEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.cess.ani"},
	{0x9645C1E6495562C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_idle.ani"},
	{0x9645D0AA0008D1F2ULL, "vxlt_0_lowtown_docks_n_aqw_0_0.vlm"},
	{0x9645FCA22553E2E1ULL, "473175350.ogg"},
	{0x964644E4DEA33170ULL, "gen_im_wep_war_two_lore_03.uti"},
	{0x964663C5A5C68BC3ULL, "gen00fl_merrill6115835.cl"},
	{0x9646892CAC0359EDULL, "follower_banter6143299.cl"},
	{0x9646E0FE324AF7FCULL, "vxca_cave_2d7_0_0.vlm"},
	{0x96478024334F8CE4ULL, "238122552.ogg"},
	{0x9648349160B79E94ULL, "vxlt_0_lowtown_a2_emg_0_0.vlm"},
	{0x9648DADD8DAECF9EULL, "hlac_lt_0_lowtown_n_-131072_5x.dds"},
	{0x9648E16665D8355EULL, "hlac_lt_shop_-1_0x.dds"},
	{0x96492F9F475EEC59ULL, "616064835.ogg"},
	{0x9649A54E3BC240F1ULL, "6082124_m.fxe"},
	{0x9649AB5F883A9829ULL, "lt_0_lowtown_docks_a3_n_206_0.mao"},
	{0x9649B75F883AACC2ULL, "lt_0_lowtown_docks_a3_n_206_0.mmh"},
	{0x9649B95F883AB078ULL, "lt_0_lowtown_docks_a3_n_206_0.msh"},
	{0x9649BA11F7D2E749ULL, "art\\vfx\\combat\\vfx_g_dual_att_left_c.impa.evt"},
	{0x9649D416C290C98CULL, "vxw_blightlands_be_0_0.vlm"},
	{0x9649D9666A2279A5ULL, "779500500.ogg"},
	{0x964A56EA36C1E075ULL, "6048492_m.fxe"},
	{0x964A87401B75D85BULL, "tacticsuserpresets.gda"},
	{0x964A9565F6E58194ULL, "vxht_keep_n_6ci_0_0.vlm"},
	{0x964B39EBCB394AC4ULL, "l_one100ar_ruin_ext.wbd"},
	{0x964B41A197A6BFEFULL, "vfx_chunkyparticle_p2.vsh.fxo"},
	{0x964B6B924F1B6701ULL, "628159394.ogg"},
	{0x964BF48F042FEA3CULL, "hlac_lt_0_lowtown_docks_65533_4z.dds"},
	{0x964C072E17E3A6E3ULL, "vxw_woundedcoast_1hc_0_0.vlm"},
	{0x964CB1C2C5A28786ULL, "vxdr_primevaltomb_w1_0_0.vlm"},
	{0x964CCCE6495B9463ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_idle.gad"},
	{0x964CE9924336FBACULL, "496558120.ogg"},
	{0x964D56F089964220ULL, "llac_lt_0_lowtown_docks_a3_-196609_1.dds"},
	{0x964F4AD8043B0536ULL, "655221715.ogg"},
	{0x964F58147734E667ULL, "hm_icarver_h3s1.mor"},
	{0x964F9BF86D673574ULL, "6076256-sisterunderattack.cut"},
	{0x965039554F2710C7ULL, "vxht_chantry_n_36l_0_0.vlm"},
	{0x9651284E72745402ULL, "hlac_lt_gallowstemplar_n_7198_3y.dds"},
	{0x9651CFDC8B64B19BULL, "vxlt_gallowstemplar_f_4so_0_0.vlm"},
	{0x9651F6ED05604A25ULL, "569350267.ogg"},
	{0x9652778AD1F1BFAFULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.cess.evt"},
	{0x965287A9946D0991ULL, "798174144.ogg"},
	{0x9652AC0229612C2EULL, "vxlt_gallowstemplar_5rk_0_0.vlm"},
	{0x9652B3715D42842DULL, "vxht_brothel_n_6g3_0_0.vlm"},
	{0x9652D02F7B9343E7ULL, "lt_0_lowtown_a2_n_46_0.rml"},
	{0x9652D30E9D812BF5ULL, "898242851.ogg"},
	{0x96530187330F045EULL, "hlac_lt_gallowscourtyard_-65535_1z.dds"},
	{0x9653500F6793396DULL, "6164880_m.fxe"},
	{0x96536907F95A34BFULL, "vxht_brothel_n_4fx_0_0.vlm"},
	{0x965378739CBF5DB7ULL, "6101133_m.fxe"},
	{0x9653A9E649617748ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_below_belt_reaction_idle.evt"},
	{0x96541B724F83038DULL, "34583307.ogg"},
	{0x9654B69F7B1797FFULL, "lt_gallowscourtyard_probe_5fy.dds"},
	{0x96552EAF7D230190ULL, "6102522_m.fxe"},
	{0x965569128F288077ULL, "6058886_m.fxe"},
	{0x96556E8CC998A25EULL, "108440784.ogg"},
	{0x965781DDCECE4C0BULL, "t3_wdg_w20.tnt"},
	{0x9657A211F7DEFBB8ULL, "art\\vfx\\combat\\vfx_g_dual_att_left_c.impa.ani"},
	{0x9657A83F336C449EULL, "6019255_m.fxe"},
	{0x9657BB1A4988557DULL, "art\\vfx\\creatures\\fxc_golem_eburst_p.mmh"},
	{0x96585D64AC369100ULL, "vxht_mansion2b_258_0_0.vlm"},
	{0x965869B2A11F0B18ULL, "vxw_bonepitdestroyed_a3_1jy_0_0.vlm"},
	{0x9658769E3146C696ULL, "vxca_cavern1_haunted_flx_0_0.vlm"},
	{0x96587E264F4F7446ULL, "vxht_0_hightown_n_avk_0_0.vlm"},
	{0x96589B7AD8CBBA5FULL, "hlac_lt_0_lowtown_a3_n_c_-131073_0z.dds"},
	{0x965921C344DD021BULL, "gen00fl_merrill6127207.cl"},
	{0x965934CD53804E94ULL, "283804113.ogg"},
	{0x9659788AD1F7E058ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.cess.gad"},
	{0x9659A8F656BA3E04ULL, "vxht_mansion2_1hl_0_0.vlm"},
	{0x9659AECA94E1DDF9ULL, "t3_shd_ash.tnt"},
	{0x9659C1254F23C5CEULL, "vxht_mansion2b_32n_0_0.vlm"},
	{0x9659EE1E86F6E5B5ULL, "466839261.ogg"},
	{0x965A03D03E636A78ULL, "gen00fl_fenris6110802.cl"},
	{0x965A0A231DA5E03FULL, "al_mage_t05.gda"},
	{0x965A49E00A1C0AC1ULL, "vxdr_slavers_1ak_0_0.vlm"},
	{0x965A790B09D9BF63ULL, "vxw_woundedcoast_1sb_0_0.vlm"},
	{0x965ADBF518C55975ULL, "w_bonepit_53_0.rml"},
	{0x965B1662D65DB968ULL, "dae000_sandal6148232.cl"},
	{0x965B4B476C34A7E2ULL, "573247554.ogg"},
	{0x965BABDCD6176257ULL, "dae000_pc_home_ambients6208492.cl"},
	{0x965C0057C338BE12ULL, "isa311_party_comments6117825.cl"},
	{0x965C46EF5B2DC7CFULL, "ave110cr_donnic_bandit_m.utc"},
	{0x965CE7931E4A2AF7ULL, "follower_banter6163488.cl"},
	{0x965D83BAF01EEE05ULL, "6143872_m.fxe"},
	{0x965DDB5A5760D3DEULL, "vxht_0_hightown_a2_c6n_0_0.vlm"},
	{0x965E45F3C7C07A07ULL, "442469426.ogg"},
	{0x965E5876DC264BF8ULL, "zz_fen_debug6165699.cl"},
	{0x965E7B3022B73DB6ULL, "mcr310_confused_guard6077013.cl"},
	{0x965E984EA4F121A6ULL, "vxlt_0_lowtown_a2_n_c_e21_0_0.vlm"},
	{0x965F2644084BA70FULL, "art\\vfx\\materials\\chunky\\vfx_c_falloff_add_vt.mao"},
	{0x96608C40C110F6E4ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_imp_p.impa.ani"},
	{0x9660913D405FDE1CULL, "6095107_m.fxe"},
	{0x9660F853C3B0D13AULL, "vxlt_darktown_2ua_0_0.vlm"},
	{0x966182928C15C664ULL, "narrative_varric6124300.cl"},
	{0x9661BE36371F807AULL, "431804374.ogg"},
	{0x9661D56A58F6758EULL, "472331578.ogg"},
	{0x9661EEADFB760599ULL, "vxlt_0_lowtown_docks_a3_n_czi_0_1.vlm"},
	{0x966225768B8882D3ULL, "zz_tbe100_barber6190999.cl"},
	{0x9662CF136BCC4241ULL, "vxht_keep_n_5hj_0_0.vlm"},
	{0x96632AC4E7CF9ECCULL, "mcr370cs_rules_surv_m.cut"},
	{0x9663DBDDDDE99135ULL, "hlac_ca_caverns1_mines_24424_10x.dds"},
	{0x96642C2337613BF5ULL, "prccontentmanager.gfx"},
	{0x96643D61A65E73F5ULL, "gen00fl_anders6119844.cl"},
	{0x9664CF620B0183F1ULL, "vxlt_undercityencounter_1dl_0_0.vlm"},
	{0x96651C5D0689B33AULL, "6167590_m.fxe"},
	{0x966523D3CBDA86B9ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_5x.dds"},
	{0x9665258E9FF7CBFBULL, "6107438_m.fxe"},
	{0x9665ADF3AF39D048ULL, "art\\vfx\\models\\vfx_lightning_ball_line.mmh"},
	{0x9665B0A67AD5C38EULL, "vxht_chantry_2yb_0_0.vlm"},
	{0x9665CBF3AF3A0372ULL, "art\\vfx\\models\\vfx_lightning_ball_line.msh"},
	{0x96664F4193F14084ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_bonepile_sa.dds"},
	{0x9666DB8F93152398ULL, "vxw_bonepitdestroyed_a3_1ov_0_0.vlm"},
	{0x9666FB8590F1F7F7ULL, "gen00fl_aveline6002293.cl"},
	{0x9667102FF8F46EA8ULL, "hlac_w_woundedcoast_-262138_0z.dds"},
	{0x96672891B772CAA2ULL, "art\\levels\\ztools\\default\\generic\\9ww_lamp.msh"},
	{0x96672A91B772CE58ULL, "art\\levels\\ztools\\default\\generic\\9ww_lamp.mmh"},
	{0x96673691B772E2FBULL, "art\\levels\\ztools\\default\\generic\\9ww_lamp.mao"},
	{0x96678940C1171046ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_imp_p.impa.gad"},
	{0x966853B933E99D12ULL, "ht_playershouse_n_mmi.gff"},
	{0x966879C2694D4468ULL, "lgt110_urn_guard6133017.cl"},
	{0x96687F1C8C3D4E0FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vent_sa.dds"},
	{0x9668F2CE317B7A1DULL, "6151884_m.fxe"},
	{0x966938C51FC52F41ULL, "vxdr_primevalentrance_1o1_0_0.vlm"},
	{0x9669DE71BA351FD5ULL, "fx_crescent01_blend.mao"},
	{0x966A138578BD069CULL, "hlac_lt_0_lowtown_a2_n_c_-3_6x.dds"},
	{0x966A66CBF80BCF58ULL, "744831821.ogg"},
	{0x966A6AA3F7BBB831ULL, "6028865_m.fxe"},
	{0x966A7C3AE1508A8FULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.dura.gad"},
	{0x966ABC4996B9D45FULL, "ser121_jethann6175423.cl"},
	{0x966ACB831BE0DD92ULL, "6179777_m.fxe"},
	{0x966BA15A7F66D79FULL, "gen_im_acc_amu_act2_07.uti"},
	{0x966BFF9B97C1C706ULL, "6072267_m.fxe"},
	{0x966C6769F28657CBULL, "562083627.ogg"},
	{0x966C7C0EBA0799ABULL, "113187552.ogg"},
	{0x966C7F3EE63D03D8ULL, "vxw_woundedcoast_1fb_0_0.vlm"},
	{0x966C87B9BFDF9977ULL, "6018796_m.fxe"},
	{0x966C9594AEB7E9E8ULL, "hlac_w_bonepit_262148_0y.dds"},
	{0x966CEBECC719ED06ULL, "gen00fl_sebastian6188590.cl"},
	{0x966D365C5D71637EULL, "pro000cs_opening_cs_end.ncs"},
	{0x966D5C642022D977ULL, "284187418.ogg"},
	{0x966D9AB2BECBEE3CULL, "art\\characters\\playercharacter\\humanmale\\eye_hlm.mao"},
	{0x966E25A20AAF7635ULL, "vxlt_0_lowtown_a3_n_c_hoy_0_0.vlm"},
	{0x966E3BAD251634C9ULL, "6026200_f.fxe"},
	{0x966E8440C11D2625ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_imp_p.impa.evt"},
	{0x966F237826F8889FULL, "ave121_jeven.bnk"},
	{0x966F6A30B5A6E9BAULL, "vxlt_darktown_2xb_0_0.vlm"},
	{0x966F9EA9EF6E330CULL, "w_blightlands_65_0.msh"},
	{0x966FA4A9EF6E3DFEULL, "w_blightlands_65_0.mmh"},
	{0x966FB0A9EF6E5195ULL, "w_blightlands_65_0.mao"},
	{0x96701EE25CD41E25ULL, "6070496_f.fxe"},
	{0x96707ED0D50CB738ULL, "481448391.ogg"},
	{0x9670913AE1551A3CULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.dura.evt"},
	{0x9670AC373D184406ULL, "292485851.ogg"},
	{0x96716C94EB3361C9ULL, "157567512.ogg"},
	{0x9671994C20116B3EULL, "t1_mue_ro2.tnt"},
	{0x967217A8CCB5997BULL, "gen_im_arm_glv_lgt_act1_01.uti"},
	{0x9672CFB0362C402EULL, "llac_lt_gallowstemplar_f_2391_2.dds"},
	{0x9673057826FC3EA7ULL, "ave121_jeven.cnv"},
	{0x967323BAE3BC9E05ULL, "hlac_w_encounter2_-65535_0z.dds"},
	{0x96733D66887C0DDFULL, "69784353.ogg"},
	{0x967368702F1106CCULL, "mcr310cr_templar_seeker.utc"},
	{0x967371A7306EBB87ULL, "6156098_m.fxe"},
	{0x96741AAD882292C7ULL, "358476183.ogg"},
	{0x96742B4172CA47A8ULL, "w_sundermount_104_0.msh"},
	{0x96743D4172CA6639ULL, "w_sundermount_104_0.mao"},
	{0x9674494172CA7AD2ULL, "w_sundermount_104_0.mmh"},
	{0x9674B4247D8739BDULL, "68715200.ogg"},
	{0x9674D119F9DBE9CAULL, "6121167_m.fxe"},
	{0x9675174B8FA81206ULL, "vxca_caverns1_cl8_0_0.vlm"},
	{0x96752CBA402F09FFULL, "ser270_emeric6149374.cl"},
	{0x967577AE9290A873ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxa.phy"},
	{0x96757A835D9BAB79ULL, "vxht_mansion2_3ex_0_0.vlm"},
	{0x9675F2E8FB471503ULL, "311911249.ogg"},
	{0x9677619EB9FC1084ULL, "6001348_m.fxe"},
	{0x9677B560F91E2644ULL, "6116606_m.fxe"},
	{0x96781EACF6E19B2CULL, "vxlt_hangedman_n_10a_0_0.vlm"},
	{0x967831B5ADDE5D53ULL, "vxht_mansion2b_1tc_0_0.vlm"},
	{0x967866CB6C373501ULL, "vxlt_blackemporium_dn_0_0.vlm"},
	{0x9678C5C293091496ULL, "llac_dr_primevalentrance_1726_6.dds"},
	{0x9679163DBA43660AULL, "6109381_m.fxe"},
	{0x967936B713F8710AULL, "llac_lt_0_lowtown_n_-65536_6.dds"},
	{0x96793E374076D615ULL, "follower_barks6153231.cl"},
	{0x9679421DB7919FE3ULL, "art\\characters\\playercharacter\\shared\\textures\\bdy_bnb_0n.dds"},
	{0x9679DF365954A3BBULL, "vxlt_backalley_g_1vp_0_0.vlm"},
	{0x967A3735E29CB208ULL, "180242311.ogg"},
	{0x967AA1191B1EBE24ULL, "475235338.ogg"},
	{0x967AB22A129DFDAFULL, "88386114.ogg"},
	{0x967AF11BD8331C58ULL, "vxw_woundedcoast_1ac_0_0.vlm"},
	{0x967AF6F2F6983309ULL, "zz_dae_debug6029285.cl"},
	{0x967B04B4F6FEA82DULL, "hlac_w_sundermount_131069_0z.dds"},
	{0x967B37B0BC7F3551ULL, "vxlt_gallowstemplar_f_3ra_0_0.vlm"},
	{0x967B5903445663EDULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrp.mmh"},
	{0x967B5D0344566A3EULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrp.mao"},
	{0x967B6B0344568273ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivygrp.msh"},
	{0x967B77EB6C3B57B4ULL, "qatest.mao"},
	{0x967BC5DCA5D68323ULL, "mrl250cr_thief_e.utc"},
	{0x967BEED7862C8035ULL, "501755136.ogg"},
	{0x967DD200B2A66F74ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_whippingpost_n.dds"},
	{0x967DD900B00EA046ULL, "hlac_lt_0_lowtown_docks_196607_1z.dds"},
	{0x967E09083324CDADULL, "vxht_brothel_n_6v0_0_0.vlm"},
	{0x967E7401F4208F13ULL, "hf_mother_h1s1.mor"},
	{0x967E993AE1616555ULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.dura.ani"},
	{0x967F6F3E59704F6DULL, "ht_0_hightown_n_57_0.rml"},
	{0x967F8772017787DCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crnf.ani"},
	{0x967FB1F678280E6EULL, "hlac_lt_0_lowtown_a2_65536_1y.dds"},
	{0x96803B334DB5F359ULL, "294036126.ogg"},
	{0x96807101C0B888B8ULL, "dae000cr_grand_cleric6148118.cl"},
	{0x9680BAE645A3413EULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeloop.ani"},
	{0x9680CF01FD9E8A93ULL, "935210562.ogg"},
	{0x96810C6B59003D22ULL, "cs_isa111cs_isa_intro.bnk"},
	{0x96810CEE0F1B1F4CULL, "6135717_m.fxe"},
	{0x968146CA601579ECULL, "w_sundermount_232_0.msh"},
	{0x96814CCA601583DEULL, "w_sundermount_232_0.mmh"},
	{0x968152529A9F7D5CULL, "vxlt_blackemporium_w2_0_0.vlm"},
	{0x9681528FDB2966CDULL, "6098048_m.fxe"},
	{0x968158CA60159875ULL, "w_sundermount_232_0.mao"},
	{0x9681AD8A5D451032ULL, "385585228.ogg"},
	{0x9681AE3715A5EE68ULL, "art\\vfx\\textures\\potion\\vfx_potion_e.dds"},
	{0x9681B26B6F32FBD2ULL, "vxht_mansion2b_1tf_0_0.vlm"},
	{0x9681C0636015BE22ULL, "vxw_sundermount_238_0_0.vlm"},
	{0x96822C5EBB2DA0CCULL, "vxht_0_hightown_c3j_0_1.vlm"},
	{0x96837F1AD6CF3DA4ULL, "isa211_party_reaction6083429.cl"},
	{0x96838EAEB5195230ULL, "34024900.ogg"},
	{0x968398B42970FB7EULL, "dae000_cerimon6163684.cl"},
	{0x9683DEF9ECD0E9EBULL, "vxlt_gallowstemplar_f_4u9_0_0.vlm"},
	{0x9683F033088B1883ULL, "hlac_ca_caverns2_5802_72y.dds"},
	{0x96844627171B351EULL, "hf_ibethany_h1s3.mor"},
	{0x9684482360B54864ULL, "6136104_f.fxe"},
	{0x96847C1EADC86ADCULL, "6217011_m.fxe"},
	{0x9684915D529A02AEULL, "llac_w_woundedcoast_-196602_0.dds"},
	{0x9684DB7E2C37217EULL, "vxht_brothel_n_612_0_0.vlm"},
	{0x9684E78F7F710138ULL, "6116865_f.fxe"},
	{0x96855D261844B922ULL, "625616222.ogg"},
	{0x96859472017C091EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crnf.gad"},
	{0x9685CE4DAE59002EULL, "bd_glow.mao"},
	{0x96861F6E80F55CDBULL, "combat_smaoe_r1_a2.gda"},
	{0x96863E7AD27DB0DBULL, "vxlt_0_lowtown_a3_n_c_hcs_0_0.vlm"},
	{0x9686ED605B71CA41ULL, "gen00fl_sebastian6180454.cl"},
	{0x9686F72AA4702B4AULL, "hlac_lt_0_lowtown_a3_n_c_131070_3z.dds"},
	{0x96871BFC56D1041EULL, "6127495_m.fxe"},
	{0x968745587079F85CULL, "dae100_viscount6118528.cl"},
	{0x968769E127655734ULL, "283043783.ogg"},
	{0x968866A86DD0DE12ULL, "mag320_bsc_banter6100768.cl"},
	{0x9688E8385617B809ULL, "247743318.ogg"},
	{0x96894F75D0A426B7ULL, "183730511.ogg"},
	{0x96897E88DFECCCEAULL, "dae000_gamlen6121582.cl"},
	{0x968A1B2FEC4134D0ULL, "406363619.ogg"},
	{0x968A219F7BB1D545ULL, "vxht_bartrand_n_4c1_0_0.vlm"},
	{0x968A2D604D59B439ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_5.ani"},
	{0x968A5D2408475F7FULL, "6178335_m.fxe"},
	{0x968A856B96DDF86AULL, "347334813.ogg"},
	{0x968B02F22D3F0006ULL, "64476869.ogg"},
	{0x968BBE012B70CF3EULL, "ch2notnt_nolit.psh.fxo"},
	{0x968C12D782BCF3F9ULL, "6072336_m.fxe"},
	{0x968C6F720181E93DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crnf.evt"},
	{0x968C9196D31D649AULL, "hlac_ca_caverns1_mines_24424_20x.dds"},
	{0x968CCCC325534317ULL, "265053662.ogg"},
	{0x968CE20E9546FA00ULL, "6123726_f.fxe"},
	{0x968D63E0C5CFCEF8ULL, "llac_w_encounter2_65539_0.dds"},
	{0x968DBE61B75F9A2BULL, "character_character_p219.psh.fxo.dx11"},
	{0x968DD2E645ADF40FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeloop.evt"},
	{0x968DDC775281279FULL, "6174533_m.fxe"},
	{0x968DE8B99BCF952AULL, "vxht_0_hightown_n_aij_0_0.vlm"},
	{0x968E00777EC66B25ULL, "mcr370_guardsman6086985.cl"},
	{0x968E1C975A253043ULL, "lgt260_eau_tmp_angelique6143963.cl"},
	{0x968E6504D6E43660ULL, "6132828_m.fxe"},
	{0x968E8A2AB609F48CULL, "w_sundermount_48_0.rml"},
	{0x968ED20000B8ACF7ULL, "vxlt_backalley_g_1lj_0_0.vlm"},
	{0x968F4AB84A66E35CULL, "hlac_lt_gallowstemplar_2382_1z.dds"},
	{0x968F74EC77226B28ULL, "6043845_m.fxe"},
	{0x96903240520BD6A2ULL, "vxw_sundermount_249_0_0.vlm"},
	{0x969038F47A885CA9ULL, "276684150.ogg"},
	{0x9690461C17665FFBULL, "art\\characters\\playercharacter\\humanfemale\\hlh_lgta.phy"},
	{0x96905C6619889D64ULL, "123090405.ogg"},
	{0x96905EFE09AFAEA5ULL, "plt_cod_hst_bonepit.plo"},
	{0x969079A71420B6A8ULL, "6107297_m.fxe"},
	{0x969108604D5F947BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_5.gad"},
	{0x9691119434959196ULL, "1052384278.ogg"},
	{0x9691340925A58CBDULL, "llac_w_bonepitdestroyed_a3_262148_0.dds"},
	{0x96915C66885292D3ULL, "zz_followers_banter6158718.cl"},
	{0x96919089DEC2F0FEULL, "dae021_male_worshipper_sta.cnv"},
	{0x9691B0F575F812E7ULL, "364163091.ogg"},
	{0x9691E25AE73272AEULL, "206518866.ogg"},
	{0x969202578AB7C85BULL, "vxw_sundermount_371_0_1.vlm"},
	{0x96928F2821D25601ULL, "561537954.ogg"},
	{0x96929A222BF36A08ULL, "sky_lt_gallowscourtyard_n.mao"},
	{0x9692D32F7BC92CFBULL, "lt_0_lowtown_a2_n_46_0.anb"},
	{0x96932976D813B84EULL, "804132376.ogg"},
	{0x96934BD206B06015ULL, "plt_cod_bks_first_children.plo"},
	{0x9693F1D9DC837F0DULL, "vxht_mansion2_2ae_0_0.vlm"},
	{0x96941A4E66542CE0ULL, "zz_mrl_debug6086556.cl"},
	{0x96943E594CA664EAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sworda_d.dds"},
	{0x969444979D627515ULL, "6018646_m.fxe"},
	{0x96947289DEC4F30EULL, "dae021_male_worshipper_sta.bnk"},
	{0x9694D3E645B414B8ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeloop.gad"},
	{0x9694EA047E94DDD5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09a.phy"},
	{0x9695B9434FE2CA13ULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.impa.ani"},
	{0x9695F3C72246C3FEULL, "255792444.ogg"},
	{0x969649F2988BBBFAULL, "6111119_m.fxe"},
	{0x96968AE9F74DCEF2ULL, "vxlt_0_lowtown_a2_n_exe_0_0.vlm"},
	{0x9696B98AEB3C71DAULL, "hlac_lt_warehouse_n_65536_3z.dds"},
	{0x9696C2F03D40D4BBULL, "vxlt_backalley_g_1xr_0_0.vlm"},
	{0x96971FE2836FAE39ULL, "177038566.ogg"},
	{0x969725604D643160ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_5.evt"},
	{0x9697A310A825C28FULL, "gen00fl_fenris6126815.cl"},
	{0x9697CF2006D49702ULL, "mer110_jansen6032876.cl"},
	{0x9697F0FAC449C0FDULL, "dae000_bartrand6034001.cl"},
	{0x9698DC4ECFB0D7CEULL, "vxlt_0_lowtown_docks_a3_n_cwt_0_0.vlm"},
	{0x969904F9D1804486ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medc_0d.dds"},
	{0x969934FB802EA5E8ULL, "gen00fl_merrill6128746.cl"},
	{0x969938178F52B416ULL, "llac_ht_brothel_n_3381_0.dds"},
	{0x96995F5320ABE8E3ULL, "459489051.ogg"},
	{0x9699698096B97B2BULL, "vxht_0_hightown_aqp_0_0.vlm"},
	{0x9699D55AA2EBDC55ULL, "vxlt_gallowstemplar_472_0_0.vlm"},
	{0x969A09AA50978D15ULL, "6052976_m.fxe"},
	{0x969A0A59481FC35AULL, "zz_vault_debug6203190.cl"},
	{0x969A204E3D921D1BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\trap_lowtownint_0.pwk"},
	{0x969AA51862F3F0F0ULL, "vxw_encounter1_az_0_0.vlm"},
	{0x969B2726DBC4FF96ULL, "vxlt_0_lowtown_n_bxr_0_0.vlm"},
	{0x969B3B1DECA7CC85ULL, "94545523.ogg"},
	{0x969BA42863D67AC4ULL, "549574127.ogg"},
	{0x969BBB7905E80D96ULL, "20262354.ogg"},
	{0x969C5FB08F5B1B26ULL, "894349447.ogg"},
	{0x969C90669619BA42ULL, "vxdr_primevalentrance_wq_0_0.vlm"},
	{0x969CE686149CF1DDULL, "794307580.ogg"},
	{0x969D96143EEC9B8EULL, "mage_stonefist.dds"},
	{0x969D9F4FB8576C6EULL, "vxht_bartrand_n_6qb_0_0.vlm"},
	{0x969DC457ED97D2F3ULL, "6062155_m.fxe"},
	{0x969DCAE8F481C0E9ULL, "484559790.ogg"},
	{0x969DF8183D8A1404ULL, "862863353.ogg"},
	{0x969EF644A61BA4FDULL, "art\\vfx\\materials\\chunky\\vfx_c_crack_tile.mao"},
	{0x969F2144729C0B0AULL, "292560145.ogg"},
	{0x969F2DA3D64355F3ULL, "6055621_m.fxe"},
	{0x969F32E53B0431D0ULL, "6166555_m.fxe"},
	{0x969F34E615406663ULL, "759011354.ogg"},
	{0x969FB0F61D1553A6ULL, "6028358_m.fxe"},
	{0x96A074347CADE0DBULL, "vxw_sundermount_322_0_1.vlm"},
	{0x96A0CF327E6041C9ULL, "hlac_dr_dungeon_3207_6y.dds"},
	{0x96A1510B9BA92BACULL, "hm_refugee_1.mor"},
	{0x96A1A18480971C8CULL, "vxw_sundermount_2or_0_0.vlm"},
	{0x96A1BDB3A513B343ULL, "llac_ht_mansion2_818_65.dds"},
	{0x96A1CDA19A707E74ULL, "vxlt_gallowscourtyard_6kw_0_0.vlm"},
	{0x96A1E9E469C384A1ULL, "hlac_ht_0_hightown_a3_65535_1z.dds"},
	{0x96A21279846D1912ULL, "6181751_f.fxe"},
	{0x96A21D632C0F4516ULL, "vxca_caverns2_7nm_0_0.vlm"},
	{0x96A263B6CE79978DULL, "vxht_mansion2_2jd_0_0.vlm"},
	{0x96A2B869E3FE8A50ULL, "vxca_caverns1_iao_0_0.vlm"},
	{0x96A2E1F1F20B275AULL, "c_doga_0t.dds"},
	{0x96A2F35F70C0FA80ULL, "6170071_m.fxe"},
	{0x96A3245C4A822287ULL, "vxca_caverns1_mines_ij0_0_0.vlm"},
	{0x96A346D8856A8BC1ULL, "lt_0_lowtown_a2_n_c_15_0.phy"},
	{0x96A3B1434FEEFA4AULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.impa.evt"},
	{0x96A3E8984A3E622EULL, "6035367_m.fxe"},
	{0x96A4D2E60B984112ULL, "6202293_m.fxe"},
	{0x96A57AC510EF5B9DULL, "vxht_brothel_n_5h1_0_0.vlm"},
	{0x96A585632D1CFB7BULL, "hlac_lt_0_lowtown_n_65534_4z.dds"},
	{0x96A63593B5C2146AULL, "llac_lt_warehouse_65537_1.dds"},
	{0x96A68625ED5E0A5FULL, "72882866.ogg"},
	{0x96A6F66B5920E740ULL, "cs_isa111cs_isa_intro.wbd"},
	{0x96A706AD4E681318ULL, "and111_donation_box6207868.cl"},
	{0x96A7C5FBEAFEF6D0ULL, "260819266.ogg"},
	{0x96A810DA263B82DFULL, "6043688_m.fxe"},
	{0x96A84ED7F8CB088BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobd_0s.dds"},
	{0x96A872C6B9DC5F19ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mede.mmh"},
	{0x96A87CC6B9DC7007ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mede.msh"},
	{0x96A90B782729D814ULL, "ave121_jeven.stg"},
	{0x96A935F9EDF64997ULL, "art\\vfx\\combat\\vfx_gib_hm_meda.msh"},
	{0x96A94BF9EDF66F09ULL, "art\\vfx\\combat\\vfx_gib_hm_meda.mmh"},
	{0x96A95889E97C6528ULL, "framebuffereffects_smallblur_p0.vsh.fxo.dx11"},
	{0x96A9676925B6FD55ULL, "mcr310_confused_guard6077067.cl"},
	{0x96A9DD6745F7417FULL, "follower_banter6155544.cl"},
	{0x96A9F3E65B7725C5ULL, "6149038_m.fxe"},
	{0x96A9F7A3BDDDEC27ULL, "llac_ht_0_hightown_a2_65533_2.dds"},
	{0x96A9F9B5BF252C2EULL, "567626156.ogg"},
	{0x96AA02008B923402ULL, "54458338.ogg"},
	{0x96AA2AF2F415742FULL, "vxca_caverns2_3ho_0_1.vlm"},
	{0x96AA3C28D6724E32ULL, "zz_approval_debug6003641.cl"},
	{0x96AA5D7653385AB1ULL, "w_encounter1_37_0.mmh"},
	{0x96AA697653386F52ULL, "w_encounter1_37_0.mao"},
	{0x96AA77765338876FULL, "w_encounter1_37_0.msh"},
	{0x96AA90434FF4E0BDULL, "art\\vfx\\environment\\vfx_pois_gas_area_med.impa.gad"},
	{0x96AAD1909F9EE265ULL, "655350513.ogg"},
	{0x96AB341B3E716140ULL, "vxht_keep_49b_0_0.vlm"},
	{0x96AB5D14F31F7988ULL, "art\\vfx\\combat\\vfx_sg_app_c.anb"},
	{0x96ABC7BC88AB4FB6ULL, "hlac_w_encounter3_-327676_0y.dds"},
	{0x96ABD212BC9BE7CAULL, "gen00fl_fenris6126746.cl"},
	{0x96ABD653A0E3F923ULL, "vxlt_backalley_n_1wn_0_0.vlm"},
	{0x96ABD95D0930E473ULL, "vxdr_primevaltomb_828_0_1.vlm"},
	{0x96AC3634DF8DEB73ULL, "6069055_m.fxe"},
	{0x96AC44BEAA821FC3ULL, "172530687.ogg"},
	{0x96AC5C9DAF21FBE7ULL, "546169617.ogg"},
	{0x96AC91882D8D7B11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsa.mmh"},
	{0x96AC9D882D8D8FB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsa.mao"},
	{0x96ACAB882D8DA7CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_floorbitsa.msh"},
	{0x96ACF88FB39F6C0EULL, "73380490.ogg"},
	{0x96AD24A94BEF4AA7ULL, "661651176.ogg"},
	{0x96AD4FFCF0AB716CULL, "vxdr_primevaltomb_34o_0_0.vlm"},
	{0x96ADC3E474CFD40CULL, "hlac_lt_0_lowtown_-1_5x.dds"},
	{0x96AE1001BAF33D8FULL, "6141601_m.fxe"},
	{0x96AE9EA8506E6E0CULL, "static_lightblend_p0.vsh.fxo.dx11"},
	{0x96AECAA602A2E5A0ULL, "hlac_w_woundedcoast_-196604_0z.dds"},
	{0x96AEDEB2114682C4ULL, "mag170cr_recruit02.utc"},
	{0x96AEF7EF4296167EULL, "422157935.ogg"},
	{0x96AF49B13630190BULL, "qun221_viscount6070166.cl"},
	{0x96AF71C3DF2CB858ULL, "hlac_lt_darktown_-1_80x.dds"},
	{0x96AFD2D43DADFA3AULL, "6113631_f.fxe"},
	{0x96B00ADE3F7897A4ULL, "vxlt_0_lowtown_docks_a3_czf_0_0.vlm"},
	{0x96B01361728D350CULL, "vxw_sundermount_2ps_0_0.vlm"},
	{0x96B05061D75D563EULL, "llac_ht_0_hightown_n_-1_2.dds"},
	{0x96B0ACABA3063C5CULL, "vxht_mansion2b_28g_0_0.vlm"},
	{0x96B0F1C8F55B7982ULL, "vxht_mansion2b_3cm_0_0.vlm"},
	{0x96B11FE8F9E4E833ULL, "29194347.ogg"},
	{0x96B1579D0923ADFCULL, "gen00fl_aveline6106387.cl"},
	{0x96B1E2DAF2586346ULL, "496548480.ogg"},
	{0x96B209A63288BD74ULL, "6033572_m.fxe"},
	{0x96B2111C84FFB456ULL, "vxw_sundermount_300_0_0.vlm"},
	{0x96B22162B3B90AF8ULL, "6125833_m.fxe"},
	{0x96B226F8D961321CULL, "6098751_m.fxe"},
	{0x96B2302318C04415ULL, "vxca_caverns1_mines_e8x_0_0.vlm"},
	{0x96B246BA857D3C36ULL, "237009085.ogg"},
	{0x96B2863515B44C56ULL, "art\\vfx\\creatures\\fxc_ashwraith_c.cess.evt"},
	{0x96B29FD66A1071FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_sacrifice.ani"},
	{0x96B3A3B333EB7A6DULL, "121087994.ogg"},
	{0x96B48A124A2088D1ULL, "vxca_caverns2_96x_0_0.vlm"},
	{0x96B4DA543BFCC7A6ULL, "6033404_m.fxe"},
	{0x96B4ECAFB4005CF9ULL, "754524282.ogg"},
	{0x96B51470DDE7749AULL, "vxw_woundedcoast_11a_0_0.vlm"},
	{0x96B534D2C751315DULL, "zz_quanaricharge.cut"},
	{0x96B5817E16723D16ULL, "hlac_lt_0_lowtown_a2_n_c_131072_0x.dds"},
	{0x96B6231CDFD51A6BULL, "6205994_m.fxe"},
	{0x96B69DCBB8C1783CULL, "500698424.ogg"},
	{0x96B6FA129C5E35D1ULL, "qun110_petrice6042211.cl"},
	{0x96B7187119369709ULL, "ch_lit.mao"},
	{0x96B7BEBB6B64BA6AULL, "art\\vfx\\combat\\vfx_gib_arm_lgta.mao"},
	{0x96B7CD141B22ED49ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_loop.ani"},
	{0x96B84C006B25AC61ULL, "vxdr_0_deeproads_etb_0_0.vlm"},
	{0x96B8DFBED4911136ULL, "hlac_ht_mansion2_818_4z.dds"},
	{0x96B9564F43F4CE18ULL, "251659202.ogg"},
	{0x96B97D3515BA5BE1ULL, "art\\vfx\\creatures\\fxc_ashwraith_c.cess.gad"},
	{0x96BA0D0F223BA5F1ULL, "vxlt_0_lowtown_docks_a3_n_aoi_0_0.vlm"},
	{0x96BA82CFB1FBE627ULL, "196432907.ogg"},
	{0x96BB36AFE9714A10ULL, "qun120_arishok6051469.cl"},
	{0x96BB5C2BDA7CED5BULL, "6053537_m.fxe"},
	{0x96BB66C7190C3790ULL, "6098957_m.fxe"},
	{0x96BBBF03BC766FFAULL, "8536310.ogg"},
	{0x96BC2D1C178B6F9AULL, "art\\characters\\playercharacter\\humanfemale\\hlh_lgta.msh"},
	{0x96BC3B1C178B87D3ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_lgta.mao"},
	{0x96BC3F1C178B8EE0ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_lgta.mmh"},
	{0x96BD5D86002C1F1CULL, "vxw_sundermount_38b_0_0.vlm"},
	{0x96BD95C222C55E68ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulley.mmh"},
	{0x96BDA1C222C5720BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulley.mao"},
	{0x96BDB3C222C59092ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_pulley.msh"},
	{0x96BDE8141B2786CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_loop.gad"},
	{0x96BE337667281F38ULL, "vxw_encounter2_ku_0_0.vlm"},
	{0x96BE77DB4F5B6C8DULL, "vxlt_0_lowtown_docks_b5u_0_1.vlm"},
	{0x96BED59CC683A541ULL, "vxw_bonepit_25z_0_0.vlm"},
	{0x96BF1E8894FC54DCULL, "vxht_mansion2b_25d_0_0.vlm"},
	{0x96BF4770B265C88DULL, "vxht_mansion2_2pb_0_0.vlm"},
	{0x96BF63A5E7519202ULL, "vxht_mansion2b_3fn_0_0.vlm"},
	{0x96BF7E3515BEC88FULL, "art\\vfx\\creatures\\fxc_ashwraith_c.cess.ani"},
	{0x96C07D2CB43D872DULL, "prp_str_mirrorstatue_01s.dds"},
	{0x96C0A7D66A1CBCF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_sacrifice.evt"},
	{0x96C0DDDE6D4A1CFEULL, "vxht_0_hightown_a2_aie_0_0.vlm"},
	{0x96C0F36685D7E70AULL, "gen_im_trs_drk_one_lsw_r1.uti"},
	{0x96C0FB599DB1B00AULL, "llac_lt_gallowstemplar_n_2384_0.dds"},
	{0x96C11116D0FBD672ULL, "491287595.ogg"},
	{0x96C16971D0812BA4ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_act_c.mmh"},
	{0x96C1D02AE71D855CULL, "zz_ser_debug6078147.cl"},
	{0x96C1FCF7B17692EBULL, "140135218.ogg"},
	{0x96C1FDA9BECE97F3ULL, "362046745.ogg"},
	{0x96C25A4A487A697EULL, "236944138.ogg"},
	{0x96C2CB5A89B45462ULL, "6054740_m.fxe"},
	{0x96C35275D4B7A229ULL, "173079033.ogg"},
	{0x96C392A25DB3292AULL, "6135885_m.fxe"},
	{0x96C3A542C94CD9D2ULL, "gen00fl_dog6107717.cl"},
	{0x96C3A79524F5A413ULL, "hlac_lt_gallowstemplar_f_2383_12y.dds"},
	{0x96C422FFB8CC84E1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopening.gad"},
	{0x96C4EBE0F2ACE4BAULL, "6172404_m.fxe"},
	{0x96C4FE72720BF4FAULL, "6174073_m.fxe"},
	{0x96C571486A9A3462ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightb.msh"},
	{0x96C573486A9A3718ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightb.mmh"},
	{0x96C57F486A9A4BBBULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightb.mao"},
	{0x96C5801375778666ULL, "vxlt_0_lowtown_docks_a3_n_bvb_0_0.vlm"},
	{0x96C596D3B463931BULL, "822563887.ogg"},
	{0x96C5E618938881FEULL, "460634579.ogg"},
	{0x96C64FE4EA768DE1ULL, "316860492.ogg"},
	{0x96C6BDDD5D1BC4E1ULL, "vxdr_0_deeproads_eqa_0_0.vlm"},
	{0x96C6C6D66A215C85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_sacrifice.gad"},
	{0x96C7B57920AEC536ULL, "6191061_m.fxe"},
	{0x96C7B79BA4F2A14DULL, "vxht_mansion_n_6o4_0_0.vlm"},
	{0x96C7C42E71EC887CULL, "6104339_m.fxe"},
	{0x96C88417DD5E4422ULL, "vxw_encounter2_5m_0_0.vlm"},
	{0x96C89973C8910EF7ULL, "vxlt_backalley_g_1xf_0_0.vlm"},
	{0x96C8C47FBD1E3488ULL, "6079674_f.fxe"},
	{0x96C8FE00AD4A5E42ULL, "vxlt_gallowscourtyard_n_63r_0_0.vlm"},
	{0x96C9AB86E4B98C7CULL, "6119650_m.fxe"},
	{0x96C9C645CC28C800ULL, "884527736.ogg"},
	{0x96CA23FFB8D0F18FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopening.ani"},
	{0x96CA2DBA67EE32C4ULL, "6066909_m.fxe"},
	{0x96CA61BEDCB06819ULL, "748599797.ogg"},
	{0x96CA6466DC63EA2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_20_walk.ani"},
	{0x96CA9EF4A35A01A6ULL, "457266332.ogg"},
	{0x96CB3A88FAF81ACBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen.evt"},
	{0x96CB4F48E2DAB63DULL, "weapon_vfx_p0.psh.fxo.dx11"},
	{0x96CBA2A0791A9E1CULL, "355296053.ogg"},
	{0x96CBA882B31204BAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_baserocks_1.mao"},
	{0x96CC8D694FF9222CULL, "w_bonepit_75_0.msh"},
	{0x96CC93694FF92C1EULL, "w_bonepit_75_0.mmh"},
	{0x96CC9F694FF940B5ULL, "w_bonepit_75_0.mao"},
	{0x96CD21C7A5D33104ULL, "llac_ht_0_hightown_-131069_0.dds"},
	{0x96CD2874960F1696ULL, "lt_0_lowtown_docks_a3_n_158_0.msh"},
	{0x96CD3274960F2794ULL, "lt_0_lowtown_docks_a3_n_158_0.mmh"},
	{0x96CD3674960F2EA7ULL, "lt_0_lowtown_docks_a3_n_158_0.mao"},
	{0x96CD4779B879BDC1ULL, "vxw_bonepit_20y_0_0.vlm"},
	{0x96CD4CFF8525DA08ULL, "lowtown_fem_prostitute_sta6161598.cl"},
	{0x96CE040FA20D735FULL, "vxlt_darktown_2w6_0_0.vlm"},
	{0x96CE8D71BEEA9776ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v3.ani"},
	{0x96CEED601674EE74ULL, "6039772_m.fxe"},
	{0x96CF0B394312222DULL, "gen00fl_varric6104545.cl"},
	{0x96CF3CE77F1E04CEULL, "6060426_m.fxe"},
	{0x96CF794DD57E2258ULL, "827745249.ogg"},
	{0x96CF7A486AA297A3ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highlightb.phy"},
	{0x96CFFA8C7205C495ULL, "boulder_mac_01.mao"},
	{0x96D01CA69E7B5B10ULL, "785553298.ogg"},
	{0x96D048417318D7CDULL, "w_sundermount_104_0.phy"},
	{0x96D08F8E87905764ULL, "981675121.ogg"},
	{0x96D17166DC6A1F6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_20_walk.gad"},
	{0x96D1B234F3080219ULL, "807909921.ogg"},
	{0x96D1BE0E1FE8F8A6ULL, "6039604_m.fxe"},
	{0x96D25388FAFE638CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen.gad"},
	{0x96D26B3B666005CBULL, "473257704.ogg"},
	{0x96D32DCCD8EF3C9DULL, "llac_lt_0_lowtown_docks_n_-196610_0.dds"},
	{0x96D35C14F340CA23ULL, "art\\vfx\\combat\\vfx_sg_app_c.mmh"},
	{0x96D36E2D86FD8231ULL, "769235883.ogg"},
	{0x96D38A958E47CCE7ULL, "6164284_m.fxe"},
	{0x96D3ED373FC1E599ULL, "6126820_m.fxe"},
	{0x96D40C022B2FE1E3ULL, "w_bonepitmining_a2_121_0.mao"},
	{0x96D410022B2FE830ULL, "w_bonepitmining_a2_121_0.mmh"},
	{0x96D41E022B30004AULL, "w_bonepitmining_a2_121_0.msh"},
	{0x96D43E531479B573ULL, "vxlt_gallowstemplar_n_6qc_0_0.vlm"},
	{0x96D44E7F45BDC0BDULL, "hlac_ht_bartrand_n_3369_0y.dds"},
	{0x96D46834CB0AE3DAULL, "6090234_m.fxe"},
	{0x96D529ADB6C8664AULL, "163181998.ogg"},
	{0x96D55D12E5157C84ULL, "cutscene_wrapper.cnv"},
	{0x96D5E8B1D54FB2E5ULL, "6164189_m.fxe"},
	{0x96D6220A559E0FAAULL, "vxht_0_hightown_n_avg_0_0.vlm"},
	{0x96D64941731D4579ULL, "w_sundermount_104_0.rml"},
	{0x96D6E7838D451FE7ULL, "859170272.ogg"},
	{0x96D72BEA76C642A3ULL, "vxlt_backalley_n_1xs_0_0.vlm"},
	{0x96D770AE92E366E8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxa.mmh"},
	{0x96D77CAE92E37A8BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxa.mao"},
	{0x96D78E761316B048ULL, "llac_lt_0_lowtown_a2_n_c_131071_7.dds"},
	{0x96D78EAE92E39912ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3co_3wayxa.msh"},
	{0x96D7A93853E8FB1CULL, "6095316_m.fxe"},
	{0x96D7BCCD1722B7D6ULL, "dae_warrior_aoe_instant.ncs"},
	{0x96D7E2BC83EB6C75ULL, "follower_banter6155601.cl"},
	{0x96D8271F5CC89BA8ULL, "pro000_colonel_ewald6040563.cl"},
	{0x96D85288FB02CD8AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.closedtoopen.ani"},
	{0x96D85BA4FAA66542ULL, "mag360_thrask6043904.cl"},
	{0x96D86C66DC70354DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_20_walk.evt"},
	{0x96D8F53EF3217729ULL, "hightown_fem_elfservant_mob6169093.cl"},
	{0x96D905D9AA9E3118ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_chantdoor_n.dds"},
	{0x96D9556F7F04029AULL, "111909009.ogg"},
	{0x96D9A59B7F3C6733ULL, "follower_banter6161055.cl"},
	{0x96D9B77E99288BE7ULL, "dae300_cullen6140630.cl"},
	{0x96D9D707108E9412ULL, "170842325.ogg"},
	{0x96DA0A9DD3452A3EULL, "6147659_m.fxe"},
	{0x96DA5D69382A6C50ULL, "ss_roguemob.cut"},
	{0x96DA66F58BAAA1CBULL, "ser120_ghyslain6108533.cl"},
	{0x96DAC860FDECDE9BULL, "hlac_ht_playershouse_n_855_5y.dds"},
	{0x96DB08076A827694ULL, "vxht_keep_n_61w_0_0.vlm"},
	{0x96DB17304B145038ULL, "vxw_encounter2_qs_0_0.vlm"},
	{0x96DB3315D8655F13ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_basic.phy"},
	{0x96DB6EC980135476ULL, "895003841.ogg"},
	{0x96DB755851BE8B64ULL, "358226367.ogg"},
	{0x96DB9571BEF52F67ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v3.evt"},
	{0x96DB9985CE445B5FULL, "6178883_m.fxe"},
	{0x96DBFD047ED100F0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09a.msh"},
	{0x96DC0B047ED1180AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye09a.mmh"},
	{0x96DC30B0603E9A96ULL, "scaleform_sf_main_p2.psh.fxo.dx11.assembly"},
	{0x96DC4EFA930E9358ULL, "one101_orders_followers6115784.cl"},
	{0x96DC6800D50E63C4ULL, "hlac_lt_0_lowtown_2_0y.dds"},
	{0x96DD4BC7142AE163ULL, "6087060_m.fxe"},
	{0x96DD4C9973B12A21ULL, "gen00fl_merrill6116010.cl"},
	{0x96DDC7CAF80832C6ULL, "mer340cr_hubert_guard.utc"},
	{0x96DDDF2413B86DF3ULL, "vxht_0_hightown_a3_bu6_0_0.vlm"},
	{0x96DED16BED2A4C33ULL, "follower_barks6153310.cl"},
	{0x96DF16C86CC9AF02ULL, "6080276_f.fxe"},
	{0x96DFD65373A705D1ULL, "vxht_0_hightown_a4q_0_0.vlm"},
	{0x96E02DCA36399DC5ULL, "57865172.ogg"},
	{0x96E06AC84DFED398ULL, "hlac_lt_0_lowtown_docks_n_-3_2y.dds"},
	{0x96E098B8EDBE990DULL, "710834590.ogg"},
	{0x96E0C4CB71E600C8ULL, "hm_ewald.mor"},
	{0x96E1708833D1D2C9ULL, "mrl150_merrill6071916.cl"},
	{0x96E184E7D3965927ULL, "vxlt_warehouse_n_x9_0_0.vlm"},
	{0x96E20362BEB20E09ULL, "hlac_ht_keep_n_3741_0x.dds"},
	{0x96E21F3684D44513ULL, "266692923.ogg"},
	{0x96E262B0A58B8FF4ULL, "6072399_m.fxe"},
	{0x96E2680611E069CBULL, "218371335.ogg"},
	{0x96E2B671BEFB85B0ULL, "art\\characters\\creatures\\mubari\\animation\\atk_v3.gad"},
	{0x96E2E061E65F15B0ULL, "character_character_p210.psh.fxo.dx11"},
	{0x96E30FECD9F6D64AULL, "6106642_f.fxe"},
	{0x96E336A196350036ULL, "hlac_ht_0_hightown_n_-65537_4y.dds"},
	{0x96E3A1974107F5E1ULL, "vxdr_0_deeproads_egg_0_0.vlm"},
	{0x96E3CAC4B287F52BULL, "1046914060.ogg"},
	{0x96E45297D4D81A09ULL, "lt_0_lowtown_docks_a3_10_0.rml"},
	{0x96E478372B998187ULL, "6104156_m.fxe"},
	{0x96E5623F294AD1ECULL, "6093909_m.fxe"},
	{0x96E5BB2DEB8EA9A2ULL, "vxw_blightlands_gd_0_0.vlm"},
	{0x96E60EED4B815D36ULL, "qun221_petrice6098713.cl"},
	{0x96E6329C34C9B027ULL, "vxlt_darktown_4t1_0_0.vlm"},
	{0x96E68C988BE72A5EULL, "gen00fl_fenris6103336.cl"},
	{0x96E68F40C8A5BD7CULL, "6113050_m.fxe"},
	{0x96E740EE35DDFBF5ULL, "6118296_m.fxe"},
	{0x96E76DF732AEAD67ULL, "6103360_m.fxe"},
	{0x96E85055BBB3E0CFULL, "vxht_0_hightown_ca2_0_0.vlm"},
	{0x96E8682EB526D684ULL, "6055082_m.fxe"},
	{0x96E870613980EF3BULL, "zz_followers_banter6159573.cl"},
	{0x96E872DB39D64B9BULL, "vxht_keep_67u_0_0.vlm"},
	{0x96E8E1FEAF76F22EULL, "vxht_0_hightown_b4a_0_0.vlm"},
	{0x96E91A786E825597ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg105a_0n.dds"},
	{0x96E95105AEDD9498ULL, "6177742_m.fxe"},
	{0x96E9630239F596B5ULL, "vxht_0_hightown_a3_9bb_0_0.vlm"},
	{0x96E982D1B2A5079FULL, "ser221_gascard6111764.cl"},
	{0x96EA6C71D0A4359BULL, "art\\vfx\\combat\\vfx_r_ibl_fli_act_c.anb"},
	{0x96EA758958273617ULL, "mer221_sabin6047672.cl"},
	{0x96EAF0A9F5FA68F9ULL, "vxw_encounter3_bx_0_0.vlm"},
	{0x96EB21516BAA32B3ULL, "llac_lt_0_lowtown_a2_n_131072_5.dds"},
	{0x96EBBBCCFC210224ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypoucha_n.dds"},
	{0x96EBF5DBDA80188BULL, "148453929.ogg"},
	{0x96EC0999EA943782ULL, "lt_0_lowtown_docks_a3_n_129_0.msh"},
	{0x96EC0B99EA943B38ULL, "lt_0_lowtown_docks_a3_n_129_0.mmh"},
	{0x96EC1799EA944FDBULL, "lt_0_lowtown_docks_a3_n_129_0.mao"},
	{0x96EC265B4F92B836ULL, "915733693.ogg"},
	{0x96EC2BCED5FB8FEFULL, "vxlt_0_lowtown_n_fe0_0_0.vlm"},
	{0x96EC510105AE8673ULL, "vxht_0_hightown_a3_bp9_0_0.vlm"},
	{0x96ECD2CE80AB9FBEULL, "99647478.ogg"},
	{0x96ED071E36B28C8DULL, "em_paivel.mor"},
	{0x96ED0C064C946F06ULL, "zz_mer_debug6051476.cl"},
	{0x96ED2A4D025CA1C8ULL, "vxdr_0_deeproads_egb_0_0.vlm"},
	{0x96ED76D40D22149DULL, "zz_and_debug6178579.cl"},
	{0x96EE078B866C54B9ULL, "llac_lt_0_lowtown_a2_n_c_131070_3.dds"},
	{0x96EE25BA08264CDEULL, "142862489.ogg"},
	{0x96EE289B31F34A1AULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowsmall.msh"},
	{0x96EE369B31F36253ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowsmall.mao"},
	{0x96EE3A9B31F36960ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_gallowsmall.mmh"},
	{0x96EE83972CBE2E45ULL, "6077137_m.fxe"},
	{0x96EEC54B27958873ULL, "vxlt_0_lowtown_a2_n_hd1_0_0.vlm"},
	{0x96EF662650AAA700ULL, "gen00fl_anders6096522.cl"},
	{0x96EF6A060B978622ULL, "325355302.ogg"},
	{0x96EF94853085308BULL, "6098059_m.fxe"},
	{0x96EFFC89FE2D772CULL, "mag370_meredith6044239.cl"},
	{0x96F06ED82AFFBB95ULL, "gen00fl_merrill6107074.cl"},
	{0x96F0A984681F4123ULL, "vxca_caverns1_dq1_0_0.vlm"},
	{0x96F0D5AA4B59CFE6ULL, "vxlt_0_lowtown_docks_a3_n_bge_0_0.vlm"},
	{0x96F0ED4E43C6250EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock3a.phy"},
	{0x96F12DE1CBC6FC08ULL, "seb321_leliana6191271.cl"},
	{0x96F160A38794D13EULL, "dae300_vashothlead6062821.cl"},
	{0x96F177A6B2BFADC0ULL, "vxlt_hangedman_26w_0_0.vlm"},
	{0x96F195F8BAAFE38AULL, "6051620_m.fxe"},
	{0x96F20D03F986E0F9ULL, "vxht_0_hightown_n_brs_0_0.vlm"},
	{0x96F2AE17C067517DULL, "vxlt_0_lowtown_a3_n_c_f6p_0_0.vlm"},
	{0x96F2E530E238B142ULL, "hlac_dr_0_deeproads_17038_6y.dds"},
	{0x96F3E5441CB5B9AEULL, "w_bonepitmining_a2_64_0.mmh"},
	{0x96F3F1441CB5CE45ULL, "w_bonepitmining_a2_64_0.mao"},
	{0x96F3FF441CB5E59CULL, "w_bonepitmining_a2_64_0.msh"},
	{0x96F42D0ADD84C222ULL, "vxw_blightlands_he_0_0.vlm"},
	{0x96F4A47926BFC8A7ULL, "vxlt_darktown_4s0_0_0.vlm"},
	{0x96F51BF420B6D1F2ULL, "6103556_m.fxe"},
	{0x96F51DFEED121B81ULL, "896450258.ogg"},
	{0x96F550BF753C00DFULL, "532976003.ogg"},
	{0x96F587486F416D3BULL, "441007847.ogg"},
	{0x96F59404078F1319ULL, "llac_ht_0_hightown_n_-131074_0.dds"},
	{0x96F5CD605B717CDBULL, "6125269_m.fxe"},
	{0x96F5E5882235F843ULL, "llac_lt_0_lowtown_docks_n_65535_2.dds"},
	{0x96F5F826C2CFD7D9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_ascend_female.ani"},
	{0x96F6019C995F9B89ULL, "79788972.ogg"},
	{0x96F60C6C8F389CB6ULL, "907835040.ogg"},
	{0x96F62BD2CB2D3027ULL, "hlac_lt_0_lowtown_a2_131072_0x.dds"},
	{0x96F6B35A7204B032ULL, "vxca_caverns2prc_abf_0_0.vlm"},
	{0x96F753DBA16D0AAEULL, "vxht_0_hightown_b1b_0_0.vlm"},
	{0x96F77B47DF4287F9ULL, "vxht_mansion2_30g_0_0.vlm"},
	{0x96F7D2F3A966BB26ULL, "6957977.ogg"},
	{0x96F7F7D10DAA1E96ULL, "w_bonepitmining_a2_74_0.mao"},
	{0x96F7FBD22390F90CULL, "6135142_m.fxe"},
	{0x96F803D10DAA3335ULL, "w_bonepitmining_a2_74_0.mmh"},
	{0x96F805D10DAA36ABULL, "w_bonepitmining_a2_74_0.msh"},
	{0x96F8B91D65CF8C43ULL, "hlac_ca_caverns1_mines_24426_5x.dds"},
	{0x96F8DB6CBCCD74F3ULL, "dae000_arishok6067699.cl"},
	{0x96F8F187FFFAE7A0ULL, "hlac_lt_blackemporium_1055_6z.dds"},
	{0x96F902AF37C01BDEULL, "vxdr_primevaltomb_847_0_1.vlm"},
	{0x96F910B2F7BD8107ULL, "llac_w_woundedcoast_8_0.dds"},
	{0x96F944A892F66FE9ULL, "w_sundermount_67_0.phy"},
	{0x96F9BDC293E04632ULL, "89895609.ogg"},
	{0x96F9D7D6629164DBULL, "vxht_brothel_n_6k1_0_0.vlm"},
	{0x96FA9FF70063F82EULL, "521681539.ogg"},
	{0x96FB081C70FDC99CULL, "one140cr_charade6136238.cl"},
	{0x96FB34F765F0D9AEULL, "ss_danzig.bnk"},
	{0x96FB3BF8E2C3BF29ULL, "plt_cod_itm_karastenbelt.plo"},
	{0x96FB984D986F0FD6ULL, "vxlt_0_lowtown_docks_a3_d2o_0_0.vlm"},
	{0x96FBADC13495A877ULL, "llac_ht_mansion2b_818_23.dds"},
	{0x96FBE90C31C9103DULL, "vxca_caverns1_mines_ioi_0_0.vlm"},
	{0x96FC3F31DA0265B0ULL, "hlac_lt_0_lowtown_-131073_1x.dds"},
	{0x96FC62F276E0F5E1ULL, "vfx_splinters.dds"},
	{0x96FC667E1006FA68ULL, "qun211cr_elf_warrior6072339.cl"},
	{0x96FCD326C2D5B71BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_ascend_female.gad"},
	{0x96FCDAD866DE846AULL, "follower_barks6153164.cl"},
	{0x96FE1815C940191EULL, "face_face_p1.psh.fxo.dx11.assembly"},
	{0x96FEA70FBCB5EF75ULL, "gen00fl_carver6108895.cl"},
	{0x96FEC534B2A88945ULL, "446462414.ogg"},
	{0x96FF380D3ABDA5C9ULL, "hlac_dr_slavers_3214_7x.dds"},
	{0x97004DA892FC9DE5ULL, "w_sundermount_67_0.rml"},
	{0x9700B5658E8B8F15ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_1z.dds"},
	{0x9700ED8B7773D902ULL, "569280275.ogg"},
	{0x9701964B804F4BB2ULL, "22669227.ogg"},
	{0x9701B090BE3467DAULL, "6116103_m.fxe"},
	{0x97025E0E09609B3AULL, "mcr370_cut_epilogue6127426.cl"},
	{0x97028821BCE034C8ULL, "6144559_m.fxe"},
	{0x97029EE7CF7ADAA2ULL, "vxw_blightlands_mf_0_0.vlm"},
	{0x9702CE00930FFE96ULL, "gen00fl_anders6125583.cl"},
	{0x9702EAB1929990E4ULL, "577201796.ogg"},
	{0x9702F1EDFDD191CCULL, "ht_mansion_n_3367_lc.mmh"},
	{0x970352A20D9DFB63ULL, "514787841.ogg"},
	{0x9703F026C2DC0700ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_ascend_female.evt"},
	{0x97041E7F2C7A6EA2ULL, "508649712.ogg"},
	{0x9704219939811389ULL, "lt_0_lowtown_docks_n_146_0.mao"},
	{0x97042D9939812822ULL, "lt_0_lowtown_docks_n_146_0.mmh"},
	{0x97042F9939812BD8ULL, "lt_0_lowtown_docks_n_146_0.msh"},
	{0x970441F9A9393D74ULL, "215814562.ogg"},
	{0x970469B16E688805ULL, "vxdr_primevaltomb_6x3_0_0.vlm"},
	{0x97049D15B040F612ULL, "6094564_m.fxe"},
	{0x970517574ED046EFULL, "6144310_m.fxe"},
	{0x97054BD885BD5EFDULL, "lt_0_lowtown_a2_n_c_15_0.mao"},
	{0x97054FD885BD6536ULL, "lt_0_lowtown_a2_n_c_15_0.mmh"},
	{0x970559D885BD7634ULL, "lt_0_lowtown_a2_n_c_15_0.msh"},
	{0x970566EDDF59C618ULL, "vxca_caverns2prc_4o3_0_1.vlm"},
	{0x970653BA91619058ULL, "6016933_m.fxe"},
	{0x9706A5AB3E83DF64ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\enter.gad"},
	{0x9706C3C0B21B419EULL, "6093477_m.fxe"},
	{0x97077BC687BFBA75ULL, "6103592_f.fxe"},
	{0x970788BD2E840E27ULL, "6110532_f.fxe"},
	{0x97079B91EE5CE829ULL, "827348636.ogg"},
	{0x9707FB5D212BE717ULL, "493556825.ogg"},
	{0x9708DD4D04FCEA1DULL, "vxht_0_hightown_n_c0l_0_0.vlm"},
	{0x970952F0F616AAD4ULL, "6076070_m.fxe"},
	{0x9709CDAA6BF0F00DULL, "145739884.ogg"},
	{0x970A5AE923BF28BDULL, "vxca_caverns1_mines_idh_0_0.vlm"},
	{0x970A73C6BA2F2B06ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mede.phy"},
	{0x970A855473885093ULL, "vxlt_gallowstemplar_4sp_0_0.vlm"},
	{0x970C4C687F548FE0ULL, "vxlt_warehouse_n_13p_0_0.vlm"},
	{0x970D3B6F70CF5624ULL, "vxlt_gamlenshouse_n_mo_0_0.vlm"},
	{0x970D565D8BBF95E7ULL, "z_dust8_si3.dds"},
	{0x970DA76BA8F8269CULL, "plt_lgt000pt_baz_shiv.nss"},
	{0x970DB4AB3E8A1742ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\enter.ani"},
	{0x970DF33860226B9DULL, "504080826.ogg"},
	{0x970DF3FBC32906BCULL, "454398818.ogg"},
	{0x970EA7A375D452A7ULL, "6105854_m.fxe"},
	{0x970EC20603030A1FULL, "vxlt_0_lowtown_a2_n_c_epb_0_0.vlm"},
	{0x970EF0BDDD7311F9ULL, "vxht_0_hightown_n_bxq_0_0.vlm"},
	{0x970EF406B53A9E3FULL, "genip_pp_hte_sustained.utp"},
	{0x970F8B2FC914E8A5ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb102a_0s.dds"},
	{0x970FE1F4EC3C131BULL, "104322098.ogg"},
	{0x971088F292E47B99ULL, "lt_0_lowtown_n.arl"},
	{0x9710BE6F36515992ULL, "99946832.ogg"},
	{0x9710C2DE4E38E536ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit_2.mao"},
	{0x9710DF873C90DEBBULL, "6080211_f.fxe"},
	{0x971110C4C170F322ULL, "vxw_blightlands_vg_0_0.vlm"},
	{0x9711289AF619F591ULL, "vxlt_shop_j7_0_0.vlm"},
	{0x97113C58498AC8EFULL, "zz_dae_debug6029852.cl"},
	{0x971219983E344A38ULL, "vxht_brothel_n_6gb_0_0.vlm"},
	{0x9712A23E48D664A3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1roothanging.mao"},
	{0x9712A63E48D66BF0ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1roothanging.mmh"},
	{0x9712B43E48D6830AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1roothanging.msh"},
	{0x9712F1DAE4CF6014ULL, "dae000_pc_home_ambients6208556.cl"},
	{0x971322A7C45130D2ULL, "422174415.ogg"},
	{0x971367100FD70DFDULL, "hlac_lt_0_lowtown_a3_n_c_-3_6x.dds"},
	{0x971417022B65D84BULL, "w_bonepitmining_a2_121_0.phy"},
	{0x97144241236B8265ULL, "257296767.ogg"},
	{0x9715F9A350B4F08DULL, "6054516_m.fxe"},
	{0x971641435358FDE3ULL, "vxw_woundedcoast_12q_0_0.vlm"},
	{0x971646B00DBB3863ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.dura.evt"},
	{0x9716567DA9597321ULL, "vxlt_0_lowtown_docks_n_c6o_0_0.vlm"},
	{0x97170F7FF6759967ULL, "hlac_lt_darktown_-1_41x.dds"},
	{0x97175197D5037229ULL, "lt_0_lowtown_docks_a3_10_0.anb"},
	{0x971766DB245C0065ULL, "6084113_m.fxe"},
	{0x9717AED9B80C1F84ULL, "607661994.ogg"},
	{0x9717D64062C32C0BULL, "vxlt_0_lowtown_a2_n_f4w_0_0.vlm"},
	{0x9717EEC36D3B76BAULL, "w_blightlands_98_0.mmh"},
	{0x9717F2C36D3B7DF1ULL, "w_blightlands_98_0.mao"},
	{0x971800C36D3B9500ULL, "w_blightlands_98_0.msh"},
	{0x97187ECB52FE8E70ULL, "vxlt_hangedman_1xn_0_0.vlm"},
	{0x9718A8243DBEFE84ULL, "6091841_f.fxe"},
	{0x971976DE5F786122ULL, "hlac_lt_0_lowtown_docks_a3_n_262143_0x.dds"},
	{0x9719D0C4E71F7BC2ULL, "hlac_w_bonepit_458755_0z.dds"},
	{0x9719E796D302DE9BULL, "vxlt_gallowstemplar_n_4ua_0_0.vlm"},
	{0x971A0834630AE143ULL, "lt_0_lowtown_docks_n_27_0.rml"},
	{0x971A5374662543C6ULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\pine_n.dds"},
	{0x971ABE45714AA860ULL, "vxlt_warehouse_n_18q_0_0.vlm"},
	{0x971B82382789BAFDULL, "263389404.ogg"},
	{0x971B9253392E1358ULL, "vxlt_darktown_5q8_0_0.vlm"},
	{0x971B99F454F1B89DULL, "qun160_party_comment_26061241.cl"},
	{0x971BBDA2742D6D62ULL, "hlac_dr_primevalentrance_1726_4x.dds"},
	{0x971BF7FCD6A5EB66ULL, "hla_ht_0_hightown_n_-131070_0z.dds"},
	{0x971C5DD3F005693CULL, "6164948_m.fxe"},
	{0x971C6CD103A1BECEULL, "isa311_isabela6091835.cl"},
	{0x971C8876D3EAD141ULL, "557084888.ogg"},
	{0x971D0C98A4EC8780ULL, "765596277.ogg"},
	{0x971D1B35A12E3655ULL, "603390664.ogg"},
	{0x971D1FB00DC114E4ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.dura.gad"},
	{0x971D47B1FC8E636CULL, "dae000_gamlen6147210.cl"},
	{0x971D48E329D43204ULL, "dae000_maren6205674.cl"},
	{0x971D8E61EE5CD42BULL, "6123388_m.fxe"},
	{0x971E040E6D06698FULL, "zz_dae_debug6029786.cl"},
	{0x971E5C9B6A7D80B0ULL, "vxlt_gallowstemplar_4y9_0_0.vlm"},
	{0x971E9CC43CF262EFULL, "301850174.ogg"},
	{0x971EDDCE4232FA70ULL, "vxlt_gallowsprison_n_5rj_0_0.vlm"},
	{0x971F3AE2A81B8C78ULL, "vxlt_0_lowtown_n_f5k_0_0.vlm"},
	{0x971F6ED1430212E2ULL, "vxlt_0_lowtown_a2_n_fee_0_0.vlm"},
	{0x971FCBFC0BD5278BULL, "203574705.ogg"},
	{0x97203595FE4FF44CULL, "llac_ht_0_hightown_a3_65535_2.dds"},
	{0x9720CCBE82E94899ULL, "171953986.ogg"},
	{0x9720EC9D87305C20ULL, "dae000_madam_lusine6151704.cl"},
	{0x9721144F6B3F4029ULL, "vxht_mansion2_1t9_0_0.vlm"},
	{0x97213D620343995FULL, "vxw_bonepitdestroyed_a3_md_0_0.vlm"},
	{0x9721906EE5DFBDA4ULL, "182145036.ogg"},
	{0x9722445B40AB529FULL, "6069954_m.fxe"},
	{0x9723559FD335FE5BULL, "gen00fl_merrill6128974.cl"},
	{0x9723AEE5F0F45533ULL, "vxlt_0_lowtown_a2_n_ehp_0_0.vlm"},
	{0x9723E223AEEC3BBDULL, "vxht_mansion2b_33v_0_0.vlm"},
	{0x97242EB00DC74CC2ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.dura.ani"},
	{0x972441AA36B2436DULL, "6135903_m.fxe"},
	{0x97247A63F0E8F8EBULL, "96028929.ogg"},
	{0x97248B293EB3C51CULL, "rdr111dg_martin6055060.cl"},
	{0x9724A267895A3128ULL, "vxw_bonepitmining_a2_eq_0_0.vlm"},
	{0x9724A364618C5DE0ULL, "6173386_m.fxe"},
	{0x9724BE8672129F0DULL, "one110_pryce6088418.cl"},
	{0x9724C85A9B4F8BA1ULL, "vxlt_0_lowtown_docks_n_c1n_0_0.vlm"},
	{0x9724D1BF4A7290C6ULL, "mcr370_desten6087322.cl"},
	{0x9724D8E973C0AE4DULL, "gen00fl_sebastian6188717.cl"},
	{0x97252DA8931B830EULL, "w_sundermount_67_0.mmh"},
	{0x972539A8931B97A5ULL, "w_sundermount_67_0.mao"},
	{0x972547A8931BAFFCULL, "w_sundermount_67_0.msh"},
	{0x972595D901F6BCF4ULL, "124163061.ogg"},
	{0x972601BB5B859113ULL, "106017237.ogg"},
	{0x972623056520E9F5ULL, "6095784_m.fxe"},
	{0x9726511C65AC8896ULL, "853185136.ogg"},
	{0x97269D146C51790AULL, "6177186_m.fxe"},
	{0x97293EF3ECEA4F1CULL, "6098540_m.fxe"},
	{0x9729AF6DBD40B0F0ULL, "vxht_mansion_n_5zm_0_0.vlm"},
	{0x9729CBFA4875293AULL, "372212285.ogg"},
	{0x972AC6413A4E37CFULL, "dae000_arishok6068571.cl"},
	{0x972B526DB13CB866ULL, "6060556_m.fxe"},
	{0x972BAABBE8BB1220ULL, "vxht_bartrand_n_5x8_0_0.vlm"},
	{0x972BC79CB4FDCBC1ULL, "vxlt_backalley_n_1l5_0_0.vlm"},
	{0x972BCFA098D95374ULL, "6118240_m.fxe"},
	{0x972BD477C15F42F9ULL, "vxht_0_hightown_n_bnw_0_0.vlm"},
	{0x972BD7450A988BC9ULL, "vxlt_gallowsprison_n_6nd_0_0.vlm"},
	{0x972BF24F4F85E4A6ULL, "6033673_m.fxe"},
	{0x972C96FDC7A3E64AULL, "6041034_m.fxe"},
	{0x972D59A9777619B9ULL, "981549210.ogg"},
	{0x972D5C206E5419C4ULL, "w_bonepitmining_a2_116_0.msh"},
	{0x972D6E206E54382DULL, "w_bonepitmining_a2_116_0.mao"},
	{0x972D72206E543F66ULL, "w_bonepitmining_a2_116_0.mmh"},
	{0x972D7A95545C73ABULL, "6201851_m.fxe"},
	{0x972DAA05816B1455ULL, "608040160.ogg"},
	{0x972DBD213D53A028ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_5x.dds"},
	{0x972E62E5B7188580ULL, "6155391_m.fxe"},
	{0x972EA0360EDF2E91ULL, "dae300_bran6143448.cl"},
	{0x972EA73F89007416ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bloodsplatxa.mao"},
	{0x972EB33F890088B5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bloodsplatxa.mmh"},
	{0x972EB53F89008C2BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bloodsplatxa.msh"},
	{0x972EBC6CB5C74721ULL, "isa321_velasco6117627.cl"},
	{0x972F0EEDFF8D000BULL, "zz_approval_debug6003088.cl"},
	{0x972F12FC333E74C0ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_meda.phy"},
	{0x972F658626CB6491ULL, "dae000_templar_knight6165159.cl"},
	{0x972F9A0B9BEC502DULL, "vxw_sundermount_2o7_0_0.vlm"},
	{0x972F9F14B241DB8AULL, "vxdr_0_deeproads_fhb_0_0.vlm"},
	{0x9730A5619CF36656ULL, "vxca_caverns1_mines_gkq_0_1.vlm"},
	{0x9730E943BB20D73EULL, "hlac_w_encounter2_camp_65534_0y.dds"},
	{0x9731258962921808ULL, "hlac_lt_0_lowtown_-196609_0y.dds"},
	{0x9731EC4E43FDBAD1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock3a.mmh"},
	{0x9731F84E43FDCF72ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock3a.mao"},
	{0x9732064E43FDE68FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_rock3a.msh"},
	{0x973220C2E2EA6DB3ULL, "vxlt_0_lowtown_a2_n_ems_0_0.vlm"},
	{0x9732FA1A6EBA3D5FULL, "lt_0_lowtown_docks_a3_n_111_0.msh"},
	{0x9733001A6EBA4701ULL, "lt_0_lowtown_docks_a3_n_111_0.mmh"},
	{0x97330C1A6EBA5BA2ULL, "lt_0_lowtown_docks_a3_n_111_0.mao"},
	{0x97332B0B579B7F50ULL, "vxca_caverns1_i7u_0_0.vlm"},
	{0x9733434C3860DA83ULL, "one120_athenril6089427.cl"},
	{0x9733443C2B131D4EULL, "art\\characters\\heraldry\\hld_circle_of_mages_2.dds"},
	{0x97334CA89327F50DULL, "w_sundermount_67_0.anb"},
	{0x9733C23B7F30E63AULL, "leave_05_l2.mao"},
	{0x9733CCE9A31CCD31ULL, "weapon_rendertoshadowmap_p1.psh.fxo.dx11"},
	{0x9733F7BF983E0846ULL, "6149961_m.fxe"},
	{0x9734120CB17CDFC5ULL, "art\\vfx\\combat\\vfx_m_bld_wound_aoe_imp_p.mmh"},
	{0x97342680D0E2DD19ULL, "art\\characters\\weapons\\longbows\\w_lbw_008a.msh"},
	{0x97343080D0E2EE07ULL, "art\\characters\\weapons\\longbows\\w_lbw_008a.mmh"},
	{0x97343480D0E2F554ULL, "art\\characters\\weapons\\longbows\\w_lbw_008a.mao"},
	{0x97344F9FCE7AB6D5ULL, "hlac_lt_0_lowtown_a2_n_65538_5y.dds"},
	{0x9734D25443403A3EULL, "character.mat"},
	{0x97351D4561BED492ULL, "ser270_templar6112634.cl"},
	{0x9735C996D54E534EULL, "vxca_caverns1_mines_i2m_0_0.vlm"},
	{0x9735CE6C07124376ULL, "dae300_bran6142895.cl"},
	{0x973668A84910B633ULL, "gen00fl_merrill6116070.cl"},
	{0x97367A28E6C84201ULL, "6138802_f.fxe"},
	{0x9736D55F1A674159ULL, "tacticsbaseconditions.gda"},
	{0x9736F19466C1FAA0ULL, "zz_followers_banter6159486.cl"},
	{0x9736FC44E26D9F08ULL, "vxw_encounter3_fx_0_1.vlm"},
	{0x97371C58B4A2D71FULL, "hlac_dr_dungeon_3214_1z.dds"},
	{0x97372F680E69A538ULL, "194897187.ogg"},
	{0x97373FAE6740B609ULL, "vxlt_darktown_4f5_0_0.vlm"},
	{0x973777F9437D26A8ULL, "hlac_w_bonepit_262147_0x.dds"},
	{0x9737820726E3B6ECULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basic.msh"},
	{0x9737880726E3C0DEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basic.mmh"},
	{0x9737940726E3D575ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_basic.mao"},
	{0x9737A816DCF8C579ULL, "mag210_arianni6024570.cl"},
	{0x9738198FBBD505F0ULL, "6092476_m.fxe"},
	{0x973852DE4F3071B4ULL, "743523985.ogg"},
	{0x97399B1E03A94B40ULL, "hlac_ca_caverns2prc_5802_26z.dds"},
	{0x9739A74498974A1DULL, "205660495.ogg"},
	{0x9739DAFD1F1B16BDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgtb_0t.dds"},
	{0x973A1246DC86E81BULL, "678511180.ogg"},
	{0x973A2D84A663D694ULL, "w_woundedcoast_40_0.rml"},
	{0x973A3979A6F3E441ULL, "vxlt_backalley_n_1w4_0_0.vlm"},
	{0x973A4BAC395EA46DULL, "6031129_f.fxe"},
	{0x973ACAF75724E055ULL, "llac_w_blightlands_327679_0.dds"},
	{0x973B7E0084687034ULL, "6083382_m.fxe"},
	{0x973BC0FB4223FDA6ULL, "gen_im_delete_me_04.uti"},
	{0x973BE15F291A4B34ULL, "6144205_m.fxe"},
	{0x973C3CA9FF0B4E71ULL, "zz_mer_debug6039439.cl"},
	{0x973C91BDF05F821CULL, "prp_str_twall_d.dds"},
	{0x973C99F29309D3AFULL, "lt_0_lowtown_n.rml"},
	{0x973CAEC4E76785FDULL, "vxca_caverns2_7e7_0_0.vlm"},
	{0x973CEA5CBFFD56D3ULL, "282446588.ogg"},
	{0x973CFB4227084664ULL, "6051851_m.fxe"},
	{0x973D0CF3C23DCC63ULL, "projdecal_projdecal_p0.psh.fxo.dx11"},
	{0x973D2C15D8B81D88ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_basic.mmh"},
	{0x973D3815D8B8322BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_basic.mao"},
	{0x973D4A15D8B850B2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ci_basic.msh"},
	{0x973D545A27E91F24ULL, "hlac_ht_keep_3732_15x.dds"},
	{0x973DC2EC49B05FA4ULL, "hlac_lt_0_lowtown_docks_-4_4y.dds"},
	{0x973DD76EE87759FCULL, "6042976_m.fxe"},
	{0x973E406CEE0DCB28ULL, "6202354_m.fxe"},
	{0x973E4D2EEBE796FFULL, "vxht_mansion2b_3ev_0_0.vlm"},
	{0x973EBF4FACEAECE8ULL, "845907234.ogg"},
	{0x973F29EBE9429BD8ULL, "gen_im_trs_krk_crack_armor.uti"},
	{0x973F96EC4E705D67ULL, "vxht_0_hightown_a3_ai5_0_0.vlm"},
	{0x9740C963820B2ABEULL, "6175041_m.fxe"},
	{0x974186216D466228ULL, "vxw_bonepitmining_a2_oo_0_0.vlm"},
	{0x97420F0CB1891792ULL, "art\\vfx\\combat\\vfx_m_bld_wound_aoe_imp_p.anb"},
	{0x9742A19EA559D77CULL, "lla_ht_0_hightown_a2_65537_0.dds"},
	{0x97434A2F1EAD5768ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.anb"},
	{0x97434E7F38A92497ULL, "6137279_m.fxe"},
	{0x974362855BD7DC64ULL, "42468662.ogg"},
	{0x9743833CF2A204BFULL, "gut_link_smo.mao"},
	{0x974456E22D275FDCULL, "hlac_lt_0_lowtown_a3_n_c_131070_4y.dds"},
	{0x974495F863244054ULL, "6150345_m.fxe"},
	{0x974505867DAA05CEULL, "zz_dae_debug6029437.cl"},
	{0x97451B3C95DE5870ULL, "llac_lt_foundry_n_320_12.dds"},
	{0x97455FF00EDC7AA0ULL, "hlac_lt_0_lowtown_docks_-65537_3z.dds"},
	{0x97462BD652C5743BULL, "518860199.ogg"},
	{0x974649B77CF6D889ULL, "llac_lt_undercitydungeon_-1_31.dds"},
	{0x9746A6038CE929EFULL, "688032219.ogg"},
	{0x9746CB8F4DBF210DULL, "mag101_grace6053100.cl"},
	{0x9747450863AE5F20ULL, "dev_res_smo.mao"},
	{0x9747B71B9A4AD1FBULL, "6128819_m.fxe"},
	{0x974842B824DB3F17ULL, "vxw_sundermount_352_0_0.vlm"},
	{0x974897481B8DD54AULL, "212032909.ogg"},
	{0x9748A308A6D7D9EAULL, "vxht_mansion2_1ss_0_0.vlm"},
	{0x9748E3EF7755662AULL, "315433492.ogg"},
	{0x974908CB9247A7B5ULL, "gen00fl_isabela6090882.cl"},
	{0x97498F414A77BED0ULL, "6092094_m.fxe"},
	{0x9749E8BA940F92B8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masd_0d.dds"},
	{0x9749EC5566E62B26ULL, "scaleform_sf_depth_p0.psh.fxo.dx11"},
	{0x974A3B7804F92905ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_4z.dds"},
	{0x974A5401A04A1DE7ULL, "hlac_ht_0_hightown_0_1y.dds"},
	{0x974AC50C8EDA18BCULL, "hlac_w_sundermount_-6_0x.dds"},
	{0x974B8694E503AF72ULL, "zz_rain_o_arrows_v2.cut"},
	{0x974BA75CBC08E0E6ULL, "6116482_m.fxe"},
	{0x974C2D1A1D09E5EBULL, "703900772.ogg"},
	{0x974C3AC2E4E3962DULL, "6156254_m.fxe"},
	{0x974C6210B9397466ULL, "vxht_bartrand_n_5ai_0_0.vlm"},
	{0x974C7DC57FD8812DULL, "vxw_sundermount_2u5_0_0.vlm"},
	{0x974CB0988CEC256EULL, "hm_qmq_minion3.mor"},
	{0x974D07DD2955C6A2ULL, "hlac_ht_brothel_n_3373_2y.dds"},
	{0x974D45F9D4144218ULL, "vxw_bonepit_1pw_0_0.vlm"},
	{0x974D627A23DAAE3BULL, "qcr220_grey_wardens6060426.cl"},
	{0x974E08C9406675E7ULL, "vxht_0_hightown_a3_av4_0_0.vlm"},
	{0x974E8B6657CB5401ULL, "hlac_dr_0_deeproads_5845_10z.dds"},
	{0x974EF6D3EB2CE738ULL, "214257546.ogg"},
	{0x974F3572B5E8170EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_exit.ani"},
	{0x974F58A0E67E6423ULL, "seb300im_mage_shield.uti"},
	{0x974F758D52D89B30ULL, "vxht_0_hightown_cb8_0_0.vlm"},
	{0x974FF7FE5F3C7AA8ULL, "vxw_bonepitmining_a2_jl_0_0.vlm"},
	{0x975072A847161BEFULL, "vxht_keep_3c8_0_0.vlm"},
	{0x97510B5FD0856A96ULL, "w_bonepit_178_0.msh"},
	{0x9751155FD0857B94ULL, "w_bonepit_178_0.mmh"},
	{0x9751195FD08582A7ULL, "w_bonepit_178_0.mao"},
	{0x97515D3F1214B5C7ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_activ_c.anb"},
	{0x9751689F1D8722E9ULL, "vxht_bartrand_n_2ny_0_1.vlm"},
	{0x97519558F752B432ULL, "hlac_ht_0_hightown_n_0_0z.dds"},
	{0x9752AD46D92D22D7ULL, "dae000_meredith6166270.cl"},
	{0x9752F8CA2D8935E4ULL, "107969093.ogg"},
	{0x97539D0E059E8BD8ULL, "vxw_woundedcoast_16r_0_0.vlm"},
	{0x9753A3A017E0083EULL, "416638477.ogg"},
	{0x97542FFC335DC255ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_meda.msh"},
	{0x975431FC335DC5CBULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_meda.mmh"},
	{0x97547CB9DE26DF78ULL, "vxlt_warehouse_n_101_0_0.vlm"},
	{0x97548DBD569758CEULL, "gen_im_cns_wrn_primeval.uti"},
	{0x9755511185B4BFFDULL, "vxlt_0_lowtown_h5g_0_0.vlm"},
	{0x97567EEC0EC6A973ULL, "vxca_cavern1_haunted_imm_0_0.vlm"},
	{0x9756A62811FDC6CEULL, "vxlt_gallowstemplar_f_37a_0_0.vlm"},
	{0x9757041182E5F07CULL, "6149461_f.fxe"},
	{0x975764221F09C0B6ULL, "hlac_dr_0_deeproads_5849_11z.dds"},
	{0x975789A788499A24ULL, "vxdr_primevaltomb_2xb_0_0.vlm"},
	{0x9758266600E71AE8ULL, "6114824_m.fxe"},
	{0x97582996E11397AFULL, "vxht_mansion_n_6nl_0_1.vlm"},
	{0x975872F2AF8951F0ULL, "796609663.ogg"},
	{0x9758FF8D1ECE8A3BULL, "qun221_viscount6084723.cl"},
	{0x975A10875FB876C4ULL, "hlac_ht_mansion2b_818_20z.dds"},
	{0x975A93504AE024D9ULL, "gen00fl_merrill6129672.cl"},
	{0x975AEB34634246BFULL, "lt_0_lowtown_docks_n_27_0.anb"},
	{0x975AF290FC633F7AULL, "and230_mage_girl6109110.cl"},
	{0x975B808D086B04C6ULL, "vxw_bonepit_26t_0_0.vlm"},
	{0x975B90323F229E87ULL, "6104347_m.fxe"},
	{0x975C2D72B5F2945FULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_exit.evt"},
	{0x975D238BBC000E27ULL, "dae210cr_human_m_01.utc"},
	{0x975D3BA3BF5C0E26ULL, "vxlt_0_lowtown_docks_a3_cyp_0_1.vlm"},
	{0x975D961837018EFDULL, "6156644_f.fxe"},
	{0x975DD051EE865E5EULL, "81526961.ogg"},
	{0x975DD0DBD1B254B8ULL, "meleering7.gda"},
	{0x975E22A60A4A5F9BULL, "202890172.ogg"},
	{0x975E7F59DBDCE891ULL, "vxlt_gallowscourtyard_751_0_0.vlm"},
	{0x975EFCF3790D8909ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor_mirror.ani"},
	{0x975F1FB3E23879CCULL, "8948018.ogg"},
	{0x975F39D549735DC6ULL, "361091251.ogg"},
	{0x975F591279B20CB9ULL, "body_tint_masks.gda"},
	{0x975FA212DB9D85A5ULL, "897818996.ogg"},
	{0x97607FD84AC6205BULL, "605438322.ogg"},
	{0x9760915496F46A6EULL, "vxht_mansion2b_2k8_0_0.vlm"},
	{0x976097C4568B9240ULL, "mcr310_argument6127583.cl"},
	{0x9760FDCE6A96A987ULL, "44626403.ogg"},
	{0x976183ED1B04027DULL, "6169876_m.fxe"},
	{0x9762133F4413ED34ULL, "6108924_m.fxe"},
	{0x9762B63F67677265ULL, "150298189.ogg"},
	{0x97630E72B5F87E48ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_exit.gad"},
	{0x9763433C7B258CC1ULL, "hlac_w_bonepitdestroyed_a3_393222_0x.dds"},
	{0x97640516FA10D569ULL, "zz_dae_debug6029797.cl"},
	{0x9764F0C900BCC1F3ULL, "vxca_cavern1_haunted_ijl_0_0.vlm"},
	{0x9764F61CE8CEB691ULL, "vxht_0_hightown_a3_bxk_0_0.vlm"},
	{0x9765325FEB1E0574ULL, "6033971_crouchtransition.cut"},
	{0x97653663A6CFB7B4ULL, "llac_lt_0_lowtown_a2_n_c_-2_1.dds"},
	{0x97656DA8F3F7DDACULL, "gen00fl_anders6114512.cl"},
	{0x97659D7AFE670002ULL, "vxlt_gallowstemplar_f_3mc_0_0.vlm"},
	{0x9765B86BA942B9C7ULL, "plt_lgt000pt_baz_shiv.plo"},
	{0x9765C92CAB11DFF7ULL, "242546525.ogg"},
	{0x976617F37913D58BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor_mirror.gad"},
	{0x97664081062D367BULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.dura.ani"},
	{0x97665C24CDA6CA67ULL, "zz_isa_debug6133246.cl"},
	{0x97669842F2DD3368ULL, "6113924_m.fxe"},
	{0x9767286CE6534853ULL, "ts_slavers_01.gda"},
	{0x9767457D7DFB515AULL, "190680558.ogg"},
	{0x97674BB9D7437595ULL, "176493454.ogg"},
	{0x976753C317AFCAF0ULL, "vxw_encounter2_camp_5g_0_0.vlm"},
	{0x9767FA17030FCC82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_loop.gad"},
	{0x976851D5D334E9D2ULL, "dae100_bran6142992.cl"},
	{0x9768734F26F58253ULL, "qun231_varnell6077244.cl"},
	{0x9768E3365E8BC34EULL, "vxht_0_hightown_n_9x9_0_0.vlm"},
	{0x9769149A512E928BULL, "gen00fl_varric6103450.cl"},
	{0x976945CA9D25A566ULL, "vxht_bartrand_n_5ok_0_0.vlm"},
	{0x976A1DDA8660A9CAULL, "6206873_m.fxe"},
	{0x976A7E2ED35239D5ULL, "mag291_demon_marethari6028912.cl"},
	{0x976A8DB209ADB544ULL, "vxlt_0_lowtown_docks_a3_n_d75_0_0.vlm"},
	{0x976B008BA33A3351ULL, "fx_z.dds"},
	{0x976B05A6DC3D3A1AULL, "94346156.ogg"},
	{0x976B1C7E8633117CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_110a_d.dds"},
	{0x976B53F2023CABABULL, "hlac_lt_0_lowtown_n_-1_5x.dds"},
	{0x976C0B45BAE5BF7FULL, "122744847.ogg"},
	{0x976C2F677A2A31BCULL, "art\\characters\\creatures\\shriek\\animation\\blendtree\\shriek.anb"},
	{0x976C3639AD9884DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7base.mmh"},
	{0x976C3A39AD988B11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7base.mao"},
	{0x976C4839AD98A320ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7base.msh"},
	{0x976C492F1ED05A03ULL, "art\\vfx\\combat\\vfx_glb_arw_ice_c.mmh"},
	{0x976C7DE074DAD51FULL, "vxlt_0_lowtown_docks_a3_9bs_0_0.vlm"},
	{0x976CE9EC18437725ULL, "one350_variel6085392.cl"},
	{0x976CF16E77A14579ULL, "vxw_encounter3_7g_0_0.vlm"},
	{0x976D14F37919EFF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\abomination_sailor_mirror.evt"},
	{0x976D1A4FEC667E14ULL, "llac_lt_hangedman_n_-1_28.dds"},
	{0x976D2F4C0C1B9BAEULL, "pro000st_aveline_wesley_int.stg"},
	{0x976D42A44FAE2531ULL, "vis120_commoner36069052.cl"},
	{0x976E3EEE74CDCBD7ULL, "art\\vfx\\materials\\vfx_water_decal_add.mao"},
	{0x976E3FDF77017279ULL, "gen00fl_fenris6106723.cl"},
	{0x976E49D651EF406AULL, "hlac_lt_0_lowtown_docks_a3_65535_5z.dds"},
	{0x976E7C6EA7C07AE5ULL, "dae000_bartrand6210162.cl"},
	{0x976E84ED404FAF62ULL, "vxht_mansion2b_2ox_0_0.vlm"},
	{0x976E9DC89AE646FAULL, "639975199.ogg"},
	{0x976E9E661DC6D341ULL, "6160952_m.fxe"},
	{0x976EC450DFAFF603ULL, "331431545.ogg"},
	{0x976EF66EFD6400D1ULL, "nukertacticpreset.gda"},
	{0x976F13CD371A160CULL, "6135335_m.fxe"},
	{0x976F2884DEF9748EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\kd_lp.ani"},
	{0x976F35DFBFCD8EE3ULL, "204263237.ogg"},
	{0x976F8D7071F179CBULL, "llac_lt_undercityencounter_-1_50.dds"},
	{0x976F8F7FE6A737AEULL, "hlac_w_bonepitmining_a2_0_5y.dds"},
	{0x976FA9E3C8D7CB2BULL, "521402384.ogg"},
	{0x976FEC85250A8B82ULL, "w_bonepit_109_0.msh"},
	{0x976FEE85250A8F38ULL, "w_bonepit_109_0.mmh"},
	{0x976FF74248293488ULL, "vxca_caverns1_mines_j26_0_2.vlm"},
	{0x976FFA85250AA3DBULL, "w_bonepit_109_0.mao"},
	{0x97708FBF74C3B8F5ULL, "hlac_dr_primevaltomb_11985_1z.dds"},
	{0x9770929079A19D1FULL, "hlac_ca_cavern1_haunted_24423_8z.dds"},
	{0x97709C4F7BF21D4BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed2.ani"},
	{0x9771546314C15A8CULL, "821642124.ogg"},
	{0x9771AA5A6277BE33ULL, "war_auraofpain.dds"},
	{0x9771AE15E400C7AAULL, "zz_qcr_debug6074963.cl"},
	{0x97721AF8195DE430ULL, "6134528_f.fxe"},
	{0x97726740A24C8431ULL, "6121894_m.fxe"},
	{0x977267C0FBC77314ULL, "440926160.ogg"},
	{0x977274E9D9239019ULL, "gen00fl_anders6119534.cl"},
	{0x977362A5F2B2DA73ULL, "vxca_cavern1_haunted_igo_0_0.vlm"},
	{0x977367F9DAC4CF11ULL, "vxht_0_hightown_a3_bwj_0_0.vlm"},
	{0x97736CC98A928DFBULL, "hlac_w_encounter3_-196608_1x.dds"},
	{0x977389E1F5E9F7CEULL, "vxlt_gallowstemplar_f_39c_0_0.vlm"},
	{0x9773B477753AE81FULL, "22281523.ogg"},
	{0x9773D212F740198DULL, "vxca_cavern1_haunted_i71_0_0.vlm"},
	{0x9773F0E78DD01769ULL, "lla_ht_0_hightown_n_65538_1.dds"},
	{0x97742D84A6954F57ULL, "w_woundedcoast_40_0.mmh"},
	{0x97743184A6955664ULL, "w_woundedcoast_40_0.mao"},
	{0x97744384A69574C9ULL, "w_woundedcoast_40_0.msh"},
	{0x9774488106398172ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.dura.evt"},
	{0x97749C777BB8DE91ULL, "6094721_m.fxe"},
	{0x9774DB91EBBD2CC5ULL, "448504932.ogg"},
	{0x9774E045D24FC69BULL, "mcr371_orsino6083318.cl"},
	{0x9775049BD4DFDB6DULL, "vxdr_primevalentrance_2l6_0_1.vlm"},
	{0x97752A0CAD490615ULL, "vxw_encounter2_camp_p4_0_1.vlm"},
	{0x97757C3E7046D95DULL, "203432386.ogg"},
	{0x97759A8404A86AA7ULL, "dae010cr_joker2.utc"},
	{0x97764989015D5484ULL, "hlac_ht_chantry_2534_15y.dds"},
	{0x97765B198C491F72ULL, "6106609_m.fxe"},
	{0x97766C471F9E7BB5ULL, "vxw_sundermount_2f4_0_0.vlm"},
	{0x97769723ED728116ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appa_0n.dds"},
	{0x9776BE03068A5552ULL, "hlac_w_encounter2_196613_0z.dds"},
	{0x97770A7E0162AD11ULL, "6105827_m.fxe"},
	{0x97771DD29DEB88FDULL, "vxht_chantry_n_2pm_0_0.vlm"},
	{0x977755D4775D53C7ULL, "ser270_emeric6149345.cl"},
	{0x9777D51BBADE5265ULL, "342750268.ogg"},
	{0x97783289B2BED219ULL, "w_bonepitdestroyed_a3_56_0.msh"},
	{0x97783C89B2BEE307ULL, "w_bonepitdestroyed_a3_56_0.mmh"},
	{0x97783EE28D533CD3ULL, "6072172_m.fxe"},
	{0x97784089B2BEEA54ULL, "w_bonepitdestroyed_a3_56_0.mao"},
	{0x97788E5D38ABE6D5ULL, "6149907_f.fxe"},
	{0x97791A3DAAD06141ULL, "lt_0_lowtown_docks_a3_3_0.anb"},
	{0x9779242DD2B3DABBULL, "mrl150_merrill6070614.cl"},
	{0x9779A173CBB142BDULL, "vxlt_0_lowtown_docks_n_aon_0_0.vlm"},
	{0x9779F5852512EFC3ULL, "w_bonepit_109_0.phy"},
	{0x977A0E41BC8DDA44ULL, "1110128.ogg"},
	{0x977A5A3F1237B5C8ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_activ_c.mmh"},
	{0x977A6781063E2105ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.dura.gad"},
	{0x977A6B1308E312AEULL, "85422456.ogg"},
	{0x977A9A72AE0C5B2CULL, "vxca_caverns2prc_a2x_0_0.vlm"},
	{0x977B0E8C03EA92A7ULL, "docks_scene_dockworkers6176483.cl"},
	{0x977B5DFC61173370ULL, "781164102.ogg"},
	{0x977B6EA6504896B0ULL, "gen_im_trs_apo_rng_r2.uti"},
	{0x977B820060273965ULL, "hlac_lt_0_lowtown_docks_-65539_0x.dds"},
	{0x977BFAD95B5883D3ULL, "6129265_m.fxe"},
	{0x977C2084DF03F1DFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\kd_lp.evt"},
	{0x977CB1D25A0631EBULL, "6207560_f.fxe"},
	{0x977CD0D50E6642DDULL, "39429626.ogg"},
	{0x977D251703227268ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_loop.ani"},
	{0x977D7386AA5D5FB1ULL, "vxht_0_hightown_n_brk_0_0.vlm"},
	{0x977D939993A353BAULL, "llac_lt_0_lowtown_a2_n_131070_6.dds"},
	{0x977DBB96470D7DD0ULL, "400865290.ogg"},
	{0x977E0CE83F03FBD9ULL, "llac_lt_0_lowtown_a3_n_c_65534_5.dds"},
	{0x977E6087614CF250ULL, "790511023.ogg"},
	{0x977E710282684170ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_1x.dds"},
	{0x977E844F7BFE31C2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed2.evt"},
	{0x977E937046CF71B6ULL, "607321263.ogg"},
	{0x977ED1473426F179ULL, "401224769.ogg"},
	{0x977EE5A20AF62580ULL, "6078726_m.fxe"},
	{0x977F1F8258E5EC2DULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_tool.msh"},
	{0x977F2D8258E60460ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_tool.mao"},
	{0x977F318258E60AB3ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blades_tool.mmh"},
	{0x977F31AC4D72B06AULL, "6123415_m.fxe"},
	{0x977F5D6F728ECA50ULL, "hlac_ht_0_hightown_n_131074_0z.dds"},
	{0x977FF84DBFAC71EFULL, "6101501_f.fxe"},
	{0x97806B8070B339BCULL, "514380182.ogg"},
	{0x978136DEE0380A33ULL, "vxht_mansion2b_3gl_0_0.vlm"},
	{0x978156C453A951C9ULL, "6075198_f.fxe"},
	{0x9781D9D6CCBAE791ULL, "vxht_0_hightown_a3_bri_0_0.vlm"},
	{0x9781DA6E3A23D6FEULL, "gen00fl_aveline6139826.cl"},
	{0x9782578949721B64ULL, "static_static_p67.psh.fxo.dx11"},
	{0x9782D5042828A0E2ULL, "6104193_m.fxe"},
	{0x97830184DF09DBC8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\kd_lp.gad"},
	{0x9783D4423D1FBD3AULL, "6121112_m.fxe"},
	{0x9783FCD13DCFB57CULL, "6208576_m.fxe"},
	{0x9784834F7C029BD5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.opentoclosed2.gad"},
	{0x9784B3E8707A51D5ULL, "vxdr_primevaltomb_4hq_0_0.vlm"},
	{0x978506FA804B0583ULL, "gen00fl_anders6119446.cl"},
	{0x9785D7160ADA5BDDULL, "603933320.ogg"},
	{0x9785DAE8D23EA36AULL, "6117012_m.fxe"},
	{0x978625308856F4FCULL, "hlac_w_encounter1_131071_0x.dds"},
	{0x97864B6D9A0C39F8ULL, "vxht_0_hightown_a2_9v2_0_0.vlm"},
	{0x97866277926E0F24ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_168bb.mmh"},
	{0x97866677926E1677ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_168bb.mao"},
	{0x9786677F7E3364FDULL, "vxca_cavern1_haunted_efz_0_0.vlm"},
	{0x978670971CF48A8EULL, "6125628_m.fxe"},
	{0x97867877926E34C6ULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_168bb.msh"},
	{0x97869A2EF38D07A4ULL, "vxht_playershouse_n_1sc_0_0.vlm"},
	{0x9786ABB18801DFABULL, "6107153_m.fxe"},
	{0x97874D8D4CF33C0CULL, "892347489.ogg"},
	{0x9787DD120428B017ULL, "6214713_f.fxe"},
	{0x9787E82907E47B12ULL, "zz_one_debug6077968.cl"},
	{0x9788628067E3C576ULL, "qun110_petrice6036235.cl"},
	{0x9788D04FC835340CULL, "136717036.ogg"},
	{0x978954ED7EDCA144ULL, "hlac_dr_slavers_3207_0x.dds"},
	{0x97895F347E0D1D84ULL, "vxca_caverns1_mines_e5z_0_0.vlm"},
	{0x9789E32B68AFFC06ULL, "zz_isa_debug6133159.cl"},
	{0x978A01D158F00F0BULL, "802111835.ogg"},
	{0x978B23AA97EE1C3BULL, "hlac_w_sundermount_262139_0y.dds"},
	{0x978B5871E3E78B21ULL, "290767406.ogg"},
	{0x978B603CE3C5FAEFULL, "hlac_lt_0_lowtown_n_65537_1x.dds"},
	{0x978BA188C3D4408FULL, "6115315_f.fxe"},
	{0x978BB922DEF8E554ULL, "follower_barks6154985.cl"},
	{0x978C417BA7F35215ULL, "6149495_f.fxe"},
	{0x978D019348B1E4F0ULL, "vxca_caverns1_j2e_0_0.vlm"},
	{0x978D4B66C395FBBFULL, "166632867.ogg"},
	{0x978DA3893F68C8EAULL, "6128515_m.fxe"},
	{0x978E2A08CE7CCED4ULL, "zz_followers_banter6157909.cl"},
	{0x978FE57852F8DBFCULL, "vxlt_undercityencounter_1k0_0_0.vlm"},
	{0x9790249E218320EDULL, "gen_im_trs_gen_med_glv_r0.uti"},
	{0x9790F311D4494982ULL, "vxlt_gallowstemplar_f_3rn_0_0.vlm"},
	{0x979193D91E25B38DULL, "12437772.ogg"},
	{0x97919548DF3E76CDULL, "vxlt_warehouse_n_um_0_1.vlm"},
	{0x97920EA233087C04ULL, "dae000_gen_turn_in_46178435.cl"},
	{0x97925539ADB988E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow7base.phy"},
	{0x979261FCB50A7273ULL, "gen00fl_carver6104538.cl"},
	{0x9792A0FCAD6EA49BULL, "232643250.ogg"},
	{0x9793232CD7020C96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_roll_shoulders.ani"},
	{0x97936BD8241F70E8ULL, "vxlt_shop_ci_0_0.vlm"},
	{0x9794970745EB384AULL, "w_woundedcoast_161_0.mao"},
	{0x97949B0745EB3E99ULL, "w_woundedcoast_161_0.mmh"},
	{0x9794A50745EB4F87ULL, "w_woundedcoast_161_0.msh"},
	{0x9794B072ADBDA03BULL, "hlac_lt_0_lowtown_a3_n_c_-65537_6y.dds"},
	{0x97957F4A8DE2BC98ULL, "vxw_bonepit_1ip_0_0.vlm"},
	{0x9795C853C694EE1DULL, "6103942_m.fxe"},
	{0x97966D9870A8CBDFULL, "lgt311_shw_mer_mettin6156470.cl"},
	{0x9796852DAF9D73BDULL, "vxlt_0_lowtown_docks_n_aql_0_0.vlm"},
	{0x97970A8167B83D75ULL, "6107414_m.fxe"},
	{0x979732C4BED7C338ULL, "vxw_blightlands_bq_0_0.vlm"},
	{0x9797FDD3D383375AULL, "vxca_caverns2_8d0_0_0.vlm"},
	{0x97980590E2F94BA8ULL, "hla_ht_0_hightown_65536_0y.dds"},
	{0x979842EE2D32D01CULL, "114154152.ogg"},
	{0x9799932FA331472FULL, "6182473_m.fxe"},
	{0x979AD6A7E78801DBULL, "dae000_merchant_rare_em.bnk"},
	{0x979ADA89AC2DAD0BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen_v2.ani"},
	{0x979BAA785F9D0847ULL, "6060445_m.fxe"},
	{0x979BBFD3BC33B4E1ULL, "425098861.ogg"},
	{0x979BFEBC0747BCCBULL, "6087194_m.fxe"},
	{0x979C56BFC2465313ULL, "gen00fl_merrill6130126.cl"},
	{0x979CEF3DF5402A12ULL, "6102519_m.fxe"},
	{0x979CF6FCD962CEDBULL, "vxw_sundermount_2g9_0_0.vlm"},
	{0x979DDAB41D997A24ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_teleport.ani"},
	{0x979DE5B444A3FB87ULL, "vxw_encounter2_qq_0_1.vlm"},
	{0x979E33755A467499ULL, "1025327089.ogg"},
	{0x979E45E344EAC459ULL, "421969413.ogg"},
	{0x979E6F5D69E3F28FULL, "hlac_w_encounter2_-131072_0z.dds"},
	{0x979E99757DDA1FC5ULL, "mag360_thrask6044474.cl"},
	{0x979EB1D5888DE8F7ULL, "dae000cr_grand_cleric6183976.cl"},
	{0x979ED8A7E78BED9BULL, "dae000_merchant_rare_em.cnv"},
	{0x979F0B1647A77E70ULL, "907831046.ogg"},
	{0x979F64EEC63F6202ULL, "vxlt_gallowstemplar_f_3yo_0_0.vlm"},
	{0x979FBF2362565C74ULL, "459514163.ogg"},
	{0x979FCBF43A75C7A9ULL, "349220520.ogg"},
	{0x979FFC2BC51E8580ULL, "6161316_m.fxe"},
	{0x97A05C55FE0B9791ULL, "vxht_keep_53d_0_0.vlm"},
	{0x97A082A9301ECF8FULL, "6075219_f.fxe"},
	{0x97A0BB50F99E1693ULL, "vxht_0_hightown_c1v_0_0.vlm"},
	{0x97A12B2CD70E5787ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_roll_shoulders.evt"},
	{0x97A1404016B6854BULL, "llac_ca_caverns2prc_5802_61.dds"},
	{0x97A1AA4C2B8C3294ULL, "vxw_encounter2_fn_0_0.vlm"},
	{0x97A1C1DDF580C535ULL, "vxw_sundermount_2q1_0_0.vlm"},
	{0x97A1DDB516158968ULL, "vxlt_shop_ff_0_0.vlm"},
	{0x97A220B278A61DB0ULL, "vxlt_0_lowtown_n_exk_0_0.vlm"},
	{0x97A255853A40B26AULL, "839340454.ogg"},
	{0x97A2736973CDD27DULL, "vxht_chantry_n_2aj_0_0.vlm"},
	{0x97A29D535B3A4B5EULL, "plt_cod_ltr_elren_orders.nss"},
	{0x97A3BA9C278D95DBULL, "t3_wdg_t01.tnt"},
	{0x97A3D7B41D9DE086ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_teleport.gad"},
	{0x97A3F6649AEDF75CULL, "zz_ser_debug6085767.cl"},
	{0x97A43A30B88B069DULL, "6104842_m.fxe"},
	{0x97A47E896BF25B2BULL, "qcr000_varric6065839.cl"},
	{0x97A4D2DD67DDEE48ULL, "dae000_keep_guard_2_sta6163824.cl"},
	{0x97A518F214BC82D0ULL, "6074376_f.fxe"},
	{0x97A53F87FD3C4067ULL, "dae000_fenarel6133349.cl"},
	{0x97A61B1CB95869BCULL, "367821335.ogg"},
	{0x97A64918B89A77CCULL, "6053884_m.fxe"},
	{0x97A6576567B69B7DULL, "gen_im_trs_krk_junk1_01.uti"},
	{0x97A66F0D8EB46FEFULL, "6176492_m.fxe"},
	{0x97A67ADB07BAD493ULL, "mag321_comtesse_de_launcet6064896.cl"},
	{0x97A6C75C586B81B9ULL, "6167243_m.fxe"},
	{0x97A6ED6C09BA00ABULL, "vxdr_primevaltomb_7lp_0_0.vlm"},
	{0x97A7581EF133E149ULL, "vxlt_0_lowtown_n_fg8_0_0.vlm"},
	{0x97A784FA5DF56D76ULL, "264332943.ogg"},
	{0x97A7B6F4CEDAC5DFULL, "6059261_m.fxe"},
	{0x97A84C2CD714AED0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_roll_shoulders.gad"},
	{0x97A85113699046E7ULL, "mrl000_marethari6069858.cl"},
	{0x97A8C289AC39C182ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen_v2.evt"},
	{0x97A9AC6D74487465ULL, "207334583.ogg"},
	{0x97A9AEC1CE040E4DULL, "6099592_f.fxe"},
	{0x97AA71C433415431ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.dura.ani"},
	{0x97AA79700AD84D41ULL, "6116194_f.fxe"},
	{0x97AA91ADF27F5A0BULL, "clf260_harley6093811.cl"},
	{0x97AAD2B41DA3F765ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_teleport.evt"},
	{0x97AB68D9CB58E75BULL, "vxw_sundermount_2h8_0_0.vlm"},
	{0x97ABA52D745CFEA2ULL, "vxca_caverns1_fn7_0_0.vlm"},
	{0x97AD22AF9F41EF68ULL, "76494383.ogg"},
	{0x97AD2A830292865AULL, "874061712.ogg"},
	{0x97AD4934040E8C24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_wary.ani"},
	{0x97ADFDA7FE5309C3ULL, "gen_im_trs_bst_drg_gland_01.uti"},
	{0x97AE00A53C00E28FULL, "seb321ar_keep_nt.are"},
	{0x97AE037A267955C4ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.anb"},
	{0x97AE041EE4D65E73ULL, "hlac_dr_slavers_3214_7z.dds"},
	{0x97AE1837C67D7533ULL, "gen00fl_varric6118955.cl"},
	{0x97AF2D2DEB942F13ULL, "vxht_0_hightown_c4y_0_0.vlm"},
	{0x97AF4F503D58A057ULL, "character_character_p215.psh.fxo.dx11"},
	{0x97AF7FF0CED67536ULL, "hlac_ht_0_hightown_-131074_2x.dds"},
	{0x97AFC189AC3FDE95ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen_v2.gad"},
	{0x97AFC5B351E7FA99ULL, "246791118.ogg"},
	{0x97AFD6302C5148B7ULL, "mag101_aombanter6052757.cl"},
	{0x97B01D53DCBD5A5DULL, "llac_lt_0_lowtown_a3_n_c_65535_0.dds"},
	{0x97B02993119C0525ULL, "dae200cr_viscount6087195.cl"},
	{0x97B033BAE776DDB5ULL, "vxw_sundermount_2z0_0_0.vlm"},
	{0x97B0CBDD94F1FC1FULL, "227893888.ogg"},
	{0x97B0D96C4A1AC415ULL, "vxht_0_hightown_a2_7g2_0_0.vlm"},
	{0x97B17CC4334785D3ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.dura.gad"},
	{0x97B18CEFE7B55621ULL, "243237905.ogg"},
	{0x97B1D7BB0D8E37E7ULL, "vxht_keep_5gn_0_0.vlm"},
	{0x97B2639BF42E90ADULL, "qun221_viscount6084707.cl"},
	{0x97B27752267F0F62ULL, "l_dae250ar_sundermount.bnk"},
	{0x97B2FB3DAB01A4D1ULL, "lt_0_lowtown_docks_a3_3_0.rml"},
	{0x97B346340412F286ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_wary.gad"},
	{0x97B43A426762015FULL, "mcr370_templar16086970.cl"},
	{0x97B48490F5FE1A42ULL, "llac_lt_0_lowtown_docks_-2_5.dds"},
	{0x97B4AB65218DF3A3ULL, "6144358_m.fxe"},
	{0x97B4CA08FC60711DULL, "gen_im_arm_hlm_lgt_set_med.uti"},
	{0x97B4E0E106769AADULL, "6153345_m.fxe"},
	{0x97B4E809D281EDE1ULL, "6053327_f.fxe"},
	{0x97B4E8BA60A20DBFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_framechantry.phy"},
	{0x97B4FB98C81DB99CULL, "377544842.ogg"},
	{0x97B5E759A6257509ULL, "vxlt_gallowscourtyard_n_4cc_0_0.vlm"},
	{0x97B653CF85F43D8AULL, "vxlt_0_lowtown_a2_ekj_0_1.vlm"},
	{0x97B6C0A863B9CB7CULL, "vxdr_primevaltomb_8mb_0_1.vlm"},
	{0x97B6DD05A930915FULL, "w_bonepit_131_0.msh"},
	{0x97B6E305A9309B01ULL, "w_bonepit_131_0.mmh"},
	{0x97B6EF05A930AFA2ULL, "w_bonepit_131_0.mao"},
	{0x97B6FF558D92082FULL, "vxlt_gallowstemplar_f_3p5_0_0.vlm"},
	{0x97B75795DEE8CC98ULL, "vxht_0_hightown_a2_cb4_0_0.vlm"},
	{0x97B779C4334BEC38ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.dura.evt"},
	{0x97B8148FD3F20898ULL, "6174660_m.fxe"},
	{0x97B81681764E3F42ULL, "6146928_m.fxe"},
	{0x97B834A98211F67DULL, "gen00fl_fenris6126493.cl"},
	{0x97B97819877429E2ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_blueglowworms_sa.dds"},
	{0x97B9B2894BC2FCECULL, "hlac_ca_caverns2prc_5802_28x.dds"},
	{0x97B9B48DB3145A65ULL, "vxlt_foundry_n_119_0_0.vlm"},
	{0x97B9CB12A0ADD2E4ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_4y.dds"},
	{0x97BA18F8F51F773EULL, "vxw_bonepit_1s4_0_0.vlm"},
	{0x97BA251253B91E51ULL, "6103344_m.fxe"},
	{0x97BA413404190965ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_wary.evt"},
	{0x97BA425EE56B3BD6ULL, "vxlt_0_lowtown_docks_a3_n_ary_0_0.vlm"},
	{0x97BA99DBCC997ADAULL, "vxlt_gallowstemplar_2os_0_0.vlm"},
	{0x97BB227A268414C7ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.mmh"},
	{0x97BBC6BA43ABE89FULL, "gen00fl_aveline6114208.cl"},
	{0x97BBD882C87EF95AULL, "framebuffereffects_dofblend_p0.psh.fxo.dx11"},
	{0x97BBE9D6D100273DULL, "vxlt_0_lowtown_a2_n_eia_0_0.vlm"},
	{0x97BBEDB6CA1A7039ULL, "vxlt_gallowstemplar_520_0_0.vlm"},
	{0x97BBF58A7F5E9A66ULL, "6032933_m.fxe"},
	{0x97BC01040C54A873ULL, "171523042.ogg"},
	{0x97BC7D7D1C93AD0EULL, "vxca_caverns1_mines_dpv_0_0.vlm"},
	{0x97BC93FAE2CE2AE2ULL, "6072567_f.fxe"},
	{0x97BD70DB5ECB2FD0ULL, "vxlt_hangedman_n_1o0_0_0.vlm"},
	{0x97BDD9CCF9E49CD9ULL, "ca_caverns1_24421_lc.mmh"},
	{0x97BDDA1F87C3BCA6ULL, "6135607_m.fxe"},
	{0x97BE727C7AE56E01ULL, "qun211_braggart6075460.cl"},
	{0x97BEF9A53C0F4A4CULL, "seb321ar_keep_nt.lst"},
	{0x97BF40E963D82C6EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wmda_0s.dds"},
	{0x97BFA1E7E37740CEULL, "and111_lirene6084754.cl"},
	{0x97BFE298CB05A0F1ULL, "6149984_m.fxe"},
	{0x97BFEA2096B31D53ULL, "64329775.ogg"},
	{0x97BFFE5B7E5F29CAULL, "donate.ncs"},
	{0x97C0961F69D503B1ULL, "6150487_m.fxe"},
	{0x97C0C4CBFE6B90A4ULL, "306090215.ogg"},
	{0x97C16B372E9980B4ULL, "6141714_m.fxe"},
	{0x97C184538DAB59C8ULL, "6042150_m.fxe"},
	{0x97C1C164B9B6DC43ULL, "363251522.ogg"},
	{0x97C24A47D57D1403ULL, "hlac_lt_undercityencounter_-1_12z.dds"},
	{0x97C257C3E7F3D893ULL, "145038138.ogg"},
	{0x97C25F009F90C252ULL, "pro000_merchant6057572.cl"},
	{0x97C29C0399C2ABA6ULL, "6112125_m.fxe"},
	{0x97C352BDF86CB32DULL, "6156045_m.fxe"},
	{0x97C366DCF128EC56ULL, "vxlt_gallowscourtyard_n_5z6_0_0.vlm"},
	{0x97C38066655B7DD3ULL, "vxht_keep_36w_0_0.vlm"},
	{0x97C3D125EDA631ABULL, "vxdr_primevaltomb_7br_0_0.vlm"},
	{0x97C3DEDD25F632E8ULL, "vxlt_gallowscourtyard_76m_0_0.vlm"},
	{0x97C403C55211D7B9ULL, "vrc321_party_barks6095733.cl"},
	{0x97C40F5DD83927E0ULL, "290183289.ogg"},
	{0x97C44D64306B69F6ULL, "hlac_ht_0_hightown_a2_-65537_0x.dds"},
	{0x97C4881D3367E69FULL, "6055438_m.fxe"},
	{0x97C506BBFC04F3D1ULL, "hlac_lt_0_lowtown_docks_a3_-4_4y.dds"},
	{0x97C54BCCA1B2024CULL, "vxht_0_hightown_bi8_0_0.vlm"},
	{0x97C56A2A23220A15ULL, "hlac_ca_cavern1_haunted_24423_3y.dds"},
	{0x97C57D02C1F84099ULL, "vxca_caverns1_mines_iis_0_0.vlm"},
	{0x97C5D9A53C1532F5ULL, "seb321ar_keep_nt.ncs"},
	{0x97C5DB504B32F7C0ULL, "mer241ar_cavern_of_dead.are"},
	{0x97C5E9BA60B082C4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_framechantry.mmh"},
	{0x97C5EDBA60B08917ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_framechantry.mao"},
	{0x97C5FFBA60B0A866ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_framechantry.msh"},
	{0x97C6885E684457C2ULL, "6143628_m.fxe"},
	{0x97C6DF23089B98FEULL, "vxw_sundermount_2d5_0_0.vlm"},
	{0x97C7A6D951ACF832ULL, "6055080_m.fxe"},
	{0x97C7E2195C7BFECDULL, "383933682.ogg"},
	{0x97C80AEEA2D81F12ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.mmh"},
	{0x97C85BF55DF48D5DULL, "6001183_m.fxe"},
	{0x97C8D2504B351E42ULL, "mer241ar_cavern_of_dead.ncs"},
	{0x97C908C862422B00ULL, "hlac_dr_primevalentrance_1724_0x.dds"},
	{0x97C91F01B19C901BULL, "6122444_m.fxe"},
	{0x97C9765F82B3EBD8ULL, "690924109.ogg"},
	{0x97C992321EF51523ULL, "348241597.ogg"},
	{0x97C9BFA3115F0139ULL, "mer320_hubert6081994.cl"},
	{0x97CA457361EA9D02ULL, "vxca_caverns2_96w_0_0.vlm"},
	{0x97CA5F8ADD8758A1ULL, "40422599.ogg"},
	{0x97CAC9C265214F29ULL, "229619493.ogg"},
	{0x97CAEC2FFD12337FULL, "vxdr_dungeon_210_0_0.vlm"},
	{0x97CB32E7270AA0B9ULL, "6086256_m.fxe"},
	{0x97CBF3D8D476A86BULL, "6139380_f.fxe"},
	{0x97CCFA396BFD9B3AULL, "gen00fl_fenris6126931.cl"},
	{0x97CD3F41F67F018BULL, "vxlt_hangedman_n_1w6_0_1.vlm"},
	{0x97CE191537A87869ULL, "vxw_encounter3_fe_0_1.vlm"},
	{0x97CE796C4E18A0DEULL, "146664632.ogg"},
	{0x97CE923934593F89ULL, "870635077.ogg"},
	{0x97CF90AA5937D39BULL, "vxht_keep_6ge_0_0.vlm"},
	{0x97CFEA504B3B65D3ULL, "mer241ar_cavern_of_dead.lst"},
	{0x97D084EFE7193C75ULL, "zz_mag_debug6053403.cl"},
	{0x97D0957EC1A7436AULL, "297172121.ogg"},
	{0x97D17239ED219672ULL, "art\\characters\\playercharacter\\humanmale\\hn_arm_lgtb.mao"},
	{0x97D19A028A858A50ULL, "301454410.ogg"},
	{0x97D1B906E36BDF20ULL, "vxht_mansion2b_1td_0_0.vlm"},
	{0x97D1D6F23028C465ULL, "620144118.ogg"},
	{0x97D21C7898F1E538ULL, "vxw_sundermount_2ah_0_0.vlm"},
	{0x97D236F5D5C03A38ULL, "497291585.ogg"},
	{0x97D29D4E31D42F4CULL, "vxht_mansion_n_4ye_0_0.vlm"},
	{0x97D335CD1DABE62CULL, "vxw_woundedcoast_1fv_0_0.vlm"},
	{0x97D37438EC2EDF66ULL, "202610179.ogg"},
	{0x97D3A92E11C80BF6ULL, "6048442_m.fxe"},
	{0x97D3F8C7482BE2C2ULL, "ht_0_hightown_a3_21_0_ll.mmh"},
	{0x97D43383F36A796BULL, "vxlt_0_lowtown_n_h2c_0_0.vlm"},
	{0x97D496199AE1D02BULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_blightfire.mmh"},
	{0x97D4CA0EE66291D9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25.mmh"},
	{0x97D5489EB34CE024ULL, "506042812.ogg"},
	{0x97D550FFFA91B17EULL, "vxw_sundermount_2i6_0_0.vlm"},
	{0x97D55CD977CBD4C0ULL, "dae100_alistair_drunk6135159.cl"},
	{0x97D5DF634EAF71F6ULL, "zz_followers_banter6159228.cl"},
	{0x97D5EC05A90A732EULL, "vxlt_warehouse_n_12q_0_0.vlm"},
	{0x97D61F2F39D19CD9ULL, "gen00fl_varric6118948.cl"},
	{0x97D651D47F713046ULL, "1065630566.ogg"},
	{0x97D65E7FD24D47C7ULL, "164556448.ogg"},
	{0x97D6C8D739B84C05ULL, "vxw_bonepit_9b_0_0.vlm"},
	{0x97D738549953A934ULL, "ht_0_hightown_a2_29_0_lc.mmh"},
	{0x97D76152269E0580ULL, "l_dae250ar_sundermount.wbd"},
	{0x97D77414B437A91DULL, "ht_0_hightown_a3_33_0.anb"},
	{0x97D7D642C98EF53AULL, "hlac_lt_gallowstemplar_7198_5x.dds"},
	{0x97D84868609D9264ULL, "6098337_m.fxe"},
	{0x97D8634A27ED414FULL, "mer120_hubert6070042.cl"},
	{0x97D87DD2EB8EBF53ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.cess.evt"},
	{0x97D924710EB9A02AULL, "vxw_bonepit_26x_0_0.vlm"},
	{0x97DA1400CDBF9ED5ULL, "6079948_m.fxe"},
	{0x97DA3F0C0F89AA03ULL, "pro000_merchant6051270.cl"},
	{0x97DA8DD9E7336DC9ULL, "6110667_m.fxe"},
	{0x97DA92FA44806AB4ULL, "6050287_m.fxe"},
	{0x97DB26FB390A9B92ULL, "hlac_w_sundermount_131075_0x.dds"},
	{0x97DB71BFF3649494ULL, "vxw_encounter2_rr_0_0.vlm"},
	{0x97DBB0FDBC02AE76ULL, "vxca_cave_me_0_1.vlm"},
	{0x97DC0BC6E9162C2BULL, "6053003_m.fxe"},
	{0x97DC6C2B7325624CULL, "vxlt_gallowstemplar_5k2_0_0.vlm"},
	{0x97DCF4605286CB3FULL, "fen110_fenris6118019.cl"},
	{0x97DD009640E9CDAFULL, "vxlt_0_lowtown_docks_n_b0n_0_0.vlm"},
	{0x97DD5E321C1F617AULL, "vxlt_0_lowtown_docks_a3_n_d3g_0_0.vlm"},
	{0x97DDA234D1F76766ULL, "284931384.ogg"},
	{0x97DE032AED60868EULL, "vxlt_shop_em_0_0.vlm"},
	{0x97DEA9F0E4EC4816ULL, "490351415.ogg"},
	{0x97DEDF941B7C5E7FULL, "195427717.ogg"},
	{0x97DEFF5227EFF859ULL, "vxht_0_hightown_a3_bnt_0_1.vlm"},
	{0x97DF76D2EB94D194ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.cess.gad"},
	{0x97DFB2E7346E9AE9ULL, "310161555.ogg"},
	{0x97E01481AA228C63ULL, "hightown_male_noble_gossip6171397.cl"},
	{0x97E0749769B33DC0ULL, "batchexport2.arl"},
	{0x97E08E558AE7FDB8ULL, "vxw_sundermount_2ji_0_0.vlm"},
	{0x97E152A0C9F89044ULL, "art\\vfx\\combat\\vfx_r_gn4_escape_act_c.mmh"},
	{0x97E152FD40EFD93FULL, "gui_guipixelshader_p1.psh.fxo.dx11"},
	{0x97E1A7AA0FA1FEACULL, "vxw_woundedcoast_1aw_0_0.vlm"},
	{0x97E1ACFF4C4B1313ULL, "6179255_m.fxe"},
	{0x97E1B9FAEA1B074EULL, "hlac_lt_0_lowtown_docks_n_131071_0x.dds"},
	{0x97E1D349056CD533ULL, "633346840.ogg"},
	{0x97E260BCA5E47199ULL, "vxca_caverns1_mines_icq_0_0.vlm"},
	{0x97E277199AEDD890ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_blightfire.anb"},
	{0x97E28EC413D36219ULL, "6016756_m.fxe"},
	{0x97E33A138DE63F07ULL, "hlac_lt_gallowstemplar_n_2383_11x.dds"},
	{0x97E3C2DCEC87C9FEULL, "vxw_sundermount_2r7_0_0.vlm"},
	{0x97E4573F9719A1FDULL, "vxca_caverns2prc_5xk_0_0.vlm"},
	{0x97E63A10A9D36D3AULL, "6111757_m.fxe"},
	{0x97E65294653BC75FULL, "6107306_m.fxe"},
	{0x97E685D2EB9B0A72ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.cess.ani"},
	{0x97E68AB4C69066D5ULL, "6134250_m.fxe"},
	{0x97E6A96EBC0F023DULL, "6095072_m.fxe"},
	{0x97E6BCB193B25D47ULL, "one300_zevran6089105.cl"},
	{0x97E6DFDB30B4F06DULL, "hlac_lt_0_lowtown_docks_65534_4z.dds"},
	{0x97E6EC849901BF1AULL, "hlac_lt_0_lowtown_docks_n_-65536_0y.dds"},
	{0x97E756C6157DBA68ULL, "156829629.ogg"},
	{0x97E852EE7DE3E809ULL, "473176948.ogg"},
	{0x97E8F2B08C703848ULL, "6208355_f.fxe"},
	{0x97E91B12145407BAULL, "1034651381.ogg"},
	{0x97E929BCFEEBAEF5ULL, "vxlt_gallowscourtyard_61d_0_0.vlm"},
	{0x97E98CF9586E2937ULL, "vxlt_gallowstemplar_3gc_0_0.vlm"},
	{0x97E9B7E8D064425FULL, "vxht_mansion2b_1il_0_0.vlm"},
	{0x97E9DBD70F6AC477ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgtp_0t.dds"},
	{0x97EA27C9B662F69AULL, "dae000_meredith6085307.cl"},
	{0x97EA4472A6F07B43ULL, "6019104_m.fxe"},
	{0x97EA48D4F4D46994ULL, "hlac_lt_hangedman_n_-1_25y.dds"},
	{0x97EA562F80BB3D54ULL, "fen210_orana6133054.cl"},
	{0x97EADE08651B7ACCULL, "vxlt_gallowstemplar_5t3_0_0.vlm"},
	{0x97EBD3B306B3FE54ULL, "6172301_m.fxe"},
	{0x97EC07A3F763540FULL, "zz_followers_banter6157287.cl"},
	{0x97EC42A5C7102331ULL, "w_bonepitdestroyed_a3_179_0.mmh"},
	{0x97EC448DBC4A2878ULL, "vxht_keep_n_3yj_0_0.vlm"},
	{0x97EC4EA5C71037D2ULL, "w_bonepitdestroyed_a3_179_0.mao"},
	{0x97EC5CA5C7104FEFULL, "w_bonepitdestroyed_a3_179_0.msh"},
	{0x97EC6B434B248829ULL, "hlac_w_bonepit_131073_0z.dds"},
	{0x97EC6DD4E7C41FD4ULL, "611890221.ogg"},
	{0x97ED82535B7A08A9ULL, "plt_cod_ltr_elren_orders.plo"},
	{0x97EE2FE576AD4FE3ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_5z.dds"},
	{0x97EE7B366DCB5DAEULL, "6051660_m.fxe"},
	{0x97EE8470B3CF446BULL, "6000887_m.fxe"},
	{0x97EEDACC162C2E53ULL, "plt_cod_cha_varric1.plo"},
	{0x97EF00327CDE1638ULL, "vxw_sundermount_2of_0_0.vlm"},
	{0x97EF049BF9754FD8ULL, "pro000_merchant6051590.cl"},
	{0x97EF7FBC3A44B756ULL, "lowtown_male_alienage1_sta6162798.cl"},
	{0x97EFBEDED15E5F78ULL, "gen_im_trs_qun_lgt_boo_r1.uti"},
	{0x97EFE441A3C6D80AULL, "hlac_w_woundedcoast_-262138_0x.dds"},
	{0x97EFF869117990E6ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_180_c.anb"},
	{0x97F019870198172CULL, "vxw_woundedcoast_1lt_0_0.vlm"},
	{0x97F0BD9B966ED160ULL, "var_area.gda"},
	{0x97F109EEA2FB22B9ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.anb"},
	{0x97F13FE7DA87EABEULL, "6121637_m.fxe"},
	{0x97F149EB3CDEB940ULL, "hlac_w_bonepit_393218_0x.dds"},
	{0x97F20B424C5652B0ULL, "653884802.ogg"},
	{0x97F27B8D289CE2D8ULL, "sse_ss_aveline.gda"},
	{0x97F33320668DA486ULL, "6057099_m.fxe"},
	{0x97F346E833759D26ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.dura.ani"},
	{0x97F50E7E987AA450ULL, "6105898_f.fxe"},
	{0x97F5895D004CBF7CULL, "vxht_bartrand_n_5rf_0_0.vlm"},
	{0x97F59A3D432978F2ULL, "hlac_ht_0_hightown_n_-196609_0x.dds"},
	{0x97F5C53E1A39C3C4ULL, "6113274_m.fxe"},
	{0x97F63BBA678F84CBULL, "6119763_m.fxe"},
	{0x97F6666463CFA5DCULL, "6201462_player_home.cut"},
	{0x97F6AFD53D0247A3ULL, "ht_0_hightown_n_42_0_ll.mmh"},
	{0x97F6F5130D8FD7A0ULL, "dae000_bartrand6107537.cl"},
	{0x97F74BD920B7F7E1ULL, "vxlt_gallowstemplar_f_4tx_0_0.vlm"},
	{0x97F760969A97CC93ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st01a_0e.dds"},
	{0x97F77876502B251DULL, "dae000_flemeth6013276.cl"},
	{0x97F7BF60DA184912ULL, "hlac_w_sundermount_-131070_0z.dds"},
	{0x97F939A73CAB34CFULL, "6097339_m.fxe"},
	{0x97F94D0F8BF83CEBULL, "vxlt_gallowstemplar_n_437_0_0.vlm"},
	{0x97F95A27BF7111DEULL, "and121_karl.cnv"},
	{0x97FAA44F55C5EB80ULL, "118602507.ogg"},
	{0x97FABB42A3AA7BADULL, "dae000_arishok6078183.cl"},
	{0x97FAE867594250D5ULL, "91795288.ogg"},
	{0x97FB3B8B4ACD7C90ULL, "6174642_m.fxe"},
	{0x97FBB2DB85B868B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_beseech.gad"},
	{0x97FBED6C140D91E4ULL, "fen310_fenris6118544.cl"},
	{0x97FBEED8E0806F38ULL, "6028165_m.fxe"},
	{0x97FBF49AD45B7812ULL, "vxht_keep_n_73f_0_0.vlm"},
	{0x97FC37B687A88FE1ULL, "mag120_macha6080606.cl"},
	{0x97FC5ECA27744749ULL, "vxca_caverns2prc_4g7_0_0.vlm"},
	{0x97FCFB6911841F91ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_180_c.mmh"},
	{0x97FD10155CA8A548ULL, "hlac_dr_0_deeproads_5849_0x.dds"},
	{0x97FD3C27BF74C7EEULL, "and121_karl.bnk"},
	{0x97FD720F6ED42EB8ULL, "vxw_sundermount_2pg_0_0.vlm"},
	{0x97FD977801E4790BULL, "hlac_dr_0_deeproads_5845_10x.dds"},
	{0x97FE5A1E90E588C2ULL, "440065634.ogg"},
	{0x97FEAE029848D399ULL, "6088899_f.fxe"},
	{0x97FF0F1538E8F711ULL, "6170283_m.fxe"},
	{0x97FFC532646C323AULL, "6145125_m.fxe"},
	{0x98000452FD41300EULL, "vxlt_gallowscourtyard_77h_0_0.vlm"},
	{0x98002EE8337FFE97ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.dura.evt"},
	{0x98010F705B6899B9ULL, "6085419_m.fxe"},
	{0x98017C534324804DULL, "vxw_sundermount_220_0_0.vlm"},
	{0x98018A32DF56C229ULL, "397871734.ogg"},
	{0x9801BF639280AFDCULL, "202636227.ogg"},
	{0x98021630585DC92EULL, "294333242.ogg"},
	{0x980234A08430204FULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_ballistabolt01_0_d.dds"},
	{0x980288927DCAD61CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofl.mao"},
	{0x980294927DCAEABFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofl.mmh"},
	{0x98029A927DCAF561ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofl.msh"},
	{0x9802B5DB85BE8CE9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_beseech.evt"},
	{0x9802D47989FEA529ULL, "vxw_encounter2_camp_p0_0_1.vlm"},
	{0x9802FE6CE4FF9492ULL, "qcr000_varric6208246.cl"},
	{0x98034A9216F00D1AULL, "299914848.ogg"},
	{0x980438E07D547614ULL, "6037099_m.fxe"},
	{0x9804A68281184A17ULL, "vxw_bonepitdestroyed_a3_oz_0_0.vlm"},
	{0x980521D2A8792CCDULL, "401881832.ogg"},
	{0x98069BA2B450735FULL, "vxht_mansion2b_1gn_0_0.vlm"},
	{0x98072FE833861F20ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_imp_lp_c.dura.gad"},
	{0x98079A74AF3D9A6FULL, "6110724_m.fxe"},
	{0x9807C1C24907ABCCULL, "vxlt_gallowstemplar_5z1_0_0.vlm"},
	{0x980855D62093F0C5ULL, "vxlt_0_lowtown_docks_a3_n_d3l_0_0.vlm"},
	{0x9808794C0F319F0DULL, "6054227_f.fxe"},
	{0x9808B049C369A176ULL, "6148720_m.fxe"},
	{0x9808EB7BB9BF3B91ULL, "vxca_cavern1_haunted_ia3_0_0.vlm"},
	{0x98099B1F366624DFULL, "hlac_w_encounter2_camp_131071_0z.dds"},
	{0x9809A7FD4354931FULL, "vxlt_undercityserialkiller_1n8_0_0.vlm"},
	{0x9809F0CF74E6D3A0ULL, "693124296.ogg"},
	{0x980A55A0CA1B9A3BULL, "art\\vfx\\combat\\vfx_r_gn4_escape_act_c.anb"},
	{0x980A9AEE6F9A6071ULL, "ser221_gamlen6134997.cl"},
	{0x980AEDCD20555770ULL, "zz_da2.arl"},
	{0x980BE3EC60CA4738ULL, "vxw_sundermount_2ud_0_0.vlm"},
	{0x980BF27A6287E482ULL, "gen00fl_dog6170979.cl"},
	{0x980C94F3070AEF4FULL, "130866090.ogg"},
	{0x980CE4FA297CFBACULL, "art\\levels\\races\\proxy\\props\\_textures\\p_armoire_01_d.dds"},
	{0x980CEEE29AB9031AULL, "125569381.ogg"},
	{0x980CFD40E584482CULL, "vxw_woundedcoast_1rj_0_0.vlm"},
	{0x980E25C5223A5DE6ULL, "root_war_defense.dds"},
	{0x980E35CEA3589DBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_bicep_flex_right.evt"},
	{0x980F1E22EED13B84ULL, "vfx_vfx_p3.vsh.fxo"},
	{0x980F27D83D3BF9F4ULL, "dae000_arishok6075356.cl"},
	{0x980F466BD402BC48ULL, "vxlt_gallowstemplar_f_4yz_0_0.vlm"},
	{0x980F9DDB85C8EE58ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_beseech.ani"},
	{0x980FD134BCAB1545ULL, "mer000ip_hubert_store.ncs"},
	{0x980FDB85A9E14FBFULL, "1000144629.ogg"},
	{0x980FDC9AD8DF4E7FULL, "vxlt_gallowstemplar_f_3qv_0_0.vlm"},
	{0x98100E0F5717A990ULL, "ht_mansion2_818_ll.mmh"},
	{0x98101C487F333001ULL, "qcr220_orsino6061355.cl"},
	{0x9810AF867F2A1D29ULL, "vxht_brothel_n_2ny_0_1.vlm"},
	{0x9810B2AF2C7538A5ULL, "766054461.ogg"},
	{0x9810D4385C7FF9A5ULL, "gen00fl_bethany6101238.cl"},
	{0x9810DDDFAAA3A374ULL, "vxht_mansion_n_3fc_0_0.vlm"},
	{0x981122DEDAEF7213ULL, "6045609_f.fxe"},
	{0x98117514B4692315ULL, "ht_0_hightown_a3_33_0.rml"},
	{0x9811E20063B631B5ULL, "art\\vfx\\combat\\vfx_gib_cora.msh"},
	{0x9811E40063B6352BULL, "art\\vfx\\combat\\vfx_gib_cora.mmh"},
	{0x9811E5CEC35E9920ULL, "6053112_m.fxe"},
	{0x9811F8FA2E4A7E87ULL, "llac_ca_caverns1_mines_24427_0.dds"},
	{0x9812164A451AC5C1ULL, "hm_harlan.mor"},
	{0x98126D16E438F07CULL, "vxht_bartrand_n_5xd_0_0.vlm"},
	{0x9813A76D0EECD2CFULL, "vxlt_0_lowtown_n_e0p_0_0.vlm"},
	{0x9813F8CDE28FFB3CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose_v2.ani"},
	{0x98142ECEA35CFDFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_bicep_flex_right.gad"},
	{0x9814EA534260FE16ULL, "6104576_f.fxe"},
	{0x981528C0A98EEFA8ULL, "zz_followers_banter6159183.cl"},
	{0x9815337239C055D4ULL, "vxht_keep_3ii_0_0.vlm"},
	{0x9815504B3020C943ULL, "hlac_ht_0_hightown_a3_-1_0y.dds"},
	{0x981590F16C3B4154ULL, "hlac_lt_0_lowtown_docks_n_-65537_2z.dds"},
	{0x9815B33A8EFC0771ULL, "ca_caverns1_mines_mmap.dds"},
	{0x9815F2002E38199EULL, "sib131_slave_a6165254.cl"},
	{0x981612DD2FB0152DULL, "lgt310_mag_melson6152203.cl"},
	{0x98171EC1161AB188ULL, "vxht_bartrand_n_67b_0_0.vlm"},
	{0x9817947A4CAC7B05ULL, "vxlt_0_lowtown_glc_0_0.vlm"},
	{0x98179531E1371184ULL, "s_ballista_01.dds"},
	{0x9817C6EE5BD6AEF3ULL, "llac_lt_0_lowtown_docks_-196610_5.dds"},
	{0x9817D26D0B56170AULL, "251108672.ogg"},
	{0x981819C775CBE37AULL, "l_pro000ar_blightlands_fake.bnk"},
	{0x9818290B3508997CULL, "6151027_m.fxe"},
	{0x98183B30F43BBCDDULL, "vxw_bonepitdestroyed_a3_234_0_0.vlm"},
	{0x98186591F99CC061ULL, "vxlt_gallowscourtyard_n_74a_0_0.vlm"},
	{0x9818659534B3A0B9ULL, "vxca_caverns2_8o2_0_0.vlm"},
	{0x9818F91697A88879ULL, "w_woundedcoast_178_0.msh"},
	{0x9819031697A89967ULL, "w_woundedcoast_178_0.mmh"},
	{0x9819071697A89FB4ULL, "w_woundedcoast_178_0.mao"},
	{0x98197A12B64C26B6ULL, "vxlt_0_lowtown_n_cfi_0_0.vlm"},
	{0x98197FB19E49855BULL, "lgt250_way_dalish_conflict6147116.cl"},
	{0x98199C1C155EF919ULL, "378634510.ogg"},
	{0x9819B5476B0FAD5CULL, "vxlt_0_lowtown_n_eqh_0_0.vlm"},
	{0x9819DE6C39A7656FULL, "vxca_caverns2prc_488_0_0.vlm"},
	{0x981A0F55482696E0ULL, "gen00fl_isabela6085685.cl"},
	{0x981A4D712F15D81DULL, "6042021_m.fxe"},
	{0x981A97AED1E14CBBULL, "498518419.ogg"},
	{0x981B05CDE296307EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose_v2.gad"},
	{0x981B2DCEA3631AFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_bicep_flex_right.ani"},
	{0x981C5E957C26416DULL, "vxht_mansion2b_23w_0_0.vlm"},
	{0x981D2CABC7552C02ULL, "6177212_m.fxe"},
	{0x981DB3B92B7A312AULL, "vrc221_varric6043971.cl"},
	{0x981F7314824A05FAULL, "6147879_m.fxe"},
	{0x981FFF82D9020633ULL, "llac_w_bonepit_131073_0.dds"},
	{0x9820AEE5EDCB10A7ULL, "dae000_flemeth6000762.cl"},
	{0x9820C234BCB96E9CULL, "mer000ip_hubert_store.utp"},
	{0x9820C72671C63CBDULL, "vxlt_gallowstemplar_f_34s_0_0.vlm"},
	{0x9820DB70D284D60CULL, "hm_bloodmage1.mor"},
	{0x9820E604D87AA714ULL, "vxht_brothel_n_6cb_0_0.vlm"},
	{0x9821B0331F76619DULL, "172905313.ogg"},
	{0x9821E0CDE29C0F9DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddopentoclose_v2.evt"},
	{0x9821EE47BAF16C48ULL, "vxlt_gallowscourtyard_n_74d_0_0.vlm"},
	{0x9821EEF06D0F37C3ULL, "dae150_chandan6148640.cl"},
	{0x98220E8F7293AA31ULL, "vxlt_0_lowtown_a2_ewx_0_0.vlm"},
	{0x9822432FA804E3DDULL, "hlac_lt_0_lowtown_a2_-196608_3x.dds"},
	{0x98224A0307039239ULL, "lgt310_mag_melson6152122.cl"},
	{0x9822D6460DE36329ULL, "192033032.ogg"},
	{0x98231884119AD342ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_hole.phy"},
	{0x98235265DE8E360DULL, "6152320_m.fxe"},
	{0x9823777932A14AC4ULL, "art\\vfx\\combat\\vfx_w_gn2_taunt_c.impa.ani"},
	{0x9824324E16659D83ULL, "ser191_emeric6120515.cl"},
	{0x982437DE456F06B8ULL, "art\\levels\\races\\design\\_textures\\prp_bigtreasure_n.dds"},
	{0x9824781795C0AF91ULL, "fx_icecrystals.mao"},
	{0x98249E7348DBFEAEULL, "vxht_0_hightown_a2_9zp_0_0.vlm"},
	{0x982610CD9161D3C0ULL, "vxlt_0_lowtown_a2_hdi_0_0.vlm"},
	{0x982665E5599C2D39ULL, "258943341.ogg"},
	{0x982672F09525B0D8ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic_1.mao"},
	{0x9826A6CC34E55D21ULL, "hlac_dr_primevaltomb_9621_7x.dds"},
	{0x9826CE1C7A4DECB9ULL, "vxht_0_hightown_be5_0_0.vlm"},
	{0x9826E3FC0DE40AC2ULL, "vxlt_shop_j8_0_0.vlm"},
	{0x982721B2AE7FED51ULL, "vxdr_primevaltomb_75u_0_0.vlm"},
	{0x98276E4C5F4B97B8ULL, "hlac_w_sundermount_327676_0z.dds"},
	{0x982827245D05C5DCULL, "vxlt_0_lowtown_n_eji_0_0.vlm"},
	{0x98286B6F9121CB5DULL, "zz_mer_debug6180872.cl"},
	{0x98287497C866435FULL, "llac_lt_0_lowtown_n_-196608_4.dds"},
	{0x9828AF214BC3DADAULL, "vxlt_0_lowtown_a2_hgz_0_0.vlm"},
	{0x9828E79DBC10794EULL, "ef_qun_crazed_warrior.mor"},
	{0x982906E37DCC8A95ULL, "798797570.ogg"},
	{0x9829B6B3DC258634ULL, "6117186_m.fxe"},
	{0x982A2CFC6275AFDAULL, "one140cr_charade6166008.cl"},
	{0x982A376FBD5D7E90ULL, "lit_whispy.mao"},
	{0x982A48CB1011EDF3ULL, "887148296.ogg"},
	{0x982A52802AFC7CD6ULL, "mag370_meredith6046958.cl"},
	{0x982A52D42E268B61ULL, "vxlt_0_lowtown_docks_a3_n_d45_0_0.vlm"},
	{0x982A6C9990BFB8A2ULL, "vxlt_foundry_n_z4_0_0.vlm"},
	{0x982A747932A76426ULL, "art\\vfx\\combat\\vfx_w_gn2_taunt_c.impa.gad"},
	{0x982AE052F5AE94E6ULL, "gen00fl_isabela6101077.cl"},
	{0x982B14C5F8BD6560ULL, "zz_followers_banter6158346.cl"},
	{0x982B4610CDD402E9ULL, "510772494.ogg"},
	{0x982D0D90779E7FADULL, "zz_vault_debug6203375.cl"},
	{0x982DD0ABCA528400ULL, "vxw_bonepitmining_a2_lr_0_0.vlm"},
	{0x982E3BE236F3E793ULL, "mnd_pro_03.mao"},
	{0x982E50CA27DDCA44ULL, "mag200_feynriel6023951.cl"},
	{0x982E7B2DE146D245ULL, "6121967_f.fxe"},
	{0x982F02B60EC2CB87ULL, "vxlt_warehouse_8l_0_0.vlm"},
	{0x982F9C31E14BAFD2ULL, "s_ballista_01.mao"},
	{0x982FD91CACF3F255ULL, "hlac_ca_caverns1_mines_24422_3x.dds"},
	{0x983162ABB2940F13ULL, "dae000ip_crowd_mf2_talk18.utp"},
	{0x983162EFA407120DULL, "ser191_emeric6108160.cl"},
	{0x983164C6D2344304ULL, "807270460.ogg"},
	{0x98316F7932AD7A05ULL, "art\\vfx\\combat\\vfx_w_gn2_taunt_c.impa.evt"},
	{0x9831B33F2EDF42E0ULL, "vxlt_darktown_6jv_0_0.vlm"},
	{0x9831BEFDC3AE207EULL, "ser221_party_comments6166200.cl"},
	{0x9831C7830D243B15ULL, "character_character_p48.psh.fxo.dx11.assembly"},
	{0x98328CD349E43E19ULL, "ave200_aveline_bar6116428.cl"},
	{0x98337A9051D9314CULL, "6128377_m.fxe"},
	{0x9833920248170C03ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoorcobble_d.dds"},
	{0x9833D72494410416ULL, "6134868_m.fxe"},
	{0x98344431B3947DB7ULL, "gen00fl_anders6111095.cl"},
	{0x98344F80372D81B1ULL, "611442263.ogg"},
	{0x983461EA57F15596ULL, "vxlt_warehouse_11g_0_0.vlm"},
	{0x983495561A5C3A4EULL, "vxlt_gallowstemplar_376_0_0.vlm"},
	{0x9834C137A70D610FULL, "gen00fl_varric6077818.cl"},
	{0x98354059E49897A4ULL, "893568092.ogg"},
	{0x98355F0A0FAB18ECULL, "vxlt_0_lowtown_a2_n_c_gji_0_0.vlm"},
	{0x9835F7A1A0DC606DULL, "zz_followers_banter6158217.cl"},
	{0x98362A04E02CEB1EULL, "hlac_dr_primevalentrance_1726_11x.dds"},
	{0x9836E8B08DB7BB16ULL, "6069624_f.fxe"},
	{0x9837CB0EE6B710C6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25.phy"},
	{0x98385648C5C2D5FAULL, "follower_banter6155961.cl"},
	{0x98389CD174BAF2F9ULL, "gen00fl_anders6142552.cl"},
	{0x98397AB457EB93EFULL, "zz_qcr_debug6052124.cl"},
	{0x9839A083C5F927ACULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunaritent_dp.dds"},
	{0x9839EFF4BEF99DA0ULL, "6025349_f.fxe"},
	{0x9839F15154863316ULL, "6111480_f.fxe"},
	{0x983A0F480529409DULL, "w_encounter1_23_0.rml"},
	{0x983B498DA20F35BEULL, "vxca_cavern1_haunted_jfn_0_0.vlm"},
	{0x983C0898670E1674ULL, "vxlt_0_lowtown_exx_0_0.vlm"},
	{0x983C08ECA2AFF472ULL, "6125808_m.fxe"},
	{0x983C1F571289AE81ULL, "vxlt_0_lowtown_n_eq9_0_0.vlm"},
	{0x983C4288BC489C80ULL, "vxw_bonepitmining_a2_ks_0_0.vlm"},
	{0x983C4C5BF468235DULL, "vxlt_0_lowtown_f1i_0_0.vlm"},
	{0x983CDBCFBC337706ULL, "710803946.ogg"},
	{0x983D33C775EB2B18ULL, "l_pro000ar_blightlands_fake.wbd"},
	{0x983DC3F7E164E622ULL, "llac_w_sundermount_65536_0.dds"},
	{0x983E0F1CA970FFDBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mc.mmh"},
	{0x983E1B1CA9711478ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mc.mao"},
	{0x983E2D1CA97132C5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12mc.msh"},
	{0x983EF8185B5F5196ULL, "938864747.ogg"},
	{0x983F4BABD04442EEULL, "876431236.ogg"},
	{0x983FFFB5451B3A70ULL, "vxw_encounter2_camp_jv_0_0.vlm"},
	{0x984044F19A726729ULL, "zz_bet_debug6113390.cl"},
	{0x98407BA76CFF048CULL, "6108255_f.fxe"},
	{0x98408601ECB588E1ULL, "gen_im_npc_arm_hum_glv_tmpc.uti"},
	{0x9840B11B7418A0FDULL, "vxlt_0_lowtown_a2_n_c_hli_0_0.vlm"},
	{0x9840C43AE28D8931ULL, "620485752.ogg"},
	{0x9840CF683305EE8EULL, "dae000ip_crowd_mf3_talk16.utp"},
	{0x984125F5AA0A0921ULL, "zz_dae_debug6029898.cl"},
	{0x9841B368DD2A443BULL, "93852730.ogg"},
	{0x9842D3C749E76E16ULL, "vxlt_warehouse_14h_0_0.vlm"},
	{0x9842FB3F7FEC9336ULL, "lt_0_lowtown_docks_a3_n_222_0.msh"},
	{0x9843053F7FECA434ULL, "lt_0_lowtown_docks_a3_n_222_0.mmh"},
	{0x9843093F7FECAB47ULL, "lt_0_lowtown_docks_a3_n_222_0.mao"},
	{0x98432D94ECEFAEBCULL, "vxht_store_bs_0_0.vlm"},
	{0x98439EE843309F07ULL, "6098716_m.fxe"},
	{0x9843BCBE8F9BFFB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\scared_loop.ani"},
	{0x9843F0972B400F9DULL, "203679006.ogg"},
	{0x984497640353EF79ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_diagonal.msh"},
	{0x9844A16403540067ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_diagonal.mmh"},
	{0x9844A564035406B4ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_diagonal.mao"},
	{0x9844C0D582343549ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_3z.dds"},
	{0x9844EC1697CDAC9CULL, "w_woundedcoast_178_0.phy"},
	{0x9844FF60A4897EC5ULL, "hf_ser_marlein.mor"},
	{0x98451E6DCB2E7AD4ULL, "vxdr_0_deeproads_9iu_0_1.vlm"},
	{0x9845202F31D1E588ULL, "6106892_m.fxe"},
	{0x984539E410887FA9ULL, "vrc321ip_spooky7_obj_10.utp"},
	{0x9845614AA05380D6ULL, "vxht_mansion2b_1aq_0_0.vlm"},
	{0x9845B47157298336ULL, "984482553.ogg"},
	{0x9846228EF819098BULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_003a_s.dds"},
	{0x98462B20EDE0B323ULL, "6079017_f.fxe"},
	{0x98466A59A03FD437ULL, "hlac_lt_0_lowtown_docks_65535_2y.dds"},
	{0x98469928A1CAAC76ULL, "clf260_harley6069655.cl"},
	{0x9846AB55247E9100ULL, "hlac_lt_0_lowtown_a2_65535_4y.dds"},
	{0x9846BC9B84D8F861ULL, "6160758_m.fxe"},
	{0x984786D2584DA5E9ULL, "6040148_m.fxe"},
	{0x9847B4997B13B128ULL, "hlac_lt_0_lowtown_docks_a3_65536_0z.dds"},
	{0x9847E05B0BB2D4A5ULL, "llac_w_bonepitdestroyed_a3_393217_0.dds"},
	{0x9847F38411B9B0F6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_hole.mao"},
	{0x9847FF8411B9C495ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_hole.mmh"},
	{0x9848018411B9C80BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_hole.msh"},
	{0x98486D81DAF08BD3ULL, "11345207.ogg"},
	{0x9848C9EDEEF81417ULL, "dae000_flemeth6001332.cl"},
	{0x9849B4D7750EECAEULL, "vxlt_backalley_n_1k3_0_0.vlm"},
	{0x984A1433CD1805C0ULL, "vxht_keep_n_57v_0_0.vlm"},
	{0x984A7AC994A60CF2ULL, "6122908_m.fxe"},
	{0x984A97BE8FA1DFFBULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\scared_loop.gad"},
	{0x984A98EF29F15E02ULL, "vxlt_gallowstemplar_n_3qo_0_0.vlm"},
	{0x984AA6B13379ECD0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblepile.mao"},
	{0x984AAAB13379F3E3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblepile.mmh"},
	{0x984AB8B1337A0B3DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblepile.msh"},
	{0x984B87B16ADABFD9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_coffer.msh"},
	{0x984B91B16ADAD0C7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_coffer.mmh"},
	{0x984B95B16ADAD714ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_coffer.mao"},
	{0x984D869AB82DC164ULL, "fxa_pois_split.dds"},
	{0x984E1115B203C987ULL, "gen00fl_anders6112977.cl"},
	{0x984E347B9B7936AEULL, "341138567.ogg"},
	{0x984F3B0239C50AFBULL, "282902033.ogg"},
	{0x984F436ED7212D9BULL, "follower_banter6136738.cl"},
	{0x984F868FAA71249EULL, "vxlt_gallowstemplar_5un_0_0.vlm"},
	{0x985070B2AC59D130ULL, "6127123_m.fxe"},
	{0x985070DEE7FDBE54ULL, "lgt200im_eom_book.uti"},
	{0x98507CD8D579EEE8ULL, "355132920.ogg"},
	{0x9850A3236395ECBDULL, "143628278.ogg"},
	{0x9850C68525EA8596ULL, "llac_lt_0_lowtown_docks_n_131072_1.dds"},
	{0x9850CBAE37469B02ULL, "6054883_m.fxe"},
	{0x98511D9D758D1A86ULL, "253969899.ogg"},
	{0x98518F214350AC44ULL, "dae000_flemeth6002516.cl"},
	{0x9851B333EC15F30BULL, "750282973.ogg"},
	{0x9852D49ABA27D5DAULL, "184381067.ogg"},
	{0x9853D32792499956ULL, "vxht_mansion2b_1dr_0_0.vlm"},
	{0x98544D43332E7C0CULL, "lgt000ar_gang_hideout.ncs"},
	{0x9854646ABB78A7FBULL, "vxlt_warehouse_n_ro_0_0.vlm"},
	{0x98546B120F9A30B6ULL, "vxlt_gallowstemplar_5c0_0_0.vlm"},
	{0x9855208F28498835ULL, "dae010_chase_guard.utc"},
	{0x98557AAD0A1DB530ULL, "dae000_mother6136927.cl"},
	{0x9855CC70E433654BULL, "vxlt_undercityencounter_1qq_0_0.vlm"},
	{0x985671B366BD2ED7ULL, "vxht_0_hightown_an5_0_0.vlm"},
	{0x98569D59461FDDC8ULL, "character_character_p56.psh.fxo.dx11.assembly"},
	{0x98569D83BA37CBA4ULL, "vxdr_0_deeproads_eqh_0_0.vlm"},
	{0x98569E09A873AE29ULL, "483402853.ogg"},
	{0x9856B49FE38B747AULL, "hlac_lt_0_lowtown_docks_a3_n_-1_2x.dds"},
	{0x9856C66EF7B10B92ULL, "vxca_caverns2_7w6_0_0.vlm"},
	{0x9856E6F887EA5F93ULL, "vxlt_0_lowtown_docks_n_cvt_0_1.vlm"},
	{0x985706B5C1C708D4ULL, "vxw_bonepit_yw_0_0.vlm"},
	{0x98574805B0559202ULL, "618045992.ogg"},
	{0x98575A3D26990E37ULL, "vxht_0_hightown_aq4_0_0.vlm"},
	{0x985786E48952473BULL, "6177950_m.fxe"},
	{0x9857B82959B0A167ULL, "vxht_brothel_n_5dw_0_0.vlm"},
	{0x9857D70FB1AD563DULL, "tbe100st_crafted_a3.utm"},
	{0x98580BC65120DA28ULL, "vxlt_gallowscourtyard_695_0_0.vlm"},
	{0x985829F3C179D0C5ULL, "zz_and_debug6134074.cl"},
	{0x9858AFED65910623ULL, "vxdr_0_deeproads_f9p_0_0.vlm"},
	{0x9858CAB70F1F3545ULL, "em_faj.mor"},
	{0x9858CED9F6C2A3D2ULL, "mcr370_templar26087650.cl"},
	{0x9858D4EB38A95094ULL, "tbe100im_treasure_map_1.uti"},
	{0x9858DDA833A2B194ULL, "gen00fl_isabela6101556.cl"},
	{0x985940D896F76DC1ULL, "ico_armor_medium_3.dds"},
	{0x9859C08E8754A2B9ULL, "mag100_cullen6140950.cl"},
	{0x985B07EA158AAD07ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_act_c.anb"},
	{0x985B0D2423F3D335ULL, "vxlt_0_lowtown_a2_n_c_bxw_0_0.vlm"},
	{0x985B1B478953C844ULL, "553131558.ogg"},
	{0x985B636BB83A5FE9ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev1_b.ani"},
	{0x985BABB133886808ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_3rubblepile.phy"},
	{0x985BF9CE703955CAULL, "hlac_ht_0_hightown_a3_-131074_2y.dds"},
	{0x985C09DE83C1B457ULL, "llac_lt_0_lowtown_docks_n_-196610_2.dds"},
	{0x985C47D8BD23FCDCULL, "816055920.ogg"},
	{0x985C9DB13F69E156ULL, "22153519.ogg"},
	{0x985D830938A9BDA8ULL, "19757621.ogg"},
	{0x985DC25C5B8410E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_reaction.ani"},
	{0x985DC7C53A7B360DULL, "vxht_0_hightown_a2_c5p_0_0.vlm"},
	{0x985DE67228DC8F6CULL, "hlac_lt_gallowstemplar_2382_0y.dds"},
	{0x985E1C4DE9C135AFULL, "zz_o_11_gallows_nt_ct2.cut"},
	{0x985E2ABB3D7C36BDULL, "502443599.ogg"},
	{0x985E51636F0FC577ULL, "173435343.ogg"},
	{0x985EB6BB74D17696ULL, "6162248_f.fxe"},
	{0x985F0166ED572D1BULL, "vxca_caverns1_i9x_0_0.vlm"},
	{0x985F0329FE547AF2ULL, "qun131_party_comment6117300.cl"},
	{0x985F074FD2CE2D95ULL, "6115596_m.fxe"},
	{0x985F12114CCC791FULL, "dae000_mother6099175.cl"},
	{0x985F40C40C32DA75ULL, "follower_banter6136466.cl"},
	{0x985F480FF49CE09DULL, "vxlt_0_lowtown_docks_a3_cod_0_0.vlm"},
	{0x985FD92211E26D68ULL, "vxca_cavern1_haunted_iin_0_0.vlm"},
	{0x9860227932B9B23CULL, "seekersbulwark.dds"},
	{0x98603569022A0A52ULL, "6068432_f.fxe"},
	{0x98605200DBBB0BA4ULL, "323091332.ogg"},
	{0x986062A30A37531EULL, "gen00fl_fenris6126351.cl"},
	{0x98608AB07582601EULL, "ico_dagger_3.dds"},
	{0x9860BF107D1B65D1ULL, "6106656_f.fxe"},
	{0x986136CC2B59CE8CULL, "435149985.ogg"},
	{0x986151AA6F674998ULL, "vxw_bonepitmining_a2_1lq_0_0.vlm"},
	{0x986189605D5A2E4AULL, "6079279_m.fxe"},
	{0x9861AE3F7CC0444EULL, "vxlt_0_lowtown_n_d2n_0_0.vlm"},
	{0x9861D89872A48AD1ULL, "vxlt_0_lowtown_n_gc5_0_0.vlm"},
	{0x9861FFEF844BAFEFULL, "lla_ht_0_hightown_65534_0.dds"},
	{0x98623355F144042FULL, "vxlt_gallowstemplar_n_3x5_0_0.vlm"},
	{0x986235C845E4B363ULL, "540304797.ogg"},
	{0x98627E6BB840AC6BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev1_b.gad"},
	{0x9862D647AD6EC07BULL, "vxlt_warehouse_n_yn_0_0.vlm"},
	{0x986300601477F4BBULL, "911897563.ogg"},
	{0x98634512C634D142ULL, "295062820.ogg"},
	{0x9863902AC3486DE9ULL, "mcr371_angry_mage_m6087725.cl"},
	{0x986413B03503FADCULL, "hlac_lt_0_lowtown_docks_a3_1_0x.dds"},
	{0x9864B6D23AC58D85ULL, "vxht_chantry_n_36f_0_0.vlm"},
	{0x98658CC3F903D8CAULL, "mag170_thrask6016947.cl"},
	{0x98658D9AB8425EB2ULL, "fxa_pois_split.mao"},
	{0x9865CC1A188F26B7ULL, "vxht_0_hightown_at7_0_0.vlm"},
	{0x9866564E94C4370EULL, "175965188.ogg"},
	{0x9866902509CE5CCFULL, "dae000cr_grand_cleric6147899.cl"},
	{0x9866B346A62FB102ULL, "hlac_lt_0_lowtown_docks_-65536_0z.dds"},
	{0x9866CF2DFF806E7DULL, "gen00fl_carver6102549.cl"},
	{0x98670C318D7FE922ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_1y.dds"},
	{0x98674F58A1072ABFULL, "rdr110dg_party_comments.bnk"},
	{0x986770F0803979B9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_001a_0s.dds"},
	{0x98678A94B77CA10CULL, "pro000_refugee_male6052103.cl"},
	{0x986823D63274B76BULL, "art\\vfx\\combat\\vfx_d_single_att_c.impa.ani"},
	{0x9868327325D3EEE5ULL, "180789099.ogg"},
	{0x986860F92BA05A3DULL, "6151624_m.fxe"},
	{0x986890D9B9B8F69DULL, "llac_lt_0_lowtown_0_6.dds"},
	{0x98689C001D303E77ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_0x.dds"},
	{0x9869242A660C96F3ULL, "vxlt_0_lowtown_docks_b63_0_0.vlm"},
	{0x98697D8DF873D911ULL, "art\\characters\\creatures\\stonegolem\\animation\\p.ani"},
	{0x98698047A8097451ULL, "lt_gallowscourtyard_37_0_ll.mmh"},
	{0x9869B9871DBCC3C1ULL, "w_bonepit_204_0.phy"},
	{0x986A3158A1092DC7ULL, "rdr110dg_party_comments.cnv"},
	{0x986A628FB05D82C1ULL, "mag371_elsa_st.stg"},
	{0x986ADA5C5B8EC3DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_reaction.evt"},
	{0x986ADC5714134458ULL, "vxlt_0_lowtown_docks_a3_n_ctg_0_0.vlm"},
	{0x986B296A505E2821ULL, "qun120cr_javaris6043991.cl"},
	{0x986B93D1DB474454ULL, "gen00fl_aveline6139679.cl"},
	{0x986BACAB18C7F9B8ULL, "115581757.ogg"},
	{0x986E189C31E7FF04ULL, "zz_dre_debug6168296.cl"},
	{0x986F00A3522273F3ULL, "lt_0_lowtown_docks_150_0.phy"},
	{0x986F3FB7A86D400EULL, "vxw_bonepitdestroyed_a3_17h_0_0.vlm"},
	{0x9870888DF87A0AB3ULL, "art\\characters\\creatures\\stonegolem\\animation\\p.gad"},
	{0x9870B6E17635CA56ULL, "vxht_mansion2b_1rt_0_0.vlm"},
	{0x9870EE3506702EC8ULL, "2643092.ogg"},
	{0x9871CE2269709BBFULL, "689967620.ogg"},
	{0x9871D95C5B94E0CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throat_slit_reaction.gad"},
	{0x9871E004D3B668C1ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_007a_d.dds"},
	{0x987281DDA3F3F826ULL, "gen00fl_anders6121640.cl"},
	{0x9872A1ED1FE2D429ULL, "365062562.ogg"},
	{0x9872C1D2FB2D1AE1ULL, "6126392_m.fxe"},
	{0x9872CF3615606C4EULL, "lt_0_lowtown_docks_n_142_0.mmh"},
	{0x9872DB36156080E5ULL, "lt_0_lowtown_docks_n_142_0.mao"},
	{0x9872E9361560983CULL, "lt_0_lowtown_docks_n_142_0.msh"},
	{0x98730FB79D406018ULL, "qun220cr_javaris6071718.cl"},
	{0x9873813D9E23FCA4ULL, "vxdr_0_deeproads_egn_0_0.vlm"},
	{0x9873A7C620A95E68ULL, "follower_banter6136339.cl"},
	{0x9873BF96E14EB7FDULL, "6118735_m.fxe"},
	{0x987457064F0E8EC7ULL, "co_bl_dot.mao"},
	{0x9874C6190368F734ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_enter.ani"},
	{0x98750BD6327F18E2ULL, "art\\vfx\\combat\\vfx_d_single_att_c.impa.evt"},
	{0x9875E3889D50A184ULL, "zz_followers_banter6158823.cl"},
	{0x987609008C9A87DAULL, "651286886.ogg"},
	{0x987628518EAC00A3ULL, "ss_player_female.wbd"},
	{0x9876442B157D8794ULL, "6103925_m.fxe"},
	{0x98765E040B67E04EULL, "caustic_flip_a.dds"},
	{0x9877139D313F1DE3ULL, "llac_lt_0_lowtown_a2_n_c_131070_1.dds"},
	{0x9877BE57DD53219BULL, "581254071.ogg"},
	{0x987839D670C23BAFULL, "6143966_m.fxe"},
	{0x98787AB16B0197FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1co_coffer.phy"},
	{0x987898C4C71FF6F5ULL, "6085911_meredith_righteous.cut"},
	{0x9878C090FE24327CULL, "6178462_m.fxe"},
	{0x987908482B770711ULL, "vxht_keep_5xu_0_0.vlm"},
	{0x98799088234D8E81ULL, "vxht_0_hightown_a2_c3w_0_1.vlm"},
	{0x987A8C212CF11ADAULL, "ave310_brennan6137248.cl"},
	{0x987AB17E0098CBC2ULL, "vxht_0_hightown_a3_bxd_0_0.vlm"},
	{0x987AEECAD14A41B5ULL, "269846811.ogg"},
	{0x987B21D728F4A067ULL, "hlac_lt_0_lowtown_a2_1_5z.dds"},
	{0x987B26AE041B20D4ULL, "vxlt_0_lowtown_n_f1s_0_0.vlm"},
	{0x987B4D0F178BA16CULL, "zz_followers_banter6157912.cl"},
	{0x987B685BDAFEDCF2ULL, "art\\characters\\creatures\\mubari\\animation\\eat2pause.gad"},
	{0x987B79E987C38508ULL, "308916209.ogg"},
	{0x987BC319036F1096ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_enter.gad"},
	{0x987C0AD6328535F5ULL, "art\\vfx\\combat\\vfx_d_single_att_c.impa.gad"},
	{0x987CBCDBF5CE9E68ULL, "vxca_cavern1_haunted_icl_0_0.vlm"},
	{0x987D28B4FB88A251ULL, "110581915.ogg"},
	{0x987D315A2835AB8CULL, "vxlt_gallowscourtyard_n_4wf_0_0.vlm"},
	{0x987D716987C5DD4CULL, "6147723_m.fxe"},
	{0x987E19D940555572ULL, "456709540.ogg"},
	{0x987E246946C851BBULL, "art\\characters\\creatures\\darkspawnogre\\cn_hed_ogra.mmh"},
	{0x987E306946C86658ULL, "art\\characters\\creatures\\darkspawnogre\\cn_hed_ogra.mao"},
	{0x987E34FA4BB816F5ULL, "root_rog_assassin.dds"},
	{0x987E426946C884A5ULL, "art\\characters\\creatures\\darkspawnogre\\cn_hed_ogra.msh"},
	{0x987EA2A6ACE52F86ULL, "6076379_m.fxe"},
	{0x987EB8C00884FE0DULL, "vxlt_backalley_g_1u2_0_1.vlm"},
	{0x987ED4FB77C24947ULL, "292429298.ogg"},
	{0x987EF16B15BF1526ULL, "559086252.ogg"},
	{0x987F170FD530352FULL, "vxlt_gallowstemplar_n_3r7_0_0.vlm"},
	{0x987F28BE682BE2D6ULL, "vxht_mansion2b_1uu_0_0.vlm"},
	{0x987FDEB17B1B0D25ULL, "lt_undercityencounter_probe_1m6.dds"},
	{0x98800E2C98A85932ULL, "902695801.ogg"},
	{0x98807B14646C49A4ULL, "6115455_f.fxe"},
	{0x988096854ED2AA56ULL, "269842500.ogg"},
	{0x9880B1F1925F5DB8ULL, "gen_im_wep_war_one_act3_08.uti"},
	{0x9880BA416E0AEED6ULL, "isa211_isabela6202290.cl"},
	{0x9880FEEBB94D6E21ULL, "vxdr_0_deeproads_cum_0_2.vlm"},
	{0x98817DDD5160F979ULL, "976730571.ogg"},
	{0x9881B22FFBB69769ULL, "hlac_ht_bartrand_n_3371_0x.dds"},
	{0x98824612E3C073C9ULL, "vxlt_0_lowtown_a3_n_c_e1j_0_0.vlm"},
	{0x9882580B7969055AULL, "vxht_0_hightown_a2_b9r_0_0.vlm"},
	{0x98825AB6732E2D34ULL, "416306240.ogg"},
	{0x98826B5BDB050029ULL, "art\\characters\\creatures\\mubari\\animation\\eat2pause.evt"},
	{0x9882ABB7EC6873BAULL, "6067666_f.fxe"},
	{0x9882CE9056CD8A14ULL, "647488056.ogg"},
	{0x9882DE1903755DF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drinking_enter.evt"},
	{0x98835C5B673A9AE5ULL, "6189401_m.fxe"},
	{0x9883A6C19A340412ULL, "dr_slavers_3207.anb"},
	{0x9883E6C92918926EULL, "423315457.ogg"},
	{0x988404EA15ADAD08ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_act_c.mmh"},
	{0x9884061E4964CFDBULL, "148937289.ogg"},
	{0x98846474570016D4ULL, "wm_ico_chantry_roll.dds"},
	{0x9884A9E4535C75D9ULL, "gen_im_trs_apo_junk2_04.uti"},
	{0x98854B233AB56521ULL, "ser270_emeric6111272.cl"},
	{0x9885994F35169232ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_disap_p.impa.gad"},
	{0x9885F56843B48E76ULL, "one140cr_charade6128510.cl"},
	{0x988608CA5897CE43ULL, "6075528_m.fxe"},
	{0x98872C6C3C496437ULL, "llac_lt_0_lowtown_docks_a3_-196610_4.dds"},
	{0x988752094A93880BULL, "art\\vfx\\combat\\vfx_w_rev_devour_imp_c.mmh"},
	{0x988796E9F8437579ULL, "507636978.ogg"},
	{0x9887A31C0BE6D011ULL, "113964470.ogg"},
	{0x9887C010F7FF7558ULL, "vxlt_0_lowtown_docks_a3_n_cze_0_0.vlm"},
	{0x9887F9707EC7E0ECULL, "zz_prmpt_debug6200927.cl"},
	{0x988896E8F00E5066ULL, "6164739_m.fxe"},
	{0x9889235AF28EE442ULL, "vxht_0_hightown_a3_bwe_0_0.vlm"},
	{0x9889988AF6113954ULL, "vxlt_0_lowtown_n_f4p_0_0.vlm"},
	{0x988A3603D370FAAFULL, "687801913.ogg"},
	{0x988A7A45D1E4DC1BULL, "mag311_keran6041392.cl"},
	{0x988B2A01D3E1558DULL, "hlac_lt_0_lowtown_docks_65535_3x.dds"},
	{0x988B719026A3612DULL, "gen00fl_aveline6132664.cl"},
	{0x988C0C28518B6BA0ULL, "gen00fl_bethany6097844.cl"},
	{0x988C9C4F351CB669ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_disap_p.impa.evt"},
	{0x988CBF5346559B4FULL, "sse_ss_bethany.gda"},
	{0x988D3ABDB2A3A552ULL, "lt_0_lowtown_a2_n_c_33_0_lc.mmh"},
	{0x988DC50BEE1B78EFULL, "hlac_lt_undercitydungeon_-1_6x.dds"},
	{0x988DF0D8B73CCA8AULL, "vxht_mansion2_22o_0_0.vlm"},
	{0x988E2161E9ABB26EULL, "w_woundedcoast_188_0.mmh"},
	{0x988E2D61E9ABC605ULL, "w_woundedcoast_188_0.mao"},
	{0x988E3B61E9ABDE5CULL, "w_woundedcoast_188_0.msh"},
	{0x988E6B46C70E0E88ULL, "298212439.ogg"},
	{0x988EE2E466BBB4ADULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_stump02_d.dds"},
	{0x988EFDA41DE863AFULL, "137535991.ogg"},
	{0x988F28215167BD4FULL, "6110517_m.fxe"},
	{0x988F4619A3F72182ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\mov_p_injured.gad"},
	{0x988F47E6BB372D73ULL, "828456147.ogg"},
	{0x988F7CB20ECA1D00ULL, "6170259_m.fxe"},
	{0x988FF1CC1F6BBB78ULL, "6165389_m.fxe"},
	{0x98903C6D3841A39CULL, "vxdr_primevaltomb_8pz_0_2.vlm"},
	{0x989043A24BD96E89ULL, "vxlt_hangedman_n_m5_0_1.vlm"},
	{0x9890535BDB111498ULL, "art\\characters\\creatures\\mubari\\animation\\eat2pause.ani"},
	{0x9890AB4AE1671D29ULL, "746979825.ogg"},
	{0x9890E3C811C596E4ULL, "813513542.ogg"},
	{0x989105EE940A28E4ULL, "vxw_bonepitdestroyed_a3_ll_0_0.vlm"},
	{0x98915575AC40BFF3ULL, "dae000_gamlen6131359.cl"},
	{0x98915CC18E896C69ULL, "819470567.ogg"},
	{0x98918B645EF2AF5BULL, "hlac_lt_gallowstemplar_f_2382_0y.dds"},
	{0x9891B79086762495ULL, "vxdr_primevalentrance_2zi_0_0.vlm"},
	{0x98921C5F431BE9A1ULL, "vxdr_primevalentrance_179_0_0.vlm"},
	{0x98923D04261B8DD5ULL, "vxlt_0_lowtown_n_by2_0_0.vlm"},
	{0x98925D6E01691901ULL, "620531015.ogg"},
	{0x98927753957B9AF9ULL, "334647189.ogg"},
	{0x9893032A9FF2DF63ULL, "art\\characters\\playercharacter\\dwarfmale\\hlm_lgtb.mao"},
	{0x98946C0B9CAA169AULL, "art\\characters\\playercharacter\\humanfemale\\boo_maso.mmh"},
	{0x98947AFD406FF6DDULL, "gen00fl_anders6125700.cl"},
	{0x98947E0B9CAA35E0ULL, "art\\characters\\playercharacter\\humanfemale\\boo_maso.msh"},
	{0x98952B08894E31AAULL, "283120767.ogg"},
	{0x989533094A9F9170ULL, "art\\vfx\\combat\\vfx_w_rev_devour_imp_c.anb"},
	{0x98956AE4C763F5C7ULL, "6056813_m.fxe"},
	{0x9895716E197B1EEEULL, "ht_0_hightownexplosion_n_46_0.rml"},
	{0x989586F3074F9D69ULL, "707547292.ogg"},
	{0x98959C66806F06C3ULL, "905681529.ogg"},
	{0x9895AC5177967EC9ULL, "6065839_m.fxe"},
	{0x98960B48057798BDULL, "w_encounter1_23_0.mao"},
	{0x98960F4805779FF6ULL, "w_encounter1_23_0.mmh"},
	{0x989619480577B0F4ULL, "w_encounter1_23_0.msh"},
	{0x98966919A3FD7C79ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\mov_p_injured.evt"},
	{0x98966D7024000F39ULL, "45969501.ogg"},
	{0x9896C74A719DC378ULL, "vxlt_darktown_6c5_0_0.vlm"},
	{0x98976901E67C6792ULL, "hlac_w_bonepitmining_a2_131076_0z.dds"},
	{0x989788F5C6E21F1CULL, "353654325.ogg"},
	{0x98988686460D78F8ULL, "llac_ca_cavern1_haunted_24426_8.dds"},
	{0x9899844F352717D8ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_disap_p.impa.ani"},
	{0x9899DE21DBF79D3FULL, "vxlt_0_lowtown_a2_n_c_gvb_0_0.vlm"},
	{0x989A33716AB15835ULL, "qun221_petrice6077449.cl"},
	{0x989A59BD968A215CULL, "w_encounter2_camp_19_0_lc.mmh"},
	{0x989A70416EC0ABA5ULL, "818774127.ogg"},
	{0x989B07F929C534FFULL, "genip_node_g1.utp"},
	{0x989B6E1C49B52219ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04a.msh"},
	{0x989B781C49B53307ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04a.mmh"},
	{0x989BD8856314B24AULL, "236290952.ogg"},
	{0x989C1A518ECCB851ULL, "ss_player_female.bnk"},
	{0x989C49609D903134ULL, "llac_ht_0_hightown_n_0_2.dds"},
	{0x989C99D942585D7BULL, "llac_lt_darktown_-1_12.dds"},
	{0x989CBC5A9B642F79ULL, "630134266.ogg"},
	{0x989D12A70F2F41EBULL, "dae000ip_crowd_lm2_talk5.utp"},
	{0x989E20C2D0E0AE6FULL, "hlac_lt_0_lowtown_a3_n_c_-65537_2y.dds"},
	{0x989E3AD773300F2DULL, "vxca_caverns2_7tu_0_1.vlm"},
	{0x989E643794376A58ULL, "vxlt_0_lowtown_docks_n_as3_0_0.vlm"},
	{0x98A00450C32A726CULL, "llac_lt_backalley_n_0_3.dds"},
	{0x98A0B5D0CB58609BULL, "gen00fl_sebastian6174037.cl"},
	{0x98A0C43BE702DFC7ULL, "vxht_bartrand_n_662_0_0.vlm"},
	{0x98A1B8613578074EULL, "hlac_dr_0_deeproads_5842_9x.dds"},
	{0x98A21A339EF38F04ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutc.mmh"},
	{0x98A21E339EF39657ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutc.mao"},
	{0x98A230339EF3B4A6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutc.msh"},
	{0x98A26187451B8899ULL, "vxlt_0_lowtown_a2_n_euy_0_0.vlm"},
	{0x98A27D2AC11BC107ULL, "6122439_m.fxe"},
	{0x98A2962FBE2051BEULL, "6169185_m.fxe"},
	{0x98A2B871BA6F07A5ULL, "t3_cth_cm4.tnt"},
	{0x98A30717DB8FB115ULL, "vxht_0_hightown_n_apv_0_0.vlm"},
	{0x98A321CC1B1BF174ULL, "vxht_chantry_n_2e3_0_0.vlm"},
	{0x98A363B022BCF4D5ULL, "vxlt_0_lowtown_h8x_0_0.vlm"},
	{0x98A37BDE01537518ULL, "6094827_m.fxe"},
	{0x98A3CC0F7F5B4B29ULL, "554417476.ogg"},
	{0x98A42CF9CB2F0E33ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_001a_0s.dds"},
	{0x98A435C8C3CB652DULL, "16054300.ogg"},
	{0x98A47119A409C768ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\mov_p_injured.ani"},
	{0x98A4AB23D8374CD5ULL, "72618018.ogg"},
	{0x98A4BC74DE38FF33ULL, "dae000_bodhan6129103.cl"},
	{0x98A5D0E58BC2DD6AULL, "501837556.ogg"},
	{0x98A5FF2810F6CBB3ULL, "6099705_f.fxe"},
	{0x98A5FF31591E2014ULL, "vxw_blightlands_fq_0_0.vlm"},
	{0x98A60714D67B1542ULL, "vxht_0_hightown_a3_big_0_0.vlm"},
	{0x98A6477E93FE11B0ULL, "981000585.ogg"},
	{0x98A696A98003947CULL, "ss_coterie_mage_f.wbd"},
	{0x98A6D7BA91E80509ULL, "783641556.ogg"},
	{0x98A72D3BF5E429BAULL, "llac_w_woundedcoast_0_9.dds"},
	{0x98A813A942595489ULL, "darktown_male_elf_sta6174569.cl"},
	{0x98A81AEF9EEDAF1AULL, "6017811_m.fxe"},
	{0x98A8594E6FFDA6CBULL, "mag360_samson6042090.cl"},
	{0x98A893449778A392ULL, "vxdr_primevaltomb_7e8_0_2.vlm"},
	{0x98A89F69D82F008BULL, "vxht_0_hightown_c6k_0_0.vlm"},
	{0x98A90699FE957A0CULL, "ht_keep_n_mmap_ph.dds"},
	{0x98A91CD88C8AB157ULL, "vxlt_0_lowtown_n_gpj_0_0.vlm"},
	{0x98A93EFD4AF329ECULL, "103052885.ogg"},
	{0x98AAD4929B28FB8AULL, "vxht_mansion2_28i_0_0.vlm"},
	{0x98AC41B60C57E872ULL, "vxht_0_hightown_ap2_0_0.vlm"},
	{0x98ACD58625370AF3ULL, "gen_im_npc_elf_arm_boo_dala.uti"},
	{0x98ACFE8DD4452E3DULL, "gen00fl_merrill6116146.cl"},
	{0x98AE3A5C1E2C7FEEULL, "llac_lt_0_lowtown_a3_n_c_131072_5.dds"},
	{0x98AEBA761B520FD2ULL, "6146970_f.fxe"},
	{0x98AF37EF3676E5A8ULL, "168891838.ogg"},
	{0x98AF7FC32CAE7B9BULL, "vxlt_hangedman_n_1y9_0_0.vlm"},
	{0x98AFF2D2D52453F1ULL, "ts_chantry_02.gda"},
	{0x98B01C01C35B46FCULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_enter.ani"},
	{0x98B0FD686E26A05AULL, "vxdr_primevaltomb_3hu_0_0.vlm"},
	{0x98B178F4CD85C995ULL, "vxht_0_hightown_n_auu_0_0.vlm"},
	{0x98B194C11D8D0435ULL, "6118501_f.fxe"},
	{0x98B27F46DFA4E46CULL, "gen00fl_anders6119890.cl"},
	{0x98B2A54F6D625CF3ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls108a_0d.dds"},
	{0x98B2BE22B95BD597ULL, "t1_mue_gn1.tnt"},
	{0x98B2F2D492AA1FCEULL, "788988033.ogg"},
	{0x98B314116A373661ULL, "vxlt_gallowstemplar_f_45k_0_0.vlm"},
	{0x98B32A942B15F404ULL, "20598947.ogg"},
	{0x98B389A48E832B97ULL, "sib111_mother6098971.cl"},
	{0x98B3D4643A2FEC15ULL, "hlac_ht_mansion_n_3367_1y.dds"},
	{0x98B40399C2BA4C9BULL, "mrl210_merrill_cave.stg"},
	{0x98B44AA448CC72F5ULL, "vxlt_hangedman_n_1o1_0_0.vlm"},
	{0x98B51ACFE6AFE6C7ULL, "549598247.ogg"},
	{0x98B5F0EDBC342C03ULL, "717330648.ogg"},
	{0x98B6098BCC177807ULL, "vxht_keep_n_67z_0_0.vlm"},
	{0x98B61DA4F3328162ULL, "vxca_caverns1_j2e_0_2.vlm"},
	{0x98B65ECD1094FD92ULL, "73368038.ogg"},
	{0x98B69132E826098AULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_5z.dds"},
	{0x98B6A9C19A5F634EULL, "dr_slavers_3207.rml"},
	{0x98B6F7D222434B99ULL, "art\\vfx\\combat\\vfx_w_rev_devour_act_c.mmh"},
	{0x98B72901C3617B3EULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_enter.gad"},
	{0x98B804346FCDDED7ULL, "t3_csp_b01.tnt"},
	{0x98B87B61DEE9BD7CULL, "vxht_mansion2b_1t8_0_0.vlm"},
	{0x98B8956405E2E5E4ULL, "453759263.ogg"},
	{0x98B899C5D0A2BADCULL, "vxlt_0_lowtown_n_e8g_0_0.vlm"},
	{0x98B98B0B9CC967A5ULL, "art\\characters\\playercharacter\\humanfemale\\boo_maso.phy"},
	{0x98BA4C4D7DE61E32ULL, "vxlt_0_lowtown_docks_a3_n_d73_0_0.vlm"},
	{0x98BA539C3D0D6E2CULL, "vxw_woundedcoast_16f_0_0.vlm"},
	{0x98BABCE8AAA9CD72ULL, "vxca_caverns2_9f7_0_0.vlm"},
	{0x98BAC6F856E9F280ULL, "flare02.dds"},
	{0x98BB971AD9BA70E5ULL, "vxht_0_hightown_a2_b7w_0_0.vlm"},
	{0x98BBB1FC9A555D30ULL, "isa311_castillon6092089.cl"},
	{0x98BC7F92B737AE6EULL, "711098316.ogg"},
	{0x98BC9CC72B8BE248ULL, "vxlt_gallowstemplar_f_45n_0_0.vlm"},
	{0x98BE0401C3675B5DULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_enter.evt"},
	{0x98BE1A6800E06764ULL, "llac_lt_0_lowtown_n_131072_4.dds"},
	{0x98BE3E04BB9E4990ULL, "gen00fl_isabela_init6136304.cl"},
	{0x98BE3F90157F1619ULL, "645688277.ogg"},
	{0x98BE44D85C4027CEULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.impa.ani"},
	{0x98BE601ABC21D9B3ULL, "6080163_f.fxe"},
	{0x98BEC0C9CD33E9A0ULL, "404473064.ogg"},
	{0x98BF0DF92CBBB528ULL, "6136738_m.fxe"},
	{0x98BF79E9A20C82BEULL, "6163385_m.fxe"},
	{0x98BFD4876240832EULL, "401247192.ogg"},
	{0x98C00743B0C95EC7ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgta_n.dds"},
	{0x98C00CF577BC265EULL, "dae200st_solivitus_plus.utm"},
	{0x98C04C0AEE4F900AULL, "hlac_w_bonepit_262147_0z.dds"},
	{0x98C0E02C100C30BEULL, "vxlt_backalley_g_1ud_0_0.vlm"},
	{0x98C13368DC001088ULL, "ser221_quentin6123356.cl"},
	{0x98C1D1EADA69ACEFULL, "65740577.ogg"},
	{0x98C1E5607E3A33F6ULL, "art\\levels\\ztools\\default\\generic\\1oo_basic.msh"},
	{0x98C1EF607E3A44F4ULL, "art\\levels\\ztools\\default\\generic\\1oo_basic.mmh"},
	{0x98C1F3607E3A4B07ULL, "art\\levels\\ztools\\default\\generic\\1oo_basic.mao"},
	{0x98C2487919008C9DULL, "vxht_playershouse_n_1lo_0_0.vlm"},
	{0x98C2E2E1F4E2FCB3ULL, "6093505_f.fxe"},
	{0x98C2E2EB3D0A5114ULL, "vxw_blightlands_lw_0_0.vlm"},
	{0x98C2EACEBA674642ULL, "vxht_0_hightown_a3_bca_0_0.vlm"},
	{0x98C335DE45A55EE5ULL, "vxdr_primevalentrance_2z9_0_0.vlm"},
	{0x98C33C9B812FD5E1ULL, "6015920_m.fxe"},
	{0x98C3B7441D9E8C73ULL, "198719943.ogg"},
	{0x98C3DD650615C6E7ULL, "537037226.ogg"},
	{0x98C4170E6B5B6A63ULL, "art\\characters\\creatures\\mubari\\animation\\com_bark.evt"},
	{0x98C47B68BE0D9087ULL, "vxht_keep_n_62u_0_0.vlm"},
	{0x98C4C96003C7C2DBULL, "659253694.ogg"},
	{0x98C4D1C2600CD047ULL, "vxht_mansion2b_2iw_0_0.vlm"},
	{0x98C5456947048CF0ULL, "art\\characters\\creatures\\darkspawnogre\\cn_hed_ogra.phy"},
	{0x98C55FA7F01366BAULL, "231440444.ogg"},
	{0x98C56EFBCECD9BE5ULL, "149131697.ogg"},
	{0x98C5A00958D6E42CULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.cess.ani"},
	{0x98C5D37E81BE44B4ULL, "958038603.ogg"},
	{0x98C5F841E4E41FA4ULL, "zz_fen_debug6134327.cl"},
	{0x98C64D97D1AA2D9FULL, "dae000_templar_knight6165175.cl"},
	{0x98C65ABD23F1E3C5ULL, "6087329_m.fxe"},
	{0x98C66F612A0FD657ULL, "518963997.ogg"},
	{0x98C68E685A75975FULL, "t3_har_lbr.tnt"},
	{0x98C69AFAEBADD7FDULL, "6165469_m.fxe"},
	{0x98C6A9DE8DEDDF53ULL, "737605502.ogg"},
	{0x98C6EB997EA8456EULL, "6109107_m.fxe"},
	{0x98C7611C49DA473CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth04a.phy"},
	{0x98C83D88CBD4DD91ULL, "vxht_brothel_n_5cd_0_0.vlm"},
	{0x98C83FE07EBBED3BULL, "lt_0_lowtown_n_33_0.rml"},
	{0x98C8C4169CDE08D7ULL, "6149238_m.fxe"},
	{0x98C951F6D74A61A3ULL, "vxdr_0_deeproads_cf0_0_0.vlm"},
	{0x98C97E8BC545F999ULL, "vxca_caverns1_mines_i3a_0_0.vlm"},
	{0x98C9842FFA049152ULL, "vxlt_0_lowtown_docks_n_b6t_0_1.vlm"},
	{0x98C9C8B6CBEB5082ULL, "gen00fl_aveline6131606.cl"},
	{0x98C9E9B9A8C8CA59ULL, "gen_im_trs_bst_drg_gland_03.uti"},
	{0x98CA08F7CBB08965ULL, "vxht_0_hightown_a2_b8v_0_0.vlm"},
	{0x98CA23F91B013BFCULL, "mag311_nyssa6041769.cl"},
	{0x98CA40F7305C06B4ULL, "vxdr_primevalentrance_1qv_0_0.vlm"},
	{0x98CA730590F986AAULL, "zz_dre_debug6127484.cl"},
	{0x98CAF00E6B6146E4ULL, "art\\characters\\creatures\\mubari\\animation\\com_bark.gad"},
	{0x98CB62FEC6058E28ULL, "lgt120_orz_bln_official6170492.cl"},
	{0x98CBBE871E0F96FDULL, "w_bonepit_204_0.mao"},
	{0x98CBC2871E0F9D36ULL, "w_bonepit_204_0.mmh"},
	{0x98CBCC871E0FAE34ULL, "w_bonepit_204_0.msh"},
	{0x98CC3CD85C4C571FULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.impa.evt"},
	{0x98CC6B4F8EC33BB3ULL, "vxdr_primevalentrance_2vz_0_1.vlm"},
	{0x98CCAD0958DD196EULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.cess.gad"},
	{0x98CDB2B43C562E27ULL, "717720274.ogg"},
	{0x98CDE6D57655662AULL, "63243107.ogg"},
	{0x98CE00FEAE2D4FB2ULL, "vxdr_0_deeproads_bp7_0_1.vlm"},
	{0x98CE2E0FE1D69067ULL, "90512385.ogg"},
	{0x98CEDBD3089DF853ULL, "hlac_lt_backalley_n_0_13x.dds"},
	{0x98CF092ADDA2EABDULL, "gen00fl_aveline6116653.cl"},
	{0x98CFAC5F7D853798ULL, "6100762_m.fxe"},
	{0x98CFEFD9776AA171ULL, "6187869_m.fxe"},
	{0x98D00580C65858F2ULL, "hlac_ht_0_hightown_a3_131070_0x.dds"},
	{0x98D04E38D379BCBAULL, "6118992_m.fxe"},
	{0x98D0F9A352753268ULL, "lt_0_lowtown_docks_150_0.mmh"},
	{0x98D105A35275460BULL, "lt_0_lowtown_docks_150_0.mao"},
	{0x98D117A352756492ULL, "lt_0_lowtown_docks_150_0.msh"},
	{0x98D11E1181204125ULL, "209166294.ogg"},
	{0x98D154C82F006994ULL, "vxw_blightlands_qt_0_0.vlm"},
	{0x98D1ED350BDB5474ULL, "character_character_p285.psh.fxo.dx11.assembly"},
	{0x98D1FF0E6B677EC2ULL, "art\\characters\\creatures\\mubari\\animation\\com_bark.ani"},
	{0x98D2EB5EF583D717ULL, "qun221_petrice6083710.cl"},
	{0x98D2F94F76CC6A10ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslba_0n.dds"},
	{0x98D31AB9F0D22F6EULL, "6147682_m.fxe"},
	{0x98D31DD85C524108ULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.impa.gad"},
	{0x98D3439F5202E8C7ULL, "vxht_mansion2b_2bv_0_0.vlm"},
	{0x98D394A9802A7EF6ULL, "ss_coterie_mage_f.bnk"},
	{0x98D3A80958E32F4DULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.cess.evt"},
	{0x98D3C0D017E965C3ULL, "6080476_m.fxe"},
	{0x98D437574BA32928ULL, "6121064_m.fxe"},
	{0x98D535797AC2FB7BULL, "popuplayer.gfx"},
	{0x98D5949CB8CB6539ULL, "6208192_f.fxe"},
	{0x98D61A6F9DD05A6AULL, "37300818.ogg"},
	{0x98D61C2D310D3007ULL, "6067668_m.fxe"},
	{0x98D630B7FB204E8EULL, "zz_dae_debug6050002.cl"},
	{0x98D64CB4043ED3AEULL, "6095978_m.fxe"},
	{0x98D668F874697F14ULL, "6020063_f.fxe"},
	{0x98D7825E09667D1AULL, "hlac_w_sundermount_327676_0x.dds"},
	{0x98D7F455207D7B7FULL, "vxlt_gallowstemplar_n_3sp_0_0.vlm"},
	{0x98D8B0452AAC9F8EULL, "vxlt_gallowscourtyard_7bg_0_0.vlm"},
	{0x98D8BFBFECF3F8ECULL, "llac_ca_caverns2prc_5802_48.dds"},
	{0x98D94781ED3737ACULL, "gen00fl_varric6076531.cl"},
	{0x98D9539DFE712C1EULL, "735790041.ogg"},
	{0x98D9B52AC9EA675AULL, "vxht_brothel_n_4fq_0_0.vlm"},
	{0x98D9B8EDDB577DE7ULL, "vxht_brothel_n_57n_0_0.vlm"},
	{0x98DA470E322043B3ULL, "6090213_m.fxe"},
	{0x98DAF0BBD4BB0750ULL, "763614153.ogg"},
	{0x98DB03CF595BEAE8ULL, "886460314.ogg"},
	{0x98DB03D5781CBC01ULL, "vxlt_0_lowtown_n_e86_0_0.vlm"},
	{0x98DB04E6B9C3A63EULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_cobble_sa.dds"},
	{0x98DB3C0DF4885768ULL, "gen00fl_dog6117682.cl"},
	{0x98DB457DDB2309FAULL, "gen00fl_isabela6099321.cl"},
	{0x98DB8EB46E862EB6ULL, "vxdr_primevalentrance_15x_0_0.vlm"},
	{0x98DBF73473ADA724ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_002a_0d.dds"},
	{0x98DC05DCE80E6E69ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur4_flemeth_fake.ani"},
	{0x98DC0ABFAE33A642ULL, "382082535.ogg"},
	{0x98DC9B9A3DE9EBE4ULL, "294167243.ogg"},
	{0x98DCD07415E49CF1ULL, "ht_0_hightown_a3_60_0.rml"},
	{0x98DD70CF364EA548ULL, "6164236_m.fxe"},
	{0x98DD7601D91545E1ULL, "vxht_0_hightown_n_61z_0_0.vlm"},
	{0x98DDAB816757510DULL, "hf_isabela.mor"},
	{0x98DE1BC1993391EBULL, "695945490.ogg"},
	{0x98DE28B18C247CDAULL, "6054573_m.fxe"},
	{0x98DE5BE524FD57B6ULL, "hlac_lt_0_lowtown_n_65535_4x.dds"},
	{0x98DF72DD15A06F6CULL, "6173139_m.fxe"},
	{0x98DFF4D222664B8EULL, "art\\vfx\\combat\\vfx_w_rev_devour_act_c.anb"},
	{0x98E0709D28AB68E8ULL, "vxw_bonepitdestroyed_a3_1ra_0_0.vlm"},
	{0x98E0A0C5AADB8919ULL, "hlac_lt_0_lowtown_n_-196609_1z.dds"},
	{0x98E0DD785D1FDB90ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebc.mao"},
	{0x98E0E1785D1FE2A3ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebc.mmh"},
	{0x98E0EEA263A356C3ULL, "6208492_m.fxe"},
	{0x98E0EF785D1FFAFDULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderwebc.msh"},
	{0x98E138A6D1DA7F39ULL, "fen110_fenris6098902.cl"},
	{0x98E15FB97F710F58ULL, "6053263_f.fxe"},
	{0x98E2B4F3E9B8A8ABULL, "vxlt_darktown_4u6_0_0.vlm"},
	{0x98E2C883057ABDF7ULL, "6104447_f.fxe"},
	{0x98E2CFC57CF54BBBULL, "vxw_bonepitmining_a2_8q_0_0.vlm"},
	{0x98E312AB500CCDD9ULL, "docks_fem_sailor_mob6176299.cl"},
	{0x98E320DCE814BAEBULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur4_flemeth_fake.gad"},
	{0x98E3F0CC320E3E04ULL, "mcr370_cullen6087427.cl"},
	{0x98E437C35595C7A7ULL, "6026942_m.fxe"},
	{0x98E4DD45F19B6A54ULL, "445631748.ogg"},
	{0x98E50A852583CA60ULL, "6118074_f.fxe"},
	{0x98E525EF6B6F2B71ULL, "mcr370_meredith6083056.cl"},
	{0x98E55DDAD11776BDULL, "584413871.ogg"},
	{0x98E5660C332D106FULL, "dae100_alistair_drunk6135132.cl"},
	{0x98E5795752E141D0ULL, "vxlt_ruinedhouse_n_88_0_0.vlm"},
	{0x98E5C5E605532DB8ULL, "vrc321ip_spooky6_obj_5.utp"},
	{0x98E60DA38AD40E34ULL, "6026084_m.fxe"},
	{0x98E638AF0186E1B1ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.dura.ani"},
	{0x98E64111AF0ECEC5ULL, "6041834_f.fxe"},
	{0x98E643B9D464194DULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\2048texel.dds"},
	{0x98E662A96304BF36ULL, "hlac_ht_0_hightown_a2_-65534_0y.dds"},
	{0x98E67414AA734A35ULL, "6033982_m.fxe"},
	{0x98E6A3ABC8A2788DULL, "vxca_cavern1_haunted_cjp_0_0.vlm"},
	{0x98E703079A59A838ULL, "246288316.ogg"},
	{0x98E724B1144837B4ULL, "vxdr_primevalentrance_1kt_0_0.vlm"},
	{0x98E74C5CB065E144ULL, "vxlt_0_lowtown_a2_n_emp_0_0.vlm"},
	{0x98E7F3C49B686CACULL, "dae000_bartrand6201350.cl"},
	{0x98E874B44D38EEC8ULL, "fx_fadecloudb.mao"},
	{0x98E8CA6D26CDD635ULL, "6028404_f.fxe"},
	{0x98E8D41BCDD2AA60ULL, "hlac_lt_0_lowtown_a2_n_131072_6z.dds"},
	{0x98E8DD9FD025571AULL, "155845415.ogg"},
	{0x98E91E7E04A2F7C3ULL, "vxht_mansion2b_20t_0_0.vlm"},
	{0x98E9742968FB4E1CULL, "823297744.ogg"},
	{0x98E9E31F5981B21DULL, "vxdr_primevaltomb_2sj_0_0.vlm"},
	{0x98EA3797486A640CULL, "6027696_m.fxe"},
	{0x98EA7DB2C8E11CEEULL, "vxht_0_hightown_cbv_0_0.vlm"},
	{0x98EA8F1CD1AF4C82ULL, "hlac_ca_caverns2_5802_67z.dds"},
	{0x98EABB3A6E46B484ULL, "vxlt_0_lowtown_a2_hdm_0_0.vlm"},
	{0x98EB0114066920E0ULL, "w_bonepit_19_0_lc.mmh"},
	{0x98EB34D2DE27D54CULL, "follower_banter6156063.cl"},
	{0x98EBD1E2352E9EFCULL, "6135887_m.fxe"},
	{0x98ECDDB26CE86D8CULL, "6167592_m.fxe"},
	{0x98ECF7D87021A550ULL, "495986829.ogg"},
	{0x98ED3BD5F8CD20CAULL, "6016914_f.fxe"},
	{0x98ED43AF018D1353ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.dura.gad"},
	{0x98EDB04E5FB8FBE8ULL, "6103204_m.fxe"},
	{0x98EE0A7551A22116ULL, "llac_lt_0_lowtown_a2_n_c_-2_3.dds"},
	{0x98EE6EE093EF60E7ULL, "dae021_sister_samea_sta6150176.cl"},
	{0x98EE8F69B892C323ULL, "vxlt_gallowscourtyard_n_69e_0_0.vlm"},
	{0x98EEC3969174EF0AULL, "480046109.ogg"},
	{0x98EED33A83523856ULL, "zz_vault_debug6200304.cl"},
	{0x98EF8706B52BDB74ULL, "6020055_m.fxe"},
	{0x98EF8BDE265D69D6ULL, "489314076.ogg"},
	{0x98EF8D6A192A9956ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1a.msh"},
	{0x98EF976A192AAA54ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1a.mmh"},
	{0x98EF9B6A192AB167ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1a.mao"},
	{0x98EF9E9C2E18C162ULL, "vxht_0_hightown_a2_5wg_0_0.vlm"},
	{0x98EFE44D263EA4C8ULL, "llac_ca_caverns1_24427_2.dds"},
	{0x98F00AACB2549E21ULL, "hlac_ca_caverns2_5802_51x.dds"},
	{0x98F0E29E63B86A7EULL, "vxlt_gallowstemplar_f_5v8_0_0.vlm"},
	{0x98F10B5191C03922ULL, "vxht_0_hightown_a3_caj_0_0.vlm"},
	{0x98F16C12B0A9082AULL, "6129291_m.fxe"},
	{0x98F1A0F3CEF97E24ULL, "hla_ht_0_hightown_a3_2_0y.dds"},
	{0x98F1E72133AEE8F7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_windarkbase.msh"},
	{0x98F1F92133AF075AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_windarkbase.mao"},
	{0x98F1FD2133AF0E69ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_windarkbase.mmh"},
	{0x98F221CB00AD65CAULL, "hlac_lt_gallowscourtyard_65536_0y.dds"},
	{0x98F29481E2BE42F6ULL, "qun231_varnell6077229.cl"},
	{0x98F2B93A3541F86EULL, "6043912_f.fxe"},
	{0x98F2F85F71789643ULL, "rog_confusion.dds"},
	{0x98F3275E4648F75FULL, "follower_banter6167857.cl"},
	{0x98F340AF019179B8ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.dura.evt"},
	{0x98F3874F2063BC2EULL, "gen00fl_aveline6121957.cl"},
	{0x98F38CE72C4CEDF3ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_3x.dds"},
	{0x98F3931FA1B72711ULL, "vxht_brothel_n_5ly_0_0.vlm"},
	{0x98F3F930745C76B3ULL, "gen00cs_cs_end_music_int_4.ncs"},
	{0x98F47F4F3A61E925ULL, "gen00fl_fenris6113813.cl"},
	{0x98F4EFD9FA5BE88FULL, "ave200_donnic6110650.cl"},
	{0x98F564908BDB9DFDULL, "lt_0_lowtown_docks_33_0_ll.mmh"},
	{0x98F593FF0E98D08EULL, "vxlt_gallowscourtyard_7hy_0_0.vlm"},
	{0x98F5DC43941FCD54ULL, "6076218_m.fxe"},
	{0x98F5EAB61A86DFE7ULL, "vxht_0_hightown_a2_c4a_0_0.vlm"},
	{0x98F5F16840C79202ULL, "mag130_danzig6016428.cl"},
	{0x98F61C9C403DBCA4ULL, "6107437_m.fxe"},
	{0x98F73EC6CD467C52ULL, "dae200_cullen6140324.cl"},
	{0x98F7C28F245954F7ULL, "vxw_bonepitmining_a2_6k_0_0.vlm"},
	{0x98F8C026700036A8ULL, "1019709089.ogg"},
	{0x98F8DABE923A6B1FULL, "102664668.ogg"},
	{0x98F8FB572610FF86ULL, "6135346_m.fxe"},
	{0x98F92869473C6C88ULL, "vxht_0_hightown_a3_a4q_0_0.vlm"},
	{0x98F943223899A63EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enrage.ani"},
	{0x98FB715D550C2F13ULL, "vxca_caverns2_74j_0_0.vlm"},
	{0x98FBB5535DF2EE63ULL, "gen00fl_sebastian6188113.cl"},
	{0x98FBC7538B86BD6AULL, "vxlt_hangedman_21d_0_0.vlm"},
	{0x98FC0FECE0D8EE9DULL, "vxht_playershouse_n_1pk_0_0.vlm"},
	{0x98FC1F08E2BCC5ABULL, "hlac_lt_0_lowtown_n_-65536_4y.dds"},
	{0x98FC222B51AF1468ULL, "6104304_m.fxe"},
	{0x98FC3CAE75FEE4F0ULL, "ave310_donnic6152592.cl"},
	{0x98FC8CD3CE9E2A6AULL, "gen_im_arm_boo_hvy_act3_01.uti"},
	{0x98FCAC43A00E438DULL, "vxht_0_hightown_a2_cbe_0_0.vlm"},
	{0x98FD03B74612FBC3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.ani"},
	{0x98FDF2CF136A452EULL, "vxca_cavern1_haunted_gir_0_0.vlm"},
	{0x98FE2F72D8C3AF51ULL, "hlac_lt_warehouse_0_2z.dds"},
	{0x98FEAD99A308A986ULL, "plt_cod_ltr_ggt_note2.plo"},
	{0x98FEBC5EFB34A6A1ULL, "6020414_m.fxe"},
	{0x98FF3F65F08DE557ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rockpile.phy"},
	{0x98FF547B55AE82FEULL, "vxlt_gallowstemplar_f_5y9_0_0.vlm"},
	{0x98FF78777C83F623ULL, "vxht_chantry_2q3_0_0.vlm"},
	{0x98FFF4D41EF6A728ULL, "gen00fl_varric6033598.cl"},
	{0x99001F529A7A9E9AULL, "one300_murder_followers6086731.cl"},
	{0x9901023EA39F8233ULL, "79294015.ogg"},
	{0x9902077684EF24D5ULL, "6039275_f.fxe"},
	{0x9902083746F51E93ULL, "hlac_lt_gallowstemplar_f_2394_9y.dds"},
	{0x990268F2FD792258ULL, "lt_undercityserialkiller_3567_lc.mmh"},
	{0x9902905DEBB62920ULL, "433074377.ogg"},
	{0x9902AC2644627985ULL, "docks_male_human_worker_sta6176003.cl"},
	{0x9902C5C82C9FF085ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masc_0d.dds"},
	{0x9902F0B91ECF6578ULL, "6157192_f.fxe"},
	{0x9902F72840D0A355ULL, "11446498.ogg"},
	{0x9902FDFEB2197DF4ULL, "gen_im_trs_krk_drk_junk2_01.uti"},
	{0x990302F9E7B09F96ULL, "dae000_flemeth6000884.cl"},
	{0x9903038AC1545AC7ULL, "qun110_petrice6037370.cl"},
	{0x99033ABFABACAB81ULL, "zz_otl_debug6060599.cl"},
	{0x99035512C632AF69ULL, "codex_col.gda"},
	{0x9903C0F6F79665E1ULL, "gen00fl_aveline6122786.cl"},
	{0x9903F0BCD770EF5EULL, "vxlt_undercityserialkiller_1it_0_0.vlm"},
	{0x9904366DD2DE011FULL, "3p_200_a_frame.stg"},
	{0x99049B7DD08C8831ULL, "gen00fl_aveline6139399.cl"},
	{0x99064A5F7A6F3E4CULL, "vxw_bonepit_zr_0_0.vlm"},
	{0x9906500D5E88623BULL, "439518377.ogg"},
	{0x990651198A515A82ULL, "834925694.ogg"},
	{0x99065B2238A4590FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enrage.evt"},
	{0x99065D6299944939ULL, "llac_ca_caverns1_24424_8.dds"},
	{0x9906D6895CB88F93ULL, "97479117.ogg"},
	{0x9906E20B86C60998ULL, "6030469_m.fxe"},
	{0x9906F1B431229B1EULL, "gen00fl_aveline6115860.cl"},
	{0x9906F21E62841867ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.dura.gad"},
	{0x99072D0238FA0A84ULL, "482176435.ogg"},
	{0x99073471690A7C6AULL, "ser221_quentin6123365.cl"},
	{0x990745B44D534F2AULL, "fx_fadecloudb.dds"},
	{0x99074BC9F12380DEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_aveline.anb"},
	{0x99075C204125C350ULL, "6061337_m.fxe"},
	{0x99076D3418071806ULL, "6130046_m.fxe"},
	{0x99078F6A2549C7B3ULL, "6116128_m.fxe"},
	{0x9907B780E45E6C68ULL, "6074431_f.fxe"},
	{0x990832BD39757F36ULL, "6148126_m.fxe"},
	{0x990842E07EF1D607ULL, "lt_0_lowtown_n_33_0.anb"},
	{0x99088B54E910536DULL, "75879050.ogg"},
	{0x9909421AF79F124AULL, "vxlt_0_lowtown_docks_n_b0g_0_0.vlm"},
	{0x99096F6B78109C45ULL, "llac_w_sundermount_4_0.dds"},
	{0x99099D075F2AB3D8ULL, "289078387.ogg"},
	{0x9909B384681110F4ULL, "6075556_m.fxe"},
	{0x9909DEC2F7259AAFULL, "ser270_emeric6116095.cl"},
	{0x990A76FFBC23C3A1ULL, "6070217_m.fxe"},
	{0x990AA6EC38751BA3ULL, "84938074.ogg"},
	{0x990AC2348827B0F8ULL, "gencr_dalish_warrior_2w_f.utc"},
	{0x990AFCBC2AE6E31CULL, "vxlt_0_lowtown_a2_gxf_0_0.vlm"},
	{0x990B1BB7461F61BAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.evt"},
	{0x990B39A551DFF3EBULL, "vxca_cave_1nu_0_0.vlm"},
	{0x990B4D64E420FB12ULL, "genip_pp_hl_sustained.utp"},
	{0x990B58A888BA74C6ULL, "character_character_p47.psh.fxo.dx11.assembly"},
	{0x990B6D0CF1B5FD73ULL, "aip_coteriethug_base.gda"},
	{0x990BDBE90DE852BCULL, "vxca_caverns1_mines_i38_0_0.vlm"},
	{0x990C79D0525F549FULL, "6153390_m.fxe"},
	{0x990CE71E62887174ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.dura.evt"},
	{0x990CF0B7853985E4ULL, "lt_0_lowtown_docks_10_0.anb"},
	{0x990D2E3BED2ABF21ULL, "6027514_m.fxe"},
	{0x990D5C2238AA79B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enrage.gad"},
	{0x990D69C1F4C23855ULL, "vxht_chantry_n_2dy_0_0.vlm"},
	{0x990DFCB17838B661ULL, "6096499_m.fxe"},
	{0x990E7D8EC4884B97ULL, "vxlt_0_lowtown_a2_n_gwc_0_0.vlm"},
	{0x990EBEA0A9FCE684ULL, "vxht_keep_n_57r_0_0.vlm"},
	{0x990EF1B72A12B759ULL, "538270970.ogg"},
	{0x990F988C70A46632ULL, "6153728_m.fxe"},
	{0x990FA33B3E9A3E28ULL, "487442975.ogg"},
	{0x990FBF57DAE04A8EULL, "vxlt_0_lowtown_docks_bdp_0_0.vlm"},
	{0x9910113261E62DE9ULL, "gen00fl_anders6116122.cl"},
	{0x99101E313861EABAULL, "hf_gen_slaver.mor"},
	{0x9910706F01CF8048ULL, "w_sundermount_229_0.msh"},
	{0x99107631737B8CAAULL, "gen00fl_aveline6114823.cl"},
	{0x9910826F01CF9ED9ULL, "w_sundermount_229_0.mao"},
	{0x99108E6F01CFB372ULL, "w_sundermount_229_0.mmh"},
	{0x9910D293CE4C8A66ULL, "401502198.ogg"},
	{0x9910EC20DEF45717ULL, "278665735.ogg"},
	{0x99111AB74623CBADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.gad"},
	{0x99119A89DB855D38ULL, "lgt111tr_bladetrap1.ncs"},
	{0x9911A3FD77A44CF1ULL, "llac_lt_0_lowtown_a2_n_c_65534_0.dds"},
	{0x9911DF05BE358C47ULL, "6037607_m.fxe"},
	{0x9911F5FBC8365D89ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basic.phy"},
	{0x991220BFAB7ED91FULL, "eye_eyelash_p0.psh.fxo.dx11.assembly"},
	{0x991243CDA9C5FC30ULL, "601358674.ogg"},
	{0x99126299C96707DEULL, "vxlt_undercityserialkiller_1nu_0_0.vlm"},
	{0x99127302695A28BAULL, "one300dg_vanard6063952.cl"},
	{0x99127B85191E137AULL, "qcr000_cut2d_thechampion6025133.cl"},
	{0x9912A45673E9D87FULL, "61211749.ogg"},
	{0x9912BEA6AB452A60ULL, "231095831.ogg"},
	{0x9912C0176863CA22ULL, "vxlt_gallowstemplar_n_5j8_0_0.vlm"},
	{0x9912D490A166B27DULL, "6147850_m.fxe"},
	{0x9912DEFCB2E7BA58ULL, "vxht_mansion2b_3em_0_0.vlm"},
	{0x9912EA97D01A6307ULL, "108155766.ogg"},
	{0x99138593AA684C59ULL, "hlac_lt_0_lowtown_docks_-131075_2y.dds"},
	{0x99138C39C3C8F265ULL, "6020447_m.fxe"},
	{0x9913997BC37A57EDULL, "vxlt_0_lowtown_docks_a3_d2q_0_1.vlm"},
	{0x9913A882C1ECAAD0ULL, "696975141.ogg"},
	{0x991424210115FEE4ULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_2y.dds"},
	{0x9914A649084585F7ULL, "vxw_bonepitmining_a2_8e_0_0.vlm"},
	{0x9914E54005E208A8ULL, "6041766_m.fxe"},
	{0x9914FAB9F2BD178BULL, "6177569_m.fxe"},
	{0x99159D592943DF79ULL, "208077197.ogg"},
	{0x9915B5257C21B70EULL, "6120817_f.fxe"},
	{0x9915FAFF0512E1E5ULL, "hlac_lt_gallowstemplar_2383_12z.dds"},
	{0x991621FCA18A980EULL, "516831653.ogg"},
	{0x9916A4B3D2B6EF11ULL, "lt_0_lowtown_docks_113_0.msh"},
	{0x9916B2B3D2B7072CULL, "lt_0_lowtown_docks_113_0.mao"},
	{0x9916BEB3D2B71BCFULL, "lt_0_lowtown_docks_113_0.mmh"},
	{0x99170A6DE142E156ULL, "934809100.ogg"},
	{0x99173FAF481C4866ULL, "vxht_0_hightown_bot_0_0.vlm"},
	{0x99174904F47814E6ULL, "vxca_caverns2_8pp_0_0.vlm"},
	{0x99175CB45467C37BULL, "6126871_m.fxe"},
	{0x99175D5508E78B9DULL, "lt_0_lowtown_docks_n_160_0.mao"},
	{0x9917615508E792D6ULL, "lt_0_lowtown_docks_n_160_0.mmh"},
	{0x99176B5508E7A3D4ULL, "lt_0_lowtown_docks_n_160_0.msh"},
	{0x9917B45B23C89469ULL, "vxht_0_hightown_cf1_0_0.vlm"},
	{0x9917E3D10C650C8EULL, "one3510cr_dead_fighter_5.utc"},
	{0x9917F9AF39E90C21ULL, "1004104511.ogg"},
	{0x9917FCECADD87311ULL, "hm_taryn.mor"},
	{0x99184AF62C43E9E7ULL, "513049364.ogg"},
	{0x991902DFBD589C7EULL, "sky_lt_undercitydungeon.mao"},
	{0x99195ECD3A8E3108ULL, "965980672.ogg"},
	{0x9919D7998C5E4FB2ULL, "6089478_m.fxe"},
	{0x991ADF1E6294A0ADULL, "art\\vfx\\creatures\\vfx_c_drag_spit_p.dura.ani"},
	{0x991B58F9E21F7979ULL, "zz_vault_debug6178916.cl"},
	{0x991BF9EED752401CULL, "45625093.ogg"},
	{0x991C6BDFB9736FCFULL, "llac_ca_caverns2prc_5802_21.dds"},
	{0x991C940C93A6A956ULL, "vxlt_0_lowtown_a2_gz6_0_0.vlm"},
	{0x991CDB3E22866AC4ULL, "vxca_caverns1_mines_i0y_0_0.vlm"},
	{0x991CEE607E86EEEFULL, "art\\levels\\ztools\\default\\generic\\1oo_basic.phy"},
	{0x991D2D6CF9BBCA42ULL, "35716356.ogg"},
	{0x991E4A915DE53614ULL, "mag101_grace6035864.cl"},
	{0x991E651BA98B1C31ULL, "hlac_lt_0_lowtown_1_9z.dds"},
	{0x991F2E901750FF3DULL, "ply_glow.mao"},
	{0x991F46DA84B80F65ULL, "6070262_m.fxe"},
	{0x991F55E8E117F183ULL, "vxw_encounter1_cw_0_0.vlm"},
	{0x991F5E9FE7715883ULL, "vxlt_hangedman_n_rk_0_0.vlm"},
	{0x991F613CC876559BULL, "cs_mcr370cs_arrival_gallows.bnk"},
	{0x991FA56AECC2CED0ULL, "zz_dog_debug6127913.cl"},
	{0x991FE34B4807E315ULL, "6171463_f.fxe"},
	{0x9920002BA85C5F71ULL, "vxlt_gallowstemplar_5y9_0_0.vlm"},
	{0x9920A628E7F53E86ULL, "follower_banter6143364.cl"},
	{0x9920AD4D76FD5808ULL, "gen00fl_aveline6140147.cl"},
	{0x9920B90845A3BCF9ULL, "follower_banter6170639.cl"},
	{0x9920F8B980B208CBULL, "89486723.ogg"},
	{0x99213B31162267BBULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_bdaa_0n.dds"},
	{0x9921466D935CCAFDULL, "6140950_m.fxe"},
	{0x99224F7B4B868765ULL, "6108486_m.fxe"},
	{0x99225BE9A60C983AULL, "vxht_mansion_n_4yc_0_0.vlm"},
	{0x99226F21200CD097ULL, "80755767.ogg"},
	{0x9922D84C0D56A0D0ULL, "llac_w_encounter2_camp_196608_0.dds"},
	{0x992377FA4FF863E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_1.evt"},
	{0x9923B68A64EF9DA0ULL, "864971795.ogg"},
	{0x9923CFFEE50C0807ULL, "llac_lt_0_lowtown_a2_65534_3.dds"},
	{0x9923ED6B588E8B52ULL, "gen00fl_isabela6135687.cl"},
	{0x99248E2266CD70F4ULL, "6160777_m.fxe"},
	{0x992493C13591CAB1ULL, "gen_im_trs_dal_one_lsw_r0.uti"},
	{0x9924E59F0BE25619ULL, "592889781.ogg"},
	{0x9925B7E8F30B63BFULL, "hm_gustav.mor"},
	{0x9925F96BCB0475B9ULL, "vxht_keep_3zq_0_0.vlm"},
	{0x9925FCA2C4678F79ULL, "347138074.ogg"},
	{0x99263745083F430CULL, "61967548.ogg"},
	{0x992680A2D757A2C2ULL, "one000_ilen6205657.cl"},
	{0x992687DB5FB2AFE8ULL, "vxdr_primevaltomb_8oa_0_0.vlm"},
	{0x9926F328D69439CDULL, "pro000_colonel_ewald6039973.cl"},
	{0x99270FB04CEAC711ULL, "557756899.ogg"},
	{0x99276424280D423CULL, "vxlt_0_lowtown_docks_n_bgd_0_0.vlm"},
	{0x99279F01C85FA52CULL, "hlac_lt_0_lowtown_docks_n_-2_6z.dds"},
	{0x9928A5495BB17E16ULL, "378680716.ogg"},
	{0x9928A7A61573A05BULL, "mcr370_assignments6081692.cl"},
	{0x992907F96E67C258ULL, "ht_0_hightown_a2_20_0_ll.mmh"},
	{0x99291BB5C61E8721ULL, "vxlt_warehouse_n_1br_0_0.vlm"},
	{0x99293621916C3FDAULL, "llac_lt_ruinedhouse_n_-1_4.dds"},
	{0x9929B189DB9A1635ULL, "lgt111tr_bladetrap1.utt"},
	{0x9929D51ABBBE568FULL, "vxlt_0_lowtown_n_db3_0_0.vlm"},
	{0x992A50FA4FFE4064ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_1.gad"},
	{0x992A86C21FB23188ULL, "ico_dagger_5.dds"},
	{0x992ABFA3514EB3BCULL, "dae000_bartrand6031391.cl"},
	{0x992B0518C3DEB8CAULL, "vxca_caverns2_956_0_0.vlm"},
	{0x992B8D89764C23A1ULL, "6169047_m.fxe"},
	{0x992B9EA03B29746EULL, "gen00fl_sebastian6180819.cl"},
	{0x992C01CA06684483ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_fancy.evt"},
	{0x992C2AC6E75DCB3AULL, "vxlt_gallowstemplar_5k4_0_0.vlm"},
	{0x992C5B979E680396ULL, "w_bonepit_243_0.mao"},
	{0x992C5F8607780341ULL, "vxdr_0_deeproads_cwu_0_0.vlm"},
	{0x992C67979E681835ULL, "w_bonepit_243_0.mmh"},
	{0x992C69979E681BABULL, "w_bonepit_243_0.msh"},
	{0x992C901584088C7CULL, "lt_0_lowtown_a3_n_c_40_0.rml"},
	{0x992CD5571CCFD80AULL, "vxht_mansion2_2mb_0_0.vlm"},
	{0x992D180F019F4E90ULL, "358486392.ogg"},
	{0x992D444F98890047ULL, "6000968_m.fxe"},
	{0x992D752B330401D3ULL, "vxlt_0_lowtown_n_f0o_0_0.vlm"},
	{0x992DB8B776AE27E5ULL, "6075162_m.fxe"},
	{0x992DC119B3892FB7ULL, "1020294322.ogg"},
	{0x992DC7C5D30E0A03ULL, "vxw_encounter1_hv_0_0.vlm"},
	{0x992E089252A59602ULL, "and100_anders6095808.cl"},
	{0x992E600E198FA866ULL, "vxlt_gallowstemplar_5az_0_0.vlm"},
	{0x992F4DD21FE40F78ULL, "6195834_m.fxe"},
	{0x992F9C6E3B4999C9ULL, "vxlt_0_lowtown_a3_n_c_emv_0_0.vlm"},
	{0x992FD75EFEBA18E8ULL, "vxlt_0_lowtown_gc3_0_0.vlm"},
	{0x993037E7DC88C100ULL, "hlac_ht_keep_4519_14y.dds"},
	{0x9930B6AF105104B4ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_002a_0t.dds"},
	{0x9931564847C22A46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_left_in.ani"},
	{0x99315FFA50047842ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_1.ani"},
	{0x99319FF6505187CCULL, "vxw_bonepit_kq_0_0.vlm"},
	{0x9931A95AC03C67C0ULL, "cn_bdy_glka_0t.dds"},
	{0x9931DFDACC982F0AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a_1.mao"},
	{0x993212E6682A827CULL, "6034365_f.fxe"},
	{0x9932284AE3B8D8CDULL, "6154098_m.fxe"},
	{0x9932CAD6A7105DF0ULL, "6169154_m.fxe"},
	{0x9932DACA066E2104ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_fancy.gad"},
	{0x9933D7545BE5F2E1ULL, "6016500_f.fxe"},
	{0x99342F3F90517440ULL, "ca_caverns1_24422_lc.mmh"},
	{0x99343D80550AE8ADULL, "tools_merchant.mmh"},
	{0x99344F80550B0733ULL, "tools_merchant.msh"},
	{0x99345241AEF258F3ULL, "gen_im_wep_war_two_act3_05.uti"},
	{0x99348A3CD0F48C92ULL, "6171579_m.fxe"},
	{0x9935920C71A7D1F9ULL, "vxht_brothel_n_61y_0_0.vlm"},
	{0x99362909B655CC4BULL, "vxht_mansion2b_1wb_0_0.vlm"},
	{0x99366278ADB37DC4ULL, "vxht_mansion2b_3go_0_0.vlm"},
	{0x99369D25B856D928ULL, "6111001_m.fxe"},
	{0x9936D1A2CDF75FA7ULL, "659627019.ogg"},
	{0x9936F310B8B21089ULL, "hlac_lt_0_lowtown_docks_n_-262144_2x.dds"},
	{0x9936F5E0F458A052ULL, "vxlt_0_lowtown_n_f0j_0_0.vlm"},
	{0x9937306FD6CF4E41ULL, "vxht_chantry_2ow_0_0.vlm"},
	{0x993800E909596424ULL, "follower_banter6170485.cl"},
	{0x99385718466039F9ULL, "vxw_encounter2_gl_0_0.vlm"},
	{0x9938B4B99024A5ABULL, "dae000tr_location_market.utt"},
	{0x9938F376C93FDAC0ULL, "hlac_w_sundermount_262148_0z.dds"},
	{0x99399091B5451062ULL, "hlac_lt_0_lowtown_docks_-2_3x.dds"},
	{0x9939E9CA067459E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_bow_fancy.ani"},
	{0x993A30E06DC5CE18ULL, "vxlt_blackemporium_ih_0_0.vlm"},
	{0x993A336664D07C9DULL, "follower_banter6156275.cl"},
	{0x993A9CA3D953E3BAULL, "vxlt_gallowstemplar_5t5_0_0.vlm"},
	{0x993AA69B96DA0B90ULL, "6108991_m.fxe"},
	{0x993AD162F96E1BC1ULL, "vxdr_0_deeproads_clt_0_0.vlm"},
	{0x993B3551E0305EACULL, "6197865_m.fxe"},
	{0x993B46137DFC4947ULL, "hlac_lt_0_lowtown_docks_65535_3z.dds"},
	{0x993C0750C767B25AULL, "hlac_lt_warehouse_1_0x.dds"},
	{0x993C31115534CC90ULL, "vxht_keep_5hh_0_0.vlm"},
	{0x993C4D47363E2CBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_exit.ani"},
	{0x993CDD13C64A0182ULL, "vxdr_primevaltomb_75z_0_0.vlm"},
	{0x993D4B04A553042AULL, "6149297_m.fxe"},
	{0x993D602225CC2942ULL, "208252162.ogg"},
	{0x993D6B49F4FD3202ULL, "gen00fl_isabela6110215.cl"},
	{0x993DDEFBC85B71AEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basic.mmh"},
	{0x993DEAFBC85B8645ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basic.mao"},
	{0x993DF8FBC85B9D9CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_basic.msh"},
	{0x993E52DD41104DC8ULL, "hlac_lt_gallowscourtyard_-65537_0y.dds"},
	{0x993E88A47717FEEDULL, "mer210_seldrek6078802.cl"},
	{0x993EC5A99281ABA1ULL, "488661536.ogg"},
	{0x993F0B458DAA001BULL, "332494046.ogg"},
	{0x993F33352F72B865ULL, "6106686_m.fxe"},
	{0x993FA3DFDE6FA640ULL, "vxht_keep_n_5hg_0_0.vlm"},
	{0x993FF13F73F90D68ULL, "mcr310_party_banter6087823.cl"},
	{0x9940524DD5FB260EULL, "mag100_cullen6171199.cl"},
	{0x994086BB207238FDULL, "hlac_lt_0_lowtown_docks_-65538_1x.dds"},
	{0x9940A34C35FA5BAEULL, "297288451.ogg"},
	{0x9940EEECB41DB993ULL, "vxht_0_hightown_9b5_0_0.vlm"},
	{0x99419D76A3582E36ULL, "731022391.ogg"},
	{0x9942307C3CD8C7DCULL, "6090838_m.fxe"},
	{0x994263AF9D50FDF2ULL, "hightown_fem_noble1_sta6174604.cl"},
	{0x99430DBE33CED001ULL, "vxlt_0_lowtown_n_h4g_0_0.vlm"},
	{0x994319DF52A81BB4ULL, "206814104.ogg"},
	{0x9943217E23E0F8D0ULL, "vxlt_0_lowtown_a3_n_c_hkt_0_0.vlm"},
	{0x99435A47364461FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_exit.gad"},
	{0x9943D1591591FCCEULL, "vxlt_0_lowtown_n_eqh_0_2.vlm"},
	{0x9943D293262B605AULL, "13513054.ogg"},
	{0x994466C76323DAC9ULL, "6154285_m.fxe"},
	{0x9944E68EB9C08E78ULL, "hlac_lt_undercityserialkiller_-1_29y.dds"},
	{0x9945106B9E75DDE5ULL, "709732133.ogg"},
	{0x99451E6DDCB07018ULL, "isa311_isabela6092304.cl"},
	{0x994536805519506AULL, "tools_merchant.phy"},
	{0x9945390DD9C3F833ULL, "hlac_w_encounter3_-4_0y.dds"},
	{0x99463F4847D45F40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_left_in.gad"},
	{0x9946A1B2C9D6577DULL, "vfx_vfx_p4.vsh.fxo.dx11.assembly"},
	{0x9946D7DF7A96BBB9ULL, "hlac_lt_gallowstemplar_f_2394_8x.dds"},
	{0x994711D1F3B13EA0ULL, "6153991_m.fxe"},
	{0x99471465F0CB51AFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rockpile.mao"},
	{0x99472065F0CB664CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rockpile.mmh"},
	{0x99472665F0CB703EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rockpile.msh"},
	{0x99475B0222840919ULL, "dae000_gamlen6102412.cl"},
	{0x99475BDA222FCDA2ULL, "isa311_isabela6091826.cl"},
	{0x994764DE0E826B32ULL, "ht_0_hightown_51_0.rml"},
	{0x9947E0D985CBF42CULL, "6051617_m.fxe"},
	{0x9947E9029520E712ULL, "hlac_lt_gallowstemplar_f_2388_2x.dds"},
	{0x9947EBA9D5D8AF8EULL, "vxw_bonepitdestroyed_a3_1bg_0_0.vlm"},
	{0x9948866E0EA0F88FULL, "ser210tr_amb_witness.ncs"},
	{0x9948E43C5510206AULL, "nat320_delilah6065933.cl"},
	{0x99498E99A34860C5ULL, "plt_cod_ltr_ggt_note2.nss"},
	{0x9949BC3F9A4B6ADFULL, "vxlt_hangedman_1vb_0_1.vlm"},
	{0x9949C1B990332F5AULL, "dae000tr_location_market.ncs"},
	{0x994A3547364A411DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_exit.evt"},
	{0x994B280599C2BD3CULL, "718151971.ogg"},
	{0x994B35966B876E7AULL, "6032285_m.fxe"},
	{0x994B5EAB347BB0DDULL, "413185539.ogg"},
	{0x994B7B4E325A3D14ULL, "6208478_f.fxe"},
	{0x994B7CB1A1520924ULL, "6053747_m.fxe"},
	{0x994B966A1979074FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1a.phy"},
	{0x994BDA5156AA0516ULL, "vxlt_gallowstemplar_f_4o2_0_0.vlm"},
	{0x994C2A0D913F69DEULL, "vxlt_undercityserialkiller_1ri_0_0.vlm"},
	{0x994C860AB9E8AA86ULL, "mag110_samson6206391.cl"},
	{0x994C9A9DE1242211ULL, "6026971_m.fxe"},
	{0x994D0FB7856F9E44ULL, "lt_0_lowtown_docks_10_0.rml"},
	{0x994D3DAF8FD820EDULL, "6170588_m.fxe"},
	{0x994D7C1BB6443857ULL, "6136944_m.fxe"},
	{0x994D964B0A17C49FULL, "vxca_caverns1_mines_jfs_0_0.vlm"},
	{0x994DEF61FB2FCB6DULL, "6102810_f.fxe"},
	{0x994E424225714514ULL, "hightown_market_guard_sta.cnv"},
	{0x994F011B5A70A1F3ULL, "6029051_m.fxe"},
	{0x994F1E366F2B34C2ULL, "hlac_ca_cave_2194_7x.dds"},
	{0x994F32858D778FE0ULL, "404742775.ogg"},
	{0x994F897324806385ULL, "mag330_walter6045984.cl"},
	{0x994FBABC296598A2ULL, "730155935.ogg"},
	{0x994FC30FE0CE589CULL, "584571598.ogg"},
	{0x994FD50DE4DF57C2ULL, "hlac_ca_caverns1_24423_0z.dds"},
	{0x995027C670C76664ULL, "6109022_m.fxe"},
	{0x99511104A0297150ULL, "6122160_m.fxe"},
	{0x99512045490CAA28ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpouta.msh"},
	{0x99512525F03E0D55ULL, "302430115.ogg"},
	{0x99513245490CC8B9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpouta.mao"},
	{0x99513E45490CDD52ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpouta.mmh"},
	{0x9951F325973E4440ULL, "6102067_m.fxe"},
	{0x99520F682B510B55ULL, "vxlt_0_lowtown_a2_n_c_er6_0_0.vlm"},
	{0x995224422574FB54ULL, "hightown_market_guard_sta.bnk"},
	{0x9952934147F679AEULL, "215987581.ogg"},
	{0x995313A143852540ULL, "vxca_caverns2_80w_0_1.vlm"},
	{0x99532C39F58C6DECULL, "432714454.ogg"},
	{0x99533FC5D9FD7C48ULL, "vxht_keep_6ux_0_0.vlm"},
	{0x99534096A3024ADDULL, "101895380.ogg"},
	{0x995518E7676F75BEULL, "hlac_ca_caverns1_mines_24427_3x.dds"},
	{0x99554463D3A9C547ULL, "vxht_mansion2b_27i_0_0.vlm"},
	{0x9955B4F52F2A216FULL, "mcr370_cut_player_survives6098272.cl"},
	{0x995621188BD0B168ULL, "661795381.ogg"},
	{0x99565D86C7CEC80EULL, "vxw_bonepitdestroyed_a3_1gx_0_0.vlm"},
	{0x995698CBD166D11FULL, "395143967.ogg"},
	{0x9956AE8F7471E6D8ULL, "6156325_m.fxe"},
	{0x9956FAFF8DFB08DFULL, "vxht_0_hightown_5i8_0_0.vlm"},
	{0x9957A8CA15AA005FULL, "gen00pt_random.ncs"},
	{0x9957FC081E9E1B69ULL, "dae021_torelin_mob6150216.cl"},
	{0x995820243BDD3B06ULL, "zz_jp_enters.cut"},
	{0x99585423E9BBCAAEULL, "6015282_m.fxe"},
	{0x99585B152903C288ULL, "follower_banter6170597.cl"},
	{0x9958F82FD8FA3EEEULL, "art\\characters\\weapons\\shortbows\\w_sbw_dssba.mmh"},
	{0x9959122FD8FA6ADCULL, "art\\characters\\weapons\\shortbows\\w_sbw_dssba.msh"},
	{0x99595012DE38C09DULL, "vxlt_undercityserialkiller_193_0_0.vlm"},
	{0x995996D855EC331CULL, "659803823.ogg"},
	{0x995A43DA76D10EBFULL, "ave100_donnic6105934.cl"},
	{0x995A4D3876A0A457ULL, "533289320.ogg"},
	{0x995A69DF31FD2675ULL, "196846573.ogg"},
	{0x995A96569E2A89E4ULL, "ht_0_hightown_21_0_lc.mmh"},
	{0x995A9BEA8335825EULL, "vxlt_undercityserialkiller_1wj_0_0.vlm"},
	{0x995AACF671E77BD8ULL, "6094477_f.fxe"},
	{0x995B48C198DC6F20ULL, "6110381_m.fxe"},
	{0x995B56D134261645ULL, "6019111_m.fxe"},
	{0x995B6263DAFE0E45ULL, "vxlt_warehouse_hv_0_0.vlm"},
	{0x995C190B48137810ULL, "vxht_0_hightown_c73_0_0.vlm"},
	{0x995C1E3CF4B92FA0ULL, "1034057453.ogg"},
	{0x995CAADB631D0DA1ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scroll.mao"},
	{0x995D411E22CCA065ULL, "6163232_m.fxe"},
	{0x995DBF750A042055ULL, "vxca_caverns2prc_5zy_0_0.vlm"},
	{0x995E0A6F4001DC32ULL, "102073304.ogg"},
	{0x995E63BC14BD7752ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\appear.gad"},
	{0x995E9F5A436ADF7AULL, "6034815_f.fxe"},
	{0x995F1AA48E929EE9ULL, "710766747.ogg"},
	{0x995FDEFCF4F7B963ULL, "782520546.ogg"},
	{0x995FF317CC2A69F8ULL, "vxca_caverns1_mines_j25_0_1.vlm"},
	{0x996020241CCB10F2ULL, "6207701_f.fxe"},
	{0x99604C788905E9C6ULL, "vxht_chantry_26w_0_0.vlm"},
	{0x996059F928F0F9D9ULL, "6122860_m.fxe"},
	{0x99610F72FD1BEA55ULL, "274906656.ogg"},
	{0x99612B97EFE5A43CULL, "vxlt_0_lowtown_docks_n_bsx_0_0.vlm"},
	{0x99614A9D9394BA7AULL, "19091857.ogg"},
	{0x996158D2AFEF92ECULL, "370737579.ogg"},
	{0x9961B9DCDC5D0107ULL, "vxlt_gallowstemplar_3r8_0_0.vlm"},
	{0x9961BE3533FDC416ULL, "follower_banter6141616.cl"},
	{0x9961D5E4D88BB626ULL, "6123327_m.fxe"},
	{0x99627C912456D4E4ULL, "6104875_f.fxe"},
	{0x99627D17DFBF5A48ULL, "gen_im_trs_qun_two_2hm_r1.uti"},
	{0x9962AD5136FDAD0EULL, "vxdr_primevaltomb_8nd_0_0.vlm"},
	{0x9962EBDB86CA2676ULL, "vxca_caverns1_iia_0_0.vlm"},
	{0x9963663A79AAD386ULL, "599309032.ogg"},
	{0x996375E0EBC30CBBULL, "376132607.ogg"},
	{0x9963F192A0469203ULL, "6153790_m.fxe"},
	{0x9963F28733566B3FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_enter.gad"},
	{0x996403D5FBB24277ULL, "vxlt_0_lowtown_n_ew9_0_0.vlm"},
	{0x996444B408476233ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_rnf.evt"},
	{0x99646176725E6B0AULL, "art\\vfx\\combat\\vfx_r_brd_shout_lar_imp_p.mmh"},
	{0x996466BC14C1E789ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\appear.evt"},
	{0x9964CF63B9C4E08EULL, "vxw_bonepitdestroyed_a3_1hy_0_0.vlm"},
	{0x99653C4E2E163678ULL, "397520543.ogg"},
	{0x9965CEA8598849AAULL, "dae000_party_night_comment6201650.cl"},
	{0x9965DAFE458ACB5FULL, "vrc321ip_projectile_source.ncs"},
	{0x996626ED81963703ULL, "vxht_mansion_n_5ca_0_0.vlm"},
	{0x99663460EFCCC663ULL, "w_blightlands_225_0.mao"},
	{0x99663860EFCCCCB0ULL, "w_blightlands_225_0.mmh"},
	{0x99664660EFCCE4CAULL, "w_blightlands_225_0.msh"},
	{0x99664F3CC8B23999ULL, "cs_mcr370cs_arrival_gallows.wbd"},
	{0x996688E70CC5DE54ULL, "gen00fl_fenris6126843.cl"},
	{0x99669135F203460BULL, "156206461.ogg"},
	{0x9966EEAD6B8177B6ULL, "and341ar_mines.are"},
	{0x996713C2DA48683DULL, "gen00fl_varric6093681.cl"},
	{0x996719CF833C1DFBULL, "6104779_f.fxe"},
	{0x99676267C562DF19ULL, "hlac_lt_0_lowtown_a2_n_c_-1_8y.dds"},
	{0x9967736E0EBB8936ULL, "ser210tr_amb_witness.utt"},
	{0x9968237E4AB3EC43ULL, "6096488_m.fxe"},
	{0x996842CF4CF73E7BULL, "6162761_m.fxe"},
	{0x996866CA8D942C91ULL, "462870323.ogg"},
	{0x9968774B57E6D3E6ULL, "gen00fl_dog6115112.cl"},
	{0x99690BAF08C881A8ULL, "6175664_f.fxe"},
	{0x99696F6A79C8101FULL, "326384867.ogg"},
	{0x99697AAF0685A702ULL, "hlac_lt_0_lowtown_a2_n_131071_6y.dds"},
	{0x996A9A9F3BA4D7D7ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic.phy"},
	{0x996AE787335C772CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_enter.evt"},
	{0x996B21F11642177AULL, "6140621_m.fxe"},
	{0x996B3DB4084D7574ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_rnf.gad"},
	{0x996B4918AA0EC39BULL, "1056117817.ogg"},
	{0x996B5B48C09AACE9ULL, "vxht_keep_5gp_0_0.vlm"},
	{0x996B898E62C9F549ULL, "6084355_m.fxe"},
	{0x996BC1C1C23DBC84ULL, "follower_banter6163618.cl"},
	{0x996C08F369C97727ULL, "lgt311_shw_mer_agatha6174288.cl"},
	{0x996C41C7683C961DULL, "6081654_f.fxe"},
	{0x996CCCBEE7798DBBULL, "ser191_emeric6120050.cl"},
	{0x996D54CAECA47CDCULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_dummy.mao"},
	{0x996D60CAECA4917FULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_dummy.mmh"},
	{0x996D66CAECA49B21ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_dummy.msh"},
	{0x996E197BB526D439ULL, "vxdr_primevaltomb_6v9_0_0.vlm"},
	{0x996E21E4F806DBFCULL, "167109215.ogg"},
	{0x996E2E79AAFB4CC6ULL, "llac_lt_0_lowtown_n_131072_6.dds"},
	{0x996E2F4C17EAE189ULL, "vxw_bonepitdestroyed_a3_2af_0_0.vlm"},
	{0x996EAB93A4625C0CULL, "6031714_m.fxe"},
	{0x996EBE557AFC0246ULL, "vxht_chantry_23x_0_0.vlm"},
	{0x996EE56B7F2E6F86ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.cess.ani"},
	{0x996EEDA2B30FCA5DULL, "57947489.ogg"},
	{0x996EEE64D7AC1FA6ULL, "6076300_m.fxe"},
	{0x996F025459D1A667ULL, "blood_splatter_04.dds"},
	{0x996F3113A1D8A143ULL, "llac_lt_0_lowtown_docks_a3_n_-2_6.dds"},
	{0x996F6A211F77D87DULL, "mer120_hubert6059423.cl"},
	{0x996FD8363BB88EA1ULL, "hlac_ht_0_hightown_n_131073_0z.dds"},
	{0x996FF48F9C3814A2ULL, "zz_dae_debug6029428.cl"},
	{0x997056B3C18CF1C8ULL, "ave200_party_comment6116364.cl"},
	{0x99715DB878C03EF6ULL, "vxca_caverns1_ilb_0_0.vlm"},
	{0x9971B37B70C32136ULL, "lgt320_shw_ors_secret_meet6156357.cl"},
	{0x9971D2AD45EE1B79ULL, "6107205_m.fxe"},
	{0x9971F1B4DCCB5E69ULL, "331617721.ogg"},
	{0x99724CB40853AD52ULL, "art\\characters\\creatures\\spider\\animation\\p_tn180r_rnf.ani"},
	{0x99724EBC14CDFCF8ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\appear.ani"},
	{0x9972636F923CAA83ULL, "6156490_m.fxe"},
	{0x9972786DDD07A990ULL, "6152116_m.fxe"},
	{0x9972C20A7E9D0809ULL, "dr_primevaltomb_661.rml"},
	{0x9972FF12AFB6D211ULL, "dae000cr_grand_cleric6124739.cl"},
	{0x997331BA27E0E0E1ULL, "vxw_encounter3_ee_0_0.vlm"},
	{0x99734140ABBAF90EULL, "vxw_bonepitdestroyed_a3_1mz_0_0.vlm"},
	{0x997392D23AD65B1EULL, "240558796.ogg"},
	{0x9973C68CC6E972D2ULL, "s_proj_link.mao"},
	{0x99746F06327DD93FULL, "6121148_m.fxe"},
	{0x9974EB7778FADB01ULL, "w_bonepitdestroyed_a3_169_0.phy"},
	{0x9974F0EA70DEF11CULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_2y.dds"},
	{0x99761EB9C753D7CEULL, "802112996.ogg"},
	{0x997630F2951AC02EULL, "hlac_ht_mansion2_818_24z.dds"},
	{0x99766D519376DA9EULL, "m_mer141ar_caverns.bnk"},
	{0x997672076DEC2DE2ULL, "vxlt_0_lowtown_n_eox_0_0.vlm"},
	{0x99775FB529C3F5EAULL, "hlac_w_bonepitmining_a2_0_1y.dds"},
	{0x99778EAD499C392AULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgte.mao"},
	{0x997792AD499C4079ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgte.mmh"},
	{0x99779CAD499C5167ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgte.msh"},
	{0x9977B6499C6FF301ULL, "6037075_f.fxe"},
	{0x9977F910AC03ADB8ULL, "6205593_m.fxe"},
	{0x9978111397B7BDAEULL, "215906366.ogg"},
	{0x9978461DBEEA3F45ULL, "vxlt_warehouse_vt_0_0.vlm"},
	{0x9978CF8733688B85ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_enter.ani"},
	{0x9979362331BEC251ULL, "mag311_emile_de_launcet6064722.cl"},
	{0x997982C341C60E6FULL, "hlac_lt_0_lowtown_a2_n_65537_3x.dds"},
	{0x9979B50A9406EE16ULL, "495248165.ogg"},
	{0x9979CD25B290C569ULL, "vxht_keep_5jw_0_0.vlm"},
	{0x997A36A4187A8B71ULL, "337113007.ogg"},
	{0x997A7154350BC013ULL, "6016660_m.fxe"},
	{0x997B8266A81D0A15ULL, "785755440.ogg"},
	{0x997B883C52389E67ULL, "6139989_m.fxe"},
	{0x997C01CD4CF2150FULL, "qun110_petrice_ketojan.stg"},
	{0x997CCD6B7F3A84F7ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.cess.evt"},
	{0x997CFC4BA9656426ULL, "310711944.ogg"},
	{0x997D0E3484BF497BULL, "alphaclouds01.mao"},
	{0x997DFBB8AEBCC387ULL, "118151677.ogg"},
	{0x997E502D89CC662EULL, "lt_0_lowtown_docks_a3_249_0.mmh"},
	{0x997E5C2D89CC7AC5ULL, "lt_0_lowtown_docks_a3_249_0.mao"},
	{0x997E6A2D89CC921CULL, "lt_0_lowtown_docks_a3_249_0.msh"},
	{0x997E8D648CAB082CULL, "hlac_ht_0_hightown_-196610_0y.dds"},
	{0x997EAA4A5CD550AEULL, "545117759.ogg"},
	{0x997EC69CA8AE2267ULL, "vxca_cavern1_haunted_cka_0_0.vlm"},
	{0x997F0394F83276BEULL, "6151490_m.fxe"},
	{0x997F0C14557EC6DEULL, "fen210_fenris6108740.cl"},
	{0x997F910B1AE9DE0EULL, "vxdr_primevaltomb_8df_0_0.vlm"},
	{0x997F9DE60264B507ULL, "6135831_m.fxe"},
	{0x997FB31793925D71ULL, "disabled_disarm_pressed.cur"},
	{0x997FC5CDD762483DULL, "wm_ico_woundedcoast_roll.dds"},
	{0x997FE3E874B79BBBULL, "564024547.ogg"},
	{0x99804BE1ED365E03ULL, "vxlt_hangedman_1pa_0_0.vlm"},
	{0x998058F9467D097BULL, "245284594.ogg"},
	{0x9980BAAA6BAB9902ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06b.phy"},
	{0x9980CF5EE248144FULL, "vxca_caverns2_7rd_0_0.vlm"},
	{0x9981BD63AAD8E3E8ULL, "lt_0_lowtown_a2_n_21_0_ll.mmh"},
	{0x99825D0008651935ULL, "119838805.ogg"},
	{0x99826A3127944898ULL, "vxlt_blackemporium_ve_0_0.vlm"},
	{0x9982B623D741B8BCULL, "325875167.ogg"},
	{0x9982CE6B7F3EF180ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.cess.gad"},
	{0x9982D6071F480E50ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate.mao"},
	{0x9982DA071F481563ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate.mmh"},
	{0x9982E58B4B42FA71ULL, "one300_zevran6089752.cl"},
	{0x9982E8071F482CBDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate.msh"},
	{0x99834376A0690E10ULL, "711929674.ogg"},
	{0x99837E83E0813261ULL, "vxht_keep_6aq_0_0.vlm"},
	{0x9983A692F698B0A3ULL, "6040865_m.fxe"},
	{0x9983C9C01AEAFE7DULL, "vxlt_hangedman_n_2id_0_0.vlm"},
	{0x9983D94862B5E7C4ULL, "dae000_hate_follower_react6187866.cl"},
	{0x998459EF2AB0E058ULL, "21212291.ogg"},
	{0x9984E3E45FE24662ULL, "vxlt_0_lowtown_n_ehy_0_0.vlm"},
	{0x998580F9389A6AE4ULL, "431433316.ogg"},
	{0x9985FDF31C693B18ULL, "mrl150_merrill6078332.cl"},
	{0x998682CBD3D550C3ULL, "6136157_m.fxe"},
	{0x9986AF6ACFA2926BULL, "1053802619.ogg"},
	{0x9987126F748F6C1CULL, "art\\characters\\creatures\\desiredemon\\animation\\appr.ani"},
	{0x9987758116260FECULL, "vxlt_backalley_n_1pa_0_1.vlm"},
	{0x9987EB12FC8AEE75ULL, "vxlt_undercitydungeon_16w_0_0.vlm"},
	{0x9987FD6408B8A597ULL, "vxlt_0_lowtown_docks_n_c4c_0_0.vlm"},
	{0x99887369C4A8A6D8ULL, "6036229_m.fxe"},
	{0x99889D12199A2755ULL, "zz_one_debug6069467.cl"},
	{0x9988BA92C990FCE5ULL, "538240458.ogg"},
	{0x9988DFE6548E4677ULL, "llac_w_blightlands_327682_0.dds"},
	{0x9988E50FC60F54B6ULL, "vxdr_primevalentrance_1qd_0_0.vlm"},
	{0x99891DFA88B24FFCULL, "591313305.ogg"},
	{0x998922B15598C620ULL, "6071026_m.fxe"},
	{0x998935C4EB9C4BCCULL, "794754862.ogg"},
	{0x9989479A6CEFED8AULL, "6048576_m.fxe"},
	{0x9989CEB92ECA04A9ULL, "vxht_0_hightown_a2_bu9_0_0.vlm"},
	{0x998A134702D9E7D8ULL, "ser221_party_comments6209891.cl"},
	{0x998B42ADB1880D62ULL, "vxht_mansion2_3gp_0_1.vlm"},
	{0x998B5B2650E03804ULL, "follower_banter6172335.cl"},
	{0x998BF5AD6BA09F81ULL, "and341ar_mines.lst"},
	{0x998CD64653EFA82AULL, "default_test_area.lst"},
	{0x998D0739A1D5DE48ULL, "vxht_keep_6at_0_0.vlm"},
	{0x998D61F3CAEDB5B6ULL, "6116994_m.fxe"},
	{0x998D6C258204A6E7ULL, "vxht_keep_n_706_0_0.vlm"},
	{0x998DC7B93DF13165ULL, "6115344_f.fxe"},
	{0x998E0785FB82BC27ULL, "80746011.ogg"},
	{0x998E1F6F7495A15EULL, "art\\characters\\creatures\\desiredemon\\animation\\appr.gad"},
	{0x998EC3A943411C42ULL, "vxht_keep_5xz_0_0.vlm"},
	{0x998F229A92111AD0ULL, "592057466.ogg"},
	{0x998F253F1CB2ACFFULL, "rand_denerim.gda"},
	{0x998F64409EEC7BE4ULL, "art\\vfx\\combat\\vfx_str_scream_act_c.anb"},
	{0x9990382A78E2BC93ULL, "character_character_p109.psh.fxo.dx11"},
	{0x9990916810944632ULL, "vxht_mansion2b_30x_0_0.vlm"},
	{0x9990EB4EF6010EA5ULL, "806866818.ogg"},
	{0x9991E14F40056EF0ULL, "ht_0_hightown_a3_54_0.rml"},
	{0x99920737EE3048F4ULL, "553762923.ogg"},
	{0x999280A98E54A9ADULL, "mag320_orsino6071196.cl"},
	{0x999280BE0660E678ULL, "vxht_brothel_n_3ss_0_0.vlm"},
	{0x9992965F2247E989ULL, "6134801_m.fxe"},
	{0x99929B25BCBBC0E8ULL, "6095103_m.fxe"},
	{0x999307E2FB519C5FULL, "vxlt_gallowsprison_n_6n2_0_0.vlm"},
	{0x99932E4F2EF40913ULL, "937792697.ogg"},
	{0x999345B17ECCC5E9ULL, "gen00fl_merrill6128958.cl"},
	{0x999381CD768A8559ULL, "mage_glyphparalyze.dds"},
	{0x999387EDD4AA478FULL, "6147672_m.fxe"},
	{0x9993AC089462F2EBULL, "hlac_dr_dungeon_3214_5z.dds"},
	{0x9993DB071F568988ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate.phy"},
	{0x9994B3D05EE85701ULL, "444813517.ogg"},
	{0x9994FA6F749B817DULL, "art\\characters\\creatures\\desiredemon\\animation\\appr.evt"},
	{0x99957BCE77AA8828ULL, "44723325.ogg"},
	{0x9996DCF2645A95EBULL, "vxht_mansion_n_6b5_0_0.vlm"},
	{0x99970DF6839EB303ULL, "art\\characters\\playercharacter\\humanmale\\rob_appc.mao"},
	{0x999711F6839EBA50ULL, "art\\characters\\playercharacter\\humanmale\\rob_appc.mmh"},
	{0x99971FF6839ED26AULL, "art\\characters\\playercharacter\\humanmale\\rob_appc.msh"},
	{0x999731CE0F5989E1ULL, "vxht_0_hightown_n_byn_0_0.vlm"},
	{0x99975828B6E90A31ULL, "vxca_caverns1_cju_0_0.vlm"},
	{0x9997FF3D0F47A550ULL, "6035310_m.fxe"},
	{0x9998562B6CEE6FACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_reaction.ani"},
	{0x9998F003833E9EC9ULL, "vxlt_backalley_g_1yj_0_1.vlm"},
	{0x999992B342B08511ULL, "hlac_lt_0_lowtown_-65536_4x.dds"},
	{0x9999B51FA56F70E9ULL, "vxht_0_hightown_c58_0_0.vlm"},
	{0x999A121BE720003CULL, "hlac_w_bonepitdestroyed_a3_393216_0z.dds"},
	{0x999A338018DDCA45ULL, "156080535.ogg"},
	{0x999A87A6863A6D31ULL, "llac_w_bonepitdestroyed_a3_327683_0.dds"},
	{0x999BB557052E021AULL, "6086286_m.fxe"},
	{0x999BD66B26F9F0D3ULL, "146065647.ogg"},
	{0x999C0877DAC7DE5CULL, "238248324.ogg"},
	{0x999C12913D5A129AULL, "vxht_0_hightown_a7o_0_0.vlm"},
	{0x999C336D82F58631ULL, "1055555429.ogg"},
	{0x999C83409EF73AE7ULL, "art\\vfx\\combat\\vfx_str_scream_act_c.mmh"},
	{0x999D11D0B3C54A91ULL, "dae000_fem_commoner2_mob6174581.cl"},
	{0x999D39E45F68790CULL, "mag200_feynriel6023960.cl"},
	{0x999DD982765F52DFULL, "vxht_mansion_n_4z3_0_0.vlm"},
	{0x999DDA8AD84D53DEULL, "lla_ht_0_hightown_a2_65538_1.dds"},
	{0x999E0CEA94273D1DULL, "6110501_f.fxe"},
	{0x999E31A624081221ULL, "dae021_sister_alema_mob6144774.cl"},
	{0x999E5EDE7910BED1ULL, "684951709.ogg"},
	{0x999EB5099A5C873FULL, "zz_one_debug6058977.cl"},
	{0x999EB59EA88B4179ULL, "284242645.ogg"},
	{0x999F632B6CF4A4EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_reaction.gad"},
	{0x999F6836781EB741ULL, "53946886.ogg"},
	{0x999FE70F671E063FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_doubledoor.msh"},
	{0x999FED0F671E10E1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_doubledoor.mmh"},
	{0x999FF90F671E2482ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_doubledoor.mao"},
	{0x99A051FD9AA80A7AULL, "6050181_f.fxe"},
	{0x99A0672456C58742ULL, "6122504_m.fxe"},
	{0x99A06C431776658BULL, "vxht_0_hightown_n_c91_0_0.vlm"},
	{0x99A11FFA7359C9F2ULL, "6129435_m.fxe"},
	{0x99A1849A97418761ULL, "609396707.ogg"},
	{0x99A18949D3E690F3ULL, "6051892_m.fxe"},
	{0x99A29568074CBDC5ULL, "mcr310_sibling6076382.cl"},
	{0x99A3562641AFE092ULL, "llac_w_encounter1_65534_0.dds"},
	{0x99A3E3621A2AD124ULL, "6202174_m.fxe"},
	{0x99A400FA579D8199ULL, "vxlt_gallowstemplar_n_35d_0_0.vlm"},
	{0x99A43D0FCBCF6C1CULL, "595169786.ogg"},
	{0x99A44A754457B0D8ULL, "306508709.ogg"},
	{0x99A44E71CAAD9C85ULL, "6036798_m.fxe"},
	{0x99A4834579FA1F9AULL, "zz_followers_banter6158757.cl"},
	{0x99A493A28C1FA625ULL, "476193131.ogg"},
	{0x99A4A597C38E387EULL, "284376848.ogg"},
	{0x99A510CFDB1118BFULL, "237651115.ogg"},
	{0x99A5A506D102B1DFULL, "hlac_w_blightlands_-65537_0y.dds"},
	{0x99A5C57DB470B8FBULL, "hlac_dr_primevaltomb_657_3z.dds"},
	{0x99A5D4AD5CFC1009ULL, "vxdr_primevaltomb_9eg_0_0.vlm"},
	{0x99A5F436BE4D4075ULL, "60038529.ogg"},
	{0x99A65E2B6CFABACDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_reaction.evt"},
	{0x99A6A1AA6BCC3E55ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06b.mmh"},
	{0x99A6A3AA6BCC41CBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06b.msh"},
	{0x99A725E4E64C7092ULL, "vxlt_0_lowtown_docks_a3_9rb_0_0.vlm"},
	{0x99A77590764A9329ULL, "6029138_m.fxe"},
	{0x99A786F38A21B49BULL, "one350_vinell6087487.cl"},
	{0x99A79FFBBCB88613ULL, "358430923.ogg"},
	{0x99A7B5161185C9B5ULL, "vxlt_0_lowtown_docks_n_bff_0_0.vlm"},
	{0x99A7C71144B3A0CBULL, "one100_party_comments6133500.cl"},
	{0x99A7D3457E278CCEULL, "dae000ip_crowd_ff1_talk4.utp"},
	{0x99A7DCED0D4B22CAULL, "ts_beasts_spiders_03.gda"},
	{0x99A7DEB896F1837EULL, "vxht_bartrand_n_3r2_0_0.vlm"},
	{0x99A8500C2D2F22B6ULL, "6126755_m.fxe"},
	{0x99A8A0C1312ED9DAULL, "704945254.ogg"},
	{0x99A91FE84D988D46ULL, "6201582_m.fxe"},
	{0x99A9F3ED0A3A1591ULL, "487265258.ogg"},
	{0x99AA1C462B1D21C0ULL, "vxw_bonepit_20t_0_0.vlm"},
	{0x99AA6781D4F5BCCEULL, "6134752_m.fxe"},
	{0x99AA887BB9607EFEULL, "vxlt_gallowstemplar_n_5q9_0_0.vlm"},
	{0x99AB1726067FCAF2ULL, "qcr220_orsino6060731.cl"},
	{0x99AB18E8AF7C467EULL, "439971625.ogg"},
	{0x99AB4834FF5DE50FULL, "hlac_lt_0_lowtown_a2_-131072_8y.dds"},
	{0x99ABA773F50D06D5ULL, "vxlt_0_lowtown_a2_n_c_hox_0_0.vlm"},
	{0x99ABB8EAE17D8BC1ULL, "45678289.ogg"},
	{0x99ABCFB29572D69AULL, "hlac_lt_darktown_-1_30y.dds"},
	{0x99ACB58C391316B2ULL, "8167.ogg"},
	{0x99AD09DAB86AA2BBULL, "vxw_bonepitmining_a2_fo_0_0.vlm"},
	{0x99AD4BACAFD18ED6ULL, "vxlt_0_lowtown_epc_0_1.vlm"},
	{0x99AD9438126C0C2CULL, "211103697.ogg"},
	{0x99AE215171E4AB22ULL, "zz_approval_debug6003568.cl"},
	{0x99AE446FB64685F5ULL, "342356550.ogg"},
	{0x99B09B7E80AB892AULL, "6082616_f.fxe"},
	{0x99B11A3D1E8F135BULL, "6120763_m.fxe"},
	{0x99B15D726A629E6AULL, "vxht_0_hightown_a2_brs_0_0.vlm"},
	{0x99B18ACD1D7770FBULL, "vxht_mansion2_1wj_0_0.vlm"},
	{0x99B1939744D7F9C6ULL, "vxlt_gallowscourtyard_n_54h_0_0.vlm"},
	{0x99B19728A0BB6749ULL, "gen00fl_fenris6108093.cl"},
	{0x99B1ACDED75A5851ULL, "plt_lgt000pt_ultra_4_05.nss"},
	{0x99B1BF8266BDC38EULL, "lgt210_magu_selby6142051.cl"},
	{0x99B23B292CDCA562ULL, "6143273_m.fxe"},
	{0x99B23EA8950DCED2ULL, "268861339.ogg"},
	{0x99B26F9F3BE2442FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic.mao"},
	{0x99B27B9F3BE258CCULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic.mmh"},
	{0x99B2819F3BE262BEULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic.msh"},
	{0x99B2D5465410758DULL, "default_test_area.are"},
	{0x99B302FD4DD3E999ULL, "6088150_f.fxe"},
	{0x99B308E8D1F6732EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightownbooks_sa.dds"},
	{0x99B359604930F90EULL, "art\\vfx\\combat\\vfx_g_dual_att_right_c.mmh"},
	{0x99B366E0BA226173ULL, "gen00fl_merrill6128732.cl"},
	{0x99B3CE1D83C03365ULL, "747755316.ogg"},
	{0x99B3D35F955B7A38ULL, "743145417.ogg"},
	{0x99B3F075AED13265ULL, "lgt220cr_mag_apost_apost2.utc"},
	{0x99B4FCFCE11BE6FAULL, "6172259_m.fxe"},
	{0x99B52396DB1D53A4ULL, "hlac_lt_0_lowtown_131072_3x.dds"},
	{0x99B60ECF94BA4211ULL, "291811753.ogg"},
	{0x99B626F3037BE235ULL, "vxlt_0_lowtown_docks_n_bce_0_0.vlm"},
	{0x99B68939290541E4ULL, "hlac_ca_caverns1_24421_8z.dds"},
	{0x99B6B9DF98EAE9EEULL, "follower_banter6172706.cl"},
	{0x99B6CA923ABECA8EULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglehead.msh"},
	{0x99B6D8923ABEE2DFULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglehead.mao"},
	{0x99B6E4923ABEF77CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_eaglehead.mmh"},
	{0x99B7371D4D9949E7ULL, "lowtown_fem_cityguard_mob6164809.cl"},
	{0x99B73F5193AE23A4ULL, "m_mer141ar_caverns.wbd"},
	{0x99B74EFC86CCE105ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgta_0n.dds"},
	{0x99B81E6AF415AD73ULL, "6049228_m.fxe"},
	{0x99B825659B76ECB4ULL, "vxlt_backalley_n_1ky_0_0.vlm"},
	{0x99B893E11F12454FULL, "gen00fl_aveline6114051.cl"},
	{0x99B8C158B3AE064BULL, "gen00fl_varric6199022.cl"},
	{0x99B906B8B9A50FC9ULL, "dae000ip_crowd_lm2_talk7.utp"},
	{0x99B94D1B14168145ULL, "182376565.ogg"},
	{0x99B9587EE2C2400EULL, "vxdr_primevaltomb_8pz_0_0.vlm"},
	{0x99B98E5739BB7C21ULL, "vxlt_warehouse_n_14h_0_0.vlm"},
	{0x99B9C5A17F586DF2ULL, "vxdr_primevaltomb_7m2_0_0.vlm"},
	{0x99B9DA4A33B20BECULL, "725828077.ogg"},
	{0x99B9F009180C2052ULL, "6129138_m.fxe"},
	{0x99BA35A47F3DCB64ULL, "6054845_m.fxe"},
	{0x99BA641A3646720CULL, "one200ip_st_solivitus.utp"},
	{0x99BAAB78DACA8EE2ULL, "hlac_lt_0_lowtown_a2_n_-131071_2z.dds"},
	{0x99BB22F1AB3D0CACULL, "hlac_lt_0_lowtown_n_131071_6y.dds"},
	{0x99BB7BB7AA60BB3BULL, "vxw_bonepitmining_a2_mn_0_0.vlm"},
	{0x99BC25A78547DCD4ULL, "zz_mrl_debug6184180.cl"},
	{0x99BC5A9518C014DCULL, "413533016.ogg"},
	{0x99BD2ACF49CAE130ULL, "622466877.ogg"},
	{0x99BDA164B58D9A1DULL, "vxdr_0_deeproads_ewc_0_0.vlm"},
	{0x99BDB7EF2941067BULL, "321908783.ogg"},
	{0x99BDD467EE7A7E98ULL, "hlac_lt_undercitydungeon_-1_26y.dds"},
	{0x99BDF78891C0E517ULL, "follower_barks6169900.cl"},
	{0x99BE57D7AB0B3389ULL, "vxlt_hangedman_2xa_0_1.vlm"},
	{0x99BE5CFAC6B6EB8AULL, "6180463_m.fxe"},
	{0x99BEAC0FABEE65C3ULL, "gen00fl_merrill6133251.cl"},
	{0x99BEC805F67DB326ULL, "vxht_keep_61z_0_0.vlm"},
	{0x99BF0A3B320F8F7FULL, "llac_lt_0_lowtown_docks_a3_-1_0.dds"},
	{0x99BF7AFBF81398DCULL, "vxht_mansion_n_338_0_0.vlm"},
	{0x99BF847F32F0FE44ULL, "w_blightlands_212_0.msh"},
	{0x99BF8BACBB74EF47ULL, "framebuffereffects_bloomdiagonalblur_p0.psh.fxo.dx11"},
	{0x99BF967F32F11CADULL, "w_blightlands_212_0.mao"},
	{0x99BF9A7F32F123E6ULL, "w_blightlands_212_0.mmh"},
	{0x99C0F88F227E50B1ULL, "vxht_brothel_n_61a_0_0.vlm"},
	{0x99C17860493D6B0DULL, "art\\vfx\\combat\\vfx_g_dual_att_right_c.anb"},
	{0x99C21297C3EE7CCCULL, "vxw_bonepit_9o_0_0.vlm"},
	{0x99C23AD549BD063FULL, "vxdr_0_deeproads_fi4_0_0.vlm"},
	{0x99C25B0C336ED7E6ULL, "dm_gen_4.mor"},
	{0x99C25EB2DF1D26EFULL, "6075609_m.fxe"},
	{0x99C26CD80CBD6225ULL, "hlac_lt_gallowstemplar_2395_0x.dds"},
	{0x99C28764E53BD361ULL, "vxht_0_hightown_n_bjc_0_0.vlm"},
	{0x99C30EA6C0F44B90ULL, "6043969_f.fxe"},
	{0x99C36F1DBA5F6AFDULL, "523869527.ogg"},
	{0x99C3A343382694C9ULL, "601529787.ogg"},
	{0x99C3E103C49B2213ULL, "llac_lt_0_lowtown_a2_n_-131072_4.dds"},
	{0x99C3F87353DB9B01ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar6_mcr370_playerrules.gad"},
	{0x99C49696C28E3940ULL, "vxdr_primevaltomb_526_0_1.vlm"},
	{0x99C498CFF571FAB5ULL, "vxlt_0_lowtown_docks_n_bxd_0_0.vlm"},
	{0x99C4B905DF84A476ULL, "zz_amb_test_man.utc"},
	{0x99C5133DDF57DE7CULL, "86487728.ogg"},
	{0x99C516F80C9E6F6FULL, "hlac_lt_0_lowtown_docks_-65538_8y.dds"},
	{0x99C52DC5BD6B0510ULL, "vxht_brothel_n_5ge_0_0.vlm"},
	{0x99C533C6111B53B6ULL, "6128555_m.fxe"},
	{0x99C564F4FF561048ULL, "168018925.ogg"},
	{0x99C56F8DA75EBE56ULL, "6090447_m.fxe"},
	{0x99C5B480DF068443ULL, "vxht_0_hightown_ca6_0_0.vlm"},
	{0x99C5D6F28C4571E9ULL, "gen00fl_isabela6075256.cl"},
	{0x99C5D89E1C472025ULL, "lt_0_lowtown_a2_n_c_43_0_lc.mmh"},
	{0x99C5DE56DD1EA0D0ULL, "vxw_woundedcoast_1c5_0_0.vlm"},
	{0x99C61EFB820C1041ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_001a_0d.dds"},
	{0x99C6E4BDFF8EB175ULL, "fen310_varania6175044.cl"},
	{0x99C6F4C8AE718D7EULL, "579300208.ogg"},
	{0x99C71AB4E5B04A27ULL, "6126682_m.fxe"},
	{0x99C731EB033D4BCAULL, "168862683.ogg"},
	{0x99C77FD2F810B876ULL, "hlac_w_woundedcoast_262148_0z.dds"},
	{0x99C7EBE07769A2D6ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_act_c.impa.evt"},
	{0x99C81E2551347028ULL, "gencr_stone_golem_statue.utc"},
	{0x99C87AD8470A6E96ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha02a.phy"},
	{0x99C88046EEC5625DULL, "100975379.ogg"},
	{0x99C8C197CFACCB49ULL, "gen_im_npc_elf_arm_boo_dalc.uti"},
	{0x99C8E26FD57D536CULL, "hlac_w_bonepit_4_0x.dds"},
	{0x99C8E5A9DB51E6C8ULL, "6089848_f.fxe"},
	{0x99C91D15E4416379ULL, "vxw_blightlands_9m_0_0.vlm"},
	{0x99C92C7E61864A6BULL, "ht_0_hightownexplosion_n_21_0.rml"},
	{0x99C95F41AC38B10DULL, "6188640_m.fxe"},
	{0x99C98A5887571B75ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightr.phy"},
	{0x99C9ED949C56D3BBULL, "vxw_bonepitmining_a2_ha_0_0.vlm"},
	{0x99CA57D795ABE613ULL, "887803294.ogg"},
	{0x99CA93330B835DF7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock2a.phy"},
	{0x99CAF97353E1BBAFULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar6_mcr370_playerrules.ani"},
	{0x99CB18DF029653D3ULL, "6112083_m.fxe"},
	{0x99CB7F79073C5FD2ULL, "dae000_bartrand6064045.cl"},
	{0x99CBC6ED5FED9FC0ULL, "397606339.ogg"},
	{0x99CBE0D18A724917ULL, "mrl150ip_pile_a.utp"},
	{0x99CBEE0F67434D5EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9dd_doubledoor.phy"},
	{0x99CBFDB2BF836284ULL, "76549427.ogg"},
	{0x99CC101AA6907F48ULL, "vxht_0_hightown_n_bjf_0_0.vlm"},
	{0x99CC53011E24BDB6ULL, "vfx_chunkyparticle_p1.psh.fxo.dx11"},
	{0x99CD240197DC157AULL, "gen00fl_anders6125499.cl"},
	{0x99CD5F918ABEFC72ULL, "68881801.ogg"},
	{0x99CD776E75F78178ULL, "797687721.ogg"},
	{0x99CD8EF6B224681AULL, "1037252127.ogg"},
	{0x99CE067323AB84ADULL, "hf_ave1_raiderf.mor"},
	{0x99CEA1EFCAAF1950ULL, "vxca_cave_1bo_0_1.vlm"},
	{0x99CEE2E0776FB261ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_act_c.impa.gad"},
	{0x99CEED59FDC4FABCULL, "hlac_w_blightlands_393215_0z.dds"},
	{0x99D0306BE2D4EEF7ULL, "lowtown_fem_prostitute_sta6152359.cl"},
	{0x99D069CAECF91DC4ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_dummy.phy"},
	{0x99D088AB25E61A8AULL, "hlac_lt_0_lowtown_docks_a3_65536_0x.dds"},
	{0x99D0BA9DD1DCA1BBULL, "180161310.ogg"},
	{0x99D0FF7522475F4AULL, "mag321cr_comte.utc"},
	{0x99D1305AE86E7FCFULL, "hlac_ht_playershouse_n_855_1y.dds"},
	{0x99D151F7881F7190ULL, "35089283.ogg"},
	{0x99D21C8736EEF945ULL, "vxlt_blackemporium_15t_0_0.vlm"},
	{0x99D263D847129C37ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha02a.msh"},
	{0x99D279D84712C1A9ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha02a.mmh"},
	{0x99D2DD8C64C32DC8ULL, "6110588_f.fxe"},
	{0x99D318431B04126FULL, "ht_0_hightown_n_25_0_lc.mmh"},
	{0x99D377754310D901ULL, "6144774_m.fxe"},
	{0x99D41F1E78BC41CDULL, "isa311_isabela6094455.cl"},
	{0x99D4265DD0FC9CC3ULL, "vxht_0_hightown_cd9_0_0.vlm"},
	{0x99D46AD61020BAA0ULL, "follower_banter6160835.cl"},
	{0x99D4879017D7D145ULL, "95550317.ogg"},
	{0x99D4CC442AC9F55BULL, "798636202.ogg"},
	{0x99D4E3E077741F0FULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_act_c.impa.ani"},
	{0x99D510DF429978BBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_gesture_forward.ani"},
	{0x99D52B6FD7228443ULL, "dae000_pc_home_ambients6208672.cl"},
	{0x99D56DE8A7139635ULL, "hightown_male_noble_gossip6174086.cl"},
	{0x99D65D921529C0BBULL, "6034871_m.fxe"},
	{0x99D67AC3167AEA76ULL, "vxca_caverns1_i1i_0_0.vlm"},
	{0x99D68F55BAE90494ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_left_hand.ani"},
	{0x99D69EB85D66BA61ULL, "30087065.ogg"},
	{0x99D718F683D4AA6BULL, "art\\characters\\playercharacter\\humanmale\\rob_appc.phy"},
	{0x99D7F077794F613DULL, "w_bonepitdestroyed_a3_169_0.mao"},
	{0x99D7F477794F6876ULL, "w_bonepitdestroyed_a3_169_0.mmh"},
	{0x99D7F7E168A434D9ULL, "hlac_lt_0_lowtown_docks_n_-4_3z.dds"},
	{0x99D7FE77794F7974ULL, "w_bonepitdestroyed_a3_169_0.msh"},
	{0x99D82809AD1D3D84ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_broken.msh"},
	{0x99D83A09AD1D5CEDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_broken.mao"},
	{0x99D83E09AD1D6326ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_broken.mmh"},
	{0x99D88FD19AE1A664ULL, "6042034_f.fxe"},
	{0x99D8DFB3AF847F64ULL, "6199993_m.fxe"},
	{0x99D91C5055E5A62AULL, "fen110st_hunter.stg"},
	{0x99D92B2C9EBA3796ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mugpewter01_0_d.dds"},
	{0x99DA272A024D8F4AULL, "llac_lt_0_lowtown_n_1_10.dds"},
	{0x99DA7A58EA723B11ULL, "mag110_arianni6016798.cl"},
	{0x99DA93AD49F0BF66ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgte.phy"},
	{0x99DAA76FCF8A6EF8ULL, "vxlt_0_lowtown_a2_n_elj_0_1.vlm"},
	{0x99DAB16511D870BBULL, "6035854_m.fxe"},
	{0x99DB14EAFE13C25AULL, "w_woundedcoast_153_0.mmh"},
	{0x99DB18EAFE13C891ULL, "w_woundedcoast_153_0.mao"},
	{0x99DB26EAFE13E0A0ULL, "w_woundedcoast_153_0.msh"},
	{0x99DBE9B8EE0EA8BBULL, "793074061.ogg"},
	{0x99DC764B8ECCF518ULL, "45306570.ogg"},
	{0x99DCB4AC5B530E82ULL, "vfx_rendertoshadowmap_p2.psh.fxo.dx11"},
	{0x99DCDAC0E7781D82ULL, "rdr111dg_assistant6050916.cl"},
	{0x99DCFBEF0BB77536ULL, "6123311_f.fxe"},
	{0x99DD4AB5731BC47CULL, "vxht_0_hightown_bds_0_0.vlm"},
	{0x99DD8C55BAEF1EF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_left_hand.gad"},
	{0x99DDF8E9E39EF2C2ULL, "vxht_brothel_n_5ck_0_0.vlm"},
	{0x99DE3696E680810FULL, "gen00fl_merrill6128659.cl"},
	{0x99DE5E023628A675ULL, "6067770_m.fxe"},
	{0x99DED50517ECADA4ULL, "651292654.ogg"},
	{0x99DF14DE293E7E6FULL, "dae000_bartrand6031156.cl"},
	{0x99DF47550036C28FULL, "llac_lt_0_lowtown_n_131070_5.dds"},
	{0x99DFBC6C3ECA2F4BULL, "6061367_m.fxe"},
	{0x99E064FF7B3B3763ULL, "vxca_caverns2_97a_0_0.vlm"},
	{0x99E0B75BC938CAC8ULL, "llac_ht_0_hightown_a3_-196609_0.dds"},
	{0x99E132DA1C311089ULL, "gen00fl_anders6125234.cl"},
	{0x99E1926EE51F93DAULL, "16503788.ogg"},
	{0x99E1BE33A79BA589ULL, "vxw_bonepitdestroyed_a3_29n_0_0.vlm"},
	{0x99E1D50C55167CA6ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_1y.dds"},
	{0x99E1F5503E0F36C0ULL, "qun131_taskmaster6039719.cl"},
	{0x99E218DF42A410B2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_gesture_forward.evt"},
	{0x99E23509AD25A8F1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_broken.phy"},
	{0x99E2C210C10AD1D0ULL, "vxw_woundedcoast_1i3_0_0.vlm"},
	{0x99E2E867C930F7F9ULL, "vxht_brothel_n_6ue_0_0.vlm"},
	{0x99E2E87BAF6B5E3FULL, "6126916_m.fxe"},
	{0x99E36743166E5366ULL, "6125167_m.fxe"},
	{0x99E3EBC649723B2BULL, "6001332_m.fxe"},
	{0x99E3FB9ADE5C8D48ULL, "738157373.ogg"},
	{0x99E4028F9DF89DD2ULL, "hlac_lt_0_lowtown_n_-196608_3x.dds"},
	{0x99E405E181C830FEULL, "216160475.ogg"},
	{0x99E414FFAC53F4E8ULL, "lgt210cr_debt_thug1.utc"},
	{0x99E43B6D32BBD594ULL, "747353427.ogg"},
	{0x99E48CB0FB63D17AULL, "w_blightlands_144_0.msh"},
	{0x99E49AB0FB63E8B3ULL, "w_blightlands_144_0.mao"},
	{0x99E49EB0FB63EFC0ULL, "w_blightlands_144_0.mmh"},
	{0x99E4A755BAF56B55ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_left_hand.evt"},
	{0x99E551E910C633E7ULL, "vxlt_0_lowtown_a2_n_c_e82_0_0.vlm"},
	{0x99E59AC22C5CE7D0ULL, "vxht_chantry_n_2t9_0_1.vlm"},
	{0x99E5B086EE109D75ULL, "nat380_nathaniel6067317.cl"},
	{0x99E66CD1BCED1034ULL, "928535772.ogg"},
	{0x99E6777495AB05D8ULL, "6087144_m.fxe"},
	{0x99E680A0875C9066ULL, "hlac_dr_primevaltomb_659_11x.dds"},
	{0x99E7230FDA6F5479ULL, "vxw_sundermount_26p_0_0.vlm"},
	{0x99E74806E7C0904DULL, "vxlt_gallowstemplar_f_5c0_0_0.vlm"},
	{0x99E7BF79ADD79317ULL, "art\\levels\\races\\design\\prp_chestc_0.pwk"},
	{0x99E7DA8BCDFAF63BULL, "mcr310_argument6182297.cl"},
	{0x99E7EA65AA830D9CULL, "6033756_m.fxe"},
	{0x99E7F496796CCEB0ULL, "vxca_caverns2_7un_0_0.vlm"},
	{0x99E890BF427F11D6ULL, "6115432_m.fxe"},
	{0x99E8BAC34A526A0EULL, "hlac_ht_keep_4519_0x.dds"},
	{0x99E90DBD7ACB21E4ULL, "gen00fl_bethany6028819.cl"},
	{0x99E937DF42AA6445ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_gesture_forward.gad"},
	{0x99E94E9B08FFBCFCULL, "377934139.ogg"},
	{0x99EA08C3C532E48EULL, "vxht_0_hightown_n_c54_0_0.vlm"},
	{0x99EA0D33D62B2036ULL, "mag291_demon_marethari6152612.cl"},
	{0x99EB5240E54FD2FBULL, "vxht_mansion2_1kn_0_0.vlm"},
	{0x99EB850D4F94E919ULL, "665756136.ogg"},
	{0x99EBE11E4C5700EFULL, "86515437.ogg"},
	{0x99EC6239814A9DAAULL, "vxht_chantry_23t_0_0.vlm"},
	{0x99EC6AC6D5950B42ULL, "vxht_brothel_n_5ft_0_0.vlm"},
	{0x99EC8B837E130B25ULL, "face_face_p0.psh.fxo"},
	{0x99ED173C4DED10D3ULL, "6095795_m.fxe"},
	{0x99ED523B35B0F577ULL, "vxw_bonepitmining_a2_mz_0_0.vlm"},
	{0x99ED67B4A15808C1ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\barkburn.mao"},
	{0x99EDC40CB24FA67CULL, "304566578.ogg"},
	{0x99EDFAFF91FEA3F1ULL, "vxdr_primevaltomb_52z_0_0.vlm"},
	{0x99EE3E1FBF5E8D66ULL, "gen00fl_aveline6110360.cl"},
	{0x99EE7161523DFA6FULL, "hlac_w_bonepit_327686_0y.dds"},
	{0x99EE769F2F6E57ACULL, "vxlt_0_lowtown_docks_a3_n_czq_0_0.vlm"},
	{0x99EEC3DED78E2D3AULL, "plt_lgt000pt_ultra_4_05.plo"},
	{0x99EF29F4D11CE364ULL, "297041287.ogg"},
	{0x99EF2F8DF3280AB0ULL, "hlac_lt_0_lowtown_a2_65538_0z.dds"},
	{0x99EFDD505A540874ULL, "vxht_0_hightown_a3_9uv_0_0.vlm"},
	{0x99F030109991BE09ULL, "vxw_bonepitdestroyed_a3_24m_0_0.vlm"},
	{0x99F0C683884EBC59ULL, "583170876.ogg"},
	{0x99F133EDB300EA50ULL, "vxw_woundedcoast_1t0_0_0.vlm"},
	{0x99F16CEB648B6E57ULL, "llac_lt_0_lowtown_0_4.dds"},
	{0x99F16FF6D3FFA70FULL, "art\\levels\\ztest\\chad\\w_gax_sotf.mao"},
	{0x99F1DB60F68ADAE0ULL, "6076988_m.fxe"},
	{0x99F1E5F7761E6049ULL, "6117791_f.fxe"},
	{0x99F2190A37F127D0ULL, "hlac_lt_hangedman_n_-1_29y.dds"},
	{0x99F21AC5F8F1AE32ULL, "hlac_ht_0_hightown_a3_131073_0y.dds"},
	{0x99F22D7FB8D3A6DAULL, "325356369.ogg"},
	{0x99F240B822763B56ULL, "6137014_m.fxe"},
	{0x99F25E418919B8ADULL, "6122446_m.fxe"},
	{0x99F2D6B5E57D0EAEULL, "432578423.ogg"},
	{0x99F33924108242C2ULL, "zz_vault_debug6178940.cl"},
	{0x99F3510B02194316ULL, "ser191_emeric6116980.cl"},
	{0x99F3B2567292CF78ULL, "572263227.ogg"},
	{0x99F4103180E27AEBULL, "181370096.ogg"},
	{0x99F4279E5E10F23CULL, "art\\vfx\\combat\\vfx_w_tmp_clean_lar_aoe_p.impa.ani"},
	{0x99F431D8A2CA07A4ULL, "w_encounter3_65_0.msh"},
	{0x99F443D8A2CA260DULL, "w_encounter3_65_0.mao"},
	{0x99F447D8A2CA2D46ULL, "w_encounter3_65_0.mmh"},
	{0x99F454C7F57781ADULL, "462463129.ogg"},
	{0x99F4D39A07AD6AEDULL, "hlac_ht_brothel_n_3369_2y.dds"},
	{0x99F5B8FCFD171409ULL, "mag170_thrask6016933.cl"},
	{0x99F5F5A9DE7EBAA7ULL, "366126421.ogg"},
	{0x99F666736B62E730ULL, "vxca_caverns2_7jo_0_0.vlm"},
	{0x99F6731A0E686075ULL, "zz_area_signage_18.utp"},
	{0x99F68E7139BBF9AEULL, "vxdr_0_deeproads_f55_0_0.vlm"},
	{0x99F6EC96DFD86135ULL, "hlac_lt_gallowstemplar_2394_10x.dds"},
	{0x99F7062408E369BAULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.cess.gad"},
	{0x99F70A769A138164ULL, "art\\levels\\ztools\\default\\generic\\9ff_bench.msh"},
	{0x99F71C769A139FCDULL, "art\\levels\\ztools\\default\\generic\\9ff_bench.mao"},
	{0x99F720769A13A606ULL, "art\\levels\\ztools\\default\\generic\\9ff_bench.mmh"},
	{0x99F7542486B767ACULL, "hlac_lt_undercitydungeon_-1_36x.dds"},
	{0x99F7D97691DF7579ULL, "ser221_party_comments6164315.cl"},
	{0x99F7FAA5D1BA9F6BULL, "gen_im_npc_arm_qun_hlm_maga.uti"},
	{0x99F8107421F6D2D2ULL, "6167323_m.fxe"},
	{0x99F823230A827DF5ULL, "sky_motionvectorssky_p0.vsh.fxo.dx11"},
	{0x99F87AA0B728FD0EULL, "vxht_0_hightown_n_c25_0_0.vlm"},
	{0x99F8A7AC5B19C70BULL, "6066476_m.fxe"},
	{0x99F8B8AE1380D27AULL, "6201350_f.fxe"},
	{0x99F8B9628B2E3710ULL, "gen_im_npc_arm_hum_boo_thga.uti"},
	{0x99F8ED40816FD9FBULL, "bug.ncs"},
	{0x99F9306C8A87C3D5ULL, "vxlt_hangedman_n_21t_0_0.vlm"},
	{0x99F9AF2AC5E40113ULL, "91409336.ogg"},
	{0x99FA582CF57ECA08ULL, "287942400.ogg"},
	{0x99FA628F63A030A2ULL, "6116557_f.fxe"},
	{0x99FA8A54395FC8FDULL, "demo_warrior_shield.utc"},
	{0x99FADCA3C78B23C2ULL, "vxht_brothel_n_5iu_0_0.vlm"},
	{0x99FB08A202B93F71ULL, "hlac_ca_caverns1_mines_24421_6z.dds"},
	{0x99FB349E5E17277EULL, "art\\vfx\\combat\\vfx_w_tmp_clean_lar_aoe_p.impa.gad"},
	{0x99FB72863A3130A7ULL, "6111081_f.fxe"},
	{0x99FB7B7CAD521612ULL, "6127997_f.fxe"},
	{0x99FD544F78BE054EULL, "6053878_f.fxe"},
	{0x99FDE640D29F4FF0ULL, "619786726.ogg"},
	{0x99FDF082DF703759ULL, "6180608_m.fxe"},
	{0x99FE092408E98DD1ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.cess.evt"},
	{0x99FE69B7F7D8BCCEULL, "6125206_m.fxe"},
	{0x99FF017534FA7925ULL, "hlac_ca_caverns1_mines_24426_4y.dds"},
	{0x99FF3ED8A2D42511ULL, "w_encounter3_65_0.phy"},
	{0x99FF8BA8CDD06F1AULL, "vxlt_gallowstemplar_n_3sy_0_0.vlm"},
	{0x99FFD5A98C327C15ULL, "699127117.ogg"},
	{0x99FFD81FDE86FA55ULL, "6085833_m.fxe"},
	{0x9A001DD5ED40F02AULL, "230853795.ogg"},
	{0x9A00B879181DA92BULL, "799288813.ogg"},
	{0x9A0117769A1BEBD1ULL, "art\\levels\\ztools\\default\\generic\\9ff_bench.phy"},
	{0x9A0133EAFE34C665ULL, "w_woundedcoast_153_0.phy"},
	{0x9A01D77637AF74F7ULL, "llac_dr_dungeon_3207_7.dds"},
	{0x9A01EC3FC8C4F930ULL, "vxlt_0_lowtown_a2_n_c_e38_0_0.vlm"},
	{0x9A020F9E5E1D069DULL, "art\\vfx\\combat\\vfx_w_tmp_clean_lar_aoe_p.impa.evt"},
	{0x9A029FFF609DC58FULL, "6118356_f.fxe"},
	{0x9A02C9DA1AC6FDE9ULL, "vxht_brothel_n_5aj_0_0.vlm"},
	{0x9A02CFB0AEE5B1B0ULL, "gen00fl_anders6110062.cl"},
	{0x9A034B089D9F2709ULL, "hlac_ca_caverns1_24424_4x.dds"},
	{0x9A034EECEE3D6083ULL, "402489086.ogg"},
	{0x9A044E4CBB548704ULL, "vxht_keep_n_5hc_0_0.vlm"},
	{0x9A0451FDD78881B8ULL, "6067928_m.fxe"},
	{0x9A0484952B338C53ULL, "vxlt_0_lowtown_docks_n_b5d_0_1.vlm"},
	{0x9A052B0F0B1785F8ULL, "6174358_m.fxe"},
	{0x9A053A21EB843652ULL, "vxlt_0_lowtown_a2_n_c_i0o_0_0.vlm"},
	{0x9A0569B93A3A053AULL, "942468263.ogg"},
	{0x9A057543BF5B11DAULL, "hlac_w_sundermount_196609_0x.dds"},
	{0x9A05895478B1C23FULL, "6174601_m.fxe"},
	{0x9A0596AB4D6C093FULL, "plt_cod_ltr_ggt_note4.nss"},
	{0x9A05A77E9AD0D224ULL, "ht_mansion_n_3371_ll.mmh"},
	{0x9A05C6BDA82E20B6ULL, "hlac_lt_ruinedhouse_n_-1_1y.dds"},
	{0x9A05CF89A6D40B44ULL, "vxlt_gallowstemplar_n_37c_0_0.vlm"},
	{0x9A06103366D6FB00ULL, "vxht_mansion_n_3wh_0_0.vlm"},
	{0x9A0677837FDF9389ULL, "433019579.ogg"},
	{0x9A0762ED2A23F66EULL, "vxlt_0_lowtown_n_e23_0_0.vlm"},
	{0x9A07958657BD0DBFULL, "mcr370_negotiation6080195.cl"},
	{0x9A07EB5D7B183685ULL, "vxht_bartrand_n_6r4_0_0.vlm"},
	{0x9A0869ACAC8AB034ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\left_hand_torightarm_loop.ani"},
	{0x9A089C6927A12FC0ULL, "vxlt_backalley_n_1l8_0_0.vlm"},
	{0x9A08A04C48FE0DFCULL, "vxht_0_hightown_bwv_0_0.vlm"},
	{0x9A08FC77FE569B2BULL, "llac_lt_0_lowtown_docks_n_-131075_4.dds"},
	{0x9A09328A27818EA6ULL, "ht_0_hightown_35_0_ll.mmh"},
	{0x9A095964A4F80B5BULL, "443883359.ogg"},
	{0x9A09BC6FA1BDE519ULL, "6042127_f.fxe"},
	{0x9A09CD165EE305E5ULL, "vxdr_0_deeproads_f1o_0_1.vlm"},
	{0x9A09EB676B2B5F63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_haste.evt"},
	{0x9A0A20BFD1BE59FBULL, "941046827.ogg"},
	{0x9A0A4521F718DD7DULL, "character_character_p55.psh.fxo"},
	{0x9A0A68FE7FAA418FULL, "static_passthrough_p0.psh.fxo"},
	{0x9A0A7E08E5340021ULL, "lgt311_shw_mer_intro.bnk"},
	{0x9A0ADCE8249FBFBCULL, "dae000_keep_male_servant_mo6160918.cl"},
	{0x9A0B07543E760DB0ULL, "w_bonepitmining_a2_237_0.msh"},
	{0x9A0B15543E7625CAULL, "w_bonepitmining_a2_237_0.mmh"},
	{0x9A0B19543E762C01ULL, "w_bonepitmining_a2_237_0.mao"},
	{0x9A0B212408F440A0ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.cess.ani"},
	{0x9A0BB98F3F195F5BULL, "6208507_f.fxe"},
	{0x9A0C2D8BFAA77446ULL, "6085874_m.fxe"},
	{0x9A0C50A708FC3886ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgtp_0n.dds"},
	{0x9A0C5F71A1BBEA34ULL, "90268359.ogg"},
	{0x9A0C673D241D10DEULL, "vrc221ip_servant_corpse_3.utp"},
	{0x9A0CBABC43CCB480ULL, "6127040_m.fxe"},
	{0x9A0D5BA14671AE98ULL, "mrl150_merrill6093520.cl"},
	{0x9A0E81F3476ECA3CULL, "vxlt_0_lowtown_docks_n_b7t_0_0.vlm"},
	{0x9A0E88BCE1ECF1DFULL, "vxlt_0_lowtown_n_fgn_0_0.vlm"},
	{0x9A0EA008E5382291ULL, "lgt311_shw_mer_intro.cnv"},
	{0x9A0F55FA7283D077ULL, "432411194.ogg"},
	{0x9A0F66ACAC90C996ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\left_hand_torightarm_loop.gad"},
	{0x9A0F878FD5167206ULL, "79384593.ogg"},
	{0x9A0F9DED2E73C248ULL, "gen_im_trs_gem_opal_04.uti"},
	{0x9A0FBA46737F5E7AULL, "zz_followers_banter6159501.cl"},
	{0x9A0FD22FA629C6ECULL, "sib131_letter_react6127310.cl"},
	{0x9A0FDF14AFD13D3DULL, "6142251_m.fxe"},
	{0x9A102541DFB2C502ULL, "6016636_m.fxe"},
	{0x9A1086EDA4A82270ULL, "vxlt_0_lowtown_a3_n_c_i0l_0_0.vlm"},
	{0x9A109D5887933D90ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightr.msh"},
	{0x9A10AB58879355AAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightr.mmh"},
	{0x9A10AF5887935CE1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightr.mao"},
	{0x9A10C4676B313BE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_haste.gad"},
	{0x9A10F689DCFE545EULL, "778920636.ogg"},
	{0x9A1102BBD8F2D1F2ULL, "6177937_m.fxe"},
	{0x9A11052AE0C11A83ULL, "6092552_m.fxe"},
	{0x9A114E0A3AF67EB7ULL, "qun160_vashothlead6055585.cl"},
	{0x9A115017C8D8C44CULL, "vxlt_0_lowtown_gko_0_0.vlm"},
	{0x9A11BAB9611A8C9EULL, "llac_ht_keep_3743_0.dds"},
	{0x9A1268330BC0CA4FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock2a.mao"},
	{0x9A1274330BC0DEECULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock2a.mmh"},
	{0x9A1274442EC46695ULL, "636067029.ogg"},
	{0x9A127A330BC0E8DEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock2a.msh"},
	{0x9A1328B31EBBCE98ULL, "609441329.ogg"},
	{0x9A137F3EBFF1BBE3ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazier.mao"},
	{0x9A13833EBFF1C230ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazier.mmh"},
	{0x9A13913EBFF1DA4AULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazier.msh"},
	{0x9A139E47D6B7C124ULL, "gen00fl_merrill6111225.cl"},
	{0x9A13FFD0F503D9B9ULL, "art\\vfx\\combat\\vfx_g_appearance_c.mmh"},
	{0x9A14131373D397E6ULL, "qun110cr_mugger.utc"},
	{0x9A141F7244014CAFULL, "qun221_petrice6075980.cl"},
	{0x9A1497468D0A8A5BULL, "6128078_f.fxe"},
	{0x9A14A6ACED793C55ULL, "vxht_bartrand_n_4fj_0_0.vlm"},
	{0x9A150811A58A7D4FULL, "ef_gen_act1.mor"},
	{0x9A152CF74E1E0CC7ULL, "3412003.ogg"},
	{0x9A152F93E1F53646ULL, "6068261-ifearwehavecome.cut"},
	{0x9A154C2D268AE0D9ULL, "vxw_bonepit_1xc_0_1.vlm"},
	{0x9A15E18D175068ABULL, "qcr000_arishok6059294.cl"},
	{0x9A16531A3694B435ULL, "one200ip_st_solivitus.ncs"},
	{0x9A16CF45F0BB1F5EULL, "vxlt_gallowstemplar_f_4qd_0_0.vlm"},
	{0x9A16E3519488726BULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.cess.evt"},
	{0x9A170E4619974840ULL, "vxlt_backalley_n_1w9_0_0.vlm"},
	{0x9A1712293AF4267CULL, "vxht_0_hightown_bpw_0_0.vlm"},
	{0x9A1732C65BB80A6CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_8.pwk"},
	{0x9A173696603645A8ULL, "vxca_caverns2prc_adz_0_0.vlm"},
	{0x9A17888EF5EB06FAULL, "vxlt_warehouse_8m_0_0.vlm"},
	{0x9A17D3676B3773C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_haste.ani"},
	{0x9A17E163B84C0DE2ULL, "6162324_m.fxe"},
	{0x9A182EB2187E1312ULL, "vxdr_0_deeproads_ebw_0_0.vlm"},
	{0x9A185F36913E4712ULL, "6129797_f.fxe"},
	{0x9A1889ACD1B60256ULL, "gen00fl_varric6088855.cl"},
	{0x9A18AC986E39484AULL, "102532379.ogg"},
	{0x9A18BFAE8739651CULL, "plt_cod_cha_aveline3.plo"},
	{0x9A1982AC73C8C3ACULL, "817675.ogg"},
	{0x9A1990774190063CULL, "vxca_caverns1_mines_ckf_0_0.vlm"},
	{0x9A19ADB240377998ULL, "mag120_macha6043033.cl"},
	{0x9A19B2AFA9C83E59ULL, "7041711.ogg"},
	{0x9A19C83DE06443C0ULL, "vxw_blightlands_11a_0_0.vlm"},
	{0x9A19DF12F3D8A4F0ULL, "vxca_caverns2prc_8kc_0_1.vlm"},
	{0x9A1A89AB4D7E4F50ULL, "plt_cod_ltr_ggt_note4.plo"},
	{0x9A1A92A94EB25512ULL, "hlac_w_encounter1_3_0x.dds"},
	{0x9A1AA22269B43652ULL, "mcr371_scared_mage_m6087699.cl"},
	{0x9A1AAAE75C395173ULL, "character_character_p152.psh.fxo.dx11.assembly"},
	{0x9A1ABEE191426C08ULL, "305400631.ogg"},
	{0x9A1AEA1D48E59B3AULL, "follower_banter6163120.cl"},
	{0x9A1AF3322D400FF3ULL, "vxlt_backalley_g_1ks_0_0.vlm"},
	{0x9A1B269C9F15C8D4ULL, "897422273.ogg"},
	{0x9A1BC8DDF63A1A8BULL, "6109403_m.fxe"},
	{0x9A1BCDA89B9F0D4BULL, "vxlt_0_lowtown_a2_gc8_0_0.vlm"},
	{0x9A1BDE761EB120D1ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_001a_0t.dds"},
	{0x9A1C53578F97624AULL, "vxlt_0_lowtown_f0u_0_0.vlm"},
	{0x9A1C78C5910D1BFDULL, "6075451_f.fxe"},
	{0x9A1CC70B503FB00EULL, "vxw_woundedcoast_nq_0_0.vlm"},
	{0x9A1CF8024478F401ULL, "6082422_f.fxe"},
	{0x9A1CFC51948D082CULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.cess.gad"},
	{0x9A1D80BB823785AFULL, "hlac_lt_0_lowtown_a2_n_65535_3z.dds"},
	{0x9A1F353BDA83A881ULL, "hlac_lt_undercityserialkiller_-1_43x.dds"},
	{0x9A1F6979761AD78EULL, "6095928_f.fxe"},
	{0x9A1F85F16AE065C9ULL, "6142703_m.fxe"},
	{0x9A1F8A1473512444ULL, "808583002.ogg"},
	{0x9A1FD32BFF0D45C5ULL, "vxlt_house_e6_0_0.vlm"},
	{0x9A2007FCEB2A9990ULL, "art\\characters\\weapons\\longswords\\w_lsw_020a.mao"},
	{0x9A200BFCEB2AA0A3ULL, "art\\characters\\weapons\\longswords\\w_lsw_020a.mmh"},
	{0x9A2019FCEB2AB8FDULL, "art\\characters\\weapons\\longswords\\w_lsw_020a.msh"},
	{0x9A20672D88D58062ULL, "6129732_m.fxe"},
	{0x9A2075633910BB0BULL, "6167528_m.fxe"},
	{0x9A20A3D6CC3A2DC8ULL, "6107287_f.fxe"},
	{0x9A21BB3E7609AE0CULL, "darktown_fem_fer_urch_sta6174805.cl"},
	{0x9A21D6FAD9C05475ULL, "6122752_m.fxe"},
	{0x9A221BA2029ECC21ULL, "194988008.ogg"},
	{0x9A224CB3249FDD91ULL, "t3_rob_pl2.tnt"},
	{0x9A2293CCF21186C3ULL, "qun221_petrice6077991.cl"},
	{0x9A229CD01F816AA5ULL, "6093617_f.fxe"},
	{0x9A23449796F7FA27ULL, "6116410_m.fxe"},
	{0x9A23B607E833FA9BULL, "376201482.ogg"},
	{0x9A23F8123DB52EC5ULL, "260486010.ogg"},
	{0x9A23FB519493252AULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.cess.ani"},
	{0x9A2430191EBAC315ULL, "llac_w_blightlands_262143_0.dds"},
	{0x9A24FAAE5B1ADCE9ULL, "77319774.ogg"},
	{0x9A25186A4A919D78ULL, "bla_dust.mao"},
	{0x9A251D7248629A96ULL, "llac_ht_0_hightown_n_0_0.dds"},
	{0x9A254122E2B137DEULL, "vxlt_gallowstemplar_f_4ve_0_0.vlm"},
	{0x9A25971DBDEBBD91ULL, "120673275.ogg"},
	{0x9A26A542F0A7A7DAULL, "6060003_f.fxe"},
	{0x9A27608ADD6547EDULL, "809175865.ogg"},
	{0x9A281F5DB9DA6E7EULL, "6144108_m.fxe"},
	{0x9A2828DFE5CE060EULL, "76727517.ogg"},
	{0x9A283A1AD25A5C40ULL, "vxw_blightlands_14n_0_0.vlm"},
	{0x9A2856694B3C6F14ULL, "6144043_m.fxe"},
	{0x9A28991B32A831B2ULL, "vxlt_0_lowtown_a2_n_ehu_0_0.vlm"},
	{0x9A28C1FFA4C75866ULL, "vxlt_undercityserialkiller_yo_0_0.vlm"},
	{0x9A2934A533E1A28BULL, "dae000_gamlen6005038.cl"},
	{0x9A294B0F01FB6CAAULL, "916138608.ogg"},
	{0x9A29650F1F362873ULL, "vxlt_backalley_g_1pr_0_0.vlm"},
	{0x9A299743C0B502A0ULL, "mcr370_varric6098300.cl"},
	{0x9A29A4C4222C6A74ULL, "vxht_0_hightown_a3_9az_0_0.vlm"},
	{0x9A29CB6A889CE883ULL, "hlac_ca_cavern1_haunted_24425_8x.dds"},
	{0x9A29E13A09362F5AULL, "dre180ar_keep_fake.ncs"},
	{0x9A2B38E84235C88EULL, "vxw_woundedcoast_kr_0_0.vlm"},
	{0x9A2B3C0D4531CFCDULL, "498665306.ogg"},
	{0x9A2B896A1BEC2375ULL, "vxlt_0_lowtown_a3_n_c_hnv_0_0.vlm"},
	{0x9A2B9D5FEBBF3F28ULL, "dae200_bran6094874.cl"},
	{0x9A2BEC39E22A6594ULL, "6048736_m.fxe"},
	{0x9A2C3EE4892ECE1DULL, "307387355.ogg"},
	{0x9A2C5918440EE813ULL, "6067303_m.fxe"},
	{0x9A2C82A09895CDE7ULL, "seb000pt_gift_bow.ncs"},
	{0x9A2E116424C4161FULL, "qun111_ketojan6059227.cl"},
	{0x9A2E928C58699679ULL, "hlac_lt_gallowstemplar_2383_7z.dds"},
	{0x9A2EE651F6BD404FULL, "hlac_ht_0_hightown_-1_2x.dds"},
	{0x9A2F082BBEB9BD13ULL, "607167389.ogg"},
	{0x9A2FA46CD69AF482ULL, "6069807_m.fxe"},
	{0x9A2FCB2834C76DB4ULL, "6147590_m.fxe"},
	{0x9A303C096B7AA5E0ULL, "483967814.ogg"},
	{0x9A3087A236137EE5ULL, "767778962.ogg"},
	{0x9A30BBCAEAA2E39FULL, "lt_0_lowtown_docks_n_116_0.msh"},
	{0x9A30C1CAEAA2EE41ULL, "lt_0_lowtown_docks_n_116_0.mmh"},
	{0x9A30CDCAEAA302E2ULL, "lt_0_lowtown_docks_n_116_0.mao"},
	{0x9A30D22356B29438ULL, "vxht_mansion2_1s1_0_0.vlm"},
	{0x9A30D93A093C404BULL, "dre180ar_keep_fake.lst"},
	{0x9A30EE90044E103AULL, "gen00fl_fenris6106634.cl"},
	{0x9A30F5268D5C1310ULL, "w_encounter1_52_0.msh"},
	{0x9A3103268D5C2B2AULL, "w_encounter1_52_0.mmh"},
	{0x9A31059ED8415B5DULL, "gen00fl_varric6093577.cl"},
	{0x9A3107268D5C3261ULL, "w_encounter1_52_0.mao"},
	{0x9A310CFCEB3915C8ULL, "art\\characters\\weapons\\longswords\\w_lsw_020a.phy"},
	{0x9A316C3C9EA4B7BCULL, "qun221_petrice6083812.cl"},
	{0x9A31849C32B63D5DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mbn.mao"},
	{0x9A31889C32B64396ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mbn.mmh"},
	{0x9A318A66D1656D55ULL, "vxht_bartrand_n_4ph_0_0.vlm"},
	{0x9A31929C32B65494ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16mbn.msh"},
	{0x9A3246BB42810DB4ULL, "124800937.ogg"},
	{0x9A329CB35B808580ULL, "gen00fl_isabela6090434.cl"},
	{0x9A32B4E91504A90CULL, "vxw_encounter1_cl_0_0.vlm"},
	{0x9A32BDA01B5E100CULL, "vxlt_hangedman_n_rp_0_0.vlm"},
	{0x9A32D2D06A17BDD1ULL, "661245917.ogg"},
	{0x9A3315958E73CC86ULL, "gen00fl_aveline6128532.cl"},
	{0x9A3323DACC47303CULL, "786346147.ogg"},
	{0x9A33FEA23677502AULL, "185665500.ogg"},
	{0x9A340F4707203B5AULL, "hlac_lt_0_lowtown_n_-131072_6y.dds"},
	{0x9A344149F7671B7CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopen_v2.gad"},
	{0x9A349457C4F9BBC1ULL, "sky_w_bonepit.mao"},
	{0x9A34FA17D0B3A0FCULL, "hlac_ht_0_hightown_n_65535_2z.dds"},
	{0x9A351533D407DE3EULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.impa.evt"},
	{0x9A352BEE190BBD50ULL, "lt_0_lowtown_docks_a3_12_0_lc.mmh"},
	{0x9A35E9A545C69748ULL, "hlac_lt_undercityserialkiller_-1_35x.dds"},
	{0x9A364B6E7BE7AE0EULL, "433554532.ogg"},
	{0x9A36913AABD086FEULL, "6141208_m.fxe"},
	{0x9A370AF8249E4A32ULL, "vxlt_0_lowtown_a2_n_emv_0_0.vlm"},
	{0x9A376B32EB244262ULL, "6065759_wander1.cut"},
	{0x9A3780805C54F079ULL, "6126806_m.fxe"},
	{0x9A37DFA78495BE8FULL, "vrc221_bartrand6065388.cl"},
	{0x9A38AE9B360D0E2EULL, "weapon_distortionmask_p0.psh.fxo.dx11.assembly"},
	{0x9A39C9EAE1AA8AFAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk_n.dds"},
	{0x9A39D2805C63D2C6ULL, "gen_im_trs_slv_shd_wdn_r1.uti"},
	{0x9A39FB470DE23BF5ULL, "vxlt_0_lowtown_a3_n_c_hku_0_0.vlm"},
	{0x9A3A4B08DD59E7C1ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyn_0t.dds"},
	{0x9A3ADE8F5AE4A729ULL, "6087804_m.fxe"},
	{0x9A3B4049F76D387AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopen_v2.ani"},
	{0x9A3B6C82BDB022F1ULL, "vxw_bonepit_mz_0_0.vlm"},
	{0x9A3BFCD0F52526AEULL, "art\\vfx\\combat\\vfx_g_appearance_c.anb"},
	{0x9A3C2C33D40E2469ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.impa.gad"},
	{0x9A3C688DD3A6F35BULL, "vxlt_gallowstemplar_n_518_0_0.vlm"},
	{0x9A3C6F3E9BCC488EULL, "6149882_m.fxe"},
	{0x9A3CBA4AC4792D5BULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masc.phy"},
	{0x9A3CFE473B9C7F25ULL, "vxdr_primevaltomb_7eh_0_1.vlm"},
	{0x9A3D6F650E1F1D7DULL, "hightown_male_noble1_sta6170790.cl"},
	{0x9A3D731596297C5BULL, "808234626.ogg"},
	{0x9A3E057277066D41ULL, "259704710.ogg"},
	{0x9A3E1135914E4229ULL, "465798943.ogg"},
	{0x9A3E3D0526BD8634ULL, "6140490_m.fxe"},
	{0x9A3E5F74B77410D6ULL, "zz_dae_debug6049981.cl"},
	{0x9A3E7F4A57A0C38CULL, "vxca_caverns1_mines_e96_0_0.vlm"},
	{0x9A3E847CCBEBAABFULL, "6034057_m.fxe"},
	{0x9A3F720D6CD721BAULL, "vxdr_primevaltomb_6x8_0_0.vlm"},
	{0x9A3FE534F748FDC3ULL, "6180923_m.fxe"},
	{0x9A3FFED1FE47DA0EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crnf_rdy.ani"},
	{0x9A40058FECB050A6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_table_01_1_n.dds"},
	{0x9A405AC03DC1F8A6ULL, "gen00fl_anders6119548.cl"},
	{0x9A4139388B8DF2B7ULL, "342536732.ogg"},
	{0x9A4156AE45F371B3ULL, "486838037.ogg"},
	{0x9A41952EB0B683C7ULL, "vxlt_gallowstemplar_n_2mw_0_0.vlm"},
	{0x9A41EC43DB166FF9ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.cess.ani"},
	{0x9A421625E3038B63ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v3.ani"},
	{0x9A428032FA5F9645ULL, "vxlt_warehouse_8f_0_0.vlm"},
	{0x9A42A02F202EF548ULL, "lightning4.mao"},
	{0x9A42F953A4252D31ULL, "vxht_brothel_n_6bx_0_0.vlm"},
	{0x9A42FD33D413F2B7ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s2_c.impa.ani"},
	{0x9A4366280D56C4DAULL, "222514554.ogg"},
	{0x9A43CE11C635024CULL, "6044651_m.fxe"},
	{0x9A4443DE9033F63CULL, "hlac_dr_0_deeproads_5849_7y.dds"},
	{0x9A44944D7703891FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_forward_1.gad"},
	{0x9A44FFD36B807198ULL, "dae000_gamlen6149443.cl"},
	{0x9A4573A1E3F1A193ULL, "lt_0_lowtown_14_0.rml"},
	{0x9A4683CF1F117FFBULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.cess.ani"},
	{0x9A4693AB9D16B865ULL, "6101333_m.fxe"},
	{0x9A46C666C7AE0720ULL, "dae000_mother_blightlands_1.stg"},
	{0x9A46E4288A268FD9ULL, "vxca_caverns1_mines_dqb_0_0.vlm"},
	{0x9A471D75367E0EDAULL, "6000820_m.fxe"},
	{0x9A4836CE07CDEFF2ULL, "vxlt_darktown_40t_0_0.vlm"},
	{0x9A483C25A33A51DBULL, "260589289.ogg"},
	{0x9A4842977C01E63AULL, "104280052.ogg"},
	{0x9A48C743DB1C4F3BULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.cess.gad"},
	{0x9A48D9378310C3B3ULL, "dae000_orana6149078.cl"},
	{0x9A48DD86367F27FDULL, "796562044.ogg"},
	{0x9A48FFED152513A4ULL, "art\\vfx\\combat\\vfx_gib_hm_masc.msh"},
	{0x9A4915ED15253946ULL, "art\\vfx\\combat\\vfx_gib_hm_masc.mmh"},
	{0x9A491E8E440AD497ULL, "lgt220_shp_mag_guard6149884.cl"},
	{0x9A492FACB3DB1EA8ULL, "gen00fl_dog6127281.cl"},
	{0x9A4976EE8D5AB389ULL, "25243883.ogg"},
	{0x9A49B3642E883A47ULL, "810616441.ogg"},
	{0x9A49D4312BFEDD3FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10a.msh"},
	{0x9A49DA312BFEE7E1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10a.mmh"},
	{0x9A49DE5FAFA63B71ULL, "vxw_bonepit_hy_0_0.vlm"},
	{0x9A4ADDCDBAFA941BULL, "t3_wdg_w15.tnt"},
	{0x9A4AE11B8DC2610EULL, "6142982_m.fxe"},
	{0x9A4B11CEDF5236B9ULL, "vxlt_0_lowtown_docks_a3_cw1_0_1.vlm"},
	{0x9A4B368845A84337ULL, "403674988.ogg"},
	{0x9A4B4FFF55975621ULL, "575046716.ogg"},
	{0x9A4B793F0F9B79DEULL, "dre000_sandal6031375.cl"},
	{0x9A4B88891A46AF3DULL, "fo_02_f.mao"},
	{0x9A4B894D7709950CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_forward_1.evt"},
	{0x9A4C28C1AA955278ULL, "250954111.ogg"},
	{0x9A4CBF9F03F22F7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_noway.ani"},
	{0x9A4CF659BDE1C33FULL, "6031357_m.fxe"},
	{0x9A4D058CB7E01D1EULL, "927639080.ogg"},
	{0x9A4D44BCA0F7BF03ULL, "vxht_mansion_n_53q_0_0.vlm"},
	{0x9A4D483AFABBA28CULL, "6105884_m.fxe"},
	{0x9A4D5A471BE24E1BULL, "vxlt_undercitydungeon_1nq_0_0.vlm"},
	{0x9A4DB1D3797FE115ULL, "104208301.ogg"},
	{0x9A4DC9B5E7547D46ULL, "dae000_mother6121051.cl"},
	{0x9A4E9E7DBA7247DDULL, "6110615_f.fxe"},
	{0x9A4F2E25E30E3E5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v3.evt"},
	{0x9A4F7708755E27C9ULL, "358696479.ogg"},
	{0x9A4FE443DB229F20ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_lp_c.cess.evt"},
	{0x9A500092C4FBBE4CULL, "vxht_chantry_n_36k_0_0.vlm"},
	{0x9A50193190FF1766ULL, "vxdr_primevaltomb_6yw_0_0.vlm"},
	{0x9A503E5359681949ULL, "gen_im_wep_war_two_act3_07.uti"},
	{0x9A50C4299B1552D5ULL, "6085852_f.fxe"},
	{0x9A518E23DCD8E7E5ULL, "521322547.ogg"},
	{0x9A5221906945AF94ULL, "dre000_carver6037605.cl"},
	{0x9A52366DC915F498ULL, "llac_lt_0_lowtown_a3_n_c_131072_3.dds"},
	{0x9A525DE131D3B25AULL, "hlac_lt_0_lowtown_-196609_1x.dds"},
	{0x9A52936CDB9AB6DBULL, "hlac_lt_undercitydungeon_-1_49y.dds"},
	{0x9A52A31F83DD5236ULL, "static_static_p7.psh.fxo.dx11.assembly"},
	{0x9A52B7E41A67CF77ULL, "889765068.ogg"},
	{0x9A52D3B8BAF7E4ABULL, "6027779_f.fxe"},
	{0x9A5320E314E8BC91ULL, "llac_lt_0_lowtown_a2_n_-65539_0.dds"},
	{0x9A53B5E80098C9F9ULL, "937440837.ogg"},
	{0x9A53CD31E22B6347ULL, "docks_male_human_worker_sta6175952.cl"},
	{0x9A53D7D1FE584148ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_crnf_rdy.gad"},
	{0x9A541C249C742E43ULL, "6143333_m.fxe"},
	{0x9A54618A0AFED7A6ULL, "hlac_ca_caverns1_mines_24424_4x.dds"},
	{0x9A546F48D67EE9F7ULL, "6141717_m.fxe"},
	{0x9A547E238458C883ULL, "6139570_m.fxe"},
	{0x9A5483989517E231ULL, "vxw_blightlands_9u_0_0.vlm"},
	{0x9A548BCF1F1DCAF2ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.cess.evt"},
	{0x9A54ADE350FE0C69ULL, "dev_res_clu.mao"},
	{0x9A54BE63637D8B38ULL, "6152761_m.fxe"},
	{0x9A5573F32F3E5E01ULL, "vxht_mansion_n_3ka_0_0.vlm"},
	{0x9A5598229F6FFC6EULL, "6116075-mrl_pc_wander3.cut"},
	{0x9A5608FAEEA7A73AULL, "vxw_bonepit_zt_0_0.vlm"},
	{0x9A562D25E3145B4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v3.gad"},
	{0x9A568E7F1818120EULL, "vxw_woundedcoast_zu_0_0.vlm"},
	{0x9A57F213744389FDULL, "6047219_m.fxe"},
	{0x9A585D503FEF8FBCULL, "6128657_f.fxe"},
	{0x9A5881B0EA041FD3ULL, "hlac_lt_0_lowtown_docks_a3_196608_0x.dds"},
	{0x9A58DBBF13093318ULL, "gen00fl_aveline6151884.cl"},
	{0x9A59005BEADF44E6ULL, "vxht_brothel_n_4z1_0_0.vlm"},
	{0x9A59714D7715AA65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_forward_1.ani"},
	{0x9A59DCD7270DEAB3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_aoe.evt"},
	{0x9A5AAACF1F226A85ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.cess.gad"},
	{0x9A5AB0A8B9D9B199ULL, "838792118.ogg"},
	{0x9A5AD79F03FE954FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_noway.evt"},
	{0x9A5B0443972E2D78ULL, "hlac_lt_gallowscourtyard_-65536_2x.dds"},
	{0x9A5B48DC694631E8ULL, "6098544_m.fxe"},
	{0x9A5C506A6CE62464ULL, "hlac_dr_primevaltomb_659_0z.dds"},
	{0x9A5C7BCA4342111AULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.dura.gad"},
	{0x9A5CB1C13E0BA0D7ULL, "art\\characters\\weapons\\staff\\w_stf_012a.mmh"},
	{0x9A5CB5C13E0BA7E4ULL, "art\\characters\\weapons\\staff\\w_stf_012a.mao"},
	{0x9A5CC2CAEAC82ABEULL, "lt_0_lowtown_docks_n_116_0.phy"},
	{0x9A5CC7C13E0BC649ULL, "art\\characters\\weapons\\staff\\w_stf_012a.msh"},
	{0x9A5CEA3A09619868ULL, "dre180ar_keep_fake.are"},
	{0x9A5D42E1A54F5AD0ULL, "mag291_demon_marethari6028847.cl"},
	{0x9A5D44B4E14A5559ULL, "one220_yevhen6102806.cl"},
	{0x9A5D882B35BC2492ULL, "llac_lt_0_lowtown_a2_65537_5.dds"},
	{0x9A5DBE745464155EULL, "hlac_w_bonepitmining_a2_5_0y.dds"},
	{0x9A5E23D3A7899317ULL, "6140071_m.fxe"},
	{0x9A5E474C4BA0231AULL, "dae000_arishok6068464.cl"},
	{0x9A5E533AC9B8CA70ULL, "vxlt_0_lowtown_docks_a3_d3j_0_0.vlm"},
	{0x9A5E5F7A5DA6A65CULL, "vxht_mansion_n_3z7_0_0.vlm"},
	{0x9A5EFB4418F1F9F6ULL, "hlac_lt_0_lowtown_docks_-4_5x.dds"},
	{0x9A5F70C89349F7A0ULL, "624452191.ogg"},
	{0x9A60628A3CBC2528ULL, "hlac_lt_0_lowtown_a2_n_c_65537_4x.dds"},
	{0x9A60A88B078047B1ULL, "hlac_lt_0_lowtown_docks_a3_65533_0x.dds"},
	{0x9A60D5D72713FDF4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_aoe.gad"},
	{0x9A60D8A19A1C8905ULL, "6096028_m.fxe"},
	{0x9A614655F2B9C35BULL, "6101285_m.fxe"},
	{0x9A61C3F0C0DC26B4ULL, "hlac_lt_0_lowtown_2_1z.dds"},
	{0x9A61D89F0404B5F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_noway.gad"},
	{0x9A624F315B974E76ULL, "6104433_f.fxe"},
	{0x9A625CB4FB993103ULL, "face_face_p4.psh.fxo.dx11.assembly"},
	{0x9A6294AD29E56BBAULL, "6137519_m.fxe"},
	{0x9A630BDEA1F772BAULL, "6116653_m.fxe"},
	{0x9A6337FDAFA51435ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_1z.dds"},
	{0x9A637ECA43483531ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.dura.evt"},
	{0x9A63EE0BAFEABC2BULL, "qun110_petrice6037136.cl"},
	{0x9A64580643FBA1D1ULL, "mag291_demon_marethari6027966.cl"},
	{0x9A6484876776B8D6ULL, "6025344_m.fxe"},
	{0x9A64A3A8D109E092ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower3_.gad"},
	{0x9A64C763784AFFC6ULL, "568282135.ogg"},
	{0x9A64D8ADF01B3E9CULL, "666400312.ogg"},
	{0x9A65005C0A0E2A8EULL, "vxw_woundedcoast_wv_0_0.vlm"},
	{0x9A6574380CAADB9AULL, "401116599.ogg"},
	{0x9A65AD44456B3506ULL, "6020783_m.fxe"},
	{0x9A6650F75FA48EC3ULL, "1003569717.ogg"},
	{0x9A669D4AD3202746ULL, "hlac_ca_caverns1_24421_8x.dds"},
	{0x9A66D9E87BCE554EULL, "art\\characters\\creatures\\spider\\animation\\rn_p_rf.evt"},
	{0x9A672670759042CDULL, "vxlt_gallowstemplar_n_5z3_0_0.vlm"},
	{0x9A67E4D7271A35D2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_aoe.ani"},
	{0x9A67F5279779A5EBULL, "weaponhighlight.psh.fxo"},
	{0x9A6812C6A513DE0AULL, "8059453.ogg"},
	{0x9A682E512983647EULL, "zz_firestaff2.cut"},
	{0x9A6849E7BCD28C11ULL, "one220_yevhen6098778.cl"},
	{0x9A6853DE117EC8A5ULL, "vxdr_primevaltomb_7tw_0_1.vlm"},
	{0x9A685DE35F15F09AULL, "143228336.ogg"},
	{0x9A68A14AC49E3DFAULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masc.msh"},
	{0x9A68B34AC49E5C40ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masc.mmh"},
	{0x9A68CCEA585FCEA6ULL, "351055805.ogg"},
	{0x9A69389ED446743EULL, "6129285_f.fxe"},
	{0x9A69C36A7EFD2634ULL, "vxlt_darktown_6k1_0_0.vlm"},
	{0x9A6A287684E3F003ULL, "vxht_mansion_n_59s_0_0.vlm"},
	{0x9A6A36305A0DF6ACULL, "hlac_ht_mansion_n_3365_1z.dds"},
	{0x9A6A442561898FA5ULL, "llac_w_blightlands_0_6.dds"},
	{0x9A6A7376557E2CECULL, "gen_im_npc_dwf_arm_cht_cxt3.uti"},
	{0x9A6A9D733273C94BULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_leafs.mao"},
	{0x9A6ABE99DDF88545ULL, "lt_0_lowtown_docks_a3_32_0_ll.mmh"},
	{0x9A6B267CB98E97A2ULL, "qun131_taskmaster6059001.cl"},
	{0x9A6BC2E7F3FB64DFULL, "6105825_m.fxe"},
	{0x9A6C59CB4D943D09ULL, "20150094.ogg"},
	{0x9A6C5CCE9A1ADDB0ULL, "injuries.gda"},
	{0x9A6C635D39D24534ULL, "vxca_cavern1_haunted_gi8_0_0.vlm"},
	{0x9A6C6BBC25FA77F3ULL, "175143134.ogg"},
	{0x9A6C75A069AB0BD3ULL, "lgt310_mis_burn_rival_thug.bnk"},
	{0x9A6CC517BBAEE2F0ULL, "vxlt_0_lowtown_docks_a3_d4k_0_0.vlm"},
	{0x9A6D416D5FC2941DULL, "face0_ds2y.psh.fxo"},
	{0x9A6D94F9FEC71EC2ULL, "295375692.ogg"},
	{0x9A6D970BEC65D3F7ULL, "6050284_m.fxe"},
	{0x9A6DD0E87BD464F9ULL, "art\\characters\\creatures\\spider\\animation\\rn_p_rf.gad"},
	{0x9A6DFB661032278CULL, "6111755_m.fxe"},
	{0x9A6E2257B2585B28ULL, "133134946.ogg"},
	{0x9A6EC6B308020284ULL, "vxw_bonepit_20p_0_0.vlm"},
	{0x9A6F1CA971001088ULL, "backgrounds.gda"},
	{0x9A6F97A069AD7B23ULL, "lgt310_mis_burn_rival_thug.cnv"},
	{0x9A7018BA38DF26ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_surrounding_female.ani"},
	{0x9A702D91276321B1ULL, "6060731_m.fxe"},
	{0x9A70A26542829B37ULL, "vxlt_warehouse_n_hy_0_0.vlm"},
	{0x9A70CFAA0330A973ULL, "680270435.ogg"},
	{0x9A70D1BE40330911ULL, "vxw_woundedcoast_5o_0_0.vlm"},
	{0x9A70E37F386AF95EULL, "zz_mag_debug6089452.cl"},
	{0x9A718E09933609B8ULL, "vxlt_gallowstemplar_n_53i_0_0.vlm"},
	{0x9A7196CA43549B00ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.dura.ani"},
	{0x9A71D22F70375C10ULL, "vxht_0_hightown_a2_c6p_0_0.vlm"},
	{0x9A721C948814FE3AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mc.mmh"},
	{0x9A72209488150571ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mc.mao"},
	{0x9A722E9488151C80ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12mc.msh"},
	{0x9A723B7C9E78D487ULL, "vxw_encounter2_7r_0_0.vlm"},
	{0x9A7289928B45D331ULL, "vxw_sundermount_26h_0_0.vlm"},
	{0x9A735A64D6274A9AULL, "vxca_cavern1_haunted_dpr_0_0.vlm"},
	{0x9A73DFB7F545914EULL, "gen00fl_merrill6121097.cl"},
	{0x9A73E1E87BD8ED47ULL, "art\\characters\\creatures\\spider\\animation\\rn_p_rf.ani"},
	{0x9A74126A56E83076ULL, "vxlt_0_lowtown_euq_0_0.vlm"},
	{0x9A741F2137614D86ULL, "6027683_m.fxe"},
	{0x9A74F515B25C7CABULL, "vxht_mansion2_1n1_0_0.vlm"},
	{0x9A752686B35076B0ULL, "hlac_ht_keep_n_3731_0z.dds"},
	{0x9A753AAA5FF139DAULL, "gen00fl_fenris6112121.cl"},
	{0x9A75747232D0F904ULL, "root_rog_general2.dds"},
	{0x9A75DB312C24245EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10a.phy"},
	{0x9A75FB92A79A7D08ULL, "vxlt_0_lowtown_docks_arw_0_0.vlm"},
	{0x9A7625BA38E3A82EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_surrounding_female.gad"},
	{0x9A764DF4D5CD698BULL, "vxlt_darktown_5kl_0_1.vlm"},
	{0x9A769F8448C42B80ULL, "121112756.ogg"},
	{0x9A76C97E312A3AB6ULL, "vxlt_backalley_n_1kk_0_0.vlm"},
	{0x9A789A5153C41CDFULL, "w_sundermount_6_0.msh"},
	{0x9A78A05153C42681ULL, "w_sundermount_6_0.mmh"},
	{0x9A78A1D25FD7BC12ULL, "6161163_m.fxe"},
	{0x9A78AC5153C43B22ULL, "w_sundermount_6_0.mao"},
	{0x9A7902A3F254D6F9ULL, "6139868_m.fxe"},
	{0x9A7908C806C963CEULL, "47309243.ogg"},
	{0x9A7914AAD8D4D40CULL, "6148227_m.fxe"},
	{0x9A7942E73727B2D7ULL, "6119895_m.fxe"},
	{0x9A795E713684CF44ULL, "vxlt_gallowstemplar_n_3ok_0_0.vlm"},
	{0x9A7961AAC0405556ULL, "one100pt_herb_glans.ncs"},
	{0x9A7985EDB582EE8CULL, "vxht_mansion_n_5cz_0_0.vlm"},
	{0x9A798EA8D11C1938ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower3_.ani"},
	{0x9A79958AEA569CACULL, "vxht_mansion2b_3ey_0_0.vlm"},
	{0x9A7A236AEED8F533ULL, "6174032_m.fxe"},
	{0x9A7AD2FDCD00950AULL, "dre000_dougal6129833.cl"},
	{0x9A7AED102AB0AAC0ULL, "llac_lt_0_lowtown_a3_n_c_65534_0.dds"},
	{0x9A7B24BD8F720ED7ULL, "llac_w_bonepitdestroyed_a3_131073_0.dds"},
	{0x9A7B2ED6E871D00AULL, "hlac_lt_0_lowtown_131071_5y.dds"},
	{0x9A7BB96C29EC4819ULL, "623598048.ogg"},
	{0x9A7BC52A58F828CAULL, "797113231.ogg"},
	{0x9A7C5FF21A42036BULL, "6142796_m.fxe"},
	{0x9A7D20BA38E9BE0DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_surrounding_female.evt"},
	{0x9A7D647CCBDE7B8EULL, "54164200.ogg"},
	{0x9A7DCC5E1000A3D3ULL, "hlac_ht_chantry_n_2534_43x.dds"},
	{0x9A7E218280AEC956ULL, "hlac_ht_mansion2b_818_52y.dds"},
	{0x9A7E9419F4BFD49CULL, "6112664_m.fxe"},
	{0x9A7ECDAA8D856CE8ULL, "w_bonepitmining_a2_57_0.mmh"},
	{0x9A7ED9AA8D85808BULL, "w_bonepitmining_a2_57_0.mao"},
	{0x9A7EEBAA8D859F12ULL, "w_bonepitmining_a2_57_0.msh"},
	{0x9A7F14423478B3B7ULL, "vxlt_warehouse_n_ox_0_0.vlm"},
	{0x9A7F23286A2951D0ULL, "731145124.ogg"},
	{0x9A7F439B32292191ULL, "vxw_woundedcoast_2n_0_0.vlm"},
	{0x9A7FFFE6852C2238ULL, "vxlt_gallowstemplar_n_56f_0_0.vlm"},
	{0x9A8065C03D5F6492ULL, "hlac_lt_0_lowtown_n_131070_0z.dds"},
	{0x9A80735804E06C82ULL, "6106968_m.fxe"},
	{0x9A80A60C49B45AACULL, "llac_lt_0_lowtown_a2_-196608_2.dds"},
	{0x9A80AD59906EED07ULL, "vxw_encounter2_2m_0_0.vlm"},
	{0x9A80DAA57DA9FCDAULL, "dae000ip_crowd_em2_talk1.utp"},
	{0x9A81067921A3D5C1ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal.phy"},
	{0x9A813835DF0FA238ULL, "vxlt_0_lowtown_docks_be4_0_1.vlm"},
	{0x9A815E91C489F0BAULL, "vxw_bonepit_kw_0_0.vlm"},
	{0x9A816CEFDDE61AB5ULL, "fx_black_line.dds"},
	{0x9A8182D7C0B737A5ULL, "429104878.ogg"},
	{0x9A81A67E660C82DEULL, "vxca_caverns1_mines_jw1_0_0.vlm"},
	{0x9A81BE2765583B40ULL, "pro000_athenril6049466.cl"},
	{0x9A81F34B30DCBC38ULL, "hlac_lt_0_lowtown_-131073_2y.dds"},
	{0x9A81FB0F3B79C21FULL, "lgt200_eau_tome6180579.cl"},
	{0x9A822C076BD68900ULL, "mag211_marethari6025329.cl"},
	{0x9A831616FB6A3840ULL, "6108533_m.fxe"},
	{0x9A83BEC28598301BULL, "lt_0_lowtown_a2_n_c_35_0.rml"},
	{0x9A842A19C9E081E4ULL, "hlac_dr_0_deeproads_5843_2x.dds"},
	{0x9A846322739EDE7EULL, "vxht_0_hightown_n_5y2_0_0.vlm"},
	{0x9A84E3EBCAB91D1EULL, "hlac_lt_gallowscourtyard_-65536_3y.dds"},
	{0x9A853B5B23205336ULL, "vxlt_backalley_n_1vl_0_0.vlm"},
	{0x9A854CEF8BEB6770ULL, "108230185.ogg"},
	{0x9A85563AB543F58BULL, "vxca_caverns2_7xv_0_0.vlm"},
	{0x9A85961B3722763EULL, "static_static_p87.psh.fxo.dx11"},
	{0x9A85B655408B431FULL, "vxlt_warehouse_cy_0_0.vlm"},
	{0x9A8610FAEC5F5D84ULL, "6098197_m.fxe"},
	{0x9A86211421EBE6B2ULL, "vxht_mansion2b_3gi_0_0.vlm"},
	{0x9A86440AEF73C436ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_002a_0s.dds"},
	{0x9A86CFD3102CE328ULL, "91865881.ogg"},
	{0x9A86FC7C689183A0ULL, "vxlt_0_lowtown_n_f0h_0_0.vlm"},
	{0x9A874ACA957CE889ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtalentrapidattk.ani"},
	{0x9A88C68C0CF9F58FULL, "144951164.ogg"},
	{0x9A89245080DF6A17ULL, "6048249_m.fxe"},
	{0x9A895C184B8AE9B3ULL, "6031821_m.fxe"},
	{0x9A89ABC07A84661AULL, "53591070.ogg"},
	{0x9A8A0035D09FD382ULL, "qcr000_arishok6060015.cl"},
	{0x9A8A5E5F517A91D8ULL, "734040248.ogg"},
	{0x9A8A5EC4B37EA7FAULL, "106118289.ogg"},
	{0x9A8AFF1D1F6F96EAULL, "hlac_lt_0_lowtown_-65534_0y.dds"},
	{0x9A8B1FD68C32D324ULL, "6098838_m.fxe"},
	{0x9A8B69D9C47F666AULL, "6069670_m.fxe"},
	{0x9A8B9CF0E6B73A82ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.cess.gad"},
	{0x9A8BA8B7D6626790ULL, "hlac_dr_dungeon_3216_12y.dds"},
	{0x9A8C65B6762C0C67ULL, "60946114.ogg"},
	{0x9A8D66785E011FB1ULL, "131550688.ogg"},
	{0x9A8D9B39EC9E2A24ULL, "vxht_mansion2_3bl_0_0.vlm"},
	{0x9A8DF430E6AC44FDULL, "ave331_jeven6138724.cl"},
	{0x9A8E4C72AB4E5916ULL, "79173764.ogg"},
	{0x9A8E65CA9583350BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtalentrapidattk.gad"},
	{0x9A8EB538BF9E2586ULL, "mrl000_marethari6129340.cl"},
	{0x9A8F667CB57E6533ULL, "atl_achievemen_dxt5_dat.xml"},
	{0x9A8F7BAD9C867A6DULL, "268065000.ogg"},
	{0x9A900A8379FEF851ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyn_0d.dds"},
	{0x9A902BCE89362537ULL, "hlac_ca_caverns2_5802_29y.dds"},
	{0x9A9044ACC0A5FD21ULL, "gfxfontlib.gfx"},
	{0x9A9073883AAF07CCULL, "one100pt_magistrate.ncs"},
	{0x9A90992E6E580533ULL, "6018423_f.fxe"},
	{0x9A91800D8408FDAEULL, "326993921.ogg"},
	{0x9A92468C5F098B31ULL, "6167301_f.fxe"},
	{0x9A927C074B728C4DULL, "vxlt_gallowstemplar_n_5k0_0_0.vlm"},
	{0x9A92BFF0E6BD9579ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.cess.evt"},
	{0x9A933C7B8E63F2E9ULL, "vxht_brothel_n_57p_0_0.vlm"},
	{0x9A937E7FCB45C109ULL, "98887399.ogg"},
	{0x9A94283232815B9FULL, "vxlt_warehouse_dx_0_0.vlm"},
	{0x9A9444AF7001843DULL, "follower_banter6173007.cl"},
	{0x9A944F3047774A17ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03a.phy"},
	{0x9A947D1EF0D96206ULL, "14967872.ogg"},
	{0x9A9562CA95894F70ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtalentrapidattk.evt"},
	{0x9A957158150D3314ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_left_hand.ani"},
	{0x9A95D8DC631765ADULL, "659607712.ogg"},
	{0x9A967213446B8726ULL, "vxca_caverns1_ijd_0_0.vlm"},
	{0x9A9697C03F1A7E0CULL, "vxlt_hangedman_212_0_0.vlm"},
	{0x9A96EF99A66003D5ULL, "ser221_gascard6133753.cl"},
	{0x9A974F327ADA465AULL, "narrative_varric6172499.cl"},
	{0x9A98CC7DB87F9FB2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_dummy_d.dds"},
	{0x9A98E37C1B37AF01ULL, "58535771.ogg"},
	{0x9A9A2790BE8C74C2ULL, "6170352_m.fxe"},
	{0x9A9AB0A27633520AULL, "zz_stever_test.cut"},
	{0x9A9AB7A276335E15ULL, "zz_stever_test.cnv"},
	{0x9A9B5972068668D9ULL, "6064784_m.fxe"},
	{0x9A9B7ED2353657A5ULL, "vxht_mansion_n_3yg_0_0.vlm"},
	{0x9A9BBA0655CC4730ULL, "gen00fl_isabela6096814.cl"},
	{0x9A9BD98963106A78ULL, "montage_subs_demohack6217105.cl"},
	{0x9A9BDAAFB0694820ULL, "hlac_w_bonepitdestroyed_a3_458757_0z.dds"},
	{0x9A9BF7FC1864E4B7ULL, "vxlt_warehouse_n_yz_0_0.vlm"},
	{0x9A9C6E5815134D76ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_left_hand.gad"},
	{0x9A9CA1C551FAA8F5ULL, "llac_ht_mansion2_818_63.dds"},
	{0x9A9CADCF42D5A886ULL, "vxw_bonepitdestroyed_a3_1wq_0_1.vlm"},
	{0x9A9CD32E0C070456ULL, "hm_azurejamos.mor"},
	{0x9A9CF5080C0A38AAULL, "216466823.ogg"},
	{0x9A9D805153E30CAAULL, "w_sundermount_6_0.rml"},
	{0x9A9D9113745B1E07ULL, "vxw_encounter2_8o_0_0.vlm"},
	{0x9A9E4C489471B8DAULL, "292422975.ogg"},
	{0x9A9E66D8D77D3037ULL, "6064154_f.fxe"},
	{0x9A9EF15F06D56B91ULL, "dre000_varric6167457.cl"},
	{0x9A9F0421B1C446CEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoortrim_sa.dds"},
	{0x9A9FDEB7879FA399ULL, "641233561.ogg"},
	{0x9AA02BD85B51C90BULL, "vxlt_0_lowtown_docks_a3_ct3_0_0.vlm"},
	{0x9AA034B309711A97ULL, "6040920_m.fxe"},
	{0x9AA085EFDE00F5BFULL, "fx_black_line.mao"},
	{0x9AA0A5B079983BB7ULL, "vxlt_darktown_4ur_0_0.vlm"},
	{0x9AA0AF76D764068CULL, "vxca_caverns1_mines_fmq_0_0.vlm"},
	{0x9AA0C7F0E6C9E068ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.cess.ani"},
	{0x9AA0F00B4E86B7F1ULL, "one280_emrys6137199.cl"},
	{0x9AA0FB74F8B4703BULL, "920384315.ogg"},
	{0x9AA12A59B328D2B6ULL, "gen00fl_isabela6122908.cl"},
	{0x9AA20FB689D58452ULL, "vxlt_0_lowtown_docks_a3_bt7_0_0.vlm"},
	{0x9AA22995D78F06F4ULL, "gen_im_arm_cht_nobc_t3.uti"},
	{0x9AA2D8AEAD7E46EBULL, "88000081.ogg"},
	{0x9AA2EB4823EE92F9ULL, "vxw_sundermount_2wc_0_0.vlm"},
	{0x9AA32DD478EFE9DBULL, "follower_banter6155658.cl"},
	{0x9AA345CF1833AB3CULL, "hlac_w_blightlands_-65533_0z.dds"},
	{0x9AA38958151999D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_gestures_left_hand.evt"},
	{0x9AA3F93CA8164617ULL, "825322081.ogg"},
	{0x9AA4BDF3CE15AF0BULL, "dae000_bartrand6206866.cl"},
	{0x9AA4E5D6C66D1CA5ULL, "vxlt_hangedman_n_1ht_0_0.vlm"},
	{0x9AA4FB78CBEDB070ULL, "6000703_m.fxe"},
	{0x9AA524E0384ED593ULL, "zz_clearfx.ncs"},
	{0x9AA54596D6848C52ULL, "498372074.ogg"},
	{0x9AA55F3435D5BC3EULL, "art\\cinematics\\follower plots\\sebastian\\cs_sebastian_sebastianintro.ani"},
	{0x9AA5F0BA8B62C658ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgta.phy"},
	{0x9AA679C1071E9716ULL, "113955785.ogg"},
	{0x9AA754099DF66B9CULL, "vxlt_gallowsprison_n_62i_0_0.vlm"},
	{0x9AA7CCCBDF530A03ULL, "vxw_encounter2_camp_ri_0_0.vlm"},
	{0x9AA83E8F0C415CDAULL, "mer140_bonepit_comments6062987.cl"},
	{0x9AA8A1538C1B0812ULL, "vxdr_0_deeproads_e0y_0_0.vlm"},
	{0x9AA8AF9DAF3253C0ULL, "6060359_f.fxe"},
	{0x9AA9109C777E696BULL, "vfx_distortionmask_p5.vsh.fxo.dx11.assembly"},
	{0x9AAA44F3D53E2CF3ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.dura.evt"},
	{0x9AAADC95A63FB50CULL, "gen00fl_isabela6079952.cl"},
	{0x9AAAF7ADCF439B78ULL, "vxht_mansion2b_2id_0_0.vlm"},
	{0x9AAB2EA53DC0CDEFULL, "vxdr_0_deeproads_fbi_0_0.vlm"},
	{0x9AAB904E34237CADULL, "dae300cr_vashothlead.utc"},
	{0x9AAC564C90956B07ULL, "vxw_blightlands_ou_0_0.vlm"},
	{0x9AAC742FE7C806CCULL, "6049740_m.fxe"},
	{0x9AAD0141A535CBFAULL, "vxht_keep_n_62t_0_0.vlm"},
	{0x9AAD50F68BE5892EULL, "gen_im_wep_war_two_lore_05.uti"},
	{0x9AADC12568BE6768ULL, "class_abilities_fenris.gda"},
	{0x9AAE00A56BC09027ULL, "gen_im_trs_crc_sta_phy_r2.uti"},
	{0x9AAE2B3C2EDD95DEULL, "hlac_w_sundermount_-65539_0x.dds"},
	{0x9AAE48E89214D6F7ULL, "6118085_m.fxe"},
	{0x9AAE6517E8A5904EULL, "blood_cycle_l2.mao"},
	{0x9AAE86F47460F94FULL, "6113296_f.fxe"},
	{0x9AAF6B9B796CE235ULL, "mag170_thrask6020536.cl"},
	{0x9AAF95BB962A4CCBULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar4_mcr370_playerflees.ani"},
	{0x9AB1317338B2EB19ULL, "atl_warninglig_dxt5_00.dds"},
	{0x9AB13DF3D5443F34ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.dura.gad"},
	{0x9AB14BC2B2AF9D30ULL, "w_woundedcoast_96_0.msh"},
	{0x9AB159C2B2AFB54AULL, "w_woundedcoast_96_0.mmh"},
	{0x9AB15DC2B2AFBB81ULL, "w_woundedcoast_96_0.mao"},
	{0x9AB16BDA8F772E0EULL, "qun111_ketojan6040709.cl"},
	{0x9AB22C38D51493CAULL, "6061696_m.fxe"},
	{0x9AB29B286491B927ULL, "vxca_caverns2_8pm_0_0.vlm"},
	{0x9AB2AAA82AB203B8ULL, "405379189.ogg"},
	{0x9AB37288504CD4BBULL, "hf_patron1.mor"},
	{0x9AB3907C55695870ULL, "hlac_w_encounter2_camp_262144_0y.dds"},
	{0x9AB4472062AA3E3EULL, "6139265_m.fxe"},
	{0x9AB49A2CDCBF6138ULL, "vxlt_0_lowtown_a3_n_c_enw_0_1.vlm"},
	{0x9AB4C3A29BC71CD6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_loop.ani"},
	{0x9AB541B8D2A15C85ULL, "vxht_bartrand_n_668_0_0.vlm"},
	{0x9AB5A0B2CC4B817AULL, "hm_bar_addict.mor"},
	{0x9AB5A38E7CBC94F9ULL, "vxht_chantry_1yj_0_0.vlm"},
	{0x9AB5F432FD2E9B8DULL, "593029056.ogg"},
	{0x9AB621E52A176A4AULL, "and321tr_endstage_approval.ncs"},
	{0x9AB63EA8D1492283ULL, "vxw_encounter2_camp_mh_0_0.vlm"},
	{0x9AB6A46666C92FD4ULL, "hightown_fem_cityguard_mob6168488.cl"},
	{0x9AB6DB1FEAD8C501ULL, "428209994.ogg"},
	{0x9AB711BE2DA61217ULL, "follower_banter6149178.cl"},
	{0x9AB7345F7820A075ULL, "llac_w_encounter2_262142_0.dds"},
	{0x9AB74CF3D548C412ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.dura.ani"},
	{0x9AB7AEFA693BFB1AULL, "141256591.ogg"},
	{0x9AB7D26FB5AA72CCULL, "one100_solivitus6048116.cl"},
	{0x9AB8107F38B35E50ULL, "6125499_m.fxe"},
	{0x9AB827722CD2120BULL, "vxlt_0_lowtown_n_bza_0_0.vlm"},
	{0x9AB8E0F5085FC7ABULL, "vxht_bartrand_n_5qp_0_0.vlm"},
	{0x9AB8E418EE003A1FULL, "mag121_idunna6084228.cl"},
	{0x9AB8F11A4B8AD1FFULL, "795414739.ogg"},
	{0x9AB9079BFC4ADA63ULL, "vxht_0_hightown_awr_0_0.vlm"},
	{0x9AB93519AA15A6D9ULL, "llac_lt_0_lowtown_n_-2_0.dds"},
	{0x9AB93AD9D147D543ULL, "lla_ht_0_hightown_a3_0_0.dds"},
	{0x9AB9783435E68FB8ULL, "art\\cinematics\\follower plots\\sebastian\\cs_sebastian_sebastianintro.gad"},
	{0x9AB9FFAD831AB4FCULL, "zz_fol_debug6092320.cl"},
	{0x9ABA7B55DF9DBAF6ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.cess.ani"},
	{0x9ABA9255C932E45DULL, "215349393.ogg"},
	{0x9ABADD235876ABE6ULL, "6052953_m.fxe"},
	{0x9ABADFDAA3295006ULL, "vxdr_0_deeproads_dpd_0_0.vlm"},
	{0x9ABAE60CD9BE1F4CULL, "6044440_m.fxe"},
	{0x9ABB423879D265AEULL, "6180423_m.fxe"},
	{0x9ABB4E7A5A00BF17ULL, "629670861.ogg"},
	{0x9ABBDB3DADFEDF0AULL, "vxht_0_hightown_n_c7z_0_0.vlm"},
	{0x9ABBE4219122D1DEULL, "vxht_keep_62k_0_0.vlm"},
	{0x9ABC7B3E679D5A55ULL, "6176029_m.fxe"},
	{0x9ABCAA6D01DAE2A6ULL, "vxlt_warehouse_q6_0_0.vlm"},
	{0x9ABCB81FC71FE3F5ULL, "vxw_woundedcoast_1c4_0_0.vlm"},
	{0x9ABCE2A62653D8C0ULL, "lt_0_lowtown_26_0.anb"},
	{0x9ABD92F532F0FD3FULL, "vxht_0_hightown_a3_b5r_0_0.vlm"},
	{0x9ABF7DC90863A51FULL, "vxlt_warehouse_we_0_0.vlm"},
	{0x9ABF821DA5BD6CE4ULL, "vxdr_0_deeproads_6z7_0_1.vlm"},
	{0x9AC0905DE5E4C4DBULL, "hlac_ca_caverns1_24422_7y.dds"},
	{0x9AC09B2B60C065FFULL, "w_bonepitmining_a2_199_0.msh"},
	{0x9AC0A12B60C06FA1ULL, "w_bonepitmining_a2_199_0.mmh"},
	{0x9AC0AD2B60C08442ULL, "w_bonepitmining_a2_199_0.mao"},
	{0x9AC0B6B3D6D8EE53ULL, "hlac_ht_0_hightown_a3_-196608_0z.dds"},
	{0x9AC0EC2B2FE80ABFULL, "6081375_f.fxe"},
	{0x9AC1B1ADEA8D153EULL, "396489876.ogg"},
	{0x9AC1CBA29BD1B4C7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_loop.evt"},
	{0x9AC1DED01B6E0438ULL, "vxca_caverns1_k2q_0_1.vlm"},
	{0x9AC1E44E3A6E0D75ULL, "324108331.ogg"},
	{0x9AC20346546D1B5EULL, "vxlt_gallowstemplar_n_4yd_0_0.vlm"},
	{0x9AC2843766C7F402ULL, "6086039_m.fxe"},
	{0x9AC2E7960B79CF02ULL, "6141598_m.fxe"},
	{0x9AC3378AE8A10B8EULL, "vxca_caverns1_eft_0_0.vlm"},
	{0x9AC337C3A8A8C34DULL, "vxlt_0_lowtown_docks_n_bdx_0_0.vlm"},
	{0x9AC35B7B7146740CULL, "6041025_m.fxe"},
	{0x9AC36778CA4CFDEDULL, "66735921.ogg"},
	{0x9AC37CBB963ACB55ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar4_mcr370_playerflees.gad"},
	{0x9AC3BF2CE1A72C00ULL, "w_encounter2_camp_probe_qh.dds"},
	{0x9AC3C784635F14D2ULL, "207671342.ogg"},
	{0x9AC426915E1C1E7CULL, "16449299.ogg"},
	{0x9AC432CEA7D12C5CULL, "121716808.ogg"},
	{0x9AC434756E84524EULL, "6201471_f.fxe"},
	{0x9AC4B085C33F3B03ULL, "vxw_encounter2_camp_hk_0_0.vlm"},
	{0x9AC4C1C285CFCCE7ULL, "lt_0_lowtown_a2_n_c_35_0.anb"},
	{0x9AC4F861C565CC72ULL, "304862225.ogg"},
	{0x9AC4FC2214281E27ULL, "438200578.ogg"},
	{0x9AC52F58A0AF88F5ULL, "gen_im_arm_cht_nobb.uti"},
	{0x9AC56668CA47ED03ULL, "vxw_bonepit_1wp_0_1.vlm"},
	{0x9AC5833395E8B025ULL, "6083736_m.fxe"},
	{0x9AC58B0F9820326BULL, "526339735.ogg"},
	{0x9AC6267F074CF321ULL, "6139136_m.fxe"},
	{0x9AC68A304C1BBA99ULL, "vxht_0_hightown_av3_0_0.vlm"},
	{0x9AC7083110F1E828ULL, "6080630_f.fxe"},
	{0x9AC71F8E3BD54178ULL, "dae300_alistair_king6134889.cl"},
	{0x9AC752D1FA55E02BULL, "vxht_bartrand_n_5ts_0_0.vlm"},
	{0x9AC7659BF242BCC1ULL, "vxht_mansion2_264_0_0.vlm"},
	{0x9AC78355DFA852E7ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.cess.evt"},
	{0x9AC7BEDEE8B9BF8DULL, "vxlt_blackemporium_v4_0_0.vlm"},
	{0x9AC800BD10637C47ULL, "dae000_katriela6166398.cl"},
	{0x9AC8125F21ACFEEFULL, "vxdr_0_deeproads_fhk_0_0.vlm"},
	{0x9AC840064209A04FULL, "cod_hst_seekers.utp"},
	{0x9AC8D2796758BC2BULL, "hlac_ht_0_hightown_a3_0_3z.dds"},
	{0x9AC8ECA29BD80B10ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_loop.gad"},
	{0x9AC927C032CCEC89ULL, "vxht_0_hightown_n_a2t_0_0.vlm"},
	{0x9AC98708B7B442C5ULL, "vxht_keep_n_67p_0_0.vlm"},
	{0x9AC98EBB2100CEF2ULL, "271285746.ogg"},
	{0x9ACA044CC7A4EFA5ULL, "620001815.ogg"},
	{0x9ACA1B34EE214B43ULL, "static_static_p71.psh.fxo.dx11"},
	{0x9ACAADA2765C7144ULL, "zz_stever_test.utc"},
	{0x9ACB1C49F3D0FB26ULL, "vxlt_warehouse_z7_0_0.vlm"},
	{0x9ACB29FCB915FC75ULL, "vxw_woundedcoast_1n3_0_0.vlm"},
	{0x9ACB2CA2760107F7ULL, "6112285_m.fxe"},
	{0x9ACBADCA47980B2EULL, "544126040.ogg"},
	{0x9ACBAE74E5BF9EB8ULL, "w_encounter3_46_0.rml"},
	{0x9ACCA43037FB87B0ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medh.mao"},
	{0x9ACCA83037FB8EC3ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medh.mmh"},
	{0x9ACCB63037FBA61DULL, "art\\characters\\playercharacter\\humanmale\\hlf_medh.msh"},
	{0x9ACCBEDC8DA94EC7ULL, "vxlt_0_lowtown_n_eiw_0_0.vlm"},
	{0x9ACCC35205DF0208ULL, "llac_w_encounter2_196611_0.dds"},
	{0x9ACD22F2D4DC88CBULL, "lt_0_lowtown_n_21_0_lc.mmh"},
	{0x9ACD4420D909D0D7ULL, "749926972.ogg"},
	{0x9ACDEFA5FA59BD9FULL, "vxlt_warehouse_xd_0_0.vlm"},
	{0x9ACE414D849CCD8BULL, "hlac_lt_undercityserialkiller_-1_43z.dds"},
	{0x9ACE6FBABE4509D5ULL, "gen00fl_fenris6126686.cl"},
	{0x9ACE99A0625C59A2ULL, "1002992409.ogg"},
	{0x9ACEA455DFAEA930ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.cess.gad"},
	{0x9ACF08ED1EC5EA11ULL, "vxca_caverns1_e8i_0_0.vlm"},
	{0x9ACF6BAC3E268E72ULL, "217717315.ogg"},
	{0x9ACFABD645C5F084ULL, "gen00fl_fenris6127000.cl"},
	{0x9ACFE3F1256984E3ULL, "hlac_lt_gallowstemplar_f_2394_8z.dds"},
	{0x9AD04FB6BC15E9DFULL, "follower_banter6155496.cl"},
	{0x9AD0A54D58C37082ULL, "llac_w_encounter2_camp_131072_0.dds"},
	{0x9AD1A9A09A9EDBCDULL, "vxlt_0_lowtown_docks_n_baw_0_0.vlm"},
	{0x9AD28F17B914F349ULL, "6075834_m.fxe"},
	{0x9AD2B761AA348F8BULL, "gen_im_trs_crc_sta_nat_r2.uti"},
	{0x9AD2D34746A74E50ULL, "hlac_ht_0_hightown_n_-1_1y.dds"},
	{0x9AD30E9C4B12C605ULL, "6101396_f.fxe"},
	{0x9AD33D0F65533927ULL, "6104615_m.fxe"},
	{0x9AD399BB5CFEA457ULL, "hlac_ht_brothel_n_3370_2y.dds"},
	{0x9AD45AEAA70D06BEULL, "ts_reactive_stamina.gda"},
	{0x9AD47F7FAEA617C9ULL, "hlac_dr_0_deeproads_5845_15y.dds"},
	{0x9AD48DCB18BB86F6ULL, "lt_0_lowtown_docks_9_0_ll.mmh"},
	{0x9AD4B27FAAE1CCD9ULL, "401794228.ogg"},
	{0x9AD4D7CEFBE94A08ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16ma.mmh"},
	{0x9AD4E3CEFBE95EABULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16ma.mao"},
	{0x9AD4EB76BCD7A68DULL, "hlac_ca_caverns2_5802_28x.dds"},
	{0x9AD4F5CEFBE97D32ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16ma.msh"},
	{0x9AD4FC0D3E11D319ULL, "vxht_0_hightown_ay2_0_0.vlm"},
	{0x9AD514E5F9841397ULL, "hlac_w_sundermount_393212_0z.dds"},
	{0x9AD5FF9F8BE63292ULL, "mag221ip_corthwaite_store.utp"},
	{0x9AD6595E4865C72AULL, "39119346.ogg"},
	{0x9AD66CE09F7A6462ULL, "hlac_lt_0_lowtown_a2_n_c_-65538_0z.dds"},
	{0x9AD754750D59A037ULL, "412407418.ogg"},
	{0x9AD7CE3D515D1080ULL, "gen00fl_isabela6118367.cl"},
	{0x9AD80520D7DAF518ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud4high.phy"},
	{0x9AD8240E7B9F6527ULL, "vxlt_gallowstemplar_5jf_0_0.vlm"},
	{0x9AD82BBF275DC008ULL, "vxlt_0_lowtown_docks_b46_0_0.vlm"},
	{0x9AD86F1EE3F2F105ULL, "999955641.ogg"},
	{0x9AD8BEF791EB100AULL, "vxht_0_hightown_n_c9t_0_0.vlm"},
	{0x9AD8EC427D07C68DULL, "gen00fl_sebastian6198198.cl"},
	{0x9AD8F8C791B7483DULL, "640531831.ogg"},
	{0x9AD90BB3568FAC39ULL, "gencr_harvester.utc"},
	{0x9AD94925CA4DD8C7ULL, "6122365_m.fxe"},
	{0x9AD96574E8E78837ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_slow_zone.gad"},
	{0x9AD99BD9AB0C14F5ULL, "vxw_woundedcoast_1i2_0_0.vlm"},
	{0x9AD9FE8EEF4CAB6CULL, "641109609.ogg"},
	{0x9ADA1B157E3A7D51ULL, "hf_ave3_insurgent2.mor"},
	{0x9ADA7AC9B9DEED01ULL, "hlac_w_encounter3_0_3y.dds"},
	{0x9ADAEF16DDFA1135ULL, "69424248.ogg"},
	{0x9ADB303047B3180CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03a.mmh"},
	{0x9ADB363047B322FEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03a.msh"},
	{0x9ADB4DED11CE91B9ULL, "6177431_m.fxe"},
	{0x9ADB91F8BC9D15BEULL, "101442228.ogg"},
	{0x9ADB9C827AFFD5D6ULL, "6025133_m.fxe"},
	{0x9ADBC82C3D40B714ULL, "474298730.ogg"},
	{0x9ADC86A1AAF0076EULL, "mag360_samson6042039.cl"},
	{0x9ADC9A0622CCBDA4ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyo_0s.dds"},
	{0x9ADD3E7B93E1151DULL, "137039543.ogg"},
	{0x9ADDA930380A03E8ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medh.phy"},
	{0x9ADEC9024A1813C2ULL, "43798366.ogg"},
	{0x9ADEF52DE9462F21ULL, "vxlt_darktown_6iv_0_0.vlm"},
	{0x9ADF09E2848F9B81ULL, "6101838_m.fxe"},
	{0x9ADF14EB47B16FA2ULL, "llac_lt_0_lowtown_a3_n_c_65533_1.dds"},
	{0x9ADF7A74E8EC17C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_slow_zone.evt"},
	{0x9ADFBE7263382156ULL, "vxdr_primevaltomb_50g_0_0.vlm"},
	{0x9ADFC5ADD69F65E8ULL, "vxdr_primevaltomb_3vt_0_1.vlm"},
	{0x9AE0DDFDA9A1AC84ULL, "173589704.ogg"},
	{0x9AE13CE5ADBE91EEULL, "dre120pt_reqs_money.ncs"},
	{0x9AE16184CAD7EB6DULL, "vxht_0_hightown_9zs_0_0.vlm"},
	{0x9AE166D5D94510E5ULL, "615267505.ogg"},
	{0x9AE1833C745926D5ULL, "386794761.ogg"},
	{0x9AE1F1A07065D31EULL, "docks_male_elven_worker_sta6176119.cl"},
	{0x9AE21E2D74188DB1ULL, "946485896.ogg"},
	{0x9AE2AD883DB8688FULL, "642962185.ogg"},
	{0x9AE30B7921F6A8FDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal.mao"},
	{0x9AE30F7921F6AF36ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal.mmh"},
	{0x9AE316CA6A2F11EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_spit_wipe.gad"},
	{0x9AE3197921F6C034ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal.msh"},
	{0x9AE342D580E4361BULL, "vxw_woundedcoast_1b9_0_0.vlm"},
	{0x9AE3D19FC12C4833ULL, "pro000_refugee_female6052095.cl"},
	{0x9AE3EA296C365A32ULL, "mrl000_marethari6069945.cl"},
	{0x9AE55DE9EAA1B6E5ULL, "791009722.ogg"},
	{0x9AE5703B4D7EE1F1ULL, "6121498_m.fxe"},
	{0x9AE5B1DB8504305EULL, "gen00fl_player6207777.cl"},
	{0x9AE615641F6488DAULL, "mcr310_sibling6076408.cl"},
	{0x9AE651A9E2A102C8ULL, "6126884_f.fxe"},
	{0x9AE658D143BBF7A3ULL, "vxht_mansion2_2jz_0_0.vlm"},
	{0x9AE6EC57E220C50CULL, "follower_banter6155523.cl"},
	{0x9AE710C74D501BCCULL, "26111199.ogg"},
	{0x9AE78B935F509152ULL, "vxw_encounter2_camp_q0_0_1.vlm"},
	{0x9AE7A740AFCD92DEULL, "843424599.ogg"},
	{0x9AE7B7415ACF150AULL, "art\\vfx\\materials\\vfx_sleep_z.mao"},
	{0x9AE7BB56DBC8F785ULL, "ss_fell_orden.bnk"},
	{0x9AE7BFFA10F3E631ULL, "gen_im_trs_dal_med_boo_r0.uti"},
	{0x9AE7E760D7A49DE1ULL, "284222795.ogg"},
	{0x9AE7FD418AEB21E6ULL, "195919260.ogg"},
	{0x9AE88D50180B2AA5ULL, "444481811.ogg"},
	{0x9AE937BC5A48AC97ULL, "533068015.ogg"},
	{0x9AE998F0F11584B7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pickaxe_d.dds"},
	{0x9AE9AC74E5D89895ULL, "w_encounter3_46_0.msh"},
	{0x9AE9AE74E5D89C0BULL, "w_encounter3_46_0.mmh"},
	{0x9AE9BA74E5D8B0A8ULL, "w_encounter3_46_0.mao"},
	{0x9AE9F8A02686ECD9ULL, "6180799_f.fxe"},
	{0x9AEA2BCA6A35539CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_spit_wipe.evt"},
	{0x9AEA2EDF8E10456EULL, "hlac_ca_caverns2prc_5802_22x.dds"},
	{0x9AEAC35E5FDF63E5ULL, "395610762.ogg"},
	{0x9AEC1E9D6B584223ULL, "gen00fl_sebastian6179777.cl"},
	{0x9AEC2BE0354584C4ULL, "888954107.ogg"},
	{0x9AEC3B5855FC4E21ULL, "gen00fl_anders6125160.cl"},
	{0x9AECE487AF3F3D04ULL, "vxw_blightlands_14j_0_0.vlm"},
	{0x9AED300A5071B409ULL, "primitives_primitive_p5.vsh.fxo.dx11"},
	{0x9AED7CC7DDEECD2DULL, "qun120cr_javaris6042978.cl"},
	{0x9AED9274E8F87DBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_slow_zone.ani"},
	{0x9AED9D6C2CBFDCC5ULL, "gen_im_acc_amu_act2_05.uti"},
	{0x9AEDA22B60E75F1EULL, "w_bonepitmining_a2_199_0.phy"},
	{0x9AEDC4AAA883DB0FULL, "hf_zmother_h3s1.mor"},
	{0x9AEE2E8E8FED0EE5ULL, "lgt360_mag_orlanna6155389.cl"},
	{0x9AEE43DA32C07610ULL, "prop2.psh.fxo"},
	{0x9AEEB663A1F16144ULL, "dae000_meredith6080988.cl"},
	{0x9AEF758BA15985DFULL, "34577056.ogg"},
	{0x9AEFA1DB65CD93EAULL, "gen00fl_bethany6005719.cl"},
	{0x9AEFB546E257B03AULL, "vxdr_primevaltomb_5ud_0_0.vlm"},
	{0x9AF029ECDA310303ULL, "978364927.ogg"},
	{0x9AF039D674F64111ULL, "380476757.ogg"},
	{0x9AF0430E5047909CULL, "vxw_bonepit_p0_0_0.vlm"},
	{0x9AF0762816C4271BULL, "art\\characters\\playercharacter\\humanmale\\hn_glv_lgtb.mao"},
	{0x9AF0C999CF1CEDF5ULL, "llac_lt_0_lowtown_docks_n_65535_4.dds"},
	{0x9AF0D09A41618DB6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_ascend_male.evt"},
	{0x9AF0DE62593A3285ULL, "435804980.ogg"},
	{0x9AF0DF74FE23BEF2ULL, "vxlt_warehouse_n_xp_0_0.vlm"},
	{0x9AF10B98323F7643ULL, "vxlt_0_lowtown_n_e0t_0_0.vlm"},
	{0x9AF1343AA75DD3D3ULL, "mcr370_negotiation6080278.cl"},
	{0x9AF17126F2152CFBULL, "6037038_m.fxe"},
	{0x9AF1B4B272DA4E9BULL, "vxw_woundedcoast_1m8_0_0.vlm"},
	{0x9AF1EC280250450CULL, "llac_lt_gallowstemplar_n_2391_8.dds"},
	{0x9AF1EEAA6F501A78ULL, "442383961.ogg"},
	{0x9AF226BABA931F5AULL, "mrl351_pridedemon_drain.stg"},
	{0x9AF29B32F6D20FE5ULL, "hlac_lt_darktown_-1_49z.dds"},
	{0x9AF2D2434273918CULL, "seb321_leliana6180901.cl"},
	{0x9AF2E94203771202ULL, "211423764.ogg"},
	{0x9AF2E9805383E395ULL, "6112351_m.fxe"},
	{0x9AF3008B07406BA4ULL, "vxlt_0_lowtown_a2_e1k_0_0.vlm"},
	{0x9AF30C90B81E7309ULL, "hlac_ht_0_hightown_a3_65533_3z.dds"},
	{0x9AF32D082E4AAEE9ULL, "art\\characters\\playercharacter\\humanmale\\rob_appa.phy"},
	{0x9AF3545EB4884A79ULL, "mcr370_cut_marriage6189913.cl"},
	{0x9AF3910ED129D17EULL, "vxht_keep_4ho_0_0.vlm"},
	{0x9AF3B803FF2C98B5ULL, "gen00fl_fenris6127973.cl"},
	{0x9AF3BC85DDDBCC37ULL, "hlac_lt_gallowstemplar_f_2394_2x.dds"},
	{0x9AF417F5D7218FEEULL, "528138607.ogg"},
	{0x9AF464D65CAB384DULL, "532853700.ogg"},
	{0x9AF48A570CECAFB1ULL, "vxw_sundermount_2cg_0_0.vlm"},
	{0x9AF5C942EA7BBBDEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_enter.ani"},
	{0x9AF634763A25F0B1ULL, "669011709.ogg"},
	{0x9AF6AB427AC78F0AULL, "vxht_0_hightown_a2_c1o_0_0.vlm"},
	{0x9AF6DF94C4FF9621ULL, "tbe100ip_shop_codex_11.utp"},
	{0x9AF6EA9C12A914BDULL, "6033432_m.fxe"},
	{0x9AF733CA6A3FEBB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_spit_wipe.ani"},
	{0x9AF7C79A41679D41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_ascend_male.gad"},
	{0x9AF7CF746BAFA2B3ULL, "hlac_ht_keep_n_3741_0z.dds"},
	{0x9AF8297B944AB5B4ULL, "isa111_isabela6164595.cl"},
	{0x9AF86E1FD917DB96ULL, "lt_0_lowtown_docks_154_0.msh"},
	{0x9AF8781FD917EC94ULL, "lt_0_lowtown_docks_154_0.mmh"},
	{0x9AF87C1FD917F3A7ULL, "lt_0_lowtown_docks_154_0.mao"},
	{0x9AF8931D93594586ULL, "6031705_m.fxe"},
	{0x9AF8A298BDA23F73ULL, "762434128.ogg"},
	{0x9AF8B19104EF3647ULL, "6081369_m.fxe"},
	{0x9AF9005CD94FF637ULL, "qun160_vashothlead6064686.cl"},
	{0x9AF998720CFB95C6ULL, "vxlt_gallowstemplar_f_6qj_0_0.vlm"},
	{0x9AFA4440886AE7CCULL, "follower_banter6172304.cl"},
	{0x9AFC930F4ACE05FEULL, "llac_lt_gallowscourtyard_131071_1.dds"},
	{0x9AFCC486BFFD9C4DULL, "804666443.ogg"},
	{0x9AFD53019ECC3A7DULL, "gen00fl_varric6081202.cl"},
	{0x9AFD82094A3774CDULL, "540757074.ogg"},
	{0x9AFDC1A6268B37E8ULL, "lt_0_lowtown_26_0.rml"},
	{0x9AFDDD401A46D2A9ULL, "m_qun221ar_chantry_nt.bnk"},
	{0x9AFE1BFEE1AECF9CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_loop.ani"},
	{0x9AFEB656B801E662ULL, "vxlt_0_lowtown_a3_n_c_fe2_0_0.vlm"},
	{0x9AFEC89A416DBDEFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stairs_ascend_male.ani"},
	{0x9AFF25F190F61FA2ULL, "672293607.ogg"},
	{0x9AFF99F56CCBFF7FULL, "6129046_f.fxe"},
	{0x9B0082F607DCBD1BULL, "6054374_m.fxe"},
	{0x9B00F1942B34362BULL, "6207668_m.fxe"},
	{0x9B011A45C22E422BULL, "vxht_bartrand_n_5hw_0_0.vlm"},
	{0x9B011F423F90FF43ULL, "mcr371_hostile_aveline6082183.cl"},
	{0x9B0132E804DB10FDULL, "follower_banter6141529.cl"},
	{0x9B017DADAEFCC9F5ULL, "w_encounter1_44_0.msh"},
	{0x9B017FADAEFCCD6BULL, "w_encounter1_44_0.mmh"},
	{0x9B018BADAEFCE108ULL, "w_encounter1_44_0.mao"},
	{0x9B01DDBA8BB104EDULL, "art\\characters\\playercharacter\\humanmale\\arm_lgta.msh"},
	{0x9B01EBBA8BB11C20ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgta.mao"},
	{0x9B01EFBA8BB12373ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgta.mmh"},
	{0x9B02B74ECF4B262FULL, "dre000_anders.bnk"},
	{0x9B02CD5342A7BB9EULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_ars_0s.dds"},
	{0x9B02D66FA92ED7DAULL, "zz_qun_debug6088085.cl"},
	{0x9B02E142EA866EAFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_enter.evt"},
	{0x9B02FC33FEE2C831ULL, "vxw_sundermount_2df_0_0.vlm"},
	{0x9B03116084FCE7A4ULL, "vxw_encounter3_el_0_0.vlm"},
	{0x9B0374B0D96FB3FEULL, "streaks01.dds"},
	{0x9B03DD78057AC83AULL, "zz_followers_banter6158241.cl"},
	{0x9B03FF06514D2E72ULL, "vxlt_darktown_4sy_0_0.vlm"},
	{0x9B04027A3A77A27BULL, "plt_cod_aow_ability_mapping.plo"},
	{0x9B04251BD94297F8ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_2y.dds"},
	{0x9B0461F53B45A1A1ULL, "campaign_base_de-de.cre"},
	{0x9B0478C0AB7F8ACEULL, "hlac_w_encounter2_-131071_0z.dds"},
	{0x9B04C47B3E02C1FAULL, "hlac_lt_0_lowtown_65535_4y.dds"},
	{0x9B04CE7FC5521AB3ULL, "hlac_lt_backalley_g_65537_4y.dds"},
	{0x9B04FD85FFBF251BULL, "17655265.ogg"},
	{0x9B05289C1ADC9AA8ULL, "6042805_f.fxe"},
	{0x9B0528FEE1B504DEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_loop.gad"},
	{0x9B054FFBDE67FCF0ULL, "gen00fl_varric6093619.cl"},
	{0x9B0589D88AFCA94BULL, "vxdr_0_deeproads_9ga_0_0.vlm"},
	{0x9B058EED3AC95E4DULL, "343821080.ogg"},
	{0x9B0596ED9D013E84ULL, "isa211_isabela6072554.cl"},
	{0x9B05D94ECF4D9597ULL, "dre000_anders.cnv"},
	{0x9B05FB84A8AD5890ULL, "vxca_caverns2_814_0_1.vlm"},
	{0x9B06675544541BE0ULL, "llac_lt_0_lowtown_docks_a3_-65537_0.dds"},
	{0x9B068BBE91417F01ULL, "6026184_m.fxe"},
	{0x9B06F52ACDCC8E3AULL, "dae_mage_modal.ncs"},
	{0x9B07222637007EB7ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierplate.mmh"},
	{0x9B072626370085C4ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierplate.mao"},
	{0x9B0738263700A429ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazierplate.msh"},
	{0x9B07B85510896F14ULL, "6095600_m.fxe"},
	{0x9B085973211D3EAEULL, "vxht_bartrand_n_64a_0_0.vlm"},
	{0x9B08E1908947C312ULL, "6030621_m.fxe"},
	{0x9B08E78D94BE13BCULL, "vxht_0_hightown_avj_0_0.vlm"},
	{0x9B09B1CDAD4AAB4DULL, "abi_m_tempest.wbd"},
	{0x9B09E242EA8C8F58ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_enter.gad"},
	{0x9B09E7245A959B1FULL, "dae000cr_grand_cleric6124715.cl"},
	{0x9B09EF4B4EE70DECULL, "dae000_elegant6166901.cl"},
	{0x9B0A00FAD41C4B26ULL, "vxca_caverns1_i2l_0_0.vlm"},
	{0x9B0A2EE65A45BA8EULL, "mer320_hubert6053682.cl"},
	{0x9B0AB23401463208ULL, "seb221st_ruxton.stg"},
	{0x9B0AD0436D92EF14ULL, "896046403.ogg"},
	{0x9B0B019C8C3B4454ULL, "zz_one_debug6051441.cl"},
	{0x9B0B63372B33AEECULL, "zz_one_debug6079304.cl"},
	{0x9B0BD89734806FD9ULL, "pro000_refugee_female6052088.cl"},
	{0x9B0C22EE61260559ULL, "839948433.ogg"},
	{0x9B0C647C170109C6ULL, "hlac_dr_primevaltomb_659_0x.dds"},
	{0x9B0C6804DFD6DE26ULL, "627104569.ogg"},
	{0x9B0C74A20E1CA895ULL, "6140143_m.fxe"},
	{0x9B0CAAC7444146F1ULL, "vxlt_0_lowtown_gbq_0_0.vlm"},
	{0x9B0CED014FC0A7BEULL, "vxw_bonepitmining_a2_7a_0_0.vlm"},
	{0x9B0D5A4EA80BAFA3ULL, "isa311_party_comments6117813.cl"},
	{0x9B0D7E3BA0307332ULL, "follower_banter6141832.cl"},
	{0x9B0DEC099A1BB1ABULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_coma_0d.dds"},
	{0x9B0E0E5203240145ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_medp_0n.dds"},
	{0x9B0EB66F3568DE28ULL, "vxw_woundedcoast_1eg_0_0.vlm"},
	{0x9B0ED4C6559D9E83ULL, "vxlt_0_lowtown_docks_cs1_0_0.vlm"},
	{0x9B0F0E2771C7860CULL, "character_character_p87.psh.fxo.dx11"},
	{0x9B0F19BC5779C0AAULL, "247742044.ogg"},
	{0x9B0F3DEDAF0875C9ULL, "and111_lirene6084522.cl"},
	{0x9B0F74520C0F6A39ULL, "vxdr_primevaltomb_2w3_0_1.vlm"},
	{0x9B1002FA15C08A0BULL, "art\\vfx\\combat\\vfx_ha_hd_overwhelm_act_c.impa.ani"},
	{0x9B100722752E6C2CULL, "dae000_bartrand6107559.cl"},
	{0x9B111E4F2FAD4DC8ULL, "gen00fl_anders6113041.cl"},
	{0x9B1134E52A64365BULL, "and321tr_endstage_approval.utt"},
	{0x9B116E10F0D8E0B1ULL, "vxw_sundermount_2ie_0_0.vlm"},
	{0x9B119E230B42FDF3ULL, "6013285_m.fxe"},
	{0x9B11D1333826E5BAULL, "vxw_bonepit_9i_0_0.vlm"},
	{0x9B11E80BB4FE07DDULL, "hlac_dr_dungeon_3214_9x.dds"},
	{0x9B12904F23AA9E90ULL, "zz_followers_banter6159327.cl"},
	{0x9B13740791EFE58EULL, "283982697.ogg"},
	{0x9B1374E024F7D77AULL, "mcr370_negotiation6095510.cl"},
	{0x9B13A9B2CA4D8A80ULL, "hlac_lt_0_lowtown_a3_n_c_0_4y.dds"},
	{0x9B1405670DDEB79AULL, "20324359.ogg"},
	{0x9B14B563F7FBF1D9ULL, "130244803.ogg"},
	{0x9B14C32FA8442037ULL, "lgt311_shw_mer_agatha6156495.cl"},
	{0x9B1590BF491F0EF9ULL, "follower_banter6155487.cl"},
	{0x9B160C8B76136623ULL, "llac_ht_0_hightown_a2_-65534_0.dds"},
	{0x9B161E6A6879992DULL, "vxlt_hangedman_24x_0_0.vlm"},
	{0x9B162437BBFE937CULL, "vxdr_primevaltomb_6pv_0_0.vlm"},
	{0x9B16A8FFDAB5D851ULL, "vxw_bonepitdestroyed_a3_p1_0_0.vlm"},
	{0x9B1742E821BE1CC4ULL, "140734771.ogg"},
	{0x9B176D83458B4F2AULL, "bl_pol_l_si.mao"},
	{0x9B177E804F6BD843ULL, "6066609_m.fxe"},
	{0x9B1787D6D3F62F93ULL, "hlac_lt_0_lowtown_a3_n_c_65537_6z.dds"},
	{0x9B17B15C6EB7690CULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masa.msh"},
	{0x9B17B75C6EB773FEULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masa.mmh"},
	{0x9B17C35C6EB78795ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masa.mao"},
	{0x9B18B6A8507DF362ULL, "633319078.ogg"},
	{0x9B18B6B74E5FBB98ULL, "zz_lgt_1_debug6148335.cl"},
	{0x9B18C15F2A2ED02DULL, "cod_ltr_rdr_woodrows.utp"},
	{0x9B18E5519F633D61ULL, "6136704_m.fxe"},
	{0x9B18F7951C2501E5ULL, "dae000_porfiria6155292.cl"},
	{0x9B1925B72D8737FFULL, "850314879.ogg"},
	{0x9B1999B47CC772A2ULL, "vxw_bonepitdestroyed_a3_27c_0_0.vlm"},
	{0x9B19AB5288247DFCULL, "6129293_m.fxe"},
	{0x9B19D091D0129B58ULL, "hlac_lt_gallowstemplar_2388_1x.dds"},
	{0x9B19E02550C507BAULL, "zz_dae_debug6049790.cl"},
	{0x9B1A7A5F4392262AULL, "llac_ht_mansion2_818_24.dds"},
	{0x9B1A8B0CFD97FD27ULL, "lgt320_shw_ors_secret_meet.bnk"},
	{0x9B1B352DE008283EULL, "6079679_m.fxe"},
	{0x9B1BDA12F720EA99ULL, "gen00fl_anders6116827.cl"},
	{0x9B1C0ED732EA07DDULL, "vxw_sundermount_376_0_0.vlm"},
	{0x9B1CFC5ED1B9AE12ULL, "479030698.ogg"},
	{0x9B1D55C6CAB4152AULL, "sse_ss_player_female.gda"},
	{0x9B1DB54CC07AFCBCULL, "qcr000_arishok6060138.cl"},
	{0x9B1DEAFA15CC9E82ULL, "art\\vfx\\combat\\vfx_ha_hd_overwhelm_act_c.impa.evt"},
	{0x9B1E04023E0089E3ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.dura.evt"},
	{0x9B1E7E0B11E66D0FULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrda_0s.dds"},
	{0x9B1E82208320D11BULL, "577977131.ogg"},
	{0x9B1EAD0CFD9C1F9FULL, "lgt320_shw_ors_secret_meet.cnv"},
	{0x9B1EC470CA25928AULL, "vxca_caverns2_718_0_0.vlm"},
	{0x9B1F16082E6FC20EULL, "art\\characters\\playercharacter\\humanmale\\rob_appa.mmh"},
	{0x9B1F20AD20B954E7ULL, "llac_lt_gallowsprison_n_-1_0.dds"},
	{0x9B1F22082E6FD6A5ULL, "art\\characters\\playercharacter\\humanmale\\rob_appa.mao"},
	{0x9B1F235B0A0A2F5BULL, "261620774.ogg"},
	{0x9B1F30082E6FEEFCULL, "art\\characters\\playercharacter\\humanmale\\rob_appa.msh"},
	{0x9B1F725F732B9A41ULL, "gen_im_arm_hlm_med_act2_01.uti"},
	{0x9B1F9BFC4B3A6EA2ULL, "vxlt_0_lowtown_a2_f4m_0_0.vlm"},
	{0x9B1FDBB6DE3ED394ULL, "6135362_m.fxe"},
	{0x9B1FDFEDE2CEF931ULL, "vxw_sundermount_2rd_0_0.vlm"},
	{0x9B2039840A706700ULL, "lgt260_eau_tmp_roderick6165096.cl"},
	{0x9B2065E000ED0BD1ULL, "219955714.ogg"},
	{0x9B208BE3BDB77104ULL, "vxca_caverns1_dqb_0_0.vlm"},
	{0x9B20C385ECA65664ULL, "rdr110dg_guard_warehouse6016716.cl"},
	{0x9B20EDCEEDEFAED7ULL, "gen00fl_varric6126323.cl"},
	{0x9B2197234B2F3D50ULL, "vxlt_0_lowtown_a2_n_c_bye_0_0.vlm"},
	{0x9B227811DCDFC0A8ULL, "233603094.ogg"},
	{0x9B228F549B0D83B6ULL, "hlac_lt_0_lowtown_65536_8x.dds"},
	{0x9B229EC4ED834E5BULL, "hlac_lt_0_lowtown_-65536_4z.dds"},
	{0x9B230CBC62BB23FAULL, "6044653_m.fxe"},
	{0x9B2490475A6FB1ADULL, "vxlt_hangedman_21w_0_0.vlm"},
	{0x9B249614ADF4ABFCULL, "vxdr_primevaltomb_6ww_0_0.vlm"},
	{0x9B24C830C35E84B5ULL, "dae010_wandering_priest_f.utc"},
	{0x9B24DD023E066664ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.dura.gad"},
	{0x9B24E9FA15D2BB95ULL, "art\\vfx\\combat\\vfx_ha_hd_overwhelm_act_c.impa.gad"},
	{0x9B24FCD88F02AA08ULL, "6135344_m.fxe"},
	{0x9B25A34061D9BCA8ULL, "189811356.ogg"},
	{0x9B2668BBF8E62BA7ULL, "vxlt_0_lowtown_a2_hi5_0_0.vlm"},
	{0x9B26B3E8BD7B9274ULL, "vxw_woundedcoast_13a_0_0.vlm"},
	{0x9B271834C7883A1BULL, "6179912_m.fxe"},
	{0x9B272D8B7C2DD4E1ULL, "vxht_mansion_n_6ij_0_0.vlm"},
	{0x9B2752BA793FB3D5ULL, "171733353.ogg"},
	{0x9B278997475EEC91ULL, "461517881.ogg"},
	{0x9B27A49E3879291FULL, "hlac_lt_0_lowtown_n_-65535_0x.dds"},
	{0x9B280D90A5CFA8F0ULL, "801163012.ogg"},
	{0x9B286E2F223266DCULL, "hlac_lt_0_lowtown_a3_n_c_131073_0x.dds"},
	{0x9B2888C3A15A8AE0ULL, "hlac_ht_keep_3732_12y.dds"},
	{0x9B28D82F59A82A3CULL, "dae011_drunk2_sta6144221.cl"},
	{0x9B295E2D77A713EFULL, "6070130_m.fxe"},
	{0x9B2A0A82181FF29CULL, "vxw_bonepit_l4_0_0.vlm"},
	{0x9B2A3E2315C938DFULL, "follower_banter6151254.cl"},
	{0x9B2A80B424E0205DULL, "vxw_sundermount_325_0_0.vlm"},
	{0x9B2AA7E79A1C4635ULL, "gencr_trap_explosives_3.utc"},
	{0x9B2AD0C08D6B8DFCULL, "167903846.ogg"},
	{0x9B2AE440D27C34A2ULL, "437064.ogg"},
	{0x9B2AF29F999ABC6DULL, "6185100_m.fxe"},
	{0x9B2B75B04EAACBDFULL, "gen_im_trs_gen_hvy_glv_r1.uti"},
	{0x9B2B9A2919550F28ULL, "vxw_woundedcoast_1ke_0_0.vlm"},
	{0x9B2BEC023E0C9E42ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.dura.ani"},
	{0x9B2C1F1D5B80FD47ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv1.gad"},
	{0x9B2C5C2F7A2B4C3FULL, "follower_banter6160683.cl"},
	{0x9B2E13BD36F6F1E5ULL, "vxlt_0_lowtown_a2_n_c_gmq_0_0.vlm"},
	{0x9B2E3DA383BFD361ULL, "6079346_m.fxe"},
	{0x9B2E4F46A7B82ACEULL, "hlac_lt_0_lowtown_a2_n_c_131070_6x.dds"},
	{0x9B2EA7FDF2B93A2DULL, "vxca_caverns1_mines_fm4_0_0.vlm"},
	{0x9B2EDECAFF0A266EULL, "6138774_m.fxe"},
	{0x9B2F5B702B41795AULL, "gen00fl_dog6167650.cl"},
	{0x9B2FDFE22F1DE49EULL, "6149266_m.fxe"},
	{0x9B301E33D17078C6ULL, "lla_ht_0_hightown_a2_131072_0.dds"},
	{0x9B3046AA225A5DB4ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.dura.ani"},
	{0x9B30809A00AEFA6FULL, "6136107_f.fxe"},
	{0x9B30B6413D8280C8ULL, "vxht_mansion_n_6io_0_0.vlm"},
	{0x9B30CC446E99A624ULL, "hlac_w_encounter3_-65533_0x.dds"},
	{0x9B3146444A8E4B61ULL, "6101162_m.fxe"},
	{0x9B3205156F574C45ULL, "llac_lt_0_lowtown_a2_n_-131072_2.dds"},
	{0x9B328F1C1EFB151CULL, "vxca_caverns2prc_9qy_0_0.vlm"},
	{0x9B32B587D1C17F84ULL, "6157199_m.fxe"},
	{0x9B33141D5B870954ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv1.evt"},
	{0x9B33974A112921CBULL, "mag100_cullen6171202.cl"},
	{0x9B33BF1816349632ULL, "544154694.ogg"},
	{0x9B33F220D82933ADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud4high.msh"},
	{0x9B340020D8294BE0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud4high.mao"},
	{0x9B3402EF6AF87FF3ULL, "gen00fl_fenris6126781.cl"},
	{0x9B340420D8295233ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ww_extrud4high.mmh"},
	{0x9B341FECFF928692ULL, "lgt110_gxa_coa_meeting6136217.cl"},
	{0x9B343E2B73FB6746ULL, "hlac_dr_0_deeproads_5843_2z.dds"},
	{0x9B3441D75DA191AFULL, "hlac_lt_0_lowtown_a2_-1_8z.dds"},
	{0x9B346C33ED518756ULL, "vxlt_0_lowtown_n_f1a_0_0.vlm"},
	{0x9B34D834A83D06D4ULL, "dae000_dalish_apprentices6147740.cl"},
	{0x9B34E36F5B6A0981ULL, "162846947.ogg"},
	{0x9B3515CBACB47993ULL, "240031980.ogg"},
	{0x9B359F686E23ED61ULL, "vxht_mansion_n_6li_0_0.vlm"},
	{0x9B365A65F5FCFAEAULL, "6018704_m.fxe"},
	{0x9B36DE05DEE2A491ULL, "6125326_m.fxe"},
	{0x9B3735D209EAD42AULL, "542912001.ogg"},
	{0x9B3743AA22607716ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.dura.gad"},
	{0x9B3879D98470CB15ULL, "6102905_m.fxe"},
	{0x9B395A40294FE55FULL, "vxlt_warehouse_n_13s_0_0.vlm"},
	{0x9B399CB08817827FULL, "llac_lt_gallowscourtyard_65536_0.dds"},
	{0x9B39C09F52A3CFA3ULL, "mag320_bsc_banter6100770.cl"},
	{0x9B39CD4EBFF1E5EAULL, "hlac_lt_darktown_-1_64x.dds"},
	{0x9B3A8501B7AAF5BBULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_bld_flma_0s.dds"},
	{0x9B3AC0CAD2693D14ULL, "732788405.ogg"},
	{0x9B3B01849375E443ULL, "6165737_m.fxe"},
	{0x9B3B192E990F271FULL, "gen00fl_varric6093672.cl"},
	{0x9B3B3A5B496F5858ULL, "hlac_ca_caverns1_mines_24425_12z.dds"},
	{0x9B3B700DC0050F6DULL, "vxht_chantry_n_26k_0_0.vlm"},
	{0x9B3C0D69E6C52FD5ULL, "vxlt_darktown_5oo_0_0.vlm"},
	{0x9B3CB319E5583BACULL, "vxlt_gallowstemplar_f_3jx_0_0.vlm"},
	{0x9B3CC3A7C6BB2A31ULL, "vxw_sundermount_2xz_0_0.vlm"},
	{0x9B3CEDBB2C9DD4EBULL, "mag121_idunna6084189.cl"},
	{0x9B3D32FCFE962A3AULL, "918211242.ogg"},
	{0x9B3DA2D4BA1561ADULL, "34029564.ogg"},
	{0x9B3DC67A19259072ULL, "vxlt_darktown_4oe_0_0.vlm"},
	{0x9B3DF1B1197486FEULL, "6140593_m.fxe"},
	{0x9B3E3EE07EC86873ULL, "6135731_f.fxe"},
	{0x9B3E51BF2113FD1EULL, "6142366_m.fxe"},
	{0x9B3E58FD172738BAULL, "zz_qun_debug6028887.cl"},
	{0x9B3E5EAA2266C475ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_lp_c.dura.evt"},
	{0x9B3E80CFDF1CC512ULL, "vxlt_gallowscourtyard_n_76t_0_0.vlm"},
	{0x9B3F10160C5265C2ULL, "w_bonepitmining_a2_6_0.msh"},
	{0x9B3F12160C526978ULL, "w_bonepitmining_a2_6_0.mmh"},
	{0x9B3F1E160C527D1BULL, "w_bonepitmining_a2_6_0.mao"},
	{0x9B3F281E2F789948ULL, "vxht_mansion_n_6ll_0_0.vlm"},
	{0x9B3F7FA01F10E635ULL, "hlac_w_encounter1_-131068_0z.dds"},
	{0x9B400C1D5B91858DULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv1.ani"},
	{0x9B40248E86815AF2ULL, "ave200_donnic6111070.cl"},
	{0x9B410A112D9313B1ULL, "vxht_chantry_1yr_0_0.vlm"},
	{0x9B41129F3C195120ULL, "6170332_m.fxe"},
	{0x9B4197621F07BA11ULL, "dae000cr_grand_cleric6185495.cl"},
	{0x9B4199C67D0D919DULL, "z_cloud_v.dds"},
	{0x9B41A4F76F6A1841ULL, "mag170_hugh6080717.cl"},
	{0x9B41E1DB3BF10608ULL, "648187336.ogg"},
	{0x9B425EB3D7D05089ULL, "t3_fat_f00.tnt"},
	{0x9B42AF015C9675BCULL, "vxht_0_hightown_ajn_0_0.vlm"},
	{0x9B4312160C56518BULL, "w_bonepitmining_a2_6_0.rml"},
	{0x9B4323E5F775686CULL, "darktown_male_fer_urch_sta6174947.cl"},
	{0x9B437E232C5409CCULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargecelti.msh"},
	{0x9B4384232C5413BEULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargecelti.mmh"},
	{0x9B4390232C542855ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargecelti.mao"},
	{0x9B4397A2A167C374ULL, "vxw_woundedcoast_19g_0_0.vlm"},
	{0x9B43A389A74ABE6DULL, "6050130_m.fxe"},
	{0x9B43AC3660372ACDULL, "hlac_lt_0_lowtown_docks_n_-131073_4x.dds"},
	{0x9B43AF26A121076FULL, "1049681187.ogg"},
	{0x9B43B887B0C5B281ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipetail.ani"},
	{0x9B43E753FCA7F86EULL, "follower_barks6153348.cl"},
	{0x9B44603D17945D18ULL, "static_static_p56.psh.fxo.dx11"},
	{0x9B44761A0B848111ULL, "hlac_dr_slavers_3214_11y.dds"},
	{0x9B4518C62EAD7C89ULL, "vxca_caverns2prc_8ok_0_0.vlm"},
	{0x9B45B26CFEFD6883ULL, "globals.gda"},
	{0x9B45BB401A844EFBULL, "m_qun221ar_chantry_nt.wbd"},
	{0x9B46128C9ADEDBEDULL, "377060851.ogg"},
	{0x9B4659C9C2A07F63ULL, "6065654_f.fxe"},
	{0x9B466757E5A4EFF9ULL, "follower_barks6153442.cl"},
	{0x9B46D5642AC2D038ULL, "llac_lt_darktown_-1_4.dds"},
	{0x9B46E6189CFE252DULL, "follower_banter6172798.cl"},
	{0x9B46E7BDDE4A2909ULL, "gut_spla.mao"},
	{0x9B4700723C0A2E34ULL, "6121572_m.fxe"},
	{0x9B471D7DB1BBCDA2ULL, "6081092_m.fxe"},
	{0x9B4738C556F618AEULL, "348607639.ogg"},
	{0x9B47CC1D1B45FDDFULL, "vxlt_warehouse_n_18r_0_0.vlm"},
	{0x9B47D95B185149CFULL, "510986620.ogg"},
	{0x9B48187E5DABF162ULL, "plt_lgt000pt_ultra_1_06.plo"},
	{0x9B4879108DE6000CULL, "w_sundermount_181_0.phy"},
	{0x9B487DE2FD414028ULL, "vxw_woundedcoast_1qc_0_0.vlm"},
	{0x9B490F46E35C1065ULL, "zz_prmpt_debug6200959.cl"},
	{0x9B4980FF6882E84EULL, "vxw_encounter2_p3_0_1.vlm"},
	{0x9B49C387B0CA3123ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipetail.gad"},
	{0x9B49F00611123EC7ULL, "322735261.ogg"},
	{0x9B4A3FB5C3853C7AULL, "vxlt_gallowstemplar_3r9_0_0.vlm"},
	{0x9B4A41AA8D3AF1A7ULL, "lt_gallowstemplar_f_2390.rml"},
	{0x9B4A9A2504F57228ULL, "dae000_gamlen6026177.cl"},
	{0x9B4ABBCDAD825377ULL, "abi_m_tempest.bnk"},
	{0x9B4ACECA57C6683CULL, "220642457.ogg"},
	{0x9B4AF52834291035ULL, "6050189_m.fxe"},
	{0x9B4B24F6D74E542CULL, "vxlt_gallowstemplar_f_3qy_0_0.vlm"},
	{0x9B4B5E4379FCCDDDULL, "hlac_ca_cavern1_haunted_24423_0x.dds"},
	{0x9B4CD2FC0170C76DULL, "w_bonepitdestroyed_a3_83_0.mmh"},
	{0x9B4CD6FC0170CDBEULL, "w_bonepitdestroyed_a3_83_0.mao"},
	{0x9B4CE4FC0170E5F3ULL, "w_bonepitdestroyed_a3_83_0.msh"},
	{0x9B4CF2ACD112DD92ULL, "vxlt_gallowscourtyard_n_79u_0_0.vlm"},
	{0x9B4D0FCDE9F6A024ULL, "hlac_lt_0_lowtown_a2_131071_4z.dds"},
	{0x9B4D2EB425ED5BFCULL, "fireball01.mao"},
	{0x9B4D4EC41E7BB667ULL, "and111cr_refugee_boy.utc"},
	{0x9B4DA7F3BC08E6D9ULL, "llac_lt_0_lowtown_a3_n_c_-65535_5.dds"},
	{0x9B4DE8C784EDBCCBULL, "lgt320cr_shw_ors_marlein.utc"},
	{0x9B4E65887A2C30DDULL, "216663331.ogg"},
	{0x9B4EAAA5ED19CE2DULL, "gen00fl_aveline6129062.cl"},
	{0x9B4EEA042AF75E8CULL, "dae000_arishok6072850.cl"},
	{0x9B4F0416870EB399ULL, "460209635.ogg"},
	{0x9B4F90BBF98B0793ULL, "vxht_bartrand_n_4h6_0_0.vlm"},
	{0x9B4FDD45552135DBULL, "6133684_m.fxe"},
	{0x9B4FEB04B5B1E69BULL, "qun210cr_guard6083839.cl"},
	{0x9B505AC3985900F9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkback.ani"},
	{0x9B507073A28E3A51ULL, "vxw_bonepitdestroyed_a3_l5_0_0.vlm"},
	{0x9B5076AD9B9BAEAFULL, "356560714.ogg"},
	{0x9B508A042E1D24C3ULL, "744932293.ogg"},
	{0x9B50A087B0D01408ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipetail.evt"},
	{0x9B5120DE4E8C8E3CULL, "vxht_0_hightown_amo_0_0.vlm"},
	{0x9B51B9F9E79FEA3DULL, "mag291cr_young_feynriel.utc"},
	{0x9B51DD6D18FB5DF8ULL, "w_blightlands_91_0.phy"},
	{0x9B52686BD4E57A86ULL, "hlac_ht_keep_3732_15z.dds"},
	{0x9B527659C23BA00AULL, "mcr310_confused_guard6077040.cl"},
	{0x9B529CD64B1B14BAULL, "vxlt_0_lowtown_glh_0_0.vlm"},
	{0x9B52E9CA2F21A464ULL, "8498922.ogg"},
	{0x9B53454EA8998630ULL, "hlac_lt_shop_-1_4z.dds"},
	{0x9B536F151625611FULL, "6151099_m.fxe"},
	{0x9B5393C46E88CBCDULL, "6171576_m.fxe"},
	{0x9B54202539D35D27ULL, "289853893.ogg"},
	{0x9B5469AE0A5C0CE9ULL, "6110981_f.fxe"},
	{0x9B54771FD966498FULL, "lt_0_lowtown_docks_154_0.phy"},
	{0x9B54B99B8FE85C62ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_lgtd_t.dds"},
	{0x9B5540BD5FBD19A0ULL, "vxlt_0_lowtown_a2_n_c_i0m_0_0.vlm"},
	{0x9B555B675572E8A7ULL, "one300cr_ggt_thug4.utc"},
	{0x9B55CC7A3AF1B374ULL, "w_bonepitmining_a2_14_0.rml"},
	{0x9B55CE7A380CD81DULL, "fen210_hadriana6110390.cl"},
	{0x9B55F7E71E3FF218ULL, "da2terrain_terrain_p0.vsh.fxo"},
	{0x9B56352877A4DBDCULL, "vxlt_0_lowtown_a2_eop_0_1.vlm"},
	{0x9B564609AC8C546CULL, "6064091_m.fxe"},
	{0x9B565AB4EF12B818ULL, "6081256_m.fxe"},
	{0x9B56B2637C4DC0FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v4.evt"},
	{0x9B573145AFD85807ULL, "vxdr_0_deeproads_gb4_0_0.vlm"},
	{0x9B5735C3985EE03BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkback.gad"},
	{0x9B575FED79DFC563ULL, "mer340cr_hubert_combat.utc"},
	{0x9B582BBB05FE9800ULL, "63680265.ogg"},
	{0x9B58539FE40D9AC5ULL, "vxlt_gallowstemplar_3x4_0_0.vlm"},
	{0x9B588E5C6EEEC499ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masa.phy"},
	{0x9B598A7A1D953A7CULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medb_s.dds"},
	{0x9B59C1961B0506CFULL, "vista3.psh.fxo"},
	{0x9B59E8759E3C4860ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair2_n.dds"},
	{0x9B5A8855664FE48DULL, "vxlt_undercityserialkiller_1tt_0_0.vlm"},
	{0x9B5A99DCA949D8FBULL, "836301033.ogg"},
	{0x9B5B00F623ED0594ULL, "gen_im_fla_aveline_02.uti"},
	{0x9B5B17B41DD1BCB9ULL, "555918497.ogg"},
	{0x9B5B19EF35700CE7ULL, "pro000_merchant6051577.cl"},
	{0x9B5BE0C51A39C4AAULL, "6126508_m.fxe"},
	{0x9B5BE3EFF859A7DFULL, "hightown_market_guard_sta6174751.cl"},
	{0x9B5C15C929F46D33ULL, "llac_lt_undercitydungeon_-1_33.dds"},
	{0x9B5C62287046B044ULL, "mcr370_meredith6083020.cl"},
	{0x9B5CAE72AF7F581FULL, "hlac_w_encounter3_-196606_0y.dds"},
	{0x9B5CE1B34DEAAB4FULL, "hf_mother_h3s3.mor"},
	{0x9B5D2F58D1F9CCDDULL, "6140657_m.fxe"},
	{0x9B5DC9637C540629ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v4.gad"},
	{0x9B5DEAA9BADA14C9ULL, "llac_lt_backalley_g_0_16.dds"},
	{0x9B5E00E9F29BD95EULL, "nat380_nathaniel6066715.cl"},
	{0x9B5E52C398653020ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkback.evt"},
	{0x9B5E85872D80ABB7ULL, "zz_followers_banter6158145.cl"},
	{0x9B5EC1784E6A1BF1ULL, "196710760.ogg"},
	{0x9B5F89B08BD933ECULL, "vxlt_0_lowtown_a2_n_hgw_0_0.vlm"},
	{0x9B5FC86232726923ULL, "vfx_lightnig_dot2.dds"},
	{0x9B6010F6A6153DACULL, "944211416.ogg"},
	{0x9B601752B919FFA3ULL, "llac_w_blightlands_65535_0.dds"},
	{0x9B6029EFE2A674EEULL, "6125941_m.fxe"},
	{0x9B605E8F809ACB59ULL, "vxlt_0_lowtown_docks_bqt_0_0.vlm"},
	{0x9B60A3BCD4E4768CULL, "vxht_mansion_n_53j_0_0.vlm"},
	{0x9B624EB713D4932CULL, "640571184.ogg"},
	{0x9B630474CF3F8D0DULL, "gen00fl_dog6167163.cl"},
	{0x9B632DEC429F2527ULL, "vxht_0_hightown_a2_b93_0_0.vlm"},
	{0x9B638F546370C8EDULL, "hlac_lt_hangedman_n_-1_13y.dds"},
	{0x9B63976998709E24ULL, "733148154.ogg"},
	{0x9B63A61B9B3261D7ULL, "1014412428.ogg"},
	{0x9B649A637C59D577ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v4.ani"},
	{0x9B64A81DD00CB1E0ULL, "635925057.ogg"},
	{0x9B66523CEBFB6106ULL, "one280_emrys6104329.cl"},
	{0x9B667976E1D64440ULL, "6177399_m.fxe"},
	{0x9B66826F269500FFULL, "fen120ip_mansion_chest_1.utp"},
	{0x9B66C184054F4E5DULL, "follower_barks6153558.cl"},
	{0x9B672B86FDBB4F3BULL, "llac_ca_caverns2prc_5802_5.dds"},
	{0x9B6777F7CC01B018ULL, "hlac_ca_cavern1_haunted_24424_17z.dds"},
	{0x9B67C2E5A025DCACULL, "318412267.ogg"},
	{0x9B67EB37A15E7B94ULL, "6147217_m.fxe"},
	{0x9B681787239FE4B0ULL, "hm_lgt_apostate02.mor"},
	{0x9B681B6407296A4AULL, "vxlt_0_lowtown_a3_n_c_gjj_0_0.vlm"},
	{0x9B6832FE7489CAE7ULL, "gen_im_trs_bst_liver_01.uti"},
	{0x9B683EFC245071F1ULL, "vxht_chantry_n_31s_0_0.vlm"},
	{0x9B68A501B485EAE4ULL, "plt_lgt300pt_night.plo"},
	{0x9B68C935C12629D0ULL, "hlac_lt_blackemporium_1055_19x.dds"},
	{0x9B6A5B2C3B640AB9ULL, "175160812.ogg"},
	{0x9B6A8D7A5504CDC9ULL, "632125631.ogg"},
	{0x9B6B002FC7594299ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgred.msh"},
	{0x9B6B0A2FC7595387ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgred.mmh"},
	{0x9B6B0E2FC7595AD4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgred.mao"},
	{0x9B6B30AAAF8EB719ULL, "art\\vfx\\cinematic\\vfx_abom_trans_sw_p.impa.ani"},
	{0x9B6C703BACB98A78ULL, "zz_qcr_debug6106189.cl"},
	{0x9B6D273CE75D8AB1ULL, "6100320_f.fxe"},
	{0x9B6DB1801B9DA6B1ULL, "6127096_m.fxe"},
	{0x9B6DBED7640F61FCULL, "vxlt_0_lowtown_n_f04_0_0.vlm"},
	{0x9B6E86621B9791BDULL, "393705703.ogg"},
	{0x9B6EBCF47AED04CCULL, "6139387_f.fxe"},
	{0x9B6F087089A40A58ULL, "llac_lt_gallowscourtyard_n_-65537_0.dds"},
	{0x9B6F6CF4DFAF4CFCULL, "758963545.ogg"},
	{0x9B7050954391D8D6ULL, "6134886_m.fxe"},
	{0x9B7070602EBB2060ULL, "6050311_m.fxe"},
	{0x9B7188CD8A359AD9ULL, "vxlt_gallowscourtyard_7fp_0_0.vlm"},
	{0x9B71B797B4A04AE1ULL, "410633150.ogg"},
	{0x9B72074C856B99FAULL, "dae000_arishok.cnv"},
	{0x9B720BAAAF94975BULL, "art\\vfx\\cinematic\\vfx_abom_trans_sw_p.impa.gad"},
	{0x9B724273138071BFULL, "6049447_f.fxe"},
	{0x9B724D46D554170DULL, "lgt320pt_shw_ors_main.ncs"},
	{0x9B72596E236762B1ULL, "11019533.ogg"},
	{0x9B730F7959C16B09ULL, "557687461.ogg"},
	{0x9B732C6F8FC80284ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1ww_basic.msh"},
	{0x9B733E6F8FC821EDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1ww_basic.mao"},
	{0x9B73426F8FC82826ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1ww_basic.mmh"},
	{0x9B7382A842B09B92ULL, "18810239.ogg"},
	{0x9B739C1367251B5CULL, "6140506_m.fxe"},
	{0x9B73FC216EC4927CULL, "6142395_m.fxe"},
	{0x9B743348A192EF8FULL, "6127488_f.fxe"},
	{0x9B7485500467FE10ULL, "gen00fl_aveline6053272.cl"},
	{0x9B75294C856E0942ULL, "dae000_arishok.bnk"},
	{0x9B753759C7F9CBC5ULL, "vxlt_gallowstemplar_3r2_0_0.vlm"},
	{0x9B754D74B5FBAE0EULL, "90920442.ogg"},
	{0x9B7619D358C19FC2ULL, "6078463_m.fxe"},
	{0x9B765C83149FEF10ULL, "ser221_quentin6123389.cl"},
	{0x9B766E5BEFA51F81ULL, "zz_qun_debug6045388.cl"},
	{0x9B76B1F080551C7FULL, "6129231_f.fxe"},
	{0x9B76ECD96CAE6F7EULL, "6028941_m.fxe"},
	{0x9B7717A86EB33542ULL, "vxlt_0_lowtown_a2_n_c_fgi_0_0.vlm"},
	{0x9B7730349F149965ULL, "hlac_ca_caverns2_5802_71z.dds"},
	{0x9B77320E830262A2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\arm_bdaa_0s.dds"},
	{0x9B773C56134A312AULL, "vxlt_gallowstemplar_f_6qf_0_0.vlm"},
	{0x9B77F9D632078F8BULL, "vxlt_0_lowtown_a2_n_c_eqm_0_0.vlm"},
	{0x9B78EAA2BF6D5423ULL, "6078862_m.fxe"},
	{0x9B7928AAAF9AE740ULL, "art\\vfx\\cinematic\\vfx_abom_trans_sw_p.impa.evt"},
	{0x9B794F4E28238DCBULL, "hlac_w_bonepitdestroyed_a3_393222_0z.dds"},
	{0x9B7952E3F22149CDULL, "wm_ico_hightown_roll.dds"},
	{0x9B79A7110CDE7477ULL, "1039688772.ogg"},
	{0x9B79BD0984CF6841ULL, "vxlt_shop_5n_0_0.vlm"},
	{0x9B79F9204DDE62ECULL, "vxht_0_hightown_a3_bem_0_0.vlm"},
	{0x9B7A010F39C6593FULL, "hlac_lt_0_lowtown_n_-2_6x.dds"},
	{0x9B7AC86CD8C98F9CULL, "vxca_caverns2prc_9hr_0_0.vlm"},
	{0x9B7BC60197909920ULL, "em_junar.mor"},
	{0x9B7C29387A03378EULL, "vxw_bonepitmining_a2_1kr_0_0.vlm"},
	{0x9B7C4DF4473D75A8ULL, "zz_fen_debug6165664.cl"},
	{0x9B7C8719BC3B52F8ULL, "vxlt_gallowstemplar_f_2ed_0_0.vlm"},
	{0x9B7CCBC1E3F8DED8ULL, "llac_lt_gallowstemplar_f_2391_4.dds"},
	{0x9B7CD7FD30F08CC5ULL, "294506500.ogg"},
	{0x9B7D0AFF34B675DFULL, "llac_ht_0_hightown_a2_65535_0.dds"},
	{0x9B7D563A4E9B2370ULL, "art\\vfx\\cinematic\\vfx_shield_break_p.mmh"},
	{0x9B7D78B68707F126ULL, "vxht_keep_n_6j0_0_0.vlm"},
	{0x9B7D8776B8D0A78CULL, "vxht_mansion_n_59h_0_0.vlm"},
	{0x9B7D963344F3E19FULL, "6067660_m.fxe"},
	{0x9B7DA82F2AD69A9AULL, "lt_0_lowtown_docks_145_0.mmh"},
	{0x9B7DAC2F2AD6A1D1ULL, "lt_0_lowtown_docks_145_0.mao"},
	{0x9B7DBA2F2AD6B9E0ULL, "lt_0_lowtown_docks_145_0.msh"},
	{0x9B7E0127D8377E6AULL, "hlac_ht_mansion2_818_28z.dds"},
	{0x9B7E50D3076C63AFULL, "519777744.ogg"},
	{0x9B7E7EF55F760F24ULL, "hlac_ca_caverns2prc_5802_33x.dds"},
	{0x9B7F47EE9A68E726ULL, "follower_banter6169711.cl"},
	{0x9B7FB85F59C2455DULL, "llac_lt_0_lowtown_a3_n_c_-65534_0.dds"},
	{0x9B801397114BA2B8ULL, "vxw_encounter1_cx_0_0.vlm"},
	{0x9B802D5E32FD024BULL, "6146809_f.fxe"},
	{0x9B8040985E057506ULL, "vxw_encounter1_c6_0_0.vlm"},
	{0x9B80A0BC081856F2ULL, "6171041_m.fxe"},
	{0x9B80EE9FB196F2EEULL, "vxlt_0_lowtown_docks_a3_cyi_0_0.vlm"},
	{0x9B810B804077F903ULL, "291292998.ogg"},
	{0x9B811923EA3492BAULL, "w_bonepitdestroyed_a3_97_0.mmh"},
	{0x9B811D23EA3499F1ULL, "w_bonepitdestroyed_a3_97_0.mao"},
	{0x9B812B23EA34B100ULL, "w_bonepitdestroyed_a3_97_0.msh"},
	{0x9B81929B92F1F040ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bookcase.phy"},
	{0x9B81DDDF65873EDDULL, "t3_wdg_w13.tnt"},
	{0x9B82D7EB41E6FD1AULL, "character_character_p12.psh.fxo"},
	{0x9B832BDF6A2A3518ULL, "6108623_m.fxe"},
	{0x9B8402288945BA29ULL, "449836070.ogg"},
	{0x9B844E7C60870464ULL, "690927276.ogg"},
	{0x9B845B232C8B6559ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargecelti.phy"},
	{0x9B8476CB65D2D340ULL, "hlac_lt_0_lowtown_docks_a3_131071_6z.dds"},
	{0x9B84775864EFA7ACULL, "vxca_caverns2_93v_0_0.vlm"},
	{0x9B84C2A31C35E782ULL, "632931189.ogg"},
	{0x9B84E9EE3A21A3E5ULL, "317547408.ogg"},
	{0x9B84FF1DEB159B4AULL, "vxlt_0_lowtown_a3_n_c_gdd_0_0.vlm"},
	{0x9B85078866F29A47ULL, "hlac_ca_caverns2_5802_28z.dds"},
	{0x9B850E71C9FF2E6AULL, "468109365.ogg"},
	{0x9B850ED8D049E011ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2_0.pwk"},
	{0x9B8637A480DDAA63ULL, "prm000im_lion_shield.uti"},
	{0x9B863A609F3C016BULL, "hlac_lt_gallowscourtyard_65539_0y.dds"},
	{0x9B864EC65E10D282ULL, "983488541.ogg"},
	{0x9B869D5B3E88BD67ULL, "727390733.ogg"},
	{0x9B871DE3EE64840AULL, "vxht_0_hightown_a2_cca_0_0.vlm"},
	{0x9B872214D400E559ULL, "art\\characters\\weapons\\dagger\\w_dgr_104.mao"},
	{0x9B87BF7BD432ECFDULL, "vxlt_0_lowtown_a2_n_emm_0_0.vlm"},
	{0x9B88A82A38EE752DULL, "561034329.ogg"},
	{0x9B88ED5ECB1FE9DCULL, "one300_zevran6075778.cl"},
	{0x9B88F3E9542B3A27ULL, "vxlt_0_lowtown_docks_a3_n_czg_0_1.vlm"},
	{0x9B8934EDF19314BBULL, "dae000_arishok6095658.cl"},
	{0x9B8950108E1D51D7ULL, "w_sundermount_181_0.mmh"},
	{0x9B8954108E1D58E4ULL, "w_sundermount_181_0.mao"},
	{0x9B8966108E1D7749ULL, "w_sundermount_181_0.msh"},
	{0x9B898C04556679C6ULL, "392266613.ogg"},
	{0x9B89BD65E847ABB9ULL, "1010607062.ogg"},
	{0x9B89C30FB1293D17ULL, "572445433.ogg"},
	{0x9B89FADFB3AC98A2ULL, "hm_ser_quentin.mor"},
	{0x9B8A02E1330EA862ULL, "s_aoe_cr1.mao"},
	{0x9B8A2A7484551671ULL, "714316609.ogg"},
	{0x9B8A4154904D563FULL, "vxlt_0_lowtown_a2_n_hgx_0_0.vlm"},
	{0x9B8A72AAA6FEE08EULL, "1007528997.ogg"},
	{0x9B8AE6510A1E62BDULL, "6039463_m.fxe"},
	{0x9B8B2CA3AB5B0877ULL, "vxlt_0_lowtown_a3_n_c_f4t_0_0.vlm"},
	{0x9B8B3579BEC50BC7ULL, "262252699.ogg"},
	{0x9B8B4728568C1342ULL, "zz_one_debug6079324.cl"},
	{0x9B8B793A4EA79B8FULL, "art\\vfx\\cinematic\\vfx_shield_break_p.anb"},
	{0x9B8BA5BB03B4F614ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01b.phy"},
	{0x9B8C4033F12E4685ULL, "6113834_f.fxe"},
	{0x9B8C5D1AE958E91FULL, "vxlt_0_lowtown_a2_emp_0_0.vlm"},
	{0x9B8C8705C1D018F7ULL, "6065648_m.fxe"},
	{0x9B8CA07261C82BDAULL, "919075786.ogg"},
	{0x9B8D6CD936BC6455ULL, "6109364_m.fxe"},
	{0x9B8E85740341BB38ULL, "vxw_encounter1_hy_0_0.vlm"},
	{0x9B8EAF239C9EE27EULL, "335742370.ogg"},
	{0x9B8EB2754FFB8D86ULL, "vxw_encounter1_h7_0_0.vlm"},
	{0x9B8EBF15D80843C6ULL, "6208173_f.fxe"},
	{0x9B8EF06755CBF3AEULL, "854157738.ogg"},
	{0x9B8F8239F5E9B7CEULL, "lgt260cr_way_merc_lead2.utc"},
	{0x9B8FCC7A3B232B37ULL, "w_bonepitmining_a2_14_0.mmh"},
	{0x9B8FD07A3B233244ULL, "w_bonepitmining_a2_14_0.mao"},
	{0x9B8FE27A3B2350A9ULL, "w_bonepitmining_a2_14_0.msh"},
	{0x9B9013DD6C3FD070ULL, "ser270_templar6112571.cl"},
	{0x9B9046DFCC273CF7ULL, "6207739_f.fxe"},
	{0x9B9072CC73FBEA12ULL, "mag311_emile_de_launcet6043737.cl"},
	{0x9B90B50D02B553B2ULL, "6152640_m.fxe"},
	{0x9B90D26210BF7BA2ULL, "6106627_m.fxe"},
	{0x9B914210E9D0C5B8ULL, "612345683.ogg"},
	{0x9B91D7947A4A6625ULL, "vxca_cave_2i8_0_1.vlm"},
	{0x9B92E6286FAD5184ULL, "796907822.ogg"},
	{0x9B931740726F366EULL, "class_abilities_anders.gda"},
	{0x9B933C45A7094FECULL, "6085050_m.fxe"},
	{0x9B9398C2DE420F56ULL, "lt_0_lowtown_33_0.rml"},
	{0x9B93EBB1BFF971A3ULL, "182793117.ogg"},
	{0x9B943C4C5A01DC67ULL, "gen00fl_dog6146951.cl"},
	{0x9B94F01F9C0C2D42ULL, "93248271.ogg"},
	{0x9B9582BB03BD0F11ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01b.msh"},
	{0x9B959CBB03BD3BCFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth01b.mmh"},
	{0x9B9626CA53999FAFULL, "hlac_lt_0_lowtown_a2_n_c_-262143_0z.dds"},
	{0x9B967FF8C9CD743EULL, "6168418_m.fxe"},
	{0x9B96F32FC77E67BCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgred.phy"},
	{0x9B9788391FF8006AULL, "vxw_bonepit_27k_0_0.vlm"},
	{0x9B981B8FB65D18A4ULL, "6088800_m.fxe"},
	{0x9B987774F7292331ULL, "731101068.ogg"},
	{0x9B98B3EA12196166ULL, "711193255.ogg"},
	{0x9B98F6907D0A7391ULL, "vxlt_undercityserialkiller_18d_0_0.vlm"},
	{0x9B9A09C1102EC86FULL, "zz_dae_debug6029108.cl"},
	{0x9B9A169A5DEE967FULL, "dae000ip_crowd_lm2_talk10.utp"},
	{0x9B9A260348732D59ULL, "vxlt_0_lowtown_docks_bex_0_0.vlm"},
	{0x9B9AB2D3C7AE37C6ULL, "6034711_m.fxe"},
	{0x9B9AC5C5D15BD3A3ULL, "hm_icarver_h1s3.mor"},
	{0x9B9AEAB6F84EFC95ULL, "gen00fl_anders6125928.cl"},
	{0x9B9AF2DDE3A8EB15ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rr_basic.phy"},
	{0x9B9B2125C331B4D4ULL, "lgt351ar_rand_1.are"},
	{0x9B9BE898A5DB4069ULL, "vxlt_0_lowtown_a2_n_c_fid_0_0.vlm"},
	{0x9B9C35157B7122F3ULL, "706951115.ogg"},
	{0x9B9C619520DB7FC2ULL, "377281891.ogg"},
	{0x9B9C89144C6AB2D5ULL, "304618467.ogg"},
	{0x9B9CC6B7281FBD40ULL, "dae000ip_crowd_em2_talk3.utp"},
	{0x9B9CF3FD4546066CULL, "hlac_ht_mansion_n_3381_1x.dds"},
	{0x9B9CF979090D2266ULL, "art\\vfx\\creatures\\vfxroot.dura.ani"},
	{0x9B9D845CA77DA1FEULL, "vxca_caverns1_mines_e6m_0_0.vlm"},
	{0x9B9E37E528F1E976ULL, "vxlt_0_lowtown_a3_n_c_hpc_0_0.vlm"},
	{0x9B9ECA4783D64258ULL, "vxdr_primevaltomb_73v_0_0.vlm"},
	{0x9B9F566E39381EECULL, "art\\characters\\playercharacter\\dwarfmale\\dm_bdy_lgta.mao"},
	{0x9B9F7EED869DC149ULL, "vxw_bonepitmining_a2_26u_0_0.vlm"},
	{0x9B9F7FE17F18D123ULL, "6155640_m.fxe"},
	{0x9B9FBF2E3934273AULL, "vxht_chantry_n_36m_0_0.vlm"},
	{0x9BA0029083286706ULL, "297900007.ogg"},
	{0x9BA00F994E166618ULL, "6106423_m.fxe"},
	{0x9BA07654B9C1EE40ULL, "vxht_chantry_27r_0_0.vlm"},
	{0x9BA0DE30D69B20F9ULL, "6111286_m.fxe"},
	{0x9BA1409F3617D893ULL, "follower_banter6167040.cl"},
	{0x9BA1600D689E1660ULL, "gen00fl_dog6167087.cl"},
	{0x9BA185EE30472617ULL, "zz_vault_debug6177779.cl"},
	{0x9BA1A8C3D100944FULL, "vxlt_darktown_57u_0_0.vlm"},
	{0x9BA2317567C365DCULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench2_merillintro.gad"},
	{0x9BA28213A9E0F89BULL, "6166589_m.fxe"},
	{0x9BA2C444A22F2219ULL, "47897765.ogg"},
	{0x9BA3726D67EF25B5ULL, "vxw_encounter3_f_0_0.vlm"},
	{0x9BA373EFD936870EULL, "vxht_playershouse_n_1k1_0_0.vlm"},
	{0x9BA429361BE0529DULL, "428194096.ogg"},
	{0x9BA4B87CF99EA0F0ULL, "536780964.ogg"},
	{0x9BA4C23DACA459E0ULL, "389796220.ogg"},
	{0x9BA4CF951DF5428FULL, "zz_setplot.ncs"},
	{0x9BA4DA981F2A3C69ULL, "220906480.ogg"},
	{0x9BA536F939C4F4C8ULL, "41619317.ogg"},
	{0x9BA5A23A04E01842ULL, "6166333_m.fxe"},
	{0x9BA5AC97450F8EDAULL, "793178702.ogg"},
	{0x9BA5BA01B4B9BC2BULL, "plt_lgt300pt_night.nss"},
	{0x9BA5CFCB698DD8B9ULL, "llac_w_encounter2_camp_3_0.dds"},
	{0x9BA6A7A2D16B6979ULL, "qun231_varnell6078078.cl"},
	{0x9BA6AF9B93113DD5ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bookcase.msh"},
	{0x9BA6B19B9311414BULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bookcase.mmh"},
	{0x9BA6B794234B19BFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_basic.phy"},
	{0x9BA6BD9B931155E8ULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_bookcase.mao"},
	{0x9BA745D50CFF5741ULL, "vxca_cavern1_haunted_ihx_0_0.vlm"},
	{0x9BA77ECF4FE5810EULL, "vxw_bonepitmining_a2_1zu_0_0.vlm"},
	{0x9BA7EC070FAE8023ULL, "vxw_bonepitdestroyed_a3_1v2_0_0.vlm"},
	{0x9BA819949EDA8749ULL, "424613588.ogg"},
	{0x9BA82DCA2099D218ULL, "vxlt_gallowscourtyard_6eq_0_0.vlm"},
	{0x9BA8307567C7CFDAULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench2_merillintro.ani"},
	{0x9BA83A3F85529C32ULL, "tal_tau_03.mao"},
	{0x9BA8FB06BBAF8529ULL, "gen00fl_merrill6121232.cl"},
	{0x9BA9A8697301C78DULL, "gen00fl_fenris6113840.cl"},
	{0x9BA9CE1E4A6E24F7ULL, "6164349_m.fxe"},
	{0x9BAA5592C80A4B4DULL, "vxlt_0_lowtown_docks_n_b4h_0_0.vlm"},
	{0x9BAA67C31638826DULL, "hlac_lt_darktown_-1_2y.dds"},
	{0x9BAA71F58C8D13E7ULL, "vxlt_undercityserialkiller_1ln_0_0.vlm"},
	{0x9BAAE179091936D7ULL, "art\\vfx\\creatures\\vfxroot.dura.evt"},
	{0x9BAAEC0A00AD9EF7ULL, "lt_0_lowtown_docks_a3_174_0.msh"},
	{0x9BAAFE0A00ADBD5AULL, "lt_0_lowtown_docks_a3_174_0.mao"},
	{0x9BAB020A00ADC469ULL, "lt_0_lowtown_docks_a3_174_0.mmh"},
	{0x9BAB0852D73B3E1CULL, "262307042.ogg"},
	{0x9BAB28388D54EC57ULL, "vxlt_undercityencounter_1lr_0_0.vlm"},
	{0x9BAC22100CFEC8CCULL, "vxht_brothel_n_3sg_0_0.vlm"},
	{0x9BACC76761A214ECULL, "6051042_m.fxe"},
	{0x9BACCA6D1947E88DULL, "w_blightlands_91_0.msh"},
	{0x9BACD3D09208D512ULL, "ts_prog_tigers_eye.gda"},
	{0x9BACD86D194800C0ULL, "w_blightlands_91_0.mao"},
	{0x9BACDC6D19480713ULL, "w_blightlands_91_0.mmh"},
	{0x9BADCD47184B7CE7ULL, "vxdr_primevaltomb_34y_0_1.vlm"},
	{0x9BADD2F99F5C96E2ULL, "lgt110_gxa_coa_spy6117413.cl"},
	{0x9BADE81D30BF0B41ULL, "hlac_w_bonepitmining_a2_458755_0y.dds"},
	{0x9BAE117E5E029829ULL, "plt_lgt000pt_ultra_1_06.nss"},
	{0x9BAE720E8E22C237ULL, "gen00fl_anders6196829.cl"},
	{0x9BAEAD531E01BB9FULL, "hlac_ht_0_hightown_a3_1_0x.dds"},
	{0x9BAEE831ABB806C0ULL, "vxht_chantry_20s_0_0.vlm"},
	{0x9BAEF38F4DD184EEULL, "dae000_meredith6092596.cl"},
	{0x9BAFA5FEFF127066ULL, "6151326_m.fxe"},
	{0x9BAFB497FF526E34ULL, "221788761.ogg"},
	{0x9BAFCDED3E704346ULL, "6045016_m.fxe"},
	{0x9BB01AA0C2F6ACCFULL, "vxlt_darktown_58t_0_0.vlm"},
	{0x9BB08B2B3B5E3A31ULL, "489649580.ogg"},
	{0x9BB08D574CC86AAFULL, "smallblur.vsh.fxo"},
	{0x9BB0ACAC368414EEULL, "6062132_m.fxe"},
	{0x9BB0D24B51A5063AULL, "585002673.ogg"},
	{0x9BB10BFF49F1D698ULL, "796704177.ogg"},
	{0x9BB1147A8E02365CULL, "vxlt_hangedman_n_25u_0_0.vlm"},
	{0x9BB1225FCF1DA802ULL, "dre000_dougal6130992.cl"},
	{0x9BB15F2D6194AEECULL, "hlac_lt_0_lowtown_a3_n_c_131071_6x.dds"},
	{0x9BB1E279091F5760ULL, "art\\vfx\\creatures\\vfxroot.dura.gad"},
	{0x9BB1EFA10EAEE4A5ULL, "ter_cracked01_n.dds"},
	{0x9BB2A80D1EA70290ULL, "vxdr_primevalentrance_1ka_0_1.vlm"},
	{0x9BB33F8A91587F4EULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingc_d.dds"},
	{0x9BB36EF3B07ADEA8ULL, "6116985_m.fxe"},
	{0x9BB38C2646359DAFULL, "vxw_bonepitdestroyed_a3_9p_0_0.vlm"},
	{0x9BB392E15860803CULL, "zz_car_debug6122521.cl"},
	{0x9BB395A092577F2DULL, "vxca_caverns1_mines_eic_0_0.vlm"},
	{0x9BB42FE9EC807332ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_en.gad"},
	{0x9BB4425C4B17E876ULL, "130010810.ogg"},
	{0x9BB452382BBF8B0EULL, "combat_2h_nmv_default.gda"},
	{0x9BB45BBC0BED19E0ULL, "qun211_party_comment6072259.cl"},
	{0x9BB46AAED5869A0CULL, "256226360.ogg"},
	{0x9BB56E9B5F7F6159ULL, "vxw_bonepitdestroyed_a3_1ws_0_0.vlm"},
	{0x9BB5B667D461F163ULL, "one110_tree_minder6140863.cl"},
	{0x9BB5DA4A60F6A491ULL, "vxlt_undercityserialkiller_12z_0_0.vlm"},
	{0x9BB5F0AC41DB998EULL, "vxw_bonepitmining_a2_1wv_0_0.vlm"},
	{0x9BB6DC0E3EDCD9A4ULL, "586408443.ogg"},
	{0x9BB78415E01285A4ULL, "vxht_0_hightown_n_b7m_0_0.vlm"},
	{0x9BB79E9E033C5EA3ULL, "hlac_lt_gallowstemplar_2383_7x.dds"},
	{0x9BB7B89423598EC4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_basic.mmh"},
	{0x9BB7BC9423599517ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_basic.mao"},
	{0x9BB7CE942359B466ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1pp_basic.msh"},
	{0x9BB806CFFC79FBB7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int2_sa.dds"},
	{0x9BB857398F2D68E0ULL, "396850380.ogg"},
	{0x9BB8A0247D732122ULL, "hlac_lt_0_lowtown_docks_131070_2z.dds"},
	{0x9BB8E3D27E832C67ULL, "vxlt_undercityserialkiller_1ai_0_0.vlm"},
	{0x9BB8E69A9150BDD5ULL, "llac_w_encounter3_-262146_0.dds"},
	{0x9BB8F6C798B3E953ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_roundtable.phy"},
	{0x9BB8F85B4A5F014FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_isabella_flicker.gad"},
	{0x9BB94747073C3F15ULL, "vxht_chantry_25y_0_0.vlm"},
	{0x9BB9C6889C80015EULL, "vxlt_warehouse_r7_0_0.vlm"},
	{0x9BBA6837AC8DA7BAULL, "llac_lt_gallowstemplar_n_2393_0.dds"},
	{0x9BBB2BCC133FC18CULL, "vxw_encounter2_camp_rr_0_0.vlm"},
	{0x9BBB32E9EC869769ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_en.evt"},
	{0x9BBB7EE44EA04845ULL, "vxw_encounter2_mv_0_0.vlm"},
	{0x9BBBA666DD6838A7ULL, "6029367_m.fxe"},
	{0x9BBBB930A7BED9E9ULL, "vxlt_0_lowtown_a2_n_c_e22_0_0.vlm"},
	{0x9BBBEEC7AE4FA59BULL, "vxw_woundedcoast_176_0_0.vlm"},
	{0x9BBBF87A86021EF0ULL, "vxht_chantry_214_0_0.vlm"},
	{0x9BBC332336349478ULL, "236026272.ogg"},
	{0x9BBCAAC04879C349ULL, "vxlt_0_lowtown_a2_n_ff7_0_0.vlm"},
	{0x9BBDA8EF2FB8A00AULL, "vxca_caverns2_7x3_0_0.vlm"},
	{0x9BBDAA9B2214F747ULL, "6208349_f.fxe"},
	{0x9BBE326715F72DF0ULL, "lt_0_lowtown_docks_21_0.rml"},
	{0x9BBE76FFC74E6A00ULL, "186583885.ogg"},
	{0x9BBE7C4F3F7EB345ULL, "170784714.ogg"},
	{0x9BBEA7E7490A1B72ULL, "hlac_lt_0_lowtown_a2_n_c_-3_2z.dds"},
	{0x9BBED6726B20DC8BULL, "870765521.ogg"},
	{0x9BBEFD7CB3687EC8ULL, "6169198_m.fxe"},
	{0x9BBF70F5A6575E50ULL, "6124764_m.fxe"},
	{0x9BBF7117B44B02BFULL, "vxht_mansion2_2l4_0_0.vlm"},
	{0x9BBFE8B62C61407DULL, "qun120_arishok_guard6060608.cl"},
	{0x9BC00D5B4A6543FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_isabella_flicker.evt"},
	{0x9BC0A5A56EEFFC2CULL, "6191306_m.fxe"},
	{0x9BC0FFC2D5665933ULL, "192030211.ogg"},
	{0x9BC12025C352824FULL, "lgt351ar_rand_1.lst"},
	{0x9BC16DA08ECCB5B3ULL, "lgt150im_ntb_wr_ingrdnt_4.uti"},
	{0x9BC1E1013F867B0BULL, "clf260cr_guard2.utc"},
	{0x9BC2538A8D1E42FEULL, "341656193.ogg"},
	{0x9BC307D722CDCCC2ULL, "darktown_male_fer_urch_sta6174963.cl"},
	{0x9BC3162A9F0C0898ULL, "vxht_0_hightown_n_3ub_0_0.vlm"},
	{0x9BC3694FAE984FBAULL, "gen00fl_aveline6139787.cl"},
	{0x9BC3A9019BA1CFBBULL, "dae011_corff_sta6142451.cl"},
	{0x9BC3BFB4D91680FFULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_conj_c.mmh"},
	{0x9BC42975F6EF388CULL, "vxca_caverns2prc_8be_0_0.vlm"},
	{0x9BC44C2752ECBD11ULL, "vxlt_undercityserialkiller_17y_0_0.vlm"},
	{0x9BC52F124F03DF7BULL, "223180209.ogg"},
	{0x9BC543BF7D2FBD71ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basicnb.msh"},
	{0x9BC551BF7D2FD48CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basicnb.mao"},
	{0x9BC55DBF7D2FE92FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basicnb.mmh"},
	{0x9BC57A14697DD225ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.mmh"},
	{0x9BC5A229F61753CBULL, "one140cr_charade6139393.cl"},
	{0x9BC5D125ABAB9E36ULL, "art\\characters\\playercharacter\\humanmale\\arm_wmda.msh"},
	{0x9BC5DB25ABABAF34ULL, "art\\characters\\playercharacter\\humanmale\\arm_wmda.mmh"},
	{0x9BC5DF25ABABB647ULL, "art\\characters\\playercharacter\\humanmale\\arm_wmda.mao"},
	{0x9BC5E160846C41F3ULL, "ht_0_hightownexplosion_n_51_0_lc.mmh"},
	{0x9BC6BD1EB742D31BULL, "6096866_m.fxe"},
	{0x9BC6E49BB1C9E838ULL, "vxht_mansion_n_5cn_0_0.vlm"},
	{0x9BC75BD264E2F5C8ULL, "vxht_brothel_n_60c_0_0.vlm"},
	{0x9BC7632BCF9F5B99ULL, "765586844.ogg"},
	{0x9BC76E4044D189DCULL, "llac_dr_slavers_3215_3.dds"},
	{0x9BC81AE9EC90F8D8ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_en.ani"},
	{0x9BC8732737BF0E6FULL, "6005017_m.fxe"},
	{0x9BC874A9CC62D117ULL, "qcr000_arishok6039300.cl"},
	{0x9BC8955CE6730853ULL, "252750251.ogg"},
	{0x9BC8B82938FA5C7AULL, "82111524.ogg"},
	{0x9BC8B9BA66B83B6FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magc_0s.dds"},
	{0x9BC8EF8529241188ULL, "762238411.ogg"},
	{0x9BC900AE09061F0FULL, "hlac_ca_caverns1_24422_3y.dds"},
	{0x9BC93A7DFD1CDEE8ULL, "hlac_ht_mansion2b_818_6x.dds"},
	{0x9BC99DA90535DA0CULL, "vxw_encounter2_camp_ms_0_0.vlm"},
	{0x9BC9DA085A338516ULL, "928237359.ogg"},
	{0x9BC9F0C1409660C5ULL, "vxw_encounter2_hu_0_0.vlm"},
	{0x9BCA0CAFE3D05B0AULL, "vxlt_0_lowtown_docks_a3_cqu_0_0.vlm"},
	{0x9BCA473BB91512FEULL, "vxlt_hangedman_27v_0_0.vlm"},
	{0x9BCA5D322FDBCAE0ULL, "532114325.ogg"},
	{0x9BCB05273740D2C3ULL, "vxht_0_hightown_9wz_0_0.vlm"},
	{0x9BCB3FDF9908D6A3ULL, "6208345_f.fxe"},
	{0x9BCB858FA48B1834ULL, "6020433_m.fxe"},
	{0x9BCBA1D02C3BFBEBULL, "7936107.ogg"},
	{0x9BCBC9079D991DE6ULL, "hlac_w_bonepitdestroyed_a3_0_8z.dds"},
	{0x9BCBCC3C3840154FULL, "mag320_orsino6070441.cl"},
	{0x9BCC4B0DF430B81DULL, "w_bonepitdestroyed_a3_132_0.mao"},
	{0x9BCC4F0DF430BF56ULL, "w_bonepitdestroyed_a3_132_0.mmh"},
	{0x9BCC590DF430D054ULL, "w_bonepitdestroyed_a3_132_0.msh"},
	{0x9BCCB43146C95D12ULL, "6129098_f.fxe"},
	{0x9BCCF7A7D292DF9DULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs104a_0s.dds"},
	{0x9BCD155B4A6FDB15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_isabella_flicker.ani"},
	{0x9BCD423ADC5B2193ULL, "background_desc.gda"},
	{0x9BCD95C2DE73820AULL, "lt_0_lowtown_33_0.anb"},
	{0x9BCDA17AAF16183CULL, "dae100_bran6118320.cl"},
	{0x9BCDA720B97342E7ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\vst_cmountain_d.dds"},
	{0x9BCDF0BC1363AD5FULL, "hlac_lt_0_lowtown_docks_n_0_1y.dds"},
	{0x9BCED5BBC10F6E9FULL, "plt_cod_ltr_dissent_ella.nss"},
	{0x9BCEE9605E78E0FBULL, "llac_lt_undercityencounter_-1_47.dds"},
	{0x9BCF1BF612116857ULL, "vxlt_0_lowtown_a2_n_goj_0_1.vlm"},
	{0x9BCF79ACEBEB1BCFULL, "vrc221_varric6045035.cl"},
	{0x9BCFAC4D4B625528ULL, "81447987.ogg"},
	{0x9BD04796B5DAAE0BULL, "6174663_m.fxe"},
	{0x9BD0629F328E3FF8ULL, "260715195.ogg"},
	{0x9BD181BE54102985ULL, "927020469.ogg"},
	{0x9BD1B80E4B17C3BCULL, "follower_banter6173358.cl"},
	{0x9BD1D0A9FDBF2B09ULL, "vxht_brothel_n_4d0_0_0.vlm"},
	{0x9BD258EEDD62AFA2ULL, "w_woundedcoast_198_0.phy"},
	{0x9BD264C529777CB6ULL, "vxlt_gallowscourtyard_n_6ep_0_0.vlm"},
	{0x9BD271D60CBAE210ULL, "lgt110_gxa_coa_spy6117518.cl"},
	{0x9BD27714698857F2ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.anb"},
	{0x9BD318868865D9C2ULL, "hlac_lt_gallowstemplar_n_2394_11y.dds"},
	{0x9BD3419DE590C9A3ULL, "vxw_bonepitdestroyed_a3_1a7_0_0.vlm"},
	{0x9BD393E14438A3B9ULL, "llac_ht_0_hightown_65533_2.dds"},
	{0x9BD44753E582E9DDULL, "6140020_m.fxe"},
	{0x9BD557D5FFAE5788ULL, "6119283_m.fxe"},
	{0x9BD5D667AF9D3690ULL, "gen00fl_aveline6140136.cl"},
	{0x9BD5F572F4ACBB91ULL, "dae000_gamlen6102711.cl"},
	{0x9BD6559E3548D0DEULL, "6047238_m.fxe"},
	{0x9BD669D0E3208DF5ULL, "dae121st_viscount_2.stg"},
	{0x9BD6B19DA7BE28BFULL, "hlac_lt_undercityserialkiller_-1_47z.dds"},
	{0x9BD6D597D180897DULL, "vxht_mansion2b_3gr_0_0.vlm"},
	{0x9BD6E0B4749A0D5BULL, "vxlt_0_lowtown_f4t_0_0.vlm"},
	{0x9BD73BE08F6149F0ULL, "vxw_sundermount_37n_0_1.vlm"},
	{0x9BD7BCAB693DE948ULL, "6034713_m.fxe"},
	{0x9BD7FF58F0CA4B76ULL, "vxlt_0_lowtown_a3_n_c_hlg_0_0.vlm"},
	{0x9BD80F85F72BF28CULL, "vxw_encounter2_camp_hp_0_0.vlm"},
	{0x9BD84E9231C30245ULL, "958167511.ogg"},
	{0x9BD8FC7FA46F24E4ULL, "vxht_0_hightown_b8w_0_0.vlm"},
	{0x9BD90C2D76FF2211ULL, "llac_lt_0_lowtown_docks_a3_n_-131076_0.dds"},
	{0x9BD959C6EE281E3FULL, "6204166_m.fxe"},
	{0x9BD95C5304C22F33ULL, "vxlt_darktown_4zs_0_0.vlm"},
	{0x9BD95EA1A8383BCEULL, "hlac_lt_0_lowtown_131071_1y.dds"},
	{0x9BD961F0B5F8A6B9ULL, "w_bonepit_195_0.msh"},
	{0x9BD96BF0B5F8B7A7ULL, "w_bonepit_195_0.mmh"},
	{0x9BD96FF0B5F8BEF4ULL, "w_bonepit_195_0.mao"},
	{0x9BD974482477E583ULL, "zz_isa_debug6133206.cl"},
	{0x9BDA5D6BD47B5A87ULL, "vxlt_warehouse_n_v7_0_0.vlm"},
	{0x9BDAFF3D87ACE574ULL, "6208337_f.fxe"},
	{0x9BDB69143392FEE3ULL, "847526627.ogg"},
	{0x9BDB78A237433048ULL, "6167880_m.fxe"},
	{0x9BDBE48CF491ED98ULL, "6084721_m.fxe"},
	{0x9BDCEB7FD2D98D29ULL, "vxlt_backalley_n_1wl_0_0.vlm"},
	{0x9BDCF9780F22A327ULL, "vxdr_primevaltomb_6rk_0_0.vlm"},
	{0x9BDD01E6D8ACBFB2ULL, "834582547.ogg"},
	{0x9BDD5D630BB1B6E5ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_104a_0t.dds"},
	{0x9BDD8814BB66E9FFULL, "fen210_orana6133068.cl"},
	{0x9BDD9C1F1FE83828ULL, "6021301_f.fxe"},
	{0x9BDDE7BBCECCD6A8ULL, "dae000_bartrand6206922.cl"},
	{0x9BDE24C47AEF11EDULL, "352202195.ogg"},
	{0x9BDEDD95FACCE008ULL, "one100cr_elren.utc"},
	{0x9BDF58F4BEE4342EULL, "284176338.ogg"},
	{0x9BE01D037E913CDBULL, "6064702_m.fxe"},
	{0x9BE03290E45B2BD7ULL, "dae370ip_armor_store.utp"},
	{0x9BE0B1DFDE8FB48CULL, "6127010_f.fxe"},
	{0x9BE0C4323561AAD9ULL, "vxw_bonepitdestroyed_a3_1fp_0_0.vlm"},
	{0x9BE0DA3B3914E736ULL, "w_encounter1_59_0.phy"},
	{0x9BE0DF1D9332229CULL, "vxlt_0_lowtown_a3_n_c_ewi_0_0.vlm"},
	{0x9BE109901D6A0F38ULL, "vxlt_hangedman_1pn_0_0.vlm"},
	{0x9BE11F28EEE11250ULL, "vxlt_0_lowtown_docks_a3_n_cy4_0_0.vlm"},
	{0x9BE13BEC4ABE9A3BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_drink.ani"},
	{0x9BE141DCC0E914A2ULL, "vxdr_primevalentrance_27i_0_0.vlm"},
	{0x9BE142B650FDA58FULL, "6131574_m.fxe"},
	{0x9BE14C8E94E059EFULL, "68707605.ogg"},
	{0x9BE1B37AD786E223ULL, "vxw_bonepitdestroyed_a3_1j6_0_0.vlm"},
	{0x9BE205DDE3E50E30ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rr_basic.msh"},
	{0x9BE213DDE3E5264AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rr_basic.mmh"},
	{0x9BE217DDE3E52C81ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\7rr_basic.mao"},
	{0x9BE2855AE4D482DDULL, "hlac_lt_0_lowtown_docks_-1_3z.dds"},
	{0x9BE2AB56D241F918ULL, "qcr220_grey_wardens6060467.cl"},
	{0x9BE2C8BBC12001B0ULL, "plt_cod_ltr_dissent_ella.plo"},
	{0x9BE2E574B6D67A26ULL, "6132762_f.fxe"},
	{0x9BE31A28C661833EULL, "qun111_ketojan6039006.cl"},
	{0x9BE331FFCE639273ULL, "vxht_chantry_352_0_0.vlm"},
	{0x9BE3BFEDAA43AF0AULL, "gen00fl_anders6183815.cl"},
	{0x9BE3C85595B61938ULL, "vxht_mansion_n_5il_0_0.vlm"},
	{0x9BE4236F881B4A06ULL, "vxlt_hangedman_21h_0_0.vlm"},
	{0x9BE4449921C11A39ULL, "dae000ip_crowd_em1_talk10.utp"},
	{0x9BE47908257EC97AULL, "mag121_viveka6083382.cl"},
	{0x9BE4C44E368FFF42ULL, "vxca_caverns1_e8f_0_0.vlm"},
	{0x9BE5080677ED8D3AULL, "w_encounter3_97_0.phy"},
	{0x9BE5A0ED66C7FEF2ULL, "1029644194.ogg"},
	{0x9BE6806956B8DAD5ULL, "hlac_lt_0_lowtown_docks_a3_0_6x.dds"},
	{0x9BE68162E9220B0CULL, "vxw_encounter2_camp_cq_0_0.vlm"},
	{0x9BE68323A91FDEFAULL, "88498478.ogg"},
	{0x9BE6B436731E23F4ULL, "517591101.ogg"},
	{0x9BE6CE5E7F0F1F19ULL, "gen_im_arm_cht_elf_comb.uti"},
	{0x9BE764183C5AD923ULL, "6075278_m.fxe"},
	{0x9BE79DA47CBA467CULL, "zz_dre_debug6069201.cl"},
	{0x9BE7BF10A9C486C6ULL, "follower_banter6156119.cl"},
	{0x9BE87D19A0278CB2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_setprops_d.dds"},
	{0x9BE929EA8B7225ECULL, "573911806.ogg"},
	{0x9BE9369BE78E8E8AULL, "hlac_lt_0_lowtown_a2_n_c_65537_4z.dds"},
	{0x9BE953B5B7BEEA4EULL, "6000796_m.fxe"},
	{0x9BE967B64194C60DULL, "gen00fl_bethany6102767.cl"},
	{0x9BE99891C58226FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\cuddle_sitting_male_loop.ani"},
	{0x9BE9BE1E919C3889ULL, "606590168.ogg"},
	{0x9BE9E1AEC06F6E82ULL, "532761237.ogg"},
	{0x9BEA1D3408177E78ULL, "mag370_meredith6092197.cl"},
	{0x9BEA3367161C6A58ULL, "lt_0_lowtown_docks_21_0.anb"},
	{0x9BEA69B3D3DC1F6DULL, "hlac_dr_primevaltomb_661_2z.dds"},
	{0x9BEA98026BAE9016ULL, "hlac_lt_0_lowtown_2_1x.dds"},
	{0x9BEA9D1A6E3E477EULL, "vxlt_gallowsprison_n_5r4_0_0.vlm"},
	{0x9BEAB0CFD6FE50E6ULL, "hlac_lt_0_lowtown_docks_131073_1z.dds"},
	{0x9BEAC33B391D14D7ULL, "w_encounter1_59_0.msh"},
	{0x9BEAD53B391D333AULL, "w_encounter1_59_0.mao"},
	{0x9BEAD93B391D3A49ULL, "w_encounter1_59_0.mmh"},
	{0x9BEBAD409305018EULL, "vxht_playershouse_n_1r6_0_0.vlm"},
	{0x9BEC5C298649340DULL, "vrc221_hugin6041112.cl"},
	{0x9BECBFD0297B81DEULL, "gen00fl_aveline6132870.cl"},
	{0x9BECC0B4D939873CULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_conj_c.anb"},
	{0x9BED1DA0F4AD5A09ULL, "6092547_m.fxe"},
	{0x9BED495D162855B7ULL, "6071177_m.fxe"},
	{0x9BEDBF1D025DDFB4ULL, "hlac_lt_0_lowtown_docks_a3_-196610_5x.dds"},
	{0x9BEDCF9D70A876F7ULL, "vxlt_0_lowtown_h7o_0_0.vlm"},
	{0x9BEDD805D3A45332ULL, "665092549.ogg"},
	{0x9BEE43EC4AC93232ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_drink.evt"},
	{0x9BEEE5ABE5A2E709ULL, "926711537.ogg"},
	{0x9BEEEE56132DC0F7ULL, "llac_ht_0_hightown_65537_0.dds"},
	{0x9BEF25666CD58940ULL, "gen00fl_merrill6142104.cl"},
	{0x9BEF360F2757C359ULL, "vxw_bonepitdestroyed_a3_1kw_0_0.vlm"},
	{0x9BEFE83784F04704ULL, "63562673.ogg"},
	{0x9BF0A591C5885B3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\cuddle_sitting_male_loop.gad"},
	{0x9BF0CF3E060A60FBULL, "6153440_m.fxe"},
	{0x9BF1604D81132233ULL, "6152366_m.fxe"},
	{0x9BF1B855797C1017ULL, "one220_yevhen6099021.cl"},
	{0x9BF1C867A49D36FDULL, "229619259.ogg"},
	{0x9BF1DD5755432455ULL, "847061111.ogg"},
	{0x9BF21651FE0E67A0ULL, "vxlt_0_lowtown_docks_a3_bt5_0_0.vlm"},
	{0x9BF23A3287AC31B8ULL, "vxht_mansion_n_5rm_0_0.vlm"},
	{0x9BF27A51E4AFBF18ULL, "mer221_sabin6044052.cl"},
	{0x9BF293E2F57F2EC2ULL, "dre000_hireling16038651.cl"},
	{0x9BF2AB46465B8E67ULL, "vxlt_undercityserialkiller_1ue_0_0.vlm"},
	{0x9BF32FD2361C7777ULL, "qcr000_arishok6057106.cl"},
	{0x9BF3351E054E674CULL, "vxht_bartrand_n_6r9_0_0.vlm"},
	{0x9BF3728922B33A46ULL, "390666857.ogg"},
	{0x9BF390749344D277ULL, "vxca_cavern1_haunted_eia_0_0.vlm"},
	{0x9BF53006C2A35020ULL, "dae000_mother6156402.cl"},
	{0x9BF562EC4ACF85C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_bar_drink.gad"},
	{0x9BF598A3D980D57BULL, "hlac_w_encounter3_-65535_0y.dds"},
	{0x9BF5BFA7D95C8FEBULL, "lgt311_shw_mer_mettin6156452.cl"},
	{0x9BF6439F57C435A5ULL, "6071122_m.fxe"},
	{0x9BF68977139F13CCULL, "6177921_f.fxe"},
	{0x9BF6A93533EC3FB2ULL, "llac_lt_0_lowtown_docks_n_-4_6.dds"},
	{0x9BF6B6D88157B5F4ULL, "6136723_m.fxe"},
	{0x9BF6E84F96D1C6B4ULL, "vxdr_primevaltomb_35q_0_1.vlm"},
	{0x9BF6E975D4DDB59AULL, "llac_lt_darktown_-1_6.dds"},
	{0x9BF723EBFB2E1E94ULL, "6168143_f.fxe"},
	{0x9BF7A27AC8C42C46ULL, "352816483.ogg"},
	{0x9BF7C9D2074D66F3ULL, "440323737.ogg"},
	{0x9BF7E051C6D5F85CULL, "vxht_bartrand_n_4qv_0_0.vlm"},
	{0x9BF833EEDD834056ULL, "w_woundedcoast_198_0.mao"},
	{0x9BF83FEEDD8354F5ULL, "w_woundedcoast_198_0.mmh"},
	{0x9BF841EEDD83586BULL, "w_woundedcoast_198_0.msh"},
	{0x9BF8719C394D1754ULL, "vxht_mansion2b_2pw_0_0.vlm"},
	{0x9BF9E6D94D567679ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_2_sylvesta.gad"},
	{0x9BFA4017EED8B7DFULL, "mag291_demon_marethari6027946.cl"},
	{0x9BFA5D3F2A98E9D1ULL, "mrl000_marethari6142456.cl"},
	{0x9BFABC7070EA2471ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_flurry.ani"},
	{0x9BFACC6627B4AD7FULL, "299861095.ogg"},
	{0x9BFB00F992750000ULL, "vxw_woundedcoast_1fz_0_0.vlm"},
	{0x9BFC6FC201B1F66BULL, "vxw_woundedcoast_10c_0_0.vlm"},
	{0x9BFC752238ABDFA3ULL, "vxdr_primevaltomb_8f5_0_0.vlm"},
	{0x9BFD4C36D97767E2ULL, "one140cr_charade6139431.cl"},
	{0x9BFDA7EC194DDBD9ULL, "vxw_bonepitdestroyed_a3_1lv_0_0.vlm"},
	{0x9BFDF13E450DADB1ULL, "6072318_m.fxe"},
	{0x9BFE45D51B256D55ULL, "325746745.ogg"},
	{0x9BFE7A2FBE197E04ULL, "w_bonepit_39_0.msh"},
	{0x9BFE7EA26C61A803ULL, "llac_dr_slavers_3207_5.dds"},
	{0x9BFE8C2FBE199D6DULL, "w_bonepit_39_0.mao"},
	{0x9BFE902FBE19A3A6ULL, "w_bonepit_39_0.mmh"},
	{0x9BFF411AF800797BULL, "6156740_m.fxe"},
	{0x9BFF5C40FD79A145ULL, "273590094.ogg"},
	{0x9BFF810972B7A599ULL, "6134755_m.fxe"},
	{0x9BFFF7D94D5AFEC7ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_2_sylvesta.ani"},
	{0x9C003FC212282A14ULL, "gui_passthrough_p0.psh.fxo"},
	{0x9C009C30033DEC7BULL, "llac_ca_caverns2prc_5802_54.dds"},
	{0x9C00B35E718F984EULL, "634430756.ogg"},
	{0x9C01A90B49394B7AULL, "6151308_m.fxe"},
	{0x9C01C77070F05513ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_flurry.gad"},
	{0x9C020768FCDF8C63ULL, "6133824_m.fxe"},
	{0x9C025846451CE8E2ULL, "hlac_lt_0_lowtown_-1_1z.dds"},
	{0x9C026C9AB4558FD6ULL, "6129551_f.fxe"},
	{0x9C02D2B136AE7F6AULL, "gen00flst_bethany.stg"},
	{0x9C02FBB6FE93E94FULL, "dae000_hate_enemies6167560.cl"},
	{0x9C036A2FD2828BE4ULL, "lt_0_lowtown_n_3_0.rml"},
	{0x9C03F53B0AD44E3EULL, "150978897.ogg"},
	{0x9C0451055A68899BULL, "follower_barks6153794.cl"},
	{0x9C04B1E9DAA478B3ULL, "vxlt_darktown_4kt_0_0.vlm"},
	{0x9C04D8C56366361DULL, "vxlt_gallowscourtyard_n_7er_0_0.vlm"},
	{0x9C052AF7697735D0ULL, "23567850.ogg"},
	{0x9C05E6EDF70F1E9CULL, "6209980_m.fxe"},
	{0x9C05EC943728E1D5ULL, "6051605_m.fxe"},
	{0x9C062B20E1D8F408ULL, "956989602.ogg"},
	{0x9C06A9D64BEF51FAULL, "241885305.ogg"},
	{0x9C06C93B43BF821CULL, "270681886.ogg"},
	{0x9C06E3792B432FD4ULL, "vxht_mansion2b_2mt_0_0.vlm"},
	{0x9C0704C8465AE637ULL, "gen00fl_varric6105412.cl"},
	{0x9C071986AB0B5B22ULL, "fxe_lava_spark.dds"},
	{0x9C072591BC5569EEULL, "942571727.ogg"},
	{0x9C072BAE547E7704ULL, "1003361077.ogg"},
	{0x9C07541EE3E6C128ULL, "503039235.ogg"},
	{0x9C0768B506FF1B19ULL, "gen00fl_anders6110119.cl"},
	{0x9C084116A8BBD6A9ULL, "vxlt_backalley_n_1xq_0_0.vlm"},
	{0x9C08828EDB1DC75CULL, "168036046.ogg"},
	{0x9C08C47070F66F78ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_flurry.evt"},
	{0x9C09D7CB99660C17ULL, "6097853_f.fxe"},
	{0x9C09F7B6887CC0D9ULL, "vxdr_primevaltomb_8er_0_0.vlm"},
	{0x9C0A36C831C05DACULL, "gen00fl_isabela6079268.cl"},
	{0x9C0A67AF792434DFULL, "llac_ca_caverns2_5802_4.dds"},
	{0x9C0AB20785AA2176ULL, "dae000ip_crowd_mf3_talk5.utp"},
	{0x9C0ACC1AB85F4E63ULL, "vxdr_primevaltomb_5uh_0_1.vlm"},
	{0x9C0B1BF06BC5ABCBULL, "262769074.ogg"},
	{0x9C0B244D865A3296ULL, "6164335_m.fxe"},
	{0x9C0B645EFB65EA46ULL, "hlac_lt_0_lowtown_-65535_5y.dds"},
	{0x9C0B981DC5978D09ULL, "vxht_brothel_n_4x4_0_0.vlm"},
	{0x9C0C2C38F14FDEB6ULL, "vxlt_gallowscourtyard_n_6yt_0_0.vlm"},
	{0x9C0C30D0239995F9ULL, "target_creature_pressed.cur"},
	{0x9C0C4709EA753FFFULL, "6043396_m.fxe"},
	{0x9C0C49D4C628B48EULL, "qcr220cr_carver.utc"},
	{0x9C0CB6BB64F2CD73ULL, "llac_lt_backalley_g_0_14.dds"},
	{0x9C0CFF6B0C6C634CULL, "374698261.ogg"},
	{0x9C0D29A845DBF588ULL, "6121820_m.fxe"},
	{0x9C0DE382865CC9BFULL, "dae000_mother6135365.cl"},
	{0x9C0E16D1FB941DA8ULL, "435377656.ogg"},
	{0x9C0E4053BB2ED6A5ULL, "vxht_keep_n_8ax_0_0.vlm"},
	{0x9C0EC4D82919DA63ULL, "vxdr_primevaltomb_7sy_0_0.vlm"},
	{0x9C0EEF7D8D74ABFEULL, "follower_barks6153858.cl"},
	{0x9C0EFF5764F5FD32ULL, "dae021st_brothel_ambients.stg"},
	{0x9C0F17CB2C82C3F4ULL, "6101181_m.fxe"},
	{0x9C0F288A1EEC18ADULL, "dae000cr_grand_cleric6187945.cl"},
	{0x9C0F72EBA25BB47CULL, "vxht_keep_n_3g0_0_0.vlm"},
	{0x9C0F902FBE28176DULL, "w_bonepit_39_0.rml"},
	{0x9C0F93AE54EC24FEULL, "vxlt_hangedman_1i3_0_0.vlm"},
	{0x9C10065992497268ULL, "6016689_m.fxe"},
	{0x9C10331D4C5E2664ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_grand.msh"},
	{0x9C1043D20456CA4FULL, "6039973_m.fxe"},
	{0x9C10451D4C5E44CDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_grand.mao"},
	{0x9C10491D4C5E4B06ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3pp_grand.mmh"},
	{0x9C105ACBBBBD34B7ULL, "gui_guipixelshader_p9.psh.fxo.dx11"},
	{0x9C1099AB79774910ULL, "6081443_m.fxe"},
	{0x9C10D6DFA3859D07ULL, "963938106.ogg"},
	{0x9C113575B3E79675ULL, "505548903.ogg"},
	{0x9C11740C32F388CBULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.dura.evt"},
	{0x9C11D392D43B3CC3ULL, "vxht_keep_n_733_0_0.vlm"},
	{0x9C134AA2555C4E9DULL, "vxlt_gallowscourtyard_n_7hq_0_0.vlm"},
	{0x9C13B0FBEAE7A0FAULL, "541827997.ogg"},
	{0x9C13F069A1FEF955ULL, "hlac_ca_caverns1_mines_24423_1y.dds"},
	{0x9C14027868A136E3ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_rf.evt"},
	{0x9C1476EFD28D83FFULL, "520243393.ogg"},
	{0x9C14A7DB91D2FC2CULL, "6180472_m.fxe"},
	{0x9C153A74038EC634ULL, "6163264_m.fxe"},
	{0x9C15D5612B01865BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_008a_s.dds"},
	{0x9C16137180ED28B9ULL, "hm_noble_gen_2.mor"},
	{0x9C166B7907BE993CULL, "vxdr_primevaltomb_7kj_0_0.vlm"},
	{0x9C1692BDE2C3ED77ULL, "242551521.ogg"},
	{0x9C1870EC35308075ULL, "vxlt_0_lowtown_a2_n_c_byd_0_0.vlm"},
	{0x9C188D0C32F9D18CULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.dura.gad"},
	{0x9C189FC79B514D3FULL, "362868213.ogg"},
	{0x9C18E2F8A116BEE8ULL, "vxlt_blackemporium_cj_0_0.vlm"},
	{0x9C1932915FE88392ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wic_0n.dds"},
	{0x9C1937EBDD230F41ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_2x.dds"},
	{0x9C19536BDD1660D7ULL, "949858943.ogg"},
	{0x9C19A2CDF13DF9A9ULL, "64534269.ogg"},
	{0x9C1A42B265110353ULL, "263190598.ogg"},
	{0x9C1A4C3806F1EFF8ULL, "1003102627.ogg"},
	{0x9C1A6A7CB51EAA11ULL, "hlac_w_blightlands_131069_0x.dds"},
	{0x9C1ACC0B3C64C50CULL, "vrc321_housekeeper6085997.cl"},
	{0x9C1AD103E5427138ULL, "vxlt_hangedman_1lj_0_0.vlm"},
	{0x9C1ADB7868A71364ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_rf.gad"},
	{0x9C1B2850D62399E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_exit.ani"},
	{0x9C1B6936C0616203ULL, "mrl000_marethari6069810.cl"},
	{0x9C1BA1EB6069CBAEULL, "hlac_lt_0_lowtown_a2_-131073_2y.dds"},
	{0x9C1BBDB23C905F03ULL, "hm_gamlen_h3s1.mor"},
	{0x9C1BEFC799085AC8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_roundtable.mmh"},
	{0x9C1BFBC799086F6BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_roundtable.mao"},
	{0x9C1C0DC799088DF2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_roundtable.msh"},
	{0x9C1C26C7CA1A09D9ULL, "vxlt_0_lowtown_docks_b0g_0_0.vlm"},
	{0x9C1CCC7D7DBB2197ULL, "zz_approval_debug6003736.cl"},
	{0x9C1D0206F5649161ULL, "79450024.ogg"},
	{0x9C1D04E1D64E7C20ULL, "6078443_m.fxe"},
	{0x9C1D58BE280A580AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_dir_cast.gad"},
	{0x9C1D66C4663F6745ULL, "mag370_meredith6064854.cl"},
	{0x9C1DA7BD0AED4121ULL, "449753849.ogg"},
	{0x9C1E1A64B4F0A989ULL, "6020651_f.fxe"},
	{0x9C1E2886AB1E53C0ULL, "fxe_lava_spark.mao"},
	{0x9C1E8C0C32FE3B8AULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.dura.ani"},
	{0x9C1EC285712D11A1ULL, "182504922.ogg"},
	{0x9C1F8B047F14E9A3ULL, "6166963_m.fxe"},
	{0x9C1F92865975D6C0ULL, "ser120_ghyslain6118436.cl"},
	{0x9C1FAE9387BCD087ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_monique.gad"},
	{0x9C1FC2640938F4CAULL, "1013750344.ogg"},
	{0x9C2002F832C1E53DULL, "6126428_m.fxe"},
	{0x9C20281E03ABADDDULL, "hm_gencr_coterie_alchemist.mor"},
	{0x9C203C374E80166BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleft_3.mao"},
	{0x9C20731973A64E8BULL, "zz_approval_debug6003664.cl"},
	{0x9C2084F531C6F080ULL, "plt_seb120pt_duty_wounded.plo"},
	{0x9C20E195CBDF22B6ULL, "clf260_harley6069574.cl"},
	{0x9C215E6E599E56DAULL, "hlac_dr_primevalentrance_1726_12y.dds"},
	{0x9C21DA25ABFA0C2FULL, "art\\characters\\playercharacter\\humanmale\\arm_wmda.phy"},
	{0x9C21EA7868AD4B42ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_rf.ani"},
	{0x9C23A0645F6A2F5BULL, "vxlt_undercitydungeon_21c_0_0.vlm"},
	{0x9C23A317F5D747B6ULL, "vxdr_0_deeproads_f5m_0_0.vlm"},
	{0x9C23B78456DAAB16ULL, "vxht_brothel_n_57r_0_1.vlm"},
	{0x9C23E741ADF8615AULL, "272556249.ogg"},
	{0x9C2409E860E414DDULL, "559073426.ogg"},
	{0x9C241DF6E0456DB4ULL, "745810252.ogg"},
	{0x9C243BBE281045A1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_dir_cast.evt"},
	{0x9C2473F55045F737ULL, "vxht_mansion2_1kz_0_0.vlm"},
	{0x9C24815DE62DA93EULL, "dae000_merchant_dm_mob6174307.cl"},
	{0x9C2509140CDA8554ULL, "6151478_m.fxe"},
	{0x9C252D8DC5CB9A0DULL, "vxlt_gallowstemplar_f_4si_0_0.vlm"},
	{0x9C259579FEE6C377ULL, "gen00fl_isabela6079045.cl"},
	{0x9C259A64694917AAULL, "vxlt_gallowscourtyard_70u_0_0.vlm"},
	{0x9C262550F1874F36ULL, "darktown_fem_fer_urch_sta6174797.cl"},
	{0x9C266696ED0DD714ULL, "6202031_m.fxe"},
	{0x9C26722F0F3A3578ULL, "vxht_0_hightown_ca9_0_0.vlm"},
	{0x9C268931421E1359ULL, "gen00fl_carver6108947.cl"},
	{0x9C26B33415B6DBAEULL, "llac_lt_0_lowtown_n_65534_3.dds"},
	{0x9C271390CF67F9D3ULL, "hlac_w_encounter1_65540_0x.dds"},
	{0x9C275CECE51B75EFULL, "gen00fl_merrill6127272.cl"},
	{0x9C277D0BADA08832ULL, "187381114.ogg"},
	{0x9C27F6D6E80FEEEAULL, "dae000ip_crowd_lm1_talk10.utp"},
	{0x9C282DC3A06F9BDFULL, "hlac_w_sundermount_262141_0z.dds"},
	{0x9C2887BE5A7A5C4AULL, "693355003.ogg"},
	{0x9C2927EB9BB4229AULL, "6123644_m.fxe"},
	{0x9C295FC2AB84FF73ULL, "6098514_m.fxe"},
	{0x9C296DADD29B21B5ULL, "dae011_drunk1_sta6144380.cl"},
	{0x9C29A1C571C262D2ULL, "hlac_w_sundermount_-131075_0y.dds"},
	{0x9C29DDF44ADB7938ULL, "ser210_witness6168156.cl"},
	{0x9C29F36BBE5FE141ULL, "co_bl_sh.mao"},
	{0x9C2A2D1FEA99DABEULL, "art\\vfx\\combat\\vfx_m_spi_crh_aoe_p.mmh"},
	{0x9C2A98A4BC102259ULL, "vxlt_0_lowtown_docks_b7f_0_0.vlm"},
	{0x9C2B0F1510A4A88FULL, "hlac_lt_0_lowtown_-65536_0z.dds"},
	{0x9C2B7377A1EC0710ULL, "vxlt_0_lowtown_a2_n_c_ep1_0_0.vlm"},
	{0x9C2B90D2B3153A01ULL, "vxca_caverns2prc_a2y_0_0.vlm"},
	{0x9C2B95953C5D6D77ULL, "llac_lt_0_lowtown_docks_196607_0.dds"},
	{0x9C2C53BE47F26453ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stand_n.dds"},
	{0x9C2C725C0CE8A50DULL, "6037348_m.fxe"},
	{0x9C2C793CDC6ED789ULL, "isa111_cut_isabela_intro6069877.cl"},
	{0x9C2C923472C71C6CULL, "vxlt_0_lowtown_docks_bmy_0_0.vlm"},
	{0x9C2D23DE3CA5922CULL, "944377684.ogg"},
	{0x9C2D5D5C27BE8B50ULL, "fxe_waterfall_add.mao"},
	{0x9C2E2233591AE5A5ULL, "hlac_lt_0_lowtown_a2_n_-2_4z.dds"},
	{0x9C2E74D524B7FDBDULL, "6123728_m.fxe"},
	{0x9C2E9C2DBEE7807FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_drunk.gad"},
	{0x9C2F02254A26B8BEULL, "ach_exp_codex.dds"},
	{0x9C2F1D8053C8436EULL, "6131965_f.fxe"},
	{0x9C2F3F50D63469CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_exit.gad"},
	{0x9C2FA7E7A99AD9AFULL, "vrc321_varric6086930.cl"},
	{0x9C300DB2331180A3ULL, "995042929.ogg"},
	{0x9C303E7C2E5D9CA9ULL, "6095249_m.fxe"},
	{0x9C3041AD2DB3D9F6ULL, "dae000_oldmangossip6174979.cl"},
	{0x9C30B25BB31C7819ULL, "static1_nolit.psh.fxo"},
	{0x9C3133BE281AC2F0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_dir_cast.ani"},
	{0x9C315E5939A61F8FULL, "llac_lt_gallowscourtyard_n_-65536_2.dds"},
	{0x9C31D3D790A45358ULL, "450055603.ogg"},
	{0x9C321B28CC4A29D6ULL, "6162446_f.fxe"},
	{0x9C3248AE66F84092ULL, "vxht_mansion2_3gq_0_0.vlm"},
	{0x9C32877E3A9B26AAULL, "one300_zevran6070906.cl"},
	{0x9C32B509B5F03BD5ULL, "vxca_caverns1_mines_iht_0_0.vlm"},
	{0x9C3307857506B152ULL, "vis121_saemus6167898.cl"},
	{0x9C330BA12D38BA8BULL, "6042485_m.fxe"},
	{0x9C330EE830755B18ULL, "follower_banter6171600.cl"},
	{0x9C33130F43BB17FEULL, "vxw_encounter2_camp_59_0_0.vlm"},
	{0x9C33221BBBC2483DULL, "vxlt_warehouse_n_ul_0_0.vlm"},
	{0x9C3351B38C89663FULL, "6060337_m.fxe"},
	{0x9C33B874E9BB4490ULL, "ave310_donnic6062088.cl"},
	{0x9C340C415B3F302AULL, "vxlt_gallowscourtyard_75v_0_0.vlm"},
	{0x9C34332946489B58ULL, "art\\characters\\playercharacter\\humanmale\\mt_hlm_hn_eyesm1.mao"},
	{0x9C3489E94FFDB0C0ULL, "vxdr_0_deeproads_cfg_0_0.vlm"},
	{0x9C34912DBEEBD96CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_drunk.evt"},
	{0x9C349B9387CF0CCDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_monique.ani"},
	{0x9C349FAA04608637ULL, "vrc321_party_barks6096266.cl"},
	{0x9C34B77F603B4529ULL, "397937502.ogg"},
	{0x9C34C86D5A4D6200ULL, "vxw_woundedcoast_1rf_0_0.vlm"},
	{0x9C34CCFD412BFD77ULL, "753738837.ogg"},
	{0x9C34EF0B262E1583ULL, "lt_gallowscourtyard_39_0.rml"},
	{0x9C35F88E0351643AULL, "vxlt_0_lowtown_a3_n_c_eov_0_1.vlm"},
	{0x9C360A0FCD3A2C8DULL, "1272209.ogg"},
	{0x9C363C96008441A3ULL, "vxdr_primevaltomb_8r9_0_0.vlm"},
	{0x9C364F2D15BE6B41ULL, "vxw_bonepitdestroyed_a3_1vh_0_0.vlm"},
	{0x9C366EF6836377F9ULL, "llac_lt_0_lowtown_-131073_2.dds"},
	{0x9C369A27B35C5554ULL, "dae000_arishok6075200.cl"},
	{0x9C36F87675EAB710ULL, "vxlt_0_lowtown_a2_n_c_f4v_0_0.vlm"},
	{0x9C370138FC0B24DCULL, "780970087.ogg"},
	{0x9C3734301248A673ULL, "narrative_banter6165530.cl"},
	{0x9C37481E16EA67E2ULL, "6086211_m.fxe"},
	{0x9C376D1D52AFC173ULL, "55245396.ogg"},
	{0x9C378CC0F9F66AB1ULL, "framebuffereffects_bloomdiagonalblur_p0.psh.fxo"},
	{0x9C3799C88DAA3B1AULL, "6128744_m.fxe"},
	{0x9C380C1FEAA5E07DULL, "art\\vfx\\combat\\vfx_m_spi_crh_aoe_p.anb"},
	{0x9C384B13AE3386AEULL, "qcr220_grey_wardens6060393.cl"},
	{0x9C384CA0E41AE48BULL, "vxdr_0_deeproads_ev6_0_0.vlm"},
	{0x9C38B905CBCF5BE1ULL, "941064850.ogg"},
	{0x9C38F850DBC4D65FULL, "vxca_cavern1_haunted_j49_0_0.vlm"},
	{0x9C3954D75A25C981ULL, "hla_ht_0_hightown_-131070_0x.dds"},
	{0x9C3BD7FC8AFCE617ULL, "vxlt_0_lowtown_docks_a3_czh_0_1.vlm"},
	{0x9C3BED0A1C2AB112ULL, "llac_w_blightlands_196609_0.dds"},
	{0x9C3C08EB9007873CULL, "6208500_f.fxe"},
	{0x9C3C21B4108B1957ULL, "vxht_bartrand_n_5z9_0_0.vlm"},
	{0x9C3C5446934BAA3FULL, "mag110_arianni6016886.cl"},
	{0x9C3CAB843778B37BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_coma_0t.dds"},
	{0x9C3CAE366B18ECA9ULL, "460597480.ogg"},
	{0x9C3D5C2936ACF955ULL, "llac_lt_0_lowtown_a2_n_65538_2.dds"},
	{0x9C3D63C602120AD1ULL, "vxlt_0_lowtown_a2_n_gx8_0_0.vlm"},
	{0x9C3D928DE0AD9578ULL, "vxlt_gallowstemplar_4o3_0_0.vlm"},
	{0x9C3DB9F52965B669ULL, "vxlt_0_lowtown_a2_n_c_eo9_0_0.vlm"},
	{0x9C3E1F7194C9DC22ULL, "992179614.ogg"},
	{0x9C3F150295D046D3ULL, "272412962.ogg"},
	{0x9C3F93FA6E2C6F72ULL, "vxht_mansion2_1vt_0_0.vlm"},
	{0x9C3FF442C1578126ULL, "892797442.ogg"},
	{0x9C403CFEB1BF1188ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs102a.msh"},
	{0x9C404EFEB1BF3019ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs102a.mao"},
	{0x9C40511013EF335BULL, "923478624.ogg"},
	{0x9C405AFEB1BF44B2ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs102a.mmh"},
	{0x9C40FCA8E7F51102ULL, "6087173_m.fxe"},
	{0x9C410106783BE063ULL, "w_encounter3_97_0.msh"},
	{0x9C410F06783BF7BDULL, "w_encounter3_97_0.mmh"},
	{0x9C411306783BFECEULL, "w_encounter3_97_0.mao"},
	{0x9C415307D3BCD9DFULL, "6026973_m.fxe"},
	{0x9C416C4D5A57EBE4ULL, "6191254_m.fxe"},
	{0x9C4193F8ADB860BDULL, "vxlt_warehouse_n_xk_0_0.vlm"},
	{0x9C419A322A47C8F0ULL, "vxdr_0_deeproads_cg8_0_0.vlm"},
	{0x9C41B810B9AB3C3FULL, "fen110_fenris6096201.cl"},
	{0x9C420046C36CF81DULL, "6149566_m.fxe"},
	{0x9C422DB0CC3E8A59ULL, "gen00fl_varric6033553.cl"},
	{0x9C42792DBEF7EDC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_drunk.ani"},
	{0x9C428A023AC62B92ULL, "dre000_vampire6034304.cl"},
	{0x9C42DF78EA857408ULL, "follower_banter6163456.cl"},
	{0x9C42DFF6DA46B4F2ULL, "vxca_caverns1_mines_hii_0_0.vlm"},
	{0x9C42F058B7084F8BULL, "vxlt_warehouse_18u_0_0.vlm"},
	{0x9C43E8E0711C20D8ULL, "535046076.ogg"},
	{0x9C4400232A24E4F4ULL, "vxw_woundedcoast_1if_0_1.vlm"},
	{0x9C44B5C1241CBD12ULL, "gen00fl_carver6097228.cl"},
	{0x9C44C50124C488D8ULL, "6155931_m.fxe"},
	{0x9C44F121F4488C14ULL, "567734626.ogg"},
	{0x9C44FCD4930703C0ULL, "qun110_petrice6018725.cl"},
	{0x9C4514A8A4080A8FULL, "6024984_m.fxe"},
	{0x9C45415FDAF8F3F9ULL, "6129429_m.fxe"},
	{0x9C454D3D484424CAULL, "6164836_m.fxe"},
	{0x9C457F0708B6D042ULL, "43406049.ogg"},
	{0x9C459ACDB6E47EFBULL, "vxht_0_hightown_a2_bao_0_0.vlm"},
	{0x9C459CE9F83DB8C7ULL, "vxht_0_hightown_9wf_0_0.vlm"},
	{0x9C462A53729264C1ULL, "vxca_cavern1_haunted_i7u_0_0.vlm"},
	{0x9C463B860AFA764BULL, "6129100_m.fxe"},
	{0x9C4641AD0EEBB7BDULL, "vxlt_darktown_4tc_0_0.vlm"},
	{0x9C473F16A2B527C6ULL, "6065930_f.fxe"},
	{0x9C477BAD3EE69654ULL, "449608043.ogg"},
	{0x9C482D6189426C56ULL, "hlac_lt_0_lowtown_a2_n_1_4y.dds"},
	{0x9C48C8A4A8CE6814ULL, "combat_2h_range.gda"},
	{0x9C4924532E94F52DULL, "vxw_encounter2_camp_r7_0_0.vlm"},
	{0x9C4987C7D9271670ULL, "mag291cr_betrayer.ncs"},
	{0x9C4A740004007F17ULL, "814418636.ogg"},
	{0x9C4A89963D023350ULL, "1038813761.ogg"},
	{0x9C4AA97B09F08D43ULL, "lgt260_eau_tmp_roderick6143729.cl"},
	{0x9C4B847E9FF0CF22ULL, "hlac_lt_darktown_-1_73x.dds"},
	{0x9C4B9009E6CD1543ULL, "dae000ip_crowd_ff6_talk2.utp"},
	{0x9C4B9F2EC38F47A8ULL, "w_blightlands_201_0.msh"},
	{0x9C4BA16A4DD99E08ULL, "488363709.ogg"},
	{0x9C4BB12EC38F6639ULL, "w_blightlands_201_0.mao"},
	{0x9C4BBD2EC38F7AD2ULL, "w_blightlands_201_0.mmh"},
	{0x9C4C8D48826CF4A0ULL, "6166313_m.fxe"},
	{0x9C4C8E1F5A7E0974ULL, "6041779_m.fxe"},
	{0x9C4CDE66B59CE35EULL, "vxlt_0_lowtown_a2_ev8_0_0.vlm"},
	{0x9C4D84B64A192EF5ULL, "follower_banter6161000.cl"},
	{0x9C4DCAD92DCB4702ULL, "mag120_macha6087857.cl"},
	{0x9C4DECF6A7694D2AULL, "hlac_w_sundermount_-65535_0x.dds"},
	{0x9C4DF04686C1387DULL, "992651445.ogg"},
	{0x9C4E05D7602287F2ULL, "vxht_mansion2_1yu_0_0.vlm"},
	{0x9C4E4C1B78D436D3ULL, "hlac_w_encounter1_-65535_1x.dds"},
	{0x9C4EEF173C2A0B25ULL, "vxlt_ruinedhouse_n_jc_0_0.vlm"},
	{0x9C4EF6060FE3DBBBULL, "ser221_party_comments6116132.cl"},
	{0x9C4EFF1F2889BA4FULL, "vxlt_darktown_5sy_0_0.vlm"},
	{0x9C4F26E4C9FAAC66ULL, "727177777.ogg"},
	{0x9C4F93A94146C06BULL, "dre000_hireling1.bnk"},
	{0x9C4FBB9F166D51FDULL, "hlac_lt_0_lowtown_docks_131071_6z.dds"},
	{0x9C4FEC5688DE5826ULL, "6120933_m.fxe"},
	{0x9C5032ECCF96FB3CULL, "vxdr_primevaltomb_7wn_0_0.vlm"},
	{0x9C505AF90459C39DULL, "665353690.ogg"},
	{0x9C50F38ADEEED9C2ULL, "vis160cr_ginnis.utc"},
	{0x9C51214C3AD95BDEULL, "vxht_mansion_n_61a_0_0.vlm"},
	{0x9C515EFD5BF1F464ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_c.impa.ani"},
	{0x9C5313994D425605ULL, "vxht_0_hightown_a3_at8_0_0.vlm"},
	{0x9C5315F755186FDFULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.impa.gad"},
	{0x9C534B2214E2A6D5ULL, "hlac_w_woundedcoast_2_0x.dds"},
	{0x9C536F07FE2AC506ULL, "qcr000_arishok6059345.cl"},
	{0x9C5395A9414AAC4BULL, "dre000_hireling1.cnv"},
	{0x9C53D7D8DAB35654ULL, "w_woundedcoast_170_0.phy"},
	{0x9C5402A025B49643ULL, "6144901_m.fxe"},
	{0x9C545170716F180AULL, "art\\vfx\\combat\\vfx_c_mer_rush_strike_c.impa.gad"},
	{0x9C5492A637EE6208ULL, "lgt160_wk_avern_bandits.cnv"},
	{0x9C557165F4A18007ULL, "one100dg_kelder6055625.cl"},
	{0x9C55A2D19DA6970BULL, "genip_pp_lte_sustained.utp"},
	{0x9C55AA0B3E651DD7ULL, "172976450.ogg"},
	{0x9C55DC0EB8D75021ULL, "hlac_lt_0_lowtown_131070_5y.dds"},
	{0x9C55E30E44119CB6ULL, "hf_templar1.mor"},
	{0x9C560FB7EBC7D7D2ULL, "314739689.ogg"},
	{0x9C56AF632535B4B3ULL, "vxlt_warehouse_n_5r_0_0.vlm"},
	{0x9C56B107A1A95FBCULL, "408246484.ogg"},
	{0x9C56D89DF3B73A55ULL, "6051628_f.fxe"},
	{0x9C56E55DD6599C5EULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robd_0n.dds"},
	{0x9C5710C76CF774F2ULL, "crt_aoe_cycle2.mao"},
	{0x9C572851F21FDC3AULL, "mag121_idunna6084252.cl"},
	{0x9C574E512F6BCE94ULL, "tal_tau_01.mao"},
	{0x9C575208133F56A2ULL, "6177249_m.fxe"},
	{0x9C579630208B0DADULL, "vxw_encounter2_camp_m6_0_0.vlm"},
	{0x9C57C27FEDE4E11AULL, "396024569.ogg"},
	{0x9C57E1563FF5891FULL, "116739226.ogg"},
	{0x9C5812F2063671F4ULL, "follower_banter6170283.cl"},
	{0x9C585BFD5BF80DC6ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_c.impa.gad"},
	{0x9C589115938901FEULL, "176989750.ogg"},
	{0x9C58B4A637F28570ULL, "lgt160_wk_avern_bandits.bnk"},
	{0x9C5935982FC80FB6ULL, "767861148.ogg"},
	{0x9C5989D78C098CFFULL, "59525975.ogg"},
	{0x9C5A0AF7551E7BCCULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.impa.evt"},
	{0x9C5A351CE7BAC614ULL, "vxht_0_hightown_n_57f_0_0.vlm"},
	{0x9C5AB5FE1A71C54CULL, "mage_groupheal.dds"},
	{0x9C5AC15585A10FFAULL, "vxw_encounter2_7s_0_0.vlm"},
	{0x9C5ADDA7A30D49F7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_choke.gad"},
	{0x9C5AE092B4B2346BULL, "206919727.ogg"},
	{0x9C5AEA440C901CA5ULL, "ach_soc_narrative.dds"},
	{0x9C5B3470717505A1ULL, "art\\vfx\\combat\\vfx_c_mer_rush_strike_c.impa.evt"},
	{0x9C5B3F09FED12F79ULL, "vxlt_gallowstemplar_n_5qr_0_0.vlm"},
	{0x9C5C1655636D40B6ULL, "art\\characters\\creatures\\shade\\textures\\c_shadeb_0d.dds"},
	{0x9C5C3E8D03412CC0ULL, "vxht_chantry_2tg_0_0.vlm"},
	{0x9C5D02BC84BD9CE8ULL, "llac_lt_gallowstemplar_2382_1.dds"},
	{0x9C5D5D4396BA8E90ULL, "hlac_ht_mansion2b_818_33z.dds"},
	{0x9C5D98E85CA8E29DULL, "6015836_m.fxe"},
	{0x9C5D99ADD2794518ULL, "vxht_keep_6g2_0_0.vlm"},
	{0x9C5D9BBB0D3150EAULL, "6102552_f.fxe"},
	{0x9C5E0C0165250F1BULL, "gen_im_wep_rog_dua_act2_04.uti"},
	{0x9C5E38C2451F2B9CULL, "hlac_w_sundermount_-262150_0z.dds"},
	{0x9C5E3A46E1A6BDE6ULL, "84525899.ogg"},
	{0x9C5E4725DDCF6753ULL, "1015759256.ogg"},
	{0x9C5E70E42EAE752BULL, "mag100_cullen6095179.cl"},
	{0x9C5EB4D8DABD2251ULL, "w_woundedcoast_170_0.msh"},
	{0x9C5EC116D8DA985FULL, "vxlt_darktown_6ju_0_0.vlm"},
	{0x9C5EC2D8DABD3A6CULL, "w_woundedcoast_170_0.mao"},
	{0x9C5ECED8DABD4E0FULL, "w_woundedcoast_170_0.mmh"},
	{0x9C5F183CEC689602ULL, "37007375.ogg"},
	{0x9C5F295AA5D8B53EULL, "6047403_m.fxe"},
	{0x9C5F56FD5BFE23A5ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_c.impa.evt"},
	{0x9C5FD4129AF4808BULL, "vxlt_warehouse_12w_0_0.vlm"},
	{0x9C600B7B2FF4147BULL, "138809729.ogg"},
	{0x9C6039AF28186452ULL, "6060141_m.fxe"},
	{0x9C6040403F1FA95AULL, "qcr220_orsino6060740.cl"},
	{0x9C6042A2320EC260ULL, "vxht_0_hightown_c1q_0_0.vlm"},
	{0x9C60527FBE379F08ULL, "6128975_m.fxe"},
	{0x9C60667FF2F3127AULL, "mag320_orsino6071122.cl"},
	{0x9C6074DC9A941DAEULL, "6053925_m.fxe"},
	{0x9C612600877B7D77ULL, "636442725.ogg"},
	{0x9C618AA8F3A45AEEULL, "vxdr_primevalentrance_30l_0_1.vlm"},
	{0x9C61BAE71E2877F2ULL, "vxlt_0_lowtown_docks_a3_n_bs5_0_0.vlm"},
	{0x9C61F2A7A3138B84ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_choke.evt"},
	{0x9C62111E72F8D3C3ULL, "6086180_m.fxe"},
	{0x9C625E46B90F55EEULL, "6116359_f.fxe"},
	{0x9C62A74948075DC5ULL, "35078862.ogg"},
	{0x9C62C22C37E48124ULL, "6092563_f.fxe"},
	{0x9C62D850A3B97F1BULL, "dae000_arishok6068513.cl"},
	{0x9C63F1E4307EA8F3ULL, "6017501_f.fxe"},
	{0x9C645000EC8A70EBULL, "dae000_meredith6170927.cl"},
	{0x9C64A2DA865A0F53ULL, "6089865_f.fxe"},
	{0x9C64BEF429C226C0ULL, "6093902_f.fxe"},
	{0x9C64D7BAF8F4CDEDULL, "gui_guipixelshader_p5.psh.fxo"},
	{0x9C652140172BCD33ULL, "vxlt_warehouse_n_8u_0_0.vlm"},
	{0x9C65840B741F0ED0ULL, "hlac_ht_0_hightown_n_-65534_0z.dds"},
	{0x9C65A670CA0A8FB7ULL, "hlac_w_encounter2_65540_0y.dds"},
	{0x9C65AE0205A44B66ULL, "vxlt_0_lowtown_a2_n_c_gv7_0_0.vlm"},
	{0x9C65E372EFE060DAULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyo.mmh"},
	{0x9C65E772EFE06711ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyo.mao"},
	{0x9C65F572EFE07F20ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyo.msh"},
	{0x9C66080D1281262DULL, "vxw_encounter2_camp_h5_0_0.vlm"},
	{0x9C668ABB67DE07EDULL, "vxdr_0_deeproads_9el_0_1.vlm"},
	{0x9C66C72321D7FBB1ULL, "vxlt_gallowstemplar_f_5jo_0_0.vlm"},
	{0x9C66F2F75528DD25ULL, "art\\vfx\\combat\\vfx_f_and_vengeance_c.impa.ani"},
	{0x9C670B6C5E3067BFULL, "qun211cr_final2.utc"},
	{0x9C670E5A135D93FEULL, "zz_dre_debug6168210.cl"},
	{0x9C671788CC7D97E8ULL, "follower_barks6153658.cl"},
	{0x9C677BF1C1204E25ULL, "lt_darktown.arl"},
	{0x9C678B64FB831C5BULL, "vxlt_0_lowtown_n_gko_0_0.vlm"},
	{0x9C681429C0AA6C34ULL, "240665254.ogg"},
	{0x9C682C70717F82F0ULL, "art\\vfx\\combat\\vfx_c_mer_rush_strike_c.impa.ani"},
	{0x9C6962CF89525703ULL, "hlac_lt_0_lowtown_docks_n_-196610_6x.dds"},
	{0x9C69A4DE3D03BEE5ULL, "239590117.ogg"},
	{0x9C6A6B121FA6415BULL, "vxht_mansion_n_540_0_0.vlm"},
	{0x9C6ABA959593399FULL, "vxlt_undercitydungeon_176_0_0.vlm"},
	{0x9C6AE991440EB8F2ULL, "vxht_mansion2_1gw_0_0.vlm"},
	{0x9C6B09B46C0E5DE0ULL, "bld_pro_03.mao"},
	{0x9C6B4C2CF0CA61DDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platformcrak.mao"},
	{0x9C6B502CF0CA6816ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platformcrak.mmh"},
	{0x9C6B52D146A3A51FULL, "qun100_mage_party_comment6019099.cl"},
	{0x9C6B5A2CF0CA7914ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_platformcrak.msh"},
	{0x9C6B900F63932B25ULL, "6053096_m.fxe"},
	{0x9C6B934F2BE07B9BULL, "6110907_f.fxe"},
	{0x9C6C02ADCEBB06A7ULL, "vrc221_bartrand6065393.cl"},
	{0x9C6C445780762DB6ULL, "hlac_lt_undercityencounter_-1_37z.dds"},
	{0x9C6C7CD688DC88F4ULL, "ht_bartrand_n_9156_ll.mmh"},
	{0x9C6CA566F790D043ULL, "vxca_caverns2_7wc_0_0.vlm"},
	{0x9C6CE29DDBEEF42BULL, "llac_ht_playershouse_n_855_33.dds"},
	{0x9C6D423432853DE5ULL, "zz_mrl_debug6086524.cl"},
	{0x9C6DD9ED4F7A4440ULL, "6170664_m.fxe"},
	{0x9C6DE13660AEE4EFULL, "vxlt_hangedman_s5_0_0.vlm"},
	{0x9C6ED1900AD575ACULL, "vxht_0_hightown_a2_at4_0_0.vlm"},
	{0x9C6EF2B4193F14B5ULL, "art\\vfx\\models\\vfx_tube.msh"},
	{0x9C6EF4B4193F182BULL, "art\\vfx\\models\\vfx_tube.mmh"},
	{0x9C6F0AA7A31E3F7DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_choke.ani"},
	{0x9C6F396EFFEBB8ABULL, "and100_anders6095847.cl"},
	{0x9C6F4CEA43F56ECBULL, "ico_armor_medium_1.dds"},
	{0x9C6F50B72C14FD2AULL, "llac_dr_0_deeproads_5845_0.dds"},
	{0x9C6F60B1052A89FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v3.gad"},
	{0x9C6F87BF8D3B3B7BULL, "art\\characters\\creatures\\spider\\animation\\k_bk.evt"},
	{0x9C6F9F729F5CC783ULL, "hlac_w_bonepitdestroyed_a3_0_10x.dds"},
	{0x9C7008C8776FBC0DULL, "dae000_gen_turn_in_26178489.cl"},
	{0x9C7034284189AF04ULL, "6053332_m.fxe"},
	{0x9C70654BD80C4C59ULL, "bld_drian2.dds"},
	{0x9C70914F6349912CULL, "dm_gen_carta_thug.mor"},
	{0x9C70D9E61A25A744ULL, "810966053.ogg"},
	{0x9C71A282AA75E311ULL, "vxlt_undercityserialkiller_1su_0_0.vlm"},
	{0x9C71B5BA12D0DE3EULL, "251504980.ogg"},
	{0x9C71BA04B78DFB71ULL, "art\\levels\\races\\design\\_textures\\prp_chesta_d.dds"},
	{0x9C720FA4D36E6671ULL, "442529993.ogg"},
	{0x9C729459A7AC809BULL, "vxlt_foundry_n_yd_0_0.vlm"},
	{0x9C72FBC773494538ULL, "vxca_caverns1_mines_ei2_0_0.vlm"},
	{0x9C7479EA04773EADULL, "vxw_encounter2_camp_c4_0_0.vlm"},
	{0x9C748BB8BCCB6C3EULL, "ht_0_hightown_42_0.rml"},
	{0x9C7563B1052EFA11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v3.evt"},
	{0x9C7581810C04DF21ULL, "6127597_m.fxe"},
	{0x9C75E927D8637B57ULL, "vxht_bartrand_n_5n5_0_0.vlm"},
	{0x9C763E9253C49831ULL, "archertacticpreset.gda"},
	{0x9C7680BF8D414DBCULL, "art\\characters\\creatures\\spider\\animation\\k_bk.gad"},
	{0x9C76D1B1983F5AB7ULL, "vxht_bartrand_n_5q4_0_0.vlm"},
	{0x9C76F778F2D6FB13ULL, "vxlt_darktown_5m3_0_0.vlm"},
	{0x9C774409B2725BD6ULL, "6201241_m.fxe"},
	{0x9C776ADA23E2AAB9ULL, "6044272_m.fxe"},
	{0x9C7777092C269601ULL, "vxlt_darktown_502_0_0.vlm"},
	{0x9C779AA86A29069FULL, "6090068_m.fxe"},
	{0x9C77E412E707D2A7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hightlighta_d.dds"},
	{0x9C77F86B93D1757AULL, "vxw_blightlands_az_0_0.vlm"},
	{0x9C7807350A7CBB23ULL, "6005057_m.fxe"},
	{0x9C7871F53211460FULL, "plt_seb120pt_duty_wounded.nss"},
	{0x9C78E118956F9328ULL, "vxlt_0_lowtown_a2_n_c_bz4_0_0.vlm"},
	{0x9C78F056CB4FC4E5ULL, "59955264.ogg"},
	{0x9C7937FCEF4846F1ULL, "lt_0_lowtown_docks_175_0.msh"},
	{0x9C7945FCEF485E0CULL, "lt_0_lowtown_docks_175_0.mao"},
	{0x9C7951FCEF4872AFULL, "lt_0_lowtown_docks_175_0.mmh"},
	{0x9C79602659756C1DULL, "vxca_caverns1_ik8_0_0.vlm"},
	{0x9C7973DE7059A1ACULL, "6060129_m.fxe"},
	{0x9C799D710A12134DULL, "6100770_m.fxe"},
	{0x9C7A66EA7924061FULL, "vxca_cavern1_haunted_hl3_0_0.vlm"},
	{0x9C7A7D67B6657618ULL, "vxht_keep_6m0_0_0.vlm"},
	{0x9C7A94D221AD0FDEULL, "vxht_keep_n_6m1_0_0.vlm"},
	{0x9C7ABF395A73CEFBULL, "6025359_f.fxe"},
	{0x9C7B5DFDC288700DULL, "756535985.ogg"},
	{0x9C7C2728B28501F1ULL, "vxht_brothel_n_3et_0_0.vlm"},
	{0x9C7C7C7F372819ACULL, "6118385_m.fxe"},
	{0x9C7C9A09B5BD2130ULL, "hlac_ht_0_hightown_n_65537_1y.dds"},
	{0x9C7CBECE3CC91970ULL, "6127224_m.fxe"},
	{0x9C7CF44703D949A2ULL, "vxca_caverns2prc_8mf_0_0.vlm"},
	{0x9C7D1D6E7951E83CULL, "hlac_lt_undercityserialkiller_-1_32y.dds"},
	{0x9C7D2456C32B99AEULL, "hlac_w_sundermount_0_5x.dds"},
	{0x9C7D3B37C27909D3ULL, "58711243.ogg"},
	{0x9C7D4B1CE4957C94ULL, "6055157_f.fxe"},
	{0x9C7D5746071ACBABULL, "vxca_caverns1_mines_eb8_0_0.vlm"},
	{0x9C7D7FBF8D476ABAULL, "art\\characters\\creatures\\spider\\animation\\k_bk.ani"},
	{0x9C7E1ADA9EE5361CULL, "372729963.ogg"},
	{0x9C7E50F39F876597ULL, "188521493.ogg"},
	{0x9C7E887DCF8CE5C1ULL, "vxw_bonepitdestroyed_a3_1oc_0_0.vlm"},
	{0x9C7EFE22AB99CF03ULL, "6080555_m.fxe"},
	{0x9C7F6043234C0EABULL, "ser221_gascard6149156.cl"},
	{0x9C7F62A8DF7DE179ULL, "rdr111dg_martin6054491.cl"},
	{0x9C7FB863EB8C03EAULL, "786742325.ogg"},
	{0x9C800D04B3CC9B39ULL, "vxht_bartrand_n_6q9_0_0.vlm"},
	{0x9C80130592F90F77ULL, "mag170_hugh6092222.cl"},
	{0x9C806E26B99BE714ULL, "vxw_encounter1_bk_0_0.vlm"},
	{0x9C8071A92A0DB29DULL, "gen00fl_sebastian6195845.cl"},
	{0x9C80812FFC3D511DULL, "vxlt_gallowstemplar_n_46j_0_0.vlm"},
	{0x9C808E9E32E48CD2ULL, "6116815_m.fxe"},
	{0x9C809FDCE8F89201ULL, "one200_gold_followers6120148.cl"},
	{0x9C8200C5DE75E3FFULL, "dae000_gen_turn_in_36178448.cl"},
	{0x9C8204C57155DFE8ULL, "vxht_mansion2b_283_0_0.vlm"},
	{0x9C8307EDEDAE5786ULL, "vxlt_hangedman_2xs_0_0.vlm"},
	{0x9C8353FAA6003D79ULL, "gen00fl_varric6078703.cl"},
	{0x9C837BB1053B60E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v3.ani"},
	{0x9C842C2D22B4C501ULL, "vxw_woundedcoast_1fw_0_0.vlm"},
	{0x9C845A4FF2405B13ULL, "769517673.ogg"},
	{0x9C84FD4408BE976FULL, "and111cr_refugee_2.utc"},
	{0x9C853C87D78B21D7ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.impa.gad"},
	{0x9C85438E8A357337ULL, "vxht_bartrand_n_5t7_0_0.vlm"},
	{0x9C8548A860E13EF6ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower3_varric_.ani"},
	{0x9C855A02D878839DULL, "vxlt_0_lowtown_a2_e8a_0_0.vlm"},
	{0x9C85BB50891A21E6ULL, "6129484_m.fxe"},
	{0x9C85D4DDF5783AE0ULL, "vxlt_gallowstemplar_42x_0_0.vlm"},
	{0x9C85EA658BF4DF37ULL, "vxlt_hangedman_n_21o_0_1.vlm"},
	{0x9C8616EC5B83597AULL, "vxw_encounter2_8n_0_0.vlm"},
	{0x9C8694A0D4B378F9ULL, "vxlt_gallowstemplar_n_5bw_0_0.vlm"},
	{0x9C86D0D377A814DEULL, "follower_banter6164558.cl"},
	{0x9C86DF297FA97408ULL, "vxlt_0_lowtown_docks_a3_cou_0_0.vlm"},
	{0x9C870B6D27435CACULL, "vxca_cavern1_haunted_dpt_0_0.vlm"},
	{0x9C874A64157510E4ULL, "llac_lt_0_lowtown_n_65536_7.dds"},
	{0x9C87CD4B27FAE9F2ULL, "vxht_mansion2_1mq_0_0.vlm"},
	{0x9C87D2034B6B849DULL, "vxca_caverns1_in7_0_0.vlm"},
	{0x9C87D9A8FA2D2A01ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsrock.phy"},
	{0x9C87DC48880F3189ULL, "pro000ip_corpse_weapons.utp"},
	{0x9C87E9B83972F879ULL, "6060447_m.fxe"},
	{0x9C8876A196D37BABULL, "6127018_m.fxe"},
	{0x9C88A9C8E3A9465BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_017_0n.dds"},
	{0x9C88CBE552CC9353ULL, "qun131_taskmaster6039404.cl"},
	{0x9C88E081778FDFEAULL, "follower_banter6163465.cl"},
	{0x9C88E97385BED05AULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh3a.mmh"},
	{0x9C88FB7385BEEEA0ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh3a.msh"},
	{0x9C8906AF13A3285EULL, "vxht_keep_n_6h2_0_0.vlm"},
	{0x9C894799B7D2072EULL, "vxht_0_hightown_a3_alo_0_0.vlm"},
	{0x9C897FE13D203414ULL, "llac_w_encounter2_327680_0.dds"},
	{0x9C8B3B1353C9A79BULL, "vxlt_0_lowtown_docks_n_b69_0_1.vlm"},
	{0x9C8C0272F00164E5ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyo.phy"},
	{0x9C8C513DAD25C6C4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglehead_n.dds"},
	{0x9C8C5187D7916464ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.impa.evt"},
	{0x9C8C5F78FFFF99A9ULL, "one300cr_ggt_thug2.utc"},
	{0x9C8C8977D1F60145ULL, "6099347_f.fxe"},
	{0x9C8C950402F5708FULL, "lt_0_lowtown_docks_n_232_0.mao"},
	{0x9C8C98A796791EF6ULL, "lla_ht_0_hightown_a3_-131072_0.dds"},
	{0x9C8CA10402F5852CULL, "lt_0_lowtown_docks_n_232_0.mmh"},
	{0x9C8CA70402F58F1EULL, "lt_0_lowtown_docks_n_232_0.msh"},
	{0x9C8DB524CD8F0C92ULL, "582316201.ogg"},
	{0x9C8DD947FEAEC863ULL, "charhighlight.psh.fxo"},
	{0x9C8DEE5E305AC659ULL, "371103479.ogg"},
	{0x9C8E21996A34120AULL, "hlac_ca_caverns1_24421_3y.dds"},
	{0x9C8ECFE2403A1E0EULL, "vxlt_blackemporium_tm_0_0.vlm"},
	{0x9C8EE003AB91FF94ULL, "vxw_encounter1_gh_0_0.vlm"},
	{0x9C8F4D327D312E5DULL, "800994906.ogg"},
	{0x9C8F83CAB5B68955ULL, "shelf_n.dds"},
	{0x9C8FDF8157357638ULL, "vxca_caverns1_mines_ec0_0_0.vlm"},
	{0x9C8FEA92DF687E5BULL, "180033260.ogg"},
	{0x9C9076A2634BF868ULL, "vxht_mansion2b_250_0_0.vlm"},
	{0x9C90BECD6E170408ULL, "zz_mrl_debug6109297.cl"},
	{0x9C90DAD763688572ULL, "822892391.ogg"},
	{0x9C9144D06DFF7BC8ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls107a_0d.dds"},
	{0x9C9185920C7AF9B1ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_k_bk.ani"},
	{0x9C91D16336B9341CULL, "994916312.ogg"},
	{0x9C922A23A5EEB28AULL, "vxht_playershouse_n_1gy_0_0.vlm"},
	{0x9C9262B41663C1B5ULL, "llac_dr_slavers_3207_3.dds"},
	{0x9C929E0A14AADD81ULL, "vxw_woundedcoast_1av_0_0.vlm"},
	{0x9C934A89E1426172ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_pot_sa.dds"},
	{0x9C936EA875D6DB46ULL, "vxca_caverns2prc_9ld_0_0.vlm"},
	{0x9C939E6AAEF87058ULL, "genip_pp_hte_one_shot.utp"},
	{0x9C93B08871431C8EULL, "6107116_m.fxe"},
	{0x9C942AD67C0BB7C5ULL, "vxw_encounter2_2g_0_0.vlm"},
	{0x9C9454E0D2DAE908ULL, "6162685_m.fxe"},
	{0x9C94706266074815ULL, "421816171.ogg"},
	{0x9C94DC2577BDA67AULL, "vxw_blightlands_ot_0_0.vlm"},
	{0x9C957D6B5489CC61ULL, "hlac_ht_keep_3732_3x.dds"},
	{0x9C95C0FC987B82A6ULL, "vxlt_0_lowtown_a2_ekf_0_1.vlm"},
	{0x9C966464FA7628FFULL, "beckmann.dds"},
	{0x9C97788C059940DEULL, "vxht_keep_n_6c3_0_0.vlm"},
	{0x9C977D5D51A95112ULL, "251150798.ogg"},
	{0x9C97834D3E4D4449ULL, "52407836.ogg"},
	{0x9C9795E3F24F4651ULL, "hlac_lt_0_lowtown_docks_a3_-1_1y.dds"},
	{0x9C97BDED7F3126D9ULL, "6054750_m.fxe"},
	{0x9C97C38DC7C6A356ULL, "6118750_m.fxe"},
	{0x9C9817732DB11732ULL, "face2_ds1y.psh.fxo"},
	{0x9C9867DD5837F08EULL, "hlac_lt_0_lowtown_a2_n_131072_2x.dds"},
	{0x9C9890920C812B53ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_k_bk.gad"},
	{0x9C9913940F8D12CDULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs103a_0d.dds"},
	{0x9C9914D20A5B430CULL, "335403860.ogg"},
	{0x9C99275EAEE86DD2ULL, "667330421.ogg"},
	{0x9C99483EE0C87A7BULL, "vxlt_0_lowtown_a2_hcc_0_0.vlm"},
	{0x9C9971A860F22D30ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower3_varric_.gad"},
	{0x9C9A01169EEAF978ULL, "vxht_0_hightown_c91_0_0.vlm"},
	{0x9C9A58FF0ADA1CE9ULL, "and110_refugee6090304.cl"},
	{0x9C9A6987D79DCA5DULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.impa.ani"},
	{0x9C9A7D8409037A98ULL, "nat380_sibling6103560.cl"},
	{0x9C9AC92834DB1815ULL, "hightown_male_noble2_mob6170904.cl"},
	{0x9C9AF48D14CB065EULL, "hlac_ht_brothel_n_3364_1x.dds"},
	{0x9C9B73361E45A9F1ULL, "170576537.ogg"},
	{0x9C9B7CB69525F9D2ULL, "6126496_m.fxe"},
	{0x9C9C1AB300AB9576ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_rtlb.anb"},
	{0x9C9C59FEB20DA1ADULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs102a.phy"},
	{0x9C9C75419774F19EULL, "6033574_m.fxe"},
	{0x9C9C873259C1779DULL, "vxht_0_hightown_n_bzg_0_0.vlm"},
	{0x9C9CC0854AA3F626ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\standard_13.mao"},
	{0x9C9CE1100283A70AULL, "vxdr_primevaltomb_2pt_0_0.vlm"},
	{0x9C9D67CF96EFB5E9ULL, "llac_ht_bartrand_n_2995_7.dds"},
	{0x9C9DA06EF27450B0ULL, "828197917.ogg"},
	{0x9C9DA86D01BE3725ULL, "llac_ht_keep_n_3741_0.dds"},
	{0x9C9DBDD827B47378ULL, "268347466.ogg"},
	{0x9C9E3C75E54EB10DULL, "6016907_m.fxe"},
	{0x9C9E586DCDDCBC25ULL, "qun110_petrice6041324.cl"},
	{0x9C9E771FE3530584ULL, "vxlt_0_lowtown_a3_n_c_eox_0_1.vlm"},
	{0x9C9E8D920C8591B8ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_k_bk.evt"},
	{0x9C9EA7C857DB710DULL, "6163481_m.fxe"},
	{0x9C9EF9658A72FA17ULL, "llac_lt_0_lowtown_a3_n_c_65535_2.dds"},
	{0x9C9F27ACBC06C98EULL, "583947352.ogg"},
	{0x9C9F9E098F46B73BULL, "vxca_caverns2prc_aar_0_0.vlm"},
	{0x9CA08B795CD78D4CULL, "vxht_bartrand_n_665_0_0.vlm"},
	{0x9CA08BE8A3974D35ULL, "gen00fl_anders6125555.cl"},
	{0x9CA09C0097E4CB0AULL, "vxht_playershouse_n_1zz_0_0.vlm"},
	{0x9CA0CA5B9BFC8858ULL, "vxlt_darktown_424_0_0.vlm"},
	{0x9CA0CDDB95AEF12EULL, "mag101_alain6032136.cl"},
	{0x9CA0E51890311F28ULL, "t1_stb_blk.tnt"},
	{0x9CA0ED285C47BBDDULL, "vxht_mansion2b_1iv_0_0.vlm"},
	{0x9CA10FE706A0F601ULL, "vxw_woundedcoast_1lu_0_0.vlm"},
	{0x9CA116FD67FD3571ULL, "6138795_m.fxe"},
	{0x9CA15323ABF7233EULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l180_rf.ani"},
	{0x9CA1BCDFAFBD9623ULL, "vxht_0_hightown_a2_c5n_0_0.vlm"},
	{0x9CA1E08567CCF3C6ULL, "vxca_caverns2prc_9ke_0_0.vlm"},
	{0x9CA27C61617AC603ULL, "vxca_caverns1_gjk_0_0.vlm"},
	{0x9CA2F00F984604C5ULL, "vxw_blightlands_aq_0_0.vlm"},
	{0x9CA34E0269B3BEFAULL, "vxw_blightlands_pu_0_0.vlm"},
	{0x9CA357D33908149CULL, "6128755_m.fxe"},
	{0x9CA3D562742A504FULL, "vxca_caverns1_mines_jl1_0_0.vlm"},
	{0x9CA401889883C90DULL, "6050417_m.fxe"},
	{0x9CA42A5605E83631ULL, "plt_cod_hst_dalish2.plo"},
	{0x9CA4483DDCDA3A14ULL, "hlac_lt_undercitydungeon_-1_35y.dds"},
	{0x9CA45EB83DDE69ECULL, "vxca_caverns1_mines_k2t_0_1.vlm"},
	{0x9CA490694FF814EFULL, "lgt261cr_rogue.utc"},
	{0x9CA4E279396DCDC5ULL, "zz_followers_banter6158802.cl"},
	{0x9CA504062A3C35DBULL, "6127414_m.fxe"},
	{0x9CA5063E64F1FE19ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_die_loop.gad"},
	{0x9CA552391EF4CD53ULL, "6067761_m.fxe"},
	{0x9CA556FE13584EC7ULL, "dae000_bodhan6129086.cl"},
	{0x9CA5C6231D64D057ULL, "follower_banter6141804.cl"},
	{0x9CA5E7D72F040FE5ULL, "vxht_mansion2_2or_0_1.vlm"},
	{0x9CA6179CED1F4657ULL, "6041853_m.fxe"},
	{0x9CA62B539BBE382EULL, "vxht_0_hightown_a3_ava_0_0.vlm"},
	{0x9CA6422714281616ULL, "w_bonepitmining_a2_53_0.msh"},
	{0x9CA64C2714282714ULL, "w_bonepitmining_a2_53_0.mmh"},
	{0x9CA6502714282E27ULL, "w_bonepitmining_a2_53_0.mao"},
	{0x9CA6D2717DC43B1DULL, "hla_ht_0_hightown_a2_131072_0y.dds"},
	{0x9CA6E404BB6A5EECULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1walk.msh"},
	{0x9CA6EA04BB6A68DEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1walk.mmh"},
	{0x9CA6EC8A9883EF37ULL, "pro000_opening6207394.cl"},
	{0x9CA6F604BB6A7D75ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1walk.mao"},
	{0x9CA6FFFB6C1BD3CCULL, "6127577_f.fxe"},
	{0x9CA796025B3DBEA3ULL, "vxlt_0_lowtown_db4_0_0.vlm"},
	{0x9CA796537047761CULL, "hlac_lt_0_lowtown_a2_n_c_65537_3y.dds"},
	{0x9CA7B40BF944AA88ULL, "6174869_f.fxe"},
	{0x9CA7B96FF870356CULL, "hlac_lt_0_lowtown_n_65538_3x.dds"},
	{0x9CA7BA1BD2BE92FBULL, "vxlt_0_lowtown_a2_hhb_0_0.vlm"},
	{0x9CA7F35D1C0CCD95ULL, "vxlt_0_lowtown_a2_gve_0_0.vlm"},
	{0x9CA8688B43638457ULL, "vxht_mansion2b_1by_0_0.vlm"},
	{0x9CA8A86D795AF37CULL, "1072110242.ogg"},
	{0x9CA922CCBA1674E4ULL, "vxlt_warehouse_n_14a_0_0.vlm"},
	{0x9CAA4846E927D088ULL, "vxlt_gamlenshouse_n_nn_0_0.vlm"},
	{0x9CAA5FCBF8DC4BFFULL, "701644421.ogg"},
	{0x9CAA6F1C4324996CULL, "qun120_arishok_guard6049215.cl"},
	{0x9CAB2ED65454CC7FULL, "vxht_0_hightown_a3_5yi_0_0.vlm"},
	{0x9CAC008D95C4391DULL, "hlac_w_woundedcoast_196609_0y.dds"},
	{0x9CAC173E64F83A67ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_die_loop.ani"},
	{0x9CAC660E91AD6C14ULL, "vxdr_primevaltomb_9cp_0_0.vlm"},
	{0x9CAD535244C08DA5ULL, "vxw_woundedcoast_1bw_0_0.vlm"},
	{0x9CAD62B182C00F09ULL, "ss_midclass_thugs_m_1.wbd"},
	{0x9CAD88B8BCFB2CE2ULL, "ht_0_hightown_42_0.anb"},
	{0x9CADCFFB71698390ULL, "hlac_w_bonepit_131075_0y.dds"},
	{0x9CAE6B23AC01D60FULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l180_rf.evt"},
	{0x9CAE753FC3482F8EULL, "507485447.ogg"},
	{0x9CAE92F8F46DF8ECULL, "vxlt_0_lowtown_docks_b8v_0_0.vlm"},
	{0x9CAF087385DFD465ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh3a.phy"},
	{0x9CAF1D959A8191EDULL, "dae000_unhappy_mage6165341.cl"},
	{0x9CAF3C388DF2A0D8ULL, "vxlt_darktown_495_0_0.vlm"},
	{0x9CB02EBCA1B3AEA3ULL, "vxht_0_hightown_a2_c0q_0_0.vlm"},
	{0x9CB10643B5EDBBE5ULL, "607646660.ogg"},
	{0x9CB10E905FF7E8C5ULL, "vxw_encounter2_8e_0_0.vlm"},
	{0x9CB13CAD466F788EULL, "608567758.ogg"},
	{0x9CB1444822F09F95ULL, "6076241_f.fxe"},
	{0x9CB15B023A73A5A6ULL, "vxca_cavern1_haunted_hzp_0_0.vlm"},
	{0x9CB161EC8A3C1D45ULL, "vxw_blightlands_jp_0_0.vlm"},
	{0x9CB17519685FF1F4ULL, "hlac_lt_shop_-1_8z.dds"},
	{0x9CB2E4C7FCE019F8ULL, "vxlt_0_lowtown_n_eid_0_0.vlm"},
	{0x9CB3141A10EA5EF7ULL, "376518993.ogg"},
	{0x9CB36C0BCAE8209BULL, "cd_one100dg_elren.bnk"},
	{0x9CB39C6C71ED204EULL, "6082039_m.fxe"},
	{0x9CB3C37551635574ULL, "vxlt_0_lowtown_a2_n_ex3_0_0.vlm"},
	{0x9CB3D6372F4245E4ULL, "6072188_m.fxe"},
	{0x9CB44972E89F1D12ULL, "vxht_mansion2_32z_0_0.vlm"},
	{0x9CB45863C9D28623ULL, "qcr000_varric6208252.cl"},
	{0x9CB4A747ABB2D456ULL, "6112550_m.fxe"},
	{0x9CB56C23AC07F6B8ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l180_rf.gad"},
	{0x9CB577CC25CAB999ULL, "gen00fl_fenris6127102.cl"},
	{0x9CB627D4899E5261ULL, "vxw_encounter2_38_0_0.vlm"},
	{0x9CB69A076D570C26ULL, "vxca_caverns1_ck1_0_0.vlm"},
	{0x9CB6F54EC8BF30A4ULL, "6016704_m.fxe"},
	{0x9CB70EF0E85705D6ULL, "w_sundermount_118_0.rml"},
	{0x9CB726223F805E3DULL, "471707254.ogg"},
	{0x9CB794A9AC0C8D64ULL, "vxlt_warehouse_n_19f_0_0.vlm"},
	{0x9CB80E46C9CD087FULL, "6033392_f.fxe"},
	{0x9CB82FCBF7344980ULL, "6054183_f.fxe"},
	{0x9CB86F24104D9D1EULL, "430811881.ogg"},
	{0x9CB88956BD9708B7ULL, "hlac_lt_0_lowtown_a2_65536_10x.dds"},
	{0x9CB8F4B113EADD3EULL, "vxdr_primevaltomb_4iq_0_0.vlm"},
	{0x9CB927C2E74E857EULL, "6169291_m.fxe"},
	{0x9CB977F68BD16324ULL, "573213456.ogg"},
	{0x9CB9D49C757C5BFCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a_2.mao"},
	{0x9CB9EA5214602186ULL, "614233630.ogg"},
	{0x9CBAE7DB55558E91ULL, "hlac_w_sundermount_-262140_0z.dds"},
	{0x9CBB019F381CC84EULL, "vxca_caverns1_mines_hma_0_0.vlm"},
	{0x9CBB4C2588296A35ULL, "vxw_encounter2_camp_l1_0_0.vlm"},
	{0x9CBBAEE4E33C1A8BULL, "6094824_m.fxe"},
	{0x9CBBBE21045DD079ULL, "dae000_flemeth6000971.cl"},
	{0x9CBBC52F36B6A625ULL, "vxw_woundedcoast_1mv_0_0.vlm"},
	{0x9CBBD08C743D340DULL, "vxht_keep_67s_0_0.vlm"},
	{0x9CBBED4CB690523DULL, "gen00fl_anders6125656.cl"},
	{0x9CBC1664E017E3EDULL, "6214635_m.fxe"},
	{0x9CBC60C58A545365ULL, "4132656.ogg"},
	{0x9CBC619C422F9A9EULL, "6141558_m.fxe"},
	{0x9CBCCB795177F822ULL, "vxht_0_hightown_n_aqz_0_0.vlm"},
	{0x9CBCD37FC340E28BULL, "vxlt_0_lowtown_docks_a3_95j_0_0.vlm"},
	{0x9CBD05B8458B97A6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_enter.evt"},
	{0x9CBD23506CF16DC6ULL, "586663259.ogg"},
	{0x9CBDD0E24033ECDDULL, "vxht_mansion2b_1gt_0_0.vlm"},
	{0x9CBDD25ED7FC26BAULL, "1005049540.ogg"},
	{0x9CBDF3A0EA8D2701ULL, "vxw_woundedcoast_1rk_0_0.vlm"},
	{0x9CBE07F0E85D18FAULL, "w_sundermount_118_0.phy"},
	{0x9CBE6E73D5E01BD0ULL, "262970099.ogg"},
	{0x9CBEFB55B59FCB12ULL, "354803506.ogg"},
	{0x9CBF1B5C5BDD13A5ULL, "splatter_decal.mao"},
	{0x9CBF5315AC770A93ULL, "vxlt_0_lowtown_docks_a3_n_d3j_0_0.vlm"},
	{0x9CBFCCDF2C69BE26ULL, "vxca_cavern1_haunted_huq_0_0.vlm"},
	{0x9CBFD2CAB21BE236ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc05a.phy"},
	{0x9CC005A54768B6B4ULL, "character_character_p214.psh.fxo.dx11"},
	{0x9CC05E59B1FCDE92ULL, "vxlt_warehouse_sh_0_0.vlm"},
	{0x9CC08D0853A81EE5ULL, "qunarimale.fxa"},
	{0x9CC0B75A1665C2B6ULL, "6024093_m.fxe"},
	{0x9CC0B7FDCDD5D074ULL, "hlac_lt_0_lowtown_-1_6y.dds"},
	{0x9CC0E0523A628CEFULL, "lgt210_eau_senestra6144873.cl"},
	{0x9CC17935714C5EC9ULL, "253465824.ogg"},
	{0x9CC18FCF4B56BB30ULL, "lt_0_lowtown_a3_n_c_20_0_ll.mmh"},
	{0x9CC26220B24E6F68ULL, "6115899_f.fxe"},
	{0x9CC31CB845902AF1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_enter.gad"},
	{0x9CC35D753AE3CC36ULL, "6166997_m.fxe"},
	{0x9CC3B020BD26FE12ULL, "298858293.ogg"},
	{0x9CC4320751D8D709ULL, "216196086.ogg"},
	{0x9CC436F4B55ED8C6ULL, "6094195_m.fxe"},
	{0x9CC47A922DB53240ULL, "dae011_corff_sta6149600.cl"},
	{0x9CC49A45254D9E7BULL, "6031141_f.fxe"},
	{0x9CC49FBC06942CC1ULL, "vxdr_primevaltomb_8rc_0_0.vlm"},
	{0x9CC4E0C5BE5F6836ULL, "qun110_petrice6018745.cl"},
	{0x9CC50BE45F4D24A6ULL, "vxca_caverns1_cn2_0_0.vlm"},
	{0x9CC5234C68AD3702ULL, "w_bonepitmining_a2_40_0.msh"},
	{0x9CC5254C68AD3AB8ULL, "w_bonepitmining_a2_40_0.mmh"},
	{0x9CC5314C68AD4F5BULL, "w_bonepitmining_a2_40_0.mao"},
	{0x9CC5B44B5038D8CEULL, "73087537.ogg"},
	{0x9CC5ED412FF492C6ULL, "zz_followers_banter6159462.cl"},
	{0x9CC654701968FC12ULL, "ser221_gamlen6135731.cl"},
	{0x9CC79F1BCB08E2DBULL, "258808525.ogg"},
	{0x9CC8254C68AF70CBULL, "w_bonepitmining_a2_40_0.rml"},
	{0x9CC83E761EBA6163ULL, "dae000_merchant_generic_hm6172598.cl"},
	{0x9CC856E827B5C750ULL, "758761488.ogg"},
	{0x9CC8FB14374FB236ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_5x.dds"},
	{0x9CC94B667E268152ULL, "hlac_w_bonepitmining_a2_131073_0y.dds"},
	{0x9CC955FADB650B2EULL, "lt_0_lowtown_docks_n_221_0.msh"},
	{0x9CC963FADB65237FULL, "lt_0_lowtown_docks_n_221_0.mao"},
	{0x9CC96FFADB65371CULL, "lt_0_lowtown_docks_n_221_0.mmh"},
	{0x9CC98F65F7BD1BF6ULL, "and230st_mage_girl.stg"},
	{0x9CC9BBCAB2240FD7ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc05a.msh"},
	{0x9CC9BE027A1F82B5ULL, "vxw_encounter2_camp_g0_0_0.vlm"},
	{0x9CC9D1CAB2243549ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bc05a.mmh"},
	{0x9CC9DA8CEE33D44AULL, "zz_followers_banter6159255.cl"},
	{0x9CC9FDB437E8D78DULL, "6094757_f.fxe"},
	{0x9CCA1DB845964A9FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_enter.ani"},
	{0x9CCA370C28ACBEA5ULL, "vxw_woundedcoast_1hu_0_0.vlm"},
	{0x9CCA426966334C8DULL, "vxht_keep_60r_0_0.vlm"},
	{0x9CCABE9DAADEABBFULL, "6116418_m.fxe"},
	{0x9CCADFA6CBFF0827ULL, "6173323_m.fxe"},
	{0x9CCAFA471C7FC444ULL, "llac_ht_playershouse_n_855_12.dds"},
	{0x9CCBA0FCBA125539ULL, "hlac_lt_0_lowtown_docks_a3_-262144_2z.dds"},
	{0x9CCBAC35CF2BCB5CULL, "gen_im_acc_amu_act1_01.uti"},
	{0x9CCBB92A7DFE34F2ULL, "vxht_bartrand_n_5p2_0_0.vlm"},
	{0x9CCC0D89A0692228ULL, "6103653_m.fxe"},
	{0x9CCC42BF322A055DULL, "vxht_mansion2b_1zs_0_0.vlm"},
	{0x9CCCB1D3E46307C2ULL, "6085526_m.fxe"},
	{0x9CCCE156D5C6ABD7ULL, "vxlt_0_lowtown_a2_n_gvl_0_0.vlm"},
	{0x9CCCFC630A3398F0ULL, "121307969.ogg"},
	{0x9CCD512B33C49F55ULL, "vxlt_0_lowtown_docks_n_apw_0_0.vlm"},
	{0x9CCDBD37AB17755BULL, "vxca_caverns1_jl8_0_0.vlm"},
	{0x9CCDDB4A80B1ED22ULL, "992725932.ogg"},
	{0x9CCDEE9C227E8603ULL, "6036864_m.fxe"},
	{0x9CCE3AAAD9E3D8A2ULL, "hlac_lt_0_lowtown_-2_3y.dds"},
	{0x9CCEEBB7B786B65BULL, "art\\characters\\playercharacter\\humanfemale\\arm_masd.mmh"},
	{0x9CCEF7B7B786CAF8ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masd.mao"},
	{0x9CCF09B7B786E945ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masd.msh"},
	{0x9CCF332EF55E4229ULL, "pro000_aveline_intro6109325.cl"},
	{0x9CCF3E2D87E9F38CULL, "ser211ip_dead_alessa.utp"},
	{0x9CCF65AAFAF36F26ULL, "10897692.ogg"},
	{0x9CD0BAC9A4C39338ULL, "docks_male_elven_worker_sta6175813.cl"},
	{0x9CD14615B594F386ULL, "vxht_brothel_n_704_0_0.vlm"},
	{0x9CD187C87052E969ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_5z.dds"},
	{0x9CD1CC8134B73675ULL, "lgt110cr_coa_kirk_rep.utc"},
	{0x9CD1F97C794F657CULL, "6164580_m.fxe"},
	{0x9CD2306CD1A709EAULL, "art\\characters\\creatures\\mubari\\c_dog_eyes.mao"},
	{0x9CD234B182DEDDEBULL, "ss_midclass_thugs_m_1.bnk"},
	{0x9CD270508DF34C96ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsplinter.phy"},
	{0x9CD325CB13996E65ULL, "ser270_templar6122500.cl"},
	{0x9CD3505F355F6879ULL, "llac_w_woundedcoast_-262144_0.dds"},
	{0x9CD3BE221945CDD7ULL, "vxht_mansion2b_1sz_0_0.vlm"},
	{0x9CD3DEFDD16AA451ULL, "sf_textcolor.psh.fxo"},
	{0x9CD43F153C8EDA0AULL, "730297194.ogg"},
	{0x9CD454AB526A5E9EULL, "dae000_arishok6067843.cl"},
	{0x9CD470438A601C52ULL, "6059408_m.fxe"},
	{0x9CD48FA39E1C43ABULL, "hlac_lt_undercitydungeon_-1_17z.dds"},
	{0x9CD49178ED1C0C9FULL, "vxlt_0_lowtown_a3_n_c_hq5_0_0.vlm"},
	{0x9CD4AA590E69234AULL, "mag200_feynriel6030441.cl"},
	{0x9CD55D6E98362A49ULL, "one220_yevhen6103131.cl"},
	{0x9CD57749B5B4C6B9ULL, "6122514_m.fxe"},
	{0x9CD5C482D2BE2A84ULL, "vxlt_0_lowtown_gob_0_1.vlm"},
	{0x9CD602D208F15E0DULL, "gen00fl_bethany6098059.cl"},
	{0x9CD68FC20A7C1923ULL, "vxw_bonepit_1p8_0_0.vlm"},
	{0x9CD709AF53D4B7F5ULL, "6085547_m.fxe"},
	{0x9CD74B18B44D14AAULL, "vxca_caverns2prc_9pl_0_0.vlm"},
	{0x9CD7F8F1D8E25575ULL, "vxlt_0_lowtown_docks_a3_n_cy5_0_0.vlm"},
	{0x9CD80E27980241E1ULL, "isa311_isabela6094464.cl"},
	{0x9CD8401CFE51E627ULL, "27914473.ogg"},
	{0x9CD883EEC999F6F2ULL, "6053525_m.fxe"},
	{0x9CD8863E614B1ABEULL, "443870851.ogg"},
	{0x9CD8907260BA2502ULL, "112900562.ogg"},
	{0x9CD8A8E91AA2D725ULL, "vxw_woundedcoast_1st_0_0.vlm"},
	{0x9CD8F635281E54D1ULL, "dae221_viscount_chantry.bnk"},
	{0x9CD90A1FD52068E8ULL, "ca_caverns1_mines_24422_ll.mmh"},
	{0x9CD92FA75F0C57F7ULL, "vxlt_0_lowtown_a2_n_c_hjm_0_0.vlm"},
	{0x9CD94A9CF8E21B3FULL, "vxlt_0_lowtown_docks_b8y_0_0.vlm"},
	{0x9CD95BC517297B9DULL, "6002523_f.fxe"},
	{0x9CD971AADB3C07E5ULL, "vxlt_0_lowtown_a2_gv5_0_0.vlm"},
	{0x9CD973902D85A47CULL, "gen00fl_isabela6088071.cl"},
	{0x9CDA07886AB6D242ULL, "vxlt_undercityserialkiller_17v_0_0.vlm"},
	{0x9CDA766383C55E19ULL, "6173026_m.fxe"},
	{0x9CDA7F66925F3AA8ULL, "6104753_m.fxe"},
	{0x9CDAC1A3F6640E7EULL, "vxca_caverns1_jg4_0_0.vlm"},
	{0x9CDB3CFB4BE5B1A9ULL, "6077986_m.fxe"},
	{0x9CDB6AA0B18DBAE1ULL, "221922201.ogg"},
	{0x9CDBD835282057E1ULL, "dae221_viscount_chantry.cnv"},
	{0x9CDC4ADB4350435BULL, "character_character_p77.psh.fxo.dx11.assembly"},
	{0x9CDC59508DFB7A37ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsplinter.msh"},
	{0x9CDC5AE3AE053DE2ULL, "qun111_ketojan6040712.cl"},
	{0x9CDC6F508DFB9FA9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsplinter.mmh"},
	{0x9CDC6F8AFAE9BA39ULL, "ht_mansion_n_probe_54o.dds"},
	{0x9CDC7014FD34D772ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad_0.pwk"},
	{0x9CDC7CCA1F8A2566ULL, "6144518_f.fxe"},
	{0x9CDC7E0DE4EDD04EULL, "art\\vfx\\combat\\vfx_gib_hm_masc_rtlb.anb"},
	{0x9CDD2CC29704170BULL, "gen00fl_isabela_init6076100.cl"},
	{0x9CDD421395E90F92ULL, "vxlt_warehouse_yj_0_0.vlm"},
	{0x9CDE09352EAE4BA1ULL, "llac_lt_0_lowtown_docks_-131074_4.dds"},
	{0x9CDE72087E59ABC6ULL, "858049477.ogg"},
	{0x9CDE75E21742CC48ULL, "qun111_ketojan6039813.cl"},
	{0x9CDE94BD686F815CULL, "hlac_lt_0_lowtown_docks_a3_n_0_4z.dds"},
	{0x9CDF25E5B67787D0ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_1z.dds"},
	{0x9CDF615494BA6C03ULL, "da2terrain_rendertodepth_p0.psh.fxo.dx11.assembly"},
	{0x9CE0600FDD270B16ULL, "340381991.ogg"},
	{0x9CE061B2B7C4DB2EULL, "6211925_m.fxe"},
	{0x9CE0A62BFADA9082ULL, "gen_im_trs_chn_amu_r1.uti"},
	{0x9CE0D6D18DF7C936ULL, "gen00flst_isa_act1.stg"},
	{0x9CE0F52A85AAF4D9ULL, "vxw_woundedcoast_yc_0_0.vlm"},
	{0x9CE112F10D100AFBULL, "35772935.ogg"},
	{0x9CE1527B47E3F628ULL, "vxht_store_c8_0_0.vlm"},
	{0x9CE186523230ED68ULL, "qun221_viscount6073889.cl"},
	{0x9CE1E2D8B06C7E1DULL, "vxdr_primevalentrance_3bd_0_0.vlm"},
	{0x9CE21EC8B1D68800ULL, "vxw_woundedcoast_16j_0_0.vlm"},
	{0x9CE22FFF0B3BE657ULL, "vxht_mansion2b_1vu_0_0.vlm"},
	{0x9CE268D761EB3379ULL, "llac_lt_0_lowtown_a2_131071_1.dds"},
	{0x9CE26A514245B653ULL, "6041152_m.fxe"},
	{0x9CE2A960380C0036ULL, "411020824.ogg"},
	{0x9CE3040B47DCCB03ULL, "vxlt_0_lowtown_docks_a3_d1g_0_0.vlm"},
	{0x9CE31B6627613B0FULL, "vxht_0_hightown_a3_c8l_0_0.vlm"},
	{0x9CE32BF72CD5BEDFULL, "w_encounter2_camp_19_0_ll.mmh"},
	{0x9CE35AAFEF419D51ULL, "6054634_f.fxe"},
	{0x9CE36539D447B00AULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.cess.gad"},
	{0x9CE3FCB12D945B6EULL, "hlac_lt_gallowstemplar_n_2391_3y.dds"},
	{0x9CE4B82BDE4E5A0EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11b.phy"},
	{0x9CE4C04FD4268B46ULL, "6075293_m.fxe"},
	{0x9CE4FF52C04F8DEFULL, "6165767_f.fxe"},
	{0x9CE550AACC204907ULL, "dae000ip_crowd_em1_talk16.utp"},
	{0x9CE5A983D5EC101FULL, "w_bonepitdestroyed_a3_184_0.msh"},
	{0x9CE5AF83D5EC1AC1ULL, "w_bonepitdestroyed_a3_184_0.mmh"},
	{0x9CE5BB83D5EC2F62ULL, "w_bonepitdestroyed_a3_184_0.mao"},
	{0x9CE5BCA6EB56ECFFULL, "vxlt_0_lowtown_docks_a3_n_bun_0_0.vlm"},
	{0x9CE5BEF2258993A2ULL, "723061845.ogg"},
	{0x9CE5D2D5B42D7D1FULL, "vxw_encounter2_fy_0_0.vlm"},
	{0x9CE6A58CAFA7AE31ULL, "vxlt_gallowstemplar_n_5qj_0_0.vlm"},
	{0x9CE74A89EB825E7CULL, "vxlt_0_lowtown_docks_a3_byn_0_0.vlm"},
	{0x9CE74ED0E945DF4CULL, "1031270048.ogg"},
	{0x9CE77E7F7B343D2DULL, "hlac_lt_0_lowtown_a2_n_c_-131071_1z.dds"},
	{0x9CE7B2172950411DULL, "llac_lt_gallowstemplar_f_2383_8.dds"},
	{0x9CE7DF665F6DDD0BULL, "6100761_m.fxe"},
	{0x9CE7F32522801636ULL, "308577804.ogg"},
	{0x9CE7F6D7F9BC9D74ULL, "vxlt_0_lowtown_docks_n_bev_0_0.vlm"},
	{0x9CE84F94DC4D0194ULL, "vxw_bonepitdestroyed_a3_23y_0_0.vlm"},
	{0x9CE85495B3118315ULL, "958728031.ogg"},
	{0x9CE860E9043EEE8BULL, "hla_ht_0_hightown_-131070_0z.dds"},
	{0x9CE917A667B6CA0AULL, "401355707.ogg"},
	{0x9CE91E1B8721B961ULL, "p_leather_stretcher_01d.dds"},
	{0x9CE92ABBA990BC92ULL, "287174458.ogg"},
	{0x9CE965A95E9D3B77ULL, "166477754.ogg"},
	{0x9CE9AAF9A308968CULL, "6036455_f.fxe"},
	{0x9CE9CE3A943D96F5ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgta.phy"},
	{0x9CE9D3BBBACCE74AULL, "6173177_m.fxe"},
	{0x9CEA34E517B0D055ULL, "vxlt_0_lowtown_docks_n_azu_0_0.vlm"},
	{0x9CEA4839D44D9DA1ULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.cess.evt"},
	{0x9CEA5B6973DEDB25ULL, "567980288.ogg"},
	{0x9CEADEA8FA81B03DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsrock.mao"},
	{0x9CEAE2A8FA81B776ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsrock.mmh"},
	{0x9CEAECA8FA81C874ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01ruinsrock.msh"},
	{0x9CEBBB20CD81347CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgta_0d.dds"},
	{0x9CEC1DE32CDC5820ULL, "hlac_lt_0_lowtown_docks_n_131071_4z.dds"},
	{0x9CECC1517F5B976BULL, "721546420.ogg"},
	{0x9CED23EA55411EEEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecartfram.msh"},
	{0x9CED31EA5541363FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecartfram.mao"},
	{0x9CED3DEA55414ADCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecartfram.mmh"},
	{0x9CED40F3DAB4315AULL, "6034676_m.fxe"},
	{0x9CED45480D69C70DULL, "vxlt_gallowstemplar_n_4ug_0_0.vlm"},
	{0x9CEE07A474C48F23ULL, "581163825.ogg"},
	{0x9CEE4D0A7D37E8EEULL, "6161589_m.fxe"},
	{0x9CEE7B166E9CF6FDULL, "816043993.ogg"},
	{0x9CEECF75EFC259B8ULL, "vxlt_backalley_g_1yg_0_1.vlm"},
	{0x9CEF025370718066ULL, "door_pressed.cur"},
	{0x9CEFA99957B7B1CAULL, "6090335_m.fxe"},
	{0x9CF0124904266EC6ULL, "gen_im_wep_war_two_act0_01.uti"},
	{0x9CF02FCD7F52EDA8ULL, "vxlt_0_lowtown_docks_a3_n_coj_0_0.vlm"},
	{0x9CF090A5A3CCA080ULL, "vxw_woundedcoast_11k_0_0.vlm"},
	{0x9CF0BB8AC48D2CDDULL, "vxlt_0_lowtown_a2_n_gnf_0_0.vlm"},
	{0x9CF0FE1EC03C4A0DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_tram.mmh"},
	{0x9CF1021EC03C515EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_tram.mao"},
	{0x9CF1101EC03C6893ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_tram.msh"},
	{0x9CF149ACFBAAE7FEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerpot01_0_sa.dds"},
	{0x9CF228FE6FD38D27ULL, "hlac_ca_caverns1_mines_24426_12z.dds"},
	{0x9CF22B6DE8A36274ULL, "47856367.ogg"},
	{0x9CF284334C4312D9ULL, "vxw_bonepit_1zt_0_0.vlm"},
	{0x9CF29F1883905289ULL, "256931441.ogg"},
	{0x9CF2D379D199B2D9ULL, "framebuffereffects_illuminationmask_p0.psh.fxo"},
	{0x9CF388F3F4E45680ULL, "zz_prmpt_debug6198152.cl"},
	{0x9CF3F30D1FCADAFFULL, "cod_bks_lite_vials_4.utp"},
	{0x9CF403B6D8400909ULL, "vxw_sundermount_34o_0_0.vlm"},
	{0x9CF41B39F8CDA612ULL, "llac_dr_primevaltomb_659_2.dds"},
	{0x9CF42ED2983945AAULL, "vxca_caverns2prc_9zn_0_0.vlm"},
	{0x9CF444B2A623959FULL, "vxw_encounter2_ax_0_0.vlm"},
	{0x9CF48BDF8064C6BCULL, "547837088.ogg"},
	{0x9CF4A4AA09354013ULL, "6148740_f.fxe"},
	{0x9CF4F0172DC610BAULL, "gen00fl_isabela6099703.cl"},
	{0x9CF5569EFC271854ULL, "vxlt_gallowstemplar_f_2oz_0_0.vlm"},
	{0x9CF668B4EBB2B5F4ULL, "vxlt_0_lowtown_docks_n_bbw_0_0.vlm"},
	{0x9CF6AB7061E40E97ULL, "vxdr_0_deeproads_cwa_0_2.vlm"},
	{0x9CF6C57659633DBCULL, "6059697_m.fxe"},
	{0x9CF74039D4581AF0ULL, "art\\vfx\\combat\\vfx_d_qun_immune_cast_c.cess.ani"},
	{0x9CF79B059A1FDAA4ULL, "hlac_lt_backalley_n_65537_8y.dds"},
	{0x9CF7B20C36192018ULL, "snow_01_l2.mao"},
	{0x9CF7BFA08E43696DULL, "vxdr_primevaltomb_50y_0_1.vlm"},
	{0x9CF859DB02A76C26ULL, "vxlt_undercitydungeon_1gs_0_0.vlm"},
	{0x9CF8F17528FA1032ULL, "6066363_m.fxe"},
	{0x9CF934562CE29184ULL, "vxdr_0_deeproads_cfc_0_0.vlm"},
	{0x9CF97CF730F65B54ULL, "hm_jansen.mor"},
	{0x9CF9D5896D6200A1ULL, "dwarf_male_heads.gda"},
	{0x9CF9F6426B245457ULL, "zz_mcr_debug6071530.cl"},
	{0x9CFA1A744E9EA6CCULL, "gen_im_cut_isabella_dagger.uti"},
	{0x9CFA5A0BCB240499ULL, "cd_one100dg_elren.wbd"},
	{0x9CFA78CA93FBF959ULL, "savescreen.gfx"},
	{0x9CFA87E2F3EFAB1DULL, "vxw_encounter3_9u_0_0.vlm"},
	{0x9CFA94913DEDFA76ULL, "gen_im_arm_boo_lgt_act3_03.uti"},
	{0x9CFAED9A4F5154C8ULL, "vxw_sundermount_2ba_0_0.vlm"},
	{0x9CFB4C2714706717ULL, "w_bonepitmining_a2_53_0.rml"},
	{0x9CFB6BFEFF541861ULL, "551550695.ogg"},
	{0x9CFB9F7EB0BCAD16ULL, "69931318.ogg"},
	{0x9CFBAC9366C81AE9ULL, "zz_followers_banter6159597.cl"},
	{0x9CFC1E0B3267C235ULL, "vxlt_0_lowtown_a3_n_c_goj_0_0.vlm"},
	{0x9CFC84D788D1B988ULL, "734167240.ogg"},
	{0x9CFC9BCCC16CF26DULL, "6106512_m.fxe"},
	{0x9CFCF5F12D513CE2ULL, "6067830_m.fxe"},
	{0x9CFD069C82F77163ULL, "dae000_bartrand6027829.cl"},
	{0x9CFD4D247F306EFEULL, "vxw_encounter2_camp_o7_0_0.vlm"},
	{0x9CFDB8AA0CDDE601ULL, "war_taunt.dds"},
	{0x9CFDD3A2B0C9E464ULL, "lt_gallowscourtyard_n_51_0.anb"},
	{0x9CFDD8E4699725D9ULL, "vxw_woundedcoast_sa_0_0.vlm"},
	{0x9CFE22E0F1E0F44DULL, "vxw_encounter2_camp_92_0_0.vlm"},
	{0x9CFEDFFB2ED22EC9ULL, "zz_dae_debug6142942.cl"},
	{0x9CFEEB6228FE09B0ULL, "6124148_m.fxe"},
	{0x9CFF4137CBE9376EULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_110a_sa.dds"},
	{0x9CFF6574F6E76C77ULL, "dae200cr_viscount6087225.cl"},
	{0x9CFFE3339190EBA6ULL, "vxw_encounter2_n4_0_0.vlm"},
	{0x9CFFFF200B4D6C0FULL, "vxht_0_hightown_a3_c6n_0_0.vlm"},
	{0x9D001584D32A2410ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockd.mao"},
	{0x9D001984D32A2B23ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockd.mmh"},
	{0x9D002784D32A437DULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockd.msh"},
	{0x9D0036C0440D2867ULL, "pro000cr_meeran_entourage.utc"},
	{0x9D004D7C638D0634ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_partition_1.mao"},
	{0x9D00ACB68C36ADA5ULL, "6108073_m.fxe"},
	{0x9D0144A2BF8CF415ULL, "hlac_w_sundermount_-65541_0x.dds"},
	{0x9D0184D115B4FC46ULL, "10513079.ogg"},
	{0x9D01D1644018CEADULL, "llac_lt_hangedman_-1_8.dds"},
	{0x9D01D787B68A0B7DULL, "vxht_0_hightown_a2_a85_0_0.vlm"},
	{0x9D01ED5D1A81C4AFULL, "hlac_lt_0_lowtown_n_65534_0z.dds"},
	{0x9D01F71EC04AB1CAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_tram.phy"},
	{0x9D024DA899DD586DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\prp_spruce2_n.dds"},
	{0x9D025F0853B6AB59ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax106a_0s.dds"},
	{0x9D027593CA362189ULL, "vxw_sundermount_33n_0_0.vlm"},
	{0x9D027B5AF177993AULL, "art\\effects\\cloudshadows\\w_sundermount\\cloud2.dds"},
	{0x9D029641B0F43308ULL, "6172411_m.fxe"},
	{0x9D02EEA8A78F0A81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_talk_and_exit.gad"},
	{0x9D036301DCC0C888ULL, "dae000_arishok6075323.cl"},
	{0x9D04005582A90AD6ULL, "6168919_m.fxe"},
	{0x9D0475AE304988B0ULL, "hlac_ht_chantry_2534_18z.dds"},
	{0x9D04FD38981F56EFULL, "221020736.ogg"},
	{0x9D053400A22D371EULL, "vxht_bartrand_n_67d_0_0.vlm"},
	{0x9D05CA17628E5DB5ULL, "6093841_m.fxe"},
	{0x9D076FEB5B78CD8EULL, "6179773_m.fxe"},
	{0x9D080F24E9E5E17DULL, "23815772.ogg"},
	{0x9D0885646B1CBA17ULL, "110598159.ogg"},
	{0x9D088A7A0F86BB38ULL, "vxw_encounter2_camp_rf_0_0.vlm"},
	{0x9D08A183D25D3337ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs102a_0s.dds"},
	{0x9D08F9BFE5E5C39DULL, "vxw_encounter3_2t_0_0.vlm"},
	{0x9D0928D0DB93E263ULL, "mcr370_templar26087648.cl"},
	{0x9D093893FA9C1968ULL, "lt_0_lowtown_docks_a3_250_0.mmh"},
	{0x9D094493FA9C2D0BULL, "lt_0_lowtown_docks_a3_250_0.mao"},
	{0x9D095693FA9C4B92ULL, "lt_0_lowtown_docks_a3_250_0.msh"},
	{0x9D0983845866834AULL, "631489999.ogg"},
	{0x9D09949B8887EEE6ULL, "116056243.ogg"},
	{0x9D09EFA8A7952B2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_talk_and_exit.ani"},
	{0x9D0A196460F5E7B7ULL, "lt_0_lowtown_a2_n_c_26_0.anb"},
	{0x9D0ACD170B5FDCEAULL, "6137216_m.fxe"},
	{0x9D0AF9419087183AULL, "dae000cr_commoner_m1.utc"},
	{0x9D0B10E2394CBE19ULL, "53007693.ogg"},
	{0x9D0B591F9B2D2215ULL, "hla_ht_0_hightown_a2_65536_0x.dds"},
	{0x9D0B8BB0C90725C4ULL, "6024100_f.fxe"},
	{0x9D0BCD75C127A189ULL, "character_character_p244.psh.fxo.dx11"},
	{0x9D0C94BDE3D70CCDULL, "vxw_encounter2_camp_41_0_0.vlm"},
	{0x9D0D0C4FBBBCE3D4ULL, "l_fen200ar_slaver_ambush2.bnk"},
	{0x9D0D1387633F1EA8ULL, "vxlt_0_lowtown_docks_a3_n_cuh_0_0.vlm"},
	{0x9D0D7565BBFE65C0ULL, "follower_banter6136740.cl"},
	{0x9D0D924FA1487B09ULL, "follower_banter6170002.cl"},
	{0x9D0D9F4E05BDFC67ULL, "gen00fl_sebastian6179940.cl"},
	{0x9D0DA19C1A907F80ULL, "19403067.ogg"},
	{0x9D0E5487752C22EDULL, "vxlt_hangedman_n_27f_0_0.vlm"},
	{0x9D0F8EFB1D5ECBEAULL, "6110974_m.fxe"},
	{0x9D103E296B34BC58ULL, "174124171.ogg"},
	{0x9D106F2910E7C0ECULL, "barr_smok.mao"},
	{0x9D109FEA94949C7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_sitting_table.anb"},
	{0x9D10E666AF7E95D6ULL, "49695188.ogg"},
	{0x9D111F4155BA8397ULL, "hlac_lt_0_lowtown_a2_-196608_3z.dds"},
	{0x9D112F70DB451A06ULL, "gen00fl_isabela6106285.cl"},
	{0x9D117C967517CE39ULL, "mag131_taraneh6088689.cl"},
	{0x9D11B083D611573EULL, "w_bonepitdestroyed_a3_184_0.phy"},
	{0x9D11FB238589F7B1ULL, "vxlt_gallowstemplar_n_5bo_0_0.vlm"},
	{0x9D121CEA55602EA7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9ff_minecartfram.phy"},
	{0x9D121D4C7BE40438ULL, "zz_isa_debug6133197.cl"},
	{0x9D12F71B26E4B9F7ULL, "vxlt_0_lowtown_a2_n_c_hvi_0_0.vlm"},
	{0x9D132C5D691C86D3ULL, "qun200_party_comment6090447.cl"},
	{0x9D13AA168FA44F95ULL, "6080266_f.fxe"},
	{0x9D13AC3142C52F71ULL, "6105904_m.fxe"},
	{0x9D13D5DAE4786235ULL, "6131171_f.fxe"},
	{0x9D13D756C6B6CF63ULL, "588823878.ogg"},
	{0x9D1433561BFEE2DDULL, "isa211_isabela6164478.cl"},
	{0x9D147936BBDF0BABULL, "vxlt_0_lowtown_a2_hc3_0_0.vlm"},
	{0x9D14962E2E86FEB7ULL, "6106284_m.fxe"},
	{0x9D14BB1C3510E536ULL, "6178323_m.fxe"},
	{0x9D15130513E21DFDULL, "hlac_lt_gallowstemplar_2384_0z.dds"},
	{0x9D153D94E6939D26ULL, "vxlt_undercitydungeon_1mu_0_0.vlm"},
	{0x9D1594B2B056A577ULL, "6085989_m.fxe"},
	{0x9D15A89B72CE0AD2ULL, "qun160_vashothlead6053953.cl"},
	{0x9D15E463AB82CF13ULL, "qun221_petrice6077322.cl"},
	{0x9D163B3BFDC5879EULL, "6118246_m.fxe"},
	{0x9D166B4BD56E9EF7ULL, "395580896.ogg"},
	{0x9D16834F680C4710ULL, "6139277_f.fxe"},
	{0x9D16A6F913B991D5ULL, "gen00fl_aveline6114008.cl"},
	{0x9D16FC57017CD3B8ULL, "vxw_encounter2_camp_mg_0_0.vlm"},
	{0x9D170CB7B7C4A390ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masd.phy"},
	{0x9D1713903B5AA597ULL, "168717481.ogg"},
	{0x9D1785CD84E84D62ULL, "6153294_m.fxe"},
	{0x9D17ADE90C1A2ABBULL, "hlac_w_encounter3_-65536_0x.dds"},
	{0x9D17C34E31DA5526ULL, "zz_area_signage7.utp"},
	{0x9D17C6E90B126A1DULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scl_b.mao"},
	{0x9D17E791F0477790ULL, "vxlt_undercityserialkiller_1ij_0_0.vlm"},
	{0x9D17FC2AD7BB965AULL, "168488001.ogg"},
	{0x9D185C555E592242ULL, "6153409_m.fxe"},
	{0x9D18DD88B1B6834FULL, "dae251ar_mount_caverns.are"},
	{0x9D1900F0E8A9B823ULL, "w_sundermount_118_0.msh"},
	{0x9D190EF0E8A9D07DULL, "w_sundermount_118_0.mmh"},
	{0x9D190F683A600E6EULL, "vxdr_0_deeproads_eun_0_0.vlm"},
	{0x9D1912F0E8A9D68EULL, "w_sundermount_118_0.mao"},
	{0x9D1938503E3EF74CULL, "368791932.ogg"},
	{0x9D198443B279EE11ULL, "sw_dust.mao"},
	{0x9D1A34C32A019892ULL, "vxlt_blackemporium_10h_0_0.vlm"},
	{0x9D1ABC4FD6DEDFB5ULL, "6079077_f.fxe"},
	{0x9D1B0402D6A48208ULL, "combat_smcc_r4_a1.gda"},
	{0x9D1BAC771FD1A753ULL, "art\\vfx\\textures\\vfx_d_noise.dds"},
	{0x9D1BCCD2F99B62F4ULL, "55324120.ogg"},
	{0x9D1BD5F0027856ECULL, "dae000_junar6147926.cl"},
	{0x9D1CB0836839AA05ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medh_0d.dds"},
	{0x9D1CF30C46DC0ABDULL, "729048359.ogg"},
	{0x9D1D9FE3A02F1555ULL, "hlac_lt_0_lowtown_a2_-3_5y.dds"},
	{0x9D1DCA2B1F6F560FULL, "zz_vault_debug6200361.cl"},
	{0x9D1E2C98A987532BULL, "art\\vfx\\textures\\vfx_shapes_cs.dds"},
	{0x9D1E72399CD1DAEDULL, "ss_zevran.bnk"},
	{0x9D1ED1401E526C67ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopening.gad"},
	{0x9D1F178C6FC103CBULL, "581146837.ogg"},
	{0x9D1F84696E1B8F2AULL, "vxca_caverns2prc_9ki_0_0.vlm"},
	{0x9D1F9A497C05DF1FULL, "vxw_encounter2_re_0_0.vlm"},
	{0x9D2095A1F9220241ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_001a_0d.dds"},
	{0x9D20D1950C165E4DULL, "lt_0_lowtown_41_0.rml"},
	{0x9D211DD89B5F7FB9ULL, "crt_cycle7.dds"},
	{0x9D21640CCCD508FBULL, "lt_0_lowtown_docks_n_192_0.msh"},
	{0x9D2168F818DAD277ULL, "vxlt_0_lowtown_a2_n_c_hqh_0_0.vlm"},
	{0x9D21760CCCD52746ULL, "lt_0_lowtown_docks_n_192_0.mao"},
	{0x9D21820CCCD53BE5ULL, "lt_0_lowtown_docks_n_192_0.mmh"},
	{0x9D21BE4BC194FF74ULL, "vxlt_0_lowtown_docks_n_bqz_0_0.vlm"},
	{0x9D21CA2775B63CC8ULL, "6075291_m.fxe"},
	{0x9D22191054D2F3ABULL, "gen00fl_anders6119284.cl"},
	{0x9D227C1E12559194ULL, "6149581_m.fxe"},
	{0x9D22E1A194D126EDULL, "gen00fl_merrill6090262.cl"},
	{0x9D22E616C23E6854ULL, "lgt110_arl_bann_teagan6115593.cl"},
	{0x9D22EB13ADD5242BULL, "vxlt_0_lowtown_a2_hh2_0_0.vlm"},
	{0x9D231D21CCAE8AACULL, "536270007.ogg"},
	{0x9D23945E21090254ULL, "204128906.ogg"},
	{0x9D239DB6EEF0DAA8ULL, "llac_ht_keep_n_3732_10.dds"},
	{0x9D23A7066BE58208ULL, "vxht_playershouse_n_1pz_0_0.vlm"},
	{0x9D23C4CBFFEE819FULL, "ave310_donnic6134715.cl"},
	{0x9D23D0B122E30E4DULL, "vxca_caverns1_j60_0_0.vlm"},
	{0x9D243E813E0B911DULL, "vxlt_0_lowtown_a2_eqv_0_0.vlm"},
	{0x9D2479F9C0C050C6ULL, "hm_lgt_avern_mess.mor"},
	{0x9D250172A0E74027ULL, "llac_lt_0_lowtown_docks_-3_5.dds"},
	{0x9D256E33F372EC38ULL, "vxw_encounter2_camp_hd_0_0.vlm"},
	{0x9D25B72BDE85EFD1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11b.mmh"},
	{0x9D25D12BDE861B8FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos11b.msh"},
	{0x9D25EA7E60CFDB3CULL, "6048224_m.fxe"},
	{0x9D25F7AA76DE65E2ULL, "6156194_m.fxe"},
	{0x9D26596EE23D9010ULL, "vxlt_undercityserialkiller_1nk_0_0.vlm"},
	{0x9D26950E814871B5ULL, "943096545.ogg"},
	{0x9D269EE5D1082883ULL, "6155727_m.fxe"},
	{0x9D26EF3DF7783C85ULL, "508532194.ogg"},
	{0x9D27376D22BFB856ULL, "llac_ht_0_hightown_n_131072_0.dds"},
	{0x9D275E3245D77C48ULL, "cod_hst_wilds.utp"},
	{0x9D27CB986B9ACF6EULL, "genlock_alpha_dual_axes.utc"},
	{0x9D27F71C5C1BD43EULL, "gen00fl_aveline6137697.cl"},
	{0x9D2842668C94BCB1ULL, "6125300_m.fxe"},
	{0x9D28436D4C47F37DULL, "vxlt_gallowstemplar_n_3ju_0_0.vlm"},
	{0x9D2876CF852502E8ULL, "6085844_f.fxe"},
	{0x9D2887D82E9C5460ULL, "w_sundermount_28_0_ll.mmh"},
	{0x9D293E46CC92637EULL, "vxht_0_hightown_n_aox_0_0.vlm"},
	{0x9D2943C7D9D5C6FAULL, "hlac_lt_0_lowtown_docks_a3_n_-4_7x.dds"},
	{0x9D297FE68FBA0E02ULL, "6140613_f.fxe"},
	{0x9D29BF9980FDDA92ULL, "vxw_encounter3_5z_0_0.vlm"},
	{0x9D29C235DF59E5ADULL, "6054376_m.fxe"},
	{0x9D29CAC7D81D9B6DULL, "vxht_bartrand_n_4dl_0_0.vlm"},
	{0x9D29D688B1C4EA0CULL, "dae251ar_mount_caverns.lst"},
	{0x9D2A02D6776DB58EULL, "art\\characters\\weapons\\dagger\\w_dger_101a.phy"},
	{0x9D2A208163325AE3ULL, "764747626.ogg"},
	{0x9D2A3C0FCA5CB6BAULL, "143200091.ogg"},
	{0x9D2B12075F311271ULL, "601325789.ogg"},
	{0x9D2B2D2D883835B5ULL, "ser211ip_dead_alessa.ncs"},
	{0x9D2B7C3665E6D51AULL, "w_bonepitmining_a2_68_0.mmh"},
	{0x9D2B803665E6DC51ULL, "w_bonepitmining_a2_68_0.mao"},
	{0x9D2B8E3665E6F460ULL, "w_bonepitmining_a2_68_0.msh"},
	{0x9D2BA0254BCC62A1ULL, "243160247.ogg"},
	{0x9D2C1687420B18E6ULL, "6015000_m.fxe"},
	{0x9D2C4BA7141B74D9ULL, "vxw_bonepit_1nx_0_0.vlm"},
	{0x9D2C53089A8CCF70ULL, "694078738.ogg"},
	{0x9D2C5502EF659C1DULL, "195514112.ogg"},
	{0x9D2C61CD0D8459D7ULL, "t1_skn_e02.tnt"},
	{0x9D2C8F48AE662658ULL, "vxlt_warehouse_n_p2_0_0.vlm"},
	{0x9D2CB670A9143854ULL, "art\\cinematics\\follower plots\\sebastian\\cs_cleric_sebastianintro.gad"},
	{0x9D2D044A9E93ABB1ULL, "6104622_m.fxe"},
	{0x9D2D1B3A46548E89ULL, "gen00fl_aveline6116631.cl"},
	{0x9D2E5C26E16A76FFULL, "45617833.ogg"},
	{0x9D2E66D313544894ULL, "vxdr_primevaltomb_90y_0_0.vlm"},
	{0x9D2E89114F969F03ULL, "rdr111dg_martin6055058.cl"},
	{0x9D2F15C447D75C12ULL, "ser270_templar6116090.cl"},
	{0x9D2F19F039853EF9ULL, "vxht_0_hightown_a3_c37_0_0.vlm"},
	{0x9D2FD7418626D896ULL, "dr_0_deeproads_5841.rml"},
	{0x9D2FF9D2BB0BEBFDULL, "zz_dae_debug6029124.cl"},
	{0x9D302005B5CD8141ULL, "vxht_0_hightown_a2_c54_0_0.vlm"},
	{0x9D30C7A3BC493850ULL, "art\\characters\\weapons\\spears\\w_spr_003a.mao"},
	{0x9D30CBA3BC493F63ULL, "art\\characters\\weapons\\spears\\w_spr_003a.mmh"},
	{0x9D30CD032CE41036ULL, "6126988_m.fxe"},
	{0x9D30D9A3BC4956BDULL, "art\\characters\\weapons\\spears\\w_spr_003a.msh"},
	{0x9D30E13A9479B910ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgta.msh"},
	{0x9D30E9E934204528ULL, "dae220ip_at_lowtown.utp"},
	{0x9D30EF3A9479D12AULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgta.mmh"},
	{0x9D30F33A9479D861ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_lgta.mao"},
	{0x9D311086EDE28718ULL, "124397284.ogg"},
	{0x9D311141DD520F0FULL, "vxdr_0_deeproads_dvw_0_0.vlm"},
	{0x9D3114FF2FB507C2ULL, "gen_im_cns_flu_sebastian_03.uti"},
	{0x9D32313732FBFFECULL, "vxht_0_hightown_n_c30_0_0.vlm"},
	{0x9D32429D16F84BA5ULL, "6154021_m.fxe"},
	{0x9D326BBD74005D36ULL, "dae000_tranquil_16165115.cl"},
	{0x9D32BE401E62F4ADULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl1_pcopening.ani"},
	{0x9D33AABD91F5EE60ULL, "160312830.ogg"},
	{0x9D33C570A91A7032ULL, "art\\cinematics\\follower plots\\sebastian\\cs_cleric_sebastianintro.ani"},
	{0x9D33E010E56904B8ULL, "vxw_encounter2_camp_ce_0_0.vlm"},
	{0x9D33E037CD844231ULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.anb"},
	{0x9D34B993C8465148ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_4z.dds"},
	{0x9D34CFA40BA187B4ULL, "634454169.ogg"},
	{0x9D350ECDCB85E95BULL, "hlac_dr_primevaltomb_9621_6y.dds"},
	{0x9D35455D9D1C3E5BULL, "321227266.ogg"},
	{0x9D3551657C37D8CDULL, "206228571.ogg"},
	{0x9D35697C810CF2A2ULL, "hlac_lt_gallowstemplar_f_2390_1x.dds"},
	{0x9D356E19DD388CB6ULL, "vxht_bartrand_n_649_0_0.vlm"},
	{0x9D35C19E917C9955ULL, "llac_w_encounter3_-393215_0.dds"},
	{0x9D360F5F28855FC7ULL, "one140cr_charade6128508.cl"},
	{0x9D361664611B1DCBULL, "lt_0_lowtown_a2_n_c_26_0.rml"},
	{0x9D362A8277996B65ULL, "828653922.ogg"},
	{0x9D366DA9C88E31D9ULL, "215576528.ogg"},
	{0x9D3709940155E8C2ULL, "701354897.ogg"},
	{0x9D375E75BF8FF646ULL, "llac_lt_0_lowtown_n_65536_5.dds"},
	{0x9D37724096BA9744ULL, "34868294.ogg"},
	{0x9D37A63F63D73D84ULL, "gen00fl_anders6125787.cl"},
	{0x9D383F3F7785678FULL, "llac_lt_gamlenshouse_n_-1_3.dds"},
	{0x9D38541C4DBABE2EULL, "vxht_0_hightown_c5s_0_0.vlm"},
	{0x9D38584F8301E209ULL, "6132893_f.fxe"},
	{0x9D38B73042155054ULL, "112276821.ogg"},
	{0x9D38FDF11CC9B6ADULL, "mer320_hubert6089688.cl"},
	{0x9D38FF881F98B9FEULL, "vxlt_0_lowtown_docks_col_0_0.vlm"},
	{0x9D392E9FCACFBB5AULL, "fxe_caustic_diff.mao"},
	{0x9D395111827607D7ULL, "704015855.ogg"},
	{0x9D39BE9AE3DC35A2ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_imp_p.mmh"},
	{0x9D3A008E26F93C18ULL, "36937079.ogg"},
	{0x9D3AED217BB1AE1DULL, "sse_ss_circlemages_m_1.gda"},
	{0x9D3B0125A05C3ED8ULL, "vxlt_warehouse_n_w3_0_0.vlm"},
	{0x9D3B60CCFA1160CBULL, "guardian_dmg_resist.dds"},
	{0x9D3BE64A1371DE70ULL, "6171380_m.fxe"},
	{0x9D3C6047F896CAA1ULL, "gen_im_trs_slv_junk2_01.uti"},
	{0x9D3C7CC2B478A36AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxa.mao"},
	{0x9D3C80C2B478A9B9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxa.mmh"},
	{0x9D3C8AC2B478BAA7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxa.msh"},
	{0x9D3C9A6DEFBD9453ULL, "224301416.ogg"},
	{0x9D3CBE0B2DECA2A7ULL, "173792020.ogg"},
	{0x9D3D8BCD2B7B5779ULL, "vxht_0_hightown_a3_c46_0_0.vlm"},
	{0x9D3DF2A2B0FFFCC4ULL, "lt_gallowscourtyard_n_51_0.rml"},
	{0x9D3DFA3D29AB06CBULL, "6179788_f.fxe"},
	{0x9D3E62AAA7918852ULL, "922711949.ogg"},
	{0x9D3E80B2266E293CULL, "17780810.ogg"},
	{0x9D3F174C2E3C69C2ULL, "qun231cr_viscount_guard4.utc"},
	{0x9D3F41EDF26D2F6FULL, "6133020_m.fxe"},
	{0x9D40033D3D9FBC27ULL, "137047319.ogg"},
	{0x9D401E438AF4026EULL, "gen00fl_isabela_init6136104.cl"},
	{0x9D402352B8E3362AULL, "vxht_keep_5u1_0_0.vlm"},
	{0x9D405B6343006D9CULL, "6111055_f.fxe"},
	{0x9D405FE42248A39AULL, "mag120pt_enemies.ncs"},
	{0x9D40A31424F2186CULL, "vxht_0_hightown_n_c81_0_0.vlm"},
	{0x9D410137CD8F04AAULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.mmh"},
	{0x9D417EE1970C829CULL, "hlac_lt_0_lowtown_n_131071_7z.dds"},
	{0x9D41816D1823B536ULL, "qun211_braggart6075443.cl"},
	{0x9D4194D1B18A42C7ULL, "493450910.ogg"},
	{0x9D41CCA3BC57B388ULL, "art\\characters\\weapons\\spears\\w_spr_003a.phy"},
	{0x9D41EB91F65CF7D5ULL, "6174504_m.fxe"},
	{0x9D425F9EAC061747ULL, "vxlt_gamlenshouse_n_b9_0_0.vlm"},
	{0x9D427341D03E8232ULL, "793389689.ogg"},
	{0x9D42CC8A13A1ECABULL, "vxht_0_hightown_a2_b26_0_0.vlm"},
	{0x9D42D94321D6CAE0ULL, "art\\vfx\\materials\\chunky\\vfx_c_branch.mao"},
	{0x9D42FF0992699939ULL, "vxht_mansion2b_2kc_0_0.vlm"},
	{0x9D435769707D0EF8ULL, "vxlt_0_lowtown_n_e7z_0_0.vlm"},
	{0x9D43BB006E1A3C23ULL, "673096800.ogg"},
	{0x9D43E2E89285AF50ULL, "dae000ip_crowd_lm1_talk12.utp"},
	{0x9D441EB09A0CE180ULL, "6059150_m.fxe"},
	{0x9D442163993BD295ULL, "vxlt_gallowstemplar_f_6sw_0_0.vlm"},
	{0x9D45045AD923A812ULL, "vxlt_0_lowtown_a2_erx_0_0.vlm"},
	{0x9D455E15D3C20E34ULL, "gen00fl_anders6125897.cl"},
	{0x9D456D205F8218DEULL, "art\\characters\\weapons\\unique\\aveline\\textures\\w_shd_009a_0n.dds"},
	{0x9D45897A61CDD7C9ULL, "271623396.ogg"},
	{0x9D45998F92D0DC62ULL, "w_sundermount_36_0.msh"},
	{0x9D459B8F92D0DF18ULL, "w_sundermount_36_0.mmh"},
	{0x9D45A78F92D0F3BBULL, "w_sundermount_36_0.mao"},
	{0x9D45C6BB8A8FE0FBULL, "art\\characters\\playercharacter\\humanfemale\\hf_glv_lgte.phy"},
	{0x9D46612C19F705D7ULL, "vxlt_0_lowtown_docks_n_c5l_0_0.vlm"},
	{0x9D4680D901CB01EBULL, "6195837_m.fxe"},
	{0x9D46EB07662CA350ULL, "vxlt_0_lowtown_docks_n_b8e_0_0.vlm"},
	{0x9D46FA7DA72E9CAAULL, "hlac_lt_0_lowtown_a2_n_c_-65537_6y.dds"},
	{0x9D4761AECC14B904ULL, "6088866_m.fxe"},
	{0x9D4784794814DB0BULL, "6152137_m.fxe"},
	{0x9D47B71732C3033DULL, "744144791.ogg"},
	{0x9D47F81969C2DCD9ULL, "dae150_radha6148946.cl"},
	{0x9D489B8F92D3152BULL, "w_sundermount_36_0.rml"},
	{0x9D48DFEC9A9C4BDCULL, "vxw_woundedcoast_166_0_0.vlm"},
	{0x9D4916930B39B5C3ULL, "815320631.ogg"},
	{0x9D49730292525758ULL, "vxlt_warehouse_n_z0_0_0.vlm"},
	{0x9D4A348FEFD53599ULL, "6189510_f.fxe"},
	{0x9D4A555E525E1A36ULL, "hlac_lt_0_lowtown_docks_a3_-131075_6y.dds"},
	{0x9D4A5866A94A973AULL, "455473316.ogg"},
	{0x9D4A981A89B95C8EULL, "vxlt_blackemporium_5z_0_0.vlm"},
	{0x9D4B7C231ACAD346ULL, "vxht_brothel_n_5l2_0_0.vlm"},
	{0x9D4BB2BF7F303CB6ULL, "hlac_lt_undercityencounter_-1_1y.dds"},
	{0x9D4BFAA397E60598ULL, "vxlt_hangedman_n_1vo_0_0.vlm"},
	{0x9D4C563E19A112AAULL, "468480471.ogg"},
	{0x9D4CAA4F21FAEEBBULL, "6094473_m.fxe"},
	{0x9D4CBF2265AFA533ULL, "hlac_lt_0_lowtown_docks_n_-262144_2z.dds"},
	{0x9D4D09B0D4441106ULL, "6101796_m.fxe"},
	{0x9D4D264FBBF2F386ULL, "l_fen200ar_slaver_ambush2.wbd"},
	{0x9D4D975891C17709ULL, "376479016.ogg"},
	{0x9D4DB2C300AA50A4ULL, "vxht_brothel_n_60o_0_0.vlm"},
	{0x9D4DD546EB27D139ULL, "2692305.ogg"},
	{0x9D4DE8F40659CA1EULL, "455790817.ogg"},
	{0x9D4E60AE99B918E7ULL, "645221630.ogg"},
	{0x9D4E8A729A94EF62ULL, "vxlt_warehouse_13u_0_0.vlm"},
	{0x9D4ECB657ABF55E0ULL, "dae000_bodhan6129160.cl"},
	{0x9D4EE08B447D52BEULL, "art\\levels\\4cinematics\\ferelden\\_textures\\letterfront_d.dds"},
	{0x9D4EE3A47ADDD611ULL, "lt_0_lowtown_docks_a3_213_0.msh"},
	{0x9D4EF1A47ADDEE2CULL, "lt_0_lowtown_docks_a3_213_0.mao"},
	{0x9D4EFDA47ADE02CFULL, "lt_0_lowtown_docks_a3_213_0.mmh"},
	{0x9D4F295B203E8928ULL, "6093804_m.fxe"},
	{0x9D4F413ABBDD4637ULL, "163329872.ogg"},
	{0x9D4F4B943A69DAD8ULL, "6092372_f.fxe"},
	{0x9D4F9F2CFDBA95FEULL, "452843734.ogg"},
	{0x9D4FA28F92D93FA3ULL, "w_sundermount_36_0.phy"},
	{0x9D50440E3EAF1FB6ULL, "art\\characters\\weapons\\unique\\arishok\\textures\\w_gsw_002a_n.dds"},
	{0x9D50453FDC9112E5ULL, "dae000_mother6099209.cl"},
	{0x9D50F57E6B84ACC3ULL, "97646539.ogg"},
	{0x9D518080C0718D6AULL, "gen00fl_sebastian6179925.cl"},
	{0x9D51B94A1574F5EAULL, "901185671.ogg"},
	{0x9D51BEE51EAC8BA8ULL, "gen00fl_anders6183826.cl"},
	{0x9D5208B1A90AFB77ULL, "764086291.ogg"},
	{0x9D520E8FA7EF484AULL, "hlac_ht_mansion2b_818_6z.dds"},
	{0x9D52BBA395AA9082ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.open.gad"},
	{0x9D52D4F32A94CF05ULL, "dae100_gamlen_letter6146977.cl"},
	{0x9D532D0F9D5F3BD9ULL, "plt_cod_ltr_sk3_circlelette.plo"},
	{0x9D537637CB19C092ULL, "vxlt_0_lowtown_a2_euy_0_0.vlm"},
	{0x9D53A7FF503A1960ULL, "isa211_isabela6202212.cl"},
	{0x9D53C6C1B1C22A2FULL, "vxlt_0_lowtown_docks_n_ctk_0_0.vlm"},
	{0x9D53DB001E2709ECULL, "610376319.ogg"},
	{0x9D53F37D5448DDE7ULL, "vxht_keep_64d_0_0.vlm"},
	{0x9D5467881A665487ULL, "vxht_mansion2_1a4_0_0.vlm"},
	{0x9D549A99CCBD6B0DULL, "767644885.ogg"},
	{0x9D54A5C5ABAD2AB3ULL, "398663194.ogg"},
	{0x9D550545049F120FULL, "ach_secret.dds"},
	{0x9D550BA1497E72CCULL, "vxht_mansion2b_27n_0_0.vlm"},
	{0x9D55205EADFFE4EDULL, "vxht_bartrand_n_4si_0_0.vlm"},
	{0x9D555CE45822BBD0ULL, "vxlt_0_lowtown_docks_n_b5b_0_0.vlm"},
	{0x9D559A291C80C61CULL, "6119113_m.fxe"},
	{0x9D55D18CB3094DDAULL, "vxlt_0_lowtown_docks_n_b7s_0_1.vlm"},
	{0x9D55D38BBE0AD184ULL, "6083966_m.fxe"},
	{0x9D55E3420384EAFEULL, "vxlt_0_lowtown_docks_cun_0_0.vlm"},
	{0x9D55F66A57894EB6ULL, "1043061412.ogg"},
	{0x9D561FA719899DB8ULL, "6170079_m.fxe"},
	{0x9D5751C98C92645CULL, "vxw_woundedcoast_117_0_0.vlm"},
	{0x9D5799E10A655484ULL, "6126854_m.fxe"},
	{0x9D57B7E29BB5733EULL, "lowtown_argument6169435.cl"},
	{0x9D57BC3B0171637BULL, "docks_male_human_worker_sta6175949.cl"},
	{0x9D57D328CCBA1F74ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzcot_0.pwk"},
	{0x9D580EF458E96299ULL, "vxlt_darktown_4ix_0_0.vlm"},
	{0x9D581412A36444FBULL, "art\\levels\\races\\natural\\blightlands\\vista\\_textures\\bridge_d.dds"},
	{0x9D58491304767E0FULL, "vxht_keep_4ay_0_0.vlm"},
	{0x9D584F6AA340B61DULL, "vxht_0_hightown_n_b9z_0_0.vlm"},
	{0x9D591056E5F002F8ULL, "hlac_w_bonepit_393216_0z.dds"},
	{0x9D59206FC7498A14ULL, "vxlt_undercitydungeon_221_0_0.vlm"},
	{0x9D5A00892F18F225ULL, "character_character_p262.psh.fxo.dx11"},
	{0x9D5A00F84582448AULL, "dm_lgt_car_loy.mor"},
	{0x9D5A344366C38108ULL, "753591785.ogg"},
	{0x9D5A85840B908C16ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_rtl.anb"},
	{0x9D5A8CBB78D13BBEULL, "vxht_chantry_n_25y_0_0.vlm"},
	{0x9D5AC2A9EB502BF3ULL, "6171901_m.fxe"},
	{0x9D5BDB79C855117CULL, "lt_0_lowtown_docks_18_0.rml"},
	{0x9D5C1489D4A794DBULL, "lowtown_fem_prostitute_sta6161601.cl"},
	{0x9D5C9FF2EF0B6CE3ULL, "llac_ht_0_hightown_65533_0.dds"},
	{0x9D5CDF1C1455DE38ULL, "6110244_m.fxe"},
	{0x9D5DEF651A20EB47ULL, "hlac_lt_warehouse_0_9y.dds"},
	{0x9D5E1EAFF367AE4EULL, "vxlt_0_lowtown_a2_n_deg_0_0.vlm"},
	{0x9D5E6DF03E69D5AFULL, "zz_mag_debug6054611.cl"},
	{0x9D5EDA2835500D83ULL, "w_bonepit_55_0.mao"},
	{0x9D5EDE28355014D0ULL, "w_bonepit_55_0.mmh"},
	{0x9D5EE74E420F70A2ULL, "primitives_primitive_p2.psh.fxo"},
	{0x9D5EE7C7B03598C0ULL, "gen00fl_sebastian6174029.cl"},
	{0x9D5EEC2835502CEAULL, "w_bonepit_55_0.msh"},
	{0x9D5F35DA4073899EULL, "fx_dis_blank_m.mao"},
	{0x9D5F6CC6FA6FBECBULL, "966150413.ogg"},
	{0x9D5F807D500588F9ULL, "6083448_m.fxe"},
	{0x9D5F9DE068DDE5B9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_kirkwallgates_sa.dds"},
	{0x9D5FBFFD98EA6D3CULL, "grass.dds"},
	{0x9D5FE2C37655CA39ULL, "vxht_mansion2b_2aa_0_0.vlm"},
	{0x9D5FE8B76DB0A366ULL, "gen_im_cns_flu_anders_02.uti"},
	{0x9D5FEE28AFD971EBULL, "gen00fl_dog6167239.cl"},
	{0x9D6020A2CCFF3314ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_left.phy"},
	{0x9D60D9F487577C31ULL, "follower_banter6155784.cl"},
	{0x9D61026A7DF91280ULL, "6053350_m.fxe"},
	{0x9D622A85D49FECCAULL, "follower_banter6173051.cl"},
	{0x9D62389EA3B842AFULL, "vxlt_0_lowtown_docks_n_coj_0_0.vlm"},
	{0x9D6254D8817940EBULL, "zz_betpt_debug.ncs"},
	{0x9D626359B2E9D3F5ULL, "zz_ave_debug6064127.cl"},
	{0x9D62655A463EF667ULL, "vxht_keep_61g_0_0.vlm"},
	{0x9D629C10698EEF37ULL, "vxca_caverns1_mines_e7o_0_0.vlm"},
	{0x9D62BD9AE3FF3949ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_imp_p.anb"},
	{0x9D637D7E3B748B4CULL, "vxht_mansion2b_20o_0_0.vlm"},
	{0x9D63D0E88FD9CC5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_loop.ani"},
	{0x9D63D3434CD8E3A2ULL, "6018596_f.fxe"},
	{0x9D63F5833F9988E9ULL, "vxlt_undercityserialkiller_1lp_0_0.vlm"},
	{0x9D643CDA36B20B80ULL, "524981331.ogg"},
	{0x9D644568B000EA04ULL, "6086666_m.fxe"},
	{0x9D647C08375B8FFDULL, "vxlt_undercityencounter_y4_0_0.vlm"},
	{0x9D64C7ADE38397CEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_alcove.msh"},
	{0x9D64D5ADE383AF1FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_alcove.mao"},
	{0x9D64E1ADE383C3BCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_alcove.mmh"},
	{0x9D64F2E8108C0441ULL, "vxw_bonepit_1pb_0_0.vlm"},
	{0x9D650E8E611BF19FULL, "vxht_brothel_n_6cu_0_0.vlm"},
	{0x9D65539A598FCBA6ULL, "vxw_blightlands_70_0_0.vlm"},
	{0x9D65E56FA3BCC233ULL, "follower_banter6141758.cl"},
	{0x9D65E810DEF5E0BBULL, "ht_0_hightown_a2_50_0.rml"},
	{0x9D663273DB545D79ULL, "mrl150_merrill6202031.cl"},
	{0x9D67170F6B5EF755ULL, "dae000_elegant.bnk"},
	{0x9D678925ABF0D32EULL, "hlac_lt_0_lowtown_a3_n_c_-196608_4z.dds"},
	{0x9D67B8B70E72B8D1ULL, "pro000cr_refugee6.utc"},
	{0x9D67E6A395BD3668ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.open.ani"},
	{0x9D6833BCF0AD2517ULL, "and110cr_angry_refugee.utc"},
	{0x9D68464D6EE7EA42ULL, "495610946.ogg"},
	{0x9D6854DB8112AA4EULL, "zz_isa_debug6134420.cl"},
	{0x9D68784974F2309AULL, "478996722.ogg"},
	{0x9D68B3FBD8AF1664ULL, "6067821_m.fxe"},
	{0x9D68E2E0331D5304ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.ani"},
	{0x9D69217178177CA7ULL, "333057455.ogg"},
	{0x9D69382A41C62CE6ULL, "vxca_caverns1_gkl_0_1.vlm"},
	{0x9D695B7B3FBC9A19ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keep_1.mao"},
	{0x9D69FDA2CD074C11ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_left.msh"},
	{0x9D6A0BA2CD07642CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_left.mao"},
	{0x9D6A15254EBF96F5ULL, "llac_lt_0_lowtown_docks_a3_n_-2_0.dds"},
	{0x9D6A17A2CD0778CFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_left.mmh"},
	{0x9D6A2FE06D4E1B1AULL, "419104300.ogg"},
	{0x9D6A390F6B6166ADULL, "dae000_elegant.cnv"},
	{0x9D6AE2B9A3D7CF40ULL, "6161525_m.fxe"},
	{0x9D6B01D677A54B51ULL, "art\\characters\\weapons\\dagger\\w_dger_101a.mmh"},
	{0x9D6B0ACE8099FCF4ULL, "116147680.ogg"},
	{0x9D6B1BD677A5770FULL, "art\\characters\\weapons\\dagger\\w_dger_101a.msh"},
	{0x9D6B3A7437E89DD6ULL, "dae000_bartrand6030677.cl"},
	{0x9D6B49FC42164D01ULL, "vxw_woundedcoast_16g_0_0.vlm"},
	{0x9D6B80D991A61969ULL, "6113288_m.fxe"},
	{0x9D6B842B63C20085ULL, "hlac_ca_caverns1_24425_1x.dds"},
	{0x9D6BF485EE5EC82DULL, "vxlt_0_lowtown_docks_c0b_0_0.vlm"},
	{0x9D6BF835300AB087ULL, "813788020.ogg"},
	{0x9D6C5E0A9541DA7AULL, "429744813.ogg"},
	{0x9D6C77D1F7DA0B9DULL, "vxlt_0_lowtown_a2_ehq_0_0.vlm"},
	{0x9D6C8B48CD0186CFULL, "vxlt_gallowscourtyard_n_7i4_0_0.vlm"},
	{0x9D6D64E4B3676DD3ULL, "lt_0_lowtown_a2_n_c_6_0.phy"},
	{0x9D6DBD7BBE352851ULL, "6148455_m.fxe"},
	{0x9D6E6640B9EB1AFFULL, "mcr371cr_templar_lieutenant.utc"},
	{0x9D6EA65BC772D4B1ULL, "6059426_f.fxe"},
	{0x9D6F0A9EEC6CD08BULL, "6141784_m.fxe"},
	{0x9D6F1E082CD98757ULL, "6096422_m.fxe"},
	{0x9D6F313722982129ULL, "739162423.ogg"},
	{0x9D6F9CDA1CB8ED9BULL, "224367157.ogg"},
	{0x9D6FDFE033236D66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.gad"},
	{0x9D6FFB2B8FF575CCULL, "llac_lt_0_lowtown_a2_n_c_-3_0.dds"},
	{0x9D701B47C1A4816CULL, "art\\vfx\\materials\\vfx_arc_cl.mao"},
	{0x9D701CCB349E1851ULL, "544014768.ogg"},
	{0x9D709F555E8F8F0AULL, "vxdr_0_deeproads_evs_0_0.vlm"},
	{0x9D70BBB811476774ULL, "lt_backalley_n.rml"},
	{0x9D71ADBB8AB4F09AULL, "art\\characters\\playercharacter\\humanfemale\\hf_glv_lgte.msh"},
	{0x9D71BFBB8AB50FE0ULL, "art\\characters\\playercharacter\\humanfemale\\hf_glv_lgte.mmh"},
	{0x9D733847AF43CD79ULL, "6059911_m.fxe"},
	{0x9D73615BAFFEE3ECULL, "vxw_woundedcoast_ar_0_0.vlm"},
	{0x9D73C5774B85E426ULL, "vxw_blightlands_81_0_0.vlm"},
	{0x9D743D0A67AC68F5ULL, "6048121_f.fxe"},
	{0x9D74A8256C768496ULL, "6040550_m.fxe"},
	{0x9D74C3DAA40AF4C2ULL, "108298022.ogg"},
	{0x9D75280F9D7C19EEULL, "plt_cod_ltr_sk3_circlelette.nss"},
	{0x9D759EEA7A681EE4ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_4z.dds"},
	{0x9D75DAE03327D045ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.evt"},
	{0x9D76009031A2C326ULL, "framebuffereffects_diffusiondof_p0.psh.fxo.dx11"},
	{0x9D76551C95C712D1ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_001a_0t.dds"},
	{0x9D76932EAD304916ULL, "hlac_lt_0_lowtown_docks_a3_-196610_5z.dds"},
	{0x9D76B29AA2E83D65ULL, "gen_im_trs_cot_blt_r0.uti"},
	{0x9D771FE224626250ULL, "vxca_caverns2_965_0_0.vlm"},
	{0x9D77A492C8A77A8AULL, "llac_lt_warehouse_n_1_5.dds"},
	{0x9D77E731A329FDBDULL, "598381672.ogg"},
	{0x9D7834FA23C03080ULL, "hlac_lt_hangedman_n_-1_28z.dds"},
	{0x9D78F7E88FEC6AE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_loop.gad"},
	{0x9D79A67DBF6FD580ULL, "318738046.ogg"},
	{0x9D79BBD9340C6581ULL, "vxw_woundedcoast_11f_0_0.vlm"},
	{0x9D79D6A1B09CBAA0ULL, "gen00fl_fenris6106618.cl"},
	{0x9D79E244477559BDULL, "llac_ca_caverns1_24422_6.dds"},
	{0x9D7A4394DAC8FBABULL, "ht_chantry_n_mmi.gff"},
	{0x9D7ACC3637D32D2BULL, "cs_pro000cs_innblood_carv.wbd"},
	{0x9D7B8A285A9B7043ULL, "art\\vfx\\textures\\vfx_lightning_single.dds"},
	{0x9D7BB43C7AE2711BULL, "6188563_m.fxe"},
	{0x9D7C0362DEB62445ULL, "699402414.ogg"},
	{0x9D7C491350A8BEDDULL, "dae000_elegant6166887.cl"},
	{0x9D7C55D888C2F27CULL, "6090863_m.fxe"},
	{0x9D7C5905490CC91FULL, "clf260_harley6091413.cl"},
	{0x9D7D1A9C9EB7BAADULL, "isa321_velasco6117689.cl"},
	{0x9D7D1F2542855FB2ULL, "mrl200im_statue_gift.uti"},
	{0x9D7E74BDEFBE6882ULL, "vxlt_0_lowtown_a2_n_go2_0_0.vlm"},
	{0x9D7E980E5BCD5A87ULL, "vxlt_gallowscourtyard_n_43m_0_0.vlm"},
	{0x9D7FF8A7076A560FULL, "mcr371_hostile_fenris6082223.cl"},
	{0x9D8093E3C6B17FC5ULL, "llac_lt_foundry_n_320_8.dds"},
	{0x9D809B5D9636B78EULL, "one280_golem6115969.cl"},
	{0x9D80DEBCD2210760ULL, "698459692.ogg"},
	{0x9D80F0E9DD2A4185ULL, "art\\vfx\\environment\\vfx_spike_holes_p.mmh"},
	{0x9D80F9C35F0DCEBDULL, "vrc321_varric6087081.cl"},
	{0x9D811DFED2777842ULL, "450240511.ogg"},
	{0x9D819CCAC3772D95ULL, "6206162_m.fxe"},
	{0x9D81AA24A139E5F9ULL, "6056811_m.fxe"},
	{0x9D81C1E63E5C84C9ULL, "6076516_m.fxe"},
	{0x9D81F2484508229FULL, "vxht_brothel_n_6iw_0_0.vlm"},
	{0x9D81FA21C3C6A1CCULL, "llac_lt_0_lowtown_n_-131072_1.dds"},
	{0x9D826287B9E7A7DCULL, "6016779_m.fxe"},
	{0x9D8290B0E7717175ULL, "281644079.ogg"},
	{0x9D82ED12E6D655CEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmla_1.mao"},
	{0x9D8317D75F56EBA4ULL, "6147157_f.fxe"},
	{0x9D834AB6DF06CEDDULL, "575929756.ogg"},
	{0x9D85023847398B76ULL, "189317087.ogg"},
	{0x9D859ABA11A96551ULL, "hlac_ht_mansion_n_3371_0y.dds"},
	{0x9D85FF44722BFD25ULL, "vxw_woundedcoast_17x_0_0.vlm"},
	{0x9D86450585863937ULL, "lt_gallowstemplar_probe_2qy.dds"},
	{0x9D875DE3C23FF842ULL, "vxlt_undercityserialkiller_1sz_0_0.vlm"},
	{0x9D8787C74BA8746FULL, "537706890.ogg"},
	{0x9D87DC79C87A4DCCULL, "lt_0_lowtown_docks_18_0.anb"},
	{0x9D88AC057283373BULL, "vxw_woundedcoast_1gf_0_0.vlm"},
	{0x9D88C7C4A6CF51DEULL, "106558656.ogg"},
	{0x9D88E7DD9BEB4239ULL, "158613148.ogg"},
	{0x9D88F06923A6EF63ULL, "llac_dr_0_deeproads_5849_9.dds"},
	{0x9D89AB4C3580C642ULL, "113776151.ogg"},
	{0x9D89AE65A2372AB7ULL, "vxdr_dungeon_1lx_0_0.vlm"},
	{0x9D89C0ADE3A2A787ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_alcove.phy"},
	{0x9D89C51997869A29ULL, "vxdr_0_deeproads_f9r_0_0.vlm"},
	{0x9D89EEEF97B03413ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightl.phy"},
	{0x9D8A93D78E0FB6A9ULL, "175718332.ogg"},
	{0x9D8ABC258A889097ULL, "ht_0_hightown_mmap_ph.dds"},
	{0x9D8AC1F8FCDE7010ULL, "65201185.ogg"},
	{0x9D8AC7B57AB90AFCULL, "6095763_m.fxe"},
	{0x9D8B880A5FC7853CULL, "336366980.ogg"},
	{0x9D8D1720F4BCD9E5ULL, "art\\vfx\\combat\\vfx_gib_corm_rblt.anb"},
	{0x9D8D1720F4BCD9EEULL, "art\\vfx\\combat\\vfx_gib_corm_rblt.ani"},
	{0x9D8D597CDA9021E6ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_lp.ani"},
	{0x9D8DC957FC147D03ULL, "hlac_ht_chantry_2534_47y.dds"},
	{0x9D8DF5F7535E3308ULL, "6096502_m.fxe"},
	{0x9D8E0AB57369B466ULL, "vxdr_0_deeproads_fia_0_0.vlm"},
	{0x9D8E5CFBB5A236A5ULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.mmh"},
	{0x9D8EC2386DCC85A2ULL, "27849730.ogg"},
	{0x9D8F2F9CB22939FBULL, "vxlt_0_lowtown_docks_c6v_0_0.vlm"},
	{0x9D8F9BA53E1E6C8CULL, "1031626582.ogg"},
	{0x9D8F9E8029200592ULL, "vxht_keep_n_6gc_0_0.vlm"},
	{0x9D8FA722C27CEDBFULL, "vxw_woundedcoast_dr_0_0.vlm"},
	{0x9D9040D1C0361BEAULL, "249482465.ogg"},
	{0x9D90487E1DC1113DULL, "vxht_mansion2_1vo_0_0.vlm"},
	{0x9D9050EF36B80F87ULL, "hlac_lt_0_lowtown_docks_n_65533_3y.dds"},
	{0x9D9059B1369B4B1FULL, "vxlt_0_lowtown_a3_n_c_h28_0_0.vlm"},
	{0x9D912DD7196FE179ULL, "vxht_mansion_n_5cb_0_1.vlm"},
	{0x9D913EEEF24E978DULL, "vrc321_party_barks6095786.cl"},
	{0x9D9194C7C1BCD83BULL, "vxca_caverns1_mines_gk8_0_1.vlm"},
	{0x9D91B5B5223C3F64ULL, "6156185_m.fxe"},
	{0x9D920C39BEEFEEA2ULL, "hlac_ht_0_hightown_a3_196607_0y.dds"},
	{0x9D922DF89705180FULL, "779770327.ogg"},
	{0x9D936241688D1A6CULL, "vxw_bonepit_1wz_0_0.vlm"},
	{0x9D93B2CF5C7CAB33ULL, "plt_lgt210pt_frame.plo"},
	{0x9D9421EF0092F519ULL, "gen00fl_aveline6131399.cl"},
	{0x9D947121642215A5ULL, "vxw_woundedcoast_12g_0_0.vlm"},
	{0x9D95203F389C8A53ULL, "gol_flare.mao"},
	{0x9D958684FE60ACD1ULL, "774025240.ogg"},
	{0x9D95CE52D5EF57A2ULL, "vxca_cave_2j3_0_0.vlm"},
	{0x9D95DD66AF0D9935ULL, "hlac_lt_0_lowtown_a2_n_c_65534_6z.dds"},
	{0x9D967E4C38B632B5ULL, "6145118_f.fxe"},
	{0x9D974D92422C0AE7ULL, "6116814_m.fxe"},
	{0x9D9752CB12FA934CULL, "ave100_donnic6105953.cl"},
	{0x9D976D294EBD9D33ULL, "6188606_m.fxe"},
	{0x9D978FE0B28F8C50ULL, "water_water_p3.psh.fxo.dx11.assembly"},
	{0x9D97DE28357FD993ULL, "w_bonepit_55_0.rml"},
	{0x9D9826C715692B0AULL, "vxlt_warehouse_122_0_0.vlm"},
	{0x9D98B5FFF98C6845ULL, "store_dae_210_store4.utm"},
	{0x9D98D964DE746908ULL, "1020519604.ogg"},
	{0x9D9948EA209D0CE1ULL, "vxca_caverns1_flv_0_0.vlm"},
	{0x9D9983DF9EE19852ULL, "gen00fl_fenris6112060.cl"},
	{0x9D9A0B46E65B8393ULL, "lt_warehouse_4_0_lc.mmh"},
	{0x9D9A42F67C2FD031ULL, "350809244.ogg"},
	{0x9D9B417CDA9C3657ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_lp.evt"},
	{0x9D9B59FBB5ACBC72ULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.anb"},
	{0x9D9B7C9743D254FFULL, "plt_cod_aow_journal.plo"},
	{0x9D9BEAB3FAA7004FULL, "93296563.ogg"},
	{0x9D9CD8B8116CD326ULL, "lt_backalley_n.arl"},
	{0x9D9CDCB8116CDA74ULL, "lt_backalley_n.anb"},
	{0x9D9D2C580F399214ULL, "vxlt_hangedman_10b_0_0.vlm"},
	{0x9D9D6ECD2D9747B8ULL, "hlac_lt_gallowscourtyard_-65537_1z.dds"},
	{0x9D9D74D3ADAB364AULL, "518735276.ogg"},
	{0x9D9DA179A41F527BULL, "vxlt_0_lowtown_docks_c3i_0_0.vlm"},
	{0x9D9DBFBE41611D5EULL, "6110116_m.fxe"},
	{0x9D9DC021646D06C0ULL, "6087056_m.fxe"},
	{0x9D9E105D1B161E12ULL, "vxht_keep_n_6bl_0_0.vlm"},
	{0x9D9E81C2B4CB75A6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basicxa.phy"},
	{0x9D9E8EFB5BE3F795ULL, "ht_bartrand_n_3364_ll.mmh"},
	{0x9D9EA006B8491270ULL, "qun111_ketojan6120456.cl"},
	{0x9D9EABFDA3A375AFULL, "825982052.ogg"},
	{0x9D9EB6F285E12D6CULL, "vxw_woundedcoast_pq_0_0.vlm"},
	{0x9D9EBCFEE9A48AA9ULL, "695698600.ogg"},
	{0x9D9FAE3637F216F9ULL, "cs_pro000cs_innblood_carv.bnk"},
	{0x9D9FF2D3AF152BD6ULL, "6041406_f.fxe"},
	{0x9DA03324C039417DULL, "dae000_male_commoner2_mob6174445.cl"},
	{0x9DA05800548EBF60ULL, "hlac_lt_0_lowtown_131072_4y.dds"},
	{0x9DA0972EE32E03AAULL, "lt_blackemporium_1055.rml"},
	{0x9DA0F020F4CD4028ULL, "art\\vfx\\combat\\vfx_gib_corm_rblt.gad"},
	{0x9DA1AC366A7786B3ULL, "6056736_m.fxe"},
	{0x9DA1D9C5707DC589ULL, "gen00fl_aveline6131889.cl"},
	{0x9DA1EB02EB223ABBULL, "747101502.ogg"},
	{0x9DA2427CDAA256E0ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_lp.gad"},
	{0x9DA28B51CF872F71ULL, "hlac_lt_gallowstemplar_2391_6z.dds"},
	{0x9DA2EE5B939EBC0DULL, "vxht_keep_6gc_0_0.vlm"},
	{0x9DA31E1BA6515485ULL, "49254522.ogg"},
	{0x9DA3768E5FF1735BULL, "hlac_w_blightlands_131069_0z.dds"},
	{0x9DA38A414E69CDDCULL, "static_static_p7.psh.fxo"},
	{0x9DA3D6BD08FA28BCULL, "447256392.ogg"},
	{0x9DA3E62CAF38033DULL, "6123337_m.fxe"},
	{0x9DA43A6864E620C0ULL, "vxht_mansion2_269_0_0.vlm"},
	{0x9DA44EFB5DCBD1FEULL, "498884272.ogg"},
	{0x9DA48101A907272FULL, "vxlt_0_lowtown_a2_n_gmy_0_0.vlm"},
	{0x9DA4B95873EBCBC9ULL, "mag291_demon_marethari6028941.cl"},
	{0x9DA5165096826CB5ULL, "6055233_f.fxe"},
	{0x9DA57C2F6EBCA06FULL, "vxdr_primevalentrance_39n_0_1.vlm"},
	{0x9DA5978707533D37ULL, "6034286_m.fxe"},
	{0x9DA5A97CA143BF85ULL, "hlac_w_sundermount_65540_0x.dds"},
	{0x9DA5EE7DB5F52C66ULL, "6207815_f.fxe"},
	{0x9DA60B5F25986393ULL, "mag130_danzig6016430.cl"},
	{0x9DA674E14651EF4BULL, "72655918.ogg"},
	{0x9DA68F14011EAF91ULL, "91574030.ogg"},
	{0x9DA74D5FF80DE475ULL, "hlac_lt_undercityencounter_-1_23x.dds"},
	{0x9DA7CB364C0B887AULL, "gen_im_cft_reg_l.uti"},
	{0x9DA7DDF7F4FBCCBAULL, "gen00fl_merrill6127503.cl"},
	{0x9DA8A18C19131D6FULL, "727861835.ogg"},
	{0x9DA8D76B94D24C99ULL, "hm_sailor02.mor"},
	{0x9DA8E370028A9862ULL, "llac_ht_chantry_2534_45.dds"},
	{0x9DA8FD72A8B73575ULL, "abi_m_hex_misdirect.bnk"},
	{0x9DA900B44DDC5E2DULL, "dae000_quintus6154713.cl"},
	{0x9DA96B4F4335181BULL, "gen00fl_aveline6152585.cl"},
	{0x9DAA75555A0BAF60ULL, "gen00fl_isabela6135717.cl"},
	{0x9DAAD9541F529A30ULL, "243944925.ogg"},
	{0x9DAB15BEC09E1FDFULL, "hlac_lt_0_lowtown_a2_n_c_-262144_3z.dds"},
	{0x9DAB2909BA699DD3ULL, "vxlt_hangedman_nv_0_0.vlm"},
	{0x9DAB395AD037E9ABULL, "vxdr_primevaltomb_3a7_0_0.vlm"},
	{0x9DAB6BA6B629D3C4ULL, "vxlt_darktown_4lv_0_0.vlm"},
	{0x9DAD2C38CA5A7EC1ULL, "vxw_bonepit_1ie_0_0.vlm"},
	{0x9DADCD6507454277ULL, "vxlt_darktown_531_0_0.vlm"},
	{0x9DADD1FFBAFD6010ULL, "vxdr_primevaltomb_7ej_0_2.vlm"},
	{0x9DAE297A332B7785ULL, "220416767.ogg"},
	{0x9DAE39B275906DCBULL, "110913271.ogg"},
	{0x9DAEE0AD433C53FEULL, "vxlt_hangedman_1vo_0_1.vlm"},
	{0x9DAF29D292B6B295ULL, "ht_0_hightown_a2_35_0_lc.mmh"},
	{0x9DB0305A08014A11ULL, "6090365_m.fxe"},
	{0x9DB08C136657A102ULL, "hlac_lt_0_lowtown_a2_n_c_0_4x.dds"},
	{0x9DB098B089B5254EULL, "ts_beasts_spdr_prog_silk.gda"},
	{0x9DB122600442EA90ULL, "25073010.ogg"},
	{0x9DB154DB480E46A5ULL, "vxw_woundedcoast_18e_0_0.vlm"},
	{0x9DB185DE75AD1784ULL, "1053355334.ogg"},
	{0x9DB1C0366B8E2536ULL, "782635030.ogg"},
	{0x9DB2F2DE9AFD3FAFULL, "vxlt_0_lowtown_a2_n_gjx_0_0.vlm"},
	{0x9DB37F24A6B66811ULL, "424741193.ogg"},
	{0x9DB39538217CC22CULL, "127901527.ogg"},
	{0x9DB3A8C52428B2E6ULL, "gen00fl_dog6166820.cl"},
	{0x9DB41374F83FC345ULL, "vxlt_warehouse_n_b1_0_0.vlm"},
	{0x9DB4E2850DAA0576ULL, "vxlt_0_lowtown_docks_a3_n_8ue_0_0.vlm"},
	{0x9DB5162D021394A8ULL, "vxht_mansion2b_3dn_0_0.vlm"},
	{0x9DB5305554C1B816ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.dura.evt"},
	{0x9DB5306BE50422B3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_exit.evt"},
	{0x9DB5432C34559452ULL, "prestartmenu.gfx"},
	{0x9DB5A3122A68E8F4ULL, "285619146.ogg"},
	{0x9DB5F75C087A6F9CULL, "vxlt_backalley_g_1ef_0_0.vlm"},
	{0x9DB6154195C255C0ULL, "llac_lt_0_lowtown_n_-2_5.dds"},
	{0x9DB67D64C8B04DCEULL, "230130812.ogg"},
	{0x9DB6EAC5186F55F0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_noba_0n.dds"},
	{0x9DB8670EEFDB70B6ULL, "555672226.ogg"},
	{0x9DB8B63B3239024EULL, "1057840627.ogg"},
	{0x9DB94AC9D7B3E257ULL, "125297094.ogg"},
	{0x9DB95D42BEB760C0ULL, "w_bonepit_20_0_lc.mmh"},
	{0x9DBA79394001B8AAULL, "hm_walter.mor"},
	{0x9DBA8288FDD37177ULL, "vxw_woundedcoast_1gr_0_0.vlm"},
	{0x9DBAF8AE8D7CDAE0ULL, "vxca_caverns2_74m_0_0.vlm"},
	{0x9DBB1928FD77264EULL, "692456697.ogg"},
	{0x9DBB9E14F3A35ABDULL, "vxht_mansion2_1gl_0_0.vlm"},
	{0x9DBBC0ADA84AD02FULL, "personality.gda"},
	{0x9DBBC3062FEF2BC0ULL, "549857804.ogg"},
	{0x9DBC013D40D10A6BULL, "zz_vault_debug6203395.cl"},
	{0x9DBC275554C7C7A1ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.dura.gad"},
	{0x9DBC296BE50A35F4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_exit.gad"},
	{0x9DBC42E283E96624ULL, "750717162.ogg"},
	{0x9DBC4F095D840E70ULL, "172614719.ogg"},
	{0x9DBC5D814AA33514ULL, "6121121_m.fxe"},
	{0x9DBC6B804AA5D6DCULL, "6105592_m.fxe"},
	{0x9DBC7760F4DCEEE1ULL, "6093650_m.fxe"},
	{0x9DBD01438FFD436BULL, "gen00fl_aveline6120801.cl"},
	{0x9DBD04DCE62C3543ULL, "vxlt_0_lowtown_docks_a3_9r7_0_0.vlm"},
	{0x9DBD41691D47A5C4ULL, "lt_0_lowtown_a2_37_0_ll.mmh"},
	{0x9DBDF6B2A6AD2E1FULL, "cd_qun131_warrens_thug.bnk"},
	{0x9DBE05A780EF15BCULL, "829328651.ogg"},
	{0x9DBE19E56D013CBFULL, "vxw_bonepit_1wu_0_0.vlm"},
	{0x9DBE3E78DC8F4A79ULL, "vxht_keep_73b_0_0.vlm"},
	{0x9DBEC3825282EE3BULL, "hlac_ca_cavern1_haunted_24425_6z.dds"},
	{0x9DBEE1B485E0264BULL, "6152560_m.fxe"},
	{0x9DBEF6C84DA077B7ULL, "173876377.ogg"},
	{0x9DBF022A6264B06DULL, "vxw_woundedcoast_1d9_0_0.vlm"},
	{0x9DBF7B082E364023ULL, "llac_lt_0_lowtown_-131073_0.dds"},
	{0x9DBFBE5EB350A98DULL, "framebuffereffects_ps3zcullreload_p0.psh.fxo.dx11"},
	{0x9DBFD215778AED0DULL, "vxht_keep_6ma_0_0.vlm"},
	{0x9DBFF3AF514A0071ULL, "vxht_mansion2b_2nd_0_0.vlm"},
	{0x9DC0512CEEE0F4ADULL, "vxw_bonepit_27x_0_0.vlm"},
	{0x9DC13473B07EE33CULL, "zz_mag_debug6054597.cl"},
	{0x9DC18C2A40EC2E3FULL, "pro000_friedrich6051056.cl"},
	{0x9DC1A836B7C54EF0ULL, "6165239_m.fxe"},
	{0x9DC1A8FBF92035A9ULL, "6174588_m.fxe"},
	{0x9DC1AD96C0DC2D88ULL, "6079079_m.fxe"},
	{0x9DC28551EA35DBC5ULL, "vxlt_warehouse_n_i0_0_0.vlm"},
	{0x9DC3285554CDE84FULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.dura.ani"},
	{0x9DC32F952B19AF2AULL, "lgt000cr_hate_thugs.utc"},
	{0x9DC3386BE5106DD2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_exit.ani"},
	{0x9DC399CF5CA5A4D0ULL, "plt_lgt210pt_frame.nss"},
	{0x9DC3AC90ED5B6BD6ULL, "pro000_merchant6051624.cl"},
	{0x9DC403EF35EDBEE9ULL, "mag160_feynriel6020466.cl"},
	{0x9DC48273BFDD9F3DULL, "432918844.ogg"},
	{0x9DC4BDF423625ED3ULL, "882630673.ogg"},
	{0x9DC56DDED40D3FDDULL, "301707191.ogg"},
	{0x9DC58184AE7299A4ULL, "art\\characters\\creatures\\ragedemon\\animation\\ctn180l.gad"},
	{0x9DC65936C1A9E8D4ULL, "6117627_m.fxe"},
	{0x9DC690302D478422ULL, "816227910.ogg"},
	{0x9DC7427BA819F87DULL, "602765838.ogg"},
	{0x9DC78141A64D8598ULL, "vxht_mansion_n_6s1_0_0.vlm"},
	{0x9DC7A8C4D70F031CULL, "aip_mercenarywarrior2h_base.gda"},
	{0x9DC8AFB55D219801ULL, "hairalpha_hair_p2.psh.fxo.dx11.assembly"},
	{0x9DC91C786D7FD3FDULL, "39007240.ogg"},
	{0x9DC920C1A19A0FCBULL, "6088450_m.fxe"},
	{0x9DC923E6E6F45B94ULL, "hm_amb_merchant_wander.mor"},
	{0x9DC97AE84DBA8E9CULL, "1061613909.ogg"},
	{0x9DCA37A88819AF06ULL, "hlac_lt_0_lowtown_131072_3z.dds"},
	{0x9DCA50617B0D5D69ULL, "894984337.ogg"},
	{0x9DCA9BC07FC8A072ULL, "941925624.ogg"},
	{0x9DCAAAC3E1D39DEAULL, "dae000cr_grand_cleric6183957.cl"},
	{0x9DCAB11EEB317377ULL, "vxlt_darktown_593_0_0.vlm"},
	{0x9DCB713EE26F38BFULL, "lgt320_mis_resp_swordhunter6150268.cl"},
	{0x9DCB99ADFE1D745EULL, "6071297_m.fxe"},
	{0x9DCBF9D4D87B9544ULL, "narrative_banter6136816.cl"},
	{0x9DCC1CCFA6C39262ULL, "vxlt_undercitydungeon_wn_0_0.vlm"},
	{0x9DCC9084AE78D182ULL, "art\\characters\\creatures\\ragedemon\\animation\\ctn180l.ani"},
	{0x9DCC9A9EF59E5CC7ULL, "hlac_lt_0_lowtown_a2_n_65533_0y.dds"},
	{0x9DCCD6640C305B2DULL, "llac_ht_mansion2_818_12.dds"},
	{0x9DCCD96677E02694ULL, "dae000_arishok6067687.cl"},
	{0x9DCCF0A4D38F051FULL, "vxw_encounter2_6i_0_0.vlm"},
	{0x9DCD632DCAA46936ULL, "6116107_m.fxe"},
	{0x9DCD7407545AC8EDULL, "vxw_woundedcoast_1o8_0_0.vlm"},
	{0x9DCE80E019A66DC6ULL, "6004377_f.fxe"},
	{0x9DCEFD5B11D398AEULL, "6126899_f.fxe"},
	{0x9DCEFF33A4328986ULL, "qun230cr_guild_contact6068696.cl"},
	{0x9DCF1F648859A8E7ULL, "6128697_m.fxe"},
	{0x9DCF30D1473C2A0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_enter.gad"},
	{0x9DCF8A7E8E3FAD67ULL, "174661372.ogg"},
	{0x9DCFB3BDEF7ADF85ULL, "vis121_saemus6099484.cl"},
	{0x9DCFC9991F1D6A7BULL, "gen_im_cns_tom_war_reaver.uti"},
	{0x9DCFCE814062CFB3ULL, "plt_prm000pt_isaac_armor_.plo"},
	{0x9DD00B05D2532B96ULL, "533217968.ogg"},
	{0x9DD02B77B9FAA405ULL, "hlac_ht_keep_3743_0y.dds"},
	{0x9DD05DE4B3BBDF48ULL, "lt_0_lowtown_a2_n_c_6_0.mmh"},
	{0x9DD069E4B3BBF3EBULL, "lt_0_lowtown_a2_n_c_6_0.mao"},
	{0x9DD07BE4B3BC1272ULL, "lt_0_lowtown_a2_n_c_6_0.msh"},
	{0x9DD0E4DDC0FAB87FULL, "6206018_m.fxe"},
	{0x9DD115B811406B71ULL, "art\\vfx\\combat\\vfx_c_pride_cast_p_c.anb"},
	{0x9DD13CC6BEC41D85ULL, "ht_0_hightown_a3_28_0.anb"},
	{0x9DD1867EB8EC2C54ULL, "vxlt_hangedman_4k_0_0.vlm"},
	{0x9DD1A5393204DE1BULL, "w_encounter1_7_0.rml"},
	{0x9DD21EA7596EAB42ULL, "wispy_smoke_darker.dds"},
	{0x9DD227C7405B3FEFULL, "6126462_m.fxe"},
	{0x9DD25208F67F025AULL, "701952549.ogg"},
	{0x9DD2BF1FBD5BE2DDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masf_0d.dds"},
	{0x9DD371AB06068E39ULL, "vxht_mansion2b_29i_0_0.vlm"},
	{0x9DD3F8C513C60197ULL, "vxdr_0_deeproads_ete_0_1.vlm"},
	{0x9DD447623D5FC9E3ULL, "gen00fl_isabela6135671.cl"},
	{0x9DD4B27DE16D458FULL, "6079652_m.fxe"},
	{0x9DD4C3565B5CA7B1ULL, "dae300_cullen6135333.cl"},
	{0x9DD4E721BC3B6BF9ULL, "652882556.ogg"},
	{0x9DD52372A8DCB0FFULL, "abi_m_hex_misdirect.wbd"},
	{0x9DD545D14740B9BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_enter.evt"},
	{0x9DD5CC627B48B61AULL, "mcr370_templar2.cnv"},
	{0x9DD6008963304E56ULL, "vxlt_backalley_g_1l7_0_0.vlm"},
	{0x9DD6595FBC4AFF28ULL, "556741631.ogg"},
	{0x9DD69286E5917545ULL, "877463628.ogg"},
	{0x9DD71E4E65ED117CULL, "llac_lt_0_lowtown_a2_-3_7.dds"},
	{0x9DD80073CB79684CULL, "6171098_m.fxe"},
	{0x9DD8378C2C90C9E9ULL, "hf_qmq_emeline.mor"},
	{0x9DD881CED78F8BBDULL, "vxht_mansion2_1mj_0_0.vlm"},
	{0x9DD88271F7041C93ULL, "249612342.ogg"},
	{0x9DD8DB7A047FB5C9ULL, "207947770.ogg"},
	{0x9DD8EE627B4B2662ULL, "mcr370_templar2.bnk"},
	{0x9DD9279690DFA990ULL, "vxdr_primevaltomb_7tu_0_2.vlm"},
	{0x9DDA946BA55689F1ULL, "6170492_m.fxe"},
	{0x9DDAD7F17E684E2DULL, "combat_lc_range.gda"},
	{0x9DDB06BCB47D84B9ULL, "vxlt_0_lowtown_a2_n_f1v_0_0.vlm"},
	{0x9DDB6281C5851D9FULL, "vxw_encounter2_1h_0_0.vlm"},
	{0x9DDBA4E89AEF9825ULL, "6015685_m.fxe"},
	{0x9DDBB015694FE82FULL, "239112764.ogg"},
	{0x9DDBD57466562B7DULL, "6099233_f.fxe"},
	{0x9DDBE5E44650E16DULL, "vxw_woundedcoast_1j7_0_0.vlm"},
	{0x9DDC2B6876E9C9CBULL, "pro000_opening6178533.cl"},
	{0x9DDC8A5C560DBD5CULL, "sse_ss_meredith.gda"},
	{0x9DDCD76935363171ULL, "vxht_mansion2b_2dz_0_0.vlm"},
	{0x9DDD4C15732E815EULL, "vxlt_0_lowtown_a3_n_c_hqx_0_0.vlm"},
	{0x9DDD8B23250D0CF5ULL, "qcr000_arishok6060164.cl"},
	{0x9DDDE333B08E3245ULL, "335095130.ogg"},
	{0x9DDE0DD25FA92DBDULL, "follower_banter6171041.cl"},
	{0x9DDF1B47CED5144BULL, "art\\vfx\\combat\\vfx_m_ent_horrow_imp_c.impa.evt"},
	{0x9DDF36B8114CE0EAULL, "art\\vfx\\combat\\vfx_c_pride_cast_p_c.mmh"},
	{0x9DDFCD9D911527B4ULL, "23951297.ogg"},
	{0x9DE040FC8A2AEFA4ULL, "686134699.ogg"},
	{0x9DE06E23DA2AF36EULL, "6113947_f.fxe"},
	{0x9DE08292659B1B96ULL, "lgt320_shw_ors_black_op6156334.cl"},
	{0x9DE0B15ADD5EC660ULL, "art\\vfx\\textures\\vfx_crack_cyc.dds"},
	{0x9DE0EDE105BD6FECULL, "1013621499.ogg"},
	{0x9DE17099F78F6EB0ULL, "fx_blizz_swirl.dds"},
	{0x9DE18A819CC9E32DULL, "6091964_m.fxe"},
	{0x9DE1A505E33113C1ULL, "6101139_m.fxe"},
	{0x9DE1E387F7FCA6B9ULL, "vxht_mansion2b_22h_0_0.vlm"},
	{0x9DE1EB842D38C3AFULL, "mag321_nella6064940.cl"},
	{0x9DE24146216BD183ULL, "llac_w_encounter3_-196608_1.dds"},
	{0x9DE27453ADB2C433ULL, "seb201_lady_harimann6177306.cl"},
	{0x9DE2D09C1CD161B5ULL, "follower_banter6168601.cl"},
	{0x9DE300ED385D33EFULL, "442899322.ogg"},
	{0x9DE34DD1474D04D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\overlay_enter.ani"},
	{0x9DE37493309029BAULL, "the_antivan_garrote.dds"},
	{0x9DE38BD99C5C8D03ULL, "863464109.ogg"},
	{0x9DE3B3547883834FULL, "6139413_m.fxe"},
	{0x9DE3D1C9726B5951ULL, "dae000_madam_lusine6154906.cl"},
	{0x9DE44D3163AF0B63ULL, "976909243.ogg"},
	{0x9DE4CB37948C9C88ULL, "96583346.ogg"},
	{0x9DE4F05F91DDC693ULL, "hlac_ht_keep_3731_2x.dds"},
	{0x9DE599A793AA98B0ULL, "mag100tr_eau_wilmod.ncs"},
	{0x9DE5D6CE2F90064AULL, "llac_lt_gallowstemplar_2382_3.dds"},
	{0x9DE63447CEDB5D0CULL, "art\\vfx\\combat\\vfx_m_ent_horrow_imp_c.impa.gad"},
	{0x9DE66A2ADB8C1920ULL, "6087715_m.fxe"},
	{0x9DE67020BD4FC8F7ULL, "gen00fl_merrill6108631.cl"},
	{0x9DE682F26E66FB23ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfallr_3.mao"},
	{0x9DE6C1681614B852ULL, "hlac_lt_darktown_-1_63y.dds"},
	{0x9DE7D39CCBBED721ULL, "329322620.ogg"},
	{0x9DE7E77146FE947AULL, "317055744.ogg"},
	{0x9DE7EC233E6F038CULL, "788406940.ogg"},
	{0x9DE8384E601C2A26ULL, "6101210_f.fxe"},
	{0x9DE8EECDBB0813F0ULL, "vrc321ip_spooky7_obj_8.utp"},
	{0x9DE9DCB2A6D23D55ULL, "cd_qun131_warrens_thug.wbd"},
	{0x9DEA27BAE1BF6A9FULL, "vxw_blightlands_ff_0_0.vlm"},
	{0x9DEAAFCFDCF6E4F1ULL, "hlac_lt_gallowstemplar_f_7198_7z.dds"},
	{0x9DEB993DE40C5610ULL, "zz_lgt_2_debug6211874.cl"},
	{0x9DEBE7EF9802F288ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightl.mmh"},
	{0x9DEBF3EF9803072BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightl.mao"},
	{0x9DEC0050B4441EC4ULL, "mrl000_marethari6129319.cl"},
	{0x9DEC056F6896B0EDULL, "6069722_f.fxe"},
	{0x9DEC05EF980325B2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_straightl.msh"},
	{0x9DEC27392E74EE99ULL, "747955592.ogg"},
	{0x9DEC3347CEDFC70AULL, "art\\vfx\\combat\\vfx_m_ent_horrow_imp_c.impa.ani"},
	{0x9DEC3E911D4BF88EULL, "vxlt_gallowstemplar_f_4u8_0_0.vlm"},
	{0x9DEC3EBC7F107E11ULL, "art\\characters\\weapons\\shields\\w_shd_seas.msh"},
	{0x9DEC516429FA7DCEULL, "hlac_ca_caverns1_24421_7y.dds"},
	{0x9DEC58BC7F10AACFULL, "art\\characters\\weapons\\shields\\w_shd_seas.mmh"},
	{0x9DED19C5877E6681ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_011a_0d.dds"},
	{0x9DED3B848679A09BULL, "hlac_lt_warehouse_0_2x.dds"},
	{0x9DED602B31D9EED8ULL, "zz_qcr_debug6053504.cl"},
	{0x9DEDBA3E0324CC4AULL, "dae000_serendipity6161425.cl"},
	{0x9DEE21C8D5F0D963ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.impa.ani"},
	{0x9DEE3FC07A002360ULL, "857628488.ogg"},
	{0x9DEE60006ECE0BC6ULL, "lowtown_male_beggar_sta6151036.cl"},
	{0x9DEE6526D704E535ULL, "6095897_f.fxe"},
	{0x9DEE765DB41AB41CULL, "character_character_p100.psh.fxo"},
	{0x9DEE8D723A3FF1DEULL, "vxw_encounter2_f4_0_0.vlm"},
	{0x9DEE91516704A34EULL, "vxlt_0_lowtown_a2_n_d7q_0_0.vlm"},
	{0x9DEED5FBC52ADC23ULL, "6133068_m.fxe"},
	{0x9DEF0B81244770EFULL, "6128662_m.fxe"},
	{0x9DEF32597720A081ULL, "hlac_w_encounter2_1_0x.dds"},
	{0x9DEF610A0719393FULL, "6093968_m.fxe"},
	{0x9DEFB10568B019A9ULL, "642206823.ogg"},
	{0x9DF04A14D10FF63AULL, "vxht_bartrand_n_663_0_0.vlm"},
	{0x9DF08DB26EDA73B3ULL, "dae000_dalish_apprentices.bnk"},
	{0x9DF0DA673EF666E3ULL, "vxlt_undercityencounter_119_0_0.vlm"},
	{0x9DF107B398AA7745ULL, "172725890.ogg"},
	{0x9DF10D8420A95DF8ULL, "lgt000ip_ultra_dwf_remains.ncs"},
	{0x9DF135735B0A55DAULL, "vxht_0_hightown_a3_byo_0_0.vlm"},
	{0x9DF13780391BA815ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_basic.phy"},
	{0x9DF1C627AF3BD8E7ULL, "follower_barks6153589.cl"},
	{0x9DF1ECD974F37D54ULL, "vxlt_0_lowtown_n_ewv_0_0.vlm"},
	{0x9DF27F71EF3AB915ULL, "vxlt_gallowscourtyard_72x_0_0.vlm"},
	{0x9DF2DB00850B1B4DULL, "hlac_w_blightlands_196611_0x.dds"},
	{0x9DF3AD23A62085C0ULL, "ser270_emeric6112663.cl"},
	{0x9DF3AFB26EDCE303ULL, "dae000_dalish_apprentices.cnv"},
	{0x9DF3C9069C50A7F6ULL, "mer210_coterie_contact6035280.cl"},
	{0x9DF40F649618D383ULL, "47121413.ogg"},
	{0x9DF490407941BDECULL, "follower_banter6154216.cl"},
	{0x9DF4F566312CE245ULL, "hlac_lt_0_lowtown_docks_a3_-262144_3y.dds"},
	{0x9DF5EC62126E1A2CULL, "art\\characters\\playercharacter\\humanfemale\\hf_hoo_chab.mao"},
	{0x9DF605DE9E69952EULL, "743984719.ogg"},
	{0x9DF60E6F5271BFC3ULL, "541306574.ogg"},
	{0x9DF655F0D762D09FULL, "511260523.ogg"},
	{0x9DF716EAC994C2D6ULL, "502690305.ogg"},
	{0x9DF721F865F540C7ULL, "6082247_m.fxe"},
	{0x9DF72BE7CD120163ULL, "451515211.ogg"},
	{0x9DF767BF1D31B0F5ULL, "plt_lgt160pt_cod_avern_lett.plo"},
	{0x9DF77093C6B897AFULL, "t3_eye_hzl.tnt"},
	{0x9DF7C18D58B0A470ULL, "370243319.ogg"},
	{0x9DF7CFDEC69AFBAFULL, "6142547_m.fxe"},
	{0x9DF7E2627B65C257ULL, "mcr370_templar2.utc"},
	{0x9DF849F4CA5B6168ULL, "dr_primevaltomb_657_ll.mmh"},
	{0x9DF89997D3B5831FULL, "vxw_blightlands_ka_0_0.vlm"},
	{0x9DF8C99E2A3D126DULL, "vxw_woundedcoast_1p5_0_0.vlm"},
	{0x9DF8D97961B5B737ULL, "hlac_w_bonepitmining_a2_262144_0z.dds"},
	{0x9DF8EED7C319E075ULL, "dre000_varric6035138.cl"},
	{0x9DF9102B24ED2784ULL, "hlac_w_blightlands_131072_0y.dds"},
	{0x9DF96C74FD842FFCULL, "6140256_m.fxe"},
	{0x9DF97F99F7A41A4EULL, "fx_blizz_swirl.mao"},
	{0x9DFA6723093FF5A9ULL, "vxdr_0_deeproads_cf2_0_0.vlm"},
	{0x9DFA873A02154CCFULL, "6150324_m.fxe"},
	{0x9DFAC5DAE0F1CAA7ULL, "6082311_m.fxe"},
	{0x9DFB39C8D5FB8C5AULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.impa.evt"},
	{0x9DFB7BC188248F83ULL, "1041675867.ogg"},
	{0x9DFBE439D7CCC7A6ULL, "hlac_lt_gallowstemplar_2394_4x.dds"},
	{0x9DFC5719A0C297A3ULL, "gen00fl_sebastian6189011.cl"},
	{0x9DFC66EE9ADF0D09ULL, "gen00fl_anders6112889.cl"},
	{0x9DFCFF4F2C360A5EULL, "vxw_encounter2_a5_0_0.vlm"},
	{0x9DFD031724628E0BULL, "gen00fl_fenris6107999.cl"},
	{0x9DFD237F711E80C4ULL, "6000956_m.fxe"},
	{0x9DFD45CE3B3609AAULL, "423656936.ogg"},
	{0x9DFD58629CAAFF1BULL, "250736983.ogg"},
	{0x9DFD85240A5F495EULL, "6179600_f.fxe"},
	{0x9DFDABC5BDA3C476ULL, "sst_ss_shadow_f_1.gda"},
	{0x9DFDE65080E9A3D0ULL, "6054489_m.fxe"},
	{0x9DFE1E45980B7C79ULL, "6037605_m.fxe"},
	{0x9DFE6CF0301D9ED5ULL, "art\\vfx\\combat\\vfx_gib_ogre.msh"},
	{0x9DFE6EF0301DA24BULL, "art\\vfx\\combat\\vfx_gib_ogre.mmh"},
	{0x9DFE72AE48F52A5BULL, "hlac_ca_caverns2_5802_50y.dds"},
	{0x9DFE73C2B001C983ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v1.evt"},
	{0x9DFE7AF0301DB6E8ULL, "art\\vfx\\combat\\vfx_gib_ogre.mao"},
	{0x9DFEFF667C38D97FULL, "vxw_woundedcoast_xg_0_1.vlm"},
	{0x9DFF4C9518E15868ULL, "ico_1h_axe_1.dds"},
	{0x9DFF649221F28477ULL, "vxlt_hangedman_n_2an_0_0.vlm"},
	{0x9DFFA614EEC455BAULL, "one100dg_vanard6047404.cl"},
	{0x9DFFB581408BC950ULL, "plt_prm000pt_isaac_armor_.nss"},
	{0x9DFFBC53D40DA15BULL, "art\\levels\\ztools\\default\\generic\\0ff_basic.mmh"},
	{0x9DFFC853D40DB5F8ULL, "art\\levels\\ztools\\default\\generic\\0ff_basic.mao"},
	{0x9DFFDA53D40DD445ULL, "art\\levels\\ztools\\default\\generic\\0ff_basic.msh"},
	{0x9E00359090178A71ULL, "vxw_sundermount_37c_0_1.vlm"},
	{0x9E00763DEC8E6560ULL, "vxht_0_hightown_9b2_0_0.vlm"},
	{0x9E00FF8D4BB9866EULL, "6154909_m.fxe"},
	{0x9E016C4F1A1DE1D0ULL, "hlac_w_sundermount_0_0y.dds"},
	{0x9E021D172831EDDEULL, "327498005.ogg"},
	{0x9E0238C8D601A94DULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.impa.gad"},
	{0x9E02A503705C5854ULL, "6077284_m.fxe"},
	{0x9E02B3D1D20B9FB3ULL, "428547159.ogg"},
	{0x9E02BA2EF2319DB6ULL, "6174969_m.fxe"},
	{0x9E03E65263EC2EA8ULL, "321035585.ogg"},
	{0x9E043CABC4EE7665ULL, "zz_followers_banter6158334.cl"},
	{0x9E04778FBFA74340ULL, "747938422.ogg"},
	{0x9E054CC2B007A604ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v1.gad"},
	{0x9E05BE19C059278FULL, "159592097.ogg"},
	{0x9E063A0DC454E0A1ULL, "lt_0_lowtown_docks_a3_n_116_0.phy"},
	{0x9E06533626CFB73FULL, "vxw_bonepit_1lp_0_0.vlm"},
	{0x9E06BCCD72A078E0ULL, "125107620.ogg"},
	{0x9E06FF529F6DED57ULL, "vxht_0_hightown_bf2_0_0.vlm"},
	{0x9E071EA7C97889DAULL, "zz_tbe_debug6189390.cl"},
	{0x9E073FBC6F993A96ULL, "337212231.ogg"},
	{0x9E079475716AF38CULL, "vxlt_darktown_2vj_0_0.vlm"},
	{0x9E0799BB4FCA48BDULL, "6075644_f.fxe"},
	{0x9E079B52E9D69EF2ULL, "follower_banter6155820.cl"},
	{0x9E07E316D05A61BEULL, "vxht_chantry_n_2yu_0_0.vlm"},
	{0x9E08185EBF2AB7A2ULL, "lt_0_lowtown_a2_24_0.rml"},
	{0x9E086D706EEE94E6ULL, "zz_ave_debug6174991.cl"},
	{0x9E08BAE6BEAC253EULL, "vxca_cavern1_haunted_ehg_0_0.vlm"},
	{0x9E08C5544F55B602ULL, "combat_2h_r2_a1.gda"},
	{0x9E08E68093A10B8CULL, "ave100_donnic6058930.cl"},
	{0x9E09248420BE16F1ULL, "lgt000ip_ultra_dwf_remains.utp"},
	{0x9E093F04FDADFCE4ULL, "gen00fl_varric6093606.cl"},
	{0x9E0A187F9716C9A8ULL, "lt_0_lowtown_docks_n_38_0_lc.mmh"},
	{0x9E0A3DC6BEF3E48DULL, "ht_0_hightown_a3_28_0.rml"},
	{0x9E0A95B8150AFA33ULL, "612596817.ogg"},
	{0x9E0AAC2DBE242015ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_track_1.mao"},
	{0x9E0ADAA656AFB2BDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_floor_tile_d.dds"},
	{0x9E0AF59D4C82CE0AULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chac.mao"},
	{0x9E0AF99D4C82D559ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chac.mmh"},
	{0x9E0B039D4C82E647ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chac.msh"},
	{0x9E0B0C67BD0BB019ULL, "124675761.ogg"},
	{0x9E0B712C1E2C22DEULL, "vxw_encounter2_l6_0_0.vlm"},
	{0x9E0BDAFCF2B12844ULL, "vxlt_warehouse_n_5q_0_0.vlm"},
	{0x9E0BFBE138895287ULL, "626031993.ogg"},
	{0x9E0C5BC2B00DDEE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_quick_die_v1.ani"},
	{0x9E0C750066980443ULL, "402359378.ogg"},
	{0x9E0C93B0B20E2CB5ULL, "mer240cr_miner5.utc"},
	{0x9E0D4D52E18EF7D6ULL, "art\\vfx\\models\\vfx_rockarmor_rshin.msh"},
	{0x9E0D53FA98A9E09AULL, "vxlt_0_lowtown_docks_a3_n_bug_0_0.vlm"},
	{0x9E0D5752E18F08D4ULL, "art\\vfx\\models\\vfx_rockarmor_rshin.mmh"},
	{0x9E0DA3C7A0BBFC9CULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_distort1.dds"},
	{0x9E0E6ED40BC6EADBULL, "vxdr_slavers_212_0_0.vlm"},
	{0x9E0E73D1FDDA9067ULL, "6156072_m.fxe"},
	{0x9E0ED09358DFAE54ULL, "vxlt_0_lowtown_n_emt_0_0.vlm"},
	{0x9E0F26F066D52916ULL, "880960564.ogg"},
	{0x9E0F6E4730A67CDCULL, "hlac_ht_mansion2b_818_1y.dds"},
	{0x9E0FBCD5183F25D5ULL, "193966902.ogg"},
	{0x9E107F3B6CDEF210ULL, "gen00fl_merrill6124423.cl"},
	{0x9E108D7760DD88DAULL, "vxlt_0_lowtown_a2_n_hgq_0_0.vlm"},
	{0x9E10A6AB8BF1D432ULL, "6162557_m.fxe"},
	{0x9E10B46BDE3FB404ULL, "1072535852.ogg"},
	{0x9E110251D7EBCCD3ULL, "vxht_mansion2_2e7_0_0.vlm"},
	{0x9E1209872CCF436BULL, "6165709_f.fxe"},
	{0x9E1228F9F270398EULL, "6106090_m.fxe"},
	{0x9E124B7627ED0EE0ULL, "vxht_chantry_n_271_0_0.vlm"},
	{0x9E12677D6E6F2BB6ULL, "vxht_mansion2_27l_0_0.vlm"},
	{0x9E12835B15D1D5ACULL, "follower_banter6164769.cl"},
	{0x9E13157A822A21E4ULL, "6114046_m.fxe"},
	{0x9E1317DCF062B107ULL, "6038615_m.fxe"},
	{0x9E1340373A0ED5C9ULL, "137032230.ogg"},
	{0x9E1383FF52E276EAULL, "vxdr_0_deeproads_cry_0_0.vlm"},
	{0x9E13B90533073637ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ce_transition.gad"},
	{0x9E13CB9CC454B8DDULL, "626323799.ogg"},
	{0x9E144139C2E1BA66ULL, "plt_lgt000pt_ultra_2_04.nss"},
	{0x9E15427BD44DFAF0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtb_0t.dds"},
	{0x9E154847E0CC024BULL, "373713779.ogg"},
	{0x9E156BEECFDF1186ULL, "vxht_keep_n_5av_0_0.vlm"},
	{0x9E157D51B7A1B41FULL, "vxw_blightlands_qc_0_0.vlm"},
	{0x9E158C9A1BBB1B3BULL, "fxa_cen_volt3.dds"},
	{0x9E15FC685FA0E21FULL, "1892214.ogg"},
	{0x9E16065263610C0CULL, "vxlt_darktown_2yk_0_0.vlm"},
	{0x9E171AF2D46FEE79ULL, "vxw_encounter1_ao_0_0.vlm"},
	{0x9E17508BDB09EB12ULL, "6098825_m.fxe"},
	{0x9E1765086DDE946EULL, "vxlt_darktown_5b7_0_0.vlm"},
	{0x9E17F1ACAB91A293ULL, "dae000_bartrand6201368.cl"},
	{0x9E1810563CFFC6E1ULL, "vxlt_foundry_n_xq_0_0.vlm"},
	{0x9E18323922E2B482ULL, "hlac_ht_0_hightown_-65536_0y.dds"},
	{0x9E187374E4C50B7EULL, "vxht_keep_n_5if_0_0.vlm"},
	{0x9E19171A4A9CBBB3ULL, "hlac_ca_caverns1_24424_4z.dds"},
	{0x9E192829B1867816ULL, "vxw_encounter2_3s_0_0.vlm"},
	{0x9E193F571F16DC52ULL, "6089507_f.fxe"},
	{0x9E198E129F767512ULL, "6132839_m.fxe"},
	{0x9E19A4C4AAF053ABULL, "plt_cod_crt_skeleton.plo"},
	{0x9E19B0775969886EULL, "6081979_m.fxe"},
	{0x9E19CE05330BC5C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ce_transition.evt"},
	{0x9E1A19ACB035187FULL, "606044410.ogg"},
	{0x9E1A4CD9E4A740C4ULL, "vxlt_warehouse_n_8v_0_0.vlm"},
	{0x9E1A68DDEE4B7A5EULL, "6177400_f.fxe"},
	{0x9E1A830799406B52ULL, "vxht_mansion2_2e2_0_0.vlm"},
	{0x9E1ADA748CC4FDBAULL, "101862297.ogg"},
	{0x9E1AFB60469B3AA2ULL, "lt_0_lowtown_docks_n_29_0_lc.mmh"},
	{0x9E1AFDF6F608001DULL, "lowtown_guard_chase.bnk"},
	{0x9E1D0AB52BD1E765ULL, "dre000_varric6208046.cl"},
	{0x9E1D9AAF957D2520ULL, "199681820.ogg"},
	{0x9E1DB2F3B25BD62BULL, "6088112_f.fxe"},
	{0x9E1EDFF6F60BB625ULL, "lowtown_guard_chase.cnv"},
	{0x9E1EF44B6988EAFEULL, "1004066865.ogg"},
	{0x9E1F18887DE7ECB2ULL, "6167657_m.fxe"},
	{0x9E1F3C5D11D391F8ULL, "qcr220_grey_wardens6135927.cl"},
	{0x9E1F5943183E9C8DULL, "hlac_w_sundermount_262149_0x.dds"},
	{0x9E1FDD8A7AA6F24BULL, "mag291_vincento6027389.cl"},
	{0x9E204124AE51A166ULL, "hlac_lt_darktown_-1_15x.dds"},
	{0x9E20D373F1F9ABE5ULL, "panning_h.cur"},
	{0x9E20D95A60654436ULL, "vxht_mansion2_20m_0_0.vlm"},
	{0x9E213437BACD4A70ULL, "zz_followers_banter6158012.cl"},
	{0x9E215C7193C2898EULL, "6162614_m.fxe"},
	{0x9E21794BC5A38D43ULL, "6079570_f.fxe"},
	{0x9E2187FD4DA83438ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_lt_ash.mmh"},
	{0x9E21990BFE5472C8ULL, "vxlt_foundry_n_xt_0_0.vlm"},
	{0x9E221BC6BC8371A9ULL, "vxht_chantry_2t2_0_0.vlm"},
	{0x9E2220BF6F4C9833ULL, "962728218.ogg"},
	{0x9E226435DB29AA42ULL, "hlac_ca_caverns1_24426_5z.dds"},
	{0x9E229A5939DE0019ULL, "100344531.ogg"},
	{0x9E22C9783C0FFF3AULL, "vxht_mansion2b_2bw_0_0.vlm"},
	{0x9E23A17EAC8C84A1ULL, "cd_one110_gustav.bnk"},
	{0x9E23DDCBC1D52A06ULL, "vxht_keep_n_5jw_0_0.vlm"},
	{0x9E23EC622FB5EAD7ULL, "vxca_caverns2_8s0_0_0.vlm"},
	{0x9E24C6D0B44692BEULL, "vxht_chantry_n_2ow_0_0.vlm"},
	{0x9E2566E39739F0E7ULL, "169831926.ogg"},
	{0x9E25890430EB31DCULL, "gen00fl_isabela_init6077461.cl"},
	{0x9E258CCFC66606F9ULL, "vxw_encounter1_fn_0_0.vlm"},
	{0x9E25B2FC24830446ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_4x.dds"},
	{0x9E260CF0AB6DA1DFULL, "6082118_f.fxe"},
	{0x9E2636EA5DD65830ULL, "vxlt_warehouse_b9_0_0.vlm"},
	{0x9E26A059689ADAB2ULL, "6088864_m.fxe"},
	{0x9E26FBF52CFB9057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_6.gad"},
	{0x9E270C042A588171ULL, "w_bonepitmining_a2_76_0.msh"},
	{0x9E271A042A58988CULL, "w_bonepitmining_a2_76_0.mao"},
	{0x9E2726042A58AD2FULL, "w_bonepitmining_a2_76_0.mmh"},
	{0x9E27AE35B062DA3EULL, "gen_im_flw_two_fenris_02.uti"},
	{0x9E27E60533182BBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ce_transition.ani"},
	{0x9E281D341A02B3C3ULL, "26703650.ogg"},
	{0x9E28371511EB8806ULL, "309862390.ogg"},
	{0x9E28B8F442C41E7DULL, "vis121_saemus6118462.cl"},
	{0x9E28BE281ADDC709ULL, "vxdr_0_deeproads_cuf_0_1.vlm"},
	{0x9E28D8AB7D0242DCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovemid.mao"},
	{0x9E28E4AB7D02577FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovemid.mmh"},
	{0x9E28EAAB7D026121ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovemid.msh"},
	{0x9E295639C2F439A1ULL, "plt_lgt000pt_ultra_2_04.plo"},
	{0x9E297004B4B738BDULL, "6128000_m.fxe"},
	{0x9E297C6AB05788C8ULL, "l_mer241ar_cavern_of_dead.bnk"},
	{0x9E29ECD41A9B16E3ULL, "590849697.ogg"},
	{0x9E2A54ADA2A8C446ULL, "plt_sib100pt_birthright.plo"},
	{0x9E2AFCE722E2B7DAULL, "vxht_0_hightown_a3_bek_0_0.vlm"},
	{0x9E2B5AACCF508E38ULL, "vxlt_0_lowtown_docks_c03_0_0.vlm"},
	{0x9E2C214639F63C3FULL, "qun211_party_comment6179256.cl"},
	{0x9E2C23124BB2A70CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hanginglight_sa.dds"},
	{0x9E2C50C22063AD39ULL, "vxht_0_hightown_a3_a88_0_0.vlm"},
	{0x9E2C82EFC695C29DULL, "vxht_mansion_n_6jk_0_0.vlm"},
	{0x9E2C9770BCF51292ULL, "ch2_s2.psh.fxo"},
	{0x9E2D0AC82660C5D5ULL, "and100_anders6114232.cl"},
	{0x9E2D13A5DDEE09C6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rail8m_1.mao"},
	{0x9E2DB859570E085DULL, "6139328_f.fxe"},
	{0x9E2DC88A2968C161ULL, "6205703_m.fxe"},
	{0x9E2DCE0AC3C55ABEULL, "6128866_f.fxe"},
	{0x9E2E5552FC592C2EULL, "6167489_f.fxe"},
	{0x9E2E61C960FD2C65ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobe_0s.dds"},
	{0x9E2E7D1FB91A7D16ULL, "1016615608.ogg"},
	{0x9E2EC66D7F72A794ULL, "vxca_caverns2prc_8kv_0_0.vlm"},
	{0x9E2F435D4ACD5913ULL, "vxlt_warehouse_vr_0_0.vlm"},
	{0x9E2F867CE10F6BB6ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk01.phy"},
	{0x9E2FFBE7683C8B4BULL, "gen_im_trs_tmp_hvy_cht_r0.uti"},
	{0x9E308CB17A2A3AB1ULL, "6092193_f.fxe"},
	{0x9E308DA3AE798A29ULL, "vxht_chantry_2q1_0_0.vlm"},
	{0x9E31D27DB2E67BCFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_lp.gad"},
	{0x9E31E57ABD7863D9ULL, "vxw_bonepitmining_a2_1yc_0_0.vlm"},
	{0x9E3222CF97378D40ULL, "6097044_m.fxe"},
	{0x9E32A28C8805BD6EULL, "689607419.ogg"},
	{0x9E32B62D7979E2CEULL, "121321790.ogg"},
	{0x9E32D78D6ABFBED5ULL, "hlac_lt_undercitydungeon_-1_18y.dds"},
	{0x9E3301B60E40C4B0ULL, "247117345.ogg"},
	{0x9E3390A793ECAABDULL, "mag100tr_eau_wilmod.utt"},
	{0x9E33FEACB85C1F79ULL, "vxw_encounter1_km_0_0.vlm"},
	{0x9E344DCAED7D5DEDULL, "vxlt_darktown_5cs_0_0.vlm"},
	{0x9E345ABF1D6548D2ULL, "plt_lgt160pt_cod_avern_lett.nss"},
	{0x9E34ABF3162CEEA0ULL, "follower_banter6141694.cl"},
	{0x9E34BA80A245453AULL, "lt_0_lowtown_n_36_0.anb"},
	{0x9E3512365A90F332ULL, "6083964_m.fxe"},
	{0x9E355B3293DF8B88ULL, "118435176.ogg"},
	{0x9E35EE1959BFA4BDULL, "hlac_ht_0_hightown_n_65533_3z.dds"},
	{0x9E36899DAC9A72FCULL, "vxht_bartrand_n_6pw_0_0.vlm"},
	{0x9E3697891254D566ULL, "crt_desire_demon.bnk"},
	{0x9E36CCF533EE2CD5ULL, "199178033.ogg"},
	{0x9E36F5E0B4F59E22ULL, "6116095_m.fxe"},
	{0x9E36F689F8506D9AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupa.mmh"},
	{0x9E36FA89F85074D1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupa.mao"},
	{0x9E370889F8508CE0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupa.msh"},
	{0x9E379A58558D926DULL, "746425535.ogg"},
	{0x9E37B9E36806E997ULL, "vxlt_darktown_67z_0_0.vlm"},
	{0x9E37CB9E3EC5584FULL, "dae021_male_worshipper_sta6141155.cl"},
	{0x9E384A803957CB30ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_basic.msh"},
	{0x9E3858803957E34AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_basic.mmh"},
	{0x9E385C803957E981ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_basic.mao"},
	{0x9E388468B757EE69ULL, "follower_banter6167190.cl"},
	{0x9E38A9916E8C2A1AULL, "vxlt_0_lowtown_docks_a3_n_bbb_0_0.vlm"},
	{0x9E38BB4FD3D42C93ULL, "6122105_m.fxe"},
	{0x9E38E77DB2ECBE7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_lp.evt"},
	{0x9E38F8E55B894B3AULL, "500993804.ogg"},
	{0x9E396F7CE1179957ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk01.msh"},
	{0x9E39857CE117BEC9ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk01.mmh"},
	{0x9E3A084CC37174DFULL, "6122224_m.fxe"},
	{0x9E3A0F67942CCA67ULL, "vxca_caverns1_ijy_0_0.vlm"},
	{0x9E3A47ED8326FB8FULL, "6175817_m.fxe"},
	{0x9E3A492B13325B56ULL, "hlac_lt_shop_-1_8x.dds"},
	{0x9E3A4D37E7BAF14DULL, "hlac_ca_cavern1_haunted_24424_3x.dds"},
	{0x9E3AA539325DC7E8ULL, "w_encounter1_7_0.mmh"},
	{0x9E3AB139325DDB8BULL, "w_encounter1_7_0.mao"},
	{0x9E3AC339325DFA12ULL, "w_encounter1_7_0.msh"},
	{0x9E3AE7D08ABB1EAFULL, "zz_mag_debug6102189.cl"},
	{0x9E3AF4CCB88BDB1DULL, "vxht_mansion_n_6mj_0_0.vlm"},
	{0x9E3B05B0A266DF96ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_showcase.msh"},
	{0x9E3B0FB0A266F094ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_showcase.mmh"},
	{0x9E3B13B0A266F7A7ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_showcase.mao"},
	{0x9E3B3BFC0520E9DEULL, "gen00fl_varric6105749.cl"},
	{0x9E3B8E5D5D64188EULL, "189862998.ogg"},
	{0x9E3C02C9A04BA2BCULL, "91616319.ogg"},
	{0x9E3C21838FB5CFD4ULL, "ht_0_hightown_18_0_lc.mmh"},
	{0x9E3C28F52D0E38DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_6.ani"},
	{0x9E3C32F2A0D28086ULL, "6164568_m.fxe"},
	{0x9E3C6588E9911B68ULL, "llac_lt_gallowstemplar_f_2383_12.dds"},
	{0x9E3CBEF7D1026304ULL, "136179283.ogg"},
	{0x9E3D711CDE57D80FULL, "vxlt_hangedman_25q_0_0.vlm"},
	{0x9E3D7F1DEEA7438FULL, "hlac_dr_primevaltomb_9621_2y.dds"},
	{0x9E3D9F2562F21A32ULL, "invalid_move.cur"},
	{0x9E3DAF47571B055EULL, "vxlt_undercitydungeon_1yi_0_0.vlm"},
	{0x9E3DBED51240EAD4ULL, "pro000cr_merchant.utc"},
	{0x9E3EAF538CF38765ULL, "hlac_lt_0_lowtown_a2_n_-3_3x.dds"},
	{0x9E3EF09F8DEE1E08ULL, "hlac_w_bonepitdestroyed_a3_262145_0z.dds"},
	{0x9E3F227B4F639CA6ULL, "486355573.ogg"},
	{0x9E3F794E3E07771AULL, "994536864.ogg"},
	{0x9E3FCEE99CB7BFA6ULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressurepusheddown.ani"},
	{0x9E3FD0FF25EEA238ULL, "ico_potion_lifeward.dds"},
	{0x9E405757AF6E7C59ULL, "vxw_bonepitmining_a2_1vb_0_0.vlm"},
	{0x9E40A2BFA9E4E4A4ULL, "16996937.ogg"},
	{0x9E40B89CC016406CULL, "vxw_bonepit_13v_0_0.vlm"},
	{0x9E415DAABFD75F22ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_dssta_0n.dds"},
	{0x9E41629232510791ULL, "w_woundedcoast_255_0.msh"},
	{0x9E41709232511FACULL, "w_woundedcoast_255_0.mao"},
	{0x9E417C923251344FULL, "w_woundedcoast_255_0.mmh"},
	{0x9E41D9402421C411ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_011a_0t.dds"},
	{0x9E424FBFEC90319CULL, "vxw_bonepitmining_a2_21i_0_0.vlm"},
	{0x9E425D3B57FB2574ULL, "vxlt_0_lowtown_docks_a3_n_cya_0_1.vlm"},
	{0x9E42E726A7407D7AULL, "llac_lt_undercityencounter_-1_4.dds"},
	{0x9E431EB23D8D6198ULL, "zz_followers_banter6158874.cl"},
	{0x9E434F548369631FULL, "15526466.ogg"},
	{0x9E435E26EF51D6B3ULL, "147154227.ogg"},
	{0x9E43F3ED002857DBULL, "hlac_w_sundermount_-262140_0x.dds"},
	{0x9E4424AC6EFDA829ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugthrone.phy"},
	{0x9E44331E54596E5BULL, "el_lr_new_day2.mmh"},
	{0x9E451549E4DA4BB9ULL, "6120462_m.fxe"},
	{0x9E453475ABDF33E8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovetop.mmh"},
	{0x9E453DD7DE734E22ULL, "97422273.ogg"},
	{0x9E454075ABDF478BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovetop.mao"},
	{0x9E455275ABDF6612ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovetop.msh"},
	{0x9E4593D97E807E04ULL, "static_static_p5.vsh.fxo.dx11"},
	{0x9E45EF7DB2F75595ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_lp.ani"},
	{0x9E46077AE6BA5924ULL, "vxw_encounter2_31_0_0.vlm"},
	{0x9E46490617B32259ULL, "llac_w_bonepitmining_a2_262144_0.dds"},
	{0x9E4744C89CB5DA92ULL, "vxht_0_hightown_a2_a73_0_0.vlm"},
	{0x9E47583C9F47F265ULL, "vxht_0_hightown_a3_bup_0_0.vlm"},
	{0x9E476930AAF35DFCULL, "w_bonepitmining_a2_22_0_lc.mmh"},
	{0x9E47A410A41EE383ULL, "hlac_lt_0_lowtown_a2_n_c_65536_3y.dds"},
	{0x9E47DD53D44B8E90ULL, "art\\levels\\ztools\\default\\generic\\0ff_basic.phy"},
	{0x9E47ED1CE4BD906BULL, "hlac_lt_hangedman_-1_6x.dds"},
	{0x9E47F013070D3644ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock2a.msh"},
	{0x9E480213070D54ADULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock2a.mao"},
	{0x9E480613070D5BE6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock2a.mmh"},
	{0x9E48473C4479E371ULL, "zz_mcr_debug6080789.cl"},
	{0x9E48553A1F5DFA68ULL, "dae000_bodhan6146909.cl"},
	{0x9E4881448622E2E7ULL, "vxca_caverns1_imx_0_0.vlm"},
	{0x9E48AA52EAE4F391ULL, "mag110_samson6018711.cl"},
	{0x9E48BA356715FC48ULL, "394329417.ogg"},
	{0x9E490CC27B9180C5ULL, "gen00fl_anders6112993.cl"},
	{0x9E49AFBE586A5351ULL, "6060164_m.fxe"},
	{0x9E49E10DE502B913ULL, "follower_banter6163145.cl"},
	{0x9E49FDDBC01FCBE0ULL, "vxca_cavern1_haunted_foc_0_0.vlm"},
	{0x9E4A20ED138044CAULL, "798547178.ogg"},
	{0x9E4A881BA7B595C0ULL, "llac_lt_0_lowtown_a3_n_c_-65535_0.dds"},
	{0x9E4AB11131FAD769ULL, "6036699_m.fxe"},
	{0x9E4BE2F9D04DF08FULL, "vxlt_hangedman_22p_0_0.vlm"},
	{0x9E4C327AC39EDE5CULL, "art\\levels\\races\\natural\\proxy\\foliage\\_textures\\medbush_dp.dds"},
	{0x9E4C36AD4223FEF5ULL, "vxlt_0_lowtown_docks_n_b5c_0_0.vlm"},
	{0x9E4CC26377AFD7BEULL, "183332992.ogg"},
	{0x9E4CDE252732EA28ULL, "vxlt_gallowstemplar_f_3dr_0_0.vlm"},
	{0x9E4CE0DE881447ADULL, "vxw_blightlands_12a_0_0.vlm"},
	{0x9E4D509506D0806CULL, "mrl250cr_skeleton_ice.utc"},
	{0x9E4D82DED7B22D1AULL, "hlac_lt_gallowscourtyard_-65537_1x.dds"},
	{0x9E4D94B21CE388ABULL, "669402398.ogg"},
	{0x9E4DC99A857F25F3ULL, "1055562779.ogg"},
	{0x9E4DEB7CCCCCA8B4ULL, "gen_im_trs_krk_junk1_04.uti"},
	{0x9E4E0D4075DEA22EULL, "vxht_0_hightown_a2_c44_0_0.vlm"},
	{0x9E4E5234579599FAULL, "67935723.ogg"},
	{0x9E4E6E6AB0768D72ULL, "l_mer241ar_cavern_of_dead.wbd"},
	{0x9E4EC934A16494D9ULL, "vxw_bonepitmining_a2_1sa_0_0.vlm"},
	{0x9E501C98A6017654ULL, "gen00fl_isabela6099352.cl"},
	{0x9E513184D1698EEDULL, "vxlt_darktown_5iq_0_0.vlm"},
	{0x9E521A7F5E4EE4E2ULL, "qun221_petrice6185136.cl"},
	{0x9E5239948FB4BB88ULL, "vxht_chantry_337_0_0.vlm"},
	{0x9E529A8AB08E228DULL, "385567555.ogg"},
	{0x9E52FADD98ABF22DULL, "522690469.ogg"},
	{0x9E53981929BD8C03ULL, "hlac_ht_chantry_2534_0x.dds"},
	{0x9E53B7E99CC841A0ULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressurepusheddown.gad"},
	{0x9E53BFC524EE6D63ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.dura.evt"},
	{0x9E53E4C27B96DBF2ULL, "578620730.ogg"},
	{0x9E5402984F7CE6EDULL, "6164989_m.fxe"},
	{0x9E544E881D6A026FULL, "vxht_0_hightown_n_c7s_0_0.vlm"},
	{0x9E547BD2A4DD0C78ULL, "hlac_lt_0_lowtown_n_2_0x.dds"},
	{0x9E54AD0E64E51E63ULL, "hlac_lt_0_lowtown_docks_a3_-262144_2x.dds"},
	{0x9E5629698DE2C0BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_chair.anb"},
	{0x9E56709495190377ULL, "31540569.ogg"},
	{0x9E567D165861FE5CULL, "823759745.ogg"},
	{0x9E57A9235145041BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrb_0s.dds"},
	{0x9E57DFB245084D9EULL, "vxht_0_hightown_a3_bm7_0_0.vlm"},
	{0x9E58872699A005E2ULL, "art\\characters\\creatures\\textures\\cn_gut_0d.dds"},
	{0x9E58EF3732765700ULL, "hlac_ht_0_hightown_n_-1_0z.dds"},
	{0x9E5936EA7C256A44ULL, "78395480.ogg"},
	{0x9E59B69A4AB57725ULL, "gen00fl_aveline6132725.cl"},
	{0x9E59D64B3C1D96FFULL, "vxlt_hangedman_264_0_0.vlm"},
	{0x9E59E5DE0F6AB3B0ULL, "453646718.ogg"},
	{0x9E5A08DFC06047E5ULL, "lgt150_nat_were_supporter6110962.cl"},
	{0x9E5A3AF9A189420DULL, "810222530.ogg"},
	{0x9E5A3F04FAF1CDEEULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernlavaedge_d.dds"},
	{0x9E5A66E0F0B0A7BCULL, "6167949_m.fxe"},
	{0x9E5A6ADA8273B7A1ULL, "vxlt_gallowstemplar_4xw_0_0.vlm"},
	{0x9E5A7F2ACF606BECULL, "vxw_woundedcoast_1b_0_0.vlm"},
	{0x9E5A98C524F449E4ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.dura.gad"},
	{0x9E5B4D985C997467ULL, "723393264.ogg"},
	{0x9E5BAB30B869D05FULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinstower.msh"},
	{0x9E5BB130B869DA01ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinstower.mmh"},
	{0x9E5BBD30B869EEA2ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinstower.mao"},
	{0x9E5C6D85ABAAD3A1ULL, "dae000_gen_player_turn_in6177845.cl"},
	{0x9E5C75D156A70420ULL, "gen00fl_merrill6128711.cl"},
	{0x9E5C78BB99811384ULL, "417306748.ogg"},
	{0x9E5C7F1D67D4BAAEULL, "vxht_0_hightown_a2_c35_0_0.vlm"},
	{0x9E5CDCB9976C3E24ULL, "230188350.ogg"},
	{0x9E5CEED176B5425BULL, "6065886_m.fxe"},
	{0x9E5D5346B7B5CE54ULL, "hlac_lt_0_lowtown_a3_n_c_131071_5y.dds"},
	{0x9E5DCDD2DCD8E81BULL, "124097840.ogg"},
	{0x9E5E291BBE2F2145ULL, "814859592.ogg"},
	{0x9E5E80257BAA0405ULL, "10830331.ogg"},
	{0x9E5E910C89CA2A0CULL, "vxlt_hangedman_lj_0_1.vlm"},
	{0x9E5EB07A6B5A5259ULL, "llac_ca_cavern1_haunted_24421_6.dds"},
	{0x9E5ED82DB6DD0DF1ULL, "vxht_mansion2b_29q_0_0.vlm"},
	{0x9E5FA361C35FA76DULL, "vxlt_darktown_5rp_0_0.vlm"},
	{0x9E604812B82CDDBBULL, "ht_0_hightown_25_0_ll.mmh"},
	{0x9E605E567482E3CDULL, "qun110_petrice6042295.cl"},
	{0x9E606EE310B9B0C9ULL, "hlac_w_encounter3_0_0x.dds"},
	{0x9E60A490A6FA7D76ULL, "272962040.ogg"},
	{0x9E60EE0F9D3D8722ULL, "vxht_0_hightown_dw_0_0.vlm"},
	{0x9E6124697B84A069ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_grcleric_chan_ex3.gad"},
	{0x9E61372F55AB7346ULL, "219897756.ogg"},
	{0x9E6173C31E7F23B3ULL, "l_env_wal_hightown_int_brothel_bed.bnk"},
	{0x9E61A7C524FA81C2ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.dura.ani"},
	{0x9E61BBBAA4DC3739ULL, "hlac_ht_0_hightown_65535_2z.dds"},
	{0x9E61BC59D4E9CE7BULL, "dae000_arishok6070879.cl"},
	{0x9E61E341925EF97EULL, "vxw_encounter2_camp_x_0_0.vlm"},
	{0x9E61E97C2003F13CULL, "6147565_m.fxe"},
	{0x9E61F0BA83A7AFCCULL, "hlac_ht_bartrand_n_3370_0z.dds"},
	{0x9E621A285026C7FEULL, "737577110.ogg"},
	{0x9E62A578F00FF474ULL, "hlac_ht_chantry_2534_14z.dds"},
	{0x9E6328B56C802511ULL, "6082344_m.fxe"},
	{0x9E633CF9420EDB18ULL, "6085278_m.fxe"},
	{0x9E6376C9F0304A7DULL, "456355025.ogg"},
	{0x9E63E05C9A6D930DULL, "623295943.ogg"},
	{0x9E65340E1DBC7CA5ULL, "gen00fl_bethany6097876.cl"},
	{0x9E654192AE143F12ULL, "zz_mage2.cut"},
	{0x9E655E8ED1800E0FULL, "gen00fl_varric6087466.cl"},
	{0x9E658DFBFF8FE9FCULL, "dae000_orana6149105.cl"},
	{0x9E65B6B1E1331A3FULL, "454237743.ogg"},
	{0x9E65BD33E04D5A41ULL, "333158015.ogg"},
	{0x9E65F1559D4A007EULL, "6002516_m.fxe"},
	{0x9E6608214D8655FEULL, "vxw_bonepit_29u_0_0.vlm"},
	{0x9E6632043076C035ULL, "vxw_bonepitmining_a2_1wh_0_1.vlm"},
	{0x9E663C47CDF6ACF1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1d.mmh"},
	{0x9E664847CDF6C092ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1d.mao"},
	{0x9E665647CDF6D8AFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1d.msh"},
	{0x9E66772CD9F049BEULL, "709951351.ogg"},
	{0x9E66D2C19BB5E374ULL, "786839724.ogg"},
	{0x9E671A2DA1BB95D0ULL, "643438143.ogg"},
	{0x9E6785DF45C4024BULL, "6144591-mrl_pt3_end2.cut"},
	{0x9E67BD80A270A466ULL, "lt_0_lowtown_n_36_0.rml"},
	{0x9E67F5697B8A6EB7ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_grcleric_chan_ex3.ani"},
	{0x9E683F0DC4A7B3DDULL, "lt_0_lowtown_docks_a3_n_116_0.mao"},
	{0x9E68430DC4A7BA16ULL, "lt_0_lowtown_docks_a3_n_116_0.mmh"},
	{0x9E6848282E13AF7FULL, "vxlt_hangedman_237_0_0.vlm"},
	{0x9E684D0DC4A7CB14ULL, "lt_0_lowtown_docks_a3_n_116_0.msh"},
	{0x9E687C6F6B27D6D3ULL, "6161287_m.fxe"},
	{0x9E68E0610689851AULL, "vxht_bartrand_n_5sc_0_0.vlm"},
	{0x9E68EDE0F344364EULL, "dre000_dougal6129734.cl"},
	{0x9E695A9414D6A14FULL, "gen00fl_aveline6139700.cl"},
	{0x9E698D0623BF71B9ULL, "618469131.ogg"},
	{0x9E69C1F670F0EB40ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgtb_0d.dds"},
	{0x9E6A0FB9768797C8ULL, "653249414.ogg"},
	{0x9E6AB71CD6E109EEULL, "vxlt_gallowscourtyard_n_70n_0_0.vlm"},
	{0x9E6B74E90CBDFBA3ULL, "llac_lt_0_lowtown_a2_131071_3.dds"},
	{0x9E6BAF7EACCA5253ULL, "cd_one110_gustav.wbd"},
	{0x9E6BE620393CE32EULL, "700844935.ogg"},
	{0x9E6CC961BCFC17D4ULL, "vxw_bonepitmining_a2_f4_0_0.vlm"},
	{0x9E6D4A0AA8D32671ULL, "vxht_mansion2b_22p_0_0.vlm"},
	{0x9E6DF03E1B20CCC8ULL, "plt_one200pt_herb_blush.plo"},
	{0x9E6DF9AB0E63A575ULL, "vxca_caverns2_964_0_0.vlm"},
	{0x9E6DFA9D4CD75446ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_chac.phy"},
	{0x9E6DFBDC1355D201ULL, "489762041.ogg"},
	{0x9E6E41987F04E5A3ULL, "547617414.ogg"},
	{0x9E6FAD2D96A574F2ULL, "19577902.ogg"},
	{0x9E700DAC6F22BC4EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugthrone.mmh"},
	{0x9E701855442A88DAULL, "hlac_lt_gallowscourtyard_-65536_2z.dds"},
	{0x9E7019AC6F22D0E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugthrone.mao"},
	{0x9E7027AC6F22E83CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugthrone.msh"},
	{0x9E70375E0171E3AAULL, "300384659.ogg"},
	{0x9E70CA15A0FB2986ULL, "6149056_f.fxe"},
	{0x9E719A925E763D91ULL, "6085244_m.fxe"},
	{0x9E71A8D98D1E3DBBULL, "vxlt_0_lowtown_docks_a3_n_d58_0_0.vlm"},
	{0x9E71C39E5B373BAEULL, "987769718.ogg"},
	{0x9E71E8B11D8FF699ULL, "258459499.ogg"},
	{0x9E721BC48D664DFEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.impa.evt"},
	{0x9E72220C57C950EAULL, "6177844_f.fxe"},
	{0x9E730F21E2BF576BULL, "hlac_ht_mansion2_818_50x.dds"},
	{0x9E7313CFEECF01CBULL, "231274057.ogg"},
	{0x9E7435630CF3FF4EULL, "gen00fl_fenris6126966.cl"},
	{0x9E743D8C7F3B0F86ULL, "6137520_m.fxe"},
	{0x9E7463328F4018FEULL, "6005416_m.fxe"},
	{0x9E7535ADA2E87A85ULL, "plt_sib100pt_birthright.nss"},
	{0x9E758E6AA98D011DULL, "lgt251tr_nexus_fight.ncs"},
	{0x9E75C8A671B11251ULL, "qun110_petrice6150075.cl"},
	{0x9E7630F2F7403950ULL, "lt_0_lowtown_docks_n_253_0.mao"},
	{0x9E7634F2F7404063ULL, "lt_0_lowtown_docks_n_253_0.mmh"},
	{0x9E7642F2F74057BDULL, "lt_0_lowtown_docks_n_253_0.msh"},
	{0x9E777DF8ECB6B544ULL, "1067141417.ogg"},
	{0x9E779989128C6F0CULL, "crt_desire_demon.wbd"},
	{0x9E77BC7EE734DCEAULL, "atl_saveload_dxt5_00.dds"},
	{0x9E782DA73E9A0953ULL, "432507446.ogg"},
	{0x9E786AD38F4DFA34ULL, "vxht_keep_5gq_0_0.vlm"},
	{0x9E788E3FD7A1E35AULL, "gen00fl_fenris6113834.cl"},
	{0x9E78B61EFF55B6F7ULL, "6075137_m.fxe"},
	{0x9E7932C48D6C9329ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.impa.gad"},
	{0x9E794486C92F0808ULL, "6041054_m.fxe"},
	{0x9E794A040B4838FAULL, "gen00fl_anders6112916.cl"},
	{0x9E7964ECF75BA459ULL, "arl101cr_undead.utc"},
	{0x9E79E21DB6807B3FULL, "vxw_bonepit_14x_0_0.vlm"},
	{0x9E7BC862A4038C89ULL, "vxht_chantry_2ve_0_1.vlm"},
	{0x9E7BD3DA7707D67BULL, "vxlt_backalley_n_1u0_0_0.vlm"},
	{0x9E7C28395E5E73D8ULL, "6086169_m.fxe"},
	{0x9E7C7CC8A4011249ULL, "art\\vfx\\combat\\vfx_str_spit2_aoe_p.impa.ani"},
	{0x9E7C9A88560E6370ULL, "hlac_w_encounter2_camp_-131068_0x.dds"},
	{0x9E7CB80EEE60BC10ULL, "220302069.ogg"},
	{0x9E7D029BD8187E53ULL, "172214855.ogg"},
	{0x9E7D91E3021138EAULL, "hm_gencr_city_guard_archer.mor"},
	{0x9E7DA97572B831B5ULL, "6104536_m.fxe"},
	{0x9E7DDC9B1277BB6EULL, "6116868_m.fxe"},
	{0x9E7E33337BA3752BULL, "hlac_ca_cavern1_haunted_24424_26y.dds"},
	{0x9E7E3FA85082A116ULL, "hlac_ht_keep_n_3732_4y.dds"},
	{0x9E7E5B7AD867FC1EULL, "hlac_lt_0_lowtown_n_2_1y.dds"},
	{0x9E7E6E948667908DULL, "lt_0_lowtown_n_52_0_lc.mmh"},
	{0x9E7E8DD23DD4D1FBULL, "vxw_blightlands_13h_0_0.vlm"},
	{0x9E7EC59EF06C3BDCULL, "6095391_f.fxe"},
	{0x9E7EFC358F85CF16ULL, "159138830.ogg"},
	{0x9E7F9BD3D965074AULL, "vrc321_housekeeper6095657.cl"},
	{0x9E7FE4551C8C8D06ULL, "vxw_bonepit_27g_0_0.vlm"},
	{0x9E8003C48D726277ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.impa.ani"},
	{0x9E80084668776ECAULL, "vxlt_0_lowtown_docks_n_ctb_0_0.vlm"},
	{0x9E8049418EF3C09BULL, "rog_decoy.dds"},
	{0x9E808DDCCE8A6017ULL, "405301039.ogg"},
	{0x9E8094DAD7DC1100ULL, "hlac_ht_mansion2_818_20x.dds"},
	{0x9E8100DE54D8B722ULL, "dae300_petrice6097903.cl"},
	{0x9E81448AEB18AFA4ULL, "vxw_sundermount_2bm_0_0.vlm"},
	{0x9E81A388A8691314ULL, "art\\characters\\creatures\\strider\\animation\\die_loop_v1.gad"},
	{0x9E81A5DC5AC4DDAFULL, "qun221_petrice6077973.cl"},
	{0x9E81D845EECC2373ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldb_n.dds"},
	{0x9E822030D6EB773CULL, "mag110_ws_banter6079652.cl"},
	{0x9E8297C8A405ABCBULL, "art\\vfx\\combat\\vfx_str_spit2_aoe_p.impa.gad"},
	{0x9E82B314F84D8F05ULL, "vxdr_primevaltomb_50b_0_2.vlm"},
	{0x9E82DD279801C70EULL, "vrc321tr_door_shift_1.ncs"},
	{0x9E82DFAF54305B13ULL, "war_magicresistance.dds"},
	{0x9E833563E4803EE4ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_fire02.mao"},
	{0x9E834816C9C2FBADULL, "464353280.ogg"},
	{0x9E841AA17FA2BB40ULL, "vxca_cavern1_haunted_ihu_0_0.vlm"},
	{0x9E845B92971EB62DULL, "vxlt_0_lowtown_docks_a3_d19_0_0.vlm"},
	{0x9E84833DFB342228ULL, "vxlt_gallowstemplar_n_34z_0_0.vlm"},
	{0x9E84A0F8929EFD42ULL, "260396945.ogg"},
	{0x9E85438B093856D2ULL, "llac_ht_chantry_2534_52.dds"},
	{0x9E854E58928AA1ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_down.ani"},
	{0x9E855AD570A0E166ULL, "ser191_emeric6108132.cl"},
	{0x9E861FAEEEB0A0C1ULL, "dae000tr_display_district.ncs"},
	{0x9E867B50632859FCULL, "gen00fl_isabela6079013.cl"},
	{0x9E86AE59615088ECULL, "hlac_ht_mansion2_818_13z.dds"},
	{0x9E86DCB0814412B4ULL, "vxht_keep_5jv_0_0.vlm"},
	{0x9E872FEE2D922E84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_escape_human_female.ani"},
	{0x9E87B230B88F177EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinstower.phy"},
	{0x9E87B8197A40A717ULL, "259916753.ogg"},
	{0x9E87E63DA6FCE2A9ULL, "317232131.ogg"},
	{0x9E88B288A86F4BF2ULL, "art\\characters\\creatures\\strider\\animation\\die_loop_v1.ani"},
	{0x9E88C2BA3B5ADE35ULL, "6173400_m.fxe"},
	{0x9E8919C6C2D21313ULL, "vxlt_blackemporium_15r_0_0.vlm"},
	{0x9E891F059D6B5B6EULL, "vxca_caverns1_mines_jgq_0_0.vlm"},
	{0x9E893B2ECD35B03AULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_w_5p_ex.gad"},
	{0x9E898E1190E90E9AULL, "269629934.ogg"},
	{0x9E8994C8A40BC530ULL, "art\\vfx\\combat\\vfx_str_spit2_aoe_p.impa.evt"},
	{0x9E89D755CB0CD226ULL, "one200it_tattoo.uti"},
	{0x9E89F1DF9EF25E68ULL, "vxlt_0_lowtown_n_e83_0_0.vlm"},
	{0x9E8A1CF8FECB34A1ULL, "w_bonepit_136_0.phy"},
	{0x9E8A517EA339ED0EULL, "vxht_chantry_2th_0_1.vlm"},
	{0x9E8A841A0B770E53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_reaction_left_shoulder.ani"},
	{0x9E8AD31443204A7CULL, "vxca_caverns2_78w_0_2.vlm"},
	{0x9E8BA5941EC9B8F4ULL, "vxht_mansion_n_2lt_0_0.vlm"},
	{0x9E8BC39777B2E736ULL, "art\\vfx\\models\\vfx_rockarmor_lt.mmh"},
	{0x9E8BC54AC9C8815EULL, "art\\characters\\weapons\\dagger\\w_dgr_106a.mao"},
	{0x9E8BCD9777B2F834ULL, "art\\vfx\\models\\vfx_rockarmor_lt.msh"},
	{0x9E8BEDAB7D56E3C4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_alcovemid.phy"},
	{0x9E8BF63E80E69113ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potc_0.pwk"},
	{0x9E8C5B589290D6EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_down.gad"},
	{0x9E8C9A8C2B4164A2ULL, "541824956.ogg"},
	{0x9E8CE3E3CE9B9A87ULL, "167762464.ogg"},
	{0x9E8D058F399B2E12ULL, "vxlt_undercityencounter_v2_0_0.vlm"},
	{0x9E8D2C0D919F2DF8ULL, "918561170.ogg"},
	{0x9E8D7B9863FADADEULL, "141477471.ogg"},
	{0x9E8D7ECBE6495043ULL, "vxlt_darktown_58p_0_0.vlm"},
	{0x9E8D84C1D3592FB6ULL, "isa111_isabela6067065.cl"},
	{0x9E8DB219A1875F2CULL, "hlac_lt_0_lowtown_n_131072_1y.dds"},
	{0x9E8DB57041CFDDD2ULL, "lla_ht_0_hightown_n_-2_0.dds"},
	{0x9E8E2CEE2D9848E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_escape_human_female.gad"},
	{0x9E8E6285C2098B3BULL, "hlac_w_bonepitmining_a2_458758_0x.dds"},
	{0x9E8E7A235A6D874AULL, "vxlt_0_lowtown_docks_n_coc_0_0.vlm"},
	{0x9E8F178DBECDE3E0ULL, "6132723_m.fxe"},
	{0x9E8F1E8CEA32B26FULL, "vxht_0_hightown_a2_c1f_0_0.vlm"},
	{0x9E8F7E5F1FF1248DULL, "620515167.ogg"},
	{0x9E8F922EC5922401ULL, "6136476_m.fxe"},
	{0x9E8FD0D2757C0CA5ULL, "combat_deathblow_04.gda"},
	{0x9E8FEAE2C0335AF0ULL, "vxw_bonepit_nb_0_0.vlm"},
	{0x9E8FED3E1B3DAD87ULL, "plt_one200pt_herb_blush.nss"},
	{0x9E9029E3F02101A4ULL, "6116534_m.fxe"},
	{0x9E903E2ECD3BD451ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_w_5p_ex.evt"},
	{0x9E905FFA97257225ULL, "one101_orders_followers6115787.cl"},
	{0x9E90B43E995DBF58ULL, "6129150_m.fxe"},
	{0x9E90DF3150E3DB41ULL, "art\\characters\\weapons\\shields\\w_shd_010a_.mao"},
	{0x9E912AEA59ECD370ULL, "vxca_cavern1_haunted_ii6_0_0.vlm"},
	{0x9E9157ECF7702077ULL, "arl101cr_undead.ncs"},
	{0x9E916002474955FDULL, "332169267.ogg"},
	{0x9E91B75D7934DD7FULL, "6041255_m.fxe"},
	{0x9E91BA6C2BF91D7AULL, "vxht_0_hightown_a3_cdi_0_0.vlm"},
	{0x9E923B1D151411A2ULL, "lt_0_lowtown_docks_n_175_0.msh"},
	{0x9E923D1D15141558ULL, "lt_0_lowtown_docks_n_175_0.mmh"},
	{0x9E92491D151429FBULL, "lt_0_lowtown_docks_n_175_0.mao"},
	{0x9E928453BB0385A5ULL, "vxdr_0_deeproads_a28_0_0.vlm"},
	{0x9E92F1954FBAB13CULL, "var_trigger_station.gda"},
	{0x9E932B6471AF0202ULL, "756761985.ogg"},
	{0x9E932F65D484A241ULL, "143045237.ogg"},
	{0x9E9356589296ECCDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_front_down.evt"},
	{0x9E93875D88C56479ULL, "vxca_caverns1_mines_fna_0_0.vlm"},
	{0x9E93F96259C38898ULL, "6101376_m.fxe"},
	{0x9E9427EE2D9CABC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_escape_human_female.evt"},
	{0x9E944071DBF1D54CULL, "262130099.ogg"},
	{0x9E94436393CF4D17ULL, "6067398_m.fxe"},
	{0x9E945F788F577FBAULL, "ser121_jethann6107098.cl"},
	{0x9E948460C827F79BULL, "vxca_caverns2_979_0_0.vlm"},
	{0x9E948A0E0743831BULL, "6071216_m.fxe"},
	{0x9E949D111A457B37ULL, "830993979.ogg"},
	{0x9E954E8D733A2B34ULL, "vxht_keep_5mw_0_0.vlm"},
	{0x9E95566DC1023308ULL, "6095899_m.fxe"},
	{0x9E95567EF2911CD0ULL, "6167483_f.fxe"},
	{0x9E956DBA7961D190ULL, "6125824_m.fxe"},
	{0x9E95760C13A229AAULL, "324437773.ogg"},
	{0x9E959C1EB3F11279ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_001a_0n.dds"},
	{0x9E95D3DC0A712131ULL, "w_bonepit_96_0.msh"},
	{0x9E95E1DC0A71394CULL, "w_bonepit_96_0.mao"},
	{0x9E95EDDC0A714DEFULL, "w_bonepit_96_0.mmh"},
	{0x9E9684B6A3387BBFULL, "gen00fl_anders.utc"},
	{0x9E96FE0922A41C7FULL, "vxca_caverns1_eia_0_1.vlm"},
	{0x9E970EB0A2B54D8FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_showcase.phy"},
	{0x9E970F3B6CDC0B60ULL, "fen210_hadriana6110591.cl"},
	{0x9E9720EE3DFD0027ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar6_mcr370_playerflees.gad"},
	{0x9E974CA2FC2FAF24ULL, "1035144972.ogg"},
	{0x9E977C1A0B818A8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_reaction_left_shoulder.evt"},
	{0x9E980C29D90A2AB0ULL, "llac_ca_cavern1_haunted_24421_10.dds"},
	{0x9E982788025B07D8ULL, "vxht_mansion2_1t4_0_0.vlm"},
	{0x9E98AD0F862AD651ULL, "ave310_brennan6137227.cl"},
	{0x9E99458C03737252ULL, "hlac_lt_blackemporium_1055_13x.dds"},
	{0x9E994CD3EFCEC083ULL, "359498803.ogg"},
	{0x9E997123062C7797ULL, "148545253.ogg"},
	{0x9E9AF56BD2898003ULL, "vxdr_0_deeproads_cr4_0_0.vlm"},
	{0x9E9AFBC7C65E3363ULL, "854019709.ogg"},
	{0x9E9B24B9CF178736ULL, "549062921.ogg"},
	{0x9E9B3D12630B496BULL, "400994023.ogg"},
	{0x9E9BA03C05C6D437ULL, "vxht_0_hightown_a2_bou_0_0.vlm"},
	{0x9E9BD87E9D4BC340ULL, "dae300_cullen6140631.cl"},
	{0x9E9C45950F8F309AULL, "vxht_keep_4b5_0_0.vlm"},
	{0x9E9CEB93CD1D0C15ULL, "vxca_cavern1_haunted_ijz_0_0.vlm"},
	{0x9E9D562ECD468720ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_w_5p_ex.ani"},
	{0x9E9DBE69E05DF0C0ULL, "853378231.ogg"},
	{0x9E9E3E6B2139C58BULL, "hlac_w_encounter2_1_0z.dds"},
	{0x9E9E5B1A0B8771FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_reaction_left_shoulder.gad"},
	{0x9E9E5CBFB2297370ULL, "vxw_bonepit_ic_0_0.vlm"},
	{0x9E9ECE891CEB208CULL, "341765705.ogg"},
	{0x9E9F01552185E4FBULL, "280913165.ogg"},
	{0x9E9F1199C6D11AC9ULL, "hlac_dr_0_deeproads_5838_1z.dds"},
	{0x9E9F6CFF34773681ULL, "vxlt_0_lowtown_docks_a3_n_czp_0_0.vlm"},
	{0x9E9FB57373C539C5ULL, "gen00fl_merrill6115983.cl"},
	{0x9E9FE0EFFA30DB69ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ri_basicoutcrn.phy"},
	{0x9E9FED4F0C34F78FULL, "6083389_m.fxe"},
	{0x9EA06F4E59E615B0ULL, "7047972.ogg"},
	{0x9EA07B635AF6BEEBULL, "mnd_cycle_f2.dds"},
	{0x9EA0F26BF14D846CULL, "follower_barks6153311.cl"},
	{0x9EA14570EC4CC82FULL, "hlac_lt_gallowscourtyard_-131071_1x.dds"},
	{0x9EA160B75F1FB680ULL, "vxht_mansion2b_3eu_0_0.vlm"},
	{0x9EA186E40B04565DULL, "one000_ilen6057352.cl"},
	{0x9EA2817585466D31ULL, "vxw_encounter1_aw_0_0.vlm"},
	{0x9EA2F63DBA1E101BULL, "vxca_caverns2_988_0_0.vlm"},
	{0x9EA32330FB453B1FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_2.gad"},
	{0x9EA3658E4F454621ULL, "7681861.ogg"},
	{0x9EA42E3E6ADFF11FULL, "931947350.ogg"},
	{0x9EA432A23EF85429ULL, "vxdr_primevaltomb_5b3_0_1.vlm"},
	{0x9EA440AC43145771ULL, "29160931.ogg"},
	{0x9EA4531AF554C753ULL, "zz_vault_debug6178958.cl"},
	{0x9EA46B09E62F6BEFULL, "92635671.ogg"},
	{0x9EA4EC9A45318FC8ULL, "6148425_m.fxe"},
	{0x9EA559712B9C9585ULL, "docks_male_sailor_mob6176656.cl"},
	{0x9EA5ECBFD42557CFULL, "791939108.ogg"},
	{0x9EA64BD205218E1AULL, "lt_0_lowtown_docks_183_0.msh"},
	{0x9EA659D20521A653ULL, "lt_0_lowtown_docks_183_0.mao"},
	{0x9EA65DD20521AD60ULL, "lt_0_lowtown_docks_183_0.mmh"},
	{0x9EA669C90EBC3499ULL, "310733385.ogg"},
	{0x9EA6C4AA7422BE34ULL, "vxht_0_hightown_c59_0_0.vlm"},
	{0x9EA6D8A717E8D584ULL, "one100dg_elren6014248.cl"},
	{0x9EA726DA3CE55F8FULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_act_c.impa.gad"},
	{0x9EA729F9F9132E6AULL, "6141194_m.fxe"},
	{0x9EA74378B7FB3932ULL, "llac_lt_0_lowtown_a2_n_c_65536_1.dds"},
	{0x9EA763DB9D771D64ULL, "mag320_orsino6071209.cl"},
	{0x9EA78C672CFC78EDULL, "362335379.ogg"},
	{0x9EA79E65A68080A3ULL, "gen00fl_isabela6123236.cl"},
	{0x9EA83E85AB742F77ULL, "6163267_m.fxe"},
	{0x9EA84E6431B733F2ULL, "vxht_0_hightown_a2_bwt_0_0.vlm"},
	{0x9EA86913E0BF66C2ULL, "280575366.ogg"},
	{0x9EA891C31EBB5911ULL, "l_env_wal_hightown_int_brothel_bed.wbd"},
	{0x9EA8D89E6F9F16CDULL, "t3_arm_lst.tnt"},
	{0x9EA996BA8BF5E52EULL, "72512858.ogg"},
	{0x9EAA1830FB4B470CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_2.evt"},
	{0x9EAA67E08FD59B39ULL, "mag311_keran6041400.cl"},
	{0x9EAADAE843E746D5ULL, "hlac_lt_0_lowtown_a3_n_c_131072_6x.dds"},
	{0x9EAB5D70BF132495ULL, "vxca_cavern1_haunted_igy_0_0.vlm"},
	{0x9EAB5F975E8A79E9ULL, "gen00fl_varric6087477.cl"},
	{0x9EABD06446158268ULL, "qun131_party_comment6117287.cl"},
	{0x9EAC0B165F141302ULL, "demo_weapon_swap6207142.cl"},
	{0x9EAC0DEE3E0F3C6DULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar6_mcr370_playerflees.ani"},
	{0x9EAC18731FF6F850ULL, "hlac_lt_undercitydungeon_-1_39y.dds"},
	{0x9EAC1992F86D0563ULL, "al_trees_seb.gda"},
	{0x9EAC549304F3B72AULL, "gen00fl_sebastian6174328.cl"},
	{0x9EAC71C899C3440AULL, "llac_ht_keep_n_3732_12.dds"},
	{0x9EACA6D9A26311A8ULL, "vxlt_0_lowtown_docks_n_b05_0_0.vlm"},
	{0x9EACFE00B7B065A5ULL, "zz_vrc.ncs"},
	{0x9EAD1C1245DC1682ULL, "71008383.ogg"},
	{0x9EAD307746E52D6CULL, "hlac_lt_0_lowtown_docks_a3_n_0_5y.dds"},
	{0x9EAD3BDA3CE9EF3CULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_act_c.impa.evt"},
	{0x9EAD5FDB14297BD5ULL, "183875689.ogg"},
	{0x9EAD9E4F23DD8E67ULL, "vxca_caverns1_i2q_0_0.vlm"},
	{0x9EADCC1CE8AC6E06ULL, "vxlt_blackemporium_3u_0_1.vlm"},
	{0x9EAE18E9E8335F55ULL, "dae000_gamlen6048422.cl"},
	{0x9EAE91AEABB9D01CULL, "vxht_mansion2b_2o6_0_0.vlm"},
	{0x9EAEEA7BBD867808ULL, "hlac_w_blightlands_-65536_0y.dds"},
	{0x9EAF062F7E24A6FDULL, "t3_wer_h01.tnt"},
	{0x9EAF1E4DE546AD8EULL, "seb221_brett_harimann6187783.cl"},
	{0x9EAFFABF9A39CD1AULL, "190252496.ogg"},
	{0x9EB0099ED7669651ULL, "follower_banter6161259.cl"},
	{0x9EB024E691A7A9BCULL, "gen00fl_varric6107323.cl"},
	{0x9EB05535BD8A8476ULL, "848963789.ogg"},
	{0x9EB0770C40A8F061ULL, "mrl351cr_fakefemaleelf.utc"},
	{0x9EB08C2CF5B43111ULL, "236398215.ogg"},
	{0x9EB0F352773C85B1ULL, "vxw_encounter1_fv_0_0.vlm"},
	{0x9EB16C89E6B580E7ULL, "mag321_comte_de_launcet6064930.cl"},
	{0x9EB1A8D5A1080AFCULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extplaster.dds"},
	{0x9EB1D85FA866BCF9ULL, "vxlt_hangedman_1by_0_0.vlm"},
	{0x9EB1EF11BDAE7259ULL, "w_sundermount_22_0.msh"},
	{0x9EB1F911BDAE8347ULL, "w_sundermount_22_0.mmh"},
	{0x9EB1FD11BDAE8994ULL, "w_sundermount_22_0.mao"},
	{0x9EB1FEE04D9CC534ULL, "gen00fl_anders6119414.cl"},
	{0x9EB24B84386761D6ULL, "77383103.ogg"},
	{0x9EB27354B7A09F05ULL, "mer110_jansen6032918.cl"},
	{0x9EB324C72AFFDD82ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\cyclone.gad"},
	{0x9EB347D9EB9CC1EDULL, "ser221_gascard6165439.cl"},
	{0x9EB35272C775CE66ULL, "qun200_party_comment6095403.cl"},
	{0x9EB37DDFB265617BULL, "mrl150_merrill6197593.cl"},
	{0x9EB3D375F0408D0CULL, "978838968.ogg"},
	{0x9EB3EAE52CE42E1DULL, "isa311_party_comments6092028.cl"},
	{0x9EB41DBE3EED7572ULL, "178218636.ogg"},
	{0x9EB50B41E64738D8ULL, "vxht_mansion2_1b2_0_0.vlm"},
	{0x9EB5B1EAFC8FE8A9ULL, "art\\vfx\\combat\\vfx_sg_app_c.cess.ani"},
	{0x9EB6C04123AD4C72ULL, "vxht_0_hightown_a2_bxu_0_0.vlm"},
	{0x9EB736070BF96EC5ULL, "33725395.ogg"},
	{0x9EB7469B7D892B4FULL, "vxlt_0_lowtown_a2_ef5_0_0.vlm"},
	{0x9EB7904305700373ULL, "gen_im_trs_slv_sta_cld_r1.uti"},
	{0x9EB7A26857DBF26DULL, "vfx_vfx_p2.psh.fxo.dx11.assembly"},
	{0x9EB80030FB575C65ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_2.ani"},
	{0x9EB93C26B25318F2ULL, "vxlt_gallowstemplar_2jd_0_0.vlm"},
	{0x9EB949299E4907D1ULL, "dae000_hate_follower_react6103241.cl"},
	{0x9EB94B498893EBA4ULL, "hlac_lt_0_lowtown_docks_a3_-65538_2y.dds"},
	{0x9EB9D26155DA8A5EULL, "6121093_m.fxe"},
	{0x9EBA47C72B063879ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\cyclone.evt"},
	{0x9EBA8B0C416DB17AULL, "hlac_lt_ruinedhouse_n_-1_8z.dds"},
	{0x9EBA9229A2C3340BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robe_d.dds"},
	{0x9EBAEDC102B99DB0ULL, "vxw_encounter3_72_0_0.vlm"},
	{0x9EBB28FBB050893EULL, "hlac_lt_0_lowtown_docks_a3_n_-2_3z.dds"},
	{0x9EBB43DA3CF63A55ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_act_c.impa.ani"},
	{0x9EBBD0EDBECE44FDULL, "one100dg_kelder6055418.cl"},
	{0x9EBC652AB0E772BDULL, "dae200_bran6094796.cl"},
	{0x9EBC7F6AA9C8EAE0ULL, "lgt251tr_nexus_fight.utt"},
	{0x9EBCCCEAFC96352BULL, "art\\vfx\\combat\\vfx_sg_app_c.cess.gad"},
	{0x9EBD1D7A05005005ULL, "vxlt_warehouse_19i_0_1.vlm"},
	{0x9EBD67CC5ADFE800ULL, "lgt210_eau_symp_templar6144461.cl"},
	{0x9EBD7DE1345A8DD4ULL, "vxca_caverns2_a4w_0_0.vlm"},
	{0x9EBD8339857244E3ULL, "6068690_m.fxe"},
	{0x9EBDAD46B4C110D9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\prp_ccliff02_d.dds"},
	{0x9EBDCE5AA2C47BD1ULL, "27814137.ogg"},
	{0x9EBE19999C973664ULL, "1026020685.ogg"},
	{0x9EBE6CF5C0D209FDULL, "934999857.ogg"},
	{0x9EBEC78262F61B57ULL, "vxlt_0_lowtown_n_fex_0_0.vlm"},
	{0x9EBF652F69329E31ULL, "vxw_encounter1_ku_0_0.vlm"},
	{0x9EC085EEFD31863EULL, "mag291_torpor.cnv"},
	{0x9EC0E942331E5086ULL, "570278946.ogg"},
	{0x9EC113E01E2AE715ULL, "vxca_caverns2prc_9xo_0_0.vlm"},
	{0x9EC132481B6366C7ULL, "one280_iwan6096879.cl"},
	{0x9EC14545E0D352E4ULL, "hlac_ht_keep_n_3734_1z.dds"},
	{0x9EC18D30E7DFA868ULL, "507596958.ogg"},
	{0x9EC1E5817116423AULL, "hlac_lt_gallowsprison_n_-1_5z.dds"},
	{0x9EC1F0688199EC08ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0rw_plain.msh"},
	{0x9EC20268819A0A99ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0rw_plain.mao"},
	{0x9EC20E33974D16A0ULL, "644654863.ogg"},
	{0x9EC20E68819A1F32ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0rw_plain.mmh"},
	{0x9EC2C9EAFC9A9B90ULL, "art\\vfx\\combat\\vfx_sg_app_c.cess.evt"},
	{0x9EC3237EBDAFE024ULL, "6182020_f.fxe"},
	{0x9EC350EA9E2C163AULL, "follower_banter6172997.cl"},
	{0x9EC351193DE20307ULL, "vxw_bonepitdestroyed_a3_29p_0_0.vlm"},
	{0x9EC3B544781A7807ULL, "pro000_friedrich6051027.cl"},
	{0x9EC3F81659C88637ULL, "vxlt_0_lowtown_a2_hft_0_0.vlm"},
	{0x9EC3FD6E9E63DB52ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female_gob.gad"},
	{0x9EC467EEFD353C4EULL, "mag291_torpor.bnk"},
	{0x9EC4CE1D65E36DB6ULL, "vxht_0_hightown_a3_ai8_0_0.vlm"},
	{0x9EC4E9E7CFA1AC29ULL, "vrc321_party_barks6096314.cl"},
	{0x9EC50215D48070EBULL, "dae000_madam_lusine6152274.cl"},
	{0x9EC58926BD46EEC7ULL, "vxdr_0_deeproads_eh8_0_0.vlm"},
	{0x9EC59A11EF7726B6ULL, "bw_plasma.mao"},
	{0x9EC5B8786F7F43CFULL, "vxlt_0_lowtown_a2_ei4_0_0.vlm"},
	{0x9EC5BDFC6DD06B5DULL, "vxlt_0_lowtown_n_fgt_0_0.vlm"},
	{0x9EC60C4E93F2BB7AULL, "follower_banter6141717.cl"},
	{0x9EC615DE8119309BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curved.mmh"},
	{0x9EC621DE81194538ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curved.mao"},
	{0x9EC633DE81196385ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curved.msh"},
	{0x9EC7018A0616FBC4ULL, "ca_cave_2193_ll.mmh"},
	{0x9EC7B0D7FD59CE77ULL, "60915903.ogg"},
	{0x9EC7DCE80A5F75E3ULL, "mer120_hubert6070208.cl"},
	{0x9EC815D8106C1268ULL, "497780364.ogg"},
	{0x9EC8194D05AD26CFULL, "vxht_keep_n_6g2_0_0.vlm"},
	{0x9EC833D275A5FC6FULL, "hlac_ca_cavern1_haunted_24425_2z.dds"},
	{0x9EC84FC72B128368ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\cyclone.ani"},
	{0x9EC89ED0B099DD67ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_lgtc_0d.dds"},
	{0x9EC8A194273B857EULL, "gen00fl_fenris6127086.cl"},
	{0x9EC8FB8FE54B1DD6ULL, "follower_banter6164059.cl"},
	{0x9EC978406811AD00ULL, "dae000_bodhan6146912.cl"},
	{0x9EC9A901209D21E7ULL, "vxw_encounter3_35_0_0.vlm"},
	{0x9EC9B9F7AA086164ULL, "6135098_m.fxe"},
	{0x9ECA006E9E684B89ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female_gob.evt"},
	{0x9ECB8A008F703AE4ULL, "ht_0_hightownexplosion_n_53_0.rml"},
	{0x9ECBC9EFFA55EE8EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ri_basicoutcrn.mmh"},
	{0x9ECBD5EFFA560325ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ri_basicoutcrn.mao"},
	{0x9ECBE3EFFA561B7CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ri_basicoutcrn.msh"},
	{0x9ECC4979BE1A6D22ULL, "plt_one100pt_herb_fang.nss"},
	{0x9ECCB3B56C8D8883ULL, "6139884_m.fxe"},
	{0x9ECCE1F1D81A4CCEULL, "mag360_thrask.cnv"},
	{0x9ECD680DFE7D155EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action.evt"},
	{0x9ECDA90F6EDF36F5ULL, "vxlt_shop_1w_0_1.vlm"},
	{0x9ECDE0B763508A37ULL, "gen_im_arm_boo_med_act2_02.uti"},
	{0x9ECDF38F393D4C03ULL, "359900900.ogg"},
	{0x9ECE28AEEEEE65D4ULL, "dae000tr_display_district.utt"},
	{0x9ECE408D3C5D6A97ULL, "82929728.ogg"},
	{0x9ECE69D76EA2B828ULL, "vxca_caverns2_95d_0_0.vlm"},
	{0x9ECE9ADC066FAED6ULL, "vxht_mansion2b_2fg_0_0.vlm"},
	{0x9ECF7554C2599047ULL, "hlac_w_sundermount_262149_0z.dds"},
	{0x9ECF77CA2B3F6CB0ULL, "769695615.ogg"},
	{0x9ECFDAF6A9F95A10ULL, "vxht_keep_62u_0_0.vlm"},
	{0x9ECFFD53F208CDEFULL, "follower_barks6153884.cl"},
	{0x9ED003F1D81CBC3EULL, "mag360_thrask.bnk"},
	{0x9ED006FE1A5D2BFDULL, "plt_prm000pt_resilient_ring.plo"},
	{0x9ED03FA53D726FADULL, "hm_feynriel.mor"},
	{0x9ED07569DCC6D933ULL, "and230_mage_girl6135828.cl"},
	{0x9ED0C2865FBC6BCAULL, "vxlt_0_lowtown_a2_n_gmp_0_0.vlm"},
	{0x9ED0FFA055D06AFAULL, "llac_w_encounter3_-2_0.dds"},
	{0x9ED16D82501A061DULL, "hlac_w_sundermount_-131067_0y.dds"},
	{0x9ED18C862D444964ULL, "fen110_fenris6118119.cl"},
	{0x9ED1912CC7CD26F0ULL, "w_blightlands_183_0.mao"},
	{0x9ED1952CC7CD2D03ULL, "w_blightlands_183_0.mmh"},
	{0x9ED1A32CC7CD455DULL, "w_blightlands_183_0.msh"},
	{0x9ED1C2F62FD81B87ULL, "vxw_bonepitdestroyed_a3_24s_0_0.vlm"},
	{0x9ED1EEFBCA3369D8ULL, "vxht_mansion2_1h0_0_0.vlm"},
	{0x9ED225960854443FULL, "vxw_bonepitmining_a2_1os_0_0.vlm"},
	{0x9ED269F34BBE9EB7ULL, "vxlt_0_lowtown_a2_hcw_0_0.vlm"},
	{0x9ED2B8D4C6EB0865ULL, "t3_com_c03.tnt"},
	{0x9ED33FFA57D98636ULL, "vxht_0_hightown_a3_av9_0_0.vlm"},
	{0x9ED3A3FB07997D72ULL, "vxht_0_hightown_a2_bfw_0_0.vlm"},
	{0x9ED400121BF09C5AULL, "244688169.ogg"},
	{0x9ED47F0DFE835A89ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action.gad"},
	{0x9ED4F8E703AD8222ULL, "dae000_merchant_access_hf6135895.cl"},
	{0x9ED585DE7A60B40DULL, "6097830_m.fxe"},
	{0x9ED585FD7964E749ULL, "chu_bl_dot.mao"},
	{0x9ED59F799AB79BB2ULL, "llac_w_woundedcoast_0_12.dds"},
	{0x9ED5CE232DA57CD3ULL, "vxht_0_hightown_n_ca1_0_0.vlm"},
	{0x9ED5FA1B6D43A390ULL, "vxlt_gallowstemplar_n_4yz_0_0.vlm"},
	{0x9ED61C9059FBFB72ULL, "mrl210_merrill.cnv"},
	{0x9ED68B27E34421B4ULL, "6175937_m.fxe"},
	{0x9ED6E9270405E4AEULL, "vxca_cavern1_haunted_ck4_0_0.vlm"},
	{0x9ED753B40D6E55C0ULL, "vxht_keep_5k1_0_0.vlm"},
	{0x9ED78E07F6894548ULL, "759459861.ogg"},
	{0x9ED7AAA1A6B1846DULL, "6075096_m.fxe"},
	{0x9ED7DF0D7898D6A7ULL, "631665172.ogg"},
	{0x9ED7E86E9E7460F8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\stairs_descend_female_gob.ani"},
	{0x9ED7F57404BF53F4ULL, "vxca_caverns2_9fa_0_0.vlm"},
	{0x9ED8054FD790740CULL, "art\\vfx\\combat\\vfx_f_and_fade_c.dura.ani"},
	{0x9ED831B4DF882433ULL, "6044612_m.fxe"},
	{0x9ED87F106DF126E3ULL, "vxht_bartrand_n_5wr_0_0.vlm"},
	{0x9ED9165C39D376A8ULL, "193556269.ogg"},
	{0x9ED93E9059FE6ACAULL, "mrl210_merrill.bnk"},
	{0x9ED9B2C248D051BBULL, "vxlt_0_lowtown_docks_a3_n_c3i_0_0.vlm"},
	{0x9ED9C5E70D5EDAC8ULL, "vxht_0_hightown_a3_cb9_0_0.vlm"},
	{0x9EDB500DFE8929D7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action.ani"},
	{0x9EDB8E2AC92169D5ULL, "fontshader.psh.fxo"},
	{0x9EDB8EB3B45E00D4ULL, "vxlt_0_lowtown_n_eoo_0_1.vlm"},
	{0x9EDC01A0CFED4925ULL, "pro000_cut2d_voyage6004232.cl"},
	{0x9EDC2CE41ED69709ULL, "lt_ruinedhouse_n_3567.rml"},
	{0x9EDD61C7F4ADB2D7ULL, "vxlt_warehouse_n_19h_0_1.vlm"},
	{0x9EDD68EABC8DA6E5ULL, "gen00fl_merrill6129984.cl"},
	{0x9EDDFAE1EF1B0C21ULL, "and271_lyrium_smuggler6108511.cl"},
	{0x9EDE468846760BB2ULL, "vxlt_darktown_4zv_0_0.vlm"},
	{0x9EDE48CBD385FE39ULL, "ave331_jeven6151383.cl"},
	{0x9EDE894DF8A58629ULL, "gen00fl_varric6107287.cl"},
	{0x9EDE98388054FEF4ULL, "792903981.ogg"},
	{0x9EDEB3D949D16986ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\right_hand_gesture_left.ani"},
	{0x9EDED27D21D5D978ULL, "wa_rock.mao"},
	{0x9EDEE827985038AFULL, "vrc321tr_door_shift_1.utt"},
	{0x9EDEFF891DB23664ULL, "890233630.ogg"},
	{0x9EDF124FD796A94EULL, "art\\vfx\\combat\\vfx_f_and_fade_c.dura.gad"},
	{0x9EDF346351B2844AULL, "vxlt_0_lowtown_a2_n_gjq_0_0.vlm"},
	{0x9EDF4ED8EEFA1B52ULL, "vxht_0_hightown_n_ca4_0_0.vlm"},
	{0x9EE0B94DF3ED64C8ULL, "vxlt_gallowstemplar_f_53a_0_0.vlm"},
	{0x9EE0D6B3E96FFA36ULL, "6030428_f.fxe"},
	{0x9EE12AED1EE50151ULL, "6156092_m.fxe"},
	{0x9EE15E035E69543BULL, "art\\levels\\races\\design\\_textures\\prp_chestc_sa.dds"},
	{0x9EE161872DD6FB9AULL, "318590570.ogg"},
	{0x9EE17EB6A3785D52ULL, "gen00fl_anders.cnv"},
	{0x9EE2962CC7DBA228ULL, "w_blightlands_183_0.phy"},
	{0x9EE29BA25C544C2FULL, "vxlt_0_lowtown_docks_n_8ug_0_0.vlm"},
	{0x9EE2AC596B4089B1ULL, "6043835_m.fxe"},
	{0x9EE2C013B8CE1D6CULL, "439466189.ogg"},
	{0x9EE32FE0E7055087ULL, "t3_arm_stl.tnt"},
	{0x9EE367338D1D363EULL, "mrl000_marethari6129432.cl"},
	{0x9EE36DE26124946BULL, "vxht_0_hightown_a6i_0_0.vlm"},
	{0x9EE3801A30B7A761ULL, "vxlt_gallowstemplar_n_5ro_0_0.vlm"},
	{0x9EE3C99D892BC879ULL, "vxw_encounter2_camp_5r_0_0.vlm"},
	{0x9EE3EE5A890E4E92ULL, "vxht_brothel_n_62w_0_0.vlm"},
	{0x9EE4A0B6A37ACCAAULL, "gen00fl_anders.bnk"},
	{0x9EE4A5C827286779ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02a.msh"},
	{0x9EE4AFC827287867ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02a.mmh"},
	{0x9EE4F911BDD9DCA4ULL, "w_sundermount_22_0.rml"},
	{0x9EE568FB850E0252ULL, "mag160_feynriel6020063.cl"},
	{0x9EE58EAFDD5C692EULL, "vxht_keep_n_70c_0_0.vlm"},
	{0x9EE60D4FD79CBF2DULL, "art\\vfx\\combat\\vfx_f_and_fade_c.dura.evt"},
	{0x9EE72E6C34509C02ULL, "6075778_m.fxe"},
	{0x9EE75526498CC999ULL, "182374556.ogg"},
	{0x9EE81B3CF4BAFCA9ULL, "6101272_m.fxe"},
	{0x9EE8A3A5AF63F1C9ULL, "llac_lt_0_lowtown_docks_a3_n_-131073_5.dds"},
	{0x9EE92762DB2E3BB5ULL, "dae000_sandal6148231.cl"},
	{0x9EE97853A7DB2A92ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.cess.gad"},
	{0x9EE9B78288C1C6D0ULL, "vxdr_primevaltomb_758_0_0.vlm"},
	{0x9EE9D76C0B2914FBULL, "dae300_bran6142892.cl"},
	{0x9EEA18ABF52B8297ULL, "vxw_bonepitmining_a2_241_0_0.vlm"},
	{0x9EEA30F846F072BEULL, "art\\levels\\races\\proxy\\props\\_textures\\cotbedroll_n.dds"},
	{0x9EEA350A2A9A5E2CULL, "hlac_lt_warehouse_n_1_5x.dds"},
	{0x9EEA88DEA85EB57DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairleft.phy"},
	{0x9EEAA1B24EFCB015ULL, "lgt310ip_body_3.utp"},
	{0x9EEAD296AE031BEAULL, "943367129.ogg"},
	{0x9EEAF549FE0DA4D6ULL, "243349427.ogg"},
	{0x9EEB0C47C79F8A94ULL, "w_blightlands_86_0.phy"},
	{0x9EEB6C8B07EF5BD3ULL, "6111221_f.fxe"},
	{0x9EEBD4BCA9DDAD1BULL, "6201496_f.fxe"},
	{0x9EEC00A29AD0C844ULL, "qun120_arishok6051713.cl"},
	{0x9EEC21F8FF1E07DDULL, "w_bonepit_136_0.mao"},
	{0x9EEC235611ABE066ULL, "hlac_lt_hangedman_-1_10z.dds"},
	{0x9EEC25F8FF1E0E16ULL, "w_bonepit_136_0.mmh"},
	{0x9EEC2FF8FF1E1F14ULL, "w_bonepit_136_0.msh"},
	{0x9EEC5D3E9CA4230EULL, "701256856.ogg"},
	{0x9EEC9BD949DD7EF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\right_hand_gesture_left.evt"},
	{0x9EED08CFF20C5348ULL, "vxlt_gallowstemplar_n_5rj_0_0.vlm"},
	{0x9EED2FA68ACC1DA0ULL, "714972248.ogg"},
	{0x9EED44A9E8935422ULL, "vxlt_undercityencounter_1b1_0_0.vlm"},
	{0x9EED9C96F9970321ULL, "vxht_keep_5e6_0_0.vlm"},
	{0x9EEDB66636AFF313ULL, "zz_approval_debug6003923.cl"},
	{0x9EEDE1528831FE9AULL, "follower_banter6143399.cl"},
	{0x9EEDFBC04F66617CULL, "431444376.ogg"},
	{0x9EEE5A198A73AABEULL, "vxlt_0_lowtown_a2_n_c_hkk_0_0.vlm"},
	{0x9EEE892186B8859FULL, "ico_gloves_heavy_2.dds"},
	{0x9EEF3C62B0B49262ULL, "6059329_m.fxe"},
	{0x9EEFA742BDAE9C6CULL, "vxw_bonepitmining_a2_1pq_0_0.vlm"},
	{0x9EEFF786F1430243ULL, "450388405.ogg"},
	{0x9EF07B53A7E14EC9ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.cess.evt"},
	{0x9EF090AA6A880428ULL, "hlac_dr_dungeon_3214_11z.dds"},
	{0x9EF0ABFC9B36B6E9ULL, "w_encounter3_20_0_lc.mmh"},
	{0x9EF0F4D3F679B284ULL, "41107210.ogg"},
	{0x9EF118061386868AULL, "vxlt_0_lowtown_docks_a3_n_d19_0_0.vlm"},
	{0x9EF16F2969105F27ULL, "llac_ht_mansion2_818_5.dds"},
	{0x9EF1C705174CBC20ULL, "hlac_dr_0_deeproads_5840_2y.dds"},
	{0x9EF1DD9C74F1549EULL, "static_lava_p1.psh.fxo.dx11"},
	{0x9EF1F1F722ADBFE1ULL, "vxlt_gallowstemplar_n_5un_0_0.vlm"},
	{0x9EF2357530B75FD5ULL, "6209897_m.fxe"},
	{0x9EF25AEF9412B9D0ULL, "6146834_m.fxe"},
	{0x9EF286862AF04FF2ULL, "dae000_templar_knight6165158.cl"},
	{0x9EF29951C7CAC9C6ULL, "14892484.ogg"},
	{0x9EF29CD949E1EB80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\right_hand_gesture_left.gad"},
	{0x9EF2E8281F971A60ULL, "dae000_orana6149017.cl"},
	{0x9EF2F31B9BD4F33FULL, "w_encounter3_80_0.msh"},
	{0x9EF2F91B9BD4FDE1ULL, "w_encounter3_80_0.mmh"},
	{0x9EF3051B9BD51182ULL, "w_encounter3_80_0.mao"},
	{0x9EF36421189C8901ULL, "6021686_m.fxe"},
	{0x9EF3A4B9683E8EBCULL, "31825883.ogg"},
	{0x9EF3F4EA010243BBULL, "749437069.ogg"},
	{0x9EF4708C08104116ULL, "vxlt_undercityserialkiller_1lr_0_1.vlm"},
	{0x9EF47381AFB81F99ULL, "vxht_bartrand_n_5y2_0_0.vlm"},
	{0x9EF4C8F114933E85ULL, "art\\vfx\\combat\\vfx_str_att_front_r_arm_c.anb"},
	{0x9EF4E5986178ED1EULL, "6051999_f.fxe"},
	{0x9EF4E947C7A7A391ULL, "w_blightlands_86_0.msh"},
	{0x9EF4F747C7A7BBACULL, "w_blightlands_86_0.mao"},
	{0x9EF50347C7A7D04FULL, "w_blightlands_86_0.mmh"},
	{0x9EF5BF5C4421C8AFULL, "vxlt_gallowstemplar_5ce_0_0.vlm"},
	{0x9EF66AB97CF502E6ULL, "gen00fl_anders6125842.cl"},
	{0x9EF72976B0382E2FULL, "llac_dr_0_deeproads_17639_1.dds"},
	{0x9EF74C6A9FCDA284ULL, "lt_0_lowtown_n_12_0.rml"},
	{0x9EF7C3B0CE0C4A09ULL, "838916206.ogg"},
	{0x9EF7C85B78080D22ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow5base.msh"},
	{0x9EF7CA5B780810D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow5base.mmh"},
	{0x9EF7D65B7808257BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow5base.mao"},
	{0x9EF7D7593FD8AF5CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25_0.pwk"},
	{0x9EF7E3D1B578B117ULL, "genip_node_j2.utp"},
	{0x9EF82C14360367CAULL, "6171265_m.fxe"},
	{0x9EF8EB70F230E4F9ULL, "vxlt_foundry_n_102_0_0.vlm"},
	{0x9EF92331C31F30D3ULL, "vxlt_gallowstemplar_46s_0_0.vlm"},
	{0x9EF92449DE9B2398ULL, "731349975.ogg"},
	{0x9EF95993934C726CULL, "899163248.ogg"},
	{0x9EF9670F7EAA93D5ULL, "static_static_p28.psh.fxo.dx11.assembly"},
	{0x9EF98712F96B1C02ULL, "w_bonepit_211_0.msh"},
	{0x9EF98912F96B1FB8ULL, "w_bonepit_211_0.mmh"},
	{0x9EF99512F96B345BULL, "w_bonepit_211_0.mao"},
	{0x9EF9A4B1B5501DC7ULL, "vxlt_darktown_57m_0_0.vlm"},
	{0x9EF9AE61EECF99B8ULL, "llac_w_bonepitdestroyed_a3_6_0.dds"},
	{0x9EF9B403FB02B6C8ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_010a_0s.dds"},
	{0x9EF9BECA18EB632BULL, "vxw_blightlands_138_0_0.vlm"},
	{0x9EFA0E4038F5BAECULL, "vxht_playershouse_n_1d2_0_0.vlm"},
	{0x9EFA46FE624F888CULL, "6029355_m.fxe"},
	{0x9EFA4EB4A462910BULL, "vxht_0_hightown_n_bu0_0_0.vlm"},
	{0x9EFA56B17E080828ULL, "llac_w_bonepitdestroyed_a3_65537_0.dds"},
	{0x9EFA57994E82343BULL, "6125931_m.fxe"},
	{0x9EFB16B9946DC3EFULL, "vxca_cavern1_haunted_cpd_0_0.vlm"},
	{0x9EFB2C6C2ACA9B8BULL, "vxlt_gallowstemplar_n_451_0_0.vlm"},
	{0x9EFB3679BE41BDA5ULL, "plt_one100pt_herb_fang.plo"},
	{0x9EFB7AACE4026BC8ULL, "vxlt_gallowstemplar_n_5uk_0_0.vlm"},
	{0x9EFB7DDEA86D15E9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairleft.mao"},
	{0x9EFB89DEA86D2982ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairleft.mmh"},
	{0x9EFB8BDEA86D2D38ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4so_stairleft.msh"},
	{0x9EFB8C902CF7870AULL, "gen_im_trs_apo_junk1_02.uti"},
	{0x9EFBAE830274B265ULL, "llac_ca_caverns1_mines_24423_6.dds"},
	{0x9EFBBAA6A42C3305ULL, "vxca_caverns2_7wu_0_0.vlm"},
	{0x9EFC0E73EB8D1BA1ULL, "vxht_keep_5h5_0_0.vlm"},
	{0x9EFC142312404A85ULL, "mag330_walter6045243.cl"},
	{0x9EFCC4228CDE4794ULL, "6174978_m.fxe"},
	{0x9EFCD44F5B3B60FBULL, "257585063.ogg"},
	{0x9EFD076B7D1716FBULL, "6075685_m.fxe"},
	{0x9EFE23ACCC25BB17ULL, "dae000_mother6121067.cl"},
	{0x9EFE6353A7ED6338ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.cess.ani"},
	{0x9EFE69509798ACD0ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.mmh"},
	{0x9EFE6D0C51D57E08ULL, "rdr111dg_martin6054153.cl"},
	{0x9EFEA036D164FAD6ULL, "ave100_donnic6105986.cl"},
	{0x9EFEB62D797A2681ULL, "6077991_m.fxe"},
	{0x9EFEF8573B8E38C2ULL, "art\\vfx\\combat\\vfx_m_bld_control_imp_p.impa.gad"},
	{0x9EFF0328D1D2F626ULL, "llac_dr_primevaltomb_664_2.dds"},
	{0x9F0039D6931D77E4ULL, "585816191.ogg"},
	{0x9F0117640B8D5E5FULL, "clf221_jalen6070335.cl"},
	{0x9F01B65E5AB007B0ULL, "zz_dae_debug6036578.cl"},
	{0x9F01D5FE0A92A181ULL, "6024986_m.fxe"},
	{0x9F0206CB13BD94BCULL, "6121334_m.fxe"},
	{0x9F0217E5C818BE56ULL, "pro000_varric6170670.cl"},
	{0x9F02271FDBF9070BULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody5_.ani"},
	{0x9F02A3E78473CF52ULL, "vxlt_gallowstemplar_46v_0_0.vlm"},
	{0x9F02B75BE4F46809ULL, "6155754_m.fxe"},
	{0x9F02D15B78122463ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow5base.phy"},
	{0x9F02F830B407684BULL, "512189173.ogg"},
	{0x9F03776E9164D9B1ULL, "vxlt_hangedman_1vm_0_0.vlm"},
	{0x9F037F2459766A8BULL, "vxht_keep_6c9_0_0.vlm"},
	{0x9F038B963210F45EULL, "gen00fl_merrill6108265.cl"},
	{0x9F039530BF484127ULL, "4592850.ogg"},
	{0x9F04C52C958AAF6CULL, "vxdr_primevaltomb_2q1_0_0.vlm"},
	{0x9F0576AF4414802BULL, "vxlt_backalley_n_1pc_0_0.vlm"},
	{0x9F05CC5AE32E4A63ULL, "6144145_m.fxe"},
	{0x9F05E2AADCDC51BCULL, "88949543.ogg"},
	{0x9F05F74E05EB039BULL, "art\\vfx\\combat\\vfx_gib_ogr_p.mmh"},
	{0x9F06034E05EB1838ULL, "art\\vfx\\combat\\vfx_gib_ogr_p.mao"},
	{0x9F0608DDDCC59BA3ULL, "6054583_m.fxe"},
	{0x9F06154E05EB3685ULL, "art\\vfx\\combat\\vfx_gib_ogr_p.msh"},
	{0x9F061B573B9492B9ULL, "art\\vfx\\combat\\vfx_m_bld_control_imp_p.impa.evt"},
	{0x9F0641D8FDABE954ULL, "vfx_blt_lava_small.bnk"},
	{0x9F069DF127F9804AULL, "6174165_m.fxe"},
	{0x9F06E2FE6C16848FULL, "hlac_ca_caverns2_5802_54y.dds"},
	{0x9F073B0D36CB6C1DULL, "963953187.ogg"},
	{0x9F0794306F268BF3ULL, "6086265_m.fxe"},
	{0x9F07B4CEE6549C35ULL, "vxlt_darktown_2x9_0_0.vlm"},
	{0x9F07B96EEF802303ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocklarge.msh"},
	{0x9F07C76EEF803B5DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocklarge.mmh"},
	{0x9F07CB6EEF80426EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocklarge.mao"},
	{0x9F080C9CB68622EFULL, "6103294_m.fxe"},
	{0x9F0813D2DFD87020ULL, "501948902.ogg"},
	{0x9F08168EA7463647ULL, "vxlt_darktown_58l_0_0.vlm"},
	{0x9F08A2E4510AA6CDULL, "227101118.ogg"},
	{0x9F08C0919658A98BULL, "vxht_0_hightown_n_bp3_0_0.vlm"},
	{0x9F08D8A2651C07B3ULL, "hlac_ht_0_hightown_a3_65533_3x.dds"},
	{0x9F09043E3E5EA1F8ULL, "gen00fl_aveline6137596.cl"},
	{0x9F0915ED40E2BF46ULL, "6139683_m.fxe"},
	{0x9F0928B17455C298ULL, "37366214.ogg"},
	{0x9F09776DBA2DE279ULL, "vxca_caverns1_j2p_0_0.vlm"},
	{0x9F09EC89D5F88448ULL, "vxlt_gallowstemplar_n_5hh_0_0.vlm"},
	{0x9F0B223A00E2D1D1ULL, "416129226.ogg"},
	{0x9F0B8C5097A372EFULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.anb"},
	{0x9F0BA2091ACCA01BULL, "6071007_m.fxe"},
	{0x9F0BE46CE3DB811CULL, "6120816_m.fxe"},
	{0x9F0C470569060DB8ULL, "vxw_blightlands_120_0_0.vlm"},
	{0x9F0C4C882B3AA556ULL, "375448900.ogg"},
	{0x9F0C58BFFAFEAF19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_r_fist_clench.ani"},
	{0x9F0C9C2F028ABC44ULL, "w_woundedcoast_210_0.msh"},
	{0x9F0CAE2F028ADAADULL, "w_woundedcoast_210_0.mao"},
	{0x9F0CB22F028AE1E6ULL, "w_woundedcoast_210_0.mmh"},
	{0x9F0CC84D1E672B39ULL, "ht_0_hightown_a3_29_0_lc.mmh"},
	{0x9F0D36DE81556BD0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\4so_curved.phy"},
	{0x9F0D45C585B3BA3BULL, "lt_0_lowtown_docks_a3_n_124_0.msh"},
	{0x9F0D4AE0DF25A3ECULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.impa.ani"},
	{0x9F0D57C585B3D886ULL, "lt_0_lowtown_docks_a3_n_124_0.mao"},
	{0x9F0D5A33FD9D8C50ULL, "gr_dis.mao"},
	{0x9F0D63C585B3ED25ULL, "lt_0_lowtown_docks_a3_n_124_0.mmh"},
	{0x9F0D84D1041F130EULL, "mag291_demon_marethari6029136.cl"},
	{0x9F0DDD5B15E5D0E8ULL, "6093769_m.fxe"},
	{0x9F0E7EC1FC5F640BULL, "dae300_alistair_king6134934.cl"},
	{0x9F0EA0EC875365A5ULL, "art\\characters\\creatures\\buzzard\\textures\\c_buzzard_0s.dds"},
	{0x9F0ED5B10699F0E1ULL, "vxlt_gallowstemplar_n_5cl_0_0.vlm"},
	{0x9F0EDAF349BCA7E5ULL, "mag101_alain6032285.cl"},
	{0x9F0EF9BDCBB9A51EULL, "6160771_m.fxe"},
	{0x9F0F2BE41F01EF29ULL, "lt_ruinedhouse_n_3567.anb"},
	{0x9F0F3AF4BAD26973ULL, "hlac_w_encounter3_0_0z.dds"},
	{0x9F0F77B73515773BULL, "143733510.ogg"},
	{0x9F0FD2F5A0498FCCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railplain.phy"},
	{0x9F0FD964E97A4B34ULL, "vxht_brothel_n_5ak_0_0.vlm"},
	{0x9F1098C8274D8B9CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet02a.phy"},
	{0x9F10DA2DA39ADEC3ULL, "llac_w_woundedcoast_131074_0.dds"},
	{0x9F111F516BD98EEFULL, "vxht_0_hightown_a2_cdi_0_0.vlm"},
	{0x9F11AED400931F8EULL, "qun211_party_comment6072262.cl"},
	{0x9F11BFC65E147612ULL, "ave310_brennan6136218.cl"},
	{0x9F11C26570B5CEF0ULL, "vxw_bonepitdestroyed_a3_1r9_0_0.vlm"},
	{0x9F1208FD33A90012ULL, "vxw_encounter2_ko_0_0.vlm"},
	{0x9F13183830E6F8ACULL, "383130480.ogg"},
	{0x9F1323573B9F2AA8ULL, "art\\vfx\\combat\\vfx_m_bld_control_imp_p.impa.ani"},
	{0x9F1323F98861CF40ULL, "vxw_bonepitdestroyed_a3_1ve_0_0.vlm"},
	{0x9F1333BFFB048F5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_r_fist_clench.gad"},
	{0x9F1357E0DF2A252EULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.impa.gad"},
	{0x9F13752A61F6B6F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_blink.ani"},
	{0x9F13B475BDA586F8ULL, "art\\vfx\\materials\\vfx_crack_cyc_add.mao"},
	{0x9F13DAD9307287E3ULL, "6086944_m.fxe"},
	{0x9F145838640B695EULL, "zz_whereami.ncs"},
	{0x9F1521C58B47438AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_down.gad"},
	{0x9F158A42B8FE19CDULL, "vxca_cave_2e5_0_0.vlm"},
	{0x9F15B7D71CC566DAULL, "6016672_m.fxe"},
	{0x9F15DADB2036B4D8ULL, "vxht_bartrand_n_5sq_0_0.vlm"},
	{0x9F169145A090D909ULL, "706487720.ogg"},
	{0x9F16A92F029326B1ULL, "w_woundedcoast_210_0.phy"},
	{0x9F16DC51878644DFULL, "hlac_lt_0_lowtown_n_65533_3z.dds"},
	{0x9F16DD4B4D128BD1ULL, "follower_banter6172893.cl"},
	{0x9F170E1FDC0B3895ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody5_.gad"},
	{0x9F185E66C7EE9CC8ULL, "vxlt_gallowstemplar_n_5ci_0_0.vlm"},
	{0x9F18DF8F3F876072ULL, "vxlt_0_lowtown_a2_hgr_0_0.vlm"},
	{0x9F1920E465FA4B98ULL, "47233870.ogg"},
	{0x9F19B29E23B3CCDAULL, "195496154.ogg"},
	{0x9F19E7AC6ABE960EULL, "6067100_m.fxe"},
	{0x9F1A502A61FC963BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_blink.gad"},
	{0x9F1A50BFFB0ADF40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hand_r_fist_clench.evt"},
	{0x9F1A52E0DF303B0DULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.impa.evt"},
	{0x9F1A7F0D0FF1E715ULL, "224761009.ogg"},
	{0x9F1AB7CB8877D6E2ULL, "gen_im_wep_war_shd_act3_01.uti"},
	{0x9F1AE678A6CC2BC5ULL, "mag160_feynriel6107435.cl"},
	{0x9F1B30C01C0D0856ULL, "one100dg_nabil6054976.cl"},
	{0x9F1B723AD0648DC3ULL, "hlac_w_blightlands_0_0x.dds"},
	{0x9F1B8DBF355FB965ULL, "rdr111dg_martin6057076.cl"},
	{0x9F1BA76019F6A4CEULL, "con_ico_direct.dds"},
	{0x9F1BAA37F23F7849ULL, "one280cr_golem.utc"},
	{0x9F1BAA8CAEB803C2ULL, "follower_banter6161074.cl"},
	{0x9F1BE8E7F94B6B69ULL, "vxht_keep_61y_0_0.vlm"},
	{0x9F1BF66DDE6BC721ULL, "vxht_0_hightown_c7n_0_0.vlm"},
	{0x9F1BFF80E4C97EE9ULL, "mag100_cullen6206258.cl"},
	{0x9F1C04C58B4D3121ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_down.evt"},
	{0x9F1C382FC28E3A5EULL, "6111090_m.fxe"},
	{0x9F1C618E0965D2E9ULL, "452490715.ogg"},
	{0x9F1CC9F114B492C6ULL, "art\\vfx\\combat\\vfx_str_att_front_r_arm_c.mmh"},
	{0x9F1D0B6C1F266965ULL, "hlac_ca_cavern1_haunted_24425_9y.dds"},
	{0x9F1D30A91ACD3FAAULL, "6016553_m.fxe"},
	{0x9F1E066F1604B148ULL, "788599734.ogg"},
	{0x9F1E0D6881E87C2DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0rw_plain.phy"},
	{0x9F1E571734E07FB3ULL, "6206212_m.fxe"},
	{0x9F1ED5549B539619ULL, "718763825.ogg"},
	{0x9F1EFA1B9BFA3A5EULL, "w_encounter3_80_0.phy"},
	{0x9F1F450F6A4FC918ULL, "hlac_ht_0_hightown_a2_-3_0y.dds"},
	{0x9F1F954D70F57FD1ULL, "278897644.ogg"},
	{0x9F1FC918859A6919ULL, "vxht_bartrand_n_5j7_0_0.vlm"},
	{0x9F205B2875510AB1ULL, "vxlt_hangedman_1xc_0_0.vlm"},
	{0x9F20B86127BFA163ULL, "vxht_bartrand_n_5ny_0_0.vlm"},
	{0x9F2114F31A04122FULL, "vxlt_gallowstemplar_5rf_0_0.vlm"},
	{0x9F213D2801FDA709ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_arh_0n.dds"},
	{0x9F2150843BBEB845ULL, "hlac_lt_undercityserialkiller_-1_16z.dds"},
	{0x9F21EA7828562B05ULL, "qun211_braggart6075598.cl"},
	{0x9F21EF43140585BFULL, "6129312_f.fxe"},
	{0x9F2261F34996517FULL, "gen00fl_merrill6129999.cl"},
	{0x9F227FFEB76F30AAULL, "rdr111dg_assistant6055111.cl"},
	{0x9F230E5BD410A757ULL, "vxht_keep_3cp_0_0.vlm"},
	{0x9F235CA625F2AABEULL, "gr_aoe_clu2.mao"},
	{0x9F236A2CC95CC459ULL, "6054844_m.fxe"},
	{0x9F237246D797A0F5ULL, "789013476.ogg"},
	{0x9F23FC3AEFCF0FCDULL, "w_sundermount_19_0.mmh"},
	{0x9F24003AEFCF161EULL, "w_sundermount_19_0.mao"},
	{0x9F240E3AEFCF2E53ULL, "w_sundermount_19_0.msh"},
	{0x9F24A59B52DC5FCBULL, "character_character_p103.psh.fxo"},
	{0x9F2533D151C85827ULL, "718657870.ogg"},
	{0x9F259839B0EFCEC2ULL, "hla_ht_0_hightown_a2_-131073_0z.dds"},
	{0x9F265A1E87631C7DULL, "w_encounter2_camp_6_0.phy"},
	{0x9F28132E6555E34BULL, "vxlt_0_lowtown_a2_n_feh_0_0.vlm"},
	{0x9F28208C359C7EC6ULL, "framebuffereffects_vsmblur_p1.psh.fxo"},
	{0x9F285F020664603CULL, "vxht_bartrand_n_5vj_0_0.vlm"},
	{0x9F28FCC58B57AE70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_down.ani"},
	{0x9F29AFD6457FD25EULL, "hlac_lt_gallowscourtyard_-131072_2y.dds"},
	{0x9F29FD8A2ACE7CCEULL, "6135791_m.fxe"},
	{0x9F2A1286F5C4177EULL, "vxlt_0_lowtown_docks_a3_d34_0_0.vlm"},
	{0x9F2A8BEAEF5F7B81ULL, "vxht_mansion2b_3ex_0_0.vlm"},
	{0x9F2B0D7334E16D84ULL, "zz_isa_debug6133957.cl"},
	{0x9F2B359A5551CAE4ULL, "gen00fl_varric6103451.cl"},
	{0x9F2BBA5EEAC89BF3ULL, "136187825.ogg"},
	{0x9F2C5B1E87678909ULL, "w_encounter2_camp_6_0.rml"},
	{0x9F2CACB4C22CC9E6ULL, "hlac_lt_gallowstemplar_f_2384_0z.dds"},
	{0x9F2D1CF0F6B710CDULL, "6108617_m.fxe"},
	{0x9F2D971DE8080C16ULL, "zz_lgt_1_debug6114387.cl"},
	{0x9F2D9D3D712434BBULL, "hlac_lt_undercityserialkiller_-1_42y.dds"},
	{0x9F2DFC3AEFD76446ULL, "w_sundermount_19_0.rml"},
	{0x9F2E28C861CDF8EEULL, "vxw_sundermount_2zn_0_1.vlm"},
	{0x9F2E6306DDF24C3DULL, "dead01.mao"},
	{0x9F2E7C70459E4B0CULL, "362067835.ogg"},
	{0x9F2E8B43DDC6EA68ULL, "6051632_f.fxe"},
	{0x9F2EECB717953112ULL, "vxw_encounter2_qi_0_0.vlm"},
	{0x9F2F0D5DE5D1A433ULL, "6016031_m.fxe"},
	{0x9F2F2F7CC90E59D1ULL, "hlac_lt_foundry_n_320_15z.dds"},
	{0x9F2F4A62378AA6BDULL, "vxdr_dungeon_1mm_0_0.vlm"},
	{0x9F2F653EFA703555ULL, "hlac_lt_0_lowtown_a2_n_c_1_11z.dds"},
	{0x9F2F89F4BE022E2FULL, "hlac_ht_0_hightown_a3_-65538_1z.dds"},
	{0x9F2FC462BA114AF3ULL, "vxlt_hangedman_n_1mm_0_0.vlm"},
	{0x9F303B2C9B26B2F0ULL, "fake_im_arm_boo_med_champ.uti"},
	{0x9F307963DFC5EDD8ULL, "gui_guipixelshader_p4.psh.fxo.dx11"},
	{0x9F30AD10813EC296ULL, "lt_0_lowtown_docks_a3_254_0.msh"},
	{0x9F30B710813ED394ULL, "lt_0_lowtown_docks_a3_254_0.mmh"},
	{0x9F30BB10813EDAA7ULL, "lt_0_lowtown_docks_a3_254_0.mao"},
	{0x9F30C67E8DE74632ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldbase_1.mao"},
	{0x9F3115B7F3FA2C2EULL, "lgt320_mis_resp_swordhunter6150322.cl"},
	{0x9F320F0539872384ULL, "706167513.ogg"},
	{0x9F32D2E08573148DULL, "629724977.ogg"},
	{0x9F32E6615DF65AD4ULL, "809874525.ogg"},
	{0x9F339F07C75B7D9BULL, "6004335_m.fxe"},
	{0x9F33B135CF5ECFE1ULL, "hlac_ca_caverns1_24427_5z.dds"},
	{0x9F340B233FBAF4DCULL, "hlac_w_encounter3_-393215_0y.dds"},
	{0x9F34221D0930F2A7ULL, "dae251ip_at_sundermount_b.utp"},
	{0x9F34768330B171A2ULL, "6129555_f.fxe"},
	{0x9F350A8CF039CD6EULL, "hlac_lt_0_lowtown_docks_a3_131071_2x.dds"},
	{0x9F353BFF83243F41ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_2_ian.gad"},
	{0x9F35655CE0052E57ULL, "vxlt_undercityserialkiller_x1_0_0.vlm"},
	{0x9F35EEC015A80789ULL, "280195820.ogg"},
	{0x9F364EF9D05DD0F9ULL, "271694710.ogg"},
	{0x9F365B9C304BE761ULL, "gencr_pride_demon.utc"},
	{0x9F36A0A735F6CDA7ULL, "hlac_lt_warehouse_n_0_5x.dds"},
	{0x9F36B930B426D166ULL, "vxht_0_hightown_a3_by3_0_0.vlm"},
	{0x9F36CD91C72776BBULL, "genip_pp_hti_standing.utp"},
	{0x9F370EEC25014836ULL, "vxht_keep_5gc_0_0.vlm"},
	{0x9F3727F63D258162ULL, "vxht_0_hightown_a3_7vs_0_0.vlm"},
	{0x9F374F1E87717CE9ULL, "w_encounter2_camp_6_0.mao"},
	{0x9F375B1E87719082ULL, "w_encounter2_camp_6_0.mmh"},
	{0x9F375D1E87719438ULL, "w_encounter2_camp_6_0.msh"},
	{0x9F37AEFDE7B13B67ULL, "dae000_sabina.bnk"},
	{0x9F385B0B8C19C92EULL, "6111219_f.fxe"},
	{0x9F388463E7BA2FFEULL, "vxlt_0_lowtown_docks_a3_d45_0_0.vlm"},
	{0x9F38B4E02950EE2DULL, "ch_cl.mao"},
	{0x9F38BF78E5ECF530ULL, "6078331_m.fxe"},
	{0x9F39B52BC270CA5FULL, "527865363.ogg"},
	{0x9F3A34EE0E1F5A33ULL, "dre000_sibling_has_blight6036473.cl"},
	{0x9F3AB476953EB127ULL, "hlac_ca_caverns2prc_5802_47x.dds"},
	{0x9F3AC81C1A9E0E28ULL, "6178354_m.fxe"},
	{0x9F3AD0FDE7B3AADFULL, "dae000_sabina.cnv"},
	{0x9F3B45E46FD5C37FULL, "character_character_p170.psh.fxo.dx11"},
	{0x9F3B460FA4CF085FULL, "295718340.ogg"},
	{0x9F3B7D38BB65E941ULL, "204761465.ogg"},
	{0x9F3BD07EB1F6AD4AULL, "vxlt_0_lowtown_docks_n_c3w_0_0.vlm"},
	{0x9F3C3CFF832A5FEFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_2_ian.ani"},
	{0x9F3C6B806BFA96FBULL, "qun110_petrice6036232.cl"},
	{0x9F3C931B23E5880EULL, "vxlt_0_lowtown_n_h2b_0_0.vlm"},
	{0x9F3CB71BA1F3B3F1ULL, "vxlt_gallowstemplar_n_2jd_0_0.vlm"},
	{0x9F3CBB1AC014D020ULL, "6036711_m.fxe"},
	{0x9F3CED61018E8FFAULL, "vxht_mansion2_1a5_0_0.vlm"},
	{0x9F3D74736C1DF62EULL, "vxlt_0_lowtown_a2_n_c_e8g_0_0.vlm"},
	{0x9F3E0EF0E7EE6347ULL, "6064539_m.fxe"},
	{0x9F3E2F8C080F7D00ULL, "docks_scene_dockworkers6176482.cl"},
	{0x9F3E4653CA84E6E4ULL, "w_bonepitdestroyed_a3_228_0.mmh"},
	{0x9F3E4A53CA84ED37ULL, "w_bonepitdestroyed_a3_228_0.mao"},
	{0x9F3E5C53CA850B86ULL, "w_bonepitdestroyed_a3_228_0.msh"},
	{0x9F3E6DC948583C85ULL, "512513954.ogg"},
	{0x9F3E80E457373E1DULL, "dae000_arishok6067837.cl"},
	{0x9F3F2DAAD43F3F27ULL, "lt_0_lowtown_a3_n_c_32_0_ll.mmh"},
	{0x9F3FA2B24F44F264ULL, "lgt310ip_body_3.ncs"},
	{0x9F3FF72E943396B6ULL, "6174051_m.fxe"},
	{0x9F40558ACAEB7565ULL, "6013377_m.fxe"},
	{0x9F408B1E0C5B0BB7ULL, "6039810_m.fxe"},
	{0x9F40D017B654AFD3ULL, "sib131_letter_react6127328.cl"},
	{0x9F40FAFD8E90E32FULL, "gen_im_arm_cht_lgt_set_high.uti"},
	{0x9F414D8E70D84AADULL, "280217728.ogg"},
	{0x9F415875A2F37631ULL, "vxht_0_hightown_4as_0_0.vlm"},
	{0x9F4223455FEB5A4FULL, "484430398.ogg"},
	{0x9F4243929B817E17ULL, "vrc321_party_barks6095753.cl"},
	{0x9F42FFB21D7DA51AULL, "and100_anders6096056.cl"},
	{0x9F432B710D4A0C71ULL, "19466412.ogg"},
	{0x9F4367D1F268DA4CULL, "26514329.ogg"},
	{0x9F437B838E024B99ULL, "al_warrior_t05.gda"},
	{0x9F44B456AF59F725ULL, "12881478.ogg"},
	{0x9F44CE1D28D234F1ULL, "em_gen_act2.mor"},
	{0x9F4525C0E7697624ULL, "vxlt_undercitydungeon_1ib_0_1.vlm"},
	{0x9F45B3E7281F0A38ULL, "47822683.ogg"},
	{0x9F45E0561B960186ULL, "hlac_w_encounter3_-65533_0z.dds"},
	{0x9F465028C867A995ULL, "6138799_m.fxe"},
	{0x9F465BD8FDE1F906ULL, "vfx_blt_lava_small.wbd"},
	{0x9F4681467BBBD4B8ULL, "vxlt_darktown_2sa_0_0.vlm"},
	{0x9F47B0920182157DULL, "w_encounter2_57_0.phy"},
	{0x9F47D924A0949AEDULL, "vxca_caverns1_mines_inf_0_0.vlm"},
	{0x9F47FAE0574A4E71ULL, "615382297.ogg"},
	{0x9F48C2EC3291C117ULL, "6110874_m.fxe"},
	{0x9F48FD4D41AD3E7AULL, "face_face_p5.psh.fxo.dx11.assembly"},
	{0x9F49A277F174AEB7ULL, "ht_keep_3731.anb"},
	{0x9F49A62E074AB500ULL, "nat380_nathaniel6066414.cl"},
	{0x9F4A33B2C38E682FULL, "406707698.ogg"},
	{0x9F4A5B43AF266238ULL, "1162594.ogg"},
	{0x9F4A6F3FB2B5EC7FULL, "6100242_m.fxe"},
	{0x9F4AE3A5E5D15A6CULL, "56598603.ogg"},
	{0x9F4B2460006C437EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_reaction.evt"},
	{0x9F4BC2078FBD8444ULL, "508617757.ogg"},
	{0x9F4C85F2A846AD67ULL, "10111861.ogg"},
	{0x9F4CA6347E51059BULL, "77952537.ogg"},
	{0x9F4CEB6D503A3140ULL, "vxw_bonepitdestroyed_a3_1ja_0_0.vlm"},
	{0x9F4D0D3832EF362EULL, "6121007_f.fxe"},
	{0x9F4D16BF2F407117ULL, "6113853_m.fxe"},
	{0x9F4D1789A28ABB0DULL, "vxlt_hangedman_n_171_0_0.vlm"},
	{0x9F4D27E17A95D43BULL, "539553179.ogg"},
	{0x9F4D74FEA0704427ULL, "575913203.ogg"},
	{0x9F4D76DD990BB1E6ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.impa.evt"},
	{0x9F4DB19201868209ULL, "w_encounter2_57_0.rml"},
	{0x9F4DD9957DED633BULL, "ts_slavers_arm_props.gda"},
	{0x9F4DDD08914AB578ULL, "gen00fl_varric6099411.cl"},
	{0x9F4DDDAB70E9D373ULL, "hlac_dr_0_deeproads_5838_1x.dds"},
	{0x9F4DFCE479A7C7ECULL, "ser120_ghyslain6104240.cl"},
	{0x9F4EBD5336A97928ULL, "vxlt_gallowstemplar_n_3rl_0_0.vlm"},
	{0x9F4ED3EF61F669A6ULL, "6104953_m.fxe"},
	{0x9F4F0FD6A97EF128ULL, "6090456_m.fxe"},
	{0x9F4F4B17EF2E6108ULL, "ser221_gascard6128295.cl"},
	{0x9F4F820FA03A2B1CULL, "dae000_serendipity6161359.cl"},
	{0x9F4FA9F5A07F2D97ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railplain.mmh"},
	{0x9F4FADF5A07F34A4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railplain.mao"},
	{0x9F4FBFF5A07F5309ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railplain.msh"},
	{0x9F4FEFB6B406487FULL, "vxw_bonepitmining_a2_1xg_0_1.vlm"},
	{0x9F508922C8C04E16ULL, "zz_store_dae210_store5.utm"},
	{0x9F50AEAA058AD901ULL, "hlac_lt_gallowscourtyard_n_-131073_1y.dds"},
	{0x9F51658BC8624990ULL, "vxca_caverns1_mines_j60_0_0.vlm"},
	{0x9F518816E544F8F3ULL, "vxlt_0_lowtown_ep3_0_0.vlm"},
	{0x9F522452AB2755D8ULL, "seb321_leliana6180923.cl"},
	{0x9F523B60007288A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_reaction.gad"},
	{0x9F52AA0FBA724955ULL, "6135679_m.fxe"},
	{0x9F5337C492BFAA63ULL, "vxht_mansion2b_1b5_0_0.vlm"},
	{0x9F539B2261932C9DULL, "vxht_brothel_n_6au_0_0.vlm"},
	{0x9F53AD894E9154D4ULL, "hlac_lt_0_lowtown_n_65538_0y.dds"},
	{0x9F53B428E5C01254ULL, "618911879.ogg"},
	{0x9F53D0CA3D4FEFD6ULL, "496864319.ogg"},
	{0x9F53EBAF56DFF955ULL, "vxht_0_hightown_a2_akd_0_0.vlm"},
	{0x9F5424F5C568FC0DULL, "ca_caverns1_26035_ll.mmh"},
	{0x9F54762298D333DBULL, "mag100_karras6034911.cl"},
	{0x9F548DDD9911F731ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.impa.gad"},
	{0x9F54AABFD7D9199AULL, "vxca_caverns1_mines_e54_0_0.vlm"},
	{0x9F54B39FB306D5FAULL, "ave200_donnic_patrol_1.stg"},
	{0x9F54B409B4898D0AULL, "vxht_keep_63l_0_0.vlm"},
	{0x9F54C720B561137FULL, "lgt200cr_gang_2_ht.utc"},
	{0x9F54CB83C2CF3E1CULL, "vxlt_gallowstemplar_f_2p3_0_0.vlm"},
	{0x9F54FA5201D7BBCAULL, "542552579.ogg"},
	{0x9F5511F95A46D450ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_p.mmh"},
	{0x9F552024BA6BBF86ULL, "vxlt_darktown_2v0_0_0.vlm"},
	{0x9F55A2BD4F2DD4FCULL, "434617713.ogg"},
	{0x9F55A918933F3359ULL, "llac_lt_0_lowtown_docks_n_-65539_5.dds"},
	{0x9F55C3361383E081ULL, "llac_ht_keep_n_3732_6.dds"},
	{0x9F5623402B6CCBDBULL, "vxw_bonepitdestroyed_a3_do_0_0.vlm"},
	{0x9F564B01928AB36DULL, "vxca_caverns1_mines_ike_0_0.vlm"},
	{0x9F566CC62B4C7D61ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medh_0n.dds"},
	{0x9F5674E981A69954ULL, "hlac_w_sundermount_327683_0z.dds"},
	{0x9F56DD1B97A49405ULL, "vxlt_gamlenshouse_n_b3_0_0.vlm"},
	{0x9F56F6B1861E7E9DULL, "779136044.ogg"},
	{0x9F57750BA782F8B6ULL, "353536058.ogg"},
	{0x9F578761F0D9585EULL, "art\\characters\\creatures\\mubari\\animation\\com_gag.evt"},
	{0x9F57D9A3A7BD15F0ULL, "vxdr_primevaltomb_8o9_0_0.vlm"},
	{0x9F57E9F385669E27ULL, "follower_banter6173364.cl"},
	{0x9F580FE25C22FB13ULL, "330461138.ogg"},
	{0x9F5831C9A654EE6BULL, "6129917_m.fxe"},
	{0x9F5851EA752CCB71ULL, "vxht_0_hightown_cb5_0_0.vlm"},
	{0x9F588702D0218FDFULL, "hlac_lt_0_lowtown_n_-65536_0y.dds"},
	{0x9F5897A1F98CB4AAULL, "vxdr_primevaltomb_81x_0_1.vlm"},
	{0x9F58A592019075E9ULL, "w_encounter2_57_0.mao"},
	{0x9F58AC0A50FF7D7AULL, "vxlt_gallowscourtyard_n_46s_0_0.vlm"},
	{0x9F58B19201908982ULL, "w_encounter2_57_0.mmh"},
	{0x9F58B39201908D38ULL, "w_encounter2_57_0.msh"},
	{0x9F58C8480D339387ULL, "hlac_lt_0_lowtown_docks_n_-131073_4z.dds"},
	{0x9F58CE8B254CF4BFULL, "6103575_m.fxe"},
	{0x9F58D6611EF0DA82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_malepc.gad"},
	{0x9F590C60007857F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_reaction.ani"},
	{0x9F5941152EEF5B5CULL, "zz_qun_debug6028243.cl"},
	{0x9F594866FE817E4CULL, "hlac_lt_0_lowtown_65535_6y.dds"},
	{0x9F597D357615436AULL, "vxht_mansion_n_4w9_0_0.vlm"},
	{0x9F59B04A3197C1CCULL, "qun211cr_commoner1.utc"},
	{0x9F5A582D500A272EULL, "vxlt_0_lowtown_a2_n_c_e2y_0_0.vlm"},
	{0x9F5ABDF30324A9E8ULL, "38210105.ogg"},
	{0x9F5AD2A2355B68FDULL, "ss_alienage.dds"},
	{0x9F5B425537FEAC75ULL, "plotactionsets.gda"},
	{0x9F5B5D4A423049C0ULL, "vxw_bonepitdestroyed_a3_1on_0_0.vlm"},
	{0x9F5B8EDD991817DFULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.impa.ani"},
	{0x9F5B999866E99132ULL, "vxlt_warehouse_n_5w_0_0.vlm"},
	{0x9F5BA00EDF0F40A4ULL, "hlac_ht_mansion2_818_40z.dds"},
	{0x9F5C1DDAB2CCD2B8ULL, "464485643.ogg"},
	{0x9F5C7810DA43DF78ULL, "mcr370_templar16087631.cl"},
	{0x9F5C937673C918E4ULL, "hlac_w_bonepitmining_a2_327680_0x.dds"},
	{0x9F5CC5979D964638ULL, "6117696_m.fxe"},
	{0x9F5DE81D29D5B4F6ULL, "967446066.ogg"},
	{0x9F5E7EFB5EBC856EULL, "vxht_mansion2_3ec_0_0.vlm"},
	{0x9F5E9E61F0DF9D89ULL, "art\\characters\\creatures\\mubari\\animation\\com_gag.gad"},
	{0x9F5EB144753F42A3ULL, "w_sundermount_115_0.rml"},
	{0x9F5EBF12EC712C9CULL, "dae000_merchant_access_hf6162952.cl"},
	{0x9F5F585F3A4BEDE7ULL, "hlac_lt_0_lowtown_a2_n_c_65536_8x.dds"},
	{0x9F5F686A184E5FE9ULL, "gen00fl_bethany6101388.cl"},
	{0x9F5F7C7A1F97451EULL, "6027342_m.fxe"},
	{0x9F5F7E0EC612D5A6ULL, "562792988.ogg"},
	{0x9F5FF85E83BDF495ULL, "vxca_caverns2prc_91t_0_0.vlm"},
	{0x9F5FF9611EF73579ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_malepc.evt"},
	{0x9F60685075B38997ULL, "hairalpha_hair_p0.vsh.fxo"},
	{0x9F60DB1AABB148E2ULL, "345645892.ogg"},
	{0x9F61292746C47EF0ULL, "vxlt_0_lowtown_gck_0_0.vlm"},
	{0x9F613F61DC287B9FULL, "art\\characters\\playercharacter\\humanmale\\boo_nuda.msh"},
	{0x9F614561DC288641ULL, "art\\characters\\playercharacter\\humanmale\\boo_nuda.mmh"},
	{0x9F615161DC289AE2ULL, "art\\characters\\playercharacter\\humanmale\\boo_nuda.mao"},
	{0x9F61ACA6BC734B8EULL, "vxlt_gallowstemplar_n_436_0_0.vlm"},
	{0x9F61CAE04D750749ULL, "gen00fl_isabela6102460.cl"},
	{0x9F620CFF5389451DULL, "vxht_brothel_n_6jt_0_0.vlm"},
	{0x9F6223DF96F2FB86ULL, "hlac_lt_0_lowtown_a2_131071_4x.dds"},
	{0x9F6234F95A519A6FULL, "art\\vfx\\combat\\vfx_m_frc_attwave_p.anb"},
	{0x9F624DFF8A2619FCULL, "gen00fl_dog6170432.cl"},
	{0x9F6284861B7878D3ULL, "6111070_f.fxe"},
	{0x9F639201AC61D806ULL, "vxlt_darktown_2y1_0_0.vlm"},
	{0x9F63BF52C6ADE806ULL, "vxlt_0_lowtown_docks_a3_alp_0_0.vlm"},
	{0x9F63D8F253AB9049ULL, "6041851_m.fxe"},
	{0x9F63FC99284A7306ULL, "591332852.ogg"},
	{0x9F645A24999A6534ULL, "926844023.ogg"},
	{0x9F64B8447543BA1BULL, "w_sundermount_115_0.phy"},
	{0x9F64E822754F92EEULL, "6061227_m.fxe"},
	{0x9F6522D6DD70D8CCULL, "6164254_m.fxe"},
	{0x9F656F61F0E56CD7ULL, "art\\characters\\creatures\\mubari\\animation\\com_gag.ani"},
	{0x9F6590A361047230ULL, "6153762_m.fxe"},
	{0x9F65EF1F0AB52B12ULL, "vxht_brothel_n_6gx_0_0.vlm"},
	{0x9F65FE33F91C0658ULL, "688490542.ogg"},
	{0x9F66096AC95106EEULL, "6104184_m.fxe"},
	{0x9F6615366533030CULL, "vxlt_0_lowtown_n_eow_0_1.vlm"},
	{0x9F67406F77437F34ULL, "131964343.ogg"},
	{0x9F6767CC60EE4087ULL, "zz_mag_debug.ncs"},
	{0x9F678C4ACEFB79DCULL, "338411765.ogg"},
	{0x9F67915C00D26AE7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_9.pwk"},
	{0x9F67AD14B15F2EC0ULL, "vxdr_primevaltomb_8fb_0_0.vlm"},
	{0x9F67CC0BDA091EB6ULL, "442618977.ogg"},
	{0x9F67FA6DAA19E2BAULL, "712061764.ogg"},
	{0x9F6847DF920B6DB0ULL, "6166601_m.fxe"},
	{0x9F685BCC56EEBDBDULL, "zz_dre_debug6205408.cl"},
	{0x9F686FDE088F6965ULL, "567584868.ogg"},
	{0x9F688D51C89A8651ULL, "vxlt_warehouse_n_148_0_0.vlm"},
	{0x9F68CA0A42003FAEULL, "vxlt_0_lowtown_a2_n_c_e7z_0_0.vlm"},
	{0x9F68F570DA38A46DULL, "vxlt_0_lowtown_a2_n_e1i_0_0.vlm"},
	{0x9F69371DB10D7B57ULL, "hlac_lt_0_lowtown_n_-65536_8y.dds"},
	{0x9F69583AD8344E85ULL, "6118054_m.fxe"},
	{0x9F69B35ED8E77AE0ULL, "hlac_ca_caverns2prc_5802_34y.dds"},
	{0x9F69CF255559AFAFULL, "225418933.ogg"},
	{0x9F6A0B7558DFA9B2ULL, "vxlt_warehouse_n_8p_0_0.vlm"},
	{0x9F6A19B5F5C73A39ULL, "llac_w_bonepit_131075_0.dds"},
	{0x9F6A3C188CBF07FDULL, "tm_war_d03.tnt"},
	{0x9F6A420BF0790E09ULL, "vxlt_gallowstemplar_f_65u_0_0.vlm"},
	{0x9F6A5262D9B4223DULL, "hlac_lt_darktown_-1_52z.dds"},
	{0x9F6A89A30D794EA0ULL, "vxht_mansion2_2e0_0_0.vlm"},
	{0x9F6AFD6332A39312ULL, "zz_dre_debug6168229.cl"},
	{0x9F6B40F218EE5E86ULL, "182392403.ogg"},
	{0x9F6B534E44742FC3ULL, "dae000_bartrand6027542.cl"},
	{0x9F6B6A98691109CCULL, "vxlt_0_lowtown_n_e0o_0_0.vlm"},
	{0x9F6BB05D6EDF3C38ULL, "seb321_leliana6197880.cl"},
	{0x9F6BEA60875C1DB8ULL, "gen00fl_merrill6151012.cl"},
	{0x9F6C7295F4D40359ULL, "art\\vfx\\creatures\\vfx_c_drag_tswipe_c.impa.ani"},
	{0x9F6C9D029666B5EEULL, "282989903.ogg"},
	{0x9F6CE3FA99F42B03ULL, "lt_gamlenshouse_n_mmap.dds"},
	{0x9F6D28EB67F4D0B9ULL, "6137686_m.fxe"},
	{0x9F6D7E1BBC67749DULL, "qalevel1.arl"},
	{0x9F6D8D6D7658A709ULL, "994067233.ogg"},
	{0x9F6DDFF595FA361CULL, "177406676.ogg"},
	{0x9F6DF1476B83BA21ULL, "hlac_ht_0_hightown_a2_65535_2x.dds"},
	{0x9F6E01611F038068ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_malepc.ani"},
	{0x9F6E0F15CEF64E5BULL, "6106639_m.fxe"},
	{0x9F6E136E769A3B42ULL, "6098296_m.fxe"},
	{0x9F6E2AF66D0A6CD0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtclosed.msh"},
	{0x9F6E38F66D0A84EAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtclosed.mmh"},
	{0x9F6E3CF66D0A8B21ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curtclosed.mao"},
	{0x9F6E410D1A228704ULL, "6076301_f.fxe"},
	{0x9F6E6BD0C93129F3ULL, "vxlt_0_lowtown_ef5_0_0.vlm"},
	{0x9F6E8DBBD0FB5D8BULL, "6081255_m.fxe"},
	{0x9F703937CAA0AD6CULL, "gen00fl_varric6118954.cl"},
	{0x9F709852C032DABCULL, "vxht_bartrand_n_5mo_0_0.vlm"},
	{0x9F70DAAE2BBD04ABULL, "vxlt_0_lowtown_docks_a3_bfd_0_0.vlm"},
	{0x9F70FC8F70C8FC51ULL, "6113635_m.fxe"},
	{0x9F714BD8FA25A798ULL, "vxw_woundedcoast_1bn_0_0.vlm"},
	{0x9F7159578AEE3846ULL, "hlac_ht_keep_n_3734_1x.dds"},
	{0x9F715ACC60F67F14ULL, "zz_mag_debug.cnv"},
	{0x9F7197C39875BE0AULL, "vxht_keep_69n_0_0.vlm"},
	{0x9F71AE8C19357183ULL, "hlac_ht_keep_4519_9z.dds"},
	{0x9F71F62023DFCB1CULL, "12308244.ogg"},
	{0x9F722045C9FDD0E2ULL, "6119831_f.fxe"},
	{0x9F7288796D2C8AF2ULL, "vxht_0_hightown_a2_b9z_0_0.vlm"},
	{0x9F7293C0EEA66AFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_left.gad"},
	{0x9F72F0234C176E00ULL, "vxlt_gallowstemplar_f_3qu_0_0.vlm"},
	{0x9F734D95F4D9E29BULL, "art\\vfx\\creatures\\vfx_c_drag_tswipe_c.impa.gad"},
	{0x9F737F7D1AB96B71ULL, "vxht_0_hightown_a3_ay3_0_0.vlm"},
	{0x9F7399933772C42CULL, "6067627_m.fxe"},
	{0x9F7485E94126A535ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgtp_0t.dds"},
	{0x9F751B2BE2803F4BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kitb_0t.dds"},
	{0x9F759F77F199E4CBULL, "ht_keep_3731.rml"},
	{0x9F75A551AAEFE8B2ULL, "568062812.ogg"},
	{0x9F75BF29B82E0A65ULL, "426769210.ogg"},
	{0x9F75CD74B6DB54FEULL, "gen00fl_aveline6139387.cl"},
	{0x9F76575A96664949ULL, "838058359.ogg"},
	{0x9F766CB0366D5975ULL, "275794713.ogg"},
	{0x9F76980EEE661935ULL, "6102783_m.fxe"},
	{0x9F76AA05E92CB680ULL, "gen00flst_varric_dreintro.stg"},
	{0x9F76C01BCA5B61DBULL, "6039714_f.fxe"},
	{0x9F770DA27432CCF2ULL, "6106700_m.fxe"},
	{0x9F771E1FD07FB300ULL, "lgt160_ntb_elf_conflict6163349.cl"},
	{0x9F77376E6C55C98AULL, "gen00fl_aveline6116704.cl"},
	{0x9F77CA8B1D95F558ULL, "ave221cr_brennan.utc"},
	{0x9F77D22506D389E9ULL, "716391321.ogg"},
	{0x9F781B94C832386AULL, "6118039_m.fxe"},
	{0x9F7837D183A083F2ULL, "6169794_m.fxe"},
	{0x9F7860B921A10DBEULL, "373441938.ogg"},
	{0x9F786D99788C52C2ULL, "gen00fl_merrill6116011.cl"},
	{0x9F789688C7BAF1C6ULL, "qun200tr_follower_amb_end.utt"},
	{0x9F7896C0EEAADB11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_left.evt"},
	{0x9F7898DF19828A30ULL, "44441647.ogg"},
	{0x9F78FB7FFF6F6720ULL, "vxht_mansion2_2n1_0_0.vlm"},
	{0x9F79063881D7B94CULL, "6103330_m.fxe"},
	{0x9F790B323D35BB88ULL, "6101794_m.fxe"},
	{0x9F7945AD9C6A1C39ULL, "6207860_m.fxe"},
	{0x9F79BA6E994D64D4ULL, "dae000_bodhan6129119.cl"},
	{0x9F79C477AB1C9FF7ULL, "vxht_mansion2b_32k_0_0.vlm"},
	{0x9F79EB0EA32C5526ULL, "pro000_colonel_ewald6039943.cl"},
	{0x9F7A5DE260C9061CULL, "6125640_m.fxe"},
	{0x9F7A6A95F4E03280ULL, "art\\vfx\\creatures\\vfx_c_drag_tswipe_c.impa.evt"},
	{0x9F7AC5AE683ED3BEULL, "r_u_c.dds"},
	{0x9F7B2605EE6BE0ABULL, "hlac_lt_0_lowtown_a2_n_65536_0z.dds"},
	{0x9F7B55065165510BULL, "zz_mer_debug6051471.cl"},
	{0x9F7BA8842B27AC0CULL, "554483861.ogg"},
	{0x9F7BC905D034C371ULL, "hlac_ht_0_hightown_a2_-65536_0x.dds"},
	{0x9F7C281947BA4390ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keepthin_1.mao"},
	{0x9F7C362E788F204BULL, "6068504_m.fxe"},
	{0x9F7C648A03AD27BDULL, "one300_zevran6086573.cl"},
	{0x9F7C7945D7412AA5ULL, "lowtown_male_lowlife1_sta6164984.cl"},
	{0x9F7C8DAB7AF3451CULL, "6076002_f.fxe"},
	{0x9F7D77E0834E828EULL, "454166240.ogg"},
	{0x9F7DA256D043F31FULL, "6086984_m.fxe"},
	{0x9F7E06B2A1AF6810ULL, "38953857.ogg"},
	{0x9F7FB6AA98BF4382ULL, "75270535.ogg"},
	{0x9F7FBDB5EC1BC018ULL, "vxw_woundedcoast_1mo_0_0.vlm"},
	{0x9F80347B2A036DA5ULL, "6111304_f.fxe"},
	{0x9F80844843975C93ULL, "vxht_keep_n_67v_0_0.vlm"},
	{0x9F81082C6F76F463ULL, "hlac_lt_undercityserialkiller_-1_15y.dds"},
	{0x9F810CAAEDB0537BULL, "art\\characters\\creatures\\mubari\\animation\\basepose.evt"},
	{0x9F811ABD8ED5943DULL, "893321412.ogg"},
	{0x9F819F3745AF3D16ULL, "follower_barks6153230.cl"},
	{0x9F8285389724E4EFULL, "gen00fl_varric6033722.cl"},
	{0x9F82F8D690B61D3DULL, "738009538.ogg"},
	{0x9F832B6148B3D29DULL, "vxw_sundermount_25e_0_0.vlm"},
	{0x9F83AAA2A4F12065ULL, "6128942_f.fxe"},
	{0x9F843C10BD3AE9B2ULL, "119949702.ogg"},
	{0x9F84CF8BCBA781A2ULL, "hm_sebastian_vael.mor"},
	{0x9F84D58E93A8C219ULL, "art\\characters\\creatures\\highdragon\\cn_bdy_fdr_arm_1.mao"},
	{0x9F84F556296EEDE8ULL, "6093538_m.fxe"},
	{0x9F8509EBE05C31B5ULL, "6107883_m.fxe"},
	{0x9F850B2853C35421ULL, "vxht_brothel_n_5g8_0_0.vlm"},
	{0x9F8554D01617B11EULL, "714207834.ogg"},
	{0x9F85796E8D4AC642ULL, "355480331.ogg"},
	{0x9F859B275499984BULL, "vxht_mansion2b_2ht_0_0.vlm"},
	{0x9F8611D68DF0E9D5ULL, "6196683_m.fxe"},
	{0x9F8683FC51EE2EB8ULL, "zz_and_debug6134068.cl"},
	{0x9F86AEC0EEB741E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_left.ani"},
	{0x9F879337B514F795ULL, "vxht_brothel_n_5k0_0_0.vlm"},
	{0x9F87B67A398ED15FULL, "332027116.ogg"},
	{0x9F87B98430F891DDULL, "llac_lt_0_lowtown_docks_n_-65538_0.dds"},
	{0x9F87F6FCD36AF554ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2.phy"},
	{0x9F8805AAEDB665BCULL, "art\\characters\\creatures\\mubari\\animation\\basepose.gad"},
	{0x9F8820A6C3B3C1CAULL, "ico_1h_axe_3.dds"},
	{0x9F884798D5998F72ULL, "dae021_brothel_onenter.cnv"},
	{0x9F88A6989885941BULL, "vxca_cavern1_haunted_dq8_0_0.vlm"},
	{0x9F88CFBF52BF1E9CULL, "6122740_m.fxe"},
	{0x9F88F35FFF7326DDULL, "gen00fl_merrill6105656.cl"},
	{0x9F890D79A98EC21AULL, "hlac_lt_shop_-1_3y.dds"},
	{0x9F89DC3DBCC94789ULL, "vxlt_0_lowtown_a2_gwv_0_0.vlm"},
	{0x9F8A467876F7E3D8ULL, "ave121_brennan6105836.cl"},
	{0x9F8B10241BF6C445ULL, "hlac_ht_0_hightown_65535_3y.dds"},
	{0x9F8B67AA2A0FB99CULL, "hlac_w_sundermount_-131072_0z.dds"},
	{0x9F8B6998D59BFECAULL, "dae021_brothel_onenter.bnk"},
	{0x9F8BBB620FD57537ULL, "280876585.ogg"},
	{0x9F8BD8A6FE27AB5BULL, "hlac_w_encounter1_-65537_0y.dds"},
	{0x9F8CB36A60AA8CB0ULL, "6087632_m.fxe"},
	{0x9F8CB5F20788D3EFULL, "175234883.ogg"},
	{0x9F8CFF385A980C63ULL, "vxht_mansion2b_1v9_0_0.vlm"},
	{0x9F8D0BB78CD3A21BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\kneeling_faint_loop.ani"},
	{0x9F8D4661DC4DC2BEULL, "art\\characters\\playercharacter\\humanmale\\boo_nuda.phy"},
	{0x9F8D796EAD2D9492ULL, "da2terrain_terrain_p0.dsh.fxo.dx11"},
	{0x9F8DA6B5F9725B03ULL, "art\\cinematics\\additional animation\\cs_malenobleiii_cinematicanim.ani"},
	{0x9F8DAC448877020DULL, "6189034_m.fxe"},
	{0x9F8DD179CDAEF9EEULL, "782499441.ogg"},
	{0x9F8E2F92DE11D898ULL, "vxw_woundedcoast_1hl_0_0.vlm"},
	{0x9F8E31726B39144FULL, "vxht_0_hightown_a2_a2m_0_0.vlm"},
	{0x9F8EEB52C66261CDULL, "l_dae210ar_docks_nt.wbd"},
	{0x9F8EF625358D7513ULL, "vxht_keep_n_62y_0_0.vlm"},
	{0x9F8EF65D5C5610FCULL, "6175476_m.fxe"},
	{0x9F8F04AAEDBC82BAULL, "art\\characters\\creatures\\mubari\\animation\\basepose.ani"},
	{0x9F8F26221CDFD5A7ULL, "hlac_w_woundedcoast_-196605_0z.dds"},
	{0x9F8F2FE9EB33BF12ULL, "6109730_m.fxe"},
	{0x9F8F667FA0D31C52ULL, "6125721_m.fxe"},
	{0x9F8F8652E672628FULL, "mag110_vincento6017111.cl"},
	{0x9F8FD028C6ABECAAULL, "391481664.ogg"},
	{0x9F903394720F9CB2ULL, "static1_ds2.psh.fxo"},
	{0x9F9038D7138BD70EULL, "6105802_m.fxe"},
	{0x9F904DD016367426ULL, "m_mag121ar_store.bnk"},
	{0x9F90E3F1F26B1D74ULL, "hairalpha_hair_p3.vsh.fxo.dx11"},
	{0x9F90ED4BD56181F6ULL, "647122642.ogg"},
	{0x9F9100E4DE7BD760ULL, "594368328.ogg"},
	{0x9F910D77A614DE03ULL, "6121059_f.fxe"},
	{0x9F911878524040CEULL, "hlac_lt_0_lowtown_-65535_0z.dds"},
	{0x9F9128B0C02A0E3BULL, "w_bonepit_104_0.msh"},
	{0x9F913AB0C02A2C86ULL, "w_bonepit_104_0.mao"},
	{0x9F91439AB056C67CULL, "vxlt_gallowstemplar_f_2r5_0_0.vlm"},
	{0x9F9146B0C02A4125ULL, "w_bonepit_104_0.mmh"},
	{0x9F914C62B79A99F4ULL, "hlac_w_bonepit_0_11x.dds"},
	{0x9F919F44756A7DBAULL, "w_sundermount_115_0.msh"},
	{0x9F91AD44756A95F3ULL, "w_sundermount_115_0.mao"},
	{0x9F91B144756A9C00ULL, "w_sundermount_115_0.mmh"},
	{0x9F91D09604CC76C2ULL, "dae321ar_keep.are"},
	{0x9F9229F41787D739ULL, "199211257.ogg"},
	{0x9F92B584F348EFB0ULL, "vxlt_gallowstemplar_34q_0_0.vlm"},
	{0x9F92E2765B48BC5DULL, "vxlt_0_lowtown_a2_n_c_eod_0_1.vlm"},
	{0x9F92EDFCD374ED0FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2.mmh"},
	{0x9F9343E49FA0B922ULL, "zz_isa_debug6134419.cl"},
	{0x9F935C321C81B315ULL, "ht_bartrand_n_4152_ll.mmh"},
	{0x9F9393070C726A86ULL, "hlac_ca_caverns2prc_5802_33z.dds"},
	{0x9F93E801C0FFCF8AULL, "vxw_encounter2_camp_or_0_1.vlm"},
	{0x9F93E8988FFCE984ULL, "vxlt_gallowscourtyard_7ad_0_0.vlm"},
	{0x9F940D04468FB0CBULL, "vxht_mansion2b_2ew_0_0.vlm"},
	{0x9F9448A42E565CAAULL, "6016742_m.fxe"},
	{0x9F947C48E905C762ULL, "art\\vfx\\materials\\vfx_cen_splash.mao"},
	{0x9F949C0A0064AD53ULL, "415797152.ogg"},
	{0x9F956D38F40AAF29ULL, "vxht_mansion2_29q_0_0.vlm"},
	{0x9F959715D84CC4F5ULL, "ave331cr_insurgent_m.utc"},
	{0x9F9633AA42050437ULL, "197393858.ogg"},
	{0x9F972177840018E6ULL, "mcr370_guardsman6086984.cl"},
	{0x9F97BFAFA1BFD5D4ULL, "zz_dae_debug6050747.cl"},
	{0x9F9833E79751E8F5ULL, "character_character_p99.psh.fxo"},
	{0x9F986B31820FB305ULL, "690763806.ogg"},
	{0x9F9889301AE32B47ULL, "vxlt_darktown_5nf_0_0.vlm"},
	{0x9F98AD0FF1C45275ULL, "vxw_bonepitdestroyed_a3_1wv_0_1.vlm"},
	{0x9F99A5BCC907FFD5ULL, "gen00fl_merrill6129849.cl"},
	{0x9F99E482F8D9BC1DULL, "hlac_ca_caverns1_mines_24423_2x.dds"},
	{0x9F9A171F640C0AC1ULL, "vxlt_blackemporium_vp_0_0.vlm"},
	{0x9F9A2D928CD48849ULL, "art\\vfx\\creatures\\vfx_c_corpse_c.mmh"},
	{0x9F9B5EE64C12DA8BULL, "6108620_m.fxe"},
	{0x9F9BC79604D4BB95ULL, "dae321ar_keep.lst"},
	{0x9F9BFE20EA533684ULL, "vxht_keep_5k5_0_0.vlm"},
	{0x9F9C4525E8EE821BULL, "6001202_m.fxe"},
	{0x9F9C6827F1D5288FULL, "vxht_0_hightown_n_aic_0_0.vlm"},
	{0x9F9D4F738B8D0884ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobc_0t.dds"},
	{0x9F9D6D5E8EBF70B7ULL, "6118702_m.fxe"},
	{0x9F9DFDE63CA0DED3ULL, "6085570_m.fxe"},
	{0x9F9E5DAABE51B1FBULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockcxxl.msh"},
	{0x9F9E6FAABE51D046ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockcxxl.mao"},
	{0x9F9E7874BE05A0CEULL, "768669422.ogg"},
	{0x9F9E7BAABE51E4E5ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockcxxl.mmh"},
	{0x9F9EE33A67094817ULL, "follower_banter6155856.cl"},
	{0x9F9EFF81CBF3978EULL, "6094509_f.fxe"},
	{0x9F9F5C378251B9D0ULL, "6111213_f.fxe"},
	{0x9F9F6E012526B2DDULL, "hlac_ht_0_hightown_a3_-131071_1x.dds"},
	{0x9F9F864CB4831B89ULL, "413761027.ogg"},
	{0x9F9FA19695D845EDULL, "llac_ht_mansion_n_3364_1.dds"},
	{0x9F9FB9660924339FULL, "gen_im_arm_cht_comb.uti"},
	{0x9F9FEF9B905B3974ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_cobble.phy"},
	{0x9FA04C065D174034ULL, "vxht_brothel_n_57q_0_0.vlm"},
	{0x9FA059718CD030C7ULL, "mag100_cullen6082272.cl"},
	{0x9FA0C0AC7C236AADULL, "6059331_m.fxe"},
	{0x9FA0FCCD5E6F111CULL, "6139444_f.fxe"},
	{0x9FA132B78CE48DA5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\kneeling_faint_loop.gad"},
	{0x9FA16EC04562C58FULL, "hlac_dr_primevaltomb_665_0z.dds"},
	{0x9FA1BDB5F9832BEDULL, "art\\cinematics\\additional animation\\cs_malenobleiii_cinematicanim.gad"},
	{0x9FA1FE4DA79E392DULL, "6090371_f.fxe"},
	{0x9FA2BF9604DACC84ULL, "dae321ar_keep.ncs"},
	{0x9FA345AC81577B75ULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_wep_ggg_0n.dds"},
	{0x9FA3D60488F09EABULL, "712893702.ogg"},
	{0x9FA4C00346FFD363ULL, "art\\vfx\\combat\\vfx_m_sta_ele_range_c.impa.ani"},
	{0x9FA4C5AACF2AC994ULL, "gen00fl_fenris6106661.cl"},
	{0x9FA5102CCAB01FECULL, "303348557.ogg"},
	{0x9FA52FB580A004ACULL, "gen_im_trs_slv_sta_phy_r0.uti"},
	{0x9FA5331E90DB8332ULL, "420881115.ogg"},
	{0x9FA55CD37DE80961ULL, "ave200_aveline_bar6116194.cl"},
	{0x9FA6A4FC803C7223ULL, "67284244.ogg"},
	{0x9FA7163315656524ULL, "mcr370_meredith6084432.cl"},
	{0x9FA74716E80319E2ULL, "567128177.ogg"},
	{0x9FA75CBA9DF5A1DCULL, "hlac_w_bonepitdestroyed_a3_0_0x.dds"},
	{0x9FA7EFFC9A1BCDA4ULL, "vxlt_foundry_n_xx_0_0.vlm"},
	{0x9FA7F93579B29C11ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft_3.mao"},
	{0x9FA7FD069EDE9845ULL, "905353299.ogg"},
	{0x9FA82919CF46B674ULL, "vxlt_0_lowtown_docks_n_b80_0_1.vlm"},
	{0x9FA83AB07A3FA9BBULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_lp.ani"},
	{0x9FA8506579C3C719ULL, "follower_barks6153693.cl"},
	{0x9FA95EE71EA5AA31ULL, "6126828_m.fxe"},
	{0x9FA9CC9B90635271ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_cobble.msh"},
	{0x9FA9DA9B9063698CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_cobble.mao"},
	{0x9FA9E69B90637E2FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_cobble.mmh"},
	{0x9FAA10E9F28166B5ULL, "6116135_m.fxe"},
	{0x9FAA75D1087233B3ULL, "isa311_isabela6091832.cl"},
	{0x9FAA9877C38F2D23ULL, "vxlt_blackemporium_97_0_0.vlm"},
	{0x9FAABAA5C9134A07ULL, "vxht_0_hightown_n_a2j_0_0.vlm"},
	{0x9FAAFD0F9AFA4DF1ULL, "vxlt_0_lowtown_n_exf_0_0.vlm"},
	{0x9FAC383809DD0E71ULL, "vxdr_0_deeproads_8ve_0_0.vlm"},
	{0x9FAC3B50B8371804ULL, "6165971_f.fxe"},
	{0x9FAC689D0E510229ULL, "gen00fl_aveline6106384.cl"},
	{0x9FACAFDDC407F066ULL, "vxht_mansion_n_750_0_0.vlm"},
	{0x9FACCD52D8079B23ULL, "6149480_f.fxe"},
	{0x9FACF100D76B6AABULL, "gen00fl_sebastian6174666.cl"},
	{0x9FAE0EF228CE4BF5ULL, "608237616.ogg"},
	{0x9FAE2773AD25E154ULL, "hlac_lt_0_lowtown_docks_a3_-65539_0x.dds"},
	{0x9FAE5ACA4B0A4212ULL, "513283866.ogg"},
	{0x9FAE5F40EE8D44CFULL, "fen210_orana6109713.cl"},
	{0x9FAE824A65BBE295ULL, "172117974.ogg"},
	{0x9FAE89DCF0924C0EULL, "215299442.ogg"},
	{0x9FAE916A3DB9FAA8ULL, "vxlt_0_lowtown_n_daq_0_0.vlm"},
	{0x9FAEA1E436F32BBAULL, "art\\levels\\races\\design\\_textures\\prp_herbs_e.dds"},
	{0x9FAF3D9D8C014B8DULL, "dae000_madam_lusine6151707.cl"},
	{0x9FB0B050992E6590ULL, "fen210stg_dying_magister.stg"},
	{0x9FB143300D9520F8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medd_0d.dds"},
	{0x9FB1D803470A865AULL, "art\\vfx\\combat\\vfx_m_sta_ele_range_c.impa.evt"},
	{0x9FB210509E401697ULL, "627514240.ogg"},
	{0x9FB21A71F753F6F2ULL, "203861811.ogg"},
	{0x9FB2849CCA5CB7B1ULL, "6125570_f.fxe"},
	{0x9FB2C6BF5D5392DFULL, "6097844_f.fxe"},
	{0x9FB32F221044EC7FULL, "467933201.ogg"},
	{0x9FB41EBA57B90E35ULL, "189199048.ogg"},
	{0x9FB449ABF1464178ULL, "741504324.ogg"},
	{0x9FB4D449F2AD1843ULL, "1008868642.ogg"},
	{0x9FB542B07A4A41B2ULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_lp.evt"},
	{0x9FB54A4B6E4BEBABULL, "cod_mgc_maleficar.utp"},
	{0x9FB5E71FD2FD819AULL, "lt_0_lowtown_docks_a3_245_0.mmh"},
	{0x9FB5EB1FD2FD88D1ULL, "lt_0_lowtown_docks_a3_245_0.mao"},
	{0x9FB5F91FD2FDA0E0ULL, "lt_0_lowtown_docks_a3_245_0.msh"},
	{0x9FB69AF6C13CCEF4ULL, "vxlt_0_lowtown_docks_n_b57_0_1.vlm"},
	{0x9FB6EDC323CBC49BULL, "hlac_ht_keep_3732_2y.dds"},
	{0x9FB73FBF686C11D1ULL, "vxht_0_hightown_a2_9x9_0_0.vlm"},
	{0x9FB75DAF46064B02ULL, "gen_im_arm_cht_hvy_start.uti"},
	{0x9FB784839861A317ULL, "zz_vault_debug6203258.cl"},
	{0x9FB7BAF459CE943EULL, "qun160_party_comment_26061240.cl"},
	{0x9FB7F1353BDF3BD2ULL, "activatehostile.ncs"},
	{0x9FB81E99198376FCULL, "hlac_lt_0_lowtown_n_-262144_4x.dds"},
	{0x9FB87E01D177C668ULL, "6126551_m.fxe"},
	{0x9FB8BEA193B479E9ULL, "hlac_w_bonepit_458756_0x.dds"},
	{0x9FB8D7034710A34DULL, "art\\vfx\\combat\\vfx_m_sta_ele_range_c.impa.gad"},
	{0x9FB8EEBFBB1A0827ULL, "23620552.ogg"},
	{0x9FBA3CD626C997C5ULL, "vxlt_gallowscourtyard_4uv_0_0.vlm"},
	{0x9FBA45EF6EC29A56ULL, "dae000_gamlen6124242.cl"},
	{0x9FBAA2F8E59F984BULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_hurl.evt"},
	{0x9FBB219B7CE188ACULL, "vxht_0_hightown_a3_5jc_0_0.vlm"},
	{0x9FBB272616B78041ULL, "6089781_m.fxe"},
	{0x9FBB7034B3AAFB34ULL, "store_dae270cr_mage.utm"},
	{0x9FBB73196ECCE955ULL, "hlac_lt_gallowstemplar_2393_1y.dds"},
	{0x9FBB9034AC2DF617ULL, "6116289_m.fxe"},
	{0x9FBBC8137A4247B5ULL, "hlac_lt_0_lowtown_-196608_1x.dds"},
	{0x9FBC21E44C745048ULL, "6126407_m.fxe"},
	{0x9FBC61B07A509545ULL, "art\\characters\\creatures\\mubari\\animation\\pp_rst_lp.gad"},
	{0x9FBC6BE70D43D1C4ULL, "w_woundedcoast_mmi.gff"},
	{0x9FBC758FB69737F9ULL, "vxw_encounter2_camp_jc_0_0.vlm"},
	{0x9FBC85CE4E3A7C04ULL, "vxlt_0_lowtown_docks_bbl_0_0.vlm"},
	{0x9FBD1F44DF5A4EE6ULL, "6053594_m.fxe"},
	{0x9FBD479758A1FE67ULL, "art\\vfx\\combat\\vfx_g_spider_att_f1_c.impa.gad"},
	{0x9FBE2806D7DACCD1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_4.gad"},
	{0x9FBE769FD81126F4ULL, "gen00fl_merrill6128975.cl"},
	{0x9FBF15F8E25853CCULL, "lit_up_05_l2.mao"},
	{0x9FBF3EBACACF6A03ULL, "sst_ss_templars_m_1.gda"},
	{0x9FBFEB87AC036587ULL, "497591468.ogg"},
	{0x9FC0E4B5486C077FULL, "vxht_keep_n_584_0_0.vlm"},
	{0x9FC14DD36AB0F361ULL, "6164071_m.fxe"},
	{0x9FC184893A7D24BAULL, "dre000_dougal6205764.cl"},
	{0x9FC1BBF8E5A5E10CULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_hurl.gad"},
	{0x9FC224AE28DE9EACULL, "w_woundedcoast_232_0.msh"},
	{0x9FC22AAE28DEA89EULL, "w_woundedcoast_232_0.mmh"},
	{0x9FC236AE28DEBD35ULL, "w_woundedcoast_232_0.mao"},
	{0x9FC32A928CF7887EULL, "art\\vfx\\creatures\\vfx_c_corpse_c.anb"},
	{0x9FC33C9758A65774ULL, "art\\vfx\\combat\\vfx_g_spider_att_f1_c.impa.evt"},
	{0x9FC3BFA9D9CCB5DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\aveline_inspect_blade.gad"},
	{0x9FC42906D7DF3A7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_4.ani"},
	{0x9FC46CD6B9E9752FULL, "6034343_m.fxe"},
	{0x9FC5399A946384CAULL, "llac_lt_gallowstemplar_f_2383_10.dds"},
	{0x9FC56434428C1130ULL, "58990616.ogg"},
	{0x9FC58865A423C8F9ULL, "6122367_m.fxe"},
	{0x9FC5F01921AFB510ULL, "zz_isa_debug6133988.cl"},
	{0x9FC65377A5D0AA49ULL, "vxht_0_hightown_9bg_0_0.vlm"},
	{0x9FC6B60E98B9CB82ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_lgtc_0s.dds"},
	{0x9FC6F52D46E3B230ULL, "vis121_saemus6118414.cl"},
	{0x9FC7599F42B30258ULL, "llac_lt_gallowstemplar_f_2394_9.dds"},
	{0x9FC78988C7FEAF9FULL, "qun200tr_follower_amb_end.ncs"},
	{0x9FC79F6BBD7D2E26ULL, "26835848.ogg"},
	{0x9FC7BAF8E5AA4B0AULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_hurl.ani"},
	{0x9FC850AE239BF64BULL, "408628010.ogg"},
	{0x9FC8EB5041AADBEEULL, "vxlt_0_lowtown_a2_n_c_el9_0_2.vlm"},
	{0x9FC9CD9DDA4BD7A8ULL, "t3_dog_ow2.tnt"},
	{0x9FC9D269CB66EFF6ULL, "zz_followers_banter6159549.cl"},
	{0x9FC9D64C23C9F90EULL, "55004958.ogg"},
	{0x9FC9DD6CBAA26FC2ULL, "isa321_velasco6117626.cl"},
	{0x9FC9FC61AD5F73C1ULL, "6147902_m.fxe"},
	{0x9FCA206EA94DC3F8ULL, "510801250.ogg"},
	{0x9FCABB3B91B725CBULL, "6146806_m.fxe"},
	{0x9FCAC09305843839ULL, "52233639.ogg"},
	{0x9FCAC2A9D9D2D9F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\aveline_inspect_blade.evt"},
	{0x9FCAE641289A52DFULL, "t3_wer_h03.tnt"},
	{0x9FCB4A10AC2584BDULL, "6041950_m.fxe"},
	{0x9FCB89DB56211B76ULL, "llac_lt_0_lowtown_docks_a3_-4_0.dds"},
	{0x9FCB8B852B9B9D1FULL, "129528101.ogg"},
	{0x9FCBEE9638394E9AULL, "mcr371_angry_mage_m.cnv"},
	{0x9FCC3A2A05602535ULL, "374150672.ogg"},
	{0x9FCC3A5304C7BDFCULL, "1001461236.ogg"},
	{0x9FCCD42238AA18D6ULL, "889203813.ogg"},
	{0x9FCD245BBB15C9A6ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgta_0n.dds"},
	{0x9FCD72DF91AC6579ULL, "lgt260cr_eau_tmp_roderick.utc"},
	{0x9FCDB09ED33CDB56ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestshield.msh"},
	{0x9FCDBA9ED33CEC54ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestshield.mmh"},
	{0x9FCDBE9ED33CF367ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestshield.mao"},
	{0x9FCE442AC35F967CULL, "mag211_marethari6025361.cl"},
	{0x9FCE6A4B2C5404D6ULL, "sse_ss_coterie_mage_m.gda"},
	{0x9FCEF21C7D2EE7ABULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.dura.ani"},
	{0x9FCF1096383BBEE2ULL, "mcr371_angry_mage_m.bnk"},
	{0x9FCFAB231EEF0376ULL, "fence_sa.dds"},
	{0x9FCFDA35D7FBADB0ULL, "vxlt_0_lowtown_by4_0_0.vlm"},
	{0x9FCFF552C69A09F7ULL, "l_dae210ar_docks_nt.bnk"},
	{0x9FD03C34A63A0949ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_loop.ani"},
	{0x9FD04FD0166C5BCCULL, "m_mag121ar_store.wbd"},
	{0x9FD089006FD33776ULL, "vxlt_gallowstemplar_n_393_0_0.vlm"},
	{0x9FD119E74B9944D0ULL, "vxlt_0_lowtown_a3_n_c_gcg_0_0.vlm"},
	{0x9FD1349758B286ADULL, "art\\vfx\\combat\\vfx_g_spider_att_f1_c.impa.ani"},
	{0x9FD20DD75F121889ULL, "6139117_f.fxe"},
	{0x9FD253331F899B48ULL, "vfx_base_one.gda"},
	{0x9FD282FAFFC4F5A6ULL, "dae300_alistair_king6134946.cl"},
	{0x9FD29D42F19C1B65ULL, "vxlt_0_lowtown_docks_b6u_0_0.vlm"},
	{0x9FD2C61ECA6C32B1ULL, "1004938718.ogg"},
	{0x9FD2EA5C60078689ULL, "vfxchunky0.psh.fxo"},
	{0x9FD2F322E077C771ULL, "zz_isa_debug6133161.cl"},
	{0x9FD3072C51F9598AULL, "6139877_f.fxe"},
	{0x9FD3C3EB887309D3ULL, "ser221_party_comments6209919.cl"},
	{0x9FD3EF8841F7A512ULL, "288225461.ogg"},
	{0x9FD3F1978A2BFE83ULL, "gen00fl_isabela_init6175649.cl"},
	{0x9FD3F521F71B42EEULL, "vxw_encounter3_bc_0_0.vlm"},
	{0x9FD418F3821756CFULL, "gen_im_ammo_quiver_06.uti"},
	{0x9FD561BBB563F768ULL, "6128751_m.fxe"},
	{0x9FD5645E7899BE54ULL, "gen00fl_isabela6135704.cl"},
	{0x9FD5D9D2603DDB27ULL, "6174717_m.fxe"},
	{0x9FD6075EC2D3BE1DULL, "6126604_m.fxe"},
	{0x9FD651C5B81E357BULL, "w_bonepitdestroyed_a3_100_0.msh"},
	{0x9FD65734A63EA2CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_loop.gad"},
	{0x9FD663C5B81E53C6ULL, "w_bonepitdestroyed_a3_100_0.mao"},
	{0x9FD66FC5B81E6865ULL, "w_bonepitdestroyed_a3_100_0.mmh"},
	{0x9FD6D8AF398BD93AULL, "74441095.ogg"},
	{0x9FD70BA7EA5D46B8ULL, "6166627_m.fxe"},
	{0x9FD753A3344D147AULL, "hlac_lt_0_lowtown_65536_3y.dds"},
	{0x9FD769C10AEB2A23ULL, "zz_and_debug6134098.cl"},
	{0x9FD77AC4066A2A5AULL, "6055567_m.fxe"},
	{0x9FD7CE666546B004ULL, "vxw_bonepitdestroyed_a3_1va_0_0.vlm"},
	{0x9FD7DAA9D9DD8CC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\aveline_inspect_blade.ani"},
	{0x9FD80ADFFAA3B141ULL, "6087581_m.fxe"},
	{0x9FD81B499D65E01CULL, "6126718_f.fxe"},
	{0x9FD81FB0C06601D2ULL, "w_bonepit_104_0.phy"},
	{0x9FD9398276C315E8ULL, "gen_im_trs_tmp_junk2_02.uti"},
	{0x9FD953FDFD9AD50BULL, "gen00fl_sebastian6187299.cl"},
	{0x9FDA19FF390D070AULL, "6095817_m.fxe"},
	{0x9FDA65BC3AA077B9ULL, "lt_gallowstemplar_f_2392.rml"},
	{0x9FDADF9A988E007BULL, "inside_day_rig.msh"},
	{0x9FDB05245D1E2BD7ULL, "388980832.ogg"},
	{0x9FDB0A3B995FE10DULL, "lt_0_lowtown_docks_a3_n_146_0.mmh"},
	{0x9FDB0E3B995FE85EULL, "lt_0_lowtown_docks_a3_n_146_0.mao"},
	{0x9FDB1C3B995FFF93ULL, "lt_0_lowtown_docks_a3_n_146_0.msh"},
	{0x9FDB38F535B0C303ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodplanks.mao"},
	{0x9FDB3CF535B0CA50ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodplanks.mmh"},
	{0x9FDB4AF535B0E26AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodplanks.msh"},
	{0x9FDB9D3C06049616ULL, "893660855.ogg"},
	{0x9FDBA3E3D78F43B0ULL, "6037583_m.fxe"},
	{0x9FDBD0280FB213E9ULL, "hlac_lt_0_lowtown_131070_6x.dds"},
	{0x9FDC11BAF1E6B28CULL, "6164823_m.fxe"},
	{0x9FDC2F3AA74CA210ULL, "vxlt_warehouse_w6_0_0.vlm"},
	{0x9FDC58F1D1B103AEULL, "vxlt_0_lowtown_a2_n_c_eor_0_0.vlm"},
	{0x9FDCDA1C7D3AFC22ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.dura.evt"},
	{0x9FDCEF5A7123B1B6ULL, "vxht_brothel_n_5dz_0_0.vlm"},
	{0x9FDD50AFC217C774ULL, "271577623.ogg"},
	{0x9FDD8EC03B398454ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_carver_flemeth_real.ani"},
	{0x9FDDD26FF4D0D3E6ULL, "6043245_m.fxe"},
	{0x9FDDF4F2362FC367ULL, "855412819.ogg"},
	{0x9FDE2B3895429EAFULL, "6117927_m.fxe"},
	{0x9FDE9378D51034B8ULL, "6089410_m.fxe"},
	{0x9FDF8FB7B5AA225FULL, "hlac_ht_0_hightown_a3_2_0y.dds"},
	{0x9FE08A8B221F1E17ULL, "w_woundedcoast_156_0.mmh"},
	{0x9FE08E8B221F2524ULL, "w_woundedcoast_156_0.mao"},
	{0x9FE0914B72C309F5ULL, "vxdr_primevaltomb_759_0_0.vlm"},
	{0x9FE0A08B221F4389ULL, "w_woundedcoast_156_0.msh"},
	{0x9FE12ABCE66153DAULL, "6037101_m.fxe"},
	{0x9FE17DD19BEF7940ULL, "6061247_m.fxe"},
	{0x9FE1A33BE0C22FFFULL, "772058859.ogg"},
	{0x9FE1C7F5F3740B9BULL, "framebuffereffects_twopassblurcocvertical_p0.psh.fxo.dx11.assembly"},
	{0x9FE334FE080054DAULL, "vxht_0_hightown_n_c36_0_0.vlm"},
	{0x9FE38F8448548D68ULL, "llac_ht_chantry_n_2534_17.dds"},
	{0x9FE39F6C333E539CULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_hed_sw.mao"},
	{0x9FE3D91C7D411935ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.dura.gad"},
	{0x9FE3DB52F4B5D99BULL, "w_encounter3_55_0.mmh"},
	{0x9FE3E752F4B5EE38ULL, "w_encounter3_55_0.mao"},
	{0x9FE3F952F4B60C85ULL, "w_encounter3_55_0.msh"},
	{0x9FE44D8DB123C144ULL, "980551310.ogg"},
	{0x9FE48BC03B3F9DB6ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_carver_flemeth_real.gad"},
	{0x9FE4A178ABA3A30FULL, "vxht_0_hightown_n_avn_0_0.vlm"},
	{0x9FE4C3B7E4FBD75FULL, "hlac_w_encounter3_-196605_0x.dds"},
	{0x9FE5150BA6EC6B28ULL, "dae300_alistair_king6134874.cl"},
	{0x9FE51CD5A61CCAFEULL, "gen00fl_fenris6107163.cl"},
	{0x9FE6251EBF348389ULL, "vxht_0_hightown_5jc_0_0.vlm"},
	{0x9FE654AABE8F5792ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockcxxl.phy"},
	{0x9FE669A4E6E3CB1BULL, "936464263.ogg"},
	{0x9FE6EB9FA8F7A6BBULL, "lgt310pt_fixer_3.ncs"},
	{0x9FE707FDC7C71123ULL, "ser221_party_comments6166207.cl"},
	{0x9FE748FB7FB11498ULL, "gen00fl_anders6112907.cl"},
	{0x9FE80417092B25CEULL, "vxlt_0_lowtown_docks_cw4_0_0.vlm"},
	{0x9FE878D99EAB3BBFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_ornate.msh"},
	{0x9FE87ED99EAB4661ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_ornate.mmh"},
	{0x9FE88928C576F1E9ULL, "gen00fl_anders6215041.cl"},
	{0x9FE88AD99EAB5A02ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_ornate.mao"},
	{0x9FE8B579B0186991ULL, "vxw_woundedcoast_1sy_0_1.vlm"},
	{0x9FE8D6D6CBB9DF12ULL, "plt_cod_cha_connor.plo"},
	{0x9FE8FBD4423A41E8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelred_0.pwk"},
	{0x9FE90184CF01696AULL, "6107545_m.fxe"},
	{0x9FE951A4402035DBULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robe_t.dds"},
	{0x9FE971A7F36ED6BFULL, "6173158_m.fxe"},
	{0x9FE99905AA590ED7ULL, "vxlt_warehouse_n_15m_0_0.vlm"},
	{0x9FE9D0A225992539ULL, "vxlt_0_lowtown_n_e2h_0_0.vlm"},
	{0x9FEAA1179942BA90ULL, "vxlt_warehouse_x7_0_0.vlm"},
	{0x9FEAA677E97682AEULL, "w_encounter2_42_0.msh"},
	{0x9FEAB477E9769AFFULL, "w_encounter2_42_0.mao"},
	{0x9FEAC077E976AE9CULL, "w_encounter2_42_0.mmh"},
	{0x9FEAC7CC4FAF0063ULL, "hlac_ht_0_hightown_65535_2x.dds"},
	{0x9FEAD46B08F11A44ULL, "hlac_ht_0_hightown_n_-3_0x.dds"},
	{0x9FEB084B8B58A412ULL, "vxw_blightlands_bk_0_0.vlm"},
	{0x9FEB0CB84C90B23AULL, "hlac_w_blightlands_65534_0z.dds"},
	{0x9FEB39ABC47F4C15ULL, "qun120cr_javaris6049236.cl"},
	{0x9FEB56B5F0905778ULL, "follower_banter6151372.cl"},
	{0x9FEB798A9AE25DD6ULL, "hlac_ht_chantry_2534_14x.dds"},
	{0x9FEC29861444BA3FULL, "gen00fl_isabela6201224.cl"},
	{0x9FECE7F80E246692ULL, "vxlt_0_lowtown_docks_n_b5a_0_1.vlm"},
	{0x9FED09D62690DAA8ULL, "vxht_chantry_n_13r_0_0.vlm"},
	{0x9FED173BA26FD006ULL, "mcr370_meredith6084409.cl"},
	{0x9FED20E2EABD9E72ULL, "6044961_m.fxe"},
	{0x9FEDA20A5ABBF367ULL, "gen00fl_merrill6128765.cl"},
	{0x9FEE7A7108121AACULL, "945789309.ogg"},
	{0x9FEE953C41C94C17ULL, "nat380_nathaniel6068369.cl"},
	{0x9FEEAAA554AD8BF9ULL, "363768729.ogg"},
	{0x9FEEE8C45B5C81ADULL, "mcr310_argument6127580.cl"},
	{0x9FEF740B3FF0DADCULL, "112783677.ogg"},
	{0x9FF08F7FB5335EDDULL, "vxlt_warehouse_n_13i_0_0.vlm"},
	{0x9FF1179DD72AF2C1ULL, "93937428.ogg"},
	{0x9FF1A6DAF9F66D5AULL, "vxht_0_hightown_n_c87_0_0.vlm"},
	{0x9FF1EAA2648D5764ULL, "ave310_donnic6134743.cl"},
	{0x9FF1EB6361F15127ULL, "and100_anders6095911.cl"},
	{0x9FF230A79EB7FAF3ULL, "498910789.ogg"},
	{0x9FF270096C7D906AULL, "vxlt_gallowscourtyard_74a_0_0.vlm"},
	{0x9FF29DD6AC992103ULL, "vxht_keep_6h5_0_1.vlm"},
	{0x9FF2F891DBB571CCULL, "lt_0_lowtown_docks_255_0.msh"},
	{0x9FF2FE91DBB57BBEULL, "lt_0_lowtown_docks_255_0.mmh"},
	{0x9FF30A91DBB59055ULL, "lt_0_lowtown_docks_255_0.mao"},
	{0x9FF30EEE28321914ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobc_0d.dds"},
	{0x9FF382C782B35C5DULL, "hlac_w_sundermount_-131068_0x.dds"},
	{0x9FF392F547393F69ULL, "vxca_caverns1_ijg_0_0.vlm"},
	{0x9FF3A6FBC2290900ULL, "226241069.ogg"},
	{0x9FF420BDC1DC76A4ULL, "6063861_m.fxe"},
	{0x9FF4FF038152111CULL, "6128518_f.fxe"},
	{0x9FF58AE71FDF737CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p_wkf.ani"},
	{0x9FF5FD879EDE7120ULL, "vxlt_hangedman_21o_0_1.vlm"},
	{0x9FF71A7D55F451DFULL, "vxlt_gallowstemplar_4st_0_0.vlm"},
	{0x9FF817FF060EC089ULL, "fen110_fenris6117979.cl"},
	{0x9FF873F416157164ULL, "vxht_bartrand_n_68w_0_0.vlm"},
	{0x9FF8CFD830382296ULL, "gen00fl_merrill6107075.cl"},
	{0x9FF9B548E2D02E48ULL, "39672759.ogg"},
	{0x9FF9BC3BEA5E0E73ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.dura.ani"},
	{0x9FF9E3D1A5DE02A1ULL, "hlac_ht_chantry_2534_24x.dds"},
	{0x9FFA17F8ACC75905ULL, "195733971.ogg"},
	{0x9FFA450217352531ULL, "6162738_m.fxe"},
	{0x9FFA6430C14FB143ULL, "108948669.ogg"},
	{0x9FFA8A4391C55AA9ULL, "vxht_mansion2_2dz_0_0.vlm"},
	{0x9FFA95180CF16514ULL, "vxw_sundermount_32x_0_0.vlm"},
	{0x9FFAA228A5E17DFBULL, "clf260_harley6069652.cl"},
	{0x9FFABB869DBAC5CBULL, "zz_approval_debug6003725.cl"},
	{0x9FFADC3942C6068DULL, "vrc321tr_door_shift_3.utt"},
	{0x9FFAFFD302D178DEULL, "6034377_m.fxe"},
	{0x9FFB4261C1BC3BFCULL, "47328802.ogg"},
	{0x9FFC1B56DC573301ULL, "vxlt_gallowstemplar_f_3qx_0_0.vlm"},
	{0x9FFC6F7E217B8E50ULL, "vxlt_0_lowtown_a3_n_c_gpb_0_0.vlm"},
	{0x9FFC97E71FE5A7BEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p_wkf.gad"},
	{0x9FFDBA80B9532A5CULL, "649047965.ogg"},
	{0x9FFDC636F5470BE7ULL, "6209904_f.fxe"},
	{0x9FFDD8DBD8C0FD33ULL, "6069730_m.fxe"},
	{0x9FFDE6AA17F176E8ULL, "gen00fl_sebastian6174876.cl"},
	{0x9FFF6FCECB96DE48ULL, "604724742.ogg"},
	{0xA000597725932584ULL, "sib100_father_contact6108954.cl"},
	{0xA000A63040059AE0ULL, "vxht_bartrand_n_4c0_0_0.vlm"},
	{0xA000C41D62B3A797ULL, "hlac_dr_dungeon_3214_9z.dds"},
	{0xA000E59484255B9BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rb_basic.mmh"},
	{0xA000F19484257038ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rb_basic.mao"},
	{0xA000F3EFF4BEF221ULL, "325920714.ogg"},
	{0xA001039484258E85ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rb_basic.msh"},
	{0xA00122EBBBFBAECAULL, "dre000_bodahn6028138.cl"},
	{0xA001A448B5E6435DULL, "llac_lt_0_lowtown_a2_n_c_65537_3.dds"},
	{0xA001D3EE3003CF09ULL, "6182761_m.fxe"},
	{0xA001EBC888F2ADF6ULL, "lt_0_lowtown_a2_n_c_2_0.msh"},
	{0xA001F5C888F2BEF4ULL, "lt_0_lowtown_a2_n_c_2_0.mmh"},
	{0xA001F9C888F2C507ULL, "lt_0_lowtown_a2_n_c_2_0.mao"},
	{0xA00204D2392F57E9ULL, "vxca_caverns1_imf_0_0.vlm"},
	{0xA0025D8615DEFC07ULL, "vxlt_0_lowtown_a2_he1_0_0.vlm"},
	{0xA00265F19E959E8AULL, "zz_bet_debug6113391.cl"},
	{0xA002B9C9A6803F9EULL, "6218939_m.fxe"},
	{0xA002FE64B9813A87ULL, "401587255.ogg"},
	{0xA00301AE2915FA39ULL, "w_woundedcoast_232_0.phy"},
	{0xA00323FD3B28F984ULL, "vxw_bonepitdestroyed_a3_1ad_0_0.vlm"},
	{0xA003344E31A94994ULL, "62551155.ogg"},
	{0xA00371895C297ECCULL, "zz_facefx_test6088483.cl"},
	{0xA003BB3074F7AB3EULL, "438469147.ogg"},
	{0xA00411D242C198BCULL, "425590517.ogg"},
	{0xA0042D316FBB6E69ULL, "hlac_lt_0_lowtown_a2_n_c_65534_2z.dds"},
	{0xA004CF47BD9D058BULL, "llac_ht_keep_n_3732_4.dds"},
	{0xA0050C959CCF9370ULL, "6111316_f.fxe"},
	{0xA0052654F86C7C18ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archupper.phy"},
	{0xA0056BCCBAE118E2ULL, "203077647.ogg"},
	{0xA0061015A84D8F76ULL, "distortionvfx.vsh.fxo"},
	{0xA006344A4343A01CULL, "lt_0_lowtown.anb"},
	{0xA006404A4343B48EULL, "lt_0_lowtown.arl"},
	{0xA0064C57EC3489DEULL, "6118813_m.fxe"},
	{0xA00667C87D111443ULL, "6202098_m.fxe"},
	{0xA0067917B47A219AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\exit.gad"},
	{0xA006B43BEA688AAAULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.dura.evt"},
	{0xA006E515C7266E5AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_basic.phy"},
	{0xA00774C4D341EE3DULL, "vxht_0_hightown_a2_bxn_0_0.vlm"},
	{0xA007949E0B5CB774ULL, "llac_lt_0_lowtown_a2_n_c_-196607_0.dds"},
	{0xA008080FC1F00976ULL, "gen00fl_carver6108894.cl"},
	{0xA0086D94A998AF6EULL, "540201653.ogg"},
	{0xA009C077E9915FEFULL, "w_encounter2_42_0.rml"},
	{0xA00A80AE2312AD17ULL, "6168738_m.fxe"},
	{0xA00B3347C11A7BC1ULL, "dae000_tomwise6167301.cl"},
	{0xA00B848DDECBBA67ULL, "vxlt_undercitydungeon_11t_0_0.vlm"},
	{0xA00BBA7E131E4507ULL, "19768163.ogg"},
	{0xA00BCEB6060E5D62ULL, "gen00fl_varric6099163.cl"},
	{0xA00BD0BAE83A1DDAULL, "6149889_f.fxe"},
	{0xA00BEDB434C49E16ULL, "art\\characters\\playercharacter\\humanfemale\\boo_meda.phy"},
	{0xA00CA7881DE3FE46ULL, "hlac_w_bonepitmining_a2_327680_0z.dds"},
	{0xA00CBCFB8F7E2D9AULL, "639583446.ogg"},
	{0xA00CE246AF9F3C1FULL, "140641249.ogg"},
	{0xA00CF4D22870E540ULL, "vxht_0_hightown_a2_c59_0_0.vlm"},
	{0xA00D7C17B48045B1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\exit.evt"},
	{0xA00D933BEA6E711DULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.dura.gad"},
	{0xA00DBD9B9834A6F0ULL, "6026237_m.fxe"},
	{0xA00DE80B8360D76DULL, "hlac_ht_0_hightown_131073_0y.dds"},
	{0xA00DF105212188BEULL, "hlac_ht_mansion2b_818_37x.dds"},
	{0xA00EA2BA4DB31F28ULL, "844928276.ogg"},
	{0xA00ED9D415506E3EULL, "gen_im_npc_arm_hum_glv_mrca.uti"},
	{0xA00EDDAF905D2F2BULL, "6155877_m.fxe"},
	{0xA00F3B2E09C896C6ULL, "lt_0_lowtown_docks_32_0_ll.mmh"},
	{0xA00F825382257701ULL, "6001896_f.fxe"},
	{0xA00FDB15796210C9ULL, "vxca_cavern1_haunted_fo6_0_0.vlm"},
	{0xA0103E06FBDD0655ULL, "vxw_encounter2_41_0_0.vlm"},
	{0xA010669977BDC977ULL, "dae000_pc_home_ambients6208625.cl"},
	{0xA0109F77E9974667ULL, "w_encounter2_42_0.phy"},
	{0xA010C5A0A1465674ULL, "vxdr_primevaltomb_7yh_0_0.vlm"},
	{0xA01100B0BFD1254BULL, "mcr371_emile_de_launcet6087016.cl"},
	{0xA0113B4858271133ULL, "zz_approval_debug6003570.cl"},
	{0xA0117E575010961FULL, "gen00fl_fenris6126561.cl"},
	{0xA012A734EDA9822BULL, "gen00fl_aveline6121907.cl"},
	{0xA013FCB2F447F32CULL, "isa311_castillon6092126.cl"},
	{0xA01424D548F699C1ULL, "hlac_dr_slavers_3217_10y.dds"},
	{0xA0147FD99ED082DEULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_ornate.phy"},
	{0xA0148F044D59D73EULL, "vxlt_hangedman_n_2fn_0_0.vlm"},
	{0xA014FD842C4D46DAULL, "174709848.ogg"},
	{0xA0156BD4C91F4540ULL, "dae000cr_grand_cleric6147681.cl"},
	{0xA015D6B434CCCBB7ULL, "art\\characters\\playercharacter\\humanfemale\\boo_meda.msh"},
	{0xA015E761E7E8F2B9ULL, "lt_0_lowtown_a3_n_c_34_0.anb"},
	{0xA015E8B434CCEA1AULL, "art\\characters\\playercharacter\\humanfemale\\boo_meda.mao"},
	{0xA015ECB434CCF129ULL, "art\\characters\\playercharacter\\humanfemale\\boo_meda.mmh"},
	{0xA0160CDD9F3DEA69ULL, "277671094.ogg"},
	{0xA0162F32C4E2388BULL, "42103232.ogg"},
	{0xA016DF118A426185ULL, "6140077_m.fxe"},
	{0xA016E25AAA85F877ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkfront.gad"},
	{0xA0176DFD75B18BA9ULL, "vxht_mansion2_2jx_0_0.vlm"},
	{0xA01772B9031D6C48ULL, "llac_ht_keep_3734_0.dds"},
	{0xA017AB01ED89F85BULL, "6179053_m.fxe"},
	{0xA01898B201BC2419ULL, "dae000_gamlen6147213.cl"},
	{0xA018CF1506799193ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.cess.evt"},
	{0xA0192DD4E8A0E73EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_clench_fist.evt"},
	{0xA019422EEED41624ULL, "dae000_mother6035932.cl"},
	{0xA01976A4D56CDBDAULL, "6085510_f.fxe"},
	{0xA019C53942E09000ULL, "vrc321tr_door_shift_3.ncs"},
	{0xA019CB4F04A89AE9ULL, "vxlt_warehouse_q5_0_0.vlm"},
	{0xA019EFECACB60FFCULL, "llac_lt_gallowstemplar_n_2394_5.dds"},
	{0xA01A1BD9F36ACED1ULL, "hlac_lt_0_lowtown_a2_n_0_5z.dds"},
	{0xA01ADCEC945F4E12ULL, "vxlt_0_lowtown_a2_n_c_d46_0_0.vlm"},
	{0xA01AE9FFDF82CD65ULL, "zz_dae_debug6029308.cl"},
	{0xA01B0EF35C42EB1EULL, "zz_fen_debug6166700.cl"},
	{0xA01B43F535E6BA6BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_woodplanks.phy"},
	{0xA01B54C13D52E1B1ULL, "647750857.ogg"},
	{0xA01B9417B48CAB80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\exit.ani"},
	{0xA01BBC852D0FA4D0ULL, "682527668.ogg"},
	{0xA01C352ACC006946ULL, "vxlt_undercityserialkiller_172_0_0.vlm"},
	{0xA01C6F08B4BA9A6CULL, "773672990.ogg"},
	{0xA01C7C10F5D06450ULL, "6071116_m.fxe"},
	{0xA01DC28B165FE678ULL, "vxlt_0_lowtown_docks_a3_9r8_0_0.vlm"},
	{0xA01DC481CD8103CAULL, "hlac_dr_0_deeproads_5845_1y.dds"},
	{0xA01DF75AAA8C3A04ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkfront.evt"},
	{0xA01E808FF4FD06D6ULL, "6018609_m.fxe"},
	{0xA01EC815067DF1D4ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.cess.gad"},
	{0xA01F377D933C6EF4ULL, "vxdr_primevaltomb_7ri_0_0.vlm"},
	{0xA01F9560A40B13F0ULL, "511101674.ogg"},
	{0xA0201B3E068934A1ULL, "784406664.ogg"},
	{0xA0201FA237D8FF51ULL, "dae000_gen_turn_in_46178436.cl"},
	{0xA0203A3A375AE19EULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.cess.ani"},
	{0xA02044D4E8A72D69ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_clench_fist.gad"},
	{0xA0204541BD1D2D00ULL, "qun110_petrice6018664.cl"},
	{0xA02080CDE7070552ULL, "6170028_m.fxe"},
	{0xA020B7F85649368DULL, "vxlt_undercitydungeon_1nw_0_0.vlm"},
	{0xA0210502E9C5F2ACULL, "vxlt_0_lowtown_docks_a3_n_bua_0_0.vlm"},
	{0xA0212987BE4FF55BULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar6_.ani"},
	{0xA0218BC8A52D1575ULL, "722964397.ogg"},
	{0xA02200E0FAD70D4AULL, "vxlt_gallowstemplar_5cl_0_0.vlm"},
	{0xA022067F2FD4C955ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_eye_hlk_0d.dds"},
	{0xA0220FDFAE46E01DULL, "vxw_sundermount_2lz_0_0.vlm"},
	{0xA0221914DF9F611FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge02_sa.dds"},
	{0xA0222E25CBD18E79ULL, "vxht_0_hightown_n_at6_0_0.vlm"},
	{0xA022F15CF7A9AED5ULL, "546218531.ogg"},
	{0xA022F26B6060FDA5ULL, "ss_qunari_m_3.wbd"},
	{0xA02300E13F4FEFBEULL, "vxlt_hangedman_n_2mo_0_0.vlm"},
	{0xA0232768101CA647ULL, "951488.ogg"},
	{0xA023BD70E4A626FDULL, "one100dg_lia6061027.cl"},
	{0xA024587EB72E1F3DULL, "vxht_0_hightown_a2_bfl_0_0.vlm"},
	{0xA0246522850338DAULL, "vxw_woundedcoast_at_0_0.vlm"},
	{0xA02564E486F69F08ULL, "vxw_encounter2_camp_dq_0_1.vlm"},
	{0xA025D715068429B2ULL, "art\\vfx\\cinematic\\vfx_dragon_bd_glow_c.cess.ani"},
	{0xA025F1FF6572C612ULL, "vxw_sundermount_2ar_0_0.vlm"},
	{0xA025F3C15888CDF7ULL, "6139156_m.fxe"},
	{0xA0263987DB0E5E4FULL, "lt_0_lowtown_docks_38_0_lc.mmh"},
	{0xA02665277821950BULL, "44571024.ogg"},
	{0xA026ADEE5A88D213ULL, "6173679_m.fxe"},
	{0xA026CC4E3BE4ED4BULL, "299137565.ogg"},
	{0xA026E27D7FF8B1EFULL, "vxdr_primevaltomb_2ur_0_0.vlm"},
	{0xA02715D4E8ACFBB7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\rightarm_clench_fist.ani"},
	{0xA0273BD85C79AE0EULL, "221401053.ogg"},
	{0xA027E5BE294A24E0ULL, "ht_0_hightown_36_0_lc.mmh"},
	{0xA0283D2BF69EB369ULL, "vxlt_warehouse_z4_0_0.vlm"},
	{0xA029D12F338C306CULL, "hlac_lt_backalley_n_0_5x.dds"},
	{0xA029E200C864C0F6ULL, "llac_ht_brothel_n_2995_6.dds"},
	{0xA029EAA81D93DE6DULL, "zz_followers_banter6158796.cl"},
	{0xA02A3959FA96ECAFULL, "vxdr_primevalentrance_1q1_0_0.vlm"},
	{0xA02A85C5CA99F281ULL, "6139244_m.fxe"},
	{0xA02AA81164F6A488ULL, "vxht_mansion2_3co_0_0.vlm"},
	{0xA02AB9419BFD7BEBULL, "vxdr_primevaltomb_77m_0_0.vlm"},
	{0xA02B020C3CB55AB3ULL, "vxht_0_hightown_a2_b8p_0_0.vlm"},
	{0xA02B0F5AAA96EDFDULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkfront.ani"},
	{0xA02B114B3673D81CULL, "hlac_w_bonepitmining_a2_131074_0z.dds"},
	{0xA02B2061A683CD31ULL, "hlac_ht_chantry_n_2534_18z.dds"},
	{0xA02B99705B4D6530ULL, "6144429_m.fxe"},
	{0xA02BE6257B61D452ULL, "gen00fl_aveline6137688.cl"},
	{0xA02BF59A67C752AEULL, "vxlt_gamlenshouse_n_3o_0_0.vlm"},
	{0xA02C349953DCE593ULL, "vxlt_gallowstemplar_3r4_0_0.vlm"},
	{0xA02C6FDEEE72C988ULL, "vxlt_gallowscourtyard_n_7ec_0_0.vlm"},
	{0xA02CE095E1A22FC1ULL, "vxlt_undercityserialkiller_1t0_0_0.vlm"},
	{0xA02D5F829EC9C3A4ULL, "large_n.dds"},
	{0xA02D5FFBE1F4FC8AULL, "vxht_keep_6fc_0_0.vlm"},
	{0xA02DC838AD4B59ABULL, "mag291_demon_marethari.bnk"},
	{0xA02E1C09E7CF811FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_loop.gad"},
	{0xA02E477F61CDA70AULL, "6174428_m.fxe"},
	{0xA02E523A3767486FULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.cess.evt"},
	{0xA02EBF9E81E07B83ULL, "6113848_m.fxe"},
	{0xA02F0856417A5E61ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v2.ani"},
	{0xA02F0BC33F2DEDBBULL, "vxht_mansion2b_3bo_0_0.vlm"},
	{0xA02F3B4F9AF6F30CULL, "art\\vfx\\models\\vfx_twist_two_tail.msh"},
	{0xA02F414F9AF6FDFEULL, "art\\vfx\\models\\vfx_twist_two_tail.mmh"},
	{0xA02F43FD8A21104BULL, "hlac_lt_0_lowtown_docks_a3_n_131071_2z.dds"},
	{0xA02F55746332FEA0ULL, "vxht_0_hightown_n_ca6_0_0.vlm"},
	{0xA02F849D7644253DULL, "vxht_0_hightown_bu9_0_0.vlm"},
	{0xA02F866EEC9C5783ULL, "6099157_m.fxe"},
	{0xA02FB7942BF649E8ULL, "vxht_mansion_n_3zc_0_0.vlm"},
	{0xA02FED06316D8F7BULL, "778767029.ogg"},
	{0xA03081BCA03CF89DULL, "vxw_sundermount_2qy_0_0.vlm"},
	{0xA030CA38AD4D928BULL, "mag291_demon_marethari.cnv"},
	{0xA031DB52F4F7FB28ULL, "w_encounter3_55_0.rml"},
	{0xA031E402C5501186ULL, "760665336.ogg"},
	{0xA0323141741B6E1DULL, "6116496_m.fxe"},
	{0xA03240255B47969EULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v4.evt"},
	{0xA0333D8673F25496ULL, "vxht_0_hightown_a3_cbw_0_0.vlm"},
	{0xA0336109FEB5E11BULL, "vxlt_0_lowtown_n_gos_0_0.vlm"},
	{0xA0337F20A53F460AULL, "vxlt_gallowstemplar_n_45t_0_0.vlm"},
	{0xA033879AC4C225D8ULL, "584661738.ogg"},
	{0xA033CFF0D2A4AB66ULL, "hlac_lt_hangedman_n_-1_4x.dds"},
	{0xA033FAEBD58ABF19ULL, "mer210_coterie_contact6035274.cl"},
	{0xA034533A376BB518ULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.cess.gad"},
	{0xA03463DC5768DE92ULL, "vxw_sundermount_2js_0_0.vlm"},
	{0xA035AC98A9D7B482ULL, "isa000im_gft_ship_in_bottle.uti"},
	{0xA035C48B15905259ULL, "vxlt_0_lowtown_eom_0_1.vlm"},
	{0xA0360E8EB4F58867ULL, "gen00pt_achievement.ncs"},
	{0xA0361038DAFBDC06ULL, "479195674.ogg"},
	{0xA036135641808F03ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v2.gad"},
	{0xA0365087BE6293E5ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar6_.gad"},
	{0xA036ABDD32EDA4C0ULL, "gen_im_wep_war_shd_act3_03.uti"},
	{0xA036B70458DA9239ULL, "one220_yevhen6102954.cl"},
	{0xA036BA8759B38D34ULL, "follower_banter6164062.cl"},
	{0xA0370D8D944742EFULL, "hlac_lt_undercityserialkiller_-1_46y.dds"},
	{0xA0370E0B5CBE8934ULL, "mcr370_negotiation6078742.cl"},
	{0xA03745E5E2C199B0ULL, "467545467.ogg"},
	{0xA0375D75AFB80255ULL, "hlac_w_woundedcoast_-262139_0x.dds"},
	{0xA0376A37A2296F9EULL, "character_character_p85.psh.fxo"},
	{0xA03792E3B845C2CEULL, "ser211ar_foundry.ncs"},
	{0xA037D933C67261E4ULL, "6122137_m.fxe"},
	{0xA0381D327FDE07BFULL, "24004340.ogg"},
	{0xA038331F55C16CBDULL, "vxlt_gallowstemplar_5j4_0_0.vlm"},
	{0xA038517D211F3643ULL, "llac_lt_warehouse_n_0_1.dds"},
	{0xA03857255B4C29C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v4.gad"},
	{0xA038979A3D0AB44DULL, "6132836_m.fxe"},
	{0xA038B32EEA797BA0ULL, "hlac_lt_0_lowtown_a2_n_c_131070_2z.dds"},
	{0xA038FD4CD7A2C9A8ULL, "190365639.ogg"},
	{0xA03919EE56ECBD08ULL, "vxht_mansion2_3fl_0_0.vlm"},
	{0xA0395C2653B983D6ULL, "qun221_petrice6098766.cl"},
	{0xA0396A3575779E31ULL, "vxw_encounter2_camp_od_0_0.vlm"},
	{0xA03994D69BB04E02ULL, "mcr310_argument6074407.cl"},
	{0xA039D7FE1D7D53C3ULL, "vxdr_dungeon_1m2_0_1.vlm"},
	{0xA03A0B4D4D437DB0ULL, "6141729_m.fxe"},
	{0xA03A3A5527B26D97ULL, "hlac_ca_cavern1_haunted_24423_0z.dds"},
	{0xA03A722F684FBBD6ULL, "6044657_m.fxe"},
	{0xA03A80CB2A8E5A4FULL, "zz_isa_debug6175877.cl"},
	{0xA03B04085415FD69ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_nud_0d.dds"},
	{0xA03B07DD93B59B00ULL, "vxlt_gallowstemplar_n_3ss_0_0.vlm"},
	{0xA03B73C34D54A48FULL, "42302422.ogg"},
	{0xA03B7D96FE5E11C3ULL, "vxlt_gallowscourtyard_n_7f1_0_0.vlm"},
	{0xA03B87CDC2C23A4FULL, "vxht_0_hightown_a2_avq_0_0.vlm"},
	{0xA03C7DF159EE1C21ULL, "618637479.ogg"},
	{0xA03CF056418672E8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_attk_v2.evt"},
	{0xA03D36786EEEEEE7ULL, "mag131_keran6094509.cl"},
	{0xA03DB5B95CF9AEEAULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_dwn_c.anb"},
	{0xA03E4FE9EC05025CULL, "vxht_0_hightown_a79_0_0.vlm"},
	{0xA03E8AE3B84BD3A7ULL, "ser211ar_foundry.lst"},
	{0xA03EF3999233111DULL, "vxw_sundermount_2zx_0_0.vlm"},
	{0xA03F28255B51F817ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v4.ani"},
	{0xA03F33BAEE6CF9FEULL, "llac_ht_0_hightown_n_-196607_0.dds"},
	{0xA04097E0DAE1EB9EULL, "vxht_keep_n_493_0_0.vlm"},
	{0xA040B20563322373ULL, "vxca_caverns2prc_adu_0_0.vlm"},
	{0xA041134610CD540AULL, "vxht_0_hightown_n_bpv_0_0.vlm"},
	{0xA0411C373FD6838FULL, "gen00fl_anders6113097.cl"},
	{0xA041C7B4832F6840ULL, "vxw_bonepit_1po_0_0.vlm"},
	{0xA041DF9463CCD952ULL, "hlac_w_encounter2_393214_0x.dds"},
	{0xA04222AD0560BE6BULL, "zz_one000_crafter6179357.cl"},
	{0xA042D6CD093BFCA6ULL, "744980815.ogg"},
	{0xA042F909E7E1A265ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_action_loop.ani"},
	{0xA043368B18DFCFFCULL, "llac_w_blightlands_-65537_0.dds"},
	{0xA043FC8BF1071E72ULL, "6016166_m.fxe"},
	{0xA04438A606B69A20ULL, "vxlt_gallowstemplar_43w_0_0.vlm"},
	{0xA044EBD03B24C6D6ULL, "follower_banter6154113.cl"},
	{0xA04554D984BE7FB0ULL, "hlac_lt_0_lowtown_n_1_4y.dds"},
	{0xA045848840B1256CULL, "vxht_chantry_31u_0_0.vlm"},
	{0xA045D12C994740F2ULL, "6105622_m.fxe"},
	{0xA045D89B46B0E285ULL, "808191307.ogg"},
	{0xA045DC14941F881EULL, "character_character_p74.psh.fxo"},
	{0xA045F1B08243AC5FULL, "follower_banter6161275.cl"},
	{0xA045F9391D7AD4CDULL, "plt_mrl150pt_longway.nss"},
	{0xA046329A67B046D4ULL, "vxca_caverns2_7mr_0_0.vlm"},
	{0xA046C0536DBDD08BULL, "mag130_danzig6018803.cl"},
	{0xA0473244C6021A51ULL, "425109064.ogg"},
	{0xA047334A437B35ACULL, "lt_0_lowtown.rml"},
	{0xA0478127F94606EFULL, "vxht_playershouse_n_1qr_0_0.vlm"},
	{0xA047953B2DE1F12FULL, "hlac_ht_0_hightown_1_1z.dds"},
	{0xA0479852C58A73D0ULL, "692494990.ogg"},
	{0xA047BD1888F3CAD1ULL, "hlac_lt_foundry_n_320_6z.dds"},
	{0xA0480694846196D0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\7rb_basic.phy"},
	{0xA048313D02231E5EULL, "6067752_f.fxe"},
	{0xA0485A132ED8F785ULL, "mer120_hubert6017872.cl"},
	{0xA048CA4280104543ULL, "w_encounter2_0_0.mao"},
	{0xA048CE4280104B90ULL, "w_encounter2_0_0.mmh"},
	{0xA048DC42801063AAULL, "w_encounter2_0_0.msh"},
	{0xA049C25AD2755A03ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01a.mmh"},
	{0xA049D05AD275725DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01a.msh"},
	{0xA049EAAC010BC232ULL, "hlac_ht_brothel_n_3381_0z.dds"},
	{0xA049EF73F0542A43ULL, "vxlt_gallowscourtyard_n_7i0_0_0.vlm"},
	{0xA049F7CB27911718ULL, "vxw_woundedcoast_17a_0_0.vlm"},
	{0xA049F9AAB4B852CFULL, "vxht_0_hightown_a2_aqp_0_0.vlm"},
	{0xA04A3609BDAD21C6ULL, "610338218.ogg"},
	{0xA04A856214B84C29ULL, "6089099_m.fxe"},
	{0xA04AC6DE0014D9A1ULL, "qun221_viscount6152866.cl"},
	{0xA04AE5F479A61008ULL, "210838510.ogg"},
	{0xA04B5E4857196256ULL, "vxlt_gallowstemplar_n_2bg_0_0.vlm"},
	{0xA04BA4B9BAD0D2E6ULL, "art\\vfx\\cinematic\\vfx_meredith_smokey.impa.ani"},
	{0xA04BDEC5D9F6DF7DULL, "vxlt_darktown_4lk_0_0.vlm"},
	{0xA04BF730312E2450ULL, "mag101_aombanter6052756.cl"},
	{0xA04C12157126B5DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r180_rf.gad"},
	{0xA04C36C0BA82E69CULL, "6217345_m.fxe"},
	{0xA04CDC5DAE7ACD81ULL, "hlac_lt_0_lowtown_a2_n_-65536_7x.dds"},
	{0xA04D17AABD110F2BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hed_frsa_0d.dds"},
	{0xA04D5677D0B956CAULL, "vxlt_gallowstemplar_5ro_0_0.vlm"},
	{0xA04D9368E97E40F0ULL, "llac_w_encounter1_-65537_0.dds"},
	{0xA04E428C715C1D14ULL, "dre180ip_deep_roads_chest_3.utp"},
	{0xA04E85DF4B4A2BA0ULL, "6082162_f.fxe"},
	{0xA04EB2D23850F403ULL, "640437964.ogg"},
	{0xA04ED5989171AF83ULL, "gen00fl_fenris6103331.cl"},
	{0xA04EDED2C101B98EULL, "one000_ilen6205639.cl"},
	{0xA04F54441D57DA30ULL, "170724616.ogg"},
	{0xA04F8BE3B85A48DCULL, "ser211ar_foundry.are"},
	{0xA04FBAB95AE5825AULL, "vxw_woundedcoast_pw_0_0.vlm"},
	{0xA04FE861E81A6CB9ULL, "lt_0_lowtown_a3_n_c_34_0.rml"},
	{0xA0503991752580C0ULL, "vxw_bonepit_1ul_0_0.vlm"},
	{0xA05045CF7BEE4D15ULL, "ser211ip_note_circle.utp"},
	{0xA050545D81C59C08ULL, "87803736.ogg"},
	{0xA0505637084D8734ULL, "llac_ht_playershouse_n_855_27.dds"},
	{0xA0509B426C3F4938ULL, "mcr370_templar16086971.cl"},
	{0xA050B0019AD08688ULL, "hlac_w_encounter3_-327677_0x.dds"},
	{0xA05147EDA728CE9EULL, "vxw_woundedcoast_i8_0_0.vlm"},
	{0xA0514BFE08F72CFCULL, "vxht_mansion2_2el_0_0.vlm"},
	{0xA051C0BD03236785ULL, "17976632.ogg"},
	{0xA051E21C90B1604AULL, "zz_ave_debug6134014.cl"},
	{0xA052143F85F43FFAULL, "gen_im_trs_rdr_junk2_01.uti"},
	{0xA0523207B928A192ULL, "541509723.ogg"},
	{0xA0528063057DB190ULL, "366202199.ogg"},
	{0xA052AA82F8ACB2A0ULL, "vxlt_gallowstemplar_46t_0_0.vlm"},
	{0xA052DBBBBFDC3FDFULL, "703231544.ogg"},
	{0xA05303B3D3038143ULL, "6092134_m.fxe"},
	{0xA0531515712CD9F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r180_rf.evt"},
	{0xA0533EA8969C0598ULL, "6205889_m.fxe"},
	{0xA053C7D27DB56B68ULL, "gen00fl_carver6102670.cl"},
	{0xA0544C8A98B16627ULL, "sky_sky2_p0.vsh.fxo.dx11"},
	{0xA0544DF18615A09EULL, "character_character_p27.psh.fxo"},
	{0xA0549209136A1D6DULL, "6061807_f.fxe"},
	{0xA05534B06444BAD4ULL, "llac_ca_caverns1_mines_24425_1.dds"},
	{0xA05552B1887B77CEULL, "770397702.ogg"},
	{0xA055F70B5C767610ULL, "6064033_m.fxe"},
	{0xA0560836AA3D7BEDULL, "hm_harbormaster_liam.mor"},
	{0xA0562CF476B6D9F3ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobb.phy"},
	{0xA056419200A3C1B3ULL, "6041264_m.fxe"},
	{0xA0564DEF5963CF31ULL, "vxw_encounter2_camp_ez_0_0.vlm"},
	{0xA056918839669EAAULL, "mrl150_merrill6071917.cl"},
	{0xA0569EC46D68EF22ULL, "lt_0_lowtown_a2_n_14_0.rml"},
	{0xA056B20BA461FBCEULL, "art\\toolset\\_materials\\tools_light.mao"},
	{0xA056D3CB64192D8CULL, "isa111_isabela6067122.cl"},
	{0xA056F5C8893AFEF7ULL, "lt_0_lowtown_a2_n_c_2_0.rml"},
	{0xA05799ED5B888C5EULL, "889649792.ogg"},
	{0xA057AAC40F8006A5ULL, "665729707.ogg"},
	{0xA057E04DDC43F003ULL, "vxlt_shop_d3_0_0.vlm"},
	{0xA05844E31E5DDAFAULL, "6103332_m.fxe"},
	{0xA058CAB3DFEFB555ULL, "vxw_blightlands_w7_0_0.vlm"},
	{0xA058CCDC55FC8B19ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_loop.gad"},
	{0xA058D25241B1CA9DULL, "gen00fl_fenris6108026.cl"},
	{0xA058D3945F120257ULL, "6106618_f.fxe"},
	{0xA058EFF0EC29E905ULL, "6040926_m.fxe"},
	{0xA05982A64FA4BFCDULL, "mcr310_argument6075139.cl"},
	{0xA0598CB9BADCE757ULL, "art\\vfx\\cinematic\\vfx_meredith_smokey.impa.evt"},
	{0xA05A740C2C1D9FF5ULL, "625635125.ogg"},
	{0xA05AC35AD283CF28ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear01a.phy"},
	{0xA05B6FCABB166671ULL, "w_bonepit_241_0.msh"},
	{0xA05B7DCABB167D8CULL, "w_bonepit_241_0.mao"},
	{0xA05B89CABB16922FULL, "w_bonepit_241_0.mmh"},
	{0xA05BCD7B1F8407E0ULL, "6105668_m.fxe"},
	{0xA05BD4D26ABD2756ULL, "vxw_bonepitmining_a2_88_0_0.vlm"},
	{0xA05CF4C8893F68EFULL, "lt_0_lowtown_a2_n_c_2_0.phy"},
	{0xA05D08F818D95917ULL, "231585997.ogg"},
	{0xA05D9AE86BFE9DF0ULL, "ts_demonic_arm_table.gda"},
	{0xA05DF6FFF4B9850AULL, "vxht_0_hightown_n_bfp_0_0.vlm"},
	{0xA05E06C23D0AD4EEULL, "6098776_f.fxe"},
	{0xA05EED26D3D6350DULL, "w_bonepit_166_0.mmh"},
	{0xA05EF126D3D63C5EULL, "w_bonepit_166_0.mao"},
	{0xA05EFF26D3D65393ULL, "w_bonepit_166_0.msh"},
	{0xA05F14C5458C3AEAULL, "6043326_m.fxe"},
	{0xA05F9AFC34BAFD94ULL, "follower_banter6141665.cl"},
	{0xA05FBDDAFAED457CULL, "vxht_mansion2_2nm_0_0.vlm"},
	{0xA05FDDDC5602C767ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_loop.ani"},
	{0xA0601CD7A92635A4ULL, "vxht_0_hightown_a3_cb5_0_0.vlm"},
	{0xA06024C556EEB35AULL, "vrc321_party_barks6095732.cl"},
	{0xA0602D1571378CC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r180_rf.ani"},
	{0xA0608DB9BAE307E0ULL, "art\\vfx\\cinematic\\vfx_meredith_smokey.impa.gad"},
	{0xA060B156AAC82830ULL, "hlac_ht_0_hightown_n_65536_3x.dds"},
	{0xA060B9F9937F0381ULL, "vxht_0_hightown_a7h_0_0.vlm"},
	{0xA060D2DDC1F3847FULL, "mcr370_meredith6085908.cl"},
	{0xA0611354F8BABAADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archupper.msh"},
	{0xA0612154F8BAD2E0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archupper.mao"},
	{0xA0612554F8BAD933ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archupper.mmh"},
	{0xA061ED62193FB678ULL, "one110_meeran6089897.cl"},
	{0xA06219B5B9007116ULL, "w_encounter1_3_0.msh"},
	{0xA06223B5B9008214ULL, "w_encounter1_3_0.mmh"},
	{0xA062261B5D95FDE4ULL, "w_bonepit_62_0.msh"},
	{0xA06227B5B9008927ULL, "w_encounter1_3_0.mao"},
	{0xA062381B5D961C4DULL, "w_bonepit_62_0.mao"},
	{0xA0623C1B5D962286ULL, "w_bonepit_62_0.mmh"},
	{0xA06268395DBD65A3ULL, "mag291_torpor6026960.cl"},
	{0xA062DE15C774C083ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_basic.msh"},
	{0xA062E297805D8F99ULL, "vxlt_gallowstemplar_519_0_0.vlm"},
	{0xA062EC15C774D8DDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_basic.mmh"},
	{0xA062F015C774DFEEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_basic.mao"},
	{0xA0633FAE99CE09C5ULL, "vxw_bonepitdestroyed_a3_2ar_0_0.vlm"},
	{0xA06365CAFCF47DCEULL, "hlac_lt_0_lowtown_0_7x.dds"},
	{0xA0636885FE132CADULL, "6085232_f.fxe"},
	{0xA0636A337C2B0167ULL, "vxca_cavern1_haunted_hul_0_0.vlm"},
	{0xA063B3D71D486076ULL, "dae000_flemeth6000747.cl"},
	{0xA063CC6B6098544FULL, "ss_qunari_m_3.bnk"},
	{0xA063CCA0628706A9ULL, "qun210cr_guard6072158.cl"},
	{0xA0640A46E88D4993ULL, "vxw_encounter2_bq_0_0.vlm"},
	{0xA064DE9B4AC1FAE2ULL, "vxht_mansion2b_3g9_0_0.vlm"},
	{0xA0659C703387684BULL, "vxlt_gallowstemplar_f_2iw_0_0.vlm"},
	{0xA06630AA722A537CULL, "t3_shl_rfi.tnt"},
	{0xA06681BE6ED475D2ULL, "6085614_m.fxe"},
	{0xA066B06329FD16F1ULL, "6181752_f.fxe"},
	{0xA066B8B95D1CB8DDULL, "art\\vfx\\combat\\vfx_m_sta_atk_l_dwn_c.mmh"},
	{0xA06721DCD6EA0369ULL, "vxca_caverns1_i2o_0_0.vlm"},
	{0xA067DFEF463A514BULL, "gen00fl_sebastian6198130.cl"},
	{0xA06847A2C0E2ACCEULL, "vxdr_slavers_1t7_0_0.vlm"},
	{0xA0692AE654296AF0ULL, "395368464.ogg"},
	{0xA069BDA562C4CDBCULL, "820107448.ogg"},
	{0xA069E7D36BECFD17ULL, "840141559.ogg"},
	{0xA06A3A31B4A587CAULL, "vxlt_gallowstemplar_5xi_0_0.vlm"},
	{0xA06A5A5F2FE3EC9EULL, "zz_mrl_debug6109303.cl"},
	{0xA06A7D1693CFD64AULL, "177835481.ogg"},
	{0xA06AA288DD0BB465ULL, "vxw_sundermount_1zg_0_0.vlm"},
	{0xA06ADDC50032BA02ULL, "1047005560.ogg"},
	{0xA06B2540415D4BDCULL, "llac_ht_mansion_n_3373_0.dds"},
	{0xA06B719E32FAF747ULL, "follower_banter6163552.cl"},
	{0xA06BB1C3799C7857ULL, "6104826_m.fxe"},
	{0xA06CE31D995364B1ULL, "hlac_lt_hangedman_n_-1_14x.dds"},
	{0xA06D54DA4F728ACCULL, "gen00fl_fenris6116552.cl"},
	{0xA06DE792F9ACFBF6ULL, "633394165.ogg"},
	{0xA06DF9E6D78BF57BULL, "vxdr_0_deeproads_cqq_0_0.vlm"},
	{0xA06E251CBB1BD431ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medp_0n.dds"},
	{0xA06E2FB7ECE35DFCULL, "vxht_mansion2_2kb_0_0.vlm"},
	{0xA06E6B6A26ED1B3BULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyn_0t.dds"},
	{0xA06E7AE2E19E90C0ULL, "zz_lgt_1_debug6114617.cl"},
	{0xA06EFA310A79C96AULL, "llac_ca_caverns2_5802_70.dds"},
	{0xA06FA45E317E4142ULL, "173141795.ogg"},
	{0xA06FE626D3E49CCAULL, "w_bonepit_166_0.phy"},
	{0xA06FFD7D144E1F1DULL, "vxlt_0_lowtown_a2_eki_0_1.vlm"},
	{0xA0701D4B099FB8EAULL, "vxw_bonepitdestroyed_a3_19b_0_0.vlm"},
	{0xA07040105D05AD3BULL, "llac_lt_0_lowtown_a2_n_65537_7.dds"},
	{0xA070FC52002D1BE6ULL, "6093819_m.fxe"},
	{0xA0711E57FE7904E7ULL, "6067682_m.fxe"},
	{0xA0722D570E41A877ULL, "6117509_m.fxe"},
	{0xA0727B415299E47AULL, "gen_im_flw_one_aveline_01.uti"},
	{0xA0727C23DA836213ULL, "vxw_encounter2_mp_0_0.vlm"},
	{0xA072F257ABEE1DBCULL, "hm_wesley_inf.mor"},
	{0xA0733E23E5148210ULL, "6165392_m.fxe"},
	{0xA0734F04BFB6DC7DULL, "gen00fl_isabela_init6136307.cl"},
	{0xA0735FC08DD3AE24ULL, "vxlt_gallowstemplar_n_5rf_0_0.vlm"},
	{0xA07416ACCECDA98CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_door.msh"},
	{0xA0741A79D79B6C86ULL, "weapon_weapon_p1.psh.fxo"},
	{0xA0741CACCECDB47EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_door.mmh"},
	{0xA07428ACCECDC815ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_door.mao"},
	{0xA074308E540540B8ULL, "6043795_m.fxe"},
	{0xA07466D6A983B11AULL, "ht_bartrand_n_3367.anb"},
	{0xA074BA82835488DEULL, "895087485.ogg"},
	{0xA07533D54241C85CULL, "vxca_caverns2_6wa_0_0.vlm"},
	{0xA0753724D66FA4F1ULL, "vxlt_0_lowtown_n_e2p_0_0.vlm"},
	{0xA07554615442EF8DULL, "mag291_torpor6116225.cl"},
	{0xA0758A69E76780D3ULL, "486285790.ogg"},
	{0xA0759C19FC8F69BCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junke.mao"},
	{0xA075A819FC8F7E5FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junke.mmh"},
	{0xA075AE19FC8F8801ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junke.msh"},
	{0xA076A4FC0CCE87D1ULL, "260831834.ogg"},
	{0xA076A97A6DF6A0F7ULL, "vxlt_gallowstemplar_f_37d_0_0.vlm"},
	{0xA07794852D0412A3ULL, "vxlt_gallowstemplar_f_4xv_0_0.vlm"},
	{0xA077B492D0578194ULL, "zz_followers_banter6158232.cl"},
	{0xA077E6FEBE127EBEULL, "follower_banter6143333.cl"},
	{0xA07964BC155A6D8AULL, "hlac_dr_dungeon_3214_11x.dds"},
	{0xA0798317052CB2BBULL, "hlac_ht_0_hightown_65536_3z.dds"},
	{0xA0798B4DE87CC539ULL, "hlac_ca_cavern1_haunted_24421_4z.dds"},
	{0xA079A5E0DAFCB53CULL, "347897372.ogg"},
	{0xA079BCBB9B769153ULL, "vxht_0_hightown_a2_cf7_0_0.vlm"},
	{0xA079E19E9D4387E1ULL, "276319818.ogg"},
	{0xA07A70CD8C555F38ULL, "zz_followers_banter6157996.cl"},
	{0xA07B42DAEA59F3A1ULL, "dae000_pc_home_ambients6208555.cl"},
	{0xA07BDD431416A295ULL, "zz_vault_debug6200338.cl"},
	{0xA07BF0206115605CULL, "93261034.ogg"},
	{0xA07C916957A5972CULL, "vxht_bartrand_n_5se_0_0.vlm"},
	{0xA07CF9874ED5AE56ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.impa.ani"},
	{0xA07DB0B9C9A07487ULL, "603282648.ogg"},
	{0xA07DB7D2B523887EULL, "hlac_w_blightlands_262141_0y.dds"},
	{0xA07DC98F12F0CA09ULL, "llac_lt_0_lowtown_docks_a3_65533_4.dds"},
	{0xA07E2D7C0569D15AULL, "vxw_bonepit_1kx_0_0.vlm"},
	{0xA07E2EC5D8D2F2BCULL, "ch1alphnotnt_vl.psh.fxo"},
	{0xA07E48B2E0649A82ULL, "vxlt_warehouse_n_147_0_0.vlm"},
	{0xA07E710048D257E2ULL, "vxht_0_hightown_n_9zs_0_0.vlm"},
	{0xA07E8C449F398621ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_laundry_d.dds"},
	{0xA07E932FB9967895ULL, "6120109_m.fxe"},
	{0xA07EC70155D4C0CCULL, "fen110_fenris6096226.cl"},
	{0xA07EC79FFF295992ULL, "6051466_m.fxe"},
	{0xA07F64B9D2C58894ULL, "6139793_m.fxe"},
	{0xA07FF8716673B26AULL, "vxca_cavern1_haunted_ine_0_0.vlm"},
	{0xA07FFD3CA2C7C804ULL, "6144379_m.fxe"},
	{0xA08028028825B9F7ULL, "6133547_m.fxe"},
	{0xA0802F5E40A3AA69ULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf_origional.ani"},
	{0xA0809604046DE1C9ULL, "vxlt_0_lowtown_h4x_0_0.vlm"},
	{0xA080EE00CC797A93ULL, "vxw_encounter2_hs_0_0.vlm"},
	{0xA081402F82018BBDULL, "91606395.ogg"},
	{0xA08166267C297317ULL, "571318334.ogg"},
	{0xA0818196441FE651ULL, "vxlt_0_lowtown_docks_a3_bqw_0_0.vlm"},
	{0xA081C394A32C808AULL, "781027155.ogg"},
	{0xA0820EA7E4D8FED4ULL, "6114495_m.fxe"},
	{0xA082223E41043FD1ULL, "vxca_cavern1_haunted_gj8_0_0.vlm"},
	{0xA08236B1E28FD748ULL, "285022300.ogg"},
	{0xA0828C83BFCF81F7ULL, "ts_money.gda"},
	{0xA082A5F2135F9DF2ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier5_arishokresponse.gad"},
	{0xA082CE428041C453ULL, "w_encounter2_0_0.rml"},
	{0xA082EFA09184A914ULL, "111271535.ogg"},
	{0xA0833D715CCB2FD2ULL, "vxht_0_hightown_a2_cf2_0_0.vlm"},
	{0xA0834800261E5BB7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleftcrn.msh"},
	{0xA0835A00261E7A1AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleftcrn.mao"},
	{0xA0835E00261E8129ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandleftcrn.mmh"},
	{0xA083952ACC0AA50FULL, "vxdr_primevalentrance_1i9_0_0.vlm"},
	{0xA0840596BAD63469ULL, "vxca_caverns1_i8m_0_0.vlm"},
	{0xA0840E9766E6E94AULL, "vxht_0_hightown_a2_4ap_0_0.vlm"},
	{0xA0849A0B6DE06D9BULL, "art\\characters\\playercharacter\\humanmale\\hlh_medf.mmh"},
	{0xA084A60B6DE08238ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medf.mao"},
	{0xA084B80B6DE0A085ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medf.msh"},
	{0xA08544A193F103EAULL, "454282346.ogg"},
	{0xA08546131260C70DULL, "dae000_bartrand6107534.cl"},
	{0xA085526A4EE5AF66ULL, "art\\vfx\\combat\\vfx_glb_ele_f_extreme_c.anb"},
	{0xA0861C2A3734B3DFULL, "tm_war_d01.tnt"},
	{0xA086536AD1789230ULL, "vxw_encounter2_g9_0_0.vlm"},
	{0xA086752503F9976FULL, "dae000ip_crowd_lm3_talk17.utp"},
	{0xA0867F1B51F41DADULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saintb.mmh"},
	{0xA086831B51F424FEULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saintb.mao"},
	{0xA08683CACB035EACULL, "lt_warehouse_n_1_0.phy"},
	{0xA086911B51F43C33ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saintb.msh"},
	{0xA0869B7AC68C171CULL, "661667765.ogg"},
	{0xA086DA6674A5373BULL, "207320786.ogg"},
	{0xA08718DDD0C850FEULL, "6119964_m.fxe"},
	{0xA0874A5E40A9F6EBULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf_origional.gad"},
	{0xA087A4D650F59967ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_sand01_n.dds"},
	{0xA087AD862E1F0173ULL, "457987096.ogg"},
	{0xA08820A12F8F25C1ULL, "452563750.ogg"},
	{0xA08868612228468BULL, "vxht_keep_n_60k_0_0.vlm"},
	{0xA089420420199A7DULL, "dae000_gamlen6048413.cl"},
	{0xA08942A4D79F460EULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_aoe_p.impa.evt"},
	{0xA089615E52F985CBULL, "lgt260_eau_tmp_roderick6143651.cl"},
	{0xA08993C423F008B6ULL, "vxht_0_hightown_a2_c3m_0_0.vlm"},
	{0xA089D5428047EEABULL, "w_encounter2_0_0.phy"},
	{0xA089F265F840BE0FULL, "206167241.ogg"},
	{0xA08A01053CFDE2C0ULL, "vxw_bonepit_1ih_0_0.vlm"},
	{0xA08A01874EE04647ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.impa.evt"},
	{0xA08A6E19D2BE1722ULL, "6189028_m.fxe"},
	{0xA08ACC04A22292D6ULL, "vxlt_hangedman_4y_0_0.vlm"},
	{0xA08AFCF08C55D06FULL, "follower_banter6171894.cl"},
	{0xA08B21832FB95872ULL, "qun110_petrice6018514.cl"},
	{0xA08BB8244BF9EAE0ULL, "6093520_f.fxe"},
	{0xA08BE2E6C595F7C5ULL, "one350_variel6076871.cl"},
	{0xA08C98BA39D512AFULL, "cntr_dot.mao"},
	{0xA08CB9147EC670BAULL, "6174727_m.fxe"},
	{0xA08CBA8FD25AB302ULL, "vxlt_warehouse_n_190_0_0.vlm"},
	{0xA08CBEE1886963A7ULL, "vxht_0_hightown_a3_a7k_0_0.vlm"},
	{0xA08CCBF3B5410ABCULL, "74788558.ogg"},
	{0xA08CE6BAE564C051ULL, "513136536.ogg"},
	{0xA08D5079DC0A17DFULL, "vxht_0_hightown_c1r_0_0.vlm"},
	{0xA08D5ACACB0937B4ULL, "lt_warehouse_n_1_0.rml"},
	{0xA08E3C7F77840A59ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_ext_c.mmh"},
	{0xA08E475E40B01050ULL, "art\\characters\\creatures\\stonegolem\\animation\\crnf_origional.evt"},
	{0xA08E6A4E5869CAEAULL, "vxca_cavern1_haunted_ikf_0_0.vlm"},
	{0xA08E77F863427E33ULL, "6123332_f.fxe"},
	{0xA08F4977BC421CBBULL, "763123913.ogg"},
	{0xA08FA53F1BD94116ULL, "vxdr_0_deeproads_d3a_0_0.vlm"},
	{0xA08FD104B33FCC64ULL, "6143971_f.fxe"},
	{0xA08FEEE942D59E3EULL, "8105631.ogg"},
	{0xA09039A4D7A555B9ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_aoe_p.impa.gad"},
	{0xA0904249B3DD4489ULL, "971398530.ogg"},
	{0xA09068EEAB98505CULL, "6076085_m.fxe"},
	{0xA0911F0A89F64574ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_left.gad"},
	{0xA09122874EE69C90ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.impa.gad"},
	{0xA0912EF64E0C18A7ULL, "182977526.ogg"},
	{0xA0915F0796CEE2E8ULL, "mer221_sabin6047527.cl"},
	{0xA09173D347F7DC81ULL, "6079650_m.fxe"},
	{0xA091D630ED437271ULL, "6206562_m.fxe"},
	{0xA092556A4EF03E11ULL, "art\\vfx\\combat\\vfx_glb_ele_f_extreme_c.mmh"},
	{0xA0925C8557363CCCULL, "hlac_w_woundedcoast_1_0x.dds"},
	{0xA093F62F01455E73ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdblc.phy"},
	{0xA094250E729CE968ULL, "zz_dae_debug6029787.cl"},
	{0xA09475F498C3564BULL, "vxlt_hangedman_rj_0_0.vlm"},
	{0xA094BDD6B70F5607ULL, "6125416_m.fxe"},
	{0xA094FCD2B3020FFBULL, "mcr310_argument6074383.cl"},
	{0xA0954FC6B3840C67ULL, "ts_generic_prog_statuette.gda"},
	{0xA0956037E469D1CAULL, "15918699.ogg"},
	{0xA09585ED0050ECE0ULL, "llac_lt_0_lowtown_docks_a3_-4_6.dds"},
	{0xA095AE8A8ACF717BULL, "vxht_chantry_2cq_0_0.vlm"},
	{0xA0960193CAF5517EULL, "6150343_m.fxe"},
	{0xA09675102B6C72CDULL, "ave310cr_guard_f.utc"},
	{0xA0968C1615A490DEULL, "6089791_m.fxe"},
	{0xA0969CBC7F9E8F80ULL, "vxht_0_hightown_a2_at8_0_0.vlm"},
	{0xA096AEF25E90B573ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.dura.evt"},
	{0xA096F374B61FC639ULL, "863771683.ogg"},
	{0xA0974AA4D7AB9107ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_aoe_p.impa.ani"},
	{0xA09750A1DA48CDF5ULL, "25120012.ogg"},
	{0xA0976E88C9E70856ULL, "894163166.ogg"},
	{0xA097781B5202856AULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_saintb.phy"},
	{0xA09790F21371D598ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier5_arishokresponse.ani"},
	{0xA097DBA5EA17B6E6ULL, "hlac_lt_backalley_n_1_4z.dds"},
	{0xA0982E0A89FC7D52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\eyes_left.ani"},
	{0xA09858B68C856A82ULL, "mag120_macha6080607.cl"},
	{0xA099575D99CCC7EEULL, "657617257.ogg"},
	{0xA099854EC2C5A77CULL, "vxht_mansion2_2za_0_0.vlm"},
	{0xA0999B0EF56A1D9CULL, "vxlt_0_lowtown_n_ey0_0_0.vlm"},
	{0xA09A401B4E743491ULL, "zz_fenris_glow_v1.cut"},
	{0xA09AA7D556838951ULL, "263832595.ogg"},
	{0xA09AC7A47690D4BCULL, "ave200_donnic6128942.cl"},
	{0xA09B94D8E4DD0BCEULL, "vxlt_0_lowtown_a3_n_c_coy_0_0.vlm"},
	{0xA09B95CC409E9E0FULL, "plt_prm_cod_cliffs_tome.plo"},
	{0xA09C19C7C9B39CA4ULL, "lgt320_shw_ors_marlein6156490.cl"},
	{0xA09C99A16147C53FULL, "257990410.ogg"},
	{0xA09C9E435EE00FE1ULL, "hlac_lt_0_lowtown_docks_n_65535_4x.dds"},
	{0xA09CFC5DF758EBF3ULL, "227721354.ogg"},
	{0xA09D3284A2C7A489ULL, "459438224.ogg"},
	{0xA09D374BF1471DB7ULL, "mrl310_party_comments6142845.cl"},
	{0xA09D4E5FD171425CULL, "210496046.ogg"},
	{0xA09D845FD0CB37E2ULL, "650441035.ogg"},
	{0xA09DA7F25E96C7B4ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.dura.gad"},
	{0xA09DB104229FB057ULL, "llac_w_sundermount_-65536_0.dds"},
	{0xA09DE8391DC52DDEULL, "plt_mrl150pt_longway.plo"},
	{0xA09E5BD6C35D9D48ULL, "dae000cr_flemeth.utc"},
	{0xA09E76CA0B5D5573ULL, "dae000cr_grand_cleric6173650.cl"},
	{0xA09EB55763B5F7A4ULL, "vxlt_gallowstemplar_n_5ce_0_0.vlm"},
	{0xA09F246C2B820E22ULL, "gen00fl_fenris6126997.cl"},
	{0xA0A01C57CAB61B66ULL, "fen110_fenris6118005.cl"},
	{0xA0A01E37FD556208ULL, "vxw_encounter3_do_0_0.vlm"},
	{0xA0A03F0DE82CBA08ULL, "6137522_m.fxe"},
	{0xA0A0BA1EACD502A6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkback.ani"},
	{0xA0A14EFA003FE672ULL, "238871922.ogg"},
	{0xA0A1C008D19CF649ULL, "6143186_m.fxe"},
	{0xA0A23FE0F4D87C3FULL, "6082319_m.fxe"},
	{0xA0A2875CA003F0D9ULL, "hlac_w_sundermount_196611_0z.dds"},
	{0xA0A339C0AE5F72F5ULL, "zz_followers_banter6159180.cl"},
	{0xA0A393FCF709D81DULL, "hlac_lt_0_lowtown_a2_-3_6x.dds"},
	{0xA0A3B6F25E9B4C92ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.dura.ani"},
	{0xA0A3DAA67FDD401BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kitb_0d.dds"},
	{0xA0A3EC744FE62323ULL, "579364074.ogg"},
	{0xA0A43A289FB6995CULL, "vxlt_undercityencounter_zh_0_0.vlm"},
	{0xA0A5F78BA099F03FULL, "blood_conj_04_l2.mao"},
	{0xA0A61388E54E47FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_3windows.mao"},
	{0xA0A61F88E54E5B9FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_3windows.mmh"},
	{0xA0A62588E54E6641ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_3windows.msh"},
	{0xA0A67C640208AF68ULL, "w_sundermount_250_0.msh"},
	{0xA0A68E640208CDF9ULL, "w_sundermount_250_0.mao"},
	{0xA0A69A640208E192ULL, "w_sundermount_250_0.mmh"},
	{0xA0A6A0CE113FCB76ULL, "vxlt_0_lowtown_fha_0_0.vlm"},
	{0xA0A769D6A9AF1046ULL, "ht_bartrand_n_3367.rml"},
	{0xA0A818ABCBE8D36BULL, "6136166_m.fxe"},
	{0xA0A82BADF7C563C8ULL, "6113645_m.fxe"},
	{0xA0A8339D28AB47DBULL, "vxlt_0_lowtown_a2_n_gck_0_0.vlm"},
	{0xA0A87B67020EA6F6ULL, "6028848_m.fxe"},
	{0xA0A899FC186CF411ULL, "vxca_cavern1_haunted_hkx_0_0.vlm"},
	{0xA0A8A9B7DF9C4C51ULL, "532458943.ogg"},
	{0xA0A980711B258333ULL, "150134852.ogg"},
	{0xA0A9B89EC5BCA27EULL, "hlac_w_sundermount_-196611_0x.dds"},
	{0xA0AA77CB489C436AULL, "399879371.ogg"},
	{0xA0AB325F5E29E3D6ULL, "vxlt_0_lowtown_n_ewd_0_0.vlm"},
	{0xA0ABAD33488FDC98ULL, "hlac_ca_caverns2_5802_43z.dds"},
	{0xA0ABD6F2250A3E7AULL, "vxw_bonepitdestroyed_a3_29q_0_0.vlm"},
	{0xA0AC62153ED5154EULL, "hasachievement.ncs"},
	{0xA0ADA21EACDF6417ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkback.evt"},
	{0xA0AE9014EF4B7A88ULL, "vxw_encounter3_al_0_0.vlm"},
	{0xA0AED4232E445FF6ULL, "6017496_f.fxe"},
	{0xA0AED6999C0C7680ULL, "vxca_cavern1_haunted_dpx_0_0.vlm"},
	{0xA0AEDA529BD0A953ULL, "vxca_cavern1_haunted_fly_0_0.vlm"},
	{0xA0AFF69D60AB42ACULL, "vxht_keep_4b3_0_0.vlm"},
	{0xA0B04DE21C39E783ULL, "332866784.ogg"},
	{0xA0B07A2172AA1B7BULL, "lgt100cr_gang_1_docks.utc"},
	{0xA0B08897B853D944ULL, "243115181.ogg"},
	{0xA0B0E898F900A799ULL, "vxca_caverns2_8m4_0_0.vlm"},
	{0xA0B1ECA535F289E6ULL, "479479411.ogg"},
	{0xA0B292446EBBA27BULL, "vxht_chantry_2ys_0_0.vlm"},
	{0xA0B30253B14F77D2ULL, "105271585.ogg"},
	{0xA0B36FCFF990C1DEULL, "6087591_m.fxe"},
	{0xA0B380E68EEB9F1BULL, "llac_lt_0_lowtown_a2_n_c_2_1.dds"},
	{0xA0B3A6E8968D6549ULL, "llac_ht_0_hightown_a3_65532_0.dds"},
	{0xA0B3BB8C4C1B1E31ULL, "hlac_lt_darktown_-1_5x.dds"},
	{0xA0B42211BCFACAABULL, "fighting80a.mao"},
	{0xA0B4A31EACE584A0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkback.gad"},
	{0xA0B4EA3671820153ULL, "gen_im_acc_rng_act3_13.uti"},
	{0xA0B4F3ACCF050519ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_door.phy"},
	{0xA0B4FA740F03F376ULL, "189970600.ogg"},
	{0xA0B54202340ED1E8ULL, "983897392.ogg"},
	{0xA0B5749FDC04563CULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_inquisitionsoldier1_.ani"},
	{0xA0B5A008E719258BULL, "vxht_mansion_n_6n1_0_0.vlm"},
	{0xA0B612D76645D0C1ULL, "hlac_lt_0_lowtown_-65536_2z.dds"},
	{0xA0B6232C640F3E9AULL, "art\\vfx\\materials\\vfx_debris_blurred.mao"},
	{0xA0B643B9EC9FF443ULL, "6163151_m.fxe"},
	{0xA0B696DE26AAAFF7ULL, "6165175_m.fxe"},
	{0xA0B697F4A0C16285ULL, "203009115.ogg"},
	{0xA0B6C85FC1D3B161ULL, "6142439_m.fxe"},
	{0xA0B723B5B948C217ULL, "w_encounter1_3_0.rml"},
	{0xA0B7397F77A70A4EULL, "art\\vfx\\combat\\vfx_glb_nat_imp_ext_c.anb"},
	{0xA0B7487D16CEE21EULL, "6080837_m.fxe"},
	{0xA0B7A0753D3BA55CULL, "vxlt_gallowstemplar_3qr_0_0.vlm"},
	{0xA0B7CE24D425A429ULL, "307531110.ogg"},
	{0xA0B825F477099868ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobb.mmh"},
	{0xA0B831F47709AC0BULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobb.mao"},
	{0xA0B843F47709CA92ULL, "art\\characters\\playercharacter\\humanmale\\hoo_nobb.msh"},
	{0xA0B8A651EA0A6E10ULL, "ico_mace_1.dds"},
	{0xA0B900AB69B22455ULL, "6119611_m.fxe"},
	{0xA0B94A9316D5B2E6ULL, "dae200cr_viscount6087194.cl"},
	{0xA0B97158B87C28EFULL, "211155217.ogg"},
	{0xA0B97BCACB2EAAB4ULL, "lt_warehouse_n_1_0.anb"},
	{0xA0BA370FFC499D13ULL, "250673037.ogg"},
	{0xA0BA48CF170056FAULL, "vxw_bonepitdestroyed_a3_24r_0_0.vlm"},
	{0xA0BAA763BDCA82E4ULL, "combat_ar_r5_a1.gda"},
	{0xA0BABFF908744C7DULL, "vxlt_warehouse_n_8k_0_0.vlm"},
	{0xA0BAF0F688FDDBD2ULL, "594368326.ogg"},
	{0xA0BB2BE30F9A70EBULL, "vxdr_primevaltomb_7eg_0_0.vlm"},
	{0xA0BB31A504817D9FULL, "848852875.ogg"},
	{0xA0BBFAED553F552CULL, "llac_lt_gallowstemplar_n_7198_4.dds"},
	{0xA0BC69B51EC709FAULL, "dae000_mother6103051.cl"},
	{0xA0BC819FDC0A8B7EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_inquisitionsoldier1_.gad"},
	{0xA0BC9953EE66CB44ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barreltiny.msh"},
	{0xA0BCAB53EE66E9ADULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barreltiny.mao"},
	{0xA0BCAF53EE66F0E6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barreltiny.mmh"},
	{0xA0BE07B7C242642CULL, "vxlt_undercityencounter_139_0_1.vlm"},
	{0xA0BE0FF071EC137CULL, "dae150_chandan6148641.cl"},
	{0xA0BE22B5B94EDF0FULL, "w_encounter1_3_0.phy"},
	{0xA0BE29FAF1BE3E73ULL, "vxlt_darktown_5lu_0_1.vlm"},
	{0xA0BE5369752ADB8EULL, "410650994.ogg"},
	{0xA0BE86C9489EC916ULL, "519638853.ogg"},
	{0xA0BF75E3DF0A46E8ULL, "6121994_m.fxe"},
	{0xA0BF7C545B790BEDULL, "vxlt_house_dq_0_0.vlm"},
	{0xA0BF9C601BA68F91ULL, "one120cr_magistrate.utc"},
	{0xA0BFE703EC9557B8ULL, "6127048_m.fxe"},
	{0xA0C01D5D88CD2B66ULL, "hlac_lt_0_lowtown_docks_n_-1_3z.dds"},
	{0xA0C037B38CB769A8ULL, "6111319_m.fxe"},
	{0xA0C075FB18751A98ULL, "72775853.ogg"},
	{0xA0C0AA6B7A9D6B42ULL, "w_encounter2_1_0.rml"},
	{0xA0C1295CCBBED789ULL, "hlac_w_blightlands_327676_0y.dds"},
	{0xA0C12C7328E824B4ULL, "nat380_nathaniel6090042.cl"},
	{0xA0C1B7BE91F514A2ULL, "vxlt_0_lowtown_a2_gik_0_0.vlm"},
	{0xA0C1D5707D6FD5D0ULL, "llac_lt_gallowstemplar_2390_0.dds"},
	{0xA0C1EDD90CD7722DULL, "6128893_m.fxe"},
	{0xA0C213317C6BF340ULL, "plt_lgt323pt_shw_ors_blckop.plo"},
	{0xA0C2BF6927D4EB47ULL, "vxht_mansion2_3g8_0_0.vlm"},
	{0xA0C2CB03686C8E07ULL, "152427066.ogg"},
	{0xA0C33CFD1CC12CBDULL, "vxht_0_hightown_a2_b9a_0_0.vlm"},
	{0xA0C36B75278A58EFULL, "6065833_m.fxe"},
	{0xA0C388A515A93B7AULL, "535072653.ogg"},
	{0xA0C391F7014D2B1BULL, "6170313_m.fxe"},
	{0xA0C3AD7BE61030E3ULL, "highleveloadhints.gda"},
	{0xA0C4A13180134934ULL, "6119179_m.fxe"},
	{0xA0C4F0A5CB4C671CULL, "vxlt_0_lowtown_n_eh7_0_0.vlm"},
	{0xA0C5561785D29E46ULL, "6119656_m.fxe"},
	{0xA0C5BF836FABC6A4ULL, "ht_0_hightown_a2_mmi.gff"},
	{0xA0C6093E3704E60CULL, "gen00fl_aveline6139253.cl"},
	{0xA0C6A653EE6F35B1ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barreltiny.phy"},
	{0xA0C6EF02BBB77E86ULL, "mcr310_argument6074515.cl"},
	{0xA0C6FF0078F62E82ULL, "hlac_w_woundedcoast_0_8y.dds"},
	{0xA0C7026D36B8794BULL, "6110485_m.fxe"},
	{0xA0C719B0BC8E3B03ULL, "6015060_f.fxe"},
	{0xA0C75ACACB3AB077ULL, "lt_warehouse_n_1_0.mmh"},
	{0xA0C75ECACB3AB684ULL, "lt_warehouse_n_1_0.mao"},
	{0xA0C770CACB3AD5E9ULL, "lt_warehouse_n_1_0.msh"},
	{0xA0C772885BA83CD5ULL, "6114111_m.fxe"},
	{0xA0C7AB6B7AA38BF6ULL, "w_encounter2_1_0.phy"},
	{0xA0C7CFE211F37748ULL, "mer210_seldrek6078795.cl"},
	{0xA0C7F3B0359A87F5ULL, "vxlt_0_lowtown_a3_n_c_gcf_0_0.vlm"},
	{0xA0C8324C6220D5B5ULL, "lowtown_fem_alienage1_sta6162602.cl"},
	{0xA0C84982250CF73DULL, "lt_0_lowtown_docks_mmi.gff"},
	{0xA0C8784889F7AC11ULL, "hlac_lt_0_lowtown_docks_a3_-2_3z.dds"},
	{0xA0C89C2F64CB48B6ULL, "fx_lightning03.dds"},
	{0xA0C8B8C8E33FFD64ULL, "hlac_ht_0_hightown_n_65532_0y.dds"},
	{0xA0C8C8BB76722734ULL, "vxlt_0_lowtown_a2_n_c_e23_0_0.vlm"},
	{0xA0C90684BEDFD604ULL, "vxlt_0_lowtown_def_0_0.vlm"},
	{0xA0C92AB1A0151B41ULL, "art\\levels\\ztools\\default\\generic\\1pp_basic.phy"},
	{0xA0C9909228409D28ULL, "pro000cr_athenril_entouragm.utc"},
	{0xA0C9CFC0404DFD71ULL, "6103065_m.fxe"},
	{0xA0CA28051DA56E7BULL, "ht_brothel_n_3370_lc.mmh"},
	{0xA0CADF7CA22255ABULL, "gen00fl_aveline6139959.cl"},
	{0xA0CB1BFB99E8FFEAULL, "vxlt_gallowscourtyard_7cp_0_0.vlm"},
	{0xA0CB1F378E90A726ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_s_p.anb"},
	{0xA0CB40C89899BB38ULL, "gen00fl_anders6125400.cl"},
	{0xA0CBBB0B6E1CA8D0ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medf.phy"},
	{0xA0CCFB2080362602ULL, "vxht_0_hightown_a2_9x6_0_0.vlm"},
	{0xA0CDAFD5ACEE8102ULL, "gen00flst_fenris_romance.stg"},
	{0xA0CDC9406DF00AE3ULL, "mag330_walter6045968.cl"},
	{0xA0CDF92B8E5397ECULL, "6127115_f.fxe"},
	{0xA0CE72718832FF3DULL, "vxca_caverns2_832_0_0.vlm"},
	{0xA0CECAA7C7B7662EULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_3x.dds"},
	{0xA0CEEC357C5103CAULL, "follower_banter6172250.cl"},
	{0xA0CF0011996C8EDAULL, "6064686_m.fxe"},
	{0xA0CF3D871CE6E621ULL, "mer340cr_hubert_guard_2.utc"},
	{0xA0CF8CA4491C47FCULL, "hlac_lt_0_lowtown_a2_n_-131073_0x.dds"},
	{0xA0CFB74F3B89BD55ULL, "vxlt_darktown_43j_0_0.vlm"},
	{0xA0D0299B83EB2D22ULL, "vxlt_0_lowtown_a2_glt_0_0.vlm"},
	{0xA0D1346D4EA6D317ULL, "193024066.ogg"},
	{0xA0D1616C928A2297ULL, "hlac_lt_0_lowtown_docks_-1_3x.dds"},
	{0xA0D1946B7AABB897ULL, "w_encounter2_1_0.msh"},
	{0xA0D1A66B7AABD7FAULL, "w_encounter2_1_0.mao"},
	{0xA0D1AA6B7AABDE09ULL, "w_encounter2_1_0.mmh"},
	{0xA0D1D841742F3464ULL, "llac_lt_0_lowtown_n_65535_0.dds"},
	{0xA0D1DCF99418C0F2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratetarp_sa.dds"},
	{0xA0D2CFD303D7A809ULL, "cd_vrc221_bartrand.wbd"},
	{0xA0D3AB103181B008ULL, "vxlt_gallowscourtyard_651_0_0.vlm"},
	{0xA0D3F0C7A6AC4D02ULL, "zz_and_debug6134127.cl"},
	{0xA0D45EE9BFB97FB3ULL, "dae000_pc_home_ambients6208575.cl"},
	{0xA0D4BDD17997CD9AULL, "gen00fl_anders6142553.cl"},
	{0xA0D54562D9E37078ULL, "6081310_f.fxe"},
	{0xA0D58E7A2F614559ULL, "46079971.ogg"},
	{0xA0D5E4654D9E5555ULL, "6113011_m.fxe"},
	{0xA0D6377EA4F60A83ULL, "mag101_alain6052999.cl"},
	{0xA0D67C56F4D0B8A8ULL, "386440236.ogg"},
	{0xA0D67FAFD6C57CEBULL, "406160831.ogg"},
	{0xA0D687F6340C2D56ULL, "vxlt_0_lowtown_n_efg_0_0.vlm"},
	{0xA0D6A68071C65C6CULL, "qcr220_grey_wardens6060438.cl"},
	{0xA0D6CE96297ECDC5ULL, "vxw_bonepitdestroyed_a3_29z_0_0.vlm"},
	{0xA0D7032A9ABCD655ULL, "qun131_taskmaster6019219.cl"},
	{0xA0D7B119FCE257A4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_junke.phy"},
	{0xA0D7C35F555A3A94ULL, "79565686.ogg"},
	{0xA0D7D2A5B8B93EB8ULL, "68425881.ogg"},
	{0xA0D7F68FC83B242DULL, "zz_vrcpt_debug.ncs"},
	{0xA0D8044764EFE79CULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_fleshpod_02d.dds"},
	{0xA0D922378E9CE9D1ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_s_p.mmh"},
	{0xA0D9601D226F58A1ULL, "ser270_emeric6113511.cl"},
	{0xA0D976D4D6B3F9CAULL, "6000910_m.fxe"},
	{0xA0DA06A3A8E56B85ULL, "vxdr_0_deeproads_eh2_0_0.vlm"},
	{0xA0DA305328B013E2ULL, "gen00fl_anders6104014.cl"},
	{0xA0DA5F49AD0E0317ULL, "6091870_m.fxe"},
	{0xA0DA839DD3C90297ULL, "vxlt_gallowstemplar_4qj_0_0.vlm"},
	{0xA0DA999889D5A969ULL, "vxht_keep_n_6j3_0_0.vlm"},
	{0xA0DAFC0CD3F8740EULL, "276955077.ogg"},
	{0xA0DB0A25BCA5350CULL, "vxlt_gallowsprison_n_6ho_0_0.vlm"},
	{0xA0DB8D2613D49E83ULL, "follower_banter6149318.cl"},
	{0xA0DB9D5DB2ACF6D3ULL, "vxca_caverns2_7tj_0_0.vlm"},
	{0xA0DBF91091367B45ULL, "qun100cr_qunari_warrior4.utc"},
	{0xA0DC1886D212E6EEULL, "qun160ip_at_cave.utp"},
	{0xA0DCDE7DAEBD48F6ULL, "hlac_lt_0_lowtown_docks_a3_131070_4y.dds"},
	{0xA0DCEDAC3A0E0F9BULL, "lt_0_lowtown_docks_a3_27_0.rml"},
	{0xA0DD56F2849675ECULL, "character_character_p120.psh.fxo.dx11.assembly"},
	{0xA0DD7651903399C3ULL, "6172492_m.fxe"},
	{0xA0DE621D667ACFBFULL, "vxlt_0_lowtown_docks_a3_n_9rd_0_0.vlm"},
	{0xA0DEA510588386F0ULL, "zz_approval_debug6003694.cl"},
	{0xA0DF0E2C87B10564ULL, "292672156.ogg"},
	{0xA0DF0EAFDE140CC5ULL, "184925873.ogg"},
	{0xA0DF333DCFC34A23ULL, "385562055.ogg"},
	{0xA0E01A173ECCEAEEULL, "qun131_taskmaster6039780.cl"},
	{0xA0E1609E6732DA25ULL, "6118776_f.fxe"},
	{0xA0E17EF6FE6D76D9ULL, "112389270.ogg"},
	{0xA0E1C6B8FD22B1A6ULL, "mind_distortion.mao"},
	{0xA0E1F5633C23CE96ULL, "6112675_m.fxe"},
	{0xA0E36CD20D0D35CDULL, "gen_im_wep_war_one_act1_03.uti"},
	{0xA0E455DE440D0004ULL, "198824149.ogg"},
	{0xA0E5C6192888F043ULL, "390535237.ogg"},
	{0xA0E5EB3BE31A82D6ULL, "isa211_isabela6070576.cl"},
	{0xA0E6EAB542C1908DULL, "6108603_f.fxe"},
	{0xA0E7A3DF64BF338FULL, "vxw_bonepit_26q_0_0.vlm"},
	{0xA0E7AF9B8BEB4BBEULL, "34538355.ogg"},
	{0xA0E7EBFEA5033FDBULL, "vxlt_0_lowtown_docks_n_c4w_0_0.vlm"},
	{0xA0E7EDC710820F25ULL, "6097625_m.fxe"},
	{0xA0E7FFB57DD530EAULL, "vxlt_gallowscourtyard_7ir_0_0.vlm"},
	{0xA0E84D032AD80FF2ULL, "6144948_m.fxe"},
	{0xA0E84F74C7057747ULL, "l_dae311ar_gamlens_house_nt.bnk"},
	{0xA0E8A5E91C3B6A73ULL, "886210341.ogg"},
	{0xA0E8F247C6A5ED13ULL, "hf_one_charade.mor"},
	{0xA0E9B959134795ABULL, "sky_lt_undercityencounter.mao"},
	{0xA0E9BBFDF76E48BDULL, "6052107_m.fxe"},
	{0xA0E9F1345E40170AULL, "vxht_mansion_n_60j_0_0.vlm"},
	{0xA0E9FDDECDA5C09FULL, "hlac_lt_0_lowtown_a3_n_c_-196609_2z.dds"},
	{0xA0EA1A0AA50FEAB3ULL, "vxdr_primevaltomb_39w_0_0.vlm"},
	{0xA0EA6E232DFA989CULL, "6096220_f.fxe"},
	{0xA0EB646EDBFE0934ULL, "follower_banter6136739.cl"},
	{0xA0EB9E6278EBF0F9ULL, "art\\characters\\creatures\\hurlock\\animation\\blendtree\\hrlk_cb.anb"},
	{0xA0EBA645438E285CULL, "vxht_0_hightown_as5_0_0.vlm"},
	{0xA0EBD20F56D0C04FULL, "art\\characters\\creatures\\shade\\textures\\c_shade_0s.dds"},
	{0xA0EC31ED93051095ULL, "hla_ht_0_hightown_n_131072_0z.dds"},
	{0xA0ED6E3E5D80EB06ULL, "vxdr_0_deeproads_cut_0_0.vlm"},
	{0xA0EE2921A1843F04ULL, "nat380_sibling6068029.cl"},
	{0xA0EE3513F98C305FULL, "6139862_m.fxe"},
	{0xA0EEE1CA6A56B2FCULL, "sib111_mother6098831.cl"},
	{0xA0EF22F2803B4963ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em01.mmh"},
	{0xA0EF30F2803B60BDULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em01.msh"},
	{0xA0EF80792DE876D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_wipe_eyes.ani"},
	{0xA0EFFA9AEEF8422DULL, "gen00fl_isabela6079696.cl"},
	{0xA0F015052C8D3235ULL, "gen_im_ammo_quiver_04.uti"},
	{0xA0F06AD4CB9A3EB1ULL, "gen_im_arm_cht_elf_comb_t1.uti"},
	{0xA0F07249C5C7EFBFULL, "qun211_braggart6075484.cl"},
	{0xA0F075C47266E4FEULL, "vxdr_0_deeproads_cmd_0_0.vlm"},
	{0xA0F076980DD8714AULL, "757917598.ogg"},
	{0xA0F0CE61C086A4BCULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0co_galow.mao"},
	{0xA0F0DA61C086B95FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0co_galow.mmh"},
	{0xA0F0E061C086C301ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0co_galow.msh"},
	{0xA0F13055555AA434ULL, "665600278.ogg"},
	{0xA0F1DBA78C1BAA42ULL, "and230cr_templar.utc"},
	{0xA0F218AD931B7BC3ULL, "6027013_m.fxe"},
	{0xA0F26F814C4F8815ULL, "vxlt_0_lowtown_a2_n_c_cmm_0_0.vlm"},
	{0xA0F27014D3EFB3D8ULL, "vxht_0_hightown_a2_bp7_0_0.vlm"},
	{0xA0F345F641A300BCULL, "vxca_caverns2_8mm_0_0.vlm"},
	{0xA0F349470B7CD175ULL, "vxlt_0_lowtown_a3_n_c_gpc_0_0.vlm"},
	{0xA0F382718927CC0BULL, "dae000_arishok6067758.cl"},
	{0xA0F3A20FDA4A37B1ULL, "6051421_m.fxe"},
	{0xA0F3A8387DC4A08DULL, "936606999.ogg"},
	{0xA0F3B24DC3E26F11ULL, "6059919_m.fxe"},
	{0xA0F3DE1272949260ULL, "llac_ht_brothel_n_2995_0.dds"},
	{0xA0F3E7C14CF247F7ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb103a_0n.dds"},
	{0xA0F3EE1430BE7ECAULL, "6034812_m.fxe"},
	{0xA0F4055BF17DF400ULL, "442262350.ogg"},
	{0xA0F415CEA194E987ULL, "6166826_m.fxe"},
	{0xA0F41D830F1D1E41ULL, "6076624_m.fxe"},
	{0xA0F427F47970AB37ULL, "w_bonepitmining_a2_183_0.phy"},
	{0xA0F4472F2BC61990ULL, "361542838.ogg"},
	{0xA0F47B4E8E04526CULL, "641246348.ogg"},
	{0xA0F49EE498CAC461ULL, "vxht_mansion_n_5aa_0_0.vlm"},
	{0xA0F4ADA59E032FF9ULL, "gen00fl_aveline6139117.cl"},
	{0xA0F58B792DECF573ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_wipe_eyes.gad"},
	{0xA0F5B115CE0E0971ULL, "mag311_keran6041501.cl"},
	{0xA0F5E18E8C317E5AULL, "mag100_cullen6140951.cl"},
	{0xA0F5EF2F01981CE8ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdblc.mmh"},
	{0xA0F5FB2F0198308BULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdblc.mao"},
	{0xA0F60D2F01984F12ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdblc.msh"},
	{0xA0F60FB808EFAEA3ULL, "gen00fl_carver6097230.cl"},
	{0xA0F615BC56B54C0FULL, "vxw_bonepit_21p_0_0.vlm"},
	{0xA0F647522C8FD4E3ULL, "134860724.ogg"},
	{0xA0F679EF66068ACCULL, "6048804_m.fxe"},
	{0xA0F6A5D50B3A42D5ULL, "6123359_m.fxe"},
	{0xA0F6A872A2164569ULL, "art\\characters\\creatures\\desiredemon\\animation\\wk_p_rf.ani"},
	{0xA0F733427BF282A1ULL, "prm000im_lost_tome.uti"},
	{0xA0F7A1D303F676EBULL, "cd_vrc221_bartrand.bnk"},
	{0xA0F7ABBE2AF633E1ULL, "6109635_f.fxe"},
	{0xA0F7AF8CD1313456ULL, "6206046_m.fxe"},
	{0xA0F802A48BD1CDC3ULL, "gen_im_trs_gen_sta_nat_r2.uti"},
	{0xA0F88117969927D3ULL, "vxca_caverns2_7nl_0_0.vlm"},
	{0xA0F943EA870F7EDAULL, "923700748.ogg"},
	{0xA0F956AE4D4BC077ULL, "llac_dr_0_deeproads_5840_4.dds"},
	{0xA0FA2F6DE0C902E5ULL, "isa311_isabela6092307.cl"},
	{0xA0FA731151A80EF8ULL, "dae000_mother6099174.cl"},
	{0xA0FA7C7BC8F55AEBULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.dura.evt"},
	{0xA0FADDF459F68689ULL, "vxlt_gallowsprison_n_64j_0_0.vlm"},
	{0xA0FB0B9107A380D2ULL, "vxlt_0_lowtown_n_bxv_0_0.vlm"},
	{0xA0FB61947FA5B946ULL, "6080745_m.fxe"},
	{0xA0FB6431D27E3253ULL, "en_dust_f.mao"},
	{0xA0FBDA5575215291ULL, "w_bonepitmining_a2_9_0_lc.mmh"},
	{0xA0FBE86F5893F28BULL, "hlac_lt_0_lowtown_a2_n_-65536_7z.dds"},
	{0xA0FC003F136DB86DULL, "6139086_m.fxe"},
	{0xA0FC88792DF30ED8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_wipe_eyes.evt"},
	{0xA0FCD0423F24A640ULL, "hlac_ht_bartrand_n_3372_0x.dds"},
	{0xA0FCD2A6B1D752A5ULL, "mag311_nyssa6051213.cl"},
	{0xA0FD1114B26EA85AULL, "vxw_woundedcoast_4k_0_0.vlm"},
	{0xA0FD9A09EC10F640ULL, "844836935.ogg"},
	{0xA0FDC372A21C91EBULL, "art\\characters\\creatures\\desiredemon\\animation\\wk_p_rf.gad"},
	{0xA0FE279A5A1F7048ULL, "vxht_mansion_n_5ad_0_0.vlm"},
	{0xA0FE776EE04A0353ULL, "hlac_dr_dungeon_3215_4z.dds"},
	{0xA0FE790AA907DBC9ULL, "998130476.ogg"},
	{0xA0FEA12569BA3B26ULL, "vxlt_gallowscourtyard_4v4_0_0.vlm"},
	{0xA0FEE7A1645CFD7EULL, "vxdr_0_deeproads_cje_0_0.vlm"},
	{0xA0FEF6197009B02AULL, "vxlt_undercityserialkiller_1j5_0_0.vlm"},
	{0xA0FF440E48E0EBDAULL, "w_bonepit_88_0.mmh"},
	{0xA0FF480E48E0F211ULL, "w_bonepit_88_0.mao"},
	{0xA0FF55A9FB2C1E7CULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_005a_0n.dds"},
	{0xA0FF560E48E10A20ULL, "w_bonepit_88_0.msh"},
	{0xA0FFB051AEF88205ULL, "vxlt_darktown_52x_0_0.vlm"},
	{0xA0FFD84D2FC7E95AULL, "vxlt_0_lowtown_docks_a3_cx2_0_0.vlm"},
	{0xA10023F28049BD88ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em01.phy"},
	{0xA1002E408710ADDAULL, "vxw_bonepit_14q_0_0.vlm"},
	{0xA100957BC8F9F0ACULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.dura.gad"},
	{0xA1009D5D20EDF218ULL, "vxlt_foundry_n_y3_0_0.vlm"},
	{0xA100E1F1C5E5CC58ULL, "vxht_0_hightown_a2_bu4_0_0.vlm"},
	{0xA1010867989EC03AULL, "6094973_m.fxe"},
	{0xA1012B8A7A465298ULL, "hla_ht_0_hightown_a2_0_0x.dds"},
	{0xA101AE7D873D4CBCULL, "vxht_0_hightown_bgf_0_0.vlm"},
	{0xA101D678A8BB36D7ULL, "dae000_dalish_apprentices6147565.cl"},
	{0xA101F9979AEA71EFULL, "hlac_w_blightlands_327680_0z.dds"},
	{0xA10201CD57EDC652ULL, "vxca_caverns2_7ni_0_0.vlm"},
	{0xA102EA4A4B283A03ULL, "1013556114.ogg"},
	{0xA1030643A7053BBAULL, "gen00fl_varric6033633.cl"},
	{0xA1032DF222C2C857ULL, "vxca_cavern1_haunted_gkm_0_0.vlm"},
	{0xA103683B6D3235EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedflurrish.evt"},
	{0xA103F488A22124D1ULL, "zz_followers_banter6158820.cl"},
	{0xA103FDCDC02132B6ULL, "vxca_cavern1_haunted_ckl_0_0.vlm"},
	{0xA104090DC284F27FULL, "vxw_bonepit_254_0_0.vlm"},
	{0xA10456CA17865DA9ULL, "vxw_bonepitdestroyed_a3_1a5_0_0.vlm"},
	{0xA104C072A222AB50ULL, "art\\characters\\creatures\\desiredemon\\animation\\wk_p_rf.evt"},
	{0xA104EBCC57FCA34CULL, "6043904_m.fxe"},
	{0xA1058F237E6706B0ULL, "vxht_brothel_n_6b5_0_0.vlm"},
	{0xA105B6B00815010EULL, "vxlt_hangedman_n_14o_0_0.vlm"},
	{0xA1064610848C9473ULL, "308131450.ogg"},
	{0xA106722160144904ULL, "vxw_bonepit_1pk_0_0.vlm"},
	{0xA107197F28E5B7CFULL, "arc_mit_03.mao"},
	{0xA1074EFE9C7CB6FCULL, "407499038.ogg"},
	{0xA107947BC9000DAAULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.dura.ani"},
	{0xA107E3C8997A5F66ULL, "hlac_lt_gallowstemplar_n_2384_2x.dds"},
	{0xA107E831ECFE4E8BULL, "6088303_f.fxe"},
	{0xA1080166909E6E20ULL, "vxht_0_hightown_n_c49_0_0.vlm"},
	{0xA1080E6D23815959ULL, "926673310.ogg"},
	{0xA10826C181CB4ABBULL, "gen00fl_aveline6121143.cl"},
	{0xA108BDC710F3C258ULL, "932387741.ogg"},
	{0xA109129FCDFAB26FULL, "vxdr_0_deeproads_evz_0_0.vlm"},
	{0xA1092888E5A2E8E4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_3windows.phy"},
	{0xA1095F3B6D369199ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedflurrish.gad"},
	{0xA10979F23B791188ULL, "vxdr_primevalentrance_3bu_0_0.vlm"},
	{0xA10A4DA8A98EAF18ULL, "374791320.ogg"},
	{0xA10B10371B4092B7ULL, "zz_followers_banter6158880.cl"},
	{0xA10CC8BC9B9C321BULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_15ring.msh"},
	{0xA10CDABC9B9C5166ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_15ring.mao"},
	{0xA10CE6BC9B9C6505ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_15ring.mmh"},
	{0xA10D2D540988C41DULL, "hlac_ht_keep_n_4519_15z.dds"},
	{0xA10D67F661FFC8AAULL, "vxlt_undercityserialkiller_1o6_0_0.vlm"},
	{0xA10D82431BCCB242ULL, "107954792.ogg"},
	{0xA10E29E992E73233ULL, "345302662.ogg"},
	{0xA10E3783DB75ACA3ULL, "vxht_keep_60l_0_0.vlm"},
	{0xA10E8F344B9C1C07ULL, "262285325.ogg"},
	{0xA10EC3D9046AEB61ULL, "vxlt_hangedman_1ra_0_0.vlm"},
	{0xA10EC4D123F93EC6ULL, "mag170_thrask6171446.cl"},
	{0xA10F01C96F32C66AULL, "vxw_bonepitdestroyed_a3_1pu_0_0.vlm"},
	{0xA10F0F3A12E40A98ULL, "vxlt_foundry_n_t0_0_0.vlm"},
	{0xA10F2B93A87B4E79ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_ex.ani"},
	{0xA10F5F0470918BFDULL, "gen00fl_dog6167682.cl"},
	{0xA10F799F93351E6AULL, "prm000im_ashes_amulet.uti"},
	{0xA110703B6D3CCDE7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedflurrish.ani"},
	{0xA110D1CC19586848ULL, "6114154_m.fxe"},
	{0xA110FA754C2155E8ULL, "amb_scene_guardmarch.bnk"},
	{0xA11198D69DE2C0FEULL, "vxlt_0_lowtown_docks_a3_d3e_0_1.vlm"},
	{0xA111E1E078A77D4EULL, "499624527.ogg"},
	{0xA11277BFC7DCD2CCULL, "gen00fl_merrill6130127.cl"},
	{0xA1127AEAB47B0AFFULL, "vxw_bonepit_207_0_0.vlm"},
	{0xA1129A29E58CD610ULL, "6195849_m.fxe"},
	{0xA1136F89E3295B76ULL, "gen00fl_aveline6132658.cl"},
	{0xA113B657A0E81D6FULL, "vxlt_warehouse_189_0_0.vlm"},
	{0xA113E6CE3C47C687ULL, "6041192_m.fxe"},
	{0xA1144574C72AA17DULL, "l_dae311ar_gamlens_house_nt.wbd"},
	{0xA1144838A09CFD72ULL, "follower_banter6169984.cl"},
	{0xA114E3FE520A6184ULL, "vxw_bonepit_1uh_0_0.vlm"},
	{0xA114E787EAB4AA4EULL, "6032201_m.fxe"},
	{0xA115B1B24808DBD9ULL, "6117944_f.fxe"},
	{0xA115F65664BE11DDULL, "vxlt_darktown_6jo_0_0.vlm"},
	{0xA11605088D92660FULL, "hlac_lt_0_lowtown_docks_n_65533_4x.dds"},
	{0xA1160693A8812DBBULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_ex.gad"},
	{0xA11617CF5EB0DCFCULL, "vxht_0_hightown_n_caj_0_0.vlm"},
	{0xA116D86416BDE0BBULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedright_enter.evt"},
	{0xA116EC6E5762F910ULL, "778952925.ogg"},
	{0xA11710FC0E4352B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_elemental_mark.ani"},
	{0xA11719686F59BBB0ULL, "vxw_blightlands_98_0_0.vlm"},
	{0xA1179585C92E4597ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrotha.phy"},
	{0xA117A13206537E70ULL, "vxlt_0_lowtown_a3_n_c_epa_0_1.vlm"},
	{0xA117CDBDDB80A8DEULL, "vxlt_gallowscourtyard_n_4v7_0_0.vlm"},
	{0xA117DD39262B775AULL, "ca_caverns1_mines_26037.rml"},
	{0xA117E4B7A1731265ULL, "614542144.ogg"},
	{0xA117FAE5DA926106ULL, "6140579_m.fxe"},
	{0xA117FB9C0E7DEC73ULL, "393067832.ogg"},
	{0xA11816B60D1659B5ULL, "lla_ht_0_hightown_a3_65537_0.dds"},
	{0xA11938EBF41E5DA5ULL, "6064850_m.fxe"},
	{0xA1198A1FF24D4665ULL, "6115127_m.fxe"},
	{0xA119D68A1CDDC098ULL, "hlac_w_encounter1_-65532_0y.dds"},
	{0xA119FA619603209BULL, "6077355_m.fxe"},
	{0xA11A2074626D0356ULL, "hlac_w_bonepit_0_11z.dds"},
	{0xA11A2FB5A0F08F76ULL, "vxlt_0_lowtown_f0i_0_0.vlm"},
	{0xA11A419A4C50845CULL, "vxht_0_hightown_a2_ajz_0_0.vlm"},
	{0xA11A46DE6F26B7E6ULL, "vxht_0_hightown_n_c6i_0_0.vlm"},
	{0xA11A9A1C6D428EBDULL, "vxht_brothel_n_4fb_0_0.vlm"},
	{0xA11AC9121333D8B1ULL, "vxht_keep_n_5u1_0_0.vlm"},
	{0xA11B00317CB7FBCFULL, "plt_lgt323pt_shw_ors_blckop.nss"},
	{0xA11B0F5F41E86A4FULL, "vxlt_0_lowtown_docks_a3_n_cyf_0_1.vlm"},
	{0xA11B1F4AE6E98877ULL, "241078858.ogg"},
	{0xA11B3B9ED3651135ULL, "6129357_m.fxe"},
	{0xA11B7B8225D67B1AULL, "vxlt_0_lowtown_docks_cq6_0_0.vlm"},
	{0xA11BB0ACA7FED0D5ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em02.phy"},
	{0xA11BE8F7538ED7EBULL, "zz_dae_debug6029333.cl"},
	{0xA11C940B92E4B305ULL, "vxlt_darktown_58f_0_0.vlm"},
	{0xA11CD16416C240FCULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedright_enter.gad"},
	{0xA11D2393A8877DA0ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_ex.evt"},
	{0xA11DC5ABA9D1FD58ULL, "vxht_0_hightown_a2_bc2_0_0.vlm"},
	{0xA11DE14EAC9923F2ULL, "6030395_m.fxe"},
	{0xA11DEDB4C9B6A9AFULL, "gen00fl_aveline6105440.cl"},
	{0xA11DF0AC3A45AC67ULL, "lt_0_lowtown_docks_a3_27_0.anb"},
	{0xA11E31BF072526F5ULL, "708422118.ogg"},
	{0xA11E39D16A8F4AA4ULL, "6103693_m.fxe"},
	{0xA11E4566E451BFA1ULL, "6149306_m.fxe"},
	{0xA11E8B2837E44E25ULL, "sib131cr_slaver.utc"},
	{0xA11EC3D68315A897ULL, "6068006.cut"},
	{0xA11ED8ACC734B7D5ULL, "hlac_ht_playershouse_n_855_15z.dds"},
	{0xA11F14105A27688EULL, "zz_vault_debug6200353.cl"},
	{0xA11FA338FF8B4D8CULL, "vxht_mansion_n_4db_0_0.vlm"},
	{0xA11FA84470512B91ULL, "vxlt_0_lowtown_a3_n_c_hdd_0_0.vlm"},
	{0xA11FB1F57413D5E5ULL, "396818419.ogg"},
	{0xA1203B1FE5213A76ULL, "hlac_w_woundedcoast_0_1z.dds"},
	{0xA121278182BA5720ULL, "follower_banter6155574.cl"},
	{0xA12133A97B809FF7ULL, "w_bonepitmining_a2_92_0.msh"},
	{0xA12145A97B80BE5AULL, "w_bonepitmining_a2_92_0.mao"},
	{0xA12149A97B80C569ULL, "w_bonepitmining_a2_92_0.mmh"},
	{0xA1215FCE96317FCDULL, "358368723.ogg"},
	{0xA121D53550BE1723ULL, "mcr310_sibling6076322.cl"},
	{0xA1221B9A78840A49ULL, "lgt120_orz_bln_off_messenge6140221.cl"},
	{0xA1221EFC7EFC3E88ULL, "vxw_encounter3_9d_0_0.vlm"},
	{0xA12220909D4DA3E4ULL, "hlac_lt_undercitydungeon_-1_29x.dds"},
	{0xA1223AEC56496367ULL, "219630738.ogg"},
	{0xA12272DD625337B0ULL, "vxht_brothel_n_6h3_0_0.vlm"},
	{0xA122889D3E5B3F8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_enter.ani"},
	{0xA122AA8F89307CB7ULL, "dae000_player_generic_find6179062.cl"},
	{0xA122BC408E40A367ULL, "vxht_brothel_n_6k5_0_0.vlm"},
	{0xA122C094A3AC33D7ULL, "hlac_ca_caverns1_mines_24423_2z.dds"},
	{0xA122CEBD40B2A638ULL, "dae000_ineria6137239.cl"},
	{0xA122E9742F68D09CULL, "vxw_bonepitmining_a2_cw_0_0.vlm"},
	{0xA1232FCE14A3A248ULL, "487446672.ogg"},
	{0xA123D06416C85DFAULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\humangrabbedright_enter.ani"},
	{0xA124334478532A0EULL, "530978631.ogg"},
	{0xA124E8C1412A63BDULL, "289964924.ogg"},
	{0xA1250BC9EBE7EA3DULL, "vxlt_0_lowtown_a2_n_eov_0_1.vlm"},
	{0xA12518FC0E4F9DA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_elemental_mark.evt"},
	{0xA1267F72F91F72E1ULL, "565437294.ogg"},
	{0xA126BE6BB7B5AFC8ULL, "vxlt_hangedman_1ye_0_0.vlm"},
	{0xA12740B94AA6C53CULL, "vxlt_gallowstemplar_5by_0_0.vlm"},
	{0xA12757BAF7BF9619ULL, "llac_w_sundermount_131074_0.dds"},
	{0xA127785EF96B4C5AULL, "117075670.ogg"},
	{0xA127AAC8E6147625ULL, "6063950_m.fxe"},
	{0xA127DC73C58F96C7ULL, "vxht_mansion2_381_0_0.vlm"},
	{0xA128A19292E6A7F6ULL, "vxlt_0_lowtown_f5j_0_0.vlm"},
	{0xA128B218682F9A71ULL, "444642171.ogg"},
	{0xA128B8BB611CD066ULL, "vxht_0_hightown_n_c3j_0_0.vlm"},
	{0xA128CD328549F516ULL, "vxht_0_hightown_a3_c7h_0_0.vlm"},
	{0xA12917A57BDD700DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_basic.mmh"},
	{0xA1291BA57BDD775EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_basic.mao"},
	{0xA12922361D84E645ULL, "vxlt_0_lowtown_cmt_0_0.vlm"},
	{0xA12929A57BDD8E93ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_basic.msh"},
	{0xA1298AA42532C411ULL, "hlac_lt_gallowstemplar_2383_9x.dds"},
	{0xA1298ED75BFC0D2BULL, "282522776.ogg"},
	{0xA129959D3E6174CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_enter.gad"},
	{0xA129B93C89410CCEULL, "mrl000_marethari6129274.cl"},
	{0xA129C3B2CE3D9047ULL, "vxw_encounter3_g9_0_0.vlm"},
	{0xA12A34B793064290ULL, "lgt320_shw_ors_marlein6156379.cl"},
	{0xA12A48C03657BBEBULL, "qun210ip_loot_a.utp"},
	{0xA12A826911F54A27ULL, "flare3.dds"},
	{0xA12ACE860A198E37ULL, "6150265_m.fxe"},
	{0xA12ADD8F3C8812CBULL, "lav_river00a.dds"},
	{0xA12AFD647B502701ULL, "vxca_caverns2_9df_0_0.vlm"},
	{0xA12B2FB1A067EE7DULL, "art\\levels\\ztools\\default\\generic\\1pp_basic.mao"},
	{0xA12B33B1A067F4B6ULL, "art\\levels\\ztools\\default\\generic\\1pp_basic.mmh"},
	{0xA12B39FC0E5441F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_elemental_mark.gad"},
	{0xA12B3DB1A06805B4ULL, "art\\levels\\ztools\\default\\generic\\1pp_basic.msh"},
	{0xA12CABAE5C856324ULL, "6104793_m.fxe"},
	{0xA12D1D6EA6EA1BC8ULL, "vxlt_gallowsprison_n_5sa_0_0.vlm"},
	{0xA12D4C02300E3E80ULL, "vxdr_primevaltomb_73n_0_0.vlm"},
	{0xA12D4ECCBDFCD86AULL, "vxca_cavern1_haunted_i2y_0_0.vlm"},
	{0xA12D695BB3B3C3A4ULL, "6015089_f.fxe"},
	{0xA12D8A4E6AA173A9ULL, "vxdr_primevaltomb_8f7_0_0.vlm"},
	{0xA12DC1722F21418FULL, "gen00fl_isabela6099461.cl"},
	{0xA12E2D556D6BE519ULL, "vxlt_backalley_n_1wm_0_1.vlm"},
	{0xA12EC63E8C8DE0FBULL, "15867686.ogg"},
	{0xA12F2FA12577809DULL, "vxlt_gallowstemplar_f_5pi_0_0.vlm"},
	{0xA12FD802CC299806ULL, "790396794.ogg"},
	{0xA130909D3E678AADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_enter.evt"},
	{0xA13090D970F25708ULL, "vxw_encounter3_2e_0_0.vlm"},
	{0xA130AF45C3837A58ULL, "llac_lt_0_lowtown_n_65534_5.dds"},
	{0xA130E60A2284AFC6ULL, "dae000_arishok6067652.cl"},
	{0xA130F52EACF76BE0ULL, "art\\levels\\4cinematics\\qunari\\_textures\\prp_qunarirelic_d.dds"},
	{0xA1313A6282AC2723ULL, "vxlt_gallowstemplar_n_4uy_0_0.vlm"},
	{0xA1315B51215EE91CULL, "vxw_bonepitmining_a2_nt_0_0.vlm"},
	{0xA13186A623F8E535ULL, "722461560.ogg"},
	{0xA131885180B28BA0ULL, "6066431_m.fxe"},
	{0xA13188FD17768EE0ULL, "381372011.ogg"},
	{0xA131EB44DE18550FULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.mmh"},
	{0xA1329A4D5F8DB6D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_exit.ani"},
	{0xA13418CC6A02331BULL, "264524059.ogg"},
	{0xA1344BAC43E28862ULL, "vxlt_gallowstemplar_n_375_0_0.vlm"},
	{0xA13488A3DCC73C2AULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_csand02_da.dds"},
	{0xA134C789A638E0E8ULL, "6080691_m.fxe"},
	{0xA134CBAAE71012DCULL, "w_encounter1_48_0.mao"},
	{0xA134D7AAE710277FULL, "w_encounter1_48_0.mmh"},
	{0xA134DDAAE7103121ULL, "w_encounter1_48_0.msh"},
	{0xA1353048A9ABC848ULL, "vxlt_hangedman_1db_0_0.vlm"},
	{0xA1354C8A1B6C4365ULL, "vxlt_0_lowtown_a2_n_c_fed_0_0.vlm"},
	{0xA13564369BCE0931ULL, "614635248.ogg"},
	{0xA135DF9E7DC3109FULL, "art\\cinematics\\additional animation\\cs_malenoblei_cinematicanim.gad"},
	{0xA135EC754C405992ULL, "amb_scene_guardmarch.wbd"},
	{0xA1362012B3DBBD72ULL, "dae000cr_grand_cleric6124738.cl"},
	{0xA1367BC2147DA03EULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformbladestab.evt"},
	{0xA136B6F4CF0D7E03ULL, "6078296_f.fxe"},
	{0xA136C6DA1FEC7D6AULL, "107579199.ogg"},
	{0xA1372A985312E8E6ULL, "vxht_0_hightown_n_c8k_0_0.vlm"},
	{0xA137B3D786268068ULL, "vxlt_0_lowtown_docks_bdw_0_1.vlm"},
	{0xA137CC666C0DCDEAULL, "47884322.ogg"},
	{0xA13801CD2C4C3B81ULL, "vxca_cavern1_haunted_dpu_0_0.vlm"},
	{0xA13831DE3864BBF3ULL, "vxw_woundedcoast_1eh_0_0.vlm"},
	{0xA138321B422CC801ULL, "6152158_f.fxe"},
	{0xA13853CDAA91C54DULL, "gen00flst_isa_init_1.stg"},
	{0xA13928F67529881BULL, "6040563_f.fxe"},
	{0xA139649A83146F77ULL, "6108515_f.fxe"},
	{0xA13981FF89906BC6ULL, "6042970_m.fxe"},
	{0xA13A10A57BEBD7CAULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_basic.phy"},
	{0xA13AE686DADBDDDCULL, "8633967.ogg"},
	{0xA13AFCF479AC648FULL, "w_bonepitmining_a2_183_0.mao"},
	{0xA13B07D5064FA028ULL, "10821048.ogg"},
	{0xA13B08F479AC792CULL, "w_bonepitmining_a2_183_0.mmh"},
	{0xA13B0EF32081CDE5ULL, "mrl150_merrill6078331.cl"},
	{0xA13B0EF479AC831EULL, "w_bonepitmining_a2_183_0.msh"},
	{0xA13BE202920CBE10ULL, "vxht_keep_n_613_0_0.vlm"},
	{0xA13C369EB94CACBBULL, "1009871326.ogg"},
	{0xA13C9D8B22347A19ULL, "6088248_f.fxe"},
	{0xA13D3F4E6E850F7EULL, "6172307_m.fxe"},
	{0xA13D92C21483E669ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformbladestab.gad"},
	{0xA13E244706F0C6A5ULL, "ser221_party_comments6209892.cl"},
	{0xA13E2FE06D01E92CULL, "ts_immortal.ncs"},
	{0xA13ECF850D4CB514ULL, "6095427_f.fxe"},
	{0xA13EF3DD9D50002FULL, "252433590.ogg"},
	{0xA13F51012006D4A4ULL, "art\\cinematics\\props\\letter.mmh"},
	{0xA13F67012006FA46ULL, "art\\cinematics\\props\\letter.msh"},
	{0xA13F924D5F98330AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_exit.evt"},
	{0xA13FA846C159D673ULL, "6082183_f.fxe"},
	{0xA13FEC05B29BE907ULL, "6083958_m.fxe"},
	{0xA1404B7799ECDE38ULL, "vxht_chantry_2v9_0_0.vlm"},
	{0xA14090B6417F65B2ULL, "ave121_brennan6105904.cl"},
	{0xA141B876510D2A10ULL, "art\\vfx\\materials\\chunky\\vfx_c_lightbeam_dust.mao"},
	{0xA1421891DECCAABBULL, "vxlt_undercitydungeon_1c7_0_1.vlm"},
	{0xA14237E6082B3C61ULL, "6121199_m.fxe"},
	{0xA14249CC2EBFBD6BULL, "store_dae000cr_hubert.utm"},
	{0xA142F7F7CDB24B3DULL, "one220_yevhen6103294.cl"},
	{0xA142FB62B3E83806ULL, "6048961_m.fxe"},
	{0xA142FB85284F369EULL, "vxht_0_hightown_a2_c8q_0_1.vlm"},
	{0xA14302E1984DBE88ULL, "zz_one_debug6099759.cl"},
	{0xA1430FAD578BC768ULL, "qun231_party_comment6182828.cl"},
	{0xA14359AF3C7515C7ULL, "vxlt_gallowscourtyard_5a5_0_0.vlm"},
	{0xA143AF2503169BE4ULL, "130612276.ogg"},
	{0xA143ECC74EEE965DULL, "mag160st_feynriel_cave.stg"},
	{0xA1440D940C14100FULL, "vfx_steamdrain.wbd"},
	{0xA14438490B75E073ULL, "995965608.ogg"},
	{0xA14463C21489B4B7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\transformbladestab.ani"},
	{0xA144751B114B507BULL, "hlac_w_bonepitmining_a2_327685_0x.dds"},
	{0xA1448276CF0934F8ULL, "art\\vfx\\materials\\chunky\\vfx_c_lightbeam_falloff_add.mao"},
	{0xA144C02DA97BC7C7ULL, "vxht_mansion2_323_0_0.vlm"},
	{0xA144E72A6DEB138AULL, "955291427.ogg"},
	{0xA145044A846A6FD4ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.dura.ani"},
	{0xA14535A6759BE62CULL, "6166281_m.fxe"},
	{0xA1457D6CF9BDA8C8ULL, "zz_followers_banter6158175.cl"},
	{0xA1466D3FDD3B016BULL, "70917191.ogg"},
	{0xA146714D5F9E1A7DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_exit.gad"},
	{0xA146BC17B5F592DBULL, "254275831.ogg"},
	{0xA1476F9F8D868016ULL, "554680284.ogg"},
	{0xA147ACBA987F6DFEULL, "128439908.ogg"},
	{0xA147D72DF48FA5B6ULL, "6180382_m.fxe"},
	{0xA148800651F7587CULL, "252747612.ogg"},
	{0xA148A0A186372F11ULL, "6088351_m.fxe"},
	{0xA148C93D370B286AULL, "vxw_bonepitdestroyed_a3_1dq_0_0.vlm"},
	{0xA148FD73F81D7000ULL, "6087733_m.fxe"},
	{0xA1496BD61EBFE756ULL, "ttx_steps_curve_b_1.msh"},
	{0xA14975D61EBFF854ULL, "ttx_steps_curve_b_1.mmh"},
	{0xA14A85FC85AD4C28ULL, "22747644.ogg"},
	{0xA14ABC9E7DD531E5ULL, "art\\cinematics\\additional animation\\cs_malenoblei_cinematicanim.ani"},
	{0xA14ACE76FFEA1C2BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1ww_basic_1.mao"},
	{0xA14AF51E6EDDB3F3ULL, "character_character_p163.psh.fxo.dx11"},
	{0xA14AFDDB46337511ULL, "vxlt_0_lowtown_a3_n_c_hky_0_0.vlm"},
	{0xA14B281607EED193ULL, "vxw_encounter2_2a_0_0.vlm"},
	{0xA14BD59AC9182488ULL, "vxlt_0_lowtown_a2_eqg_0_0.vlm"},
	{0xA14C014A84708936ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.dura.gad"},
	{0xA14C0B52795C0331ULL, "vxdr_slavers_1b6_0_0.vlm"},
	{0xA14C6C32B07F3A50ULL, "dae000_arishok6075051.cl"},
	{0xA14C7568C524AEC0ULL, "mag100_thrask6032069.cl"},
	{0xA14CBE64F3E56FFCULL, "lt_gallowscourtyard_n_59_0.rml"},
	{0xA14DEE2A3D48D0E3ULL, "vxdr_0_deeproads_9c5_0_0.vlm"},
	{0xA14E49C4EF48328AULL, "96874187.ogg"},
	{0xA14E9BF1C8DEADFFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupa.msh"},
	{0xA14EA1F1C8DEB7A1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupa.mmh"},
	{0xA14EAB7219E2C384ULL, "vxw_bonepit_1il_0_0.vlm"},
	{0xA14EADF1C8DECC42ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupa.mao"},
	{0xA14EEA55D89CC906ULL, "vxht_chantry_2sx_0_0.vlm"},
	{0xA14F94B652454DCCULL, "421764062.ogg"},
	{0xA150229219D3AD01ULL, "688143676.ogg"},
	{0xA1509D8CD49831AEULL, "zz_fol_debug6057004.cl"},
	{0xA150CD6DDFCF5F4DULL, "764529741.ogg"},
	{0xA1514663F48EA501ULL, "6144467_f.fxe"},
	{0xA1514F4EE586E883ULL, "vxlt_0_lowtown_a2_byo_0_0.vlm"},
	{0xA151A72F86441DE1ULL, "fen210_orana6133057.cl"},
	{0xA151FD7EF6BB4B59ULL, "6077325_f.fxe"},
	{0xA15215B858E4679AULL, "llac_w_encounter1_-131072_0.dds"},
	{0xA1521C4A84752295ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.dura.evt"},
	{0xA152E344A964EE3DULL, "w_sundermount_1_0.rml"},
	{0xA152E361C0D992A4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0co_galow.phy"},
	{0xA152EDFCF9A05A15ULL, "vxlt_backalley_g_1ke_0_0.vlm"},
	{0xA153819A3287752BULL, "887745460.ogg"},
	{0xA154539781C60D5EULL, "6085749_f.fxe"},
	{0xA15498A876551477ULL, "vxlt_gallowstemplar_373_0_0.vlm"},
	{0xA154EA77BB160939ULL, "vxlt_0_lowtown_docks_a3_cv3_0_0.vlm"},
	{0xA1550F14428CA329ULL, "bink.mat.dx11"},
	{0xA15515981C50ECF3ULL, "vxw_woundedcoast_1kj_0_0.vlm"},
	{0xA1558F6ED2FD5178ULL, "hlac_lt_0_lowtown_docks_a3_-65537_5y.dds"},
	{0xA15610813BB29305ULL, "6080562_m.fxe"},
	{0xA15648546700A077ULL, "6106715_f.fxe"},
	{0xA15677116F0F5A32ULL, "follower_banter6151239.cl"},
	{0xA156EDE4BA59C1BCULL, "777799796.ogg"},
	{0xA157127E782D4791ULL, "6083251_m.fxe"},
	{0xA1571B6D246C07CAULL, "423037207.ogg"},
	{0xA15859AB3301DFBCULL, "vxht_0_hightown_byx_0_0.vlm"},
	{0xA15863F29A63A75EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_exit.evt"},
	{0xA15871B8D8E476DAULL, "vxw_bonepitmining_a2_1oz_0_0.vlm"},
	{0xA158BB32E3F3D0EBULL, "bld_drian_m.dds"},
	{0xA15952BBFEDD6C75ULL, "rdr111dg_martin6054965.cl"},
	{0xA1595B63A9492F08ULL, "321809191.ogg"},
	{0xA159EC44DE39A94CULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.anb"},
	{0xA159ED4F24291E93ULL, "vxw_blightlands_aw_0_0.vlm"},
	{0xA15A94C78AE5F90CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potplant_sa.dds"},
	{0xA15A9D867E996FDBULL, "12222950.ogg"},
	{0xA15AA0D43D4505F4ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelftop.phy"},
	{0xA15B05F080AF5E31ULL, "fen210_hadriana6110008.cl"},
	{0xA15B7C3C4DF35B55ULL, "vxlt_warehouse_n_qh_0_0.vlm"},
	{0xA15B90A257070140ULL, "llac_ca_cavern1_haunted_24421_3.dds"},
	{0xA15BED8C87D63D31ULL, "6083093_m.fxe"},
	{0xA15BF951238A297AULL, "6130114_m.fxe"},
	{0xA15C2D5F0446DD19ULL, "979457398.ogg"},
	{0xA15CA6124F5E38A1ULL, "hlac_lt_0_lowtown_a3_n_c_65533_2x.dds"},
	{0xA15CB45A0FA05B6AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave7statue.mao"},
	{0xA15CB85A0FA061B9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave7statue.mmh"},
	{0xA15CC25A0FA072A7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave7statue.msh"},
	{0xA15E6A85C969FFEFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrotha.mao"},
	{0xA15E7685C96A138CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrotha.mmh"},
	{0xA15E7C85C96A1E7EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrotha.msh"},
	{0xA15F31442A7632D1ULL, "follower_barks6153194.cl"},
	{0xA15F48B4F57D4ECFULL, "rdr111dg_martin6014673.cl"},
	{0xA15F7AF29A69EC89ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_exit.gad"},
	{0xA16068260B073B63ULL, "869944465.ogg"},
	{0xA1615FD9EB967295ULL, "vxlt_backalley_g_1pd_0_0.vlm"},
	{0xA1617B14EF5C806BULL, "6180912_m.fxe"},
	{0xA161AE97D685C378ULL, "dae000_templar_knight6165174.cl"},
	{0xA161B01EC9086DBEULL, "gen00fl_merrill6129017.cl"},
	{0xA162334B3A390FFAULL, "vxca_caverns1_mines_ea8_0_0.vlm"},
	{0xA16247BD67E8B017ULL, "vxht_0_hightown_c8a_0_0.vlm"},
	{0xA162C3ACA83AF3F0ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em02.msh"},
	{0xA162D1ACA83B0B0AULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em02.mmh"},
	{0xA162DF7435E64865ULL, "hlac_w_bonepitdestroyed_a3_0_11y.dds"},
	{0xA163013FE1BFE058ULL, "vxlt_0_lowtown_a2_hd1_0_0.vlm"},
	{0xA1635C54AD0C21B9ULL, "vxlt_0_lowtown_docks_a3_cy2_0_0.vlm"},
	{0xA1641024C757E787ULL, "mcr371_orsino6083358.cl"},
	{0xA1641F4322E54179ULL, "gen_im_cns_cft_pot_03.uti"},
	{0xA1647DD43D4D1EF1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelftop.msh"},
	{0xA1648BD43D4D360CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelftop.mao"},
	{0xA16497D43D4D4AAFULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelftop.mmh"},
	{0xA1649A17C9B50913ULL, "211521760.ogg"},
	{0xA1649C69CAE32917ULL, "670708579.ogg"},
	{0xA164DD026141C373ULL, "6032002_m.fxe"},
	{0xA164E2727151BCF5ULL, "646453100.ogg"},
	{0xA164F73A6917C2D9ULL, "mer141_jansen6031841.cl"},
	{0xA164FDA2EBA8572FULL, "6139630_f.fxe"},
	{0xA1650BF0800A7343ULL, "6110789_f.fxe"},
	{0xA1656CF4F1632EA6ULL, "vxlt_0_lowtown_a2_fh5_0_0.vlm"},
	{0xA1664BF29A6FBBD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_exit.ani"},
	{0xA166711AF626CA8CULL, "art\\characters\\creatures\\pridedemon\\animation\\rdy.ani"},
	{0xA166804D47DBDE92ULL, "509971232.ogg"},
	{0xA1668C340FC5CD00ULL, "statecnt_trigger.gda"},
	{0xA16697ABFFBC28BAULL, "hlac_lt_backalley_n_65537_2y.dds"},
	{0xA166D7D544043474ULL, "6153854_m.fxe"},
	{0xA1679C3C0A41DBB6ULL, "vxlt_0_lowtown_a2_n_c_fiw_0_0.vlm"},
	{0xA167ACD317491691ULL, "6081346_f.fxe"},
	{0xA167B8F267C2D25EULL, "gen00fl_merrill6107004.cl"},
	{0xA167F587FF8AA739ULL, "hlac_ht_mansion_n_3369_2z.dds"},
	{0xA1680BCFEBDB0293ULL, "vxw_encounter2_8c_0_0.vlm"},
	{0xA16819FA392A04C8ULL, "6047672_m.fxe"},
	{0xA1685BC828F52BACULL, "vxca_caverns1_mines_e52_0_0.vlm"},
	{0xA1685F2C161F3713ULL, "vxw_blightlands_jv_0_0.vlm"},
	{0xA168A1C4116CF476ULL, "follower_banter6136467.cl"},
	{0xA168A6B469C9641CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeama.mao"},
	{0xA168B2B469C978BFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeama.mmh"},
	{0xA168B8B469C98361ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_closebeama.msh"},
	{0xA168DD06A77EB6EFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_diningtable_d.dds"},
	{0xA16916D6354434B6ULL, "art\\characters\\creatures\\crowds\\humanmale\\prp_atlas_jimmy.mao"},
	{0xA1694D0DF958D9B5ULL, "gen00fl_dog6117681.cl"},
	{0xA1696132B0CBB566ULL, "t3_bow_elm.tnt"},
	{0xA16970E265312F74ULL, "vxlt_darktown_4j0_0_0.vlm"},
	{0xA169E81148D6D824ULL, "one100_party_comments6133501.cl"},
	{0xA169EE193FE973D5ULL, "vxlt_warehouse_n_tw_0_0.vlm"},
	{0xA169FDA2C3D3EF63ULL, "dae000_cora6162439.cl"},
	{0xA16AA8C55C540BD1ULL, "one110cr_harimann_guard_1.utc"},
	{0xA16AE25B4E042F6EULL, "vxlt_0_lowtown_byz_0_0.vlm"},
	{0xA16B0C39FD729E69ULL, "vxht_keep_n_689_0_0.vlm"},
	{0xA16B1BA4C14C144FULL, "rdr111dg_martin6057144.cl"},
	{0xA16C0AADBB124061ULL, "vxw_encounter2_camp_6o_0_0.vlm"},
	{0xA16C1E85E6B8B234ULL, "art\\vfx\\combat\\vfx_glb_ele_c_major_c.anb"},
	{0xA16C6395F326F6CAULL, "llac_ht_chantry_n_2534_15.dds"},
	{0xA16D583786C2BBA8ULL, "851421296.ogg"},
	{0xA16D7E1AF62CFFCEULL, "art\\characters\\creatures\\pridedemon\\animation\\rdy.gad"},
	{0xA16E21A99710A73CULL, "6041560_m.fxe"},
	{0xA16E246892EC5E5AULL, "hlac_w_bonepit_393216_0x.dds"},
	{0xA16E2851306A4478ULL, "gen_im_trs_rdr_junk2_03.uti"},
	{0xA16E3E190B9211E7ULL, "gen00fl_fenris6126734.cl"},
	{0xA16E41E652719648ULL, "692271470.ogg"},
	{0xA16ED805F27AA336ULL, "prp_str_slime_n.dds"},
	{0xA16F508A4F29F2D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_right.ani"},
	{0xA16F5C4358F18E6DULL, "6064700_m.fxe"},
	{0xA16F69530D30F06DULL, "dae000_pickpocket_scene6175070.cl"},
	{0xA16F78CBD3FBAA07ULL, "mer120_hubert6034796.cl"},
	{0xA16FDC42A93F478EULL, "dae000_arishok6078182.cl"},
	{0xA16FFE0EB01D822CULL, "6055073_m.fxe"},
	{0xA17047549D9DB309ULL, "100804303.ogg"},
	{0xA1705529422B8E2CULL, "w_encounter2_probe_qh.dds"},
	{0xA1714C1A25F6BC1FULL, "w_blightlands_189_0.msh"},
	{0xA17151B7605023E9ULL, "6111097_m.fxe"},
	{0xA171521A25F6C6C1ULL, "w_blightlands_189_0.mmh"},
	{0xA1715E1A25F6DB62ULL, "w_blightlands_189_0.mao"},
	{0xA1718D6AF46C71CCULL, "genip_trap_blade.utp"},
	{0xA171F952003D1DF3ULL, "vxw_woundedcoast_1ql_0_0.vlm"},
	{0xA172265AB17F9C44ULL, "gen_im_wep_war_two_act0_03.uti"},
	{0xA172B0A2D8A3D69CULL, "126524787.ogg"},
	{0xA1738B8EA500F2E7ULL, "6016448_m.fxe"},
	{0xA173DC4DDCE097B9ULL, "vxca_caverns2_7jz_0_0.vlm"},
	{0xA173FF4C6D88B318ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf01.phy"},
	{0xA17439676EBD5C68ULL, "llac_lt_0_lowtown_131071_6.dds"},
	{0xA174DA8FB04F8FFBULL, "214878892.ogg"},
	{0xA174F4690BBBF99EULL, "vxlt_0_lowtown_docks_a3_cos_0_0.vlm"},
	{0xA17587E0BE4598ACULL, "gen00fl_merrill6128733.cl"},
	{0xA17593637C66EC48ULL, "vxw_encounter2_camp_6j_0_0.vlm"},
	{0xA1771D843D4558DCULL, "plt_cod_mgc_maleficar.plo"},
	{0xA1777C2743256AB8ULL, "gen00fl_sebastian6188628.cl"},
	{0xA177C1FB807EEC1FULL, "hlac_lt_0_lowtown_a2_-1_6x.dds"},
	{0xA1782A552CFCF31BULL, "gen_im_trs_qun_two_2ha_r2.uti"},
	{0xA178BF64F40AAC4CULL, "lt_gallowscourtyard_n_59_0.anb"},
	{0xA178C4D7BA049983ULL, "blood_dot_l2.mao"},
	{0xA178F24D97F15704ULL, "163928715.ogg"},
	{0xA1794C594A5C9587ULL, "myriadpro-regular.ttf"},
	{0xA179621AACCA39E4ULL, "vxht_0_hightown_a3_apq_0_0.vlm"},
	{0xA1797E16EF68B6E9ULL, "vxht_keep_n_638_0_0.vlm"},
	{0xA179B931C131069EULL, "vxca_caverns1_e64_0_0.vlm"},
	{0xA179BF57F0D96437ULL, "6147759_m.fxe"},
	{0xA179FD85E6C4B7F7ULL, "art\\vfx\\combat\\vfx_glb_ele_c_major_c.mmh"},
	{0xA17A7C8AAD0858E1ULL, "vxw_encounter2_camp_1n_0_0.vlm"},
	{0xA17AE258B7D13DA4ULL, "gen00fl_varric6199023.cl"},
	{0xA17B06315A393ED9ULL, "459432864.ogg"},
	{0xA17B26B95DC72D4DULL, "mag110_samson6053549.cl"},
	{0xA17B3C91AD9B1C96ULL, "zz_approval_debug6003058.cl"},
	{0xA17BF952500F7123ULL, "qapath4.arl"},
	{0xA17C3EC04CD23422ULL, "dae000_gamlen6048496.cl"},
	{0xA17C588A4F348AC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_right.evt"},
	{0xA17C884111A364F3ULL, "824250172.ogg"},
	{0xA17DA21D99C39E52ULL, "431772662.ogg"},
	{0xA17DC7B39A54834CULL, "hlac_lt_0_lowtown_docks_a3_n_-4_1x.dds"},
	{0xA17E111039092B83ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn_alt.ani"},
	{0xA17E48131EEA3389ULL, "81804334.ogg"},
	{0xA17E6FEBA2139AACULL, "6052173_m.fxe"},
	{0xA17EDA66E4E7B660ULL, "vxlt_0_lowtown_docks_a3_n_d3m_0_0.vlm"},
	{0xA17F1664903FFE4BULL, "vxlt_0_lowtown_n_eou_0_0.vlm"},
	{0xA17F2B774BD4E117ULL, "vxht_0_hightown_c2c_0_0.vlm"},
	{0xA17F6CEAA2F5317FULL, "vxlt_backalley_n_1wz_0_0.vlm"},
	{0xA17F6F170688411EULL, "hlac_lt_0_lowtown_docks_a3_-65537_2z.dds"},
	{0xA17F837A9D7B5AD5ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_stn_0s.dds"},
	{0xA17FD7D4225F45D2ULL, "vxht_playershouse_n_1o9_0_0.vlm"},
	{0xA17FDED02A75433CULL, "follower_banter6172865.cl"},
	{0xA17FE6842E4D4988ULL, "vxht_0_hightown_n_b9k_0_0.vlm"},
	{0xA1801BC68F3F45C0ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\harvester_movement.anb"},
	{0xA1804BAA7F20ACB9ULL, "vxca_caverns1_mines_ics_0_2.vlm"},
	{0xA18051D85451208CULL, "hlac_lt_gallowstemplar_7198_3x.dds"},
	{0xA180973A63D98299ULL, "docks_male_elven_worker_sta.bnk"},
	{0xA180FD92472B07A1ULL, "gen00fl_carver6102343.cl"},
	{0xA18146A36CE2BFE1ULL, "979562281.ogg"},
	{0xA18173DAA823E228ULL, "mcr370_cullen.cnv"},
	{0xA181C0BC452DF473ULL, "6038602_m.fxe"},
	{0xA181EC531E4A19C6ULL, "llac_lt_0_lowtown_n_65535_2.dds"},
	{0xA1831DCF4EE1F757ULL, "lt_0_lowtown_docks_140_0.mmh"},
	{0xA18321CF4EE1FE64ULL, "lt_0_lowtown_docks_140_0.mao"},
	{0xA18333CF4EE21CC9ULL, "lt_0_lowtown_docks_140_0.msh"},
	{0xA183798A4F3AE110ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_right.gad"},
	{0xA183B93A63DBF2F9ULL, "docks_male_elven_worker_sta.cnv"},
	{0xA183D98C3D4CF83EULL, "vxca_caverns1_mines_iap_0_0.vlm"},
	{0xA18405406E5D04C8ULL, "vxw_encounter2_camp_1k_0_0.vlm"},
	{0xA18427940C4A20E5ULL, "vfx_steamdrain.bnk"},
	{0xA1846C4AB8CF9465ULL, "vxlt_backalley_g_1k5_0_0.vlm"},
	{0xA18495DAA8265190ULL, "mcr370_cullen.bnk"},
	{0xA18542E5FA0B6813ULL, "vxw_blightlands_px_0_0.vlm"},
	{0xA185D9C1701694D0ULL, "6086796_m.fxe"},
	{0xA185ED074692D8A7ULL, "73133187.ogg"},
	{0xA1861906DAFB3009ULL, "vxca_caverns2prc_9wz_0_0.vlm"},
	{0xA18622DBF7909636ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.cess.ani"},
	{0xA18694E813B693F3ULL, "6126563_m.fxe"},
	{0xA187637E22705969ULL, "6075110_f.fxe"},
	{0xA1876C07347C3690ULL, "vxlt_undercitydungeon_17e_0_0.vlm"},
	{0xA187883F05ECEA80ULL, "hlac_w_sundermount_0_1z.dds"},
	{0xA18805D2726E8171ULL, "273284555.ogg"},
	{0xA1886E22A75E1BBFULL, "45597591.ogg"},
	{0xA18894A4DBF98EB2ULL, "516436223.ogg"},
	{0xA188CD49C831768DULL, "6082999_m.fxe"},
	{0xA188D9C7D6BDF3C9ULL, "vxlt_0_lowtown_a2_n_c_hcx_0_0.vlm"},
	{0xA189EBAC4F2D66DCULL, "plt_and300pt_justice.nss"},
	{0xA18A1358284F2718ULL, "6066353_f.fxe"},
	{0xA18A4B011585DB4AULL, "dae200_bran6094722.cl"},
	{0xA18ABBD6BA49922CULL, "23613290.ogg"},
	{0xA18B29103913DF7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn_alt.evt"},
	{0xA18B7AB6BD1FEE08ULL, "vxdr_primevalentrance_37j_0_0.vlm"},
	{0xA18B7E855E5D98F7ULL, "co_smok.mao"},
	{0xA18BDD167FA75FFDULL, "6048113_m.fxe"},
	{0xA18BDFEB743BB793ULL, "1039885939.ogg"},
	{0xA18BE1698B33638EULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_buzzarda.phy"},
	{0xA18C7CA5000A33ABULL, "214636438.ogg"},
	{0xA18D20F87C100EF8ULL, "vxlt_darktown_5sr_0_0.vlm"},
	{0xA18D40762011510CULL, "vxlt_gallowscourtyard_747_0_0.vlm"},
	{0xA18D8841823616CBULL, "vxlt_0_lowtown_n_eht_0_0.vlm"},
	{0xA18DE54628453194ULL, "vxht_mansion2b_1g9_0_0.vlm"},
	{0xA18E0298AE05370AULL, "art\\characters\\creatures\\rubblegolem\\textures\\c_rbg_0s.dds"},
	{0xA18E28CCEF8F7F41ULL, "vxht_keep_677_0_0.vlm"},
	{0xA18E6D98C6FF7AADULL, "follower_banter6153979.cl"},
	{0xA18E839A9687F3C2ULL, "150423573.ogg"},
	{0xA18EC375C36D1A79ULL, "hm_qun_maecon.mor"},
	{0xA18EEEE9D683F2BEULL, "vxht_mansion2b_33a_0_0.vlm"},
	{0xA18F2499EAF4B112ULL, "528089738.ogg"},
	{0xA18F32326C265C94ULL, "hlac_lt_0_lowtown_a3_n_c_-1_5z.dds"},
	{0xA18F4C78E8B68E0DULL, "vxw_woundedcoast_170_0_0.vlm"},
	{0xA18F7A806EB09597ULL, "zz_inftrans.ncc"},
	{0xA18F9E69C7EC94FBULL, "6107801_m.fxe"},
	{0xA18FE380896F47B7ULL, "lowtown_fem_lowlife_mob6169183.cl"},
	{0xA1916479BC8021B8ULL, "6175393_m.fxe"},
	{0xA191772EE51BCDBFULL, "static_static_p40.psh.fxo.dx11.assembly"},
	{0xA19179F59599120CULL, "store_dae321_bodahn.utm"},
	{0xA191B87DEC145A61ULL, "vxca_caverns1_j5m_0_0.vlm"},
	{0xA191C00A870122BBULL, "hlac_ht_keep_n_3732_11z.dds"},
	{0xA191F3895353060EULL, "249893669.ogg"},
	{0xA191F6091BAA8C10ULL, "6127310_m.fxe"},
	{0xA19228103919FC6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn_alt.gad"},
	{0xA192CA4F2FDC3B2FULL, "llac_lt_backalley_n_0_17.dds"},
	{0xA1932CB69F47ABE5ULL, "mcr371_scared_mage_m6086910.cl"},
	{0xA193407EB03B857AULL, "vxht_0_hightown_n_a2k_0_0.vlm"},
	{0xA19353247AE8CB16ULL, "471664037.ogg"},
	{0xA193732EB2407CC1ULL, "zz_dae_debug6029825.cl"},
	{0xA1938D47035F29D3ULL, "gen00fl_aveline6131230.cl"},
	{0xA1942ADBF79CE127ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.cess.evt"},
	{0xA19499E6A3C829F4ULL, "6142456_m.fxe"},
	{0xA19510227718F1CDULL, "6139842_m.fxe"},
	{0xA19522C2E24ACF3AULL, "vxht_0_hightown_9wg_0_0.vlm"},
	{0xA195653C4FACF2A1ULL, "zz_mer_debug6023995.cl"},
	{0xA1956F8FCE050B6EULL, "6143264_m.fxe"},
	{0xA19595CC9A335B1DULL, "vxlt_gallowscourtyard_n_3x5_0_0.vlm"},
	{0xA195F55D2F17CE16ULL, "hlac_lt_gamlenshouse_n_-1_7y.dds"},
	{0xA1967335EED9CF83ULL, "vxw_bonepitdestroyed_a3_1p8_0_0.vlm"},
	{0xA196799E8096755BULL, "art\\levels\\ztools\\default\\generic\\3wo_angle.msh"},
	{0xA1968B9E809693A6ULL, "art\\levels\\ztools\\default\\generic\\3wo_angle.mao"},
	{0xA196979E8096A845ULL, "art\\levels\\ztools\\default\\generic\\3wo_angle.mmh"},
	{0xA197292380BD3094ULL, "hlac_lt_warehouse_n_1_6y.dds"},
	{0xA1973F26BA278F0DULL, "6085899_m.fxe"},
	{0xA197BE587C2EC599ULL, "6219197_m.fxe"},
	{0xA197BFE5D5E9C06BULL, "fen110_fenris6100281.cl"},
	{0xA197E0AAE764B3C4ULL, "w_encounter1_48_0.phy"},
	{0xA1985D86753A561EULL, "vxlt_gamlenshouse_n_nh_0_0.vlm"},
	{0xA1986C33EBE82C09ULL, "llac_lt_0_lowtown_n_65538_4.dds"},
	{0xA19870418B2F8184ULL, "169451545.ogg"},
	{0xA198E830A17A0452ULL, "gen_im_arm_glv_lgt_start.uti"},
	{0xA19938BE576BDC9FULL, "gen00fl_aveline6101721.cl"},
	{0xA1994B2CBF322563ULL, "6084223_m.fxe"},
	{0xA199CE6FE8C0A4E7ULL, "ser191_emeric6104760.cl"},
	{0xA199EC93AF160688ULL, "vxdr_primevalentrance_30k_0_0.vlm"},
	{0xA199FD9586FD9211ULL, "6000962_m.fxe"},
	{0xA19A46A43579C5DBULL, "gen_im_cns_flu_merril_03.uti"},
	{0xA19A4BDBF7A18570ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.cess.gad"},
	{0xA19A4E80EE12F9A6ULL, "zz_qun_debug6028296.cl"},
	{0xA19A4EF3719D787DULL, "6043013_m.fxe"},
	{0xA19AEE89737AC01BULL, "862288179.ogg"},
	{0xA19B198DE1A5A060ULL, "553805929.ogg"},
	{0xA19B4133AD690648ULL, "vxca_caverns1_j5h_0_0.vlm"},
	{0xA19C327FE5EB48DFULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.dura.gad"},
	{0xA19C3BF511F7CFB7ULL, "oil_si.dds"},
	{0xA19C9AA9E18597C1ULL, "vxht_keep_606_0_0.vlm"},
	{0xA19CD0E223E12BE8ULL, "vxlt_0_lowtown_docks_b9p_0_1.vlm"},
	{0xA19CD775169241D2ULL, "gen00fl_fenris6113913.cl"},
	{0xA19CE09AC27F3CDCULL, "vxw_encounter1_ek_0_1.vlm"},
	{0xA19D531A261C033EULL, "w_blightlands_189_0.phy"},
	{0xA19DA6C9CA666F3FULL, "6207552_m.fxe"},
	{0xA19DC282C0D34BB9ULL, "face_skinnedtsb_p0.vsh.fxo.dx11"},
	{0xA19DFD233DCBF950ULL, "6108230_m.fxe"},
	{0xA19E745A80F0BFBEULL, "vxlt_gallowstemplar_6qa_0_0.vlm"},
	{0xA19E747D63BF2603ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldc.msh"},
	{0xA19E827D63BF3E5DULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldc.mmh"},
	{0xA19E867D63BF456EULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_shieldc.mao"},
	{0xA19EA4E153A15F3AULL, "fx_smoke.mao"},
	{0xA19EC94FED5DE5D2ULL, "1017417865.ogg"},
	{0xA19F1633DA43A4BBULL, "mag291_demon_marethari6152615.cl"},
	{0xA19F38B56907EDBDULL, "vxdr_primevaltomb_34z_0_0.vlm"},
	{0xA19F4FE6ECC417CEULL, "vxca_caverns2_6wa_0_2.vlm"},
	{0xA19F9DDCA048B5D4ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_floor.dds"},
	{0xA1A0372B5AB02442ULL, "vxlt_undercitydungeon_13k_0_0.vlm"},
	{0xA1A0937A09F16EB7ULL, "zz_seb_debug6185066.cl"},
	{0xA1A0AB0992B2F15AULL, "vxw_bonepitmining_a2_1pw_0_0.vlm"},
	{0xA1A0C28DFF2DDF02ULL, "121996901.ogg"},
	{0xA1A1C0AB4B0DFBF4ULL, "6015806_m.fxe"},
	{0xA1A22E907D334B8EULL, "zz_vault_debug6203374.cl"},
	{0xA1A290E450791811ULL, "vxdr_0_deeproads_a25_0_1.vlm"},
	{0xA1A2B94B090795CAULL, "87973701.ogg"},
	{0xA1A2DFB75F975245ULL, "hlac_ca_cavern1_haunted_24421_5y.dds"},
	{0xA1A31CA392D1E743ULL, "vxlt_0_lowtown_a2_ei0_0_0.vlm"},
	{0xA1A325DE361D718AULL, "gen_im_arm_cht_coma_t5.uti"},
	{0xA1A3277FE5F154CCULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.dura.evt"},
	{0xA1A384A44484EB74ULL, "hlac_lt_0_lowtown_n_1_8y.dds"},
	{0xA1A3CF3EE9D0F521ULL, "6142017_m.fxe"},
	{0xA1A3D69D83424D47ULL, "gen00fl_varric6099133.cl"},
	{0xA1A444413F3E399AULL, "vxht_keep_n_5y2_0_0.vlm"},
	{0xA1A4CAF94059CF94ULL, "fxa_pow_shine.mao"},
	{0xA1A5014C946C8B0EULL, "mcr370_cut_marriage.cnv"},
	{0xA1A5523DA2CDB51DULL, "mag371ar_gallows_templar.lst"},
	{0xA1A57D7828FFCA43ULL, "vxht_keep_n_6g6_0_0.vlm"},
	{0xA1A6099C16795A1DULL, "61187385.ogg"},
	{0xA1A610AF8C710267ULL, "hlac_w_bonepitdestroyed_a3_196614_0x.dds"},
	{0xA1A6383D217FA313ULL, "hlac_ht_chantry_n_2534_2y.dds"},
	{0xA1A6A289DC1AB671ULL, "323284175.ogg"},
	{0xA1A72B53F021DF15ULL, "def_dirt_s.dds"},
	{0xA1A74455CA5D425DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masa_0n.dds"},
	{0xA1A80558CE57EB01ULL, "vxdr_primevaltomb_76l_0_0.vlm"},
	{0xA1A8234C946EFB7EULL, "mcr370_cut_marriage.bnk"},
	{0xA1A8BA3CD62C629EULL, "zz_followers_banter6157600.cl"},
	{0xA1A91F2213E69530ULL, "85434387.ogg"},
	{0xA1A92B9BFFD7CA22ULL, "388185504.ogg"},
	{0xA1A96965F1760B9CULL, "6137643_m.fxe"},
	{0xA1A9DC08C2403FA5ULL, "6036673_m.fxe"},
	{0xA1AA01D3C525076DULL, "vxht_0_hightown_a81_0_0.vlm"},
	{0xA1AAC28178D77AFFULL, "vxlt_backalley_n_1xg_0_0.vlm"},
	{0xA1AAC2EEC85D11A8ULL, "character_character_p3.psh.fxo.dx11.assembly"},
	{0xA1AAE237ACA3E0E8ULL, "gen00fl_varric6077819.cl"},
	{0xA1AAE52D3530A6EEULL, "dae200_bran6094818.cl"},
	{0xA1AAF3E27F5FC503ULL, "hlac_lt_0_lowtown_docks_a3_-3_5y.dds"},
	{0xA1AB1DCC31917610ULL, "6072683_m.fxe"},
	{0xA1AB4A3DA2D2130CULL, "mag371ar_gallows_templar.ncs"},
	{0xA1AB5277B475555CULL, "vxw_encounter1_jj_0_1.vlm"},
	{0xA1AB5426BF92C8A7ULL, "vxca_caverns1_hm5_0_0.vlm"},
	{0xA1ABBA716DE24093ULL, "lt_0_lowtown_docks_n_139_0.phy"},
	{0xA1ABC4774FE46CC7ULL, "vxw_encounter3_80_0_0.vlm"},
	{0xA1ABCBE4E5C632D8ULL, "6129777_f.fxe"},
	{0xA1ABDFC8345707C3ULL, "hlac_w_blightlands_131068_0y.dds"},
	{0xA1AC7EC8C70092F6ULL, "one121cr_note_thug.utc"},
	{0xA1ACE63772E6D83EULL, "vxlt_gallowstemplar_6tb_0_0.vlm"},
	{0xA1ACE82AA36779DDULL, "25005735.ogg"},
	{0xA1ACE89B5F968845ULL, "vxdr_primevalentrance_28c_0_0.vlm"},
	{0xA1ACFF234E1BF121ULL, "gen_im_trs_gen_junk2_04.uti"},
	{0xA1AD06AD0681A805ULL, "vxlt_darktown_5nl_0_0.vlm"},
	{0xA1AD550646981DB5ULL, "vxht_0_hightown_bie_0_0.vlm"},
	{0xA1ADC81E3304D216ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunkxlo.msh"},
	{0xA1ADD046E4D2A519ULL, "gen00fl_anders6119893.cl"},
	{0xA1ADD21E3304E314ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunkxlo.mmh"},
	{0xA1ADD61E3304EA27ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww43_cellchunkxlo.mao"},
	{0xA1ADDF44A9B1935DULL, "w_sundermount_1_0.mao"},
	{0xA1ADE344A9B19996ULL, "w_sundermount_1_0.mmh"},
	{0xA1ADED44A9B1AA94ULL, "w_sundermount_1_0.msh"},
	{0xA1AFBE4D7BCF8E55ULL, "gen00fl_aveline6140144.cl"},
	{0xA1AFE2281064FB12ULL, "ht_0_hightownexplosion_n_28_0.rml"},
	{0xA1B0089625E5F286ULL, "484452067.ogg"},
	{0xA1B00F7FE5FBB625ULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.dura.ani"},
	{0xA1B00FC10A56C407ULL, "vxdr_0_deeproads_cu8_0_1.vlm"},
	{0xA1B025CD3C03F56DULL, "325008034.ogg"},
	{0xA1B1226F8A6B1380ULL, "644535838.ogg"},
	{0xA1B1D7812ABA6C61ULL, "1014118207.ogg"},
	{0xA1B21DCEB4A00988ULL, "vxlt_0_lowtown_ew9_0_0.vlm"},
	{0xA1B2D6EE55B8E40EULL, "mag170_hugh6080705.cl"},
	{0xA1B2F89F093E2FE8ULL, "gen00fl_anders6125970.cl"},
	{0xA1B36DAA13A003FEULL, "mag210_arianni6031696.cl"},
	{0xA1B3D4EA72AF759EULL, "follower_banter6154122.cl"},
	{0xA1B3EF551AF5E2C3ULL, "vxht_keep_n_6b9_0_0.vlm"},
	{0xA1B3F29AC34E9344ULL, "hm_bar_gossip2.mor"},
	{0xA1B4123ADD6819B9ULL, "404112131.ogg"},
	{0xA1B419A7B70F6CFAULL, "vxht_0_hightown_a2_be6_0_0.vlm"},
	{0xA1B478359885CF7BULL, "vxht_0_hightown_a3_b1b_0_0.vlm"},
	{0xA1B4C0393DA3E183ULL, "vxca_cavern1_haunted_i24_0_0.vlm"},
	{0xA1B4E7FA57979829ULL, "hlac_lt_0_lowtown_65533_2z.dds"},
	{0xA1B50D3D2B58B97CULL, "crt_star.dds"},
	{0xA1B58120D326B4D4ULL, "zz_chantryexplodetest.cut"},
	{0xA1B62993D1AF1456ULL, "ts_gems_02.gda"},
	{0xA1B676899DA4BAF7ULL, "188539625.ogg"},
	{0xA1B705D597D85E50ULL, "60485241.ogg"},
	{0xA1B74E677EFFBBD2ULL, "dre000_carver6037584.cl"},
	{0xA1B779F9E6FC541AULL, "zz_vault_debug6178917.cl"},
	{0xA1B793ACA323B50DULL, "267623461.ogg"},
	{0xA1B84B00FCFF8957ULL, "w_blightlands_130_0.msh"},
	{0xA1B85D00FCFFA7BAULL, "w_blightlands_130_0.mao"},
	{0xA1B86100FCFFAEC9ULL, "w_blightlands_130_0.mmh"},
	{0xA1B954D5502D0488ULL, "6103756_m.fxe"},
	{0xA1B976D18137046FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mc.mao"},
	{0xA1B982D18137180CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mc.mmh"},
	{0xA1B988D1813722FEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16mc.msh"},
	{0xA1BA0DC7A1C92DE2ULL, "6173656_m.fxe"},
	{0xA1BAA291FFF5CDD2ULL, "6129017_m.fxe"},
	{0xA1BAA687D6BA4AB5ULL, "vxht_mansion_n_4y8_0_0.vlm"},
	{0xA1BBDC3C9B4EC8F8ULL, "zz_approval_debug6003971.cl"},
	{0xA1BC0828FFAE0D83ULL, "6091849_m.fxe"},
	{0xA1BC87E405D1D27CULL, "6016740_m.fxe"},
	{0xA1BCAEF1C1E3544CULL, "l_lgt210ar_gang_hideout_d.bnk"},
	{0xA1BD4E2C6BA0C02AULL, "454958986.ogg"},
	{0xA1BD668AE9D4CE89ULL, "20120206.ogg"},
	{0xA1BD7DE153BBCDECULL, "fx_smoke.dds"},
	{0xA1BD8BAD0FA7DB9DULL, "dae000_mother6136924.cl"},
	{0xA1BDF23A1013168BULL, "90169538.ogg"},
	{0xA1BE2BF34BF6AFB5ULL, "hlac_w_sundermount_-262144_0z.dds"},
	{0xA1BE47405513FB10ULL, "dae000_paivel6134509.cl"},
	{0xA1BE4EFF4CADCBF4ULL, "gen_im_trs_cot_lgt_cht_r2.uti"},
	{0xA1BE555D30370942ULL, "vxca_cavern1_haunted_hkw_0_0.vlm"},
	{0xA1BE6F41C8CA44DDULL, "atl_shared_dxt1_03.dds"},
	{0xA1BE819DFC4CDC87ULL, "vxdr_0_deeproads_cr9_0_1.vlm"},
	{0xA1BEA5CCE95D4B4FULL, "6115464_m.fxe"},
	{0xA1BF603B4103C807ULL, "mag101_aombanter6052601.cl"},
	{0xA1BFD1360CAE2C25ULL, "145482340.ogg"},
	{0xA1C033E57BD89870ULL, "hlac_ca_caverns2prc_5802_19x.dds"},
	{0xA1C08FABA6962208ULL, "vxlt_0_lowtown_er6_0_0.vlm"},
	{0xA1C09AC355AAB61BULL, "6161272_m.fxe"},
	{0xA1C0A7181C37C518ULL, "6180238_m.fxe"},
	{0xA1C0BBCEAF3E23A8ULL, "895558530.ogg"},
	{0xA1C0BC7D888DE07DULL, "hlac_ca_caverns1_24426_8z.dds"},
	{0xA1C0EB73F5F942AFULL, "vxht_mansion2b_1hr_0_1.vlm"},
	{0xA1C10255B030ABE4ULL, "6196897_m.fxe"},
	{0xA1C14E7E6CC366CCULL, "fb5.cut"},
	{0xA1C25C0EDCE40FA4ULL, "dae320ar_hightown_nt.lst"},
	{0xA1C26CA14A8A4165ULL, "mrl150_merrill6093523.cl"},
	{0xA1C2BBEB3B703AEDULL, "261016498.ogg"},
	{0xA1C33215B7989660ULL, "gen00fl_anders6112976.cl"},
	{0xA1C43DFF7F56A5D5ULL, "6208556_f.fxe"},
	{0xA1C45EDC1BE55B4EULL, "fx_large_canter.mao"},
	{0xA1C4C7D2A437133BULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\_textures\\vst_gallows_d.dds"},
	{0xA1C4DFE71612335BULL, "6149094_m.fxe"},
	{0xA1C55E6E8E7236C0ULL, "vxlt_0_lowtown_docks_a3_cyf_0_1.vlm"},
	{0xA1C565C120F09750ULL, "mer240_jansen6050405.cl"},
	{0xA1C5E5F71C7CF53CULL, "6176728_f.fxe"},
	{0xA1C601ECDACE797AULL, "da2terrain_rendertodepth_p1.psh.fxo.dx11.assembly"},
	{0xA1C676CCF86A711AULL, "art\\vfx\\combat\\vfx_gib_cora_ltr.gad"},
	{0xA1C74F9AE1B08436ULL, "6175148_m.fxe"},
	{0xA1C76E0A28FDBE19ULL, "gen00fl_anders6125487.cl"},
	{0xA1C83E543C0E04B3ULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.dura.ani"},
	{0xA1C85C0EDCE87B0DULL, "dae320ar_hightown_nt.ncs"},
	{0xA1C9715BAF526F97ULL, "6135295_m.fxe"},
	{0xA1C9EA66EA6DD905ULL, "vxlt_darktown_5tj_0_0.vlm"},
	{0xA1C9FC723C87B648ULL, "vxdr_primevalentrance_228_0_0.vlm"},
	{0xA1CA4EAF09D9A866ULL, "283145212.ogg"},
	{0xA1CABEADB3024FBCULL, "dae000ip_crowd_mf5_talk5.utp"},
	{0xA1CAF4C145006DD3ULL, "art\\characters\\playercharacter\\children\\animation\\blendtree\\child_exp_movement.anb"},
	{0xA1CB102F6247688EULL, "hlac_ht_0_hightown_a3_-196609_0x.dds"},
	{0xA1CB1589AEFA81AFULL, "150406546.ogg"},
	{0xA1CB5E708A95A656ULL, "48974383.ogg"},
	{0xA1CCA7316A97176DULL, "vxlt_gallowstemplar_3jr_0_0.vlm"},
	{0xA1CCA9FFA6A62D45ULL, "vxht_0_hightown_n_a7o_0_0.vlm"},
	{0xA1CCE0698B6AF951ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_buzzarda.mmh"},
	{0xA1CCEC698B6B0DF2ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_buzzarda.mao"},
	{0xA1CCFA698B6B250FULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_buzzarda.msh"},
	{0xA1CD14B33FA4593AULL, "6037668_m.fxe"},
	{0xA1CDBAB3FF711D11ULL, "vxw_woundedcoast_1c8_0_0.vlm"},
	{0xA1CDEF02711F9C60ULL, "gen00fl_varric6119012.cl"},
	{0xA1CE023F78C98FB5ULL, "mcr310_party_banter6087820.cl"},
	{0xA1CE2E9A8BB3C9B1ULL, "hlac_lt_0_lowtown_docks_a3_-131074_7z.dds"},
	{0xA1CF0CA047A0CE9BULL, "6166372_m.fxe"},
	{0xA1CF36E8ED4BF189ULL, "vxht_chantry_n_356_0_1.vlm"},
	{0xA1CFB7781EFC3AE6ULL, "art\\vfx\\combat\\vfx_m_modal_act_c.anb"},
	{0xA1CFC157F3A33D6DULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha05a.mmh"},
	{0xA1CFD357F3A35BF3ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha05a.msh"},
	{0xA1CFEC4C6DD6F1ADULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf01.msh"},
	{0xA1CFFE4C6DD71033ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf01.mmh"},
	{0xA1D035312DAA87A6ULL, "vxw_encounter1_e0_0_0.vlm"},
	{0xA1D04EB0769035F9ULL, "464866245.ogg"},
	{0xA1D0BCE827E7A74FULL, "298698584.ogg"},
	{0xA1D15B3DA2F2FFFAULL, "mag371ar_gallows_templar.are"},
	{0xA1D16308FE35B7F9ULL, "zz_dae_debug6049975.cl"},
	{0xA1D1C343D04C96E1ULL, "vxht_0_hightown_n_a82_0_0.vlm"},
	{0xA1D1E6199CE72760ULL, "107304537.ogg"},
	{0xA1D1EC09B89F2FBBULL, "6020395_m.fxe"},
	{0xA1D21F0CC38F0E53ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgtp_0d.dds"},
	{0xA1D234A247688946ULL, "hlac_lt_undercitydungeon_-1_29z.dds"},
	{0xA1D287EF522F34C5ULL, "6097221_m.fxe"},
	{0xA1D29FC269C71C05ULL, "mcr370_meredith6136172.cl"},
	{0xA1D2D20BE1B32A29ULL, "vxht_0_hightown_a2_c5l_0_0.vlm"},
	{0xA1D3E0EED88D209BULL, "zz_fenris_slashes.cut"},
	{0xA1D46BFEDC50EF50ULL, "6170062_m.fxe"},
	{0xA1D5969C6E866687ULL, "vxw_sundermount_34q_0_0.vlm"},
	{0xA1D5BC360B7BA699ULL, "vxht_brothel_n_4cu_0_0.vlm"},
	{0xA1D5F8B982FF836BULL, "730383467.ogg"},
	{0xA1D636543C1A34EAULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.dura.evt"},
	{0xA1D6407248248408ULL, "qun221_petrice6075981.cl"},
	{0xA1D6E06E67550026ULL, "mag370_meredith6172389.cl"},
	{0xA1D75816EC168707ULL, "226708192.ogg"},
	{0xA1D79152AA7983BDULL, "6085482_m.fxe"},
	{0xA1D7E11D13663EF8ULL, "ts_generic_arm_props.gda"},
	{0xA1D8AA9D1C7A6735ULL, "vxht_0_hightown_bxb_0_0.vlm"},
	{0xA1D8AB2E8D3352A9ULL, "vxlt_0_lowtown_db6_0_0.vlm"},
	{0xA1D8C6A74283C54AULL, "6102951_m.fxe"},
	{0xA1D92A48B31FF096ULL, "llac_ht_playershouse_n_855_25.dds"},
	{0xA1DA63A189297C9DULL, "vxlt_gallowstemplar_n_5xi_0_0.vlm"},
	{0xA1DA83A157D585F6ULL, "hlac_lt_0_lowtown_docks_a3_-65538_3x.dds"},
	{0xA1DB190E5C8D2FEDULL, "vxlt_gallowstemplar_3gq_0_0.vlm"},
	{0xA1DB28245B5D93C4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany.ani"},
	{0xA1DB4BF991A142C8ULL, "vxht_0_hightown_n_a87_0_0.vlm"},
	{0xA1DB4F95F4EDC227ULL, "vxlt_0_lowtown_docks_bg8_0_0.vlm"},
	{0xA1DB7DFCEA738F16ULL, "w_encounter3_156_0.mao"},
	{0xA1DB89FCEA73A3B5ULL, "w_encounter3_156_0.mmh"},
	{0xA1DB8BFCEA73A72BULL, "w_encounter3_156_0.msh"},
	{0xA1DB91CCF87CFB00ULL, "art\\vfx\\combat\\vfx_gib_cora_ltr.ani"},
	{0xA1DB91CCF87CFB0BULL, "art\\vfx\\combat\\vfx_gib_cora_ltr.anb"},
	{0xA1DBB1250F6328A8ULL, "dae000cr_grand_cleric6147898.cl"},
	{0xA1DBE7D757400F12ULL, "hlac_lt_gallowscourtyard_n_-2_0x.dds"},
	{0xA1DBE956EBC34D1AULL, "550934710.ogg"},
	{0xA1DBF4FD74CAE073ULL, "6032613_m.fxe"},
	{0xA1DC24BDED5B018BULL, "167586802.ogg"},
	{0xA1DC6B6C1F61514FULL, "gen_im_wep_war_shd_act2_07.uti"},
	{0xA1DC8A7425E11683ULL, "vxdr_primevaltomb_8m8_0_0.vlm"},
	{0xA1DC9B5B601B8923ULL, "fen210_dying_magister6108580.cl"},
	{0xA1DCBA781F06C991ULL, "art\\vfx\\combat\\vfx_m_modal_act_c.mmh"},
	{0xA1DD15543C201B5DULL, "art\\vfx\\combat\\vfx_c_mer_crust_c.dura.gad"},
	{0xA1DD4D85F9C99D18ULL, "674964598.ogg"},
	{0xA1DD768E2A954011ULL, "6185104_m.fxe"},
	{0xA1DD773760A7F30AULL, "6072322_m.fxe"},
	{0xA1DE709E80D41BF2ULL, "art\\levels\\ztools\\default\\generic\\3wo_angle.phy"},
	{0xA1DEA70E1FA0A026ULL, "vxw_encounter1_j1_0_0.vlm"},
	{0xA1DED80C35F2C082ULL, "hlac_dr_slavers_3217_7z.dds"},
	{0xA1DF4626BDC5FFECULL, "728072888.ogg"},
	{0xA1DF562196BB9F39ULL, "728781352.ogg"},
	{0xA1DF7D7F0E1D9D89ULL, "vxca_caverns1_hu3_0_0.vlm"},
	{0xA1DFFA7E3C3BD4B6ULL, "seb221_random_servants6187831.cl"},
	{0xA1E02DA60A30CE44ULL, "vxht_0_hightown_a2_b8s_0_0.vlm"},
	{0xA1E07E1D9F9F64B7ULL, "6175321_m.fxe"},
	{0xA1E0BA57F3B1A42AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha05a.phy"},
	{0xA1E1313D907E9DDBULL, "vxht_mansion_n_4x5_0_0.vlm"},
	{0xA1E1591EFBF7AEB5ULL, "vxlt_gallowstemplar_5qm_0_0.vlm"},
	{0xA1E1C84910154F9EULL, "mag291_demon_marethari6028922.cl"},
	{0xA1E225245B63AD26ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany.gad"},
	{0xA1E2FCAC4F79AC07ULL, "plt_and300pt_justice.plo"},
	{0xA1E3264D675F1469ULL, "vxca_cavern1_haunted_hin_0_0.vlm"},
	{0xA1E39EAF3C9BCAE3ULL, "289090324.ogg"},
	{0xA1E3C3AED8E55E49ULL, "92770026.ogg"},
	{0xA1E40879607C7F07ULL, "vxw_sundermount_33p_0_0.vlm"},
	{0xA1E42E12FD71BF19ULL, "vxht_brothel_n_4ht_0_0.vlm"},
	{0xA1E4756233ADF253ULL, "vfx_chunkyparticle_p2.psh.fxo.dx11"},
	{0xA1E4F8905DD0AEF6ULL, "486728340.ogg"},
	{0xA1E54FBA593BC86FULL, "287747253.ogg"},
	{0xA1E585D2E353E3F9ULL, "follower_banter6156060.cl"},
	{0xA1E62A26E984AE6AULL, "vxca_cavern1_haunted_j26_0_0.vlm"},
	{0xA1E636E6BA6A6DAAULL, "dae000_bartrand6031146.cl"},
	{0xA1E69D99F6321897ULL, "425841274.ogg"},
	{0xA1E6C4B281CA38C8ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.anb"},
	{0xA1E6E8FE85886FD4ULL, "vxlt_warehouse_15i_0_0.vlm"},
	{0xA1E6EF317374CF6AULL, "cry_val02.mao"},
	{0xA1E72D7B8AE68B84ULL, "llac_lt_darktown_-1_55.dds"},
	{0xA1E7630EDD033797ULL, "dae320ar_hightown_nt.are"},
	{0xA1E76A3C0D845C8FULL, "gen_im_npc_arm_hum_cht_mrcb.uti"},
	{0xA1E76C152DD5F8D5ULL, "follower_banter6170594.cl"},
	{0xA1E7710E0A66F15CULL, "framebuffereffects_vsmblur_p1.psh.fxo.dx11.assembly"},
	{0xA1E806863778305FULL, "vxca_caverns2_74n_0_0.vlm"},
	{0xA1E846DA8E226863ULL, "fxa_evl_smoke.mao"},
	{0xA1E8C3B5731BB568ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medd_0t.dds"},
	{0xA1E8C90522431AD8ULL, "hlac_w_encounter1_65537_0y.dds"},
	{0xA1E920245B69C305ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\injured_supported_bethany.evt"},
	{0xA1EAE35EFD07377AULL, "vxlt_0_lowtown_a2_he0_0_0.vlm"},
	{0xA1EB07E555214039ULL, "hlac_ht_chantry_n_2534_3x.dds"},
	{0xA1EBDD31CAF4EC22ULL, "1044416369.ogg"},
	{0xA1EBEDC513F3D404ULL, "947894195.ogg"},
	{0xA1EC33F331AD4F4FULL, "abi_base.gda"},
	{0xA1EC6BF060864226ULL, "42225997.ogg"},
	{0xA1ECFB98DCD07E4CULL, "gen00fl_varric6033569.cl"},
	{0xA1ED25B49C4C5F3EULL, "one280ip_iwan_corpse.utp"},
	{0xA1EDF2BD618DD00AULL, "vxht_mansion_n_6nt_0_0.vlm"},
	{0xA1EE41904D504FD3ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.impa.evt"},
	{0xA1EEC3C18C1C4C7CULL, "vxht_0_hightown_n_c4e_0_0.vlm"},
	{0xA1EEE5544AA76CD7ULL, "409952437.ogg"},
	{0xA1EEFE5033CBAE5FULL, "gen00fl_merrill6131352.cl"},
	{0xA1EFA1A4A5C0FBB7ULL, "vxw_encounter2_p6_0_1.vlm"},
	{0xA1EFA2D2A0F0BC6DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_grand.mmh"},
	{0xA1EFA6D2A0F0C2BEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_grand.mao"},
	{0xA1EFB4D2A0F0DAF3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_grand.msh"},
	{0xA1F07FB77CB5E588ULL, "469657489.ogg"},
	{0xA1F092BE56EC0EFFULL, "6114070_m.fxe"},
	{0xA1F0D0F39BC0FEF1ULL, "vxca_caverns2_7uc_0_0.vlm"},
	{0xA1F0E114BAE60985ULL, "801338352.ogg"},
	{0xA1F0E79AAF06EB30ULL, "vxw_sundermount_2w6_0_0.vlm"},
	{0xA1F11A3965661873ULL, "461203564.ogg"},
	{0xA1F142EA8C050990ULL, "vxlt_0_lowtown_a3_n_c_hqf_0_0.vlm"},
	{0xA1F175E191171B13ULL, "dae000_friend_turn_in6179317.cl"},
	{0xA1F21F7308AA347FULL, "6129984_m.fxe"},
	{0xA1F254F198317E18ULL, "215695668.ogg"},
	{0xA1F25ECF7DEC83C7ULL, "vxlt_0_lowtown_a2_ex3_0_0.vlm"},
	{0xA1F2988A98D90414ULL, "6091994_m.fxe"},
	{0xA1F2A3818978898FULL, "pro000_merchant6051641.cl"},
	{0xA1F2C4AF21033AACULL, "llac_lt_0_lowtown_n_131069_0.dds"},
	{0xA1F4619BF5AC6456ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_2.evt"},
	{0xA1F49F389B07BAC2ULL, "vxlt_ruinedhouse_n_59_0_0.vlm"},
	{0xA1F4AEE38274B7B6ULL, "83541869.ogg"},
	{0xA1F4BEE3333C1188ULL, "vxca_caverns1_gkr_0_1.vlm"},
	{0xA1F4D8CAE1A4C063ULL, "6110319_m.fxe"},
	{0xA1F4F61E0733656EULL, "vxca_caverns1_cjv_0_0.vlm"},
	{0xA1F53A904D566214ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.impa.gad"},
	{0xA1F623FCD548DAB1ULL, "class_abilities_mage.gda"},
	{0xA1F679C809E6066EULL, "763884624.ogg"},
	{0xA1F6AB1EBC75945AULL, "6099592_m.fxe"},
	{0xA1F7622A3ACDDA0AULL, "art\\vfx\\combat\\vfx_glb_phys_imp_min_c.impa.gad"},
	{0xA1F791DCA01D8E0CULL, "677988928.ogg"},
	{0xA1F7F5223BAC56BFULL, "qun100cr_qunari_warrior6.utc"},
	{0xA1F7FCC8407960EDULL, "vxlt_gallowstemplar_3yo_0_0.vlm"},
	{0xA1F8010395F32970ULL, "6125842_m.fxe"},
	{0xA1F80F97CEFF7D93ULL, "1046210528.ogg"},
	{0xA1F81BAE2C7083EAULL, "plt_ave100pt_promotion.plo"},
	{0xA1F8567076535206ULL, "dae000cr_grand_cleric6124679.cl"},
	{0xA1F86222B2FC8FC7ULL, "vxlt_house_hg_0_0.vlm"},
	{0xA1F8ADE61BD9FB0CULL, "138604785.ogg"},
	{0xA1F8F811CAB5DB78ULL, "gen00fl_fenris6112088.cl"},
	{0xA1F937D52B85DB2DULL, "vxlt_shop_dm_0_0.vlm"},
	{0xA1F941896505F644ULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_gen_c.dura.ani"},
	{0xA1F9C2B72A3A6EA1ULL, "vxlt_0_lowtown_a2_n_c_euc_0_0.vlm"},
	{0xA1F9C81EFBC3B231ULL, "6060429_m.fxe"},
	{0xA1F9D19545FE6125ULL, "6147199_m.fxe"},
	{0xA1F9F2ECD84898CBULL, "vxw_bonepitmining_a2_n7_0_0.vlm"},
	{0xA1FA10D36F990465ULL, "dae000_gamlen6149440.cl"},
	{0xA1FA2B77EDEE707FULL, "w_woundedcoast_182_0.phy"},
	{0xA1FA3BC51E9ECB86ULL, "vxlt_0_lowtown_eqd_0_1.vlm"},
	{0xA1FA6BA7533F4239ULL, "51805262.ogg"},
	{0xA1FB49F313C5027BULL, "gen00fl_aveline6122003.cl"},
	{0xA1FB589BF5B273E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_2.gad"},
	{0xA1FB8A4C588768E3ULL, "gen00fl_isabela6099502.cl"},
	{0xA1FBCB64CADE0636ULL, "zz_one_debug6099951.cl"},
	{0xA1FC2F8155A57013ULL, "vxlt_darktown_467_0_0.vlm"},
	{0xA1FC49904D5C9AF2ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.impa.ani"},
	{0xA1FCCEEC989B6C86ULL, "6081675_f.fxe"},
	{0xA1FD1748B7A9BFC9ULL, "6025097_m.fxe"},
	{0xA1FD19066CD0A764ULL, "vxht_0_hightown_c0f_0_0.vlm"},
	{0xA1FD5AAD2376F82AULL, "hlac_lt_warehouse_65537_0y.dds"},
	{0xA1FD6A7996949E83ULL, "187233526.ogg"},
	{0xA1FDB8F1C21AFC7EULL, "l_lgt210ar_gang_hideout_d.wbd"},
	{0xA1FDD69D77411D0BULL, "vxw_sundermount_244_0_0.vlm"},
	{0xA1FDE0AFCDDA2914ULL, "plt_cod_itm_stonehammer.plo"},
	{0xA1FE20B02CB4948FULL, "dae000_bodhan_letter6144248.cl"},
	{0xA1FE427AE163C7F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v1.evt"},
	{0xA1FE452A3AD3C7A1ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_min_c.impa.evt"},
	{0xA1FEBD73BBEADC54ULL, "520769942.ogg"},
	{0xA1FF049B48E2277FULL, "6113170_m.fxe"},
	{0xA1FF457427E7484BULL, "6108035_m.fxe"},
	{0xA1FF5977A0FD03B0ULL, "vxw_sundermount_2x7_0_0.vlm"},
	{0xA1FF941C722C1447ULL, "744781756.ogg"},
	{0xA1FFBD63DC27C949ULL, "6165793_m.fxe"},
	{0xA1FFE0BF037D482DULL, "ss_merrillhawke.cut"},
	{0xA2003B24781CACDEULL, "zz_ave_debug6064067.cl"},
	{0xA2003E89650C0FA6ULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_gen_c.dura.gad"},
	{0xA20040197C589FB3ULL, "589441387.ogg"},
	{0xA2004E0973075744ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_wood_d.dds"},
	{0xA20055DA8E371319ULL, "fxa_evl_smoke.dds"},
	{0xA200770E186FCD7AULL, "hlac_lt_undercityencounter_-1_8z.dds"},
	{0xA2009BD2A0FF232AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_grand.phy"},
	{0xA2010A678ACF1C94ULL, "6096894_m.fxe"},
	{0xA20117CAC6111118ULL, "154183563.ogg"},
	{0xA20155530100403EULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_cwk.ani"},
	{0xA201599BF5B6E08FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_2.ani"},
	{0xA20193535C523F6BULL, "vxca_cave_2d4_0_0.vlm"},
	{0xA201E36C947CFABEULL, "6033714_m.fxe"},
	{0xA20215C6E1BF6003ULL, "hlac_lt_0_lowtown_a2_n_-1_7x.dds"},
	{0xA20234C2DF2390DEULL, "gen00fl_varric6093680.cl"},
	{0xA2025A475D03EA2AULL, "mer210_smith6051418.cl"},
	{0xA2028F56B3AC1E7AULL, "hlac_lt_0_lowtown_docks_a3_65534_6x.dds"},
	{0xA202975F934F8E63ULL, "hlac_ca_cavern1_haunted_24421_4x.dds"},
	{0xA202D20ED704842AULL, "769812816.ogg"},
	{0xA2031E374A1517ABULL, "hlac_w_sundermount_393214_0x.dds"},
	{0xA20327FEA5BDCEFCULL, "758175757.ogg"},
	{0xA2033B873AEF1260ULL, "dae000_fem_priest_mob6174183.cl"},
	{0xA2036BB78DE23924ULL, "prp_str_dbone_n.dds"},
	{0xA20387617739F640ULL, "247746118.ogg"},
	{0xA2039B9429FDE66EULL, "6129743_m.fxe"},
	{0xA203B661D6F62C46ULL, "dre180cs_thewayout.cut"},
	{0xA203D66D7F8748FCULL, "art\\characters\\creatures\\rubblegolem\\textures\\c_rbgb_0e.dds"},
	{0xA2040E05CB100763ULL, "vrc321tr_varric_crazy.utt"},
	{0xA20425D663F85228ULL, "lgt250cr_way_dalish_templar.utc"},
	{0xA20458BF7BB54DBFULL, "hlac_lt_0_lowtown_a2_n_-65536_3z.dds"},
	{0xA204B7F814754916ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp02.phy"},
	{0xA204F30C3BE45478ULL, "w_encounter3_165_0.phy"},
	{0xA2053B7AE169DA34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v1.gad"},
	{0xA20588261DB5626EULL, "467983089.ogg"},
	{0xA206F194F3C8B092ULL, "vxht_bartrand_n_3qw_0_0.vlm"},
	{0xA207398965122585ULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_gen_c.dura.evt"},
	{0xA207775CE9080737ULL, "vxlt_0_lowtown_docks_a3_b11_0_0.vlm"},
	{0xA2078B0234157B5AULL, "6103354_m.fxe"},
	{0xA20793AF1B22C2E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_local_corpse_idle.anb"},
	{0xA207964DE11CEE05ULL, "380347929.ogg"},
	{0xA207D8D4CC0E579BULL, "318958922.ogg"},
	{0xA207F71232078407ULL, "hlac_w_blightlands_196611_0z.dds"},
	{0xA208D19C27E1473DULL, "lt_0_lowtown_docks_n_32_0_ll.mmh"},
	{0xA2098FB020167BF4ULL, "dae000_merchant_weapons_dm6151116.cl"},
	{0xA209B7D4ACC958BFULL, "one140cr_charade6134310.cl"},
	{0xA209BF36750B0C40ULL, "6101493_nethany_struggling.cut"},
	{0xA20A599445097F48ULL, "hlac_lt_0_lowtown_n_0_5y.dds"},
	{0xA20A5D3F91ADEC98ULL, "6004159_f.fxe"},
	{0xA20AEDB7BD33798CULL, "vxw_bonepitdestroyed_a3_194_0_0.vlm"},
	{0xA20AF69E76E71883ULL, "709777217.ogg"},
	{0xA20B2C77EDFCE484ULL, "w_woundedcoast_182_0.mmh"},
	{0xA20B3077EDFCEBD7ULL, "w_woundedcoast_182_0.mao"},
	{0xA20B3D2A3ADE44F0ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_min_c.impa.ani"},
	{0xA20B4277EDFD0A26ULL, "w_woundedcoast_182_0.msh"},
	{0xA20B4A7AE16E5F12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v1.ani"},
	{0xA20B70C6F4E1F29DULL, "llac_lt_gallowscourtyard_0_1.dds"},
	{0xA20B8AE35EC6BFE4ULL, "vxht_0_hightown_c7g_0_0.vlm"},
	{0xA20BA64E283F80F6ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_2z.dds"},
	{0xA20C104C1CC876D6ULL, "mag371_elsa6171591.cl"},
	{0xA20C487A6937358BULL, "vxw_sundermount_297_0_0.vlm"},
	{0xA20C66BABBD5A6AAULL, "dae000_arishok6067861.cl"},
	{0xA20CAAEAFA8A810DULL, "zz_followers_banter6159669.cl"},
	{0xA20D16A52BD12FD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_exit.gad"},
	{0xA20D873CC4EA7A6CULL, "6125994_m.fxe"},
	{0xA20DB3716E34FF08ULL, "lt_0_lowtown_docks_n_139_0.mmh"},
	{0xA20DB4AD7FAD2FF1ULL, "vxca_caverns2_7oa_0_0.vlm"},
	{0xA20DBF716E3513ABULL, "lt_0_lowtown_docks_n_139_0.mao"},
	{0xA20DC00A9EEA0BEAULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_movement.anb"},
	{0xA20DD1716E353232ULL, "lt_0_lowtown_docks_n_139_0.msh"},
	{0xA20DF9D8A4BEDE9BULL, "rdr110dg_guard_sleeping6016739.cl"},
	{0xA20E24A2023A8C87ULL, "l_tbe100ar_emporium.bnk"},
	{0xA20E6D53010AF30FULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_cwk.evt"},
	{0xA20EA0F8147D76B7ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp02.msh"},
	{0xA20EB6F8147D9C29ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp02.mmh"},
	{0xA20F45720241376EULL, "fen110_fenris6118074.cl"},
	{0xA20F577FA6108D3BULL, "hlac_w_bonepitmining_a2_327686_0y.dds"},
	{0xA20FA95ECA2A3F1CULL, "6171159_m.fxe"},
	{0xA20FC3B281ED3C63ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.mmh"},
	{0xA21005304E4857EBULL, "vxca_cave_2i7_0_0.vlm"},
	{0xA210491E94EA1EF5ULL, "plt_dae00pt_lightcontent.plo"},
	{0xA21055A0F0512906ULL, "6090317_m.fxe"},
	{0xA2105AB2A04B5572ULL, "vxlt_0_lowtown_a2_n_c_gvc_0_0.vlm"},
	{0xA2105C142E82ABAFULL, "6022740_m.fxe"},
	{0xA210B7CB8B035ED0ULL, "6016483_m.fxe"},
	{0xA210C80FF797CA84ULL, "mcr310_loyalist_mages6076858.cl"},
	{0xA210CF31C7FC79A7ULL, "485007014.ogg"},
	{0xA210DD906926920BULL, "523773927.ogg"},
	{0xA212B8ABEBB58F66ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.cess.ani"},
	{0xA21301482ED245D7ULL, "dre000_varric6028426.cl"},
	{0xA21319A52BD5A009ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_exit.evt"},
	{0xA213B9426DBB36E1ULL, "vxca_caverns1_jfd_0_0.vlm"},
	{0xA213BAE97024C469ULL, "vxca_caverns1_ck2_0_0.vlm"},
	{0xA213D39BD621AB46ULL, "6151021_m.fxe"},
	{0xA213EEB636478E59ULL, "gen_im_trs_gen_sta_nat_r0.uti"},
	{0xA2140E098D56D51CULL, "hlac_lt_0_lowtown_n_131072_0z.dds"},
	{0xA21477AB8DB50E78ULL, "6078804_f.fxe"},
	{0xA214BEC4DDEF8C3FULL, "6207743_m.fxe"},
	{0xA214E082246591EDULL, "vxlt_gallowstemplar_3sm_0_0.vlm"},
	{0xA214ECF609CC85BCULL, "sky_lt_blackemporium.mao"},
	{0xA2155F63ED2AFBDFULL, "252309037.ogg"},
	{0xA2156E53011113B8ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_cwk.gad"},
	{0xA21575D295E80FCAULL, "6187831_m.fxe"},
	{0xA21578812A77EE6EULL, "gen00fl_merrill6127280.cl"},
	{0xA215D1E4D9621CF9ULL, "6103897_f.fxe"},
	{0xA21616CB2E67C5A9ULL, "vxw_bonepit_1u9_0_0.vlm"},
	{0xA2161CAA7FCF9699ULL, "198560859.ogg"},
	{0xA216A024AFD01A26ULL, "dae310ar_lowtown_nt.are"},
	{0xA2172EB806051660ULL, "mrl000_marethari6092372.cl"},
	{0xA2176AFFD330BEC4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blendinside.msh"},
	{0xA21771D9B44CC77BULL, "vxht_mansion2_213_0_0.vlm"},
	{0xA2177CFFD330DD2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blendinside.mao"},
	{0xA21780FFD330E466ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blendinside.mmh"},
	{0xA217AF3659EA7CF3ULL, "6177428_m.fxe"},
	{0xA217C430E31F448FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand.gad"},
	{0xA21801ADCB301BCCULL, "6171807_m.fxe"},
	{0xA21801DF31D744D5ULL, "hlac_lt_0_lowtown_docks_a3_65537_0x.dds"},
	{0xA21834AD7981A9FEULL, "rog_assassinate.dds"},
	{0xA218634AC6FA9EB1ULL, "vxlt_0_lowtown_a3_n_c_i0a_0_0.vlm"},
	{0xA2189FBBFFE1E2F7ULL, "173531968.ogg"},
	{0xA218DBE1D1A64FE4ULL, "llac_ht_0_hightown_a2_-131071_1.dds"},
	{0xA218F5F9056F8D9CULL, "6072214_m.fxe"},
	{0xA21909D23F587290ULL, "20942438.ogg"},
	{0xA2195EB7E2642BF0ULL, "6077802_m.fxe"},
	{0xA21A7ADEBF4120FDULL, "970564848.ogg"},
	{0xA21AC8DE1A07730CULL, "vxca_cavern1_haunted_in3_0_0.vlm"},
	{0xA21AEA15A01DCC5BULL, "6100233_m.fxe"},
	{0xA21B38FD3B6684ADULL, "vxlt_gallowscourtyard_74r_0_0.vlm"},
	{0xA21B8023C4677AEFULL, "mag291_vincento6027114.cl"},
	{0xA21BDB17EB149FE2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_rack.phy"},
	{0xA21C6CB1BFD02179ULL, "648452404.ogg"},
	{0xA21C7B66DE23C80BULL, "6109708_m.fxe"},
	{0xA21CC1C3242DE972ULL, "llac_lt_backalley_n_1_4.dds"},
	{0xA21CEDC12F377669ULL, "vxca_cavern1_haunted_hur_0_0.vlm"},
	{0xA21D31211868EF72ULL, "6116396_m.fxe"},
	{0xA21D3C0C02E4D9B0ULL, "gen_im_trs_gen_lgt_cht_r2.uti"},
	{0xA21D41F82F0FE2C8ULL, "vxca_caverns1_jfa_0_0.vlm"},
	{0xA21DD930E323D43CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand.evt"},
	{0xA21DE030E0FDB25BULL, "pro000_cut2d_thefirstyear6005326.cl"},
	{0xA21E34A947EFD46AULL, "darktown_male_elf_sta6174568.cl"},
	{0xA21E658EF0FD8612ULL, "vxlt_gallowscourtyard_n_4wh_0_0.vlm"},
	{0xA21E968472C8ABEEULL, "vxht_brothel_n_61b_0_0.vlm"},
	{0xA21E9A172768772DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_showcase_d.dds"},
	{0xA21EC72678692BBEULL, "6039514_m.fxe"},
	{0xA21F48E8F034294AULL, "vxht_0_hightown_a2_9zt_0_0.vlm"},
	{0xA21F93D99E513C73ULL, "one220_yevhen6135860.cl"},
	{0xA21F9C384F7708DFULL, "gen00fl_isabela6079658.cl"},
	{0xA21FA77FA59FF4AAULL, "vxca_cavern1_haunted_j24_0_2.vlm"},
	{0xA2203163CF95555BULL, "qun120_arishok6051910.cl"},
	{0xA22091C1A0B3F436ULL, "6107505_m.fxe"},
	{0xA220A0ABEBC1A3D7ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.cess.evt"},
	{0xA220EA1FE1F3C6F0ULL, "6132640_m.fxe"},
	{0xA22101A52BE1B578ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_exit.ani"},
	{0xA22159D246495657ULL, "6043635_m.fxe"},
	{0xA2215B90478113EEULL, "vxw_encounter2_gw_0_0.vlm"},
	{0xA22177FFD3392931ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_blendinside.phy"},
	{0xA221B685E2D0DE99ULL, "lgt261ar_rescue_cave.lst"},
	{0xA221BA929F8F7AA5ULL, "hlac_w_blightlands_0_1y.dds"},
	{0xA2227E234FCE2CE9ULL, "llac_lt_gallowscourtyard_-65537_0.dds"},
	{0xA222E3EC45C350F0ULL, "vxht_mansion2b_2je_0_0.vlm"},
	{0xA2237E286E76194CULL, "86215004.ogg"},
	{0xA223920E7FFE34C7ULL, "vxdr_primevalentrance_335_0_0.vlm"},
	{0xA223B2991C11B36AULL, "6206068_m.fxe"},
	{0xA22436C20BD4A6E3ULL, "635230273.ogg"},
	{0xA224601917EDB848ULL, "806674380.ogg"},
	{0xA2249213D9D2BF6BULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_corruptlyrium_d.dds"},
	{0xA225F3F4BFCF82C1ULL, "202271340.ogg"},
	{0xA2265AAFA292F984ULL, "hlac_lt_gallowstemplar_2382_3x.dds"},
	{0xA22672C91DC9A8D0ULL, "dae011_wench_mob6149865.cl"},
	{0xA226FDAA7883E9AFULL, "538569265.ogg"},
	{0xA22724AE2C9803E1ULL, "plt_ave100pt_promotion.nss"},
	{0xA2277B93387AE64AULL, "vxlt_backalley_n_1e5_0_0.vlm"},
	{0xA227A1ABEBC7C460ULL, "art\\vfx\\combat\\vfx_r_gn2_impact_bl_loop_c.cess.gad"},
	{0xA227EF06A98A831DULL, "6095528_m.fxe"},
	{0xA2280BD2966B670EULL, "hlac_lt_0_lowtown_docks_-196611_1z.dds"},
	{0xA228EA52BF2495EEULL, "hlac_w_sundermount_131066_0x.dds"},
	{0xA2294EE6F0A773FBULL, "gen00fl_fenris6127131.cl"},
	{0xA22A1610EA3B85FBULL, "hlac_lt_0_lowtown_a2_131070_2x.dds"},
	{0xA22AD4A7C2DAB571ULL, "static_static_p39.psh.fxo"},
	{0xA22AF28FBC974144ULL, "hlac_w_blightlands_131075_0x.dds"},
	{0xA22B4BE5858FDB64ULL, "vrc321_varric6094460.cl"},
	{0xA22B62C4A398F51AULL, "dre000_vampire.cnv"},
	{0xA22B6D189D2F963AULL, "llac_w_sundermount_-327677_0.dds"},
	{0xA22B78101FCD03F0ULL, "486840598.ogg"},
	{0xA22BE130E3301F55ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_sweep_right_hand.ani"},
	{0xA22C637269339BF5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\hf_paperdoll.anb"},
	{0xA22C700EA1DC4FD7ULL, "hlac_lt_0_lowtown_a2_-3_6z.dds"},
	{0xA22C7CD42E18E062ULL, "llac_dr_slavers_3217_10.dds"},
	{0xA22CF6890C59E04EULL, "6081089_m.fxe"},
	{0xA22D9D2C6984F6AFULL, "122156172.ogg"},
	{0xA22DD68E645E0853ULL, "488494235.ogg"},
	{0xA22E84C4A39B6562ULL, "dre000_vampire.bnk"},
	{0xA22E8855D0078854ULL, "art\\characters\\creatures\\dragonling\\c_dragonling_eyes.mao"},
	{0xA22F0D0B1588D615ULL, "gen00fl_varric6099395.cl"},
	{0xA22F3A8B6B975D82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_mage_ready.anb"},
	{0xA230497ACDDB135BULL, "art\\vfx\\models\\vfx_rockarmor_rs.mmh"},
	{0xA23064E0042917ACULL, "6095806_m.fxe"},
	{0xA230677ACDDB4645ULL, "art\\vfx\\models\\vfx_rockarmor_rs.msh"},
	{0xA230A173E7EA87B8ULL, "gen_im_trs_cot_med_hlm_r1.uti"},
	{0xA231BA8D137EEAD2ULL, "6129206_m.fxe"},
	{0xA231EFDD80D8FCC9ULL, "20992966.ogg"},
	{0xA231FA918C96B6EDULL, "571093538.ogg"},
	{0xA23230706A9F983FULL, "6187752_f.fxe"},
	{0xA23234630BC60100ULL, "hlac_lt_undercitydungeon_-1_38z.dds"},
	{0xA23248733822E2A3ULL, "war_holysmite.dds"},
	{0xA232AF85E2DF465EULL, "lgt261ar_rescue_cave.are"},
	{0xA232EBD1B0DE8AE0ULL, "207554559.ogg"},
	{0xA2331A37498A6768ULL, "vxw_woundedcoast_1fr_0_0.vlm"},
	{0xA233B307DB48536DULL, "hlac_w_woundedcoast_327683_0y.dds"},
	{0xA2344346049C876DULL, "6127416_m.fxe"},
	{0xA23443A46FF9F989ULL, "500358017.ogg"},
	{0xA23469084F323D89ULL, "vxlt_0_lowtown_a3_n_c_eqj_0_0.vlm"},
	{0xA23476A8AC91B932ULL, "6092470_m.fxe"},
	{0xA234A67DC7DF83CFULL, "vxlt_warehouse_n_az_0_0.vlm"},
	{0xA2353E0B15715656ULL, "vxca_caverns2_77r_0_0.vlm"},
	{0xA2356879EA8DF6BEULL, "vxlt_0_lowtown_a2_n_erw_0_0.vlm"},
	{0xA23591886102A366ULL, "llac_lt_0_lowtown_docks_a3_-65535_0.dds"},
	{0xA23683626EE124ECULL, "vxlt_gallowscourtyard_6yq_0_0.vlm"},
	{0xA236B8D712A20D17ULL, "387166286.ogg"},
	{0xA236EA2D96EE4C66ULL, "vxlt_0_lowtown_docks_a3_cum_0_0.vlm"},
	{0xA2370B092DD08E23ULL, "follower_banter6172811.cl"},
	{0xA2382F4BE32E044AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_enter.gad"},
	{0xA2388F6A1102E9A5ULL, "1045630541.ogg"},
	{0xA238F21FC2353E8DULL, "754011168.ogg"},
	{0xA2393AF9952C6BF8ULL, "6153137_m.fxe"},
	{0xA2393E03AFC064BCULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v2.ani"},
	{0xA239752AAB9547D9ULL, "art\\characters\\creatures\\strider\\animation\\dazed.ani"},
	{0xA2397C3B45D38410ULL, "vxlt_0_lowtown_a3_n_c_hjc_0_0.vlm"},
	{0xA23B10913218736FULL, "tal_thre_03.mao"},
	{0xA23B5A1B7F6FF23EULL, "266104375.ogg"},
	{0xA23B8FAFCC761C46ULL, "vxw_bonepitmining_a2_gi_0_0.vlm"},
	{0xA23BA724AFEF4211ULL, "dae310ar_lowtown_nt.lst"},
	{0xA23BCA29B0BACCF1ULL, "91431500.ogg"},
	{0xA23C4963D3F7B3FEULL, "130670043.ogg"},
	{0xA23C4BBB1AEEDC10ULL, "vxw_encounter2_r6_0_0.vlm"},
	{0xA23C5FBCECACC972ULL, "one100dg_nabil6176027.cl"},
	{0xA23C9F3109346ADFULL, "342304195.ogg"},
	{0xA23D0B96282DE0BFULL, "face_face_p8.psh.fxo.dx11.assembly"},
	{0xA23D6AE7450AF8AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ea_transition.evt"},
	{0xA23DA5846A005F6AULL, "vxw_bonepit_1rs_0_0.vlm"},
	{0xA23DB52ACFEF766EULL, "vxht_0_hightown_a2_as3_0_0.vlm"},
	{0xA23DEA73E650644DULL, "6101364_m.fxe"},
	{0xA23E0C2244BDB398ULL, "6068103_m.fxe"},
	{0xA23E21746DCB47A5ULL, "dae000ip_crowd_mf6_talk12.utp"},
	{0xA23ED33828DE276FULL, "6149609_m.fxe"},
	{0xA23F124BE333F1E1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_enter.evt"},
	{0xA23F35A4EA1D375DULL, "43027267.ogg"},
	{0xA23F4A6026BC40ADULL, "6108136_m.fxe"},
	{0xA23F594DD1E8645FULL, "vxht_chantry_n_272_0_0.vlm"},
	{0xA2404B03AFC699FEULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v2.gad"},
	{0xA240502AAB9B271BULL, "art\\characters\\creatures\\strider\\animation\\dazed.gad"},
	{0xA2406CA092A86DA7ULL, "vxlt_0_lowtown_docks_b81_0_0.vlm"},
	{0xA2407BEE51ED407DULL, "6201545_m.fxe"},
	{0xA240B617EB337C96ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_rack.mao"},
	{0xA240C217EB339135ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_rack.mmh"},
	{0xA240C417EB3394ABULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_rack.msh"},
	{0xA240D001364FC29BULL, "6127128_m.fxe"},
	{0xA240D94AD0ABDD24ULL, "ca_caverns1_mines_26035.rml"},
	{0xA240EDA1B6C499BEULL, "s_blood_sh.mao"},
	{0xA241149BDAA62779ULL, "199924951.ogg"},
	{0xA2411C3BCBBEE2A6ULL, "hlac_lt_gallowstemplar_n_7198_0x.dds"},
	{0xA24123DE8F15B9DFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg05.msh"},
	{0xA24129DE8F15C381ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg05.mmh"},
	{0xA2418C143B807FE8ULL, "vxw_woundedcoast_1as_0_0.vlm"},
	{0xA241AFB9D1A7ECC5ULL, "genip_node_e1.utp"},
	{0xA241D95172A9D006ULL, "vxht_mansion_n_4w5_0_0.vlm"},
	{0xA24275029215C2ACULL, "vxlt_hangedman_2z7_0_0.vlm"},
	{0xA24294D20B8B23E6ULL, "vxw_woundedcoast_47_0_0.vlm"},
	{0xA242BF24AFF589A8ULL, "dae310ar_lowtown_nt.ncs"},
	{0xA243027FED17C9DCULL, "tbe100st_magic_a2.utm"},
	{0xA24325FAD81898EAULL, "hlac_w_bonepitdestroyed_a3_327682_0x.dds"},
	{0xA243AD2C5EB79C73ULL, "zz_fol_debug6096567.cl"},
	{0xA243DA56DC840F3EULL, "vxlt_0_lowtown_a2_n_ewh_0_0.vlm"},
	{0xA24420430AD5BD5FULL, "hm_wilmod.mor"},
	{0xA2445F4D1C67174AULL, "vxlt_backalley_n_1k7_0_0.vlm"},
	{0xA24461E745110859ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ea_transition.gad"},
	{0xA2455C0A88E464E6ULL, "vxlt_0_lowtown_docks_a3_cxn_0_0.vlm"},
	{0xA245F20AE213690DULL, "vxlt_foundry_n_yb_0_0.vlm"},
	{0xA2467DC5F7E9C2BEULL, "65782822.ogg"},
	{0xA24682CDAEA807AEULL, "vxca_caverns1_mines_jdd_0_0.vlm"},
	{0xA246E82F18A75C52ULL, "mag131_keran6088226.cl"},
	{0xA246FD951C6A9805ULL, "354947487.ogg"},
	{0xA2471D1A157D5978ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_firestatue_n.dds"},
	{0xA2472603AFCC791DULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v2.evt"},
	{0xA2473B8AD4D9A59EULL, "hlac_w_encounter2_camp_262147_0z.dds"},
	{0xA2473CBA4F2D387AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockac.phy"},
	{0xA24762C99F652BB4ULL, "6095257_m.fxe"},
	{0xA2476D2AABA17700ULL, "art\\characters\\creatures\\strider\\animation\\dazed.evt"},
	{0xA247F1027A40C26BULL, "6071888_f.fxe"},
	{0xA2488CDBF54A0E83ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_diningtable_sa.dds"},
	{0xA248C47E88B136D7ULL, "6047378_m.fxe"},
	{0xA248DFA713BBF6FFULL, "6103368_m.fxe"},
	{0xA2499D457862F0BDULL, "mag321_comtesse_de_launcet6171525.cl"},
	{0xA249DDD683B2D6E3ULL, "gen00fl_fenris6112078.cl"},
	{0xA24A1CB6EA53662EULL, "6127024_m.fxe"},
	{0xA24AA3AB4A16508FULL, "579400176.ogg"},
	{0xA24ADABE951A43DCULL, "art\\vfx\\combat\\vfx_w_jug_smash_act_c.impa.ani"},
	{0xA24B122C7BDC6420ULL, "vxlt_0_lowtown_n_bxt_0_0.vlm"},
	{0xA24B72E7451743A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ea_transition.ani"},
	{0xA24B82CB53F07BE9ULL, "dae000_female_tranquil6165142.cl"},
	{0xA24BB861816168A5ULL, "1019289418.ogg"},
	{0xA24BFB05CB4D9C02ULL, "vrc321tr_varric_crazy.ncs"},
	{0xA24C18B815A18607ULL, "zz_seb_debug6215787.cl"},
	{0xA24C2707C1E58EEEULL, "vxht_0_hightown_a2_an4_0_0.vlm"},
	{0xA24C3D94287FA1A8ULL, "w_blightlands_32_0.msh"},
	{0xA24C3EB49EFF4FADULL, "vxht_0_hightown_a3_cbi_0_1.vlm"},
	{0xA24C4F94287FC039ULL, "w_blightlands_32_0.mao"},
	{0xA24C5B94287FD4D2ULL, "w_blightlands_32_0.mmh"},
	{0xA24CE5F0F1270EFBULL, "vxlt_gallowscourtyard_77y_0_0.vlm"},
	{0xA24D0A4BE3402130ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_enter.ani"},
	{0xA24D3C1E951DB6D2ULL, "plt_dae00pt_lightcontent.nss"},
	{0xA24D4E405779CFBFULL, "702102524.ogg"},
	{0xA24DB12BF9A34FF2ULL, "6153513_m.fxe"},
	{0xA24E0A61EE1F2389ULL, "927603262.ogg"},
	{0xA24E88B01D6934C5ULL, "6172483_m.fxe"},
	{0xA24E8907CCBBE2D5ULL, "6069435_m.fxe"},
	{0xA24E9238A2688B69ULL, "vxw_encounter2_c0_0_0.vlm"},
	{0xA24E9F4788AE923BULL, "6051801_m.fxe"},
	{0xA24ED9E7B90E96CAULL, "6048392_m.fxe"},
	{0xA24EF67314EFCA3FULL, "qun221_petrice6077344.cl"},
	{0xA24F694994B75A07ULL, "hlac_ca_cavern1_haunted_24424_3z.dds"},
	{0xA24F8EF0EC77D3E1ULL, "6083111_f.fxe"},
	{0xA24FB65DE4AC811CULL, "314799465.ogg"},
	{0xA24FF66934FB36C2ULL, "hlac_lt_0_lowtown_a2_65536_2x.dds"},
	{0xA24FF7CD9F2664D1ULL, "6085679_m.fxe"},
	{0xA24FFDF12D769868ULL, "vxw_woundedcoast_1lp_0_0.vlm"},
	{0xA2500382023FB155ULL, "cod_hst_anderfels.utp"},
	{0xA250195AA0936269ULL, "6034370_m.fxe"},
	{0xA25020C8449A977EULL, "showstrid.ncs"},
	{0xA25054E040024EA0ULL, "6000747_m.fxe"},
	{0xA2508949651AB29FULL, "vxlt_0_lowtown_f13_0_0.vlm"},
	{0xA250E6DF840BDB2CULL, "vxlt_hangedman_2w4_0_0.vlm"},
	{0xA250E7BE951EC51EULL, "art\\vfx\\combat\\vfx_w_jug_smash_act_c.impa.gad"},
	{0xA250F189EFE429CAULL, "ch2notnt_ds2y.psh.fxo"},
	{0xA25106AEFD813C66ULL, "vxw_woundedcoast_18_0_0.vlm"},
	{0xA251845A34CA755BULL, "hlac_lt_0_lowtown_docks_a3_-2_3x.dds"},
	{0xA25187F4A7AE7F78ULL, "vxht_0_hightown_a2_bnt_0_1.vlm"},
	{0xA251949B7D2E3114ULL, "isa311_castillon6095039.cl"},
	{0xA251DD2451C407ABULL, "1066092845.ogg"},
	{0xA252372273561292ULL, "6111986_m.fxe"},
	{0xA252475F6C0771C2ULL, "vxlt_0_lowtown_a3_n_c_hnu_0_0.vlm"},
	{0xA2524C33CE7A27BEULL, "vxlt_0_lowtown_a2_n_exi_0_0.vlm"},
	{0xA25289407EEBE240ULL, "llac_lt_0_lowtown_docks_n_-65539_0.dds"},
	{0xA252BE7DF70C47DEULL, "836364028.ogg"},
	{0xA252E15CAC414CF9ULL, "hlac_lt_gallowscourtyard_-131071_3x.dds"},
	{0xA252E2AE567532B9ULL, "ht_0_hightown_a3_11_0.rml"},
	{0xA25312A01EA9AE50ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_6y.dds"},
	{0xA2534FBE6F871D84ULL, "945041739.ogg"},
	{0xA2535F8B10C3F00BULL, "vxw_blightlands_gi_0_0.vlm"},
	{0xA2539F44C8753561ULL, "975878252.ogg"},
	{0xA2541A0951A25608ULL, "vxht_mansion_n_6jz_0_0.vlm"},
	{0xA254565D36554218ULL, "lt_gallowstemplar_n_mmi.gff"},
	{0xA25463E7D409818DULL, "vxlt_foundry_n_ta_0_0.vlm"},
	{0xA25475A8FDA1721AULL, "6054227_m.fxe"},
	{0xA254BDC6A7CD4D88ULL, "6164742_m.fxe"},
	{0xA254F687B88AF777ULL, "6093551_m.fxe"},
	{0xA255206864722CC8ULL, "1056648.ogg"},
	{0xA255EA898AE6A168ULL, "gen00fl_merrill6147099.cl"},
	{0xA25693C5931C47E3ULL, "hlac_lt_warehouse_n_65537_4y.dds"},
	{0xA257C2BE9524A53DULL, "art\\vfx\\combat\\vfx_w_jug_smash_act_c.impa.evt"},
	{0xA258C9C3A444BA9EULL, "827407309.ogg"},
	{0xA259C6582D801946ULL, "hlac_w_encounter2_-131067_0y.dds"},
	{0xA25A0CC22362AE0FULL, "vxlt_blackemporium_3p_0_1.vlm"},
	{0xA25A305925376DB5ULL, "hlac_lt_0_lowtown_a2_n_-65536_8y.dds"},
	{0xA25A46EF7103F752ULL, "mcr370_meredith6083057.cl"},
	{0xA25A7810AAE82CB7ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_pc_mcr370_playerflees.gad"},
	{0xA25A7909013C6934ULL, "588264148.ogg"},
	{0xA25A98E4B3DBA76EULL, "vxht_0_hightown_a2_ai5_0_0.vlm"},
	{0xA25ABA546A7BADEFULL, "vxlt_gallowscourtyard_n_73z_0_0.vlm"},
	{0xA25B3E86006791A9ULL, "37200223.ogg"},
	{0xA25B7C13BC0C482AULL, "250458226.ogg"},
	{0xA25BDE19A99DD6DEULL, "ss_caress.bnk"},
	{0xA25C0CB2A1581881ULL, "971074375.ogg"},
	{0xA25C27CA508808A6ULL, "vxlt_0_lowtown_f4m_0_0.vlm"},
	{0xA25C478F10582E5FULL, "mer140_bonepit_comments6062988.cl"},
	{0xA25C7E9B8085EA6FULL, "mag100_thrask6032002.cl"},
	{0xA25D0415945EA3E9ULL, "vxw_encounter2_n7_0_0.vlm"},
	{0xA25D2D16140B62CEULL, "714987365.ogg"},
	{0xA25D39D13A662390ULL, "mcr370_cut_marriage6098256.cl"},
	{0xA25D4BC4AB04AF4AULL, "6043292_m.fxe"},
	{0xA25D6BF650134C37ULL, "6096314_m.fxe"},
	{0xA25D7995C8D6E96AULL, "one280_merin6116970.cl"},
	{0xA25DCE60DA0DCDF1ULL, "llac_lt_backalley_n_0_19.dds"},
	{0xA25E232BDB622DF5ULL, "6171525_m.fxe"},
	{0xA25E3BA1697722DEULL, "hlac_lt_0_lowtown_a2_n_c_-65535_3x.dds"},
	{0xA25EBD0B56960106ULL, "vxht_mansion_n_4a7_0_0.vlm"},
	{0xA25F2A510431268FULL, "324671273.ogg"},
	{0xA25F2F7C47CE6692ULL, "llac_lt_0_lowtown_n_-262144_4.dds"},
	{0xA25F6974B3418B32ULL, "gen00fl_varric6021133.cl"},
	{0xA25F7766B17AA2B1ULL, "vxlt_0_lowtown_a3_n_c_f66_0_2.vlm"},
	{0xA25FE00C3C30DF0DULL, "w_encounter3_165_0.msh"},
	{0xA25FEE0C3C30F740ULL, "w_encounter3_165_0.mao"},
	{0xA25FF20C3C30FD93ULL, "w_encounter3_165_0.mmh"},
	{0xA260B93C5DFD8A42ULL, "vxlt_0_lowtown_a3_n_c_hkv_0_0.vlm"},
	{0xA2612025C91F2439ULL, "502359445.ogg"},
	{0xA2617CC4A3C6B16DULL, "dre000_vampire.stg"},
	{0xA261D16802BA088BULL, "vxw_blightlands_hh_0_0.vlm"},
	{0xA261F361565E40F4ULL, "llac_lt_gallowscourtyard_n_65536_0.dds"},
	{0xA2628AFA54D3E761ULL, "vxw_sundermount_30j_0_1.vlm"},
	{0xA263AD7A7FE016E0ULL, "6153559_m.fxe"},
	{0xA263D2322D3150D2ULL, "890790012.ogg"},
	{0xA2647B5B91B712D4ULL, "vxlt_undercitydungeon_yf_0_0.vlm"},
	{0xA264AE4F1973A00CULL, "vxlt_0_lowtown_a2_byt_0_0.vlm"},
	{0xA264B71E397C281DULL, "vxht_keep_7ah_0_0.vlm"},
	{0xA2651CF946C96248ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_0.pwk"},
	{0xA2655B0E6CA53792ULL, "6091888_m.fxe"},
	{0xA265A789335A84E2ULL, "6120826_f.fxe"},
	{0xA265ADD2BF2307BDULL, "130601949.ogg"},
	{0xA2662F97EC0B46B1ULL, "vxlt_0_lowtown_docks_a3_d3g_0_0.vlm"},
	{0xA266890FD16DD61DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_walls_r.dds"},
	{0xA266D561DFD5A5C9ULL, "vxlt_hangedman_n_2wz_0_0.vlm"},
	{0xA266DEF3D238E764ULL, "dae000_bartrand6206867.cl"},
	{0xA266ED0CD9166E7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\anders_staff_idle.ani"},
	{0xA266EE1B1487A3DDULL, "6072884_m.fxe"},
	{0xA267F447D8A0E18FULL, "6208341_f.fxe"},
	{0xA26815B5831CCA55ULL, "6178423_m.fxe"},
	{0xA2690FCA4184586DULL, "vxht_0_hightown_a2_amv_0_0.vlm"},
	{0xA26926943C7328B5ULL, "vxca_caverns2_89l_0_0.vlm"},
	{0xA2692D91D6A49856ULL, "mag100_karras6034549.cl"},
	{0xA2692FEBB6B04160ULL, "502078202.ogg"},
	{0xA2698FE451D967DAULL, "hlac_lt_0_lowtown_n_2_0z.dds"},
	{0xA26A08B6543BF32AULL, "follower_banter6170421.cl"},
	{0xA26A1FCF61D08529ULL, "vxca_cavern1_haunted_cos_0_0.vlm"},
	{0xA26A335D03844B04ULL, "hlac_ht_brothel_n_2995_0z.dds"},
	{0xA26A3DC650EFA696ULL, "vxw_bonepitdestroyed_a3_25q_0_0.vlm"},
	{0xA26AC67F200D8373ULL, "396523773.ogg"},
	{0xA26AE092F8374810ULL, "1024090789.ogg"},
	{0xA26AE177C8A3FC5AULL, "114164733.ogg"},
	{0xA26B0BF9ED3AB89BULL, "dae000_flemeth6000883.cl"},
	{0xA26B10B48422DD13ULL, "vxlt_warehouse_n_b7_0_0.vlm"},
	{0xA26B516BE7B01690ULL, "lowtown_fem_prostitute_sta6152358.cl"},
	{0xA26C34CF1720602FULL, "314441893.ogg"},
	{0xA26CE1AB1162C968ULL, "vxw_woundedcoast_1rn_0_0.vlm"},
	{0xA26D2ADE8F3B00FEULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg05.phy"},
	{0xA26D8F6EAC16F50DULL, "463278584.ogg"},
	{0xA26E11E9F83D3211ULL, "dae000_arishok6072862.cl"},
	{0xA26E18A89CDC11DDULL, "lowtown_male_urchin_sta.bnk"},
	{0xA26E3AB436AB6E43ULL, "gen00fl_aveline6115867.cl"},
	{0xA26EA510AAF9223DULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_pc_mcr370_playerflees.ani"},
	{0xA26EC3127D778F02ULL, "271126458.ogg"},
	{0xA26F1C4B2AAAF103ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2win.mao"},
	{0xA26F204B2AAAF850ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2win.mmh"},
	{0xA26F2E4B2AAB106AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2win.msh"},
	{0xA26FCEBAC3559432ULL, "isa211_party_reaction6202329.cl"},
	{0xA270CE1CDD6F24B1ULL, "vxca_caverns1_dr6_0_0.vlm"},
	{0xA270FAA89CDE14E5ULL, "lowtown_male_urchin_sta.cnv"},
	{0xA271050E0E4CD634ULL, "vxlt_undercityserialkiller_1lq_0_0.vlm"},
	{0xA2714C6FDBFF5FFCULL, "dae000_pc_home_ambients6208673.cl"},
	{0xA2716A1D24AF6244ULL, "vxca_caverns2_6he_0_1.vlm"},
	{0xA271ED0DEEB22400ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_0x.dds"},
	{0xA2721F5771D62F60ULL, "6156659_m.fxe"},
	{0xA2725E76E9615569ULL, "vxlt_undercitydungeon_1mv_0_0.vlm"},
	{0xA2729F38D8C4327FULL, "one110_pryce6088395.cl"},
	{0xA2734E7C0A596B52ULL, "gen00fl_aveline6137464.cl"},
	{0xA2743D93C5E784F6ULL, "llac_dr_dungeon_3215_3.dds"},
	{0xA2749F9DD41E4CE4ULL, "vxht_brothel_n_5g1_0_0.vlm"},
	{0xA274A174DE015F31ULL, "vxlt_0_lowtown_docks_a3_d4f_0_0.vlm"},
	{0xA275CC57970CD0D8ULL, "vxlt_0_lowtown_docks_c58_0_0.vlm"},
	{0xA27687927512E2D5ULL, "6173523_m.fxe"},
	{0xA27698061AF1226CULL, "504610006.ogg"},
	{0xA276AB6D6B5D8E4EULL, "6106656_m.fxe"},
	{0xA276F06763DA680DULL, "219290118.ogg"},
	{0xA27867B3319CDC47ULL, "6188660_m.fxe"},
	{0xA2786BC65A39156DULL, "art\\characters\\weapons\\staff\\w_stf_005a.msh"},
	{0xA27879C65A392CA0ULL, "art\\characters\\weapons\\staff\\w_stf_005a.mao"},
	{0xA2787DC65A3933F3ULL, "art\\characters\\weapons\\staff\\w_stf_005a.mmh"},
	{0xA278FB0C90277CFEULL, "gen00fl_anders6116202.cl"},
	{0xA279FA00AE8037A9ULL, "6143699_f.fxe"},
	{0xA27A330574F21A84ULL, "881306608.ogg"},
	{0xA27AF53AECAF0705ULL, "53292723.ogg"},
	{0xA27AFE6B7DD9988BULL, "70147118.ogg"},
	{0xA27B09C900086FD1ULL, "w_bonepitdestroyed_a3_22_0_ll.mmh"},
	{0xA27B140CD9275905ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\anders_staff_idle.gad"},
	{0xA27B1D6B0506C1FFULL, "fx_0033_diff.dds"},
	{0xA27BA751DE7E0D34ULL, "92603545.ogg"},
	{0xA27BF4B2B0331CB8ULL, "6077351_f.fxe"},
	{0xA27C28CCDC553224ULL, "6110812_f.fxe"},
	{0xA27C7803162BDABFULL, "follower_banter6143244.cl"},
	{0xA27CA48C8ACC773AULL, "binknoalpha.psh.fxo"},
	{0xA27EB521E6A6398BULL, "vxw_blightlands_vj_0_0.vlm"},
	{0xA27EF1A05C91D2D7ULL, "154789256.ogg"},
	{0xA27FB10A00059336ULL, "hlac_ht_mansion2_818_32y.dds"},
	{0xA27FFEA518B239A9ULL, "ss_torpor.bnk"},
	{0xA2804BC2876DC5EEULL, "llac_lt_0_lowtown_-2_6.dds"},
	{0xA2809CF1A69F4542ULL, "art\\vfx\\combat\\vfx_r_var_pinpoint_str_c.impa.gad"},
	{0xA2809D336C57AF30ULL, "6080581_m.fxe"},
	{0xA280D4C68A440D1EULL, "6055167_m.fxe"},
	{0xA2810959929DFDE5ULL, "hlac_ht_chantry_2534_44z.dds"},
	{0xA2813E743B23FE46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_no_way.ani"},
	{0xA2819B0C3D070757ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drgb_0i.dds"},
	{0xA281A62F74C19681ULL, "gen00fl_anders6096518.cl"},
	{0xA282018B21848245ULL, "6142435_m.fxe"},
	{0xA282347F353E4B8DULL, "733652360.ogg"},
	{0xA282609098A7FD46ULL, "967657578.ogg"},
	{0xA28272329A37A433ULL, "535554718.ogg"},
	{0xA282A587741A38B2ULL, "6063845_m.fxe"},
	{0xA282EC177F4E9C76ULL, "llac_lt_0_lowtown_a3_n_c_-131073_3.dds"},
	{0xA28307D5D3295B09ULL, "vxlt_0_lowtown_docks_b76_0_0.vlm"},
	{0xA283117AC6146564ULL, "vxht_brothel_n_5j6_0_0.vlm"},
	{0xA2838C3B4073BEC7ULL, "vxlt_gallowstemplar_39m_0_0.vlm"},
	{0xA283B91BC3C1D6C9ULL, "vxlt_hangedman_n_2yx_0_0.vlm"},
	{0xA283C01F57077A7BULL, "411144389.ogg"},
	{0xA283D9B4A8BE78B9ULL, "6059227_m.fxe"},
	{0xA283DE31EAADD037ULL, "604616996.ogg"},
	{0xA2840B3AF6DDF5B6ULL, "vxw_sundermount_33l_0_1.vlm"},
	{0xA2847E8AF5E6CB24ULL, "vxht_mansion_n_5ah_0_0.vlm"},
	{0xA2849012EB901D2BULL, "vxca_caverns1_mines_dp7_0_0.vlm"},
	{0xA284B994952BC1D2ULL, "6179940_m.fxe"},
	{0xA2853643E211C32EULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgtb.phy"},
	{0xA2853D221E79E315ULL, "440706356.ogg"},
	{0xA285E4D6396113B8ULL, "mag320_noblewoman_26060282.cl"},
	{0xA2863872A5843044ULL, "599519086.ogg"},
	{0xA2869B91636FC9A1ULL, "mag101_grace6035867.cl"},
	{0xA286B8B5FE1AF1F4ULL, "6090449_m.fxe"},
	{0xA2872EB0F39CAA0DULL, "640461262.ogg"},
	{0xA2879B5D3F042FDCULL, "vxlt_darktown_47s_0_0.vlm"},
	{0xA287BFF1A6A59F39ULL, "art\\vfx\\combat\\vfx_r_var_pinpoint_str_c.impa.evt"},
	{0xA288A6FA645B59FFULL, "gen_im_acc_rng_act2_07.uti"},
	{0xA288ABA41EB425EDULL, "681167529.ogg"},
	{0xA288C5B5A90D4545ULL, "follower_banter6171139.cl"},
	{0xA289A10D012C47E3ULL, "278941981.ogg"},
	{0xA289A52B52B5B5C9ULL, "t1_skn_d05.tnt"},
	{0xA28A24319165ECAAULL, "hlac_ht_mansion2_818_5y.dds"},
	{0xA28A2FEA2D971AC0ULL, "139303984.ogg"},
	{0xA28A73D9DED1C701ULL, "6134540_m.fxe"},
	{0xA28AC59EC27B3853ULL, "vxlt_warehouse_11f_0_0.vlm"},
	{0xA28AD0417F863160ULL, "6087009_m.fxe"},
	{0xA28B0C25BE09059DULL, "ch1_ds1y.psh.fxo"},
	{0xA28B4AEE7529F608ULL, "hlac_ht_0_hightown_a2_-65535_0x.dds"},
	{0xA28BD3EC995DF3BEULL, "vxlt_gallowstemplar_f_34d_0_0.vlm"},
	{0xA28BDBB38876EAE9ULL, "vxca_caverns1_gko_0_1.vlm"},
	{0xA28C659BC01B60D5ULL, "zz_ser_debug6091916.cl"},
	{0xA28CE7B8D8856FE4ULL, "hlac_ht_brothel_n_3373_1z.dds"},
	{0xA28CF4B639E2F225ULL, "vxdr_primevaltomb_7so_0_0.vlm"},
	{0xA28CF4DBB5546068ULL, "vxw_woundedcoast_wq_0_1.vlm"},
	{0xA28D2688AA50C763ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.impa.evt"},
	{0xA28E0EEE9BAA3F59ULL, "vxlt_0_lowtown_ha7_0_0.vlm"},
	{0xA28E9FFFF9EBE5BFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_en.gad"},
	{0xA28EAED848392FA9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopen.gad"},
	{0xA28F0CB3E43904CAULL, "495283003.ogg"},
	{0xA28F26743B3012B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_no_way.evt"},
	{0xA28F284FC8CA5445ULL, "6015865_m.fxe"},
	{0xA28F4D1AAB2A13CFULL, "vxlt_0_lowtown_docks_n_b50_0_1.vlm"},
	{0xA28F5D6E7CD1977EULL, "6018533_m.fxe"},
	{0xA28FB7F364A061AFULL, "6096219_m.fxe"},
	{0xA28FC4CBB9936196ULL, "558708786.ogg"},
	{0xA28FCD1C7A99E0ADULL, "6083985_m.fxe"},
	{0xA28FE8BB4F99FF18ULL, "plt_cod_cha_merrill2.plo"},
	{0xA28FF31BE19CB9E3ULL, "dre000_varric6036955.cl"},
	{0xA2905FFDFA857726ULL, "764960460.ogg"},
	{0xA290613199C5E2DAULL, "one120_athenril6113288.cl"},
	{0xA29079417F200882ULL, "vxdr_0_deeproads_f0r_0_0.vlm"},
	{0xA2909E26F8C30B88ULL, "6176510_m.fxe"},
	{0xA2910DC87BBE6E06ULL, "clf360cr_evets_archer.utc"},
	{0xA29152AA48CD40BBULL, "hlac_ca_cavern1_haunted_24426_1z.dds"},
	{0xA291CCBA1C399168ULL, "378521651.ogg"},
	{0xA291E57796132DBAULL, "vxlt_gamlenshouse_n_b8_0_0.vlm"},
	{0xA2924F8CA1F81C0EULL, "llac_lt_foundry_n_320_21.dds"},
	{0xA29252BC88DAE1A4ULL, "vxlt_0_lowtown_a3_n_c_d8u_0_0.vlm"},
	{0xA2925527CDB860C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\blendtree\\trooper_movement.anb"},
	{0xA292A2D18F6358ABULL, "vxht_mansion2_21c_0_0.vlm"},
	{0xA293974CDEC38D15ULL, "qun221cr_petrice_mob2.utc"},
	{0xA293DF8B902D25E7ULL, "vxw_bonepit_1wl_0_1.vlm"},
	{0xA293FA7696BC501AULL, "ser221_party_comments6164314.cl"},
	{0xA293FF88AA56A3E4ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.impa.gad"},
	{0xA29410C05889ACEAULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_infa_0d.dds"},
	{0xA294465483CFD6D2ULL, "vxlt_warehouse_11c_0_0.vlm"},
	{0xA294C7F1A6B03728ULL, "art\\vfx\\combat\\vfx_r_var_pinpoint_str_c.impa.ani"},
	{0xA295441874FC04ECULL, "vxht_keep_6y8_0_0.vlm"},
	{0xA2954711A3C749B3ULL, "lt_0_lowtown_docks_a3_n_30_0_lc.mmh"},
	{0xA2954DA9F693EDF2ULL, "llac_ca_caverns2_5802_34.dds"},
	{0xA2957FD8483EFEF7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsopen.ani"},
	{0xA29594FFF9F1F1ACULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_en.evt"},
	{0xA2959AAC08F6E37DULL, "235670925.ogg"},
	{0xA2960965B45B3BD7ULL, "hlac_ht_keep_n_4519_15x.dds"},
	{0xA296134367599772ULL, "hlac_lt_0_lowtown_n_131071_3x.dds"},
	{0xA2961720896E1D14ULL, "hlac_w_sundermount_-131073_0y.dds"},
	{0xA29627743B363340ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_no_way.gad"},
	{0xA2967307989B8590ULL, "vxht_0_hightown_a2_c3j_0_1.vlm"},
	{0xA296EE37F017814EULL, "vxlt_0_lowtown_a2_n_c_gwl_0_0.vlm"},
	{0xA297032872866413ULL, "6180431_m.fxe"},
	{0xA29715B68E70945EULL, "zz_followers_banter6159282.cl"},
	{0xA2971D17862387A4ULL, "vxw_bonepitdestroyed_a3_253_0_0.vlm"},
	{0xA298321464B85118ULL, "460919461.ogg"},
	{0xA2983C6A176FE065ULL, "vxlt_undercityserialkiller_15o_0_0.vlm"},
	{0xA2984C4B5160F6E2ULL, "gen_im_trs_bks_trueprophet.uti"},
	{0xA2988AE464963AB9ULL, "mag200_feynriel6023963.cl"},
	{0xA298E63ED888AD2DULL, "vxw_bonepitdestroyed_a3_19q_0_0.vlm"},
	{0xA299377BB47150D3ULL, "vxlt_warehouse_14i_0_0.vlm"},
	{0xA299D1D3F28BF84DULL, "vxlt_0_lowtown_docks_bgk_0_1.vlm"},
	{0xA299E9B226D6BE1CULL, "gen00fl_aveline6142251.cl"},
	{0xA299FACD5D3952CCULL, "vxdr_0_deeproads_ey8_0_0.vlm"},
	{0xA29A01F3767D9BEBULL, "265294188.ogg"},
	{0xA29A5B40B5A8645FULL, "art\\vfx\\combat\\vfx_g_spider_fb_aoe_p.impa.gad"},
	{0xA29AB1DD7CE4228CULL, "zz_dog_debug6204278.cl"},
	{0xA29AEF81DAA8D7D1ULL, "zz_otl_debug6060942.cl"},
	{0xA29AF3ACB3C7E679ULL, "868827044.ogg"},
	{0xA29B0E88AA5CDBC2ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.impa.ani"},
	{0xA29B305AAA2430EFULL, "hlac_ht_keep_n_3732_15z.dds"},
	{0xA29B615035622D68ULL, "6126417_f.fxe"},
	{0xA29BD18250A1405EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.evt"},
	{0xA29C2331E6BEFDFEULL, "dre000_stroud6036691.cl"},
	{0xA29C8535F68ED5F7ULL, "6208625_m.fxe"},
	{0xA29D5C5C6D1D5065ULL, "gen00fl_dog6117788.cl"},
	{0xA29DD1279CBD98CDULL, "6051469_m.fxe"},
	{0xA29E3C43AF891CFBULL, "307876280.ogg"},
	{0xA29E413A68005223ULL, "art\\levels\\races\\proxy\\props\\door\\animation\\dooropen.ani"},
	{0xA29EA0A141B6666EULL, "917920932.ogg"},
	{0xA29EA37F6186F224ULL, "vxlt_hangedman_1rh_0_0.vlm"},
	{0xA29F5422A3FF26B9ULL, "6068353_m.fxe"},
	{0xA29F8F4A0C4BFD59ULL, "hlac_lt_gallowstemplar_f_2382_2y.dds"},
	{0xA2A034B2C41EBC04ULL, "gen00fl_dog6115099.cl"},
	{0xA2A15040B5AE704CULL, "art\\vfx\\combat\\vfx_g_spider_fb_aoe_p.impa.evt"},
	{0xA2A15058DDE3A979ULL, "6102776_m.fxe"},
	{0xA2A169D6512CEA2FULL, "gen_im_trs_slv_lgt_glv_r2.uti"},
	{0xA2A16A817C787BD7ULL, "llac_w_sundermount_2_0.dds"},
	{0xA2A1E08573649FAFULL, "mag110_vincento6017667.cl"},
	{0xA2A2005B946BF5E5ULL, "gen00fl_anders6096626.cl"},
	{0xA2A235BA4F79D7A3ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockac.msh"},
	{0xA2A243BA4F79EFFDULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockac.mmh"},
	{0xA2A2465A393291C7ULL, "lgt110_urn_con_leader6133020.cl"},
	{0xA2A247BA4F79F60EULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockac.mao"},
	{0xA2A2511BCA705BD8ULL, "zz_mer_debug6032459.cl"},
	{0xA2A260794ABEBA23ULL, "verticalblur5.vsh.fxo"},
	{0xA2A27E241575BD82ULL, "abi_m_crushing_prison.bnk"},
	{0xA2A2B410877481ECULL, "vxca_caverns1_ihd_0_0.vlm"},
	{0xA2A2B83175C5EF52ULL, "vxlt_warehouse_14l_0_0.vlm"},
	{0xA2A2E88250A78589ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.gad"},
	{0xA2A3140C2CC6EF3CULL, "6079945_m.fxe"},
	{0xA2A37CFFF9FE0605ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_grdstand_en.ani"},
	{0xA2A393C61CE5A033ULL, "6097876_m.fxe"},
	{0xA2A3B5F566F21D6CULL, "vxht_keep_6b9_0_0.vlm"},
	{0xA2A48BF139E510C0ULL, "zz_enc_test.lst"},
	{0xA2A4F9C88BE3271BULL, "vxlt_undercityencounter_11a_0_0.vlm"},
	{0xA2A537E1E7922CABULL, "pro000st_opening_pc_2.stg"},
	{0xA2A582568CEE4C08ULL, "480982773.ogg"},
	{0xA2A5829BE139C462ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobb_0t.dds"},
	{0xA2A58EF47819A024ULL, "vxw_bonepitdestroyed_a3_200_0_0.vlm"},
	{0xA2A5ABEF8C73436BULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyn_0d.dds"},
	{0xA2A7043E5048251AULL, "vxlt_backalley_n_1ws_0_0.vlm"},
	{0xA2A715518E0F627CULL, "6137618_m.fxe"},
	{0xA2A7B8BC9963DDF1ULL, "379921310.ogg"},
	{0xA2A8546862CD9A28ULL, "hlac_lt_darktown_-1_29y.dds"},
	{0xA2A8C165355CA6ADULL, "vxca_cavern1_haunted_inv_0_0.vlm"},
	{0xA2A92C53890A90C8ULL, "w_bonepit_22_0.msh"},
	{0xA2A92ED2670E71D5ULL, "tal_thre_roll.mao"},
	{0xA2A93E53890AAF59ULL, "w_bonepit_22_0.mao"},
	{0xA2A94A53890AC3F2ULL, "w_bonepit_22_0.mmh"},
	{0xA2A9B98250AD54D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.ani"},
	{0xA2AA1323D53F3F83ULL, "198263664.ogg"},
	{0xA2AA7D58542005B5ULL, "follower_banter6143101.cl"},
	{0xA2AA8BF139E97C11ULL, "zz_enc_test.ncs"},
	{0xA2AAD8E8578CD690ULL, "dae000tr_exit_.utt"},
	{0xA2AB1A3B01F27A03ULL, "6135282_m.fxe"},
	{0xA2AB324B704D084BULL, "plt_lgt000pt_ultra_2_06.plo"},
	{0xA2AC1607AFA0A02EULL, "vxht_keep_641_0_0.vlm"},
	{0xA2AC2C693BAC2E0BULL, "6101375_m.fxe"},
	{0xA2AC43DC4471B95EULL, "6137655_f.fxe"},
	{0xA2AC9B1C2E4E028CULL, "6031967_m.fxe"},
	{0xA2ACCCEAF33BF8A1ULL, "6152734_f.fxe"},
	{0xA2ACED0272E0A1ADULL, "286436897.ogg"},
	{0xA2ACFFECBD5AE93BULL, "300769224.ogg"},
	{0xA2AD55BD0E4D65CFULL, "hlac_ht_keep_3732_6y.dds"},
	{0xA2AD6F0A3FD35A50ULL, "qun160_vashothlead6055584.cl"},
	{0xA2AE0CF592ED1BC9ULL, "829847735.ogg"},
	{0xA2AE3840B5B8D1A5ULL, "art\\vfx\\combat\\vfx_g_spider_fb_aoe_p.impa.ani"},
	{0xA2AE5B9428D29F79ULL, "w_blightlands_32_0.rml"},
	{0xA2AE82F139ED558BULL, "zz_enc_test.are"},
	{0xA2AE8BECD160B017ULL, "vxca_caverns2_78r_0_0.vlm"},
	{0xA2AF274B2AE0E86BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2win.phy"},
	{0xA2AF36BC13BAD38BULL, "follower_banter6169828.cl"},
	{0xA2AFC235CFD9C1F9ULL, "6107876_m.fxe"},
	{0xA2B05E4176861A50ULL, "6078713_m.fxe"},
	{0xA2B0A23E6494E41DULL, "361143906.ogg"},
	{0xA2B0E068E38E7F75ULL, "t3_shd_blr.tnt"},
	{0xA2B12B87EF910764ULL, "6095954_m.fxe"},
	{0xA2B13F0FC8AB0CCBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge02_d.dds"},
	{0xA2B14160989AA4F3ULL, "art\\vfx\\cinematic\\vfx_m_sta_fire_close_imp_p.anb"},
	{0xA2B15C5F3BA7BD1DULL, "one280_emrys6137205.cl"},
	{0xA2B16DDE9F869E88ULL, "genpt_crafting.ncs"},
	{0xA2B1BEAD711838D9ULL, "45446264.ogg"},
	{0xA2B2218DE5623D6EULL, "vxw_blightlands_9v_0_0.vlm"},
	{0xA2B25E035C0990FAULL, "mer341_jansen6054876.cl"},
	{0xA2B296B5D0058D5BULL, "hlac_lt_gallowstemplar_2383_9z.dds"},
	{0xA2B2FD160C258A49ULL, "874802734.ogg"},
	{0xA2B3243E8E8E3EA3ULL, "6045192_m.fxe"},
	{0xA2B34734155E49FFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\abomination.anb"},
	{0xA2B3583A6812D50DULL, "art\\levels\\races\\proxy\\props\\door\\animation\\dooropen.gad"},
	{0xA2B35FE838386AEBULL, "gen00fl_fenris6110858.cl"},
	{0xA2B3650658CEE043ULL, "6182455_m.fxe"},
	{0xA2B38C88EAAFA48AULL, "6068480_m.fxe"},
	{0xA2B3B7AA87A8AEF3ULL, "art\\characters\\creatures\\strider\\animation\\turnright30.ani"},
	{0xA2B54A22FFD2CD4AULL, "lit_conj_cloud02.mao"},
	{0xA2B54F79DE2685ABULL, "zz_followers_banter6158097.cl"},
	{0xA2B57DB9794FCBB2ULL, "6084921_m.fxe"},
	{0xA2B59901DF405C27ULL, "vxht_0_hightown_n_c6t_0_0.vlm"},
	{0xA2B59E1726360238ULL, "228644024.ogg"},
	{0xA2B5F8BDC85D1DCEULL, "6204981_m.fxe"},
	{0xA2B61D3528EDBCCCULL, "gen00fl_isabela6106423.cl"},
	{0xA2B6394045495AC3ULL, "vxdr_primevaltomb_w0_0_0.vlm"},
	{0xA2B69C20F216C486ULL, "896177976.ogg"},
	{0xA2B6C6323F707C3FULL, "6051027_m.fxe"},
	{0xA2B70EB76C2BBDF7ULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_c.anb"},
	{0xA2B764F53716C871ULL, "llac_lt_0_lowtown_-131072_3.dds"},
	{0xA2B7B04C282A92B5ULL, "1064145983.ogg"},
	{0xA2B7B74599490B93ULL, "6137621_f.fxe"},
	{0xA2B7FFD5E7B66868ULL, "gen00fl_merrill6150679.cl"},
	{0xA2B88DC792EEE817ULL, "lt_backalley_n_3_0_ll.mmh"},
	{0xA2B89574DA2D6376ULL, "420296052.ogg"},
	{0xA2B8A29EAFF22455ULL, "vxht_mansion2_1rk_0_0.vlm"},
	{0xA2B8CAFD59317386ULL, "vxlt_shop_dr_0_0.vlm"},
	{0xA2B8E268A049AB45ULL, "315482468.ogg"},
	{0xA2B9A54F263992A9ULL, "hlac_ca_caverns1_24427_0y.dds"},
	{0xA2B9AE5FF08FC275ULL, "dae200_bran6094877.cl"},
	{0xA2BA15C44A1286AAULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extboards.dds"},
	{0xA2BA65D6EBC4FECDULL, "vxht_0_hightown_bnx_0_0.vlm"},
	{0xA2BA73CFE2304D5FULL, "6129423_m.fxe"},
	{0xA2BA87E4A196B8AEULL, "vxht_keep_612_0_0.vlm"},
	{0xA2BB89BB02AD14DCULL, "vxdr_0_deeproads_gfw_0_0.vlm"},
	{0xA2BC21559C531EFBULL, "vxw_bonepitdestroyed_a3_17y_0_0.vlm"},
	{0xA2BD6ADF414890B9ULL, "mnd_con_dis02.mao"},
	{0xA2BD711EF43A3D95ULL, "6054754_m.fxe"},
	{0xA2BD908F84D4FBEBULL, "hlac_lt_0_lowtown_a2_n_65535_1x.dds"},
	{0xA2BDE4CB033A4B5DULL, "vxw_bonepitmining_a2_269_0_0.vlm"},
	{0xA2BDF487B026FD22ULL, "vxca_caverns2prc_9qw_0_0.vlm"},
	{0xA2BEBBE12AFC611AULL, "219768870.ogg"},
	{0xA2BEEC6DEC2CED22ULL, "fex_bd_d_end.msh"},
	{0xA2BEEE6DEC2CF0D8ULL, "fex_bd_d_end.mmh"},
	{0xA2BF4F4EEFC60FDDULL, "6136101_f.fxe"},
	{0xA2BF52091796FD4BULL, "ico_mm_worldmap.dds"},
	{0xA2BF5E6098A7135CULL, "art\\vfx\\cinematic\\vfx_m_sta_fire_close_imp_p.mmh"},
	{0xA2BFE6D632C30E48ULL, "6173566_m.fxe"},
	{0xA2BFFBBC7970273FULL, "vxht_keep_6y7_0_0.vlm"},
	{0xA2C078D089FED0A8ULL, "6147868_m.fxe"},
	{0xA2C0AFAA87B32B2AULL, "art\\characters\\creatures\\strider\\animation\\turnright30.evt"},
	{0xA2C0C506AECB9EC4ULL, "zz_dre_debug6054667.cl"},
	{0xA2C0EFF19A0BE5FEULL, "vxlt_gallowscourtyard_76k_0_0.vlm"},
	{0xA2C1C8049981539AULL, "875064350.ogg"},
	{0xA2C23C9A1EE27FC9ULL, "906477992.ogg"},
	{0xA2C2747C5C5D83E9ULL, "vxw_blightlands_73_0_0.vlm"},
	{0xA2C3AFFFBB15B929ULL, "hlac_lt_gallowscourtyard_n_-65538_4z.dds"},
	{0xA2C3FF55266E4AB7ULL, "vxdr_0_deeproads_ckg_0_0.vlm"},
	{0xA2C40ADED13674A7ULL, "vxht_0_hightown_n_c3w_0_0.vlm"},
	{0xA2C46A33FAB337E6ULL, "523515829.ogg"},
	{0xA2C4CE8F96357AC2ULL, "one110_pryce6088402.cl"},
	{0xA2C4F7622A5EDFAFULL, "vxdr_0_deeproads_f58_0_0.vlm"},
	{0xA2C4FF1519C15FEEULL, "mag311_keran6062055.cl"},
	{0xA2C53543E247A5F1ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgtb.mmh"},
	{0xA2C54143E247B992ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgtb.mao"},
	{0xA2C54F43E247D1AFULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgtb.msh"},
	{0xA2C55C4AD49CDB2FULL, "497833027.ogg"},
	{0xA2C56EFF10D07C2DULL, "w_bonepitmining_a2_19_0_ll.mmh"},
	{0xA2C570A3CD4BD075ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_hammer_n.dds"},
	{0xA2C58ABBBFB21B21ULL, "qcr000_aveline6054362.cl"},
	{0xA2C5E7B4E71F7DEEULL, "llac_lt_0_lowtown_docks_a3_0_5.dds"},
	{0xA2C619C895B2E648ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock0a.msh"},
	{0xA2C62BC895B304D9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock0a.mao"},
	{0xA2C62C4DB710D943ULL, "llac_ht_keep_3732_10.dds"},
	{0xA2C637C895B31972ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock0a.mmh"},
	{0xA2C6A5AF497F8104ULL, "vxdr_0_deeproads_bhh_0_0.vlm"},
	{0xA2C6B085B0766FB5ULL, "statecnt_container.gda"},
	{0xA2C6C1D4C0AD1252ULL, "221184251.ogg"},
	{0xA2C6F599D7C7989AULL, "203374311.ogg"},
	{0xA2C70175A7BA6918ULL, "dae000_bodhan6146875.cl"},
	{0xA2C7147BA1E83CD5ULL, "vxht_mansion2_1uj_0_0.vlm"},
	{0xA2C768241594B4A0ULL, "abi_m_crushing_prison.wbd"},
	{0xA2C78EAA87B9119DULL, "art\\characters\\creatures\\strider\\animation\\turnright30.gad"},
	{0xA2C7A14F27C3317DULL, "zz_followers_banter6159324.cl"},
	{0xA2C7B7091208645CULL, "hlac_lt_hangedman_-1_18x.dds"},
	{0xA2C801998F92AF48ULL, "vxht_mansion2b_2os_0_1.vlm"},
	{0xA2C8963731871AB9ULL, "329254033.ogg"},
	{0xA2C8D2B8E093D821ULL, "801404310.ogg"},
	{0xA2C903E229D85303ULL, "6087811_m.fxe"},
	{0xA2C94ACEAB8EBBB5ULL, "vxdr_primevaltomb_6ud_0_0.vlm"},
	{0xA2C97264299FABF8ULL, "qun111_ketojan6059226.cl"},
	{0xA2C9729A83CD84E9ULL, "gen00fl_fenris6126369.cl"},
	{0xA2C9A30DB6123741ULL, "vxlt_0_lowtown_e8a_0_0.vlm"},
	{0xA2C9E6FA4AF83D34ULL, "mag100_cullen6140980.cl"},
	{0xA2CA5742F6E85A5AULL, "lgt260_eau_tmp_dockworker6173550.cl"},
	{0xA2CA93328E49377BULL, "vxw_bonepitdestroyed_a3_18x_0_0.vlm"},
	{0xA2CBBDE15765D8C4ULL, "6016164_m.fxe"},
	{0xA2CBC52C6B63CF2AULL, "w_encounter2_camp_13_0_lc.mmh"},
	{0xA2CD269EDD1E36FEULL, "gen00fl_varric6093576.cl"},
	{0xA2CD26FFDBC881B0ULL, "cs_and111cs_anders_intro.wbd"},
	{0xA2CD3CD00EE8500AULL, "6059968_f.fxe"},
	{0xA2CD514CC84FA65EULL, "vxw_encounter1_b1_0_0.vlm"},
	{0xA2CD814FEBC4DA3CULL, "hlac_lt_0_lowtown_docks_131072_0x.dds"},
	{0xA2CDD9E7124EBEE1ULL, "gen00fl_fenris6126840.cl"},
	{0xA2CE643BEDC7948FULL, "ave200_aveline_bar.bnk"},
	{0xA2CE68A41999887AULL, "321739270.ogg"},
	{0xA2CE6D996B663FBFULL, "vxht_keep_6b6_0_0.vlm"},
	{0xA2CED03EF5C595B7ULL, "vxht_chantry_2ce_0_0.vlm"},
	{0xA2CEEAAD7BF4E928ULL, "6140968_m.fxe"},
	{0xA2D00BC6F5AA3994ULL, "gen_im_wep_rog_arc_act3_02.uti"},
	{0xA2D02787DB902E6EULL, "vxw_sundermount_26k_0_0.vlm"},
	{0xA2D05EB5A749623CULL, "hlac_lt_0_lowtown_docks_n_-3_5x.dds"},
	{0xA2D0C838B05A7B92ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_tram_1.mao"},
	{0xA2D0E6594E539C69ULL, "vxw_blightlands_82_0_0.vlm"},
	{0xA2D174CB30AE7A50ULL, "6161323_m.fxe"},
	{0xA2D1E6F49F51D3E0ULL, "vxlt_warehouse_hw_0_0.vlm"},
	{0xA2D27CBBC32C8D27ULL, "vxht_0_hightown_n_c8v_0_0.vlm"},
	{0xA2D2863BEDCBB7F7ULL, "ave200_aveline_bar.cnv"},
	{0xA2D34E34012B05ABULL, "6174191_m.fxe"},
	{0xA2D394725F523C3BULL, "mag160_varian_ilithis6171463.cl"},
	{0xA2D3D3A236ED73C1ULL, "136970230.ogg"},
	{0xA2D42681080400B5ULL, "vxlt_warehouse_n_1cq_0_0.vlm"},
	{0xA2D4B809E5A07821ULL, "209058847.ogg"},
	{0xA2D4C933E1FEF6F9ULL, "6116180_m.fxe"},
	{0xA2D4DEEB0081EEB3ULL, "6064665_m.fxe"},
	{0xA2D5865893DE5555ULL, "vxht_mansion2_1xi_0_0.vlm"},
	{0xA2D5A145B9C3C237ULL, "vxht_0_hightown_n_7fx_0_0.vlm"},
	{0xA2D5ABEFD6BD73E9ULL, "6161083_m.fxe"},
	{0xA2D5C54889AD4FD4ULL, "hlac_lt_backalley_n_0_6y.dds"},
	{0xA2D61AC3B9F82DDAULL, "346120507.ogg"},
	{0xA2D67A41FDBFE7BCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0di_left_1.mao"},
	{0xA2D6BBAB84CA2E9FULL, "141596599.ogg"},
	{0xA2D6EAD30192993AULL, "w_bonepitmining_a2_132_0.msh"},
	{0xA2D6F8D30192B173ULL, "w_bonepitmining_a2_132_0.mao"},
	{0xA2D6FCD30192B780ULL, "w_bonepitmining_a2_132_0.mmh"},
	{0xA2D7021CE5905F4FULL, "art\\characters\\creatures\\dragon\\animation\\cwfr.gad"},
	{0xA2D72FE9603E1484ULL, "6125297_m.fxe"},
	{0xA2D73CE613736805ULL, "vxht_mansion2_3g2_0_0.vlm"},
	{0xA2D749D06FA5213AULL, "408218108.ogg"},
	{0xA2D76D398ED964A6ULL, "promoitem_one.gda"},
	{0xA2D77A505E907B49ULL, "6055159_m.fxe"},
	{0xA2D7BB5061B9D5A9ULL, "lgt160_wk_soph_hunter.bnk"},
	{0xA2D7BB835D928547ULL, "vxht_0_hightown_a2_a7z_0_0.vlm"},
	{0xA2D7BCAB9D84D435ULL, "vxdr_primevaltomb_6xc_0_0.vlm"},
	{0xA2D81F6E2F4D0E44ULL, "llac_ht_0_hightown_65535_1.dds"},
	{0xA2D8BC1662ED985EULL, "594638346.ogg"},
	{0xA2D8E7616B8BCD23ULL, "hlac_dr_0_deeproads_17035_0y.dds"},
	{0xA2D902C7CD176990ULL, "616162116.ogg"},
	{0xA2D9409A343C4AF4ULL, "803649284.ogg"},
	{0xA2D99C654327BF7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_down_loop.gad"},
	{0xA2D9B6D4D9A10B83ULL, "qun230cr_guild_contact6068880.cl"},
	{0xA2D9CA0E36161673ULL, "ser221cr_last_corpse.utc"},
	{0xA2D9E5CA6EE2325CULL, "one300dg_vanard6063873.cl"},
	{0xA2DACC2569121D06ULL, "vxlt_gallowscourtyard_74m_0_0.vlm"},
	{0xA2DB167379E7593BULL, "art\\characters\\creatures\\ashwraith\\animation\\abilities.ani"},
	{0xA2DB4EC367CDD45DULL, "llac_ca_cavern1_haunted_24423_4.dds"},
	{0xA2DB9D5061BD8BE9ULL, "lgt160_wk_soph_hunter.cnv"},
	{0xA2DBAB05E6129B10ULL, "1031093647.ogg"},
	{0xA2DBC329BA45BEDEULL, "vxw_encounter1_g2_0_0.vlm"},
	{0xA2DC12B5EC255F4BULL, "dae000_mother6121056.cl"},
	{0xA2DC2BF28F2CE7A9ULL, "285946674.ogg"},
	{0xA2DCB81C836F40B6ULL, "dae000cr_grand_cleric6147885.cl"},
	{0xA2DDF5B91AA6D993ULL, "gen_im_trs_gen_dua_dag_r1.uti"},
	{0xA2DE171CE596A1FCULL, "art\\characters\\creatures\\dragon\\animation\\cwfr.evt"},
	{0xA2DE44A301EFC714ULL, "vxht_0_hightown_5io_0_0.vlm"},
	{0xA2DE5A68E4966C27ULL, "305504392.ogg"},
	{0xA2DE832AC335F2D1ULL, "343557037.ogg"},
	{0xA2E00BB76C4EBDF8ULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_c.mmh"},
	{0xA2E0FDD0E9B9F175ULL, "gen00fl_anders6119439.cl"},
	{0xA2E122406016A654ULL, "153020384.ogg"},
	{0xA2E154E197DADC9FULL, "vxlt_hangedman_10u_0_0.vlm"},
	{0xA2E1BF9ABEA2CB9EULL, "vxdr_0_deeproads_epu_0_0.vlm"},
	{0xA2E1EFCE2471960FULL, "mag371_elsa6041152.cl"},
	{0xA2E21DCF827181E7ULL, "hlac_lt_gallowsprison_n_-1_15z.dds"},
	{0xA2E292DBA23E2262ULL, "6122738_f.fxe"},
	{0xA2E2B57D97647AD1ULL, "6016812_f.fxe"},
	{0xA2E2D91E3B833FE6ULL, "hlac_dr_0_deeproads_17038_5x.dds"},
	{0xA2E2DC6D41D81E51ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_brick.dds"},
	{0xA2E33B933A3BC998ULL, "489563603.ogg"},
	{0xA2E361383FCB30F9ULL, "6082245_m.fxe"},
	{0xA2E3C3ACC181F97AULL, "895346564.ogg"},
	{0xA2E3F83585D46DD5ULL, "vxht_mansion2_1ch_0_0.vlm"},
	{0xA2E45DDBAF46F43DULL, "static_static_p33.psh.fxo.dx11.assembly"},
	{0xA2E481C2577F0366ULL, "hla_ht_0_hightown_a3_-1_0y.dds"},
	{0xA2E482E65229DDACULL, "hlac_lt_warehouse_1_6x.dds"},
	{0xA2E4DADC05983928ULL, "6033161_m.fxe"},
	{0xA2E4EB1F22EF3A3EULL, "951112298.ogg"},
	{0xA2E53F8E48E7B030ULL, "lgt220_shp_mag_guard6149885.cl"},
	{0xA2E55D5AC322B588ULL, "dae200cr_viscount6087215.cl"},
	{0xA2E57478B38344CCULL, "and331ip_at_darktown.utp"},
	{0xA2E62D604F889DC7ULL, "vxht_0_hightown_a2_a2u_0_0.vlm"},
	{0xA2E636D17895DF21ULL, "vxlt_undercityserialkiller_1n6_0_0.vlm"},
	{0xA2E6568380375114ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_1.ani"},
	{0xA2E69A46B4441093ULL, "406082973.ogg"},
	{0xA2E6EACA7A1A969AULL, "ser221_party_comments6209956.cl"},
	{0xA2E73D8D0CC77E99ULL, "757576552.ogg"},
	{0xA2E77C72D647B0C6ULL, "6135881_m.fxe"},
	{0xA2E7C0C41694DEFAULL, "vxht_0_hightown_a2_aob_0_1.vlm"},
	{0xA2E7C527BE54C231ULL, "shd_cld.mao"},
	{0xA2E7D4CC3499D3FDULL, "6206188_m.fxe"},
	{0xA2E81E7379F1F132ULL, "art\\characters\\creatures\\ashwraith\\animation\\abilities.evt"},
	{0xA2E8606DD3C915DDULL, "923770793.ogg"},
	{0xA2E8A9C13BF2ADC4ULL, "vxlt_undercityencounter_za_0_1.vlm"},
	{0xA2E8C583A623D87EULL, "780678927.ogg"},
	{0xA2E8F1B5D813BB74ULL, "vxht_bartrand_n_5pd_0_0.vlm"},
	{0xA2E90B4D72133256ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dmg_v1.ani"},
	{0xA2E91F929EEEA078ULL, "hlac_lt_0_lowtown_a2_65535_7x.dds"},
	{0xA2E946F452CECD45ULL, "6025339_f.fxe"},
	{0xA2E94A1E860946A2ULL, "vxca_caverns2prc_9fr_0_0.vlm"},
	{0xA2EA4C1D1A9F7A90ULL, "w_bonepitdestroyed_a3_29_0_ll.mmh"},
	{0xA2EAFA12392EB2CCULL, "vxlt_darktown_52u_0_0.vlm"},
	{0xA2EB1F1CE5A13915ULL, "art\\characters\\creatures\\dragon\\animation\\cwfr.ani"},
	{0xA2EB4E6A88F589C0ULL, "6118140_f.fxe"},
	{0xA2EB51534F5270BFULL, "vxht_keep_6h0_0_0.vlm"},
	{0xA2EBB3F8D9B1C6B7ULL, "vxht_chantry_2yg_0_0.vlm"},
	{0xA2EBE510A32C4404ULL, "lla_ht_0_hightown_-65534_0.dds"},
	{0xA2EC9D4A2B1A041FULL, "6044278_m.fxe"},
	{0xA2ED01FAD37A07F3ULL, "ht_keep_n_mmap.dds"},
	{0xA2ED275423F4B06FULL, "vxht_keep_63e_0_0.vlm"},
	{0xA2ED5383803D6B76ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_1.gad"},
	{0xA2ED796543382CC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_down_loop.ani"},
	{0xA2EE94810762C885ULL, "578414747.ogg"},
	{0xA2EEB41082B24757ULL, "6095603_m.fxe"},
	{0xA2EECAAE833E04E0ULL, "vxlt_warehouse_vu_0_0.vlm"},
	{0xA2EEDC9E825E6E30ULL, "zz_one_debug6051522.cl"},
	{0xA2EEEC756223C316ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sandrune_sa.dds"},
	{0xA2EEF9D5C8449B18ULL, "6024565_m.fxe"},
	{0xA2EF339EE13637FFULL, "friendly.mao"},
	{0xA2EF3D7379F844C5ULL, "art\\characters\\creatures\\ashwraith\\animation\\abilities.gad"},
	{0xA2EF46AE8B7DFE0FULL, "200142189.ogg"},
	{0xA2EFCB549A207581ULL, "character_character_p154.psh.fxo.dx11"},
	{0xA2F08CB049C1097CULL, "hlac_lt_0_lowtown_a2_65533_2y.dds"},
	{0xA2F0E806DEF23FBFULL, "124607708.ogg"},
	{0xA2F0F8A57512C624ULL, "gen00fl_anders6125314.cl"},
	{0xA2F10199AA5D7063ULL, "hlac_ht_mansion_n_3369_2x.dds"},
	{0xA2F195F863CD343EULL, "hlac_lt_0_lowtown_a3_n_c_1_8x.dds"},
	{0xA2F1D15616B55CB2ULL, "llac_lt_0_lowtown_a2_n_c_65535_6.dds"},
	{0xA2F1E7F209314AC2ULL, "vxw_woundedcoast_1n0_0_0.vlm"},
	{0xA2F1EBF21222CD59ULL, "hightown_male_noble_gossip6171412.cl"},
	{0xA2F22F5E93A91EB2ULL, "vxht_0_hightown_n_ak8_0_0.vlm"},
	{0xA2F25EB421B6897DULL, "vxlt_hangedman_204_0_0.vlm"},
	{0xA2F26A1277CA8655ULL, "vxht_mansion2_1fw_0_0.vlm"},
	{0xA2F28584021CFF1FULL, "vxht_mansion_n_52v_0_0.vlm"},
	{0xA2F2A6D634C8D3ADULL, "vxlt_0_lowtown_a2_by1_0_0.vlm"},
	{0xA2F2D9FD316E2D56ULL, "679936837.ogg"},
	{0xA2F2E3455FE7F3E0ULL, "14399604.ogg"},
	{0xA2F35D07F9C5D504ULL, "vxlt_blackemporium_it_0_0.vlm"},
	{0xA2F3A631BCF362EFULL, "83298799.ogg"},
	{0xA2F3E4EF68D1F42DULL, "vxlt_gallowscourtyard_7cc_0_0.vlm"},
	{0xA2F3E7056C5A86BCULL, "vrc321_housekeeper6087326.cl"},
	{0xA2F42D4AB39D60CDULL, "vxht_0_hightown_bzt_0_0.vlm"},
	{0xA2F456F9F627DF6CULL, "149149570.ogg"},
	{0xA2F46E838043B7D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_stop_1.evt"},
	{0xA2F4745B96C03114ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig_n.dds"},
	{0xA2F4A8AE6A8BF7A1ULL, "vxlt_undercityserialkiller_1c5_0_0.vlm"},
	{0xA2F4BBB4EEA63898ULL, "hlac_lt_0_lowtown_docks_n_-65535_0z.dds"},
	{0xA2F4F09CF0D8B81AULL, "art\\levels\\ztools\\default\\generic\\0rl_basic.phy"},
	{0xA2F5062AA39FD4F4ULL, "6076024_f.fxe"},
	{0xA2F529601E726B94ULL, "cd_mag291_vincento.wbd"},
	{0xA2F5D73A8AB496F4ULL, "448563945.ogg"},
	{0xA2F6134D721DCA47ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dmg_v1.evt"},
	{0xA2F67C906D819CDCULL, "vxlt_warehouse_n_cs_0_0.vlm"},
	{0xA2F6884F15B57A5AULL, "vxht_chantry_31s_0_0.vlm"},
	{0xA2F7567A74968724ULL, "store_dae170cr_armor.utm"},
	{0xA2F7BBFB77FF5F22ULL, "vxca_caverns2prc_9ms_0_0.vlm"},
	{0xA2F7E6BCC0E2DE37ULL, "927123102.ogg"},
	{0xA2F7F4EA038AC1D3ULL, "6103662_f.fxe"},
	{0xA2F8677DC9D72CB5ULL, "gen_im_wep_war_shd_act2_05.uti"},
	{0xA2F8F5FFBEEBF521ULL, "6170687_m.fxe"},
	{0xA2F928D0D73B4E68ULL, "w_encounter2_camp_62_0.msh"},
	{0xA2F9390179F515C9ULL, "28952377.ogg"},
	{0xA2F93AD0D73B6CF9ULL, "w_encounter2_camp_62_0.mao"},
	{0xA2F946D0D73B8092ULL, "w_encounter2_camp_62_0.mmh"},
	{0xA2F9870D5E578ECAULL, "23316810.ogg"},
	{0xA2F9B08B77217916ULL, "6172732_m.fxe"},
	{0xA2FA16CF019E97F7ULL, "gen00fl_aveline6105471.cl"},
	{0xA2FA226C9F35A831ULL, "hlac_ht_bartrand_n_3369_1z.dds"},
	{0xA2FA48B914A1CC5EULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyn_0s.dds"},
	{0xA2FA83BDC839249BULL, "qun221_petrice6093180.cl"},
	{0xA2FAC650FBEA5E84ULL, "hlac_ca_cavern1_haunted_24427_1z.dds"},
	{0xA2FBDB84D70DA52DULL, "lt_0_lowtown_a2_n_c_3_0.mmh"},
	{0xA2FBDF84D70DAC7EULL, "lt_0_lowtown_a2_n_c_3_0.mao"},
	{0xA2FBEA54182E4724ULL, "vxw_encounter2_camp_6f_0_0.vlm"},
	{0xA2FBED84D70DC3B3ULL, "lt_0_lowtown_a2_n_c_3_0.msh"},
	{0xA2FC8E735A68DAE1ULL, "llac_lt_0_lowtown_-65536_0.dds"},
	{0xA2FC96C6F238E483ULL, "one300_zevran6075854.cl"},
	{0xA2FD0D79198FC5CAULL, "llac_lt_0_lowtown_131071_4.dds"},
	{0xA2FD1F5A45D7C291ULL, "6053531_m.fxe"},
	{0xA2FD2F3D89A024E5ULL, "lgt310pt_fixer.ncs"},
	{0xA2FD2FA48AD0B1BEULL, "lla_ht_0_hightown_0_0.dds"},
	{0xA2FD344D72242090ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dmg_v1.gad"},
	{0xA2FD42B430B4AF9EULL, "542944811.ogg"},
	{0xA2FDB80997C0B857ULL, "791128955.ogg"},
	{0xA2FE0B63B80429E0ULL, "one100dg_vanard6048267.cl"},
	{0xA2FFB643C4FFBC3FULL, "qun200_party_comment6095387.cl"},
	{0xA30059CEFB276342ULL, "vxw_woundedcoast_1i1_0_0.vlm"},
	{0xA300A13B859F3732ULL, "vxht_0_hightown_n_al9_0_0.vlm"},
	{0xA300AC6BDF55173DULL, "gen_im_npc_arm_hum_glv_rdrb.uti"},
	{0xA300C9F092F2DB08ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thorn.mmh"},
	{0xA300D5F092F2EFABULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thorn.mao"},
	{0xA300DBEF69C09ED5ULL, "vxht_mansion2_1iv_0_0.vlm"},
	{0xA300DE76A5423566ULL, "gen00fl_isabela6089535.cl"},
	{0xA300E7F092F30E32ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thorn.msh"},
	{0xA300EEC3DED4B5F0ULL, "gen00fl_isabela6079570.cl"},
	{0xA301653306E46FC6ULL, "673839914.ogg"},
	{0xA301960C7DDC6C27ULL, "hlac_lt_0_lowtown_a2_n_-65535_2x.dds"},
	{0xA302137C4C313599ULL, "219841284.ogg"},
	{0xA302A262FCEAA16BULL, "922814375.ogg"},
	{0xA302C2BF5D90B3A2ULL, "dae000ip_crowd_mf5_talk3.utp"},
	{0xA302DDD0260E0422ULL, "6143350_m.fxe"},
	{0xA30361C4BC400009ULL, "343559040.ogg"},
	{0xA303831017EAF456ULL, "lt_0_lowtown_a2_45_0_lc.mmh"},
	{0xA303DAB0CD0C3C6FULL, "vxlt_gallowstemplar_f_43s_0_0.vlm"},
	{0xA303E6858CD955C3ULL, "42418103.ogg"},
	{0xA304FFBED5D8728AULL, "496971174.ogg"},
	{0xA3051D10946B3A67ULL, "108768755.ogg"},
	{0xA3053B034E3958A7ULL, "vxw_woundedcoast_a5_0_0.vlm"},
	{0xA30552B6A22A0C38ULL, "648642110.ogg"},
	{0xA30586DF04CEEEDEULL, "809040890.ogg"},
	{0xA3058CFFC21A9B47ULL, "vxlt_gallowstemplar_3nd_0_0.vlm"},
	{0xA305DB84D715F9A6ULL, "lt_0_lowtown_a2_n_c_3_0.rml"},
	{0xA305F97E607B594EULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnleft.evt"},
	{0xA306088CF32F128CULL, "6171425_m.fxe"},
	{0xA3067D80BE268FFFULL, "one101cr_spider_loot.utc"},
	{0xA3068444000965ECULL, "vxlt_0_lowtown_e0x_0_0.vlm"},
	{0xA3068F64E3953FBEULL, "6136912_m.fxe"},
	{0xA306B91C806C0E96ULL, "6128726_m.fxe"},
	{0xA306D7725067ABD7ULL, "6135361_m.fxe"},
	{0xA306D92C80A86790ULL, "vxw_blightlands_f5_0_0.vlm"},
	{0xA3076C47DD997123ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.dura.evt"},
	{0xA307AAD4EACB4F8CULL, "vxlt_0_lowtown_ghr_0_0.vlm"},
	{0xA307B2C62AEFEB64ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_act_c.anb"},
	{0xA307DDCC1D1AE3CCULL, "vxlt_darktown_58k_0_0.vlm"},
	{0xA307E816398103AFULL, "803324898.ogg"},
	{0xA307F554D193367CULL, "643419920.ogg"},
	{0xA307F800ADF77F8FULL, "dae000_pickpocket_scene6175102.cl"},
	{0xA30889579752AB69ULL, "6087596_m.fxe"},
	{0xA308D4F9431FD77FULL, "vxht_chantry_271_0_0.vlm"},
	{0xA3092A99E94EF8D4ULL, "119738104.ogg"},
	{0xA3094AC1CB2F5C12ULL, "459584256.ogg"},
	{0xA30983FAFE6A75D8ULL, "6106954_f.fxe"},
	{0xA30A3632CD46AA3CULL, "6119702_m.fxe"},
	{0xA30A48D222FE9121ULL, "1051703600.ogg"},
	{0xA30A5C310A245FA4ULL, "vxw_encounter2_camp_1g_0_0.vlm"},
	{0xA30ACDEA4ADFD3F3ULL, "vxht_store_c7_0_0.vlm"},
	{0xA30B0349D984F92FULL, "zz_mage2_simple.cut"},
	{0xA30B043B4F9C93C3ULL, "vxw_encounter2_e8_0_0.vlm"},
	{0xA30B232FAB578799ULL, "sib131_letter_react6127313.cl"},
	{0xA30B4A53895D8D99ULL, "w_bonepit_22_0.rml"},
	{0xA30BD8EA6F4AFE98ULL, "vxca_caverns1_mines_cjq_0_0.vlm"},
	{0xA30C15BD363EE9D7ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_bdy_0e.dds"},
	{0xA30CB6E3F1A467F4ULL, "hlac_w_bonepit_0_4x.dds"},
	{0xA30CD484D71C0CEAULL, "lt_0_lowtown_a2_n_c_3_0.phy"},
	{0xA30CD5D5488D5825ULL, "hlac_lt_0_lowtown_-131072_3x.dds"},
	{0xA30CF07E608168F9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnleft.gad"},
	{0xA30CF08198FEE6BEULL, "871122148.ogg"},
	{0xA30D15319485151EULL, "vxdr_0_deeproads_ech_0_0.vlm"},
	{0xA30D5434E14C85C7ULL, "410996282.ogg"},
	{0xA30D6A86E5CB16F5ULL, "hlac_lt_0_lowtown_a2_n_c_65535_1x.dds"},
	{0xA30DC018C63FF027ULL, "gen00fl_anders6111231.cl"},
	{0xA30E4547DD9F4DA4ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.dura.gad"},
	{0xA30E7868A2FB6463ULL, "520087665.ogg"},
	{0xA30ECBABED1D7BC2ULL, "vxw_woundedcoast_1t2_0_0.vlm"},
	{0xA30F51BF87CBF48EULL, "llac_ca_caverns1_24424_3.dds"},
	{0xA30F5A767AEF3CDAULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.anb"},
	{0xA3100F276A292AADULL, "pro000_athenril6049465.cl"},
	{0xA3105A268F29B9ECULL, "art\\levels\\races\\dwarf\\props\\lyrium\\0ww_broken_1.mao"},
	{0xA3112AEDBD31A585ULL, "6058952_m.fxe"},
	{0xA311494B70A3E218ULL, "plt_lgt000pt_ultra_2_06.nss"},
	{0xA31152D194BD2488ULL, "gen00fl_anders6125206.cl"},
	{0xA311B05900A8E48AULL, "zz_mcr_debug6080850.cl"},
	{0xA311B1F1B6CBBBBEULL, "527607243.ogg"},
	{0xA311C9110770AFEAULL, "rog_stunnedsneaktattack.dds"},
	{0xA31243F222867D94ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg103a_0s.dds"},
	{0xA31248C804039A2AULL, "llac_lt_0_lowtown_a2_65536_4.dds"},
	{0xA312DC892970C7C5ULL, "qun221_petrice6083745.cl"},
	{0xA312ED5DEA6EAA08ULL, "zz_car_debug6122482.cl"},
	{0xA312F48FAB157C47ULL, "6150385_f.fxe"},
	{0xA312FC913B488459ULL, "vxlt_0_lowtown_eqf_0_0.vlm"},
	{0xA312FF3AD279901EULL, "hlac_lt_0_lowtown_a2_65535_6y.dds"},
	{0xA313017E6085F147ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderturnleft.ani"},
	{0xA313129DA4EF7B21ULL, "6117266_m.fxe"},
	{0xA31372A76713FF67ULL, "gen00fl_anders6125250.cl"},
	{0xA313FCFFDC043D72ULL, "cs_and111cs_anders_intro.bnk"},
	{0xA314474CADF604F4ULL, "vxht_bartrand_n_5ak_0_0.vlm"},
	{0xA3147F1ED27CFAECULL, "6120113_m.fxe"},
	{0xA3149256E8F8D8CEULL, "816559819.ogg"},
	{0xA31492B1FAB8C0B9ULL, "vxlt_warehouse_10w_0_0.vlm"},
	{0xA314D1C62AFAAA67ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_act_c.mmh"},
	{0xA3155447DDA58582ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.dura.ani"},
	{0xA31773651C63B905ULL, "6043454_f.fxe"},
	{0xA317BBFAE54D7A3BULL, "w_encounter1_25_0.rml"},
	{0xA318188CDC21F099ULL, "plt_lgt000pt_ultra_5_02.nss"},
	{0xA318211047F2373CULL, "160673102.ogg"},
	{0xA318A80FBF3CC2BCULL, "6114002_m.fxe"},
	{0xA3195BBF2BF78919ULL, "art\\vfx\\materials\\vfx_streak01.mao"},
	{0xA319AB3C824ECB73ULL, "vxlt_0_lowtown_docks_a3_n_coe_0_0.vlm"},
	{0xA31A0F0BFF5EF1E1ULL, "6111343_m.fxe"},
	{0xA31A380668EBEF68ULL, "vxw_woundedcoast_16b_0_0.vlm"},
	{0xA31A51866F465A94ULL, "vxlt_gallowsprison_n_6ne_0_0.vlm"},
	{0xA31A7D71FB6F90D5ULL, "541300840.ogg"},
	{0xA31A89AED1674AEBULL, "6062388_m.fxe"},
	{0xA31ACFD050A950F0ULL, "em_amb_jaken.mor"},
	{0xA31AF5C118A90840ULL, "hlac_ca_caverns1_mines_24424_28z.dds"},
	{0xA31B09253E769668ULL, "w_woundedcoast_104_0.msh"},
	{0xA31B1B253E76B4F9ULL, "w_woundedcoast_104_0.mao"},
	{0xA31B27253E76C892ULL, "w_woundedcoast_104_0.mmh"},
	{0xA31BB512BFD88614ULL, "gen00fl_merrill6127530.cl"},
	{0xA31BCAE9D7A9E3F1ULL, "486279485.ogg"},
	{0xA31C280A536E1221ULL, "605845984.ogg"},
	{0xA31C3B91137D281DULL, "93595226.ogg"},
	{0xA31C5BDA34D2DE56ULL, "vxca_caverns2_7gb_0_0.vlm"},
	{0xA31CFA3D14E67712ULL, "vxdr_0_deeproads_dw1_0_0.vlm"},
	{0xA31D48C172462381ULL, "art\\vfx\\combat\\vfx_gib_qun_rblt.ani"},
	{0xA31D9429245F5CB2ULL, "vxdr_primevalentrance_16a_0_0.vlm"},
	{0xA31E53A08AC48142ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyn_0n.dds"},
	{0xA31E5A851F4250D9ULL, "386529356.ogg"},
	{0xA31E7401A238EEA5ULL, "6020609_m.fxe"},
	{0xA31EA47E19AA867CULL, "6116721_m.fxe"},
	{0xA31EC2FAE553A493ULL, "w_encounter1_25_0.phy"},
	{0xA31EFF8CDC27E572ULL, "plt_lgt000pt_ultra_5_02.plo"},
	{0xA31F996B33676A1FULL, "gen00fl_sebastian6177748.cl"},
	{0xA320051F7517DCAFULL, "mcr371cr_harvester3.utc"},
	{0xA32007A9A4388FFBULL, "def_i_fla.mao"},
	{0xA32060D2364AAA2EULL, "lt_gallowscourtyard_58_0_ll.mmh"},
	{0xA320BAAD84FC74ACULL, "vxlt_0_lowtown_docks_ct3_0_0.vlm"},
	{0xA320E7C720264A7DULL, "hlac_lt_undercityserialkiller_-1_7z.dds"},
	{0xA320EDF24AE95CA8ULL, "lgt310_mis_burn_murder16152758.cl"},
	{0xA32117601E97882EULL, "cd_mag291_vincento.bnk"},
	{0xA321847A96E593A1ULL, "6112166_m.fxe"},
	{0xA3218F69D82FBA86ULL, "801896530.ogg"},
	{0xA321982449306124ULL, "vxca_caverns1_j5d_0_0.vlm"},
	{0xA321CAD8EB5A0641ULL, "ht_0_hightown_a2_45_0.anb"},
	{0xA321CF38B39E5545ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3co_basic_2.mao"},
	{0xA321FD4D47230A5BULL, "gen00fl_varric6076457.cl"},
	{0xA322216C88FE8C3EULL, "dae000ip_crowd_ff8_talk12.utp"},
	{0xA322715B037782E1ULL, "977284808.ogg"},
	{0xA3228D6441696918ULL, "vxca_caverns2_97n_0_0.vlm"},
	{0xA322B284374DBD01ULL, "follower_banter6165659.cl"},
	{0xA322C3387DD378B1ULL, "one110_meeran6089861.cl"},
	{0xA322EC0DDFE19F3AULL, "clf260_harley6069682.cl"},
	{0xA32329069FCA23C1ULL, "6066265_m.fxe"},
	{0xA32353C1724AA223ULL, "art\\vfx\\combat\\vfx_gib_qun_rblt.gad"},
	{0xA3237411A31B3016ULL, "vxdr_primevalentrance_22v_0_0.vlm"},
	{0xA32376A78AD0BD61ULL, "zz_mrl_debug6184183.cl"},
	{0xA323BCE664949890ULL, "vxw_blightlands_l3_0_0.vlm"},
	{0xA324235AA8BDD388ULL, "gen_im_arm_hlm_lgt_act1_02.uti"},
	{0xA32467653287F87DULL, "hlac_ca_caverns2prc_5802_1z.dds"},
	{0xA32498E36DF72D50ULL, "w_bonepitmining_a2_122_0.mao"},
	{0xA3249CE36DF73463ULL, "w_bonepitmining_a2_122_0.mmh"},
	{0xA324AAE36DF74BBDULL, "w_bonepitmining_a2_122_0.msh"},
	{0xA324EAA6CC6ABF2DULL, "vxht_0_hightown_a3_c9x_0_1.vlm"},
	{0xA3252325EA541114ULL, "vxht_mansion_n_4zd_0_0.vlm"},
	{0xA32591E31E927E7BULL, "vxlt_gallowscourtyard_7bv_0_0.vlm"},
	{0xA325DA786CB73335ULL, "385916656.ogg"},
	{0xA326235BDAFA70DCULL, "634468953.ogg"},
	{0xA3262B4DBDA11B54ULL, "hlac_lt_0_lowtown_a2_65534_0z.dds"},
	{0xA326354467F060C7ULL, "6108435_m.fxe"},
	{0xA3268C49E129FFD6ULL, "vxlt_0_lowtown_h21_0_0.vlm"},
	{0xA326B54EBAE5FF88ULL, "6089858_m.fxe"},
	{0xA326C297502D741BULL, "111199074.ogg"},
	{0xA32719ECB132DB3CULL, "6113102_m.fxe"},
	{0xA3272F93D1AD8497ULL, "453123960.ogg"},
	{0xA3274058ED4A0BDFULL, "ser221_gamlen6133132.cl"},
	{0xA32743D8AD428D93ULL, "6072272_f.fxe"},
	{0xA3274E57348B4F3BULL, "hlac_dr_primevaltomb_658_6x.dds"},
	{0xA327B774FFBFAF0AULL, "vxw_sundermount_23p_0_0.vlm"},
	{0xA327F34B46F066E5ULL, "dae000_bodhan6129152.cl"},
	{0xA327F5F5E7AFFDF3ULL, "gen00fl_fenris6126820.cl"},
	{0xA328036AC1D2A5B4ULL, "718311199.ogg"},
	{0xA32835F4F9750ECEULL, "621192458.ogg"},
	{0xA32847EFEE3039BAULL, "hlac_lt_0_lowtown_docks_a3_-65538_8y.dds"},
	{0xA3284FDD9EAAEA1CULL, "vxht_chantry_n_2zj_0_0.vlm"},
	{0xA328A894DBDE9C2AULL, "hlac_dr_0_deeproads_5849_10y.dds"},
	{0xA328A9E35AE207E8ULL, "vxw_woundedcoast_11c_0_0.vlm"},
	{0xA328C363613C7314ULL, "vxlt_gallowsprison_n_6kj_0_0.vlm"},
	{0xA32937264C3E9462ULL, "413229302.ogg"},
	{0xA32973806D81B991ULL, "vxlt_0_lowtown_a2_n_c_byh_0_0.vlm"},
	{0xA329F9ED2168C4C1ULL, "seb221_flora_harimann6177276.cl"},
	{0xA32A1530EB891F9EULL, "ave331_jeven6138725.cl"},
	{0xA32A1D1A3914CF96ULL, "gen00fl_carver6104847.cl"},
	{0xA32AABAE9FA198D4ULL, "llac_ht_bartrand_n_3381_0.dds"},
	{0xA32AC0B0DE2ADCC5ULL, "351523192.ogg"},
	{0xA32AE81E91222E88ULL, "6156888_m.fxe"},
	{0xA32AFED3FD5CB7D9ULL, "6028527_f.fxe"},
	{0xA32B193F2C3E94F8ULL, "6800666.ogg"},
	{0xA32B388C1D10E5E1ULL, "lgt251ip_door_trap1.ncs"},
	{0xA32B921AE0D16187ULL, "character_character_p134.psh.fxo.dx11"},
	{0xA32C2591C3392605ULL, "6033646_f.fxe"},
	{0xA32CE5DE942F0725ULL, "6027709_m.fxe"},
	{0xA32CE96A0B1023A8ULL, "hlac_ca_caverns2_5802_11y.dds"},
	{0xA32CF8AB8170F4BAULL, "111034132.ogg"},
	{0xA32CFC055E1BFB21ULL, "vxht_bartrand_n_70o_0_0.vlm"},
	{0xA32D219094D09930ULL, "isa311_isabela6095367.cl"},
	{0xA32D62A7DB19DB75ULL, "mag360_thrask6053531.cl"},
	{0xA32D7837F0B948ADULL, "qcr270cr_abomination.utc"},
	{0xA32DA09CCE098FAEULL, "vxca_caverns1_mines_j4t_0_0.vlm"},
	{0xA32DA4D5EF5CFEE9ULL, "art\\characters\\creatures\\corpse\\uni_cora.phy"},
	{0xA32DCE3A336A2A2EULL, "10382264.ogg"},
	{0xA32E73EB45B626D5ULL, "220234057.ogg"},
	{0xA32F1F9D1A0C8EA0ULL, "ht_0_hightown_a3_probe_9x4.dds"},
	{0xA32F456F541CE7FDULL, "1006662448.ogg"},
	{0xA32F4C4672982453ULL, "gen_im_trs_rdr_med_hlm_r0.uti"},
	{0xA32FA42927BFB17DULL, "81395816.ogg"},
	{0xA32FD3FA2172CAA1ULL, "mag291tr_nt_ragedemons.ncs"},
	{0xA32FEC417E693732ULL, "vxht_0_hightown_a2_b8u_0_0.vlm"},
	{0xA33044043559FEDCULL, "vxlt_warehouse_n_ww_0_0.vlm"},
	{0xA3306AC22095A91BULL, "vxw_bonepitdestroyed_a3_kr_0_0.vlm"},
	{0xA330E0CD7C1EF5B4ULL, "dae000_gamlen6102536.cl"},
	{0xA330F79D8B810DDFULL, "hlac_lt_undercitydungeon_-1_13z.dds"},
	{0xA33102BDB7F7AA0BULL, "commands.gda"},
	{0xA3312B0691E235F4ULL, "vxht_bartrand_n_5oi_0_0.vlm"},
	{0xA3313BE8047D883FULL, "vxlt_0_lowtown_e0w_0_0.vlm"},
	{0xA3315A74F2D79CCEULL, "lt_0_lowtown_docks_1_0_ll.mmh"},
	{0xA3322EC3568AB110ULL, "vxw_blightlands_q0_0_0.vlm"},
	{0xA33267CF271DB76AULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgta_0t.dds"},
	{0xA332A56AA042BCE6ULL, "491225821.ogg"},
	{0xA33386A79B9BF7E3ULL, "vxlt_backalley_n_1yd_0_0.vlm"},
	{0xA33403C0108896FBULL, "vxlt_gallowscourtyard_7gi_0_0.vlm"},
	{0xA33466ADD9C244B7ULL, "hlac_lt_gallowscourtyard_n_65536_1z.dds"},
	{0xA33471F4594297B9ULL, "str_cracked01_n.dds"},
	{0xA334CEA9DD92BEA9ULL, "follower_banter6155739.cl"},
	{0xA334F5682DB6DE38ULL, "197795169.ogg"},
	{0xA334FE26D3201856ULL, "vxlt_0_lowtown_h72_0_0.vlm"},
	{0xA33513AB87993DCDULL, "969040546.ogg"},
	{0xA3352D019D662E7FULL, "gen00fl_anders6125496.cl"},
	{0xA3359DE36E05A888ULL, "w_bonepitmining_a2_122_0.phy"},
	{0xA335C0C195D23B94ULL, "hlac_w_bonepitmining_a2_131075_0y.dds"},
	{0xA3360E460B2AC9EEULL, "305923163.ogg"},
	{0xA3362951F1B5C78AULL, "vxw_sundermount_24q_0_0.vlm"},
	{0xA3364A0E1BFE7612ULL, "sse_ss_abomination_female.gda"},
	{0xA3368EF6663AFC73ULL, "vxlt_0_lowtown_docks_a3_n_cug_0_0.vlm"},
	{0xA3375D767B10938DULL, "art\\vfx\\combat\\vfx_r_gn2_impact_loop_c.mmh"},
	{0xA33795845520D2D0ULL, "466022570.ogg"},
	{0xA338071C60DC7050ULL, "lit_tep_02.mao"},
	{0xA338848ABD1FE2F7ULL, "327244403.ogg"},
	{0xA338B7660E10BE72ULL, "vxdr_0_deeproads_fi5_0_0.vlm"},
	{0xA3393406A8F99BADULL, "vxca_cavern1_haunted_i8l_0_0.vlm"},
	{0xA339645B9CC42D89ULL, "vxw_woundedcoast_y3_0_0.vlm"},
	{0xA33A41F560476298ULL, "932227638.ogg"},
	{0xA33A6003DE6409C7ULL, "vxlt_undercityencounter_xu_0_0.vlm"},
	{0xA33A8C678A80349AULL, "707923767.ogg"},
	{0xA33AB6C8D977D601ULL, "417938269.ogg"},
	{0xA33AE4123AA86465ULL, "hlac_lt_0_lowtown_a2_n_c_65536_0z.dds"},
	{0xA33B6A4AC6952207ULL, "750907210.ogg"},
	{0xA33B9658B3944F84ULL, "vxlt_blackemporium_vy_0_0.vlm"},
	{0xA33BE2014265A942ULL, "376104988.ogg"},
	{0xA33CEA6C10EF5EF0ULL, "6163417_m.fxe"},
	{0xA33CFBCA82A05546ULL, "hlac_ht_brothel_n_3373_1x.dds"},
	{0xA33DD36D4FEF182BULL, "hlac_lt_darktown_-1_42y.dds"},
	{0xA33E22F9BD3AD6B5ULL, "hlac_ht_0_hightown_n_131071_2z.dds"},
	{0xA33E39A570175660ULL, "llac_dr_dungeon_3215_5.dds"},
	{0xA33E412FB9DC0D05ULL, "vxw_encounter3_g3_0_0.vlm"},
	{0xA33E5C53C94C6398ULL, "6154309_m.fxe"},
	{0xA33EAA41F41A5691ULL, "6062792_f.fxe"},
	{0xA33F1A0EE2AD4FB3ULL, "112617557.ogg"},
	{0xA33F33E601C0204EULL, "art\\vfx\\creatures\\fxc_ogre_stmp_c.mmh"},
	{0xA33FC0C83349CBBBULL, "876945017.ogg"},
	{0xA33FEB6ACD10ADCDULL, "89528205.ogg"},
	{0xA3404A9D685936DAULL, "qun110_petrice6018543.cl"},
	{0xA3404ED41BC7411EULL, "vxht_chantry_331_0_0.vlm"},
	{0xA34206219E56C70AULL, "664449405.ogg"},
	{0xA3420DFB86AD46BFULL, "hla_ht_0_hightown_-2_0x.dds"},
	{0xA3427804FAFF291AULL, "art\\levels\\races\\proxy\\props\\armoire\\prp_zzarmoire.mmh"},
	{0xA3427C04FAFF3051ULL, "art\\levels\\races\\proxy\\props\\armoire\\prp_zzarmoire.mao"},
	{0xA3428A04FAFF4860ULL, "art\\levels\\races\\proxy\\props\\armoire\\prp_zzarmoire.msh"},
	{0xA342F576D636C964ULL, "6089739_m.fxe"},
	{0xA34318FE4BDC91C7ULL, "6106635_m.fxe"},
	{0xA343592E10545D3EULL, "hlac_lt_0_lowtown_a2_n_131069_0z.dds"},
	{0xA34396A89E606BD7ULL, "gen00fl_merrill6129906.cl"},
	{0xA344051599768FFDULL, "466639735.ogg"},
	{0xA3448659A547A29CULL, "6178366_f.fxe"},
	{0xA344B7A09E29FB49ULL, "vxlt_0_lowtown_docks_a3_n_d38_0_0.vlm"},
	{0xA3452023155280E8ULL, "isa211_isabela6069631.cl"},
	{0xA3456CDCD4066CF8ULL, "vxlt_warehouse_t3_0_0.vlm"},
	{0xA345C321F3661283ULL, "vxw_bonepitmining_a2_2ar_0_0.vlm"},
	{0xA3463C9A3011D32DULL, "vxw_bonepitdestroyed_a3_1ue_0_0.vlm"},
	{0xA3463D6ECA7E49EBULL, "vxlt_0_lowtown_docks_a3_n_bqn_0_0.vlm"},
	{0xA34665A2EA70124FULL, "6134850_m.fxe"},
	{0xA346AA5A004FA1E9ULL, "6115788_f.fxe"},
	{0xA346CF216A467AA7ULL, "vxdr_dungeon_1t1_0_1.vlm"},
	{0xA346D81AC0470CE5ULL, "hlac_ht_chantry_2534_1y.dds"},
	{0xA346EDAF1BDC50C1ULL, "art\\vfx\\combat\\vfx_d_arc_aoe_cast_c.impa.ani"},
	{0xA3472846AC72F864ULL, "6156509_m.fxe"},
	{0xA347D6388EBA4609ULL, "vxw_woundedcoast_v2_0_0.vlm"},
	{0xA348A117FB42BEE0ULL, "vxlt_blackemporium_15u_0_0.vlm"},
	{0xA34949D96EF6B6E3ULL, "6121943_f.fxe"},
	{0xA3494F0EB2E341BFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_win.phy"},
	{0xA349F73799885308ULL, "6041112_f.fxe"},
	{0xA34A06A2F67791C4ULL, "6122814_m.fxe"},
	{0xA34A3BCF317CB0E8ULL, "mer120_hubert6077583.cl"},
	{0xA34AACFC22C5C7C0ULL, "533051405.ogg"},
	{0xA34B708EB98384FDULL, "dae000_mother6028511.cl"},
	{0xA34C52E601CADF4DULL, "art\\vfx\\creatures\\fxc_ogre_stmp_c.anb"},
	{0xA34CAB1859F316E5ULL, "vxht_chantry_35t_0_0.vlm"},
	{0xA34CAD6027BB2819ULL, "84151774.ogg"},
	{0xA34CDF5D28EB0D37ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st105a_0t.dds"},
	{0xA34D05DDF0EC304CULL, "zz_vrc_debug6200185.cl"},
	{0xA34D4A077EC879D5ULL, "zz_bet_debug6113431.cl"},
	{0xA34DF8AF1BE28263ULL, "art\\vfx\\combat\\vfx_d_arc_aoe_cast_c.impa.gad"},
	{0xA34E2F18DCD010FBULL, "vxlt_0_lowtown_docks_ban_0_0.vlm"},
	{0xA34E3D02556B892DULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc101a_0t.dds"},
	{0xA34E5246C1AF5C44ULL, "6199752_f.fxe"},
	{0xA34E601D9F9217D0ULL, "6163835_m.fxe"},
	{0xA34E7C28F31034C3ULL, "ht_0_hightown_n_59_0.rml"},
	{0xA34EE6F3138753D7ULL, "6147214_m.fxe"},
	{0xA34FE022A8F6CD81ULL, "hlac_lt_undercitydungeon_-1_50y.dds"},
	{0xA3500C5EF10BCE8EULL, "6125343_m.fxe"},
	{0xA3507DCED6581D74ULL, "120967797.ogg"},
	{0xA350E9015A44E844ULL, "14563287.ogg"},
	{0xA350E99CF1270B43ULL, "art\\levels\\ztools\\default\\generic\\0rl_basic.msh"},
	{0xA350F79CF127229DULL, "art\\levels\\ztools\\default\\generic\\0rl_basic.mmh"},
	{0xA350FB9CF12729AEULL, "art\\levels\\ztools\\default\\generic\\0rl_basic.mao"},
	{0xA3512B4EE2BF832BULL, "one280_emrys6105383.cl"},
	{0xA35142495AD5F456ULL, "217545417.ogg"},
	{0xA3514BB134016483ULL, "897431898.ogg"},
	{0xA3520491559C689FULL, "hlac_lt_0_lowtown_a3_n_c_-2_6y.dds"},
	{0xA35293FD2EAA7CB6ULL, "55074896.ogg"},
	{0xA3529D6494993D39ULL, "434331328.ogg"},
	{0xA352E71DCFFA78E9ULL, "vxw_blightlands_e9_0_0.vlm"},
	{0xA352FD4B5AFFBAEBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubrush_n.dds"},
	{0xA3530ECAAFC2A3CAULL, "mag210_arianni6021231.cl"},
	{0xA353203C2A3CDF21ULL, "vxca_caverns2prc_9dv_0_1.vlm"},
	{0xA353DEB9C5FC8578ULL, "vxlt_warehouse_y0_0_0.vlm"},
	{0xA353FEB869F554D8ULL, "vxlt_backalley_n_1wa_0_0.vlm"},
	{0xA3540D28F07DA474ULL, "lgt160_ntb_elf_conflict6163356.cl"},
	{0xA354A845A7CF1BA2ULL, "mrl000_marethari6129245.cl"},
	{0xA354AF4BBC74626BULL, "vxlt_0_lowtown_docks_a3_n_bta_0_0.vlm"},
	{0xA354D5AF1BE86548ULL, "art\\vfx\\combat\\vfx_d_arc_aoe_cast_c.impa.evt"},
	{0xA35545D0D789DD8DULL, "w_encounter2_camp_62_0.phy"},
	{0xA35561366E39BA82ULL, "dae000_arishok6082028.cl"},
	{0xA35617C08CE5CCADULL, "vxca_cavern1_haunted_i2j_0_0.vlm"},
	{0xA356450DD4EAB967ULL, "vxlt_gallowscourtyard_n_3wz_0_0.vlm"},
	{0xA356481580B05E89ULL, "vxw_woundedcoast_s1_0_0.vlm"},
	{0xA3565D4C5E2513C0ULL, "gen00fl_dog6146950.cl"},
	{0xA356B8B030C25058ULL, "vxw_blightlands_119_0_0.vlm"},
	{0xA356E05BCD8AA74FULL, "vxlt_gallowscourtyard_n_4w9_0_0.vlm"},
	{0xA357C8155CC26E89ULL, "463434137.ogg"},
	{0xA358796706344586ULL, "vxw_bonepitdestroyed_a3_19n_0_0.vlm"},
	{0xA358CC8EEA2F8E78ULL, "character_character_p218.psh.fxo.dx11"},
	{0xA3593DAA951A4485ULL, "vxht_mansion2_329_0_0.vlm"},
	{0xA3595AD5F358FD3FULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_1x.dds"},
	{0xA3598DD5EF82120EULL, "art\\characters\\creatures\\corpse\\uni_cora.mmh"},
	{0xA359A7D5EF823EFCULL, "art\\characters\\creatures\\corpse\\uni_cora.msh"},
	{0xA359BEDB3C12422FULL, "vxdr_primevalentrance_f7_0_0.vlm"},
	{0xA359F1A0B7442AB0ULL, "follower_banter6163658.cl"},
	{0xA35A0D1BB95BFB72ULL, "gen00fl_merrill6108474.cl"},
	{0xA35A500EB2F1B6C4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_win.mmh"},
	{0xA35A540EB2F1BD17ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_win.mao"},
	{0xA35A660EB2F1DC66ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_win.msh"},
	{0xA35A6B7ABF864F34ULL, "6115320_m.fxe"},
	{0xA35A74D3E7EFD217ULL, "vxca_cavern1_haunted_fmv_0_0.vlm"},
	{0xA35A9D5B88963A7EULL, "hlac_ht_0_hightown_a2_-65537_2z.dds"},
	{0xA35AA7894DF176B4ULL, "dae000ip_crowd_em3_talk2.utp"},
	{0xA35AF8A21AC53B2AULL, "ca_caverns1_mines_24424_ll.mmh"},
	{0xA35B6F0D6DAE88A7ULL, "6154571_m.fxe"},
	{0xA35BABD8EB8B49D1ULL, "ht_0_hightown_a2_45_0.rml"},
	{0xA35BB4C5F4AC5D23ULL, "80135235.ogg"},
	{0xA35C11544FB6DDC8ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyn_0n.dds"},
	{0xA35C46D0D78FFE39ULL, "w_encounter2_camp_62_0.rml"},
	{0xA35CA0F5CEC6297BULL, "vxlt_0_lowtown_docks_bda_0_0.vlm"},
	{0xA35CEA3685F8207EULL, "vxca_cavern1_haunted_iih_0_0.vlm"},
	{0xA35D1EDA0970939CULL, "307689706.ogg"},
	{0xA35D668CE9301178ULL, "6092164_m.fxe"},
	{0xA35D71AF6EEABFE1ULL, "mcr310_argument6075100.cl"},
	{0xA35D98C771BC1485ULL, "6209854_f.fxe"},
	{0xA35DBCD64A9673D1ULL, "gen00fl_fenris6127003.cl"},
	{0xA35E06036DCF4894ULL, "425944400.ogg"},
	{0xA35E13A844A0BF91ULL, "w_bonepitdestroyed_a3_67_0.msh"},
	{0xA35E21A844A0D7ACULL, "w_bonepitdestroyed_a3_67_0.mao"},
	{0xA35E2DA844A0EC4FULL, "w_bonepitdestroyed_a3_67_0.mmh"},
	{0xA35E3E6F8BAD53FBULL, "vxht_0_hightown_n_cb4_0_0.vlm"},
	{0xA35E44569AE6219AULL, "1015190014.ogg"},
	{0xA35E8346F3B0B838ULL, "mag320_orsino6071159.cl"},
	{0xA35F0B30278E4D24ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.dura.ani"},
	{0xA35F34DB9DFA3782ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobd.phy"},
	{0xA35FD930BA22FA64ULL, "6087539_m.fxe"},
	{0xA36009517A156249ULL, "art\\vfx\\combat\\vfx_str_att_front_r_leg_c.impa.ani"},
	{0xA3606204C31AD2F1ULL, "hlac_ht_mansion2_818_65z.dds"},
	{0xA3609A294F2D236EULL, "vxw_sundermount_2de_0_0.vlm"},
	{0xA360C38FEA371CD6ULL, "193742516.ogg"},
	{0xA36158FAC1F09169ULL, "vxw_blightlands_n8_0_0.vlm"},
	{0xA361835256C342B3ULL, "gen00fl_carver6108909.cl"},
	{0xA3622337031906B6ULL, "qun110_petrice_alley.stg"},
	{0xA36260D7326E112DULL, "gen00fl_anders6119404.cl"},
	{0xA362BC4F0CFCCD5CULL, "dae000_dalish_apprentices6147533.cl"},
	{0xA363484E156BC737ULL, "lt_0_lowtown_docks_a3_30_0.anb"},
	{0xA36386FACF67FD8EULL, "6026971_f.fxe"},
	{0xA363A791D45DC7C1ULL, "vxlt_hangedman_n_1vz_0_0.vlm"},
	{0xA36467F742F1BB25ULL, "620222450.ogg"},
	{0xA364CF59322D050EULL, "vxw_woundedcoast_10b_0_0.vlm"},
	{0xA36508302792B386ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.dura.gad"},
	{0xA3651327B6959574ULL, "gen00fl_carver6108845.cl"},
	{0xA3651863F405FC18ULL, "gen_im_trs_tmp_insignia_04.uti"},
	{0xA3652A8D22B868D8ULL, "vxw_blightlands_146_0_0.vlm"},
	{0xA3656618D18C92A2ULL, "qun110cr_arishok_guard.utc"},
	{0xA3656AD6F363C811ULL, "6081363_m.fxe"},
	{0xA365A02145220464ULL, "dae000_bartrand6032022.cl"},
	{0xA3661F3D562E00EDULL, "gen00fl_isabela6118364.cl"},
	{0xA36624517A19FBCBULL, "art\\vfx\\combat\\vfx_str_att_front_r_leg_c.impa.gad"},
	{0xA3662E32B39537D8ULL, "368975990.ogg"},
	{0xA36701679306DF79ULL, "896837767.ogg"},
	{0xA36759A1C63CF8A0ULL, "554918545.ogg"},
	{0xA3679704FB1E7A25ULL, "art\\levels\\races\\proxy\\props\\armoire\\prp_zzarmoire.phy"},
	{0xA367F9633295197DULL, "6121663_m.fxe"},
	{0xA3689D6EC34A8F43ULL, "6155309_m.fxe"},
	{0xA36A010F88769319ULL, "6032206_m.fxe"},
	{0xA36A6E510AC77FA4ULL, "character_character_p58.psh.fxo.dx11"},
	{0xA36A73CAA7F50784ULL, "t3_wdg_b00.tnt"},
	{0xA36B0401AAC64B28ULL, "443169827.ogg"},
	{0xA36BA3E7FD6046D9ULL, "vxht_chantry_n_2c8_0_0.vlm"},
	{0xA36C03302798CA65ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.dura.evt"},
	{0xA36C4489245E0AA2ULL, "vxlt_0_lowtown_a3_n_c_ewg_0_0.vlm"},
	{0xA36CB042DE527C96ULL, "dae011_conspiracy_mob6149560.cl"},
	{0xA36CC05EFAA9A788ULL, "vxht_brothel_n_6dg_0_0.vlm"},
	{0xA36CCCFEDF332084ULL, "6163350_m.fxe"},
	{0xA36D1B2A49ADAD1EULL, "6051645_f.fxe"},
	{0xA36D21517A201530ULL, "art\\vfx\\combat\\vfx_str_att_front_r_leg_c.impa.evt"},
	{0xA36D7539C09609A7ULL, "vxw_bonepit_1l8_0_0.vlm"},
	{0xA36DB8FDA0005BA3ULL, "ht_0_hightown_n_18_0_lc.mmh"},
	{0xA36DD6F054B9A771ULL, "dae000_brothel_orana6160881.cl"},
	{0xA36DE42DDA1CCEA8ULL, "6180397_m.fxe"},
	{0xA36DEF16258A8E7EULL, "hlac_w_bonepitmining_a2_196613_0x.dds"},
	{0xA36DF0BDDEEA94F8ULL, "hlac_lt_0_lowtown_n_-65537_1z.dds"},
	{0xA36E123D6277C226ULL, "llac_ca_caverns1_24426_5.dds"},
	{0xA36E1F6C00775FF7ULL, "dae310ip_at_mrl_home.utp"},
	{0xA36F0C0641233BEEULL, "vxw_sundermount_2if_0_0.vlm"},
	{0xA36F60A75240F77BULL, "follower_banter6163549.cl"},
	{0xA36F63693C26599EULL, "qun111_ketojan6040011.cl"},
	{0xA36FA8BB890CBBF4ULL, "6031152_m.fxe"},
	{0xA36FFCAC5A5C3631ULL, "6037627_m.fxe"},
	{0xA3702C3CCBC9732AULL, "lt_0_lowtown_docks_a3_n_lc.mmh"},
	{0xA370339F37037D09ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_meda_0s.dds"},
	{0xA3705ED6720B3D38ULL, "lt_0_lowtown_docks_n_11_0.rml"},
	{0xA370A4335AD68454ULL, "530588069.ogg"},
	{0xA370A92FAB48D68EULL, "gen00fl_merrill6106807.cl"},
	{0xA370B4208C0B9C06ULL, "hlac_ht_mansion2_818_40x.dds"},
	{0xA370B82C4CEE7544ULL, "gen00fl_isabela_init6110256.cl"},
	{0xA371571DB600DA72ULL, "248365497.ogg"},
	{0xA371EF7C8A29512EULL, "mrl000_marethari6142152.cl"},
	{0xA372418C1D4CF7F0ULL, "lgt251ip_door_trap1.utp"},
	{0xA37288B6AB22568CULL, "vxca_cavern1_haunted_j7c_0_0.vlm"},
	{0xA372B1690AD67C12ULL, "6050111_m.fxe"},
	{0xA372BE458CEFF80CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardleft.ani"},
	{0xA372EEF70813F21FULL, "vxca_caverns2prc_8ka_0_0.vlm"},
	{0xA3735D9C5715C999ULL, "vxht_chantry_2t3_0_1.vlm"},
	{0xA37413F6FFF40963ULL, "hlac_ht_chantry_n_2534_3z.dds"},
	{0xA3742BAAD699DAFAULL, "6177353_m.fxe"},
	{0xA374DF08953A3978ULL, "244884600.ogg"},
	{0xA375003290394D61ULL, "vxca_cave_2gk_0_0.vlm"},
	{0xA3754BA61ED501E5ULL, "465647324.ogg"},
	{0xA375D57EAA4FA022ULL, "bedsingle_d.dds"},
	{0xA3763764E1A009CDULL, "vxca_caverns2_80o_0_0.vlm"},
	{0xA376C66A54FE5786ULL, "vxca_cavern1_haunted_ikj_0_0.vlm"},
	{0xA376DB70B022C8AFULL, "head_tint_masks.gda"},
	{0xA376DCFA21AEDCB4ULL, "mag291tr_nt_ragedemons.utt"},
	{0xA376E4ED3F30AAECULL, "6116314_m.fxe"},
	{0xA3772447271333C1ULL, "vxht_mansion_n_3yk_0_0.vlm"},
	{0xA37726253EC5258DULL, "w_woundedcoast_104_0.phy"},
	{0xA377E2775DCE9EB6ULL, "vxlt_undercityserialkiller_1uh_0_0.vlm"},
	{0xA37816002181E0BAULL, "follower_banter6161148.cl"},
	{0xA37870AD51C0CA8EULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.impa.ani"},
	{0xA378D441115F1CECULL, "llac_dr_0_deeproads_5842_5.dds"},
	{0xA378F933C87AA8F9ULL, "gen00fl_aveline6106915.cl"},
	{0xA3790A4813F9716AULL, "416960583.ogg"},
	{0xA3790D89E5D4A209ULL, "zz_vrc_debug6200205.cl"},
	{0xA37918679A7078C3ULL, "6089686_m.fxe"},
	{0xA3799D03A073C462ULL, "sib131_slaver_lord6165207.cl"},
	{0xA379CB458CF62D4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardleft.gad"},
	{0xA379CE313217449EULL, "9095085.ogg"},
	{0xA37ABE7BE29B9A6FULL, "gen00fl_anders6107661.cl"},
	{0xA37AD34A5F696E19ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrga_1.mao"},
	{0xA37B323BEC9FC008ULL, "vxht_brothel_n_6ad_0_0.vlm"},
	{0xA37BB5FF34781DD8ULL, "mag101_grace6034221.cl"},
	{0xA37BCA907BC46120ULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure_12.pwk"},
	{0xA37C362BCAA63827ULL, "hlac_w_bonepitdestroyed_a3_458756_0z.dds"},
	{0xA37C560ACC12E728ULL, "6185097_m.fxe"},
	{0xA37CB1FB436F305BULL, "vxlt_warehouse_n_11u_0_0.vlm"},
	{0xA37CE6C5481B8B99ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_coasttree.msh"},
	{0xA37CF0C5481B9C87ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_coasttree.mmh"},
	{0xA37CF4C5481BA3D4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_coasttree.mao"},
	{0xA37D2DFEB9E0D1EBULL, "one350_variel6076828.cl"},
	{0xA37D44EA75AF0A8EULL, "ss_hawkecu2.cut"},
	{0xA37D6EC10ECFFB08ULL, "6043863_m.fxe"},
	{0xA37D7DE33319546EULL, "vxw_sundermount_2rg_0_0.vlm"},
	{0xA37D9DBB967AD1CEULL, "809111792.ogg"},
	{0xA37DA3D3CE9CCAC3ULL, "hlac_dr_0_deeproads_5842_0z.dds"},
	{0xA37DCBB226E97596ULL, "one220_yevhen6103022.cl"},
	{0xA37DE1FB752301ADULL, "vxdr_primevaltomb_8mf_0_0.vlm"},
	{0xA37E7CA8DA8DDEADULL, "6153622_m.fxe"},
	{0xA37EADDB2970D32EULL, "mag291_ntbanter6056739.cl"},
	{0xA3808B4BB849F8C1ULL, "vxlt_hangedman_n_1hx_0_0.vlm"},
	{0xA38092ACD7401B5BULL, "gen00fl_varric6088852.cl"},
	{0xA380BBFAE5A66308ULL, "w_encounter1_25_0.mmh"},
	{0xA380C6458CFC432DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkforwardleft.evt"},
	{0xA380C7FAE5A677ABULL, "w_encounter1_25_0.mao"},
	{0xA380D9FAE5A69632ULL, "w_encounter1_25_0.msh"},
	{0xA38104266FA80E29ULL, "llac_lt_0_lowtown_a2_n_65534_2.dds"},
	{0xA3810CE477ADA399ULL, "llac_w_sundermount_-131077_0.dds"},
	{0xA38140C339FAB346ULL, "138195599.ogg"},
	{0xA38204B2FAE2D61FULL, "6138818_m.fxe"},
	{0xA38258D35EC95159ULL, "one350_terath6087120.cl"},
	{0xA38274A45B58B81BULL, "6057106_m.fxe"},
	{0xA382BEEA965B9C46ULL, "vxlt_0_lowtown_a3_n_c_fhe_0_0.vlm"},
	{0xA3836EDC86C71465ULL, "6076780_f.fxe"},
	{0xA38398D2A3D9F159ULL, "follower_barks6153730.cl"},
	{0xA383C47344FE8D4FULL, "6111169_m.fxe"},
	{0xA384056984E2F729ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_009a_0d.dds"},
	{0xA38474805FBD0C8AULL, "6202154_m.fxe"},
	{0xA3850FDB9E1AC836ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobd.mao"},
	{0xA3851BDB9E1ADCD5ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobd.mmh"},
	{0xA3851DDB9E1AE04BULL, "art\\characters\\playercharacter\\humanmale\\cth_nobd.msh"},
	{0xA385638B95BEEEF8ULL, "6027656_m.fxe"},
	{0xA38568AD51CB47DFULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.impa.evt"},
	{0xA3856D6AA03C0CE3ULL, "mcr310_argument6182446.cl"},
	{0xA385E69B6A83A2E0ULL, "mag371_elsa6041139.cl"},
	{0xA3865AE0371B1D12ULL, "416478071.ogg"},
	{0xA386CB0665EFC454ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_ltrb.anb"},
	{0xA3873275A0ED43EAULL, "541110400.ogg"},
	{0xA3873F9D70336ADCULL, "gen00fl_sebastian6179776.cl"},
	{0xA3876F774CB3332AULL, "167627599.ogg"},
	{0xA387A0A86265DB1EULL, "follower_banter6163999.cl"},
	{0xA387C4B442A656FBULL, "vxca_cavern1_haunted_i1q_0_0.vlm"},
	{0xA388036A42F208DBULL, "dae000_orana6149092.cl"},
	{0xA3882749C3C2C7FAULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgta_0d.dds"},
	{0xA38846D9F96FDDC2ULL, "6105446_m.fxe"},
	{0xA388543703A5212EULL, "armycontrol.gfx"},
	{0xA388C9BCFEFD7A8CULL, "6094971_m.fxe"},
	{0xA3895B785BFC934BULL, "vxlt_0_lowtown_a3_n_c_f66_0_0.vlm"},
	{0xA389A418DE95D888ULL, "vxht_brothel_n_6je_0_0.vlm"},
	{0xA389B6CCCB96C180ULL, "hlac_lt_0_lowtown_docks_a3_-262143_0x.dds"},
	{0xA389D96C56F6023BULL, "art\\characters\\weapons\\staff\\w_stf_010a.msh"},
	{0xA389EB6C56F62086ULL, "art\\characters\\weapons\\staff\\w_stf_010a.mao"},
	{0xA389F76C56F63525ULL, "art\\characters\\weapons\\staff\\w_stf_010a.mmh"},
	{0xA389FD731AFF03FBULL, "6042283_m.fxe"},
	{0xA38A226868047494ULL, "6151175_m.fxe"},
	{0xA38B4888B3B02CB8ULL, "564693603.ogg"},
	{0xA38B802DAE956F93ULL, "6068557_m.fxe"},
	{0xA38C49AD51D131C8ULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.impa.gad"},
	{0xA38DEC1684704DF4ULL, "86924926.ogg"},
	{0xA38DECB90ADCE755ULL, "vxht_0_hightown_n_8pg_0_0.vlm"},
	{0xA38E5588D36175C9ULL, "vrc321_varric6086886.cl"},
	{0xA38EE5E868561A59ULL, "dae011_gossips6143754.cl"},
	{0xA38F454E1590FD4BULL, "lt_0_lowtown_docks_a3_30_0.rml"},
	{0xA38F4F9EE01FC393ULL, "6098086_m.fxe"},
	{0xA38FD2B0EC00231FULL, "vxca_caverns2prc_8qc_0_0.vlm"},
	{0xA39027C75CAD833AULL, "lgt320_mis_resp_swordhunter6150340.cl"},
	{0xA390506E67BF0A33ULL, "810607414.ogg"},
	{0xA390E43738A2B6BDULL, "6104943_m.fxe"},
	{0xA391270712E80EE1ULL, "6111134_m.fxe"},
	{0xA3912C67497DF7A8ULL, "374261382.ogg"},
	{0xA391477B414AB165ULL, "llac_lt_0_lowtown_a2_-1_4.dds"},
	{0xA39148CF2DD72A47ULL, "vxht_brothel_n_688_0_0.vlm"},
	{0xA39157B5EECD7707ULL, "dae000_epic_merchant_36179494.cl"},
	{0xA39197673AD52D43ULL, "hlac_w_sundermount_-65532_0x.dds"},
	{0xA391CE7D95B8621EULL, "hlac_w_encounter2_-65531_0y.dds"},
	{0xA391E3EC74257E61ULL, "vxca_cave_2mi_0_0.vlm"},
	{0xA3929EB568E18DF5ULL, "6105427_m.fxe"},
	{0xA392E65D51B3250AULL, "6209054_m.fxe"},
	{0xA39383BBA711C3C7ULL, "vxht_0_hightown_a2_avi_0_0.vlm"},
	{0xA3939BB4C52B70BEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\sister_pcopening_rog.ani"},
	{0xA393A7EDA1D1C1D1ULL, "isa211_isabela6072557.cl"},
	{0xA393A866FCA2FE20ULL, "vfx_vfx_p3.vsh.fxo.dx11"},
	{0xA3942C1B21E3A756ULL, "one110_gustav6089619.cl"},
	{0xA3948C738C5F4228ULL, "gen00fl_anders6113113.cl"},
	{0xA394E4F03D0B3686ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_c.anb"},
	{0xA3951D2CE67B6CD0ULL, "389911137.ogg"},
	{0xA3959E8643769F51ULL, "177703644.ogg"},
	{0xA395C9CEEBCD93F2ULL, "873124315.ogg"},
	{0xA396008F9514596CULL, "hlac_lt_0_lowtown_-3_1z.dds"},
	{0xA3962120A4E32FCBULL, "dae000cr_merchant_wandering.utc"},
	{0xA39657BCC3610B60ULL, "vxlt_house_e7_0_0.vlm"},
	{0xA396CDCA89F79C03ULL, "hlac_lt_house_-1_3z.dds"},
	{0xA396FB78FAD00B42ULL, "615943310.ogg"},
	{0xA39701BEEA17E1B8ULL, "6155715_m.fxe"},
	{0xA397778A23744EFDULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffy2_n.dds"},
	{0xA397924A6FDCFAB1ULL, "233571327.ogg"},
	{0xA397B1DB1005A6B6ULL, "lgt120_orz_bln_official6176131.cl"},
	{0xA397BC56D65A8BE5ULL, "qcr220_grey_wardens6060464.cl"},
	{0xA397E25B9554401FULL, "hlac_ht_0_hightown_65533_2z.dds"},
	{0xA397EC8BB86650B7ULL, "hlac_ht_chantry_n_2534_9x.dds"},
	{0xA397FD01E69ECBC7ULL, "197466976.ogg"},
	{0xA3980AFCBE73E35AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_1groupcrag.msh"},
	{0xA39810F6B5F4FF0BULL, "vxlt_foundry_n_z9_0_0.vlm"},
	{0xA39818FCBE73FA93ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_1groupcrag.mao"},
	{0xA3981CFCBE7401A0ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_1groupcrag.mmh"},
	{0xA39839873465E9E5ULL, "214649931.ogg"},
	{0xA398506D90B7D8CCULL, "vxlt_darktown_5na_0_0.vlm"},
	{0xA3986EC6B3E316E6ULL, "6125985_m.fxe"},
	{0xA3989C75AA6EB0A1ULL, "6112737_m.fxe"},
	{0xA398F618DD0542BFULL, "225326151.ogg"},
	{0xA39937E65F17E373ULL, "mer320_hubert6053685.cl"},
	{0xA39A619D1705856EULL, "vxw_sundermount_2xy_0_0.vlm"},
	{0xA39A92D5AB887829ULL, "328274535.ogg"},
	{0xA39AC2FA6BF04EEFULL, "hlac_ca_cavern1_haunted_24426_5z.dds"},
	{0xA39ACA062C9EE718ULL, "llac_w_bonepitmining_a2_6_0.dds"},
	{0xA39B0D2C60C2B9ECULL, "hlac_ca_caverns1_mines_24424_26x.dds"},
	{0xA39B7888A4733FE7ULL, "6179941_m.fxe"},
	{0xA39BA1F55D54211DULL, "6018615_m.fxe"},
	{0xA39C387BCD92E630ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_basic_1.mao"},
	{0xA39C3FD6723043F0ULL, "lt_0_lowtown_docks_n_11_0.anb"},
	{0xA39C4257261D8E53ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston_12.pwk"},
	{0xA39C509DE1CA4D88ULL, "vxw_sundermount_20w_0_0.vlm"},
	{0xA39CB03B9670322FULL, "197588773.ogg"},
	{0xA39CE4665580F2D9ULL, "vxca_cave_18i_0_0.vlm"},
	{0xA39D255C926E0B17ULL, "vxlt_0_lowtown_n_byx_0_0.vlm"},
	{0xA39D96924B7E1CABULL, "6153439_m.fxe"},
	{0xA39DB5E037E6431CULL, "6018709_f.fxe"},
	{0xA39E448DDDF63B9FULL, "vxca_caverns2prc_8zb_0_0.vlm"},
	{0xA39EA8067DDD6EEEULL, "vxht_chantry_1yq_0_0.vlm"},
	{0xA39F92CB69F3906BULL, "vxlt_0_lowtown_a2_n_gon_0_1.vlm"},
	{0xA39FF5D5FD7A5DC8ULL, "568448461.ogg"},
	{0xA39FF844C18C4F42ULL, "674289347.ogg"},
	{0xA39FF934AB8F4071ULL, "vxlt_warehouse_10o_0_0.vlm"},
	{0xA3A037E69D0C6E5DULL, "dae000_gen_turn_in_26198696.cl"},
	{0xA3A0C0D45FA8AC9FULL, "ch1em.psh.fxo"},
	{0xA3A0E1C6781D434DULL, "382268044.ogg"},
	{0xA3A12F884E244A17ULL, "vxlt_0_lowtown_a2_n_c_e1i_0_0.vlm"},
	{0xA3A1A83DA9F32542ULL, "6108953_m.fxe"},
	{0xA3A1F5989907DC47ULL, "vxht_0_hightown_a2_aqh_0_0.vlm"},
	{0xA3A2357D9B97305DULL, "6085842_m.fxe"},
	{0xA3A23E05C133E7D3ULL, "vxlt_gallowstemplar_f_4o3_0_0.vlm"},
	{0xA3A2E7F03D177931ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_c.mmh"},
	{0xA3A398E8CAD73DA4ULL, "vxca_caverns1_jfm_0_0.vlm"},
	{0xA3A40FDDEAFDDA48ULL, "face_tsb_p0.gsh.fxo.dx11"},
	{0xA3A456EA4E15A887ULL, "140046358.ogg"},
	{0xA3A4D5E73C3E372CULL, "art\\characters\\creatures\\pridedemon\\animation\\op.ani"},
	{0xA3A5C4C71DAD0A91ULL, "606959646.ogg"},
	{0xA3A5F47F58A78D4EULL, "1028548958.ogg"},
	{0xA3A61B26711548CBULL, "hlac_ht_playershouse_n_855_34y.dds"},
	{0xA3A6EC0D3CAF9264ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_2z.dds"},
	{0xA3A6F997395B987AULL, "pro000_refugee_female6052089.cl"},
	{0xA3A70A03EB517E6CULL, "84731256.ogg"},
	{0xA3A764F8206B0B6CULL, "6153591_m.fxe"},
	{0xA3A79BE126BE192BULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_act_c.impa.evt"},
	{0xA3A7B4B4C53C4438ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\sister_pcopening_rog.gad"},
	{0xA3A7D67E7481C12BULL, "vxca_caverns1_hl2_0_0.vlm"},
	{0xA3A8013A80931333ULL, "339924328.ogg"},
	{0xA3A87B4EACE6D85CULL, "isa311_party_comments6117812.cl"},
	{0xA3A972D0FE223F27ULL, "abi_r_global.wbd"},
	{0xA3A999FAA9A071BBULL, "688196759.ogg"},
	{0xA3A9A2F86B41AC31ULL, "combat_deathblow_2.gda"},
	{0xA3A9CB5D979B17F5ULL, "vxlt_blackemporium_3r_0_1.vlm"},
	{0xA3A9CF59F1ED0AA1ULL, "mag321_comtesse_de_launcet6043260.cl"},
	{0xA3A9F46F64A41F09ULL, "192526815.ogg"},
	{0xA3AAA129D22ABC46ULL, "dae000_bartrand6032019.cl"},
	{0xA3AAC27AD3C06608ULL, "vxw_sundermount_25t_0_0.vlm"},
	{0xA3AAC5A2246EC64BULL, "mag291_demon_marethari6029045.cl"},
	{0xA3AAD0784E528A65ULL, "llac_lt_0_lowtown_docks_131071_4.dds"},
	{0xA3AB1F53F862235BULL, "837584473.ogg"},
	{0xA3AB3C3D7E496362ULL, "6085745_f.fxe"},
	{0xA3ABE2E73C446C6EULL, "art\\characters\\creatures\\pridedemon\\animation\\op.gad"},
	{0xA3ABEEBE53364097ULL, "weapon_rendertoshadowmap_p1.vsh.fxo.dx11"},
	{0xA3AC086F3D74352BULL, "6156739_m.fxe"},
	{0xA3AC258AB21BFF2EULL, "zz_mag_debug6051790.cl"},
	{0xA3ACCB2CC35FAB79ULL, "hlac_ca_caverns1_24424_2z.dds"},
	{0xA3AD42B162492F92ULL, "572792235.ogg"},
	{0xA3AD5C83753A94E3ULL, "vxlt_0_lowtown_docks_a3_brh_0_0.vlm"},
	{0xA3ADB2E67D81EDA9ULL, "1025721379.ogg"},
	{0xA3AE316F5C83F662ULL, "194601247.ogg"},
	{0xA3AE395D9694C489ULL, "art\\vfx\\combat\\vfx_c_ogr_prone_p.mmh"},
	{0xA3AE4C088F0F7217ULL, "vxlt_0_lowtown_h82_0_0.vlm"},
	{0xA3AE5E444B6DD4A7ULL, "6049466_m.fxe"},
	{0xA3AE80AEFF548ADBULL, "vxht_0_hightown_a2_9up_0_0.vlm"},
	{0xA3AEB4E126C462ECULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_act_c.impa.gad"},
	{0xA3AF5438008DD15FULL, "follower_banter6172316.cl"},
	{0xA3AF8EEF2AE38BFBULL, "vxdr_primevaltomb_8nu_0_0.vlm"},
	{0xA3B03BB22A5406EBULL, "vxht_bartrand_n_5re_0_0.vlm"},
	{0xA3B042B20410725BULL, "dae021_sister_samea_sta.bnk"},
	{0xA3B067758AFDF4C7ULL, "vxht_0_hightown_a2_alk_0_0.vlm"},
	{0xA3B08164C14B594DULL, "fenris_4.cut"},
	{0xA3B084CC5900FFE5ULL, "one140cr_charade6137032.cl"},
	{0xA3B0AFE2B32A0053ULL, "vxlt_gallowstemplar_f_4t2_0_0.vlm"},
	{0xA3B1242A3D3DC693ULL, "hlac_lt_0_lowtown_n_65537_2y.dds"},
	{0xA3B298926B5115B5ULL, "49946476.ogg"},
	{0xA3B2C7E433FA478BULL, "one220_yevhen6104150.cl"},
	{0xA3B31F915F83105BULL, "39301137.ogg"},
	{0xA3B34B1649C3BA38ULL, "vxlt_gallowscourtyard_7c2_0_0.vlm"},
	{0xA3B3DA9F57928BECULL, "vxlt_0_lowtown_ett_0_0.vlm"},
	{0xA3B444B204145E1BULL, "dae021_sister_samea_sta.cnv"},
	{0xA3B4D0BA242896AAULL, "6098216_m.fxe"},
	{0xA3B4EF849569103FULL, "6179047_f.fxe"},
	{0xA3B510AD88480929ULL, "zz_fol_debug6092323.cl"},
	{0xA3B5342774A409CCULL, "vxlt_darktown_5tg_0_0.vlm"},
	{0xA3B5469412B4E613ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\dodge_object.ani"},
	{0xA3B56AB06291C43CULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.cess.ani"},
	{0xA3B5865B8E533211ULL, "llac_lt_0_lowtown_docks_a3_-131075_4.dds"},
	{0xA3B5B3E126CA7FEAULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_act_c.impa.ani"},
	{0xA3B650FE9ED2EE2BULL, "one350_vinell6087339.cl"},
	{0xA3B74A4045F5752BULL, "vxht_keep_5ih_0_0.vlm"},
	{0xA3B7A96F3CFB63ADULL, "vxdr_primevaltomb_8yb_0_0.vlm"},
	{0xA3B7F4F4ADA21F47ULL, "mcr370_cut_animate_guardian.bnk"},
	{0xA3B84CBA45FAE38BULL, "vxlt_gallowstemplar_f_5t3_0_0.vlm"},
	{0xA3B8BD6C0B6D06FCULL, "vxlt_0_lowtown_docks_a3_c3u_0_0.vlm"},
	{0xA3B8E437F9BC4177ULL, "hlac_lt_0_lowtown_docks_n_-131073_5y.dds"},
	{0xA3B904C3BEB35BE0ULL, "vxlt_warehouse_8g_0_0.vlm"},
	{0xA3B98F53BF393FA8ULL, "380911262.ogg"},
	{0xA3BA192679D6BF33ULL, "dae000_arishok6068544.cl"},
	{0xA3BB16F4ADA48EBFULL, "mcr370_cut_animate_guardian.cnv"},
	{0xA3BB37F2D99D8557ULL, "vxlt_0_lowtown_a2_n_f4s_0_0.vlm"},
	{0xA3BBF0A99C374E8FULL, "6097358_f.fxe"},
	{0xA3BC0203498CE05FULL, "6152212_m.fxe"},
	{0xA3BC0B171EC10A27ULL, "328604695.ogg"},
	{0xA3BC2348E867898CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_103a_d.dds"},
	{0xA3BC6034FD4CC0F5ULL, "hlac_lt_gallowscourtyard_n_-65538_0z.dds"},
	{0xA3BC77B06297F97EULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.cess.gad"},
	{0xA3BC9E6C710219C7ULL, "880547412.ogg"},
	{0xA3BD0ECEF2CC8A70ULL, "gen00fl_varric6126322.cl"},
	{0xA3BD2276FFC0964FULL, "vxht_mansion2_1sz_0_0.vlm"},
	{0xA3BDDF9F4598A27CULL, "6054269_m.fxe"},
	{0xA3BE0E9B2E22081BULL, "hlac_lt_0_lowtown_docks_n_65535_5y.dds"},
	{0xA3BE1C7555AEA1FAULL, "vxw_sundermount_34p_0_0.vlm"},
	{0xA3BF1CCA2E5800B4ULL, "test_example.ncc"},
	{0xA3BF654911D0D15EULL, "dae100_viscount6118504.cl"},
	{0xA3BF9429334E68BCULL, "gen00fl_varric6033701.cl"},
	{0xA3C01D25B9B094F7ULL, "gen00fl_carver6102697.cl"},
	{0xA3C01DE03A39431BULL, "6061763_m.fxe"},
	{0xA3C0A69302CC8478ULL, "6163123_m.fxe"},
	{0xA3C0A76A67F07391ULL, "lla_ht_0_hightown_n_-131073_0.dds"},
	{0xA3C0A865816AADC2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gurney_n.dds"},
	{0xA3C0CE9C5ECB2460ULL, "dae000_quintus6154729.cl"},
	{0xA3C1882628EA43B1ULL, "hf.cut"},
	{0xA3C1B0F87E6146FBULL, "6205983_m.fxe"},
	{0xA3C1BE006C80D4FBULL, "vxlt_0_lowtown_docks_b9f_0_0.vlm"},
	{0xA3C24C7C4988A46CULL, "vxlt_0_lowtown_eou_0_0.vlm"},
	{0xA3C26C28C151F397ULL, "lt_0_lowtown_docks_a3_23_0.rml"},
	{0xA3C3141A0B97AAABULL, "gen00fl_varric6033689.cl"},
	{0xA3C314CB9B08F73EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closed.ani"},
	{0xA3C33E9412C1164AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\dodge_object.evt"},
	{0xA3C352B0629DD89DULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.cess.evt"},
	{0xA3C3535BF482CB3EULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.dura.ani"},
	{0xA3C359CC31327D9FULL, "gen00fl_merrill6119651.cl"},
	{0xA3C37CA99168CA10ULL, "vxdr_primevaltomb_6px_0_2.vlm"},
	{0xA3C384CECE4373AFULL, "dae000_male_templar_mob6174370.cl"},
	{0xA3C3D4D7A8549B31ULL, "vxht_chantry_219_0_0.vlm"},
	{0xA3C42339812F82EEULL, "6072152_f.fxe"},
	{0xA3C4562D0D970EA0ULL, "hlac_dr_0_deeproads_5843_5y.dds"},
	{0xA3C49EAF6194D13DULL, "vxlt_gallowscourtyard_n_6qs_0_0.vlm"},
	{0xA3C5363A6BD1B196ULL, "vxht_0_hightown_n_anc_0_0.vlm"},
	{0xA3C6F2064C9AA954ULL, "6074410_m.fxe"},
	{0xA3C7E9C25EC5844DULL, "vxdr_primevaltomb_6ix_0_0.vlm"},
	{0xA3C819D9BE6B6F2CULL, "124391004.ogg"},
	{0xA3C8A2756674B752ULL, "vxlt_gallowstemplar_f_4y0_0_0.vlm"},
	{0xA3C8C28306F3AB7DULL, "da2terrain.mat"},
	{0xA3C8DE36D3FD7C20ULL, "281271049.ogg"},
	{0xA3C8E25B47A4543BULL, "follower_barks6154166.cl"},
	{0xA3C8EE31E7C02F20ULL, "docks_male_human_worker_sta6175953.cl"},
	{0xA3C9D6EEC49C1859ULL, "hlac_lt_0_lowtown_65534_1z.dds"},
	{0xA3C9E2FF2D1A43E7ULL, "425620574.ogg"},
	{0xA3C9E63C131DD606ULL, "vxdr_primevaltomb_85q_0_0.vlm"},
	{0xA3CA0783BF494CA3ULL, "678014207.ogg"},
	{0xA3CA1D9412C6FCBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\dodge_object.gad"},
	{0xA3CB62506EF6332CULL, "mrl210_merrill6166782.cl"},
	{0xA3CC2D2DC473211EULL, "vxlt_undercityserialkiller_1ie_0_1.vlm"},
	{0xA3CC6413197F5548ULL, "dae000_male_commoner2_mob6174428.cl"},
	{0xA3CC8E5247A4BA7AULL, "vxw_sundermount_33q_0_0.vlm"},
	{0xA3CD1D999BC15B62ULL, "zz_vault_debug6203345.cl"},
	{0xA3CD1F6C0E4037EBULL, "vxht_bartrand_n_5xg_0_0.vlm"},
	{0xA3CD61331B8D5362ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.impa.gad"},
	{0xA3CD776A26E06A96ULL, "character_character_p16.psh.fxo"},
	{0xA3CE5FCF312A7484ULL, "vxlt_undercityserialkiller_1t9_0_0.vlm"},
	{0xA3CEB3D12D790205ULL, "vxw_encounter3_59_0_0.vlm"},
	{0xA3CED7D6DEE0CB5AULL, "hlac_w_blightlands_65538_1y.dds"},
	{0xA3CF295113B5C4BCULL, "108755507.ogg"},
	{0xA3CF2EFB4AFFF1AFULL, "182100757.ogg"},
	{0xA3CF6DB1BC239017ULL, "521375695.ogg"},
	{0xA3CFFD26D1B63C63ULL, "w_bonepitdestroyed_a3_15_0.mao"},
	{0xA3D00126D1B642B0ULL, "w_bonepitdestroyed_a3_15_0.mmh"},
	{0xA3D00F26D1B65ACAULL, "w_bonepitdestroyed_a3_15_0.msh"},
	{0xA3D02ED02DAFEB38ULL, "vxlt_gallowscourtyard_7i0_0_0.vlm"},
	{0xA3D06B5BF48D7E0FULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.dura.evt"},
	{0xA3D07CB5074608CEULL, "170110479.ogg"},
	{0xA3D0C1311B6AAD28ULL, "667390331.ogg"},
	{0xA3D23C510707F0A7ULL, "vxlt_0_lowtown_docks_a3_cup_0_0.vlm"},
	{0xA3D2AC2556F8D078ULL, "hlac_ht_keep_4519_17x.dds"},
	{0xA3D32E1BEB7D14F1ULL, "dae000_bodhan_letter6144299.cl"},
	{0xA3D34673DF180DF6ULL, "6126966_m.fxe"},
	{0xA3D3731B7FDC15ABULL, "6152274_m.fxe"},
	{0xA3D384331B91FA59ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.impa.evt"},
	{0xA3D3E91DDDCC8392ULL, "vxw_encounter1_cb_0_0.vlm"},
	{0xA3D42640984E2506ULL, "20833642.ogg"},
	{0xA3D427EFDBB3777DULL, "art\\vfx\\materials\\vfx_shockwave_blend.mao"},
	{0xA3D464CEDAA2CB60ULL, "541488444.ogg"},
	{0xA3D48F11A3200CAAULL, "318315386.ogg"},
	{0xA3D4B2ED573CE25FULL, "6083442_m.fxe"},
	{0xA3D4BDD7553BF978ULL, "hlac_lt_0_lowtown_a2_n_65534_3z.dds"},
	{0xA3D4E19F577EF85CULL, "mag320_bsc_banter6100771.cl"},
	{0xA3D5460CDB8530D4ULL, "78259054.ogg"},
	{0xA3D558F2E33C92B7ULL, "vrc221_hawke6063643.cl"},
	{0xA3D59502BB287B0FULL, "atl_worldmap_dxt5_00.dds"},
	{0xA3D6413D897D7836ULL, "gen00fl_dog6146970.cl"},
	{0xA3D6C26411C024AFULL, "vxht_bartrand_n_64l_0_0.vlm"},
	{0xA3D72DCB9B19CAB8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.closed.gad"},
	{0xA3D7365D96B7C4BEULL, "art\\vfx\\combat\\vfx_c_ogr_prone_p.anb"},
	{0xA3D76C5BF4939EB8ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.dura.gad"},
	{0xA3D7F4E1E2C55698ULL, "ef_elf_servant.mor"},
	{0xA3D83654048C42A4ULL, "hlac_ht_0_hightown_n_-131071_1y.dds"},
	{0xA3D8AB2509C5F575ULL, "dae000_gamlen6026174.cl"},
	{0xA3D8CA5DEA44E7E3ULL, "dae000_merchant_dm_mob6174300.cl"},
	{0xA3D8E58FE09C2EE2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_rdy_crnf.gad"},
	{0xA3D8E99C35A497E4ULL, "qun110_petrice6041279.cl"},
	{0xA3D96F6766C3E52BULL, "vxht_0_hightown_n_cbd_0_0.vlm"},
	{0xA3DA50B70726BEC6ULL, "632314285.ogg"},
	{0xA3DA6C4AC7FE89E5ULL, "gen00fl_varric6133763.cl"},
	{0xA3DA7AD719275D04ULL, "vxlt_blackemporium_9d_0_0.vlm"},
	{0xA3DB110ADDC3B216ULL, "703965800.ogg"},
	{0xA3DB2E50F59FD3BBULL, "darktown_fem_fer_urch_sta6174790.cl"},
	{0xA3DBC45BDEF0EF4EULL, "qun211cr_commoner3.utc"},
	{0xA3DBE354BD5B4BF5ULL, "6072555_f.fxe"},
	{0xA3DBED6E57141523ULL, "hlac_lt_gallowscourtyard_-131071_3z.dds"},
	{0xA3DC0A71D073BE02ULL, "37622952.ogg"},
	{0xA3DC8ADC91E266ACULL, "hlac_lt_0_lowtown_n_-131072_4y.dds"},
	{0xA3DCA8D77B1F5E07ULL, "pro000_opening_pc6001190.cl"},
	{0xA3DD453604174C75ULL, "ts_kirk_hightown_03.gda"},
	{0xA3DDE52BED21CF68ULL, "vxw_bonepitmining_a2_1wq_0_1.vlm"},
	{0xA3DE919C050F7273ULL, "vxht_mansion2b_3da_0_0.vlm"},
	{0xA3DEECEAAB36BAEAULL, "gen00fl_aveline6115455.cl"},
	{0xA3DFB37F20689E12ULL, "vxca_caverns1_mines_e98_0_0.vlm"},
	{0xA3E00949937497ABULL, "hlac_ht_keep_3741_0y.dds"},
	{0xA3E0D45E28CD6186ULL, "71108771.ogg"},
	{0xA3E0F857D8A90272ULL, "6181756_m.fxe"},
	{0xA3E13F6149B21ECBULL, "character_character_p174.psh.fxo.dx11"},
	{0xA3E146BDDE8CF15BULL, "6087345_m.fxe"},
	{0xA3E149AC0370D476ULL, "6138760_f.fxe"},
	{0xA3E14C5C37E21B9BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube.mmh"},
	{0xA3E155CC4B954A47ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_004a_0s.dds"},
	{0xA3E1585C37E23038ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube.mao"},
	{0xA3E16A5C37E24E85ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube.msh"},
	{0xA3E18C331B9E4548ULL, "art\\vfx\\combat\\vfx_d_single_att_proj_p.impa.ani"},
	{0xA3E25AFACFC29C12ULL, "vxw_encounter1_hc_0_0.vlm"},
	{0xA3E28857EA81CA9AULL, "follower_barks6153443.cl"},
	{0xA3E2A81A577ECB57ULL, "lowtown_fem_alienage1_sta6162598.cl"},
	{0xA3E2B700354C50EFULL, "vxht_keep_6lt_0_0.vlm"},
	{0xA3E2CBCC4C048712ULL, "609391749.ogg"},
	{0xA3E306B403701FA3ULL, "lgt160_wk_avern_bandits6112751.cl"},
	{0xA3E3174DBF29C759ULL, "482433092.ogg"},
	{0xA3E338AC2CEE98C7ULL, "art\\characters\\creatures\\spider\\animation\\atk_v3.gad"},
	{0xA3E36945C3C6EF06ULL, "w_bonepit_22_0_lc.mmh"},
	{0xA3E3BE270465C54DULL, "vxw_bonepit_1wn_0_1.vlm"},
	{0xA3E3E0FCA6992482ULL, "hlac_dr_primevalentrance_1724_5y.dds"},
	{0xA3E41CD5D0D5ECE2ULL, "lt_0_lowtown_docks_a3_n_107_0.msh"},
	{0xA3E41ED5D0D5EF98ULL, "lt_0_lowtown_docks_a3_n_107_0.mmh"},
	{0xA3E42AD5D0D6043BULL, "lt_0_lowtown_docks_a3_n_107_0.mao"},
	{0xA3E448DC295BAC32ULL, "6081319_m.fxe"},
	{0xA3E453530CAE79BFULL, "6171066_m.fxe"},
	{0xA3E491C6C1BEDF49ULL, "vxca_caverns1_mines_imo_0_0.vlm"},
	{0xA3E4BB3812BC5921ULL, "lt_0_lowtown_docks_a3_18_0.anb"},
	{0xA3E529221D87517EULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comf_0d.dds"},
	{0xA3E540FD47090798ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_4y.dds"},
	{0xA3E5E4411933ECCCULL, "837374938.ogg"},
	{0xA3E6657001A27EB8ULL, "hlac_ht_0_hightown_n_65536_0y.dds"},
	{0xA3E7660BC6829A9FULL, "6083950_m.fxe"},
	{0xA3E81FE8348DEDE5ULL, "follower_banter6171603.cl"},
	{0xA3E8253283E0CB16ULL, "gen00fl_anders6125600.cl"},
	{0xA3E86F80E400699CULL, "follower_banter6165454.cl"},
	{0xA3E88F176B85EBACULL, "vrc221_bartrand6043284.cl"},
	{0xA3E896D1BC30E470ULL, "ch1emmultnt_vl.psh.fxo"},
	{0xA3E8C12A9DFBD173ULL, "vxca_caverns2_6hf_0_1.vlm"},
	{0xA3E8F942D6BB6701ULL, "tbe100im_star_ring.uti"},
	{0xA3E8FB8611338E91ULL, "vxlt_0_lowtown_docks_a3_n_cy9_0_0.vlm"},
	{0xA3E941EEF1CBBBE8ULL, "framebuffereffects_horizontalblurshadows_p0.vsh.fxo"},
	{0xA3E98FF065262A52ULL, "139308107.ogg"},
	{0xA3E99164A609E326ULL, "6031653_m.fxe"},
	{0xA3E99C660EB8E26DULL, "65704571.ogg"},
	{0xA3E9BB41285EB8FAULL, "174198127.ogg"},
	{0xA3E9DD0E18AFF746ULL, "vxw_encounter2_x_0_0.vlm"},
	{0xA3E9EE967216365EULL, "vxlt_0_lowtown_n_ghw_0_0.vlm"},
	{0xA3EA0F7E1DED515EULL, "vxlt_hangedman_108_0_0.vlm"},
	{0xA3EA1DF9A5102A76ULL, "543105480.ogg"},
	{0xA3EA2DAC2CF4A4D4ULL, "art\\characters\\creatures\\spider\\animation\\atk_v3.evt"},
	{0xA3EA7CD0FE59E75DULL, "abi_r_global.bnk"},
	{0xA3EAF940997C6426ULL, "art\\characters\\creatures\\desiredemon\\animation\\crnf.ani"},
	{0xA3EB0E978DA83134ULL, "6149365_m.fxe"},
	{0xA3EC24058D8C17ABULL, "6083424_m.fxe"},
	{0xA3EC39EA65395F4CULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpb.msh"},
	{0xA3EC3FEA6539693EULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpb.mmh"},
	{0xA3ECAD202D21CE13ULL, "864965490.ogg"},
	{0xA3ECC42AF718585FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1light_1.mao"},
	{0xA3ECD79A6CB84460ULL, "153529177.ogg"},
	{0xA3ECDA3A4E8A3F8CULL, "396324751.ogg"},
	{0xA3ED04204DFCC6BFULL, "vxlt_0_lowtown_eoz_0_0.vlm"},
	{0xA3ED108FE0AD20C8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_rdy_crnf.ani"},
	{0xA3ED379381ED3252ULL, "vxlt_0_lowtown_docks_a3_n_d4z_0_0.vlm"},
	{0xA3ED6D9B9C80C8B6ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_7y.dds"},
	{0xA3ED6DA5529BB128ULL, "6201474_m.fxe"},
	{0xA3ED79ACC2B3A141ULL, "follower_banter6141598.cl"},
	{0xA3EDC4973BB78AA2ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_0z.dds"},
	{0xA3EE25D5D0DE5023ULL, "lt_0_lowtown_docks_a3_n_107_0.phy"},
	{0xA3EE3FCEBF28D0CFULL, "82840605.ogg"},
	{0xA3EE5DC34A015862ULL, "16276486.ogg"},
	{0xA3EEA94BC24ECB18ULL, "one100dg_kelder6055615.cl"},
	{0xA3EEDF507B3B5E96ULL, "6114054_f.fxe"},
	{0xA3EEE9FF214C147DULL, "11639195.ogg"},
	{0xA3EF467380F6846DULL, "llac_lt_0_lowtown_a2_n_-196609_1.dds"},
	{0xA3EF98A69FCF754FULL, "6163858_m.fxe"},
	{0xA3F037C3B51F881BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_enter.ani"},
	{0xA3F04DFEBC6B24A2ULL, "605128884.ogg"},
	{0xA3F05BAA93901675ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_boiler_sa.dds"},
	{0xA3F0B45C9A580F4DULL, "458107101.ogg"},
	{0xA3F128DD2742696FULL, "vxht_keep_6iw_0_0.vlm"},
	{0xA3F138E6E114244AULL, "vxdr_0_deeproads_f51_0_0.vlm"},
	{0xA3F1B2B296584681ULL, "241442750.ogg"},
	{0xA3F1C8E7ADBE1108ULL, "vrc321_varric6086931.cl"},
	{0xA3F1E0A5244F0EC6ULL, "vxlt_hangedman_n_1xs_0_0.vlm"},
	{0xA3F2BECCE9FED2A0ULL, "vxlt_warehouse_14n_0_0.vlm"},
	{0xA3F2C52FFEA4923FULL, "6174366_m.fxe"},
	{0xA3F2F9EC3055C1B4ULL, "922321412.ogg"},
	{0xA3F3050C38845962ULL, "1068541283.ogg"},
	{0xA3F34FEAA7A9C881ULL, "mcr310_argument6074356.cl"},
	{0xA3F396ADE98C8A1CULL, "zz_followers_banter6158904.cl"},
	{0xA3F3FCC031A9491CULL, "6178880_m.fxe"},
	{0xA3F495408D96F679ULL, "follower_banter6172307.cl"},
	{0xA3F5E74CA85E71F6ULL, "6174617_m.fxe"},
	{0xA3F5F3D270DF63B9ULL, "hlac_lt_0_lowtown_n_65537_3x.dds"},
	{0xA3F6069ACEF91CD4ULL, "956935043.ogg"},
	{0xA3F64005CB6B0BBDULL, "zz_qun_debug6028262.cl"},
	{0xA3F64F8E952508A6ULL, "lgt360_mag_orlanna6155388.cl"},
	{0xA3F665D9CB91D5C7ULL, "vxht_brothel_n_6g1_0_0.vlm"},
	{0xA3F66A4A6424A86FULL, "vxht_0_hightown_n_bfy_0_0.vlm"},
	{0xA3F6B76DFD2A4EAFULL, "6091720_m.fxe"},
	{0xA3F725AC2CFF210DULL, "art\\characters\\creatures\\spider\\animation\\atk_v3.ani"},
	{0xA3F730E41F0D80B8ULL, "vxw_bonepitdestroyed_a3_1p7_0_0.vlm"},
	{0xA3F73DBE1E52584DULL, "vxlt_undercityserialkiller_wz_0_0.vlm"},
	{0xA3F75DE56BC75306ULL, "vxw_bonepitdestroyed_a3_1py_0_0.vlm"},
	{0xA3F785F64C1949C5ULL, "vxw_sundermount_33z_0_0.vlm"},
	{0xA3F7E1409986C597ULL, "art\\characters\\creatures\\desiredemon\\animation\\crnf.evt"},
	{0xA3F921F23138776EULL, "hlac_lt_0_lowtown_a2_n_c_-65537_2y.dds"},
	{0xA3F92A2B19763B58ULL, "vxlt_0_lowtown_docks_a3_bsx_0_0.vlm"},
	{0xA3FAADDDCC4D8053ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_enter.evt"},
	{0xA3FB098B5DFD31A5ULL, "gen00fl_isabela6101814.cl"},
	{0xA3FB225823474E6BULL, "6141821_m.fxe"},
	{0xA3FB75FD3FF2DF3FULL, "vxlt_0_lowtown_eje_0_0.vlm"},
	{0xA3FC0041601EA691ULL, "mag291_ntbanter6116207.cl"},
	{0xA3FC605A0615316BULL, "6042056_m.fxe"},
	{0xA3FC87FFC2BD6135ULL, "6167482_f.fxe"},
	{0xA3FC8BCD8A83C01EULL, "hlac_ht_keep_4519_16y.dds"},
	{0xA3FD3FC3B52A2012ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_enter.evt"},
	{0xA3FD595D982A01B7ULL, "6206098_m.fxe"},
	{0xA3FE141CEE61C7E9ULL, "lt_gallowscourtyard_25_0_lc.mmh"},
	{0xA3FE5BBEAC52FAF8ULL, "353757406.ogg"},
	{0xA3FEAB8BF639F3BCULL, "802179881.ogg"},
	{0xA3FEB4EA153BFDDAULL, "wm_ico_kirkwall_roll.dds"},
	{0xA3FEDA1B4BE481FCULL, "vxht_playershouse_n_1vx_0_0.vlm"},
	{0xA3FEE240998CE620ULL, "art\\characters\\creatures\\desiredemon\\animation\\crnf.gad"},
	{0xA3FF56E72028BDF3ULL, "6048307_f.fxe"},
	{0xA3FF78CEF9A4CABAULL, "6093859_m.fxe"},
	{0xA3FFB4F79F5BCCCEULL, "12463183.ogg"},
	{0xA3FFC11125E98E9CULL, "dus_clo_drak.mao"},
	{0xA400BE0C6E394FF7ULL, "dae200_saemus6091693.cl"},
	{0xA4010C9C5560F2A3ULL, "1066483818.ogg"},
	{0xA4017580A5AB1049ULL, "vxca_caverns1_mines_igm_0_0.vlm"},
	{0xA401A6DDCC539294ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_enter.gad"},
	{0xA401FDF6D2B1FB95ULL, "lowtown_male_urchin_sta6164236.cl"},
	{0xA4023AFC7A8651FDULL, "lt_0_lowtown_a2_26_0_lc.mmh"},
	{0xA402E2840A2C29FEULL, "follower_barks6153559.cl"},
	{0xA4030BBD9963ABDCULL, "hlac_ht_chantry_n_2534_38y.dds"},
	{0xA403479CF1C6663DULL, "art\\characters\\creatures\\hurlock\\textures\\w_lsw_hlk_0s.dds"},
	{0xA4045EC3B53073A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_hit_chest_enter.gad"},
	{0xA4048F4BCD9F7A68ULL, "w_bonepitdestroyed_a3_206_0.mmh"},
	{0xA4049430E02635B2ULL, "lowtown_male_cityguard_mob6150509.cl"},
	{0xA4049B4BCD9F8E0BULL, "w_bonepitdestroyed_a3_206_0.mao"},
	{0xA404A48CAF0A50C7ULL, "211774485.ogg"},
	{0xA404AD4BCD9FAC92ULL, "w_bonepitdestroyed_a3_206_0.msh"},
	{0xA404DC27561AC0EFULL, "vxht_0_hightown_n_bax_0_0.vlm"},
	{0xA4058F9F47D5A204ULL, "hlac_w_encounter2_camp_262148_0y.dds"},
	{0xA405A2C111039938ULL, "vxw_bonepitdestroyed_a3_1u4_0_0.vlm"},
	{0xA405A7605CDCC09DULL, "vxlt_hangedman_1tc_0_0.vlm"},
	{0xA405B55D568B3C05ULL, "gen00fl_carver6103942.cl"},
	{0xA405CFC25DBD6B86ULL, "vxw_bonepitdestroyed_a3_1uz_0_0.vlm"},
	{0xA406090831099810ULL, "901939003.ogg"},
	{0xA406311E6810B89EULL, "gen00fl_merrill6119582.cl"},
	{0xA4063B2659337B78ULL, "vxht_keep_n_6g9_0_0.vlm"},
	{0xA4064FCB49CBE964ULL, "218560974.ogg"},
	{0xA4068913F7DB5E43ULL, "385633548.ogg"},
	{0xA406A777A290AC28ULL, "74564811.ogg"},
	{0xA406B43730616F99ULL, "zz_one_debug6079307.cl"},
	{0xA406DC4C698B0085ULL, "vxlt_0_lowtown_a2_ex9_0_0.vlm"},
	{0xA4071F7D0E1FAEB2ULL, "78135086.ogg"},
	{0xA40744E3D5E8E1F6ULL, "gen_im_npc_arm_hum_hlm_tmpd.uti"},
	{0xA407504C14CC9E8EULL, "119130867.ogg"},
	{0xA4075725DFAC6A9BULL, "vxht_mansion2b_1bm_0_0.vlm"},
	{0xA4086EF7B0ADFE77ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof4extrude.mmh"},
	{0xA40872F7B0AE0484ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof4extrude.mao"},
	{0xA40884F7B0AE23E9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roof4extrude.msh"},
	{0xA4089264C1E923DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_diepose1.gad"},
	{0xA408A2802C4CB25EULL, "vxlt_0_lowtown_docks_n_b5m_0_1.vlm"},
	{0xA408B5DDCC59CB72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_enter.ani"},
	{0xA408FF3A7F9F94D0ULL, "llac_dr_primevaltomb_664_4.dds"},
	{0xA409FE00E75FBECBULL, "hlac_lt_gallowstemplar_n_2383_12y.dds"},
	{0xA40A0126D1E7BB73ULL, "w_bonepitdestroyed_a3_15_0.rml"},
	{0xA40A5EF44C0DF8F5ULL, "zz_fen_debug6165667.cl"},
	{0xA40A7F60124961A5ULL, "hla_ht_0_hightown_a2_-65535_0x.dds"},
	{0xA40AA5B066E1BC64ULL, "w_blightlands_249_0.msh"},
	{0xA40AB7B066E1DACDULL, "w_blightlands_249_0.mao"},
	{0xA40AB96B5D114CBAULL, "198867906.ogg"},
	{0xA40ABBB066E1E106ULL, "w_blightlands_249_0.mmh"},
	{0xA40AD42B5A9AD7A9ULL, "6071485_m.fxe"},
	{0xA40B5FED6EAC4258ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lute_n.dds"},
	{0xA40C1AACFF517381ULL, "vrc221cr_fake_commander.utc"},
	{0xA40C27AD1DF6E228ULL, "6140630_m.fxe"},
	{0xA40C3BB77C88FE2BULL, "929175327.ogg"},
	{0xA40C7F2A407D827FULL, "351882307.ogg"},
	{0xA40CD5E5F947BDD8ULL, "126075535.ogg"},
	{0xA40CDC969EC71A95ULL, "vxca_caverns1_mines_gk7_0_0.vlm"},
	{0xA40CDCAD549B5D28ULL, "statecnt_puzzle.gda"},
	{0xA40CDF9F9E9B0C85ULL, "vxlt_house_hm_0_0.vlm"},
	{0xA40D032ECE8E73C1ULL, "207336141.ogg"},
	{0xA40D63F20A4F7001ULL, "follower_banter6170280.cl"},
	{0xA40D69DB9B2ADB33ULL, "6103373_m.fxe"},
	{0xA40EC45F083B3FC6ULL, "vxlt_hangedman_n_1ru_0_0.vlm"},
	{0xA40EC6E058C4331CULL, "vxlt_0_lowtown_a2_n_c_hll_0_0.vlm"},
	{0xA40ED8D9A9DCC0C2ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\poundground.gad"},
	{0xA40F1A5DD693434EULL, "220731674.ogg"},
	{0xA41013AB18EA6BA7ULL, "820062719.ogg"},
	{0xA4105ACE48932E56ULL, "dae010_lowlife_m1.utc"},
	{0xA410666E4AFE438EULL, "309225227.ogg"},
	{0xA410C499E2E32938ULL, "seb321_leliana6181779.cl"},
	{0xA410DE9F9CA90785ULL, "hlac_lt_gallowstemplar_2383_6y.dds"},
	{0xA4122B0A6F1650CCULL, "6120148_m.fxe"},
	{0xA4123DC7870F77DFULL, "vxlt_hangedman_n_249_0_0.vlm"},
	{0xA4132AE6973833B2ULL, "art\\characters\\creatures\\dragon\\animation\\tal_spt.gad"},
	{0xA41359AE89FECBAEULL, "vxw_encounter3_6o_0_0.vlm"},
	{0xA413B15A9F5F8CB7ULL, "6092465_f.fxe"},
	{0xA413E65CBC9ED600ULL, "6020525_m.fxe"},
	{0xA4144DD1A0C5C3A0ULL, "638221440.ogg"},
	{0xA4148A25C9F935A2ULL, "trap_detection.ncs"},
	{0xA414AD034B2993F8ULL, "vxht_keep_n_6b6_0_0.vlm"},
	{0xA414BE6FDDD50E17ULL, "vxlt_0_lowtown_a2_n_c_eia_0_0.vlm"},
	{0xA414F795B3F13B43ULL, "vrc321st_spooky_scene_6.stg"},
	{0xA4150EFFA754053AULL, "vxdr_0_deeproads_eh9_0_0.vlm"},
	{0xA4150FC2F6546AEDULL, "hlac_w_woundedcoast_-65535_0y.dds"},
	{0xA4153DD8480A97E1ULL, "vxlt_0_lowtown_a3_n_c_i1f_0_0.vlm"},
	{0xA41556779B344EFFULL, "6122123_m.fxe"},
	{0xA415AAE8BA916506ULL, "vxca_cavern1_haunted_i2u_0_0.vlm"},
	{0xA415E89A4BBBA7CDULL, "6175994_m.fxe"},
	{0xA415FBD9A9E31AB9ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\poundground.evt"},
	{0xA4163B57A8645EC0ULL, "hlac_lt_hangedman_-1_17y.dds"},
	{0xA416C11ED6FC2F11ULL, "6049455_m.fxe"},
	{0xA4179EDA49562034ULL, "one300cr_nuncio_guard.utc"},
	{0xA4188B92CC7DAF2BULL, "311420206.ogg"},
	{0xA418EB54B1B94ABDULL, "6090848_m.fxe"},
	{0xA4199B99E9D406E8ULL, "character_character_p69.psh.fxo"},
	{0xA41A2DE6973E57E9ULL, "art\\characters\\creatures\\dragon\\animation\\tal_spt.evt"},
	{0xA41A3B6F08685791ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_maso_0s.dds"},
	{0xA41A7C84B812CDEEULL, "6086202_m.fxe"},
	{0xA41ABA227E4AC912ULL, "vxht_mansion_n_5ct_0_0.vlm"},
	{0xA41ABB2F45EA6BD8ULL, "255738107.ogg"},
	{0xA41AFEA090898D2AULL, "117663190.ogg"},
	{0xA41B417A777FE50CULL, "dae000_oldmangossip6174915.cl"},
	{0xA41B4946515679D6ULL, "hlac_ht_0_hightown_n_0_1y.dds"},
	{0xA41BDBB88D20F3B3ULL, "6104273_m.fxe"},
	{0xA41BDFC909B9BE9AULL, "6107105_m.fxe"},
	{0xA41C5F298169B6F0ULL, "vxlt_0_lowtown_n_eqd_0_0.vlm"},
	{0xA41C67F96779B566ULL, "441953361.ogg"},
	{0xA41C8DB9B2ED973FULL, "6185159_m.fxe"},
	{0xA41CAD64C1F9FAC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_diepose1.ani"},
	{0xA41CB5856FF28550ULL, "mag320cr_mage_conspirator.utc"},
	{0xA41D363BFA315846ULL, "vxlt_hangedman_n_1ov_0_0.vlm"},
	{0xA41D5E148019799BULL, "vxlt_0_lowtown_a2_n_gyy_0_0.vlm"},
	{0xA41DF721B3478A49ULL, "277592023.ogg"},
	{0xA41E137EC1A0969CULL, "120556120.ogg"},
	{0xA41E1C3331649571ULL, "772918708.ogg"},
	{0xA41E20F876B7FBB1ULL, "vxdr_0_deeproads_cp0_0_0.vlm"},
	{0xA41E3B8E76C44B1BULL, "hlac_lt_foundry_n_320_15x.dds"},
	{0xA41E49790D9B6999ULL, "1042838746.ogg"},
	{0xA41E4F557F3DFCEFULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_sister_victim_sis.gad"},
	{0xA41E593A89974149ULL, "vxca_caverns1_mines_iys_0_0.vlm"},
	{0xA41E701C96CB382BULL, "vxw_bonepitdestroyed_a3_1g9_0_0.vlm"},
	{0xA41E9A24B4D71662ULL, "fen210_tevinter_ambusher6108320.cl"},
	{0xA41E9C3812ED9CB1ULL, "lt_0_lowtown_docks_a3_18_0.rml"},
	{0xA41EE24491E63913ULL, "vxht_mansion2_1a8_0_0.vlm"},
	{0xA41FA0FD7D8DF300ULL, "6070506_m.fxe"},
	{0xA420DB6DA97419F1ULL, "lt_0_lowtown_docks_n_215_0.mmh"},
	{0xA420E76DA9742D92ULL, "lt_0_lowtown_docks_n_215_0.mao"},
	{0xA420F56DA97445AFULL, "lt_0_lowtown_docks_n_215_0.msh"},
	{0xA421116D81B1BEDAULL, "llac_dr_0_deeproads_5849_11.dds"},
	{0xA4211DD20513CA1DULL, "vxca_caverns1_mines_flw_0_0.vlm"},
	{0xA42120C9001CDB32ULL, "hlac_lt_0_lowtown_docks_a3_n_0_0x.dds"},
	{0xA42148307DEBCA30ULL, "6137648_f.fxe"},
	{0xA421BB70A1741F47ULL, "vxht_brothel_n_6v2_0_0.vlm"},
	{0xA421FE0D67D484A0ULL, "6083351_m.fxe"},
	{0xA42205D10C664F32ULL, "ht_mansion2_818.anb"},
	{0xA42210BEE7267A14ULL, "195578887.ogg"},
	{0xA422B85DFA8DD939ULL, "1052370498.ogg"},
	{0xA422C9B609AB99EDULL, "425083102.ogg"},
	{0xA42303D9A9EDB2A8ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\poundground.ani"},
	{0xA42398EFD3C862FEULL, "vxdr_primevaltomb_8og_0_0.vlm"},
	{0xA42417E095A2C3D9ULL, "plt_lgt000pt_lite_key_03.plo"},
	{0xA4242C5966AD364AULL, "331205347.ogg"},
	{0xA4243ADFC3989B9BULL, "vxht_mansion2b_1xo_0_0.vlm"},
	{0xA42447D432573A98ULL, "llac_lt_0_lowtown_-2_0.dds"},
	{0xA42459A7B47CC3A4ULL, "dae000_orsino6085523.cl"},
	{0xA4253CDB15895AF3ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm3.ani"},
	{0xA4253E1049D9B743ULL, "mag160_feynriel6020428.cl"},
	{0xA42581513180D6DBULL, "6161467_m.fxe"},
	{0xA425AF052A9E8063ULL, "plt_cod_mgc_darkspawn.plo"},
	{0xA425FCFB7E8F36B1ULL, "616495578.ogg"},
	{0xA4263CF98B3976F0ULL, "6101556_m.fxe"},
	{0xA426857C638242A3ULL, "nat380_sibling6067991.cl"},
	{0xA426F75975985EFDULL, "hlac_ca_caverns2prc_5802_50z.dds"},
	{0xA42715E69748B958ULL, "art\\characters\\creatures\\dragon\\animation\\tal_spt.ani"},
	{0xA4275D31A3AF633DULL, "6095748_m.fxe"},
	{0xA42792AF5126538FULL, "mag160cr_feynriel.utc"},
	{0xA427EC70133116A6ULL, "lt_0_lowtown_docks_a3_n_18_0_ll.mmh"},
	{0xA4285D5B7228337DULL, "hlac_w_encounter2_131077_0x.dds"},
	{0xA4288630BE4C5F2DULL, "578956535.ogg"},
	{0xA428B571E1F62EBCULL, "6126314_f.fxe"},
	{0xA428C6C8CC56F44FULL, "gen00fl_merrill6107050.cl"},
	{0xA428CD5534F737C8ULL, "6149668_m.fxe"},
	{0xA428DB99DD1CA048ULL, "6181813_m.fxe"},
	{0xA428F872C7FEB8B4ULL, "vxht_keep_61x_0_0.vlm"},
	{0xA4293876B06E3ECCULL, "gen00fl_isabela6201243.cl"},
	{0xA4294D4514DA7444ULL, "w_bonepitdestroyed_a3_22_0.msh"},
	{0xA4295A1815EC4012ULL, "vxlt_hangedman_243_0_0.vlm"},
	{0xA4295F4514DA92ADULL, "w_bonepitdestroyed_a3_22_0.mao"},
	{0xA429634514DA99E6ULL, "w_bonepitdestroyed_a3_22_0.mmh"},
	{0xA42968AF4BE72775ULL, "dae000_gen_player_turn_in6177830.cl"},
	{0xA42AAF74D16DE0A2ULL, "zz_personality_debug6004629.cl"},
	{0xA42ACF56108A6CEEULL, "vxw_encounter2_camp_ql_0_1.vlm"},
	{0xA42AD106735FCF70ULL, "vxlt_0_lowtown_n_eje_0_0.vlm"},
	{0xA42B181AF5A60A10ULL, "hlac_lt_0_lowtown_n_65535_2x.dds"},
	{0xA42B378283DC58CCULL, "lt_0_lowtown_docks_a3_155_0.msh"},
	{0xA42B3D8283DC62BEULL, "lt_0_lowtown_docks_a3_155_0.mmh"},
	{0xA42B498283DC7755ULL, "lt_0_lowtown_docks_a3_155_0.mao"},
	{0xA42B5D4699C9991CULL, "mag210_arianni6030477.cl"},
	{0xA42B7CCC29FF23B5ULL, "6187717_m.fxe"},
	{0xA42B81994AC63CE2ULL, "gen00fl_aveline6132676.cl"},
	{0xA42BD78EF4CBD0CCULL, "hlac_lt_0_lowtown_65536_1y.dds"},
	{0xA42C3A4185BB5B08ULL, "6202332_m.fxe"},
	{0xA42C471063A37243ULL, "gen00fl_anders6055276.cl"},
	{0xA42C72829B03EDC1ULL, "hlac_dr_primevaltomb_665_2z.dds"},
	{0xA42CC33F55674288ULL, "vxw_sundermount_2fm_0_0.vlm"},
	{0xA42CC803CC5B675BULL, "w_bonepitmining_a2_67_0.msh"},
	{0xA42CDA03CC5B85A6ULL, "w_bonepitmining_a2_67_0.mao"},
	{0xA42CE603CC5B9A45ULL, "w_bonepitmining_a2_67_0.mmh"},
	{0xA42CE87372CFCBFEULL, "vxlt_0_lowtown_gc5_0_0.vlm"},
	{0xA42D16EA6570BAD9ULL, "art\\characters\\playercharacter\\humanfemale\\rob_tmpb.phy"},
	{0xA42D28B9DA52E6C4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masd_0s.dds"},
	{0xA42D778E0673AA23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanfemale_ambient_cheerwinning_2_monique.ani"},
	{0xA42D82DACF395F72ULL, "465085990.ogg"},
	{0xA42D920B3D4A3C61ULL, "vxlt_undercityencounter_18x_0_0.vlm"},
	{0xA42DA0F2A74CF78BULL, "646845721.ogg"},
	{0xA42DB407EA00F05FULL, "vxht_bartrand_n_4c3_0_0.vlm"},
	{0xA42DC428C7CAC9E0ULL, "460824365.ogg"},
	{0xA42E09C9617A6FD5ULL, "6055069_m.fxe"},
	{0xA42E77E2BA47DAC3ULL, "631464779.ogg"},
	{0xA42ED9B98035A91CULL, "6135124_m.fxe"},
	{0xA42EE274FB985A76ULL, "6034045_m.fxe"},
	{0xA42F5E6D673D65FBULL, "mcr370_negotiation6080292.cl"},
	{0xA42FFAF7AB98E597ULL, "gen00fl_merrill6106985.cl"},
	{0xA4300E340C93608EULL, "vxdr_0_deeproads_bhc_0_1.vlm"},
	{0xA43055C692B14FB8ULL, "t1_mub_ro3.tnt"},
	{0xA430A000584C8F37ULL, "wm_ico_tavern_roll.dds"},
	{0xA430F857E6E5E2B8ULL, "vxw_bonepitdestroyed_a3_1d3_0_0.vlm"},
	{0xA43110125F8721F5ULL, "6049218_f.fxe"},
	{0xA4320ED9CC02EAFDULL, "6034756_m.fxe"},
	{0xA43256306478BA3EULL, "6119861_m.fxe"},
	{0xA4326C557F4ED6B5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_sister_victim_sis.ani"},
	{0xA432A2494C4847AFULL, "lowtown_male_beggar_sta6156764.cl"},
	{0xA432ACBCB58EB41BULL, "vxht_mansion2b_1sn_0_0.vlm"},
	{0xA432B6D0E1CB2B46ULL, "694502213.ogg"},
	{0xA432B993FB800949ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.ani"},
	{0xA433358ECE062BF5ULL, "llac_lt_warehouse_n_0_7.dds"},
	{0xA433A06EE97BE2EFULL, "6117729_m.fxe"},
	{0xA433D8C1A142BC03ULL, "hlac_w_bonepitdestroyed_a3_327685_0x.dds"},
	{0xA43427B29C5E7EE4ULL, "103507314.ogg"},
	{0xA43493FA514FA0A3ULL, "fen110_party_comment6095069.cl"},
	{0xA434D6DD9A5CD409ULL, "vxlt_backalley_g_1jv_0_0.vlm"},
	{0xA43583205A83DC75ULL, "6148091_m.fxe"},
	{0xA43633939CB4901EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_die_v1_lp.ani"},
	{0xA4370423FF556420ULL, "924825161.ogg"},
	{0xA437A0C064674218ULL, "797350689.ogg"},
	{0xA437EEE9292F64E9ULL, "framebuffereffects_dof_p0.psh.fxo"},
	{0xA438552D3F3A803FULL, "vxht_mansion_n_6p3_0_0.vlm"},
	{0xA438D493FB84A2CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.gad"},
	{0xA43913DB1599BD9DULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm3.gad"},
	{0xA43960AFA3A4DD5EULL, "243196490.ogg"},
	{0xA439B9021BA9CE7CULL, "610396524.ogg"},
	{0xA43A634514E90CADULL, "w_bonepitdestroyed_a3_22_0.rml"},
	{0xA43AA89CBA2BF4CEULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comf_0t.dds"},
	{0xA43AB8ADAD671BBAULL, "vxlt_darktown_52s_0_0.vlm"},
	{0xA43AC831806C0517ULL, "dae000cr_grand_cleric6181909.cl"},
	{0xA43AD90421A48BFEULL, "vxht_mansion_n_4bw_0_1.vlm"},
	{0xA43B04B25AA42CB1ULL, "vxdr_0_deeproads_cj6_0_0.vlm"},
	{0xA43B351C475D5B08ULL, "vxw_sundermount_2kj_0_0.vlm"},
	{0xA43BD3221780E4DDULL, "vxlt_0_lowtown_a2_eou_0_1.vlm"},
	{0xA43C0C6AEAC5E968ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avel_2.mao"},
	{0xA43C7BA653708855ULL, "6052169_m.fxe"},
	{0xA43CE7A0050F9BE3ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.impa.ani"},
	{0xA43CFC8F52A1F9D3ULL, "dae000_meredith6092591.cl"},
	{0xA43D48225614C7B8ULL, "vxdr_primevaltomb_8m7_0_0.vlm"},
	{0xA43DA6EE352401B0ULL, "zz_vault_debug6177778.cl"},
	{0xA43E551FDF224CA3ULL, "llac_ht_keep_n_3743_0.dds"},
	{0xA43E5AE45DF654E5ULL, "vxca_caverns1_mines_gkg_0_0.vlm"},
	{0xA43E8010FE89790EULL, "vxdr_0_deeproads_bcb_0_1.vlm"},
	{0xA43E84E8651B37A3ULL, "6103657_f.fxe"},
	{0xA43EDE29E268E39AULL, "vxht_chantry_209_0_0.vlm"},
	{0xA43F2EE1854BCEC2ULL, "vxlt_0_lowtown_a2_n_c_byg_0_0.vlm"},
	{0xA43F2F5C306BAEAFULL, "follower_banter6136640.cl"},
	{0xA43F6A34D8DBFB38ULL, "vxw_bonepitdestroyed_a3_1i0_0_0.vlm"},
	{0xA43FD193FB8ABC30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_2.evt"},
	{0xA43FE92C6956C068ULL, "vxlt_0_lowtown_docks_a3_n_cqt_0_0.vlm"},
	{0xA4400B997C080860ULL, "vxht_keep_n_67q_0_0.vlm"},
	{0xA440DC425C7A2622ULL, "hlac_w_encounter3_-262148_0y.dds"},
	{0xA4411E99A784CC9BULL, "vxht_mansion2b_1va_0_0.vlm"},
	{0xA441533259572A2EULL, "vxht_mansion_n_637_0_0.vlm"},
	{0xA44237F1F4DD5A79ULL, "6072093_m.fxe"},
	{0xA4428E8E06862D0DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\humanfemale_ambient_cheerwinning_2_monique.gad"},
	{0xA442D308AFC1ADC3ULL, "sib131_slaver_guard6165154.cl"},
	{0xA44321FFD9E5C608ULL, "dae150_chandan6148663.cl"},
	{0xA44348BA8C52EC89ULL, "vxlt_backalley_g_1wu_0_0.vlm"},
	{0xA4442891F10747DEULL, "w_bonepitmining_a2_37_0_lc.mmh"},
	{0xA44441A7FD0766E8ULL, "gen00fl_merrill6150994.cl"},
	{0xA4446403567BEBA1ULL, "one100dg_vanard6054827.cl"},
	{0xA4446A2624B30D9FULL, "6170344_m.fxe"},
	{0xA44483E58157CAA6ULL, "zz_followers_banter6157943.cl"},
	{0xA4453549D6F05734ULL, "zz_followers_banter6158661.cl"},
	{0xA445F3E8C2E2A179ULL, "6129322_m.fxe"},
	{0xA44612E095BFA1EEULL, "plt_lgt000pt_lite_key_03.nss"},
	{0xA4467BA95BD7F235ULL, "vxht_chantry_n_335_0_0.vlm"},
	{0xA446C42E5F5286C8ULL, "pro000_db_ogre_warr.cut"},
	{0xA446EEE8D41BEB62ULL, "lt_0_lowtown_docks_a3_180_0.msh"},
	{0xA446F0E8D41BEE18ULL, "lt_0_lowtown_docks_a3_180_0.mmh"},
	{0xA446FCE8D41C02BBULL, "lt_0_lowtown_docks_a3_180_0.mao"},
	{0xA44793D4230E0CE8ULL, "mag330_walter6044823.cl"},
	{0xA447C1E8560C4667ULL, "655144660.ogg"},
	{0xA447E1FBAD23C853ULL, "6179530_m.fxe"},
	{0xA448B204C135C216ULL, "llac_w_bonepit_0_8.dds"},
	{0xA4490D94839EB506ULL, "lt_0_lowtown_docks_a3_n_45_0.rml"},
	{0xA449171D017E7E51ULL, "hlac_lt_0_lowtown_a2_-196608_4y.dds"},
	{0xA4492A8A9F5D343AULL, "vxlt_darktown_57l_0_0.vlm"},
	{0xA449434D6403F0DFULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredidies.gad"},
	{0xA4495C48EAE1A1E9ULL, "zz_dae_debug6029872.cl"},
	{0xA44987950955614AULL, "dae000pt_talkedto.ncs"},
	{0xA449A6F939537388ULL, "vxw_sundermount_2lk_0_0.vlm"},
	{0xA449E15143B41B7BULL, "428897302.ogg"},
	{0xA449FFA0051A4EDAULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.impa.evt"},
	{0xA44A1938D4345F33ULL, "495266195.ogg"},
	{0xA44A4C939CC56398ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\trooper_die_v1_lp.gad"},
	{0xA44A930E92FF9DD0ULL, "gen00fl_anders6196828.cl"},
	{0xA44C0BA97BB94232ULL, "6163352_m.fxe"},
	{0xA44C945D1B129E8EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thinornate.msh"},
	{0xA44CA25D1B12B6DFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thinornate.mao"},
	{0xA44CAE5D1B12CB7CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thinornate.mmh"},
	{0xA44CD6B7DE5B9D85ULL, "153623320.ogg"},
	{0xA44D18E554579BF8ULL, "vxw_sundermount_343_0_1.vlm"},
	{0xA44D3CD988DB5655ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_crouch_one_knee.anb"},
	{0xA44DC346B2EA10ACULL, "vxw_bonepit_208_0_0.vlm"},
	{0xA44DC99DF300BB39ULL, "fen110_anso6094973.cl"},
	{0xA44DE2FE5C3EAD2EULL, "vxht_keep_n_6ye_0_0.vlm"},
	{0xA44DFADB8A8EB583ULL, "gen00fl_player6207770.cl"},
	{0xA44E279C2FCFBC82ULL, "290347999.ogg"},
	{0xA44ED51C2E648AF9ULL, "gen00fl_sebastian6180308.cl"},
	{0xA44EEEAE0F5A88ECULL, "ch2_ds2.psh.fxo"},
	{0xA45067DFBE954618ULL, "267564510.ogg"},
	{0xA450855B25B24414ULL, "hlac_lt_0_lowtown_a2_131073_2z.dds"},
	{0xA450B17F9EBD4DE1ULL, "857489486.ogg"},
	{0xA450D1B2202C0430ULL, "face_face_p5.psh.fxo"},
	{0xA450E7EACC54C27DULL, "vxdr_slavers_135_0_0.vlm"},
	{0xA450FEA005206BCDULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.impa.gad"},
	{0xA45172C1C3A4D897ULL, "vxlt_0_lowtown_a2_n_c_h25_0_0.vlm"},
	{0xA451D767D93ECC1CULL, "one110_tree_minder6140862.cl"},
	{0xA451FFE3C2ACECA3ULL, "6170771_m.fxe"},
	{0xA452257FDAB3E42DULL, "6133282_m.fxe"},
	{0xA4527BBBF1430796ULL, "zz_followers_banter6157253.cl"},
	{0xA45389F805A807B5ULL, "gen00fl_merrill6113642.cl"},
	{0xA4538AF83EBF366FULL, "one300dg_solivitus6060361.cl"},
	{0xA453BB885B5AC1AFULL, "6164954_m.fxe"},
	{0xA453E2725B4E5131ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampfree.msh"},
	{0xA453F0725B4E694CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampfree.mao"},
	{0xA453FC725B4E7DEFULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lampfree.mmh"},
	{0xA45508D10C91AE6EULL, "ht_mansion2_818.rml"},
	{0xA4552168A7178EF4ULL, "447628825.ogg"},
	{0xA455570D36E2D02AULL, "cd_isa120_hayder.wbd"},
	{0xA4555E8E602318F6ULL, "gen00fl_merrill6128645.cl"},
	{0xA455744FB77CB7DBULL, "78547842.ogg"},
	{0xA455C9DEBC4D44D4ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_c.anb"},
	{0xA455D6D9F494D86CULL, "vxht_mansion_n_6o1_0_0.vlm"},
	{0xA455E601D99434EAULL, "vxlt_blackemporium_zc_0_0.vlm"},
	{0xA4565EF866019450ULL, "dae000_bartrand6031126.cl"},
	{0xA45667DB8D0DA921ULL, "hm_gencr_city_guard_cmndr.mor"},
	{0xA456B601BE3A3C6CULL, "104318399.ogg"},
	{0xA456C4B658E54840ULL, "hlac_lt_gallowstemplar_f_2384_7y.dds"},
	{0xA457001EE6CA77C6ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_2x.dds"},
	{0xA457310CEFB28A44ULL, "character_character_p3.vsh.fxo.dx11"},
	{0xA4578D1CB8A9AA43ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_baserocks.msh"},
	{0xA45793CF606D2E20ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_maj_c.mmh"},
	{0xA4579B1CB8A9C19DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_baserocks.mmh"},
	{0xA4579C6791534CBAULL, "vxlt_darktown_58m_0_0.vlm"},
	{0xA4579F1CB8A9C8AEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_baserocks.mao"},
	{0xA4579FCEDF4B0A5FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appp_0d.dds"},
	{0xA457D69BCB287F92ULL, "one280_emrys6104295.cl"},
	{0xA45818D62B498C08ULL, "vxw_sundermount_2qh_0_0.vlm"},
	{0xA4585EB8F2160E5BULL, "6087132_m.fxe"},
	{0xA45922222752CA0CULL, "vxw_bonepitmining_a2_2ai_0_0.vlm"},
	{0xA45A25A4895C5B56ULL, "vxlt_0_lowtown_a2_n_c_hv4_0_0.vlm"},
	{0xA45A25F9BC40D3BEULL, "hlac_lt_0_lowtown_a2_n_131074_0x.dds"},
	{0xA45AB174CC08026BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair1_n.dds"},
	{0xA45B92A1525D7C56ULL, "llac_lt_gallowscourtyard_131070_0.dds"},
	{0xA45B97D91B2AE09EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_major_v1.ani"},
	{0xA45C614773EEE804ULL, "138396990.ogg"},
	{0xA45C936A45C68216ULL, "hlac_lt_gallowstemplar_n_2394_6y.dds"},
	{0xA45CFD4661198C87ULL, "vxlt_0_lowtown_a2_gpi_0_0.vlm"},
	{0xA45D204D64145E25ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_player_meredidies.ani"},
	{0xA45E9E8B7792C9AFULL, "pla_another.dds"},
	{0xA45EAFAEB6F57F8BULL, "vxht_0_hightown_a3_cag_0_0.vlm"},
	{0xA45FCA01A07EAB54ULL, "dae011_corff_sta6142450.cl"},
	{0xA46000CEB53F1CF3ULL, "vxlt_undercityencounter_zb_0_1.vlm"},
	{0xA4604865F9641E70ULL, "storage_d.dds"},
	{0xA4609379C5B61063ULL, "6161096_m.fxe"},
	{0xA4609DFFBD35A27AULL, "follower_banter6173756.cl"},
	{0xA461265E43C37ADBULL, "art\\vfx\\combat\\material__38.mao"},
	{0xA46153A39B4634D5ULL, "vxw_bonepitdestroyed_a3_263_0_0.vlm"},
	{0xA461AF7E80956022ULL, "gen_im_npc_arm_hum_cht_tmpb.uti"},
	{0xA461CC77BCA55608ULL, "hlac_lt_0_lowtown_n_-65537_0y.dds"},
	{0xA4620D1818282FB9ULL, "oneev_item_acquired.ncs"},
	{0xA4629DC7E383990EULL, "qun120cr_javaris6042979.cl"},
	{0xA462A8DEBC579697ULL, "art\\vfx\\combat\\vfx_c_mer_holy_smt_c.mmh"},
	{0xA462B1EC43DC8221ULL, "6043602_m.fxe"},
	{0xA462DAAAB31DCA26ULL, "isa211_isabela6069422.cl"},
	{0xA4630A256D1B3FE1ULL, "vxlt_0_lowtown_docks_a3_d2d_0_0.vlm"},
	{0xA463178D3AEAC2F6ULL, "6062122_f.fxe"},
	{0xA46332F2C48DABD1ULL, "gen00fl_varric6076231.cl"},
	{0xA463ACBB45770F7DULL, "455557332.ogg"},
	{0xA463FA7B38A829FBULL, "vxht_brothel_n_598_0_0.vlm"},
	{0xA464067D4C3C89B7ULL, "vxlt_shop_eh_0_0.vlm"},
	{0xA4640CBE7A10188DULL, "gen00fl_anders6125502.cl"},
	{0xA4641F9B99DF0390ULL, "6033724_f.fxe"},
	{0xA46495A9D13FACB0ULL, "qcr000_arishok6039301.cl"},
	{0xA464AD12CB8B96A3ULL, "vxlt_undercitydungeon_1ni_0_0.vlm"},
	{0xA46576CF60793A3FULL, "art\\vfx\\combat\\vfx_glb_ice_imp_maj_c.anb"},
	{0xA4660CFBF8DBB575ULL, "mcr370_negotiation6078723.cl"},
	{0xA466238CB1EF0FF7ULL, "gen00fl_anders6114492.cl"},
	{0xA46650865EC7A40CULL, "6148146_m.fxe"},
	{0xA4667B78D4761242ULL, "gen_im_trs_chn_dua_dag_r1.uti"},
	{0xA4668AB31D3FA488ULL, "vxw_sundermount_2zi_0_0.vlm"},
	{0xA466A69F7B7A1465ULL, "230706878.ogg"},
	{0xA46702FADC84562EULL, "hlac_w_blightlands_327683_0z.dds"},
	{0xA46755EB69B930D5ULL, "vxlt_0_lowtown_a2_f51_0_0.vlm"},
	{0xA46780F21EFE92AAULL, "vxw_encounter2_t_0_0.vlm"},
	{0xA467D0C957D16A9EULL, "6099479_m.fxe"},
	{0xA467FFC10B4C40E2ULL, "w_bonepit_127_0.msh"},
	{0xA46801C10B4C4398ULL, "w_bonepit_127_0.mmh"},
	{0xA4680DC10B4C583BULL, "w_bonepit_127_0.mao"},
	{0xA4686A71291214C5ULL, "501763586.ogg"},
	{0xA46897817B5273D6ULL, "vxlt_0_lowtown_a2_n_c_hq5_0_0.vlm"},
	{0xA468A8A198BEACBAULL, "985783763.ogg"},
	{0xA468E658E10BFE19ULL, "gen00fl_isabela_init6077952.cl"},
	{0xA4695045C543E2A7ULL, "hlac_lt_backalley_n_0_10y.dds"},
	{0xA469AFD91B37476FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_major_v1.evt"},
	{0xA469BFF3EDD6A58AULL, "w_bonepitdestroyed_a3_225_0.mao"},
	{0xA469C3F3EDD6ACD9ULL, "w_bonepitdestroyed_a3_225_0.mmh"},
	{0xA469CA70C8F85080ULL, "lgt110_urn_barker6132993.cl"},
	{0xA469CDF3EDD6BDC7ULL, "w_bonepitdestroyed_a3_225_0.msh"},
	{0xA469E3C69208F298ULL, "llac_lt_0_lowtown_docks_a3_0_3.dds"},
	{0xA46A05FFE5EEF2ACULL, "hlac_lt_0_lowtown_docks_a3_-65540_0x.dds"},
	{0xA46B2E08D5555086ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_enter.evt"},
	{0xA46B3098FEE7A3BFULL, "202472024.ogg"},
	{0xA46B5848696F3DDDULL, "one300dg_solivitus6048307.cl"},
	{0xA46B6F23530FA507ULL, "vxlt_0_lowtown_a2_gwh_0_0.vlm"},
	{0xA46BF8BBD39D59F5ULL, "dae000_bartrand6206921.cl"},
	{0xA46C14828413B459ULL, "lt_0_lowtown_docks_a3_155_0.phy"},
	{0xA46C4E1A86A52D63ULL, "6153695_m.fxe"},
	{0xA46C57F989C4EE9AULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody4_.gad"},
	{0xA46C92DB2E6FEBC8ULL, "vxlt_0_lowtown_docks_a3_d2a_0_0.vlm"},
	{0xA46CCCBD4E540BF6ULL, "vxca_caverns1_mines_hu1_0_0.vlm"},
	{0xA46CF21EC9E50ED0ULL, "vxlt_0_lowtown_n_cw0_0_0.vlm"},
	{0xA46D020BCBF93282ULL, "art\\characters\\creatures\\dragon\\textures\\cn_eye_drgb_0d.dds"},
	{0xA46D6F3B3689DB10ULL, "hlac_w_encounter3_-393211_0y.dds"},
	{0xA46E9A951FFCA792ULL, "72124899.ogg"},
	{0xA46FB0D91B3BB418ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_fire_major_v1.gad"},
	{0xA46FD066B8C2A72BULL, "gen00fl_fenris6126503.cl"},
	{0xA46FEC3B982E526FULL, "6077148_m.fxe"},
	{0xA4706F93C98F8BF7ULL, "647701664.ogg"},
	{0xA470986F932766B4ULL, "6142034_m.fxe"},
	{0xA470C054486E4294ULL, "vxlt_0_lowtown_a3_n_c_e1k_0_0.vlm"},
	{0xA470F6D03E4C9230ULL, "6169059_m.fxe"},
	{0xA4712136394FD6AEULL, "vxht_0_hightown_n_b8j_0_0.vlm"},
	{0xA47140AA10EA51CDULL, "6111989_m.fxe"},
	{0xA4716012057A2E1DULL, "vxw_encounter2_camp_7h_0_0.vlm"},
	{0xA471BA6685E1737BULL, "vxht_0_hightown_a2_c70_0_0.vlm"},
	{0xA472386AAD06E23BULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.impa.ani"},
	{0xA4724508D55B96D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_enter.gad"},
	{0xA4724A7A7811F223ULL, "6142378_m.fxe"},
	{0xA4727B8550997D36ULL, "vxht_mansion2_50_0_0.vlm"},
	{0xA4728D5D1B336247ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thinornate.phy"},
	{0xA473B6883D4EF244ULL, "580618543.ogg"},
	{0xA473CD6E2EEE80FAULL, "hlac_lt_0_lowtown_docks_-196610_1y.dds"},
	{0xA473F3D9D2506588ULL, "sib111_mother6098813.cl"},
	{0xA474007F20259157ULL, "hlac_lt_warehouse_-65535_0y.dds"},
	{0xA47457B73338330DULL, "17115165.ogg"},
	{0xA4755D5B780E35B0ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_ric.mmh"},
	{0xA4756B5B780E4DCAULL, "art\\characters\\playercharacter\\humanfemale\\hlh_ric.msh"},
	{0xA475D1E7298453EAULL, "444899048.ogg"},
	{0xA4765C08E92E92D0ULL, "g_wispy.mao"},
	{0xA4766EE233B1DAF3ULL, "6048116_f.fxe"},
	{0xA4769595DEAE3197ULL, "llac_lt_0_lowtown_docks_n_-65538_2.dds"},
	{0xA4770F961DED29B8ULL, "vxdr_primevaltomb_8y3_0_0.vlm"},
	{0xA477EAA6466CB8F7ULL, "189946525.ogg"},
	{0xA4781A86361BA77BULL, "vxw_woundedcoast_kc_0_0.vlm"},
	{0xA4783A2DBFE0776EULL, "6105386_f.fxe"},
	{0xA4784608D560047FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kissing_male_female_action_enter.ani"},
	{0xA479769FD4E3E451ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.anb"},
	{0xA479E1004505BD87ULL, "vxlt_0_lowtown_a2_gjk_0_0.vlm"},
	{0xA47A419DBE65E355ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_012a_0t.dds"},
	{0xA47AF39DF42B31E6ULL, "mag100_cullen6140937.cl"},
	{0xA47AF530C4843B9FULL, "800623693.ogg"},
	{0xA47B1A20183B5AC3ULL, "gen_im_trs_krk_arc_lbw_r0.uti"},
	{0xA47BE7D1555B1FAEULL, "526853775.ogg"},
	{0xA47C59490224465BULL, "hlac_ht_0_hightown_a2_65535_3y.dds"},
	{0xA47C7DC3D8402596ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.dura.ani"},
	{0xA47C82E7FB1E63DBULL, "pro000_opening_pc6142523.cl"},
	{0xA47D13DA0521423EULL, "6076542_m.fxe"},
	{0xA47D1642C49D9730ULL, "vxht_0_hightown_a3_c3b_0_0.vlm"},
	{0xA47D766671A678ADULL, "gen00fl_merrill6142107.cl"},
	{0xA47E3598E43BC935ULL, "664362934.ogg"},
	{0xA47E5587D8476262ULL, "static_static_p1.psh.fxo.dx11"},
	{0xA47EE78E87CC96F2ULL, "537482120.ogg"},
	{0xA47F12342A8DD2EDULL, "vxlt_warehouse_n_ri_0_0.vlm"},
	{0xA47F24DE314067E9ULL, "6028511_m.fxe"},
	{0xA47F406AAD117A32ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.impa.evt"},
	{0xA47FD1EEF770469DULL, "vxw_encounter2_camp_2g_0_0.vlm"},
	{0xA4802EED0ECB3F66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_1.ani"},
	{0xA480421D0388538FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masc_0n.dds"},
	{0xA48064748D99108BULL, "vxlt_gallowstemplar_n_5j5_0_0.vlm"},
	{0xA480651E6DAD9FA0ULL, "6152170_m.fxe"},
	{0xA480A0C52DFDD9FDULL, "vxht_0_hightown_a2_b8n_0_0.vlm"},
	{0xA480B59053EB16E1ULL, "hlac_lt_0_lowtown_docks_n_65536_0y.dds"},
	{0xA48126BB6D922387ULL, "345288106.ogg"},
	{0xA4812AE6179505E8ULL, "23308705.ogg"},
	{0xA4813218C247B428ULL, "48707415.ogg"},
	{0xA48172F989D77880ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody4_.ani"},
	{0xA48180EA5BE91127ULL, "vxw_bonepitmining_a2_1p6_0_0.vlm"},
	{0xA481E94067ABAAAAULL, "6202059_m.fxe"},
	{0xA48205D983E525CFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_exit.gad"},
	{0xA48218F9D1931CEAULL, "1007720274.ogg"},
	{0xA4822A775FE9BDF5ULL, "gen00fl_fenris6126475.cl"},
	{0xA48247BD7A267359ULL, "917715825.ogg"},
	{0xA482647ECDAB3A6FULL, "vxht_mansion_n_60c_0_0.vlm"},
	{0xA482A9FE19E37FBFULL, "art\\vfx\\textures\\potion\\vfx_potion_n.dds"},
	{0xA4838106C7743F8DULL, "dae000_mother6156401.cl"},
	{0xA483B617C44B7C54ULL, "lgt110_urn_guard6170687.cl"},
	{0xA483D5D6681721E0ULL, "857923040.ogg"},
	{0xA484D32F313140ADULL, "art\\characters\\playercharacter\\qunarimale\\hed_nud.mao"},
	{0xA485499D8DE92C2EULL, "vxlt_0_lowtown_a2_eoc_0_0.vlm"},
	{0xA485649D32A10416ULL, "285662968.ogg"},
	{0xA485F8DC130F9286ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bush_n.dds"},
	{0xA48618A158B65915ULL, "vxw_bonepitmining_a2_eh_0_0.vlm"},
	{0xA48634621447B789ULL, "mcr310_argument6182479.cl"},
	{0xA4865F6AAD17CDC5ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.impa.gad"},
	{0xA486FDCE06FA4062ULL, "llac_dr_primevaltomb_9621_4.dds"},
	{0xA48732B115DEB989ULL, "6135297_m.fxe"},
	{0xA487594B53B95E29ULL, "vxlt_undercityserialkiller_1ke_0_0.vlm"},
	{0xA4876F35C4CDA66BULL, "215872389.ogg"},
	{0xA487979FD4F059CAULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.mmh"},
	{0xA487AA7224170F2EULL, "vxht_keep_n_6ey_0_0.vlm"},
	{0xA4888F12BC2152B4ULL, "279655280.ogg"},
	{0xA4891AD983EB687CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_exit.evt"},
	{0xA4894273A989939AULL, "hlac_lt_gallowstemplar_n_2393_1y.dds"},
	{0xA4897C1A50CD3EB7ULL, "face_face_p2.vsh.fxo.dx11"},
	{0xA489AD0C2E8476E1ULL, "6117631_m.fxe"},
	{0xA489F19EA4F1B1DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_neck_crack_both_sides.ani"},
	{0xA48A379627443600ULL, "167009670.ogg"},
	{0xA48A3C8CD7083F2BULL, "zz_mer_debug6017187.cl"},
	{0xA48A85C3D84C7087ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.dura.evt"},
	{0xA48A9D05F8607305ULL, "997765839.ogg"},
	{0xA48AF9CA26A5B26AULL, "657111009.ogg"},
	{0xA48B260493AE3E23ULL, "zz_ave_debug6175036.cl"},
	{0xA48B5C596FC7AB9EULL, "463524860.ogg"},
	{0xA48B881FB693AFB0ULL, "vxht_0_hightown_a3_c4c_0_0.vlm"},
	{0xA48BD5152EAB302FULL, "788708948.ogg"},
	{0xA48C30439E4858AAULL, "6173032_m.fxe"},
	{0xA48C58D02D13C547ULL, "m_dae121ar_chantry.bnk"},
	{0xA48CCA399A2A05D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_other.anb"},
	{0xA48CD2D412B5C6D9ULL, "fake_im_arm_boo_hvy_champ.uti"},
	{0xA48D1A617D40BF7FULL, "genev_item_acquired.ncs"},
	{0xA48D6341605D0B22ULL, "zz_ip_at_to_encounter.utp"},
	{0xA48D7882D5559090ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzardb_n.dds"},
	{0xA48D84111C83EB6DULL, "vxlt_warehouse_n_yh_0_0.vlm"},
	{0xA48D9CE668D7F7B7ULL, "108424219.ogg"},
	{0xA48DAF83E9AE0502ULL, "vxht_mansion2_25r_0_0.vlm"},
	{0xA48E16ED0ED753D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_1.evt"},
	{0xA48E4A12D2DC5297ULL, "hlac_ht_0_hightown_a3_-131071_1z.dds"},
	{0xA48E5C06F186C8DEULL, "189571698.ogg"},
	{0xA48E9E2069CDA47BULL, "vxht_0_hightown_a2_c92_0_0.vlm"},
	{0xA48E9FED76A8777BULL, "hlac_lt_0_lowtown_docks_a3_n_131071_3y.dds"},
	{0xA48EB5FF799649DFULL, "6000960_m.fxe"},
	{0xA48EE083148BE0D5ULL, "gen00fl_varric6078967.cl"},
	{0xA48F51856F5CF3D4ULL, "6048234_f.fxe"},
	{0xA48FCA88FFEDE45CULL, "6115496_m.fxe"},
	{0xA490BD7DBFB21DA4ULL, "6198198_m.fxe"},
	{0xA49180603E229A1DULL, "lgt110_gxa_vgk_conspirator46117116.cl"},
	{0xA491A6C3D852C7D0ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.dura.gad"},
	{0xA491AB9AF87E27BAULL, "dae000ip_crowd_em3_talk4.utp"},
	{0xA491FB9F746C2EA3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext_overlay.ani"},
	{0xA492AA8FC8A27B71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_crouching.anb"},
	{0xA492C09203472F1AULL, "cod_ltr_sk3_circlelette.utp"},
	{0xA492C126B7A28EDAULL, "vxlt_gallowstemplar_f_476_0_0.vlm"},
	{0xA4931179701BCE2EULL, "6016323_f.fxe"},
	{0xA4931FB6758A0D4AULL, "443176102.ogg"},
	{0xA4935395CF1FAB73ULL, "one300_nuncio6182738.cl"},
	{0xA49386DA0245FFA7ULL, "175766533.ogg"},
	{0xA493B9FDE3354338ULL, "vxht_chantry_n_35j_0_0.vlm"},
	{0xA493E38D379AA27BULL, "w_sundermount_206_0.mmh"},
	{0xA493EF8D379AB618ULL, "w_sundermount_206_0.mao"},
	{0xA493FACAF6F1F1DCULL, "ser221_gamlen6123415.cl"},
	{0xA494018D379AD565ULL, "w_sundermount_206_0.msh"},
	{0xA4940AF03CFB3B8CULL, "vxht_brothel_n_3uu_0_0.vlm"},
	{0xA49438DE29DB0563ULL, "art\\vfx\\combat\\vfx_glb_confused_c.anb"},
	{0xA4943D49F1ABE002ULL, "vxlt_darktown_4t9_0_1.vlm"},
	{0xA49517ED0EDD7460ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_1.gad"},
	{0xA4951DF68FFBAD00ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_pmta_0t.dds"},
	{0xA49553E08D9C26A6ULL, "gen00fl_anders6125459.cl"},
	{0xA4957B15C694F6A4ULL, "210126859.ogg"},
	{0xA4958AF59C76D156ULL, "hlac_w_bonepit_0_4z.dds"},
	{0xA495D9AAA8904C89ULL, "vxw_bonepitmining_a2_27j_0_0.vlm"},
	{0xA49622D983F5FF95ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_exit.ani"},
	{0xA49674CF70191D38ULL, "580769131.ogg"},
	{0xA49679272D8EE111ULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_c.anb"},
	{0xA497427905CFC295ULL, "art\\vfx\\combat\\vfx_c_pride_s_p.anb"},
	{0xA497F99EA4FDFCD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_neck_crack_both_sides.evt"},
	{0xA497FA75261DD63CULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_en.ani"},
	{0xA49809B82C123B85ULL, "vxlt_gallowstemplar_39g_0_0.vlm"},
	{0xA498C0F891DFFDA2ULL, "637718653.ogg"},
	{0xA498C688C687B509ULL, "vxw_bonepitmining_a2_1v2_0_0.vlm"},
	{0xA498D1AA86F5FABAULL, "rdr110dg_dockworker6015920.cl"},
	{0xA4996EEEEA875077ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs105a_0t.dds"},
	{0xA4997A81DFE50DDCULL, "6137074_m.fxe"},
	{0xA499E299683B6398ULL, "ht_0_hightownexplosion_n_37_0_ll.mmh"},
	{0xA49A9A19B8AB804BULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.impa.evt"},
	{0xA49B1F401439E70AULL, "gen_im_arm_boo_med_lore_01.uti"},
	{0xA49B7C60950191A0ULL, "hlac_lt_0_lowtown_-65535_4x.dds"},
	{0xA49BF0A5696655E8ULL, "6208548_f.fxe"},
	{0xA49C2934ADC5E761ULL, "dae000_dalish_apprentices6147743.cl"},
	{0xA49C350D371E9940ULL, "cd_isa120_hayder.bnk"},
	{0xA49C5C3F9CA6C88AULL, "zz_isa_debug6133238.cl"},
	{0xA49C8B5872C5AC42ULL, "art\\vfx\\textures\\vfx_blood02.dds"},
	{0xA49C9E2C0B13C095ULL, "853669998.ogg"},
	{0xA49CA587394FBEC4ULL, "mag210_arianni6024115.cl"},
	{0xA49DA4F2CFD8AEB0ULL, "lgt331st_hybris.stg"},
	{0xA49DAD0DCE5CC074ULL, "6174079_m.fxe"},
	{0xA49E2A98EEE3687DULL, "lt_0_lowtown_docks_n_120_0.phy"},
	{0xA49E2BD2703EF32BULL, "356558919.ogg"},
	{0xA49E7D6C67AF8650ULL, "hlac_lt_0_lowtown_a2_-65539_0z.dds"},
	{0xA49EF5A3AA3153D0ULL, "mrl250cr_thief_b.utc"},
	{0xA49F077526240B7EULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_en.gad"},
	{0xA49F189EA5045065ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_neck_crack_both_sides.gad"},
	{0xA49F3C7920651306ULL, "gen00fl_anders6096490.cl"},
	{0xA4A0139F7478949AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext_overlay.evt"},
	{0xA4A06D203F15D0BEULL, "6126517_m.fxe"},
	{0xA4A0FEE822C6B064ULL, "926140262.ogg"},
	{0xA4A109D8B73EDCF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_enter.gad"},
	{0xA4A115DE29E5538CULL, "art\\vfx\\combat\\vfx_glb_confused_c.mmh"},
	{0xA4A1B319B8B1C90CULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.impa.gad"},
	{0xA4A1CDAB2404BE92ULL, "hlac_lt_0_lowtown_n_-131073_2z.dds"},
	{0xA4A2BF0770A5C9B3ULL, "6036998_m.fxe"},
	{0xA4A2CCD1D37CDC3CULL, "8365181.ogg"},
	{0xA4A2EA8FDFA1EAE0ULL, "dae000_bodhan6146844.cl"},
	{0xA4A2FC5B3CA28A15ULL, "vxw_bonepitmining_a2_ov_0_0.vlm"},
	{0xA4A36BD8B05401E3ULL, "548850322.ogg"},
	{0xA4A3701D0BFDF0FBULL, "vxw_woundedcoast_zd_0_0.vlm"},
	{0xA4A39A272D99A38AULL, "art\\vfx\\combat\\vfx_m_sta_fire_range_c.mmh"},
	{0xA4A45C95CC6EB100ULL, "vxht_bartrand_n_5si_0_0.vlm"},
	{0xA4A4B848D3D6052FULL, "dae000ip_crowd_mf4_talk12.utp"},
	{0xA4A4CCC61E0F7A9DULL, "hlac_ht_playershouse_n_855_18y.dds"},
	{0xA4A4D3EA0D8EBAF2ULL, "6097218_f.fxe"},
	{0xA4A4DBEE8B6552AEULL, "6049671_f.fxe"},
	{0xA4A4F2767EC7A9C4ULL, "6116394_m.fxe"},
	{0xA4A525E5D42C10FFULL, "t3_arm_tmp.tnt"},
	{0xA4A54549E6AE20BAULL, "6038562_m.fxe"},
	{0xA4A5E2752629EA9DULL, "art\\characters\\creatures\\mubari\\animation\\tal_ovlm_en.evt"},
	{0xA4A5EF50F5A2B54EULL, "hlac_w_sundermount_196605_0x.dds"},
	{0xA4A60A0914E83B4DULL, "sse_ss_dalish_m_1.gda"},
	{0xA4A63C67C9C22ADDULL, "hlac_lt_hangedman_-1_24z.dds"},
	{0xA4A661FD1CB151BFULL, "zz_qun_debug6028888.cl"},
	{0xA4A68D48E09FF67BULL, "vxw_bonepit_1vj_0_0.vlm"},
	{0xA4A6D94E7C9F2F2BULL, "64825420.ogg"},
	{0xA4A6E54C5269E55FULL, "gen_im_trs_dem_rng_r1.uti"},
	{0xA4A7129F747EB18DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_ext_overlay.gad"},
	{0xA4A73865B87DCD89ULL, "vxw_bonepitmining_a2_1s1_0_0.vlm"},
	{0xA4A7B219B8B6330AULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.impa.ani"},
	{0xA4A80CD8B7450029ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_enter.evt"},
	{0xA4A871E424B1E3FBULL, "gen_im_wep_war_one_act3_10.uti"},
	{0xA4A93DBFA79DFBD3ULL, "611733549.ogg"},
	{0xA4A95DD27FF0B366ULL, "653752842.ogg"},
	{0xA4A9A777D880D7ECULL, "hla_ht_0_hightown_a3_-131071_0y.dds"},
	{0xA4AA4E207B44F8DCULL, "vxht_mansion2_1bv_0_0.vlm"},
	{0xA4AAAD9EE285CF90ULL, "gen00fl_anders6111137.cl"},
	{0xA4AACC688820B9B8ULL, "187245564.ogg"},
	{0xA4AAF00AC9E47B33ULL, "586346641.ogg"},
	{0xA4AAF63DDCC27F9DULL, "hlac_lt_0_lowtown_docks_-65539_3y.dds"},
	{0xA4AB0D6E7A6C25D1ULL, "vxht_keep_5he_0_0.vlm"},
	{0xA4AB270C4413556DULL, "1000700261.ogg"},
	{0xA4AB6736197EFA23ULL, "nat380_sibling6207573.cl"},
	{0xA4ABC1D5997FA44BULL, "967706934.ogg"},
	{0xA4ADC3E78BC48B5FULL, "mer221_sabin6047430.cl"},
	{0xA4ADD6F289280254ULL, "198108207.ogg"},
	{0xA4AE4FBCE1D7CB5DULL, "vxlt_gallowstemplar_4sn_0_0.vlm"},
	{0xA4AE5125818AF006ULL, "vxca_caverns2_1zo_0_0.vlm"},
	{0xA4AF1F98EEF1C8E9ULL, "lt_0_lowtown_docks_n_120_0.mao"},
	{0xA4AF2B98EEF1DC82ULL, "lt_0_lowtown_docks_n_120_0.mmh"},
	{0xA4AF2D98EEF1E038ULL, "lt_0_lowtown_docks_n_120_0.msh"},
	{0xA4AF52522ECECCDEULL, "303787936.ogg"},
	{0xA4AF91534754D05CULL, "vxlt_0_lowtown_docks_a3_n_bu1_0_0.vlm"},
	{0xA4AFE6C3317DC2ECULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_end.ani"},
	{0xA4B03A7BC65A4C53ULL, "311111227.ogg"},
	{0xA4B0484919B76C3EULL, "6172703_m.fxe"},
	{0xA4B05A7CF7F48ABEULL, "580599262.ogg"},
	{0xA4B16E382E98A295ULL, "vxw_bonepitmining_a2_ju_0_0.vlm"},
	{0xA4B19838472DA5B4ULL, "91614450.ogg"},
	{0xA4B1FDEF2D5AF001ULL, "hlac_lt_0_lowtown_docks_65535_4y.dds"},
	{0xA4B1FEA25539FA53ULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.impa.ani"},
	{0xA4B21B59942F54ABULL, "vxht_0_hightown_n_c5u_0_0.vlm"},
	{0xA4B23A6D107EA622ULL, "872499205.ogg"},
	{0xA4B2B5EFBD94AD84ULL, "6075259_m.fxe"},
	{0xA4B2BD648C7C7D89ULL, "vxw_bonepitmining_a2_29h_0_0.vlm"},
	{0xA4B385BE0B81B899ULL, "826134478.ogg"},
	{0xA4B388F947934493ULL, "hlac_lt_undercitydungeon_-1_5y.dds"},
	{0xA4B3CD9B2BB010ABULL, "vxdr_primevaltomb_2xh_0_1.vlm"},
	{0xA4B5645B784425CBULL, "art\\characters\\playercharacter\\humanfemale\\hlh_ric.phy"},
	{0xA4B591EB9274CDA2ULL, "follower_banter6170560.cl"},
	{0xA4B5DCBEC4439455ULL, "vxdr_primevaltomb_922_0_0.vlm"},
	{0xA4B5F3C33182442EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_end.gad"},
	{0xA4B5F4D8B7511498ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_enter.ani"},
	{0xA4B5FBFD9DC6F2DAULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.dura.gad"},
	{0xA4B60A5245775FF0ULL, "6092274_m.fxe"},
	{0xA4B6F79E3DCDFB9AULL, "939252477.ogg"},
	{0xA4B70E71D33F3224ULL, "6094805_m.fxe"},
	{0xA4B71B4329858494ULL, "follower_banter6168912.cl"},
	{0xA4B73D7EAD127927ULL, "vxht_playershouse_n_1uv_0_0.vlm"},
	{0xA4B796EF17E59365ULL, "vxw_bonepitmining_a2_e8_0_0.vlm"},
	{0xA4B7D55EBAF8DFACULL, "vxdr_primevaltomb_6pw_0_1.vlm"},
	{0xA4B84ED02D38EF7DULL, "m_dae121ar_chantry.wbd"},
	{0xA4B883710FE10D08ULL, "143863445.ogg"},
	{0xA4B8C02753C712ADULL, "hlac_lt_warehouse_-65535_1x.dds"},
	{0xA4B8ED7D8297FD30ULL, "zz_approval_debug6003737.cl"},
	{0xA4B9FF4BB8707C40ULL, "679808381.ogg"},
	{0xA4BA3F9F134BA5ADULL, "gen00fl_anders6128372.cl"},
	{0xA4BA498F523F310AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic.mao"},
	{0xA4BA4D8F523F3859ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic.mmh"},
	{0xA4BA578F523F4947ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic.msh"},
	{0xA4BAA3B65100A9D2ULL, "pou_rock.mao"},
	{0xA4BAB256EBFC12F5ULL, "6116603_m.fxe"},
	{0xA4BAB546A164372CULL, "vxlt_backalley_n_1wu_0_0.vlm"},
	{0xA4BAE03330C07ABDULL, "6098591_f.fxe"},
	{0xA4BB43A7E6FCC470ULL, "vxlt_0_lowtown_n_e8k_0_0.vlm"},
	{0xA4BC22A2B078D993ULL, "6211933_m.fxe"},
	{0xA4BCC43759C16748ULL, "6107729_carver_open.cut"},
	{0xA4BCFEFD9DCD16F1ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.dura.evt"},
	{0xA4BE0195D4B31DE4ULL, "vxlt_0_lowtown_a2_n_c_efg_0_0.vlm"},
	{0xA4BE04ACB0E73161ULL, "vxlt_undercityencounter_1jv_0_0.vlm"},
	{0xA4BE3393CF554F50ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm03.mmh"},
	{0xA4BE4193CF55676AULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm03.msh"},
	{0xA4BEF6A25544768AULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.impa.evt"},
	{0xA4BF3A728548AA1BULL, "6036734_m.fxe"},
	{0xA4BF66004C6E6716ULL, "art\\characters\\playercharacter\\elffemale\\ef_arm_lgte.msh"},
	{0xA4BF70004C6E7814ULL, "art\\characters\\playercharacter\\elffemale\\ef_arm_lgte.mmh"},
	{0xA4BF702F7C107A79ULL, "vxlt_gallowsprison_n_6jh_0_0.vlm"},
	{0xA4BF8F0584D6D6FAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainsack_0.pwk"},
	{0xA4C03C0D5AFC4530ULL, "lt_0_lowtown_docks_196_0.msh"},
	{0xA4C0437905F2C9D6ULL, "art\\vfx\\combat\\vfx_c_pride_s_p.mmh"},
	{0xA4C04A0D5AFC5D4AULL, "lt_0_lowtown_docks_196_0.mmh"},
	{0xA4C04E0D5AFC6381ULL, "lt_0_lowtown_docks_196_0.mao"},
	{0xA4C0527A849B0CD0ULL, "hlac_w_encounter3_-262142_0y.dds"},
	{0xA4C08D3686256D2BULL, "vxht_0_hightown_n_c2t_0_0.vlm"},
	{0xA4C0B67A03C1EB10ULL, "gen00fl_isabela6079044.cl"},
	{0xA4C0DAF806E0A06AULL, "lt_0_lowtown_docks_a3_probe_cts.dds"},
	{0xA4C0DEEE640FC01AULL, "vxht_chantry_2e2_0_0.vlm"},
	{0xA4C124D46518EB12ULL, "gen00fl_aveline6139574.cl"},
	{0xA4C1764B44E22749ULL, "vxlt_0_lowtown_a2_n_equ_0_0.vlm"},
	{0xA4C1D33F97CCF0C5ULL, "dre000_stroud6066431.cl"},
	{0xA4C1E4B83900BD32ULL, "855512857.ogg"},
	{0xA4C21AE77589B883ULL, "llac_lt_0_lowtown_docks_n_-2_1.dds"},
	{0xA4C2EE377BDBAD75ULL, "60730476.ogg"},
	{0xA4C32D5CCBF90C88ULL, "946549252.ogg"},
	{0xA4C36437DB620D0AULL, "mcr370_assignments6081707.cl"},
	{0xA4C375469C5453B0ULL, "t3_wdg_tw3.tnt"},
	{0xA4C485C8FFBCD91DULL, "18256584.ogg"},
	{0xA4C56176CF7EF5AFULL, "vxca_cavern1_haunted_icu_0_2.vlm"},
	{0xA4C5D5A2554A5DFDULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.impa.gad"},
	{0xA4C649E9F8265E83ULL, "nat380_nathaniel6066712.cl"},
	{0xA4C652C985F40946ULL, "6041588_m.fxe"},
	{0xA4C6679D7A9DC862ULL, "hlac_w_encounter2_camp_-131071_0x.dds"},
	{0xA4C67D8CFEB48B1FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_2roothanging.msh"},
	{0xA4C6838CFEB495C1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_2roothanging.mmh"},
	{0xA4C68F8CFEB4AA62ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_2roothanging.mao"},
	{0xA4C691E392B13B68ULL, "hlac_ht_mansion2b_818_51z.dds"},
	{0xA4C6E6F68F631065ULL, "433474325.ogg"},
	{0xA4C729515F64DE26ULL, "6041715_m.fxe"},
	{0xA4C731DA5F3129DCULL, "vxht_mansion2_1ht_0_0.vlm"},
	{0xA4C78D62723BDD48ULL, "vxlt_undercityencounter_1js_0_0.vlm"},
	{0xA4C7A00310B32C82ULL, "6092184_m.fxe"},
	{0xA4C7F6A9D4376C32ULL, "dae000_gamlen6149434.cl"},
	{0xA4C80440815A6FAEULL, "mag210_arianni6024526.cl"},
	{0xA4C8999132F6BA88ULL, "hlac_lt_0_lowtown_a2_n_65534_2y.dds"},
	{0xA4C90EE45647437CULL, "65357646.ogg"},
	{0xA4C9607F3E04EFFFULL, "vxca_cave_14j_0_1.vlm"},
	{0xA4C96BF601D5AFE9ULL, "527190627.ogg"},
	{0xA4CA16FD9DD7C9C0ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.dura.ani"},
	{0xA4CA8B0C623602BAULL, "6016315_m.fxe"},
	{0xA4CACEDF1C4DE504ULL, "ave121_jeven6105683.cl"},
	{0xA4CB0BCC7EFDAB9AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilea.msh"},
	{0xA4CB19CC7EFDC3D3ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilea.mao"},
	{0xA4CB1DCC7EFDCAE0ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepilea.mmh"},
	{0xA4CB4E9C9F53D993ULL, "6135189_m.fxe"},
	{0xA4CBA8A62E387432ULL, "6132766_f.fxe"},
	{0xA4CBC4F5091F7719ULL, "6106651_m.fxe"},
	{0xA4CC0677CCDE4254ULL, "follower_banter6160981.cl"},
	{0xA4CC1EE91343D1CBULL, "hlac_lt_0_lowtown_-65536_2x.dds"},
	{0xA4CC445310AE2B6DULL, "zz_spider_stomp_v1.cut"},
	{0xA4CC58A11B53A424ULL, "fxa_cen_aoe_shine2.mao"},
	{0xA4CD2A844331EB73ULL, "gen00fl_anders6125944.cl"},
	{0xA4CD32959B3A95D1ULL, "vxht_0_hightown_n_ccq_0_0.vlm"},
	{0xA4CD38B8B7B92C12ULL, "hlac_lt_0_lowtown_65535_3x.dds"},
	{0xA4CD71C38C063556ULL, "300607640.ogg"},
	{0xA4CE5941995F68B8ULL, "vxca_cavern1_haunted_j2t_0_0.vlm"},
	{0xA4CE5A1770CBDF66ULL, "6088290_f.fxe"},
	{0xA4CE8324FB5A0987ULL, "6127652_m.fxe"},
	{0xA4CE9443933E12D9ULL, "hlac_ca_caverns1_24422_5y.dds"},
	{0xA4CE963B79F75DD4ULL, "vxht_0_hightown_n_9cn_0_0.vlm"},
	{0xA4CECE121724CF15ULL, "plt_prc000pt_generic_01.nss"},
	{0xA4CED830A8616060ULL, "vxlt_gallowstemplar_3x5_0_0.vlm"},
	{0xA4CEE6D088BE82F2ULL, "vxlt_gallowstemplar_1ij_0_0.vlm"},
	{0xA4CF8E3D37A62C6CULL, "vxlt_0_lowtown_a3_n_c_hte_0_0.vlm"},
	{0xA4CFAA533718BDCDULL, "vxht_keep_n_3c9_0_0.vlm"},
	{0xA4CFC04E09E133C0ULL, "gen00fl_sebastian6179941.cl"},
	{0xA4CFDEDA6F006DC9ULL, "100305222.ogg"},
	{0xA4D0B30E4AE2C074ULL, "vxht_0_hightown_bbq_0_0.vlm"},
	{0xA4D18D34C707EFA8ULL, "art\\characters\\creatures\\crowds\\elfmale\\textures\\prp_hightown_jimmy_sa.dds"},
	{0xA4D26F08A669BB8FULL, "vrc221_varric6043600.cl"},
	{0xA4D2A03B109E219DULL, "vxht_mansion_n_57s_0_0.vlm"},
	{0xA4D2D6F564186F6EULL, "vxlt_gallowstemplar_f_5mm_0_0.vlm"},
	{0xA4D34C740A3BA1C7ULL, "vxw_sundermount_2o5_0_0.vlm"},
	{0xA4D4B9189EE510ACULL, "vxdr_primevaltomb_6zu_0_1.vlm"},
	{0xA4D5086FB5919F6CULL, "6061926_m.fxe"},
	{0xA4D509FAF292C44AULL, "6084405_f.fxe"},
	{0xA4D591362F715E04ULL, "art\\vfx\\combat\\vfx_str_att_front_l_arm_c.mmh"},
	{0xA4D59B2CF2B1DD2CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkb.ani"},
	{0xA4D6BB22A9D63F84ULL, "680893125.ogg"},
	{0xA4D6D4E90754D48FULL, "hlac_w_encounter3_-65531_0y.dds"},
	{0xA4D6F485214855F3ULL, "209269267.ogg"},
	{0xA4D74C64DC403800ULL, "965209670.ogg"},
	{0xA4D837B000A087FCULL, "lgt210_eau_senestra6144933.cl"},
	{0xA4D867DC90E6A8BAULL, "339236569.ogg"},
	{0xA4D92AAE6BBBEE36ULL, "vxht_keep_64i_0_0.vlm"},
	{0xA4D98127E57CB0FCULL, "vxlt_warehouse_142_0_0.vlm"},
	{0xA4D9E5E9A2B0DB91ULL, "llac_lt_0_lowtown_a2_n_c_-131072_0.dds"},
	{0xA4D9F376B2012A29ULL, "zz_mag_debug6089448.cl"},
	{0xA4DA1375915A0603ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_hlm_0d.dds"},
	{0xA4DA436F88F35852ULL, "follower_banter6156245.cl"},
	{0xA4DA44479FF4D364ULL, "art\\characters\\playercharacter\\humanmale\\hlf_mede.msh"},
	{0xA4DA56479FF4F1CDULL, "art\\characters\\playercharacter\\humanmale\\hlf_mede.mao"},
	{0xA4DA5A479FF4F806ULL, "art\\characters\\playercharacter\\humanmale\\hlf_mede.mmh"},
	{0xA4DA66DB474A94F7ULL, "6095254_m.fxe"},
	{0xA4DA78A0D35BD402ULL, "gen_im_trs_krk_hvy_cht_r2.uti"},
	{0xA4DA830A2AF9DB26ULL, "hlac_ca_caverns2_5802_66y.dds"},
	{0xA4DA9B2D69083F5FULL, "hlac_ca_cavern1_haunted_24424_22y.dds"},
	{0xA4DACB36C8398F02ULL, "vrc221_varric.cnv"},
	{0xA4DAF09899A4F6A3ULL, "6149029_m.fxe"},
	{0xA4DAF6FE49D91F4BULL, "339853376.ogg"},
	{0xA4DB48CAEDC986AAULL, "gen00fl_sebastian6188697.cl"},
	{0xA4DB706406827DA0ULL, "art\\characters\\weapons\\axes\\w_axe_102a.mao"},
	{0xA4DBDAE4DD18EE96ULL, "470433838.ogg"},
	{0xA4DC0222A6D1E1C2ULL, "hlac_w_sundermount_196602_0y.dds"},
	{0xA4DC401C603312E3ULL, "gen00fl_aveline6137690.cl"},
	{0xA4DC97BF3F9DBD94ULL, "hlac_lt_gallowscourtyard_n_-65536_3x.dds"},
	{0xA4DCA82CF2B8126EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkb.gad"},
	{0xA4DD4EB0D119B2FAULL, "gen00fl_varric6033552.cl"},
	{0xA4DD4F4D04855E7CULL, "vxdr_0_deeproads_egv_0_0.vlm"},
	{0xA4DDFFAF6F21050CULL, "54540143.ogg"},
	{0xA4DE1986227EB5C6ULL, "375795576.ogg"},
	{0xA4DE3A2B8EA0B3B7ULL, "347413050.ogg"},
	{0xA4DEAD36C83D453AULL, "vrc221_varric.bnk"},
	{0xA4DEF59A694335F3ULL, "75727447.ogg"},
	{0xA4DEFB4DDB0C65EBULL, "ht_0_hightown_42_0_lc.mmh"},
	{0xA4DF24EB3CD8D8F4ULL, "vxht_0_hightown_bov_0_0.vlm"},
	{0xA4DF2B7313BEBB2BULL, "vxht_brothel_n_59h_0_0.vlm"},
	{0xA4DF2E40E934A574ULL, "vxca_caverns2_8pr_0_0.vlm"},
	{0xA4DF7AFB6BF834EAULL, "hm_alistair_king.mor"},
	{0xA4DF8C04E165F66EULL, "468948616.ogg"},
	{0xA4DFC581C8BADA0BULL, "dae000_dalish_apprentices6147594.cl"},
	{0xA4E049FDA10AE015ULL, "wpnstand_03.mao"},
	{0xA4E04B2503E0BCDFULL, "llac_lt_0_lowtown_131070_4.dds"},
	{0xA4E0AAE2F2A907EAULL, "mag360_alain6044633.cl"},
	{0xA4E1062F92DFE0B7ULL, "vxht_mansion2_2o9_0_0.vlm"},
	{0xA4E112D3A9B9632BULL, "gen00fl_aveline6122740.cl"},
	{0xA4E11E48F93856BCULL, "follower_banter6141922.cl"},
	{0xA4E140E1E8357646ULL, "6033729_m.fxe"},
	{0xA4E162A64DA60C23ULL, "cen_smk_02.mao"},
	{0xA4E1BE50FC31BA47ULL, "vxw_sundermount_2p4_0_0.vlm"},
	{0xA4E21769C1A64B26ULL, "6033996_m.fxe"},
	{0xA4E22D91CE79758AULL, "vxlt_gallowstemplar_f_5qu_0_0.vlm"},
	{0xA4E23E5669B32604ULL, "6163729_m.fxe"},
	{0xA4E25ED93068FC17ULL, "hlac_w_sundermount_-131068_0z.dds"},
	{0xA4E285DA83F83B61ULL, "hlac_w_bonepitmining_a2_65540_0y.dds"},
	{0xA4E2A4BD84B34E2CULL, "vxca_caverns1_do4_0_0.vlm"},
	{0xA4E2A9F69094AE72ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.cess.gad"},
	{0xA4E2E08668E15C42ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_fireplace.mmh"},
	{0xA4E34ACEB8957CF6ULL, "hlac_lt_gallowscourtyard_-65538_4z.dds"},
	{0xA4E3A32CF2BE284DULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkb.evt"},
	{0xA4E3DAF15EA44253ULL, "qcr220_orsino6041904.cl"},
	{0xA4E44F0948630614ULL, "dae130ip_at_lowtown.utp"},
	{0xA4E451479FFD3DD1ULL, "art\\characters\\playercharacter\\humanmale\\hlf_mede.phy"},
	{0xA4E4521561CC5E26ULL, "mag100_thrask6206228.cl"},
	{0xA4E4D9A11F3C202BULL, "vxw_woundedcoast_n0_0_0.vlm"},
	{0xA4E537F518468BB7ULL, "6182738_m.fxe"},
	{0xA4E60ADD774680ACULL, "vxlt_backalley_n_1xh_0_0.vlm"},
	{0xA4E64DC7E8030EE1ULL, "6101944_m.fxe"},
	{0xA4E71BC0018BA622ULL, "vxca_cavern1_haunted_jgm_0_0.vlm"},
	{0xA4E7872943E579DAULL, "vxlt_gallowscourtyard_6yw_0_0.vlm"},
	{0xA4E79C8B5DB206B6ULL, "vxht_keep_61j_0_0.vlm"},
	{0xA4E7FCCA69E9A52EULL, "hlac_ht_0_hightown_65537_1z.dds"},
	{0xA4E80F0904F041BAULL, "vxlt_darktown_5ng_0_0.vlm"},
	{0xA4E8F1E6420C8252ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_021a_0n.dds"},
	{0xA4E977A24977E1BBULL, "802122352.ogg"},
	{0xA4E9ACF6909AD1A9ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.cess.evt"},
	{0xA4E9E5C8AD0BB2C1ULL, "vxht_keep_n_8at_0_0.vlm"},
	{0xA4EA751865D44C65ULL, "zz_dae_debug6049953.cl"},
	{0xA4EB170614C0B754ULL, "ser221_party_comments6116133.cl"},
	{0xA4EB4C1FC7A9A6C7ULL, "follower_banter6151456.cl"},
	{0xA4EB66DEDF463A87ULL, "6129852_m.fxe"},
	{0xA4EBAF8E44A27FDCULL, "hlac_lt_darktown_-1_33z.dds"},
	{0xA4EBBBEA8C4D9160ULL, "vxlt_gallowstemplar_3r3_0_0.vlm"},
	{0xA4EC60A5BAE5DAD3ULL, "dae000_mother6103074.cl"},
	{0xA4ED815D21E58E95ULL, "6102127_m.fxe"},
	{0xA4EDB1E80770D4B1ULL, "6061156_m.fxe"},
	{0xA4EEB6B03D3F285FULL, "zz_approval_debug6003790.cl"},
	{0xA4EEBD5C66C032CCULL, "6097047_f.fxe"},
	{0xA4EEC6999A6E70FBULL, "vxw_bonepit_1oa_0_0.vlm"},
	{0xA4EF15D24D38E56BULL, "llac_lt_darktown_-1_70.dds"},
	{0xA4EF5B7826FE0626ULL, "follower_banter6155688.cl"},
	{0xA4EF9818CAE80EF6ULL, "hla_ht_0_hightown_65535_0z.dds"},
	{0xA4EFBC5E76295A31ULL, "11848166.ogg"},
	{0xA4F0302DEE27D2C7ULL, "vxw_sundermount_2u7_0_0.vlm"},
	{0xA4F0805A639BE317ULL, "llac_lt_0_lowtown_a2_n_c_65537_1.dds"},
	{0xA4F0ADE71DD9B9ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_lhand.evt"},
	{0xA4F0CD121741B476ULL, "plt_prc000pt_generic_01.plo"},
	{0xA4F0F3307A7DA666ULL, "679318651.ogg"},
	{0xA4F1E3B2F270E562ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_006a_d.dds"},
	{0xA4F206865635FAD2ULL, "6053107_m.fxe"},
	{0xA4F243EFA1B2DA30ULL, "vxht_keep_4hq_0_0.vlm"},
	{0xA4F2DDD2A9C8825DULL, "373360238.ogg"},
	{0xA4F2FBEDC7BB8C44ULL, "rdr111dg_martin6054298.cl"},
	{0xA4F30C4D23547705ULL, "6115772_m.fxe"},
	{0xA4F31038D19B81C1ULL, "vxlt_undercitydungeon_1n3_0_0.vlm"},
	{0xA4F3AFD678A9CA9FULL, "6027775_f.fxe"},
	{0xA4F415DE2F532BDEULL, "vxht_brothel_n_29v_0_0.vlm"},
	{0xA4F4B665455B5E7FULL, "vxlt_shop_s2_0_0.vlm"},
	{0xA4F4FB815E7AB5B7ULL, "w_bonepit_72_0.msh"},
	{0xA4F50D815E7AD41AULL, "w_bonepit_72_0.mao"},
	{0xA4F511815E7ADB29ULL, "w_bonepit_72_0.mmh"},
	{0xA4F5430A07494337ULL, "vxw_bonepitdestroyed_a3_17m_0_0.vlm"},
	{0xA4F5B3F5152D46DAULL, "vxdr_primevalentrance_146_0_1.vlm"},
	{0xA4F6454A0E4C38F4ULL, "vxlt_gallowstemplar_5ax_0_0.vlm"},
	{0xA4F68A6D1C3C39BBULL, "qun211_braggart6075444.cl"},
	{0xA4F6F33D49B709F1ULL, "vxca_caverns1_mines_jg2_0_0.vlm"},
	{0xA4F745036858A458ULL, "6170040_m.fxe"},
	{0xA4F74CC0099F4F6BULL, "hlac_lt_0_lowtown_65536_10z.dds"},
	{0xA4F77EC083938FEFULL, "hf_ser_moira.mor"},
	{0xA4F794F690A6E618ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.cess.ani"},
	{0xA4F7C6E71DE0036CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_lhand.gad"},
	{0xA4F858A17B34E1B9ULL, "hlac_lt_undercitydungeon_-1_4x.dds"},
	{0xA4F95873627E72E8ULL, "313060324.ogg"},
	{0xA4F967835D261B3BULL, "dae011_corff_sta6149627.cl"},
	{0xA4F97D9E89C95D1AULL, "703450083.ogg"},
	{0xA4F99C4929A724DDULL, "6027459_m.fxe"},
	{0xA4F9D085E5A95843ULL, "mag101_grace6034044.cl"},
	{0xA4F9D420DE52101FULL, "473185070.ogg"},
	{0xA4F9EF97A76C1878ULL, "6076491_f.fxe"},
	{0xA4FA08FDFDDC1C27ULL, "896310392.ogg"},
	{0xA4FA882D0C754965ULL, "496349165.ogg"},
	{0xA4FAC94FE41BCA90ULL, "6112656_f.fxe"},
	{0xA4FB193BDDFEFFECULL, "811763835.ogg"},
	{0xA4FB5CF58373CD29ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanr.ani"},
	{0xA4FB5D3B52EE1D2BULL, "vxht_0_hightown_a2_c2c_0_0.vlm"},
	{0xA4FBC556F6C169A0ULL, "vxlt_gallowstemplar_f_4o4_0_0.vlm"},
	{0xA4FBFD7689A1B7B7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngan.mmh"},
	{0xA4FC017689A1BEC4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngan.mao"},
	{0xA4FC137689A1DD29ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphngan.msh"},
	{0xA4FCB204B129A267ULL, "gen_im_trs_gen_junk3_01.uti"},
	{0xA4FCB4760F280340ULL, "pro000_athenril6048762.cl"},
	{0xA4FCC8D02EF6FDA1ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_st.phy"},
	{0xA4FD71705717AEEEULL, "6166580_m.fxe"},
	{0xA4FD94362F92B5FBULL, "art\\vfx\\combat\\vfx_str_att_front_l_arm_c.anb"},
	{0xA4FE3A93CF8B3F6BULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm03.phy"},
	{0xA4FE97E24AF27865ULL, "follower_banter6156006.cl"},
	{0xA4FEB5CEC8A0B51CULL, "6107305_m.fxe"},
	{0xA4FEB6E728FDA3C2ULL, "vxlt_0_lowtown_docks_a3_n_cy6_0_0.vlm"},
	{0xA4FEC5E71DE6206AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_up_lhand.ani"},
	{0xA4FEE7A67041CB5DULL, "6110090_m.fxe"},
	{0xA4FF48879FE5EC46ULL, "vxlt_undercitydungeon_191_0_0.vlm"},
	{0xA4FFEA4C20A27426ULL, "gen_im_flw_sta_bethany_02.uti"},
	{0xA4FFF5EF54B6BE86ULL, "6116598_m.fxe"},
	{0xA5005AD669FC1EF3ULL, "w_encounter3_131_0.phy"},
	{0xA500BDE554A10E72ULL, "vxlt_0_lowtown_a2_n_hgy_0_0.vlm"},
	{0xA50161D8BEFE68ADULL, "ss_coterie_assassin_f.wbd"},
	{0xA5019B30F5C97E1DULL, "6134971_m.fxe"},
	{0xA502582E8522CEF0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_chantryfloor_n.dds"},
	{0xA50277F5837A19ABULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanr.gad"},
	{0xA502EAF728F826EAULL, "572680514.ogg"},
	{0xA50303E8C875694AULL, "vxht_bartrand_n_64e_0_0.vlm"},
	{0xA5030D8CFFB1DFD2ULL, "aip_follower_varric.gda"},
	{0xA5035DD6DD18E8D2ULL, "hlac_ht_0_hightown_n_131072_0z.dds"},
	{0xA5036DCB5BB2FC56ULL, "vxlt_darktown_54m_0_0.vlm"},
	{0xA503A23569AEA3F8ULL, "vxlt_undercityencounter_vh_0_0.vlm"},
	{0xA503B4E6F93F5BB7ULL, "vxw_bonepitdestroyed_a3_18l_0_0.vlm"},
	{0xA504054862F31F21ULL, "vxht_0_hightown_n_du_0_0.vlm"},
	{0xA5042718FC7065E8ULL, "493536019.ogg"},
	{0xA5043589AB33571BULL, "126688440.ogg"},
	{0xA5046243C2552754ULL, "6110824_m.fxe"},
	{0xA5046ECEC6CFFBC7ULL, "dae110ip_at_to_trench_b.utp"},
	{0xA50490AB2E477A8BULL, "vxdr_slavers_11x_0_1.vlm"},
	{0xA504A7B53B3263B7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1b.phy"},
	{0xA504F2C2E8DC72BAULL, "vxlt_darktown_5ta_0_0.vlm"},
	{0xA505C28F0ADAB524ULL, "730822169.ogg"},
	{0xA506788FDAE49F8BULL, "815293124.ogg"},
	{0xA5069BB8E15CADC1ULL, "qun221_petrice6076806.cl"},
	{0xA506AC0ABB0FDCD2ULL, "follower_banter6148743.cl"},
	{0xA506B3231CCF02BFULL, "one101ar_ruin_int.are"},
	{0xA506B3D14B7F3AF8ULL, "qun100_mage_party_comment6019098.cl"},
	{0xA506CE35092298B3ULL, "458098229.ogg"},
	{0xA506F1543C686DDAULL, "w_encounter2_29_0.rml"},
	{0xA50710824C22D228ULL, "dre000_hireling16038598.cl"},
	{0xA507268DE81AD0AAULL, "hlac_lt_0_lowtown_a2_n_-262144_2z.dds"},
	{0xA507F6C8CBFF88A9ULL, "537160025.ogg"},
	{0xA507FBBC701ECA55ULL, "6025800_m.fxe"},
	{0xA508092A3B95A1EBULL, "331550147.ogg"},
	{0xA5084BD33C7E8EDBULL, "hlac_ht_chantry_2534_25y.dds"},
	{0xA50874F5837E8010ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured_leanr.evt"},
	{0xA508B4585DFA1739ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_4y.dds"},
	{0xA508EA8C18BCD821ULL, "599107408.ogg"},
	{0xA50927EBA120C01BULL, "hlac_lt_0_lowtown_a2_n_0_5x.dds"},
	{0xA509F4A093E46D0BULL, "213894453.ogg"},
	{0xA50AAB67FA341555ULL, "vxlt_hangedman_2yg_0_0.vlm"},
	{0xA50AAC5CE1EE19CDULL, "vxlt_0_lowtown_a2_n_c_i17_0_0.vlm"},
	{0xA50AEFC8583282DEULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured.evt"},
	{0xA50B1245661F0699ULL, "6204986_m.fxe"},
	{0xA50B1DB9B53474DCULL, "vxht_bartrand_n_5s5_0_0.vlm"},
	{0xA50BA42326C66759ULL, "82279340.ogg"},
	{0xA50BDF8669045FE9ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_fireplace.anb"},
	{0xA50C00E29FCCB1E3ULL, "lowtown_argument6169432.cl"},
	{0xA50C27550D05409FULL, "zz_isa_debug6133189.cl"},
	{0xA50C27C274DFCCEFULL, "6043514_m.fxe"},
	{0xA50C63EABA59A4CFULL, "vxw_encounter1_c3_0_0.vlm"},
	{0xA50C842E93950E67ULL, "vxht_keep_709_0_0.vlm"},
	{0xA50D360C6D82A531ULL, "599676395.ogg"},
	{0xA50D569EA29D826EULL, "815251619.ogg"},
	{0xA50D8328A45BBF0AULL, "vxlt_gallowstemplar_f_5bp_0_0.vlm"},
	{0xA50DBA33663A2AA0ULL, "vxdr_0_deeproads_crr_0_1.vlm"},
	{0xA50DC6ED4622B050ULL, "follower_banter6143407.cl"},
	{0xA50DD2543C6E57CEULL, "w_encounter2_29_0.phy"},
	{0xA50DEA258D8E9090ULL, "l_env_wal_lowtown_wal_calm_p2p.bnk"},
	{0xA50DEAD3528618DEULL, "dae021cr_sister_lorena.utc"},
	{0xA50E61BD5F8FBF52ULL, "hlac_ca_caverns1_24425_13z.dds"},
	{0xA50E8984C723D322ULL, "gen00fl_bethany6102902.cl"},
	{0xA50EA4804BC98C2FULL, "hlac_lt_0_lowtown_docks_-131072_0x.dds"},
	{0xA50EC5F2410C7F12ULL, "6108777_m.fxe"},
	{0xA50F145D07E2A534ULL, "vxca_caverns1_img_0_0.vlm"},
	{0xA50F33ABCD99633CULL, "662220143.ogg"},
	{0xA50F3A03289840A8ULL, "mer141_jansen6077790.cl"},
	{0xA50F7B888AF7D3D4ULL, "6149916_m.fxe"},
	{0xA50FE2CC664BBC7BULL, "hlac_lt_0_lowtown_a2_n_-131072_2z.dds"},
	{0xA510687FDDF033B2ULL, "dae000cr_grand_cleric6124697.cl"},
	{0xA510CF649C56E285ULL, "hlac_lt_0_lowtown_a2_n_-65537_4z.dds"},
	{0xA510E866ECDE0953ULL, "6126923_m.fxe"},
	{0xA5110511881A17CBULL, "6179978_m.fxe"},
	{0xA511091678D0D5FFULL, "dae000_orana6149038.cl"},
	{0xA511E7E24742DDCDULL, "6072695_m.fxe"},
	{0xA511EFAE1C45B470ULL, "134141831.ogg"},
	{0xA51206C85838C809ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured.gad"},
	{0xA51228517E2D0AFCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_mountainc.mao"},
	{0xA51234517E2D1E9FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_mountainc.mmh"},
	{0xA5123A517E2D2941ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_mountainc.msh"},
	{0xA51247D5944EEA85ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_001a_0s.dds"},
	{0xA512A7630596E868ULL, "736941488.ogg"},
	{0xA513084AE2C3996DULL, "216264374.ogg"},
	{0xA513F3D2094D459CULL, "hlac_lt_0_lowtown_a2_-65536_5x.dds"},
	{0xA514039CD26E329FULL, "vxw_bonepitdestroyed_a3_87_0_0.vlm"},
	{0xA5142BE4ABE5464EULL, "vxlt_shop_j4_0_0.vlm"},
	{0xA514786946D4E5BEULL, "36923995.ogg"},
	{0xA514C19F8611826DULL, "plt_cod_mgc_mana.plo"},
	{0xA514EB1A33438A28ULL, "6121752_m.fxe"},
	{0xA51610D430A7FB4DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_bark01_sa.dds"},
	{0xA516941D77E1160EULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.mmh"},
	{0xA516D3C617C51046ULL, "566154580.ogg"},
	{0xA51757D0FC42DC23ULL, "280981852.ogg"},
	{0xA517AC231CDD6A5CULL, "one101ar_ruin_int.lst"},
	{0xA517AFFCB0750655ULL, "vxca_caverns1_mines_edl_0_0.vlm"},
	{0xA5181D1397B4042EULL, "vxht_keep_n_63s_0_0.vlm"},
	{0xA5182584A3E7F5A6ULL, "hlac_dr_slavers_3215_5z.dds"},
	{0xA51851439E421EF7ULL, "lt_0_lowtown_docks_a3_n_232_0.mmh"},
	{0xA51855439E422504ULL, "lt_0_lowtown_docks_a3_n_232_0.mao"},
	{0xA51867439E424469ULL, "lt_0_lowtown_docks_a3_n_232_0.msh"},
	{0xA51895374BE516D8ULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetold_dp.dds"},
	{0xA518BD50F1092A1EULL, "vxht_mansion2_3ci_0_0.vlm"},
	{0xA518D7C8583E9757ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_injured.ani"},
	{0xA518E922A9CEE45FULL, "gen_im_acc_blt_act3_09.uti"},
	{0xA5194B44881244E8ULL, "104374753.ogg"},
	{0xA51AD5C7AC4FBD4FULL, "vxw_encounter1_h2_0_0.vlm"},
	{0xA51AFFF48CBEBCE6ULL, "fx_glyph_repul.dds"},
	{0xA51B0AE9E03A227FULL, "vxht_mansion2_23e_0_0.vlm"},
	{0xA51B6B9794B6982CULL, "gen_im_flw_arc_varric_02.uti"},
	{0xA51B6F004CBCD50FULL, "art\\characters\\playercharacter\\elffemale\\ef_arm_lgte.phy"},
	{0xA51B83A8E45ABC1AULL, "rdr111dg_martin6054490.cl"},
	{0xA51CC0DCEDD56DA2ULL, "one200_gold_followers6120149.cl"},
	{0xA51CC5B00E1CB567ULL, "456124520.ogg"},
	{0xA51CE29DEF87AC15ULL, "802989461.ogg"},
	{0xA51D23368F715200ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightsand_n.dds"},
	{0xA51D4E8F5293B746ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic.phy"},
	{0xA51D853BF39151FAULL, "6162301_f.fxe"},
	{0xA51DC0E17AED7956ULL, "vxlt_undercitydungeon_yt_0_0.vlm"},
	{0xA51E16AF9FF76894ULL, "vxlt_0_lowtown_a3_n_c_emw_0_0.vlm"},
	{0xA51E54764F57E2DAULL, "art\\characters\\playercharacter\\humanmale\\bdy_masc.mmh"},
	{0xA51E58764F57E911ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masc.mao"},
	{0xA51E61C5E3532BD8ULL, "dae000_gen_turn_in_36178449.cl"},
	{0xA51E66764F580120ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masc.msh"},
	{0xA51E8B05CE7D02FBULL, "vxca_caverns1_ead_0_0.vlm"},
	{0xA51EA114E714822BULL, "vxw_woundedcoast_z4_0_0.vlm"},
	{0xA51EAC231CE388A5ULL, "one101ar_ruin_int.ncs"},
	{0xA51EBBAB81988212ULL, "vxht_mansion_n_59f_0_0.vlm"},
	{0xA51EBD7FECB53894ULL, "hlac_lt_0_lowtown_n_131073_0y.dds"},
	{0xA51FA1B2C94C6BDCULL, "6082200_m.fxe"},
	{0xA51FD18AB660A380ULL, "hlac_ht_0_hightown_65532_0x.dds"},
	{0xA520A21978308072ULL, "22053787.ogg"},
	{0xA520E2108286A229ULL, "pro000ip_corpse_demo.utp"},
	{0xA52134D977B6C69AULL, "465263798.ogg"},
	{0xA521844CDDCB2482ULL, "6154725_m.fxe"},
	{0xA521A0413094F2C7ULL, "zz_ser_debug6078573.cl"},
	{0xA521B33454233E17ULL, "324039859.ogg"},
	{0xA521B436BA176BE1ULL, "vxlt_0_lowtown_docks_a3_n_arr_0_0.vlm"},
	{0xA521B616D4C3994EULL, "22723972.ogg"},
	{0xA523419C307CEFBFULL, "400156090.ogg"},
	{0xA52393D84DC523BAULL, "gen00fl_aveline6122071.cl"},
	{0xA5240C234A186E5EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkl.ani"},
	{0xA5242BDFAE5EE4A5ULL, "mag110_samson6016164.cl"},
	{0xA524B31D77ED880DULL, "art\\vfx\\combat\\vfx_r_brd_ssong_loop_c.anb"},
	{0xA5255B4D9FB8AA58ULL, "gen00fl_aveline6139235.cl"},
	{0xA525A8961C5BBC2FULL, "gen00fl_bethany6109232.cl"},
	{0xA526078DF51E078EULL, "245901571.ogg"},
	{0xA5266A7A97F0F8BDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03deadtree.phy"},
	{0xA52689EE29B4471CULL, "vxlt_hangedman_n_2xb_0_0.vlm"},
	{0xA526B6D645AEE24CULL, "241473647.ogg"},
	{0xA526DAD9D35BE834ULL, "vxlt_warehouse_q4_0_0.vlm"},
	{0xA5272F2DE2FF429EULL, "vxht_mansion2_3fj_0_0.vlm"},
	{0xA52846235F40DBAEULL, "vxlt_0_lowtown_bxm_0_0.vlm"},
	{0xA52847676DB89655ULL, "w_blightlands_129_0.msh"},
	{0xA52849676DB899CBULL, "w_blightlands_129_0.mmh"},
	{0xA52855676DB8AE68ULL, "w_blightlands_129_0.mao"},
	{0xA5285FBF5A435721ULL, "6028803_m.fxe"},
	{0xA52899921A177F63ULL, "vxdr_0_deeproads_fbm_0_0.vlm"},
	{0xA528D5ED5162C2FEULL, "vxht_brothel_n_322_0_0.vlm"},
	{0xA528F6FB6C7B679EULL, "vxlt_gallowscourtyard_n_7hf_0_0.vlm"},
	{0xA52905C7DF4DE268ULL, "atl_optionsmen_dxt5_dat.xml"},
	{0xA52959520F581250ULL, "77679819.ogg"},
	{0xA529E481AC540D48ULL, "37007997.ogg"},
	{0xA52A6471E5245234ULL, "mag100_thrask6032076.cl"},
	{0xA52B079E1AF76881ULL, "art\\vfx\\combat\\vfx_gib_qun_ltrb.ani"},
	{0xA52BFD911758338DULL, "ser221_gascard6122470.cl"},
	{0xA52C485CFE8EE590ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basica.mao"},
	{0xA52C4C5CFE8EECA3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basica.mmh"},
	{0xA52C5A5CFE8F04FDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basica.msh"},
	{0xA52CA34A95152BFFULL, "1018765311.ogg"},
	{0xA52D12F1D90A9AABULL, "vxw_woundedcoast_w7_0_0.vlm"},
	{0xA52D3E44E5031416ULL, "follower_banter6172276.cl"},
	{0xA52D612232152B74ULL, "vxlt_0_lowtown_docks_n_cw4_0_1.vlm"},
	{0xA52D75DFE8D4649FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_basic.phy"},
	{0xA52D775CBB9E59B1ULL, "6036418_f.fxe"},
	{0xA52D87C95CAE7601ULL, "vxht_bartrand_n_5sd_0_0.vlm"},
	{0xA52D9EC9AE622FFEULL, "gencr_harvester_head.utc"},
	{0xA52DAA630DD73CF2ULL, "6169318_m.fxe"},
	{0xA52DBB0F6028302EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv1.ani"},
	{0xA52DFED85A342604ULL, "116163149.ogg"},
	{0xA52E138FBB42845CULL, "6085300_m.fxe"},
	{0xA52E28370907D589ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v2.ani"},
	{0xA52F1633F5E6F5C5ULL, "static1_alpha.psh.fxo"},
	{0xA52F7215D1069F6BULL, "361237137.ogg"},
	{0xA52F8C99C416F8B2ULL, "dae000_keep_servant_1_sta6160884.cl"},
	{0xA52FAE7CA343A8AEULL, "cod_hst_dalish2.utp"},
	{0xA53024FEEC8F3D1FULL, "mag111ip_hh_at_lowtown.utp"},
	{0xA5304BC4A39C971FULL, "vxw_sundermount_37l_0_0.vlm"},
	{0xA530CC7FF9536B47ULL, "6169661_m.fxe"},
	{0xA530FF42A2FF92ACULL, "mer120_hubert6034432.cl"},
	{0xA531129E1AFBE723ULL, "art\\vfx\\combat\\vfx_gib_qun_ltrb.gad"},
	{0xA531378C84A3F72CULL, "123296971.ogg"},
	{0xA5319FC7A1F6A891ULL, "llac_w_sundermount_-262146_0.dds"},
	{0xA531CEADD7D388B6ULL, "dae011_drunk1_sta6144381.cl"},
	{0xA531D05F933B6FACULL, "qcr220_meredith6054725.cl"},
	{0xA532037512F40DE0ULL, "vxw_encounter2_mw_0_0.vlm"},
	{0xA53290B2B2778562ULL, "mag320_orsino6071188.cl"},
	{0xA53306F48CD35A34ULL, "fx_glyph_repul.mao"},
	{0xA533CC258DAF2D5AULL, "l_env_wal_lowtown_wal_calm_p2p.wbd"},
	{0xA5340A22E2A25E51ULL, "vxht_brothel_n_5gh_0_0.vlm"},
	{0xA534A20487A538C3ULL, "hm_gen_coterie_archer.mor"},
	{0xA5352CB42A62FB24ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srdb_0n.dds"},
	{0xA5354337090E220BULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v2.gad"},
	{0xA5354CB6C55200B4ULL, "vxlt_warehouse_z5_0_0.vlm"},
	{0xA5355111D62F4DFEULL, "llac_ht_mansion_n_3366_0.dds"},
	{0xA5355BA96455C088ULL, "163416911.ogg"},
	{0xA535BE2FD9972AC3ULL, "465077910.ogg"},
	{0xA535D73E6E3273A3ULL, "hlac_ca_caverns1_24424_2x.dds"},
	{0xA535EDA8789EBAF2ULL, "vxht_mansion2_2l5_0_0.vlm"},
	{0xA5363BC06D1158E3ULL, "6125693_f.fxe"},
	{0xA5365F3AC8158EFBULL, "265145082.ogg"},
	{0xA536711382220EB2ULL, "prx_cutoff.dds"},
	{0xA536B6A51EE14111ULL, "57309687.ogg"},
	{0xA536C7FAEE4FB6E2ULL, "mcr370cr_mage_escort.utc"},
	{0xA536C92A36A9BC1BULL, "hlac_lt_foundry_n_320_6x.dds"},
	{0xA53724DF12084525ULL, "follower_banter6161109.cl"},
	{0xA537560F83E94C5DULL, "hlac_ca_caverns1_mines_24425_3y.dds"},
	{0xA5375D953B6199C6ULL, "vxca_cavern1_haunted_e8o_0_0.vlm"},
	{0xA5375F7A97FF5929ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03deadtree.mao"},
	{0xA5376B7A97FF6DC2ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03deadtree.mmh"},
	{0xA5376D7A97FF7178ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_03deadtree.msh"},
	{0xA537B41C1D7DFA4EULL, "829345638.ogg"},
	{0xA537CC0AD167B385ULL, "vrc221_varric6040926.cl"},
	{0xA53825234A2941D8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkl.gad"},
	{0xA53830340024959DULL, "6033968_m.fxe"},
	{0xA538A12A2A7719E4ULL, "vxlt_0_lowtown_docks_a3_9bt_0_0.vlm"},
	{0xA538C9552A191351ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_2.ani"},
	{0xA538DDC2F47B784FULL, "llac_lt_hangedman_-1_24.dds"},
	{0xA5393FA2D1E33E21ULL, "6172649_m.fxe"},
	{0xA5395577315B9A17ULL, "6092182_f.fxe"},
	{0xA539738B12B0735BULL, "static_static_p2.psh.fxo.dx11.assembly"},
	{0xA539A3D8E2B64758ULL, "lt_0_lowtown_35_0.rml"},
	{0xA539B1BF06C45F7FULL, "lt_warehouse_5_0_ll.mmh"},
	{0xA539E0DC225E3A66ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.impa.evt"},
	{0xA539E23708010A47ULL, "6141942_m.fxe"},
	{0xA53A05A8CE9F90C1ULL, "dae000_arishok6078419.cl"},
	{0xA53A2BAD1F8D7AB6ULL, "883501492.ogg"},
	{0xA53A4B15D8FC8ECBULL, "vxlt_warehouse_19f_0_0.vlm"},
	{0xA53A82050B47FEA1ULL, "hlac_ht_0_hightown_-131073_0x.dds"},
	{0xA53B320468508660ULL, "mag100_cullen6082350.cl"},
	{0xA53B9A2740EF851DULL, "vxw_encounter2_camp_qz_0_0.vlm"},
	{0xA53BB30F6034607FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv1.evt"},
	{0xA53C287A5883E954ULL, "hlac_lt_backalley_g_0_4x.dds"},
	{0xA53C403709143C70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v2.evt"},
	{0xA53C729799501204ULL, "6169778_m.fxe"},
	{0xA53D2E0C70ECEAE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90.gad"},
	{0xA53D3E2EF62615A0ULL, "vxlt_0_lowtown_docks_a3_n_d4x_0_0.vlm"},
	{0xA53D3F041BF6BF27ULL, "dre000_dougal6129827.cl"},
	{0xA53D4D5CFE9D61C8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basica.phy"},
	{0xA53D6FFC0222F40EULL, "526215858.ogg"},
	{0xA53D8F3F5D134157ULL, "vxht_0_hightown_c3l_0_0.vlm"},
	{0xA53E76DFE8E2D9A4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_basic.mmh"},
	{0xA53E7ADFE8E2E0F7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_basic.mao"},
	{0xA53E8CDFE8E2FF46ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_basic.msh"},
	{0xA53E926D3925FB5BULL, "llac_lt_0_lowtown_docks_a3_-131075_6.dds"},
	{0xA53EA2574E2FB574ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.ces.gad"},
	{0xA53EBDA19592AF9FULL, "vxw_sundermount_32o_0_0.vlm"},
	{0xA53EBF364C64B91DULL, "875207569.ogg"},
	{0xA53ED4552A1D91F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_2.gad"},
	{0xA53FB79A0F67D41BULL, "6107188_m.fxe"},
	{0xA53FC8FB3E29E666ULL, "hlac_ca_caverns1_24423_1y.dds"},
	{0xA540755204EA2660ULL, "vxw_encounter2_ht_0_0.vlm"},
	{0xA540A1A97899229FULL, "hlac_ca_caverns1_mines_24426_1x.dds"},
	{0xA540ECB05A31B439ULL, "vxht_mansion2_3e8_0_0.vlm"},
	{0xA540ED3C3DD4E128ULL, "mag320_orsino6070440.cl"},
	{0xA540F7DC22647FB1ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.impa.gad"},
	{0xA54105359F2583C7ULL, "l_fen121ar_base.wbd"},
	{0xA54134EDA82BB093ULL, "vxlt_darktown_581_0_1.vlm"},
	{0xA5426BD8BF3610D7ULL, "ss_coterie_assassin_f.bnk"},
	{0xA5427BFFD49876D1ULL, "vxht_brothel_n_5jo_0_0.vlm"},
	{0xA542940F603A4A68ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\rangedv1.gad"},
	{0xA542FFB06E8C0C2FULL, "6018716_m.fxe"},
	{0xA543510C70F191D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90.evt"},
	{0xA543E196AAD38751ULL, "6086986_m.fxe"},
	{0xA543F9127B6874CEULL, "640575098.ogg"},
	{0xA5444381FFF2096EULL, "vxlt_gallowstemplar_n_5qi_0_0.vlm"},
	{0xA5445F856A94D372ULL, "vxht_mansion2_2i6_0_0.vlm"},
	{0xA54473764F78E6E5ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masc.phy"},
	{0xA544876B1F50754CULL, "vxht_mansion2_324_0_0.vlm"},
	{0xA5453938DCAFE1FEULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.dura.ani"},
	{0xA5457D4BFE03B063ULL, "vxdr_0_deeproads_fho_0_0.vlm"},
	{0xA545B1574E35ED52ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.ces.ani"},
	{0xA545C77C85BA66D2ULL, "plt_cod_itm_aurvars_prize.plo"},
	{0xA545CD7DD657078BULL, "vxht_0_hightown_a3_c1w_0_0.vlm"},
	{0xA545D1552A23AB58ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_talk_2.evt"},
	{0xA54613D3E344A750ULL, "815354921.ogg"},
	{0xA546793B297A5F03ULL, "vxht_mansion2b_23i_0_0.vlm"},
	{0xA547BE36FAE4D8AFULL, "hlac_w_bonepitdestroyed_a3_65542_0y.dds"},
	{0xA547F8DC226AA05FULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.impa.ani"},
	{0xA548073AC37CC582ULL, "6170227-mrl351-pridedemon-h.cut"},
	{0xA5481EBE066F1751ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree03_dp.dds"},
	{0xA5482F354446BC4EULL, "llac_ca_caverns1_24424_25.dds"},
	{0xA5487BCBAE282843ULL, "610705242.ogg"},
	{0xA548CC46F9FF5157ULL, "6179589_m.fxe"},
	{0xA549091598992631ULL, "273073128.ogg"},
	{0xA549B9FDCF93C01AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic.phy"},
	{0xA549E1D5434BAA01ULL, "374901612.ogg"},
	{0xA549ED1E7FB7F9C4ULL, "6127426_m.fxe"},
	{0xA54A0C0432E59D9DULL, "vxw_encounter2_camp_ly_0_0.vlm"},
	{0xA54A718B64044042ULL, "6040987_m.fxe"},
	{0xA54AD6B22CE6F931ULL, "vxlt_gallowsprison_n_6jp_0_0.vlm"},
	{0xA54B7CB53B6E1D0FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1b.mao"},
	{0xA54B88B53B6E31ACULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1b.mmh"},
	{0xA54B8EB53B6E3B9EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1b.msh"},
	{0xA54BAC712656BB55ULL, "w_sundermount_221_0.phy"},
	{0xA54C011C4F0959D7ULL, "vxht_0_hightown_c6o_0_0.vlm"},
	{0xA54C0AC3BB16C2F4ULL, "20772945.ogg"},
	{0xA54DC3198AAD37C9ULL, "vxlt_house_jc_0_0.vlm"},
	{0xA54DC749468E323EULL, "25247017.ogg"},
	{0xA54DCA62CD26800EULL, "6125326_f.fxe"},
	{0xA54DD1543CA43991ULL, "w_encounter2_29_0.mmh"},
	{0xA54DD8A5373A6C6EULL, "204860356.ogg"},
	{0xA54DDD543CA44E32ULL, "w_encounter2_29_0.mao"},
	{0xA54DEB543CA4664FULL, "w_encounter2_29_0.msh"},
	{0xA54EC764887F1F91ULL, "vxlt_0_lowtown_docks_n_b8h_0_0.vlm"},
	{0xA54F4FA895A395E9ULL, "6119581_m.fxe"},
	{0xA54FC52CB306C4D4ULL, "llac_ht_mansion2b_818_15.dds"},
	{0xA5500265E0182BB8ULL, "6068914_m.fxe"},
	{0xA550ED3CEF480C18ULL, "vxdr_primevaltomb_7sv_0_0.vlm"},
	{0xA551590C70FDDCC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_90.ani"},
	{0xA55198CC2AA7953AULL, "gen00fl_fenris6127103.cl"},
	{0xA5522A0C8886FCDDULL, "635405831.ogg"},
	{0xA5524D24CA38286FULL, "6209914_m.fxe"},
	{0xA552BDED432BD1E4ULL, "620703516.ogg"},
	{0xA552F79659E3A9DDULL, "vxht_0_hightown_c4g_0_0.vlm"},
	{0xA553166DEF7574C2ULL, "6123229_f.fxe"},
	{0xA5535138DCBC47CFULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.dura.evt"},
	{0xA5538EC304B42C06ULL, "rdr111dg_assistant6016644.cl"},
	{0xA554474E1AFCEFE6ULL, "6144402_f.fxe"},
	{0xA555465B49E1C6CDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_channelwater_d.dds"},
	{0xA555B121EEFA5CE7ULL, "pro000cr_desperate_refugee.utc"},
	{0xA555E632E25950A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_en.evt"},
	{0xA55650D73447DBB2ULL, "hm_ave2_smuggler.mor"},
	{0xA556AEEE7EBA3EA5ULL, "vxht_bartrand_n_5wd_0_0.vlm"},
	{0xA5572ECFBCE8BFCBULL, "vxlt_warehouse_13x_0_0.vlm"},
	{0xA55738CA9300B72DULL, "6081373_f.fxe"},
	{0xA55758D7B18DACD9ULL, "6061343_m.fxe"},
	{0xA55759FDFABE3BC1ULL, "6068844_f.fxe"},
	{0xA55793F97AD6CC37ULL, "atl_load_contr_dxt5_dat.xml"},
	{0xA557C9DE51EB32D7ULL, "lgt110_orz_harrowmont6172677.cl"},
	{0xA558335E59372CC9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_enter.ani"},
	{0xA558594C042D6271ULL, "483864574.ogg"},
	{0xA5587DE124DBB61DULL, "vxw_encounter2_camp_gx_0_0.vlm"},
	{0xA558976AF918E85CULL, "t3_eey_bl3.tnt"},
	{0xA558B14E8D1F0352ULL, "477469940.ogg"},
	{0xA558B4E461C58394ULL, "vxdr_primevalentrance_2sj_0_1.vlm"},
	{0xA558F7E7EFB85BB8ULL, "zz_dae_debug6029350.cl"},
	{0xA5596E9C5FD035A6ULL, "hlac_lt_0_lowtown_a2_n_131071_5x.dds"},
	{0xA559D1410E29C9BEULL, "hlac_w_sundermount_131068_0z.dds"},
	{0xA55A5238DCC26878ULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.dura.gad"},
	{0xA55AFCD48D350FD0ULL, "vxht_0_hightown_b8c_0_0.vlm"},
	{0xA55B180EA3C21F65ULL, "6128080_m.fxe"},
	{0xA55B445F451B59BBULL, "hlac_lt_0_lowtown_a2_n_-65536_6y.dds"},
	{0xA55B9FFB3CC20900ULL, "6155607_m.fxe"},
	{0xA55BFF7E8EC9ECD5ULL, "6179990_m.fxe"},
	{0xA55CBF32E25F2D24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_en.gad"},
	{0xA55CE5FE901F03A8ULL, "hlac_ht_playershouse_n_855_25x.dds"},
	{0xA55D0CF557A98A5CULL, "hlac_w_sundermount_327674_0x.dds"},
	{0xA55D39417A753811ULL, "vxlt_0_lowtown_docks_n_b5o_0_0.vlm"},
	{0xA55E38B9A7A1789EULL, "qun111_ketojan6040958.cl"},
	{0xA55E40817D9288A2ULL, "vxlt_backalley_g_1oj_0_0.vlm"},
	{0xA55E4E5E593BC64BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_enter.gad"},
	{0xA55EB3978988A2B7ULL, "vxlt_gallowscourtyard_7bb_0_0.vlm"},
	{0xA55ED1D02F49D716ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_st.mmh"},
	{0xA55ED315E5F52732ULL, "character_character_p241.psh.fxo.dx11"},
	{0xA55EDBD02F49E814ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_st.msh"},
	{0xA55F1FE8EF2F896FULL, "llac_lt_0_lowtown_a3_n_c_-262144_4.dds"},
	{0xA55F569AFFA88EDBULL, "6084537_m.fxe"},
	{0xA55F5A28DE6E9069ULL, "6048179_m.fxe"},
	{0xA55FA471DCBF49F8ULL, "392791878.ogg"},
	{0xA55FA4853BFF8FF7ULL, "dae000_arishok6078354.cl"},
	{0xA5606F2DA92AFB12ULL, "hlac_ht_0_hightown_n_-65537_0y.dds"},
	{0xA560A526F2DA12F9ULL, "hla_ht_0_hightown_a3_65537_0x.dds"},
	{0xA560DFA0E5F4AAA9ULL, "lgt310_mis_burn_rival_thug6152778.cl"},
	{0xA561273BE3DE3A6EULL, "vxlt_gallowstemplar_n_5ok_0_0.vlm"},
	{0xA56182A391A8394EULL, "vxlt_gallowscourtyard_752_0_0.vlm"},
	{0xA561D4E4F279614FULL, "vxlt_hangedman_24b_0_0.vlm"},
	{0xA5621FB23C02D8CEULL, "vfx_alignquadtoshadowdirection_p0.gsh.fxo.dx11"},
	{0xA562452975CC8F21ULL, "vxlt_warehouse_cw_0_0.vlm"},
	{0xA56253D66A4EDD68ULL, "w_encounter3_131_0.mmh"},
	{0xA5625FD66A4EF10BULL, "w_encounter3_131_0.mao"},
	{0xA56265AE667DC80DULL, "lgt110_orz_harrowmont6108999.cl"},
	{0xA56271D66A4F0F92ULL, "w_encounter3_131_0.msh"},
	{0xA56284C28E483111ULL, "vxca_caverns2prc_7o9_0_1.vlm"},
	{0xA5628AAFF639F62CULL, "155480595.ogg"},
	{0xA562B32C1143E346ULL, "vxca_cavern1_haunted_e7r_0_0.vlm"},
	{0xA562EDC27A27886EULL, "296380516.ogg"},
	{0xA563CE32E2656502ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_weep_en.ani"},
	{0xA5644A948CDD7998ULL, "6162281_m.fxe"},
	{0xA56520CB70B05725ULL, "vxht_bartrand_n_5zc_0_0.vlm"},
	{0xA5654B5E5941DFB0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\kissing_male_female_reaction_enter.evt"},
	{0xA5655DAABFB71312ULL, "vxca_caverns2prc_8bk_0_0.vlm"},
	{0xA56584D8E2DB4D10ULL, "lt_0_lowtown_35_0.anb"},
	{0xA565C65398013E38ULL, "967028323.ogg"},
	{0xA565D4BC668ED341ULL, "w_encounter2_62_0.phy"},
	{0xA56602A887F7ACEDULL, "lgt250_way_dalish_conflict6147125.cl"},
	{0xA5662E9FB18E8138ULL, "qun220cr_javaris6071721.cl"},
	{0xA566C2D260FA1DB1ULL, "gen_im_trs_cot_lgt_glv_r1.uti"},
	{0xA567AADF34395D21ULL, "follower_banter6160829.cl"},
	{0xA567BA6C10D32A31ULL, "vxlt_gallowsprison_n_6pv_0_0.vlm"},
	{0xA567F9669E9A074EULL, "vxht_0_hightown_a3_bxh_0_0.vlm"},
	{0xA568E65017D8863AULL, "6103404_m.fxe"},
	{0xA5691331C62F7DBAULL, "w_blightlands_236_0.msh"},
	{0xA5692131C62F95F3ULL, "w_blightlands_236_0.mao"},
	{0xA5692531C62F9C00ULL, "w_blightlands_236_0.mmh"},
	{0xA5693432B981BD42ULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf.gad"},
	{0xA56945B045F1FBC6ULL, "mag121_viveka6082741.cl"},
	{0xA569EA75EF912095ULL, "vxca_caverns1_ii6_0_0.vlm"},
	{0xA569F7917585022BULL, "vxw_bonepit_1o1_0_0.vlm"},
	{0xA56A20955EA1972EULL, "vxca_caverns1_cpd_0_0.vlm"},
	{0xA56A452687442C7CULL, "hlac_lt_0_lowtown_a2_n_c_-2_5z.dds"},
	{0xA56AB602EF31C033ULL, "zz_followers_banter6159630.cl"},
	{0xA56AE80B3A4BD3B7ULL, "mcr370_negotiation6080160.cl"},
	{0xA56BC1E97FD044F8ULL, "vxw_encounter2_e7_0_0.vlm"},
	{0xA56BDDBC66934E5DULL, "w_encounter2_62_0.rml"},
	{0xA56C8468095E55D3ULL, "6034080_m.fxe"},
	{0xA56C86E21C134E95ULL, "qun111_ketojan6039810.cl"},
	{0xA56DC9B33FB5CEA0ULL, "hlac_lt_0_lowtown_0_3z.dds"},
	{0xA56E48F63EA43D14ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookinquisb_sa.dds"},
	{0xA56FAA5623F272D6ULL, "6103336_m.fxe"},
	{0xA56FB29CFE115815ULL, "6179262_m.fxe"},
	{0xA570132CD8DE1165ULL, "vxca_caverns1_ixy_0_0.vlm"},
	{0xA5701EB32E2CD272ULL, "ash_t_f.mao"},
	{0xA57046C1E46F79CFULL, "vxlt_hangedman_21m_0_0.vlm"},
	{0xA5705732B9881739ULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf.evt"},
	{0xA570C29B76AAB237ULL, "21210434.ogg"},
	{0xA570F1BEB8860508ULL, "6137604_f.fxe"},
	{0xA57124DD113C660AULL, "6167613_m.fxe"},
	{0xA5718883B2CB6551ULL, "380879899.ogg"},
	{0xA57188C35016F02CULL, "hlac_ht_mansion2b_818_28z.dds"},
	{0xA5719D761FE2A35AULL, "vxht_keep_6h9_0_0.vlm"},
	{0xA5720B497E96A380ULL, "vxca_cavern1_haunted_gkg_0_1.vlm"},
	{0xA57216C5F9D9EF8DULL, "gen00fl_anders6103906.cl"},
	{0xA5721E9ADE9A4C71ULL, "dae000_sandal6148194.cl"},
	{0xA572B33B675AEAE3ULL, "vxlt_darktown_58a_0_1.vlm"},
	{0xA572C4668397D1C5ULL, "6200887_m.fxe"},
	{0xA5735500224A84F0ULL, "gen00fl_anders6119182.cl"},
	{0xA57402014C402B06ULL, "lgt211ar_gang_hideout_lt.lst"},
	{0xA5743167A59D50B4ULL, "vxca_caverns1_i2n_0_0.vlm"},
	{0xA574732EA4C9DFFFULL, "zz_one_debug6079316.cl"},
	{0xA57494448C611E2FULL, "vxht_0_hightown_a3_avl_0_0.vlm"},
	{0xA57521BC0628EC37ULL, "vxdr_primevaltomb_86i_0_0.vlm"},
	{0xA575CDDEF17AE3C7ULL, "mag170_hugh6080720.cl"},
	{0xA575E35A51E4B95EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklback.ani"},
	{0xA5766B4390901FCEULL, "vxht_0_hightown_a3_bwi_0_0.vlm"},
	{0xA5768B8F1A023AEDULL, "follower_barks6153626.cl"},
	{0xA576F7E7E7A1C54AULL, "ser270_emeric6113231.cl"},
	{0xA5772D179C82DB30ULL, "6101318_m.fxe"},
	{0xA5782EA50C8B39D3ULL, "dre000_varric6035154.cl"},
	{0xA57838C239A556F3ULL, "vxlt_gallowstemplar_n_3rc_0_0.vlm"},
	{0xA57853EB758AFF0AULL, "vrc221cr_hugin.utc"},
	{0xA5794B1710C17376ULL, "733844340.ogg"},
	{0xA579C5ACC29DF013ULL, "835575488.ogg"},
	{0xA57A20214F411F09ULL, "hlac_ht_playershouse_n_855_11z.dds"},
	{0xA57AB334ACBFC335ULL, "6044603_m.fxe"},
	{0xA57B22A66592A76BULL, "vxca_caverns2_8az_0_0.vlm"},
	{0xA57B8F8475C8A39EULL, "6208255_m.fxe"},
	{0xA57BD2439204AF4CULL, "llac_lt_hangedman_-1_16.dds"},
	{0xA57C834981940A16ULL, "vxw_sundermount_30p_0_0.vlm"},
	{0xA57C944B0AD28F15ULL, "112139042.ogg"},
	{0xA57CE8E1EF9B61FAULL, "hm_amb_keep_bureaucrat2.mor"},
	{0xA57D5DFFD494F9ABULL, "vxht_0_hightown_a2_caz_0_0.vlm"},
	{0xA57D5F32B992AF28ULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf.ani"},
	{0xA57E11014C489911ULL, "lgt211ar_gang_hideout_lt.are"},
	{0xA57E8509CAD429E5ULL, "vxca_caverns1_icx_0_0.vlm"},
	{0xA57EFC3F46B4DE91ULL, "vxca_caverns2_968_0_0.vlm"},
	{0xA57EFFE41BB22CE3ULL, "hlac_lt_0_lowtown_n_65537_3z.dds"},
	{0xA580A28645B1379CULL, "676533563.ogg"},
	{0xA580B4A4C41BFCC6ULL, "6131561_f.fxe"},
	{0xA580E583B1ED74BBULL, "6127973_m.fxe"},
	{0xA5810F359F5B78FDULL, "l_fen121ar_base.bnk"},
	{0xA58176932CF636F3ULL, "6120459_m.fxe"},
	{0xA58185D982511721ULL, "vxht_mansion2_33z_0_0.vlm"},
	{0xA581FAFF437477B0ULL, "vxw_encounter1_k8_0_0.vlm"},
	{0xA5821C5186665D88ULL, "vxlt_gallowstemplar_f_5aw_0_0.vlm"},
	{0xA5824695AFC28474ULL, "po_clo.mao"},
	{0xA5827CD7ACB7CB6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_reaction.ani"},
	{0xA582AAF4E84D8537ULL, "vxca_caverns1_hvp_0_0.vlm"},
	{0xA582E7BFB106DB4FULL, "gen00fl_varric6020609.cl"},
	{0xA582FB5A51EF6C2FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklback.evt"},
	{0xA58306217E5736AFULL, "vxht_0_hightown_a3_aso_0_0.vlm"},
	{0xA5835B295A05CD03ULL, "vxlt_0_lowtown_docks_n_bdw_0_1.vlm"},
	{0xA583A0621C2D5D31ULL, "6108730_m.fxe"},
	{0xA5842D4F1AB33E79ULL, "129551781.ogg"},
	{0xA584C409A6D9573DULL, "vxht_0_hightown_a3_b5h_0_0.vlm"},
	{0xA584DD208286384EULL, "vxht_0_hightown_a3_brj_0_0.vlm"},
	{0xA584E96C296DA8DDULL, "573910624.ogg"},
	{0xA58526CE473146C7ULL, "6095406_f.fxe"},
	{0xA585364234B0C48BULL, "225943426.ogg"},
	{0xA5856A96131A2F1DULL, "one110_tree_minder6133737.cl"},
	{0xA5857B49C6A4AABCULL, "6176024_m.fxe"},
	{0xA5862B0D6DA93448ULL, "678122310.ogg"},
	{0xA5865F497CA80B2FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appp_0t.dds"},
	{0xA586A36BE5DFA37AULL, "6069960_m.fxe"},
	{0xA586C7B26B59E704ULL, "ht_0_hightown.anb"},
	{0xA586CA17E19997A9ULL, "117320362.ogg"},
	{0xA586CB6CFEDC2D05ULL, "85655136.ogg"},
	{0xA586E3B26B5A1696ULL, "ht_0_hightown.arl"},
	{0xA58728574692BEF3ULL, "gen00fl_sebastian6180463.cl"},
	{0xA58750CD47C63537ULL, "vxlt_0_lowtown_docks_baz_0_0.vlm"},
	{0xA587D436552E9742ULL, "dae200_bran6094821.cl"},
	{0xA587FF75468E8B99ULL, "llac_ca_caverns2prc_5802_41.dds"},
	{0xA588675F9D913DD9ULL, "qun221_petrice6093300.cl"},
	{0xA58889D7ACBC4CAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_reaction.gad"},
	{0xA588A5A363BC75F8ULL, "vxw_encounter2_k5_0_0.vlm"},
	{0xA588BF6F68E5F86AULL, "zz_followers_banter6159084.cl"},
	{0xA588D39D37250C93ULL, "art\\levels\\races\\vfx_models\\prp_camplogsml.phy"},
	{0xA588E29235897605ULL, "vxht_brothel_n_624_0_0.vlm"},
	{0xA588E4E03A3FEF8AULL, "6126812_m.fxe"},
	{0xA5897BBA6B935800ULL, "69494581.ogg"},
	{0xA589AAA57D3BE396ULL, "121068381.ogg"},
	{0xA589D153E6C9440CULL, "hlac_ht_0_hightown_a3_-131074_1z.dds"},
	{0xA589FC5A51F58CD8ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklback.gad"},
	{0xA58A1A4F9B430F2CULL, "9316322.ogg"},
	{0xA58A1FA62F1FC7E2ULL, "6149896_f.fxe"},
	{0xA58A7A90FE5B8E27ULL, "6120756_m.fxe"},
	{0xA58A8FCA76B47275ULL, "6013293_f.fxe"},
	{0xA58A9E11A133A4E4ULL, "6068525_m.fxe"},
	{0xA58AF16405C7DD1EULL, "336239974.ogg"},
	{0xA58AFF641160968DULL, "6177911_f.fxe"},
	{0xA58C1E6A90C3381BULL, "mcr310_sibling6076263.cl"},
	{0xA58C40FE4B6B57ACULL, "6086162_f.fxe"},
	{0xA58CA26BAC2F0D47ULL, "vxlt_warehouse_n_ab_0_0.vlm"},
	{0xA58D6192322B718AULL, "vxlt_gallowstemplar_n_5yu_0_0.vlm"},
	{0xA58D70209648EE9AULL, "377971909.ogg"},
	{0xA58D9AC04BAED8A1ULL, "vxlt_warehouse_rt_0_0.vlm"},
	{0xA58E2CB245E0FCF4ULL, "269772621.ogg"},
	{0xA58E33778967E79EULL, "vxw_encounter3_di_0_0.vlm"},
	{0xA58EB673E77F9E45ULL, "static_static_p64.psh.fxo.dx11"},
	{0xA58F6FC56FCD11B1ULL, "vxht_mansion_n_4bu_0_0.vlm"},
	{0xA58F84D7ACC2628DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_scare_handler_reaction.evt"},
	{0xA58F88696891ABB4ULL, "6137478_m.fxe"},
	{0xA58FAD0342AF0D87ULL, "zz_rdr_debug6066644.cl"},
	{0xA59101DC379FC456ULL, "402078230.ogg"},
	{0xA5911521898981B4ULL, "vxca_caverns1_i8l_0_0.vlm"},
	{0xA59176FFA84CB014ULL, "hlac_lt_warehouse_1_5y.dds"},
	{0xA5919DEA190B0317ULL, "770010772.ogg"},
	{0xA5926AA53A70F7BEULL, "vxdr_primevaltomb_7c0_0_0.vlm"},
	{0xA59270479E46FD06ULL, "ico_potion_rockarmor.dds"},
	{0xA592ACCD5A8E560BULL, "dae000_loyal_mage6165235.cl"},
	{0xA592BF712692DE70ULL, "w_sundermount_221_0.msh"},
	{0xA592CD712692F58AULL, "w_sundermount_221_0.mmh"},
	{0xA592D1712692FCC1ULL, "w_sundermount_221_0.mao"},
	{0xA59322FFD2C0EA20ULL, "vxht_0_hightown_a3_7lo_0_0.vlm"},
	{0xA5941DD448F6D7F5ULL, "6056929_m.fxe"},
	{0xA59437F56A250E64ULL, "vxdr_primevaltomb_2wr_0_1.vlm"},
	{0xA594B6701AFF59C6ULL, "871498177.ogg"},
	{0xA595A3472195D0C4ULL, "lt_0_lowtown_docks_a3_n_217_0.mmh"},
	{0xA595A7472195D717ULL, "lt_0_lowtown_docks_a3_n_217_0.mao"},
	{0xA595B9472195F666ULL, "lt_0_lowtown_docks_a3_n_217_0.msh"},
	{0xA595CFB9D2FDE7F1ULL, "gen00fl_anders6110058.cl"},
	{0xA595EAFAABF72161ULL, "179589404.ogg"},
	{0xA595F9BBD1B45FDBULL, "641471246.ogg"},
	{0xA59666B958026B18ULL, "lgt310ip_stones.ncs"},
	{0xA59683A43D7304B6ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev1.evt"},
	{0xA596945737FC5131ULL, "gen00fl_dog6165635.cl"},
	{0xA596E56EDCD504C0ULL, "6206151_f.fxe"},
	{0xA5970631C2B11B59ULL, "gen_im_trs_krk_arc_lbw_r2.uti"},
	{0xA5970EE5A4FC965EULL, "hlac_lt_0_lowtown_docks_-131073_5y.dds"},
	{0xA59748E8A1F4D8CBULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyn_0d.dds"},
	{0xA5974D52BCD5D5B9ULL, "6174690_m.fxe"},
	{0xA5976AA428938241ULL, "vxw_woundedcoast_1mz_0_0.vlm"},
	{0xA59810B3F015BDC8ULL, "6118459_m.fxe"},
	{0xA598989B94C7DB07ULL, "6036749_m.fxe"},
	{0xA598FD2C1BA5924CULL, "hlac_w_bonepit_458757_0z.dds"},
	{0xA5990C3713889BC0ULL, "pro000_athenril6050125.cl"},
	{0xA599279C87D44C1CULL, "dae000_bartrand6027828.cl"},
	{0xA59991E879A98C5DULL, "follower_banter6173098.cl"},
	{0xA599DA49D4222482ULL, "6071103_m.fxe"},
	{0xA59A0D52364C1DE9ULL, "6076220_m.fxe"},
	{0xA59B09B14E3BB26FULL, "art\\vfx\\cinematic\\vfx_dust_imp_med_p.mmh"},
	{0xA59B68C3AEC05AE5ULL, "vxca_caverns1_iif_0_0.vlm"},
	{0xA59B9838C9D6E662ULL, "gen00fl_merrill6106818.cl"},
	{0xA59C0C9D3DA4F121ULL, "vxlt_warehouse_wk_0_0.vlm"},
	{0xA59C1A8F3BCDC169ULL, "fen210_hadriana6110538.cl"},
	{0xA59C373D7806AAB0ULL, "vxca_caverns1_mines_e6s_0_0.vlm"},
	{0xA59CA5547B5E001EULL, "vxw_encounter3_aj_0_0.vlm"},
	{0xA59CAD7E194A1D69ULL, "lt_0_lowtown_docks_mmap.dds"},
	{0xA59CCF9F1C33E1E9ULL, "145399297.ogg"},
	{0xA59D7AA43D791441ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev1.gad"},
	{0xA59DC93D95D9A5BBULL, "6129773_m.fxe"},
	{0xA59E3EACBF773D48ULL, "6149291_m.fxe"},
	{0xA59E57BE75D83655ULL, "237833227.ogg"},
	{0xA59E6749DF5EFC7FULL, "one300_zevran6076507.cl"},
	{0xA59EE83F3888B925ULL, "vxht_bartrand_n_5no_0_0.vlm"},
	{0xA59F000B6A528E88ULL, "vxlt_gallowstemplar_f_5ou_0_0.vlm"},
	{0xA59F46B09B3DDA13ULL, "hlac_lt_0_lowtown_a2_n_65537_0y.dds"},
	{0xA59F48B7630135A3ULL, "hlac_lt_0_lowtown_196607_0y.dds"},
	{0xA5A036F52159A951ULL, "zz_approval_debug6003598.cl"},
	{0xA5A06C814E4CBB66ULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.cess.evt"},
	{0xA5A06D6746DFBCB1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainlink.msh"},
	{0xA5A07B6746DFD4CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainlink.mao"},
	{0xA5A0876746DFE96FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_chainlink.mmh"},
	{0xA5A08E8740C30EDAULL, "486438850.ogg"},
	{0xA5A13C9ACBB322EFULL, "655513167.ogg"},
	{0xA5A141435E1BDF9CULL, "lgt150cr_nat_were_supporter.utc"},
	{0xA5A143286830900DULL, "6152526_m.fxe"},
	{0xA5A16702E0E6F3A1ULL, "gen00flst_fenris_final.stg"},
	{0xA5A1FD311D1F3AEBULL, "hlac_lt_0_lowtown_n_65534_2x.dds"},
	{0xA5A299655ED26937ULL, "vxw_bonepitdestroyed_a3_1sa_0_0.vlm"},
	{0xA5A35F9FB3D6E84EULL, "vxht_keep_53g_0_0.vlm"},
	{0xA5A3E22CF91A1195ULL, "hlac_lt_gallowstemplar_n_2383_4x.dds"},
	{0xA5A41D5D43016264ULL, "811232027.ogg"},
	{0xA5A42FA80E851FF2ULL, "6109244_m.fxe"},
	{0xA5A47BA43D7F34EFULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\diev1.ani"},
	{0xA5A4A11281209019ULL, "6106198_aveline_barracks.cut"},
	{0xA5A4F3BEAB6299D6ULL, "bee.dds"},
	{0xA5A4F9D3316AE40CULL, "and100_anders6096018.cl"},
	{0xA5A52A2FEF362E02ULL, "235751807.ogg"},
	{0xA5A549D0E070BD1CULL, "mcr370_templar26087649.cl"},
	{0xA5A5B2FDCFE21343ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic.msh"},
	{0xA5A5BFB2A019CEF0ULL, "war_battlesynergy.dds"},
	{0xA5A5C0FDCFE22A9DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic.mmh"},
	{0xA5A5C4FDCFE231AEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic.mao"},
	{0xA5A5DC811A899AC1ULL, "vxw_woundedcoast_1hy_0_0.vlm"},
	{0xA5A63DA262163ED4ULL, "792251020.ogg"},
	{0xA5A677DA28B59B40ULL, "vxlt_0_lowtown_e8l_0_0.vlm"},
	{0xA5A6ADD170539606ULL, "vxw_bonepitmining_a2_2a3_0_0.vlm"},
	{0xA5A6D29DB6DD8C4BULL, "art\\vfx\\combat\\vfx_w_tmp_clean_aoe_p.mmh"},
	{0xA5A783814E5300B1ULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.cess.gad"},
	{0xA5A7FEE6EA909A11ULL, "w_encounter3_172_0.msh"},
	{0xA5A80CE6EA90B22CULL, "w_encounter3_172_0.mao"},
	{0xA5A8187F53A32FCEULL, "mag320cr_crowd_5.utc"},
	{0xA5A818E6EA90C6CFULL, "w_encounter3_172_0.mmh"},
	{0xA5A849F1F43917DDULL, "6051634_m.fxe"},
	{0xA5A8ECE8A8CFB436ULL, "gen00fl_anders6125554.cl"},
	{0xA5A92FF264A61F1EULL, "vxht_mansion2_35c_0_0.vlm"},
	{0xA5A9629AB6C7EB24ULL, "vxw_sundermount_302_0_0.vlm"},
	{0xA5A97525A48EB211ULL, "564798802.ogg"},
	{0xA5A9B53EED24A6CFULL, "vxlt_0_lowtown_ch9_0_0.vlm"},
	{0xA5A9DAA0A0B67365ULL, "vxca_caverns1_ile_0_0.vlm"},
	{0xA5AA14C21EB92BB2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masb_0s.dds"},
	{0xA5AA7E7A2F9B09A1ULL, "vxlt_warehouse_xj_0_0.vlm"},
	{0xA5AB13D99A041DC3ULL, "6079668_f.fxe"},
	{0xA5AB3AEAD806ADCFULL, "vxht_0_hightown_4c7_0_0.vlm"},
	{0xA5AB49FAD42691D6ULL, "dae000_keep_noble_sta6161547.cl"},
	{0xA5AB8BB4C5B91A4CULL, "vxw_encounter3_87_0_0.vlm"},
	{0xA5ABB3E432D5E0ADULL, "6057104_m.fxe"},
	{0xA5ABC96D8A79C9C4ULL, "vxlt_gallowscourtyard_n_66o_0_0.vlm"},
	{0xA5ABE83E8EE30BDFULL, "vxlt_0_lowtown_docks_a3_n_d3n_0_0.vlm"},
	{0xA5AC3BD8C536E361ULL, "69393441.ogg"},
	{0xA5ACD32274FFB0E1ULL, "vxht_mansion2b_1az_0_0.vlm"},
	{0xA5AD1C72DF9A44FAULL, "106961803.ogg"},
	{0xA5AD67FE53E3317EULL, "lt_0_lowtown_docks_a3_n_10_0_ll.mmh"},
	{0xA5AD7DB958157111ULL, "lgt310ip_stones.utp"},
	{0xA5ADC4FA1D0158FAULL, "6031283_f.fxe"},
	{0xA5ADC85547C50D3CULL, "hlac_lt_undercitydungeon_-1_21x.dds"},
	{0xA5ADD4D6E5C5A961ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_archdiv.phy"},
	{0xA5ADDD769D6E2A98ULL, "mag101_grace6035859.cl"},
	{0xA5AE84814E59215FULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.cess.ani"},
	{0xA5AE9441355C846BULL, "hlac_lt_0_lowtown_a3_n_c_-2_2y.dds"},
	{0xA5AF19B89B17E7C0ULL, "llac_lt_0_lowtown_a2_n_-1_8.dds"},
	{0xA5AF67D4ED4FCB00ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\shuffle.evt"},
	{0xA5B0D4BBB9A9C236ULL, "vxca_caverns1_igs_0_0.vlm"},
	{0xA5B0DB4CCCE985A9ULL, "screenshake.gda"},
	{0xA5B11E3E2CF5A272ULL, "204312671.ogg"},
	{0xA5B13C19C95EFD2EULL, "follower_barks6154976.cl"},
	{0xA5B1B3AAD894AF25ULL, "51857838.ogg"},
	{0xA5B1E8D93EC101D4ULL, "vxlt_gallowstemplar_f_3k7_0_0.vlm"},
	{0xA5B220F0CFEA51F0ULL, "vxht_0_hightown_9us_0_0.vlm"},
	{0xA5B2917C4C3A12E5ULL, "zz_followers_banter6159108.cl"},
	{0xA5B3446AFEF98159ULL, "lt_0_lowtown_docks_a3_190_0.msh"},
	{0xA5B34E6AFEF99247ULL, "lt_0_lowtown_docks_a3_190_0.mmh"},
	{0xA5B3526AFEF99894ULL, "lt_0_lowtown_docks_a3_190_0.mao"},
	{0xA5B3A8C3192797F6ULL, "gen00fl_merrill6116046.cl"},
	{0xA5B3DB26385EF00BULL, "674616009.ogg"},
	{0xA5B43F28189A0C1CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraitc_n.dds"},
	{0xA5B44E5E0C7FB341ULL, "vxw_woundedcoast_1sx_0_0.vlm"},
	{0xA5B4505CEB1AB8E6ULL, "967277882.ogg"},
	{0xA5B457C2D78E2203ULL, "art\\vfx\\combat\\vfx_gib_arm_meda.mao"},
	{0xA5B470BF029FED2AULL, "vxlt_ruinedhouse_n_5q_0_0.vlm"},
	{0xA5B4B39DB6E994B0ULL, "art\\vfx\\combat\\vfx_w_tmp_clean_aoe_p.anb"},
	{0xA5B4C23B8663701EULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_player_fenrintro.ani"},
	{0xA5B5017941B0352AULL, "vxca_cavern1_haunted_e8c_0_0.vlm"},
	{0xA5B52F39B6435C58ULL, "darktown_fem_fereldan_sta.cnv"},
	{0xA5B61DF044BD99F8ULL, "hlac_lt_0_lowtown_a3_n_c_0_7x.dds"},
	{0xA5B660854D460404ULL, "476068658.ogg"},
	{0xA5B662D6F29071B9ULL, "vxw_encounter3_b8_0_0.vlm"},
	{0xA5B6958E1B5D1B3EULL, "crt_sw_cyc.mao"},
	{0xA5B6A18048352998ULL, "825052163.ogg"},
	{0xA5B6C142329411D0ULL, "w_blightlands_226_0.mao"},
	{0xA5B6C542329418E3ULL, "w_blightlands_226_0.mmh"},
	{0xA5B6D3423294303DULL, "w_blightlands_226_0.msh"},
	{0xA5B73E54DD19D2B4ULL, "218504891.ogg"},
	{0xA5B744595AEC9891ULL, "990752203.ogg"},
	{0xA5B74E52CD7BCE08ULL, "w_woundedcoast_229_0.msh"},
	{0xA5B76052CD7BEC99ULL, "w_woundedcoast_229_0.mao"},
	{0xA5B764613A0BD553ULL, "qun120cr_javaris6043988.cl"},
	{0xA5B76C52CD7C0132ULL, "w_woundedcoast_229_0.mmh"},
	{0xA5B783A365F200EFULL, "6112010_m.fxe"},
	{0xA5B798FDF7C112DDULL, "zz_prmpt_debug6217999.cl"},
	{0xA5B8651E975D3001ULL, "dae200cr_meredith.utc"},
	{0xA5B8B729080DBB0AULL, "vxlt_gallowstemplar_n_5jp_0_0.vlm"},
	{0xA5B91139B6471260ULL, "darktown_fem_fereldan_sta.bnk"},
	{0xA5B937782429DE33ULL, "vxlt_0_lowtown_a2_n_hdq_0_0.vlm"},
	{0xA5B99467F181CBEEULL, "llac_lt_gallowstemplar_n_2388_0.dds"},
	{0xA5B9A9FBE039BD9DULL, "618741820.ogg"},
	{0xA5B9FD91B7AF32CCULL, "vxw_encounter3_54_0_0.vlm"},
	{0xA5BA2BF694C1E6E0ULL, "6109286_m.fxe"},
	{0xA5BAB08F01B1E38FULL, "6117776_m.fxe"},
	{0xA5BAC7C2E4A63FA7ULL, "gen_im_cns_flu_isabela_01.uti"},
	{0xA5BB5CBFF708DE57ULL, "lt_0_lowtown_docks_a3_240_0.mmh"},
	{0xA5BB60BFF708E564ULL, "lt_0_lowtown_docks_a3_240_0.mao"},
	{0xA5BB6EEAE532EFFDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\tools_flatnormal_n.dds"},
	{0xA5BB72BFF70903C9ULL, "lt_0_lowtown_docks_a3_240_0.msh"},
	{0xA5BBD986C1BBB22BULL, "6127255_f.fxe"},
	{0xA5BC02BEDAB22816ULL, "ca_cave_2194_lc.mmh"},
	{0xA5BC0EB297A32C4CULL, "hlac_lt_0_lowtown_131071_6z.dds"},
	{0xA5BC80EC61C9C72EULL, "zz_one000_merchant6179344.cl"},
	{0xA5BCE9D3AF55B680ULL, "vxlt_hangedman_267_0_0.vlm"},
	{0xA5BDC03C1053413EULL, "vxdr_primevaltomb_7r3_0_0.vlm"},
	{0xA5BF04B0E5BB7489ULL, "llac_lt_0_lowtown_a2_0_4.dds"},
	{0xA5BF117695554CBFULL, "llac_lt_0_lowtown_n_-1_1.dds"},
	{0xA5BF4698AB9FDAB6ULL, "vxca_caverns1_ijt_0_0.vlm"},
	{0xA5BF821CC143422FULL, "6085394_m.fxe"},
	{0xA5C00CB4E4FA5D0FULL, "6136935_m.fxe"},
	{0xA5C02D077F1F8076ULL, "vxlt_0_lowtown_a2_n_gpk_0_0.vlm"},
	{0xA5C041F44FB99494ULL, "vxca_cavern1_haunted_j4n_0_0.vlm"},
	{0xA5C0435997C3194EULL, "vxht_keep_59y_0_0.vlm"},
	{0xA5C0C8CF92363A02ULL, "vxht_keep_5hj_0_0.vlm"},
	{0xA5C0CBC30A662A2EULL, "vxca_caverns1_cnz_0_0.vlm"},
	{0xA5C164D628CDB75EULL, "zz_followers_banter6157262.cl"},
	{0xA5C1F3A780D03540ULL, "hlac_lt_backalley_n_1_3y.dds"},
	{0xA5C224DE0E6DA270ULL, "vxht_mansion_n_334_0_0.vlm"},
	{0xA5C25B84E109E329ULL, "6174481_m.fxe"},
	{0xA5C2DA406837D892ULL, "vxlt_0_lowtown_a2_n_gos_0_1.vlm"},
	{0xA5C352A17305556BULL, "vxht_0_hightown_bwu_0_0.vlm"},
	{0xA5C3738E02731D79ULL, "art\\vfx\\combat\\vfx_f_isa_trail_shimmer_p.dura.ani"},
	{0xA5C37E2B5EDCAA19ULL, "qun131_warrens_thug6038991.cl"},
	{0xA5C38980C7FA2CC7ULL, "gen00fl_aveline6132640.cl"},
	{0xA5C40AB14E5EB8ACULL, "art\\vfx\\cinematic\\vfx_dust_imp_med_p.anb"},
	{0xA5C41F4EC9F78774ULL, "6101963_m.fxe"},
	{0xA5C41FD51B33A0D3ULL, "415164651.ogg"},
	{0xA5C490449B0F4CB3ULL, "art\\characters\\creatures\\dog\\textures\\c_doga_0n.dds"},
	{0xA5C5350AF438F4B3ULL, "vxht_0_hightown_a3_box_0_0.vlm"},
	{0xA5C54CB45FD8668EULL, "gen_im_trs_rdr_shd_wdn_r2.uti"},
	{0xA5C55065D17520DDULL, "407926272.ogg"},
	{0xA5C5903B70F784FDULL, "gen00fl_merrill6124420.cl"},
	{0xA5C5B73F68A7C0D1ULL, "gen00fl_anders6125784.cl"},
	{0xA5C675998DF4B325ULL, "6124286_m.fxe"},
	{0xA5C67ECE13BE569BULL, "724018696.ogg"},
	{0xA5C6B6947A591DEAULL, "art\\characters\\playercharacter\\humanfemale\\glv_magc.mao"},
	{0xA5C6BA947A592439ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magc.mmh"},
	{0xA5C6C4947A593527ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magc.msh"},
	{0xA5C72072ABA67933ULL, "vrc321ip_spooky_4_owner.utp"},
	{0xA5C7847E8A86088AULL, "w_encounter3_116_0.mao"},
	{0xA5C7887E8A860FD9ULL, "w_encounter3_116_0.mmh"},
	{0xA5C7927E8A8620C7ULL, "w_encounter3_116_0.msh"},
	{0xA5C794C640945052ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava2_sa.dds"},
	{0xA5C7D9BC66E1A67DULL, "w_encounter2_62_0.mao"},
	{0xA5C7DDBC66E1ACB6ULL, "w_encounter2_62_0.mmh"},
	{0xA5C7E6B26B91B364ULL, "ht_0_hightown.rml"},
	{0xA5C7E7BC66E1BDB4ULL, "w_encounter2_62_0.msh"},
	{0xA5C8C521DFC6323AULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.cess.gad"},
	{0xA5C8DB3B86744398ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_player_fenrintro.gad"},
	{0xA5C8F20DE8554417ULL, "ico_1h_sword_5.dds"},
	{0xA5C8FA1C899E1BB0ULL, "vxlt_hangedman_27h_0_0.vlm"},
	{0xA5C90AD66DE9167BULL, "vxw_bonepitmining_a2_1kc_0_0.vlm"},
	{0xA5C921C2AF0FCFD1ULL, "dae000cr_grand_cleric6197650.cl"},
	{0xA5C9DA03E460822CULL, "hlac_dr_primevalentrance_1726_10y.dds"},
	{0xA5CA4E8E0278FCBBULL, "art\\vfx\\combat\\vfx_f_isa_trail_shimmer_p.dura.gad"},
	{0xA5CA8E0E20D014FFULL, "w_blightlands_167_0.msh"},
	{0xA5CA940E20D01EA1ULL, "w_blightlands_167_0.mmh"},
	{0xA5CAA00E20D03342ULL, "w_blightlands_167_0.mao"},
	{0xA5CAAA1154737ABCULL, "rdr111dg_martin6055059.cl"},
	{0xA5CAF82EADBB5C36ULL, "vxht_mansion2_317_0_0.vlm"},
	{0xA5CB5BB0A14BCF00ULL, "vxlt_hangedman_234_0_0.vlm"},
	{0xA5CBAD83732F98FBULL, "llac_ht_mansion2b_818_30.dds"},
	{0xA5CC1A6CD67D87CFULL, "qun131_party_comment6117299.cl"},
	{0xA5CC666B442515C5ULL, "816100102.ogg"},
	{0xA5CC73CA3995F03FULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_cass.mmh"},
	{0xA5CC79CA3995FAE1ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_cass.msh"},
	{0xA5CD3D0834F10810ULL, "530514684.ogg"},
	{0xA5CD6116F7F3F1F9ULL, "vxdr_primevaltomb_56v_0_0.vlm"},
	{0xA5CD6A22C80B69D1ULL, "hlac_w_encounter2_camp_65539_0y.dds"},
	{0xA5CDB52B7DBC19A2ULL, "vxw_bonepitmining_a2_21g_0_0.vlm"},
	{0xA5CDB8759D95F336ULL, "vxca_caverns1_imu_0_0.vlm"},
	{0xA5CDEB3EE7B22450ULL, "vxdr_primevalentrance_1m6_0_0.vlm"},
	{0xA5CDEEFC34B4B2B7ULL, "vxw_bonepitdestroyed_a3_1bb_0_0.vlm"},
	{0xA5CE5DA55E51D9EAULL, "zz_followers_banter6158919.cl"},
	{0xA5CE876B2A90AC85ULL, "gen00fl_aveline6139481.cl"},
	{0xA5CE9EE4711598F6ULL, "vxlt_0_lowtown_a2_n_gml_0_0.vlm"},
	{0xA5CEB25EBAC028CAULL, "w_sundermount_60_0.mao"},
	{0xA5CEB3D141AFAD14ULL, "vxca_cavern1_haunted_j3o_0_0.vlm"},
	{0xA5CEB65EBAC02F19ULL, "w_sundermount_60_0.mmh"},
	{0xA5CEB9800A07A22CULL, "dae000_orsino6085493.cl"},
	{0xA5CEC05EBAC04007ULL, "w_sundermount_60_0.msh"},
	{0xA5CF149C133FADABULL, "vxw_bonepit_126_0_0.vlm"},
	{0xA5CF526780A4E66EULL, "dae000_worthy6166652.cl"},
	{0xA5CF7F99C03B2ABEULL, "lgt000cr_aspire.utc"},
	{0xA5CFB433ADC26347ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_9.pwk"},
	{0xA5CFBD94D1481485ULL, "dre000st_siblingdies.stg"},
	{0xA5CFC821DFCC5651ULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.cess.evt"},
	{0xA5CFD4A4DCE505DDULL, "clf260_harley6093808.cl"},
	{0xA5D0C1C02BD8CABCULL, "hlac_ht_0_hightown_n_2_0z.dds"},
	{0xA5D0FDB4BD707B98ULL, "6042705_m.fxe"},
	{0xA5D106C857803FB6ULL, "694026630.ogg"},
	{0xA5D115A7FC643612ULL, "gen00fl_varric6093587.cl"},
	{0xA5D169F15C4FD434ULL, "6118305_m.fxe"},
	{0xA5D16B8E027F4CA0ULL, "art\\vfx\\combat\\vfx_f_isa_trail_shimmer_p.dura.evt"},
	{0xA5D1B057FA547859ULL, "llac_lt_0_lowtown_docks_-2_2.dds"},
	{0xA5D1C47E64FB6DEBULL, "vxht_0_hightown_bpt_0_0.vlm"},
	{0xA5D1FC6704D18D9EULL, "679969055.ogg"},
	{0xA5D2444A2893EF38ULL, "6080281_f.fxe"},
	{0xA5D283D45A3DCF64ULL, "6083751_m.fxe"},
	{0xA5D28FF0BBD3E410ULL, "pro000_colonel_ewald6005407.cl"},
	{0xA5D2C73078F2E39BULL, "212608468.ogg"},
	{0xA5D32013144E2455ULL, "vxca_caverns2prc_8w7_0_0.vlm"},
	{0xA5D429E27E4BDB33ULL, "6143382_m.fxe"},
	{0xA5D447DB1A2996E2ULL, "6135653_f.fxe"},
	{0xA5D4849032B2F8A9ULL, "gen00fl_isabela6088072.cl"},
	{0xA5D4A96C9B8F9B11ULL, "hlac_lt_undercityserialkiller_-1_50x.dds"},
	{0xA5D542530FD5004CULL, "vxlt_0_lowtown_a2_er6_0_0.vlm"},
	{0xA5D6D32229B960FFULL, "vxht_mansion2_2ph_0_0.vlm"},
	{0xA5D70B306502FC2AULL, "hlac_lt_ruinedhouse_n_-1_2x.dds"},
	{0xA5D724CC7C01A109ULL, "w_encounter2_camp_2_0.phy"},
	{0xA5D78DFFDBD0AADFULL, "hlac_lt_0_lowtown_a2_n_65536_4z.dds"},
	{0xA5D7A8D79FE0E283ULL, "w_encounter3_12_0.rml"},
	{0xA5D863365988E02FULL, "vxht_keep_5g6_0_0.vlm"},
	{0xA5D92E900A78B700ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_7z.dds"},
	{0xA5DA20C45B6FAA6CULL, "829265970.ogg"},
	{0xA5DA997A94F9BCD2ULL, "6214416_m.fxe"},
	{0xA5DB3E57AA3482B5ULL, "vxht_bartrand_n_6rd_0_0.vlm"},
	{0xA5DB79071236023BULL, "99518874.ogg"},
	{0xA5DB863ABE2E413EULL, "73298085.ogg"},
	{0xA5DBA9BD16CD8222ULL, "243087734.ogg"},
	{0xA5DBB4B9CA1FF97BULL, "seb321_leliana6197856.cl"},
	{0xA5DC38968C8C9BEDULL, "vxht_mansion2_1wl_0_0.vlm"},
	{0xA5DC70D9F03CE407ULL, "vxlt_gallowsprison_n_64t_0_0.vlm"},
	{0xA5DCA143CFC5A9D6ULL, "6061822_f.fxe"},
	{0xA5DCADBB105A50B6ULL, "6082149_m.fxe"},
	{0xA5DCCB716D0054D2ULL, "art\\vfx\\combat\\vfx_q_ari_heal_act_c.impa.gad"},
	{0xA5DCE021DFD70920ULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.cess.ani"},
	{0xA5DD10C1630BB176ULL, "vxlt_0_lowtown_a2_n_gjm_0_0.vlm"},
	{0xA5DD1C657F90454DULL, "dae000_bodhan6129163.cl"},
	{0xA5DD50DB21D5A840ULL, "6104574_f.fxe"},
	{0xA5DD74CF9975EA4AULL, "pro000_db_ogre_mage.cut"},
	{0xA5DDD8F3FC157C91ULL, "and230st_base_writing.stg"},
	{0xA5DE2DCC7C07CF05ULL, "w_encounter2_camp_2_0.rml"},
	{0xA5DF2C00740C8054ULL, "298469978.ogg"},
	{0xA5DF500821728B43ULL, "425222068.ogg"},
	{0xA5DF77FDE76F4DE6ULL, "596607427.ogg"},
	{0xA5DFFD987848899EULL, "hlac_lt_0_lowtown_a3_n_c_0_6y.dds"},
	{0xA5E01F75A9680013ULL, "vxlt_0_lowtown_cmr_0_0.vlm"},
	{0xA5E0A9E69267E9CFULL, "6064889_m.fxe"},
	{0xA5E0B0C98514E582ULL, "vxht_0_hightown_c71_0_0.vlm"},
	{0xA5E0F858DAFB93F6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_portcullis.msh"},
	{0xA5E10258DAFBA4F4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_portcullis.mmh"},
	{0xA5E104D3079809DCULL, "hm_alain.mor"},
	{0xA5E10658DAFBAB07ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_portcullis.mao"},
	{0xA5E16A7429356836ULL, "6016676_m.fxe"},
	{0xA5E183B013DDBCB0ULL, "vxca_caverns1_mines_fn4_0_0.vlm"},
	{0xA5E1B16F0C36A648ULL, "vxht_mansion2b_1rz_0_0.vlm"},
	{0xA5E1B3704FEF06A3ULL, "vxw_woundedcoast_121_0_0.vlm"},
	{0xA5E1CD133FCB2867ULL, "vxca_caverns1_mines_fo8_0_0.vlm"},
	{0xA5E2C7E3596383B6ULL, "6120630_m.fxe"},
	{0xA5E2CE716D04C509ULL, "art\\vfx\\combat\\vfx_q_ari_heal_act_c.impa.evt"},
	{0xA5E3623ABC15D080ULL, "778431396.ogg"},
	{0xA5E3B43001CB18CCULL, "vxlt_0_lowtown_a2_eu7_0_0.vlm"},
	{0xA5E3C1F7AE45341DULL, "twopassblurbloom7.psh.fxo"},
	{0xA5E41658CEDF7739ULL, "hlac_lt_0_lowtown_a2_n_65537_1x.dds"},
	{0xA5E4C67C3D614F1DULL, "6174791_m.fxe"},
	{0xA5E4E005422C15DAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicx.mmh"},
	{0xA5E4E405422C1C11ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicx.mao"},
	{0xA5E4F205422C3420ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicx.msh"},
	{0xA5E58213556A89CEULL, "20536000.ogg"},
	{0xA5E5B92E2BF356E7ULL, "follower_barks6153228.cl"},
	{0xA5E632DCA59713A7ULL, "cla_base.gda"},
	{0xA5E657F631E446ABULL, "lt_0_lowtown_a2_n_c_38_0_lc.mmh"},
	{0xA5E6A68262CE4AEAULL, "581148614.ogg"},
	{0xA5E6D639F1C83684ULL, "fen110_cut_fenris_intro6066812.cl"},
	{0xA5E7A3A5F1CDC09BULL, "6043588_f.fxe"},
	{0xA5E7A9235888F6B4ULL, "vxht_keep_n_6j2_0_0.vlm"},
	{0xA5E7F4527F8E0E0AULL, "950630092.ogg"},
	{0xA5E87CC0FD08381AULL, "6071381_m.fxe"},
	{0xA5E8B0A395D9B037ULL, "vxdr_primevaltomb_8na_0_0.vlm"},
	{0xA5E8D4C7CBFF3339ULL, "vxht_0_hightown_n_bc0_0_0.vlm"},
	{0xA5E8FB359EB1AB35ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_sepulcher_d.dds"},
	{0xA5E95D31F97ABFDEULL, "fen210_fenris.cnv"},
	{0xA5E96B31AE4FEF6DULL, "516574773.ogg"},
	{0xA5E96D7F0C6F94F5ULL, "vxdr_0_deeproads_cuq_0_2.vlm"},
	{0xA5E97CE2AD38D0A2ULL, "hlac_w_encounter2_camp_-131072_0y.dds"},
	{0xA5E9DC32F950DFA6ULL, "vxw_bonepitdestroyed_a3_ow_0_0.vlm"},
	{0xA5EA3C33B9E24F74ULL, "503082927.ogg"},
	{0xA5EAC816500D4D60ULL, "319825578.ogg"},
	{0xA5EACC9D3777CB08ULL, "art\\levels\\races\\vfx_models\\prp_camplogsml.mmh"},
	{0xA5EAD89D3777DFABULL, "art\\levels\\races\\vfx_models\\prp_camplogsml.mao"},
	{0xA5EAEA9D3777FE32ULL, "art\\levels\\races\\vfx_models\\prp_camplogsml.msh"},
	{0xA5EB707130CC55D6ULL, "75522933.ogg"},
	{0xA5EBE1C0819147E7ULL, "6059306_m.fxe"},
	{0xA5EBF551BEA49F73ULL, "580709840.ogg"},
	{0xA5EC30576E8F4C66ULL, "gen_im_trs_gem_onyx_04.uti"},
	{0xA5EC347B223F27C2ULL, "fen110_fenris6118045.cl"},
	{0xA5ECCDF2628B6B3DULL, "vxht_0_hightown_a3_a79_0_0.vlm"},
	{0xA5ECDFE131584444ULL, "vxlt_gallowstemplar_f_687_0_0.vlm"},
	{0xA5ECE05F6B0F3E27ULL, "6116155_m.fxe"},
	{0xA5ED3F31F97E75EEULL, "fen210_fenris.bnk"},
	{0xA5ED51B6D25F17A3ULL, "6039766_m.fxe"},
	{0xA5ED6F8C676BB877ULL, "6115323_m.fxe"},
	{0xA5EDAA5FB0105CACULL, "260663401.ogg"},
	{0xA5EDBD64613E605EULL, "6120599_m.fxe"},
	{0xA5EDD5B77F8DE0CAULL, "226362513.ogg"},
	{0xA5EE88551B0AE8E7ULL, "fx_blood02.mao"},
	{0xA5EF3EF11E3C289CULL, "vxw_encounter2_pq_0_1.vlm"},
	{0xA5EFD60BA037E8FCULL, "llac_ht_0_hightown_a2_-131070_0.dds"},
	{0xA5F015D042A434EFULL, "6099213_f.fxe"},
	{0xA5F038104DBE8D86ULL, "6148153_m.fxe"},
	{0xA5F03F28D86D50BBULL, "vxht_0_hightown_a3_c80_0_0.vlm"},
	{0xA5F0534467AF5D83ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robc_0n.dds"},
	{0xA5F062B9AA8F152EULL, "llac_lt_0_lowtown_docks_n_196607_0.dds"},
	{0xA5F063E626EE4AC1ULL, "mag291_torpor6026988.cl"},
	{0xA5F07B5335E54BAAULL, "vxlt_undercitydungeon_1a5_0_0.vlm"},
	{0xA5F09D3A91120527ULL, "810725134.ogg"},
	{0xA5F0A73B9983E8F4ULL, "vxlt_undercityserialkiller_ym_0_0.vlm"},
	{0xA5F0B6716D10DA78ULL, "art\\vfx\\combat\\vfx_q_ari_heal_act_c.impa.ani"},
	{0xA5F0FA587FB5906DULL, "qun221_viscount6073891.cl"},
	{0xA5F13B64BA547D95ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnr90pause.evt"},
	{0xA5F2260CF3C1314CULL, "vxlt_0_lowtown_a2_ex4_0_0.vlm"},
	{0xA5F2D810190C5328ULL, "154586087.ogg"},
	{0xA5F2DA31AF9FC62CULL, "112909295.ogg"},
	{0xA5F2DD6B89E9E891ULL, "6120765_m.fxe"},
	{0xA5F329AB637014D4ULL, "hlac_lt_0_lowtown_a3_n_c_1_0z.dds"},
	{0xA5F383FE94421199ULL, "ave200_donnic6112139.cl"},
	{0xA5F49474B94F6C73ULL, "825028657.ogg"},
	{0xA5F4B6153AA928AAULL, "959786550.ogg"},
	{0xA5F4C5F77AA3A4A4ULL, "335548431.ogg"},
	{0xA5F4DCA94F5589FEULL, "hlac_ht_keep_3732_10y.dds"},
	{0xA5F54B3B711CCA7EULL, "hlac_lt_0_lowtown_docks_-4_7z.dds"},
	{0xA5F58AFC363173F9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_1_ian.gad"},
	{0xA5F5C30008506370ULL, "mag100ar_encounter.are"},
	{0xA5F5CC8C3F35A027ULL, "408117462.ogg"},
	{0xA5F5E06E35EBE9A5ULL, "vxlt_0_lowtown_a2_n_gv6_0_0.vlm"},
	{0xA5F6169468B7D8A3ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.dura.evt"},
	{0xA5F6FC001AAFD4A1ULL, "vxw_encounter3_4z_0_0.vlm"},
	{0xA5F702E8B0AA509BULL, "336188581.ogg"},
	{0xA5F704241B01C9BBULL, "132416514.ogg"},
	{0xA5F7174B713A2BFEULL, "6169840_m.fxe"},
	{0xA5F782264BA17D62ULL, "768379356.ogg"},
	{0xA5F78F31C78B971EULL, "vxdr_primevalentrance_3bs_0_0.vlm"},
	{0xA5F7950E20F70E1EULL, "w_blightlands_167_0.phy"},
	{0xA5F82FB241A5F465ULL, "zz_followers_banter6158877.cl"},
	{0xA5F87B3FF6209A09ULL, "w_encounter2_53_0.phy"},
	{0xA5F91FC36CDC31D6ULL, "vxlt_0_lowtown_docks_a3_n_d0r_0_0.vlm"},
	{0xA5F977E130264BDFULL, "hlac_lt_0_lowtown_a3_n_c_-262143_0y.dds"},
	{0xA5F9DA0008547372ULL, "mag100ar_encounter.ncs"},
	{0xA5F9DE832CF7BD1BULL, "dae000_pc_home_ambients6208179.cl"},
	{0xA5F9EA2E39EFDDCFULL, "hlac_w_encounter3_-65532_0x.dds"},
	{0xA5FA54F1500AA8EFULL, "llac_w_woundedcoast_0_0.dds"},
	{0xA5FA9B5C5ECF213BULL, "art\\characters\\creatures\\mubari\\animation\\eatloop.evt"},
	{0xA5FAA9EA05DFEF24ULL, "6050067_m.fxe"},
	{0xA5FB01552B88676AULL, "art\\characters\\playercharacter\\humanmale\\boo_lgta.mao"},
	{0xA5FB05552B886DB9ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgta.mmh"},
	{0xA5FB0F552B887EA7ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgta.msh"},
	{0xA5FB1A5A39150F20ULL, "zz_followers_banter6159231.cl"},
	{0xA5FB23A6E9B9E6C2ULL, "mcr310_argument6072706.cl"},
	{0xA5FB29655A0B5893ULL, "151697990.ogg"},
	{0xA5FB635F0A12D98EULL, "479305350.ogg"},
	{0xA5FB9BFC3635FC47ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_4_1_ian.ani"},
	{0xA5FBAFDB2ABFE539ULL, "hlac_dr_slavers_3214_1z.dds"},
	{0xA5FC38751D129A37ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_die_v1.gad"},
	{0xA5FC805B3EC5912FULL, "274579241.ogg"},
	{0xA5FC8C389A3E8213ULL, "lgt210_eau_symp_templar6154722.cl"},
	{0xA5FC924F21836B54ULL, "dae000_porfiria6162653.cl"},
	{0xA5FC9AE3FA0FFDB3ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_15ring_uhm.mao"},
	{0xA5FCBB9682BCC411ULL, "vxlt_0_lowtown_fea_0_0.vlm"},
	{0xA5FCEF9468BDB524ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.dura.gad"},
	{0xA5FD0D0290531531ULL, "vxlt_gallowscourtyard_768_0_0.vlm"},
	{0xA5FD4D6022CC8C2FULL, "art\\levels\\races\\proxy\\props\\_textures\\corpses_n.dds"},
	{0xA5FD6C4C55C88EB6ULL, "455730374.ogg"},
	{0xA5FD9246795AE651ULL, "vxht_0_hightown_a2_c8s_0_0.vlm"},
	{0xA5FDA803EF9D57BCULL, "dae021_sister_samea_sta6150137.cl"},
	{0xA5FE08360BB26544ULL, "vxlt_0_lowtown_a2_gm1_0_0.vlm"},
	{0xA5FE4C0C3E450DBFULL, "art\\vfx\\combat\\vfx_r_gn2_accu_act_c.mmh"},
	{0xA5FE529E56C2F168ULL, "vxca_caverns2_6h8_0_0.vlm"},
	{0xA5FEA5A9A64CA6D8ULL, "zz_ztest.cut"},
	{0xA5FF1FC9420B4156ULL, "gen00fl_anders6128324.cl"},
	{0xA5FF4C71B4CB484AULL, "andcr_templar_captain.ncs"},
	{0xA5FF68496246CF12ULL, "262466079.ogg"},
	{0xA5FF843FF626C805ULL, "w_encounter2_53_0.rml"},
	{0xA5FFC51EA0494BE7ULL, "vxlt_0_lowtown_f4p_0_0.vlm"},
	{0xA5FFD2000858D163ULL, "mag100ar_encounter.lst"},
	{0xA60039718AB6D0DBULL, "673729678.ogg"},
	{0xA6004FF8A316AF0BULL, "170771792.ogg"},
	{0xA600945C5ED3817CULL, "art\\characters\\creatures\\mubari\\animation\\eatloop.gad"},
	{0xA600954AAA67211DULL, "llac_ht_0_hightown_a3_-1_1.dds"},
	{0xA601090FB731A106ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.impa.ani"},
	{0xA6020AC181CC3203ULL, "hlac_lt_0_lowtown_docks_n_-131075_2z.dds"},
	{0xA60238451967C589ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.ani"},
	{0xA6024D751D1729C4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_die_v1.evt"},
	{0xA60257D933FADD40ULL, "6111482_f.fxe"},
	{0xA602A8881A182D8EULL, "1154867.ogg"},
	{0xA602AE40EE3ED90FULL, "vxca_caverns2prc_9zg_0_0.vlm"},
	{0xA602CC11B0A0D92CULL, "hlac_lt_gallowscourtyard_n_-131072_2z.dds"},
	{0xA602D24A35C1787BULL, "vxw_bonepitmining_a2_1wg_0_0.vlm"},
	{0xA6030DCC7C26B52EULL, "w_encounter2_camp_2_0.mmh"},
	{0xA60319CC7C26C9C5ULL, "w_encounter2_camp_2_0.mao"},
	{0xA60327CC7C26E11CULL, "w_encounter2_camp_2_0.msh"},
	{0xA60352D802050344ULL, "gen00fl_anders6125462.cl"},
	{0xA6036C7E129FD007ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishock_pcoverride.anb"},
	{0xA603D9123D284B24ULL, "gen00fl_anders6113496.cl"},
	{0xA603FE9468C3ED02ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_mod_c.dura.ani"},
	{0xA60426F850FC5C41ULL, "6134852_m.fxe"},
	{0xA60430A0C054212EULL, "art\\vfx\\textures\\vfx_noise.dds"},
	{0xA604C66CEDF96B74ULL, "247660562.ogg"},
	{0xA6057E5C402FA3FDULL, "6136876_m.fxe"},
	{0xA605945D79C5E137ULL, "vxdr_primevaltomb_8dc_0_0.vlm"},
	{0xA605B881AFEB6439ULL, "vxht_0_hightown_n_bi6_0_0.vlm"},
	{0xA60619CF4529ABABULL, "gen00fl_sebastian6188720.cl"},
	{0xA606ADBA20B38629ULL, "78464963.ogg"},
	{0xA606B99A46B429C1ULL, "6075979_m.fxe"},
	{0xA607350E507EDED1ULL, "246183357.ogg"},
	{0xA6076BA0B85AE4D3ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.dura.evt"},
	{0xA607935C5ED99E7AULL, "art\\characters\\creatures\\mubari\\animation\\eatloop.ani"},
	{0xA607B4A1CD051B06ULL, "6099408_m.fxe"},
	{0xA608D8B4670BD18EULL, "llac_dr_0_deeproads_5849_0.dds"},
	{0xA6095345196E120BULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.gad"},
	{0xA60A37DABFC84A28ULL, "6206246_m.fxe"},
	{0xA60A697CF0E2770CULL, "588843642.ogg"},
	{0xA60A80A7FC140DCFULL, "qun160_party_comment_16061233.cl"},
	{0xA60A96D7A00C1D9AULL, "w_encounter3_12_0.msh"},
	{0xA60AA4D7A00C35D3ULL, "w_encounter3_12_0.mao"},
	{0xA60AA8D7A00C3CE0ULL, "w_encounter3_12_0.mmh"},
	{0xA60AAF67405F7D60ULL, "vxw_encounter2_2f_0_0.vlm"},
	{0xA60AB324824DBE16ULL, "mer240_jansen6045113.cl"},
	{0xA60AFF05424D19E5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_basicx.phy"},
	{0xA60B7E9AC9AFDF64ULL, "185709204.ogg"},
	{0xA60B8EA170DA6D8CULL, "mag210_arianni6024124.cl"},
	{0xA60B97C5F72024D1ULL, "190436745.ogg"},
	{0xA60BD20393992C5AULL, "gen00fl_varric6200022.cl"},
	{0xA60BDD4C88D9E4FAULL, "6148441_m.fxe"},
	{0xA60C06762E3C452EULL, "zz_qun_debug6028902.cl"},
	{0xA60D22E2BC5981BBULL, "vxht_0_hightown_a3_c62_0_0.vlm"},
	{0xA60D2BA879FF6C4CULL, "zz_mcrpt_debug.ncs"},
	{0xA60D5F0D19D17CAAULL, "vxlt_undercitydungeon_1o7_0_0.vlm"},
	{0xA60D964F68BBA5E1ULL, "6128350_m.fxe"},
	{0xA60E1208D6413083ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_flat.evt"},
	{0xA60E64A0B860F714ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.dura.gad"},
	{0xA60EBDFEA65B426EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldidle.ani"},
	{0xA60EF10FB73DB677ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.impa.evt"},
	{0xA60EF7AB3E711EE5ULL, "vxca_caverns1_i1n_0_0.vlm"},
	{0xA60F28AF039EDAC8ULL, "6118648_m.fxe"},
	{0xA60F851D62BF6964ULL, "276111505.ogg"},
	{0xA60FA803BA54A5E8ULL, "vxht_0_hightown_a7m_0_0.vlm"},
	{0xA60FA80AF3072BF8ULL, "6133812_m.fxe"},
	{0xA610343C0B0EC41EULL, "vxw_encounter3_9b_0_0.vlm"},
	{0xA610504519742C70ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_rf_p_lfoot.evt"},
	{0xA6105A4F3996A910ULL, "hm_gen_coterie_rogue.mor"},
	{0xA61065751D238FBDULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_die_v1.ani"},
	{0xA6106B7B0171AD1AULL, "6044828_m.fxe"},
	{0xA6107F687C2BA99BULL, "vxdr_0_deeproads_9dl_0_0.vlm"},
	{0xA610BC235F9FE90FULL, "hlac_ht_chantry_2534_21y.dds"},
	{0xA610D9D6E61A2E9DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_archdiv.mao"},
	{0xA610DDD6E61A35D6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_archdiv.mmh"},
	{0xA610E7D6E61A46D4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_archdiv.msh"},
	{0xA611420EE223A4FBULL, "6132809_m.fxe"},
	{0xA61154247478162BULL, "905104923.ogg"},
	{0xA611B906D1598D0CULL, "hlac_w_bonepit_262144_0y.dds"},
	{0xA6120D4275121437ULL, "17717530.ogg"},
	{0xA612454211E201BAULL, "vxht_mansion2_3g9_0_0.vlm"},
	{0xA61255E4D63A1C35ULL, "dae000cr_grand_cleric6173667.cl"},
	{0xA6136B52CDCA5E2DULL, "w_woundedcoast_229_0.phy"},
	{0xA6146C108FED7881ULL, "mag211_marethari6025335.cl"},
	{0xA614EB08D6470D04ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_flat.gad"},
	{0xA614F20FB7422300ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_l_c.impa.gad"},
	{0xA6150E11EBAFC073ULL, "rdr100st_martin.utm"},
	{0xA6152826AEE81396ULL, "one100dg_nabil6033126.cl"},
	{0xA61573A0B8672FF2ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.dura.ani"},
	{0xA615BAB324AC0780ULL, "lowtown_male_alienage1_sta6162781.cl"},
	{0xA615EEAD7BA1FE59ULL, "100986269.ogg"},
	{0xA616000A5464FDEDULL, "vxht_mansion2_1kh_0_0.vlm"},
	{0xA61688FF42638CBDULL, "character_character_p172.psh.fxo.dx11"},
	{0xA6172DF1DAD217D9ULL, "201215851.ogg"},
	{0xA6175CA0A78FE4A1ULL, "638249612.ogg"},
	{0xA617ADD3FA3D7BBEULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1.evt"},
	{0xA617BA054DE85EF8ULL, "clf260_harley6091412.cl"},
	{0xA617FE6D85BBA7F4ULL, "vxdr_primevaltomb_6yu_0_0.vlm"},
	{0xA61840D0CCA68E27ULL, "hlac_ht_chantry_n_2534_40y.dds"},
	{0xA618904C44FA535CULL, "6088097_m.fxe"},
	{0xA61915C31EE70507ULL, "6107175_m.fxe"},
	{0xA619347BB2BE066CULL, "vis110_commoner26069055.cl"},
	{0xA61952282DFE0A81ULL, "vxca_caverns1_mines_e53_0_0.vlm"},
	{0xA619650A09560D0AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_maga_0n.dds"},
	{0xA61974A05C99CA60ULL, "vxw_blightlands_ap_0_0.vlm"},
	{0xA61A0C5170388C21ULL, "187904079.ogg"},
	{0xA61A4EA57A0143E9ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm01.phy"},
	{0xA61A7E582E1FB019ULL, "w_bonepitdestroyed_a3_119_0.msh"},
	{0xA61A88582E1FC107ULL, "w_bonepitdestroyed_a3_119_0.mmh"},
	{0xA61A8C582E1FC854ULL, "w_bonepitdestroyed_a3_119_0.mao"},
	{0xA61AD29A23A570F9ULL, "vxw_bonepit_iv_0_0.vlm"},
	{0xA61B587668539CF7ULL, "follower_banter6136671.cl"},
	{0xA61B6A5873754691ULL, "lgt110_orz_harr_carta6109055.cl"},
	{0xA61B90D026E4599FULL, "gen00fl_dog6170613.cl"},
	{0xA61BB5FEA665BEBFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldidle.evt"},
	{0xA61BFA08D64D45E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_flat.ani"},
	{0xA61C1AC363E8F75EULL, "vrc321_varric6087080.cl"},
	{0xA61C8050DF7C1CC2ULL, "718905547.ogg"},
	{0xA61C83BB5A3439F4ULL, "6033118_m.fxe"},
	{0xA61DC863D53A9205ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgtp_0d.dds"},
	{0xA61E3C012022CADEULL, "hlac_ca_caverns1_24424_16x.dds"},
	{0xA61E4FC88D6A3B0AULL, "mer210pt_cave_crawling.ncs"},
	{0xA61E643BC358012DULL, "6127603_m.fxe"},
	{0xA61EA618FD04DC9EULL, "vxw_encounter3_2c_0_0.vlm"},
	{0xA61EC4D3FA43C1E9ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1.gad"},
	{0xA61EF08DE385D9D3ULL, "6094877_m.fxe"},
	{0xA61F91FAD22B0A0FULL, "vxca_caverns2prc_9la_0_0.vlm"},
	{0xA61FFEA35F4ABBAAULL, "168384585.ogg"},
	{0xA620C8E673F32BB0ULL, "248209007.ogg"},
	{0xA620F3439121ABACULL, "955984761.ogg"},
	{0xA621023C37421921ULL, "139009887.ogg"},
	{0xA62129A28FA723DAULL, "mrl210_merrill6102117.cl"},
	{0xA6212AD40851F752ULL, "fxe_tree_beam.dds"},
	{0xA62186FDEE01138EULL, "w_blightlands_probe_go.dds"},
	{0xA62193199969C7A2ULL, "lgt210_shp_kill_qunari16156858.cl"},
	{0xA621A736FA759BD0ULL, "vxca_caverns2_8cz_0_1.vlm"},
	{0xA621BB36D0BEA1ECULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_flourish_additive.ani"},
	{0xA622186E3D420A75ULL, "art\\levels\\races\\proxy\\props\\_textures\\vfxsign_d.dds"},
	{0xA6222D96242FF2CFULL, "6118063_m.fxe"},
	{0xA62296FEA66BA8A8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldidle.gad"},
	{0xA622EC663609F766ULL, "vxht_mansion2_3dx_0_0.vlm"},
	{0xA6230178833EA488ULL, "4976143.ogg"},
	{0xA623594BF5D2E24CULL, "vxdr_primevalentrance_332_0_0.vlm"},
	{0xA623AAF7981B7A5CULL, "vxlt_hangedman_26k_0_0.vlm"},
	{0xA623AFABBC3615F6ULL, "6060325_m.fxe"},
	{0xA623CD94B0635C15ULL, "plt_mag321pt_bestservedcold.nss"},
	{0xA6243C18B6D3AEE5ULL, "vxlt_undercitydungeon_1ig_0_1.vlm"},
	{0xA6243E95A0165DCEULL, "dae000_unhappy_mage6165340.cl"},
	{0xA624449C0A50E60EULL, "155705182.ogg"},
	{0xA624643FF645AE2EULL, "w_encounter2_53_0.mmh"},
	{0xA624703FF645C2C5ULL, "w_encounter2_53_0.mao"},
	{0xA6247E3FF645DA1CULL, "w_encounter2_53_0.msh"},
	{0xA624985BB0F14C06ULL, "6093608_m.fxe"},
	{0xA62505F3FEC629C8ULL, "316969268.ogg"},
	{0xA62595D3FA499037ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v1.ani"},
	{0xA625998B736375D6ULL, "287267639.ogg"},
	{0xA625A3B77884C946ULL, "vxw_woundedcoast_1cs_0_0.vlm"},
	{0xA62673A549AEB84FULL, "art\\levels\\races\\design\\_textures\\prp_bigtreasure_e.dds"},
	{0xA626B07879E6EC8EULL, "vxca_caverns1_mines_cmr_0_0.vlm"},
	{0xA626DC108CA0A67DULL, "6060836_m.fxe"},
	{0xA626E78EE01A40A1ULL, "557319494.ogg"},
	{0xA626F9F76161C5E6ULL, "hlac_w_encounter1_-65531_0z.dds"},
	{0xA62716FDC9DE1343ULL, "hlac_ht_keep_3736_1x.dds"},
	{0xA62747E807EE3891ULL, "781289165.ogg"},
	{0xA6274D0C3E6814FCULL, "art\\vfx\\combat\\vfx_r_gn2_accu_act_c.anb"},
	{0xA627C836D0C3232EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_flourish_additive.gad"},
	{0xA627E67D4E8FE2E0ULL, "vxw_blightlands_jq_0_0.vlm"},
	{0xA6283F171ABEFD54ULL, "vxht_0_hightown_a2_a6v_0_0.vlm"},
	{0xA628AE5795A5F355ULL, "vxdr_primevaltomb_3iq_0_0.vlm"},
	{0xA628BB947AABF026ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magc.phy"},
	{0xA6292663D922EEE9ULL, "art\\characters\\playercharacter\\qunarimale\\uni_ars.phy"},
	{0xA62942AE3F935CA5ULL, "art\\vfx\\combat\\vfx_str_app_c.anb"},
	{0xA629A8352CA53AE8ULL, "vxca_caverns2_6y7_0_0.vlm"},
	{0xA629E971E34DF07FULL, "rog_shadowsneakattack.dds"},
	{0xA62A4D37F3FCAEAEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter_walk.ani"},
	{0xA62A7C133DBC7A6BULL, "6124423_f.fxe"},
	{0xA62A897E8ADA8EC6ULL, "w_encounter3_116_0.phy"},
	{0xA62AB65EBB0E8E72ULL, "w_sundermount_60_0.rml"},
	{0xA62B2321A2DC5FD0ULL, "vxlt_warehouse_n_14u_0_0.vlm"},
	{0xA62B4B35A58ADF60ULL, "hlac_ca_cavern1_haunted_24425_10z.dds"},
	{0xA62B54FA38CD6A1DULL, "658296210.ogg"},
	{0xA62B55DC1405DEA6ULL, "463429349.ogg"},
	{0xA62C03C854422396ULL, "w_bonepitdestroyed_a3_202_0.msh"},
	{0xA62C0DC854423494ULL, "w_bonepitdestroyed_a3_202_0.mmh"},
	{0xA62C11C854423BA7ULL, "w_bonepitdestroyed_a3_202_0.mao"},
	{0xA62C39549DD005E2ULL, "mag330_walter6045866.cl"},
	{0xA62C6170D3F1DCC7ULL, "361809736.ogg"},
	{0xA62D5051EA185988ULL, "vxlt_gallowstemplar_n_5iw_0_0.vlm"},
	{0xA62DC7F128EDBA10ULL, "dae310ip_store_access.utp"},
	{0xA62DDA0BD393E627ULL, "hightown_market_guard_sta6169586.cl"},
	{0xA62E6A4B35C36DC0ULL, "337926752.ogg"},
	{0xA62E7CCA39E8C984ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_cass.phy"},
	{0xA62EC336D0C9390DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_flourish_additive.evt"},
	{0xA62ECD13E8DCD878ULL, "dre000_sibling_has_blight6036533.cl"},
	{0xA62F2D8DD28476FDULL, "81287724.ogg"},
	{0xA62F8EA79AC0CD63ULL, "6027963_m.fxe"},
	{0xA62F9EB6BB7F6C8CULL, "art\\characters\\playercharacter\\humanfemale\\glv_dksa.msh"},
	{0xA62FA4B6BB7F777EULL, "art\\characters\\playercharacter\\humanfemale\\glv_dksa.mmh"},
	{0xA6300DC8D609FD51ULL, "zz_followers_banter6157379.cl"},
	{0xA6308AA63573DE0BULL, "mag311_keran6064111.cl"},
	{0xA630B75EBB12FB06ULL, "w_sundermount_60_0.phy"},
	{0xA630DAD31B752829ULL, "vxht_keep_67o_0_0.vlm"},
	{0xA63111FA6FBA0BC8ULL, "gen00fl_sebastian6174043.cl"},
	{0xA631A54613D3C748ULL, "292428607.ogg"},
	{0xA631C6A1CC5055FAULL, "803665204.ogg"},
	{0xA63208C1D514A445ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_3x.dds"},
	{0xA63220AD0860E7FFULL, "6129131_m.fxe"},
	{0xA633009DCA002937ULL, "character_rendertodepth_p1.vsh.fxo"},
	{0xA6333A69B785AC40ULL, "6108589_f.fxe"},
	{0xA63343FF991A7794ULL, "486559667.ogg"},
	{0xA633567907B586FEULL, "372874694.ogg"},
	{0xA63415946A7AE1C6ULL, "vxw_woundedcoast_1nt_0_0.vlm"},
	{0xA6349CB0B5774F50ULL, "hlac_lt_foundry_n_320_23z.dds"},
	{0xA634EC0034C0E1B6ULL, "6023957_m.fxe"},
	{0xA63518718975333FULL, "hlac_lt_0_lowtown_a3_n_c_-131072_9z.dds"},
	{0xA635C55E35F45011ULL, "404359257.ogg"},
	{0xA6362DF55D44D970ULL, "6057318_m.fxe"},
	{0xA636338CBB029CD5ULL, "mer320_hubert6053049.cl"},
	{0xA63642707B83FD15ULL, "330549245.ogg"},
	{0xA636484B0A053403ULL, "art\\vfx\\combat\\vfx_w_gn1_second_wind_c.impa.evt"},
	{0xA636B7502E0DC95AULL, "dae000_ineria6140550.cl"},
	{0xA636EB155EB8F7C8ULL, "hlac_ca_caverns2prc_5802_24x.dds"},
	{0xA637053C2BBDF094ULL, "dre000st_varric_wayaroundb.stg"},
	{0xA63778DF877261E3ULL, "llac_lt_0_lowtown_a2_131069_0.dds"},
	{0xA637ACDDF01C64BAULL, "165514316.ogg"},
	{0xA638022CF461D638ULL, "vxlt_0_lowtown_a2_n_eow_0_1.vlm"},
	{0xA6382F2E411BA886ULL, "vxlt_0_lowtown_a2_n_eo9_0_1.vlm"},
	{0xA6383F0FE2005A5EULL, "6202265_f.fxe"},
	{0xA6384537F408DEFFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter_walk.evt"},
	{0xA638C6555EDC9ECDULL, "gen00fl_isabela6135714.cl"},
	{0xA638F9D4086635D0ULL, "fxe_tree_beam.mao"},
	{0xA639C4DC4049B2B9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mase_0t.dds"},
	{0xA639EADA552AAA1EULL, "vxlt_0_lowtown_a2_eqa_0_0.vlm"},
	{0xA639F31813A1BF4EULL, "561268633.ogg"},
	{0xA639F5BCE3136D5BULL, "qun111_ketojan6041028.cl"},
	{0xA639F71A3CE5B822ULL, "vxht_mansion2b_2o8_0_0.vlm"},
	{0xA63AB9EEF3F76C90ULL, "6149181_m.fxe"},
	{0xA63B10DF36DC95C2ULL, "6122227-shortscene.cut"},
	{0xA63B47F5A744322DULL, "6129803_m.fxe"},
	{0xA63BBACCD0ABE0D2ULL, "204018715.ogg"},
	{0xA63BC22EDC0E7208ULL, "vxlt_gallowstemplar_n_5lt_0_0.vlm"},
	{0xA63C0158DB484EEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_portcullis.phy"},
	{0xA63C3A641D3DC12AULL, "vxca_caverns2prc_a0d_0_0.vlm"},
	{0xA63C5424C5161C1EULL, "dae000_male_commoner2_mob6174444.cl"},
	{0xA63CB26874FA5C58ULL, "dae120ip_chant_board_ht_nt.utp"},
	{0xA63CC3F075899B99ULL, "hlac_ht_playershouse_n_855_7y.dds"},
	{0xA63D214B0A0B1084ULL, "art\\vfx\\combat\\vfx_w_gn1_second_wind_c.impa.gad"},
	{0xA63E4E3A12EC01F3ULL, "939176934.ogg"},
	{0xA63F2637F40EC8E8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_enter_walk.gad"},
	{0xA63F8316143D1FF4ULL, "836408704.ogg"},
	{0xA63FE62345126F19ULL, "vxht_0_hightown_a3_b8s_0_0.vlm"},
	{0xA6406E55461A19B6ULL, "vxw_encounter3_67_0_0.vlm"},
	{0xA6414E4ADF4C6E20ULL, "hlac_w_woundedcoast_196612_0x.dds"},
	{0xA641D39F4D6C5362ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier4_epilogue.gad"},
	{0xA641F0BE00DD65D1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_wmda_0s.dds"},
	{0xA642441E1969F5CEULL, "vxht_keep_5dr_0_0.vlm"},
	{0xA64287715C70FA46ULL, "vxw_woundedcoast_1iu_0_0.vlm"},
	{0xA642B52639382D8FULL, "6148552_m.fxe"},
	{0xA6436079B20309EAULL, "6114000_m.fxe"},
	{0xA643CDF58E50BED7ULL, "vxht_0_hightown_n_c95_0_0.vlm"},
	{0xA643F3F6DC9B351AULL, "nat380_nathaniel6066672.cl"},
	{0xA644049AC7023E76ULL, "6116671_m.fxe"},
	{0xA644304B0A114962ULL, "art\\vfx\\combat\\vfx_w_gn1_second_wind_c.impa.ani"},
	{0xA644793FA8F62939ULL, "vxht_0_hightown_n_9uy_0_0.vlm"},
	{0xA64488A12B40AE37ULL, "llac_lt_gallowscourtyard_n_-1_0.dds"},
	{0xA645831F4841C7FAULL, "vxlt_0_lowtown_a2_gph_0_0.vlm"},
	{0xA645BC2A48BA19FCULL, "6075380_f.fxe"},
	{0xA645CC94B0804176ULL, "plt_mag321pt_bestservedcold.plo"},
	{0xA645F3166105D054ULL, "6113119_m.fxe"},
	{0xA6460B76E558EB32ULL, "259140838.ogg"},
	{0xA64637A57A26570EULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm01.mmh"},
	{0xA64651A57A2683FCULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm01.msh"},
	{0xA64671582E44D53CULL, "w_bonepitdestroyed_a3_119_0.phy"},
	{0xA646DECF7DC42233ULL, "vxlt_hangedman_31t_0_0.vlm"},
	{0xA6470145B6E3D954ULL, "hlac_ca_caverns1_mines_24424_25y.dds"},
	{0xA64733F339E53561ULL, "6175103_m.fxe"},
	{0xA6473FB9104F57E0ULL, "hlac_ht_chantry_n_2534_37x.dds"},
	{0xA648A7E7E84A6753ULL, "and100_anders6111664.cl"},
	{0xA648D7E161DD2376ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_right.ani"},
	{0xA649282B8F547F52ULL, "llac_lt_0_lowtown_docks_a3_n_-262144_0.dds"},
	{0xA649C583FA6C7282ULL, "vxht_brothel_n_5gg_0_0.vlm"},
	{0xA64A33F754C37CA3ULL, "6118258_m.fxe"},
	{0xA64A9D7181B210CBULL, "qcr000_varric6065800.cl"},
	{0xA64AAC7436582103ULL, "hlac_ht_keep_3735_0z.dds"},
	{0xA64AE4EDA8C74482ULL, "w_bonepitdestroyed_a3_233_0.msh"},
	{0xA64AE6EDA8C74838ULL, "w_bonepitdestroyed_a3_233_0.mmh"},
	{0xA64AF2EDA8C75CDBULL, "w_bonepitdestroyed_a3_233_0.mao"},
	{0xA64AFFA0ABFC342AULL, "llac_ca_caverns1_mines_24426_8.dds"},
	{0xA64B1B456CFC8A0EULL, "118455576.ogg"},
	{0xA64B3826FC687179ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj02.msh"},
	{0xA64B4226FC688267ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj02.mmh"},
	{0xA64BC264B277EFDFULL, "vxca_caverns1_mines_im9_0_0.vlm"},
	{0xA64BCC8B6FF37B70ULL, "mcr310_party_banter6087773.cl"},
	{0xA64BCD79919FA30EULL, "vxht_0_hightown_a6h_0_0.vlm"},
	{0xA64CA839D0BC6F4CULL, "mag360_thrask6044293.cl"},
	{0xA64CABE02D6F470CULL, "6119566_m.fxe"},
	{0xA64CE13DEBC66B63ULL, "23597814.ogg"},
	{0xA64E0597C60B0C59ULL, "character_character_p183.psh.fxo.dx11.assembly"},
	{0xA64E143D7237666FULL, "111678654.ogg"},
	{0xA64E4681D2765BB6ULL, "9739969.ogg"},
	{0xA64ED7C6379953DFULL, "vxdr_0_deeproads_8we_0_1.vlm"},
	{0xA64EE03238103236ULL, "vxw_encounter3_38_0_0.vlm"},
	{0xA64F29CC565D3899ULL, "381470066.ogg"},
	{0xA64F561F75807B79ULL, "qun110_petrice6037422.cl"},
	{0xA64F65F53D83A4CAULL, "hlac_ht_mansion2b_818_51x.dds"},
	{0xA64FAF3D0D626E4BULL, "qun120cr_javaris6044004.cl"},
	{0xA64FC72DBD35D372ULL, "vxlt_0_lowtown_docks_b8x_0_0.vlm"},
	{0xA64FE49050100496ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.dura.ani"},
	{0xA64FFBEDE1441213ULL, "gen00fl_aveline6122771.cl"},
	{0xA6502C46E3B8CB98ULL, "follower_banter6168545.cl"},
	{0xA6509CBA3BB3F24EULL, "vxdr_primevaltomb_846_0_0.vlm"},
	{0xA650EFB8BEC087D4ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse4naked_orsinoadefiance.ani"},
	{0xA65118F3B30E8904ULL, "follower_banter6163175.cl"},
	{0xA651D256A3F9226AULL, "6113100_m.fxe"},
	{0xA651D69D773652F5ULL, "vxht_0_hightown_b8b_0_0.vlm"},
	{0xA65243AE3FB663E6ULL, "art\\vfx\\combat\\vfx_str_app_c.mmh"},
	{0xA6525EE705253ACEULL, "crt_cycle_3.dds"},
	{0xA652A6015385C952ULL, "vxht_0_hightown_a3_aok_0_1.vlm"},
	{0xA652DCDADFD68960ULL, "6209855_f.fxe"},
	{0xA653109A7EB8B724ULL, "6201294_m.fxe"},
	{0xA65334302B1FD77FULL, "fxa_cen_aoe_ring.mao"},
	{0xA653D8CE63CBBE13ULL, "hlac_lt_0_lowtown_docks_-3_1x.dds"},
	{0xA653F4FC3A37E07AULL, "vxlt_0_lowtown_a2_gwi_0_0.vlm"},
	{0xA6544C2CA75FA4C4ULL, "171741083.ogg"},
	{0xA65472BACF1C1288ULL, "mcr310_argument6127680.cl"},
	{0xA654A1A4B0F381FFULL, "dae000_gamlen6026217.cl"},
	{0xA654AA3F50350CF3ULL, "hlac_ht_0_hightown_-1_1y.dds"},
	{0xA654CD0D49C5A608ULL, "dae000_gamlen6131105.cl"},
	{0xA6550F63D948020EULL, "art\\characters\\playercharacter\\qunarimale\\uni_ars.mmh"},
	{0xA6552963D9482EFCULL, "art\\characters\\playercharacter\\qunarimale\\uni_ars.msh"},
	{0xA65546A7C366D100ULL, "368003974.ogg"},
	{0xA655DFE161E7BB67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_right.evt"},
	{0xA655FE9F4D7D4548ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier4_epilogue.ani"},
	{0xA656152CBA86B41AULL, "vxlt_shop_g7_0_0.vlm"},
	{0xA656649822C63C64ULL, "hlac_lt_gallowstemplar_f_2388_4x.dds"},
	{0xA656F883A1028FE4ULL, "225517709.ogg"},
	{0xA6571342742ADACCULL, "hlac_lt_0_lowtown_docks_a3_65534_0x.dds"},
	{0xA65721C751C8EDA4ULL, "follower_banter6149166.cl"},
	{0xA65736785E25A34FULL, "vxht_mansion_n_597_0_0.vlm"},
	{0xA657ECB8BEC6A136ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse4naked_orsinoadefiance.gad"},
	{0xA658663FE1C90CA6ULL, "dae000_mother6099208.cl"},
	{0xA6589C71C2B4E8CFULL, "6140651_m.fxe"},
	{0xA658B4B299267390ULL, "hlac_ht_0_hightown_a3_-65537_1x.dds"},
	{0xA658C5412AAD383AULL, "123616797.ogg"},
	{0xA658D7458644381CULL, "plt_isa200pt_opening.nss"},
	{0xA659EA87FAC74E1FULL, "6116757_m.fxe"},
	{0xA65A00391374770AULL, "6120418_m.fxe"},
	{0xA65A0AD4DD4C1791ULL, "narrative_banter6136815.cl"},
	{0xA65A2340E5FC7FDFULL, "6081361_m.fxe"},
	{0xA65A3441A46E085FULL, "vxca_caverns1_mines_ij8_0_0.vlm"},
	{0xA65A5F71B518144CULL, "andcr_templar_captain.utc"},
	{0xA65A91422EA1D296ULL, "6035373_f.fxe"},
	{0xA65B1F208A9409D7ULL, "6041071_m.fxe"},
	{0xA65B2F269958AC6FULL, "hlac_lt_0_lowtown_a2_-262144_2z.dds"},
	{0xA65B632425C17350ULL, "gen00fl_anders6119490.cl"},
	{0xA65BC2076C87F369ULL, "vxlt_gallowscourtyard_4v7_0_0.vlm"},
	{0xA65C0EFEF91DAA18ULL, "6105660_m.fxe"},
	{0xA65C42DF9F6F2E08ULL, "art\\vfx\\cinematic\\vfx_d_aoe_shockwave_p.anb"},
	{0xA65CDAE6C2D1FFA1ULL, "vxht_keep_n_6hq_0_0.vlm"},
	{0xA65D00E161EE11B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_right.gad"},
	{0xA65D06552BDB39A6ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgta.phy"},
	{0xA65D076A1F76D903ULL, "311812478.ogg"},
	{0xA65D3A57424CE849ULL, "247838007.ogg"},
	{0xA65D60182AED5CC0ULL, "vxlt_hangedman_n_1hu_0_0.vlm"},
	{0xA65DAB0EA241ED81ULL, "4437784.ogg"},
	{0xA65DE0A9D68A2AA4ULL, "follower_banter6163621.cl"},
	{0xA65DE6A3CD2E2FC4ULL, "blood_conj_02.dds"},
	{0xA65DEC90501C4F87ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.dura.evt"},
	{0xA65DFB1105AC34AEULL, "vxht_brothel_n_6en_0_0.vlm"},
	{0xA65E528DE6634BE4ULL, "1048571734.ogg"},
	{0xA65EBA907BF62DA6ULL, "22002626.ogg"},
	{0xA65F0A6D91C71855ULL, "hlac_lt_undercityencounter_-1_15y.dds"},
	{0xA65F723FC18412EBULL, "vxdr_0_deeproads_9p8_0_0.vlm"},
	{0xA65FDD5ADA00E075ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmc.phy"},
	{0xA6608849864F7257ULL, "gen_im_trs_gen_arc_sbw_r1.uti"},
	{0xA661137904A8D12EULL, "vxw_blightlands_7x_0_0.vlm"},
	{0xA6619B45A99ADE76ULL, "349760160.ogg"},
	{0xA661A26CE651638FULL, "344702836.ogg"},
	{0xA661A4B96408D941ULL, "vxw_woundedcoast_17t_0_0.vlm"},
	{0xA661A4F1BB6C19F8ULL, "vxw_blightlands_b1_0_0.vlm"},
	{0xA66266D92C2DF8FAULL, "vxlt_0_lowtown_a2_gjj_0_0.vlm"},
	{0xA66289C1140DF1C6ULL, "qcr220_cut_distraction.cnv"},
	{0xA662CE0583DB72D2ULL, "vxht_0_hightown_car_0_1.vlm"},
	{0xA662EFAB79ED60B1ULL, "hlac_lt_gallowstemplar_n_2392_1y.dds"},
	{0xA662F1FB4D83A4DBULL, "llac_lt_0_lowtown_a2_n_c_-131072_2.dds"},
	{0xA66321A06F48AC64ULL, "lgt000ip_who_she_bookshelf.utp"},
	{0xA66337FC534082CFULL, "mcr310_argument6074376.cl"},
	{0xA66352BD63AB1F5FULL, "6214633_m.fxe"},
	{0xA663E5DAEB79A213ULL, "vxdr_primevalentrance_28e_0_0.vlm"},
	{0xA6645AA17D19763FULL, "art\\characters\\creatures\\spider\\animation\\dmg_v1.gad"},
	{0xA6646E67946590BFULL, "670068411.ogg"},
	{0xA664956C72F0B92DULL, "hlac_w_woundedcoast_327686_0z.dds"},
	{0xA664B4AF683E67EFULL, "hlac_lt_0_lowtown_a2_n_-65536_2y.dds"},
	{0xA664CACEEF705E73ULL, "470310851.ogg"},
	{0xA664DE662C96E0DAULL, "vxlt_0_lowtown_etr_0_0.vlm"},
	{0xA664ECE73728A8D6ULL, "6135210_m.fxe"},
	{0xA6650D905022A6D0ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.dura.gad"},
	{0xA665193C8C6D5C9BULL, "vxht_keep_3yb_0_0.vlm"},
	{0xA66530041E8C324AULL, "866099156.ogg"},
	{0xA665672724E4F2FAULL, "638885756.ogg"},
	{0xA6668BC11411DD86ULL, "qcr220_cut_distraction.bnk"},
	{0xA666D23EE74ACD98ULL, "lgt320_mis_resp_swordhunter6150269.cl"},
	{0xA6670831FEEEC180ULL, "qcr000_arishok6060105.cl"},
	{0xA667177A5CC16FBAULL, "20688909.ogg"},
	{0xA66768AABA7A0507ULL, "one220_yevhen6137026.cl"},
	{0xA667B2CD46CFF9E0ULL, "6177283_m.fxe"},
	{0xA667B73A2743D403ULL, "zz_dae_debug6029085.cl"},
	{0xA66813021E9031F7ULL, "hlac_w_blightlands_196611_1y.dds"},
	{0xA66957D2DDBB5C5BULL, "gen00fl_merrill6114058.cl"},
	{0xA669758431907938ULL, "hlac_lt_0_lowtown_n_0_4z.dds"},
	{0xA6699D5D6F8E6A1AULL, "839128111.ogg"},
	{0xA66A0E4675740001ULL, "gen00fl_isabela6118377.cl"},
	{0xA66A295A5D2B1DCDULL, "lt_0_lowtown_docks_21_0_lc.mmh"},
	{0xA66A78F879675503ULL, "vxlt_0_lowtown_docks_n_b4g_0_1.vlm"},
	{0xA66A79DC347BF208ULL, "652991851.ogg"},
	{0xA66A8057ACA2C45FULL, "vxlt_0_lowtown_n_eqf_0_1.vlm"},
	{0xA66B01C24077FD6BULL, "hla_ht_0_hightown_a3_65536_0y.dds"},
	{0xA66B0E13E5FEC4DBULL, "hlac_lt_0_lowtown_a3_n_c_65533_3y.dds"},
	{0xA66B2247059A7C04ULL, "vxlt_0_lowtown_e8h_0_0.vlm"},
	{0xA66B4FA17D1F822CULL, "art\\characters\\creatures\\spider\\animation\\dmg_v1.evt"},
	{0xA66BF7D75763592AULL, "6043893_f.fxe"},
	{0xA66C329B02BA77DDULL, "6118578_f.fxe"},
	{0xA66C48673B7F2531ULL, "vxw_bonepitdestroyed_a3_168_0_0.vlm"},
	{0xA66C939EBEE8F1A0ULL, "717853696.ogg"},
	{0xA66D840CA3C15013ULL, "6086737_m.fxe"},
	{0xA66DF0C668F1006DULL, "149188125.ogg"},
	{0xA66DFBC3BA5FCF1EULL, "vxht_0_hightown_n_b9m_0_0.vlm"},
	{0xA66E0A1AC457E02EULL, "6167653_m.fxe"},
	{0xA66E1014960C7FB0ULL, "w_bonepitmining_a2_99_0.msh"},
	{0xA66E1E14960C97CAULL, "w_bonepitmining_a2_99_0.mmh"},
	{0xA66E2214960C9E01ULL, "w_bonepitmining_a2_99_0.mao"},
	{0xA66E33C7CE3C0728ULL, "714507565.ogg"},
	{0xA66F277D85738DB9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinstower.msh"},
	{0xA66F3120EFFE51E2ULL, "962809901.ogg"},
	{0xA66F317D85739EA7ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinstower.mmh"},
	{0xA66F357D8573A5F4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinstower.mao"},
	{0xA66F43B4205FB254ULL, "llac_dr_0_deeproads_5842_16.dds"},
	{0xA66F60437A1C43A5ULL, "vxlt_0_lowtown_docks_n_c56_0_0.vlm"},
	{0xA66F7B3965F6C870ULL, "vxht_mansion_n_3w8_0_0.vlm"},
	{0xA66F91651EE1F12BULL, "vxw_bonepitmining_a2_1z4_0_0.vlm"},
	{0xA66FCA47166B905FULL, "6087089_f.fxe"},
	{0xA6700C16240EE7B7ULL, "vxht_0_hightown_n_c51_0_0.vlm"},
	{0xA670169655FEF1C1ULL, "vxw_woundedcoast_12k_0_0.vlm"},
	{0xA6706862423CA49CULL, "gen00fl_isabela6135670.cl"},
	{0xA6707AC34CB65745ULL, "vxlt_0_lowtown_a2_gpc_0_0.vlm"},
	{0xA6707BB6BBB6C819ULL, "art\\characters\\playercharacter\\humanfemale\\glv_dksa.phy"},
	{0xA670906950158170ULL, "hlac_lt_0_lowtown_docks_-131074_7x.dds"},
	{0xA670C80D84136B3EULL, "239372343.ogg"},
	{0xA670DCCD7DA2C8C0ULL, "vxht_mansion_n_3sd_0_0.vlm"},
	{0xA6711F6143DA4786ULL, "hlac_ht_chantry_n_2534_36y.dds"},
	{0xA6721D772C2B80C6ULL, "673024637.ogg"},
	{0xA672D7A49B5D7440ULL, "633080053.ogg"},
	{0xA672D94B0D77B5BEULL, "vxlt_warehouse_n_dx_0_0.vlm"},
	{0xA67350431E8CF95AULL, "vxlt_0_lowtown_eos_0_0.vlm"},
	{0xA673871EAD2CB5CDULL, "1004293987.ogg"},
	{0xA67391400BC470FDULL, "llac_ca_cave_2194_2.dds"},
	{0xA673D3AE1D57C5B3ULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_en.ani"},
	{0xA674209B1144EC5EULL, "6151447_m.fxe"},
	{0xA674EDA3CD411A12ULL, "blood_conj_02.mao"},
	{0xA67504AFE48B215FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_015a_0n.dds"},
	{0xA67528E35A8E6345ULL, "zz_mer_debug6073697.cl"},
	{0xA67532B3BC0146FBULL, "vxdr_0_deeproads_bal_0_0.vlm"},
	{0xA675355E831EB39FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closed.gad"},
	{0xA675C59C46A82D15ULL, "866051032.ogg"},
	{0xA675C5B27293CD26ULL, "154243780.ogg"},
	{0xA6760BED0D12CE50ULL, "hlac_w_bonepitmining_a2_4_0x.dds"},
	{0xA676310A7867AD95ULL, "vxlt_hangedman_n_1jz_0_0.vlm"},
	{0xA6764E98D9D0AAB9ULL, "llac_ht_mansion_n_2995_0.dds"},
	{0xA676CB1421D34CE0ULL, "671124278.ogg"},
	{0xA676DD5E71E1F514ULL, "gen00fl_varric6033600.cl"},
	{0xA6771DDDF32D02BAULL, "477397900.ogg"},
	{0xA6772B26FC8D959CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj02.phy"},
	{0xA6772D952308743CULL, "gencr_gate_slave_l3.utc"},
	{0xA67791559EF12AEBULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.impa.ani"},
	{0xA6781BC4CC25EBB4ULL, "vxht_keep_n_688_0_0.vlm"},
	{0xA67897E09D5E3105ULL, "dae200_saemus6142797.cl"},
	{0xA67937A17D2B9685ULL, "art\\characters\\creatures\\spider\\animation\\dmg_v1.ani"},
	{0xA679473AF353C85EULL, "172088022.ogg"},
	{0xA679498BB84AA6BAULL, "vxw_encounter3_g8_0_0.vlm"},
	{0xA67958666210B605ULL, "hlac_w_encounter3_-65538_0x.dds"},
	{0xA6798FF64932739EULL, "vxdr_primevalentrance_37l_0_0.vlm"},
	{0xA679FF18E8F5E2B0ULL, "gen00fl_isabela6075318.cl"},
	{0xA67A0F1069FCD516ULL, "art\\levels\\races\\darkspawn\\props\\prp_ogreskull_0.pwk"},
	{0xA67A1B9C9ED78142ULL, "622140999.ogg"},
	{0xA67A2E4C816E2A85ULL, "zz_isa_debug6133194.cl"},
	{0xA67A9A177A82FC18ULL, "937662441.ogg"},
	{0xA67AA477F4B0905DULL, "6117478_f.fxe"},
	{0xA67ACD87FD2A089BULL, "6093534_f.fxe"},
	{0xA67AE385FE99A631ULL, "vxw_bonepit_xi_0_0.vlm"},
	{0xA67AE9A2778AAB04ULL, "mcr371cr_fenris_cut.utc"},
	{0xA67B3753B71E8B5AULL, "vxca_cave_1cw_0_0.vlm"},
	{0xA67C0B91E14257CEULL, "vxht_keep_5xv_0_0.vlm"},
	{0xA67C3A9725DC7249ULL, "577670021.ogg"},
	{0xA67CE9D4A613379BULL, "267595294.ogg"},
	{0xA67CFB88F7EA1715ULL, "plt_one100pt_gamlen.plo"},
	{0xA67CFF7781AE8797ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ferelden_jimbrn_da.dds"},
	{0xA67DA7C6B1387DD3ULL, "gen00fl_aveline6139159.cl"},
	{0xA67DD2206C125C25ULL, "vxlt_0_lowtown_docks_n_c05_0_0.vlm"},
	{0xA67E12A9358D9A77ULL, "6149366_m.fxe"},
	{0xA67E203250CBB20AULL, "vxht_0_hightown_a3_c12_0_0.vlm"},
	{0xA67E914E17BAB29BULL, "408084195.ogg"},
	{0xA67ED2E0ECC44DAAULL, "vrc221_bartrand6043174.cl"},
	{0xA67EECA03EAC6FC5ULL, "vxlt_0_lowtown_a2_gwb_0_0.vlm"},
	{0xA67F3CD48AAC98B9ULL, "art\\characters\\weapons\\staff\\w_stf_017.msh"},
	{0xA67F46D48AACA9A7ULL, "art\\characters\\weapons\\staff\\w_stf_017.mmh"},
	{0xA67F4A93101A29FAULL, "vxlt_0_lowtown_a2_gdd_0_0.vlm"},
	{0xA67F4AD48AACB0F4ULL, "art\\characters\\weapons\\staff\\w_stf_017.mao"},
	{0xA680314C3CF1E4D6ULL, "6038588_m.fxe"},
	{0xA68063FB5C7AFA93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_incense.evt"},
	{0xA680AD42A0784020ULL, "572448716.ogg"},
	{0xA68164B32607AAE3ULL, "hlac_lt_undercitydungeon_-1_4z.dds"},
	{0xA6819120C22AF090ULL, "gen00fl_merrill6108630.cl"},
	{0xA6821E56CFA45B62ULL, "vxlt_0_lowtown_a2_n_df2_0_0.vlm"},
	{0xA68290266CA024BBULL, "follower_banter6143284.cl"},
	{0xA682E6620B47A3BFULL, "393346447.ogg"},
	{0xA6830C5E7FB145A2ULL, "ttx_blocker_a_1.msh"},
	{0xA6830E5E7FB14958ULL, "ttx_blocker_a_1.mmh"},
	{0xA6834DE657ABA410ULL, "513015872.ogg"},
	{0xA6837A6752197137ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga105a_0d.dds"},
	{0xA6837CCC461F3E7DULL, "6074521_m.fxe"},
	{0xA683987D14E959EFULL, "777414195.ogg"},
	{0xA6839DAF350F5F5CULL, "6174013_m.fxe"},
	{0xA683C945A7518DBDULL, "6043826_m.fxe"},
	{0xA683F20F5F2A31CAULL, "vxht_0_hightown_a3_alk_0_0.vlm"},
	{0xA684395F53A168C2ULL, "6125155_f.fxe"},
	{0xA68441DF9F907EA3ULL, "art\\vfx\\cinematic\\vfx_d_aoe_shockwave_p.mmh"},
	{0xA68479559EFB8C62ULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.impa.evt"},
	{0xA684EE4DAC2572FBULL, "6117007_f.fxe"},
	{0xA685A76E7176B493ULL, "6108255_m.fxe"},
	{0xA685CA1DED5B69AEULL, "6126852_m.fxe"},
	{0xA685F340CE6E6B47ULL, "6195811_m.fxe"},
	{0xA6863DBA5EAD34F4ULL, "43103858.ogg"},
	{0xA6865CFB5C7F5AD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_incense.gad"},
	{0xA6868DA1BE1C0434ULL, "vxht_keep_n_639_0_0.vlm"},
	{0xA6873E11A1FCB90BULL, "vxw_bonepitdestroyed_a3_27n_0_0.vlm"},
	{0xA687AE442C0875B0ULL, "706124655.ogg"},
	{0xA6880CC85490918FULL, "w_bonepitdestroyed_a3_202_0.phy"},
	{0xA68813B26C26B4D5ULL, "zz_undead_1.cut"},
	{0xA68848D5FD8E40DCULL, "144826571.ogg"},
	{0xA6884D5D6EB1528CULL, "qun200_party_comment6090446.cl"},
	{0xA6888B5C0AC2B48DULL, "gen_im_trs_slv_hvy_cht_r0.uti"},
	{0xA6889305E39AA451ULL, "follower_banter6160652.cl"},
	{0xA688AAAE1D69DC5DULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_en.gad"},
	{0xA689964C78409681ULL, "hlac_lt_0_lowtown_docks_a3_n_196609_0z.dds"},
	{0xA689F0AFDC729C66ULL, "vxw_encounter3_fw_0_0.vlm"},
	{0xA689F97EC2265EAFULL, "hla_ht_0_hightown_n_-131071_0x.dds"},
	{0xA68A125E8330D4E5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.closed.ani"},
	{0xA68A3716C7C748E1ULL, "lgt110_arl_bann_teagan6115590.cl"},
	{0xA68A7FC1C23EBB52ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_deeproadsdetail_n.dds"},
	{0xA68A8747B136AD9CULL, "890728080.ogg"},
	{0xA68A8A281E780690ULL, "llac_w_encounter3_0_12.dds"},
	{0xA68A8DE9AC040626ULL, "763273677.ogg"},
	{0xA68AF00B32F75335ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_2y.dds"},
	{0xA68B1A902CAEE844ULL, "6036302_m.fxe"},
	{0xA68B78559F01A975ULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.impa.gad"},
	{0xA68BABF77785DA7FULL, "hlac_lt_0_lowtown_docks_a3_n_131071_6z.dds"},
	{0xA68BCB9CA31CD566ULL, "vxht_keep_6yb_0_0.vlm"},
	{0xA68C868F45CEAB89ULL, "dae300_alistair_king6134990.cl"},
	{0xA68C8C12727743F7ULL, "170167469.ogg"},
	{0xA68CAACCC21B4AF3ULL, "907865228.ogg"},
	{0xA68CC89475654791ULL, "81630609.ogg"},
	{0xA68CFA5039EB22C1ULL, "vxw_woundedcoast_18i_0_0.vlm"},
	{0xA68D5E7D30A28845ULL, "vxlt_0_lowtown_a2_gja_0_0.vlm"},
	{0xA68D6BFB5C8592B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_incense.ani"},
	{0xA68D8A45A1FB4A15ULL, "168541899.ogg"},
	{0xA68D8E9A29A78548ULL, "vxlt_0_lowtown_a2_gz8_0_0.vlm"},
	{0xA68F1D09D7090502ULL, "hlac_lt_0_lowtown_1_0z.dds"},
	{0xA68F5FD7CD9E579EULL, "719822256.ogg"},
	{0xA690053D0EF2196BULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.cess.evt"},
	{0xA690379ECE79EA4BULL, "lt_0_lowtown_a3_n_c_19_0_lc.mmh"},
	{0xA690555F41783106ULL, "6126742_m.fxe"},
	{0xA6906FCF27325AE7ULL, "vxlt_0_lowtown_n_gkk_0_0.vlm"},
	{0xA6909DC29B44DE68ULL, "hlac_lt_0_lowtown_a2_n_1_7z.dds"},
	{0xA690A4B7B94036EBULL, "6036338_m.fxe"},
	{0xA6911A839453B2B7ULL, "follower_banter6167952.cl"},
	{0xA6912997919A9681ULL, "102812445.ogg"},
	{0xA69159317CCA495CULL, "mrl000_marethari6129188.cl"},
	{0xA69195A7A1968E29ULL, "6110866_f.fxe"},
	{0xA691A3CB1826A1F9ULL, "ave100_donnic6105950.cl"},
	{0xA691BFF7DDE5F164ULL, "212244360.ogg"},
	{0xA6920B4330193D65ULL, "gen00fl_aveline6106933.cl"},
	{0xA6920CD53EDFB6B2ULL, "dae221ip_exit_keep.ncs"},
	{0xA692166D9FED77FBULL, "vxdr_0_deeproads_bwn_0_0.vlm"},
	{0xA6921E53F62C0548ULL, "follower_barks6153885.cl"},
	{0xA69263EC51204A4AULL, "vxht_0_hightown_a3_aid_0_0.vlm"},
	{0xA692C3A6B44D9A67ULL, "ser120_ghyslain6104339.cl"},
	{0xA692D3A31418C8B1ULL, "vxlt_0_lowtown_a3_n_c_e2w_0_0.vlm"},
	{0xA692DF8826822C3AULL, "vxlt_gallowscourtyard_5a4_0_0.vlm"},
	{0xA6931A878FC6D185ULL, "6091876_m.fxe"},
	{0xA693212BE8B41CABULL, "vxdr_primevaltomb_2vw_0_0.vlm"},
	{0xA69356B4CA064FE7ULL, "zz_ave_debug6059248.cl"},
	{0xA69379D2545146F8ULL, "6079286_m.fxe"},
	{0xA693A2E48D455580ULL, "vxlt_0_lowtown_n_cox_0_0.vlm"},
	{0xA693AC8BE5F91793ULL, "pro000cr_guard_archer.utc"},
	{0xA6943343BFFB184EULL, "and231cr_lyrium_smuggler3.utc"},
	{0xA69446069388DE3AULL, "vxht_mansion2_322_0_0.vlm"},
	{0xA694ACA2E0EAB97FULL, "6086359_m.fxe"},
	{0xA694B7A05E7EF3C2ULL, "vxca_caverns2_967_0_0.vlm"},
	{0xA694F9F4BF0C07B7ULL, "6141645_m.fxe"},
	{0xA695630FF38934A8ULL, "vxht_0_hightown_b14_0_0.vlm"},
	{0xA695ABF32B9B980FULL, "566919054.ogg"},
	{0xA696184E91F446F1ULL, "vxht_0_hightown_n_9am_0_0.vlm"},
	{0xA6961E3D0EF6AF2CULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.cess.gad"},
	{0xA69688C3A753BB2CULL, "plt_mrl350pt_pridedemon.plo"},
	{0xA696EF6107A5E77CULL, "fx_impact01.dds"},
	{0xA697068FF3B3C135ULL, "vxht_bartrand_n_63w_0_0.vlm"},
	{0xA697B033913474E3ULL, "mrl000_marethari6129435.cl"},
	{0xA697C73FE285D731ULL, "vxw_bonepit_no_0_0.vlm"},
	{0xA698628CCE68B4E6ULL, "vxw_encounter3_cx_0_0.vlm"},
	{0xA698703946D4190EULL, "art\\characters\\weapons\\spears\\w_spr_001a.phy"},
	{0xA6993CF03ED03DB4ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_gear.phy"},
	{0xA699DD3C90FFAB04ULL, "6127220_m.fxe"},
	{0xA69A35990BB79F33ULL, "vxca_caverns2_7v5_0_1.vlm"},
	{0xA69A3D799512EDE6ULL, "vxht_keep_6bc_0_0.vlm"},
	{0xA69A79E783D85C19ULL, "hlac_ht_0_hightown_-1_0x.dds"},
	{0xA69B1A7D8598B2DCULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinstower.phy"},
	{0xA69B53A2E278683BULL, "262105470.ogg"},
	{0xA69BC8CD4738B50DULL, "139563534.ogg"},
	{0xA69BEFD6B69A718AULL, "vxlt_hangedman_n_4p_0_0.vlm"},
	{0xA69C0C002D8F6122ULL, "vxlt_darktown_58d_0_1.vlm"},
	{0xA69D1D3D0EFCCC2AULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.cess.ani"},
	{0xA69D6DC3A759AC63ULL, "plt_mrl350pt_pridedemon.nss"},
	{0xA69D918B89C5A765ULL, "450198224.ogg"},
	{0xA69DCFD9C0680AA4ULL, "6102892_m.fxe"},
	{0xA69E33E415E945E5ULL, "hlac_lt_0_lowtown_docks_a3_-3_6z.dds"},
	{0xA69E35DCF56B24B6ULL, "vxht_0_hightown_n_b82_0_0.vlm"},
	{0xA69E6BDF871A974BULL, "hlac_ht_keep_3734_0y.dds"},
	{0xA69EC6E374522717ULL, "mrl150_merrill6199322.cl"},
	{0xA69F51DD03C090DDULL, "follower_banter6163756.cl"},
	{0xA6A1F5E17E36DD14ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\choke.ani"},
	{0xA6A21CC33A91ABE7ULL, "hlac_ht_0_hightown_-65537_1x.dds"},
	{0xA6A231EB8AF769BBULL, "gen00fl_isabela_init6076485.cl"},
	{0xA6A25CB95E6009C5ULL, "w_sundermount_49_0.rml"},
	{0xA6A269F2DB6C0E2CULL, "6107033_m.fxe"},
	{0xA6A2CA5DAF1BB337ULL, "vxw_woundedcoast_nt_0_0.vlm"},
	{0xA6A2EA528C4A831FULL, "follower_banter6143396.cl"},
	{0xA6A2F46842B3E77CULL, "vxdr_primevaltomb_7ue_0_1.vlm"},
	{0xA6A309845E149891ULL, "6101283_m.fxe"},
	{0xA6A319F03ED856B1ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_gear.msh"},
	{0xA6A322E11B6B9CFAULL, "vxht_mansion2_1hs_0_1.vlm"},
	{0xA6A327F03ED86ECCULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_gear.mao"},
	{0xA6A333F03ED8836FULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_gear.mmh"},
	{0xA6A34D7BC688128EULL, "follower_banter6170119.cl"},
	{0xA6A4BA5AC4BE55E5ULL, "195511876.ogg"},
	{0xA6A4E425EDFC5E50ULL, "mcr310_sibling6076301.cl"},
	{0xA6A51BD46D45C309ULL, "6151628_m.fxe"},
	{0xA6A59D7EB368BB14ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_glow01.dds"},
	{0xA6A63691F298A584ULL, "vxca_caverns1_i93_0_0.vlm"},
	{0xA6A645285F584C3CULL, "hlac_lt_0_lowtown_a2_-2_2z.dds"},
	{0xA6A65AC4AA7D2533ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_roba_0d.dds"},
	{0xA6A6725F85A3BD1EULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy.ani"},
	{0xA6A6D0FB6E64CA42ULL, "dae000_gamlen6026180.cl"},
	{0xA6A6DB2B95FF11C6ULL, "1030590611.ogg"},
	{0xA6A6F05ADA3D0290ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmc.msh"},
	{0xA6A6FE5ADA3D1AAAULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbmc.mmh"},
	{0xA6A76E60F6942220ULL, "6081212_m.fxe"},
	{0xA6A793DCB83E06EFULL, "170392193.ogg"},
	{0xA6A7CD9012B8B014ULL, "hlac_lt_0_lowtown_-196609_4y.dds"},
	{0xA6A8691B755619BFULL, "vxca_caverns1_icx_0_2.vlm"},
	{0xA6A87910ECE2D48CULL, "961652407.ogg"},
	{0xA6A8D2831FF07A07ULL, "vxlt_0_lowtown_docks_bt5_0_0.vlm"},
	{0xA6A8D9B12FE3CD2AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closed.gad"},
	{0xA6A8F2E17E3CF776ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\choke.gad"},
	{0xA6A91A2B64C1EA7DULL, "290965486.ogg"},
	{0xA6A998E057189B7FULL, "llac_lt_0_lowtown_docks_a3_n_-131074_0.dds"},
	{0xA6A9E13884C2911DULL, "154724646.ogg"},
	{0xA6AA048C7E5F72ADULL, "vxlt_blackemporium_um_0_0.vlm"},
	{0xA6AA1F3F60C90A6AULL, "6104218_m.fxe"},
	{0xA6AA585D087158B2ULL, "6124300_m.fxe"},
	{0xA6AA711100764496ULL, "806891262.ogg"},
	{0xA6AB30600B793B32ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cast_loop.gad"},
	{0xA6AB9826841E74F2ULL, "hlac_w_sundermount_-196613_0z.dds"},
	{0xA6AB9E776E2A4F90ULL, "w_sundermount_237_0.mao"},
	{0xA6ABA2776E2A56A3ULL, "w_sundermount_237_0.mmh"},
	{0xA6ABB0776E2A6EFDULL, "w_sundermount_237_0.msh"},
	{0xA6AC8F527EDE64E5ULL, "vxlt_0_lowtown_euv_0_0.vlm"},
	{0xA6ACAB543F8CEC65ULL, "l_env_wal_gallows_frightgrp.wbd"},
	{0xA6ACCFE829E621BEULL, "hlac_ht_0_hightown_-131074_0z.dds"},
	{0xA6ACDD59F2D15536ULL, "6037105_m.fxe"},
	{0xA6ACE55BAFB08A87ULL, "zz_one_debug6099967.cl"},
	{0xA6AD3E72DCF20726ULL, "hlac_lt_undercityserialkiller_-1_26z.dds"},
	{0xA6AD4B53767FE497ULL, "atl_shared_dxt1_01.dds"},
	{0xA6AEB034251491BAULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bed_n.dds"},
	{0xA6AF2E6F1FE63ECBULL, "64726165.ogg"},
	{0xA6AF8539814007F7ULL, "hlac_ca_cavern1_haunted_24424_2y.dds"},
	{0xA6AFB00C5611CC40ULL, "zz_fen_debug6166023.cl"},
	{0xA6AFE53D2B100807ULL, "vxca_caverns2prc_9nc_0_0.vlm"},
	{0xA6B00DE17E4343D5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\choke.evt"},
	{0xA6B0200EBB061B4BULL, "vxlt_0_lowtown_docks_a3_n_d2a_0_0.vlm"},
	{0xA6B0E845868ECB47ULL, "plt_isa200pt_opening.plo"},
	{0xA6B13C3AA111CBB7ULL, "vxw_woundedcoast_kw_0_0.vlm"},
	{0xA6B13DD7A600E29AULL, "w_bonepitdestroyed_a3_217_0.mmh"},
	{0xA6B141D7A600E9D1ULL, "w_bonepitdestroyed_a3_217_0.mao"},
	{0xA6B14FD7A60101E0ULL, "w_bonepitdestroyed_a3_217_0.msh"},
	{0xA6B17923671C8CACULL, "mcr310cr_circle_highmage.utc"},
	{0xA6B186F347B3DFC5ULL, "6101948_m.fxe"},
	{0xA6B19924458FD649ULL, "vxlt_gallowstemplar_3rf_0_0.vlm"},
	{0xA6B1B6F7CCA4B0B7ULL, "gen00fl_varric6076075.cl"},
	{0xA6B1B97870C97D2AULL, "vxw_woundedcoast_1nx_0_0.vlm"},
	{0xA6B1DA17952B9857ULL, "871642861.ogg"},
	{0xA6B211A566AB5FABULL, "vxlt_0_lowtown_a2_n_c_f5j_0_0.vlm"},
	{0xA6B2465FB2025070ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexb.msh"},
	{0xA6B2545FB202678AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexb.mmh"},
	{0xA6B2585FB2026EC1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobodrapexb.mao"},
	{0xA6B2720DE16D0246ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkd_n.dds"},
	{0xA6B28666EAE68329ULL, "243641975.ogg"},
	{0xA6B35A247D46DAD5ULL, "converse_pressed.cur"},
	{0xA6B35B1D1709D022ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_exit.gad"},
	{0xA6B3DF1090902017ULL, "818899389.ogg"},
	{0xA6B3E1565C370D2EULL, "mcr371_hostile_merrill6082149.cl"},
	{0xA6B3EA49D79FF235ULL, "vxht_bartrand_n_69u_0_0.vlm"},
	{0xA6B3FB5D2F1729ECULL, "follower_banter6164035.cl"},
	{0xA6B407A945E6659BULL, "6066414_m.fxe"},
	{0xA6B4769518A3A94FULL, "hlac_lt_0_lowtown_docks_n_65535_1y.dds"},
	{0xA6B48A5F85B023EFULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy.evt"},
	{0xA6B4D14B49F31DF6ULL, "gen_im_cft_reg_a3.uti"},
	{0xA6B4D43CF2F8A2A3ULL, "dae021_sister_samea_sta6150183.cl"},
	{0xA6B4F0861683DDF5ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh05.phy"},
	{0xA6B519936168BEADULL, "539485296.ogg"},
	{0xA6B525FE12535894ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwound.ani"},
	{0xA6B55B77529E02F6ULL, "rdr301ar_caverns.lst"},
	{0xA6B567F4A578BDA8ULL, "w_bonepitdestroyed_a3_31_0.msh"},
	{0xA6B579F4A578DC39ULL, "w_bonepitdestroyed_a3_31_0.mao"},
	{0xA6B585F4A578F0D2ULL, "w_bonepitdestroyed_a3_31_0.mmh"},
	{0xA6B5C24801855D50ULL, "412264694.ogg"},
	{0xA6B70CA603B08C95ULL, "582819181.ogg"},
	{0xA6B7213378FF1EE6ULL, "vxht_keep_6he_0_0.vlm"},
	{0xA6B8CDC444DFC1B8ULL, "lt_0_lowtown_a2_n_42_0_lc.mmh"},
	{0xA6B90B8D9DC9092BULL, "989277667.ogg"},
	{0xA6B90CE952D828D3ULL, "plt_cod_hst_amell.plo"},
	{0xA6B90F1351FBF3DBULL, "270157313.ogg"},
	{0xA6B91B5084E5CCC6ULL, "996736446.ogg"},
	{0xA6B9CD214CC14A8FULL, "6148743_m.fxe"},
	{0xA6B9F789A15B822AULL, "plt_zz_onept_debug.plo"},
	{0xA6BA4B4823182BECULL, "hlac_lt_darktown_-1_32y.dds"},
	{0xA6BA7E1D17102A19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_exit.evt"},
	{0xA6BA8B5F85B49098ULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy.gad"},
	{0xA6BAA25E8EA12C3FULL, "gen00fl_aveline6121972.cl"},
	{0xA6BADBD3BEF6AA8EULL, "91342629.ogg"},
	{0xA6BAEE88F81F62F2ULL, "plt_one100pt_gamlen.nss"},
	{0xA6BB19303B465308ULL, "vxht_bartrand_n_34b_0_0.vlm"},
	{0xA6BBF1984B17A688ULL, "6171454_m.fxe"},
	{0xA6BC0C18EFF1CA4DULL, "6143506_m.fxe"},
	{0xA6BC22FE125972F6ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwound.gad"},
	{0xA6BC4573B5AC3869ULL, "zz_mag_debug6054594.cl"},
	{0xA6BCB4B12FF43710ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_basicdoor.closed.ani"},
	{0xA6BD2EE99CC24242ULL, "598032571.ogg"},
	{0xA6BD82E944D2D9F1ULL, "530486964.ogg"},
	{0xA6BDCD58302B3B61ULL, "vxlt_0_lowtown_a2_gn1_0_0.vlm"},
	{0xA6BDFC0ECD2C2F0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_wcry.ani"},
	{0xA6BE0D92769CD36EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustainexit.evt"},
	{0xA6BE368DCDBF5C62ULL, "llac_lt_0_lowtown_docks_-131075_2.dds"},
	{0xA6BE50D91CB16C7BULL, "427638326.ogg"},
	{0xA6BE8CDAE22493CEULL, "face_face_p1.vsh.fxo"},
	{0xA6BF1B600B89C0D8ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cast_loop.ani"},
	{0xA6BF694191549820ULL, "288571852.ogg"},
	{0xA6BF6A7752A670E1ULL, "rdr301ar_caverns.are"},
	{0xA6BFBC495B2A402DULL, "514972131.ogg"},
	{0xA6BFDFCC5096E155ULL, "vxdr_slavers_1jb_0_0.vlm"},
	{0xA6C0298F0F47A7A9ULL, "528860563.ogg"},
	{0xA6C02ABF6F2AD26FULL, "vxw_sundermount_23y_0_0.vlm"},
	{0xA6C0441D44BDDE68ULL, "vxht_0_hightown_a2_ay3_0_0.vlm"},
	{0xA6C0A6F161321771ULL, "w_bonepitdestroyed_a3_89_0.msh"},
	{0xA6C0B4F161322E8CULL, "w_bonepitdestroyed_a3_89_0.mao"},
	{0xA6C0C0F16132432FULL, "w_bonepitdestroyed_a3_89_0.mmh"},
	{0xA6C0F3010D944156ULL, "vxht_mansion2b_326_0_0.vlm"},
	{0xA6C1105F63F7CEF5ULL, "llac_ht_keep_3732_16.dds"},
	{0xA6C16DE567D69071ULL, "vxht_0_hightown_n_9pb_0_0.vlm"},
	{0xA6C1B9AE6CD71486ULL, "prop1_s1.psh.fxo"},
	{0xA6C2B78293337DA5ULL, "6082145_f.fxe"},
	{0xA6C2D639E59FEF07ULL, "156382215.ogg"},
	{0xA6C2FCF31AFDDE0DULL, "follower_banter6141697.cl"},
	{0xA6C3359B92256CF7ULL, "zz_vrc_debug6200229.cl"},
	{0xA6C33DFE125FBF55ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwound.evt"},
	{0xA6C4049276A12F19ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustainexit.gad"},
	{0xA6C4646B9AD7F7B7ULL, "gen00fl_varric6118771.cl"},
	{0xA6C4C507D1B36775ULL, "vxdr_primevalentrance_1m7_0_0.vlm"},
	{0xA6C4E6BB0054590FULL, "953673020.ogg"},
	{0xA6C5090ECD32644EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_wcry.gad"},
	{0xA6C52FD7F10E42C2ULL, "lgt110pt_arl_connor.ncs"},
	{0xA6C55CACD052EA66ULL, "qun110_petrice6018561.cl"},
	{0xA6C59EBA492330A5ULL, "ser270_emeric6149376.cl"},
	{0xA6C5B071FA433A8FULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.dura.gad"},
	{0xA6C608633F51D99BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyn_0t.dds"},
	{0xA6C6F985211683D4ULL, "hlac_w_sundermount_-131074_0x.dds"},
	{0xA6C7560DF17FE748ULL, "vxlt_0_lowtown_a2_gn4_0_0.vlm"},
	{0xA6C7861D171AC208ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_behind_back_exit.ani"},
	{0xA6C79F9BB0994B13ULL, "163078277.ogg"},
	{0xA6C7C47DB8DB9FB9ULL, "6213199_m.fxe"},
	{0xA6C7C772396E3592ULL, "6166935_m.fxe"},
	{0xA6C7D2EA78CCC43FULL, "827298647.ogg"},
	{0xA6C7E45FEDF1BA8CULL, "778225841.ogg"},
	{0xA6C83CB95E80A2EEULL, "w_sundermount_49_0.mmh"},
	{0xA6C848B95E80B685ULL, "w_sundermount_49_0.mao"},
	{0xA6C856B95E80CEDCULL, "w_sundermount_49_0.msh"},
	{0xA6C863D86277ECBFULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.mmh"},
	{0xA6C8A562D5090CB2ULL, "6177814_m.fxe"},
	{0xA6C92CB66492288FULL, "622246360.ogg"},
	{0xA6C92D988528E98FULL, "vxht_brothel_n_5i4_0_0.vlm"},
	{0xA6C9DB935969482AULL, "hlac_lt_0_lowtown_docks_131071_3z.dds"},
	{0xA6C9EF1C966251B6ULL, "plt_cod_prc_xenon.plo"},
	{0xA6CA38E9F14EB651ULL, "253392261.ogg"},
	{0xA6CA7A7BEB48CA10ULL, "483312078.ogg"},
	{0xA6CACD27A4EEB83EULL, "568726185.ogg"},
	{0xA6CAF4DB828B237FULL, "376072583.ogg"},
	{0xA6CB159276A76B67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustainexit.ani"},
	{0xA6CB357F227515E5ULL, "one100_solivitus6048133.cl"},
	{0xA6CBC571FA47CA3CULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.dura.evt"},
	{0xA6CBE3943F19A65BULL, "6180978_m.fxe"},
	{0xA6CC040ECD387A2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\tal_wcry.evt"},
	{0xA6CC358D059272FBULL, "939777328.ogg"},
	{0xA6CCA97423F6545BULL, "vxca_caverns1_mines_j2b_0_1.vlm"},
	{0xA6CCC5E5D70BF483ULL, "aa_t_hawke_sky_look.cut"},
	{0xA6CD6D7764C91A06ULL, "6140175_m.fxe"},
	{0xA6CDA40D1A687746ULL, "6164586_m.fxe"},
	{0xA6CE0840FA53367BULL, "zz_mcr_debug6098526.cl"},
	{0xA6CE89611E55AFF1ULL, "follower_banter6166409.cl"},
	{0xA6CE8EDD941EA20DULL, "359839817.ogg"},
	{0xA6CE91D31EC3D112ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masa_0s.dds"},
	{0xA6CEB5FA36B3F6E8ULL, "vxht_0_hightown_a2_at0_0_0.vlm"},
	{0xA6CEEBF657892A8EULL, "hlac_dr_0_deeproads_5845_5y.dds"},
	{0xA6CF3D79801EA134ULL, "topple_pressed.cur"},
	{0xA6CFB46CD63E05CFULL, "6140026_m.fxe"},
	{0xA6D074F0D64D35ECULL, "6130978_m.fxe"},
	{0xA6D09303B58149EFULL, "lt_0_lowtown_docks_n_218_0.mao"},
	{0xA6D09F03B5815D8CULL, "lt_0_lowtown_docks_n_218_0.mmh"},
	{0xA6D0A503B581687EULL, "lt_0_lowtown_docks_n_218_0.msh"},
	{0xA6D0D68CE605DAEBULL, "vxdr_primevaltomb_6pu_0_0.vlm"},
	{0xA6D1FECD5EF9A3A2ULL, "vxlt_0_lowtown_a2_n_f55_0_0.vlm"},
	{0xA6D20B69373EAE66ULL, "dae021_sister_lorena_sta6154074.cl"},
	{0xA6D24175131C1B6DULL, "6016827_f.fxe"},
	{0xA6D27F4D75A8D960ULL, "657433426.ogg"},
	{0xA6D28307AFDF875DULL, "469937887.ogg"},
	{0xA6D2B3764C6107F8ULL, "337480412.ogg"},
	{0xA6D2BE4853E2FEA7ULL, "dae221ar_keep.ncs"},
	{0xA6D30F3C98768E41ULL, "593365361.ogg"},
	{0xA6D38CDA9C539B92ULL, "vxw_bonepit_273_0_0.vlm"},
	{0xA6D409A74860E5D5ULL, "qcr000_arishok6059968.cl"},
	{0xA6D41C6095EF52D0ULL, "gen00fl_anders6095497.cl"},
	{0xA6D48FF6B52BD22CULL, "vxht_chantry_2e4_0_0.vlm"},
	{0xA6D4E63A249E3562ULL, "hlac_lt_gallowstemplar_7198_7z.dds"},
	{0xA6D51CF2391E503DULL, "vrc221_bartrand6065319.cl"},
	{0xA6D59F9168C68931ULL, "6110390_f.fxe"},
	{0xA6D666ECBF407F9DULL, "6133304_f.fxe"},
	{0xA6D6B7AEAB06FA94ULL, "6082158_m.fxe"},
	{0xA6D6D2FC63E80D71ULL, "6166571_f.fxe"},
	{0xA6D701D4A87FEB7FULL, "vxw_sundermount_2z3_0_1.vlm"},
	{0xA6D7248C69E22122ULL, "ef_katriela.mor"},
	{0xA6D79F75771F020FULL, "vxht_brothel_n_5l7_0_0.vlm"},
	{0xA6D7A8F17DB3BE39ULL, "6037577_m.fxe"},
	{0xA6D7D4BDF4B2DA46ULL, "vis121_saemus6099485.cl"},
	{0xA6D7E3FBA61B1F9BULL, "236954417.ogg"},
	{0xA6D9581403E40605ULL, "vxw_sundermount_2j0_0_0.vlm"},
	{0xA6D95EE586C631ADULL, "vxht_mansion2b_268_0_0.vlm"},
	{0xA6D96F39470BAED1ULL, "art\\characters\\weapons\\spears\\w_spr_001a.mmh"},
	{0xA6D97B39470BC372ULL, "art\\characters\\weapons\\spears\\w_spr_001a.mao"},
	{0xA6D98939470BDA8FULL, "art\\characters\\weapons\\spears\\w_spr_001a.msh"},
	{0xA6D99E4853E8E756ULL, "dae221ar_keep.lst"},
	{0xA6D9B2AF739646CEULL, "vxht_brothel_n_5fx_0_0.vlm"},
	{0xA6D9CD71FA541555ULL, "art\\vfx\\combat\\vfx_d_bld_wound_aoe_p.dura.ani"},
	{0xA6D9D23235D76DEFULL, "6121904_m.fxe"},
	{0xA6DB5544727A7778ULL, "ht_0_hightownexplosion_n_13_0.phy"},
	{0xA6DBA58B6A402C32ULL, "gen00fl_aveline6139127.cl"},
	{0xA6DBBEDCBC3A3ACFULL, "901330272.ogg"},
	{0xA6DBC72CCAFA188BULL, "vxdr_primevaltomb_91a_0_0.vlm"},
	{0xA6DC1227C5DAE980ULL, "pro000_cut2d_thefirstyear6005316.cl"},
	{0xA6DC8C354E283CF1ULL, "plt_one200pt_herb_heart.nss"},
	{0xA6DC91D176F41537ULL, "vxw_woundedcoast_zp_0_0.vlm"},
	{0xA6DCC7F575CA0017ULL, "gen00fl_dog6106119.cl"},
	{0xA6DD4D8717067F30ULL, "6032136_f.fxe"},
	{0xA6DDA70ED954CEF0ULL, "dre000_bodahn6031321.cl"},
	{0xA6DE0A80BFE702D7ULL, "one300_zevran6187711.cl"},
	{0xA6DE0D261546EDF7ULL, "333607247.ogg"},
	{0xA6DE2173088CFBA1ULL, "lt_0_lowtown_a3_n_c_9_0.phy"},
	{0xA6DE489ADEA98CECULL, "65356137.ogg"},
	{0xA6DEC34A3C4127CFULL, "6014248_m.fxe"},
	{0xA6DF105A22C9FBAAULL, "zz_spy130cr_bookerthug.utc"},
	{0xA6DF4869D7FBF36BULL, "vxdr_primevaltomb_6wt_0_0.vlm"},
	{0xA6DF60863B2A5A13ULL, "286476482.ogg"},
	{0xA6DF6AB9286366C6ULL, "ser121_jethann6104622.cl"},
	{0xA6DF9A8CC82B9755ULL, "hlac_dr_slavers_3214_10z.dds"},
	{0xA6DFD52EAB15250FULL, "hlac_lt_warehouse_65536_1x.dds"},
	{0xA6E1292E9149379CULL, "lgt210cr_eau_margitte.utc"},
	{0xA6E12ADAB5442D58ULL, "vxht_mansion2_26a_0_0.vlm"},
	{0xA6E15444727EE090ULL, "ht_0_hightownexplosion_n_13_0.rml"},
	{0xA6E186BBF320239AULL, "vxht_mansion2_2jg_0_0.vlm"},
	{0xA6E1DF3EEB81C9D0ULL, "6082232_f.fxe"},
	{0xA6E30E35735F3810ULL, "hlac_ca_cave_2194_6y.dds"},
	{0xA6E3AD4853F15541ULL, "dae221ar_keep.are"},
	{0xA6E420B2DAED7781ULL, "738604360.ogg"},
	{0xA6E45E37CD3C04F2ULL, "250698581.ogg"},
	{0xA6E48C4E6F4853AEULL, "zz_mrl_debug6086554.cl"},
	{0xA6E4B632FE17FAA7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wrba_0n.dds"},
	{0xA6E4F184ED28531CULL, "sse_ss_midclass_thugs_m_1.gda"},
	{0xA6E4F1DD0215D876ULL, "hlac_lt_0_lowtown_docks_a3_65535_1z.dds"},
	{0xA6E583DC9AB50331ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_dead_loop.ani"},
	{0xA6E596B456D468D8ULL, "zz_followers_banter6158328.cl"},
	{0xA6E5BA103AF16D0AULL, "hlac_ht_playershouse_n_855_25z.dds"},
	{0xA6E5E363AE64FF98ULL, "ach_comp_funding.dds"},
	{0xA6E5F49DC55D4B4CULL, "vxht_0_hightown_a2_a2r_0_0.vlm"},
	{0xA6E61ACE6FA9D6B9ULL, "6030477_m.fxe"},
	{0xA6E68FFA0CCD417BULL, "w_encounter2_camp_38_0.mmh"},
	{0xA6E69BFA0CCD5518ULL, "w_encounter2_camp_38_0.mao"},
	{0xA6E6ADFA0CCD7465ULL, "w_encounter2_camp_38_0.msh"},
	{0xA6E6EFD75321DDE8ULL, "vxca_cavern1_haunted_dpp_0_0.vlm"},
	{0xA6E6F621C12100E2ULL, "follower_banter6172782.cl"},
	{0xA6E768CB7C3959B5ULL, "414599605.ogg"},
	{0xA6E7933C953589FBULL, "ht_0_hightownexplosion_n_29_0_ll.mmh"},
	{0xA6E7B2F368ED73DCULL, "vxht_0_hightown_a2_bps_0_0.vlm"},
	{0xA6E82B5B1953A839ULL, "follower_banter6143387.cl"},
	{0xA6E879B25CC5D2ADULL, "6036736_m.fxe"},
	{0xA6E8C449D043F293ULL, "vxht_0_hightown_n_ajn_0_0.vlm"},
	{0xA6E9A03A6F1BD309ULL, "668051726.ogg"},
	{0xA6E9B1389DACDB23ULL, "hla_ht_0_hightown_a3_65537_0z.dds"},
	{0xA6E9E8C635A2FBB2ULL, "hlac_w_bonepitdestroyed_a3_0_4z.dds"},
	{0xA6EA0089A184B521ULL, "plt_zz_onept_debug.nss"},
	{0xA6EA464F2989A1FBULL, "l_env_wal_lowtown_int_tavern_laugh.bnk"},
	{0xA6EB03AE68EA2DB7ULL, "vxw_woundedcoast_ws_0_0.vlm"},
	{0xA6EB4C233D059663ULL, "vxlt_hangedman_s1_0_0.vlm"},
	{0xA6EB536CC8BBA645ULL, "dae000_flemeth6013311.cl"},
	{0xA6EC32EF93AF8B5BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupa.mmh"},
	{0xA6EC3EEF93AF9FF8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupa.mao"},
	{0xA6EC50EF93AFBE45ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupa.msh"},
	{0xA6EC85543FC28F0FULL, "l_env_wal_gallows_frightgrp.bnk"},
	{0xA6EC8EDC9ABB34D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_dead_loop.gad"},
	{0xA6ECC9B231DEC642ULL, "primitives_primitive_p6.psh.fxo.dx11"},
	{0xA6ED153D666687A5ULL, "294642035.ogg"},
	{0xA6ED1D07ABE5BA57ULL, "914406674.ogg"},
	{0xA6ED1E0D08214456ULL, "6048802_m.fxe"},
	{0xA6ED3C8E863DD849ULL, "508389037.ogg"},
	{0xA6EE3C1B7E0F77B7ULL, "6113808_m.fxe"},
	{0xA6EE99199041CD9CULL, "6033135_f.fxe"},
	{0xA6EEB092789A574EULL, "art\\characters\\weapons\\longswords\\w_lsw_022a.phy"},
	{0xA6EEDC90D9A5A47DULL, "149402821.ogg"},
	{0xA6EEDFDA070AB4D8ULL, "179531525.ogg"},
	{0xA6EF49D3B048B50DULL, "6156272_m.fxe"},
	{0xA6EF4AB0B2F3FB1BULL, "6124143_m.fxe"},
	{0xA6EF9D4E9269830AULL, "6166395_m.fxe"},
	{0xA6EFFDE586F39C1AULL, "6050417_f.fxe"},
	{0xA6F035D8361261B4ULL, "vxca_cavern1_haunted_hio_0_0.vlm"},
	{0xA6F0650E17ED2663ULL, "mag100_karras6035299.cl"},
	{0xA6F08B8083BB2FBFULL, "6053263_m.fxe"},
	{0xA6F164D8629AF3FCULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.anb"},
	{0xA6F1A8A4D4414AF5ULL, "gen_im_wep_rog_arc_act0_02.uti"},
	{0xA6F207BCFDB8642CULL, "240555159.ogg"},
	{0xA6F209C6B70F6007ULL, "409478359.ogg"},
	{0xA6F24290B486200AULL, "6111292_m.fxe"},
	{0xA6F2E28AC2FCA7A5ULL, "6110854_m.fxe"},
	{0xA6F32780990AF878ULL, "zz_mag_debug6054478.cl"},
	{0xA6F34AA6A32CB09DULL, "6139504_f.fxe"},
	{0xA6F375869BA7AD92ULL, "mcr371_arthuris.cnv"},
	{0xA6F37E47B35584AAULL, "gen00fl_isabela_init6076969.cl"},
	{0xA6F3CE2FAD9F64D1ULL, "1026211044.ogg"},
	{0xA6F44258BE34DE4DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken8xa_1.mao"},
	{0xA6F4D5C5D4884968ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01a.msh"},
	{0xA6F4EE4C04BBC304ULL, "zz_approval_debug6003953.cl"},
	{0xA6F4F3C5D4887B92ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01a.mmh"},
	{0xA6F555C87848E0E5ULL, "hlac_lt_0_lowtown_a2_n_-1_6y.dds"},
	{0xA6F56A2B85EC16DEULL, "vxw_bonepitdestroyed_a3_mi_0_0.vlm"},
	{0xA6F570B8D9C7797BULL, "mrl350cr_pridedemon.ncs"},
	{0xA6F594C7C9545115ULL, "llac_ca_caverns1_mines_24425_4.dds"},
	{0xA6F5D561632FBC39ULL, "264608177.ogg"},
	{0xA6F5FBE3F6BF1F14ULL, "vxw_sundermount_36u_0_1.vlm"},
	{0xA6F60279A9A3E2E8ULL, "788018891.ogg"},
	{0xA6F624D05AE38C5CULL, "vxht_0_hightown_a2_bup_0_0.vlm"},
	{0xA6F62A34363AD85EULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_infa_0d.dds"},
	{0xA6F65C9955706369ULL, "mag311_nyssa6064837.cl"},
	{0xA6F669EDAD38C982ULL, "vxht_0_hightown_a2_c6r_0_0.vlm"},
	{0xA6F69669578277CEULL, "vxht_brothel_n_5lz_0_0.vlm"},
	{0xA6F6968C0AAD1B89ULL, "126009858.ogg"},
	{0xA6F6B864715BE154ULL, "vxlt_0_lowtown_n_db4_0_0.vlm"},
	{0xA6F6F2E7390ABACFULL, "follower_banter6136374.cl"},
	{0xA6F737AE308C4DD9ULL, "267667810.ogg"},
	{0xA6F755A2E2CFB8BBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headlow.ani"},
	{0xA6F7694952FF8BFCULL, "vxht_mansion2_3bt_0_0.vlm"},
	{0xA6F78E85B085FC42ULL, "dae000_gen_player_turn_in6177844.cl"},
	{0xA6F795762048EBAEULL, "312254432.ogg"},
	{0xA6F797869BABD0EAULL, "mcr371_arthuris.bnk"},
	{0xA6F7B028A745A42AULL, "474480388.ogg"},
	{0xA6F7B1D694DA8D72ULL, "vxw_woundedcoast_or_0_0.vlm"},
	{0xA6F8A034224182ECULL, "hlac_dr_primevaltomb_658_12z.dds"},
	{0xA6F8C311CDDB60B5ULL, "6041271_m.fxe"},
	{0xA6F9470CDD83C5CEULL, "6054634_m.fxe"},
	{0xA6F951813FDD0245ULL, "688434400.ogg"},
	{0xA6F96EABD3CDDCF6ULL, "hlac_ht_0_hightown_n_-131071_0x.dds"},
	{0xA6F9A1DA1B99743BULL, "vxlt_0_lowtown_a2_n_d9b_0_0.vlm"},
	{0xA6F9BF826CEDA384ULL, "663311084.ogg"},
	{0xA6F9BF8F73248786ULL, "art\\vfx\\combat\\vfx_w_gn1_second_wind_c.anb"},
	{0xA6F9C272A91DDCFCULL, "gen00fl_bethany6101394.cl"},
	{0xA6F9DF9D5C1030A8ULL, "vxdr_primevalentrance_1qj_0_0.vlm"},
	{0xA6F9FF01C05018E9ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.mmh"},
	{0xA6FA7D2B25FD910BULL, "vxw_encounter2_camp_32_0_0.vlm"},
	{0xA6FB4A5039F1833AULL, "vxw_encounter3_81_0_0.vlm"},
	{0xA6FB4BEAB749C4BAULL, "w_encounter2_10_0.rml"},
	{0xA6FB8153AE7DE8BBULL, "6106950_m.fxe"},
	{0xA6FBB3D68754E5F8ULL, "835546359.ogg"},
	{0xA6FBBF6528650D21ULL, "vxht_keep_n_61n_0_0.vlm"},
	{0xA6FBDF4ABFC88411ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.ani"},
	{0xA6FC038616C00010ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh05.msh"},
	{0xA6FC118616C0182AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh05.mmh"},
	{0xA6FC6825F4057984ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_loop_c.mmh"},
	{0xA6FD16897B2EC3EAULL, "179710249.ogg"},
	{0xA6FD182FAC21108DULL, "276361677.ogg"},
	{0xA6FDAEA57FD70A55ULL, "6043595_f.fxe"},
	{0xA6FDB93762479935ULL, "one300_zevran6076945.cl"},
	{0xA6FE33A44BEAFBDAULL, "hlac_lt_0_lowtown_a2_65535_7z.dds"},
	{0xA6FE6A75D70C549AULL, "vxht_mansion2_2pa_0_0.vlm"},
	{0xA6FE8E28CF94F4A6ULL, "one300_zevran6076548.cl"},
	{0xA6FE912408D370C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_exit.gad"},
	{0xA6FE9551E3F07F0BULL, "dae000_bodhan6146795.cl"},
	{0xA6FEDF2F1A216097ULL, "6148651_m.fxe"},
	{0xA6FFB331FB9F2757ULL, "rog_emergbackup.dds"},
	{0xA6FFDEC203265005ULL, "867027533.ogg"},
	{0xA7001F6AD590C86EULL, "vxw_encounter1_at_0_0.vlm"},
	{0xA7004467D79E7820ULL, "gen00fl_aveline6139204.cl"},
	{0xA7008A036BD5D4ECULL, "llac_ht_0_hightown_a2_-2_0.dds"},
	{0xA700A1AFD89FFFB5ULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_0.pwk"},
	{0xA700DC231D9390D1ULL, "540077330.ogg"},
	{0xA701619AAAF1BE16ULL, "vxht_mansion2b_1n0_0_0.vlm"},
	{0xA7018F447D30737DULL, "vxlt_0_lowtown_a2_n_c_gcg_0_0.vlm"},
	{0xA7019733073767BDULL, "vxlt_0_lowtown_docks_bsn_0_0.vlm"},
	{0xA701CF313A4FFBF9ULL, "qcr220cr_bethany_circle.utc"},
	{0xA701F3E326AC6C91ULL, "hlac_lt_0_lowtown_docks_-196608_0x.dds"},
	{0xA7021A02AEADC679ULL, "hlac_w_woundedcoast_-262140_0x.dds"},
	{0xA702251C2EB6F9A6ULL, "vxlt_0_lowtown_a2_n_glo_0_0.vlm"},
	{0xA7022CEAB74FAEAEULL, "w_encounter2_10_0.phy"},
	{0xA7025370AAF2875CULL, "w_blightlands_143_0.mao"},
	{0xA7025F70AAF29BFFULL, "w_blightlands_143_0.mmh"},
	{0xA7026570AAF2A5A1ULL, "w_blightlands_143_0.msh"},
	{0xA702773DF5CFF2B9ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.cess.ani"},
	{0xA70283412F3555C7ULL, "gen00fl_merrill6121133.cl"},
	{0xA7029694E2F070C3ULL, "family_rules_headfeatures.gda"},
	{0xA702EA4ABFCEB5B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.gad"},
	{0xA7039C813B0CCFF7ULL, "6142037_m.fxe"},
	{0xA703D391370E0EE8ULL, "vxca_cavern1_haunted_dnn_0_0.vlm"},
	{0xA703DED7FE3134A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_reaction.ani"},
	{0xA7045DA2E2DA50B2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headlow.evt"},
	{0xA7048A5D450B8062ULL, "63547182.ogg"},
	{0xA70526CB23E60C84ULL, "vxlt_hangedman_n_1vs_0_0.vlm"},
	{0xA70532E2334D94CDULL, "vxht_bartrand_n_6qu_0_0.vlm"},
	{0xA70623B386D0A5F2ULL, "vxw_woundedcoast_ds_0_0.vlm"},
	{0xA7063552AF3E34A5ULL, "seb321_leliana6180920.cl"},
	{0xA70678A9CCE121C6ULL, "hlac_lt_gallowstemplar_f_2388_4z.dds"},
	{0xA7067CCDC5022321ULL, "436691381.ogg"},
	{0xA706DAA0C9992D5AULL, "vxht_mansion_n_6o7_0_0.vlm"},
	{0xA7072CDCAA2D408BULL, "6024566_m.fxe"},
	{0xA707BCE045255FB3ULL, "str_slimebone_n.dds"},
	{0xA707C28F7330CA31ULL, "art\\vfx\\combat\\vfx_w_gn1_second_wind_c.mmh"},
	{0xA707E2ED6115E247ULL, "980919821.ogg"},
	{0xA707EDC9E9537169ULL, "6096002_m.fxe"},
	{0xA708D6A94D218766ULL, "amb_scene_couplesex.bnk"},
	{0xA708F4988E22C5F9ULL, "vxlt_0_lowtown_gpc_0_0.vlm"},
	{0xA709523DF5D5D2FBULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.cess.gad"},
	{0xA70952F53784E17EULL, "6118721_m.fxe"},
	{0xA7098B288ACADAFAULL, "300275348.ogg"},
	{0xA709BC2D2BE79BBAULL, "vxw_encounter3_52_0_0.vlm"},
	{0xA709E74ABFD4CF18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_enter.evt"},
	{0xA70A2DCBE137E0CDULL, "6214653_m.fxe"},
	{0xA70A71E241AF9030ULL, "vxw_encounter2_camp_j6_0_0.vlm"},
	{0xA70AC2A93E5AD479ULL, "6026243_m.fxe"},
	{0xA70B5D7B1C951653ULL, "hlac_ht_0_hightown_-131071_0x.dds"},
	{0xA70B6E85FF157FDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v5.gad"},
	{0xA70B7CA2E2E0A445ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headlow.gad"},
	{0xA70B826FE9BEF348ULL, "gen_im_trs_dwv_junk4_04.uti"},
	{0xA70C9482A26EC819ULL, "vxlt_0_lowtown_a2_n_hbu_0_0.vlm"},
	{0xA70C9D68FCF2A4AEULL, "6206910_m.fxe"},
	{0xA70CDFE06BB0E5D8ULL, "zz_neel_test3.cut"},
	{0xA70D2CEB51C5C8AAULL, "6206855_m.fxe"},
	{0xA70D86F2888E902BULL, "mer341_jansen6064238.cl"},
	{0xA70E69D45C31F72EULL, "vxw_blightlands_st_0_0.vlm"},
	{0xA70E9147C786E0EEULL, "vxw_encounter1_fu_0_0.vlm"},
	{0xA70ED81317ADDDEAULL, "hlac_ht_0_hightown_a2_65536_0x.dds"},
	{0xA70F5D2169CF3062ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse2naked_orsinosdefiance.gad"},
	{0xA70F6F3DF5DA6FE0ULL, "art\\vfx\\combat\\vfx_glb_confusion_crust_c.cess.evt"},
	{0xA70FD3779CE7D696ULL, "vxht_mansion2b_1a1_0_0.vlm"},
	{0xA710090FF92D803DULL, "vxlt_0_lowtown_docks_bvm_0_0.vlm"},
	{0xA7102B5F64C6A039ULL, "vxlt_gallowstemplar_34d_0_0.vlm"},
	{0xA7107A180961C700ULL, "6172921_m.fxe"},
	{0xA711AE970B2AC9B2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbench_sa.dds"},
	{0xA711DC7DF4971953ULL, "vxlt_undercityencounter_un_0_0.vlm"},
	{0xA711F6D7FE3D9A9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_reaction.evt"},
	{0xA7126385FF1B8BCCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v5.evt"},
	{0xA71276F79A86D942ULL, "vxlt_0_lowtown_fen_0_0.vlm"},
	{0xA7129381C8E5C3D5ULL, "vfx0_dba.vsh.fxo"},
	{0xA712BC2408E462A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_exit.ani"},
	{0xA712CB3C11BD96B6ULL, "art\\levels\\races\\design\\prp_smalltreasur.phy"},
	{0xA712EC8700467806ULL, "vxlt_blackemporium_kl_0_0.vlm"},
	{0xA712F74DDB4BA76EULL, "mrl351_phantomelves6170348.cl"},
	{0xA713088A3ECFBD5CULL, "vxht_0_hightown_a2_bcv_0_0.vlm"},
	{0xA7135ED30AB4C483ULL, "135732510.ogg"},
	{0xA713663894ED06E5ULL, "6163197_m.fxe"},
	{0xA7137D45DC0969EDULL, "examine_pressed.cur"},
	{0xA713A4BF2543AD4DULL, "vxht_bartrand_n_6zt_0_0.vlm"},
	{0xA714814298120AC5ULL, "mrl150_merrill6078227.cl"},
	{0xA714A3886145EC4FULL, "hlac_lt_foundry_n_320_19x.dds"},
	{0xA714BEAE751C3DA6ULL, "one101ip_at_hightown.utp"},
	{0xA714C576B7307063ULL, "6093887_m.fxe"},
	{0xA714CEB432312587ULL, "vxca_caverns1_k2s_0_0.vlm"},
	{0xA7151E64E4101FC7ULL, "vxht_mansion2_1sb_0_0.vlm"},
	{0xA715F40D152C71FEULL, "vxlt_blackemporium_cl_0_0.vlm"},
	{0xA715FA112658AA9AULL, "6094757_m.fxe"},
	{0xA715FBEDE5789B07ULL, "vxlt_0_lowtown_a3_n_c_eqt_0_0.vlm"},
	{0xA7164A481C8C3E2DULL, "llac_lt_0_lowtown_docks_196609_0.dds"},
	{0xA716939D29E39574ULL, "6172020_m.fxe"},
	{0xA71785F4A5CBBB79ULL, "w_bonepitdestroyed_a3_31_0.rml"},
	{0xA718C579DA5EB197ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_lichen.mao"},
	{0xA718E3BF33A5A8B0ULL, "vxw_encounter2_camp_e7_0_0.vlm"},
	{0xA718F5D7FE43B78DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hugging_kill_reaction.gad"},
	{0xA7192D225F931467ULL, "vxw_encounter2_camp_f9_0_0.vlm"},
	{0xA7193AA3A9731ADCULL, "vxlt_hangedman_2cx_0_0.vlm"},
	{0xA719827816F29E4BULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v3.ani"},
	{0xA7198995DBAB5E9AULL, "hlac_lt_0_lowtown_n_0_4x.dds"},
	{0xA719A3354E5C11DAULL, "plt_one200pt_herb_heart.plo"},
	{0xA719B994BFB40964ULL, "vxca_caverns1_mines_jft_0_0.vlm"},
	{0xA71A0A99E5490333ULL, "vxht_brothel_n_5kf_0_0.vlm"},
	{0xA71A156A9EC4A6E9ULL, "356213878.ogg"},
	{0xA71B2840B9DDEA55ULL, "6135073_m.fxe"},
	{0xA71B5D33B5EBB7D2ULL, "vxlt_undercityencounter_uk_0_0.vlm"},
	{0xA71B8B342390C0AAULL, "plt_ser221pt_remains.plo"},
	{0xA71BD8862D395A5FULL, "vxlt_warehouse_vv_0_0.vlm"},
	{0xA71C628B95EF8784ULL, "777776221.ogg"},
	{0xA71C76316681882FULL, "vxht_0_hightown_a2_c49_0_0.vlm"},
	{0xA71C82DFA21F1F49ULL, "vxca_caverns2_62g_0_0.vlm"},
	{0xA71CB43C11C5C457ULL, "art\\levels\\races\\design\\prp_smalltreasur.msh"},
	{0xA71CB8F0449D0D0FULL, "rdr111dg_martin6055018.cl"},
	{0xA71CC63C11C5E2BAULL, "art\\levels\\races\\design\\prp_smalltreasur.mao"},
	{0xA71CCA3C11C5E9C9ULL, "art\\levels\\races\\design\\prp_smalltreasur.mmh"},
	{0xA71CDBB14E280FAEULL, "vxw_blightlands_tu_0_0.vlm"},
	{0xA71D26F2BA1A28DCULL, "dae000ip_crowd_mf1_drunk1.utp"},
	{0xA71D75C7ADA08CB9ULL, "6026935_m.fxe"},
	{0xA71D7B5197684194ULL, "558137102.ogg"},
	{0xA71DFDF16069CFAEULL, "vxw_sundermount_2gh_0_0.vlm"},
	{0xA71E21B453712A0EULL, "dae000_quintus6154712.cl"},
	{0xA71E34811F68F9A5ULL, "genip_enchantments.ncs"},
	{0xA71E72FE611CA47DULL, "vxlt_0_lowtown_a2_n_c_gye_0_0.vlm"},
	{0xA71E7354DF5D4E4EULL, "demosplashscreen.gfx"},
	{0xA71E806C5B66C0E7ULL, "dae000cr_grand_cleric6178462.cl"},
	{0xA71EEBF4FB57DF80ULL, "6175821_m.fxe"},
	{0xA71F2DF17FA97875ULL, "t3_eye_dbr.tnt"},
	{0xA71F4B85FF25ED25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v5.ani"},
	{0xA71F8FF90972FA45ULL, "vxht_mansion_n_43g_0_1.vlm"},
	{0xA71FDD3F0C303A76ULL, "vxlt_gallowstemplar_5cp_0_0.vlm"},
	{0xA72012D40CC1A89EULL, "94865828.ogg"},
	{0xA720ACE425D88425ULL, "gen00fl_merrill6119608.cl"},
	{0xA720CA743ED811B4ULL, "vxca_caverns1_ck3_0_0.vlm"},
	{0xA7217A6730C5D5DCULL, "vxht_0_hightown_a2_bdw_0_0.vlm"},
	{0xA72185B1BE600F0BULL, "gen00fl_merrill6129915.cl"},
	{0xA72195E2B9F8DF25ULL, "vxca_caverns2_8bm_0_0.vlm"},
	{0xA721E4F0DDE90E17ULL, "890276649.ogg"},
	{0xA721FC01C071651EULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.anb"},
	{0xA721FCEA8CDDA2F5ULL, "vxlt_warehouse_n_14t_0_0.vlm"},
	{0xA7227BB98544F2E2ULL, "335163041.ogg"},
	{0xA722887C384D7B30ULL, "llac_lt_0_lowtown_n_-65537_5.dds"},
	{0xA72294D0F8720995ULL, "208592595.ogg"},
	{0xA722A8E1A35C1B54ULL, "and111_lirene6084749.cl"},
	{0xA7230B77FC0D221BULL, "art\\characters\\playercharacter\\children\\mt_bas_kf.msh"},
	{0xA7232977FC0D5505ULL, "art\\characters\\playercharacter\\children\\mt_bas_kf.mmh"},
	{0xA72368865CE17C6CULL, "vxlt_undercitydungeon_1ij_0_1.vlm"},
	{0xA723882169E02248ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse2naked_orsinosdefiance.ani"},
	{0xA723C0BB73356F6AULL, "428226006.ogg"},
	{0xA723F83DBBA42700ULL, "vxht_keep_n_587_0_0.vlm"},
	{0xA7246B25F426D17BULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_loop_c.anb"},
	{0xA7250444A1DA7245ULL, "hlac_dr_slavers_3214_0y.dds"},
	{0xA7255624C015FC05ULL, "ht_0_hightown_a3_18_0_lc.mmh"},
	{0xA72622DB42362D4EULL, "art\\vfx\\textures\\vfx_cen_splash.dds"},
	{0xA72641A5F7E7A77BULL, "vxlt_undercityserialkiller_1m9_0_0.vlm"},
	{0xA726976AFAFA57E1ULL, "6103586_m.fxe"},
	{0xA7269E5358E838DCULL, "llac_lt_0_lowtown_a2_n_c_131071_3.dds"},
	{0xA726E7ADDF422DDAULL, "hlac_lt_0_lowtown_a2_-131071_0y.dds"},
	{0xA7276A7816FEB2C2ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v3.evt"},
	{0xA7279EFF51892CE7ULL, "vxw_encounter2_camp_a8_0_0.vlm"},
	{0xA727E84E05649537ULL, "vxw_bonepit_1ou_0_0.vlm"},
	{0xA727F1DE15303545ULL, "729490839.ogg"},
	{0xA72831515A2A8EBEULL, "6096478_m.fxe"},
	{0xA728760772E0B3F5ULL, "vxht_0_hightown_a2_9aj_0_0.vlm"},
	{0xA728B7ED2C1E4ED7ULL, "vxca_caverns2_7xr_0_0.vlm"},
	{0xA7291CD165F9FA4CULL, "w_woundedcoast_9_0.msh"},
	{0xA72922D165FA043EULL, "w_woundedcoast_9_0.mmh"},
	{0xA7292EBB97AA91A0ULL, "6055625_f.fxe"},
	{0xA7292ED165FA18D5ULL, "w_woundedcoast_9_0.mao"},
	{0xA72950A35366DB33ULL, "6177312_m.fxe"},
	{0xA729A904CD2FF8F1ULL, "vxht_keep_n_6i6_0_0.vlm"},
	{0xA729F6FA742E0A08ULL, "l_dae321ar_brothel_nt.bnk"},
	{0xA72A82542D2D5FA5ULL, "488505261.ogg"},
	{0xA72A9805BC9388E0ULL, "vxlt_warehouse_n_b0_0_0.vlm"},
	{0xA72A9EF266FD5A5FULL, "fen110_anso6173733.cl"},
	{0xA72AE80E5877A0AFULL, "vxht_0_hightown_a2_c38_0_0.vlm"},
	{0xA72B14C34BC12CB9ULL, "vxdr_dungeon_2hi_0_0.vlm"},
	{0xA72B1674BAF445A3ULL, "gen00fl_aveline6139380.cl"},
	{0xA72B8446C7170B72ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medc_d.dds"},
	{0xA72B99F0D53126C1ULL, "6118708_m.fxe"},
	{0xA72CC170FFF7DF1DULL, "pro000im_merchant_key.uti"},
	{0xA72D561330783FE9ULL, "vxw_encounter1_e3_0_0.vlm"},
	{0xA72D697817031CD5ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v3.gad"},
	{0xA72DB0FA0D097BB0ULL, "w_encounter2_camp_38_0.phy"},
	{0xA72EAF9278D03911ULL, "art\\characters\\weapons\\longswords\\w_lsw_022a.mmh"},
	{0xA72EBB9278D04DB2ULL, "art\\characters\\weapons\\longswords\\w_lsw_022a.mao"},
	{0xA72EC99278D065CFULL, "art\\characters\\weapons\\longswords\\w_lsw_022a.msh"},
	{0xA72F25811F7753FCULL, "genip_enchantments.utp"},
	{0xA72F30D204AC2151ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev3_b.gad"},
	{0xA72F3C5130CE2A34ULL, "vxca_caverns1_cn0_0_0.vlm"},
	{0xA72F563736BCFAB1ULL, "lt_backalley_g_mmi.gff"},
	{0xA72F98BD312608D9ULL, "628071131.ogg"},
	{0xA731223D4665D644ULL, "zz_vault_debug6203394.cl"},
	{0xA731250447CB8FB0ULL, "fen310_fenris_danarius.stg"},
	{0xA731DA814720D9F3ULL, "300465009.ogg"},
	{0xA731F75AAD999C53ULL, "702085474.ogg"},
	{0xA73223A186357970ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw1sm.msh"},
	{0xA7322602310F37B2ULL, "vxht_mansion2_1gf_0_1.vlm"},
	{0xA73231A18635908AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw1sm.mmh"},
	{0xA732344F29C738F9ULL, "l_env_wal_lowtown_int_tavern_laugh.wbd"},
	{0xA73235A1863597C1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw1sm.mao"},
	{0xA7328A6987E0D1E3ULL, "lgt000ip_keg_static.utp"},
	{0xA732AC721B82981CULL, "hlac_lt_backalley_n_65536_2x.dds"},
	{0xA732FE91F512E8CAULL, "vxht_0_hightown_c5w_0_0.vlm"},
	{0xA73367336308DF8FULL, "lgt261ip_treasure.utp"},
	{0xA73372CB760381BAULL, "vxlt_0_lowtown_a2_eu1_0_0.vlm"},
	{0xA733965762A0ECB3ULL, "vxlt_0_lowtown_docks_b6s_0_0.vlm"},
	{0xA733AB4B83D882B5ULL, "630660749.ogg"},
	{0xA733CA041825FC5CULL, "330035752.ogg"},
	{0xA73453EF93ED7890ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_ltrockgroupa.phy"},
	{0xA73466B14D84F287ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.dura.gad"},
	{0xA7348FFA0D0F6208ULL, "w_encounter2_camp_38_0.rml"},
	{0xA73496763154F4F2ULL, "vxw_bonepit_1wt_0_0.vlm"},
	{0xA734F4966A07EFE8ULL, "6074383_f.fxe"},
	{0xA7351BADE03609D7ULL, "vxlt_0_lowtown_docks_a3_cq4_0_0.vlm"},
	{0xA7351FC12611B8BEULL, "6179518_f.fxe"},
	{0xA7352A667D6ABA21ULL, "dae000_arishok6067684.cl"},
	{0xA73531D204B08EFFULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev3_b.ani"},
	{0xA73575CF566A0F99ULL, "vxht_0_hightown_a3_bgb_0_0.vlm"},
	{0xA735BA0D28F82E71ULL, "vxlt_0_lowtown_a2_en3_0_0.vlm"},
	{0xA735CC68B8FDD256ULL, "ico_boots_light_5.dds"},
	{0xA735D7C641F529B4ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw102a.mao"},
	{0xA73640E7B65BBA92ULL, "6163543_m.fxe"},
	{0xA736424472C7020DULL, "ht_0_hightownexplosion_n_13_0.msh"},
	{0xA73649BCC317411DULL, "vxca_caverns1_gex_0_0.vlm"},
	{0xA736504472C71A40ULL, "ht_0_hightownexplosion_n_13_0.mao"},
	{0xA736544472C72093ULL, "ht_0_hightownexplosion_n_13_0.mmh"},
	{0xA7365D258B55DA22ULL, "235685319.ogg"},
	{0xA736D4592EE8F46EULL, "vrc321cr_golem.utc"},
	{0xA73701D1660609FDULL, "w_woundedcoast_9_0.anb"},
	{0xA737AA02C6AD7207ULL, "vxw_bonepitdestroyed_a3_1l8_0_0.vlm"},
	{0xA737B268F021C2B0ULL, "vxlt_0_lowtown_docks_cor_0_0.vlm"},
	{0xA7386300CB569CB7ULL, "tm_hlk_b02.tnt"},
	{0xA738A25E2259BB8BULL, "hlac_lt_0_lowtown_docks_a3_n_196609_0x.dds"},
	{0xA738AAD8A5CE14CFULL, "206777620.ogg"},
	{0xA73909E2AE89A160ULL, "vxlt_warehouse_n_i1_0_0.vlm"},
	{0xA739615A6FDB2188ULL, "follower_banter6163870.cl"},
	{0xA7396553EC8F9747ULL, "6081813_m.fxe"},
	{0xA739B4541C734FB1ULL, "46027855.ogg"},
	{0xA739B8E50724BC4AULL, "character_rendertodepth_p0.vsh.fxo.dx11"},
	{0xA739E1644AE604B5ULL, "qun131_party_comment6117284.cl"},
	{0xA73A9BA88AF7C5B1ULL, "156886348.ogg"},
	{0xA73AE1AB445600AEULL, "vxw_sundermount_2mj_0_0.vlm"},
	{0xA73B0F8D66BDC053ULL, "6169899_f.fxe"},
	{0xA73B52ACEFCCCCECULL, "dae110ip_at_and_shop.utp"},
	{0xA73B5BB14D8AFE94ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.dura.evt"},
	{0xA73B5EA6CF0FC9BDULL, "vxlt_0_lowtown_docks_bgj_0_0.vlm"},
	{0xA73B80861E3C7A22ULL, "qun231_varnell6078101.cl"},
	{0xA73B8FDE9B6EAD57ULL, "933807243.ogg"},
	{0xA73BB98A4DBE9F97ULL, "746365587.ogg"},
	{0xA73BC7F0226E5869ULL, "vxw_encounter1_j2_0_0.vlm"},
	{0xA73BCC72C2AB3C47ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_004a_0s.dds"},
	{0xA73BF4DC022E1497ULL, "dae210tr_gate_guard_on.utt"},
	{0xA73CB2C8D21BB7A4ULL, "vxht_mansion2b_1as_0_0.vlm"},
	{0xA73CC0F8F01C6B76ULL, "vxlt_gallowstemplar_5ir_0_0.vlm"},
	{0xA73D5F904DF43B49ULL, "llac_ca_caverns2prc_5802_34.dds"},
	{0xA73DDCC4F87BFC83ULL, "lt_0_lowtown_docks_253_0.mao"},
	{0xA73DE0C4F87C03D0ULL, "lt_0_lowtown_docks_253_0.mmh"},
	{0xA73DEEC4F87C1BEAULL, "lt_0_lowtown_docks_253_0.msh"},
	{0xA73E274DEA18D673ULL, "seb221_brett_harimann6187784.cl"},
	{0xA73E3A0CF9CC8628ULL, "dre000_varric6027790.cl"},
	{0xA73E91C5D759A960ULL, "ef_marethari.mor"},
	{0xA73F00E1333D62E5ULL, "154326042.ogg"},
	{0xA73F3631C4CB7046ULL, "vxca_caverns2_8le_0_1.vlm"},
	{0xA73F54F48DF4B5B2ULL, "vxlt_0_lowtown_a2_gv6_0_0.vlm"},
	{0xA7401345DC3A94B7ULL, "prty_bethany_armor_2_circle.dds"},
	{0xA740267308DFCEDDULL, "lt_0_lowtown_a3_n_c_9_0.mao"},
	{0xA7402A7308DFD516ULL, "lt_0_lowtown_a3_n_c_9_0.mmh"},
	{0xA740347308DFE614ULL, "lt_0_lowtown_a3_n_c_9_0.msh"},
	{0xA740C5F807C1014BULL, "mcr310_loyalist_mages6076860.cl"},
	{0xA741CAA39088EA9CULL, "dae300_vashothlead6062827.cl"},
	{0xA741E4A867F99A3AULL, "vxlt_0_lowtown_a2_ex2_0_0.vlm"},
	{0xA74222558B8B8C28ULL, "6041234_m.fxe"},
	{0xA7422BEAB7859171ULL, "w_encounter2_10_0.mmh"},
	{0xA74237EAB785A512ULL, "w_encounter2_10_0.mao"},
	{0xA74245EAB785BD2FULL, "w_encounter2_10_0.msh"},
	{0xA74255763BF6F08CULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\closed.gad"},
	{0xA74278BD998ABD71ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_3z.dds"},
	{0xA742B67F980C0EEBULL, "gen00fl_dog6167671.cl"},
	{0xA742EC5E6CE6E871ULL, "vxdr_dungeon_20y_0_0.vlm"},
	{0xA74312B378DF3759ULL, "hlac_dr_primevaltomb_9621_4y.dds"},
	{0xA74315CEE8C4CEE0ULL, "gen00fl_meriill_romance.stg"},
	{0xA7447810A306101AULL, "6066339_fakemeredith_arrive.cut"},
	{0xA7447E44BC13A682ULL, "isa211tr_approval_midstage.utt"},
	{0xA7448F19E21BA209ULL, "147757965.ogg"},
	{0xA745867195C4A719ULL, "vxht_keep_55n_0_0.vlm"},
	{0xA745C8ECBCA0F98FULL, "6111001_f.fxe"},
	{0xA7461955BFBE301BULL, "342143079.ogg"},
	{0xA7462E0724CFFC3BULL, "264746462.ogg"},
	{0xA746775C65976A6CULL, "ts_chantry_arm_props.gda"},
	{0xA74688E094B276DAULL, "mag311_keran6041401.cl"},
	{0xA746909EE34BF33EULL, "6136185_m.fxe"},
	{0xA746D5B280D95D5EULL, "art\\vfx\\textures\\vfx_crack.dds"},
	{0xA746F1C869F61684ULL, "art\\vfx\\combat\\vfx_glb_ele_s_major_c.impa.ani"},
	{0xA7470CC20E5199BEULL, "gen_im_wep_rog_arc_act2_03.uti"},
	{0xA7471A97DE3BD5D9ULL, "6198696_m.fxe"},
	{0xA7474969CAD485A1ULL, "vxlt_undercitydungeon_178_0_0.vlm"},
	{0xA7482B3F2B45CABCULL, "ch1notnt_ds1y.psh.fxo"},
	{0xA74849B538808395ULL, "197206839.ogg"},
	{0xA74854763BFB5A8AULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\closed.ani"},
	{0xA74887A046A1F51BULL, "vxlt_warehouse_n_15y_0_0.vlm"},
	{0xA7495388364C192EULL, "vxw_sundermount_2vk_0_0.vlm"},
	{0xA74953B14D972ECDULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.dura.ani"},
	{0xA7499AEB64C03896ULL, "vxht_mansion2b_1u5_0_0.vlm"},
	{0xA749D8A94D59210CULL, "amb_scene_couplesex.wbd"},
	{0xA74A0F4E62D0018DULL, "29129305.ogg"},
	{0xA74A11AE61807DECULL, "vxlt_gallowstemplar_f_5r6_0_0.vlm"},
	{0xA74A3A4075481B43ULL, "6165086_f.fxe"},
	{0xA74AC2B2C5C3FC1BULL, "dae000_orsino6085434.cl"},
	{0xA74B065E46C46B18ULL, "worldmap.gfx"},
	{0xA74B1D04309A260EULL, "622625577.ogg"},
	{0xA74B943423B9F3A1ULL, "plt_ser221pt_remains.nss"},
	{0xA74CC5CD4B97990BULL, "300885869.ogg"},
	{0xA74CD542297E52E6ULL, "vxht_0_hightown_n_b5y_0_0.vlm"},
	{0xA74CEB798FDB28A5ULL, "vxca_caverns2_8sj_0_0.vlm"},
	{0xA74D1159AB272048ULL, "hlac_ht_0_hightown_a2_-131071_1z.dds"},
	{0xA74D504E9FB3F4C8ULL, "606324667.ogg"},
	{0xA74D8AF06863CE66ULL, "1014348593.ogg"},
	{0xA74DA80EB6C188C6ULL, "vxca_caverns2_8kd_0_1.vlm"},
	{0xA74DEEC869FC30E6ULL, "art\\vfx\\combat\\vfx_glb_ele_s_major_c.impa.gad"},
	{0xA74EE639E61E5B0BULL, "vxdr_primevalentrance_27d_0_0.vlm"},
	{0xA74FE8FA744EC1B2ULL, "l_dae321ar_brothel_nt.wbd"},
	{0xA75046460DD22C67ULL, "251260803.ogg"},
	{0xA7508A6C97CE6F68ULL, "6123369_m.fxe"},
	{0xA750C9214AF205F9ULL, "470012133.ogg"},
	{0xA750E80427432A8BULL, "6133672_f.fxe"},
	{0xA750F2C5D4D6D88DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye01a.phy"},
	{0xA75107CAC97D91CBULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_sarcophagus_n.dds"},
	{0xA7514370A8F0689EULL, "6153852_m.fxe"},
	{0xA7524C61C894BD0CULL, "6016317_m.fxe"},
	{0xA752C86889F2F6C8ULL, "wm_ico_chantry.dds"},
	{0xA7531851DA2C94FDULL, "w_bonepitdestroyed_a3_172_0.phy"},
	{0xA7533162985FD449ULL, "llac_lt_0_lowtown_docks_n_-131073_4.dds"},
	{0xA753E9C86A0093C5ULL, "art\\vfx\\combat\\vfx_glb_ele_s_major_c.impa.evt"},
	{0xA7542F1CDC02933CULL, "949033536.ogg"},
	{0xA7549622D40DF3B0ULL, "vxlt_0_lowtown_docks_cup_0_0.vlm"},
	{0xA755341BF224D123ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatrun.evt"},
	{0xA7554EA258656C7AULL, "6049225_m.fxe"},
	{0xA755B66B9D577C38ULL, "lt_0_lowtown_docks_a3_n_26_0_lc.mmh"},
	{0xA755D540B8DE7F1CULL, "vxlt_0_lowtown_a2_n_emh_0_0.vlm"},
	{0xA755D5592F03A63CULL, "vrc321cr_golem.ncs"},
	{0xA7568755CD63D10FULL, "132457438.ogg"},
	{0xA756A5783DBA1835ULL, "303433043.ogg"},
	{0xA7570EC0D23ED5C3ULL, "hair0.psh.fxo"},
	{0xA757121C7A80DBEFULL, "hlac_w_encounter3_-131071_0x.dds"},
	{0xA75714CE186C41A0ULL, "hlac_w_bonepitmining_a2_327682_0z.dds"},
	{0xA757531CF23D9E14ULL, "dae000_gamlen6102420.cl"},
	{0xA757ACBA39732737ULL, "vxca_caverns1_eap_0_0.vlm"},
	{0xA757C9406CE29D6DULL, "dae000_bodhan6146911.cl"},
	{0xA758CCF519885213ULL, "553188975.ogg"},
	{0xA758E00329292DDEULL, "l_lgt311ar_gang_hideout_lt.bnk"},
	{0xA758FEF710D26545ULL, "hlac_lt_0_lowtown_docks_-65538_9x.dds"},
	{0xA759303491E5B46BULL, "hlac_lt_blackemporium_1055_23z.dds"},
	{0xA75962FBA3605D95ULL, "6016650_f.fxe"},
	{0xA7596A57E6196B90ULL, "and100cr_anders.utc"},
	{0xA75A8129F414D3CEULL, "627527139.ogg"},
	{0xA75ADA4F0E5EEA4EULL, "228481737.ogg"},
	{0xA75B43181573F951ULL, "fen121_cut_fenris_finds.bnk"},
	{0xA75B61046C5EAC95ULL, "752695533.ogg"},
	{0xA75B7DCA167D60D8ULL, "6179928_m.fxe"},
	{0xA75B947137F35292ULL, "hlac_lt_0_lowtown_n_65535_8x.dds"},
	{0xA75BA1C3711BD8B4ULL, "vxw_encounter2_c1_0_0.vlm"},
	{0xA75C0D1BF22AADA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatrun.gad"},
	{0xA75C105EF157A957ULL, "mcr310_sibling6076375.cl"},
	{0xA75C3F47CB8F2122ULL, "ttx_pillar_top_1.mmh"},
	{0xA75C4147CB8F24D8ULL, "ttx_pillar_top_1.msh"},
	{0xA75C6D57D1BE3095ULL, "515472599.ogg"},
	{0xA75C9B6E921708FEULL, "121799018.ogg"},
	{0xA75CCE9A6DA1CD7CULL, "vxlt_0_lowtown_a2_e1s_0_0.vlm"},
	{0xA75CF65BC441E690ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfourglw.mao"},
	{0xA75CFA5BC441EDA3ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfourglw.mmh"},
	{0xA75D085BC44205FDULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfourglw.msh"},
	{0xA75D90799C6B4098ULL, "llac_lt_gallowstemplar_n_2388_6.dds"},
	{0xA75DB57E4662645BULL, "hlac_lt_undercityserialkiller_-1_50z.dds"},
	{0xA75DDBA76E29AC92ULL, "406250803.ogg"},
	{0xA75E25181575FC61ULL, "fen121_cut_fenris_finds.cnv"},
	{0xA75E409348274357ULL, "678166524.ogg"},
	{0xA75E72CD79004833ULL, "vxdr_0_deeproads_cug_0_2.vlm"},
	{0xA75EA67B11043385ULL, "creeper_creepingaura.dds"},
	{0xA75F0A5B4ABCDCBDULL, "vxht_keep_n_4jm_0_0.vlm"},
	{0xA75F501B6AF6E801ULL, "hlac_lt_gallowstemplar_f_2394_0x.dds"},
	{0xA75FB600CA0F9231ULL, "vxdr_primevaltomb_8o4_0_0.vlm"},
	{0xA75FF728E1BE5AF7ULL, "611335824.ogg"},
	{0xA76000EE438B0D0DULL, "6053676_m.fxe"},
	{0xA76054B5DD7A3B5BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutc.msh"},
	{0xA76066B5DD7A59A6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutc.mao"},
	{0xA76072B5DD7A6E45ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpoutc.mmh"},
	{0xA7610E40FF419F54ULL, "411193540.ogg"},
	{0xA761139C345D0A5BULL, "606554557.ogg"},
	{0xA7611563E1AF34EEULL, "vxlt_hangedman_1vn_0_0.vlm"},
	{0xA761380B7A5BB839ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgta.msh"},
	{0xA7613E8DD1EFA6C0ULL, "genip_pp_gle_sustained.utp"},
	{0xA761420B7A5BC927ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgta.mmh"},
	{0xA761460B7A5BD074ULL, "art\\characters\\playercharacter\\elffemale\\glv_lgta.mao"},
	{0xA761CD323BE48CD8ULL, "6070444_f.fxe"},
	{0xA761F306855A65CEULL, "mcr310_argument6074469.cl"},
	{0xA7629103E2689D6DULL, "6142196_m.fxe"},
	{0xA7629E05C4CA1493ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.anb"},
	{0xA762B9612FDC2311ULL, "static_static_p11.psh.fxo.dx11.assembly"},
	{0xA762C1290317BCE7ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.dura.gad"},
	{0xA762E211CF230A93ULL, "vxw_encounter3_gd_0_1.vlm"},
	{0xA762E6CC7DB1DC9CULL, "hlac_lt_0_lowtown_n_-131072_5z.dds"},
	{0xA7631C1BF230E582ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatrun.ani"},
	{0xA763B976A0CD6E97ULL, "6060193_m.fxe"},
	{0xA7640D51DA3AF569ULL, "w_bonepitdestroyed_a3_172_0.mao"},
	{0xA76411A991A86876ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkb.ani"},
	{0xA7641951DA3B0902ULL, "w_bonepitdestroyed_a3_172_0.mmh"},
	{0xA7641B51DA3B0CB8ULL, "w_bonepitdestroyed_a3_172_0.msh"},
	{0xA7648784778E15C8ULL, "6135036_m.fxe"},
	{0xA764A6F30B72CC79ULL, "6188662_m.fxe"},
	{0xA764C8ED4AD4E4CBULL, "dae000_sabina6161499.cl"},
	{0xA764DC8F3D46D854ULL, "cloud_f_w.dds"},
	{0xA7654743FDBE5E63ULL, "6137632_f.fxe"},
	{0xA7656C9760E97645ULL, "6096180_f.fxe"},
	{0xA765DBAD5A9D9CA5ULL, "fen110_anso6094991.cl"},
	{0xA76657E22CAC376AULL, "6208699_m.fxe"},
	{0xA766A3262D7AA254ULL, "hlac_w_bonepitmining_a2_131072_0x.dds"},
	{0xA767427480598D41ULL, "gen_im_arm_cht_nobb_t5.uti"},
	{0xA767C145B083FF96ULL, "standard.cur"},
	{0xA768085FA7FE0124ULL, "vxht_mansion2b_1rv_0_0.vlm"},
	{0xA76843A6784371ABULL, "hlac_ca_cavern1_haunted_24425_0x.dds"},
	{0xA7684D49D53F842AULL, "hlac_ca_caverns1_24423_8z.dds"},
	{0xA7684DEE3A336DC9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_exit.ani"},
	{0xA7687DC74914AE69ULL, "765126187.ogg"},
	{0xA76893CCA8C64854ULL, "782364114.ogg"},
	{0xA7689F89AC3873B0ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_7z.dds"},
	{0xA768A26B5DA3E9F2ULL, "hlac_lt_0_lowtown_n_131072_4x.dds"},
	{0xA768A34C2407B86EULL, "6119182_m.fxe"},
	{0xA768B629031C15F4ULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.dura.evt"},
	{0xA768D74C493EAD78ULL, "467682870.ogg"},
	{0xA768F97180C2114CULL, "797345283.ogg"},
	{0xA76905846916FB31ULL, "vxlt_0_lowtown_gcf_0_0.vlm"},
	{0xA7696789C24041B6ULL, "6097895_m.fxe"},
	{0xA7697236A3A01C5AULL, "6152543_m.fxe"},
	{0xA7698023B9AF1241ULL, "hlac_lt_0_lowtown_docks_n_-131073_3y.dds"},
	{0xA76997B4AC0B0B06ULL, "vxlt_blackemporium_ur_0_0.vlm"},
	{0xA769D8BD76319C3FULL, "6094477_m.fxe"},
	{0xA76A0277FC4915B2ULL, "art\\characters\\playercharacter\\children\\mt_bas_kf.phy"},
	{0xA76A13A06311F134ULL, "vxw_encounter2_n6_0_0.vlm"},
	{0xA76A3AF33D9C5515ULL, "vxca_caverns2prc_aid_0_0.vlm"},
	{0xA76A6BAD1FDED4A8ULL, "184437634.ogg"},
	{0xA76AE7BB685EEEC3ULL, "genip_node_f2.utp"},
	{0xA76B322008D8A1F8ULL, "922873861.ogg"},
	{0xA76B64999ECB7BD4ULL, "vxlt_0_lowtown_a2_n_c_deg_0_0.vlm"},
	{0xA76B867709F993AFULL, "484597798.ogg"},
	{0xA76C025C5D49DAA4ULL, "pro000_captain_wright6047470.cl"},
	{0xA76C03F45DE5D2E3ULL, "qun160_party_comment_26061247.cl"},
	{0xA76C3743D56F2B5AULL, "gen_im_fla_anders_01.uti"},
	{0xA76C5513E1F7E029ULL, "vxlt_0_lowtown_e89_0_0.vlm"},
	{0xA76C80912F454BB4ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_rdy.ani"},
	{0xA76CDF6FEB5E497CULL, "vxw_encounter3_et_0_0.vlm"},
	{0xA76D13CFA7035C99ULL, "hlac_lt_0_lowtown_a2_n_c_65533_1z.dds"},
	{0xA76D226A79B24063ULL, "hlac_lt_0_lowtown_a2_n_65537_1z.dds"},
	{0xA76D4CCDF3C672B0ULL, "vxdr_primevalentrance_2ee_0_0.vlm"},
	{0xA76D5203F67FC01EULL, "follower_banter6170780.cl"},
	{0xA76D68B2F8980BF0ULL, "follower_banter6169850.cl"},
	{0xA76DA994EA43B8B8ULL, "qun110_petrice6037090.cl"},
	{0xA76DC46F74543A88ULL, "1011037737.ogg"},
	{0xA76DD3B441B19259ULL, "vxlt_0_lowtown_a2_n_eu9_0_0.vlm"},
	{0xA76DF981514427C8ULL, "473866796.ogg"},
	{0xA76E29C8ACE061C1ULL, "vxdr_0_deeproads_a24_0_0.vlm"},
	{0xA76E68EE3A38074BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_exit.gad"},
	{0xA76E6B9FA97888CBULL, "6166317_f.fxe"},
	{0xA76E876716F5AF47ULL, "1062696726.ogg"},
	{0xA76E9A726B8B61F3ULL, "follower_banter6155931.cl"},
	{0xA76F059E1B3AFCFCULL, "574524632.ogg"},
	{0xA76F4F640BBEA087ULL, "vxdr_primevaltomb_8v9_0_0.vlm"},
	{0xA76F675C5C7B2292ULL, "6120780_m.fxe"},
	{0xA76FBB05C4D4D0FCULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.mmh"},
	{0xA76FC40033F4811AULL, "808249843.ogg"},
	{0xA76FD5E09EBFD357ULL, "510745771.ogg"},
	{0xA77002D16635D0F5ULL, "w_woundedcoast_9_0.rml"},
	{0xA7702AFA268CD454ULL, "34440605.ogg"},
	{0xA7710268695CA8D6ULL, "zz_mcr_debug6080876.cl"},
	{0xA77119A991B30067ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkb.evt"},
	{0xA7717148B9DAECBFULL, "vxlt_foundry_n_v1_0_0.vlm"},
	{0xA771D1736589DC1BULL, "lgt110_urn_barker6164292.cl"},
	{0xA7723230D3766851ULL, "hlac_ca_cavern1_haunted_24423_7y.dds"},
	{0xA7725E9D902636EEULL, "dae000_madam_lusine6151706.cl"},
	{0xA77287EEA0758CEAULL, "gen00fl_anders6112888.cl"},
	{0xA772C8571B65D878ULL, "vxlt_warehouse_n_aq_0_0.vlm"},
	{0xA77317E76A0B4B3AULL, "vxdr_primevalentrance_334_0_0.vlm"},
	{0xA7737D912F4B6516ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_rdy.gad"},
	{0xA7739926CBEA9855ULL, "vxw_sundermount_314_0_0.vlm"},
	{0xA7742C52BFBFA6CCULL, "vxht_brothel_n_629_0_0.vlm"},
	{0xA77477F7CC23B297ULL, "vxca_caverns1_mines_j5b_0_0.vlm"},
	{0xA7749993CB1703EBULL, "6092167_m.fxe"},
	{0xA774FB6F48D5F86CULL, "6109039_m.fxe"},
	{0xA7752F5A6FB60E52ULL, "dae300_bran6143517.cl"},
	{0xA7755376B1CECE7FULL, "vxht_0_hightown_a2_c5z_0_0.vlm"},
	{0xA77565EE3A3E20B0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_exit.evt"},
	{0xA77567453762C76CULL, "vxlt_gallowstemplar_f_5c5_0_0.vlm"},
	{0xA775AEEFA53E145FULL, "vxlt_blackemporium_15v_0_0.vlm"},
	{0xA7767A3C99F419A4ULL, "vxht_mansion2b_1uw_0_0.vlm"},
	{0xA776AE290328452DULL, "art\\vfx\\combat\\vfx_f_mer_creep_loop_c.dura.ani"},
	{0xA777B28EA22C34E0ULL, "hlac_lt_0_lowtown_n_-196608_2y.dds"},
	{0xA777E24EFFC086EFULL, "vxca_caverns1_gki_0_1.vlm"},
	{0xA7783AA991B956B0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_cwkb.gad"},
	{0xA7789EA2018ED8E5ULL, "gen00fl_sebastian6174938.cl"},
	{0xA778CD48E097C23FULL, "145494723.ogg"},
	{0xA7791BE72D85CCB2ULL, "6127222_m.fxe"},
	{0xA7797BF0C6D8177BULL, "6149685_m.fxe"},
	{0xA77990F3AD249E9DULL, "vxlt_darktown_2rs_0_0.vlm"},
	{0xA77A6F59AA5EFF75ULL, "481109633.ogg"},
	{0xA77A751B57563AD1ULL, "557178523.ogg"},
	{0xA77A98912F51B275ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_rdy.evt"},
	{0xA77AC6F0D3EDF8A9ULL, "vxlt_0_lowtown_e38_0_0.vlm"},
	{0xA77B0835400ED1E9ULL, "gen00fl_aveline6139856.cl"},
	{0xA77B9635BE099734ULL, "6092384_m.fxe"},
	{0xA77BA2A9289E20B1ULL, "vxlt_0_lowtown_a2_f9h_0_1.vlm"},
	{0xA77C6B631AB427F7ULL, "vxlt_hangedman_259_0_0.vlm"},
	{0xA77CCFC6EB236F72ULL, "vxw_bonepit_1lq_0_0.vlm"},
	{0xA77D1FC40CD74A9AULL, "dae000_gamlen6149463.cl"},
	{0xA77D25F957852086ULL, "6099831_m.fxe"},
	{0xA77D3130DC18CC69ULL, "mag110_ws_banter6079651.cl"},
	{0xA77E1FDBE5C9B9DDULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_stone.dds"},
	{0xA77E263CFEE33E2AULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v4_c.anb"},
	{0xA77E3D8859E063B7ULL, "vxlt_0_lowtown_docks_a3_cs2_0_0.vlm"},
	{0xA77EE3ED39892897ULL, "6135619_m.fxe"},
	{0xA77F058208B45E1EULL, "6142450_m.fxe"},
	{0xA77F27A96AE7A295ULL, "236289917.ogg"},
	{0xA77F6E01B814A2B4ULL, "vxlt_undercitydungeon_1mw_0_0.vlm"},
	{0xA77F945CBAE6D991ULL, "151998848.ogg"},
	{0xA7800F591AD73AD8ULL, "vxht_mansion2_2at_0_0.vlm"},
	{0xA78092040CAA815AULL, "vxlt_0_lowtown_a3_n_c_htc_0_0.vlm"},
	{0xA7813928246809CDULL, "dae000_orana6149014.cl"},
	{0xA781B285AC0E5CF8ULL, "6129082_m.fxe"},
	{0xA7822320B1501E5AULL, "6083573_m.fxe"},
	{0xA78228B7E2BA475DULL, "6074518_m.fxe"},
	{0xA78304CF8BE6F05AULL, "hlac_lt_0_lowtown_n_-65537_1x.dds"},
	{0xA7830B70BD0D1C6BULL, "6097267_m.fxe"},
	{0xA7836D4C3ACC10ECULL, "6102139_m.fxe"},
	{0xA7839D2FBD38E4C3ULL, "128407276.ogg"},
	{0xA78409DC026BEE46ULL, "dae210tr_gate_guard_on.ncs"},
	{0xA7844CBF24CD22F4ULL, "261704400.ogg"},
	{0xA784A3A736E770A0ULL, "862789741.ogg"},
	{0xA784BBECD592AE63ULL, "hlac_dr_slavers_3214_1x.dds"},
	{0xA784E628384630A4ULL, "288317524.ogg"},
	{0xA7854133DEF25775ULL, "vxca_caverns2prc_62g_0_1.vlm"},
	{0xA785A2FDB253EC4EULL, "mag311_keran6063760.cl"},
	{0xA78672461BD21CC6ULL, "6095765_m.fxe"},
	{0xA78696E4A3BAF877ULL, "283586641.ogg"},
	{0xA7871EAD21888615ULL, "vxca_caverns2prc_acb_0_0.vlm"},
	{0xA78740249D23E5AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackleft.evt"},
	{0xA7874963733D505CULL, "vxht_brothel_n_46w_0_0.vlm"},
	{0xA7874D5D175DB845ULL, "qcr220_grey_wardens6135924.cl"},
	{0xA787EDCDB8CE2FFBULL, "6142785_m.fxe"},
	{0xA78802D09F1AB71DULL, "vxlt_darktown_2ur_0_0.vlm"},
	{0xA7883AF02932BC39ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.impa.ani"},
	{0xA788B647A4BEE9BAULL, "hlac_w_sundermount_1_0x.dds"},
	{0xA788D281B3AA80C4ULL, "vxht_0_hightown_a3_bjx_0_0.vlm"},
	{0xA788D9D82EDC9042ULL, "vxca_cavern1_haunted_dr0_0_0.vlm"},
	{0xA78986F7B34C01B9ULL, "vxca_caverns2_6sd_0_0.vlm"},
	{0xA789C852F0F18696ULL, "w_encounter3_135_0.msh"},
	{0xA789D252F0F19794ULL, "w_encounter3_135_0.mmh"},
	{0xA789D652F0F19EA7ULL, "w_encounter3_135_0.mao"},
	{0xA78ADD400CAA4077ULL, "vxlt_hangedman_228_0_0.vlm"},
	{0xA78B67255DCC8879ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tabletrough.msh"},
	{0xA78B71255DCC9967ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tabletrough.mmh"},
	{0xA78B75255DCC9FB4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tabletrough.mao"},
	{0xA78B9E1695F4B37CULL, "vxht_keep_6ja_0_0.vlm"},
	{0xA78C2C46AFFA301EULL, "6015800_m.fxe"},
	{0xA78CDDCBBD575262ULL, "111109269.ogg"},
	{0xA78D2B0B7A80DD5CULL, "art\\characters\\playercharacter\\elffemale\\glv_lgta.phy"},
	{0xA78D87B6D761AB7EULL, "vxht_mansion2b_256_0_0.vlm"},
	{0xA78DAB983D6CB274ULL, "6172253_m.fxe"},
	{0xA78DF92EEB4B9014ULL, "vxw_bonepitmining_a2_26t_0_0.vlm"},
	{0xA78E37249D29F559ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackleft.gad"},
	{0xA78E81360CCD5358ULL, "vxht_mansion2_2ju_0_0.vlm"},
	{0xA78EEB4F0549130DULL, "hlac_dr_dungeon_3207_7z.dds"},
	{0xA78F15F029389C7BULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.impa.gad"},
	{0xA78F32F956DE142AULL, "hlac_lt_gallowstemplar_2384_5z.dds"},
	{0xA78F3CFDA0D6E3A4ULL, "hlac_lt_darktown_-1_61y.dds"},
	{0xA78FC0E9C275162CULL, "485836354.ogg"},
	{0xA790BBBEF63ECAB6ULL, "vxlt_undercitydungeon_11y_0_0.vlm"},
	{0xA7911C19CB89C7E9ULL, "124422125.ogg"},
	{0xA791A073C07C0B11ULL, "6133682_m.fxe"},
	{0xA791BCE82B678FA8ULL, "6120056_m.fxe"},
	{0xA791EB8EDCE25F53ULL, "6174071_f.fxe"},
	{0xA79200DB4A70E0EBULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_003a_0s.dds"},
	{0xA7921FB6BFD41670ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexa.mao"},
	{0xA79221AB7B7A3D9AULL, "ser120_ghyslain6104237.cl"},
	{0xA79223B6BFD41C83ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexa.mmh"},
	{0xA79231B6BFD434DDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexa.msh"},
	{0xA7926343FE9F215CULL, "vxlt_0_lowtown_docks_cnm_0_0.vlm"},
	{0xA7929944BC55F65BULL, "isa211tr_approval_midstage.ncs"},
	{0xA79318D8CC788A8EULL, "vxdr_primevaltomb_77l_0_0.vlm"},
	{0xA7933F38A7039A1AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj04.phy"},
	{0xA79345F1AD571CF4ULL, "hlac_lt_gallowstemplar_7198_0y.dds"},
	{0xA7934AF91CE6714CULL, "vxht_0_hightown_a2_avn_0_0.vlm"},
	{0xA793720796DFF09CULL, "fen110_fenris6117947.cl"},
	{0xA793917BCF121B9EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.activetoinactive.evt"},
	{0xA794605F8E5244E3ULL, "vxw_encounter3_g4_0_1.vlm"},
	{0xA7948DE59F78B169ULL, "rdr111dg_martin6054998.cl"},
	{0xA79532F0293D3960ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.impa.evt"},
	{0xA79548249D3030A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkbackleft.ani"},
	{0xA7954CBC8D3C57FBULL, "follower_banter6155607.cl"},
	{0xA7958477ACA1052DULL, "vxw_sundermount_30n_0_1.vlm"},
	{0xA79589A5B8CFD054ULL, "275834328.ogg"},
	{0xA795C5BC58E21C75ULL, "vxlt_backalley_n_1i2_0_0.vlm"},
	{0xA79655ABD05D18BCULL, "vxht_keep_584_0_0.vlm"},
	{0xA7968E133DBD2120ULL, "framebuffereffects_dofblur_p0.psh.fxo"},
	{0xA797974044915D12ULL, "vxlt_0_lowtown_docks_a3_d1r_0_0.vlm"},
	{0xA7981A1F9B95C6E2ULL, "30229820.ogg"},
	{0xA7982F8567E9C6CDULL, "hlac_lt_backalley_g_1_2x.dds"},
	{0xA798429E7A1D706EULL, "679075835.ogg"},
	{0xA79856697A9C4E3CULL, "6167398_m.fxe"},
	{0xA7986947AC92C7BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_45.evt"},
	{0xA7987C1B44C36C7CULL, "553910829.ogg"},
	{0xA798B203295EC3E4ULL, "l_lgt311ar_gang_hideout_lt.wbd"},
	{0xA79914C374961EDEULL, "83805145.ogg"},
	{0xA7991AF597EDE0B6ULL, "vxlt_0_lowtown_a3_n_c_ekf_0_1.vlm"},
	{0xA79980FA9C5F1FE6ULL, "one101_orders_followers6115786.cl"},
	{0xA799A87BCF16AEC9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.activetoinactive.gad"},
	{0xA79A65D5AA35BBA3ULL, "gen00fl_fenris6107164.cl"},
	{0xA79B499A207084C8ULL, "409514667.ogg"},
	{0xA79B87B0951BE2AFULL, "378361535.ogg"},
	{0xA79BA0C7D7AAA3A8ULL, "hlac_lt_0_lowtown_docks_a3_-65535_0x.dds"},
	{0xA79C436BF67B4519ULL, "follower_barks6153312.cl"},
	{0xA79C6A9DB9FB9298ULL, "6064569_m.fxe"},
	{0xA79CFB7293BECED6ULL, "vxca_caverns1_mines_i25_0_0.vlm"},
	{0xA79D03B837E0D1C0ULL, "gen00fl_player6207739.cl"},
	{0xA79D7AB295C25478ULL, "llac_lt_hangedman_n_-1_24.dds"},
	{0xA79E1971BA5DA447ULL, "6027826_m.fxe"},
	{0xA79E5604F419A12FULL, "168301770.ogg"},
	{0xA79E58DDFB895A55ULL, "w_encounter2_camp_53_0.phy"},
	{0xA79E6247AC9727FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_45.gad"},
	{0xA79E7F31D5026D42ULL, "6155865_m.fxe"},
	{0xA79EBCDE90C2BC53ULL, "art\\characters\\creatures\\stonegolem\\animation\\blendtree\\stonegolem_overlay.anb"},
	{0xA79EF66AF33DD16EULL, "vxht_0_hightown_4ap_0_0.vlm"},
	{0xA79FB3FA3FDB359EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sibling_dead_loop.ani"},
	{0xA7A0797BCF1C7D17ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.activetoinactive.ani"},
	{0xA7A112EE7F280115ULL, "gen_im_trs_crc_sta_spi_r1.uti"},
	{0xA7A11FE897CD8A05ULL, "vxlt_warehouse_n_ah_0_0.vlm"},
	{0xA7A18F412C33414FULL, "vxht_keep_n_5as_0_0.vlm"},
	{0xA7A23230D8E57EC9ULL, "6172476_m.fxe"},
	{0xA7A2A87948B4E916ULL, "396459503.ogg"},
	{0xA7A324B6BFE291A8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gentablexa.phy"},
	{0xA7A37565545DCFACULL, "200945226.ogg"},
	{0xA7A4051A5C7AFD8CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoright.msh"},
	{0xA7A40B1A5C7B087EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoright.mmh"},
	{0xA7A4171A5C7B1C15ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoright.mao"},
	{0xA7A419C7C3AA549CULL, "hlac_ht_playershouse_n_855_22y.dds"},
	{0xA7A4321225648EBEULL, "6173696_m.fxe"},
	{0xA7A4EB5CB62C1B9CULL, "vxht_mansion2b_3gw_0_0.vlm"},
	{0xA7A5367F91B8E395ULL, "vxlt_0_lowtown_a3_n_c_hp4_0_0.vlm"},
	{0xA7A55217F4A8CAAFULL, "vxca_cavern1_haunted_ck9_0_0.vlm"},
	{0xA7A56147AC9D44FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_right_45.ani"},
	{0xA7A579DDFB8FB081ULL, "w_encounter2_camp_53_0.rml"},
	{0xA7A5EC3E04F4CC5CULL, "6187907_m.fxe"},
	{0xA7A60141C749DF92ULL, "hlac_lt_0_lowtown_docks_a3_-196607_0y.dds"},
	{0xA7A615A192E61ABCULL, "vxht_0_hightown_a3_auv_0_0.vlm"},
	{0xA7A63BCFAD91FB17ULL, "666411366.ogg"},
	{0xA7A6B86196DEE359ULL, "character_character_p208.psh.fxo.dx11"},
	{0xA7A7293CFF06481DULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v4_c.mmh"},
	{0xA7A736E959AE1038ULL, "vxht_mansion2b_23f_0_0.vlm"},
	{0xA7A75659C8AAD64DULL, "1072140699.ogg"},
	{0xA7A7B17150D0D601ULL, "1027828292.ogg"},
	{0xA7A89411D6BDAFC6ULL, "hlac_lt_0_lowtown_docks_131070_3y.dds"},
	{0xA7A897293BFD1FF7ULL, "6191385_m.fxe"},
	{0xA7A8A9784576A782ULL, "w_encounter3_122_0.msh"},
	{0xA7A8AB784576AB38ULL, "w_encounter3_122_0.mmh"},
	{0xA7A8B7784576BFDBULL, "w_encounter3_122_0.mao"},
	{0xA7A8C8AA3A32CE12ULL, "vxlt_darktown_2vd_0_0.vlm"},
	{0xA7A8F34A1F023E0DULL, "qun221_viscount6095598.cl"},
	{0xA7A9248B97D2A8B6ULL, "6063782_m.fxe"},
	{0xA7A9EDBA7A364A0FULL, "dae000_arishok6078435.cl"},
	{0xA7AA2B1DEEC1A7E5ULL, "vxca_caverns2prc_ad3_0_0.vlm"},
	{0xA7AA3A2238FDF153ULL, "gen_im_npc_dwf_arm_cht_lega.uti"},
	{0xA7AAB2D9990D9D3FULL, "ch2_s2y.psh.fxo"},
	{0xA7AAC836D662F511ULL, "6064533_m.fxe"},
	{0xA7AB371F496067A4ULL, "vxlt_undercityencounter_13q_0_1.vlm"},
	{0xA7AB4786D0B0DF3DULL, "llac_lt_0_lowtown_n_131071_6.dds"},
	{0xA7AB64EFF0B98458ULL, "vxht_mansion2_2ps_0_0.vlm"},
	{0xA7AC0248536B38F0ULL, "1006954799.ogg"},
	{0xA7AC0F2BAAB4ECBDULL, "clf260cr_harley.utc"},
	{0xA7AC40B367AE7C1CULL, "vxlt_gallowstemplar_f_35e_0_0.vlm"},
	{0xA7AC445EC50695ECULL, "one220_yevhen6120874.cl"},
	{0xA7ACC3C67ACC1D62ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_0.pwk"},
	{0xA7ACCE5E3DA87634ULL, "663530429.ogg"},
	{0xA7AD21F3A1070D26ULL, "hlac_w_bonepitdestroyed_a3_458758_0y.dds"},
	{0xA7AD6E5AFFD11C97ULL, "6123726_m.fxe"},
	{0xA7AD7A7130D49046ULL, "docks_male_sailor_mob6176657.cl"},
	{0xA7ADFCF038A73B6FULL, "601959449.ogg"},
	{0xA7AE64924FA10DE3ULL, "ts_kirk_darktown_prog_junk2.gda"},
	{0xA7AE7BFB512EADC5ULL, "580956619.ogg"},
	{0xA7AF40EF95783209ULL, "dae000_gen_player_turn_in.bnk"},
	{0xA7AF46FDE28B525CULL, "vxlt_0_lowtown_docks_ctc_0_0.vlm"},
	{0xA7AF6D7DFE381D88ULL, "vxlt_hangedman_n_lj_0_0.vlm"},
	{0xA7AFE47AC4F4512BULL, "vxlt_undercityserialkiller_1bh_0_0.vlm"},
	{0xA7B0011E1E2959CFULL, "vxht_keep_n_5jr_0_0.vlm"},
	{0xA7B02EB300D2A24CULL, "vxht_0_hightown_a2_all_0_0.vlm"},
	{0xA7B0E50F66143153ULL, "670874591.ogg"},
	{0xA7B122E7CD861159ULL, "1067970513.ogg"},
	{0xA7B23FD32F58BE8DULL, "89751878.ogg"},
	{0xA7B2893D8AD6BACDULL, "vxht_bartrand_n_65i_0_0.vlm"},
	{0xA7B322EF957BE849ULL, "dae000_gen_player_turn_in.cnv"},
	{0xA7B3BE5CA45C3FF0ULL, "vxlt_hangedman_n_25y_0_0.vlm"},
	{0xA7B3C3F4E69EE32FULL, "vxca_cavern1_haunted_cl8_0_0.vlm"},
	{0xA7B3CCFA3FEC0918ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sibling_dead_loop.gad"},
	{0xA7B43F663FF0D57AULL, "vxht_0_hightown_50n_0_0.vlm"},
	{0xA7B43FA05697A529ULL, "die_wispy.mao"},
	{0xA7B4568F169097A8ULL, "249557040.ogg"},
	{0xA7B4679420838BF5ULL, "328801326.ogg"},
	{0xA7B4B3EFE410256CULL, "one100ip_ggt_trap.ncs"},
	{0xA7B4D175C8929C36ULL, "vxca_caverns1_k2o_0_1.vlm"},
	{0xA7B4D2371532FB0CULL, "zz_followers_banter6159525.cl"},
	{0xA7B507F78532B3E0ULL, "hlac_lt_0_lowtown_docks_a3_n_131073_0y.dds"},
	{0xA7B5390A004EE568ULL, "dre000_dougal6131122.cl"},
	{0xA7B5FEAF7BDA11B6ULL, "vxlt_0_lowtown_a3_n_c_eqh_0_1.vlm"},
	{0xA7B695EAABDA06D6ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf.evt"},
	{0xA7B6CE24D69AF12DULL, "zz_isa_debug6133244.cl"},
	{0xA7B7175D77023DF1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkleft.ani"},
	{0xA7B73A87C91C39E1ULL, "6173189_m.fxe"},
	{0xA7B75A255DF1AC9CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tabletrough.phy"},
	{0xA7B769B873BB04FDULL, "200891311.ogg"},
	{0xA7B7C9A640EEE704ULL, "174335805.ogg"},
	{0xA7B8043C4B66ED47ULL, "fx_grass_small.mao"},
	{0xA7B82B50E84E9979ULL, "mag200_feynriel6056916.cl"},
	{0xA7B83C7C0CAE6315ULL, "seb201_lady_harimann6177586.cl"},
	{0xA7B869F8E6228BEFULL, "vxlt_0_lowtown_docks_a3_d0m_0_1.vlm"},
	{0xA7B87131FCD15129ULL, "213684575.ogg"},
	{0xA7B8A01AC131E5A6ULL, "llac_ca_caverns1_24425_2.dds"},
	{0xA7B8AE162B2CB58CULL, "6206105_m.fxe"},
	{0xA7B8C79ACD00EFD1ULL, "369527319.ogg"},
	{0xA7B8D513F1F8F2A5ULL, "follower_banter6143189.cl"},
	{0xA7B91E7339B1F0D1ULL, "zz_isa_debug6133954.cl"},
	{0xA7B948AC5355C0E9ULL, "vxlt_0_lowtown_f4n_0_0.vlm"},
	{0xA7B9C094746CEA2EULL, "dae000ip_crowd_mf6_talk4.utp"},
	{0xA7B9C7D7DF1B79B8ULL, "hlac_lt_gallowscourtyard_-131072_3x.dds"},
	{0xA7B9DF61483479CDULL, "451038026.ogg"},
	{0xA7BAA7EA46891D9FULL, "gen00fl_sebastian6180683.cl"},
	{0xA7BAD92496E13834ULL, "mer340_bonepit_comments6083160.cl"},
	{0xA7BADA75AF2FE59EULL, "6040712_m.fxe"},
	{0xA7BAF5E000C745EBULL, "6169320_m.fxe"},
	{0xA7BB3B4DC228B3C7ULL, "418028411.ogg"},
	{0xA7BC9AE532BB7933ULL, "54023496.ogg"},
	{0xA7BCB76DB8C12DA2ULL, "zz_attack.ncs"},
	{0xA7BCC8E8E0555F86ULL, "452628858.ogg"},
	{0xA7BCE885958B218EULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_c.impa.evt"},
	{0xA7BD6060BBD6ED15ULL, "vxlt_undercitydungeon_1c8_0_0.vlm"},
	{0xA7BD6436A4D5A050ULL, "ts_darkspawn_01.gda"},
	{0xA7BD8CEAABE01661ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf.gad"},
	{0xA7BDD803D06E4256ULL, "vxlt_gallowstemplar_f_379_0_0.vlm"},
	{0xA7BE03176F1CD2D3ULL, "315110167.ogg"},
	{0xA7BE225D77086E93ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkleft.gad"},
	{0xA7BE3E89A6239BA7ULL, "vxht_mansion2_3de_0_0.vlm"},
	{0xA7BF325CEA78F242ULL, "6125196_m.fxe"},
	{0xA7BFE560832F40A9ULL, "966174602.ogg"},
	{0xA7C0A83926D2C019ULL, "853410570.ogg"},
	{0xA7C0B6E6762F824EULL, "mag120_macha6080669.cl"},
	{0xA7C1277FEE69D520ULL, "gen00fl_sebastian6174821.cl"},
	{0xA7C163EE4D8B6AD9ULL, "dae000_keep_bureaucrat_2_st6165386.cl"},
	{0xA7C16ED4B142421CULL, "720599017.ogg"},
	{0xA7C193458DF578F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_arms_crossed_female.anb"},
	{0xA7C19C8B0209AD7EULL, "hlac_lt_0_lowtown_a2_n_c_131073_1x.dds"},
	{0xA7C24C7D29217BC0ULL, "dae020cr_ef_courtesan3.utc"},
	{0xA7C261A9E96EC921ULL, "vxw_encounter2_fw_0_0.vlm"},
	{0xA7C30C73E1B06866ULL, "hlac_lt_gallowstemplar_n_2394_12x.dds"},
	{0xA7C38DEAABE4830FULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_runf.ani"},
	{0xA7C3DF8595913139ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_c.impa.gad"},
	{0xA7C47C03FFB77B1DULL, "54833374.ogg"},
	{0xA7C486165D84CCE9ULL, "hla_ht_0_hightown_a2_-131070_0x.dds"},
	{0xA7C48E6E3097E576ULL, "vxca_caverns2_81s_0_0.vlm"},
	{0xA7C4E1B120EA29D0ULL, "6066501_m.fxe"},
	{0xA7C4E21B04BF0A03ULL, "vxlt_0_lowtown_n_h89_0_0.vlm"},
	{0xA7C51F5D770E88F8ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkleft.evt"},
	{0xA7C5322AB146FCEDULL, "hlac_lt_0_lowtown_a2_n_-2_2x.dds"},
	{0xA7C540DCB8886C7DULL, "6113950_m.fxe"},
	{0xA7C59163870E568FULL, "atl_itemupgrad_dxt5_00.dds"},
	{0xA7C59B8CDFBE70DFULL, "767248605.ogg"},
	{0xA7C632FA1E6E1181ULL, "art\\vfx\\creatures\\fxc_fade_ghost_c.anb"},
	{0xA7C6BA951FAF16C0ULL, "vxht_keep_n_8ay_0_0.vlm"},
	{0xA7C6E20395BEF6A6ULL, "732699091.ogg"},
	{0xA7C742F4EF7EC872ULL, "hlac_w_bonepitmining_a2_262145_0x.dds"},
	{0xA7C76FBDDC5876ACULL, "6167602_m.fxe"},
	{0xA7C77B3764BFD7FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_lp.gad"},
	{0xA7C7BEDCD0BDD1D5ULL, "vxca_caverns1_dpd_0_0.vlm"},
	{0xA7C7FBBB9F91A97BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosed.ani"},
	{0xA7C9B3D749232200ULL, "nat380_sibling6103586.cl"},
	{0xA7CA2540CC62FBEBULL, "vxht_keep_n_3g3_0_0.vlm"},
	{0xA7CA2AD515837417ULL, "llac_ca_cavern1_haunted_24423_6.dds"},
	{0xA7CA3E1814CB4C03ULL, "760819228.ogg"},
	{0xA7CA4C20A6859FDCULL, "585314651.ogg"},
	{0xA7CA5C50C760633CULL, "llac_ht_0_hightown_n_-1_0.dds"},
	{0xA7CA6CCDC2B8BC1DULL, "76305194.ogg"},
	{0xA7CABAD1A94DFD3FULL, "dae000st_grandcleric.stg"},
	{0xA7CAC24A62CC956FULL, "atl_achievemen_dxt5_01.dds"},
	{0xA7CAD85177A402DAULL, "154066606.ogg"},
	{0xA7CAE051E33EE5A8ULL, "vxlt_backalley_n_1eg_0_0.vlm"},
	{0xA7CAF08595976C87ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_c.impa.ani"},
	{0xA7CAF8659C90D684ULL, "6107056_f.fxe"},
	{0xA7CB0C1D9E1E5F0AULL, "ts_kirkwall_prog_junk2.gda"},
	{0xA7CB9B9D4E1085B3ULL, "vxw_sundermount_1za_0_0.vlm"},
	{0xA7CBC4B96106B203ULL, "23487852.ogg"},
	{0xA7CC0ED1D7A1E471ULL, "rob_kwl.tnt"},
	{0xA7CC1916B2620630ULL, "6096915_m.fxe"},
	{0xA7CC1CDFFB8F68D8ULL, "ser121_jethann6107113.cl"},
	{0xA7CC397AC25A1DA7ULL, "rdr111dg_martin6054080.cl"},
	{0xA7CC4B4F255A7ABBULL, "weapon_rendertodepth_p0.vsh.fxo"},
	{0xA7CC808C0CE06D6DULL, "docks_scene_dockworkers6176481.cl"},
	{0xA7CCCE158A4CA26EULL, "vxw_encounter2_camp_5i_0_0.vlm"},
	{0xA7CD7E3764C44811ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_lp.evt"},
	{0xA7CE4C1C6BE99175ULL, "6104451_m.fxe"},
	{0xA7CE62C124E620DFULL, "1003404823.ogg"},
	{0xA7CE74A3522B152EULL, "6105521_m.fxe"},
	{0xA7CE9194A2B8C0CDULL, "llac_lt_undercityencounter_-1_12.dds"},
	{0xA7CEAC345535F969ULL, "6018743_m.fxe"},
	{0xA7CF017519656E4CULL, "hlac_ca_caverns1_24421_0z.dds"},
	{0xA7CF07E9A3833B54ULL, "llac_ca_cavern1_haunted_24425_10.dds"},
	{0xA7CF64969DF9F2C7ULL, "6036235_m.fxe"},
	{0xA7CF84072B10D134ULL, "vxw_blightlands_72_0_0.vlm"},
	{0xA7D00B3C4B7B8A9DULL, "fx_grass_small.dds"},
	{0xA7D05407197FFB95ULL, "6191251_m.fxe"},
	{0xA7D08CA0F9A1C249ULL, "vxw_encounter2_23_0_0.vlm"},
	{0xA7D0A73B1B9169F2ULL, "395226041.ogg"},
	{0xA7D0D080920E2A07ULL, "art\\vfx\\combat\\vfx_d_arc_direct_cast_c.anb"},
	{0xA7D0D386DB64E1A1ULL, "vxw_encounter2_av_0_0.vlm"},
	{0xA7D187517ACA3208ULL, "3764372.ogg"},
	{0xA7D1CBEDED45379DULL, "vxlt_undercityencounter_1fn_0_0.vlm"},
	{0xA7D1D7927321764AULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_maj_c.anb"},
	{0xA7D2446E4317F05CULL, "gen00fl_aveline6139820.cl"},
	{0xA7D24CCE71E6E49CULL, "art\\characters\\creatures\\mubari\\animation\\enter.ani"},
	{0xA7D2EAD63C8D5464ULL, "combat_smr_range.gda"},
	{0xA7D2FDB667C1193AULL, "lgt110_gxa_coa_meeting.cnv"},
	{0xA7D321B0C3F45CA4ULL, "mcr371_emile_de_launcet6087017.cl"},
	{0xA7D376AB459FF39EULL, "ts_beasts_dragons_01.gda"},
	{0xA7D395DF86BF91DBULL, "vxlt_hangedman_we_0_0.vlm"},
	{0xA7D3EC0FD52BC197ULL, "vxlt_warehouse_ak_0_0.vlm"},
	{0xA7D413FA1E7A1AFAULL, "art\\vfx\\creatures\\fxc_fade_ghost_c.mmh"},
	{0xA7D4890ECDAA4784ULL, "zz_crowd_plc_test_2.utp"},
	{0xA7D4E4CF5267D61CULL, "171683011.ogg"},
	{0xA7D50AC365DB8209ULL, "6169586_m.fxe"},
	{0xA7D563AFA9474635ULL, "abilitytype.gda"},
	{0xA7D5E38F8ECFA11AULL, "667916236.ogg"},
	{0xA7D6B4CD50E81DE5ULL, "vxlt_0_lowtown_a3_n_c_hpd_0_0.vlm"},
	{0xA7D6C545CC3BFF70ULL, "6136154_m.fxe"},
	{0xA7D71FB667C53B82ULL, "lgt110_gxa_coa_meeting.bnk"},
	{0xA7D73CDDB71DF341ULL, "qun110_petrice6018739.cl"},
	{0xA7D7D999E4B376EBULL, "6163120_m.fxe"},
	{0xA7D8F6446DEBC111ULL, "vxlt_0_lowtown_a3_n_c_gcj_0_0.vlm"},
	{0xA7D93F77FB6B2754ULL, "6082925_m.fxe"},
	{0xA7D959CE71ED19DEULL, "art\\characters\\creatures\\mubari\\animation\\enter.gad"},
	{0xA7D98C7B96D37379ULL, "vxht_brothel_n_32y_0_0.vlm"},
	{0xA7DB302FA94C3FF0ULL, "mrl000_marethari6129359.cl"},
	{0xA7DB3A119AD69AABULL, "vxlt_undercityserialkiller_1qm_0_0.vlm"},
	{0xA7DB7F8ED7168DE8ULL, "gen00fl_varric6087467.cl"},
	{0xA7DB963764D0AEE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_lp.ani"},
	{0xA7DBABDC800D043EULL, "hlac_lt_0_lowtown_n_65535_6z.dds"},
	{0xA7DBCE8AB1E33C59ULL, "vxlt_0_lowtown_docks_n_bvb_0_0.vlm"},
	{0xA7DBE8A6CC864AD2ULL, "185383914.ogg"},
	{0xA7DBFE6BE69A5DAAULL, "6049819_f.fxe"},
	{0xA7DC0404FC6705E4ULL, "6034332_f.fxe"},
	{0xA7DC22BB9FA29405ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosed.gad"},
	{0xA7DC40B5785F0968ULL, "9069501.ogg"},
	{0xA7DC61788BAFEA96ULL, "mcr371_post_harvester6083679.cl"},
	{0xA7DC813E5CFBC6A0ULL, "one300_nuncio6077259.cl"},
	{0xA7DC9D5982D64AB0ULL, "6114531_f.fxe"},
	{0xA7DCC6E749973BBEULL, "63209884.ogg"},
	{0xA7DD696BB9017D2CULL, "6083917_m.fxe"},
	{0xA7DD975B47FA9535ULL, "6069875_m.fxe"},
	{0xA7DD99CC0899B1A2ULL, "6153162_m.fxe"},
	{0xA7DDB8BF662E05E2ULL, "543113142.ogg"},
	{0xA7DE11927FE36953ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_other.evt"},
	{0xA7DE27A3EE92CDB1ULL, "6082036_m.fxe"},
	{0xA7DE3A9568B3EF20ULL, "6120143_f.fxe"},
	{0xA7DE6492A05E59B0ULL, "vrc321_party_barks6095752.cl"},
	{0xA7DEA1CCBF2E4D29ULL, "t3_wdg_w22.tnt"},
	{0xA7DEFDF359914595ULL, "vxlt_0_lowtown_a3_n_c_hl0_0_0.vlm"},
	{0xA7DF2FC1055C002CULL, "6126865_m.fxe"},
	{0xA7DF4418174D8866ULL, "440104878.ogg"},
	{0xA7DF4563CD5AFA21ULL, "vxw_encounter2_lu_0_0.vlm"},
	{0xA7E022EA7BFB6C01ULL, "vxlt_0_lowtown_docks_cq1_0_0.vlm"},
	{0xA7E02F442C004180ULL, "gen00fl_varric6076449.cl"},
	{0xA7E034CE71F2F9FDULL, "art\\characters\\creatures\\mubari\\animation\\enter.evt"},
	{0xA7E0F95D19C38E4FULL, "darktown_male_lowlife_mob6174412.cl"},
	{0xA7E130EADF9AEEEFULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_light_n.dds"},
	{0xA7E1750856DDC6D7ULL, "56148120.ogg"},
	{0xA7E19414E10986D7ULL, "vxlt_0_lowtown_a2_n_c_emm_0_0.vlm"},
	{0xA7E1A89E6FADF4AEULL, "zz_mrl_debug6184199.cl"},
	{0xA7E1AE24110BDB1DULL, "883437480.ogg"},
	{0xA7E1ECD188DE1AF2ULL, "vxw_bonepit_13x_0_0.vlm"},
	{0xA7E240A23BFC36B2ULL, "dae000_gen_turn_in_46178437.cl"},
	{0xA7E255431B641AB1ULL, "vxca_cavern1_haunted_j5t_0_0.vlm"},
	{0xA7E2B7BA8D96EFD5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_spitoon.phy"},
	{0xA7E2F3F304EEC925ULL, "vxht_0_hightown_a2_4n2_0_0.vlm"},
	{0xA7E3ACCF7FD2A34CULL, "349037251.ogg"},
	{0xA7E44FC1A7D6F696ULL, "vxlt_gallowstemplar_f_64y_0_0.vlm"},
	{0xA7E4AB5F2986D7ADULL, "6120771_f.fxe"},
	{0xA7E4E21A5CB25919ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\4so22_galoright.phy"},
	{0xA7E50A927FE97B94ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_other.gad"},
	{0xA7E56BDDFBC57D70ULL, "w_encounter2_camp_53_0.msh"},
	{0xA7E579DDFBC5948AULL, "w_encounter2_camp_53_0.mmh"},
	{0xA7E57DDDFBC59BC1ULL, "w_encounter2_camp_53_0.mao"},
	{0xA7E5B361A39ABBBFULL, "713883720.ogg"},
	{0xA7E6129B68AEB15FULL, "vxlt_warehouse_8d_0_0.vlm"},
	{0xA7E77F571DB36022ULL, "hlac_w_blightlands_327677_0y.dds"},
	{0xA7E7C03703F52BDAULL, "hlac_ht_keep_4519_17z.dds"},
	{0xA7E7C40BC72B16A8ULL, "vxlt_backalley_n_1ke_0_0.vlm"},
	{0xA7E88A8EEBE3D6DDULL, "vxlt_0_lowtown_a2_n_c_eki_0_0.vlm"},
	{0xA7E8A164939E11B6ULL, "vxca_cavern1_haunted_hua_0_0.vlm"},
	{0xA7E8AF3C0D33B37CULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r180_rf.ani"},
	{0xA7E9131A56108A21ULL, "6168039_m.fxe"},
	{0xA7E9C815DDAC4843ULL, "vxw_encounter1_c7_0_0.vlm"},
	{0xA7E9D1E9023854DAULL, "hlac_lt_0_lowtown_a2_n_65534_3x.dds"},
	{0xA7EAC47B306BBCD2ULL, "6108311_m.fxe"},
	{0xA7EAC789EA197EECULL, "w_woundedcoast_74_0.msh"},
	{0xA7EACD89EA1988DEULL, "w_woundedcoast_74_0.mmh"},
	{0xA7EAD989EA199D75ULL, "w_woundedcoast_74_0.mao"},
	{0xA7EC0BBF188447D7ULL, "vis121_saemus6101332.cl"},
	{0xA7EC19927FEFB472ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_other.ani"},
	{0xA7EC1E396F237E5DULL, "vxht_chantry_n_2b7_0_0.vlm"},
	{0xA7EC2BA1F80CF3F3ULL, "llac_ca_caverns2prc_5802_36.dds"},
	{0xA7EC602A3DB9B418ULL, "6117487_m.fxe"},
	{0xA7EC8B23DFA83C6AULL, "vxht_chantry_n_219_0_0.vlm"},
	{0xA7ED4182A6322DF2ULL, "vxw_woundedcoast_4c_0_0.vlm"},
	{0xA7ED5429C1A6770DULL, "hlac_dr_slavers_3214_8y.dds"},
	{0xA7ED7B2A494275ECULL, "6051738_m.fxe"},
	{0xA7ED816EFCE0EC7DULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf.phy"},
	{0xA7EE148569B05609ULL, "vxlt_gallowscourtyard_6z1_0_0.vlm"},
	{0xA7EECA918D9A5068ULL, "nat320_delilah6156660.cl"},
	{0xA7EF3838A751ED43ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj04.msh"},
	{0xA7EF4638A752049DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hj04.mmh"},
	{0xA7EF5B735E9409ACULL, "static_static_p81.psh.fxo.dx11"},
	{0xA7EF6A2DAC2D3089ULL, "fen110_fenris6101958.cl"},
	{0xA7EF72DE1162CC0EULL, "dre000_varric6035126.cl"},
	{0xA7EFBC3C0D39E7BEULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r180_rf.gad"},
	{0xA7F00D0318001685ULL, "hlac_w_bonepitmining_a2_458756_0x.dds"},
	{0xA7F05EAE7AD43372ULL, "vxw_bonepit_14y_0_0.vlm"},
	{0xA7F0728E286B064FULL, "967641662.ogg"},
	{0xA7F0CFC9B917F297ULL, "vxlt_warehouse_oe_0_0.vlm"},
	{0xA7F0FD3F8CBAFD54ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium2_c.anb"},
	{0xA7F161F704204660ULL, "807858990.ogg"},
	{0xA7F170ADA6256B02ULL, "6112987_m.fxe"},
	{0xA7F18C82E0D3C1C7ULL, "vxht_keep_4jl_0_0.vlm"},
	{0xA7F1F194769ED3A7ULL, "52194653.ogg"},
	{0xA7F27BE5BB4EBC6EULL, "vxca_caverns1_j2k_0_0.vlm"},
	{0xA7F28466BBE84181ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urvb_0s.dds"},
	{0xA7F2AE9BA54A3976ULL, "446292294.ogg"},
	{0xA7F2B7350E248D52ULL, "862024071.ogg"},
	{0xA7F2C9B056022965ULL, "6131132_m.fxe"},
	{0xA7F2FB8A7F9C4373ULL, "6098051_m.fxe"},
	{0xA7F333D6C126DA25ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_diagonal_1.mao"},
	{0xA7F3A45498F547E6ULL, "qun231cr_varnell_minion4.utc"},
	{0xA7F3BAADE6E92237ULL, "vxw_bonepitmining_a2_1nt_0_0.vlm"},
	{0xA7F3DAA85261AF19ULL, "gen00fl_merrill6116072.cl"},
	{0xA7F4603BB67D3FA3ULL, "6092079_m.fxe"},
	{0xA7F4D5D51D6A1F06ULL, "qcr220ar_hightown_nt.are"},
	{0xA7F4E668DC1E38FDULL, "153426013.ogg"},
	{0xA7F517059E9D41C6ULL, "6140250_m.fxe"},
	{0xA7F545C0ED4C5789ULL, "dre000_varric6033312.cl"},
	{0xA7F60E37C221022EULL, "w_bonepitdestroyed_a3_189_0.msh"},
	{0xA7F61C37C2211A7FULL, "w_bonepitdestroyed_a3_189_0.mao"},
	{0xA7F62837C2212E1CULL, "w_bonepitdestroyed_a3_189_0.mmh"},
	{0xA7F62D174166834CULL, "vxht_brothel_n_6g6_0_0.vlm"},
	{0xA7F6973C0D3FC7DDULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_r180_rf.evt"},
	{0xA7F75CF1188C4A1FULL, "lt_backalley_g_2_0.rml"},
	{0xA7F7708B0FC76C7BULL, "atl_purchaseco_dxt5_dat.xml"},
	{0xA7F7AC89EA23DA9DULL, "w_woundedcoast_74_0.anb"},
	{0xA7F81AD732BD9003ULL, "gen_im_trs_dal_arc_sbw_r2.uti"},
	{0xA7F839F2CFA260C3ULL, "vxw_encounter1_h6_0_0.vlm"},
	{0xA7F87091D8081D30ULL, "blood_conj_01_l2.mao"},
	{0xA7F881D1242A2FC6ULL, "294807208.ogg"},
	{0xA7F8AD7B0B824395ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_flower.msh"},
	{0xA7F8AF7B0B82470BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_flower.mmh"},
	{0xA7F9931CC544C703ULL, "ave121tr_jeven_fired.ncs"},
	{0xA7F9CD8092312A08ULL, "art\\vfx\\combat\\vfx_d_arc_direct_cast_c.mmh"},
	{0xA7FA337E4B276FB4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.ani"},
	{0xA7FA92A24DC79DB6ULL, "vxlt_0_lowtown_a3_n_c_hwv_0_0.vlm"},
	{0xA7FAC3ADFB7314B1ULL, "clf260_harley6093813.cl"},
	{0xA7FADA927344803DULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_maj_c.mmh"},
	{0xA7FB1944D5694B3DULL, "hlac_ht_keep_n_3743_1y.dds"},
	{0xA7FB4E7B8FED77F3ULL, "859377507.ogg"},
	{0xA7FB762075B9DE30ULL, "static_static_p92.psh.fxo.dx11"},
	{0xA7FB7F0D9F972CC2ULL, "107341815.ogg"},
	{0xA7FBB35F98284672ULL, "vxw_woundedcoast_1l_0_0.vlm"},
	{0xA7FBE4E54CF51D03ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower2_.ani"},
	{0xA7FBEF7D114908BCULL, "one200dg_solivitus6066479.cl"},
	{0xA7FC291DB1472B21ULL, "vxw_encounter2_rk_0_0.vlm"},
	{0xA7FC3C953DA7810BULL, "vxdr_primevalentrance_2sp_0_0.vlm"},
	{0xA7FC4EF7905664CEULL, "zz_and_debug6134149.cl"},
	{0xA7FC687894E198BFULL, "ser121_jethann6107097.cl"},
	{0xA7FCD5816FC470F1ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_act_c.impa.ani"},
	{0xA7FCD726944F1C7FULL, "25708300.ogg"},
	{0xA7FD177B3A69EB55ULL, "vxw_bonepit_2ab_0_0.vlm"},
	{0xA7FD2AF0E3D98F6AULL, "483334007.ogg"},
	{0xA7FD8254842A0166ULL, "557331907.ogg"},
	{0xA7FDDC3F8CC54F17ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium2_c.mmh"},
	{0xA7FE13874F1E5A07ULL, "vxw_woundedcoast_g3_0_0.vlm"},
	{0xA7FE2F5D3FF610FFULL, "vxca_cavern1_haunted_cjz_0_0.vlm"},
	{0xA7FE5D20D5ADDB25ULL, "gen00fl_isabela6100141.cl"},
	{0xA7FE7BF118929DB7ULL, "lt_backalley_g_2_0.phy"},
	{0xA7FE826EFCEF6082ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf.mmh"},
	{0xA7FE846EFCEF6438ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf.msh"},
	{0xA7FF419DC674353AULL, "766375457.ogg"},
	{0xA7FF41A6AB0E0B17ULL, "vxlt_warehouse_pd_0_0.vlm"},
	{0xA7FF5AAA84A373E3ULL, "llac_ht_mansion_n_2995_2.dds"},
	{0xA7FF8FE58BA2614FULL, "hlac_lt_0_lowtown_a2_n_0_9x.dds"},
	{0xA7FFDB55CCD51A0DULL, "552047673.ogg"},
	{0xA800040F22004997ULL, "t3_cor_h01.tnt"},
	{0xA800118B808CEF6FULL, "6044281_m.fxe"},
	{0xA80059E954B35D81ULL, "6117503_m.fxe"},
	{0xA800D763DF159164ULL, "vxca_caverns1_mines_j6d_0_0.vlm"},
	{0xA80104396E3DDAB9ULL, "vxht_brothel_n_619_0_0.vlm"},
	{0xA80108303546B63DULL, "mag101_aombanter6052755.cl"},
	{0xA8012B6B59CEFEF9ULL, "952695236.ogg"},
	{0xA801307E4B2D8916ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.gad"},
	{0xA801F96D66591264ULL, "lt_warehouse_mmap.dds"},
	{0xA8022C8AD8DF3AB7ULL, "vxw_bonepitmining_a2_1kw_0_0.vlm"},
	{0xA8028E4487E22674ULL, "genip_pp_lte_one_shot.utp"},
	{0xA80299C9689680FAULL, "6067815_f.fxe"},
	{0xA802E6CCB05B6B77ULL, "gen00fl_fenris6126381.cl"},
	{0xA802ED51943E5606ULL, "53685074.ogg"},
	{0xA80346F63E95561CULL, "zz_mcr_debug6072383.cl"},
	{0xA803E0816FCAA193ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_act_c.impa.gad"},
	{0xA803E86B1E4139D3ULL, "hlac_ht_mansion2b_818_27x.dds"},
	{0xA804352317784546ULL, "mag330_walter6045242.cl"},
	{0xA8044BDB43CE0A91ULL, "vxlt_0_lowtown_a3_n_c_gpo_0_0.vlm"},
	{0xA8055D7945EDB93EULL, "6115340_m.fxe"},
	{0xA8056022DC183C1DULL, "6147032_m.fxe"},
	{0xA805A98400A9F558ULL, "follower_banter6163500.cl"},
	{0xA805B1F5D0AEB253ULL, "6102886_m.fxe"},
	{0xA805ED788B763989ULL, "qcr220_grey_wardens6207454.cl"},
	{0xA80608D37B39AFB4ULL, "6172626_m.fxe"},
	{0xA8063C2C9C43B194ULL, "126055546.ogg"},
	{0xA80675D57D621C60ULL, "6142166_m.fxe"},
	{0xA806961F37EA64A0ULL, "968422010.ogg"},
	{0xA806EC82A831E24BULL, "69195205.ogg"},
	{0xA8081E7262E8F30CULL, "6005048_f.fxe"},
	{0xA8084B7E4B33D675ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_4.evt"},
	{0xA8085D516D5CD07AULL, "vxdr_dungeon_1sq_0_1.vlm"},
	{0xA808E15371E108E4ULL, "mag130_danzig6018802.cl"},
	{0xA8090CB41BA8F91CULL, "356090476.ogg"},
	{0xA809763CEBD15E58ULL, "vxlt_0_lowtown_docks_bu5_0_0.vlm"},
	{0xA80A0A42338815D7ULL, "vxlt_undercityserialkiller_1c2_0_1.vlm"},
	{0xA80ADD816FD0BBF8ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_act_c.impa.evt"},
	{0xA80B7C2A782C0903ULL, "vxht_0_hightown_9zm_0_0.vlm"},
	{0xA80BDD645CC0DC2FULL, "vxlt_0_lowtown_a2_n_c_e8j_0_0.vlm"},
	{0xA80C07A778E0BCCFULL, "vxw_bonepit_27b_0_0.vlm"},
	{0xA80C1009B36FE897ULL, "vxlt_0_lowtown_a2_n_es1_0_0.vlm"},
	{0xA80C7693F64E7859ULL, "hlac_ca_caverns2_5802_52y.dds"},
	{0xA80C769DFC4BB0EBULL, "hlac_w_woundedcoast_-262141_0y.dds"},
	{0xA80CA13A31EC297FULL, "vxca_cavern1_haunted_coe_0_0.vlm"},
	{0xA80CB89300634FF6ULL, "210860629.ogg"},
	{0xA80CDAEC0A60AE39ULL, "mcr370_meredith6136121.cl"},
	{0xA80CEAC1C256BC40ULL, "ht_0_hightown_a3_36_0.rml"},
	{0xA80D06D4B2877851ULL, "175928827.ogg"},
	{0xA80D0D93C0232030ULL, "6094503_f.fxe"},
	{0xA80D431C57F3B386ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_floortile_d.dds"},
	{0xA80DB314BB5230DEULL, "vxlt_0_lowtown_a2_n_gnq_0_0.vlm"},
	{0xA80DB3839D042397ULL, "vxlt_warehouse_ug_0_0.vlm"},
	{0xA80DF2137D92794FULL, "29047960.ogg"},
	{0xA80E1A86891381E4ULL, "hlac_lt_0_lowtown_-2_0z.dds"},
	{0xA80E5C2D15100D0BULL, "hlac_lt_gallowstemplar_f_2394_0z.dds"},
	{0xA80E74F348687ABDULL, "134327245.ogg"},
	{0xA80EA94571319A28ULL, "gen00fl_varric6107297.cl"},
	{0xA80EB0CB108A6361ULL, "6182062_m.fxe"},
	{0xA80EFC9E7218EE02ULL, "art\\vfx\\combat\\vfx_gib_boo_masc.mao"},
	{0xA80F026242B0459AULL, "w_encounter3_146_0.mmh"},
	{0xA80F066242B04CD1ULL, "w_encounter3_146_0.mao"},
	{0xA80F07D19287C768ULL, "6053066_m.fxe"},
	{0xA80F146242B064E0ULL, "w_encounter3_146_0.msh"},
	{0xA80F20CE36F9548FULL, "radialquickbar.gfx"},
	{0xA80FE26BE6299C22ULL, "lgt110cr_arl_bann_teagan.utc"},
	{0xA80FFBE54D05EDEDULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_follower2_.gad"},
	{0xA81036C50F33933AULL, "vxlt_0_lowtown_bz7_0_0.vlm"},
	{0xA8107D58A560CB03ULL, "follower_banter6172943.cl"},
	{0xA81120F197E779EDULL, "vxlt_0_lowtown_docks_a3_n_d0l_0_1.vlm"},
	{0xA81136354E5C08AAULL, "840647852.ogg"},
	{0xA81192678A4B1567ULL, "785911321.ogg"},
	{0xA811E02F1DBB0FB0ULL, "w_bonepitmining_a2_207_0.mao"},
	{0xA811E42F1DBB16C3ULL, "w_bonepitmining_a2_207_0.mmh"},
	{0xA811F22F1DBB2E1DULL, "w_bonepitmining_a2_207_0.msh"},
	{0xA81231F7B9C7E6C9ULL, "hlac_lt_gallowstemplar_n_2383_8x.dds"},
	{0xA8124BE9749CEA0BULL, "424777262.ogg"},
	{0xA8125BD0661D4896ULL, "vxlt_0_lowtown_a2_gky_0_0.vlm"},
	{0xA812748441C650F9ULL, "6108437_m.fxe"},
	{0xA813051215E5F721ULL, "749952151.ogg"},
	{0xA813116484844534ULL, "6128052_m.fxe"},
	{0xA813350785CE0CD1ULL, "vxlt_undercityserialkiller_11o_0_0.vlm"},
	{0xA8134F1E8EC57F04ULL, "hlac_w_woundedcoast_-196608_0z.dds"},
	{0xA81363B2AC7E2F95ULL, "hlac_ca_caverns2_5802_20x.dds"},
	{0xA8138298CD84AF98ULL, "6064738_m.fxe"},
	{0xA813B2111FA09C57ULL, "dae000_hate_follower_react6103236.cl"},
	{0xA813E133ADDAEC58ULL, "203635168.ogg"},
	{0xA813E58AC56EE024ULL, "vxlt_0_lowtown_a2_gz4_0_0.vlm"},
	{0xA813F9CA4596B698ULL, "6087236_m.fxe"},
	{0xA8141DEEE3154EC8ULL, "vxht_keep_3xv_0_0.vlm"},
	{0xA81437F51761DBFFULL, "vxht_0_hightown_a2_cbo_0_0.vlm"},
	{0xA81444265DCC6258ULL, "vxht_keep_5ki_0_0.vlm"},
	{0xA814C36DCDF75F90ULL, "zz_show_funds.ncs"},
	{0xA814D3E1CDFF975CULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_overlay.anb"},
	{0xA815304272E8B2DAULL, "w_sundermount_34_0.mmh"},
	{0xA815344272E8B911ULL, "w_sundermount_34_0.mao"},
	{0xA815424272E8D120ULL, "w_sundermount_34_0.msh"},
	{0xA815AD576A787062ULL, "plt_cod_aow_enchantments.plo"},
	{0xA815B74A07F86490ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_3x.dds"},
	{0xA81601FA601661B8ULL, "6066410_f.fxe"},
	{0xA8160217D03340C0ULL, "6218663_m.fxe"},
	{0xA816D85D5DA44DDCULL, "mag131_keran.cnv"},
	{0xA817F45B5DF038CCULL, "follower_banter6156188.cl"},
	{0xA8188EA4E9A5F2E5ULL, "520293189.ogg"},
	{0xA818FFEF9247FDBCULL, "35996833.ogg"},
	{0xA8192B427A541202ULL, "97416920.ogg"},
	{0xA81971D4461747CAULL, "hlac_lt_0_lowtown_a2_n_1_7x.dds"},
	{0xA8199C6FFAE7ABFCULL, "mag121_viveka6082592.cl"},
	{0xA819FA5D5DA6BD2CULL, "mag131_keran.bnk"},
	{0xA81AC9256ED8EF90ULL, "vxlt_0_lowtown_n_gbw_0_0.vlm"},
	{0xA81ADCD51D8AFAF1ULL, "qcr220ar_hightown_nt.lst"},
	{0xA81AEE673D927109ULL, "72774648.ogg"},
	{0xA81B011310E779F7ULL, "6072279_m.fxe"},
	{0xA81B4D35830439BBULL, "348536873.ogg"},
	{0xA81B83F106673236ULL, "hlac_ht_0_hightown_n_-131072_0y.dds"},
	{0xA81BE362EC8BA4CFULL, "132788242.ogg"},
	{0xA81C0737C241C5E7ULL, "w_bonepitdestroyed_a3_189_0.phy"},
	{0xA81C17F5F26D51D7ULL, "vxlt_warehouse_1bc_0_0.vlm"},
	{0xA81C3991B6B03808ULL, "6148151_m.fxe"},
	{0xA81C50E34C5B9413ULL, "art\\vfx\\creatures\\vfxroot.cess.evt"},
	{0xA81CFFD9DB599344ULL, "mrl150_merrill6072770.cl"},
	{0xA81D2FCF8AC017BDULL, "6034543_m.fxe"},
	{0xA81E1FCB568335B7ULL, "gen_im_trs_qun_lgt_hlm_r1.uti"},
	{0xA81E46900C33E78FULL, "qun111_ketojan6040984.cl"},
	{0xA81E8E7372900783ULL, "w_bonepitmining_a2_115_0.msh"},
	{0xA81E9C7372901FDDULL, "w_bonepitmining_a2_115_0.mmh"},
	{0xA81EA073729026EEULL, "w_bonepitmining_a2_115_0.mao"},
	{0xA81F3DDCB0F32C01ULL, "zz_followers_banter6158565.cl"},
	{0xA81F40174D0AD82BULL, "6075245_m.fxe"},
	{0xA81F543E804F7CF3ULL, "art\\characters\\creatures\\dragon\\animation\\exit.ani"},
	{0xA81F8D61C0F4B3CAULL, "plt_cod_hst_llomerynaccord.plo"},
	{0xA81F964AD9803793ULL, "6041699_m.fxe"},
	{0xA8209B459B14C36CULL, "vxlt_gallowstemplar_n_5k5_0_0.vlm"},
	{0xA820CDAD58136116ULL, "vxlt_0_lowtown_a2_gnz_0_0.vlm"},
	{0xA820F4D51D8F8E88ULL, "qcr220ar_hightown_nt.ncs"},
	{0xA821D5BE2B68D520ULL, "6189400_m.fxe"},
	{0xA821DF1FFAF57A84ULL, "vxlt_0_lowtown_a2_n_enx_0_1.vlm"},
	{0xA821EA481388E96AULL, "mer240_jansen6050384.cl"},
	{0xA82249E34C5FF454ULL, "art\\vfx\\creatures\\vfxroot.cess.gad"},
	{0xA823469B36241B0EULL, "lt_0_lowtown_docks_a3_water_16557.mmh"},
	{0xA823529B36242FA5ULL, "lt_0_lowtown_docks_a3_water_16557.mao"},
	{0xA823609B362447FCULL, "lt_0_lowtown_docks_a3_water_16557.msh"},
	{0xA8239E293952E351ULL, "532355071.ogg"},
	{0xA823F7C687FFCEB8ULL, "6144317_m.fxe"},
	{0xA8240778AC05D886ULL, "mag160_feynriel6107434.cl"},
	{0xA8240BFD0DEFE51FULL, "283195447.ogg"},
	{0xA8242728B84AE456ULL, "testinstruction.sub.xml"},
	{0xA8245E02AE7C230AULL, "vxw_encounter2_camp_8v_0_0.vlm"},
	{0xA8249F28DFC6610FULL, "998119729.ogg"},
	{0xA824C4C402CB59E0ULL, "vxht_0_hightown_a3_ccb_0_0.vlm"},
	{0xA8250F4C31AA3C1AULL, "gen00fl_anders6096641.cl"},
	{0xA82542D34C8C40A7ULL, "vxw_encounter3_fj_0_0.vlm"},
	{0xA8256A7788D0FAEBULL, "mcr370_guardsman6086983.cl"},
	{0xA8258FAF7A0E3B71ULL, "6110556_f.fxe"},
	{0xA8264503F1A76FCDULL, "hlac_ca_caverns1_mines_24422_0y.dds"},
	{0xA826877FEE15C79FULL, "6064953_m.fxe"},
	{0xA826D89FD1DFC7CAULL, "223687033.ogg"},
	{0xA826DBC3F2B42E12ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appc_0t.dds"},
	{0xA826F88AE4D8C062ULL, "vxht_0_hightown_a2_7g1_0_0.vlm"},
	{0xA827467BC997CF59ULL, "303050563.ogg"},
	{0xA82799DDD9487593ULL, "hlac_lt_0_lowtown_a2_1_1z.dds"},
	{0xA827A056F73449EEULL, "6162988_m.fxe"},
	{0xA8281CBABABE3335ULL, "vxht_0_hightown_cap_0_0.vlm"},
	{0xA8285E22C8A26FAFULL, "fen310_fenris6118202.cl"},
	{0xA828ABF6EC6C2398ULL, "6077064_m.fxe"},
	{0xA828BFAECBA58C9DULL, "character_character_p84.psh.fxo.dx11.assembly"},
	{0xA82958E34C662C32ULL, "art\\vfx\\creatures\\vfxroot.cess.ani"},
	{0xA8298969FBD51D6DULL, "vxca_caverns1_mines_j27_0_2.vlm"},
	{0xA82998C4EEEA4924ULL, "6070192_f.fxe"},
	{0xA829C520B2B33412ULL, "944083433.ogg"},
	{0xA829CABA8DD312F0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_spitoon.msh"},
	{0xA829D8BA8DD32A0AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_spitoon.mmh"},
	{0xA829DCBA8DD33141ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_spitoon.mao"},
	{0xA82A9002C38BCA2CULL, "6286187.ogg"},
	{0xA82AED26A11D153BULL, "vxht_0_hightown_a2_7vk_0_0.vlm"},
	{0xA82BF8AEC4D835AAULL, "413464941.ogg"},
	{0xA82C28FE8CBB8B8DULL, "hlac_ht_mansion_n_3366_0y.dds"},
	{0xA82C4C3E8059F92AULL, "art\\characters\\creatures\\dragon\\animation\\exit.evt"},
	{0xA82C9733A07D38D5ULL, "vxw_encounter1_jg_0_1.vlm"},
	{0xA82C9766F9162F0BULL, "21640141.ogg"},
	{0xA82CB3A7AE199A7BULL, "139900676.ogg"},
	{0xA82D023357BC2406ULL, "305150385.ogg"},
	{0xA82D3984A2AC7455ULL, "vxdr_primevaltomb_9da_0_1.vlm"},
	{0xA82D8221AEC18437ULL, "vxw_bonepitmining_a2_1zp_0_0.vlm"},
	{0xA82D933DFA60DBF4ULL, "vxlt_0_lowtown_a2_n_c_gv5_0_0.vlm"},
	{0xA82D9B99D1F89D38ULL, "563288073.ogg"},
	{0xA82DBC7A54D7DA82ULL, "6078720_m.fxe"},
	{0xA82F2FC8DFE9F47EULL, "6043704_m.fxe"},
	{0xA82F583E25B281F9ULL, "6106637_m.fxe"},
	{0xA82FA03D5BD1D836ULL, "vxlt_undercitydungeon_1jd_0_0.vlm"},
	{0xA82FA6FB8A378F36ULL, "gen00fl_merrill6128744.cl"},
	{0xA82FD61D9B918BC6ULL, "6059585_m.fxe"},
	{0xA82FD8F2A400D1F9ULL, "383696112.ogg"},
	{0xA82FFD3F60355E08ULL, "vxht_mansion_n_5mp_0_0.vlm"},
	{0xA8306E8346CDAF28ULL, "6074450_m.fxe"},
	{0xA8307E3956972753ULL, "vxht_0_hightown_n_cf9_0_1.vlm"},
	{0xA830B8E6FE3D0DC8ULL, "hlac_lt_gallowstemplar_n_2391_9y.dds"},
	{0xA830BFB744FC97E8ULL, "6107301_m.fxe"},
	{0xA8310C3115B8E1C5ULL, "gencr_qunari_maskless_thral.utc"},
	{0xA831AD89EA555495ULL, "w_woundedcoast_74_0.rml"},
	{0xA831DD46E62D167EULL, "character_character_p52.psh.fxo"},
	{0xA83248379794CAA4ULL, "147344678.ogg"},
	{0xA832677BD5DA9D4CULL, "9881411.ogg"},
	{0xA832A0392F954478ULL, "6165424_m.fxe"},
	{0xA832B2965192A4B2ULL, "920833029.ogg"},
	{0xA832C363932790AFULL, "860913200.ogg"},
	{0xA832CFDFA0723B8AULL, "vxw_encounter2_camp_3w_0_0.vlm"},
	{0xA83311E14ACA9ED9ULL, "lgt000cr_hate_guards.utc"},
	{0xA8332B3E805FDF9DULL, "art\\characters\\creatures\\dragon\\animation\\exit.gad"},
	{0xA8335F315670599DULL, "clf221_jalen6070359.cl"},
	{0xA8337091BCB1A943ULL, "hlac_lt_0_lowtown_docks_-262144_3x.dds"},
	{0xA833B4B03E825927ULL, "vxw_encounter3_ce_0_0.vlm"},
	{0xA833DCB3D72288C0ULL, "hlac_w_bonepit_458755_0x.dds"},
	{0xA834216242CF96A5ULL, "w_encounter3_146_0.phy"},
	{0xA8345A16159FFFB6ULL, "vxlt_0_lowtown_a3_n_c_hcz_0_0.vlm"},
	{0xA834CB4C99BA7EB3ULL, "6167335_m.fxe"},
	{0xA834CBD3C1B3A7D8ULL, "vxlt_0_lowtown_docks_bb0_0_0.vlm"},
	{0xA834F054916EA0F6ULL, "and100_anders6095930.cl"},
	{0xA835590EB728A03EULL, "115444243.ogg"},
	{0xA835B3393995B43AULL, "vxht_0_hightown_a2_a2t_0_0.vlm"},
	{0xA835C99164307A60ULL, "hlac_dr_primevalentrance_1726_4z.dds"},
	{0xA835EACFF7A2EC61ULL, "vxht_keep_3ip_0_0.vlm"},
	{0xA8363290A6BC26C0ULL, "6070432_m.fxe"},
	{0xA83718772C723DACULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_weed.mmh"},
	{0xA8371E772C72479EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_weed.msh"},
	{0xA8377FF118C26983ULL, "lt_backalley_g_2_0.anb"},
	{0xA8384363042E91B2ULL, "6171386_m.fxe"},
	{0xA83843DFF47A663EULL, "follower_banter6170476.cl"},
	{0xA838517D5B9AF454ULL, "879459859.ogg"},
	{0xA83908082A31813EULL, "mag131_taraneh6044442.cl"},
	{0xA8391C265B62828CULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_cobble_d.dds"},
	{0xA83A2315DA153CC4ULL, "dae000_madam_lusine6152275.cl"},
	{0xA83A62DA66D9F2C4ULL, "6102346_m.fxe"},
	{0xA83A6FC2B6CF40E7ULL, "pro000_wesley.bnk"},
	{0xA83B45F48D0F7DEDULL, "vxca_caverns1_mines_egm_0_0.vlm"},
	{0xA83B4F427309B6E5ULL, "w_sundermount_34_0.phy"},
	{0xA83BF3FEA0B79CB7ULL, "vxw_bonepitmining_a2_1ws_0_0.vlm"},
	{0xA83D88CA3D4B092AULL, "hlac_lt_undercityencounter_-1_36y.dds"},
	{0xA83D91C2B6D1B05FULL, "pro000_wesley.cnv"},
	{0xA83E0142A13F8381ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_highlandplant01_dp.dds"},
	{0xA83E430CFC7A2FE8ULL, "305664051.ogg"},
	{0xA83E5316AAA48377ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs08a_0d.dds"},
	{0xA83E87B8EFE193ACULL, "6167475_m.fxe"},
	{0xA83E89BA8D814341ULL, "hlac_ht_0_hightown_-65538_0x.dds"},
	{0xA83EB7C886C68784ULL, "6179475_f.fxe"},
	{0xA83F0760AF6406C7ULL, "hlac_dr_dungeon_3207_7x.dds"},
	{0xA83F433DDB4D6AE4ULL, "6154042_m.fxe"},
	{0xA83F9A89DFDEC46DULL, "vxlt_0_lowtown_docks_9fj_0_0.vlm"},
	{0xA83FA5EF06AF3DECULL, "6202776_m.fxe"},
	{0xA83FE15D9A0AAA45ULL, "hm_gamlen2.mor"},
	{0xA840E4898A473AC6ULL, "vxlt_0_lowtown_docks_a3_n_cyr_0_0.vlm"},
	{0xA8412A13CCA8F0D6ULL, "624156424.ogg"},
	{0xA841C1218E359DA4ULL, "hlac_lt_blackemporium_1055_15x.dds"},
	{0xA8423042730FA0F1ULL, "w_sundermount_34_0.rml"},
	{0xA8427E944801EECBULL, "hlac_dr_primevaltomb_665_2x.dds"},
	{0xA842BBD591E1E366ULL, "vxca_cave_164_0_0.vlm"},
	{0xA842CCDCB04910ECULL, "6143293_m.fxe"},
	{0xA842F27922B79A23ULL, "6160745_m.fxe"},
	{0xA8437C00C887FC96ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_chg_p.impa.ani"},
	{0xA843DCB91E517837ULL, "zz_followers_banter6158448.cl"},
	{0xA843E9DAFAB7D8CEULL, "hlac_ht_playershouse_n_855_29z.dds"},
	{0xA8442A66515FEF5AULL, "551405705.ogg"},
	{0xA84471657EA0A2BAULL, "follower_barks6153692.cl"},
	{0xA84487E3C8B5E6AEULL, "6041782_m.fxe"},
	{0xA8451E793DF59162ULL, "6162676_m.fxe"},
	{0xA84550F118CE570FULL, "lt_backalley_g_2_0.mao"},
	{0xA8455CF118CE6BACULL, "lt_backalley_g_2_0.mmh"},
	{0xA84562F118CE759EULL, "lt_backalley_g_2_0.msh"},
	{0xA8456707377B2994ULL, "dae000_bodhan6129094.cl"},
	{0xA845DB3B5744BD50ULL, "6174043_f.fxe"},
	{0xA845F2278DFCD0ADULL, "hlac_w_woundedcoast_-65530_0z.dds"},
	{0xA84649B203A5AE01ULL, "vxw_sundermount_355_0_1.vlm"},
	{0xA846A42303406713ULL, "llac_w_bonepitdestroyed_a3_196610_0.dds"},
	{0xA846B53FF624AA32ULL, "6174086_m.fxe"},
	{0xA8471464DF2F3D45ULL, "zz_seb_debug6215763.cl"},
	{0xA847304A37E45161ULL, "vxdr_primevaltomb_50l_0_0.vlm"},
	{0xA84748E8529E49DCULL, "gen_im_wep_mag_sta_act0_04.uti"},
	{0xA847560F39A97110ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglebody_d.dds"},
	{0xA84763C13C6855F5ULL, "vxlt_darktown_6ij_0_0.vlm"},
	{0xA847E276591F3F63ULL, "dae000_flemeth6013270.cl"},
	{0xA847F4AF70EBE840ULL, "mcr371cr_templar_cut.utc"},
	{0xA84829C5B71675E5ULL, "6021242_m.fxe"},
	{0xA8482DDDFA77FE30ULL, "17482949.ogg"},
	{0xA848665A446F2F3DULL, "vxw_woundedcoast_oi_0_0.vlm"},
	{0xA848A226CCA7FC72ULL, "vxw_bonepitmining_a2_1or_0_0.vlm"},
	{0xA848B7B61D36CCCAULL, "vxht_0_hightown_a2_c40_0_0.vlm"},
	{0xA849681ACF2757C9ULL, "439514810.ogg"},
	{0xA849C19D727DF2EDULL, "hlac_lt_undercityencounter_-1_20y.dds"},
	{0xA84A88B89AA3B07CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockaxxl.mao"},
	{0xA84A94B89AA3C41FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockaxxl.mmh"},
	{0xA84A9AB89AA3CEC1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockaxxl.msh"},
	{0xA84ACC0B1FC83B6AULL, "vxlt_hangedman_n_1qd_0_0.vlm"},
	{0xA84B47ABD0FF52B1ULL, "723324134.ogg"},
	{0xA84B8BB6CD3BEC22ULL, "character_character_p274.psh.fxo.dx11"},
	{0xA84BDF69C9871765ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_0x.dds"},
	{0xA84C3C8CAE46D57DULL, "vxlt_gallowscourtyard_n_66r_0_0.vlm"},
	{0xA84C5E5E22B5A8E6ULL, "vxht_mansion_n_6ok_0_0.vlm"},
	{0xA84D1DF0907907C8ULL, "follower_banter6171895.cl"},
	{0xA84D91AF4F59B11DULL, "vxw_encounter2_camp_oq_0_1.vlm"},
	{0xA84DAAED07C21D43ULL, "vxlt_hangedman_21i_0_0.vlm"},
	{0xA84DACFE8D474224ULL, "vxlt_0_lowtown_a2_gn8_0_0.vlm"},
	{0xA84E1AE2FB011F4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_loop.ani"},
	{0xA84E82EADE8C207EULL, "vxlt_gallowstemplar_n_2p9_0_0.vlm"},
	{0xA84E981CC58D0F9AULL, "ave121tr_jeven_fired.utt"},
	{0xA84EC87DF92950ABULL, "6163735_m.fxe"},
	{0xA84ED3040F003A82ULL, "6161310_m.fxe"},
	{0xA84EE52310534C8EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v3.ani"},
	{0xA84EEF3221A2A817ULL, "6155391_f.fxe"},
	{0xA84FA1E31A38A855ULL, "vxlt_0_lowtown_a2_n_c_gnx_0_0.vlm"},
	{0xA850202EC95667C5ULL, "128790428.ogg"},
	{0xA85056115C784F22ULL, "vxlt_0_lowtown_n_eh9_0_0.vlm"},
	{0xA8505CFF10EAB19CULL, "6107548_f.fxe"},
	{0xA850A2FE363AB845ULL, "1045354452.ogg"},
	{0xA8510FBE6832EA8EULL, "107641041.ogg"},
	{0xA8518400C8944787ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_chg_p.impa.evt"},
	{0xA851FA870D5CD7B4ULL, "668478735.ogg"},
	{0xA85215D41B2016A7ULL, "hlac_lt_0_lowtown_-65537_1z.dds"},
	{0xA8527356E918C327ULL, "vxht_keep_6hx_0_0.vlm"},
	{0xA8532233E5C6EB31ULL, "plt_cod_ltr_pro_athenril.nss"},
	{0xA853360710F969E0ULL, "360442238.ogg"},
	{0xA8533CB6B81345E7ULL, "hm_slaver01.mor"},
	{0xA8533DA1F5C952F6ULL, "isa211_isabela6070107.cl"},
	{0xA853403987A2ADC2ULL, "zz_approval_debug6003511.cl"},
	{0xA85390562FEBA9E2ULL, "6167576_m.fxe"},
	{0xA8543232EC8C63FBULL, "qun110_petrice6018643.cl"},
	{0xA854B1C71C22CEA2ULL, "6153311_m.fxe"},
	{0xA854CF72696DFE42ULL, "6163213_m.fxe"},
	{0xA85527E2FB07538EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_loop.gad"},
	{0xA8554E2DAB3193D8ULL, "gen00fl_sebastian6180431.cl"},
	{0xA855A22729DA69E1ULL, "vxdr_primevaltomb_55c_0_0.vlm"},
	{0xA8561701295DCA1BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_psych_self.evt"},
	{0xA8566362A3C72FE9ULL, "pro000_meeran6050079.cl"},
	{0xA856F2D20BA40E04ULL, "6019208_m.fxe"},
	{0xA85729930F2CE54AULL, "vxht_0_hightown_a2_c31_0_0.vlm"},
	{0xA857B6DB9B069E70ULL, "vxlt_darktown_50o_0_0.vlm"},
	{0xA85809BCE8E01022ULL, "243319394.ogg"},
	{0xA85855B87C01B5FAULL, "141736257.ogg"},
	{0xA85877883D60BAD8ULL, "vxht_keep_n_4i5_0_0.vlm"},
	{0xA858A500C89A9ED0ULL, "art\\vfx\\cinematic\\vfx_meredith_swd_chg_p.impa.gad"},
	{0xA8591B9CDBD21777ULL, "hlac_ca_caverns2_5802_35x.dds"},
	{0xA85936D768599DB3ULL, "6086360_m.fxe"},
	{0xA8596110A3114842ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medb_0s.dds"},
	{0xA859BF870F744AC0ULL, "6084237_m.fxe"},
	{0xA859D012F19E81C9ULL, "dae000_merchant_access_hf6162951.cl"},
	{0xA85A976451DDD75CULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\prp_barkburn_n.dds"},
	{0xA85AD59E86BC2D1EULL, "vxht_brothel_n_6da_0_0.vlm"},
	{0xA85B545819ACADACULL, "vxlt_shop_l0_0_0.vlm"},
	{0xA85BDD23105DC9DFULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v3.evt"},
	{0xA85C2A7349FFB941ULL, "character_character_p110.psh.fxo.dx11.assembly"},
	{0xA85C537BC700994CULL, "63263969.ogg"},
	{0xA85CCA08CA660390ULL, "6128859_f.fxe"},
	{0xA85D10012963DD5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_psych_self.gad"},
	{0xA85D1FF05A8DDD66ULL, "745549840.ogg"},
	{0xA85D44E100F65302ULL, "6166210_m.fxe"},
	{0xA85D52EE07EDE18EULL, "vxdr_primevaltomb_7qz_0_0.vlm"},
	{0xA85DD46B1DBF0E1EULL, "6123391_f.fxe"},
	{0xA85E6554ADD2F831ULL, "6083991_f.fxe"},
	{0xA85E7A0FF7E7A51FULL, "vxlt_gallowsprison_n_6nr_0_0.vlm"},
	{0xA85E811C7A496EA1ULL, "vxht_brothel_n_6iy_0_0.vlm"},
	{0xA85EF313AD000988ULL, "vxlt_gallowsprison_n_5wm_0_0.vlm"},
	{0xA85EF5C77197D499ULL, "vxht_brothel_n_5bo_0_0.vlm"},
	{0xA85F6F54F087C982ULL, "317218491.ogg"},
	{0xA85F7B9D0A211C2EULL, "mag311_keran6041737.cl"},
	{0xA85FC8DEEA4112ECULL, "vxlt_0_lowtown_docks_n_bar_0_0.vlm"},
	{0xA860372381D431AEULL, "hlac_ca_caverns2_5802_63z.dds"},
	{0xA860552AE4938B72ULL, "288374921.ogg"},
	{0xA860806C4782DF14ULL, "follower_banter6165623.cl"},
	{0xA8608675D79C6DCAULL, "335284814.ogg"},
	{0xA860A246E1ABAE21ULL, "6145100_m.fxe"},
	{0xA860CBFB0495D7ABULL, "dae300_alistair_king6134941.cl"},
	{0xA8614EF9FA68F20CULL, "art\\levels\\ztools\\default\\generic\\0wo_angle.msh"},
	{0xA86154F9FA68FCFEULL, "art\\levels\\ztools\\default\\generic\\0wo_angle.mmh"},
	{0xA86160F9FA691095ULL, "art\\levels\\ztools\\default\\generic\\0wo_angle.mao"},
	{0xA8623BEA16D79A57ULL, "6060359_m.fxe"},
	{0xA862BE231063B3C8ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v3.gad"},
	{0xA862E0C5EDEDD318ULL, "vxht_bartrand_n_5nv_0_0.vlm"},
	{0xA8630F012968475AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_psych_self.ani"},
	{0xA8636AA52800A697ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pile_rags_0_n.dds"},
	{0xA863B17BF4120331ULL, "6085367_m.fxe"},
	{0xA863D7663C4672CCULL, "zz_approval_debug6003922.cl"},
	{0xA863E48A8C971712ULL, "853048279.ogg"},
	{0xA8645F910CB470CEULL, "vxlt_0_lowtown_a2_n_c_fge_0_0.vlm"},
	{0xA864BB021E0F5EB0ULL, "vxw_bonepit_29k_0_0.vlm"},
	{0xA864C4661BB862B1ULL, "one300_zevran6075575.cl"},
	{0xA864F2FE47A9F231ULL, "vxht_brothel_n_32a_0_0.vlm"},
	{0xA8650BA1C06B4816ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadb.phy"},
	{0xA8654A14285B603DULL, "vxw_woundedcoast_ag_0_0.vlm"},
	{0xA86576529DB3CF12ULL, "vxdr_primevalentrance_1m5_0_1.vlm"},
	{0xA865A8EE9A241828ULL, "one200_gold_followers6120125.cl"},
	{0xA865AB2A166A2BE5ULL, "seb201_lady_harimann6177379.cl"},
	{0xA865CD93A940CDC0ULL, "vxw_woundedcoast_1bv_0_0.vlm"},
	{0xA865E690BD56909CULL, "vxlt_gallowstemplar_n_38z_0_0.vlm"},
	{0xA866030C14F33B2CULL, "6206967_m.fxe"},
	{0xA8660FA995FF1385ULL, "and100_anders6096044.cl"},
	{0xA866305861320A48ULL, "follower_banter6141904.cl"},
	{0xA866607544EE60D7ULL, "6139790_m.fxe"},
	{0xA8668153DBD01CD5ULL, "ca_cavern1_haunted_26038.rml"},
	{0xA866A43926D44A54ULL, "37560078.ogg"},
	{0xA8671C9635BD1200ULL, "vxw_bonepit_257_0_0.vlm"},
	{0xA868D6F59FD4CCEFULL, "vxlt_foundry_n_zu_0_0.vlm"},
	{0xA868E6926377CD87ULL, "mag170_thrask6020549.cl"},
	{0xA868F799094BFDB0ULL, "mag291_demon_marethari6029057.cl"},
	{0xA868F99486E5CF41ULL, "hm_iconicmale.mop"},
	{0xA868F99486E5CF43ULL, "hm_iconicmale.mor"},
	{0xA869000776466201ULL, "6118674_m.fxe"},
	{0xA8695076BB582EFEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopening_rog.ani"},
	{0xA86951FD621B3B5CULL, "vxdr_slavers_1az_0_0.vlm"},
	{0xA869C6350BA2C62CULL, "vxlt_shop_g1_0_0.vlm"},
	{0xA869DBE989365F1AULL, "hlac_lt_gallowscourtyard_-131072_3z.dds"},
	{0xA86A0EDD5202F864ULL, "192828653.ogg"},
	{0xA86A43A2400C1B79ULL, "817332733.ogg"},
	{0xA86A51232235101DULL, "59675829.ogg"},
	{0xA86A5BAFE0D0E5C0ULL, "327780476.ogg"},
	{0xA86A699A6832EC3DULL, "6118664_f.fxe"},
	{0xA86B273A014DCF27ULL, "lt_0_lowtown_a2_mmap_ph.dds"},
	{0xA86B63CF2A249B20ULL, "vxlt_undercityencounter_ui_0_0.vlm"},
	{0xA86B7A48D2247E7FULL, "506994954.ogg"},
	{0xA86B8A37CFCE6E19ULL, "gen00fl_varric6118957.cl"},
	{0xA86BD2F263479790ULL, "6077046_m.fxe"},
	{0xA86BF5FF9072A575ULL, "gen00fl_isabela6103953.cl"},
	{0xA86C7DE7E92C6F71ULL, "6045786_f.fxe"},
	{0xA86CABE60289BE7EULL, "hlac_lt_0_lowtown_131072_6y.dds"},
	{0xA86CE8856BDC9721ULL, "vxlt_0_lowtown_a3_n_c_h26_0_0.vlm"},
	{0xA86D197126035DDFULL, "vxht_keep_n_67r_0_0.vlm"},
	{0xA86D82F86D15CB4BULL, "vxht_mansion2b_3g4_0_0.vlm"},
	{0xA86DAC9195903803ULL, "904335858.ogg"},
	{0xA86DC99E7739EDA0ULL, "lowtown_male_alienage2_mob6166551.cl"},
	{0xA86E63D4090CDD38ULL, "zz_followers_banter6157417.cl"},
	{0xA86E688590A3DEB4ULL, "vxw_blightlands_n9_0_0.vlm"},
	{0xA86E751821051E5AULL, "ca_caverns1_mmap.dds"},
	{0xA86EA11C2175CDCEULL, "hlac_ht_0_hightown_a2_131071_2x.dds"},
	{0xA86EDFEFD1D20FAAULL, "6030658_m.fxe"},
	{0xA86EF4A1C07375B7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadb.msh"},
	{0xA86F06A1C073941AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadb.mao"},
	{0xA86F0AA1C0739B29ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_buzzheadb.mmh"},
	{0xA86F8BDE1C105AD3ULL, "vxw_encounter2_3r_0_0.vlm"},
	{0xA86FC8F6DA8FBBDDULL, "sum_conj_star.mao"},
	{0xA86FE769EEE86BEBULL, "hlac_lt_0_lowtown_a2_131072_3y.dds"},
	{0xA870008F7825D1DDULL, "rock_large.mao"},
	{0xA870048F7825D816ULL, "rock_large.mmh"},
	{0xA8700E8F7825E914ULL, "rock_large.msh"},
	{0xA870489725F9A9EDULL, "vxht_0_hightown_a2_bai_0_0.vlm"},
	{0xA870751D384EA75FULL, "6070390_m.fxe"},
	{0xA870BB0235B33D0AULL, "6116514_m.fxe"},
	{0xA8711A4F02AC0FE4ULL, "vxht_mansion2_33s_0_0.vlm"},
	{0xA8714BAA76AA82C8ULL, "335329973.ogg"},
	{0xA8714CF1DCA20C09ULL, "ser221_gamlen6131969.cl"},
	{0xA871607853657C20ULL, "llac_lt_gallowscourtyard_n_-65538_3.dds"},
	{0xA871B092F1E104B7ULL, "vxht_playershouse_n_1r3_0_0.vlm"},
	{0xA87279DD7580A8A0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stackshelf_n.dds"},
	{0xA8728AC10FC652DCULL, "zz_and_debug6134099.cl"},
	{0xA8728F69B0F0B6E0ULL, "65144734.ogg"},
	{0xA872A91F4315FB4BULL, "626752890.ogg"},
	{0xA8731760340EE7F8ULL, "mer320_hubert6090362.cl"},
	{0xA87325DF4A80A335ULL, "199149297.ogg"},
	{0xA873260BABBCEE75ULL, "dae300_alistair_king6134877.cl"},
	{0xA8739E5039E6867FULL, "6180417_m.fxe"},
	{0xA873D507F5FE59CEULL, "vxw_woundedcoast_5l_0_0.vlm"},
	{0xA873EF6ADEBA0587ULL, "vxca_cavern1_haunted_jsp_0_0.vlm"},
	{0xA873EFF579491443ULL, "6104339_f.fxe"},
	{0xA873FD166E181718ULL, "hm_amb_merchant.mor"},
	{0xA8743F709B36E640ULL, "vxw_woundedcoast_1mw_0_0.vlm"},
	{0xA874C8A65D4650CCULL, "hlac_dr_primevaltomb_659_8z.dds"},
	{0xA874DCAEB7A5464FULL, "36190736.ogg"},
	{0xA874EB05ED1D80E4ULL, "6032899_f.fxe"},
	{0xA87502AF581AB75DULL, "6148600_m.fxe"},
	{0xA8752DCBE3CBE60BULL, "6206388_m.fxe"},
	{0xA8758E7327B32A80ULL, "vxw_bonepit_204_0_0.vlm"},
	{0xA875A493033EC139ULL, "6110642_f.fxe"},
	{0xA875A4B4C09B9784ULL, "hlac_lt_0_lowtown_docks_a3_n_0_6x.dds"},
	{0xA875CFE0E34CD1A3ULL, "6080450_m.fxe"},
	{0xA875E05EFD56A7E8ULL, "6209377_m.fxe"},
	{0xA875EC1D7CCBEB6BULL, "dae000_madam_lusine6162376.cl"},
	{0xA87679C24A2FA2B1ULL, "221652593.ogg"},
	{0xA8767AC8A323D31CULL, "vxlt_gallowstemplar_f_3sw_0_0.vlm"},
	{0xA876B952C5E45FACULL, "6016568_m.fxe"},
	{0xA877043704E17C90ULL, "vxlt_0_lowtown_a2_ep1_0_1.vlm"},
	{0xA877661607D65558ULL, "hf_ibethany_h1s1.mor"},
	{0xA877AE240479E537ULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown.mao"},
	{0xA877B9586AA85E1EULL, "vxht_brothel_n_6jc_0_0.vlm"},
	{0xA877BE137FA99F61ULL, "vxw_bonepit_yn_0_0.vlm"},
	{0xA877D4DCF6A92743ULL, "733050331.ogg"},
	{0xA877FA0F99F80451ULL, "201964238.ogg"},
	{0xA8780E4F104460D0ULL, "llac_ca_caverns1_24426_3.dds"},
	{0xA8781E7B5D61FCD2ULL, "hm_qsw_warrens3.mor"},
	{0xA878927452DC2CD3ULL, "355479688.ogg"},
	{0xA878C61FD3E5E417ULL, "59503478.ogg"},
	{0xA8790C93DD64F952ULL, "vxw_encounter2_3w_0_0.vlm"},
	{0xA8793E8A97EB3F23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up_alt.ani"},
	{0xA87A2917D9572AC2ULL, "vxca_caverns2prc_7uq_0_0.vlm"},
	{0xA87A50484EFACC52ULL, "ts_darkspawn_03.gda"},
	{0xA87BD98155840599ULL, "vxht_brothel_n_5hm_0_0.vlm"},
	{0xA87C0EA5FCBF1CBFULL, "vxlt_0_lowtown_docks_n_bdr_0_0.vlm"},
	{0xA87C7DCC3736A1F0ULL, "184879683.ogg"},
	{0xA87C9443630A02D6ULL, "6109113_f.fxe"},
	{0xA87C9B55F37D45C9ULL, "gen00fl_aveline6105396.cl"},
	{0xA87CAEFF8E0F9DCCULL, "290333011.ogg"},
	{0xA87E1EDE924448BCULL, "hlac_ca_caverns2prc_5802_23y.dds"},
	{0xA87E5FFD72680D4CULL, "genip_pp_hti_one_shot.utp"},
	{0xA87E6976BB6AB578ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl3_pcopening_rog.gad"},
	{0xA87E7AA69646B237ULL, "6039241_m.fxe"},
	{0xA87ED2D973AD5823ULL, "6172597_m.fxe"},
	{0xA87EE6FA2A44BFDFULL, "6075090_m.fxe"},
	{0xA87F42F495BC6367ULL, "vxdr_primevalentrance_396_0_1.vlm"},
	{0xA87F53C79496475AULL, "336233385.ogg"},
	{0xA87F9BF41DC47B2EULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef01.msh"},
	{0xA87FB5F41DC4A71CULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef01.mmh"},
	{0xA87FFBA2695DD9B1ULL, "ave310_donnic6134740.cl"},
	{0xA87FFD085303440BULL, "hlac_lt_0_lowtown_a3_n_c_65538_4y.dds"},
	{0xA88028AC559D994CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\pause_injured.ani"},
	{0xA880490DB5C18B5CULL, "vxht_keep_n_5mx_0_0.vlm"},
	{0xA8809718E3A75F82ULL, "llac_ca_caverns1_mines_24421_3.dds"},
	{0xA88108045C5310B0ULL, "hlac_w_encounter2_camp_65535_0x.dds"},
	{0xA88146C940FE4B48ULL, "vxw_bonepit_yk_0_0.vlm"},
	{0xA88228FDCCA239DCULL, "ser221_party_comments6166206.cl"},
	{0xA88246E4E7F4724EULL, "vxw_woundedcoast_2m_0_0.vlm"},
	{0xA88293EADF7C8226ULL, "one280cr_iwan.utc"},
	{0xA882B14D8D2CFEC0ULL, "vxw_woundedcoast_1ht_0_0.vlm"},
	{0xA882D1A48B96AD6DULL, "789194841.ogg"},
	{0xA882FF49BB241B92ULL, "hlac_lt_ruinedhouse_n_-1_5y.dds"},
	{0xA8836155830158D4ULL, "one100dg_elren6058886.cl"},
	{0xA8837268CA8E0122ULL, "300464919.ogg"},
	{0xA884BB2767A06F49ULL, "7108283.ogg"},
	{0xA884DF069B9505E9ULL, "mrl151ip_chest_a.utp"},
	{0xA884DF9D323B3A80ULL, "llac_ca_caverns2prc_5802_44.dds"},
	{0xA884FAB4A78E43EFULL, "hlac_lt_gallowscourtyard_65534_0z.dds"},
	{0xA8854AF9E0E996BDULL, "vxw_bonepit_1wo_0_0.vlm"},
	{0xA8857631BC89EB3DULL, "gen00fl_anders6111097.cl"},
	{0xA8862FF0719FB7E1ULL, "vxw_bonepit_dm_0_0.vlm"},
	{0xA886C3EF157ED0D1ULL, "vxlt_undercityserialkiller_1ig_0_0.vlm"},
	{0xA886C56FB380AB1DULL, "6059784_m.fxe"},
	{0xA88735AC55A3CD8EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\pause_injured.gad"},
	{0xA887568A97F7A51AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up_alt.evt"},
	{0xA887F3F216777EDAULL, "zz_neel_test5.cut"},
	{0xA8880193D9E32126ULL, "6125224_m.fxe"},
	{0xA888CAD5E1D5228CULL, "vxlt_0_lowtown_a2_csg_0_0.vlm"},
	{0xA8899B464ED470C5ULL, "vxca_caverns2prc_9rm_0_0.vlm"},
	{0xA889ACB460DF4D95ULL, "zz_qcr_debug6052126.cl"},
	{0xA889DBDFA0983DD8ULL, "vxw_bonepitdestroyed_a3_1j9_0_0.vlm"},
	{0xA889F782DF6B25BAULL, "static_static_p7.vsh.fxo.dx11"},
	{0xA88A47700E0D587AULL, "l_env_wal_gallows_cry.bnk"},
	{0xA88A5BEB6F6E4CDDULL, "vxlt_0_lowtown_a2_n_c_hm6_0_0.vlm"},
	{0xA88A65DD6DDCD31EULL, "vxw_sundermount_20q_0_0.vlm"},
	{0xA88A6F240E10C5A0ULL, "art\\characters\\creatures\\arh_abm_rev\\c_revenant.phy"},
	{0xA88AFFF4D17F35DBULL, "hlac_lt_0_lowtown_docks_-196608_0z.dds"},
	{0xA88B0FB7C1C782B6ULL, "vxht_brothel_n_6e6_0_0.vlm"},
	{0xA88B261459808EA3ULL, "hlac_w_woundedcoast_-262140_0z.dds"},
	{0xA88B3B3A70083A75ULL, "hlac_w_encounter3_-327682_0y.dds"},
	{0xA88BA112BDFFE9E4ULL, "vxlt_warehouse_wb_0_0.vlm"},
	{0xA88BA716A03922A4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potc_sa.dds"},
	{0xA88BF7AA1CC1F935ULL, "gen00fl_sebastian6174875.cl"},
	{0xA88CB70B7EBC3065ULL, "mag320_orsino6060235.cl"},
	{0xA88CF55DE8A24416ULL, "6095942_f.fxe"},
	{0xA88D30AC55A8316DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\pause_injured.evt"},
	{0xA88D44D79B4F1993ULL, "905363690.ogg"},
	{0xA88D810438B1243BULL, "441355188.ogg"},
	{0xA88D8C590771D58DULL, "vxht_bartrand_n_5oe_0_1.vlm"},
	{0xA88DB126A267BA06ULL, "413896625.ogg"},
	{0xA88E32A733941771ULL, "hlac_lt_0_lowtown_a3_n_c_-3_1y.dds"},
	{0xA88E558A97FDC20DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_r_side_up_alt.gad"},
	{0xA88E672E77949D6EULL, "203217944.ogg"},
	{0xA88E9C4E66E17623ULL, "1009760.ogg"},
	{0xA88EE5FC2E49FFEDULL, "599723422.ogg"},
	{0xA88F2FC7F3CCC271ULL, "vxw_encounter2_g4_0_0.vlm"},
	{0xA88F407AABFFE2BEULL, "6177840_f.fxe"},
	{0xA88FD37F458019F6ULL, "m_qun211ar_poison_encounter.wbd"},
	{0xA8901A78288AA88DULL, "vxht_0_hightown_a3_5i3_0_0.vlm"},
	{0xA89027AEEFAFAC70ULL, "6180920_m.fxe"},
	{0xA8903754FF77AB59ULL, "sky_x_partypicker.mao"},
	{0xA8903933E5FAC01AULL, "plt_cod_ltr_pro_athenril.plo"},
	{0xA8909FAAFE3DA9BDULL, "vxw_woundedcoast_pd_0_0.vlm"},
	{0xA890D97A7676F6A8ULL, "6195885_m.fxe"},
	{0xA890F5B468E46AB1ULL, "893579499.ogg"},
	{0xA891232A7F231740ULL, "vxw_woundedcoast_1su_0_0.vlm"},
	{0xA891AF50801B36E1ULL, "gen_im_trs_prf_sta_phy_r2.uti"},
	{0xA8922660DAA76910ULL, "genip_pp_cav_sustained.utp"},
	{0xA8927928045C3260ULL, "446010142.ogg"},
	{0xA892BB07D0024943ULL, "6113050_f.fxe"},
	{0xA893753576953A23ULL, "6027354_m.fxe"},
	{0xA89379FB2CE5B9F1ULL, "pla_s.dds"},
	{0xA8938B0223938D6BULL, "vxlt_0_lowtown_docks_a3_btj_0_0.vlm"},
	{0xA89422E8B5817C51ULL, "311105431.ogg"},
	{0xA89492E5BBF875CDULL, "vxht_0_hightown_a3_a6v_0_0.vlm"},
	{0xA894A1CD6395D061ULL, "vxw_bonepit_ol_0_0.vlm"},
	{0xA894CB9C6DBF8202ULL, "hlac_w_bonepitmining_a2_0_6x.dds"},
	{0xA89535CC0774E951ULL, "vxlt_undercityserialkiller_1nf_0_0.vlm"},
	{0xA8954622AE83D5F6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerpot01_0_n.dds"},
	{0xA895B55AEB30C8FAULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgte.mmh"},
	{0xA895B95AEB30CF31ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgte.mao"},
	{0xA895C75AEB30E740ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgte.msh"},
	{0xA895CBE50D3F09EDULL, "vxlt_0_lowtown_a2_n_eki_0_2.vlm"},
	{0xA896459E8C7114B3ULL, "vxlt_0_lowtown_a2_n_c_feb_0_0.vlm"},
	{0xA8967FE9C3DCB7ECULL, "dae000_pc_home_ambients6208574.cl"},
	{0xA8968B5B30DAA9CEULL, "hlac_w_sundermount_-196606_0x.dds"},
	{0xA897316E9B21D321ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int2_n.dds"},
	{0xA89796C2FB627AD9ULL, "vxlt_0_lowtown_docks_n_b7m_0_0.vlm"},
	{0xA8979E35613AE348ULL, "gen00fl_anders6119685.cl"},
	{0xA897A43172F857C7ULL, "6129931_m.fxe"},
	{0xA897F49F96E3C5AFULL, "50221489.ogg"},
	{0xA89818D7744B657BULL, "llac_ca_caverns1_mines_24427_4.dds"},
	{0xA8986B161238437AULL, "vxca_caverns2prc_9nb_0_0.vlm"},
	{0xA898D7BA5FD2EB9EULL, "vxw_sundermount_25r_0_0.vlm"},
	{0xA8991DCDCFF1F04FULL, "qun220cr_javaris6182707.cl"},
	{0xA8998ADF93846EFAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_secretwheel.mmh"},
	{0xA8998EDF93847531ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_secretwheel.mao"},
	{0xA899904EB32AEDE9ULL, "35683035.ogg"},
	{0xA8999CDF93848D40ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_secretwheel.msh"},
	{0xA89A12EFAFF60264ULL, "vxlt_warehouse_xc_0_0.vlm"},
	{0xA89A806D4CACBB73ULL, "383708491.ogg"},
	{0xA89AF09204FAB269ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_attraction_wave.ani"},
	{0xA89B1E1C8DAE2C74ULL, "dae250ar_sundermount.are"},
	{0xA89B9D570091C16BULL, "414638427.ogg"},
	{0xA89BABA24A6EF735ULL, "vxht_0_hightown_c9x_0_0.vlm"},
	{0xA89BB9B38031F8CDULL, "qun110_petrice6042138.cl"},
	{0xA89BC731D8CCB5D6ULL, "vxlt_gallowstemplar_37n_0_0.vlm"},
	{0xA89BD510B28CC4B5ULL, "gen00fl_fenris6126817.cl"},
	{0xA89BDE0BC1FDB761ULL, "vxlt_0_lowtown_n_f1j_0_0.vlm"},
	{0xA89CBD674F1DF764ULL, "w_encounter3_232_0.mmh"},
	{0xA89CC1674F1DFDB7ULL, "w_encounter3_232_0.mao"},
	{0xA89CD3674F1E1C06ULL, "w_encounter3_232_0.msh"},
	{0xA89CDBA8B5BD06A3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medb_0n.dds"},
	{0xA89CEACA5F21C21AULL, "vxht_0_hightown_a2_c5s_0_0.vlm"},
	{0xA89D48A27F7B2A2BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candlecluster_n.dds"},
	{0xA89D891744673F83ULL, "6128657_m.fxe"},
	{0xA89DC950AC1716E2ULL, "627425220.ogg"},
	{0xA89E2A8324EA7C48ULL, "vxw_bonepit_oi_0_0.vlm"},
	{0xA89E782079BB0FC0ULL, "696870152.ogg"},
	{0xA89F1CECC6357E4EULL, "493178914.ogg"},
	{0xA89FC0CC406DDB74ULL, "hlac_dr_dungeon_3217_12x.dds"},
	{0xA89FDD02CE9EA022ULL, "vxw_bonepitmining_a2_dx_0_0.vlm"},
	{0xA8A11036B1AE5620ULL, "dae310ip_store5.utp"},
	{0xA8A124913348BE70ULL, "mag311_keran6054382.cl"},
	{0xA8A1527FD39F807AULL, "gen00fl_aveline6137510.cl"},
	{0xA8A15324BA3A2DBCULL, "vxht_brothel_n_5b6_0_0.vlm"},
	{0xA8A1E32A821ED5EBULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.cess.evt"},
	{0xA8A1F9A5FAD54163ULL, "fxe_water_pan2.dds"},
	{0xA8A20B920500FEEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_attraction_wave.gad"},
	{0xA8A22BF9FAA04D99ULL, "art\\levels\\ztools\\default\\generic\\0wo_angle.phy"},
	{0xA8A23D4BA472F86BULL, "hlac_dr_primevaltomb_657_5x.dds"},
	{0xA8A30649056B0045ULL, "tools_light.msh"},
	{0xA8A313C77483F3F8ULL, "vis121_saemus6086689.cl"},
	{0xA8A3BCD4CDF034ADULL, "dae000cr_grand_cleric6147682.cl"},
	{0xA8A42CF6B15CE4AEULL, "6031998_m.fxe"},
	{0xA8A4429634CB3EC9ULL, "fen000pt_gift_book.ncs"},
	{0xA8A45E9E94DADCF0ULL, "6143330_m.fxe"},
	{0xA8A4B77B7E672D33ULL, "vxlt_0_lowtown_a2_n_c_fhe_0_0.vlm"},
	{0xA8A4E4C774E21FF4ULL, "314719876.ogg"},
	{0xA8A552D7846EFED0ULL, "695905431.ogg"},
	{0xA8A566C183526348ULL, "vxlt_0_lowtown_n_f1o_0_0.vlm"},
	{0xA8A57A07B7B811EEULL, "vxw_encounter2_camp_jx_0_0.vlm"},
	{0xA8A594F41DE53EE7ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef01.phy"},
	{0xA8A5D2C34E26DF2FULL, "art\\characters\\creatures\\shriek\\textures\\cn_hed_ems_0n.dds"},
	{0xA8A5F67313C6E26FULL, "vxlt_gallowstemplar_f_5by_0_0.vlm"},
	{0xA8A67825A7AB3AEEULL, "6016294_m.fxe"},
	{0xA8A6AF1D2BC59EC8ULL, "vxca_caverns2prc_9t2_0_0.vlm"},
	{0xA8A6C27253F3748EULL, "876424807.ogg"},
	{0xA8A703D99EC11C97ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2eggsac.mmh"},
	{0xA8A707D99EC123A4ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2eggsac.mao"},
	{0xA8A719D99EC14209ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2eggsac.msh"},
	{0xA8A7358374C9C27EULL, "ico_mm_player.dds"},
	{0xA8A74F2E10164EE7ULL, "6106363_m.fxe"},
	{0xA8A7509E542C75E5ULL, "art\\vfx\\textures\\vfx_glyph_paral.dds"},
	{0xA8A76908827CDA08ULL, "vxw_encounter2_camp_2v_0_0.vlm"},
	{0xA8A78C187A502B8BULL, "6132728_m.fxe"},
	{0xA8A7AF22F59D413FULL, "6016145_m.fxe"},
	{0xA8A7B0A65D66BF4DULL, "gen_im_npc_dwf_arm_cht_wara.uti"},
	{0xA8A7F371A5B3B3B6ULL, "vxht_brothel_n_6k8_0_0.vlm"},
	{0xA8A7FC2A82236BACULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.cess.gad"},
	{0xA8A8387B4317A418ULL, "388326994.ogg"},
	{0xA8A8649942871DB6ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar4_.ani"},
	{0xA8A8F568A72E65CCULL, "zz_followers_banter6158265.cl"},
	{0xA8A9089205071850ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_attraction_wave.evt"},
	{0xA8A97F7908D05121ULL, "vxw_encounter2_6g_0_0.vlm"},
	{0xA8A983C39DDD63BFULL, "6105601_m.fxe"},
	{0xA8A99B1E5DB532BBULL, "vxdr_primevalentrance_19u_0_0.vlm"},
	{0xA8A9A432AA21D138ULL, "lt_0_lowtown_docks_a3_n_122_0.phy"},
	{0xA8A9B729A914E982ULL, "747404381.ogg"},
	{0xA8A9F41CCC9D6815ULL, "vxw_woundedcoast_1qz_0_0.vlm"},
	{0xA8AA6EF07DF57F3EULL, "hf_grace.mor"},
	{0xA8AA9F805A2A6060ULL, "ht_0_hightown_n_27_0.anb"},
	{0xA8AB0D1ED9A09160ULL, "6070991_m.fxe"},
	{0xA8AB5CA75117DA9AULL, "vxht_0_hightown_a2_c0l_0_0.vlm"},
	{0xA8AB62835C5CFDBBULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrg.mmh"},
	{0xA8AB6DD9ED6EDE36ULL, "565088044.ogg"},
	{0xA8AB6E835C5D1258ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrg.mao"},
	{0xA8AB80835C5D30A5ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrplrg.msh"},
	{0xA8AB89627C3B24F1ULL, "vxht_0_hightown_a2_asp_0_0.vlm"},
	{0xA8AB997FD38534C3ULL, "gen_im_trs_qun_blt_r0.uti"},
	{0xA8ABCC87509ED5D7ULL, "463833615.ogg"},
	{0xA8ABF63EF15E0519ULL, "zz_ave_debug6064112.cl"},
	{0xA8AC29FF2055C259ULL, "6089738_m.fxe"},
	{0xA8AC369C02C5DC66ULL, "pro000_merchant6051592.cl"},
	{0xA8ACDF5754EC2BF8ULL, "gen00fl_fenris6126560.cl"},
	{0xA8AD1C4416A92551ULL, "vxca_caverns1_mines_gkj_0_1.vlm"},
	{0xA8AD399B08F9DCEDULL, "w_encounter1_20_0.mmh"},
	{0xA8AD3D9B08F9E33EULL, "w_encounter1_20_0.mao"},
	{0xA8AD4B563912ADFAULL, "one220_yevhen6120869.cl"},
	{0xA8AD4B9B08F9FB73ULL, "w_encounter1_20_0.msh"},
	{0xA8AD7FAD21081867ULL, "6094680_m.fxe"},
	{0xA8AD83FA2292A5F0ULL, "qun110_petrice6042267.cl"},
	{0xA8AE30E8DE818406ULL, "449610228.ogg"},
	{0xA8AE9641C1EE1D6DULL, "qun110_petrice6018667.cl"},
	{0xA8AEABBF9453128EULL, "face.mat"},
	{0xA8AEB1A4263B13A6ULL, "vxht_mansion_n_4c1_0_0.vlm"},
	{0xA8AECE5A6F92BCFAULL, "6013288_m.fxe"},
	{0xA8AED8886AD309C3ULL, "6071285_m.fxe"},
	{0xA8AEFB2A822988AAULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.cess.ani"},
	{0xA8AF3C12C70195A8ULL, "gen00fl_fenris6126748.cl"},
	{0xA8AF61700E2CA018ULL, "l_env_wal_gallows_cry.wbd"},
	{0xA8AF70AB6A665BAAULL, "157080191.ogg"},
	{0xA8AF8221AD80D2EEULL, "hlac_lt_warehouse_65537_4y.dds"},
	{0xA8AF8C240E301335ULL, "art\\characters\\creatures\\arh_abm_rev\\c_revenant.msh"},
	{0xA8AF8E240E3016ABULL, "art\\characters\\creatures\\arh_abm_rev\\c_revenant.mmh"},
	{0xA8AFC501AC30463CULL, "vxht_brothel_n_5e7_0_0.vlm"},
	{0xA8B00B35DEED7482ULL, "vxlt_0_lowtown_a2_n_c_ep3_0_0.vlm"},
	{0xA8B05CCEF691F159ULL, "6156508_m.fxe"},
	{0xA8B0C4D5FAAD302CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs102a_0d.dds"},
	{0xA8B0C951FCFC7E3FULL, "pro000_meeran6050107.cl"},
	{0xA8B0F8E10C41EBECULL, "6129712_m.fxe"},
	{0xA8B15506EC0F1690ULL, "6089263_m.fxe"},
	{0xA8B1B8219A7B6E4FULL, "atl_quickbar_dxt5_02.dds"},
	{0xA8B243E33CDA0D18ULL, "vxht_mansion2b_1x8_0_0.vlm"},
	{0xA8B2953CD7055D39ULL, "vxw_sundermount_260_0_0.vlm"},
	{0xA8B322D882EDC515ULL, "hlac_lt_0_lowtown_docks_a3_-4_5z.dds"},
	{0xA8B3272450254B88ULL, "319921476.ogg"},
	{0xA8B3272D14A9D793ULL, "6111755_f.fxe"},
	{0xA8B3AE1EEEA6D2F4ULL, "gen00fl_isabela6096780.cl"},
	{0xA8B3B4822B50CCA8ULL, "381091215.ogg"},
	{0xA8B3D983D4FCDEBDULL, "6111251_m.fxe"},
	{0xA8B44123EA6783C2ULL, "hlac_lt_gallowstemplar_n_2391_0x.dds"},
	{0xA8B45126EC4868ABULL, "vxlt_gallowstemplar_f_4uy_0_0.vlm"},
	{0xA8B473CAFF54E3D3ULL, "dae000_mother6127519.cl"},
	{0xA8B492BB2D34186CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall.phy"},
	{0xA8B49A846D879CA9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolema.anb"},
	{0xA8B4D9E76B8BF3B3ULL, "401846541.ogg"},
	{0xA8B543942B6287B7ULL, "621696858.ogg"},
	{0xA8B570A5DFA69316ULL, "lt_0_lowtown_docks_n_229_0.msh"},
	{0xA8B57861722D40B9ULL, "fen310st_fenris.stg"},
	{0xA8B57AA5DFA6A414ULL, "lt_0_lowtown_docks_n_229_0.mmh"},
	{0xA8B57EA5DFA6AB27ULL, "lt_0_lowtown_docks_n_229_0.mao"},
	{0xA8B5842B5DE78CEDULL, "vxlt_0_lowtown_a2_hff_0_0.vlm"},
	{0xA8B5BDE9ECCA31EEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_vael_0s.dds"},
	{0xA8B5BFC40FFC9649ULL, "807305862.ogg"},
	{0xA8B5F89DFFD3A24AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_stomach.gad"},
	{0xA8B5FBC902EA2C4DULL, "865420691.ogg"},
	{0xA8B6B3BBE18B7966ULL, "follower_banter6160860.cl"},
	{0xA8B735BEF9A60514ULL, "6126766_m.fxe"},
	{0xA8B7399B09023166ULL, "w_encounter1_20_0.rml"},
	{0xA8B76A9D4392D8DAULL, "6116828_m.fxe"},
	{0xA8B7F155FAC669A1ULL, "vxw_encounter2_1f_0_0.vlm"},
	{0xA8B939C86485CC6EULL, "llac_lt_warehouse_n_0_8.dds"},
	{0xA8B999A22E71C9EDULL, "238200677.ogg"},
	{0xA8B9CA2F356F9888ULL, "45445677.ogg"},
	{0xA8B9D0D2D6C8EAC7ULL, "vxlt_hangedman_24z_0_0.vlm"},
	{0xA8BA051C8DC8B256ULL, "dae250ar_sundermount.ncs"},
	{0xA8BA70C0510B03ABULL, "lowtown_male_lowlife2_mob6166880.cl"},
	{0xA8BAD45AEB501905ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgte.phy"},
	{0xA8BB0325E89C85F5ULL, "365413456.ogg"},
	{0xA8BB4C35213FADD3ULL, "94058595.ogg"},
	{0xA8BB9034B2EC2482ULL, "vxlt_0_lowtown_a2_n_c_f4t_0_0.vlm"},
	{0xA8BBDB2ACAD1E1E0ULL, "439127448.ogg"},
	{0xA8BBF97F45A59564ULL, "m_qun211ar_poison_encounter.bnk"},
	{0xA8BC8D9942980CF0ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar4_.gad"},
	{0xA8BCC0BCB28AD122ULL, "vxw_bonepitmining_a2_nz_0_0.vlm"},
	{0xA8BCC6BA63A0B03EULL, "w_bonepitmining_a2_4_0_lc.mmh"},
	{0xA8BCDB9DFFD98FE1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_stomach.evt"},
	{0xA8BD318F717C63C9ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_reva_0s.dds"},
	{0xA8BD919853396746ULL, "vxht_mansion2b_3ci_0_0.vlm"},
	{0xA8BDA367989588F3ULL, "testconvo.bnk"},
	{0xA8BDB9245486BB22ULL, "vxw_encounter3_g1_0_1.vlm"},
	{0xA8BDC4204B04FB0BULL, "vxdr_0_deeproads_cwv_0_1.vlm"},
	{0xA8BDFB7D03D0078EULL, "crowdctacticpreset.gda"},
	{0xA8BE2E98332E6746ULL, "hlac_lt_0_lowtown_-2_0x.dds"},
	{0xA8BE36DE9E265EBCULL, "vxht_brothel_n_5h4_0_0.vlm"},
	{0xA8BE4A9C3ADBB9B2ULL, "zz_dre_debug6168294.cl"},
	{0xA8BE85206EB3F8A5ULL, "6048193_m.fxe"},
	{0xA8BE9C752695FAB6ULL, "alpharing.mao"},
	{0xA8BEA4FBC188DB88ULL, "vxca_caverns1_j4w_0_0.vlm"},
	{0xA8BEA9DF93A3BF05ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_secretwheel.phy"},
	{0xA8BF08612331FBC5ULL, "vxlt_gallowscourtyard_7f4_0_0.vlm"},
	{0xA8BF358B44513283ULL, "hlac_dr_dungeon_3217_2z.dds"},
	{0xA8BF8E74C208D1D2ULL, "204321545.ogg"},
	{0xA8BFF12B9BEACAF9ULL, "gen_im_trs_cot_junk3_03.uti"},
	{0xA8C049B14AA47BCAULL, "gen_im_cns_pot_technique_01.uti"},
	{0xA8C053BF3201DD22ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_5z.dds"},
	{0xA8C0A5B9EBBA894CULL, "6125521_m.fxe"},
	{0xA8C0C0A046BCE3B2ULL, "384462067.ogg"},
	{0xA8C0C8DE0FB4C1A0ULL, "6108487_m.fxe"},
	{0xA8C0DB5D64770070ULL, "803829950.ogg"},
	{0xA8C0FE0963E09F73ULL, "hlac_lt_gallowstemplar_n_2383_8z.dds"},
	{0xA8C109055EA550DAULL, "follower_banter6163159.cl"},
	{0xA8C11D1C8DCEF9EFULL, "dae250ar_sundermount.lst"},
	{0xA8C1C5679899AB43ULL, "testconvo.cnv"},
	{0xA8C20E060C3C08A4ULL, "follower_banter6168000.cl"},
	{0xA8C2365804334A5BULL, "vxht_0_hightown_c8u_0_0.vlm"},
	{0xA8C23B4014DBC31EULL, "mag131_keran6086268.cl"},
	{0xA8C25B46AE6B130FULL, "140286730.ogg"},
	{0xA8C2BDE17F77F3B0ULL, "489349087.ogg"},
	{0xA8C2C3C5345C2734ULL, "static_billboard_p0.vsh.fxo"},
	{0xA8C2EDD9A2993F15ULL, "457032109.ogg"},
	{0xA8C33213A50E45FCULL, "gen00fl_isabela6075219.cl"},
	{0xA8C3DB65D888163FULL, "lowtown_fem_alien_elf_2_mob6166348.cl"},
	{0xA8C3EA1234B17C97ULL, "455746705.ogg"},
	{0xA8C3F6084FDDA56DULL, "vxlt_0_lowtown_a2_hce_0_0.vlm"},
	{0xA8C48910DFCE05C5ULL, "mcr370_templar16087632.cl"},
	{0xA8C54F7CC7E3F7B8ULL, "6176495_m.fxe"},
	{0xA8C5F0B0918946F2ULL, "p_glow.mao"},
	{0xA8C6655FFBF9D304ULL, "and121_karl6093468.cl"},
	{0xA8C6B00962851328ULL, "hlac_ht_keep_n_3732_7z.dds"},
	{0xA8C6B68F1700B6A1ULL, "vxw_blightlands_fh_0_0.vlm"},
	{0xA8C6EBD4C190182FULL, "hlac_w_blightlands_196610_0y.dds"},
	{0xA8C78E1CE3B89A5EULL, "ser270_cut_aftermath6162275.cl"},
	{0xA8C7BB334A781CE5ULL, "6163698_m.fxe"},
	{0xA8C7E50ECCB368D7ULL, "static_static_p97.psh.fxo.dx11"},
	{0xA8C82EA2CABEC09BULL, "art\\vfx\\combat\\vfx_r_gn2_hbomb_act_c.mmh"},
	{0xA8C83D9E821F4570ULL, "dae000ip_standing_think_d.utp"},
	{0xA8C842AFC8BF0347ULL, "vxlt_hangedman_21u_0_0.vlm"},
	{0xA8C85159EF6D3D4CULL, "hlac_lt_0_lowtown_docks_-196610_3y.dds"},
	{0xA8C8620F65936AC7ULL, "llac_w_encounter2_-65535_0.dds"},
	{0xA8C8B764B61891FEULL, "ca_caverns2_5802_lc.mmh"},
	{0xA8C8C7AB3AC8FD45ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_sitting_table_hands_clasped_1.anb"},
	{0xA8CAC8BD9C5FF408ULL, "6123440_m.fxe"},
	{0xA8CAD39DFFE5BF30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_stomach.ani"},
	{0xA8CBDFE05C9D5582ULL, "299100966.ogg"},
	{0xA8CBF11A438F81AFULL, "6164460_m.fxe"},
	{0xA8CC0375452F7FC6ULL, "vxht_mansion2b_3fj_0_0.vlm"},
	{0xA8CC19CCDF060FFEULL, "6117108_m.fxe"},
	{0xA8CC321A3ADC6220ULL, "6075163_m.fxe"},
	{0xA8CC93473816BBFFULL, "smstest.mao"},
	{0xA8CCD3809262DBB1ULL, "vis110_commoner3.utc"},
	{0xA8CCF6FD60AA1125ULL, "6043093_m.fxe"},
	{0xA8CD4764E41F3C39ULL, "6085997_f.fxe"},
	{0xA8CD91359D3A7C2EULL, "6170554_m.fxe"},
	{0xA8CD9F3EA956D709ULL, "49429999.ogg"},
	{0xA8CDDDBA1E82D5DCULL, "vxw_bonepit_25k_0_0.vlm"},
	{0xA8CE72D961BBCAF6ULL, "vxlt_warehouse_n_12i_0_0.vlm"},
	{0xA8CF279D20C63E18ULL, "vxht_mansion2b_1v6_0_0.vlm"},
	{0xA8CF3633B2C476C4ULL, "vxht_brothel_n_5ke_0_0.vlm"},
	{0xA8CF435E202B0DF1ULL, "w_bonepitmining_a2_164_0.mmh"},
	{0xA8CF4F5E202B2192ULL, "w_bonepitmining_a2_164_0.mao"},
	{0xA8CF5D5E202B39AFULL, "w_bonepitmining_a2_164_0.msh"},
	{0xA8CFB3C50EF9083EULL, "vxht_bartrand_n_5o3_0_0.vlm"},
	{0xA8CFED4C9B359D9CULL, "vxht_0_hightown_a2_a80_0_0.vlm"},
	{0xA8D0DFA21C671460ULL, "852839526.ogg"},
	{0xA8D1939CAB5E0F4AULL, "vxca_cavern1_haunted_ck0_0_0.vlm"},
	{0xA8D1B974F9C0B3E3ULL, "vxht_0_hightown_n_c7w_0_0.vlm"},
	{0xA8D1EFCCAC2A529DULL, "t3_shd_drb.tnt"},
	{0xA8D20C36D4913ACEULL, "vxca_caverns1_e8j_0_0.vlm"},
	{0xA8D25A6661088BC7ULL, "vxca_caverns1_mines_fm6_0_0.vlm"},
	{0xA8D267E541D3BDEDULL, "vxlt_0_lowtown_a2_hhd_0_0.vlm"},
	{0xA8D281486FF20BA2ULL, "577616303.ogg"},
	{0xA8D2824FAD6BC485ULL, "457492961.ogg"},
	{0xA8D2953C293CC75DULL, "vxlt_0_lowtown_a2_n_c_ht1_0_0.vlm"},
	{0xA8D313B180104F4CULL, "vxw_sundermount_2o2_0_0.vlm"},
	{0xA8D33D9F40857CB6ULL, "6103350_m.fxe"},
	{0xA8D3896B79992B74ULL, "llac_lt_darktown_-1_46.dds"},
	{0xA8D41B1A07A0D259ULL, "gen00fl_aveline6113658.cl"},
	{0xA8D42C6A31A2921CULL, "6023821_m.fxe"},
	{0xA8D54FA2CAC98300ULL, "art\\vfx\\combat\\vfx_r_gn2_hbomb_act_c.anb"},
	{0xA8D559312E8F8852ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\runpausefl.evt"},
	{0xA8D633F8153C088FULL, "6080630_m.fxe"},
	{0xA8D6DC7FAEF329EAULL, "w_bonepitdestroyed_a3_122_0.mao"},
	{0xA8D6E07FAEF33039ULL, "w_bonepitdestroyed_a3_122_0.mmh"},
	{0xA8D6EA7FAEF34127ULL, "w_bonepitdestroyed_a3_122_0.msh"},
	{0xA8D6F38237DE4D33ULL, "vxlt_backalley_g_1n0_0_0.vlm"},
	{0xA8D75064EA3AB7E3ULL, "lt_0_lowtown_a3_n_c_25_0.rml"},
	{0xA8D794EADFC4C384ULL, "one280cr_iwan.ncs"},
	{0xA8D7A74EA5DA05ECULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sider.msh"},
	{0xA8D7AD4EA5DA0FDEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sider.mmh"},
	{0xA8D7B94EA5DA2475ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sider.mao"},
	{0xA8D7CF0F14988C6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\dropthrust.evt"},
	{0xA8D834436DC92EAAULL, "spt_leafmeshes.vsh.fxo"},
	{0xA8D8B85FA47FB747ULL, "qun110tr_gate_guard_on.ncs"},
	{0xA8D8C8519003DB5AULL, "6084868_m.fxe"},
	{0xA8D9E157A9ED604DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_meat_n.dds"},
	{0xA8DA363B4FC182DCULL, "25137783.ogg"},
	{0xA8DAC4BC990CD9CEULL, "vxlt_0_lowtown_docks_b5b_0_1.vlm"},
	{0xA8DAEC45DAD903C1ULL, "mag311_keran6041390.cl"},
	{0xA8DB14BB5A26AB1EULL, "890408647.ogg"},
	{0xA8DB151741411B09ULL, "lit_up_05.mao"},
	{0xA8DB2D395EDEFCBEULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.impa.evt"},
	{0xA8DBAE535AF53DD7ULL, "vxdr_0_deeproads_ev2_0_0.vlm"},
	{0xA8DBC409FA0A7CCFULL, "6113924_f.fxe"},
	{0xA8DC607EF60C2FA7ULL, "278481739.ogg"},
	{0xA8DCDA7EC0D0110BULL, "751421928.ogg"},
	{0xA8DCF13071345EA6ULL, "dae000_flemeth6000956.cl"},
	{0xA8DCF3AA14274A97ULL, "vxht_keep_n_6ah_0_0.vlm"},
	{0xA8DD22D54A9146C0ULL, "305983884.ogg"},
	{0xA8DD7C490379BC61ULL, "6131381_m.fxe"},
	{0xA8DDE02F11946D80ULL, "test_petert.ncs"},
	{0xA8DDE80F149D222CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\dropthrust.gad"},
	{0xA8DE64FAC9959D58ULL, "vxdr_primevaltomb_8r6_0_0.vlm"},
	{0xA8DE9D1A57C7FCC3ULL, "w_bonepit_37_0.rml"},
	{0xA8DEB8BA28821D56ULL, "482275730.ogg"},
	{0xA8E00EB4CDD9E108ULL, "gen00fl_aveline6105441.cl"},
	{0xA8E02CEB717D1549ULL, "follower_banter6164502.cl"},
	{0xA8E0392584B493CAULL, "vxht_keep_n_70g_0_0.vlm"},
	{0xA8E1286190DA0950ULL, "6135765_f.fxe"},
	{0xA8E1437716BFEF4FULL, "vxlt_darktown_2v5_0_0.vlm"},
	{0xA8E155C65EF1135AULL, "art\\levels\\4cinematics\\ferelden\\prp_bookofinquis.mmh"},
	{0xA8E167C65EF131A0ULL, "art\\levels\\4cinematics\\ferelden\\prp_bookofinquis.msh"},
	{0xA8E1858E720667CCULL, "vxw_sundermount_2p3_0_0.vlm"},
	{0xA8E1BADF0246BBDDULL, "hlac_lt_0_lowtown_a3_n_c_65536_10z.dds"},
	{0xA8E1D8D28285EDB5ULL, "gen00fl_carver6102673.cl"},
	{0xA8E22A5F99E63648ULL, "6202134_m.fxe"},
	{0xA8E244395EE542E9ULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.impa.gad"},
	{0xA8E28073162D5100ULL, "vxlt_backalley_n_1wy_0_0.vlm"},
	{0xA8E30994911EDA3AULL, "vxht_0_hightown_a2_avh_0_0.vlm"},
	{0xA8E39A48FAECE7A1ULL, "vxw_blightlands_ln_0_0.vlm"},
	{0xA8E3C5114A2A76CDULL, "vxw_encounter3_f9_0_1.vlm"},
	{0xA8E4E70F14A33F2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\dropthrust.ani"},
	{0xA8E55900CB0CC083ULL, "723542273.ogg"},
	{0xA8E5655F29D465B3ULL, "vxlt_backalley_g_1k3_0_0.vlm"},
	{0xA8E5777C3604A3E4ULL, "vxlt_blackemporium_14b_0_0.vlm"},
	{0xA8E6520B2DE2A8D7ULL, "vxlt_warehouse_18q_0_0.vlm"},
	{0xA8E67641F04893EAULL, "1062682876.ogg"},
	{0xA8E67E1FB2C37809ULL, "vxlt_0_lowtown_hag_0_0.vlm"},
	{0xA8E690759B0ACD64ULL, "vxw_encounter3_4s_0_0.vlm"},
	{0xA8E69EEB5094C921ULL, "plt_prm_cod_seeker_shield.plo"},
	{0xA8E6FBF2648FA19AULL, "fx_ice_splat3.mao"},
	{0xA8E7AF78DEDAD359ULL, "868558694.ogg"},
	{0xA8E85FAB272362A0ULL, "66179197.ogg"},
	{0xA8E8786C8581D81FULL, "llac_dr_0_deeproads_5845_9.dds"},
	{0xA8E8C1F1301C6044ULL, "vxlt_0_lowtown_docks_b6p_0_0.vlm"},
	{0xA8E9047C182FA168ULL, "art\\characters\\creatures\\ragedemon\\animation\\blendtree\\ragedemon_movement.anb"},
	{0xA8E915395EEB1137ULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.impa.ani"},
	{0xA8E9181327DA7D76ULL, "196804446.ogg"},
	{0xA8E94AFEA6847D77ULL, "hlac_w_blightlands_327681_0y.dds"},
	{0xA8E9583A22181DB2ULL, "hlac_w_bonepitmining_a2_262146_0y.dds"},
	{0xA8E99717DE4F5319ULL, "6035867_m.fxe"},
	{0xA8E9F7DB09B472F7ULL, "6117417_m.fxe"},
	{0xA8EA66F411C0EFCAULL, "gen00fl_anders6119324.cl"},
	{0xA8EA7E805A600B88ULL, "ht_0_hightown_n_27_0.rml"},
	{0xA8EAF698964E8B3CULL, "gen00fl_fenris6103330.cl"},
	{0xA8EB4C4E58AF24BFULL, "vxdr_primevaltomb_6qk_0_1.vlm"},
	{0xA8EB535682EB856AULL, "fen110_anso6095034.cl"},
	{0xA8EB6E7230FEE9F8ULL, "6096807_m.fxe"},
	{0xA8EB8085E2572FC4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_keepthin.msh"},
	{0xA8EB9285E2574E2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_keepthin.mao"},
	{0xA8EB9685E2575566ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_keepthin.mmh"},
	{0xA8EC3B28165BAFB6ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh01a.msh"},
	{0xA8EC4528165BC0B4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh01a.mmh"},
	{0xA8EC62C0A836E945ULL, "conversation_emotions.gda"},
	{0xA8ED832D9EE9E83CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_lgtc_0s.dds"},
	{0xA8EDD2BF92D8DC16ULL, "442499051.ogg"},
	{0xA8EE150FA3958132ULL, "default_warrior_female.utc"},
	{0xA8EE25533F8803C1ULL, "162526713.ogg"},
	{0xA8EE9D2EDDACE4E3ULL, "vxht_0_hightown_n_c9y_0_0.vlm"},
	{0xA8EEB1A585B5D642ULL, "vxlt_0_lowtown_a3_n_c_gce_0_0.vlm"},
	{0xA8EEBF798A0A4C07ULL, "781365496.ogg"},
	{0xA8EEE980927FE7ECULL, "vis110_commoner3.cnv"},
	{0xA8EEE9E388DE24F1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofoc.msh"},
	{0xA8EEF7E388DE3C0CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofoc.mao"},
	{0xA8EF03E388DE50AFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofoc.mmh"},
	{0xA8EF37AFD97FE2C2ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheelb.phy"},
	{0xA8EF6FA5E9CD28F2ULL, "art\\vfx\\combat\\vfx_v_fir_single_att_c.mmh"},
	{0xA8EFA2EA4D709AB1ULL, "dre000_sibling_has_blight6036545.cl"},
	{0xA8EFB55408B607CFULL, "vxlt_darktown_2y4_0_0.vlm"},
	{0xA8EFBDDCA658C8ADULL, "vxlt_0_lowtown_n_e37_0_0.vlm"},
	{0xA8EFF5DD8C3664EFULL, "985287480.ogg"},
	{0xA8EFF76B63FC804CULL, "vxw_sundermount_2u0_0_0.vlm"},
	{0xA8F0753F75D5CDC9ULL, "vxlt_warehouse_n_be_0_0.vlm"},
	{0xA8F12FF492D8D94FULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_c.mmh"},
	{0xA8F17B718314F2BAULL, "vxht_0_hightown_a2_aqi_0_0.vlm"},
	{0xA8F1C53B5CAFB8BBULL, "vxht_0_hightown_n_av2_0_0.vlm"},
	{0xA8F1CB809281EA1CULL, "vis110_commoner3.bnk"},
	{0xA8F1D59C24A2714FULL, "896608677.ogg"},
	{0xA8F1D61325B48AA5ULL, "677397383.ogg"},
	{0xA8F20C25ECE30021ULL, "vxw_blightlands_qm_0_0.vlm"},
	{0xA8F2C867354AE706ULL, "dae000_epic_merchant_2.cnv"},
	{0xA8F33A35CE6BA2E1ULL, "6117799_f.fxe"},
	{0xA8F39E9B41184BBFULL, "145772854.ogg"},
	{0xA8F3D23D81094548ULL, "zz_followers_banter6159309.cl"},
	{0xA8F502528D00E5E4ULL, "vxw_encounter3_1p_0_0.vlm"},
	{0xA8F537C4443C35ACULL, "vxht_mansion2_2ja_0_0.vlm"},
	{0xA8F5471084B5CB5EULL, "924892231.ogg"},
	{0xA8F569BB2D6B6937ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall.mmh"},
	{0xA8F56DBB2D6B7044ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall.mao"},
	{0xA8F57FBB2D6B8EA9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall.msh"},
	{0xA8F58D85E25F9A31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_keepthin.phy"},
	{0xA8F5B93CF0B206FDULL, "lt_0_lowtown_a2_n_35_0.anb"},
	{0xA8F5CA67354D20C6ULL, "dae000_epic_merchant_2.bnk"},
	{0xA8F68576AACC9FD5ULL, "dae000_bartrand6033989.cl"},
	{0xA8F776BFF3F7F760ULL, "art\\levels\\ztools\\default\\generic\\0wi_basic.phy"},
	{0xA8F79BB597D72615ULL, "6067824_m.fxe"},
	{0xA8F928DC76CD548EULL, "6083469_m.fxe"},
	{0xA8F93287F5439E38ULL, "6092418_m.fxe"},
	{0xA8FA2998D0F6008FULL, "273592731.ogg"},
	{0xA8FB44505346EAF8ULL, "vxht_mansion2_1sq_0_0.vlm"},
	{0xA8FB68228354421FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven.phy"},
	{0xA8FBBB57BF0BEF74ULL, "6044659_m.fxe"},
	{0xA8FC601EDE00BDFCULL, "740957332.ogg"},
	{0xA8FC7DD7CC8D7C38ULL, "901457686.ogg"},
	{0xA8FC86E6DBD5480DULL, "520928638.ogg"},
	{0xA8FCE7F3CB3C2E7BULL, "6169632_m.fxe"},
	{0xA8FD0A78C3E7413DULL, "6083084_m.fxe"},
	{0xA8FD548D195960FAULL, "vxca_caverns1_k2r_0_0.vlm"},
	{0xA8FD5DF28C582E5EULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_com_0d.dds"},
	{0xA8FD84F905F4113AULL, "hlac_ca_cavern1_haunted_24422_7y.dds"},
	{0xA8FDC28961558D79ULL, "zz_facefx_test6088480.cl"},
	{0xA8FE81C6CCA0D67AULL, "vxlt_0_lowtown_a3_n_c_equ_0_0.vlm"},
	{0xA8FEB9A6B7F2EF81ULL, "vxw_bonepit_209_0_0.vlm"},
	{0xA8FEC396D476568AULL, "plt_zz_rdrpt.plo"},
	{0xA8FF114BC1698C73ULL, "58475802.ogg"},
	{0xA8FF7EFB12397E7CULL, "6162425_m.fxe"},
	{0xA8FFE19BDE6E83B3ULL, "6040699_f.fxe"},
	{0xA900BF609ADCDAB1ULL, "llac_lt_0_lowtown_a2_n_131070_1.dds"},
	{0xA900FA77DFED5CF3ULL, "392692505.ogg"},
	{0xA901A518876E90DFULL, "618319885.ogg"},
	{0xA901C3C88AC57E2DULL, "hla_ht_0_hightown_n_131070_0x.dds"},
	{0xA901C55F5C9D5722ULL, "6092153_m.fxe"},
	{0xA90284168C77855EULL, "6164436_m.fxe"},
	{0xA902A6AC81C9EAE8ULL, "x_narrativeroom.arl"},
	{0xA903279E0D0EE036ULL, "6125274_f.fxe"},
	{0xA9032FD88A427C43ULL, "6149461_m.fxe"},
	{0xA9037FCE1F08FCF7ULL, "210241692.ogg"},
	{0xA903A9A136324E2CULL, "vxht_mansion2_2wf_0_0.vlm"},
	{0xA903C8735AFC918DULL, "vxw_bonepitdestroyed_a3_kt_0_0.vlm"},
	{0xA903D3F349A179ECULL, "vxlt_hangedman_n_55_0_0.vlm"},
	{0xA90453EC224E06EBULL, "one350_variel6085394.cl"},
	{0xA904D692FCF5D07FULL, "follower_banter6161046.cl"},
	{0xA9050D36F28FC470ULL, "vxlt_darktown_5ls_0_0.vlm"},
	{0xA9059132AA700FCDULL, "lt_0_lowtown_docks_a3_n_122_0.msh"},
	{0xA9059F32AA702700ULL, "lt_0_lowtown_docks_a3_n_122_0.mao"},
	{0xA905A332AA702E53ULL, "lt_0_lowtown_docks_a3_n_122_0.mmh"},
	{0xA905EA6C1554C388ULL, "vxlt_hangedman_19e_0_0.vlm"},
	{0xA907042C441935D7ULL, "sky_sky2_p2.psh.fxo.dx11.assembly"},
	{0xA90723C275B1E850ULL, "art\\vfx\\materials\\vfx_d_shockwave_add.mao"},
	{0xA907D6AF213420E2ULL, "vxdr_primevalentrance_2zj_0_0.vlm"},
	{0xA907EBEA374B1A41ULL, "6147753_m.fxe"},
	{0xA90807696523FFF7ULL, "vxlt_0_lowtown_docks_cw1_0_0.vlm"},
	{0xA90812C230A2236DULL, "conversation_shake_settings.gda"},
	{0xA908A9318DA0CD88ULL, "vxht_mansion_n_50c_0_0.vlm"},
	{0xA9097F6184502FAFULL, "320159722.ogg"},
	{0xA909F54D328C306CULL, "llac_ca_caverns1_mines_24424_0.dds"},
	{0xA90A66103836EE9BULL, "dae300_cullen6140586.cl"},
	{0xA90A92300721F1F1ULL, "w_encounter3_154_0.msh"},
	{0xA90AA0300722090CULL, "w_encounter3_154_0.mao"},
	{0xA90AAC3007221DAFULL, "w_encounter3_154_0.mmh"},
	{0xA90ADE1D03146AEDULL, "951864527.ogg"},
	{0xA90BA49EA14356FBULL, "804810583.ogg"},
	{0xA90BCCAEF8D19FD6ULL, "gen00fl_anders6116080.cl"},
	{0xA90BCDB83D9BF39EULL, "plt_cod_itm_urzaras_tooth.plo"},
	{0xA90BE029921FC166ULL, "mcr370_varric6098333.cl"},
	{0xA90C69228362B724ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven.mmh"},
	{0xA90C6D228362BE77ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven.mao"},
	{0xA90C7F228362DCC6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven.msh"},
	{0xA90C8979C68963E3ULL, "vxht_0_hightown_a2_c1b_0_0.vlm"},
	{0xA90C9D34A3AC90B9ULL, "vxlt_0_lowtown_a2_n_c_d8u_0_0.vlm"},
	{0xA90CE0A6D60B990EULL, "6059970_m.fxe"},
	{0xA90D040C01FB6044ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_meda_t.dds"},
	{0xA90DC7A58030B871ULL, "vxdr_primevaltomb_4zv_0_1.vlm"},
	{0xA90DD609EC0F9A80ULL, "vxlt_backalley_n_1xd_0_0.vlm"},
	{0xA90DE8201C3F93E6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_medj_0d.dds"},
	{0xA90E09E73A2A2D1CULL, "aip_follower_bethany.gda"},
	{0xA90E33EDA7A90227ULL, "fen110_anso6170062.cl"},
	{0xA90E6B80A0BB36F1ULL, "vxht_brothel_n_6b8_0_0.vlm"},
	{0xA90F05F7F63BD02DULL, "t3_shd_ibk.tnt"},
	{0xA90F4DB2F975B4D9ULL, "isa311_castillon6092125.cl"},
	{0xA90F7ADDFEDBF3F9ULL, "character_character_p103.psh.fxo.dx11"},
	{0xA91017F5E52574D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext_override.ani"},
	{0xA9101CC4A712B48CULL, "prm000im_cliffs_tome.uti"},
	{0xA910984170795E75ULL, "14603380.ogg"},
	{0xA910BAF5FFB6AF33ULL, "vxlt_backalley_g_1z4_0_0.vlm"},
	{0xA910CE5AD42AC08BULL, "vxht_chantry_2ti_0_1.vlm"},
	{0xA910FE0BF6D0E9E9ULL, "6075208_f.fxe"},
	{0xA911600BEA71DAC5ULL, "vxw_bonepitdestroyed_a3_1x6_0_0.vlm"},
	{0xA9118B1A57F337DAULL, "w_bonepit_37_0.msh"},
	{0xA911991A57F34F13ULL, "w_bonepit_37_0.mao"},
	{0xA9119D1A57F35620ULL, "w_bonepit_37_0.mmh"},
	{0xA911A1B57C3894C3ULL, "6142561_m.fxe"},
	{0xA911B96AB49B4048ULL, "vxw_bonepit_7a_0_0.vlm"},
	{0xA9121B7E282866ACULL, "vxht_mansion2_2pg_0_0.vlm"},
	{0xA912981D0A7E4CA5ULL, "vxlt_warehouse_n_157_0_0.vlm"},
	{0xA913114891ADE2F8ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ds001a_0s.dds"},
	{0xA9132CA8607D75A8ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.br_atk_v3.evt"},
	{0xA9133FA31873F4F8ULL, "mcr370cr_templar_messenger.utc"},
	{0xA91406E8DD335099ULL, "gen_im_trs_dal_arc_sbw_r0.uti"},
	{0xA91412AFD99EC076ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheelb.mao"},
	{0xA9141EAFD99ED415ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheelb.mmh"},
	{0xA91420AFD99ED78BULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesewheelb.msh"},
	{0xA9142C69A0B3A3ACULL, "zz_followers_banter6159213.cl"},
	{0xA91454631F26BB0FULL, "6042805_m.fxe"},
	{0xA91459F83217B2CBULL, "6066322_m.fxe"},
	{0xA9148692AB36801FULL, "6068359_m.fxe"},
	{0xA914B2CD5DCE3C7DULL, "llac_lt_0_lowtown_docks_a3_n_-131075_3.dds"},
	{0xA914BB4D52D89A78ULL, "179122965.ogg"},
	{0xA914D3349F162137ULL, "vxlt_blackemporium_th_0_0.vlm"},
	{0xA915B8F3E81D6D0FULL, "gen00flst_post_night_terror.stg"},
	{0xA9161FBB7EE89BFBULL, "static_static_p40.psh.fxo.dx11"},
	{0xA91669C330A33187ULL, "vxlt_backalley_g_1jh_0_0.vlm"},
	{0xA9167946571A1877ULL, "vxlt_0_lowtown_docks_cp0_0_0.vlm"},
	{0xA916CC7D3831683FULL, "hlac_lt_gallowstemplar_f_2394_4z.dds"},
	{0xA916F2F5E52B541BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext_override.gad"},
	{0xA9176EA5E9EE7899ULL, "art\\vfx\\combat\\vfx_v_fir_single_att_c.anb"},
	{0xA917844EA60FAE79ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sider.phy"},
	{0xA917A54EC97B91B2ULL, "6068747_m.fxe"},
	{0xA91819905C8CEF06ULL, "6096683_m.fxe"},
	{0xA9183364EA721D5FULL, "lt_0_lowtown_a3_n_c_25_0.anb"},
	{0xA91889F08A7123ADULL, "6124141_m.fxe"},
	{0xA9188BE253FEE904ULL, "487124558.ogg"},
	{0xA918A6BE2820687CULL, "6031354_m.fxe"},
	{0xA918A8361BD56B0FULL, "6041800_m.fxe"},
	{0xA918C9C23648E6DFULL, "vxlt_gallowstemplar_3r0_0_0.vlm"},
	{0xA9192636764DAD62ULL, "vxht_0_hightown_n_apu_0_0.vlm"},
	{0xA9199C0CF45BD2F7ULL, "6127985_f.fxe"},
	{0xA919BB264E790604ULL, "59191441.ogg"},
	{0xA919D16930797D5BULL, "mcr310_confused_guard6077061.cl"},
	{0xA91A0AF13346A540ULL, "gen00fl_aveline6062171.cl"},
	{0xA91A18F23E259D7BULL, "w_sundermount_64_0_lc.mmh"},
	{0xA91A30F492FBDF8CULL, "art\\vfx\\combat\\vfx_glb_loot_shimmer_c.anb"},
	{0xA91A58CC7CA9E0CDULL, "6108550_m.fxe"},
	{0xA91A89EBB951A495ULL, "rdrev_item_acquired.ncs"},
	{0xA91A9B7B7B39ACFEULL, "hlac_w_sundermount_3_0z.dds"},
	{0xA91B212E1639657DULL, "hlac_ca_caverns1_mines_24421_2z.dds"},
	{0xA91B4F50E74F695CULL, "6147785_m.fxe"},
	{0xA91B99D10F32F5AAULL, "hlac_lt_0_lowtown_docks_a3_-131075_5x.dds"},
	{0xA91BCAD64BB81749ULL, "vxlt_warehouse_n_sj_0_0.vlm"},
	{0xA91C919B054E3A6AULL, "dae300_vashothlead6062818.cl"},
	{0xA91CC7F86A5A4B57ULL, "lt_0_lowtown_docks_26_0_ll.mmh"},
	{0xA91CF7323C05182AULL, "hlac_lt_undercityencounter_-1_2x.dds"},
	{0xA91D2547018D5676ULL, "vxht_0_hightown_a2_9zh_0_0.vlm"},
	{0xA91D93BFF418F7F5ULL, "art\\levels\\ztools\\default\\generic\\0wi_basic.msh"},
	{0xA91D95BFF418FB6BULL, "art\\levels\\ztools\\default\\generic\\0wi_basic.mmh"},
	{0xA91DA1BFF4190F08ULL, "art\\levels\\ztools\\default\\generic\\0wi_basic.mao"},
	{0xA91E0FF5E531A400ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_ext_override.evt"},
	{0xA91E22B5652E0CE8ULL, "143837471.ogg"},
	{0xA91EF96D870D2115ULL, "hlac_lt_gallowscourtyard_n_-131071_3y.dds"},
	{0xA91F2BC9FB0EAE09ULL, "153033723.ogg"},
	{0xA91F3F3576896975ULL, "1067797026.ogg"},
	{0xA91F6ADE476744BFULL, "hlac_lt_0_lowtown_docks_a3_131073_0y.dds"},
	{0xA91F70079B9F6535ULL, "mer221_sabin6047524.cl"},
	{0xA91F801A57FF623FULL, "w_bonepit_37_0.anb"},
	{0xA91FFE14182BCFC5ULL, "6024115_m.fxe"},
	{0xA92009167CBD217DULL, "72633431.ogg"},
	{0xA9201F000D4B45DAULL, "6153386_m.fxe"},
	{0xA92035AE9F5510C3ULL, "vxht_0_hightown_n_c6y_0_1.vlm"},
	{0xA9206C2DC6947819ULL, "vxw_woundedcoast_fr_0_0.vlm"},
	{0xA92152442E996A32ULL, "follower_barks6153195.cl"},
	{0xA9217B7B269A2778ULL, "6086972_m.fxe"},
	{0xA921903DFBA2DE37ULL, "fen210_fenris6108732.cl"},
	{0xA921955471FD60AFULL, "505519055.ogg"},
	{0xA92199669233ED3CULL, "vxht_0_hightown_n_c0i_0_0.vlm"},
	{0xA921D30A7489D04DULL, "hlac_w_sundermount_-262146_0x.dds"},
	{0xA9224EEF744CB35BULL, "dae000_gamlen6124245.cl"},
	{0xA9225B48F92B119CULL, "6030432_m.fxe"},
	{0xA922874B1CA738A4ULL, "art\\vfx\\combat\\vfx_gib_drkspwn_bld_p.anb"},
	{0xA9228916291A0760ULL, "247029879.ogg"},
	{0xA922D75AAFA049FDULL, "war_clearfrontarc.dds"},
	{0xA9230AFFE4BAC726ULL, "zz_dae_debug6029309.cl"},
	{0xA92318E83D21AF15ULL, "cd_mag291_demon_marethari.bnk"},
	{0xA923D2AE935ACA41ULL, "776700880.ogg"},
	{0xA9241CEC4AB22481ULL, "hlac_lt_0_lowtown_a2_65536_12x.dds"},
	{0xA92474D9827D0D0AULL, "hlac_lt_0_lowtown_docks_a3_-65535_0z.dds"},
	{0xA9248F0613F8FFB7ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_024a_0n.dds"},
	{0xA92499325BC553B0ULL, "em_amb_merchant_rare.mor"},
	{0xA924DBA022994A07ULL, "vxlt_backalley_g_1wk_0_0.vlm"},
	{0xA9256583B413A7A4ULL, "vfx_chunkyparticle_p3.vsh.fxo.dx11.assembly"},
	{0xA925C2E53B71F0D4ULL, "681745900.ogg"},
	{0xA925C6B0ED49049AULL, "vxca_cavern1_haunted_cjs_0_0.vlm"},
	{0xA9275AB80415AEACULL, "zz_dae_debug6050004.cl"},
	{0xA92798136843C5E2ULL, "vxht_0_hightown_n_auv_0_0.vlm"},
	{0xA927C1A2076FAA2CULL, "vxca_cavern1_haunted_ih9_0_0.vlm"},
	{0xA928052A32EC6B92ULL, "vxlt_backalley_n_1n1_0_1.vlm"},
	{0xA9299BE1CFAE9C85ULL, "vxht_mansion2_1sh_0_0.vlm"},
	{0xA92A3CB33DAE2FC9ULL, "vxlt_warehouse_n_vi_0_0.vlm"},
	{0xA92A5BB70FF1C69AULL, "924493540.ogg"},
	{0xA92A78008625AE84ULL, "vxw_woundedcoast_1br_0_0.vlm"},
	{0xA92AE5807F5F24FDULL, "90146070.ogg"},
	{0xA92BF87032C04342ULL, "vxlt_0_lowtown_a2_ep3_0_1.vlm"},
	{0xA92C33A88E8139D3ULL, "145083086.ogg"},
	{0xA92C8B7654C2DC2DULL, "aip_desiredemon_base.gda"},
	{0xA92CAE0015445652ULL, "6166320_m.fxe"},
	{0xA92CB8F97544C65BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzarda_n.dds"},
	{0xA92D30691D854110ULL, "follower_banter6169524.cl"},
	{0xA92D568770B68B33ULL, "6062821_m.fxe"},
	{0xA92D5F5599090025ULL, "hlac_ht_keep_3736_0y.dds"},
	{0xA92D871DE4982538ULL, "w_bonepit_102_0.phy"},
	{0xA92DA55FA4C7D69EULL, "qun110tr_gate_guard_on.utt"},
	{0xA92E43C5CE5E0BC5ULL, "vxw_bonepitdestroyed_a3_1f4_0_0.vlm"},
	{0xA92E6557CF86FD6BULL, "fen110_fenris6118002.cl"},
	{0xA92E98EEFB5244EBULL, "514714319.ogg"},
	{0xA92EBA3CF0E1CDF5ULL, "lt_0_lowtown_a2_n_35_0.rml"},
	{0xA92ECC96D49F8981ULL, "plt_zz_rdrpt.nss"},
	{0xA92EDE0AB88A9099ULL, "vxw_woundedcoast_cq_0_0.vlm"},
	{0xA92F062D4C6DD89DULL, "vis121_saemus6118417.cl"},
	{0xA92F19FBE0DC7B06ULL, "664420303.ogg"},
	{0xA92F4D1E31A155FCULL, "vxht_mansion_n_6lx_0_0.vlm"},
	{0xA93008F54B022D8EULL, "fen210_fenris6109440.cl"},
	{0xA9302898452C4143ULL, "6208576_f.fxe"},
	{0xA930584D236350BFULL, "def_rock2_s.dds"},
	{0xA9308EEF44F0C6C0ULL, "lgt000cr_hate_gen_racist.utc"},
	{0xA930A5E156802725ULL, "gen00fl_anders6119505.cl"},
	{0xA930A64B1CB3AAA7ULL, "art\\vfx\\combat\\vfx_gib_drkspwn_bld_p.mmh"},
	{0xA930C166EB1D5D60ULL, "vxlt_gallowscourtyard_4uw_0_0.vlm"},
	{0xA930D0F1E313602AULL, "6135782_m.fxe"},
	{0xA9311DD2B7DEEA94ULL, "mcr310_argument6074382.cl"},
	{0xA93165BE0C1B84DAULL, "709282776.ogg"},
	{0xA931BD69F777F818ULL, "hlac_lt_0_lowtown_65537_5z.dds"},
	{0xA932201DA2FF9807ULL, "vxw_bonepitmining_a2_1m1_0_0.vlm"},
	{0xA932C914F2DB995BULL, "hightown_market_guard_sta6169595.cl"},
	{0xA9333E924F360786ULL, "390687346.ogg"},
	{0xA93348619355616AULL, "llac_lt_0_lowtown_n_0_2.dds"},
	{0xA933740E412094ECULL, "435083926.ogg"},
	{0xA934A2A815E76763ULL, "vxca_caverns2_715_0_0.vlm"},
	{0xA935B5C15BA391ABULL, "llac_ca_caverns2_5802_39.dds"},
	{0xA935D5C34B26694CULL, "vrc321ip_spooky6_obj_1.utp"},
	{0xA9364D3B312D4B5BULL, "dre180ar_bodahnambush_st.stg"},
	{0xA936D656E5BF51D0ULL, "265850053.ogg"},
	{0xA936E68939D8A233ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_2hands.ani"},
	{0xA937426A03D6E827ULL, "vxlt_0_lowtown_ey0_0_0.vlm"},
	{0xA937BC38C565F91EULL, "vxw_sundermount_2lm_0_0.vlm"},
	{0xA937DEADA6962289ULL, "1026857611.ogg"},
	{0xA9381EE9137620CAULL, "vxlt_0_lowtown_a2_n_c_e8c_0_0.vlm"},
	{0xA938E17FAF45FC26ULL, "w_bonepitdestroyed_a3_122_0.phy"},
	{0xA938E9DD781BC704ULL, "vxw_woundedcoast_1ms_0_0.vlm"},
	{0xA9397E94392710E5ULL, "6124682_m.fxe"},
	{0xA93AA2C21D8DD246ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_exit.ani"},
	{0xA93B357C598802AAULL, "vxht_mansion2b_3ce_0_0.vlm"},
	{0xA93B44DD71E6D796ULL, "dae000_keep_guard_2_sta6163826.cl"},
	{0xA93B91BDAC269E32ULL, "vxca_caverns2prc_9v2_0_0.vlm"},
	{0xA93BC86462ACA3B3ULL, "6069921_m.fxe"},
	{0xA93BCF48C4B4CF9AULL, "vxht_0_hightown_a2_cbf_0_0.vlm"},
	{0xA93BE2BD61435B1AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_player_chant_ex.gad"},
	{0xA93C4BA61B6FC76EULL, "hlac_ht_mansion2_818_1y.dds"},
	{0xA93CB55E7E229EE1ULL, "gen00fl_isabela6135707.cl"},
	{0xA93CD13E2D8B132EULL, "61766550.ogg"},
	{0xA93CF009663658CBULL, "vxlt_gallowstemplar_n_378_0_0.vlm"},
	{0xA93D47B961EA64AEULL, "mag110_samson6053548.cl"},
	{0xA93D4FE7AA80A919ULL, "vxw_woundedcoast_xp_0_0.vlm"},
	{0xA93DF60655C6CFBDULL, "vxw_woundedcoast_4x_0_0.vlm"},
	{0xA93E21FFE3B81642ULL, "6101352_m.fxe"},
	{0xA93E710E1A777C61ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_f_pcopeningmage.ani"},
	{0xA93E7985D6AC3D40ULL, "vxw_woundedcoast_17y_0_0.vlm"},
	{0xA93F12645A0D82CBULL, "vxlt_gallowstemplar_f_3s4_0_0.vlm"},
	{0xA93F5D4B4389DC6BULL, "500824534.ogg"},
	{0xA93FBA2191417430ULL, "56319343.ogg"},
	{0xA940560FD08ACC81ULL, "vxlt_backalley_n_1fw_0_0.vlm"},
	{0xA9408952E20E52D5ULL, "6071085_m.fxe"},
	{0xA94091FA94F5B087ULL, "vxw_bonepitmining_a2_1j0_0_0.vlm"},
	{0xA941BF5A06857B07ULL, "vxlt_backalley_g_1ye_0_0.vlm"},
	{0xA942F0EDF396D63DULL, "6048451_f.fxe"},
	{0xA9437EC8E847E2C2ULL, "126516807.ogg"},
	{0xA9439EF4B80F750AULL, "46433927.ogg"},
	{0xA944DE8939E4D26AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_2hands.evt"},
	{0xA9457C0E1A7DAD03ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\player_f_pcopeningmage.gad"},
	{0xA945EBB2B59EEC3AULL, "vxht_brothel_n_6g0_0_0.vlm"},
	{0xA9461545EA1D3F35ULL, "6097867_m.fxe"},
	{0xA9462E15B75C119EULL, "vxw_sundermount_2qn_0_0.vlm"},
	{0xA9471AD5C19EB5C8ULL, "vxht_keep_n_3fs_0_0.vlm"},
	{0xA947407AC042E447ULL, "dae000cr_grand_cleric6187921.cl"},
	{0xA9475BBA6A11DF84ULL, "vxw_woundedcoast_1hp_0_0.vlm"},
	{0xA947D7A16A05248AULL, "hlac_lt_0_lowtown_docks_a3_-196610_4y.dds"},
	{0xA948442816AA1DAFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh01a.phy"},
	{0xA9488AC21D99E6B7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_exit.evt"},
	{0xA948A9ECFD38E8D1ULL, "lt_0_lowtown_docks_n_20_0_lc.mmh"},
	{0xA948F5E7F11951F0ULL, "6071703_m.fxe"},
	{0xA9492AB722F571F7ULL, "655661426.ogg"},
	{0xA949325C15704A4AULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_blueglowworms_n.dds"},
	{0xA94963D7F8BBD308ULL, "plt_cod_ltr_isabelanote.plo"},
	{0xA949A7594B7E1B2AULL, "vxht_mansion2b_3ff_0_0.vlm"},
	{0xA949AD7C1A2DB0ADULL, "623622659.ogg"},
	{0xA949D89A9477AD1EULL, "dae000_gamlen6121568.cl"},
	{0xA949E54478CEF1D8ULL, "vxht_0_hightown_a2_c5a_0_0.vlm"},
	{0xA949FE4B4CB93F82ULL, "6118798_m.fxe"},
	{0xA94A0382D18A1CE2ULL, "art\\vfx\\combat\\vfx_w_rev_devour_imp_c.impa.gad"},
	{0xA94ADBA99146FB9EULL, "6174947_m.fxe"},
	{0xA94B277FB24A3CC5ULL, "vxw_bonepitdestroyed_a3_1l2_0_0.vlm"},
	{0xA94B87BA3EF21225ULL, "744292537.ogg"},
	{0xA94BBD8939EAB8DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_2hands.gad"},
	{0xA94BC1C49C76C199ULL, "vxw_woundedcoast_uw_0_0.vlm"},
	{0xA94C67E347BCE83DULL, "vxw_woundedcoast_1w_0_0.vlm"},
	{0xA94C6F5C432CF864ULL, "vxht_keep_n_6hx_0_0.vlm"},
	{0xA94CA0E77AB4E3FFULL, "6059401_m.fxe"},
	{0xA94CEB62C8A255C0ULL, "vxw_woundedcoast_12f_0_0.vlm"},
	{0xA94DAB102B03C7EBULL, "437234109.ogg"},
	{0xA94E3D3BCCAAC740ULL, "dae000_arishok6081171.cl"},
	{0xA94F3EE83D472A9FULL, "cd_mag291_demon_marethari.wbd"},
	{0xA94F8BC21DA00740ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_pose_exit.gad"},
	{0xA94F9670053651D7ULL, "gen00fl_anders6096488.cl"},
	{0xA94F9FE4845D9F54ULL, "hlac_lt_0_lowtown_n_131074_0x.dds"},
	{0xA9502682D18EC3D9ULL, "art\\vfx\\combat\\vfx_w_rev_devour_imp_c.impa.evt"},
	{0xA95035F3D02FF81CULL, "hlac_dr_primevalentrance_1726_11z.dds"},
	{0xA950C7A81B2ED6CDULL, "6170122_m.fxe"},
	{0xA950FDBD6155E500ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_player_chant_ex.ani"},
	{0xA951217FF2B4E995ULL, "734573455.ogg"},
	{0xA951282F5570087CULL, "gen00fl_aveline6140910.cl"},
	{0xA9519110F0F76307ULL, "dae000_brothel_gamlen6157084.cl"},
	{0xA952114543D7454DULL, "vxht_0_hightown_cba_0_0.vlm"},
	{0xA95211B1ABB35866ULL, "art\\characters\\creatures\\mubari\\animation\\p_sit.evt"},
	{0xA95224CB6946EF39ULL, "isa111_isabela6067121.cl"},
	{0xA952262A25A2797CULL, "vxlt_undercityencounter_u5_0_0.vlm"},
	{0xA95252ACEA616C2EULL, "vxlt_0_lowtown_docks_a3_n_d3b_0_1.vlm"},
	{0xA95267DCC4A9E07DULL, "hlac_ca_caverns1_24426_13z.dds"},
	{0xA952B7908851DFB0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_basic_1.mao"},
	{0xA952DB684897A2B9ULL, "vxw_sundermount_30b_0_1.vlm"},
	{0xA9542623E7C31927ULL, "vxlt_0_lowtown_eo2_0_0.vlm"},
	{0xA9549D1276A748C8ULL, "325162649.ogg"},
	{0xA9549FF2A9522A1EULL, "vxw_sundermount_2zo_0_0.vlm"},
	{0xA954B83249DB5438ULL, "6190903_m.fxe"},
	{0xA9553802C088608BULL, "mcr310_argument6074512.cl"},
	{0xA95568617F347F34ULL, "62928670.ogg"},
	{0xA955CD975C07F804ULL, "vxw_woundedcoast_1sq_0_0.vlm"},
	{0xA955DF46E39271DDULL, "dae000_meredith6166272.cl"},
	{0xA955E0E216C3F995ULL, "mer210_seldrek6078796.cl"},
	{0xA955EFF1682644F4ULL, "vxdr_0_deeproads_fic_0_0.vlm"},
	{0xA9562FD91188BE62ULL, "511533163.ogg"},
	{0xA956355F2D8E9ADAULL, "1052156899.ogg"},
	{0xA956EEFEE565E89EULL, "6118970_m.fxe"},
	{0xA957480900A6D0C7ULL, "vxca_caverns1_mines_eia_0_0.vlm"},
	{0xA9577C7D87383889ULL, "llac_w_woundedcoast_-131071_0.dds"},
	{0xA957CBEC03F6D9ADULL, "fex_bd_c_bs.mmh"},
	{0xA957D53CF2E6DBFAULL, "vxdr_primevaltomb_8v8_0_0.vlm"},
	{0xA957DDEC03F6F833ULL, "fex_bd_c_bs.msh"},
	{0xA9584EC917D7B26DULL, "6027828_m.fxe"},
	{0xA95880649EF52DE4ULL, "vxlt_0_lowtown_a2_emw_0_0.vlm"},
	{0xA958992986F3D064ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_1.ani"},
	{0xA958B94F3F151668ULL, "dre000_dougal6142201.cl"},
	{0xA958D570A27FD622ULL, "729615323.ogg"},
	{0xA95928B1ABB99DB1ULL, "art\\characters\\creatures\\mubari\\animation\\p_sit.gad"},
	{0xA959354CE41A1C0AULL, "642864387.ogg"},
	{0xA95984750E7B3636ULL, "vxw_sundermount_2h1_0_0.vlm"},
	{0xA95A0D0742743C09ULL, "vxlt_0_lowtown_h9o_0_0.vlm"},
	{0xA95A1456EDF8F01EULL, "6031154_m.fxe"},
	{0xA95A33A18E6CDA19ULL, "vxw_woundedcoast_rv_0_0.vlm"},
	{0xA95B10C3001B0855ULL, "ser270_emeric6116097.cl"},
	{0xA95B9C96527D1C60ULL, "6170916_m.fxe"},
	{0xA95C0D348054AB27ULL, "vxca_cave_1c6_0_0.vlm"},
	{0xA95C718D29344661ULL, "qun221_viscount6084721.cl"},
	{0xA95C81626F1200BBULL, "w_bonepitmining_a2_104_0.mmh"},
	{0xA95C8D626F121558ULL, "w_bonepitmining_a2_104_0.mao"},
	{0xA95C935D82FD47C1ULL, "69065751.ogg"},
	{0xA95C9F626F1233A5ULL, "w_bonepitmining_a2_104_0.msh"},
	{0xA95CD07B82D56610ULL, "hlac_lt_0_lowtown_docks_n_-131074_6z.dds"},
	{0xA95CF9E5606BA489ULL, "vxlt_0_lowtown_ev7_0_0.vlm"},
	{0xA95D06B3CEB98DD7ULL, "vxht_bartrand_n_73l_0_0.vlm"},
	{0xA95D53E3F5A45AC8ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_006a_0t.dds"},
	{0xA95D65851037C3B2ULL, "130072586.ogg"},
	{0xA95DD90C3D386A2BULL, "764320878.ogg"},
	{0xA95DFD298FA24122ULL, "mrl351_pridedemon.cnv"},
	{0xA95E2E82D19B0EC8ULL, "art\\vfx\\combat\\vfx_w_rev_devour_imp_c.impa.ani"},
	{0xA95ED7927DA24D66ULL, "6085953_m.fxe"},
	{0xA95EF46E95CE2717ULL, "vxht_keep_n_64c_0_0.vlm"},
	{0xA95F114CBCEF635FULL, "vxw_encounter2_mt_0_0.vlm"},
	{0xA95F94D4062E9C18ULL, "6160683_m.fxe"},
	{0xA95F962986F9E9C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_1.gad"},
	{0xA95FA425BCCBC23BULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.impa.evt"},
	{0xA96029B1ABBFBE5FULL, "art\\characters\\creatures\\mubari\\animation\\p_sit.ani"},
	{0xA96083598412D851ULL, "and121_karl6093908.cl"},
	{0xA960DF298FA4445AULL, "mrl351_pridedemon.bnk"},
	{0xA960FF9836D8D2B4ULL, "follower_banter6155754.cl"},
	{0xA9616DD5AF6E7293ULL, "6087648_m.fxe"},
	{0xA961C3869E9EBDD4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_meda_d.dds"},
	{0xA961D8BB1797A335ULL, "hlac_lt_0_lowtown_docks_n_0_0x.dds"},
	{0xA9624D3CBBF0755FULL, "92366874.ogg"},
	{0xA96264A2530E0239ULL, "6108631_m.fxe"},
	{0xA96283B505C6BC8AULL, "957473643.ogg"},
	{0xA962C28AF121E90AULL, "vxht_0_hightown_n_aq2_0_0.vlm"},
	{0xA962D2C0F0FBC516ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arrow_act_c.impa.evt"},
	{0xA9631FB820350FF4ULL, "103262871.ogg"},
	{0xA9633140CBFC1888ULL, "vxw_encounter2_camp_qk_0_0.vlm"},
	{0xA96362C668EFDB4DULL, "fxc_puke_sp_white.dds"},
	{0xA9641B522B1BC87FULL, "follower_banter6170163.cl"},
	{0xA964D794A34CEFDCULL, "hlac_lt_0_lowtown_a2_65538_4z.dds"},
	{0xA964F000A78B2492ULL, "lgt260cr_eau_tmp_terrell.utc"},
	{0xA9659D25BCD0227CULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.impa.gad"},
	{0xA96639A25400F52DULL, "vxlt_0_lowtown_n_h8g_0_0.vlm"},
	{0xA966640BB0AEE953ULL, "471280104.ogg"},
	{0xA966912986FFFFA5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_1.evt"},
	{0xA966DEDD9A4405A6ULL, "hlac_ht_mansion2b_818_13z.dds"},
	{0xA967272E2E5210C7ULL, "vxlt_hangedman_2xn_0_0.vlm"},
	{0xA967A5DA549E9979ULL, "gen00fl_fenris6116551.cl"},
	{0xA967F65200714EB6ULL, "vxw_sundermount_2m2_0_0.vlm"},
	{0xA9682FB438F46216ULL, "vxlt_gallowstemplar_n_6sh_0_0.vlm"},
	{0xA968A3B4BA9A0078ULL, "399277276.ogg"},
	{0xA968BF75F881426FULL, "241710703.ogg"},
	{0xA968E15DB33E86D0ULL, "llac_ht_0_hightown_a3_-196608_1.dds"},
	{0xA9697DAC50435BBDULL, "gen_im_trs_bst_drag_gland_04.uti"},
	{0xA9698AE96A4327F2ULL, "dae300_petrice6097859.cl"},
	{0xA969C9C0F101D4A1ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arrow_act_c.impa.gad"},
	{0xA969CF1CAC8E86C0ULL, "vxw_woundedcoast_18d_0_0.vlm"},
	{0xA969F2EA8CCD4E0EULL, "vxca_cave_2d5_0_0.vlm"},
	{0xA96AE6E0670C3998ULL, "6154994_m.fxe"},
	{0xA96B2EC69F1E0A7EULL, "6140005_m.fxe"},
	{0xA96B60D7F8D8B4C7ULL, "plt_cod_ltr_isabelanote.nss"},
	{0xA96BABA6A66D1601ULL, "vxlt_backalley_n_1wt_0_0.vlm"},
	{0xA96C04571129E6BCULL, "dae120ip_at_chantry.utp"},
	{0xA96C9C25BCD63F7AULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.impa.ani"},
	{0xA96D5A4F67A56B3DULL, "vxca_caverns2_8w7_0_1.vlm"},
	{0xA96D8329AEE57BDFULL, "vxw_encounter2_hw_0_0.vlm"},
	{0xA96DCB76A9C600EAULL, "172083885.ogg"},
	{0xA96E5B115B5C08B3ULL, "6043823_f.fxe"},
	{0xA96E7739FB6E2926ULL, "874674730.ogg"},
	{0xA96E7CC759E9853DULL, "zz_pro_debug6059751.cl"},
	{0xA96E9D10CBEBEE13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_medium.evt"},
	{0xA96EB30EB7A4AC5BULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant_0.pwk"},
	{0xA96F003D2077DAF2ULL, "6185495_m.fxe"},
	{0xA96F7C1B4E322AA9ULL, "6080607_m.fxe"},
	{0xA96FA7CD1E805C6FULL, "gen00fl_aveline6139775.cl"},
	{0xA96FCE9DE3EC60F2ULL, "hlac_w_bonepitmining_a2_196611_0z.dds"},
	{0xA970195F35C7A72BULL, "vxlt_gallowstemplar_f_46t_0_0.vlm"},
	{0xA97045837A4D4ACEULL, "6059071_m.fxe"},
	{0xA97069D629727262ULL, "68800363.ogg"},
	{0xA97085FDB8ED42EDULL, "6087347_m.fxe"},
	{0xA970932DC6775666ULL, "416722716.ogg"},
	{0xA970BBD13DB56502ULL, "art\\characters\\weapons\\maces\\w_mce_mc104.mao"},
	{0xA970CAC0F107F54FULL, "art\\vfx\\combat\\vfx_r_arc_rain_arrow_act_c.impa.ani"},
	{0xA970DA65EFE00FC4ULL, "6143213_m.fxe"},
	{0xA9710447AEDCF889ULL, "6052587_m.fxe"},
	{0xA9714E00C364221DULL, "art\\vfx\\materials\\chunky\\vfx_c_branch_weapon.mao"},
	{0xA971710C1993DAE9ULL, "pro000_merchant6051272.cl"},
	{0xA971A31DBDF23108ULL, "vxw_encounter2_camp_lh_0_0.vlm"},
	{0xA971E24DE595DBAAULL, "dae000_meredith6156735.cl"},
	{0xA9723D4A5C44EDC6ULL, "vxlt_gallowstemplar_f_3mg_0_0.vlm"},
	{0xA972587EA9D2E63CULL, "mag101_alain6052998.cl"},
	{0xA972AC2D0250C0C7ULL, "vxlt_hangedman_141_0_0.vlm"},
	{0xA9749610CBF04E54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_medium.gad"},
	{0xA974F554C57EB862ULL, "ser121_jethann6106893.cl"},
	{0xA9750E9EBC0D41FFULL, "dae000_meredith6092573.cl"},
	{0xA9759487CB8C6537ULL, "6110791_m.fxe"},
	{0xA975990B20482947ULL, "vxlt_hangedman_2ui_0_0.vlm"},
	{0xA9759A404DB9A15DULL, "art\\vfx\\environment\\vfx_lt_firewall_4m.mmh"},
	{0xA975E462AA8A061FULL, "414506654.ogg"},
	{0xA97622B4E92CE4E5ULL, "201317780.ogg"},
	{0xA976AE46F90FD42FULL, "qcr000_arishok6166992.cl"},
	{0xA9775993C8226BE1ULL, "vxht_keep_n_3zj_0_0.vlm"},
	{0xA977606ACDB98DB5ULL, "6064652_m.fxe"},
	{0xA9786707637E33EEULL, "vxht_keep_n_5u2_0_0.vlm"},
	{0xA97879B4AE01998AULL, "6099484_m.fxe"},
	{0xA978E9B3A145E48CULL, "vxht_chantry_n_26n_0_0.vlm"},
	{0xA97917661EF188FBULL, "dae000_elegant6166931.cl"},
	{0xA97959470240EAC7ULL, "pro000_captain_wright6004373.cl"},
	{0xA979DA075F0CB6CEULL, "6085892_m.fxe"},
	{0xA979F293886D4A54ULL, "6167857_m.fxe"},
	{0xA97A05D3E31F6924ULL, "468416737.ogg"},
	{0xA97A7448FEDB5C50ULL, "vxht_bartrand_n_68c_0_0.vlm"},
	{0xA97A969F65856D1CULL, "ico_dagger_1.dds"},
	{0xA97B207D79814C7CULL, "vxht_bartrand_n_5ta_0_1.vlm"},
	{0xA97BA510CBF68632ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_medium.ani"},
	{0xA97C87AEE22A04EBULL, "355894582.ogg"},
	{0xA97C95D28D6C9667ULL, "lt_gallowscourtyard_n_lc.mmh"},
	{0xA97CA7AD5682AACFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_doorcrest_sa.dds"},
	{0xA97D26C68750BAA8ULL, "vxlt_gallowscourtyard_n_6yz_0_0.vlm"},
	{0xA97DA9EA88A6F4A4ULL, "260284284.ogg"},
	{0xA97DB376BAB911B4ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_2y.dds"},
	{0xA97E1898CEC85EF9ULL, "308373395.ogg"},
	{0xA97EAD6C2FDE6079ULL, "58131252.ogg"},
	{0xA97EAF92EF847AA9ULL, "6020083_m.fxe"},
	{0xA97EB826D38D74D8ULL, "vxw_bonepit_1p7_0_0.vlm"},
	{0xA97F07AA571C0A67ULL, "vxw_bonepitdestroyed_a3_dk_0_0.vlm"},
	{0xA97F2D4EE24A47D1ULL, "ts_darkspawn_wpn_props.gda"},
	{0xA97F40285E555EA4ULL, "hlac_lt_gallowscourtyard_n_-196608_0z.dds"},
	{0xA980045BC13FFD36ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic_3.mao"},
	{0xA98014FAAFE84988ULL, "vxw_encounter2_camp_gi_0_0.vlm"},
	{0xA9806DE76427637BULL, "vxca_caverns2_99f_0_1.vlm"},
	{0xA980E249897717C8ULL, "vxht_keep_n_3zo_0_0.vlm"},
	{0xA982CCE0F75B6B45ULL, "vxdr_primevaltomb_8v3_0_0.vlm"},
	{0xA9843E43194F0996ULL, "zz_vault_debug6200339.cl"},
	{0xA984C0DF40C149AFULL, "6118005_m.fxe"},
	{0xA985D86F6463559AULL, "215426187.ogg"},
	{0xA9866320F2C1E75CULL, "vxlt_0_lowtown_docks_a3_aqm_0_0.vlm"},
	{0xA9866B840E6E712BULL, "llac_w_blightlands_131070_0.dds"},
	{0xA986B1A4DD5B8717ULL, "6038769_m.fxe"},
	{0xA98702F41308117EULL, "vxlt_0_lowtown_n_e85_0_0.vlm"},
	{0xA987B657E93E1277ULL, "861692926.ogg"},
	{0xA987E1089C56471DULL, "mag121_idunna6076564.cl"},
	{0xA9880F4970684761ULL, "6101547_f.fxe"},
	{0xA9884AF51CDB86F7ULL, "vxdr_slavers_1t2_0_0.vlm"},
	{0xA9886D5876249F3FULL, "hlac_lt_0_lowtown_a3_n_c_65538_0y.dds"},
	{0xA988FEE66E640E77ULL, "zz_fol_debug6097847.cl"},
	{0xA989741DE4E663CDULL, "w_bonepit_102_0.msh"},
	{0xA989821DE4E67B00ULL, "w_bonepit_102_0.mao"},
	{0xA989861DE4E68253ULL, "w_bonepit_102_0.mmh"},
	{0xA98A4DEF9DB4FFA7ULL, "llac_ht_0_hightown_a2_1_1.dds"},
	{0xA98A6EF3AC83F369ULL, "zz_strider_stomp_v1.cut"},
	{0xA98A7246476E3636ULL, "19916613.ogg"},
	{0xA98A81160EE5DB17ULL, "6028878_m.fxe"},
	{0xA98B3B634D21C7A3ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.impa.evt"},
	{0xA98B4401B87563DCULL, "12231203.ogg"},
	{0xA98B74161B2EAFEDULL, "6071966_m.fxe"},
	{0xA98C2306170B7148ULL, "hlac_lt_0_lowtown_a2_n_c_1_5z.dds"},
	{0xA98C35EA5D2C6B70ULL, "485703126.ogg"},
	{0xA98C69A92203BACCULL, "vxlt_warehouse_n_ae_0_0.vlm"},
	{0xA98C7DEEFFE3551DULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02skeleton.mao"},
	{0xA98C81EEFFE35C56ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02skeleton.mmh"},
	{0xA98C8BEEFFE36D54ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_02skeleton.msh"},
	{0xA98CA67591962927ULL, "6062088_f.fxe"},
	{0xA98CE20FA3E874A2ULL, "mcr310_argument6074472.cl"},
	{0xA98D2A03C5838D58ULL, "vxw_bonepit_1u4_0_0.vlm"},
	{0xA98D45F94C86079AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_grenade_toss.gad"},
	{0xA98D7987491222E7ULL, "vxw_bonepitdestroyed_a3_oj_0_0.vlm"},
	{0xA98DEE9FE1F93D42ULL, "follower_banter6155895.cl"},
	{0xA98E0FC17A41AC27ULL, "6055594_f.fxe"},
	{0xA98E1D30DA8E6E94ULL, "788208669.ogg"},
	{0xA98E24671AD62FCBULL, "13627488.ogg"},
	{0xA98E5EC0AAE7BD4AULL, "vxht_0_hightown_a3_cas_0_1.vlm"},
	{0xA98E86D7A1DE6208ULL, "vxw_encounter2_camp_bf_0_0.vlm"},
	{0xA98E8A3E48304063ULL, "vxht_0_hightown_a2_cdm_0_0.vlm"},
	{0xA98EDE177C305D2BULL, "6077737_m.fxe"},
	{0xA98FEC32FC3EB3B3ULL, "hlac_ht_playershouse_n_855_11x.dds"},
	{0xA98FF255272BAF67ULL, "304793452.ogg"},
	{0xA98FFE9A87640289ULL, "hlac_lt_0_lowtown_-65536_7y.dds"},
	{0xA9910142AA2F870CULL, "gen00fl_dog6117181.cl"},
	{0xA99130B80DCAD75CULL, "gen00fl_carver6097231.cl"},
	{0xA991FC5D48B1C459ULL, "vxht_0_hightown_avs_0_0.vlm"},
	{0xA99214634D27A424ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.impa.gad"},
	{0xA992C3FA94BD5F51ULL, "static_static_p19.psh.fxo.dx11"},
	{0xA992F4FB9732C823ULL, "vxca_caverns2_9g1_0_0.vlm"},
	{0xA993E7298FCFAB05ULL, "mrl351_pridedemon.stg"},
	{0xA99448F94C8C2BB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_grenade_toss.evt"},
	{0xA994504A22C2DB07ULL, "vxw_bonepitmining_a2_29v_0_0.vlm"},
	{0xA99460F46A4AD432ULL, "6093489_f.fxe"},
	{0xA9947D65E556BAA2ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_hcliff2_n.dds"},
	{0xA99494F2B2DDA066ULL, "vxw_sundermount_27p_0_0.vlm"},
	{0xA9949DA58BD823A8ULL, "vxdr_0_deeproads_f5c_0_0.vlm"},
	{0xA994B63BC13C1970ULL, "fxa_cen_aoe_volt.mao"},
	{0xA994CC70375081ADULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggerc_n.dds"},
	{0xA995037E4054593BULL, "seb221_random_servants6187836.cl"},
	{0xA99503E6CACFA486ULL, "one350_variel6076870.cl"},
	{0xA99518797777CA97ULL, "hlac_lt_hangedman_-1_24x.dds"},
	{0xA9954172991F0025ULL, "81238393.ogg"},
	{0xA9954FC9E31CFD6DULL, "vxht_keep_n_5zx_0_0.vlm"},
	{0xA9955D6E91EDCA8AULL, "6093684_m.fxe"},
	{0xA995C23EFD812AA9ULL, "vxlt_undercitydungeon_1nk_0_0.vlm"},
	{0xA995CA4052F65732ULL, "6042995_f.fxe"},
	{0xA997013D7C4F5F81ULL, "vxlt_backalley_n_1xi_0_0.vlm"},
	{0xA9978D3672C650E0ULL, "vxlt_0_lowtown_a2_fit_0_0.vlm"},
	{0xA997C1EE50AD505FULL, "6101281_m.fxe"},
	{0xA998219D3AD37255ULL, "426384475.ogg"},
	{0xA9984EA7D9CAC209ULL, "mag330_walter6045881.cl"},
	{0xA998AE7B0FF8422AULL, "gen00fl_varric6079154.cl"},
	{0xA99923634D2DDC02ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.impa.ani"},
	{0xA9996EEC8AA1B1E9ULL, "6115633_m.fxe"},
	{0xA99A25AF6D905D39ULL, "lgt320_mis_resp_swordhunter6150338.cl"},
	{0xA99A3B673D53E25FULL, "hlac_lt_darktown_-1_46y.dds"},
	{0xA99A47858AA1796CULL, "cd_gen00fl_merril.bnk"},
	{0xA99A74DF7EDCA9A4ULL, "vxht_keep_3xz_0_0.vlm"},
	{0xA99AE50B37D4B63DULL, "6060432_m.fxe"},
	{0xA99B3185E7D8C612ULL, "mag121_idunna6092299.cl"},
	{0xA99B469578ED05A0ULL, "w_bonepitmining_a2_139_0.phy"},
	{0xA99B9B4E585C7231ULL, "w_encounter3_93_0.mmh"},
	{0xA99BA74E585C86D2ULL, "w_encounter3_93_0.mao"},
	{0xA99BB54E585C9EEFULL, "w_encounter3_93_0.msh"},
	{0xA99C10A829E396A8ULL, "fen310_fenris6144523.cl"},
	{0xA99C1E5426BFAE33ULL, "668534533.ogg"},
	{0xA99C48D79790C770ULL, "qun221_petrice6185484.cl"},
	{0xA99C7F502D48E502ULL, "vxlt_undercityserialkiller_1ih_0_0.vlm"},
	{0xA99C962A75FFE42DULL, "6064523_f.fxe"},
	{0xA99CC65C1D4A24D5ULL, "vxht_chantry_2df_0_0.vlm"},
	{0xA99CFC1B3A2658E3ULL, "vxht_0_hightown_a2_ccl_0_0.vlm"},
	{0xA99D92E132273746ULL, "vxlt_gallowstemplar_f_3rj_0_0.vlm"},
	{0xA99D93FA3A96B0ECULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08b.phy"},
	{0xA99D96C6BAD979BCULL, "6047385_f.fxe"},
	{0xA99DEDD00476CF3DULL, "vxlt_gamlenshouse_n_c8_0_0.vlm"},
	{0xA99E4A72B8D25C43ULL, "vxw_bonepit_1x1_0_1.vlm"},
	{0xA99E7B89A1584EF4ULL, "48524094.ogg"},
	{0xA99E874986C4D751ULL, "318288007.ogg"},
	{0xA99F336EAF1C1651ULL, "6088852_m.fxe"},
	{0xA99F6833FDA581D2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_lgtc_0d.dds"},
	{0xA99F829CE58FC141ULL, "377124030.ogg"},
	{0xA99FC7E0AE037E92ULL, "hlac_lt_0_lowtown_65536_4x.dds"},
	{0xA9A00A045AC82C52ULL, "6037066_m.fxe"},
	{0xA9A06047D9BD0AE0ULL, "247582538.ogg"},
	{0xA9A092CE1261D202ULL, "504580233.ogg"},
	{0xA9A0DAFCACE3930FULL, "face0_d.psh.fxo"},
	{0xA9A18133659369C9ULL, "6174979_m.fxe"},
	{0xA9A18EF914004E19ULL, "562036254.ogg"},
	{0xA9A1F9CF15D0194DULL, "23767549.ogg"},
	{0xA9A1FD3441523195ULL, "6114043_m.fxe"},
	{0xA9A22FD68EAF697DULL, "6013261_m.fxe"},
	{0xA9A260F94C989180ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_grenade_toss.ani"},
	{0xA9A2713E9119C3FEULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cess.evt"},
	{0xA9A2D40A009602CEULL, "amulet_ivofamilycrest.dds"},
	{0xA9A2F9FA3C2B42B8ULL, "dae000_light_tombs_boss6179455.cl"},
	{0xA9A306CFA4D3B8E6ULL, "vxw_sundermount_28q_0_0.vlm"},
	{0xA9A329AAFB913D47ULL, "6112567_f.fxe"},
	{0xA9A35E2116F0B261ULL, "vxlt_backalley_g_1vc_0_1.vlm"},
	{0xA9A377E49754C6AFULL, "6118200_f.fxe"},
	{0xA9A379EBB93254DFULL, "tbe_quest_given.ncs"},
	{0xA9A3A2626F4E3BF0ULL, "w_bonepitmining_a2_104_0.phy"},
	{0xA9A47D652D085E9BULL, "6083374_m.fxe"},
	{0xA9A4BC6A69A4C7CDULL, "vxlt_0_lowtown_n_ex2_0_0.vlm"},
	{0xA9A4D7A939B1EE57ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1d.mmh"},
	{0xA9A4DBA939B1F564ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1d.mao"},
	{0xA9A4EDA939B213C9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1d.msh"},
	{0xA9A61C3958150D5CULL, "llac_w_blightlands_327681_0.dds"},
	{0xA9A69C39ACC238D5ULL, "vxw_encounter2_camp_pt_0_1.vlm"},
	{0xA9A6F453C52D01C6ULL, "922421518.ogg"},
	{0xA9A717859ADB407BULL, "595886864.ogg"},
	{0xA9A796FAAD4D5022ULL, "69315646.ogg"},
	{0xA9A8072F1EDF1209ULL, "vxca_cavern1_haunted_edx_0_0.vlm"},
	{0xA9A8587D2854980FULL, "vxlt_undercityserialkiller_177_0_0.vlm"},
	{0xA9A894D5CC686E81ULL, "6119848_m.fxe"},
	{0xA9A8A055E8ECCABFULL, "6170652_m.fxe"},
	{0xA9A8BCAD2AAED106ULL, "6024080_m.fxe"},
	{0xA9A9883E91200929ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cess.gad"},
	{0xA9A9A947C9A65E6EULL, "vxdr_slavers_1b5_0_0.vlm"},
	{0xA9AAD5CEC7F671F6ULL, "art\\characters\\creatures\\strider\\animation\\aoe.ani"},
	{0xA9AAE8FEFF35EE1CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.dura.ani"},
	{0xA9AAF12D1F3EFD82ULL, "vxlt_undercityserialkiller_1ni_0_0.vlm"},
	{0xA9AAF6F4863719E0ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_med.phy"},
	{0xA9AAFF69AA7AFF86ULL, "6069605_m.fxe"},
	{0xA9AB379CCA2C9E41ULL, "58434623.ogg"},
	{0xA9AB62C84F2CDA86ULL, "6166984_m.fxe"},
	{0xA9ABA5052103B57CULL, "isa211_isabela6202065.cl"},
	{0xA9AC0E4732586868ULL, "hlac_ht_0_hightown_n_65537_2x.dds"},
	{0xA9AC28144B737AFEULL, "gen00fl_aveline6115482.cl"},
	{0xA9AC6C96B8E71C80ULL, "vxht_0_hightown_c3f_0_1.vlm"},
	{0xA9AC89CC8544F7B3ULL, "playss.ncs"},
	{0xA9AD6C1E1E76B4FFULL, "34425489.ogg"},
	{0xA9ADFB1AE548AE05ULL, "vis160cr_winter_archer.utc"},
	{0xA9AE1D105EF99173ULL, "zz_vault_debug6200354.cl"},
	{0xA9AE52DED1D14DDEULL, "dae210cr_elf_f_01.utc"},
	{0xA9AEF31F5994A749ULL, "dre000_sandal6033765.cl"},
	{0xA9AF1890370CD539ULL, "one110_tree_minder6082495.cl"},
	{0xA9AF7881878B468DULL, "follower_banter6155577.cl"},
	{0xA9AF7AAF665636EEULL, "345158778.ogg"},
	{0xA9AFA1CB30232628ULL, "zz_isa_debug6175876.cl"},
	{0xA9AFF9D6ECDC0667ULL, "llac_ca_caverns1_24424_6.dds"},
	{0xA9AFFC810C782F7BULL, "hlac_lt_0_lowtown_a2_n_c_65536_11z.dds"},
	{0xA9B03512248C1EB5ULL, "368942012.ogg"},
	{0xA9B0430A3B0DADB6ULL, "gen00fl_fenris6126755.cl"},
	{0xA9B0472D8AD326B9ULL, "796360475.ogg"},
	{0xA9B0593E9125D877ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cess.ani"},
	{0xA9B090D63254795DULL, "vxlt_hangedman_n_2d3_0_0.vlm"},
	{0xA9B09206D0E1FB22ULL, "mag311_keran6063759.cl"},
	{0xA9B0CE1F4359D387ULL, "16721677.ogg"},
	{0xA9B0CEBB18F945BCULL, "6171160_m.fxe"},
	{0xA9B166F05D0034AEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03a.phy"},
	{0xA9B16B03BEF34C50ULL, "6066769_m.fxe"},
	{0xA9B1ABF11812921CULL, "lt_undercityencounter_3567_lc.mmh"},
	{0xA9B1F5FEFF3C235EULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.dura.gad"},
	{0xA9B2151374EADA86ULL, "vxdr_primevaltomb_7zr_0_1.vlm"},
	{0xA9B2DD0491EFA966ULL, "701541048.ogg"},
	{0xA9B321B4B6DD0A4BULL, "llac_w_encounter3_-196609_0.dds"},
	{0xA9B32E475B9AE04DULL, "vxlt_0_lowtown_n_eu1_0_0.vlm"},
	{0xA9B352A0C3B0521BULL, "6141155_m.fxe"},
	{0xA9B41658B72E1431ULL, "6072075_m.fxe"},
	{0xA9B431256E1A3498ULL, "ch1alphemarmtnt.psh.fxo"},
	{0xA9B4E5A3ADFFDFB3ULL, "305773059.ogg"},
	{0xA9B556A50CF0FF6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_stomach_hold.ani"},
	{0xA9B58381C0615ACDULL, "art\\vfx\\combat\\vfx_gib_ogre_p_v1.anb"},
	{0xA9B5A0FEF1F05F7FULL, "vxlt_0_lowtown_docks_n_b6u_0_1.vlm"},
	{0xA9B5DFE29CBF539DULL, "78829784.ogg"},
	{0xA9B5EF0E8D4A20B4ULL, "6103913_f.fxe"},
	{0xA9B611317DF210C6ULL, "follower_banter6143359.cl"},
	{0xA9B61CA2FFF4DF05ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_lava1_n.dds"},
	{0xA9B65E2491E0FADEULL, "625430398.ogg"},
	{0xA9B6FCC0659BDC07ULL, "70973870.ogg"},
	{0xA9B743AD0AF58A44ULL, "zz_one000_crafter6179356.cl"},
	{0xA9B7DDCEC80109E7ULL, "art\\characters\\creatures\\strider\\animation\\aoe.evt"},
	{0xA9B8AC4271C96F85ULL, "mcr370_templar16086972.cl"},
	{0xA9B8D0FEFF42037DULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s1_c.dura.evt"},
	{0xA9B8DD9416C1F54EULL, "vxlt_undercityserialkiller_12y_0_0.vlm"},
	{0xA9B93495715506D6ULL, "6041360_m.fxe"},
	{0xA9B9AA16013655D5ULL, "vxht_chantry_2zd_0_0.vlm"},
	{0xA9B9DD66213E64EDULL, "vxlt_0_lowtown_docks_c6p_0_0.vlm"},
	{0xA9B9E058B1666CD3ULL, "230355274.ogg"},
	{0xA9BA14AC8E22CA58ULL, "gen00fl_varric6102791.cl"},
	{0xA9BA2F603039BD57ULL, "vxht_mansion2_a1_0_0.vlm"},
	{0xA9BAF413CF4FFFE4ULL, "hlac_lt_gallowstemplar_f_2392_1x.dds"},
	{0xA9BB53DB89A94D5EULL, "6118071_m.fxe"},
	{0xA9BBB13B451B5812ULL, "6165260_m.fxe"},
	{0xA9BCE3F2412676BEULL, "6041477_f.fxe"},
	{0xA9BD4AD650D620F1ULL, "75841295.ogg"},
	{0xA9BD6F825AADF42EULL, "vxlt_gallowstemplar_n_51i_0_0.vlm"},
	{0xA9BDE3A20CD70924ULL, "799245123.ogg"},
	{0xA9BDF919288621C9ULL, "follower_banter6166975.cl"},
	{0xA9BE1AEAC547BB3EULL, "hlac_dr_0_deeproads_5840_0y.dds"},
	{0xA9BE78D7175D0AB2ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_008a_0n.dds"},
	{0xA9BE91000EDAE14CULL, "6207072_m.fxe"},
	{0xA9BEFECEC8076030ULL, "art\\characters\\creatures\\strider\\animation\\aoe.gad"},
	{0xA9BF2F0A275591CBULL, "dae000_arishok6067655.cl"},
	{0xA9BF5F55357248DDULL, "gen00fl_isabela6075296.cl"},
	{0xA9BF6F9D94CF4FEAULL, "vxlt_hangedman_n_226_0_0.vlm"},
	{0xA9BFFED4193A0E70ULL, "775482535.ogg"},
	{0xA9C01F5E4DDAEA83ULL, "391722461.ogg"},
	{0xA9C028D11C54200CULL, "162445561.ogg"},
	{0xA9C037DD49087B88ULL, "6091378_m.fxe"},
	{0xA9C03AFDF0DA63F3ULL, "6059232_m.fxe"},
	{0xA9C047C5258037ADULL, "hlac_ca_caverns1_mines_24424_13y.dds"},
	{0xA9C06395790C5335ULL, "w_bonepitmining_a2_139_0.msh"},
	{0xA9C06595790C56ABULL, "w_bonepitmining_a2_139_0.mmh"},
	{0xA9C07195790C6B48ULL, "w_bonepitmining_a2_139_0.mao"},
	{0xA9C07B31CCD0456DULL, "6153200_m.fxe"},
	{0xA9C07E4D61292B74ULL, "6112551_f.fxe"},
	{0xA9C084FA8D9D641DULL, "art\\characters\\weapons\\maces\\textures\\w_mce_dsmca_0s.dds"},
	{0xA9C15A3E3C32A7B9ULL, "gen00fl_aveline6139250.cl"},
	{0xA9C1A0244D90F8CDULL, "vxlt_0_lowtown_n_en0_0_0.vlm"},
	{0xA9C1E7F58FF641C9ULL, "ave331_post_fight6138815.cl"},
	{0xA9C21839D4B554C3ULL, "6134937_m.fxe"},
	{0xA9C24EA50CFB7BBFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_stomach_hold.evt"},
	{0xA9C26773B98EB1D6ULL, "art\\characters\\weapons\\shields\\w_shd_twrb.msh"},
	{0xA9C27173B98EC2D4ULL, "art\\characters\\weapons\\shields\\w_shd_twrb.mmh"},
	{0xA9C27573B98EC9E7ULL, "art\\characters\\weapons\\shields\\w_shd_twrb.mao"},
	{0xA9C29781FB72BBF6ULL, "gen00fl_anders6125136.cl"},
	{0xA9C3326DFDD740EDULL, "6166565_m.fxe"},
	{0xA9C3454CE42D0EF4ULL, "416830819.ogg"},
	{0xA9C3F06704A8E57AULL, "w_sundermount_241_0.msh"},
	{0xA9C3FE6704A8FCB3ULL, "w_sundermount_241_0.mao"},
	{0xA9C4026704A903C0ULL, "w_sundermount_241_0.mmh"},
	{0xA9C4131914797F89ULL, "vxht_chantry_n_2ch_0_0.vlm"},
	{0xA9C420A8D963A7C3ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs101a_0d.dds"},
	{0xA9C4DE61B8FAC0B6ULL, "vxlt_backalley_g_1ww_0_1.vlm"},
	{0xA9C51CB75CB38018ULL, "dre000_dougal6129761.cl"},
	{0xA9C560ACCB1E1299ULL, "6189038_m.fxe"},
	{0xA9C6F1778D5BEF58ULL, "vxw_bonepit_1i0_0_0.vlm"},
	{0xA9C785917DFAF559ULL, "gen_im_trs_qun_blt_r2.uti"},
	{0xA9C78BDDEDD8827AULL, "340115912.ogg"},
	{0xA9C80B9A60ABC19EULL, "static_static_p5.psh.fxo.dx11"},
	{0xA9C9132F519DDCA0ULL, "vxw_encounter2_3u_0_0.vlm"},
	{0xA9C92FA50D0165A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_twitch_stomach_hold.gad"},
	{0xA9C9A20598F51658ULL, "663081611.ogg"},
	{0xA9CA2883B497EF22ULL, "vxlt_0_lowtown_a3_n_c_fer_0_0.vlm"},
	{0xA9CA73100BA3F781ULL, "356494222.ogg"},
	{0xA9CAD32293CA3900ULL, "6129553_f.fxe"},
	{0xA9CB12A3ECB3DB9BULL, "6108912_m.fxe"},
	{0xA9CB3201F1B9F55AULL, "hlac_lt_0_lowtown_a3_n_c_0_7z.dds"},
	{0xA9CB96E36D665402ULL, "417678775.ogg"},
	{0xA9CBC3D1108A2659ULL, "vxht_0_hightown_ajw_0_0.vlm"},
	{0xA9CBCAD5486D1EB1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\deathloop.gad"},
	{0xA9CD1027FEC4EE82ULL, "332536480.ogg"},
	{0xA9CD2BDAB170B26BULL, "864856521.ogg"},
	{0xA9CD315CC97592C5ULL, "6018624_m.fxe"},
	{0xA9CF0593D1EE4AC4ULL, "gen00fl_isabela6110172.cl"},
	{0xA9CF219009F502CDULL, "vxdr_primevalentrance_394_0_1.vlm"},
	{0xA9CFA453FE8D8397ULL, "217163063.ogg"},
	{0xA9CFC4DA730E11A1ULL, "hlac_w_woundedcoast_131079_0y.dds"},
	{0xA9D04682436A5DECULL, "hlac_lt_0_lowtown_n_-3_4z.dds"},
	{0xA9D0793DCC89E1D9ULL, "5944465.ogg"},
	{0xA9D07FDA517A7B20ULL, "693748322.ogg"},
	{0xA9D08D7D037456ABULL, "134420197.ogg"},
	{0xA9D0934C67593CB6ULL, "lowtown_fem_alienage1_sta6162603.cl"},
	{0xA9D09AD7660B54A8ULL, "6054683_m.fxe"},
	{0xA9D0CA39C4969667ULL, "dae011_conspiracy_mob6149578.cl"},
	{0xA9D0FD70D887BC2FULL, "vxlt_undercityserialkiller_1l6_0_0.vlm"},
	{0xA9D113F486581A75ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_med.msh"},
	{0xA9D115F486581DEBULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_med.mmh"},
	{0xA9D12C6FB03034D5ULL, "6001280_m.fxe"},
	{0xA9D17E0672DE8B05ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_ceiling_d.dds"},
	{0xA9D181FA71855A8FULL, "158286720.ogg"},
	{0xA9D1B6E586E8D57BULL, "hlac_lt_0_lowtown_docks_n_-2_1x.dds"},
	{0xA9D2CBD548733F5FULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\deathloop.ani"},
	{0xA9D308AF5CD3125EULL, "6201483_m.fxe"},
	{0xA9D328FDF4CB498BULL, "hlac_lt_0_lowtown_a2_65536_12z.dds"},
	{0xA9D3357DE5BC5CA5ULL, "hlac_lt_0_lowtown_docks_n_-131076_1x.dds"},
	{0xA9D3739DD4F7A8FFULL, "3571502.ogg"},
	{0xA9D394CBB5D34B09ULL, "mag371_elsa6064652.cl"},
	{0xA9D47FEBDB80E64EULL, "6180957_m.fxe"},
	{0xA9D4826FFD3A4918ULL, "489054260.ogg"},
	{0xA9D4D0C292B90933ULL, "llac_lt_0_lowtown_a2_0_6.dds"},
	{0xA9D51B2913DE5072ULL, "seb321_leliana6180954.cl"},
	{0xA9D5F660B24BA5C3ULL, "812748278.ogg"},
	{0xA9D6705E6F3387DFULL, "qun110_petrice6041300.cl"},
	{0xA9D69FF49D3DC1FBULL, "lt_0_lowtown_a2_n_c_40_0.rml"},
	{0xA9D6B5CBFD55F04EULL, "vis121_saemus6101292.cl"},
	{0xA9D6F5021CC21661ULL, "hlac_lt_0_lowtown_docks_-1_1x.dds"},
	{0xA9D6FFC2E74CFE61ULL, "6195858_f.fxe"},
	{0xA9D7036779092E4AULL, "vxlt_0_lowtown_a2_n_c_eov_0_0.vlm"},
	{0xA9D96A78D5176209ULL, "gen00fl_anders6113011.cl"},
	{0xA9DA1D5AE246BAFBULL, "6161614_m.fxe"},
	{0xA9DA35AE02803ED9ULL, "vxht_0_hightown_amv_0_0.vlm"},
	{0xA9DA51858AD76D9EULL, "cd_gen00fl_merril.wbd"},
	{0xA9DA57A6A58A16F4ULL, "6151302_m.fxe"},
	{0xA9DA5D09CB94249FULL, "550869956.ogg"},
	{0xA9DA750F537B4D9BULL, "6030488_m.fxe"},
	{0xA9DA948EDACFC5DFULL, "llac_ht_0_hightown_a2_-65538_0.dds"},
	{0xA9DAB4D71A13530DULL, "vxht_mansion2b_1bk_0_0.vlm"},
	{0xA9DAD5171E630EF1ULL, "gen00fl_fenris6126633.cl"},
	{0xA9DAE82EDE14B9DBULL, "mag291_demon_marethari6028914.cl"},
	{0xA9DB954A32529075ULL, "mer120_hubert6070040.cl"},
	{0xA9DD3CE21B5596A0ULL, "723559155.ogg"},
	{0xA9DD7DB4E765222DULL, "692183497.ogg"},
	{0xA9DDD921EB6A6825ULL, "796751581.ogg"},
	{0xA9DE01BD11CD3023ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_lava.msh"},
	{0xA9DE0FBD11CD487DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_lava.mmh"},
	{0xA9DE13BD11CD4E8EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_lava.mao"},
	{0xA9DE3E927F4BE1DFULL, "266890493.ogg"},
	{0xA9DE6AFA3ACE01B7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08b.mmh"},
	{0xA9DE80FA3ACE2729ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos08b.msh"},
	{0xA9DEFE7E784FBA9DULL, "6131346_f.fxe"},
	{0xA9DF6F4DCA7DD4AFULL, "vxlt_undercityserialkiller_1a1_0_0.vlm"},
	{0xA9E006018201532CULL, "art\\characters\\weapons\\dagger\\w_dger_107a.phy"},
	{0xA9E0360B90F63D25ULL, "art\\vfx\\cinematic\\prerendered\\vfx_fog_ground_p.mmh"},
	{0xA9E05665EDDB7F68ULL, "115904578.ogg"},
	{0xA9E096BA25E4843BULL, "6156865_m.fxe"},
	{0xA9E18FFB360B18BDULL, "vxw_bonepitmining_a2_1pd_0_0.vlm"},
	{0xA9E24C2E463F9E9EULL, "6111366_m.fxe"},
	{0xA9E2F903B337FA47ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_0y.dds"},
	{0xA9E30EF04056488BULL, "420984600.ogg"},
	{0xA9E372571CAE666AULL, "hlac_lt_darktown_-1_57z.dds"},
	{0xA9E38732E60F5778ULL, "target_cone.cur"},
	{0xA9E45DED530F4A02ULL, "261815658.ogg"},
	{0xA9E4E1380B8B909EULL, "168739570.ogg"},
	{0xA9E4FE86E8A1DB1DULL, "hlac_lt_undercityencounter_-1_16x.dds"},
	{0xA9E50D0E6144F2A0ULL, "6101332_m.fxe"},
	{0xA9E581CD86F670A4ULL, "vxdr_primevaltomb_55j_0_0.vlm"},
	{0xA9E5867EEDE6122AULL, "llac_w_encounter1_-65531_0.dds"},
	{0xA9E59C77D35BFC0CULL, "vxlt_hangedman_n_1q2_0_0.vlm"},
	{0xA9E5A84DB6104ADAULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_enter.gad"},
	{0xA9E5BC9223B2C202ULL, "583555399.ogg"},
	{0xA9E5CACE5C725949ULL, "133525714.ogg"},
	{0xA9E5F6E9358A0DA0ULL, "vxw_encounter2_9s_0_0.vlm"},
	{0xA9E604EB872D6E49ULL, "vxlt_warehouse_n_9t_0_0.vlm"},
	{0xA9E6690311EA26ECULL, "vxlt_0_lowtown_fid_0_0.vlm"},
	{0xA9E699603646957AULL, "vxw_woundedcoast_g2_0_0.vlm"},
	{0xA9E727DBBB0F9808ULL, "llac_lt_gallowstemplar_f_2382_0.dds"},
	{0xA9E79DF8CC765FEDULL, "6087130_m.fxe"},
	{0xA9E7A5967D07A90EULL, "art\\vfx\\combat\\vfx_glb_ele_n_minor_c.impa.evt"},
	{0xA9E7B08192CF4D27ULL, "vxht_mansion_n_6ov_0_0.vlm"},
	{0xA9E83891BC30293FULL, "271499227.ogg"},
	{0xA9E88F37D43CD37BULL, "6166714_m.fxe"},
	{0xA9E8B6DB80AC16CEULL, "6143347_m.fxe"},
	{0xA9E8FED45E400902ULL, "genip_ladder.utp"},
	{0xA9E9DEF50975AA95ULL, "character_character_p127.psh.fxo"},
	{0xA9EA251FCEF76C5BULL, "w_sundermount_50_0.rml"},
	{0xA9EA929DB7A1EF3BULL, "743233032.ogg"},
	{0xA9EA96C5275B8917ULL, "w_blightlands_52_0_ll.mmh"},
	{0xA9EAB4AC9A324826ULL, "hlac_dr_slavers_3216_2z.dds"},
	{0xA9EB53DAA8C005E4ULL, "vxht_keep_n_61g_0_0.vlm"},
	{0xA9EB58A0D59C7408ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.anb"},
	{0xA9EB69A0943D6FD6ULL, "302779833.ogg"},
	{0xA9EB6DE1E57C2F7DULL, "6176313_m.fxe"},
	{0xA9EBEEC295C305ABULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr180runf.evt"},
	{0xA9EBFE5F2E63B478ULL, "458029060.ogg"},
	{0xA9ECAB4DB6166EF1ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_enter.evt"},
	{0xA9ECC7E62432588FULL, "6021301_m.fxe"},
	{0xA9ECD6BB9CE446C7ULL, "vxw_encounter2_camp_r5_0_0.vlm"},
	{0xA9ECF981E35AE3EAULL, "6197874_m.fxe"},
	{0xA9ED330B9100C2F2ULL, "art\\vfx\\cinematic\\prerendered\\vfx_fog_ground_p.anb"},
	{0xA9ED46698386FA51ULL, "vxca_caverns1_icu_0_1.vlm"},
	{0xA9EE2242DC5C1562ULL, "ser221_quentin.cnv"},
	{0xA9EE3AFD909F0094ULL, "vxlt_gallowstemplar_f_6dj_0_1.vlm"},
	{0xA9EE954242243384ULL, "754411705.ogg"},
	{0xA9EE9C967D0DB8B9ULL, "art\\vfx\\combat\\vfx_glb_ele_n_minor_c.impa.gad"},
	{0xA9EEC61DBF0247C8ULL, "follower_banter6170588.cl"},
	{0xA9EEF183FC63879FULL, "635942466.ogg"},
	{0xA9EF3F2E8AA12788ULL, "hlac_ht_0_hightown_n_-1_2x.dds"},
	{0xA9EF93F3CB845F8BULL, "zz_and_debug6134072.cl"},
	{0xA9F02C1FCEFBE3B3ULL, "w_sundermount_50_0.phy"},
	{0xA9F0651726DE5830ULL, "945410535.ogg"},
	{0xA9F0B09D68CE2466ULL, "836329676.ogg"},
	{0xA9F13BF2D6A1BA6BULL, "vxht_bartrand_n_6pt_0_0.vlm"},
	{0xA9F165F05D361771ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03a.mmh"},
	{0xA9F17FF05D36432FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03a.msh"},
	{0xA9F1B66104C51982ULL, "vxht_bartrand_n_700_0_0.vlm"},
	{0xA9F20442DC5FCA9AULL, "ser221_quentin.bnk"},
	{0xA9F23CB165EC5540ULL, "6166890_m.fxe"},
	{0xA9F2D8DA079F0ED3ULL, "llac_ht_0_hightown_a2_-196610_0.dds"},
	{0xA9F2F2F45D0A8BB8ULL, "6040822_m.fxe"},
	{0xA9F2F483F24C980FULL, "dark_fog_a.mao"},
	{0xA9F307C295C94F6CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr180runf.gad"},
	{0xA9F3527F6D10F681ULL, "6076567_m.fxe"},
	{0xA9F36E21E0CDB04CULL, "609382750.ogg"},
	{0xA9F3CCA9FEFCF065ULL, "hlac_w_woundedcoast_-262136_0y.dds"},
	{0xA9F3D9B011467DF6ULL, "vxlt_darktown_6kc_0_0.vlm"},
	{0xA9F3F6F0B4F84518ULL, "vxht_0_hightown_bfq_0_0.vlm"},
	{0xA9F3FEF4FEAAFFACULL, "llac_lt_0_lowtown_docks_a3_n_-4_7.dds"},
	{0xA9F4004661541198ULL, "vxca_caverns2_8iu_0_0.vlm"},
	{0xA9F553976C9A2B89ULL, "vxca_cave_189_0_0.vlm"},
	{0xA9F5A7CD8C7EEF1AULL, "613369934.ogg"},
	{0xA9F5AD967D13F407ULL, "art\\vfx\\combat\\vfx_glb_ele_n_minor_c.impa.ani"},
	{0xA9F64482A6EFD7F8ULL, "hlac_lt_0_lowtown_docks_-131074_4y.dds"},
	{0xA9F6901646AC289FULL, "vxlt_0_lowtown_a2_bxt_0_0.vlm"},
	{0xA9F6CF884DCE8DD5ULL, "330159590.ogg"},
	{0xA9F6D37A9AB49E0EULL, "hlac_ht_chantry_n_2534_33z.dds"},
	{0xA9F6EA0D6B553B6DULL, "6189012_m.fxe"},
	{0xA9F79890FDFF840DULL, "vxht_mansion2b_1xi_0_0.vlm"},
	{0xA9F7BD6B8A6B1F5AULL, "vxlt_foundry_n_v8_0_0.vlm"},
	{0xA9F82FD192D52675ULL, "combat_st_r1_a4.gda"},
	{0xA9F8B5F507B7C0B7ULL, "178567689.ogg"},
	{0xA9F935B68149BF66ULL, "gen00fl_isabela_init6127537.cl"},
	{0xA9F98CA80BE96B0DULL, "808451645.ogg"},
	{0xA9F9AE1C128A20A5ULL, "hlac_w_blightlands_393212_0z.dds"},
	{0xA9F9C34DB62121C0ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_kd_enter.ani"},
	{0xA9F9DCBA819D6220ULL, "205204153.ogg"},
	{0xA9F9E25B117C37E5ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comb_0t.dds"},
	{0xA9FA06C295CF6C6AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnr180runf.ani"},
	{0xA9FA1A336A3F2837ULL, "24426506.ogg"},
	{0xA9FA45F393855683ULL, "vxlt_ruinedhouse_n_61_0_0.vlm"},
	{0xA9FAB2C8A8648BDCULL, "6119306_m.fxe"},
	{0xA9FB15753684D2DAULL, "vxlt_gallowstemplar_f_5r0_0_0.vlm"},
	{0xA9FB48988EDA5F47ULL, "vxw_encounter2_camp_m4_0_0.vlm"},
	{0xA9FB61236018A341ULL, "624252408.ogg"},
	{0xA9FB7E055DE7F973ULL, "6179761_m.fxe"},
	{0xA9FB81E044C03BA9ULL, "451505429.ogg"},
	{0xA9FBB6C9DAD1F00AULL, "735268193.ogg"},
	{0xA9FC354BC8A1ED19ULL, "vxw_bonepit_11u_0_0.vlm"},
	{0xA9FCBB0E72D735CCULL, "6090056_m.fxe"},
	{0xA9FCDF0E272181E5ULL, "hlac_w_encounter2_camp_131071_0x.dds"},
	{0xA9FCEEE5353C2903ULL, "6080717_m.fxe"},
	{0xA9FD73C9D4940494ULL, "6119517_m.fxe"},
	{0xA9FDB44A26897891ULL, "663865371.ogg"},
	{0xA9FE172C40211134ULL, "404395999.ogg"},
	{0xA9FE3AA4783417ACULL, "6084469_f.fxe"},
	{0xA9FEE542AE117073ULL, "dae000_arishok6078185.cl"},
	{0xA9FEF5D77B4EDCE7ULL, "lowtown_male_alienage2_mob6166525.cl"},
	{0xA9FEFC94C1873E5AULL, "pro000_refugee_male6052101.cl"},
	{0xA9FF4EBC27BC933EULL, "770587961.ogg"},
	{0xA9FF7A07493E7973ULL, "llac_dr_slavers_3215_8.dds"},
	{0xA9FF7CBC5EBD5F0DULL, "6178489_m.fxe"},
	{0xA9FFB4D0FD919A0DULL, "6169447_m.fxe"},
	{0xA9FFC24D4D2E9F8FULL, "llac_ca_cavern1_haunted_24426_3.dds"},
	{0xAA001BB7CDAEC917ULL, "795354558.ogg"},
	{0xAA00407E6100FF77ULL, "zz_vault_debug6200387.cl"},
	{0xAA00B2967F13675DULL, "6134319_m.fxe"},
	{0xAA00F85DE7F54AC8ULL, "vxlt_0_lowtown_docks_a3_n_d2g_0_1.vlm"},
	{0xAA01291A870F24EDULL, "6057342_m.fxe"},
	{0xAA0129C7F7256E5DULL, "6113657_m.fxe"},
	{0xAA01F04418452D48ULL, "dre000_bodahn6031260.cl"},
	{0xAA01F9111F5D2BA6ULL, "vxlt_warehouse_n_11l_0_0.vlm"},
	{0xAA024B8D033C9676ULL, "vxlt_darktown_6fd_0_0.vlm"},
	{0xAA0283205E7A12FFULL, "hlac_ht_playershouse_n_855_30y.dds"},
	{0xAA02E9826FB323ACULL, "lgt210_magu_selby6142057.cl"},
	{0xAA038845478C57BCULL, "6118396_m.fxe"},
	{0xAA03891522501E6AULL, "908972571.ogg"},
	{0xAA04CC6D25E5EE1FULL, "vxlt_warehouse_n_133_0_0.vlm"},
	{0xAA053AE9218DE9C6ULL, "6120869_m.fxe"},
	{0xAA055B465231575EULL, "w_sundermount_77_0.rml"},
	{0xAA057DC500689E30ULL, "28420576.ogg"},
	{0xAA05E92B13B88BD4ULL, "6179371_m.fxe"},
	{0xAA05F3A6B7110066ULL, "mag311_nyssa6051212.cl"},
	{0xAA06001085AA85CDULL, "6105676_m.fxe"},
	{0xAA060A6DEFF59C8DULL, "vxht_mansion2b_1sh_0_0.vlm"},
	{0xAA0635A184ECB74EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l180_rf.evt"},
	{0xAA06403B0CB57351ULL, "145852514.ogg"},
	{0xAA06FDD95AE5CBF8ULL, "ch1_s1.psh.fxo"},
	{0xAA06FFAD34241A47ULL, "follower_banter6161037.cl"},
	{0xAA073E6D7BA68F5DULL, "hlac_w_bonepitdestroyed_a3_131073_0z.dds"},
	{0xAA079DB9DCC5A624ULL, "vxw_bonepit_yg_0_0.vlm"},
	{0xAA08073AAF84D982ULL, "qun131_party_comment6117270.cl"},
	{0xAA0826267C2D4326ULL, "hlac_ht_0_hightown_-65536_3x.dds"},
	{0xAA0833BF310779CDULL, "rdr110dg_guard_sleeping6054678.cl"},
	{0xAA08DBBDB9637F15ULL, "bind_cyc.mao"},
	{0xAA0901AE4D69BC2FULL, "lgt220_shp_mag_guard6149493.cl"},
	{0xAA096C0FE45E78DCULL, "516245030.ogg"},
	{0xAA099D91C6FA1207ULL, "vxdr_0_deeproads_bx7_0_0.vlm"},
	{0xAA09B634F3DB964EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl03.msh"},
	{0xAA09BA7580D077C7ULL, "vxw_encounter2_camp_h7_0_0.vlm"},
	{0xAA09D034F3DBC23CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl03.mmh"},
	{0xAA09DCF815389E4DULL, "dae000_hate_follower_react6167489.cl"},
	{0xAA0A1B3892B03D53ULL, "6132812_m.fxe"},
	{0xAA0A1CB40D8351F8ULL, "6205654_m.fxe"},
	{0xAA0A1CCD9F97E2C1ULL, "mcr371_arthuris6087097.cl"},
	{0xAA0A9909285ED4F6ULL, "923944489.ogg"},
	{0xAA0B16807748A8E4ULL, "340424597.ogg"},
	{0xAA0B458A0BE8FF1FULL, "vxht_0_hightown_a2_9b5_0_0.vlm"},
	{0xAA0B5968BD54160BULL, "161222872.ogg"},
	{0xAA0B8680567DDFC9ULL, "119290647.ogg"},
	{0xAA0BF22286B72988ULL, "65722945.ogg"},
	{0xAA0C74465237A022ULL, "w_sundermount_77_0.phy"},
	{0xAA0D2CA184F2C6F9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l180_rf.gad"},
	{0xAA0D3F8EF802B7B9ULL, "978776799.ogg"},
	{0xAA0E355320238E68ULL, "398985255.ogg"},
	{0xAA0E5ECDAD10D07EULL, "22692027.ogg"},
	{0xAA0FD7F7CFBF9BE2ULL, "fxp_pillar_sym03.dds"},
	{0xAA0FE6E77BB44B00ULL, "vrc321_varric6094305.cl"},
	{0xAA10E42489D216FDULL, "hlac_lt_0_lowtown_a3_n_c_-131072_1x.dds"},
	{0xAA1191043ABB24C5ULL, "vxw_woundedcoast_g9_0_0.vlm"},
	{0xAA1195CFA5873684ULL, "vxw_woundedcoast_12b_0_0.vlm"},
	{0xAA11A111DC119603ULL, "6076309_m.fxe"},
	{0xAA11A561E0F046E4ULL, "hlac_lt_0_lowtown_a2_0_8y.dds"},
	{0xAA11BE651D65391DULL, "31784675.ogg"},
	{0xAA125823347438E6ULL, "hlac_ca_caverns1_24426_4y.dds"},
	{0xAA12667E33F14288ULL, "dre000_sandal6033432.cl"},
	{0xAA12683819A2AC30ULL, "zz_followers_banter6159477.cl"},
	{0xAA13383599CC873AULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackright.gad"},
	{0xAA133DA184F74F47ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l180_rf.ani"},
	{0xAA1357A0D5BDC4A3ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.mmh"},
	{0xAA13E48D786D5BE7ULL, "6078471_m.fxe"},
	{0xAA1466E664026FA7ULL, "20197318.ogg"},
	{0xAA146ACF53E6C1E0ULL, "6087311_m.fxe"},
	{0xAA147C4AE1EBB50DULL, "vxht_mansion2b_1vg_0_0.vlm"},
	{0xAA15F01A39C33D78ULL, "6110529_m.fxe"},
	{0xAA15FBD9D7EC1899ULL, "vxw_woundedcoast_3a_0_0.vlm"},
	{0xAA16285B7DD1BD4AULL, "6187718_m.fxe"},
	{0xAA16A3A9A755A51BULL, "806162550.ogg"},
	{0xAA16FC5F9BC9C7FAULL, "6125772_m.fxe"},
	{0xAA17219E7F831083ULL, "198621958.ogg"},
	{0xAA173A5F0FC768CFULL, "1014039153.ogg"},
	{0xAA1791E10B6E4FB0ULL, "354869827.ogg"},
	{0xAA17DCA22AA63061ULL, "6092269_f.fxe"},
	{0xAA17F69093F033FCULL, "6111395_m.fxe"},
	{0xAA182C5272C69047ULL, "vxw_encounter2_camp_c6_0_0.vlm"},
	{0xAA185628A42DD6CEULL, "w_encounter3_19_0_ll.mmh"},
	{0xAA18977D975BA476ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinscol.msh"},
	{0xAA18A17D975BB574ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinscol.mmh"},
	{0xAA18A57D975BBB87ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinscol.mao"},
	{0xAA195C992FEB67F2ULL, "6071954_m.fxe"},
	{0xAA1980111D5974E5ULL, "gen00fl_anders6045338.cl"},
	{0xAA19A7A50882727FULL, "6094827_f.fxe"},
	{0xAA19AF4B781C41FBULL, "vxlt_0_lowtown_e1l_0_0.vlm"},
	{0xAA1A141261DA9AE0ULL, "6128918_f.fxe"},
	{0xAA1A169DF3DAE277ULL, "zz_prmpt_debug6198261.cl"},
	{0xAA1A1F3C62772C3BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_exit.ani"},
	{0xAA1A3B3599D2AB51ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackright.evt"},
	{0xAA1A47B96D7168E0ULL, "756932163.ogg"},
	{0xAA1AE2A09F229011ULL, "vxw_bonepitdestroyed_a3_lu_0_0.vlm"},
	{0xAA1B36C53875D2AAULL, "vxlt_gallowstemplar_f_3rf_0_0.vlm"},
	{0xAA1C3F282A94317EULL, "hlac_w_blightlands_196604_0y.dds"},
	{0xAA1D4564704B36E7ULL, "403730565.ogg"},
	{0xAA1DBD48614AEC59ULL, "6072282_m.fxe"},
	{0xAA1E30A797487E07ULL, "mag100cr_wilmod.utc"},
	{0xAA1E4B4091676D57ULL, "6093902_m.fxe"},
	{0xAA1EF4CF26FCDFB2ULL, "vxht_brothel_n_5kc_0_0.vlm"},
	{0xAA1F068CDA6A9328ULL, "ser221_gamlen6135067.cl"},
	{0xAA1F5E3ED87CE023ULL, "vxlt_0_lowtown_n_gpf_0_0.vlm"},
	{0xAA1F6B7088D27D3AULL, "zz_prmpt_debug6200925.cl"},
	{0xAA1F8B9E3C9503ADULL, "abi_w_ss_shield_def.bnk"},
	{0xAA20108C73820177ULL, "cod_hst_kirk_arch_1a.ncs"},
	{0xAA20242ADA7AC79EULL, "hlac_lt_0_lowtown_docks_-131074_3z.dds"},
	{0xAA20DD018238A4F7ULL, "art\\characters\\weapons\\dagger\\w_dger_107a.mmh"},
	{0xAA20F3018238CA69ULL, "art\\characters\\weapons\\dagger\\w_dger_107a.msh"},
	{0xAA2176457C94245BULL, "w_blightlands_114_0.mmh"},
	{0xAA2182457C9438F8ULL, "w_blightlands_114_0.mao"},
	{0xAA218F42C35612CFULL, "smoke_ciga_f2.dds"},
	{0xAA2194457C945745ULL, "w_blightlands_114_0.msh"},
	{0xAA21AB1B64FADD1CULL, "6116633_m.fxe"},
	{0xAA2204765019048AULL, "vfx_vfx_p9.psh.fxo.dx11.assembly"},
	{0xAA22D24CF448B83AULL, "vxw_sundermount_2o4_0_0.vlm"},
	{0xAA22E28B17FBB659ULL, "6174225_m.fxe"},
	{0xAA243CB3EC123794ULL, "64326346.ogg"},
	{0xAA2494AA6621F243ULL, "w_bonepit_0_0.rml"},
	{0xAA2496E1648ECB1BULL, "vxlt_0_lowtown_a2_n_fit_0_0.vlm"},
	{0xAA249A386FC7D5CAULL, "6188618_m.fxe"},
	{0xAA249F561DD65F5FULL, "llac_w_blightlands_65539_0.dds"},
	{0xAA24AB480B21E7ADULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_elftarp_dp.dds"},
	{0xAA24AE30722F30B9ULL, "6027500_f.fxe"},
	{0xAA24F70BD8D25B8FULL, "hlac_lt_backalley_g_65537_8y.dds"},
	{0xAA250794267A9566ULL, "vxw_sundermount_2ez_0_0.vlm"},
	{0xAA259CB252053C47ULL, "arm_flare.mao"},
	{0xAA259FC934F2BB38ULL, "vxlt_0_lowtown_n_h86_0_0.vlm"},
	{0xAA25D992F06C6B37ULL, "crt_spblm_line.mao"},
	{0xAA25DDA6E3076493ULL, "6127010_m.fxe"},
	{0xAA266B0C0C671C5AULL, "vxlt_gallowstemplar_f_5c3_0_0.vlm"},
	{0xAA269BB5498FE6C9ULL, "qun211cr_elf_warrior6072196.cl"},
	{0xAA269E25A11B8377ULL, "vxw_bonepit_269_0_0.vlm"},
	{0xAA26A0100F1F7D4AULL, "vxlt_gamlenshouse_n_3k_0_0.vlm"},
	{0xAA26BF9A1B3C0055ULL, "ser191_emeric6104735.cl"},
	{0xAA26D170276E1971ULL, "hlac_lt_0_lowtown_n_65536_12y.dds"},
	{0xAA27273C6281C432ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_exit.evt"},
	{0xAA273E00AEE473A2ULL, "zz_and_debug6134152.cl"},
	{0xAA27533599DD5E20ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\walklbackright.ani"},
	{0xAA27E6F7CFD44680ULL, "fxp_pillar_sym03.mao"},
	{0xAA2894DDEB4044D6ULL, "hlac_dr_dungeon_3217_12z.dds"},
	{0xAA289780F2E3DBABULL, "zz_qun_debug6028291.cl"},
	{0xAA28A5614D6839CBULL, "vxca_caverns2prc_1yj_0_0.vlm"},
	{0xAA28C233AA0992B7ULL, "440049372.ogg"},
	{0xAA28E1850507B12BULL, "vxht_0_hightown_n_bwv_0_0.vlm"},
	{0xAA29354611E22B6AULL, "312225520.ogg"},
	{0xAA29651E9D8B8628ULL, "vxdr_primevalentrance_fl_0_0.vlm"},
	{0xAA29BF3C8B6FBBABULL, "vxlt_gallowstemplar_n_43w_0_0.vlm"},
	{0xAA29DAB623D5027FULL, "vxlt_undercityserialkiller_1ks_0_0.vlm"},
	{0xAA2A2CFF73B430BFULL, "862607167.ogg"},
	{0xAA2AC0496815CD31ULL, "vxht_mansion2b_2jh_0_0.vlm"},
	{0xAA2B01E18F3EA0FCULL, "49677605.ogg"},
	{0xAA2BBDB21F19BE24ULL, "vxlt_0_lowtown_n_f1c_0_0.vlm"},
	{0xAA2C09895DAA7D60ULL, "293216362.ogg"},
	{0xAA2C2D0982E80625ULL, "vxca_cavern1_haunted_gk7_0_0.vlm"},
	{0xAA2C911C4E24ACB3ULL, "51447433.ogg"},
	{0xAA2CBE22CB7BE0D4ULL, "413198594.ogg"},
	{0xAA2CE582C7C8E9A9ULL, "150573486.ogg"},
	{0xAA2D0A9F9372046EULL, "vrc321ip_spooky_1_owner.utp"},
	{0xAA2D22FACC8947E5ULL, "128412935.ogg"},
	{0xAA2D7C878C1BA36FULL, "6165799_m.fxe"},
	{0xAA2E463C628817C5ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_exit.gad"},
	{0xAA2E84F0416A3239ULL, "147386462.ogg"},
	{0xAA2EAF34F3FAA607ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl03.phy"},
	{0xAA2F186D6B592F98ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic_1.mao"},
	{0xAA2F46DE32323815ULL, "dae000_bartrand6031154.cl"},
	{0xAA2F942EB71D5862ULL, "zz_dae_debug6029824.cl"},
	{0xAA2F98F8E27A082CULL, "6060464_m.fxe"},
	{0xAA2FB358F293E706ULL, "1045893468.ogg"},
	{0xAA303747E31761A0ULL, "gen00fl_sebastian6180400.cl"},
	{0xAA30FC2B5F198E21ULL, "6065307-feelgoodpullingtrig.cut"},
	{0xAA3139D71D613579ULL, "llac_lt_0_lowtown_n_-131072_8.dds"},
	{0xAA3191E9F3495B45ULL, "vxw_bonepit_1q1_0_0.vlm"},
	{0xAA3195FC8E1E43BAULL, "223130359.ogg"},
	{0xAA31F635084AD2BEULL, "llac_w_bonepit_327680_0.dds"},
	{0xAA324F46525830D6ULL, "w_sundermount_77_0.mao"},
	{0xAA3255FA9AF4C52FULL, "6188691_m.fxe"},
	{0xAA325B4652584575ULL, "w_sundermount_77_0.mmh"},
	{0xAA325D46525848EBULL, "w_sundermount_77_0.msh"},
	{0xAA32EEFFEF697ABEULL, "6072256_m.fxe"},
	{0xAA32F350B2A7EFA4ULL, "vxw_bonepit_jj_0_0.vlm"},
	{0xAA332FCE7F74D7BBULL, "154792054.ogg"},
	{0xAA33418E10AB65E5ULL, "436169822.ogg"},
	{0xAA3351C89E23E185ULL, "gen00fl_anders6125403.cl"},
	{0xAA33CCC7C9B31F47ULL, "6140960_f.fxe"},
	{0xAA33D3D889F5ECF2ULL, "6067856_m.fxe"},
	{0xAA346321260594EFULL, "vxlt_hangedman_n_4y_0_0.vlm"},
	{0xAA34FB173295BBE7ULL, "6161221_m.fxe"},
	{0xAA35100293119BF7ULL, "vxw_bonepit_218_0_0.vlm"},
	{0xAA3581F3EF8BAAEAULL, "865993460.ogg"},
	{0xAA3582B854EC8680ULL, "hair0_pass1.vsh.fxo"},
	{0xAA3585684D9F5CF3ULL, "6202285_m.fxe"},
	{0xAA36074486E279FEULL, "hlac_ca_caverns2prc_5802_36y.dds"},
	{0xAA36287B7AD9B685ULL, "character_character_p143.psh.fxo.dx11.assembly"},
	{0xAA363EEBFD03ECD1ULL, "6206922_m.fxe"},
	{0xAA369D993D3404E3ULL, "512700491.ogg"},
	{0xAA36DDDF4B2F7E64ULL, "vxlt_undercitydungeon_171_0_0.vlm"},
	{0xAA36EB3C28934398ULL, "plt_gen00pt_ai.plo"},
	{0xAA36F7D0DF22AF0FULL, "art\\characters\\weapons\\staff\\textures\\w_stf_013a_0s.dds"},
	{0xAA374FAD91D7986AULL, "190099208.ogg"},
	{0xAA37A23CB2AFDDE8ULL, "mcr310_argument6173612.cl"},
	{0xAA37AC43A753EB1EULL, "6041496_m.fxe"},
	{0xAA37B77DC96DA4FDULL, "vxht_0_hightown_a3_bad_0_0.vlm"},
	{0xAA37BD3EEA5AD2DFULL, "vxw_encounter2_2e_0_0.vlm"},
	{0xAA37C3F251338265ULL, "53253504.ogg"},
	{0xAA37C84209703EF9ULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyn.msh"},
	{0xAA37D24209704FE7ULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyn.mmh"},
	{0xAA37D64209705634ULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyn.mao"},
	{0xAA3849E4B56F2507ULL, "908681155.ogg"},
	{0xAA386F2D22804083ULL, "vxlt_hangedman_n_27x_0_0.vlm"},
	{0xAA38915C0B192254ULL, "249020694.ogg"},
	{0xAA38A2BE039820A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_right_hand.evt"},
	{0xAA38D1D5BAB93051ULL, "361008175.ogg"},
	{0xAA38F516744EDC8DULL, "vxlt_0_lowtown_a2_n_gbz_0_0.vlm"},
	{0xAA38FF2FAA6A81C5ULL, "6045323_m.fxe"},
	{0xAA3921A1AB41A133ULL, "zz_followers_banter6158211.cl"},
	{0xAA3984274488CC79ULL, "6016921_m.fxe"},
	{0xAA39A374BB4BF3A7ULL, "gen00fl_aveline6116736.cl"},
	{0xAA39A9A58B00293AULL, "575709070.ogg"},
	{0xAA39F125A48AD62AULL, "850129494.ogg"},
	{0xAA3A89F2AC25CA50ULL, "vxht_mansion2_250_0_0.vlm"},
	{0xAA3ACE925343E947ULL, "ht_0_hightown_a3_12_0_lc.mmh"},
	{0xAA3BA4D1C326F5E2ULL, "dae021_torelin_mob.cnv"},
	{0xAA3BD6EB089AB9A9ULL, "dae021_sister_alema_mob6144976.cl"},
	{0xAA3BEC4F6CD800ACULL, "226681911.ogg"},
	{0xAA3BFF75F54528E2ULL, "6170200_m.fxe"},
	{0xAA3C51A134F59F0FULL, "7723411.ogg"},
	{0xAA3C7F80D07D5510ULL, "vxlt_undercitydungeon_1jc_0_1.vlm"},
	{0xAA3CB47A0ECE4A50ULL, "zz_seb_debug6185067.cl"},
	{0xAA3D232FB7EC1908ULL, "6093862_m.fxe"},
	{0xAA3D7ADB2DEB9F1BULL, "6095921_f.fxe"},
	{0xAA3E1D072E9E8F9BULL, "art\\characters\\creatures\\dragon\\animation\\atk_v3.ani"},
	{0xAA3EA4306D91B606ULL, "723143653.ogg"},
	{0xAA3EAD42ECD9D946ULL, "6067711_m.fxe"},
	{0xAA3EB753CF30B351ULL, "vxht_keep_62x_0_0.vlm"},
	{0xAA3F2423602DDFF5ULL, "t3_csp_b03.tnt"},
	{0xAA3F58465262CA02ULL, "w_sundermount_77_0.anb"},
	{0xAA3F7BBE039DFD24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_right_hand.gad"},
	{0xAA3F8389F8B1BFF9ULL, "72677488.ogg"},
	{0xAA3F86D1C32AAB1AULL, "dae021_torelin_mob.bnk"},
	{0xAA3FB606D834E93AULL, "vxw_sundermount_2u6_0_0.vlm"},
	{0xAA4003C6E53F73C5ULL, "vxw_bonepit_1z0_0_0.vlm"},
	{0xAA4032ADB8CDD37DULL, "264389630.ogg"},
	{0xAA40C569C85F497CULL, "gen00fl_dog6167072.cl"},
	{0xAA4128B820C1ECE4ULL, "lt_0_lowtown_docks_224_0.msh"},
	{0xAA413AB820C20B4DULL, "lt_0_lowtown_docks_224_0.mao"},
	{0xAA413EB820C21186ULL, "lt_0_lowtown_docks_224_0.mmh"},
	{0xAA4183E1B54A57F2ULL, "hlac_w_sundermount_-393212_0z.dds"},
	{0xAA41B93FEE4078F8ULL, "16560774.ogg"},
	{0xAA41DFEE5A8B0CF3ULL, "mag170_hugh6080702.cl"},
	{0xAA42235FA79E8870ULL, "nat380tr_sibling_fight.utt"},
	{0xAA42286B1FF37857ULL, "test2.cut"},
	{0xAA432FF325B9C7A6ULL, "mrl150_merrill6078330.cl"},
	{0xAA435A30BB13B135ULL, "vxdr_primevaltomb_77r_0_1.vlm"},
	{0xAA435BD288ECB654ULL, "408011055.ogg"},
	{0xAA4395B4B9C70C2BULL, "6139655_m.fxe"},
	{0xAA446012D1CBE0BEULL, "dae000_bodhan_letter6144287.cl"},
	{0xAA44F736AA1EA5B4ULL, "dre180ip_primeval_chest_3.utp"},
	{0xAA4559D8D5AF8AA1ULL, "mcr310_argument6072282.cl"},
	{0xAA456B205998D422ULL, "dre000_dougal6167440.cl"},
	{0xAA4576EA84787B27ULL, "gen00fl_sebastian6177915.cl"},
	{0xAA46827806951FDFULL, "vxw_blightlands_as_0_0.vlm"},
	{0xAA468ABE03A43502ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_right_hand.ani"},
	{0xAA46B530F7917B5AULL, "framebuffereffects_hslmatrix_p0.vsh.fxo.dx11.assembly"},
	{0xAA46D44737C1DEECULL, "tm_glk_b02.tnt"},
	{0xAA4745470C2A7466ULL, "ser221_party_comments6209893.cl"},
	{0xAA47804082986BC7ULL, "primitive.mao"},
	{0xAA47A999DCD53EA8ULL, "792032933.ogg"},
	{0xAA47F50C04DF7F40ULL, "6027571-whatareyouplanning.cut"},
	{0xAA48195F2489C41EULL, "134487986.ogg"},
	{0xAA48600D4C279316ULL, "vxca_caverns2prc_9oa_0_0.vlm"},
	{0xAA4893BA93EF1F10ULL, "mag320_orsino6071249.cl"},
	{0xAA48EA06A1E88ADBULL, "hlac_ht_0_hightown_-131073_1y.dds"},
	{0xAA49666681767E1DULL, "vxht_mansion_n_4yp_0_0.vlm"},
	{0xAA4990B2FD077F62ULL, "168564927.ogg"},
	{0xAA49A7BCC887E751ULL, "dust_02.mao"},
	{0xAA49BD4518A28C4BULL, "art\\characters\\creatures\\arh_abm_rev\\c_rev_body.mao"},
	{0xAA49D9078C52DE84ULL, "atl_startmenu2_dxt5_00.dds"},
	{0xAA4AFE951F83C5C0ULL, "vxdr_0_deeproads_a29_0_0.vlm"},
	{0xAA4B25072EA92792ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v3.evt"},
	{0xAA4B4155DFBB50F8ULL, "vxht_keep_42w_0_0.vlm"},
	{0xAA4B713859EC8844ULL, "vxlt_0_lowtown_a2_n_c_fea_0_0.vlm"},
	{0xAA4B819E3CBA2DD7ULL, "abi_w_ss_shield_def.wbd"},
	{0xAA4BDAB3A8EE0CEDULL, "532422869.ogg"},
	{0xAA4C3F296A24E4E7ULL, "gen00fl_anders6125618.cl"},
	{0xAA4C5D8B2877F092ULL, "77456041.ogg"},
	{0xAA4D4A40798E39ADULL, "vxht_mansion_n_3hl_0_0.vlm"},
	{0xAA4D9075BABEC885ULL, "vxlt_hangedman_n_m0_0_0.vlm"},
	{0xAA4D93228F0E8138ULL, "152630250.ogg"},
	{0xAA4DADC8A992CD69ULL, "300176211.ogg"},
	{0xAA4DB5271DF61CD7ULL, "vxlt_0_lowtown_docks_a3_n_d6y_0_0.vlm"},
	{0xAA4DD5196B7D39ABULL, "230173629.ogg"},
	{0xAA4DEB06619A360FULL, "gen00fl_isabela6100033.cl"},
	{0xAA4E16AC550C8918ULL, "mag370_meredith6046840.cl"},
	{0xAA4E40FCFBB7F8AAULL, "zz_vault_debug6203161.cl"},
	{0xAA4E9BDDD113A272ULL, "767241364.ogg"},
	{0xAA4EB742CECB9145ULL, "vxlt_backalley_g_1oh_0_1.vlm"},
	{0xAA4F75583B345926ULL, "gen_im_trs_tmp_med_hlm_r0.uti"},
	{0xAA4F841B0D577C8AULL, "hlac_ht_keep_n_3732_7x.dds"},
	{0xAA510CEE34DA4CA0ULL, "framebuffereffects_radialblur_p0.vsh.fxo.dx11.assembly"},
	{0xAA518315BC6985CDULL, "gen00fl_anders6112975.cl"},
	{0xAA51A6074EAF1595ULL, "vxlt_0_lowtown_docks_a3_n_cos_0_0.vlm"},
	{0xAA51ED00E640836AULL, "zz_vault_debug6177602.cl"},
	{0xAA520EB4A0572E7BULL, "6195842_m.fxe"},
	{0xAA5244072EAF7B25ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v3.gad"},
	{0xAA53251FCF505528ULL, "w_sundermount_50_0.mmh"},
	{0xAA53311FCF5069CBULL, "w_sundermount_50_0.mao"},
	{0xAA53431FCF508852ULL, "w_sundermount_50_0.msh"},
	{0xAA539C80A5481D60ULL, "hlac_lt_0_lowtown_docks_-2_3z.dds"},
	{0xAA53CD5C36A29A9FULL, "vxlt_0_lowtown_a3_n_c_eja_0_0.vlm"},
	{0xAA53F816CCBCD4DBULL, "vxht_0_hightown_9v1_0_0.vlm"},
	{0xAA540578C04AD4F9ULL, "vxlt_gallowstemplar_f_2xr_0_0.vlm"},
	{0xAA54330B247BA390ULL, "zz_dae_debug6029909.cl"},
	{0xAA54A0F8CE4703DFULL, "vxw_encounter2_8g_0_0.vlm"},
	{0xAA54DEAE2ACE87F2ULL, "vxlt_0_lowtown_docks_a3_n_9re_0_0.vlm"},
	{0xAA54F454F88B385FULL, "vxw_blightlands_jr_0_0.vlm"},
	{0xAA567A8132CF6260ULL, "199901559.ogg"},
	{0xAA56C13CCA6C09B8ULL, "zz_mrl_debug6185305.cl"},
	{0xAA56D1EA3E1DAB96ULL, "vxca_caverns2prc_9jb_0_0.vlm"},
	{0xAA5714724C33C529ULL, "hlac_lt_0_lowtown_docks_a3_65533_3y.dds"},
	{0xAA57509C1E73698AULL, "329232218.ogg"},
	{0xAA5782AA664D2D5AULL, "w_bonepit_0_0.msh"},
	{0xAA5790AA664D4493ULL, "w_bonepit_0_0.mao"},
	{0xAA5794AA664D4BA0ULL, "w_bonepit_0_0.mmh"},
	{0xAA579E72AB929235ULL, "6114094_m.fxe"},
	{0xAA57C2127C5502CEULL, "vxlt_undercityserialkiller_1yt_0_0.vlm"},
	{0xAA58078660B13565ULL, "mcr370_negotiation6080197.cl"},
	{0xAA5843EA67482AAAULL, "6119454_m.fxe"},
	{0xAA585937208E71BCULL, "vxw_woundedcoast_38_0_0.vlm"},
	{0xAA5895A515D8F20CULL, "901155993.ogg"},
	{0xAA58D15182C47203ULL, "follower_barks6153260.cl"},
	{0xAA58E46946D3C557ULL, "1015153146.ogg"},
	{0xAA58F7A0FFD611F8ULL, "vxw_encounter1_h9_0_0.vlm"},
	{0xAA5923F705808503ULL, "270263856.ogg"},
	{0xAA5959266ECC6FEBULL, "dae000_gamlen6005809.cl"},
	{0xAA598190986AC5E4ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcbl.msh"},
	{0xAA599390986AE44DULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcbl.mao"},
	{0xAA599790986AEA86ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockcbl.mmh"},
	{0xAA59989445144D3FULL, "29197364.ogg"},
	{0xAA59ACAD1484B73EULL, "dae000_mother6136925.cl"},
	{0xAA5ABE198F386CF0ULL, "ser270_templar6127416.cl"},
	{0xAA5B163FB8BF48C1ULL, "vxht_0_hightown_anc_0_0.vlm"},
	{0xAA5C400275F08BCDULL, "gen00fl_varric6119011.cl"},
	{0xAA5C5D7D246B94D6ULL, "llac_lt_darktown_-1_44.dds"},
	{0xAA5C904DB4ED3EB1ULL, "vxw_sundermount_34v_0_1.vlm"},
	{0xAA5CA190A0D4E54CULL, "llac_lt_gallowstemplar_2391_5.dds"},
	{0xAA5CDDE7ED604086ULL, "242764351.ogg"},
	{0xAA5CE46498187EC8ULL, "vxht_brothel_n_34b_0_0.vlm"},
	{0xAA5CF0D8FE6061F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_standing"},
	{0xAA5D1530FC02AD12ULL, "vxht_0_hightown_5yi_0_0.vlm"},
	{0xAA5D4832CC2533AFULL, "6048230_m.fxe"},
	{0xAA5DA71827C5728EULL, "llac_ht_keep_4519_6.dds"},
	{0xAA5DCBC6E912B77AULL, "479935409.ogg"},
	{0xAA5E704DAA08F152ULL, "hlac_lt_gallowstemplar_n_2388_0z.dds"},
	{0xAA5EAFCBFAECDDC6ULL, "dae230st_coterie_contact.stg"},
	{0xAA5F01BF500FDE77ULL, "nat380_sibling6207656.cl"},
	{0xAA6021254132D8C1ULL, "hlac_ca_cavern1_haunted_24424_4y.dds"},
	{0xAA60B1E7E9FEA68CULL, "330385044.ogg"},
	{0xAA60ECE5713D6EE5ULL, "vxht_mansion2b_347_0_0.vlm"},
	{0xAA61AB639011C4EAULL, "vxw_woundedcoast_1dj_0_0.vlm"},
	{0xAA622AAC3C835746ULL, "character_character_p1.vsh.fxo"},
	{0xAA6250CB14C44E09ULL, "vxlt_0_lowtown_a2_n_c_hno_0_0.vlm"},
	{0xAA62EAA10C6BCE99ULL, "vxw_bonepitmining_a2_1is_0_0.vlm"},
	{0xAA63452102B76254ULL, "6105996_m.fxe"},
	{0xAA638F0A2DDA99BAULL, "gen00fl_anders6125486.cl"},
	{0xAA63BB420995631CULL, "art\\characters\\playercharacter\\elffemale\\arm_hvyn.phy"},
	{0xAA63CCA3149D01FCULL, "gen00fl_fenris6126357.cl"},
	{0xAA640FAE08E38CBEULL, "vxht_0_hightown_a2_bwx_0_0.vlm"},
	{0xAA64600367C9C460ULL, "6153230_m.fxe"},
	{0xAA64A6B8EEECD0AEULL, "vxht_keep_70l_0_0.vlm"},
	{0xAA64ADC714407C5AULL, "6128292_m.fxe"},
	{0xAA6500586959FB7BULL, "vxht_mansion_n_625_0_0.vlm"},
	{0xAA652321FCBB172DULL, "vxlt_hangedman_n_1tx_0_0.vlm"},
	{0xAA6550842B54DCA6ULL, "art\\levels\\races\\proxy\\props\\_textures\\logpile_sa.dds"},
	{0xAA6551724CF6BA55ULL, "qun221_petrice6075982.cl"},
	{0xAA6583ED9D6A9C70ULL, "216943405.ogg"},
	{0xAA658F4835981400ULL, "6115962_m.fxe"},
	{0xAA65A4599B71FF07ULL, "6023866_m.fxe"},
	{0xAA65D5CE98C73BBDULL, "zz_qcr_debug6052117.cl"},
	{0xAA67125FA7BD872DULL, "nat380tr_sibling_fight.ncs"},
	{0xAA67173F98D0BFE0ULL, "6166537_m.fxe"},
	{0xAA679813FE121562ULL, "6067110_f.fxe"},
	{0xAA6843738EE21A91ULL, "676521099.ogg"},
	{0xAA688833D99A5A72ULL, "hlac_lt_0_lowtown_a2_-65536_1z.dds"},
	{0xAA68CD1DB89C47E0ULL, "zz_dog_debug6204128.cl"},
	{0xAA68F6C4036986A8ULL, "mag170_thrask6016949.cl"},
	{0xAA690C33F9687D5BULL, "407568434.ogg"},
	{0xAA690C8BF3B64AF9ULL, "sh_cyc.mao"},
	{0xAA692A255F2E41C0ULL, "hlac_ca_cave_2194_7z.dds"},
	{0xAA69E0FCD4E264C0ULL, "hlac_ca_caverns1_24423_0x.dds"},
	{0xAA6A5925796AFD3DULL, "gen_im_trs_gem_rosediam_01.uti"},
	{0xAA6B022AA6E35731ULL, "vxw_sundermount_33w_0_1.vlm"},
	{0xAA6B03F05A851A61ULL, "6085082_m.fxe"},
	{0xAA6B0825796AE546ULL, "hlac_lt_gallowstemplar_f_2392_1z.dds"},
	{0xAA6B0CCA5E4DBFDFULL, "394101577.ogg"},
	{0xAA6B4458018BC380ULL, "443183522.ogg"},
	{0xAA6B98B8AF33BCDCULL, "6106672_m.fxe"},
	{0xAA6BBA0FBE1B4C2FULL, "6043330_m.fxe"},
	{0xAA6BDE45CC9670F5ULL, "6129432_m.fxe"},
	{0xAA6BE881B657F700ULL, "vxht_chantry_272_0_0.vlm"},
	{0xAA6C0C1645A65F30ULL, "6110064_m.fxe"},
	{0xAA6C39D8A85FBA38ULL, "vxht_0_hightown_9zb_0_0.vlm"},
	{0xAA6C673B6CB29963ULL, "489894641.ogg"},
	{0xAA6CD9D4F5B31A32ULL, "vrc321ip_spooky6_obj_7.utp"},
	{0xAA6D0B59A5BF23B9ULL, "vxlt_gallowstemplar_f_5v3_0_0.vlm"},
	{0xAA6DD00410A65D97ULL, "vxw_encounter2_pg_0_0.vlm"},
	{0xAA6E084DC8E4AA3FULL, "lt_0_lowtown_docks_3_0_lc.mmh"},
	{0xAA6E89C1329B4695ULL, "vxlt_0_lowtown_docks_a3_n_cuq_0_0.vlm"},
	{0xAA6EEF5FDF16CCF3ULL, "dre000_stroud6036739.cl"},
	{0xAA6F24D6577282BCULL, "hlac_ca_caverns1_mines_24427_3z.dds"},
	{0xAA7000D0844C7935ULL, "6124371-mrl_pre_romance.cut"},
	{0xAA701D408207DD6AULL, "vxw_woundedcoast_1ok_0_0.vlm"},
	{0xAA7031911278E31EULL, "vxlt_gallowstemplar_f_5aq_0_0.vlm"},
	{0xAA704C0713A45339ULL, "27119069.ogg"},
	{0xAA704C1B24C5A1B9ULL, "127317358.ogg"},
	{0xAA706B98BDA96572ULL, "follower_banter6160774.cl"},
	{0xAA708CA748E9DE5FULL, "64790322.ogg"},
	{0xAA70B1161A8ECB9FULL, "vxlt_0_lowtown_a3_n_c_epc_0_0.vlm"},
	{0xAA70C927695DEA03ULL, "vxht_bartrand_n_4nt_0_0.vlm"},
	{0xAA70F771DE9D7CFEULL, "llac_w_sundermount_196607_0.dds"},
	{0xAA714DF258DC5DCFULL, "6104304_f.fxe"},
	{0xAA714E11E8DC9F75ULL, "vxht_bartrand_n_68b_0_0.vlm"},
	{0xAA7169C8CA8A38C8ULL, "6131098_m.fxe"},
	{0xAA71857A6A3BDE98ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb106a_0n.dds"},
	{0xAA71CF272755D66BULL, "145952775.ogg"},
	{0xAA71D64915BB2067ULL, "508399074.ogg"},
	{0xAA72397D0951A8A0ULL, "600862244.ogg"},
	{0xAA7394FEEEB12FADULL, "vxlt_hangedman_n_1qw_0_0.vlm"},
	{0xAA739D39DBD3C2D4ULL, "6119716_f.fxe"},
	{0xAA73A07D97A85F6FULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03p1ruinscol.phy"},
	{0xAA743BCCAFF2EBB5ULL, "vxht_keep_n_73u_0_0.vlm"},
	{0xAA745F459CF262AFULL, "280391561.ogg"},
	{0xAA747FE6BF3B4FAFULL, "dae000_bartrand6031149.cl"},
	{0xAA74A5CC604133CEULL, "vxlt_undercityserialkiller_1sv_0_0.vlm"},
	{0xAA754DCDF0052400ULL, "792605872.ogg"},
	{0xAA75537E6A6C7316ULL, "hlac_lt_0_lowtown_a2_n_c_-3_3y.dds"},
	{0xAA75CA7CB2AE588EULL, "mag111ip_backroom_door1.utp"},
	{0xAA768FD1FCA37E7DULL, "vxca_cavern1_haunted_im3_0_0.vlm"},
	{0xAA78016BBCF73658ULL, "1027369004.ogg"},
	{0xAA78042981D3E3BCULL, "vxlt_0_lowtown_docks_a3_coa_0_0.vlm"},
	{0xAA78D390DED665D4ULL, "dae000_arishok6078463.cl"},
	{0xAA78E3104316A10EULL, "325382665.ogg"},
	{0xAA799190B9D51ACFULL, "vxdr_0_deeproads_cuq_0_0.vlm"},
	{0xAA7998E75C230285ULL, "vxht_keep_42n_0_0.vlm"},
	{0xAA7B90BB3185765CULL, "ser270_templar6116089.cl"},
	{0xAA7B9DFE2F57B1B4ULL, "6120957_m.fxe"},
	{0xAA7C3E5A856D0F92ULL, "vxlt_gallowsprison_n_6ha_0_0.vlm"},
	{0xAA7CE347EB8BB5CFULL, "6074431_m.fxe"},
	{0xAA7D3FAC8FF7704AULL, "362873829.ogg"},
	{0xAA7DA283B1FF14ABULL, "t1_mue_gr1.tnt"},
	{0xAA7DF1DBBBACD234ULL, "6096533_m.fxe"},
	{0xAA7E37FA4F0F8941ULL, "mag100_cullen6140983.cl"},
	{0xAA7E584CDB4C0E35ULL, "6061417_m.fxe"},
	{0xAA7E8F1D73FDF5EAULL, "vxw_woundedcoast_1jd_0_0.vlm"},
	{0xAA7E92FEBB589B86ULL, "766628513.ogg"},
	{0xAA7EB768315CAEC1ULL, "llac_w_encounter2_-131072_0.dds"},
	{0xAA7F12FC7745348EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_head.evt"},
	{0xAA7F818310DD3B5EULL, "follower_banter6154045.cl"},
	{0xAA7FCD72BAA3BEFDULL, "895697706.ogg"},
	{0xAA805E549CCC2E56ULL, "6105077_m.fxe"},
	{0xAA80996D85036BEDULL, "x_partypicker.arl"},
	{0xAA80AB5D5E18C505ULL, "lt_warehouse_2_0_lc.mmh"},
	{0xAA8134B2947F3703ULL, "6208500_m.fxe"},
	{0xAA8179A7FF87B113ULL, "6035951_m.fxe"},
	{0xAA81C6C2C0E27FC8ULL, "hlac_lt_0_lowtown_docks_a3_131071_4x.dds"},
	{0xAA82EB60987BE1BBULL, "470435213.ogg"},
	{0xAA833D6F50E8217EULL, "6150062_m.fxe"},
	{0xAA8345E355F39CEEULL, "vxca_caverns1_ghx_0_0.vlm"},
	{0xAA8380E3AA61C60DULL, "w_encounter3_220_0.mmh"},
	{0xAA8384E3AA61CD5EULL, "w_encounter3_220_0.mao"},
	{0xAA8392E3AA61E493ULL, "w_encounter3_220_0.msh"},
	{0xAA83B8D01D16C00EULL, "dae000_madam_lusine6151218.cl"},
	{0xAA83E702479BEFB3ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.dura.evt"},
	{0xAA844AB227D036C1ULL, "574685329.ogg"},
	{0xAA8501AEEE9996FDULL, "vxca_cavern1_haunted_ij2_0_0.vlm"},
	{0xAA85722D7C3596C1ULL, "140229083.ogg"},
	{0xAA8588D31E5DF219ULL, "dae010_alienage_elf_f1.utc"},
	{0xAA85ABAD54A7E278ULL, "dae011_gossips6143010.cl"},
	{0xAA8609FC774B4439ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_head.gad"},
	{0xAA863E9DE3AAE40BULL, "vxlt_undercityencounter_1q1_0_0.vlm"},
	{0xAA871AC2C689AD30ULL, "gen_im_trs_bst_kidney_04.uti"},
	{0xAA877AC78A98FBCBULL, "177636511.ogg"},
	{0xAA87EE60EE0E9C68ULL, "6055059_m.fxe"},
	{0xAA88873CBDB9EA71ULL, "6125425_m.fxe"},
	{0xAA88952D0355938EULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_spiderweb_da.dds"},
	{0xAA89358653B59659ULL, "6063948_f.fxe"},
	{0xAA8946C220403CBDULL, "vxdr_primevaltomb_6j0_0_1.vlm"},
	{0xAA896477FBE6863CULL, "hlac_w_blightlands_65537_0y.dds"},
	{0xAA89A13ADE7C3753ULL, "vxht_0_hightown_a3_cbv_0_0.vlm"},
	{0xAA8A82894307A5D1ULL, "333837378.ogg"},
	{0xAA8AE00247A202F4ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.dura.gad"},
	{0xAA8B0F694BAFC507ULL, "6105450_m.fxe"},
	{0xAA8B6020C1322636ULL, "603595222.ogg"},
	{0xAA8B84718CDFFE02ULL, "6097608_m.fxe"},
	{0xAA8BD891D34C0331ULL, "vxlt_blackemporium_w3_0_0.vlm"},
	{0xAA8C2D7D6522F5B1ULL, "662289266.ogg"},
	{0xAA8C85027B25DC0AULL, "6057034_f.fxe"},
	{0xAA8C9E7E5B8A8571ULL, "20138940.ogg"},
	{0xAA8CA351D44F6CAAULL, "vxlt_gallowstemplar_n_3nj_0_0.vlm"},
	{0xAA8CCEF5DF6DEB4AULL, "ts_qunari_acc_props.gda"},
	{0xAA8D154AF665141EULL, "vxlt_gallowstemplar_f_5os_0_0.vlm"},
	{0xAA8D1AFC77517F87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_scratch_head.ani"},
	{0xAA8D6549417BFD43ULL, "vxht_keep_n_5jv_0_0.vlm"},
	{0xAA8DA32D28AFD6D5ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_hed_0s.dds"},
	{0xAA8DA57CB2C2AC73ULL, "mag111ip_backroom_door1.ncs"},
	{0xAA8E5555DC141327ULL, "6035209_m.fxe"},
	{0xAA8EF6E70D48E0E7ULL, "6034432_f.fxe"},
	{0xAA8F0FD33D0AF810ULL, "lowtown_argument6169411.cl"},
	{0xAA8F52212B5B7048ULL, "970647155.ogg"},
	{0xAA8F99B046EF5BF3ULL, "576741245.ogg"},
	{0xAA8FD96CD1814DF7ULL, "6095658_m.fxe"},
	{0xAA900B3896DE8822ULL, "315832064.ogg"},
	{0xAA9053D861FACEA5ULL, "gen_im_trs_krk_hgh_junk3_03.uti"},
	{0xAA912FCE3BDF077EULL, "gen00fl_anders6119203.cl"},
	{0xAA9169862095755DULL, "6080751_m.fxe"},
	{0xAA91C11C8787C53BULL, "dae000cr_grand_cleric6147882.cl"},
	{0xAA91C9828C65010DULL, "6062789_f.fxe"},
	{0xAA91EF0247A83AD2ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.dura.ani"},
	{0xAA920E4AA6B88966ULL, "743243870.ogg"},
	{0xAA9251C16DD3A923ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_glyph_of_paralysis.evt"},
	{0xAA92B7A10BB7B1D1ULL, "hm_cgn_003.mop"},
	{0xAA92C3D76EE07C58ULL, "vxw_bonepitmining_a2_1l1_0_0.vlm"},
	{0xAA92C92B81E25D0EULL, "199570492.ogg"},
	{0xAA9323EF71ADEFAEULL, "mcr370_meredith6085925.cl"},
	{0xAA941BB0E585E824ULL, "zz_followers_banter6157303.cl"},
	{0xAA9449ACB45992E8ULL, "6090877_f.fxe"},
	{0xAA9491C6DCB7547CULL, "6080195_f.fxe"},
	{0xAA94A5779F9BA3BDULL, "6056824_m.fxe"},
	{0xAA94EA304F8E567CULL, "w_sundermount_13_0.rml"},
	{0xAA95548751DBA36BULL, "6136647_m.fxe"},
	{0xAA95746175FE4D6FULL, "hlac_lt_0_lowtown_a3_n_c_-131072_8y.dds"},
	{0xAA95C79306386900ULL, "vxht_0_hightown_n_c71_0_1.vlm"},
	{0xAA9601359C4BED89ULL, "clf260_harley6173874.cl"},
	{0xAA96AD079A92DAF2ULL, "1600805.ogg"},
	{0xAA97D73D8F3BC5B7ULL, "one220_yevhen6099018.cl"},
	{0xAA97FDA93C8B2AC3ULL, "6135887_f.fxe"},
	{0xAA9830A920D0B9B0ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_011a_0s.dds"},
	{0xAA98D0304F9212D1ULL, "w_sundermount_13_0.msh"},
	{0xAA98DE304F922AECULL, "w_sundermount_13_0.mao"},
	{0xAA98EA304F923E8FULL, "w_sundermount_13_0.mmh"},
	{0xAA9916264E6B564BULL, "docks_male_human_worker_sta6176005.cl"},
	{0xAA992AC16DD985A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_glyph_of_paralysis.gad"},
	{0xAA9A83E638CAB54CULL, "6106921_f.fxe"},
	{0xAA9A90F8F74FD944ULL, "vxlt_backalley_g_1k0_0_0.vlm"},
	{0xAA9A9E5FCFF442D5ULL, "gen00fl_isabela6079076.cl"},
	{0xAA9AB3FC337943ECULL, "vxw_encounter2_camp_44_0_0.vlm"},
	{0xAA9B05687CA54C9BULL, "mag200_feynriel6030420.cl"},
	{0xAA9B0F69890448F8ULL, "963975726.ogg"},
	{0xAA9B152EC645852AULL, "vxlt_gallowstemplar_n_3ek_0_0.vlm"},
	{0xAA9B4DB6A47FA5A6ULL, "mcr371_scared_mage_m6086911.cl"},
	{0xAA9B6D0F1C330A82ULL, "hlac_ht_mansion2_818_25y.dds"},
	{0xAA9B72D757EA26EAULL, "vxw_woundedcoast_1pf_0_0.vlm"},
	{0xAA9BABF009A85C8CULL, "216867066.ogg"},
	{0xAA9C32C01E6258F8ULL, "6059392_m.fxe"},
	{0xAA9C3818E76C9F54ULL, "854062887.ogg"},
	{0xAA9C7D886F0CA17DULL, "vxlt_hangedman_31j_0_0.vlm"},
	{0xAA9CABABD0DA388CULL, "6129349_m.fxe"},
	{0xAA9D9B673999F9C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_right_hand.ani"},
	{0xAA9E26EB5AEF30A7ULL, "vxht_mansion2_2wp_0_1.vlm"},
	{0xAA9F417C4B3533D3ULL, "art\\characters\\creatures\\pridedemon\\animation\\atkbl.ani"},
	{0xAA9F64CC39D9818EULL, "vxht_0_hightown_a3_5tc_0_0.vlm"},
	{0xAA9F8048BBB249D9ULL, "vxdr_primevaltomb_7ks_0_0.vlm"},
	{0xAA9F80A0C8FBACDFULL, "6064531_m.fxe"},
	{0xAA9FA6C6E1D8ABE0ULL, "vxlt_0_lowtown_cmu_0_0.vlm"},
	{0xAA9FCA32F002A590ULL, "97695217.ogg"},
	{0xAA9FD197E7507CD2ULL, "vxdr_0_deeproads_ese_0_1.vlm"},
	{0xAAA039C16DDFBD82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_glyph_of_paralysis.ani"},
	{0xAAA048D2A2D84928ULL, "w_bonepitmining_a2_3_0_ll.mmh"},
	{0xAAA061330ED0032EULL, "dae000ip_crowd_ff3_talk17.utp"},
	{0xAAA0A141F73A213CULL, "zz_mer120ip_at_bone_pit.utp"},
	{0xAAA0B4C2A2B85164ULL, "vxlt_0_lowtown_a3_n_c_hq2_0_0.vlm"},
	{0xAAA0DBC22B4A725DULL, "6013258_m.fxe"},
	{0xAAA1F95E6719A67DULL, "6040579_m.fxe"},
	{0xAAA1F9B2AFEDC474ULL, "823795532.ogg"},
	{0xAAA209ECC2831BC8ULL, "ico_mm_plotdest.dds"},
	{0xAAA2CF9297D8D4B9ULL, "rdr110dg_dockworker6015958.cl"},
	{0xAAA30B3894AC5CCBULL, "577641535.ogg"},
	{0xAAA33C979700F3DFULL, "6139080_m.fxe"},
	{0xAAA34D1521C1F207ULL, "hlac_ht_0_hightown_a3_65533_0y.dds"},
	{0xAAA3E27234B7C168ULL, "gen00fl_isabela6099460.cl"},
	{0xAAA446CF839929C9ULL, "vxht_mansion2_1aj_0_0.vlm"},
	{0xAAA4986739A01326ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_right_hand.gad"},
	{0xAAA4A0A560855CDEULL, "sse_ss_raider_m_2.gda"},
	{0xAAA50A2B5866C37CULL, "6204165_m.fxe"},
	{0xAAA59712801FC0C1ULL, "837469583.ogg"},
	{0xAAA5B4C7DA1A77F8ULL, "mrl250_pol6101505.cl"},
	{0xAAA5B6ADABD64B17ULL, "follower_banter6163897.cl"},
	{0xAAA6076C29CB904EULL, "hlac_lt_0_lowtown_docks_a3_-196610_0y.dds"},
	{0xAAA615762C69AE77ULL, "6128051_m.fxe"},
	{0xAAA649759A36CEDAULL, "vxlt_gallowstemplar_n_5z0_0_0.vlm"},
	{0xAAA668A7733968A8ULL, "vxht_bartrand_n_647_0_0.vlm"},
	{0xAAA681A3E61D016DULL, "vxlt_backalley_n_1u6_0_0.vlm"},
	{0xAAA6D2CD6D9785B9ULL, "vxlt_gallowstemplar_f_5j7_0_0.vlm"},
	{0xAAA6DAFF5F829B7EULL, "542407022.ogg"},
	{0xAAA70F7F9BEFD022ULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_bonepile_d.dds"},
	{0xAAA74F21A71DF264ULL, "6096228_f.fxe"},
	{0xAAA77EC215153177ULL, "6175725_m.fxe"},
	{0xAAA7CF4D89ABC766ULL, "mer320_hubert6053206.cl"},
	{0xAAA7EAD402C2F417ULL, "vxw_blightlands_6n_0_0.vlm"},
	{0xAAA89780BE55BC6DULL, "lowtown_fem_prostitute_sta6161618.cl"},
	{0xAAA96C1F7FC14F6CULL, "571234556.ogg"},
	{0xAAA9A441B5032642ULL, "6131889_m.fxe"},
	{0xAAAA1AD8A99BBA34ULL, "rdr110dg_guard_sleeping6016738.cl"},
	{0xAAAA875FAA20143FULL, "428522877.ogg"},
	{0xAAAA8A1BCE92C889ULL, "vxlt_0_lowtown_a2_n_c_hur_0_0.vlm"},
	{0xAAAA8F6035C7AFB7ULL, "dre000_varric6027777.cl"},
	{0xAAAA9D8A7C2869C6ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.impa.evt"},
	{0xAAAAACD93638396BULL, "6079285_m.fxe"},
	{0xAAAAC30636C44F34ULL, "6131164_m.fxe"},
	{0xAAAB23F1C63A4919ULL, "vxw_bonepitmining_a2_1rv_0_0.vlm"},
	{0xAAAB5E6389994DC7ULL, "503214125.ogg"},
	{0xAAAB6B118592469BULL, "art\\characters\\creatures\\dog\\textures\\c_eye_dogami_0n.dds"},
	{0xAAAB8CB463C4CDA0ULL, "6170758_m.fxe"},
	{0xAAAB936739A62905ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_gesture_right_hand.evt"},
	{0xAAAC397C4B3FB00AULL, "art\\characters\\creatures\\pridedemon\\animation\\atkbl.evt"},
	{0xAAAC7DA94CC0B66FULL, "darktown_male_elf_sta6174567.cl"},
	{0xAAAD291AB87EBE9CULL, "vfx_glow_p.mmh"},
	{0xAAADA42CCAF7E833ULL, "vxht_chantry_35r_0_0.vlm"},
	{0xAAADBE055F894212ULL, "gen00fl_merrill6150703.cl"},
	{0xAAADE0B66C7AEFF3ULL, "vrc321cr_guard_ghost.ncs"},
	{0xAAAE689B37F5CE78ULL, "vxlt_hangedman_21f_0_0.vlm"},
	{0xAAAE89F7D694DFA5ULL, "6135236_f.fxe"},
	{0xAAAECC41FB8C3A22ULL, "vxlt_0_lowtown_h5l_0_0.vlm"},
	{0xAAAEEA60807F23E2ULL, "dae100_bran6118313.cl"},
	{0xAAAFCB04A00D5019ULL, "pro000_colonel_ewald6040537.cl"},
	{0xAAAFFEAFDCFEC5D4ULL, "vxht_0_hightown_a3_c8k_0_0.vlm"},
	{0xAAB05E8D57A3E708ULL, "vxca_cave_2kz_0_0.vlm"},
	{0xAAB0755570187DF6ULL, "hlac_lt_darktown_-1_60x.dds"},
	{0xAAB09BD2B03DD42CULL, "vxht_0_hightown_a2_c5u_0_0.vlm"},
	{0xAAB0B6D2372CEC03ULL, "w_blightlands_211_0.msh"},
	{0xAAB0C4D2372D045DULL, "w_blightlands_211_0.mmh"},
	{0xAAB0C8D2372D0B6EULL, "w_blightlands_211_0.mao"},
	{0xAAB17A8763B5AAB4ULL, "6121118_m.fxe"},
	{0xAAB1839A45ECE469ULL, "274420350.ogg"},
	{0xAAB19873DDF7525DULL, "982436344.ogg"},
	{0xAAB1A63D4BE7B24AULL, "450730164.ogg"},
	{0xAAB1A97B292ACCB4ULL, "6079066_m.fxe"},
	{0xAAB1B48A7C2EAF11ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.impa.gad"},
	{0xAAB1FE1F692F691FULL, "916828795.ogg"},
	{0xAAB212F20A906D40ULL, "140804045.ogg"},
	{0xAAB232EFBE538054ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_coma_0s.dds"},
	{0xAAB253986B5F8270ULL, "6086183_m.fxe"},
	{0xAAB2E5ECFBDD3119ULL, "491404982.ogg"},
	{0xAAB3187C4B45977DULL, "art\\characters\\creatures\\pridedemon\\animation\\atkbl.gad"},
	{0xAAB326E0DA8A1AB7ULL, "art\\vfx\\combat\\vfx_m_bld_sacrifice_imp_c.anb"},
	{0xAAB36CBCCE074631ULL, "276890931.ogg"},
	{0xAAB3B5492FD6D222ULL, "vxht_chantry_n_2zd_0_0.vlm"},
	{0xAAB3C2044E1266E8ULL, "6094724_m.fxe"},
	{0xAAB3D33862FA7E87ULL, "6030475_m.fxe"},
	{0xAAB4654C3FC03704ULL, "500257663.ogg"},
	{0xAAB5246C612B1A47ULL, "6143364_m.fxe"},
	{0xAAB56880DEED7786ULL, "6116114_m.fxe"},
	{0xAAB5F112C11FDE98ULL, "docks_fem_sailor_mob6176313.cl"},
	{0xAAB6300AD3BC3568ULL, "6105822_m.fxe"},
	{0xAAB66FBC4EFCA146ULL, "268420471.ogg"},
	{0xAAB67A37E3A69FE3ULL, "vxht_playershouse_n_1zw_0_0.vlm"},
	{0xAAB6E73079A7E3FBULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screena_1.mao"},
	{0xAAB6F9C345F74DBFULL, "vxw_encounter2_camp_98_0_0.vlm"},
	{0xAAB74EC97C1B8E07ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium_trim1.gad"},
	{0xAAB7F5E0EE2EF11BULL, "zz_followers_banter6157571.cl"},
	{0xAAB8A24AB6FE9DACULL, "6153472_m.fxe"},
	{0xAAB8B58A7C34CFBFULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.impa.ani"},
	{0xAAB8B9524FC36769ULL, "1030670185.ogg"},
	{0xAAB9890F5B64A42EULL, "173483949.ogg"},
	{0xAAB9C0AD3F6B2616ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsbuild.phy"},
	{0xAABA37E4C9B915ECULL, "gen00fl_isabela6103965.cl"},
	{0xAABA636E843A6C5BULL, "vxca_cavern1_haunted_fmb_0_0.vlm"},
	{0xAABA8621ACD81FAAULL, "541723408.ogg"},
	{0xAABB2856FADFAE85ULL, "391774011.ogg"},
	{0xAABB5252442F4271ULL, "hlac_lt_0_lowtown_docks_-65538_6y.dds"},
	{0xAABB7B4FA183C201ULL, "386079853.ogg"},
	{0xAABB91948DAE6982ULL, "6127046_m.fxe"},
	{0xAABBF2884FD13A76ULL, "6167027_f.fxe"},
	{0xAABBF56033F3E83AULL, "241114483.ogg"},
	{0xAABC8AB72DCCA6D0ULL, "vxht_mansion2_2f9_0_0.vlm"},
	{0xAABD88EC84036658ULL, "185302520.ogg"},
	{0xAABD8B884D29A973ULL, "504205494.ogg"},
	{0xAABDA37A572201DEULL, "vxlt_0_lowtown_a3_n_c_e11_0_0.vlm"},
	{0xAABE43C97C219A14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium_trim1.evt"},
	{0xAABF0DAFA233ECACULL, "vxht_0_hightown_a2_c0j_0_0.vlm"},
	{0xAABF28DA2B48D6E4ULL, "143698759.ogg"},
	{0xAABF361C1F1BD78EULL, "vxw_woundedcoast_1sz_0_1.vlm"},
	{0xAABF5FFF6EC4C907ULL, "525607459.ogg"},
	{0xAAC055D174280AFAULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrock_da.dds"},
	{0xAAC06781C3B2CD37ULL, "6205975_f.fxe"},
	{0xAAC07ADA7640E17BULL, "vxca_cave_1ne_0_0.vlm"},
	{0xAAC0D76DC8532812ULL, "vxht_mansion_n_4dl_0_0.vlm"},
	{0xAAC10A7945122EB8ULL, "977780437.ogg"},
	{0xAAC2C952A604205DULL, "qun120_arishok6049636.cl"},
	{0xAAC3253CC89413AEULL, "6064941_m.fxe"},
	{0xAAC34AE428539FA9ULL, "336106017.ogg"},
	{0xAAC371C1A724E7A1ULL, "6137534_m.fxe"},
	{0xAAC3A9AD3F7353B7ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsbuild.msh"},
	{0xAAC3BBAD3F73721AULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsbuild.mao"},
	{0xAAC3BFAD3F737929ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsbuild.mmh"},
	{0xAAC3CD8D3C55A1EEULL, "vxlt_0_lowtown_docks_a3_d3d_0_0.vlm"},
	{0xAAC4ED3B65FA996EULL, "6146793_m.fxe"},
	{0xAAC648619D8CDE38ULL, "vxht_chantry_n_26z_0_0.vlm"},
	{0xAAC66B8D74F3CAC5ULL, "qcr210cr_guardman.utc"},
	{0xAAC6727C2406273EULL, "784919439.ogg"},
	{0xAAC74ED2D8592916ULL, "598825669.ogg"},
	{0xAAC7F9F501C750F8ULL, "174938563.ogg"},
	{0xAAC8095E45C647D3ULL, "hlac_lt_gallowscourtyard_-131071_2y.dds"},
	{0xAAC81DB6150AF3D0ULL, "236591343.ogg"},
	{0xAAC8806BC009E458ULL, "vxht_playershouse_n_1iq_0_0.vlm"},
	{0xAAC8ED5558BE6B5CULL, "hlac_lt_shop_-1_0z.dds"},
	{0xAAC960CE9BE38F87ULL, "625247874.ogg"},
	{0xAAC9A4C7CDA8DAB1ULL, "vxht_mansion2b_25e_0_0.vlm"},
	{0xAAC9CE8D3BC13EEFULL, "gen00fl_anders6111118.cl"},
	{0xAAC9F439494D6E9EULL, "llac_w_bonepitmining_a2_4_0.dds"},
	{0xAACA239D4ED1D03EULL, "838824759.ogg"},
	{0xAACA8DA14FC23B26ULL, "mrl150_merrill6093522.cl"},
	{0xAACA95BE8D1E044BULL, "563850543.ogg"},
	{0xAACAB8C03E156F79ULL, "6077191_m.fxe"},
	{0xAACBF50FD8C38D2EULL, "vxw_encounter1_ex_0_0.vlm"},
	{0xAACC23A1574AD8E2ULL, "plt_lgt310pt_fixer_2.plo"},
	{0xAACC2D7EC5841BEEULL, "vxdr_primevaltomb_5am_0_1.vlm"},
	{0xAACC3BC97C2DCA4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_weightshift_medium_trim1.ani"},
	{0xAACCE269C0EAF6D4ULL, "vxht_0_hightown_a3_c6i_0_0.vlm"},
	{0xAACD021C910FE194ULL, "llac_lt_gallowstemplar_f_2388_2.dds"},
	{0xAACD0BC2B95BF371ULL, "gen_im_trs_qun_sta_nat_r2.uti"},
	{0xAACDA3CB2E149CF9ULL, "gen00fl_anders6125829.cl"},
	{0xAACE06912CF8D125ULL, "hlac_w_blightlands_-65534_0z.dds"},
	{0xAACE820A78BF7BE5ULL, "849379467.ogg"},
	{0xAAD11807545A8667ULL, "vxlt_0_lowtown_a2_n_gco_0_0.vlm"},
	{0xAAD170D30214C6A2ULL, "vxca_cavern1_haunted_eap_0_0.vlm"},
	{0xAAD19F0C7019185AULL, "vxlt_gallowstemplar_n_5k3_0_0.vlm"},
	{0xAAD1D8F51FAC90BCULL, "seb201_lady_harimann6177408.cl"},
	{0xAAD23F6A2E4BBA6EULL, "vxlt_0_lowtown_docks_a3_d4e_0_0.vlm"},
	{0xAAD24B629E2A10A2ULL, "6072848_f.fxe"},
	{0xAAD26EA9256D9899ULL, "vxht_playershouse_n_1p7_0_0.vlm"},
	{0xAAD2A9A59F1DBADCULL, "vxht_chantry_27n_0_0.vlm"},
	{0xAAD30D7625F53A5CULL, "hlac_lt_gallowscourtyard_-65535_1x.dds"},
	{0xAAD35DF1C792D0E3ULL, "vxht_playershouse_n_1ty_0_0.vlm"},
	{0xAAD37571CCB9303CULL, "zz_t_sir_i_three_v1.cut"},
	{0xAAD38F5733F1ACB9ULL, "6155853_m.fxe"},
	{0xAAD3A3FDC3AC6520ULL, "6133513_m.fxe"},
	{0xAAD426F6E9929E29ULL, "zz_followers_banter6158532.cl"},
	{0xAAD4B767CACD9BE0ULL, "hlac_lt_0_lowtown_a2_n_-131071_2x.dds"},
	{0xAAD4E73FDC3CC906ULL, "vxht_chantry_n_215_0_0.vlm"},
	{0xAAD5F1E71E920B07ULL, "81581844.ogg"},
	{0xAAD6228314D7F3F5ULL, "vrc321ip_spooky7_obj_1.utp"},
	{0xAAD6633F7E01F6B6ULL, "mcr310_party_banter6087821.cl"},
	{0xAAD6B51505FD726FULL, "6037092_m.fxe"},
	{0xAAD6FEC056517894ULL, "42560922.ogg"},
	{0xAAD71D5EFFEAD91DULL, "qun221_petrice6083712.cl"},
	{0xAAD75D3DDED80801ULL, "zz_warrior1.cut"},
	{0xAAD7821451FBB231ULL, "204290642.ogg"},
	{0xAAD7859C27E0A5DBULL, "ts_beasts_prog_kidney.gda"},
	{0xAAD79E65EDE5E876ULL, "vxlt_gallowscourtyard_n_6y4_0_0.vlm"},
	{0xAAD7E9F512C29D62ULL, "6089427_m.fxe"},
	{0xAAD829CBA781B451ULL, "6077595_m.fxe"},
	{0xAAD88E6D05287465ULL, "gen00fl_isabela6085574.cl"},
	{0xAAD8DAD7CB17F357ULL, "822335918.ogg"},
	{0xAAD91628A97185FAULL, "262874911.ogg"},
	{0xAAD92A9D300B87F9ULL, "6078091_m.fxe"},
	{0xAAD963A05E3D06FAULL, "mag101_grace6034177.cl"},
	{0xAADA4AB158957DE3ULL, "ht_0_hightown_a3_26_0.rml"},
	{0xAADA66ECCAB9A5AEULL, "vxw_encounter1_jy_0_0.vlm"},
	{0xAADA7C366A90B364ULL, "vxlt_0_lowtown_a3_n_c_hm6_0_0.vlm"},
	{0xAADB1533CDC7C6F0ULL, "dae000ip_crowd_em4_talk12.utp"},
	{0xAADB3950A4B3428FULL, "6103653_f.fxe"},
	{0xAADBECABC7B79FD0ULL, "6108803_m.fxe"},
	{0xAADC189BBDBCA66AULL, "gen00fl_aveline6053353.cl"},
	{0xAADC23E0DAAD1AB8ULL, "art\\vfx\\combat\\vfx_m_bld_sacrifice_imp_c.mmh"},
	{0xAADC2844963C23BAULL, "vxlt_warehouse_n_ac_0_0.vlm"},
	{0xAADC41E6FB6470C9ULL, "861329561.ogg"},
	{0xAADC75D383841868ULL, "vxht_bartrand_n_5sa_0_0.vlm"},
	{0xAADC7BE9A2195173ULL, "qbar.ncs"},
	{0xAADC92140EB22185ULL, "vxht_mansion_n_5i5_0_0.vlm"},
	{0xAADCF3BDC56D8459ULL, "406428698.ogg"},
	{0xAADE3EAD6F870E43ULL, "6034365_m.fxe"},
	{0xAADE7EB5B1A3F3D0ULL, "vxht_0_hightown_a3_ang_0_0.vlm"},
	{0xAADEC59E7F30F6F7ULL, "dae200_saemus6142715.cl"},
	{0xAADEEE1F0DF618D5ULL, "6106167_m.fxe"},
	{0xAADF5844D815B447ULL, "rdr110dg_guard_warehouse6016689.cl"},
	{0xAADFC511B6A4B82CULL, "6087782_m.fxe"},
	{0xAAE0315044708CABULL, "vxht_mansion_n_62e_0_0.vlm"},
	{0xAAE03D5D73CC17E2ULL, "200770257.ogg"},
	{0xAAE0A1962FE67DA5ULL, "llac_lt_0_lowtown_docks_a3_n_131070_3.dds"},
	{0xAAE0CB7E15669CD6ULL, "6067077_f.fxe"},
	{0xAAE0E0861763B119ULL, "vxht_playershouse_n_1w6_0_0.vlm"},
	{0xAAE11B829113D35CULL, "vxht_chantry_20o_0_0.vlm"},
	{0xAAE11C8C6C38FAB2ULL, "llac_lt_0_lowtown_a3_n_c_196607_0.dds"},
	{0xAAE176871F69EF6AULL, "6095912_m.fxe"},
	{0xAAE1ABF925BE8E5CULL, "6044633_m.fxe"},
	{0xAAE26BF9D18FEB28ULL, "vis120_commoner3.utc"},
	{0xAAE2A4D24AD763B4ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower2_meredithsintro_varric.ani"},
	{0xAAE301B6822F220BULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.impa.ani"},
	{0xAAE352FFE28AE7FEULL, "llac_lt_0_lowtown_a2_n_c_-65536_2.dds"},
	{0xAAE3DF44311C6CE2ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2_tool.phy"},
	{0xAAE3F432855FB877ULL, "147977800.ogg"},
	{0xAAE45A7C77E917BFULL, "hlac_w_sundermount_393210_0x.dds"},
	{0xAAE484BC9CE1C2FFULL, "dae000_sketch6179007.cl"},
	{0xAAE48A5312C5BC4EULL, "dae000_pickpocket_scene6175071.cl"},
	{0xAAE5D1D56E6A6781ULL, "6119170_m.fxe"},
	{0xAAE5DFEA3D7F4609ULL, "444474969.ogg"},
	{0xAAE5F3732E89404BULL, "mag330_walter6045982.cl"},
	{0xAAE5FED6888FB19CULL, "gen00fl_fenris6112079.cl"},
	{0xAAE65FD5F6E60D54ULL, "6093744_f.fxe"},
	{0xAAE693467618DF67ULL, "197442743.ogg"},
	{0xAAE6B18D370E029BULL, "dae000_arishok6096347.cl"},
	{0xAAE6BDA516C16962ULL, "453298852.ogg"},
	{0xAAE760B092D78C99ULL, "255528884.ogg"},
	{0xAAE7DF34A83F7DF1ULL, "6166815_m.fxe"},
	{0xAAE90F17C933AC0FULL, "6095984_m.fxe"},
	{0xAAE94190A747F36CULL, "hlac_ca_cave_2194_9x.dds"},
	{0xAAE95EFF00584A53ULL, "dae021_torelin_mob6147944.cl"},
	{0xAAE96BD29C336043ULL, "vxw_bonepit_27f_0_0.vlm"},
	{0xAAE9A1D24ADD7D16ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower2_meredithsintro_varric.gad"},
	{0xAAEA19CA4C1D3A40ULL, "mag170_thrask6016952.cl"},
	{0xAAEA74FC8035B297ULL, "vrc321ip_spooky5_obj_9.utp"},
	{0xAAEB129A517A117CULL, "vxw_sundermount_2bu_0_0.vlm"},
	{0xAAEBBD9331A223E4ULL, "vxw_encounter2_rb_0_0.vlm"},
	{0xAAEC247078D825B6ULL, "w_encounter3_207_0.msh"},
	{0xAAEC2E7078D836B4ULL, "w_encounter3_207_0.mmh"},
	{0xAAEC327078D83DC7ULL, "w_encounter3_207_0.mao"},
	{0xAAEC880998E9F06DULL, "192008782.ogg"},
	{0xAAECE865C5F43785ULL, "6163782_m.fxe"},
	{0xAAECF092A39A0C50ULL, "vxht_0_hightown_a3_akd_0_0.vlm"},
	{0xAAED66377DD99A1EULL, "vxht_bartrand_n_3uw_0_0.vlm"},
	{0xAAEE3FA4C4B27F83ULL, "hlac_lt_0_lowtown_a3_n_c_65537_0x.dds"},
	{0xAAEEE19862771D2FULL, "vrc321_party_barks6096248.cl"},
	{0xAAF001F28FDA9A55ULL, "543321735.ogg"},
	{0xAAF01511F4F0D620ULL, "596029045.ogg"},
	{0xAAF028D82E60B33AULL, "gen00fl_aveline6139660.cl"},
	{0xAAF087B7A7989620ULL, "6139313_m.fxe"},
	{0xAAF088182FBC978DULL, "945860173.ogg"},
	{0xAAF09F645C6FA138ULL, "zz_followers_banter6159150.cl"},
	{0xAAF0E9B6823B3682ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.impa.evt"},
	{0xAAF120CAD5B1884CULL, "6129089_f.fxe"},
	{0xAAF148A29AC0CAD4ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.anb"},
	{0xAAF1D01B8566B7B3ULL, "ser221_gamlen6135295.cl"},
	{0xAAF1E7731E440B4EULL, "ave200_donnic6061283.cl"},
	{0xAAF1FD3A8DB779F4ULL, "zz_vault_debug6200302.cl"},
	{0xAAF3C2ED1004836FULL, "37466177.ogg"},
	{0xAAF44282496831B9ULL, "6125182_m.fxe"},
	{0xAAF48CE9A22DFFA5ULL, "qbar.gda"},
	{0xAAF4CDAEF6AECE62ULL, "6087627_m.fxe"},
	{0xAAF5570C805C5A85ULL, "dae000_arishok6072783.cl"},
	{0xAAF65710DF68A4A9ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzscaffold_0.pwk"},
	{0xAAF675310BDF78D6ULL, "66498694.ogg"},
	{0xAAF702886727C641ULL, "6084442_f.fxe"},
	{0xAAF7117C8FA41C26ULL, "602069425.ogg"},
	{0xAAF757CBA843838DULL, "vxca_caverns2_83s_0_1.vlm"},
	{0xAAF77F266F8C2C88ULL, "6206325_m.fxe"},
	{0xAAF7B14F2ACAB0CCULL, "gen00fl_aveline6121951.cl"},
	{0xAAF7DEBF7AD84E5CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\trap_lowtownext_12.pwk"},
	{0xAAF7E8B682415395ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.impa.gad"},
	{0xAAF7F258ADF1BA0BULL, "art\\vfx\\combat\\vfx_glb_ele_n_major_c.impa.ani"},
	{0xAAF8A88F15357638ULL, "mer140_bonepit_comments6062989.cl"},
	{0xAAF94A9F2612A5A8ULL, "6103341_m.fxe"},
	{0xAAF9FCC0853EB3E4ULL, "hlac_w_bonepitmining_a2_65537_0z.dds"},
	{0xAAFA29A9773FEA0AULL, "468341855.ogg"},
	{0xAAFD13F72EB8A55BULL, "vxlt_0_lowtown_n_byl_0_0.vlm"},
	{0xAAFD1896D3EFB5D0ULL, "vxht_0_hightown_cem_0_0.vlm"},
	{0xAAFD19D0097A65D4ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.anb"},
	{0xAAFD440288BE81B3ULL, "l_dae311ar_tavern_nt.bnk"},
	{0xAAFDF862282D5C8DULL, "114790098.ogg"},
	{0xAAFE27A29ACB1C97ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.mmh"},
	{0xAAFE726ACAD3757FULL, "combat_2h_r1_a1.gda"},
	{0xAAFE73F0E706B08DULL, "hf_mother_h1s3.mor"},
	{0xAAFE9E6CA8E361F4ULL, "vxht_0_hightown_a3_by1_0_0.vlm"},
	{0xAAFEDA7EF0DE96B5ULL, "vxlt_gamlenshouse_n_qf_0_0.vlm"},
	{0xAAFEEF9C17CB6CFAULL, "vxlt_backalley_g_1ji_0_0.vlm"},
	{0xAAFF100607A33B61ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\til_hightown_main_sa.dds"},
	{0xAAFFAD5A21001EA4ULL, "618656206.ogg"},
	{0xAAFFAF23B3DE1B50ULL, "hlac_lt_0_lowtown_docks_-4_3x.dds"},
	{0xAB000B00377C8C50ULL, "hlac_dr_primevaltomb_663_1x.dds"},
	{0xAB00DF5429F5D764ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_chantdoor_sa.dds"},
	{0xAB0118A196C8B425ULL, "6081396_m.fxe"},
	{0xAB014ABB88831E19ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.dura.ani"},
	{0xAB0153F9D1AA7379ULL, "vis120_commoner3.bnk"},
	{0xAB01A921D7871522ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_blackbox.msh"},
	{0xAB01AB21D78718D8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_blackbox.mmh"},
	{0xAB01B721D7872D7BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_blackbox.mao"},
	{0xAB02030579495E92ULL, "lt_0_lowtown_docks_a3_17_0.anb"},
	{0xAB02203BD19FE8B5ULL, "249194642.ogg"},
	{0xAB02711ED16E17A8ULL, "6039228_m.fxe"},
	{0xAB029F38E6B8B53CULL, "vxdr_primevaltomb_2so_0_0.vlm"},
	{0xAB039FBDD2CF94FFULL, "zz_approval_debug6003141.cl"},
	{0xAB03AF80030164E0ULL, "vxca_caverns1_mines_ea2_0_0.vlm"},
	{0xAB03ED1911D54694ULL, "fx_rain01.dds"},
	{0xAB046A7DD85ADF1FULL, "vxca_cavern1_haunted_j4y_0_0.vlm"},
	{0xAB0474FA1FB999F0ULL, "hlac_dr_0_deeproads_5842_14x.dds"},
	{0xAB0475F9D1ACE2D9ULL, "vis120_commoner3.cnv"},
	{0xAB04AA0DFDE4EC6AULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lp.gad"},
	{0xAB04E1E8FCE2EB59ULL, "vxlt_0_lowtown_a3_n_c_gni_0_0.vlm"},
	{0xAB052FBD1ED7AC21ULL, "hlac_w_encounter3_-2_0y.dds"},
	{0xAB05DA58ADFDCE82ULL, "art\\vfx\\combat\\vfx_glb_ele_n_major_c.impa.evt"},
	{0xAB05F43B051A9343ULL, "hlac_ht_bartrand_n_2995_3z.dds"},
	{0xAB069C25A448D038ULL, "gen00fl_anders6125185.cl"},
	{0xAB06C8B3CB95DB16ULL, "ave200_party_comment6116366.cl"},
	{0xAB06E1E40D604B9AULL, "vxlt_blackemporium_v7_0_0.vlm"},
	{0xAB06E88E4B68A911ULL, "hi_cloud.mao"},
	{0xAB071AD4BFA08CECULL, "hlac_lt_backalley_g_0_13z.dds"},
	{0xAB072BDD5ECE01B9ULL, "llac_lt_gallowstemplar_n_2391_1.dds"},
	{0xAB0767240E7FF2B5ULL, "hightown_market_guard_sta6169953.cl"},
	{0xAB07C6ECC3E90A4FULL, "6095103_f.fxe"},
	{0xAB07D5F2FEA66091ULL, "vxlt_0_lowtown_a2_n_gxx_0_0.vlm"},
	{0xAB07E8A54398731CULL, "21974333.ogg"},
	{0xAB0825BB8888FE5BULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.dura.gad"},
	{0xAB08BA44313B4996ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2_tool.mao"},
	{0xAB08BDF5CAAF7176ULL, "mag311_nyssa6042429.cl"},
	{0xAB08C644313B5E35ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2_tool.mmh"},
	{0xAB08C844313B61ABULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2_tool.msh"},
	{0xAB095E28F496A281ULL, "lgt160_ntb_elf_conflict6163355.cl"},
	{0xAB09F8D00984B797ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.mmh"},
	{0xAB0A5BB43B8849FCULL, "gen00fl_aveline6115866.cl"},
	{0xAB0ABFBD273B1D15ULL, "lt_0_lowtown_docks_a3_2_0_lc.mmh"},
	{0xAB0B8D0DFDEAD901ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lp.evt"},
	{0xAB0BB0D28133194AULL, "vxlt_undercityserialkiller_1a8_0_0.vlm"},
	{0xAB0C2454A78CE7FEULL, "ring_of_thought.dds"},
	{0xAB0CD958AE03EB95ULL, "art\\vfx\\combat\\vfx_glb_ele_n_major_c.impa.gad"},
	{0xAB0D46623494E37AULL, "w_blightlands_105_0.phy"},
	{0xAB0D617909C1857AULL, "vxlt_backalley_g_1wj_0_0.vlm"},
	{0xAB0D723EF2567E8DULL, "636555834.ogg"},
	{0xAB0D839424E4DAF2ULL, "mag121_idunna6083972.cl"},
	{0xAB0DDFE238B816C6ULL, "fx_st_cloudb_add.mao"},
	{0xAB0E5821E859473FULL, "art\\vfx\\cinematic\\vfx_barrel_explosion_p.mmh"},
	{0xAB0EA0F2B25BA415ULL, "jagged_n.dds"},
	{0xAB0F13418A7E540CULL, "hlac_lt_backalley_n_65537_0y.dds"},
	{0xAB0F42BB888F4E40ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.dura.evt"},
	{0xAB10033CDBB535C3ULL, "zz_followers_banter6157607.cl"},
	{0xAB10C82BEFF36B09ULL, "hlac_ht_brothel_n_3365_1x.dds"},
	{0xAB10CCA5BF4567DAULL, "vxlt_0_lowtown_docks_n_bat_0_0.vlm"},
	{0xAB10D068D0B27E67ULL, "vxlt_0_lowtown_docks_n_c4s_0_0.vlm"},
	{0xAB10E5699AD9CB6DULL, "ht_0_hightown_n_17_0_ll.mmh"},
	{0xAB111886B574F543ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd07a.mmh"},
	{0xAB111998966D3A3DULL, "6037630_m.fxe"},
	{0xAB112686B5750C9DULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd07a.msh"},
	{0xAB11952C1CED4ADBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left.phy"},
	{0xAB12619E6927F09FULL, "hlac_lt_blackemporium_1055_7z.dds"},
	{0xAB12C89C81E49AA9ULL, "vxw_woundedcoast_123_0_0.vlm"},
	{0xAB12DC5ACA50F79FULL, "vxca_cavern1_haunted_j3x_0_0.vlm"},
	{0xAB12F50112EC41E0ULL, "vfx_temp3.mao"},
	{0xAB132FC934B0B916ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.ani"},
	{0xAB138D97C8FF2552ULL, "6002007_f.fxe"},
	{0xAB13F7E1176C0028ULL, "773289286.ogg"},
	{0xAB1437226CE08BD1ULL, "135031314.ogg"},
	{0xAB14B0D80DD795CEULL, "624114242.ogg"},
	{0xAB14CC066E0FF1BCULL, "vxht_playershouse_n_1pn_0_0.vlm"},
	{0xAB14DA13B43B91D0ULL, "388474084.ogg"},
	{0xAB14DEC08B02FF14ULL, "6049652_m.fxe"},
	{0xAB14E5A259E4CD4CULL, "vxht_mansion2_1se_0_0.vlm"},
	{0xAB150C078B5D5695ULL, "454641296.ogg"},
	{0xAB151B2F3D41DF23ULL, "6136719_m.fxe"},
	{0xAB15311223CE6A52ULL, "gen_im_arm_glv_med_act1_03.uti"},
	{0xAB153E445C10F5FBULL, "lgt200cr_tome_revenant.utc"},
	{0xAB159B387BAF3155ULL, "501976726.ogg"},
	{0xAB16052FDAF30651ULL, "gen00fl_bethany6101201.cl"},
	{0xAB161448CC8B8E43ULL, "w_woundedcoast_14_0.mao"},
	{0xAB161848CC8B9490ULL, "w_woundedcoast_14_0.mmh"},
	{0xAB162648CC8BACAAULL, "w_woundedcoast_14_0.msh"},
	{0xAB162C484886BC37ULL, "6020557_m.fxe"},
	{0xAB168C66D16FBC95ULL, "vxht_mansion2b_3dw_0_0.vlm"},
	{0xAB16989BDC84C9DFULL, "6064706_m.fxe"},
	{0xAB169A69AA6BBF37ULL, "hlac_lt_0_lowtown_docks_n_-65539_3x.dds"},
	{0xAB17B480A25B7032ULL, "vfx_cs_blend_temp.mao"},
	{0xAB18595B69B09214ULL, "81438691.ogg"},
	{0xAB1892BC25ED63A5ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\prp_zzhumanhair_d.dds"},
	{0xAB189B495ACF158EULL, "vxht_bartrand_n_5rd_0_0.vlm"},
	{0xAB1927917C5046D6ULL, "142245602.ogg"},
	{0xAB19850DFDF70950ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lp.ani"},
	{0xAB198C5951549419ULL, "dr_primevalentrance_probe_2kg.dds"},
	{0xAB199737132F147BULL, "qun120_arishok_guard6049225.cl"},
	{0xAB1A583120AC103BULL, "6155958_m.fxe"},
	{0xAB1A6078271F3BDFULL, "156182680.ogg"},
	{0xAB1AA5F68A27D37AULL, "vxw_bonepitmining_a2_1m0_0_0.vlm"},
	{0xAB1AF41911E831C2ULL, "fx_rain01.mao"},
	{0xAB1B2DB158CCE35FULL, "ht_0_hightown_a3_26_0.anb"},
	{0xAB1B4F45992FEA86ULL, "vxht_mansion_n_3et_0_0.vlm"},
	{0xAB1B89576E70C969ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_loop.ani"},
	{0xAB1B9E4ACF4221F0ULL, "ser200_aveline6201532.cl"},
	{0xAB1BCE68B7FF5FE0ULL, "vxht_0_hightown_a2_bq0_0_0.vlm"},
	{0xAB1D034F195D586BULL, "168891415.ogg"},
	{0xAB1D23E1037EA3C2ULL, "550741936.ogg"},
	{0xAB1D8469FC52A37AULL, "hlac_lt_darktown_-1_28x.dds"},
	{0xAB1E2BEFE51F397DULL, "28390647.ogg"},
	{0xAB1E6FC4DA48DAC3ULL, "170968415.ogg"},
	{0xAB1F268A8A229C25ULL, "6020502_m.fxe"},
	{0xAB1F31BA9421C54EULL, "hlac_lt_0_lowtown_n_196606_0x.dds"},
	{0xAB1F7021BB3BB155ULL, "lt_0_lowtown_docks_a3_n_98_0.msh"},
	{0xAB1F7221BB3BB4CBULL, "lt_0_lowtown_docks_a3_n_98_0.mmh"},
	{0xAB1F7E21BB3BC968ULL, "lt_0_lowtown_docks_a3_n_98_0.mao"},
	{0xAB2020913823E80FULL, "gen00fl_sebastian6180148.cl"},
	{0xAB2137C934BD0407ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.evt"},
	{0xAB2172A3439B9A49ULL, "hlac_lt_0_lowtown_docks_a3_-4_1z.dds"},
	{0xAB21753931012C27ULL, "vxw_sundermount_22r_0_0.vlm"},
	{0xAB218602A763D2FAULL, "334196347.ogg"},
	{0xAB21BC979158EF02ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_win.msh"},
	{0xAB21BE979158F2B8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_win.mmh"},
	{0xAB21CA979159075BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_win.mao"},
	{0xAB220C9F50349205ULL, "vxlt_hangedman_1k2_0_0.vlm"},
	{0xAB221986B5836A68ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd07a.phy"},
	{0xAB222ABF88343962ULL, "mag360_thrask6043535.cl"},
	{0xAB224539B1DF2675ULL, "378583643.ogg"},
	{0xAB225CFAF58F8618ULL, "vxlt_undercityserialkiller_xr_0_0.vlm"},
	{0xAB2288FC1A377D5DULL, "art\\characters\\creatures\\hurlock\\textures\\cn_bdy_hlkb_0d.dds"},
	{0xAB22A4576E7715EBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_loop.gad"},
	{0xAB22D462D763C35FULL, "65589420.ogg"},
	{0xAB22FF369140536EULL, "94548026.ogg"},
	{0xAB231989C87F868FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookgen_n.dds"},
	{0xAB234FBFD1D67A21ULL, "sky_w_bonepitdestroyed_a3.mao"},
	{0xAB236D2C23EA4DE0ULL, "vxdr_primevalentrance_28b_0_0.vlm"},
	{0xAB23F60E45187EA0ULL, "plt_dae200pt_find_mansion.nss"},
	{0xAB2423B6953BC9B2ULL, "308714342.ogg"},
	{0xAB245676799A4E63ULL, "6115626_m.fxe"},
	{0xAB2475199E221A26ULL, "6115490_m.fxe"},
	{0xAB248591B7F68ADEULL, "6116625_f.fxe"},
	{0xAB249E253A7CD4B7ULL, "6027457_m.fxe"},
	{0xAB24AF0C03C44F68ULL, "379656796.ogg"},
	{0xAB24B64A2A66AF86ULL, "vxlt_hangedman_n_1tg_0_0.vlm"},
	{0xAB24E9F7211C6487ULL, "mcr310_argument6074444.cl"},
	{0xAB2569509C5A0270ULL, "814450561.ogg"},
	{0xAB2647073F488EF4ULL, "vxht_mansion2b_49_0_0.vlm"},
	{0xAB26A951FDFFB57CULL, "gen00fl_anders6131328.cl"},
	{0xAB275FE7CEE7DEA6ULL, "vxlt_darktown_6lf_0_0.vlm"},
	{0xAB277802D8997113ULL, "vxlt_0_lowtown_a2_gpe_0_0.vlm"},
	{0xAB2825FF88163B52ULL, "532970646.ogg"},
	{0xAB2858C934C35B50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_2.gad"},
	{0xAB2875EF51013754ULL, "zz_mag200pt_nt_debug.ncs"},
	{0xAB289FE962E41BB0ULL, "hlac_ca_caverns1_24424_12z.dds"},
	{0xAB28EED64EBAB5B9ULL, "6077880_f.fxe"},
	{0xAB2917D37C1DEBFAULL, "vxw_bonepitmining_a2_1j1_0_0.vlm"},
	{0xAB29377EBC10F8C2ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masc.msh"},
	{0xAB29397EBC10FC78ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masc.mmh"},
	{0xAB29457EBC11101BULL, "art\\characters\\playercharacter\\humanfemale\\arm_masc.mao"},
	{0xAB294CAD837247EFULL, "hlac_w_bonepitmining_a2_65538_0x.dds"},
	{0xAB2960B344D1B7CEULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.dura.ani"},
	{0xAB29A1576E7D2F50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_loop.evt"},
	{0xAB29EB988469D097ULL, "vxw_blightlands_ca_0_0.vlm"},
	{0xAB29EBADD9AAF15FULL, "gr_sp_fl01.mao"},
	{0xAB29FDDF24EABC53ULL, "712096235.ogg"},
	{0xAB2A4045A9F57860ULL, "vxht_0_hightown_a2_bz1_0_0.vlm"},
	{0xAB2A4532EDADB67AULL, "vxlt_backalley_g_1yd_0_0.vlm"},
	{0xAB2A55A14C8EEFA6ULL, "6129541_m.fxe"},
	{0xAB2A71F067E57F62ULL, "hla_ht_0_hightown_0_0z.dds"},
	{0xAB2A98508B1E3ED2ULL, "vxlt_darktown_584_0_1.vlm"},
	{0xAB2AE0F7A4EF6771ULL, "544689737.ogg"},
	{0xAB2B7546F8FE9BEBULL, "gen_im_trs_cot_med_glv_r2.uti"},
	{0xAB2BC59791615343ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_win.phy"},
	{0xAB2C6386552BD94EULL, "941635578.ogg"},
	{0xAB2CBA3B10D7179AULL, "narrative_varric6172486.cl"},
	{0xAB2CC274D6E752C0ULL, "6153685_m.fxe"},
	{0xAB2CD3369045C01FULL, "vxca_cavern1_haunted_gf2_0_0.vlm"},
	{0xAB2D0694F2F530FBULL, "6035898_m.fxe"},
	{0xAB2D4708AE56D717ULL, "6127040_f.fxe"},
	{0xAB2E066F599177E3ULL, "6137248_m.fxe"},
	{0xAB2E1DA6EAAF7AC1ULL, "vxca_caverns2prc_7o8_0_0.vlm"},
	{0xAB2E74AAE8D92B92ULL, "vxlt_gallowscourtyard_749_0_0.vlm"},
	{0xAB2ED6C0A7804693ULL, "6036455_m.fxe"},
	{0xAB2F605D752BF430ULL, "708649459.ogg"},
	{0xAB2F70C25B93CDB6ULL, "ave200_donnic6111654.cl"},
	{0xAB2FAC5665D0CBA9ULL, "vxw_woundedcoast_181_0_0.vlm"},
	{0xAB2FAECF22CD6A99ULL, "vxlt_undercityserialkiller_19o_0_0.vlm"},
	{0xAB2FE71622F744A7ULL, "vxw_sundermount_27m_0_0.vlm"},
	{0xAB2FFAC44DDAADEFULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_013a_0n.dds"},
	{0xAB300FA4296F5C35ULL, "6069609_f.fxe"},
	{0xAB30224835A9E43EULL, "seb221_flora_harimann6177703.cl"},
	{0xAB3131D816B83664ULL, "vxht_brothel_n_6cr_0_0.vlm"},
	{0xAB315BDE98C940BEULL, "vxht_0_hightown_a3_c3m_0_1.vlm"},
	{0xAB3163BB96270D75ULL, "vxht_mansion2_251_0_0.vlm"},
	{0xAB321CA157A17FA9ULL, "plt_lgt310pt_fixer_2.nss"},
	{0xAB323C805A0F3884ULL, "6027586_m.fxe"},
	{0xAB324FA4805C5336ULL, "317173974.ogg"},
	{0xAB32C45EFA542F0DULL, "zz_slate2_frontfire.cut"},
	{0xAB32E7123131639AULL, "6108603_m.fxe"},
	{0xAB3328271C5CC806ULL, "vxlt_hangedman_n_1qh_0_0.vlm"},
	{0xAB332A9861EAEFC5ULL, "6134940_m.fxe"},
	{0xAB33407EBC195C03ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masc.phy"},
	{0xAB33440A4F139C5BULL, "vis121_saemus6085696.cl"},
	{0xAB33734E81C29BECULL, "hlac_lt_0_lowtown_a2_65538_5y.dds"},
	{0xAB33DFB46522D5DFULL, "6165287_m.fxe"},
	{0xAB342D0A18F08715ULL, "916028564.ogg"},
	{0xAB34E258DD2AD1CAULL, "hlac_ht_0_hightown_n_65537_2z.dds"},
	{0xAB3506057974BDCEULL, "lt_0_lowtown_docks_a3_17_0.rml"},
	{0xAB352313ED5D50B9ULL, "vxlt_gallowstemplar_n_5t1_0_0.vlm"},
	{0xAB357F033EBB468EULL, "vxht_bartrand_n_5xf_0_0.vlm"},
	{0xAB3587E2E417FBAEULL, "6043394_f.fxe"},
	{0xAB35E9DFCA8F8993ULL, "vxlt_0_lowtown_a2_gwd_0_0.vlm"},
	{0xAB36C3972AF49CA3ULL, "w_blightlands_27_0.rml"},
	{0xAB36C4C76072BFABULL, "6042258_m.fxe"},
	{0xAB37087456EAF9F4ULL, "art\\vfx\\combat\\vfx_glb_ground_dust_p.mmh"},
	{0xAB3758B344DDE71FULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.dura.evt"},
	{0xAB375921E87C4E7CULL, "art\\vfx\\cinematic\\vfx_barrel_explosion_p.anb"},
	{0xAB37E07C26869CEFULL, "6099324_m.fxe"},
	{0xAB3825F42BFC8397ULL, "w_bonepitmining_a2_169_0.phy"},
	{0xAB38591D0E3614C5ULL, "vxlt_backalley_g_1wa_0_0.vlm"},
	{0xAB3865E070BBC3F4ULL, "vxht_0_hightown_a3_be5_0_0.vlm"},
	{0xAB386A99C76B019AULL, "221156645.ogg"},
	{0xAB38A6C8378A3EECULL, "qun211_braggart6182758.cl"},
	{0xAB39700E346CD6A6ULL, "gen00fl_varric6118989.cl"},
	{0xAB3977B93E6516ACULL, "vxca_cavern1_haunted_cju_0_0.vlm"},
	{0xAB39B8E98BDC9625ULL, "hlac_lt_0_lowtown_docks_-262144_2y.dds"},
	{0xAB39BFD479837C8BULL, "614483744.ogg"},
	{0xAB39FF6811574C8BULL, "mcr310_sibling6076384.cl"},
	{0xAB3A451F63D5A0FCULL, "6126689_m.fxe"},
	{0xAB3A51C550E54DBCULL, "vxlt_gamlenshouse_n_7q_0_0.vlm"},
	{0xAB3A5F8CAAB879E6ULL, "188175075.ogg"},
	{0xAB3A639A4A837A74ULL, "6082161_m.fxe"},
	{0xAB3A691537E439FFULL, "art\\characters\\playercharacter\\humanmale\\hlh_masd.phy"},
	{0xAB3AE2015037FE89ULL, "hlac_dr_slavers_3214_14z.dds"},
	{0xAB3B3451C8DD6B40ULL, "6156585_m.fxe"},
	{0xAB3B3A859202157EULL, "vxht_mansion_n_3wf_0_0.vlm"},
	{0xAB3B402EACC1A62BULL, "hlac_ht_keep_n_3732_17x.dds"},
	{0xAB3B5FC28E7AA447ULL, "ser221ip_chest_2.utp"},
	{0xAB3B67A35F7733DCULL, "vxdr_0_deeproads_cwd_0_0.vlm"},
	{0xAB3B69DB2A85939CULL, "art\\characters\\weapons\\staff\\w_stf_009a.mao"},
	{0xAB3B75DB2A85A83FULL, "art\\characters\\weapons\\staff\\w_stf_009a.mmh"},
	{0xAB3B7BDB2A85B2E1ULL, "art\\characters\\weapons\\staff\\w_stf_009a.msh"},
	{0xAB3BE5F465E72318ULL, "6046895_m.fxe"},
	{0xAB3BF13D5DDAED83ULL, "6059697_f.fxe"},
	{0xAB3C549C6645411FULL, "zz_followers_banter6158310.cl"},
	{0xAB3D300B9254B30FULL, "one130_thug6131571.cl"},
	{0xAB3D344853F784A6ULL, "6082017_m.fxe"},
	{0xAB3D4128E3879AE6ULL, "vxlt_foundry_n_vd_0_0.vlm"},
	{0xAB3D7C2C1D125B7AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left.msh"},
	{0xAB3D8A2C1D1272B3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left.mao"},
	{0xAB3D8E2C1D1279C0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left.mmh"},
	{0xAB3D90553B074E30ULL, "6189916_m.fxe"},
	{0xAB3E34A20B427F68ULL, "6170286_m.fxe"},
	{0xAB3E39B344E3D108ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.dura.gad"},
	{0xAB3E58F314ED5D27ULL, "vxw_sundermount_28l_0_0.vlm"},
	{0xAB3E59EAE5C9FAEBULL, "mag270cr_recruit02.utc"},
	{0xAB3EA00FC0DD648AULL, "275198046.ogg"},
	{0xAB3EF3438B2C23A4ULL, "follower_banter6155844.cl"},
	{0xAB3F7E8BED3981F4ULL, "1035503472.ogg"},
	{0xAB3FA3BE991CECFCULL, "367512572.ogg"},
	{0xAB3FA661802C8FD1ULL, "238216340.ogg"},
	{0xAB3FB0153735E896ULL, "dae000_arishok6067788.cl"},
	{0xAB4006BC5B5F533AULL, "dae000_flemeth6000714.cl"},
	{0xAB4099328FA14E66ULL, "vxlt_gallowstemplar_f_5rl_0_0.vlm"},
	{0xAB40B2E4AED8D810ULL, "6092119_m.fxe"},
	{0xAB415D898DD2B6A3ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.dura.evt"},
	{0xAB41B294462FBC65ULL, "260299526.ogg"},
	{0xAB41B2A4817D3FB3ULL, "mer210_seldrek6078804.cl"},
	{0xAB41ECA0407FF51CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic.mao"},
	{0xAB41F8A0408009BFULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic.mmh"},
	{0xAB41FEA040801461ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic.msh"},
	{0xAB4211D08C8604FAULL, "710331560.ogg"},
	{0xAB4248A7F10E7C6BULL, "6106929_m.fxe"},
	{0xAB433E7283767C2DULL, "466136191.ogg"},
	{0xAB437A35FE270BDBULL, "vrc00pt_n_friend_mid.ncs"},
	{0xAB4418333E6002B7ULL, "544324144.ogg"},
	{0xAB44620288FAB711ULL, "l_dae311ar_tavern_nt.wbd"},
	{0xAB446EF5A26B2928ULL, "lgt110_gxa_vgk_conspirator26117089.cl"},
	{0xAB451F5A9A341EA5ULL, "llac_lt_0_lowtown_docks_a3_-131073_2.dds"},
	{0xAB4540873DDD68B5ULL, "6036985_m.fxe"},
	{0xAB45B05DBC410D4FULL, "262381393.ogg"},
	{0xAB4633A04A45AD12ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_9x.dds"},
	{0xAB4657E0A467283CULL, "dae000_hate_enemies6167532.cl"},
	{0xAB46CF5268560197ULL, "vxw_blightlands_ic_0_0.vlm"},
	{0xAB46FB16F9312A48ULL, "vxlt_backalley_g_1l9_0_0.vlm"},
	{0xAB46FF2BF6719CA4ULL, "6030292_m.fxe"},
	{0xAB471498FA776F6EULL, "llac_ht_keep_3732_19.dds"},
	{0xAB471E885B601EA4ULL, "555920248.ogg"},
	{0xAB478A9AB5BAF4C2ULL, "and230_mage_girl6174727.cl"},
	{0xAB47E996305B2F2CULL, "vxca_cavern1_haunted_coj_0_0.vlm"},
	{0xAB4836898DD89324ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.dura.gad"},
	{0xAB48647E2A47AB54ULL, "494745830.ogg"},
	{0xAB48BCC365B6B9FAULL, "646484994.ogg"},
	{0xAB48C932AF734CEAULL, "vxw_woundedcoast_14z_0_0.vlm"},
	{0xAB4915A99F12DEC1ULL, "vxlt_gallowstemplar_f_5lx_0_1.vlm"},
	{0xAB498632A4E54617ULL, "vxlt_0_lowtown_a3_n_c_by3_0_0.vlm"},
	{0xAB49B44E81FF6BCDULL, "160365415.ogg"},
	{0xAB4B16DA2D5C632EULL, "cli220cr_army_ds.ncs"},
	{0xAB4B5392F8E7BF3BULL, "vxlt_0_lowtown_a2_n_c_gky_0_0.vlm"},
	{0xAB4B6A1537F2AE04ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masd.mmh"},
	{0xAB4B6E1537F2B557ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masd.mao"},
	{0xAB4B801537F2D3A6ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masd.msh"},
	{0xAB4BCE2D566CF80CULL, "vxlt_hangedman_n_27c_0_0.vlm"},
	{0xAB4BE0931635209CULL, "6041206_f.fxe"},
	{0xAB4BF51858D0AC1CULL, "102548703.ogg"},
	{0xAB4BF5EFDA5E0DEBULL, "vxlt_gallowstemplar_f_2r6_0_0.vlm"},
	{0xAB4C214DD361161EULL, "vxlt_0_lowtown_docks_n_bf9_0_0.vlm"},
	{0xAB4C928906B99B99ULL, "vxlt_undercityserialkiller_13m_0_0.vlm"},
	{0xAB4D2521C5137C4EULL, "177056907.ogg"},
	{0xAB4E1591FAA46764ULL, "vxht_brothel_n_6ip_0_0.vlm"},
	{0xAB4E2254C05B5535ULL, "vxlt_0_lowtown_a2_n_c_ehx_0_0.vlm"},
	{0xAB4ED606B39C2111ULL, "zz_dre_debug6054664.cl"},
	{0xAB4EEB097BCBD156ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackright.evt"},
	{0xAB4F0B0F819766E6ULL, "vxlt_gallowstemplar_f_5um_0_0.vlm"},
	{0xAB4F12D72EFF404FULL, "105069936.ogg"},
	{0xAB4F2A839DF84B5AULL, "6125505_m.fxe"},
	{0xAB4F45898DDECB02ULL, "art\\vfx\\combat\\vfx_glb_arw_fire_c.dura.ani"},
	{0xAB4F70637C100E52ULL, "275766539.ogg"},
	{0xAB500A2C552FB292ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closed.gad"},
	{0xAB500F6D38A98236ULL, "6102454_m.fxe"},
	{0xAB501848CCBD0D53ULL, "w_woundedcoast_14_0.rml"},
	{0xAB5072EE79761E80ULL, "625287918.ogg"},
	{0xAB50EE1986E103B1ULL, "gen_im_arm_hlm_hvy_champion.uti"},
	{0xAB51A2692846E67CULL, "art\\vfx\\models\\vfx_isa_pose1.mao"},
	{0xAB51AE692846FA1FULL, "art\\vfx\\models\\vfx_isa_pose1.mmh"},
	{0xAB51B469284704C1ULL, "art\\vfx\\models\\vfx_isa_pose1.msh"},
	{0xAB51EA2E7565239EULL, "6117658_m.fxe"},
	{0xAB5200F975FC957EULL, "6155784_m.fxe"},
	{0xAB525E0747600F2DULL, "6078338_m.fxe"},
	{0xAB528DEEAC70C9ACULL, "one100dg_lia6055157.cl"},
	{0xAB52B7CF920B3CFEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_coma_0n.dds"},
	{0xAB52D8713248CB8FULL, "315388322.ogg"},
	{0xAB531024830D85FAULL, "82306470.ogg"},
	{0xAB532633D2B44174ULL, "vxht_mansion2_2lc_0_0.vlm"},
	{0xAB536906D9F8C1A5ULL, "ico_shield_5.dds"},
	{0xAB53699206A8465CULL, "gen00fl_carver6103844.cl"},
	{0xAB53CA8AEBF7FFBBULL, "6070440_m.fxe"},
	{0xAB540AA50EF511FAULL, "vxlt_undercityserialkiller_22j_0_0.vlm"},
	{0xAB541949EEE58084ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_en.ani"},
	{0xAB5472B4E6FAC782ULL, "vxht_keep_62w_0_0.vlm"},
	{0xAB5516B961084C86ULL, "vxht_mansion_n_3qx_0_0.vlm"},
	{0xAB55412F5A4C1A17ULL, "vxw_blightlands_rb_0_0.vlm"},
	{0xAB556CF3EB2742C8ULL, "vxlt_backalley_g_1i6_0_0.vlm"},
	{0xAB55E2097BD1E0E1ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackright.gad"},
	{0xAB561C4176A18178ULL, "and100pt_tranquility.ncs"},
	{0xAB56B5F3CC2893DDULL, "vxca_caverns1_mines_j2e_0_0.vlm"},
	{0xAB570D2C5535D6C9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closed.evt"},
	{0xAB5728D947C1D1FCULL, "6128489_m.fxe"},
	{0xAB5741E87AE5C9C9ULL, "art\\vfx\\combat\\vfx_g_spear_whirlwind_c.mmh"},
	{0xAB574C6DCC124D57ULL, "6087056_f.fxe"},
	{0xAB577F0521BF5F53ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_magb_0s.dds"},
	{0xAB57948CBC847EA4ULL, "zz_vault_debug6203246.cl"},
	{0xAB57A5DD668EDE5FULL, "vxlt_warehouse_n_b3_0_0.vlm"},
	{0xAB584B5D436364DCULL, "vxdr_0_deeproads_ciz_0_0.vlm"},
	{0xAB5857CC0FE25FF7ULL, "6119293_m.fxe"},
	{0xAB587A4D1DC36F19ULL, "vxw_bonepitmining_a2_16b_0_0.vlm"},
	{0xAB58F03259AF147BULL, "341160192.ogg"},
	{0xAB5933C396FA5D5BULL, "fxa_evl_distortion.mao"},
	{0xAB59A460B8F94D50ULL, "proj_ball_blue.mao"},
	{0xAB59A7D1F6733D5AULL, "t3_hlk_b01.tnt"},
	{0xAB5A0F70B477A78FULL, "llac_ht_0_hightown_a3_-131072_0.dds"},
	{0xAB5A15D3D47C8962ULL, "lt_0_lowtown_docks_n_134_0.phy"},
	{0xAB5A1A8CAEB8EBEAULL, "lt_house_3567_lc.mmh"},
	{0xAB5A7B682E80F810ULL, "fen110_anso6095018.cl"},
	{0xAB5A7C35615E6BA1ULL, "876957588.ogg"},
	{0xAB5AC1342FF4DC88ULL, "pro000_aveline6047256.cl"},
	{0xAB5B133C9354581EULL, "509807717.ogg"},
	{0xAB5B1649EEEB9AE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_en.gad"},
	{0xAB5B3CC9887C956AULL, "28371937.ogg"},
	{0xAB5B7A57253B437FULL, "art\\characters\\playercharacter\\humanmale\\hm_arm_bda.mao"},
	{0xAB5B92ECFDD6FFB0ULL, "430315716.ogg"},
	{0xAB5BE3097BD64D8FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackright.ani"},
	{0xAB5CEF2039167EBEULL, "vxdr_primevaltomb_3hq_0_0.vlm"},
	{0xAB5D5DF2CAD7F88BULL, "art\\characters\\playercharacter\\humanfemale\\da_shader.mao"},
	{0xAB5D7CEC738D7F66ULL, "vxlt_gallowstemplar_f_5xn_0_0.vlm"},
	{0xAB5E775B3EB3FB3FULL, "6092372_m.fxe"},
	{0xAB5E7FB8F1124929ULL, "759403166.ogg"},
	{0xAB5ECB37AF1461ADULL, "156561410.ogg"},
	{0xAB5ED0067809839FULL, "hlac_lt_0_lowtown_docks_-65537_2z.dds"},
	{0xAB5F0B74570C50ABULL, "art\\vfx\\combat\\vfx_glb_ground_dust_p.anb"},
	{0xAB5F78984778EAB7ULL, "6167872_m.fxe"},
	{0xAB608BB9A20CA7E4ULL, "6087650_m.fxe"},
	{0xAB60EED34C0F4340ULL, "932492041.ogg"},
	{0xAB60EF0E454C2163ULL, "plt_dae200pt_find_mansion.plo"},
	{0xAB611149EEEFFDC5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_grdstand_en.evt"},
	{0xAB6171893396E4A9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basicnb.phy"},
	{0xAB6230E0C907539CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galowint.mao"},
	{0xAB623C67DDEE183BULL, "6075740_m.fxe"},
	{0xAB623CE0C907683FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galowint.mmh"},
	{0xAB6242E0C90772E1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galowint.msh"},
	{0xAB62EEA0491EA3DAULL, "dae210cr_human_f_01.utc"},
	{0xAB63601299FEAE01ULL, "294306807.ogg"},
	{0xAB63DC8757EF07D4ULL, "lightprobeproj.vsh.fxo"},
	{0xAB63FE381C72AF90ULL, "47342990.ogg"},
	{0xAB6400E51692D49CULL, "105368177.ogg"},
	{0xAB6403955D304647ULL, "vxw_bonepit_27z_0_0.vlm"},
	{0xAB649B8E58815E71ULL, "hlac_ca_caverns1_mines_24424_14x.dds"},
	{0xAB64B1C93B916436ULL, "6095093_m.fxe"},
	{0xAB64DD5210926B1BULL, "542030065.ogg"},
	{0xAB64E6298AA928F0ULL, "vxlt_house_fr_0_0.vlm"},
	{0xAB64F52C5541EB38ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.closed.ani"},
	{0xAB653AFC5F1E04F7ULL, "192855513.ogg"},
	{0xAB65EA205CB31F0DULL, "vxlt_gallowstemplar_n_1ij_0_0.vlm"},
	{0xAB6617BA5884F6DFULL, "vxlt_warehouse_n_i2_0_0.vlm"},
	{0xAB6720FC2A17F417ULL, "rdr111dg_harbormaster6016148.cl"},
	{0xAB67446DB95E708CULL, "vxht_bartrand_n_4di_0_0.vlm"},
	{0xAB6796E196AD9ACCULL, "dae000_friend_turn_in6179316.cl"},
	{0xAB67AB908339D9C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_scared.anb"},
	{0xAB67E1A935495ACCULL, "402609986.ogg"},
	{0xAB67EA539455B7EAULL, "llac_lt_0_lowtown_n_-131071_0.dds"},
	{0xAB67FDAAC8C425F4ULL, "mer110cr_jansen.utc"},
	{0xAB683F6234E182A3ULL, "w_blightlands_105_0.msh"},
	{0xAB684D6234E19AFDULL, "w_blightlands_105_0.mmh"},
	{0xAB68516234E1A10EULL, "w_blightlands_105_0.mao"},
	{0xAB68C4818F0F0968ULL, "pro000_merchant6051640.cl"},
	{0xAB68D1923B3B40B4ULL, "vxlt_gallowscourtyard_4v6_0_0.vlm"},
	{0xAB6948D9386E0DE7ULL, "10900355.ogg"},
	{0xAB69AA185350E5DCULL, "vxdr_0_deeproads_fa3_0_0.vlm"},
	{0xAB69B1972B1FD7BAULL, "w_blightlands_27_0.msh"},
	{0xAB69BF972B1FEFF3ULL, "w_blightlands_27_0.mao"},
	{0xAB69C3972B1FF600ULL, "w_blightlands_27_0.mmh"},
	{0xAB6A7DC2E4AE1603ULL, "gen00fl_varric6093687.cl"},
	{0xAB6AAAF68DC7A8CEULL, "destroy_pressed.cur"},
	{0xAB6B15BF655D23F2ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_lf.gad"},
	{0xAB6B76DA8A3A76E3ULL, "516523249.ogg"},
	{0xAB6B8BBDBAC27152ULL, "art\\levels\\ztest\\chad\\w_dgr_sotf.mao"},
	{0xAB6BB7CB473E8695ULL, "gen00fl_varric6078123.cl"},
	{0xAB6BD2867C208501ULL, "6150542_m.fxe"},
	{0xAB6BDBA6282DAAE1ULL, "671852884.ogg"},
	{0xAB6BE14B624C81C4ULL, "gen00fl_dog6115114.cl"},
	{0xAB6BEEC9658397E6ULL, "vxlt_gallowstemplar_f_5co_0_0.vlm"},
	{0xAB6C42723B1E982DULL, "745565816.ogg"},
	{0xAB6C4AC653B085AFULL, "hlac_lt_0_lowtown_a2_n_-131072_6z.dds"},
	{0xAB6CD59B82BDCDBBULL, "mag170_thrask6020530.cl"},
	{0xAB6CE6A22891FDA4ULL, "mag291_demon_marethari6029044.cl"},
	{0xAB6CE93830AA13BFULL, "6097072_m.fxe"},
	{0xAB6D2D748EDE6CD3ULL, "zz_mrl_debug6094095.cl"},
	{0xAB6D855B7FF0B14EULL, "llac_dr_dungeon_3214_13.dds"},
	{0xAB6DDA98937452D7ULL, "hlac_lt_undercityencounter_-1_16z.dds"},
	{0xAB6E0A3461932EB4ULL, "gen00fl_varric6079945.cl"},
	{0xAB6E2391098ECD9DULL, "hlac_lt_0_lowtown_a2_n_-3_0y.dds"},
	{0xAB6EB1EBA06873ACULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_partycall_sa.dds"},
	{0xAB6FA790FAB039BAULL, "6143480_m.fxe"},
	{0xAB6FA952D691827DULL, "vxht_brothel_n_5kx_0_0.vlm"},
	{0xAB6FD50BA3F41653ULL, "626876457.ogg"},
	{0xAB6FD893A5E55CB5ULL, "436940034.ogg"},
	{0xAB70A7E14FB5F533ULL, "241002059.ogg"},
	{0xAB70C40781767458ULL, "6151395_m.fxe"},
	{0xAB7109C76BB51847ULL, "6140638_f.fxe"},
	{0xAB711817D8F9CDE4ULL, "plt_qun100pt_blackform.plo"},
	{0xAB711E23DA2D4CD7ULL, "6151176_m.fxe"},
	{0xAB7163E91FBFA098ULL, "6182276_m.fxe"},
	{0xAB71BA359B5E7310ULL, "922334397.ogg"},
	{0xAB71E15E9C58CCDEULL, "character_character_p281.psh.fxo.dx11"},
	{0xAB720C2C6B6FC3BCULL, "vxlt_undercityserialkiller_196_0_0.vlm"},
	{0xAB7218BF65634729ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_lf.evt"},
	{0xAB7247F5FCD12A59ULL, "hlac_lt_0_lowtown_131070_0z.dds"},
	{0xAB72B6893FEC0312ULL, "6092550_f.fxe"},
	{0xAB7341A16DD4A668ULL, "follower_banter6160952.cl"},
	{0xAB73B3353AB08616ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax102a_0t.dds"},
	{0xAB746E5AC7F4EBD5ULL, "dae200cr_viscount6087216.cl"},
	{0xAB7486DAB4CA6EDDULL, "894215954.ogg"},
	{0xAB749519B8C480F4ULL, "vxht_mansion_n_752_0_0.vlm"},
	{0xAB74DFEC101E90CAULL, "532855697.ogg"},
	{0xAB753B85F3CE90F7ULL, "6113493_m.fxe"},
	{0xAB75E082BC819A40ULL, "hlac_w_bonepitdestroyed_a3_327683_0y.dds"},
	{0xAB761BB5A0A2E383ULL, "lt_0_lowtown_docks_a3_153_0.mao"},
	{0xAB761FB5A0A2EAD0ULL, "lt_0_lowtown_docks_a3_153_0.mmh"},
	{0xAB762DB5A0A302EAULL, "lt_0_lowtown_docks_a3_153_0.msh"},
	{0xAB7739D4FFB4E061ULL, "vxlt_backalley_g_1x0_0_0.vlm"},
	{0xAB7742C8BA4D7636ULL, "qcr000_aveline6041071.cl"},
	{0xAB774BC41AC106D2ULL, "one300_zevran.cnv"},
	{0xAB77578154B023BBULL, "6084267_m.fxe"},
	{0xAB77980E9DF03BD8ULL, "155328031.ogg"},
	{0xAB77A6EF751268BBULL, "850571775.ogg"},
	{0xAB77E08740CD9F4DULL, "gencr_mercenary_warrior_wsf.utc"},
	{0xAB77F76B18BF00B3ULL, "zz_one_debug6099949.cl"},
	{0xAB78E34222772651ULL, "6088734_f.fxe"},
	{0xAB7A3FBFBF325DD2ULL, "vxht_0_hightown_a3_c2c_0_0.vlm"},
	{0xAB7A6DC41AC3762AULL, "one300_zevran.bnk"},
	{0xAB7AA693CBD856C7ULL, "220525540.ogg"},
	{0xAB7ADE46E3A7012EULL, "lt_0_lowtown_docks_n_124_0.mmh"},
	{0xAB7AEA46E3A715C5ULL, "lt_0_lowtown_docks_n_124_0.mao"},
	{0xAB7AF846E3A72D1CULL, "lt_0_lowtown_docks_n_124_0.msh"},
	{0xAB7B21E6EFDC613AULL, "537972793.ogg"},
	{0xAB7B4C2A7EBDAC83ULL, "vxlt_darktown_5io_0_0.vlm"},
	{0xAB7B4D5027E941FEULL, "meleering9.gda"},
	{0xAB7B523B9E0E2EEAULL, "930397090.ogg"},
	{0xAB7B688E1DF06D55ULL, "vxw_sundermount_35i_0_1.vlm"},
	{0xAB7C3A267DF9F76CULL, "dae000_arishok6068545.cl"},
	{0xAB7CC7A0A70A3F9AULL, "vxlt_undercityserialkiller_1zu_0_0.vlm"},
	{0xAB7CD62309EB167AULL, "vxw_bonepitmining_a2_29w_0_0.vlm"},
	{0xAB7D5235144357F6ULL, "724746970.ogg"},
	{0xAB7E9930058CF20EULL, "6081346_m.fxe"},
	{0xAB7EF0D3D49B6616ULL, "lt_0_lowtown_docks_n_134_0.mao"},
	{0xAB7EF32BEEA4902CULL, "ch1_ds1.psh.fxo"},
	{0xAB7EFAF42C383DEFULL, "w_bonepitmining_a2_169_0.mao"},
	{0xAB7EFCD3D49B7AB5ULL, "lt_0_lowtown_docks_n_134_0.mmh"},
	{0xAB7EFED3D49B7E2BULL, "lt_0_lowtown_docks_n_134_0.msh"},
	{0xAB7EFFD34F9E68C3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedenter.ani"},
	{0xAB7F06F42C38518CULL, "w_bonepitmining_a2_169_0.mmh"},
	{0xAB7F0CF42C385C7EULL, "w_bonepitmining_a2_169_0.msh"},
	{0xAB7F195B55969250ULL, "vxht_0_hightown_c0r_0_0.vlm"},
	{0xAB7F30E6EFD3D7D3ULL, "vxht_0_hightown_a3_c7i_0_0.vlm"},
	{0xAB8000BF656F5B98ULL, "art\\characters\\creatures\\mubari\\animation\\crn_rdy_lf.ani"},
	{0xAB8009A579E4FC71ULL, "gen00fl_anders6125317.cl"},
	{0xAB803EE87B08C9FEULL, "art\\vfx\\combat\\vfx_g_spear_whirlwind_c.anb"},
	{0xAB80773944B5C6E6ULL, "6117518_f.fxe"},
	{0xAB807C85AFD35E51ULL, "168719618.ogg"},
	{0xAB80C28AC1098C48ULL, "vxlt_backalley_g_1x5_0_0.vlm"},
	{0xAB8166CED3DB934BULL, "vfx_vfx_p0.vsh.fxo"},
	{0xAB819FEA3BD3A666ULL, "lit_up_04_l3.mao"},
	{0xAB81C4F4923C9DE0ULL, "sse_ss_karras.gda"},
	{0xAB81E66DF37E87ADULL, "6166608_m.fxe"},
	{0xAB82B4AB14B1BE1FULL, "1027231953.ogg"},
	{0xAB8376145FE627BCULL, "fen210_fenris6108746.cl"},
	{0xAB83ADD711F18107ULL, "mag120_macha6080644.cl"},
	{0xAB83D5DEC056D5EFULL, "280793512.ogg"},
	{0xAB8428279D4AA18CULL, "vxht_bartrand_n_4no_0_0.vlm"},
	{0xAB85264A2BB512EFULL, "6217153_m.fxe"},
	{0xAB858656BA34B101ULL, "vxht_chantry_2e5_0_0.vlm"},
	{0xAB85A4F250CB34BAULL, "dae000_flemeth6001289.cl"},
	{0xAB85A5CED266AB08ULL, "dae000_male_templar_mob6174371.cl"},
	{0xAB85A6E294DA45BEULL, "vxht_brothel_n_4fu_0_0.vlm"},
	{0xAB86F3B22439AA9CULL, "hlac_lt_0_lowtown_docks_196608_1y.dds"},
	{0xAB872C987051AF0AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railhalfsize.mao"},
	{0xAB8730987051B659ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railhalfsize.mmh"},
	{0xAB873A987051C747ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railhalfsize.msh"},
	{0xAB8749200A9BB03DULL, "w_encounter3_87_0.mao"},
	{0xAB874AB67CCC310DULL, "525949793.ogg"},
	{0xAB874D200A9BB776ULL, "w_encounter3_87_0.mmh"},
	{0xAB874FB6F62D5991ULL, "qun160_vashothlead6064679.cl"},
	{0xAB8757200A9BC874ULL, "w_encounter3_87_0.msh"},
	{0xAB875CD2855E6B33ULL, "gen_im_trs_krk_low_crossbow.uti"},
	{0xAB8826D470E8DCEEULL, "gen_im_trs_bst_kidney_02.uti"},
	{0xAB887ED89FC4B3F2ULL, "hm_one_mo_kelder.mor"},
	{0xAB88B19CB1287652ULL, "vxht_0_hightown_a3_c7l_0_0.vlm"},
	{0xAB89351053E96BFDULL, "vxlt_0_lowtown_docks_b6m_0_0.vlm"},
	{0xAB894A8B6502858DULL, "914874145.ogg"},
	{0xAB8980138B1A1B51ULL, "6037047_m.fxe"},
	{0xAB89B41E4AD4E45FULL, "vxht_chantry_13r_0_0.vlm"},
	{0xAB89BE0770B3C503ULL, "vxlt_darktown_5rn_0_0.vlm"},
	{0xAB8A632700EF9BF9ULL, "vxlt_0_lowtown_n_etr_0_0.vlm"},
	{0xAB8D0CF216FDF5FAULL, "hightown_male_noble_gossip6171413.cl"},
	{0xAB8D17D34FAACEBAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedenter.evt"},
	{0xAB8D5A8933BBF8CEULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basicnb.mmh"},
	{0xAB8D668933BC0D65ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basicnb.mao"},
	{0xAB8D748933BC24BCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wo_basicnb.msh"},
	{0xAB8D8B38478CAAD0ULL, "vxht_0_hightown_c7s_0_0.vlm"},
	{0xAB8DF471C718B442ULL, "w_woundedcoast_15_0.rml"},
	{0xAB8E089E8E00B92FULL, "6034624_m.fxe"},
	{0xAB8E2E0213A327D3ULL, "one300ip_st_solivitus.utp"},
	{0xAB8EA45812FB9239ULL, "hlac_ht_keep_3740_0x.dds"},
	{0xAB8EEB0580223145ULL, "938944745.ogg"},
	{0xAB8EEFE64F5BF4BCULL, "vxlt_undercityserialkiller_134_0_0.vlm"},
	{0xAB8F0DD35142B3EBULL, "vxht_0_hightown_a3_a2t_0_0.vlm"},
	{0xAB8F2776AA13176BULL, "gen00fl_isabela6089532.cl"},
	{0xAB8F67BD5F812278ULL, "zz_ave_debug6057575.cl"},
	{0xAB8F6E0C7E43680AULL, "6173054_m.fxe"},
	{0xAB911E4A6BEEEB8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_reaction.ani"},
	{0xAB913B62FFDFD2DAULL, "6067676_m.fxe"},
	{0xAB9168BA8F925257ULL, "6125179_m.fxe"},
	{0xAB91756B224FEE00ULL, "dae000_bartrand6030608.cl"},
	{0xAB91FB6C86142CF4ULL, "and100_anders6095969.cl"},
	{0xAB920FC6D093BC4EULL, "raven.mao"},
	{0xAB9210551964152BULL, "hlac_ht_chantry_2534_8x.dds"},
	{0xAB929A048F40BA0CULL, "vxht_bartrand_n_4sl_0_0.vlm"},
	{0xAB92F54D7BC69F81ULL, "6080632_m.fxe"},
	{0xAB9316D34FAF38ADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\roguedualwieldselectedenter.gad"},
	{0xAB933213FF92FE66ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.impa.ani"},
	{0xAB9393D49547805DULL, "243762399.ogg"},
	{0xAB93A876F9249DD1ULL, "645255042.ogg"},
	{0xAB94CCDF1A54166FULL, "zz_followers_banter6157760.cl"},
	{0xAB94E0030D0DE960ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha03a.phy"},
	{0xAB94F624FE00B678ULL, "hlac_ht_bartrand_n_3381_1y.dds"},
	{0xAB9546C48F5649B6ULL, "86621654.ogg"},
	{0xAB9743788AF608EEULL, "399627873.ogg"},
	{0xAB977D2272382F59ULL, "lt_0_lowtown_a2_n_c_14_0.msh"},
	{0xAB97872272384047ULL, "lt_0_lowtown_a2_n_c_14_0.mmh"},
	{0xAB978B2272384694ULL, "lt_0_lowtown_a2_n_c_14_0.mao"},
	{0xAB97D56D7E0DFFCFULL, "vxlt_gallowsprison_n_6b6_0_0.vlm"},
	{0xAB982B4A6BF520CEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_reaction.gad"},
	{0xAB98D503F2E5B479ULL, "vxlt_0_lowtown_n_eqq_0_0.vlm"},
	{0xAB98E58C8DF54F6BULL, "6129081_m.fxe"},
	{0xAB98E76FA89B354AULL, "6179005_m.fxe"},
	{0xAB98F4C8738BB5E6ULL, "358612689.ogg"},
	{0xAB99424A5A97BA4FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev6.gad"},
	{0xAB9982CD525E9720ULL, "hlac_lt_0_lowtown_docks_a3_n_262143_0z.dds"},
	{0xAB9A22B0B94A3EFFULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg1a.mmh"},
	{0xAB9A28B0B94A48A1ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg1a.msh"},
	{0xAB9A28E7FB1A8543ULL, "215512285.ogg"},
	{0xAB9A2C4BE9F01B22ULL, "vxlt_0_lowtown_a2_n_c_hlb_0_0.vlm"},
	{0xAB9A5D8834D44DA5ULL, "hlac_dr_primevalentrance_1725_2x.dds"},
	{0xAB9AA8680299688DULL, "6166235_m.fxe"},
	{0xAB9B1743C9DB5118ULL, "qun200_party_comment6095386.cl"},
	{0xAB9B2FD3CE24F132ULL, "vxlt_0_lowtown_a2_n_c_feg_0_0.vlm"},
	{0xAB9B8BD8C2808DF5ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.mmh"},
	{0xAB9B9280802F4058ULL, "6076680_f.fxe"},
	{0xAB9B9D7AE2C741ADULL, "vxlt_ruinedhouse_n_6o_0_0.vlm"},
	{0xAB9BDABA29BADD55ULL, "792720457.ogg"},
	{0xAB9C8107244BE883ULL, "389977772.ogg"},
	{0xAB9CA2F0250B071FULL, "vfx_chunkyparticle_p4.psh.fxo"},
	{0xAB9DA891346533D7ULL, "lt_foundry_n_mmi.gff"},
	{0xAB9DAD113775811BULL, "922872072.ogg"},
	{0xAB9DB0671B3CEA57ULL, "vxht_0_hightown_n_bu4_0_0.vlm"},
	{0xAB9E67B4AD2DAF9AULL, "hlac_lt_0_lowtown_a2_n_c_-65535_0y.dds"},
	{0xAB9E67FA011F4743ULL, "llac_lt_blackemporium_1055_18.dds"},
	{0xAB9E8E1EA1A4F4D7ULL, "vxlt_gallowstemplar_n_455_0_0.vlm"},
	{0xAB9EDE71C7270197ULL, "w_woundedcoast_15_0.msh"},
	{0xAB9EE71F006039D1ULL, "hlac_w_encounter3_0_8x.dds"},
	{0xAB9EF071C72720FAULL, "w_woundedcoast_15_0.mao"},
	{0xAB9EF471C7272709ULL, "w_woundedcoast_15_0.mmh"},
	{0xAB9F264A6BFB36ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_reaction.evt"},
	{0xAB9F53F8CEFB2633ULL, "vxca_caverns1_mines_gka_0_0.vlm"},
	{0xAB9FD7A428BA26F2ULL, "6110552_m.fxe"},
	{0xAB9FF762471D7C37ULL, "mrl150_merrill6084473.cl"},
	{0xABA002A15001F1DBULL, "vxlt_0_lowtown_a2_n_c_exv_0_0.vlm"},
	{0xABA05E50247F814CULL, "493019837.ogg"},
	{0xABA063D1998F5AD5ULL, "gen00fl_anders6125205.cl"},
	{0xABA08301F9F55A3EULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.cess.ani"},
	{0xABA11A13FF9F12D7ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.impa.evt"},
	{0xABA1280D38313208ULL, "vxlt_darktown_2rb_0_0.vlm"},
	{0xABA130AF88F8C2C8ULL, "427632785.ogg"},
	{0xABA138A23491D8E2ULL, "character_character_p85.psh.fxo.dx11"},
	{0xABA1439FCA7D4119ULL, "pro000_refugee_female6052097.cl"},
	{0xABA1A37ED878AE75ULL, "179003489.ogg"},
	{0xABA20746DA3FDB6EULL, "134332132.ogg"},
	{0xABA214A6DD1D80ADULL, "lt_0_lowtown_docks_n_25_0_ll.mmh"},
	{0xABA2196BA4952CFAULL, "hlac_ca_caverns2_5802_10x.dds"},
	{0xABA24932E17B5333ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf.evt"},
	{0xABA280964F9BB188ULL, "hlac_ht_0_hightown_n_-196607_0x.dds"},
	{0xABA2E051468FDAB4ULL, "hlac_w_sundermount_196612_0z.dds"},
	{0xABA32AAE742CB783ULL, "gen00fl_anders6119709.cl"},
	{0xABA32CDD81506458ULL, "99845991.ogg"},
	{0xABA3408970E18959ULL, "6043357_f.fxe"},
	{0xABA401A040D2E304ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic.phy"},
	{0xABA6F8F773AAA820ULL, "vfx_distortionmask_p2.vsh.fxo.dx11.assembly"},
	{0xABA71A5D81CA7D7FULL, "lt_gallowscourtyard_57_0.rml"},
	{0xABA81B13FFA53360ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.impa.gad"},
	{0xABA888D8C28B1282ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.anb"},
	{0xABA94232E1816674ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf.gad"},
	{0xABA9E91F4E71CA6DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoresign.mmh"},
	{0xABA9ED1F4E71D0BEULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoresign.mao"},
	{0xABA9FB1F4E71E8F3ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoresign.msh"},
	{0xABAA10D22F599946ULL, "hlac_w_bonepitmining_a2_65537_0x.dds"},
	{0xABAA87A0163C93CBULL, "6161218_m.fxe"},
	{0xABAB0A9FD9926B58ULL, "17572664.ogg"},
	{0xABABDD088ACA6A55ULL, "6016701_m.fxe"},
	{0xABABE49F59DF6FEDULL, "6146814.cut"},
	{0xABAC0DA92A72B758ULL, "one000_ilen6205642.cl"},
	{0xABAC14F8A5DACF19ULL, "hlac_lt_undercityencounter_-1_29z.dds"},
	{0xABAC22440D3302D7ULL, "vxht_0_hightown_n_bp7_0_0.vlm"},
	{0xABAC378897B52F50ULL, "swashbklr_hitbonus.dds"},
	{0xABACBBB424E23893ULL, "atl_prestartme_dxt5_dat.xml"},
	{0xABACDD0893C17B2DULL, "qun110_petrice6206002.cl"},
	{0xABAD492560265E0EULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzraisedstage.msh"},
	{0xABAD4B9714E62F87ULL, "hlac_lt_backalley_g_1_2z.dds"},
	{0xABAD57256026765FULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzraisedstage.mao"},
	{0xABAD5F4A5AA89415ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev6.ani"},
	{0xABAD632560268AFCULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzraisedstage.mmh"},
	{0xABAD9B01FA000D0FULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.cess.evt"},
	{0xABAE00E0683C22E2ULL, "33916.ogg"},
	{0xABAE2D17D92D9F2BULL, "plt_qun100pt_blackform.nss"},
	{0xABAE3FFDA518D323ULL, "1023522329.ogg"},
	{0xABAE93B9941B6616ULL, "lgt310ip_drop_zone_1.ncs"},
	{0xABAEF4FA15E90EA5ULL, "ht_0_hightown_a2_30_0.anb"},
	{0xABAF712412F25461ULL, "vxlt_undercityserialkiller_261_0_0.vlm"},
	{0xABAF99EA2A274A88ULL, "vxlt_darktown_2uc_0_0.vlm"},
	{0xABAFA54CFE5BBD9DULL, "vxlt_0_lowtown_n_f5j_0_0.vlm"},
	{0xABAFB0F56B9950ACULL, "qun221_viscount6069792.cl"},
	{0xABB05132E1879E52ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf.ani"},
	{0xABB15DE248E41FBDULL, "718821530.ogg"},
	{0xABB1CC617402A4F4ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_enter.ani"},
	{0xABB24E2A22CAEB85ULL, "llac_lt_0_lowtown_docks_a3_-1_2.dds"},
	{0xABB265AC69023CC6ULL, "one280_emrys6116797.cl"},
	{0xABB28EC442BEAFDAULL, "llac_lt_hangedman_n_-1_26.dds"},
	{0xABB3385BA3960F06ULL, "plt_lgt000pt_ultra_1_02.plo"},
	{0xABB38790A44F20A9ULL, "store_dae110_tailor.utm"},
	{0xABB3E37A8F061ADEULL, "lt_0_lowtown_26_0_ll.mmh"},
	{0xABB406DD667471A2ULL, "6133865_m.fxe"},
	{0xABB4249090CB72B7ULL, "vxlt_0_lowtown_n_h7d_0_0.vlm"},
	{0xABB44C436780A1A6ULL, "6082206_m.fxe"},
	{0xABB49C01FA062DB8ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.cess.gad"},
	{0xABB4B92761FA30CFULL, "vxlt_gallowsprison_n_6h0_0_0.vlm"},
	{0xABB4CDB90A7E7914ULL, "vxht_0_hightown_9bf_0_0.vlm"},
	{0xABB4F67AF1383CF6ULL, "mag160_varian_ilithis6171459.cl"},
	{0xABB57A9A9E3FC463ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enter.evt"},
	{0xABB58D941235D608ULL, "t1_skn_e05.tnt"},
	{0xABB626B5A0D8DAEBULL, "lt_0_lowtown_docks_a3_153_0.phy"},
	{0xABB64B04975BAF64ULL, "vxw_blightlands_fa_0_0.vlm"},
	{0xABB65F9D3F46C3CFULL, "cart_d.dds"},
	{0xABB68A3E429F66B8ULL, "zz_amb_wall.utp"},
	{0xABB6DFB6B4FB9000ULL, "mag360_thrask6053510.cl"},
	{0xABB6E80D20748EAAULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.dura.gad"},
	{0xABB6F18A4C105B77ULL, "vxlt_0_lowtown_a2_n_c_gwi_0_0.vlm"},
	{0xABB7866EA7C28501ULL, "dae100ip_at_to_aomcave.utp"},
	{0xABB7A04CB94EF11EULL, "vxht_bartrand_n_33e_0_0.vlm"},
	{0xABB7C96174070B56ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_enter.gad"},
	{0xABB84B3568FFB586ULL, "571966138.ogg"},
	{0xABB8B4A164C20653ULL, "294134499.ogg"},
	{0xABB9056F124490BCULL, "primitives_primitive_p0.psh.fxo"},
	{0xABB906C6711172C4ULL, "vxlt_backalley_n_1x1_0_1.vlm"},
	{0xABB97AE653D09876ULL, "56385076.ogg"},
	{0xABB9A06FAAD2D03AULL, "465171987.ogg"},
	{0xABB9D548DD3FB529ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockbxxl.mao"},
	{0xABB9E148DD3FC9C2ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockbxxl.mmh"},
	{0xABB9E348DD3FCD78ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockbxxl.msh"},
	{0xABB9E3DFA3502A8CULL, "follower_banter6172700.cl"},
	{0xABBA5385614633CFULL, "6110805_m.fxe"},
	{0xABBA6BC497FA4FF8ULL, "store_dae370cr_weapons.utm"},
	{0xABBAE21F4E80312AULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whoresign.phy"},
	{0xABBAFA6B3842FFF8ULL, "gen00fl_sebastian6177749.cl"},
	{0xABBAFD030D2EE9F5ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha03a.msh"},
	{0xABBAFF030D2EED6BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha03a.mmh"},
	{0xABBB678F41D94593ULL, "6152670_m.fxe"},
	{0xABBC40CEFC77D3D4ULL, "vxdr_0_deeproads_dqd_0_1.vlm"},
	{0xABBC539A9E45A0E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enter.gad"},
	{0xABBCD2F327EA6317ULL, "278885014.ogg"},
	{0xABBCE65B33EE22DBULL, "vxlt_0_lowtown_a2_n_c_erh_0_0.vlm"},
	{0xABBCEBF2F6FC2612ULL, "vxht_0_hightown_n_4n6_0_0.vlm"},
	{0xABBD43085852D583ULL, "llac_lt_0_lowtown_a2_n_c_65533_1.dds"},
	{0xABBD5E130828DB08ULL, "ave200_donnic6110624.cl"},
	{0xABBDCB0D207A7C41ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.dura.evt"},
	{0xABBDCB28583AD90CULL, "6047688_f.fxe"},
	{0xABBEA7A23A1292C3ULL, "vxlt_darktown_2v1_0_0.vlm"},
	{0xABBECE266C2E4EB3ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd1a.phy"},
	{0xABBEE461740D57B5ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_enter.evt"},
	{0xABBF0A4BC81FC9A9ULL, "570295863.ogg"},
	{0xABBF8CB99D84FB91ULL, "426753004.ogg"},
	{0xABBF92F99BA3F601ULL, "6071503_f.fxe"},
	{0xABC059B7969439A7ULL, "vxw_sundermount_2eg_0_0.vlm"},
	{0xABC12C855015B44EULL, "hlac_lt_darktown_-1_77x.dds"},
	{0xABC17861026CE656ULL, "440985482.ogg"},
	{0xABC22EC373756946ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_rblt.anb"},
	{0xABC26F3BE75E0DD5ULL, "vxlt_shop_6o_0_0.vlm"},
	{0xABC2966D82C18B37ULL, "vxlt_0_lowtown_n_h2g_0_0.vlm"},
	{0xABC29B9C358EE90DULL, "hlac_dr_0_deeproads_5845_14z.dds"},
	{0xABC2BE205976BA26ULL, "6085306_m.fxe"},
	{0xABC3629A9E4BD8C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\enter.ani"},
	{0xABC368701EFC8527ULL, "hlac_lt_0_lowtown_docks_a3_n_65532_0x.dds"},
	{0xABC390A830514145ULL, "27107130.ogg"},
	{0xABC3A158F22753B8ULL, "ser221_gamlen6133133.cl"},
	{0xABC3AD2A0ECCC825ULL, "vxdr_0_deeproads_fhy_0_0.vlm"},
	{0xABC445E0C95A4184ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1pp_galowint.phy"},
	{0xABC47022725F077CULL, "lt_0_lowtown_a2_n_c_14_0.phy"},
	{0xABC4DA111CEA59E3ULL, "ser121_jethann6105767.cl"},
	{0xABC563673E0673F7ULL, "vxlt_0_lowtown_a2_n_c_gph_0_0.vlm"},
	{0xABC5A53EA688134AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eagleheadsml_n.dds"},
	{0xABC5B7500D5D6B9EULL, "gen00fl_aveline6053270.cl"},
	{0xABC61AED2645A062ULL, "seb221_flora_harimann6177277.cl"},
	{0xABC64B0B925F5EBEULL, "hlac_lt_0_lowtown_a2_n_c_-131072_7z.dds"},
	{0xABC6A2F145DD2294ULL, "vxht_chantry_n_27e_0_0.vlm"},
	{0xABC6C94116783906ULL, "gencr_slaver_ws_f.utc"},
	{0xABC6CBBCE35DDCAEULL, "hlac_ht_0_hightown_a3_65537_2z.dds"},
	{0xABC720C72A768AAAULL, "vxdr_primevaltomb_75r_0_0.vlm"},
	{0xABC783CBF7B9D32DULL, "vxca_cavern1_haunted_e6z_0_0.vlm"},
	{0xABC78DEA65541512ULL, "hlac_lt_0_lowtown_docks_a3_-131075_2y.dds"},
	{0xABC7F171C74A273EULL, "w_woundedcoast_15_0.anb"},
	{0xABC81A0E174A2CD6ULL, "pro000cs_aveline_intro_end.ncs"},
	{0xABC8403254D3ACE8ULL, "vxca_caverns1_mines_e56_0_0.vlm"},
	{0xABC8757F4FDE33A7ULL, "hlac_dr_primevaltomb_664_1z.dds"},
	{0xABC8EB4B92263792ULL, "hlac_lt_undercityencounter_-1_5y.dds"},
	{0xABC8FD69020E3E01ULL, "454280215.ogg"},
	{0xABC905FDF11F33D7ULL, "vxht_0_hightown_n_bf1_0_0.vlm"},
	{0xABC92C1EA7D7D35FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_fingers_together.gad"},
	{0xABC9A6351B9FFD92ULL, "6068017_m.fxe"},
	{0xABC9C63D1CF4DC3BULL, "hlac_ht_mansion2_818_14x.dds"},
	{0xABCA1CB211DB41EEULL, "t3_wer_b00.tnt"},
	{0xABCA3312D16C96D7ULL, "389705858.ogg"},
	{0xABCA8722726399A4ULL, "lt_0_lowtown_a2_n_c_14_0.rml"},
	{0xABCAC30D2084F890ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.dura.ani"},
	{0xABCB3818DA7B4315ULL, "vxlt_gallowstemplar_f_3ni_0_0.vlm"},
	{0xABCB583825E43B5BULL, "vxlt_0_lowtown_a2_n_c_ewk_0_0.vlm"},
	{0xABCB92A683D2A2D0ULL, "6018904_m.fxe"},
	{0xABCC3D054A30F18CULL, "6042988_f.fxe"},
	{0xABCC824753C7F5E0ULL, "w_sundermount_77_0_ll.mmh"},
	{0xABCD0AD370127430ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_cupwood_01_0_sa.dds"},
	{0xABCD197F2C08AB43ULL, "vxlt_darktown_2y0_0_0.vlm"},
	{0xABCD1B714775FB0EULL, "vxht_mansion2b_1q9_0_0.vlm"},
	{0xABCD1C23348E37DDULL, "gen_im_trs_dal_dua_dag_r0.uti"},
	{0xABCDBB6EE4B8C253ULL, "6125838_m.fxe"},
	{0xABCDC7B3BE755FCDULL, "hlac_lt_gallowstemplar_2392_0y.dds"},
	{0xABCDFCA3CEFDFFCFULL, "vxht_mansion_n_4w0_0_0.vlm"},
	{0xABCE2003A937622DULL, "6144257_m.fxe"},
	{0xABCE2D91D1FBB216ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_visa.mao"},
	{0xABCE426A2EA9762AULL, "dae000ip_crowd_mf3_talk9.utp"},
	{0xABCECB94888A5227ULL, "vxw_sundermount_2nf_0_0.vlm"},
	{0xABCEF38EFCCE9F99ULL, "6085895_m.fxe"},
	{0xABCFDA2CD4414D5EULL, "6147527_m.fxe"},
	{0xABCFFE34DD1EF8D6ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravea.msh"},
	{0xABD00834DD1F09D4ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravea.mmh"},
	{0xABD00C34DD1F10E7ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravea.mao"},
	{0xABD0211EA7DDDF4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_fingers_together.evt"},
	{0xABD0265526C875EBULL, "llac_lt_0_lowtown_a2_n_65535_1.dds"},
	{0xABD0352C7917C285ULL, "mag160_feynriel6167006.cl"},
	{0xABD03EB4EDB452F3ULL, "hlac_lt_0_lowtown_docks_a3_-4_1x.dds"},
	{0xABD06B78BE41A143ULL, "113401834.ogg"},
	{0xABD0D04328066DDEULL, "gen00fl_aveline6114809.cl"},
	{0xABD115B885B15201ULL, "6042807_m.fxe"},
	{0xABD1A22498957AB3ULL, "104171180.ogg"},
	{0xABD20CA581EFD586ULL, "vxlt_hangedman_n_18s_0_0.vlm"},
	{0xABD27F4B779DDC3DULL, "6042839_f.fxe"},
	{0xABD29A4D5B32D550ULL, "hlac_lt_0_lowtown_a2_n_c_-196607_0z.dds"},
	{0xABD2CF5DF7D261E1ULL, "6149922_m.fxe"},
	{0xABD2DFF127EE38EAULL, "gen00fl_aveline6115396.cl"},
	{0xABD32EBE7B47E064ULL, "vxw_blightlands_lg_0_0.vlm"},
	{0xABD34225604721C7ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzraisedstage.phy"},
	{0xABD35C3DEB90CC7FULL, "vxht_keep_67y_0_0.vlm"},
	{0xABD365D39B512633ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangsign_n.dds"},
	{0xABD36964A8598328ULL, "art\\levels\\ztools\\default\\generic\\9ff_signarrow.msh"},
	{0xABD37B64A859A1B9ULL, "art\\levels\\ztools\\default\\generic\\9ff_signarrow.mao"},
	{0xABD38764A859B652ULL, "art\\levels\\ztools\\default\\generic\\9ff_signarrow.mmh"},
	{0xABD3A862573C5F7FULL, "vxw_encounter2_camp_q1_0_1.vlm"},
	{0xABD3F3BFB1C87D22ULL, "vxlt_0_lowtown_a2_n_c_hxn_0_0.vlm"},
	{0xABD3F84895D928F9ULL, "screencolorclear.psh.fxo"},
	{0xABD49EB9943C4823ULL, "lgt310ip_drop_zone_1.utp"},
	{0xABD521CC2EC0ABF4ULL, "sse_ss_coterie_thug_m.gda"},
	{0xABD5430213DF4D86ULL, "one300ip_st_solivitus.ncs"},
	{0xABD5587E9BA536F5ULL, "vxht_0_hightown_a3_anf_0_0.vlm"},
	{0xABD56B3CC49C798DULL, "vxdr_0_deeproads_9ck_0_0.vlm"},
	{0xABD5E8B757B04826ULL, "zz_lgt_1_debug6148337.cl"},
	{0xABD624DF1AC0638FULL, "vxw_bonepitmining_a2_bo_0_0.vlm"},
	{0xABD63FF4A6C2A35EULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_meda_0n.dds"},
	{0xABD65D7E42C49D3CULL, "vxdr_0_deeproads_ewf_0_0.vlm"},
	{0xABD700EB136CAA08ULL, "6069607_m.fxe"},
	{0xABD76449B81E8508ULL, "6166986_m.fxe"},
	{0xABD7BDD5BCE270B5ULL, "art\\levels\\races\\proxy\\props\\_textures\\qa_warning_d.dds"},
	{0xABD7C2BEC1944955ULL, "176963483.ogg"},
	{0xABD7DF86C9495286ULL, "plt_one300pt_murder_crows.nss"},
	{0xABD8EF9B0E8931CCULL, "718491507.ogg"},
	{0xABD97E6EBB07D4FDULL, "230042192.ogg"},
	{0xABD9C7E1473A4990ULL, "851082480.ogg"},
	{0xABD9EA187315ECF9ULL, "vxw_bonepitmining_a2_p3_0_0.vlm"},
	{0xABDA545BD92F2A59ULL, "ts_slavers_03.gda"},
	{0xABDB22EFCC7D095CULL, "zz_harvestertest.cut"},
	{0xABDB50BADA5B70E7ULL, "277666809.ogg"},
	{0xABDBBDAC544751EFULL, "132988752.ogg"},
	{0xABDC87890F253221ULL, "6103740_m.fxe"},
	{0xABDCCCFD8FD26DA4ULL, "dae000_flemeth6001889.cl"},
	{0xABDCEC509B6C584DULL, "hlac_lt_0_lowtown_docks_a3_0_5y.dds"},
	{0xABDD091EA7E840A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_fingers_together.ani"},
	{0xABDD3D717A806AA7ULL, "vxw_sundermount_2sa_0_0.vlm"},
	{0xABDD9187825DF1D4ULL, "vxca_caverns1_mines_e4y_0_0.vlm"},
	{0xABDD99F3F0956D96ULL, "follower_banter6166714.cl"},
	{0xABDE15E32E69747BULL, "249906052.ogg"},
	{0xABDE1A7D66133777ULL, "ser100_party_comments6120599.cl"},
	{0xABDE5B5314934253ULL, "gen00fl_fenris6107199.cl"},
	{0xABDE65BE411A4028ULL, "hlac_lt_gamlenshouse_n_-1_4z.dds"},
	{0xABDED101257B1224ULL, "370017532.ogg"},
	{0xABDF4CBD7EC06CD2ULL, "hlac_w_bonepit_262150_0z.dds"},
	{0xABDF7427E0005C67ULL, "hlac_lt_0_lowtown_n_-1_1x.dds"},
	{0xABDFB7A8A33D4770ULL, "gen00fl_merrill6129907.cl"},
	{0xABDFDC28C17EB15DULL, "vxlt_0_lowtown_a2_n_eq9_0_0.vlm"},
	{0xABE06E8D53FE7A93ULL, "6125386_m.fxe"},
	{0xABE07E8273E5EE06ULL, "vxlt_hangedman_n_15t_0_0.vlm"},
	{0xABE08BA211C689CFULL, "scale.ncs"},
	{0xABE0C18620A616C8ULL, "733812891.ogg"},
	{0xABE1712396092A27ULL, "llac_lt_gallowsprison_n_-1_13.dds"},
	{0xABE17B3DB570DCB1ULL, "art\\vfx\\creatures\\fxc_arc_shd_c.anb"},
	{0xABE1A09B6D3DF8E4ULL, "vxw_blightlands_qd_0_0.vlm"},
	{0xABE1CE1ADD86E4FFULL, "vxht_keep_60x_0_0.vlm"},
	{0xABE20C7F658A77DAULL, "6139842_f.fxe"},
	{0xABE2A97B9F03EC5FULL, "ht_0_hightown_n_11_0.rml"},
	{0xABE3288C1409B920ULL, "vxht_0_hightown_a3_cbq_0_0.vlm"},
	{0xABE33B5533DFD9A4ULL, "vxht_brothel_n_34n_0_0.vlm"},
	{0xABE33B83A897621AULL, "6085899_f.fxe"},
	{0xABE39512F5AF55C1ULL, "w_encounter3_25_0.rml"},
	{0xABE3ADA64FB52987ULL, "llac_lt_undercityencounter_-1_10.dds"},
	{0xABE3ADFDBC2372BFULL, "575290939.ogg"},
	{0xABE3DD19B692920DULL, "vxdr_0_deeproads_9dj_0_0.vlm"},
	{0xABE4106961D0F731ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar5_.ani"},
	{0xABE432D947BE45ABULL, "vxht_mansion_n_6hw_0_0.vlm"},
	{0xABE4AE4C62F6032DULL, "gen00fl_dog6146953.cl"},
	{0xABE4CF5B34BAB5BCULL, "vxdr_0_deeproads_epg_0_0.vlm"},
	{0xABE4D33A76B5E061ULL, "42357211.ogg"},
	{0xABE4E7243467109DULL, "vxlt_warehouse_n_100_0_0.vlm"},
	{0xABE55D52EE95C08EULL, "hlac_dr_primevalentrance_1726_0x.dds"},
	{0xABE57780693293D1ULL, "645587075.ogg"},
	{0xABE5E9B7D50B64D7ULL, "vxht_0_hightown_n_bl3_0_0.vlm"},
	{0xABE5F21104B3C6A7ULL, "6102150_m.fxe"},
	{0xABE6514F2488779FULL, "llac_ca_caverns1_24421_3.dds"},
	{0xABE7AFF91B4F4EB7ULL, "vxlt_gallowstemplar_n_433_0_0.vlm"},
	{0xABE7B30BD04DA88AULL, "vxht_0_hightown_a3_8ay_0_0.vlm"},
	{0xABE7ECB14A7ADB30ULL, "6075211_m.fxe"},
	{0xABE86000F6A3098AULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_ars_0s.dds"},
	{0xABE8F5FA161A88ADULL, "ht_0_hightown_a2_30_0.rml"},
	{0xABE96A2BFC78D54AULL, "6054876_m.fxe"},
	{0xABE97897ECB0D75AULL, "vxlt_0_lowtown_docks_n_b4k_0_0.vlm"},
	{0xABEA319870A63546ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railhalfsize.phy"},
	{0xABEA4F946B884232ULL, "vxlt_backalley_g_1k6_0_0.vlm"},
	{0xABEA825DCE41AF6AULL, "vxca_caverns1_mines_in5_0_0.vlm"},
	{0xABEA904E6A6B78CDULL, "vxca_caverns2_7u7_0_0.vlm"},
	{0xABEB1B6961D728D3ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerrules_extratemplar5_.gad"},
	{0xABEB7206A6AD422EULL, "vxht_mansion_n_54a_0_0.vlm"},
	{0xABEB7A418FDFAE5EULL, "773224937.ogg"},
	{0xABEBF486C95A1FC1ULL, "plt_one300pt_murder_crows.plo"},
	{0xABEBF8BFB2158A32ULL, "vxca_caverns1_e4r_0_0.vlm"},
	{0xABEC1381EA183CAFULL, "vxlt_gallowscourtyard_n_7do_0_0.vlm"},
	{0xABECEE6F6209F61CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.open2.ani"},
	{0xABED16C53F07DE59ULL, "llac_lt_backalley_g_65536_0.dds"},
	{0xABED1D3712B24736ULL, "gen00fl_isabela6118358.cl"},
	{0xABED96E40A24AB01ULL, "203435431.ogg"},
	{0xABEDF5D63EEB3A05ULL, "mag320_noblewoman_26060281.cl"},
	{0xABEE2CB3937E030BULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.cess.evt"},
	{0xABEE36056EEFBD56ULL, "194782667.ogg"},
	{0xABEE8E7BBFCC5039ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg106a_0s.dds"},
	{0xABEE9C3DB57B9F2AULL, "art\\vfx\\creatures\\fxc_arc_shd_c.mmh"},
	{0xABEEA9C133AEE53BULL, "w_blightlands_208_0.mmh"},
	{0xABEEB5C133AEF9D8ULL, "w_blightlands_208_0.mao"},
	{0xABEEC7C133AF1825ULL, "w_blightlands_208_0.msh"},
	{0xABEEE10EA896EA73ULL, "vxlt_gallowstemplar_5c1_0_0.vlm"},
	{0xABEF2466F8C17CB6ULL, "6177371_m.fxe"},
	{0xABEF8E3606D7A89AULL, "lt_gallowscourtyard_water_96.msh"},
	{0xABEF9C3606D7C0D3ULL, "lt_gallowscourtyard_water_96.mao"},
	{0xABEFA03606D7C7E0ULL, "lt_gallowscourtyard_water_96.mmh"},
	{0xABEFACE1E9859529ULL, "6034529_m.fxe"},
	{0xABEFAFE82A45A169ULL, "w_blightlands_195_0.mao"},
	{0xABEFB7987DB78DC7ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_008a_0d.dds"},
	{0xABEFBBE82A45B502ULL, "w_blightlands_195_0.mmh"},
	{0xABEFBDE82A45B8B8ULL, "w_blightlands_195_0.msh"},
	{0xABEFF9D8CD918E53ULL, "6044651_f.fxe"},
	{0xABF012785F341164ULL, "vxw_blightlands_ze_0_0.vlm"},
	{0xABF0254418165B15ULL, "llac_ht_mansion2b_818_10.dds"},
	{0xABF078CF8F7D7160ULL, "hlac_lt_0_lowtown_a2_n_c_-65538_0x.dds"},
	{0xABF153C056A86F4CULL, "vxht_keep_4jk_0_0.vlm"},
	{0xABF1A11A3FA03F8BULL, "dae000_hate_follower_react6103225.cl"},
	{0xABF1A81FC9C43B44ULL, "gen00fl_aveline6110366.cl"},
	{0xABF1AD3225D5F224ULL, "vxht_brothel_n_33o_0_0.vlm"},
	{0xABF21940E655F24FULL, "213109632.ogg"},
	{0xABF2815255AF1E04ULL, "art\\levels\\races\\design\\_textures\\prp_bigtreasure_sa.dds"},
	{0xABF2D7B7F9E86611ULL, "408368235.ogg"},
	{0xABF2DA107C42C7D8ULL, "vxlt_0_lowtown_a2_n_c_eoe_0_1.vlm"},
	{0xABF2FE24378DEE04ULL, "274807705.ogg"},
	{0xABF34A55FBDF3B26ULL, "dae021_sister_lorena_sta6150096.cl"},
	{0xABF3CB64EA1BA633ULL, "6041704_f.fxe"},
	{0xABF3F25FBDF0F8F5ULL, "vxht_0_hightown_cel_0_0.vlm"},
	{0xABF3FB6F62102B5EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.open2.gad"},
	{0xABF4219300E51EF3ULL, "740175267.ogg"},
	{0xABF4C349506B7F08ULL, "lowtown_male_beggar_sta6156765.cl"},
	{0xABF4EB35B515DFEAULL, "146170715.ogg"},
	{0xABF4F7A2D1021FFEULL, "mag120st_geoff_kerod.stg"},
	{0xABF50FDDF8066349ULL, "follower_banter6156176.cl"},
	{0xABF545B393844CCCULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.cess.gad"},
	{0xABF5ACDC09B98277ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanr.gad"},
	{0xABF67F9A2B1F5278ULL, "fen110_fenris6100305.cl"},
	{0xABF6A61B5BAB8293ULL, "plt_cod_hst_kirk_gallows.plo"},
	{0xABF75CD1B7BBF434ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_1x.dds"},
	{0xABF7CD35D44608F0ULL, "llac_lt_undercityserialkiller_-1_51.dds"},
	{0xABF82A00158F9320ULL, "3698066.ogg"},
	{0xABF832F1324C2567ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.dura.gad"},
	{0xABF85EFA417D2A99ULL, "1013638175.ogg"},
	{0xABF8684E353057BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_exit.ani"},
	{0xABF8719CDAEB3D38ULL, "character_rendertodepth_p2.psh.fxo.dx11.assembly"},
	{0xABF8F43AC037C7EAULL, "vxca_caverns1_mines_ik6_0_0.vlm"},
	{0xABF90A48ABA8AFF7ULL, "llac_lt_0_lowtown_docks_a3_65535_0.dds"},
	{0xABFA1A163EB53F38ULL, "6101079_m.fxe"},
	{0xABFA212B5E6C9BA7ULL, "vxw_sundermount_2yc_0_0.vlm"},
	{0xABFA8C1D361B47B9ULL, "vxht_0_hightown_a3_c3w_0_0.vlm"},
	{0xABFAA201DD793926ULL, "21097964.ogg"},
	{0xABFAC3627A042375ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcheese_0.pwk"},
	{0xABFB39D0B2DF187CULL, "282757315.ogg"},
	{0xABFBA3C3FD1FE7F8ULL, "art\\characters\\playercharacter\\children\\kf_cth_grla.mao"},
	{0xABFBC3532D1FC739ULL, "517382331.ogg"},
	{0xABFBCD31F3835EBFULL, "vxlt_0_lowtown_a3_n_c_gzc_0_0.vlm"},
	{0xABFBE32DCDA0A79EULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_boilerb_1.mao"},
	{0xABFBEC2D7494F656ULL, "452633177.ogg"},
	{0xABFC098DBEAC75AFULL, "vxca_caverns1_mines_j4y_0_0.vlm"},
	{0xABFC2AA554FA22D0ULL, "vxdr_primevaltomb_1cl_0_0.vlm"},
	{0xABFC44B3938A69CAULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.cess.ani"},
	{0xABFC616AD9208CE6ULL, "vxlt_gallowstemplar_f_51y_0_0.vlm"},
	{0xABFCA9B94A8F46EBULL, "dae300_alistair_king6133535.cl"},
	{0xABFCC1DC09BFC404ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanr.evt"},
	{0xABFD2BB0B99ECB44ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hg1a.phy"},
	{0xABFD5BE53C23A8CEULL, "art\\vfx\\combat\\vfx_r_gn3_embu_act_c.impa.ani"},
	{0xABFDD2E2850C879FULL, "855991776.ogg"},
	{0xABFDD540CABCAE91ULL, "774514470.ogg"},
	{0xABFDF47DFA0FAA34ULL, "6118998_m.fxe"},
	{0xABFE195BA3D5C645ULL, "plt_lgt000pt_ultra_1_02.nss"},
	{0xABFE27F132507E74ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.dura.evt"},
	{0xABFE36382FC44D48ULL, "static_static_p7.psh.fxo.dx11"},
	{0xABFE5556DE0FDF94ULL, "vxca_cavern1_haunted_fo7_0_0.vlm"},
	{0xABFF5EB693FB1983ULL, "zz_followers_banter6159285.cl"},
	{0xABFF9E58F0095CBCULL, "mcr370_assignments6081749.cl"},
	{0xABFFB38AEA95B213ULL, "329390959.ogg"},
	{0xABFFBCA6CDD04F4CULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgtp_0t.dds"},
	{0xABFFF87C8EF9C613ULL, "mrl000_marethari6142155.cl"},
	{0xAC009DDEF2DC346AULL, "441492219.ogg"},
	{0xAC00E046D9CE9676ULL, "vxht_bartrand_n_64y_0_0.vlm"},
	{0xAC013AC7BC031545ULL, "162163985.ogg"},
	{0xAC01505C348EE608ULL, "follower_banter6136641.cl"},
	{0xAC018365FFDAA749ULL, "222901577.ogg"},
	{0xAC018F9FF1A10E23ULL, "darktown_fem_elf_sta6174526.cl"},
	{0xAC0196BBB5D21FD4ULL, "458207530.ogg"},
	{0xAC01BD39C5070E9CULL, "310800358.ogg"},
	{0xAC01D7ACEE9B282AULL, "vxlt_0_lowtown_n_gc2_0_0.vlm"},
	{0xAC02612007CC3BDAULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf03.mmh"},
	{0xAC026D4619DAAF71ULL, "follower_banter6170705.cl"},
	{0xAC02732007CC5A20ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf03.msh"},
	{0xAC03094E049F41B2ULL, "6020484_m.fxe"},
	{0xAC0352E32658B60DULL, "lt_0_lowtown_docks_n_147_0.msh"},
	{0xAC0360E32658CE40ULL, "lt_0_lowtown_docks_n_147_0.mao"},
	{0xAC0364E32658D493ULL, "lt_0_lowtown_docks_n_147_0.mmh"},
	{0xAC0396C5C454F8D6ULL, "6029058_m.fxe"},
	{0xAC03BDE2A0BF9FE2ULL, "dae100_alistair_drunk6135144.cl"},
	{0xAC03F61EE32FA68FULL, "83194156.ogg"},
	{0xAC03FA87F5A3FB5CULL, "6037327_m.fxe"},
	{0xAC040684E148118BULL, "vxlt_0_lowtown_a2_gxe_0_0.vlm"},
	{0xAC04682ED5C6746FULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_0.pwk"},
	{0xAC04CFB41939ABCBULL, "6083042_m.fxe"},
	{0xAC04D0F984BD5F7FULL, "vxlt_gallowstemplar_n_4um_0_0.vlm"},
	{0xAC04EB11DFDD4CF3ULL, "6177749_m.fxe"},
	{0xAC053FAF48F83679ULL, "vxw_bonepitmining_a2_a0_0_0.vlm"},
	{0xAC056026433429AEULL, "6125291_m.fxe"},
	{0xAC05704E353AEFB2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_exit.evt"},
	{0xAC057C4F9B99D063ULL, "6042474_m.fxe"},
	{0xAC05C8D809FDDEC7ULL, "zz_approval_debug6003112.cl"},
	{0xAC05E283E23B7A92ULL, "vxlt_0_lowtown_a2_byz_0_0.vlm"},
	{0xAC067B8B33F0CECCULL, "6054311_m.fxe"},
	{0xAC06A9EB0283CDD3ULL, "466147214.ogg"},
	{0xAC06B47505AAEF36ULL, "l_qun100ar_plot_start.wbd"},
	{0xAC06F92D41A9225DULL, "vxca_caverns1_mines_i71_0_0.vlm"},
	{0xAC07471E0E2164B7ULL, "llac_lt_backalley_g_0_18.dds"},
	{0xAC0753201EF7675CULL, "6069755_m.fxe"},
	{0xAC0774084E57A9CDULL, "vxca_caverns2_7o5_0_0.vlm"},
	{0xAC08B0ECD73F7B66ULL, "vxlt_gallowstemplar_n_5pn_0_0.vlm"},
	{0xAC08D6D5592BF5BCULL, "6071181_m.fxe"},
	{0xAC08EFDD91EDE603ULL, "lt_0_lowtown_docks_44_0.rml"},
	{0xAC08FDFA28116039ULL, "vxht_0_hightown_a3_c4v_0_0.vlm"},
	{0xAC09B7889BC4E841ULL, "hlac_ca_caverns2_5802_37x.dds"},
	{0xAC09D9DC09CA77FDULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanr.ani"},
	{0xAC09E9CF6C6C6FB9ULL, "support_versions.gfx"},
	{0xAC0A15A7BC2DDC67ULL, "vxht_mansion_n_4x1_0_0.vlm"},
	{0xAC0A1E36913FAF9EULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.dura.evt"},
	{0xAC0A21E667F03A4CULL, "6100053_f.fxe"},
	{0xAC0A3C0517CC1048ULL, "mag130_danzig6167371.cl"},
	{0xAC0A546693F562D4ULL, "vxlt_hangedman_25v_0_0.vlm"},
	{0xAC0A7313402D005DULL, "6178452_m.fxe"},
	{0xAC0A7B6AB0A28E2FULL, "vxca_caverns1_mines_j3x_0_0.vlm"},
	{0xAC0B53E53C2FD81FULL, "art\\vfx\\combat\\vfx_r_gn3_embu_act_c.impa.evt"},
	{0xAC0BC3F331A775E2ULL, "mag360st_thrask.stg"},
	{0xAC0BD8FC7367A5EBULL, "6076877_m.fxe"},
	{0xAC0BFE5332FE746BULL, "vxht_mansion2_2eo_0_0.vlm"},
	{0xAC0C1FF1325CADADULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.dura.ani"},
	{0xAC0C2AA916B487DCULL, "llac_ht_0_hightown_65534_0.dds"},
	{0xAC0C3FFB1A548554ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsharda.phy"},
	{0xAC0C8F4E35414345ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_exit.gad"},
	{0xAC0C962840F1BBDBULL, "w_woundedcoast_115_0.phy"},
	{0xAC0D2C42DEE6EA44ULL, "773450677.ogg"},
	{0xAC0DB5E533163E67ULL, "vxlt_gallowsprison_n_6m3_0_0.vlm"},
	{0xAC0DB8E3DF93A2C8ULL, "mrl000_marethari6154500.cl"},
	{0xAC0E041592B1B8FAULL, "art\\vfx\\materials\\gui\\vfx_gui_disorient.mao"},
	{0xAC0E55E0BCDAFA0AULL, "303963107.ogg"},
	{0xAC0EC80E5A12A335ULL, "w_bonepitmining_a2_105_0.phy"},
	{0xAC0FEA4680A45E64ULL, "133737321.ogg"},
	{0xAC103536914442C9ULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.dura.gad"},
	{0xAC10870A1CE60E5DULL, "752991969.ogg"},
	{0xAC10A29A734CB5A2ULL, "868051789.ogg"},
	{0xAC10D42FD2788F05ULL, "268387452.ogg"},
	{0xAC10F3EC25A24B7CULL, "6104004_f.fxe"},
	{0xAC119861A4B21EEAULL, "148142950.ogg"},
	{0xAC11F60FE1778014ULL, "dae000ip_crowd_em3_talk13.utp"},
	{0xAC1234E53C35C208ULL, "art\\vfx\\combat\\vfx_r_gn3_embu_act_c.impa.gad"},
	{0xAC1257E1BDF90FDAULL, "1066976184.ogg"},
	{0xAC1262679F894D0CULL, "6208588_f.fxe"},
	{0xAC1292C6FB3A261FULL, "399887043.ogg"},
	{0xAC13AF664319044CULL, "6174466_m.fxe"},
	{0xAC13B18C3AEE4EF9ULL, "vxw_bonepitmining_a2_l7_0_0.vlm"},
	{0xAC13D661DF7FF009ULL, "822569076.ogg"},
	{0xAC149B12D406066BULL, "1000157346.ogg"},
	{0xAC149D5DE688961BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree01_dp.dds"},
	{0xAC151A33CD57839AULL, "gen00fl_aveline6106914.cl"},
	{0xAC15C9D45E97C6ADULL, "564663313.ogg"},
	{0xAC1639F67D4E372EULL, "mag121_idunna6084119.cl"},
	{0xAC167412F5DA7751ULL, "w_encounter3_25_0.anb"},
	{0xAC16B5D7F04C6B1DULL, "vxht_0_hightown_a2_9v3_0_0.vlm"},
	{0xAC16D429FF0CC171ULL, "one140cr_charade6139391.cl"},
	{0xAC170636914A1117ULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.dura.ani"},
	{0xAC170F1EF0848F40ULL, "vxca_caverns2_97v_0_0.vlm"},
	{0xAC171CFB1A5E5151ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsharda.msh"},
	{0xAC1736FB1A5E7D0FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashsharda.mmh"},
	{0xAC1748C5D61F11C5ULL, "art\\levels\\ztest\\chad\\w_lsw_sotf.mao"},
	{0xAC17682E3D902A7BULL, "6019122_m.fxe"},
	{0xAC17A7EB9EA2131AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_reaction_db.gad"},
	{0xAC17A92498D9C49CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_player_chant_ex2.gad"},
	{0xAC17D047E11EC352ULL, "gen00fl_merrill6111227.cl"},
	{0xAC17F6B86BCA9E23ULL, "fen210_dying_magister6108603.cl"},
	{0xAC185BDE7DBD0AD7ULL, "6173132_m.fxe"},
	{0xAC18FF0F7F905FE3ULL, "lowtown_male_alienage1_sta6162735.cl"},
	{0xAC1928F9F72FADE1ULL, "vxht_mansion2b_2pn_0_0.vlm"},
	{0xAC19798E574AF3A1ULL, "vxlt_0_lowtown_a2_fgw_0_0.vlm"},
	{0xAC19D00C7447B0FFULL, "222003673.ogg"},
	{0xAC19F4AA90E89B07ULL, "gen00fl_fenris6126587.cl"},
	{0xAC19FB45455A6690ULL, "hlac_dr_dungeon_3216_6z.dds"},
	{0xAC19FC9F972AF822ULL, "vxlt_warehouse_135_0_0.vlm"},
	{0xAC1A334D9CF07A18ULL, "193798892.ogg"},
	{0xAC1A36A57E7933F3ULL, "vxlt_gallowstemplar_5r2_0_0.vlm"},
	{0xAC1A81035DC53577ULL, "vxlt_gallowscourtyard_n_6rd_0_0.vlm"},
	{0xAC1B823D0A777460ULL, "art\\vfx\\creatures\\fxc_drg_brth_m_c.mmh"},
	{0xAC1B84B1B029D952ULL, "108484054.ogg"},
	{0xAC1CC079319DF7E4ULL, "6096633_f.fxe"},
	{0xAC1CEF93E03F7F26ULL, "vxlt_darktown_67w_0_0.vlm"},
	{0xAC1D192B68EB9E6BULL, "61262246.ogg"},
	{0xAC1D41EE9C1C4E09ULL, "6160642_m.fxe"},
	{0xAC1D4F05BAF98940ULL, "6139425_m.fxe"},
	{0xAC1D5D7994A1F6DFULL, "1031541194.ogg"},
	{0xAC1D9154167075C2ULL, "vxlt_0_lowtown_a2_n_gxw_0_0.vlm"},
	{0xAC1D9E3F646E1E2DULL, "vxlt_ruinedhouse_n_af_0_0.vlm"},
	{0xAC1DD621D65C6C95ULL, "zz_mrl150ip_fence.utp"},
	{0xAC1EA82498DFE19AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_player_chant_ex2.ani"},
	{0xAC1EAAEB9EA83731ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_reaction_db.evt"},
	{0xAC1F5D6CF60C95A7ULL, "hlac_w_blightlands_65532_0y.dds"},
	{0xAC21529E395ECD21ULL, "127034068.ogg"},
	{0xAC21A9E9D428B910ULL, "vxlt_0_lowtown_a2_n_gno_0_0.vlm"},
	{0xAC21C7266C82C028ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd1a.mmh"},
	{0xAC21E5266C82F352ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd1a.msh"},
	{0xAC227926763732B4ULL, "narrative_flemeth6002781.cl"},
	{0xAC229D73912FC575ULL, "gen00fl_anders6113110.cl"},
	{0xAC2339A4843616E4ULL, "and231ar_gallowsbasement.are"},
	{0xAC23688E2234E293ULL, "746560083.ogg"},
	{0xAC238121557F6C59ULL, "vxlt_darktown_4i8_0_0.vlm"},
	{0xAC248712F5E6D4B0ULL, "w_encounter3_25_0.msh"},
	{0xAC249512F5E6ECCAULL, "w_encounter3_25_0.mmh"},
	{0xAC249912F5E6F301ULL, "w_encounter3_25_0.mao"},
	{0xAC24C625D08D4BA7ULL, "dr_primevalentrance.arl"},
	{0xAC2553F2C8B0E332ULL, "gen00fl_varric6076230.cl"},
	{0xAC2566ACDFB70428ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medg_0d.dds"},
	{0xAC25C6B3D50509F7ULL, "6167599_m.fxe"},
	{0xAC262862F2F9550DULL, "ser121_jethann6104429.cl"},
	{0xAC265094419EC08EULL, "lt_0_lowtown_docks_n_254_0.msh"},
	{0xAC265E94419ED8DFULL, "lt_0_lowtown_docks_n_254_0.mao"},
	{0xAC2662A3B786BF24ULL, "6064037_m.fxe"},
	{0xAC266A94419EED7CULL, "lt_0_lowtown_docks_n_254_0.mmh"},
	{0xAC26737754D0595FULL, "and230st_to_dungeon.stg"},
	{0xAC272DBE7E3503EEULL, "gen00fl_anders6125503.cl"},
	{0xAC2756CE09EE3128ULL, "vxht_0_hightown_a3_al9_0_0.vlm"},
	{0xAC27B3467B7447F4ULL, "vxlt_0_lowtown_docks_apm_0_0.vlm"},
	{0xAC28649741E5F837ULL, "art\\vfx\\combat\\vfx_m_spi_md_blast_c.impa.gad"},
	{0xAC286E7C892110A2ULL, "vxlt_warehouse_166_0_0.vlm"},
	{0xAC28802007ED3FE5ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf03.phy"},
	{0xAC2880E83DCD36C4ULL, "gen00fl_fenris6110859.cl"},
	{0xAC28C1F0B24B9E08ULL, "6178332_m.fxe"},
	{0xAC28E20D16EAA56BULL, "vxht_mansion2_2ka_0_0.vlm"},
	{0xAC29653D0A83807FULL, "art\\vfx\\creatures\\fxc_drg_brth_m_c.anb"},
	{0xAC2ACD8809196047ULL, "vxlt_darktown_2s2_0_0.vlm"},
	{0xAC2AFAD923657047ULL, "vxlt_0_lowtown_docks_a3_ars_0_0.vlm"},
	{0xAC2BDE91B16CCC18ULL, "vfx_alignquadtoshadowdirection_p0.gsh.fxo.dx11.assembly"},
	{0xAC2C0734DD6D66CFULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravea.phy"},
	{0xAC2CA6984739B978ULL, "6023985_f.fxe"},
	{0xAC2CC2EB9EB49D00ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_reaction_db.ani"},
	{0xAC2CE42F31218DDDULL, "vxlt_blackemporium_15l_0_0.vlm"},
	{0xAC2CF58B07A70960ULL, "hlac_lt_blackemporium_1055_12y.dds"},
	{0xAC2CF5BB81F32633ULL, "6154706_m.fxe"},
	{0xAC2D75BBA7FBAF18ULL, "vxlt_backalley_n_1bn_0_0.vlm"},
	{0xAC2D7F8C10BF4DA4ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgta_0s.dds"},
	{0xAC2E1074311A9E3CULL, "dae000ip_crowd_em1_talk4.utp"},
	{0xAC2E3B99DC4FD9B2ULL, "hlac_lt_darktown_-1_37x.dds"},
	{0xAC2E799741EA87C4ULL, "art\\vfx\\combat\\vfx_m_spi_md_blast_c.impa.evt"},
	{0xAC2E863792F849E4ULL, "vfx_chunkyparticle_p3.psh.fxo"},
	{0xAC2ECE07A3FF51EEULL, "static_static_p54.psh.fxo.dx11"},
	{0xAC2EF8937616A147ULL, "one200dg_solivitus6048193.cl"},
	{0xAC2EFD5FF81B8180ULL, "nat380_nathaniel6066379.cl"},
	{0xAC2F2233D2182D79ULL, "6018564_f.fxe"},
	{0xAC2F36F9F778C6ACULL, "vxht_playershouse_n_1ie_0_0.vlm"},
	{0xAC2F6665A1FA3B62ULL, "vxht_chantry_25z_0_0.vlm"},
	{0xAC2F830C47E27836ULL, "fighting60.mao"},
	{0xAC2F8664A8A8134DULL, "art\\levels\\ztools\\default\\generic\\9ff_signarrow.phy"},
	{0xAC2F874E4D14A8F9ULL, "mcr370_meredith6085898.cl"},
	{0xAC2F9E58F1B890A7ULL, "88229455.ogg"},
	{0xAC300F03972267FDULL, "9409946.ogg"},
	{0xAC30D5F4D0B77349ULL, "167251272.ogg"},
	{0xAC32AD6395B24BB1ULL, "vxlt_0_lowtown_n_ejh_0_0.vlm"},
	{0xAC32CF8BD8262BE6ULL, "196148508.ogg"},
	{0xAC32DA7505D06AA4ULL, "l_qun100ar_plot_start.bnk"},
	{0xAC331F83D0FD2EB9ULL, "one350cr_vinell.utc"},
	{0xAC338428CF559C9CULL, "qun111_ketojan6039000.cl"},
	{0xAC33A567A2D438ACULL, "mag110_vincento6016319.cl"},
	{0xAC340683AD21C4E6ULL, "vxlt_gallowstemplar_n_5aq_0_0.vlm"},
	{0xAC340F504910D011ULL, "hlac_lt_0_lowtown_a2_n_65538_1y.dds"},
	{0xAC345E945C29DC9FULL, "vxht_brothel_n_3qv_0_0.vlm"},
	{0xAC34FC45B8586D35ULL, "50585642.ogg"},
	{0xAC35F8D735041980ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_basic.phy"},
	{0xAC3608801D593EDAULL, "31088149.ogg"},
	{0xAC360EB365E641CEULL, "gen00fl_isabela6090436.cl"},
	{0xAC362696A684F5EFULL, "883769615.ogg"},
	{0xAC3643440B7DE78CULL, "vxw_woundedcoast_1p0_0_0.vlm"},
	{0xAC36451ADD3613E1ULL, "llac_lt_0_lowtown_a2_131070_0.dds"},
	{0xAC36784D3DC3DF72ULL, "vxht_keep_6y6_0_0.vlm"},
	{0xAC368F4212A8BABEULL, "290646498.ogg"},
	{0xAC369C94B5BDBC2FULL, "10310341.ogg"},
	{0xAC36E2A80F2F8614ULL, "llac_w_bonepit_196608_0.dds"},
	{0xAC372A6832F8D339ULL, "art\\vfx\\models\\vfx_gui_staggered.mmh"},
	{0xAC37346832F8E427ULL, "art\\vfx\\models\\vfx_gui_staggered.msh"},
	{0xAC3740E66BEDC14DULL, "vxlt_gamlenshouse_n_iv_0_0.vlm"},
	{0xAC378EE9AF06F83DULL, "vxca_caverns1_mines_ghr_0_0.vlm"},
	{0xAC37C2CA888EAE2FULL, "8229783.ogg"},
	{0xAC387D284116CC7AULL, "w_woundedcoast_115_0.msh"},
	{0xAC388B284116E3B3ULL, "w_woundedcoast_115_0.mao"},
	{0xAC388F284116EAC0ULL, "w_woundedcoast_115_0.mmh"},
	{0xAC38C3966069FC3AULL, "hlac_lt_0_lowtown_a3_n_c_65535_7z.dds"},
	{0xAC38D965294ABC39ULL, "80221989.ogg"},
	{0xAC39852D4818A79FULL, "mag121_eau_banter6121090.cl"},
	{0xAC39FA9E40EF5D2AULL, "vxca_caverns2prc_7u9_0_0.vlm"},
	{0xAC3A659DFF2B8D34ULL, "vxw_encounter1_e2_0_0.vlm"},
	{0xAC3A91774F88A065ULL, "44425323.ogg"},
	{0xAC3A91B64A8A2653ULL, "gen00fl_bethany6102761.cl"},
	{0xAC3A96ABB6FF1C3EULL, "696261051.ogg"},
	{0xAC3AD2CE6E8DC42BULL, "hlac_ca_cavern1_haunted_24426_7x.dds"},
	{0xAC3B4FE1A010DE84ULL, "zz_fol_debug6066198.cl"},
	{0xAC3C1EC16175C584ULL, "206562340.ogg"},
	{0xAC3C919741F6EDBDULL, "art\\vfx\\combat\\vfx_m_spi_md_blast_c.impa.ani"},
	{0xAC3CBDCFCABCB192ULL, "vxw_bonepit_4r_0_0.vlm"},
	{0xAC3D86298F3E9453ULL, "vrc221_hugin6041114.cl"},
	{0xAC3D97AB9F845BD2ULL, "art\\vfx\\textures\\vfx_blastmask.dds"},
	{0xAC3E436BE10F05C9ULL, "502300487.ogg"},
	{0xAC3E8ED8F21BC2C4ULL, "hlac_lt_0_lowtown_n_131071_5x.dds"},
	{0xAC3EAAF5A2C68BEFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_broken_1.mao"},
	{0xAC3F4C7EDAB08142ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5.phy"},
	{0xAC3F69D569EF584DULL, "mrl250cr_corpse.utc"},
	{0xAC3F9044BBF0A5A0ULL, "vxdr_0_deeproads_d3g_0_0.vlm"},
	{0xAC3F95699C708AC8ULL, "vxht_mansion2b_2fi_0_0.vlm"},
	{0xAC40207563FC6D75ULL, "6000953_m.fxe"},
	{0xAC4066567275DFB4ULL, "lt_0_lowtown_docks_a3_n_135_0.phy"},
	{0xAC40DEFDCE54C90BULL, "6077067_m.fxe"},
	{0xAC40F8D15D115058ULL, "6115566_m.fxe"},
	{0xAC412136B803B6EDULL, "6207001_f.fxe"},
	{0xAC41275A38A28C36ULL, "qun120cr_javaris6042827.cl"},
	{0xAC41379D5A96701EULL, "6207671_f.fxe"},
	{0xAC418E1221341327ULL, "6054891_m.fxe"},
	{0xAC4232550086D50EULL, "502476541.ogg"},
	{0xAC423871FAEC797AULL, "6129040_m.fxe"},
	{0xAC429A0D99415806ULL, "mag110_arianni6016812.cl"},
	{0xAC4361D23F1031FDULL, "qcr000_arishok6057104.cl"},
	{0xAC442E1634600F95ULL, "vxw_encounter2_oq_0_1.vlm"},
	{0xAC4497005F6078F4ULL, "vxlt_0_lowtown_docks_azs_0_0.vlm"},
	{0xAC44EA2A2FB9F7F2ULL, "vxht_keep_6b7_0_0.vlm"},
	{0xAC44EE4FB585A34FULL, "vxlt_shop_dw_0_0.vlm"},
	{0xAC44EFD739BC138FULL, "w_blightlands_179_0.mao"},
	{0xAC44FBD739BC282CULL, "w_blightlands_179_0.mmh"},
	{0xAC4501D739BC321EULL, "w_blightlands_179_0.msh"},
	{0xAC473F2CAB4D23ABULL, "431090558.ogg"},
	{0xAC47D1B5A4CAA0BDULL, "vxht_mansion2_75_0_0.vlm"},
	{0xAC4838A48455317FULL, "and231ar_gallowsbasement.lst"},
	{0xAC483E9CB0D69BB8ULL, "6070137_m.fxe"},
	{0xAC48BC6DBDABD60BULL, "483413303.ogg"},
	{0xAC48D77AF121A5B4ULL, "vxw_encounter1_j3_0_0.vlm"},
	{0xAC4987BF18E095F1ULL, "6087530_m.fxe"},
	{0xAC49F42FF9F5FD4EULL, "592183314.ogg"},
	{0xAC4A42C1925792FFULL, "gen_im_trs_dal_junk1_04.uti"},
	{0xAC4A4356727DF8B1ULL, "lt_0_lowtown_docks_a3_n_135_0.msh"},
	{0xAC4A5156727E10CCULL, "lt_0_lowtown_docks_a3_n_135_0.mao"},
	{0xAC4A5D56727E256FULL, "lt_0_lowtown_docks_a3_n_135_0.mmh"},
	{0xAC4A7682D708BDF0ULL, "hlac_lt_0_lowtown_docks_a3_196607_1y.dds"},
	{0xAC4B45EE7847FE16ULL, "205421758.ogg"},
	{0xAC4BB538056B1938ULL, "follower_banter6172317.cl"},
	{0xAC4BD1C9EE9B3697ULL, "6058709_m.fxe"},
	{0xAC4BF99A065BB30CULL, "gen_im_cft_rec_pot_02.uti"},
	{0xAC4CBF7DDED9F29EULL, "vxca_caverns1_mines_co8_0_0.vlm"},
	{0xAC4D1022C945FA21ULL, "dre000_cut_sibling_dies6036302.cl"},
	{0xAC4D1F5593C28FF8ULL, "649432510.ogg"},
	{0xAC4D2FFBBD6F249CULL, "574629379.ogg"},
	{0xAC4D99182B887570ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medd.mao"},
	{0xAC4D9D182B887B83ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medd.mmh"},
	{0xAC4DAB182B8893DDULL, "art\\characters\\playercharacter\\humanfemale\\glv_medd.msh"},
	{0xAC4E076266620F42ULL, "477834515.ogg"},
	{0xAC4E446FBFB35C1AULL, "one100_solivitus6048114.cl"},
	{0xAC4EE8DE7C8186E6ULL, "vxdr_0_deeproads_avc_0_0.vlm"},
	{0xAC4F510BFCBCA945ULL, "6034352_m.fxe"},
	{0xAC5004F35D738D63ULL, "lt_0_lowtown_a2_n_c_52_0_lc.mmh"},
	{0xAC506888B2DA020BULL, "art\\characters\\playercharacter\\qunarimale\\hed_com.mao"},
	{0xAC508F427F8536F3ULL, "89464907.ogg"},
	{0xAC50CCED5346C588ULL, "llac_w_woundedcoast_131077_0.dds"},
	{0xAC50CE27555A2699ULL, "bink_bink_p0.psh.fxo.dx11.assembly"},
	{0xAC50ED3844716C8BULL, "918817298.ogg"},
	{0xAC51109C18A83DBBULL, "fex_ruin_wl_c.msh"},
	{0xAC512E9C18A870A5ULL, "fex_ruin_wl_c.mmh"},
	{0xAC51496C088EA08EULL, "tools_way_pt05.msh"},
	{0xAC51636C088ECD7CULL, "tools_way_pt05.mmh"},
	{0xAC51D29BACC6AF5FULL, "vxht_0_hightown_a3_cca_0_0.vlm"},
	{0xAC52D0FFE3126637ULL, "vxlt_0_lowtown_e1x_0_0.vlm"},
	{0xAC52D1D3C86A3F8CULL, "397756045.ogg"},
	{0xAC52F1E7F53ABD99ULL, "431469475.ogg"},
	{0xAC535A79CAF6B395ULL, "869816452.ogg"},
	{0xAC53613D27989C26ULL, "6051303_m.fxe"},
	{0xAC53A9DBAEB697ADULL, "6141720_m.fxe"},
	{0xAC5465317586F468ULL, "6030436_m.fxe"},
	{0xAC546BF2D6FEA5AEULL, "vxw_encounter2_camp_pj_0_0.vlm"},
	{0xAC5495CC3A7F444BULL, "hlac_ht_0_hightown_-65538_0z.dds"},
	{0xAC5545FD79D2105EULL, "6122239_f.fxe"},
	{0xAC55939073AB6BC2ULL, "dre000_carver6037607.cl"},
	{0xAC55BDF028E12A4CULL, "vxlt_hangedman_n_w3_0_0.vlm"},
	{0xAC566683E4D8FEBAULL, "w_encounter3_0_0.rml"},
	{0xAC56B079A2CCB438ULL, "6075237_m.fxe"},
	{0xAC56B49A76737B0AULL, "vxht_mansion2b_1hj_0_0.vlm"},
	{0xAC56D5333B31F906ULL, "hlac_lt_blackemporium_1055_15z.dds"},
	{0xAC56DB0E5A507950ULL, "w_bonepitmining_a2_105_0.msh"},
	{0xAC56E90E5A50916AULL, "w_bonepitmining_a2_105_0.mmh"},
	{0xAC56ED0E5A5097A1ULL, "w_bonepitmining_a2_105_0.mao"},
	{0xAC573F31EC911E8DULL, "docks_male_human_worker_sta6175950.cl"},
	{0xAC57873045274606ULL, "399063093.ogg"},
	{0xAC57B3F2647CC06FULL, "ca_caverns1_24423_lc.mmh"},
	{0xAC587D9AD370344FULL, "vxlt_gallowscourtyard_73i_0_0.vlm"},
	{0xAC58833CD262FE88ULL, "follower_banter6156221.cl"},
	{0xAC5976AFBF97C492ULL, "1974308.ogg"},
	{0xAC59EA560CD73F85ULL, "vxlt_0_lowtown_docks_cvx_0_0.vlm"},
	{0xAC5A50AFCF77A7D6ULL, "6016802_f.fxe"},
	{0xAC5A686EFD6C01A9ULL, "168274234.ogg"},
	{0xAC5B15D735236715ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_basic.msh"},
	{0xAC5B17D735236A8BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_basic.mmh"},
	{0xAC5B23D735237F28ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff22_basic.mao"},
	{0xAC5B30A31072DA06ULL, "vxlt_ruinedhouse_n_6p_0_0.vlm"},
	{0xAC5B52E1672A7502ULL, "ico_helm_heavy_1.dds"},
	{0xAC5B7FF822381F5CULL, "6130050_f.fxe"},
	{0xAC5B933CD973879BULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_morr.msh"},
	{0xAC5BB13CD973BA85ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_morr.mmh"},
	{0xAC5BC9A2FF4AF022ULL, "llac_ca_caverns2_5802_52.dds"},
	{0xAC5C99EEC1420EE2ULL, "157078978.ogg"},
	{0xAC5CC0470D87D521ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft30.ani"},
	{0xAC5CC5633CA55418ULL, "ht_bartrand_n_3367_ll.mmh"},
	{0xAC5D05C6338668E8ULL, "362233137.ogg"},
	{0xAC5DCD4F122A2189ULL, "dae000_dalish_apprentices6147530.cl"},
	{0xAC5E55828F50F165ULL, "dae000_mother6135367.cl"},
	{0xAC5E9E182B96F0A8ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medd.phy"},
	{0xAC5EBACC360E1378ULL, "gen00fl_merrill6119650.cl"},
	{0xAC5F04AC5E03F830ULL, "6075000_m.fxe"},
	{0xAC5F45FE3184B1E9ULL, "vxlt_0_lowtown_a2_n_gll_0_0.vlm"},
	{0xAC5F61F22EDD5F23ULL, "151700768.ogg"},
	{0xAC5FBE9EE69E627DULL, "gen00fl_anders6111134.cl"},
	{0xAC6006FE6F5979F3ULL, "gen00fl_isabela_init6077693.cl"},
	{0xAC6012EDAB96DB71ULL, "llac_lt_undercityencounter_-1_3.dds"},
	{0xAC60C6F8E44E0F82ULL, "vxlt_warehouse_w4_0_0.vlm"},
	{0xAC60C71AABB5113CULL, "569828310.ogg"},
	{0xAC612BDB7E97076CULL, "214646485.ogg"},
	{0xAC613BD842734744ULL, "lt_0_lowtown_a2_46_0_ll.mmh"},
	{0xAC616C3CEB717019ULL, "915501381.ogg"},
	{0xAC61CDE413A628F2ULL, "vxht_keep_6h1_0_0.vlm"},
	{0xAC620D63B82A273DULL, "dae000_orsino6154309.cl"},
	{0xAC623FA9EA3A7236ULL, "900879199.ogg"},
	{0xAC6254F6E6B38871ULL, "57860849.ogg"},
	{0xAC628712A362A578ULL, "6195868_f.fxe"},
	{0xAC62B3A3773D5442ULL, "w_bonepitdestroyed_a3_139_0.phy"},
	{0xAC62DDCFC8F4BE2EULL, "vxw_encounter2_camp_kk_0_0.vlm"},
	{0xAC62F439B8EBB76AULL, "mag110_arianni6016908.cl"},
	{0xAC63421685219890ULL, "70098680.ogg"},
	{0xAC63B8C76140DCD5ULL, "vxlt_0_lowtown_docks_bcv_0_0.vlm"},
	{0xAC63CB470D8E06C3ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft30.gad"},
	{0xAC63CE8BF1891ACFULL, "146045421.ogg"},
	{0xAC63F42A4F671F37ULL, "dae000_junar6147862.cl"},
	{0xAC64337EDACF7295ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5.mmh"},
	{0xAC645092EA72261AULL, "vxlt_gallowstemplar_f_4sj_0_0.vlm"},
	{0xAC64717660DD0F78ULL, "103345163.ogg"},
	{0xAC64A43DCE1D363AULL, "vxht_mansion2_1sc_0_0.vlm"},
	{0xAC6526776869938AULL, "vxht_mansion2b_1ck_0_0.vlm"},
	{0xAC652C8378D0028FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12ma.mao"},
	{0xAC65388378D0172CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12ma.mmh"},
	{0xAC653E8378D0211EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12ma.msh"},
	{0xAC658FEA107EA02BULL, "6047680_m.fxe"},
	{0xAC65A80555BC57F8ULL, "125288927.ogg"},
	{0xAC65CF97DE06C742ULL, "448733171.ogg"},
	{0xAC65EA764B1A9F85ULL, "mrl210_merrill6166860.cl"},
	{0xAC661397C1878A2AULL, "918539692.ogg"},
	{0xAC662925FE7744B4ULL, "420887973.ogg"},
	{0xAC679E7A5E8065BEULL, "189032766.ogg"},
	{0xAC68A53586688687ULL, "538441574.ogg"},
	{0xAC68B583D2876797ULL, "6052109_m.fxe"},
	{0xAC692A1E389E3185ULL, "fxe_water_fall.dds"},
	{0xAC69AFA8A0C62163ULL, "633531534.ogg"},
	{0xAC6A41F0125CDF23ULL, "6124789_m.fxe"},
	{0xAC6AA8470D93E9A8ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnleft30.evt"},
	{0xAC6AF7B65316289EULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d02_0t.dds"},
	{0xAC6B6DDE7A43E95FULL, "vxlt_gallowscourtyard_n_4iw_0_0.vlm"},
	{0xAC6BF9CBC7EA8F32ULL, "vxca_caverns2_815_0_2.vlm"},
	{0xAC6C0B77CE3DCD8FULL, "art\\cinematics\\act1\\cs_fp_smile.gad"},
	{0xAC6C2EECFE1AD97BULL, "vxlt_ruinedhouse_n_hw_0_0.vlm"},
	{0xAC6C4C78E242F42FULL, "vxht_mansion2_31b_0_0.vlm"},
	{0xAC6C637DB9AEBC0AULL, "6098977_m.fxe"},
	{0xAC6C957FFDB16F33ULL, "914221334.ogg"},
	{0xAC6CB4155DFA0456ULL, "6049221_m.fxe"},
	{0xAC6CC74DA074FE83ULL, "ca_caverns1_mines_24427.rml"},
	{0xAC6D7FA6D18019D3ULL, "dirty_stuff.mao"},
	{0xAC6DB81D0197743BULL, "hlac_lt_gallowstemplar_f_2394_1y.dds"},
	{0xAC6DBC536432BC34ULL, "gen00fl_dog6114944.cl"},
	{0xAC6E2581D84626EDULL, "dre000_bodahn6033412.cl"},
	{0xAC6E272F0F6BB44EULL, "6141864_m.fxe"},
	{0xAC6E599E3851197CULL, "lgt310_mag_grigor.cnv"},
	{0xAC6F0B75F8A4CE08ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wrba_0n.dds"},
	{0xAC6F1045691BB094ULL, "vxlt_0_lowtown_h4y_0_0.vlm"},
	{0xAC6F38D5D6442802ULL, "vxlt_warehouse_x5_0_0.vlm"},
	{0xAC6F63F56720B43BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofse.msh"},
	{0xAC6F75F56720D286ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofse.mao"},
	{0xAC6F81F56720E725ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofse.mmh"},
	{0xAC6FB9CA13308AB9ULL, "vxht_keep_4hd_0_0.vlm"},
	{0xAC6FCEABD9EBC3C3ULL, "dae210_gate_guard6085043.cl"},
	{0xAC7067647A73418BULL, "6161517_m.fxe"},
	{0xAC70A5197C9A0931ULL, "zz_approval_debug6003666.cl"},
	{0xAC70F173E012F9BBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v2.ani"},
	{0xAC714FACBAEAD6AEULL, "vxw_encounter2_camp_fl_0_0.vlm"},
	{0xAC7155D4D27E5703ULL, "367473735.ogg"},
	{0xAC717B9E385388CCULL, "lgt310_mag_grigor.bnk"},
	{0xAC71F3CA4DE8B883ULL, "vxca_caverns1_mines_inx_0_0.vlm"},
	{0xAC722AA45336F555ULL, "vxlt_0_lowtown_docks_bfu_0_0.vlm"},
	{0xAC728AC5AD229560ULL, "80583105.ogg"},
	{0xAC72B7ADDFA9DCC7ULL, "hlac_dr_0_deeproads_5845_14x.dds"},
	{0xAC73398E6A107A11ULL, "77686409.ogg"},
	{0xAC73B4887740570AULL, "hlac_lt_0_lowtown_docks_1_0y.dds"},
	{0xAC7401C70274B8ABULL, "6042980_m.fxe"},
	{0xAC7407C09166BC84ULL, "547621878.ogg"},
	{0xAC743E61A97D4BF3ULL, "130693571.ogg"},
	{0xAC7517B2E87A95CCULL, "37566289.ogg"},
	{0xAC756154B75C654FULL, "vxlt_gallowscourtyard_79k_0_0.vlm"},
	{0xAC7561A65AF31B98ULL, "vxca_cavern1_haunted_k2p_0_1.vlm"},
	{0xAC75AF0C61CA2998ULL, "vxlt_backalley_n_1yk_0_0.vlm"},
	{0xAC75CFE97972162CULL, "dae011_wench_mob6147250.cl"},
	{0xAC75F2933EA7E059ULL, "1007214606.ogg"},
	{0xAC76AB624B837A35ULL, "6143393_m.fxe"},
	{0xAC76AFAC252A0C34ULL, "440457527.ogg"},
	{0xAC7716835C9FBB08ULL, "527861067.ogg"},
	{0xAC77426C08AF6447ULL, "tools_way_pt05.phy"},
	{0xAC774E7EC7109281ULL, "dae000_party_night_comment6201606.cl"},
	{0xAC774F50FA7CAF54ULL, "darktown_fem_fer_urch_sta6174791.cl"},
	{0xAC774FEE54665755ULL, "6037032_m.fxe"},
	{0xAC776AE87DB0DACAULL, "75568302.ogg"},
	{0xAC77F07A66C57B21ULL, "6164754_m.fxe"},
	{0xAC77F26F49F04A66ULL, "gen00fl_fenris6126539.cl"},
	{0xAC7833D69ABD3844ULL, "6098547_f.fxe"},
	{0xAC787811C4FD4F73ULL, "477472488.ogg"},
	{0xAC788EA099492637ULL, "vxlt_0_lowtown_a2_n_c_eki_0_2.vlm"},
	{0xAC788FCE189CBD30ULL, "6005032_m.fxe"},
	{0xAC790152191E34A2ULL, "1047017835.ogg"},
	{0xAC791186BF7C860EULL, "vxdr_0_deeproads_a26_0_1.vlm"},
	{0xAC7ABE55D4390CAFULL, "vxht_mansion2_34m_0_0.vlm"},
	{0xAC7ADA5EC2EB59FBULL, "vxht_0_hightown_a3_9pb_0_0.vlm"},
	{0xAC7BFEA8D3D9BA57ULL, "156287853.ogg"},
	{0xAC7C488D14C49E47ULL, "gen_im_trs_gen_med_glv_r2.uti"},
	{0xAC7C6BA8896B5635ULL, "252396219.ogg"},
	{0xAC7C748B17FF4C43ULL, "dae000ip_letter_desk.ncs"},
	{0xAC7CE83F57A3F3C5ULL, "1070999303.ogg"},
	{0xAC7CEF445DFCF041ULL, "llac_lt_0_lowtown_docks_n_-3_2.dds"},
	{0xAC7D1B4305E3266FULL, "6015064_f.fxe"},
	{0xAC7D6D104CF5D02FULL, "hm_captain_reiner.mor"},
	{0xAC7D93864779EBCCULL, "6054100_m.fxe"},
	{0xAC7DF973E01D91B2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v2.evt"},
	{0xAC7E1900B38DFF68ULL, "dae000_pickpocket_scene6175103.cl"},
	{0xAC7E1E7253724200ULL, "vxht_mansion2_23f_0_0.vlm"},
	{0xAC7E263C85E34CE5ULL, "201382610.ogg"},
	{0xAC7E663C0004BB4BULL, "675538386.ogg"},
	{0xAC7E8356759242FEULL, "qcr000_arishok6042495.cl"},
	{0xAC7E8A3F7A1F5057ULL, "qcr220_orsino6061807.cl"},
	{0xAC7EA03B0D65907AULL, "mcr370_cut_player_rules6171820.cl"},
	{0xAC7EA904331028AAULL, "6169093_m.fxe"},
	{0xAC7EFC694010844EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_45ang.phy"},
	{0xAC7F798D3DD88187ULL, "pro000ar_gallow_court.lst"},
	{0xAC802877CE4EA855ULL, "art\\cinematics\\act1\\cs_fp_smile.ani"},
	{0xAC80394DC0B44B8CULL, "vxlt_darktown_5dw_0_0.vlm"},
	{0xAC806AA8E32AF1D5ULL, "vxca_cavern1_haunted_i9g_0_0.vlm"},
	{0xAC80832A3EC3A07CULL, "6119456_m.fxe"},
	{0xAC80D9E7B2904755ULL, "vrc321_varric6086932.cl"},
	{0xAC80F23B66EC5F76ULL, "llac_lt_0_lowtown_a3_n_c_131070_3.dds"},
	{0xAC8101EEE98C6F60ULL, "one140cr_charade6134320.cl"},
	{0xAC81B0CE122BAE2EULL, "w_blightlands_184_0.msh"},
	{0xAC81BECE122BC67FULL, "w_blightlands_184_0.mao"},
	{0xAC81CACE122BDA1CULL, "w_blightlands_184_0.mmh"},
	{0xAC820C09CFFDACBFULL, "gen_im_wep_mag_sta_act3_02.uti"},
	{0xAC821B1E754472CBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\pp_varric.anb"},
	{0xAC82BC431AFECFDDULL, "zz_one_debug6077406.cl"},
	{0xAC82F17E438F4008ULL, "one300_zevran6070908.cl"},
	{0xAC82F3B8AFF61A0DULL, "zz_mrlpt_debug.ncs"},
	{0xAC8323C3A39462CCULL, "mrl210_merrill6102159.cl"},
	{0xAC83BEB8135CBCA2ULL, "dae000_arishok6068495.cl"},
	{0xAC8422156CB5E834ULL, "79292483.ogg"},
	{0xAC8433F6A7E747ACULL, "vxlt_0_lowtown_a2_n_c_f6p_0_0.vlm"},
	{0xAC8464D74E75CCA1ULL, "824739316.ogg"},
	{0xAC846501318DEDF5ULL, "vxht_0_hightown_c7r_0_0.vlm"},
	{0xAC84C84D7364E49FULL, "126433365.ogg"},
	{0xAC84D29E5B4EE3DAULL, "plt_ave200pt_long_road.nss"},
	{0xAC84FFD56C8C840DULL, "genim_arm_glv_inq.uti"},
	{0xAC8503685459684EULL, "llac_lt_0_lowtown_docks_a3_n_-3_6.dds"},
	{0xAC851873E023E545ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_attk_v2.gad"},
	{0xAC8521FE3358BD5CULL, "972515322.ogg"},
	{0xAC858EA6D194C569ULL, "dirty_stuff.dds"},
	{0xAC85A50D6C4BE3E3ULL, "314278147.ogg"},
	{0xAC85BAEEDFB3DDC1ULL, "llac_w_sundermount_327677_0.dds"},
	{0xAC85E812F2B05698ULL, "mrl000_marethari6129221.cl"},
	{0xAC85F29FE6F1C771ULL, "vxht_0_hightown_a3_c3o_0_0.vlm"},
	{0xAC8619AA54D2144AULL, "and111_lirene6084390.cl"},
	{0xAC8638E04EEB2CB8ULL, "6163747_m.fxe"},
	{0xAC86CBF445F7E1DDULL, "21727419.ogg"},
	{0xAC86D33C9D96CE2FULL, "vxlt_warehouse_qs_0_0.vlm"},
	{0xAC871B387C85775BULL, "6146864_m.fxe"},
	{0xAC8729BBB2B6E818ULL, "dae000ip_crowd_ff4_talk18.utp"},
	{0xAC8740D3EBCFACF2ULL, "hlac_lt_backalley_n_65536_6z.dds"},
	{0xAC878EA3775C31F6ULL, "w_bonepitdestroyed_a3_139_0.mao"},
	{0xAC879AA3775C4595ULL, "w_bonepitdestroyed_a3_139_0.mmh"},
	{0xAC879CA3775C490BULL, "w_bonepitdestroyed_a3_139_0.msh"},
	{0xAC87C19E6FB993C1ULL, "6058927_m.fxe"},
	{0xAC87F602FF656097ULL, "vxlt_darktown_4w4_0_0.vlm"},
	{0xAC87F922903C4DD7ULL, "6110184_m.fxe"},
	{0xAC88420D87A08454ULL, "vxlt_foundry_n_12e_0_0.vlm"},
	{0xAC88D5BDED1FDE9CULL, "vxlt_warehouse_n_8n_0_0.vlm"},
	{0xAC88DEBFA0A6D1F8ULL, "6039780_f.fxe"},
	{0xAC88FA32E2146D80ULL, "vxht_playershouse_n_1q0_0_1.vlm"},
	{0xAC893CB737AFE6C8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_5.pwk"},
	{0xAC8A30CAB8F86F6EULL, "vxdr_0_deeproads_cm4_0_0.vlm"},
	{0xAC8A3CE974443123ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.dura.evt"},
	{0xAC8A44434C7E2EDAULL, "seb321_leliana6180903.cl"},
	{0xAC8A44D874228BAAULL, "mag111tr_sailors.utt"},
	{0xAC8A74575EFAB92DULL, "vxw_woundedcoast_1dy_0_0.vlm"},
	{0xAC8ABC794ED06D32ULL, "592213871.ogg"},
	{0xAC8B10C822337FCFULL, "lt_0_lowtown_29_0.rml"},
	{0xAC8B8C0C62E36CDFULL, "1067805093.ogg"},
	{0xAC8BE43106D21E2FULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.cess.gad"},
	{0xAC8BEFD0E46FCA8DULL, "w_blightlands_55_0_lc.mmh"},
	{0xAC8C9CC561D5F22FULL, "art\\vfx\\materials\\vfx_crack_tile_add.mao"},
	{0xAC8D5CEC94B582A6ULL, "6137484_f.fxe"},
	{0xAC8DC450F55C77AFULL, "qun210cr_guard6072084.cl"},
	{0xAC8DEF7CE80C6C5BULL, "vxht_bartrand_n_4k7_0_0.vlm"},
	{0xAC8E5F76314643BAULL, "6109002_m.fxe"},
	{0xAC8E8D130A2270B0ULL, "6163530_m.fxe"},
	{0xAC8EAB2AB2AA640CULL, "vxlt_darktown_5it_0_0.vlm"},
	{0xAC8F2FCD68BE2CB0ULL, "vxlt_shop_cq_0_0.vlm"},
	{0xAC8FBB03EA0779DFULL, "771253190.ogg"},
	{0xAC8FBE674BD547FEULL, "music_stinger_triggers.bnk"},
	{0xAC8FD4D6EAC407B8ULL, "hlac_lt_gallowstemplar_n_2391_8z.dds"},
	{0xAC90137295619AF8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_neutral.anb"},
	{0xAC9037EF09A0CAE3ULL, "llac_lt_gallowstemplar_n_2391_3.dds"},
	{0xAC906047A8415D25ULL, "hlac_dr_dungeon_3214_1x.dds"},
	{0xAC907A8D3DE6F6BCULL, "pro000ar_gallow_court.are"},
	{0xAC9093E79286AF92ULL, "854629176.ogg"},
	{0xAC90FBA4CB6A52D2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_cst_lp.gad"},
	{0xAC91137F7E1A2B61ULL, "aip_strider_base.gda"},
	{0xAC9115E9744A0DA4ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.dura.gad"},
	{0xAC91848B1ACDEA00ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_mag_0n.dds"},
	{0xAC918B111970C4F9ULL, "char_stage.lst"},
	{0xAC919A2F1B7888BBULL, "213854004.ogg"},
	{0xAC91BB5FAB081EB6ULL, "vxht_keep_704_0_0.vlm"},
	{0xAC92F93106D860DCULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.cess.evt"},
	{0xAC93692EA5E4CA43ULL, "hlac_lt_0_lowtown_docks_-196609_0y.dds"},
	{0xAC93923DCCFE9573ULL, "vxca_caverns2_812_0_0.vlm"},
	{0xAC93F3CD66754ED2ULL, "6036511_m.fxe"},
	{0xAC94647CD8E7DFF1ULL, "vxht_0_hightown_a3_c4n_0_0.vlm"},
	{0xAC94B1A5FB6E7C22ULL, "ave221_brennan6110460.cl"},
	{0xAC94BD692B069614ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_1.ani"},
	{0xAC951C050475AE2EULL, "76149601.ogg"},
	{0xAC9550C239DE422EULL, "hlac_lt_backalley_g_65536_2y.dds"},
	{0xAC956780A0E25704ULL, "t3_psp_b01.tnt"},
	{0xAC960C675F336216ULL, "888493136.ogg"},
	{0xAC962C606E919A1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\blendtree\\trooper_overlay.anb"},
	{0xAC96798814115A56ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm06.msh"},
	{0xAC96838814116B54ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm06.mmh"},
	{0xAC9686D651268EABULL, "573500627.ogg"},
	{0xAC96D8E41214D674ULL, "gen00fl_fenris6105589.cl"},
	{0xAC9729F49ACF7D06ULL, "propalpha1_ds2y.psh.fxo"},
	{0xAC972E43F1F081C4ULL, "vxlt_0_lowtown_docks_a3_n_9bs_0_0.vlm"},
	{0xAC9760FF542F2E06ULL, "lt_0_lowtown_n_26_0_lc.mmh"},
	{0xAC976CC9E6EE7BDAULL, "vxlt_0_lowtown_fib_0_0.vlm"},
	{0xAC97890F3734176CULL, "hlac_lt_shop_-1_1y.dds"},
	{0xAC981EB800E2C702ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_basic_2.mao"},
	{0xAC9824E974504582ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.dura.ani"},
	{0xAC985142499B2F29ULL, "mcr371_hostile_aveline6082181.cl"},
	{0xAC98A2A7AAEE87EEULL, "vxdr_0_deeproads_cj5_0_0.vlm"},
	{0xAC98A7B046A1206DULL, "vxdr_0_deeproads_cqw_0_0.vlm"},
	{0xAC98E63450F0D1ADULL, "vxw_woundedcoast_1ox_0_0.vlm"},
	{0xAC992CDB52B3F1B8ULL, "vxlt_hangedman_n_27w_0_0.vlm"},
	{0xAC999EC7DDC677F3ULL, "319355852.ogg"},
	{0xAC99D41BBEB2A0B3ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_enter.ani"},
	{0xAC99E06A35FB5BECULL, "vxht_0_hightown_al9_0_1.vlm"},
	{0xAC9A2B2AA0B992ECULL, "hlac_ca_caverns1_mines_24421_12z.dds"},
	{0xAC9A3AD872F0942BULL, "lt_0_lowtown_a2_n_c_44_0.anb"},
	{0xAC9A500D169DF9F5ULL, "103545476.ogg"},
	{0xAC9A7FAA1E13CF1EULL, "vxht_0_hightown_7g1_0_0.vlm"},
	{0xAC9A8B8132C4EB37ULL, "combat_smr3_range.gda"},
	{0xAC9AF550498E843AULL, "llac_lt_0_lowtown_docks_a3_n_65533_0.dds"},
	{0xAC9AFD3BC22935AEULL, "6163356_m.fxe"},
	{0xAC9B4D793C6BE25DULL, "ht_0_hightown_a2_12_0.anb"},
	{0xAC9B8CBE16FC59AAULL, "6163293_m.fxe"},
	{0xAC9BBA692B0CB076ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_1.gad"},
	{0xAC9BBBA849FFDE18ULL, "gen00fl_isabela6090845.cl"},
	{0xAC9BC69D221BD365ULL, "1165327.ogg"},
	{0xAC9C09775E77F2E3ULL, "t1_mue_ye2.tnt"},
	{0xAC9C7DD9F95C279BULL, "vxht_0_hightown_c3x_0_0.vlm"},
	{0xAC9C8B41EF4CD2E9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospiderdeadlp.ani"},
	{0xAC9CD153233C5C3AULL, "6102102_m.fxe"},
	{0xAC9D0A001790855AULL, "ico_maul_4.dds"},
	{0xAC9D0F3FA15678D6ULL, "ht_0_hightown_26_0_ll.mmh"},
	{0xAC9D1D07A4A07C8CULL, "vxlt_darktown_5ru_0_0.vlm"},
	{0xAC9D4683E514CB71ULL, "w_encounter3_0_0.mmh"},
	{0xAC9D4E62C71722D5ULL, "vxca_cavern1_haunted_i3e_0_0.vlm"},
	{0xAC9D5283E514DF12ULL, "w_encounter3_0_0.mao"},
	{0xAC9D6083E514F72FULL, "w_encounter3_0_0.msh"},
	{0xAC9DA6264ACBAD21ULL, "llac_ht_mansion2_818_23.dds"},
	{0xAC9DE02D0E753543ULL, "6118126_m.fxe"},
	{0xAC9E115761A69936ULL, "hlac_lt_0_lowtown_n_131070_1y.dds"},
	{0xAC9EEDEB3041F58AULL, "vxht_mansion2b_1wg_0_0.vlm"},
	{0xAC9F2459F11D40E9ULL, "hlac_w_encounter2_262148_0z.dds"},
	{0xAC9F26BA9DF06940ULL, "engineevents.gda"},
	{0xAC9FC8882D2C9BEAULL, "6043498_m.fxe"},
	{0xAC9FD989B406CD3DULL, "6096467_f.fxe"},
	{0xAC9FE0059DA7DFBFULL, "6129150_f.fxe"},
	{0xACA01AEE8A0132BEULL, "6028840_m.fxe"},
	{0xACA08CAFAF39F19AULL, "fxe_day_sunshine.mao"},
	{0xACA0F91982815B6DULL, "mag100_karras6034908.cl"},
	{0xACA1013106E4ABF5ULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.cess.ani"},
	{0xACA12584777E7DA1ULL, "vxw_encounter1_br_0_0.vlm"},
	{0xACA13E911B448D54ULL, "760705072.ogg"},
	{0xACA1614B58F1AA3AULL, "dae000_elegant6166903.cl"},
	{0xACA169B48ACFECABULL, "vxlt_undercitydungeon_1ih_0_0.vlm"},
	{0xACA17601B6809A43ULL, "gen00fl_sebastian6189028.cl"},
	{0xACA17F932F8AEADEULL, "6108600_f.fxe"},
	{0xACA1A1232160ABB9ULL, "hlac_lt_undercityserialkiller_-1_40y.dds"},
	{0xACA21FD87436DE93ULL, "mag111tr_sailors.ncs"},
	{0xACA2441E4AC72050ULL, "6144308_m.fxe"},
	{0xACA26BBAABDBB57FULL, "634324771.ogg"},
	{0xACA28411197F2BBEULL, "char_stage.are"},
	{0xACA28A3CD9AF7B32ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_morr.phy"},
	{0xACA2A259C5392707ULL, "dae000_sandal6148228.cl"},
	{0xACA2ABDD2A521014ULL, "hlac_lt_0_lowtown_a2_n_c_0_9y.dds"},
	{0xACA2D537353E4B3AULL, "zz_one_debug6079306.cl"},
	{0xACA2D5692B12FCD5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_1.evt"},
	{0xACA3445C6FB08ED0ULL, "vxht_0_hightown_a2_c60_0_0.vlm"},
	{0xACA38C4EB2142C89ULL, "isa311_party_comments6117811.cl"},
	{0xACA3A641EF531F6BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualtospiderdeadlp.gad"},
	{0xACA3D908071EDF64ULL, "qcr000_arishok6059343.cl"},
	{0xACA4BAA8FD4164FBULL, "6070606_m.fxe"},
	{0xACA4C467D3764AFAULL, "6086268_m.fxe"},
	{0xACA4CFA4860AD20AULL, "zz_dre_debug6069203.cl"},
	{0xACA4D9BCE3519197ULL, "vxlt_darktown_4y6_0_0.vlm"},
	{0xACA4E6A4CB7AD878ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_cst_lp.ani"},
	{0xACA4ED86DDE7E4DDULL, "hlac_lt_gallowscourtyard_n_-131071_0x.dds"},
	{0xACA583064C96E842ULL, "vxlt_0_lowtown_docks_n_c6p_0_0.vlm"},
	{0xACA5CE0BAA0AF611ULL, "75699560.ogg"},
	{0xACA5FA494BB6734CULL, "383457956.ogg"},
	{0xACA62C319EA58633ULL, "dre000_cut_sibling_dies6036326.cl"},
	{0xACA63D891521AAB5ULL, "art\\characters\\weapons\\longbows\\w_lbw_lw106a.mmh"},
	{0xACA63F891521AE2BULL, "art\\characters\\weapons\\longbows\\w_lbw_lw106a.msh"},
	{0xACA649CCFC4406DCULL, "6207779_m.fxe"},
	{0xACA6AB8338ACF13DULL, "art\\levels\\ztools\\default\\generic\\9ff_torch.phy"},
	{0xACA6C07D56F8CACAULL, "vxht_keep_645_0_0.vlm"},
	{0xACA7581142E6EA2DULL, "vxw_woundedcoast_1jw_0_0.vlm"},
	{0xACA7904F39B63716ULL, "gen00fl_anders6113043.cl"},
	{0xACA7956B3CD288E6ULL, "vxlt_gallowstemplar_n_59y_0_0.vlm"},
	{0xACA79EB844AA0A38ULL, "vxlt_hangedman_n_22t_0_0.vlm"},
	{0xACA7A9CE124C71E7ULL, "w_blightlands_184_0.phy"},
	{0xACA7CC1BBEBED0EAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_enter.evt"},
	{0xACA80979BB022428ULL, "vxht_keep_5gm_0_0.vlm"},
	{0xACA83E8D0A753428ULL, "hm_qsw_warrens5.mor"},
	{0xACA87282F411A2A0ULL, "6093581_m.fxe"},
	{0xACA980CBCC1BE367ULL, "llac_dr_0_deeproads_5849_5.dds"},
	{0xACAA875FB9450C00ULL, "6149606_m.fxe"},
	{0xACAC235320C58D29ULL, "183366824.ogg"},
	{0xACAC520A006B4DC3ULL, "6117026_m.fxe"},
	{0xACACB0A3150B5DA0ULL, "llac_w_sundermount_262144_0.dds"},
	{0xACAD156975C3F731ULL, "lt_0_lowtown_docks_a3_198_0.msh"},
	{0xACAD236975C40F4CULL, "lt_0_lowtown_docks_a3_198_0.mao"},
	{0xACAD2F6975C423EFULL, "lt_0_lowtown_docks_a3_198_0.mmh"},
	{0xACAD599DD801B7B5ULL, "vxdr_0_deeproads_ehb_0_0.vlm"},
	{0xACAD8881AD386DA7ULL, "dae000_gamlen6082930.cl"},
	{0xACAD992AD7ED8DC8ULL, "307490063.ogg"},
	{0xACADD9C92D9D4CD2ULL, "414364077.ogg"},
	{0xACAEA7D93BA4CFE2ULL, "dae000_madam_lusine6151221.cl"},
	{0xACAEAB1BBEC4B75DULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_enter.gad"},
	{0xACAF976169749621ULL, "vxw_encounter1_gq_0_0.vlm"},
	{0xACAF9B919059F515ULL, "vxht_chantry_n_1xf_0_0.vlm"},
	{0xACAFCCE45FBBFFFDULL, "803066653.ogg"},
	{0xACAFD7BEBDA43A4AULL, "6071207_f.fxe"},
	{0xACB026C808DBFAABULL, "hlac_ht_keep_4519_1z.dds"},
	{0xACB0338D0605A115ULL, "vxlt_gallowstemplar_n_3ru_0_0.vlm"},
	{0xACB07CB3C0A29ABBULL, "6125492_m.fxe"},
	{0xACB1E762260529A6ULL, "6026998_m.fxe"},
	{0xACB24815E68ED438ULL, "one100cr_solivitus.ncs"},
	{0xACB326A3DA2179F9ULL, "415025855.ogg"},
	{0xACB36AA7B94EF9F1ULL, "dae000_orsino6085520.cl"},
	{0xACB3921BDDE956CEULL, "llac_lt_darktown_-1_35.dds"},
	{0xACB39289904C8980ULL, "vxlt_gallowscourtyard_710_0_0.vlm"},
	{0xACB39A7B6F029105ULL, "385519154.ogg"},
	{0xACB3F4E33E8D00C2ULL, "vxlt_0_lowtown_docks_n_c1q_0_0.vlm"},
	{0xACB44390DE71517BULL, "6069979_f.fxe"},
	{0xACB4678AA6D2F9A1ULL, "follower_banter6155586.cl"},
	{0xACB495E010AF24B2ULL, "123082721.ogg"},
	{0xACB4D01D025F3C86ULL, "vxht_bartrand_n_4d3_0_0.vlm"},
	{0xACB509F0E6D1E97AULL, "684428381.ogg"},
	{0xACB5325A48EEE34AULL, "vxht_keep_616_0_0.vlm"},
	{0xACB534756345010AULL, "492968103.ogg"},
	{0xACB5E12841C78938ULL, "em_gen_dalish_archer.mor"},
	{0xACB618807B2A2776ULL, "llac_lt_0_lowtown_docks_a3_n_-131071_0.dds"},
	{0xACB64D79FC32F040ULL, "143791434.ogg"},
	{0xACB6B377CA0C3F57ULL, "w_bonepitdestroyed_a3_212_0.mmh"},
	{0xACB6B777CA0C4664ULL, "w_bonepitdestroyed_a3_212_0.mao"},
	{0xACB6C977CA0C64C9ULL, "w_bonepitdestroyed_a3_212_0.msh"},
	{0xACB77B2E4DFEF172ULL, "hlac_lt_0_lowtown_n_-131072_1x.dds"},
	{0xACB78026512C8910ULL, "vxlt_0_lowtown_a2_gy9_0_0.vlm"},
	{0xACB7A08338BB51A9ULL, "art\\levels\\ztools\\default\\generic\\9ff_torch.mao"},
	{0xACB7AC8338BB6642ULL, "art\\levels\\ztools\\default\\generic\\9ff_torch.mmh"},
	{0xACB7AE8338BB69F8ULL, "art\\levels\\ztools\\default\\generic\\9ff_torch.msh"},
	{0xACB7E061A930404DULL, "82903326.ogg"},
	{0xACB7E7711E8A0696ULL, "232664199.ogg"},
	{0xACB825BDCBEE0857ULL, "gen00fl_fenris6127034.cl"},
	{0xACB8A5CC5E38F9A6ULL, "one140cr_charade6137033.cl"},
	{0xACB91F979E766B16ULL, "113294628.ogg"},
	{0xACBA01240ADB572DULL, "738494600.ogg"},
	{0xACBA2023FACF1954ULL, "381674278.ogg"},
	{0xACBA216D4DF96EBFULL, "plt_lgt000pt_ultra_1_04.nss"},
	{0xACBA4452A93B68A9ULL, "lt_0_lowtown_docks_a3_n_mmi.gff"},
	{0xACBA7E86A9E4C6D4ULL, "6148623_m.fxe"},
	{0xACBA80732137A456ULL, "mcr310_sibling6076259.cl"},
	{0xACBAB3F23DAF4FE9ULL, "llac_lt_0_lowtown_a2_65537_2.dds"},
	{0xACBAE194A79A79E7ULL, "740472702.ogg"},
	{0xACBB5E0F922949E7ULL, "931658153.ogg"},
	{0xACBC329D5734C314ULL, "vxht_chantry_n_2bz_0_0.vlm"},
	{0xACBC42C52884E857ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright.msh"},
	{0xACBC54C5288506BAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright.mao"},
	{0xACBC58C528850DC9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright.mmh"},
	{0xACBD0B4F9719F8E8ULL, "6037701_m.fxe"},
	{0xACBD9D24E8F458D2ULL, "6119880_f.fxe"},
	{0xACBDBA0041D0E032ULL, "gen00fl_isabela6097558.cl"},
	{0xACBDDA93E1978EAFULL, "qun131_taskmaster6177184.cl"},
	{0xACBE4D6E6EBC1DABULL, "vxlt_undercitydungeon_1wj_0_0.vlm"},
	{0xACBE522D1AE40929ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsackwheat.phy"},
	{0xACBE8B11A15941DCULL, "vxlt_shop_sm_0_0.vlm"},
	{0xACBEFB6940466611ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_45ang.mmh"},
	{0xACBF076940467AB2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_45ang.mao"},
	{0xACBF1569404692CFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_45ang.msh"},
	{0xACC00FD5634A5861ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_008a_0n.dds"},
	{0xACC05C53F9E49C6DULL, "vxht_chantry_2cw_0_0.vlm"},
	{0xACC08B40E472D0CEULL, "455408358.ogg"},
	{0xACC0A17ABB798953ULL, "vxw_bonepitdestroyed_a3_25p_0_0.vlm"},
	{0xACC12F3F4306FB0CULL, "one000pt_herbalist.ncs"},
	{0xACC145D2D2DB09CAULL, "6167006_m.fxe"},
	{0xACC14E8C5239DD4CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_bars.msh"},
	{0xACC1548C5239E73EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_bars.mmh"},
	{0xACC15F104EB692FCULL, "mag160_feynriel6020429.cl"},
	{0xACC1608C5239FBD5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_bars.mao"},
	{0xACC29C2C6EFE8450ULL, "llac_ca_caverns1_24425_4.dds"},
	{0xACC2A3DDE307557DULL, "hlac_lt_gallowstemplar_2391_2z.dds"},
	{0xACC32005094F40B1ULL, "vxlt_0_lowtown_n_e8f_0_0.vlm"},
	{0xACC366A61F1F9629ULL, "vxw_encounter2_dt_0_1.vlm"},
	{0xACC3E743B295E1C1ULL, "6005381_m.fxe"},
	{0xACC3EA95231EE7FBULL, "54471882.ogg"},
	{0xACC43BCB26D31B2DULL, "vxw_woundedcoast_1pu_0_0.vlm"},
	{0xACC4567466E097FBULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_evilchya_0.pwk"},
	{0xACC4638E2A9D3150ULL, "gen00fl_fenris6110882.cl"},
	{0xACC4C8015F510DDBULL, "qun210cr_random_thug.bnk"},
	{0xACC6055D4BE7A39EULL, "w_encounter1_28_0.rml"},
	{0xACC6583722090E34ULL, "vxlt_0_lowtown_f4o_0_0.vlm"},
	{0xACC68432A8D823E1ULL, "mcr370_meredith6083610.cl"},
	{0xACC6AFC5D710F954ULL, "hlac_dr_0_deeproads_5849_4x.dds"},
	{0xACC6F4792D448D55ULL, "dae000_cora6162464.cl"},
	{0xACC737D873177D77ULL, "lt_0_lowtown_a2_n_c_44_0.rml"},
	{0xACC777BB5DA00DBFULL, "302757245.ogg"},
	{0xACC7B2996976C63EULL, "gen00fl_anders6152729.cl"},
	{0xACC7F7A12C25DF4AULL, "hla_ht_0_hightown_a2_-2_0y.dds"},
	{0xACC809ED3E93B0CCULL, "vxlt_hangedman_21r_0_0.vlm"},
	{0xACC8CA015F54F99BULL, "qun210cr_random_thug.cnv"},
	{0xACC8CEF67B20F399ULL, "760612283.ogg"},
	{0xACC8FA8BF46DD0BCULL, "hlac_lt_0_lowtown_docks_a3_131071_3y.dds"},
	{0xACC91355707AE02DULL, "12506010.ogg"},
	{0xACC9812F1A68B795ULL, "isa311_isabela6092501.cl"},
	{0xACC9D67F6D402325ULL, "hlac_ca_cavern1_haunted_24423_8x.dds"},
	{0xACCA22307CCE27D2ULL, "vxw_bonepitdestroyed_a3_25u_0_0.vlm"},
	{0xACCA465B336B4120ULL, "vxht_0_hightown_a3_c2a_0_0.vlm"},
	{0xACCA5022DA1C864CULL, "vrc321tr_spooky_scenes_5.ncs"},
	{0xACCA7BCAA008884EULL, "vxca_caverns1_ioc_0_0.vlm"},
	{0xACCB7D2C89101168ULL, "6030601_m.fxe"},
	{0xACCC16AEBADE6B29ULL, "lgt150ip_ntb_wr_ing_chest_2.utp"},
	{0xACCC7FE23CC242CFULL, "269886875.ogg"},
	{0xACCC9C7307DE57E6ULL, "650690862.ogg"},
	{0xACCCA773CD70EB36ULL, "339815274.ogg"},
	{0xACCD17D7B72C2D02ULL, "one300_zevran6086763.cl"},
	{0xACCD4729797CA1B0ULL, "w_sundermount_105_0.mao"},
	{0xACCD4B29797CA8C3ULL, "w_sundermount_105_0.mmh"},
	{0xACCD520794F88524ULL, "vxlt_backalley_g_1l5_0_0.vlm"},
	{0xACCD5929797CC01DULL, "w_sundermount_105_0.msh"},
	{0xACCD79791D5C5390ULL, "640335106.ogg"},
	{0xACCDBCD6C555F8B6ULL, "vxlt_hangedman_n_o1_0_0.vlm"},
	{0xACCE2641ACF44CB1ULL, "w_bonepit_115_0.msh"},
	{0xACCE3441ACF464CCULL, "w_bonepit_115_0.mao"},
	{0xACCE4041ACF4796FULL, "w_bonepit_115_0.mmh"},
	{0xACCEE981F026777AULL, "mag370_meredith.cnv"},
	{0xACCF146D4E0BB4D0ULL, "plt_lgt000pt_ultra_1_04.plo"},
	{0xACCF295D126BF84FULL, "579795874.ogg"},
	{0xACCFB4FA562AC95CULL, "fen110_party_comment6095068.cl"},
	{0xACCFF29F5CAC4C89ULL, "mag320_bsc_banter6100772.cl"},
	{0xACD045059BECC9FDULL, "mag170_hugh6092220.cl"},
	{0xACD0609655B6E993ULL, "hlac_ht_0_hightown_n_-196608_2x.dds"},
	{0xACD090674C0C9004ULL, "music_stinger_triggers.wbd"},
	{0xACD0A7D89B419327ULL, "one220_dera6104087.cl"},
	{0xACD0B75A40993E24ULL, "898809516.ogg"},
	{0xACD0CA2488B8246BULL, "vxht_0_hightown_n_cai_0_0.vlm"},
	{0xACD12C8DB832FF69ULL, "6043605_f.fxe"},
	{0xACD1542A86489811ULL, "119149172.ogg"},
	{0xACD16E015712FE23ULL, "mag370_meredith6092178.cl"},
	{0xACD1D8831115AEA9ULL, "vxw_encounter2_ou_0_1.vlm"},
	{0xACD20B81F028E6C2ULL, "mag370_meredith.bnk"},
	{0xACD21E72E3E9C455ULL, "demoral_space.mao"},
	{0xACD221A3D6DD394CULL, "vxca_caverns1_mines_fm1_0_0.vlm"},
	{0xACD221B873ECDA9FULL, "vxw_bonepitmining_a2_26c_0_0.vlm"},
	{0xACD232FFDEB7FC55ULL, "dae150_chandan6148660.cl"},
	{0xACD240AA39598A08ULL, "dae000ip_crowd_em2_talk10.utp"},
	{0xACD25289F7875731ULL, "6059674_m.fxe"},
	{0xACD252A801D7E935ULL, "gen00fl_merrill6150997.cl"},
	{0xACD25D4A34AEA0DEULL, "6111573_m.fxe"},
	{0xACD2969DE95CB2B1ULL, "art\\cinematics\\deathblows\\highdragon\\c_highdragon.tal_swp.ani"},
	{0xACD3368D4017EA6FULL, "hlac_ht_mansion2_818_18x.dds"},
	{0xACD365CA2C724BFEULL, "60063122.ogg"},
	{0xACD3A97C60B7E6C2ULL, "achievements_col.gda"},
	{0xACD3D4808B5F4BD3ULL, "em_on_mo_elren.mor"},
	{0xACD54E793C9D5C55ULL, "ht_0_hightown_a2_12_0.rml"},
	{0xACD5C9F16D9D6904ULL, "gen_im_npc_arm_hum_hlm_gwda.uti"},
	{0xACD6546F708053D2ULL, "6005828_f.fxe"},
	{0xACD684A2D5E35743ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_black.dds"},
	{0xACD74FF1AA53BFFEULL, "dae000cr_grand_cleric6124778.cl"},
	{0xACD75DC6EEDE857FULL, "vxht_keep_6mk_0_0.vlm"},
	{0xACD781D73802DD0EULL, "gen00fl_anders6119405.cl"},
	{0xACD7979133C59B56ULL, "838886896.ogg"},
	{0xACD7C2CD8AEE2FA3ULL, "vxlt_0_lowtown_docks_a3_n_a0n_0_0.vlm"},
	{0xACD7F4C876B46F89ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_101a_s.dds"},
	{0xACD84A637E465FADULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgte.mmh"},
	{0xACD84E637E4666FEULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgte.mao"},
	{0xACD85C637E467E33ULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgte.msh"},
	{0xACD8934C53A13F27ULL, "vxlt_foundry_n_vy_0_0.vlm"},
	{0xACD8B838256159A0ULL, "vxht_0_hightown_a3_c7n_0_0.vlm"},
	{0xACD8DC4678F2B344ULL, "6166577_m.fxe"},
	{0xACD8EF091E659427ULL, "hlac_w_bonepit_65537_0y.dds"},
	{0xACD980F2F26C7E6CULL, "757705632.ogg"},
	{0xACD9A19DE962E453ULL, "art\\cinematics\\deathblows\\highdragon\\c_highdragon.tal_swp.gad"},
	{0xACDA0B0A747D81A5ULL, "gen00fl_anders6116834.cl"},
	{0xACDA2AED85A9A7A1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodfull.phy"},
	{0xACDA307CFDC20C97ULL, "6217009_m.fxe"},
	{0xACDA3F555824D58DULL, "gen00fl_varric6033618.cl"},
	{0xACDA95E6018004BAULL, "darktown_male_fer_urch_sta6174945.cl"},
	{0xACDB403D5BC2CCCEULL, "gen00fl_isabela6118365.cl"},
	{0xACDB666D9ED89A89ULL, "lgt220tr_templar_captain.ncs"},
	{0xACDBB98BCD697004ULL, "vxca_caverns2_97r_0_0.vlm"},
	{0xACDBEB55FFBAF2A7ULL, "vxlt_gallowstemplar_f_5rq_0_0.vlm"},
	{0xACDC0C95F7F3B749ULL, "6093900_m.fxe"},
	{0xACDCB33FAD324FC6ULL, "cod_bks_chantry_hist_1.utp"},
	{0xACDD400DDDD0CD6DULL, "vxht_chantry_2yu_0_0.vlm"},
	{0xACDD4FB5BB89AACBULL, "797865452.ogg"},
	{0xACDD5AEB872C1F61ULL, "hlac_lt_hangedman_n_-1_12z.dds"},
	{0xACDD74B7C65B8D0AULL, "6177708_m.fxe"},
	{0xACDDB4D0B66C831BULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzcratefruit_0.pwk"},
	{0xACDE4C29798B1DE8ULL, "w_sundermount_105_0.phy"},
	{0xACDE90C283B5F758ULL, "895540328.ogg"},
	{0xACDF5327047FE718ULL, "21365135.ogg"},
	{0xACE02961740A10D9ULL, "642355631.ogg"},
	{0xACE032B36B1807D8ULL, "123948946.ogg"},
	{0xACE09FAF783FDB29ULL, "122684886.ogg"},
	{0xACE1ABA27141A4C9ULL, "vxht_0_hightown_a2_bqe_0_0.vlm"},
	{0xACE1C29F7AA1B1AFULL, "vxw_encounter3_6b_0_0.vlm"},
	{0xACE1EA404E0FCE00ULL, "214206856.ogg"},
	{0xACE20123BDA98BF6ULL, "6094951_m.fxe"},
	{0xACE22C1A63E34874ULL, "vxht_0_hightown_n_c6k_0_0.vlm"},
	{0xACE2DF9B1B481BE3ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic.mao"},
	{0xACE2E39B1B482230ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic.mmh"},
	{0xACE2F19B1B483A4AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic.msh"},
	{0xACE3288A9CC4AEAEULL, "2960593.ogg"},
	{0xACE397F82745C1A4ULL, "llac_w_sundermount_-2_2.dds"},
	{0xACE3FE366F29A834ULL, "llac_lt_backalley_g_0_2.dds"},
	{0xACE49CB96288754FULL, "vxw_bonepitdestroyed_a3_19k_0_0.vlm"},
	{0xACE4B3EF8DD2C77FULL, "mag100_karras6035310.cl"},
	{0xACE4BB18993A8C77ULL, "vxlt_gallowscourtyard_n_68z_0_0.vlm"},
	{0xACE4BD920CC02750ULL, "6048181_m.fxe"},
	{0xACE50A4B763F7247ULL, "and100_anders6095928.cl"},
	{0xACE54B2979913AE0ULL, "w_sundermount_105_0.rml"},
	{0xACE5889FAE23C222ULL, "206501666.ogg"},
	{0xACE5A54B8F7A4EC7ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.cess.gad"},
	{0xACE659077D313664ULL, "393018060.ogg"},
	{0xACE694B2A22B2B5AULL, "1023919894.ogg"},
	{0xACE6BF359915E700ULL, "hlac_lt_0_lowtown_docks_-65536_0x.dds"},
	{0xACE6E50F9DCD0EDBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wi_galowint.phy"},
	{0xACE6FA63519DB4BFULL, "vrc221_bartrand6044859.cl"},
	{0xACE70F76DB840DDEULL, "6121505_m.fxe"},
	{0xACE75D3B08255E23ULL, "art\\levels\\ztools\\default\\generic\\vista_basic.mao"},
	{0xACE7613B08256570ULL, "art\\levels\\ztools\\default\\generic\\vista_basic.mmh"},
	{0xACE76F3B08257C8AULL, "art\\levels\\ztools\\default\\generic\\vista_basic.msh"},
	{0xACE775294C905523ULL, "sse_ss_coterie_assassin_m.gda"},
	{0xACE78F84AFB62412ULL, "vxht_chantry_n_35p_0_0.vlm"},
	{0xACE7C996207EF388ULL, "gen00fl_bethany6109233.cl"},
	{0xACE7C9EAE1291B62ULL, "76658181.ogg"},
	{0xACE7FEF59A80F9B9ULL, "478756305.ogg"},
	{0xACE870D3558D3C4BULL, "100688453.ogg"},
	{0xACE8F4B6B2CF20E9ULL, "6048261_m.fxe"},
	{0xACE93885B1D163B0ULL, "6142009_m.fxe"},
	{0xACE943637E54C76AULL, "art\\characters\\playercharacter\\humanmale\\hm_boo_lgte.phy"},
	{0xACE94A65DD2C92C9ULL, "vxdr_primevalentrance_287_0_0.vlm"},
	{0xACE95E60D5A78309ULL, "vxca_caverns1_mines_iic_0_0.vlm"},
	{0xACE9DEDB0F9E304EULL, "1062751057.ogg"},
	{0xACE9EDA008BDA6BCULL, "637269951.ogg"},
	{0xACE9F25C759FB77CULL, "dae000_worthy6166562.cl"},
	{0xACE9FC623C23A156ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_loop.ani"},
	{0xACE9FFC5F1911E3CULL, "w_bonepit_41_0.mao"},
	{0xACEA09B7B9FC265CULL, "784478255.ogg"},
	{0xACEA0BC5F19132DFULL, "w_bonepit_41_0.mmh"},
	{0xACEA11C5F1913C81ULL, "w_bonepit_41_0.msh"},
	{0xACEA1BDB8F6B913CULL, "gen00fl_player6207771.cl"},
	{0xACEA3B2D1B091D4EULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsackwheat.mmh"},
	{0xACEA472D1B0931E5ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsackwheat.mao"},
	{0xACEA552D1B09493CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsackwheat.msh"},
	{0xACEA5D32F1B10B27ULL, "vxlt_gallowstemplar_f_5up_0_0.vlm"},
	{0xACEA97C6FC05B952ULL, "hlac_lt_0_lowtown_docks_a3_n_-196608_0y.dds"},
	{0xACEAEF9E5BA5C84DULL, "plt_ave200pt_long_road.plo"},
	{0xACEB64E71F41DECAULL, "hlac_w_bonepit_0_10y.dds"},
	{0xACEC4F6D9EE6E76CULL, "lgt220tr_templar_captain.utt"},
	{0xACEC6C4FF0CE980FULL, "6072271_m.fxe"},
	{0xACEC6F01D45585B5ULL, "hlac_lt_0_lowtown_n_-65536_1x.dds"},
	{0xACEC7A2F2F401A26ULL, "vxlt_0_lowtown_a2_n_fhz_0_0.vlm"},
	{0xACEC97DA95CE246FULL, "578982161.ogg"},
	{0xACEC9A4B8F805AD4ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.cess.evt"},
	{0xACECFFFC40B0A956ULL, "vxdr_0_deeproads_d0t_0_0.vlm"},
	{0xACED2AE6118C58BCULL, "zz_followers_banter6158298.cl"},
	{0xACED7D5C83918E28ULL, "zz_mag_debug6051370.cl"},
	{0xACEDA4B393D6F5D5ULL, "6060494_m.fxe"},
	{0xACEDE6D05B18652EULL, "w_encounter1_38_0.mmh"},
	{0xACEDF2D05B1879C5ULL, "w_encounter1_38_0.mao"},
	{0xACEE00D05B18911CULL, "w_encounter1_38_0.msh"},
	{0xACEE10D523295657ULL, "isa211_isabela6072560.cl"},
	{0xACEE234A390CF2CFULL, "rob_mer.tnt"},
	{0xACEEA7ADEEB9DF49ULL, "zz_followers_banter6158907.cl"},
	{0xACEF057257D90B9FULL, "vxw_bonepitmining_a2_28m_0_0.vlm"},
	{0xACEF19450C271059ULL, "mag311_keran6041699.cl"},
	{0xACEF2E89EB6B21EAULL, "zz_vrc_debug6200204.cl"},
	{0xACEFE705C200FF35ULL, "vxlt_darktown_586_0_0.vlm"},
	{0xACEFF113A93C8285ULL, "framebuffereffects_cocdownsample_p0.psh.fxo"},
	{0xACF01D7F6337BD49ULL, "vxht_0_hightown_a2_bzd_0_0.vlm"},
	{0xACF0347C6C97CA2FULL, "vxw_encounter3_3m_0_0.vlm"},
	{0xACF07EAF77F1BF5BULL, "369600117.ogg"},
	{0xACF0BB3CB3BC5BF6ULL, "225986204.ogg"},
	{0xACF0C7A2D7998089ULL, "6079024_f.fxe"},
	{0xACF123AAB7EEAC2BULL, "isa211_isabela6069425.cl"},
	{0xACF1832CDAB41943ULL, "art\\vfx\\combat\\vfx_projectile_p.mmh"},
	{0xACF1F95D4C0CCA16ULL, "w_encounter1_28_0.mao"},
	{0xACF2055D4C0CDEB5ULL, "w_encounter1_28_0.mmh"},
	{0xACF2075D4C0CE22BULL, "w_encounter1_28_0.msh"},
	{0xACF2179E73F34776ULL, "lt_0_lowtown_docks_a3_mmi.gff"},
	{0xACF23035452021C9ULL, "136820904.ogg"},
	{0xACF27533F3ABDE4CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ff_basic_1.mao"},
	{0xACF28288145FC84FULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm06.phy"},
	{0xACF29C2E6DC59C0FULL, "vxht_mansion_n_34r_0_0.vlm"},
	{0xACF2B09DEA387E81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ext.ani"},
	{0xACF2EBAF34A161A7ULL, "456305329.ogg"},
	{0xACF2EE10DF9F25FAULL, "6174464_m.fxe"},
	{0xACF34A8274293112ULL, "gen_im_arm_glv_med_act2_02.uti"},
	{0xACF450DEC13C9D3BULL, "l_lgt211ar_gang_hideout_lt.bnk"},
	{0xACF45E1C86AFDCB2ULL, "6125844_m.fxe"},
	{0xACF4C10BE72C966AULL, "926703713.ogg"},
	{0xACF52757F6F0E25BULL, "pro000st_varric.stg"},
	{0xACF591B3726293F2ULL, "llac_dr_0_deeproads_5843_2.dds"},
	{0xACF59BEE0652A871ULL, "plt_mrl200pt_n_postquest.nss"},
	{0xACF5B27AD3307D35ULL, "art\\vfx\\models\\vfx_isa_pose7.mmh"},
	{0xACF5B47AD33080ABULL, "art\\vfx\\models\\vfx_isa_pose7.msh"},
	{0xACF60161A1AC3C92ULL, "vxht_chantry_n_32q_0_0.vlm"},
	{0xACF6486459B38FD1ULL, "vxlt_undercitydungeon_17h_0_0.vlm"},
	{0xACF6A80D3696713BULL, "ach_soc_rival.dds"},
	{0xACF758FE0B9E2DBEULL, "6132876_f.fxe"},
	{0xACF7720E53397A3CULL, "6119265_m.fxe"},
	{0xACF807ADCD59AA22ULL, "lt_0_lowtown_docks_30_0.rml"},
	{0xACF81B70CDC940EDULL, "lgt110_urn_barker6132990.cl"},
	{0xACF852856931A3DCULL, "vxlt_shop_ga_0_0.vlm"},
	{0xACF8619ACE88CC17ULL, "303275852.ogg"},
	{0xACF86559EC040E9BULL, "hlac_lt_0_lowtown_docks_-1_0y.dds"},
	{0xACF8A1DA85A6A34EULL, "6016812_m.fxe"},
	{0xACF8BB9DEA3CFD23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ext.gad"},
	{0xACF93275C1FB26F2ULL, "vxlt_0_lowtown_docks_a3_n_czn_0_1.vlm"},
	{0xACF97FDC192ED3A7ULL, "character_character_p0.vsh.fxo"},
	{0xACF9924B8F8AD70DULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.cess.ani"},
	{0xACF9E4A2B35A192FULL, "178511482.ogg"},
	{0xACFA79AE93402DADULL, "95883185.ogg"},
	{0xACFAA55947D7023CULL, "6067638_m.fxe"},
	{0xACFAE356E7F47B39ULL, "static_lava_p0.psh.fxo.dx11"},
	{0xACFAF39198E3C01AULL, "vxht_keep_67p_0_0.vlm"},
	{0xACFB091AFCFE1C5EULL, "crt_spblm_cld.mao"},
	{0xACFB51F56BE0D137ULL, "628956836.ogg"},
	{0xACFC454226DEF8E1ULL, "hlac_w_sundermount_4_0x.dds"},
	{0xACFC46E71C568024ULL, "847406144.ogg"},
	{0xACFC9AED7B2ED566ULL, "lgt351tr_firejet_trap.utt"},
	{0xACFD2F9AED0838CEULL, "vxlt_gallowscourtyard_61k_0_0.vlm"},
	{0xACFD9F1568DF3179ULL, "combat_st_nmv_default.gda"},
	{0xACFDC57CEBE13A94ULL, "hm_guard02.mor"},
	{0xACFDFBDE45AACBFCULL, "326667635.ogg"},
	{0xACFE25623C348F90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_loop.gad"},
	{0xACFE44AFDDEB89A9ULL, "1002216847.ogg"},
	{0xACFE918659C5C7DBULL, "vxht_bartrand_n_72w_0_0.vlm"},
	{0xACFE9ABE2F508622ULL, "hlac_lt_0_lowtown_131071_2x.dds"},
	{0xACFF0FD447CF7974ULL, "vxht_0_hightown_n_c8i_0_0.vlm"},
	{0xACFF5D66E6A44958ULL, "zz_vault_debug6203321.cl"},
	{0xACFF989DEA42E008ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_overlay_ext.evt"},
	{0xAD00498BE084BCCFULL, "938699142.ogg"},
	{0xAD00521EDC425D5AULL, "zz_vrc_debug6075358.cl"},
	{0xAD00B7954D7ADEBFULL, "628916942.ogg"},
	{0xAD00BD01A931E843ULL, "gen00fl_varric6081204.cl"},
	{0xAD00E48BAE8914FBULL, "6080494_m.fxe"},
	{0xAD019B402CF30BB5ULL, "gen00fl_dog6167022.cl"},
	{0xAD01C39C58602894ULL, "6126769_f.fxe"},
	{0xAD02569765ACB9C9ULL, "darktown_fem_elf_sta6174519.cl"},
	{0xAD0261C21B9A071FULL, "gen00fl_sebastian6177300.cl"},
	{0xAD02A17912E5DC6BULL, "6096804_m.fxe"},
	{0xAD02E9BCB152874FULL, "vxca_cavern1_haunted_iko_0_0.vlm"},
	{0xAD0304D9D7229BD5ULL, "sib111_mother6098810.cl"},
	{0xAD032393DC0C3048ULL, "6099188_f.fxe"},
	{0xAD039DB05A5CFAB4ULL, "500912276.ogg"},
	{0xAD03ACEF9F934156ULL, "dae000_maren.cnv"},
	{0xAD03AFB5FCDE769EULL, "zz_ser_debug6091926.cl"},
	{0xAD04202DEB7AC354ULL, "mcr371_orsino6083346.cl"},
	{0xAD042A8B6336DF66ULL, "gen00fl_isabela6101815.cl"},
	{0xAD044657D51150D0ULL, "t3_wdg_o01.tnt"},
	{0xAD046C01EF38F508ULL, "383988216.ogg"},
	{0xAD04F2272168ECB5ULL, "hla_ht_0_hightown_131072_0y.dds"},
	{0xAD05B4015F8882CAULL, "qun210cr_random_thug.utc"},
	{0xAD05C3DE16C58298ULL, "hlac_lt_0_lowtown_n_65535_1y.dds"},
	{0xAD05E6E1E3F9B719ULL, "561509205.ogg"},
	{0xAD06294FAC5901C5ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_ext_c.mmh"},
	{0xAD06AEEF9F957A36ULL, "dae000_maren.bnk"},
	{0xAD06F63D5F485B63ULL, "vxlt_0_lowtown_docks_a3_n_d14_0_0.vlm"},
	{0xAD07197B5CD0E724ULL, "vxlt_backalley_g_1x9_0_0.vlm"},
	{0xAD0740ECD59D3C27ULL, "vxlt_gallowstemplar_f_5cr_0_0.vlm"},
	{0xAD0779486E4C197EULL, "one300dg_solivitus6048306.cl"},
	{0xAD07F0A1C9380B40ULL, "one100it_warehouse_note.uti"},
	{0xAD083DBD674287A1ULL, "mcr371cr_city_guard.utc"},
	{0xAD0852D4DD20BDBDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgtb_0n.dds"},
	{0xAD09310EA1F34DF7ULL, "mag291_torpor6025990.cl"},
	{0xAD09656E8AD9D89AULL, "vxht_keep_60q_0_0.vlm"},
	{0xAD099229B5FD22F1ULL, "vxht_keep_4ms_0_0.vlm"},
	{0xAD09D662E14D7CA6ULL, "6171315_m.fxe"},
	{0xAD0A0D10440B4B8BULL, "374286311.ogg"},
	{0xAD0A21F1D43166BAULL, "dre000_hireling16038633.cl"},
	{0xAD0A5EF6D7EA6296ULL, "lowtown_male_urchin_sta6164237.cl"},
	{0xAD0A7325F436E603ULL, "mcr310_argument6074329.cl"},
	{0xAD0AC85BF0770AFDULL, "hlac_lt_gallowstemplar_f_7198_3z.dds"},
	{0xAD0AD3B511BB44B3ULL, "vxca_caverns1_ilc_0_0.vlm"},
	{0xAD0AEEB6536A1A7BULL, "follower_banter6161006.cl"},
	{0xAD0B793DC4BD97AFULL, "vxlt_0_lowtown_a2_epc_0_0.vlm"},
	{0xAD0B88610351703DULL, "vxlt_gallowscourtyard_6wf_0_0.vlm"},
	{0xAD0BA78639CB7280ULL, "gen00fl_fenris6126454.cl"},
	{0xAD0BE34953A99576ULL, "340523966.ogg"},
	{0xAD0BFBD3E7B930F6ULL, "373383340.ogg"},
	{0xAD0BFF21A49F31BDULL, "6125691_f.fxe"},
	{0xAD0C0987315B264AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comc_0t.dds"},
	{0xAD0C1DF3A0B4ED2BULL, "163034659.ogg"},
	{0xAD0C879A3E4205E4ULL, "art\\vfx\\combat\\vfx_r_arc_front_max5_c.mmh"},
	{0xAD0CF08F4AA2B494ULL, "270426065.ogg"},
	{0xAD0D643D0D3C603EULL, "hlac_lt_0_lowtown_docks_n_-131074_2x.dds"},
	{0xAD0D72C9E1B22620ULL, "lgt360_mag_orlanna6155319.cl"},
	{0xAD0DAAD5B8C71702ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_stepwall.msh"},
	{0xAD0DACD5B8C71AB8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_stepwall.mmh"},
	{0xAD0DB8D5B8C72F5BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_stepwall.mao"},
	{0xAD0DF9A29EA31E55ULL, "6116760_m.fxe"},
	{0xAD0E3CC62889805EULL, "llac_lt_0_lowtown_a3_n_c_65536_8.dds"},
	{0xAD0EA20FD73D1AF7ULL, "6120954_m.fxe"},
	{0xAD0F02973EE5B17FULL, "pro000_refugee_female6052086.cl"},
	{0xAD0F149F1D411DC8ULL, "hlac_dr_dungeon_3216_11x.dds"},
	{0xAD0FA3CB856F4895ULL, "llac_dr_0_deeproads_5842_13.dds"},
	{0xAD101F0F1B9B182AULL, "gen_im_arm_boo_lgt_act1_01.uti"},
	{0xAD10F98A2FBEE246ULL, "151352946.ogg"},
	{0xAD11BC375369DF5CULL, "6116044_m.fxe"},
	{0xAD12CC0F9DF21F7AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wi_galowint.msh"},
	{0xAD12DA0F9DF236B3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wi_galowint.mao"},
	{0xAD12DE0F9DF23DC0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0wi_galowint.mmh"},
	{0xAD130A73FC4F2346ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.dura.ani"},
	{0xAD139DA8FD7A463CULL, "579655571.ogg"},
	{0xAD139E995B0F87D6ULL, "hlac_ht_brothel_n_3381_1y.dds"},
	{0xAD13F7B842224B74ULL, "504552849.ogg"},
	{0xAD141E31213AA2F9ULL, "face2_nolit.psh.fxo"},
	{0xAD14264FAC653992ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_ext_c.anb"},
	{0xAD15427BA5C941DAULL, "60991249.ogg"},
	{0xAD160A7097E0F5F9ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_encn_0n.dds"},
	{0xAD160AC5C754C854ULL, "qun110_petrice6018743.cl"},
	{0xAD16927A7CADA6B9ULL, "dae000_oldmangossip6174916.cl"},
	{0xAD16C05C05073AB4ULL, "art\\vfx\\combat\\vfx_v_nature_act_c.mmh"},
	{0xAD16E3B75E78289FULL, "6088201_m.fxe"},
	{0xAD170EE554948B1AULL, "vxdr_0_deeproads_cpe_0_2.vlm"},
	{0xAD1771B7F9D2C046ULL, "6174288_m.fxe"},
	{0xAD17A3E7FFF98C74ULL, "pro000_opening_pc6142522.cl"},
	{0xAD17B069BDCE5B63ULL, "hlac_ht_keep_3740_0z.dds"},
	{0xAD17B3D5B8CF7B43ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_stepwall.phy"},
	{0xAD17CA424867A5EAULL, "llac_w_encounter3_0_3.dds"},
	{0xAD17E2A2F220629AULL, "fx_dis_twirl.mao"},
	{0xAD1832E994D2F827ULL, "zz_ave_debug6062242.cl"},
	{0xAD185DA1B46BBA7EULL, "6074416_m.fxe"},
	{0xAD18667FE6C5B499ULL, "20100998.ogg"},
	{0xAD186F22DA5EDC49ULL, "vrc321tr_spooky_scenes_5.utt"},
	{0xAD188BA40EEBC688ULL, "468333074.ogg"},
	{0xAD19295A880224B4ULL, "llac_ca_caverns2_5802_21.dds"},
	{0xAD194A75D12F0BAEULL, "gen00fl_anders6096616.cl"},
	{0xAD19C606F39C0A89ULL, "lowtown_male_alienage1_sta6162728.cl"},
	{0xAD1A4F0EBAC39846ULL, "6116727_m.fxe"},
	{0xAD1A92EC5E7C5DACULL, "vxlt_blackemporium_v1_0_0.vlm"},
	{0xAD1A9877FC2C936FULL, "load_town.gfx"},
	{0xAD1ABC9510ABA501ULL, "vxlt_shop_g0_0_0.vlm"},
	{0xAD1AD395FF77375AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_revival.gad"},
	{0xAD1AE1C45DFD7EAAULL, "dae000_flemeth6001344.cl"},
	{0xAD1AE8BB824990D3ULL, "6139387_m.fxe"},
	{0xAD1B3B531D294BB7ULL, "86683848.ogg"},
	{0xAD1C40B3C75DE961ULL, "6048306_m.fxe"},
	{0xAD1C6435351FF5E0ULL, "hlac_lt_0_lowtown_-1_1x.dds"},
	{0xAD1C74B483863D05ULL, "6060195_m.fxe"},
	{0xAD1C878A185D3D09ULL, "440330564.ogg"},
	{0xAD1C91BCE63F7117ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_enter.gad"},
	{0xAD1CEB7C62AA8DF0ULL, "vxht_0_hightown_a2_ajv_0_0.vlm"},
	{0xAD1D02F0B9054F80ULL, "vxht_mansion2_2jm_0_0.vlm"},
	{0xAD1D12836AE129B0ULL, "lt_0_lowtown_n_52_0.rml"},
	{0xAD1D41F3B4D91D79ULL, "395270449.ogg"},
	{0xAD1D4738D2B07A32ULL, "6132646_m.fxe"},
	{0xAD1DD8365E17BB08ULL, "hlac_w_sundermount_0_3x.dds"},
	{0xAD1DF6E7D6BD81CFULL, "and100_anders6173578.cl"},
	{0xAD1E127BB6CAC9E2ULL, "vxlt_0_lowtown_a2_gvf_0_0.vlm"},
	{0xAD1E4D9A8045A33DULL, "vxht_mansion2_3d7_0_0.vlm"},
	{0xAD1E6D9D454E12FAULL, "dae000ip_crowd_mf6_drunk2.utp"},
	{0xAD1F13BE4F55D68FULL, "vxca_caverns1_i76_0_0.vlm"},
	{0xAD1F58CE3090D4D4ULL, "6060121_m.fxe"},
	{0xAD1F7D554A54B505ULL, "6118339_m.fxe"},
	{0xAD1F9EB95FE6E1C8ULL, "6097890_f.fxe"},
	{0xAD1FBCE9F6156E58ULL, "vxlt_gallowstemplar_f_4xy_0_0.vlm"},
	{0xAD20A419F2C78ABEULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_1z.dds"},
	{0xAD20B46A12BA115FULL, "453135743.ogg"},
	{0xAD20EC8F02EBD706ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_in_c.anb"},
	{0xAD20F273FC5B37B7ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.dura.evt"},
	{0xAD21173DE6065F1FULL, "37340859.ogg"},
	{0xAD2126FE194D4C07ULL, "mag170_hugh6206276.cl"},
	{0xAD2173D703D9436AULL, "vxlt_undercityserialkiller_1jd_0_1.vlm"},
	{0xAD21D695FF7D5B71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_revival.evt"},
	{0xAD21DB8CEDA8FFA2ULL, "ht_0_hightownexplosion_n_44_0_ll.mmh"},
	{0xAD221EAA7F10D258ULL, "6042510_m.fxe"},
	{0xAD222EC865E04992ULL, "vxca_cavern1_haunted_j2n_0_0.vlm"},
	{0xAD22337AF68D43D6ULL, "vxca_caverns1_mines_hm9_0_0.vlm"},
	{0xAD2264595187EAE8ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb101a_0s.dds"},
	{0xAD22EA9B1B7E124BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic.phy"},
	{0xAD22EFA4F817BACDULL, "685888648.ogg"},
	{0xAD237627C711EB7EULL, "vxlt_darktown_501_0_0.vlm"},
	{0xAD2379475F85973FULL, "356437717.ogg"},
	{0xAD23A6BCE645B3A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_enter.evt"},
	{0xAD2417926DA05E1FULL, "6100758_m.fxe"},
	{0xAD242D5FC79E3801ULL, "vxht_mansion2_29i_0_0.vlm"},
	{0xAD24335A39F3D8E1ULL, "hlac_lt_darktown_-1_3z.dds"},
	{0xAD244EB0B1E6D75DULL, "vxht_keep_n_67h_0_0.vlm"},
	{0xAD249F584FE8861EULL, "hla_ca_cave_2194_0z.dds"},
	{0xAD257CD73BE2B145ULL, "245363303.ogg"},
	{0xAD25847EA9DB25F3ULL, "dae100_gamlen_letter.bnk"},
	{0xAD2602F326B51278ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basic.phy"},
	{0xAD26341A167C862EULL, "hlac_lt_0_lowtown_a2_n_c_0_0x.dds"},
	{0xAD271720C5BDA6E2ULL, "hlac_w_encounter2_camp_393214_0z.dds"},
	{0xAD2731B4C7DDF49DULL, "vxlt_darktown_6c4_0_0.vlm"},
	{0xAD27712512F0A881ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_loop.ani"},
	{0xAD27C6B18481252EULL, "pro000cr_guard_amb.utc"},
	{0xAD27CCC7526C148AULL, "333458349.ogg"},
	{0xAD27F373FC615840ULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.dura.gad"},
	{0xAD284B127561BA4AULL, "vxca_caverns1_mines_j4p_0_0.vlm"},
	{0xAD287D8C6CDAB65EULL, "vxlt_darktown_65i_0_0.vlm"},
	{0xAD291A1D53873EF9ULL, "6014673_m.fxe"},
	{0xAD293B270BB05F47ULL, "6124738_m.fxe"},
	{0xAD29792F396C9440ULL, "vxht_0_hightown_aka_0_0.vlm"},
	{0xAD2994720EC4C64CULL, "art\\cinematics\\act1\\cs_fp_scream.ani"},
	{0xAD29A67EA9DF4843ULL, "dae100_gamlen_letter.cnv"},
	{0xAD29F775367C20A5ULL, "sky_lt_gamlenshouse_n.mao"},
	{0xAD2A9E0FF83EBC33ULL, "136037692.ogg"},
	{0xAD2AA53FA177AA8FULL, "zz_isa_debug6133237.cl"},
	{0xAD2AB6873E204111ULL, "mag210_arianni6024116.cl"},
	{0xAD2B158EBFF5B02DULL, "vxlt_darktown_5r0_0_0.vlm"},
	{0xAD2B380E738426EAULL, "6060138_m.fxe"},
	{0xAD2B6F72D2E74B10ULL, "one300_zevran6089702.cl"},
	{0xAD2B9B4A2D8AC951ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnip.mmh"},
	{0xAD2BA74A2D8ADDF2ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnip.mao"},
	{0xAD2BB238A380640EULL, "vxht_0_hightown_bwt_0_0.vlm"},
	{0xAD2BB462F1624416ULL, "hlac_w_sundermount_196612_0x.dds"},
	{0xAD2BB54A2D8AF50FULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnip.msh"},
	{0xAD2BCA1399029FD7ULL, "vxht_keep_n_60o_0_0.vlm"},
	{0xAD2C6E968C9CE2C1ULL, "198320659.ogg"},
	{0xAD2C8458A8C0E262ULL, "vxlt_0_lowtown_a2_gmg_0_0.vlm"},
	{0xAD2D4912748025A5ULL, "llac_lt_0_lowtown_docks_-131074_8.dds"},
	{0xAD2D5EC29FF7D1B1ULL, "gen00fl_isabela_init6076102.cl"},
	{0xAD2D7C2512F52723ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_loop.gad"},
	{0xAD2D824D2CFDFB23ULL, "dae000_flemeth6000829.cl"},
	{0xAD2D85792535F50BULL, "gen00fl_anders6096497.cl"},
	{0xAD2D887815079012ULL, "118940292.ogg"},
	{0xAD2D99F0DFF89E6BULL, "gen_im_npc_arm_hum_glv_tmpa.uti"},
	{0xAD2DF4A49A1E7BE4ULL, "6036467_m.fxe"},
	{0xAD2E247478D12E91ULL, "333170567.ogg"},
	{0xAD2E89276D401149ULL, "587859540.ogg"},
	{0xAD2EEE95FF880E40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_revival.ani"},
	{0xAD2EEF8F02F819B1ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_in_c.mmh"},
	{0xAD2F7ACFE295A8AEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnl90runf.evt"},
	{0xAD2F98CD278A0B91ULL, "hlac_ca_caverns1_mines_24422_7x.dds"},
	{0xAD2FD7F4E8CF94C2ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar4_.gad"},
	{0xAD301D0C465EAA07ULL, "6108633_m.fxe"},
	{0xAD302B4BBDA01A7EULL, "vxw_woundedcoast_1ft_0_0.vlm"},
	{0xAD30A1720ECAFA8EULL, "art\\cinematics\\act1\\cs_fp_scream.gad"},
	{0xAD30BEBCE650669DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_enter.ani"},
	{0xAD30C7585BD443EBULL, "gen00fl_varric6076589.cl"},
	{0xAD30F36EC8E26544ULL, "static1.psh.fxo"},
	{0xAD3104ADCD896A1EULL, "lt_0_lowtown_docks_30_0.anb"},
	{0xAD313B8FE472DA4DULL, "dae000_bodhan6146847.cl"},
	{0xAD32B2EE06867D5AULL, "plt_mrl200pt_n_postquest.plo"},
	{0xAD32C08DA3DCEFDDULL, "vxht_keep_n_62g_0_0.vlm"},
	{0xAD32C9AF51214176ULL, "dae000_gen_player_turn_in6177831.cl"},
	{0xAD339AC66233AFE7ULL, "6036763_m.fxe"},
	{0xAD34592512FB0A08ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\pose\\follower_ef_loop.evt"},
	{0xAD3488916EDB2CB9ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_roba_0t.dds"},
	{0xAD358A9A3E650FDBULL, "art\\vfx\\combat\\vfx_r_arc_front_max5_c.anb"},
	{0xAD363BAFDF1A1D77ULL, "6074365_m.fxe"},
	{0xAD36BCEF6757D2CAULL, "vxca_caverns1_mines_j3q_0_0.vlm"},
	{0xAD36EF1C21863907ULL, "hlac_w_sundermount_-262146_0z.dds"},
	{0xAD36EF695ED0CEDEULL, "vxlt_darktown_60j_0_0.vlm"},
	{0xAD3899DE93516B7AULL, "fx_smoke_vol.dds"},
	{0xAD38AC925E6F2431ULL, "947475115.ogg"},
	{0xAD38AE82088119EBULL, "6086971_m.fxe"},
	{0xAD391AFF56D7EE8EULL, "art\\vfx\\combat\\vfx_c_pride_pois_aoe_p.mmh"},
	{0xAD393FA65E1CFC8FULL, "hlac_ca_caverns1_24424_8z.dds"},
	{0xAD396871D7F52D1CULL, "6753458.ogg"},
	{0xAD3969F35EE33C77ULL, "hlac_lt_0_lowtown_docks_n_65534_1z.dds"},
	{0xAD3A0441421C32FBULL, "993387792.ogg"},
	{0xAD3A1F2DC505EC34ULL, "vxlt_0_lowtown_n_gqv_0_0.vlm"},
	{0xAD3A241595767C8EULL, "vxht_0_hightown_bpu_0_0.vlm"},
	{0xAD3A291E6EB909F7ULL, "gen00fl_merrill6151093.cl"},
	{0xAD3A628938323443ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw6sm.msh"},
	{0xAD3A64DB7CCF2FE7ULL, "gen00fl_merrill6122367.cl"},
	{0xAD3A708938324B9DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw6sm.mmh"},
	{0xAD3A7489383252AEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw6sm.mao"},
	{0xAD3AA49D5217FFC1ULL, "6074433_m.fxe"},
	{0xAD3ABBD37E51FA2CULL, "gencr_templar.utc"},
	{0xAD3AFD81D860697EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxb_1.mao"},
	{0xAD3B3EDEC1788139ULL, "l_lgt211ar_gang_hideout_lt.wbd"},
	{0xAD3B78ECAD1B957FULL, "vxca_caverns1_i6u_0_0.vlm"},
	{0xAD3BAAA318A9BE55ULL, "227041319.ogg"},
	{0xAD3C0D28F2282139ULL, "llac_lt_undercityserialkiller_-1_47.dds"},
	{0xAD3C2FED85FC7ADDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodfull.mao"},
	{0xAD3C33ED85FC8116ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodfull.mmh"},
	{0xAD3C3DED85FC9214ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodfull.msh"},
	{0xAD3C610F2E40E605ULL, "6112139_m.fxe"},
	{0xAD3CB28C2C48B76AULL, "6117294_m.fxe"},
	{0xAD3D25208139BF84ULL, "6169613_m.fxe"},
	{0xAD3D28348A0E3AAAULL, "art\\vfx\\combat\\vfx_r_asn_as_arch_act_c.anb"},
	{0xAD3D45E61687805AULL, "vxw_encounter2_camp_91_0_0.vlm"},
	{0xAD3DC717DEA9B877ULL, "6131167_m.fxe"},
	{0xAD3DF3E84D12EFFAULL, "723088628.ogg"},
	{0xAD3E00F0CC8BC645ULL, "6141694_m.fxe"},
	{0xAD3E262BF6B1F4C5ULL, "468942873.ogg"},
	{0xAD3E307752B8F58DULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_sml.mmh"},
	{0xAD3E427752B91413ULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_sml.msh"},
	{0xAD3E532F2494AEF5ULL, "lt_gallowscourtyard_58_0_lc.mmh"},
	{0xAD3F2B8375757EF1ULL, "441590411.ogg"},
	{0xAD3F47E8DEBA58F6ULL, "vxlt_0_lowtown_docks_n_9rb_0_0.vlm"},
	{0xAD3F8A2729F2BCAFULL, "vxht_0_hightown_n_b8g_0_0.vlm"},
	{0xAD3F9B69F2D1BC4FULL, "vxlt_gallowscourtyard_7cy_0_0.vlm"},
	{0xAD3FC35C052A456BULL, "art\\vfx\\combat\\vfx_v_nature_act_c.anb"},
	{0xAD3FE8E894DEED1AULL, "hlac_lt_gallowstemplar_n_2391_8x.dds"},
	{0xAD40251C271A3FBDULL, "6029651_m.fxe"},
	{0xAD403EC642F83F73ULL, "vxdr_primevalentrance_1kg_0_0.vlm"},
	{0xAD40785A67EA6FCEULL, "art\\vfx\\materials\\vfx_fire_torch_fb_add.mao"},
	{0xAD40DCFAD71FD8F5ULL, "fex_ruin_arch_lrg_f.mmh"},
	{0xAD40DE07E7821602ULL, "hlac_lt_0_lowtown_a2_-131073_1x.dds"},
	{0xAD40DEFAD71FDC6BULL, "fex_ruin_arch_lrg_f.msh"},
	{0xAD40E920AC5A4A57ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_mash_0s.dds"},
	{0xAD41125BCAE0D83AULL, "vxht_keep_4jm_0_0.vlm"},
	{0xAD412AB5A149BEC6ULL, "233233316.ogg"},
	{0xAD41E292AE12A916ULL, "166586565.ogg"},
	{0xAD42A66392EA5DF9ULL, "598031803.ogg"},
	{0xAD42FC08E661A82FULL, "hlac_lt_backalley_n_0_17x.dds"},
	{0xAD4402F4E8E086A8ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar4_.ani"},
	{0xAD442EFB180075AFULL, "150514054.ogg"},
	{0xAD447574631D3A9EULL, "391048609.ogg"},
	{0xAD44AD8CD11035F7ULL, "6079065_m.fxe"},
	{0xAD44AE7868DD7106ULL, "999443673.ogg"},
	{0xAD4552E46655061BULL, "986327194.ogg"},
	{0xAD45539CDE98D9DBULL, "t1_skn_003.tnt"},
	{0xAD458B198FA410F2ULL, "936061889.ogg"},
	{0xAD45B571B56DD5DFULL, "art\\characters\\weapons\\shields\\textures\\w_shd_seaa_0d.dds"},
	{0xAD45F7F088F9E2CEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_die_v1.evt"},
	{0xAD4643DD68AAE586ULL, "hlac_lt_0_lowtown_-1_0y.dds"},
	{0xAD466A7A728D3496ULL, "326440751.ogg"},
	{0xAD469B6341AC61B4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauserunf.ani"},
	{0xAD470F22C7288B16ULL, "6151708_m.fxe"},
	{0xAD4739FF56E4608DULL, "art\\vfx\\combat\\vfx_c_pride_pois_aoe_p.anb"},
	{0xAD479660CD3BFB21ULL, "275782602.ogg"},
	{0xAD47BD8DD03EF1A3ULL, "vxw_bonepitdestroyed_a3_kj_0_0.vlm"},
	{0xAD47D2F930BA42E9ULL, "6176128_m.fxe"},
	{0xAD48069C3A5DF852ULL, "llac_w_sundermount_131077_0.dds"},
	{0xAD4913836B066618ULL, "lt_0_lowtown_n_52_0.anb"},
	{0xAD4949574AB5F62CULL, "gen00fl_sebastian6180462.cl"},
	{0xAD495BEBED600B17ULL, "mer340cr_angry_miner.utc"},
	{0xAD49F3E31EE82E8FULL, "6178354_f.fxe"},
	{0xAD4A077F8CA65186ULL, "vxw_woundedcoast_1df_0_0.vlm"},
	{0xAD4A2980CC671178ULL, "vxw_bonepit_27i_0_0.vlm"},
	{0xAD4A3D8802ED7945ULL, "806185292.ogg"},
	{0xAD4A456F4D558DD5ULL, "6074987_m.fxe"},
	{0xAD4A8DED7B70E03FULL, "lgt351tr_firejet_trap.ncs"},
	{0xAD4AA53ED808EE30ULL, "19829971.ogg"},
	{0xAD4AABF9D20B344BULL, "760714882.ogg"},
	{0xAD4AEE4D111C30E0ULL, "llac_lt_0_lowtown_a3_n_c_131070_5.dds"},
	{0xAD4B9C074410F20CULL, "hlac_dr_0_deeproads_5845_2z.dds"},
	{0xAD4BB7C3087D98DAULL, "vxw_encounter2_camp_42_0_0.vlm"},
	{0xAD4BC9AE9ED6F798ULL, "llac_ht_0_hightown_-131072_0.dds"},
	{0xAD4BF13A55FFEBBDULL, "gen00fl_fenris6108079.cl"},
	{0xAD4C4574B21598D2ULL, "vxw_encounter3_ga_0_1.vlm"},
	{0xAD4CE867DE6C2149ULL, "one110_tree_minder6140861.cl"},
	{0xAD4CEEF088FFF279ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_die_v1.gad"},
	{0xAD4D00C62A4ADB07ULL, "6106433_m.fxe"},
	{0xAD4D085339128F1EULL, "vxlt_gallowsprison_n_5wk_0_0.vlm"},
	{0xAD4D0F05A18C4B7EULL, "vxw_woundedcoast_1lv_0_0.vlm"},
	{0xAD4D2F087FC60A08ULL, "791500907.ogg"},
	{0xAD4D986341B27B16ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauserunf.gad"},
	{0xAD4DA965F5FE6CDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_loop.ani"},
	{0xAD4DE224C1C25175ULL, "6107692_m.fxe"},
	{0xAD4ED4D3F6C86B6FULL, "6173914_m.fxe"},
	{0xAD4F0A62CCADA313ULL, "vxca_caverns2prc_9yj_0_0.vlm"},
	{0xAD4F297752C75D4AULL, "art\\characters\\creatures\\rubblegolem\\c_rubblegolem_sml.phy"},
	{0xAD4F6B235878C2A4ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_aim.ani"},
	{0xAD4F8CCE7C38C400ULL, "vxca_cavern1_haunted_ih5_0_0.vlm"},
	{0xAD5048B8075F2F12ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_enco_0n.dds"},
	{0xAD50979971F54B6AULL, "one300_zevran6086510.cl"},
	{0xAD50CC13152AFE20ULL, "gen_im_acc_amu_act1_05.uti"},
	{0xAD5112B175C9EF60ULL, "art\\characters\\weapons\\longswords\\w_lsw_006a.phy"},
	{0xAD5177EF91637123ULL, "6034663_f.fxe"},
	{0xAD5184E86387355AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scale_n.dds"},
	{0xAD51C9D780B42AE0ULL, "pro000_opening_pc6001191.cl"},
	{0xAD51CFDE5382612AULL, "gen00fl_aveline6120792.cl"},
	{0xAD521053FBC0EAFFULL, "6084721_f.fxe"},
	{0xAD523CCD7F11D149ULL, "vxlt_0_lowtown_a2_heo_0_0.vlm"},
	{0xAD529B33A80C96A3ULL, "ht_0_hightown_a2_33_0_lc.mmh"},
	{0xAD52AC9DC8E6C3E5ULL, "vxlt_0_lowtown_f0n_0_0.vlm"},
	{0xAD52FFF089047AC7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_die_v1.ani"},
	{0xAD530BC5F1EA1BACULL, "w_bonepit_41_0.rml"},
	{0xAD531566A8608EB1ULL, "hlac_lt_undercityencounter_-1_27x.dds"},
	{0xAD53E6277D1398B8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medg_0t.dds"},
	{0xAD54559D1EDD6E8FULL, "6042096_m.fxe"},
	{0xAD547F929C7699C4ULL, "vxca_caverns1_mines_gkb_0_0.vlm"},
	{0xAD54B050980B895AULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_vfx_suca_0n.dds"},
	{0xAD54B36341B8C875ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauserunf.evt"},
	{0xAD54CDCD5EB18E64ULL, "dae000_loyal_mage6165234.cl"},
	{0xAD55190525C403E7ULL, "llac_lt_0_lowtown_a2_65533_0.dds"},
	{0xAD5550CBEF85F53CULL, "6088901_m.fxe"},
	{0xAD556823587D2906ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_aim.gad"},
	{0xAD55969EE25E320EULL, "6062792_m.fxe"},
	{0xAD55AC0EC715722CULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffa.msh"},
	{0xAD55B20EC7157C1EULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffa.mmh"},
	{0xAD55BE0EC71590B5ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_staffa.mao"},
	{0xAD56098E6F7205CFULL, "vxdr_primevaltomb_85t_0_0.vlm"},
	{0xAD567F2A165F12A7ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear2.gad"},
	{0xAD5688948ED04D22ULL, "art\\characters\\creatures\\mubari\\animation\\dazed.gad"},
	{0xAD5781C088FF8DB1ULL, "6081086_m.fxe"},
	{0xAD57A02DACE75FF8ULL, "6106391_m.fxe"},
	{0xAD57AF8412540726ULL, "nat380_sibling6103562.cl"},
	{0xAD586C0D41FE746EULL, "6086871_m.fxe"},
	{0xAD5901BB5DF37ED7ULL, "vxht_mansion_n_6n5_0_0.vlm"},
	{0xAD5945E0E74D21DBULL, "zz_mer_debug6017204.cl"},
	{0xAD59AFEA82716CAFULL, "one330cr_ggt_thug1.utc"},
	{0xAD59F4944CB49725ULL, "one100pt_loose_end_meer.ncs"},
	{0xAD5A1B390F29DD8AULL, "413909565.ogg"},
	{0xAD5ABEAF31585A44ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbottom.msh"},
	{0xAD5AC96F9095BADCULL, "426794934.ogg"},
	{0xAD5AD0AF315878ADULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbottom.mao"},
	{0xAD5AD4AF31587FE6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbottom.mmh"},
	{0xAD5B4D301C5A82B0ULL, "llac_lt_0_lowtown_docks_0_2.dds"},
	{0xAD5BB90FE1936BF7ULL, "6084516_f.fxe"},
	{0xAD5C12D7302577B1ULL, "2205158.ogg"},
	{0xAD5C632358833FE5ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_tal_aim.evt"},
	{0xAD5C86869163D6A7ULL, "63046222.ogg"},
	{0xAD5CE6B100BBF039ULL, "hlac_w_encounter3_0_6z.dds"},
	{0xAD5D46B0E8BE83EFULL, "6176614_m.fxe"},
	{0xAD5D4FBD60B129F9ULL, "vxht_chantry_n_28i_0_0.vlm"},
	{0xAD5DAB948ED6A719ULL, "art\\characters\\creatures\\mubari\\animation\\dazed.evt"},
	{0xAD5DCF3E9518B2DFULL, "vxlt_gallowscourtyard_4ut_0_0.vlm"},
	{0xAD5DEFD10A8A6759ULL, "w_bonepit_205_0.msh"},
	{0xAD5DF9D10A8A7847ULL, "w_bonepit_205_0.mmh"},
	{0xAD5DFDD10A8A7E94ULL, "w_bonepit_205_0.mao"},
	{0xAD5E660257C3B240ULL, "6054833_f.fxe"},
	{0xAD5E99033C87CE99ULL, "268798715.ogg"},
	{0xAD5F98AA1E76F6E0ULL, "vxw_sundermount_34z_0_0.vlm"},
	{0xAD601C85A6EDA9A1ULL, "591686161.ogg"},
	{0xAD603B5A1FB11303ULL, "ss_fenris2.cut"},
	{0xAD6068E269632E8EULL, "37787536.ogg"},
	{0xAD60AEAA7107E9C9ULL, "vxlt_0_lowtown_a2_hbn_0_0.vlm"},
	{0xAD60C901CDFE83DAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comc_0d.dds"},
	{0xAD611E7ABADCDC65ULL, "vxlt_0_lowtown_f5m_0_0.vlm"},
	{0xAD612D2CFB29AC80ULL, "169787077.ogg"},
	{0xAD61C265F60F4058ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_loop.gad"},
	{0xAD6257F5257CE0B2ULL, "zz_approval_debug6003599.cl"},
	{0xAD626AE9FD033A3CULL, "nat380_nathaniel6066713.cl"},
	{0xAD62935D14249C69ULL, "and271_lyrium_smuggler.bnk"},
	{0xAD62D797ACC977C2ULL, "character_character_p81.psh.fxo"},
	{0xAD62D931AF878C11ULL, "vxht_0_hightown_b8n_0_0.vlm"},
	{0xAD62E59BD180A329ULL, "hm_mer_ambush_lead.mor"},
	{0xAD633A4551D5955AULL, "381500977.ogg"},
	{0xAD635AC6E32E241FULL, "hlac_w_blightlands_0_4x.dds"},
	{0xAD637B39F703C3B3ULL, "6122462_m.fxe"},
	{0xAD63F269A1D84128ULL, "vxca_caverns2_43g_0_0.vlm"},
	{0xAD64277AF9D2439AULL, "art\\vfx\\combat\\vfx_w_bsk_haste_activ_c.anb"},
	{0xAD642E1FAC327793ULL, "350597369.ogg"},
	{0xAD643F819C946CE9ULL, "6199022_m.fxe"},
	{0xAD646BDA7906258EULL, "6166931_m.fxe"},
	{0xAD64CBAF3160C4B1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0rw21_railbottom.phy"},
	{0xAD651015D379D1EDULL, "gen00fl_isabela6096798.cl"},
	{0xAD656798CF1DD6C3ULL, "vxdr_0_deeproads_crt_0_0.vlm"},
	{0xAD65755D14269EA9ULL, "and271_lyrium_smuggler.cnv"},
	{0xAD65A61DCB979417ULL, "65122522.ogg"},
	{0xAD661485DBA90222ULL, "dae000ip_crowd_em1_talk2.utp"},
	{0xAD662B348A31449DULL, "art\\vfx\\combat\\vfx_r_asn_as_arch_act_c.mmh"},
	{0xAD667F700A759384ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.dura.ani"},
	{0xAD66A680EB44AF0EULL, "98096202.ogg"},
	{0xAD66EB3970928286ULL, "vxw_woundedcoast_1jh_0_0.vlm"},
	{0xAD6739CFEBECA98AULL, "hlac_lt_gamlenshouse_n_-1_4x.dds"},
	{0xAD679C69C36A0B59ULL, "mrl150_hunter6071694.cl"},
	{0xAD679FC47F1D766EULL, "319069033.ogg"},
	{0xAD67BED868EBCBAAULL, "llac_lt_0_lowtown_n_65536_9.dds"},
	{0xAD67FE2BAEB43554ULL, "727778035.ogg"},
	{0xAD685C220FC576B7ULL, "vxht_mansion_n_6t7_0_0.vlm"},
	{0xAD6866C0B19292ADULL, "134075625.ogg"},
	{0xAD6915A8F83F8427ULL, "hlac_lt_hangedman_-1_2x.dds"},
	{0xAD694435C6352C56ULL, "vxht_0_hightown_a3_c8y_0_0.vlm"},
	{0xAD694B71A772B354ULL, "vxht_0_hightown_n_aid_0_0.vlm"},
	{0xAD69635C08F36AD0ULL, "mag291_ntbanter6056827.cl"},
	{0xAD69A5CA40507009ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_exit.ani"},
	{0xAD69CC6F802BC921ULL, "vxlt_0_lowtown_cfi_0_0.vlm"},
	{0xAD69F2BF85787C7EULL, "vxw_woundedcoast_1rh_0_0.vlm"},
	{0xAD6A9241004B61E3ULL, "vxht_keep_63a_0_0.vlm"},
	{0xAD6A9B0BC4958280ULL, "hlac_ca_caverns2_5802_67x.dds"},
	{0xAD6AB3948EE13F08ULL, "art\\characters\\creatures\\mubari\\animation\\dazed.ani"},
	{0xAD6AE0EBF2053414ULL, "6102910_f.fxe"},
	{0xAD6B0C8CA27FA1F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fd_transition.evt"},
	{0xAD6B1822ED8B3499ULL, "vxht_mansion_n_3y2_0_1.vlm"},
	{0xAD6B4051AB3E83D8ULL, "281932720.ogg"},
	{0xAD6B4443075F7A01ULL, "llac_ht_0_hightown_n_1_2.dds"},
	{0xAD6B4F0DC050F437ULL, "67787866.ogg"},
	{0xAD6B6C2A16714EEDULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear2.ani"},
	{0xAD6B8E60B2C0D08CULL, "140066508.ogg"},
	{0xAD6C0262F96FB200ULL, "6075831_m.fxe"},
	{0xAD6C62712DC41436ULL, "6120128_m.fxe"},
	{0xAD6C643BFF3211E6ULL, "llac_ht_keep_4519_17.dds"},
	{0xAD6CB7769A397AA8ULL, "120771774.ogg"},
	{0xAD6D06A35BFCBF40ULL, "bld_pro_01.dds"},
	{0xAD6D7C700A7BADE6ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.dura.gad"},
	{0xAD6E2F3CCE217743ULL, "hlac_dr_dungeon_3215_2x.dds"},
	{0xAD6E40404C1A1B8BULL, "800790146.ogg"},
	{0xAD6E44FE2A836945ULL, "hlac_lt_0_lowtown_n_-2_6z.dds"},
	{0xAD6E4651D9BCBD23ULL, "qcr000_arishok6087459.cl"},
	{0xAD6EAF3453B7CE71ULL, "vxht_keep_40t_0_0.vlm"},
	{0xAD6EAF4C97F7D16FULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.mmh"},
	{0xAD6EF14E48E80C6BULL, "hlac_lt_blackemporium_1055_3z.dds"},
	{0xAD6F143C51233DB8ULL, "6124368_m.fxe"},
	{0xAD6F1A77FB0AD376ULL, "follower_banter6164002.cl"},
	{0xAD704021183524F3ULL, "544707042.ogg"},
	{0xAD704EEE2571D2E5ULL, "llac_ht_0_hightown_a2_65535_2.dds"},
	{0xAD70A6482E821669ULL, "zz_isa_debug6133204.cl"},
	{0xAD70C0CA4056BC8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_exit.gad"},
	{0xAD70C6CF4BE39407ULL, "lt_0_lowtown_a2_n_c_53_0.rml"},
	{0xAD70D398B000567EULL, "hlac_lt_0_lowtown_docks_a3_-65538_1z.dds"},
	{0xAD71038CA283FD81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fd_transition.gad"},
	{0xAD711BC2BC889836ULL, "lt_0_lowtown_docks_n_39_0.anb"},
	{0xAD7149749EBF9042ULL, "character_character_p70.psh.fxo"},
	{0xAD7192267E5C412DULL, "vxw_woundedcoast_14i_0_0.vlm"},
	{0xAD722D2BA48A55EEULL, "481443301.ogg"},
	{0xAD723FA8B687445DULL, "hlac_lt_warehouse_0_6z.dds"},
	{0xAD724D2CD246D330ULL, "vxlt_0_lowtown_a3_n_c_e88_0_0.vlm"},
	{0xAD726CB7239D3DAEULL, "6078809_m.fxe"},
	{0xAD731772D0B3C101ULL, "ss_denerim.dds"},
	{0xAD73184ACF6DBA0CULL, "hlac_ht_mansion2_818_48z.dds"},
	{0xAD7377700A8010C5ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.dura.evt"},
	{0xAD7377B6B9DE1835ULL, "hlac_ht_0_hightown_a3_2_1x.dds"},
	{0xAD73DC6780E5B08EULL, "487294604.ogg"},
	{0xAD73DFE875B27F87ULL, "vxlt_0_lowtown_a3_n_c_fga_0_0.vlm"},
	{0xAD73E3BFA5615CEFULL, "6039017_f.fxe"},
	{0xAD741F00D3D7EF44ULL, "isa311_isabela.cnv"},
	{0xAD744D605BF4523FULL, "805272030.ogg"},
	{0xAD749ECB46C5F89DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_meda_0n.dds"},
	{0xAD74BE2124B810E2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_workstool.msh"},
	{0xAD74C02124B81398ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_workstool.mmh"},
	{0xAD74C9170159C6E8ULL, "hlac_lt_0_lowtown_n_1_7x.dds"},
	{0xAD74CC2124B8283BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_workstool.mao"},
	{0xAD74EFA79DAE6B3EULL, "mcr371_hostile_anders6082132.cl"},
	{0xAD755237AFD4FADCULL, "vxht_mansion2_2a0_0_0.vlm"},
	{0xAD75572FF2768E9EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spike_sa.dds"},
	{0xAD7572C2618450E8ULL, "w_bonepitmining_a2_16_0_ll.mmh"},
	{0xAD7605A0AF151593ULL, "vxlt_gallowscourtyard_7f2_0_0.vlm"},
	{0xAD769EC0C8431260ULL, "llac_lt_gallowstemplar_n_2383_9.dds"},
	{0xAD771455CA42C6FCULL, "hlac_w_sundermount_0_4y.dds"},
	{0xAD7719F1A35FA5D5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basic.msh"},
	{0xAD771BF1A35FA94BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basic.mmh"},
	{0xAD7727F1A35FBDE8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3pp_basic.mao"},
	{0xAD77286D24BA826FULL, "hlac_lt_gallowstemplar_f_2394_5y.dds"},
	{0xAD772FB175EAEFF5ULL, "art\\characters\\weapons\\longswords\\w_lsw_006a.msh"},
	{0xAD7731B175EAF36BULL, "art\\characters\\weapons\\longswords\\w_lsw_006a.mmh"},
	{0xAD773DB175EB0708ULL, "art\\characters\\weapons\\longswords\\w_lsw_006a.mao"},
	{0xAD774D6769284394ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayout.phy"},
	{0xAD7754093B6BC294ULL, "vxlt_0_lowtown_a2_n_c_hcy_0_0.vlm"},
	{0xAD77669B097AE5F7ULL, "follower_banter6169868.cl"},
	{0xAD77BDCA405CD6F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_exit.evt"},
	{0xAD77BE0CF4DB7BA1ULL, "cod_dal_vallaslin.utp"},
	{0xAD77C7DAE817418BULL, "6160680_m.fxe"},
	{0xAD78048CA28A1E2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fd_transition.ani"},
	{0xAD784100D3DC11A4ULL, "isa311_isabela.bnk"},
	{0xAD784928DB5A6C51ULL, "vxlt_undercitydungeon_1bg_0_0.vlm"},
	{0xAD78611C650FED9CULL, "gen00fl_aveline6137691.cl"},
	{0xAD78FF452F8E1D92ULL, "mag320cr_templar.utc"},
	{0xAD7906415177B891ULL, "llac_ht_mansion2_818_16.dds"},
	{0xAD793AD37A78E10CULL, "847681411.ogg"},
	{0xAD793C542125D13BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masb_0n.dds"},
	{0xAD796A4595FF83FEULL, "217573816.ogg"},
	{0xAD79D7C60DA512D3ULL, "vxdr_primevalentrance_22w_0_0.vlm"},
	{0xAD7A0311DD7A291DULL, "412475251.ogg"},
	{0xAD7A8597086FC334ULL, "6104937_m.fxe"},
	{0xAD7AA0355CECC243ULL, "hlac_lt_0_lowtown_n_-65535_3y.dds"},
	{0xAD7BB12CDA46691CULL, "mcr371_hostile_fenris.cnv"},
	{0xAD7C657F73D28F71ULL, "vxw_encounter1_f6_0_0.vlm"},
	{0xAD7CAA7023179268ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml102a_0s.dds"},
	{0xAD7D4925BAF848D3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerbunch01_0_sa.dds"},
	{0xAD7D82F16C497F82ULL, "6124047_m.fxe"},
	{0xAD7E4BA0FBA838CEULL, "284093794.ogg"},
	{0xAD7E64204700C212ULL, "844376590.ogg"},
	{0xAD7E73065084C74FULL, "51629471.ogg"},
	{0xAD7EC72124C07423ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_workstool.phy"},
	{0xAD7ED32CDA48D96CULL, "mcr371_hostile_fenris.bnk"},
	{0xAD7F819EE36ECD97ULL, "ice_crust.dds"},
	{0xAD7FBB5190B5A8C2ULL, "character_character_p23.psh.fxo"},
	{0xAD7FC7F8E0402E74ULL, "157518995.ogg"},
	{0xAD7FF51945922DDFULL, "6026249_m.fxe"},
	{0xAD8030E3628E5D96ULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_1z.dds"},
	{0xAD8061E67BB84E73ULL, "596475238.ogg"},
	{0xAD8088BE26D00FB9ULL, "vxdr_primevaltomb_505_0_1.vlm"},
	{0xAD80A2FAB55636CDULL, "vxdr_0_deeproads_euy_0_0.vlm"},
	{0xAD80ACD29F96A250ULL, "187265218.ogg"},
	{0xAD80E5476A1B86CCULL, "vxlt_hangedman_137_0_0.vlm"},
	{0xAD80EFF327019D0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basic.msh"},
	{0xAD80FDF32701B540ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basic.mao"},
	{0xAD8101F32701BB93ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basic.mmh"},
	{0xAD812A6769305C91ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayout.msh"},
	{0xAD813867693074ACULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayout.mao"},
	{0xAD814256E8DBA242ULL, "640249122.ogg"},
	{0xAD8144676930894FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkwayout.mmh"},
	{0xAD81459DA962EDE3ULL, "vxlt_gallowstemplar_f_43w_0_0.vlm"},
	{0xAD818E39BF9D81E1ULL, "6017409_m.fxe"},
	{0xAD81BDF2D4B595FEULL, "gen00fl_aveline6121980.cl"},
	{0xAD824D826A70F291ULL, "594489282.ogg"},
	{0xAD82544D85CBB34AULL, "21686362.ogg"},
	{0xAD82D5D54E9E74AFULL, "6060904_m.fxe"},
	{0xAD82DF2B09B5FC89ULL, "577202583.ogg"},
	{0xAD82F9D730431B1BULL, "437737569.ogg"},
	{0xAD83027C3357AF9CULL, "vxht_0_hightown_a3_c58_0_0.vlm"},
	{0xAD8303AD4167ED3AULL, "dae000_gamlen6026225.cl"},
	{0xAD8319E748F4DF0BULL, "6154997_m.fxe"},
	{0xAD832E9D1806995AULL, "168455645.ogg"},
	{0xAD836F2D9628ECDEULL, "art\\vfx\\cinematic\\vfx_blood_dripping_c.impa.evt"},
	{0xAD83B1EEA4D04804ULL, "follower_banter6169717.cl"},
	{0xAD83C414A1CB135CULL, "vxht_mansion2_2j1_0_0.vlm"},
	{0xAD83CEF3547EB386ULL, "vxw_woundedcoast_1pj_0_0.vlm"},
	{0xAD843D434D87FAE9ULL, "hlac_ht_0_hightown_a2_-65536_3y.dds"},
	{0xAD846DAA2865C1BBULL, "vxlt_gallowstemplar_f_6t0_0_0.vlm"},
	{0xAD847E4008F93B00ULL, "hlac_dr_0_deeproads_17039_1y.dds"},
	{0xAD848045A6B9C03BULL, "vxdr_primevaltomb_5ao_0_2.vlm"},
	{0xAD84C8BE7CCAFC67ULL, "608347896.ogg"},
	{0xAD8515A35C116AFEULL, "bld_pro_01.mao"},
	{0xAD8532209976BBCFULL, "6016689_f.fxe"},
	{0xAD8551E426AE483DULL, "vxw_encounter3_f8_0_0.vlm"},
	{0xAD8552CA81D5700CULL, "vxca_caverns1_mines_inc_0_0.vlm"},
	{0xAD859E37A580AF7AULL, "art\\characters\\creatures\\ragedemon\\animation\\p.gad"},
	{0xAD85B20079381795ULL, "hlac_lt_0_lowtown_docks_a3_-1_0z.dds"},
	{0xAD865E5270E38422ULL, "265333897.ogg"},
	{0xAD866DC8770324B9ULL, "90764410.ogg"},
	{0xAD86B50A858B4371ULL, "dae200_bran6092471.cl"},
	{0xAD86EDC91F0A8E78ULL, "hlac_ca_caverns1_mines_24424_0z.dds"},
	{0xAD87A0F0BFEC76FDULL, "pro000_colonel_ewald6005404.cl"},
	{0xAD88452B3D37F370ULL, "vxdr_primevaltomb_761_0_0.vlm"},
	{0xAD888E8E319C76CEULL, "character_character_p128.psh.fxo"},
	{0xAD88DC36E67FCEEFULL, "6208070_m.fxe"},
	{0xAD899D1A94DCA575ULL, "art\\vfx\\textures\\vfx_wis_smo.dds"},
	{0xAD89AD951A0C8CCAULL, "gen_im_trs_dal_med_cht_r2.uti"},
	{0xAD8A25E1FD42ADB6ULL, "art\\characters\\weapons\\mauls\\w_mal_107a.phy"},
	{0xAD8A862D962F3209ULL, "art\\vfx\\cinematic\\vfx_blood_dripping_c.impa.gad"},
	{0xAD8ACA18B5A5F332ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_hv_dmg.gad"},
	{0xAD8AD75C65C8A7F1ULL, "vxw_encounter1_k5_0_0.vlm"},
	{0xAD8AE2D10AB13F7CULL, "w_bonepit_205_0.phy"},
	{0xAD8B0395BAA41580ULL, "185550451.ogg"},
	{0xAD8BBC6996F2DE39ULL, "vxlt_hangedman_1i8_0_0.vlm"},
	{0xAD8BC79E4C747E56ULL, "art\\characters\\playercharacter\\shared\\textures\\default_proxy.dds"},
	{0xAD8C24E6705F83A4ULL, "vxht_mansion2b_2za_0_0.vlm"},
	{0xAD8C29D1D2CFAB95ULL, "gen00fl_anders6128318.cl"},
	{0xAD8C62BA8DC47599ULL, "142569602.ogg"},
	{0xAD8D2A7AF9F54E4DULL, "art\\vfx\\combat\\vfx_w_bsk_haste_activ_c.mmh"},
	{0xAD8D2B9D55850372ULL, "623961902.ogg"},
	{0xAD8D3FDB29545529ULL, "6137466_m.fxe"},
	{0xAD8D570BA1745AC7ULL, "298276474.ogg"},
	{0xAD8D73ECA3E53CAEULL, "674665709.ogg"},
	{0xAD8D76033B6B8030ULL, "gen00fl_merrill6108483.cl"},
	{0xAD8D97265E100F6DULL, "art\\levels\\ztools\\default\\generic\\1wi_angle.mmh"},
	{0xAD8D9B265E1015BEULL, "art\\levels\\ztools\\default\\generic\\1wi_angle.mao"},
	{0xAD8DA434BA8C7355ULL, "vxw_woundedcoast_es_0_0.vlm"},
	{0xAD8DA9265E102DF3ULL, "art\\levels\\ztools\\default\\generic\\1wi_angle.msh"},
	{0xAD8DAD33990E28AAULL, "278323968.ogg"},
	{0xAD8DF47D68375359ULL, "255237180.ogg"},
	{0xAD8E14757DC2A700ULL, "805648753.ogg"},
	{0xAD8E356C4CF8F679ULL, "gen00fl_dog6167178.cl"},
	{0xAD8EB377FD7D0DFEULL, "abi_r_decoy.bnk"},
	{0xAD8EC88C7E852D9EULL, "615515025.ogg"},
	{0xAD8F10F4A2D10E68ULL, "vxlt_0_lowtown_docks_cq4_0_0.vlm"},
	{0xAD8F1740079F1C85ULL, "hlac_ca_caverns1_24424_3y.dds"},
	{0xAD8F970C6070D7A6ULL, "art\\vfx\\cinematic\\vfx_fire_trail_p.mmh"},
	{0xAD8FBDE77D6FF0B9ULL, "6091839_f.fxe"},
	{0xAD8FD794D0551677ULL, "pro000_meeran6050091.cl"},
	{0xAD8FE8F771EB2D9CULL, "2613897.ogg"},
	{0xAD903AF2507B5934ULL, "6137513_m.fxe"},
	{0xAD90B223C88D6034ULL, "6116659_m.fxe"},
	{0xAD91366F9F28DC07ULL, "vxlt_0_lowtown_e25_0_0.vlm"},
	{0xAD9136EAC9301BD6ULL, "758665232.ogg"},
	{0xAD91572D96350157ULL, "art\\vfx\\cinematic\\vfx_blood_dripping_c.impa.ani"},
	{0xAD916364B55E4108ULL, "13923082.ogg"},
	{0xAD916709CE4CB546ULL, "6116576_m.fxe"},
	{0xAD91CA58C0EFC9D2ULL, "vxdr_primevalentrance_29s_0_0.vlm"},
	{0xAD91CD18B5AC1769ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_hv_dmg.evt"},
	{0xAD925488AF9F4C3EULL, "vxlt_0_lowtown_a2_n_c_gjk_0_0.vlm"},
	{0xAD92AB6D21191554ULL, "qun211_braggart6075445.cl"},
	{0xAD92FA8A940E5D26ULL, "plt_cod_hst_kirk_alienage.plo"},
	{0xAD930B6F5A251A8DULL, "6096842_m.fxe"},
	{0xAD936FC243A8BB00ULL, "179988341.ogg"},
	{0xAD93C3C118A460BDULL, "vxw_encounter3_c7_0_0.vlm"},
	{0xAD940343D931B4F0ULL, "218404690.ogg"},
	{0xAD940EE1FD4ADB57ULL, "art\\characters\\weapons\\mauls\\w_mal_107a.msh"},
	{0xAD9424E1FD4B00C9ULL, "art\\characters\\weapons\\mauls\\w_mal_107a.mmh"},
	{0xAD9437ADF6414B27ULL, "hlac_ht_mansion2_818_54x.dds"},
	{0xAD94543A0B51BA06ULL, "hlac_lt_gallowscourtyard_n_-196608_0x.dds"},
	{0xAD94CF243FA839FCULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower2_mcr370_playerflees.ani"},
	{0xAD957E480E8F5F9EULL, "vxw_encounter2_camp_2p_0_0.vlm"},
	{0xAD9591D9E27F970EULL, "llac_dr_primevaltomb_658_7.dds"},
	{0xAD95B3508EA655F8ULL, "lgt220_mag_tmplr_templar6140556.cl"},
	{0xAD95F185EA8633FCULL, "mag101_grace6034045.cl"},
	{0xAD96749E65A08C1FULL, "gen_im_delete_me_09.uti"},
	{0xAD96DA60CE99D469ULL, "6028804_f.fxe"},
	{0xAD97B04C981AD7ACULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.anb"},
	{0xAD97BC47574EFC01ULL, "vxht_mansion2_2aa_0_0.vlm"},
	{0xAD9825E3F292F77DULL, "6086281_m.fxe"},
	{0xAD988DFD98F838CAULL, "vxht_mansion2_31k_0_0.vlm"},
	{0xAD9901F5E348BAAEULL, "6155667_m.fxe"},
	{0xAD9904C804EDF922ULL, "6076843_m.fxe"},
	{0xAD9949BB2411538DULL, "w_encounter1_19_0.mmh"},
	{0xAD994DBB24115ADEULL, "w_encounter1_19_0.mao"},
	{0xAD995BBB24117213ULL, "w_encounter1_19_0.msh"},
	{0xAD999D5A2700F2E3ULL, "ht_0_hightown_n_60_0.rml"},
	{0xAD999F4E0BD834A2ULL, "hlac_lt_0_lowtown_docks_a3_65534_4z.dds"},
	{0xAD99B937A5918660ULL, "art\\characters\\creatures\\ragedemon\\animation\\p.ani"},
	{0xAD99FCBBCC2497B8ULL, "gen00fl_aveline6053283.cl"},
	{0xAD9A1E2559B1D1F5ULL, "vxht_0_hightown_a2_c61_0_0.vlm"},
	{0xAD9A96C362559C24ULL, "vxht_mansion2b_2wf_0_0.vlm"},
	{0xAD9AA73306B7F25BULL, "art\\vfx\\combat\\vfx_g_dual_flurry_act_c.impa.evt"},
	{0xAD9AAFACCAC7A438ULL, "292249190.ogg"},
	{0xAD9AE713BDC52F10ULL, "ser270_emeric6111468.cl"},
	{0xAD9AE7B1648F6E8FULL, "vxlt_hangedman_n_1rp_0_0.vlm"},
	{0xAD9B07BB632240AAULL, "hlac_lt_undercityserialkiller_-1_2z.dds"},
	{0xAD9B0A1E0394BFD7ULL, "6061247_f.fxe"},
	{0xAD9B54ED1663D363ULL, "6118045_m.fxe"},
	{0xAD9B6022FC5DECBFULL, "618745022.ogg"},
	{0xAD9B7F2E8F2F558EULL, "llac_w_blightlands_-131069_0.dds"},
	{0xAD9BA7B8B101C99DULL, "gencr_bloodmage.utc"},
	{0xAD9BD84FD401B602ULL, "hlac_ht_chantry_n_2534_11z.dds"},
	{0xAD9BDC243FAE6E3EULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_follower2_mcr370_playerflees.gad"},
	{0xAD9BFC86AED3D844ULL, "ch1alphemmultnt.psh.fxo"},
	{0xAD9C1611AC828BD5ULL, "vxw_woundedcoast_br_0_0.vlm"},
	{0xAD9C454B9F2840EFULL, "584406109.ogg"},
	{0xAD9CBF77D00D8C11ULL, "254163619.ogg"},
	{0xAD9D0BDA721EAF8BULL, "411913356.ogg"},
	{0xAD9D13C40CD5D926ULL, "tbe100ip_shop_codex_9.utp"},
	{0xAD9D3E54AFCC40E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\flurry.ani"},
	{0xAD9D4B032D68C3F5ULL, "mer141_jansen6077793.cl"},
	{0xAD9DA6785D77DA47ULL, "art\\levels\\races\\design\\_textures\\plc_metaltrp010_0_d.dds"},
	{0xAD9E87597583EEAAULL, "hlac_ca_caverns2prc_5802_4z.dds"},
	{0xAD9E90265E1E762AULL, "art\\levels\\ztools\\default\\generic\\1wi_angle.phy"},
	{0xAD9EB518B5B678D8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_hv_dmg.ani"},
	{0xAD9F76EA470D4CFAULL, "6165577_f.fxe"},
	{0xAD9FDE11AC9EFFB5ULL, "t1_mue_gn3.tnt"},
	{0xADA014774B7D4B84ULL, "6118680_m.fxe"},
	{0xADA0402C38766264ULL, "zz_lgt_2_debug6174311.cl"},
	{0xADA0684E92CF20F6ULL, "vxht_0_hightown_cca_0_0.vlm"},
	{0xADA0B5DF6BFF204CULL, "vxlt_hangedman_2dm_0_0.vlm"},
	{0xADA181F5222C7730ULL, "boots_frozen_wastes.dds"},
	{0xADA1A03306BE049CULL, "art\\vfx\\combat\\vfx_g_dual_flurry_act_c.impa.gad"},
	{0xADA1B672D53FB7BDULL, "633346284.ogg"},
	{0xADA1CE3B16E594CFULL, "vxca_cavern1_haunted_i2p_0_0.vlm"},
	{0xADA1FECD07F8B3EDULL, "mer320cr_lady_hubert.utc"},
	{0xADA22395D79DD8B7ULL, "vxht_mansion_n_6h3_0_0.vlm"},
	{0xADA256A811B8D664ULL, "378300465.ogg"},
	{0xADA25E1C155F03A6ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_8x.dds"},
	{0xADA2B53E1AFB9FD3ULL, "6177921_m.fxe"},
	{0xADA2B9462C309CD4ULL, "vxdr_primevalentrance_24s_0_1.vlm"},
	{0xADA2BCB8E6398962ULL, "qun221_petrice6076807.cl"},
	{0xADA2C5E10A045E80ULL, "vxht_keep_5kq_0_0.vlm"},
	{0xADA33F1A341CCB7BULL, "vxw_woundedcoast_15p_0_0.vlm"},
	{0xADA358D436FDAD00ULL, "265766008.ogg"},
	{0xADA36BA7CD762C09ULL, "mnd_imp_dis.mao"},
	{0xADA3FEC2BCB3C0CAULL, "lt_0_lowtown_docks_n_39_0.rml"},
	{0xADA419FF592E402EULL, "isa211_isabela6202210.cl"},
	{0xADA43AED9AD6C383ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzscaffold.mao"},
	{0xADA43EED9AD6CAD0ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzscaffold.mmh"},
	{0xADA449BB241B5B06ULL, "w_encounter1_19_0.rml"},
	{0xADA44CED9AD6E2EAULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzscaffold.msh"},
	{0xADA4546D1532F475ULL, "617908516.ogg"},
	{0xADA46CF3FCB4914BULL, "6197585_m.fxe"},
	{0xADA4C4A0E575E54CULL, "hla_ht_0_hightown_a2_-131071_0z.dds"},
	{0xADA511AC2F6D787AULL, "mag110_arianni6016773.cl"},
	{0xADA596656F1E6EE7ULL, "vxlt_warehouse_n_11q_0_0.vlm"},
	{0xADA59DD7B4C182E0ULL, "1024592155.ogg"},
	{0xADA5B52E6FE05469ULL, "788940041.ogg"},
	{0xADA62A8E2224E0CDULL, "follower_banter6154192.cl"},
	{0xADA66F16BB79798CULL, "vxlt_0_lowtown_a2_n_e1l_0_0.vlm"},
	{0xADA69B7CF6759C11ULL, "6141948_m.fxe"},
	{0xADA6BAAC04434514ULL, "193251777.ogg"},
	{0xADA6C905CC221C64ULL, "character_character_p166.psh.fxo.dx11"},
	{0xADA768F85593581BULL, "938232329.ogg"},
	{0xADA7CBB1BF7C377CULL, "6132842_m.fxe"},
	{0xADA7EE0BB290EFD8ULL, "vxht_keep_67b_0_0.vlm"},
	{0xADA83D863D839D00ULL, "hlac_dr_slavers_3215_2y.dds"},
	{0xADA8981CC41A4302ULL, "mag121_viveka6082576.cl"},
	{0xADA89C9D7889DD0BULL, "mag370_meredith6185094.cl"},
	{0xADA89F3306C4219AULL, "art\\vfx\\combat\\vfx_g_dual_flurry_act_c.impa.ani"},
	{0xADA8B0EC9D9429E4ULL, "pro000_cut_innocent_blood6004071.cl"},
	{0xADA8B5F95D673881ULL, "203377047.ogg"},
	{0xADA8DD8D0F0E2A5CULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_lp.ani"},
	{0xADA908A0544BB4A4ULL, "vxht_mansion2b_2pg_0_0.vlm"},
	{0xADA95F2B74DE9A09ULL, "vxlt_0_lowtown_a2_n_eok_0_0.vlm"},
	{0xADA98D902840912AULL, "hla_ht_0_hightown_a2_65534_0y.dds"},
	{0xADA9F400460B4F59ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_abma_0n.dds"},
	{0xADAA42BB241FBB4AULL, "w_encounter1_19_0.phy"},
	{0xADAA9DEF8BAC232FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_zbm_0d.dds"},
	{0xADAAAFDD2E54A219ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_icon.mmh"},
	{0xADAAB9DD2E54B307ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_icon.msh"},
	{0xADAACBD9D8E5C993ULL, "hlac_ht_0_hightown_a3_131071_2x.dds"},
	{0xADAB2CD2D19591C4ULL, "vxdr_0_deeproads_cfr_0_1.vlm"},
	{0xADAB33BAA0371A86ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax102a_0d.dds"},
	{0xADAB84C47CCA9365ULL, "716067462.ogg"},
	{0xADAC3F6513A49D82ULL, "6131328_m.fxe"},
	{0xADAC48D54BC9E3DCULL, "932392462.ogg"},
	{0xADAC8689DD9BBB81ULL, "lowtown_fem_prostitute_sta6161603.cl"},
	{0xADACBAF41EAD5CC2ULL, "zz_followers_banter6159651.cl"},
	{0xADAD377711679FADULL, "625723518.ogg"},
	{0xADAE73AE74020CC7ULL, "53217793.ogg"},
	{0xADAE93DFDE300E3EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wc_sa.dds"},
	{0xADAEB344807A1D5DULL, "6074385_m.fxe"},
	{0xADAEC730B80C97F3ULL, "50438580.ogg"},
	{0xADAED1BE038E3872ULL, "vxht_mansion_n_6p2_0_0.vlm"},
	{0xADAED6CE9DFDCC4FULL, "vxw_bonepitdestroyed_a3_1cy_0_0.vlm"},
	{0xADAEDA2B84C53976ULL, "vxht_0_hightown_cfb_0_0.vlm"},
	{0xADAEEA8D0F12AB9EULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_lp.gad"},
	{0xADAF3BCA85C1D139ULL, "vxw_encounter2_camp_52_0_0.vlm"},
	{0xADAF752106244D5BULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.dura.ani"},
	{0xADAF9FF0475F4355ULL, "zz_mag_debug6054613.cl"},
	{0xADAFDEC4603B256DULL, "hlac_lt_0_lowtown_a2_n_c_65535_2y.dds"},
	{0xADB09BF56F7D72F9ULL, "75859089.ogg"},
	{0xADB0C9CF4C197D1BULL, "lt_0_lowtown_a2_n_c_53_0.anb"},
	{0xADB13C59DE4321DBULL, "6071142_m.fxe"},
	{0xADB147924B5803E6ULL, "t3_psp_b03.tnt"},
	{0xADB15554AFDD10CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\flurry.gad"},
	{0xADB173D1BA5810D9ULL, "vxht_bartrand_n_5vs_0_0.vlm"},
	{0xADB184C261412DD4ULL, "vxht_0_hightown_n_avi_0_0.vlm"},
	{0xADB18BF4B09861F9ULL, "vxlt_0_lowtown_a3_n_c_f55_0_0.vlm"},
	{0xADB1B0F72612E3FBULL, "vxw_woundedcoast_10s_0_0.vlm"},
	{0xADB1FC6F3F0A1B88ULL, "qun111_ketojan6059190.cl"},
	{0xADB20FB304240751ULL, "743891685.ogg"},
	{0xADB23E0B466F625AULL, "vrc321_housekeeper6085995.cl"},
	{0xADB2631A5C7D4823ULL, "vxht_bartrand_n_5z5_0_0.vlm"},
	{0xADB289A50C7574DBULL, "w_bonepitdestroyed_a3_218_0.msh"},
	{0xADB28E620CABC4B2ULL, "vxht_chantry_35w_0_0.vlm"},
	{0xADB29BA50C759326ULL, "w_bonepitdestroyed_a3_218_0.mao"},
	{0xADB2A7A50C75A7C5ULL, "w_bonepitdestroyed_a3_218_0.mmh"},
	{0xADB304DE3F51A72FULL, "766918284.ogg"},
	{0xADB314C1544C12CAULL, "vxlt_warehouse_qz_0_0.vlm"},
	{0xADB32B9BD4EEA1D4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudwindow.phy"},
	{0xADB3A6876A74FFFBULL, "6136175_m.fxe"},
	{0xADB3D8073DB5EBC4ULL, "static1_punch_ds2y.psh.fxo"},
	{0xADB3F6B5C17B849BULL, "w_blightlands_108_0.phy"},
	{0xADB426F8E07FB20EULL, "character_character_p126.psh.fxo.dx11.assembly"},
	{0xADB4734421D4A4B2ULL, "hlac_lt_0_lowtown_-131073_1z.dds"},
	{0xADB4919C6E4D07BFULL, "gen00fl_varric6118847.cl"},
	{0xADB4A0013B3B2D01ULL, "vxht_mansion2_2wg_0_0.vlm"},
	{0xADB55016103AC9A4ULL, "hla_ht_0_hightown_n_131073_0x.dds"},
	{0xADB55639E1585FDBULL, "vxlt_darktown_5p0_0_0.vlm"},
	{0xADB5C58D0F188BBDULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_lp.evt"},
	{0xADB601961207798EULL, "6169747_m.fxe"},
	{0xADB65FE8A4870858ULL, "vxht_keep_60c_0_0.vlm"},
	{0xADB6EBD8AE610B68ULL, "865603387.ogg"},
	{0xADB77A7D4641CD24ULL, "vxht_mansion2b_2md_0_0.vlm"},
	{0xADB871D358439C43ULL, "vxht_chantry_2sy_0_0.vlm"},
	{0xADB8941BE342EAB7ULL, "194455393.ogg"},
	{0xADB89B052018D949ULL, "191383459.ogg"},
	{0xADB8F9CB906EBCD5ULL, "vxw_woundedcoast_tp_0_0.vlm"},
	{0xADB9445AC33FEC7CULL, "vxlt_backalley_n_1pq_0_1.vlm"},
	{0xADB9832708EE9E07ULL, "220776465.ogg"},
	{0xADB9A7F09658C664ULL, "1070487931.ogg"},
	{0xADB9EB6E845EDC13ULL, "vxw_bonepitdestroyed_a3_1w1_0_0.vlm"},
	{0xADBA012D45D605BDULL, "300455035.ogg"},
	{0xADBA408233F9CEE2ULL, "vxlt_gallowstemplar_f_6st_0_0.vlm"},
	{0xADBA5CAECCF1E1ECULL, "rog_obscured.dds"},
	{0xADBA7B6FD1289BB4ULL, "vxca_caverns2prc_7nr_0_0.vlm"},
	{0xADBAE910D10C8CEAULL, "6068563_m.fxe"},
	{0xADBB5EB6196A638CULL, "6188609_m.fxe"},
	{0xADBBCB9A78D1FEBDULL, "vxlt_gallowscourtyard_5z2_0_0.vlm"},
	{0xADBC856792E02A1CULL, "plt_dae400pt_post_game.nss"},
	{0xADBCCBCDF4869737ULL, "490802495.ogg"},
	{0xADBD7D2106309852ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.dura.evt"},
	{0xADBE089BD4F86DD1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudwindow.msh"},
	{0xADBE169BD4F885ECULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudwindow.mao"},
	{0xADBE229BD4F8998FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrudwindow.mmh"},
	{0xADBE918D105C123CULL, "94128404.ogg"},
	{0xADBF261ECB052CB8ULL, "vxlt_darktown_5cb_0_0.vlm"},
	{0xADBF3959A2B37C34ULL, "379616644.ogg"},
	{0xADBF7A9320DF37FCULL, "vxht_0_hightown_a3_c72_0_0.vlm"},
	{0xADBFE5AEAC4E2959ULL, "vxht_bartrand_n_5yr_0_0.vlm"},
	{0xADBFF338B94555A2ULL, "plt_ser120pt_sacrifice_offe.nss"},
	{0xADC02AA1C05DCF2EULL, "ca_caverns2prc_5802.anb"},
	{0xADC079C0CF0E2B82ULL, "vxht_keep_n_611_0_0.vlm"},
	{0xADC112E98E06AC22ULL, "83207013.ogg"},
	{0xADC192794700CD66ULL, "vxht_0_hightown_7j2_0_0.vlm"},
	{0xADC1C5D535C114B6ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.impa.evt"},
	{0xADC1F79C18BAB73DULL, "190290578.ogg"},
	{0xADC24E8AE8FFBB13ULL, "vxht_keep_n_4h9_0_0.vlm"},
	{0xADC26CA7D22D3123ULL, "mcr310_cut_chantry_blows_up.bnk"},
	{0xADC2812C48678633ULL, "380950774.ogg"},
	{0xADC31BA1AC2E8BC7ULL, "vxdr_0_deeproads_cxn_0_0.vlm"},
	{0xADC31D9B869286F7ULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgc.phy"},
	{0xADC346AA3F695761ULL, "vxw_bonepitmining_a2_83_0_0.vlm"},
	{0xADC3489DEDA3D63CULL, "172860277.ogg"},
	{0xADC3A16F9D664423ULL, "dre000_varric6027759.cl"},
	{0xADC3D305C4025F52ULL, "isa311_castillon6092094.cl"},
	{0xADC4167EA43B5D04ULL, "399300264.ogg"},
	{0xADC49B0E1CC1A65FULL, "vxlt_0_lowtown_a2_fiw_0_0.vlm"},
	{0xADC49C210636EBE5ULL, "art\\vfx\\combat\\vfx_str_spit1_proj_p.dura.gad"},
	{0xADC59DC335244D0CULL, "6056739_m.fxe"},
	{0xADC5E8C51BD7C7A7ULL, "fillstamina.ncs"},
	{0xADC5EC5A3837E5A4ULL, "vxht_mansion2b_2fe_0_0.vlm"},
	{0xADC64EA7D230E753ULL, "mcr310_cut_chantry_blows_up.cnv"},
	{0xADC6759B8C87B6BEULL, "6202068_m.fxe"},
	{0xADC750994CD0AEBDULL, "vxlt_gallowscourtyard_66y_0_0.vlm"},
	{0xADC75B955397D490ULL, "qun221_petrice6076724.cl"},
	{0xADC75BBF4CDCE0FBULL, "art\\vfx\\creatures\\fxc_fade_ghost_c.dura.evt"},
	{0xADC7AED0273996C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palms_together.ani"},
	{0xADC7F0B8823ED7BFULL, "hlac_lt_undercityserialkiller_-1_19y.dds"},
	{0xADC811054909D587ULL, "vxlt_0_lowtown_hay_0_0.vlm"},
	{0xADC85D4B7654F493ULL, "vxw_bonepitdestroyed_a3_1x0_0_0.vlm"},
	{0xADC8BCAAC635B59EULL, "6127001_f.fxe"},
	{0xADC8BCD535C72441ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.impa.gad"},
	{0xADC8DC4B6F24B14DULL, "vxdr_0_deeproads_ejt_0_0.vlm"},
	{0xADC912C3B6DAEFD7ULL, "6134991_m.fxe"},
	{0xADC972B93F0298F6ULL, "6000921_m.fxe"},
	{0xADC9931BF50619A2ULL, "6054380_m.fxe"},
	{0xADC9AA3C2F08ADDFULL, "w_bonepitmining_a2_177_0.msh"},
	{0xADC9B03C2F08B781ULL, "w_bonepitmining_a2_177_0.mmh"},
	{0xADC9BC3C2F08CC22ULL, "w_bonepitmining_a2_177_0.mao"},
	{0xADC9BC58F7815884ULL, "509360964.ogg"},
	{0xADCA6C4A3B5A2853ULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.impa.ani"},
	{0xADCAF58C987B38C5ULL, "hlac_lt_undercityserialkiller_-1_47x.dds"},
	{0xADCB61ED30BC0980ULL, "552600169.ogg"},
	{0xADCB678620B40D5FULL, "6187768_m.fxe"},
	{0xADCB8C9AEC2D21E4ULL, "6969691.ogg"},
	{0xADCBD223209E7A35ULL, "6177303_m.fxe"},
	{0xADCC21581896C833ULL, "6125733_m.fxe"},
	{0xADCC2568952F931AULL, "6124245_m.fxe"},
	{0xADCCC3BBD1375271ULL, "6076667_f.fxe"},
	{0xADCCCF6000BE0348ULL, "vxw_bonepitmining_a2_86_0_0.vlm"},
	{0xADCD1F6AE0679D77ULL, "llac_w_encounter1_0_0.dds"},
	{0xADCD4E5F8BF4C67FULL, "623402749.ogg"},
	{0xADCDC700FF8D546FULL, "zz_dre_debug6037643.cl"},
	{0xADCDEE88316CEB9DULL, "zz_followers_banter6159399.cl"},
	{0xADCDFCA65E3BA563ULL, "hlac_lt_0_lowtown_n_-2_5y.dds"},
	{0xADCE54BF4CE2F33CULL, "art\\vfx\\creatures\\fxc_fade_ghost_c.dura.gad"},
	{0xADCE6288BD32878EULL, "6170129_m.fxe"},
	{0xADCEABD0273FB026ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palms_together.gad"},
	{0xADCECBF98F3D4F26ULL, "hlac_dr_slavers_3217_6y.dds"},
	{0xADCF43A2201491EEULL, "10663868.ogg"},
	{0xADCF8577FDB45604ULL, "abi_r_decoy.wbd"},
	{0xADCFB09C3CC5A359ULL, "115912741.ogg"},
	{0xADCFBDD535CD44EFULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.impa.ani"},
	{0xADCFEAAC88CB6A28ULL, "vxht_brothel_n_57m_0_0.vlm"},
	{0xADD02A00DC2B3BEEULL, "vxht_mansion_n_4ht_0_0.vlm"},
	{0xADD07C2FF5090B09ULL, "vxca_caverns1_mines_i9s_0_0.vlm"},
	{0xADD08FF9C502EB54ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.dura.ani"},
	{0xADD096F0AFFC5B97ULL, "hlac_lt_0_lowtown_a3_n_c_65536_10x.dds"},
	{0xADD0B2D87A583A3CULL, "6031362_f.fxe"},
	{0xADD0CDF9338909B1ULL, "106420.ogg"},
	{0xADD18D7E9E24A447ULL, "vxdr_0_deeproads_cui_0_0.vlm"},
	{0xADD1D679ABF24BF1ULL, "vxlt_hangedman_27e_0_0.vlm"},
	{0xADD204CD6B026713ULL, "vxlt_0_lowtown_a3_n_c_eny_0_0.vlm"},
	{0xADD23F2BB5F99B95ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sidetable_d.dds"},
	{0xADD348B005CDDC29ULL, "lgt210_eau_senestra6144930.cl"},
	{0xADD359480FCA199EULL, "764103531.ogg"},
	{0xADD39877DFDDBC7DULL, "gen00fl_varric6075551.cl"},
	{0xADD4B70CF20B9478ULL, "849867356.ogg"},
	{0xADD4DBBAE6C4F60DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1b.mmh"},
	{0xADD4DFBAE6C4FD5EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1b.mao"},
	{0xADD4E99AB6C3EB39ULL, "vxca_caverns1_j20_0_0.vlm"},
	{0xADD4EDBAE6C51493ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1b.msh"},
	{0xADD532A293BDF55FULL, "mrl210_merrill6102118.cl"},
	{0xADD553BF4CE9103AULL, "art\\vfx\\creatures\\fxc_fade_ghost_c.dura.ani"},
	{0xADD5A2C7A2CFA20BULL, "vxdr_primevaltomb_537_0_0.vlm"},
	{0xADD5A6D02745C605ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palms_together.evt"},
	{0xADD6B2D8D85399D1ULL, "one300dg_vanard6063908.cl"},
	{0xADD6DE232251D681ULL, "dae000_flemeth6013298.cl"},
	{0xADD7161C44F93EAFULL, "sky_dr_primevalentrance.mao"},
	{0xADD7448C63D940A2ULL, "vxht_0_hightown_a3_7g0_0_0.vlm"},
	{0xADD7644A3B64A48AULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.impa.evt"},
	{0xADD765192B03C967ULL, "vxht_0_hightown_a2_9ut_0_0.vlm"},
	{0xADD78CF9C50904B6ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.dura.gad"},
	{0xADD7CFFB50563534ULL, "llac_w_encounter1_-131070_0.dds"},
	{0xADD7F06ABB4AA3B5ULL, "610943811.ogg"},
	{0xADD869A84622334EULL, "llac_ht_0_hightown_a3_65535_0.dds"},
	{0xADD89B63D6C37F34ULL, "static_static_p96.psh.fxo.dx11"},
	{0xADD8B8020C788901ULL, "vxca_cavern1_haunted_ih8_0_0.vlm"},
	{0xADD8BC1861BE6C17ULL, "6054183_m.fxe"},
	{0xADD8D5C50AE2D604ULL, "701945802.ogg"},
	{0xADD928ADC3796795ULL, "fex_ruin_wl_a.msh"},
	{0xADD92AADC3796B0BULL, "fex_ruin_wl_a.mmh"},
	{0xADD9374AB0012B08ULL, "mag131cr_keran_templar.utc"},
	{0xADDA4D2943550A8AULL, "gen00fl_isabela6135607.cl"},
	{0xADDA7D841AB9180AULL, "6162409_m.fxe"},
	{0xADDB8A7791742388ULL, "823397956.ogg"},
	{0xADDBB5069218B7C3ULL, "6017418_m.fxe"},
	{0xADDBBF77529D3B7CULL, "art\\vfx\\textures\\vfx_lightning_arc_fb.dds"},
	{0xADDBF055279C1454ULL, "24782601.ogg"},
	{0xADDC7579F56ADADDULL, "663543848.ogg"},
	{0xADDDA7F9C50D9E15ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.dura.evt"},
	{0xADDE309B4187B648ULL, "and000pt_gift_amulet.ncs"},
	{0xADDE37FBDA85A656ULL, "dae000cr_cora.utc"},
	{0xADDE434A3B6A8BFDULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.impa.gad"},
	{0xADDE71E2A34A6F56ULL, "vxht_keep_4al_0_0.vlm"},
	{0xADDF7BBA0E5E1970ULL, "vxht_keep_n_670_0_1.vlm"},
	{0xADDF8C71469CAFE4ULL, "6125392_m.fxe"},
	{0xADDFA6FCD8A720CAULL, "dae100_alistair_drunk6135113.cl"},
	{0xADE076AA5CF87F93ULL, "vxlt_0_lowtown_a3_n_c_eqx_0_0.vlm"},
	{0xADE0A9C637580496ULL, "w_woundedcoast_235_0.phy"},
	{0xADE0DDB5C1A2483AULL, "w_blightlands_108_0.msh"},
	{0xADE0EBB5C1A26073ULL, "w_blightlands_108_0.mao"},
	{0xADE0EFB5C1A26680ULL, "w_blightlands_108_0.mmh"},
	{0xADE1A5E296C449D4ULL, "w_sundermount_31_0.rml"},
	{0xADE206CDB5BCE441ULL, "zz_lgt_1_debug6135729.cl"},
	{0xADE208494FE1E9DFULL, "478728038.ogg"},
	{0xADE23EB542BAB048ULL, "287203424.ogg"},
	{0xADE243FFF76D1537ULL, "6020491_f.fxe"},
	{0xADE2481A5E3A4387ULL, "follower_banter6161198.cl"},
	{0xADE24F2EA425F30BULL, "jagged_sa.dds"},
	{0xADE2FA1D22DE472EULL, "807978618.ogg"},
	{0xADE32819F96AE0A7ULL, "lt_0_lowtown_a3_n_c_43_0_ll.mmh"},
	{0xADE3C362CF0A77A7ULL, "77876824.ogg"},
	{0xADE3E13D1DAA6C9CULL, "6135906_m.fxe"},
	{0xADE445ED9B0CBAEBULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzscaffold.phy"},
	{0xADE5163861481AFCULL, "1022487295.ogg"},
	{0xADE524769D8F7185ULL, "gen_im_npc_elf_arm_glv_dald.uti"},
	{0xADE5371CEF3314BAULL, "mag101_alain6032279.cl"},
	{0xADE541055A412593ULL, "vxw_bonepitdestroyed_a3_1f2_0_0.vlm"},
	{0xADE5A268059219F6ULL, "309908313.ogg"},
	{0xADE5F166BE592604ULL, "gen00fl_fenris6126502.cl"},
	{0xADE650EE4A9AD20CULL, "6048286_m.fxe"},
	{0xADE6A368DCBE61B8ULL, "6101251_m.fxe"},
	{0xADE6DD09C5011147ULL, "mag100_karras6035321.cl"},
	{0xADE7876B9EC44C96ULL, "6105833_m.fxe"},
	{0xADE7B1D0067A3C69ULL, "190706196.ogg"},
	{0xADE7B506B6BF7753ULL, "304247762.ogg"},
	{0xADE7BCE95D5947E6ULL, "ser221_gamlen6135274.cl"},
	{0xADE82BE2F8315AF4ULL, "character_character_p106.psh.fxo.dx11.assembly"},
	{0xADE85B50090E4E1AULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_brickround01_d.dds"},
	{0xADE8B6401187A8B1ULL, "vxht_chantry_n_28q_0_0.vlm"},
	{0xADE8C20962D4279BULL, "vxca_caverns2_71m_0_0.vlm"},
	{0xADE9354EB52AA3C1ULL, "364583846.ogg"},
	{0xADE94CD706A8BF0BULL, "vxw_bonepit_p3_0_0.vlm"},
	{0xADE98284B6A46D23ULL, "816759386.ogg"},
	{0xADE99074B6D60968ULL, "hlac_w_sundermount_-131077_0y.dds"},
	{0xADE996FA08CD2983ULL, "vxlt_house_do_0_0.vlm"},
	{0xADE9E94FF90497BDULL, "hla_ht_0_hightown_a3_-131072_0y.dds"},
	{0xADEA1DBAC9848BA9ULL, "372627105.ogg"},
	{0xADEA37426C54D986ULL, "vfx0_cs_dba.vsh.fxo"},
	{0xADEA555A5DA37047ULL, "llac_lt_0_lowtown_docks_1_0.dds"},
	{0xADEA600D3631ECF1ULL, "vxca_caverns1_mines_fn9_0_0.vlm"},
	{0xADEA7AAB3422CB9CULL, "art\\levels\\races\\proxy\\props\\_textures\\shelf_n.dds"},
	{0xADEA7BB5A0E77638ULL, "vxlt_darktown_5ra_0_0.vlm"},
	{0xADEA92C637603237ULL, "w_woundedcoast_235_0.msh"},
	{0xADEAA4C63760509AULL, "w_woundedcoast_235_0.mao"},
	{0xADEAA8C6376057A9ULL, "w_woundedcoast_235_0.mmh"},
	{0xADEB0717C9D45CE1ULL, "lgt110_urn_guard6170684.cl"},
	{0xADEB0EE8CEBB6BD0ULL, "mag330cr_evelina.ncs"},
	{0xADEB466BAE84EFF3ULL, "6100769_m.fxe"},
	{0xADEB9DD2FB00BE27ULL, "hlac_lt_0_lowtown_a2_n_c_131072_5z.dds"},
	{0xADEBAFECA5E4E9D4ULL, "art\\vfx\\environment\\vfx_water_drips_p.anb"},
	{0xADEC1C74C7FACC28ULL, "vxht_0_hightown_a2_c4b_0_0.vlm"},
	{0xADEC24F9D7E119E2ULL, "6036480_f.fxe"},
	{0xADEC9DF912B8F44CULL, "vxw_encounter2_camp_r2_0_0.vlm"},
	{0xADECDB2B595B24ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_laughing.ani"},
	{0xADECE9F24B0D57D5ULL, "dre000_varric6036988.cl"},
	{0xADECF789BFE3CEEEULL, "main_plot_screenshots.gda"},
	{0xADEE107DCDA7214DULL, "106387495.ogg"},
	{0xADEE5655DDC59D35ULL, "266574883.ogg"},
	{0xADEEE038B96CA625ULL, "plt_ser120pt_sacrifice_offe.plo"},
	{0xADEEE20B8191BD7CULL, "gen00fl_isabela6097358.cl"},
	{0xADEF5A418C0C477DULL, "6060238_m.fxe"},
	{0xADF001250F175BCFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallb.mao"},
	{0xADF00D250F17706CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallb.mmh"},
	{0xADF013250F177A5EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstallb.msh"},
	{0xADF108FF3A09AE4AULL, "206483919.ogg"},
	{0xADF118CB0CBB4746ULL, "246541194.ogg"},
	{0xADF129D033E2E2BCULL, "gen00fl_aveline6132876.cl"},
	{0xADF1A316288947A1ULL, "mag311_emile_de_launcet6043673.cl"},
	{0xADF23DDC76CEC2E1ULL, "vxlt_hangedman_4r_0_0.vlm"},
	{0xADF271C964F1CDBAULL, "6041269_m.fxe"},
	{0xADF27C7E334E8421ULL, "llac_dr_0_deeproads_5845_7.dds"},
	{0xADF283833F37DF82ULL, "381214281.ogg"},
	{0xADF2B4B8BBCEA88AULL, "6102762_m.fxe"},
	{0xADF2BB85B420E175ULL, "w_encounter3_60_0.msh"},
	{0xADF2BD85B420E4EBULL, "w_encounter3_60_0.mmh"},
	{0xADF2C985B420F888ULL, "w_encounter3_60_0.mao"},
	{0xADF2E82B595FA62EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_laughing.gad"},
	{0xADF2F462B8AFF290ULL, "6133827_m.fxe"},
	{0xADF2FF2E02A26233ULL, "736788712.ogg"},
	{0xADF3051423D69AF5ULL, "849361585.ogg"},
	{0xADF32DA1C0892D92ULL, "ca_caverns2prc_5802.rml"},
	{0xADF365FC995204C2ULL, "mag100_thrask6031743.cl"},
	{0xADF3B2E24C373E13ULL, "vxw_bonepitdestroyed_a3_1k5_0_0.vlm"},
	{0xADF3D2D1D69DAB72ULL, "mcr370_meredith6136151.cl"},
	{0xADF41A6976B59B33ULL, "259274972.ogg"},
	{0xADF42DCD5D7128AEULL, "dae000ip_crowd_ff4_talk16.utp"},
	{0xADF48718D9197AF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_2.ani"},
	{0xADF4AB3BB67F052CULL, "art\\characters\\playercharacter\\humanmale\\hlm_mede.mao"},
	{0xADF4C2CB3C90EA8CULL, "6043386_m.fxe"},
	{0xADF5137E3EDA44B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_enter.ani"},
	{0xADF5B1B3B8284B5DULL, "zz_dre_debug6204371.cl"},
	{0xADF5DE6B2CE92455ULL, "vxw_bonepit_1hv_0_0.vlm"},
	{0xADF73FC2854F6B40ULL, "zz_area_signage9.utp"},
	{0xADF745DE7EBC8144ULL, "gen00fl_anders6125383.cl"},
	{0xADF768E6610EF40CULL, "6143482_m.fxe"},
	{0xADF83220DB462C6EULL, "art\\vfx\\combat\\vfx_gib_hm_masc_lbrt.anb"},
	{0xADF86CFC7865E23DULL, "vxlt_0_lowtown_e0m_0_0.vlm"},
	{0xADF86E19CADDD1E3ULL, "zz_isa_debug6133178.cl"},
	{0xADF88EECA5EF3B97ULL, "art\\vfx\\environment\\vfx_water_drips_p.mmh"},
	{0xADF8AE78D9B52E8AULL, "vxht_keep_30l_0_0.vlm"},
	{0xADF8AF5157A82B1CULL, "cod_bks_chantry_hist_3.utp"},
	{0xADF8F34D485DD3F6ULL, "llac_ht_chantry_2534_41.dds"},
	{0xADF946D25C40718FULL, "vxca_cavern1_haunted_ebd_0_0.vlm"},
	{0xADF94D81FEA836E7ULL, "119816119.ogg"},
	{0xADF9504553404362ULL, "hlac_lt_0_lowtown_docks_-196611_0y.dds"},
	{0xADF9B3959152BA63ULL, "788631705.ogg"},
	{0xADF9E32B5965BC0DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_laughing.evt"},
	{0xADFA5656F01AA492ULL, "vxw_bonepitmining_a2_2ag_0_0.vlm"},
	{0xADFA7E2AAE5E2F9FULL, "lt_undercityencounter_3567_ll.mmh"},
	{0xADFA8E51B9F0E4A8ULL, "vxht_0_hightown_a2_c3c_0_0.vlm"},
	{0xADFAACC357ED827AULL, "6033394_m.fxe"},
	{0xADFB0FD604AF0CCCULL, "vxw_encounter2_camp_m3_0_0.vlm"},
	{0xADFB389CBA9B7451ULL, "6063851_m.fxe"},
	{0xADFB72A7ECE94BAFULL, "plt_cod_hst_mgc_orichalcum.plo"},
	{0xADFB9218D91FAB93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_2.gad"},
	{0xADFBC69238236EC8ULL, "vxlt_hangedman_4w_0_0.vlm"},
	{0xADFBCED75947E4D8ULL, "852875908.ogg"},
	{0xADFBEE7E3EE024FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_enter.gad"},
	{0xADFC7E2BC778A370ULL, "isa311_isabela6092384.cl"},
	{0xADFC957908478D84ULL, "vxht_0_hightown_anj_0_0.vlm"},
	{0xADFCC5D4A1ED212DULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_fema_0n.dds"},
	{0xADFD010C03E2543BULL, "level_buildtest_int.arl"},
	{0xADFD69FE9C6BAB59ULL, "261981178.ogg"},
	{0xADFDAB9E69DF91E5ULL, "llac_ca_caverns2_5802_6.dds"},
	{0xADFDB5A40C321924ULL, "37561583.ogg"},
	{0xADFECA8E628F11DEULL, "85964185.ogg"},
	{0xADFECF698BAADCF4ULL, "vxlt_0_lowtown_docks_a3_cuo_0_0.vlm"},
	{0xADFF3D4B50A6FD22ULL, "mag320_orsino6071063.cl"},
	{0xADFF55B47FF92C05ULL, "plt_cod_hst_surfacedwarves.plo"},
	{0xADFFF06C45D2E09CULL, "vxw_sundermount_31h_0_1.vlm"},
	{0xAE00C9C858DDE86FULL, "6130069_m.fxe"},
	{0xAE00CCF979EA5F2CULL, "gen_im_flw_shd_aveline_02.uti"},
	{0xAE00E1B654720A69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkbackleft.gad"},
	{0xAE00E6581B0F1CABULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_pc.evt"},
	{0xAE0112976613AA77ULL, "720607412.ogg"},
	{0xAE0133F9334472B2ULL, "346224807.ogg"},
	{0xAE01E0201DA9C637ULL, "6028391_m.fxe"},
	{0xAE020B7E3EE4C1E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_enter.evt"},
	{0xAE025FE8610CC7F8ULL, "6117973_m.fxe"},
	{0xAE028250593EBE94ULL, "302375736.ogg"},
	{0xAE028F18D925C5F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_self_2.evt"},
	{0xAE02F2B592AA1F8CULL, "statecnt_dialog.gda"},
	{0xAE030B321BBDA28CULL, "vxlt_0_lowtown_docks_n_aln_0_0.vlm"},
	{0xAE0372359695F558ULL, "6166348_m.fxe"},
	{0xAE03AC343C8E19ABULL, "zz_mrl_debug6086522.cl"},
	{0xAE03AC650F0FDF8AULL, "gen00fl_dog6114928.cl"},
	{0xAE03ECD7C1CC16EEULL, "337075281.ogg"},
	{0xAE043FB3552452C4ULL, "6084527_m.fxe"},
	{0xAE0450481EDF3CD5ULL, "vxw_bonepit_1mu_0_0.vlm"},
	{0xAE067EB3738B23A0ULL, "follower_banter6155619.cl"},
	{0xAE06A315816D0891ULL, "6149092_m.fxe"},
	{0xAE06D217081D24DFULL, "vxw_bonepitdestroyed_a3_2ap_0_0.vlm"},
	{0xAE073701A96C2FA0ULL, "919288157.ogg"},
	{0xAE07B2B65477D8B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkbackleft.ani"},
	{0xAE07B8AF4E368A0FULL, "vxca_cavern1_haunted_egg_0_0.vlm"},
	{0xAE07C035234B4483ULL, "lgt120_orz_bln_off_messenge.bnk"},
	{0xAE07D4AF56DCC45DULL, "vxw_encounter2_bo_0_0.vlm"},
	{0xAE07F28DBB78A1D4ULL, "hlac_lt_gallowscourtyard_-65537_4y.dds"},
	{0xAE07F724E835EF42ULL, "hlac_lt_0_lowtown_docks_0_4z.dds"},
	{0xAE07FF581B15666CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_pc.gad"},
	{0xAE0884179B487A74ULL, "6043265_m.fxe"},
	{0xAE0936E41138351FULL, "lgt110_arl_bann_teagan6115574.cl"},
	{0xAE0981B2F6A5254CULL, "vxw_encounter2_camp_h0_0_0.vlm"},
	{0xAE09B120EFAD07D8ULL, "llac_w_blightlands_327678_0.dds"},
	{0xAE0A04745DA5A6ADULL, "213775536.ogg"},
	{0xAE0AA235234D47B3ULL, "lgt120_orz_bln_off_messenge.cnv"},
	{0xAE0AF2735D52DEC0ULL, "vxca_caverns2prc_7o5_0_0.vlm"},
	{0xAE0AF29B86CFF34FULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgc.mao"},
	{0xAE0AFE9B86D007ECULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgc.mmh"},
	{0xAE0B049B86D011DEULL, "art\\characters\\weapons\\dagger\\w_dgr_dsdgc.msh"},
	{0xAE0B33470DED748BULL, "art\\vfx\\combat\\vfx_c_mer_thrust_c.impa.ani"},
	{0xAE0B5DDA4D5C7FE6ULL, "vxht_mansion_n_5bs_0_0.vlm"},
	{0xAE0BD2862908CBE9ULL, "vxht_mansion_n_4c2_0_0.vlm"},
	{0xAE0BD7C00FF6BE87ULL, "1002002907.ogg"},
	{0xAE0C03C5717DA402ULL, "vxlt_undercitydungeon_17g_0_0.vlm"},
	{0xAE0C067CC7D74AB8ULL, "6080292.cut"},
	{0xAE0C2D4837A2C1B3ULL, "boulder_mac_center.mao"},
	{0xAE0C424EB7370F30ULL, "6139601_m.fxe"},
	{0xAE0C5EC3755732E8ULL, "142830097.ogg"},
	{0xAE0C9AFE601D1311ULL, "6156316_f.fxe"},
	{0xAE0CB0DD2EA76E06ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_icon.phy"},
	{0xAE0CEFDFFF6445EDULL, "vxlt_0_lowtown_a2_n_h2i_0_0.vlm"},
	{0xAE0D0A80B0228DB3ULL, "isa200cr_badpoet.utc"},
	{0xAE0D1FEE96F377BBULL, "838802664.ogg"},
	{0xAE0D41467DA0F574ULL, "vxlt_0_lowtown_docks_a3_cxl_0_0.vlm"},
	{0xAE0D61962F069AB6ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg105a_0s.dds"},
	{0xAE0D8866EFE33633ULL, "vxlt_0_lowtown_eup_0_0.vlm"},
	{0xAE0DE8997924BB48ULL, "hlac_dr_primevaltomb_659_14y.dds"},
	{0xAE0E04243156F64AULL, "vxw_encounter3_6k_0_0.vlm"},
	{0xAE0ED59984FA25CFULL, "ht_0_hightown_a2_43_0_ll.mmh"},
	{0xAE0EF28B9BAAD045ULL, "6118137_m.fxe"},
	{0xAE0EFE581B1B836AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncloop_pc.ani"},
	{0xAE0F0671131DA1AEULL, "qcr220_grey_wardens6060414.cl"},
	{0xAE0F4FEB571CDB39ULL, "6165340_m.fxe"},
	{0xAE10008B0E59E517ULL, "one110_gustav6089821.cl"},
	{0xAE1010D0D8236FF1ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.impa.ani"},
	{0xAE10742429DA053DULL, "gen00fl_anders6119493.cl"},
	{0xAE10AA0C36912339ULL, "vxlt_hangedman_2mo_0_0.vlm"},
	{0xAE113F7ACA26F473ULL, "vxlt_backalley_n_1kj_0_0.vlm"},
	{0xAE1147FD43C325C6ULL, "566372717.ogg"},
	{0xAE11C0C17ED47424ULL, "339173976.ogg"},
	{0xAE11D47ED4E5B637ULL, "110664621.ogg"},
	{0xAE12A2722867C6E7ULL, "llac_dr_primevaltomb_659_12.dds"},
	{0xAE12A43732D613E1ULL, "vxlt_0_lowtown_n_ewo_0_0.vlm"},
	{0xAE1394D7D48F6B2AULL, "hlac_w_woundedcoast_5_0z.dds"},
	{0xAE13D3EE82328EACULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapupper.msh"},
	{0xAE13D9EE8232989EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapupper.mmh"},
	{0xAE13E5EE8232AD35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_wincapupper.mao"},
	{0xAE14780266BBA257ULL, "w_encounter3_143_0.mmh"},
	{0xAE147C0266BBA964ULL, "w_encounter3_143_0.mao"},
	{0xAE148E0266BBC7C9ULL, "w_encounter3_143_0.msh"},
	{0xAE149667932ABD47ULL, "plt_dae400pt_post_game.plo"},
	{0xAE153A7DC80D7016ULL, "fen310_varania6118589.cl"},
	{0xAE1558BC5EDC089AULL, "follower_banter6173183.cl"},
	{0xAE16468C48D2DCDDULL, "vxw_encounter2_mn_0_0.vlm"},
	{0xAE1679697140E6DEULL, "gen00fl_anders6113072.cl"},
	{0xAE16915D5FC7C5A2ULL, "6138742_f.fxe"},
	{0xAE16C114478E9FCCULL, "313336036.ogg"},
	{0xAE171BD0D829A093ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.impa.gad"},
	{0xAE17491ADD01A27EULL, "vxw_woundedcoast_16d_0_0.vlm"},
	{0xAE17A73FC2481DE8ULL, "corpses_d.dds"},
	{0xAE17BD5C5DBD7C41ULL, "36086860.ogg"},
	{0xAE17C888B2CC19BAULL, "vxlt_hangedman_21t_0_0.vlm"},
	{0xAE17F38FE89B3DCCULL, "vxw_encounter2_camp_c1_0_0.vlm"},
	{0xAE1823F2DA5FAE50ULL, "zz_followers_banter6157893.cl"},
	{0xAE191B470DF98902ULL, "art\\vfx\\combat\\vfx_c_mer_thrust_c.impa.evt"},
	{0xAE19603BF4D17DA4ULL, "w_woundedcoast_41_0.msh"},
	{0xAE19723BF4D19C0DULL, "w_woundedcoast_41_0.mao"},
	{0xAE19763BF4D1A346ULL, "w_woundedcoast_41_0.mmh"},
	{0xAE19BDE6E9953BEEULL, "hlac_ca_cave_2194_3x.dds"},
	{0xAE1A28CBF1070B20ULL, "vxw_bonepitdestroyed_a3_25w_0_0.vlm"},
	{0xAE1A292677AE1D2BULL, "hlac_lt_0_lowtown_a2_n_131070_4y.dds"},
	{0xAE1A5054B1789F0BULL, "llac_ht_0_hightown_n_1_0.dds"},
	{0xAE1A5CE77AA7E2D0ULL, "gen00fl_aveline6122154.cl"},
	{0xAE1A74BE5F495EEEULL, "hlac_ca_caverns1_24423_4z.dds"},
	{0xAE1AA5E296F40E97ULL, "w_sundermount_31_0.mmh"},
	{0xAE1AA9E296F415A4ULL, "w_sundermount_31_0.mao"},
	{0xAE1ABBE296F43409ULL, "w_sundermount_31_0.msh"},
	{0xAE1AF161653ED010ULL, "6160971_m.fxe"},
	{0xAE1B38741F9A8DAAULL, "llac_ht_0_hightown_a2_-131074_2.dds"},
	{0xAE1B64F8F85A3F72ULL, "334657950.ogg"},
	{0xAE1B71B2912AF8E5ULL, "hlac_w_sundermount_262141_0x.dds"},
	{0xAE1B99EED65B9453ULL, "vxht_0_hightown_n_anb_0_0.vlm"},
	{0xAE1BBC66D3B94B86ULL, "6117613_f.fxe"},
	{0xAE1BDBFEFF2C7A27ULL, "171414942.ogg"},
	{0xAE1BE77FFABBA151ULL, "vxlt_0_lowtown_a3_n_c_h2f_0_0.vlm"},
	{0xAE1C03721946950EULL, "433393261.ogg"},
	{0xAE1C2CECF42ABFC8ULL, "vxlt_0_lowtown_n_ewj_0_0.vlm"},
	{0xAE1C7601234D0ECAULL, "vxw_encounter3_3d_0_0.vlm"},
	{0xAE1D2BD4E17102F2ULL, "narrative_banter6136814.cl"},
	{0xAE1D6815630D8048ULL, "vxdr_0_deeproads_8sk_0_0.vlm"},
	{0xAE1D94DEF09D393CULL, "vxlt_gallowscourtyard_n_7ew_0_0.vlm"},
	{0xAE1DA11705A79A44ULL, "6089890_m.fxe"},
	{0xAE1E18D0D82FBAF8ULL, "art\\vfx\\combat\\vfx_m_spi_md_imp_c.impa.evt"},
	{0xAE1EA9F18F55B323ULL, "vxht_mansion2b_1bu_0_0.vlm"},
	{0xAE1F6933AD26A3E4ULL, "qun230cr_guild_contact6068690.cl"},
	{0xAE1FB157BC1D0CF3ULL, "vxlt_backalley_n_1vm_0_0.vlm"},
	{0xAE1FEDC851525FECULL, "617886211.ogg"},
	{0xAE201A470DFFA615ULL, "art\\vfx\\combat\\vfx_c_mer_thrust_c.impa.gad"},
	{0xAE206B271A71A73CULL, "622807241.ogg"},
	{0xAE209FC35664C279ULL, "6084539_m.fxe"},
	{0xAE20FE7D8822239DULL, "zz_approval_debug6003734.cl"},
	{0xAE214FBAEF6A446BULL, "6205657_m.fxe"},
	{0xAE21C5B86ED39696ULL, "6145172_m.fxe"},
	{0xAE21DA57AB75D309ULL, "285276874.ogg"},
	{0xAE21E03F4B15A23AULL, "vxca_caverns1_mines_fm7_0_0.vlm"},
	{0xAE227C17B7C5C1F3ULL, "6030462_m.fxe"},
	{0xAE23144ACE81210BULL, "vxw_bonepit_l7_0_0.vlm"},
	{0xAE24DADBF3D21ABEULL, "art\\vfx\\combat\\vfx_m_ent_sleep_act_c.impa.ani"},
	{0xAE2501F90AACC5EBULL, "390292608.ogg"},
	{0xAE251AA497B032D2ULL, "vxht_0_hightown_n_ang_0_0.vlm"},
	{0xAE2543B013599993ULL, "150635201.ogg"},
	{0xAE2582CEB8FCA6E6ULL, "vxlt_hangedman_1qs_0_0.vlm"},
	{0xAE259B140C1D05D3ULL, "ser270_emeric6113508.cl"},
	{0xAE25A2D760F7B655ULL, "vxca_caverns1_edq_0_0.vlm"},
	{0xAE25BAF7CEF7BAFEULL, "vxw_woundedcoast_11e_0_0.vlm"},
	{0xAE25DA476512BCF4ULL, "408344496.ogg"},
	{0xAE25F52596FC9FFDULL, "ai_enable.ncs"},
	{0xAE263BC2084732FAULL, "art\\cinematics\\act3\\romances\\isabella\\cs_pcfem_isarom.gad"},
	{0xAE2726E73E983DB8ULL, "hlac_lt_0_lowtown_docks_-131073_4x.dds"},
	{0xAE273CD3BC3767AFULL, "687469861.ogg"},
	{0xAE27BFF58DF0EFB7ULL, "follower_banter6171573.cl"},
	{0xAE2841943148B0E6ULL, "vxht_mansion_n_5hu_0_0.vlm"},
	{0xAE287FDBFF00140EULL, "hlac_w_bonepit_65538_0y.dds"},
	{0xAE28C65BAA503AE7ULL, "672110880.ogg"},
	{0xAE29117B391AEEA9ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medj.phy"},
	{0xAE291817F8908D6BULL, "gen00fl_aveline6206185.cl"},
	{0xAE2A0EB256B167DCULL, "vxlt_0_lowtown_docks_bbt_0_0.vlm"},
	{0xAE2A4F606F5103F4ULL, "llac_dr_dungeon_3214_8.dds"},
	{0xAE2A74E133D57B3CULL, "raceicons.gda"},
	{0xAE2A763BF4E0160DULL, "w_woundedcoast_41_0.rml"},
	{0xAE2AD49EEA8E7548ULL, "zz_followers_banter6158496.cl"},
	{0xAE2B3539129D8C33ULL, "con_ico_helpful.dds"},
	{0xAE2B880CAB849116ULL, "6088931_f.fxe"},
	{0xAE2B98E4D55C438BULL, "vxca_caverns2prc_9qz_0_0.vlm"},
	{0xAE2C06BBE29351BCULL, "vxlt_gallowscourtyard_n_7ht_0_0.vlm"},
	{0xAE2C5042A82D5459ULL, "mer120_hubert6034431.cl"},
	{0xAE2CED0AE7965D5AULL, "980122101.ogg"},
	{0xAE2E492B44C9F04EULL, "177044979.ogg"},
	{0xAE2EDF7940B996C9ULL, "6151174_m.fxe"},
	{0xAE2F3D99DC57E11EULL, "525873738.ogg"},
	{0xAE2F517B0B684282ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic_1.mao"},
	{0xAE2F787BFD236408ULL, "937621541.ogg"},
	{0xAE2F87F116C244E1ULL, "vxlt_0_lowtown_n_emm_0_0.vlm"},
	{0xAE2FBF881C691176ULL, "58618981.ogg"},
	{0xAE2FCBFFCEF02DAEULL, "w_bonepitdestroyed_a3_176_0.mmh"},
	{0xAE2FD7FFCEF04245ULL, "w_bonepitdestroyed_a3_176_0.mao"},
	{0xAE2FE5FFCEF0599CULL, "w_bonepitdestroyed_a3_176_0.msh"},
	{0xAE306E700CB765D2ULL, "6065703_f.fxe"},
	{0xAE30DAF0066C15F6ULL, "vxht_0_hightown_c5k_0_0.vlm"},
	{0xAE311BC23C8F8C35ULL, "412249830.ogg"},
	{0xAE31254EAC07D986ULL, "vxw_woundedcoast_14v_0_0.vlm"},
	{0xAE31CA14080F3C8BULL, "6036341_m.fxe"},
	{0xAE31F2DBF3DCCD8FULL, "art\\vfx\\combat\\vfx_m_ent_sleep_act_c.impa.evt"},
	{0xAE3246E57CFBF195ULL, "591619763.ogg"},
	{0xAE32546370971AC1ULL, "vxht_bartrand_n_5wh_0_0.vlm"},
	{0xAE32DFD1EF2143A2ULL, "fen110_fenris6150715.cl"},
	{0xAE32FF7EC52EDC11ULL, "vxlt_warehouse_n_14x_0_0.vlm"},
	{0xAE33DE8CBFE49296ULL, "w_bonepitdestroyed_a3_162_0.mao"},
	{0xAE33E20179D6E959ULL, "549746129.ogg"},
	{0xAE33EA8CBFE4A735ULL, "w_bonepitdestroyed_a3_162_0.mmh"},
	{0xAE33EC8CBFE4AAABULL, "w_bonepitdestroyed_a3_162_0.msh"},
	{0xAE340CE47F8EB6EDULL, "vxca_caverns1_mines_dpa_0_0.vlm"},
	{0xAE343DBBC8FC7509ULL, "vxlt_0_lowtown_docks_n_bv2_0_0.vlm"},
	{0xAE34818CD80FBA0BULL, "zz_dre_debug6168279.cl"},
	{0xAE350500DB28024CULL, "vxdr_slavers_19r_0_0.vlm"},
	{0xAE35740A9315865BULL, "dae000_male_commoner2_mob6174432.cl"},
	{0xAE3698F49146EF09ULL, "vxca_caverns2_91t_0_1.vlm"},
	{0xAE36A220C643837DULL, "gen00fl_merrill6108633.cl"},
	{0xAE36DBE889B86F61ULL, "336408306.ogg"},
	{0xAE36EBB4982A0040ULL, "vxca_caverns1_j2d_0_1.vlm"},
	{0xAE36FCAD8BA106ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cf_transition.ani"},
	{0xAE378DE7B97BB8CFULL, "6115899_m.fxe"},
	{0xAE37FD846E935AE7ULL, "dae000_paivel6134467.cl"},
	{0xAE3801FE9D4D7B87ULL, "vxca_cave_1o2_0_0.vlm"},
	{0xAE384DFD6DCE7986ULL, "6201601_m.fxe"},
	{0xAE38F3DBF3E2EE38ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_act_c.impa.gad"},
	{0xAE38FDE428B81B74ULL, "llac_lt_0_lowtown_a3_n_c_-131072_8.dds"},
	{0xAE3905E8CEFD7DCAULL, "mag330cr_evelina.utc"},
	{0xAE3910A6D816F0C8ULL, "vxlt_0_lowtown_n_emh_0_0.vlm"},
	{0xAE39478501AC1232ULL, "6163411_m.fxe"},
	{0xAE3958F69BEB3EBCULL, "6034233_f.fxe"},
	{0xAE39CA20833E8513ULL, "vxdr_primevaltomb_8v5_0_0.vlm"},
	{0xAE3A0C27BDC21082ULL, "dae000_arishok6075202.cl"},
	{0xAE3A56C2085809E0ULL, "art\\cinematics\\act3\\romances\\isabella\\cs_pcfem_isarom.ani"},
	{0xAE3A69E6D67BCDE3ULL, "6147140_m.fxe"},
	{0xAE3A96CEA3290226ULL, "6187901_m.fxe"},
	{0xAE3ADCAD29561671ULL, "vxca_caverns1_mines_j20_0_1.vlm"},
	{0xAE3AEF41A54FD387ULL, "prp_str_rock_d.dds"},
	{0xAE3B592DCB15C613ULL, "6162541_m.fxe"},
	{0xAE3B8DAB7341E423ULL, "vxht_mansion2b_1xw_0_0.vlm"},
	{0xAE3B9FECD8BA9987ULL, "vxlt_0_lowtown_h9q_0_0.vlm"},
	{0xAE3BFDA800AF54CAULL, "vxht_mansion_n_633_0_0.vlm"},
	{0xAE3C287151C7040DULL, "6127406_f.fxe"},
	{0xAE3C41ABA8A4EB12ULL, "hlac_lt_0_lowtown_docks_-196610_6x.dds"},
	{0xAE3C7513B89A7B4CULL, "qcr220_grey_wardens6060395.cl"},
	{0xAE3C8B77F10C8485ULL, "vxlt_0_lowtown_docks_a3_d11_0_0.vlm"},
	{0xAE3D09AD8BA5882EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cf_transition.gad"},
	{0xAE3E11A711173E30ULL, "vxca_caverns2prc_9pw_0_1.vlm"},
	{0xAE3E4B833EDA6B49ULL, "lt_0_lowtown_docks_n_245_0.mao"},
	{0xAE3E57833EDA7FE2ULL, "lt_0_lowtown_docks_n_245_0.mmh"},
	{0xAE3E590429710DC3ULL, "hlac_lt_hangedman_-1_9y.dds"},
	{0xAE3E59833EDA8298ULL, "lt_0_lowtown_docks_n_245_0.msh"},
	{0xAE3E5BB63D92CA45ULL, "38175283.ogg"},
	{0xAE3E5F4686EC63B0ULL, "6170699_m.fxe"},
	{0xAE3E7623ABDE4D3EULL, "vxdr_primevalentrance_2ol_0_1.vlm"},
	{0xAE3F3A974C056C5DULL, "812285812.ogg"},
	{0xAE3F4889E068625CULL, "vxht_keep_n_4bp_0_0.vlm"},
	{0xAE3F4B04F4B7DD05ULL, "vxlt_darktown_2s8_0_0.vlm"},
	{0xAE3F78560F03ED05ULL, "vxlt_0_lowtown_docks_a3_ary_0_0.vlm"},
	{0xAE401341AEB91715ULL, "6095951_m.fxe"},
	{0xAE40A9AB6730492EULL, "vxca_caverns1_gik_0_0.vlm"},
	{0xAE41009ACFE1276EULL, "llac_ht_0_hightown_a3_-3_0.dds"},
	{0xAE41AF77E9EA9C90ULL, "dae000_bartrand6027514.cl"},
	{0xAE426BBBBCD73519ULL, "one220_yevhen6103724.cl"},
	{0xAE42F0FAE156CD85ULL, "6107211_f.fxe"},
	{0xAE434B72EE84560DULL, "vxlt_0_lowtown_docks_a3_bw5_0_0.vlm"},
	{0xAE435E2C442E71C6ULL, "dae121ip_at_keep_hightown.utp"},
	{0xAE43BBB81C3C7732ULL, "779644555.ogg"},
	{0xAE4404AD8BAB9E0DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cf_transition.evt"},
	{0xAE442BDCBCB5589CULL, "vxlt_0_lowtown_a2_n_d5u_0_0.vlm"},
	{0xAE456EE19FAA9E10ULL, "mag311_nyssa6042364.cl"},
	{0xAE45A5881A97FFBEULL, "828458966.ogg"},
	{0xAE45ADF91166F940ULL, "static_static_p49.psh.fxo.dx11"},
	{0xAE45DAED55BBD0A8ULL, "one300pt_herb_dragon.ncs"},
	{0xAE45FFC4AB71D14FULL, "6110799_m.fxe"},
	{0xAE46629E6B898A1BULL, "6123365_f.fxe"},
	{0xAE467E9855003F03ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\enrage.ani"},
	{0xAE469A6671235C0EULL, "1021668181.ogg"},
	{0xAE46A7686C2B575EULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar2_.ani"},
	{0xAE46AFAE80BFB3DDULL, "gen_im_arm_boo_med_act1_01.uti"},
	{0xAE46CB8440F907DFULL, "one100dg_kelder6173950.cl"},
	{0xAE46E4A525EA5E79ULL, "clf260_harley6088761.cl"},
	{0xAE478283CA0D0948ULL, "vxlt_0_lowtown_n_efi_0_0.vlm"},
	{0xAE479C730F3C14BAULL, "hlac_w_encounter3_-262143_0x.dds"},
	{0xAE47A8010877020EULL, "vxdr_primevaltomb_6wu_0_0.vlm"},
	{0xAE47E587BB47CA26ULL, "gen00fl_fenris6105558.cl"},
	{0xAE484D9974918E72ULL, "hlac_dr_slavers_3215_9z.dds"},
	{0xAE484FAB76A3B4BEULL, "follower_banter6162097.cl"},
	{0xAE486488ABF063E1ULL, "6135271_m.fxe"},
	{0xAE488D740396D7CAULL, "vxht_keep_n_6ya_0_0.vlm"},
	{0xAE489008AC003B68ULL, "vrc221_varric6043601.cl"},
	{0xAE49CB0ABD0BDE93ULL, "6167641_m.fxe"},
	{0xAE49FF886537FCA3ULL, "vxht_mansion2b_1sv_0_0.vlm"},
	{0xAE4A89EC669A3C13ULL, "zz_one000_merchant6179343.cl"},
	{0xAE4A8F81DA33FA60ULL, "6101134_m.fxe"},
	{0xAE4AE8E9E5392E99ULL, "412014467.ogg"},
	{0xAE4CEC8180025067ULL, "879019351.ogg"},
	{0xAE4CFD81AE9685FBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofsw.mmh"},
	{0xAE4D0981AE969998ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofsw.mao"},
	{0xAE4D1B81AE96B8E5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofsw.msh"},
	{0xAE4D76689D18A056ULL, "240841775.ogg"},
	{0xAE4DE958D76F6BB2ULL, "251721110.ogg"},
	{0xAE4DEE4DCB9C8580ULL, "213834724.ogg"},
	{0xAE4EB17200103C1BULL, "6075102_m.fxe"},
	{0xAE4F3BDAD9374D3DULL, "dae021_sister_lorena_sta6150062.cl"},
	{0xAE4F51B85DE83332ULL, "hlac_ht_mansion2b_818_56y.dds"},
	{0xAE4FABCB4C4ECFD8ULL, "6144771_m.fxe"},
	{0xAE4FBEE29E0A1619ULL, "vxca_caverns1_ikd_0_0.vlm"},
	{0xAE4FE9A18F41A0E7ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_5y.dds"},
	{0xAE50D88371788DB9ULL, "35136773.ogg"},
	{0xAE50F57CBCDA6EAAULL, "hlac_ht_mansion2_818_31x.dds"},
	{0xAE514680580E9E1EULL, "vxw_encounter2_camp_qm_0_0.vlm"},
	{0xAE51A63FF450C9E1ULL, "additivemagicarc.mao"},
	{0xAE51D824BF8A2D1EULL, "gen_im_acc_amu_act1_03.uti"},
	{0xAE5216DC4D98CBBAULL, "6031297_m.fxe"},
	{0xAE5268061A4997E1ULL, "ser221_party_comments6116130.cl"},
	{0xAE526A14AF6BBBD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_left.evt"},
	{0xAE532BC273785FD3ULL, "6127577_m.fxe"},
	{0xAE53561CD4CCD4D3ULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.cess.ani"},
	{0xAE539698550AF2FAULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\enrage.evt"},
	{0xAE53CC8E79F1ED9FULL, "hlac_lt_warehouse_n_0_2y.dds"},
	{0xAE542FDE2FD049BEULL, "blank_exterior_mmap_ph.dds"},
	{0xAE54C9C1E2E0A8EDULL, "hlac_ca_caverns1_24427_1z.dds"},
	{0xAE54FA7B394002CEULL, "art\\characters\\playercharacter\\humanfemale\\boo_medj.mmh"},
	{0xAE55067B39401765ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medj.mao"},
	{0xAE55147B39402EBCULL, "art\\characters\\playercharacter\\humanfemale\\boo_medj.msh"},
	{0xAE553270CA91955CULL, "vxca_cavern1_haunted_i7d_0_0.vlm"},
	{0xAE553C3E0147AA36ULL, "558507789.ogg"},
	{0xAE55C195390F1906ULL, "807900701.ogg"},
	{0xAE55D2ABADE6EAE9ULL, "87000329.ogg"},
	{0xAE55DE34EDBC746AULL, "527800565.ogg"},
	{0xAE564E37EC2D7DABULL, "vxca_caverns1_fms_0_0.vlm"},
	{0xAE5696093427EB2BULL, "gen00fl_sebastian6179786.cl"},
	{0xAE56F1B20E90068BULL, "6068009_m.fxe"},
	{0xAE56FF05948F6934ULL, "hlac_lt_0_lowtown_docks_a3_65534_3y.dds"},
	{0xAE579A55158674D9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_spider_movement.anb"},
	{0xAE57E05DD939D11FULL, "hlac_ca_caverns2prc_5802_40y.dds"},
	{0xAE583FB18691FF61ULL, "6147605_m.fxe"},
	{0xAE587165572E1523ULL, "vxht_mansion2b_1vy_0_0.vlm"},
	{0xAE587E7397AF5B51ULL, "vxw_encounter2_67_0_0.vlm"},
	{0xAE58E365FF073C4DULL, "one100dg_kelder6055627.cl"},
	{0xAE593A35B7B08985ULL, "l_env_wal_lowtown_cough.wbd"},
	{0xAE593C44D25B8F2BULL, "mrl210_merrill6102099.cl"},
	{0xAE5962CA4E8E332AULL, "llac_lt_0_lowtown_a2_n_c_0_2.dds"},
	{0xAE596314AF71CE14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_left.gad"},
	{0xAE5A5E4C12FBCA60ULL, "995579505.ogg"},
	{0xAE5A6EE8673476DDULL, "6208318_f.fxe"},
	{0xAE5A959855110FEDULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperwarrior\\enrage.gad"},
	{0xAE5AB3B399C10103ULL, "89630449.ogg"},
	{0xAE5AC0686C3C2AD8ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar2_.gad"},
	{0xAE5B4F3412DB5186ULL, "mag370_meredith6092195.cl"},
	{0xAE5B60517941DE85ULL, "follower_banter6163239.cl"},
	{0xAE5BE9CC303E0F11ULL, "349120468.ogg"},
	{0xAE5CA6A56322BBEFULL, "zz_followers_banter6158916.cl"},
	{0xAE5D70807CF8AB2FULL, "w_sundermount_43_0_ll.mmh"},
	{0xAE5D75138E61BD2AULL, "dae000_gamlen6048470.cl"},
	{0xAE5D7F64647E4524ULL, "gen_im_trs_gen_bracelet_01.uti"},
	{0xAE5DD666FE5CE4E0ULL, "6176755_f.fxe"},
	{0xAE5E16F5D7697940ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cartempty_0.pwk"},
	{0xAE5E18789A6F1CEFULL, "6128513_f.fxe"},
	{0xAE5E4EFCF9BD5B1EULL, "gen00fl_fenris6115292.cl"},
	{0xAE5EFC952004D64BULL, "425710970.ogg"},
	{0xAE5F65C892B5EBC5ULL, "gen_im_trs_apo_lgt_glv_r1.uti"},
	{0xAE5FB85D4A04B69EULL, "vxw_encounter2_camp_ln_0_0.vlm"},
	{0xAE6021ED11A30263ULL, "vxht_keep_6lp_0_0.vlm"},
	{0xAE6047126B71B81BULL, "190017913.ogg"},
	{0xAE604E1CD4D7510AULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.cess.evt"},
	{0xAE607214AF7806F2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_left.ani"},
	{0xAE61561EC842839BULL, "6169052_m.fxe"},
	{0xAE616DD8359D7790ULL, "6136639_m.fxe"},
	{0xAE61B13B75D45F9EULL, "gen00fl_merrill6124421.cl"},
	{0xAE6265EAE7579C3FULL, "gen_im_trs_gem_snowobsid_01.uti"},
	{0xAE6336B06E204408ULL, "6031839_m.fxe"},
	{0xAE63B1B29617CA6EULL, "526925221.ogg"},
	{0xAE63B9A76358EBE8ULL, "hlac_ht_0_hightown_n_0_2z.dds"},
	{0xAE63DB478DD098EFULL, "zz_lgt_1_debug6114353.cl"},
	{0xAE6461955112D10AULL, "272587445.ogg"},
	{0xAE6498126138BFBBULL, "vxht_0_hightown_a3_bwx_0_0.vlm"},
	{0xAE64A3A584FBEA2AULL, "lt_gallowscourtyard_45_0.rml"},
	{0xAE64E87EB21306D0ULL, "art\\levels\\races\\design\\prp_chestb.mao"},
	{0xAE64EC7EB2130DE3ULL, "art\\levels\\races\\design\\prp_chestb.mmh"},
	{0xAE64FA7EB213253DULL, "art\\levels\\races\\design\\prp_chestb.msh"},
	{0xAE668E56BD387F25ULL, "vxht_0_hightown_c36_0_0.vlm"},
	{0xAE66CCF57CC68286ULL, "269837074.ogg"},
	{0xAE66F05089A573D1ULL, "vxw_encounter2_16_0_0.vlm"},
	{0xAE672D1CD4DD387DULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.cess.gad"},
	{0xAE675402812D4921ULL, "6126886_m.fxe"},
	{0xAE67A50201CB881DULL, "gen_im_trs_chn_blt_r1.uti"},
	{0xAE67FC9EAE74A311ULL, "1053489605.ogg"},
	{0xAE6825EB7F90D721ULL, "914147351.ogg"},
	{0xAE683C93BAA9011BULL, "6005310_m.fxe"},
	{0xAE685C031EE8611CULL, "vxca_caverns1_mines_e85_0_0.vlm"},
	{0xAE686745DF1516AFULL, "6067343_m.fxe"},
	{0xAE68B13CDEDA9574ULL, "vxca_caverns1_e97_0_0.vlm"},
	{0xAE6BC0ED01F68657ULL, "art\\levels\\races\\proxy\\props\\armoire\\prp_zzarmoire_0.pwk"},
	{0xAE6BCBABE0A8014AULL, "vxht_0_hightown_n_b8n_0_0.vlm"},
	{0xAE6BFFBC30371C67ULL, "252637677.ogg"},
	{0xAE6C62AA24BE0216ULL, "llac_dr_0_deeproads_5845_19.dds"},
	{0xAE6C632A12DD4779ULL, "gen00fl_fenris6126916.cl"},
	{0xAE6C89907B937383ULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown1.msh"},
	{0xAE6C97907B938BDDULL, "art\\characters\\creatures\\crowds\\elfmale\\crd_hightown1.mmh"},
	{0xAE6CD24819FC1196ULL, "llac_lt_backalley_g_0_0.dds"},
	{0xAE6D08F08DE8ACFBULL, "6180901_m.fxe"},
	{0xAE6D855D94523F4FULL, "gen_im_npc_wep_sta_gend.uti"},
	{0xAE6DBC1F72CB2166ULL, "vxlt_hangedman_1hx_0_0.vlm"},
	{0xAE6E2A3A3BFACF1EULL, "vxw_encounter2_camp_go_0_0.vlm"},
	{0xAE6E4EBA04E8D28AULL, "vxdr_0_deeproads_9gd_0_0.vlm"},
	{0xAE6F4AB1AD8070F0ULL, "fxe_waterblend.dds"},
	{0xAE6FA5635421951FULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_merr.mmh"},
	{0xAE6FAB6354219FC1ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_merr.msh"},
	{0xAE704C56C86EA15DULL, "follower_banter6164652.cl"},
	{0xAE7061D0F812DC62ULL, "dre000_bethany6037333.cl"},
	{0xAE70AE62E678D43DULL, "vxdr_primevalentrance_395_0_0.vlm"},
	{0xAE70FD800C831188ULL, "vxht_mansion_n_4ya_0_0.vlm"},
	{0xAE7112F891AD9302ULL, "mcr370_meredith6085936.cl"},
	{0xAE719CA9DA58CF65ULL, "lgt110_arl_bann_teagan6115936.cl"},
	{0xAE7309D229B8B750ULL, "nat380_temmerin6120942.cl"},
	{0xAE736B6F0A52BAD1ULL, "and100_anders6095845.cl"},
	{0xAE73B6D698923695ULL, "hlac_ht_mansion_n_3371_1z.dds"},
	{0xAE73D2358F2ABE0BULL, "vxca_caverns2prc_9hq_0_0.vlm"},
	{0xAE742B5AE67886BBULL, "340568012.ogg"},
	{0xAE746E7648CB3264ULL, "w_bonepitdestroyed_a3_79_0.msh"},
	{0xAE7474EC52CAD6AFULL, "art\\characters\\weapons\\shields\\textures\\w_shd_seaa_0t.dds"},
	{0xAE74807648CB50CDULL, "w_bonepitdestroyed_a3_79_0.mao"},
	{0xAE74847648CB5706ULL, "w_bonepitdestroyed_a3_79_0.mmh"},
	{0xAE74BFA9719084BFULL, "44698374.ogg"},
	{0xAE74C11296445883ULL, "vxlt_gallowstemplar_35e_0_0.vlm"},
	{0xAE74F85B56220EA8ULL, "vxca_cavern1_haunted_ckb_0_0.vlm"},
	{0xAE75163B9F8BFF0EULL, "639890588.ogg"},
	{0xAE756E3A385D8925ULL, "ai.ncs"},
	{0xAE75B589A5DFC0D1ULL, "vxw_blightlands_f8_0_0.vlm"},
	{0xAE75ED7EB2218208ULL, "art\\levels\\races\\design\\prp_chestb.phy"},
	{0xAE7696AAD5ABE117ULL, "6033315_m.fxe"},
	{0xAE7724D1363C4568ULL, "crt_adult_dragon.wbd"},
	{0xAE77A4C281E75FE0ULL, "gen00fl_isabela6090874.cl"},
	{0xAE78019B3B8E382CULL, "vxlt_gallowstemplar_f_4sl_0_0.vlm"},
	{0xAE78372F1B12529FULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_bdy_0n.dds"},
	{0xAE789492C751A142ULL, "vxht_0_hightown_b8a_0_0.vlm"},
	{0xAE78AFCCD402F6ECULL, "850124129.ogg"},
	{0xAE78CC5E83C86EB3ULL, "4188602.ogg"},
	{0xAE78DD22DEAEB1F8ULL, "dae000cr_grand_cleric6148154.cl"},
	{0xAE7949ADB6D350C2ULL, "612478378.ogg"},
	{0xAE7967A8C8E8D3E4ULL, "lt_0_lowtown_docks_a3_124_0.msh"},
	{0xAE7979A8C8E8F24DULL, "lt_0_lowtown_docks_a3_124_0.mao"},
	{0xAE797D732DE6EB1BULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3_tool.phy"},
	{0xAE797DA8C8E8F886ULL, "lt_0_lowtown_docks_a3_124_0.mmh"},
	{0xAE797EF02D689C52ULL, "hla_ht_0_hightown_-1_0x.dds"},
	{0xAE799F2B105A9D55ULL, "257198548.ogg"},
	{0xAE79AC930D224916ULL, "6095879_m.fxe"},
	{0xAE79ED32C7A21246ULL, "6133747_f.fxe"},
	{0xAE7A5B68046514EDULL, "705872129.ogg"},
	{0xAE7A60FE97CDE8DFULL, "vxw_bonepitdestroyed_a3_29x_0_0.vlm"},
	{0xAE7A920F41C8CD71ULL, "42507894.ogg"},
	{0xAE7A9D32EF9D6292ULL, "408753173.ogg"},
	{0xAE7AA552218E0638ULL, "773050725.ogg"},
	{0xAE7AC982FCC54F88ULL, "hlac_ca_caverns1_mines_24424_1y.dds"},
	{0xAE7ACC5D4DD44488ULL, "vxlt_gallowstemplar_5k6_0_0.vlm"},
	{0xAE7AF3C13EC8417FULL, "6207606_m.fxe"},
	{0xAE7BD8D2EE39C436ULL, "vxca_caverns1_jl1_0_0.vlm"},
	{0xAE7C47B060140FE2ULL, "mag110_arianni6088208.cl"},
	{0xAE7C934642EA63E5ULL, "qun211_party_comment6179254.cl"},
	{0xAE7D2787A8D8DFCBULL, "6110706_m.fxe"},
	{0xAE7D7A01D0BA53CCULL, "6059352_m.fxe"},
	{0xAE7DD36F88B214CDULL, "6036524_f.fxe"},
	{0xAE7E4FFB4EC98A0AULL, "ave200_aveline_bar6116313.cl"},
	{0xAE7EC1F6415E3CF3ULL, "654781130.ogg"},
	{0xAE7EFA3A26A45FBFULL, "6109058_m.fxe"},
	{0xAE7F49BA5D4ABA30ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyn_t.dds"},
	{0xAE7F747B547C0049ULL, "lowtown_fem_prostitute_sta6152339.cl"},
	{0xAE7FCC326A84D743ULL, "29835653.ogg"},
	{0xAE7FD7A29AE308C1ULL, "525598438.ogg"},
	{0xAE807638A56823BAULL, "448528025.ogg"},
	{0xAE8099FDDFFC5F58ULL, "vxht_keep_6zu_0_0.vlm"},
	{0xAE81257F368601EBULL, "179727127.ogg"},
	{0xAE81760F3E62D8DAULL, "fx_st_cloudb.mao"},
	{0xAE8180999319F3B6ULL, "vxw_encounter2_camp_p2_0_0.vlm"},
	{0xAE81DCA7A676A471ULL, "zz_one_debug6051518.cl"},
	{0xAE81FCD04F44AB22ULL, "art\\characters\\playercharacter\\elfmale\\cth_comb.phy"},
	{0xAE820D23BDD2D7B4ULL, "vxlt_0_lowtown_docks_n_b5w_0_1.vlm"},
	{0xAE82599E99AFB03FULL, "mag370_meredith6064953.cl"},
	{0xAE82AF41CF394B88ULL, "plt_lgt220pt_bounty_hunter.nss"},
	{0xAE82C9A36FA3FBE6ULL, "vxw_encounter2_camp_9m_0_0.vlm"},
	{0xAE833F4C39614999ULL, "644458055.ogg"},
	{0xAE8392432DB310D1ULL, "ser221_gascard6149154.cl"},
	{0xAE839DE42F789328ULL, "vxht_keep_n_705_0_0.vlm"},
	{0xAE839E3926A4C49CULL, "856585005.ogg"},
	{0xAE83AB759356BBC3ULL, "vxlt_gallowsprison_n_6o5_0_0.vlm"},
	{0xAE83E8D5B6F47ECBULL, "hlac_ca_cavern1_haunted_24421_10y.dds"},
	{0xAE8414A18691965CULL, "672964220.ogg"},
	{0xAE8459ABAEBBE238ULL, "190086116.ogg"},
	{0xAE84880529E40A77ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.cess.gad"},
	{0xAE84B93C4AC6D755ULL, "hlac_w_sundermount_-327683_0z.dds"},
	{0xAE84F4AAE8F13158ULL, "hlac_ht_keep_3732_11x.dds"},
	{0xAE85633D0AB871D8ULL, "hlac_lt_0_lowtown_docks_-4_0y.dds"},
	{0xAE85BF198E56E2D8ULL, "hlac_dr_primevaltomb_663_2y.dds"},
	{0xAE863B1A77D67D45ULL, "hlac_w_encounter3_0_7y.dds"},
	{0xAE866BD2F5FAD4C6ULL, "vxht_0_hightown_n_caq_0_1.vlm"},
	{0xAE875A04D2476A52ULL, "vxdr_primevalentrance_2ll_0_0.vlm"},
	{0xAE87FCB8FA92E177ULL, "vxht_0_hightown_a3_bxm_0_0.vlm"},
	{0xAE880D2810228E03ULL, "vxdr_primevaltomb_6iw_0_1.vlm"},
	{0xAE88DDB279D5876EULL, "6053049_f.fxe"},
	{0xAE893E3A3FCA5D08ULL, "vxlt_gallowstemplar_5t7_0_0.vlm"},
	{0xAE89915931936F1BULL, "lt_gallowscourtyard.arl"},
	{0xAE899A14B5256F87ULL, "vxca_cavern1_haunted_edf_0_0.vlm"},
	{0xAE89B841CF3F799BULL, "plt_lgt220pt_bounty_hunter.plo"},
	{0xAE8A0A7BD8728CBCULL, "gen_im_arm_cht_nobe_t1.uti"},
	{0xAE8A29137693F078ULL, "hlac_dr_0_deeproads_5842_17y.dds"},
	{0xAE8A4D7718CAA065ULL, "vxlt_undercityencounter_13t_0_1.vlm"},
	{0xAE8A5D2EAF7D9459ULL, "hlac_ht_0_hightown_a2_65535_1y.dds"},
	{0xAE8A987371F379C3ULL, "6080230_m.fxe"},
	{0xAE8B2074B6AA2617ULL, "vxlt_0_lowtown_a2_fek_0_0.vlm"},
	{0xAE8B25D6EF15E368ULL, "426082954.ogg"},
	{0xAE8B9D0529EA4C04ULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.cess.evt"},
	{0xAE8BE797F89574B2ULL, "llac_lt_0_lowtown_docks_-65537_5.dds"},
	{0xAE8C09216BB5AC4FULL, "art\\vfx\\textures\\vfx_distortion_ball.dds"},
	{0xAE8C84E62BCB2662ULL, "mag291_torpor6026989.cl"},
	{0xAE8D909B278C165CULL, "vxw_bonepitdestroyed_a3_1or_0_0.vlm"},
	{0xAE8DA42BB76EB33FULL, "6131192_m.fxe"},
	{0xAE8DC4FED9F985B2ULL, "6117782_m.fxe"},
	{0xAE8DC817324D5748ULL, "243423631.ogg"},
	{0xAE8E0489F3248082ULL, "vxlt_undercitydungeon_1bh_0_0.vlm"},
	{0xAE8E7047360B3C90ULL, "mage_glyphrepulsion.dds"},
	{0xAE8F0BDAD1F277D8ULL, "vxht_keep_6gr_0_0.vlm"},
	{0xAE8FA4FE991EED3AULL, "ave200_donnic6112138.cl"},
	{0xAE8FB7F5D6CA98D3ULL, "llac_lt_0_lowtown_docks_a3_-2_3.dds"},
	{0xAE8FDF800CA0F543ULL, "vxw_encounter3_fg_0_1.vlm"},
	{0xAE8FF27685100C36ULL, "vxw_encounter2_camp_k3_0_0.vlm"},
	{0xAE906E848C2AF79FULL, "mag291_torpor6027053.cl"},
	{0xAE90AA12BB920DBEULL, "hlac_dr_0_deeproads_5843_7y.dds"},
	{0xAE90CDF872963B5DULL, "6095865_m.fxe"},
	{0xAE90D0C3BE91B80EULL, "vxlt_0_lowtown_n_f06_0_0.vlm"},
	{0xAE91341D4809ECFFULL, "lla_ht_0_hightown_n_-131071_0.dds"},
	{0xAE913B80619A1466ULL, "vxw_encounter2_camp_4n_0_0.vlm"},
	{0xAE9150AA6E5F767BULL, "40670651.ogg"},
	{0xAE91D90B32B06EA1ULL, "561124150.ogg"},
	{0xAE91DE198C9A8FCFULL, "gen00fl_aveline6114873.cl"},
	{0xAE921D52854CD443ULL, "vxlt_gallowsprison_n_6h4_0_0.vlm"},
	{0xAE924BE15654BF84ULL, "6195863_m.fxe"},
	{0xAE9267BAAB1C6B96ULL, "gen00fl_aveline6140044.cl"},
	{0xAE92D5C24570F093ULL, "167996281.ogg"},
	{0xAE92F177D7177FCFULL, "hlac_lt_0_lowtown_docks_a3_n_131070_1x.dds"},
	{0xAE931A26561517CEULL, "vxdr_primevaltomb_4i0_0_1.vlm"},
	{0xAE93B1244137E09DULL, "1061616099.ogg"},
	{0xAE94382E24798BC4ULL, "632422980.ogg"},
	{0xAE94F5E88607F0AAULL, "54398195.ogg"},
	{0xAE95707F221ACB25ULL, "vxlt_0_lowtown_docks_aqk_0_0.vlm"},
	{0xAE95AD9AE7DBEEA7ULL, "vxlt_hangedman_n_5r_0_0.vlm"},
	{0xAE95FF8331D498B4ULL, "dae000_pc_home_ambients6208178.cl"},
	{0xAE96B4D0263EAB74ULL, "gen00fl_aveline6129032.cl"},
	{0xAE96C1413629BEA0ULL, "zz_ser_debug6078572.cl"},
	{0xAE96D70D6B945856ULL, "vxca_caverns1_mines_e4k_0_0.vlm"},
	{0xAE972D906B817C4EULL, "plt_prm000pt_cliffs_tome.plo"},
	{0xAE9730ECDC1F2CF8ULL, "80712769.ogg"},
	{0xAE973421B8C9C487ULL, "vxht_chantry_n_2hw_0_0.vlm"},
	{0xAE973D2A2FA5FBA0ULL, "plt_cod_cha_isabela2.plo"},
	{0xAE977BC6252F9D24ULL, "vxlt_gallowstemplar_f_2je_0_0.vlm"},
	{0xAE978FFCADB61F94ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.cess.ani"},
	{0xAE9862422435DAFDULL, "734860320.ogg"},
	{0xAE98B50529F4FFFDULL, "art\\vfx\\combat\\vfx_w_def_def_mode_loop_c.cess.ani"},
	{0xAE9915AD1C903916ULL, "isa311_isabela6092148.cl"},
	{0xAE999412CB4EDEF3ULL, "6030633_m.fxe"},
	{0xAE999D7BD25A3043ULL, "qun160cr_vashothlead.utc"},
	{0xAE99C7C6750C7374ULL, "697112143.ogg"},
	{0xAE9A06A05964FA64ULL, "llac_lt_0_lowtown_a2_n_c_65538_5.dds"},
	{0xAE9A0FF4559D9A46ULL, "ser210_witness6168154.cl"},
	{0xAE9A1435B7E7E02FULL, "l_env_wal_lowtown_cough.bnk"},
	{0xAE9A788100192964ULL, "w_bonepit_187_0.msh"},
	{0xAE9A8A81001947CDULL, "w_bonepit_187_0.mao"},
	{0xAE9A8E8100194E06ULL, "w_bonepit_187_0.mmh"},
	{0xAE9AA6F6BC9E8A6EULL, "6107181_m.fxe"},
	{0xAE9BC8F9F6C457D3ULL, "zz_end_choose.cut"},
	{0xAE9BE60326DA5B16ULL, "180607574.ogg"},
	{0xAE9BF0F0EAA7784EULL, "vxlt_undercitydungeon_16x_0_0.vlm"},
	{0xAE9C108EE767A2F1ULL, "vxlt_hangedman_2mw_0_0.vlm"},
	{0xAE9C7BFF1D2D64AAULL, "642470085.ogg"},
	{0xAE9CA3D32401C579ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_1y.dds"},
	{0xAE9CA530DBAB2E06ULL, "58988750.ogg"},
	{0xAE9D20F77CFB006AULL, "hlac_lt_backalley_g_65536_6y.dds"},
	{0xAE9DA2968C524C4FULL, "6085844_m.fxe"},
	{0xAE9DCC320C7CA857ULL, "6171471_m.fxe"},
	{0xAE9DE5D0DA15C0B5ULL, "vxlt_gallowscourtyard_n_7fw_0_0.vlm"},
	{0xAE9DECFEB29A9DDDULL, "llac_lt_0_lowtown_a2_131074_0.dds"},
	{0xAE9DEE40D5F6EB08ULL, "192326193.ogg"},
	{0xAE9E161C65AF86BCULL, "vxlt_0_lowtown_a2_e8d_0_0.vlm"},
	{0xAE9E569B025DA9E2ULL, "1018838054.ogg"},
	{0xAE9E5E065560E7FDULL, "6045856_f.fxe"},
	{0xAE9E5E0C01D91D8BULL, "qun111_ketojan6019087.cl"},
	{0xAE9E6453770624B6ULL, "vxw_encounter2_camp_f4_0_0.vlm"},
	{0xAE9E74AA4078CBE5ULL, "vxlt_0_lowtown_a3_n_c_gjq_0_0.vlm"},
	{0xAE9E8CFCADBC39F6ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.cess.gad"},
	{0xAE9E8D021DF16F0AULL, "ser131ip_to_darktown_2.utp"},
	{0xAE9E9C061D52BDDEULL, "6101392_m.fxe"},
	{0xAE9F08D0841BE10CULL, "6109211_m.fxe"},
	{0xAE9F2152645F856BULL, "137055222.ogg"},
	{0xAE9F4F0F3E7B938CULL, "fx_st_cloudb.dds"},
	{0xAE9F531E96669400ULL, "mag291_vincento6027663.cl"},
	{0xAE9F549679E2C780ULL, "6117726_m.fxe"},
	{0xAE9F6DDF39BA4F64ULL, "410748059.ogg"},
	{0xAE9F7651810FD2B7ULL, "634543231.ogg"},
	{0xAE9FB514F175EA3CULL, "vxw_encounter3_ax_0_0.vlm"},
	{0xAEA018BFF85C9E6CULL, "zz_ave_debug6062236.cl"},
	{0xAEA02C1A66810277ULL, "422996559.ogg"},
	{0xAEA08E1CD8A287BCULL, "vxca_caverns1_in2_0_0.vlm"},
	{0xAEA0B8F094D20839ULL, "vxca_caverns2prc_92x_0_1.vlm"},
	{0xAEA0CC4C788CF7DBULL, "482657186.ogg"},
	{0xAEA0D9E0788C69D4ULL, "6119603_m.fxe"},
	{0xAEA1174DB894552AULL, "hlac_ca_caverns2prc_5802_55z.dds"},
	{0xAEA1384FEE1B7657ULL, "vxw_sundermount_240_0_0.vlm"},
	{0xAEA146CF889AC44DULL, "hlac_lt_0_lowtown_a3_n_c_131072_5y.dds"},
	{0xAEA162A081D6FD14ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180l.ani"},
	{0xAEA1FD6C32D48E16ULL, "llac_ca_caverns2_5802_23.dds"},
	{0xAEA2A839DCD537FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_help_up_goodbye_malepc.ani"},
	{0xAEA2F1783D71BED6ULL, "gen00fl_sebastian6179914.cl"},
	{0xAEA3462DCB6A2FF0ULL, "6092137_m.fxe"},
	{0xAEA3CDAD5423F5F3ULL, "lt_0_lowtown_docks_n_117_0.phy"},
	{0xAEA3F4D184DCD9AAULL, "217458575.ogg"},
	{0xAEA40BB3297CF7EDULL, "lowtown_male_alienage1_sta6162782.cl"},
	{0xAEA4373EF91797F3ULL, "638732141.ogg"},
	{0xAEA43E2E10AF02B2ULL, "vxca_caverns1_mines_gkd_0_0.vlm"},
	{0xAEA48581002193D1ULL, "w_bonepit_187_0.phy"},
	{0xAEA4E01015E09AE8ULL, "fen210_fenris6109415.cl"},
	{0xAEA4EB12F5A5F430ULL, "vxca_cavern1_haunted_j59_0_0.vlm"},
	{0xAEA4F7DBA015E5CCULL, "mag101_alain6032130.cl"},
	{0xAEA503CB57C8F428ULL, "hlac_lt_gallowstemplar_n_2394_5z.dds"},
	{0xAEA580BCFE731901ULL, "932163326.ogg"},
	{0xAEA5A5FEAABFDD07ULL, "vxht_chantry_n_2cv_0_0.vlm"},
	{0xAEA5A7FCADC28655ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_repul_p.cess.evt"},
	{0xAEA5C1A6C4829148ULL, "gen_im_trs_dal_med_cht_r0.uti"},
	{0xAEA5FB02E6E380C0ULL, "arial12.fnt.xml"},
	{0xAEA664732E0DAEBAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3_tool.msh"},
	{0xAEA672732E0DC6F3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3_tool.mao"},
	{0xAEA676732E0DCD00ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3_tool.mmh"},
	{0xAEA68451CD6ADBBBULL, "6101447_m.fxe"},
	{0xAEA709BB46822B48ULL, "hlac_lt_backalley_n_0_9x.dds"},
	{0xAEA7A39BE29E92D5ULL, "vxca_caverns1_e7x_0_0.vlm"},
	{0xAEA7D7D04F653BD6ULL, "art\\characters\\playercharacter\\elfmale\\cth_comb.mao"},
	{0xAEA7E3D04F655075ULL, "art\\characters\\playercharacter\\elfmale\\cth_comb.mmh"},
	{0xAEA7E5D04F6553EBULL, "art\\characters\\playercharacter\\elfmale\\cth_comb.msh"},
	{0xAEA839EA84BE9820ULL, "6150709_m.fxe"},
	{0xAEA85FA081DD1776ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180l.gad"},
	{0xAEA8C8FE1DBE0A0DULL, "dae000_bodhan6129084.cl"},
	{0xAEA986B52967DA7DULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_005a_0d.dds"},
	{0xAEA9B539DCDB6C3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_help_up_goodbye_malepc.gad"},
	{0xAEA9F8C515E3D3B2ULL, "follower_banter6171116.cl"},
	{0xAEAA7C126D6ECE50ULL, "hlac_lt_0_lowtown_a2_n_c_0_5y.dds"},
	{0xAEAA8FE0428FB435ULL, "vxht_mansion2_3gb_0_0.vlm"},
	{0xAEAAEDE28FC44F43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_petrify.evt"},
	{0xAEAB1E906B920BDDULL, "plt_prm000pt_cliffs_tome.nss"},
	{0xAEAB238CDDF80213ULL, "vxlt_0_lowtown_df0_0_0.vlm"},
	{0xAEAB3CC6BC59CBB9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_monique.gad"},
	{0xAEAB5067E4825A79ULL, "gen00fl_aveline6137568.cl"},
	{0xAEAB5DD4CA2D4BDDULL, "6129281_m.fxe"},
	{0xAEACAB914E2C9866ULL, "ser191_emeric6107887.cl"},
	{0xAEACB74F3B314651ULL, "hlac_w_woundedcoast_131074_0z.dds"},
	{0xAEACD63068FC3D36ULL, "vxw_encounter2_camp_a5_0_0.vlm"},
	{0xAEAD15114DD349DFULL, "llac_w_bonepitdestroyed_a3_393219_0.dds"},
	{0xAEAD5020EDC73EEDULL, "hlac_ca_caverns2prc_5802_49x.dds"},
	{0xAEAD6C1F19F332DEULL, "6062077_m.fxe"},
	{0xAEADD2FE617185D4ULL, "6105999_f.fxe"},
	{0xAEADDB09E7A2FF87ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_p.impa.gad"},
	{0xAEADE7D15F9CDEEEULL, "gen00fl_merrill6128713.cl"},
	{0xAEAE145423D98475ULL, "vxdr_primevaltomb_7zu_0_1.vlm"},
	{0xAEAE934F73B1078BULL, "vxlt_0_lowtown_a3_n_c_efi_0_0.vlm"},
	{0xAEAF65F881579686ULL, "6201410_m.fxe"},
	{0xAEAF7AA081E363D5ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180l.evt"},
	{0xAEAFAA2CE0118ED7ULL, "vxw_sundermount_293_0_0.vlm"},
	{0xAEB00E0E6FAF2264ULL, "vxlt_gallowstemplar_n_5uv_0_1.vlm"},
	{0xAEB01F308E4C0662ULL, "pro000_athenril6049455.cl"},
	{0xAEB02AD73F376B7CULL, "dre000st_varricrtrnsurface.stg"},
	{0xAEB0366C063B445BULL, "vxlt_0_lowtown_a2_n_e26_0_0.vlm"},
	{0xAEB083C1CA6EB903ULL, "cod_hst_kirk_darktown.utp"},
	{0xAEB09039DCE14C5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sister_help_up_goodbye_malepc.evt"},
	{0xAEB0EA4D73973671ULL, "hlac_ht_0_hightown_n_-131070_0x.dds"},
	{0xAEB177761F54253EULL, "hlac_w_bonepitdestroyed_a3_262150_0z.dds"},
	{0xAEB18D71ED409FC4ULL, "vxca_caverns1_iic_0_0.vlm"},
	{0xAEB1C6E28FCA2BC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_petrify.gad"},
	{0xAEB224F2C2E91BCFULL, "vxw_bonepit_1rz_0_0.vlm"},
	{0xAEB24DC6BC600707ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_monique.ani"},
	{0xAEB3B388B0EDC35AULL, "lt_0_lowtown_docks_a3_n_105_0.mmh"},
	{0xAEB3B788B0EDC991ULL, "lt_0_lowtown_docks_a3_n_105_0.mao"},
	{0xAEB3C588B0EDE1A0ULL, "lt_0_lowtown_docks_a3_n_105_0.msh"},
	{0xAEB3DBC96ED948E8ULL, "llac_w_encounter2_camp_65537_0.dds"},
	{0xAEB408BBAA1EE01EULL, "hlac_dr_0_deeproads_5842_10z.dds"},
	{0xAEB40E1F8C6C4005ULL, "fx_snowcloud.mao"},
	{0xAEB4647F51E2B44FULL, "vxht_bartrand_n_681_0_1.vlm"},
	{0xAEB4700E10FD1EC9ULL, "216278700.ogg"},
	{0xAEB472D94B9B4628ULL, "hlac_w_sundermount_-131078_0x.dds"},
	{0xAEB487291A5CBE54ULL, "vxw_bonepit_26v_0_0.vlm"},
	{0xAEB4D009E7A90B94ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_p.impa.evt"},
	{0xAEB5280F8B35B0D6ULL, "vxlt_hangedman_28c_0_0.vlm"},
	{0xAEB52843619F7339ULL, "vxlt_0_lowtown_docks_cog_0_0.vlm"},
	{0xAEB5A765A7D21638ULL, "qcr220_grey_wardens6135938.cl"},
	{0xAEB62E21BF79D991ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning1sail.msh"},
	{0xAEB63C21BF79F1ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning1sail.mao"},
	{0xAEB64821BF7A064FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_awning1sail.mmh"},
	{0xAEB67329A7048EB1ULL, "974617559.ogg"},
	{0xAEB679766D2EC490ULL, "follower_banter6136670.cl"},
	{0xAEB6A81456AC75AAULL, "vxw_bonepitdestroyed_a3_p6_0_0.vlm"},
	{0xAEB6AD07185F273AULL, "vxlt_hangedman_2xo_0_0.vlm"},
	{0xAEB743C9C6DA9A7DULL, "ht_0_hightown_a3.rml"},
	{0xAEB749D78DD542D6ULL, "6125527_m.fxe"},
	{0xAEB76870A81154A2ULL, "6083605_m.fxe"},
	{0xAEB76B80027FAF41ULL, "ser221_party_comments6111580.cl"},
	{0xAEB7D5E28FCEB0A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_petrify.ani"},
	{0xAEB8246A69747800ULL, "lgt000cr_dwf_turn_in_2.utc"},
	{0xAEB8483E695BFC42ULL, "vxlt_warehouse_n_1bm_0_0.vlm"},
	{0xAEB8A4DED25CD4DBULL, "hlac_ca_caverns1_mines_24422_7z.dds"},
	{0xAEB8C5815B4AF6D0ULL, "354673602.ogg"},
	{0xAEB8F468E0C57D84ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_wep_d.dds"},
	{0xAEB9089ED5595A8EULL, "dae000ip_crowd_mf6_merchant.utp"},
	{0xAEB917E0FC37E36CULL, "dre000_dougal6129732.cl"},
	{0xAEB9B02DF4F5A729ULL, "326063632.ogg"},
	{0xAEBA3DE57CCE407EULL, "hlac_lt_darktown_-1_86x.dds"},
	{0xAEBA5D5D396A2A8DULL, "65732426.ogg"},
	{0xAEBAB73637798B53ULL, "1063129597.ogg"},
	{0xAEBAD980AF6E0305ULL, "6113077_m.fxe"},
	{0xAEBB9A263B56BEAFULL, "vxw_encounter2_cv_0_0.vlm"},
	{0xAEBBA4DA1A12B222ULL, "dre000_bodahn6028156.cl"},
	{0xAEBC3262B8FB6AF1ULL, "6083758_m.fxe"},
	{0xAEBC4CBB36A6DE12ULL, "deadspace_sirissac.dds"},
	{0xAEBD2BC0A56DD91FULL, "vxw_bonepitmining_a2_5j_0_0.vlm"},
	{0xAEBD525AAA3CFFE9ULL, "plc_door01.utp"},
	{0xAEBDB5EE64DB4A78ULL, "496282971.ogg"},
	{0xAEBE13F982D4F182ULL, "vxlt_0_lowtown_a3_n_c_hja_0_0.vlm"},
	{0xAEBE22993BDBA8F4ULL, "weapon2_ds1.psh.fxo"},
	{0xAEBE43C358B6426EULL, "412326525.ogg"},
	{0xAEBE4AD13678882AULL, "crt_adult_dragon.bnk"},
	{0xAEBE4F574427D51EULL, "gen00fl_fenris6127994.cl"},
	{0xAEBE5F842D59F24FULL, "771626767.ogg"},
	{0xAEBE66C096B640F9ULL, "g_cl.mao"},
	{0xAEBEA5D064DEDE8CULL, "llac_lt_0_lowtown_a2_n_65533_0.dds"},
	{0xAEBF22FA748A8E15ULL, "gen00fl_sebastian6174040.cl"},
	{0xAEBF5C53D6CA6CA5ULL, "544875493.ogg"},
	{0xAEBF65FA3219C349ULL, "mer320_hubert6051999.cl"},
	{0xAEBFB94E4C019826ULL, "vxlt_gallowstemplar_4qg_0_0.vlm"},
	{0xAEC02B8E289F9B0DULL, "6208191_m.fxe"},
	{0xAEC0E37957F04982ULL, "vxw_encounter2_r4_0_0.vlm"},
	{0xAEC194F93E43F32DULL, "215140385.ogg"},
	{0xAEC2080077519159ULL, "21356028.ogg"},
	{0xAEC23205EC5DD73AULL, "vxlt_hangedman_140_0_0.vlm"},
	{0xAEC282AA9E6357AFULL, "vxlt_undercitydungeon_1g6_0_0.vlm"},
	{0xAEC2955533A6F9D9ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgecrgrck.msh"},
	{0xAEC29F5533A70AC7ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgecrgrck.mmh"},
	{0xAEC2A35533A71114ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgecrgrck.mao"},
	{0xAEC2C71523B3E721ULL, "388665791.ogg"},
	{0xAEC2C809E7B53BCDULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_f_p.impa.ani"},
	{0xAEC2EF5EF28DA560ULL, "llac_ht_chantry_2534_47.dds"},
	{0xAEC365AFEF2120E3ULL, "lt_gallowstemplar_n_2394.rml"},
	{0xAEC37436E8F6B72BULL, "gen_im_trs_slv_junk2_03.uti"},
	{0xAEC3765931C4B9B1ULL, "lt_gallowscourtyard.rml"},
	{0xAEC380E797B5C88FULL, "vxca_cavern1_haunted_e8j_0_0.vlm"},
	{0xAEC399EC7D2BC956ULL, "vxlt_hangedman_25d_0_0.vlm"},
	{0xAEC45F6315E7AC6CULL, "gen00fl_fenris6126960.cl"},
	{0xAEC468F2246F15D1ULL, "genip_crate.utp"},
	{0xAEC4735E7E43BCD0ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.mmh"},
	{0xAEC49C9A1BC63746ULL, "zz_ave_debug6056790.cl"},
	{0xAEC4EC00B1FD95CEULL, "mag360cr_thrask.utc"},
	{0xAEC4FCED2F7A4C8BULL, "vxdr_0_deeproads_bcc_0_1.vlm"},
	{0xAEC507B4CEC1E178ULL, "vrc221_varric6175229.cl"},
	{0xAEC5193A9CFAEA63ULL, "llac_lt_undercityserialkiller_-1_45.dds"},
	{0xAEC58CB5C656B48FULL, "art\\vfx\\combat\\vfx_w_fns_fenris2_aoe_p.mmh"},
	{0xAEC5DA4D9A076D22ULL, "6086905_m.fxe"},
	{0xAEC61965DDEB286BULL, "vxlt_0_lowtown_docks_a3_cwz_0_0.vlm"},
	{0xAEC63DAC3EB94822ULL, "6141284_m.fxe"},
	{0xAEC6992C1DB4CF5CULL, "6027861_m.fxe"},
	{0xAEC6ABFA6BB2DCA1ULL, "6167496_m.fxe"},
	{0xAEC6ADF3B896C52BULL, "289402569.ogg"},
	{0xAEC6D8B31271F923ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01a.mmh"},
	{0xAEC6E6B31272117DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01a.msh"},
	{0xAEC7B1685508E1D8ULL, "hlac_lt_gallowscourtyard_-131073_1y.dds"},
	{0xAEC84488D63CD443ULL, "vxht_mansion_n_4a6_0_0.vlm"},
	{0xAEC87AC8813ED75FULL, "gen_im_trs_dwv_dua_dag_r2.uti"},
	{0xAEC8C9B73E1F994DULL, "and321st_chantry.stg"},
	{0xAECA0A800F3563D9ULL, "dae000_orsino6085490.cl"},
	{0xAECA0C032D4CD72FULL, "vxw_encounter2_nq_0_0.vlm"},
	{0xAECAA85D7FB47393ULL, "6102679_m.fxe"},
	{0xAECACB545A9F0AE2ULL, "custic00c.dds"},
	{0xAECB146F57DF4D23ULL, "vxlt_0_lowtown_docks_n_c0c_0_0.vlm"},
	{0xAECC4ADF869012FBULL, "art\\characters\\playercharacter\\humanmale\\boo_maga.msh"},
	{0xAECC5CDF86903146ULL, "art\\characters\\playercharacter\\humanmale\\boo_maga.mao"},
	{0xAECC68DF869045E5ULL, "art\\characters\\playercharacter\\humanmale\\boo_maga.mmh"},
	{0xAECCB49E8BD4015FULL, "vxlt_0_lowtown_cmv_0_0.vlm"},
	{0xAECD034C871DA66FULL, "74291808.ogg"},
	{0xAECD362E3E7B160EULL, "combat_impact_1.gda"},
	{0xAECDC1C1DB005D3DULL, "140036171.ogg"},
	{0xAECDEC73B79D09A8ULL, "132538778.ogg"},
	{0xAECDF143D5253A4DULL, "atl_deathscree_dxt5_00.dds"},
	{0xAECDFB0FB6524AEAULL, "418045019.ogg"},
	{0xAECE6A4E0A0D9698ULL, "gen00fl_anders6111353.cl"},
	{0xAECF4442D77A3A48ULL, "379978995.ogg"},
	{0xAECF50CDA665503BULL, "mrl250_pol6101795.cl"},
	{0xAECF637A91018A0FULL, "6111319_f.fxe"},
	{0xAECFBD6E5617854BULL, "289734635.ogg"},
	{0xAECFE1C81D921E74ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingb_1.mao"},
	{0xAECFE755B886A975ULL, "624937255.ogg"},
	{0xAECFF2A42FF09B91ULL, "836284732.ogg"},
	{0xAECFFB2DDA13CBEDULL, "6167359_m.fxe"},
	{0xAED0074B1F1ADB72ULL, "6039957_f.fxe"},
	{0xAED06CC92EAEF207ULL, "6122080_m.fxe"},
	{0xAED0FCA28078EE29ULL, "cs_dre180_db_rock_mage.wbd"},
	{0xAED124384DE42379ULL, "6098686_m.fxe"},
	{0xAED1965E7E4E82EFULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.anb"},
	{0xAED19A514120A1D8ULL, "1007729486.ogg"},
	{0xAED1EBF07413320CULL, "qcr000_arishok6060189.cl"},
	{0xAED241EE86F9488AULL, "6105428_m.fxe"},
	{0xAED2AE6354762264ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_merr.phy"},
	{0xAED2DE79CAD63C06ULL, "vxca_caverns1_mines_in9_0_0.vlm"},
	{0xAED311BC9D344373ULL, "follower_banter6161019.cl"},
	{0xAED34F806EF69BFDULL, "combat_2h_r3_a1.gda"},
	{0xAED3542531B767F9ULL, "gen00fl_isabela6212002.cl"},
	{0xAED35C7C18C720E7ULL, "hlac_ht_keep_3731_1y.dds"},
	{0xAED398CE797DB037ULL, "6125851_f.fxe"},
	{0xAED3FE1E21A6A4B7ULL, "lt_0_lowtown_docks_238_0.msh"},
	{0xAED40143C554ED05ULL, "6080760_m.fxe"},
	{0xAED4101E21A6C31AULL, "lt_0_lowtown_docks_238_0.mao"},
	{0xAED4141E21A6CA29ULL, "lt_0_lowtown_docks_238_0.mmh"},
	{0xAED45E52691D39EEULL, "art\\characters\\playercharacter\\humanfemale\\cth_comb.msh"},
	{0xAED46692686E4CDBULL, "vxca_caverns1_mines_j5v_0_0.vlm"},
	{0xAED46C52691D513FULL, "art\\characters\\playercharacter\\humanfemale\\cth_comb.mao"},
	{0xAED476846AF4F295ULL, "713108650.ogg"},
	{0xAED47852691D65DCULL, "art\\characters\\playercharacter\\humanfemale\\cth_comb.mmh"},
	{0xAED498757F0A24EAULL, "593701179.ogg"},
	{0xAED4C934F9EF5D80ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyn_d.dds"},
	{0xAED51C50B8987D66ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_enter.ani"},
	{0xAED594D72FDFA2CAULL, "ave310cr_guard_m.utc"},
	{0xAED63CFCF2B9917AULL, "53653055.ogg"},
	{0xAED6EFD9FC0C7527ULL, "87391283.ogg"},
	{0xAED71ED6F1FC5FFCULL, "ts_coterie_01.gda"},
	{0xAED73AF8E8B3231AULL, "hlac_lt_0_lowtown_docks_-131073_4z.dds"},
	{0xAED77B01BDB7D1D1ULL, "459181450.ogg"},
	{0xAED7D9B312806E48ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos01a.phy"},
	{0xAED8314AD9D42EECULL, "95190532.ogg"},
	{0xAED866FF2506EE00ULL, "vxht_0_hightown_a2_c5y_0_0.vlm"},
	{0xAED87DE01F42EFAFULL, "vxw_encounter2_ip_0_0.vlm"},
	{0xAED896C9A042B7A8ULL, "vxlt_0_lowtown_a2_n_c_fiy_0_0.vlm"},
	{0xAED8BBBEDBB55EB5ULL, "plt_cod_itm_bardins_folly.plo"},
	{0xAED8BF396CA43439ULL, "follower_banter6155946.cl"},
	{0xAED9B3F533532C1BULL, "cs_mcr370cs_player_flees.bnk"},
	{0xAED9C422A1D1C601ULL, "dae000ip_warm_hands_d.utp"},
	{0xAED9D288B10EC765ULL, "lt_0_lowtown_docks_a3_n_105_0.phy"},
	{0xAED9E8E64993510BULL, "vxlt_0_lowtown_a3_n_c_ewj_0_0.vlm"},
	{0xAED9FCD4C4AC63BAULL, "846061654.ogg"},
	{0xAEDA46811E8A1720ULL, "zz_vault_debug6203310.cl"},
	{0xAEDA7D2970E2D28FULL, "364078092.ogg"},
	{0xAEDA969A91F197D5ULL, "6064964_m.fxe"},
	{0xAEDB6183DA64F67BULL, "795627784.ogg"},
	{0xAEDB72E7330036B1ULL, "37270848.ogg"},
	{0xAEDB83A7BE3A2028ULL, "vxlt_0_lowtown_a2_n_c_e81_0_0.vlm"},
	{0xAEDBF2127C192E8DULL, "32715292.ogg"},
	{0xAEDC6126B8A509E4ULL, "hf_talwain.mor"},
	{0xAEDC80B870798071ULL, "598658408.ogg"},
	{0xAEDE592F46B858DCULL, "6064640_m.fxe"},
	{0xAEDEB508DEC48996ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_inkwell_d.dds"},
	{0xAEDEC6A60D500007ULL, "vxlt_0_lowtown_a2_n_c_h1x_0_0.vlm"},
	{0xAEDF8D101B4C692FULL, "6037086_f.fxe"},
	{0xAEDF90D90324B2A7ULL, "820541117.ogg"},
	{0xAEE0032BC0684E54ULL, "411852900.ogg"},
	{0xAEE01AF034417A80ULL, "193633203.ogg"},
	{0xAEE098E2BCDAC43CULL, "vxht_0_hightown_n_bka_0_0.vlm"},
	{0xAEE09DEABACDC288ULL, "106188754.ogg"},
	{0xAEE10BAADED838DAULL, "350235471.ogg"},
	{0xAEE1B545AEFA36C6ULL, "877644565.ogg"},
	{0xAEE25C095541576AULL, "6084495_m.fxe"},
	{0xAEE25FB2557C6C22ULL, "728767306.ogg"},
	{0xAEE28F3D565F5DDEULL, "hlac_lt_hangedman_-1_15y.dds"},
	{0xAEE2B1DB50B305A0ULL, "6019889_f.fxe"},
	{0xAEE2F6F83C6E3308ULL, "vxlt_0_lowtown_elj_0_2.vlm"},
	{0xAEE30450B8A491D7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_enter.evt"},
	{0xAEE3684C7913A7D8ULL, "vxlt_0_lowtown_e84_0_0.vlm"},
	{0xAEE3EEC85F1A2309ULL, "6197976_m.fxe"},
	{0xAEE4FF621976AED7ULL, "gencr_mercenary_archer.utc"},
	{0xAEE5F2C2AB8EB963ULL, "hlac_w_encounter3_0_6x.dds"},
	{0xAEE618BEC3BB40D2ULL, "z_dust3.dds"},
	{0xAEE6BD86271BF919ULL, "6083119-mages_killed.cut"},
	{0xAEE6D8DC16FD0680ULL, "vxht_0_hightown_a2_c0f_0_0.vlm"},
	{0xAEE6F35840DA19EEULL, "vxw_bonepit_ck_0_0.vlm"},
	{0xAEE7369072A74CCFULL, "dae000_party_night_comment6201626.cl"},
	{0xAEE7D8374797E285ULL, "22661479.ogg"},
	{0xAEE7E0F6029552E7ULL, "vxlt_darktown_6bu_0_0.vlm"},
	{0xAEE832EC5C003998ULL, "hlac_dr_0_deeproads_5840_1x.dds"},
	{0xAEE866219DB13313ULL, "152103755.ogg"},
	{0xAEE87D8FB3D4A515ULL, "ts_profane_wpn_props.gda"},
	{0xAEE88B9836789DA9ULL, "112820102.ogg"},
	{0xAEE88F297F2C7513ULL, "vxw_bonepit_1q7_0_0.vlm"},
	{0xAEE8CC72D1DAFF5AULL, "and100_anders6173779.cl"},
	{0xAEE8E15281734104ULL, "zz_approval_debug6003610.cl"},
	{0xAEE9036D1EED5DD0ULL, "w_blightlands_209_0.msh"},
	{0xAEE9116D1EED75EAULL, "w_blightlands_209_0.mmh"},
	{0xAEE9156D1EED7C21ULL, "w_blightlands_209_0.mao"},
	{0xAEE9512C71307A2CULL, "gen_im_trs_gen_one_lsw_r2.uti"},
	{0xAEE98FE59CA6685CULL, "6121135_f.fxe"},
	{0xAEE9F6ED2AB48D40ULL, "llac_lt_backalley_g_65536_5.dds"},
	{0xAEEA0550B8AAB260ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_enter.gad"},
	{0xAEEA42C9C705F195ULL, "ht_0_hightown_a3.anb"},
	{0xAEEA46C9C705F8AFULL, "ht_0_hightown_a3.arl"},
	{0xAEEA58F16DD42443ULL, "llac_w_sundermount_-262139_0.dds"},
	{0xAEEA5E4A5D07DA6AULL, "6123356_m.fxe"},
	{0xAEEAD61D57E13252ULL, "framebuffereffects_ps3downsamplemsaa_p0.psh.fxo"},
	{0xAEEB82FD21CEB5C2ULL, "vxdr_primevalentrance_1m4_0_0.vlm"},
	{0xAEEBAC8F1F9706CEULL, "follower_barks6153627.cl"},
	{0xAEED0DB438E49969ULL, "703810260.ogg"},
	{0xAEEDFF2DAC68A2A3ULL, "188873206.ogg"},
	{0xAEEE3FE73B62EF4DULL, "757907619.ogg"},
	{0xAEEE4C9A5C482CFCULL, "vxht_mansion_n_5ap_0_0.vlm"},
	{0xAEEED67E6D98120FULL, "vxca_cavern1_haunted_e7w_0_0.vlm"},
	{0xAEEF461E1712A916ULL, "6060743_m.fxe"},
	{0xAEEF885533CDD1FCULL, "art\\levels\\races\\natural\\highlands\\props\\prp_02hgecrgrck.phy"},
	{0xAEEF96A8FE37AB4DULL, "vxht_mansion2_6r_0_0.vlm"},
	{0xAEEFB03EC754E3EEULL, "hlac_lt_undercityencounter_-1_32y.dds"},
	{0xAEF06F881E84D7AAULL, "vxw_bonepitdestroyed_a3_l2_0_0.vlm"},
	{0xAEF0747AE037893AULL, "vxlt_hangedman_2dk_0_0.vlm"},
	{0xAEF0D9F6D5CC52C9ULL, "348727471.ogg"},
	{0xAEF0F65A4D09B687ULL, "mag210_arianni6152624.cl"},
	{0xAEF10836798AC7B8ULL, "6126465_m.fxe"},
	{0xAEF12D994FACA67EULL, "ave310_donnic6134778.cl"},
	{0xAEF17D012D633C13ULL, "650700933.ogg"},
	{0xAEF1DA296B09C058ULL, "vxlt_0_lowtown_e35_0_0.vlm"},
	{0xAEF1EC93EDC453AFULL, "hlac_ht_chantry_n_2534_47x.dds"},
	{0xAEF2652A7DE0161BULL, "895699258.ogg"},
	{0xAEF2C7FEE4FD66D9ULL, "hlac_lt_0_lowtown_docks_n_-196609_1y.dds"},
	{0xAEF2D007EB411A8BULL, "700523167.ogg"},
	{0xAEF2F0E413FA22A3ULL, "one000_ilen6057354.cl"},
	{0xAEF33B5852AA012DULL, "vxlt_backalley_g_1xe_0_1.vlm"},
	{0xAEF34B9397334E54ULL, "6162080_f.fxe"},
	{0xAEF34BABE15DDD72ULL, "474208997.ogg"},
	{0xAEF3CDC4B7C07338ULL, "vxlt_0_lowtown_a2_n_e1x_0_0.vlm"},
	{0xAEF3FAC6047A4586ULL, "vxlt_0_lowtown_a2_n_e16_0_0.vlm"},
	{0xAEF467CFF95C80A5ULL, "317696212.ogg"},
	{0xAEF478AFE53D173EULL, "vxdr_0_deeproads_fi9_0_0.vlm"},
	{0xAEF481C1DC332E06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_enter.evt"},
	{0xAEF4870D9EF9DE1EULL, "zz_eric_soundtst2.cut"},
	{0xAEF5CEA28097BC0BULL, "cs_dre180_db_rock_mage.bnk"},
	{0xAEF60EA4FAD35FBFULL, "hlac_ht_0_hightown_a2_2_0z.dds"},
	{0xAEF61A7F295C2F17ULL, "hlac_w_bonepitdestroyed_a3_131073_0x.dds"},
	{0xAEF65E787BA59D34ULL, "plt_cod_cha_bethany.plo"},
	{0xAEF679FE3F2F7512ULL, "hlac_ht_0_hightown_a2_65538_0z.dds"},
	{0xAEF67A5A070D9AFCULL, "426524294.ogg"},
	{0xAEF6A53CCC89AAEEULL, "vxca_cavern1_haunted_ii8_0_0.vlm"},
	{0xAEF6C13D879E4957ULL, "vxw_blightlands_gm_0_0.vlm"},
	{0xAEF6FAF3FF4A12CEULL, "zz_prmpt_debug6198150.cl"},
	{0xAEF7010671228D93ULL, "vxw_bonepit_1z6_0_0.vlm"},
	{0xAEF709D4C081C03CULL, "6055064_m.fxe"},
	{0xAEF72E100C11B8BFULL, "gen00fl_varric6105373.cl"},
	{0xAEF7485233E280F2ULL, "hlac_lt_0_lowtown_a2_n_2_0x.dds"},
	{0xAEF79E871DDAF8BBULL, "6083529_m.fxe"},
	{0xAEF8F0E2532C2C8FULL, "seb321st_leliana.stg"},
	{0xAEF93DE55D8F20B9ULL, "qun131_taskmaster6039406.cl"},
	{0xAEF95752693C49A7ULL, "art\\characters\\playercharacter\\humanfemale\\cth_comb.phy"},
	{0xAEF9D0726954D175ULL, "rdr111dg_martin6054543.cl"},
	{0xAEFA9EF23DA3813CULL, "vxdr_primevalentrance_3ba_0_0.vlm"},
	{0xAEFB3EFA9BF621ABULL, "hlac_lt_0_lowtown_a2_-262144_0x.dds"},
	{0xAEFB98C1DC397451ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_enter.gad"},
	{0xAEFBC997677E934FULL, "gen00fl_varric6087479.cl"},
	{0xAEFBD043FA90D4D4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sidetable.phy"},
	{0xAEFC25E70C8D0175ULL, "296300308.ogg"},
	{0xAEFC6EABDD768DEDULL, "755229681.ogg"},
	{0xAEFD28427589A35DULL, "zz_mcr_debug6071532.cl"},
	{0xAEFD9D28AC2C304AULL, "hlac_lt_0_lowtown_n_1_7z.dds"},
	{0xAEFDB7B79278995DULL, "vxw_blightlands_ai_0_0.vlm"},
	{0xAEFDC2D92CBA77F6ULL, "vxdr_dungeon_11w_0_1.vlm"},
	{0xAEFDF6F4A50D29D4ULL, "1067949295.ogg"},
	{0xAEFE46D7159EC601ULL, "1062842785.ogg"},
	{0xAEFE568455B5DC12ULL, "6167091_m.fxe"},
	{0xAEFED5C7E12B18C6ULL, "6072403_m.fxe"},
	{0xAEFF082D6CD643F9ULL, "6001193_f.fxe"},
	{0xAEFF20E6D6B8F3A0ULL, "zz_mag_debug6053417.cl"},
	{0xAEFF35F37E1A450EULL, "943843391.ogg"},
	{0xAEFF51BFE8F17882ULL, "hlac_ca_caverns1_mines_24424_8x.dds"},
	{0xAEFF8955C02EEB54ULL, "hlac_lt_undercityserialkiller_-1_31x.dds"},
	{0xAEFF93BAE732D457ULL, "gen_im_arm_glv_lgt_champion.uti"},
	{0xAEFFCF6BAA01EBE3ULL, "vxht_mansion_n_60g_0_0.vlm"},
	{0xAF00484DEE3C0DACULL, "seb221_brett_harimann6187785.cl"},
	{0xAF0063D9BF6374ABULL, "mcr370_cut_player_survives6171829.cl"},
	{0xAF00D1F7D6407536ULL, "581143677.ogg"},
	{0xAF01297EE26417D6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven_12.pwk"},
	{0xAF01999863EB62BAULL, "624569350.ogg"},
	{0xAF0199C1DC3DE1FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\cower_enter.ani"},
	{0xAF019CFB417657A1ULL, "hlac_ht_keep_n_4519_17x.dds"},
	{0xAF021252FC4A81DBULL, "llac_ht_mansion2_818_14.dds"},
	{0xAF023645CE657FADULL, "142374105.ogg"},
	{0xAF0270E05690ECE2ULL, "gen00fl_anders6119416.cl"},
	{0xAF03354A938713EAULL, "llac_lt_0_lowtown_docks_n_131071_2.dds"},
	{0xAF0362E92EEB68ADULL, "1062563714.ogg"},
	{0xAF03AC301C45556DULL, "hf_bethany_inf.mor"},
	{0xAF03B90705CF6D6CULL, "163881397.ogg"},
	{0xAF03CF1AD859CB7AULL, "678901430.ogg"},
	{0xAF047CAB2E517E82ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_reaction.gad"},
	{0xAF0498D575E4EFE2ULL, "vxca_cavern1_haunted_imx_0_0.vlm"},
	{0xAF04E3F4A2570D7DULL, "mag360_alain6044618.cl"},
	{0xAF04E79511D54835ULL, "504421049.ogg"},
	{0xAF052CD3BA3FADE0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45sw.phy"},
	{0xAF05331A799461D7ULL, "vxw_blightlands_hl_0_0.vlm"},
	{0xAF05829AB301D725ULL, "art\\effects\\lightattenuation\\default_lightattenuation.dds"},
	{0xAF05C6AD5476B468ULL, "lt_0_lowtown_docks_n_117_0.mmh"},
	{0xAF05D2AD5476C80BULL, "lt_0_lowtown_docks_n_117_0.mao"},
	{0xAF05E4AD5476E692ULL, "lt_0_lowtown_docks_n_117_0.msh"},
	{0xAF067E376E027C2BULL, "40664706.ogg"},
	{0xAF068633E506A2F0ULL, "6163152_m.fxe"},
	{0xAF06AD43FA9AA0D1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sidetable.msh"},
	{0xAF06BB43FA9AB8ECULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sidetable.mao"},
	{0xAF06C743FA9ACC8FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sidetable.mmh"},
	{0xAF073FD1F8CC8ABFULL, "1067482139.ogg"},
	{0xAF0770ABC706CA7AULL, "120788952.ogg"},
	{0xAF079894AEB27E80ULL, "592561323.ogg"},
	{0xAF07C449784D90CCULL, "vxht_keep_n_5jm_0_0.vlm"},
	{0xAF07FFFEF489FB9FULL, "hlac_w_encounter2_camp_327683_0z.dds"},
	{0xAF08337CFA4E4EB4ULL, "art\\vfx\\combat\\vfx_glb_ele_e_deathblow_c.mmh"},
	{0xAF08718EC7E853FCULL, "vxlt_hangedman_1rp_0_0.vlm"},
	{0xAF088823C7DFDB48ULL, "fen210_fenris6108760.cl"},
	{0xAF089F1F6BBDD408ULL, "hlac_lt_gallowscourtyard_-2_1z.dds"},
	{0xAF09634583654C16ULL, "vxw_bonepitmining_a2_6v_0_0.vlm"},
	{0xAF0A19647B481304ULL, "vxdr_primevaltomb_7sj_0_0.vlm"},
	{0xAF0A1AE767D27F00ULL, "33726430.ogg"},
	{0xAF0A9ED8399C8470ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_dual.anb"},
	{0xAF0AF0B9D892B3D2ULL, "gen00fl_anders6110059.cl"},
	{0xAF0B9FAB2E57D979ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_reaction.evt"},
	{0xAF0BBDD1F3F29A8DULL, "hlac_w_bonepitdestroyed_a3_65538_0z.dds"},
	{0xAF0C27D3DB4C9FE8ULL, "vxlt_backalley_n_1lp_0_0.vlm"},
	{0xAF0C2994846EB1DDULL, "vxw_blightlands_jh_0_0.vlm"},
	{0xAF0C3C39D476C316ULL, "dae000ip_crowd_mf5_talk14.utp"},
	{0xAF0D0F23D8E77CEFULL, "90871903.ogg"},
	{0xAF0DF57C1D444C8AULL, "vxdr_primevaltomb_53r_0_0.vlm"},
	{0xAF0EA499E9FFD22CULL, "vxht_keep_67v_0_0.vlm"},
	{0xAF0F0C1909D3851FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_knife_to_throat.gad"},
	{0xAF0F353B48A39398ULL, "gen00fl_anders6209331.cl"},
	{0xAF0F6E4825F2F03CULL, "6026991_f.fxe"},
	{0xAF0FACEAE04A5FB5ULL, "vxht_mansion2_38k_0_0.vlm"},
	{0xAF0FD3654CDE6D07ULL, "6099561_f.fxe"},
	{0xAF0FFB12233F612AULL, "550741819.ogg"},
	{0xAF100C814FA1831DULL, "vxht_mansion2b_2vh_0_0.vlm"},
	{0xAF104E2FA924D7DBULL, "w_encounter3_149_0.msh"},
	{0xAF10602FA924F626ULL, "w_encounter3_149_0.mao"},
	{0xAF10604DAA1B8690ULL, "llac_ht_keep_4519_11.dds"},
	{0xAF106C2FA9250AC5ULL, "w_encounter3_149_0.mmh"},
	{0xAF10A85E9C95F351ULL, "vxlt_0_lowtown_a2_n_cms_0_0.vlm"},
	{0xAF10E765C664210EULL, "follower_banter6136742.cl"},
	{0xAF1184CF13FD87A2ULL, "6150228_m.fxe"},
	{0xAF118D80B112C822ULL, "hlac_ca_caverns1_24421_4x.dds"},
	{0xAF119429E8F85935ULL, "vxw_encounter3_gc_0_0.vlm"},
	{0xAF11BEF4DB8D93A9ULL, "zz_dae_debug6045834.cl"},
	{0xAF11E7FA26B916C4ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliff01_n.dds"},
	{0xAF127F5884FDE52BULL, "hlac_w_bonepitdestroyed_a3_196612_0z.dds"},
	{0xAF12B3FE2B3576E8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medg_0n.dds"},
	{0xAF12CC6AD6DDB017ULL, "vxw_encounter1_dn_0_0.vlm"},
	{0xAF130AB267DB0862ULL, "vxca_cavern1_haunted_ijy_0_0.vlm"},
	{0xAF133DCAF8A49D5CULL, "one220_yevhen6099473.cl"},
	{0xAF1343FC8126AE3CULL, "vxw_encounter3_9p_0_0.vlm"},
	{0xAF13605257A7656EULL, "6143066_m.fxe"},
	{0xAF13A4F76B8A7A57ULL, "vxw_blightlands_mo_0_0.vlm"},
	{0xAF1441DF86CDB892ULL, "art\\characters\\playercharacter\\humanmale\\boo_maga.phy"},
	{0xAF14B87B5E8300E6ULL, "gen00fl_varric6118808.cl"},
	{0xAF15231346DFB967ULL, "zz_clearvfx.ncs"},
	{0xAF156F1A9C3ADBD5ULL, "vxw_bonepitmining_a2_207_0_0.vlm"},
	{0xAF15D040995617D7ULL, "777876388.ogg"},
	{0xAF16011909D9910CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_knife_to_throat.evt"},
	{0xAF161AFA6C31135EULL, "6133784_m.fxe"},
	{0xAF16364AEB8A99A5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_saint_sa.dds"},
	{0xAF164FB8E9D7DCD9ULL, "hlac_ht_chantry_2534_27y.dds"},
	{0xAF16859BB433C1DAULL, "lt_0_lowtown_docks_a3_182_0.mmh"},
	{0xAF16899BB433C811ULL, "lt_0_lowtown_docks_a3_182_0.mao"},
	{0xAF16979BB433E020ULL, "lt_0_lowtown_docks_a3_182_0.msh"},
	{0xAF16C9A7880C110BULL, "6180808_m.fxe"},
	{0xAF173D02E0930EDDULL, "344557346.ogg"},
	{0xAF1776204062F803ULL, "art\\levels\\races\\design\\prp_rubble_0.pwk"},
	{0xAF17FE507B9B6FA0ULL, "dre000_stroud6036716.cl"},
	{0xAF18E4593ECCA208ULL, "t1_mue_or2.tnt"},
	{0xAF18E9529B688199ULL, "hlac_ht_keep_3732_0y.dds"},
	{0xAF190B94302F9EDCULL, "gen00fl_fenris6127080.cl"},
	{0xAF19A7AB2E642468ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\custom\\hugging_kill_reaction.ani"},
	{0xAF1A9B717664CA5DULL, "vxw_blightlands_og_0_0.vlm"},
	{0xAF1AB6E4DB728336ULL, "dae000cr_grand_cleric6173666.cl"},
	{0xAF1AFCD61DEA81FCULL, "lt_0_lowtown_39_0.anb"},
	{0xAF1B823D8134E143ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x4portcull.mao"},
	{0xAF1B863D8134E790ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x4portcull.mmh"},
	{0xAF1B943D8134FFAAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x4portcull.msh"},
	{0xAF1BD88D7C73165FULL, "lgt110_orz_harrowmont6109028.cl"},
	{0xAF1C0129D4548B71ULL, "vxht_keep_n_53g_0_0.vlm"},
	{0xAF1CBFBDA05695D2ULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_pcfem.gad"},
	{0xAF1CC89422B57C37ULL, "6076460_m.fxe"},
	{0xAF1D1676DBF5EAACULL, "vxht_keep_60w_0_0.vlm"},
	{0xAF1D3CD7A0079415ULL, "hlac_ht_chantry_2534_33x.dds"},
	{0xAF1D65BB087E4F9AULL, "vxw_sundermount_374_0_1.vlm"},
	{0xAF1D7D9EB8EE7EA6ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_c.impa.evt"},
	{0xAF1DD6861271B3C4ULL, "dae000_cora6162360.cl"},
	{0xAF1F14288C378334ULL, "418375312.ogg"},
	{0xAF1FBE3C3B324418ULL, "368927416.ogg"},
	{0xAF20169A332C7E8BULL, "6080986_m.fxe"},
	{0xAF20485977BE579AULL, "898931738.ogg"},
	{0xAF204EF393804BC7ULL, "gui_reticle_p1.psh.fxo"},
	{0xAF20A1F5338F1019ULL, "cs_mcr370cs_player_flees.wbd"},
	{0xAF20D56377922636ULL, "hlac_lt_0_lowtown_a2_0_9x.dds"},
	{0xAF213E47C8D3C897ULL, "vxw_encounter1_ii_0_0.vlm"},
	{0xAF21414685DB1A37ULL, "one120_athenril6088316.cl"},
	{0xAF215F0A21ECECBAULL, "284564881.ogg"},
	{0xAF217C8F59D120E2ULL, "vxca_cavern1_haunted_igz_0_0.vlm"},
	{0xAF21B5D9731CC6BCULL, "vxw_encounter3_2q_0_0.vlm"},
	{0xAF21FC2B29D65835ULL, "zz_vault_debug6200363.cl"},
	{0xAF2216D45D8092D7ULL, "vxw_blightlands_vn_0_0.vlm"},
	{0xAF22527AD8D846C3ULL, "vxlt_shop_ds_0_0.vlm"},
	{0xAF22A7B5E164E90DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_05tree_1.mao"},
	{0xAF22AA9FC582A8EBULL, "lgt310pt_reining.ncs"},
	{0xAF23189C06E0475EULL, "hlac_lt_gallowstemplar_f_2384_5y.dds"},
	{0xAF23949EB8F311F1ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_c.impa.gad"},
	{0xAF23E91909E5A665ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_knife_to_throat.ani"},
	{0xAF2443B05D5C6FD7ULL, "836560489.ogg"},
	{0xAF24740B7EB3201BULL, "vis120_commoner1.bnk"},
	{0xAF247C0959BDEFF0ULL, "vxlt_warehouse_n_pz_0_0.vlm"},
	{0xAF2492A8899B168CULL, "6102100_m.fxe"},
	{0xAF24C57275BC60EEULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd05a.phy"},
	{0xAF24E016AE5B3FE9ULL, "17853616.ogg"},
	{0xAF251C597CF163C4ULL, "vxca_caverns1_i1k_0_0.vlm"},
	{0xAF254C65AA729F83ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic8x.mao"},
	{0xAF255065AA72A6D0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic8x.mmh"},
	{0xAF2553BE1E513E58ULL, "616644452.ogg"},
	{0xAF255E65AA72BEEAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic8x.msh"},
	{0xAF25C2D664515C25ULL, "6076417_m.fxe"},
	{0xAF25CDA21E384E22ULL, "art\\levels\\races\\design\\_textures\\prp_chestb_d.dds"},
	{0xAF25EF8336A00D45ULL, "896318673.ogg"},
	{0xAF2646FF67517D16ULL, "vxw_bonepitmining_a2_8x_0_0.vlm"},
	{0xAF264B105F39F5D1ULL, "gen00fl_anders6119286.cl"},
	{0xAF26746ED0C2C132ULL, "127860097.ogg"},
	{0xAF26943D9CF0FFB3ULL, "hlac_ht_brothel_n_3365_1z.dds"},
	{0xAF272821B40CAD19ULL, "366784627.ogg"},
	{0xAF27760B7EB559DBULL, "vis120_commoner1.cnv"},
	{0xAF27F6FBAAD1FC10ULL, "vxht_0_hightown_a2_9bf_0_0.vlm"},
	{0xAF290D4E685AE2DDULL, "vxw_blightlands_pf_0_0.vlm"},
	{0xAF29ADD62E583C83ULL, "zz_followers_banter6157265.cl"},
	{0xAF2A023AB7247934ULL, "hlac_lt_0_lowtown_a2_n_c_65537_0x.dds"},
	{0xAF2A2A98E4A899C1ULL, "vxw_encounter2_camp_dt_0_1.vlm"},
	{0xAF2A67C164A0F888ULL, "fx_snowflake01.mao"},
	{0xAF2A959EB8F9319FULL, "art\\vfx\\creatures\\vfx_c_hd_spit_c.impa.ani"},
	{0xAF2B36D4B41390C3ULL, "w_sundermount_177_0.mao"},
	{0xAF2B3AD4B4139710ULL, "w_sundermount_177_0.mmh"},
	{0xAF2B48D4B413AF2AULL, "w_sundermount_177_0.msh"},
	{0xAF2B49D3BA60AE75ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45sw.msh"},
	{0xAF2B4BD3BA60B1EBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45sw.mmh"},
	{0xAF2B4FDFAC817563ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d.ani"},
	{0xAF2B57D3BA60C588ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_45sw.mao"},
	{0xAF2BAF05F7A8686FULL, "vxlt_undercityencounter_1bp_0_0.vlm"},
	{0xAF2BE369B3487B35ULL, "6033141_f.fxe"},
	{0xAF2C2A20CF5EE4DBULL, "one100_party_comments.bnk"},
	{0xAF2C6401F64F1AAFULL, "vxht_brothel_n_6ec_0_0.vlm"},
	{0xAF2C90A4C43690B5ULL, "vxht_mansion2_32i_0_0.vlm"},
	{0xAF2D554BDAF44D1AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockc.msh"},
	{0xAF2D634BDAF46553ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockc.mao"},
	{0xAF2D674BDAF46C60ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockc.mmh"},
	{0xAF2DA0F2AC19D71BULL, "face_rendertodepth_p1.psh.fxo.dx11"},
	{0xAF2DA11EAF32E209ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtone.mao"},
	{0xAF2DAD1EAF32F6A2ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtone.mmh"},
	{0xAF2DAF1EAF32FA58ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtone.msh"},
	{0xAF2DC4CB1D037C9AULL, "ave100_donnic6105951.cl"},
	{0xAF2E372B21C3CBDCULL, "hlac_w_sundermount_-65539_0z.dds"},
	{0xAF2E54C47CF39959ULL, "vxlt_gallowstemplar_51y_0_0.vlm"},
	{0xAF2E8863EB765AFEULL, "vxlt_0_lowtown_docks_a3_n_ctq_0_0.vlm"},
	{0xAF2ECBE67DF80836ULL, "74969289.ogg"},
	{0xAF2EF34E01999F8FULL, "gen00fl_varric6107289.cl"},
	{0xAF2F2D642F1C5A7AULL, "491485160.ogg"},
	{0xAF2F3AB6D650849BULL, "hlac_w_blightlands_131073_0y.dds"},
	{0xAF2F7C69F54BB72FULL, "vxw_blightlands_7u_0_0.vlm"},
	{0xAF2FA69B449C7A2DULL, "448692707.ogg"},
	{0xAF2FD8559C8D6100ULL, "ser270_emeric6113585.cl"},
	{0xAF2FEFBBB6C50390ULL, "vxlt_warehouse_n_18a_0_0.vlm"},
	{0xAF302C20CF62D09BULL, "one100_party_comments.cnv"},
	{0xAF30343715D83D79ULL, "vxlt_hangedman_n_25l_0_0.vlm"},
	{0xAF30AABDA0671B78ULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_pcfem.ani"},
	{0xAF30BD57ACFCBA5FULL, "vxca_caverns1_mines_ea1_0_0.vlm"},
	{0xAF30CD84BF6F9C71ULL, "vxht_0_hightown_n_aok_0_0.vlm"},
	{0xAF30E924129501D0ULL, "6218700_m.fxe"},
	{0xAF30F39ED5D210CBULL, "vxca_caverns1_mines_e8v_0_0.vlm"},
	{0xAF31054265131C2DULL, "lgt110_gxa_vgk_silver_order6117055.cl"},
	{0xAF31367CFA71596BULL, "art\\vfx\\combat\\vfx_glb_ele_e_deathblow_c.anb"},
	{0xAF31A2E11285B582ULL, "vxlt_0_lowtown_a3_n_c_h2i_0_0.vlm"},
	{0xAF31B1739E3A0BBAULL, "gen00fl_merrill6128692.cl"},
	{0xAF326988B9E98204ULL, "6112079_m.fxe"},
	{0xAF3284D7FCDE0A8EULL, "vxht_keep_n_3g2_0_0.vlm"},
	{0xAF32D123CCA7C5CFULL, "vxw_bonepitmining_a2_2a6_0_0.vlm"},
	{0xAF32D6A4F6EA02ACULL, "6083600_f.fxe"},
	{0xAF330AA628F3B9E5ULL, "191101697.ogg"},
	{0xAF3331E202D7CA14ULL, "vxlt_0_lowtown_docks_a3_n_d39_0_0.vlm"},
	{0xAF336075BA45EC18ULL, "hlac_ht_0_hightown_a3_-65537_2y.dds"},
	{0xAF338F3F4BF23CEAULL, "6071183_m.fxe"},
	{0xAF33AA3CF9DE41FDULL, "gen00fl_isabela6099522.cl"},
	{0xAF353EC9B775FBFFULL, "llac_lt_gallowstemplar_n_2382_3.dds"},
	{0xAF3732FEB8E8E362ULL, "vxca_caverns2prc_9xl_0_0.vlm"},
	{0xAF379673EB64175AULL, "w_bonepit_125_0.mmh"},
	{0xAF379A73EB641D91ULL, "w_bonepit_125_0.mao"},
	{0xAF37A873EB6435A0ULL, "w_bonepit_125_0.msh"},
	{0xAF37ACE4ACAB2F5CULL, "pro000_captain_wright6004408.cl"},
	{0xAF37ADE1C1ECEF33ULL, "vxlt_0_lowtown_a3_n_c_hpb_0_0.vlm"},
	{0xAF37CA389F601BEEULL, "918893646.ogg"},
	{0xAF383C2A24EAB7D3ULL, "144370215.ogg"},
	{0xAF3867DFAC8C285AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d.evt"},
	{0xAF38FB7919FC249AULL, "art\\levels\\ztools\\default\\generic\\0oo_basic.phy"},
	{0xAF3963690C47E212ULL, "gen00fl_aveline6106873.cl"},
	{0xAF3A07FEB136B2A1ULL, "6034518_m.fxe"},
	{0xAF3AC85DB20AAFFCULL, "vxdr_0_deeproads_f1r_0_1.vlm"},
	{0xAF3AD107BC454963ULL, "llac_w_encounter1_4_0.dds"},
	{0xAF3B62182583CC9DULL, "vxht_mansion2b_2ie_0_0.vlm"},
	{0xAF3BDF9F82BACA48ULL, "6072401_m.fxe"},
	{0xAF3C289666954507ULL, "vxca_caverns2prc_aej_0_0.vlm"},
	{0xAF3C28FE7945F769ULL, "6121502_m.fxe"},
	{0xAF3C5C9486F15D3AULL, "vxw_encounter2_camp_r4_0_0.vlm"},
	{0xAF3C7024CB2CB740ULL, "hlac_ca_caverns1_24426_5x.dds"},
	{0xAF3CFA9ACB42943CULL, "vxlt_0_lowtown_a2_eqs_0_0.vlm"},
	{0xAF3D5674768DAE80ULL, "hlac_lt_undercityserialkiller_-1_21y.dds"},
	{0xAF3DAF1D6CA0924CULL, "zz_ave_debug.ncs"},
	{0xAF3DDD882332E5E8ULL, "autolevelparty.ncs"},
	{0xAF3DEE46E741CFAFULL, "vxw_blightlands_8t_0_0.vlm"},
	{0xAF3E010C5AE2BBADULL, "zz_fen_debug6166020.cl"},
	{0xAF3E115CCFB1A530ULL, "llac_dr_slavers_3216_0.dds"},
	{0xAF3E57B85EF3A4A9ULL, "prm000im_resilient_ring.uti"},
	{0xAF3EFB27D1B9AF83ULL, "vxlt_gallowstemplar_3ss_0_0.vlm"},
	{0xAF3F5230E03C03CAULL, "mag110_ws_banter6079650.cl"},
	{0xAF3F66DFAC92454DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d.gad"},
	{0xAF3F8F480A08AD4EULL, "vxlt_0_lowtown_a3_n_c_hny_0_0.vlm"},
	{0xAF401955832C652CULL, "vxlt_gallowstemplar_n_4ub_0_0.vlm"},
	{0xAF405ED32FF3D906ULL, "6092583_f.fxe"},
	{0xAF4082AF7D526D1BULL, "vxlt_0_lowtown_a2_n_c_emy_0_0.vlm"},
	{0xAF409DCD4C7E1721ULL, "526133081.ogg"},
	{0xAF4117F09171B441ULL, "gen00fl_aveline6114031.cl"},
	{0xAF413B6CDC1253A8ULL, "qun131_party_comment6117298.cl"},
	{0xAF415FC33DAA20F0ULL, "vxlt_warehouse_n_zx_0_0.vlm"},
	{0xAF425B278B822A70ULL, "6027010_m.fxe"},
	{0xAF428927B091CF68ULL, "vxca_caverns1_do0_0_0.vlm"},
	{0xAF42AFC688E19924ULL, "vxlt_gallowstemplar_n_2be_0_0.vlm"},
	{0xAF431A94C0C719FAULL, "lt_0_lowtown_a2_n_c_45_0.anb"},
	{0xAF434DC8C66B6A09ULL, "vxca_caverns1_mines_cp2_0_0.vlm"},
	{0xAF43642FD9AE13BDULL, "6090431_m.fxe"},
	{0xAF43E3B0AB0AAB87ULL, "vxlt_0_lowtown_a2_n_c_hnq_0_0.vlm"},
	{0xAF441A78017A545AULL, "381258955.ogg"},
	{0xAF444354414010C2ULL, "gen00fl_fenris6126940.cl"},
	{0xAF44D6BCAC93D651ULL, "gen_im_trs_gen_hvy_boo_r0.uti"},
	{0xAF455EFC282ECAD0ULL, "6020536_m.fxe"},
	{0xAF4599CAF6DDFEF9ULL, "dae300_alistair_king6133559.cl"},
	{0xAF45EF47A32501E8ULL, "vxlt_backalley_n_1xl_0_0.vlm"},
	{0xAF463157371FBAD2ULL, "125933108.ogg"},
	{0xAF4662B30A430320ULL, "830096437.ogg"},
	{0xAF4717687F8A6AB5ULL, "vxlt_warehouse_tz_0_0.vlm"},
	{0xAF477D6DE02145A6ULL, "vxht_0_hightown_c8l_0_0.vlm"},
	{0xAF47C53B36C9800AULL, "436216228.ogg"},
	{0xAF4838C164B9A6EAULL, "fx_snowflake01.dds"},
	{0xAF489FF9E8DC0DE5ULL, "mcr370_assignments6176725.cl"},
	{0xAF48BADFDCF8F142ULL, "vxlt_warehouse_n_14w_0_0.vlm"},
	{0xAF4938DBFD1692E5ULL, "s_ballista_06_l3.mao"},
	{0xAF4940760512CA1FULL, "hlac_ht_0_hightown_a3_65536_4z.dds"},
	{0xAF4947BBDA3B4BAFULL, "vxht_brothel_n_6km_0_0.vlm"},
	{0xAF49D6D760BA23CDULL, "rob_pl2.tnt"},
	{0xAF4A52CDC56DEFF1ULL, "6164083_f.fxe"},
	{0xAF4A5F97693A4308ULL, "6119446_f.fxe"},
	{0xAF4A8BFFAE813BFFULL, "141843679.ogg"},
	{0xAF4AB9D27343548FULL, "vxca_caverns2prc_a0m_0_0.vlm"},
	{0xAF4ACE7178E775BAULL, "vxw_encounter2_camp_m5_0_0.vlm"},
	{0xAF4B66ECD4DE89B1ULL, "gen00fl_sebastian6188595.cl"},
	{0xAF4BDBAD5AD3B5E6ULL, "579993008.ogg"},
	{0xAF4BF8F48C19889AULL, "vxht_mansion_n_6sc_0_0.vlm"},
	{0xAF4C946B3D989591ULL, "vxca_cavern1_haunted_gjx_0_0.vlm"},
	{0xAF4CBE9471C03C00ULL, "6113530_f.fxe"},
	{0xAF4D3F7CEF69BC85ULL, "zz_slate2.cut"},
	{0xAF4D6CBA7C109356ULL, "follower_banter6026234.cl"},
	{0xAF4DD7F7D1818C50ULL, "gen00fl_varric6076074.cl"},
	{0xAF4DEFC7A6026C83ULL, "character_character_p13.psh.fxo"},
	{0xAF4DF24D38D028D0ULL, "197224290.ogg"},
	{0xAF4E2037505469EBULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.impa.evt"},
	{0xAF4E3B91E6B812F5ULL, "414934697.ogg"},
	{0xAF4E9A1E6F7CCE7BULL, "6139927_m.fxe"},
	{0xAF4ED74CAC39F627ULL, "llac_ca_caverns1_24424_20.dds"},
	{0xAF4EF0DBBD89EA26ULL, "291674955.ogg"},
	{0xAF4EF558CFDF4D46ULL, "qun200_party_comment6090332.cl"},
	{0xAF4FE9530A89CBA5ULL, "vxht_0_hightown_n_c7a_0_0.vlm"},
	{0xAF503FAC8DA7A96DULL, "vxlt_gallowscourtyard_7bp_0_0.vlm"},
	{0xAF50736B0A727429ULL, "vxlt_0_lowtown_n_gpd_0_0.vlm"},
	{0xAF507B03CDE5A35FULL, "vxdr_primevalentrance_28a_0_0.vlm"},
	{0xAF50AE12FD359333ULL, "hlac_dr_slavers_3214_14x.dds"},
	{0xAF50BE6455628CDBULL, "vxlt_warehouse_oq_0_0.vlm"},
	{0xAF51470100C0D169ULL, "llac_dr_primevaltomb_659_5.dds"},
	{0xAF514A62A6E84209ULL, "and121_karl6045419.cl"},
	{0xAF5175D3315B2D27ULL, "vxlt_hangedman_n_ta_0_0.vlm"},
	{0xAF5219EE20E2C9E7ULL, "and100_anders6173563.cl"},
	{0xAF52AAB4D53EB7EDULL, "character_character_p6.psh.fxo.dx11"},
	{0xAF52C57FDAAF08C3ULL, "vxlt_gallowscourtyard_79o_0_0.vlm"},
	{0xAF530696EA009F1EULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.mmh"},
	{0xAF5368ED76A3F889ULL, "static_motionvectors_p0.psh.fxo.dx11"},
	{0xAF53C4FD18D35830ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleft.mao"},
	{0xAF53C8FD18D35F43ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleft.mmh"},
	{0xAF53D6FD18D3769DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleft.msh"},
	{0xAF5427CA4C8B883CULL, "vxht_0_hightown_n_b3y_0_0.vlm"},
	{0xAF5439375058FFACULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.impa.gad"},
	{0xAF549D3AF05D7DA9ULL, "6047390_m.fxe"},
	{0xAF54AD47DDA4A5DAULL, "vxca_cave_2e6_0_0.vlm"},
	{0xAF54CC3985A1FC35ULL, "scaleform_multiply.mao"},
	{0xAF55675EE22CC63EULL, "dae200_cullen6140283.cl"},
	{0xAF55742024F2FF18ULL, "vrc321tr_loud_noise.ncs"},
	{0xAF55894571808335ULL, "vxlt_warehouse_yy_0_0.vlm"},
	{0xAF561529ACEEC586ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.cess.ani"},
	{0xAF56AEE5A39BE8CAULL, "rdr111dg_martin6054999.cl"},
	{0xAF56D510B022CBB3ULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.dura.ani"},
	{0xAF56F3E36224D4C4ULL, "dae000_adriano6163669.cl"},
	{0xAF574E1D53F0CDDFULL, "vxht_0_hightown_a2_bz2_0_0.vlm"},
	{0xAF57BB3AC1118AE8ULL, "vxlt_0_lowtown_docks_a3_aq9_0_0.vlm"},
	{0xAF5832DC4117AA19ULL, "427830043.ogg"},
	{0xAF58452FA9627E72ULL, "w_encounter3_149_0.phy"},
	{0xAF58513994A49B5BULL, "zz_cave_accolades_3_cinetes.cut"},
	{0xAF58639C990FD1E7ULL, "qun110_petrice.bnk"},
	{0xAF58C2A6D8A0209FULL, "vxht_0_hightown_c12_0_0.vlm"},
	{0xAF592FF7F8A14B93ULL, "246437579.ogg"},
	{0xAF59404E6ADD8E3AULL, "vxw_encounter2_camp_h6_0_0.vlm"},
	{0xAF595301BC3614D5ULL, "637191066.ogg"},
	{0xAF5977955DEF3841ULL, "vxlt_0_lowtown_docks_n_c09_0_0.vlm"},
	{0xAF5982A6E0341821ULL, "vxlt_0_lowtown_a2_n_epc_0_1.vlm"},
	{0xAF599A544905CCD1ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\custom enter\\second_enter.ani"},
	{0xAF5A8BB1D7D2CF21ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_018a_0n.dds"},
	{0xAF5A9C3C13CB68C0ULL, "910892135.ogg"},
	{0xAF5AFBD61E20638CULL, "lt_0_lowtown_39_0.rml"},
	{0xAF5B3837505F1CAAULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.impa.ani"},
	{0xAF5B66FF48F5CB8AULL, "vxw_bonepit_lr_0_0.vlm"},
	{0xAF5B859C9912415FULL, "qun110_petrice.cnv"},
	{0xAF5B94D829DEFCE9ULL, "747229208.ogg"},
	{0xAF5BF1C5655F0D6DULL, "6179873_m.fxe"},
	{0xAF5C6C85EA1D14BEULL, "6121178_m.fxe"},
	{0xAF5C8D3D816C8AABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9dd_4x4portcull.phy"},
	{0xAF5C9ACD9BED82DFULL, "vxht_0_hightown_a2_aok_0_1.vlm"},
	{0xAF5DB573EB851B65ULL, "w_bonepit_125_0.phy"},
	{0xAF5E681463D9579AULL, "hlac_ca_caverns1_24424_15y.dds"},
	{0xAF5EA4656E1D9D36ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgtp.msh"},
	{0xAF5EAE656E1DAE34ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgtp.mmh"},
	{0xAF5EB15549455F31ULL, "6101006_f.fxe"},
	{0xAF5EB2656E1DB547ULL, "art\\characters\\playercharacter\\elfmale\\boo_lgtp.mao"},
	{0xAF5F19F3222AE3E9ULL, "vxlt_warehouse_n_11o_0_0.vlm"},
	{0xAF5F20DA3A97A5A0ULL, "llac_lt_0_lowtown_a3_n_c_65533_3.dds"},
	{0xAF5F30414758A55BULL, "vxlt_warehouse_pp_0_0.vlm"},
	{0xAF5FA554490A4B73ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\custom enter\\second_enter.gad"},
	{0xAF5FBEF677BA0CC6ULL, "vxlt_0_lowtown_a2_n_gx3_0_0.vlm"},
	{0xAF5FD3E216C1812BULL, "hlac_lt_0_lowtown_a2_n_131072_3z.dds"},
	{0xAF6054693C0F906BULL, "dae021_sister_lorena_sta6154073.cl"},
	{0xAF6088F555CFAB70ULL, "gen_im_cns_msc_backpack_01.uti"},
	{0xAF60895034BF4F3FULL, "vxht_brothel_n_5e9_0_1.vlm"},
	{0xAF60B056B13B9AF2ULL, "6005643_m.fxe"},
	{0xAF60E5716E282019ULL, "vxlt_0_lowtown_ewt_0_0.vlm"},
	{0xAF60E596EA0CA4DDULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.anb"},
	{0xAF6142E5B32E3080ULL, "vxca_cavern1_haunted_cjy_0_0.vlm"},
	{0xAF61845D7B1761B7ULL, "hlac_ht_keep_n_4519_0z.dds"},
	{0xAF619232B546B301ULL, "vxht_0_hightown_a2_c5t_0_0.vlm"},
	{0xAF61A010F2626F38ULL, "item_sets.gda"},
	{0xAF61B2D6198B5678ULL, "189586728.ogg"},
	{0xAF61E8DD15134337ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_hallastatue.mmh"},
	{0xAF61ECDD15134A44ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_hallastatue.mao"},
	{0xAF61FEDD151368A9ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_hallastatue.msh"},
	{0xAF62CA52A437408DULL, "79846139.ogg"},
	{0xAF63A49C0D464D56ULL, "lt_0_lowtown_docks_a3_10_0_lc.mmh"},
	{0xAF63C50870787C3EULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_isabella.ani"},
	{0xAF63FD29ACFADAF7ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.cess.evt"},
	{0xAF6408F92AF6596FULL, "vxlt_0_lowtown_docks_a3_n_akk_0_0.vlm"},
	{0xAF6420543F85D44CULL, "531811793.ogg"},
	{0xAF644444E1AED05FULL, "mer210_smith6051390.cl"},
	{0xAF646127C40D76A6ULL, "vxht_0_hightown_c2n_0_0.vlm"},
	{0xAF64C9FD18E1D468ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1wi_grandleft.phy"},
	{0xAF64CD10B02EFBEAULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.dura.evt"},
	{0xAF64F84AD523EA78ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkout.phy"},
	{0xAF65285E3169A759ULL, "mcr371cr_harvester_stunt.utc"},
	{0xAF652B3AE58793E4ULL, "51970228.ogg"},
	{0xAF653A0D5F77C348ULL, "6135119_m.fxe"},
	{0xAF6548519E68C306ULL, "vxlt_0_lowtown_docks_a3_n_cxe_0_0.vlm"},
	{0xAF655765AAA896EBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic8x.phy"},
	{0xAF65600B7EEA950AULL, "vis120_commoner1.utc"},
	{0xAF65AF97A6A47A6EULL, "lt_gallowstemplar_2390.rml"},
	{0xAF65C47275F3F5B1ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd05a.mmh"},
	{0xAF65DE7275F4226FULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd05a.msh"},
	{0xAF660990BB91E4D7ULL, "genip_cin_book_dagger.utp"},
	{0xAF661B58854A5C4EULL, "qun221_viscount6073890.cl"},
	{0xAF66418EB113E620ULL, "vxdr_slavers_14i_0_0.vlm"},
	{0xAF66562428CEC06CULL, "140140983.ogg"},
	{0xAF66A254491064D8ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\custom enter\\second_enter.evt"},
	{0xAF6718FF9A37619AULL, "6129457_m.fxe"},
	{0xAF67843FDB74075AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidleft.phy"},
	{0xAF67B22B5CD3A6BAULL, "vxw_encounter2_camp_c7_0_0.vlm"},
	{0xAF67BAA658E1A301ULL, "gen00fl_merrill6128639.cl"},
	{0xAF67DD032EDE21CDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottlewine_d.dds"},
	{0xAF685BE51B2D3E77ULL, "376700639.ogg"},
	{0xAF687E6A34A41615ULL, "lgt260tr_way_blood_mage.utt"},
	{0xAF6940D611BF7D07ULL, "6119827_f.fxe"},
	{0xAF69C20B6AB352F5ULL, "243484127.ogg"},
	{0xAF69D41D6CC60C7FULL, "zz_ave_debug.cnv"},
	{0xAF69FE29ACFF4780ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.cess.gad"},
	{0xAF6A3AA18C95FA6BULL, "281360234.ogg"},
	{0xAF6A94F88747FE71ULL, "vxht_0_hightown_n_aso_0_0.vlm"},
	{0xAF6AE139A8DC5E1BULL, "642506590.ogg"},
	{0xAF6B018E2D97177BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldpost.phy"},
	{0xAF6B39B7601379D5ULL, "6036567_m.fxe"},
	{0xAF6B79624769F9C9ULL, "gen00fl_isabela6135673.cl"},
	{0xAF6BAC10B034E25DULL, "art\\vfx\\combat\\vfx_w_wns_sheild_defen_l_c.dura.gad"},
	{0xAF6BAD9283EA266EULL, "205343997.ogg"},
	{0xAF6C056E8AD83313ULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfl.ani"},
	{0xAF6C3BFC0798B217ULL, "vxw_bonepitdestroyed_a3_26y_0_0.vlm"},
	{0xAF6C871842A0C5CCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerable.ani"},
	{0xAF6C8B2025060515ULL, "vrc321tr_loud_noise.utt"},
	{0xAF6CCD0CEE75FCA5ULL, "vxht_0_hightown_n_c9o_0_0.vlm"},
	{0xAF6CDAC1C2743673ULL, "vxlt_gallowscourtyard_n_6eq_0_0.vlm"},
	{0xAF6D5724EE5EA529ULL, "vxlt_0_lowtown_n_gjj_0_0.vlm"},
	{0xAF6D5B15CECA19E5ULL, "232792701.ogg"},
	{0xAF6DB3B92D3448CBULL, "ser121_jethann6104625.cl"},
	{0xAF6DC3411D633BB5ULL, "408849958.ogg"},
	{0xAF6E3529B6349BB8ULL, "6042208_m.fxe"},
	{0xAF6E39D4AF796F82ULL, "vxw_encounter2_68_0_0.vlm"},
	{0xAF6E6446F71B27F5ULL, "art\\vfx\\combat\\vfx_q_ari_charge_en_c.anb"},
	{0xAF6E95F8911B2E98ULL, "888896076.ogg"},
	{0xAF6EC08DB01C8395ULL, "hlac_lt_0_lowtown_docks_65533_1x.dds"},
	{0xAF6EEB2990C237D8ULL, "633967530.ogg"},
	{0xAF6F4536FAE1794EULL, "mag211_marethari6021301.cl"},
	{0xAF70040FA73CCB81ULL, "vxht_0_hightown_a2_c0k_0_0.vlm"},
	{0xAF703C6814F5485FULL, "vxht_0_hightown_n_a7m_0_0.vlm"},
	{0xAF714EEBD27ABF17ULL, "vxdr_primevaltomb_55d_0_1.vlm"},
	{0xAF71755589C55133ULL, "vxlt_0_lowtown_a3_n_c_hlf_0_0.vlm"},
	{0xAF718016ACA785F1ULL, "qcr220_orsino6060716.cl"},
	{0xAF71AC0758EDE063ULL, "ser221_quentin6131093.cl"},
	{0xAF723CAFC4FB0C4CULL, "vxw_bonepit_22r_0_0.vlm"},
	{0xAF723DC5C7CDDCA9ULL, "6072440_m.fxe"},
	{0xAF727AD61CEC71EFULL, "vxlt_0_lowtown_docks_a3_n_apj_0_0.vlm"},
	{0xAF7376DB97DD977AULL, "mrl210_merrill6102162.cl"},
	{0xAF73941842A6FA0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerable.gad"},
	{0xAF73FB05688B34DCULL, "vxw_bonepit_1ps_0_0.vlm"},
	{0xAF747A28FF2D224FULL, "vxht_mansion2_25b_0_1.vlm"},
	{0xAF751FADFC3E0B6DULL, "vxdr_primevaltomb_81z_0_0.vlm"},
	{0xAF75E1FC87B58E2EULL, "follower_banter6156036.cl"},
	{0xAF761D94C0F27826ULL, "lt_0_lowtown_a2_n_c_45_0.rml"},
	{0xAF767A0FC6FDA3D8ULL, "559627738.ogg"},
	{0xAF76DB7C00882A57ULL, "283431961.ogg"},
	{0xAF76EFFB1D85C505ULL, "412388878.ogg"},
	{0xAF774B7D8A99FB80ULL, "vxht_0_hightown_a2_cbl_0_0.vlm"},
	{0xAF7780B30C50CC8FULL, "6128059_m.fxe"},
	{0xAF77DE0870894FB8ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_isabella.gad"},
	{0xAF77F954E9BC4334ULL, "6145120_f.fxe"},
	{0xAF78226885C8B45BULL, "32579571.ogg"},
	{0xAF784AEF86E70A88ULL, "755312757.ogg"},
	{0xAF78790DB72E2262ULL, "22030591.ogg"},
	{0xAF78E8F57AA6DBB0ULL, "gen00fl_dog6106118.cl"},
	{0xAF794A47A412BC0CULL, "208971122.ogg"},
	{0xAF798C258C4C1C3DULL, "251818202.ogg"},
	{0xAF798F1842AB5DEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerable.evt"},
	{0xAF799C66512CA67AULL, "6053150_m.fxe"},
	{0xAF79AE35D83AF78CULL, "vxlt_gallowstemplar_38y_0_0.vlm"},
	{0xAF79BC488651177AULL, "vxlt_0_lowtown_e24_0_0.vlm"},
	{0xAF79FD6E8AE4634AULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfl.evt"},
	{0xAF7A5FCA5BAD5DF6ULL, "dae000ip_crowd_mf1_talk10.utp"},
	{0xAF7AADD8F98ECA97ULL, "vxw_bonepitdestroyed_a3_21x_0_0.vlm"},
	{0xAF7AEA01AC4040D8ULL, "374312101.ogg"},
	{0xAF7AFE7F0C32DCB2ULL, "6033589_m.fxe"},
	{0xAF7B06908EB6381BULL, "vxlt_warehouse_1bw_0_0.vlm"},
	{0xAF7B857E8DA35723ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3plank.msh"},
	{0xAF7B937E8DA36F7DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3plank.mmh"},
	{0xAF7B977E8DA3758EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn3plank.mao"},
	{0xAF7BB7AA8EC20DE0ULL, "1034536914.ogg"},
	{0xAF7C0A0C18B2436BULL, "vxdr_primevaltomb_2xy_0_0.vlm"},
	{0xAF7CA3DED15ADD82ULL, "6042527_m.fxe"},
	{0xAF7CABB1A16F8802ULL, "vxw_encounter2_19_0_0.vlm"},
	{0xAF7D00F6CA1237CCULL, "6149418_f.fxe"},
	{0xAF7D8530BE9D6CCCULL, "368360679.ogg"},
	{0xAF7D8D03EC44F1E7ULL, "gen00fl_merrill6131561.cl"},
	{0xAF7D9EF439EAA6A9ULL, "6176674_m.fxe"},
	{0xAF7DC92B52145119ULL, "vxlt_0_lowtown_emz_0_0.vlm"},
	{0xAF7ED319DAC34DAEULL, "mag121_viveka6083368.cl"},
	{0xAF7EE9872DF83207ULL, "297602658.ogg"},
	{0xAF7F57B4220572F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_enter.ani"},
	{0xAF7FA1A801A8D9A8ULL, "qun160_party_comment_16061232.cl"},
	{0xAF8076B584093A75ULL, "244115604.ogg"},
	{0xAF808AB50796183EULL, "fen310_fenris6185232.cl"},
	{0xAF809A9BD051E142ULL, "gen_im_arm_cht_elf_comb_t6.uti"},
	{0xAF80B8E0A2962BC6ULL, "dae200_saemus6142796.cl"},
	{0xAF80DC6E8AEA49BDULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkfl.gad"},
	{0xAF80E561D64AED37ULL, "vxlt_0_lowtown_docks_b5g_0_1.vlm"},
	{0xAF80FC28BDECFB2AULL, "110304993.ogg"},
	{0xAF815FC49EA59446ULL, "823398304.ogg"},
	{0xAF8188105F0581B9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_006a_0t.dds"},
	{0xAF81C747B82666AFULL, "gen00fl_isabela_init6076966.cl"},
	{0xAF81F3A97E974BEDULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtc.mmh"},
	{0xAF81F7A97E97523EULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtc.mao"},
	{0xAF8205A97E976A73ULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtc.msh"},
	{0xAF824F4C86A62546ULL, "zz_isa_debug6133195.cl"},
	{0xAF8256A4A8248BF3ULL, "zz_slate_3_track.cut"},
	{0xAF825927B05C3230ULL, "cd_one300_nuncio.bnk"},
	{0xAF8307907EDFD240ULL, "999827112.ogg"},
	{0xAF832D0653FE4292ULL, "6149073_m.fxe"},
	{0xAF839E5D73DF1439ULL, "qun200_party_comment6090445.cl"},
	{0xAF84005498BAADA4ULL, "hlac_lt_undercityencounter_-1_49y.dds"},
	{0xAF843FEFCC768909ULL, "dae011_conspiracy_mob6141284.cl"},
	{0xAF8463C369737C83ULL, "vrc321_varric6087087.cl"},
	{0xAF84B6F47B67C94FULL, "vfx_vfx_p4.psh.fxo.dx11.assembly"},
	{0xAF84C2DB08D41772ULL, "vxca_caverns2_9dx_0_1.vlm"},
	{0xAF8533EB85110EBCULL, "vxlt_0_lowtown_a2_eht_0_0.vlm"},
	{0xAF8554B42209D956ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_enter.gad"},
	{0xAF859B98FB87CA93ULL, "120995892.ogg"},
	{0xAF85B03BEAA5E319ULL, "vxca_cave_1av_0_0.vlm"},
	{0xAF85D03E2A583727ULL, "mcr370_assignments6081713.cl"},
	{0xAF866C9F64A2CD60ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_splash.mmh"},
	{0xAF86FBC2CA8C088AULL, "art\\characters\\heraldry\\hld_viscount_1.dds"},
	{0xAF8776982CB57294ULL, "ss_circlemages_m_1.wbd"},
	{0xAF88A86F4445B9BBULL, "hlac_lt_0_lowtown_196606_0z.dds"},
	{0xAF88EC38FAC27EEDULL, "vxht_0_hightown_n_cb2_0_0.vlm"},
	{0xAF890C14BD4BCB4BULL, "vxlt_hangedman_20r_0_0.vlm"},
	{0xAF8977B984FB7EBBULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em06.phy"},
	{0xAF8A21EE7323092AULL, "56525087.ogg"},
	{0xAF8A3ADED24AD629ULL, "vxlt_0_lowtown_n_gdh_0_0.vlm"},
	{0xAF8A6493A0FC8707ULL, "lla_ht_0_hightown_a2_-65537_0.dds"},
	{0xAF8A7BE90AA85BEBULL, "vxdr_primevaltomb_2wx_0_0.vlm"},
	{0xAF8A97EFFB3D333CULL, "hf_gen_act3.mor"},
	{0xAF8AB957D7527BA5ULL, "vxht_0_hightown_a2_c1t_0_0.vlm"},
	{0xAF8AC4363CFE609CULL, "fen110_fenris6101944.cl"},
	{0xAF8B1AB953D05497ULL, "mcr370_meredith6136185.cl"},
	{0xAF8B70EABDA9D502ULL, "vxw_blightlands_f7_0_0.vlm"},
	{0xAF8C035CA43F5504ULL, "w_bonepitmining_a2_15_0.msh"},
	{0xAF8C155CA43F746DULL, "w_bonepitmining_a2_15_0.mao"},
	{0xAF8C195CA43F7AA6ULL, "w_bonepitmining_a2_15_0.mmh"},
	{0xAF8C1D0164D92607ULL, "vxlt_0_lowtown_a2_n_c_hul_0_0.vlm"},
	{0xAF8C3A8EC731471AULL, "vxlt_hangedman_n_172_0_0.vlm"},
	{0xAF8C3B08440A6999ULL, "vxlt_0_lowtown_ehy_0_0.vlm"},
	{0xAF8C434759B2F91EULL, "828460657.ogg"},
	{0xAF8C4F4D33E5AF1BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_45basic.phy"},
	{0xAF8C5A984BD389A2ULL, "6016806_f.fxe"},
	{0xAF8C6FB4221025B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_behind_back_enter.evt"},
	{0xAF8C91D51CE71B81ULL, "6103915_m.fxe"},
	{0xAF8D6287B977C890ULL, "vxlt_0_lowtown_a3_n_c_ep0_0_0.vlm"},
	{0xAF8DB53C31B28A47ULL, "919863924.ogg"},
	{0xAF8E16B9D59C366DULL, "vxht_keep_n_4al_0_0.vlm"},
	{0xAF8E1CC06FE1DBCDULL, "631761380.ogg"},
	{0xAF8E6078EA08DB93ULL, "lgt350cr_arcane_horror.ncs"},
	{0xAF8EC5D08D74E098ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrlb_0s.dds"},
	{0xAF8EEA26957A1531ULL, "6169685_m.fxe"},
	{0xAF919548308DC7A0ULL, "6086261_f.fxe"},
	{0xAF919EE345F45B12ULL, "6142173_m.fxe"},
	{0xAF91DC6A1A1238CBULL, "primitives_primitive_p3.vsh.fxo.dx11.assembly"},
	{0xAF927D0F92559792ULL, "one300_zevran6075398.cl"},
	{0xAF928919AACB1489ULL, "gen00fl_sebastian6189013.cl"},
	{0xAF92ECA97EA5B2AAULL, "art\\characters\\playercharacter\\humanmale\\arm_lgtc.phy"},
	{0xAF934143FF6A4328ULL, "6174534_m.fxe"},
	{0xAF936913F8E47CF0ULL, "zz_jump_forward.ncs"},
	{0xAF9394EE6A9F35B5ULL, "vxw_encounter3_8j_0_0.vlm"},
	{0xAF944F9F64AED97FULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_splash.anb"},
	{0xAF94F4791A4A77C3ULL, "art\\levels\\ztools\\default\\generic\\0oo_basic.msh"},
	{0xAF9502791A4A8F1DULL, "art\\levels\\ztools\\default\\generic\\0oo_basic.mmh"},
	{0xAF9506791A4A962EULL, "art\\levels\\ztools\\default\\generic\\0oo_basic.mao"},
	{0xAF9512721280AE2BULL, "6202189_f.fxe"},
	{0xAF951554F9DFAE26ULL, "plt_cod_cha_cullen.plo"},
	{0xAF9559EE36197DC3ULL, "132026614.ogg"},
	{0xAF95E1D42410CF3EULL, "137842902.ogg"},
	{0xAF95FC6C6C476BD3ULL, "276026921.ogg"},
	{0xAF96011EBAB6827AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pew_d.dds"},
	{0xAF96879B852EAA61ULL, "art\\characters\\creatures\\buzzard\\animation\\perched.ani"},
	{0xAF9690AB667CC920ULL, "hlac_lt_gallowstemplar_f_2388_3y.dds"},
	{0xAF96D5831EA2EF70ULL, "hlac_lt_warehouse_n_65536_1z.dds"},
	{0xAF96E88E2DBC271AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldpost.msh"},
	{0xAF96F68E2DBC3F53ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldpost.mao"},
	{0xAF96FA8E2DBC4660ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldpost.mmh"},
	{0xAF976546F73E2E36ULL, "art\\vfx\\combat\\vfx_q_ari_charge_en_c.mmh"},
	{0xAF97C21A593B1221ULL, "133233770.ogg"},
	{0xAF97E88355EA204AULL, "6067812_m.fxe"},
	{0xAF98103DC6C08B07ULL, "6139826_m.fxe"},
	{0xAF988AC8617D383BULL, "hlac_w_bonepit_131078_0y.dds"},
	{0xAF98B60F4120C840ULL, "618682056.ogg"},
	{0xAF98CD589C05BDC0ULL, "693801049.ogg"},
	{0xAF99908709A5392EULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm01.msh"},
	{0xAF99AA8709A5651CULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm01.mmh"},
	{0xAF9A561B792DC41DULL, "dae000cr_osric.utc"},
	{0xAF9A75118A6AA4ECULL, "878607370.ogg"},
	{0xAF9A9411FBAD5512ULL, "229595345.ogg"},
	{0xAF9BC8657897FDB1ULL, "vxca_caverns1_k2t_0_1.vlm"},
	{0xAF9C3D2528BA28B8ULL, "vxdr_primevaltomb_6j1_0_1.vlm"},
	{0xAF9C6420B0288868ULL, "mer220_hubert6041202.cl"},
	{0xAF9C69E53536D883ULL, "194637650.ogg"},
	{0xAF9C6DCCF147DE42ULL, "610498064.ogg"},
	{0xAF9C7147B5990785ULL, "470577636.ogg"},
	{0xAF9CC2165AD623A2ULL, "ts_circleofmagi_arm_table.gda"},
	{0xAF9D116119409ABBULL, "gen_im_trs_krk_one_mac_r0.uti"},
	{0xAF9D195CA44DEE6DULL, "w_bonepitmining_a2_15_0.rml"},
	{0xAF9D26DE5BBF0804ULL, "mer120_hubert6059392.cl"},
	{0xAF9D6C6F0B9279DFULL, "6208187_f.fxe"},
	{0xAF9D8ED6C8CFA890ULL, "908843109.ogg"},
	{0xAF9D929B8534DB03ULL, "art\\characters\\creatures\\buzzard\\animation\\perched.gad"},
	{0xAF9DF6C543CE85F1ULL, "30329712.ogg"},
	{0xAF9E1F29415A3FBEULL, "6105864_m.fxe"},
	{0xAF9E59ADB1597CE5ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_3x.dds"},
	{0xAF9E7128C4F22E80ULL, "6148775_m.fxe"},
	{0xAF9E9234F735F92BULL, "cd_mrl350_merrill.wbd"},
	{0xAF9EC17F97A03488ULL, "gen00fl_sebastian6195811.cl"},
	{0xAF9ED7536F415BCAULL, "505698526.ogg"},
	{0xAF9F2BD6E9CF8EE4ULL, "vxw_encounter1_gx_0_0.vlm"},
	{0xAF9F2BE3F6B6968CULL, "174028169.ogg"},
	{0xAF9F9CBAED232850ULL, "llac_ht_keep_3735_1.dds"},
	{0xAF9FB6C5B3F3F30AULL, "gen00fl_merrill6115836.cl"},
	{0xAFA019ADA5EFE1FDULL, "rog_assassinsneakbonus.dds"},
	{0xAFA07E4CE1A23EB0ULL, "94823033.ogg"},
	{0xAFA0A56CA6401D7FULL, "and100_anders6206364.cl"},
	{0xAFA179B73737D26CULL, "dae000_gamlen6048483.cl"},
	{0xAFA1A93E3F3E4568ULL, "61046151.ogg"},
	{0xAFA1BBC5BA4C05D2ULL, "llac_w_sundermount_-131074_0.dds"},
	{0xAFA1D206C2AE617EULL, "qun111_ketojan6120454.cl"},
	{0xAFA1D582067857DAULL, "w_woundedcoast_213_0.phy"},
	{0xAFA1E6BB609ED73DULL, "dae000_meredith6085232.cl"},
	{0xAFA1EF672050E1D4ULL, "qun200_party_comment6095346.cl"},
	{0xAFA206CB5C954E35ULL, "vxw_encounter3_5i_0_0.vlm"},
	{0xAFA283DD8FF21AA4ULL, "vxlt_0_lowtown_n_ewf_0_0.vlm"},
	{0xAFA3206E0E10765AULL, "vxw_bonepitdestroyed_a3_7c_0_0.vlm"},
	{0xAFA33349E09F5B21ULL, "art\\vfx\\combat\\vfx_impact_p.impa.ani"},
	{0xAFA39D24CA9EEF43ULL, "dae000_male_commoner2_mob6174443.cl"},
	{0xAFA3BA87FEE11C37ULL, "art\\characters\\creatures\\dragon\\animation\\dblw_dmg_v1.gad"},
	{0xAFA3C784C1F9585DULL, "ter_sand01_s.dds"},
	{0xAFA3CF1A4453BB07ULL, "6093530_f.fxe"},
	{0xAFA3E2FEFB673BE9ULL, "6182707_m.fxe"},
	{0xAFA41D7B829DBD1FULL, "llac_lt_warehouse_0_6.dds"},
	{0xAFA496FC56D350CFULL, "water_bump_n.dds"},
	{0xAFA4AA91551E2088ULL, "vxw_bonepit_zv_0_0.vlm"},
	{0xAFA6139169581AADULL, "73976546.ogg"},
	{0xAFA656B77EC88217ULL, "6095746_m.fxe"},
	{0xAFA6A2358A4C9873ULL, "vxlt_gallowscourtyard_n_6yu_0_0.vlm"},
	{0xAFA6AFFDBD762427ULL, "vxht_mansion_n_5bn_0_0.vlm"},
	{0xAFA6EF6E18FDED92ULL, "da2terrain_terrain_p2.vsh.fxo.dx11"},
	{0xAFA6FD502EA99AE3ULL, "691715242.ogg"},
	{0xAFA7008CAB6E0833ULL, "270373384.ogg"},
	{0xAFA72D3C9FBAB779ULL, "qun231_varnell6099651.cl"},
	{0xAFA77FBD14078C27ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_4y.dds"},
	{0xAFA814AE5405F21CULL, "605061042.ogg"},
	{0xAFA83B27B07CCEFAULL, "cd_one300_nuncio.wbd"},
	{0xAFA92DA557592F01ULL, "6072273_m.fxe"},
	{0xAFA9CF87FEE5ABC4ULL, "art\\characters\\creatures\\dragon\\animation\\dblw_dmg_v1.evt"},
	{0xAFAA3E49E0A58CC3ULL, "art\\vfx\\combat\\vfx_impact_p.impa.gad"},
	{0xAFAB1A9383F5B0EBULL, "vxht_playershouse_n_1sx_0_0.vlm"},
	{0xAFABB820C336B4D6ULL, "539538793.ogg"},
	{0xAFABCC20F8C45D2DULL, "6016935_f.fxe"},
	{0xAFABF7814DD8280CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_calf_scratch.ani"},
	{0xAFABF9201AB8100EULL, "lt_0_lowtown_docks_n_241_0.mmh"},
	{0xAFAC05201AB824A5ULL, "lt_0_lowtown_docks_n_241_0.mao"},
	{0xAFAC13201AB83CFCULL, "lt_0_lowtown_docks_n_241_0.msh"},
	{0xAFAC24EA099F688AULL, "vxca_caverns2prc_9h4_0_0.vlm"},
	{0xAFAC8C2CC2C68AF8ULL, "6097552_m.fxe"},
	{0xAFAC8DC3F901C75CULL, "vxca_caverns2_4g7_0_0.vlm"},
	{0xAFAD83780DC7EA06ULL, "6114479_m.fxe"},
	{0xAFADB2B9B1CF2948ULL, "12822146.ogg"},
	{0xAFADCDC37F301DB7ULL, "dae000_merchant_tailor_hm6164355.cl"},
	{0xAFAE11D4B7075242ULL, "vxlt_backalley_n_1n0_0_0.vlm"},
	{0xAFAED8EE3E08101AULL, "vxca_cavern1_haunted_icl_0_2.vlm"},
	{0xAFAF242417D867B2ULL, "987761509.ogg"},
	{0xAFAF9FDE0760EDE8ULL, "ico_mm_hostilearrow.dds"},
	{0xAFAFC46D25053620ULL, "21793589.ogg"},
	{0xAFAFD26A2142794FULL, "gen00fl_bethany6101386.cl"},
	{0xAFAFF9E185839F26ULL, "6147783_m.fxe"},
	{0xAFB014E105D043D8ULL, "564897285.ogg"},
	{0xAFB096DE303210FAULL, "vxlt_0_lowtown_hax_0_0.vlm"},
	{0xAFB09F639B06FD82ULL, "follower_banter6155910.cl"},
	{0xAFB0A74F68950DB5ULL, "vxca_caverns2_8kc_0_1.vlm"},
	{0xAFB0F3927CE3A143ULL, "320279105.ogg"},
	{0xAFB0FD5E03068CD9ULL, "85077497.ogg"},
	{0xAFB11B49E0AB6FA8ULL, "art\\vfx\\combat\\vfx_impact_p.impa.evt"},
	{0xAFB1492B7C657224ULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_lp.gad"},
	{0xAFB1C709D4D8F80FULL, "vxw_woundedcoast_1cv_0_0.vlm"},
	{0xAFB2F67EFF1C1DD2ULL, "zz_followers_banter6157684.cl"},
	{0xAFB304814DDE5D4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_calf_scratch.gad"},
	{0xAFB3465C67D1E1FDULL, "250396338.ogg"},
	{0xAFB350C41723D5BBULL, "hlac_dr_0_deeproads_5839_0x.dds"},
	{0xAFB364982CDA8F2EULL, "ss_circlemages_m_1.bnk"},
	{0xAFB37B7966384352ULL, "llac_lt_0_lowtown_a2_n_196606_0.dds"},
	{0xAFB3BDAD921C0409ULL, "167832996.ogg"},
	{0xAFB400FDAEAF0417ULL, "w_sundermount_176_0.msh"},
	{0xAFB412FDAEAF237AULL, "w_sundermount_176_0.mao"},
	{0xAFB416FDAEAF2989ULL, "w_sundermount_176_0.mmh"},
	{0xAFB41BCD388218C2ULL, "149574187.ogg"},
	{0xAFB4DF6DBE3A3BFAULL, "6177425_m.fxe"},
	{0xAFB518B86721FBF3ULL, "hlac_w_bonepitmining_a2_196608_0y.dds"},
	{0xAFB543A1B514F281ULL, "137046255.ogg"},
	{0xAFB55EB985208F5AULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em06.msh"},
	{0xAFB570B98520ADA0ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em06.mmh"},
	{0xAFB5DED41FB744E8ULL, "222654997.ogg"},
	{0xAFB638CD168CE835ULL, "171597186.ogg"},
	{0xAFB6B202D2FA8D96ULL, "348823640.ogg"},
	{0xAFB6C681938C1E82ULL, "vxw_blightlands_q2_0_0.vlm"},
	{0xAFB712C52E8E60C4ULL, "gen00fl_dog6166826.cl"},
	{0xAFB7B36A0D5B6F78ULL, "vxdr_0_deeproads_cuz_0_0.vlm"},
	{0xAFB7E787FEF211BDULL, "art\\characters\\creatures\\dragon\\animation\\dblw_dmg_v1.ani"},
	{0xAFB80F4AE0BFD8A0ULL, "556960214.ogg"},
	{0xAFB8319CC76B3DEEULL, "vxlt_0_lowtown_gdd_0_0.vlm"},
	{0xAFB8582B7C6BAA02ULL, "art\\characters\\creatures\\ragedemon\\animation\\rst_lp.ani"},
	{0xAFB9364D340C72BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_45basic.msh"},
	{0xAFB9444D340C8AF3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_45basic.mao"},
	{0xAFB9484D340C9100ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_45basic.mmh"},
	{0xAFB9697BE58D578BULL, "midleveloadhints.gda"},
	{0xAFB96D6B6FCAE1DFULL, "vxlt_0_lowtown_a2_he9_0_0.vlm"},
	{0xAFB96DA15248CE3BULL, "vxht_brothel_n_5cv_0_0.vlm"},
	{0xAFB9BA8404C0D425ULL, "follower_banter6163503.cl"},
	{0xAFB9FF814DE4732DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_calf_scratch.evt"},
	{0xAFBA0F5F7BB738CFULL, "182840512.ogg"},
	{0xAFBAAD656E6C0B2FULL, "art\\characters\\playercharacter\\elfmale\\boo_lgtp.phy"},
	{0xAFBAF7ABEA10533FULL, "6129777_m.fxe"},
	{0xAFBAFCE034E364DCULL, "hlac_lt_0_lowtown_n_1_0y.dds"},
	{0xAFBB92F1C7ED6428ULL, "vxlt_gamlenshouse_n_88_0_0.vlm"},
	{0xAFBB9AB9C6AEBC31ULL, "vxlt_hangedman_n_25t_0_0.vlm"},
	{0xAFBBD9E39E0D8E47ULL, "hlac_w_bonepitdestroyed_a3_65538_0x.dds"},
	{0xAFBC27038EE2B3DAULL, "130408437.ogg"},
	{0xAFBC34562259AF5CULL, "vxw_bonepit_1it_0_0.vlm"},
	{0xAFBC702E253E1630ULL, "bink_bink_p1.psh.fxo.dx11.assembly"},
	{0xAFBC83B1A8FD6AC2ULL, "vxlt_backalley_n_1i1_0_0.vlm"},
	{0xAFBD0C13720417F0ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf06.mmh"},
	{0xAFBD1A1372042F0AULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf06.msh"},
	{0xAFBD4B8CB4BB1BB5ULL, "vxlt_0_lowtown_a2_erk_0_0.vlm"},
	{0xAFBD53F61A59ED6EULL, "vxdr_primevaltomb_8o7_0_0.vlm"},
	{0xAFBDC8E7EDDF330CULL, "and100_anders6111665.cl"},
	{0xAFBDF6546E4DA88EULL, "l_qun231ar_varnell_hideout.wbd"},
	{0xAFBE09E5FB1F2673ULL, "312087970.ogg"},
	{0xAFBE16F245A57F4AULL, "338071563.ogg"},
	{0xAFBE19148ED40173ULL, "vxdr_0_deeproads_f5l_0_0.vlm"},
	{0xAFBE5784FAAF2F25ULL, "vxca_caverns2_4wz_0_0.vlm"},
	{0xAFBEB20E5955A42AULL, "hlac_lt_0_lowtown_a3_n_c_-65535_4x.dds"},
	{0xAFBEEB65A9B5D647ULL, "vxlt_gallowscourtyard_74p_0_0.vlm"},
	{0xAFBF2F5D14D0BC25ULL, "6028429_m.fxe"},
	{0xAFBF451986B014D6ULL, "487322989.ogg"},
	{0xAFBF462313742B96ULL, "framebuffereffects_verticalblurshadows_p0.vsh.fxo.dx11"},
	{0xAFBF898709C5FCE7ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm01.phy"},
	{0xAFBF9EDB03CDD68EULL, "263023002.ogg"},
	{0xAFBFE54AD570750DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkout.msh"},
	{0xAFBFEA76C2EE3FEBULL, "hlac_w_blightlands_0_8x.dds"},
	{0xAFBFEAFB5609C6E3ULL, "w_bonepitmining_a2_252_0.msh"},
	{0xAFBFF34AD5708D40ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkout.mao"},
	{0xAFBFF74AD5709393ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkout.mmh"},
	{0xAFBFF8FB5609DE3DULL, "w_bonepitmining_a2_252_0.mmh"},
	{0xAFBFFCFB5609E54EULL, "w_bonepitmining_a2_252_0.mao"},
	{0xAFC038E6C6CF108FULL, "vxw_woundedcoast_1nq_0_0.vlm"},
	{0xAFC08AC42A4EF1E8ULL, "6095598_m.fxe"},
	{0xAFC08CE83000C267ULL, "vxlt_darktown_67j_0_0.vlm"},
	{0xAFC0C7EBAF4403F7ULL, "w_encounter3_118_0.mmh"},
	{0xAFC0CBEBAF440A04ULL, "w_encounter3_118_0.mao"},
	{0xAFC0DDEBAF442969ULL, "w_encounter3_118_0.msh"},
	{0xAFC1152EF57E2F0FULL, "hlac_ca_caverns1_mines_24422_3z.dds"},
	{0xAFC173B2A2CEACBCULL, "gen00fl_anders6121196.cl"},
	{0xAFC1C33217F62FD3ULL, "123235654.ogg"},
	{0xAFC1DBF02CF3A3A2ULL, "ser120_ghyslain6105490.cl"},
	{0xAFC200E485DB18C3ULL, "vxw_bonepitdestroyed_a3_19o_0_0.vlm"},
	{0xAFC2C1C3294B6230ULL, "mag100_cullen6095138.cl"},
	{0xAFC368687ED37056ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masa.msh"},
	{0xAFC372687ED38154ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masa.mmh"},
	{0xAFC37434F754E2F9ULL, "cd_mrl350_merrill.bnk"},
	{0xAFC37D3FDBC25983ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidleft.msh"},
	{0xAFC38B3FDBC271DDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidleft.mmh"},
	{0xAFC38F3FDBC278EEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gatemidleft.mao"},
	{0xAFC393707CB82FBEULL, "qun131_party_comment.cnv"},
	{0xAFC393B7A1625527ULL, "vxht_mansion_n_5hh_0_0.vlm"},
	{0xAFC3A1D11E3F265EULL, "6165181_m.fxe"},
	{0xAFC4C64DD209A7C7ULL, "65548908.ogg"},
	{0xAFC57C00460BC016ULL, "llac_w_woundedcoast_65541_0.dds"},
	{0xAFC6128DD9B2ADCDULL, "vxht_mansion2_3et_0_0.vlm"},
	{0xAFC6A379B961566EULL, "vxlt_0_lowtown_gce_0_0.vlm"},
	{0xAFC775707CBBE5CEULL, "qun131_party_comment.bnk"},
	{0xAFC79D121D300CE6ULL, "vxlt_0_lowtown_a3_n_c_gz6_0_0.vlm"},
	{0xAFC7DF4861C0FA5FULL, "vxlt_0_lowtown_a2_hb8_0_0.vlm"},
	{0xAFC8725A74AD57D5ULL, "follower_banter6163873.cl"},
	{0xAFC8933EFEB620F2ULL, "6077329_m.fxe"},
	{0xAFC96664EE0B9FC2ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_1x.dds"},
	{0xAFCA31E5A4424580ULL, "plot_core.ncs"},
	{0xAFCA3DCE761FC219ULL, "6087097_m.fxe"},
	{0xAFCA4022C50F8AF0ULL, "dre000_vampire6034346.cl"},
	{0xAFCA5450CDCA80D6ULL, "6084409_f.fxe"},
	{0xAFCA8F06CFC160BEULL, "hlac_dr_dungeon_3216_2x.dds"},
	{0xAFCAEEDDEA601FEEULL, "6109412_f.fxe"},
	{0xAFCB7FFCF489D16EULL, "llac_w_woundedcoast_-262136_0.dds"},
	{0xAFCB87065A590BB1ULL, "zz_mer_debug6051473.cl"},
	{0xAFCBA897F36F3B17ULL, "339390242.ogg"},
	{0xAFCBBD5C9E5DEF19ULL, "gen00fl_fenris6126400.cl"},
	{0xAFCC67433849C3CFULL, "537552376.ogg"},
	{0xAFCC78AF21B189F7ULL, "gen00fl_sebastian6179759.cl"},
	{0xAFCD5D429BABEEC7ULL, "vxlt_gallowscourtyard_79s_0_0.vlm"},
	{0xAFCD88FCA292BDC4ULL, "6112663_f.fxe"},
	{0xAFCDE1944BF75827ULL, "866207909.ogg"},
	{0xAFCDECEE2AC9D1EDULL, "vxw_bonepitdestroyed_a3_1go_0_0.vlm"},
	{0xAFCE2C38329EB6FFULL, "vxw_woundedcoast_1b5_0_0.vlm"},
	{0xAFCE474FB6FBF6FFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobc_0s.dds"},
	{0xAFCE4B846CA59A36ULL, "zz_followers_banter6159264.cl"},
	{0xAFCE5BA909F2B904ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_mered_chant_ex.ani"},
	{0xAFCEAAC3B8C5290FULL, "vxw_woundedcoast_1ip_0_0.vlm"},
	{0xAFCF13B437926909ULL, "dae000_cerimon6163681.cl"},
	{0xAFCF49B42CE9A75FULL, "vxca_caverns2prc_8jr_0_0.vlm"},
	{0xAFD000282B006A08ULL, "vxw_bonepit_ku_0_0.vlm"},
	{0xAFD001860472248AULL, "621745311.ogg"},
	{0xAFD058C71ED120A0ULL, "149214573.ogg"},
	{0xAFD0B1B2FB7379A1ULL, "6111206_m.fxe"},
	{0xAFD16B44C1F5D415ULL, "vxca_caverns1_j2f_0_2.vlm"},
	{0xAFD184A494ACBA0BULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.cess.ani"},
	{0xAFD1C3DAEE66C563ULL, "vxlt_hangedman_2xc_0_1.vlm"},
	{0xAFD204FB7869212DULL, "lt_0_lowtown_a3_n_c_52_0_ll.mmh"},
	{0xAFD2D70A45953B1CULL, "6129420_m.fxe"},
	{0xAFD2E6BE8F56ACA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ledgedropland.ani"},
	{0xAFD2E6FDEC1DECD3ULL, "vxw_sundermount_30q_0_0.vlm"},
	{0xAFD2EF9F8DC26985ULL, "ts_coterie_wpn_props.gda"},
	{0xAFD30C08266CAE14ULL, "vxca_caverns1_mines_imn_0_0.vlm"},
	{0xAFD3C18C559BD383ULL, "zz_mag_debug6050909.cl"},
	{0xAFD3EE29C6BD0FD4ULL, "6091693_m.fxe"},
	{0xAFD3F7E911B0A3B2ULL, "vxht_0_hightown_a3_bns_0_1.vlm"},
	{0xAFD42A59ED3AC853ULL, "art\\characters\\weapons\\maces\\w_mce_mc106a.phy"},
	{0xAFD42CB60C829BF9ULL, "vxlt_darktown_50z_0_0.vlm"},
	{0xAFD478BC36040FB5ULL, "vxht_0_hightown_n_b4a_0_0.vlm"},
	{0xAFD52DD9EA1AE8C6ULL, "gen00pt_combat.ncs"},
	{0xAFD534EAA053963CULL, "hlac_lt_0_lowtown_docks_a3_n_-2_3x.dds"},
	{0xAFD558A909F8D366ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_mered_chant_ex.gad"},
	{0xAFD5B7C9CDE73526ULL, "215337953.ogg"},
	{0xAFD5D590B1A14A90ULL, "dae000_madam_lusine6151487.cl"},
	{0xAFD5EE7D01DD21ECULL, "lt_0_lowtown_docks_n_154_0.phy"},
	{0xAFD5FD6E685CD794ULL, "313584408.ogg"},
	{0xAFD6FBC0FFD0699AULL, "6177911_m.fxe"},
	{0xAFD760A046804DA0ULL, "vxdr_primevalentrance_2ln_0_0.vlm"},
	{0xAFD7991E97CD0885ULL, "vxdr_0_deeproads_cxd_0_0.vlm"},
	{0xAFD7ABA711D4EA37ULL, "989514261.ogg"},
	{0xAFD7C747E7913F18ULL, "hlac_ca_caverns1_mines_24423_10z.dds"},
	{0xAFD7C9621DD3C9F1ULL, "vxht_keep_n_5pr_0_0.vlm"},
	{0xAFD7F24D57297A10ULL, "6076809_m.fxe"},
	{0xAFD85101A9132077ULL, "167580488.ogg"},
	{0xAFD86122A55FD822ULL, "t3_shd_hid.tnt"},
	{0xAFD86BCD9B292E89ULL, "6095850_m.fxe"},
	{0xAFD88234CCB7D893ULL, "dae000_sandal6180615.cl"},
	{0xAFD9421A88E401A1ULL, "332144190.ogg"},
	{0xAFD9664BED6EAF17ULL, "mrl351cr_fakechandan.utc"},
	{0xAFD968C0493DD87FULL, "6085278_f.fxe"},
	{0xAFD981A6E5B6D7FEULL, "hlac_w_encounter3_-262141_0z.dds"},
	{0xAFD98C5BA01322C5ULL, "740969145.ogg"},
	{0xAFD9C5011B3CC77DULL, "mag121_viveka_st.stg"},
	{0xAFD9FF488AAE3DA4ULL, "one200dg_solivitus6048224.cl"},
	{0xAFDA01BE8F5CF92BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ledgedropland.gad"},
	{0xAFDA2F4698A74CB5ULL, "vxlt_0_lowtown_a2_exi_0_0.vlm"},
	{0xAFDA3CB899F8B6EDULL, "vxdr_primevaltomb_8ns_0_0.vlm"},
	{0xAFDA81178B6A3DEBULL, "vxlt_undercityserialkiller_1c6_0_1.vlm"},
	{0xAFDAB90A4133000FULL, "cd_mag121_idunna.wbd"},
	{0xAFDAC46F23B79CB1ULL, "265614397.ogg"},
	{0xAFDAD8133645C905ULL, "244392258.ogg"},
	{0xAFDB2C47C4C350C3ULL, "mag370_meredith6172401.cl"},
	{0xAFDB3EFAC999BF12ULL, "dae000cr_grand_cleric6124741.cl"},
	{0xAFDB4A2A745CE248ULL, "trap_leghold.ncs"},
	{0xAFDB7B08DE282CEAULL, "mrl310_party_comments6143078.cl"},
	{0xAFDB7E04EFA28067ULL, "llac_ht_bartrand_n_3364_0.dds"},
	{0xAFDB8E8234A6A045ULL, "vxlt_0_lowtown_has_0_0.vlm"},
	{0xAFDB9D712736E5F7ULL, "hlac_ht_mansion_n_3370_0x.dds"},
	{0xAFDBE947020E1AA5ULL, "6019094_m.fxe"},
	{0xAFDC2013AD4FF7BDULL, "hlac_ht_0_hightown_n_-65536_1z.dds"},
	{0xAFDC52087300E048ULL, "6135308_m.fxe"},
	{0xAFDC5ECB1CBFEA6DULL, "vxw_bonepitdestroyed_a3_1hn_0_0.vlm"},
	{0xAFDC60CE9F12EFB7ULL, "vxw_woundedcoast_15d_0_0.vlm"},
	{0xAFDC67B3AD728B52ULL, "vxw_sundermount_30t_0_0.vlm"},
	{0xAFDC6E7F68246AFBULL, "6013291_f.fxe"},
	{0xAFDC9E152494CF7FULL, "vxw_woundedcoast_1m4_0_0.vlm"},
	{0xAFDCB9E8022187CEULL, "t3_wpn_klr.tnt"},
	{0xAFDCD2C54CD4DD2FULL, "vxw_blightlands_sy_0_0.vlm"},
	{0xAFDD46E595B44FF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_dstrk.ani"},
	{0xAFDD88F971432FE3ULL, "6116642_f.fxe"},
	{0xAFDDB88495021946ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_1y.dds"},
	{0xAFDE0CA1735FBF5CULL, "vxht_0_hightown_a2_cb0_0_0.vlm"},
	{0xAFDE15AD17730C1DULL, "6118029_m.fxe"},
	{0xAFDEBB51AF290322ULL, "lt_0_lowtown_docks_a3_n_240_0.msh"},
	{0xAFDEBD51AF2906D8ULL, "lt_0_lowtown_docks_a3_n_240_0.mmh"},
	{0xAFDEC951AF291B7BULL, "lt_0_lowtown_docks_a3_n_240_0.mao"},
	{0xAFDEF2DA16366B44ULL, "dae_rogue_singletarget.ncs"},
	{0xAFDEF43740C9EE5BULL, "852360184.ogg"},
	{0xAFDF05B45BB714ECULL, "630844313.ogg"},
	{0xAFDF2E5E776C88A1ULL, "gen00fl_varric6033603.cl"},
	{0xAFDF3EFE589E4688ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med1_c.anb"},
	{0xAFDF6CA494B8CE82ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.cess.evt"},
	{0xAFDFBE6C10CBC16DULL, "338725577.ogg"},
	{0xAFDFFEBE8F615F90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ledgedropland.evt"},
	{0xAFE01B60220198D4ULL, "3702800.ogg"},
	{0xAFE02D97CC72E98AULL, "6149081_f.fxe"},
	{0xAFE04DE7D4A52AC3ULL, "vxca_cavern1_haunted_ikk_0_0.vlm"},
	{0xAFE076C2859C0F8BULL, "gen00fl_anders6112995.cl"},
	{0xAFE0999C58D62856ULL, "w_sundermount_54_0.msh"},
	{0xAFE0A39C58D63954ULL, "w_sundermount_54_0.mmh"},
	{0xAFE0A79C58D64067ULL, "w_sundermount_54_0.mao"},
	{0xAFE0C7C5B2FFC9FAULL, "200478267.ogg"},
	{0xAFE134A8B33E44BAULL, "vxdr_primevalentrance_234_0_1.vlm"},
	{0xAFE16E7AE6F59059ULL, "vxht_0_hightown_bya_0_0.vlm"},
	{0xAFE17DE51862C694ULL, "vxca_caverns1_mines_ijo_0_0.vlm"},
	{0xAFE17FA65B3C0065ULL, "hlac_lt_0_lowtown_a3_n_c_65537_1y.dds"},
	{0xAFE1E533BC49E452ULL, "vxw_encounter2_qx_0_1.vlm"},
	{0xAFE1EA00F8C81324ULL, "561338909.ogg"},
	{0xAFE2BCC534E0E402ULL, "qun211_party_comment6072244.cl"},
	{0xAFE3E668392F3A10ULL, "hlac_lt_0_lowtown_a2_65536_3y.dds"},
	{0xAFE4130CF4BDBF5AULL, "llac_w_woundedcoast_-131069_0.dds"},
	{0xAFE476DA2BB4B285ULL, "gen_im_trs_dwv_dua_dag_r0.uti"},
	{0xAFE563DFA248E6A1ULL, "6121014_m.fxe"},
	{0xAFE568C4354D18D9ULL, "hlac_lt_warehouse_65536_3x.dds"},
	{0xAFE5B2F83FD99965ULL, "73532674.ogg"},
	{0xAFE60AFB89C32105ULL, "vxdr_0_deeproads_cuc_0_0.vlm"},
	{0xAFE6207F970381FFULL, "762280187.ogg"},
	{0xAFE66BA494BEEB95ULL, "art\\vfx\\combat\\vfx_m_wpn_spi_crust_c.cess.gad"},
	{0xAFE691A42D163DE4ULL, "fxe_water_form.dds"},
	{0xAFE6B78C27B94E40ULL, "follower_banter6173048.cl"},
	{0xAFE6BA05798DE337ULL, "vxca_caverns1_mines_j2e_0_2.vlm"},
	{0xAFE6FEAEC305A41CULL, "765919377.ogg"},
	{0xAFE6FF54110A1CE0ULL, "vxlt_0_lowtown_a2_gpb_0_0.vlm"},
	{0xAFE76F976852978AULL, "lgt260_eau_tmp_angelique6143960.cl"},
	{0xAFE77B79A665DC8CULL, "vxht_0_hightown_a84_0_0.vlm"},
	{0xAFE7C35BBE013784ULL, "dae000_arishok6068444.cl"},
	{0xAFE8990BA276ACC3ULL, "mrl310_merrill6120009.cl"},
	{0xAFE8AE958BEECF6DULL, "vxdr_primevaltomb_8ir_0_0.vlm"},
	{0xAFE8B4BBCFFF25EAULL, "688534862.ogg"},
	{0xAFE8D0018675BF2EULL, "113835564.ogg"},
	{0xAFE9373E2FE9CA5CULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.cess.ani"},
	{0xAFE9D98B2071E607ULL, "llac_lt_0_lowtown_131074_0.dds"},
	{0xAFE9DF381B40739EULL, "122542231.ogg"},
	{0xAFE9EC546E72D2BCULL, "l_qun231ar_varnell_hideout.bnk"},
	{0xAFEA298B28947864ULL, "vxw_bonepitdestroyed_a3_28h_0_0.vlm"},
	{0xAFEA4EE595BEE7E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_dstrk.evt"},
	{0xAFEA6E19436DF581ULL, "vxca_cavern1_haunted_cjt_0_0.vlm"},
	{0xAFEA7F22CCC5A708ULL, "fen310_fenris6118203.cl"},
	{0xAFEAD0A80EB602EDULL, "vxw_bonepitdestroyed_a3_1mm_0_0.vlm"},
	{0xAFEAD2AB91090837ULL, "vxw_woundedcoast_10g_0_0.vlm"},
	{0xAFEB44A23ECAF5AFULL, "vxw_blightlands_tx_0_0.vlm"},
	{0xAFEC27E8A00FC371ULL, "vxdr_slavers_1sq_0_0.vlm"},
	{0xAFEC2D7F5957358FULL, "55910571.ogg"},
	{0xAFEC66E2510CB5AFULL, "vxw_sundermount_2ge_0_0.vlm"},
	{0xAFEC8DB90E2B554AULL, "hlac_ht_0_hightown_n_0_2x.dds"},
	{0xAFED68DE05DA9AFEULL, "6141082_m.fxe"},
	{0xAFEE3BB610ACB7B5ULL, "85118633.ogg"},
	{0xAFEEC76D46AB6A6EULL, "art\\vfx\\combat\\vfx_gib_ogr_p_3.mao"},
	{0xAFEEF012AAA0B6BDULL, "hlac_ca_caverns1_24425_2y.dds"},
	{0xAFEF20AEE6F228A0ULL, "605798469.ogg"},
	{0xAFEF27E4E5DE9E6FULL, "vxht_mansion2b_1hc_0_0.vlm"},
	{0xAFEF443E2FEE4B9EULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.cess.gad"},
	{0xAFEF565C9A192E93ULL, "vxlt_backalley_g_1wg_0_0.vlm"},
	{0xAFEF5F823378B0DFULL, "vxdr_primevalentrance_2qj_0_0.vlm"},
	{0xAFEFEFC20A58DF14ULL, "vxca_caverns1_mines_igl_0_0.vlm"},
	{0xAFF04B96EAAC5FB2ULL, "gen_im_trs_slv_hvy_hlm_r1.uti"},
	{0xAFF07A69047AB4CEULL, "vxht_chantry_n_2t7_0_1.vlm"},
	{0xAFF08114A2C6D416ULL, "vxw_bonepitmining_a2_ff_0_0.vlm"},
	{0xAFF0927107AF7EA1ULL, "vxw_bonepitdestroyed_a3_89_0_0.vlm"},
	{0xAFF1209A54C006EBULL, "gen00fl_aveline6132723.cl"},
	{0xAFF16FE595C53E30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_dstrk.gad"},
	{0xAFF21D6CE2129061ULL, "6171820_m.fxe"},
	{0xAFF2589A8AA91520ULL, "vxlt_hangedman_n_15s_0_1.vlm"},
	{0xAFF2C54621F034A5ULL, "280064294.ogg"},
	{0xAFF33CDF704FC354ULL, "dae000_bartrand6201383.cl"},
	{0xAFF341AA4C299B03ULL, "155445010.ogg"},
	{0xAFF40315F613A33FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_004a_0d.dds"},
	{0xAFF47CD87BB93985ULL, "vxdr_0_deeproads_crb_0_0.vlm"},
	{0xAFF4894E81CFC3C3ULL, "w_sundermount_83_0.rml"},
	{0xAFF5713103003560ULL, "vxlt_0_lowtown_a2_gwc_0_0.vlm"},
	{0xAFF5B000038B3909ULL, "follower_banter6141572.cl"},
	{0xAFF5C26909615A2CULL, "vxht_keep_6gf_0_0.vlm"},
	{0xAFF61F3E2FF42BBDULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.cess.evt"},
	{0xAFF624DFDC82E067ULL, "735249993.ogg"},
	{0xAFF64D87E4021BCAULL, "vxw_encounter2_nx_0_0.vlm"},
	{0xAFF65C8B3E2CEEBDULL, "6082270_m.fxe"},
	{0xAFF69D576513E94CULL, "hlac_ca_caverns2_5802_16x.dds"},
	{0xAFF77111A597D075ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_flowergroupb.msh"},
	{0xAFF77311A597D3EBULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_flowergroupb.mmh"},
	{0xAFF83D0A08450FAEULL, "vxht_0_hightown_a3_ccm_0_1.vlm"},
	{0xAFF8F8A02D56E9BBULL, "6172386_m.fxe"},
	{0xAFF981CF0881007FULL, "vxw_woundedcoast_1s6_0_0.vlm"},
	{0xAFFA6F49760C3A14ULL, "6104578_f.fxe"},
	{0xAFFAF1739F219999ULL, "1011434364.ogg"},
	{0xAFFAF966FB5F3D1DULL, "6112229_m.fxe"},
	{0xAFFB135C621C10F1ULL, "pro000_captain_wright6047473.cl"},
	{0xAFFB6823919C63C8ULL, "dae000_flemeth6000877.cl"},
	{0xAFFB904E81D5EE3BULL, "w_sundermount_83_0.phy"},
	{0xAFFBC61320A8AC1FULL, "zz_mer_debug6062741.cl"},
	{0xAFFBFB691FD06218ULL, "6156369_m.fxe"},
	{0xAFFC0345753904E5ULL, "904549174.ogg"},
	{0xAFFC9A1FB1F93270ULL, "hlac_lt_gallowstemplar_f_2383_7x.dds"},
	{0xAFFCC9DDD22BA958ULL, "698175895.ogg"},
	{0xAFFCD48D54182896ULL, "6030676_m.fxe"},
	{0xAFFCDC0781850F05ULL, "llac_lt_0_lowtown_docks_a3_-2_5.dds"},
	{0xAFFD3B57ABD5BB41ULL, "vxca_caverns2_8ba_0_0.vlm"},
	{0xAFFD99C1D7D4B6EFULL, "vxht_mansion2b_1cb_0_0.vlm"},
	{0xAFFDB2564FD2F3E2ULL, "vxht_brothel_n_5k3_0_0.vlm"},
	{0xAFFDCE8206C6AA03ULL, "w_woundedcoast_213_0.msh"},
	{0xAFFDDC8206C6C25DULL, "w_woundedcoast_213_0.mmh"},
	{0xAFFDE08206C6C96EULL, "w_woundedcoast_213_0.mao"},
	{0xAFFDF6570215D954ULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s2_c.mmh"},
	{0xAFFE1313723BBA0BULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf06.phy"},
	{0xAFFE435159C84670ULL, "dae000_sabina6161502.cl"},
	{0xAFFEF2F194BCEC96ULL, "vxw_bonepitmining_a2_mg_0_0.vlm"},
	{0xAFFF21664911421AULL, "dae000_bodhan6146811.cl"},
	{0xB0000E83C0162B88ULL, "zz_prmpt_debug6198251.cl"},
	{0xB000239757642E2CULL, "qun221_petrice6083496.cl"},
	{0xB00076B11AD9C081ULL, "vxht_0_hightown_a2_cba_0_0.vlm"},
	{0xB0014C0D617908B2ULL, "vxca_caverns1_iie_0_0.vlm"},
	{0xB001E034344FBB22ULL, "vxlt_gallowstemplar_f_3sy_0_0.vlm"},
	{0xB001EC1FFBC62BF2ULL, "qcr000_arishok6059338.cl"},
	{0xB001ED77953D7036ULL, "zz_cave_accolades_2_v1.cut"},
	{0xB00223BCBB00E784ULL, "vxca_caverns2_5xn_0_1.vlm"},
	{0xB0022C7A821021FFULL, "llac_w_sundermount_-262144_0.dds"},
	{0xB0022D6527F8C8C7ULL, "509261376.ogg"},
	{0xB00286EE866B9F55ULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpa.phy"},
	{0xB002A78F4B652B6AULL, "dae300_alistair_king6134991.cl"},
	{0xB002E1AC0B2A6C99ULL, "vxht_mansion_n_6j7_0_0.vlm"},
	{0xB0031AD0E89982DBULL, "vxht_keep_n_6ux_0_0.vlm"},
	{0xB0033D69C9EEF016ULL, "w_woundedcoast_33_0.rml"},
	{0xB0037067542515E6ULL, "follower_banter6155547.cl"},
	{0xB003A44C34A06431ULL, "mer110_jansen6032921.cl"},
	{0xB003D0F4AD4FA4E3ULL, "vxht_mansion_n_6ny_0_0.vlm"},
	{0xB003D9CEB3A875D3ULL, "vxlt_undercityencounter_124_0_0.vlm"},
	{0xB00410450F42DB1EULL, "34307966.ogg"},
	{0xB004145606D8FFB8ULL, "pro000_colonel_ewald6005397.cl"},
	{0xB0043EDE7D68D513ULL, "gen_im_arm_cht_nobd_t1.uti"},
	{0xB0044305665DAA5FULL, "hlac_lt_0_lowtown_docks_n_-196610_2x.dds"},
	{0xB004BF64D5F8344AULL, "vxw_encounter2_iy_0_0.vlm"},
	{0xB005197F9D717A74ULL, "plt_cod_itm_shield_unspoken.plo"},
	{0xB0051BBD64B49D3EULL, "sky_ht_0_hightown.mao"},
	{0xB0059BEE6B8FFE3CULL, "zz_rockwraith.cut"},
	{0xB0061A3767800036ULL, "one300_zevran6076944.cl"},
	{0xB0069080924AE56FULL, "vxlt_0_lowtown_a2_n_f6p_0_1.vlm"},
	{0xB006F9BBB0F6ECCBULL, "vxlt_foundry_n_11g_0_0.vlm"},
	{0xB0073DFE58BF9723ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med1_c.mmh"},
	{0xB0078FBB9C18C019ULL, "hlac_w_bonepitdestroyed_a3_393217_0x.dds"},
	{0xB0079C348359440EULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.mmh"},
	{0xB007E5B204B77935ULL, "vxht_keep_n_6gp_0_0.vlm"},
	{0xB00824F462C2AD9CULL, "qun160_party_comment_26061246.cl"},
	{0xB0085CE5BB88C4E4ULL, "gen_im_npc_arm_hum_cht_cotc.uti"},
	{0xB0085FACC7F93189ULL, "hlac_w_woundedcoast_196616_0y.dds"},
	{0xB008BB652F2BDA87ULL, "dae000_hate_follower_react6103193.cl"},
	{0xB008DAF07A63B70FULL, "6202212_m.fxe"},
	{0xB009497010396DCBULL, "6218732_m.fxe"},
	{0xB0094A9C34F8E6AFULL, "vxw_sundermount_2mg_0_0.vlm"},
	{0xB009769942E386B9ULL, "mrl351_pridedemon6165555.cl"},
	{0xB00993393FFBBEDEULL, "208929789.ogg"},
	{0xB009A69C58F9430BULL, "w_sundermount_54_0.anb"},
	{0xB009C1CEEA155EF4ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v3.impa.ani"},
	{0xB009EE55AB9AF294ULL, "hlac_w_encounter1_0_3z.dds"},
	{0xB00A0231028AC5CEULL, "halfroundcombatanimsb.gda"},
	{0xB00A09432CDEB488ULL, "14013593.ogg"},
	{0xB00A779DFD951EF3ULL, "270793816.ogg"},
	{0xB00B11BE20F4A503ULL, "6176728_m.fxe"},
	{0xB00BA7E93904CD4AULL, "vxlt_undercitydungeon_1m9_0_0.vlm"},
	{0xB00BB251ED689E1CULL, "3949964.ogg"},
	{0xB00C0B9EC9CACF6FULL, "vxht_mansion2b_1fm_0_0.vlm"},
	{0xB00C3A167E055F93ULL, "vxlt_backalley_g_1yi_0_0.vlm"},
	{0xB00C864E666EAFCBULL, "gen_im_arm_hlm_med_act2_03.uti"},
	{0xB00CDA98D8DDE321ULL, "vxw_sundermount_37b_0_0.vlm"},
	{0xB00CF5FBBB5BF403ULL, "vxdr_primevaltomb_740_0_0.vlm"},
	{0xB00D3025570F0F16ULL, "810262447.ogg"},
	{0xB00D541A1B475ED2ULL, "art\\characters\\creatures\\mubari\\animation\\pause2eat.gad"},
	{0xB00D5A8474FD1452ULL, "vxlt_undercityencounter_121_0_0.vlm"},
	{0xB00D65B5FAAEB490ULL, "fex_bd_c_rf.mmh"},
	{0xB00D66F831389F7CULL, "vxca_caverns1_jfu_0_0.vlm"},
	{0xB00D6787016EDD77ULL, "hlac_lt_backalley_g_1_3y.dds"},
	{0xB00D6DE79C037E49ULL, "llac_ht_0_hightown_a2_131072_0.dds"},
	{0xB00D73B5FAAECCAAULL, "fex_bd_c_rf.msh"},
	{0xB00D7D55029F3E55ULL, "270293174.ogg"},
	{0xB00ECFD6532578ADULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_desk.mmh"},
	{0xB00ED3D653257FFEULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_desk.mao"},
	{0xB00EE1D653259733ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_desk.msh"},
	{0xB00FBECEEA19C556ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v3.impa.gad"},
	{0xB010F7165D38FB59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\fire_death.ani"},
	{0xB0115388FD208519ULL, "vxht_mansion_n_6m6_0_0.vlm"},
	{0xB011D28FD7ED1B35ULL, "6152575_m.fxe"},
	{0xB012036FF282A2E0ULL, "dae000_flemeth6013415.cl"},
	{0xB012E44858C47C63ULL, "lgt300cr_gang_1_lt.utc"},
	{0xB013571A1B4BCF09ULL, "art\\characters\\creatures\\mubari\\animation\\pause2eat.evt"},
	{0xB013E3E72EC7D08DULL, "6052994_m.fxe"},
	{0xB014042C61D118EDULL, "vxdr_primevaltomb_8zo_0_0.vlm"},
	{0xB0143EF47C0D10B7ULL, "vxca_caverns1_fm7_0_0.vlm"},
	{0xB014AC532D806418ULL, "6087319_m.fxe"},
	{0xB015660427DC5C9CULL, "w_sundermount_112_0.mao"},
	{0xB015720427DC713FULL, "w_sundermount_112_0.mmh"},
	{0xB015780427DC7BE1ULL, "w_sundermount_112_0.msh"},
	{0xB015BB348365B60DULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.anb"},
	{0xB015F730AF51DFEAULL, "vxdr_primevaltomb_7bw_0_0.vlm"},
	{0xB01626F1363A43E5ULL, "vxw_bonepitdestroyed_a3_9b_0_0.vlm"},
	{0xB0164B38030C6F54ULL, "6119218_m.fxe"},
	{0xB0166C9474884A53ULL, "zz_followers_banter6159483.cl"},
	{0xB016AB5AA4F225C4ULL, "lt_0_lowtown_docks_n_199_0.msh"},
	{0xB016BD5AA4F2442DULL, "lt_0_lowtown_docks_n_199_0.mao"},
	{0xB016C15AA4F24B66ULL, "lt_0_lowtown_docks_n_199_0.mmh"},
	{0xB016C57D021472B7ULL, "lt_0_lowtown_docks_n_154_0.mmh"},
	{0xB016C97D021479C4ULL, "lt_0_lowtown_docks_n_154_0.mao"},
	{0xB016D9CEEA2011B5ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v3.impa.evt"},
	{0xB016DB7D02149829ULL, "lt_0_lowtown_docks_n_154_0.msh"},
	{0xB016DE0DA4B27072ULL, "96176487.ogg"},
	{0xB016F63031357E61ULL, "mcr310_confused_guard6077016.cl"},
	{0xB0170D205766AABCULL, "819702823.ogg"},
	{0xB017232F74749117ULL, "plt_isa300pt_find_isabela.plo"},
	{0xB017BC7926EEFF2FULL, "vxw_sundermount_2vf_0_0.vlm"},
	{0xB017D2165D3EDA9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\fire_death.gad"},
	{0xB0195D97667BD632ULL, "hlac_lt_0_lowtown_a2_n_1_0y.dds"},
	{0xB0198505AD0F4C08ULL, "hlac_lt_0_lowtown_a2_65537_3z.dds"},
	{0xB0199A3D7C0636EEULL, "345710736.ogg"},
	{0xB019F84BB7FD0550ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_vael.mmh"},
	{0xB01A064BB7FD1D6AULL, "art\\characters\\playercharacter\\humanmale\\hm_har_vael.msh"},
	{0xB01AD30A416910E5ULL, "cd_mag121_idunna.bnk"},
	{0xB01B4BD13839F8D4ULL, "llac_ca_caverns1_mines_24425_12.dds"},
	{0xB01B4C75CAD3FBA1ULL, "vxw_sundermount_32a_0_0.vlm"},
	{0xB01B60654412CE52ULL, "wos_04.mao"},
	{0xB01C40F19EFB0700ULL, "water_water_p1.psh.fxo.dx11"},
	{0xB01C9E6F345BC6C3ULL, "ser191_emeric6120559.cl"},
	{0xB01CBAD91E8E4387ULL, "zz_otl_debug6060661.cl"},
	{0xB01CBCF7327917B9ULL, "zz_followers_banter6157968.cl"},
	{0xB01CDA7321974901ULL, "6087216_f.fxe"},
	{0xB01D210C4F2D92B6ULL, "vxw_bonepitdestroyed_a3_4v_0_0.vlm"},
	{0xB01D7CD677A72C1DULL, "ht_0_hightown_a3_43_0_lc.mmh"},
	{0xB01E1BC6DFBDA027ULL, "vxw_encounter2_camp_9p_0_0.vlm"},
	{0xB01E615B334D07B0ULL, "6099666_m.fxe"},
	{0xB01E6B1674B883BEULL, "406593890.ogg"},
	{0xB01EEF165D452A80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\fire_death.evt"},
	{0xB01EFB3B1479CE2BULL, "follower_banter6151269.cl"},
	{0xB01F1C02FAE49842ULL, "1027737104.ogg"},
	{0xB01F3AB87D026765ULL, "6134979_m.fxe"},
	{0xB01F71687F21DE4FULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masa.phy"},
	{0xB01FC39A48C2E94BULL, "hlac_ca_caverns2_5802_37z.dds"},
	{0xB01FC8D65333E06AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_desk.phy"},
	{0xB01FD26F519B8305ULL, "vxdr_0_deeproads_cao_0_0.vlm"},
	{0xB01FFE8ACC85B3DBULL, "vxht_keep_n_6kz_0_0.vlm"},
	{0xB02002B2D7A24471ULL, "vxht_keep_n_5iu_0_0.vlm"},
	{0xB0201C896BCF09C1ULL, "6072880_m.fxe"},
	{0xB0203BB65CE2564BULL, "812613260.ogg"},
	{0xB0208A08147D87D0ULL, "w_bonepitdestroyed_a3_16_0.mao"},
	{0xB0208E08147D8EE3ULL, "w_bonepitdestroyed_a3_16_0.mmh"},
	{0xB0209C08147DA63DULL, "w_bonepitdestroyed_a3_16_0.msh"},
	{0xB0213F1A1B57E478ULL, "art\\characters\\creatures\\mubari\\animation\\pause2eat.ani"},
	{0xB021EE92F28702B4ULL, "566758210.ogg"},
	{0xB0222A7845FFC16AULL, "6119267_m.fxe"},
	{0xB02237AE9970816EULL, "vxht_chantry_n_30c_0_0.vlm"},
	{0xB022760953C7316DULL, "vxdr_primevaltomb_8un_0_0.vlm"},
	{0xB022A0F07CDE996CULL, "6069580_m.fxe"},
	{0xB02309EC49E7C691ULL, "6087343_m.fxe"},
	{0xB0233F03F860C277ULL, "6137510_m.fxe"},
	{0xB02425C5BFE2D4FAULL, "vxlt_0_lowtown_h9p_0_0.vlm"},
	{0xB024BD14F50287D4ULL, "6114518_m.fxe"},
	{0xB024C4D6FBBCD45EULL, "934639457.ogg"},
	{0xB0260356B3D1390EULL, "554743516.ogg"},
	{0xB02615FC045CEBA1ULL, "sst_ss_circlemages_m_2.gda"},
	{0xB026F0C9AA03CBE8ULL, "hlac_ca_caverns2_5802_2y.dds"},
	{0xB026F9570238E30BULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s2_c.anb"},
	{0xB0274AB8C33C5D56ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.impa.evt"},
	{0xB0276DF583AFC4ADULL, "6150991_m.fxe"},
	{0xB027774E81FAFEDAULL, "w_sundermount_83_0.msh"},
	{0xB027854E81FB1613ULL, "w_sundermount_83_0.mao"},
	{0xB027894E81FB1D20ULL, "w_sundermount_83_0.mmh"},
	{0xB0279617C6C5C72DULL, "974183249.ogg"},
	{0xB027C9FAA13001EBULL, "one101_orders_followers6115781.cl"},
	{0xB027D9FFADBE6EE7ULL, "6080387_f.fxe"},
	{0xB027DD53C9566713ULL, "mag311_nyssa6051145.cl"},
	{0xB028B173C5DAE944ULL, "zz_followers_banter6159132.cl"},
	{0xB028FFC169B1AA44ULL, "mag170_thrask6016039.cl"},
	{0xB029384B7EEE526CULL, "dae000ip_crowd_mf5_talk16.utp"},
	{0xB02950BAF3740D08ULL, "pro000ip_corpse_nonint.utp"},
	{0xB02A0678E90B72EFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupc.mao"},
	{0xB02A1278E90B868CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupc.mmh"},
	{0xB02A1878E90B917EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockgroupc.msh"},
	{0xB02A74CDA263E829ULL, "zz_followers_banter6157278.cl"},
	{0xB02A94C57EEE0DCCULL, "vxlt_darktown_2s5_0_0.vlm"},
	{0xB02ABCE9515DC867ULL, "786122485.ogg"},
	{0xB02AC216993A1DCCULL, "vxlt_0_lowtown_docks_a3_art_0_0.vlm"},
	{0xB02AC3EBD8737166ULL, "262251804.ogg"},
	{0xB02B322F7485538CULL, "plt_isa300pt_find_isabela.nss"},
	{0xB02B524461DDA91BULL, "6080974_f.fxe"},
	{0xB02C1C0263E40769ULL, "334781005.ogg"},
	{0xB02CBE86C1A0399FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_wf_injured.gad"},
	{0xB02D47CF62BD43E7ULL, "w_sundermount_43_0.rml"},
	{0xB02D9BDB64F2E43AULL, "370727730.ogg"},
	{0xB02DD7DD029B5D8AULL, "hlac_lt_gallowstemplar_n_2394_5x.dds"},
	{0xB02E2D46773966D7ULL, "603795001.ogg"},
	{0xB02E41B8C3426CE1ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.impa.gad"},
	{0xB02E753F13DF577AULL, "hlac_ht_mansion2b_818_67y.dds"},
	{0xB02E8A4A783F2885ULL, "415187588.ogg"},
	{0xB02FA09427948EBEULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_throw.ani"},
	{0xB02FA5E5DA9F1E99ULL, "vxlt_undercitydungeon_15n_0_0.vlm"},
	{0xB0301D604D412DC0ULL, "lgt260_eau_tmp_roderick6143715.cl"},
	{0xB03045F13143A240ULL, "art\\characters\\playercharacter\\humanmale\\boo_magc.phy"},
	{0xB0307FF063075ED5ULL, "500216063.ogg"},
	{0xB03172402CE64568ULL, "6107208_m.fxe"},
	{0xB031BF0CD1E28927ULL, "6087161_m.fxe"},
	{0xB03234D4F7CD68ADULL, "986453460.ogg"},
	{0xB032D443ABDAFA7FULL, "734449965.ogg"},
	{0xB0330E6906D7E576ULL, "vxlt_0_lowtown_a2_n_c_fi7_0_0.vlm"},
	{0xB0337FF01BD358DCULL, "6072830_m.fxe"},
	{0xB033B10785192E3FULL, "6165632_m.fxe"},
	{0xB033E4244BBA5417ULL, "vxdr_primevalentrance_26w_0_0.vlm"},
	{0xB03442B8C346D98FULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.impa.ani"},
	{0xB03446CF62C360FFULL, "w_sundermount_43_0.phy"},
	{0xB0344CC0107F756FULL, "6179581_m.fxe"},
	{0xB03470DBF0574F47ULL, "224512490.ogg"},
	{0xB034D46AF9229544ULL, "hlac_dr_dungeon_3214_14y.dds"},
	{0xB03515049FCCBE37ULL, "prm000im_lothering_tome.uti"},
	{0xB0356EC3196FBCE1ULL, "40198940.ogg"},
	{0xB0358127BD40664BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_staff_swarm.evt"},
	{0xB035A39C591E7957ULL, "w_sundermount_54_0.rml"},
	{0xB035DE6C955F05A5ULL, "287994327.ogg"},
	{0xB0367A841BCCAB69ULL, "hlac_w_bonepit_327684_0y.dds"},
	{0xB036D22E2EFC63A0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodburn_n.dds"},
	{0xB0372359ED8F39C8ULL, "art\\characters\\weapons\\maces\\w_mce_mc106a.mmh"},
	{0xB0374159ED8F6CF2ULL, "art\\characters\\weapons\\maces\\w_mce_mc106a.msh"},
	{0xB03761359FAD18EFULL, "vxht_mansion2b_1wn_0_0.vlm"},
	{0xB037625EFA396479ULL, "dae000_worthy6166648.cl"},
	{0xB037704A32A931A7ULL, "hlac_w_bonepitmining_a2_327683_0z.dds"},
	{0xB037CFBF9B08F637ULL, "zz_followers_banter6158967.cl"},
	{0xB0380F39D865CB93ULL, "hlac_w_encounter3_-262144_0x.dds"},
	{0xB038646BFB5820BAULL, "follower_barks6153313.cl"},
	{0xB0388E0814922000ULL, "w_bonepitdestroyed_a3_16_0.rml"},
	{0xB038AC1370595720ULL, "hlac_lt_0_lowtown_docks_131070_2x.dds"},
	{0xB038F84EE76679B6ULL, "230706275.ogg"},
	{0xB0392F343B936F5CULL, "6031119_f.fxe"},
	{0xB0396A2C4ACA5F97ULL, "hlac_lt_0_lowtown_a2_n_-2_3y.dds"},
	{0xB03A761303B0B5C8ULL, "86988036.ogg"},
	{0xB03A973A6D47876FULL, "333361622.ogg"},
	{0xB03B0E8AFC153C0EULL, "6064725_m.fxe"},
	{0xB03B2DC7A1588ECBULL, "974819579.ogg"},
	{0xB03B5016B59A2094ULL, "424513845.ogg"},
	{0xB03B88B472116B4DULL, "vxw_encounter3_bt_0_0.vlm"},
	{0xB03BC37135A5724BULL, "docks_male_sailor_mob6176650.cl"},
	{0xB03C55DBAB9D2A28ULL, "hlac_lt_0_lowtown_docks_-131076_1x.dds"},
	{0xB03C6AF9B2BDBA99ULL, "hlac_w_blightlands_393214_0x.dds"},
	{0xB03C8B0C25BC165FULL, "6068451_f.fxe"},
	{0xB03C98A746D7FEC7ULL, "vxw_bonepitdestroyed_a3_19s_0_0.vlm"},
	{0xB03C9A27BD46AF0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_staff_swarm.gad"},
	{0xB03CA29C5924964FULL, "w_sundermount_54_0.phy"},
	{0xB03CA707291B13AEULL, "hlac_w_woundedcoast_0_4y.dds"},
	{0xB03CB894279F418FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_throw.evt"},
	{0xB03CDBA395B63FC9ULL, "dae300_vashothlead6062824.cl"},
	{0xB03D46EAF66DAF8AULL, "hlac_w_sundermount_-131078_0z.dds"},
	{0xB03D49C7F7D4DFC7ULL, "hlac_dr_0_deeproads_5838_2y.dds"},
	{0xB03E0129A8EC7CAAULL, "930875648.ogg"},
	{0xB03E140A63B01A2DULL, "gen00fl_merrill6128767.cl"},
	{0xB03E17C2CC953719ULL, "vxlt_undercitydungeon_18m_0_0.vlm"},
	{0xB03EB2A65191D7E4ULL, "vxdr_primevaltomb_90i_0_0.vlm"},
	{0xB03F7A4C304A0412ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar5_mcr370_playerflees.gad"},
	{0xB03F8424941B1711ULL, "vxlt_darktown_69v_0_0.vlm"},
	{0xB03FA428A0F2DBC0ULL, "hla_ht_0_hightown_a2_-131073_0x.dds"},
	{0xB03FCEAA5DCCABF8ULL, "hlac_w_encounter2_-2_0y.dds"},
	{0xB03FDF700D774F6FULL, "lt_0_lowtown_a2_n_33_0.anb"},
	{0xB03FF3386BAE75EEULL, "vxca_cavern1_haunted_j5w_0_0.vlm"},
	{0xB03FFA2FC33A9376ULL, "27754892.ogg"},
	{0xB04044AD20B3539CULL, "vxht_0_hightown_a2_b7j_0_0.vlm"},
	{0xB040A611CFB35510ULL, "654613861.ogg"},
	{0xB041097FA3CF05FAULL, "vxlt_0_lowtown_h3r_0_0.vlm"},
	{0xB0414E261A334E08ULL, "6034911_m.fxe"},
	{0xB0416D9A2E68C6F3ULL, "742005218.ogg"},
	{0xB041929AFC16DF40ULL, "hlac_ht_0_hightown_a2_65537_0y.dds"},
	{0xB0419B86C1B25AE5ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_wf_injured.ani"},
	{0xB041C4DE84B01362ULL, "hla_ht_0_hightown_a3_1_0z.dds"},
	{0xB041C537A32F76FFULL, "67768436.ogg"},
	{0xB04251749BAD2B76ULL, "hlac_lt_0_lowtown_a2_n_c_-4_0y.dds"},
	{0xB0429927BD4B190AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_staff_swarm.ani"},
	{0xB042F33CBBFD0160ULL, "lt_0_lowtown_n_42_0_lc.mmh"},
	{0xB0436BE6DACD4409ULL, "hlac_lt_0_lowtown_docks_n_-131073_0x.dds"},
	{0xB0438F26D38061DDULL, "601634674.ogg"},
	{0xB043B99427A56238ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_gam_dice_throw.gad"},
	{0xB043C964EC54CE76ULL, "581411398.ogg"},
	{0xB04477289A5B147AULL, "301029232.ogg"},
	{0xB044F98ACAE74568ULL, "6164881_m.fxe"},
	{0xB045113DB34859D1ULL, "838639507.ogg"},
	{0xB0453C8CCF9344E6ULL, "491269731.ogg"},
	{0xB0453F0953CCC5BCULL, "vxht_mansion_n_6jn_0_0.vlm"},
	{0xB04540A5CA2B8F1BULL, "vxlt_warehouse_18e_0_0.vlm"},
	{0xB04547CF62D1D504ULL, "w_sundermount_43_0.mmh"},
	{0xB0454BCF62D1DC57ULL, "w_sundermount_43_0.mao"},
	{0xB0455DCF62D1FAA6ULL, "w_sundermount_43_0.msh"},
	{0xB045AEBBCD6A809FULL, "hlac_lt_0_lowtown_a2_n_131073_0x.dds"},
	{0xB0468401AC1C197FULL, "llac_lt_0_lowtown_-1_4.dds"},
	{0xB046ACD3EE52D84DULL, "vxlt_gallowscourtyard_5gq_0_0.vlm"},
	{0xB046D80E76FB5CA6ULL, "454179392.ogg"},
	{0xB0474ECE5C084783ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bookmid.msh"},
	{0xB0475CCE5C085FDDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bookmid.mmh"},
	{0xB04760CE5C0866EEULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bookmid.mao"},
	{0xB04895BC8D04CD15ULL, "6174221_m.fxe"},
	{0xB048B2AA514478ABULL, "follower_banter6178585.cl"},
	{0xB048CB9FF9527BB0ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_basic.mao"},
	{0xB048CF9FF95282C3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_basic.mmh"},
	{0xB048DD9FF9529A1DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_basic.msh"},
	{0xB0497FEBA61771BFULL, "hlac_ca_cavern1_haunted_24421_0x.dds"},
	{0xB04999EE86A7C270ULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpa.msh"},
	{0xB049A7EE86A7D98AULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpa.mmh"},
	{0xB049ABEE86A7E0C1ULL, "art\\characters\\playercharacter\\humanmale\\bdy_tmpa.mao"},
	{0xB04A6A599DD8B289ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_bark_d.dds"},
	{0xB04AF450AECFB363ULL, "lowtown_male_lowlife2_mob6166758.cl"},
	{0xB04B21F4F69238C7ULL, "919515092.ogg"},
	{0xB04B4F7068051F53ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_imp_p.impa.ani"},
	{0xB04B7F7B88EA4645ULL, "vxca_cave_1o8_0_0.vlm"},
	{0xB04B88131EDE9DA8ULL, "567270892.ogg"},
	{0xB04BED86A6B03371ULL, "zz_approval_debug6003727.cl"},
	{0xB04BF615DA365A1EULL, "vxca_caverns1_geo_0_0.vlm"},
	{0xB04C5FD5F0281C30ULL, "543903699.ogg"},
	{0xB04C73CDA3ABF847ULL, "vxca_cavern1_haunted_int_0_0.vlm"},
	{0xB04CC1633F723127ULL, "lgt210_qun_blk_qunari6142323.cl"},
	{0xB04D03E9AC91C47CULL, "vxdr_slavers_145_0_0.vlm"},
	{0xB04D217E2F50DF6BULL, "842842836.ogg"},
	{0xB04D44DC54222A60ULL, "617710419.ogg"},
	{0xB04D8D5BDA65D584ULL, "168084003.ogg"},
	{0xB04DF38F8269A1E2ULL, "ser120_ghyslain6118405.cl"},
	{0xB04E346C140CBBA1ULL, "6094995_f.fxe"},
	{0xB04E665283DDD683ULL, "715912972.ogg"},
	{0xB04EB68A12A96C1CULL, "vxht_0_hightown_a2_b8k_0_0.vlm"},
	{0xB04F24B1908F875BULL, "6169816_m.fxe"},
	{0xB04F492303E6AF90ULL, "seb221_brett_harimann6177243.cl"},
	{0xB04F9659D4FEB6E2ULL, "6109756_m.fxe"},
	{0xB04FEEEA91F70FD4ULL, "6178435_m.fxe"},
	{0xB050E381704FFECEULL, "6127294-romance_after_intro.cut"},
	{0xB0511C13FA78FF64ULL, "lt_backalley_n_3_0_lc.mmh"},
	{0xB0513A663AE40730ULL, "6165925_m.fxe"},
	{0xB051513517193383ULL, "gen_im_trs_tmp_one_lsw_r2.uti"},
	{0xB051944302376B08ULL, "mage_arcanebolt.dds"},
	{0xB0519CEC0182598DULL, "746591724.ogg"},
	{0xB051F40B24028DBDULL, "vxht_brothel_n_6f0_0_0.vlm"},
	{0xB053B0E645C2DE3CULL, "vxht_mansion_n_6mo_0_0.vlm"},
	{0xB0544C50ED2B741AULL, "mag200_feynriel6056917.cl"},
	{0xB054654C305C3CB8ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar5_mcr370_playerflees.ani"},
	{0xB0549B7113C6FB3BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_spare.evt"},
	{0xB054F1D9C6572FD6ULL, "mag121_idunna6084021.cl"},
	{0xB05562F13162EFD5ULL, "art\\characters\\playercharacter\\humanmale\\boo_magc.msh"},
	{0xB05564F13162F34BULL, "art\\characters\\playercharacter\\humanmale\\boo_magc.mmh"},
	{0xB05570F1316307E8ULL, "art\\characters\\playercharacter\\humanmale\\boo_magc.mao"},
	{0xB055FFB76D58C0DEULL, "hlac_lt_0_lowtown_a2_n_c_65538_1x.dds"},
	{0xB056144691FC778EULL, "lgt110ip_at_docks_nt.utp"},
	{0xB056634713699B16ULL, "rdr110st_guard_warehouse.stg"},
	{0xB0568294B1262913ULL, "hlac_dr_primevaltomb_658_9y.dds"},
	{0xB056B2CB419575A1ULL, "lgt310ip_drop_zone_3.utp"},
	{0xB05782E4EC85F9BAULL, "vxht_keep_n_5jk_0_0.vlm"},
	{0xB057A55AF38DBB84ULL, "6019103_m.fxe"},
	{0xB0580A507C7D978CULL, "6167013_m.fxe"},
	{0xB0584770680F9B8AULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_imp_p.impa.evt"},
	{0xB058521FB7DBDBBBULL, "vxht_brothel_n_54c_0_0.vlm"},
	{0xB0586EF512BA28B9ULL, "vxlt_0_lowtown_a3_n_c_i09_0_0.vlm"},
	{0xB0589046DDCEDF2DULL, "vxht_0_hightown_a2s_0_0.vlm"},
	{0xB058A586AD045F4AULL, "vxht_brothel_n_6ej_0_0.vlm"},
	{0xB0594AE4A6096002ULL, "6108830_m.fxe"},
	{0xB059D09FF960F7E8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_basic.phy"},
	{0xB059D3B020EEF9DAULL, "sib100im_giftbet_portrait.uti"},
	{0xB059FF4BB832F56BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_vael.phy"},
	{0xB05A02748C9119D9ULL, "llac_lt_0_lowtown_docks_a3_n_196607_0.dds"},
	{0xB05A70E5EC53828DULL, "dr_primevaltomb_661_ll.mmh"},
	{0xB05A93A9F6F7D4F2ULL, "931719453.ogg"},
	{0xB05A947113CB5B7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_spare.gad"},
	{0xB05AEC0F723DDA1CULL, "6085831_f.fxe"},
	{0xB05BAE6CBCD53CD7ULL, "vxlt_gallowstemplar_f_5t7_0_0.vlm"},
	{0xB05C2520F96BBAC7ULL, "6146868_m.fxe"},
	{0xB05C5C83710DC428ULL, "character_character_p270.psh.fxo.dx11.assembly"},
	{0xB05C84EE5266937AULL, "dae000_keep_bureaucrat_2_st6165387.cl"},
	{0xB05D2BDEF1EB9F6EULL, "art\\characters\\playercharacter\\humanfemale\\glv_maso.phy"},
	{0xB05DDAB38455302EULL, "qun110_petrice6042139.cl"},
	{0xB05DF740E49091F9ULL, "gen00fl_isabela6110221.cl"},
	{0xB05E2C823401D730ULL, "6129603_f.fxe"},
	{0xB05E411C5FED4B4FULL, "42334337.ogg"},
	{0xB05ECB069A42BF93ULL, "vxw_bonepitmining_a2_29z_0_0.vlm"},
	{0xB05EDCC59395BFA6ULL, "gen00fl_anders6111298.cl"},
	{0xB05F2670681582FDULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_imp_p.impa.gad"},
	{0xB05F2CC746BB303AULL, "vxlt_0_lowtown_docks_c0a_0_0.vlm"},
	{0xB05F65464A1A867BULL, "mag320_nobleman_1.bnk"},
	{0xB05F7C6EE3668F9CULL, "219094984.ogg"},
	{0xB05FBD11E8D8D75AULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkrf.gad"},
	{0xB05FDAFD7AEFC29AULL, "vxlt_undercitydungeon_1nt_0_0.vlm"},
	{0xB060065847042EBFULL, "qcr000_arishok6087275.cl"},
	{0xB060131250A1418FULL, "hightown_male_noble_gossip6174053.cl"},
	{0xB06072C99E9D5F08ULL, "vxw_bonepit_9k_0_0.vlm"},
	{0xB0607B72159767D0ULL, "959421271.ogg"},
	{0xB0607EBD8AA95E65ULL, "gen00fl_varric6104985.cl"},
	{0xB060E562EB6E2643ULL, "vxw_bonepitdestroyed_a3_1px_0_0.vlm"},
	{0xB0610C8F7542AF6AULL, "6094551_m.fxe"},
	{0xB06152CE6A384669ULL, "prty_anders_armor_1.dds"},
	{0xB061937113D1787AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_spare.ani"},
	{0xB061E38043F704B2ULL, "fen110_party_comment6095219.cl"},
	{0xB0620BBBCFA48434ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm03.phy"},
	{0xB062245FAE17C01BULL, "vxlt_warehouse_12g_0_0.vlm"},
	{0xB062435C5B68D613ULL, "6171299_m.fxe"},
	{0xB0624FCC5562AAC2ULL, "vxca_cavern1_haunted_gjw_0_0.vlm"},
	{0xB06256DA85C45B38ULL, "zz_followers_banter6157354.cl"},
	{0xB0627834665D78F2ULL, "hlac_ht_keep_n_3732_0y.dds"},
	{0xB0634F04DE06E229ULL, "hlac_lt_hangedman_n_-1_17y.dds"},
	{0xB06367464A1E723BULL, "mag320_nobleman_1.cnv"},
	{0xB063E6A3BA5E3B4FULL, "vxdr_0_deeproads_boz_0_0.vlm"},
	{0xB064A18741636698ULL, "6153238_m.fxe"},
	{0xB064CADC4DE53FBFULL, "vxht_0_hightown_bes_0_0.vlm"},
	{0xB064EB2715D98EBFULL, "follower_banter6171942.cl"},
	{0xB065EEE39582132EULL, "vxlt_0_lowtown_a2_n_bxv_0_0.vlm"},
	{0xB06605CCF303A7B7ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fish.msh"},
	{0xB06617CCF303C61AULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fish.mao"},
	{0xB0661BCCF303CD29ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fish.mmh"},
	{0xB0662F69CA435663ULL, "w_woundedcoast_33_0.msh"},
	{0xB0663514192137DBULL, "6161351_m.fxe"},
	{0xB0663D69CA436DBDULL, "w_woundedcoast_33_0.mmh"},
	{0xB0664169CA4374CEULL, "w_woundedcoast_33_0.mao"},
	{0xB066C011E8DEFB71ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkrf.evt"},
	{0xB066CF68594BACCAULL, "hlac_w_bonepit_0_5y.dds"},
	{0xB066E9F82592DD61ULL, "vxca_caverns2_7mk_0_0.vlm"},
	{0xB066FADCB7ED8E1FULL, "hlac_w_encounter2_131076_0y.dds"},
	{0xB067447F9CDBAC11ULL, "6161461_m.fxe"},
	{0xB0679F0360DB0A68ULL, "gen_im_trs_dwv_med_glv_r2.uti"},
	{0xB067A4D4D3ADE69CULL, "vxht_0_hightown_n_au8_0_0.vlm"},
	{0xB067FEECE84C5449ULL, "118036055.ogg"},
	{0xB068284F90C7E8DEULL, "6172862_m.fxe"},
	{0xB0686C6F1289EF25ULL, "11048636.ogg"},
	{0xB068A7330193F49AULL, "6152200_m.fxe"},
	{0xB068BD922C636F2DULL, "459158575.ogg"},
	{0xB068E583BEECF5C3ULL, "176772207.ogg"},
	{0xB068FF1EF2BFD101ULL, "gen00fl_isabela6096783.cl"},
	{0xB06903DC2F1EA7CAULL, "6061418_m.fxe"},
	{0xB069572E34A5147CULL, "6207551_m.fxe"},
	{0xB069727F853C3271ULL, "vxw_encounter2_camp_k_0_0.vlm"},
	{0xB069CF5857B6E996ULL, "zz_dae_debug6029851.cl"},
	{0xB06A189B6DF0D807ULL, "vxlt_hangedman_n_2er_0_0.vlm"},
	{0xB06A56A1627A36C4ULL, "montage_subs_demohack6217138.cl"},
	{0xB06B19F037BE0B1BULL, "923447796.ogg"},
	{0xB06B5CB25B2E87B5ULL, "vxlt_gallowstemplar_397_0_0.vlm"},
	{0xB06BA459E70A072FULL, "878369499.ogg"},
	{0xB06BBC700D9C4E13ULL, "lt_0_lowtown_a2_n_33_0.rml"},
	{0xB06BE8BBCFAC9D31ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm03.msh"},
	{0xB06C02BBCFACC9EFULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm03.mmh"},
	{0xB06C0A5F807012B8ULL, "qun120_arishok6051689.cl"},
	{0xB06C13E73E9F86A8ULL, "follower_banter6136375.cl"},
	{0xB06C4C1BB3284F77ULL, "hlac_lt_gallowscourtyard_65535_0y.dds"},
	{0xB06C5F9B01CD172BULL, "298961444.ogg"},
	{0xB06C8CA232516025ULL, "vxlt_0_lowtown_docks_a3_ctm_0_0.vlm"},
	{0xB06C91A3A7E3B88AULL, "follower_banter6161319.cl"},
	{0xB06D0AE89CD97C3EULL, "ts_coterie_03.gda"},
	{0xB06D4ACF62F32CFBULL, "w_sundermount_43_0.anb"},
	{0xB06D5751888540BFULL, "one110_meeran6089485.cl"},
	{0xB06D975F40638EF1ULL, "vxw_bonepit_29f_0_0.vlm"},
	{0xB06DA1EBF043DCA7ULL, "7499543.ogg"},
	{0xB06DBE52234B9736ULL, "vxca_caverns1_gi3_0_0.vlm"},
	{0xB06E0EA41EB8D854ULL, "dae300_bran6143490.cl"},
	{0xB06E21B68C83A442ULL, "vxlt_darktown_6ii_0_0.vlm"},
	{0xB06EA626B5FB9C91ULL, "673122554.ogg"},
	{0xB06EEEACC0B4758DULL, "579540663.ogg"},
	{0xB06FA3E9DFA564DFULL, "plt_cod_crt_rock_wraith.plo"},
	{0xB06FEF2050A8ADDEULL, "vxht_0_hightown_c4p_0_0.vlm"},
	{0xB07072ADE6E78948ULL, "vxca_caverns2_7mn_0_0.vlm"},
	{0xB070ABB06DD15543ULL, "mcr370_cut_player_survives.bnk"},
	{0xB0710F2264CFFABAULL, "6064668_m.fxe"},
	{0xB0720343234177BCULL, "vxlt_undercitydungeon_157_0_0.vlm"},
	{0xB072100D1790C9A3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrgc_1.mao"},
	{0xB0721FED9C4DAAFAULL, "vxca_cavern1_haunted_edg_0_0.vlm"},
	{0xB0724744563E2DC0ULL, "dae000_serendipity6161434.cl"},
	{0xB072680E29E6EAE8ULL, "6173377_m.fxe"},
	{0xB072A32525D1060BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_daggera_n.dds"},
	{0xB072ED4DCEB727B9ULL, "gen_im_wep_rog_dua_act1_03.uti"},
	{0xB07308EC4A4314FAULL, "6146817_f.fxe"},
	{0xB07311347215A979ULL, "vxlt_darktown_5kg_0_0.vlm"},
	{0xB0733E6DB60362C0ULL, "959478145.ogg"},
	{0xB073460EEC3724BFULL, "vxca_caverns2_8qv_0_0.vlm"},
	{0xB073CB57B71B0615ULL, "vxca_caverns1_mines_huf_0_0.vlm"},
	{0xB073D2EAC69682ABULL, "gen00fl_merrill6129982.cl"},
	{0xB073D811E8E9AE40ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkrf.ani"},
	{0xB0740118A588782EULL, "479291608.ogg"},
	{0xB0745CDD4BDB882BULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.impa.ani"},
	{0xB07477949D4C6425ULL, "hlac_dr_dungeon_3215_3y.dds"},
	{0xB0748DB06DD50B73ULL, "mcr370_cut_player_survives.cnv"},
	{0xB074A6F10B17505BULL, "6166251_m.fxe"},
	{0xB074EDC09162D02AULL, "6127232_f.fxe"},
	{0xB075120562BC225FULL, "follower_banter6163156.cl"},
	{0xB07559A62026A5E4ULL, "781639552.ogg"},
	{0xB075894090F0904AULL, "vxht_brothel_n_6kd_0_0.vlm"},
	{0xB075B0982FB15704ULL, "535939558.ogg"},
	{0xB075DC6182C1A409ULL, "6208336_f.fxe"},
	{0xB075F28B83301173ULL, "plt_rumor_act_of_mercy.nss"},
	{0xB075F6B44479B298ULL, "480706425.ogg"},
	{0xB0760510D3E4C821ULL, "vxlt_darktown_2wx_0_0.vlm"},
	{0xB07606AFD2D4EDB7ULL, "6103221_m.fxe"},
	{0xB076622539FFD325ULL, "dae000ip_crowd_lm3_drunk2.utp"},
	{0xB077F0B1D78D64ADULL, "6075100_m.fxe"},
	{0xB0783DF0F2C5CBC2ULL, "747008253.ogg"},
	{0xB0786198916AC81BULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\enter_walk_elf_female.evt"},
	{0xB078666E017C90D0ULL, "6101062_m.fxe"},
	{0xB0789354838E5979ULL, "839953709.ogg"},
	{0xB078A80BECC048B1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_tmpa_n.dds"},
	{0xB078D91D7850EC2EULL, "qabasic.arl"},
	{0xB07911F9935812D9ULL, "hlac_lt_0_lowtown_a3_n_c_65533_1y.dds"},
	{0xB07932F03CE33E01ULL, "779380010.ogg"},
	{0xB07979B06642BF20ULL, "vxca_cave_2b9_0_0.vlm"},
	{0xB079E06A25754C7BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vent_d.dds"},
	{0xB079F91CB4E3E696ULL, "6173858_m.fxe"},
	{0xB07A9290C968A085ULL, "1042590589.ogg"},
	{0xB07B62ABCA67ECE2ULL, "sst_ss_castillon.gda"},
	{0xB07DB40415AF99D8ULL, "gen00fl_anders6112918.cl"},
	{0xB07E720428355A0CULL, "w_sundermount_112_0.rml"},
	{0xB07F04F4FDB0EE49ULL, "vxlt_gallowscourtyard_n_5a4_0_0.vlm"},
	{0xB07F5A989170DB5CULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\enter_walk_elf_female.gad"},
	{0xB07F9EFE5A5FCEE0ULL, "war_shockwave.dds"},
	{0xB08052F94C889DD9ULL, "vxht_0_hightown_b0n_0_0.vlm"},
	{0xB08075201537903CULL, "vxlt_undercitydungeon_184_0_0.vlm"},
	{0xB080E88D2C17AF25ULL, "hlac_lt_0_lowtown_n_-65535_0z.dds"},
	{0xB080E96C04ADD852ULL, "vxlt_undercityencounter_1j9_0_0.vlm"},
	{0xB0812F5F5150B1CEULL, "w_encounter3_25_0_lc.mmh"},
	{0xB0814DE47AD87FB8ULL, "55286178.ogg"},
	{0xB08170FD4FC0E191ULL, "6156104_m.fxe"},
	{0xB0819BA33FFD82C5ULL, "6091713_m.fxe"},
	{0xB081AC0564B9EEF9ULL, "224432008.ogg"},
	{0xB0821ABF70B996C6ULL, "art\\vfx\\combat\\vfx_r_var_grenade_toss_c.impa.evt"},
	{0xB08244DD4BE79CA2ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.impa.evt"},
	{0xB082BF859CCE9604ULL, "hlac_lt_darktown_-1_31x.dds"},
	{0xB083AD89C81C9967ULL, "592927977.ogg"},
	{0xB084151A351D7425ULL, "6126481_f.fxe"},
	{0xB084D6D8663740C7ULL, "lt_0_lowtown_docks_6_0.rml"},
	{0xB0853FCDE6E6A7FAULL, "gen00fl_anders6125521.cl"},
	{0xB08559989175455AULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\enter_walk_elf_female.ani"},
	{0xB0857DEEB6B24D6BULL, "405048424.ogg"},
	{0xB0858382C79B3F50ULL, "559910571.ogg"},
	{0xB085DE9FAAAC230CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgta_0s.dds"},
	{0xB085F0A0ECA6DDC9ULL, "397439190.ogg"},
	{0xB086DCEC58FC5D79ULL, "hf_zmother_h2s2.mor"},
	{0xB086FEDFB7D4B296ULL, "ht_mansion2b.arl"},
	{0xB08717B8E35672DDULL, "vxlt_darktown_4f9_0_0.vlm"},
	{0xB0875D5BCFA8DDD5ULL, "233034374.ogg"},
	{0xB087800B65B2B0E2ULL, "mcr370_negotiation6078740.cl"},
	{0xB0879D7172E21919ULL, "269890809.ogg"},
	{0xB087FF96D1688C83ULL, "6095737_m.fxe"},
	{0xB0882CBF1D612370ULL, "vis121_saemus6101333.cl"},
	{0xB0884B365961673AULL, "exporttestlevel.arl"},
	{0xB088E2F8C9142508ULL, "dae000cr_grand_cleric6066113.cl"},
	{0xB08931BF70BFDC11ULL, "art\\vfx\\combat\\vfx_r_var_grenade_toss_c.impa.gad"},
	{0xB0893707722FA8B2ULL, "6156349_m.fxe"},
	{0xB08943DD4BEDB9B5ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.impa.gad"},
	{0xB08970FBAB3F0EE9ULL, "33697263.ogg"},
	{0xB08A2CF13D23F146ULL, "vxlt_0_lowtown_docks_a3_cze_0_1.vlm"},
	{0xB08A3492C3F13E4DULL, "static_static_p33.psh.fxo.dx11"},
	{0xB08A84B61F4D09A5ULL, "6084860_f.fxe"},
	{0xB08A9541D26555C2ULL, "art\\vfx\\models\\vfx_rockarmor_rt.msh"},
	{0xB08A9741D2655978ULL, "art\\vfx\\models\\vfx_rockarmor_rt.mmh"},
	{0xB08B29994D27D8F3ULL, "one220_yevhen6135756.cl"},
	{0xB08C0DDD2F3C87ADULL, "162531656.ogg"},
	{0xB08C722843BAB16AULL, "mag160_feynriel6020412.cl"},
	{0xB08C7DBFC0985F1EULL, "vxlt_0_lowtown_docks_a3_d6k_0_0.vlm"},
	{0xB08C92312765351BULL, "6165100_m.fxe"},
	{0xB08CA38036E2C8E8ULL, "art\\characters\\playercharacter\\qunarimale\\bdy_ars.mao"},
	{0xB08CC0364925E754ULL, "mrl000_marethari6129269.cl"},
	{0xB08CDF43EB972DE1ULL, "6042079_f.fxe"},
	{0xB08D189AA851FDD0ULL, "6155610_m.fxe"},
	{0xB08D7600D3CBBDA6ULL, "6104409_m.fxe"},
	{0xB08DF330AE162B1BULL, "hlac_w_encounter3_0_8z.dds"},
	{0xB08E07867FA7D1D2ULL, "6172978_m.fxe"},
	{0xB08E2BD791E8187AULL, "vxht_chantry_n_2cw_0_0.vlm"},
	{0xB08E478D2ECB345EULL, "hlac_lt_backalley_n_1_1y.dds"},
	{0xB08E5FDD07C57090ULL, "vxw_encounter2_camp_p5_0_1.vlm"},
	{0xB08E83079C0D45C9ULL, "fen110_fenris6117944.cl"},
	{0xB08EABA95124E078ULL, "lt_gallowstemplar_2392.rml"},
	{0xB08EF47F333EF2C1ULL, "pro000_athenril6048758.cl"},
	{0xB08F5A277933EA25ULL, "6119699_m.fxe"},
	{0xB08FAD1C55B8026AULL, "561526111.ogg"},
	{0xB0900D8031F39B45ULL, "6096613_m.fxe"},
	{0xB09017837F015D67ULL, "vxlt_0_lowtown_a2_n_fhg_0_0.vlm"},
	{0xB09032BF70C5FCBFULL, "art\\vfx\\combat\\vfx_r_var_grenade_toss_c.impa.ani"},
	{0xB0906E542AF38D69ULL, "m_dae311_gamlens_house_nt.bnk"},
	{0xB09080868BCB7399ULL, "zz_dae_debug6029432.cl"},
	{0xB090A2EF20DA18D4ULL, "15958015.ogg"},
	{0xB090F79F97E8AF99ULL, "hlac_ht_mansion_n_3373_3z.dds"},
	{0xB091A8E0A69F8234ULL, "505731980.ogg"},
	{0xB09257AFF3389309ULL, "6151909_m.fxe"},
	{0xB09263AD9B36A234ULL, "gen_im_cns_arn_celestial.uti"},
	{0xB092DA2E778532F4ULL, "hlac_ht_mansion2b_818_2x.dds"},
	{0xB092FB560AFAF14BULL, "llac_lt_0_lowtown_docks_n_-3_0.dds"},
	{0xB093FBD881FBEFA7ULL, "llac_w_encounter3_-196612_0.dds"},
	{0xB094337D0F83F86BULL, "vxlt_warehouse_141_0_0.vlm"},
	{0xB0945672B33AB9EDULL, "vxw_woundedcoast_nf_0_0.vlm"},
	{0xB094DDA5BF416E1DULL, "107935787.ogg"},
	{0xB09636549167C1AAULL, "6174367_m.fxe"},
	{0xB096593EA3182547ULL, "vxlt_0_lowtown_a2_by3_0_0.vlm"},
	{0xB0967D6DC84AB609ULL, "6054181_m.fxe"},
	{0xB0976809A460D84AULL, "static3.psh.fxo"},
	{0xB09772D3E75D51E7ULL, "6207448_m.fxe"},
	{0xB0978AC4B86D59CFULL, "nat380_nathaniel6071280.cl"},
	{0xB0979757D72145C7ULL, "vxlt_gallowscourtyard_7ca_0_0.vlm"},
	{0xB09894EA3E8F0579ULL, "775548944.ogg"},
	{0xB098AF17858581FBULL, "vxht_0_hightown_a6y_0_0.vlm"},
	{0xB098C17A8056B6CFULL, "229453288.ogg"},
	{0xB09967CF63B7F505ULL, "6150498_m.fxe"},
	{0xB099A898854C6FFEULL, "vxlt_gallowstemplar_f_3jz_0_0.vlm"},
	{0xB099AC379EE46017ULL, "gen00fl_isabela_init6120904.cl"},
	{0xB099AC8275C008A5ULL, "mag110tr_vincentoarrest.utt"},
	{0xB099DBAB98BF8912ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.mmh"},
	{0xB099E342F81A88CDULL, "gen00fl_carver6108925.cl"},
	{0xB099F8D76736B2CDULL, "878548797.ogg"},
	{0xB09B50F7F8D24DAAULL, "6035851_m.fxe"},
	{0xB09B523CE4C7C639ULL, "hlac_dr_primevaltomb_658_8x.dds"},
	{0xB09BA05FB0AFB761ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.impa.ani"},
	{0xB09BCABC387AF058ULL, "vxlt_0_lowtown_a2_eot_0_1.vlm"},
	{0xB09BE2BC857BDC2EULL, "236226777.ogg"},
	{0xB09C29303A21DEDEULL, "mag101_aombanter6052754.cl"},
	{0xB09C682FF4F69E6DULL, "vxht_0_hightown_a2_c76_0_0.vlm"},
	{0xB09C6896E50080B8ULL, "780959873.ogg"},
	{0xB09C7B4D6BC07D53ULL, "707700486.ogg"},
	{0xB09CBA4576021D75ULL, "gen00fl_varric6107294.cl"},
	{0xB09CC04083E75203ULL, "1017565134.ogg"},
	{0xB09D1791A0C23A45ULL, "vxca_cavern1_haunted_edl_0_0.vlm"},
	{0xB09D92A866F0CFCEULL, "161041955.ogg"},
	{0xB09E28A1372FDCDAULL, "80939995.ogg"},
	{0xB09E2ADEF2233431ULL, "art\\characters\\playercharacter\\humanfemale\\glv_maso.mmh"},
	{0xB09E2E34494E20EFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portrait_sa.dds"},
	{0xB09E44DEF22360EFULL, "art\\characters\\playercharacter\\humanfemale\\glv_maso.msh"},
	{0xB09E89ED59FF9DE1ULL, "qun221_petrice6098716.cl"},
	{0xB09ED4E755AA8D04ULL, "vxht_0_hightown_a3_56w_0_0.vlm"},
	{0xB09F4A7EA7B17E8EULL, "dae300_cullen6140633.cl"},
	{0xB09F9502499DAB23ULL, "dae000ip_crowd_lm3_talk13.utp"},
	{0xB09FC911771F8173ULL, "mer120_hubert6077230.cl"},
	{0xB0A026B0811CFCC3ULL, "632287261.ogg"},
	{0xB0A08B2994DB710CULL, "979692123.ogg"},
	{0xB0A15BD14850B3D6ULL, "351636875.ogg"},
	{0xB0A1E7F17FDFAAB2ULL, "6087339_m.fxe"},
	{0xB0A230D5CA68A3E1ULL, "vxca_cavern1_haunted_ei7_0_0.vlm"},
	{0xB0A24B50249185B2ULL, "6140298_m.fxe"},
	{0xB0A293CBA0756ED4ULL, "gen00fl_isabela6090885.cl"},
	{0xB0A2AB5FB0B5E803ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.impa.gad"},
	{0xB0A2C9A748FAC854ULL, "matdef_dcd_diffuse.dds"},
	{0xB0A35AB28B461343ULL, "vxca_caverns2prc_90a_0_0.vlm"},
	{0xB0A3A36A3D1A20B4ULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_act_c.mmh"},
	{0xB0A3B70AE4F433CFULL, "6051632_m.fxe"},
	{0xB0A3B9990929232FULL, "cd_fen110_cut_fenris_intro.wbd"},
	{0xB0A3C0DF4497CE7BULL, "6129197_m.fxe"},
	{0xB0A3F0C670093E25ULL, "guiquadmatobject.mao"},
	{0xB0A3FB72C742A3DDULL, "vxlt_darktown_4h7_0_0.vlm"},
	{0xB0A40E0DF9314464ULL, "lt_undercitydungeon_3567.anb"},
	{0xB0A43AD01C52379CULL, "vxlt_0_lowtown_a2_gof_0_1.vlm"},
	{0xB0A47AFB8901A126ULL, "gen00fl_anders6112905.cl"},
	{0xB0A4C607B06B8E7FULL, "493599183.ogg"},
	{0xB0A4D13A4F3479C7ULL, "lgt110cr_coa_captain_2.utc"},
	{0xB0A4DAB6F09AD057ULL, "6118140_m.fxe"},
	{0xB0A628D8441375BBULL, "344712420.ogg"},
	{0xB0A6AEA6F259A216ULL, "mer240_jansen6050450.cl"},
	{0xB0A6CD56878911A3ULL, "dre000_sibling_has_blight6037038.cl"},
	{0xB0A6E8D8C4978C55ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_108a_sa.dds"},
	{0xB0A763A03AFA49B2ULL, "hlac_lt_0_lowtown_a2_65538_0x.dds"},
	{0xB0A77F44D6682A60ULL, "zz_lgtpt_2_debug.ncs"},
	{0xB0A78B7D0082ED3AULL, "art\\characters\\playercharacter\\humanfemale\\cth_homa.mmh"},
	{0xB0A78F7D0082F471ULL, "art\\characters\\playercharacter\\humanfemale\\cth_homa.mao"},
	{0xB0A79D7D00830B80ULL, "art\\characters\\playercharacter\\humanfemale\\cth_homa.msh"},
	{0xB0A9282055C23939ULL, "vxlt_0_lowtown_docks_a3_apg_0_0.vlm"},
	{0xB0A93B591FB893AEULL, "sse_ss_dalish_f_1.gda"},
	{0xB0A93C524AA5E4E3ULL, "gen00fl_fenris6108020.cl"},
	{0xB0A9522595EE7A93ULL, "6170560_m.fxe"},
	{0xB0A9885FB0BBCBE8ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.impa.evt"},
	{0xB0AA277390CE9BA9ULL, "hlac_lt_darktown_-1_6y.dds"},
	{0xB0AABD7AD8048376ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_3.ani"},
	{0xB0AACC2E8CB48F22ULL, "19420149.ogg"},
	{0xB0AADA0CE6ECB6EDULL, "vxht_0_hightown_a2_c25_0_0.vlm"},
	{0xB0AB689D8F16A5FBULL, "191979528.ogg"},
	{0xB0ABA22C353C5533ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_exit.evt"},
	{0xB0ABB98B8BBD4FC8ULL, "vxca_cavern1_haunted_ei2_0_0.vlm"},
	{0xB0ABBBCB41DDC568ULL, "lgt310ip_drop_zone_3.ncs"},
	{0xB0AC20AAF907AB32ULL, "214052808.ogg"},
	{0xB0AC2106A82FE89EULL, "vxlt_darktown_2xf_0_0.vlm"},
	{0xB0AC29748FC6A6D7ULL, "619681657.ogg"},
	{0xB0AC48B426033CE5ULL, "65132103.ogg"},
	{0xB0AC5A6268CEFFF8ULL, "mrl000_marethari6129333.cl"},
	{0xB0ACE308908EAFA8ULL, "6148450_m.fxe"},
	{0xB0ACF802919F0270ULL, "vxht_mansion2_1bz_0_0.vlm"},
	{0xB0AD2BC3B73274F6ULL, "hlac_w_bonepitmining_a2_262149_0x.dds"},
	{0xB0AD5C02A2EC8955ULL, "vxlt_0_lowtown_a3_n_c_eom_0_1.vlm"},
	{0xB0AD782B4D772164ULL, "vxlt_hangedman_10r_0_0.vlm"},
	{0xB0ADC92E6F0D6A09ULL, "and230cr_ser_alrik.utc"},
	{0xB0AE249A58C4CED3ULL, "da2terrain_terrain_p1.vsh.fxo.dx11.assembly"},
	{0xB0AE25F7F3CDEBACULL, "dae000ip_crowd_mf2_talk2.utp"},
	{0xB0AED73E3574FACCULL, "6166893_f.fxe"},
	{0xB0AF0124BEDB9E0BULL, "vxlt_backalley_g_1ee_0_0.vlm"},
	{0xB0AF21E9BE4CE5DFULL, "6163469_m.fxe"},
	{0xB0AF8B6040E5D2A6ULL, "564800328.ogg"},
	{0xB0AFA03FFA07C8D9ULL, "vxht_brothel_n_4h4_0_0.vlm"},
	{0xB0AFB049A387A688ULL, "vxlt_0_lowtown_gkk_0_0.vlm"},
	{0xB0AFD311247D77F0ULL, "dae000_hate_follower_react6103237.cl"},
	{0xB0AFD3C5815F0646ULL, "vxlt_0_lowtown_n_cmt_0_0.vlm"},
	{0xB0AFF08CA98FF614ULL, "vxlt_0_lowtown_a2_n_eqt_0_0.vlm"},
	{0xB0B02D6DF9069EC2ULL, "6142043_m.fxe"},
	{0xB0B03573E276581CULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_4_ian.ani"},
	{0xB0B05C3E59B2096BULL, "hlac_lt_warehouse_n_0_6y.dds"},
	{0xB0B0712C4022A8DBULL, "vxht_keep_n_69t_0_0.vlm"},
	{0xB0B087BE497C2AABULL, "llac_lt_0_lowtown_65533_2.dds"},
	{0xB0B104CE6617176EULL, "366638084.ogg"},
	{0xB0B109A0978989BAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_viscdesk_n.dds"},
	{0xB0B116CA2E67FC39ULL, "6059775_m.fxe"},
	{0xB0B130F7087F01ACULL, "conversation_followers.gda"},
	{0xB0B1D84AC5285889ULL, "cs_mcr370cs_player_rules.bnk"},
	{0xB0B2001C23F86B7DULL, "558983306.ogg"},
	{0xB0B20E7C0C27642BULL, "6036953_f.fxe"},
	{0xB0B25078B0D6BA8BULL, "mag160_feynriel6107433.cl"},
	{0xB0B29B2C35426874ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_exit.gad"},
	{0xB0B2D64C61F6E296ULL, "hlac_lt_0_lowtown_a2_n_c_65537_0z.dds"},
	{0xB0B30D7795611E2CULL, "6051154_m.fxe"},
	{0xB0B3476761E91928ULL, "10441411.ogg"},
	{0xB0B3DDD7263C0C94ULL, "dae000_flemeth6000741.cl"},
	{0xB0B3F1BC2706AB7BULL, "hlac_ht_0_hightown_-65538_1y.dds"},
	{0xB0B41365CF501CACULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.dura.ani"},
	{0xB0B4700DB8F9ECEBULL, "hlac_ca_caverns2prc_5802_44y.dds"},
	{0xB0B47B11BB0D76C7ULL, "vxlt_gallowscourtyard_7ic_0_0.vlm"},
	{0xB0B49EE06D026F61ULL, "vfx_vfx_p6.psh.fxo.dx11.assembly"},
	{0xB0B4D467103590D6ULL, "616607323.ogg"},
	{0xB0B5AFF166A0A8A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_reaction.ani"},
	{0xB0B65494D9588B22ULL, "hlac_lt_0_lowtown_-65535_1y.dds"},
	{0xB0B6DC99657A669DULL, "vxht_0_hightown_ca8_0_0.vlm"},
	{0xB0B71A124DA05930ULL, "zz_seb_debug6183195.cl"},
	{0xB0B74273E27C8D5EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_4_ian.gad"},
	{0xB0B7C57AD80F1B67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_3.evt"},
	{0xB0B7DF2E43FA4575ULL, "6044966_m.fxe"},
	{0xB0B848EE560D538FULL, "vfx_rendertoshadowmap_p3.vsh.fxo.dx11.assembly"},
	{0xB0B8BE9E5855D801ULL, "625532716.ogg"},
	{0xB0B920304E5BE88EULL, "463257456.ogg"},
	{0xB0B94BE9D8E2CF6DULL, "vxht_0_hightown_a2_c94_0_0.vlm"},
	{0xB0B9AA2C3548A052ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_exit.ani"},
	{0xB0B9F5E14C58740EULL, "6101283_f.fxe"},
	{0xB0BB2065CF5651EEULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.dura.gad"},
	{0xB0BB4900A18E3F41ULL, "gen00fl_anders6125586.cl"},
	{0xB0BB7F9B27BE3A78ULL, "596346469.ogg"},
	{0xB0BC2B311FD7CF13ULL, "ts_dwarven_acc_props.gda"},
	{0xB0BC4351E4007A8CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wc_d.dds"},
	{0xB0BC7F2F2BCD673EULL, "vxlt_0_lowtown_a2_n_hgu_0_0.vlm"},
	{0xB0BCBAF166A6DA43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_reaction.gad"},
	{0xB0BD43E36235B00CULL, "vxca_caverns2_61k_0_0.vlm"},
	{0xB0BD8978F3B847CDULL, "vxw_encounter3_7k_0_0.vlm"},
	{0xB0BDF71CCD62CDE6ULL, "786004942.ogg"},
	{0xB0BE6CBE186E10CDULL, "ach_comp_craft.dds"},
	{0xB0BEA8CDB7B943E4ULL, "vxht_mansion_n_52q_0_0.vlm"},
	{0xB0BEE67AD81571B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_2_gesture_3.gad"},
	{0xB0BF11D04913FF41ULL, "74237419.ogg"},
	{0xB0BF148FAE54D4E1ULL, "vxca_cavern1_haunted_ec5_0_0.vlm"},
	{0xB0BFCF2ABB4D2083ULL, "dae200_bran6094790.cl"},
	{0xB0BFD511E18C4E0DULL, "60561469.ogg"},
	{0xB0C0268694872B55ULL, "llac_ht_0_hightown_n_-131070_0.dds"},
	{0xB0C04FCC2D41ECC4ULL, "vxca_caverns2prc_ad6_0_0.vlm"},
	{0xB0C0CCA4532DC770ULL, "hlac_lt_darktown_-1_21y.dds"},
	{0xB0C12EFBAFC91748ULL, "ts_raiders_prog_junk1.gda"},
	{0xB0C15842D3A436CCULL, "6125468_m.fxe"},
	{0xB0C21B65CF5C67CDULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.dura.evt"},
	{0xB0C2D6C6BA719852ULL, "683550816.ogg"},
	{0xB0C2DAAB98E28CB9ULL, "art\\vfx\\combat\\vfx_glb_stagger_crust_c.anb"},
	{0xB0C397F166ACBD28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_reaction.evt"},
	{0xB0C3A9E46CAB9A9EULL, "141885294.ogg"},
	{0xB0C44C9779EB9F7EULL, "1049850335.ogg"},
	{0xB0C49A427E2C2AE2ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.mmh"},
	{0xB0C5A7F08D6A5AF4ULL, "6204806_m.fxe"},
	{0xB0C621BB04AD0FA7ULL, "6097393_m.fxe"},
	{0xB0C66DE8986C42F6ULL, "vxlt_0_lowtown_a2_n_c_eji_0_0.vlm"},
	{0xB0C67323ECA6153CULL, "397555293.ogg"},
	{0xB0C68AF4739DED50ULL, "6070128_m.fxe"},
	{0xB0C74C7EF4783B7AULL, "vxlt_0_lowtown_a2_n_c_h1y_0_0.vlm"},
	{0xB0C82ED3EEB86A04ULL, "28917494.ogg"},
	{0xB0C83415DEE5BF11ULL, "dae000_madam_lusine6152276.cl"},
	{0xB0C84615932AAFB7ULL, "hairalpha_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0xB0C89D456FA980C8ULL, "vxca_cavern1_haunted_ec0_0_0.vlm"},
	{0xB0C95054BB1F420CULL, "185983046.ogg"},
	{0xB0C95FF306423C22ULL, "vxht_chantry_n_1yj_0_0.vlm"},
	{0xB0C99603BAD66F85ULL, "657610721.ogg"},
	{0xB0C9C041EC2BA0D9ULL, "6087651_m.fxe"},
	{0xB0C9F64CDB171B9DULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comc_0t.dds"},
	{0xB0CA6570327CF14FULL, "vxlt_blackemporium_zj_0_0.vlm"},
	{0xB0CAFEF08FE1AEFCULL, "genip_table.utp"},
	{0xB0CB6654C2016D8CULL, "6034195_m.fxe"},
	{0xB0CB906D92D70520ULL, "hlac_lt_darktown_-1_73z.dds"},
	{0xB0CBAF8ECC704E87ULL, "738762255.ogg"},
	{0xB0CBE4DEA2C7CF0BULL, "vxlt_backalley_g_1og_0_0.vlm"},
	{0xB0CBF7F6BF853C0EULL, "hlac_lt_0_lowtown_-1_5z.dds"},
	{0xB0CC7D994472D558ULL, "6147675_m.fxe"},
	{0xB0CC8F8C06301BE2ULL, "w_sundermount_ll.mmh"},
	{0xB0CCA66A3D3D2B6BULL, "art\\vfx\\combat\\vfx_m_pri_rockarmor_act_c.anb"},
	{0xB0CCAA7D00A23E45ULL, "art\\characters\\playercharacter\\humanfemale\\cth_homa.phy"},
	{0xB0CCC492062D5549ULL, "128900513.ogg"},
	{0xB0CCEA407277694EULL, "dae000_bodhan6146910.cl"},
	{0xB0CCEF57C65674A4ULL, "dae000ip_crowd_ff1_talk6.utp"},
	{0xB0CCF24B4A78053BULL, "mer120_hubert6020744.cl"},
	{0xB0CD12B4EBF35C82ULL, "lt_0_lowtown_a3_n_c_37_0.rml"},
	{0xB0CD556E48454489ULL, "gen00fl_aveline6139823.cl"},
	{0xB0CD987403724FBBULL, "89177706.ogg"},
	{0xB0CDFCF836EDC69DULL, "vxlt_gallowstemplar_4o2_0_0.vlm"},
	{0xB0CE1DE67B131BEDULL, "vxw_woundedcoast_zb_0_0.vlm"},
	{0xB0CE214179BA2617ULL, "568279187.ogg"},
	{0xB0CE23F029AD2C84ULL, "6097318_m.fxe"},
	{0xB0CE5F9D043E4805ULL, "943106868.ogg"},
	{0xB0CE87967EAB4CF8ULL, "321608208.ogg"},
	{0xB0CEAD4861780A19ULL, "zz_approval_debug6003572.cl"},
	{0xB0CEC1A91F380544ULL, "vxca_caverns2prc_ac7_0_0.vlm"},
	{0xB0CF63436A04592CULL, "hlac_lt_0_lowtown_a2_-131071_2y.dds"},
	{0xB0CF8ACFD4695840ULL, "6142632_m.fxe"},
	{0xB0CFB2614B4C4BA8ULL, "112460086.ogg"},
	{0xB0CFE41676354673ULL, "vxht_bartrand_n_648_0_0.vlm"},
	{0xB0D037289274E88FULL, "character_rendertodepth_p1.psh.fxo.dx11.assembly"},
	{0xB0D073EE6E35DEC2ULL, "mrl150_merrill6089925.cl"},
	{0xB0D13B124FB5216DULL, "vxw_bonepit_1vl_0_0.vlm"},
	{0xB0D34DFF4E7C075AULL, "t3_shd_stl.tnt"},
	{0xB0D34FD0C0255DF4ULL, "hlac_w_bonepitdestroyed_a3_131074_0z.dds"},
	{0xB0D37A5D91303BA6ULL, "ts_dwarven_prog_junk4.gda"},
	{0xB0D3B98C5B9C9DD4ULL, "277228884.ogg"},
	{0xB0D3F08345E44470ULL, "1015126707.ogg"},
	{0xB0D43C38845CAF27ULL, "vxca_caverns1_jdd_0_0.vlm"},
	{0xB0D479B2FE207F5DULL, "follower_banter6169853.cl"},
	{0xB0D487485C3FFE1FULL, "gen00fl_sebastian6188665.cl"},
	{0xB0D493FD4B1A774EULL, "framebuffereffects_diffusiondof_p8.psh.fxo.dx11"},
	{0xB0D4B884B855B48EULL, "w_bonepitmining_a2_125_0.msh"},
	{0xB0D4C684B855CCDFULL, "w_bonepitmining_a2_125_0.mao"},
	{0xB0D4D284B855E17CULL, "w_bonepitmining_a2_125_0.mmh"},
	{0xB0D4E84B256544EAULL, "and121_karl6093871.cl"},
	{0xB0D5009B9F7171A8ULL, "mcr310_cut_chantry_blows_up6087736.cl"},
	{0xB0D55B9AE17CAE24ULL, "211544999.ogg"},
	{0xB0D5A1C69B13BFFCULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_bookstand.mao"},
	{0xB0D5ADC69B13D39FULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_bookstand.mmh"},
	{0xB0D5B3C69B13DE41ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_bookstand.msh"},
	{0xB0D5BA94EFCDDF05ULL, "qun110_petrice6037093.cl"},
	{0xB0D60F4CA82369FEULL, "6201551_m.fxe"},
	{0xB0D6209DA65027C9ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.mmh"},
	{0xB0D66692C3702A8BULL, "follower_banter6173139.cl"},
	{0xB0D670131B562753ULL, "dae000_bartrand6107532.cl"},
	{0xB0D67C4EA6E6E6FEULL, "6135261_m.fxe"},
	{0xB0D69C698F694597ULL, "186793841.ogg"},
	{0xB0D6A278FE6F9232ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_pick_up_from_desk.gad"},
	{0xB0D6D9D1A3C96FE8ULL, "129094278.ogg"},
	{0xB0D71454022DC128ULL, "125304029.ogg"},
	{0xB0D74C542B2F55BBULL, "m_dae311_gamlens_house_nt.wbd"},
	{0xB0D77935C62AEBDAULL, "vxlt_undercityencounter_1p7_0_0.vlm"},
	{0xB0D7F059BDAEE78DULL, "6071063_m.fxe"},
	{0xB0D8AFC85360C8AAULL, "6050168_m.fxe"},
	{0xB0D8D2166E92E968ULL, "811891163.ogg"},
	{0xB0D8EA7C0A277952ULL, "vxw_sundermount_2ac_0_1.vlm"},
	{0xB0D9066A126E381EULL, "vxlt_undercityencounter_1ix_0_0.vlm"},
	{0xB0D90B686EE6C1DBULL, "zz_mcr_debug6080871.cl"},
	{0xB0D90EFE58DA1FB9ULL, "809360861.ogg"},
	{0xB0D9187350B8ACBBULL, "hlac_lt_0_lowtown_-3_6x.dds"},
	{0xB0D9B5D199C4961AULL, "737582262.ogg"},
	{0xB0D9D831B3F7860CULL, "6031295_m.fxe"},
	{0xB0DAA64FA621DD68ULL, "w_encounter2_camp_30_0.anb"},
	{0xB0DADF289F01D487ULL, "412961875.ogg"},
	{0xB0DAF5D6CFEA1259ULL, "vxht_brothel_n_4w9_0_0.vlm"},
	{0xB0DB7D25AC6B0C47ULL, "vxw_bonepitdestroyed_a3_1pd_0_0.vlm"},
	{0xB0DBED8B8386BBDCULL, "plt_rumor_act_of_mercy.plo"},
	{0xB0DC8FC36D09346DULL, "vxw_woundedcoast_wa_0_0.vlm"},
	{0xB0DD0E30A1895521ULL, "art\\levels\\races\\design\\_textures\\prp_herbs_r.dds"},
	{0xB0DDA578FE75B669ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_pick_up_from_desk.evt"},
	{0xB0DDFFDEEC1F438AULL, "w_encounter1_60_0.mao"},
	{0xB0DE03DEEC1F4AD9ULL, "w_encounter1_60_0.mmh"},
	{0xB0DE0DDEEC1F5BC7ULL, "w_encounter1_60_0.msh"},
	{0xB0DE59FA8E786DCAULL, "gen00fl_anders6119445.cl"},
	{0xB0DE6C1C28E6C734ULL, "370135311.ogg"},
	{0xB0DE8E99B5568843ULL, "vxlt_0_lowtown_d3s_0_0.vlm"},
	{0xB0DEF2604B3D23C0ULL, "vxlt_0_lowtown_docks_a3_n_bfd_0_0.vlm"},
	{0xB0DF331416CA2EDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.gad"},
	{0xB0DF881B118B7679ULL, "hlac_w_sundermount_131070_0x.dds"},
	{0xB0DFB6BCBA46631BULL, "vxlt_undercitydungeon_xn_0_0.vlm"},
	{0xB0E06F29482E317CULL, "6207740_m.fxe"},
	{0xB0E0915FA60E0BCEULL, "lgt220cr_shp_mag_tate.utc"},
	{0xB0E1A8A1D7A4234BULL, "vxht_keep_n_55s_0_0.vlm"},
	{0xB0E30E3592C3E6CFULL, "art\\vfx\\creatures\\fxc_rage_burst_p.dura.gad"},
	{0xB0E35F8F697BC1A6ULL, "6108474_m.fxe"},
	{0xB0E3A47C8ACA2DDBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keepthin.mmh"},
	{0xB0E3AE19CCC8D76BULL, "6134906_m.fxe"},
	{0xB0E3B07C8ACA4278ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keepthin.mao"},
	{0xB0E3C27C8ACA60C5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keepthin.msh"},
	{0xB0E42D0DF9675CC4ULL, "lt_undercitydungeon_3567.rml"},
	{0xB0E4D3990960E605ULL, "cd_fen110_cut_fenris_intro.bnk"},
	{0xB0E5BF5BB3EE7817ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_order_2.gad"},
	{0xB0E6281416D03ACCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.evt"},
	{0xB0E64019D14EF411ULL, "hlac_lt_0_lowtown_docks_a3_0_2x.dds"},
	{0xB0E6DD9EC524C6C7ULL, "mag360_thrask6053528.cl"},
	{0xB0E736ECAE65A321ULL, "ht_0_hightown_a3_19_0.rml"},
	{0xB0E740173DD42DCDULL, "vxlt_0_lowtown_a2_ewl_0_0.vlm"},
	{0xB0E750F16A0F8673ULL, "hlac_w_woundedcoast_0_10y.dds"},
	{0xB0E79334BBB9C2E5ULL, "ht_brothel_n_3382.rml"},
	{0xB0E7EDD97E2EA4F2ULL, "vxlt_foundry_n_v0_0_0.vlm"},
	{0xB0E869A3A2039F04ULL, "hlac_lt_0_lowtown_n_-131071_2z.dds"},
	{0xB0E89E4EF06B1EBAULL, "6087191_m.fxe"},
	{0xB0E8C013C2F60F8FULL, "6162691_m.fxe"},
	{0xB0E8DD04F3CF7AA8ULL, "vxlt_undercityserialkiller_1uf_0_0.vlm"},
	{0xB0E9AF56B908A6B2ULL, "framebuffereffects_horizontalblurshadows_p0.vsh.fxo.dx11.assembly"},
	{0xB0E9E6423B6790FDULL, "mcr310_argument6092082.cl"},
	{0xB0E9EF029E6124C7ULL, "vxw_bonepitdestroyed_a3_1ug_0_0.vlm"},
	{0xB0EA233592CA297CULL, "art\\vfx\\creatures\\fxc_rage_burst_p.dura.evt"},
	{0xB0EA4C036CC0633CULL, "vxlt_0_lowtown_a3_n_c_hx1_0_0.vlm"},
	{0xB0EA8D78FE8017D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_pick_up_from_desk.ani"},
	{0xB0EAEEB4BCF6381EULL, "character_character_p81.psh.fxo.dx11"},
	{0xB0EB45E32A485872ULL, "vxlt_gallowstemplar_f_5r8_0_0.vlm"},
	{0xB0ECB5B1D1C67C6DULL, "440199075.ogg"},
	{0xB0ECD97B8F6862C4ULL, "vxlt_0_lowtown_a3_n_c_hpa_0_0.vlm"},
	{0xB0ED007911B3705CULL, "316788836.ogg"},
	{0xB0ED4C776A22639FULL, "mer120_hubert6020656.cl"},
	{0xB0ED99427E4F2D89ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.anb"},
	{0xB0EDAB96A464F34AULL, "gen_im_cns_pot_health_01.uti"},
	{0xB0EE60C49FAAA088ULL, "vxht_chantry_n_36o_0_0.vlm"},
	{0xB0EE694BFA3C8B3DULL, "mrl310_party_comments6142847.cl"},
	{0xB0EE8CB170D14C43ULL, "character_character_p156.psh.fxo.dx11"},
	{0xB0EEB58276085868ULL, "mag110tr_vincentoarrest.ncs"},
	{0xB0EF1E26FFEF069BULL, "6042368_m.fxe"},
	{0xB0EF68DE9F02C2F2ULL, "6127176_m.fxe"},
	{0xB0EFE5BAD49AAE25ULL, "lt_0_lowtown_docks_a3_18_0_ll.mmh"},
	{0xB0F00811506225D8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate.phy"},
	{0xB0F0630D2917DD18ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_0y.dds"},
	{0xB0F11AE96A0EC3D7ULL, "376839058.ogg"},
	{0xB0F1BE64CA742EDAULL, "one300_zevran6071480.cl"},
	{0xB0F215365BDAE302ULL, "6042138_m.fxe"},
	{0xB0F22B3581969236ULL, "dae000_serendipity6161414.cl"},
	{0xB0F2F4047E1296D7ULL, "mag110_arianni6016806.cl"},
	{0xB0F3101416DA9C25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_exit.ani"},
	{0xB0F3A5B58F59E8EEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelcloth.msh"},
	{0xB0F3ACE808B5CDD2ULL, "6088222_m.fxe"},
	{0xB0F3B3B58F5A003FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelcloth.mao"},
	{0xB0F3BFB58F5A14DCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelcloth.mmh"},
	{0xB0F3D2D2738FADB1ULL, "llac_w_woundedcoast_65537_0.dds"},
	{0xB0F406A309E2598FULL, "6033161_f.fxe"},
	{0xB0F4415866028C95ULL, "follower_banter6141907.cl"},
	{0xB0F497F4A6562CF9ULL, "hlac_lt_backalley_n_0_15x.dds"},
	{0xB0F4C257C0E37156ULL, "426213831.ogg"},
	{0xB0F58DF6CF4B36BBULL, "lt_0_lowtown_docks_n_233_0.msh"},
	{0xB0F59FF6CF4B5506ULL, "lt_0_lowtown_docks_n_233_0.mao"},
	{0xB0F5ABF6CF4B69A5ULL, "lt_0_lowtown_docks_n_233_0.mmh"},
	{0xB0F5BACAF269B10CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curvel.msh"},
	{0xB0F5BB07085F3ACFULL, "vxlt_blackemporium_uw_0_0.vlm"},
	{0xB0F5C0CAF269BBFEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curvel.mmh"},
	{0xB0F5CCCAF269CF95ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curvel.mao"},
	{0xB0F5EA23F65A691EULL, "vxlt_undercityencounter_1cz_0_0.vlm"},
	{0xB0F5EB49581D6115ULL, "6135690_f.fxe"},
	{0xB0F63FFDCF42A8EAULL, "vfx_chunkyparticle_p1.psh.fxo"},
	{0xB0F655F5FA4CD590ULL, "6083647_m.fxe"},
	{0xB0F6C7B90D290FA7ULL, "vxw_encounter2_camp_na_0_0.vlm"},
	{0xB0F72B3592D4C095ULL, "art\\vfx\\creatures\\fxc_rage_burst_p.dura.ani"},
	{0xB0F731F0B4EC280FULL, "6167591_m.fxe"},
	{0xB0F781640238FC2FULL, "static_static_p17.psh.fxo.dx11"},
	{0xB0F86ADCB3F2AA2EULL, "6042264_m.fxe"},
	{0xB0F88F1AD4A257FDULL, "45154233.ogg"},
	{0xB0F8AC09D4421D06ULL, "llac_w_sundermount_-2_0.dds"},
	{0xB0F94A4FCFB5D69AULL, "6180005_m.fxe"},
	{0xB0F9898EC92CB6E2ULL, "llac_lt_0_lowtown_-262143_0.dds"},
	{0xB0F9B64AC565D4DBULL, "cs_mcr370cs_player_rules.wbd"},
	{0xB0F9D4B9875AB0C4ULL, "gen00fl_anders6125844.cl"},
	{0xB0F9EC5BB3FF6D9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_order_2.ani"},
	{0xB0FA5C5085F3142CULL, "272364073.ogg"},
	{0xB0FA7D0549B6EC7CULL, "llac_lt_0_lowtown_docks_a3_n_65534_5.dds"},
	{0xB0FA9852B5AD70B2ULL, "str_rock_sm.mao"},
	{0xB0FAA26D1194094AULL, "vxw_blightlands_nw_0_0.vlm"},
	{0xB0FAB184B8767847ULL, "w_bonepitmining_a2_125_0.phy"},
	{0xB0FAC96072D50C3DULL, "ser121_jethann6104546.cl"},
	{0xB0FAF13484B34349ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_shd_reva_0s.dds"},
	{0xB0FAFCA4DAB0325BULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lrgchair_sa.dds"},
	{0xB0FB0372B0645EA1ULL, "361671704.ogg"},
	{0xB0FB6A3D5C66A3ACULL, "453960269.ogg"},
	{0xB0FB95ABE8E5B590ULL, "972968393.ogg"},
	{0xB0FB96C24A579430ULL, "6091644_m.fxe"},
	{0xB0FC14A85666B5E0ULL, "504045745.ogg"},
	{0xB0FCAC1A4E8C1551ULL, "6156543_f.fxe"},
	{0xB0FCE99612FB8E12ULL, "hlac_ht_mansion2_818_36y.dds"},
	{0xB0FD6BEFA0EF3702ULL, "489800542.ogg"},
	{0xB0FDA0EAFFEA3AB7ULL, "dre000_dougal6130973.cl"},
	{0xB0FE4F2F8CD58467ULL, "llac_ht_keep_4519_3.dds"},
	{0xB0FE60CB2FDD07B9ULL, "820745153.ogg"},
	{0xB0FEA0F1AE13B3CAULL, "6027395_m.fxe"},
	{0xB0FF1D9DA67327FEULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.anb"},
	{0xB0FF82110503DB08ULL, "112728564.ogg"},
	{0xB0FFB1B29BBAE5EEULL, "681937189.ogg"},
	{0xB0FFCD7E0FCEC42DULL, "lgt000ip_rusty_market_key.utp"},
	{0xB1000A3879E3296DULL, "6059321_f.fxe"},
	{0xB10043494D67F2A6ULL, "6106674_m.fxe"},
	{0xB100E1A4ABFEA235ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexwild_n.dds"},
	{0xB10142D216B13D9EULL, "vxw_sundermount_370_0_1.vlm"},
	{0xB1014D23A9728CD0ULL, "vxht_keep_625_0_0.vlm"},
	{0xB101583759B3C138ULL, "mag110_arianni6015770.cl"},
	{0xB1016309BA2727A2ULL, "6081965_f.fxe"},
	{0xB101D1E8EFFC1FD3ULL, "hlac_lt_0_lowtown_a3_n_c_1_10z.dds"},
	{0xB101FCA9E22700D1ULL, "darktown_male_fereldan_sta6175146.cl"},
	{0xB10211612DCFD8D7ULL, "vxht_0_hightown_a3_cac_0_0.vlm"},
	{0xB1027F835F59EFCAULL, "art\\characters\\playercharacter\\shared\\textures\\default_flatnm.dds"},
	{0xB102854E4E7E5FA9ULL, "dae000_bartrand6027540.cl"},
	{0xB1029F9B4025B505ULL, "6145095_f.fxe"},
	{0xB103A14FA644D9F0ULL, "w_encounter2_camp_30_0.mao"},
	{0xB103A54FA644E003ULL, "w_encounter2_camp_30_0.mmh"},
	{0xB103B34FA644F85DULL, "w_encounter2_camp_30_0.msh"},
	{0xB103D381A7AECC42ULL, "qun230cr_guild_contact6068869.cl"},
	{0xB10456A3314240ECULL, "vxca_caverns2_80j_0_0.vlm"},
	{0xB1058DF99609611FULL, "hlac_lt_0_lowtown_docks_n_131071_5z.dds"},
	{0xB1063510F3BF6230ULL, "vxlt_blackemporium_nu_0_0.vlm"},
	{0xB106379C831237C2ULL, "6105680_m.fxe"},
	{0xB106D2BC824D55C7ULL, "vxw_bonepitdestroyed_a3_1ca_0_0.vlm"},
	{0xB10705D25230024EULL, "541196257.ogg"},
	{0xB1070FB4EC24D07EULL, "lt_0_lowtown_a3_n_c_37_0.anb"},
	{0xB1072FBD50AC943CULL, "vxlt_0_lowtown_a3_n_c_hb7_0_0.vlm"},
	{0xB107BC68DDF274D2ULL, "6115112_f.fxe"},
	{0xB107EE35A4A15DFBULL, "6042045_m.fxe"},
	{0xB107F14211A041A5ULL, "hlac_ht_0_hightown_a3_1_0z.dds"},
	{0xB107F5963C78553CULL, "gen00fl_merrill6108263.cl"},
	{0xB10838ED43B51952ULL, "6127279_m.fxe"},
	{0xB108C69308D6E09EULL, "dae000_pc_home_ambients6208345.cl"},
	{0xB1090875DE63505AULL, "6143506_f.fxe"},
	{0xB109144A038A21CAULL, "vxw_blightlands_sp_0_0.vlm"},
	{0xB1098FC606820113ULL, "zz_followers_banter6158343.cl"},
	{0xB109A037C5FAEAA5ULL, "6160668_m.fxe"},
	{0xB10A20407841A1E5ULL, "hlac_lt_0_lowtown_n_65533_3x.dds"},
	{0xB10A3267F136B8B1ULL, "431278695.ogg"},
	{0xB10ACB7ACE7E6D1FULL, "vxht_chantry_n_27r_0_0.vlm"},
	{0xB10AE72BDEEDEB88ULL, "plt_qun000pt_amb_test.nss"},
	{0xB10B1CE7F198AE99ULL, "and111_lirene6084751.cl"},
	{0xB10B71E86EE068FFULL, "6036473_m.fxe"},
	{0xB10C54CB44402BDFULL, "vxdr_0_deeproads_gfy_0_1.vlm"},
	{0xB10C9F1E50C6759BULL, "387283004.ogg"},
	{0xB10CB9182E6A58FEULL, "vxht_0_hightown_a7k_0_0.vlm"},
	{0xB10CD577949CF258ULL, "vxlt_undercitydungeon_1nf_0_0.vlm"},
	{0xB10CF3DEA6902072ULL, "w_sundermount_probe_3a2.dds"},
	{0xB10D14A901476AC6ULL, "513020845.ogg"},
	{0xB10D168F5C494D0CULL, "mag101_grace6053107.cl"},
	{0xB10D466DF93020D6ULL, "6019884_m.fxe"},
	{0xB10D5E84361730D4ULL, "ts_apostates_02.gda"},
	{0xB10D678D04A1208CULL, "490954971.ogg"},
	{0xB10DC79BCB8E7E70ULL, "vxht_bartrand_n_5s9_0_0.vlm"},
	{0xB10E009734F24ECEULL, "vxlt_0_lowtown_docks_a3_n_8pf_0_0.vlm"},
	{0xB10E982B0AD69F6AULL, "mrl000_marethari6129429.cl"},
	{0xB10EBAD74A43C575ULL, "6016161_m.fxe"},
	{0xB10ECD0AF7CCA0DCULL, "hlac_lt_gallowstemplar_n_2395_1y.dds"},
	{0xB10EF60FD3823A03ULL, "vxlt_darktown_41r_0_0.vlm"},
	{0xB10F292FE33A7EC0ULL, "vxht_bartrand_n_5we_0_0.vlm"},
	{0xB10FB4AF08A7561EULL, "vxw_sundermount_323_0_1.vlm"},
	{0xB11020EC5FEEC6F6ULL, "dae000cr_grand_cleric6148084.cl"},
	{0xB1102D6719A7C67AULL, "498922778.ogg"},
	{0xB110CF35471B46D9ULL, "art\\characters\\creatures\\ragedemon\\animation\\appr.ani"},
	{0xB1111E00E8661544ULL, "hlac_lt_0_lowtown_n_131072_2x.dds"},
	{0xB111F02BDEF4199BULL, "plt_qun000pt_amb_test.plo"},
	{0xB11226CF1DFCCB42ULL, "gen00fl_anders6103918.cl"},
	{0xB112729C319712B2ULL, "vxlt_0_lowtown_euu_0_0.vlm"},
	{0xB112A4A8136A2338ULL, "dae000ip_crowd_ff5_talk6.utp"},
	{0xB112CC97AA54F5ABULL, "6149032_m.fxe"},
	{0xB112D96401DD0C03ULL, "dae000_gen_turn_in_56178428.cl"},
	{0xB1133C447EA53D9EULL, "186416806.ogg"},
	{0xB1137738AA584407ULL, "696539638.ogg"},
	{0xB11382F81CDF1C54ULL, "6090027_m.fxe"},
	{0xB113AB72F11540A7ULL, "vxw_encounter2_camp_dc_0_0.vlm"},
	{0xB113B01A1C951291ULL, "hlac_ca_caverns1_mines_24423_5y.dds"},
	{0xB113E345E14BBA6BULL, "lowtown_male_lowlife1_sta6164982.cl"},
	{0xB114314114314A5DULL, "vxlt_gallowscourtyard_n_4ln_0_0.vlm"},
	{0xB114A64FA6535528ULL, "w_encounter2_camp_30_0.phy"},
	{0xB114A6EDE5B57AB0ULL, "vxlt_blackemporium_cr_0_0.vlm"},
	{0xB11528283E121F0FULL, "746395759.ogg"},
	{0xB11550983AF01EC1ULL, "follower_banter6155757.cl"},
	{0xB1156F56BC8EB756ULL, "77838434.ogg"},
	{0xB1158B0E0DA5651DULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb10a.phy"},
	{0xB115D3C3413A9485ULL, "ser270cr_templar.utc"},
	{0xB1161095D2CB6DA6ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.impa.evt"},
	{0xB116F9C98B0A5B2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\enter.ani"},
	{0xB117166BD781716EULL, "6066335_fakemeredith_arrive.cut"},
	{0xB117192E523494EDULL, "279985638.ogg"},
	{0xB11779B82AA0415DULL, "vxht_0_hightown_a3_cdz_0_0.vlm"},
	{0xB1178626F5803A4AULL, "vxw_blightlands_tq_0_0.vlm"},
	{0xB117A7AFA235D548ULL, "6128939_m.fxe"},
	{0xB117AA354721261BULL, "art\\characters\\creatures\\ragedemon\\animation\\appr.gad"},
	{0xB118236F0B7A7786ULL, "vxht_0_hightown_a2l_0_0.vlm"},
	{0xB1189EB58F78F8A7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelcloth.phy"},
	{0xB118A86707C1FA4AULL, "vxw_sundermount_2gl_0_0.vlm"},
	{0xB118AFC202B4963FULL, "6106954_m.fxe"},
	{0xB1191D8B6627DB96ULL, "832903919.ogg"},
	{0xB1195ABF03BF188FULL, "6041909_m.fxe"},
	{0xB119746309839BEDULL, "vxw_bonepit_1og_0_0.vlm"},
	{0xB119EBA7032B9F51ULL, "6000711_m.fxe"},
	{0xB11A35360D3E33C4ULL, "vrc321_party_barks6096137.cl"},
	{0xB11A658328B7ECC5ULL, "1006666848.ogg"},
	{0xB11A9234BBE51A0DULL, "ht_brothel_n_3382.anb"},
	{0xB11AA54FA657BF20ULL, "w_encounter2_camp_30_0.rml"},
	{0xB11B007C07B13C5CULL, "hlac_ht_brothel_n_3364_1z.dds"},
	{0xB11B604C024CF02EULL, "6129986_f.fxe"},
	{0xB11B682861194D51ULL, "mag121_idunna6084208.cl"},
	{0xB11B8BF2FB340CDEULL, "679454544.ogg"},
	{0xB11BF0DE19A8552FULL, "37991738.ogg"},
	{0xB11C2795D2D000F1ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.impa.gad"},
	{0xB11C3B97913F7897ULL, "art\\levels\\atmosphere\\sky\\sb03_back.dds"},
	{0xB11C55F57F61D7D1ULL, "gen_im_trs_drk_junk1_04.uti"},
	{0xB11CE598ED78DF53ULL, "vxw_encounter2_rp_0_1.vlm"},
	{0xB11D05D484F659EFULL, "vxlt_0_lowtown_a2_gc4_0_0.vlm"},
	{0xB11D9B0CD5309740ULL, "vxht_bartrand_n_5zb_0_0.vlm"},
	{0xB11E06C98B10906EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\enter.gad"},
	{0xB11E30DD8D5EBDD0ULL, "vxht_keep_683_0_0.vlm"},
	{0xB11EC73547277600ULL, "art\\characters\\creatures\\ragedemon\\animation\\appr.evt"},
	{0xB11EFE5B17A9D1EFULL, "539738567.ogg"},
	{0xB11FD1B6FFDFAABCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_bootser_n.dds"},
	{0xB12043FF9C57482AULL, "dae000ip_crowd_mf4_talk6.utp"},
	{0xB12068C870B6413DULL, "vxw_sundermount_2e5_0_0.vlm"},
	{0xB1209A82C4A01B27ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_right.gad"},
	{0xB121310768D2A56CULL, "vxht_0_hightown_b5l_0_0.vlm"},
	{0xB121512B2C9A9F8FULL, "6103408_player_addmissive.cut"},
	{0xB12167EC627C9DE8ULL, "wm_ico_deeproads_roll.dds"},
	{0xB121731C8BF656ACULL, "6127211_m.fxe"},
	{0xB1219463E3725347ULL, "vxdr_primevaltomb_8md_0_0.vlm"},
	{0xB1223328E4C3B45CULL, "183839020.ogg"},
	{0xB1226618AD580FF8ULL, "6093893_m.fxe"},
	{0xB122A45988360FB2ULL, "and121_karl6093909.cl"},
	{0xB122AB331D1AECEEULL, "950785200.ogg"},
	{0xB122C2909370A40FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_004a_0t.dds"},
	{0xB122F1F0C26FC415ULL, "vxw_encounter1_ir_0_1.vlm"},
	{0xB1232895D2D6209FULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.impa.ani"},
	{0xB123685F9BDA72CDULL, "vxht_keep_n_6ib_0_0.vlm"},
	{0xB123C627047369D7ULL, "vxlt_gallowstemplar_n_5j1_0_0.vlm"},
	{0xB12419FA18FE3B17ULL, "777125152.ogg"},
	{0xB12467D76A65E8C3ULL, "664205958.ogg"},
	{0xB124AE6F4DBD807AULL, "vxca_caverns2prc_aek_0_0.vlm"},
	{0xB124CF35AF6FAE9EULL, "6082167_m.fxe"},
	{0xB124ED9DC15F5827ULL, "hlac_lt_gallowscourtyard_-1_1x.dds"},
	{0xB12501C98B16A64DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\enter.evt"},
	{0xB1250A88AA2BB92BULL, "hlac_lt_gallowscourtyard_65534_2x.dds"},
	{0xB1250D748454E5A2ULL, "900424948.ogg"},
	{0xB125CFB65CC2F1B8ULL, "zz_vault_debug6203233.cl"},
	{0xB1261B6306278684ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_headintro.ani"},
	{0xB12670E3A99DA7D9ULL, "gen00fl_anders6116782.cl"},
	{0xB1268C0E0DB3DA22ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb10a.mmh"},
	{0xB1268E0E0DB3DDD8ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb10a.msh"},
	{0xB126A0EF5740C4C4ULL, "vxlt_0_lowtown_a3_n_c_hle_0_0.vlm"},
	{0xB126A6B61D63F796ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_exit.ani"},
	{0xB126A77E19B7B224ULL, "829615054.ogg"},
	{0xB126BA5F988EE42BULL, "mer140_bonepit_comments6077843.cl"},
	{0xB126DFAFA5504DDEULL, "872282452.ogg"},
	{0xB1278F82C4A62734ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_right.evt"},
	{0xB127CAF9E125376AULL, "6016430_m.fxe"},
	{0xB127E7EF5E3F22C3ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_007a_0t.dds"},
	{0xB127FCF4112495BCULL, "hlac_ht_mansion2b_818_37z.dds"},
	{0xB12812356C395A2EULL, "hlac_lt_0_lowtown_docks_a3_n_-3_7y.dds"},
	{0xB1281BE3DA54EA28ULL, "817316338.ogg"},
	{0xB12853B9AE38216CULL, "gen00fl_fenris6106647.cl"},
	{0xB128B403BA576352ULL, "6108940_m.fxe"},
	{0xB128D1DDDBC72F0DULL, "gen00fl_fenris6127128.cl"},
	{0xB129023706708578ULL, "vxht_mansion_n_4a9_0_0.vlm"},
	{0xB129430EA53B2509ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2cocoon.mao"},
	{0xB1294F0EA53B39A2ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2cocoon.mmh"},
	{0xB129510EA53B3D58ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2cocoon.msh"},
	{0xB12966247A88CFA4ULL, "ser111ar_foundry.lst"},
	{0xB12990170899BAA8ULL, "zz_dae_debug6029798.cl"},
	{0xB129A35F52374033ULL, "fx_firetorchfb_blend.mao"},
	{0xB12A00A8065ADD9EULL, "755166677.ogg"},
	{0xB12A0FEFF5470294ULL, "zz_mag_debug6053404.cl"},
	{0xB12A6B80F52B2A9DULL, "vxht_0_hightown_c90_0_0.vlm"},
	{0xB12B3FB2ED70E583ULL, "gen00fl_merrill6130067.cl"},
	{0xB12BC57C8B081B10ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keepthin.phy"},
	{0xB12C69899232E6FAULL, "vxlt_0_lowtown_a2_n_c_hnp_0_0.vlm"},
	{0xB12C6E4F21AB6EA0ULL, "vxw_sundermount_30v_0_0.vlm"},
	{0xB12C76CA234D0A4EULL, "vxlt_0_lowtown_a2_n_c_byk_0_0.vlm"},
	{0xB12C8F8947398D24ULL, "vxca_caverns2_77p_0_0.vlm"},
	{0xB12D1863062DA0E6ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_headintro.gad"},
	{0xB12D20999DB2D599ULL, "hlac_w_woundedcoast_0_11x.dds"},
	{0xB12D3B29DB195D96ULL, "lt_0_lowtown_docks_n_113_0.msh"},
	{0xB12D4529DB196E94ULL, "lt_0_lowtown_docks_n_113_0.mmh"},
	{0xB12D48E7CBD815F1ULL, "hlac_lt_undercityserialkiller_-1_48y.dds"},
	{0xB12D4929DB1975A7ULL, "lt_0_lowtown_docks_n_113_0.mao"},
	{0xB12D5A98FB22B5B2ULL, "hlac_ht_keep_n_3731_0x.dds"},
	{0xB12D7D062A90F9B5ULL, "604540866.ogg"},
	{0xB12E7078EBD87E78ULL, "hlac_lt_gallowstemplar_2394_0z.dds"},
	{0xB12ECA88EA5FD863ULL, "6182712_m.fxe"},
	{0xB12F0E4735B6633FULL, "vxca_caverns2_80e_0_0.vlm"},
	{0xB12F65F27B0D50BCULL, "78601107.ogg"},
	{0xB12F66247A8D3B0DULL, "ser111ar_foundry.ncs"},
	{0xB12FE1C9F1EA6007ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_p.impa.gad"},
	{0xB13005AA51A7713AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cart.phy"},
	{0xB1304716F8835232ULL, "110229731.ogg"},
	{0xB13071CAF454D44BULL, "6077944_m.fxe"},
	{0xB130EF255F0E93A5ULL, "vxdr_primevaltomb_2po_0_0.vlm"},
	{0xB1318AA7C9A1ABB0ULL, "vxlt_blackemporium_yp_0_0.vlm"},
	{0xB1318E7DBACB6446ULL, "qcr270_fake_meredith6066350.cl"},
	{0xB1321A9F87DB4F11ULL, "ts_kirkwall_acc_props.gda"},
	{0xB1324E4444B9CFCFULL, "follower_banter6136697.cl"},
	{0xB13280997A6E0BE3ULL, "vxlt_0_lowtown_a2_n_c_df9_0_0.vlm"},
	{0xB132EB0104D1970CULL, "6100252_m.fxe"},
	{0xB132EE4B126EC2F6ULL, "vxlt_gamlenshouse_n_6x_0_0.vlm"},
	{0xB1331363063203C5ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_headintro.evt"},
	{0xB134AEB61D704287ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_exit.evt"},
	{0xB134C87781EF75DBULL, "577228242.ogg"},
	{0xB134DAFEAB6456EAULL, "vxlt_darktown_4u3_0_0.vlm"},
	{0xB1353FB7FE6120CCULL, "lgt320_mis_resp_swordhunter6150324.cl"},
	{0xB1358782C4B2576DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_right.ani"},
	{0xB1358C20EBAE2B4AULL, "vxw_sundermount_2mn_0_0.vlm"},
	{0xB135943B7E7263B6ULL, "qun110_petrice6018679.cl"},
	{0xB1368882F733D692ULL, "163427438.ogg"},
	{0xB136A6431AC33E40ULL, "1056460288.ogg"},
	{0xB136AB1E08D45306ULL, "267414695.ogg"},
	{0xB136D6C9F1F06C14ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_p.impa.evt"},
	{0xB136D99316307D0CULL, "lt_0_lowtown_a2_n_c_53_0_lc.mmh"},
	{0xB1377413F8669DF8ULL, "vxht_mansion_n_4d6_0_0.vlm"},
	{0xB137B04F6CAC5416ULL, "vxlt_0_lowtown_a3_n_c_fib_0_0.vlm"},
	{0xB137B5C00C0128E7ULL, "470400158.ogg"},
	{0xB137BA98E679DC79ULL, "gen00fl_isabela6090820.cl"},
	{0xB1382D45D6592867ULL, "qcr220_grey_wardens6207430.cl"},
	{0xB138A236D94D00E8ULL, "6139865_m.fxe"},
	{0xB138B6C69B6860E4ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_bookstand.phy"},
	{0xB138BB08BB8A02D1ULL, "519877574.ogg"},
	{0xB138CF46FD330B88ULL, "qcr000_arishok6166993.cl"},
	{0xB138EFF18A0BB0B6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\pause.ani"},
	{0xB13959A8BA0BBBD7ULL, "hlac_ht_chantry_2534_43y.dds"},
	{0xB139A1F19DCC6712ULL, "follower_banter6173082.cl"},
	{0xB139A8F8E74BBB27ULL, "lgt260_eau_tmp_roderick6143699.cl"},
	{0xB13A7F9C0796BE6BULL, "pro000_merchant6051595.cl"},
	{0xB13B0CA79C188F9DULL, "6034715_f.fxe"},
	{0xB13B25AD392F5953ULL, "571241322.ogg"},
	{0xB13B5988A2F1600BULL, "vxw_bonepitmining_a2_21j_0_0.vlm"},
	{0xB13B721475104FEDULL, "hlac_w_blightlands_0_11x.dds"},
	{0xB13BC1710348A33BULL, "w_woundedcoast_206_0.mmh"},
	{0xB13BCD710348B7D8ULL, "w_woundedcoast_206_0.mao"},
	{0xB13BCFB61D7699D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_exit.gad"},
	{0xB13BDF710348D625ULL, "w_woundedcoast_206_0.msh"},
	{0xB13C4992C7AFF356ULL, "vxdr_primevaltomb_352_0_0.vlm"},
	{0xB13CCCED935EE40EULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef02.msh"},
	{0xB13CDDABF1B68AC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\basestaff\\blendtree\\devastator_movement.anb"},
	{0xB13CE6ED935F10FCULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef02.mmh"},
	{0xB13D4C8254A2723DULL, "vxw_sundermount_2s3_0_0.vlm"},
	{0xB13DE8D485E7FA6EULL, "ser270_emeric6149346.cl"},
	{0xB13ED734EBD9D46AULL, "882294803.ogg"},
	{0xB13F4E4F6C47B27DULL, "gr_ft_clo.mao"},
	{0xB13FF5F0C8C67F23ULL, "art\\vfx\\combat\\vfx_c_profane_s_crust_c.dura.evt"},
	{0xB13FFC84BB97C430ULL, "vxlt_blackemporium_rq_0_0.vlm"},
	{0xB13FFD6B5B5D16DBULL, "mcr310_loyalist_templars6076911.cl"},
	{0xB140191FE070AFBAULL, "vxw_blightlands_12b_0_0.vlm"},
	{0xB14022F464BE6BF0ULL, "dae000_meredith6085244.cl"},
	{0xB1403E3C944A9AEBULL, "447421034.ogg"},
	{0xB1406BDEE9F857CEULL, "art\\vfx\\cinematic\\vfx_dust_impact_p.impa.evt"},
	{0xB140B871FB2D2972ULL, "ave121_jeven6105502.cl"},
	{0xB14125A505B0E981ULL, "hlac_w_bonepit_131076_0y.dds"},
	{0xB141E65F7D585F9DULL, "6130957_m.fxe"},
	{0xB14211147BA4956EULL, "117666913.ogg"},
	{0xB1421804185D798DULL, "975408853.ogg"},
	{0xB14278B428CDF7F3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_d.dds"},
	{0xB142D44190D3FB1CULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.dura.ani"},
	{0xB1432687EA5A427EULL, "plt_cod_bks_sermons.plo"},
	{0xB143541635A43742ULL, "vxw_encounter2_camp_p7_0_1.vlm"},
	{0xB14468920975C721ULL, "lgt000cr_baz_car_loy.utc"},
	{0xB144CEC9F1FC9C4DULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_p.impa.ani"},
	{0xB144FEDB09C53BE9ULL, "dae021_brothel_onenter6196702.cl"},
	{0xB1455168B7F4F852ULL, "hlac_lt_gallowsprison_n_-1_8y.dds"},
	{0xB14561A5FB7050F0ULL, "hlac_ca_caverns2_5802_19y.dds"},
	{0xB145902468B9A0A4ULL, "260603119.ogg"},
	{0xB145DDE92DD36366ULL, "rev_clo.mao"},
	{0xB1461A15E4DA6B8DULL, "llac_ht_0_hightown_a3_-65534_1.dds"},
	{0xB1466470AE285E74ULL, "ip_restrict_level.ncs"},
	{0xB1467C2C1E2A9079ULL, "377850980.ogg"},
	{0xB146CEF0C8CC5BA4ULL, "art\\vfx\\combat\\vfx_c_profane_s_crust_c.dura.gad"},
	{0xB146D674134B8115ULL, "zz_qun_debug6028334.cl"},
	{0xB146F82CE4E4259CULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_foam01.mao"},
	{0xB14722B9CCC21CF9ULL, "art\\characters\\creatures\\buzzard\\c_buzzardg.msh"},
	{0xB1472CB9CCC22DE7ULL, "art\\characters\\creatures\\buzzard\\c_buzzardg.mmh"},
	{0xB14762DEE9FE6779ULL, "art\\vfx\\cinematic\\vfx_dust_impact_p.impa.gad"},
	{0xB147E0EF92A1237EULL, "one140cr_charade6137604.cl"},
	{0xB147F9D9B473A7DDULL, "773292838.ogg"},
	{0xB148074BEE64D46AULL, "6044802_m.fxe"},
	{0xB14852FE4B7FA05AULL, "dust.mao"},
	{0xB14881075014577CULL, "379305283.ogg"},
	{0xB148C1C73B29061FULL, "6155859_m.fxe"},
	{0xB148E0C3E5C240E7ULL, "hlac_lt_0_lowtown_docks_n_-65540_0y.dds"},
	{0xB148F4B5AA81CFAEULL, "vxlt_gallowsprison_n_6gd_0_0.vlm"},
	{0xB148F92CDF2AF5DDULL, "ser221_quentin6162446.cl"},
	{0xB1493E2F683B3623ULL, "one110_meeran6089878.cl"},
	{0xB14963F06AE0E50DULL, "gen00fl_anders6125681.cl"},
	{0xB14976EE22BB0DA7ULL, "vrc321_party_barks6096305.cl"},
	{0xB149E14190DA305EULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.dura.gad"},
	{0xB149E1D59FF35299ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_lgtc_0t.dds"},
	{0xB14A0573EE28EE56ULL, "lightbulb01.dds"},
	{0xB14A26868567AB3EULL, "vxlt_hangedman_30j_0_0.vlm"},
	{0xB14A2AC17E9B6DAAULL, "6219068_m.fxe"},
	{0xB14ABCE703139424ULL, "lgt160_wk_soph_hunter6119730.cl"},
	{0xB14B497079E19BD3ULL, "hlac_ca_caverns2prc_5802_61x.dds"},
	{0xB14B80934E8ADE3FULL, "vfx_test.arl"},
	{0xB14BBE5F46988ABDULL, "vxw_sundermount_2t2_0_0.vlm"},
	{0xB14BCC05B9FD6093ULL, "t3_hey_am1.tnt"},
	{0xB14BE8AB6D46C7BFULL, "vxht_0_hightown_b5c_0_0.vlm"},
	{0xB14BF51150B0646DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate.msh"},
	{0xB14C031150B07BA0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate.mao"},
	{0xB14C071150B082F3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate.mmh"},
	{0xB14C1C4F2F9E7E82ULL, "lt_0_lowtown_docks_n_100_0.msh"},
	{0xB14C1E4F2F9E8238ULL, "lt_0_lowtown_docks_n_100_0.mmh"},
	{0xB14C2A4F2F9E96DBULL, "lt_0_lowtown_docks_n_100_0.mao"},
	{0xB14C3396097D5907ULL, "53816572.ogg"},
	{0xB14C3EB495E9B4D2ULL, "follower_banter6162088.cl"},
	{0xB14D18F18A1C9FF0ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\pause.gad"},
	{0xB14D5E4E4F23FFEFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_nicerchair_0.pwk"},
	{0xB14D6CEC2D8FA1EFULL, "6062776_f.fxe"},
	{0xB14D73DEEA02EFC7ULL, "art\\vfx\\cinematic\\vfx_dust_impact_p.impa.ani"},
	{0xB14DDC956BC61F17ULL, "vxht_brothel_n_4b4_0_0.vlm"},
	{0xB14DDDF0C8D29382ULL, "art\\vfx\\combat\\vfx_c_profane_s_crust_c.dura.ani"},
	{0xB14DDF3DD2230F4DULL, "6071610_m.fxe"},
	{0xB14E6D247AA7F797ULL, "ser111ar_foundry.are"},
	{0xB14EA7004CDD825DULL, "w_encounter2_22_0.phy"},
	{0xB14F93962B741561ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_lgta_0s.dds"},
	{0xB14F9FDC982E910FULL, "6044894_m.fxe"},
	{0xB1500543646E2A02ULL, "416959095.ogg"},
	{0xB1501076D4CC41AEULL, "vxht_keep_371_0_0.vlm"},
	{0xB15018B10E741113ULL, "6100134_m.fxe"},
	{0xB1503815BBE66C94ULL, "w_bonepit_24_0_lc.mmh"},
	{0xB1505B07ADD96A08ULL, "6068297_m.fxe"},
	{0xB150AFD02F3BD01FULL, "w_sundermount_204_0.msh"},
	{0xB150B5D02F3BDAC1ULL, "w_sundermount_204_0.mmh"},
	{0xB150BC4190E0107DULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.dura.evt"},
	{0xB150C1D02F3BEF62ULL, "w_sundermount_204_0.mao"},
	{0xB15153A1CAE40426ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_dwn_c.impa.ani"},
	{0xB15165ADFE11595FULL, "6100237_m.fxe"},
	{0xB151FD0C83DCD8F6ULL, "183588713.ogg"},
	{0xB1528C5FDD5D10BBULL, "vxlt_0_lowtown_a2_n_c_fgt_0_0.vlm"},
	{0xB1531141303E03D8ULL, "702129245.ogg"},
	{0xB15320F557D267C7ULL, "6027044_m.fxe"},
	{0xB153662BD348D1B1ULL, "vxw_woundedcoast_167_0_0.vlm"},
	{0xB153698B135866B7ULL, "237913725.ogg"},
	{0xB153871EDFDE31F8ULL, "vxca_caverns2prc_62n_0_1.vlm"},
	{0xB1542A78DF63BE08ULL, "6098324_m.fxe"},
	{0xB154774FCF04C07BULL, "vxht_0_hightown_aut_0_0.vlm"},
	{0xB1548012E617A5AEULL, "hlac_dr_slavers_3217_3z.dds"},
	{0xB154C4B7554513C2ULL, "6070892_m.fxe"},
	{0xB155A8004CE3A2E9ULL, "w_encounter2_22_0.rml"},
	{0xB1561A5D1F585A28ULL, "darktown_male_lowlife_mob6174413.cl"},
	{0xB156254F2FA6E2C3ULL, "lt_0_lowtown_docks_n_100_0.phy"},
	{0xB15691DA563BB57BULL, "lt_warehouse_2_0.mmh"},
	{0xB1569DDA563BC918ULL, "lt_warehouse_2_0.mao"},
	{0xB156AFDA563BE865ULL, "lt_warehouse_2_0.msh"},
	{0xB157187C76937B20ULL, "6142063_m.fxe"},
	{0xB15762809D08F940ULL, "vxht_bartrand_n_5nn_0_0.vlm"},
	{0xB1577720EC1ABAA5ULL, "genip_door_highxt_gate.utp"},
	{0xB1577F85481EE638ULL, "vxlt_gallowstemplar_3jc_0_0.vlm"},
	{0xB157BA923329AE78ULL, "vxlt_shop_ka_0_0.vlm"},
	{0xB157C54F157BA0D3ULL, "vxht_mansion2b_1n1_0_0.vlm"},
	{0xB157E5EC058CBA68ULL, "vxdr_0_deeproads_foe_0_0.vlm"},
	{0xB1581507B2BC1071ULL, "391134804.ogg"},
	{0xB158879C9326EE6DULL, "6156218_m.fxe"},
	{0xB1588FC94E6600A9ULL, "hlac_w_bonepit_327683_0x.dds"},
	{0xB158BB663F6D2273ULL, "w_bonepitdestroyed_a3_185_0.phy"},
	{0xB1590805D4DBAA85ULL, "6060457_f.fxe"},
	{0xB1598A1D8D487C99ULL, "vxw_sundermount_259_0_0.vlm"},
	{0xB1599B8A071CC90EULL, "vxht_bartrand_n_6pu_0_0.vlm"},
	{0xB1599DE500811156ULL, "6129044_m.fxe"},
	{0xB15A2B2CFB8CD243ULL, "w_bonepit_19_0.mao"},
	{0xB15A2F2CFB8CD890ULL, "w_bonepit_19_0.mmh"},
	{0xB15A303C388EA33DULL, "vxw_sundermount_2y1_0_0.vlm"},
	{0xB15A3D2CFB8CF0AAULL, "w_bonepit_19_0.msh"},
	{0xB15A98D905DF1F75ULL, "mer210cr_coterie_thug_2.utc"},
	{0xB15AD3DBE7F403D7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medc_0s.dds"},
	{0xB15AF8219ADDB43AULL, "308803772.ogg"},
	{0xB15B63EBF2DE9641ULL, "6085846_m.fxe"},
	{0xB15BF5FAA19C67E4ULL, "vxca_caverns2prc_8qd_0_0.vlm"},
	{0xB15C217D506831DBULL, "vxlt_gallowstemplar_3p2_0_0.vlm"},
	{0xB15C7F4CF5333807ULL, "373619305.ogg"},
	{0xB15C96289C97AC01ULL, "194188552.ogg"},
	{0xB15D511A13FC2409ULL, "vxlt_0_lowtown_a3_n_c_gn9_0_0.vlm"},
	{0xB15D5E65A7CC3618ULL, "dae120cr_hm_mage_1.utc"},
	{0xB15D611FE935F7A0ULL, "vxlt_undercityencounter_123_0_0.vlm"},
	{0xB15D7DBC5967017FULL, "vxlt_foundry_n_yh_0_0.vlm"},
	{0xB15DD5791A1B8E64ULL, "616490819.ogg"},
	{0xB15DE3AC0E41E6B4ULL, "vxlt_0_lowtown_a2_n_gwl_0_0.vlm"},
	{0xB15DE55EFCD4173CULL, "vxht_brothel_n_6ds_0_0.vlm"},
	{0xB15E29FCDD62E6D1ULL, "hlac_ca_caverns1_mines_24425_2z.dds"},
	{0xB15E3BA1CAEE6597ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_dwn_c.impa.evt"},
	{0xB15E42E16E0529D9ULL, "419478888.ogg"},
	{0xB15ED446AAD1EF13ULL, "zz_qcr_debug6056847.cl"},
	{0xB15F0768456FBC8CULL, "6069867_m.fxe"},
	{0xB15F127C72884578ULL, "6139945_f.fxe"},
	{0xB15F2ED12CED5824ULL, "mag170_thrask6171440.cl"},
	{0xB15F3C65DD63AB18ULL, "lowtown_fem_alien_elf_2_mob6166349.cl"},
	{0xB15F498D6DF8A2B4ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_patternfloor.dds"},
	{0xB15F5BFA01F3D293ULL, "lgt150_nat_were_supporter6110955.cl"},
	{0xB15F75D90CE8EDE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_enter.gad"},
	{0xB15F9C004CEBE2C9ULL, "w_encounter2_22_0.mao"},
	{0xB15FA8004CEBF762ULL, "w_encounter2_22_0.mmh"},
	{0xB15FAA004CEBFA18ULL, "w_encounter2_22_0.msh"},
	{0xB15FDA6BF64B3DA3ULL, "vxlt_warehouse_n_15a_0_0.vlm"},
	{0xB1603C0B61C09D7CULL, "vxca_cave_2be_0_0.vlm"},
	{0xB160A3C0B7B05E7BULL, "zz_followers_banter6159186.cl"},
	{0xB160E5EC023CC0DCULL, "vxlt_blackemporium_il_0_0.vlm"},
	{0xB1613FC84AD25C2AULL, "llac_w_bonepitmining_a2_65539_0.dds"},
	{0xB1616B8E1965DF15ULL, "6083754_m.fxe"},
	{0xB16171144F8C6BA3ULL, "gen00fl_aveline6115485.cl"},
	{0xB161D808C53EEA31ULL, "vxw_woundedcoast_116_0_0.vlm"},
	{0xB161D8E6ADDCA952ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sackclose_sa.dds"},
	{0xB162765458470B60ULL, "hm_nobleman_2.mor"},
	{0xB162C5ED937FA7C7ULL, "art\\characters\\playercharacter\\elffemale\\mt_ant_ef02.phy"},
	{0xB162EB660762E6E9ULL, "832637006.ogg"},
	{0xB163175EF15B7075ULL, "lt_gamlenshouse_n_probe_o3.dds"},
	{0xB1634F4EE093C072ULL, "llac_lt_foundry_n_320_10.dds"},
	{0xB1636A2DF77C16F6ULL, "6173418_m.fxe"},
	{0xB163CB1F6D0C788AULL, "451223597.ogg"},
	{0xB1645EE1ED6E6C8FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbaseb_d.dds"},
	{0xB1646FD001512AD0ULL, "vxlt_gallowscourtyard_n_7ge_0_0.vlm"},
	{0xB164733FAF2A7228ULL, "6114868_m.fxe"},
	{0xB164B2DA56482AE0ULL, "lt_warehouse_2_0.anb"},
	{0xB164D1CB6832DD4EULL, "follower_banter6155622.cl"},
	{0xB16532CEBE23795CULL, "6059059_m.fxe"},
	{0xB1653CA1CAF48620ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_dwn_c.impa.gad"},
	{0xB16578E6A5F669CBULL, "6155622_m.fxe"},
	{0xB16598D90CED94D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_enter.evt"},
	{0xB165D86F8E6E00AEULL, "vxlt_gallowsprison_n_6mf_0_0.vlm"},
	{0xB16696E10FEB852EULL, "6068743_m.fxe"},
	{0xB16753797334794AULL, "6089525_f.fxe"},
	{0xB1676DA96F5476C1ULL, "hlac_lt_0_lowtown_a3_n_c_196607_0y.dds"},
	{0xB167BF1AB890B116ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowbutress.phy"},
	{0xB1684D3A876C6FA1ULL, "669363419.ogg"},
	{0xB16874C9B47A0AB9ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4so_basic_4.mao"},
	{0xB1688E9FE70D8462ULL, "vxlt_0_lowtown_a2_n_c_cmn_0_0.vlm"},
	{0xB168D1B99A445E18ULL, "character_character_p175.psh.fxo.dx11"},
	{0xB168D6C634608EABULL, "6100291_m.fxe"},
	{0xB168F750C3B5F29BULL, "zz_dog_debug6127945.cl"},
	{0xB16A134112297A29ULL, "1004380162.ogg"},
	{0xB16AA3602E2EA766ULL, "6054968_m.fxe"},
	{0xB16C573BEECA2FBCULL, "vxht_brothel_n_6ap_0_0.vlm"},
	{0xB16C7E53CF3B194BULL, "69439370.ogg"},
	{0xB16D1F5169D59AFEULL, "6043933_m.fxe"},
	{0xB16D3D46C18EF419ULL, "gen_im_trs_tmp_one_lsw_r0.uti"},
	{0xB16DCCAF039324D6ULL, "vxw_bonepit_26d_0_0.vlm"},
	{0xB16E14153B0FC1FBULL, "art\\characters\\creatures\\mubari\\animation\\kd_exit.evt"},
	{0xB16E35A5D12D29F2ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.cess.gad"},
	{0xB16E511BBD78C803ULL, "hlac_dr_primevalentrance_1726_9z.dds"},
	{0xB16ED3348ECE526BULL, "6106392_m.fxe"},
	{0xB16F1CDEF1E17828ULL, "6132791_f.fxe"},
	{0xB16F1FB4D2AC1755ULL, "gen00fl_aveline6105442.cl"},
	{0xB16F7E8762A38788ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_5x.dds"},
	{0xB16FAEE477A0AD88ULL, "379246304.ogg"},
	{0xB16FB7E1C8C657D2ULL, "vxht_mansion2b_1a5_0_0.vlm"},
	{0xB1700CFA1011E54DULL, "1067113802.ogg"},
	{0xB171A81AB898DEB7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowbutress.msh"},
	{0xB171BA1AB898FD1AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowbutress.mao"},
	{0xB171BE1AB8990429ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galowbutress.mmh"},
	{0xB1727C9AB00DEE98ULL, "1050323624.ogg"},
	{0xB172A3C5B3F56E7FULL, "gen00fl_bethany6102783.cl"},
	{0xB173874C34A7B525ULL, "380153463.ogg"},
	{0xB173A0D90CF9DFC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_pray_enter.ani"},
	{0xB173DC8F9225EA3DULL, "dae000_player_generic_find6179060.cl"},
	{0xB17417CBEE1327A5ULL, "gen_im_npc_arm_hum_cht_gwdb.uti"},
	{0xB17427545D93BA2CULL, "mag200_feynriel6034447.cl"},
	{0xB174420F35D16427ULL, "ss_karras.bnk"},
	{0xB174A2DA4C01617AULL, "vxlt_0_lowtown_a2_n_c_hum_0_0.vlm"},
	{0xB1750D153B15D43CULL, "art\\characters\\creatures\\mubari\\animation\\kd_exit.gad"},
	{0xB17538A5D1334D29ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.cess.evt"},
	{0xB1763BDF9BDC9648ULL, "gen00fl_aveline6132850.cl"},
	{0xB1771825C1206781ULL, "vxca_caverns1_mines_j1z_0_2.vlm"},
	{0xB17735E939CE0475ULL, "mag311_emile_de_launcet6043462.cl"},
	{0xB1774F49A6593D46ULL, "ser121_jethann6175420.cl"},
	{0xB17766F04FD1C1C1ULL, "477412369.ogg"},
	{0xB178063F7092CEC7ULL, "vxht_chantry_n_359_0_0.vlm"},
	{0xB178FFD1EC66168FULL, "6031717_m.fxe"},
	{0xB179E33C9234B9ECULL, "mrl000_marethari6129272.cl"},
	{0xB179ED5AF6FB0562ULL, "vxlt_0_lowtown_a2_n_c_fe7_0_0.vlm"},
	{0xB179FA30FB3FE1BBULL, "llac_ht_0_hightown_-3_0.dds"},
	{0xB17A5360F32676BAULL, "vxlt_darktown_2s3_0_0.vlm"},
	{0xB17A780D15B6BDA5ULL, "1810935.ogg"},
	{0xB17A80B20D7286BAULL, "vxlt_0_lowtown_docks_a3_arr_0_0.vlm"},
	{0xB17A900B474A38F1ULL, "118740392.ogg"},
	{0xB17AC918E0C0483CULL, "vxht_brothel_n_6jq_0_0.vlm"},
	{0xB17B5432C1CC8CB5ULL, "vxht_0_hightown_a2_av3_0_0.vlm"},
	{0xB17BB5E6F8698B87ULL, "gen00fl_sebastian6180005.cl"},
	{0xB17C0C153B1BF13AULL, "art\\characters\\creatures\\mubari\\animation\\kd_exit.ani"},
	{0xB17C3E8BF5893D56ULL, "vxw_bonepit_21e_0_0.vlm"},
	{0xB17C65464B2614A2ULL, "6066529_m.fxe"},
	{0xB17C72AFF15B764CULL, "follower_banter6170203.cl"},
	{0xB17C907160E8CAFFULL, "788115045.ogg"},
	{0xB17C9D707D761113ULL, "716083167.ogg"},
	{0xB17CA7AC8645DCFEULL, "follower_banter6169791.cl"},
	{0xB17CAA3D58A5FA58ULL, "plt_cod_aow_playerstates1.plo"},
	{0xB17CB6D02F61173EULL, "w_sundermount_204_0.phy"},
	{0xB17CD3C4757DFB0AULL, "dae000_orsino6085419.cl"},
	{0xB17D39FDD1CF8E09ULL, "ser221_party_comments6166205.cl"},
	{0xB17D55E2212C4336ULL, "w_encounter3_228_0.mao"},
	{0xB17D61E2212C57D5ULL, "w_encounter3_228_0.mmh"},
	{0xB17D63E2212C5B4BULL, "w_encounter3_228_0.msh"},
	{0xB17E1950EDAD3D2DULL, "hlac_ht_chantry_2534_42x.dds"},
	{0xB17EFEC0D57A70CAULL, "145436256.ogg"},
	{0xB17F408A46CA1EFDULL, "llac_lt_0_lowtown_docks_n_-131075_2.dds"},
	{0xB17F6314F03F893EULL, "453308854.ogg"},
	{0xB17FB2F6FE9D9B07ULL, "ht_0_hightown_n_mmap.dds"},
	{0xB17FC14C9DC1BD47ULL, "al_mage_s02.gda"},
	{0xB18014EE51BAC2E1ULL, "vxca_caverns2prc_aa0_0_0.vlm"},
	{0xB180E1265FE0D833ULL, "gen00fl_anders6096527.cl"},
	{0xB1815E97844A04ECULL, "vxdr_primevaltomb_6vp_0_0.vlm"},
	{0xB18250A6212DD891ULL, "vxht_bartrand_n_68n_0_0.vlm"},
	{0xB1825B0D0AA6A0A6ULL, "vxlt_0_lowtown_n_byu_0_0.vlm"},
	{0xB1827696A9AF217CULL, "6051024_m.fxe"},
	{0xB182911CEA811A71ULL, "110970894.ogg"},
	{0xB18320A5D13F6198ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.cess.ani"},
	{0xB1833B7C9764CDB2ULL, "6129941_m.fxe"},
	{0xB18439ADD7E886E2ULL, "952808824.ogg"},
	{0xB1859C4C6B71C13BULL, "lowtown_fem_alienage1_sta6162604.cl"},
	{0xB185C937501631B4ULL, "follower_barks6153236.cl"},
	{0xB1862A212F38873CULL, "6035927_m.fxe"},
	{0xB1864938F81D318AULL, "framebuffereffects_diffusiondof_p4.psh.fxo.dx11"},
	{0xB186C058800EAE6FULL, "dae100_viscount6118525.cl"},
	{0xB187A563A0A245FDULL, "6075392_f.fxe"},
	{0xB188036148C38602ULL, "zz_followers_banter6159570.cl"},
	{0xB188183BEDDF2DE9ULL, "6139183_m.fxe"},
	{0xB1883AF813A313E6ULL, "vxlt_0_lowtown_docks_a3_d0y_0_0.vlm"},
	{0xB188BA4A5485ADDCULL, "674199086.ogg"},
	{0xB188D54B2AC3DAA8ULL, "6060161_m.fxe"},
	{0xB188DDD5794C8F76ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.dura.evt"},
	{0xB188F499F2A556F7ULL, "6082680_f.fxe"},
	{0xB18931F7AC001B36ULL, "lgt260_way_blood_mage6148363.cl"},
	{0xB189C60FB3C2A535ULL, "vxht_0_hightown_a2_aq2_0_0.vlm"},
	{0xB18A04C1C3ED89BCULL, "49656103.ogg"},
	{0xB18A1244F19E1812ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_bookcase_n.dds"},
	{0xB18A4926D43DE828ULL, "617800561.ogg"},
	{0xB18AC889676B13BEULL, "mer221_sabin6047671.cl"},
	{0xB18B6BAE98A7BA57ULL, "character_character_p107.psh.fxo"},
	{0xB18B7FC670F35841ULL, "dae000_hate_enemies6167540.cl"},
	{0xB18BFEAA51F5C463ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cart.msh"},
	{0xB18C0CAA51F5DBBDULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cart.mmh"},
	{0xB18C10AA51F5E2CEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_cart.mao"},
	{0xB18C491FF3F5BF83ULL, "vxlt_warehouse_n_hu_0_0.vlm"},
	{0xB18C63479BB6B1F2ULL, "pro000_friedrich.cnv"},
	{0xB18CA681C8724C5FULL, "vxw_sundermount_34y_0_0.vlm"},
	{0xB18CE8637083E48FULL, "llac_w_encounter2_131075_0.dds"},
	{0xB18D0398E9557E75ULL, "6122811_m.fxe"},
	{0xB18D48D0D92963E2ULL, "gen00fl_sebastian6174034.cl"},
	{0xB18D759DE3F4BD3CULL, "vxw_sundermount_20c_0_0.vlm"},
	{0xB18DA06919C9B03BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_008a_0s.dds"},
	{0xB18DBC772E789F87ULL, "ts_books_03.gda"},
	{0xB18ED4D57950EB01ULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.dura.gad"},
	{0xB18F85479BB9214AULL, "pro000_friedrich.bnk"},
	{0xB18F86F3C8A05328ULL, "406581107.ogg"},
	{0xB1904F25955248A4ULL, "6207388_m.fxe"},
	{0xB19059205BD8DB74ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_diagonal_3.mao"},
	{0xB19137CD4BC74012ULL, "6099155_f.fxe"},
	{0xB191D0E9C90A7899ULL, "dae000_pc_home_ambients6208577.cl"},
	{0xB1920DB46617B496ULL, "zz_qcr_debug6052127.cl"},
	{0xB19237FF1A481009ULL, "353570590.ogg"},
	{0xB1924D7E8CFA802CULL, "hlac_lt_0_lowtown_a2_n_65534_5x.dds"},
	{0xB19289D9D6FEDDADULL, "678312611.ogg"},
	{0xB19293B2927B3775ULL, "vxca_caverns1_mines_jfa_0_0.vlm"},
	{0xB192BFE3AFC259F7ULL, "105286106.ogg"},
	{0xB1934CC5D6752850ULL, "static_billboard_p0.vsh.fxo.dx11"},
	{0xB193783D8FF6AF35ULL, "6071870_m.fxe"},
	{0xB193B5388DC6D200ULL, "6150514_m.fxe"},
	{0xB193C05001881BD5ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabv3.evt"},
	{0xB193C9988B9D8497ULL, "hlac_lt_gallowscourtyard_n_-131071_0z.dds"},
	{0xB193D85A51F733F7ULL, "vxw_woundedcoast_5i_0_0.vlm"},
	{0xB1941E9E26A34678ULL, "lt_0_lowtown_a2_28_0_ll.mmh"},
	{0xB1942A4BD699D9EEULL, "isa311_isabela6092258.cl"},
	{0xB1942F2CFBBE5153ULL, "w_bonepit_19_0.rml"},
	{0xB194EE3BB9732BB8ULL, "vxlt_0_lowtown_a2_n_c_gos_0_1.vlm"},
	{0xB194F905E044BFD1ULL, "rdr100pt_keepers_note.ncs"},
	{0xB1958790A2D03E8CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\bethany_staff_idle.ani"},
	{0xB195A90B6B41DAA0ULL, "vxw_woundedcoast_1pm_0_1.vlm"},
	{0xB195D5D579570BAFULL, "art\\vfx\\combat\\vfx_r_var_kback_proj_p.dura.ani"},
	{0xB195E296D9759B1EULL, "785185475.ogg"},
	{0xB197AE56531A5F37ULL, "346013770.ogg"},
	{0xB197AEB706449D57ULL, "w_sundermount_253_0.msh"},
	{0xB197C0B70644BBBAULL, "w_sundermount_253_0.mao"},
	{0xB197C4B70644C2C9ULL, "w_sundermount_253_0.mmh"},
	{0xB197EE2670E0550CULL, "6092202_f.fxe"},
	{0xB1981E1507FBD3A9ULL, "hf_ibethany_h2s3.mor"},
	{0xB199299A7024E60BULL, "vxht_mansion2b_2r2_0_0.vlm"},
	{0xB199C85DB6BFFC7DULL, "276242488.ogg"},
	{0xB199E3896632681AULL, "zz_facefx_test6088481.cl"},
	{0xB19A0B97DDDD266CULL, "7769511.ogg"},
	{0xB19A292996F0A36BULL, "mcr370_varric6098334.cl"},
	{0xB19A8B778E65C6C4ULL, "mcr370_guardsman6086982.cl"},
	{0xB19ABAFCE5EBD803ULL, "vxlt_warehouse_n_ot_0_0.vlm"},
	{0xB19ACE685DF7D166ULL, "6142211_f.fxe"},
	{0xB19B0D789EA8A152ULL, "vxht_mansion2b_1r0_0_0.vlm"},
	{0xB19B185EBA6864DFULL, "vxw_sundermount_33x_0_0.vlm"},
	{0xB19B31AFAC259102ULL, "zz_dae_debug6050745.cl"},
	{0xB19B9150C2615F98ULL, "gen_im_trs_dal_blt_r2.uti"},
	{0xB19BB99E20B6C993ULL, "298030878.ogg"},
	{0xB19BD39FC0245A4AULL, "t3_dog_reg.tnt"},
	{0xB19BF17EF2202DBFULL, "lgt250_way_dalish_conflict6147170.cl"},
	{0xB19C01DACC06E9DAULL, "hlac_ca_caverns1_mines_24424_0x.dds"},
	{0xB19C05246A6F4D2FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandleft_1.mao"},
	{0xB19C9490A2D673CEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\bethany_staff_idle.gad"},
	{0xB19C9E32B972F4DEULL, "dae000_arishok6075053.cl"},
	{0xB19CA514CD1DA972ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_1z.dds"},
	{0xB19CB0A088D33AFBULL, "vxht_0_hightown_alk_0_0.vlm"},
	{0xB19CD1FB6BE8E93DULL, "6050088_m.fxe"},
	{0xB19D16CFAA8125F9ULL, "107000344.ogg"},
	{0xB19D432303A83653ULL, "284617946.ogg"},
	{0xB19D824240CA8191ULL, "430419833.ogg"},
	{0xB19DAAFA9C4D1A96ULL, "6103732_m.fxe"},
	{0xB19DB2DA5677EFB0ULL, "lt_warehouse_2_0.phy"},
	{0xB19E1567EC1453CAULL, "603488689.ogg"},
	{0xB19E1D48C7B99D96ULL, "vfx_grad01.dds"},
	{0xB19E3F395D2EFED1ULL, "gen_im_wep_war_two_act1_05.uti"},
	{0xB19EB84E2A72C4E0ULL, "47093939.ogg"},
	{0xB19F9A7E5075F0C5ULL, "vxlt_0_lowtown_a2_n_c_huf_0_0.vlm"},
	{0xB19FD3997960B643ULL, "6052758_m.fxe"},
	{0xB19FE5D74EA73306ULL, "vxlt_gallowstemplar_3s2_0_0.vlm"},
	{0xB19FEB97184D592CULL, "vxlt_hangedman_n_174_0_0.vlm"},
	{0xB19FFE9FCF4A1B53ULL, "vxht_mansion2b_1u4_0_0.vlm"},
	{0xB1A042A458A988DDULL, "49192833.ogg"},
	{0xB1A0918D2A992C2EULL, "lt_0_lowtown_35_0_ll.mmh"},
	{0xB1A0D7C20C8603F5ULL, "lt_0_lowtown_a3_n_c_41_0.rml"},
	{0xB1A1CA18411BD3D9ULL, "285378462.ogg"},
	{0xB1A1DEE87B90DFDBULL, "vxht_0_hightown_a2_au8_0_0.vlm"},
	{0xB1A24A3743ED4C77ULL, "vxw_woundedcoast_2h_0_0.vlm"},
	{0xB1A2661B0C1937C5ULL, "vxlt_hangedman_n_2xn_0_1.vlm"},
	{0xB1A2B6C4AD9AD770ULL, "llac_lt_0_lowtown_docks_-4_5.dds"},
	{0xB1A2F5DD937D9105ULL, "vxlt_shop_af_0_0.vlm"},
	{0xB1A2FA8CB838F349ULL, "6034305_m.fxe"},
	{0xB1A3CB719735EC0DULL, "mag100_cullen6082270.cl"},
	{0xB1A468D75B4A0697ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_act_c.anb"},
	{0xB1A47735D4E0A46EULL, "ser221_gascard6111802.cl"},
	{0xB1A4778DF2B25498ULL, "llac_lt_0_lowtown_docks_a3_-196608_0.dds"},
	{0xB1A491DA567DD608ULL, "lt_warehouse_2_0.rml"},
	{0xB1A4EA262C223345ULL, "vxlt_hangedman_n_20w_0_0.vlm"},
	{0xB1A5115118ECA178ULL, "vxlt_gallowscourtyard_74c_0_0.vlm"},
	{0xB1A56C8D5D36A866ULL, "410068025.ogg"},
	{0xB1A5A3A8705FF45AULL, "1065081937.ogg"},
	{0xB1A5D025E4D2E5EBULL, "hlac_ht_0_hightown_a3_65536_0z.dds"},
	{0xB1A645B2E95C03D8ULL, "184224401.ogg"},
	{0xB1A657A1AAE5B988ULL, "art\\vfx\\materials\\vfx_cyc_dark.mao"},
	{0xB1A66FF04328F3C7ULL, "hlac_lt_0_lowtown_docks_a3_131068_0y.dds"},
	{0xB1A687103D13CA34ULL, "dae300_cullen6140587.cl"},
	{0xB1A6D81A18D12D4DULL, "vxw_encounter2_camp_s5_0_1.vlm"},
	{0xB1A71939DCFB379CULL, "6000963_m.fxe"},
	{0xB1A733E8F32CA4B5ULL, "vxw_encounter2_qz_0_0.vlm"},
	{0xB1A75AF37C444416ULL, "6125314_m.fxe"},
	{0xB1A79B77621AFE8BULL, "vxht_mansion2b_2o5_0_0.vlm"},
	{0xB1A7A57B2A239063ULL, "245412807.ogg"},
	{0xB1A837AAD99085C2ULL, "gen00fl_fenris6106663.cl"},
	{0xB1A88AA70D5EABA9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\til_lowtown_main_sa.dds"},
	{0xB1A88CAFC3C33035ULL, "308637828.ogg"},
	{0xB1A91F3CBC0B3B5CULL, "vxlt_blackemporium_va_0_0.vlm"},
	{0xB1A92CD9D7E1F083ULL, "vxlt_warehouse_n_rw_0_0.vlm"},
	{0xB1A943254FE12A63ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_ballistabolt01_0_sa.dds"},
	{0xB1A97F55909EB9D2ULL, "vxht_mansion2b_1u1_0_0.vlm"},
	{0xB1A9853626065536ULL, "136412146.ogg"},
	{0xB1AA7182D154192EULL, "vxht_mansion_n_4cy_0_0.vlm"},
	{0xB1AACC4E3DC599B2ULL, "vxca_cave_17s_0_0.vlm"},
	{0xB1AB48331FCC6752ULL, "mcr370_meredith6084430.cl"},
	{0xB1AB491913963BA1ULL, "hlac_ht_0_hightown_131070_0x.dds"},
	{0xB1AB5BF407230CC6ULL, "179188703.ogg"},
	{0xB1ABA3FE19F760D5ULL, "9376761.ogg"},
	{0xB1ABD55169002231ULL, "vxlt_backalley_g_1wl_0_1.vlm"},
	{0xB1AC13CA73A73E8AULL, "sib111_mother6098833.cl"},
	{0xB1AC6B87F38922D2ULL, "200290591.ogg"},
	{0xB1ACDD7A076BE573ULL, "vxht_mansion2_27m_0_0.vlm"},
	{0xB1ACE725613027E9ULL, "6113837_m.fxe"},
	{0xB1ADB8073D05BD21ULL, "dae000_bodhan6129097.cl"},
	{0xB1AE0C5B426C0945ULL, "vxlt_0_lowtown_a2_n_c_hpe_0_0.vlm"},
	{0xB1AE5AAF9A6C29C0ULL, "vxlt_0_lowtown_a2_n_gv7_0_0.vlm"},
	{0xB1AF58A61185AFF9ULL, "6086177_m.fxe"},
	{0xB1AF8205D70BEB11ULL, "vrc321_spooky_scenes6215549.cl"},
	{0xB1AFCAEA768DE01FULL, "hlac_lt_0_lowtown_a2_n_-131072_8x.dds"},
	{0xB1B08B5C62299401ULL, "and330ip_at_sewers.ncs"},
	{0xB1B0A722A1BC68A1ULL, "llac_lt_0_lowtown_a2_n_2_1.dds"},
	{0xB1B0C01F30CFC5E8ULL, "6085241_m.fxe"},
	{0xB1B0D92AD60B948EULL, "art\\vfx\\combat\\vfx_g_dual_mode_c.impa.ani"},
	{0xB1B0DAE4C904C6DBULL, "vxw_encounter2_jq_0_0.vlm"},
	{0xB1B12300E1D26CD1ULL, "gen00fl_sebastian6174664.cl"},
	{0xB1B12AD6B3544A3DULL, "10594947.ogg"},
	{0xB1B167BA8573A985ULL, "vxlt_shop_de_0_0.vlm"},
	{0xB1B19140F8F293F5ULL, "fen210_orana6109711.cl"},
	{0xB1B275392CD81C9AULL, "lt_0_lowtown_docks_n_128_0.mmh"},
	{0xB1B279392CD823D1ULL, "lt_0_lowtown_docks_n_128_0.mao"},
	{0xB1B287392CD83BE0ULL, "lt_0_lowtown_docks_n_128_0.msh"},
	{0xB1B2CF05EABAC811ULL, "ser211ip_note_scribbles1.utp"},
	{0xB1B2FB3EB3D13CDDULL, "18469287.ogg"},
	{0xB1B36C4AF6AE1BBEULL, "6030647_m.fxe"},
	{0xB1B4C72846D47F2AULL, "896266760.ogg"},
	{0xB1B5313AA0A1C65FULL, "6071071_f.fxe"},
	{0xB1B53F97E4B7A878ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdga_0n.dds"},
	{0xB1B5756B73F328A0ULL, "hlac_lt_gallowstemplar_f_2390_1z.dds"},
	{0xB1B5C04C201B6A03ULL, "lt_0_lowtown_docks_17_0_ll.mmh"},
	{0xB1B6914DDDE40251ULL, "676701312.ogg"},
	{0xB1B6CE372F984324ULL, "6071721_m.fxe"},
	{0xB1B74DAA71C6304FULL, "89832562.ogg"},
	{0xB1B7565653EE65DFULL, "993385026.ogg"},
	{0xB1B79EB6C9D80903ULL, "vxlt_warehouse_n_yv_0_0.vlm"},
	{0xB1B88D52161B2D58ULL, "one100dg_vanard6048249.cl"},
	{0xB1B8CBC41A60A094ULL, "follower_banter6136461.cl"},
	{0xB1B8E35FC34A31AEULL, "vxht_mansion_n_4fz_0_0.vlm"},
	{0xB1B93B6D8A8346C4ULL, "694726949.ogg"},
	{0xB1B94EB453E5A37DULL, "follower_banter6141491.cl"},
	{0xB1B9B25D04C33525ULL, "499626553.ogg"},
	{0xB1B9C8DD698E49ECULL, "386163892.ogg"},
	{0xB1B9D62D8ADEFACAULL, "one111ip_ggt_mallet.ncs"},
	{0xB1BA558A438AF47BULL, "vxdr_primevalentrance_1ja_0_1.vlm"},
	{0xB1BAB4663FBFE0E8ULL, "w_bonepitdestroyed_a3_185_0.mmh"},
	{0xB1BAC0663FBFF48BULL, "w_bonepitdestroyed_a3_185_0.mao"},
	{0xB1BACDDE3C2841BCULL, "6072346_m.fxe"},
	{0xB1BAD2663FC01312ULL, "w_bonepitdestroyed_a3_185_0.msh"},
	{0xB1BB159EAC351B9BULL, "vxlt_0_lowtown_docks_a3_n_d1l_0_0.vlm"},
	{0xB1BB2B4C2D4D320FULL, "437663758.ogg"},
	{0xB1BB4E6ABC427A49ULL, "vxlt_0_lowtown_a2_gw6_0_0.vlm"},
	{0xB1BB4F56F961FDF3ULL, "vxht_mansion2_20l_0_0.vlm"},
	{0xB1BD538FF83B817CULL, "6094948_m.fxe"},
	{0xB1BDD12AD61611DFULL, "art\\vfx\\combat\\vfx_g_dual_mode_c.impa.evt"},
	{0xB1BEB019148638ADULL, "gen00fl_fenris6126736.cl"},
	{0xB1BF1165D21F4A81ULL, "6135143_m.fxe"},
	{0xB1BF682CA6F9E082ULL, "582195145.ogg"},
	{0xB1BFB416330CE225ULL, "472900768.ogg"},
	{0xB1BFBE9A873A822CULL, "460055646.ogg"},
	{0xB1C06C36188C6908ULL, "hlac_ht_0_hightown_n_-131073_0x.dds"},
	{0xB1C09295EAD3E607ULL, "223687887.ogg"},
	{0xB1C150B2C4E8BFADULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_flsa_0e.dds"},
	{0xB1C15131FDC8903BULL, "1002033352.ogg"},
	{0xB1C190905F0EEA78ULL, "character_character_p34.psh.fxo.dx11.assembly"},
	{0xB1C1C427787EC771ULL, "aip_rubblegolema_base.gda"},
	{0xB1C1E5F9E2424033ULL, "vxht_mansion2b_341_0_0.vlm"},
	{0xB1C2B8D6699418E5ULL, "tbe100im_mystic_chest.uti"},
	{0xB1C2E2DF2203394CULL, "vxdr_0_deeproads_cxi_0_0.vlm"},
	{0xB1C38FD93793F30AULL, "599130265.ogg"},
	{0xB1C393E6625B369DULL, "qun131_taskmaster6039514.cl"},
	{0xB1C3F6DC90BF0CA9ULL, "gen00fl_merrill6144978.cl"},
	{0xB1C47F3146072F8BULL, "vxht_mansion2b_2e7_0_0.vlm"},
	{0xB1C4B22AD61BFBC8ULL, "art\\vfx\\combat\\vfx_g_dual_mode_c.impa.gad"},
	{0xB1C4BBC7FAD6F696ULL, "571007305.ogg"},
	{0xB1C4D36993BCE8E5ULL, "vxlt_0_lowtown_docks_n_cox_0_0.vlm"},
	{0xB1C529ED566F938AULL, "hlac_lt_0_lowtown_docks_-131076_1z.dds"},
	{0xB1C57FA039791938ULL, "gen00fl_varric6191294.cl"},
	{0xB1C58B9095FFBB00ULL, "ser221_gascard6133747.cl"},
	{0xB1C61A18E8CB07C1ULL, "gencr_corpse_arc_onground.utc"},
	{0xB1C69384CD160992ULL, "mag291_ambient_1.cnv"},
	{0xB1C6EDBC80CC1332ULL, "6027108_m.fxe"},
	{0xB1C7030331D66B33ULL, "6127069_m.fxe"},
	{0xB1C8028771632406ULL, "1072507129.ogg"},
	{0xB1C805B017157752ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_001a_0d.dds"},
	{0xB1C8BAA6CECE937CULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_2y.dds"},
	{0xB1C8CF59E6B166EAULL, "114800497.ogg"},
	{0xB1C8D16C4D0D72A1ULL, "follower_banter6165620.cl"},
	{0xB1C90D1B945E5662ULL, "vxlt_backalley_g_1kf_0_0.vlm"},
	{0xB1C93E347CAC8A6DULL, "gen_im_trs_qun_horn_03.uti"},
	{0xB1C95AE192D72770ULL, "6089685_m.fxe"},
	{0xB1C99EFBAF44C3EAULL, "llac_w_bonepitmining_a2_458755_0.dds"},
	{0xB1C9C047AE3892C9ULL, "vxlt_0_lowtown_a2_gj5_0_0.vlm"},
	{0xB1CA0B772EBF3C80ULL, "6020068_m.fxe"},
	{0xB1CA34A9FA174B9EULL, "hlac_lt_0_lowtown_a2_n_1_10x.dds"},
	{0xB1CA6427BD372506ULL, "hla_ht_0_hightown_n_131073_0z.dds"},
	{0xB1CA731539B32F6AULL, "zz_qun_debug6028241.cl"},
	{0xB1CAB584CD1A2CEAULL, "mag291_ambient_1.bnk"},
	{0xB1CB17351768A5AEULL, "118260112.ogg"},
	{0xB1CB22AE518CF388ULL, "lgt220_shp_mag_guard6149492.cl"},
	{0xB1CB55B7B76F78CDULL, "llac_ht_mansion2b_818_47.dds"},
	{0xB1CB894C63853C81ULL, "vxlt_blackemporium_v0_0_0.vlm"},
	{0xB1CBC70EAAC9A9FEULL, "vxht_0_hightown_a2_at6_0_0.vlm"},
	{0xB1CBDD591F862AC4ULL, "89684045.ogg"},
	{0xB1CBFDF8195BD5AEULL, "dae000_hate_follower_react6167488.cl"},
	{0xB1CCAE41E2DFCD36ULL, "6132934_m.fxe"},
	{0xB1CCE4961D8EB322ULL, "llac_lt_0_lowtown_a2_131072_1.dds"},
	{0xB1CCF9E94604E756ULL, "6072142_m.fxe"},
	{0xB1CD07DBE69B698FULL, "vxlt_0_lowtown_docks_a3_n_cyw_0_0.vlm"},
	{0xB1CD65D75B6D0698ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_act_c.mmh"},
	{0xB1CD70CDA3136FC2ULL, "368251476.ogg"},
	{0xB1CDAA10E507B286ULL, "mcr370_templar16087633.cl"},
	{0xB1CDBE9EACF0F7DBULL, "vxw_encounter2_ps_0_0.vlm"},
	{0xB1CDC6A345D9CAAAULL, "489908134.ogg"},
	{0xB1CDD296DD0A08FAULL, "6055167_runoff.cut"},
	{0xB1CDF58403752A93ULL, "6094971_f.fxe"},
	{0xB1CE6F7E0D5D1A1AULL, "vxw_woundedcoast_173_0_0.vlm"},
	{0xB1CEBC5C5B3A8B5FULL, "6001003_f.fxe"},
	{0xB1CED693565774F0ULL, "hlac_w_blightlands_262140_0x.dds"},
	{0xB1CEF6E59FCA2374ULL, "cd_mag110_vincento.bnk"},
	{0xB1CF2709CA9C1535ULL, "6068519_m.fxe"},
	{0xB1CF7459447E6F32ULL, "6080452_f.fxe"},
	{0xB1D03C002E816182ULL, "6127034_f.fxe"},
	{0xB1D0412BD54A959AULL, "art\\characters\\playercharacter\\humanfemale\\hlf_wmda.phy"},
	{0xB1D0B5E0140A2E56ULL, "6129233_m.fxe"},
	{0xB1D108990ED6241DULL, "mag291_demon_marethari6029054.cl"},
	{0xB1D154BC13F951CCULL, "vxdr_0_deeproads_cun_0_0.vlm"},
	{0xB1D20B55F619F842ULL, "w_bonepit_18_0.rml"},
	{0xB1D23295F5A8C5AFULL, "vxlt_gallowstemplar_f_56e_0_0.vlm"},
	{0xB1D275EDC8FA57CDULL, "vxlt_0_lowtown_bym_0_0.vlm"},
	{0xB1D32633F34ADF39ULL, "168757740.ogg"},
	{0xB1D368E295545305ULL, "vxlt_gallowscourtyard_74z_0_0.vlm"},
	{0xB1D3BEA6083FE3B8ULL, "vxht_0_hightown_aq7_0_0.vlm"},
	{0xB1D3D39CC01F5F84ULL, "vxht_bartrand_n_5wa_0_0.vlm"},
	{0xB1D3D665A487F80DULL, "922339203.ogg"},
	{0xB1D3D9F3DEED72D0ULL, "vxlt_warehouse_cj_0_0.vlm"},
	{0xB1D45BBB60453124ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml_b.anb"},
	{0xB1D485A4B5474120ULL, "follower_banter6163290.cl"},
	{0xB1D4B22F07F5FCA2ULL, "6071051_m.fxe"},
	{0xB1D4DD2B4E2936E3ULL, "one280_emrys6104308.cl"},
	{0xB1D4DD6987E71737ULL, "6151033_m.fxe"},
	{0xB1D50176E560A12EULL, "794035107.ogg"},
	{0xB1D53AEB327F4459ULL, "6097430_m.fxe"},
	{0xB1D55D7BD4947D2BULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrock_d.dds"},
	{0xB1D568FC02EECCC4ULL, "6126552_f.fxe"},
	{0xB1D58705142E35E6ULL, "6173663_m.fxe"},
	{0xB1D656516DB9A925ULL, "150739169.ogg"},
	{0xB1D674D40E970385ULL, "zz_followers_banter6157414.cl"},
	{0xB1D67BC237F81524ULL, "282116729.ogg"},
	{0xB1D6967144CB4523ULL, "8732960.ogg"},
	{0xB1D69DBAE61DAA6FULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_2x.dds"},
	{0xB1D6BE1880345372ULL, "388781700.ogg"},
	{0xB1D6ECFB0A2C5684ULL, "dae300_alistair_king6134940.cl"},
	{0xB1D6FC83163B7C7EULL, "vxca_caverns2_9de_0_0.vlm"},
	{0xB1D712A8D666675BULL, "dae300_cullen6140663.cl"},
	{0xB1D76381527240B4ULL, "vxlt_0_lowtown_docks_n_c4l_0_0.vlm"},
	{0xB1D807A6D5B73465ULL, "hlac_w_encounter3_-196605_0z.dds"},
	{0xB1D815AAB99384F6ULL, "hlac_ca_caverns1_24425_12y.dds"},
	{0xB1D81D8996869D11ULL, "one200dg_solivitus6060337.cl"},
	{0xB1D8A27DE1DB6E0DULL, "vxht_mansion2_2v0_0_0.vlm"},
	{0xB1D8E18CB0FE710DULL, "6115574_m.fxe"},
	{0xB1D8F20DB7C9E3ACULL, "l_dae210ar_docks.bnk"},
	{0xB1D95258C95DE042ULL, "llac_ht_0_hightown_n_-3_0.dds"},
	{0xB1D9BE5C3E3B7A5FULL, "mer221_sabin6080112.cl"},
	{0xB1DA3F619EA65036ULL, "vxlt_0_lowtown_docks_n_cuq_0_0.vlm"},
	{0xB1DAF0925E12F402ULL, "6089535_m.fxe"},
	{0xB1DB2B3F2DA5DC36ULL, "vxht_mansion2b_1ko_0_0.vlm"},
	{0xB1DBB44E455E8164ULL, "hlac_lt_0_lowtown_65535_5x.dds"},
	{0xB1DBD89C5DC1078EULL, "zz_dre_debug6037565.cl"},
	{0xB1DBF14607E7DF86ULL, "t1_mue_bl3.tnt"},
	{0xB1DBFBAB5A88FE39ULL, "plt_isa000pt_main.nss"},
	{0xB1DC924DA8144664ULL, "mag200_feynriel6023898.cl"},
	{0xB1DD08554A6D8307ULL, "6165342_m.fxe"},
	{0xB1DD77BE14EED4C3ULL, "hlac_ca_cavern1_haunted_24424_12z.dds"},
	{0xB1DDA024405ECE31ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_basic_1.mao"},
	{0xB1DDC8EB129879CFULL, "wm_ico_shop.dds"},
	{0xB1DDE90751343423ULL, "309035584.ogg"},
	{0xB1DDEAF975DE2D44ULL, "mag360cr_hostage_anders.utc"},
	{0xB1DE20F15F08D2FDULL, "pro000st_athenril.stg"},
	{0xB1DE3650E64C9EB5ULL, "vxht_brothel_n_6gk_0_0.vlm"},
	{0xB1DE4D66A47EF3E8ULL, "774451307.ogg"},
	{0xB1DE4F93A0CB1B62ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.cess.gad"},
	{0xB1DEB91B9856EFA7ULL, "hlac_lt_0_lowtown_a2_n_65535_5x.dds"},
	{0xB1DEF18348A4C6CFULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\_textures\\vst_gallows_sa.dds"},
	{0xB1DEF2C0FFF5C20FULL, "247623803.ogg"},
	{0xB1DF61E46ADCD420ULL, "vxlt_undercityencounter_1g8_0_0.vlm"},
	{0xB1DFBF2E3AF95ACBULL, "6092187_wander1.cut"},
	{0xB1DFD0FA13DD3565ULL, "6063643_f.fxe"},
	{0xB1DFFDFC97C403AEULL, "1002410808.ogg"},
	{0xB1E00E01915D7D82ULL, "hlac_w_encounter2_327680_0x.dds"},
	{0xB1E06B432219BCE9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_die_lp_v1.ani"},
	{0xB1E1874C6A810821ULL, "vxlt_0_lowtown_a2_n_gn2_0_0.vlm"},
	{0xB1E1DABF874A6B85ULL, "vxlt_gallowscourtyard_79y_0_0.vlm"},
	{0xB1E2179D03A286CEULL, "6178482_m.fxe"},
	{0xB1E23082FA35FC38ULL, "vxht_0_hightown_at4_0_0.vlm"},
	{0xB1E23773C4F69A5DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockfxxl.phy"},
	{0xB1E27ABB6051A327ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiresml_b.mmh"},
	{0xB1E27F00188C20D7ULL, "vxdr_primevaltomb_9cn_0_1.vlm"},
	{0xB1E2A7FC70F3EF85ULL, "one140cr_charade6166005.cl"},
	{0xB1E2C16C92447D05ULL, "222538696.ogg"},
	{0xB1E2E2AB5A8EF212ULL, "plt_isa000pt_main.plo"},
	{0xB1E2F555F6284597ULL, "w_bonepit_18_0.msh"},
	{0xB1E30755F62864FAULL, "w_bonepit_18_0.mao"},
	{0xB1E30B55F6286B09ULL, "w_bonepit_18_0.mmh"},
	{0xB1E31178A09ED79DULL, "w_encounter2_camp_30_0_lc.mmh"},
	{0xB1E353AE8C0A5F9DULL, "t1_skn_005.tnt"},
	{0xB1E3D420DD1740C0ULL, "936452030.ogg"},
	{0xB1E400EBC73790FEULL, "vxca_cavern1_haunted_dpv_0_0.vlm"},
	{0xB1E46F8546BCDCD7ULL, "6119514_m.fxe"},
	{0xB1E47293A0CFC259ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.cess.evt"},
	{0xB1E4835B016E20E0ULL, "459710929.ogg"},
	{0xB1E4A6F8C7406637ULL, "238487593.ogg"},
	{0xB1E52404A2FE3034ULL, "998448345.ogg"},
	{0xB1E546C2EB0828EDULL, "vxw_bonepitmining_a2_1nf_0_0.vlm"},
	{0xB1E5785781B0312AULL, "ht_keep.arl"},
	{0xB1E5BD7C0E1DE37EULL, "321598808.ogg"},
	{0xB1E663C3A01726FAULL, "6138824_m.fxe"},
	{0xB1E6CA4F43E598B5ULL, "dre000_dougal6142202.cl"},
	{0xB1E72A3434E0E489ULL, "6176731_m.fxe"},
	{0xB1E786432220096BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_die_lp_v1.gad"},
	{0xB1E892460BA7659BULL, "art\\characters\\playercharacter\\humanmale\\bdy_cara.msh"},
	{0xB1E8A4460BA784E6ULL, "art\\characters\\playercharacter\\humanmale\\bdy_cara.mao"},
	{0xB1E8B0460BA79885ULL, "art\\characters\\playercharacter\\humanmale\\bdy_cara.mmh"},
	{0xB1E8DDBD84329770ULL, "1021155230.ogg"},
	{0xB1E92F304D3160D7ULL, "vxht_0_hightown_a3_c1s_0_0.vlm"},
	{0xB1E9A37C7B5FF0C1ULL, "6065924_f.fxe"},
	{0xB1E9E4B65F2960EBULL, "hlac_w_sundermount_131071_0y.dds"},
	{0xB1EAD1A22AC687F4ULL, "vxht_mansion2_3dz_0_0.vlm"},
	{0xB1EAEC15EFD4F03AULL, "133704039.ogg"},
	{0xB1EB1A9635D31954ULL, "6166289_m.fxe"},
	{0xB1EBC525B4B43373ULL, "pro000_colonel_ewald6040579.cl"},
	{0xB1EDA9DC33CB9954ULL, "follower_banter6170504.cl"},
	{0xB1EE3875F7E582CCULL, "vxdr_0_deeproads_col_0_0.vlm"},
	{0xB1EE3F5458050505ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_mag_0s.dds"},
	{0xB1EE784F06C8DF5FULL, "hlac_ht_chantry_2534_4x.dds"},
	{0xB1EEBB3CFFEB312CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robb_t.dds"},
	{0xB1EED32A12D5D790ULL, "vxw_bonepitmining_a2_260_0_0.vlm"},
	{0xB1EEDA60A2103680ULL, "6046954_m.fxe"},
	{0xB1EF3D71C172584DULL, "61179858.ogg"},
	{0xB1EF94872632B669ULL, "882205613.ogg"},
	{0xB1F0A8E1DB5BB25BULL, "seb321_leliana6191274.cl"},
	{0xB1F129BAC81C3F00ULL, "15793849.ogg"},
	{0xB1F19307E60B2B90ULL, "gen00fl_sebastian6178353.cl"},
	{0xB1F1ADFA21321DE0ULL, "vxca_caverns2prc_9wo_0_0.vlm"},
	{0xB1F1DB329E58E71FULL, "688417360.ogg"},
	{0xB1F20A314AFD5572ULL, "mag360_thrask6043669.cl"},
	{0xB1F20F89308A14F7ULL, "vxca_caverns1_e8o_0_0.vlm"},
	{0xB1F27A93A0DC0D48ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose2_p.cess.ani"},
	{0xB1F28B36C112D4F2ULL, "vxlt_0_lowtown_docks_n_bds_0_0.vlm"},
	{0xB1F2931BE3E2CBCBULL, "831164976.ogg"},
	{0xB1F2BC03B2C05E66ULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_ambient"},
	{0xB1F2CB050D7B56F1ULL, "43433169.ogg"},
	{0xB1F2EA5DE9110C07ULL, "vxht_mansion2b_35n_0_0.vlm"},
	{0xB1F32C73C504FAC9ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockfxxl.mao"},
	{0xB1F3346162162257ULL, "hlac_lt_0_lowtown_a2_n_c_65536_7y.dds"},
	{0xB1F33873C5050F62ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockfxxl.mmh"},
	{0xB1F33A73C5051218ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockfxxl.msh"},
	{0xB1F3933964799579ULL, "6162055_m.fxe"},
	{0xB1F401FA9FB4B326ULL, "108202793.ogg"},
	{0xB1F4E3141F1E3830ULL, "lt_0_lowtown_a2_33_0_ll.mmh"},
	{0xB1F5D2E9568672DBULL, "6156487_m.fxe"},
	{0xB1F5D940BAB3AB34ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_2.ani"},
	{0xB1F649425157F43AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basick.mmh"},
	{0xB1F64D425157FB71ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basick.mao"},
	{0xB1F65B4251581280ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basick.msh"},
	{0xB1F6F0F93136BFB9ULL, "1010182446.ogg"},
	{0xB1F6F6DF33DFAD00ULL, "hlac_ht_mansion2b_818_14y.dds"},
	{0xB1F7338E10C43E45ULL, "vxlt_ruinedhouse_n_i6_0_0.vlm"},
	{0xB1F749964FC5A312ULL, "193343612.ogg"},
	{0xB1F7945C6265A610ULL, "and330ip_at_sewers.utp"},
	{0xB1F7D44C41FF2502ULL, "6087735_m.fxe"},
	{0xB1F85B0C4A454584ULL, "ser210_witness6168185.cl"},
	{0xB1F8928D2E112102ULL, "qun221_viscount6084720.cl"},
	{0xB1F91A56E445DFC4ULL, "6124706_m.fxe"},
	{0xB1F93C71FF6C0410ULL, "6110532_m.fxe"},
	{0xB1F93D9C0166AD32ULL, "6150743_m.fxe"},
	{0xB1F995A0A451B2F2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medd_0t.dds"},
	{0xB1F9A892140E3F3FULL, "llac_ca_cavern1_haunted_24424_5.dds"},
	{0xB1F9B7D4926626ECULL, "lt_0_lowtown_docks_174_0.msh"},
	{0xB1F9BDD4926630DEULL, "lt_0_lowtown_docks_174_0.mmh"},
	{0xB1F9C9D492664575ULL, "lt_0_lowtown_docks_174_0.mao"},
	{0xB1F9D9CA3EADC040ULL, "163334961.ogg"},
	{0xB1FA0BA637E9BB50ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_bdy_snga_0d.dds"},
	{0xB1FA7176058843B0ULL, "llac_ht_mansion2b_818_28.dds"},
	{0xB1FB049C8EC39C3EULL, "6076784_m.fxe"},
	{0xB1FBFBDD75490DCAULL, "sse_ss_wryme.gda"},
	{0xB1FC22AC14D33C79ULL, "vxlt_0_lowtown_docks_a3_n_cz8_0_0.vlm"},
	{0xB1FC7A49B26ED067ULL, "vxlt_hangedman_wa_0_0.vlm"},
	{0xB1FC81FF1076226FULL, "fen110_fenris6117977.cl"},
	{0xB1FC946E94EDA78AULL, "6087372_f.fxe"},
	{0xB1FC996C56E5D18FULL, "6201474_f.fxe"},
	{0xB1FCD640BAB9C496ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_2.gad"},
	{0xB1FCF9D83A9F1734ULL, "gen00fl_merrill6107073.cl"},
	{0xB1FDDFD0C9AF735BULL, "728623832.ogg"},
	{0xB1FE4F74EC66D3BAULL, "dae000_bodhan6129100.cl"},
	{0xB1FE594442AEA554ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_0z.dds"},
	{0xB1FE6F49C2421A0BULL, "6165654_m.fxe"},
	{0xB1FEE676B00506D6ULL, "dae000_bartrand6033988.cl"},
	{0xB1FF2F8AB4CFBC50ULL, "vxlt_warehouse_ri_0_0.vlm"},
	{0xB1FF7053CFB46E16ULL, "vxlt_0_lowtown_h7r_0_0.vlm"},
	{0xB2005B26DFD60D9FULL, "vxht_chantry_n_2bm_0_0.vlm"},
	{0xB2005B5846BC4890ULL, "6105854_f.fxe"},
	{0xB200D569360375E1ULL, "505457328.ogg"},
	{0xB200E4A5AB23C1FEULL, "vxca_cavern1_haunted_dnh_0_0.vlm"},
	{0xB20125DDCEC2A31DULL, "ave200_donnic6110766.cl"},
	{0xB201B7A4E6B2796CULL, "6040850_f.fxe"},
	{0xB203120277CF22EFULL, "qcr220_orsino6061793.cl"},
	{0xB20327AC59251BE5ULL, "mag370_meredith6046843.cl"},
	{0xB203F140BAC011F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_2.evt"},
	{0xB203FD5BA07DCC7DULL, "art\\characters\\weapons\\staff\\textures\\w_stf_005a_0d.dds"},
	{0xB204564FB14E2AEFULL, "vrc321_varric6096683.cl"},
	{0xB204846080A42C39ULL, "44750263.ogg"},
	{0xB204A1802B384A2FULL, "mrl251ip_loot_e.utp"},
	{0xB2053B7B74A4FCA4ULL, "vxca_caverns2_7bc_0_0.vlm"},
	{0xB2059BF137E2B17FULL, "840019647.ogg"},
	{0xB2059FCF59713EA6ULL, "83610130.ogg"},
	{0xB205E96B6BF2D400ULL, "167775755.ogg"},
	{0xB2069CC2D9C0D9D1ULL, "6207812_f.fxe"},
	{0xB2071BCC990BCD65ULL, "vxlt_0_lowtown_a2_n_gmk_0_0.vlm"},
	{0xB2077D9429362E4CULL, "616574155.ogg"},
	{0xB207A05C436B0661ULL, "vxlt_warehouse_15p_0_0.vlm"},
	{0xB207D96E9F0D8E5BULL, "vxw_blightlands_6z_0_0.vlm"},
	{0xB207E3F9EFF06E78ULL, "zz_vault_debug6178919.cl"},
	{0xB207F6C339B54EF8ULL, "753245020.ogg"},
	{0xB208FA5DCC4E86D3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge.phy"},
	{0xB2095EF71CB29AA7ULL, "hlac_lt_0_lowtown_n_-131072_8z.dds"},
	{0xB2096FBB71F3B3ABULL, "vxht_0_hightown_ai8_0_0.vlm"},
	{0xB209BA0E58836A77ULL, "255937154.ogg"},
	{0xB20A1BE10D4BDB81ULL, "141858174.ogg"},
	{0xB20A44C4104B8C32ULL, "vxw_bonepit_1xd_0_1.vlm"},
	{0xB20A8EA3DF0A8022ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_ltbarproxi.phy"},
	{0xB20AC8EA69A5EE60ULL, "ts_darkspawn_prog_junk2.gda"},
	{0xB20B5CA9B631F48FULL, "443306854.ogg"},
	{0xB20B6C0E0F322870ULL, "471836830.ogg"},
	{0xB20B76826B56A45FULL, "rdr111dg_martin6053588.cl"},
	{0xB20B7960F06BC532ULL, "hlac_lt_shop_-1_4x.dds"},
	{0xB20C20426031FEB5ULL, "290668817.ogg"},
	{0xB20C3131382F6060ULL, "gen_im_trs_crc_lgt_glv_r0.uti"},
	{0xB20C8F544E193DCCULL, "6143870_f.fxe"},
	{0xB20CC8BDA0AA60FAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh03a.mmh"},
	{0xB20CDABDA0AA7F40ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh03a.msh"},
	{0xB20CF9F49B6F4A61ULL, "6136872_m.fxe"},
	{0xB20D3DFD8A0E0B64ULL, "6107540_f.fxe"},
	{0xB20D44085374B1F8ULL, "follower_banter6170636.cl"},
	{0xB20DA167A6C5D4D0ULL, "vxlt_warehouse_wv_0_0.vlm"},
	{0xB20DF8978DBC9594ULL, "hlac_lt_0_lowtown_a2_65533_1x.dds"},
	{0xB20F10E5A0003326ULL, "cd_mag110_vincento.wbd"},
	{0xB20F7D69A5E16559ULL, "zz_followers_banter6159210.cl"},
	{0xB20F80956B4CF4A5ULL, "weapon_weapon_p0.psh.fxo.dx11.assembly"},
	{0xB210354EEFFA6943ULL, "vxw_bonepitmining_a2_2a2_0_0.vlm"},
	{0xB210A293F4FCA32AULL, "57373641.ogg"},
	{0xB210C8A13B9EEB9FULL, "ts_demonic_02.gda"},
	{0xB210D41CBC3E0CB5ULL, "hlac_ht_chantry_n_2534_4y.dds"},
	{0xB210F13BF856F61DULL, "559778163.ogg"},
	{0xB211291204BFB248ULL, "vxlt_warehouse_15u_0_0.vlm"},
	{0xB211528D1BD5FDEDULL, "456127596.ogg"},
	{0xB2116F9EC0EA89D8ULL, "dae000_meredith6092572.cl"},
	{0xB211EFE26DC0729DULL, "hairalpha_hair_p0.vsh.fxo.dx11.assembly"},
	{0xB212CA439D7CA3C8ULL, "559272646.ogg"},
	{0xB212D9B5CFE7F1C8ULL, "fex_bd_b_bs_wl.mmh"},
	{0xB212F7B5CFE824F2ULL, "fex_bd_b_bs_wl.msh"},
	{0xB21379B969F6C899ULL, "follower_banter6169677.cl"},
	{0xB213A415C08CBD2EULL, "gen00fl_anders6112974.cl"},
	{0xB213E1D9BFEEE598ULL, "vxht_0_hightown_n_c7x_0_0.vlm"},
	{0xB213E9CF0C4D5D8CULL, "57737229.ogg"},
	{0xB2146F193A14B58BULL, "60768916.ogg"},
	{0xB21494FA281B195DULL, "qun110_petrice6042264.cl"},
	{0xB214DBCB97E86952ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masb_0n.dds"},
	{0xB214E92D8B2BC6DFULL, "one111ip_ggt_mallet.utp"},
	{0xB214F0575A765245ULL, "gen00fl_fenris6126563.cl"},
	{0xB2150AA6776000F5ULL, "vxht_keep_682_0_0.vlm"},
	{0xB21539015FDDAA1AULL, "fen110_fenris6096224.cl"},
	{0xB2158DA98B01E5E5ULL, "vxlt_0_lowtown_a2_n_gjj_0_0.vlm"},
	{0xB215F9926D81324BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_meredith.ani"},
	{0xB2160C172645EF6EULL, "296101373.ogg"},
	{0xB2164C0E8F88C7F2ULL, "mag130ip_at_eau_mages.utp"},
	{0xB216D62D8293A047ULL, "261581932.ogg"},
	{0xB217899B7A307B97ULL, "36401113.ogg"},
	{0xB217F7A32FB7B549ULL, "hlac_ca_cavern1_haunted_24421_9y.dds"},
	{0xB218DDB909F36467ULL, "6083323_m.fxe"},
	{0xB2192DF07C3906E1ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\femalemage2_.gad"},
	{0xB219976440FDCE24ULL, "6142085_m.fxe"},
	{0xB21998DB0A38AA61ULL, "l_qcr270ar_gallows.wbd"},
	{0xB2199F5F54FD543FULL, "6208382_m.fxe"},
	{0xB219FC0DB8018BDEULL, "l_dae210ar_docks.wbd"},
	{0xB21A063810FE2666ULL, "llac_lt_0_lowtown_docks_n_65533_1.dds"},
	{0xB21A51C5AB0C8EFCULL, "935443813.ogg"},
	{0xB21A524D9504DA47ULL, "dre000_carver.bnk"},
	{0xB21A63F98AE6CE15ULL, "hlac_dr_0_deeproads_5842_7y.dds"},
	{0xB21AF61856747F68ULL, "character_character_p23.psh.fxo.dx11.assembly"},
	{0xB21B2CEBAD4B5704ULL, "123047535.ogg"},
	{0xB21B3DB94336F170ULL, "286134484.ogg"},
	{0xB21B684251774545ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basick.phy"},
	{0xB21BAE4022579C56ULL, "hlac_ht_mansion2b_818_2z.dds"},
	{0xB21C0CED79EDDA04ULL, "vxht_bartrand_n_5nj_0_0.vlm"},
	{0xB21C134498BBED50ULL, "vxlt_warehouse_xw_0_0.vlm"},
	{0xB21C8A025512C7A4ULL, "694723915.ogg"},
	{0xB21CA6B94322828DULL, "621636630.ogg"},
	{0xB21CECB5B8C3AC1CULL, "character_character_p137.psh.fxo.dx11.assembly"},
	{0xB21CF75673C4E74AULL, "framebuffereffects_ssao_p0.vsh.fxo.dx11"},
	{0xB21D29F4F4BC5141ULL, "genip_pp_dr_one_shot.utp"},
	{0xB21D5F941AC52E91ULL, "hlac_lt_0_lowtown_a2_-3_1y.dds"},
	{0xB21D66915F854D8BULL, "combat_ar_mv_default.gda"},
	{0xB21D744D950749BFULL, "dre000_carver.cnv"},
	{0xB21DDDC812F236A4ULL, "vxlt_0_lowtown_h23_0_0.vlm"},
	{0xB21E2743B2B33BB3ULL, "660509869.ogg"},
	{0xB21E911307698F6DULL, "gen_im_trs_gen_hvy_cht_r1.uti"},
	{0xB21EB1A24D38221BULL, "vxht_keep_639_0_0.vlm"},
	{0xB21EC1F3C1D78C1FULL, "hlac_lt_0_lowtown_a2_n_-65535_5y.dds"},
	{0xB21EC7FEA72C10C3ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb105.mao"},
	{0xB21EEE8809C73DE9ULL, "789614642.ogg"},
	{0xB21F2EF07C3D738FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\femalemage2_.ani"},
	{0xB21F33DC966D8EEEULL, "6170231_m.fxe"},
	{0xB22024893E99695CULL, "547916587.ogg"},
	{0xB221DC6A9A4EA32DULL, "vxdr_primevaltomb_7lf_0_0.vlm"},
	{0xB2221507A02E75F4ULL, "6069052_m.fxe"},
	{0xB2224C32C9933F88ULL, "hlac_lt_gallowstemplar_2394_1y.dds"},
	{0xB2226236CC40F914ULL, "6130970_m.fxe"},
	{0xB222BC91A09508A9ULL, "ht_0_hightown_n_17_0.rml"},
	{0xB2230C389F57E19DULL, "6025541_m.fxe"},
	{0xB22326AEAD1FFC4EULL, "625379055.ogg"},
	{0xB223E1926D8D46C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_meredith.evt"},
	{0xB2241FE6733F3610ULL, "zz_fol_debug6097846.cl"},
	{0xB2245E4E8F9A8F63ULL, "hlac_dr_primevaltomb_658_8z.dds"},
	{0xB22549B72BEFED26ULL, "vxlt_0_lowtown_a2_n_c_exo_0_0.vlm"},
	{0xB225E56656836618ULL, "ts_kirkwall_arm_props.gda"},
	{0xB225F7ECCEE0F912ULL, "6189077_m.fxe"},
	{0xB226091E7A619DB5ULL, "6034683_f.fxe"},
	{0xB22654D9AE7583F0ULL, "vxlt_hangedman_26g_0_0.vlm"},
	{0xB2266F138084599BULL, "6080106_m.fxe"},
	{0xB226BE6672A1719CULL, "follower_banter6156272.cl"},
	{0xB2276A7AA0E09404ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffb_n.dds"},
	{0xB227DA84DE9C83C0ULL, "vxlt_0_lowtown_docks_n_c57_0_0.vlm"},
	{0xB227F0633F2D29A0ULL, "gen_im_flw_two_carver_02.uti"},
	{0xB227FE67A814409CULL, "dae000cr_maren.utc"},
	{0xB22824663C347E25ULL, "vxlt_gallowstemplar_n_5us_0_1.vlm"},
	{0xB229131F9131741FULL, "91982468.ogg"},
	{0xB229252B87AE5942ULL, "312710516.ogg"},
	{0xB22941BC8C87325DULL, "hlac_lt_0_lowtown_-131072_0y.dds"},
	{0xB229E0926D91B0D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncexit_meredith.gad"},
	{0xB229FF0AAA9CFD12ULL, "vxlt_0_lowtown_a2_csi_0_0.vlm"},
	{0xB22A21A82EB419F5ULL, "fen310_fenris6144520.cl"},
	{0xB22A34E40C89B973ULL, "seb221_flora_harimann6177281.cl"},
	{0xB22A369FEB29BB85ULL, "wa_so.mao"},
	{0xB22ADA90C7958407ULL, "art\\vfx\\textures\\vfx_fire_single.dds"},
	{0xB22B621F08687F11ULL, "948228354.ogg"},
	{0xB22B9DB97A696E2CULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga105a_0s.dds"},
	{0xB22C3A2BD598E8C3ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_wmda.msh"},
	{0xB22C482BD599001DULL, "art\\characters\\playercharacter\\humanfemale\\hlf_wmda.mmh"},
	{0xB22D39D82CF8457DULL, "213968662.ogg"},
	{0xB22DA5E37F6B5795ULL, "hlac_w_encounter2_camp_-131067_0x.dds"},
	{0xB22DE909A2D8DAE3ULL, "zz_jebs_test.cut"},
	{0xB22DF1C7BCE3D521ULL, "6037685_m.fxe"},
	{0xB22E57A8E7E04CE1ULL, "6060832_m.fxe"},
	{0xB22E82C7DC598024ULL, "w_sundermount_180_0.msh"},
	{0xB22E94C7DC599E8DULL, "w_sundermount_180_0.mao"},
	{0xB22E98C7DC59A5C6ULL, "w_sundermount_180_0.mmh"},
	{0xB22F4B503DF9A7E4ULL, "weapon_weapon_p3.psh.fxo.dx11"},
	{0xB22F649ED44961DAULL, "152343657.ogg"},
	{0xB22F89460BE35932ULL, "art\\characters\\playercharacter\\humanmale\\bdy_cara.phy"},
	{0xB22FCB9180405898ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cartempty_01_0_n.dds"},
	{0xB22FCF3CC06D1A7DULL, "238364566.ogg"},
	{0xB22FD71CE9431F83ULL, "ser270_cut_aftermath6162272.cl"},
	{0xB2305AB66EF160B7ULL, "hm_one_har_guard2.mor"},
	{0xB23069A3DF2B10D6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_ltbarproxi.mao"},
	{0xB23069BE46F842F9ULL, "104635327.ogg"},
	{0xB23075A3DF2B2575ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_ltbarproxi.mmh"},
	{0xB23077A3DF2B28EBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_ltbarproxi.msh"},
	{0xB2309802593DD700ULL, "hlac_lt_0_lowtown_a2_n_c_0_4z.dds"},
	{0xB230A9F80BB8C9BAULL, "892978645.ogg"},
	{0xB230C593A3DB1698ULL, "vxht_0_hightown_n_c9v_0_0.vlm"},
	{0xB230F5A3616A63A7ULL, "gen00fl_aveline6106395.cl"},
	{0xB2314438398C8743ULL, "6137222_m.fxe"},
	{0xB231CFDDC641A383ULL, "zz_ser_debug6077998.cl"},
	{0xB231E7BDA0C9B105ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh03a.phy"},
	{0xB231E7D1BE8D6797ULL, "695990193.ogg"},
	{0xB23299B912E2C378ULL, "vxca_cavern1_haunted_ing_0_0.vlm"},
	{0xB232D71C943D6B82ULL, "518321022.ogg"},
	{0xB23307213B6F5A85ULL, "ave310_brennan6137245.cl"},
	{0xB2338DF04641A055ULL, "hlac_lt_house_-1_4y.dds"},
	{0xB2350D931C90B9C7ULL, "gen00fl_fenris6113891.cl"},
	{0xB2358CF1E785C160ULL, "vxlt_gallowscourtyard_7f5_0_0.vlm"},
	{0xB2364C61D0929C40ULL, "vxlt_0_lowtown_docks_n_c04_0_0.vlm"},
	{0xB236A8723BDD6E26ULL, "vxlt_0_lowtown_a2_n_c_h28_0_0.vlm"},
	{0xB236AC7A8DD39529ULL, "868912788.ogg"},
	{0xB237378DB56CAD2DULL, "6115798_m.fxe"},
	{0xB237F40995B7A3ADULL, "hlac_lt_0_lowtown_a2_n_c_65536_6x.dds"},
	{0xB23895268993EC75ULL, "6141838_m.fxe"},
	{0xB238AB198C153E31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_tw.ani"},
	{0xB23935417C892E81ULL, "isa211_isabela6202295.cl"},
	{0xB2398FC7DC639D91ULL, "w_sundermount_180_0.phy"},
	{0xB23A19897F4DDEEFULL, "96572262.ogg"},
	{0xB23AB7EE13805AD8ULL, "113485565.ogg"},
	{0xB23B6800C8B4BB6EULL, "602252788.ogg"},
	{0xB23B943C475150CFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medc_0n.dds"},
	{0xB23BBE4DCD5F0776ULL, "635489289.ogg"},
	{0xB23C981703A0CBB2ULL, "vxlt_0_lowtown_a3_n_c_hpg_0_0.vlm"},
	{0xB23CE503C92F6D93ULL, "680123957.ogg"},
	{0xB23D77DFA404D48CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlc.phy"},
	{0xB23D7A68E8A1B5E2ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appc.phy"},
	{0xB23DC6783A2D2A24ULL, "mer231_brekker_st.stg"},
	{0xB23DE99B28A9984CULL, "rdr110dg_dockworker6015944.cl"},
	{0xB23E2515F3334A96ULL, "862406826.ogg"},
	{0xB23E591A858B01ECULL, "317654417.ogg"},
	{0xB23E70685232CE21ULL, "one140cr_charade6128515.cl"},
	{0xB23EBADB0A580063ULL, "l_qcr270ar_gallows.bnk"},
	{0xB23F1DCCECAFFA80ULL, "6208473_m.fxe"},
	{0xB23F681FF4BFD44CULL, "6120456_m.fxe"},
	{0xB23F6C4575A865A1ULL, "6148121_m.fxe"},
	{0xB23F8085D0FE3A3BULL, "dae111st_dog_player_plus.stg"},
	{0xB23FA32A842EEA0BULL, "vxlt_0_lowtown_a2_n_f58_0_0.vlm"},
	{0xB23FB6198C1B6FD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_tw.gad"},
	{0xB2410B9604D8DBF8ULL, "vxca_cavern1_haunted_ikd_0_0.vlm"},
	{0xB2410F9624B08ED9ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredidol.ani"},
	{0xB241B480F46E9C9BULL, "vxlt_gallowstemplar_n_4t2_0_0.vlm"},
	{0xB241EB10B2ADE792ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cleanse_area.gad"},
	{0xB241F236F3F2E68FULL, "76158480.ogg"},
	{0xB242A3072F7A62BFULL, "905780741.ogg"},
	{0xB2430F9801A49640ULL, "758623922.ogg"},
	{0xB2434A614FF1ADDCULL, "6162278_m.fxe"},
	{0xB2435961D97B72D4ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.anb"},
	{0xB243FA6B7D4140DEULL, "hlac_lt_gallowstemplar_n_2382_2z.dds"},
	{0xB2440674502733FEULL, "hlac_lt_0_lowtown_n_-196608_0y.dds"},
	{0xB244237AE4B3D5FCULL, "vxlt_undercityencounter_12o_0_0.vlm"},
	{0xB2446AF6E9BE64A0ULL, "815517386.ogg"},
	{0xB244773C6C37D1EAULL, "hlac_w_woundedcoast_0_0y.dds"},
	{0xB244E5D2754DD902ULL, "art\\characters\\playercharacter\\humanmale\\hlh_meda.msh"},
	{0xB244E7D2754DDCB8ULL, "art\\characters\\playercharacter\\humanmale\\hlh_meda.mmh"},
	{0xB244F3D2754DF15BULL, "art\\characters\\playercharacter\\humanmale\\hlh_meda.mao"},
	{0xB2454E374A3D85B5ULL, "gen00fl_anders6113095.cl"},
	{0xB2455D33216EE73FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02a.msh"},
	{0xB245617735B4DA9DULL, "6085925_m.fxe"},
	{0xB2456333216EF1E1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02a.mmh"},
	{0xB245A861FA5CA2BFULL, "lgt160_wk_soph_sophia6119823.cl"},
	{0xB245B3198C1FD638ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_std_beg_tw.evt"},
	{0xB245E01CE2746D70ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magicloker_n.dds"},
	{0xB246E238CFA5F872ULL, "507446855.ogg"},
	{0xB2474C1CA0BBCB59ULL, "one300_zevran6076930.cl"},
	{0xB2474CDB996B1A0CULL, "art\\characters\\creatures\\crowds\\elfmale\\textures\\prp_hightown_jimmy_n.dds"},
	{0xB247A91DA019CC48ULL, "hlac_lt_darktown_-1_10y.dds"},
	{0xB247A9F15EEAE0A7ULL, "648854178.ogg"},
	{0xB247EA9624B66E1BULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow2_meredidol.gad"},
	{0xB24836E3480CD6B7ULL, "hlac_dr_primevaltomb_658_2x.dds"},
	{0xB2486A0F38BAC808ULL, "zz_fen_debug6134341.cl"},
	{0xB248EE10B2B40BC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cleanse_area.evt"},
	{0xB2499270692C359EULL, "869103771.ogg"},
	{0xB249B4EAAF8D960FULL, "vxht_mansion2b_3cl_0_0.vlm"},
	{0xB24A0C6E56FE09D0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_small.mao"},
	{0xB24A106E56FE10E3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_small.mmh"},
	{0xB24A1E6E56FE283DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_small.msh"},
	{0xB24A253C01BCCA28ULL, "vxht_keep_64g_0_0.vlm"},
	{0xB24A7E1C8544472DULL, "llac_lt_gallowstemplar_f_2395_1.dds"},
	{0xB24BDB5B8668E57EULL, "mcr310_sibling6103595.cl"},
	{0xB24C77CCBC0D6E31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerg.msh"},
	{0xB24C784073B45AB2ULL, "gen_im_trs_dwv_med_cht_r2.uti"},
	{0xB24C85CCBC0D864CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerg.mao"},
	{0xB24C91CCBC0D9AEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerg.mmh"},
	{0xB24CF564D3D36654ULL, "zz_one_debug6099957.cl"},
	{0xB24D0B63448E65E5ULL, "551575092.ogg"},
	{0xB24DC6788154D71DULL, "mrl150_merrill6085739.cl"},
	{0xB24DE60E77284F5BULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbed.mmh"},
	{0xB24DF20E772863F8ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbed.mao"},
	{0xB24E040E77288245ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbed.msh"},
	{0xB24EB1DE8CB79598ULL, "vxht_0_hightown_a2_c1m_0_0.vlm"},
	{0xB24EEED275563D43ULL, "art\\characters\\playercharacter\\humanmale\\hlh_meda.phy"},
	{0xB24F8B20D48E5EA0ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_bdy_snga_0t.dds"},
	{0xB24FA4AC47617E37ULL, "llac_lt_0_lowtown_a3_n_c_-2_1.dds"},
	{0xB2503861D985C497ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire2.mmh"},
	{0xB25081661C7141B2ULL, "321145961.ogg"},
	{0xB2509F4E01D236A6ULL, "vxlt_0_lowtown_a2_n_c_ewr_0_0.vlm"},
	{0xB250ADEEAFB4C045ULL, "lt_0_lowtown_docks_a3_n_20_0.anb"},
	{0xB250D8AECC720C94ULL, "389890447.ogg"},
	{0xB251162432767B85ULL, "vxw_bonepitdestroyed_a3_19y_0_0.vlm"},
	{0xB25136E9EE8384BEULL, "hlac_lt_0_lowtown_docks_n_-3_8y.dds"},
	{0xB251A7ACADB47FD6ULL, "vxlt_0_lowtown_docks_a3_aom_0_0.vlm"},
	{0xB251B0D22AF0C490ULL, "docks_male_overseer_mob6176531.cl"},
	{0xB251BC3E41E7C65CULL, "6167178_m.fxe"},
	{0xB251D389FCED9CD8ULL, "759754834.ogg"},
	{0xB251D5EBD12F2CF4ULL, "vxw_encounter3_fu_0_0.vlm"},
	{0xB2529F0E1EED4F54ULL, "vxlt_gallowstemplar_f_6qh_0_0.vlm"},
	{0xB25343AE8F35ED2EULL, "vxca_caverns2prc_4x5_0_0.vlm"},
	{0xB25366C43303F787ULL, "zz_followers_banter6158592.cl"},
	{0xB25380E2E3741076ULL, "vxlt_0_lowtown_docks_a3_97m_0_0.vlm"},
	{0xB253EA99D0401310ULL, "gen00fl_sebastian6195820.cl"},
	{0xB2546A71A362F094ULL, "t3_rob_dal.tnt"},
	{0xB2554DEB7711E12AULL, "follower_banter6164503.cl"},
	{0xB255BB91A0C060C9ULL, "ht_0_hightown_n_17_0.anb"},
	{0xB255C515910D5993ULL, "target_cone_pressed.cur"},
	{0xB2562F2536020D78ULL, "6131223_m.fxe"},
	{0xB2564F386BCD96B6ULL, "vxlt_0_lowtown_a2_n_gcb_0_0.vlm"},
	{0xB256D610B2C02038ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cleanse_area.ani"},
	{0xB2573517F1C9E18CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02gravecairn.phy"},
	{0xB2577E5C0F505072ULL, "llac_lt_undercitydungeon_-1_29.dds"},
	{0xB2579C7942BF5D0FULL, "6105887_m.fxe"},
	{0xB257BFA991547FFCULL, "gen00fl_fenris6126494.cl"},
	{0xB25826C7A183AE8FULL, "vxht_mansion2b_3fo_0_0.vlm"},
	{0xB258724ECECB7344ULL, "6165536_m.fxe"},
	{0xB2589718F3B2E2A8ULL, "vxht_keep_61d_0_0.vlm"},
	{0xB258D9261131A8C2ULL, "m_dae311ar_hanged_man_nt.wbd"},
	{0xB25922E82BD172D7ULL, "fx_energy02.mao"},
	{0xB259E3C7CEC3D0FAULL, "883067925.ogg"},
	{0xB25A8D873BB19B6FULL, "vxca_caverns2prc_4ke_0_0.vlm"},
	{0xB25B116E570C8508ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1co_small.phy"},
	{0xB25B3726528E8DA4ULL, "docks_male_human_worker_sta6176004.cl"},
	{0xB25B4998EB526DF5ULL, "vxlt_gallowscourtyard_n_7gd_0_0.vlm"},
	{0xB25B6198DFC0A0F2ULL, "6093885_f.fxe"},
	{0xB25B956BEFD2B687ULL, "198983289.ogg"},
	{0xB25BC4A22256B8ACULL, "76913015.ogg"},
	{0xB25BC99445B279B0ULL, "mcr310_argument6182490.cl"},
	{0xB25C23E26AF7C756ULL, "hlac_w_bonepitdestroyed_a3_131074_0x.dds"},
	{0xB25C5E50669CC023ULL, "qun100_mage_party_comment.bnk"},
	{0xB25CB7DB3098A87CULL, "vxlt_0_lowtown_docks_a3_d2u_0_0.vlm"},
	{0xB25D3654EA76ACE2ULL, "431994246.ogg"},
	{0xB25D885D3F71A894ULL, "972367093.ogg"},
	{0xB25DA797D28D1959ULL, "385188776.ogg"},
	{0xB25E1CC1D6E625D8ULL, "411984997.ogg"},
	{0xB25E1E35B13370CDULL, "556842678.ogg"},
	{0xB25E2CA2C27785B9ULL, "vxdr_0_deeproads_cph_0_0.vlm"},
	{0xB25EA782B2697235ULL, "522998851.ogg"},
	{0xB25EE4128F96D1FDULL, "w_blightlands_160_0.mao"},
	{0xB25EE8128F96D836ULL, "w_blightlands_160_0.mmh"},
	{0xB25EEC98F04D9F9DULL, "6051717_m.fxe"},
	{0xB25EF2128F96E934ULL, "w_blightlands_160_0.msh"},
	{0xB25EF6BA2D2D5EE6ULL, "508779576.ogg"},
	{0xB25F4419FEE02773ULL, "vxht_0_hightown_a3_ai9_0_0.vlm"},
	{0xB25F483034D364F5ULL, "64485796.ogg"},
	{0xB260405066A07653ULL, "qun100_mage_party_comment.cnv"},
	{0xB26047C8C3254574ULL, "vxw_encounter3_cz_0_0.vlm"},
	{0xB260665EE0BEFDF0ULL, "6207995_m.fxe"},
	{0xB260666049373FAEULL, "follower_banter6141735.cl"},
	{0xB260D6E4F8915FB9ULL, "178085604.ogg"},
	{0xB260DD3F450A34CEULL, "442686008.ogg"},
	{0xB260FC1C8E0E3DD9ULL, "437585133.ogg"},
	{0xB2610083CADFC19CULL, "6051607_f.fxe"},
	{0xB2611BBA643CE417ULL, "hlac_lt_warehouse_0_6x.dds"},
	{0xB261E5241094E87EULL, "409553314.ogg"},
	{0xB261EA4AA3A0B7AFULL, "142019204.ogg"},
	{0xB261EF3636361268ULL, "266592406.ogg"},
	{0xB26222B589CF7E74ULL, "vxht_keep_6ba_0_0.vlm"},
	{0xB262380BC3981749ULL, "gen00fl_aveline6115498.cl"},
	{0xB2625568E8C09296ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appc.mao"},
	{0xB2626168E8C0A735ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appc.mmh"},
	{0xB2626368E8C0AAABULL, "art\\characters\\playercharacter\\humanfemale\\rob_appc.msh"},
	{0xB2626D7FC32AB0E5ULL, "348251851.ogg"},
	{0xB262ADFD3032C9E5ULL, "6145214_m.fxe"},
	{0xB262B3F56E4908CCULL, "vxdr_primevalentrance_15p_0_2.vlm"},
	{0xB262BF935D260F45ULL, "305053630.ogg"},
	{0xB262C4778E0F816EULL, "871239480.ogg"},
	{0xB263271FE12BB57DULL, "vxlt_0_lowtown_eun_0_0.vlm"},
	{0xB26331E91DCAD8A7ULL, "m_dae250ar_sundermount.wbd"},
	{0xB2634F7D66F017B5ULL, "vxw_blightlands_bx_0_0.vlm"},
	{0xB26371C305536F56ULL, "ser270_emeric6116096.cl"},
	{0xB263A3A8EA2202B5ULL, "hla_ht_0_hightown_a3_-65536_0x.dds"},
	{0xB264180CE554CC09ULL, "6095744_m.fxe"},
	{0xB264933F855AF204ULL, "vxlt_0_lowtown_a2_n_gy2_0_0.vlm"},
	{0xB2650F145B4E8AD4ULL, "vxht_0_hightown_a3_2xr_0_0.vlm"},
	{0xB265CB423934560AULL, "290161892.ogg"},
	{0xB266D91D0BFD56E2ULL, "6042350_m.fxe"},
	{0xB26859E94E72AD38ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_carcass_n.dds"},
	{0xB26890CA83D0B850ULL, "vxw_encounter3_4g_0_0.vlm"},
	{0xB268942CBC5E3EA3ULL, "hlac_w_sundermount_131070_0z.dds"},
	{0xB2689F8A28175A9EULL, "fx_lyrium_whispy.mao"},
	{0xB268BDD3BEE55F8CULL, "gen00fl_anders6125817.cl"},
	{0xB268FD8C91CED0B9ULL, "vxht_chantry_2lj_0_0.vlm"},
	{0xB26968141868FCCEULL, "cd_mag100_cullen.wbd"},
	{0xB269D65BDEAE4CB3ULL, "6082002_f.fxe"},
	{0xB26A15BB541D1DADULL, "vxdr_primevaltomb_7ua_0_0.vlm"},
	{0xB26A4CC6B8C0762EULL, "gen00fl_merrill6115964.cl"},
	{0xB26A90C0FD6D3D27ULL, "gen_im_wep_war_one_act1_01.uti"},
	{0xB26AA627AF76C392ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_3y.dds"},
	{0xB26B0212C5387165ULL, "docks_fem_sailor_mob6176310.cl"},
	{0xB26B5C27FE0A3508ULL, "680280646.ogg"},
	{0xB26B7171A44604F9ULL, "vxlt_gallowstemplar_663_0_0.vlm"},
	{0xB26BBA3CD8B74260ULL, "6068590_m.fxe"},
	{0xB26BF35DCCA2F848ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge.mmh"},
	{0xB26BFF5DCCA30CEBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge.mao"},
	{0xB26C115DCCA32B72ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge.msh"},
	{0xB26C9E9665C6F447ULL, "6104249_f.fxe"},
	{0xB26D048965ABB538ULL, "hlac_dr_primevaltomb_659_15z.dds"},
	{0xB26DB5F6F0D63FF3ULL, "vxht_0_hightown_a3_av8_0_0.vlm"},
	{0xB26DC45F3A499B7CULL, "zz_mrl_debug6109305.cl"},
	{0xB26E4CEEC1853787ULL, "llac_ca_cavern1_haunted_24421_8.dds"},
	{0xB26E832C6D3C3535ULL, "hlac_ca_cavern1_haunted_24424_23x.dds"},
	{0xB26EAC4E7E87E052ULL, "mag360_samson6042093.cl"},
	{0xB26EEAC336D05064ULL, "hlac_ht_0_hightown_n_-65537_2y.dds"},
	{0xB26F193B44BFB0AFULL, "413503489.ogg"},
	{0xB26F4C2B7C21BD5BULL, "hlac_lt_0_lowtown_docks_a3_0_2z.dds"},
	{0xB27029E82BE45D0DULL, "fx_energy02.dds"},
	{0xB2702BA43A5729F1ULL, "345248935.ogg"},
	{0xB27031B14798514EULL, "146132674.ogg"},
	{0xB2706D1833FB5E7BULL, "vxdr_primevaltomb_7ew_0_0.vlm"},
	{0xB270F10BC6107ABBULL, "136684090.ogg"},
	{0xB2711431A75B9F78ULL, "qcr000_arishok6059314.cl"},
	{0xB27132A68414899DULL, "324276726.ogg"},
	{0xB271643321942E5EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth02a.phy"},
	{0xB271ECE2EC044C0EULL, "zz_lgt_1_debug6114615.cl"},
	{0xB27381C135352FDDULL, "526327936.ogg"},
	{0xB273C873CC95A40BULL, "522306434.ogg"},
	{0xB2747C2FB03DF8D8ULL, "vxht_brothel_n_6ey_0_1.vlm"},
	{0xB274FC560D5B1ABFULL, "290442134.ogg"},
	{0xB27592F18212CD35ULL, "6092190_m.fxe"},
	{0xB275E475033F2801ULL, "gentr_team_move.ncs"},
	{0xB2765F8ACB792DB2ULL, "vxlt_0_lowtown_a3_n_c_hlc_0_0.vlm"},
	{0xB276DDCC266B4970ULL, "hlac_w_blightlands_65533_0y.dds"},
	{0xB276E8A20F8F4558ULL, "6018799_m.fxe"},
	{0xB27742944B12C79FULL, "art\\characters\\creatures\\dragon\\animation\\dppr.gad"},
	{0xB27744CA0FCC7F67ULL, "6152160_m.fxe"},
	{0xB277524B63FE6EABULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackleft.evt"},
	{0xB27754FDF510A0CEULL, "lgt260_way_blood_mage6148378.cl"},
	{0xB2778260C0B31E7EULL, "6035324_f.fxe"},
	{0xB277B2742E8451EDULL, "743119974.ogg"},
	{0xB2790FAAD4A3E1BEULL, "plt_vrc100pt_n_personal.nss"},
	{0xB2796F586EB4F723ULL, "vxca_cavern1_haunted_gka_0_0.vlm"},
	{0xB27AAB7603249AEAULL, "6098783_f.fxe"},
	{0xB27ACCAE33CDEF80ULL, "hlac_dr_slavers_3216_5y.dds"},
	{0xB27B103897F8A1C8ULL, "art\\characters\\creatures\\corpse\\uni_corm.msh"},
	{0xB27B14FA047A1DDBULL, "vxw_encounter2_4o_0_0.vlm"},
	{0xB27B2E3897F8D4F2ULL, "art\\characters\\creatures\\corpse\\uni_corm.mmh"},
	{0xB27B3780F2C1A750ULL, "hlac_lt_0_lowtown_docks_a3_-65538_5x.dds"},
	{0xB27B47C37EB99203ULL, "946647429.ogg"},
	{0xB27B4E3681D7DC57ULL, "dre000_stroud6066428.cl"},
	{0xB27B7C2128E45EF1ULL, "plt_cod_itm_yusaris.plo"},
	{0xB27BD04DA8A51FDDULL, "vxlt_0_lowtown_a2_fim_0_0.vlm"},
	{0xB27BF0526C98A509ULL, "vxlt_darktown_4ih_0_0.vlm"},
	{0xB27C4227B6648574ULL, "mag291_vincento6027678.cl"},
	{0xB27C8600296D99EEULL, "dae000ip_crowd_lm5_talk10.utp"},
	{0xB27D197C88CE696AULL, "6109049_f.fxe"},
	{0xB27D7C9E9512BB46ULL, "6105859_m.fxe"},
	{0xB27DA4065128F523ULL, "hlac_lt_backalley_n_0_15z.dds"},
	{0xB27DBD43465810F8ULL, "vxlt_gallowscourtyard_7cr_0_0.vlm"},
	{0xB27E37944B18D38CULL, "art\\characters\\creatures\\dragon\\animation\\dppr.evt"},
	{0xB27E4EDFA43C2657ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlc.mmh"},
	{0xB27E52DFA43C2D64ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlc.mao"},
	{0xB27E64DFA43C4BC9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlc.msh"},
	{0xB27E6B4B6404B86CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackleft.gad"},
	{0xB27EB5D729FD6C7EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0ww_basic_1.mao"},
	{0xB27F066F6DBBAF74ULL, "vxht_keep_6hg_0_0.vlm"},
	{0xB27F10F7305393A6ULL, "hlac_lt_0_lowtown_1_1y.dds"},
	{0xB27F1951A03CFCA6ULL, "custic_dark.dds"},
	{0xB27F2281D14EF141ULL, "rdr111dg_martin6054521.cl"},
	{0xB2800AD9C517E67DULL, "vxlt_0_lowtown_ekl_0_0.vlm"},
	{0xB2800C7FE9B1DCCDULL, "s_ballista_01_l2.mao"},
	{0xB2807A90D69AF8FDULL, "hlac_ht_mansion2_818_61z.dds"},
	{0xB280B4D8E0FA1ABEULL, "vxht_0_hightown_n_c61_0_0.vlm"},
	{0xB28368E920D88E8FULL, "6148224_m.fxe"},
	{0xB283928A223B379AULL, "art\\characters\\weapons\\staff\\w_stf_st104a.mmh"},
	{0xB283A30CC6598D7FULL, "character_character_p242.psh.fxo.dx11"},
	{0xB283A48A223B56E0ULL, "art\\characters\\weapons\\staff\\w_stf_st104a.msh"},
	{0xB284044B35AE495DULL, "llac_lt_0_lowtown_docks_-65539_4.dds"},
	{0xB28404CE7408E5B5ULL, "6097290_m.fxe"},
	{0xB2840E241D199CEEULL, "6078333_m.fxe"},
	{0xB28447DCAA9DEE74ULL, "vxht_0_hightown_n_bgm_0_0.vlm"},
	{0xB28483428B5B1D01ULL, "dae100_bran6058962.cl"},
	{0xB284F2C6FDD4A3F2ULL, "6152756_m.fxe"},
	{0xB285151006B9F51AULL, "vxlt_foundry_n_ya_0_0.vlm"},
	{0xB2854CF88AE4C32EULL, "dae011_corff_sta6142449.cl"},
	{0xB2856A4B640AD56AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackleft.ani"},
	{0xB285D7D3B324DEFFULL, "sst_ss_qunari_m_1.gda"},
	{0xB28605CB2EDECBA5ULL, "vxht_bartrand_n_736_0_0.vlm"},
	{0xB286615FE780EDCDULL, "ttx_pillar_base_b_1.mmh"},
	{0xB286735FE7810C53ULL, "ttx_pillar_base_b_1.msh"},
	{0xB2867D1129964622ULL, "art\\levels\\races\\design\\prp_metaltrp.phy"},
	{0xB286DE351858028CULL, "410935761.ogg"},
	{0xB2874197CFBB8179ULL, "vxht_mansion2b_3e0_0_0.vlm"},
	{0xB287CCC429D8E558ULL, "llac_w_encounter3_-262147_0.dds"},
	{0xB287E49CAEC5A060ULL, "vxw_bonepitdestroyed_a3_1x7_0_0.vlm"},
	{0xB287F2F7F1B0B300ULL, "zz_bet_debug6113388.cl"},
	{0xB2885E3EB860286EULL, "vxht_mansion2b_2jk_0_0.vlm"},
	{0xB2887CD01F1935F8ULL, "778594048.ogg"},
	{0xB2891F285C9D3E6CULL, "vxht_keep_n_3r3_0_0.vlm"},
	{0xB28989EC6A4685A5ULL, "228920673.ogg"},
	{0xB289C6B2F987A2D2ULL, "6219136_m.fxe"},
	{0xB28AAEEEAFE63A4DULL, "lt_0_lowtown_docks_a3_n_20_0.rml"},
	{0xB28AD6BF25F014F4ULL, "zz_charge_forest2.cut"},
	{0xB28AE8823E428846ULL, "vxlt_warehouse_n_143_0_0.vlm"},
	{0xB28AF2110B24F787ULL, "vxlt_0_lowtown_a2_n_eou_0_0.vlm"},
	{0xB28B54711AFA1244ULL, "71355618.ogg"},
	{0xB28B6ECFD3F9016CULL, "1479957.ogg"},
	{0xB28BA0C8B8228E2BULL, "39316032.ogg"},
	{0xB28BC85FEF7966C6ULL, "hlac_lt_0_lowtown_65535_5z.dds"},
	{0xB28BD1EC449596E0ULL, "dae000cr_gossip_old_m2.utc"},
	{0xB28BEF8D3FE47648ULL, "gen00fl_anders6111119.cl"},
	{0xB28C1F944B24E8E5ULL, "art\\characters\\creatures\\dragon\\animation\\dppr.ani"},
	{0xB28C5242AF5D48B9ULL, "gen00fl_dog6117182.cl"},
	{0xB28CD0272E1D4FB5ULL, "537756545.ogg"},
	{0xB28D0B61ECEF9D52ULL, "art\\vfx\\combat\\vfx_gib_cora_rtl.gad"},
	{0xB28D5E2D3084FE30ULL, "6060445_f.fxe"},
	{0xB28DF0643DE487F3ULL, "mag320_noblewoman_16060277.cl"},
	{0xB28E0D1244F6E523ULL, "6031146_m.fxe"},
	{0xB28E95D632C55713ULL, "hlac_lt_gallowsprison_n_-1_11z.dds"},
	{0xB28F26B5D2F0333EULL, "vxht_0_hightown_n_c32_0_0.vlm"},
	{0xB28F29CB775D459AULL, "art\\characters\\creatures\\buzzard\\c_buzzarda.phy"},
	{0xB29001441D15AF95ULL, "dre000_bodahn6031263.cl"},
	{0xB29081CA8C243FA2ULL, "379290352.ogg"},
	{0xB290A367349B8D80ULL, "qun220cr_javaris6071578.cl"},
	{0xB291F51BBDF169B6ULL, "6125710_m.fxe"},
	{0xB29286805D030473ULL, "939181662.ogg"},
	{0xB292AAA7C582BC63ULL, "vxlt_0_lowtown_docks_a3_d16_0_1.vlm"},
	{0xB2932665A4F30115ULL, "art\\characters\\weapons\\longswords\\w_lsw_012a.phy"},
	{0xB2932A0EE58AECC5ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_lsw_reva_0n.dds"},
	{0xB29386ECF8B00D9AULL, "vxlt_foundry_n_tb_0_0.vlm"},
	{0xB293F8698F4704FDULL, "6125447_m.fxe"},
	{0xB294B01EAAFEB258ULL, "hlac_lt_0_lowtown_a2_131071_0z.dds"},
	{0xB2953AD51CC01E23ULL, "590743852.ogg"},
	{0xB2955E14188E26FCULL, "cd_mag100_cullen.bnk"},
	{0xB295701C2DD8BA9EULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_c.impa.ani"},
	{0xB295A3263C7D2DE4ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.dura.ani"},
	{0xB296070E77663C90ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbed.phy"},
	{0xB29687B5668265C9ULL, "ico_belt_1.dds"},
	{0xB29688C7C7AE41C2ULL, "6117463_m.fxe"},
	{0xB296E6EF950BCDAFULL, "373392686.ogg"},
	{0xB2970AF1D5A04F4CULL, "6077583_m.fxe"},
	{0xB2974D527CE08340ULL, "sb_fade_cloudb.dds"},
	{0xB2980C0738F7EDC2ULL, "vxht_bartrand_n_68a_0_0.vlm"},
	{0xB2980C17F2013357ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02gravecairn.mmh"},
	{0xB2981017F2013A64ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02gravecairn.mao"},
	{0xB2982217F20158C9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02gravecairn.msh"},
	{0xB2982A7BC65FA30FULL, "426221185.ogg"},
	{0xB2984C1012AA835BULL, "vxw_blightlands_dt_0_0.vlm"},
	{0xB298CBD10F83B9E5ULL, "6076013_m.fxe"},
	{0xB2990C87C618F492ULL, "467896444.ogg"},
	{0xB2992BE849DD13F6ULL, "character_character_p283.psh.fxo.dx11.assembly"},
	{0xB2993044956C5F48ULL, "10581621.ogg"},
	{0xB2995A5F3038A0C6ULL, "vxlt_warehouse_n_194_0_0.vlm"},
	{0xB2996ED2BD697E21ULL, "mcr310_argument6074381.cl"},
	{0xB2997B05F68B4E1CULL, "hlac_lt_0_lowtown_docks_a3_-65537_2x.dds"},
	{0xB299D3D20C37B53AULL, "vxca_cave_u0_0_0.vlm"},
	{0xB299D63F69A8E133ULL, "665418959.ogg"},
	{0xB29A120AF9C55997ULL, "ser121_jethann6120018.cl"},
	{0xB29A3EA3551DB16FULL, "zz_dae_debug6045454.cl"},
	{0xB29AAC0F01AFBE61ULL, "hlac_lt_hangedman_-1_26x.dds"},
	{0xB29AF88415D7F8B1ULL, "6014695_m.fxe"},
	{0xB29BC2AF09DDA7FBULL, "vxdr_primevaltomb_7th_0_0.vlm"},
	{0xB29CA0263C834746ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.dura.gad"},
	{0xB29CA332434A2984ULL, "17052294.ogg"},
	{0xB29D665D7F02C8CFULL, "527262654.ogg"},
	{0xB29D7AE3A4C9EB34ULL, "6111265_m.fxe"},
	{0xB29D9892C4E64BBEULL, "vxht_0_hightown_n_c83_0_0.vlm"},
	{0xB29E1A99294E7787ULL, "lgt250_way_dalish_conflict6147141.cl"},
	{0xB29E7CF3F0778E9AULL, "6053547_m.fxe"},
	{0xB29E8BF55A14D786ULL, "146670480.ogg"},
	{0xB29EB06CF2C44010ULL, "752368250.ogg"},
	{0xB29EF3B0495FACABULL, "6112065_m.fxe"},
	{0xB29FDFC0A404006BULL, "vxdr_primevaltomb_6ni_0_2.vlm"},
	{0xB29FE726116DC378ULL, "m_dae311ar_hanged_man_nt.bnk"},
	{0xB2A01365AA3DA514ULL, "vxlt_gallowstemplar_3rg_0_0.vlm"},
	{0xB2A0184B144C39D3ULL, "tbe100im_nug_gloves.uti"},
	{0xB2A03069145E06E2ULL, "hlac_dr_0_deeproads_5845_6x.dds"},
	{0xB2A03C376BA1119BULL, "vxca_caverns1_mines_j1z_0_0.vlm"},
	{0xB2A057390F1354E5ULL, "6133251_m.fxe"},
	{0xB2A075F5CE73AD78ULL, "6096021_m.fxe"},
	{0xB2A08A574C8A9C5CULL, "404216953.ogg"},
	{0xB2A0F661ED0022F3ULL, "art\\vfx\\combat\\vfx_gib_cora_rtl.anb"},
	{0xB2A0F661ED0022F8ULL, "art\\vfx\\combat\\vfx_gib_cora_rtl.ani"},
	{0xB2A0FED7F56D4CE5ULL, "vxdr_primevaltomb_6z8_0_1.vlm"},
	{0xB2A150B8DFC859AFULL, "llac_w_woundedcoast_-131068_0.dds"},
	{0xB2A1609D28A47469ULL, "w_encounter2_26_0.phy"},
	{0xB2A2C27271B8C004ULL, "vxw_sundermount_310_0_1.vlm"},
	{0xB2A30098804834C6ULL, "lgt311_shw_mer_mettin6156473.cl"},
	{0xB2A3881C2DE5216FULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_c.impa.evt"},
	{0xB2A3887FADB8D879ULL, "hlac_ht_chantry_n_2534_45x.dds"},
	{0xB2A39B263C895D25ULL, "art\\vfx\\cinematic\\vfx_orsino_blood_p.dura.evt"},
	{0xB2A3A4E2EB853DFFULL, "dae300_alistair_king6133580.cl"},
	{0xB2A3B883F0E2DA64ULL, "6119389_f.fxe"},
	{0xB2A3D5A41E66EF84ULL, "hlac_ht_mansion2b_818_54y.dds"},
	{0xB2A42806F141AFF4ULL, "lt_0_lowtown_docks_n_136_0.phy"},
	{0xB2A43BE91E0280DDULL, "m_dae250ar_sundermount.bnk"},
	{0xB2A442FB6BA280F2ULL, "6146854_m.fxe"},
	{0xB2A4BB16CA41A1B0ULL, "w_blightlands_200_0.mao"},
	{0xB2A4BF16CA41A8C3ULL, "w_blightlands_200_0.mmh"},
	{0xB2A4CD16CA41C01DULL, "w_blightlands_200_0.msh"},
	{0xB2A538AC67FC47F6ULL, "hlac_lt_undercityencounter_-1_48x.dds"},
	{0xB2A574D40C1C49A3ULL, "art\\vfx\\textures\\vfx_d_heat.dds"},
	{0xB2A5F4A0B2643B15ULL, "vxca_caverns2prc_8cz_0_1.vlm"},
	{0xB2A5FB3F2CA05378ULL, "772678786.ogg"},
	{0xB2A6008E90B63C8CULL, "hlac_ca_cavern1_haunted_24422_5y.dds"},
	{0xB2A61B6DC7C74209ULL, "mcr371tr_trapped_fix.ncs"},
	{0xB2A63DCDA474BE62ULL, "mcr371_arthuris6087096.cl"},
	{0xB2A6B605263109A9ULL, "isa211_isabela6202066.cl"},
	{0xB2A6BDED04A09BDBULL, "vxw_blightlands_iw_0_0.vlm"},
	{0xB2A6D46FB5BF6882ULL, "hlac_lt_gallowstemplar_2394_8x.dds"},
	{0xB2A74E3A91D07801ULL, "zz_vault_debug6200301.cl"},
	{0xB2A7D5CAEF112887ULL, "vxlt_0_lowtown_a2_n_euw_0_0.vlm"},
	{0xB2A7F7075A510663ULL, "llac_lt_0_lowtown_a2_2_0.dds"},
	{0xB2A80BFBEABFFA2EULL, "cd_qcr221cs_keep_intro.wbd"},
	{0xB2A8699D28AAA265ULL, "w_encounter2_26_0.rml"},
	{0xB2A8E9BFD1E5BC1AULL, "gen00fl_merrill6130125.cl"},
	{0xB2A9060EDD8EC9EDULL, "36698304.ogg"},
	{0xB2A928E95445EAE4ULL, "6076948_m.fxe"},
	{0xB2A9891C2DE98E18ULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_c.impa.gad"},
	{0xB2A99576D13F2079ULL, "fen210_dying_magister.bnk"},
	{0xB2A9F350AD662A4FULL, "6080691_f.fxe"},
	{0xB2AA348BFBD3C07BULL, "vxdr_primevaltomb_7qk_0_0.vlm"},
	{0xB2AA4370BAC31546ULL, "6126913_m.fxe"},
	{0xB2AA83A6580185D9ULL, "6208602_f.fxe"},
	{0xB2AAF95109BEC783ULL, "6140509_m.fxe"},
	{0xB2AB5091129749C1ULL, "hm_thug02.mor"},
	{0xB2AC14D4C2BFC285ULL, "vxlt_gallowscourtyard_7ck_0_0.vlm"},
	{0xB2AC51197DE99999ULL, "736656996.ogg"},
	{0xB2AC581129B6D6D6ULL, "art\\levels\\races\\design\\prp_metaltrp.mao"},
	{0xB2AC641129B6EB75ULL, "art\\levels\\races\\design\\prp_metaltrp.mmh"},
	{0xB2AC661129B6EEEBULL, "art\\levels\\races\\design\\prp_metaltrp.msh"},
	{0xB2ACB09B8979A415ULL, "mag100_thrask6032000.cl"},
	{0xB2ACB3CAD8C2EFB7ULL, "vxca_caverns1_mines_cmw_0_0.vlm"},
	{0xB2ACB776D1418FD9ULL, "fen210_dying_magister.cnv"},
	{0xB2ACB9F35899B54DULL, "llac_lt_0_lowtown_a2_n_-1_1.dds"},
	{0xB2ACCC903F296C84ULL, "vxlt_0_lowtown_a2_n_gb7_0_0.vlm"},
	{0xB2AD07C05CBF0049ULL, "3172750.ogg"},
	{0xB2AE0506F149C8F1ULL, "lt_0_lowtown_docks_n_136_0.msh"},
	{0xB2AE1306F149E00CULL, "lt_0_lowtown_docks_n_136_0.mao"},
	{0xB2AE1F06F149F4AFULL, "lt_0_lowtown_docks_n_136_0.mmh"},
	{0xB2AE2AFDC9460371ULL, "174292534.ogg"},
	{0xB2AE3FE2F5E467A6ULL, "hlac_ca_caverns2prc_5802_15z.dds"},
	{0xB2AF70412924195EULL, "vxdr_primevaltomb_1cc_0_1.vlm"},
	{0xB2B05A8C599D792EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1_0.pwk"},
	{0xB2B05E03D7523978ULL, "958782144.ogg"},
	{0xB2B05F31F47E4113ULL, "483859424.ogg"},
	{0xB2B0D6393F33FE9BULL, "36521332.ogg"},
	{0xB2B0DF4E983119EEULL, "6153760_m.fxe"},
	{0xB2B0FA42272AE194ULL, "6120114_f.fxe"},
	{0xB2B20F158B4F6887ULL, "vxca_caverns1_mines_i9m_0_0.vlm"},
	{0xB2B2164BE623203FULL, "6153259_m.fxe"},
	{0xB2B21B9E7D06D36BULL, "vxht_mansion2_3bw_0_0.vlm"},
	{0xB2B23CA4A02CD869ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv7.ani"},
	{0xB2B2B4D85D989972ULL, "6143754_m.fxe"},
	{0xB2B2B974B2BB06D8ULL, "1013567588.ogg"},
	{0xB2B2BC1D627DE73EULL, "6116428_m.fxe"},
	{0xB2B2D6DA84FD3026ULL, "ave100_donnic6105937.cl"},
	{0xB2B2E9259A782E73ULL, "277872712.ogg"},
	{0xB2B33A3384A7E9E0ULL, "vxw_bonepitdestroyed_a3_1k2_0_0.vlm"},
	{0xB2B37E7B6526C4DBULL, "6076165_m.fxe"},
	{0xB2B3968AC938156EULL, "6166603_m.fxe"},
	{0xB2B39ED28AB4D60EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal_3.mao"},
	{0xB2B3A64FB888996AULL, "634696888.ogg"},
	{0xB2B455870AD91C8EULL, "vxlt_gallowstemplar_f_2r7_0_0.vlm"},
	{0xB2B468F1C2D3DA68ULL, "zz_firstshot.cut"},
	{0xB2B46DCDD40C48C3ULL, "hlac_lt_gallowstemplar_f_2395_1z.dds"},
	{0xB2B52FC9F696B45BULL, "vxw_blightlands_rv_0_0.vlm"},
	{0xB2B55CE52EEF08E9ULL, "pro000ar_gallows_court.ncs"},
	{0xB2B57C3FD960ADBBULL, "vxlt_undercityserialkiller_17k_0_0.vlm"},
	{0xB2B7046DC7D58EECULL, "mcr371tr_trapped_fix.utt"},
	{0xB2B7450B20F4643EULL, "art\\characters\\creatures\\mubari\\animation\\dmg_v1.ani"},
	{0xB2B7DD23667F7038ULL, "qun221_viscount6070790.cl"},
	{0xB2B805ED290EA251ULL, "vxw_bonepitdestroyed_a3_230_0_0.vlm"},
	{0xB2B83ACCB985B986ULL, "dae000_sandal6147774.cl"},
	{0xB2B87ED3E32ADAFFULL, "hlac_lt_gallowstemplar_2383_3x.dds"},
	{0xB2B9328D961DC131ULL, "hlac_lt_0_lowtown_a2_n_c_65535_5x.dds"},
	{0xB2B957A4A03324EBULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv7.gad"},
	{0xB2B95D775E02FC64ULL, "pro000cr_friedrich.utc"},
	{0xB2B9611129C17002ULL, "art\\levels\\races\\design\\prp_metaltrp.anb"},
	{0xB2B9E30970E7494DULL, "6104994_m.fxe"},
	{0xB2BA86B1B4B5DB05ULL, "vxlt_gallowscourtyard_7dj_0_0.vlm"},
	{0xB2BA9C0202AA7F62ULL, "840507457.ogg"},
	{0xB2BB1E5BB3EB036EULL, "6089430_m.fxe"},
	{0xB2BB6E961EFAD9AEULL, "w_blightlands_191_0.mmh"},
	{0xB2BB7A961EFAEE45ULL, "w_blightlands_191_0.mao"},
	{0xB2BB88961EFB059CULL, "w_blightlands_191_0.msh"},
	{0xB2BBD580DEA183C9ULL, "6148136_m.fxe"},
	{0xB2BBE1F8D9A52754ULL, "ch1notnt_ds2.psh.fxo"},
	{0xB2BBFD2C8ECE9E77ULL, "vxlt_undercityserialkiller_18b_0_0.vlm"},
	{0xB2BBFD7865052F15ULL, "ht_0_hightown_a2_52_0_ll.mmh"},
	{0xB2BC401E89A622B5ULL, "zz_dre_debug6036251.cl"},
	{0xB2BC5B5304E4C511ULL, "gen00fl_isabela6101072.cl"},
	{0xB2BC8BDA0BCEA80CULL, "vxlt_warehouse_n_rl_0_0.vlm"},
	{0xB2BCD28AF3C88086ULL, "6100247_m.fxe"},
	{0xB2BD41D72E89D269ULL, "6172865_m.fxe"},
	{0xB2BDA73459A54F79ULL, "99410938.ogg"},
	{0xB2BE075694009E5DULL, "vxw_bonepitdestroyed_a3_2aj_0_0.vlm"},
	{0xB2BE1A39B7DCEFECULL, "hla_ht_0_hightown_2_0z.dds"},
	{0xB2BE2267CF8B4442ULL, "6087212_m.fxe"},
	{0xB2BFBE7D3CFF3B20ULL, "vxht_mansion2b_1a7_0_0.vlm"},
	{0xB2BFEC5966192990ULL, "135593833.ogg"},
	{0xB2C054A4A0393E50ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv7.evt"},
	{0xB2C0CD4277016A46ULL, "mcr370_templar16086973.cl"},
	{0xB2C144DC6F5B8C7BULL, "vxca_cave_235_0_0.vlm"},
	{0xB2C1A3111BA1D778ULL, "vxw_bonepitdestroyed_a3_1pw_0_0.vlm"},
	{0xB2C1AC10769E0260ULL, "vxw_bonepitdestroyed_a3_1l3_0_0.vlm"},
	{0xB2C2F4AAD4E1EC09ULL, "plt_vrc100pt_n_personal.plo"},
	{0xB2C337E2CAB57B75ULL, "6219449_m.fxe"},
	{0xB2C391B11D61B6B4ULL, "vxht_0_hightown_5w5_0_0.vlm"},
	{0xB2C3A1A6E88CCCDBULL, "vxw_blightlands_wi_0_0.vlm"},
	{0xB2C411691116170EULL, "553591631.ogg"},
	{0xB2C4275605690F1FULL, "llac_ca_caverns1_mines_24426_1.dds"},
	{0xB2C45D0B20FF170FULL, "art\\characters\\creatures\\mubari\\animation\\dmg_v1.evt"},
	{0xB2C46714501F2311ULL, "vxht_mansion2_2f4_0_0.vlm"},
	{0xB2C473C7A7BF4A23ULL, "201112413.ogg"},
	{0xB2C49AC27419DB72ULL, "6164087_m.fxe"},
	{0xB2C4DB69EF373607ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_ready.gad"},
	{0xB2C4F4D12AA217AEULL, "ht_playershouse_n_probe_1kp.dds"},
	{0xB2C51DBD688BF98CULL, "art\\vfx\\combat\\vfx_m_spi_frc_cast_c.impa.ani"},
	{0xB2C59C45E0EE4E01ULL, "6202210_m.fxe"},
	{0xB2C60E07DF0BB210ULL, "mag210_arianni6031711.cl"},
	{0xB2C61F46207E2158ULL, "hlac_ca_caverns2prc_5802_37x.dds"},
	{0xB2C6243F665AB27CULL, "vxlt_undercityencounter_1gd_0_0.vlm"},
	{0xB2C652A8D3C09000ULL, "6096425_m.fxe"},
	{0xB2C6B0CC48916224ULL, "mag100_cullen6095141.cl"},
	{0xB2C6C6C7772AD0A0ULL, "352654377.ogg"},
	{0xB2C6CCDAABA75F59ULL, "hlac_w_encounter3_-196603_0x.dds"},
	{0xB2C705D8D2DAD83DULL, "6202200_f.fxe"},
	{0xB2C78552E792668CULL, "6097879_m.fxe"},
	{0xB2C8655232C4104DULL, "follower_barks6154155.cl"},
	{0xB2C871D40B301CC5ULL, "6151462_m.fxe"},
	{0xB2C8E0B0940B2AABULL, "hlac_ht_0_hightown_-65537_2y.dds"},
	{0xB2C8E6BB78C0F81FULL, "6093497_f.fxe"},
	{0xB2C90414E36AE623ULL, "6083751_cleric_decidefate.cut"},
	{0xB2C98A51DC11179EULL, "tbe100im_mystic_boot.uti"},
	{0xB2C98EC816EC8335ULL, "vxw_sundermount_2pn_0_0.vlm"},
	{0xB2CA185067ED3DF4ULL, "6120140_m.fxe"},
	{0xB2CA5EB710CF2BABULL, "6165745_f.fxe"},
	{0xB2CAB3BCC8EEB5F5ULL, "vxlt_warehouse_ck_0_0.vlm"},
	{0xB2CAFDB6FDC4C08CULL, "vxlt_warehouse_n_ym_0_0.vlm"},
	{0xB2CB315A8C444242ULL, "hlac_lt_darktown_-1_19x.dds"},
	{0xB2CB3B93A714CF9DULL, "llac_lt_warehouse_1_5.dds"},
	{0xB2CB4ABB3695DAA4ULL, "6090864_f.fxe"},
	{0xB2CB5E0B210537B8ULL, "art\\characters\\creatures\\mubari\\animation\\dmg_v1.gad"},
	{0xB2CBB335ED5D9F37ULL, "w_bonepitdestroyed_a3_146_0.mmh"},
	{0xB2CBB735ED5DA644ULL, "w_bonepitdestroyed_a3_146_0.mao"},
	{0xB2CBC935ED5DC4A9ULL, "w_bonepitdestroyed_a3_146_0.msh"},
	{0xB2CBD069EF3D4214ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_ready.evt"},
	{0xB2CBEB7A7E180064ULL, "dae000ip_crowd_mf3_talk14.utp"},
	{0xB2CC0B7CC97042BBULL, "439535466.ogg"},
	{0xB2CC2ABD68922ECEULL, "art\\vfx\\combat\\vfx_m_spi_frc_cast_c.impa.gad"},
	{0xB2CCCACD79016380ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potplant_n.dds"},
	{0xB2CCD609024841C8ULL, "6152813_m.fxe"},
	{0xB2CCE44984097946ULL, "6174604_m.fxe"},
	{0xB2CD499D28C9878EULL, "w_encounter2_26_0.mmh"},
	{0xB2CD559D28C99C25ULL, "w_encounter2_26_0.mao"},
	{0xB2CD639D28C9B47CULL, "w_encounter2_26_0.msh"},
	{0xB2CE305A2EF553A0ULL, "vxht_mansion2b_1d4_0_0.vlm"},
	{0xB2CE6FB12BF2E350ULL, "vxht_keep_n_6hl_0_0.vlm"},
	{0xB2CE890A32A1241CULL, "vrc221_bartrand6065341.cl"},
	{0xB2CE8A764630A503ULL, "gen00fl_varric6078529.cl"},
	{0xB2CEB5F31B16180AULL, "292720284.ogg"},
	{0xB2CF3A2F3AD2EA0BULL, "art\\characters\\creatures\\mubari\\animation\\turnl_loop.evt"},
	{0xB2CF3AE53BEA3EFEULL, "art\\vfx\\creatures\\fxc_arch_dblt_p.impa.evt"},
	{0xB2CF4844F557D0B9ULL, "hlac_lt_0_lowtown_a2_n_-65536_4y.dds"},
	{0xB2CF773B4F2AC535ULL, "6154717_m.fxe"},
	{0xB2CFA50961DC059CULL, "vxlt_0_lowtown_a2_n_c_gcb_0_0.vlm"},
	{0xB2CFB6B96151A4FBULL, "vxca_cave_244_0_0.vlm"},
	{0xB2CFD629113E9A10ULL, "gen00fl_aveline6139402.cl"},
	{0xB2D00D015A4E2282ULL, "ht_0_hightown_a3_12_0.rml"},
	{0xB2D056DA4FF3173EULL, "vxw_woundedcoast_ap_0_0.vlm"},
	{0xB2D0A3320275A522ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_8x.dds"},
	{0xB2D0B65C32643288ULL, "6167912_m.fxe"},
	{0xB2D184E8BDFE01F3ULL, "vxht_keep_5gb_0_0.vlm"},
	{0xB2D1D0CBF07BA637ULL, "vxht_0_hightown_a6m_0_0.vlm"},
	{0xB2D2939FB2D3C408ULL, "lgt320cr_shw_ors_scrt_thug.utc"},
	{0xB2D2E0EEDE68FBD6ULL, "902600322.ogg"},
	{0xB2D305CAC46792FDULL, "6198678_m.fxe"},
	{0xB2D3148BE6E5A2A8ULL, "waterflowing0.vsh.fxo"},
	{0xB2D31B69EE73DA68ULL, "vxht_mansion2_29l_0_0.vlm"},
	{0xB2D325BD689844ADULL, "art\\vfx\\combat\\vfx_m_spi_frc_cast_c.impa.evt"},
	{0xB2D4ABD8D11656ACULL, "6102889_m.fxe"},
	{0xB2D501FBEAE6D75CULL, "cd_qcr221cs_keep_intro.bnk"},
	{0xB2D506699C7CC3A9ULL, "vxht_bartrand_n_5w0_0_0.vlm"},
	{0xB2D5D13E2784A1F2ULL, "351263730.ogg"},
	{0xB2D5E2D377A79D8CULL, "art\\characters\\playercharacter\\humanmale\\hoo_robd.msh"},
	{0xB2D5E8D377A7A87EULL, "art\\characters\\playercharacter\\humanmale\\hoo_robd.mmh"},
	{0xB2D62D3898457EEDULL, "art\\characters\\creatures\\corpse\\uni_corm.phy"},
	{0xB2D651E53BF08429ULL, "art\\vfx\\creatures\\fxc_arch_dblt_p.impa.gad"},
	{0xB2D6532F3AD933CCULL, "art\\characters\\creatures\\mubari\\animation\\turnl_loop.gad"},
	{0xB2D66F22E3CAC1C0ULL, "zz_ave_debug6134027.cl"},
	{0xB2D6C8C20B465F02ULL, "hlac_lt_gallowstemplar_f_2391_2z.dds"},
	{0xB2D7DB7CE23B2BD0ULL, "vxlt_warehouse_9x_0_0.vlm"},
	{0xB2D7FD6D5731EEE1ULL, "222924411.ogg"},
	{0xB2D800A508E29BB5ULL, "vxw_sundermount_2um_0_0.vlm"},
	{0xB2D8206AFB410E83ULL, "zz_dog_debug6127916.cl"},
	{0xB2D8450413E88E72ULL, "gen00fl_sebastian6180486.cl"},
	{0xB2D87CB0A0306D4DULL, "142178730.ogg"},
	{0xB2D90BC57095FE11ULL, "follower_banter6172732.cl"},
	{0xB2D9A905E0745CF2ULL, "vxlt_0_lowtown_docks_n_b4c_0_0.vlm"},
	{0xB2D9A983093DF0EBULL, "980361626.ogg"},
	{0xB2D9C869EF49724DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_ready.ani"},
	{0xB2DA10241E791F40ULL, "vxca_caverns2_8bl_0_0.vlm"},
	{0xB2DA3965A52F2430ULL, "art\\characters\\weapons\\longswords\\w_lsw_012a.msh"},
	{0xB2DA4765A52F3C4AULL, "art\\characters\\weapons\\longswords\\w_lsw_012a.mmh"},
	{0xB2DA4B65A52F4281ULL, "art\\characters\\weapons\\longswords\\w_lsw_012a.mao"},
	{0xB2DACCE27CDA6BECULL, "follower_banter6170579.cl"},
	{0xB2DB2B31E55D971EULL, "959700512.ogg"},
	{0xB2DBBE73B582FEDDULL, "w_bonepitmining_a2_156_0.phy"},
	{0xB2DC5A7ECE9BDA70ULL, "57063669.ogg"},
	{0xB2DC60DE10814D44ULL, "6121997_f.fxe"},
	{0xB2DC67669C8B9BCFULL, "atl_guiachieve_dxt5_00.dds"},
	{0xB2DC7E0BB9006F5FULL, "6106320_f.fxe"},
	{0xB2DC8A845D4F9033ULL, "m_dae111ar_hanged_man.bnk"},
	{0xB2DCE18E1DE8FBD0ULL, "vxht_keep_n_6cm_0_0.vlm"},
	{0xB2DD1CB48DD9CBA2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexcity_d.dds"},
	{0xB2DD22E53BF65377ULL, "art\\vfx\\creatures\\fxc_arch_dblt_p.impa.ani"},
	{0xB2DD522F3ADF50CAULL, "art\\characters\\creatures\\mubari\\animation\\turnl_loop.ani"},
	{0xB2DD833033F6C426ULL, "hlac_w_sundermount_196603_0z.dds"},
	{0xB2DDAFF483A5809FULL, "vxlt_hangedman_n_24y_0_0.vlm"},
	{0xB2DE1B0847F6EC24ULL, "6106870_m.fxe"},
	{0xB2DFEE2BDC8FA90AULL, "vxht_mansion2b_2op_0_0.vlm"},
	{0xB2DFF6C5AFF41A73ULL, "vxht_keep_5je_0_0.vlm"},
	{0xB2E0181F43E4EEC8ULL, "47143227.ogg"},
	{0xB2E0235A8A869EEDULL, "83128104.ogg"},
	{0xB2E026E6880A7764ULL, "vxw_bonepitdestroyed_a3_80_0_0.vlm"},
	{0xB2E065F099ADF37AULL, "art\\vfx\\creatures\\fxc_mrtorch_c.anb"},
	{0xB2E0CB31A2B9FC38ULL, "one120_athenril6113286.cl"},
	{0xB2E14EC112BA1476ULL, "6062345_m.fxe"},
	{0xB2E15C03FC252679ULL, "vxlt_house_fg_0_0.vlm"},
	{0xB2E1C33B14EF9C1FULL, "narrative_varric6172489.cl"},
	{0xB2E1D71062E8D5FEULL, "zz_approval_debug6003696.cl"},
	{0xB2E207153CC249FEULL, "269949559.ogg"},
	{0xB2E3611968F1BB44ULL, "210688102.ogg"},
	{0xB2E378468E72DC29ULL, "vxht_bartrand_n_5z7_0_0.vlm"},
	{0xB2E46B4839C122E3ULL, "seb221_flora_harimann6177704.cl"},
	{0xB2E47FB4DEF1518CULL, "llac_ca_caverns1_mines_24424_11.dds"},
	{0xB2E641E27C2B7E63ULL, "vxlt_foundry_n_zq_0_0.vlm"},
	{0xB2E6457FA11BAA0CULL, "art\\cinematics\\additional animation\\cs_malenobleii_cinematicanim.ani"},
	{0xB2E65B12D3F67011ULL, "vxht_0_hightown_a3_anj_0_0.vlm"},
	{0xB2E6845E56541501ULL, "6119299_m.fxe"},
	{0xB2E6E259623CF6EAULL, "t1_skn_h11.tnt"},
	{0xB2E7343546044187ULL, "6088317_m.fxe"},
	{0xB2E7ECE530D38AF0ULL, "52405585.ogg"},
	{0xB2E813F1E8682C41ULL, "and110st_refugee.stg"},
	{0xB2E92BA43B4E4E58ULL, "gen_im_wep_rog_arc_act3_06.uti"},
	{0xB2E94E756F12D6A6ULL, "vxlt_hangedman_n_213_0_0.vlm"},
	{0xB2E9768723965E6EULL, "vxlt_blackemporium_w0_0_0.vlm"},
	{0xB2E99C35E626E86CULL, "hlac_ht_brothel_n_3364_0y.dds"},
	{0xB2E9E7087BC0250FULL, "art\\characters\\creatures\\spider\\animation\\tal_spit.gad"},
	{0xB2E9E933BFE559C1ULL, "6034709_m.fxe"},
	{0xB2EA023B5DDA0FADULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf03.mmh"},
	{0xB2EA143B5DDA2E33ULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf03.msh"},
	{0xB2EA2AE748137676ULL, "gen00fl_aveline6132791.cl"},
	{0xB2EAD05DBE184F16ULL, "vxlt_0_lowtown_a2_n_c_hjp_0_0.vlm"},
	{0xB2EAEA096A5882E7ULL, "ht_0_hightown_a2_29_0_ll.mmh"},
	{0xB2EB1AC22FE288E7ULL, "vxht_0_hightown_c8q_0_0.vlm"},
	{0xB2EB22CB77AB98C3ULL, "art\\characters\\creatures\\buzzard\\c_buzzarda.msh"},
	{0xB2EB30CB77ABB01DULL, "art\\characters\\creatures\\buzzard\\c_buzzarda.mmh"},
	{0xB2EB3CC9535572E8ULL, "t1_mub_te3.tnt"},
	{0xB2EC32815394D998ULL, "gen_im_trs_gen_amu_r2.uti"},
	{0xB2EC33F380BAD849ULL, "gen_im_trs_rdr_hvy_hlm_r1.uti"},
	{0xB2EC6781B9C1882AULL, "hightown_male_noble_gossip6171394.cl"},
	{0xB2EC8D31EFB3175CULL, "dre000_stroud6036697.cl"},
	{0xB2ECB373B5915F49ULL, "w_bonepitmining_a2_156_0.mao"},
	{0xB2ECBF73B59173E2ULL, "w_bonepitmining_a2_156_0.mmh"},
	{0xB2ECC173B5917698ULL, "w_bonepitmining_a2_156_0.msh"},
	{0xB2ECD7B622B54123ULL, "822560594.ogg"},
	{0xB2ED03E65593B16EULL, "hlac_ca_caverns1_24426_1z.dds"},
	{0xB2ED1C3496B3840AULL, "164150018.ogg"},
	{0xB2ED40EA46C98B65ULL, "vxht_0_hightown_a2_7j1_0_0.vlm"},
	{0xB2ED44FAB78F5144ULL, "hlac_lt_0_lowtown_131073_1y.dds"},
	{0xB2ED527FA121DF4EULL, "art\\cinematics\\additional animation\\cs_malenobleii_cinematicanim.gad"},
	{0xB2ED6DBB3E477195ULL, "544844220.ogg"},
	{0xB2ED9102A769D6D5ULL, "gen00fl_varric6063839.cl"},
	{0xB2EDA52A8DEAA0C1ULL, "6167878_m.fxe"},
	{0xB2EE147D87961E1EULL, "art\\vfx\\combat\\material__13.mao"},
	{0xB2EE3E4A7EBA09F6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv4.evt"},
	{0xB2EE4F0A2A24EDEAULL, "gen_im_cns_cft_pot_04.uti"},
	{0xB2EE6008CE85C18AULL, "vxht_mansion2b_2hq_0_0.vlm"},
	{0xB2EF118FE75ABCA2ULL, "446487584.ogg"},
	{0xB2EF7C60C200B6E2ULL, "6094991_m.fxe"},
	{0xB2EFAE36BC91E456ULL, "vxlt_0_lowtown_a2_n_c_gz3_0_0.vlm"},
	{0xB2EFFAA298098905ULL, "vxw_bonepitdestroyed_a3_1pn_0_0.vlm"},
	{0xB2EFFC087BC4B4BCULL, "art\\characters\\creatures\\spider\\animation\\tal_spit.evt"},
	{0xB2F02FAB3F3DB50AULL, "369477299.ogg"},
	{0xB2F057C008814DBEULL, "vxw_bonepit_1wx_0_0.vlm"},
	{0xB2F09F8F3CDA63DFULL, "6153031_m.fxe"},
	{0xB2F13B40273B9D81ULL, "6035284_f.fxe"},
	{0xB2F154B5A3137A8DULL, "501817768.ogg"},
	{0xB2F1A52AF32A716BULL, "framebuffereffects_illuminationmask_p0.psh.fxo.dx11.assembly"},
	{0xB2F1F765669135CCULL, "hlac_ht_keep_n_3735_1y.dds"},
	{0xB2F4354A7EBE6581ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv4.gad"},
	{0xB2F4A8385039CCE3ULL, "41953928.ogg"},
	{0xB2F4CCEFC5EC8891ULL, "vxht_0_hightown_a3_aki_0_0.vlm"},
	{0xB2F4F25187A14DBCULL, "follower_barks6153261.cl"},
	{0xB2F5CA1B8BF854AEULL, "vxdr_primevaltomb_5bi_0_0.vlm"},
	{0xB2F609539ED0FF75ULL, "vxlt_warehouse_rh_0_0.vlm"},
	{0xB2F61C0BF8EB9963ULL, "6133796_m.fxe"},
	{0xB2F62BE2AE5FAD69ULL, "624492886.ogg"},
	{0xB2F69A024ED0DFB5ULL, "6084427_m.fxe"},
	{0xB2F6F595C6448BF9ULL, "hlac_ht_mansion_n_3374_0z.dds"},
	{0xB2F75D947AA3919BULL, "vxht_keep_6f6_0_0.vlm"},
	{0xB2F76D6A81BDC28EULL, "vxht_0_hightown_a3_a2u_0_0.vlm"},
	{0xB2F8050265441DF1ULL, "vxdr_0_deeproads_cmw_0_0.vlm"},
	{0xB2F80BC91A3B4A63ULL, "vxlt_gallowstemplar_f_5qx_0_0.vlm"},
	{0xB2F89AA40702BD36ULL, "zz_followers_banter6157284.cl"},
	{0xB2F8A1124009947BULL, "870149271.ogg"},
	{0xB2F9526ED25039D0ULL, "6139638_m.fxe"},
	{0xB2F957CD354CB64FULL, "6148431_m.fxe"},
	{0xB2F9DF5F3E6A814BULL, "x_partypicker_mmi.gff"},
	{0xB2F9E914F4753E15ULL, "ico_potion_injurykit.dds"},
	{0xB2F9FF7CEDF93561ULL, "494860145.ogg"},
	{0xB2FA23C0E6FD4D6EULL, "llac_ht_chantry_2534_12.dds"},
	{0xB2FA272D1215844CULL, "336497408.ogg"},
	{0xB2FAFB3B5DE8776AULL, "art\\characters\\playercharacter\\humanfemale\\mt_aet_hf03.phy"},
	{0xB2FB02ABB6C247E5ULL, "6118168_m.fxe"},
	{0xB2FB364A7EC4862FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv4.ani"},
	{0xB2FB6A84E37A3978ULL, "vxw_bonepitdestroyed_a3_1ds_0_0.vlm"},
	{0xB2FBAC7125D560BEULL, "vxw_woundedcoast_ps_0_0.vlm"},
	{0xB2FBDA7BFD2BD757ULL, "6049651_m.fxe"},
	{0xB2FBEED86A531868ULL, "static_static_p85.psh.fxo.dx11"},
	{0xB2FBFCAF8C294EC8ULL, "451309978.ogg"},
	{0xB2FC2E80C68F3D65ULL, "708810634.ogg"},
	{0xB2FC5EE3BF3C116FULL, "vxca_caverns2prc_9q6_0_0.vlm"},
	{0xB2FC6D4E4D4E3A21ULL, "6128552_m.fxe"},
	{0xB2FC6F0E5938DBE4ULL, "vxw_sundermount_37k_0_0.vlm"},
	{0xB2FC9522862793D3ULL, "38349082.ogg"},
	{0xB2FCCD3A3EF6C0C0ULL, "623325854.ogg"},
	{0xB2FCD1E5C07BDA0AULL, "vxht_mansion2b_2er_0_0.vlm"},
	{0xB2FD8AB5ED9D647DULL, "gui_guipixelshader_p3.psh.fxo.dx11"},
	{0xB2FE04087BD0FFD5ULL, "art\\characters\\creatures\\spider\\animation\\tal_spit.ani"},
	{0xB2FE1D2B6015E951ULL, "vxdr_primevaltomb_900_0_0.vlm"},
	{0xB2FE3A8274E0E559ULL, "lgt210_magu_selby6142054.cl"},
	{0xB2FE6C7F89FFA185ULL, "vxw_bonepitdestroyed_a3_1um_0_0.vlm"},
	{0xB2FEC7816E7D0987ULL, "lgt220_shp_mag_guard6149905.cl"},
	{0xB2FEEDF44F661D6FULL, "hlac_w_encounter1_65535_0x.dds"},
	{0xB2FF116C2ED07C5FULL, "6156331_m.fxe"},
	{0xB2FF9CA2E328F647ULL, "gen00fl_carver6097421.cl"},
	{0xB2FFD865A9A479BDULL, "267799791.ogg"},
	{0xB3003F1F2F7E247FULL, "6066353_m.fxe"},
	{0xB3004282993EF807ULL, "709987191.ogg"},
	{0xB3014CA4EF0ACF00ULL, "895881980.ogg"},
	{0xB3016067837F987FULL, "6107096_m.fxe"},
	{0xB301B81E4E6A37C1ULL, "14299138.ogg"},
	{0xB301B9BCFC6580B2ULL, "rr_pro000cs_opening_pc.cut"},
	{0xB302540EC9C55E16ULL, "lt_0_lowtown_docks_a3_138_0.phy"},
	{0xB302D5EC349B68F1ULL, "vxht_chantry_2iq_0_0.vlm"},
	{0xB302DAAC3EB1FC5BULL, "mcr370_meredith6122889.cl"},
	{0xB30355A9678FA87FULL, "plt_fen100pt_bait.nss"},
	{0xB30368183A404601ULL, "6113499_m.fxe"},
	{0xB303841B74207EC4ULL, "vxlt_0_lowtown_docks_a3_n_8ug_0_0.vlm"},
	{0xB3047B3090C717F5ULL, "vxlt_warehouse_ww_0_0.vlm"},
	{0xB3052A6ABC39BC20ULL, "845605444.ogg"},
	{0xB3056FF7E95BA389ULL, "hm_qmq_patronm3.mor"},
	{0xB3058291750C1F60ULL, "seb221_ruxton_harimann6177194.cl"},
	{0xB3058E5D6A0C2A99ULL, "vxlt_gallowstemplar_f_5p5_0_0.vlm"},
	{0xB305A68E5A2E61ABULL, "vxw_sundermount_308_0_1.vlm"},
	{0xB305EFA04652EB76ULL, "48711379.ogg"},
	{0xB3061FD9D786DDA7ULL, "6035317_m.fxe"},
	{0xB306417B0C5045C0ULL, "dae000_flemeth6001005.cl"},
	{0xB306533C39E0B983ULL, "vxht_mansion2_1wr_0_0.vlm"},
	{0xB30682E2FF0FF386ULL, "6070099_f.fxe"},
	{0xB307312808E5D211ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaricommander.anb"},
	{0xB3075222A6EAC667ULL, "ht_0_hightown_n.rml"},
	{0xB307BA020201F7DBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15b.msh"},
	{0xB307D80202022AC5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15b.mmh"},
	{0xB307E2E598959A62ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxa_1.mao"},
	{0xB307E51D161FF503ULL, "t3_eye_lbl.tnt"},
	{0xB307E9D320B32624ULL, "t1_mub_bk3.tnt"},
	{0xB307F4EA5DD3CB7BULL, "vxht_brothel_n_288_0_0.vlm"},
	{0xB307FE7C13CEB9E7ULL, "vxht_0_hightown_c2s_0_0.vlm"},
	{0xB308134179F2B1A9ULL, "349625523.ogg"},
	{0xB30820B262D60CBCULL, "art\\vfx\\combat\\vfx_w_wns_front_close1_c.impa.ani"},
	{0xB3082ADB6F65B0CDULL, "vxht_brothel_n_6bl_0_0.vlm"},
	{0xB3082C957E3C3DB7ULL, "clf260_harley6093824.cl"},
	{0xB30830DDD5DC4D6DULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_hed_0d.dds"},
	{0xB30900DB6DF5D8CFULL, "477217876.ogg"},
	{0xB30968F099D0FD2DULL, "art\\vfx\\creatures\\fxc_mrtorch_c.mmh"},
	{0xB3097488A8B86065ULL, "6117068_m.fxe"},
	{0xB30A0A015A7F967EULL, "ht_0_hightown_a3_12_0.anb"},
	{0xB30A247EEDD03176ULL, "w_woundedcoast_24_0.anb"},
	{0xB30A347FF93DC1A7ULL, "dae000_oldmangossip6175760.cl"},
	{0xB30AE0EB4B2EF464ULL, "vxw_sundermount_32h_0_0.vlm"},
	{0xB30AFD2106CCEE14ULL, "vxca_caverns2_62f_0_0.vlm"},
	{0xB30B06FFCA0B75EAULL, "vxht_bartrand_n_5qu_0_0.vlm"},
	{0xB30B0AFA41B56905ULL, "dae000_light_tombs_boss6179456.cl"},
	{0xB30B1064A2D021FDULL, "6125630_m.fxe"},
	{0xB30B8F48E7DE50DAULL, "demo_weapon_swap.cnv"},
	{0xB30B98C18C3106E1ULL, "gen00fl_aveline6121141.cl"},
	{0xB30BFE97D2A55A38ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\ht_gallowsvista_d.dds"},
	{0xB30C3D0EC9CD8BB7ULL, "lt_0_lowtown_docks_a3_138_0.msh"},
	{0xB30C4F0EC9CDAA1AULL, "lt_0_lowtown_docks_a3_138_0.mao"},
	{0xB30C530EC9CDB129ULL, "lt_0_lowtown_docks_a3_138_0.mmh"},
	{0xB30C89D2DBC1FC55ULL, "453820713.ogg"},
	{0xB30C8EC1F640C4DFULL, "hlac_ht_keep_4519_5z.dds"},
	{0xB30D1963358F90C5ULL, "lt_0_lowtown_a2_n_44_0.anb"},
	{0xB30D7F2729ED2B9AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tlfrbroom_01_0_sa.dds"},
	{0xB30E47A5E6D39B05ULL, "vxca_cavern1_haunted_i2b_0_0.vlm"},
	{0xB30E7D3FFE7557DFULL, "313415147.ogg"},
	{0xB30EAA855CDC19CAULL, "455675411.ogg"},
	{0xB30EB148E7E0C022ULL, "demo_weapon_swap.bnk"},
	{0xB30EC352AC89F963ULL, "6217597_m.fxe"},
	{0xB30EEE05E04FC91DULL, "character_character_p158.psh.fxo.dx11"},
	{0xB30F2DB262DC41FEULL, "art\\vfx\\combat\\vfx_w_wns_front_close1_c.impa.gad"},
	{0xB30FFB283CDD7DA6ULL, "355076020.ogg"},
	{0xB310E5F806965145ULL, "6052745_m.fxe"},
	{0xB3111BF6934CBA4AULL, "vxlt_gallowscourtyard_n_4uz_0_0.vlm"},
	{0xB31147C926918171ULL, "vxht_chantry_2bp_0_0.vlm"},
	{0xB311B54056C0AEDDULL, "vxlt_0_lowtown_a3_n_c_eoe_0_1.vlm"},
	{0xB311DDA4EA0D75B3ULL, "6059311_m.fxe"},
	{0xB31213337C346BF8ULL, "art\\characters\\playercharacter\\humanfemale\\arm_avea.phy"},
	{0xB3122A7A07820E6CULL, "zz_mombie_aftermath.ncs"},
	{0xB3123E5B773B8522ULL, "179372955.ogg"},
	{0xB312A17A963BA23AULL, "vxdr_0_deeproads_cxo_0_0.vlm"},
	{0xB312ED0D82BD3075ULL, "vxlt_warehouse_xv_0_0.vlm"},
	{0xB313255A0171D2C4ULL, "6148199_m.fxe"},
	{0xB313601093DE9194ULL, "6116313_f.fxe"},
	{0xB313E098E0E61DF9ULL, "vxdr_dungeon_21v_0_0.vlm"},
	{0xB3142DCF46300870ULL, "gen00fl_anders6173775.cl"},
	{0xB314514EC9397A46ULL, "6031352_m.fxe"},
	{0xB314E8BC49304EF1ULL, "vxdr_0_deeproads_cgu_0_0.vlm"},
	{0xB315B26C5529C6F0ULL, "vxht_0_hightown_a2_bcz_0_0.vlm"},
	{0xB31608B262E2211DULL, "art\\vfx\\combat\\vfx_w_wns_front_close1_c.impa.evt"},
	{0xB31644E99D3D93A3ULL, "6111378_m.fxe"},
	{0xB316BFD377DEF919ULL, "art\\characters\\playercharacter\\humanmale\\hoo_robd.phy"},
	{0xB316F3A6E4F4392BULL, "vxdr_primevaltomb_7t8_0_0.vlm"},
	{0xB316FE4336873C2AULL, "vxlt_warehouse_n_198_0_0.vlm"},
	{0xB31748A967A03A90ULL, "plt_fen100pt_bait.plo"},
	{0xB31780EF97A914EBULL, "vxlt_0_lowtown_a2_e1p_0_0.vlm"},
	{0xB318217EEDDC693FULL, "w_woundedcoast_24_0.msh"},
	{0xB318277EEDDC73E1ULL, "w_woundedcoast_24_0.mmh"},
	{0xB318337EEDDC8782ULL, "w_woundedcoast_24_0.mao"},
	{0xB31833FAC1D8EE2DULL, "hlac_lt_0_lowtown_-196608_2y.dds"},
	{0xB318A967784AF5ACULL, "805099270.ogg"},
	{0xB318AB5B9884542AULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_spiderweb3_da.dds"},
	{0xB318E210F7BC1934ULL, "vxht_mansion_n_4c3_0_0.vlm"},
	{0xB31978DCBC018E6AULL, "vxht_bartrand_n_5tv_0_0.vlm"},
	{0xB319824194C63A7DULL, "6122730_m.fxe"},
	{0xB3198780BD220011ULL, "hlac_dr_0_deeproads_17639_0y.dds"},
	{0xB319B520605865CDULL, "vxw_blightlands_9a_0_0.vlm"},
	{0xB319CC75A291D555ULL, "lgt320cr_shw_ors_scrt_nob.utc"},
	{0xB319DC9A8CC19E4FULL, "gen00fl_fenris6126367.cl"},
	{0xB31A3935427C04C1ULL, "follower_banter6141613.cl"},
	{0xB31A5303729627CBULL, "6172940_m.fxe"},
	{0xB31A8A50FF37A6CFULL, "451764792.ogg"},
	{0xB31A96F85D6462E4ULL, "lgt120_orz_bln_official6140169.cl"},
	{0xB31BB933EF74D379ULL, "hm_qmq_visguard1.mor"},
	{0xB31BCC5C147F51A8ULL, "lgt000ip_store_epic3_2.utp"},
	{0xB31BE485BFCD2470ULL, "vxlt_hangedman_2ct_0_0.vlm"},
	{0xB31C2ACA854C691DULL, "403180218.ogg"},
	{0xB31C3CA8EACB19EEULL, "6161112_m.fxe"},
	{0xB31C7A943524E348ULL, "mcr310_sibling6076297.cl"},
	{0xB31D0EA60C9F7FA0ULL, "577146208.ogg"},
	{0xB31D189B0FC69A9AULL, "hlac_dr_primevaltomb_659_15x.dds"},
	{0xB31D3319FAE80546ULL, "dae000_brothel_orana6160875.cl"},
	{0xB31D4E42C5B23762ULL, "6144342_m.fxe"},
	{0xB31D87082D2700A8ULL, "dae021_torelin_mob6150219.cl"},
	{0xB31D909EE612505CULL, "gen00fl_varric6093570.cl"},
	{0xB31DFF09DB3E439FULL, "175384937.ogg"},
	{0xB31E47848E40076CULL, "seb221_random_servants6187828.cl"},
	{0xB31E8FB617A16DD4ULL, "120141094.ogg"},
	{0xB31E9C212F19AB73ULL, "6128746_f.fxe"},
	{0xB31EA6CA9145F5FAULL, "6120454_m.fxe"},
	{0xB31F1AF75DF42611ULL, "zz_dae_debug6029331.cl"},
	{0xB31F3BB8E442EBADULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown5.msh"},
	{0xB31F4DB8E4430A33ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown5.mmh"},
	{0xB31FAC2380F5B4B4ULL, "6148366_m.fxe"},
	{0xB31FC0289AF9F43AULL, "dae100cr_gamlen.utc"},
	{0xB32113578831BABAULL, "vxdr_0_deeproads_cuh_0_0.vlm"},
	{0xB32119528C0EC3C8ULL, "character_character_p83.psh.fxo.dx11"},
	{0xB32119F9AD63FAEBULL, "vxlt_gallowstemplar_n_42x_0_0.vlm"},
	{0xB3218156BFC38C79ULL, "gen00fl_carver6006072.cl"},
	{0xB321B8D5D17DE1AAULL, "6109132_m.fxe"},
	{0xB3232CB0EA581E71ULL, "zz_followers_banter6157300.cl"},
	{0xB3238D1B83BCB827ULL, "vxlt_0_lowtown_docks_a3_d0d_0_0.vlm"},
	{0xB323A8845D8BC591ULL, "m_dae111ar_hanged_man.wbd"},
	{0xB3248348E3870D41ULL, "one300_zevran6076417.cl"},
	{0xB32497D185F0B116ULL, "vxlt_0_lowtown_a2_n_c_hvt_0_0.vlm"},
	{0xB3249C9E9F9DC5BEULL, "dae000_orsino6085518.cl"},
	{0xB324F2F0021FCB2DULL, "102709314.ogg"},
	{0xB324F41E56A65E80ULL, "llac_dr_0_deeproads_5842_9.dds"},
	{0xB3257DC2A449809AULL, "w_encounter1_34_0.mmh"},
	{0xB32581C2A44987D1ULL, "w_encounter1_34_0.mao"},
	{0xB3258FC2A4499FE0ULL, "w_encounter1_34_0.msh"},
	{0xB3262D616AF280E2ULL, "11633565.ogg"},
	{0xB32652887C1D8B8DULL, "41142076.ogg"},
	{0xB326584E14F9EDACULL, "vxw_woundedcoast_1h8_0_0.vlm"},
	{0xB3267DA6B042EA72ULL, "vxht_mansion2_1b9_0_1.vlm"},
	{0xB32742BD1DF335EEULL, "vxht_mansion2b_25f_0_0.vlm"},
	{0xB32791C5156590EBULL, "vxw_encounter3_ew_0_0.vlm"},
	{0xB327B47A951E5AEFULL, "vxca_caverns2prc_9f3_0_0.vlm"},
	{0xB327CA0F9794F5F0ULL, "1019266203.ogg"},
	{0xB327CE0D4A8CB7FAULL, "ch2notnt_s2y.psh.fxo"},
	{0xB327D40379DC296FULL, "6180592_m.fxe"},
	{0xB32824FEC643130BULL, "6104985_m.fxe"},
	{0xB328C4E0648C404BULL, "6175840_m.fxe"},
	{0xB328D828DCF1FFB3ULL, "890570273.ogg"},
	{0xB328F7FB40971701ULL, "vxlt_gallowstemplar_f_4sm_0_0.vlm"},
	{0xB3290B27691171BBULL, "vxlt_undercityserialkiller_1oc_0_0.vlm"},
	{0xB3295A93312CDEA3ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v2.evt"},
	{0xB329B6338E7B660FULL, "vxlt_gallowstemplar_f_3ro_0_0.vlm"},
	{0xB329C2165FE1EB05ULL, "vxw_bonepitdestroyed_a3_1dj_0_0.vlm"},
	{0xB32A08764263C8D8ULL, "428615558.ogg"},
	{0xB32A6B1E22593B69ULL, "149943128.ogg"},
	{0xB32A9FC418D71257ULL, "vxw_bonepitdestroyed_a3_27j_0_0.vlm"},
	{0xB32AF4D2A1866E36ULL, "vxlt_0_lowtown_a2_n_eoi_0_1.vlm"},
	{0xB32BCA3B4326B622ULL, "plt_prm000pt_ashes_amulet.plo"},
	{0xB32BCCADBA4D5EAFULL, "dae000_merchant_tailor_hm.bnk"},
	{0xB32CE96C126B7AADULL, "6165313_m.fxe"},
	{0xB32D203820894855ULL, "w_sundermount_88_0_lc.mmh"},
	{0xB32D61C8E7C9F22DULL, "dae000ip_standing_bless_d.utp"},
	{0xB32DAAE6576BDE76ULL, "vxca_caverns1_mines_gj3_0_0.vlm"},
	{0xB32E5F3144EBC1A3ULL, "weapon2_ds2.psh.fxo"},
	{0xB32E7ECA8DAA5D29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_rhand.ani"},
	{0xB32EC15ECDF9A6B3ULL, "6053111_m.fxe"},
	{0xB32ED79B6602F126ULL, "6137199_m.fxe"},
	{0xB32EEEADBA4FCE17ULL, "dae000_merchant_tailor_hm.cnv"},
	{0xB32F336E72CFB3ABULL, "6087323_m.fxe"},
	{0xB32F51822945AE42ULL, "hightown_fem_noble1_sta6169493.cl"},
	{0xB32F947B2F13515EULL, "19560195.ogg"},
	{0xB32FE26AAF56186AULL, "hlac_lt_0_lowtown_docks_a3_n_-3_3y.dds"},
	{0xB32FFD9697E9160DULL, "6156548_m.fxe"},
	{0xB33033933132BB24ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v2.gad"},
	{0xB330412641A0B6A8ULL, "vxca_caverns1_ijz_0_0.vlm"},
	{0xB3306F684EC447F0ULL, "hlac_w_bonepit_65536_0y.dds"},
	{0xB330EF877C222085ULL, "vxlt_0_lowtown_gf1_0_0.vlm"},
	{0xB33116BBCA5A5046ULL, "6206851_m.fxe"},
	{0xB33177A5F34B6EF8ULL, "ht_store_187_lc.mmh"},
	{0xB331963E23B1625DULL, "vxw_bonepitdestroyed_a3_29b_0_0.vlm"},
	{0xB33199F038E9B9A5ULL, "hm_ser_guard.mor"},
	{0xB3320F8A206724D8ULL, "vxlt_foundry_n_ys_0_0.vlm"},
	{0xB332361F44EEA2E1ULL, "6052561_m.fxe"},
	{0xB332DA491D6F77DCULL, "mag161tr_slavers.utt"},
	{0xB332DE662DB6A5A1ULL, "6110079_f.fxe"},
	{0xB332E6312CA550AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ad_transition.gad"},
	{0xB3330C10271B2D1DULL, "730294527.ogg"},
	{0xB3330FA9DF8FB6AFULL, "vxht_keep_70a_0_0.vlm"},
	{0xB33310FD7A786384ULL, "follower_banter6171936.cl"},
	{0xB333504FD843ECDAULL, "vxht_0_hightown_bzw_0_0.vlm"},
	{0xB33380724547FA4DULL, "vxht_brothel_n_6ui_0_0.vlm"},
	{0xB33404F82A6C2B94ULL, "gen00fl_isabela_init6077689.cl"},
	{0xB3344B173FA41BEDULL, "hlac_w_sundermount_-327681_0x.dds"},
	{0xB334B0C0476332D2ULL, "qun160_vashothbandits.cnv"},
	{0xB334CA2B06F0062CULL, "vxw_woundedcoast_1s9_0_0.vlm"},
	{0xB334DD842487391EULL, "vxht_0_hightown_bc8_0_0.vlm"},
	{0xB33521B6A3BCD0ABULL, "494843085.ogg"},
	{0xB3354172917C6228ULL, "6082229_f.fxe"},
	{0xB335529D141895F0ULL, "6098716_f.fxe"},
	{0xB33599CA8DB0A9ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_rhand.gad"},
	{0xB335C14FBAF69B70ULL, "202231824.ogg"},
	{0xB335D4ADA9004B52ULL, "art\\vfx\\combat\\vfx_c_mer_whirlwind_c.mmh"},
	{0xB3366FC65EC8EC00ULL, "vxht_keep_67z_0_0.vlm"},
	{0xB336DA5C893AD182ULL, "one120_athenril6088287.cl"},
	{0xB336F15D7717713AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_bda_0d.dds"},
	{0xB33742933138F302ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_atk_v2.ani"},
	{0xB3378AFCDF321B84ULL, "plt_cod_aow_spellcombo6.plo"},
	{0xB337CEAF5028C463ULL, "6113822_m.fxe"},
	{0xB337D2C04765A22AULL, "qun160_vashothbandits.bnk"},
	{0xB337DE8ECE05E26BULL, "vxht_keep_6ac_0_0.vlm"},
	{0xB338281080717E8FULL, "vxlt_gallowstemplar_f_3yn_0_0.vlm"},
	{0xB3389ACC93715601ULL, "m_isa311ar_castillon.wbd"},
	{0xB339A49F2F4AD1B5ULL, "vxlt_0_lowtown_f66_0_2.vlm"},
	{0xB339C9312CAB3E41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ad_transition.evt"},
	{0xB33A0CE862510A89ULL, "hlac_ca_cavern1_haunted_24424_7x.dds"},
	{0xB33A62F176B98EB9ULL, "lgt150_nat_were_supporter6110948.cl"},
	{0xB33AC04713899B3BULL, "433982138.ogg"},
	{0xB33B335411F7B500ULL, "6143183_m.fxe"},
	{0xB33B96CA8DB51010ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_up_rhand.evt"},
	{0xB33BF08C2BCC62DBULL, "w_encounter2_camp_1_0.msh"},
	{0xB33C028C2BCC8126ULL, "w_encounter2_camp_1_0.mao"},
	{0xB33C0E8C2BCC95C5ULL, "w_encounter2_camp_1_0.mmh"},
	{0xB33C4EEB9C432637ULL, "sky_lt_0_lowtown_docks.mao"},
	{0xB33D077EEDFB590AULL, "w_woundedcoast_24_0.rml"},
	{0xB33D42551C02432BULL, "452155537.ogg"},
	{0xB33DA0FE6AE2D278ULL, "lowtown_fem_urchin_mob6165925.cl"},
	{0xB33DE6F7EFEDD188ULL, "da2terrain_terrain_p0.hsh.fxo.dx11.assembly"},
	{0xB33E58AF2C15DED9ULL, "zz_perc.arl"},
	{0xB33E779C05871883ULL, "fen210_hadriana6174813.cl"},
	{0xB33FFB33C853D167ULL, "l_env_wal_hightown_int_whis_murm_p2p.wbd"},
	{0xB340081B15A77ADDULL, "vxw_bonepitdestroyed_a3_24a_0_0.vlm"},
	{0xB3401AB001B91B83ULL, "vxht_mansion2_1kv_0_0.vlm"},
	{0xB34096F637043F3BULL, "gen00fl_dog6108768.cl"},
	{0xB340D2154F58F8EDULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_001a_0n.dds"},
	{0xB34171C8C9DE95EAULL, "pro000_cut_act_of_mercy6001043.cl"},
	{0xB341A948E80C0C2DULL, "demo_weapon_swap.stg"},
	{0xB341F77B4726F01DULL, "hlac_lt_0_lowtown_a2_n_65536_11z.dds"},
	{0xB34400F984EA6526ULL, "6120025_m.fxe"},
	{0xB344E116459FE71FULL, "370488078.ogg"},
	{0xB34553C7109B1E73ULL, "ico_belt_3.dds"},
	{0xB3460CC400116A3AULL, "161564720.ogg"},
	{0xB346506BBFFBFAEBULL, "vxht_keep_6jb_0_0.vlm"},
	{0xB34688C31E1BED60ULL, "llac_dr_slavers_3217_12.dds"},
	{0xB346AA517E249E68ULL, "vxht_mansion2_2ad_0_0.vlm"},
	{0xB346C1312CB5BA90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ad_transition.ani"},
	{0xB3471A6335C10ACDULL, "lt_0_lowtown_a2_n_44_0.rml"},
	{0xB34727EF7C0D4223ULL, "cs_qcr000cs_arishok_response.wbd"},
	{0xB3477122A720DEC1ULL, "ht_0_hightown_n.arl"},
	{0xB347EBBAB2AE83BCULL, "vxlt_gallowstemplar_f_5pl_0_0.vlm"},
	{0xB3480A07BAFCA308ULL, "hlac_lt_0_lowtown_docks_n_-65538_5z.dds"},
	{0xB3481AB6595D0C19ULL, "130779909.ogg"},
	{0xB34839F40CA1AE01ULL, "548029049.ogg"},
	{0xB3490735F9898B60ULL, "349353047.ogg"},
	{0xB34939E73814D312ULL, "worldmaps.gda"},
	{0xB3497AC11F1F09A1ULL, "vxlt_gallowsprison_n_6ks_0_0.vlm"},
	{0xB349D19CA286AE1BULL, "vxht_0_hightown_n_9bf_0_0.vlm"},
	{0xB349DDF56B3C4937ULL, "zz_fol_debug6056992.cl"},
	{0xB34A70B6FA5244F2ULL, "qun160_vashothlead6064678.cl"},
	{0xB34A9CC2A468D1A5ULL, "w_encounter1_34_0.phy"},
	{0xB34BA44782EFD174ULL, "vxdr_primevalentrance_32y_0_0.vlm"},
	{0xB34BD3A01EC80304ULL, "330343110.ogg"},
	{0xB34BFB48025D8AD0ULL, "6076405_m.fxe"},
	{0xB34C348E8A9494E6ULL, "llac_lt_0_lowtown_0_9.dds"},
	{0xB34D04AC9B21535BULL, "art\\characters\\weapons\\dagger\\w_dger_105a.msh"},
	{0xB34D0E6C1814A083ULL, "zz_dae_debug6029500.cl"},
	{0xB34D16AC9B2171A6ULL, "art\\characters\\weapons\\dagger\\w_dger_105a.mao"},
	{0xB34D22AC9B218645ULL, "art\\characters\\weapons\\dagger\\w_dger_105a.mmh"},
	{0xB34D39B440BDF7E6ULL, "355208016.ogg"},
	{0xB34DA5DDD8CF7FBFULL, "960176635.ogg"},
	{0xB34DBB699C77555EULL, "vxlt_0_lowtown_a2_n_c_ets_0_0.vlm"},
	{0xB34DBC632D57BE6FULL, "6060351_m.fxe"},
	{0xB34EC1B14FE256FCULL, "gen00fl_sebastian6189906.cl"},
	{0xB34F2DBF6D9DBD2BULL, "vxlt_0_lowtown_docks_a3_d2g_0_1.vlm"},
	{0xB34F7E5F259E6519ULL, "mag100_karras6034526.cl"},
	{0xB34FB102023F9E72ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15b.phy"},
	{0xB34FBF1DC14B4729ULL, "vxht_mansion2b_1bw_0_0.vlm"},
	{0xB35071BB447A4B57ULL, "6206151_m.fxe"},
	{0xB3509A5423AD2613ULL, "14660424.ogg"},
	{0xB350D233C75A74BDULL, "434717272.ogg"},
	{0xB350E1CF02A78AA3ULL, "zz_followers_banter6158949.cl"},
	{0xB35158E3E7913A61ULL, "en_cl.mao"},
	{0xB3517722BB249224ULL, "vxlt_0_lowtown_a2_n_c_hv6_0_0.vlm"},
	{0xB352302A6693ABC8ULL, "384320387.ogg"},
	{0xB3524721A5F663ADULL, "1011396017.ogg"},
	{0xB353283ABBB509CEULL, "663407974.ogg"},
	{0xB3533C01D4108EB0ULL, "vxw_bonepitmining_a2_ae_0_0.vlm"},
	{0xB354428A65AA94D2ULL, "sky_lt_gallowsprison_n.mao"},
	{0xB3546303764377FCULL, "vxht_0_hightown_a3_8px_0_0.vlm"},
	{0xB354A4853E7BB0CDULL, "6088064_f.fxe"},
	{0xB354C78CDEFF264CULL, "zz_fol_debug6057002.cl"},
	{0xB354F4170C0A251EULL, "850634426.ogg"},
	{0xB3551C2E701AB6E8ULL, "vxht_mansion2_2je_0_0.vlm"},
	{0xB35627437A9E6DEAULL, "vxlt_gallowscourtyard_3w0_0_0.vlm"},
	{0xB3566A94E75C79EFULL, "hlac_ht_0_hightown_a2_2_1y.dds"},
	{0xB356BF38B922C6C0ULL, "gen00fl_anders6116136.cl"},
	{0xB356F84762FD8BEDULL, "lgt220ip_shp_mag_tate.utp"},
	{0xB35772A41B15AC29ULL, "hlac_lt_0_lowtown_docks_a3_-196611_1z.dds"},
	{0xB357A2D5955FEFD3ULL, "698472201.ogg"},
	{0xB357B677AF02F073ULL, "24350681.ogg"},
	{0xB3581BC738CEC088ULL, "m_dae111ar_hanged_man_nt.wbd"},
	{0xB35821B0E6425BC2ULL, "vxht_mansion2_252_0_0.vlm"},
	{0xB3585A2734266EE5ULL, "vxlt_gallowstemplar_5cw_0_0.vlm"},
	{0xB3587DCDF9C5AE98ULL, "zz_and_debug6134132.cl"},
	{0xB358CBA99CDED5ACULL, "mag320_orsino6071191.cl"},
	{0xB359844286AF60B2ULL, "146954502.ogg"},
	{0xB35A7701F5EBAC99ULL, "334867220.ogg"},
	{0xB35AB192B404781CULL, "w_bonepitmining_a2_135_0.mao"},
	{0xB35ABD92B4048CBFULL, "w_bonepitmining_a2_135_0.mmh"},
	{0xB35AC392B4049761ULL, "w_bonepitmining_a2_135_0.msh"},
	{0xB35AEE008D17A9F0ULL, "vxlt_0_lowtown_a2_hdy_0_0.vlm"},
	{0xB35B2BBBCD28F62EULL, "vxlt_0_lowtown_a3_n_c_eos_0_0.vlm"},
	{0xB35BA33CFDD231FCULL, "6103103_carver_dumped_here.cut"},
	{0xB35BCCF4C3063AAAULL, "gen00fl_anders6125738.cl"},
	{0xB35BF7F02FA82434ULL, "6002587_m.fxe"},
	{0xB35D149D857D5A16ULL, "506347012.ogg"},
	{0xB35D46FFA5EB5BDBULL, "w_encounter2_50_0.msh"},
	{0xB35D58FFA5EB7A26ULL, "w_encounter2_50_0.mao"},
	{0xB35D64FFA5EB8EC5ULL, "w_encounter2_50_0.mmh"},
	{0xB35E1DD77A2C76CBULL, "281030692.ogg"},
	{0xB35E519A08CBD1F5ULL, "vxlt_blackemporium_14w_0_0.vlm"},
	{0xB35EBCCC93925F03ULL, "m_isa311ar_castillon.bnk"},
	{0xB35ED3ADA9234EF9ULL, "art\\vfx\\combat\\vfx_c_mer_whirlwind_c.anb"},
	{0xB35EDDA319CA5629ULL, "gen00fl_fenris6126354.cl"},
	{0xB35F12BF9895971EULL, "vxht_mansion_n_4yg_0_0.vlm"},
	{0xB35F31FC6FFBD285ULL, "vxlt_0_lowtown_d6f_0_0.vlm"},
	{0xB35F6A936DD1FB75ULL, "vxw_encounter3_4f_0_0.vlm"},
	{0xB3603C8B304088CEULL, "vxlt_0_lowtown_docks_n_b5l_0_0.vlm"},
	{0xB3604319973C3334ULL, "w_encounter2_mmap_ph.dds"},
	{0xB3605224772E2D0BULL, "679704938.ogg"},
	{0xB360919B418861E1ULL, "vxlt_0_lowtown_docks_a3_n_d0y_0_0.vlm"},
	{0xB361014A1B6E439CULL, "vxht_chantry_287_0_0.vlm"},
	{0xB3614779C4DCB29CULL, "hlac_w_encounter2_camp_262147_0x.dds"},
	{0xB36178604B807DC1ULL, "6051630_m.fxe"},
	{0xB3617B9CC22D9B2AULL, "mcr310_sibling6076286.cl"},
	{0xB361ADDEC606A730ULL, "vxw_bonepitmining_a2_lb_0_0.vlm"},
	{0xB361D3FC02004658ULL, "zz_dae_debug6029880.cl"},
	{0xB361E12778FD2A47ULL, "6130127_f.fxe"},
	{0xB3621DA975C97CEBULL, "hlac_lt_0_lowtown_docks_n_131071_1z.dds"},
	{0xB3625CB64A47D26BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv5.ani"},
	{0xB362B49530D9263EULL, "6104603_m.fxe"},
	{0xB362D29B30E9D3BBULL, "vxlt_undercityserialkiller_1sg_0_0.vlm"},
	{0xB363E4A77008262DULL, "gen00fl_anders6125252.cl"},
	{0xB364606CA608D382ULL, "569250832.ogg"},
	{0xB36496DA837CC4BBULL, "248806047.ogg"},
	{0xB364DF2ABFDF2237ULL, "857108493.ogg"},
	{0xB366564B4CC10694ULL, "323530705.ogg"},
	{0xB366CC04261C8765ULL, "vxlt_gallowstemplar_5lv_0_0.vlm"},
	{0xB366E0E37CF2990EULL, "w_blightlands_221_0.msh"},
	{0xB366EEE37CF2B15FULL, "w_blightlands_221_0.mao"},
	{0xB366FAE37CF2C5FCULL, "w_blightlands_221_0.mmh"},
	{0xB36725E1BFBD8D66ULL, "421366286.ogg"},
	{0xB36772BBEE45C2EEULL, "56147894.ogg"},
	{0xB367D5DFE0BCFC5DULL, "350145207.ogg"},
	{0xB367E481F7F57F00ULL, "vxlt_gallowstemplar_n_4un_0_0.vlm"},
	{0xB367E7585D70F13BULL, "follower_banter6143107.cl"},
	{0xB367FA3D145B8E2EULL, "116293686.ogg"},
	{0xB3680C219FD1E6F8ULL, "482410010.ogg"},
	{0xB3682BE74977DB5BULL, "6207771_f.fxe"},
	{0xB3686C521E2A2890ULL, "gen_im_trs_dwv_med_cht_r0.uti"},
	{0xB368CCE499D605B4ULL, "245121009.ogg"},
	{0xB368E19CD9E6CA1EULL, "vxlt_gallowstemplar_5k0_0_0.vlm"},
	{0xB3699D98BF1F0EAEULL, "vxlt_0_lowtown_a3_n_c_ert_0_0.vlm"},
	{0xB36A025824FE43C0ULL, "hlac_lt_0_lowtown_a2_65536_2z.dds"},
	{0xB36A775E93D5A81CULL, "qcr220st_orsino_distraction.stg"},
	{0xB36ADB46EEA4E5D3ULL, "art\\characters\\creatures\\hurlock\\textures\\w_qui_hlk_0n.dds"},
	{0xB36B52C67BDF6C6AULL, "581547354.ogg"},
	{0xB36BA5115C6BE206ULL, "dae000_mother6099176.cl"},
	{0xB36C22F2722A333CULL, "gen00fl_merrill6107002.cl"},
	{0xB36C4647C7349E03ULL, "166070157.ogg"},
	{0xB36C742F0ECFB6F5ULL, "follower_banter6172912.cl"},
	{0xB36C8437BCED0339ULL, "6106772_m.fxe"},
	{0xB36CA2D7A5377829ULL, "vxht_mansion2b_1xu_0_0.vlm"},
	{0xB36CC1C8B02417D1ULL, "vxlt_0_lowtown_n_ghd_0_0.vlm"},
	{0xB36CD5A958CE757AULL, "6151869_m.fxe"},
	{0xB36D00337C80F68DULL, "art\\characters\\playercharacter\\humanfemale\\arm_avea.msh"},
	{0xB36D0E337C810EC0ULL, "art\\characters\\playercharacter\\humanfemale\\arm_avea.mao"},
	{0xB36D12337C811513ULL, "art\\characters\\playercharacter\\humanfemale\\arm_avea.mmh"},
	{0xB36D19EF7C2DF9D1ULL, "cs_qcr000cs_arishok_response.bnk"},
	{0xB36D3038CD62C231ULL, "mrl000_marethari6129345.cl"},
	{0xB36D542F9185F0D0ULL, "vxht_keep_n_61s_0_0.vlm"},
	{0xB36DB272522F2156ULL, "qun221_petrice6075983.cl"},
	{0xB36DE35B1E9F84AAULL, "vxlt_backalley_n_1nx_0_0.vlm"},
	{0xB36F382229689FB1ULL, "841012281.ogg"},
	{0xB36F44B64A5233E2ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv5.evt"},
	{0xB36FD896F37E277BULL, "6129059_f.fxe"},
	{0xB36FD95496E93205ULL, "vxdr_primevaltomb_8yj_0_0.vlm"},
	{0xB3701FBBB7FCBFB0ULL, "vxw_bonepitmining_a2_kc_0_0.vlm"},
	{0xB3702F9304980E35ULL, "follower_banter6141792.cl"},
	{0xB37059D32B55EE5EULL, "6095013_m.fxe"},
	{0xB37060103A4A4479ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1.mmh"},
	{0xB371649BC4768FF1ULL, "vrc321_party_barks6095768.cl"},
	{0xB3716B0EA2867888ULL, "vxht_mansion2b_2bu_0_0.vlm"},
	{0xB3717F0814DD5C81ULL, "6128963_m.fxe"},
	{0xB371801314829909ULL, "lt_0_lowtown_docks_a3_20_0_lc.mmh"},
	{0xB371820AD52EE5D1ULL, "6141795_m.fxe"},
	{0xB3718B8144990A7DULL, "mag101_grace6156845.cl"},
	{0xB371FFE85406E7E8ULL, "vxht_mansion2_2pc_0_0.vlm"},
	{0xB3721729A6E6D17FULL, "6001259_m.fxe"},
	{0xB372B7E53AA95BE6ULL, "6118371_f.fxe"},
	{0xB3730D095F779169ULL, "201869142.ogg"},
	{0xB373436ECFD09B54ULL, "hlac_lt_gallowstemplar_f_2394_10x.dds"},
	{0xB3739C6756B6F214ULL, "hm_gen_slaver_mage.mor"},
	{0xB3740088CDF1301CULL, "923605874.ogg"},
	{0xB37422A2EC0D297FULL, "hlac_lt_0_lowtown_docks_a3_n_-131076_1y.dds"},
	{0xB37437E953AE6EA9ULL, "vxlt_0_lowtown_a2_n_gvb_0_0.vlm"},
	{0xB37438805F75CDFEULL, "6177177_m.fxe"},
	{0xB3745DEA6379830EULL, "vxht_mansion2_2en_0_0.vlm"},
	{0xB374DBC6FB84CCDCULL, "6051680_m.fxe"},
	{0xB374E12FCCA868DBULL, "174691649.ogg"},
	{0xB374F69100977517ULL, "zz_ave_debug6056788.cl"},
	{0xB3753DE118129FE5ULL, "vxlt_gallowstemplar_5iu_0_0.vlm"},
	{0xB3757E2E7C5C0825ULL, "ht_0_hightown_n_13_0.rml"},
	{0xB375A4F6D51C4FC7ULL, "combat_lmm_r1_a1.gda"},
	{0xB375B7067E2CBDC2ULL, "mer210_seldrek.cnv"},
	{0xB375D5B2873BA195ULL, "310221406.ogg"},
	{0xB37610E386F13F3AULL, "6122157_m.fxe"},
	{0xB37643B64A5850F5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv5.gad"},
	{0xB3769C94C3BEABDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ledgedropland.ani"},
	{0xB376AEB47799A51BULL, "521225727.ogg"},
	{0xB3771193AFBDB3C4ULL, "vxht_mansion2b_342_0_0.vlm"},
	{0xB37711C3B83C92F4ULL, "gen_im_trs_bst_drag_gland_01.uti"},
	{0xB37718EB32922E2AULL, "6059062_m.fxe"},
	{0xB3774A25948FAFAFULL, "deathhints_base.gda"},
	{0xB3775379CBDCE29EULL, "vxlt_gallowstemplar_5t1_0_0.vlm"},
	{0xB3778AFF57D090ACULL, "77529873.ogg"},
	{0xB377BC7B9F43178FULL, "hlac_lt_0_lowtown_docks_a3_0_6z.dds"},
	{0xB377FD0D6837F9EEULL, "118828943.ogg"},
	{0xB3780F75B115272EULL, "vxlt_0_lowtown_a3_n_c_euu_0_0.vlm"},
	{0xB3781494B1F2013EULL, "vfx_sparkae.dds"},
	{0xB378965131657478ULL, "vxca_caverns2_1za_0_0.vlm"},
	{0xB37899067E2EC0FAULL, "mer210_seldrek.bnk"},
	{0xB37900DD64426A38ULL, "vfx_combat_blood_01.mao"},
	{0xB379E69E840C1E90ULL, "dae200_saemus6142714.cl"},
	{0xB37A2DE944F2A5F2ULL, "hlac_lt_gamlenshouse_n_-1_3y.dds"},
	{0xB37A79381F2CD29DULL, "zz_followers_banter6159474.cl"},
	{0xB37A82C97667FEA4ULL, "llac_lt_gallowstemplar_2394_0.dds"},
	{0xB37B14B4972D90A9ULL, "vxht_mansion2b_1st_0_0.vlm"},
	{0xB37B7CF0498D9DBDULL, "239544079.ogg"},
	{0xB37B942E4AF65FD7ULL, "genip_blocker_wide_info.utp"},
	{0xB37C248F24E35353ULL, "hlac_ht_0_hightown_131072_0x.dds"},
	{0xB37C328DA221622DULL, "454171513.ogg"},
	{0xB37C92E8C0AB9F79ULL, "downarrow.dds"},
	{0xB37C9FE82118CF76ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_enter.evt"},
	{0xB37CA994C3C32D1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ledgedropland.gad"},
	{0xB37CAE453016830EULL, "plt_zz_merpt_debug.plo"},
	{0xB37D0E8C2C042CCEULL, "w_encounter2_camp_1_0.rml"},
	{0xB37D8344425972AFULL, "dae000_merchant_generic_hm6172671.cl"},
	{0xB37E368D669C3F1AULL, "gen00fl_isabela_init6076604.cl"},
	{0xB37E4B3188DF4A85ULL, "vxdr_primevaltomb_8ti_0_0.vlm"},
	{0xB37E4EB649F93AC2ULL, "199563997.ogg"},
	{0xB37E7F27AB6CCEADULL, "6168545_m.fxe"},
	{0xB37EBBF7854D4556ULL, "312576297.ogg"},
	{0xB37ECCB24258E2C6ULL, "6120840_f.fxe"},
	{0xB37F28DF41D471D7ULL, "vxdr_primevaltomb_91e_0_0.vlm"},
	{0xB37F6DC379E2787FULL, "vxca_caverns2_975_0_0.vlm"},
	{0xB3800533C889C59DULL, "l_env_wal_hightown_int_whis_murm_p2p.bnk"},
	{0xB38058F34F2547C6ULL, "816701728.ogg"},
	{0xB3806B972BB3985CULL, "160943715.ogg"},
	{0xB380890698DCE9FFULL, "6004159_m.fxe"},
	{0xB380989F9259ED71ULL, "471464898.ogg"},
	{0xB380E5BCA1BF0AD8ULL, "dae000_sketch6179006.cl"},
	{0xB380E72E6F140BAEULL, "gui_guipixelshader_p6.psh.fxo.dx11"},
	{0xB38152CD85131D62ULL, "dae000_gamlen6102534.cl"},
	{0xB381C8287E3F31E3ULL, "zz_t_trailer_shot_next.cut"},
	{0xB381DDFF4B3FA51CULL, "846757270.ogg"},
	{0xB3823B5C05824875ULL, "725419403.ogg"},
	{0xB3824E3908D388F8ULL, "hlac_lt_gallowstemplar_f_2383_4y.dds"},
	{0xB3827054031920CEULL, "799786606.ogg"},
	{0xB382746646905339ULL, "6071719_m.fxe"},
	{0xB38296E8211D2B01ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_enter.gad"},
	{0xB3829DAB5F4F5758ULL, "docks_fem_sailor_mob6176296.cl"},
	{0xB382D28D3BEADE34ULL, "dae000_arishok6096346.cl"},
	{0xB3838494C3C90D3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ledgedropland.evt"},
	{0xB383AD680B155F03ULL, "sky_ca_caverns1_mines.mao"},
	{0xB383AFBE0A08B865ULL, "vxlt_gallowstemplar_5rt_0_0.vlm"},
	{0xB383E78C2C0A0972ULL, "w_encounter2_camp_1_0.phy"},
	{0xB383E9DE17B77ADDULL, "vxlt_0_lowtown_cml_0_0.vlm"},
	{0xB38404D85F95291EULL, "mag320_nobleman_36060832.cl"},
	{0xB384EFA90FE0AFF9ULL, "124385220.ogg"},
	{0xB38532F0346869A2ULL, "63906277.ogg"},
	{0xB385515C9CBA597DULL, "gen00fl_fenris6113941.cl"},
	{0xB3859F01A65A5525ULL, "gen00fl_anders6125494.cl"},
	{0xB385D88FD0C9F0A8ULL, "ico_ring_5.dds"},
	{0xB38645CC78934233ULL, "follower_banter6136327.cl"},
	{0xB3868152A30B3FAEULL, "vxlt_0_lowtown_a3_n_c_exv_0_0.vlm"},
	{0xB386C2E9B025E3AAULL, "vxca_caverns2_8xe_0_1.vlm"},
	{0xB3871E99984BEBB0ULL, "6056813_f.fxe"},
	{0xB3874D960A5CC26EULL, "6180521_f.fxe"},
	{0xB38751FD22AAD8A9ULL, "601558538.ogg"},
	{0xB38781B3A2A1E563ULL, "79035870.ogg"},
	{0xB3879D37DAB20172ULL, "vxlt_0_lowtown_fij_0_0.vlm"},
	{0xB387CF0BD0B876C0ULL, "6039394_m.fxe"},
	{0xB3881594FF0EEA60ULL, "vxw_woundedcoast_g8_0_0.vlm"},
	{0xB38836C2A6A377D2ULL, "231421874.ogg"},
	{0xB38869649AB455B8ULL, "6080102_f.fxe"},
	{0xB388763C6BE228A2ULL, "vxlt_darktown_4zf_0_0.vlm"},
	{0xB388FF6A9A6553EFULL, "6126624_m.fxe"},
	{0xB38932E5A5D8D311ULL, "vxht_0_hightown_n_cf2_0_0.vlm"},
	{0xB389381BDE4522CBULL, "art\\vfx\\cinematic\\vfx_scaredmage_c.impa.ani"},
	{0xB38997E821234BAFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_enter.ani"},
	{0xB389A582941048D1ULL, "vxlt_0_lowtown_n_gbz_0_0.vlm"},
	{0xB389C672BEEAC89FULL, "99397950.ogg"},
	{0xB38A38267B028B90ULL, "6084220_m.fxe"},
	{0xB38A58F86EB5E0A4ULL, "gen00fl_anders6119252.cl"},
	{0xB38A5991A3FDA6F8ULL, "vxca_cavern1_haunted_j77_0_0.vlm"},
	{0xB38AD7AC9786EEC0ULL, "6141138_m.fxe"},
	{0xB38B1B78A21E853EULL, "vxlt_0_lowtown_a2_n_c_fag_0_0.vlm"},
	{0xB38B32B427233A06ULL, "6055570_m.fxe"},
	{0xB38B34E1B313C168ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_cull.msh"},
	{0xB38B52E1B313F392ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_cull.mmh"},
	{0xB38B61F29ACE5728ULL, "gen00fl_isabela6075259.cl"},
	{0xB38B812EBF010198ULL, "170181750.ogg"},
	{0xB38BF54CCE12A47AULL, "424809748.ogg"},
	{0xB38C138296CD0A60ULL, "6066268_m.fxe"},
	{0xB38C395CE729B6C1ULL, "hlac_ht_mansion_n_3370_2x.dds"},
	{0xB38C79B9BD8B9D1DULL, "6171936_m.fxe"},
	{0xB38C83BC5C314B85ULL, "vxht_chantry_n_353_0_0.vlm"},
	{0xB38D1FEEF1104301ULL, "458175694.ogg"},
	{0xB38D5CE4B0C6B0A3ULL, "6111129_m.fxe"},
	{0xB38D6343BAC3F470ULL, "vxht_keep_n_58g_0_0.vlm"},
	{0xB38DDFA06BD890FFULL, "vxca_caverns2_984_0_0.vlm"},
	{0xB38ED1491DBDC779ULL, "mag161tr_slavers.ncs"},
	{0xB3901DE5931E4D45ULL, "353471709.ogg"},
	{0xB3907AD379A1E02BULL, "dae000_gamlen6149446.cl"},
	{0xB3909CFF0347D995ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_arch.msh"},
	{0xB3909EFF0347DD0BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_arch.mmh"},
	{0xB390AAFF0347F1A8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\2ca_arch.mao"},
	{0xB391419A8489689EULL, "267008687.ogg"},
	{0xB391775644773A85ULL, "6054185_m.fxe"},
	{0xB391884C28672138ULL, "189370730.ogg"},
	{0xB3919F453028C59DULL, "plt_zz_merpt_debug.nss"},
	{0xB391AE99780DAADDULL, "576070093.ogg"},
	{0xB391C33B437D5DE9ULL, "plt_prm000pt_ashes_amulet.nss"},
	{0xB3921D3C3CA4BDEBULL, "350485970.ogg"},
	{0xB392580B50345F6FULL, "820895300.ogg"},
	{0xB392786C49B27F03ULL, "vxht_bartrand_n_5q8_0_0.vlm"},
	{0xB392CED09DCB9206ULL, "dae000ip_crowd_mf3_drunk1.utp"},
	{0xB392E864F5502E98ULL, "hlac_dr_primevaltomb_659_4z.dds"},
	{0xB393588F297E9112ULL, "hlac_ht_keep_n_3736_3x.dds"},
	{0xB39393A5143E0E3FULL, "vxca_caverns2prc_3cf_0_0.vlm"},
	{0xB3943046F987F16FULL, "cd_mag330_walter.wbd"},
	{0xB39430DF9203B7F0ULL, "6122439_f.fxe"},
	{0xB3943733AFC9139EULL, "vxlt_gallowstemplar_5z3_0_0.vlm"},
	{0xB394A91EBC3A0DF7ULL, "vxlt_undercityserialkiller_1ss_0_0.vlm"},
	{0xB394FBAC9B5EF9F2ULL, "art\\characters\\weapons\\dagger\\w_dger_105a.phy"},
	{0xB395412259355BAAULL, "hlac_ca_caverns1_mines_24421_7z.dds"},
	{0xB3955238BF00AAFDULL, "vxca_cavern1_haunted_e57_0_0.vlm"},
	{0xB395AA7896751AE7ULL, "575788483.ogg"},
	{0xB39612C4C17D3E61ULL, "gen_im_cns_tom_war_juggnaut.uti"},
	{0xB396DD27897AEED7ULL, "6104574_m.fxe"},
	{0xB397201BDE513742ULL, "art\\vfx\\cinematic\\vfx_scaredmage_c.impa.evt"},
	{0xB3978002A0870D24ULL, "vxlt_warehouse_15y_0_0.vlm"},
	{0xB3989053893A0140ULL, "art\\vfx\\textures\\vfx_bub_burst.dds"},
	{0xB398CB6E95F3BF78ULL, "vxca_cavern1_haunted_j24_0_0.vlm"},
	{0xB3993740E004DA2EULL, "w_blightlands_152_0.phy"},
	{0xB3996DC956C60A83ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikib_0.pwk"},
	{0xB399D03561F74D87ULL, "549977121.ogg"},
	{0xB39A4FC6D7E9FC67ULL, "art\\characters\\creatures\\buzzard\\c_buzzard.mao"},
	{0xB39A5E00C966732FULL, "vxw_encounter1_eu_0_0.vlm"},
	{0xB39A69FE9A5441AFULL, "hlac_ht_0_hightown_a2_-1_0x.dds"},
	{0xB39A6A1E87466DCCULL, "isa311_isabela6094452.cl"},
	{0xB39A9EA71CD684D8ULL, "218175415.ogg"},
	{0xB39AEB211E70A528ULL, "fxa_web_shine.mao"},
	{0xB39B6E34696F0940ULL, "vxht_0_hightown_a2_4n3_0_0.vlm"},
	{0xB39B741A296C7C03ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_roba_0t.dds"},
	{0xB39BC360E8E7379BULL, "hlac_w_woundedcoast_131074_0x.dds"},
	{0xB39BCC721B6810EAULL, "1051037506.ogg"},
	{0xB39C22D672923349ULL, "800158595.ogg"},
	{0xB39C87B946A86090ULL, "6071268_m.fxe"},
	{0xB39C93CB357C8D58ULL, "dae020cr_hf_noble_3.utc"},
	{0xB39CB17B0FA9DA4CULL, "vxlt_shop_dh_0_0.vlm"},
	{0xB39CCA9C5515286FULL, "vxht_keep_n_3iq_0_0.vlm"},
	{0xB39D1F1BDE55A155ULL, "art\\vfx\\cinematic\\vfx_scaredmage_c.impa.gad"},
	{0xB39D57F57CCF9AEFULL, "vxlt_0_lowtown_docks_a3_cy5_0_1.vlm"},
	{0xB39DAE1B600A73BEULL, "vxw_bonepit_13t_0_0.vlm"},
	{0xB39DDEDD9CD7EA57ULL, "vxw_encounter2_camp_36_0_0.vlm"},
	{0xB39DEE62531748EDULL, "llac_lt_0_lowtown_docks_n_65534_5.dds"},
	{0xB39E5CDDDAE72427ULL, "hlac_ht_mansion2b_818_24y.dds"},
	{0xB39E64FFA62325CEULL, "w_encounter2_50_0.rml"},
	{0xB39EBA90FB5E0004ULL, "vxca_caverns2_8bh_0_0.vlm"},
	{0xB39F41C7390B044AULL, "m_dae111ar_hanged_man_nt.bnk"},
	{0xB39F57A5C478EA9CULL, "6108717_m.fxe"},
	{0xB39F8812CCF36038ULL, "llac_w_woundedcoast_196610_0.dds"},
	{0xB39F9801A8396C82ULL, "hf_gen_act1.mor"},
	{0xB3A02B459B7B1A6CULL, "vxdr_primevaltomb_6z5_0_1.vlm"},
	{0xB3A09475A1F34533ULL, "vxlt_0_lowtown_docks_n_b0z_0_0.vlm"},
	{0xB3A0BBB785979032ULL, "6143336_m.fxe"},
	{0xB3A0D215A4ECA22EULL, "seb221pt_repentance.ncs"},
	{0xB3A0EAFBAE8A33AEULL, "vxlt_0_lowtown_gjj_0_0.vlm"},
	{0xB3A137EBBA06B3F0ULL, "882211666.ogg"},
	{0xB3A13B5A06896CB7ULL, "fen210_hadriana6110415.cl"},
	{0xB3A187AF9AF8E94DULL, "vxca_caverns1_gk3_0_1.vlm"},
	{0xB3A2A64C07E9A1E0ULL, "279386342.ogg"},
	{0xB3A2AD1EC0DED3E4ULL, "gen00fl_sebastian6188633.cl"},
	{0xB3A2FF5FD52CA9D6ULL, "gen00fl_isabela6079077.cl"},
	{0xB3A33FB2A6D1EEDDULL, "hlac_lt_foundry_n_320_11z.dds"},
	{0xB3A33FC97DED4663ULL, "vxlt_gallowstemplar_n_5yx_0_0.vlm"},
	{0xB3A35F8962C09AEAULL, "follower_banter6169885.cl"},
	{0xB3A39AA36B7B6FE4ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acast.ani"},
	{0xB3A3CF1D1418A286ULL, "vxdr_0_deeproads_gh7_0_0.vlm"},
	{0xB3A3D00DD1517AA8ULL, "vxca_caverns1_i2r_0_0.vlm"},
	{0xB3A3D20E7AE0C9F5ULL, "ht_0_hightown_a3_46_0.rml"},
	{0xB3A47D6EB6AC41C6ULL, "vxca_cavern1_haunted_gj3_0_0.vlm"},
	{0xB3A4AA008720FD5FULL, "356523966.ogg"},
	{0xB3A53DFFA6290272ULL, "w_encounter2_50_0.phy"},
	{0xB3A544061766A819ULL, "art\\levels\\races\\proxy\\props\\_textures\\cheese_sa.dds"},
	{0xB3A680D8387D197CULL, "vxht_mansion2b_1ak_0_0.vlm"},
	{0xB3A78BDBCD76616CULL, "w_bonepitmining_a2_77_0.msh"},
	{0xB3A791DBCD766B5EULL, "w_bonepitmining_a2_77_0.mmh"},
	{0xB3A79DDBCD767FF5ULL, "w_bonepitmining_a2_77_0.mao"},
	{0xB3A7AA3647E4A783ULL, "574749889.ogg"},
	{0xB3A7AC47DB2330AFULL, "llac_lt_0_lowtown_a2_n_-65537_4.dds"},
	{0xB3A7D01051FED2CEULL, "art\\characters\\weapons\\longbows\\w_lbw_lb107.mmh"},
	{0xB3A7DC1051FEE765ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb107.mao"},
	{0xB3A7EA1051FEFEBCULL, "art\\characters\\weapons\\longbows\\w_lbw_lb107.msh"},
	{0xB3A8167AB79D20E0ULL, "vxw_bonepit_1q0_0_0.vlm"},
	{0xB3A8183E974AF207ULL, "mcr371st_orsino_2.stg"},
	{0xB3A865730AA2605FULL, "6057344_m.fxe"},
	{0xB3A8A23C54B95A6FULL, "art\\levels\\races\\elf\\props\\_textures\\prp_tiki_d.dds"},
	{0xB3A8B6443FC1749CULL, "ht_mansion_n_3373.rml"},
	{0xB3A8CFDDBB5C8BAFULL, "vxw_encounter1_jt_0_0.vlm"},
	{0xB3A8D430AB41D6B3ULL, "845487906.ogg"},
	{0xB3A940B946AF5ACCULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_black.dds"},
	{0xB3A9B0ACB42C3A41ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.cess.ani"},
	{0xB3A9EAD2234EA8B7ULL, "lining_ciga_f.dds"},
	{0xB3AA48197DA6D40FULL, "119015679.ogg"},
	{0xB3AA868BD53D0F9FULL, "lt_0_lowtown_docks_n_23_0.rml"},
	{0xB3AA97A36B818946ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acast.gad"},
	{0xB3AA9A98BE7D6011ULL, "hlac_lt_0_lowtown_a3_n_c_131072_2x.dds"},
	{0xB3AAA9400869998DULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden3.msh"},
	{0xB3AABB400869B813ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden3.mmh"},
	{0xB3AB5B578C712768ULL, "mcr371_orsino6083333.cl"},
	{0xB3AB7A3A98CCE395ULL, "llac_w_sundermount_-327678_0.dds"},
	{0xB3ABCA04FAB7C7BEULL, "vxlt_0_lowtown_docks_a3_cx6_0_0.vlm"},
	{0xB3AC1A41104954F9ULL, "vxca_caverns2_6hu_0_0.vlm"},
	{0xB3AC2DE13C5E789EULL, "hlac_lt_gallowstemplar_f_2383_3z.dds"},
	{0xB3AC87F43B528711ULL, "vxlt_gallowstemplar_45p_0_0.vlm"},
	{0xB3AC92E6B02B1F01ULL, "vxca_caverns2_78u_0_1.vlm"},
	{0xB3ACF8BF986C306DULL, "isa111_isabela6068911.cl"},
	{0xB3AD076BF0269B00ULL, "lt_0_lowtown_docks_31_0_lc.mmh"},
	{0xB3ADCA1FB9D2DC21ULL, "622267183.ogg"},
	{0xB3AE255E2B51E50EULL, "vxht_mansion2_2qb_0_0.vlm"},
	{0xB3AE65D44EE2DD0DULL, "lgt101_mag_escp_terrie6156642.cl"},
	{0xB3AED53BCF8FFC95ULL, "one280_merin6105434.cl"},
	{0xB3B0BBACB4326BE3ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.cess.gad"},
	{0xB3B0E80453E878CFULL, "6114809_m.fxe"},
	{0xB3B11E9EE0AD7478ULL, "554752866.ogg"},
	{0xB3B168E2DC1AF5A8ULL, "vxw_encounter3_69_0_0.vlm"},
	{0xB3B17CC4B3A1187AULL, "hlac_w_woundedcoast_327685_0x.dds"},
	{0xB3B192A36B879F25ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acast.evt"},
	{0xB3B1F0DFF8E55BF0ULL, "vxlt_0_lowtown_docks_cto_0_0.vlm"},
	{0xB3B2D6267BDB679CULL, "gen_im_trs_bst_spd_silk_03.uti"},
	{0xB3B331B90C6E7522ULL, "647156182.ogg"},
	{0xB3B3412BAACC8AECULL, "6098089_m.fxe"},
	{0xB3B372FB63F0DAEDULL, "plt_cod_itm_qunari_talisman.plo"},
	{0xB3B39CB38FA6C156ULL, "hlac_w_bonepitdestroyed_a3_0_5y.dds"},
	{0xB3B3CBD341C47222ULL, "vxlt_darktown_4ka_0_0.vlm"},
	{0xB3B3F3461B76BF01ULL, "6167593_m.fxe"},
	{0xB3B441E922EE1B78ULL, "hlac_lt_0_lowtown_docks_a3_196606_0y.dds"},
	{0xB3B48D239255490BULL, "31665491.ogg"},
	{0xB3B49F12AB949E57ULL, "668760534.ogg"},
	{0xB3B4ACBA6A693E36ULL, "isa120_hayder6066875.cl"},
	{0xB3B5B837180BEF14ULL, "qun120_arishok_guard6049224.cl"},
	{0xB3B65BE6B110F01FULL, "vxlt_hangedman_n_2yn_0_0.vlm"},
	{0xB3B6867B96E09B60ULL, "6110433_m.fxe"},
	{0xB3B68857A9933960ULL, "vxw_bonepit_1z1_0_0.vlm"},
	{0xB3B6D79D23FE0DF5ULL, "vxht_keep_n_6mn_0_0.vlm"},
	{0xB3B71B024DE15769ULL, "6093522_m.fxe"},
	{0xB3B73E0E1D3A05F7ULL, "411421400.ogg"},
	{0xB3B798ACB4384EC8ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.cess.evt"},
	{0xB3B7EFC444F401DDULL, "dae000_loyal_mage6165224.cl"},
	{0xB3B953FB46E79584ULL, "487016538.ogg"},
	{0xB3B976CBA180FFE1ULL, "gen00fl_varric6033580.cl"},
	{0xB3B9863ACC653AE0ULL, "plt_cod_hst_kirk_hightown.plo"},
	{0xB3B9A45EEA6C40E2ULL, "llac_lt_0_lowtown_a3_n_c_65536_11.dds"},
	{0xB3B9BBDED82D334DULL, "vxw_sundermount_2cs_0_0.vlm"},
	{0xB3BAA9E513659B39ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomed_d.dds"},
	{0xB3BAC214524A8C18ULL, "hlac_dr_0_deeproads_5843_6x.dds"},
	{0xB3BAC31B4F33F3E5ULL, "dre000_sibling_has_blight6201758.cl"},
	{0xB3BB2B6D915C23E1ULL, "hlac_w_encounter3_-262146_0x.dds"},
	{0xB3BB42EB3FDC10EBULL, "cloud_f_oil.dds"},
	{0xB3BB56473FADEC96ULL, "6125380_f.fxe"},
	{0xB3BBE4259BA023D7ULL, "6111482_m.fxe"},
	{0xB3BC084A718174DEULL, "ash_01_l3.mao"},
	{0xB3BC1FBA9D3B100DULL, "vxca_caverns2_71k_0_0.vlm"},
	{0xB3BCA9C0781BC7B2ULL, "hlac_ht_chantry_2534_18x.dds"},
	{0xB3BCC692B4576604ULL, "w_bonepitmining_a2_135_0.phy"},
	{0xB3BD824C1B9B33DFULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax103a_0d.dds"},
	{0xB3BDFA69F38F0DF8ULL, "731107733.ogg"},
	{0xB3BE0372C161AE41ULL, "hlac_lt_backalley_g_1_5y.dds"},
	{0xB3BE4A9821D1FA61ULL, "6114854_m.fxe"},
	{0xB3BE58442068402FULL, "446786547.ogg"},
	{0xB3BE9368A37E44D3ULL, "vxlt_gallowstemplar_n_6si_0_0.vlm"},
	{0xB3BEB2A31F8A83CCULL, "vxlt_gallowscourtyard_74w_0_0.vlm"},
	{0xB3BEFB15D3BCD3DEULL, "qun160_vashothlead6053881.cl"},
	{0xB3BF06E25E4F9114ULL, "vxw_encounter2_22_0_0.vlm"},
	{0xB3BF1F3681DF4743ULL, "llac_lt_0_lowtown_65536_7.dds"},
	{0xB3BF9AF9EF08B101ULL, "vxht_keep_67w_0_0.vlm"},
	{0xB3C06CA22DEA0A11ULL, "cc_str500d.dds"},
	{0xB3C1E559034FADACULL, "6000880_m.fxe"},
	{0xB3C2E8A57FC64375ULL, "vxlt_0_lowtown_a2_ejc_0_0.vlm"},
	{0xB3C3208452566850ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutb.mao"},
	{0xB3C3248452566F63ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutb.mmh"},
	{0xB3C33284525686BDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarpoutb.msh"},
	{0xB3C3CDB855836F7FULL, "w_blightlands_52_0.rml"},
	{0xB3C40C066730B5E8ULL, "gen00fl_isabela6100032.cl"},
	{0xB3C44D635E29AA67ULL, "dae000ip_crowd_mf6_talk10.utp"},
	{0xB3C49416E72287B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\runforward.ani"},
	{0xB3C4A1CDBF8F9150ULL, "hlac_lt_darktown_-1_82z.dds"},
	{0xB3C50B625289E8A2ULL, "tv_door_02.dds"},
	{0xB3C5497A15F42675ULL, "vxht_keep_n_6hm_0_0.vlm"},
	{0xB3C6532F569072DBULL, "vxlt_warehouse_n_q6_0_0.vlm"},
	{0xB3C7D878EC674595ULL, "gen00fl_aveline6152532.cl"},
	{0xB3C7DBA8C2064E87ULL, "genip_node_e3.utp"},
	{0xB3C8141E64D2E352ULL, "vxlt_gallowstemplar_n_6sl_0_0.vlm"},
	{0xB3C820332AF771EFULL, "vxht_keep_n_3xr_0_0.vlm"},
	{0xB3C82DBBCA234BCDULL, "vxw_sundermount_2dr_0_0.vlm"},
	{0xB3C8B95D467F7A06ULL, "dae000ip_crowd_lm4_talk13.utp"},
	{0xB3C8EBC843755F69ULL, "6032048_m.fxe"},
	{0xB3C949DD00C6B7EAULL, "44214047.ogg"},
	{0xB3CAB56C87AFD073ULL, "darktown_fem_fereldan_sta6152799.cl"},
	{0xB3CABD92A4F28543ULL, "hlac_lt_0_lowtown_a2_-1_5y.dds"},
	{0xB3CC233F5CDF6896ULL, "vxw_woundedcoast_1go_0_0.vlm"},
	{0xB3CD1F100D2DB724ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgte.msh"},
	{0xB3CD31100D2DD58DULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgte.mao"},
	{0xB3CD35100D2DDCC6ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgte.mmh"},
	{0xB3CDF5A32581135FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_long.msh"},
	{0xB3CDFBA325811D01ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_long.mmh"},
	{0xB3CE04438FFE59F1ULL, "follower_banter6155847.cl"},
	{0xB3CE07A3258131A2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_long.mao"},
	{0xB3CE0CD6E0FEC981ULL, "vxht_keep_60v_0_0.vlm"},
	{0xB3CE7A2674613BC4ULL, "dae000_gamlen6005808.cl"},
	{0xB3CE983B7F37360FULL, "llac_lt_0_lowtown_a2_n_65538_0.dds"},
	{0xB3D04D1188F8BFA5ULL, "6197954_m.fxe"},
	{0xB3D114156B7DB7A1ULL, "hlac_w_woundedcoast_327680_0x.dds"},
	{0xB3D1D66F0E5F62FCULL, "vxht_mansion2b_1rn_0_0.vlm"},
	{0xB3D2120D29957AB3ULL, "88142624.ogg"},
	{0xB3D29C16E72ED2A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\runforward.evt"},
	{0xB3D31F322BB1DE43ULL, "vxlt_blackemporium_us_0_0.vlm"},
	{0xB3D331338FC76D5AULL, "hc_star.mao"},
	{0xB3D361103A9EC366ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat1x1.phy"},
	{0xB3D3BB5707EA3EF5ULL, "vxht_keep_n_6cl_0_0.vlm"},
	{0xB3D3DB3AA1048FB3ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench2f.ani"},
	{0xB3D44A46F9BE0145ULL, "cd_mag330_walter.bnk"},
	{0xB3D48956357867ADULL, "6127580_m.fxe"},
	{0xB3D4B7443FE6B1ECULL, "ht_mansion_n_3373.anb"},
	{0xB3D4C6EF1E7F0CC7ULL, "6136661_m.fxe"},
	{0xB3D599B210D1B6FAULL, "hlac_lt_0_lowtown_a2_0_2y.dds"},
	{0xB3D5A8CA8D31FD03ULL, "hla_ht_0_hightown_a2_131070_0y.dds"},
	{0xB3D69F98BC19644DULL, "vxw_sundermount_2iq_0_0.vlm"},
	{0xB3D82C100D37D491ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgte.phy"},
	{0xB3D8747E6CCBEE14ULL, "hlac_lt_0_lowtown_n_-65537_4y.dds"},
	{0xB3D8BAEA49E4F042ULL, "515998762.ogg"},
	{0xB3D8BD16E73376F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\runforward.gad"},
	{0xB3D90E7FA907B98CULL, "6114064_m.fxe"},
	{0xB3D93640E03ABCF1ULL, "w_blightlands_152_0.mmh"},
	{0xB3D94240E03AD092ULL, "w_blightlands_152_0.mao"},
	{0xB3D95040E03AE8AFULL, "w_blightlands_152_0.msh"},
	{0xB3D96D8B5F2DB3AEULL, "850393835.ogg"},
	{0xB3D97123DB68EE02ULL, "w_encounter2_camp_61_0.rml"},
	{0xB3DA75C6E5B3E805ULL, "227516495.ogg"},
	{0xB3DB32F210483F8BULL, "682643912.ogg"},
	{0xB3DB7036D039477AULL, "vxht_mansion2b_35o_0_0.vlm"},
	{0xB3DBC9A2116192E2ULL, "6205168_m.fxe"},
	{0xB3DBCE10A5371E49ULL, "vxlt_0_lowtown_docks_bze_0_0.vlm"},
	{0xB3DC1A62529CE140ULL, "tv_door_02.mao"},
	{0xB3DC8F8688BC178DULL, "x_partypicker_3567_ll.mmh"},
	{0xB3DCA75DA2E34D07ULL, "vxca_caverns1_mines_ck1_0_0.vlm"},
	{0xB3DD1FF726C977DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_loop.ani"},
	{0xB3DE6EC3578AAB45ULL, "vxca_caverns1_mines_iiw_0_0.vlm"},
	{0xB3DEA7DA37C9726BULL, "26628571.ogg"},
	{0xB3DEC1F27EED2A86ULL, "vxlt_warehouse_n_e3_0_0.vlm"},
	{0xB3DF278D01DF702FULL, "6181924_m.fxe"},
	{0xB3DF523A7034163BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1.evt"},
	{0xB3E003EA78651FEFULL, "6110972_f.fxe"},
	{0xB3E016AF24F8F6F2ULL, "6131218_m.fxe"},
	{0xB3E02490E45F4661ULL, "dae000_arishok6078460.cl"},
	{0xB3E0484C00557B7CULL, "vxht_mansion2b_1uo_0_0.vlm"},
	{0xB3E07223DB6F0EB6ULL, "w_encounter2_camp_61_0.phy"},
	{0xB3E0855E8BA6A2F8ULL, "llac_lt_0_lowtown_2_1.dds"},
	{0xB3E0D5791D40571BULL, "isa120ip_at_chantry.utp"},
	{0xB3E20F59815AEE4DULL, "vxlt_0_lowtown_docks_a3_d07_0_1.vlm"},
	{0xB3E220865E792C2CULL, "vxw_woundedcoast_175_0_0.vlm"},
	{0xB3E228E2C6628A16ULL, "w_bonepitdestroyed_a3_181_0.msh"},
	{0xB3E232E2C6629B14ULL, "w_bonepitdestroyed_a3_181_0.mmh"},
	{0xB3E236E2C662A227ULL, "w_bonepitdestroyed_a3_181_0.mao"},
	{0xB3E2445ECB2A661EULL, "583597887.ogg"},
	{0xB3E255D0A0F8D94BULL, "6027542_f.fxe"},
	{0xB3E2C15A36A45B83ULL, "vxht_keep_4nd_0_0.vlm"},
	{0xB3E3155A988A586EULL, "vxdr_0_deeproads_es9_0_1.vlm"},
	{0xB3E326B0246277FDULL, "llac_ht_playershouse_n_855_35.dds"},
	{0xB3E37CBEDE2AD7EEULL, "6097411_m.fxe"},
	{0xB3E3D4FE9396B08CULL, "art\\vfx\\combat\\vfx_glb_bld_imp_maj_c.impa.ani"},
	{0xB3E4023574AFABD9ULL, "vxht_bartrand_n_6sc_0_0.vlm"},
	{0xB3E487345D6CE38FULL, "hlac_w_woundedcoast_2_0z.dds"},
	{0xB3E4B230A1CE83D9ULL, "246692545.ogg"},
	{0xB3E51175AE0F7CCDULL, "vxw_sundermount_2rp_0_0.vlm"},
	{0xB3E5118DA4C5855AULL, "6167671_m.fxe"},
	{0xB3E532491A7AFE7CULL, "mag291_demon_marethari6028924.cl"},
	{0xB3E54B3A7038767CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1.gad"},
	{0xB3E567F5F100F6D9ULL, "gen00fl_fenris6126822.cl"},
	{0xB3E5CE1843261265ULL, "ser120_ghyslain6104197.cl"},
	{0xB3E66A9109B7F55EULL, "vxlt_hangedman_n_244_0_0.vlm"},
	{0xB3E67668D2F4806CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs08a_0s.dds"},
	{0xB3E69C3DF40375E7ULL, "416642864.ogg"},
	{0xB3E6A58CC156B4F1ULL, "zz_vault_debug6203245.cl"},
	{0xB3E6B4B5807E143DULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightsand_da.dds"},
	{0xB3E6CE1E940AA89BULL, "vxlt_warehouse_5s_0_0.vlm"},
	{0xB3E74DD9010301C7ULL, "418429728.ogg"},
	{0xB3E751598A5BE884ULL, "plt_cod_ltr_otl_walter.plo"},
	{0xB3E751E1B362508DULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_cull.phy"},
	{0xB3E7CA831667C083ULL, "follower_banter6154042.cl"},
	{0xB3E8330E7781CF16ULL, "vxlt_undercitydungeon_y4_0_0.vlm"},
	{0xB3E89C5BF57BE284ULL, "552059705.ogg"},
	{0xB3E8B23AA116A65DULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench2f.gad"},
	{0xB3E9699012CC50E5ULL, "gen00fl_fenris6106639.cl"},
	{0xB3E971D1BC4A1160ULL, "6148394_f.fxe"},
	{0xB3EA0393469ADED2ULL, "hlac_lt_backalley_n_65537_5x.dds"},
	{0xB3EA28B96F3754B4ULL, "vxlt_warehouse_n_1da_0_0.vlm"},
	{0xB3EA3B7F0357AB62ULL, "6208668_m.fxe"},
	{0xB3EA5B23DB773C57ULL, "w_encounter2_camp_61_0.msh"},
	{0xB3EA6D23DB775ABAULL, "w_encounter2_camp_61_0.mao"},
	{0xB3EA7123DB7761C9ULL, "w_encounter2_camp_61_0.mmh"},
	{0xB3EACCA7E46AC6FBULL, "mag360_thrask6053537.cl"},
	{0xB3EAE1FE939CE5CEULL, "art\\vfx\\combat\\vfx_glb_bld_imp_maj_c.impa.gad"},
	{0xB3EB7DB7F97750ADULL, "gen00fl_aveline6115432.cl"},
	{0xB3EBD085C85D6C3BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopea.msh"},
	{0xB3EBE285C85D8A86ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopea.mao"},
	{0xB3EBEE85C85D9F25ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopea.mmh"},
	{0xB3EBF232AE5AD728ULL, "vxht_keep_n_6y3_0_0.vlm"},
	{0xB3EBFCF21273EF87ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb105a_0n.dds"},
	{0xB3EC350F3B4409B2ULL, "t1_mue_bk2.tnt"},
	{0xB3EC4A3A703E937AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v1.ani"},
	{0xB3EC754487C894D4ULL, "vxca_caverns2prc_9lf_0_0.vlm"},
	{0xB3ECC6B61F2AF2D7ULL, "6108589_m.fxe"},
	{0xB3ED06CE242970B8ULL, "vxlt_warehouse_n_hz_0_0.vlm"},
	{0xB3ED17CD63917C95ULL, "6083840_m.fxe"},
	{0xB3ED653FBF7433DCULL, "art\\characters\\creatures\\arh_abm_rev\\c_abm_body.mao"},
	{0xB3EDA22AAF6B46D0ULL, "dae000_flemeth6013392.cl"},
	{0xB3EDAFF8E3697010ULL, "vxlt_gallowstemplar_4vd_0_0.vlm"},
	{0xB3EDBCAD5A3208C5ULL, "dae011_gossips6143013.cl"},
	{0xB3EDD7355783D6ADULL, "follower_banter6172909.cl"},
	{0xB3EE9675DA4D153AULL, "154972020.ogg"},
	{0xB3F050D8D63B5605ULL, "609721824.ogg"},
	{0xB3F065E9AC94F3D1ULL, "670164964.ogg"},
	{0xB3F075F7F3F1843CULL, "vxht_bartrand_n_3ue_0_0.vlm"},
	{0xB3F07DBB04006C28ULL, "6098719_m.fxe"},
	{0xB3F08643291DA5ACULL, "570655884.ogg"},
	{0xB3F120113197CA0CULL, "mer221_sabin6047462.cl"},
	{0xB3F192994D4C2EEDULL, "follower_banter6141789.cl"},
	{0xB3F1D27941051D02ULL, "vxlt_0_lowtown_f1b_0_0.vlm"},
	{0xB3F1DCFE93A2FBADULL, "art\\vfx\\combat\\vfx_glb_bld_imp_maj_c.impa.evt"},
	{0xB3F1EA791D4EEE3EULL, "isa120ip_at_chantry.ncs"},
	{0xB3F20A295C2FEC3DULL, "6147793_m.fxe"},
	{0xB3F246F726DC1665ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_loop.gad"},
	{0xB3F2CD69C6802F72ULL, "qun221_petrice6093200.cl"},
	{0xB3F2FCFAFE56FBBAULL, "hlac_lt_undercityencounter_-1_43y.dds"},
	{0xB3F30BCCD9AE3555ULL, "llac_lt_0_lowtown_docks_a3_n_-262143_0.dds"},
	{0xB3F344CD46D88695ULL, "gen00fl_aveline6122166.cl"},
	{0xB3F36BD3242D5508ULL, "vxht_mansion2b_27j_0_0.vlm"},
	{0xB3F370503E331C85ULL, "gen00fl_merrill6131350.cl"},
	{0xB3F3875E4B102BB2ULL, "484119083.ogg"},
	{0xB3F4057E5EB0CB49ULL, "atl_startmenu_dxt5_07.dds"},
	{0xB3F46581FC30A841ULL, "llac_lt_undercityencounter_-1_23.dds"},
	{0xB3F47C331558D099ULL, "lt_0_lowtown_n_26_0.anb"},
	{0xB3F518A8D729EB8AULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medp_s.dds"},
	{0xB3F5E3DEB10ECD2FULL, "hlac_w_sundermount_-262147_0y.dds"},
	{0xB3F681A443459FCBULL, "6113996_m.fxe"},
	{0xB3F6D58193DF8BB5ULL, "pro000_merchant6051643.cl"},
	{0xB3F70469431395C7ULL, "6087433_m.fxe"},
	{0xB3F72517CE266233ULL, "vxdr_primevalentrance_16u_0_1.vlm"},
	{0xB3F78ADCD2454A35ULL, "6125573_m.fxe"},
	{0xB3F89CAE6C725DC5ULL, "hlac_lt_0_lowtown_a2_n_-65536_3x.dds"},
	{0xB3F8C2DF086D0110ULL, "face_face_p3.psh.fxo.dx11.assembly"},
	{0xB3F9511A56394264ULL, "685688513.ogg"},
	{0xB3F966D3AF06905AULL, "narrative_varric6172521.cl"},
	{0xB3F9750D4678A67CULL, "6056609_m.fxe"},
	{0xB3F99C1A30849796ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v1.anb"},
	{0xB3F9FCA325A65A7EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_long.phy"},
	{0xB3FAC237CB4763C1ULL, "vxlt_darktown_681_0_0.vlm"},
	{0xB3FAD676331CD2E3ULL, "vxdr_0_deeproads_d16_0_0.vlm"},
	{0xB3FB4CD76D85D364ULL, "1000450134.ogg"},
	{0xB3FB527D3B76DC45ULL, "vxca_caverns1_mines_icu_0_0.vlm"},
	{0xB3FBA5AC62D95B86ULL, "vxlt_warehouse_n_o5_0_0.vlm"},
	{0xB3FBABEFBAD253A9ULL, "6118986_m.fxe"},
	{0xB3FC299B89D4E129ULL, "vxlt_0_lowtown_docks_n_c0a_0_0.vlm"},
	{0xB3FC47083CA757D5ULL, "vxca_cavern1_haunted_eh8_0_0.vlm"},
	{0xB3FD0522672871F4ULL, "hlac_w_sundermount_-196610_0y.dds"},
	{0xB3FD12A9A0A51030ULL, "mag321st_comtesse_de_launce.stg"},
	{0xB3FD3A4B923B1F04ULL, "6102820_f.fxe"},
	{0xB3FD3AD612AF7F8AULL, "vxlt_hangedman_ro_0_0.vlm"},
	{0xB3FD4F13CD9ACD22ULL, "6125559_f.fxe"},
	{0xB3FD54D6CACA3992ULL, "hlac_ht_chantry_n_2534_24z.dds"},
	{0xB3FD9BCD78DA7429ULL, "gen_im_trs_chn_rng_r2.uti"},
	{0xB3FDFA468A9ABA3FULL, "mcr310_argument6072312.cl"},
	{0xB3FE2459E0870789ULL, "llac_lt_0_lowtown_a3_n_c_1_5.dds"},
	{0xB3FF4F4FB4B08B17ULL, "qun200_party_comment6090326.cl"},
	{0xB3FF8AEF1AAC86CBULL, "6106683_f.fxe"},
	{0xB3FFB6B5FC8F562EULL, "hlac_lt_gallowscourtyard_-65538_1y.dds"},
	{0xB3FFBA5E55E86046ULL, "follower_banter6167854.cl"},
	{0xB3FFBEE0388814CEULL, "vxlt_0_lowtown_feb_0_0.vlm"},
	{0xB40069AF18B8E3FCULL, "mag131_taraneh.cnv"},
	{0xB4010A081AE7AB02ULL, "w_bonepitdestroyed_a3_196_0.msh"},
	{0xB4010C081AE7AEB8ULL, "w_bonepitdestroyed_a3_196_0.mmh"},
	{0xB40118081AE7C35BULL, "w_bonepitdestroyed_a3_196_0.mao"},
	{0xB401C1C8FE9EECD1ULL, "6050017_m.fxe"},
	{0xB401CF6532807ED0ULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_9.pwk"},
	{0xB401DDB016236D88ULL, "vxht_mansion2b_20k_0_0.vlm"},
	{0xB401F52F91FBADCDULL, "vxw_sundermount_2xn_0_0.vlm"},
	{0xB402527842AA56B6ULL, "w_woundedcoast_107_0.phy"},
	{0xB402936B8F086316ULL, "vxlt_hangedman_n_2zr_0_0.vlm"},
	{0xB40341FC2EF4CFB0ULL, "rdr111dg_harbormaster6016149.cl"},
	{0xB4036E0902300487ULL, "qun111_ketojan6041129.cl"},
	{0xB4038BAF18BB534CULL, "mag131_taraneh.bnk"},
	{0xB40394AA086A9A2EULL, "6095087_f.fxe"},
	{0xB4041A3B136A7A05ULL, "561125553.ogg"},
	{0xB404A7C82A04C0D8ULL, "llac_w_encounter3_-393214_0.dds"},
	{0xB404F0B855BB41E3ULL, "w_blightlands_52_0.anb"},
	{0xB40563FCD5FA21CBULL, "lgt200cr_kanky_hammertoe.utc"},
	{0xB4056CB440B59E29ULL, "gen00fl_aveline6115865.cl"},
	{0xB405A5D8F500AAA5ULL, "vxht_keep_69u_0_0.vlm"},
	{0xB4063B98F8D3974DULL, "vxht_chantry_1yf_0_0.vlm"},
	{0xB4069EC2E98AF1BCULL, "gen00fl_varric6093686.cl"},
	{0xB406C75060506D7FULL, "mrl150_merrill6150743.cl"},
	{0xB407090C52F1AE1FULL, "1033774239.ogg"},
	{0xB4071CDAFDCB1ECAULL, "t1_mub_te1.tnt"},
	{0xB4078767D9D660E9ULL, "6128729_m.fxe"},
	{0xB40791021241FE50ULL, "vxca_caverns1_cmm_0_0.vlm"},
	{0xB4079F1A3090DA01ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v1.mmh"},
	{0xB407DC36A54B5714ULL, "lt_0_lowtown_docks_a3_n_39_0_lc.mmh"},
	{0xB40808536D91EA2AULL, "gen00fl_sebastian6188110.cl"},
	{0xB40891DC6E37CFE4ULL, "vxlt_undercityencounter_12f_0_1.vlm"},
	{0xB408B143CF68AC95ULL, "genip_trap_bolt_acid.utp"},
	{0xB408B7F53F12E707ULL, "qcr000_varric6043181.cl"},
	{0xB408D23A23A25514ULL, "6167040_m.fxe"},
	{0xB40909B646BD4F9BULL, "w_woundedcoast_79_0.mmh"},
	{0xB4090FDE1EC57969ULL, "274548517.ogg"},
	{0xB40915B646BD6438ULL, "w_woundedcoast_79_0.mao"},
	{0xB409240012F29B3BULL, "6016488_m.fxe"},
	{0xB40927B646BD8285ULL, "w_woundedcoast_79_0.msh"},
	{0xB40948532512EB63ULL, "vxdr_0_deeproads_d49_0_0.vlm"},
	{0xB409895E4115A33CULL, "6071088_m.fxe"},
	{0xB409C45A2D6CF4C5ULL, "vxca_caverns1_mines_ixt_0_0.vlm"},
	{0xB409EA880815A1B8ULL, "vxlt_warehouse_n_rx_0_0.vlm"},
	{0xB40A178954CF7406ULL, "vxlt_warehouse_n_r6_0_0.vlm"},
	{0xB40B1FEA55BE5EBFULL, "dae000_sabina6151870.cl"},
	{0xB40BA1E753D0413CULL, "hlac_lt_0_lowtown_a3_n_c_1_8z.dds"},
	{0xB40BE9F7F4FA3BDDULL, "6015062_f.fxe"},
	{0xB40C04F9064004E0ULL, "84833857.ogg"},
	{0xB40C3B7842B28457ULL, "w_woundedcoast_107_0.msh"},
	{0xB40C4D7842B2A2BAULL, "w_woundedcoast_107_0.mao"},
	{0xB40C517842B2A9C9ULL, "w_woundedcoast_107_0.mmh"},
	{0xB40C5A9A564DFBEBULL, "hlac_lt_0_lowtown_a2_n_-131072_4x.dds"},
	{0xB40C5FBEF6BC2482ULL, "ser191_emeric6120053.cl"},
	{0xB40C8477CBF84774ULL, "37956772.ogg"},
	{0xB40C84F8E3820AC0ULL, "gen00fl_isabela6075242.cl"},
	{0xB40CA5A7DA3BDD39ULL, "vxlt_shop_cd_0_0.vlm"},
	{0xB40CF9408DCEB8CAULL, "mag291_demon_marethari6028939.cl"},
	{0xB40D49E208F4CBAAULL, "6094476_m.fxe"},
	{0xB40D869CAAB6AFCFULL, "109314447.ogg"},
	{0xB40D897651F2A89EULL, "vxht_0_hightown_b67_0_0.vlm"},
	{0xB40DB0EC2A5CAFC8ULL, "175662160.ogg"},
	{0xB40E71F1ED063C1CULL, "845472026.ogg"},
	{0xB40FC82413B859B6ULL, "hightown_market_guard_sta6169952.cl"},
	{0xB41004FCD816E3DBULL, "hlac_ht_keep_n_4519_16y.dds"},
	{0xB410663B018473B7ULL, "stage_hop.cut"},
	{0xB4108D173C8F76CFULL, "6126417_m.fxe"},
	{0xB4110E8C55A1981CULL, "isa311_party_comments6091974.cl"},
	{0xB411C1B855C57C6FULL, "w_blightlands_52_0.mao"},
	{0xB411CD4CB11A7E5DULL, "6041132_m.fxe"},
	{0xB411CDB855C5900CULL, "w_blightlands_52_0.mmh"},
	{0xB411D3B855C59AFEULL, "w_blightlands_52_0.msh"},
	{0xB411FCF22AA4911DULL, "236729970.ogg"},
	{0xB4129929B5D082C8ULL, "framebuffereffects_linearizedepth_p0.psh.fxo.dx11"},
	{0xB413033D08695AEFULL, "6073935_m.fxe"},
	{0xB4138A5AA28F18E1ULL, "6097208_f.fxe"},
	{0xB41428171B201181ULL, "hlac_dr_slavers_3214_9z.dds"},
	{0xB4142FEA3402B5D0ULL, "123530080.ogg"},
	{0xB414439A17723974ULL, "vxht_mansion_n_6oi_0_0.vlm"},
	{0xB414594DC60ED61FULL, "238296904.ogg"},
	{0xB414B6CED738E155ULL, "dae000_male_templar_mob6174372.cl"},
	{0xB41501AEBEE50C01ULL, "971662338.ogg"},
	{0xB415065E5C4B843CULL, "193188921.ogg"},
	{0xB415125FACB5B024ULL, "vxlt_undercitydungeon_yv_0_0.vlm"},
	{0xB415FB29D09F0820ULL, "6043654_m.fxe"},
	{0xB4172191942D5F51ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_chantrybars.msh"},
	{0xB4172442A281C16DULL, "190324053.ogg"},
	{0xB4172F91942D776CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_chantrybars.mao"},
	{0xB4173B91942D8B0FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9uw_chantrybars.mmh"},
	{0xB41834955BFF1736ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation2.ani"},
	{0xB4183A8513A63A9DULL, "474079145.ogg"},
	{0xB41840A78239D46BULL, "twopassblurcochorizontl.psh.fxo"},
	{0xB4185C64FA0BBA38ULL, "vxlt_warehouse_n_yy_0_0.vlm"},
	{0xB418896646C58C86ULL, "vxlt_warehouse_n_y7_0_0.vlm"},
	{0xB418936F463FBD13ULL, "6218765_m.fxe"},
	{0xB4191C2B2B548DB7ULL, "vxdr_primevaltomb_77i_0_0.vlm"},
	{0xB4192604F25411A5ULL, "611361728.ogg"},
	{0xB419B23C6DCD710CULL, "vxht_mansion2_1wi_0_0.vlm"},
	{0xB419F7E1D0801E4FULL, "752576762.ogg"},
	{0xB41AB0550E0C4742ULL, "vxht_0_hightown_ceo_0_0.vlm"},
	{0xB41B366CEBF79316ULL, "mag130cr_slaver_mage.utc"},
	{0xB41B458DF2CCBF99ULL, "vxlt_0_lowtown_docks_n_bcq_0_0.vlm"},
	{0xB41B975DB61CF455ULL, "dae000_bartrand6027526.cl"},
	{0xB41BC009D7FC3035ULL, "hlac_lt_0_lowtown_n_65537_4y.dds"},
	{0xB41C35E6E01BE1DBULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic.mmh"},
	{0xB41C41E6E01BF678ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic.mao"},
	{0xB41C53E6E01C14C5ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic.msh"},
	{0xB41CCC864FDBB94FULL, "zz_approval_debug6003835.cl"},
	{0xB41CDA304ECF5069ULL, "vxlt_gallowstemplar_4qd_0_0.vlm"},
	{0xB41D0A195B3F1138ULL, "vxlt_darktown_4uq_0_0.vlm"},
	{0xB41D8592032FD69BULL, "134725061.ogg"},
	{0xB41DB880C3EA884EULL, "lowtown_fem_prostitute_sta6161619.cl"},
	{0xB41E32591D5DC4ADULL, "vxca_cavern1_haunted_eib_0_1.vlm"},
	{0xB41E4D8EF03A12A2ULL, "6153654_m.fxe"},
	{0xB41E85073C63D9E4ULL, "cod_bks_lite_tombs_1.utp"},
	{0xB41E9E5D92D72205ULL, "vxw_bonepit_1rh_0_0.vlm"},
	{0xB41ED3AAACD315E8ULL, "ave200_donnic6111995.cl"},
	{0xB41EF8046970F630ULL, "6000968_f.fxe"},
	{0xB41F201036FDB2E8ULL, "6083445_m.fxe"},
	{0xB41F50454485DD6EULL, "llac_lt_blackemporium_1055_11.dds"},
	{0xB41F58F5B17D05BAULL, "w_sundermount_154_0.phy"},
	{0xB41F68065BE44397ULL, "art\\levels\\ztools\\default\\tilesets\\setint\\_textures\\xx_ceiling.dds"},
	{0xB41FD9792F487511ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_1_darren.gad"},
	{0xB42048AB1BF8B435ULL, "gen00fl_merrill6162251.cl"},
	{0xB420E8BE045165F9ULL, "vxlt_gallowstemplar_n_3s6_0_0.vlm"},
	{0xB420ED322EE132B1ULL, "hlac_ht_keep_n_3743_0z.dds"},
	{0xB4215E9DD421370CULL, "mcr310_argument6072713.cl"},
	{0xB4217BD3D3E4F8BEULL, "art\\vfx\\combat\\vfx_c_rw_bladestab_c.impa.ani"},
	{0xB4217D47BFDC3537ULL, "133492762.ogg"},
	{0xB421AE3F266669A7ULL, "294078427.ogg"},
	{0xB421D453CE2A32A8ULL, "dr_0_deeproads_5843.anb"},
	{0xB421E5F538A9D10EULL, "646919173.ogg"},
	{0xB4221C0FC1267CEBULL, "hlac_dr_0_deeproads_5842_8z.dds"},
	{0xB422615E16CC7AC7ULL, "vxlt_blackemporium_zr_0_0.vlm"},
	{0xB4226A0DE52B2F74ULL, "vrc221_bartrand6043385.cl"},
	{0xB4228642EDD40C21ULL, "9308311.ogg"},
	{0xB422EE7560FE0CEBULL, "qun120_arishok6051309.cl"},
	{0xB4236EAC5726E72AULL, "6244011.ogg"},
	{0xB42383C17286FEBEULL, "mag100_thrask6034825.cl"},
	{0xB42393A369D3A5ABULL, "vxca_cave_2cu_0_0.vlm"},
	{0xB42409C0BCECC273ULL, "vxht_0_hightown_a2_c3l_0_0.vlm"},
	{0xB42466598A8FBACBULL, "plt_cod_ltr_otl_walter.nss"},
	{0xB424C5264FA53031ULL, "1000495693.ogg"},
	{0xB425219A19BE9EDDULL, "6152904_m.fxe"},
	{0xB42531C9939D25C5ULL, "6072704_f.fxe"},
	{0xB42544DD6C3A79CCULL, "774683449.ogg"},
	{0xB425DD9947E466E6ULL, "vxht_0_hightown_a3_c3u_0_1.vlm"},
	{0xB425F620F4651E40ULL, "lgt220_shp_mag_guard6175903.cl"},
	{0xB426277086BAE776ULL, "vxca_caverns1_mines_j4l_0_0.vlm"},
	{0xB42652062D8D33FAULL, "227422391.ogg"},
	{0xB4267D8A7F2188CBULL, "34726243.ogg"},
	{0xB4268A8A98E138E8ULL, "vxca_cavern1_haunted_i70_0_0.vlm"},
	{0xB426DA792F4E95BFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_1_darren.ani"},
	{0xB4271267E7420051ULL, "841121316.ogg"},
	{0xB427251BEBA6E9C9ULL, "dre000_varric6036957.cl"},
	{0xB4278E081D4AA637ULL, "vxdr_primevaltomb_70h_0_0.vlm"},
	{0xB428052B6950B3ECULL, "hlac_lt_0_lowtown_n_0_2x.dds"},
	{0xB42824195FC3898CULL, "vxht_mansion2_1zn_0_0.vlm"},
	{0xB4282F3DD7C7B4CEULL, "llac_lt_0_lowtown_a2_n_c_-3_2.dds"},
	{0xB4285719F03979B5ULL, "hlac_lt_0_lowtown_a3_n_c_-3_0z.dds"},
	{0xB4288C9C0BF3FC22ULL, "6089764_m.fxe"},
	{0xB4288D3F4A3BD866ULL, "299786500.ogg"},
	{0xB42955E96BB647C4ULL, "6028529_m.fxe"},
	{0xB4296E96C857A0BBULL, "6004223_m.fxe"},
	{0xB42998E5D257F34BULL, "vxht_chantry_28t_0_0.vlm"},
	{0xB42998EA658148FBULL, "6060226_f.fxe"},
	{0xB429A3161E1F4199ULL, "982456100.ogg"},
	{0xB42A5D899008FEBDULL, "vxlt_0_lowtown_docks_n_b4x_0_0.vlm"},
	{0xB42AD806C4830F66ULL, "cli000cr_army_ds.ncs"},
	{0xB42B31289F9645BBULL, "vxdr_primevaltomb_1a8_0_1.vlm"},
	{0xB42B734D1681834DULL, "275712225.ogg"},
	{0xB42BEDF6651F2A7DULL, "131469478.ogg"},
	{0xB42BF317D57705A4ULL, "282836669.ogg"},
	{0xB42C3CA91407C82BULL, "zz_dae_debug6036588.cl"},
	{0xB42C5D955C100670ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation2.gad"},
	{0xB42C9491588BA0A3ULL, "hlac_ht_chantry_n_2534_45z.dds"},
	{0xB42D0E3CA6129C06ULL, "zz_approval_debug6003973.cl"},
	{0xB42D3D1D427024FBULL, "23574912.ogg"},
	{0xB42D629112762853ULL, "6069280_m.fxe"},
	{0xB42D76540A4B05F2ULL, "vxw_encounter2_camp_99_0_0.vlm"},
	{0xB42E78109149F0E9ULL, "w_bonepitmining_a2_152_0.phy"},
	{0xB42E7D33158A4A99ULL, "lt_0_lowtown_n_26_0.rml"},
	{0xB42E93D3D3EFAB8FULL, "art\\vfx\\combat\\vfx_c_rw_bladestab_c.impa.evt"},
	{0xB42EAEACBF34ADB4ULL, "6094979_m.fxe"},
	{0xB42EB11F19702A26ULL, "character_character_p63.psh.fxo"},
	{0xB42FFD607A00BC6DULL, "455812585.ogg"},
	{0xB4301C2A23391C08ULL, "6041252_m.fxe"},
	{0xB431A4EDB844B8B3ULL, "vxlt_0_lowtown_fga_0_0.vlm"},
	{0xB431B465CB163A1AULL, "art\\vfx\\textures\\vfx_grenade.dds"},
	{0xB431C164286C7A58ULL, "war_staminaregain.dds"},
	{0xB431CC0C03EAA485ULL, "639457300.ogg"},
	{0xB431ED8EA3951D48ULL, "vxht_mansion2b_328_0_0.vlm"},
	{0xB4324EA01EEFED1CULL, "6044754_m.fxe"},
	{0xB433F674528FACD0ULL, "vxw_encounter2_fj_0_0.vlm"},
	{0xB433F6AFCC5DAD41ULL, "6075287_m.fxe"},
	{0xB43414769DF6C40DULL, "6065003_f.fxe"},
	{0xB43457A82B41C348ULL, "gen00fl_aveline6139051.cl"},
	{0xB4348AD3A9FB9C42ULL, "723392059.ogg"},
	{0xB4351D97C08E8239ULL, "fex_bd_a_bs_cn.mmh"},
	{0xB4352797C08E9327ULL, "fex_bd_a_bs_cn.msh"},
	{0xB43536AC162DF655ULL, "vxht_chantry_n_2e6_0_0.vlm"},
	{0xB43594D3D3F5CC38ULL, "art\\vfx\\combat\\vfx_c_rw_bladestab_c.impa.gad"},
	{0xB43614AFF2C9B33CULL, "46825452.ogg"},
	{0xB4361B1CEE871BD0ULL, "w_woundedcoast_207_0.msh"},
	{0xB436291CEE8733EAULL, "w_woundedcoast_207_0.mmh"},
	{0xB4362D1CEE873A21ULL, "w_woundedcoast_207_0.mao"},
	{0xB43639F1478FA8E7ULL, "llac_dr_primevaltomb_658_2.dds"},
	{0xB43660C342F8E45AULL, "vxlt_undercityserialkiller_wy_0_0.vlm"},
	{0xB4366FAF18E67D4BULL, "mag131_taraneh.stg"},
	{0xB436B50DC00AEDB2ULL, "6165753_f.fxe"},
	{0xB436BF2285C5CF8EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bush_sa.dds"},
	{0xB4373124A1646F99ULL, "6178485_m.fxe"},
	{0xB43770B244446373ULL, "art\\vfx\\combat\\vfx_str_app_c.impa.ani"},
	{0xB4380AC2C44E0BCBULL, "vxht_chantry_25w_0_0.vlm"},
	{0xB4382947D6B8F099ULL, "vxlt_0_lowtown_docks_n_buw_0_0.vlm"},
	{0xB438566096E36F9AULL, "118064222.ogg"},
	{0xB43899D16526A412ULL, "263552148.ogg"},
	{0xB438F42A142DEF79ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comb_0s.dds"},
	{0xB43929BBB933330EULL, "vxht_0_hightown_n_cah_0_0.vlm"},
	{0xB43A2E340B98E0CEULL, "llac_lt_0_lowtown_n_-131072_3.dds"},
	{0xB43A977A0CE45465ULL, "6119534_m.fxe"},
	{0xB43AAE17B923572BULL, "6097362_m.fxe"},
	{0xB43B078BF798FF01ULL, "600576900.ogg"},
	{0xB43B115FA4134D0EULL, "937668625.ogg"},
	{0xB43B15DA7E10D219ULL, "vxht_keep_7a4_0_0.vlm"},
	{0xB43B90AD8739B0CEULL, "138956411.ogg"},
	{0xB43C2A2B5E3958E3ULL, "vxdr_0_deeproads_cud_0_1.vlm"},
	{0xB43C8DD88780950BULL, "character_character_p116.psh.fxo.dx11.assembly"},
	{0xB43CEA20D73AE404ULL, "wm_ico_hut.dds"},
	{0xB43D649FCF5A1CBAULL, "pro000_refugee_female6052096.cl"},
	{0xB43DF0CAF83B96F7ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_loop.gad"},
	{0xB43E10D117D5A03EULL, "zz_followers_banter6158472.cl"},
	{0xB43E8434250639DAULL, "6106124_m.fxe"},
	{0xB43EC26C96B3A09AULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyo.phy"},
	{0xB43EF4B55D8879D6ULL, "916353797.ogg"},
	{0xB43F4197E31D5294ULL, "hlac_lt_0_lowtown_a2_n_65534_6y.dds"},
	{0xB43FD88D711A5490ULL, "447213056.ogg"},
	{0xB4405AB8D78B3E44ULL, "173327258.ogg"},
	{0xB440775D4DBFD6ABULL, "vxca_cave_2iw_0_0.vlm"},
	{0xB44135340B97E9C3ULL, "6029028_m.fxe"},
	{0xB441B1FAB0D86336ULL, "491668611.ogg"},
	{0xB4425ED449AB65AEULL, "6084334_m.fxe"},
	{0xB44268514485C550ULL, "vxw_encounter2_ak_0_0.vlm"},
	{0xB442AB59283C39A1ULL, "6113301_m.fxe"},
	{0xB443894E35892A18ULL, "6092220_f.fxe"},
	{0xB443F71908898520ULL, "ave110cr_donnic_bandit_f.utc"},
	{0xB44468B2444EDFAAULL, "art\\vfx\\combat\\vfx_str_app_c.impa.evt"},
	{0xB444FD8B87B2A557ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_plain_1.mao"},
	{0xB44501746C75384DULL, "6189078_m.fxe"},
	{0xB44505CAF841D884ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_loop.evt"},
	{0xB44530B6B9CC806DULL, "mag360_thrask6053513.cl"},
	{0xB4458BCB38C6AB8BULL, "lt_0_lowtown_a2_20_0_lc.mmh"},
	{0xB445FD24583D93D3ULL, "6106868_m.fxe"},
	{0xB446076A231D763CULL, "l_var221ar_mansion_fake.wbd"},
	{0xB4469F3FF5D756FCULL, "170572763.ogg"},
	{0xB446BBC88482AF05ULL, "290437600.ogg"},
	{0xB44829611D826EE3ULL, "6172874_m.fxe"},
	{0xB4487098ADC08CDCULL, "vxlt_0_lowtown_n_h5e_0_0.vlm"},
	{0xB44893B62513C31FULL, "genip_blocker_wide_fire.utp"},
	{0xB448D62924B18831ULL, "vxht_0_hightown_a78_0_0.vlm"},
	{0xB449396635474FD1ULL, "310506862.ogg"},
	{0xB44A1FD9BC3E60BAULL, "gen00fl_aveline6128990.cl"},
	{0xB44A3A4800CA4371ULL, "6087659_m.fxe"},
	{0xB44AE4B0ED15EB6DULL, "6108051_m.fxe"},
	{0xB44B47B24454C61DULL, "art\\vfx\\combat\\vfx_str_app_c.impa.gad"},
	{0xB44BCFF8B1E72292ULL, "one140cr_charade6137635.cl"},
	{0xB44C18F14B1F5741ULL, "hlac_ca_caverns1_mines_24422_1z.dds"},
	{0xB44C2B48EB46154EULL, "6206922_f.fxe"},
	{0xB44C3B0B1980A5CCULL, "vxca_cavern1_haunted_ins_0_0.vlm"},
	{0xB44C6F130CFB1155ULL, "ave200_donnic6110627.cl"},
	{0xB44C9CC15DA2CFBEULL, "vxw_bonepitmining_a2_1ps_0_0.vlm"},
	{0xB44DB6F4ECAEC447ULL, "vxlt_blackemporium_uo_0_0.vlm"},
	{0xB44DC1D66807E549ULL, "vxw_bonepitdestroyed_a3_1xb_0_0.vlm"},
	{0xB44DDEEEB19E8B59ULL, "one100dg_lia6055154.cl"},
	{0xB44E336C6DFA60B4ULL, "5974496.ogg"},
	{0xB44E56FF5675BA06ULL, "dae000_arishok6187910.cl"},
	{0xB44F2382FD00ED5CULL, "art\\vfx\\textures\\vfx_crack_decal_add.dds"},
	{0xB44F2CEBF7A793C4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02bonetrack.mmh"},
	{0xB44F30EBF7A79A17ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02bonetrack.mao"},
	{0xB44F42DA7AE499D5ULL, "740916323.ogg"},
	{0xB44F42EBF7A7B966ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_02bonetrack.msh"},
	{0xB44F8AB82007424DULL, "zz_seb_debug6215785.cl"},
	{0xB44FAE535801822DULL, "6080104_m.fxe"},
	{0xB44FBE8AD5741176ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoortrim.mao"},
	{0xB44FCA8AD5742515ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoortrim.mmh"},
	{0xB44FCC8AD574288BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoortrim.msh"},
	{0xB45028ECFBE49B24ULL, "30715516.ogg"},
	{0xB45057D54F0422D9ULL, "art\\vfx\\models\\vfx_shield_uhm.msh"},
	{0xB4505A728CD24B9FULL, "follower_banter6163277.cl"},
	{0xB45061D54F0433C7ULL, "art\\vfx\\models\\vfx_shield_uhm.mmh"},
	{0xB450AAF8D698F748ULL, "6111391_m.fxe"},
	{0xB450AECDFF8DC96EULL, "hlac_ca_caverns2prc_5802_8z.dds"},
	{0xB450CFBC3FECE82BULL, "1073419284.ogg"},
	{0xB450DA2E367BDDD0ULL, "vxw_encounter2_lh_0_0.vlm"},
	{0xB451D2FC5D9D4230ULL, "314777173.ogg"},
	{0xB4520E8789394327ULL, "ser270_emeric6113607.cl"},
	{0xB4521DCAF84C8C7DULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\follower_em_loop.ani"},
	{0xB452234553A6849FULL, "hlac_lt_0_lowtown_a2_n_c_131070_3z.dds"},
	{0xB45298B5469AE04FULL, "vxht_chantry_n_261_0_0.vlm"},
	{0xB452F5117AA329F3ULL, "hlac_lt_foundry_n_320_9y.dds"},
	{0xB4530106253E476FULL, "isa211_isabela6081984.cl"},
	{0xB45301840FECFF63ULL, "follower_banter6155478.cl"},
	{0xB4538F0E89A80556ULL, "plt_cod_aow_spellcombo4.plo"},
	{0xB4546A8CD3172758ULL, "mag360_thrask6043551.cl"},
	{0xB45513889BC88164ULL, "art\\vfx\\combat\\vfx_r_brd_csong_act_c.mmh"},
	{0xB45519766982B25CULL, "6180627_m.fxe"},
	{0xB4557618BA291DB4ULL, "hlac_w_blightlands_65536_1y.dds"},
	{0xB4557E40FA45328FULL, "66577105.ogg"},
	{0xB45678D5FF9A6940ULL, "hm_dead_4.mor"},
	{0xB456AE6060102455ULL, "659217946.ogg"},
	{0xB456D35083C30610ULL, "dae000_arishok6067719.cl"},
	{0xB456EC451199A491ULL, "hlac_lt_gallowscourtyard_n_-65537_2y.dds"},
	{0xB45709B646FF7128ULL, "w_woundedcoast_79_0.rml"},
	{0xB4574233061FE3D2ULL, "6048720_f.fxe"},
	{0xB45842C4264190FDULL, "vxw_woundedcoast_1e6_0_0.vlm"},
	{0xB45847B9AA97B6BBULL, "6155389_m.fxe"},
	{0xB45872019994D8BBULL, "311189653.ogg"},
	{0xB4587A326F08C530ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsfloor.mao"},
	{0xB4587E326F08CC43ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsfloor.mmh"},
	{0xB4588C326F08E39DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_01ruinsfloor.msh"},
	{0xB458B06461F9C785ULL, "zz_followers_banter6159153.cl"},
	{0xB45937F887CAD8B3ULL, "hlac_lt_0_lowtown_docks_n_-131073_0z.dds"},
	{0xB45A6110916F040EULL, "w_bonepitmining_a2_152_0.mmh"},
	{0xB45A6D10916F18A5ULL, "w_bonepitmining_a2_152_0.mao"},
	{0xB45A7B10916F30FCULL, "w_bonepitmining_a2_152_0.msh"},
	{0xB45AACE80B76BE4CULL, "vxca_cavern1_haunted_ikp_0_0.vlm"},
	{0xB45B37E9C46CD648ULL, "hlac_lt_hangedman_n_-1_1y.dds"},
	{0xB45B95D121C7B31EULL, "6048132_m.fxe"},
	{0xB45BB12C16A8C093ULL, "vxlt_0_lowtown_e29_0_0.vlm"},
	{0xB45BD4D656C399BAULL, "llac_lt_0_lowtown_a3_n_c_0_3.dds"},
	{0xB45BFC633E13D22FULL, "l_dae210ar_lowtown_nt.wbd"},
	{0xB45C2DB3CAF6B8BFULL, "374906453.ogg"},
	{0xB45C3F3348903C43ULL, "vxlt_0_lowtown_docks_n_ba8_0_1.vlm"},
	{0xB45CBE18B88A8ACEULL, "w_sundermount_230_0.msh"},
	{0xB45CCC18B88AA21FULL, "w_sundermount_230_0.mao"},
	{0xB45CD818B88AB6BCULL, "w_sundermount_230_0.mmh"},
	{0xB45CF48F4C9675F8ULL, "873132112.ogg"},
	{0xB45D10EA71621C95ULL, "dae021_brothel_onenter6196720.cl"},
	{0xB45D1D427BC5D968ULL, "gen_im_trs_cot_lgt_hlm_r0.uti"},
	{0xB45D651D00576A3CULL, "hlac_lt_0_lowtown_a2_n_131069_0x.dds"},
	{0xB45E4CB7BA712DC6ULL, "6115354_f.fxe"},
	{0xB45E99B92AA644AAULL, "static_static_p58.psh.fxo.dx11"},
	{0xB45ED8A34931E177ULL, "162672218.ogg"},
	{0xB45F01ABF61AEB04ULL, "55433972.ogg"},
	{0xB45F512E9644FB4AULL, "vxht_keep_70h_0_0.vlm"},
	{0xB45F83325AA30998ULL, "6155562_m.fxe"},
	{0xB45FB2CAA73C83B6ULL, "zz_mer_debug6045073.cl"},
	{0xB46055E376E446A1ULL, "6093874_m.fxe"},
	{0xB4610A923890F8CFULL, "vxht_chantry_n_210_0_0.vlm"},
	{0xB461EB8D279BEB8CULL, "vxht_mansion2_1nb_0_0.vlm"},
	{0xB462628E67C53A35ULL, "gen00flst_varric_act3anders.stg"},
	{0xB462C8B0D9CB7B8FULL, "vxlt_0_lowtown_f66_0_0.vlm"},
	{0xB462D353CE61C860ULL, "dr_0_deeproads_5843.rml"},
	{0xB462E5CB680F23CDULL, "350885846.ogg"},
	{0xB463147A5341E2DCULL, "6141884_m.fxe"},
	{0xB4636C7B77630761ULL, "vxca_caverns2_92u_0_1.vlm"},
	{0xB463A25AB6567B2AULL, "6167121_m.fxe"},
	{0xB4642DF87B8733F5ULL, "vxht_keep_n_61r_0_0.vlm"},
	{0xB46456E6E059CF10ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic.phy"},
	{0xB4646136943BDBB0ULL, "6149211_m.fxe"},
	{0xB464CB4E08DBD28BULL, "hlac_dr_dungeon_3207_4y.dds"},
	{0xB46552A17805BA2EULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgta.phy"},
	{0xB46555BC16467763ULL, "979486589.ogg"},
	{0xB465752A2C74A1ECULL, "6075267_f.fxe"},
	{0xB465842963D8D7D4ULL, "716235347.ogg"},
	{0xB4660F35E82F63F0ULL, "vxht_keep_n_5gv_0_0.vlm"},
	{0xB46694794914FCAEULL, "art\\characters\\weapons\\greataxes\\w_gax_003a.msh"},
	{0xB466A279491514FFULL, "art\\characters\\weapons\\greataxes\\w_gax_003a.mao"},
	{0xB466AE794915289CULL, "art\\characters\\weapons\\greataxes\\w_gax_003a.mmh"},
	{0xB4672997FDD51AB1ULL, "492004668.ogg"},
	{0xB4672AB99AB6AA3EULL, "w_encounter1_11_0.rml"},
	{0xB4673B42D5B6289FULL, "6064045_m.fxe"},
	{0xB4677F113ABDDB80ULL, "spark_ring.dds"},
	{0xB467C8256675588AULL, "6077840_m.fxe"},
	{0xB468945F88ABB2A2ULL, "6078416_m.fxe"},
	{0xB468FC431535A99AULL, "219887771.ogg"},
	{0xB469304CCE9C6B4FULL, "qcr000_arishok6060135.cl"},
	{0xB46988276B0FAFC6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_8.pwk"},
	{0xB46A7DD5D189D3F6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_landscapes_sa.dds"},
	{0xB46B46262A265BD6ULL, "ts_dalish_prog_junk2.gda"},
	{0xB46B9261A18207A3ULL, "111947824.ogg"},
	{0xB46BFBDF787544F0ULL, "vxdr_slavers_1av_0_0.vlm"},
	{0xB46C297DF0DE3134ULL, "vxlt_warehouse_n_11n_0_0.vlm"},
	{0xB46D121911A4A77AULL, "6097590_m.fxe"},
	{0xB46D43B99ABB4002ULL, "w_encounter1_11_0.phy"},
	{0xB46DBA80BEC05B72ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.mmh"},
	{0xB46DBDE81A680ED0ULL, "vxw_encounter2_rv_0_0.vlm"},
	{0xB46DE18326FA9284ULL, "hlac_lt_0_lowtown_a2_n_1_2y.dds"},
	{0xB46E5A6E85C078ADULL, "ht_0_hightown_n_16_0_lc.mmh"},
	{0xB46E7D03B02DB252ULL, "vxca_caverns2prc_9sy_0_0.vlm"},
	{0xB470AD58EDF43E1FULL, "gen_im_trs_dal_shd_wdn_r2.uti"},
	{0xB470B2D296DD0C90ULL, "77659978.ogg"},
	{0xB471147ED4187167ULL, "art\\vfx\\combat\\vfx_gib_hm_meda_ltrb.anb"},
	{0xB471B78485521AFDULL, "752653694.ogg"},
	{0xB471BFC8605C39F5ULL, "mag330_walter6171548.cl"},
	{0xB471DD178E5B4799ULL, "mcr371_scared_mage_f6087707.cl"},
	{0xB472056A2342ADB6ULL, "l_var221ar_mansion_fake.bnk"},
	{0xB472427648BC324AULL, "art\\vfx\\combat\\vfx_glb_stun_c.anb"},
	{0xB47377E9F24D32FAULL, "vxlt_0_lowtown_a2_n_eot_0_0.vlm"},
	{0xB4737856380C2827ULL, "930641150.ogg"},
	{0xB473FC87CD3C57C3ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.impa.evt"},
	{0xB4746FACC9596DF5ULL, "6179777_f.fxe"},
	{0xB47525BE3F789265ULL, "6093827_m.fxe"},
	{0xB475267E0A2DC1FDULL, "vxw_woundedcoast_1k4_0_0.vlm"},
	{0xB4756258D6C1D3F6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoorbird.msh"},
	{0xB4756C58D6C1E4F4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoorbird.mmh"},
	{0xB4757058D6C1EB07ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoorbird.mao"},
	{0xB475741877891D90ULL, "vxht_0_hightown_c1a_0_0.vlm"},
	{0xB47691206512E75FULL, "hlac_lt_0_lowtown_a2_n_131070_0y.dds"},
	{0xB47781922B657624ULL, "6153537_m.fxe"},
	{0xB477CEE0B783D3B6ULL, "556545284.ogg"},
	{0xB477ED3CE1643474ULL, "gen_im_trs_gem_rosediam_04.uti"},
	{0xB4782451A3777CB3ULL, "6110382_m.fxe"},
	{0xB47839910F588995ULL, "vxdr_primevalentrance_1kq_0_0.vlm"},
	{0xB47909C8A43B09E9ULL, "vxdr_dungeon_2hh_0_1.vlm"},
	{0xB479176D3DEA2EC9ULL, "vxw_bonepitdestroyed_a3_1kg_0_0.vlm"},
	{0xB4792F3A8AA32DB2ULL, "6160832_m.fxe"},
	{0xB4793B7D6AEE5F10ULL, "ser100_party_comments6120598.cl"},
	{0xB4795228D4C3814CULL, "lt_0_lowtown_docks_a3_n_100_0.phy"},
	{0xB479D2D0EDF5A259ULL, "750167100.ogg"},
	{0xB479DE496C9B8727ULL, "hlac_lt_0_lowtown_docks_a3_n_262142_0z.dds"},
	{0xB479FF856CE0DDD3ULL, "6144978_m.fxe"},
	{0xB47AD587CD423444ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.impa.gad"},
	{0xB47ADC95BB7840E8ULL, "clf360st_evets.stg"},
	{0xB47AE9B40C00172EULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_app_p.impa.ani"},
	{0xB47AF3E1BB857F1EULL, "llac_ht_0_hightown_-131070_0.dds"},
	{0xB47B51A3A197A7EBULL, "hlac_lt_0_lowtown_a2_n_-65535_1y.dds"},
	{0xB47B51F5B1CB58E3ULL, "w_sundermount_154_0.msh"},
	{0xB47B5FF5B1CB703DULL, "w_sundermount_154_0.mmh"},
	{0xB47B63F5B1CB774EULL, "w_sundermount_154_0.mao"},
	{0xB47B6BDF8DB17291ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdieloop.gad"},
	{0xB47B7BDA48ABD431ULL, "192857647.ogg"},
	{0xB47C54D6FA42DB12ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p.impa.gad"},
	{0xB47DCA2544549DEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmsup.gad"},
	{0xB47E16889BEB8B5BULL, "art\\vfx\\combat\\vfx_r_brd_csong_act_c.anb"},
	{0xB47ED8AE4DE7FF79ULL, "880962073.ogg"},
	{0xB47F1F8A354A507BULL, "hlac_ca_caverns2_5802_36y.dds"},
	{0xB47FB91FCE94BD91ULL, "gen00fl_aveline6110365.cl"},
	{0xB47FCAD3452E1CC4ULL, "356959021.ogg"},
	{0xB48001A771175423ULL, "hlac_ht_mansion_n_3374_0x.dds"},
	{0xB480352831D91115ULL, "vxw_sundermount_31e_0_1.vlm"},
	{0xB480E487CD46B922ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.impa.ani"},
	{0xB4811C8B033D1BB0ULL, "270036730.ogg"},
	{0xB481C09F471F5FA4ULL, "6089267_m.fxe"},
	{0xB481D97BF4068E6EULL, "vxlt_0_lowtown_n_eqj_0_0.vlm"},
	{0xB482276A495479FBULL, "172440571.ogg"},
	{0xB48236740E6D7F5AULL, "6042563_m.fxe"},
	{0xB4825955AECC37BEULL, "vxht_playershouse_n_1u3_0_0.vlm"},
	{0xB4826CDF8DB7933FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderdieloop.ani"},
	{0xB482831298491192ULL, "1655920.ogg"},
	{0xB482F92AD45E13CCULL, "hlac_ht_mansion2b_818_65y.dds"},
	{0xB482FBB5398516E2ULL, "vxca_caverns1_mines_gk4_0_0.vlm"},
	{0xB48357D6FA48FF49ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p.impa.evt"},
	{0xB483B5B67FCFB76DULL, "dae000_bodhan_letter6144251.cl"},
	{0xB483CC3604B5DB35ULL, "6154018_m.fxe"},
	{0xB483D449553DB555ULL, "lowtown_male_beggar_sta6156766.cl"},
	{0xB483E5F5697F3610ULL, "vxht_0_hightown_c4n_0_0.vlm"},
	{0xB4844237BBC585A2ULL, "6173596_f.fxe"},
	{0xB4844EB966F037DCULL, "vxht_keep_6g6_0_0.vlm"},
	{0xB484AD25445A8A81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmsup.evt"},
	{0xB484EF2F88576DA6ULL, "follower_banter6160680.cl"},
	{0xB486494F24DCE259ULL, "712568936.ogg"},
	{0xB486CFE828EDEFADULL, "one100_party_comments6133461.cl"},
	{0xB48758E88E7B1017ULL, "mag101_grace6034084.cl"},
	{0xB4876D8BDD3D8D02ULL, "mcr310_argument6182294.cl"},
	{0xB487894A2FE04749ULL, "vxw_bonepitdestroyed_a3_1lf_0_0.vlm"},
	{0xB487B229C180B5B9ULL, "hlac_ht_bartrand_n_3368_0z.dds"},
	{0xB487C62E5890B768ULL, "6088354_m.fxe"},
	{0xB488136EA1504ED5ULL, "alphafluff.mao"},
	{0xB488308562FB06CCULL, "w_woundedcoast_171_0.msh"},
	{0xB488368562FB10BEULL, "w_woundedcoast_171_0.mmh"},
	{0xB488428562FB2555ULL, "w_woundedcoast_171_0.mao"},
	{0xB4885BB85B6BA5F1ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm7.gad"},
	{0xB488D23E85FDC899ULL, "641307796.ogg"},
	{0xB488E1B40C0C477FULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_app_p.impa.evt"},
	{0xB489AAE942D412A2ULL, "6161246_m.fxe"},
	{0xB48A1FC864F8D833ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_avea_0s.dds"},
	{0xB48AB8C5A835E476ULL, "vxca_caverns1_ij4_0_0.vlm"},
	{0xB48AD21A3FF11ECAULL, "vxlt_0_lowtown_a3_n_c_df1_0_0.vlm"},
	{0xB48B4E5DEA622BE7ULL, "vxht_0_hightown_9i4_0_0.vlm"},
	{0xB48B5D2D0B7AAB0FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_group_heal.gad"},
	{0xB48B832AE70B9BC7ULL, "vxht_mansion2b_235_0_0.vlm"},
	{0xB48BA4C097ABF865ULL, "lgt320_shw_ors_secret_nob.bnk"},
	{0xB48BD5407D890566ULL, "gen00fl_isabela6097608.cl"},
	{0xB48C5DA778DCAEACULL, "820075442.ogg"},
	{0xB48C8D794935C067ULL, "art\\characters\\weapons\\greataxes\\w_gax_003a.phy"},
	{0xB48C8EB23E936D5DULL, "vxht_0_hightown_n_ajp_0_0.vlm"},
	{0xB48CAB12BB0551B7ULL, "vxdr_primevaltomb_7oa_0_0.vlm"},
	{0xB48CE4C956A1BB9BULL, "pro000cr_refugee4.utc"},
	{0xB48CF2078FF26094ULL, "vxlt_darktown_48j_0_0.vlm"},
	{0xB48D61275DAE8C84ULL, "6106816_m.fxe"},
	{0xB48DC0F15DABF583ULL, "568447470.ogg"},
	{0xB48E0747C59AB2AEULL, "6177369_f.fxe"},
	{0xB48E2C0FD12E63DDULL, "hlac_lt_0_lowtown_a2_n_0_1z.dds"},
	{0xB48ED00C7B515587ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comb_0n.dds"},
	{0xB48ED5A3B353562CULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.dura.ani"},
	{0xB48F052A6CDBA932ULL, "236744521.ogg"},
	{0xB48F101852DCF588ULL, "4972345.ogg"},
	{0xB48F5CB85B71C59FULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm7.ani"},
	{0xB48F86C097AFAD9DULL, "lgt320_shw_ors_secret_nob.cnv"},
	{0xB48F86CBBFAEF3F4ULL, "ss_lowclass_thugs_f_1.bnk"},
	{0xB48FC2B40C123168ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_app_p.impa.gad"},
	{0xB4905BA3D63963FAULL, "vxlt_0_lowtown_a2_n_euv_0_0.vlm"},
	{0xB490615C39611C55ULL, "follower_banter6136642.cl"},
	{0xB4913FD6FA5513B8ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p.impa.ani"},
	{0xB4915890CC38D1D9ULL, "vxht_bartrand_n_67o_0_0.vlm"},
	{0xB491722D0B7F3ABCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_group_heal.evt"},
	{0xB49177B20650D921ULL, "vxlt_0_lowtown_a3_n_c_e1r_0_0.vlm"},
	{0xB491CF36F8A56BEBULL, "575474984.ogg"},
	{0xB4920A37EE19F2FDULL, "vxw_woundedcoast_1q2_0_0.vlm"},
	{0xB4928BC1C089BA13ULL, "vxlt_0_lowtown_hau_0_0.vlm"},
	{0xB492A5254466BAD0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmsup.ani"},
	{0xB492AE37D7214B73ULL, "691488191.ogg"},
	{0xB492BFD0E130A61EULL, "vxw_bonepit_zp_0_0.vlm"},
	{0xB4931EB99ADBD0B6ULL, "w_encounter1_11_0.mao"},
	{0xB4932AB99ADBE555ULL, "w_encounter1_11_0.mmh"},
	{0xB4932CB99ADBE8CBULL, "w_encounter1_11_0.msh"},
	{0xB493522082EFCBECULL, "vxht_bartrand_n_6qp_0_0.vlm"},
	{0xB4935445AFE40A6BULL, "vxht_mansion_n_6re_0_0.vlm"},
	{0xB49380B50818E4CFULL, "hlac_w_sundermount_-65541_0z.dds"},
	{0xB493C8C6FC431569ULL, "one300_zevran6075856.cl"},
	{0xB493CF974FF3207FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb02.mmh"},
	{0xB493D5974FF32A21ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb02.msh"},
	{0xB493EE7FEABF056AULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dslb_0d.dds"},
	{0xB4941E76B9134313ULL, "hlac_ht_0_hightown_a3_-131069_0x.dds"},
	{0xB49428122937CFF4ULL, "zz_one_debug6069460.cl"},
	{0xB494B59708CC068BULL, "692363868.ogg"},
	{0xB494CC065087681EULL, "hlac_lt_0_lowtown_a2_-65536_8y.dds"},
	{0xB494E0613A7FCAC3ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_20ring_uhm.mao"},
	{0xB495748F0B3D2121ULL, "llac_lt_0_lowtown_a2_65536_3.dds"},
	{0xB4957D9B7E4167AFULL, "one300_nuncio6076785.cl"},
	{0xB49590D3800A22F1ULL, "gen_im_cft_reg_c2.uti"},
	{0xB495B980BEE1AB19ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.anb"},
	{0xB495D2335D9C1BDFULL, "6084533_m.fxe"},
	{0xB495E2A3B3598B6EULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.dura.gad"},
	{0xB495ECB62F6D51DAULL, "dae000_merchant_mage_hf6164457.cl"},
	{0xB495FFFD14446E8FULL, "278131577.ogg"},
	{0xB496207C18ECD2CCULL, "vxlt_0_lowtown_a2_by4_0_0.vlm"},
	{0xB496380B4A86CFE8ULL, "6083254_m.fxe"},
	{0xB4966A5BAA5165F6ULL, "vxht_mansion2_31w_0_0.vlm"},
	{0xB4966AA42A442C03ULL, "mag210_arianni6057329.cl"},
	{0xB496CAAB5EA76BFAULL, "llac_ht_playershouse_n_855_29.dds"},
	{0xB4975E954CF5F34CULL, "vxlt_gallowscourtyard_7cf_0_0.vlm"},
	{0xB498C4B9AE467919ULL, "hlac_lt_foundry_n_320_8x.dds"},
	{0xB498C61857D6072CULL, "vxlt_foundry_n_yg_0_0.vlm"},
	{0xB498FA797E5FF78BULL, "321011926.ogg"},
	{0xB499088286ECD554ULL, "llac_lt_0_lowtown_docks_n_0_1.dds"},
	{0xB4992AA29A2BFCF6ULL, "vxca_caverns1_im5_0_0.vlm"},
	{0xB49930AF1C828800ULL, "6165796_m.fxe"},
	{0xB499A0819572A9B6ULL, "483594357.ogg"},
	{0xB499EBFF38BCF755ULL, "317245861.ogg"},
	{0xB499FF8AC4B9EAF1ULL, "vxlt_gallowstemplar_n_6dj_0_0.vlm"},
	{0xB49A2B526EFAD176ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_exit.ani"},
	{0xB49A49CC6E83900CULL, "6096497_f.fxe"},
	{0xB49A94EE7277A3FAULL, "vxca_caverns1_mines_i9l_0_0.vlm"},
	{0xB49ABB6C9701F3C3ULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyo.msh"},
	{0xB49AC96C97020B1DULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyo.mmh"},
	{0xB49ACD6C9702122EULL, "art\\characters\\playercharacter\\elfmale\\boo_hvyo.mao"},
	{0xB49AEC686BDBF435ULL, "follower_banter6168779.cl"},
	{0xB49B457648DF3C3DULL, "art\\vfx\\combat\\vfx_glb_stun_c.mmh"},
	{0xB49B600C25FA23BAULL, "18558184.ogg"},
	{0xB49BB2EE05A4D507ULL, "6053577_m.fxe"},
	{0xB49BC9E3E4642515ULL, "mrl000_marethari6154503.cl"},
	{0xB49C1C1F03C672F4ULL, "6114873_m.fxe"},
	{0xB49CB09FF67C36DCULL, "darktown_fem_elf_sta6174527.cl"},
	{0xB49CDDA3B35FA14DULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.dura.evt"},
	{0xB49CEA2F84A35121ULL, "w_bonepitmining_a2_170_0.phy"},
	{0xB49D16633E4B9505ULL, "l_dae210ar_lowtown_nt.bnk"},
	{0xB49D1910DCFDF646ULL, "vxdr_0_deeproads_f0v_0_0.vlm"},
	{0xB49D8FA9514156BFULL, "hlac_ht_0_hightown_a3_-131073_1x.dds"},
	{0xB49D9531E3B912CFULL, "hla_ht_0_hightown_a2_65536_0z.dds"},
	{0xB49DB3BD6C1B78F6ULL, "lla_ht_0_hightown_n_-131074_0.dds"},
	{0xB49E1FD291E6C218ULL, "llac_ht_chantry_2534_14.dds"},
	{0xB49E4E0C98B301E7ULL, "6081206_m.fxe"},
	{0xB49E5CC8FDC17F86ULL, "282517478.ogg"},
	{0xB49E73CA591DEC73ULL, "vxca_cavern1_haunted_ckm_0_0.vlm"},
	{0xB49E8C752053B500ULL, "vxca_caverns1_i6v_0_0.vlm"},
	{0xB49E90642DE2C885ULL, "mcr310_sibling6076405.cl"},
	{0xB49EA8B96A7947F0ULL, "gen00fl_anders6112987.cl"},
	{0xB49ECAE73E163655ULL, "hlac_lt_0_lowtown_docks_n_-131075_5y.dds"},
	{0xB49EF41957C2D77EULL, "6124750_f.fxe"},
	{0xB49EFEAC3B3E662BULL, "594584024.ogg"},
	{0xB49F11F8D0138E1AULL, "face_rendertodepth_p0.vsh.fxo.dx11"},
	{0xB49F529F00A9A441ULL, "vxdr_0_deeproads_fhu_0_0.vlm"},
	{0xB49F5CD6979633F1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw2sm.mmh"},
	{0xB49F60E3ABDA5D02ULL, "vxlt_0_lowtown_n_gbu_0_0.vlm"},
	{0xB49F68D697964792ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw2sm.mao"},
	{0xB49F76D697965FAFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw2sm.msh"},
	{0xB49F7A2D0B8B85D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_group_heal.ani"},
	{0xB49F9E47B481DA57ULL, "vxdr_primevalentrance_2st_0_0.vlm"},
	{0xB49FED5DD7BB6429ULL, "vxca_caverns2_8b9_0_0.vlm"},
	{0xB4A04D2440950921ULL, "6070166_f.fxe"},
	{0xB4A08DC105D1B4ABULL, "art\\characters\\playercharacter\\humanfemale\\glv_cas.mao"},
	{0xB4A0B4BBF97FD546ULL, "6204163_m.fxe"},
	{0xB4A132734ADC68DCULL, "vxht_keep_6m4_0_0.vlm"},
	{0xB4A13DCF1763A784ULL, "334409188.ogg"},
	{0xB4A1509D7DD56B76ULL, "306803147.ogg"},
	{0xB4A189F1401553F5ULL, "229532697.ogg"},
	{0xB4A2D19008C983F3ULL, "atl_startmenu_dxt5_05.dds"},
	{0xB4A2EBF75E064B9DULL, "lowtown_male_beggar_sta6151028.cl"},
	{0xB4A2FB1A54F28669ULL, "515091646.ogg"},
	{0xB4A3498B9464893AULL, "84969477.ogg"},
	{0xB4A34AE05217E764ULL, "207484108.ogg"},
	{0xB4A35C52EE463B8CULL, "127749297.ogg"},
	{0xB4A38A79439380C7ULL, "6133143_m.fxe"},
	{0xB4A38AD8144FA97FULL, "827993467.ogg"},
	{0xB4A39C33B8DC12DBULL, "vxht_0_hightown_a3_cfb_0_0.vlm"},
	{0xB4A3BCF847B2F015ULL, "one300dg_solivitus6060363.cl"},
	{0xB4A3D4803FB8EA30ULL, "follower_banter6160783.cl"},
	{0xB4A466C86D0E22AEULL, "hla_ht_0_hightown_n_-65536_0y.dds"},
	{0xB4A482BF2CBC8CD4ULL, "vxht_brothel_n_5l0_0_0.vlm"},
	{0xB4A48354DE66134DULL, "48708011.ogg"},
	{0xB4A49E4FE2EEFDE9ULL, "vxht_0_hightown_c8o_0_0.vlm"},
	{0xB4A4CDDE366E06E8ULL, "llac_lt_darktown_-1_59.dds"},
	{0xB4A5397838FCB4A0ULL, "w_blightlands_103_0.phy"},
	{0xB4A551A1783B9CF1ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgta.mmh"},
	{0xB4A5552053AA4F03ULL, "vxdr_dungeon_1lt_0_0.vlm"},
	{0xB4A55DA1783BB092ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgta.mao"},
	{0xB4A56BA1783BC8AFULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgta.msh"},
	{0xB4A5888E6916C514ULL, "gen00fl_merrill6128643.cl"},
	{0xB4A6640B58DBB614ULL, "llac_ht_chantry_2534_3.dds"},
	{0xB4A67E077D087571ULL, "vxlt_shop_fs_0_0.vlm"},
	{0xB4A67E9377F0DCBEULL, "121585761.ogg"},
	{0xB4A67F9235D9FC17ULL, "static_static_p0.psh.fxo"},
	{0xB4A690F86EF54EDEULL, "dae000_bartrand6031124.cl"},
	{0xB4A6B8C90E6A3901ULL, "vxht_keep_6gg_0_0.vlm"},
	{0xB4A6BA1BD3D7E958ULL, "6139241_m.fxe"},
	{0xB4A733526F056967ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_exit.evt"},
	{0xB4A7999D07FF5B7DULL, "hlac_w_blightlands_196608_0z.dds"},
	{0xB4A82AA3B6D7AF8EULL, "561372844.ogg"},
	{0xB4A8447778061452ULL, "vxca_caverns2prc_9oe_0_0.vlm"},
	{0xB4A8A66409082315ULL, "159654870.ogg"},
	{0xB4A8C1DFF9FB5F1DULL, "1045450848.ogg"},
	{0xB4A8D832E9BB3CCAULL, "237570518.ogg"},
	{0xB4A98397885BE5FEULL, "6131940_m.fxe"},
	{0xB4AA2032CFDB7D95ULL, "gen_im_wep_war_shd_act1_04.uti"},
	{0xB4AA5F5DF4794756ULL, "zz_car_debug6122480.cl"},
	{0xB4AB00F7781CF8D0ULL, "vxdr_primevalentrance_39m_0_1.vlm"},
	{0xB4AB22F681A93C42ULL, "vxht_0_hightown_c7q_0_0.vlm"},
	{0xB4AB594EC4EF3C1AULL, "1068005271.ogg"},
	{0xB4ABB6A331E5A975ULL, "6181751_m.fxe"},
	{0xB4ABE21048822576ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_ltrb.anb"},
	{0xB4AE0EB38BF6B2D2ULL, "hlac_lt_0_lowtown_docks_a3_131070_0y.dds"},
	{0xB4AE218D306A6921ULL, "llac_lt_warehouse_0_8.dds"},
	{0xB4AE2CF56CA0179DULL, "vxlt_0_lowtown_docks_a3_9r9_0_0.vlm"},
	{0xB4AE54526F0BBFB0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_exit.gad"},
	{0xB4AE770FB910D175ULL, "6202293_f.fxe"},
	{0xB4AEC067D8631F31ULL, "vxlt_undercityencounter_zi_0_0.vlm"},
	{0xB4AEE541C84438E3ULL, "507217704.ogg"},
	{0xB4AF6C274FA42C4EULL, "6054397_m.fxe"},
	{0xB4B03544EA1949F5ULL, "183110481.ogg"},
	{0xB4B047AC8B4C6BC6ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_4y.dds"},
	{0xB4B0D99C86BADF4FULL, "vxlt_hangedman_n_1tb_0_0.vlm"},
	{0xB4B0F11F65A53FF6ULL, "vxlt_warehouse_qf_0_0.vlm"},
	{0xB4B143591ACA625CULL, "6170479_m.fxe"},
	{0xB4B184E40B7D34BDULL, "vxlt_hangedman_n_t3_0_0.vlm"},
	{0xB4B210D3DCEB7519ULL, "lt_0_lowtown_docks_a3_n_250_0.msh"},
	{0xB4B21AD3DCEB8607ULL, "lt_0_lowtown_docks_a3_n_250_0.mmh"},
	{0xB4B21ED3DCEB8D54ULL, "lt_0_lowtown_docks_a3_n_250_0.mao"},
	{0xB4B34AE1C0B69BF6ULL, "73008763.ogg"},
	{0xB4B36E66EABB2825ULL, "zz_vault_debug6203322.cl"},
	{0xB4B3A5B239EB9533ULL, "vxlt_0_lowtown_f0h_0_0.vlm"},
	{0xB4B40A2F78A97205ULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.anb"},
	{0xB4B471709E30BA82ULL, "893979548.ogg"},
	{0xB4B47A02D2A28DC9ULL, "vxlt_0_lowtown_docks_b7v_0_0.vlm"},
	{0xB4B4FFB09F18A3A0ULL, "gen_im_trs_crc_junk3_02.uti"},
	{0xB4B5200F846D3A9CULL, "lowtown_male_alienage1_sta6162734.cl"},
	{0xB4B52013777A6F16ULL, "vxlt_0_lowtown_docks_a3_n_d51_0_0.vlm"},
	{0xB4B534DFA87DEC39ULL, "follower_banter6172703.cl"},
	{0xB4B5A6AD1775DA59ULL, "628452998.ogg"},
	{0xB4B6558CBAE2CA7DULL, "gen00fl_anders6114490.cl"},
	{0xB4B6B65469FC2CD2ULL, "vxca_caverns2prc_9jf_0_0.vlm"},
	{0xB4B6C8036670F1D8ULL, "mer341_jansen6054878.cl"},
	{0xB4B70DADD8166772ULL, "vxdr_dungeon_1t8_0_1.vlm"},
	{0xB4B7279D99F918A2ULL, "swashbklr_flicker.dds"},
	{0xB4B78D9818C1FCB6ULL, "w_bonepitdestroyed_a3_21_0.rml"},
	{0xB4B79E14AC28211FULL, "565263923.ogg"},
	{0xB4B8B85493F5D916ULL, "400195739.ogg"},
	{0xB4B8CE69B6984010ULL, "6126682_f.fxe"},
	{0xB4B8F6168BD06FDCULL, "497311818.ogg"},
	{0xB4B92928D4F91F17ULL, "lt_0_lowtown_docks_a3_n_100_0.mmh"},
	{0xB4B92D28D4F92624ULL, "lt_0_lowtown_docks_a3_n_100_0.mao"},
	{0xB4B93F28D4F94489ULL, "lt_0_lowtown_docks_a3_n_100_0.msh"},
	{0xB4B9BFC55C815403ULL, "6016776_m.fxe"},
	{0xB4B9DAF79750282FULL, "vxht_mansion_n_54l_0_0.vlm"},
	{0xB4BB51C7118D4C68ULL, "m_mer211ar_house.bnk"},
	{0xB4BB7A2D88B1F38EULL, "857227858.ogg"},
	{0xB4BBCD4E66B75FDDULL, "hlac_lt_foundry_n_320_2z.dds"},
	{0xB4BC364E9A537F67ULL, "lowtown_guard_chase6169512.cl"},
	{0xB4BC5B434FC191E9ULL, "llac_lt_0_lowtown_n_-262144_3.dds"},
	{0xB4BC6F6EE2D6CFFEULL, "6060800_m.fxe"},
	{0xB4BCEF3094307D19ULL, "527720378.ogg"},
	{0xB4BD869F0D61984AULL, "6200022_m.fxe"},
	{0xB4BDC3DF7C6EE528ULL, "hlac_lt_0_lowtown_a2_n_c_-3_0z.dds"},
	{0xB4BDD826197A19E3ULL, "dae000ip_crowd_mf4_talk16.utp"},
	{0xB4BE0AA017E32CE5ULL, "758274219.ogg"},
	{0xB4BE1567B712EF9EULL, "vxw_bonepit_ks_0_0.vlm"},
	{0xB4BF2B6B817655C3ULL, "195742818.ogg"},
	{0xB4BF2EAB74F175C9ULL, "vxdr_primevaltomb_8vg_0_0.vlm"},
	{0xB4BF4B7978B0F7CFULL, "vxlt_hangedman_n_1qm_0_0.vlm"},
	{0xB4BF62FC579B5876ULL, "vxlt_warehouse_zg_0_0.vlm"},
	{0xB4BFF6C0FD734D3DULL, "vxlt_hangedman_n_s2_0_0.vlm"},
	{0xB4C07CCBB3CD2507ULL, "516114672.ogg"},
	{0xB4C07E2A7F5FDB74ULL, "6060118_m.fxe"},
	{0xB4C0982B1A9A71E5ULL, "6150186_m.fxe"},
	{0xB4C0D3C4474C9449ULL, "117039008.ogg"},
	{0xB4C0F0D357DD16FBULL, "vxht_0_hightown_a3_5eq_0_0.vlm"},
	{0xB4C14240F3FB8E2FULL, "6162084_m.fxe"},
	{0xB4C163887251D684ULL, "6144906_m.fxe"},
	{0xB4C16E113E7FE925ULL, "vxht_keep_6fx_0_0.vlm"},
	{0xB4C2178F2BE1ADB3ULL, "vxlt_0_lowtown_f5k_0_0.vlm"},
	{0xB4C226B475E82B31ULL, "vxlt_gallowstemplar_3qs_0_0.vlm"},
	{0xB4C29F2233CC38C1ULL, "gen00fl_fenris6126729.cl"},
	{0xB4C2DD1D91EF495EULL, "101734321.ogg"},
	{0xB4C325BD30A12EA7ULL, "hlac_lt_0_lowtown_n_65534_6x.dds"},
	{0xB4C33428C539368BULL, "hlac_dr_slavers_3214_9x.dds"},
	{0xB4C38B18EE13FAFFULL, "6042571_m.fxe"},
	{0xB4C391F069708796ULL, "vxlt_0_lowtown_docks_a3_n_d02_0_0.vlm"},
	{0xB4C417B36AB6B6B3ULL, "gen00fl_isabela6090431.cl"},
	{0xB4C45EF0581B7DE3ULL, "794055012.ogg"},
	{0xB4C4A69E62666381ULL, "llac_lt_0_lowtown_a2_-262144_2.dds"},
	{0xB4C4E1CD76A0594BULL, "6042221_m.fxe"},
	{0xB4C520B576536FABULL, "6175480_m.fxe"},
	{0xB4C5218730D0BC64ULL, "187439550.ogg"},
	{0xB4C56414567F3C60ULL, "6060647_f.fxe"},
	{0xB4C58C9276EC3345ULL, "vxca_caverns1_mines_i9g_0_0.vlm"},
	{0xB4C692264D6852B2ULL, "6074078_m.fxe"},
	{0xB4C6A33542FF7A7DULL, "115640893.ogg"},
	{0xB4C6AF7E7B622825ULL, "vxht_0_hightown_n_bun_0_0.vlm"},
	{0xB4C6D02F23F61CB2ULL, "vxht_mansion2b_344_0_0.vlm"},
	{0xB4C78D3601CA32A5ULL, "vxlt_gallowstemplar_n_45o_0_0.vlm"},
	{0xB4C7AEB04861F3E7ULL, "follower_banter6160971.cl"},
	{0xB4C8047375EF3C11ULL, "vxht_0_hightown_a2_c83_0_0.vlm"},
	{0xB4C829152426A08CULL, "mag311_keran6062053.cl"},
	{0xB4C84D3F38CF8E3FULL, "hlac_lt_undercitydungeon_-1_8z.dds"},
	{0xB4C895A3C3E35E26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_crack_knuckles.evt"},
	{0xB4C8A5F1707F9598ULL, "vxdr_0_deeproads_evq_0_0.vlm"},
	{0xB4C8E6BFD97E7A20ULL, "vxdr_0_deeproads_9e4_0_0.vlm"},
	{0xB4C90D8563326259ULL, "w_woundedcoast_171_0.phy"},
	{0xB4C9610D3C4F48B9ULL, "vxw_bonepit_c0_0_0.vlm"},
	{0xB4C99C49F390F634ULL, "vxlt_0_lowtown_a3_n_c_eo9_0_0.vlm"},
	{0xB4CA5678391C0235ULL, "w_blightlands_103_0.msh"},
	{0xB4CA5878391C05ABULL, "w_blightlands_103_0.mmh"},
	{0xB4CA6478391C1A48ULL, "w_blightlands_103_0.mao"},
	{0xB4CA843D835A7E63ULL, "885052221.ogg"},
	{0xB4CA8A0FAB8A604AULL, "136383387.ogg"},
	{0xB4CAD38CF1F967D7ULL, "hlac_lt_0_lowtown_a2_n_65536_11x.dds"},
	{0xB4CB125E6EFE8CC0ULL, "vxlt_warehouse_n_156_0_0.vlm"},
	{0xB4CB9B15C6AACE16ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsb.phy"},
	{0xB4CBA84E51F1839AULL, "mcr370_meredith6085899.cl"},
	{0xB4CCF4E804128A81ULL, "pro000_opening_pc6142521.cl"},
	{0xB4CD0BF258CC1A8CULL, "dae000_gamlen6026199.cl"},
	{0xB4CDD188ED8797D7ULL, "vxht_0_hightown_n_aqs_0_0.vlm"},
	{0xB4CEA2A57AFFDA6DULL, "6182488_m.fxe"},
	{0xB4CEACA3C3E7F171ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_crack_knuckles.gad"},
	{0xB4CF5B86B51F49F9ULL, "361600407.ogg"},
	{0xB4CFA0CBBFE503A6ULL, "ss_lowclass_thugs_f_1.wbd"},
	{0xB4D01FF0676DFB9DULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medb.phy"},
	{0xB4D06B58D70E8EEFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_visdoorbird.phy"},
	{0xB4D0879C34DCDF70ULL, "vxw_bonepit_25g_0_0.vlm"},
	{0xB4D1250081A0C2BCULL, "6000692_f.fxe"},
	{0xB4D216CA05C00AB2ULL, "264292339.ogg"},
	{0xB4D2BD5AE4818FDCULL, "lt_0_lowtown_n_37_0_lc.mmh"},
	{0xB4D3079D22FFEEEAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgta_0s.dds"},
	{0xB4D31EEA7BED74A1ULL, "vxlt_0_lowtown_a2_bxz_0_0.vlm"},
	{0xB4D3BF3F27BEF0BDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9rr_windowbarsxa.mao"},
	{0xB4D3C33F27BEF7F6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9rr_windowbarsxa.mmh"},
	{0xB4D3CD3F27BF08F4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9rr_windowbarsxa.msh"},
	{0xB4D414BE43F9533DULL, "6121167_f.fxe"},
	{0xB4D4E62D4CF43D78ULL, "mag121_eau_banter6121091.cl"},
	{0xB4D58415C6B2FBB7ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsb.msh"},
	{0xB4D59615C6B31A1AULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsb.mao"},
	{0xB4D59A15C6B32129ULL, "art\\characters\\weapons\\greatswords\\w_gsw_dsgsb.mmh"},
	{0xB4D5ADA3C3EE111FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_crack_knuckles.ani"},
	{0xB4D5FEE16209D96CULL, "lgt120_orz_bln_official6176128.cl"},
	{0xB4D61816F755889FULL, "6168590_m.fxe"},
	{0xB4D65136FD398680ULL, "gen_im_trs_prf_dwarf_shield.uti"},
	{0xB4D6A041265F75D3ULL, "6170667_m.fxe"},
	{0xB4D6CC300B59C40DULL, "83367402.ogg"},
	{0xB4D7067269B7F861ULL, "mag160_varian_ilithis6171461.cl"},
	{0xB4D70DAB4CCD7BF6ULL, "llac_dr_primevaltomb_9621_0.dds"},
	{0xB4D77BFD0BC5DEE9ULL, "vxca_caverns2_6se_0_1.vlm"},
	{0xB4D7D2EA2E456139ULL, "vxw_bonepit_n7_0_0.vlm"},
	{0xB4D8044D30B59321ULL, "qcr220cr_noblewoman2.utc"},
	{0xB4D90072D8838357ULL, "130982222.ogg"},
	{0xB4D912F1C7FED9B9ULL, "vxca_cavern1_haunted_iic_0_0.vlm"},
	{0xB4D936D3159A9A32ULL, "6168128_m.fxe"},
	{0xB4D9696918360F0AULL, "vxlt_0_lowtown_n_erw_0_0.vlm"},
	{0xB4D9F74CDB971EACULL, "hlac_w_bonepitmining_a2_0_8x.dds"},
	{0xB4D9F869175BBBD7ULL, "mrl310_merrill6144353.cl"},
	{0xB4DA0C4FDAF58DD7ULL, "87673389.ogg"},
	{0xB4DB26E613A4B104ULL, "700734333.ogg"},
	{0xB4DB37B0B53FB48DULL, "seb321_leliana6197865.cl"},
	{0xB4DB4463223FB9CCULL, "vxw_bonepitdestroyed_a3_1pc_0_0.vlm"},
	{0xB4DB5379EFBDF17AULL, "6085504_m.fxe"},
	{0xB4DB8922818F8D04ULL, "vxlt_undercityserialkiller_xn_0_0.vlm"},
	{0xB4DB8D9669B284EBULL, "vxht_mansion_n_6ix_0_0.vlm"},
	{0xB4DC0B2F78CAC646ULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.mmh"},
	{0xB4DC5179215CFFF4ULL, "166659359.ogg"},
	{0xB4DC6147816B9C5CULL, "6104426_f.fxe"},
	{0xB4DC83BF8E93842AULL, "454727278.ogg"},
	{0xB4DC990B53389F09ULL, "6111227_f.fxe"},
	{0xB4DD8EFAD9C2F926ULL, "mer341_jansen6054865.cl"},
	{0xB4DDC5DF7AD4D41AULL, "6062789_m.fxe"},
	{0xB4DDE2F8A18D6673ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction1.evt"},
	{0xB4DEC026B0974C42ULL, "hlac_lt_hangedman_n_-1_8x.dds"},
	{0xB4DECB4ACCB6D149ULL, "120674573.ogg"},
	{0xB4DEF862A659A67BULL, "vxht_chantry_2t9_0_1.vlm"},
	{0xB4DEF97926D2F7F0ULL, "vxw_bonepit_20d_0_0.vlm"},
	{0xB4DF1057DE3C6CB2ULL, "28450186.ogg"},
	{0xB4DFBCA90A37DB51ULL, "ave110_ambush_26117514.cl"},
	{0xB4E0412DD3BE640EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03b.phy"},
	{0xB4E043C711AC5012ULL, "m_mer211ar_house.wbd"},
	{0xB4E0BCBCAC42CA3BULL, "vxht_0_hightown_a2_c3u_0_1.vlm"},
	{0xB4E120F0677C70A2ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medb.mmh"},
	{0xB4E122F0677C7458ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medb.msh"},
	{0xB4E1CE9960A0A70DULL, "t3_hey_hz3.tnt"},
	{0xB4E1D050F5491FCDULL, "zz_seb_debug6187880.cl"},
	{0xB4E1F5CA6B78FC01ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur5_flemeth_real.ani"},
	{0xB4E2704C5AD86445ULL, "vxca_caverns1_mines_i3e_0_0.vlm"},
	{0xB4E2A71FFBD07686ULL, "atl_chargen_st_dxt5_dat.xml"},
	{0xB4E2E89BD534EEAEULL, "projdecal_projdecal_p1.psh.fxo.dx11.assembly"},
	{0xB4E3193B433F5B79ULL, "mcr370_varric6098318.cl"},
	{0xB4E339679B30D136ULL, "hlac_lt_0_lowtown_-262144_3z.dds"},
	{0xB4E347D8AFE9AAE7ULL, "vxca_cavern1_haunted_fmf_0_0.vlm"},
	{0xB4E377DDFAF5199AULL, "zz_vrc_debug6200187.cl"},
	{0xB4E39B5C7EEA3D4FULL, "dae000_merchant_tailor_hm6151377.cl"},
	{0xB4E411D7F12F9A76ULL, "3266274.ogg"},
	{0xB4E4A42C628185AEULL, "6198681_f.fxe"},
	{0xB4E4DBF8A19378B4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction1.gad"},
	{0xB4E58B0CC4A7656EULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1eggsac.msh"},
	{0xB4E5990CC4A77CBFULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1eggsac.mao"},
	{0xB4E5A50CC4A7915CULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1eggsac.mmh"},
	{0xB4E644C7203B79B9ULL, "vxw_bonepit_i6_0_0.vlm"},
	{0xB4E6EB0821E8B8D0ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgtp_0n.dds"},
	{0xB4E71E20B8336B8CULL, "1012449106.ogg"},
	{0xB4E7BAF21A408C65ULL, "6080276_m.fxe"},
	{0xB4E7DB460A2C278AULL, "vxlt_0_lowtown_n_eop_0_0.vlm"},
	{0xB4E7FF769265FB21ULL, "320048295.ogg"},
	{0xB4E800CA6B7D7AA3ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur5_flemeth_real.gad"},
	{0xB4E83DFEA12C37F7ULL, "gen00fl_fenris6126604.cl"},
	{0xB4E8C40A3FD20A6AULL, "one280_emrys6104347.cl"},
	{0xB4E8E52129AE4365ULL, "6029462_m.fxe"},
	{0xB4E9052BA990A9D2ULL, "66434142.ogg"},
	{0xB4E96020A4A576C1ULL, "468310937.ogg"},
	{0xB4E9718E534EB23EULL, "lgt220_shp_mag_guard6149887.cl"},
	{0xB4E97AA962D5EE21ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmexit.ani"},
	{0xB4EAC66AB583DC00ULL, "6168148_m.fxe"},
	{0xB4EAC83A3563E656ULL, "mag211_marethari6025341.cl"},
	{0xB4EAEAF8A197FD92ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingaction1.ani"},
	{0xB4EBB2AAE1B5317CULL, "llac_lt_warehouse_-65535_0.dds"},
	{0xB4EBC138C20CCDEDULL, "6112891_f.fxe"},
	{0xB4EC41C6B8F834AFULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax103a_0t.dds"},
	{0xB4EC9B1B117CC79AULL, "6166908_m.fxe"},
	{0xB4ED02621F34D292ULL, "vxlt_hangedman_n_27m_0_0.vlm"},
	{0xB4ED127E4AB0DEE0ULL, "vrc221_hugin6042561.cl"},
	{0xB4ED263B0B192690ULL, "6111081_m.fxe"},
	{0xB4ED6160E06B60B9ULL, "914342714.ogg"},
	{0xB4ED8F25C97CD1DCULL, "6016836_m.fxe"},
	{0xB4EDD65A6B35849CULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_wep_ggg_0s.dds"},
	{0xB4EDEFD2BFEB043FULL, "746085162.ogg"},
	{0xB4EEC982880F6A60ULL, "6108083_m.fxe"},
	{0xB4EED078D24546C0ULL, "dae011_badpoet_sta6145203.cl"},
	{0xB4EEF9705ED46D19ULL, "6133853_m.fxe"},
	{0xB4EF1021ED6A0D5FULL, "hlac_w_encounter2_131075_0x.dds"},
	{0xB4EF333B51F5BA03ULL, "mag170_hugh6080795.cl"},
	{0xB4EF85B516E845BDULL, "519844631.ogg"},
	{0xB4EFA67929592D64ULL, "gen00fl_anders6096496.cl"},
	{0xB4F02D6327593198ULL, "vxlt_warehouse_120_0_0.vlm"},
	{0xB4F04BBE81C729E5ULL, "lt_gamlenshouse_n.arl"},
	{0xB4F071419EA468A4ULL, "vxlt_0_lowtown_docks_a3_n_d0p_0_0.vlm"},
	{0xB4F085A962DC1FC3ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmexit.gad"},
	{0xB4F13D39BDBC996FULL, "mag110_arianni6016907.cl"},
	{0xB4F17663DB09288BULL, "zz_isa_debug6133978.cl"},
	{0xB4F1F8A45FE3BB14ULL, "l_dae351ar_mount_caverns.bnk"},
	{0xB4F21946E14BA26AULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rock_002_n.dds"},
	{0xB4F30E9E8CC5703EULL, "zz_one_debug6051520.cl"},
	{0xB4F3AD6E92781A04ULL, "ts_kirk_lowtown_02.gda"},
	{0xB4F4F2C23504670FULL, "83586704.ogg"},
	{0xB4F4F716F662F24CULL, "215869588.ogg"},
	{0xB4F5516A62AD00F0ULL, "gen00fl_merrill6115997.cl"},
	{0xB4F59D0E05C68B04ULL, "38403705.ogg"},
	{0xB4F5ACD33240F225ULL, "qcr220_meredith6153089.cl"},
	{0xB4F5D4472312A8B1ULL, "ht_0_hightown_n_34_0_lc.mmh"},
	{0xB4F64D22FC22400AULL, "vxlt_0_lowtown_n_ehq_0_0.vlm"},
	{0xB4F6D8975047ACC4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hb02.phy"},
	{0xB4F74E782FBE76ABULL, "6015926_m.fxe"},
	{0xB4F762A962E202A8ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmexit.evt"},
	{0xB4F79A5C97B1CF03ULL, "art\\characters\\playercharacter\\humanmale\\boo_meda.mao"},
	{0xB4F79E5C97B1D650ULL, "art\\characters\\playercharacter\\humanmale\\boo_meda.mmh"},
	{0xB4F7AC5C97B1EE6AULL, "art\\characters\\playercharacter\\humanmale\\boo_meda.msh"},
	{0xB4F81295B27D5BF5ULL, "ser121_jethann6104447.cl"},
	{0xB4F85893FBB2A606ULL, "203430750.ogg"},
	{0xB4F8BC25EFB48553ULL, "cotbedroll_sa.dds"},
	{0xB4F8D09762896968ULL, "gen_im_trs_gem_tigerseye_01.uti"},
	{0xB4F9657A2692F31BULL, "and300pt_justice.ncs"},
	{0xB4F96A87EF90BE6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_haymaker_action.ani"},
	{0xB4F991667109CBCCULL, "vxca_cavern1_haunted_i2o_0_0.vlm"},
	{0xB4FABAF8663A46B4ULL, "6037662_m.fxe"},
	{0xB4FAD55DD5F00FC0ULL, "vxdr_primevaltomb_56b_0_1.vlm"},
	{0xB4FB519926FECDB6ULL, "6167279_m.fxe"},
	{0xB4FB6E110758E8BCULL, "61558490.ogg"},
	{0xB4FB735DFB64C229ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_noble2f_arishokaddress.gad"},
	{0xB4FB743F112AEB12ULL, "vxlt_hangedman_n_22n_0_0.vlm"},
	{0xB4FB7F1A01347A6AULL, "389805335.ogg"},
	{0xB4FBA7D96D1B7198ULL, "approval_data_seb.gda"},
	{0xB4FBBAEDB89AFEC6ULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa_lbrt.anb"},
	{0xB4FC37BFF3745562ULL, "vxw_bonepitmining_a2_ek_0_0.vlm"},
	{0xB4FC85DB80F26740ULL, "gen00fl_merrill6122366.cl"},
	{0xB4FCC562475D7A01ULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_aoe_p.anb"},
	{0xB4FCCFB4F08CDED8ULL, "one220_yevhen6102809.cl"},
	{0xB4FD35140F39D54CULL, "w_bonepit_120_0.phy"},
	{0xB4FD78E021DE442FULL, "genip_pp_hte_standing.utp"},
	{0xB4FD9329715AA690ULL, "zz_fen_debug6134370.cl"},
	{0xB4FDF88774FD53C5ULL, "vxlt_0_lowtown_docks_n_bg9_0_0.vlm"},
	{0xB4FE52866CEE3D0EULL, "364679038.ogg"},
	{0xB4FE698AB34AEE01ULL, "dre000_stroud.bnk"},
	{0xB4FE6ACAFC434175ULL, "vxca_caverns1_cml_0_0.vlm"},
	{0xB4FEEF2F84F6245DULL, "w_bonepitmining_a2_170_0.mao"},
	{0xB4FEF32F84F62A96ULL, "w_bonepitmining_a2_170_0.mmh"},
	{0xB4FEFD2F84F63B94ULL, "w_bonepitmining_a2_170_0.msh"},
	{0xB4FF0B23582693B1ULL, "6180958_m.fxe"},
	{0xB4FF7787EF953FAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_haymaker_action.gad"},
	{0xB4FF7F2CB3C33F35ULL, "993539292.ogg"},
	{0xB4FFBD262C025035ULL, "6170005_m.fxe"},
	{0xB5004B9D74F49851ULL, "139196581.ogg"},
	{0xB5011EFD9CB47696ULL, "dre000_sandal6028409.cl"},
	{0xB5018B8AB34D5E71ULL, "dre000_stroud.cnv"},
	{0xB501A0AE12DB2A84ULL, "plt_cod_mgc_chant_structure.plo"},
	{0xB501B64C8C3FA6A8ULL, "vxlt_undercitydungeon_11w_0_0.vlm"},
	{0xB501C096073210ABULL, "722395076.ogg"},
	{0xB502445DFB6A9177ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_noble2f_arishokaddress.ani"},
	{0xB502AAC204D890C3ULL, "w_blightlands_250_0.mao"},
	{0xB502AEC204D89710ULL, "w_blightlands_250_0.mmh"},
	{0xB502BCC204D8AF2AULL, "w_blightlands_250_0.msh"},
	{0xB50367086161BF1AULL, "dae000_meredith6088201.cl"},
	{0xB5036A78342A8609ULL, "122599067.ogg"},
	{0xB503D41C27453F0AULL, "fx_ice_mesh.mao"},
	{0xB5043CBBE74F8DBCULL, "ave100_donnic6106127.cl"},
	{0xB5047FC8E0CAF220ULL, "6125610_f.fxe"},
	{0xB504E8FBD7F3B7EDULL, "6087520_f.fxe"},
	{0xB505642280192804ULL, "378688956.ogg"},
	{0xB505CA6A61B64A6EULL, "310342377.ogg"},
	{0xB5067287EF9B558DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punch_haymaker_action.evt"},
	{0xB50699F9F822034CULL, "vxw_bonepitdestroyed_a3_1cf_0_0.vlm"},
	{0xB506EFC189B1FF65ULL, "vxlt_warehouse_n_12n_0_0.vlm"},
	{0xB507B4756D0DDE75ULL, "642724136.ogg"},
	{0xB5082FC9B0BDC128ULL, "vxca_caverns1_cp2_0_0.vlm"},
	{0xB5085E365F7B220EULL, "one280_emrys6104255.cl"},
	{0xB50896008D855E8CULL, "205423673.ogg"},
	{0xB509E33333BCC89CULL, "zz_mcr_debug6080791.cl"},
	{0xB509EAA6CCB37620ULL, "6130088_m.fxe"},
	{0xB50A386E7AA5254DULL, "6160798_m.fxe"},
	{0xB50A4517B065E614ULL, "vxlt_warehouse_n_sk_0_0.vlm"},
	{0xB50A9A0C6AA8AC17ULL, "74767662.ogg"},
	{0xB50AA6624769837AULL, "art\\vfx\\combat\\vfx_ha_bd_mblast_aoe_p.mmh"},
	{0xB50B0992CBA7C85CULL, "860708043.ogg"},
	{0xB50B585041A878D9ULL, "vxht_0_hightown_a3_b83_0_0.vlm"},
	{0xB50BEFABDEC89F7CULL, "dae210_gate_guard6085042.cl"},
	{0xB50C5431B3FE621DULL, "w_encounter3_191_0.mao"},
	{0xB50C5831B3FE6956ULL, "w_encounter3_191_0.mmh"},
	{0xB50C6231B3FE7A54ULL, "w_encounter3_191_0.msh"},
	{0xB50C9018D842EFC2ULL, "hlac_ht_mansion2_818_2x.dds"},
	{0xB50CB41C64335E44ULL, "6152754_m.fxe"},
	{0xB50CD87F4FA9C0EBULL, "vxlt_undercityencounter_u6_0_0.vlm"},
	{0xB50D71E10C4F8399ULL, "vxca_caverns1_mines_cos_0_0.vlm"},
	{0xB50DA8295135D358ULL, "6079655_m.fxe"},
	{0xB50DDF950E8EB029ULL, "vxw_sundermount_31a_0_1.vlm"},
	{0xB50E328A73092B17ULL, "557442466.ogg"},
	{0xB50E67E7C48397A2ULL, "vxht_0_hightown_a3_c56_0_0.vlm"},
	{0xB50F2F0A9090DCC9ULL, "ss_anvil.dds"},
	{0xB50FDC3A7D502E55ULL, "6042289_m.fxe"},
	{0xB50FFD54AC2C8B37ULL, "6174432_m.fxe"},
	{0xB51034A8F40D939EULL, "6201528_m.fxe"},
	{0xB5106714B8672D7EULL, "487642980.ogg"},
	{0xB5106A7FA933DC39ULL, "hlac_w_sundermount_131077_0y.dds"},
	{0xB510803CDC1C5568ULL, "vxht_0_hightown_a2_c5q_0_0.vlm"},
	{0xB511D5ADE69EA28EULL, "vxht_0_hightown_n_c4g_0_0.vlm"},
	{0xB51200316F83CADBULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hanginga_sa.dds"},
	{0xB5121BC1F7D3FFC6ULL, "ser221_party_comments6209945.cl"},
	{0xB5125F01D6434229ULL, "163108042.ogg"},
	{0xB51294CFC62D1D4BULL, "271729250.ogg"},
	{0xB5136E115BD13562ULL, "613485515.ogg"},
	{0xB5136F7ECBED6E22ULL, "dae000_party_night_comment6201607.cl"},
	{0xB51437C4AFDFD507ULL, "vxht_0_hightown_am8_0_0.vlm"},
	{0xB5143C16029D55F3ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.dura.evt"},
	{0xB5143F2610915D53ULL, "ht_0_hightown_36_0.rml"},
	{0xB5145A061CC65417ULL, "vxlt_0_lowtown_a2_ghq_0_0.vlm"},
	{0xB514F5A5FBA3FA2CULL, "gen00fl_aveline6129065.cl"},
	{0xB514F84C67DFF010ULL, "6116410_f.fxe"},
	{0xB5150BD6EA181BCCULL, "vxw_bonepitdestroyed_a3_1dg_0_0.vlm"},
	{0xB5150E045D498574ULL, "722655304.ogg"},
	{0xB51606C2AF425788ULL, "848593876.ogg"},
	{0xB5161FAD05F0BC4EULL, "722351877.ogg"},
	{0xB51694F2B067F80EULL, "vxca_caverns2_8ib_0_1.vlm"},
	{0xB516994A38C364F4ULL, "736772808.ogg"},
	{0xB517229BE4064C5AULL, "pro000_guard6013449.cl"},
	{0xB51731086ECE4B6AULL, "vxlt_0_lowtown_n_csf_0_0.vlm"},
	{0xB51788E92D2894C5ULL, "2968297.ogg"},
	{0xB5181AB9D90BC6A0ULL, "vxlt_gallowstemplar_n_6sn_0_0.vlm"},
	{0xB51869D181935D85ULL, "w_bonepitdestroyed_a3_26_0_ll.mmh"},
	{0xB518754CFB383E0FULL, "hlac_ht_keep_n_4519_12y.dds"},
	{0xB51884642AB9DDADULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_spikexa.msh"},
	{0xB51892642AB9F5E0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_spikexa.mao"},
	{0xB518957FF5E84235ULL, "gen00fl_fenris6126448.cl"},
	{0xB51896642AB9FC33ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_spikexa.mmh"},
	{0xB518B6F4A25BFE94ULL, "vxlt_warehouse_n_vh_0_0.vlm"},
	{0xB518FF7AE2DD51C8ULL, "6060175_f.fxe"},
	{0xB519136B0795A33BULL, "ser221_quentin6128859.cl"},
	{0xB51919656F5CA739ULL, "gen_im_trs_gen_one_mac_r2.uti"},
	{0xB5191B79D7608662ULL, "vxw_bonepitmining_a2_ou_0_0.vlm"},
	{0xB5191E6D4A0DBF8AULL, "6037376_m.fxe"},
	{0xB51A0698DA7DCBBDULL, "lt_gallowscourtyard_42_0.rml"},
	{0xB51A24CCCB79EECEULL, "hlac_ht_bartrand_n_3370_0x.dds"},
	{0xB51A525F34A63BF6ULL, "vxw_encounter3_3x_0_0.vlm"},
	{0xB51A6B5834E24FDAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha06a.phy"},
	{0xB51A7F9819166203ULL, "w_bonepitdestroyed_a3_21_0.msh"},
	{0xB51A8D9819167A5DULL, "w_bonepitdestroyed_a3_21_0.mmh"},
	{0xB51A91981916816EULL, "w_bonepitdestroyed_a3_21_0.mao"},
	{0xB51ADE554F1C179CULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga102a_0d.dds"},
	{0xB51B144371F134D0ULL, "vxw_encounter2_6z_0_0.vlm"},
	{0xB51B351602A36834ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.dura.gad"},
	{0xB51B9FCB711D4A52ULL, "vxca_caverns1_k6v_0_0.vlm"},
	{0xB51BFB3BD83F1D97ULL, "lt_0_lowtown_docks_a3_187_0.mmh"},
	{0xB51BFF3BD83F24A4ULL, "lt_0_lowtown_docks_a3_187_0.mao"},
	{0xB51C113BD83F4309ULL, "lt_0_lowtown_docks_a3_187_0.msh"},
	{0xB51C913FA757B9B6ULL, "vxlt_0_lowtown_a2_bz7_0_0.vlm"},
	{0xB51D2D1D22C15C2FULL, "vxca_caverns1_jld_0_0.vlm"},
	{0xB51D330E5939324FULL, "follower_banter6173355.cl"},
	{0xB51D4DCA1FFF4E05ULL, "891254404.ogg"},
	{0xB51E24D8E251AE35ULL, "hlac_lt_undercitydungeon_-1_3y.dds"},
	{0xB51E745C0D0BFCBDULL, "mag291_ntbanter6056824.cl"},
	{0xB51EF219CE126DE8ULL, "vxht_0_hightown_a2_c0n_0_0.vlm"},
	{0xB51F09E11FCD1751ULL, "plt_ssf200pt_main.plo"},
	{0xB51F765A4DAFEE70ULL, "hlac_lt_0_lowtown_a3_n_c_65536_3y.dds"},
	{0xB520478AD894BB0EULL, "vxht_0_hightown_n_c1x_0_0.vlm"},
	{0xB5209EE95823DCF7ULL, "dae111st_mother.stg"},
	{0xB520BDFA59999BEEULL, "vxlt_0_lowtown_n_e8e_0_0.vlm"},
	{0xB520D4086C3DF8D7ULL, "ttx_pillar_mid_1.msh"},
	{0xB520EA086C3E1E49ULL, "ttx_pillar_mid_1.mmh"},
	{0xB521002FE693ACFEULL, "900937701.ogg"},
	{0xB521402DD3F5F9D1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03b.mmh"},
	{0xB521441602A7ED12ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.dura.ani"},
	{0xB5215A2DD3F6258FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_mth03b.msh"},
	{0xB5215BA1594700CCULL, "vxdr_primevaltomb_8mc_0_0.vlm"},
	{0xB521A913F6F6D8A1ULL, "dae000ip_crowd_lm3_talk11.utp"},
	{0xB5221182B1277B99ULL, "vxca_caverns2prc_7oa_0_1.vlm"},
	{0xB5223C2CA1DA0157ULL, "vxw_sundermount_37e_0_1.vlm"},
	{0xB52255596E94AD10ULL, "vrc321ip_spooky5_obj_2.utp"},
	{0xB522D3129A7B4EA6ULL, "vxca_caverns1_mines_ih3_0_0.vlm"},
	{0xB522D4E3C7A18BAFULL, "vxw_encounter2_camp_pg_0_0.vlm"},
	{0xB522D63F82A2BFF5ULL, "482284812.ogg"},
	{0xB522DEC24813A2B6ULL, "vxdr_primevaltomb_5b1_0_0.vlm"},
	{0xB5232377FF21A4FBULL, "follower_banter6164005.cl"},
	{0xB523A7BB406B54B8ULL, "6042684_m.fxe"},
	{0xB523D7E49E323767ULL, "6084717_m.fxe"},
	{0xB5243B8C7C3C6AE5ULL, "hlac_lt_undercitydungeon_-1_13x.dds"},
	{0xB52494CF5DC7C2FCULL, "hm_wesley.mor"},
	{0xB524E2B6AD57AA51ULL, "294188454.ogg"},
	{0xB525049374D83643ULL, "lt_0_lowtown_docks_114_0.mao"},
	{0xB525089374D83C90ULL, "lt_0_lowtown_docks_114_0.mmh"},
	{0xB525169374D854AAULL, "lt_0_lowtown_docks_114_0.msh"},
	{0xB5251ED258EDC2E3ULL, "6099719_m.fxe"},
	{0xB52549DBA51A7261ULL, "vxlt_gallowstemplar_f_372_0_0.vlm"},
	{0xB5255972E84BB295ULL, "zz_test_convo.cnv"},
	{0xB525650671E36044ULL, "6083724_m.fxe"},
	{0xB5256FD740280EBDULL, "6108442_m.fxe"},
	{0xB525A66EA104111FULL, "vxca_caverns1_mines_ijx_0_0.vlm"},
	{0xB525F8615A67E3E7ULL, "vxlt_0_lowtown_n_byh_0_0.vlm"},
	{0xB52670B10FD5B230ULL, "vxlt_hangedman_n_1li_0_0.vlm"},
	{0xB526B6D8D91E9A4BULL, "vxlt_0_lowtown_a2_gvk_0_0.vlm"},
	{0xB526BFDD9DB23015ULL, "gen00fl_varric6076046.cl"},
	{0xB5278D56C9569EE2ULL, "vxw_bonepitmining_a2_jv_0_0.vlm"},
	{0xB5283C0A2594A9D9ULL, "vxht_0_hightown_a3_b21_0_0.vlm"},
	{0xB528DA5E75B40290ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb09a.mmh"},
	{0xB528DABD9500D843ULL, "hlac_lt_0_lowtown_docks_n_196607_1x.dds"},
	{0xB528E7D28E3D8CD2ULL, "llac_lt_house_-1_0.dds"},
	{0xB528E85E75B41AAAULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb09a.msh"},
	{0xB52938A7A1C5A9E3ULL, "mcr371_hostile_anders6082135.cl"},
	{0xB52962BBDCEE457BULL, "dae000_bartrand6206927.cl"},
	{0xB529A67ECF9ACAF9ULL, "6189337_m.fxe"},
	{0xB529C48DB8A2D87CULL, "hlac_w_sundermount_-196611_0z.dds"},
	{0xB529DBF885C365FBULL, "6183816_m.fxe"},
	{0xB52A028D161E22E7ULL, "vxlt_0_lowtown_a2_n_c_e1y_0_0.vlm"},
	{0xB52A5D76BAF5438FULL, "zz_mag_debug6089446.cl"},
	{0xB52A60A5B8E49E07ULL, "hlac_ht_playershouse_n_855_12y.dds"},
	{0xB52AD03D3C90EFF5ULL, "vxw_encounter2_fk_0_0.vlm"},
	{0xB52B10B508DE3884ULL, "229982691.ogg"},
	{0xB52B230D440B18CDULL, "hlac_lt_0_lowtown_-65536_6z.dds"},
	{0xB52BD685F5856E14ULL, "vxlt_darktown_4qq_0_0.vlm"},
	{0xB52C22458161219FULL, "6127537_m.fxe"},
	{0xB52CBF2BC87EE06EULL, "fenris_3.cut"},
	{0xB52D005012E63EC3ULL, "gen00fl_aveline6053277.cl"},
	{0xB52DD40761AC4DC5ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonc_head.mao"},
	{0xB52E2656E06367ABULL, "qcr220_grey_wardens6060462.cl"},
	{0xB52E471A437BC434ULL, "gen00fl_carver6104841.cl"},
	{0xB52E48E0BBA749B1ULL, "qun221_petrice6185476.cl"},
	{0xB52E9528D482F1C9ULL, "qun111_ketojan6039003.cl"},
	{0xB52ED291666F1E48ULL, "vxlt_gallowstemplar_f_377_0_0.vlm"},
	{0xB52F48DF31CFB02DULL, "gen_im_cft_reg_j1.uti"},
	{0xB52F4BC18B428A48ULL, "hlac_lt_gallowstemplar_n_2384_7y.dds"},
	{0xB52F4E0F2504314EULL, "vxlt_0_lowtown_n_gcv_0_0.vlm"},
	{0xB52F5E9DDC22CECDULL, "359626154.ogg"},
	{0xB52F649183948E96ULL, "266145567.ogg"},
	{0xB5305DE22836EC56ULL, "gen00fl_aveline6115464.cl"},
	{0xB530AE0993D019D7ULL, "vxw_sundermount_32f_0_1.vlm"},
	{0xB5311BE0C7EA8D82ULL, "vxca_caverns2prc_8q2_0_0.vlm"},
	{0xB5313DC000B28517ULL, "vxlt_0_lowtown_a2_gbs_0_0.vlm"},
	{0xB53146C0B997A42FULL, "vxw_encounter2_camp_kf_0_0.vlm"},
	{0xB53147C026C9801BULL, "art\\levels\\ztools\\default\\generic\\9ww_signshop.msh"},
	{0xB53159C026C99F66ULL, "art\\levels\\ztools\\default\\generic\\9ww_signshop.mao"},
	{0xB53165C026C9B305ULL, "art\\levels\\ztools\\default\\generic\\9ww_signshop.mmh"},
	{0xB5317C3581E3CF0DULL, "vfx_vfx_p4.vsh.fxo.dx11"},
	{0xB531EA080BEF61B1ULL, "qcr000_arishok6059340.cl"},
	{0xB531EB6EAA85545BULL, "vxht_mansion2_1u4_0_0.vlm"},
	{0xB532138A4F7A0ABFULL, "1036955871.ogg"},
	{0xB532F88D2FFFB346ULL, "vxlt_gallowscourtyard_n_62i_0_0.vlm"},
	{0xB53312A4601B7EC6ULL, "l_dae351ar_mount_caverns.wbd"},
	{0xB533A6AA6950295BULL, "hlac_lt_0_lowtown_a3_n_c_131072_2z.dds"},
	{0xB53466E831AD3060ULL, "zz_and_debug6134165.cl"},
	{0xB53482CDFBF86A7FULL, "w_sundermount_68_0_lc.mmh"},
	{0xB53483B0A431A4C2ULL, "lgt260cr_way_apostate_2.utc"},
	{0xB5349F26DE6BD86BULL, "6065338_m.fxe"},
	{0xB53510F9BEC07901ULL, "6111754_m.fxe"},
	{0xB53520CB40A3F83AULL, "hlac_ca_caverns2_5802_3x.dds"},
	{0xB53528B5CB14B2CBULL, "vxlt_0_lowtown_a2_gmj_0_0.vlm"},
	{0xB5353B3263C9DE1FULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmenter.gad"},
	{0xB53550713F233D99ULL, "open.cur"},
	{0xB53587B6902A8BAFULL, "lgt120cr_orz_bln_official.utc"},
	{0xB53651EA776A56C5ULL, "hla_ht_0_hightown_-2_0z.dds"},
	{0xB536611BFD418139ULL, "lt_gallowscourtyard_n_42_0_lc.mmh"},
	{0xB536A00331CFDE93ULL, "hlac_w_bonepitdestroyed_a3_196613_0y.dds"},
	{0xB53752DC8C3EBDB1ULL, "llac_lt_0_lowtown_a2_n_-262144_1.dds"},
	{0xB537758AB37AC872ULL, "dre000_stroud.stg"},
	{0xB537A55C97E7C66BULL, "art\\characters\\playercharacter\\humanmale\\boo_meda.phy"},
	{0xB5389683FDD4436EULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiresml.mmh"},
	{0xB539421A2E870875ULL, "vxw_encounter2_aj_0_0.vlm"},
	{0xB5396D516BE57715ULL, "gen00fl_varric6118742.cl"},
	{0xB539991F8F3F026BULL, "6093890_m.fxe"},
	{0xB53A0B0364815203ULL, "vxca_caverns1_mines_ehn_0_0.vlm"},
	{0xB53A51828145630AULL, "fxa_cen_aoe_slp.mao"},
	{0xB53A5944683AD366ULL, "6068541_m.fxe"},
	{0xB53AAF81A78D7C32ULL, "257073802.ogg"},
	{0xB53AD2060B66829BULL, "ht_chantry_mmap_ph.dds"},
	{0xB53AD887F116D908ULL, "106855089.ogg"},
	{0xB53AE7BE35D11B75ULL, "plt_cod_crt_rage_demon.plo"},
	{0xB53AFF7481261777ULL, "hlac_lt_0_lowtown_a3_n_c_-65538_0z.dds"},
	{0xB53B16A3FE441862ULL, "32754810.ogg"},
	{0xB53B1E1F17560E2FULL, "hlac_w_sundermount_-65534_0z.dds"},
	{0xB53B2DD6E2D4B5EEULL, "fen200cr_hadriana.utc"},
	{0xB53B667B649F4E7EULL, "dre180ip_deep_roads_chest_1.utp"},
	{0xB53BBA2AF938FF1BULL, "hlac_w_encounter2_camp_262149_0y.dds"},
	{0xB53BE23FD034EA26ULL, "target_inactive.cur"},
	{0xB53C0C790BB2B35CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_105a_n.dds"},
	{0xB53C303263CFEA0CULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmenter.evt"},
	{0xB53C7B44781B3B83ULL, "dae210_gate_guard6085128.cl"},
	{0xB53CB23F1B4D6EFCULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v4.ani"},
	{0xB53D0C140F6F7317ULL, "w_bonepit_120_0.mmh"},
	{0xB53D10140F6F7A24ULL, "w_bonepit_120_0.mao"},
	{0xB53D22140F6F9889ULL, "w_bonepit_120_0.msh"},
	{0xB53D9E78C0DC4BEBULL, "23228035.ogg"},
	{0xB53DC974A200CE5AULL, "follower_banter6163524.cl"},
	{0xB53ECE0D5DD9FF21ULL, "vxlt_0_lowtown_a3_n_c_emn_0_0.vlm"},
	{0xB53F1B92E08FE162ULL, "202428462.ogg"},
	{0xB53F2DA92C31241FULL, "vxht_mansion2b_31j_0_0.vlm"},
	{0xB53F34DD2BD9AA4BULL, "273960684.ogg"},
	{0xB53F66E9252685FFULL, "22591589.ogg"},
	{0xB53FB89DAB8DBCAFULL, "vxw_encounter2_camp_fa_0_0.vlm"},
	{0xB54009A6FAC69C3CULL, "6162537_m.fxe"},
	{0xB540155F6D6F3B52ULL, "226379614.ogg"},
	{0xB54023CCBAA75E52ULL, "vfx_alphacloudanim01.dds"},
	{0xB5402D49E8181597ULL, "6070506_f.fxe"},
	{0xB540552D4EC154ACULL, "81024531.ogg"},
	{0xB540A87BDA78F1ECULL, "vxht_bartrand_n_65l_0_0.vlm"},
	{0xB540EC7BA6F45FB0ULL, "mag121_idunna6083945.cl"},
	{0xB5417F95C78A18D3ULL, "dae000st_grandclericseb.stg"},
	{0xB5422D958906A361ULL, "vxlt_gallowstemplar_f_390_0_0.vlm"},
	{0xB54241D08AEA82ADULL, "vxht_bartrand_n_5te_0_0.vlm"},
	{0xB542539124143FBDULL, "6106642_m.fxe"},
	{0xB5429DF529E66C4CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_medp_0s.dds"},
	{0xB542C974451C441FULL, "102819433.ogg"},
	{0xB543090C378235C0ULL, "6125438_f.fxe"},
	{0xB54336F5A32B2322ULL, "226603803.ogg"},
	{0xB54361F88A1BDABEULL, "hlac_w_woundedcoast_327687_0z.dds"},
	{0xB543848A98D4D9DAULL, "hlac_lt_gallowstemplar_2394_0x.dds"},
	{0xB543884A9DDD05A0ULL, "6147235_m.fxe"},
	{0xB543BF3F1B53A33EULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v4.gad"},
	{0xB543D75CD85BA3DEULL, "hlac_lt_0_lowtown_docks_n_131070_3z.dds"},
	{0xB54414881367A027ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_wng_hdra_0s.dds"},
	{0xB5445DBFADC54380ULL, "803797499.ogg"},
	{0xB545133B204BCE33ULL, "6004160_m.fxe"},
	{0xB545A7B6E33B9433ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_beyondpossession_d.dds"},
	{0xB545B583FDDF026DULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiresml.anb"},
	{0xB54671ECB2BA59CFULL, "6103229_m.fxe"},
	{0xB546BD367217B2D0ULL, "vxw_blightlands_kr_0_0.vlm"},
	{0xB546C3B8C1754B22ULL, "114750475.ogg"},
	{0xB546F5E949A9E27DULL, "hlac_ca_cavern1_haunted_24426_12z.dds"},
	{0xB54727D03805CD36ULL, "6082563_m.fxe"},
	{0xB547473229EABE93ULL, "6132759_m.fxe"},
	{0xB547B3F7207D20F5ULL, "vxw_encounter2_li_0_0.vlm"},
	{0xB547D2486F96B16CULL, "vxht_mansion2b_2n5_0_0.vlm"},
	{0xB547F85440C3F0D5ULL, "6201291_m.fxe"},
	{0xB54804439EE28FECULL, "6107154_m.fxe"},
	{0xB548078D3CAC2A99ULL, "167136205.ogg"},
	{0xB5482B625A9EEF04ULL, "6042060_m.fxe"},
	{0xB5487A0C96F4F08FULL, "vxca_caverns2prc_7u0_0_0.vlm"},
	{0xB5488EDD7ABE1023ULL, "ser270_templar6112574.cl"},
	{0xB54892CBB9DC8FCAULL, "vxht_0_hightown_4kf_0_0.vlm"},
	{0xB548B160FA873B25ULL, "dae000_mother6099245.cl"},
	{0xB548C468C5F1CE8DULL, "llac_lt_backalley_n_0_6.dds"},
	{0xB54989FF4ECB0401ULL, "gen_im_arm_cht_elf_comc_t6.uti"},
	{0xB549CF620694A5DAULL, "vxlt_0_lowtown_a2_n_c_i14_0_0.vlm"},
	{0xB54A183263DBFF65ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\overwhelmenter.ani"},
	{0xB54A9A3F1B59835DULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v4.evt"},
	{0xB54AAAFCFCC2CEE8ULL, "gen00fl_bethany6101004.cl"},
	{0xB54B7E860170D2EEULL, "l_dae260ar_wounded_coast.wbd"},
	{0xB54B9CCAB9073A7AULL, "954344373.ogg"},
	{0xB54BB64B4A5B4F48ULL, "vxlt_gallowstemplar_f_395_0_0.vlm"},
	{0xB54C0AEF5EB32505ULL, "gui_music.wbd"},
	{0xB54C0D1ED1B87F13ULL, "104149345.ogg"},
	{0xB54C0DB943B606FBULL, "vxlt_warehouse_d_0_0.vlm"},
	{0xB54C4DC351B493D6ULL, "ht_0_hightown_n_25_0.anb"},
	{0xB54C5C17DA7895C4ULL, "6164085_m.fxe"},
	{0xB54C74FF9BD7B1A4ULL, "438721966.ogg"},
	{0xB54D0598DAA923D5ULL, "lt_gallowscourtyard_42_0.anb"},
	{0xB54D2D73863EC2C9ULL, "lt_gallowscourtyard_51_0.rml"},
	{0xB54D3A34DF20CF2AULL, "one120_athenril6088337.cl"},
	{0xB54D9EC4E4160D45ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_1z.dds"},
	{0xB54DCDEDED4558D9ULL, "ach_exp_night.dds"},
	{0xB54DF2CEB8512F49ULL, "vxw_woundedcoast_gm_0_0.vlm"},
	{0xB54E0851B5A41874ULL, "vxw_sundermount_22m_0_0.vlm"},
	{0xB54E489178B18C7FULL, "lt_backalley_n_3_0.rml"},
	{0xB54E4C381CF6C482ULL, "6004193_f.fxe"},
	{0xB54E88092AAFE49EULL, "vxw_bonepit_9m_0_0.vlm"},
	{0xB54EA570DC1A19F0ULL, "vxw_blightlands_11q_0_0.vlm"},
	{0xB54EB319D69DF11FULL, "vxlt_gallowstemplar_6sj_0_0.vlm"},
	{0xB54ECF288E71855BULL, "vxht_mansion2_1c6_0_0.vlm"},
	{0xB54EDA8EC3E932C3ULL, "dae000_mother6028517.cl"},
	{0xB54F6B50B18A8262ULL, "dae000_arishok6068510.cl"},
	{0xB54FA1C7CB232916ULL, "287918405.ogg"},
	{0xB54FDB971CEC6150ULL, "85069239.ogg"},
	{0xB550403C20DC69B2ULL, "llac_ca_cavern1_haunted_24421_12.dds"},
	{0xB550AA33A96E6FD1ULL, "hlac_lt_gallowstemplar_n_2383_0x.dds"},
	{0xB551C8A56AFCB68FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_bounce_small.gad"},
	{0xB551FB00F934D717ULL, "6217117_m.fxe"},
	{0xB5521B4F0AED0442ULL, "ht_0_hightown_37_0.rml"},
	{0xB5523369D54AEF7AULL, "320606003.ogg"},
	{0xB55244FD80022054ULL, "148116056.ogg"},
	{0xB552613BC61570A6ULL, "vxlt_0_lowtown_a2_n_e81_0_0.vlm"},
	{0xB5527E48A4E050C9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wc.phy"},
	{0xB5529407DC6190E0ULL, "dre000_dougal6167437.cl"},
	{0xB5533EC2429E2FEEULL, "hlac_ca_caverns2prc_5802_59z.dds"},
	{0xB55414D12CA55413ULL, "vxca_cavern1_haunted_edj_0_0.vlm"},
	{0xB554264DE894E37CULL, "combat_smr3_range_only.gda"},
	{0xB55445F1603EA98AULL, "vxw_bonepitmining_a2_c1_0_0.vlm"},
	{0xB554A51EB62C9D4AULL, "mag160_feynriel6033386.cl"},
	{0xB554E3C9A4B0724FULL, "6104329_m.fxe"},
	{0xB5552F13640DCB50ULL, "vxw_blightlands_ls_0_0.vlm"},
	{0xB5557D1B6EB36E2CULL, "368525739.ogg"},
	{0xB555871F99EC2299ULL, "642348398.ogg"},
	{0xB5559DFEC580C2BCULL, "617344061.ogg"},
	{0xB555A22F29838AE5ULL, "vxht_0_hightown_a2_9zo_0_0.vlm"},
	{0xB555B4F1471DE6F1ULL, "6097438_m.fxe"},
	{0xB555C00F6F6EB8C4ULL, "vxlt_0_lowtown_docks_n_b0y_0_0.vlm"},
	{0xB556079786DF40F1ULL, "vxca_cavern1_haunted_ild_0_0.vlm"},
	{0xB55623E52016CEAFULL, "dae000_gen_turn_in_36198693.cl"},
	{0xB5566D5E02C496FFULL, "vxca_caverns2prc_7qt_0_0.vlm"},
	{0xB5567195FE960866ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_up.ani"},
	{0xB55684333D5703D7ULL, "hlac_lt_0_lowtown_docks_a3_-3_1y.dds"},
	{0xB5569B9B212C1843ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.dura.evt"},
	{0xB556B0464F4CC452ULL, "ser311ar_dungeon.ncs"},
	{0xB556D54F8662B5C5ULL, "6164453_m.fxe"},
	{0xB556D57088202EFAULL, "vfx_distortionmask_p2.vsh.fxo"},
	{0xB556EEBD486D8303ULL, "vxca_caverns1_mines_ebp_0_0.vlm"},
	{0xB5571AE2642BF6A4ULL, "855136677.ogg"},
	{0xB5571FDE4D6517FBULL, "w_bonepitmining_a2_243_0.mmh"},
	{0xB5572BDE4D652B98ULL, "w_bonepitmining_a2_243_0.mao"},
	{0xB5573DDE4D654AE5ULL, "w_bonepitmining_a2_243_0.msh"},
	{0xB55755CD5249F7BBULL, "zz_followers_banter6158589.cl"},
	{0xB5579E4D9D3115AFULL, "art\\vfx\\combat\\vfx_r_dwp_leftb_close1_c.mmh"},
	{0xB557DDA56B01463CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_bounce_small.evt"},
	{0xB557E21B2AEAA804ULL, "764677456.ogg"},
	{0xB5585925BCDAA11CULL, "6166919_m.fxe"},
	{0xB55959AAFCEE6C52ULL, "26414122.ogg"},
	{0xB559774C016421E1ULL, "376507205.ogg"},
	{0xB5598D2C0A16FF58ULL, "6027710_f.fxe"},
	{0xB559ADA8445C4826ULL, "vxca_caverns1_mines_j5o_0_0.vlm"},
	{0xB559E74A137BEE39ULL, "6057099_f.fxe"},
	{0xB55A2ED6E2EF67BCULL, "fen200cr_hadriana.ncs"},
	{0xB55A572C9784F53EULL, "6173339_m.fxe"},
	{0xB55A5D56074336F0ULL, "vxlt_backalley_n_1r6_0_0.vlm"},
	{0xB55A96126656B6C5ULL, "lt_gallowscourtyard_n_mmap_ph.dds"},
	{0xB55B5140BC691719ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenb.msh"},
	{0xB55B5B40BC692807ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenb.mmh"},
	{0xB55B5F40BC692F54ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_brokenb.mao"},
	{0xB55B702841AD49DDULL, "vxht_brothel_n_4x8_0_0.vlm"},
	{0xB55BBFEC168DF2FAULL, "6117684_f.fxe"},
	{0xB55C5543A5A29BE8ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig_r.dds"},
	{0xB55C7A2EA79A30F4ULL, "vxw_sundermount_27r_0_0.vlm"},
	{0xB55C9C4A4276D9B6ULL, "vxht_0_hightown_a2_9uy_0_0.vlm"},
	{0xB55D6999ADF90122ULL, "481820423.ogg"},
	{0xB55D749B2131F4C4ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.dura.gad"},
	{0xB55DEC43023BCE6FULL, "289022732.ogg"},
	{0xB55E45D010AED271ULL, "vxlt_0_lowtown_a3_n_c_e27_0_0.vlm"},
	{0xB55E909A30A978C5ULL, "fen110_fenris6100306.cl"},
	{0xB55F6F5E59952392ULL, "gen00fl_isabela6075288.cl"},
	{0xB55FD42E958127C1ULL, "487250386.ogg"},
	{0xB5600688151D6BCDULL, "6110506_m.fxe"},
	{0xB560639FB01A11B0ULL, "hlac_lt_0_lowtown_a2_n_c_65538_5z.dds"},
	{0xB5607E1369F948A9ULL, "fen310_fenris6118222.cl"},
	{0xB5615766FCB94BFFULL, "bld_drian_tu2.dds"},
	{0xB561D64D45374EECULL, "110609807.ogg"},
	{0xB5627E570525AC81ULL, "qun110ip_compound_gate.utp"},
	{0xB562B7CE5234C20AULL, "vxw_bonepitmining_a2_n2_0_0.vlm"},
	{0xB562F2BFDC92F571ULL, "w_bonepitdestroyed_a3_160_0.msh"},
	{0xB56300BFDC930C8CULL, "w_bonepitdestroyed_a3_160_0.mao"},
	{0xB5630CBFDC93212FULL, "w_bonepitdestroyed_a3_160_0.mmh"},
	{0xB5636DC3FB480F8CULL, "77684006.ogg"},
	{0xB563839B213679A2ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe_p.dura.ani"},
	{0xB563A0F05603E3D0ULL, "vxw_blightlands_qp_0_0.vlm"},
	{0xB5644DE7B09B48C7ULL, "323916777.ogg"},
	{0xB5645995FEA21CD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_up.evt"},
	{0xB56482BE49BCC886ULL, "6129981_m.fxe"},
	{0xB56497B1046951F5ULL, "vxw_encounter2_rw_0_0.vlm"},
	{0xB564CCD3B3DF861AULL, "vrc221_bartrand6065291.cl"},
	{0xB564DF3AF4BAAF7FULL, "vxca_caverns2prc_7tw_0_0.vlm"},
	{0xB565609A3A639B83ULL, "vxca_caverns1_mines_egs_0_0.vlm"},
	{0xB5657CE74EB4F47BULL, "vxht_bartrand_n_5ru_0_0.vlm"},
	{0xB56582F0246DCF22ULL, "vxdr_0_deeproads_cua_0_1.vlm"},
	{0xB5659A9E7E936977ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb05a.phy"},
	{0xB565E5A56B0D9155ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_bounce_small.ani"},
	{0xB565E85B63962482ULL, "hlac_ht_mansion_n_3382_0x.dds"},
	{0xB56632570D05FBDDULL, "vxlt_0_lowtown_a2_h22_0_0.vlm"},
	{0xB5665EE506D5BFAEULL, "736255532.ogg"},
	{0xB56676829488EB26ULL, "dae000_mother6135366.cl"},
	{0xB5669F417D8AB9EAULL, "vxlt_0_lowtown_a2_hc6_0_0.vlm"},
	{0xB5673DEBB6CE3D06ULL, "vxht_mansion2_1w3_0_0.vlm"},
	{0xB567475DEE9A0B77ULL, "16419091.ogg"},
	{0xB5675B8185547B95ULL, "follower_banter6163468.cl"},
	{0xB5678615DD5A7193ULL, "438463290.ogg"},
	{0xB5678D81E12CDDFCULL, "vxht_playershouse_n_1re_0_1.vlm"},
	{0xB567D78179CC8584ULL, "character_character_p230.psh.fxo.dx11.assembly"},
	{0xB568488C42A250E6ULL, "dotblur_add.mao"},
	{0xB5693CC2B1758A14ULL, "6105772_m.fxe"},
	{0xB569A809F2012E76ULL, "vxht_0_hightown_n_b8r_0_0.vlm"},
	{0xB569E15E75EBA5ABULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb09a.phy"},
	{0xB56AE64BDE2CDE88ULL, "vxlt_0_lowtown_n_eit_0_0.vlm"},
	{0xB56AF69C10482B40ULL, "follower_banter6155985.cl"},
	{0xB56B5A95FEA83D60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_up.gad"},
	{0xB56B601FDEED143FULL, "vxht_bartrand_n_65c_0_0.vlm"},
	{0xB56B66B3299AA84DULL, "130703302.ogg"},
	{0xB56B814AB3DDDAFDULL, "pro000cr_wesley_tainted.utc"},
	{0xB56C45AC408C898DULL, "538794580.ogg"},
	{0xB56C9FD1FA34AE32ULL, "6106814_m.fxe"},
	{0xB56CF4B539639961ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rrogue_below_belt_reaction_enter.ani"},
	{0xB56CFCA937DEA79CULL, "gen00fl_sebastian6195842.cl"},
	{0xB56D3B9E2BFD2DBFULL, "hlac_dr_dungeon_3207_0y.dds"},
	{0xB56D6328A940B6BFULL, "6154113_m.fxe"},
	{0xB56D80972D56D5E6ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.dura.ani"},
	{0xB56DE8776E45FFA8ULL, "ht_mansion_n_3370_lc.mmh"},
	{0xB56DF3B470DF65C9ULL, "vxw_bonepit_1qe_0_0.vlm"},
	{0xB56E204F2E25C458ULL, "mag110_arianni6016898.cl"},
	{0xB56E9EFC660CDEF7ULL, "zz_dae_debug6029175.cl"},
	{0xB56EB112AB924DD9ULL, "vxca_caverns1_mines_dr7_0_0.vlm"},
	{0xB56EB1F264EC7F90ULL, "genip_pp_dr_sustained.utp"},
	{0xB56F2265DFDA44BAULL, "1037294239.ogg"},
	{0xB56F60FDDCE0D0A7ULL, "fer_ter_dirt_001.mao"},
	{0xB56FA2E371F881DFULL, "mag311_emile_de_launcet6043613.cl"},
	{0xB56FF369818BF38CULL, "gen00fl_fenris6113847.cl"},
	{0xB57020BB223FC193ULL, "vxht_chantry_n_2cz_0_0.vlm"},
	{0xB5709667AB6C4A4AULL, "458429444.ogg"},
	{0xB570DA42E8B97134ULL, "dae011_conspiracy_mob6149566.cl"},
	{0xB570F8F9402EAD3FULL, "6070444_m.fxe"},
	{0xB5710D38A0B1DBBFULL, "llac_ht_brothel_n_3374_1.dds"},
	{0xB5711954F697CA42ULL, "isa311_isabela6092261.cl"},
	{0xB5715A268F9E9468ULL, "964828931.ogg"},
	{0xB571CAB9502412C4ULL, "dae300_alistair_king6133534.cl"},
	{0xB572016FD927B41AULL, "246268460.ogg"},
	{0xB57222ADCB18174CULL, "906131215.ogg"},
	{0xB57249C2B7D716F2ULL, "vxlt_0_lowtown_a3_n_c_gzb_0_0.vlm"},
	{0xB57297A4BE0DB32FULL, "hlac_dr_dungeon_3215_8z.dds"},
	{0xB572AFAF2C5CF85CULL, "vxlt_hangedman_n_1vk_0_0.vlm"},
	{0xB572C7CBB4CB6D64ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0dd_basic.msh"},
	{0xB572CD69468C087CULL, "qun111_ketojan6040017.cl"},
	{0xB572D9CBB4CB8BCDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0dd_basic.mao"},
	{0xB572DDCBB4CB9206ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0dd_basic.mmh"},
	{0xB57303BFE4EC8EA6ULL, "6064646_m.fxe"},
	{0xB57357DAD61D7E66ULL, "isa111_isabela6067106.cl"},
	{0xB573A87ADFCA9978ULL, "hlac_lt_gallowstemplar_n_7198_4z.dds"},
	{0xB573B2B00C7F888BULL, "llac_lt_0_lowtown_a2_-262144_0.dds"},
	{0xB573E897732E9615ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopec.msh"},
	{0xB573EA97732E998BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopec.mmh"},
	{0xB573F697732EAE28ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9cc_roofslopec.mao"},
	{0xB573FFB53969CA03ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rrogue_below_belt_reaction_enter.gad"},
	{0xB57431E8AC43E0BCULL, "hlac_lt_0_lowtown_a2_n_131074_0z.dds"},
	{0xB574A1B06205E1ACULL, "6154213_m.fxe"},
	{0xB575111E6F80D26AULL, "vxlt_0_lowtown_a2_hh7_0_0.vlm"},
	{0xB575AFC8A8C45586ULL, "vxht_mansion2_1z4_0_0.vlm"},
	{0xB575E1BA59F2CEF9ULL, "leave_chunky_l2.mao"},
	{0xB57664583530A203ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha06a.msh"},
	{0xB57672583530BA5DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha06a.mmh"},
	{0xB576B403D3A8F38AULL, "6133595_m.fxe"},
	{0xB576DA96255129D5ULL, "gen00fl_bethany6109230.cl"},
	{0xB5771174BB0E67F1ULL, "6064840_f.fxe"},
	{0xB57714E120179FB6ULL, "plt_ssf200pt_main.nss"},
	{0xB5771A0981C2CCC3ULL, "6048212_m.fxe"},
	{0xB577410FEB2F67F0ULL, "vxlt_backalley_n_1x4_0_0.vlm"},
	{0xB57774860195FC1CULL, "l_dae260ar_wounded_coast.bnk"},
	{0xB577B5C197735F60ULL, "art\\vfx\\combat\\vfx_gib_cora_1.mao"},
	{0xB57815C72F759119ULL, "vxht_0_hightown_7ix_0_0.vlm"},
	{0xB578570DD779F89FULL, "hlac_ht_0_hightown_a3_-196608_2x.dds"},
	{0xB5787A63D000EF0BULL, "6160789_m.fxe"},
	{0xB57938B40D789C89ULL, "lgt160_wk_avern_bandits6112753.cl"},
	{0xB57945BA5D14CC83ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_diagonal.mao"},
	{0xB57949BA5D14D3D0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_diagonal.mmh"},
	{0xB5794DEDC46482BDULL, "6123395_quentin_impossible.cut"},
	{0xB57957BA5D14EBEAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_diagonal.msh"},
	{0xB57994CD95850056ULL, "6214423_m.fxe"},
	{0xB579C9CAE47ED952ULL, "vxca_caverns1_mines_ei9_0_1.vlm"},
	{0xB57ADCB5396FADE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rrogue_below_belt_reaction_enter.evt"},
	{0xB57B2989F49B0371ULL, "vxlt_0_lowtown_a3_n_c_e85_0_0.vlm"},
	{0xB57B53A746225D9DULL, "hlac_lt_0_lowtown_a3_n_c_-131073_3y.dds"},
	{0xB57B68972D62EA57ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.dura.evt"},
	{0xB57B7096F2867251ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.anb"},
	{0xB57C5AC0139CB054ULL, "vxlt_0_lowtown_docks_b61_0_1.vlm"},
	{0xB57C659162D57E49ULL, "vxw_bonepit_1zd_0_0.vlm"},
	{0xB57C6EE7C5476AA7ULL, "6121899_f.fxe"},
	{0xB57CD4CBB4D3D7D1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0dd_basic.phy"},
	{0xB57D0AF5CF642BEEULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potb_0.pwk"},
	{0xB57D29E2CCE4E20EULL, "dadistortion.mao"},
	{0xB57D6AB7FC1C2F77ULL, "709448463.ogg"},
	{0xB57DFF49BACB03A5ULL, "vxht_mansion_n_60q_0_0.vlm"},
	{0xB57E0A6FD7982561ULL, "llac_lt_gallowscourtyard_n_-2_1.dds"},
	{0xB57E654508D4FAD1ULL, "gen_im_arm_glv_hvy_act1_01.uti"},
	{0xB57E6952A1537C55ULL, "271657489.ogg"},
	{0xB57EADB8B914C112ULL, "6068875_m.fxe"},
	{0xB57EC22265BB3A36ULL, "vxw_bonepit_82_0_0.vlm"},
	{0xB57F0117758ED5FAULL, "vrc221_bartrand6043286.cl"},
	{0xB57F30C351DFBC6AULL, "ht_0_hightown_n_25_0.rml"},
	{0xB57F6748A50717EEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wc.mmh"},
	{0xB57F7348A5072B85ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wc.mao"},
	{0xB57F8148A50743DCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_wc.msh"},
	{0xB57F8389D82FCF62ULL, "540624762.ogg"},
	{0xB57FEFECF3479696ULL, "gen00fl_merrill6127271.cl"},
	{0xB5801508118B895EULL, "vxlt_0_lowtown_docks_a3_n_d3c_0_0.vlm"},
	{0xB580258F5C629A38ULL, "llac_ht_mansion2b_818_19.dds"},
	{0xB5802C73866A1AE9ULL, "lt_gallowscourtyard_51_0.anb"},
	{0xB5809F4D9D541CECULL, "art\\vfx\\combat\\vfx_r_dwp_leftb_close1_c.anb"},
	{0xB581218C1E5310DCULL, "vxlt_hangedman_n_1sh_0_0.vlm"},
	{0xB5815D2E5CC18A71ULL, "vxca_cavern1_haunted_icy_0_0.vlm"},
	{0xB581F23055660D55ULL, "vxht_0_hightown_a3_9us_0_0.vlm"},
	{0xB581F5B231506A34ULL, "one220_yevhen6103024.cl"},
	{0xB58260A132A1257BULL, "vxht_bartrand_n_5xw_0_0.vlm"},
	{0xB58269972D690AE0ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.dura.gad"},
	{0xB5829C2BDB574C34ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v3.anb"},
	{0xB5830CC78F1A104BULL, "236496104.ogg"},
	{0xB5833BD0D728F72DULL, "plt_cod_hst_tev_magisters.plo"},
	{0xB5845168F7D8FCDBULL, "gen_im_trs_qun_sta_fir_r2.uti"},
	{0xB5846ACF4D9A80BFULL, "691890424.ogg"},
	{0xB5847A2F8F8347ABULL, "vxw_bonepitdestroyed_a3_nm_0_0.vlm"},
	{0xB584F3F650AD4A04ULL, "289263972.ogg"},
	{0xB5854DA2192EB445ULL, "vxlt_gallowstemplar_f_3eh_0_0.vlm"},
	{0xB58677CF943E02A2ULL, "llac_lt_0_lowtown_docks_65534_5.dds"},
	{0xB586D0BB123F35FFULL, "6105555_m.fxe"},
	{0xB5876C8BDA746AB4ULL, "ts_coterie_prog_junk2.gda"},
	{0xB5877FDC4322648CULL, "art\\vfx\\combat\\vfx_gib_cora_lbrt.gad"},
	{0xB587BA4280299149ULL, "vxw_woundedcoast_sq_0_0.vlm"},
	{0xB58823267F82CE68ULL, "72910032.ogg"},
	{0xB588D8579BB3951CULL, "6070393_m.fxe"},
	{0xB588D8E9C17527A6ULL, "gen00fl_fenris6116576.cl"},
	{0xB5890D03091BF0F0ULL, "vxlt_0_lowtown_docks_a3_aqa_0_0.vlm"},
	{0xB5899196F292E7CAULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_t_p.mmh"},
	{0xB589AC0751DFC2AAULL, "6138727_m.fxe"},
	{0xB58A312FFFD3F02EULL, "421828717.ogg"},
	{0xB58B21658CE59E99ULL, "6034027_f.fxe"},
	{0xB58B74EAFF740417ULL, "w_blightlands_251_0.msh"},
	{0xB58B86EAFF74237AULL, "w_blightlands_251_0.mao"},
	{0xB58B8AEAFF742989ULL, "w_blightlands_251_0.mmh"},
	{0xB58C184F0B1E773EULL, "ht_0_hightown_37_0.anb"},
	{0xB58C1E8030E5153AULL, "vxw_bonepitdestroyed_a3_19r_0_0.vlm"},
	{0xB58C550C75A1BF42ULL, "741501765.ogg"},
	{0xB58C5C07619A2818ULL, "203803589.ogg"},
	{0xB58CBC50A7B2F679ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs05a_0s.dds"},
	{0xB58CD69F1150914FULL, "6080757_m.fxe"},
	{0xB58CE4EF5EEA7BAFULL, "gui_music.bnk"},
	{0xB58D137FAEACF6F2ULL, "6110211_m.fxe"},
	{0xB58D7EDC4326CE81ULL, "art\\vfx\\combat\\vfx_gib_cora_lbrt.anb"},
	{0xB58D7EDC4326CE8AULL, "art\\vfx\\combat\\vfx_gib_cora_lbrt.ani"},
	{0xB58DE8F5106A2686ULL, "gen_im_trs_krk_hvy_hlm_r0.uti"},
	{0xB58E5167BE1B7743ULL, "gen00fl_aveline6140133.cl"},
	{0xB58ECF9F10B94ECEULL, "6088734_m.fxe"},
	{0xB58EEB5BADEFF2D6ULL, "839054511.ogg"},
	{0xB58F53AC3DA4139EULL, "vxlt_0_lowtown_gll_0_0.vlm"},
	{0xB58F612C5899482BULL, "315821397.ogg"},
	{0xB58F6B9178E95EE3ULL, "lt_backalley_n_3_0.anb"},
	{0xB58F94835FD54830ULL, "trap_bolt.ncs"},
	{0xB58FC7302074C7FAULL, "narrative_banter6165533.cl"},
	{0xB58FD146C1D2E142ULL, "w_encounter3_247_0.msh"},
	{0xB58FD346C1D2E4F8ULL, "w_encounter3_247_0.mmh"},
	{0xB58FDF46C1D2F89BULL, "w_encounter3_247_0.mao"},
	{0xB58FEDE577BD636FULL, "art\\vfx\\combat\\vfx_r_brd_csong_act_c.impa.gad"},
	{0xB5907B2BDB6351F7ULL, "art\\vfx\\combat\\vfx_gib_hlk_bld_p_v3.mmh"},
	{0xB5907F9DDADB4D1DULL, "6060201_m.fxe"},
	{0xB5917ADF48C970E4ULL, "hlac_lt_ruinedhouse_n_-1_7y.dds"},
	{0xB591A1CB4EEECB0CULL, "vxca_caverns1_mines_e7x_0_0.vlm"},
	{0xB591AB674A8C87ECULL, "art\\characters\\weapons\\longswords\\w_lsw_024a.phy"},
	{0xB591BF88DDC723AFULL, "vrc321_varric6086888.cl"},
	{0xB59226541ACE8E52ULL, "6020453_m.fxe"},
	{0xB5928334D52F98C2ULL, "6035280_m.fxe"},
	{0xB592918E8657E021ULL, "324073406.ogg"},
	{0xB5929388B56D6F1CULL, "6060135_f.fxe"},
	{0xB592B79382734455ULL, "6102897_m.fxe"},
	{0xB592D97FCEAD2FFAULL, "hlac_w_woundedcoast_-65533_0x.dds"},
	{0xB5933F8AA654855EULL, "pro000_opening6207397.cl"},
	{0xB593AF76EC44BF1BULL, "100816041.ogg"},
	{0xB59485EFA8E6D13FULL, "6084263_m.fxe"},
	{0xB5949889A84D62F0ULL, "follower_banter6173123.cl"},
	{0xB594D226FE2CBAFEULL, "6148416_m.fxe"},
	{0xB594D2540FD29A27ULL, "375103604.ogg"},
	{0xB594EBD8F4F44CC9ULL, "249479833.ogg"},
	{0xB595F46CFA24F111ULL, "6124732_m.fxe"},
	{0xB595FB6730EE880AULL, "895667948.ogg"},
	{0xB59673048F9D0952ULL, "657680820.ogg"},
	{0xB596C8361318D07DULL, "hlac_ht_chantry_n_2534_7x.dds"},
	{0xB59702E577C3A51CULL, "art\\vfx\\combat\\vfx_r_brd_csong_act_c.impa.evt"},
	{0xB597222650A2C1FDULL, "lt_0_lowtown_docks_19_0.anb"},
	{0xB5974A348DA9AD9CULL, "6075093_m.fxe"},
	{0xB5977828CD0C6D13ULL, "6016317_f.fxe"},
	{0xB597C829FBBF576FULL, "w_encounter3_70_0.mao"},
	{0xB597D429FBBF6B0CULL, "w_encounter3_70_0.mmh"},
	{0xB597DA29FBBF75FEULL, "w_encounter3_70_0.msh"},
	{0xB597F5A07378E3C5ULL, "6113958_m.fxe"},
	{0xB5994248DE1F9FCFULL, "gen00fl_dog6165743.cl"},
	{0xB59949D84F66514FULL, "art\\characters\\weapons\\mauls\\w_mal_ml102.mao"},
	{0xB5994D3713A999FEULL, "382337765.ogg"},
	{0xB5995C236A4397E8ULL, "vxca_cavern1_haunted_cjq_0_0.vlm"},
	{0xB599D5B40E32D07BULL, "566001760.ogg"},
	{0xB59A06C80FA3DA77ULL, "257210947.ogg"},
	{0xB59A633CD34DC783ULL, "vxht_0_hightown_a2_baw_0_0.vlm"},
	{0xB59B43DB70F8852DULL, "hlac_lt_0_lowtown_docks_a3_-3_0x.dds"},
	{0xB59B906BE8F2C361ULL, "877391075.ogg"},
	{0xB59BC4DF8CC4F73BULL, "vxca_caverns1_iap_0_0.vlm"},
	{0xB59BF9A68DB90EBAULL, "vxca_cavern1_haunted_inu_0_0.vlm"},
	{0xB59C0CF62D633A07ULL, "5290892.ogg"},
	{0xB59CAE9544C7D9E0ULL, "6034281_m.fxe"},
	{0xB59CF5FCBECAB35FULL, "hlac_lt_0_lowtown_a2_n_-1_3x.dds"},
	{0xB59D6458FFF6DC6CULL, "6053004_m.fxe"},
	{0xB59D6B06EC51CF45ULL, "hightown_fem_noble1_sta6169250.cl"},
	{0xB59DAB15BBB282A1ULL, "1032213496.ogg"},
	{0xB59DB0050C6D6012ULL, "vxw_woundedcoast_1d2_0_0.vlm"},
	{0xB59DD65234D4EFD1ULL, "6108639_m.fxe"},
	{0xB59E06F049436573ULL, "152487842.ogg"},
	{0xB59E3DED9A920B17ULL, "vxlt_gallowsprison_n_6cm_0_0.vlm"},
	{0xB59E71AF99ED98CAULL, "6054054_m.fxe"},
	{0xB59E83B0564A0A64ULL, "follower_barks6153830.cl"},
	{0xB59FC1AB15CF1C39ULL, "vxdr_primevaltomb_8ol_0_0.vlm"},
	{0xB59FC22CC890AC6EULL, "llac_w_bonepitmining_a2_131076_0.dds"},
	{0xB59FD287CAFD3CADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantmed.mmh"},
	{0xB59FD687CAFD43FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantmed.mao"},
	{0xB59FE487CAFD5B33ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantmed.msh"},
	{0xB5A05812E1C191DDULL, "hlac_dr_0_deeproads_5842_4x.dds"},
	{0xB5A09AC721A521B1ULL, "164968959.ogg"},
	{0xB5A109BCF6FD4D60ULL, "llac_dr_primevaltomb_9621_6.dds"},
	{0xB5A19D6EB9FE97E7ULL, "6189663_m.fxe"},
	{0xB5A1B4E2761725CDULL, "gen00fl_isabela_init6076494.cl"},
	{0xB5A1EF8B7CCACF53ULL, "hlac_w_woundedcoast_327682_0x.dds"},
	{0xB5A1FB759EDA8DA1ULL, "character_character_p176.psh.fxo.dx11.assembly"},
	{0xB5A2087D0DE48FF8ULL, "vxdr_0_deeproads_boq_0_0.vlm"},
	{0xB5A2301149B03190ULL, "dae000ip_crowd_mf4_talk4.utp"},
	{0xB5A23DB4283BE9AEULL, "hlac_lt_0_lowtown_a2_n_c_131074_0z.dds"},
	{0xB5A27A06E0A558BEULL, "vxht_chantry_31w_0_0.vlm"},
	{0xB5A29CFCEAB99308ULL, "ts_apostates_prog_junk1.gda"},
	{0xB5A2D46999220099ULL, "6078709_m.fxe"},
	{0xB5A33D1EFFCB8CF1ULL, "hlac_w_bonepit_65541_0z.dds"},
	{0xB5A39E453E047396ULL, "390867320.ogg"},
	{0xB5A3F5E746022ABBULL, "w_bonepitmining_a2_182_0.msh"},
	{0xB5A3F8C30DA84664ULL, "rdr111dg_assistant6016642.cl"},
	{0xB5A407E746024906ULL, "w_bonepitmining_a2_182_0.mao"},
	{0xB5A40AE577CE3D35ULL, "art\\vfx\\combat\\vfx_r_brd_csong_act_c.impa.ani"},
	{0xB5A413E746025DA5ULL, "w_bonepitmining_a2_182_0.mmh"},
	{0xB5A4F0D1B31174B7ULL, "ht_keep_n_3741.anb"},
	{0xB5A51901F18A9D5CULL, "164566040.ogg"},
	{0xB5A5C33831D0D776ULL, "269462146.ogg"},
	{0xB5A63DA9D5C5238AULL, "vrc221_varric6043957.cl"},
	{0xB5A68B71007D5775ULL, "773503590.ogg"},
	{0xB5A6B53991355D0CULL, "hlac_lt_darktown_-1_69y.dds"},
	{0xB5A6EC70D574DE71ULL, "6116687_m.fxe"},
	{0xB5A6F174E2A690CEULL, "6065256_m.fxe"},
	{0xB5A75375D5FF8093ULL, "gen00fl_anders6096611.cl"},
	{0xB5A811197553C66EULL, "mag101_grace6034212.cl"},
	{0xB5A85863B8CFEADDULL, "489300276.ogg"},
	{0xB5A8AB51F562D61CULL, "vxlt_gallowscourtyard_725_0_0.vlm"},
	{0xB5A8D139756C1A07ULL, "288864465.ogg"},
	{0xB5A9416284CFE493ULL, "vxlt_0_lowtown_a2_n_c_h1t_0_0.vlm"},
	{0xB5A96280DC9F777CULL, "lt_0_lowtown_a2_n_25_0_ll.mmh"},
	{0xB5A9F1C42A1528FFULL, "w_sundermount_222_0.msh"},
	{0xB5A9F7C42A1532A1ULL, "w_sundermount_222_0.mmh"},
	{0xB5AA03C42A154742ULL, "w_sundermount_222_0.mao"},
	{0xB5AA5E0EFDD1C062ULL, "6196702_f.fxe"},
	{0xB5AA5EA9B4C8EB65ULL, "402511671.ogg"},
	{0xB5AAD175A3AF0CB4ULL, "106439528.ogg"},
	{0xB5AB60C318278BE6ULL, "vxw_bonepit_bp_0_0.vlm"},
	{0xB5AB639743C2F958ULL, "pro000_refugee_female6052087.cl"},
	{0xB5ABCBF3E49C85BAULL, "llac_w_encounter1_131071_0.dds"},
	{0xB5AC21E1FE637892ULL, "vxw_woundedcoast_1o3_0_0.vlm"},
	{0xB5AC52D1CDB3C3D2ULL, "6075150_m.fxe"},
	{0xB5ACAE656C92DE6DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masd_0s.dds"},
	{0xB5ACAFCA8C882397ULL, "vxlt_gallowsprison_n_6ll_0_0.vlm"},
	{0xB5ACB3ABCFBE220BULL, "mer220_hubert6088150.cl"},
	{0xB5AD3D20778DD7CBULL, "578401196.ogg"},
	{0xB5AD566D17CE4963ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_2y.dds"},
	{0xB5AD7B9E7ED0EA6CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb05a.mmh"},
	{0xB5AD819E7ED0F45EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb05a.msh"},
	{0xB5ADBEE685E5F201ULL, "bink_bink_p0.vsh.fxo"},
	{0xB5ADCEBC6F73A997ULL, "lt_0_lowtown_docks_115_0.msh"},
	{0xB5ADDB65C99649BAULL, "136261528.ogg"},
	{0xB5ADE0BC6F73C8FAULL, "lt_0_lowtown_docks_115_0.mao"},
	{0xB5ADE4BC6F73CF09ULL, "lt_0_lowtown_docks_115_0.mmh"},
	{0xB5ADF96F6EBF5BCDULL, "vxlt_gallowstemplar_f_55r_0_0.vlm"},
	{0xB5AE048C9D1FC3C0ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintinga_d.dds"},
	{0xB5AE6DE7FE4D0C33ULL, "378308431.ogg"},
	{0xB5AE9DAD1511DADAULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01a.phy"},
	{0xB5AEAF932CBBC072ULL, "vxht_0_hightown_a2_8pg_0_0.vlm"},
	{0xB5AF64BB41AF62E8ULL, "vxht_0_hightown_a2_cbd_0_0.vlm"},
	{0xB5AFB2B6501EF20CULL, "gen00fl_bethany6102760.cl"},
	{0xB5B0B72F321783FCULL, "4572.ogg"},
	{0xB5B12295ED5C7249ULL, "592912827.ogg"},
	{0xB5B14B7B31431A97ULL, "6116551_m.fxe"},
	{0xB5B15CDB4E4EB2EEULL, "6027746_m.fxe"},
	{0xB5B270D08689CF21ULL, "hlac_w_woundedcoast_-65536_0x.dds"},
	{0xB5B2895A217F0189ULL, "zz_dre_debug6168215.cl"},
	{0xB5B32C2DC09D3B6DULL, "ave100_donnic6105998.cl"},
	{0xB5B39849C1B7741EULL, "hlac_lt_0_lowtown_a2_n_c_196607_1x.dds"},
	{0xB5B3BFD6CE2F488AULL, "18031384.ogg"},
	{0xB5B51FBABEF3F65AULL, "6169981_m.fxe"},
	{0xB5B60BC24977D29EULL, "gen_im_trs_crc_junk3_04.uti"},
	{0xB5B62D052AADE049ULL, "vxw_bonepit_1nh_0_0.vlm"},
	{0xB5B64541F21B4025ULL, "921893087.ogg"},
	{0xB5B660932C8168BBULL, "bl_cl_still.mao"},
	{0xB5B721A20E4FB93CULL, "6162766_m.fxe"},
	{0xB5B734359014E488ULL, "6055572_m.fxe"},
	{0xB5B7347E35C3E54FULL, "and100_anders6095942.cl"},
	{0xB5B79F2E4A6F4ECDULL, "vxht_bartrand_n_3yt_0_0.vlm"},
	{0xB5B7C7153015572EULL, "cod_hst_parvollen.utp"},
	{0xB5B8447F3F5C7824ULL, "vxht_mansion2b_324_0_0.vlm"},
	{0xB5B866D19A9F113DULL, "6044621_m.fxe"},
	{0xB5B921C04D645726ULL, "vxlt_warehouse_af_0_0.vlm"},
	{0xB5B935028C26C319ULL, "llac_ca_caverns1_mines_24424_9.dds"},
	{0xB5B938B42137E665ULL, "106399470.ogg"},
	{0xB5B950BA5D4AC3EBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_diagonal.phy"},
	{0xB5B9D2A00A1DA466ULL, "vxw_bonepit_mq_0_0.vlm"},
	{0xB5B9D3CDEFFA23C4ULL, "follower_banter6151329.cl"},
	{0xB5B9F20B17239EAFULL, "gen_im_trs_rdr_arc_lbw_r2.uti"},
	{0xB5BA32E4DEB1E637ULL, "171056287.ogg"},
	{0xB5BA705A2F992C42ULL, "art\\characters\\weapons\\maces\\w_mce_dsmca.msh"},
	{0xB5BA725A2F992FF8ULL, "art\\characters\\weapons\\maces\\w_mce_dsmca.mmh"},
	{0xB5BA7E5A2F99439BULL, "art\\characters\\weapons\\maces\\w_mce_dsmca.mao"},
	{0xB5BA862F9475EE13ULL, "4162413.ogg"},
	{0xB5BAF72C357C112AULL, "6125934_m.fxe"},
	{0xB5BB651306637418ULL, "gen00fl_anders6116828.cl"},
	{0xB5BB8E825A871857ULL, "914248209.ogg"},
	{0xB5BB9DE62606FABFULL, "one100dg_lia6055167.cl"},
	{0xB5BBC30CBF483794ULL, "6037136_m.fxe"},
	{0xB5BBD99367BCED51ULL, "vxw_sundermount_372_0_0.vlm"},
	{0xB5BBDEA7ABA2ECB4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.ani"},
	{0xB5BC13480347DFA0ULL, "dae100_bran6118342.cl"},
	{0xB5BC1C66D0F1B69BULL, "hlac_lt_hangedman_-1_27y.dds"},
	{0xB5BCB77C6D8ABF89ULL, "nat380_sibling6067993.cl"},
	{0xB5BCEF5EAE0AE12EULL, "54178821.ogg"},
	{0xB5BD11ECD5F838EAULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiremed.dura.gad"},
	{0xB5BD838F50CA73FBULL, "lt_0_lowtown_25_0.rml"},
	{0xB5BDE9EF1EC022C7ULL, "hlac_ca_caverns1_mines_24425_9y.dds"},
	{0xB5BF646C6286BB07ULL, "dae100_viscount6103305.cl"},
	{0xB5BF6670D1351F47ULL, "mag100_karras6034506.cl"},
	{0xB5BF76D8908232D0ULL, "439161082.ogg"},
	{0xB5BFAA76BA77281AULL, "gen00fl_isabela6201241.cl"},
	{0xB5C0B84BFBB9E478ULL, "t3_hey_te3.tnt"},
	{0xB5C0BD9F39EB6AECULL, "6143051_m.fxe"},
	{0xB5C0D374BAE78E09ULL, "dot.mao"},
	{0xB5C0DC1AA9CD2042ULL, "vxht_0_hightown_a2_c62_0_0.vlm"},
	{0xB5C0E42E700C0E08ULL, "6037110_m.fxe"},
	{0xB5C12F5F132AC638ULL, "65860872.ogg"},
	{0xB5C1AE44C78914E2ULL, "678984994.ogg"},
	{0xB5C24FD40107031CULL, "6170101_m.fxe"},
	{0xB5C264FDE7021C9CULL, "vxlt_gallowstemplar_n_6q8_0_0.vlm"},
	{0xB5C26E1B8E1F98E2ULL, "384213814.ogg"},
	{0xB5C28FA578AA5804ULL, "llac_lt_0_lowtown_docks_a3_-131074_4.dds"},
	{0xB5C2DBA7ABA90616ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.gad"},
	{0xB5C2E1F79DF4FA6BULL, "151112530.ogg"},
	{0xB5C2E4DF564D9F4BULL, "vxlt_gallowstemplar_f_6sy_0_0.vlm"},
	{0xB5C2E66420838100ULL, "6104821_m.fxe"},
	{0xB5C2FFD0681B4AEEULL, "w_bonepitmining_a2_200_0.msh"},
	{0xB5C30DD0681B623FULL, "w_bonepitmining_a2_200_0.mao"},
	{0xB5C319D0681B76DCULL, "w_bonepitmining_a2_200_0.mmh"},
	{0xB5C325C4294354FCULL, "terrain_blend4.psh.fxo"},
	{0xB5C340C602CF4FD3ULL, "gen00fl_anders6103900.cl"},
	{0xB5C34C56EF6F5218ULL, "llac_lt_blackemporium_1055_17.dds"},
	{0xB5C3C1797C67DABDULL, "676265717.ogg"},
	{0xB5C3CD2733484CE2ULL, "3344215.ogg"},
	{0xB5C3D10E393E8ACFULL, "1020909066.ogg"},
	{0xB5C3EAF3534A6099ULL, "6161339_m.fxe"},
	{0xB5C3F4ECD5FE2581ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiremed.dura.evt"},
	{0xB5C40943BC4BE00CULL, "zz_t_isaac_slash.cut"},
	{0xB5C4100685031815ULL, "6163885_m.fxe"},
	{0xB5C42C588CADEEE8ULL, "dae150_chandan.cnv"},
	{0xB5C4795A2FA18F83ULL, "art\\characters\\weapons\\maces\\w_mce_dsmca.phy"},
	{0xB5C52D368A0B887AULL, "vxca_caverns1_mines_ck0_0_0.vlm"},
	{0xB5C6311D27CB2662ULL, "354035648.ogg"},
	{0xB5C638249046BA45ULL, "6072567_m.fxe"},
	{0xB5C6B0CC2C957075ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firecoal_d.dds"},
	{0xB5C6CBCC3B9839C5ULL, "gen00fl_merrill6119653.cl"},
	{0xB5C739CA07289C87ULL, "155521419.ogg"},
	{0xB5C74162C51977C5ULL, "733017417.ogg"},
	{0xB5C74E588CB05E50ULL, "dae150_chandan.bnk"},
	{0xB5C753835986B180ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m_tube_tool_1.mao"},
	{0xB5C778421E19D5EDULL, "124583329.ogg"},
	{0xB5C7D8FFEB09D87CULL, "distortion_nh.dds"},
	{0xB5C7F8A1883E71E5ULL, "6151233_m.fxe"},
	{0xB5C817ED57897A62ULL, "hlac_ht_chantry_n_2534_32y.dds"},
	{0xB5C8B47F5018B4D6ULL, "vxlt_gallowscourtyard_75j_0_0.vlm"},
	{0xB5C8C37F5C4EF71FULL, "vxlt_0_lowtown_a2_n_eff_0_0.vlm"},
	{0xB5C8CBA7FCABADC6ULL, "hlac_w_bonepitdestroyed_a3_5_0y.dds"},
	{0xB5C8E70B7BB13067ULL, "vxlt_0_lowtown_a2_n_c_exr_0_0.vlm"},
	{0xB5C9938470745497ULL, "vxlt_gallowsprison_n_6rn_0_0.vlm"},
	{0xB5C9B529686BEFCEULL, "isa211_isabela6069628.cl"},
	{0xB5C9C1B496C854E0ULL, "6015635_m.fxe"},
	{0xB5C9C36495180221ULL, "6103461_m.fxe"},
	{0xB5C9F6A7ABAF5375ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_gesture_3.evt"},
	{0xB5CA1AEC240719CBULL, "222552184.ogg"},
	{0xB5CA36EF3E1A04A9ULL, "101535864.ogg"},
	{0xB5CA428DF86E90A7ULL, "and111_refugee_barks6111536.cl"},
	{0xB5CA4B7059B305D1ULL, "vxw_sundermount_321_0_0.vlm"},
	{0xB5CB686B61EB4307ULL, "6135641_f.fxe"},
	{0xB5CBA478E85A91BFULL, "hlac_dr_slavers_3214_5x.dds"},
	{0xB5CBEDDE4B5C6510ULL, "6034388_m.fxe"},
	{0xB5CC05CEE640D51CULL, "vxht_0_hightown_a3_bum_0_0.vlm"},
	{0xB5CC2058D97F4DE6ULL, "lgt210_magu_selby6142028.cl"},
	{0xB5CC65071A6D6962ULL, "lt_0_lowtown_docks_n_214_0.msh"},
	{0xB5CC67071A6D6C18ULL, "lt_0_lowtown_docks_n_214_0.mmh"},
	{0xB5CC73071A6D80BBULL, "lt_0_lowtown_docks_n_214_0.mao"},
	{0xB5CCE32A2A614A15ULL, "630606284.ogg"},
	{0xB5CD34F2DE78403DULL, "zz_followers_banter6157890.cl"},
	{0xB5CD3B93B4BB077EULL, "295256865.ogg"},
	{0xB5CD843D9DDDE569ULL, "fire_ball_blue.mao"},
	{0xB5CD87B131C0F94EULL, "vxlt_0_lowtown_a3_n_c_glh_0_0.vlm"},
	{0xB5CD9B6318153177ULL, "vxca_caverns1_iad_0_0.vlm"},
	{0xB5CE487D6A79358BULL, "6110513_m.fxe"},
	{0xB5CF25327B72FC39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\look_left_turn.ani"},
	{0xB5CF2B425D0A82D6ULL, "vxht_0_hightown_a3_bx0_0_0.vlm"},
	{0xB5CF560B62022688ULL, "6030010_m.fxe"},
	{0xB5CF6DE77EC074BDULL, "gen00fl_aveline6122157.cl"},
	{0xB5CF882D05073592ULL, "1047609625.ogg"},
	{0xB5CFDEF24216BF40ULL, "dae350ip_at_scaverns_b.utp"},
	{0xB5D0232650D288F5ULL, "lt_0_lowtown_docks_19_0.rml"},
	{0xB5D0439955611CFEULL, "6088456_f.fxe"},
	{0xB5D087260EB00DAEULL, "art\\characters\\heraldry\\hld_undercity.dds"},
	{0xB5D0EDD1B336AACBULL, "ht_keep_n_3741.rml"},
	{0xB5D17CC52897D949ULL, "hlac_w_encounter2_camp_-65535_0x.dds"},
	{0xB5D1CD37806B15A1ULL, "1034459973.ogg"},
	{0xB5D1ECECD60A55D0ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfiremed.dura.ani"},
	{0xB5D1EE3E6CACDCC6ULL, "170213347.ogg"},
	{0xB5D1F2D357D31CE9ULL, "vxdr_primevaltomb_4h5_0_0.vlm"},
	{0xB5D207D8C1EC92F9ULL, "ts_kirkwall_wpn_table.gda"},
	{0xB5D225F98825BF3DULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_bdy_hdra_0d.dds"},
	{0xB5D248E2163A3951ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslba_0s.dds"},
	{0xB5D25CD045407919ULL, "6166239_m.fxe"},
	{0xB5D282674AC3D8B7ULL, "art\\characters\\weapons\\longswords\\w_lsw_024a.mmh"},
	{0xB5D286674AC3DFC4ULL, "art\\characters\\weapons\\longswords\\w_lsw_024a.mao"},
	{0xB5D298674AC3FE29ULL, "art\\characters\\weapons\\longswords\\w_lsw_024a.msh"},
	{0xB5D32C866B50980EULL, "6016940_m.fxe"},
	{0xB5D34D81096E7082ULL, "vxw_encounter3_fb_0_1.vlm"},
	{0xB5D350D9E33296A8ULL, "1013524930.ogg"},
	{0xB5D41EAF27E20183ULL, "w_bonepitdestroyed_a3_9_0_lc.mmh"},
	{0xB5D46523EB820B17ULL, "vxht_mansion_n_4vo_0_0.vlm"},
	{0xB5D47F2CEBDB3D14ULL, "vxlt_warehouse_n_9u_0_0.vlm"},
	{0xB5D4860C6B8C7E73ULL, "6152227_m.fxe"},
	{0xB5D4EBD9E3C34A4AULL, "vxlt_hangedman_16j_0_0.vlm"},
	{0xB5D53F901AF10E51ULL, "6156940_m.fxe"},
	{0xB5D5738A2D7644ACULL, "dae000cr_grand_cleric6187942.cl"},
	{0xB5D600327B78DC7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\look_left_turn.gad"},
	{0xB5D651B0A38EE38AULL, "fx_lowerlag.dds"},
	{0xB5D67E033A96B832ULL, "6042003_f.fxe"},
	{0xB5D6C83E259E153DULL, "363764822.ogg"},
	{0xB5D73A10E85EF96CULL, "dae000ip_crowd_ff3_talk8.utp"},
	{0xB5D7450547F8DCBEULL, "6090326_m.fxe"},
	{0xB5D77F946CC659C9ULL, "vxca_caverns1_egl_0_0.vlm"},
	{0xB5D79BD102DC0540ULL, "mag210_arianni6021228.cl"},
	{0xB5D7AC78BEC7F405ULL, "40097739.ogg"},
	{0xB5D806AF9F04D1D9ULL, "vxlt_0_lowtown_a2_n_hb5_0_0.vlm"},
	{0xB5D8E274BAFC39B7ULL, "dot.dds"},
	{0xB5D8F1991E85B00EULL, "6050450_m.fxe"},
	{0xB5D95C95DA5D6361ULL, "clf260_harley6069571.cl"},
	{0xB5D986E7F799CB7EULL, "vxht_bartrand_n_5sg_0_0.vlm"},
	{0xB5D9B37DDF6ECF74ULL, "gen00fl_aveline6116724.cl"},
	{0xB5D9EBE172BBACDBULL, "6004379_m.fxe"},
	{0xB5D9EC37C3F01E61ULL, "dae000ip_crowd_mf4_talk14.utp"},
	{0xB5D9FABACDE2B274ULL, "gen00fl_fenris6126681.cl"},
	{0xB5DA4729C8FB448FULL, "595905066.ogg"},
	{0xB5DA4C6EB7B7F384ULL, "85875414.ogg"},
	{0xB5DABE2AD0AA04F3ULL, "vxlt_0_lowtown_a2_n_c_f2g_0_0.vlm"},
	{0xB5DB04A2CE262651ULL, "6110630_m.fxe"},
	{0xB5DB19282B841937ULL, "ttx_steps_curve_a_1.mmh"},
	{0xB5DB2F282B843EA9ULL, "ttx_steps_curve_a_1.msh"},
	{0xB5DBDB3879D23C52ULL, "6163200_m.fxe"},
	{0xB5DBEFA25E34466CULL, "w_bonepit_240_0.msh"},
	{0xB5DBF5A25E34505EULL, "w_bonepit_240_0.mmh"},
	{0xB5DBF98E23B711CEULL, "vxlt_0_lowtown_a3_n_c_gci_0_0.vlm"},
	{0xB5DC01A25E3464F5ULL, "w_bonepit_240_0.mao"},
	{0xB5DC1445C68011ECULL, "mag311_emile_de_launcet6185107.cl"},
	{0xB5DC1D327B7D7960ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\look_left_turn.evt"},
	{0xB5DC6A8EB70145E6ULL, "framebuffereffects_smallblur_p0.psh.fxo"},
	{0xB5DC76042BBD261EULL, "vxht_chantry_n_36i_0_0.vlm"},
	{0xB5DCE56FB2BBB739ULL, "319628075.ogg"},
	{0xB5DDBD2667D8D2BBULL, "522628455.ogg"},
	{0xB5DE05D42D16F636ULL, "mag330_walter6044821.cl"},
	{0xB5DE4A0FB83003ABULL, "sw_imp_gr.mao"},
	{0xB5DE5206688CA9E7ULL, "gen_im_npc_elf_arm_cht_lgtb.uti"},
	{0xB5DE54D23DFAABE0ULL, "matt_warrior.utc"},
	{0xB5DEDFFDE44D56C0ULL, "dae300_cullen6135976.cl"},
	{0xB5DF500328005B94ULL, "6041533_m.fxe"},
	{0xB5DF63AE3BAC1DD3ULL, "6153505_m.fxe"},
	{0xB5DF995F97FFEC38ULL, "305438113.ogg"},
	{0xB5DF9C0F02C3B7FFULL, "mcr371_scared_mage_f6087716.cl"},
	{0xB5DF9ED10B3BAA83ULL, "vxlt_0_lowtown_a2_hfx_0_0.vlm"},
	{0xB5E020F1D5A69DA6ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar2_.ani"},
	{0xB5E071C9646BE187ULL, "llac_ht_mansion2b_818_45.dds"},
	{0xB5E0738C6AA0D560ULL, "hlac_w_encounter2_camp_-131071_0z.dds"},
	{0xB5E0CE491F43A4CBULL, "6164183_m.fxe"},
	{0xB5E0F9520D9C2B21ULL, "vxca_cavern1_haunted_j4w_0_0.vlm"},
	{0xB5E1128FEA38C502ULL, "792165929.ogg"},
	{0xB5E196EA6E2C9C93ULL, "hm_qsw_thug_leader.mor"},
	{0xB5E1C09D2444107AULL, "6156456_f.fxe"},
	{0xB5E1E8749F9CA217ULL, "qcr270_fake_meredith6066348.cl"},
	{0xB5E294E0DD916B41ULL, "6052558_m.fxe"},
	{0xB5E31DBBE7723963ULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_p_p.impa.ani"},
	{0xB5E36A241AC3E5D5ULL, "671078325.ogg"},
	{0xB5E3703DA4744253ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v4.ani"},
	{0xB5E45B2163F891A3ULL, "76672409.ogg"},
	{0xB5E45DFC8F77CE73ULL, "w_blightlands_188_0.phy"},
	{0xB5E477572346A0A6ULL, "vxlt_warehouse_pi_0_0.vlm"},
	{0xB5E4933D9DF0DD17ULL, "fire_ball_blue.dds"},
	{0xB5E499FAE757B824ULL, "hlac_w_bonepitdestroyed_a3_6_0z.dds"},
	{0xB5E4A74AE7684640ULL, "vxca_cavern1_haunted_gk6_0_0.vlm"},
	{0xB5E50E60CAE941F3ULL, "vxlt_gamlenshouse_n_87_0_0.vlm"},
	{0xB5E510EF8F9E54F1ULL, "vxw_bonepitdestroyed_a3_1rt_0_0.vlm"},
	{0xB5E52CDB9498E669ULL, "gen00fl_player6207772.cl"},
	{0xB5E582B4AB8FB12AULL, "422399034.ogg"},
	{0xB5E5C619822ED512ULL, "zz_approval_debug6003667.cl"},
	{0xB5E5DF178D253BBAULL, "vxlt_0_lowtown_a2_by2_0_0.vlm"},
	{0xB5E60E9A0D238BF1ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.dura.ani"},
	{0xB5E61C7C4E056CCAULL, "vxht_mansion_n_54e_0_0.vlm"},
	{0xB5E64D0A0057BE71ULL, "19738472.ogg"},
	{0xB5E68732A928CA67ULL, "vxlt_warehouse_1ef_0_0.vlm"},
	{0xB5E70C3372CD41F7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basic.msh"},
	{0xB5E71BB27EF9D221ULL, "hlac_ca_caverns1_mines_24424_12z.dds"},
	{0xB5E71E3372CD605AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basic.mao"},
	{0xB5E7223372CD6769ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wo_basic.mmh"},
	{0xB5E73C36B29783FAULL, "802709891.ogg"},
	{0xB5E750F91B7FF83DULL, "305314982.ogg"},
	{0xB5E7807D3AA56DAAULL, "dae210ip_store4.utp"},
	{0xB5E7AF62C141D31EULL, "518064587.ogg"},
	{0xB5E7B83612A578D8ULL, "6141109_m.fxe"},
	{0xB5E7F8D0683A5AA7ULL, "w_bonepitmining_a2_200_0.phy"},
	{0xB5E908B70512E14AULL, "zz_qcr_debug6054925.cl"},
	{0xB5E9ABBA00FD627CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon2_sa.dds"},
	{0xB5EA7F1CFC016277ULL, "vxca_caverns1_iof_0_0.vlm"},
	{0xB5EAB6D0CF553B4CULL, "zz_spy130_darktown_res_26055330.cl"},
	{0xB5EAE7E11DB33E9EULL, "vxht_chantry_n_33j_0_0.vlm"},
	{0xB5EAECE7463E1E52ULL, "w_bonepitmining_a2_182_0.phy"},
	{0xB5EAF0670BDAE533ULL, "6167696_m.fxe"},
	{0xB5EB5C69BDB0D8BBULL, "vxlt_gallowscourtyard_61z_0_0.vlm"},
	{0xB5EC4F328E47BFB4ULL, "gen00fl_anders6125606.cl"},
	{0xB5ECD9F178A15C9CULL, "w_blightlands_231_0.mao"},
	{0xB5ECE5F178A1713FULL, "w_blightlands_231_0.mmh"},
	{0xB5ECEBF178A17BE1ULL, "w_blightlands_231_0.msh"},
	{0xB5ECECD01B3725F4ULL, "vxw_sundermount_2ex_0_0.vlm"},
	{0xB5ED199A0D29BC93ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.dura.gad"},
	{0xB5ED27DDDF231E5FULL, "6155925_m.fxe"},
	{0xB5ED2DB5F444F76DULL, "667924058.ogg"},
	{0xB5EDA81A53EDEABBULL, "art\\vfx\\combat\\vfx_glb_levelup_c.impa.ani"},
	{0xB5EDDA4FE07F19B0ULL, "6039743_m.fxe"},
	{0xB5EE3E02E3C57488ULL, "vxca_caverns2_a9x_0_0.vlm"},
	{0xB5EF197A3D691F1DULL, "91485125.ogg"},
	{0xB5EF39008BA6DC78ULL, "zz_fol_debug6131618.cl"},
	{0xB5EF840DA9ECA54DULL, "lt_0_lowtown_a3_n_c_26_0_lc.mmh"},
	{0xB5EF89DCF240AF19ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_doorcrest_n.dds"},
	{0xB5EF91292C211C07ULL, "lt_shop_mmi.gff"},
	{0xB5EFE58D8D0A6A4FULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_bellyspew.gad"},
	{0xB5EFF63CEBA95137ULL, "vxca_cave_2d0_0_0.vlm"},
	{0xB5F035BBE77CEC5AULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_p_p.impa.evt"},
	{0xB5F0683DA47EBE8AULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v4.evt"},
	{0xB5F0D1353BC2D7C1ULL, "351329862.ogg"},
	{0xB5F150CA2BBADA7EULL, "ser210_witness6168103.cl"},
	{0xB5F238E508DADB6DULL, "hlac_lt_0_lowtown_docks_a3_65533_2z.dds"},
	{0xB5F2DAFC7F4A556AULL, "art\\vfx\\combat\\vfx_f_and_fade_c.cess.gad"},
	{0xB5F2E934153CB926ULL, "vxlt_warehouse_uj_0_0.vlm"},
	{0xB5F321D14DC56AA7ULL, "l_dae310ar_docks.wbd"},
	{0xB5F3361A79E63038ULL, "vxht_bartrand_n_5q7_0_0.vlm"},
	{0xB5F3631BC6A00286ULL, "vxht_bartrand_n_5qy_0_0.vlm"},
	{0xB5F409F1D5B71FA0ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar2_.gad"},
	{0xB5F4169A0D2FD6F8ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.dura.evt"},
	{0xB5F419163231409FULL, "vxlt_0_lowtown_a2_n_eqc_0_0.vlm"},
	{0xB5F58F0DB32474BAULL, "vxlt_gallowscourtyard_7da_0_0.vlm"},
	{0xB5F5C4A6DFCE5102ULL, "6101077_m.fxe"},
	{0xB5F617412B85AAC4ULL, "1040268478.ogg"},
	{0xB5F6745F9E9A71FFULL, "6169997_m.fxe"},
	{0xB5F68EFFBF3CA851ULL, "6084737_m.fxe"},
	{0xB5F69595BEA8D6A7ULL, "6054077_m.fxe"},
	{0xB5F6A0D3A13D649AULL, "6116198_m.fxe"},
	{0xB5F6A28CDC32F264ULL, "zz_dre_debug6168278.cl"},
	{0xB5F6C077EE032F7DULL, "dae000_bartrand6027517.cl"},
	{0xB5F6FA8D8D10ACFCULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_bellyspew.evt"},
	{0xB5F6FD26BD85EBFEULL, "503851837.ogg"},
	{0xB5F71DFD27D5DDC1ULL, "972595561.ogg"},
	{0xB5F734BBE783094DULL, "art\\vfx\\combat\\vfx_w_off_stun_chc_p_p.impa.gad"},
	{0xB5F7473DA484A5FDULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v4.gad"},
	{0xB5F76D44E84CCE91ULL, "383445868.ogg"},
	{0xB5F7A54988AC9D07ULL, "297913222.ogg"},
	{0xB5F7BD8E0372DD8EULL, "hlac_ht_0_hightown_n_196607_0y.dds"},
	{0xB5F86596FD43FA79ULL, "gen_im_acc_blt_act3_07.uti"},
	{0xB5F8CC632C49FDAFULL, "plt_rumor_wayward_son.plo"},
	{0xB5F8E07A700CC247ULL, "llac_lt_backalley_n_0_4.dds"},
	{0xB5F95056FDBEFF7DULL, "851345694.ogg"},
	{0xB5F9AA05D5D0B983ULL, "zz_qun_debug6028264.cl"},
	{0xB5F9BDFC7F504201ULL, "art\\vfx\\combat\\vfx_f_and_fade_c.cess.evt"},
	{0xB5FA2F853EA85579ULL, "6154762_m.fxe"},
	{0xB5FA96FD7716B2A1ULL, "6166990_m.fxe"},
	{0xB5FAB01A53F882B2ULL, "art\\vfx\\combat\\vfx_glb_levelup_c.impa.evt"},
	{0xB5FADDDC050D844FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile_6.pwk"},
	{0xB5FB1ECCE0EBB36BULL, "6036792_m.fxe"},
	{0xB5FB632F6AFF3E1CULL, "hlac_lt_0_lowtown_a2_131074_0y.dds"},
	{0xB5FC0A4557C85EA8ULL, "vxht_brothel_n_6e8_0_0.vlm"},
	{0xB5FC29BF1337D3CAULL, "390835882.ogg"},
	{0xB5FC47803A4A173CULL, "gen00fl_carver6102951.cl"},
	{0xB5FC828AEF27DB83ULL, "vxlt_0_lowtown_a2_hhz_0_0.vlm"},
	{0xB5FD742BE211C2A5ULL, "233243815.ogg"},
	{0xB5FD867A6D9BE266ULL, "hlac_lt_0_lowtown_a2_n_c_-131071_0z.dds"},
	{0xB5FD87E1344AC770ULL, "mrl150_merrill6093968.cl"},
	{0xB5FDA46305A49341ULL, "vxht_mansion2_22x_0_0.vlm"},
	{0xB5FDA67A03C8D845ULL, "187736098.ogg"},
	{0xB5FDAE3E74DA5665ULL, "716130920.ogg"},
	{0xB5FDDA4407A459BEULL, "6135944_m.fxe"},
	{0xB5FDEEAB0B16C02FULL, "t3_hey_hz1.tnt"},
	{0xB5FE079682E8C338ULL, "mag131_taraneh6088686.cl"},
	{0xB5FE61045881AAACULL, "art\\vfx\\cinematic\\vfx_mage_punch_c.impa.ani"},
	{0xB5FE868F51020FC7ULL, "lt_0_lowtown_25_0.anb"},
	{0xB5FEB9997E974558ULL, "99820590.ogg"},
	{0xB5FEDB9A255E2FEDULL, "lt_0_lowtown_docks_a3_n_218_0.msh"},
	{0xB5FEDD14D489A4FCULL, "vxlt_gallowstemplar_n_6s2_0_0.vlm"},
	{0xB5FEE99A255E4720ULL, "lt_0_lowtown_docks_a3_n_218_0.mao"},
	{0xB5FEED9A255E4E73ULL, "lt_0_lowtown_docks_a3_n_218_0.mmh"},
	{0xB5FF001C44663FF0ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_4y.dds"},
	{0xB5FF1ECA99097E1EULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.cess.ani"},
	{0xB5FF78393BA90258ULL, "6108928_m.fxe"},
	{0xB5FFC11C3D652D49ULL, "302163233.ogg"},
	{0xB5FFD695E6527D17ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_whiteflower.mmh"},
	{0xB5FFDA70F13544FEULL, "ht_mansion_n_3366.rml"},
	{0xB5FFEC95E652A289ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_whiteflower.msh"},
	{0xB600288BF16D798BULL, "dae300cr_bran.utc"},
	{0xB6003B16C88D0A66ULL, "vxca_caverns1_j22_0_1.vlm"},
	{0xB600EB9A499BFDD8ULL, "hlac_lt_0_lowtown_docks_a3_-65538_6y.dds"},
	{0xB60152D5BBF72347ULL, "fen110_anso6170039.cl"},
	{0xB6015DC4460FB1AEULL, "hlac_lt_0_lowtown_docks_a3_65535_4y.dds"},
	{0xB6017589D79FE901ULL, "w_sundermount_96_0.rml"},
	{0xB601A7F76BDC48B8ULL, "vxht_bartrand_n_5t4_0_0.vlm"},
	{0xB601CF1A53FED645ULL, "art\\vfx\\combat\\vfx_glb_levelup_c.impa.gad"},
	{0xB601D4F8B8961B06ULL, "vxht_bartrand_n_5tz_0_0.vlm"},
	{0xB602389F5CA28CDCULL, "1066033362.ogg"},
	{0xB602AF7A9DB45651ULL, "356844700.ogg"},
	{0xB6030DA22B855524ULL, "326837207.ogg"},
	{0xB6036E9B1B824928ULL, "hlac_lt_0_lowtown_a2_n_c_-1_7y.dds"},
	{0xB603AAB5B4CC1022ULL, "571167748.ogg"},
	{0xB603D5C323B8A3DEULL, "hlac_dr_0_deeproads_5841_4z.dds"},
	{0xB60400EAA51A8D3AULL, "vxlt_gallowscourtyard_7ib_0_0.vlm"},
	{0xB604028D8D1B4415ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_bellyspew.ani"},
	{0xB604140D48D0B573ULL, "6122735_f.fxe"},
	{0xB6041C15FF73294AULL, "dae000cr_grand_cleric6148079.cl"},
	{0xB604C510872DEA2EULL, "hlac_lt_0_lowtown_1_4z.dds"},
	{0xB604CCAD447ABF83ULL, "mag131_keran6086168.cl"},
	{0xB604CE0407C06F4CULL, "w_blightlands_mmap.dds"},
	{0xB6056E045887DFEEULL, "art\\vfx\\cinematic\\vfx_mage_punch_c.impa.gad"},
	{0xB60589FE786C20FAULL, "vxlt_gallowstemplar_n_3mc_0_0.vlm"},
	{0xB60686753B523853ULL, "6087055_m.fxe"},
	{0xB606E8E5FB40B44DULL, "64368572.ogg"},
	{0xB6079288B50A2E4CULL, "vxlt_hangedman_n_2xc_0_1.vlm"},
	{0xB6079BDEEFB6A041ULL, "6043462_m.fxe"},
	{0xB607B5FC7F5C7250ULL, "art\\vfx\\combat\\vfx_f_and_fade_c.cess.ani"},
	{0xB6080F8C304B1744ULL, "art\\characters\\creatures\\ragedemon\\animation\\ctn180r.ani"},
	{0xB6081A6144FB66DEULL, "vfx_gore.bnk"},
	{0xB6088749860345F8ULL, "vxlt_blackemporium_za_0_0.vlm"},
	{0xB608C4BA15B3AE89ULL, "vxlt_gallowstemplar_5b0_0_0.vlm"},
	{0xB6094B0D0A60E408ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14b.msh"},
	{0xB609690D0A611732ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14b.mmh"},
	{0xB6099B1E7276677CULL, "gen00fl_merrill6119584.cl"},
	{0xB609E29035FC43B5ULL, "hlac_lt_0_lowtown_a2_n_65537_6y.dds"},
	{0xB609E391786D2C97ULL, "131384519.ogg"},
	{0xB609FC40322B72B6ULL, "gen00fl_dog6167023.cl"},
	{0xB609FF391F7A1502ULL, "558030506.ogg"},
	{0xB60A0471BACE89F7ULL, "lt_gallowstemplar_n_8165.anb"},
	{0xB60A541008D86757ULL, "hlac_ht_store_187_1y.dds"},
	{0xB60A76D386A3BE02ULL, "6155601_m.fxe"},
	{0xB60A8171C6853222ULL, "792306256.ogg"},
	{0xB60A96AD15602D03ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01a.msh"},
	{0xB60AA4AD1560455DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nec01a.mmh"},
	{0xB60AC8BDD71E2352ULL, "llac_lt_0_lowtown_docks_a3_n_-196609_0.dds"},
	{0xB60ADE373AC8643FULL, "zz_one_debug6079309.cl"},
	{0xB60ADFB44C3FBD83ULL, "plt_prm000pt_collectors.plo"},
	{0xB60B389B8EB9E979ULL, "553990363.ogg"},
	{0xB60BA061E752B1E2ULL, "hlac_w_woundedcoast_196614_0x.dds"},
	{0xB60BDF0AB37E0816ULL, "885391637.ogg"},
	{0xB60BE96731896502ULL, "6126846_f.fxe"},
	{0xB60C16F957172C50ULL, "zz_followers_banter6157750.cl"},
	{0xB60C6904588DF5CDULL, "art\\vfx\\cinematic\\vfx_mage_punch_c.impa.evt"},
	{0xB60CEC1C009F76FEULL, "vxht_keep_4b1_0_0.vlm"},
	{0xB60CF7B8769BCA43ULL, "85229331.ogg"},
	{0xB60D36CA9915E4EFULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.cess.evt"},
	{0xB60D38E6F101EF82ULL, "hlac_ht_0_hightown_65537_0y.dds"},
	{0xB60D3FDABA7C0FE8ULL, "staff_of_parthalan.dds"},
	{0xB60D5B1F9F222F88ULL, "character_character_p209.psh.fxo.dx11.assembly"},
	{0xB60E3310AB069549ULL, "6206153_m.fxe"},
	{0xB60E727687DCA885ULL, "vxca_caverns1_mines_gk6_0_1.vlm"},
	{0xB60E75B3F6599111ULL, "6179877_m.fxe"},
	{0xB60EAC1AA621ADF6ULL, "llac_w_sundermount_-262141_0.dds"},
	{0xB60F0C8C305130A6ULL, "art\\characters\\creatures\\ragedemon\\animation\\ctn180r.gad"},
	{0xB60F176B5B2316F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cmov_rf.ani"},
	{0xB61000D2EA77BC13ULL, "w_encounter3_81_0.phy"},
	{0xB61020D0DF1CF169ULL, "6107282_f.fxe"},
	{0xB610A3E750655BF6ULL, "vxht_keep_n_63k_0_0.vlm"},
	{0xB610BE3B6C537EE3ULL, "hlac_ht_bartrand_n_3368_0x.dds"},
	{0xB610DF12BCA7B1BBULL, "642504031.ogg"},
	{0xB610F0BAB8D65AB2ULL, "686264419.ogg"},
	{0xB6110E0CC6720623ULL, "framebuffereffects_bloomdiagonalblur_p0.psh.fxo.dx11.assembly"},
	{0xB6111C6C70939F4FULL, "6208548_m.fxe"},
	{0xB6112F8641F22147ULL, "clf260_fell_orden6070484.cl"},
	{0xB611C2EEA9A0CB51ULL, "follower_banter6169714.cl"},
	{0xB611C2F5B4D093D3ULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.cess.ani"},
	{0xB612FBA597860D92ULL, "vxlt_darktown_5cx_0_0.vlm"},
	{0xB61337CA991A5198ULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.cess.gad"},
	{0xB6133C438C938611ULL, "hlac_ca_cave_2194_12y.dds"},
	{0xB614066464C72B96ULL, "6148399_m.fxe"},
	{0xB6153E1981A9E39AULL, "226981677.ogg"},
	{0xB615DBB52B228789ULL, "hlac_lt_0_lowtown_docks_-65540_0x.dds"},
	{0xB6160EBC3B4B8A93ULL, "224052387.ogg"},
	{0xB6160F3D95335682ULL, "6111678_m.fxe"},
	{0xB616226B5B294793ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cmov_rf.gad"},
	{0xB6167F17AB5AE9A1ULL, "hlac_lt_darktown_-1_48y.dds"},
	{0xB616A43B95E9A9E1ULL, "kf_scion_3.mor"},
	{0xB616F335E34855C8ULL, "774671628.ogg"},
	{0xB6172C863C0C1368ULL, "llac_lt_0_lowtown_65536_11.dds"},
	{0xB617649316D8FD56ULL, "vxht_0_hightown_a3_bc5_0_0.vlm"},
	{0xB617B691ADC3FD1FULL, "llac_lt_0_lowtown_n_131072_3.dds"},
	{0xB618724A07095606ULL, "art\\characters\\weapons\\arrows\\w_arw_001.cess.evt"},
	{0xB618A6579B6D4E67ULL, "art\\characters\\creatures\\ragedemon\\animation\\rdy_crnf.gad"},
	{0xB618B539F42E2BD0ULL, "664795586.ogg"},
	{0xB618CC93E34C1791ULL, "co_ele.mao"},
	{0xB618DDEC77B98A32ULL, "195005345.ogg"},
	{0xB618EDFF3BB48FA8ULL, "vxht_brothel_n_6k6_0_0.vlm"},
	{0xB6192784BC65320BULL, "mag170_thrask6054069.cl"},
	{0xB61974651CB15FCEULL, "902837245.ogg"},
	{0xB61996E0E28F3CE1ULL, "gen00fl_varric6104945.cl"},
	{0xB619AD4EB7AAAC6AULL, "isa311_party_comments6117810.cl"},
	{0xB619CE83E54B7D5CULL, "lowtown_fem_prostitute_sta6152307.cl"},
	{0xB61AA6374A81E57BULL, "25461410.ogg"},
	{0xB61B821138C54C04ULL, "889219632.ogg"},
	{0xB61BB2FC0AA86DB7ULL, "follower_banter6171032.cl"},
	{0xB61BBACE5C90822DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\blendtree\\floortrap.anb"},
	{0xB61BC49AC16AE362ULL, "585811022.ogg"},
	{0xB61BCB711EA46D52ULL, "hlac_w_sundermount_393213_0x.dds"},
	{0xB61BFDD6202613BEULL, "art\\characters\\creatures\\strider\\animation\\walkback.evt"},
	{0xB61C02291D2D960BULL, "sse_ss_coterie_thug_f.gda"},
	{0xB61C9654BC193C38ULL, "6048413_m.fxe"},
	{0xB61CF9F21A229809ULL, "vxht_0_hightown_bs3_0_0.vlm"},
	{0xB61D1F6B5B2F61F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_cmov_rf.evt"},
	{0xB61E267B9F44611BULL, "780813754.ogg"},
	{0xB61E6262C5073061ULL, "951567587.ogg"},
	{0xB61E6379EA13C3BCULL, "vxlt_hangedman_1tf_0_0.vlm"},
	{0xB61E985D41CAD111ULL, "hlac_ht_playershouse_n_855_19z.dds"},
	{0xB61EBAF5B4DB100AULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.cess.evt"},
	{0xB61EC3CB6C310222ULL, "art\\vfx\\cinematic\\vfx_meredith_dies_c.mmh"},
	{0xB61F0F89E6818AB9ULL, "vxca_caverns2_1ye_0_0.vlm"},
	{0xB61F894A070F9C51ULL, "art\\characters\\weapons\\arrows\\w_arw_001.cess.gad"},
	{0xB61FD16E9396328EULL, "6149077_m.fxe"},
	{0xB61FEF2E35206D5EULL, "vxca_caverns1_mines_im4_0_0.vlm"},
	{0xB6200C49A2FCBECAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout.mao"},
	{0xB6201049A2FCC519ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout.mmh"},
	{0xB6201A49A2FCD607ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout.msh"},
	{0xB620777A381C5C17ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0cc_basic.msh"},
	{0xB620897A381C7B7AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0cc_basic.mao"},
	{0xB6208D7A381C8189ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0cc_basic.mmh"},
	{0xB62159A6550C188BULL, "vxht_chantry_n_2zi_0_0.vlm"},
	{0xB621E7670A37ADADULL, "6018444_m.fxe"},
	{0xB6224C726E4FF24BULL, "6084119_m.fxe"},
	{0xB62314D6202C59E9ULL, "art\\characters\\creatures\\strider\\animation\\walkback.gad"},
	{0xB6239ED3AAC667F9ULL, "vxlt_darktown_49h_0_0.vlm"},
	{0xB623B0AA429FC99AULL, "character_character_p119.psh.fxo"},
	{0xB6245325E7540918ULL, "6089784_m.fxe"},
	{0xB624C1D5007DD611ULL, "hlac_lt_0_lowtown_docks_-65539_2z.dds"},
	{0xB62582952BC45C84ULL, "dae000_porfiria6155295.cl"},
	{0xB6258A4A071409FFULL, "art\\characters\\weapons\\arrows\\w_arw_001.cess.ani"},
	{0xB62599F5B4E0F77DULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.cess.gad"},
	{0xB625BF748B65A3F6ULL, "art\\vfx\\combat\\vfx_a_disengage_aoe_p.anb"},
	{0xB625D9CFBF5B27D6ULL, "vxlt_undercitydungeon_1j5_0_1.vlm"},
	{0xB6260F725E2C0C2AULL, "7608568.ogg"},
	{0xB62696AC785E8C3FULL, "mag311_keran6041712.cl"},
	{0xB626E05C2B815DB5ULL, "ht_0_hightown_41_0.rml"},
	{0xB6275FDC2DAAA828ULL, "vxht_brothel_n_6t7_0_0.vlm"},
	{0xB6276A0CF72510B2ULL, "6042062_m.fxe"},
	{0xB627A335A4D53B16ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\additive\\human_to_qunari_face_scale.ani"},
	{0xB628072E8AE35933ULL, "79515844.ogg"},
	{0xB6283DBC5BBF5941ULL, "vxlt_hangedman_m5_0_0.vlm"},
	{0xB628531F068C8007ULL, "darktown_fem_lowlife_mob6176169.cl"},
	{0xB628E6FD832F766FULL, "6134605_m.fxe"},
	{0xB6294B5A4EAD2FEFULL, "vxlt_0_lowtown_a2_n_fed_0_0.vlm"},
	{0xB62979A3BB1119DBULL, "228636581.ogg"},
	{0xB6299247BF581B56ULL, "677231677.ogg"},
	{0xB629BBECBA5EA449ULL, "vxw_bonepit_16p_0_0.vlm"},
	{0xB629D3C35B7A449EULL, "599743092.ogg"},
	{0xB629E5D620322837ULL, "art\\characters\\creatures\\strider\\animation\\walkback.ani"},
	{0xB62A071FB56017CAULL, "591947120.ogg"},
	{0xB62A7CD0356B79DDULL, "w_encounter1_17_0_lc.mmh"},
	{0xB62ADED038BCA8FDULL, "vxht_store_ci_0_0.vlm"},
	{0xB62B02FE967822BAULL, "vxw_bonepitdestroyed_a3_1pe_0_0.vlm"},
	{0xB62B282DC6A59BF1ULL, "vxdr_primevaltomb_8ot_0_0.vlm"},
	{0xB62C37810061EC3BULL, "vxca_caverns1_i7j_0_0.vlm"},
	{0xB62C5698393CB265ULL, "vxlt_darktown_6kd_0_0.vlm"},
	{0xB62C93579B7DD6ADULL, "art\\characters\\creatures\\ragedemon\\animation\\rdy_crnf.ani"},
	{0xB62CBF3D938BF011ULL, "character_character_p266.psh.fxo.dx11"},
	{0xB62CD556DC09DC3CULL, "vxlt_hangedman_1sg_0_0.vlm"},
	{0xB62D2DAEBD238BDFULL, "vxca_caverns2prc_9ro_0_0.vlm"},
	{0xB62D7F0FA569B155ULL, "sib131_slave_a6165277.cl"},
	{0xB62DEB6DEECC3851ULL, "art\\vfx\\materials\\chunky\\vfx_c_crsh_pris_blend.mao"},
	{0xB62E610B271685DEULL, "vxca_caverns1_mines_ij5_0_0.vlm"},
	{0xB62E6B0B62903E8EULL, "vxlt_0_lowtown_docks_8pf_0_0.vlm"},
	{0xB62EDB33EDCA3084ULL, "vxlt_backalley_n_1yw_0_0.vlm"},
	{0xB62F2A7146D9CB59ULL, "vxht_0_hightown_a3_auo_0_0.vlm"},
	{0xB62F2B1F5ED90DC6ULL, "569672605.ogg"},
	{0xB62F53A70AFCF520ULL, "mag360_thrask6043560.cl"},
	{0xB62F9885BB2054FCULL, "vxht_keep_3ia_0_0.vlm"},
	{0xB62FDF5F7B723E92ULL, "vxlt_darktown_5iz_0_0.vlm"},
	{0xB62FFB80CE362915ULL, "gen00fl_sebastian6179928.cl"},
	{0xB63081A27CE0B045ULL, "vxlt_gallowstemplar_n_3mh_0_0.vlm"},
	{0xB631A607629853F4ULL, "6103253_m.fxe"},
	{0xB632AC8B044E3F7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\teleportfull.ani"},
	{0xB632EA9E50D8DA16ULL, "qun120_arishok6051489.cl"},
	{0xB6331B75D7818FD3ULL, "93558245.ogg"},
	{0xB6331FC95C1B2114ULL, "6068362_carver_leave.cut"},
	{0xB633C2748B71E661ULL, "art\\vfx\\combat\\vfx_a_disengage_aoe_p.mmh"},
	{0xB6342BD14DFD12DDULL, "l_dae310ar_docks.bnk"},
	{0xB634B0E730225DA8ULL, "lgt210_eau_symp_templar6144450.cl"},
	{0xB634B7AAF8A99446ULL, "57465881.ogg"},
	{0xB634EC45205DCC46ULL, "ss_anders.bnk"},
	{0xB6356EF92891A064ULL, "vxht_mansion_n_4zt_0_0.vlm"},
	{0xB6358210E87EA88DULL, "vxlt_warehouse_12a_0_0.vlm"},
	{0xB636D1232568A1C6ULL, "follower_banter6151257.cl"},
	{0xB6370171BAF5720BULL, "lt_gallowstemplar_n_8165.rml"},
	{0xB637B21DD4B56C9EULL, "6121642_m.fxe"},
	{0xB637D846214A0243ULL, "131868189.ogg"},
	{0xB63878C215B6CA1BULL, "vxw_encounter2_5p_0_0.vlm"},
	{0xB6388A223D8034ACULL, "352782052.ogg"},
	{0xB63974DB886E3B3AULL, "vxw_bonepitdestroyed_a3_1uf_0_0.vlm"},
	{0xB63977C174185538ULL, "vxlt_0_lowtown_docks_b4w_0_1.vlm"},
	{0xB639D770F166B7A2ULL, "ht_mansion_n_3366.anb"},
	{0xB639DDABFE0EC909ULL, "vxht_0_hightown_by1_0_0.vlm"},
	{0xB63A358720F21370ULL, "698181218.ogg"},
	{0xB63AA0A6661FCA0BULL, "26663835.ogg"},
	{0xB63AA9778A831BF9ULL, "vxdr_0_deeproads_gfn_0_0.vlm"},
	{0xB63B2CF9D75C3E43ULL, "hlac_lt_0_lowtown_65538_3z.dds"},
	{0xB63BD7E0843E8BA9ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.dura.ani"},
	{0xB63C16FA8C71838FULL, "hlac_lt_0_lowtown_a3_n_c_131072_6z.dds"},
	{0xB63C44E718314AE8ULL, "6053534_m.fxe"},
	{0xB63CCC35A4E7DD50ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\additive\\human_to_qunari_face_scale.gad"},
	{0xB63DD41081570DBAULL, "536184083.ogg"},
	{0xB63E55696936E302ULL, "follower_banter6141726.cl"},
	{0xB63EBF96C1E9387FULL, "dre000_carver6037613.cl"},
	{0xB63ED0F1350A997AULL, "332062002.ogg"},
	{0xB640B48B045A8A72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\teleportfull.evt"},
	{0xB640CBF65BA159F2ULL, "hlac_lt_gallowstemplar_n_2394_2y.dds"},
	{0xB6411A5EAEFABEE3ULL, "12606516.ogg"},
	{0xB6416789D7D5B5F0ULL, "w_sundermount_96_0.msh"},
	{0xB6417589D7D5CD0AULL, "w_sundermount_96_0.mmh"},
	{0xB6417989D7D5D441ULL, "w_sundermount_96_0.mao"},
	{0xB6417B5E908C8EE0ULL, "llac_w_encounter3_-327675_0.dds"},
	{0xB6419A3E6FF21518ULL, "6140807_m.fxe"},
	{0xB641B679B85D8BA9ULL, "dae021_sister_samea_sta6142256.cl"},
	{0xB641DA40EFCA93D8ULL, "531343654.ogg"},
	{0xB642120A8491A1C6ULL, "6104618_m.fxe"},
	{0xB64213E0DF8945C7ULL, "6096826_m.fxe"},
	{0xB642F2E08444D82BULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.dura.gad"},
	{0xB6433ED0697911BDULL, "hlac_lt_0_lowtown_docks_a3_-131074_3z.dds"},
	{0xB64341CDC14446A5ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_2x.dds"},
	{0xB6435398671B15F5ULL, "limp_blue.mao"},
	{0xB643858354AAC3D9ULL, "518400245.ogg"},
	{0xB64390B7D8D084BAULL, "6174569_m.fxe"},
	{0xB643C8CC77A4BFBAULL, "6165563_m.fxe"},
	{0xB643DBC15455C5DDULL, "vxw_sundermount_34c_0_0.vlm"},
	{0xB643EB632C8A1D94ULL, "plt_rumor_wayward_son.nss"},
	{0xB6445D14E7316CEDULL, "vxlt_0_lowtown_e1j_0_0.vlm"},
	{0xB64498A87DA8215AULL, "6170572_m.fxe"},
	{0xB644A596E14F109CULL, "vxca_cavern1_haunted_im6_0_0.vlm"},
	{0xB644FACB427EAC00ULL, "6109111_f.fxe"},
	{0xB6453153DB563293ULL, "llac_lt_0_lowtown_n_-1_5.dds"},
	{0xB6456A5B8A5AB8BAULL, "371513020.ogg"},
	{0xB645E67C86AB0F80ULL, "vxca_cavern1_haunted_hvr_0_0.vlm"},
	{0xB646139F6242CC6AULL, "mag320_bsc_banter6100773.cl"},
	{0xB646543561D203EEULL, "vxht_chantry_n_28r_0_0.vlm"},
	{0xB64656FC8FCA8CE8ULL, "w_blightlands_188_0.mmh"},
	{0xB64662FC8FCAA08BULL, "w_blightlands_188_0.mao"},
	{0xB6466A7075531D0CULL, "gen00fl_aveline6137574.cl"},
	{0xB64674FC8FCABF12ULL, "w_blightlands_188_0.msh"},
	{0xB64697F127414E8AULL, "hlac_lt_0_lowtown_a2_n_c_-3_0x.dds"},
	{0xB646C54CC8434E7EULL, "hlac_lt_darktown_-1_62z.dds"},
	{0xB646D38B045F2A05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\teleportfull.gad"},
	{0xB647AC2EA8AE9006ULL, "gen00fl_varric6093671.cl"},
	{0xB647BECE4A0ED819ULL, "vxw_encounter2_camp_74_0_0.vlm"},
	{0xB647C2CB6C5405C9ULL, "art\\vfx\\cinematic\\vfx_meredith_dies_c.anb"},
	{0xB647C5548877B4FCULL, "383996866.ogg"},
	{0xB647EC614530FCE4ULL, "vfx_gore.wbd"},
	{0xB6481EFD7158205BULL, "568717878.ogg"},
	{0xB6483122EE9B8710ULL, "hlac_lt_gallowstemplar_n_2391_1y.dds"},
	{0xB64864ECF554FEBDULL, "vxw_encounter2_camp_s4_0_0.vlm"},
	{0xB648A789E2789722ULL, "lowtown_fem_prostitute_sta6161602.cl"},
	{0xB648AE16EC341063ULL, "vxw_encounter2_camp_3z_0_0.vlm"},
	{0xB648EFE084493E90ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.dura.evt"},
	{0xB648F674D8A193E1ULL, "vxht_0_hightown_a2_a6o_0_0.vlm"},
	{0xB6497AC92C02E577ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v1.gad"},
	{0xB64A4B027EDF57B7ULL, "847779096.ogg"},
	{0xB64ABC9062E37EB0ULL, "6118231_m.fxe"},
	{0xB64AE50D2DC44AFBULL, "hlac_dr_slavers_3207_3z.dds"},
	{0xB64C08DC4E0F0E75ULL, "vxw_blightlands_lr_0_0.vlm"},
	{0xB64C304FB963F735ULL, "vxw_encounter1_cq_0_0.vlm"},
	{0xB64C3906BFBD5E35ULL, "vxlt_hangedman_n_rm_0_0.vlm"},
	{0xB64CB78501FDF57DULL, "6201568_f.fxe"},
	{0xB64CC3195F5E6F92ULL, "vxlt_darktown_5wd_0_0.vlm"},
	{0xB64CEF26792BBA44ULL, "511096220.ogg"},
	{0xB64D1A70D0797433ULL, "263419943.ogg"},
	{0xB64D360E8B18D81BULL, "hlac_ca_caverns1_mines_24425_2x.dds"},
	{0xB64D4A396760A313ULL, "6005048_m.fxe"},
	{0xB64D5DAAF009CF2EULL, "llac_lt_0_lowtown_docks_65535_0.dds"},
	{0xB64D6A03986E098CULL, "vxca_caverns1_mines_ehu_0_0.vlm"},
	{0xB64DC4343513D4BBULL, "6166633_m.fxe"},
	{0xB64E4020DB4413CBULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrc_0t.dds"},
	{0xB64ECAD7847FB8A1ULL, "dae200_bran6094778.cl"},
	{0xB64F09D125696FDCULL, "zz_isaac_hawke_crane_v1.cut"},
	{0xB64F13B83C79E8ADULL, "hlac_ht_store_187_0x.dds"},
	{0xB64F16687E56D04AULL, "vxw_encounter2_camp_pn_0_0.vlm"},
	{0xB64F7F95EBED6BDDULL, "qun160_vashothlead6064639.cl"},
	{0xB6508FC92C092704ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v1.evt"},
	{0xB650B83B58AEBE23ULL, "plt_genpt_app_merril.plo"},
	{0xB650DF7FED6003A6ULL, "hlac_lt_0_lowtown_docks_65534_5z.dds"},
	{0xB650E1F71D563801ULL, "vxlt_hangedman_n_175_0_0.vlm"},
	{0xB65155BFFE7329D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_b.ani"},
	{0xB6524D9E464BDE5DULL, "vxw_sundermount_33b_0_0.vlm"},
	{0xB6527F2A99F63FC8ULL, "vxht_0_hightown_a2_a6j_0_0.vlm"},
	{0xB6528DAE4D088C05ULL, "327461921.ogg"},
	{0xB6531773D345291CULL, "vxca_cavern1_haunted_ij7_0_0.vlm"},
	{0xB65318047B9A2745ULL, "13690457.ogg"},
	{0xB653B96E425DCAE9ULL, "art\\characters\\playercharacter\\humanmale\\boo_medc.phy"},
	{0xB65442DDC2E94C0FULL, "rewards.gda"},
	{0xB6549E7CF00738C9ULL, "6104576_m.fxe"},
	{0xB654C61253C81C6EULL, "vxht_chantry_n_23s_0_0.vlm"},
	{0xB654C92AF927DDA5ULL, "vxht_keep_n_605_0_0.vlm"},
	{0xB65524B9164A9AE6ULL, "752828.ogg"},
	{0xB65576C500D4A4B5ULL, "hlac_lt_0_lowtown_docks_-65538_7z.dds"},
	{0xB655AFD823E72F9BULL, "vxw_blightlands_gy_0_0.vlm"},
	{0xB655C92EACF1A068ULL, "946866397.ogg"},
	{0xB6563CE74A0ED6D6ULL, "vxca_cavern1_haunted_iof_0_0.vlm"},
	{0xB656D6C9E74B173DULL, "vxw_encounter2_camp_n3_0_0.vlm"},
	{0xB657AA1E0CEE6B71ULL, "6147167_f.fxe"},
	{0xB657D3DD9ACEE1FAULL, "6030495_f.fxe"},
	{0xB657D84578A4576DULL, "789586095.ogg"},
	{0xB65830BFFE79091BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_b.gad"},
	{0xB658728715985B3BULL, "w_bonepitdestroyed_a3_38_0.mmh"},
	{0xB6587E8715986FD8ULL, "w_bonepitdestroyed_a3_38_0.mao"},
	{0xB658908715988E25ULL, "w_bonepitdestroyed_a3_38_0.msh"},
	{0xB658C1D66F67DA7CULL, "848855891.ogg"},
	{0xB6596FD7E4BD7865ULL, "6054883_f.fxe"},
	{0xB659A3060F273174ULL, "6176165_m.fxe"},
	{0xB659C9012D67A6D7ULL, "vxw_sundermount_38e_0_0.vlm"},
	{0xB659E7E8EB977420ULL, "568874488.ogg"},
	{0xB65A3B90FB963B6DULL, "324500245.ogg"},
	{0xB65A42C774A723E4ULL, "611886082.ogg"},
	{0xB65A7AB9400526F5ULL, "vxw_blightlands_qq_0_0.vlm"},
	{0xB65AA22CAB5A0FB5ULL, "vxw_encounter1_hp_0_0.vlm"},
	{0xB65AE3025140E34DULL, "cs_mcr310cs_scared_mage.wbd"},
	{0xB65B532300A3F4D4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownexit.ani"},
	{0xB65B7B61EEA4551FULL, "586657868.ogg"},
	{0xB65B99382E139A59ULL, "hlac_ht_brothel_n_3370_1z.dds"},
	{0xB65BA7B62408BA3DULL, "75230183.ogg"},
	{0xB65BC564FCDDEBB0ULL, "hlac_lt_0_lowtown_a2_-65537_3y.dds"},
	{0xB65BDCB02226DE0EULL, "vxca_caverns2_7n5_0_0.vlm"},
	{0xB65BE765B8054D13ULL, "vxlt_hangedman_n_20q_0_0.vlm"},
	{0xB65BF950D3245D25ULL, "6197640_m.fxe"},
	{0xB65C05FC08DC5761ULL, "286550704.ogg"},
	{0xB65C23E86A55B906ULL, "716354651.ogg"},
	{0xB65C2BAE0C91EF76ULL, "6051813_m.fxe"},
	{0xB65C74C2F075BB78ULL, "mer221_sabin6080575.cl"},
	{0xB65CB3956F0E5B2DULL, "6176296_m.fxe"},
	{0xB65D102EFF885DB2ULL, "gen_im_arm_hlm_lgt_act3_02.uti"},
	{0xB65D3B1442759390ULL, "6153696_m.fxe"},
	{0xB65D8845704CE8CAULL, "vxw_encounter2_camp_ko_0_0.vlm"},
	{0xB65DA7C92C13DAFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_atk_v1.ani"},
	{0xB65DE89A0B97DBC1ULL, "6169912_f.fxe"},
	{0xB65E63B32145986BULL, "982271194.ogg"},
	{0xB65EE13B120F0D73ULL, "gen_im_trs_chn_light_doc_02.uti"},
	{0xB65EE3FB28D8F680ULL, "hlac_dr_slavers_3217_7x.dds"},
	{0xB65F348750F5CCF9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8man.msh"},
	{0xB65F3A24AD0F7F00ULL, "41571840.ogg"},
	{0xB65F3E8750F5DDE7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8man.mmh"},
	{0xB65F428750F5E434ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8man.mao"},
	{0xB65F4DBFFE7F5900ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_b.evt"},
	{0xB65F68967F1C018BULL, "sib100st_father_contact.stg"},
	{0xB65FC35748B94299ULL, "darktown_fem_fer_urch_sta6174789.cl"},
	{0xB65FF6C3D8C11CC4ULL, "hlac_lt_0_lowtown_n_-131072_7x.dds"},
	{0xB65FFD4CB0EA8184ULL, "hlac_lt_shop_-1_2x.dds"},
	{0xB661878D7B8E7EE0ULL, "898110650.ogg"},
	{0xB661DEECCDF68679ULL, "gen00fl_varric6089660.cl"},
	{0xB6620C0D833C0810ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_manystriplow_n.dds"},
	{0xB6623471D84D4105ULL, "vxlt_blackemporium_ue_0_0.vlm"},
	{0xB6623D30FA8FC453ULL, "gen_im_trs_slv_lgt_hlm_r2.uti"},
	{0xB662502300AA0E36ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownexit.gad"},
	{0xB6625EDE4C9B3E1CULL, "6116189_m.fxe"},
	{0xB6627CAE7AE5ADA2ULL, "gen00fl_anders6125910.cl"},
	{0xB662979737D06FDCULL, "6081977_m.fxe"},
	{0xB6629AC9918116DFULL, "vxlt_gallowscourtyard_7f6_0_0.vlm"},
	{0xB6633C01AAE46682ULL, "dae011_corff_sta6142452.cl"},
	{0xB6639ABDA7609C6BULL, "11214057.ogg"},
	{0xB664169B13875E4CULL, "gen00fl_fenris6110907.cl"},
	{0xB66421B515DD481BULL, "vxw_blightlands_hx_0_0.vlm"},
	{0xB6642224232B1BBEULL, "tacticsconditions.gda"},
	{0xB6644759CE229A4BULL, "hlac_ht_0_hightown_a3_131074_0y.dds"},
	{0xB665064E0E430E0FULL, "642052156.ogg"},
	{0xB665099ACE4F2313ULL, "vxlt_0_lowtown_a2_n_c_hpc_0_0.vlm"},
	{0xB665680D0AAF742DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos14b.phy"},
	{0xB66654601A27A4EDULL, "mcr310_confused_guard6077078.cl"},
	{0xB66669A189CAA0D8ULL, "vxht_0_hightown_ak9_0_0.vlm"},
	{0xB666BCC876EC66C7ULL, "ts_dwarven_02.gda"},
	{0xB6676D9F90532496ULL, "vxca_caverns1_mines_i2u_0_0.vlm"},
	{0xB6678434BD855A88ULL, "hlac_lt_gallowstemplar_n_7198_5y.dds"},
	{0xB667F235E0634EC9ULL, "vxca_caverns1_e5z_0_0.vlm"},
	{0xB667F574934C7C1CULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf_v2.ani"},
	{0xB66815B72CDEAFDBULL, "251810129.ogg"},
	{0xB66821E258953258ULL, "fen110_party_comment6117726.cl"},
	{0xB6686B2300AEA795ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownexit.evt"},
	{0xB668C7A9DBA6AEBEULL, "qcr000_arishok6039303.cl"},
	{0xB668D14C076708C4ULL, "vxlt_0_lowtown_f0k_0_0.vlm"},
	{0xB66969482E202D9DULL, "311912880.ogg"},
	{0xB669907E1D7FD31DULL, "vxlt_hangedman_n_1w0_0_1.vlm"},
	{0xB66A4DBD7C5A3A8CULL, "vxca_caverns1_mines_ebk_0_0.vlm"},
	{0xB66A934D53B8339CULL, "vxlt_hangedman_31o_0_0.vlm"},
	{0xB66AC774BF06CBC0ULL, "198306038.ogg"},
	{0xB66B712DF103A3E1ULL, "mcr371_orsino6083345.cl"},
	{0xB66BA666E5562DA0ULL, "vxca_caverns1_k6t_0_0.vlm"},
	{0xB66BC4D56405A9C4ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_003a_0d.dds"},
	{0xB66BEDA126E83D30ULL, "vxca_cavern1_haunted_e4m_0_0.vlm"},
	{0xB66BF356A9269956ULL, "vxca_caverns2prc_9pp_0_0.vlm"},
	{0xB66BFA226243014AULL, "vxw_encounter2_camp_fh_0_0.vlm"},
	{0xB66D18271A086336ULL, "vxlt_darktown_6bx_0_0.vlm"},
	{0xB66D622C17E8FFEBULL, "6131220_m.fxe"},
	{0xB66DF1650FDAFC88ULL, "6102947_f.fxe"},
	{0xB66E215DC4F561B9ULL, "6130046_f.fxe"},
	{0xB66EE9324A8219B2ULL, "vxlt_gallowstemplar_n_397_0_0.vlm"},
	{0xB66F02749352B15EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf_v2.gad"},
	{0xB66F365C2BCE2FE5ULL, "6173277_m.fxe"},
	{0xB670FA60F5980B9AULL, "gen00fl_fenris6113878.cl"},
	{0xB6711A3CCD7F69BAULL, "vxdr_primevaltomb_8me_0_0.vlm"},
	{0xB67142EDAC4C85CBULL, "vxlt_gallowscourtyard_72v_0_0.vlm"},
	{0xB671906417CBFFCAULL, "zz_followers_banter6158652.cl"},
	{0xB671913DAA1CFB05ULL, "def_grass_d.dds"},
	{0xB671F771CB7C2142ULL, "820522991.ogg"},
	{0xB671F9D2EACA7A88ULL, "w_encounter3_81_0.mmh"},
	{0xB6720124ED2E068AULL, "6198097_m.fxe"},
	{0xB67205D2EACA8F2BULL, "w_encounter3_81_0.mao"},
	{0xB67217D2EACAADB2ULL, "w_encounter3_81_0.msh"},
	{0xB67282FD298A1F16ULL, "vxht_mansion2b_3bf_0_0.vlm"},
	{0xB672A0EAA35D1434ULL, "hlac_lt_0_lowtown_131073_0z.dds"},
	{0xB672BB0573284A5BULL, "vxw_encounter1_dz_0_0.vlm"},
	{0xB672FF23272EB222ULL, "dae000_flemeth6013299.cl"},
	{0xB6730CE7FFA9DB23ULL, "1056375576.ogg"},
	{0xB673721C6A3D42C9ULL, "gen00fl_aveline6137692.cl"},
	{0xB6738DE22C4F714BULL, "lt_ruinedhouse_n_mmi.gff"},
	{0xB673AD7DC2ED97DEULL, "vxca_caverns1_mines_j2r_0_0.vlm"},
	{0xB673DD2FFE2CD6BCULL, "hlac_w_sundermount_131068_0x.dds"},
	{0xB673EF4337552B87ULL, "character_character_p206.psh.fxo.dx11"},
	{0xB67412F1016641A5ULL, "722358374.ogg"},
	{0xB6744E83453FE505ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordb_sa.dds"},
	{0xB6748AC7F5B67174ULL, "180738173.ogg"},
	{0xB674AA3AE0D45F25ULL, "6048470_m.fxe"},
	{0xB674BAE78BCC1F49ULL, "vxlt_backalley_g_1w5_0_0.vlm"},
	{0xB6750EA795D856BCULL, "vxlt_hangedman_1jx_0_0.vlm"},
	{0xB6759C4CF557A3F9ULL, "vxlt_warehouse_n_cj_0_0.vlm"},
	{0xB675DD749358917DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_crnf_v2.evt"},
	{0xB675E1D565870A0CULL, "hlac_lt_0_lowtown_docks_a3_-65536_0y.dds"},
	{0xB6763AEDB97B9335ULL, "gen00fl_anders6183818.cl"},
	{0xB6768CD7A767A659ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medb_0t.dds"},
	{0xB6769CDA3B5DF5C9ULL, "6069624_m.fxe"},
	{0xB676BFFC996F144BULL, "t1_mul_pu2.tnt"},
	{0xB6772074089C5DDAULL, "mrl000_marethari6142149.cl"},
	{0xB6774328F95D2144ULL, "vxlt_0_lowtown_f5h_0_0.vlm"},
	{0xB67777976B4185AAULL, "darktown_fem_elf_sta6174518.cl"},
	{0xB677830741750039ULL, "100954280.ogg"},
	{0xB677BE15AC9F3E78ULL, "650775747.ogg"},
	{0xB6783D5DC3F9F851ULL, "vxw_bonepitdestroyed_a3_mz_0_0.vlm"},
	{0xB67861764BFACC6FULL, "vxht_mansion2b_2oy_0_0.vlm"},
	{0xB67968FF90E34CFEULL, "339043268.ogg"},
	{0xB679A57B022C6DC9ULL, "6111480_m.fxe"},
	{0xB67A6BFF543919CAULL, "vxw_encounter2_camp_ai_0_0.vlm"},
	{0xB67A9C7E4DC8F39FULL, "mag330_walter6044530.cl"},
	{0xB67AAEF5BAEBE5BBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_curtains_sa.dds"},
	{0xB67C2B943627B37FULL, "hlac_ca_caverns2_5802_33z.dds"},
	{0xB67C2DBADFDCB246ULL, "6119450_m.fxe"},
	{0xB67C77DE69206FDBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towera.msh"},
	{0xB67C89DE69208E26ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towera.mao"},
	{0xB67C95DE6920A2C5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towera.mmh"},
	{0xB67CBF25A74FBD74ULL, "dae000ip_crowd_ff7_talk4.utp"},
	{0xB67CDF5F5CD6C322ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_001a_0t.dds"},
	{0xB67CE933A612766FULL, "6125171_m.fxe"},
	{0xB67D1F079F09A61AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\4sd_basic_2.mao"},
	{0xB67D50ABA7714647ULL, "53918872.ogg"},
	{0xB67D5BC5CCDDA8E1ULL, "qun110_petrice6018740.cl"},
	{0xB67D5FD059B7DF75ULL, "6102343_m.fxe"},
	{0xB67E0E5DA5B35068ULL, "vxdr_primevaltomb_6zq_0_1.vlm"},
	{0xB67E2CB67EA07C15ULL, "hlac_lt_gallowscourtyard_131069_1x.dds"},
	{0xB67E3089A0F095B0ULL, "vxlt_hangedman_n_259_0_0.vlm"},
	{0xB67E83304F492B40ULL, "zz_ave_debug6207521.cl"},
	{0xB67F68473020536AULL, "563796589.ogg"},
	{0xB67FA26E4282DE0EULL, "art\\characters\\playercharacter\\humanmale\\boo_medc.mmh"},
	{0xB67FAE6E4282F2A5ULL, "art\\characters\\playercharacter\\humanmale\\boo_medc.mao"},
	{0xB67FBC6E42830AFCULL, "art\\characters\\playercharacter\\humanmale\\boo_medc.msh"},
	{0xB68048D6D2B091F3ULL, "hlac_w_encounter2_camp_-65535_0z.dds"},
	{0xB680D68663927352ULL, "hightown_male_cityguard_mob6170020.cl"},
	{0xB681056EF9C9791BULL, "vxw_blightlands_vz_0_0.vlm"},
	{0xB68118C942C19993ULL, "610575377.ogg"},
	{0xB681A0667C0D6D73ULL, "gen00fl_merrill6142101.cl"},
	{0xB681CA3D7B48E01AULL, "421748228.ogg"},
	{0xB68200D2BE18BCA5ULL, "vxht_mansion_n_6no_0_0.vlm"},
	{0xB6821149A34F9106ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockout.phy"},
	{0xB682563B68403065ULL, "art\\vfx\\environment\\vfx_steam_s_p.mmh"},
	{0xB6828675B49A8152ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_loop.gad"},
	{0xB682B681F7D4F48CULL, "181314761.ogg"},
	{0xB6837B229CDF2952ULL, "6125893_m.fxe"},
	{0xB684A3269ECECB0BULL, "405112882.ogg"},
	{0xB684AFDC72E94FBCULL, "976150376.ogg"},
	{0xB684E2DB984A596DULL, "864174450.ogg"},
	{0xB68556DE838D0391ULL, "gen00fl_anders6125380.cl"},
	{0xB6858019C7D475F6ULL, "hlac_ht_0_hightown_a2_65536_4x.dds"},
	{0xB685F720D4418BDDULL, "hlac_w_sundermount_65537_0z.dds"},
	{0xB68611A9671C0698ULL, "default_missing.dds"},
	{0xB6864EE7A24717F6ULL, "769721279.ogg"},
	{0xB6866D04940424EAULL, "6100073_f.fxe"},
	{0xB686AF3AB5F010D1ULL, "vxw_bonepitdestroyed_a3_hy_0_0.vlm"},
	{0xB686C2FB501693D2ULL, "412379011.ogg"},
	{0xB6877CF5CBE38874ULL, "6013282_m.fxe"},
	{0xB687BE6DDD869D27ULL, "vxca_caverns2_852_0_0.vlm"},
	{0xB6888DDA98B55B2FULL, "llac_dr_primevalentrance_1726_3.dds"},
	{0xB688D7108D12CA56ULL, "vxca_caverns2prc_9zr_0_0.vlm"},
	{0xB688F05567D04B13ULL, "gen00fl_isabela6135712.cl"},
	{0xB6890940BA3D1B2DULL, "766857752.ogg"},
	{0xB689E033A4693839ULL, "1015360492.ogg"},
	{0xB68A1477343DC8CBULL, "178990221.ogg"},
	{0xB68A1C2B92B1313CULL, "vxw_encounter2_camp_7m_0_0.vlm"},
	{0xB68A4307B23D1077ULL, "322196893.ogg"},
	{0xB68A8D852BCAB22DULL, "6013355_m.fxe"},
	{0xB68AF292011DE141ULL, "follower_banter6161336.cl"},
	{0xB68B05CC8BA9D794ULL, "unsyncdeathanimations.gda"},
	{0xB68B10E490AEC866ULL, "vxlt_0_lowtown_gbz_0_0.vlm"},
	{0xB68B54DA6F901153ULL, "mrl311ip_mirror_whole.utp"},
	{0xB68B5B1098088EEFULL, "6122071_m.fxe"},
	{0xB68B60FC757D9926ULL, "hlac_ht_mansion2_818_26x.dds"},
	{0xB68B7B51EF4916C0ULL, "fen200_slavers_exterior_st.stg"},
	{0xB68B9BE4B6A98AC5ULL, "mrl000_marethari6117696.cl"},
	{0xB68BA98E3FBCF58EULL, "vxlt_gamlenshouse_n_jt_0_0.vlm"},
	{0xB68BAB7FBC4CA227ULL, "llac_lt_0_lowtown_docks_a3_n_-3_3.dds"},
	{0xB68BFC4677B01199ULL, "384190967.ogg"},
	{0xB68C4C430C7AE4D6ULL, "6188590_f.fxe"},
	{0xB68C9ADB8EF8F25CULL, "974446763.ogg"},
	{0xB68CD6C6D7310CB6ULL, "6127006_m.fxe"},
	{0xB68E02BBE57D1135ULL, "qun100_mage_party_comment6019105.cl"},
	{0xB68ED7C169F2699FULL, "vxlt_house_et_0_0.vlm"},
	{0xB68F6EA68FEAD50CULL, "173098569.ogg"},
	{0xB68F75E8F5F1DA98ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_04a_0n.dds"},
	{0xB68FAB6ABCEA9418ULL, "557764877.ogg"},
	{0xB68FAE21D86F266DULL, "vxht_mansion_n_623_0_0.vlm"},
	{0xB68FE0468B948525ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_noise2_n.dds"},
	{0xB6907141FE2885C9ULL, "vxw_bonepitmining_a2_1jn_0_0.vlm"},
	{0xB690EF982840D1A9ULL, "vxlt_warehouse_n_15c_0_0.vlm"},
	{0xB6910285EFB38829ULL, "mag101_grace6034046.cl"},
	{0xB69113A01DBD7FB7ULL, "594016648.ogg"},
	{0xB6918ECAE0B40005ULL, "vxht_mansion2b_260_0_0.vlm"},
	{0xB691ED3A9C2E08C3ULL, "613596654.ogg"},
	{0xB69211249A4F594CULL, "lt_0_lowtown_a3_n_c_22_0.rml"},
	{0xB692219C77B0B5B1ULL, "vxlt_darktown_47r_0_0.vlm"},
	{0xB6929641F3B018ABULL, "hlac_lt_0_lowtown_docks_a3_n_-65536_0y.dds"},
	{0xB692BCDB7B254AACULL, "lt_blackemporium_1055_ll.mmh"},
	{0xB692BFC69B6A022EULL, "one220_yevhen6102825.cl"},
	{0xB692C110E846F894ULL, "vxht_mansion2_1ak_0_0.vlm"},
	{0xB693704514A81A47ULL, "vxlt_0_lowtown_n_e35_0_0.vlm"},
	{0xB6944CF0C58705A4ULL, "vxca_caverns1_mines_i27_0_0.vlm"},
	{0xB694D4165138D620ULL, "pro000_colonel_ewald6040556.cl"},
	{0xB694FB57E200E9DEULL, "297158543.ogg"},
	{0xB69501E254FB542BULL, "follower_banter6156000.cl"},
	{0xB69547F38CEF5E3DULL, "hlac_w_encounter2_65536_0z.dds"},
	{0xB695AE63C00B52B9ULL, "6048768_m.fxe"},
	{0xB695FB66AE3E4CB0ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_009a_0n.dds"},
	{0xB6967175B4AB06F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_loop.ani"},
	{0xB6969B0E41B8C837ULL, "vxht_chantry_n_2t2_0_1.vlm"},
	{0xB696A6E4B0E74EE9ULL, "hlac_lt_0_lowtown_a3_n_c_-3_4z.dds"},
	{0xB697BC7038CAB97BULL, "6163634_m.fxe"},
	{0xB698422C06B0E5CAULL, "vxca_cave_223_0_0.vlm"},
	{0xB6988E0884A749BCULL, "vxw_encounter2_camp_2b_0_0.vlm"},
	{0xB6991174C3C8E346ULL, "6113250_m.fxe"},
	{0xB6995F88D83018F6ULL, "follower_banter6155778.cl"},
	{0xB6998FDEF04CD943ULL, "718842302.ogg"},
	{0xB699F39C81A4AADCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasea_d.dds"},
	{0xB69A4F53478BF2C6ULL, "vxht_mansion2_25v_0_0.vlm"},
	{0xB69AE9236E1CBED3ULL, "6097047_m.fxe"},
	{0xB69B1419539D7321ULL, "hlac_ca_caverns2_5802_70y.dds"},
	{0xB69B2E1810BB3DEEULL, "vxht_mansion_n_3k2_0_0.vlm"},
	{0xB69B60EC4105EAD4ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf2a.phy"},
	{0xB69B719A89A8279FULL, "996544372.ogg"},
	{0xB69B8F46A48BBA2AULL, "mag210_arianni6030475.cl"},
	{0xB69BED0251788B77ULL, "cs_mcr310cs_scared_mage.bnk"},
	{0xB69C88E16B37A0C1ULL, "vxht_mansion2_2mu_0_0.vlm"},
	{0xB69CCB6662F61FE9ULL, "vxca_cavern1_haunted_fmx_0_0.vlm"},
	{0xB69D01E4F54FD5D2ULL, "383355513.ogg"},
	{0xB69D361392AA0B3EULL, "387171953.ogg"},
	{0xB69D4409DD055731ULL, "514143236.ogg"},
	{0xB69D7C15B693103EULL, "fen110_fenris6118085.cl"},
	{0xB69E6AE52277B149ULL, "6162248_m.fxe"},
	{0xB69EF646C8007C77ULL, "hlac_lt_0_lowtown_a2_131070_6x.dds"},
	{0xB69F0746A13D7AA3ULL, "29154035.ogg"},
	{0xB69F61A85399C8DDULL, "mer221_sabin6048392.cl"},
	{0xB69FF5BFE6C788D4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopea.phy"},
	{0xB6A0CCB4D67B89ECULL, "167051424.ogg"},
	{0xB6A0F1E3CB39ED79ULL, "vxlt_warehouse_n_po_0_0.vlm"},
	{0xB6A1A07CF9CABFEBULL, "6039390_f.fxe"},
	{0xB6A1D44AA8338028ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hf_injured_movement.anb"},
	{0xB6A25B6302F1CC0EULL, "6166720_m.fxe"},
	{0xB6A29BAFD69EAF3EULL, "vxca_caverns1_mines_ghe_0_0.vlm"},
	{0xB6A2B50C2C7598ACULL, "hla_ht_0_hightown_a3_-65535_0x.dds"},
	{0xB6A30C36DC78CC98ULL, "ser191_emeric6107126.cl"},
	{0xB6A3405FA01752D8ULL, "hlac_ht_mansion2_818_44z.dds"},
	{0xB6A372F8F211C482ULL, "6016503_f.fxe"},
	{0xB6A3FD9E7771387CULL, "vxlt_0_lowtown_n_f4x_0_0.vlm"},
	{0xB6A49063C5BD5068ULL, "6081366_m.fxe"},
	{0xB6A4B6F41F62873BULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_spiderweb4_da.dds"},
	{0xB6A4CDB149A55142ULL, "173085662.ogg"},
	{0xB6A5C4EF7B8158EBULL, "vxw_sundermount_2bv_0_0.vlm"},
	{0xB6A61618F38873D2ULL, "vxca_cavern1_haunted_jlp_0_0.vlm"},
	{0xB6A63DEC410FB6D1ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf2a.msh"},
	{0xB6A643AB8F183D96ULL, "one140cr_charade6110319.cl"},
	{0xB6A64B19C62A7310ULL, "terrain_blend8.psh.fxo"},
	{0xB6A657EC410FE28FULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf2a.mmh"},
	{0xB6A6728715DA7CC8ULL, "w_bonepitdestroyed_a3_38_0.rml"},
	{0xB6A6809766CADD5FULL, "vxw_bonepitdestroyed_a3_1w5_0_0.vlm"},
	{0xB6A6A1717460345FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_carddeck.msh"},
	{0xB6A6A77174603E01ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_carddeck.mmh"},
	{0xB6A6B371746052A2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_carddeck.mao"},
	{0xB6A6C50D8CA0C638ULL, "hlac_lt_darktown_-1_11z.dds"},
	{0xB6A6E9E6CECCF8C3ULL, "6109098_m.fxe"},
	{0xB6A6F3B51641A498ULL, "gen00fl_anders6110116.cl"},
	{0xB6A7DB7153F4BC3EULL, "art\\vfx\\environment\\vfx_trap_pot_p.dura.ani"},
	{0xB6A814015E2BF597ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_conj_c.impa.gad"},
	{0xB6A8620A21C59267ULL, "vxht_bartrand_n_4f0_0_0.vlm"},
	{0xB6A8968D597A7477ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_diagonal_1.mao"},
	{0xB6A8A597DDEA2503ULL, "vxlt_0_lowtown_docks_a3_cyq_0_1.vlm"},
	{0xB6A8B131A6FF17D1ULL, "115620086.ogg"},
	{0xB6A8CE70BD6EBE43ULL, "826954768.ogg"},
	{0xB6A928F997E8740FULL, "6164778_m.fxe"},
	{0xB6A951B7C44D2704ULL, "vxca_cavern1_haunted_gk2_0_0.vlm"},
	{0xB6A9B1569275F4A5ULL, "6148752_m.fxe"},
	{0xB6A9FD06BAE02B04ULL, "gen_im_trs_krk_hvy_hlm_r2.uti"},
	{0xB6AA157C15CFF9AAULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_sika_0n.dds"},
	{0xB6AA7044A28DBDE0ULL, "6111166_m.fxe"},
	{0xB6AAD2BFE6D154D1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopea.msh"},
	{0xB6AAE0B9D1216976ULL, "vxca_caverns1_cka_0_0.vlm"},
	{0xB6AAE0BFE6D16CECULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopea.mao"},
	{0xB6AAE1B09557CBBAULL, "vxdr_primevaltomb_8ya_0_0.vlm"},
	{0xB6AAEC06E29700A1ULL, "187500431.ogg"},
	{0xB6AAECBFE6D1808FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo33_slopea.mmh"},
	{0xB6AAFABE5D2DB941ULL, "vxht_mansion2_2vt_0_0.vlm"},
	{0xB6AB76CD3226628BULL, "vxca_caverns2prc_5xm_0_0.vlm"},
	{0xB6AB966C11A39943ULL, "6167991_m.fxe"},
	{0xB6ABD889929D099EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_stagger_1.ani"},
	{0xB6AC0FEEF956DBA3ULL, "6172269_m.fxe"},
	{0xB6AC36E502D22918ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_0z.dds"},
	{0xB6ACBAA5C6799398ULL, "vxht_keep_63n_0_0.vlm"},
	{0xB6AD3116DA1216B1ULL, "dae000_friend_turn_in6179272.cl"},
	{0xB6AD44CD5172E94DULL, "6218600_m.fxe"},
	{0xB6AD6E8F5D648739ULL, "dae000_meredith6092593.cl"},
	{0xB6AD74D7DD316C3AULL, "6054047_m.fxe"},
	{0xB6AD75E93A1B5989ULL, "vxca_caverns2_865_0_0.vlm"},
	{0xB6ADBFE0436BEFE8ULL, "6214634_m.fxe"},
	{0xB6ADC23D073A7F0CULL, "vxht_0_hightown_a2_bal_0_0.vlm"},
	{0xB6ADC77017FA2471ULL, "vxca_cavern1_haunted_j54_0_0.vlm"},
	{0xB6ADD39F45B7419AULL, "follower_banter6167043.cl"},
	{0xB6AE43763BA8D3E4ULL, "hlac_lt_0_lowtown_65536_2x.dds"},
	{0xB6AE52F1514F2E0DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmle.mmh"},
	{0xB6AE56F1514F355EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmle.mao"},
	{0xB6AE64F1514F4C93ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmle.msh"},
	{0xB6AEF5B59F023BF3ULL, "dae000_gen_player_turn_in6177829.cl"},
	{0xB6AF29015E323824ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_conj_c.impa.evt"},
	{0xB6AF5E4FF2FCED30ULL, "6208349_m.fxe"},
	{0xB6AF63C0BD3005F9ULL, "vxlt_warehouse_n_wn_0_0.vlm"},
	{0xB6AFC9FBC3B34B3CULL, "vxca_caverns1_j4c_0_0.vlm"},
	{0xB6B016C7AEFD868FULL, "464579587.ogg"},
	{0xB6B0AF1127C5FC19ULL, "349124019.ogg"},
	{0xB6B0C2037FFD191BULL, "gen00fl_anders6119305.cl"},
	{0xB6B19FF764BB52D6ULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_pcoverride.anb"},
	{0xB6B20D8668ADBAF3ULL, "ser120_ghyslain6118433.cl"},
	{0xB6B3B77D9A69C7A6ULL, "vxht_keep_n_6fd_0_0.vlm"},
	{0xB6B405F735ECAA74ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightr.phy"},
	{0xB6B471BA9BD6FDE0ULL, "315242672.ogg"},
	{0xB6B4F27458C0F5DFULL, "vxw_bonepitdestroyed_a3_1x4_0_0.vlm"},
	{0xB6B4F37153FF6F0FULL, "art\\vfx\\environment\\vfx_trap_pot_p.dura.evt"},
	{0xB6B4FE763F6DC9D3ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.impa.evt"},
	{0xB6B54CCCAE589D02ULL, "vxw_bonepitdestroyed_a3_8f_0_0.vlm"},
	{0xB6B6CEDCB46DDC3EULL, "461867561.ogg"},
	{0xB6B757251C0C24E7ULL, "6027552_m.fxe"},
	{0xB6B767D1EDF53DFDULL, "lgt311_shw_mer_intro6156408.cl"},
	{0xB6B7A8FCB05A9FEBULL, "6096022_m.fxe"},
	{0xB6B7CC6EF339F61DULL, "6113132_m.fxe"},
	{0xB6B80BD0AC9995A7ULL, "mag121_idunna6084017.cl"},
	{0xB6B8226709B25079ULL, "6074427_f.fxe"},
	{0xB6B82B557F54FE6AULL, "663366978.ogg"},
	{0xB6B8C8C8B88D9755ULL, "vxw_woundedcoast_13d_0_0.vlm"},
	{0xB6B98058BB3557B8ULL, "6118217_m.fxe"},
	{0xB6B9A93504E58B91ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothc_1.mao"},
	{0xB6BA2ADD7BC354B5ULL, "vxca_caverns1_mines_fml_0_0.vlm"},
	{0xB6BA3A3345DC90A7ULL, "llac_lt_darktown_-1_30.dds"},
	{0xB6BA49E97A0F99F5ULL, "verticalblurshadows.vsh.fxo"},
	{0xB6BA5A8F28F33D1EULL, "6171809_m.fxe"},
	{0xB6BB19D2637E780AULL, "projectile.ncs"},
	{0xB6BB3292DABBCDF4ULL, "53445148.ogg"},
	{0xB6BB9AE77A4C8448ULL, "dae000cr_fake_beth.utc"},
	{0xB6BBF47154058FB8ULL, "art\\vfx\\environment\\vfx_trap_pot_p.dura.gad"},
	{0xB6BBF7763F73DC14ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.impa.gad"},
	{0xB6BC1C5E2B2D88B7ULL, "6173868_m.fxe"},
	{0xB6BC41015E3CEB1DULL, "art\\vfx\\combat\\vfx_c_ogr_rock_conj_c.impa.ani"},
	{0xB6BCDD8AB74C1526ULL, "art\\levels\\races\\proxy\\_textures\\default_black.dds"},
	{0xB6BD28E92A69D2EEULL, "6132888_m.fxe"},
	{0xB6BD2F067CC3307FULL, "167998845.ogg"},
	{0xB6BD59571F590895ULL, "hlac_w_encounter2_262144_0z.dds"},
	{0xB6BDD59DAF261E79ULL, "vxlt_warehouse_n_zm_0_0.vlm"},
	{0xB6BDE2F735F4C371ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightr.msh"},
	{0xB6BDF0F735F4DA8CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightr.mao"},
	{0xB6BDFCF735F4EF2FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_straightr.mmh"},
	{0xB6BE59E56BDB3227ULL, "hlac_ca_caverns1_mines_24426_9x.dds"},
	{0xB6BE699F1DB29A73ULL, "gen00fl_anders6128374.cl"},
	{0xB6BEA9FFC938FC82ULL, "vxdr_primevalentrance_39o_0_1.vlm"},
	{0xB6BEEF41778A143DULL, "126800071.ogg"},
	{0xB6BEFBAFE3B453B1ULL, "vxca_cave_1o5_0_1.vlm"},
	{0xB6BF4BF1515D95CAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmle.phy"},
	{0xB6BF7942484BF336ULL, "hlac_lt_0_lowtown_n_-131073_3y.dds"},
	{0xB6BFF18992ADDD18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_stagger_1.gad"},
	{0xB6C02780C0B36141ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\enrage.ani"},
	{0xB6C0845000A90754ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_003a_0t.dds"},
	{0xB6C0A15142FA720FULL, "hlac_lt_0_lowtown_a2_n_c_1_11x.dds"},
	{0xB6C0E8A233B4D712ULL, "6129185_m.fxe"},
	{0xB6C13B99C6F65DF4ULL, "hlac_w_bonepitmining_a2_196612_0y.dds"},
	{0xB6C15D7E6FE33788ULL, "lt_0_lowtown_docks_n_182_0.msh"},
	{0xB6C16F7E6FE35619ULL, "lt_0_lowtown_docks_n_182_0.mao"},
	{0xB6C17B7E6FE36AB2ULL, "lt_0_lowtown_docks_n_182_0.mmh"},
	{0xB6C1D60505961E07ULL, "llac_lt_0_lowtown_a2_n_-1_3.dds"},
	{0xB6C20967E400ED2AULL, "one110_tree_minder6140860.cl"},
	{0xB6C2295A8C5FE026ULL, "vxht_keep_n_6ae_0_0.vlm"},
	{0xB6C273F4D1710989ULL, "vxdr_primevaltomb_3pb_0_0.vlm"},
	{0xB6C27C3BB564A3A8ULL, "vxlt_0_lowtown_docks_n_b5k_0_1.vlm"},
	{0xB6C2A4162A9579E8ULL, "mag160_feynriel6033397.cl"},
	{0xB6C2C2F4853A2FBDULL, "gen00fl_dog6107341.cl"},
	{0xB6C2FDB1A5F42FCDULL, "461603805.ogg"},
	{0xB6C306763F7A14F2ULL, "art\\vfx\\combat\\vfx_d_qun_aoe_p.impa.ani"},
	{0xB6C311C5DCFB41E7ULL, "6075151_m.fxe"},
	{0xB6C329758F39172AULL, "6170442_m.fxe"},
	{0xB6C331D62747203AULL, "llac_lt_undercityserialkiller_-1_31.dds"},
	{0xB6C36E026F911F98ULL, "vxlt_gallowstemplar_f_43x_0_0.vlm"},
	{0xB6C39E3DE27DA452ULL, "6070792_f.fxe"},
	{0xB6C3AB90AC3461C4ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumbig2_d.dds"},
	{0xB6C43D75691B5DD7ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgtb_rendermesh.mmh"},
	{0xB6C4454072ED281FULL, "541266526.ogg"},
	{0xB6C45375691B8349ULL, "art\\characters\\playercharacter\\humanmale\\hm_glv_lgtb_rendermesh.msh"},
	{0xB6C4F85253AFF277ULL, "one140cr_charade6128489.cl"},
	{0xB6C5087F927277EDULL, "6002630_m.fxe"},
	{0xB6C54E15DCBBA138ULL, "vxca_cavern1_haunted_j26_0_2.vlm"},
	{0xB6C550FBA7250512ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\enter.gad"},
	{0xB6C55146C8E95FFEULL, "vxca_caverns1_mines_e50_0_0.vlm"},
	{0xB6C5FA94C0000F5AULL, "lt_0_lowtown_docks_a3_n_190_0.msh"},
	{0xB6C60894C0002693ULL, "lt_0_lowtown_docks_a3_n_190_0.mao"},
	{0xB6C60C94C0002DA0ULL, "lt_0_lowtown_docks_a3_n_190_0.mmh"},
	{0xB6C60FFA3B599AB1ULL, "vxdr_0_deeproads_gfv_0_0.vlm"},
	{0xB6C6CA389765DEAEULL, "qun211_braggart6075692.cl"},
	{0xB6C6D1FF99040B5CULL, "6156251_m.fxe"},
	{0xB6C73280C0B992E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\enrage.gad"},
	{0xB6C786A3A18E3BD8ULL, "vxlt_0_lowtown_n_gpi_0_0.vlm"},
	{0xB6C78C185A9FE53AULL, "vxht_chantry_n_358_0_0.vlm"},
	{0xB6C817212CFDB321ULL, "vxca_cavern1_haunted_jdg_0_0.vlm"},
	{0xB6C84395A875303EULL, "llac_w_bonepitdestroyed_a3_393221_0.dds"},
	{0xB6C850DD13177E5BULL, "art\\levels\\ztools\\default\\generic\\8gr_basic.phy"},
	{0xB6C90683FDBFC5A9ULL, "vfx_distortionmask_p3.vsh.fxo.dx11"},
	{0xB6C95C47CC6C1E2DULL, "459277371.ogg"},
	{0xB6C984BC1D0E1B94ULL, "6116237_m.fxe"},
	{0xB6C99E5FAA65C498ULL, "vxht_keep_69l_0_0.vlm"},
	{0xB6C9CD7040EB9503ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_excited.ani"},
	{0xB6CA38B5C600E7C9ULL, "vxw_bonepitmining_a2_1vr_0_0.vlm"},
	{0xB6CA8F653FD72DD7ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masa.phy"},
	{0xB6CBC8471553622AULL, "gen_im_trs_gen_lgt_glv_r2.uti"},
	{0xB6CBDAC4A8C3F38AULL, "6127565_m.fxe"},
	{0xB6CC2A96A3E34C5BULL, "vxw_bonepitdestroyed_a3_26m_0_0.vlm"},
	{0xB6CC33E6A07F5263ULL, "propalpha1_s1y.psh.fxo"},
	{0xB6CC53FBA72B2949ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\enter.evt"},
	{0xB6CC7842C0E02435ULL, "genip_exploding_barrel_huge.utp"},
	{0xB6CD33A2D9AF9FB2ULL, "mer141st_jansen.stg"},
	{0xB6CDAB2214D9483FULL, "6125496_m.fxe"},
	{0xB6CDDB007C945B71ULL, "lowtown_male_beggar_sta6151033.cl"},
	{0xB6CDDCC67C9EBCDCULL, "art\\characters\\playercharacter\\humanfemale\\boo_medc.mao"},
	{0xB6CDE8C67C9ED17FULL, "art\\characters\\playercharacter\\humanfemale\\boo_medc.mmh"},
	{0xB6CDEEC67C9EDB21ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medc.msh"},
	{0xB6CE3AAFCADB3AB3ULL, "6208948_m.fxe"},
	{0xB6CE44B42E7360AEULL, "6118686_m.fxe"},
	{0xB6CE70FE940432B6ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bloomdesk.msh"},
	{0xB6CE7AFE940443B4ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bloomdesk.mmh"},
	{0xB6CE7EFE94044AC7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bloomdesk.mao"},
	{0xB6CECB798A64242EULL, "hla_ht_0_hightown_a3_65538_0z.dds"},
	{0xB6CF236EBAB54B19ULL, "gen_im_acc_rng_act2_09.uti"},
	{0xB6CFEAB377D44C6EULL, "vxlt_hangedman_n_1i8_0_0.vlm"},
	{0xB6D0087F6323D2ECULL, "vxht_0_hightown_a3_ajo_0_0.vlm"},
	{0xB6D00FA9FC6738F6ULL, "vxlt_0_lowtown_docks_b7t_0_1.vlm"},
	{0xB6D105EA741D7225ULL, "hlac_lt_0_lowtown_a2_-1_6z.dds"},
	{0xB6D1618E40D4CD68ULL, "6108713_m.fxe"},
	{0xB6D1F1849702D13AULL, "llac_lt_0_lowtown_docks_a3_-65538_1.dds"},
	{0xB6D2032DDFA9D375ULL, "hlac_ca_caverns2_5802_38y.dds"},
	{0xB6D2B8B9E03D4BA5ULL, "323654790.ogg"},
	{0xB6D32195CBF5A6DEULL, "vxlt_warehouse_hy_0_0.vlm"},
	{0xB6D336705EF19624ULL, "983866464.ogg"},
	{0xB6D38548713AA785ULL, "fen110_fenris6118022.cl"},
	{0xB6D38F13F49C19E5ULL, "llac_lt_0_lowtown_131070_6.dds"},
	{0xB6D3F944BF71D354ULL, "hlac_lt_0_lowtown_n_0_1y.dds"},
	{0xB6D54B2E573AB2C6ULL, "6118317_f.fxe"},
	{0xB6D62A863A344F61ULL, "hlac_lt_0_lowtown_a3_n_c_65536_12x.dds"},
	{0xB6D62AD1E8FA814FULL, "vxw_woundedcoast_1dc_0_0.vlm"},
	{0xB6D690DEBF9B18F4ULL, "6095176_m.fxe"},
	{0xB6D6E57040F648FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_excited.evt"},
	{0xB6D7545A13ADCBA1ULL, "6090237_m.fxe"},
	{0xB6D75E3B4818B90DULL, "139329534.ogg"},
	{0xB6D7E85FF9045105ULL, "120729811.ogg"},
	{0xB6D825B6CC610548ULL, "6123393_f.fxe"},
	{0xB6D85F3C47DF7E76ULL, "mag320_orsino6070442.cl"},
	{0xB6D8A8E6330F96A8ULL, "follower_banter6155673.cl"},
	{0xB6D919FC419057F2ULL, "60878669.ogg"},
	{0xB6D97BE3CC8FBD53ULL, "62772423.ogg"},
	{0xB6D9CF35AF92843FULL, "vxlt_gallowstemplar_4yj_0_0.vlm"},
	{0xB6D9E1F7A0EBBE77ULL, "vxlt_hangedman_n_1r8_0_0.vlm"},
	{0xB6DA3BFBA7373DB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\enter.ani"},
	{0xB6DA52F1F724A3AAULL, "lgt151ip_sundermount.utp"},
	{0xB6DA61F2AB29630DULL, "art\\characters\\playercharacter\\humanfemale\\hf_hoo_robb.mao"},
	{0xB6DA765B940C76FFULL, "6162281_f.fxe"},
	{0xB6DA9C7395D964DBULL, "vxw_bonepitdestroyed_a3_21l_0_0.vlm"},
	{0xB6DB3F5C39AFE341ULL, "follower_barks6173398.cl"},
	{0xB6DB628AB3C33D2DULL, "vxca_caverns1_mines_eh0_0_0.vlm"},
	{0xB6DBCF5865EF5D0EULL, "vxlt_0_lowtown_docks_a3_c3w_0_0.vlm"},
	{0xB6DBF94927780473ULL, "plt_mer210pt_pickaxes.plo"},
	{0xB6DC0CACFB8A84F6ULL, "vrc221_varric6045036.cl"},
	{0xB6DC10A9DDD66F0EULL, "417664943.ogg"},
	{0xB6DC158D77EF0CE9ULL, "703081648.ogg"},
	{0xB6DCADBDAC29FE9AULL, "765808895.ogg"},
	{0xB6DCB2260621461DULL, "603455947.ogg"},
	{0xB6DCC21346D44851ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzvfxsign.mmh"},
	{0xB6DCCE1346D45CF2ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzvfxsign.mao"},
	{0xB6DCDC1346D4740FULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzvfxsign.msh"},
	{0xB6DD0FCD55EB414EULL, "23579723.ogg"},
	{0xB6DDE47040FC65EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_pacing_excited.gad"},
	{0xB6DDF3F3A223E44BULL, "220520518.ogg"},
	{0xB6DE0DF893D7EA3DULL, "6069743_m.fxe"},
	{0xB6DE47905C5D58DCULL, "hlac_lt_0_lowtown_a2_n_c_-65535_3z.dds"},
	{0xB6DED2D4E1D44A69ULL, "81282954.ogg"},
	{0xB6DF00794B35C4B8ULL, "vxca_caverns1_mines_e7l_0_0.vlm"},
	{0xB6DF2463FC22EE43ULL, "359582249.ogg"},
	{0xB6DF53F606E87346ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_r_5p.evt"},
	{0xB6DF56F69A5A8256ULL, "256407431.ogg"},
	{0xB6E0308393797C6CULL, "vxht_0_hightown_cbd_0_0.vlm"},
	{0xB6E0480B2EA33F5FULL, "vxw_bonepitdestroyed_a3_1k1_0_0.vlm"},
	{0xB6E06C0EBADFE289ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he3a.phy"},
	{0xB6E17868EE71EDEDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_floordirt_sa.dds"},
	{0xB6E1B726B4379B5CULL, "w_sundermount_246_0.mao"},
	{0xB6E1C326B437AFFFULL, "w_sundermount_246_0.mmh"},
	{0xB6E1C926B437B9A1ULL, "w_sundermount_246_0.msh"},
	{0xB6E236327124F403ULL, "lla_ht_0_hightown_n_-65536_0.dds"},
	{0xB6E25DDA11507CFBULL, "vxlt_gallowstemplar_55o_0_0.vlm"},
	{0xB6E29E9A59DA43C2ULL, "6167023_f.fxe"},
	{0xB6E3F80BCE928283ULL, "vxlt_0_lowtown_docks_c6n_0_0.vlm"},
	{0xB6E40D10CDDA2E1FULL, "6065741_m.fxe"},
	{0xB6E45EDA8888E455ULL, "6084742_m.fxe"},
	{0xB6E49CAEDAF099CFULL, "vxw_woundedcoast_1ob_0_0.vlm"},
	{0xB6E4A6852F2FE22EULL, "vxht_mansion2b_24i_0_0.vlm"},
	{0xB6E4F777AC1BAAA6ULL, "915429350.ogg"},
	{0xB6E524F6EF98FB16ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_180.evt"},
	{0xB6E5B7C2DD0B3366ULL, "art\\vfx\\creatures\\vfx_c_drag_rake_c.impa.ani"},
	{0xB6E5B80903653076ULL, "vxw_encounter2_cc_0_0.vlm"},
	{0xB6E63187CC308A18ULL, "6076056_m.fxe"},
	{0xB6E66AF606EEB891ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_r_5p.gad"},
	{0xB6E687D187FD5985ULL, "bla_sh.mao"},
	{0xB6E69793BE571DC8ULL, "6128293_f.fxe"},
	{0xB6E6A51970BDF366ULL, "vxht_chantry_n_31x_0_0.vlm"},
	{0xB6E6C74BD574B26CULL, "6110293_m.fxe"},
	{0xB6E70D1A2FA841DEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_camp_fire.ani"},
	{0xB6E773382EBE7247ULL, "dae000_arishok6068566.cl"},
	{0xB6E7CD8095850E02ULL, "233223902.ogg"},
	{0xB6E82CA1A3599FECULL, "gen00fl_merrill6090265.cl"},
	{0xB6E93813B0373BD1ULL, "gen_im_wep_rog_dua_act2_06.uti"},
	{0xB6EA5335E80412F0ULL, "vxdr_0_deeproads_f66_0_0.vlm"},
	{0xB6EA55141F6A6DBAULL, "100072752.ogg"},
	{0xB6EA8423C97221E5ULL, "612550620.ogg"},
	{0xB6EAB3013F2D95F3ULL, "6185133_m.fxe"},
	{0xB6EAE6BD827E9DE1ULL, "dae000_tranquil_16165110.cl"},
	{0xB6EB14EB4780A365ULL, "6188563-sheevenwantsthat.cut"},
	{0xB6EB28A3E5560C6DULL, "vxca_cave_1nc_0_0.vlm"},
	{0xB6EB9731DD85FC83ULL, "animblend_core.fdl"},
	{0xB6EB9A024BF6C9F3ULL, "mcr370_negotiation6078758.cl"},
	{0xB6EC1BF6EF9F0AA1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_180.gad"},
	{0xB6EC2460671CF8C5ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree01_n.dds"},
	{0xB6EC4E4675A1DCBFULL, "vxht_0_hightown_a3_ama_0_0.vlm"},
	{0xB6EC50291325A9E2ULL, "vxlt_hangedman_n_1jy_0_0.vlm"},
	{0xB6EC537E7A8734CFULL, "6069402_m.fxe"},
	{0xB6ECA2CD0A65C3DDULL, "6074976_m.fxe"},
	{0xB6ED46D76DB2853CULL, "vxlt_0_lowtown_docks_n_b6j_0_1.vlm"},
	{0xB6ED6BF606F4D93FULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_r_5p.ani"},
	{0xB6EDB9E4285EE70CULL, "follower_banter6147735.cl"},
	{0xB6EE89061F267282ULL, "ser221_party_comments6116131.cl"},
	{0xB6EEB9E8209957DFULL, "vxw_bonepitdestroyed_a3_1l0_0_0.vlm"},
	{0xB6EEC9A9FE20ABA7ULL, "556859306.ogg"},
	{0xB6EF25F24137F47DULL, "follower_banter6148711.cl"},
	{0xB6EFD8DC92ACDB57ULL, "vxlt_undercityencounter_1e9_0_0.vlm"},
	{0xB6F02250EF9296C0ULL, "6155751_m.fxe"},
	{0xB6F035625DB89D59ULL, "vxlt_0_lowtown_docks_arz_0_0.vlm"},
	{0xB6F04F374DA699D3ULL, "346358749.ogg"},
	{0xB6F0BBDD0CB65D23ULL, "follower_banter6163750.cl"},
	{0xB6F0D0DB1B27F034ULL, "6017488_m.fxe"},
	{0xB6F15A66203FF0F6ULL, "vxlt_0_lowtown_a3_n_c_hf9_0_0.vlm"},
	{0xB6F1B701A8DCC110ULL, "vxht_chantry_n_2q3_0_0.vlm"},
	{0xB6F1C63E76C9D6C1ULL, "follower_banter6172930.cl"},
	{0xB6F21425DA80D0FEULL, "vxlt_hangedman_269_0_0.vlm"},
	{0xB6F22EE83B26E3C6ULL, "6112117_f.fxe"},
	{0xB6F2B636D2F8262CULL, "vxlt_0_lowtown_a3_n_c_dej_0_0.vlm"},
	{0xB6F2B9FE837FDDB1ULL, "w_encounter3_213_0.msh"},
	{0xB6F2C7FE837FF5CCULL, "w_encounter3_213_0.mao"},
	{0xB6F2D3FE83800A6FULL, "w_encounter3_213_0.mmh"},
	{0xB6F2FDD13F28620CULL, "vxlt_0_lowtown_a2_hfc_0_0.vlm"},
	{0xB6F30E8BCCE6B24FULL, "vxw_woundedcoast_1jm_0_0.vlm"},
	{0xB6F318622125FAAEULL, "vxht_mansion2b_21j_0_0.vlm"},
	{0xB6F31ABBA05021DFULL, "52108708.ogg"},
	{0xB6F31CF6EFA52B4FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_180.ani"},
	{0xB6F328B047A69685ULL, "zz_approval_debug6003792.cl"},
	{0xB6F35832DBC64858ULL, "pro000_friedrich6051048.cl"},
	{0xB6F39FC2DD1747D7ULL, "art\\vfx\\creatures\\vfx_c_drag_rake_c.impa.evt"},
	{0xB6F3BB6CE4F4F5D2ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_1z.dds"},
	{0xB6F4251A2FB2F4AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_camp_fire.evt"},
	{0xB6F429E5F55B48F6ULL, "vxw_encounter2_nd_0_0.vlm"},
	{0xB6F437DD133C8EFAULL, "art\\levels\\ztools\\default\\generic\\8gr_basic.msh"},
	{0xB6F445DD133CA633ULL, "art\\levels\\ztools\\default\\generic\\8gr_basic.mao"},
	{0xB6F449DD133CAD40ULL, "art\\levels\\ztools\\default\\generic\\8gr_basic.mmh"},
	{0xB6F4A28015A0F85DULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st104a_0s.dds"},
	{0xB6F57101097F3EB8ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.anb"},
	{0xB6F59A06566999D6ULL, "zz_followers_banter6158511.cl"},
	{0xB6F5BA8878B4D4C9ULL, "661773850.ogg"},
	{0xB6F5BCC6A9699D34ULL, "1034754497.ogg"},
	{0xB6F605FE7C9535A9ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.dura.ani"},
	{0xB6F613DF5B344B95ULL, "vxdr_0_deeproads_f5z_0_0.vlm"},
	{0xB6F67B0627335D0DULL, "6084547_f.fxe"},
	{0xB6F69A4855BF503EULL, "832487794.ogg"},
	{0xB6F756F790162D84ULL, "dae000ip_crowd_ff8_talk3.utp"},
	{0xB6F7881652DF8CAAULL, "936792020.ogg"},
	{0xB6F7A767DC189BD8ULL, "647603583.ogg"},
	{0xB6F7B2FB95791DDEULL, "6033642_m.fxe"},
	{0xB6F8C33ABA24AF5DULL, "mrl150_merrill6173799.cl"},
	{0xB6F94970B5A2087EULL, "hlac_lt_blackemporium_1055_10y.dds"},
	{0xB6F96DBA8F184082ULL, "loginconnected.gfx"},
	{0xB6FAA0C2DD1D6860ULL, "art\\vfx\\creatures\\vfx_c_drag_rake_c.impa.gad"},
	{0xB6FABD0831232C64ULL, "374688170.ogg"},
	{0xB6FB261A2FB91558ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_camp_fire.gad"},
	{0xB6FB7B982574578EULL, "117015782.ogg"},
	{0xB6FB9CCB97BDA85DULL, "llac_w_blightlands_-131071_0.dds"},
	{0xB6FC0EBC183EBA0FULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_c.mmh"},
	{0xB6FC3B6899496B89ULL, "vxdr_primevaltomb_3df_0_0.vlm"},
	{0xB6FD20FE7C9B822BULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.dura.gad"},
	{0xB6FD3CA7FE0A591AULL, "dre000_dougal6129785.cl"},
	{0xB6FD585305B15196ULL, "mag170_hugh6080820.cl"},
	{0xB6FEA39B90A4BCDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_haste.gad"},
	{0xB6FEC79711147A6BULL, "art\\characters\\playercharacter\\humanmale\\arm_urvb.mao"},
	{0xB6FEEAA7404CCC77ULL, "76196936.ogg"},
	{0xB6FF11B0F6BFC544ULL, "w_encounter2_camp_5_0_ll.mmh"},
	{0xB6FF401DD7CF995AULL, "233423092.ogg"},
	{0xB6FF8E62F06D598BULL, "vxdr_0_deeproads_a27_0_1.vlm"},
	{0xB70048297721DEF2ULL, "6163815_m.fxe"},
	{0xB70102CFA62E22B1ULL, "vxlt_warehouse_n_cr_0_0.vlm"},
	{0xB7013F502621F482ULL, "6139210_m.fxe"},
	{0xB70163250AB64DFAULL, "llac_dr_primevaltomb_661_3.dds"},
	{0xB7016BDE109579A4ULL, "gen_im_trs_drk_two_2hs_r0.uti"},
	{0xB7017893757D2EBEULL, "6138737_f.fxe"},
	{0xB70190BE7D5439DCULL, "126242575.ogg"},
	{0xB701DE6AB6C3FCFCULL, "vxw_bonepit_7u_0_0.vlm"},
	{0xB701E5B3EDC655BFULL, "665790983.ogg"},
	{0xB7020DBBD1AEBE05ULL, "gen00fl_aveline6053280.cl"},
	{0xB70224F249022417ULL, "vxca_caverns2prc_9ur_0_0.vlm"},
	{0xB7029BC2E7516176ULL, "vxw_encounter2_ie_0_0.vlm"},
	{0xB702D38632E33BDEULL, "6120798_f.fxe"},
	{0xB702E8F7D59A1E3DULL, "gen00fl_varric6076077.cl"},
	{0xB7031DFE7C9FE890ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.dura.evt"},
	{0xB703489CCE95F2A6ULL, "gen00fl_isabela_init6076622.cl"},
	{0xB703C21C3A775004ULL, "6208019_m.fxe"},
	{0xB704C662308E6127ULL, "vxdr_primevaltomb_6s4_0_0.vlm"},
	{0xB7052664D8442B74ULL, "vxlt_foundry_n_vf_0_0.vlm"},
	{0xB705989B90AAC8CCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_haste.evt"},
	{0xB705A498D015B7FBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_enter.evt"},
	{0xB705E02C30E81C4FULL, "6178454_m.fxe"},
	{0xB705F5C286B84F4DULL, "gen00fl_isabela6090877.cl"},
	{0xB7061F49EC44F3F0ULL, "286376475.ogg"},
	{0xB7062DF0FB37FD2FULL, "t3_dog_on2.tnt"},
	{0xB70652676CBD3D7AULL, "llac_w_bonepit_0_4.dds"},
	{0xB7066A950D4C873BULL, "vxca_cavern1_haunted_eb8_0_0.vlm"},
	{0xB70754ADCD2C6BE0ULL, "295249742.ogg"},
	{0xB7087E6E845DDEF4ULL, "vxlt_gallowstemplar_f_5rn_0_0.vlm"},
	{0xB708B1C8415360FFULL, "gen00fl_isabela6079265.cl"},
	{0xB708C4A7E460A55CULL, "t3_dey_am2.tnt"},
	{0xB708DEE95B675132ULL, "mrl000_marethari6129296.cl"},
	{0xB709B05EE64404E3ULL, "dae200_cullen6140284.cl"},
	{0xB70A41172D4B6AC8ULL, "hlac_dr_0_deeproads_17038_0y.dds"},
	{0xB70ABC82693CE2F4ULL, "aip_harvester_plot.gda"},
	{0xB70AE049279F4A10ULL, "plt_mer210pt_pickaxes.nss"},
	{0xB70AE82797ED9EBFULL, "vxht_0_hightown_cbk_0_0.vlm"},
	{0xB70B452A6D4FDE3DULL, "hlac_lt_0_lowtown_docks_n_262143_0z.dds"},
	{0xB70B91926325EBF4ULL, "cd_qun111_ketojan.wbd"},
	{0xB70C550EBB04F6AEULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he3a.mmh"},
	{0xB70C6F0EBB05229CULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he3a.msh"},
	{0xB70C9D98D01BCA3CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_enter.gad"},
	{0xB70CBB52DB4B143DULL, "336580194.ogg"},
	{0xB70CD25A2A636F35ULL, "6076880_m.fxe"},
	{0xB70CF64DCA274C67ULL, "ss_coterie_mage_m.wbd"},
	{0xB70D30693B6AF81EULL, "434248210.ogg"},
	{0xB70D3D329838C04BULL, "vxca_cavern1_haunted_imu_0_0.vlm"},
	{0xB70D688119327B5CULL, "639872818.ogg"},
	{0xB70DB1E6AFD54BBFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_brokewood.phy"},
	{0xB70E1AEB55FD0FE4ULL, "vxht_mansion_n_47g_0_0.vlm"},
	{0xB70E466B139297F6ULL, "llac_lt_0_lowtown_docks_-131075_6.dds"},
	{0xB70F129C57D32F2FULL, "vxca_caverns1_gif_0_0.vlm"},
	{0xB70F1EB947D73168ULL, "693962258.ogg"},
	{0xB70F46C8F6C5304FULL, "950294679.ogg"},
	{0xB70F7313F1C0F630ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_imp_c.mmh"},
	{0xB70FF245B0D2E34FULL, "vxw_woundedcoast_1po_0_0.vlm"},
	{0xB7100DA8DC754E9EULL, "mrl150pt_longway.ncs"},
	{0xB71060BAB8911FABULL, "mer320_hubert6053182.cl"},
	{0xB7106C672C3BBC78ULL, "631186371.ogg"},
	{0xB710CB702706A78AULL, "vxca_caverns2_7vy_0_0.vlm"},
	{0xB710ED68916B28A1ULL, "vxht_brothel_n_3qx_0_0.vlm"},
	{0xB7115EBF617DCF72ULL, "264403582.ogg"},
	{0xB711B465803FCBE3ULL, "vxca_cavern1_haunted_k2n_0_0.vlm"},
	{0xB711BDA1088118E4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wrba_0s.dds"},
	{0xB712212C6EA0F375ULL, "502170165.ogg"},
	{0xB7125CA8A1ED2EBFULL, "772022814.ogg"},
	{0xB712646540149A2FULL, "art\\characters\\playercharacter\\humanmale\\bdy_masa.mao"},
	{0xB71270654014AECCULL, "art\\characters\\playercharacter\\humanmale\\bdy_masa.mmh"},
	{0xB71276654014B8BEULL, "art\\characters\\playercharacter\\humanmale\\bdy_masa.msh"},
	{0xB712809B90B52A25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_haste.ani"},
	{0xB712A1FFC5BA2989ULL, "132701489.ogg"},
	{0xB712A7D31D6DE356ULL, "6108157_m.fxe"},
	{0xB7135E65FF0EB13BULL, "hlac_lt_0_lowtown_n_-3_3x.dds"},
	{0xB7139C98D021E73AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\arishok_collapse_enter.ani"},
	{0xB71426F323F33E53ULL, "follower_banter6141691.cl"},
	{0xB7150B9D861D4FF7ULL, "gen00fl_aveline6115500.cl"},
	{0xB716D70EEACC8A33ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch.phy"},
	{0xB716F04B7653F774ULL, "vxlt_gallowstemplar_f_5uo_0_0.vlm"},
	{0xB7175A2A04B50D28ULL, "w_woundedcoast_148_0.msh"},
	{0xB7176C2A04B52BB9ULL, "w_woundedcoast_148_0.mao"},
	{0xB717782A04B54052ULL, "w_woundedcoast_148_0.mmh"},
	{0xB71798BF5F1227FFULL, "6153417_m.fxe"},
	{0xB717E653BE38CE4AULL, "70888247.ogg"},
	{0xB717F6DCE353D38EULL, "hlac_w_bonepit_0_1y.dds"},
	{0xB718E568F29A6732ULL, "hlac_ht_0_hightown_n_65536_3z.dds"},
	{0xB7194D677C100DE8ULL, "llac_lt_0_lowtown_docks_a3_n_-65538_2.dds"},
	{0xB7194F3C769C6C8CULL, "l_lgt261ar_rescue_cave.bnk"},
	{0xB719540E629FAE3CULL, "1008074730.ogg"},
	{0xB719EF2F798356F1ULL, "99664500.ogg"},
	{0xB71A0C5AFD343EF7ULL, "gen00fl_aveline6139186.cl"},
	{0xB71A94E97199E622ULL, "ave200_donnic6110671.cl"},
	{0xB71A9619AF653DEEULL, "hlac_lt_0_lowtown_a2_65536_6x.dds"},
	{0xB71ABE7255851F3CULL, "50576436.ogg"},
	{0xB71B89AD616BFA7AULL, "6036542_m.fxe"},
	{0xB71BAF0F8A2ED8CBULL, "vxca_cavern1_haunted_ijt_0_0.vlm"},
	{0xB71C9613F1CBBC4FULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_imp_c.anb"},
	{0xB71CF0C54D10F73AULL, "dae200cr_viscount6087176.cl"},
	{0xB71D1428EDB7D9EEULL, "994165990.ogg"},
	{0xB71D2393AD3BDD5EULL, "435783590.ogg"},
	{0xB71D782B6925ADF7ULL, "fxc_puke_vol.dds"},
	{0xB71D7957B389DB60ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffb_sa.dds"},
	{0xB71D82B3712E6B7BULL, "6061395_f.fxe"},
	{0xB71D9232845906D2ULL, "6033798_m.fxe"},
	{0xB71E700109A24253ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.mmh"},
	{0xB71E9D488394F146ULL, "222571256.ogg"},
	{0xB71E9F109941AE5FULL, "vrc221_cassandra.bnk"},
	{0xB71EB2E6AFE3C0C4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_brokewood.mmh"},
	{0xB71EB6E6AFE3C717ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_brokewood.mao"},
	{0xB71EC8E6AFE3E666ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ww_brokewood.msh"},
	{0xB71F3D4D18FCC00AULL, "vxca_caverns2_7kz_0_0.vlm"},
	{0xB71F736582CDFC6BULL, "shi_link.mao"},
	{0xB71FCA20325BFB2DULL, "365286584.ogg"},
	{0xB7200135B3531F1DULL, "6208574_f.fxe"},
	{0xB7200B660B0891DBULL, "186041519.ogg"},
	{0xB721506CFC2D9051ULL, "hlac_lt_0_lowtown_docks_a3_-196609_0x.dds"},
	{0xB72181109943B167ULL, "vrc221_cassandra.cnv"},
	{0xB7218822874AA197ULL, "hlac_ca_caverns2prc_5802_48y.dds"},
	{0xB721ACF9917D0143ULL, "hlac_ca_caverns2_5802_46z.dds"},
	{0xB721DB4DDBAD5ACAULL, "hlac_lt_0_lowtown_a3_n_c_131070_6y.dds"},
	{0xB721EC07A666D80EULL, "vxht_playershouse_n_1vz_0_0.vlm"},
	{0xB7224326C1C2DA79ULL, "6125586_m.fxe"},
	{0xB7225828C51864A1ULL, "vxlt_backalley_g_1dd_0_0.vlm"},
	{0xB722D063C566A51BULL, "topple.cur"},
	{0xB72337C19E6BAD98ULL, "718651237.ogg"},
	{0xB723401464D2BA6EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_curvel_2.mao"},
	{0xB7235D09CA9222C9ULL, "326877992.ogg"},
	{0xB723713483EF2602ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small_trim1.gad"},
	{0xB7239349296AB151ULL, "dre000_varric6035224.cl"},
	{0xB72394D6C2EEC40EULL, "lt_0_lowtown_docks_a3_33_0_ll.mmh"},
	{0xB7240FBC18600E4CULL, "art\\vfx\\combat\\vfx_m_sta_ice_range_c.anb"},
	{0xB7244E609EE30D5EULL, "gen00fl_anders6095495.cl"},
	{0xB7248265D114B381ULL, "vxlt_undercitydungeon_1ns_0_0.vlm"},
	{0xB724BEB3A527F23EULL, "hlac_w_woundedcoast_-65531_0z.dds"},
	{0xB7256530825639B8ULL, "6115312_f.fxe"},
	{0xB72570465C2FEEAAULL, "175009637.ogg"},
	{0xB72591B0023F17B7ULL, "53642377.ogg"},
	{0xB725A4ED750C4464ULL, "518712215.ogg"},
	{0xB72657AE9889A104ULL, "gen00fl_aveline6139635.cl"},
	{0xB7265F4965AA865FULL, "vxdr_primevaltomb_8v1_0_0.vlm"},
	{0xB72694C563E39569ULL, "dae000_merchant_mage_hf6151496.cl"},
	{0xB72695EDA5EEF6BFULL, "lt_0_lowtown_n_water_21225.msh"},
	{0xB7269BEDA5EF0161ULL, "lt_0_lowtown_n_water_21225.mmh"},
	{0xB726A7EDA5EF1502ULL, "lt_0_lowtown_n_water_21225.mao"},
	{0xB72775F085750667ULL, "vxht_bartrand_n_72s_0_0.vlm"},
	{0xB728243B040CB6F6ULL, "hlac_lt_gallowstemplar_n_2388_1y.dds"},
	{0xB7284BE2EEEF78E6ULL, "6053015_m.fxe"},
	{0xB7284EF2C4FA8A8CULL, "640501662.ogg"},
	{0xB728F1E18AE2E533ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv1.evt"},
	{0xB7290D6E74543E01ULL, "hlac_ca_caverns1_24423_11y.dds"},
	{0xB729263143FC0C5EULL, "6209850_m.fxe"},
	{0xB7297020A1DC96B0ULL, "l_dae121ar_chantry.bnk"},
	{0xB7299C89A035E8CBULL, "character_character_p246.psh.fxo.dx11"},
	{0xB72A44C42B9502DDULL, "nat380_nathaniel6066611.cl"},
	{0xB72A79FE9452A0AFULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_bloomdesk.phy"},
	{0xB72A943483F580F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small_trim1.evt"},
	{0xB72A9D6DC2A9E04EULL, "6188584_m.fxe"},
	{0xB72AFF2BC9C4BA58ULL, "vxca_caverns2prc_7qo_0_0.vlm"},
	{0xB72B0841E9F00178ULL, "6089916_f.fxe"},
	{0xB72BD37EA3D57C39ULL, "pro000cs_flemeth_real_demo.cut"},
	{0xB72BE4EF7FC1F246ULL, "static_static_p3.vsh.fxo.dx11"},
	{0xB72C58667C106C31ULL, "vxlt_warehouse_n_pw_0_0.vlm"},
	{0xB72C618E758801A5ULL, "217886597.ogg"},
	{0xB72C7AF787DD0B0EULL, "gen00fl_anders6125576.cl"},
	{0xB72C8AC1CD931E41ULL, "776827617.ogg"},
	{0xB72CBD6D491AD4FCULL, "304227132.ogg"},
	{0xB72D0D4F80DC257DULL, "263910453.ogg"},
	{0xB72D1F14203AB834ULL, "6093853_m.fxe"},
	{0xB72D440AD5646BBFULL, "w_sundermount_125_0.msh"},
	{0xB72D4A0AD5647661ULL, "w_sundermount_125_0.mmh"},
	{0xB72D560AD5648A02ULL, "w_sundermount_125_0.mao"},
	{0xB72D6A849D948130ULL, "240088451.ogg"},
	{0xB72DD65BCB04EB83ULL, "fx_exp_fing_cloud.dds"},
	{0xB72E23664187D835ULL, "hlac_lt_darktown_-1_47x.dds"},
	{0xB72EDE2C8E292330ULL, "vxw_woundedcoast_1fj_0_0.vlm"},
	{0xB72EF83A56AF066DULL, "gen00fl_aveline6122805.cl"},
	{0xB72F0A26E6F32436ULL, "vxht_brothel_n_3f6_0_0.vlm"},
	{0xB72F752E150A4404ULL, "nat380_nathaniel6187719.cl"},
	{0xB72FB8BD9D198153ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_4.dds"},
	{0xB72FEAE18AE8F874ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv1.gad"},
	{0xB730502299C49B54ULL, "405504604.ogg"},
	{0xB730F1C67CF35DC4ULL, "art\\characters\\playercharacter\\humanfemale\\boo_medc.phy"},
	{0xB731364C488BC408ULL, "zz_followers_banter6157620.cl"},
	{0xB7313FC0A5D6D680ULL, "vxw_woundedcoast_1b6_0_0.vlm"},
	{0xB73207A5E338F000ULL, "gen00fl_anders6116099.cl"},
	{0xB732DBA8B946FEECULL, "vxlt_0_lowtown_n_ex7_0_0.vlm"},
	{0xB73353E43866CEABULL, "vxdr_primevaltomb_50o_0_1.vlm"},
	{0xB7336507E7034B09ULL, "vxht_brothel_n_4wi_0_0.vlm"},
	{0xB733D4055A402874ULL, "vxlt_gallowstemplar_f_5cm_0_0.vlm"},
	{0xB733F171C1226044ULL, "t3_wdg_bn2.tnt"},
	{0xB734221E14DB4DF0ULL, "6153920_m.fxe"},
	{0xB73433D6C9196EFCULL, "dae000_bartrand6027468.cl"},
	{0xB73469D67437F339ULL, "hlac_lt_0_lowtown_docks_65535_1z.dds"},
	{0xB734AE41D9226A34ULL, "gen00fl_anders6116107.cl"},
	{0xB734D12657A09EDFULL, "vxdr_primevaltomb_8q0_0_0.vlm"},
	{0xB734D3B50BAD56E3ULL, "fen310_fenris6185235.cl"},
	{0xB734F8DC3C936312ULL, "vxca_caverns1_mines_ikn_0_0.vlm"},
	{0xB7351D606F8D880FULL, "mrl350cr_dragon.utc"},
	{0xB735269BE00959E9ULL, "6167491_f.fxe"},
	{0xB735B8421C10A426ULL, "vxht_keep_n_69m_0_0.vlm"},
	{0xB736512D2B1DC2A3ULL, "816336059.ogg"},
	{0xB736F9E18AEF3052ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\attackv1.ani"},
	{0xB7377B79B5F7BB03ULL, "dae380st_nathaniel_final.stg"},
	{0xB7377F92634B078EULL, "cd_qun111_ketojan.bnk"},
	{0xB73872DAE6AEC4F3ULL, "541318559.ogg"},
	{0xB7388B9E2299669FULL, "vxht_keep_n_672_0_0.vlm"},
	{0xB7389C348401CBE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_small_trim1.ani"},
	{0xB738DB2B3FFB5D6BULL, "zz_qcr_debug6053501.cl"},
	{0xB7392962FCC14322ULL, "hlac_ca_caverns2_5802_12z.dds"},
	{0xB739C554058CAA09ULL, "6083088_m.fxe"},
	{0xB73AB73E1ADD6C5BULL, "dae100_viscount.bnk"},
	{0xB73AC569C7A35703ULL, "808886979.ogg"},
	{0xB73ACA436E0684B1ULL, "vxlt_warehouse_n_wv_0_0.vlm"},
	{0xB73B66D0BFAB6207ULL, "6169764_m.fxe"},
	{0xB73BDE109B3E2D86ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_strt.ani"},
	{0xB73C6C8AE75B9620ULL, "608095888.ogg"},
	{0xB73C9BA4CCC641D4ULL, "vxlt_backalley_g_1pp_0_1.vlm"},
	{0xB73CC3D226034BAAULL, "hlac_lt_gallowstemplar_2391_7z.dds"},
	{0xB73E075E90957758ULL, "sg_cl.mao"},
	{0xB73E118575D13D70ULL, "vxdr_0_deeproads_cwh_0_0.vlm"},
	{0xB73E6EBD5BE39C43ULL, "vxht_chantry_n_214_0_0.vlm"},
	{0xB73EB2DE82926D3EULL, "vxlt_gallowstemplar_f_472_0_0.vlm"},
	{0xB73EB93E1AE1581BULL, "dae100_viscount.cnv"},
	{0xB73ED5B87BE9AD7BULL, "vxca_caverns2_8lm_0_0.vlm"},
	{0xB73FC10DF7E43C47ULL, "lowtown_fem_cityguard_mob6164823.cl"},
	{0xB7404BE4231B42EAULL, "948364703.ogg"},
	{0xB740B7E4911AA151ULL, "vxht_mansion2b_1ir_0_0.vlm"},
	{0xB7414A2CDB57D686ULL, "zz_dre_debug6075956.cl"},
	{0xB7414D85AB3D176CULL, "vxlt_0_lowtown_n_eu4_0_0.vlm"},
	{0xB741A70EB9B7538CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pew_sa.dds"},
	{0xB74244EAEAB51867ULL, "hlac_lt_0_lowtown_docks_-3_2y.dds"},
	{0xB7425A78FA1A2A5BULL, "follower_banter6163453.cl"},
	{0xB74266CB077AA1D4ULL, "92034581.ogg"},
	{0xB742A4E2E655BF79ULL, "hm_meeran.mor"},
	{0xB742B50F891B3118ULL, "dae000cr_merchant_mage.utc"},
	{0xB7430B7DA208C5CDULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_007a_0s.dds"},
	{0xB74312354BCB7437ULL, "vxlt_0_lowtown_docks_a3_n_bte_0_0.vlm"},
	{0xB74324E515D112A2ULL, "6054369_f.fxe"},
	{0xB74336D3F807AA8FULL, "364816773.ogg"},
	{0xB74371B194825CBBULL, "396725456.ogg"},
	{0xB7442A1F0E06BCA6ULL, "vxht_keep_n_64n_0_0.vlm"},
	{0xB74436D42DF9953BULL, "222940734.ogg"},
	{0xB74455E757F420DAULL, "vxht_bartrand_n_6qv_0_0.vlm"},
	{0xB744B8FF692822DEULL, "and111_lirene6084509.cl"},
	{0xB744C704FDFFF815ULL, "388662880.ogg"},
	{0xB744D61C77671BA4ULL, "6075400_m.fxe"},
	{0xB744D7B1990A2799ULL, "mag101tr_aom_undead2.utt"},
	{0xB74500FFEBEEADB2ULL, "vxlt_0_lowtown_fhe_0_0.vlm"},
	{0xB74519232CA6FAFEULL, "follower_banter6141807.cl"},
	{0xB745204C0DAF7DB2ULL, "zz_approval_debug6003951.cl"},
	{0xB745C61242ACAE56ULL, "163145345.ogg"},
	{0xB7461AC11D5483C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v1.ani"},
	{0xB746469DE21355E2ULL, "llac_lt_0_lowtown_a2_n_0_0.dds"},
	{0xB7466E5F8FF20BF2ULL, "213322395.ogg"},
	{0xB746C09193FEB46BULL, "102628967.ogg"},
	{0xB74744475CF50B15ULL, "hlac_w_encounter2_camp_-65536_0y.dds"},
	{0xB7474E8653EFE082ULL, "lgt110_gxa_vgk_conspirator2.cnv"},
	{0xB747A40638FABDD6ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_012a_0n.dds"},
	{0xB74806F3D63DABEBULL, "mag200_feynriel6023987.cl"},
	{0xB7492ADD13CA15EAULL, "vxlt_backalley_n_1ue_0_0.vlm"},
	{0xB749C6109B4A42F7ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_strt.evt"},
	{0xB74A0C0396A6A5E7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_blueglowworms_da.dds"},
	{0xB74A90979F087BCEULL, "hlac_w_woundedcoast_1_0z.dds"},
	{0xB74B308653F396BAULL, "lgt110_gxa_vgk_conspirator2.bnk"},
	{0xB74B3167E2DBEDF4ULL, "dr_primevaltomb_12048_lc.mmh"},
	{0xB74B5AEF99AE1A33ULL, "6105564_m.fxe"},
	{0xB74B5CE256B31ED0ULL, "6115016_m.fxe"},
	{0xB74B7CDE5B3BFCC9ULL, "dae000ip_squat_d.utp"},
	{0xB74B82BBED8DDF06ULL, "177833432.ogg"},
	{0xB74BA31DA3C362A0ULL, "pet_04.mao"},
	{0xB74BBCB51042947CULL, "123779629.ogg"},
	{0xB74BC1E672155430ULL, "vxw_woundedcoast_1lh_0_0.vlm"},
	{0xB74C05EA5A48C679ULL, "hlac_ht_0_hightown_a2_-1_2x.dds"},
	{0xB74C7F7A42EA8A3DULL, "6016914_m.fxe"},
	{0xB74CF72E4DF1DE2BULL, "6136702_m.fxe"},
	{0xB74D004DCA5D409DULL, "ss_coterie_mage_m.bnk"},
	{0xB74D4B8B0EEF7164ULL, "6078865_m.fxe"},
	{0xB74E237A89C30780ULL, "vxw_woundedcoast_1h4_0_0.vlm"},
	{0xB74EEAD47BC740FFULL, "vxht_0_hightown_cas_0_1.vlm"},
	{0xB74F0300C014E7ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic"},
	{0xB74F115B10E8EA07ULL, "vxlt_darktown_4bw_0_0.vlm"},
	{0xB74F4594BF3639E6ULL, "gen_im_trs_tmp_junk2_04.uti"},
	{0xB74F5220A1FD337AULL, "l_dae121ar_chantry.wbd"},
	{0xB74F66F4E52B093DULL, "hlac_w_encounter3_-196612_0z.dds"},
	{0xB74F824D08030D3AULL, "gen_im_arm_cht_lgt_act2_03.uti"},
	{0xB74F9406FCC21869ULL, "6115390_m.fxe"},
	{0xB74FC7109B4EAF80ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_strt.gad"},
	{0xB74FD62DBEDB95B9ULL, "hlac_w_encounter1_1_0y.dds"},
	{0xB74FD8FA0F4E9F33ULL, "hlac_ca_cavern1_haunted_24424_7z.dds"},
	{0xB74FF48F79C896F6ULL, "273402844.ogg"},
	{0xB75006F5E07F1CE2ULL, "6075262_m.fxe"},
	{0xB750742897ABBF61ULL, "llac_lt_0_lowtown_n_0_5.dds"},
	{0xB750A5A8A10B73C1ULL, "vxht_0_hightown_c69_0_0.vlm"},
	{0xB750B7BF3E2C5974ULL, "vxlt_gallowstemplar_f_5is_0_0.vlm"},
	{0xB75151BD76439CA5ULL, "gen00fl_isabela6110184.cl"},
	{0xB7520BC800D7E5F5ULL, "hlac_lt_gallowstemplar_n_2392_0z.dds"},
	{0xB7522A0AD5835B8AULL, "w_sundermount_125_0.rml"},
	{0xB752B5103CCEBB21ULL, "l_rdr111ar_warehouse.bnk"},
	{0xB753338F93D23BA1ULL, "309822623.ogg"},
	{0xB7533670413F7D5EULL, "ico_shortbow_3.dds"},
	{0xB7535A8F65197297ULL, "lgt311_shw_mer_mettin6156485.cl"},
	{0xB7538C80075C8AE2ULL, "ser221_party_comments6111581.cl"},
	{0xB753C7C36AD67007ULL, "6170146_m.fxe"},
	{0xB75432C11D60E9BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v1.evt"},
	{0xB7545F13E81ACEEFULL, "6052945_m.fxe"},
	{0xB7551B09841E4B1DULL, "mcr370_meredith6083597.cl"},
	{0xB755567F57365091ULL, "106362414.ogg"},
	{0xB756493E4328878BULL, "374043848.ogg"},
	{0xB756AA9C00C11FB4ULL, "vxlt_0_lowtown_a2_n_fhx_0_0.vlm"},
	{0xB756AF8D4BC6FA4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_loop.ani"},
	{0xB756D91F36BBAB9AULL, "hlac_lt_darktown_-1_11x.dds"},
	{0xB757071A948B2F21ULL, "125647343.ogg"},
	{0xB7571ADEB34C43C0ULL, "w_woundedcoast_138_0.phy"},
	{0xB7572E8895119CF9ULL, "ser221_gascard6122449.cl"},
	{0xB75764C7A696F633ULL, "613563562.ogg"},
	{0xB75774569122228CULL, "gen00fl_bethany6098917.cl"},
	{0xB75807FC2EDF2AC9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft_2.mao"},
	{0xB7588181AB9DFC6EULL, "hlac_lt_0_lowtown_n_0_8x.dds"},
	{0xB7594B0AD589B2DEULL, "w_sundermount_125_0.phy"},
	{0xB759593C76D261BEULL, "l_lgt261ar_rescue_cave.wbd"},
	{0xB759A6B99145BBBAULL, "vxw_bonepit_1rc_0_0.vlm"},
	{0xB75A31C11D6553ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_trip_v1.gad"},
	{0xB75A6350B36E4453ULL, "vxlt_undercityserialkiller_23x_0_0.vlm"},
	{0xB75AF53F59BD6E70ULL, "vxdr_0_deeproads_civ_0_0.vlm"},
	{0xB75B1900C73270A8ULL, "dae000_flemeth6000792.cl"},
	{0xB75B558CFD6A0575ULL, "zz_followers_banter6159258.cl"},
	{0xB75BF5B1009B36D5ULL, "hlac_dr_dungeon_3217_5y.dds"},
	{0xB75BFC39FF4DCAB5ULL, "vxca_caverns1_mines_ec9_0_0.vlm"},
	{0xB75C0944B446F19FULL, "139979588.ogg"},
	{0xB75C1E77A2D943D8ULL, "578785226.ogg"},
	{0xB75C823BE4E53602ULL, "art\\characters\\weapons\\axes\\w_axe_ax102a.msh"},
	{0xB75C843BE4E539B8ULL, "art\\characters\\weapons\\axes\\w_axe_ax102a.mmh"},
	{0xB75C95577BB92000ULL, "vxw_woundedcoast_1s5_0_0.vlm"},
	{0xB75D43C57DADA5E5ULL, "6131598_m.fxe"},
	{0xB75D833802DF0287ULL, "vxlt_darktown_4iv_0_0.vlm"},
	{0xB75D934CBDBB213FULL, "vxlt_0_lowtown_n_ex8_0_0.vlm"},
	{0xB75D9B9E7506D251ULL, "vxht_mansion2b_1gp_0_0.vlm"},
	{0xB75E0C23302B4CAFULL, "vxlt_0_lowtown_docks_a3_d7e_0_0.vlm"},
	{0xB75E5787E5C3B946ULL, "hlac_lt_0_lowtown_65536_2z.dds"},
	{0xB75E6F4EB9AF59AEULL, "static_static_p1.vsh.fxo.dx11.assembly"},
	{0xB75EC6E599996043ULL, "mag360_thrask6043899.cl"},
	{0xB760237A2729FA92ULL, "354046134.ogg"},
	{0xB7602FD459599E80ULL, "819837417.ogg"},
	{0xB760A6B818574B2DULL, "sib100_father_contact6110079.cl"},
	{0xB760C222BC0DADC7ULL, "6205671_m.fxe"},
	{0xB760C669BD25F753ULL, "781594289.ogg"},
	{0xB760CBB033D6A52FULL, "vxlt_darktown_2ql_0_0.vlm"},
	{0xB7612BBA1FD9842CULL, "vxht_0_hightown_a3_c73_0_1.vlm"},
	{0xB76167E7EB5C5CB5ULL, "hlac_ht_0_hightown_0_3y.dds"},
	{0xB761B78689BEC96CULL, "490619083.ogg"},
	{0xB761C25EDD91E213ULL, "496061410.ogg"},
	{0xB761CE8252955DE1ULL, "hlac_w_blightlands_327679_0x.dds"},
	{0xB7624F80324E742BULL, "6079340_m.fxe"},
	{0xB762ABA2A8B16462ULL, "6169512_m.fxe"},
	{0xB762D8B19923275CULL, "mag101tr_aom_undead2.ncs"},
	{0xB763A89291D4A904ULL, "320112045.ogg"},
	{0xB7640C9EFDF763FFULL, "1037375698.ogg"},
	{0xB764F16B3499894BULL, "gen00fl_aveline6139487.cl"},
	{0xB7653E056180C7B5ULL, "902321510.ogg"},
	{0xB7655EE9A2C5AFF2ULL, "zz_rdr.ncs"},
	{0xB765E6F9591C05DCULL, "6189372_m.fxe"},
	{0xB7660BD7F0481C9DULL, "gen00fl_anders6111189.cl"},
	{0xB7668250088AA32BULL, "6168110_m.fxe"},
	{0xB7668B3BE4ED9A43ULL, "art\\characters\\weapons\\axes\\w_axe_ax102a.phy"},
	{0xB766DC2F59A08184ULL, "275142344.ogg"},
	{0xB766F1DF30A03D35ULL, "mcr310_argument6127571.cl"},
	{0xB766FEE40361A780ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtd.phy"},
	{0xB7679ECED38970CFULL, "dae000_bartrand6031170.cl"},
	{0xB7680C29AF504176ULL, "vxw_blightlands_8a_0_0.vlm"},
	{0xB7681DCE5C72505DULL, "vxlt_0_lowtown_a2_n_c_gx2_0_0.vlm"},
	{0xB7682371BEF7B17FULL, "dev_res_star.mao"},
	{0xB7683EEF0E627A1BULL, "6101238_m.fxe"},
	{0xB7686CA2B527D4D4ULL, "llac_ca_caverns1_24425_11.dds"},
	{0xB7689781235B068DULL, "zz_vault_debug6203313.cl"},
	{0xB768A5A056018530ULL, "vxw_woundedcoast_1rv_0_0.vlm"},
	{0xB768BAC97F5D2A23ULL, "6065943_m.fxe"},
	{0xB7692CA8925CEDB3ULL, "lgt250_way_dalish_conflict6147123.cl"},
	{0xB7695672BDAC1B56ULL, "gen00fl_dog6167715.cl"},
	{0xB769671C4BB386F0ULL, "vxdr_0_deeproads_cfw_0_0.vlm"},
	{0xB76968D88D080233ULL, "6166310_m.fxe"},
	{0xB769D5DA82CAEF74ULL, "one140cr_charade6134588.cl"},
	{0xB76A968D4BD778D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grovel_floor_loop.gad"},
	{0xB76AC6B1E151B21DULL, "653191334.ogg"},
	{0xB76B71CDAB422BD4ULL, "mrl250_pol6101794.cl"},
	{0xB76C651FCCE6C659ULL, "6028358_f.fxe"},
	{0xB76C90D562573379ULL, "405179083.ogg"},
	{0xB76CE9429B0C0702ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\crouched_behind_wall.gad"},
	{0xB76D522D7634A5FFULL, "398812556.ogg"},
	{0xB76E6EFB5DBEEEEAULL, "lt_0_lowtown_docks_a3_n_249_0.mao"},
	{0xB76E72FB5DBEF539ULL, "lt_0_lowtown_docks_a3_n_249_0.mmh"},
	{0xB76E7CFB5DBF0627ULL, "lt_0_lowtown_docks_a3_n_249_0.msh"},
	{0xB76ED7986E85B616ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_003a_0s.dds"},
	{0xB770344D77910670ULL, "hlac_lt_gallowscourtyard_-65536_0z.dds"},
	{0xB770374660EB9E52ULL, "718721950.ogg"},
	{0xB7703A467986F253ULL, "vxca_caverns2prc_8kt_0_1.vlm"},
	{0xB7704F1095C98889ULL, "ave221_brennan6059775.cl"},
	{0xB771718FA9703F01ULL, "lgt231cr_lyrium_smuggler.utc"},
	{0xB771B7B1C753C9B1ULL, "gen00fl_merrill6129917.cl"},
	{0xB771EC9E4CD58E27ULL, "qun_amb_test.stg"},
	{0xB771FDCBA33603B4ULL, "7002812.ogg"},
	{0xB772A118533A2716ULL, "6056801_m.fxe"},
	{0xB772A73EEFFB0538ULL, "hlac_lt_hangedman_-1_14x.dds"},
	{0xB772B61EAE43E218ULL, "6131000_f.fxe"},
	{0xB772E05CE363E917ULL, "llac_lt_0_lowtown_docks_-65539_6.dds"},
	{0xB773772A05039D4DULL, "w_woundedcoast_148_0.phy"},
	{0xB7738B897EE9878AULL, "mag100_cullen6206246.cl"},
	{0xB7740C429B1261F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\crouched_behind_wall.evt"},
	{0xB7744F086A325ED5ULL, "docks_male_laborer_mob.bnk"},
	{0xB7749097E93BA8DDULL, "gen00fl_aveline6139896.cl"},
	{0xB774F26E8B728441ULL, "555200710.ogg"},
	{0xB775880EB90D636AULL, "751671526.ogg"},
	{0xB7758A3B7AA0A713ULL, "fen210_hadriana6110594.cl"},
	{0xB77598D10F0AD9B8ULL, "dae000ip_crowd_lm1_talk9.utp"},
	{0xB775B86A18206573ULL, "nat380_nathaniel6090058.cl"},
	{0xB77624DA427D094DULL, "vxw_encounter1_fb_0_0.vlm"},
	{0xB776F2528643C451ULL, "zz_approval_debug6003613.cl"},
	{0xB77767E4A1BEB78AULL, "vxca_cavern1_haunted_ea8_0_0.vlm"},
	{0xB77771086A34CE2DULL, "docks_male_laborer_mob.cnv"},
	{0xB7778FD807DBA89CULL, "673373223.ogg"},
	{0xB778099DB4395B8FULL, "1011368090.ogg"},
	{0xB77893D6976642FBULL, "vxht_0_hightown_bwe_0_0.vlm"},
	{0xB778CEF47D5EA191ULL, "isa211_isabela6071435.cl"},
	{0xB7793C764869FB44ULL, "62801698.ogg"},
	{0xB779CCB307F9B157ULL, "km_amb_darktown_urchin.mor"},
	{0xB779D00EEB20FBA8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch.mmh"},
	{0xB779DC0EEB21104BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch.mao"},
	{0xB779EE0EEB212ED2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch.msh"},
	{0xB77A1F237943C4B5ULL, "vxht_mansion_n_53w_0_0.vlm"},
	{0xB77A2DACA8BDE845ULL, "351759296.ogg"},
	{0xB77A4D3C1EEF2335ULL, "vxlt_hangedman_2o1_0_0.vlm"},
	{0xB77A6878A33ED52CULL, "553796209.ogg"},
	{0xB77A77568A5ADEE8ULL, "vxlt_blackemporium_v5_0_0.vlm"},
	{0xB77ADA174E006D04ULL, "6208208_m.fxe"},
	{0xB77B24FB4BE070AAULL, "6118692_f.fxe"},
	{0xB77B648C5E365E94ULL, "6119755_m.fxe"},
	{0xB77BAB2207209064ULL, "176402336.ogg"},
	{0xB77BABE2C959FFBBULL, "vxlt_undercitydungeon_13v_0_0.vlm"},
	{0xB77BF45C16E97493ULL, "589290783.ogg"},
	{0xB77C276B734D0840ULL, "vxdr_0_deeproads_fhx_0_0.vlm"},
	{0xB77C2790FBD9AB81ULL, "pro000_merchant6051621.cl"},
	{0xB77C37DEB36B9155ULL, "w_woundedcoast_138_0.msh"},
	{0xB77C39DEB36B94CBULL, "w_woundedcoast_138_0.mmh"},
	{0xB77C45DEB36BA968ULL, "w_woundedcoast_138_0.mao"},
	{0xB77C6C1F8BFD2959ULL, "6083026_m.fxe"},
	{0xB77C9F6CA25A3511ULL, "904279572.ogg"},
	{0xB77CB69071534B3AULL, "386142554.ogg"},
	{0xB77CFF9B78A1149BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrc_0d.dds"},
	{0xB77D67078EFEFCA5ULL, "89040767.ogg"},
	{0xB77DC37D604C6CF4ULL, "6100376_f.fxe"},
	{0xB77DFC89B03F8A19ULL, "dae011_corff_sta6149618.cl"},
	{0xB77E1AC4CD722FB6ULL, "hlac_ht_0_hightown_n_-65537_3x.dds"},
	{0xB77E2AA3B0AE9AF1ULL, "6150004_f.fxe"},
	{0xB77E78E3170043E6ULL, "narrative_varric6172547.cl"},
	{0xB77F29F8FD6D3B81ULL, "lgt320ip_at_gang_hideout_ht.utp"},
	{0xB77F553D964458E1ULL, "693183434.ogg"},
	{0xB7800172F4A5EB8DULL, "vxlt_0_lowtown_docks_a3_n_d6k_0_0.vlm"},
	{0xB7802618CACCBA5CULL, "mag170_thrask6015997.cl"},
	{0xB7802E8B59E161A7ULL, "ht_bartrand_n_9156_lc.mmh"},
	{0xB780D4E5A9DBCF17ULL, "352504774.ogg"},
	{0xB780D53A9BD17149ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_r.ani"},
	{0xB7815C1860230F26ULL, "6066194_m.fxe"},
	{0xB781D77B943FD5F6ULL, "vxht_brothel_n_6ty_0_0.vlm"},
	{0xB7820ABCC10870E8ULL, "atl_login_dxt5_00.dds"},
	{0xB78214429B1EACE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\crouched_behind_wall.ani"},
	{0xB7821954769981E7ULL, "287390248.ogg"},
	{0xB7822DFF3786DB01ULL, "hlac_lt_0_lowtown_docks_n_-65539_1x.dds"},
	{0xB7826A992EBDA569ULL, "vxlt_0_lowtown_a2_n_c_exl_0_0.vlm"},
	{0xB782E57180866D48ULL, "6132931_f.fxe"},
	{0xB7831A5DBAD940D2ULL, "472844302.ogg"},
	{0xB78325118EE5A064ULL, "w_blightlands_52_0_lc.mmh"},
	{0xB7834C4697D6E6D3ULL, "hlac_w_woundedcoast_-65534_0x.dds"},
	{0xB783D32882CDFFCCULL, "plt_cod_hst_kirk_lowtown.plo"},
	{0xB7843FB77111BD9FULL, "935531298.ogg"},
	{0xB78442E8A7BB08D4ULL, "580108011.ogg"},
	{0xB7846EFEE5B6618CULL, "dae000_sandal6147794.cl"},
	{0xB78478779739AEBEULL, "350501103.ogg"},
	{0xB78496B7347321CDULL, "vxw_encounter1_ka_0_0.vlm"},
	{0xB784B44648225DA6ULL, "qun211_party_comment6179255.cl"},
	{0xB784D6C318D6BDC3ULL, "vxht_mansion_n_3eu_0_0.vlm"},
	{0xB784E8503373012AULL, "hlac_lt_gallowstemplar_f_7198_6z.dds"},
	{0xB785122930E6C46DULL, "6156485_m.fxe"},
	{0xB7854A7A209B20E8ULL, "6041475_f.fxe"},
	{0xB7856697F3E50318ULL, "vxht_keep_6fa_0_0.vlm"},
	{0xB785C00005A37E58ULL, "one100ip_ironwood.utp"},
	{0xB78611415B96B23DULL, "6120031_m.fxe"},
	{0xB786828F213EEA8EULL, "follower_banter6155745.cl"},
	{0xB786F03A9BD60ACBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_r.gad"},
	{0xB786FAC64F834FD8ULL, "w_sundermount_55_0_ll.mmh"},
	{0xB787A0BB61E115ADULL, "hlac_w_woundedcoast_-131068_0z.dds"},
	{0xB787A0FCBC603E8BULL, "vxlt_warehouse_138_0_0.vlm"},
	{0xB787EF86AB59DA54ULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_0x.dds"},
	{0xB787FFEB30D83D06ULL, "vxca_caverns1_ei4_0_1.vlm"},
	{0xB788BF1910E53BB5ULL, "vxlt_hangedman_2d0_0_0.vlm"},
	{0xB7892080FE0EC5DEULL, "vxlt_house_j8_0_0.vlm"},
	{0xB7894C06E7C97B41ULL, "6089562_f.fxe"},
	{0xB7898EEF4476A428ULL, "mag160_feynriel6020469.cl"},
	{0xB7899ACBCDB6F5DEULL, "6135636_m.fxe"},
	{0xB789D669502C7714ULL, "6114055_m.fxe"},
	{0xB78AB861A493058CULL, "23010338.ogg"},
	{0xB78ABEF14833EFC1ULL, "one100_party_comments6133458.cl"},
	{0xB78ADB586CE7AE15ULL, "vxdr_primevalentrance_fu_0_0.vlm"},
	{0xB78B85BCB72F4C98ULL, "vxlt_gallowstemplar_n_45v_0_0.vlm"},
	{0xB78BAD6BF69C2F52ULL, "6170749_m.fxe"},
	{0xB78BC4CE062E8DEDULL, "ch2_s1.psh.fxo"},
	{0xB78C1BE40380F515ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtd.msh"},
	{0xB78C1DE40380F88BULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtd.mmh"},
	{0xB78CDE4B4B659E54ULL, "hlac_w_sundermount_-65538_0y.dds"},
	{0xB78D71E6DB89E30DULL, "zz_mag_debug6053414.cl"},
	{0xB78D7DC12BE95C43ULL, "art\\characters\\creatures\\mubari\\animation\\runascend.ani"},
	{0xB78DED3A9BDC2430ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_r.evt"},
	{0xB78DF8BEDBC40C82ULL, "vxw_bonepitdestroyed_a3_mu_0_0.vlm"},
	{0xB78E0003ACC39735ULL, "129730135.ogg"},
	{0xB78E05EA0B60FAE6ULL, "204915948.ogg"},
	{0xB78E3DC7AA0D82C3ULL, "vxlt_hangedman_n_1tf_0_0.vlm"},
	{0xB78E734FE69C040DULL, "vxlt_0_lowtown_docks_a3_n_d1j_0_0.vlm"},
	{0xB78EE5360E394905ULL, "w_encounter2_camp_38_0_lc.mmh"},
	{0xB78F2B7B42158D6DULL, "vxca_caverns2prc_63a_0_0.vlm"},
	{0xB78F96DB23645A06ULL, "llac_lt_gallowstemplar_2394_2.dds"},
	{0xB78FE32B3D9DA7FFULL, "vxlt_undercityencounter_11m_0_0.vlm"},
	{0xB7907CC117AA7D74ULL, "6151081_m.fxe"},
	{0xB7908646F9601251ULL, "vxlt_0_lowtown_a2_n_gnb_0_0.vlm"},
	{0xB790D523AF2B3160ULL, "ch1armtnt_ds1.psh.fxo"},
	{0xB791C4D01A1A94EEULL, "814772863.ogg"},
	{0xB791F7F810B66EF1ULL, "mcr310_loyalist_mages6076862.cl"},
	{0xB792B8CC3C38E863ULL, "244551426.ogg"},
	{0xB792D281CD3621D6ULL, "6084243_f.fxe"},
	{0xB792F0DD985454F4ULL, "hlac_w_sundermount_-393211_0y.dds"},
	{0xB79318B00A94ECA4ULL, "vxlt_gallowstemplar_37l_0_0.vlm"},
	{0xB7931F034D14C9CDULL, "zz_rdr_debug6066646.cl"},
	{0xB79378E10ED5583EULL, "vxlt_gallowscourtyard_6ys_0_0.vlm"},
	{0xB7939D409D0CBBC3ULL, "hlac_ca_caverns1_mines_24426_2y.dds"},
	{0xB794846CCC340D60ULL, "6118705_m.fxe"},
	{0xB79541AFA7796D53ULL, "457115451.ogg"},
	{0xB795478D37A5D572ULL, "164725557.ogg"},
	{0xB795C3671446B3D5ULL, "6004405_m.fxe"},
	{0xB7969A80FBBFE154ULL, "ward_cl.mao"},
	{0xB796B7DFAAC4E42FULL, "129580558.ogg"},
	{0xB796DB2AD944E7A5ULL, "ave310_party_comment.bnk"},
	{0xB796EA0EC7A37F62ULL, "art\\characters\\playercharacter\\humanfemale\\boo_dksa.phy"},
	{0xB79702DD5D2FF5B5ULL, "vxht_mansion_n_59u_0_0.vlm"},
	{0xB7979CEAD871A45FULL, "dae000_meredith6085344.cl"},
	{0xB797D779BEF5834CULL, "hlac_ht_0_hightown_n_65534_1x.dds"},
	{0xB79800E48E9C9A5CULL, "vxw_woundedcoast_1bj_0_0.vlm"},
	{0xB7982D639268E846ULL, "19170752.ogg"},
	{0xB7988F9CAD4630BBULL, "vxlt_undercitydungeon_19h_0_0.vlm"},
	{0xB798FD9274195C41ULL, "lgt320_shw_ors_black_op6156331.cl"},
	{0xB79919E222BD5A98ULL, "vxht_0_hightown_n_bpt_0_0.vlm"},
	{0xB79949427A667EFEULL, "zz_mcr_debug6071533.cl"},
	{0xB799BD2AD946EADDULL, "ave310_party_comment.cnv"},
	{0xB79A9D3047686C65ULL, "571137232.ogg"},
	{0xB79AA0980840BEACULL, "6207568_m.fxe"},
	{0xB79AC3103D0C88D3ULL, "l_rdr111ar_warehouse.wbd"},
	{0xB79B34B2CEB823C1ULL, "dae000_orsino6085436.cl"},
	{0xB79B43B391396A45ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_ashcover_n.dds"},
	{0xB79B950E73AEC76AULL, "static_static_p108.psh.fxo.dx11"},
	{0xB79B95C12BF5C23AULL, "art\\characters\\creatures\\mubari\\animation\\runascend.evt"},
	{0xB79B98F68ECB556DULL, "fen310_fenris6118705.cl"},
	{0xB79C029E7910751FULL, "vxlt_gallowstemplar_3iu_0_0.vlm"},
	{0xB79C7E371DEE570EULL, "pro000_athenril6050127.cl"},
	{0xB79CAFA49C039B43ULL, "vxlt_hangedman_n_1qi_0_0.vlm"},
	{0xB79CDF293389422EULL, "6180415_f.fxe"},
	{0xB79D211A38FDBE96ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mindblast.ani"},
	{0xB79D2BA1F656BCA3ULL, "vxht_0_hightown_n_c77_0_0.vlm"},
	{0xB79DCC49F5663DE9ULL, "303000243.ogg"},
	{0xB79E9E9BB9AB18B7ULL, "vxht_bartrand_n_5ra_0_0.vlm"},
	{0xB79EBC181679B09CULL, "6133683_m.fxe"},
	{0xB79EBD1A5B674950ULL, "llac_lt_0_lowtown_a3_n_c_65535_9.dds"},
	{0xB79F0D0233316808ULL, "1006081978.ogg"},
	{0xB79F48DBA541F479ULL, "mag101_alain6032133.cl"},
	{0xB79F7BF8DC47A682ULL, "lgt000cr_hate_isabelathug.utc"},
	{0xB79F8D2541D90BDAULL, "793024638.ogg"},
	{0xB79FB6A88B5CB268ULL, "150954053.ogg"},
	{0xB7A104F4A733E81EULL, "mag360_alain6044619.cl"},
	{0xB7A149DEE632343FULL, "default_start_area.are"},
	{0xB7A194C12BFA2C2DULL, "art\\characters\\creatures\\mubari\\animation\\runascend.gad"},
	{0xB7A1B4232ED6F3FDULL, "hlac_lt_0_lowtown_n_65537_7x.dds"},
	{0xB7A20316CB431917ULL, "vxlt_0_lowtown_a3_n_c_ehc_0_0.vlm"},
	{0xB7A2456E9427B7CBULL, "hlac_ht_mansion_n_3370_2z.dds"},
	{0xB7A2D766CE4BA249ULL, "dae221_viscount_chantry6154257.cl"},
	{0xB7A323A5B6F3651BULL, "748579494.ogg"},
	{0xB7A3764260954946ULL, "6132844_m.fxe"},
	{0xB7A3ABF5659F3B91ULL, "407909859.ogg"},
	{0xB7A4082A9926F052ULL, "character_rendertodepth_p2.psh.fxo"},
	{0xB7A484EA49346D5BULL, "985414006.ogg"},
	{0xB7A4CDF6D1CB8FE6ULL, "6067658_m.fxe"},
	{0xB7A4E9AD1A8194A3ULL, "15238594.ogg"},
	{0xB7A4F7592820236FULL, "vrc321tr_merchant_papers.ncs"},
	{0xB7A5078E94B1E338ULL, "gen00fl_carver6005908.cl"},
	{0xB7A58304491524C9ULL, "hlac_lt_0_lowtown_a2_-196608_1z.dds"},
	{0xB7A59325A316B2AFULL, "vxw_bonepitdestroyed_a3_d3_0_0.vlm"},
	{0xB7A59B36F3287543ULL, "genim_arm_cht_inq.uti"},
	{0xB7A5D8E12E20B994ULL, "zz_approval_debug6003108.cl"},
	{0xB7A62FC0CBB5AC41ULL, "6087762_f.fxe"},
	{0xB7A672C18092B2DCULL, "vxw_woundedcoast_1mk_0_0.vlm"},
	{0xB7A6904E619CEBB9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubbucket_sa.dds"},
	{0xB7A6D9C38F3F15A1ULL, "ser221_gascard6113805.cl"},
	{0xB7A790B1EFC89CA4ULL, "147941604.ogg"},
	{0xB7A7A46EEC9D9A5BULL, "hlac_ht_playershouse_n_855_19x.dds"},
	{0xB7A87E84327C445FULL, "vxlt_backalley_n_1r4_0_1.vlm"},
	{0xB7A8BE7E35184B67ULL, "mrl150_dalish_guard6068730.cl"},
	{0xB7A917B56006315BULL, "vxht_0_hightown_a2_by7_0_0.vlm"},
	{0xB7A9EA1C52B94EBFULL, "qun120_arishok_guard6049218.cl"},
	{0xB7A9F819FF156EFBULL, "6112977_m.fxe"},
	{0xB7AA8484A0AA8B29ULL, "vxca_cavern1_haunted_gkc_0_0.vlm"},
	{0xB7AA8803179C7D81ULL, "511483436.ogg"},
	{0xB7AAE4F0299140FEULL, "dae000cr_grand_cleric6148138.cl"},
	{0xB7AB1A78FD131BA0ULL, "6116748_f.fxe"},
	{0xB7AB291A390A0987ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mindblast.evt"},
	{0xB7ABC8FDCD919BD7ULL, "6166890_f.fxe"},
	{0xB7ABE78617423611ULL, "dae000_cora6162363.cl"},
	{0xB7AC36ED276177E4ULL, "121937995.ogg"},
	{0xB7AC618413228D72ULL, "gencr_gate_slave_l5.utc"},
	{0xB7AD36518DABE942ULL, "vxlt_gamlenshouse_n_4f_0_0.vlm"},
	{0xB7AD3CE163464236ULL, "347980584.ogg"},
	{0xB7ADC030049FEEE9ULL, "vxlt_0_lowtown_a2_n_c_ewq_0_0.vlm"},
	{0xB7ADCDE6AB509F5BULL, "hlac_lt_0_lowtown_docks_-65539_2x.dds"},
	{0xB7ADCF4253B67A5CULL, "6061229_m.fxe"},
	{0xB7ADCF805A18D123ULL, "256759515.ogg"},
	{0xB7AE2F505B55F17CULL, "6067054_m.fxe"},
	{0xB7AE37351590AC23ULL, "69432810.ogg"},
	{0xB7AE373E46BECC8DULL, "vxw_encounter3_fy_0_1.vlm"},
	{0xB7AE73FADE8BD374ULL, "dae000_keep_noble_sta6161541.cl"},
	{0xB7AEB08CA92E26A1ULL, "vxw_bonepitmining_a2_26m_0_0.vlm"},
	{0xB7AED13A3E412B25ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isaa_0n.dds"},
	{0xB7AFE2312E0DCC89ULL, "dae000cr_grand_cleric6147637.cl"},
	{0xB7B066555B5132E9ULL, "6217339_m.fxe"},
	{0xB7B0864A7C073257ULL, "645154576.ogg"},
	{0xB7B08A6FB27A04ACULL, "971066760.ogg"},
	{0xB7B108FC0DB1578FULL, "dae000_orana6149108.cl"},
	{0xB7B10C642BDEBF44ULL, "744584512.ogg"},
	{0xB7B175486B44C0B6ULL, "36283866.ogg"},
	{0xB7B18520746B142DULL, "target_circle.cur"},
	{0xB7B1E2C7B6E47E81ULL, "6116650_m.fxe"},
	{0xB7B1EB47295B387CULL, "6138725_m.fxe"},
	{0xB7B242DEE6409BDCULL, "default_start_area.lst"},
	{0xB7B24A1A391060D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mindblast.gad"},
	{0xB7B3309DA07BEEB8ULL, "hlac_lt_gallowstemplar_f_2384_4x.dds"},
	{0xB7B34317FA4F5895ULL, "885101084.ogg"},
	{0xB7B3556586125CC5ULL, "llac_lt_0_lowtown_n_-1_3.dds"},
	{0xB7B4193727F8D73CULL, "796689414.ogg"},
	{0xB7B42F069A1DD0D8ULL, "6164346_m.fxe"},
	{0xB7B4C794C3454343ULL, "49039927.ogg"},
	{0xB7B4E49E7288CB5CULL, "vxw_woundedcoast_1hh_0_0.vlm"},
	{0xB7B4E668786A537EULL, "dae000_arishok6067722.cl"},
	{0xB7B5673FECBCC170ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand.msh"},
	{0xB7B5753FECBCD88AULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand.mmh"},
	{0xB7B5793FECBCDFC1ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand.mao"},
	{0xB7B59C24A1643643ULL, "6111561_m.fxe"},
	{0xB7B5FD9C06A98B98ULL, "vxht_0_hightown_n_bfr_0_0.vlm"},
	{0xB7B681D70A779DD0ULL, "fx_smoke_vol2_add.mao"},
	{0xB7B6F472B26E688AULL, "gen00fl_bethany6101396.cl"},
	{0xB7B71F47C8990CC4ULL, "vxlt_hangedman_n_1vb_0_1.vlm"},
	{0xB7B7899251FC49DBULL, "vxht_0_hightown_a2_bb6_0_0.vlm"},
	{0xB7B7C5E4ECB5263DULL, "225489403.ogg"},
	{0xB7B7D6FFC349DB4CULL, "zz_followers_banter6157974.cl"},
	{0xB7B7D96CC55B0686ULL, "vxca_caverns1_mines_eb1_0_0.vlm"},
	{0xB7B82114FD429665ULL, "one100dg_vanard6047409.cl"},
	{0xB7B8BF76DE138B5BULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_beth.msh"},
	{0xB7B8DD76DE13BE45ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_beth.mmh"},
	{0xB7B8E831A443990FULL, "zz_dae_debug6029096.cl"},
	{0xB7B8EB32216F9FD4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_4.pwk"},
	{0xB7B8F0D71E2FD192ULL, "vxlt_0_lowtown_docks_a3_n_cx1_0_0.vlm"},
	{0xB7B92C659B277E9BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed.evt"},
	{0xB7B9D2991E6EAECAULL, "6068516_m.fxe"},
	{0xB7BA03482EC63CF5ULL, "llac_lt_0_lowtown_65536_1.dds"},
	{0xB7BA0F5BDA42EDA3ULL, "vxht_0_hightown_n_c99_0_0.vlm"},
	{0xB7BA3F64EB572EDEULL, "vxlt_warehouse_8i_0_0.vlm"},
	{0xB7BA5CF6B65F7790ULL, "ser221_gamlen6135368.cl"},
	{0xB7BA9234DE82F77CULL, "dae000_merchant_generic_hm6172656.cl"},
	{0xB7BAA282432BFB74ULL, "250828065.ogg"},
	{0xB7BAD6229AAA583AULL, "vxlt_0_lowtown_docks_b5v_0_1.vlm"},
	{0xB7BB5D31E3E73A86ULL, "vxca_caverns2_5c8_0_0.vlm"},
	{0xB7BB82559D9749B7ULL, "vxht_bartrand_n_5xc_0_0.vlm"},
	{0xB7BB9FD1FA65E19CULL, "6139083_m.fxe"},
	{0xB7BBC58A5C4A9BA5ULL, "vxw_sundermount_23k_0_0.vlm"},
	{0xB7BBCB147E76C73FULL, "vxlt_0_lowtown_docks_aow_0_0.vlm"},
	{0xB7BBD10EC7C270B5ULL, "art\\characters\\playercharacter\\humanfemale\\boo_dksa.mmh"},
	{0xB7BBD30EC7C2742BULL, "art\\characters\\playercharacter\\humanfemale\\boo_dksa.msh"},
	{0xB7BC1054F12BB751ULL, "zz_followers_banter6158778.cl"},
	{0xB7BC6C1FDD121019ULL, "179743624.ogg"},
	{0xB7BD04EC47B086AEULL, "393753754.ogg"},
	{0xB7BD1A1A7893DFFAULL, "165787031.ogg"},
	{0xB7BD31EBD48F7341ULL, "603337883.ogg"},
	{0xB7BD48A1085C094FULL, "vxw_woundedcoast_14s_0_0.vlm"},
	{0xB7BD57B753CD13A4ULL, "hlac_lt_0_lowtown_docks_a3_n_131068_0y.dds"},
	{0xB7BD62468AB7F5D0ULL, "one120_athenril6088317.cl"},
	{0xB7BDC112055B2312ULL, "499632560.ogg"},
	{0xB7BE2ECAA46A47DAULL, "gen00fl_fenris6107211.cl"},
	{0xB7BEEDE6C2293BC8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1pp_basic.msh"},
	{0xB7BEFFE6C2295A59ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1pp_basic.mao"},
	{0xB7BF0BE6C2296EF2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1pp_basic.mmh"},
	{0xB7BF14DD6D469555ULL, "873163353.ogg"},
	{0xB7BF50F053437C02ULL, "6172927_m.fxe"},
	{0xB7BF60C073F6B3B3ULL, "634910756.ogg"},
	{0xB7BF7A88B796F7F4ULL, "vxht_bartrand_n_3vt_0_0.vlm"},
	{0xB7BFD2FD10767F69ULL, "hlac_lt_gallowstemplar_2383_10x.dds"},
	{0xB7C01058DA989B27ULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_bdy_gib_0s.dds"},
	{0xB7C02557B9997E9BULL, "lowtown_guard_chase6169509.cl"},
	{0xB7C025659B2D91DCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed.gad"},
	{0xB7C02BB95CD28DD1ULL, "hlac_ca_caverns2_5802_24x.dds"},
	{0xB7C07C1E394AD4AEULL, "6195865_m.fxe"},
	{0xB7C0FFFCBD30D99FULL, "111294344.ogg"},
	{0xB7C17A74972DB177ULL, "w_encounter3_231_0.mmh"},
	{0xB7C17E74972DB784ULL, "w_encounter3_231_0.mao"},
	{0xB7C19074972DD6E9ULL, "w_encounter3_231_0.msh"},
	{0xB7C1CF286B1232DDULL, "675157325.ogg"},
	{0xB7C23E8AE95D1836ULL, "zz_followers_banter6159585.cl"},
	{0xB7C2E45928390016ULL, "vrc321tr_merchant_papers.utt"},
	{0xB7C2E7B8CA87366EULL, "art\\characters\\creatures\\mubari\\animation\\sleep_tw_v1.evt"},
	{0xB7C300B8ED68B2EAULL, "vxlt_0_lowtown_docks_a3_bfa_0_0.vlm"},
	{0xB7C33DCB948BDE1AULL, "6064714_m.fxe"},
	{0xB7C354CAE2AEFB8BULL, "and340ip_at_mines.ncs"},
	{0xB7C3567B647EE3DCULL, "vxw_woundedcoast_1si_0_0.vlm"},
	{0xB7C365D7B7F4EC27ULL, "hlac_dr_primevaltomb_657_1x.dds"},
	{0xB7C373E7BA6EC2FFULL, "268487811.ogg"},
	{0xB7C3F8833D577145ULL, "6095115_m.fxe"},
	{0xB7C42E6698EF29F1ULL, "fxe_water_smoke2_blend.dds"},
	{0xB7C539DE4B41E3DCULL, "457828334.ogg"},
	{0xB7C57D93B3586DA7ULL, "gen00fl_anders6119734.cl"},
	{0xB7C624659B31FBDAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_door.opentoclosed.ani"},
	{0xB7C6DBD719E92EBCULL, "hlac_ht_0_hightown_-131074_0x.dds"},
	{0xB7C7198EED4FCFE5ULL, "6051393_m.fxe"},
	{0xB7C76794B0C6EE91ULL, "6110404_f.fxe"},
	{0xB7C78E15A917E2B7ULL, "116898222.ogg"},
	{0xB7C8599CFA835C95ULL, "prm000_varric_crest.bnk"},
	{0xB7C89987F8423F0FULL, "hlac_w_woundedcoast_-262139_0z.dds"},
	{0xB7C8DEB8CA8B9219ULL, "art\\characters\\creatures\\mubari\\animation\\sleep_tw_v1.gad"},
	{0xB7C8E1527D6DAB87ULL, "vxdr_dungeon_1ly_0_1.vlm"},
	{0xB7C9058C7C3019E7ULL, "15586712.ogg"},
	{0xB7C96F378D3523D8ULL, "323799319.ogg"},
	{0xB7C978AA20252CFBULL, "236700353.ogg"},
	{0xB7C987963043552DULL, "art\\cinematics\\additional"},
	{0xB7C9C1740C7CE2B9ULL, "16111638.ogg"},
	{0xB7C9DF22C2637560ULL, "58441275.ogg"},
	{0xB7CA4C9EAFF2F8CBULL, "42139156.ogg"},
	{0xB7CA50DA1B1D36CBULL, "6093585_f.fxe"},
	{0xB7CA9FDF0A206586ULL, "vxlt_0_lowtown_docks_9f5_0_0.vlm"},
	{0xB7CAD1EECC8C07E5ULL, "hlac_lt_0_lowtown_a2_n_65536_4x.dds"},
	{0xB7CB4383B428AC94ULL, "hlac_lt_gallowscourtyard_n_-1_3z.dds"},
	{0xB7CB4E2B97F30023ULL, "w_bonepitdestroyed_a3_59_0.mao"},
	{0xB7CB522B97F30770ULL, "w_bonepitdestroyed_a3_59_0.mmh"},
	{0xB7CB55CBFA2BDF66ULL, "llac_lt_gallowstemplar_f_2383_7.dds"},
	{0xB7CB572DA11DAFBCULL, "1037608539.ogg"},
	{0xB7CB602B97F31E8AULL, "w_bonepitdestroyed_a3_59_0.msh"},
	{0xB7CB94468D1A57A1ULL, "vxw_bonepitmining_a2_28c_0_0.vlm"},
	{0xB7CB95D47A9CC5CDULL, "vxlt_hangedman_271_0_0.vlm"},
	{0xB7CBDFD8031E126CULL, "scaleform_sf_depth_p0.vsh.fxo.dx11"},
	{0xB7CC70612A8476EBULL, "vxlt_foundry_n_xc_0_0.vlm"},
	{0xB7CC7B9CFA877FEDULL, "prm000_varric_crest.cnv"},
	{0xB7CC8C38C447CE34ULL, "dae011_corff_sta6152734.cl"},
	{0xB7CCC59F09D5A06DULL, "299510564.ogg"},
	{0xB7CCD4609BA4F695ULL, "llac_lt_0_lowtown_a2_n_c_65536_11.dds"},
	{0xB7CD689EF01427C6ULL, "vxht_keep_5hn_0_0.vlm"},
	{0xB7CD91222C1CCB3CULL, "1019999437.ogg"},
	{0xB7CDB71D6BE0655FULL, "dae000_osric6162704.cl"},
	{0xB7CE5B80CC38637FULL, "263177571.ogg"},
	{0xB7CEF7A47A0BADB1ULL, "272122872.ogg"},
	{0xB7CF21A8E7C087E3ULL, "114370793.ogg"},
	{0xB7CF33C305B7E655ULL, "hlac_lt_0_lowtown_a2_n_65538_0z.dds"},
	{0xB7CF35C9B456964BULL, "docks_male_elven_worker_sta6175816.cl"},
	{0xB7CF84D2CA4E757CULL, "zz_dae_debug6029123.cl"},
	{0xB7CFC8C71B1185E7ULL, "vxlt_darktown_5o5_0_0.vlm"},
	{0xB7CFEFB8CA91CE67ULL, "art\\characters\\creatures\\mubari\\animation\\sleep_tw_v1.ani"},
	{0xB7D07A21C538C5BBULL, "276171917.ogg"},
	{0xB7D09628CBFC0BF4ULL, "vxlt_gallowstemplar_n_5pl_0_0.vlm"},
	{0xB7D0A4AA48C1DE3BULL, "vxca_cavern1_haunted_e86_0_0.vlm"},
	{0xB7D0A86F3A6FF7D2ULL, "6139320_m.fxe"},
	{0xB7D0CA8A0580C2DCULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube_tool_1.mao"},
	{0xB7D0EFB6DB8232E4ULL, "vxlt_warehouse_n_134_0_0.vlm"},
	{0xB7D11154D3AEE839ULL, "static_static_p4.psh.fxo.dx11"},
	{0xB7D1BA7D2022AFCDULL, "vxdr_primevaltomb_6vd_0_1.vlm"},
	{0xB7D23AC6004CA77FULL, "6105660_f.fxe"},
	{0xB7D2CFAB61FD61A9ULL, "dae000_arishok6067846.cl"},
	{0xB7D2D071BA10E312ULL, "dae000ip_crowd_mf6_talk8.utp"},
	{0xB7D33F637F4B9AF7ULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm_2.phy"},
	{0xB7D35AC91FF5E3A1ULL, "mag110_vincento6017809.cl"},
	{0xB7D46368F079BB1BULL, "186134743.ogg"},
	{0xB7D4A732B79F0FB5ULL, "vxw_encounter2_camp_ro_0_0.vlm"},
	{0xB7D4B0A350D037C3ULL, "lt_gallowstemplar_7198.rml"},
	{0xB7D4C281110D47E5ULL, "ave121_brennan6105808.cl"},
	{0xB7D4E3421DDB69EBULL, "115311375.ogg"},
	{0xB7D4F28987E87A40ULL, "vrc321ip_static_book_2.utp"},
	{0xB7D524526198765FULL, "vxw_bonepit_1q3_0_0.vlm"},
	{0xB7D5B2555D1F3885ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzraisedstage_0.pwk"},
	{0xB7D5D7472A469E1DULL, "vxlt_warehouse_t2_0_0.vlm"},
	{0xB7D607AF456424CDULL, "t3_eey_br4.tnt"},
	{0xB7D6266014CAE9EFULL, "26218103.ogg"},
	{0xB7D7178F1E5C380BULL, "282663284.ogg"},
	{0xB7D74EAE1A02CC81ULL, "vxw_woundedcoast_1h9_0_0.vlm"},
	{0xB7D75CEB438B5C9BULL, "229355823.ogg"},
	{0xB7D768D3C18543E4ULL, "vxht_0_hightown_a2_9b2_0_0.vlm"},
	{0xB7D7B370E13E5513ULL, "462059395.ogg"},
	{0xB7D8427CBDC26960ULL, "llac_lt_0_lowtown_docks_-131075_0.dds"},
	{0xB7D8A9444036CCA5ULL, "vxw_sundermount_29i_0_0.vlm"},
	{0xB7D8D61B35BD9AA4ULL, "vxht_0_hightown_a2_a6f_0_0.vlm"},
	{0xB7D93EB61A7FA923ULL, "testperc.ncs"},
	{0xB7D96FBE8ADA3CB4ULL, "803957970.ogg"},
	{0xB7DA1BF5A3D55422ULL, "vxht_0_hightown_a2_b8e_0_0.vlm"},
	{0xB7DAA94A7B7C707CULL, "hlac_ht_0_hightown_a2_-65537_2x.dds"},
	{0xB7DAD34AE44F9739ULL, "6027683_f.fxe"},
	{0xB7DB3663ABCB2235ULL, "zz_rogueslate_white.cut"},
	{0xB7DB38DC68F2B256ULL, "qun221_petrice6077970.cl"},
	{0xB7DB6D5AB653DA6DULL, "285482057.ogg"},
	{0xB7DBE2D87C6C69F2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masa_0n.dds"},
	{0xB7DBECBA0363D267ULL, "6101326_m.fxe"},
	{0xB7DCF0E226D5CE9BULL, "qun111_ketojan6039816.cl"},
	{0xB7DD39148A6BFEC6ULL, "814201417.ogg"},
	{0xB7DD3E58D4B02F4BULL, "qun200_party_comment6090335.cl"},
	{0xB7DD578C0EA45E17ULL, "972825585.ogg"},
	{0xB7DD9392ABFE165CULL, "178230566.ogg"},
	{0xB7DDC1F373CB0A26ULL, "100280805.ogg"},
	{0xB7DDF140E11A32B0ULL, "vxlt_darktown_5q0_0_0.vlm"},
	{0xB7DE3AA40D079E67ULL, "vxlt_darktown_5p4_0_0.vlm"},
	{0xB7DE82B8F5987ED4ULL, "147082156.ogg"},
	{0xB7DF6193CD784B64ULL, "vxlt_warehouse_n_185_0_0.vlm"},
	{0xB7DF8C85F206E959ULL, "1048957380.ogg"},
	{0xB7DFB0CB2338C904ULL, "ser270_templar6122507.cl"},
	{0xB7DFC69E6A55750FULL, "6054683_f.fxe"},
	{0xB7E031F817F6EF82ULL, "art\\characters\\playercharacter\\elfmale\\cth_hoo_comc.phy"},
	{0xB7E05A11D85A5162ULL, "vxca_caverns2prc_aig_0_0.vlm"},
	{0xB7E0EE22E438D845ULL, "dae000cr_grand_cleric6148157.cl"},
	{0xB7E12FED7D457D4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_other.evt"},
	{0xB7E1911114FF9783ULL, "lowtown_male_urchin_sta6164201.cl"},
	{0xB7E19EBA601C9C2CULL, "6163440_m.fxe"},
	{0xB7E1B70005F1CDF9ULL, "one100ip_ironwood.ncs"},
	{0xB7E1C41BE2B8F058ULL, "combat_smm_r1_a1.gda"},
	{0xB7E1D4A040E4CE8CULL, "dae121st_bran.stg"},
	{0xB7E2267477B36C1FULL, "6013380_m.fxe"},
	{0xB7E23A8194C29BDEULL, "23520013.ogg"},
	{0xB7E259CAE2C9B576ULL, "and340ip_at_mines.utp"},
	{0xB7E2A2EC30A78363ULL, "vxht_chantry_n_354_0_1.vlm"},
	{0xB7E2C28DBF512983ULL, "6047385_m.fxe"},
	{0xB7E2DEB4C329550EULL, "genip_tray.utp"},
	{0xB7E2E6469972AB36ULL, "gen00fl_sebastian6195888.cl"},
	{0xB7E3190FA9952835ULL, "vxw_encounter2_camp_mn_0_0.vlm"},
	{0xB7E37832FD1E8C09ULL, "ht_0_hightown_a3_4_0.rml"},
	{0xB7E43BEF32DEEF9FULL, "vxlt_gallowstemplar_3rp_0_0.vlm"},
	{0xB7E449241C3CB69DULL, "vxlt_warehouse_y1_0_0.vlm"},
	{0xB7E4986372E3B622ULL, "vxht_0_hightown_n_ap5_0_0.vlm"},
	{0xB7E4BA76424964BBULL, "mag321_comtesse_de_launcet6061396.cl"},
	{0xB7E4C4D10674DBAEULL, "343950471.ogg"},
	{0xB7E4CA6E9ECD39BDULL, "gen00fl_fenris6116582.cl"},
	{0xB7E54DA61F420BDBULL, "85543921.ogg"},
	{0xB7E5C08B0BF8E501ULL, "vxw_woundedcoast_1s8_0_0.vlm"},
	{0xB7E5C4613CDCC47AULL, "framebuffereffects_vsmblur_p1.psh.fxo.dx11"},
	{0xB7E5DFECF8CCB6CFULL, "296823432.ogg"},
	{0xB7E65DC47DD635A1ULL, "hlac_ht_0_hightown_a3_-196607_0y.dds"},
	{0xB7E665F939EA6D0FULL, "6164820_m.fxe"},
	{0xB7E6BC83F4996DC8ULL, "t3_arm_blr.tnt"},
	{0xB7E76ACA671F8649ULL, "tbe100ip_mirror_transformat.utp"},
	{0xB7E7783605D7E3E0ULL, "vrc321ip_shade_proj_end.utp"},
	{0xB7E826ED7D4B8CF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_other.gad"},
	{0xB7E8411B10A859F3ULL, "398127326.ogg"},
	{0xB7E8CAAE6DA2E6D4ULL, "vxlt_undercityserialkiller_1o8_0_0.vlm"},
	{0xB7E8D8260C4DFE57ULL, "776263154.ogg"},
	{0xB7E913CBF815CD98ULL, "561258896.ogg"},
	{0xB7E9E29E82B1352CULL, "6139388_m.fxe"},
	{0xB7EA0BF99B32E179ULL, "911837145.ogg"},
	{0xB7EA0CEE30F71DF6ULL, "vxw_blightlands_ej_0_0.vlm"},
	{0xB7EA565C2C7A9A61ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv2.ani"},
	{0xB7EA86F9516DE602ULL, "416560301.ogg"},
	{0xB7EA874D09733B90ULL, "1007342449.ogg"},
	{0xB7EBCCBF882D1A17ULL, "gen00fl_anders6125416.cl"},
	{0xB7EBD3542429733CULL, "vxlt_0_lowtown_a3_n_c_gv8_0_0.vlm"},
	{0xB7EBE4BAD926AFD6ULL, "mcr310_argument6127682.cl"},
	{0xB7EC210599181FFEULL, "225966156.ogg"},
	{0xB7ECD8CE80B8E2A6ULL, "vxca_caverns1_coe_0_0.vlm"},
	{0xB7ED2436F80D0B08ULL, "w_woundedcoast_71_0.anb"},
	{0xB7ED5098E2F2D346ULL, "vxht_0_hightown_c75_0_0.vlm"},
	{0xB7ED60977294B252ULL, "lgt220_shp_mag_guard6149898.cl"},
	{0xB7EDB642E073B781ULL, "hm_qun_hench1.mor"},
	{0xB7EDFB051870C47CULL, "hlac_w_bonepitmining_a2_196613_0z.dds"},
	{0xB7EE1809DF9DB2D7ULL, "227756895.ogg"},
	{0xB7EE37ED7D501547ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_other.ani"},
	{0xB7EEB4AD1E058AE0ULL, "429668216.ogg"},
	{0xB7EEB8F986FD800CULL, "raven_cloud.mao"},
	{0xB7EECC07A5368A6FULL, "102351627.ogg"},
	{0xB7EEF22914A14260ULL, "hlac_lt_gallowstemplar_7198_7x.dds"},
	{0xB7EF2FA587BEEE82ULL, "dae000_quintus6154734.cl"},
	{0xB7EF501B2404971CULL, "vxw_sundermount_1zz_0_0.vlm"},
	{0xB7EF908E752E1D24ULL, "vxca_caverns2_6zn_0_0.vlm"},
	{0xB7EFA0AD6C985B0FULL, "cod_bks_thrask_notes.utp"},
	{0xB7EFC39D22648830ULL, "187714688.ogg"},
	{0xB7F04CEFE6BCD7C1ULL, "316678301.ogg"},
	{0xB7F0FCA8E7DD8B47ULL, "vxht_mansion_n_3hn_0_0.vlm"},
	{0xB7F135D0DE515BACULL, "527028455.ogg"},
	{0xB7F1615C2C80CB03ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv2.gad"},
	{0xB7F1A93694E2C940ULL, "lgt310_mis_burn_body.cnv"},
	{0xB7F1AC42AF7D46DAULL, "vxht_bartrand_n_65j_0_0.vlm"},
	{0xB7F1B21201E28D2CULL, "42386317.ogg"},
	{0xB7F1E7F15016A5C3ULL, "6075380_m.fxe"},
	{0xB7F30A4064D9CEA2ULL, "vxht_0_hightown_n_au6_0_0.vlm"},
	{0xB7F39BA117ED4741ULL, "hlac_lt_gallowstemplar_2392_1z.dds"},
	{0xB7F3A7C2F1509768ULL, "lt_0_lowtown_docks_a3_n_156_0.msh"},
	{0xB7F3B9C2F150B5F9ULL, "lt_0_lowtown_docks_a3_n_156_0.mao"},
	{0xB7F3C5C2F150C992ULL, "lt_0_lowtown_docks_a3_n_156_0.mmh"},
	{0xB7F3F4A312D8EB3BULL, "vxlt_0_lowtown_a3_n_c_hrt_0_0.vlm"},
	{0xB7F4538459CB0332ULL, "vxht_chantry_n_149_0_0.vlm"},
	{0xB7F48B3694E4CC78ULL, "lgt310_mis_burn_body.bnk"},
	{0xB7F4B56187630459ULL, "809820460.ogg"},
	{0xB7F5092B68C58D17ULL, "6134740_m.fxe"},
	{0xB7F518E18154B1DAULL, "728027017.ogg"},
	{0xB7F51D6639642518ULL, "6134783_m.fxe"},
	{0xB7F6891A857877B4ULL, "hlac_ca_caverns2_5802_15y.dds"},
	{0xB7F7570C027F3A0CULL, "vxlt_0_lowtown_docks_c6u_0_0.vlm"},
	{0xB7F7CFAC4DF0AFAEULL, "gen00fl_fenris6128069.cl"},
	{0xB7F7D0DCF558F933ULL, "vxca_caverns1_mines_hz1_0_0.vlm"},
	{0xB7F82FDCEA54896DULL, "w_bonepit_223_0.msh"},
	{0xB7F83DDCEA54A0A0ULL, "w_bonepit_223_0.mao"},
	{0xB7F83E5C2C86AEE8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\twitchv2.evt"},
	{0xB7F8403742C7F276ULL, "6109231_m.fxe"},
	{0xB7F841DCEA54A7F3ULL, "w_bonepit_223_0.mmh"},
	{0xB7F84236BBC24CADULL, "44906293.ogg"},
	{0xB7F842A2E5EBF626ULL, "gen00fl_aveline6132719.cl"},
	{0xB7F85DB01C64A7F6ULL, "ele.dds"},
	{0xB7F865CF4FD0E758ULL, "878774641.ogg"},
	{0xB7F87ECB22ED3676ULL, "vxw_blightlands_nk_0_0.vlm"},
	{0xB7F91F8909788789ULL, "352406485.ogg"},
	{0xB7FA5037B5209A59ULL, "vxw_woundedcoast_c1_0_0.vlm"},
	{0xB7FA6A4F24224835ULL, "mag130_danzig6018799.cl"},
	{0xB7FA6D7490AB4C94ULL, "vxht_mansion2b_3gn_0_1.vlm"},
	{0xB7FB20EB038178B8ULL, "vxht_0_hightown_a2_bax_0_0.vlm"},
	{0xB7FB74FC4E2F7D96ULL, "hlac_lt_0_lowtown_131073_0x.dds"},
	{0xB7FB8E3156463A92ULL, "llac_lt_0_lowtown_a2_n_-2_5.dds"},
	{0xB7FBA6270927F936ULL, "leave_chunky.mao"},
	{0xB7FBEBBFA1DE5574ULL, "vxlt_gallowstemplar_n_5as_0_0.vlm"},
	{0xB7FC49383D0353EDULL, "6096828_m.fxe"},
	{0xB7FC674DFC906FC1ULL, "57040890.ogg"},
	{0xB7FC93143FFF070EULL, "vxlt_warehouse_140_0_0.vlm"},
	{0xB7FC9E30FF47C7FFULL, "w_encounter2_camp_50_0.msh"},
	{0xB7FCA430FF47D1A1ULL, "w_encounter2_camp_50_0.mmh"},
	{0xB7FCAFCAC8838005ULL, "894530114.ogg"},
	{0xB7FCB030FF47E642ULL, "w_encounter2_camp_50_0.mao"},
	{0xB7FD3DCBBC468262ULL, "vxca_caverns2prc_aca_0_0.vlm"},
	{0xB7FE02BC31523D95ULL, "vxlt_0_lowtown_a2_deg_0_0.vlm"},
	{0xB7FE2B8A30A4A0F4ULL, "ser270_emeric6113376.cl"},
	{0xB7FE4B992E8630B1ULL, "art\\characters\\creatures\\mubari\\animation\\crnf.ani"},
	{0xB7FE51ABCA04C583ULL, "hlac_lt_0_lowtown_a2_1_7x.dds"},
	{0xB7FE8F2382D7C559ULL, "vxlt_darktown_514_0_1.vlm"},
	{0xB7FEE93A314E04E9ULL, "zz_dae_debug6029087.cl"},
	{0xB7FF13C141304F17ULL, "vxlt_0_lowtown_docks_b56_0_0.vlm"},
	{0xB7FF7811FBD0149EULL, "6055620_m.fxe"},
	{0xB7FF83CB760144BDULL, "803802264.ogg"},
	{0xB7FF8D246B62D806ULL, "vxlt_0_lowtown_n_csj_0_0.vlm"},
	{0xB7FFFCC98D815935ULL, "vxw_encounter2_camp_cl_0_0.vlm"},
	{0xB800505FD117F068ULL, "724305483.ogg"},
	{0xB80061395F07B48EULL, "vxdr_0_deeproads_egt_0_0.vlm"},
	{0xB800631896C18883ULL, "6072455_m.fxe"},
	{0xB800A544E68C1838ULL, "mer210_smith6051391.cl"},
	{0xB800B676DE5131F2ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_beth.phy"},
	{0xB80158FD7CA406AEULL, "6092276_m.fxe"},
	{0xB8018B3FABEDF52DULL, "6075218_m.fxe"},
	{0xB801B42C626AE6FAULL, "hlac_lt_0_lowtown_docks_a3_262142_0x.dds"},
	{0xB80249C94C7082F4ULL, "gen00fl_anders6128322.cl"},
	{0xB802C03DF4E43FC1ULL, "hlac_lt_0_lowtown_docks_a3_0_4z.dds"},
	{0xB803839B38D85390ULL, "vxlt_0_lowtown_a2_n_c_ht8_0_0.vlm"},
	{0xB8041CD162E03EBAULL, "6070882_f.fxe"},
	{0xB8045A2307FF427DULL, "seb221_brett_harimann6177240.cl"},
	{0xB80478D7D7DD18ADULL, "zz_followers_banter6157568.cl"},
	{0xB8051FA65ACAA675ULL, "6127997_m.fxe"},
	{0xB8054C6184E5B3E4ULL, "vxdr_primevaltomb_50t_0_1.vlm"},
	{0xB8054E3699BAD3F7ULL, "6050198_m.fxe"},
	{0xB80553CA67385CC0ULL, "tbe100ip_mirror_transformat.ncs"},
	{0xB80556992E8C6253ULL, "art\\characters\\creatures\\mubari\\animation\\crnf.gad"},
	{0xB805C8E8F475528CULL, "vxlt_0_lowtown_docks_c3j_0_0.vlm"},
	{0xB805EB3D6A741357ULL, "mag100_thrask6031964.cl"},
	{0xB8060CF818178036ULL, "art\\characters\\playercharacter\\elfmale\\cth_hoo_comc.mao"},
	{0xB80618F8181794D5ULL, "art\\characters\\playercharacter\\elfmale\\cth_hoo_comc.mmh"},
	{0xB8061AF81817984BULL, "art\\characters\\playercharacter\\elfmale\\cth_hoo_comc.msh"},
	{0xB80638153979DB2BULL, "hlac_lt_0_lowtown_n_-1_2y.dds"},
	{0xB806F0A814E34EF6ULL, "vxw_blightlands_sl_0_0.vlm"},
	{0xB806FF19922F5BA8ULL, "gen00fl_aveline6114872.cl"},
	{0xB80735EB20D0622CULL, "vxlt_undercitydungeon_20w_0_0.vlm"},
	{0xB8078B28DEAB02E8ULL, "one220_yevhen6098931.cl"},
	{0xB807C65D888E6A3DULL, "78936893.ogg"},
	{0xB807C75E30C9ED06ULL, "hlac_w_encounter1_65539_0y.dds"},
	{0xB808192DE6588A2EULL, "dre000_varric6034639.cl"},
	{0xB808821BC4C9A0BCULL, "1065827452.ogg"},
	{0xB808D034F75395E8ULL, "llac_ca_caverns2prc_5802_17.dds"},
	{0xB809002BD5521F07ULL, "6105692_m.fxe"},
	{0xB809148E9E2CA8C2ULL, "mag160_varian_ilithis6016486.cl"},
	{0xB8092DF9892C6FEAULL, "hlac_lt_0_lowtown_docks_a3_65535_0y.dds"},
	{0xB809328DB04BDFC3ULL, "990751770.ogg"},
	{0xB809A4BD3DFFDF44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_buildup.ani"},
	{0xB809EBB2D3424600ULL, "vxw_woundedcoast_179_0_0.vlm"},
	{0xB80A00E583BF02CEULL, "6154101_m.fxe"},
	{0xB80A5BCD5A701873ULL, "6053564_m.fxe"},
	{0xB80A7193A92EA94BULL, "llac_lt_undercityencounter_-1_21.dds"},
	{0xB80A73A31DE4672EULL, "hlac_w_bonepitmining_a2_0_2x.dds"},
	{0xB80AB1F58D2A57C0ULL, "mag160_feynriel6020472.cl"},
	{0xB80AB7B6F13AAF3AULL, "166116609.ogg"},
	{0xB80AFA7842FBD7DBULL, "gen00fl_sebastian6179913.cl"},
	{0xB80B44FB52DAEFBBULL, "325622513.ogg"},
	{0xB80B53992E90C8B8ULL, "art\\characters\\creatures\\mubari\\animation\\crnf.evt"},
	{0xB80B5D34020A0D0BULL, "809655723.ogg"},
	{0xB80BDB5CC912BD8DULL, "6070525_m.fxe"},
	{0xB80C80A3051AAA1EULL, "6069054_m.fxe"},
	{0xB80C8A231F2ECEF9ULL, "vxht_mansion2_1ti_0_0.vlm"},
	{0xB80C8F3605F72691ULL, "vrc321ip_shade_proj_end.ncs"},
	{0xB80C9545CA4AA86AULL, "hlac_lt_0_lowtown_1_8z.dds"},
	{0xB80D733F8B7DB61EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magb_0d.dds"},
	{0xB80DB2A806795CF5ULL, "qun160_party_comment_16061231.cl"},
	{0xB80DF5B66C991A92ULL, "440324359.ogg"},
	{0xB80F2935C449ABFEULL, "6113005_m.fxe"},
	{0xB80F64B01C779224ULL, "ele.mao"},
	{0xB80FEFEEAD79EDDCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ch_hole.mao"},
	{0xB80FFBEEAD7A027FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ch_hole.mmh"},
	{0xB81001EEAD7A0C21ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ch_hole.msh"},
	{0xB8102F7B53B9E032ULL, "72447554.ogg"},
	{0xB8105A5CC36D432AULL, "6097073_m.fxe"},
	{0xB810984C8B77074BULL, "zz_isa_debug6133192.cl"},
	{0xB810A1BD3E05F8A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_buildup.gad"},
	{0xB810C92380722931ULL, "lowtown_fem_alienage1_sta6162585.cl"},
	{0xB811DC3A651DCF3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female_unconscious.ani"},
	{0xB8125AB3CD1C0370ULL, "6134934_m.fxe"},
	{0xB812BA3C8761068EULL, "vxca_cavern1_haunted_gk9_0_1.vlm"},
	{0xB812FAAC45A37C72ULL, "6156495_m.fxe"},
	{0xB813F06B8D849C33ULL, "llac_lt_0_lowtown_a3_n_c_1_7.dds"},
	{0xB8140326758849CCULL, "fen210_orana6133043.cl"},
	{0xB8148C777E09C041ULL, "6119624_m.fxe"},
	{0xB814D4D5150BB144ULL, "6110213_m.fxe"},
	{0xB814D6EAB5EA0B5DULL, "zz_and_debug6134049.cl"},
	{0xB815157E5C6AD60EULL, "6153016_m.fxe"},
	{0xB8151F36F82E5490ULL, "w_woundedcoast_71_0.mao"},
	{0xB8152336F82E5BA3ULL, "w_woundedcoast_71_0.mmh"},
	{0xB8152D607A44CA89ULL, "character_character_p198.psh.fxo.dx11"},
	{0xB8153136F82E73FDULL, "w_woundedcoast_71_0.msh"},
	{0xB8155F9D30BA89B7ULL, "hlac_lt_0_lowtown_a3_n_c_131071_1z.dds"},
	{0xB815628506D96776ULL, "vxw_blightlands_tm_0_0.vlm"},
	{0xB81593A351079D3FULL, "lt_gallowstemplar_7198.anb"},
	{0xB815A2E66A1C73E7ULL, "hslmatrix.vsh.fxo"},
	{0xB815FBFBAD8AAB30ULL, "vxw_woundedcoast_16z_0_0.vlm"},
	{0xB8160332E37D5A1AULL, "613842587.ogg"},
	{0xB8162C7BCD56602AULL, "art\\characters\\creatures\\ashwraith\\animation\\blendtree\\ashwraith.anb"},
	{0xB81684C5191B2776ULL, "vxw_sundermount_2gp_0_0.vlm"},
	{0xB816C983298CCCCAULL, "6143710_m.fxe"},
	{0xB816F449E7828F14ULL, "143471510.ogg"},
	{0xB81747675D008147ULL, "vxlt_hangedman_n_1qu_0_0.vlm"},
	{0xB8179CBD3E0C0E85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_buildup.evt"},
	{0xB817A0F6556139ECULL, "673504267.ogg"},
	{0xB817CF6E86E143DAULL, "gen00fl_isabela6099307.cl"},
	{0xB817F62C5F7B6AFDULL, "vxca_caverns2_955_0_0.vlm"},
	{0xB81804A4E76DA9B8ULL, "vxht_0_hightown_a2_bov_0_0.vlm"},
	{0xB8185055F71EB3FDULL, "6107102_m.fxe"},
	{0xB818C96B8D29A265ULL, "6016636_f.fxe"},
	{0xB818CF167758BB86ULL, "gen_im_trs_bst_liver_04.uti"},
	{0xB818D1D1482C3206ULL, "hlac_lt_gallowstemplar_7198_6y.dds"},
	{0xB818E93A6524047EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female_unconscious.gad"},
	{0xB81938AF16F9176EULL, "vxlt_hangedman_n_25w_0_0.vlm"},
	{0xB8193E40C1069126ULL, "vxw_encounter2_df_0_0.vlm"},
	{0xB8198B04D67D2777ULL, "890001403.ogg"},
	{0xB81A0763CF4783AEULL, "one120_athenril6113262.cl"},
	{0xB81A0AE6C27618EDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1pp_basic.phy"},
	{0xB81A215C6216B535ULL, "vxdr_slavers_2hg_0_0.vlm"},
	{0xB81A5EC23289F768ULL, "hla_ht_0_hightown_a3_-131074_0z.dds"},
	{0xB81A9C6E5A7C928CULL, "6129107_m.fxe"},
	{0xB81AB9C731812CEFULL, "6027788_f.fxe"},
	{0xB81ACA50DED32540ULL, "lt_0_lowtown_a2_n_36_0_ll.mmh"},
	{0xB81AE88D0596365BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portrait_d.dds"},
	{0xB81B20637F891BECULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm_2.mmh"},
	{0xB81B26637F8925DEULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm_2.msh"},
	{0xB81B559E2C76DFE2ULL, "vxlt_0_lowtown_a3_n_c_hp7_0_0.vlm"},
	{0xB81B7DFE932342A9ULL, "154694383.ogg"},
	{0xB81BBFF91C15805CULL, "123691732.ogg"},
	{0xB81BFA12BF7376D3ULL, "gen00fl_varric6107278.cl"},
	{0xB81C119CC78F539FULL, "vxw_encounter2_b5_0_0.vlm"},
	{0xB81C5A5116BDF437ULL, "llac_w_encounter2_camp_196607_0.dds"},
	{0xB81D8F726FB0E40EULL, "6149092_f.fxe"},
	{0xB81DE5EA54EDE0CEULL, "vxw_bonepitdestroyed_a3_lv_0_0.vlm"},
	{0xB81E2C422E45521BULL, "w_sundermount_196_0.mmh"},
	{0xB81E38422E4566B8ULL, "w_sundermount_196_0.mao"},
	{0xB81E4A422E458505ULL, "w_sundermount_196_0.msh"},
	{0xB81EA2C07D5D6014ULL, "1029054292.ogg"},
	{0xB81F93B6E015758BULL, "vxlt_hangedman_n_2xa_0_0.vlm"},
	{0xB81FB11E1DB3A85CULL, "vxw_encounter2_p1_0_1.vlm"},
	{0xB81FBF5D78BBEFDAULL, "qun200_party_comment6090444.cl"},
	{0xB81FC43A6529E39DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragged_human_female_unconscious.evt"},
	{0xB81FED4164EC14E5ULL, "6134892_m.fxe"},
	{0xB81FF0A5EC3EDF64ULL, "vxlt_0_lowtown_a3_n_c_ejf_0_0.vlm"},
	{0xB820298030A4ABC6ULL, "6103570_m.fxe"},
	{0xB82203603EB6D775ULL, "6129797_m.fxe"},
	{0xB8228430FF686ACAULL, "w_encounter2_camp_50_0.rml"},
	{0xB822A345096A680DULL, "126797171.ogg"},
	{0xB822B1CF916F04CEULL, "6081171_m.fxe"},
	{0xB822BB509A15EA9AULL, "hlac_lt_hangedman_-1_14z.dds"},
	{0xB822C9A42FCE46AAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_alchem_d.dds"},
	{0xB822FCFAA87A2401ULL, "gen00fl_varric6078848.cl"},
	{0xB822FE546FFFD8C1ULL, "6114124_m.fxe"},
	{0xB82354B47AABF984ULL, "vxlt_0_lowtown_docks_b5x_0_1.vlm"},
	{0xB8236CD9DD98EA77ULL, "hlac_w_encounter2_camp_65538_0x.dds"},
	{0xB823875B4E60EE8EULL, "6156316_m.fxe"},
	{0xB823A785E856EB9EULL, "mag170_hugh6206068.cl"},
	{0xB823E2BCEF8504CAULL, "vxw_sundermount_35s_0_0.vlm"},
	{0xB824385C4386D3C8ULL, "vxdr_primevalentrance_b1_0_1.vlm"},
	{0xB824BF0094F0F91DULL, "hlac_lt_0_lowtown_docks_n_-131075_6x.dds"},
	{0xB824C42AF5A41460ULL, "6089487_m.fxe"},
	{0xB8252E86DAB0A88AULL, "6143865_f.fxe"},
	{0xB8254C1B7C5438B5ULL, "gen00fl_varric6206133.cl"},
	{0xB82565A57C8031ADULL, "326476414.ogg"},
	{0xB82578416A3FAC6BULL, "6149739_m.fxe"},
	{0xB8259A204AA695A6ULL, "character_character_p124.psh.fxo.dx11"},
	{0xB825A5CE8B02E3D9ULL, "vxw_woundedcoast_r6_0_0.vlm"},
	{0xB825A7006FAF07EFULL, "combat_dw_r1_a4.gda"},
	{0xB825B6144CE080F6ULL, "hlac_ht_0_hightown_n_65533_2z.dds"},
	{0xB826A956830D0FCCULL, "qun110_petrice6042292.cl"},
	{0xB826EBB473B157C2ULL, "one111ip_ggt_crate_6.utp"},
	{0xB826F27D83694A78ULL, "hlac_dr_0_deeproads_17039_2x.dds"},
	{0xB82748D9F539BE93ULL, "944694795.ogg"},
	{0xB827B40078B0AA47ULL, "6065932_m.fxe"},
	{0xB82850869D1E25A4ULL, "lt_0_lowtown_docks_141_0.msh"},
	{0xB82862869D1E440DULL, "lt_0_lowtown_docks_141_0.mao"},
	{0xB82864A5B17BB319ULL, "78765482.ogg"},
	{0xB82866869D1E4B46ULL, "lt_0_lowtown_docks_141_0.mmh"},
	{0xB8288EC6DA58C759ULL, "hlac_ht_mansion_n_3364_1x.dds"},
	{0xB82890540034EE96ULL, "follower_barks6153887.cl"},
	{0xB829023E20E325BBULL, "vxlt_undercitydungeon_1ob_0_0.vlm"},
	{0xB8295C607A324345ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_9x.dds"},
	{0xB8296DDD031AFFF9ULL, "vxht_mansion2_1bo_0_0.vlm"},
	{0xB8299E45CF34DDF8ULL, "138308166.ogg"},
	{0xB829A530FF6EC11EULL, "w_encounter2_camp_50_0.phy"},
	{0xB829C0EC330DBC24ULL, "t1_mul_re3.tnt"},
	{0xB829C759C19060A9ULL, "hlac_w_bonepitdestroyed_a3_65540_0y.dds"},
	{0xB82A0B57364E0FC5ULL, "6129732_f.fxe"},
	{0xB82A1D9DD150734CULL, "hlac_lt_0_lowtown_a2_0_0y.dds"},
	{0xB82A8379B9856C1FULL, "vxw_encounter2_m4_0_0.vlm"},
	{0xB82A8F80289BF120ULL, "hlac_lt_gallowscourtyard_n_-2_1y.dds"},
	{0xB82B0BE46508331CULL, "dae000_arishok6067830.cl"},
	{0xB82B1271A25CBD07ULL, "828208807.ogg"},
	{0xB82B5048519EFCE2ULL, "m_pro000ar_var_cass_stage.wbd"},
	{0xB82B8EC7FC66A7C1ULL, "vxht_0_hightown_n_c7m_0_0.vlm"},
	{0xB82BE686D506EE36ULL, "lgt300cr_gang_2_lt.utc"},
	{0xB82C043166E45D9EULL, "t3_har_red.tnt"},
	{0xB82C74317B3323E1ULL, "hlac_ht_0_hightown_a3_-1_2y.dds"},
	{0xB82CA828316E3AE7ULL, "zz_slate1_baddy2.cut"},
	{0xB82CC82D456FD409ULL, "448365618.ogg"},
	{0xB82CE05F401D2933ULL, "690658557.ogg"},
	{0xB82D00EAC398F51EULL, "vxht_bartrand_n_667_0_0.vlm"},
	{0xB82D15C8A5FBD9E0ULL, "vxlt_0_lowtown_a2_cow_0_0.vlm"},
	{0xB82D1D8C7DFCCB65ULL, "hlac_w_bonepitmining_a2_393221_0y.dds"},
	{0xB82D2336F842EDC0ULL, "w_woundedcoast_71_0.rml"},
	{0xB82DBC16DAB14D3BULL, "vxlt_0_lowtown_a3_n_c_hnh_0_0.vlm"},
	{0xB82E3231D9B99FA5ULL, "vxw_blightlands_h6_0_0.vlm"},
	{0xB82E6282DE34F7E4ULL, "vxlt_0_lowtown_a3_n_c_emg_0_0.vlm"},
	{0xB82E7A68AD081B9BULL, "6139869_f.fxe"},
	{0xB82ECB0785E87F29ULL, "224145009.ogg"},
	{0xB82F009BB0757450ULL, "hlac_lt_gallowstemplar_2388_3z.dds"},
	{0xB82F2315642BC525ULL, "hlac_lt_0_lowtown_docks_n_196607_0y.dds"},
	{0xB83113A6580C6DADULL, "6163768_m.fxe"},
	{0xB8311EFFB8B7E359ULL, "plt_cod_ltr_sk3_scribbles4.nss"},
	{0xB8313A10E1A0000BULL, "hlac_lt_0_lowtown_docks_n_-65539_1z.dds"},
	{0xB83167EC65B309EFULL, "6198711_m.fxe"},
	{0xB831703161F5F141ULL, "6113024_m.fxe"},
	{0xB831F1055D1E10F3ULL, "llac_w_bonepitdestroyed_a3_5_0.dds"},
	{0xB83201C94D5323B0ULL, "hlac_ht_mansion2b_818_59z.dds"},
	{0xB8325D5F746865BEULL, "vxlt_gallowscourtyard_66n_0_0.vlm"},
	{0xB83363F95616DA15ULL, "438357566.ogg"},
	{0xB833687EFD075876ULL, "vxw_sundermount_2mr_0_0.vlm"},
	{0xB833ACEF0FD57E98ULL, "gen00fl_aveline6131396.cl"},
	{0xB833DEE1BF302F2CULL, "dae000pt_amb_sunder.ncs"},
	{0xB834A2D9C66D117FULL, "108027023.ogg"},
	{0xB835392D0E7D2545ULL, "vxdr_primevaltomb_7p2_0_0.vlm"},
	{0xB835D1EB4DC0C8C6ULL, "vxlt_0_lowtown_a2_i4f_0_0.vlm"},
	{0xB835FBC5E226C24EULL, "hlac_lt_0_lowtown_docks_a3_n_-4_1z.dds"},
	{0xB83621FAA4F2C226ULL, "vxw_encounter2_jh_0_0.vlm"},
	{0xB836BB310BF2D3B7ULL, "llac_ca_cavern1_haunted_24426_12.dds"},
	{0xB837A3E09CBDA0AFULL, "6187925_f.fxe"},
	{0xB837DC01347E5531ULL, "59192562.ogg"},
	{0xB83805FFB8BDD732ULL, "plt_cod_ltr_sk3_scribbles4.plo"},
	{0xB8389E4EC0A219CBULL, "6075168_m.fxe"},
	{0xB838F556AB7B849FULL, "vxw_encounter2_h7_0_0.vlm"},
	{0xB83904F0668BE581ULL, "105936920.ogg"},
	{0xB83982571750C382ULL, "6127280_m.fxe"},
	{0xB83989C20EFF1B92ULL, "mrl310_merrill_pt3.stg"},
	{0xB839DACA79CEDBA2ULL, "25300352.ogg"},
	{0xB83A695604EFC29BULL, "6118355_m.fxe"},
	{0xB83ACDB2068FE27EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_wes_flemeth_real.ani"},
	{0xB83BA6896DEB1220ULL, "follower_banter6153994.cl"},
	{0xB83BAF2D2DE9A435ULL, "vxlt_0_lowtown_docks_cs7_0_0.vlm"},
	{0xB83C2DF3CCA765BBULL, "vxlt_0_lowtown_a3_n_c_hkk_0_0.vlm"},
	{0xB83C70F3C3B8C64EULL, "w_encounter1_mmap.dds"},
	{0xB83C9AA6046CAAAAULL, "vxht_keep_5ea_0_0.vlm"},
	{0xB83CA40ECBAFB825ULL, "vxw_blightlands_m5_0_0.vlm"},
	{0xB83CD45FD02B1064ULL, "vxlt_0_lowtown_a3_n_c_epd_0_0.vlm"},
	{0xB83D03795C411C63ULL, "mag100_karras6052920.cl"},
	{0xB83D1119B0FC0700ULL, "character_character_p29.psh.fxo.dx11"},
	{0xB83D47A07EF7FE7BULL, "6207974_m.fxe"},
	{0xB83DD21270983FC5ULL, "art\\characters\\creatures\\gallowsgateguard\\w_ggg.mao"},
	{0xB83DF3BA3421F7DCULL, "zz_qcr_debug6055519.cl"},
	{0xB83E3C17FF47D640ULL, "mag170_hugh6080754.cl"},
	{0xB83FB4448E56B9DDULL, "mag200_feynriel6023880.cl"},
	{0xB83FBE24CF7BCAFCULL, "dae000_male_commoner2_mob6174442.cl"},
	{0xB8405C7F3B6F5BB8ULL, "vxlt_0_lowtown_a2_hfw_0_0.vlm"},
	{0xB840898088292E06ULL, "vxlt_0_lowtown_a2_hf9_0_0.vlm"},
	{0xB840F198C23A6906ULL, "vxw_bonepitmining_a2_1x2_0_1.vlm"},
	{0xB8413BB48C8803BAULL, "514605180.ogg"},
	{0xB8414C86015760D0ULL, "lt_gallowstemplar_f_mmi.gff"},
	{0xB841D6D67537B1EDULL, "follower_banter6167226.cl"},
	{0xB841F7D070301D98ULL, "vxht_mansion_n_60h_0_0.vlm"},
	{0xB8429C944AACCD1EULL, "vxht_mansion2b_2lq_0_0.vlm"},
	{0xB842AEE082155677ULL, "vxlt_0_lowtown_a3_n_c_hdb_0_0.vlm"},
	{0xB842B262E95C02D4ULL, "dae000_sandal6148236.cl"},
	{0xB8430F8A8132F5FCULL, "358910983.ogg"},
	{0xB8431482777CBABDULL, "vxca_caverns1_gk2_0_0.vlm"},
	{0xB84320CC6BE0A86EULL, "vxht_bartrand_n_4gr_0_0.vlm"},
	{0xB844993BD7DF88C1ULL, "382411019.ogg"},
	{0xB844D4C841E0F322ULL, "6096070_m.fxe"},
	{0xB845AC13B2D0F2E8ULL, "6164988_m.fxe"},
	{0xB845ADB8398A71E7ULL, "280162330.ogg"},
	{0xB845D8554C755A55ULL, "vxlt_0_lowtown_a2_n_ex6_0_0.vlm"},
	{0xB8465196E70730F9ULL, "vxht_mansion2_1hm_0_0.vlm"},
	{0xB8466F5771E08EE9ULL, "vxlt_hangedman_n_210_0_0.vlm"},
	{0xB846B159413E908CULL, "565684471.ogg"},
	{0xB846DC927F1B186FULL, "follower_banter6154024.cl"},
	{0xB84799A3680AD71BULL, "mag131_taraneh6088702.cl"},
	{0xB847A97753BF93D1ULL, "705780932.ogg"},
	{0xB848B10B9DDA1508ULL, "580213832.ogg"},
	{0xB848D32DC409039EULL, "117324394.ogg"},
	{0xB84927EF8EB6936FULL, "6151356_m.fxe"},
	{0xB849EEC3840CF950ULL, "dae000_merchant_tailor_hm6164354.cl"},
	{0xB84A32B87B44E03BULL, "hlac_lt_backalley_g_0_14x.dds"},
	{0xB84A442572D4AAD3ULL, "6061244_m.fxe"},
	{0xB84A7B149530BB5BULL, "6144530_m.fxe"},
	{0xB84A9E29A9B0A4A9ULL, "239399585.ogg"},
	{0xB84AB1C3B7F3AC78ULL, "hlac_dr_slavers_3215_1x.dds"},
	{0xB84AB56EF3D87F5CULL, "dae000ip_crowd_lm4_talk11.utp"},
	{0xB84B0789D70E2694ULL, "6135817_f.fxe"},
	{0xB84B15EBBDA5D0A5ULL, "vxw_blightlands_v4_0_0.vlm"},
	{0xB84B20EBBCE94F2CULL, "vxlt_0_lowtown_n_gjs_0_0.vlm"},
	{0xB84BC0E13C2CC23AULL, "mcr370_negotiation6078717.cl"},
	{0xB84C42EEC887C39CULL, "6165284_m.fxe"},
	{0xB84C92A3ED2A58D3ULL, "one100dg_kelder6060683.cl"},
	{0xB84D45EB2B3F3844ULL, "hlac_dr_primevalentrance_1726_13x.dds"},
	{0xB84D7E32628371CCULL, "vxlt_blackemporium_uh_0_0.vlm"},
	{0xB84E02B496984983ULL, "781480926.ogg"},
	{0xB84E1FEDB958F43CULL, "6156747_m.fxe"},
	{0xB84EE83AB0A15FC6ULL, "vfx_vfx_p5.psh.fxo.dx11.assembly"},
	{0xB84F25AA8F6E389FULL, "vxca_caverns2_62q_0_0.vlm"},
	{0xB84FB8179AA1F3DEULL, "vxht_keep_n_62p_0_0.vlm"},
	{0xB84FE6B206A268F8ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_wes_flemeth_real.gad"},
	{0xB84FF7216581F0A5ULL, "178724073.ogg"},
	{0xB84FF8F3BC5E489AULL, "follower_banter6143221.cl"},
	{0xB85022C095C6E336ULL, "6143624_m.fxe"},
	{0xB850233D2F598556ULL, "mer220_hubert6050017.cl"},
	{0xB8506BBAC9FE40DEULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyo_0s.dds"},
	{0xB850C68793C218A5ULL, "6121960_m.fxe"},
	{0xB8510E713CA2E59EULL, "vxht_mansion2b_2ir_0_0.vlm"},
	{0xB8514F7843B2179FULL, "dae120st_merch_tailor.utm"},
	{0xB851865F6972D33DULL, "vxca_caverns1_gf1_0_0.vlm"},
	{0xB851B0763BD26617ULL, "161440690.ogg"},
	{0xB85218A7D9EA6DE4ULL, "hlac_lt_0_lowtown_a2_1_10y.dds"},
	{0xB85238F38F6C4168ULL, "6126699_f.fxe"},
	{0xB852A0386CFB5D97ULL, "vxca_cavern1_haunted_j6k_0_0.vlm"},
	{0xB85305B488DEF326ULL, "vxw_encounter2_pj_0_0.vlm"},
	{0xB8531DD74B2E6217ULL, "435429172.ogg"},
	{0xB853DF51DB941A71ULL, "dae300_vashothlead6062763.cl"},
	{0xB8540C71087289BCULL, "ht_0_hightown_a3_21_0.anb"},
	{0xB8544F15F32DDD73ULL, "hlac_lt_0_lowtown_a2_-196608_1x.dds"},
	{0xB8545819DABFC71AULL, "6057033_m.fxe"},
	{0xB8546E600CDF7209ULL, "follower_banter6173155.cl"},
	{0xB8551D11F44F41E2ULL, "vxlt_0_lowtown_a3_n_c_hl3_0_0.vlm"},
	{0xB855D00A60475734ULL, "7731154.ogg"},
	{0xB8561F6B2D18D0CBULL, "873112034.ogg"},
	{0xB85621BC8BFB43EEULL, "100047842.ogg"},
	{0xB8566553584A6046ULL, "vxht_brothel_n_5gc_0_0.vlm"},
	{0xB857C3790604EB41ULL, "lt_darktown_probe_67i.dds"},
	{0xB857DD95244F6A3BULL, "vxca_caverns2prc_a01_0_0.vlm"},
	{0xB858411AEC59CECEULL, "497728074.ogg"},
	{0xB8585D77AB3D093DULL, "956552517.ogg"},
	{0xB8587AC208EA093FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch_1.mao"},
	{0xB858AFBFE4888DFFULL, "vxlt_0_lowtown_docks_a3_n_aln_0_0.vlm"},
	{0xB85905F51A7C9747ULL, "383168428.ogg"},
	{0xB8590C603770F03CULL, "859505756.ogg"},
	{0xB859FBF59DE99A6CULL, "6163897_m.fxe"},
	{0xB859FFB3029D9BD8ULL, "art\\effects\\cloudshadows\\w_woundedcoast\\cloud2.dds"},
	{0xB85A749C7DB52759ULL, "art\\characters\\weapons\\staff\\w_stf_001a.msh"},
	{0xB85A7E9C7DB53847ULL, "art\\characters\\weapons\\staff\\w_stf_001a.mmh"},
	{0xB85A829C7DB53E94ULL, "art\\characters\\weapons\\staff\\w_stf_001a.mao"},
	{0xB85A90C5CED213B3ULL, "hlac_lt_warehouse_65536_2y.dds"},
	{0xB85AECCA31341B28ULL, "hlac_ca_caverns1_24424_26y.dds"},
	{0xB85BB314EE95B156ULL, "art\\vfx\\combat\\vfx_r_dwp_rightb_close1_c.impa.evt"},
	{0xB85C0442FC3DC02EULL, "gen00fl_carver6108924.cl"},
	{0xB85C5790783C18A8ULL, "dae000_party_night_comment6201627.cl"},
	{0xB85C5A16C1DF3EB9ULL, "vxht_mansion2b_2q1_0_0.vlm"},
	{0xB85CCEABCA38FFA2ULL, "vxht_brothel_n_4tg_0_0.vlm"},
	{0xB85D3357D2345D48ULL, "6103572_m.fxe"},
	{0xB85D40391F5B8CB8ULL, "vxlt_0_lowtown_a2_hhu_0_0.vlm"},
	{0xB85DBE3FEB528045ULL, "hlac_lt_0_lowtown_docks_65535_0y.dds"},
	{0xB85E0C12F74EB480ULL, "6135366_m.fxe"},
	{0xB85E29D40EB64159ULL, "qun211_party_comment6072267.cl"},
	{0xB85E983C5E1AA81FULL, "6033590_f.fxe"},
	{0xB85E98A657DA6B8FULL, "mag211_marethari6025805.cl"},
	{0xB85F2630112BFD99ULL, "vxca_cave_2k7_0_0.vlm"},
	{0xB85F3355E71F67CEULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkback.ani"},
	{0xB85F64E4559F2B4AULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs106a_0n.dds"},
	{0xB85F804E2E98FE1EULL, "vxht_mansion2b_2bs_0_0.vlm"},
	{0xB85FB9E8BABA3630ULL, "fx_dust007.dds"},
	{0xB85FD1A255BEA77CULL, "6072330_f.fxe"},
	{0xB86004864FCCD96EULL, "vxht_bartrand_n_4qt_0_0.vlm"},
	{0xB8604E4D5A5F3995ULL, "hlac_ca_cavern1_haunted_24423_6z.dds"},
	{0xB86097B6B05A9966ULL, "6162643_m.fxe"},
	{0xB860C606961DAA82ULL, "852616235.ogg"},
	{0xB860FFFBCA18BF1BULL, "vxht_0_hightown_a3_caw_0_0.vlm"},
	{0xB86136FE7E8D3CA0ULL, "272899973.ogg"},
	{0xB8614B8C0891D9D2ULL, "6165365_m.fxe"},
	{0xB8615E2A3AE8A923ULL, "976747944.ogg"},
	{0xB8619F7F2C7DF1ABULL, "one100_solivitus6048135.cl"},
	{0xB8622502F81D584BULL, "hlac_ca_caverns1_mines_24422_1x.dds"},
	{0xB8627CF99DA38802ULL, "970238502.ogg"},
	{0xB862AA14EE9BC0E1ULL, "art\\vfx\\combat\\vfx_r_dwp_rightb_close1_c.impa.gad"},
	{0xB862D5596E01A8B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\revive_reaction_temp.ani"},
	{0xB862F2BA28225A6EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_magb_0t.dds"},
	{0xB86306A3641F4B68ULL, "mag311im_bsc_notes.uti"},
	{0xB86323770546E749ULL, "77881396.ogg"},
	{0xB86344AF4A96D41AULL, "hlac_lt_gallowstemplar_f_2384_4z.dds"},
	{0xB8639BD930160BB0ULL, "nat380_nathaniel_banter6067294.cl"},
	{0xB863A42F1E7C5B01ULL, "235317330.ogg"},
	{0xB8646C8668A31CCDULL, "w_bonepitdestroyed_a3_238_0.msh"},
	{0xB8647A8668A33400ULL, "w_bonepitdestroyed_a3_238_0.mao"},
	{0xB8647E8668A33B53ULL, "w_bonepitdestroyed_a3_238_0.mmh"},
	{0xB864A76556FBCE71ULL, "hlac_w_bonepitmining_a2_262148_0x.dds"},
	{0xB864D7304A4078C6ULL, "vxht_brothel_n_5jd_0_0.vlm"},
	{0xB864DCE236B68A73ULL, "gen00fl_bethany6101030.cl"},
	{0xB8654D422E818D50ULL, "w_sundermount_196_0.phy"},
	{0xB86551582BC3C83EULL, "6174705_m.fxe"},
	{0xB8658F73A3603853ULL, "160527794.ogg"},
	{0xB86623699940DDA8ULL, "6088783_m.fxe"},
	{0xB86635AE803FAAEAULL, "195316818.ogg"},
	{0xB8665ECCC92F57C1ULL, "vxht_0_hightown_a2_c1x_0_0.vlm"},
	{0xB866E69FC868DF5DULL, "zz_seb_debug6189572.cl"},
	{0xB867219CD67EA67FULL, "vxlt_0_lowtown_docks_a3_n_aqi_0_0.vlm"},
	{0xB86799AF268CB190ULL, "gen00fl_sebastian6179758.cl"},
	{0xB8679CF26D0FABE4ULL, "6114005_f.fxe"},
	{0xB867F167FA5DB12EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_enter.ani"},
	{0xB868AB14EEA02D8FULL, "art\\vfx\\combat\\vfx_r_dwp_rightb_close1_c.impa.ani"},
	{0xB869149941115AFCULL, "ch1armtnt_s1.psh.fxo"},
	{0xB86ACBF3B3D55739ULL, "vxht_mansion2b_2j0_0_0.vlm"},
	{0xB86AF47CE9416EAAULL, "vxht_0_hightown_c79_0_0.vlm"},
	{0xB86C35D524D499D5ULL, "258161495.ogg"},
	{0xB86C7DEFE944CD00ULL, "6130066_m.fxe"},
	{0xB86D2B55E72B971FULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkback.evt"},
	{0xB86D6E0801E17E83ULL, "354955675.ogg"},
	{0xB86D980D03221619ULL, "vxca_cave_2l6_0_0.vlm"},
	{0xB86DD0862E9DC0A0ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_0x.dds"},
	{0xB86E6DE405AFCD8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_melee_attk_v2.ani"},
	{0xB86E7D84DF3B6903ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\twitchv1.evt"},
	{0xB86F55CA1E7B74AAULL, "hlac_lt_0_lowtown_n_131070_2x.dds"},
	{0xB86FE28C5A78AF3CULL, "zz_mag_debug6050908.cl"},
	{0xB87052ABD072A23AULL, "dae000_quintus6154709.cl"},
	{0xB870ACD6BC0809DCULL, "vxw_woundedcoast_17e_0_0.vlm"},
	{0xB870DD596E0DF3A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\revive_reaction_temp.evt"},
	{0xB870EEFF3EBE8D85ULL, "gencr_bronto.utc"},
	{0xB871C181755A3C58ULL, "fen110_fenris6118054.cl"},
	{0xB871E2FF35CF86CFULL, "838755948.ogg"},
	{0xB87218F90FC93225ULL, "mcr370cr_guardsman.utc"},
	{0xB8724CCFC4106879ULL, "940558653.ogg"},
	{0xB8725E4851DB1698ULL, "m_pro000ar_var_cass_stage.bnk"},
	{0xB8726075B489C51BULL, "mrl351_pridedemon6170227.cl"},
	{0xB87271B706210EAEULL, "gen00fl_isabela_init6076671.cl"},
	{0xB872E628F3A1D5A9ULL, "205123486.ogg"},
	{0xB87304EEADCE8EC4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ch_hole.phy"},
	{0xB8731E14DC8DC477ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_013a_0d.dds"},
	{0xB873490D3C369146ULL, "vxht_brothel_n_5me_0_0.vlm"},
	{0xB873A0AB04B8A7F3ULL, "vxht_keep_64h_0_0.vlm"},
	{0xB8740C55E7318108ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualattkback.gad"},
	{0xB874112973F5D8D7ULL, "589004384.ogg"},
	{0xB8742D969B3685E5ULL, "ts_demonic_acc_table.gda"},
	{0xB87469E1F4B224C0ULL, "158905129.ogg"},
	{0xB874916BEB7E9C1EULL, "hlac_dr_slavers_3215_0y.dds"},
	{0xB8755684DF414584ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\twitchv1.gad"},
	{0xB875C7BC42F20583ULL, "vxlt_0_lowtown_docks_a3_bu1_0_0.vlm"},
	{0xB875D45BC2D1BAD1ULL, "dae000_arishok6068447.cl"},
	{0xB875E967FA69E17FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_enter.evt"},
	{0xB876198C1613413DULL, "366320272.ogg"},
	{0xB876B0A7B0A1FA02ULL, "6139629_m.fxe"},
	{0xB876FE596E1297F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\revive_reaction_temp.gad"},
	{0xB87730B6D0667D21ULL, "6174056_m.fxe"},
	{0xB87738D35884F4DCULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.impa.ani"},
	{0xB8778AAE2BC6EB68ULL, "w_bonepit_176_0.msh"},
	{0xB8779CAE2BC709F9ULL, "w_bonepit_176_0.mao"},
	{0xB877A8AE2BC71D92ULL, "w_bonepit_176_0.mmh"},
	{0xB878246286B5FB9AULL, "scaleform_sf_main_p2.vsh.fxo.dx11"},
	{0xB8786F4E04477EB9ULL, "and111_lirene6084360.cl"},
	{0xB878C8974EE0E66FULL, "488342426.ogg"},
	{0xB87933EDDF82FBEAULL, "6167635_m.fxe"},
	{0xB8799022D197DD55ULL, "fen310_fenris6118200.cl"},
	{0xB879B0ADF8C80C1DULL, "312043231.ogg"},
	{0xB879EC504BDC2196ULL, "56757484.ogg"},
	{0xB879FD31CD5AE12BULL, "ht_0_hightown_n_29_0_lc.mmh"},
	{0xB87A0BB7F4CEA255ULL, "vxlt_0_lowtown_docks_n_bes_0_0.vlm"},
	{0xB87A3AB061CF573CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage5_.gad"},
	{0xB87B011E3EB66BEEULL, "593365652.ogg"},
	{0xB87B1F8DDF5F32E2ULL, "follower_banter6164781.cl"},
	{0xB87B55663482A6ADULL, "1060399580.ogg"},
	{0xB87B65E405BA4ADFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_melee_attk_v2.evt"},
	{0xB87B7436387D4127ULL, "284557194.ogg"},
	{0xB87C16D735EA1248ULL, "l_dae310ar_docks_nt.bnk"},
	{0xB87C6584DF477E62ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\twitchv1.ani"},
	{0xB87C89043B6140F4ULL, "llac_ht_0_hightown_n_65536_1.dds"},
	{0xB87C94CFCBB9A922ULL, "vxlt_0_lowtown_a3_n_c_cms_0_0.vlm"},
	{0xB87C99FF535FFD6AULL, "map_wilderness_1280x1024.dds"},
	{0xB87CCA67FA6FCB68ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_enter.gad"},
	{0xB87D37BED787DFB8ULL, "6020060_m.fxe"},
	{0xB87D45D35889761EULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.impa.gad"},
	{0xB87D75A35121A675ULL, "6064379_m.fxe"},
	{0xB87E6DFFDB91014FULL, "hla_ht_0_hightown_n_131072_0x.dds"},
	{0xB87E79135092FDC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_loop.ani"},
	{0xB87EA1F57DFEEEA6ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.dura.ani"},
	{0xB87EEFE668413685ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_6x.dds"},
	{0xB87F353E125AFD75ULL, "dae000_serendipity6161428.cl"},
	{0xB87F3E903DED5EE2ULL, "vxca_caverns2prc_a6j_0_0.vlm"},
	{0xB8804E9079974AEFULL, "vxlt_0_lowtown_n_eoy_0_0.vlm"},
	{0xB880D0FB932CEF54ULL, "gen00fl_merrill6128742.cl"},
	{0xB8811E73267D462BULL, "lt_0_lowtown_21_0.anb"},
	{0xB8812B80352551AFULL, "hlac_lt_0_lowtown_docks_a3_n_65535_3x.dds"},
	{0xB88139B061D5743AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage5_.ani"},
	{0xB881AB1E168CF38AULL, "293978044.ogg"},
	{0xB8821287F6AEC073ULL, "vxht_keep_61k_0_0.vlm"},
	{0xB88214DBBAB4CEAAULL, "vxlt_gallowstemplar_457_0_0.vlm"},
	{0xB88246E405C034C8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_melee_attk_v2.gad"},
	{0xB88307237D2D0EA8ULL, "6086983_m.fxe"},
	{0xB8839C2C5DDEBFBDULL, "vxlt_0_lowtown_docks_a3_d06_0_0.vlm"},
	{0xB883B7F6F6C441F8ULL, "vxlt_0_lowtown_a2_ci5_0_0.vlm"},
	{0xB883DCFD6995C4D9ULL, "vxca_caverns1_ec0_0_0.vlm"},
	{0xB88420D3588F563DULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.impa.evt"},
	{0xB88494135097974BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_loop.gad"},
	{0xB8851A549A624D14ULL, "and100_anders6095936.cl"},
	{0xB885F3F7418202CCULL, "vxht_mansion2b_3gv_0_1.vlm"},
	{0xB886AC8484439E1BULL, "894751652.ogg"},
	{0xB886E8FA6DB35662ULL, "zz_vrc_debug6064284.cl"},
	{0xB887D646AAB2089DULL, "6135885_f.fxe"},
	{0xB88824CE4036D71AULL, "6127406_m.fxe"},
	{0xB8887D94E6C4BAD5ULL, "vxlt_0_lowtown_docks_n_bbr_0_0.vlm"},
	{0xB88934A73D3D70CDULL, "dae000_merchant_access_hf6162982.cl"},
	{0xB889484486A36BAEULL, "pro000_friedrich6051024.cl"},
	{0xB8894EE004DFF566ULL, "ser121_jethann6107111.cl"},
	{0xB8897923966CE615ULL, "dae000_flemeth6000874.cl"},
	{0xB8897EC8DE0F3531ULL, "vxht_mansion2_1qr_0_0.vlm"},
	{0xB889E8BF30B27F58ULL, "pro000_merchant6051280.cl"},
	{0xB88A8E4937952F59ULL, "6135607_f.fxe"},
	{0xB88AE78E8ACEFDA5ULL, "llac_w_bonepit_3_0.dds"},
	{0xB88AE842AEB9F805ULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.anb"},
	{0xB88AEFBCC179D3D8ULL, "dae200cr_viscount6087167.cl"},
	{0xB88B0E58C06BD750ULL, "vxlt_0_lowtown_a2_fgj_0_0.vlm"},
	{0xB88B5515E30AAA72ULL, "dae000_madam_lusine6152277.cl"},
	{0xB88B89F57E095017ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.dura.evt"},
	{0xB88B9113509DB0B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_loop.evt"},
	{0xB88BFADB73DBEC78ULL, "hlac_lt_0_lowtown_a2_n_131071_1z.dds"},
	{0xB88C1D98E62B6BB6ULL, "hlac_lt_0_lowtown_docks_-196607_0z.dds"},
	{0xB88C220916E6C17EULL, "vxlt_0_lowtown_docks_cq2_0_0.vlm"},
	{0xB88C7974F7679ECEULL, "isa211_isabela6175169.cl"},
	{0xB88C8BA71709BF36ULL, "mcr371_hostile_fenris6082220.cl"},
	{0xB88CAC9E6AF25702ULL, "w_blightlands_175_0.msh"},
	{0xB88CAE9E6AF25AB8ULL, "w_blightlands_175_0.mmh"},
	{0xB88CBA9E6AF26F5BULL, "w_blightlands_175_0.mao"},
	{0xB88D0D34EA8BE9CAULL, "vxlt_darktown_2qe_0_0.vlm"},
	{0xB88D20847AFB89A6ULL, "vxw_blightlands_6c_0_0.vlm"},
	{0xB88D51B54C89108BULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax107a_0t.dds"},
	{0xB88D67726F3079BBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8ma.mmh"},
	{0xB88D73726F308E58ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8ma.mao"},
	{0xB88D85726F30ACA5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp8ma.msh"},
	{0xB88EC06D6B8D636FULL, "vxlt_0_lowtown_n_ehx_0_0.vlm"},
	{0xB88F0F0314B1CEBFULL, "6140862_m.fxe"},
	{0xB88F15CA8B416C4BULL, "717881443.ogg"},
	{0xB88F1F83C4E861D5ULL, "zz_prmpt_debug6198252.cl"},
	{0xB88FB852138BA312ULL, "vxlt_house_iy_0_0.vlm"},
	{0xB88FC9069091E581ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_pc_flemeth_fake.ani"},
	{0xB88FD4D7A4830E9BULL, "one110cr_harimann_guard_3.utc"},
	{0xB8904CCB24D98383ULL, "904079976.ogg"},
	{0xB8906CA895F8C75EULL, "hlac_lt_0_lowtown_a2_n_c_131071_5z.dds"},
	{0xB890701628FCEFF0ULL, "qun111_ketojan6039178.cl"},
	{0xB890C255495DFA0FULL, "6090301_m.fxe"},
	{0xB890F3F10886856AULL, "6126874_m.fxe"},
	{0xB8910110E20C000CULL, "739350007.ogg"},
	{0xB89119532BC99CB2ULL, "312611134.ogg"},
	{0xB8920E094FD4D83DULL, "vxlt_0_lowtown_docks_a3_d55_0_0.vlm"},
	{0xB8921BC454878E97ULL, "hlac_lt_foundry_n_320_11x.dds"},
	{0xB8923CBB471F8C43ULL, "llac_ht_keep_3732_5.dds"},
	{0xB8928AF57E0F70A0ULL, "art\\vfx\\combat\\vfx_w_bsk_haste_loop_c.dura.gad"},
	{0xB892C57A5843D0AEULL, "6053603_f.fxe"},
	{0xB8931DDD771A1851ULL, "6177898_f.fxe"},
	{0xB893A27C12B32A71ULL, "617407060.ogg"},
	{0xB893C7EE1C4D2EBAULL, "dre000_sibling_has_blight6036470.cl"},
	{0xB8940DA1D90A665AULL, "502987524.ogg"},
	{0xB894899EFC5CE0FBULL, "vxdr_primevaltomb_6wd_0_0.vlm"},
	{0xB8950B7108AA1F4CULL, "ht_0_hightown_a3_21_0.rml"},
	{0xB895D40690966423ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_pc_flemeth_fake.gad"},
	{0xB89637208B17F6C0ULL, "6139890_m.fxe"},
	{0xB8964509F6A9C389ULL, "780555397.ogg"},
	{0xB89678ACD616CCF7ULL, "144576088.ogg"},
	{0xB896A590AF73AC9DULL, "vxht_keep_n_6g8_0_0.vlm"},
	{0xB897F0A5D0054DB1ULL, "vxht_mansion2_1tq_0_0.vlm"},
	{0xB897F5F24465CFF4ULL, "mrl350pt_newpath.ncs"},
	{0xB897F7C0D64E9884ULL, "6037320_f.fxe"},
	{0xB8987B9B3921A88BULL, "hlac_dr_0_deeproads_5845_17y.dds"},
	{0xB8988C8D26DEF097ULL, "vxw_woundedcoast_e9_0_0.vlm"},
	{0xB898C93FEF02CDCFULL, "vxca_caverns1_mines_iz8_0_0.vlm"},
	{0xB899C7DA4E90A71EULL, "705446491.ogg"},
	{0xB899D5423DBF2CDFULL, "hlac_w_bonepitmining_a2_0_13y.dds"},
	{0xB89AC7EE6304C07EULL, "lt_0_lowtown_docks_n_40_0.rml"},
	{0xB89ACE9AA0C3CB72ULL, "hlac_w_encounter2_camp_-131068_0z.dds"},
	{0xB89B5058FFA20484ULL, "6069752_f.fxe"},
	{0xB89B7F11DC82024AULL, "vxlt_darktown_2tf_0_0.vlm"},
	{0xB89CDCCF47D3F785ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_2.dds"},
	{0xB89D4176F73E9929ULL, "6125151_m.fxe"},
	{0xB89D53BDE35EDD4AULL, "gen00fl_aveline6122101.cl"},
	{0xB89D7B0264FFA6D0ULL, "6075822_m.fxe"},
	{0xB89D87D5C8CE9CF4ULL, "895979647.ogg"},
	{0xB89E2A2F0581BB92ULL, "vxlt_house_dz_0_0.vlm"},
	{0xB89E74D49D738069ULL, "gencr_templar_archer_f.utc"},
	{0xB89E8BDD9104F705ULL, "tablewood_01.mao"},
	{0xB89E940BDABE7ADBULL, "dre000_vampire6034408.cl"},
	{0xB89F1B33355263E8ULL, "gen00fl_aveline6139305.cl"},
	{0xB89FFF13F3A0AB86ULL, "dre000_sibling_has_blight6036531.cl"},
	{0xB8A0BADE6E859506ULL, "6109330_m.fxe"},
	{0xB8A0C0B74D81F5D9ULL, "vxca_caverns1_ei6_0_0.vlm"},
	{0xB8A0CF41F1D77EC7ULL, "dae000_gamlen6145125.cl"},
	{0xB8A108D7360916F2ULL, "l_dae310ar_docks_nt.wbd"},
	{0xB8A1FC2EE69F651AULL, "characterrecord.gfx"},
	{0xB8A2035B093D8829ULL, "6104433_m.fxe"},
	{0xB8A21303279D4AE7ULL, "dae100_alistair_drunk6135107.cl"},
	{0xB8A21726D3140322ULL, "38502765.ogg"},
	{0xB8A21DEF230CDE2BULL, "6124792_m.fxe"},
	{0xB8A2217926E2774AULL, "llac_lt_0_lowtown_docks_a3_n_-65538_0.dds"},
	{0xB8A222ABB71BD2D7ULL, "lt_0_lowtown_a3_n_c_21_0_ll.mmh"},
	{0xB8A278D2C8E62338ULL, "gen_im_npc_arm_hum_hlm_tmpb.uti"},
	{0xB8A292D4ABA4B046ULL, "6169097_m.fxe"},
	{0xB8A30FF752E8B80AULL, "vxlt_warehouse_19c_0_0.vlm"},
	{0xB8A39646CACC7091ULL, "vxca_caverns1_mines_jfm_0_0.vlm"},
	{0xB8A3E68A32E2954BULL, "691634024.ogg"},
	{0xB8A418F2392C71FDULL, "hlac_lt_undercitydungeon_-1_0x.dds"},
	{0xB8A5176DA169C51DULL, "vxht_keep_n_6b7_0_0.vlm"},
	{0xB8A53CCA086F7C51ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm1.ani"},
	{0xB8A5597C2F31FBB0ULL, "w_blightlands_33_0.mao"},
	{0xB8A55D7C2F3202C3ULL, "w_blightlands_33_0.mmh"},
	{0xB8A5614ECAB0EBD5ULL, "vxlt_0_lowtown_docks_n_btp_0_0.vlm"},
	{0xB8A56B7C2F321A1DULL, "w_blightlands_33_0.msh"},
	{0xB8A5A3EB955D25E1ULL, "gen00fl_isabela_init6076487.cl"},
	{0xB8A60AD0EAF43F9CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockexxl.mao"},
	{0xB8A616D0EAF4543FULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockexxl.mmh"},
	{0xB8A61B93BA6E5C80ULL, "6147213_m.fxe"},
	{0xB8A61CD0EAF45EE1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockexxl.msh"},
	{0xB8A6372283B2A01AULL, "mer141_jansen6031839.cl"},
	{0xB8A641A8112A2682ULL, "vxlt_0_lowtown_a2_n_gnm_0_0.vlm"},
	{0xB8A6B6AE777DA3E3ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha08a.mmh"},
	{0xB8A6C4AE777DBB3DULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha08a.msh"},
	{0xB8A6CD62ACBB494FULL, "pro000_meeran6050077.cl"},
	{0xB8A6FE6A18D50917ULL, "vxw_woundedcoast_b8_0_0.vlm"},
	{0xB8A73EFCF705523DULL, "vxca_cave_14p_0_1.vlm"},
	{0xB8A79E348117A13AULL, "6067904_m.fxe"},
	{0xB8A7BFA1DD123B61ULL, "zz_spy130_darktown_res_3.cnv"},
	{0xB8A8859893E4FE9BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf04a.msh"},
	{0xB8A89DBA0238B033ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_killed.ani"},
	{0xB8A8A39893E53185ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf04a.mmh"},
	{0xB8A8E7781F217246ULL, "mag360_samson6042040.cl"},
	{0xB8A9735DB8430193ULL, "llac_w_bonepit_393219_0.dds"},
	{0xB8AA205D7135394FULL, "6074383_m.fxe"},
	{0xB8AB47CA0873FAF3ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_napalm1.gad"},
	{0xB8ABD55F26FED3EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_jump_around.evt"},
	{0xB8ABF154C8DBDB0CULL, "art\\characters\\heraldry\\hld_kirkwall.dds"},
	{0xB8ACE34A3AD304ECULL, "hlac_lt_0_lowtown_a2_n_c_-65535_2y.dds"},
	{0xB8AD18322392ACF6ULL, "gen_im_npc_wep_shd_tmpa.uti"},
	{0xB8AD5B536671D43AULL, "6059435_m.fxe"},
	{0xB8AD7255D1C520CAULL, "hlac_w_bonepitdestroyed_a3_262146_0y.dds"},
	{0xB8AE1B7326A42F77ULL, "lt_0_lowtown_21_0.rml"},
	{0xB8AE9F85C5417D2DULL, "llac_lt_0_lowtown_n_-3_5.dds"},
	{0xB8B014BF0FBEF48FULL, "hlac_ht_playershouse_n_855_15x.dds"},
	{0xB8B047DCE9727DA3ULL, "dae000_pc_home_ambients6208496.cl"},
	{0xB8B04B79629D91B9ULL, "865846439.ogg"},
	{0xB8B0575BBA1646CDULL, "zz_one_debug6099965.cl"},
	{0xB8B0652B5AA533BCULL, "6056843_m.fxe"},
	{0xB8B0A3FD6AFAAC19ULL, "ser191st_emeric.stg"},
	{0xB8B16452395A4B88ULL, "vxht_0_hightown_a3_c3k_0_1.vlm"},
	{0xB8B171B42AADAD99ULL, "257988556.ogg"},
	{0xB8B181511878DF9BULL, "6118522_m.fxe"},
	{0xB8B19C679427EC72ULL, "6092557_m.fxe"},
	{0xB8B1B8E367A7C9E2ULL, "6069686_f.fxe"},
	{0xB8B1CC5F27032F99ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_jump_around.gad"},
	{0xB8B24FBCC759F613ULL, "vxca_caverns1_mines_ics_0_0.vlm"},
	{0xB8B28460BB7357CFULL, "hlac_lt_hangedman_-1_23y.dds"},
	{0xB8B291405DE88108ULL, "ave331_jeven6138745.cl"},
	{0xB8B2E942AEDB4C46ULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.mmh"},
	{0xB8B3220C6077878EULL, "zz_fen_debug6166021.cl"},
	{0xB8B350406F9A0D8EULL, "6087115_m.fxe"},
	{0xB8B3C106934CAEDBULL, "lt_0_lowtown_docks_n_127_0.msh"},
	{0xB8B3D306934CCD26ULL, "lt_0_lowtown_docks_n_127_0.mao"},
	{0xB8B3DF06934CE1C5ULL, "lt_0_lowtown_docks_n_127_0.mmh"},
	{0xB8B43B502DA25934ULL, "llac_lt_0_lowtown_a2_-131072_0.dds"},
	{0xB8B474A91BF1A77DULL, "hlac_lt_0_lowtown_a3_n_c_-65537_0y.dds"},
	{0xB8B4C34C3595927AULL, "372621276.ogg"},
	{0xB8B4D45FB3F17EB1ULL, "vxht_mansion2_1bw_0_0.vlm"},
	{0xB8B561522177E2CAULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_en.gad"},
	{0xB8B5C04A501B0C9CULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_inquisitionsoldier2_.ani"},
	{0xB8B602DB83D827C0ULL, "256107470.ogg"},
	{0xB8B695BA0244E06AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_killed.evt"},
	{0xB8B6E59DA097D13FULL, "1061860120.ogg"},
	{0xB8B77A7EEAF09B9AULL, "vxlt_undercityencounter_11d_0_0.vlm"},
	{0xB8B7ABBB377C9E1FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_lava_e.dds"},
	{0xB8B7B1CDB2EB6FCCULL, "271463085.ogg"},
	{0xB8B7B526CC590DB0ULL, "vxca_caverns1_mines_il9_0_0.vlm"},
	{0xB8B7B7AE778C1808ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha08a.phy"},
	{0xB8B7E7B6DAEE46CBULL, "vxw_sundermount_1z9_0_0.vlm"},
	{0xB8B8A84DA5754EB1ULL, "6088412_m.fxe"},
	{0xB8B8B1DADF9136E0ULL, "art\\effects\\cloudshadows\\w_encounter1\\cloud2.dds"},
	{0xB8B8DD5F27096BE7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_jump_around.ani"},
	{0xB8B95E6B340ECD2FULL, "6119608_m.fxe"},
	{0xB8B9A339224F0BB2ULL, "26262176.ogg"},
	{0xB8B9AF0AAAD724A0ULL, "1054881746.ogg"},
	{0xB8BA38E96BDC08B8ULL, "vxdr_primevaltomb_3vu_0_0.vlm"},
	{0xB8BA3F046EF128F6ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm02.msh"},
	{0xB8BA49046EF139F4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm02.mmh"},
	{0xB8BA53D9B0F3D038ULL, "hlac_lt_hangedman_n_-1_0z.dds"},
	{0xB8BB111B9C321E55ULL, "mag100ar_eau_road.are"},
	{0xB8BB20B06BAB1D38ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\standard_18.mao"},
	{0xB8BB89956E9C3F89ULL, "32263001.ogg"},
	{0xB8BB95A5B96CB8D4ULL, "llac_lt_darktown_-1_82.dds"},
	{0xB8BC14E4C03DCD2FULL, "combat_deathblow_06.gda"},
	{0xB8BC1BE566E25DDEULL, "271980410.ogg"},
	{0xB8BC4452217DD061ULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_en.evt"},
	{0xB8BC5C2E66FFCBF3ULL, "6106561_m.fxe"},
	{0xB8BC601AA9A0EC7AULL, "6109406_m.fxe"},
	{0xB8BCCD4A502141DEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_inquisitionsoldier2_.gad"},
	{0xB8BD312E0DA375EEULL, "vxlt_darktown_50a_0_0.vlm"},
	{0xB8BD5D7C2F4694E0ULL, "w_blightlands_33_0.rml"},
	{0xB8BD651A6090F093ULL, "vxht_mansion2b_35b_0_0.vlm"},
	{0xB8BD74BA024AC6DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_killed.gad"},
	{0xB8BD75E81F0ABC63ULL, "hlac_lt_0_lowtown_docks_65535_1x.dds"},
	{0xB8BD8D63B5B4F0CEULL, "zz_danielletest.cut"},
	{0xB8BDADC7FF3C5ACBULL, "354073508.ogg"},
	{0xB8BF4E202639ABD9ULL, "static_static_p86.psh.fxo.dx11"},
	{0xB8BF6185CD3818A7ULL, "26397493.ogg"},
	{0xB8BF72B67DA8C891ULL, "gen_im_cns_flu_merril_01.uti"},
	{0xB8BF768E8F90E146ULL, "6163297_m.fxe"},
	{0xB8BFE7B6A01CE619ULL, "zz_otl_debug6060582.cl"},
	{0xB8BFFEA43CC25611ULL, "539855411.ogg"},
	{0xB8C09144CC9F2100ULL, "vxlt_foundry_n_yk_0_0.vlm"},
	{0xB8C0ACD51DC472B1ULL, "890331344.ogg"},
	{0xB8C0E27BC20A92A2ULL, "334586319.ogg"},
	{0xB8C11422A5106C49ULL, "436947523.ogg"},
	{0xB8C11F624EC2DAC8ULL, "75217264.ogg"},
	{0xB8C14874C4FCC289ULL, "gen00fl_aveline6139382.cl"},
	{0xB8C240B36C7EDAF4ULL, "vxlt_hangedman_n_5m_0_0.vlm"},
	{0xB8C2A1C251D83AA9ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.cess.ani"},
	{0xB8C2BE92EEECBD7FULL, "567247343.ogg"},
	{0xB8C2D9F2E7F62942ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizea.msh"},
	{0xB8C2DBF2E7F62CF8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizea.mmh"},
	{0xB8C2E7F2E7F6409BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizea.mao"},
	{0xB8C35579F07B1F7BULL, "mag100_karras6034515.cl"},
	{0xB8C3B2C021A1893FULL, "6101443_m.fxe"},
	{0xB8C3C68FC18279E9ULL, "hlac_lt_0_lowtown_docks_-65538_3z.dds"},
	{0xB8C40047A4985251ULL, "260188420.ogg"},
	{0xB8C41DA9E64A3832ULL, "darktown_male_fereldan_sta6175147.cl"},
	{0xB8C429EC6405987BULL, "dae000cr_grand_cleric6148083.cl"},
	{0xB8C44018B1EAA57DULL, "686796807.ogg"},
	{0xB8C448B4C1E62145ULL, "6121313_m.fxe"},
	{0xB8C4A1CD0EDE7F3BULL, "709715507.ogg"},
	{0xB8C4A64E52A1970AULL, "dae000_bartrand6027541.cl"},
	{0xB8C5321B9C3AAABBULL, "mag100ar_eau_road.ncs"},
	{0xB8C6074875C86A32ULL, "ser221_gamlen6136095.cl"},
	{0xB8C6120420361476ULL, "gen00fl_merrill6128779.cl"},
	{0xB8C6182D5AF0B2A5ULL, "6080603_m.fxe"},
	{0xB8C649FAA8C89C77ULL, "6172377_m.fxe"},
	{0xB8C67A2C9B984707ULL, "lgt000_chanter6147190.cl"},
	{0xB8C71A2A976B6731ULL, "6083512_m.fxe"},
	{0xB8C76667041458DBULL, "6142846_m.fxe"},
	{0xB8C77F6F9CE70A3BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basic.phy"},
	{0xB8C843F45206A7FCULL, "vis121_saemus6118465.cl"},
	{0xB8C88F62AFCB6D69ULL, "151344915.ogg"},
	{0xB8C8BF5EDFE3AB31ULL, "455347146.ogg"},
	{0xB8C8C6ACDAB89844ULL, "qun110_petrice6018567.cl"},
	{0xB8C8D5C161B2823AULL, "6068501_m.fxe"},
	{0xB8C8E82A7CD80096ULL, "302332025.ogg"},
	{0xB8C96047FE2CA1CCULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_ltr.anb"},
	{0xB8C994BF05145037ULL, "ave310_party_comment6134851.cl"},
	{0xB8C9BCC251DE872BULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.cess.gad"},
	{0xB8CA3C522189FFB0ULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_en.ani"},
	{0xB8CABB45B50A7A0DULL, "vxht_keep_6f0_0_0.vlm"},
	{0xB8CAC132EF6A9169ULL, "gen00fl_merrill6124419.cl"},
	{0xB8CACFCC77239ADAULL, "6036524_m.fxe"},
	{0xB8CB21BCF8B97B86ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkf.evt"},
	{0xB8CB61BEFD0FFC53ULL, "389225671.ogg"},
	{0xB8CBC92CC16ABB37ULL, "gen00fl_anders6103665.cl"},
	{0xB8CC121B9C409362ULL, "mag100ar_eau_road.lst"},
	{0xB8CC3EDFB7D31AD0ULL, "llac_ca_caverns1_24423_13.dds"},
	{0xB8CD88BE244E64A0ULL, "19076467.ogg"},
	{0xB8CE3689C4A1D5CBULL, "482745722.ogg"},
	{0xB8CE6649DC85665EULL, "118282040.ogg"},
	{0xB8CED6791196DA04ULL, "6097233_m.fxe"},
	{0xB8CF6F4BB8B2FE95ULL, "qun131_taskmaster6019255.cl"},
	{0xB8CF7F374AD127F7ULL, "246916612.ogg"},
	{0xB8CFB9C251E2ED90ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.cess.evt"},
	{0xB8D013AEB988EEFFULL, "1069651135.ogg"},
	{0xB8D05C09EB83F979ULL, "vxdr_dungeon_1lw_0_1.vlm"},
	{0xB8D06327D63D86E8ULL, "541468514.ogg"},
	{0xB8D0932C228369D3ULL, "64813379.ogg"},
	{0xB8D16052E1410C89ULL, "6056828_m.fxe"},
	{0xB8D238BCF8BFC1D1ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkf.gad"},
	{0xB8D25400EEB75297ULL, "vxw_woundedcoast_q5_0_0.vlm"},
	{0xB8D2D3B4B0AD60CAULL, "6081728_m.fxe"},
	{0xB8D32C346CBB10D2ULL, "267840074.ogg"},
	{0xB8D34D7747594053ULL, "6150676_m.fxe"},
	{0xB8D3D1C20AE7EA92ULL, "dae300_alistair_king6134937.cl"},
	{0xB8D4206317CC1C90ULL, "889502835.ogg"},
	{0xB8D4A32CDE74A8BAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rubblelarge.mmh"},
	{0xB8D4A72CDE74AFF1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rubblelarge.mao"},
	{0xB8D4B52CDE74C700ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rubblelarge.msh"},
	{0xB8D4FBCA2656BFACULL, "489505711.ogg"},
	{0xB8D511FC051B8EA3ULL, "hlac_ht_0_hightown_a2_-1_2z.dds"},
	{0xB8D53716E24FF71EULL, "6136897_m.fxe"},
	{0xB8D55E42207367C1ULL, "6147659_f.fxe"},
	{0xB8D5C1788519526BULL, "ave121_brennan6105833.cl"},
	{0xB8D6097FF2F96A64ULL, "hlac_lt_0_lowtown_docks_a3_-131075_0y.dds"},
	{0xB8D6315B90FC1E3FULL, "s_ballista_06.dds"},
	{0xB8D6422517FAFE2FULL, "6113008_m.fxe"},
	{0xB8D6492D04B83C6EULL, "hla_ht_0_hightown_n_-65535_0x.dds"},
	{0xB8D6DA1EC1ADCC66ULL, "plt_ave000pt_main.nss"},
	{0xB8D711DA899F966DULL, "642656053.ogg"},
	{0xB8D766E11FCB8CE4ULL, "hlac_lt_undercityencounter_-1_7y.dds"},
	{0xB8D77FC1D9870BC8ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsm.msh"},
	{0xB8D791C1D9872A59ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsm.mao"},
	{0xB8D79DC1D9873EF2ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushgrpsm.mmh"},
	{0xB8D7E605676CB968ULL, "vxlt_gallowstemplar_f_4sh_0_0.vlm"},
	{0xB8D7E9BE324752C7ULL, "318362837.ogg"},
	{0xB8D831A653D36972ULL, "6133375_m.fxe"},
	{0xB8D839BCF8C42F7FULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_walkf.ani"},
	{0xB8D86617D9F13CC6ULL, "vxht_brothel_n_52l_0_0.vlm"},
	{0xB8D89367CB1BB7CAULL, "706135787.ogg"},
	{0xB8D8B6DB6DF7EF44ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_1_ian.gad"},
	{0xB8D91AD631EA82E7ULL, "vxlt_0_lowtown_a2_n_e22_0_0.vlm"},
	{0xB8D95299971EBEEDULL, "hlac_w_bonepitmining_a2_458754_0z.dds"},
	{0xB8D9A0AB459C17BEULL, "6094959_m.fxe"},
	{0xB8D9BC19FD853256ULL, "qcr220_meredith6060311.cl"},
	{0xB8DB06E3B46ED556ULL, "vxw_woundedcoast_54_0_0.vlm"},
	{0xB8DBEB857C852FDFULL, "vxlt_backalley_g_1wc_0_0.vlm"},
	{0xB8DC4D6E86E21C39ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twra_0d.dds"},
	{0xB8DDADC5ECF4D2B6ULL, "383996334.ogg"},
	{0xB8DE18AD9B9F7C70ULL, "vxlt_shop_io_0_0.vlm"},
	{0xB8DE97C0C7E58C14ULL, "vxht_0_hightown_a2_bze_0_0.vlm"},
	{0xB8DEA70AE32043CDULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrla_0s.dds"},
	{0xB8DEC5DB6DFC7422ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_1_ian.ani"},
	{0xB8DF9CAC4D3461C8ULL, "6169095_m.fxe"},
	{0xB8E0A62FA01D15BCULL, "6043280_f.fxe"},
	{0xB8E0AA6F92A01F96ULL, "vxlt_0_lowtown_a2_n_c_db4_0_0.vlm"},
	{0xB8E104F4E175CB6BULL, "vxdr_0_deeproads_d38_0_0.vlm"},
	{0xB8E15B038551008DULL, "dre000_hireling16038618.cl"},
	{0xB8E18112B03FDA9EULL, "vxht_mansion2b_27l_0_0.vlm"},
	{0xB8E1D772D551E893ULL, "6129349_f.fxe"},
	{0xB8E1E142714E8A20ULL, "836609477.ogg"},
	{0xB8E20148464BC057ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisa.msh"},
	{0xB8E21348464BDEBAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisa.mao"},
	{0xB8E21748464BE5C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisa.mmh"},
	{0xB8E335CAEECC44F4ULL, "jobboards_one.gda"},
	{0xB8E3B6D879DB8716ULL, "follower_barks6153168.cl"},
	{0xB8E3FB49C75DD68CULL, "vxlt_backalley_n_1u3_0_0.vlm"},
	{0xB8E4175AAD5E47B4ULL, "314932370.ogg"},
	{0xB8E492EAB20D282FULL, "gen_im_trs_gen_junk1_03.uti"},
	{0xB8E6121DF6864CC1ULL, "zz_lgt_1_debug6114382.cl"},
	{0xB8E637D7E1E90781ULL, "710604047.ogg"},
	{0xB8E6509D521E81EFULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_eye_0d.dds"},
	{0xB8E65BBDD58FDD15ULL, "ico_helm_medium_2.dds"},
	{0xB8E66201023AF5B3ULL, "6089516_m.fxe"},
	{0xB8E6E34E7422A61AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_small_2.mao"},
	{0xB8E7897852920F70ULL, "vxht_mansion2_236_0_0.vlm"},
	{0xB8E85F914AD63441ULL, "vxht_0_hightown_a2_cdw_0_0.vlm"},
	{0xB8E8B64E79AF484CULL, "zz_mrl_debug6086552.cl"},
	{0xB8E8B7D1E7A5F413ULL, "6165229_m.fxe"},
	{0xB8E978C0A664EDD6ULL, "vxw_woundedcoast_25_0_0.vlm"},
	{0xB8EAB5894715D09AULL, "and100_anders6095813.cl"},
	{0xB8EAB6B696E23E8FULL, "lgt311cr_shw_mer_intro.utc"},
	{0xB8EAE0B8B950597DULL, "6177186_f.fxe"},
	{0xB8EB21D232A0BF38ULL, "429029020.ogg"},
	{0xB8EB34136707D5C4ULL, "36278310.ogg"},
	{0xB8EBCF90E5DC3D45ULL, "art\\vfx\\cinematic\\vfx_blood_pool_grow_p.anb"},
	{0xB8EBD4DFF7975AF6ULL, "6110270_f.fxe"},
	{0xB8EBEF1EC1C04BA1ULL, "plt_ave000pt_main.plo"},
	{0xB8ECE5CF7CCDD2E9ULL, "6127188_m.fxe"},
	{0xB8ED642F21EA2906ULL, "ser221_party_comments6209844.cl"},
	{0xB8EDB584287EE0E2ULL, "w_encounter2_11_0.msh"},
	{0xB8EDB784287EE398ULL, "w_encounter2_11_0.mmh"},
	{0xB8EDC384287EF83BULL, "w_encounter2_11_0.mao"},
	{0xB8EECA257B99EF1FULL, "505246043.ogg"},
	{0xB8EF7C989420F232ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bf04a.phy"},
	{0xB8EFF2EFA235F31EULL, "vxht_mansion2b_20m_0_0.vlm"},
	{0xB8F0436FAAEBFDCFULL, "156074440.ogg"},
	{0xB8F067E8376CD74AULL, "6153885_m.fxe"},
	{0xB8F0B784288119ABULL, "w_encounter2_11_0.rml"},
	{0xB8F0ECF142852CAAULL, "vxht_brothel_n_5mi_0_0.vlm"},
	{0xB8F0F362C8236494ULL, "w_encounter1_31_0.rml"},
	{0xB8F1011FF27A0678ULL, "6182459_f.fxe"},
	{0xB8F1264A88F2C625ULL, "hlac_ht_0_hightown_65533_2x.dds"},
	{0xB8F16CCE93041058ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_5_0.pwk"},
	{0xB8F16FEC4A5C90FBULL, "gen00fl_merrill6119563.cl"},
	{0xB8F264EBAB232191ULL, "vxlt_gallowstemplar_n_4o1_0_0.vlm"},
	{0xB8F2A96A9B4D4345ULL, "gen_im_trs_dal_shd_wdn_r0.uti"},
	{0xB8F2D6DCF387412BULL, "vxdr_primevaltomb_6m2_0_0.vlm"},
	{0xB8F3666F9D0C1ADAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basic.msh"},
	{0xB8F3746F9D0C3213ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basic.mao"},
	{0xB8F3786F9D0C3920ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basic.mmh"},
	{0xB8F39748C614882AULL, "344319482.ogg"},
	{0xB8F3BEC73D3E606AULL, "2922739.ogg"},
	{0xB8F41C3B8D1EBC04ULL, "6075466_m.fxe"},
	{0xB8F46537214E652DULL, "6149100_m.fxe"},
	{0xB8F498F4099A5AE4ULL, "932120502.ogg"},
	{0xB8F4ADD00B9D1E28ULL, "mag321_comtesse_de_launcet6064941.cl"},
	{0xB8F50D7A52198C89ULL, "wm_ico_lowtown.dds"},
	{0xB8F5425B9116EBD1ULL, "s_ballista_06.mao"},
	{0xB8F589B208CA4147ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.dura.gad"},
	{0xB8F5B7CB7010113CULL, "hlac_lt_0_lowtown_n_65535_6x.dds"},
	{0xB8F5CF1C14104B5FULL, "hlac_lt_0_lowtown_a2_n_-196608_2x.dds"},
	{0xB8F668957BF3ABD4ULL, "vxht_chantry_350_0_1.vlm"},
	{0xB8F70394D596EBAFULL, "6123340_m.fxe"},
	{0xB8F74EAF20BB0A1EULL, "802530754.ogg"},
	{0xB8F7630E6D53929EULL, "1017548570.ogg"},
	{0xB8F7A1DFEEE1965FULL, "vxw_bonepitmining_a2_1j8_0_0.vlm"},
	{0xB8F7A382A0552EAFULL, "vxlt_0_lowtown_a2_n_gc7_0_0.vlm"},
	{0xB8F7BE8428874423ULL, "w_encounter2_11_0.phy"},
	{0xB8F7E673745FF42BULL, "llac_lt_undercitydungeon_-1_24.dds"},
	{0xB8F8056E7C759768ULL, "6187940_f.fxe"},
	{0xB8F8CF3F607160DFULL, "vxlt_backalley_g_1ym_0_0.vlm"},
	{0xB8F90C3E283D70FEULL, "lt_0_lowtown_docks_a3_32_0_lc.mmh"},
	{0xB8F9C22CDE93F9C5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2rubblelarge.phy"},
	{0xB8FA24E7437009F5ULL, "follower_banter6136376.cl"},
	{0xB8FADAA3ACB49A8FULL, "follower_banter6161316.cl"},
	{0xB8FAEECA2AB071FDULL, "fen311cr_patron_1.utc"},
	{0xB8FB057C4EC2792EULL, "vxlt_0_lowtown_n_e1j_0_0.vlm"},
	{0xB8FB327963FD525FULL, "286106455.ogg"},
	{0xB8FBF25FCC95EED8ULL, "6017610_jansen_hiding.cut"},
	{0xB8FC7EB208D04D54ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.dura.evt"},
	{0xB8FCC9F8EB70A580ULL, "lgt260_eau_tmp_roderick6143698.cl"},
	{0xB8FD45AD64A82C6BULL, "fen110_anso6094997.cl"},
	{0xB8FD4AFC941D8B67ULL, "vxca_caverns1_mines_j5r_0_0.vlm"},
	{0xB8FDA6137CD315ACULL, "hlac_w_encounter2_camp_-65534_0z.dds"},
	{0xB8FDCFC03038C37AULL, "vxlt_darktown_4vz_0_0.vlm"},
	{0xB8FDD28D574E67AFULL, "llac_lt_0_lowtown_a2_n_c_-2_6.dds"},
	{0xB8FDE1F625A30FE0ULL, "vxlt_0_lowtown_a3_n_c_fgj_0_0.vlm"},
	{0xB8FE7DAB41CB1BE5ULL, "gen00fl_varric6078659.cl"},
	{0xB8FE98E5AE332987ULL, "mer220_hubert6041175.cl"},
	{0xB8FF332D032B6077ULL, "vxw_woundedcoast_1n9_0_0.vlm"},
	{0xB8FF780E848C248BULL, "dae000_gamlen6005830.cl"},
	{0xB9003215AB56114AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_vent.mao"},
	{0xB9003615AB561799ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_vent.mmh"},
	{0xB9004015AB562887ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_vent.msh"},
	{0xB9004E1440B0B80DULL, "hlac_lt_gallowstemplar_n_2383_7y.dds"},
	{0xB9004F91E9CC44A0ULL, "344664245.ogg"},
	{0xB900C4E4CAD14101ULL, "hlac_w_encounter3_-393212_0y.dds"},
	{0xB90118F935C79459ULL, "vxlt_0_lowtown_docks_a3_cox_0_0.vlm"},
	{0xB90137EA195B2CB4ULL, "qun231_varnell6099586.cl"},
	{0xB90148B9E57D59ABULL, "vxdr_primevaltomb_6p5_0_0.vlm"},
	{0xB901C11C2CECCF06ULL, "532543662.ogg"},
	{0xB901CA94C8234869ULL, "637125384.ogg"},
	{0xB90270FA6B78E3FDULL, "w_encounter3_98_0.phy"},
	{0xB902EFBD18559D1EULL, "6101030_f.fxe"},
	{0xB90309BD0B16B746ULL, "vxht_0_hightown_a2_c6v_0_0.vlm"},
	{0xB9032E1DE8F248DFULL, "6179875_m.fxe"},
	{0xB90335F467F002C9ULL, "qun160_party_comment_26061245.cl"},
	{0xB90345BAE62726B3ULL, "6087716_m.fxe"},
	{0xB903ADC5CC0A06EEULL, "vxw_woundedcoast_1r8_0_0.vlm"},
	{0xB903CEDC9F09AF78ULL, "zz_followers_banter6157408.cl"},
	{0xB90457F3F33179B2ULL, "vxlt_0_lowtown_a3_n_c_gjr_0_0.vlm"},
	{0xB9058F0EED923740ULL, "474189100.ogg"},
	{0xB905F541399B110DULL, "gen00fl_merrill6121131.cl"},
	{0xB9073290F1BD15DDULL, "363226632.ogg"},
	{0xB907980F2CA8AAB0ULL, "art\\vfx\\combat\\vfx_c_rw_cyclone_c.mmh"},
	{0xB9079BCE1AB7C6F0ULL, "cd_mag101_decimus.wbd"},
	{0xB907CC836362B0F7ULL, "979537645.ogg"},
	{0xB9081FD0EB472D84ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockexxl.phy"},
	{0xB9082EDD37F804D1ULL, "401191782.ogg"},
	{0xB9087401A7F91D3FULL, "vxw_sundermount_2zs_0_1.vlm"},
	{0xB9088C6B618A1525ULL, "6041833_m.fxe"},
	{0xB908AA19B061946AULL, "gen00fl_sebastian6189012.cl"},
	{0xB908B8518D60D598ULL, "one110_meeran6089484.cl"},
	{0xB909406725D9C261ULL, "qun200_party_comment6095345.cl"},
	{0xB909446B2B34147DULL, "vxlt_0_lowtown_f0v_0_0.vlm"},
	{0xB90976B208DAC98DULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_c.dura.ani"},
	{0xB90A3C3DEF1C9585ULL, "t3_har_org.tnt"},
	{0xB90A654E87C96FD9ULL, "art\\vfx\\combat\\vfx_g_dual_flurry_act_c.mmh"},
	{0xB90AC3084655BF1CULL, "vxca_caverns1_e4l_0_0.vlm"},
	{0xB90ACD45A22C1B98ULL, "cli_army_good_guys.gda"},
	{0xB90BB9DFC7961037ULL, "hlac_lt_0_lowtown_n_65538_4z.dds"},
	{0xB90BFBB0F6DBF9A2ULL, "vxht_chantry_n_351_0_1.vlm"},
	{0xB90C247156CA167FULL, "6036518_m.fxe"},
	{0xB90C27EB170F6E1AULL, "6208191_f.fxe"},
	{0xB90C31E5ABA57498ULL, "vxht_mansion_n_6zz_0_0.vlm"},
	{0xB90C4EECA48F2839ULL, "513952923.ogg"},
	{0xB90D0D9A32AE3A4DULL, "vxw_encounter2_camp_jo_0_0.vlm"},
	{0xB90D15C4BCD4946AULL, "dae000_flemeth6000981.cl"},
	{0xB90D2FC9CDFE2678ULL, "788428988.ogg"},
	{0xB90DA509F52178F7ULL, "vxw_woundedcoast_1i8_0_0.vlm"},
	{0xB90E2152399D4B3BULL, "01_fire_f2.dds"},
	{0xB90E3361D723E489ULL, "96892733.ogg"},
	{0xB90E8C2C6952366EULL, "111780800.ogg"},
	{0xB90E8D8946CF71E7ULL, "hlac_ht_chantry_2534_11z.dds"},
	{0xB90F116E16139271ULL, "vxlt_blackemporium_nx_0_0.vlm"},
	{0xB90F2148FCA26626ULL, "vxw_blightlands_cl_0_0.vlm"},
	{0xB90F244275531C7FULL, "vxht_mansion2b_1xc_0_0.vlm"},
	{0xB90F3326A05A583BULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_l_p.impa.ani"},
	{0xB90F3EDA6FA86F2AULL, "6051804_m.fxe"},
	{0xB90FBA96D773722BULL, "vxdr_primevaltomb_6w4_0_0.vlm"},
	{0xB91014D4F5FD25F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_back.anb"},
	{0xB9102EF43A0CD29CULL, "qun110ip_compound_gate_nt.utp"},
	{0xB9105F3D87816574ULL, "and330tr_endstage_approval.ncs"},
	{0xB91139A7EC25A91BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02a.mmh"},
	{0xB9114942DB5E9828ULL, "6109003_f.fxe"},
	{0xB91157A7EC25DC05ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02a.msh"},
	{0xB9116FDD1F31FD3DULL, "6061696_f.fxe"},
	{0xB911DA33009F54E1ULL, "w_bonepitdestroyed_a3_39_0.rml"},
	{0xB911F4A5032B289FULL, "vxw_blightlands_a3_0_0.vlm"},
	{0xB9126D5DABA8C74DULL, "6002775_m.fxe"},
	{0xB912AE87175C0887ULL, "zz_hangedman.cut"},
	{0xB912BA4EE09F6150ULL, "vxw_encounter1_bo_0_0.vlm"},
	{0xB912C776B3C4364FULL, "46296547.ogg"},
	{0xB912CC5A3364C837ULL, "938029958.ogg"},
	{0xB91365FA6B874469ULL, "w_encounter3_98_0.mao"},
	{0xB91371FA6B875802ULL, "w_encounter3_98_0.mmh"},
	{0xB91373FA6B875BB8ULL, "w_encounter3_98_0.msh"},
	{0xB9144284412FA32EULL, "974715320.ogg"},
	{0xB914BB0F2CB370CFULL, "art\\vfx\\combat\\vfx_c_rw_cyclone_c.anb"},
	{0xB914CB346563425FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxb.msh"},
	{0xB914D090E5FF4386ULL, "art\\vfx\\cinematic\\vfx_blood_pool_grow_p.mmh"},
	{0xB914D0A382D9BD29ULL, "6106668_m.fxe"},
	{0xB914D13465634C01ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxb.mmh"},
	{0xB914DD34656360A2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxb.mao"},
	{0xB915018015400BF7ULL, "human_female_heads.gda"},
	{0xB91548046F3DE3EFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm02.phy"},
	{0xB915A3F8EDE0FAB1ULL, "vxht_bartrand_n_5s4_0_0.vlm"},
	{0xB91624115FC4984DULL, "ts_kirkwall_02.gda"},
	{0xB916B95423CA12FFULL, "vxlt_0_lowtown_a2_n_gvt_0_0.vlm"},
	{0xB916E2419E61B545ULL, "dre000_bodahn6031348.cl"},
	{0xB917282969C132B2ULL, "6164008_f.fxe"},
	{0xB91737DFA9F0850FULL, "vxlt_0_lowtown_a2_n_go3_0_0.vlm"},
	{0xB9175ACE7ADA87BAULL, "6146803_m.fxe"},
	{0xB91760A6697B97C2ULL, "ht_0_hightownexplosion_n_6_0.phy"},
	{0xB91826DBBEF79480ULL, "267827453.ogg"},
	{0xB9193CB201848B2FULL, "vxdr_primevalentrance_24t_0_1.vlm"},
	{0xB91A4D6332361346ULL, "77871419.ogg"},
	{0xB91A810BBEEFB9D6ULL, "1007563149.ogg"},
	{0xB91A9C2B1BBA2617ULL, "gen00fl_aveline6120834.cl"},
	{0xB91AB15EFDA455BAULL, "gen_im_arm_glv_lgt_act2_01.uti"},
	{0xB91ABC2F1A7BCD97ULL, "6018590_m.fxe"},
	{0xB91B286F6D7A856AULL, "997460948.ogg"},
	{0xB91B31A0CE331684ULL, "897236933.ogg"},
	{0xB91B7F7724A452CDULL, "vxw_encounter2_camp_en_0_0.vlm"},
	{0xB91B8F947ACD2D5FULL, "lt_0_lowtown_n_37_0_ll.mmh"},
	{0xB91B9A914B9E1F56ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\timberpostxb.msh"},
	{0xB91BA4914B9E3054ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\timberpostxb.mmh"},
	{0xB91BA8914B9E3767ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\timberpostxb.mao"},
	{0xB91C3B26A064F032ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_l_p.impa.evt"},
	{0xB91C416B0AE2EFD3ULL, "m_dae121ar_brothel_nt.bnk"},
	{0xB91C978988AB471AULL, "vxlt_undercityencounter_1lc_0_0.vlm"},
	{0xB91CB10D98CE6184ULL, "652032437.ogg"},
	{0xB91CDE9CA5B00108ULL, "lgt110_urn_guard6133112.cl"},
	{0xB91CFBBB87B4A73EULL, "262589853.ogg"},
	{0xB91D961F674934FFULL, "vxht_mansion2b_1sb_0_0.vlm"},
	{0xB91DB175F2AAB93DULL, "6072668_m.fxe"},
	{0xB91DC1F879819914ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_2z.dds"},
	{0xB91E2C73C9698AABULL, "vxdr_primevaltomb_6z7_0_0.vlm"},
	{0xB91E579A82060FC3ULL, "vxht_bartrand_n_4d2_0_0.vlm"},
	{0xB91E5F8EC866DBB8ULL, "hlac_lt_backalley_n_1_0x.dds"},
	{0xB91E6987ADEEC385ULL, "vxlt_0_lowtown_n_h8o_0_0.vlm"},
	{0xB91E91F109634DCCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforwardleft.gad"},
	{0xB91F33CAFAFCB216ULL, "hlac_lt_0_lowtown_a2_-65536_0y.dds"},
	{0xB91F41FAE97CD7BEULL, "vxlt_0_lowtown_eow_0_0.vlm"},
	{0xB91F4F49D9DFD708ULL, "gen_im_trs_chn_andraste_04.uti"},
	{0xB91FED76EF02E846ULL, "vxht_0_hightown_a2_c8x_0_0.vlm"},
	{0xB91FF59B00065158ULL, "6085986_m.fxe"},
	{0xB920133B675B6462ULL, "vxw_woundedcoast_1qy_0_0.vlm"},
	{0xB9205386FED065C4ULL, "lowtown_male_alienage2_mob.cnv"},
	{0xB9206681F521411FULL, "vxw_blightlands_j2_0_0.vlm"},
	{0xB9212C2BD29579D0ULL, "vxw_encounter1_gl_0_0.vlm"},
	{0xB921E594478E2AE0ULL, "6150262_m.fxe"},
	{0xB9227D5830ADCF0EULL, "vxlt_gallowstemplar_f_4yl_0_0.vlm"},
	{0xB92294B181B8BC5FULL, "vxlt_undercityserialkiller_22c_0_0.vlm"},
	{0xB9235A26A06B43C5ULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_l_p.impa.gad"},
	{0xB923A4A2393D3484ULL, "6163156_m.fxe"},
	{0xB923A6A26B870F9BULL, "b_gr_fire_05.mao"},
	{0xB923FD91EB9484F3ULL, "6049480_m.fxe"},
	{0xB9247586FED48824ULL, "lowtown_male_alienage2_mob.bnk"},
	{0xB92482169F45F0D9ULL, "vxlt_gamlenshouse_n_nc_0_0.vlm"},
	{0xB924A8A16A91BA78ULL, "953986787.ogg"},
	{0xB924CC394DA1C85EULL, "vxw_encounter2_bx_0_0.vlm"},
	{0xB9255B042C08E04CULL, "zz_one_debug6073980.cl"},
	{0xB92561295C82F556ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion1.ani"},
	{0xB925640FF5DDC2B0ULL, "30338809.ogg"},
	{0xB92590F109696ACAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkforwardleft.ani"},
	{0xB9275268B564E4DCULL, "vxlt_foundry_n_y7_0_0.vlm"},
	{0xB92802271AB33333ULL, "art\\characters\\playercharacter\\humanmale\\cth_comc.phy"},
	{0xB928027AB1834871ULL, "vxw_sundermount_346_0_1.vlm"},
	{0xB9283E3D8795BEA1ULL, "and330tr_endstage_approval.utt"},
	{0xB928A92F2558A86FULL, "hlac_lt_undercitydungeon_-1_9y.dds"},
	{0xB928DE11F5E4C620ULL, "6160993_m.fxe"},
	{0xB928E0437098E176ULL, "899981633.ogg"},
	{0xB92934D1849C7185ULL, "art\\levels\\races\\proxy\\props\\_textures\\woodrope_sa.dds"},
	{0xB929B9D5245DEB37ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobb_0d.dds"},
	{0xB929E9BB1D829DB4ULL, "vxlt_gallowstemplar_4qe_0_0.vlm"},
	{0xB92AF362C854DD57ULL, "w_encounter1_31_0.mmh"},
	{0xB92AF762C854E464ULL, "w_encounter1_31_0.mao"},
	{0xB92B0962C85502C9ULL, "w_encounter1_31_0.msh"},
	{0xB92BBD9A26B6FC69ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masa_0t.dds"},
	{0xB92C07FC593F4D7FULL, "vxht_mansion2b_1vm_0_0.vlm"},
	{0xB92CA18690A84F3AULL, "zz_dae_debug6029433.cl"},
	{0xB92CB47322F354D8ULL, "llac_lt_0_lowtown_a3_n_c_1_11.dds"},
	{0xB92CDA617B8536B0ULL, "715256817.ogg"},
	{0xB92D07C637A39CDBULL, "552779248.ogg"},
	{0xB92D3B44BBFF63B0ULL, "vxlt_0_lowtown_docks_a3_n_cto_0_0.vlm"},
	{0xB92D9BBD585DD4FAULL, "gen_im_arm_cht_lgt_act1_02.uti"},
	{0xB92DB3D7DB72F03EULL, "vxlt_0_lowtown_ejh_0_0.vlm"},
	{0xB92E3F24AF3193ECULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar8_.ani"},
	{0xB92E7685A1D46A82ULL, "llac_lt_0_lowtown_docks_n_131068_0.dds"},
	{0xB92EE5C427A0EB5EULL, "630465233.ogg"},
	{0xB92F7AC72FFC10EEULL, "lgt100cr_ignacio_strand.utc"},
	{0xB92FE7EFABC16966ULL, "vxht_mansion2b_2un_0_0.vlm"},
	{0xB930900BB50DD9FBULL, "dae300_alistair_king6134871.cl"},
	{0xB93097D5B43E3889ULL, "gen00fl_fenris6107166.cl"},
	{0xB931D64C92E5655EULL, "6064717_f.fxe"},
	{0xB93269295C8D8D47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion1.evt"},
	{0xB932F3F3913C0959ULL, "vxlt_gamlenshouse_n_ab_0_0.vlm"},
	{0xB933282DE171668DULL, "6101056_m.fxe"},
	{0xB9333E163F97E0DEULL, "vxw_encounter2_my_0_0.vlm"},
	{0xB933624E87EC6FCEULL, "art\\vfx\\combat\\vfx_g_dual_flurry_act_c.anb"},
	{0xB9339D0E07B643FFULL, "vxlt_0_lowtown_a2_n_ghv_0_0.vlm"},
	{0xB933CBA22084397EULL, "6096795_m.fxe"},
	{0xB9342D7104F2B2D2ULL, "6118959_f.fxe"},
	{0xB9344C24AF36152EULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar8_.gad"},
	{0xB9345B0CE8F1DBDEULL, "zz_dae_debug6142969.cl"},
	{0xB93520E4BA159E9AULL, "vxlt_0_lowtown_docks_a3_n_br2_0_0.vlm"},
	{0xB9352534179C7F0EULL, "vxlt_gallowscourtyard_n_74r_0_0.vlm"},
	{0xB935362D4390268CULL, "6054045_m.fxe"},
	{0xB935C445A75AFD5CULL, "vxlt_foundry_n_t4_0_0.vlm"},
	{0xB935C648577530B5ULL, "579662972.ogg"},
	{0xB935EC1C6102647FULL, "genip_node_b2.utp"},
	{0xB936693A5DF75848ULL, "hlac_w_sundermount_-196614_0y.dds"},
	{0xB9372E0E6A25B55AULL, "940855081.ogg"},
	{0xB93730000899C266ULL, "vxht_bartrand_n_656_0_0.vlm"},
	{0xB93895C835FE6E74ULL, "and100_anders6114235.cl"},
	{0xB9389B0E30142D54ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_roba_0n.dds"},
	{0xB9393E129301B656ULL, "vxca_caverns1_e82_0_0.vlm"},
	{0xB9398A295C93E390ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\attackversion1.gad"},
	{0xB9399E2404A3C940ULL, "lit_aoe_02_l3.mao"},
	{0xB939AA6104C15B3FULL, "hlac_lt_0_lowtown_docks_n_-65539_5z.dds"},
	{0xB93A68DB108837CCULL, "llac_ht_keep_3736_2.dds"},
	{0xB93A6C47561FA357ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls102a_0n.dds"},
	{0xB93A76DFD284AFA6ULL, "vxw_blightlands_ro_0_0.vlm"},
	{0xB93A9EB1C220E8A4ULL, "llac_lt_0_lowtown_65534_2.dds"},
	{0xB93ADE0454666E14ULL, "6016668_f.fxe"},
	{0xB93BAD02CC90190DULL, "1014923946.ogg"},
	{0xB93C3BA6699A7576ULL, "ht_0_hightownexplosion_n_6_0.mao"},
	{0xB93C47A6699A8915ULL, "ht_0_hightownexplosion_n_6_0.mmh"},
	{0xB93C49A6699A8C8BULL, "ht_0_hightownexplosion_n_6_0.msh"},
	{0xB93C75D04D652A7CULL, "520015038.ogg"},
	{0xB93CB9A1BBD9ACCBULL, "533090553.ogg"},
	{0xB93CD2FC2328C665ULL, "573560854.ogg"},
	{0xB93D0754484ABA77ULL, "6103625_m.fxe"},
	{0xB93D38DEBD3E7637ULL, "6172540_m.fxe"},
	{0xB93D6B4B10EB6054ULL, "106333026.ogg"},
	{0xB93DDDDC21BB8F57ULL, "6103078_m.fxe"},
	{0xB93ECCE8DFE3B91AULL, "1020372412.ogg"},
	{0xB93EE6D7259B81C0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edge_2.mao"},
	{0xB93EFDC1D5025361ULL, "hlac_lt_0_lowtown_a2_n_-2_5y.dds"},
	{0xB93F989CEBE61950ULL, "6170783_m.fxe"},
	{0xB9400963EA0AABE7ULL, "zz_followers_banter6158109.cl"},
	{0xB9403F1EF1078187ULL, "hlac_lt_0_lowtown_-65536_6x.dds"},
	{0xB940444AECD8FEA0ULL, "mag320_orsino6037931.cl"},
	{0xB940B70EE3BFA014ULL, "vxlt_0_lowtown_a2_hen_0_0.vlm"},
	{0xB940C30030AC87FEULL, "6151050_m.fxe"},
	{0xB940D2346588897EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_3wayxb.phy"},
	{0xB940D27A08C599E7ULL, "zz_mag_debug6054585.cl"},
	{0xB9419A9B10AF9541ULL, "sky.sub.xml"},
	{0xB941EE568C643A5CULL, "dre000_sibling_has_blight6037039.cl"},
	{0xB9421040CA137C22ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_02a_0n.dds"},
	{0xB942EB787FA31D9AULL, "vxw_bonepitdestroyed_a3_kw_0_0.vlm"},
	{0xB94308A2BB1A215BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\stun.evt"},
	{0xB943208332C77FA7ULL, "789927661.ogg"},
	{0xB9433CFA7E3DCB5DULL, "hf_delilah.mor"},
	{0xB943D0FAFE80BA8AULL, "dae020cr_merch_armor.utc"},
	{0xB9445E9E4E64C176ULL, "dae021_male_worshipper_sta6141156.cl"},
	{0xB944CD72D6ABFE91ULL, "qun200_party_comment6095406.cl"},
	{0xB945052CFD491D55ULL, "377680024.ogg"},
	{0xB9455D524F82BF9CULL, "gen00fl_fenris6108021.cl"},
	{0xB9457EC2ADD53B57ULL, "zz_followers_banter6159021.cl"},
	{0xB945F959737DD698ULL, "vxht_mansion_n_6nv_0_0.vlm"},
	{0xB9463320FEDA5625ULL, "6172537_m.fxe"},
	{0xB9469E3A0CD0C0ADULL, "vxlt_backalley_n_1dy_0_0.vlm"},
	{0xB946ED1EE0ED2D0DULL, "969804762.ogg"},
	{0xB947AC568B1C77BCULL, "612918250.ogg"},
	{0xB947DCC2A0EA888EULL, "zz_and_debug.cnv"},
	{0xB9481DCAE97A471BULL, "vxht_0_hightown_a3_c1g_0_0.vlm"},
	{0xB94824447D4D0B6BULL, "gen_im_wep_war_shd_act1_02.uti"},
	{0xB9483217242E3C41ULL, "vxlt_0_lowtown_docks_a3_cta_0_0.vlm"},
	{0xB94842FCEABCB3A2ULL, "genip_wine_bottle.utp"},
	{0xB9485A0C54484DB8ULL, "9394553.ogg"},
	{0xB94881228E9524E9ULL, "vxht_0_hightown_a3_c3v_0_1.vlm"},
	{0xB948C4EE6B5B7DE5ULL, "793096135.ogg"},
	{0xB94953673E27B235ULL, "gen_im_trs_slv_junk1_04.uti"},
	{0xB949BC785CDEE601ULL, "vxlt_foundry_n_yf_0_0.vlm"},
	{0xB949EF6E55542D3DULL, "38956954.ogg"},
	{0xB949FDDCC33222E4ULL, "framebuffereffects_motionblurcombine_p0.psh.fxo.dx11"},
	{0xB94A01A2BB20339CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\stun.gad"},
	{0xB94A1EFE9FEB94B0ULL, "vxlt_0_lowtown_docks_a3_n_czm_0_0.vlm"},
	{0xB94A9777C0FEDF4EULL, "635348636.ogg"},
	{0xB94ACC6BEB76FF18ULL, "gen00fl_anders6119173.cl"},
	{0xB94AD739DCD0DA04ULL, "6172495_m.fxe"},
	{0xB94B4AB852A67BF3ULL, "vxca_caverns1_igr_0_0.vlm"},
	{0xB94BAABBAF348EB7ULL, "6175440_m.fxe"},
	{0xB94CC994ECEBF0C2ULL, "155721539.ogg"},
	{0xB94CCBA98FAD9A66ULL, "vxht_mansion2b_2kp_0_0.vlm"},
	{0xB94CCC83333B0769ULL, "6033645_m.fxe"},
	{0xB94CE915E1F44C9CULL, "929095219.ogg"},
	{0xB94D637D963E1191ULL, "art\\vfx\\combat\\vfx_f_mer_emove_disapp_p.impa.ani"},
	{0xB94DA21E5D6F67C3ULL, "vxca_caverns1_gkm_0_1.vlm"},
	{0xB94E49CC9D8F767CULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_qunari1_arishokresponse.ani"},
	{0xB94E4BD92309819DULL, "vxht_0_hightown_a2_a7h_0_0.vlm"},
	{0xB94EE0E0D8397012ULL, "llac_ht_0_hightown_a3_-196610_0.dds"},
	{0xB94F28EBD5B5B894ULL, "vxlt_0_lowtown_a2_hbo_0_0.vlm"},
	{0xB94F573583809147ULL, "479367496.ogg"},
	{0xB94F71CE1AF534B2ULL, "cd_mag101_decimus.bnk"},
	{0xB94FD33860BAA887ULL, "ht_0_hightown_18_0_ll.mmh"},
	{0xB94FD4E16D510D83ULL, "27614599.ogg"},
	{0xB95080C7EBA274FFULL, "vxlt_0_lowtown_a2_n_gbp_0_0.vlm"},
	{0xB95100A2BB26509AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\stun.ani"},
	{0xB9515F299985FBD9ULL, "140580270.ogg"},
	{0xB95175561F63390AULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_0y.dds"},
	{0xB9517942E70AAA4BULL, "6101339_m.fxe"},
	{0xB951E1127A4F47DBULL, "6013854_m.fxe"},
	{0xB951EC94D2D403D9ULL, "vxdr_primevaltomb_2gr_0_0.vlm"},
	{0xB9521312F8FA65E0ULL, "vxlt_0_lowtown_har_0_0.vlm"},
	{0xB952CC3300D6D3D0ULL, "w_bonepitdestroyed_a3_39_0.msh"},
	{0xB952DA3300D6EBEAULL, "w_bonepitdestroyed_a3_39_0.mmh"},
	{0xB952DE3300D6F221ULL, "w_bonepitdestroyed_a3_39_0.mao"},
	{0xB953A3089E67EE6FULL, "hlac_lt_backalley_g_0_18x.dds"},
	{0xB953D4C9B9453A4DULL, "lgt110cr_orz_harrow_guard.utc"},
	{0xB953FF5F7410C16DULL, "lgt300cr_gang_1_docks.utc"},
	{0xB9544E7A2F77F19AULL, "vxht_0_hightown_c4e_0_1.vlm"},
	{0xB9546E7D96444333ULL, "art\\vfx\\combat\\vfx_f_mer_emove_disapp_p.impa.gad"},
	{0xB9551C49FF42DFF9ULL, "308826020.ogg"},
	{0xB95556CC9D95AABEULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_qunari1_arishokresponse.gad"},
	{0xB9556589CC9B5B0BULL, "ave000ip_shield_container.utp"},
	{0xB9556CCD82A4DC13ULL, "vxlt_0_lowtown_a2_n_eoy_0_0.vlm"},
	{0xB955D6AC4FD0B531ULL, "gen_im_npc_arm_hum_hlm_rdrb.uti"},
	{0xB95630099EC7BAD4ULL, "6120427_m.fxe"},
	{0xB95643E7F0A5825CULL, "w_woundedcoast_112_0.mao"},
	{0xB9564FE7F0A596FFULL, "w_woundedcoast_112_0.mmh"},
	{0xB95655E7F0A5A0A1ULL, "w_woundedcoast_112_0.msh"},
	{0xB95655F5914E659BULL, "6111287_f.fxe"},
	{0xB956FAE9EC068C2FULL, "29027463.ogg"},
	{0xB957C4AD871FE492ULL, "6208233_m.fxe"},
	{0xB957D4DD8EDEFF66ULL, "ave200_donnic6112178.cl"},
	{0xB9585AA7EC61E450ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed02a.phy"},
	{0xB958E3598D8E282BULL, "vxlt_0_lowtown_n_f0w_0_0.vlm"},
	{0xB959A304181A8633ULL, "vxlt_0_lowtown_a2_n_gpj_0_0.vlm"},
	{0xB959B264DEE421D9ULL, "vxw_bonepitmining_a2_1x0_0_0.vlm"},
	{0xB959BC95449C9473ULL, "vxca_caverns1_iju_0_0.vlm"},
	{0xB95A1A5DCCD4FE76ULL, "hlac_lt_darktown_-1_53z.dds"},
	{0xB95A5175609AB78EULL, "character_character_p111.psh.fxo.dx11"},
	{0xB95A8DDF75D8A3E1ULL, "dae000_bartrand6201380.cl"},
	{0xB95AD810412B1641ULL, "dae300_cullen6140584.cl"},
	{0xB95B365D0092CEA3ULL, "plt_cod_itm_vir_atishan.plo"},
	{0xB95B6B7D964A5C98ULL, "art\\vfx\\combat\\vfx_f_mer_emove_disapp_p.impa.evt"},
	{0xB95BB2EF66476D2BULL, "vxht_keep_n_30f_0_0.vlm"},
	{0xB95C0FDB09F678ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\whirlwind.ani"},
	{0xB95C3A3251B2828FULL, "25282464.ogg"},
	{0xB95D0A92AF0D70D2ULL, "200784218.ogg"},
	{0xB95D34AB6B3787A9ULL, "art\\vfx\\creatures\\fxc_gol_proj_c.impa.ani"},
	{0xB95D43841CFF1D43ULL, "lt_0_lowtown_docks_a3_214_0.mao"},
	{0xB95D47841CFF2390ULL, "lt_0_lowtown_docks_a3_214_0.mmh"},
	{0xB95D55841CFF3BAAULL, "lt_0_lowtown_docks_a3_214_0.msh"},
	{0xB95D723B87F45812ULL, "gen_im_arm_cht_nobb_t2.uti"},
	{0xB95D8CE7BE7D55DFULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyo_n.dds"},
	{0xB95EE62798E756BDULL, "character_character_p240.psh.fxo.dx11"},
	{0xB95EFABB9995B3D0ULL, "eye0.psh.fxo"},
	{0xB95F54D2C4540A0FULL, "gen00fl_aveline6115401.cl"},
	{0xB95FC6F9EDEF2ED3ULL, "vxw_bonepitmining_a2_6w_0_0.vlm"},
	{0xB95FE333A14B1ED0ULL, "6142328_m.fxe"},
	{0xB9602DB0F5AE8541ULL, "vxlt_0_lowtown_n_bz3_0_0.vlm"},
	{0xB96051067D23F777ULL, "follower_barks6164411.cl"},
	{0xB960E059AF6D5CC6ULL, "586215513.ogg"},
	{0xB960F3D7AB7C1AECULL, "6166652_m.fxe"},
	{0xB96106F478BA3F99ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_tmpa_0s.dds"},
	{0xB9611A2F621E450EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_down_to_up.evt"},
	{0xB9619B41753F8175ULL, "603739157.ogg"},
	{0xB961D26324F37B8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_front.gad"},
	{0xB9623407AA2F4832ULL, "11348825.ogg"},
	{0xB9623715ABA8E386ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_vent.phy"},
	{0xB962DAF20ED1023FULL, "856758213.ogg"},
	{0xB96309FBCE58CD41ULL, "6151039_m.fxe"},
	{0xB9635F6B0B1F2531ULL, "m_dae121ar_brothel_nt.wbd"},
	{0xB96381F3F0BCF1ADULL, "vxlt_backalley_n_1jw_0_0.vlm"},
	{0xB963D9A41B66813CULL, "6092494_m.fxe"},
	{0xB964274AA018C0A3ULL, "6134951_m.fxe"},
	{0xB9644765E4EB86C7ULL, "gen00fl_fenris6103353.cl"},
	{0xB9644FAB6B3DD42BULL, "art\\vfx\\creatures\\fxc_gol_proj_c.impa.gad"},
	{0xB96471C08CF49625ULL, "vxlt_foundry_n_zg_0_0.vlm"},
	{0xB9656E2B46B815F3ULL, "vxht_mansion2_316_0_0.vlm"},
	{0xB965983217AF005CULL, "320595149.ogg"},
	{0xB965D009A2151837ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_lean_forward.gad"},
	{0xB966419A5A54D2C4ULL, "gen00fl_aveline6132722.cl"},
	{0xB966441105044138ULL, "w_encounter3_129_0.phy"},
	{0xB9667A1CBA3DEE48ULL, "hlac_ca_cave_2194_5x.dds"},
	{0xB96693DFCFFFA284ULL, "lgt150_nat_were_supporter6110965.cl"},
	{0xB9671AD68CF345E6ULL, "llac_lt_0_lowtown_-196608_2.dds"},
	{0xB9680FC91B566C8EULL, "153379082.ogg"},
	{0xB968112F622454B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_down_to_up.gad"},
	{0xB96814E10A109EB3ULL, "vxlt_0_lowtown_a2_n_gmm_0_0.vlm"},
	{0xB9682441D0DA3A59ULL, "vxw_bonepitmining_a2_1u7_0_0.vlm"},
	{0xB9682D38EE7BC22BULL, "zz_vault_debug6177788.cl"},
	{0xB9682E723692ACF3ULL, "vxca_caverns1_imt_0_0.vlm"},
	{0xB968B56324F96921ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_front.evt"},
	{0xB96925F781D02322ULL, "hlac_ht_playershouse_n_855_28y.dds"},
	{0xB96932E988736CD5ULL, "6122925_m.fxe"},
	{0xB96947AFAF43CD52ULL, "vxw_bonepitmining_a2_6r_0_0.vlm"},
	{0xB969AF637399CB66ULL, "vxht_mansion2b_2ar_0_0.vlm"},
	{0xB969BFF63FDB27D3ULL, "hlac_lt_backalley_n_0_14y.dds"},
	{0xB969C47AF2D8FFBDULL, "591543822.ogg"},
	{0xB969DFBA574DAD82ULL, "6042206_m.fxe"},
	{0xB969F7DB0A028D22ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\whirlwind.evt"},
	{0xB96A4CAB6B423A90ULL, "art\\vfx\\creatures\\fxc_gol_proj_c.impa.evt"},
	{0xB96A67EB5B0EB59AULL, "hlac_lt_hangedman_n_-1_0x.dds"},
	{0xB96A811482C0762FULL, "mrl210_merrill6116388.cl"},
	{0xB96B28181E6828D5ULL, "242124552.ogg"},
	{0xB96B35477C8CEC78ULL, "qcr220_grey_wardens6060441.cl"},
	{0xB96BBAB2FBEE9CDBULL, "scaling_level.gda"},
	{0xB96BE509A219A7C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_lean_forward.evt"},
	{0xB96C25560C632605ULL, "pro000_colonel_ewald6005394.cl"},
	{0xB96C4C2D23A3A711ULL, "vxlt_gallowscourtyard_n_7gh_0_0.vlm"},
	{0xB96D0FCBD5D495E9ULL, "hlac_ca_caverns1_mines_24424_17y.dds"},
	{0xB96D4AA11EF1ABD5ULL, "225686534.ogg"},
	{0xB96D4E8DE9424E4EULL, "346767892.ogg"},
	{0xB96E4F9CD7F21AA9ULL, "73876512.ogg"},
	{0xB96E742B05C4BF95ULL, "6125397_m.fxe"},
	{0xB96E890415DE5368ULL, "vxht_keep_67r_0_0.vlm"},
	{0xB96EF2448611D9A9ULL, "6018655_m.fxe"},
	{0xB96F0B566DD8911DULL, "llac_w_blightlands_196606_0.dds"},
	{0xB96F222F622A9007ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\toast_drink_down_to_up.ani"},
	{0xB96F51DDA5155410ULL, "llac_lt_gallowstemplar_f_2383_1.dds"},
	{0xB96F5B57A751E846ULL, "6140352_f.fxe"},
	{0xB96FB2A8F85F1230ULL, "vxw_bonepitmining_a2_l_0_0.vlm"},
	{0xB96FBB1D65D6CD2EULL, "562280496.ogg"},
	{0xB96FBC53E03F5014ULL, "vxht_brothel_n_47h_0_0.vlm"},
	{0xB970BC0400E56EFCULL, "follower_banter6170786.cl"},
	{0xB970F6DB0A08AA35ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\whirlwind.gad"},
	{0xB971834D6AB34615ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_4craggyrock.msh"},
	{0xB971854D6AB3498BULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_4craggyrock.mmh"},
	{0xB971914D6AB35E28ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_4craggyrock.mao"},
	{0xB971AB32D78DBDB2ULL, "877373652.ogg"},
	{0xB971B3A674498EEBULL, "art\\vfx\\combat\\vfx_m_bld_control_imp_p.mmh"},
	{0xB971E73700D9913AULL, "vxlt_blackemporium_zs_0_0.vlm"},
	{0xB97203FC6EAC6E5AULL, "515917873.ogg"},
	{0xB9723F8001AC3EEBULL, "ser270_emeric6201568.cl"},
	{0xB972A12270998D1AULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb104a.msh"},
	{0xB972A305DB30D672ULL, "vrc321_spooky_scenes6215548.cl"},
	{0xB972A449B70E0805ULL, "938052187.ogg"},
	{0xB972B3227099AC60ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb104a.mmh"},
	{0xB972C23E04984BBDULL, "fen210_fenris6108730.cl"},
	{0xB973494634AC9DE3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmloop.evt"},
	{0xB97370104789A3A5ULL, "6140518_m.fxe"},
	{0xB9737F91E32E0DADULL, "dae000ip_crowd_mf3_talk11.utp"},
	{0xB973C44F926649FFULL, "art\\vfx\\materials\\decal\\vfx_projdecal.mao"},
	{0xB974054D76B50160ULL, "gen_im_trs_krk_low_junk1_01.uti"},
	{0xB97409C2A1101029ULL, "zz_and_debug.ncs"},
	{0xB974408DE9DBB2BFULL, "art\\vfx\\combat\\vfx_gib_qunc.msh"},
	{0xB974468DE9DBBD61ULL, "art\\vfx\\combat\\vfx_gib_qunc.mmh"},
	{0xB97455AB28EC6597ULL, "vxw_bonepit_104_0_0.vlm"},
	{0xB9745C934D3B50E2ULL, "6139672_m.fxe"},
	{0xB9746FABF8509BD9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basic.msh"},
	{0xB97479ABF850ACC7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basic.mmh"},
	{0xB9747DABF850B314ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3wi_basic.mao"},
	{0xB97540D813827FC7ULL, "gen00fl_aveline6114046.cl"},
	{0xB975AD632503E670ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\pushed_from_front.ani"},
	{0xB975ED9424760A76ULL, "gen00fl_aveline6139703.cl"},
	{0xB9763BA72B046225ULL, "vxlt_gallowstemplar_f_5qn_0_0.vlm"},
	{0xB9764B375B2C6408ULL, "w_bonepitdestroyed_a3_114_0.msh"},
	{0xB9765D375B2C8299ULL, "w_bonepitdestroyed_a3_114_0.mao"},
	{0xB97669375B2C9732ULL, "w_bonepitdestroyed_a3_114_0.mmh"},
	{0xB97686BDFC06B733ULL, "vxlt_0_lowtown_a2_n_gjl_0_0.vlm"},
	{0xB976961EC2D052D9ULL, "vxw_bonepitmining_a2_1r6_0_0.vlm"},
	{0xB97759877D1C536EULL, "6110893_m.fxe"},
	{0xB977C4D106B08D65ULL, "vxlt_0_lowtown_docks_n_9re_0_0.vlm"},
	{0xB978053B5A1DFF6AULL, "vxdr_0_deeproads_fjx_0_0.vlm"},
	{0xB9782D690AC74317ULL, "454696287.ogg"},
	{0xB97847A44DC8D090ULL, "6080215_m.fxe"},
	{0xB9791B14C7C66E33ULL, "2031544.ogg"},
	{0xB9795DADDDDB04C0ULL, "vxht_mansion2_2mx_0_0.vlm"},
	{0xB979A2CE8111CB67ULL, "vxlt_warehouse_ou_0_0.vlm"},
	{0xB979B6232D78A17FULL, "llac_lt_undercitydungeon_-1_51.dds"},
	{0xB979C53F7762F9E7ULL, "mag310cr_controlled_4.utc"},
	{0xB979C69B0A34E7A8ULL, "378534631.ogg"},
	{0xB979FD09A2260DBDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_lean_forward.ani"},
	{0xB97A224634B27A64ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmloop.gad"},
	{0xB97A2C6B80E8A221ULL, "hlac_lt_gallowstemplar_2394_12z.dds"},
	{0xB97A2CE84AE67F6BULL, "vxht_0_hightown_a3_c71_0_0.vlm"},
	{0xB97A42B018A7FB85ULL, "vxca_caverns2prc_adb_0_1.vlm"},
	{0xB97A9A93335FEBB3ULL, "cs_mcr371_db_harvester_warr.bnk"},
	{0xB97AAD55BB4235D5ULL, "189040386.ogg"},
	{0xB97B29298E3AC6B4ULL, "342856774.ogg"},
	{0xB97BBDB44CD3FB6EULL, "887083673.ogg"},
	{0xB97BDFC367B6899DULL, "173085939.ogg"},
	{0xB97BF283155071BFULL, "lgt000cr_hate_gen_bandits.utc"},
	{0xB97C00879CDD84FBULL, "hlac_w_encounter2_camp_-2_0x.dds"},
	{0xB97C89D202CF4D13ULL, "vxca_caverns1_mines_i9a_0_0.vlm"},
	{0xB97CAAB3D1DB5FD3ULL, "vxw_bonepitmining_a2_8y_0_0.vlm"},
	{0xB97CFAE107D46BE8ULL, "vxht_keep_60s_0_0.vlm"},
	{0xB97D08217EE40397ULL, "hlac_lt_0_lowtown_a2_n_0_1x.dds"},
	{0xB97E7FCE709C1436ULL, "450015577.ogg"},
	{0xB97E94A67453E450ULL, "art\\vfx\\combat\\vfx_m_bld_control_imp_p.anb"},
	{0xB97EEFB72312CAA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\facial anims\\combat_face_additive.ani"},
	{0xB97F702D5561F2E3ULL, "vis121_saemus6118411.cl"},
	{0xB97FD6955196AD88ULL, "hlac_lt_0_lowtown_a2_n_131071_0y.dds"},
	{0xB97FEFCE85C094BFULL, "6151395_f.fxe"},
	{0xB9801271F80C57B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_left_1.ani"},
	{0xB980ADA1116552AAULL, "vxht_0_hightown_a2_c6z_0_0.vlm"},
	{0xB981314634B8B242ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\overwhelmloop.ani"},
	{0xB981E7510547F385ULL, "vxht_0_hightown_9p9_0_0.vlm"},
	{0xB98247633A2496D7ULL, "6199284_m.fxe"},
	{0xB9825823825D1F54ULL, "follower_banter6161186.cl"},
	{0xB9825C95737834D8ULL, "llac_ht_0_hightown_-196608_2.dds"},
	{0xB98284E615265318ULL, "art\\levels\\races\\proxy\\props\\_textures\\cheese_e.dds"},
	{0xB982C96524EBBCA6ULL, "art\\vfx\\cinematic\\vfx_ground_pound_p.anb"},
	{0xB983498D87B14003ULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy_crnf.evt"},
	{0xB98394D9D997BD1AULL, "704069408.ogg"},
	{0xB983A0E83772D88BULL, "968697352.ogg"},
	{0xB9844A6D4F8286F4ULL, "21029130.ogg"},
	{0xB984963C3DCBF008ULL, "6155997_m.fxe"},
	{0xB985734008EC7C02ULL, "hlac_ca_caverns2_5802_62y.dds"},
	{0xB985D597D683D0AEULL, "6019105_m.fxe"},
	{0xB9860B08D99D5BA0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolemb_movement.anb"},
	{0xB98617DC94E2440AULL, "gen00fl_merrill6144979.cl"},
	{0xB9865FF42C7A1DB7ULL, "vxht_0_hightown_bdt_0_0.vlm"},
	{0xB986DD3FD1B78020ULL, "233911856.ogg"},
	{0xB98739C9A8E2D79FULL, "hlac_lt_0_lowtown_0_2z.dds"},
	{0xB987E5B98C2B3311ULL, "gen00fl_anders6125847.cl"},
	{0xB987FBE1C3A6A197ULL, "dae000ip_crowd_mf1_talk15.utp"},
	{0xB988BC8C8CC6F4DAULL, "hlac_lt_gallowstemplar_n_7198_4x.dds"},
	{0xB988DBC72317B0DFULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isab.mmh"},
	{0xB988E1C72317BA81ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isab.msh"},
	{0xB9891FAB66AB89B6ULL, "640262420.ogg"},
	{0xB9898A4BFF17B3DEULL, "mrl310_party_comments6142846.cl"},
	{0xB989933B6522C346ULL, "hlac_lt_0_lowtown_docks_a3_65534_5y.dds"},
	{0xB989AE5FD4B3ACFFULL, "6064738_f.fxe"},
	{0xB989B8B618F58E72ULL, "946765373.ogg"},
	{0xB98A225AA7E0DEC9ULL, "em_harshal.mor"},
	{0xB98A228D87B71C84ULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy_crnf.gad"},
	{0xB98A22CA9D45B67CULL, "6125328_f.fxe"},
	{0xB98A25914CC5B3FFULL, "6087236_f.fxe"},
	{0xB98AC7A14E7CBE1BULL, "mag370_meredith6046913.cl"},
	{0xB98AFB271B07A4A8ULL, "art\\characters\\playercharacter\\humanmale\\cth_comc.mmh"},
	{0xB98B07271B07B94BULL, "art\\characters\\playercharacter\\humanmale\\cth_comc.mao"},
	{0xB98B19271B07D7D2ULL, "art\\characters\\playercharacter\\humanmale\\cth_comc.msh"},
	{0xB98B6F08F328A270ULL, "vxlt_shop_5s_0_0.vlm"},
	{0xB98BCC3D2D4C44A6ULL, "hlac_ht_chantry_n_2534_10y.dds"},
	{0xB98C26A3CB30C157ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medb_t.dds"},
	{0xB98C4C5902E365FDULL, "seb201st_lady_harimann.stg"},
	{0xB98E1C2981129868ULL, "411982713.ogg"},
	{0xB98E29C843D16CC3ULL, "vxlt_0_lowtown_docks_a3_d0i_0_1.vlm"},
	{0xB98E67B177C98438ULL, "hlac_lt_gallowstemplar_n_2384_6z.dds"},
	{0xB98E710A638A1D73ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_4y.dds"},
	{0xB98E71EA35D0D1F6ULL, "6127965_m.fxe"},
	{0xB98EBB9B74D6AF50ULL, "hlac_w_bonepitmining_a2_0_7y.dds"},
	{0xB98F8D8036B0493CULL, "vxlt_0_lowtown_docks_a3_d6y_0_0.vlm"},
	{0xB98FD6A98CD7077DULL, "hlac_lt_0_lowtown_a2_n_65537_5x.dds"},
	{0xB98FEF9EECF6FFEAULL, "combat_ws_r2_combo.gda"},
	{0xB99031940D198F74ULL, "hlac_ht_mansion2b_818_55z.dds"},
	{0xB990CC6524F7FF51ULL, "art\\vfx\\cinematic\\vfx_ground_pound_p.mmh"},
	{0xB990E080D308A7B9ULL, "gen00fl_isabela6142068.cl"},
	{0xB991318D87BD5562ULL, "art\\characters\\creatures\\stonegolem\\animation\\rdy_crnf.ani"},
	{0xB991344A687DD15AULL, "vxht_0_hightown_cbb_0_0.vlm"},
	{0xB9919CEEF4D417A3ULL, "vxlt_0_lowtown_a2_n_c_eqe_0_0.vlm"},
	{0xB991A028CCBB5E4CULL, "6174664_m.fxe"},
	{0xB991A446A225FF4AULL, "llac_ca_caverns2prc_5802_15.dds"},
	{0xB992A7360965DA6CULL, "vxlt_0_lowtown_a2_n_c_e3d_0_0.vlm"},
	{0xB9931E5BCEF7743CULL, "359824454.ogg"},
	{0xB993929304BB5217ULL, "6095467_m.fxe"},
	{0xB993B63D99A4AFD4ULL, "6016713_m.fxe"},
	{0xB993FF0F24779717ULL, "vxlt_backalley_g_1ft_0_0.vlm"},
	{0xB99406B723254D8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\facial anims\\combat_face_additive.gad"},
	{0xB9943B71F81D46F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_left_1.gad"},
	{0xB9945EA7D0A6EA3EULL, "zz_followers_banter6157338.cl"},
	{0xB99493A9114811B4ULL, "970426149.ogg"},
	{0xB994D1D11E703637ULL, "vxht_0_hightown_baw_0_0.vlm"},
	{0xB99593469DF9E0ECULL, "vxca_caverns2_7o0_0_0.vlm"},
	{0xB9962BCAB10F4D9AULL, "hlac_lt_0_lowtown_a2_n_c_65538_2y.dds"},
	{0xB99647C0442068DAULL, "gen00fl_aveline6139619.cl"},
	{0xB9968B6CA63A8B2DULL, "238004972.ogg"},
	{0xB996C8C7784BD0FDULL, "6106088_m.fxe"},
	{0xB996CDC15718DD65ULL, "vxht_mansion2b_75_0_0.vlm"},
	{0xB99718E791A142EBULL, "662290442.ogg"},
	{0xB997CC2105D3AD88ULL, "vxlt_darktown_57n_0_0.vlm"},
	{0xB9980A6F84487113ULL, "vis121_saemus6167901.cl"},
	{0xB9996D8BE6BB7E13ULL, "vxca_caverns1_mines_i3c_0_0.vlm"},
	{0xB999BA94DAAF2521ULL, "vxw_bonepitmining_a2_5d_0_0.vlm"},
	{0xB99A97791C0F0F18ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_ric_0s.dds"},
	{0xB99AC8D504B8FBC2ULL, "6099228_m.fxe"},
	{0xB99AE3149D50149FULL, "vxw_woundedcoast_1lj_0_1.vlm"},
	{0xB99C46E860E82C7EULL, "vxw_bonepit_259_0_0.vlm"},
	{0xB99CE12B0FA7816FULL, "mrl000_marethari6129426.cl"},
	{0xB99CEAFAA6C4CDC4ULL, "one101_orders_followers6115780.cl"},
	{0xB99D3CCDD6BBDABAULL, "vxlt_blackemporium_un_0_0.vlm"},
	{0xB99D915AF55183AAULL, "vxht_0_hightown_a2_c8t_0_0.vlm"},
	{0xB99DED0BAC1273B4ULL, "6102795_m.fxe"},
	{0xB99E1D29886E892AULL, "6140284_f.fxe"},
	{0xB99E76ACDB6998BEULL, "dae000_mother6121064.cl"},
	{0xB99F239C552CD804ULL, "vxht_mansion2_22q_0_0.vlm"},
	{0xB99F3C0D0947D87CULL, "145917379.ogg"},
	{0xB99FA17C4DD29DE1ULL, "6086285_m.fxe"},
	{0xB99FFDBFAF0AE668ULL, "vxlt_gallowstemplar_n_4uf_0_0.vlm"},
	{0xB9A05EDC4C894B28ULL, "art\\characters\\playercharacter\\elffemale\\arm_nuda.msh"},
	{0xB9A070DC4C8969B9ULL, "art\\characters\\playercharacter\\elffemale\\arm_nuda.mao"},
	{0xB9A07CDC4C897E52ULL, "art\\characters\\playercharacter\\elffemale\\arm_nuda.mmh"},
	{0xB9A0BEEA2C51BA3CULL, "vxlt_darktown_2uw_0_0.vlm"},
	{0xB9A0EBEB5C4303BFULL, "hlac_dr_0_deeproads_5845_13y.dds"},
	{0xB9A1591480A374BFULL, "vxlt_0_lowtown_docks_a3_cyt_0_0.vlm"},
	{0xB9A16B1ED5D91052ULL, "fen110_fenris6118096.cl"},
	{0xB9A18920DFC93A7CULL, "721735126.ogg"},
	{0xB9A1C9A9EC42F105ULL, "dae300_alistair_king6133519.cl"},
	{0xB9A24DF697B82FEDULL, "rdr111dg_martin6055001.cl"},
	{0xB9A2B460C1649695ULL, "hlac_ht_bartrand_n_2995_4y.dds"},
	{0xB9A2B5138C5A94EAULL, "439432469.ogg"},
	{0xB9A2F6B37D50A47BULL, "592050064.ogg"},
	{0xB9A30EF6ECE320CFULL, "6072114_m.fxe"},
	{0xB9A332D3DD5C60CAULL, "hla_ht_0_hightown_a3_-131074_0x.dds"},
	{0xB9A406FD9DE5E03CULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv1.ani"},
	{0xB9A4488719153CBEULL, "vis121_saemus6100130.cl"},
	{0xB9A4EB00D587ED96ULL, "741286062.ogg"},
	{0xB9A502E7AC9FA38FULL, "ser221_gascard6133672.cl"},
	{0xB9A58FB74C8DEBB6ULL, "ht_0_hightown_n_43_0.anb"},
	{0xB9A62603DACC0562ULL, "demoral_01.mao"},
	{0xB9A63DFDF7C9C608ULL, "vxlt_darktown_58o_0_0.vlm"},
	{0xB9A68E2617FB61FBULL, "107465794.ogg"},
	{0xB9A69F8462A57ECEULL, "art\\levels\\races\\design\\prp_deathroot.mmh"},
	{0xB9A6A407EA23BE7DULL, "gen00fl_sebastian6178350.cl"},
	{0xB9A6AB8462A59365ULL, "art\\levels\\races\\design\\prp_deathroot.mao"},
	{0xB9A6B98462A5AABCULL, "art\\levels\\races\\design\\prp_deathroot.msh"},
	{0xB9A6CAEDBAE867CFULL, "6129287_m.fxe"},
	{0xB9A77987C2D15BEDULL, "lt_ruinedhouse_n.arl"},
	{0xB9A783D323ED9184ULL, "182197167.ogg"},
	{0xB9A828989F9F17CAULL, "gen00fl_fenris6103332.cl"},
	{0xB9A82BD91B4ECB82ULL, "vxw_sundermount_2yk_0_1.vlm"},
	{0xB9A8A7F2FE57FE1BULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robc_0s.dds"},
	{0xB9A919F2D3FF19E3ULL, "vxlt_undercityencounter_1bt_0_0.vlm"},
	{0xB9A92314A81F0A7BULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_act_c.impa.ani"},
	{0xB9A93DF17A61818FULL, "hlac_lt_0_lowtown_docks_a3_65537_0z.dds"},
	{0xB9A9673329F19BC4ULL, "6039857_m.fxe"},
	{0xB9A9A6F2A5B8F3CAULL, "811931528.ogg"},
	{0xB9A9B8ACA80C6A49ULL, "lt_backalley_g_water_1875.mao"},
	{0xB9A9C4ACA80C7EE2ULL, "lt_backalley_g_water_1875.mmh"},
	{0xB9A9C5F56EFAA540ULL, "129716141.ogg"},
	{0xB9A9C6ACA80C8198ULL, "lt_backalley_g_water_1875.msh"},
	{0xB9AA60FDAB59F874ULL, "gen_im_cns_flu_varric_02.uti"},
	{0xB9AAA960146CFF97ULL, "hlac_lt_foundry_n_320_2x.dds"},
	{0xB9AABE47847960BFULL, "6076680_m.fxe"},
	{0xB9AB13FD9DEC157EULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv1.gad"},
	{0xB9AC8EAC2018687EULL, "6177581_f.fxe"},
	{0xB9AD98A542C71CD6ULL, "llac_w_woundedcoast_196609_0.dds"},
	{0xB9AD9B123FC370D7ULL, "hlac_lt_0_lowtown_docks_n_-131075_6z.dds"},
	{0xB9ADEC9E2C9C89AFULL, "dae000_mother6028531.cl"},
	{0xB9ADF830796E1B7FULL, "store_dae370cr_mage.utm"},
	{0xB9AE1523AE1CFBB2ULL, "6097231_m.fxe"},
	{0xB9AE9C4AC22D43A3ULL, "zz_mer_debug6017107.cl"},
	{0xB9AEFA6406B9E7BCULL, "dae000_gen_turn_in_56178429.cl"},
	{0xB9AF30C71E47D2BCULL, "vxlt_darktown_2xt_0_0.vlm"},
	{0xB9AFAA160E93B297ULL, "hlac_w_blightlands_0_10y.dds"},
	{0xB9B0E2C90863C817ULL, "vxlt_backalley_g_1hv_0_0.vlm"},
	{0xB9B1042123932CEEULL, "vxw_woundedcoast_164_0_0.vlm"},
	{0xB9B14AD83EB66341ULL, "gen00fl_merrill6107070.cl"},
	{0xB9B1ADDAB1A24B34ULL, "vxht_0_hightown_c8n_0_0.vlm"},
	{0xB9B1C5140A1BE73AULL, "760711466.ogg"},
	{0xB9B1EEFD9DF1F49DULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv1.evt"},
	{0xB9B1FCA39B4B0BAAULL, "dae300_vashothlead6062825.cl"},
	{0xB9B2724D9F564349ULL, "420359383.ogg"},
	{0xB9B2CC5894890BBEULL, "vxht_mansion_n_6o3_0_0.vlm"},
	{0xB9B2FDDFCD0E333AULL, "llac_lt_darktown_-1_48.dds"},
	{0xB9B38EFD0C4D24C3ULL, "vxw_woundedcoast_1dg_0_0.vlm"},
	{0xB9B3DBDA61E9EA67ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_loop.gad"},
	{0xB9B400F454509FA6ULL, "mer320_hubert6096331.cl"},
	{0xB9B44ADDCD82DA54ULL, "1539222.ogg"},
	{0xB9B465B92211C13CULL, "and100_anders.cnv"},
	{0xB9B4EEEDA7EA4BBEULL, "lgt110_arl_bann_teagan6115616.cl"},
	{0xB9B533C469E6ADACULL, "mcr310_argument6127587.cl"},
	{0xB9B58ADA22DBA8B4ULL, "mag210_arianni6021251.cl"},
	{0xB9B591E687797B2AULL, "552273208.ogg"},
	{0xB9B68D46E4396B2BULL, "750441130.ogg"},
	{0xB9B69066492DBEC0ULL, "6133768_m.fxe"},
	{0xB9B69C8723CFFA19ULL, "vxlt_undercityencounter_1c1_0_0.vlm"},
	{0xB9B6CABD0D83D7AEULL, "554576194.ogg"},
	{0xB9B72B14A82B5572ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_act_c.impa.evt"},
	{0xB9B7331818FCD772ULL, "89.cut"},
	{0xB9B747CC82F3EDF0ULL, "1005525576.ogg"},
	{0xB9B76DFFC66ADAE6ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.impa.evt"},
	{0xB9B7884A76EC945FULL, "vxdr_0_deeproads_cfu_0_1.vlm"},
	{0xB9B7B5345AC7C047ULL, "6151389_m.fxe"},
	{0xB9B8097B1D28746FULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge4.gad"},
	{0xB9B87E63707BF7CEULL, "and100_anders6095912.cl"},
	{0xB9B8852349855629ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_2hands_up.ani"},
	{0xB9B887B92215E38CULL, "and100_anders.bnk"},
	{0xB9B8B5C77CF8A01FULL, "6162295_m.fxe"},
	{0xB9B8C5AC403D0182ULL, "614151994.ogg"},
	{0xB9B8F5DFA4F4400AULL, "one000_ilen6036936.cl"},
	{0xB9B8F7D47B589170ULL, "vxca_cavern1_haunted_ihe_0_0.vlm"},
	{0xB9B951A7E29685C2ULL, "vxca_caverns1_mines_jfb_0_0.vlm"},
	{0xB9B960393A03A803ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_kida_0d.dds"},
	{0xB9B9A3E9F7D107F4ULL, "dae000_gamlen6048425.cl"},
	{0xB9BA0503C78E8054ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_c.anb"},
	{0xB9BA096B52EC7288ULL, "238912652.ogg"},
	{0xB9BA2BCFCE71910DULL, "vxlt_warehouse_5u_0_0.vlm"},
	{0xB9BA4B5598799365ULL, "vxht_0_hightown_a3_alp_0_0.vlm"},
	{0xB9BAF824E101E5F3ULL, "zz_isa_debug6133242.cl"},
	{0xB9BB12BA3BA6E746ULL, "lt_0_lowtown_docks_33_0_lc.mmh"},
	{0xB9BB6E53B4CA8544ULL, "vxca_caverns1_mines_hu3_0_0.vlm"},
	{0xB9BB938F82722E44ULL, "w_bonepitmining_a2_0_0.msh"},
	{0xB9BBA58F82724CADULL, "w_bonepitmining_a2_0_0.mao"},
	{0xB9BBA98F827253E6ULL, "w_bonepitmining_a2_0_0.mmh"},
	{0xB9BC112FE9E6125BULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax107a_0d.dds"},
	{0xB9BC92949597BA74ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tn90l.gad"},
	{0xB9BCD802508F8E34ULL, "6050194_f.fxe"},
	{0xB9BD4A14A82FF505ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_act_c.impa.gad"},
	{0xB9BE20CD9B987C42ULL, "35354541.ogg"},
	{0xB9BE74F7E4D2DCA5ULL, "vxlt_gallowstemplar_f_5hi_0_0.vlm"},
	{0xB9BE84FFC6712031ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.impa.gad"},
	{0xB9BEA2FF149959C8ULL, "fen110_fenris6117976.cl"},
	{0xB9BEF640A4EC7B05ULL, "vxlt_undercitydungeon_1c9_0_1.vlm"},
	{0xB9BF15A8BF148B9FULL, "750061253.ogg"},
	{0xB9BF75FE1589456EULL, "vxw_woundedcoast_115_0_0.vlm"},
	{0xB9BFA023498BA2ABULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_2hands_up.gad"},
	{0xB9C0830FD06E4921ULL, "gen00fl_carver6108891.cl"},
	{0xB9C097C65B944AC8ULL, "188262757.ogg"},
	{0xB9C0AF5A242DECECULL, "gen00cs_cs_end_music_int_3.ncs"},
	{0xB9C0ECA20E7061E4ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_0.dds"},
	{0xB9C106117C62E775ULL, "lt_gallowscourtyard_n_58_0.rml"},
	{0xB9C150A7D2BD6FC7ULL, "vxlt_ruinedhouse_n_ad_0_0.vlm"},
	{0xB9C1B893339C2111ULL, "cs_mcr371_db_harvester_warr.wbd"},
	{0xB9C200D9FE433D43ULL, "vxw_woundedcoast_1of_0_0.vlm"},
	{0xB9C2311105527FCDULL, "w_encounter3_129_0.msh"},
	{0xB9C23F1105529700ULL, "w_encounter3_129_0.mao"},
	{0xB9C2431105529E53ULL, "w_encounter3_129_0.mmh"},
	{0xB9C2457A855D6D1DULL, "925405501.ogg"},
	{0xB9C2FF5A5518A683ULL, "931337904.ogg"},
	{0xB9C361FECC33ED49ULL, "follower_banter6143336.cl"},
	{0xB9C38D3496394961ULL, "hlac_lt_undercityencounter_-1_21z.dds"},
	{0xB9C3A194959DF252ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tn90l.ani"},
	{0xB9C3E41991AE5717ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_magicchart.mmh"},
	{0xB9C3E7F0C1ACE6C5ULL, "137274744.ogg"},
	{0xB9C3E81991AE5E24ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_magicchart.mao"},
	{0xB9C3FA1991AE7C89ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_magicchart.msh"},
	{0xB9C48DEE57F0AC7FULL, "dae000_keep_bureaucrat_2_st6165388.cl"},
	{0xB9C4A8C24BFB9274ULL, "dae010_shopper_f1.utc"},
	{0xB9C4D2EDE0382B9FULL, "6108578_m.fxe"},
	{0xB9C56024BB8FF86CULL, "vxlt_gamlenshouse_n_is_0_0.vlm"},
	{0xB9C585FFC67740DFULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.impa.ani"},
	{0xB9C58B96A8936D3BULL, "90670425.ogg"},
	{0xB9C59D2349900910ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_2hands_up.evt"},
	{0xB9C5A1FA88AB29E0ULL, "vxlt_0_lowtown_h9z_0_0.vlm"},
	{0xB9C66C87E5B0F8C0ULL, "llac_w_blightlands_-131070_0.dds"},
	{0xB9C6E403C798D217ULL, "art\\vfx\\combat\\vfx_m_sta_nat_range_c.mmh"},
	{0xB9C760B2F24DC13CULL, "gen00fl_merrill6130066.cl"},
	{0xB9C7C8DA61FA72ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_loop.ani"},
	{0xB9C7E047020541D5ULL, "qcr000_arishok6166990.cl"},
	{0xB9C8F3C7ED75D1F2ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax106a_0d.dds"},
	{0xB9C9576FCD2FC605ULL, "mer341ar_caverns.are"},
	{0xB9C99203A25876D8ULL, "vxlt_undercityencounter_1lq_0_0.vlm"},
	{0xB9CA67DA8B4E8185ULL, "zz_followers_banter6157357.cl"},
	{0xB9CA79428F36DA03ULL, "48343368.ogg"},
	{0xB9CACA1AC32DD4C0ULL, "619573434.ogg"},
	{0xB9CAD2CB3E39EE9EULL, "vxht_0_hightown_a3_at7_0_0.vlm"},
	{0xB9CB651D46D451A2ULL, "104362510.ogg"},
	{0xB9CC267B1D394E35ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge4.ani"},
	{0xB9CC9B1FA187C3A7ULL, "328666358.ogg"},
	{0xB9CCA98F8280C6ADULL, "w_bonepitmining_a2_0_0.rml"},
	{0xB9CD07253F3B9CE8ULL, "669577493.ogg"},
	{0xB9CD71CB5E25B7D5ULL, "441681117.ogg"},
	{0xB9CD825944BA64D6ULL, "785950046.ogg"},
	{0xB9CDE4BBB225F1A6ULL, "6016139_m.fxe"},
	{0xB9CDECE3A331F66DULL, "vxw_woundedcoast_15v_0_0.vlm"},
	{0xB9CE4D12DA6A90C0ULL, "hlac_lt_gallowstemplar_n_2391_0z.dds"},
	{0xB9CE73A07281C11AULL, "hlac_lt_backalley_n_1_0z.dds"},
	{0xB9CE90D47CD83CE2ULL, "6133601_m.fxe"},
	{0xB9CF5932B76B032BULL, "sst_ss_torpor.gda"},
	{0xB9CFD992B285C2B5ULL, "6068432_m.fxe"},
	{0xB9D072B6F03955C3ULL, "vxw_woundedcoast_1ji_0_0.vlm"},
	{0xB9D0AC18D8F8B580ULL, "543262563.ogg"},
	{0xB9D0B57D61FB32D5ULL, "dae200_saemus6141187.cl"},
	{0xB9D0CC6938718727ULL, "qun221_petrice6075975.cl"},
	{0xB9D0CD35E572BC50ULL, "6117713_m.fxe"},
	{0xB9D0CECE3A3886CAULL, "gen00fl_varric6076021.cl"},
	{0xB9D1ACC65CC96909ULL, "vxht_0_hightown_a3_4n3_0_0.vlm"},
	{0xB9D1B178B6B226CBULL, "vxlt_backalley_g_1kk_0_0.vlm"},
	{0xB9D1D8585D41029BULL, "zz_dae_debug6029856.cl"},
	{0xB9D30E463740DA96ULL, "mrl000_marethari6069831.cl"},
	{0xB9D322D95C4AE033ULL, "vxlt_0_lowtown_docks_n_c3j_0_0.vlm"},
	{0xB9D3431642191074ULL, "vxht_mansion_n_5bq_0_0.vlm"},
	{0xB9D374202D19EFC0ULL, "6139168_m.fxe"},
	{0xB9D3A583BF7F5EABULL, "6099456_m.fxe"},
	{0xB9D440B2972D6C99ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_5x.dds"},
	{0xB9D5089189E86426ULL, "6134957_f.fxe"},
	{0xB9D53D9DACDF3EA8ULL, "252702048.ogg"},
	{0xB9D53DEA5AD3D9B0ULL, "vxht_mansion2b_3ee_0_0.vlm"},
	{0xB9D55A83086B002BULL, "326786924.ogg"},
	{0xB9D5AF0BA59036D6ULL, "6116745_m.fxe"},
	{0xB9D5E7C490C56796ULL, "6081377_m.fxe"},
	{0xB9D634125637C168ULL, "hightown_male_noble_gossip6174052.cl"},
	{0xB9D65177C394CCC0ULL, "dae000_madam_lusine6151623.cl"},
	{0xB9D6585AF441ED75ULL, "6129098_m.fxe"},
	{0xB9D7952BDD140160ULL, "6175326_m.fxe"},
	{0xB9D863BE111E6235ULL, "vxw_blightlands_129_0_0.vlm"},
	{0xB9D872B74CB9144AULL, "ht_0_hightown_n_43_0.rml"},
	{0xB9D8FA34FC33BBB2ULL, "gen00fl_aveline6121904.cl"},
	{0xB9D9110A23D82607ULL, "mag100_karras6034924.cl"},
	{0xB9D9194F43BFBDE0ULL, "follower_banter6143119.cl"},
	{0xB9D933D4AF0E5926ULL, "art\\vfx\\combat\\vfx_g_dual_att_front_c.anb"},
	{0xB9D988EBB1178A7FULL, "591268835.ogg"},
	{0xB9DA3F4D1A4505C3ULL, "6033759_m.fxe"},
	{0xB9DA586FCD3E3B32ULL, "mer341ar_caverns.lst"},
	{0xB9DAE451DF16D373ULL, "vxw_encounter3_66_0_0.vlm"},
	{0xB9DB39429C09E630ULL, "760765096.ogg"},
	{0xB9DB476D1238F7F2ULL, "vxht_0_hightown_ber_0_0.vlm"},
	{0xB9DBB651013FEF11ULL, "vxlt_warehouse_cg_0_0.vlm"},
	{0xB9DC116352634495ULL, "hlac_lt_0_lowtown_0_9y.dds"},
	{0xB9DC5EC095280EEDULL, "vxw_woundedcoast_10u_0_0.vlm"},
	{0xB9DC6AC8340C2F03ULL, "vxdr_primevalentrance_2vy_0_2.vlm"},
	{0xB9DC839C3F52AB51ULL, "art\\characters\\weapons\\shields\\w_shd_elna.msh"},
	{0xB9DC9D9C3F52D70FULL, "art\\characters\\weapons\\shields\\w_shd_elna.mmh"},
	{0xB9DCF06071009114ULL, "mage_bloodsacrifice.dds"},
	{0xB9DD3E4B034A8F66ULL, "vxht_mansion2b_29z_0_0.vlm"},
	{0xB9DD8D194569CC79ULL, "803796097.ogg"},
	{0xB9DDBACC40A05D77ULL, "ser270_templar6149398.cl"},
	{0xB9DDDB94F505D9C6ULL, "qun110_petrice6037092.cl"},
	{0xB9DE7019D487AFDBULL, "vxw_sundermount_31j_0_0.vlm"},
	{0xB9DF808CB86F01C8ULL, "405210644.ogg"},
	{0xB9DFCAB1F0FDBDF8ULL, "hlac_lt_undercityserialkiller_-1_22x.dds"},
	{0xB9DFEFE7F175F153ULL, "sib100_father_contact.bnk"},
	{0xB9E0E68DC258A69FULL, "llac_lt_0_lowtown_n_65536_0.dds"},
	{0xB9E138A5ABD424F8ULL, "gen00fl_aveline6139118.cl"},
	{0xB9E1A0558E37D3A0ULL, "rdr111cr_assistant.utc"},
	{0xB9E1B6A7C4245C9FULL, "6106778_m.fxe"},
	{0xB9E1BC02B1D3EB39ULL, "hlac_lt_gallowstemplar_f_2394_3y.dds"},
	{0xB9E1EB14EB0C9DE7ULL, "vxlt_blackemporium_13t_0_0.vlm"},
	{0xB9E210A40ADC1D7CULL, "821222566.ogg"},
	{0xB9E2C1B8E2F6096CULL, "6030454_m.fxe"},
	{0xB9E30EAFFA3DFA04ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.mmh"},
	{0xB9E311E7F17860A3ULL, "sib100_father_contact.cnv"},
	{0xB9E3DF6F8E4B53DAULL, "vxlt_0_lowtown_n_ex1_0_0.vlm"},
	{0xB9E41BA8A7D964C8ULL, "vxht_0_hightown_a3_6pb_0_0.vlm"},
	{0xB9E41BB2A90A9CA7ULL, "hlac_w_encounter1_131070_0x.dds"},
	{0xB9E4BACC9710A203ULL, "234368564.ogg"},
	{0xB9E52C0257D69FA7ULL, "dae000_orana6149113.cl"},
	{0xB9E5FB889E27E819ULL, "151663209.ogg"},
	{0xB9E60DAD179A9097ULL, "lt_0_lowtown_docks_a3_215_0.msh"},
	{0xB9E61FAD179AAFFAULL, "lt_0_lowtown_docks_a3_215_0.mao"},
	{0xB9E623AD179AB609ULL, "lt_0_lowtown_docks_a3_215_0.mmh"},
	{0xB9E65F66812BA290ULL, "zz_followers_banter6157673.cl"},
	{0xB9E736D4AF1A9BD1ULL, "art\\vfx\\combat\\vfx_g_dual_att_front_c.mmh"},
	{0xB9E7377CFFD040CDULL, "6143810_m.fxe"},
	{0xB9E738FA061F05E5ULL, "ser121_jethann6108803.cl"},
	{0xB9E7818DD49BEB6EULL, "6165109_m.fxe"},
	{0xB9E9B94A042F1072ULL, "vxht_0_hightown_bns_0_0.vlm"},
	{0xB9E9DBDAEA6D2DE3ULL, "dr_dungeon_mmap.dds"},
	{0xB9EA282DF3360791ULL, "vxlt_warehouse_df_0_0.vlm"},
	{0xB9EA4FAF76840A5FULL, "lgt320_mis_resp_swordhunter6150336.cl"},
	{0xB9EA5C43AB03D888ULL, "gen_im_cns_cft_bmb_04.uti"},
	{0xB9EA76F9A738BA03ULL, "vxht_chantry_2ci_0_0.vlm"},
	{0xB9EA88077A7C3ABCULL, "vxlt_undercityencounter_1fk_0_0.vlm"},
	{0xB9EA90AB705AE3FDULL, "834217833.ogg"},
	{0xB9EAB9311C1C8742ULL, "hlac_ca_caverns1_24423_5y.dds"},
	{0xB9EAE4C7236A8A24ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isab.phy"},
	{0xB9EB115FD0752B5CULL, "6118153_f.fxe"},
	{0xB9EBCD1F1DBCC6D4ULL, "6166380_m.fxe"},
	{0xB9EC552E6B7621FEULL, "vxlt_backalley_n_1ft_0_0.vlm"},
	{0xB9ED3284EF431844ULL, "711844706.ogg"},
	{0xB9ED5670D42586C3ULL, "vxw_woundedcoast_1pk_0_0.vlm"},
	{0xB9ED680AADB93774ULL, "vxlt_hangedman_1qq_0_0.vlm"},
	{0xB9EDC1BD6E2BCA94ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.impa.ani"},
	{0xB9EE3FD131BF8E71ULL, "mag170_thrask6171443.cl"},
	{0xB9EE588ABAB350C6ULL, "static_lightblend_p0.psh.fxo.dx11"},
	{0xB9EEFB75DA4133F8ULL, "vxlt_hangedman_n_1ti_0_0.vlm"},
	{0xB9EF14DCB2F6375EULL, "pro000st_flemeth_intro.stg"},
	{0xB9EF29418E201B53ULL, "44498523.ogg"},
	{0xB9EFA90B6E0F75CAULL, "225976355.ogg"},
	{0xB9EFACF4B697F65EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masa_0s.dds"},
	{0xB9EFB9D67E1E9201ULL, "933206363.ogg"},
	{0xB9EFE5C2F9432363ULL, "vxlt_0_lowtown_docks_a3_n_apn_0_0.vlm"},
	{0xB9F026D026054174ULL, "vxht_mansion_n_5hw_0_0.vlm"},
	{0xB9F0893DA36B8FABULL, "6093656_m.fxe"},
	{0xB9F1040317F84CDCULL, "vxht_0_hightown_n_2es_0_0.vlm"},
	{0xB9F149CA88517942ULL, "646879886.ogg"},
	{0xB9F14BAE38ABD36DULL, "430771210.ogg"},
	{0xB9F22D89BE0159B6ULL, "ser120_ghyslain6119910.cl"},
	{0xB9F2DACB6D035233ULL, "follower_banter6155625.cl"},
	{0xB9F2EF185437953EULL, "727573690.ogg"},
	{0xB9F30194992CED08ULL, "237446125.ogg"},
	{0xB9F38E5CCE6E4192ULL, "vxlt_gallowstemplar_3sf_0_0.vlm"},
	{0xB9F3AD28F1571521ULL, "follower_banter6151447.cl"},
	{0xB9F3BDBBB84B3222ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_leli.phy"},
	{0xB9F40EB5E8CC922FULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_16bb.mao"},
	{0xB9F41012D39C4A27ULL, "vxlt_0_lowtown_n_gof_0_1.vlm"},
	{0xB9F41AB5E8CCA6CCULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_16bb.mmh"},
	{0xB9F420B5E8CCB0BEULL, "art\\levels\\races\\natural\\kirkcave\\props\\0cc_16bb.msh"},
	{0xB9F456D7A0BC18D1ULL, "vxca_caverns2prc_4ss_0_0.vlm"},
	{0xB9F4BEBD6E31E4F6ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.impa.gad"},
	{0xB9F52CB19CE51F40ULL, "972103844.ogg"},
	{0xB9F5885B0392EB14ULL, "6035372_m.fxe"},
	{0xB9F632358108AC88ULL, "269950173.ogg"},
	{0xB9F688D3083A77A0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\hu_reaction_impale_ex.evt"},
	{0xB9F6E34AA1498112ULL, "6068581_m.fxe"},
	{0xB9F6E97C8D95C192ULL, "hlac_ht_mansion2b_818_60x.dds"},
	{0xB9F6FFB0A52C0436ULL, "vxlt_gallowstemplar_4tx_0_0.vlm"},
	{0xB9F709CA198976FAULL, "dae000cr_grand_cleric6173653.cl"},
	{0xB9F76337291128DEULL, "zz_followers_banter6158883.cl"},
	{0xB9F7EB1241807DE6ULL, "hlac_lt_0_lowtown_docks_0_5y.dds"},
	{0xB9F8BA4499FB0253ULL, "hlac_lt_0_lowtown_docks_-65535_0z.dds"},
	{0xB9F8F80512A2BC1CULL, "478775170.ogg"},
	{0xB9F9AE0C94541386ULL, "hlac_w_bonepitdestroyed_a3_6_0x.dds"},
	{0xB9F9B7F1813DDEA5ULL, "lt_gallowscourtyard_mmi.gff"},
	{0xB9F9F965177DB630ULL, "vxlt_0_lowtown_docks_a3_n_cvx_0_1.vlm"},
	{0xB9FA0A69F26D10B7ULL, "llac_dr_slavers_3215_4.dds"},
	{0xB9FA4966464F5C1AULL, "zz_approval_debug6003920.cl"},
	{0xB9FAE49174358BAFULL, "6096798_m.fxe"},
	{0xB9FAFEF16E10BAD2ULL, "26604936.ogg"},
	{0xB9FBD9BD6E383155ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_proj_p.impa.evt"},
	{0xB9FBF2D795552A04ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_exit.ani"},
	{0xB9FBF3F833559698ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxb.phy"},
	{0xB9FC25521D4EA69CULL, "637035295.ogg"},
	{0xB9FC2C53D610761DULL, "vxw_encounter2_e6_0_0.vlm"},
	{0xB9FC7BDC4CD7DB4DULL, "art\\characters\\playercharacter\\elffemale\\arm_nuda.phy"},
	{0xB9FCAE93166F0D8AULL, "art\\characters\\playercharacter\\humanmale\\arm_medj.mao"},
	{0xB9FCB293166F14D9ULL, "art\\characters\\playercharacter\\humanmale\\arm_medj.mmh"},
	{0xB9FCBC93166F25C7ULL, "art\\characters\\playercharacter\\humanmale\\arm_medj.msh"},
	{0xB9FCC8C0EF041098ULL, "628593076.ogg"},
	{0xB9FD070F8C94704BULL, "vxlt_backalley_g_1zl_0_0.vlm"},
	{0xB9FD3C8E24051B94ULL, "hlac_lt_0_lowtown_docks_n_-196607_0y.dds"},
	{0xB9FD66D049A30681ULL, "follower_banter6154116.cl"},
	{0xB9FD6D52CC374C78ULL, "vxlt_hangedman_n_1qf_0_0.vlm"},
	{0xB9FE22BD5EF55741ULL, "vxdr_primevaltomb_2ul_0_0.vlm"},
	{0xB9FE68AD201DC829ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.cess.ani"},
	{0xB9FE7E4F8DDFE948ULL, "6196970_m.fxe"},
	{0xB9FE9D63DBC5F219ULL, "dae121ip_at_brothel_high.utp"},
	{0xB9FF28B2A4C31CAFULL, "254542058.ogg"},
	{0xB9FFB7109E9F4537ULL, "6067703_m.fxe"},
	{0xBA00145ED3205CBCULL, "335144170.ogg"},
	{0xBA0019F038EEF60DULL, "def_rock_s.dds"},
	{0xBA006C25FA4F598CULL, "77018777.ogg"},
	{0xBA00A2819080A6D3ULL, "follower_banter6155571.cl"},
	{0xBA00B365A4565B61ULL, "88648085.ogg"},
	{0xBA00C329723784DAULL, "vxlt_0_lowtown_n_en3_0_0.vlm"},
	{0xBA0106D5C7BDE474ULL, "vxdr_0_deeproads_74p_0_1.vlm"},
	{0xBA016D0A44030DD4ULL, "gen00fl_fenris6126753.cl"},
	{0xBA0213823FE404F7ULL, "pro000_aveline_intro6109381.cl"},
	{0xBA024014F9C569C2ULL, "829801877.ogg"},
	{0xBA02B8EF0BDCE496ULL, "6108724_m.fxe"},
	{0xBA02EFD7955B4466ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_exit.gad"},
	{0xBA02F6EF5CB28913ULL, "6047688_m.fxe"},
	{0xBA03B027446F224CULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.cust.ani"},
	{0xBA03C2DFEAE0E699ULL, "llac_ca_caverns1_mines_24421_4.dds"},
	{0xBA041B5EAECDDEF0ULL, "vxlt_gallowscourtyard_7ap_0_0.vlm"},
	{0xBA042FCF4C4EAB94ULL, "lt_0_lowtown_n_20_0_lc.mmh"},
	{0xBA044CDFA0AD1895ULL, "gen00fl_aveline6132853.cl"},
	{0xBA0472303FAC6403ULL, "mag101_aombanter6052753.cl"},
	{0xBA04A49C4B49D2E7ULL, "72424813.ogg"},
	{0xBA055D15E633AA56ULL, "llac_ht_0_hightown_n_65536_3.dds"},
	{0xBA0583AD202414ABULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.cess.gad"},
	{0xBA05CF15A721475DULL, "gen00fl_sebastian6188609.cl"},
	{0xBA060C395BD1119EULL, "99193957.ogg"},
	{0xBA0631623A7C5906ULL, "vxlt_backalley_n_1df_0_0.vlm"},
	{0xBA06863B64DF940BULL, "vrc221cr_bartrand.ncs"},
	{0xBA0755D83CE344C3ULL, "786717829.ogg"},
	{0xBA075AB38B24EB03ULL, "vxht_chantry_2yk_0_0.vlm"},
	{0xBA0807F1A571F45EULL, "hlac_lt_0_lowtown_a2_-65535_2z.dds"},
	{0xBA08480E3D3E0FA7ULL, "vxht_bartrand_n_6k5_0_0.vlm"},
	{0xBA08678F61770EB9ULL, "mag101_grace6053104.cl"},
	{0xBA087F3E4DF3570BULL, "gen00fl_aveline6137593.cl"},
	{0xBA08EAD7955FA745ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_exit.evt"},
	{0xBA08FB99D458A5FDULL, "gen00fl_sebastian6195823.cl"},
	{0xBA095DF23E69D157ULL, "6130973_m.fxe"},
	{0xBA096E114BDC9309ULL, "152325830.ogg"},
	{0xBA097FB38A5E4C15ULL, "vxht_keep_5gt_0_0.vlm"},
	{0xBA09865799B902A6ULL, "6051686_m.fxe"},
	{0xBA099417CAA0A475ULL, "vxw_encounter1_bn_0_0.vlm"},
	{0xBA09AA5A2488AD9EULL, "hlac_lt_undercityserialkiller_-1_23y.dds"},
	{0xBA0AA1641D9AEAF2ULL, "vxht_chantry_2t4_0_1.vlm"},
	{0xBA0ABD274475568EULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.cust.gad"},
	{0xBA0AD73F24D24265ULL, "6033102_m.fxe"},
	{0xBA0AF3C9F11FA444ULL, "hlac_ht_keep_n_3732_2y.dds"},
	{0xBA0B11AFFA5F51FBULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_proj_p.anb"},
	{0xBA0B407DB6F7D001ULL, "6028941_f.fxe"},
	{0xBA0B80AD20287B10ULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_p.cess.evt"},
	{0xBA0B9F8980787C61ULL, "llac_w_encounter3_-1_0.dds"},
	{0xBA0BE192302A93C9ULL, "40596726.ogg"},
	{0xBA0BEA086D03505EULL, "vxw_encounter2_2h_0_0.vlm"},
	{0xBA0C660EB2643951ULL, "pro000_colonel_ewald6039946.cl"},
	{0xBA0C7210337C1EAEULL, "art\\vfx\\cinematic\\vfx_anders_childchest_c.impa.ani"},
	{0xBA0CABC4BE0225A8ULL, "nat380_nathaniel6071281.cl"},
	{0xBA0D9011E2C8D804ULL, "fonts_hu-hu.gfx"},
	{0xBA0EFD8F970112DEULL, "dae000_player_generic_find6179061.cl"},
	{0xBA0F99C7D959A8A0ULL, "gen00fl_anders6119421.cl"},
	{0xBA0FDE722028D0B5ULL, "vxlt_0_lowtown_docks_a3_d1a_0_0.vlm"},
	{0xBA0FFBF4CD4DAB97ULL, "450222284.ogg"},
	{0xBA10610A304AFF71ULL, "dae000_arishok6067657.cl"},
	{0xBA1071F8F215FBE5ULL, "440093063.ogg"},
	{0xBA107815002C7764ULL, "tm_gwp_h02.tnt"},
	{0xBA10B8274479BA6DULL, "art\\vfx\\creatures\\fxc_ohurl_conj_c.cust.evt"},
	{0xBA10C9553E681523ULL, "gen00fl_isabela6075290.cl"},
	{0xBA111515CC57EFAEULL, "hlac_ht_mansion2_818_29y.dds"},
	{0xBA1168CC4EA8A193ULL, "6042988_m.fxe"},
	{0xBA11DA0F35382CCEULL, "299845544.ogg"},
	{0xBA11E59616974097ULL, "vxw_bonepitmining_a2_25b_0_0.vlm"},
	{0xBA11EC9E45735D64ULL, "vxlt_backalley_g_1dm_0_0.vlm"},
	{0xBA1202071AF24419ULL, "243471963.ogg"},
	{0xBA1286D8A926A9FBULL, "w_sundermount_141_0.mmh"},
	{0xBA1292851CD5A01BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xb.phy"},
	{0xBA1292D8A926BD98ULL, "w_sundermount_141_0.mao"},
	{0xBA129DFF23D42935ULL, "vxlt_darktown_2sh_0_0.vlm"},
	{0xBA12A4D8A926DCE5ULL, "w_sundermount_141_0.msh"},
	{0xBA12C18204666814ULL, "gen00fl_anders6125130.cl"},
	{0xBA12F468C933DFC4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_crumble.msh"},
	{0xBA130668C933FE2DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_crumble.mao"},
	{0xBA130A68C9340566ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_crumble.mmh"},
	{0xBA139C226515FFE0ULL, "vxca_cavern1_haunted_edm_0_0.vlm"},
	{0xBA13B5D01B98CC02ULL, "79410926.ogg"},
	{0xBA13FB6A4C0932A1ULL, "199847821.ogg"},
	{0xBA1450668472E00BULL, "6079890_m.fxe"},
	{0xBA145AAD14512978ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_anda_0n.dds"},
	{0xBA146B626E592645ULL, "mrl000_marethari6129330.cl"},
	{0xBA1477FC8C03A9ADULL, "mrl150_merrill_pt6.stg"},
	{0xBA14A33F2C727186ULL, "vxlt_backalley_n_1og_0_0.vlm"},
	{0xBA14D6F235F3D098ULL, "447818248.ogg"},
	{0xBA1529C6D0245EB0ULL, "6075787_m.fxe"},
	{0xBA157B303C1B013BULL, "161654932.ogg"},
	{0xBA160BCA30D6E4A8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12man.mmh"},
	{0xBA1617CA30D6F94BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12man.mao"},
	{0xBA1629CA30D717D2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12man.msh"},
	{0xBA163203A0CDFD15ULL, "87061100.ogg"},
	{0xBA16893A437D8012ULL, "27636121.ogg"},
	{0xBA169C5ED36BDF8FULL, "6121038_m.fxe"},
	{0xBA16ADB45AD647FDULL, "47774958.ogg"},
	{0xBA16E4112A05EB5DULL, "dae000_hate_follower_react6103234.cl"},
	{0xBA17AAC541586B7EULL, "vxlt_backalley_n_1ww_0_0.vlm"},
	{0xBA17C56087A1C27DULL, "399522223.ogg"},
	{0xBA17F1907C546495ULL, "vxht_keep_5js_0_0.vlm"},
	{0xBA1805F4BC96BCF5ULL, "vxw_encounter1_gm_0_0.vlm"},
	{0xBA1885EE5E7ED89DULL, "vxht_0_hightown_a2_aqz_0_0.vlm"},
	{0xBA18E4F034F2448AULL, "vxlt_0_lowtown_docks_a3_n_d2d_0_0.vlm"},
	{0xBA19100DB9FCA71DULL, "vxw_encounter2_k4_0_0.vlm"},
	{0xBA1926B47241AB3AULL, "lt_0_lowtown_docks_133_0.phy"},
	{0xBA193C6C95D1CD0CULL, "hlac_w_bonepit_0_13z.dds"},
	{0xBA1964E4A574B6DFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containerb.msh"},
	{0xBA196AE4A574C081ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containerb.mmh"},
	{0xBA1976E4A574D522ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containerb.mao"},
	{0xBA19A4BBB86BD775ULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_leli.mmh"},
	{0xBA19A6BBB86BDAEBULL, "art\\characters\\playercharacter\\humanfemale\\mt_cbt_leli.msh"},
	{0xBA1A0609F47AD0B3ULL, "758904387.ogg"},
	{0xBA1A5644D4C5D7A4ULL, "lt_0_lowtown_docks_26_0_lc.mmh"},
	{0xBA1A6A1033884EFFULL, "art\\vfx\\cinematic\\vfx_anders_childchest_c.impa.evt"},
	{0xBA1A7B35AD81E20EULL, "vxht_mansion2_3bv_0_0.vlm"},
	{0xBA1A8FA9E7A42A4DULL, "6131881_f.fxe"},
	{0xBA1AA309FDA52E24ULL, "861452332.ogg"},
	{0xBA1AA398F3F5AA3BULL, "vxlt_0_lowtown_a2_n_c_byz_0_0.vlm"},
	{0xBA1AC234EB07569EULL, "204622459.ogg"},
	{0xBA1B3EAEF4D42A06ULL, "601624382.ogg"},
	{0xBA1B4B98D0FA0764ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ext.ani"},
	{0xBA1B5EC46F8225FFULL, "design_test_linear.arl"},
	{0xBA1BC81FAA7BAD73ULL, "vxht_0_hightown_n_b21_0_0.vlm"},
	{0xBA1BFEB41D7A9C83ULL, "612292207.ogg"},
	{0xBA1C9A21C97545B5ULL, "415016475.ogg"},
	{0xBA1CDA2E9BED91DFULL, "art\\characters\\weapons\\axes\\w_axe_03a.msh"},
	{0xBA1CE02E9BED9B81ULL, "art\\characters\\weapons\\axes\\w_axe_03a.mmh"},
	{0xBA1CE9FC5E68C31CULL, "vxlt_undercityserialkiller_15r_0_0.vlm"},
	{0xBA1CEC2E9BEDB022ULL, "art\\characters\\weapons\\axes\\w_axe_03a.mao"},
	{0xBA1D0168C93C4A31ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_crumble.phy"},
	{0xBA1D310C37BCD9F0ULL, "6141668_m.fxe"},
	{0xBA1D744AF2F0840DULL, "346327950.ogg"},
	{0xBA1E4580187ACE48ULL, "261261572.ogg"},
	{0xBA1E5E9FCBADC835ULL, "6146970_m.fxe"},
	{0xBA1EE4E1E38F1E42ULL, "vxht_keep_627_0_0.vlm"},
	{0xBA1EF0E1EAE74647ULL, "zz_vrc_debug6064272.cl"},
	{0xBA1F089821B8BF40ULL, "vxht_0_hightown_c37_0_0.vlm"},
	{0xBA1F5BE510E2D0F8ULL, "vxht_mansion_n_3qv_0_0.vlm"},
	{0xBA1F9FDF50B516E6ULL, "522726851.ogg"},
	{0xBA20FE96E17B1195ULL, "hlac_ht_0_hightown_n_1_2x.dds"},
	{0xBA210832AA7836E9ULL, "hlac_ca_caverns2_5802_73x.dds"},
	{0xBA210FDC15CA41B5ULL, "vxlt_darktown_2vw_0_0.vlm"},
	{0xBA214B10338E38E8ULL, "art\\vfx\\cinematic\\vfx_anders_childchest_c.impa.gad"},
	{0xBA21A75A878E3CB0ULL, "50564857.ogg"},
	{0xBA21A95C5A5CB984ULL, "6155427_m.fxe"},
	{0xBA21BD4C704E9CD4ULL, "lowtown_fem_alienage1_sta6162605.cl"},
	{0xBA2205ACB70152B7ULL, "follower_banter6154015.cl"},
	{0xBA224898D10020C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ext.gad"},
	{0xBA22C952C3FBBCB1ULL, "6093590_f.fxe"},
	{0xBA23151C1E688A06ULL, "vxlt_backalley_n_1jh_0_0.vlm"},
	{0xBA23210B88C50C2BULL, "mag320_orsino6060233.cl"},
	{0xBA2344512684C197ULL, "vxw_bonepitmining_a2_1q5_0_0.vlm"},
	{0xBA23A4B9B365E219ULL, "gen00fl_fenris6106644.cl"},
	{0xBA23A8B11B1E4C7DULL, "3180540.ogg"},
	{0xBA23DF244DFEFDC2ULL, "140143809.ogg"},
	{0xBA23DF2B026C2CEAULL, "llac_lt_0_lowtown_a2_n_-131073_0.dds"},
	{0xBA23EDFE03B72507ULL, "gen00fl_fenris6115387.cl"},
	{0xBA23EFA1BB0E6991ULL, "vxlt_warehouse_xz_0_0.vlm"},
	{0xBA24783FE7F18CADULL, "275868508.ogg"},
	{0xBA252BC8212A40A7ULL, "vxht_bartrand_n_6q7_0_0.vlm"},
	{0xBA25602B6385ECD0ULL, "vxdr_primevalentrance_21y_0_0.vlm"},
	{0xBA256E40EC4DC83AULL, "ave200_donnic6111203.cl"},
	{0xBA258B3B64FA4DE5ULL, "vrc221cr_bartrand.utc"},
	{0xBA261CA2334E83FEULL, "vxlt_backalley_n_1rh_0_0.vlm"},
	{0xBA2685AF0A8757B4ULL, "hlac_ht_playershouse_n_855_21x.dds"},
	{0xBA27279A84B9BE2BULL, "vxlt_0_lowtown_a2_n_c_i0r_0_0.vlm"},
	{0xBA289C7793364911ULL, "mcr370_guardsman6086981.cl"},
	{0xBA28B31639E12043ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3b.mao"},
	{0xBA28B71639E12690ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3b.mmh"},
	{0xBA28C51639E13EAAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3b.msh"},
	{0xBA294398D10636A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ext.evt"},
	{0xBA295E74ABE90E68ULL, "llac_lt_gallowscourtyard_-131071_2.dds"},
	{0xBA297E13B5C5680DULL, "tbe000im_apprec_helm.uti"},
	{0xBA29B4C1F5199B6BULL, "124909479.ogg"},
	{0xBA2A3F49AC6D6DA8ULL, "one100dg_vanard6054200.cl"},
	{0xBA2A57A071119FE8ULL, "qun210cr_guard6072157.cl"},
	{0xBA2AFF2559ECF2C3ULL, "232470643.ogg"},
	{0xBA2B09A009B4BB90ULL, "6136334_m.fxe"},
	{0xBA2B10788DF516A5ULL, "vxht_0_hightown_a3_ca9_0_0.vlm"},
	{0xBA2BA7C6D8201C33ULL, "hlac_lt_0_lowtown_docks_-65540_0z.dds"},
	{0xBA2BADD9236829DEULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.dura.evt"},
	{0xBA2C83BBD79A5FBEULL, "vxlt_0_lowtown_a3_n_c_htg_0_0.vlm"},
	{0xBA2CB8946D0B7170ULL, "vxlt_0_lowtown_docks_bbx_0_0.vlm"},
	{0xBA2D3344CA38C4C1ULL, "vxht_playershouse_n_1tc_0_0.vlm"},
	{0xBA2D7757340F17E9ULL, "91590249.ogg"},
	{0xBA2D7A82B045BCD1ULL, "6064047_m.fxe"},
	{0xBA2DF1E9CDE7543AULL, "dae000_pc_home_ambients6208576.cl"},
	{0xBA2E32EF54C48AD2ULL, "gen00fl_sebastian6198133.cl"},
	{0xBA2E8A0E885D28D9ULL, "6123327_f.fxe"},
	{0xBA2E8A152E133B3DULL, "qun110_petrice6205999.cl"},
	{0xBA2EBEA619082D27ULL, "486584309.ogg"},
	{0xBA2F5856873D7EB0ULL, "vxlt_house_f2_0_0.vlm"},
	{0xBA2F81B907C05A35ULL, "vxlt_darktown_2yv_0_0.vlm"},
	{0xBA2F90B8038D2F79ULL, "lgt320_mis_resp_swordhunter6150327.cl"},
	{0xBA2F9C3E56305A12ULL, "92585908.ogg"},
	{0xBA2FB8E2352A22A2ULL, "art\\characters\\playercharacter\\children\\mt_cbt_conn.phy"},
	{0xBA300EEE41C4F931ULL, "6071845_m.fxe"},
	{0xBA30A123082884AFULL, "vxht_mansion_n_4xi_0_0.vlm"},
	{0xBA311C824E7767CEULL, "zz_letter.ncs"},
	{0xBA31B7B53D881CA8ULL, "mag100_cullen6093730.cl"},
	{0xBA31DFA42738D65BULL, "lowtown_male_beggar_sta6151080.cl"},
	{0xBA32703022E816ACULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25_tool.msh"},
	{0xBA32763022E8209EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25_tool.mmh"},
	{0xBA32823022E83535ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25_tool.mao"},
	{0xBA32ABEE096FA85AULL, "vxht_mansion_n_57q_0_1.vlm"},
	{0xBA32B5D8B28592E3ULL, "art\\vfx\\models\\vfx_frontwave.msh"},
	{0xBA32C3D8B285AA3DULL, "art\\vfx\\models\\vfx_frontwave.mmh"},
	{0xBA32C4D9236E6F09ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.dura.gad"},
	{0xBA3441607F10E6AFULL, "vxlt_gallowsprison_n_6mk_0_0.vlm"},
	{0xBA34D54A60409595ULL, "vxht_keep_5pq_0_0.vlm"},
	{0xBA35299B5141A7C3ULL, "197636562.ogg"},
	{0xBA358BFFA2B22975ULL, "vxdr_primevaltomb_9cu_0_0.vlm"},
	{0xBA35A15B6787B1F4ULL, "vxlt_hangedman_1hz_0_0.vlm"},
	{0xBA36F8F6CB5C160CULL, "31794576.ogg"},
	{0xBA372E5022B5D1F9ULL, "hlac_w_bonepitdestroyed_a3_262148_0x.dds"},
	{0xBA37528475D01CE2ULL, "vxlt_0_lowtown_docks_a3_n_c1n_0_0.vlm"},
	{0xBA3792FB6D29CE5EULL, "vxw_blightlands_op_0_0.vlm"},
	{0xBA38ABD98E67DE73ULL, "vxht_0_hightown_n_b83_0_0.vlm"},
	{0xBA38C5C67ECE6DE0ULL, "6120778_f.fxe"},
	{0xBA3995D923743E57ULL, "art\\vfx\\combat\\vfx_m_bld_bloodwound_c.dura.ani"},
	{0xBA39F649382BA20EULL, "6087343_f.fxe"},
	{0xBA3B0E4FD2D44880ULL, "fen210_orana6109730.cl"},
	{0xBA3B65DCD944AB37ULL, "6094608_m.fxe"},
	{0xBA3BA57C19403D80ULL, "6172473_m.fxe"},
	{0xBA3BC89BC77B4F42ULL, "vxht_keep_681_0_0.vlm"},
	{0xBA3BCB54C8004CD8ULL, "980376878.ogg"},
	{0xBA3BE07B91CEC78CULL, "683792942.ogg"},
	{0xBA3BF717DDB3350FULL, "vxlt_0_lowtown_n_cmq_0_0.vlm"},
	{0xBA3BFFAA90DDFCD5ULL, "681443450.ogg"},
	{0xBA3C6C7497FD9648ULL, "hlac_lt_undercityencounter_-1_38y.dds"},
	{0xBA3C989CBF402347ULL, "vxw_woundedcoast_1oz_0_0.vlm"},
	{0xBA3CE1DECDC6D6CAULL, "6108005_m.fxe"},
	{0xBA3D0F61D874C296ULL, "llac_lt_0_lowtown_a2_-131072_2.dds"},
	{0xBA3DBBA05F419226ULL, "6125949_m.fxe"},
	{0xBA3E22F82DD920CBULL, "fxa_stun_spk.dds"},
	{0xBA3E536172A28F25ULL, "vxlt_gallowstemplar_n_5wl_0_0.vlm"},
	{0xBA3E7BC321E4699AULL, "hlac_lt_gallowstemplar_n_2384_6x.dds"},
	{0xBA3E89273EB27F95ULL, "12157917.ogg"},
	{0xBA3EA2D727C50A0DULL, "6139235_m.fxe"},
	{0xBA3EBCB053E14701ULL, "t3_har_rbr.tnt"},
	{0xBA3F79851CFC63BAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xb.msh"},
	{0xBA3F87851CFC7BF3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xb.mao"},
	{0xBA3F8B851CFC8200ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4xb.mmh"},
	{0xBA3FB6455724611BULL, "hlac_lt_gallowstemplar_n_2383_0z.dds"},
	{0xBA3FC5F3D4781A31ULL, "zz_and_debug6134070.cl"},
	{0xBA4066E63404E90AULL, "361606143.ogg"},
	{0xBA40EE5CF3E31885ULL, "6041709_m.fxe"},
	{0xBA4111D96D173467ULL, "qun221_petrice6076844.cl"},
	{0xBA4163B712E522C5ULL, "6016693_m.fxe"},
	{0xBA41B3E38245D820ULL, "6099204_m.fxe"},
	{0xBA41D709D0D387AEULL, "gen_im_wep_mag_sta_lore_02.uti"},
	{0xBA42BC34D26C9190ULL, "108104125.ogg"},
	{0xBA42DE87181F8B99ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtoidle.ani"},
	{0xBA42E56FA17A2D65ULL, "6164484_m.fxe"},
	{0xBA43005C173AB4FEULL, "vxlt_backalley_n_1xj_0_0.vlm"},
	{0xBA4356C6AC270F36ULL, "1055391253.ogg"},
	{0xBA438E35B5C3CF53ULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.impa.ani"},
	{0xBA43C357FFD78521ULL, "vxht_0_hightown_c43_0_0.vlm"},
	{0xBA43D62643CD36E3ULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v2.ani"},
	{0xBA44159BA89E1F7CULL, "308601969.ogg"},
	{0xBA446998A0411C4CULL, "6039397_f.fxe"},
	{0xBA44B27E237A4A0FULL, "llac_ht_0_hightown_-196609_0.dds"},
	{0xBA44C0EACE411928ULL, "6162310_f.fxe"},
	{0xBA451F2C6E090E4DULL, "llac_ht_0_hightown_-65538_0.dds"},
	{0xBA452D73149C289AULL, "vxht_mansion2b_1sk_0_0.vlm"},
	{0xBA456BE4A599FDFEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containerb.phy"},
	{0xBA45A553939C50E4ULL, "llac_ca_caverns2_5802_43.dds"},
	{0xBA461099F568A64DULL, "lgt210_shp_kill_qunari3.bnk"},
	{0xBA4630C90C089E71ULL, "601262660.ogg"},
	{0xBA46C027DEE7E666ULL, "zz_trans2.ncc"},
	{0xBA46E2EC31A2C1BBULL, "hlac_w_encounter2_camp_-1_0y.dds"},
	{0xBA46E4C90D50B6D3ULL, "vxw_bonepitmining_a2_fg_0_0.vlm"},
	{0xBA47E987ACDE1481ULL, "monster4.cut"},
	{0xBA484D824AB1696FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_enter.gad"},
	{0xBA48CC3747F47AFBULL, "vxlt_0_lowtown_docks_a3_btz_0_0.vlm"},
	{0xBA48E12E9C12D8FEULL, "art\\characters\\weapons\\axes\\w_axe_03a.phy"},
	{0xBA49082C4BECFA12ULL, "follower_banter6141818.cl"},
	{0xBA493299F56B15B5ULL, "lgt210_shp_kill_qunari3.cnv"},
	{0xBA49B98718256BDBULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtoidle.gad"},
	{0xBA4A099A7F20F6A1ULL, "6000823_m.fxe"},
	{0xBA4A20F2A32A776AULL, "gen_im_trs_cot_junk3_04.uti"},
	{0xBA4A3A75C7A793D5ULL, "6200410_m.fxe"},
	{0xBA4A5CD30B1DC444ULL, "dae000ip_show_world_map_l.utp"},
	{0xBA4AA76422202E39ULL, "418927998.ogg"},
	{0xBA4AAEDF64608C74ULL, "vxw_encounter2_camp_9o_0_0.vlm"},
	{0xBA4AB5FED60C435DULL, "vxlt_gallowsprison_n_64f_0_0.vlm"},
	{0xBA4AC19B38E3C8D1ULL, "6165348_m.fxe"},
	{0xBA4AFF9BA43E2380ULL, "strmt01_d.dds"},
	{0xBA4B0A79B1363BC7ULL, "vxw_woundedcoast_1ju_0_0.vlm"},
	{0xBA4B984F716EE119ULL, "6052998_m.fxe"},
	{0xBA4BA9E14DB02307ULL, "373851721.ogg"},
	{0xBA4C3B526E522250ULL, "vxlt_0_lowtown_n_gb7_0_0.vlm"},
	{0xBA4C7A1C7F45E477ULL, "vxlt_0_lowtown_a2_n_c_byn_0_0.vlm"},
	{0xBA4DEAC0869B0B40ULL, "vxlt_0_lowtown_docks_aqj_0_0.vlm"},
	{0xBA4E029C5DA3A583ULL, "6071181_f.fxe"},
	{0xBA4E65F68D241DD5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_chantrybasic.phy"},
	{0xBA4EDF096F3FAD8BULL, "vxlt_0_lowtown_a2_n_emk_0_0.vlm"},
	{0xBA4EF7829E6327A0ULL, "ht_mansion_n_3364.anb"},
	{0xBA4F62824AB7AB1CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_enter.evt"},
	{0xBA4F6572D2BBEFB9ULL, "w_sundermount_57_0.msh"},
	{0xBA4F6F72D2BC00A7ULL, "w_sundermount_57_0.mmh"},
	{0xBA4F7372D2BC07F4ULL, "w_sundermount_57_0.mao"},
	{0xBA4F923287F34D41ULL, "hlac_lt_0_lowtown_a2_131070_4x.dds"},
	{0xBA4FC271D4724EC3ULL, "w_woundedcoast_252_0.mao"},
	{0xBA4FC671D4725510ULL, "w_woundedcoast_252_0.mmh"},
	{0xBA4FD471D4726D2AULL, "w_woundedcoast_252_0.msh"},
	{0xBA5011644207BF2CULL, "mag320_noblewoman_16060276.cl"},
	{0xBA5040DE702F87CAULL, "712567156.ogg"},
	{0xBA50771759F1945DULL, "613947751.ogg"},
	{0xBA508635B5CE4B8AULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.impa.evt"},
	{0xBA50D687182BBBC0ULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\relaxtoidle.evt"},
	{0xBA50EE2643D7E9DAULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v2.evt"},
	{0xBA50EEA27F4A0B78ULL, "549525240.ogg"},
	{0xBA512D44E641349DULL, "6054978_m.fxe"},
	{0xBA52568897AA940EULL, "tools_null.mmh"},
	{0xBA528ECD9E59F14EULL, "6002790_m.fxe"},
	{0xBA5295F9FBE2722BULL, "llac_ht_keep_n_4519_2.dds"},
	{0xBA529A5B826C47DDULL, "7158062.ogg"},
	{0xBA5309221CA31C30ULL, "art\\vfx\\environment\\vfx_bl_smoke_vista_l_p.mmh"},
	{0xBA535953823D567FULL, "6163300_m.fxe"},
	{0xBA5394594F52EC08ULL, "vxlt_darktown_5tc_0_0.vlm"},
	{0xBA539F500692411AULL, "vxht_mansion2b_1vd_0_0.vlm"},
	{0xBA53DC909AD0AB6DULL, "ser221_gascard6133744.cl"},
	{0xBA546958908732A6ULL, "hlac_dr_primevaltomb_661_3z.dds"},
	{0xBA5476B55115FF5EULL, "vxw_blightlands_ur_0_0.vlm"},
	{0xBA5483048D87E8E0ULL, "6096313_m.fxe"},
	{0xBA550115DB265261ULL, "6170228-mrl351-pridedemon-d.cut"},
	{0xBA550C77A2C61C7DULL, "vxlt_0_lowtown_a3_n_c_gji_0_0.vlm"},
	{0xBA5539972E262886ULL, "6174550_m.fxe"},
	{0xBA554007F04A3BEAULL, "hlac_ht_0_hightown_n_-65536_4z.dds"},
	{0xBA5556A5FF46CF53ULL, "vxw_bonepitmining_a2_mf_0_0.vlm"},
	{0xBA557972552C0A8BULL, "6143801_m.fxe"},
	{0xBA559FE2354AC7F5ULL, "art\\characters\\playercharacter\\children\\mt_cbt_conn.mmh"},
	{0xBA55A1E2354ACB6BULL, "art\\characters\\playercharacter\\children\\mt_cbt_conn.msh"},
	{0xBA55A6D3353FDB68ULL, "vxht_keep_6gb_0_0.vlm"},
	{0xBA55EC86EF192FE0ULL, "590817433.ogg"},
	{0xBA55FEFE0C9EC04EULL, "243145160.ogg"},
	{0xBA57259EB91B9BBFULL, "889226945.ogg"},
	{0xBA5754D1EA2B0AE7ULL, "vxht_keep_n_6fy_0_0.vlm"},
	{0xBA576535B5D432FDULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.impa.gad"},
	{0xBA578E2EA400FD13ULL, "6208588_m.fxe"},
	{0xBA57E0F833A3D52DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxb.msh"},
	{0xBA57ED2643DE06CDULL, "art\\characters\\creatures\\desiredemon\\animation\\com_v2.gad"},
	{0xBA57EEF833A3ED60ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxb.mao"},
	{0xBA57F2F833A3F3B3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hobotentxb.mmh"},
	{0xBA57FDD4E85F7DADULL, "6170690_m.fxe"},
	{0xBA58151FD3EB6155ULL, "843838517.ogg"},
	{0xBA5843655BAADDE7ULL, "6092228_m.fxe"},
	{0xBA58FB8B6A3F641EULL, "vxlt_0_lowtown_n_eir_0_0.vlm"},
	{0xBA5920BC5656A4F4ULL, "vxw_encounter2_camp_4l_0_0.vlm"},
	{0xBA5969E6AD28DF8BULL, "6087587_m.fxe"},
	{0xBA59A7D8A962E430ULL, "w_sundermount_141_0.phy"},
	{0xBA59CD4551DABA8BULL, "6140804_m.fxe"},
	{0xBA5A0B06535C76D9ULL, "w_encounter2_63_0.msh"},
	{0xBA5A0C22247486A6ULL, "vxlt_warehouse_n_t0_0_0.vlm"},
	{0xBA5A1506535C87C7ULL, "w_encounter2_63_0.mmh"},
	{0xBA5A1906535C8E14ULL, "w_encounter2_63_0.mao"},
	{0xBA5A7B12D5CBD96EULL, "224865911.ogg"},
	{0xBA5B0A74F307EDA2ULL, "vxlt_0_lowtown_a2_n_c_gcl_0_0.vlm"},
	{0xBA5B320368DFAECDULL, "hlac_lt_gallowscourtyard_n_65535_0z.dds"},
	{0xBA5B371B568EC025ULL, "vxlt_gallowstemplar_n_5mj_0_0.vlm"},
	{0xBA5B5CAAA27CE8FDULL, "414641740.ogg"},
	{0xBA5B5CE2B43F562EULL, "vxw_bonepitdestroyed_a3_1x8_0_1.vlm"},
	{0xBA5BF310E9D8948BULL, "mcr370_templar16087634.cl"},
	{0xBA5C3BF82DF248ADULL, "fxa_stun_spk.mao"},
	{0xBA5C4CD53AFFBDF7ULL, "hlac_lt_hangedman_n_-1_21z.dds"},
	{0xBA5C6A824AC24335ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_enter.ani"},
	{0xBA5C75704605EDCCULL, "232673496.ogg"},
	{0xBA5CD473CB9D60A3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbaseb_sa.dds"},
	{0xBA5D2805F75FF44DULL, "dae000_bodhan_letter6144307.cl"},
	{0xBA5DC45D9D6116FDULL, "712900549.ogg"},
	{0xBA5DE7FA8CB43651ULL, "vxht_keep_n_36v_0_0.vlm"},
	{0xBA5DF930425B72F2ULL, "llac_w_encounter2_camp_262147_0.dds"},
	{0xBA5E2854AC39B71BULL, "6028816_m.fxe"},
	{0xBA5E9988E825086EULL, "6165318_m.fxe"},
	{0xBA5ED75BC09B6DD2ULL, "vxw_bonepitmining_a2_mc_0_0.vlm"},
	{0xBA5EF0E12F4D756FULL, "25382148.ogg"},
	{0xBA5EF9E7AA2CE213ULL, "weapon.mat.dx11"},
	{0xBA5FB39316C393C6ULL, "art\\characters\\playercharacter\\humanmale\\arm_medj.phy"},
	{0xBA600E9C0BE0BE62ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft3a.msh"},
	{0xBA60109C0BE0C118ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft3a.mmh"},
	{0xBA601C9C0BE0D5BBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft3a.mao"},
	{0xBA6086D8A968CA88ULL, "w_sundermount_141_0.rml"},
	{0xBA6099593CDA3DCAULL, "vxlt_0_lowtown_a2_n_bxr_0_0.vlm"},
	{0xBA60AC7A2B657280ULL, "6149585_m.fxe"},
	{0xBA6112CB0BC6570FULL, "mag360_alain6044663.cl"},
	{0xBA61857AEDCF5609ULL, "llac_dr_dungeon_3216_7.dds"},
	{0xBA61922F2D73F686ULL, "vrc321_varric6087071.cl"},
	{0xBA61C46198641571ULL, "6146803.cut"},
	{0xBA6215B581B9E93BULL, "hm_ulrichvon.mor"},
	{0xBA62D6A4BA18308DULL, "follower_banter6163293.cl"},
	{0xBA62E4911477C83EULL, "hlac_lt_gallowstemplar_f_2388_1y.dds"},
	{0xBA6400328ABF810DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medb_0n.dds"},
	{0xBA647E93E0D71FDDULL, "vxht_0_hightown_a3_bl3_0_0.vlm"},
	{0xBA64FDFB0EFCD9D1ULL, "dae300_alistair_king6134943.cl"},
	{0xBA6501BED5BE2646ULL, "6207557_m.fxe"},
	{0xBA65617E22C272F6ULL, "mag110_cut_abomination6037099.cl"},
	{0xBA65A5565CBDE796ULL, "dr_0_deeproads_17035.rml"},
	{0xBA65E7DDA8015AE5ULL, "180051680.ogg"},
	{0xBA66226DAF4F72E0ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_avea_0s.dds"},
	{0xBA6664C6A1C9AD5BULL, "36533458.ogg"},
	{0xBA667A62A4C094D7ULL, "702813958.ogg"},
	{0xBA67C8DD0BA51CC7ULL, "254911069.ogg"},
	{0xBA67EE3395226CC7ULL, "vxw_woundedcoast_1pw_0_0.vlm"},
	{0xBA6843F5EEEE8870ULL, "lla_ht_0_hightown_a3_-131070_0.dds"},
	{0xBA68D7052A54410AULL, "isa211_isabela6202067.cl"},
	{0xBA6957D82E624A65ULL, "vxht_0_hightown_c5l_0_0.vlm"},
	{0xBA6969375F3DE785ULL, "mag110_arianni6015773.cl"},
	{0xBA698032DFE2B629ULL, "6151372_m.fxe"},
	{0xBA69BE163A18C9ABULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock3b.phy"},
	{0xBA6A0A53FFD158D5ULL, "vxca_caverns2_9fd_0_0.vlm"},
	{0xBA6A179C0BE921A3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft3a.phy"},
	{0xBA6A489DFCCE9CFDULL, "zz_prmpt_debug6198263.cl"},
	{0xBA6AE19C62933073ULL, "zz_dre_debug6037562.cl"},
	{0xBA6B8F10684A8A59ULL, "zz_vault_debug6200356.cl"},
	{0xBA6BCDBE82FF0157ULL, "zz_dre_debug6056952.cl"},
	{0xBA6CA89E5BE9D868ULL, "781490371.ogg"},
	{0xBA6D2BB3779DF626ULL, "vxlt_darktown_4w9_0_0.vlm"},
	{0xBA6D4938B2918652ULL, "vxw_bonepitmining_a2_hl_0_0.vlm"},
	{0xBA6D9D59F4800ADEULL, "509632430.ogg"},
	{0xBA6ECEB19640E9B1ULL, "vxca_caverns1_mines_icx_0_1.vlm"},
	{0xBA6ED626E503502AULL, "vxht_playershouse_n_18p_0_0.vlm"},
	{0xBA6F28452D917586ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wi_basic_2.mao"},
	{0xBA6FB437DA3562BFULL, "gen00fl_varric6118959.cl"},
	{0xBA70DED369417FFAULL, "6174666_m.fxe"},
	{0xBA7100B3B88B4746ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_2.dds"},
	{0xBA713030C881829AULL, "849264612.ogg"},
	{0xBA734D30231F7239ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_25x_25_tool.phy"},
	{0xBA7356C359F9EEF0ULL, "vxw_bonepitdestroyed_a3_1ap_0_0.vlm"},
	{0xBA7378F64957772AULL, "zz_mcr_debug6072381.cl"},
	{0xBA7387E6E55CA731ULL, "6156027_m.fxe"},
	{0xBA73CCC2FE04EB69ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archright.phy"},
	{0xBA73D49BF96CEEDEULL, "6114527_m.fxe"},
	{0xBA73EFFA79130613ULL, "vxlt_gallowstemplar_n_35f_0_0.vlm"},
	{0xBA743EFB85C4DFEDULL, "llac_lt_gallowstemplar_n_2383_0.dds"},
	{0xBA7485F12D78886AULL, "vxca_caverns1_mines_k2s_0_0.vlm"},
	{0xBA74D3A6DC8CBB9FULL, "870766119.ogg"},
	{0xBA751FB4728FFE63ULL, "lt_0_lowtown_docks_133_0.msh"},
	{0xBA752DB4729015BDULL, "lt_0_lowtown_docks_133_0.mmh"},
	{0xBA7531B472901CCEULL, "lt_0_lowtown_docks_133_0.mao"},
	{0xBA755191A6FAFE0EULL, "360686515.ogg"},
	{0xBA75572A6403007CULL, "hlac_lt_0_lowtown_docks_-4_7x.dds"},
	{0xBA7584E7B7F59079ULL, "6165403_m.fxe"},
	{0xBA75B6F1E7096DEFULL, "ser221_gamlen6131967.cl"},
	{0xBA75C8970104214DULL, "vxw_bonepit_241_0_0.vlm"},
	{0xBA75D411FD69FFF9ULL, "6201582_f.fxe"},
	{0xBA7653383A322842ULL, "6208303_m.fxe"},
	{0xBA76BAFF33E76DA0ULL, "6137566_m.fxe"},
	{0xBA76C3AFF68784F9ULL, "follower_banter6170200.cl"},
	{0xBA7746AC977356E6ULL, "gen00fl_varric6102793.cl"},
	{0xBA78181B98F5689AULL, "vxlt_0_lowtown_a2_n_gby_0_0.vlm"},
	{0xBA78759CF50C60B3ULL, "art\\characters\\creatures\\stonegolem\\animation\\electr.evt"},
	{0xBA78A929573121FBULL, "204440430.ogg"},
	{0xBA79047A19A9FB57ULL, "vxdr_primevaltomb_533_0_0.vlm"},
	{0xBA7958594A8D9CB1ULL, "one350_variel6070998.cl"},
	{0xBA7A56B4128D3D8AULL, "6059004_m.fxe"},
	{0xBA7A7D7559B7AC06ULL, "lgt120cr_bhelen_supp_ass.utc"},
	{0xBA7ACCF791245D62ULL, "w_bonepitmining_a2_165_0.msh"},
	{0xBA7ACEF791246018ULL, "w_bonepitmining_a2_165_0.mmh"},
	{0xBA7ADAF7912474BBULL, "w_bonepitmining_a2_165_0.mao"},
	{0xBA7AE7A17CD78DB6ULL, "6166838_m.fxe"},
	{0xBA7B0150C4F3A7A6ULL, "6216986_m.fxe"},
	{0xBA7B04AEB42B6863ULL, "vxht_mansion2b_1cf_0_0.vlm"},
	{0xBA7B5872D2E114DCULL, "w_sundermount_57_0.phy"},
	{0xBA7B5EB8DD4A2688ULL, "6139986_m.fxe"},
	{0xBA7BE50B5B85B398ULL, "vxca_caverns2_7qk_0_1.vlm"},
	{0xBA7C0A3AE8F2F1B2ULL, "vxw_bonepitdestroyed_a3_9a_0_0.vlm"},
	{0xBA7C3680973FDDDEULL, "lowtown_fem_lowlife_mob6169180.cl"},
	{0xBA7C9360E6186DF2ULL, "follower_banter6163216.cl"},
	{0xBA7CC079D7696E5CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldraila.mao"},
	{0xBA7CCC79D76982FFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldraila.mmh"},
	{0xBA7CD279D7698CA1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldraila.msh"},
	{0xBA7D2BDD1B056062ULL, "6019022_m.fxe"},
	{0xBA7D47A572FA1898ULL, "6143091_m.fxe"},
	{0xBA7E1872BE7CEF65ULL, "art\\vfx\\combat\\vfx_c_rw_attack2_c.mmh"},
	{0xBA7E27317BD86A56ULL, "hlac_lt_0_lowtown_-131073_0y.dds"},
	{0xBA7E66CAC3DCDCEEULL, "w_sundermount_102_0.msh"},
	{0xBA7E74CAC3DCF43FULL, "w_sundermount_102_0.mao"},
	{0xBA7E80CAC3DD08DCULL, "w_sundermount_102_0.mmh"},
	{0xBA7ECE8C8758BEF4ULL, "vxca_caverns1_j3r_0_0.vlm"},
	{0xBA7F172232C14BD5ULL, "306167794.ogg"},
	{0xBA7F6E9CF51273F4ULL, "art\\characters\\creatures\\stonegolem\\animation\\electr.gad"},
	{0xBA7F7C530909B072ULL, "gen00fl_isabela6101073.cl"},
	{0xBA7FB13C0EE763ACULL, "6021688_f.fxe"},
	{0xBA7FBE1111AFB598ULL, "vxw_sundermount_23r_0_0.vlm"},
	{0xBA7FE086C824238EULL, "vxlt_0_lowtown_a2_e1q_0_0.vlm"},
	{0xBA7FF3F49D4906E9ULL, "vxca_caverns1_mines_ih0_0_0.vlm"},
	{0xBA80852F579D6391ULL, "6075969_f.fxe"},
	{0xBA808C4FC0591C2DULL, "6112157_m.fxe"},
	{0xBA8096E93F081E76ULL, "mag311_emile_de_launcet6043463.cl"},
	{0xBA80DE6668B6209DULL, "llac_lt_0_lowtown_a2_-3_2.dds"},
	{0xBA81624DC4C350DDULL, "vxht_0_hightown_a3_bf1_0_0.vlm"},
	{0xBA82078E3B6DBC42ULL, "vxlt_gallowscourtyard_n_7gg_0_0.vlm"},
	{0xBA826F72D2E75A04ULL, "w_sundermount_57_0.rml"},
	{0xBA82C6BB2D8D58D5ULL, "6169638_m.fxe"},
	{0xBA839E033AD8B4A6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_wep_sa.dds"},
	{0xBA84078AFAA75969ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appc_0s.dds"},
	{0xBA8415370536936DULL, "6182004_m.fxe"},
	{0xBA84659B243530ACULL, "864440330.ogg"},
	{0xBA85196D7BA43095ULL, "14426840.ogg"},
	{0xBA857C1B50DED42FULL, "hlac_ht_0_hightown_a2_65534_1x.dds"},
	{0xBA867D9CF518ABD2ULL, "art\\characters\\creatures\\stonegolem\\animation\\electr.ani"},
	{0xBA86B1A9581A76CEULL, "gen00fl_carver6103613.cl"},
	{0xBA86FE0653834EFCULL, "w_encounter2_63_0.phy"},
	{0xBA874C50B3ED17AFULL, "vxht_mansion_n_4fw_0_0.vlm"},
	{0xBA8760EDCCDF9034ULL, "2172185.ogg"},
	{0xBA87823295B851A1ULL, "6201633_m.fxe"},
	{0xBA87A452DBC42010ULL, "vxca_caverns1_mines_j2l_0_0.vlm"},
	{0xBA87CFD3FB5C81E5ULL, "vxlt_0_lowtown_docks_a3_cz2_0_1.vlm"},
	{0xBA890445E6E08644ULL, "lowtown_male_lowlife1_sta6164983.cl"},
	{0xBA89768BA62180E3ULL, "vxht_mansion2b_1fi_0_0.vlm"},
	{0xBA89A552CB3973E1ULL, "vxca_caverns1_mines_hm2_0_0.vlm"},
	{0xBA89DCA1E53003FFULL, "hm_lgt_captain02.mor"},
	{0xBA8AC968BEBCE805ULL, "6055111_m.fxe"},
	{0xBA8BC9E1E036DAF4ULL, "seb321_leliana6191275.cl"},
	{0xBA8C1572BE892732ULL, "art\\vfx\\combat\\vfx_c_rw_attack2_c.anb"},
	{0xBA8C8093D5F4D536ULL, "weapon_weapon_p1.vsh.fxo.dx11.assembly"},
	{0xBA8C8D765BE38DF1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2groupcrag.mmh"},
	{0xBA8C8EDB45F843EAULL, "413001416.ogg"},
	{0xBA8C99765BE3A192ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2groupcrag.mao"},
	{0xBA8CA7765BE3B9AFULL, "art\\levels\\races\\natural\\highlands\\props\\prp_2groupcrag.msh"},
	{0xBA8CD2675CF95DAEULL, "347906577.ogg"},
	{0xBA8CE1120F693436ULL, "gen_im_arm_glv_med_act3_01.uti"},
	{0xBA8D15065387E124ULL, "w_encounter2_63_0.rml"},
	{0xBA8D829322B8A4A1ULL, "6185085_m.fxe"},
	{0xBA8DE0B6624D1805ULL, "zz_vault_debug6203230.cl"},
	{0xBA8E2FEE03A5CE18ULL, "vxw_sundermount_24s_0_0.vlm"},
	{0xBA8E4F235630A48EULL, "dre000_cut_bartrandbetrayal6029358.cl"},
	{0xBA8F34248F773197ULL, "hlac_dr_0_deeproads_5842_4z.dds"},
	{0xBA8F90A68392F49BULL, "vxlt_0_lowtown_a2_f4p_0_0.vlm"},
	{0xBA8F931EE209CA6BULL, "dae000ip_crowd_lm5_talk2.utp"},
	{0xBA8FD6829E9A86C8ULL, "ht_mansion_n_3364.rml"},
	{0xBA8FF15C45E09F8EULL, "vxw_encounter3_ev_0_0.vlm"},
	{0xBA9041014FCB8E86ULL, "988184732.ogg"},
	{0xBA90454BE76221C8ULL, "w_bonepitmining_a2_129_0.mmh"},
	{0xBA90514BE762366BULL, "w_bonepitmining_a2_129_0.mao"},
	{0xBA905A3BEA064451ULL, "vxlt_warehouse_n_13m_0_0.vlm"},
	{0xBA90634BE76254F2ULL, "w_bonepitmining_a2_129_0.msh"},
	{0xBA91FFF68C09E923ULL, "872203511.ogg"},
	{0xBA9238BF63D7927CULL, "469745625.ogg"},
	{0xBA932E088C8E1FC8ULL, "vxca_caverns1_mines_hm7_0_0.vlm"},
	{0xBA9336BC361568AFULL, "339762665.ogg"},
	{0xBA93380A989132F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_in.gad"},
	{0xBA93C2B44A1537E5ULL, "863108490.ogg"},
	{0xBA93FC6DDD7B47F4ULL, "289890745.ogg"},
	{0xBA9436FCE7184E49ULL, "plt_genpt_app_varric.plo"},
	{0xBA944C9D68B7D778ULL, "156177935.ogg"},
	{0xBA9452F2ED6705D8ULL, "vxht_bartrand_n_5n6_0_0.vlm"},
	{0xBA94A7B9D3EDD2E5ULL, "6047488_m.fxe"},
	{0xBA94CA12DABFFA1CULL, "dae000_bodhan_letter6144281.cl"},
	{0xBA94F9975AA8BF9EULL, "6113975_m.fxe"},
	{0xBA95693B27E28CA2ULL, "qcr000_arishok6060118.cl"},
	{0xBA9577398DEF517BULL, "plt_zz_daept_debug.nss"},
	{0xBA9578F68D6040F0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_chantrybasic.msh"},
	{0xBA9586F68D60580AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_chantrybasic.mmh"},
	{0xBA958AF68D605F41ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff22_chantrybasic.mao"},
	{0xBA9636E8F75B4F9EULL, "6177835_f.fxe"},
	{0xBA967AE29C7AC2B7ULL, "zz_strider_stomp_v3.cut"},
	{0xBA96A602B51B493AULL, "395666210.ogg"},
	{0xBA970BB81EBF938CULL, "rdr111dg_martin6015011.cl"},
	{0xBA9720106701C6A3ULL, "zz_approval_debug6003691.cl"},
	{0xBA9794F689BD218DULL, "899714014.ogg"},
	{0xBA97E86898179963ULL, "vxht_mansion2b_1yh_0_0.vlm"},
	{0xBA97EB6DCFD5220EULL, "vxlt_gallowstemplar_n_47z_0_0.vlm"},
	{0xBA985B7955D578B0ULL, "vxlt_gallowstemplar_f_3jd_0_0.vlm"},
	{0xBA9867E0AE933B02ULL, "vxlt_gallowstemplar_f_5ux_0_1.vlm"},
	{0xBA987ECF72FBF1CCULL, "dre000_varric6208070.cl"},
	{0xBA98DEDFFCEF60A2ULL, "zz_mcr_debug6080779.cl"},
	{0xBA99490A9895BB47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_in.ani"},
	{0xBA99D488B0D7F458ULL, "gen00fl_carver6097411.cl"},
	{0xBA99E5986C37A8A5ULL, "hlac_ca_caverns1_mines_24426_1z.dds"},
	{0xBA9AACCC2BAEACC3ULL, "vxw_woundedcoast_14w_0_0.vlm"},
	{0xBA9B29212E0CF44DULL, "205476911.ogg"},
	{0xBA9B54A65409021FULL, "vxlt_darktown_4fc_0_0.vlm"},
	{0xBA9BA5DA39C10BDAULL, "vxw_encounter2_camp_qx_0_1.vlm"},
	{0xBA9BB0A89F8BCE1DULL, "vxlt_0_lowtown_docks_a3_n_cz5_0_1.vlm"},
	{0xBA9C80CAC3F6062FULL, "w_sundermount_102_0.rml"},
	{0xBA9CA521DD95B30EULL, "lgt110_urn_con_leader.cnv"},
	{0xBA9CDB5F9E256304ULL, "mer140_bonepit_comments6077842.cl"},
	{0xBA9D34914DA4745EULL, "6121888_m.fxe"},
	{0xBA9D416FE3C85430ULL, "llac_lt_0_lowtown_a2_n_c_131072_0.dds"},
	{0xBA9D65EE02533063ULL, "dae000_flemeth6001336.cl"},
	{0xBA9D9D25ADB5F3DEULL, "6125790_m.fxe"},
	{0xBA9DBDEF5E279C52ULL, "114122591.ogg"},
	{0xBA9DF5256C8D4A0BULL, "vxht_mansion2b_3gt_0_0.vlm"},
	{0xBA9E028375890D1BULL, "vxlt_0_lowtown_a2_f1s_0_0.vlm"},
	{0xBA9E339AFCA8C66AULL, "prm000im_air_ring.uti"},
	{0xBA9E3436C3AAA549ULL, "737139780.ogg"},
	{0xBA9E450229891F64ULL, "vxw_bonepitmining_a2_26d_0_0.vlm"},
	{0xBA9E47F0130900DCULL, "hlac_ca_caverns1_24424_16z.dds"},
	{0xBA9E66AB757B9311ULL, "vxht_mansion_n_612_0_0.vlm"},
	{0xBA9E71C339003D07ULL, "1063381449.ogg"},
	{0xBA9E8779F09A2D7DULL, "6067118_m.fxe"},
	{0xBA9E96BD277266A9ULL, "vrc221ar_mansion_fake.lst"},
	{0xBA9F2EC5326ED22FULL, "498732938.ogg"},
	{0xBA9FB5C2FE29FE8EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archright.mmh"},
	{0xBA9FC1C2FE2A1325ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archright.mao"},
	{0xBA9FC721DD98237EULL, "lgt110_urn_con_leader.bnk"},
	{0xBA9FC74B58DF3146ULL, "vxlt_0_lowtown_a2_n_c_ef5_0_0.vlm"},
	{0xBA9FCFC2FE2A2B7CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_archright.msh"},
	{0xBA9FF5D59BAE4A88ULL, "dre000_dougal6128883.cl"},
	{0xBAA03E25FE80F10EULL, "vxht_keep_6ab_0_0.vlm"},
	{0xBAA0AAE0DF6B5036ULL, "hlac_lt_ruinedhouse_n_-1_6x.dds"},
	{0xBAA0EE44E48A9D15ULL, "vxht_brothel_n_5an_0_0.vlm"},
	{0xBAA125C3C0C2ED2FULL, "dae100st_solivitus_plus.utm"},
	{0xBAA187D2B23980BCULL, "dae000_brat6156897.cl"},
	{0xBAA1AD15C55D3213ULL, "gen00fl_anders6112973.cl"},
	{0xBAA1CC8B53BB45EEULL, "6091883_f.fxe"},
	{0xBAA20C3B1958DAF3ULL, "666601750.ogg"},
	{0xBAA23C0B146613CFULL, "character_character_p121.psh.fxo"},
	{0xBAA265F5A41D7AA4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1k.mmh"},
	{0xBAA269F5A41D81F7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1k.mao"},
	{0xBAA27BF5A41DA046ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1k.msh"},
	{0xBAA35FCAC3FBECA7ULL, "w_sundermount_102_0.phy"},
	{0xBAA3A4769ACEFF3FULL, "714385975.ogg"},
	{0xBAA3C8721B4DEE55ULL, "vxht_0_hightown_aol_0_0.vlm"},
	{0xBAA3D8B6FF58DC59ULL, "plt_mag111pt_act_of_mercy.plo"},
	{0xBAA40F621D3DC070ULL, "nat380cr_ogre.utc"},
	{0xBAA419C2B3BA3DCBULL, "vxca_caverns1_mines_gkh_0_1.vlm"},
	{0xBAA444D636B9C45EULL, "6094519_m.fxe"},
	{0xBAA456913DAE0D7EULL, "mag311_keran6054380.cl"},
	{0xBAA54A7C1E0FAA8BULL, "vxlt_gallowstemplar_f_35f_0_0.vlm"},
	{0xBAA553DCC7C67400ULL, "104055955.ogg"},
	{0xBAA56EBD277841E0ULL, "vrc221ar_mansion_fake.ncs"},
	{0xBAA58DB64CCC30E2ULL, "vxht_mansion2b_34d_0_0.vlm"},
	{0xBAA6C73534D62CDCULL, "gen_im_trs_bks_good_breed.uti"},
	{0xBAA733F46A46DF5DULL, "dae000_meredith6085247.cl"},
	{0xBAA76F444A4E9BA8ULL, "follower_banter6136696.cl"},
	{0xBAA7E099D2DB0434ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_act_c.impa.ani"},
	{0xBAA7FCD8B34F5521ULL, "vxlt_0_lowtown_a2_n_c_gof_0_1.vlm"},
	{0xBAA8AA34BB4C8DF9ULL, "6153789_m.fxe"},
	{0xBAA8EA3DCC4F131FULL, "hla_ht_0_hightown_-131074_0z.dds"},
	{0xBAA8EC10D9D1B122ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f.gad"},
	{0xBAA8FEEF864F746BULL, "vxht_mansion_n_5as_0_0.vlm"},
	{0xBAA90598C6C9D850ULL, "qun221_viscount6070130.cl"},
	{0xBAAA02893CD04590ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeptwra.msh"},
	{0xBAAA10893CD05DAAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeptwra.mmh"},
	{0xBAAA14893CD064E1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeptwra.mao"},
	{0xBAAA5AD084E75746ULL, "6046836_f.fxe"},
	{0xBAAA75FA591C704CULL, "6083311_m.fxe"},
	{0xBAAA8E87FE8B68BCULL, "6061104_m.fxe"},
	{0xBAAAA0B95BD861C0ULL, "vxdr_0_deeproads_8wf_0_1.vlm"},
	{0xBAAB175C63A61240ULL, "mag311_emile_de_launcet6171513.cl"},
	{0xBAAC4D58DBDBFEB7ULL, "6106198_m.fxe"},
	{0xBAAD12D9376E2988ULL, "6125926_f.fxe"},
	{0xBAAD40ECD8EAB368ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.anb"},
	{0xBAADBDE72AB5DA9FULL, "ser191_emeric6108157.cl"},
	{0xBAADE11DFC602666ULL, "6077277_m.fxe"},
	{0xBAAE0349E993CEAAULL, "vxca_caverns1_mines_k2q_0_2.vlm"},
	{0xBAAEAB79EDE0ACE7ULL, "qun131_taskmaster6039747.cl"},
	{0xBAAEB002F077098EULL, "vxht_keep_6jc_0_0.vlm"},
	{0xBAAED3F948D74012ULL, "dae000cr_grand_cleric6148101.cl"},
	{0xBAAEDD99D2E11D96ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_act_c.impa.gad"},
	{0xBAAF0187CB2F01C9ULL, "hlac_ca_caverns1_24423_12x.dds"},
	{0xBAAF0F5DED37EAADULL, "523893500.ogg"},
	{0xBAAF450C7C72E223ULL, "336595756.ogg"},
	{0xBAAF5764DCC655CAULL, "709202909.ogg"},
	{0xBAAF6021D680B595ULL, "vxht_brothel_n_5dm_0_0.vlm"},
	{0xBAAF821DC8E328E9ULL, "vxlt_0_lowtown_a2_n_e82_0_0.vlm"},
	{0xBAAF8FBD2780CE4EULL, "vrc221ar_mansion_fake.are"},
	{0xBAAF9AB96ED3A43AULL, "follower_banter6169676.cl"},
	{0xBAAFA520DF29A4CEULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_4y.dds"},
	{0xBAB08B0AF29D67A9ULL, "6069729_m.fxe"},
	{0xBAB0C42AB13586B6ULL, "6167675_m.fxe"},
	{0xBAB11FE3431A543CULL, "vxlt_0_lowtown_a3_n_c_elj_0_2.vlm"},
	{0xBAB12635B360E1D8ULL, "6091823_f.fxe"},
	{0xBAB1A99519E44715ULL, "6066342_m.fxe"},
	{0xBAB1E43D4A58D0C7ULL, "vxlt_darktown_5d0_0_0.vlm"},
	{0xBAB1E6B5EFBBD824ULL, "618656138.ogg"},
	{0xBAB2C7E8EEAB42B6ULL, "vxlt_gallowstemplar_47e_0_0.vlm"},
	{0xBAB2F4CFAB3DAB6BULL, "2738667.ogg"},
	{0xBAB41F4652098BA5ULL, "dae220ip_at_player_home.utp"},
	{0xBAB453573485179CULL, "923212342.ogg"},
	{0xBAB4CC227C03CA63ULL, "vxht_mansion2b_1wj_0_0.vlm"},
	{0xBAB4EA25A3113AC5ULL, "33199831.ogg"},
	{0xBAB5759A1BC0FBB2ULL, "6096584_m.fxe"},
	{0xBAB5AFB6FDA7E79DULL, "6129291_f.fxe"},
	{0xBAB5EB3F321B55BAULL, "356249449.ogg"},
	{0xBAB5F899D2E76AF5ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_act_c.impa.evt"},
	{0xBAB6C96F78CE27DDULL, "hlac_dr_primevalentrance_1725_1y.dds"},
	{0xBAB6E38DAB01EB5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_loop.gad"},
	{0xBAB6EF5A7651BD0FULL, "6031964_m.fxe"},
	{0xBAB7B0978867E36FULL, "vxca_caverns2_7ks_0_0.vlm"},
	{0xBAB7DC05142478CEULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extstone.dds"},
	{0xBAB833D53017327BULL, "6029385_m.fxe"},
	{0xBAB8B565308F4C73ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_2y.dds"},
	{0xBAB94E4B237CB0A0ULL, "vxw_bonepitmining_a2_6p_0_0.vlm"},
	{0xBAB9533BA52B4033ULL, "vxlt_gallowstemplar_5cq_0_0.vlm"},
	{0xBAB9651E5A357CF5ULL, "84121764.ogg"},
	{0xBABA1DBBEBF12744ULL, "follower_banter6160866.cl"},
	{0xBABACD8DEF3AC041ULL, "6163385_f.fxe"},
	{0xBABBF7DC41167846ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian_0.pwk"},
	{0xBABC10E9CC40C31DULL, "6173615_m.fxe"},
	{0xBABCD637F0D951FEULL, "vxlt_0_lowtown_docks_a3_bsn_0_0.vlm"},
	{0xBABCDF5F574EBA5DULL, "6102552_m.fxe"},
	{0xBABCE28DAB06555AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_torightarm_loop.ani"},
	{0xBABCEE19FBF24EACULL, "883931747.ogg"},
	{0xBABD1710D9E2A308ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f.ani"},
	{0xBABDCCB80E1E98FAULL, "zz_dae_debug6050006.cl"},
	{0xBABDD637050F078CULL, "gen00fl_carver6122531.cl"},
	{0xBABE426EBE61722EULL, "6036968_m.fxe"},
	{0xBABF1DB732871006ULL, "6054036_m.fxe"},
	{0xBABF3976755CECB7ULL, "vxdr_primevaltomb_6pq_0_0.vlm"},
	{0xBABFE8408CA8EA8EULL, "129806873.ogg"},
	{0xBABFF4B184280C42ULL, "hlac_lt_undercityencounter_-1_31x.dds"},
	{0xBAC08A8A6E54DE08ULL, "967293297.ogg"},
	{0xBAC0945180CF3C8FULL, "893878398.ogg"},
	{0xBAC11295A67621D0ULL, "vxca_caverns2prc_8kr_0_0.vlm"},
	{0xBAC13F08C330D1F0ULL, "llac_ht_chantry_n_2534_31.dds"},
	{0xBAC18732D7971C10ULL, "mag291cr_wryme.ncs"},
	{0xBAC28C0CACD4D7AAULL, "882686509.ogg"},
	{0xBAC29078F8A7A75AULL, "gen_im_trs_tmp_med_glv_r0.uti"},
	{0xBAC33E0843C21433ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\patrolling.evt"},
	{0xBAC3BD00B34AA690ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad.mao"},
	{0xBAC3C100B34AADA3ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad.mmh"},
	{0xBAC3CF00B34AC5FDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad.msh"},
	{0xBAC3F7BC7964FAC1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkforward.ani"},
	{0xBAC41436324AE587ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_com_0s.dds"},
	{0xBAC426DC1B6FA0B7ULL, "fen310_fenris6118677.cl"},
	{0xBAC482398E16D414ULL, "plt_zz_daept_debug.plo"},
	{0xBAC4972F68DF4F30ULL, "76789623.ogg"},
	{0xBAC4C563E8909387ULL, "vxht_0_hightown_bg1_0_0.vlm"},
	{0xBAC5D3B6FF75BA6EULL, "plt_mag111pt_act_of_mercy.nss"},
	{0xBAC6494AD0E1D240ULL, "vxlt_0_lowtown_n_eqe_0_1.vlm"},
	{0xBAC65C1BD436EE9DULL, "water_wireframe.psh.fxo"},
	{0xBAC69BAB7D43B2A7ULL, "6154030_m.fxe"},
	{0xBAC6BDD88DDE8D36ULL, "6083942_m.fxe"},
	{0xBAC6D25658670ED8ULL, "vxht_mansion2b_1bz_0_0.vlm"},
	{0xBAC6F369F862D673ULL, "6163555_m.fxe"},
	{0xBAC71ABCE212ABC1ULL, "hlac_lt_0_lowtown_a3_n_c_131072_4z.dds"},
	{0xBAC7C525E96D11E2ULL, "vxlt_0_lowtown_docks_a3_n_cop_0_0.vlm"},
	{0xBAC8300410CA1246ULL, "vxlt_0_lowtown_a2_n_c_f4p_0_0.vlm"},
	{0xBAC890CB9DFDF3BDULL, "774496596.ogg"},
	{0xBAC8981BDB7D761CULL, "gencr_darkspawn_ogre.utc"},
	{0xBAC8F904419121EFULL, "hlac_lt_0_lowtown_docks_-3_5x.dds"},
	{0xBAC944A148FFE9B5ULL, "mer221st_sabin.stg"},
	{0xBAC952D8C0CFB4AFULL, "art\\characters\\weapons\\spears\\textures\\w_spr_002a_0s.dds"},
	{0xBAC95FED0FC1569DULL, "839202500.ogg"},
	{0xBACA370843C82774ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\patrolling.gad"},
	{0xBACB02BC796B2C63ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkforward.gad"},
	{0xBACB1160F5136721ULL, "65091095.ogg"},
	{0xBACB59D03896B1CFULL, "follower_banter6172868.cl"},
	{0xBACC1705DDBAD28CULL, "one110_meeran6089509.cl"},
	{0xBACD422595C96E6CULL, "vxca_caverns2_73e_0_1.vlm"},
	{0xBACD6295B17D2596ULL, "928711103.ogg"},
	{0xBACDAB5367530537ULL, "vxdr_primevaltomb_6wp_0_0.vlm"},
	{0xBACE355FDFBE307EULL, "art\\characters\\creatures\\corpse\\textures\\cn_bdy_cor_0n.dds"},
	{0xBACE9E181B0D0EC9ULL, "hlac_ca_cavern1_haunted_24423_2z.dds"},
	{0xBACEC7F72E4501C7ULL, "vxlt_darktown_5r2_0_0.vlm"},
	{0xBACF229E5AA86141ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_candleflame_glow.mmh"},
	{0xBACFC44DCD26C955ULL, "6207701_m.fxe"},
	{0xBAD005AA1523F14EULL, "938797582.ogg"},
	{0xBAD0928FB8E7E9D7ULL, "art\\characters\\weapons\\arrows\\w_arw_001.msh"},
	{0xBAD0A48FB8E8083AULL, "art\\characters\\weapons\\arrows\\w_arw_001.mao"},
	{0xBAD0A88FB8E80F49ULL, "art\\characters\\weapons\\arrows\\w_arw_001.mmh"},
	{0xBAD0AEA5DE38FAB3ULL, "6036503_m.fxe"},
	{0xBAD0B60D9AE15C61ULL, "vxht_0_hightown_a3_c8r_0_0.vlm"},
	{0xBAD0BDAB6AC7FB34ULL, "lt_undercityserialkiller_probe_1wa.dds"},
	{0xBAD1460843CE5F52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\patrolling.ani"},
	{0xBAD1DFBC79710F48ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkforward.evt"},
	{0xBAD1F6BCE3B75EE5ULL, "mag120_macha6080611.cl"},
	{0xBAD2291403F2C56EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_k_bk.ani"},
	{0xBAD23FF96C9DAA0AULL, "seb221_brett_harimann6177234.cl"},
	{0xBAD2B0C763291CEEULL, "6094982_m.fxe"},
	{0xBAD3EE1216A40C30ULL, "t3_dog_or1.tnt"},
	{0xBAD41E75E9C2DB58ULL, "hlac_lt_0_lowtown_n_-196608_1x.dds"},
	{0xBAD454865C31BDEBULL, "vxht_mansion_n_5pt_0_0.vlm"},
	{0xBAD4C200B35922C8ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\trap_gallowpad.phy"},
	{0xBAD4F546B0686ECCULL, "zz_qcr_debug6056846.cl"},
	{0xBAD51340D283720CULL, "w_bonepit_29_0_ll.mmh"},
	{0xBAD527AD71888C3AULL, "6179782_m.fxe"},
	{0xBAD54E4606C83F32ULL, "static_static_p45.psh.fxo.dx11.assembly"},
	{0xBAD57CFA078A524CULL, "lgt150_nat_were_supporter6110954.cl"},
	{0xBAD5F56845C49954ULL, "136203001.ogg"},
	{0xBAD6215627E1F58CULL, "vxdr_0_deeproads_cqr_0_0.vlm"},
	{0xBAD636F589177133ULL, "vxlt_gallowstemplar_5is_0_0.vlm"},
	{0xBAD63FECD90DB603ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.mmh"},
	{0xBAD64C0E4E455F91ULL, "vxht_keep_n_6ha_0_0.vlm"},
	{0xBAD6B26BBFDF8906ULL, "vxlt_0_lowtown_docks_a3_bup_0_0.vlm"},
	{0xBAD6F1ACCC592A54ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basic.phy"},
	{0xBAD739CC6B882DD1ULL, "w_bonepitmining_a2_111_0.mmh"},
	{0xBAD745CC6B884272ULL, "w_bonepitmining_a2_111_0.mao"},
	{0xBAD753CC6B88598FULL, "w_bonepitmining_a2_111_0.msh"},
	{0xBAD763DAFE93A779ULL, "vxlt_0_lowtown_a2_hdl_0_0.vlm"},
	{0xBAD8040B8B507CE1ULL, "gencr_rubble_golem_b.utc"},
	{0xBAD88C9ACF0DC217ULL, "w_woundedcoast_253_0.msh"},
	{0xBAD89E9ACF0DE17AULL, "w_woundedcoast_253_0.mao"},
	{0xBAD8A29ACF0DE789ULL, "w_woundedcoast_253_0.mmh"},
	{0xBAD90E94E3D81CFFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk04.phy"},
	{0xBAD955966BD1E767ULL, "vxht_keep_n_69p_0_0.vlm"},
	{0xBAD9ABF00DD12627ULL, "vxht_mansion2b_2nr_0_0.vlm"},
	{0xBADA3EC35C360848ULL, "vxht_0_hightown_a3_c8w_0_0.vlm"},
	{0xBADA836F2E4FABEDULL, "dae000ip_crowd_ff1_talk3.utp"},
	{0xBADA894C4BAF111CULL, "635253725.ogg"},
	{0xBADB49C087FE0505ULL, "dae000_arishok6081341.cl"},
	{0xBADB878422AF19AFULL, "hlac_ca_caverns2_5802_32y.dds"},
	{0xBADBC586A84D55D4ULL, "vxlt_0_lowtown_a3_n_c_fhg_0_0.vlm"},
	{0xBADC19B9E9474346ULL, "6207746_m.fxe"},
	{0xBADC67F037FED5A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_loop.ani"},
	{0xBADD94D2923B224DULL, "vxlt_0_lowtown_a3_n_c_i0u_0_0.vlm"},
	{0xBADDA57EC6A21F09ULL, "lgt260cr_eau_official.utc"},
	{0xBADF025E26866A34ULL, "vxdr_primevaltomb_4h4_0_0.vlm"},
	{0xBADF211403FD41BFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_k_bk.evt"},
	{0xBADF3AE6D51C7166ULL, "838387861.ogg"},
	{0xBADFB7DA048DA811ULL, "vxlt_warehouse_9u_0_0.vlm"},
	{0xBADFD579D7BE0F44ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scafoldraila.phy"},
	{0xBAE01FB7D56CE4B0ULL, "vxca_caverns2_93j_0_0.vlm"},
	{0xBAE0376DDA635F1DULL, "256197957.ogg"},
	{0xBAE0A82864E6870CULL, "gen00fl_bethany6097840.cl"},
	{0xBAE1080ED438E950ULL, "6202804_m.fxe"},
	{0xBAE115B7085BB149ULL, "vxlt_0_lowtown_e2k_0_0.vlm"},
	{0xBAE1A9366B8EE43AULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.dura.gad"},
	{0xBAE1B27378D89107ULL, "vxca_caverns2_8r7_0_0.vlm"},
	{0xBAE1B77E51891E95ULL, "hm_cgn_007.mop"},
	{0xBAE1CEACCC62F651ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basic.msh"},
	{0xBAE1DCACCC630E6CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basic.mao"},
	{0xBAE1E8ACCC63220FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3wi_basic.mmh"},
	{0xBAE1E9CA5775F99DULL, "hlac_dr_dungeon_3217_6x.dds"},
	{0xBAE21B85288E4A06ULL, "mag291_demon_marethari6027844.cl"},
	{0xBAE25D16AA0D9511ULL, "vxlt_0_lowtown_a2_emn_0_0.vlm"},
	{0xBAE26D1CA596F3FAULL, "one300_zevran6076931.cl"},
	{0xBAE2F8020E5E388FULL, "lgt311_shw_mer_mettin6174283.cl"},
	{0xBAE34DBCD957F64DULL, "313689285.ogg"},
	{0xBAE35D1C9FE79175ULL, "zz_ave_debug6134019.cl"},
	{0xBAE3B304FE9DA004ULL, "vxht_0_hightown_c33_0_0.vlm"},
	{0xBAE3B6103C533FD8ULL, "vxht_mansion2b_1xx_0_0.vlm"},
	{0xBAE41AE489280EF5ULL, "6071953_m.fxe"},
	{0xBAE4A8D27B0D89B3ULL, "vxlt_gallowstemplar_5rr_0_0.vlm"},
	{0xBAE4A8DFCD5942E2ULL, "vxlt_0_lowtown_docks_a3_n_cur_0_0.vlm"},
	{0xBAE53B44BC228404ULL, "vxdr_0_deeproads_bpq_0_1.vlm"},
	{0xBAE574F0FB37F410ULL, "vxlt_0_lowtown_a2_n_ef5_0_0.vlm"},
	{0xBAE59DAD91D604F6ULL, "671870829.ogg"},
	{0xBAE5B7F29058D142ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_4z.dds"},
	{0xBAE6021404032BA8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_k_bk.gad"},
	{0xBAE628586A624D5DULL, "vxlt_hangedman_327_0_0.vlm"},
	{0xBAE630B3652A01E3ULL, "876790923.ogg"},
	{0xBAE6F58F829F168DULL, "787267310.ogg"},
	{0xBAE6FD7BD8A2DA5DULL, "one200_gold_followers6101138.cl"},
	{0xBAE71671C3793290ULL, "37887184.ogg"},
	{0xBAE75C014237207DULL, "hlac_ht_0_hightown_0_0x.dds"},
	{0xBAE8709C81FDE62BULL, "420651607.ogg"},
	{0xBAE8817D8636F4FFULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_foundryfire.mmh"},
	{0xBAE8A55E026AC6BFULL, "6152775_m.fxe"},
	{0xBAE8AC366B950851ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.dura.evt"},
	{0xBAE92425696DBBE6ULL, "6153724_m.fxe"},
	{0xBAE94A42F1293FCAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_wave_goodbye_simple.gad"},
	{0xBAE995DC92E6CE30ULL, "77462934.ogg"},
	{0xBAE9971F3B09B3E3ULL, "6148660_m.fxe"},
	{0xBAE9C40707B6E826ULL, "6180621_m.fxe"},
	{0xBAE9DD2BF0771BE4ULL, "art\\levels\\races\\darkspawn\\props\\_textures\\prp_ogreskull_d.dds"},
	{0xBAE9EDC3A8BC7D68ULL, "6064536_m.fxe"},
	{0xBAEA0F94E3E69104ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk04.mmh"},
	{0xBAEA199F00C4E360ULL, "dre000_varric6037001.cl"},
	{0xBAEA2594E3E6B6A6ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk04.msh"},
	{0xBAEAAEC9307663B7ULL, "hlac_ht_mansion2_818_10x.dds"},
	{0xBAEAD5CC9270382EULL, "vxht_chantry_n_2iu_0_0.vlm"},
	{0xBAEAE7C7059DBEACULL, "zz_followers_banter6159198.cl"},
	{0xBAEB0300CF0C686DULL, "llac_lt_0_lowtown_-196609_4.dds"},
	{0xBAEB4B21EACA3658ULL, "vxlt_0_lowtown_a2_n_gvo_0_0.vlm"},
	{0xBAEB55A9B654EA0DULL, "6124726_f.fxe"},
	{0xBAEB5F6EFE545B97ULL, "6129553_m.fxe"},
	{0xBAEB776885997441ULL, "mag200_feynriel6030422.cl"},
	{0xBAEBF5618D021486ULL, "art\\characters\\creatures\\mubari\\animation\\p_dmg.evt"},
	{0xBAEBF70804F36154ULL, "vxca_caverns1_i71_0_0.vlm"},
	{0xBAEBFCAA9DB73943ULL, "48223040.ogg"},
	{0xBAEC66B5AC9A2BA2ULL, "vxht_chantry_289_0_0.vlm"},
	{0xBAEC97B58D5C4958ULL, "6174880_m.fxe"},
	{0xBAECAB23A3839857ULL, "fighting70b.mao"},
	{0xBAECF6EE3BF011FFULL, "vxdr_0_deeproads_9ca_0_0.vlm"},
	{0xBAED53174D8A0A36ULL, "6108320_m.fxe"},
	{0xBAED9596297A7039ULL, "vxw_bonepitmining_a2_kv_0_0.vlm"},
	{0xBAED99C77ECD8D61ULL, "vxht_0_hightown_a3_c6p_0_0.vlm"},
	{0xBAEDD963F4078C4DULL, "lt_0_lowtown_docks_n_235_0.msh"},
	{0xBAEDE6EB865EA6AFULL, "hlac_lt_0_lowtown_docks_n_-196608_1x.dds"},
	{0xBAEDE763F407A380ULL, "lt_0_lowtown_docks_n_235_0.mao"},
	{0xBAEDEB63F407AAD3ULL, "lt_0_lowtown_docks_n_235_0.mmh"},
	{0xBAEE02B10DA4A091ULL, "150905050.ogg"},
	{0xBAEE22A706DBE3F0ULL, "6118125_m.fxe"},
	{0xBAEE4A5D21D6D93BULL, "static_static_p6.psh.fxo.dx11"},
	{0xBAEE6F604E07B493ULL, "follower_banter6141732.cl"},
	{0xBAEEA9908B566D59ULL, "zz_vault_debug6203371.cl"},
	{0xBAEF79EBC62F01D0ULL, "6209904_m.fxe"},
	{0xBAF01AFABE72DD07ULL, "vxht_0_hightown_bv2_0_0.vlm"},
	{0xBAF02D42F12F2D61ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_wave_goodbye_simple.evt"},
	{0xBAF050F0380F57A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_loop.gad"},
	{0xBAF0678D9E05A9FBULL, "str_06_f.mao"},
	{0xBAF0B83A631699EEULL, "vxlt_0_lowtown_a2_n_c_ekl_0_2.vlm"},
	{0xBAF0F328DF10B080ULL, "vxca_caverns2_987_0_0.vlm"},
	{0xBAF171B2190A0442ULL, "vxlt_warehouse_ch_0_0.vlm"},
	{0xBAF1EA3D84B7AE55ULL, "llac_w_sundermount_131070_0.dds"},
	{0xBAF20F5C11AC4B1DULL, "vxw_blightlands_b0_0_0.vlm"},
	{0xBAF224E1F093B884ULL, "vxht_0_hightown_c60_0_0.vlm"},
	{0xBAF227ED2E495858ULL, "vxht_mansion2b_1sy_0_0.vlm"},
	{0xBAF23932A199F9B1ULL, "vxht_0_hightown_a2_bpr_0_0.vlm"},
	{0xBAF2CE7969D3677AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d2_v1.gad"},
	{0xBAF30C618D085AD1ULL, "art\\characters\\creatures\\mubari\\animation\\p_dmg.gad"},
	{0xBAF375FA135985BFULL, "fog.gda"},
	{0xBAF4FE8A02CE7949ULL, "392142323.ogg"},
	{0xBAF5A9B813C9B213ULL, "mrl000_marethari6092377.cl"},
	{0xBAF5B272236D3E09ULL, "fxe_day_shadow.mao"},
	{0xBAF5B9653DB73646ULL, "llac_ca_caverns2_5802_41.dds"},
	{0xBAF5C4366B9FBB20ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.dura.ani"},
	{0xBAF5D2D7AD7D63D4ULL, "6116710_m.fxe"},
	{0xBAF5F6DD4DBD3828ULL, "vxlt_0_lowtown_f4s_0_0.vlm"},
	{0xBAF6DAB705BC12ABULL, "vxht_mansion2b_2wl_0_1.vlm"},
	{0xBAF7098869EFABA3ULL, "vxlt_undercitydungeon_xv_0_0.vlm"},
	{0xBAF70DB8E667303AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic_1.mao"},
	{0xBAF7227D40223948ULL, "vxht_0_hightown_a3_c6u_0_0.vlm"},
	{0xBAF79DE57D2CB84FULL, "dae000_bartrand6206952.cl"},
	{0xBAF7E009CFD2719EULL, "vxlt_0_lowtown_n_e0m_0_0.vlm"},
	{0xBAF806258D7FCED7ULL, "gen00fl_varric6093647.cl"},
	{0xBAF86161CC9529A2ULL, "vxlt_gallowscourtyard_n_6zu_0_0.vlm"},
	{0xBAF8D17969D7D791ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d2_v1.evt"},
	{0xBAF90D618D0CC87FULL, "art\\characters\\creatures\\mubari\\animation\\p_dmg.ani"},
	{0xBAF947A9846650AEULL, "vxht_chantry_n_2dv_0_0.vlm"},
	{0xBAF9A58FB90B0F7EULL, "art\\characters\\weapons\\arrows\\w_arw_001.anb"},
	{0xBAFA0A42DA5E9777ULL, "dae000cr_grand_cleric6147657.cl"},
	{0xBAFA16B46BA1CD9BULL, "zz_qcr_debug6052120.cl"},
	{0xBAFA8DDD63B31600ULL, "6121125_m.fxe"},
	{0xBAFAB565DEC079E3ULL, "6202942_m.fxe"},
	{0xBAFB52D56C3A97FFULL, "gen00fl_anders6103922.cl"},
	{0xBAFBC1603EBD1D3DULL, "dre000_varric6027775.cl"},
	{0xBAFC416F4343D6CAULL, "6149763_m.fxe"},
	{0xBAFC4ACA664335F8ULL, "6075350_m.fxe"},
	{0xBAFC9483F8D1FC70ULL, "6117025_m.fxe"},
	{0xBAFCAFB511199044ULL, "159341838.ogg"},
	{0xBAFCB47754E6918DULL, "6172906_m.fxe"},
	{0xBAFDB37D896F544AULL, "gen00st_anders_romance.stg"},
	{0xBAFDB7C63B53163BULL, "zz_one_debug6051497.cl"},
	{0xBAFDD5F9DB25718CULL, "vxht_chantry_2cr_0_0.vlm"},
	{0xBAFDE57ACA62164CULL, "fxc_dirtysm.mao"},
	{0xBAFDF862FF992682ULL, "6018133_m.fxe"},
	{0xBAFE2542F13B5CB0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_wave_goodbye_simple.ani"},
	{0xBAFE648EAE192F8BULL, "6076055_m.fxe"},
	{0xBAFE850D64329B99ULL, "hlac_lt_0_lowtown_docks_a3_n_131070_3x.dds"},
	{0xBAFECDFC9BA58419ULL, "369800113.ogg"},
	{0xBAFF832E91BB7C8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_wave_goodbye_short.evt"},
	{0xBAFFE38F0B001CC2ULL, "vxlt_warehouse_di_0_0.vlm"},
	{0xBB008AC5D9214BD4ULL, "hlac_ca_caverns2prc_5802_20x.dds"},
	{0xBB0099CA203F70D8ULL, "vxht_mansion2b_1vv_0_0.vlm"},
	{0xBB00AB0F93901231ULL, "vxht_0_hightown_a2_buq_0_0.vlm"},
	{0xBB02A04AA7B82FEDULL, "340768733.ogg"},
	{0xBB02B8855287A90AULL, "6018574_f.fxe"},
	{0xBB03492769C8EAC3ULL, "w_blightlands_81_0.mao"},
	{0xBB034D2769C8F110ULL, "w_blightlands_81_0.mmh"},
	{0xBB0358C0440BEE6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_loop.gad"},
	{0xBB035B2769C9092AULL, "w_blightlands_81_0.msh"},
	{0xBB03AAE0BFCD6AE4ULL, "art\\vfx\\combat\\vfx_m_amg_conj_c.anb"},
	{0xBB03DCF2AA757104ULL, "6061811_f.fxe"},
	{0xBB03F5348481B126ULL, "459518091.ogg"},
	{0xBB047932F3E41116ULL, "684208147.ogg"},
	{0xBB0556966142EC35ULL, "fxa_deacti_ele.mao"},
	{0xBB06315C5F186BD9ULL, "mag110_arianni6049707.cl"},
	{0xBB064678D2C42477ULL, "zz_followers_banter6159375.cl"},
	{0xBB0657273E9FBC3DULL, "hf_cgn_008.mop"},
	{0xBB067A2E91C18C39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_wave_goodbye_short.gad"},
	{0xBB06E97969E43E60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_attk_d2_v1.ani"},
	{0xBB0785EC2CB35511ULL, "one350_variel6085396.cl"},
	{0xBB0859E50DFCB20BULL, "zz_isa_debug6175190.cl"},
	{0xBB08EA854F59A58DULL, "gen00fl_merrill6128679.cl"},
	{0xBB097D582E49D2A5ULL, "610275316.ogg"},
	{0xBB098485D3307C45ULL, "560500690.ogg"},
	{0xBB09B3483824F44CULL, "vxlt_0_lowtown_n_h8b_0_0.vlm"},
	{0xBB0A46F72AA55AB6ULL, "vxdr_primevaltomb_1mt_0_0.vlm"},
	{0xBB0AB63EDBEDA585ULL, "6217111_m.fxe"},
	{0xBB0AF63CE28BE6B9ULL, "gen_im_trs_krk_hgh_junk2_04.uti"},
	{0xBB0B106C28EE7845ULL, "hlac_lt_gallowstemplar_f_2394_4x.dds"},
	{0xBB0B16E837DCBA90ULL, "llac_lt_0_lowtown_-196608_4.dds"},
	{0xBB0B28774BEF59D9ULL, "185172252.ogg"},
	{0xBB0B6BF8FEF361B0ULL, "vxht_0_hightown_a7u_0_0.vlm"},
	{0xBB0BCC7CC750D356ULL, "132215967.ogg"},
	{0xBB0BD0DCC7A285C8ULL, "6085466_f.fxe"},
	{0xBB0C1DC0060B31A7ULL, "6108237_m.fxe"},
	{0xBB0C202CEC071A41ULL, "hlac_w_encounter2_196608_0y.dds"},
	{0xBB0C8600E2631107ULL, "6160693_m.fxe"},
	{0xBB0CAFF2F7B30917ULL, "6085109_m.fxe"},
	{0xBB0D2AA976F59F66ULL, "6039753_ketojan_obeys.cut"},
	{0xBB0D6542454A8A45ULL, "vxht_bartrand_n_5y6_0_0.vlm"},
	{0xBB0D8B2E91C7C787ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_wave_goodbye_short.ani"},
	{0xBB0DB237A6DF7EFBULL, "295193743.ogg"},
	{0xBB0DE680AA11DE5EULL, "hlac_ht_0_hightown_a2_65537_2y.dds"},
	{0xBB0E39AB6679F4B0ULL, "hlac_dr_0_deeproads_5841_0x.dds"},
	{0xBB0E92A993B96A10ULL, "229533930.ogg"},
	{0xBB0E9D9D9016B65DULL, "934358171.ogg"},
	{0xBB0ECE9DE595F9C4ULL, "vxlt_0_lowtown_a2_n_gpi_0_0.vlm"},
	{0xBB0EF3A68E6E6273ULL, "315951986.ogg"},
	{0xBB0F28F8C9EF3B00ULL, "hlac_lt_0_lowtown_1_0x.dds"},
	{0xBB0F7BDFF1D8BD74ULL, "art\\vfx\\environment\\vfx_prp_partycall_c.anb"},
	{0xBB0F9CD6F1C99E79ULL, "372139370.ogg"},
	{0xBB0FBBB27D87177CULL, "vxlt_gallowstemplar_f_2or_0_0.vlm"},
	{0xBB0FDB2F1129DAEFULL, "vxca_cavern1_haunted_ea1_0_0.vlm"},
	{0xBB100B4E07213FA5ULL, "6122504_f.fxe"},
	{0xBB105B8DB569FEADULL, "vxdr_0_deeproads_f4u_0_0.vlm"},
	{0xBB107E32D7DAE10AULL, "mag291cr_wryme.utc"},
	{0xBB10C9E0BFD829E7ULL, "art\\vfx\\combat\\vfx_m_amg_conj_c.mmh"},
	{0xBB1146FD46D74C2EULL, "mrl000_marethari6068793.cl"},
	{0xBB122A771A175493ULL, "665556499.ogg"},
	{0xBB12952D637FEC04ULL, "vxlt_0_lowtown_docks_aqn_0_0.vlm"},
	{0xBB1298773CCE3DCEULL, "6156543_m.fxe"},
	{0xBB133E174F00E0A2ULL, "hlac_lt_0_lowtown_a3_n_c_65536_8z.dds"},
	{0xBB13B6F0364E4DC0ULL, "487488595.ogg"},
	{0xBB14013EC2FCBBD1ULL, "hlac_lt_blackemporium_1055_8y.dds"},
	{0xBB143352AEF9ECA3ULL, "qun120_arishok6049630.cl"},
	{0xBB1466207F79C902ULL, "6115964_m.fxe"},
	{0xBB15448EECAB92C7ULL, "vxw_woundedcoast_14k_0_0.vlm"},
	{0xBB15A3C84B61E130ULL, "6136730_m.fxe"},
	{0xBB15C55898581E5BULL, "137951180.ogg"},
	{0xBB1642A250320F69ULL, "vxlt_warehouse_ae_0_0.vlm"},
	{0xBB16957789DF7E5EULL, "llac_lt_0_lowtown_a3_n_c_2_1.dds"},
	{0xBB16B887C091F353ULL, "615063750.ogg"},
	{0xBB17F24EBF68F8D7ULL, "6054833_m.fxe"},
	{0xBB1833C0441E0B50ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_forward_throne_loop.ani"},
	{0xBB183C2C72ABD9DEULL, "4925978.ogg"},
	{0xBB1855839195B64BULL, "572436619.ogg"},
	{0xBB186F011DCEC51BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjut.phy"},
	{0xBB1891823C155066ULL, "wa_sp.mao"},
	{0xBB18B081473DD353ULL, "hlac_ht_chantry_n_2534_44y.dds"},
	{0xBB1905A5B7EBF8D6ULL, "hlac_ht_mansion2b_818_55x.dds"},
	{0xBB1923918AD6F4C8ULL, "6207744_m.fxe"},
	{0xBB1937937CFDFEDCULL, "483154755.ogg"},
	{0xBB19C33754BF977AULL, "dae000ip_crowd_ff7_talk2.utp"},
	{0xBB1A61E3D6362F27ULL, "art\\vfx\\materials\\vfx_web_fb_blend.mao"},
	{0xBB1A816347C14000ULL, "vxlt_0_lowtown_a3_n_c_def_0_0.vlm"},
	{0xBB1AB31B12E063F6ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf04.phy"},
	{0xBB1AB9B3BF11A28CULL, "vxht_chantry_2yp_0_0.vlm"},
	{0xBB1AE0560FA9A681ULL, "ico_armor_heavy_5.dds"},
	{0xBB1B0DFD1648F597ULL, "6162444_m.fxe"},
	{0xBB1B49E3749C9412ULL, "hlac_lt_0_lowtown_n_131070_5y.dds"},
	{0xBB1B6EB6F075F8E0ULL, "vxlt_hangedman_n_20v_0_0.vlm"},
	{0xBB1B7688716A5711ULL, "6169469_m.fxe"},
	{0xBB1BE06189150E20ULL, "hlac_lt_0_lowtown_65535_2y.dds"},
	{0xBB1C39F44D872FF5ULL, "vxdr_primevalentrance_21x_0_0.vlm"},
	{0xBB1CC748EEEC4DC2ULL, "vxlt_warehouse_rk_0_0.vlm"},
	{0xBB1D407AD78C1244ULL, "vxlt_0_lowtown_a2_n_gmn_0_0.vlm"},
	{0xBB1D5ADFF1E4C237ULL, "art\\vfx\\environment\\vfx_prp_partycall_c.mmh"},
	{0xBB1D8EC9777C4331ULL, "vxht_0_hightown_a2_bcw_0_0.vlm"},
	{0xBB1D93697C334DBDULL, "vxca_caverns2_8on_0_0.vlm"},
	{0xBB1E8E052964E795ULL, "581589706.ogg"},
	{0xBB1E97BD24F17FE2ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_nud_0s.dds"},
	{0xBB1F858610530BB3ULL, "seb221at_to_ruin.utp"},
	{0xBB1FAE067717BA3BULL, "dae011_badpoet_sta.bnk"},
	{0xBB1FFEAA52E09554ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.dura.ani"},
	{0xBB202423A50F6A57ULL, "art\\levels\\ztools\\default\\generic\\3oi_basic.phy"},
	{0xBB20C5EEA56C2084ULL, "6182452_m.fxe"},
	{0xBB20CDFF187A1123ULL, "194196337.ogg"},
	{0xBB211CC41FEB3421ULL, "follower_banter6136462.cl"},
	{0xBB218F59E2A11522ULL, "lt_gallowstemplar_2383.rml"},
	{0xBB219612F2D3F6ADULL, "qun221tr_petrice_initiates.utt"},
	{0xBB2196185A174412ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\bsh_wideleafy_dp.dds"},
	{0xBB22789288A5864EULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em04.phy"},
	{0xBB22983A5C7CF40BULL, "273484278.ogg"},
	{0xBB22B0067719F3FBULL, "dae011_badpoet_sta.cnv"},
	{0xBB22DEA00F3DC869ULL, "vxca_caverns1_j21_0_1.vlm"},
	{0xBB23ACC76A39BAD4ULL, "6073891_m.fxe"},
	{0xBB23B2334B3AD2E5ULL, "6147713_m.fxe"},
	{0xBB243D3057033B12ULL, "440080385.ogg"},
	{0xBB248D0B97B15AD1ULL, "73779123.ogg"},
	{0xBB2490EC06BFCEE0ULL, "vxlt_0_lowtown_a2_n_c_hpd_0_0.vlm"},
	{0xBB249B5F04E8DA28ULL, "art\\cinematics\\props\\bookinquisb.msh"},
	{0xBB249C1B12E89097ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf04.msh"},
	{0xBB24B21B12E8B609ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf04.mmh"},
	{0xBB24B95F04E90D52ULL, "art\\cinematics\\props\\bookinquisb.mmh"},
	{0xBB25474E52F63CC0ULL, "vxht_0_hightown_n_c9n_0_0.vlm"},
	{0xBB25943889C7AC93ULL, "635097971.ogg"},
	{0xBB25A029A99CFA8AULL, "vxht_0_hightown_n_90e_0_0.vlm"},
	{0xBB25D46AAAA88378ULL, "isa321_velasco6091870.cl"},
	{0xBB25E34984A26DC2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45win.phy"},
	{0xBB265249203E4F75ULL, "vxlt_hangedman_10g_0_0.vlm"},
	{0xBB2696BAB24905B4ULL, "art\\characters\\creatures\\hurlock\\textures\\w_lsw_hlk_0n.dds"},
	{0xBB26FBAA52E6AEB6ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.dura.gad"},
	{0xBB26FEE614AFA307ULL, "vxca_caverns2_56z_0_0.vlm"},
	{0xBB270BAFF197AAAEULL, "zz_one_debug6100024.cl"},
	{0xBB278343754B693FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsrock.phy"},
	{0xBB27875F08DEF363ULL, "qun221_petrice6083714.cl"},
	{0xBB27DF3614FF4978ULL, "6076828_f.fxe"},
	{0xBB280E6B7700943BULL, "lgt210cr_shp_kill_qunari1.utc"},
	{0xBB281A230ADDC736ULL, "vxlt_0_lowtown_n_e12_0_0.vlm"},
	{0xBB288241B55A8B82ULL, "plt_mag110pt_wayward_son.nss"},
	{0xBB29632CE767CF66ULL, "vxlt_0_lowtown_n_exm_0_0.vlm"},
	{0xBB29CEA0DB8E8D01ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.cess.ani"},
	{0xBB29E3E6BFC2A8E0ULL, "fxp_pillar_sym01.dds"},
	{0xBB29FE2F7453CBFAULL, "lt_0_lowtown_20_0.anb"},
	{0xBB2A65399910FA3DULL, "240911244.ogg"},
	{0xBB2A7C81E37A6FE1ULL, "1462659.ogg"},
	{0xBB2A95B8DBDD1F64ULL, "45018666.ogg"},
	{0xBB2B15A7A1B118DBULL, "6115864_m.fxe"},
	{0xBB2B2AC277F33024ULL, "mcr370_meredith6136175.cl"},
	{0xBB2B31BB8F3B8473ULL, "vxlt_undercitydungeon_11x_0_0.vlm"},
	{0xBB2B3925E0E26642ULL, "vxlt_warehouse_wt_0_0.vlm"},
	{0xBB2B611FB74BCA4DULL, "vxlt_0_lowtown_docks_a3_d3s_0_0.vlm"},
	{0xBB2C00A669725BB1ULL, "vxht_0_hightown_a2_bdv_0_0.vlm"},
	{0xBB2C0929AAFB62CFULL, "vxlt_gallowstemplar_3s7_0_0.vlm"},
	{0xBB2D16AA52EB4815ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.dura.evt"},
	{0xBB2D7899BA47A992ULL, "6153830_m.fxe"},
	{0xBB2D90A03F033F85ULL, "gen00fl_varric6191297.cl"},
	{0xBB2DF757758D2521ULL, "plt_cod_itm_absolution.plo"},
	{0xBB2E04BAF68CEF30ULL, "6101196_m.fxe"},
	{0xBB2E4F6189C1ABD7ULL, "mag110_arianni6016789.cl"},
	{0xBB2EF79E8824B17DULL, "dae200_saemus6142717.cl"},
	{0xBB2EFA15859E6EC4ULL, "demoral_03.mao"},
	{0xBB2F33537EB47997ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06b.phy"},
	{0xBB2F4FD790C4ACEEULL, "6072691_f.fxe"},
	{0xBB2FD9A0DB930BA3ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.cess.gad"},
	{0xBB2FFDEC13255E25ULL, "vxlt_hangedman_n_4k_0_0.vlm"},
	{0xBB306B03E1CF2C7DULL, "zz_one_debug6074003.cl"},
	{0xBB30DD3BAB5DA7AAULL, "6102159_m.fxe"},
	{0xBB31307D4863B1E3ULL, "hightown_male_cityguard_mob6170038.cl"},
	{0xBB321E703CE567B5ULL, "vxca_cave_uk_0_0.vlm"},
	{0xBB3224103D30EB65ULL, "6140613_m.fxe"},
	{0xBB325D291D498DE5ULL, "479058563.ogg"},
	{0xBB328DF603C1F15CULL, "vxca_caverns1_mines_cju_0_0.vlm"},
	{0xBB33894AEC5EDDF3ULL, "79364745.ogg"},
	{0xBB3435C02C2FDF97ULL, "6031375_m.fxe"},
	{0xBB346E7F32C21014ULL, "vxw_blightlands_10f_0_0.vlm"},
	{0xBB34B4531BB9696CULL, "dae000_pickpocket_scene6175077.cl"},
	{0xBB34B9DFE462D8C1ULL, "abi_w_whirlwind.bnk"},
	{0xBB34C0476B990DD8ULL, "vxdr_0_deeproads_6qw_0_0.vlm"},
	{0xBB34F6094CBE03E1ULL, "llac_w_encounter3_0_4.dds"},
	{0xBB34F6BCA5D5E9A5ULL, "dae000_sketch6179005.cl"},
	{0xBB350BCBE2849DAEULL, "mer120_hubert6034795.cl"},
	{0xBB3569F3C5A64241ULL, "art\\characters\\creatures\\desiredemon\\animation\\exit.ani"},
	{0xBB3593DBB05E259BULL, "gencr_templar_f.utc"},
	{0xBB36B6A0DB98EE88ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_loop_c.cess.evt"},
	{0xBB37AA815ABB023BULL, "963141379.ogg"},
	{0xBB37CCB07DFCA950ULL, "6118208_m.fxe"},
	{0xBB37D509D95DE7E6ULL, "vxlt_0_lowtown_n_eun_0_0.vlm"},
	{0xBB385A4255568811ULL, "6163669_m.fxe"},
	{0xBB387698EAF1EC1FULL, "gen00fl_varric6033564.cl"},
	{0xBB3884437559DE44ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsrock.mmh"},
	{0xBB3888437559E497ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsrock.mao"},
	{0xBB389A43755A03E6ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinsrock.msh"},
	{0xBB38C1E440B0FB44ULL, "hlac_lt_0_lowtown_docks_n_65534_4z.dds"},
	{0xBB38F9C386BC55EBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_ex.ani"},
	{0xBB38FC6DD4F10671ULL, "art\\vfx\\materials\\vfx_shockwave_add.mao"},
	{0xBB396899EBFC6B95ULL, "llac_lt_0_lowtown_docks_n_0_4.dds"},
	{0xBB39AB02D2D87EC2ULL, "vxlt_warehouse_xu_0_0.vlm"},
	{0xBB3A3019CA7DD49CULL, "hlac_lt_0_lowtown_docks_-131074_3x.dds"},
	{0xBB3A4A528E575D15ULL, "6155808_m.fxe"},
	{0xBB3A72835B687431ULL, "vxht_0_hightown_a2_biu_0_0.vlm"},
	{0xBB3B8015E87CA11CULL, "ht_brothel_n_3371_lc.mmh"},
	{0xBB3C1047E8E0ABC6ULL, "6177951_m.fxe"},
	{0xBB3C62BDE68A4E5CULL, "85403107.ogg"},
	{0xBB3C65F74A7D0F43ULL, "vxca_caverns1_mines_in8_0_0.vlm"},
	{0xBB3C74F3C5AC73E3ULL, "art\\characters\\creatures\\desiredemon\\animation\\exit.gad"},
	{0xBB3C8FE7EED34F03ULL, "vxht_0_hightown_n_bnt_0_1.vlm"},
	{0xBB3C9DD557B12A03ULL, "6119265_f.fxe"},
	{0xBB3D7380AA5923A4ULL, "mcr310_argument6075828.cl"},
	{0xBB3D74614C79793CULL, "6070582_m.fxe"},
	{0xBB3D80E6FAEBE026ULL, "llac_dr_0_deeproads_17038_3.dds"},
	{0xBB3DAF745B13FC8DULL, "149830419.ogg"},
	{0xBB3DC13B9241490CULL, "765401414.ogg"},
	{0xBB3DED4FBAF523AEULL, "vxlt_0_lowtown_a2_n_c_er9_0_0.vlm"},
	{0xBB3E57E48F800F00ULL, "hlac_ca_caverns2prc_5802_12y.dds"},
	{0xBB3E64DDA0DDDDDBULL, "6147665_m.fxe"},
	{0xBB3E9A6F015A9E82ULL, "6095387_m.fxe"},
	{0xBB3F17F0EEBBB532ULL, "1053136746.ogg"},
	{0xBB3FC2A53149E5DCULL, "llac_lt_gallowstemplar_n_2390_1.dds"},
	{0xBB3FD1204C4EB203ULL, "6067638_f.fxe"},
	{0xBB400CDB31E1BE1EULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityloop.ani"},
	{0xBB407F12F2EE8020ULL, "qun221tr_petrice_initiates.ncs"},
	{0xBB408167198E8301ULL, "vxw_encounter2_om_0_1.vlm"},
	{0xBB4140FEEBCF0782ULL, "llac_lt_gallowstemplar_2383_2.dds"},
	{0xBB418AEB7053BC6BULL, "hlac_w_bonepitdestroyed_a3_262149_0y.dds"},
	{0xBB419839261458E9ULL, "vxlt_warehouse_pj_0_0.vlm"},
	{0xBB41F2E6BFD7539EULL, "fxp_pillar_sym01.mao"},
	{0xBB41F4B00A4056E5ULL, "rdr111dg_martin6054183.cl"},
	{0xBB4238DFEF9C62FCULL, "6179431_f.fxe"},
	{0xBB428C9BA49027A7ULL, "6170065_m.fxe"},
	{0xBB42E3AA2AA1B0D4ULL, "vxw_encounter2_z_0_0.vlm"},
	{0xBB42FAF80A8BBC94ULL, "835796077.ogg"},
	{0xBB43090C1702429EULL, "art\\characters\\creatures\\mubari\\animation\\tailwag.evt"},
	{0xBB430B660453514FULL, "mag360tr_thrask.ncs"},
	{0xBB430DF6BB078144ULL, "t1_skn_e09.tnt"},
	{0xBB4333993BD2BBC0ULL, "vxht_0_hightown_a2_c1u_0_0.vlm"},
	{0xBB4351F3C5B256C8ULL, "art\\characters\\creatures\\desiredemon\\animation\\exit.evt"},
	{0xBB43944F94227BBFULL, "w_encounter1_43_0.phy"},
	{0xBB43B50C4930FDB2ULL, "455464910.ogg"},
	{0xBB43D32436C5529CULL, "929888864.ogg"},
	{0xBB440A59405213B5ULL, "sum_cloud.dds"},
	{0xBB446411A0F63D8EULL, "vxht_0_hightown_n_b7w_0_0.vlm"},
	{0xBB447D7056C9152CULL, "hlac_ca_caverns1_mines_24424_6x.dds"},
	{0xBB4486F7A574B85AULL, "lt_gallowstemplar_f_mmap.dds"},
	{0xBB451F813B9255F6ULL, "da2_prc_one_en-us.cre"},
	{0xBB4556011DF588BAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjut.msh"},
	{0xBB4564011DF5A0F3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjut.mao"},
	{0xBB4568011DF5A700ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_gallowjut.mmh"},
	{0xBB45E1C386C6B762ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_ex.evt"},
	{0xBB46126E34175864ULL, "hlac_ht_0_hightown_a2_-196608_1y.dds"},
	{0xBB46DB227FC79139ULL, "6115273_m.fxe"},
	{0xBB46FA1CDD69A3F2ULL, "6040548_f.fxe"},
	{0xBB4713CB3A2DC376ULL, "zz_isa_debug6175874.cl"},
	{0xBB4738F877E83A55ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thincorner.phy"},
	{0xBB4755802CF98C15ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_2x.dds"},
	{0xBB47709F6EAAC31DULL, "6095960_m.fxe"},
	{0xBB47A20F8A997653ULL, "vxca_caverns2_720_0_0.vlm"},
	{0xBB47ADCD12FDC681ULL, "6173101_m.fxe"},
	{0xBB4866CC7CB67A6CULL, "follower_banter6136326.cl"},
	{0xBB49200C1706D5C9ULL, "art\\characters\\creatures\\mubari\\animation\\tailwag.gad"},
	{0xBB4A6359FBAE0E52ULL, "sse_ss_fell_orden.gda"},
	{0xBB4A80B7FFFEFC10ULL, "6018755_m.fxe"},
	{0xBB4ABD41364A4064ULL, "hlac_lt_0_lowtown_docks_-196610_0x.dds"},
	{0xBB4ABE4984C14B76ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45win.mao"},
	{0xBB4AC593260BE311ULL, "dae200cr_viscount6087191.cl"},
	{0xBB4ACA4984C15F15ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45win.mmh"},
	{0xBB4ACC4984C1628BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45win.msh"},
	{0xBB4AE7413640C854ULL, "vxlt_hangedman_n_1xq_0_0.vlm"},
	{0xBB4AF5108851A2B0ULL, "vxlt_gallowstemplar_3e2_0_0.vlm"},
	{0xBB4B1E999448CD76ULL, "6086035_m.fxe"},
	{0xBB4BACCEA39839C3ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_cst_lp.ani"},
	{0xBB4BE507110AAA3AULL, "6108030_m.fxe"},
	{0xBB4BE63E3E6F91BCULL, "6075682_m.fxe"},
	{0xBB4C0F28EADF3E40ULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetold_n.dds"},
	{0xBB4CC9FE0EA84A64ULL, "gen00fl_aveline6139507.cl"},
	{0xBB4CE0C386CCD475ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_manimal_ex.gad"},
	{0xBB4D0FF4B9A8663BULL, "651227693.ogg"},
	{0xBB4DA5DC50A18A93ULL, "dae200_cullen6140279.cl"},
	{0xBB4DB7AEAE705544ULL, "gen_im_trs_bst_claws_02.uti"},
	{0xBB4DDC7526DD15C2ULL, "co_lit_01.mao"},
	{0xBB4DEA3A27CB19EFULL, "453515859.ogg"},
	{0xBB4E24DB31EE24EFULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityloop.evt"},
	{0xBB4EAE607A18A4CEULL, "138096597.ogg"},
	{0xBB4FF10C170CA417ULL, "art\\characters\\creatures\\mubari\\animation\\tailwag.ani"},
	{0xBB500A16180A7169ULL, "vxlt_warehouse_ui_0_0.vlm"},
	{0xBB5093FBA2AFDE4AULL, "dae000_gamlen6048409.cl"},
	{0xBB50A29DA724A6DFULL, "702697361.ogg"},
	{0xBB50F069543D60A6ULL, "vxlt_gallowscourtyard_n_4qj_0_0.vlm"},
	{0xBB5129350B049F50ULL, "hlac_lt_darktown_-1_66x.dds"},
	{0xBB521A2E0F851F07ULL, "267673938.ogg"},
	{0xBB521F78F46782CBULL, "295261163.ogg"},
	{0xBB523B7ABDFB1F6DULL, "w_sundermount_80_0_lc.mmh"},
	{0xBB528F71CE5013BFULL, "vxlt_warehouse_101_0_0.vlm"},
	{0xBB52A0D77598D08BULL, "vxlt_gallowstemplar_f_3ir_0_0.vlm"},
	{0xBB52F6087AC99F9FULL, "6144930_m.fxe"},
	{0xBB5380F38A69448DULL, "art\\vfx\\combat\\material__9.mao"},
	{0xBB5411BDDBC3BBA5ULL, "zz_approval_debug6003143.cl"},
	{0xBB5414A21062AA86ULL, "479714693.ogg"},
	{0xBB5425DB31F29198ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\immunityloop.gad"},
	{0xBB542F0391D0E4BEULL, "hlac_lt_0_lowtown_a2_n_c_1_9x.dds"},
	{0xBB548632BEAF2B3DULL, "450840491.ogg"},
	{0xBB54954F9430F0C4ULL, "w_encounter1_43_0.mmh"},
	{0xBB54994F9430F717ULL, "w_encounter1_43_0.mao"},
	{0xBB54AB4F94311666ULL, "w_encounter1_43_0.msh"},
	{0xBB54FADC395479E1ULL, "follower_banter6170507.cl"},
	{0xBB54FDFC4E7DA5F3ULL, "follower_barks6153761.cl"},
	{0xBB558D54BC43CC0AULL, "623561021.ogg"},
	{0xBB55BB7EAB72640AULL, "38812349.ogg"},
	{0xBB55F37683DD3D75ULL, "709996216.ogg"},
	{0xBB561E7F7BBFB563ULL, "w_bonepit_91_0.mao"},
	{0xBB56227F7BBFBBB0ULL, "w_bonepit_91_0.mmh"},
	{0xBB56307F7BBFD3CAULL, "w_bonepit_91_0.msh"},
	{0xBB568ED704A7530BULL, "gen00fl_isabela6100213.cl"},
	{0xBB569C80D0BF2FACULL, "vxw_bonepitdestroyed_a3_kq_0_0.vlm"},
	{0xBB56D8AD6099ABBBULL, "vxht_mansion2b_3c0_0_0.vlm"},
	{0xBB56EF6811A6E178ULL, "mag320_orsino6071115.cl"},
	{0xBB579EF9FB0BBBE2ULL, "dae000_flemeth6000880.cl"},
	{0xBB57AA7C8EE62B94ULL, "6067788_m.fxe"},
	{0xBB57B61E2AFDBD8CULL, "611722063.ogg"},
	{0xBB58391D81222E20ULL, "qun111im_ketojan_cut.uti"},
	{0xBB586F41B5838F05ULL, "plt_mag110pt_wayward_son.plo"},
	{0xBB58AF20F071D9D8ULL, "420632041.ogg"},
	{0xBB59B3A7552D8BF6ULL, "vxlt_0_lowtown_a2_fhe_0_0.vlm"},
	{0xBB59C82012E30888ULL, "6198684_m.fxe"},
	{0xBB5A3D7315AD3274ULL, "hlac_lt_0_lowtown_a2_n_1_3z.dds"},
	{0xBB5AF1CD2CE75718ULL, "ht_keep_3743.rml"},
	{0xBB5B950BA0EBD0C6ULL, "20423096.ogg"},
	{0xBB5C185EF664BEF1ULL, "vxw_bonepitmining_a2_ah_0_0.vlm"},
	{0xBB5C462196A2EC7AULL, "6094821_m.fxe"},
	{0xBB5C65D680351341ULL, "6017849_m.fxe"},
	{0xBB5CB1A001120D4DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_stala1a.msh"},
	{0xBB5CBFA001122480ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_stala1a.mao"},
	{0xBB5CC3A001122BD3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_stala1a.mmh"},
	{0xBB5D012F747F2A26ULL, "lt_0_lowtown_20_0.rml"},
	{0xBB5D186CA8D233D8ULL, "451771010.ogg"},
	{0xBB5D8243842118C3ULL, "vxlt_0_lowtown_a2_n_e17_0_0.vlm"},
	{0xBB5DCD997547BA73ULL, "hlac_ca_caverns1_24423_12z.dds"},
	{0xBB5E046AC992EB33ULL, "vxca_caverns2_81r_0_0.vlm"},
	{0xBB5E12E5E473F9F2ULL, "fen110_fenris6100282.cl"},
	{0xBB5E36FF6C905299ULL, "6183973_m.fxe"},
	{0xBB5EA04394671192ULL, "6080715_f.fxe"},
	{0xBB5ECA1E0C5D3191ULL, "hlac_lt_undercityencounter_-1_11y.dds"},
	{0xBB5EF05973B7BF0BULL, "vxlt_gallowstemplar_n_38y_0_0.vlm"},
	{0xBB5F045F84BDC19BULL, "148194659.ogg"},
	{0xBB5F1C807F99E979ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\_textures\\prp_cot_sa.dds"},
	{0xBB5F77FE1C26B256ULL, "ttx_roof_int_e_1.mmh"},
	{0xBB5F81FE1C26C354ULL, "ttx_roof_int_e_1.msh"},
	{0xBB5FC3CEA3A909ADULL, "art\\characters\\creatures\\desiredemon\\animation\\d_cst_lp.gad"},
	{0xBB5FD8FB64945A53ULL, "560745020.ogg"},
	{0xBB5FEAE4F854287EULL, "gen00fl_aveline6114346.cl"},
	{0xBB5FF2FBBA3223C4ULL, "6133377_m.fxe"},
	{0xBB603246B0E55304ULL, "hlac_lt_gallowstemplar_n_2393_0x.dds"},
	{0xBB60507CFC1FD54EULL, "6101451_m.fxe"},
	{0xBB6082E535C62B4AULL, "vxlt_gallowsprison_n_6mb_0_0.vlm"},
	{0xBB60F365E3CAFFEFULL, "412374369.ogg"},
	{0xBB60F8664E5370F9ULL, "90266024.ogg"},
	{0xBB61F866046DE1F6ULL, "mag360tr_thrask.utt"},
	{0xBB621D8DF0775142ULL, "dae000_arishok6081599.cl"},
	{0xBB6223659A1A7627ULL, "661227890.ogg"},
	{0xBB624AD4BAF57A26ULL, "one140cr_charade6134313.cl"},
	{0xBB6255FBC7E0A476ULL, "161850711.ogg"},
	{0xBB626928C43933DAULL, "927733672.ogg"},
	{0xBB62779288DB6811ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em04.mmh"},
	{0xBB62919288DB94CFULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em04.msh"},
	{0xBB62A4B9C92803A2ULL, "596982829.ogg"},
	{0xBB62D20CCF60FAD0ULL, "6067682_f.fxe"},
	{0xBB63686A84400361ULL, "90162515.ogg"},
	{0xBB63C78158C1B189ULL, "vxht_0_hightown_a3_b8c_0_0.vlm"},
	{0xBB642CB73C0A53F4ULL, "vxlt_0_lowtown_a3_n_c_gm1_0_0.vlm"},
	{0xBB6439358CB8191CULL, "hlac_lt_0_lowtown_docks_n_-196611_1y.dds"},
	{0xBB6558417DDBE6E4ULL, "58249237.ogg"},
	{0xBB65C36489A5FBBDULL, "vxht_0_hightown_bqe_0_0.vlm"},
	{0xBB65DC7B6552DCCEULL, "vxht_0_hightown_b8m_0_0.vlm"},
	{0xBB66AF4FEE74D7E1ULL, "hlac_lt_warehouse_0_4x.dds"},
	{0xBB66C88AD6F05312ULL, "llac_ht_playershouse_n_855_7.dds"},
	{0xBB66F418CF3B9DFFULL, "lgt110_gxa_vgk_silver_order6117000.cl"},
	{0xBB67536E55D73234ULL, "w_bonepit_21_0_ll.mmh"},
	{0xBB67BA5A96885F90ULL, "ave331_jeven6138774.cl"},
	{0xBB67E1B2D7C18341ULL, "6074381_m.fxe"},
	{0xBB67F923A54CD6AFULL, "art\\levels\\ztools\\default\\generic\\3oi_basic.mao"},
	{0xBB680523A54CEB4CULL, "art\\levels\\ztools\\default\\generic\\3oi_basic.mmh"},
	{0xBB680B23A54CF53EULL, "art\\levels\\ztools\\default\\generic\\3oi_basic.msh"},
	{0xBB685B2218267C36ULL, "568182079.ogg"},
	{0xBB686893F9B6E6AAULL, "w_blightlands_0_0.mao"},
	{0xBB686C93F9B6EDF9ULL, "w_blightlands_0_0.mmh"},
	{0xBB687693F9B6FEE7ULL, "w_blightlands_0_0.msh"},
	{0xBB68AEBF7E95DB5BULL, "w_bonepitdestroyed_a3_175_0.msh"},
	{0xBB68BE31B05541A6ULL, "6051270_f.fxe"},
	{0xBB68C0BF7E95F9A6ULL, "w_bonepitdestroyed_a3_175_0.mao"},
	{0xBB68C8C0F4F6DB3AULL, "6106230_m.fxe"},
	{0xBB68CCBF7E960E45ULL, "w_bonepitdestroyed_a3_175_0.mmh"},
	{0xBB69506D6217D43EULL, "6103691_player_too_far.cut"},
	{0xBB695D0719D264B8ULL, "hlac_dr_0_deeproads_17038_1z.dds"},
	{0xBB6A06B913FBE905ULL, "6116557_m.fxe"},
	{0xBB6A06FC8D9CBF44ULL, "497984472.ogg"},
	{0xBB6A28AD3562B79AULL, "6109266_m.fxe"},
	{0xBB6A5F7C6933D97FULL, "hlac_lt_0_lowtown_docks_n_-262144_0x.dds"},
	{0xBB6A667ABA2922EAULL, "6153445_m.fxe"},
	{0xBB6B7D0A75DBAA24ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1b.mmh"},
	{0xBB6B810A75DBB177ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1b.mao"},
	{0xBB6B930A75DBCFC6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock1b.msh"},
	{0xBB6BFA3DACF5FE22ULL, "nat380_nathaniel6066188.cl"},
	{0xBB6CDEE2AC5793DBULL, "character_character_p255.psh.fxo.dx11"},
	{0xBB6D31118F7E244CULL, "vxht_0_hightown_9p4_0_0.vlm"},
	{0xBB6D5DE8CBC81DF0ULL, "hlac_w_blightlands_196605_0x.dds"},
	{0xBB6DA942D1EEA5E6ULL, "6030445_m.fxe"},
	{0xBB6DB81ED51803E0ULL, "vxlt_ruinedhouse_n_i7_0_0.vlm"},
	{0xBB6E85D8D5E557F2ULL, "6036861_m.fxe"},
	{0xBB6E8A2C4F9FCACFULL, "vxca_caverns2_9as_0_0.vlm"},
	{0xBB6E8A32CAC67834ULL, "art\\cinematics\\act3\\romances\\anders\\cs_pcfem_andersrom.ani"},
	{0xBB6EAC552FB625BAULL, "l_vrc321ar_mansion_haun.bnk"},
	{0xBB6EBB396D5C626AULL, "mag291_torpor6026963.cl"},
	{0xBB6EE5108F69BE30ULL, "vxlt_gallowstemplar_n_3sc_0_0.vlm"},
	{0xBB6EEEB3E7E6D689ULL, "573166661.ogg"},
	{0xBB6F5D0BF44311A7ULL, "309126406.ogg"},
	{0xBB6F84915985018BULL, "vxlt_gallowstemplar_f_3st_0_0.vlm"},
	{0xBB6FCF99F414D60DULL, "hlac_ca_caverns2_5802_23y.dds"},
	{0xBB6FD0BE5311BF6DULL, "nat380_nathaniel.bnk"},
	{0xBB70DCB721F4CA19ULL, "vxht_0_hightown_8pd_0_0.vlm"},
	{0xBB70FB76CD9674B9ULL, "qun221_petrice6076886.cl"},
	{0xBB7186D45307C86EULL, "lgt101_mag_escp_terrie6156643.cl"},
	{0xBB71E6C7644F3303ULL, "zz_pro_debug6059757.cl"},
	{0xBB7210E5413E60F8ULL, "vxlt_hangedman_n_224_0_0.vlm"},
	{0xBB72395E4AB7CA09ULL, "vxht_0_hightown_a3_b7b_0_0.vlm"},
	{0xBB7279AB03CC7E49ULL, "zz_vrc_debug6200249.cl"},
	{0xBB7282EC7BBBB31DULL, "plt_isa100pt_recruited.plo"},
	{0xBB72DB05F6907BE5ULL, "rand_mountains.gda"},
	{0xBB72F2BE53142ED5ULL, "nat380_nathaniel.cnv"},
	{0xBB73109E60291352ULL, "238110029.ogg"},
	{0xBB73969BFD8C7614ULL, "700887860.ogg"},
	{0xBB73A35BA47E4A82ULL, "vxht_keep_n_36y_0_0.vlm"},
	{0xBB7435417B9C143DULL, "vxht_0_hightown_bzd_0_0.vlm"},
	{0xBB74D33281F72B97ULL, "hlac_w_sundermount_65537_0x.dds"},
	{0xBB74DF22CDAA689DULL, "6082616_m.fxe"},
	{0xBB753B7DC4ABFBEAULL, "vxht_mansion2_2ox_0_0.vlm"},
	{0xBB757C89C1A24C8BULL, "6076622_m.fxe"},
	{0xBB758732CACC9196ULL, "art\\cinematics\\act3\\romances\\anders\\cs_pcfem_andersrom.gad"},
	{0xBB75AF4AFE05E5E7ULL, "gen_im_trs_slv_hvy_cht_r2.uti"},
	{0xBB760231BF80B502ULL, "10126000.ogg"},
	{0xBB7614537EF0478CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06b.mmh"},
	{0xBB761A537EF0527EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06b.msh"},
	{0xBB7621506ECDB7F5ULL, "6086968_m.fxe"},
	{0xBB764AA75E33EC30ULL, "vxlt_gallowstemplar_3t1_0_0.vlm"},
	{0xBB7676C16CD3FA80ULL, "662978983.ogg"},
	{0xBB768300F1647DF8ULL, "hla_ht_0_hightown_a2_-65534_0z.dds"},
	{0xBB769DB266F107C2ULL, "vxlt_blackemporium_11k_0_0.vlm"},
	{0xBB7749397E80FD46ULL, "316053112.ogg"},
	{0xBB77C38A54507BA3ULL, "339962483.ogg"},
	{0xBB780770AF89594AULL, "ser200_aveline6201471.cl"},
	{0xBB780A45478DBBB7ULL, "mrl000_marethari6129457.cl"},
	{0xBB780E2BF955056BULL, "11711048.ogg"},
	{0xBB78954EB87E66AEULL, "6143163_m.fxe"},
	{0xBB78EBCD9BE8624BULL, "zz_followers_banter6157993.cl"},
	{0xBB78F7954BD03DC1ULL, "follower_banter6173032.cl"},
	{0xBB78FC18DA50EFF1ULL, "vxw_bonepitmining_a2_kn_0_0.vlm"},
	{0xBB79003D19A23554ULL, "891049984.ogg"},
	{0xBB792196A9874A97ULL, "vxht_chantry_301_0_0.vlm"},
	{0xBB792240198B7C75ULL, "523207424.ogg"},
	{0xBB7985EC7BC1D74AULL, "plt_isa100pt_recruited.nss"},
	{0xBB799C374FC9F30DULL, "1029735060.ogg"},
	{0xBB79F4B61A6B0675ULL, "27759614.ogg"},
	{0xBB7BA6AA8D5DFA3BULL, "dae210cr_gate_guard.utc"},
	{0xBB7BC7DFE49EF373ULL, "abi_w_whirlwind.wbd"},
	{0xBB7C579F28F6DEDDULL, "233597796.ogg"},
	{0xBB7D3A8CFCF0E381ULL, "575378190.ogg"},
	{0xBB7D6A29C525C773ULL, "hlac_ca_cavern1_haunted_24423_2x.dds"},
	{0xBB7D89662958F721ULL, "dae000_elegant6166933.cl"},
	{0xBB7DA40DB9DB7CC3ULL, "1069705436.ogg"},
	{0xBB7E089C2020BC52ULL, "302089817.ogg"},
	{0xBB7E4178C118DD92ULL, "421164642.ogg"},
	{0xBB7F68488934E19DULL, "gen00fl_isabela6078985.cl"},
	{0xBB7F69699643AEE8ULL, "follower_banter6163430.cl"},
	{0xBB7FAD37FDAFAE03ULL, "vxlt_shop_ed_0_0.vlm"},
	{0xBB80F0E04CD2C752ULL, "hlac_ca_cavern1_haunted_24422_0x.dds"},
	{0xBB8126A34CE06F8BULL, "fx_equake.dds"},
	{0xBB814F28990CE4BDULL, "179520661.ogg"},
	{0xBB81B4D56E7B124BULL, "6043536_m.fxe"},
	{0xBB81B8A5210BF548ULL, "6174718_m.fxe"},
	{0xBB8236849A4C3025ULL, "6016789_m.fxe"},
	{0xBB82A09F13B5CBFDULL, "6180463_f.fxe"},
	{0xBB82B018BC40D8E9ULL, "6123311_m.fxe"},
	{0xBB831FE5B65ACD40ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_017a_0s.dds"},
	{0xBB8350FFBFA0811DULL, "6085716_m.fxe"},
	{0xBB835E5901D5EC57ULL, "980367174.ogg"},
	{0xBB83B60A5807C301ULL, "vis121_saemus6085694.cl"},
	{0xBB84028B70DF97D8ULL, "334880899.ogg"},
	{0xBB843281EE542AC8ULL, "ico_shortbow_5.dds"},
	{0xBB8435778D12F753ULL, "lowtown_fem_alien_elf_2_mob6166362.cl"},
	{0xBB844112CB47FA5DULL, "vxlt_0_lowtown_a3_n_c_eqf_0_0.vlm"},
	{0xBB8564EA9A32A31CULL, "6085426_f.fxe"},
	{0xBB861B94A0223697ULL, "6115962_f.fxe"},
	{0xBB862888BDED49C2ULL, "vxw_encounter3_4e_0_0.vlm"},
	{0xBB865DFDEE338917ULL, "dre100cr_frozen_ogre.utc"},
	{0xBB866B19A9ACD243ULL, "198115400.ogg"},
	{0xBB86B317F551E0B7ULL, "mag360cr_hostage_aveline.utc"},
	{0xBB86CDC3EB205831ULL, "vxht_keep_n_57k_0_0.vlm"},
	{0xBB8755702A09CF8EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_sbash.evt"},
	{0xBB884E7329C653C4ULL, "vxlt_0_lowtown_docks_n_c3i_0_0.vlm"},
	{0xBB88FA5B3F77CA57ULL, "1066935096.ogg"},
	{0xBB89769FFB9FEDC2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potc_d.dds"},
	{0xBB89996E171676FBULL, "vxht_bartrand_n_6jb_0_0.vlm"},
	{0xBB89B46E2EC35763ULL, "6126431_m.fxe"},
	{0xBB8A1490454F051BULL, "mag121_viveka6082618.cl"},
	{0xBB8A31681A4B0731ULL, "mcr310_sibling6076386.cl"},
	{0xBB8A447BDA163685ULL, "plt_cod_ltr_x_invitation.plo"},
	{0xBB8ACA2A4FF41070ULL, "fall_cl_f.mao"},
	{0xBB8B1ADA64AB90B0ULL, "vxlt_0_lowtown_docks_n_b84_0_1.vlm"},
	{0xBB8B374FD2849140ULL, "qun131cr_warrens_thug7.utc"},
	{0xBB8B988C5CC25F91ULL, "43078681.ogg"},
	{0xBB8BAFBBAD8F2071ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_6y.dds"},
	{0xBB8BCE80F67AA52DULL, "vxht_chantry_2c7_0_0.vlm"},
	{0xBB8BF8203DFEB6DBULL, "dae000pt_home_ambients.ncs"},
	{0xBB8BFE3118CC9A4DULL, "hlac_lt_0_lowtown_a3_n_c_-1_0x.dds"},
	{0xBB8C077971BF216CULL, "347699066.ogg"},
	{0xBB8C135064645E17ULL, "w_blightlands_80_0.msh"},
	{0xBB8C255064647D7AULL, "w_blightlands_80_0.mao"},
	{0xBB8C295064648389ULL, "w_blightlands_80_0.mmh"},
	{0xBB8C64C506DF84E4ULL, "llac_lt_0_lowtown_a2_n_c_-65537_1.dds"},
	{0xBB8CAB856C2DB8E2ULL, "vxht_mansion2b_3dt_0_0.vlm"},
	{0xBB8D32F4D6FD0530ULL, "6111009_m.fxe"},
	{0xBB8D656ADF909A6AULL, "459943433.ogg"},
	{0xBB8DBFDAE05F8F89ULL, "6173013_m.fxe"},
	{0xBB8DF4B60CCA4788ULL, "324176775.ogg"},
	{0xBB8E4BF878245D70ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thincorner.msh"},
	{0xBB8E4C702A0FDF39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_sbash.gad"},
	{0xBB8E59F87824748AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thincorner.mmh"},
	{0xBB8E5DF878247BC1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3pp_thincorner.mao"},
	{0xBB8EDE52872D856FULL, "223804839.ogg"},
	{0xBB8F71DA9F681692ULL, "38941845.ogg"},
	{0xBB8F824E11EC9D74ULL, "vxlt_0_lowtown_a3_n_c_gz4_0_0.vlm"},
	{0xBB8F8F1A30BB5743ULL, "83952349.ogg"},
	{0xBB8FA402F599A545ULL, "6182941_m.fxe"},
	{0xBB8FB769FA0F5AC8ULL, "cd_gen00fl_anders.bnk"},
	{0xBB8FC22641794912ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.mmh"},
	{0xBB8FC67F35EF50FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_action.gad"},
	{0xBB9192E3B86BC32BULL, "cd_dae111_gamlen_intro.wbd"},
	{0xBB92367FB3D7ADE7ULL, "1016117475.ogg"},
	{0xBB925C523710DB2DULL, "vxdr_0_deeproads_fgl_0_0.vlm"},
	{0xBB9349C393EACB3BULL, "hlac_lt_0_lowtown_a2_1_9z.dds"},
	{0xBB937B43D3D19C36ULL, "hlac_lt_0_lowtown_docks_131071_7z.dds"},
	{0xBB93C6552FD56E58ULL, "l_vrc321ar_mansion_haun.wbd"},
	{0xBB93D39179DD0ECDULL, "seb221_ruxton_harimann6177197.cl"},
	{0xBB94556EE57070A0ULL, "follower_banter6164017.cl"},
	{0xBB945B093D86078AULL, "dae000ip_crowd_ff8_talk5.utp"},
	{0xBB955123BE499347ULL, "6077315_m.fxe"},
	{0xBB955D702A161A87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\abilities\\sw_tal_sbash.ani"},
	{0xBB957D489FE1AFCEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03a.phy"},
	{0xBB958EAEC10878C8ULL, "631218500.ogg"},
	{0xBB95C97F35F3C111ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_action.evt"},
	{0xBB95FF7A3397F371ULL, "hlac_ht_0_hightown_a2_0_2x.dds"},
	{0xBB960FD6EEACC040ULL, "6119257_f.fxe"},
	{0xBB96B10C5D52E9C2ULL, "l_dae151ar_mount_caverns.bnk"},
	{0xBB96C2A1747D3C4BULL, "238176767.ogg"},
	{0xBB970F9EC0872692ULL, "6040770_m.fxe"},
	{0xBB97855B84B11F2EULL, "vxlt_0_lowtown_a2_n_c_how_0_0.vlm"},
	{0xBB97B9B31E658868ULL, "96702790.ogg"},
	{0xBB98264B20B96FE3ULL, "6123731_m.fxe"},
	{0xBB98757E4ABA1561ULL, "seb221_random_servants6187834.cl"},
	{0xBB98AAB60C974497ULL, "dae300_alistair_king6133812.cl"},
	{0xBB98F4E1D3015109ULL, "6137014_f.fxe"},
	{0xBB993A64055FDB35ULL, "isa211_isabela6202112.cl"},
	{0xBB998CB756A1A930ULL, "vxlt_0_lowtown_docks_n_b53_0_1.vlm"},
	{0xBB99957AC8D34CB6ULL, "mage_slowzone.dds"},
	{0xBB9A0C22B5B20AF8ULL, "172410502.ogg"},
	{0xBB9A70C1614D0B63ULL, "vxlt_0_lowtown_a2_gc0_0_0.vlm"},
	{0xBB9ABAF757204F3DULL, "and110_refugee6111609.cl"},
	{0xBB9AEB19B9D40CAAULL, "6051798_m.fxe"},
	{0xBB9B0862736FF18AULL, "zz_dog_debug6127928.cl"},
	{0xBB9B362FD291CA5FULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyo_0t.dds"},
	{0xBB9C402C6CE3BDFAULL, "zz_fol_debug6096564.cl"},
	{0xBB9C526A0EBFB8E8ULL, "vxht_mansion2_2ei_0_1.vlm"},
	{0xBB9CDDEF51DAEFC9ULL, "vxlt_0_lowtown_eqv_0_0.vlm"},
	{0xBB9CE983374BA731ULL, "vxlt_0_lowtown_docks_cnl_0_0.vlm"},
	{0xBB9CF0B731E26470ULL, "fxa_shield_spk.dds"},
	{0xBB9D82CE740DE43EULL, "6119542_m.fxe"},
	{0xBB9DE75F4978B21CULL, "one280_emrys6137202.cl"},
	{0xBB9DEA29F90764D0ULL, "338263285.ogg"},
	{0xBB9DF7A22F0D2665ULL, "bh.cut"},
	{0xBB9E508A6EB291A7ULL, "353879127.ogg"},
	{0xBB9E77730E1DF46DULL, "w_encounter3_84_0.mmh"},
	{0xBB9E7B730E1DFABEULL, "w_encounter3_84_0.mao"},
	{0xBB9E85EA9F06888EULL, "hlac_lt_gallowstemplar_7198_3z.dds"},
	{0xBB9E89730E1E12F3ULL, "w_encounter3_84_0.msh"},
	{0xBB9EADB9D43538CCULL, "cheese_n.dds"},
	{0xBB9FA5D156B5B625ULL, "518846935.ogg"},
	{0xBBA03FA34CFB4B6DULL, "fx_equake.mao"},
	{0xBBA05FDE7284F8D2ULL, "824672461.ogg"},
	{0xBBA0E877554E3735ULL, "6059146_m.fxe"},
	{0xBBA1AE600D905783ULL, "vxlt_gallowscourtyard_77a_0_0.vlm"},
	{0xBBA1B0A7DC8DBD05ULL, "131906719.ogg"},
	{0xBBA1C4A9E6AD5556ULL, "191895832.ogg"},
	{0xBBA25EA969ED9E08ULL, "gencr_rock_wraith.utc"},
	{0xBBA27205E26DB08CULL, "971650.ogg"},
	{0xBBA2B9E098CB9A58ULL, "var_trigger_teammove.gda"},
	{0xBBA2BCBBB40FEFB7ULL, "832350127.ogg"},
	{0xBBA318D849D00423ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathpose.ani"},
	{0xBBA3E17F360027E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_action.ani"},
	{0xBBA4437916999261ULL, "dae000_flemeth6013409.cl"},
	{0xBBA5315FA51E4194ULL, "hlac_lt_warehouse_n_-65536_2z.dds"},
	{0xBBA5A9D9360DED4EULL, "lowtown_male_cityguard_mob6150714.cl"},
	{0xBBA5B7BD241E4FA2ULL, "vxlt_gallowscourtyard_n_66y_0_0.vlm"},
	{0xBBA5C821F9FD55ADULL, "6019249_m.fxe"},
	{0xBBA5CCFF0CCE6FF9ULL, "dae170ip_armor_store.utp"},
	{0xBBA6159D01D05882ULL, "vxlt_warehouse_n_13b_0_0.vlm"},
	{0xBBA67D27FB02A7FBULL, "vxht_bartrand_n_6pd_0_0.vlm"},
	{0xBBA688D3918F27B9ULL, "hlac_lt_0_lowtown_a2_65536_11y.dds"},
	{0xBBA69B753B91774BULL, "art\\vfx\\creatures\\fxc_drg_brth_l_c.impa.evt"},
	{0xBBA71C476E26B59CULL, "509942113.ogg"},
	{0xBBA75B53A1634C34ULL, "271137003.ogg"},
	{0xBBA78973231BEBA6ULL, "qun221_petrice6077347.cl"},
	{0xBBA79D72F24EDE95ULL, "611760671.ogg"},
	{0xBBA84CA55BC725A8ULL, "hlac_lt_0_lowtown_docks_a3_n_196606_0z.dds"},
	{0xBBA87F0EFA2EC98AULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_poplar.mao"},
	{0xBBA8830EFA2ED0D9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_poplar.mmh"},
	{0xBBA88D0EFA2EE1C7ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_poplar.msh"},
	{0xBBA8B23ADA66D62DULL, "vxht_chantry_2y5_0_0.vlm"},
	{0xBBA8BEFB2F37BD23ULL, "art\\characters\\weapons\\longswords\\w_lsw_014a.mao"},
	{0xBBA8C2FB2F37C470ULL, "art\\characters\\weapons\\longswords\\w_lsw_014a.mmh"},
	{0xBBA8D0FB2F37DB8AULL, "art\\characters\\weapons\\longswords\\w_lsw_014a.msh"},
	{0xBBA91F1F8B26D2C6ULL, "vxlt_undercityserialkiller_1il_0_0.vlm"},
	{0xBBA944176B1F85B2ULL, "gen00fl_anders6136024.cl"},
	{0xBBA9DAF131A96D34ULL, "vxca_cavern1_haunted_fmy_0_0.vlm"},
	{0xBBAA0996339700E3ULL, "docks_fem_sailor_mob6176428.cl"},
	{0xBBAA2E93222585A0ULL, "gen00fl_fenris6113890.cl"},
	{0xBBAB7C19FFB8E74BULL, "dae000_brothel_orana6160872.cl"},
	{0xBBAB94039B05F99FULL, "vxht_mansion2_1ps_0_0.vlm"},
	{0xBBAC574999C4A410ULL, "938710075.ogg"},
	{0xBBACB78FFBBAAA4DULL, "hm_cgn_001.mor"},
	{0xBBACB78FFBBAAA4FULL, "hm_cgn_001.mop"},
	{0xBBACCC2EEC3B375DULL, "hlac_dr_dungeon_3214_2y.dds"},
	{0xBBAD089718A3AA81ULL, "hlac_ht_chantry_n_2534_51y.dds"},
	{0xBBAD8530CAEA804CULL, "qun120_arishok.stg"},
	{0xBBAD8E00D32F5A8CULL, "6143701_m.fxe"},
	{0xBBADADE5EDC93F3CULL, "hlac_ca_cave_2194_2y.dds"},
	{0xBBADB24E12394084ULL, "298943524.ogg"},
	{0xBBADB4753B97C00CULL, "art\\vfx\\creatures\\fxc_drg_brth_l_c.impa.gad"},
	{0xBBAE0F75FD8EE261ULL, "art\\vfx\\combat\\vfx_gib_qunc_ltr.anb"},
	{0xBBAE332E7D1FF716ULL, "dae000cr_grand_cleric6182028.cl"},
	{0xBBAE5210803A7EA3ULL, "6171088_m.fxe"},
	{0xBBAEACEA07997AA4ULL, "llac_lt_0_lowtown_docks_a3_131073_0.dds"},
	{0xBBAEAF11FEFDBAD6ULL, "gen_im_trs_rdr_hvy_boo_r1.uti"},
	{0xBBAEB9BBC06DF299ULL, "6004122_m.fxe"},
	{0xBBAED56360140D2FULL, "17997499.ogg"},
	{0xBBAEDDF734D45120ULL, "vxw_bonepitmining_a2_ma_0_0.vlm"},
	{0xBBAF59C0B559C116ULL, "hlac_ht_playershouse_n_855_21z.dds"},
	{0xBBAF70730E2C5B2AULL, "w_encounter3_84_0.phy"},
	{0xBBAF96C43448379FULL, "6170627_m.fxe"},
	{0xBBB01D0176390400ULL, "570661872.ogg"},
	{0xBBB1C96FC0090D46ULL, "follower_banter6170771.cl"},
	{0xBBB2328656BB77CAULL, "llac_lt_gallowscourtyard_-131071_0.dds"},
	{0xBBB24588A3C31E58ULL, "6191283_m.fxe"},
	{0xBBB2684AD20DE6A9ULL, "hlac_ht_0_hightown_a3_-1_1x.dds"},
	{0xBBB32180D96C32C6ULL, "145526841.ogg"},
	{0xBBB327F06A2F9F45ULL, "character_character_p165.psh.fxo.dx11"},
	{0xBBB3B3753B9C2A0AULL, "art\\vfx\\creatures\\fxc_drg_brth_l_c.impa.ani"},
	{0xBBB3D9B68ED18370ULL, "main_plot_screenshots2.gda"},
	{0xBBB3FFB731F55C0EULL, "fxa_shield_spk.mao"},
	{0xBBB40DF778D9D985ULL, "979825027.ogg"},
	{0xBBB4A969FA2E5F72ULL, "cd_gen00fl_anders.wbd"},
	{0xBBB4FF8347EEAA9DULL, "6179882_m.fxe"},
	{0xBBB617C8E0EB0D9FULL, "hlac_dr_dungeon_3215_6x.dds"},
	{0xBBB64F8DB7F55B89ULL, "284543056.ogg"},
	{0xBBB674E3B88AACF9ULL, "cd_dae111_gamlen_intro.bnk"},
	{0xBBB679CF6D3D6B06ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingb_sa.dds"},
	{0xBBB72FDB65F53E39ULL, "ss_sundercut_1.cut"},
	{0xBBB7531F8EB9E8BAULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_smoke01.dds"},
	{0xBBB82FD849E2870DULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathpose.gad"},
	{0xBBB8C126419C4CB9ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_l_p.anb"},
	{0xBBB8C254567C87B0ULL, "vxht_0_hightown_asy_0_0.vlm"},
	{0xBBB9244323C1508EULL, "6207391_m.fxe"},
	{0xBBB9CD3D1B37D831ULL, "vxlt_0_lowtown_docks_ctb_0_0.vlm"},
	{0xBBBA15607A3DD223ULL, "art\\characters\\weapons\\dagger\\w_dgr_demo_only.msh"},
	{0xBBBA17096B6F781EULL, "vrc221_varric6040699.cl"},
	{0xBBBA23607A3DEA7DULL, "art\\characters\\weapons\\dagger\\w_dgr_demo_only.mmh"},
	{0xBBBA2642696F34DEULL, "lt_0_lowtown_a2_n_c_43_0_ll.mmh"},
	{0xBBBA27607A3DF08EULL, "art\\characters\\weapons\\dagger\\w_dgr_demo_only.mao"},
	{0xBBBA64519FB01F31ULL, "6024530_f.fxe"},
	{0xBBBA82AA37272EB7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masa_0d.dds"},
	{0xBBBBA45D02D87973ULL, "vxlt_undercitydungeon_1gb_0_0.vlm"},
	{0xBBBBD72553777556ULL, "6036757_m.fxe"},
	{0xBBBBEEAB586ECEF7ULL, "gen_im_wep_war_one_act2_04.uti"},
	{0xBBBC9B0C5D7393E0ULL, "l_dae151ar_mount_caverns.wbd"},
	{0xBBBD0852321AF0C9ULL, "hlac_ht_mansion_n_3367_0z.dds"},
	{0xBBBD4FD426CA69A0ULL, "vxw_bonepitmining_a2_hn_0_0.vlm"},
	{0xBBBD7D4B0CDA4218ULL, "554902818.ogg"},
	{0xBBBD96D0DA79045FULL, "6146860_m.fxe"},
	{0xBBBDA256C4A0671AULL, "gen00fl_carver6006073.cl"},
	{0xBBBE7E974C6F5808ULL, "vxdr_primevalentrance_2oj_0_1.vlm"},
	{0xBBBE83EA18E330F5ULL, "lt_0_lowtown_35_0_lc.mmh"},
	{0xBBBEEFA521F11152ULL, "6013363_m.fxe"},
	{0xBBBFA8ADF30423C7ULL, "dae000_hate_enemies6167595.cl"},
	{0xBBBFE1C7F1E158C7ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_paintingb_sa.dds"},
	{0xBBC001BBAD4BD296ULL, "w_bonepitmining_a2_30_0.rml"},
	{0xBBC009A151608977ULL, "6090406_m.fxe"},
	{0xBBC03A61CD889A23ULL, "hlac_w_bonepitdestroyed_a3_262148_0z.dds"},
	{0xBBC121FD7F48E6D1ULL, "follower_banter6171935.cl"},
	{0xBBC1407EDD02061CULL, "lgt260_eau_tmp_flowergirl6143872.cl"},
	{0xBBC1D7A5070F4E5EULL, "303111483.ogg"},
	{0xBBC33B3A01483D19ULL, "873887950.ogg"},
	{0xBBC38944D036CC12ULL, "vxlt_darktown_41g_0_0.vlm"},
	{0xBBC39BD5F96CF776ULL, "272056224.ogg"},
	{0xBBC422579B4CBFC3ULL, "6111395_f.fxe"},
	{0xBBC46C93FA054C52ULL, "w_blightlands_0_0.rml"},
	{0xBBC48B4C2B46B681ULL, "mag371_elsa6171594.cl"},
	{0xBBC531875779EAB9ULL, "mcr370_cut_meredith_idol.bnk"},
	{0xBBC5345DBD799840ULL, "vxht_0_hightown_a2_cdz_0_0.vlm"},
	{0xBBC5D5FC12D4A122ULL, "ht_0_hightownexplosion_n_17_0_ll.mmh"},
	{0xBBC64AE22186799BULL, "mer210_seldrek6078790.cl"},
	{0xBBC6A23EF4FF822BULL, "one111ip_ggt_trap.utp"},
	{0xBBC6BAF22ACEB9BBULL, "6068962_m.fxe"},
	{0xBBC7336ACFEE2D9DULL, "gen00fl_isabela6196850.cl"},
	{0xBBC78DB444D8D58AULL, "gen00fl_aveline6115864.cl"},
	{0xBBC7AC8692BDC68BULL, "framebuffereffects_verticalblur_p0.vsh.fxo"},
	{0xBBC7C36D63E93127ULL, "14842949.ogg"},
	{0xBBC7CFD2AAD585B5ULL, "vxdr_0_deeproads_cr2_0_0.vlm"},
	{0xBBC8194D5B3FE6EBULL, "vxca_caverns1_jfv_0_0.vlm"},
	{0xBBC85387577C5A19ULL, "mcr370_cut_meredith_idol.cnv"},
	{0xBBC86CC8C342D654ULL, "qcr000_aveline6041077.cl"},
	{0xBBC8F7155196CC0BULL, "vxht_0_hightown_a2_a83_0_0.vlm"},
	{0xBBC92395B443746CULL, "llac_ht_0_hightown_a2_65537_1.dds"},
	{0xBBCA1639F4CE91F3ULL, "vxlt_undercitydungeon_1je_0_0.vlm"},
	{0xBBCA31133F030375ULL, "hairalpha_rendertodepth_p0.vsh.fxo"},
	{0xBBCA35045AF3721BULL, "da2terrain.mat.dx11"},
	{0xBBCB4388B0E3DAA2ULL, "hlac_lt_0_lowtown_docks_a3_-196610_3x.dds"},
	{0xBBCC1B259A7EAD1EULL, "dre180ip_primeval_chest_1.utp"},
	{0xBBCC9756E668DD51ULL, "896583752.ogg"},
	{0xBBCD0E014BEEEC84ULL, "vxht_0_hightown_n_c7d_0_0.vlm"},
	{0xBBCD3CE227913B03ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4sc_basic.mao"},
	{0xBBCD40E227914250ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4sc_basic.mmh"},
	{0xBBCD4EE227915A6AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4sc_basic.msh"},
	{0xBBCDFC21EAE7B2A7ULL, "6155385_m.fxe"},
	{0xBBCE27BB82C9609BULL, "art\\vfx\\environment\\vfx_lyrium_aurora.dura.ani"},
	{0xBBCE363A042DB710ULL, "vxlt_warehouse_8w_0_0.vlm"},
	{0xBBCE4D71DA7C1B82ULL, "302184408.ogg"},
	{0xBBCECFCF1E0F9643ULL, "vxlt_0_lowtown_docks_a3_n_cxd_0_0.vlm"},
	{0xBBCED3110A02FE8AULL, "vxdr_primevalentrance_1j4_0_1.vlm"},
	{0xBBCEE1F49FF0D53FULL, "ico_staff_0.dds"},
	{0xBBCEED572DEE36FDULL, "233584846.ogg"},
	{0xBBCF0E1340CB4D68ULL, "gen00fl_sebastian6180501.cl"},
	{0xBBCFEEB5E3159706ULL, "vxdr_0_deeproads_f4j_0_0.vlm"},
	{0xBBD08579B2652688ULL, "6129136_m.fxe"},
	{0xBBD0A9D66B1406B7ULL, "one110_meeran6089470.cl"},
	{0xBBD126DEC0B6533AULL, "50719875.ogg"},
	{0xBBD16F858DC152A7ULL, "vxht_0_hightown_a3_a86_0_0.vlm"},
	{0xBBD1965EB55C3AB8ULL, "gen_im_arm_cht_lgt_act2_01.uti"},
	{0xBBD204D32647C3FDULL, "325429871.ogg"},
	{0xBBD2686A5BDE6423ULL, "vxlt_0_lowtown_docks_a3_n_d1t_0_0.vlm"},
	{0xBBD2EDC0836AFD76ULL, "vxw_encounter1_ja_0_0.vlm"},
	{0xBBD314B87FC19183ULL, "w_woundedcoast_177_0.mao"},
	{0xBBD318B87FC198D0ULL, "w_woundedcoast_177_0.mmh"},
	{0xBBD326B87FC1B0EAULL, "w_woundedcoast_177_0.msh"},
	{0xBBD334990F7B017EULL, "622492692.ogg"},
	{0xBBD36B63D65EDE70ULL, "art\\vfx\\cinematic\\vfx_lyrium_glow_c.mmh"},
	{0xBBD4348545DFF8C9ULL, "6107537_m.fxe"},
	{0xBBD4680ADAE21405ULL, "609038939.ogg"},
	{0xBBD475BDAD7AA607ULL, "rune_armor_cold.dds"},
	{0xBBD4B9FA16397D4AULL, "355968547.ogg"},
	{0xBBD4D20F46A85BCFULL, "652289733.ogg"},
	{0xBBD50B032457FC26ULL, "follower_banter6143247.cl"},
	{0xBBD54623EC9AF961ULL, "448716201.ogg"},
	{0xBBD57C48A0179191ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03a.mmh"},
	{0xBBD59648A017BE4FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos03a.msh"},
	{0xBBD5BEB56DF05302ULL, "gen00fl_fenris6128072.cl"},
	{0xBBD62180815B11BFULL, "qcr220_grey_wardens6060435.cl"},
	{0xBBD647D9E701F48AULL, "vxht_mansion2b_3g1_0_0.vlm"},
	{0xBBD6A383BD783480ULL, "vxca_cavern1_haunted_dq7_0_0.vlm"},
	{0xBBD77E3F4DE97339ULL, "6149056_m.fxe"},
	{0xBBD792C79831A695ULL, "1003143232.ogg"},
	{0xBBD7BB0DE9442D81ULL, "vrc221_bartrand6043386.cl"},
	{0xBBD7CCC1769E3D63ULL, "mag100_thrask6034822.cl"},
	{0xBBD7ED7C02A858D5ULL, "6195848_m.fxe"},
	{0xBBD824BE4FB6024CULL, "llac_ht_keep_3739_1.dds"},
	{0xBBD8CAE5BFCFF14EULL, "vxdr_primevalentrance_1m8_0_0.vlm"},
	{0xBBD92531ACE5C5C5ULL, "qcr000_arishok6059317.cl"},
	{0xBBD97FFD53A534E5ULL, "gen00fl_aveline6122791.cl"},
	{0xBBD9BF4EF4B40567ULL, "vxw_encounter2_ju_0_0.vlm"},
	{0xBBDA0B5C32DE3A97ULL, "qun120cr_javaris6042983.cl"},
	{0xBBDA57DD5D69D225ULL, "m_one121ar_warehouse.wbd"},
	{0xBBDAB2E1850485C8ULL, "630107416.ogg"},
	{0xBBDAF1B62C295939ULL, "6137520_f.fxe"},
	{0xBBDB0B4D74D6C81CULL, "l_isa311ar_castillon_base.wbd"},
	{0xBBDB2FBB82D3F892ULL, "art\\vfx\\environment\\vfx_lyrium_aurora.dura.evt"},
	{0xBBDB8E9EADF4688DULL, "278054656.ogg"},
	{0xBBDB93F6634FC997ULL, "9951472.ogg"},
	{0xBBDBEC6047BD6E0DULL, "ts_generic_prog_junk1.gda"},
	{0xBBDC41CB3147CE4BULL, "6097133_m.fxe"},
	{0xBBDC62AF63102DE6ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drgb_0d.dds"},
	{0xBBDC63DC56072267ULL, "hlac_ca_cavern1_haunted_24425_4x.dds"},
	{0xBBDC7692A46FBE93ULL, "vxht_chantry_n_12n_0_0.vlm"},
	{0xBBDD959274195C42ULL, "vxht_keep_n_6mm_0_0.vlm"},
	{0xBBDDD875D555A5EDULL, "462113965.ogg"},
	{0xBBDDE245522BF24FULL, "739079156.ogg"},
	{0xBBDE4C084F44631EULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_cracked01_d.dds"},
	{0xBBDE67F7C68E0072ULL, "hlac_ca_caverns2prc_5802_19z.dds"},
	{0xBBDEA99B58782C57ULL, "mer240_jansen6050321.cl"},
	{0xBBDEEDC92C47E883ULL, "dae011_wench_mob6149860.cl"},
	{0xBBDF1097F79A6259ULL, "6163337_m.fxe"},
	{0xBBDF4D6AA685E16FULL, "vxca_caverns2prc_ab7_0_0.vlm"},
	{0xBBDFE8AB72A3C05BULL, "vxdr_0_deeproads_cv8_0_0.vlm"},
	{0xBBE0670C8FC5C279ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_loop.ani"},
	{0xBBE08D30CB15E737ULL, "qun120_arishok.bnk"},
	{0xBBE0D778B82E2C8CULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_001a_0n.dds"},
	{0xBBE12AF0B7099983ULL, "vxlt_0_lowtown_a2_ekg_0_1.vlm"},
	{0xBBE1811E8CF8A588ULL, "vxlt_gallowscourtyard_5a6_0_0.vlm"},
	{0xBBE18E63D66B568FULL, "art\\vfx\\cinematic\\vfx_lyrium_glow_c.anb"},
	{0xBBE196C11E776804ULL, "6111359_m.fxe"},
	{0xBBE233119012E327ULL, "6112995_m.fxe"},
	{0xBBE24EBB82DA4C25ULL, "art\\vfx\\environment\\vfx_lyrium_aurora.dura.gad"},
	{0xBBE277E17E9DC018ULL, "67651960.ogg"},
	{0xBBE2B3CC97C099B0ULL, "vxca_cavern1_haunted_dph_0_0.vlm"},
	{0xBBE2E79CF9FF5788ULL, "vxht_mansion2_320_0_0.vlm"},
	{0xBBE2F61DFEB530BAULL, "vxlt_0_lowtown_n_e34_0_0.vlm"},
	{0xBBE30A3D18FB27F0ULL, "459533102.ogg"},
	{0xBBE33492C300522CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upper.phy"},
	{0xBBE3C97988A2F36BULL, "hlac_lt_0_lowtown_0_6z.dds"},
	{0xBBE4517A7278E024ULL, "vxdr_0_deeproads_d0v_0_0.vlm"},
	{0xBBE46376CDD04025ULL, "vxlt_gallowscourtyard_470_0_0.vlm"},
	{0xBBE46F30CB199D4FULL, "qun120_arishok.cnv"},
	{0xBBE47DED510AF03EULL, "4615514.ogg"},
	{0xBBE4E5D1DAFB0429ULL, "vxht_0_hightown_c3b_0_0.vlm"},
	{0xBBE51038BDF3B62DULL, "gen00fl_anders6116135.cl"},
	{0xBBE5185714995137ULL, "vxlt_0_lowtown_a2_n_c_i06_0_0.vlm"},
	{0xBBE52FF82BF2F2B6ULL, "vxht_0_hightown_9x6_0_0.vlm"},
	{0xBBE5833B38172A57ULL, "art\\characters\\weapons\\longswords\\w_lsw_018a.phy"},
	{0xBBE5932CA92FE840ULL, "6099709_m.fxe"},
	{0xBBE5E811790FF433ULL, "vxlt_warehouse_n_1ea_0_0.vlm"},
	{0xBBE6045DE2ABF33CULL, "298326361.ogg"},
	{0xBBE64A0CA498B3A1ULL, "dre000_hireling16038603.cl"},
	{0xBBE6A7501CCCD987ULL, "307829608.ogg"},
	{0xBBE6C32E2FE7EE62ULL, "6172668_m.fxe"},
	{0xBBE7132BF169913FULL, "vxlt_0_lowtown_docks_n_b65_0_1.vlm"},
	{0xBBE7420C8FCBA1BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_loop.gad"},
	{0xBBE79E31F4E06B89ULL, "dre000_stroud6036694.cl"},
	{0xBBE7A37433BCCC0BULL, "mer210_smith6091996.cl"},
	{0xBBE80C9D5255F21AULL, "hlac_ht_0_hightown_a2_-131072_1y.dds"},
	{0xBBE8836B7E32B4DCULL, "452240564.ogg"},
	{0xBBE92F6C1CF17C3CULL, "zz_dae_debug6029501.cl"},
	{0xBBE93B62F44DB64BULL, "533520885.ogg"},
	{0xBBE95FAAAC60B83DULL, "dae111st_dog_gamlen.stg"},
	{0xBBE9C9FB2F6F668BULL, "art\\characters\\weapons\\longswords\\w_lsw_014a.phy"},
	{0xBBEAC61A7F6E12B3ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.anb"},
	{0xBBEB11F9743E98B9ULL, "904230597.ogg"},
	{0xBBEBE23B8D03261FULL, "836117256.ogg"},
	{0xBBEC076F660F74C2ULL, "vxht_keep_n_6hn_0_0.vlm"},
	{0xBBEC4790A1388816ULL, "893482978.ogg"},
	{0xBBEC86CC415D6BC6ULL, "6127547_m.fxe"},
	{0xBBEC8B79DC27A634ULL, "vxlt_0_lowtown_n_byw_0_0.vlm"},
	{0xBBED55A767B104D3ULL, "lgt160_ntb_elf_conflict6177451.cl"},
	{0xBBED7856DD91B69CULL, "w_blightlands_20_0.mao"},
	{0xBBED8456DD91CB3FULL, "w_blightlands_20_0.mmh"},
	{0xBBED8A56DD91D5E1ULL, "w_blightlands_20_0.msh"},
	{0xBBEDE19C66C8B0D5ULL, "759339298.ogg"},
	{0xBBEDE52CD58D3C54ULL, "llac_lt_undercityencounter_-1_39.dds"},
	{0xBBEE3222CFDA1C50ULL, "lgt331cr_shw_bmage_lead1.utc"},
	{0xBBEE4A820F0F1D58ULL, "llac_dr_slavers_3216_11.dds"},
	{0xBBEEB763438A2F7EULL, "dae150_chandan6148805.cl"},
	{0xBBEF2667C5A50B09ULL, "gen_im_npc_dwf_arm_glv_lgtb.uti"},
	{0xBBEF2D093F7D4A07ULL, "vxdr_primevaltomb_6j3_0_0.vlm"},
	{0xBBEF6A52FA9D4351ULL, "6027720_f.fxe"},
	{0xBBEFCDF595CB2D57ULL, "6167043_m.fxe"},
	{0xBBF003518CCEA2E9ULL, "follower_barks6153262.cl"},
	{0xBBF00743B1812252ULL, "vxca_caverns2_7y8_0_0.vlm"},
	{0xBBF092DCDEC4217DULL, "dae000_surprises6157193.cl"},
	{0xBBF0C3E65DB238CCULL, "vxht_mansion_n_3hi_0_0.vlm"},
	{0xBBF129625012E9BDULL, "mrl150_merrill6084471.cl"},
	{0xBBF177D19D8F28AEULL, "dre000_dougal6129752.cl"},
	{0xBBF1A508D7AE40C6ULL, "cry_val_cloud.mao"},
	{0xBBF259D11BECDEE3ULL, "zz_followers_banter6158475.cl"},
	{0xBBF2F8AFC1F81C1BULL, "vxht_0_hightown_n_amc_0_0.vlm"},
	{0xBBF357AECCF11CA9ULL, "vxht_0_hightown_c6a_0_0.vlm"},
	{0xBBF39E87A34F8A82ULL, "433568908.ogg"},
	{0xBBF3B627115BEBB8ULL, "hlac_lt_0_lowtown_docks_a3_-131071_0y.dds"},
	{0xBBF3C668075509C9ULL, "6098303_m.fxe"},
	{0xBBF3CF744A996D7BULL, "6065907_m.fxe"},
	{0xBBF40154493A75BEULL, "114944335.ogg"},
	{0xBBF42CEB66D9BA7CULL, "118012140.ogg"},
	{0xBBF4B21F8573CF3EULL, "vfx0_cs_addv_alt.vsh.fxo"},
	{0xBBF4BC020B963E61ULL, "hm_osric.mor"},
	{0xBBF555C33493F71DULL, "hla_ht_0_hightown_131073_1y.dds"},
	{0xBBF56164F772B757ULL, "109033214.ogg"},
	{0xBBF5725EB68F12A3ULL, "6067663_f.fxe"},
	{0xBBF5A7BCF1F2E4ACULL, "15628686.ogg"},
	{0xBBF6347088883464ULL, "hlac_lt_0_lowtown_-65535_2z.dds"},
	{0xBBF6909F6ED413AFULL, "rdr110dg_guard_sleeping.bnk"},
	{0xBBF695A2CD383823ULL, "ht_0_hightown_a2_42_0_lc.mmh"},
	{0xBBF69C1AC1975E71ULL, "vxlt_0_lowtown_n_h4w_0_0.vlm"},
	{0xBBF6A308D8A03667ULL, "vxw_encounter2_pw_0_0.vlm"},
	{0xBBF778130D2E95D1ULL, "187499984.ogg"},
	{0xBBF7E31A7F78CE1CULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.mmh"},
	{0xBBF90BCD89E0516AULL, "775084041.ogg"},
	{0xBBF90F38A647FDB5ULL, "6177691_m.fxe"},
	{0xBBF98B9F909AE3FDULL, "isa111cr_thug_1.utc"},
	{0xBBF9B29F6ED68317ULL, "rdr110dg_guard_sleeping.cnv"},
	{0xBBF9EE11613CC40BULL, "dae000_mother6099171.cl"},
	{0xBBF9FF29640ECA2EULL, "art\\vfx\\combat\\vfx_m_ent_horror_act_c.mmh"},
	{0xBBFAD152E06525C6ULL, "hlac_lt_0_lowtown_docks_-196610_0z.dds"},
	{0xBBFB1784E517B216ULL, "vxca_caverns1_mines_hhz_0_0.vlm"},
	{0xBBFB38E490EE1488ULL, "seb321_leliana6181825.cl"},
	{0xBBFB738581927456ULL, "mag110_vincento6017664.cl"},
	{0xBBFBB3DCE05F190EULL, "vxht_mansion_n_6rd_0_0.vlm"},
	{0xBBFCACEA44530503ULL, "vxht_mansion2_2o5_0_0.vlm"},
	{0xBBFCB1C4364580D2ULL, "vxlt_0_lowtown_a3_n_c_f4m_0_0.vlm"},
	{0xBBFCEA4E3FEDD561ULL, "6032066_m.fxe"},
	{0xBBFD1064C95642ADULL, "248945915.ogg"},
	{0xBBFD2B652F1C0296ULL, "686256186.ogg"},
	{0xBBFDB4C4FE7A016FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chanterb_d.dds"},
	{0xBBFDCA6F8B0AE2A9ULL, "464235041.ogg"},
	{0xBBFE3F11F145BE22ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_zevr.phy"},
	{0xBBFE73779AADF86DULL, "hlac_lt_hangedman_n_-1_16z.dds"},
	{0xBBFED64DA1D9FFF1ULL, "plt_mrl200pt_hunteramulets.nss"},
	{0xBBFF24A48ACE3899ULL, "mer210_seldrek6078806.cl"},
	{0xBBFF6BD54041CA26ULL, "vxht_mansion2_1l2_0_0.vlm"},
	{0xBBFF8E047B16B3A8ULL, "761586553.ogg"},
	{0xBBFFD001442BDA68ULL, "vxlt_hangedman_n_170_0_0.vlm"},
	{0xBC008A7D1DE21393ULL, "6188609_f.fxe"},
	{0xBC00B02899C3046EULL, "191092370.ogg"},
	{0xBC0171344A117DA8ULL, "679385435.ogg"},
	{0xBC0217CFB5F7BDD5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_5.pwk"},
	{0xBC0262B8188EAB6AULL, "gen00fl_carver6097233.cl"},
	{0xBC02AAE1D6E56AC8ULL, "follower_banter6171060.cl"},
	{0xBC02E0133F36D30EULL, "art\\characters\\weapons\\greatswords\\w_gsw_003a.msh"},
	{0xBC02EE133F36EB5FULL, "art\\characters\\weapons\\greatswords\\w_gsw_003a.mao"},
	{0xBC02FA133F36FFFCULL, "art\\characters\\weapons\\greatswords\\w_gsw_003a.mmh"},
	{0xBC0332C171A6099DULL, "vxlt_warehouse_n_ap_0_0.vlm"},
	{0xBC034D2F02FC592BULL, "1041211656.ogg"},
	{0xBC035667780F1FD1ULL, "gen_im_trs_qun_two_2ha_r0.uti"},
	{0xBC035D09CAF9EA04ULL, "6087481_m.fxe"},
	{0xBC039F20F1557CAFULL, "6166282_m.fxe"},
	{0xBC041B37BFFC4A0FULL, "follower_banter6170053.cl"},
	{0xBC04601A1EAB41ECULL, "6125389_m.fxe"},
	{0xBC04B09A595EE1AEULL, "ts_kirk_hightown_prog_junk3.gda"},
	{0xBC04CA65F22D082BULL, "vxht_bartrand_n_6j2_0_0.vlm"},
	{0xBC04D68C7230A29BULL, "6075487_m.fxe"},
	{0xBC057EFBE112EA39ULL, "llac_ht_chantry_n_2534_27.dds"},
	{0xBC05CA41E9A91E16ULL, "342457447.ogg"},
	{0xBC05D73046D4F003ULL, "349638758.ogg"},
	{0xBC0689E25A3EFAD8ULL, "451064332.ogg"},
	{0xBC07094D74FBFF96ULL, "l_isa311ar_castillon_base.bnk"},
	{0xBC076CC28CD9D183ULL, "6034062_m.fxe"},
	{0xBC076DF1B6E974E9ULL, "901571662.ogg"},
	{0xBC07B63ED6990BEBULL, "vxdr_slavers_146_0_0.vlm"},
	{0xBC07DE0618B79C84ULL, "vxht_0_hightown_a2_c1q_0_0.vlm"},
	{0xBC081E29641B3C2DULL, "art\\vfx\\combat\\vfx_m_ent_horror_act_c.anb"},
	{0xBC082168FA0467F7ULL, "llac_w_woundedcoast_-65530_0.dds"},
	{0xBC0839F0C1C00E69ULL, "6128022_m.fxe"},
	{0xBC088EB97CE11358ULL, "vxca_caverns2_8g7_0_0.vlm"},
	{0xBC08A3D95D67AA0FULL, "602030826.ogg"},
	{0xBC08EC9C15E9F289ULL, "sib111_mother6098968.cl"},
	{0xBC095DA6C176AE44ULL, "mag311_nyssa6051214.cl"},
	{0xBC097573BB86DC4FULL, "6090877_m.fxe"},
	{0xBC09A7B2C4EB484BULL, "hlac_lt_gallowstemplar_2392_1x.dds"},
	{0xBC09B099BF7A2D06ULL, "mcr371ip_to_prison.utp"},
	{0xBC09BA7813889857ULL, "llac_lt_0_lowtown_a2_-3_0.dds"},
	{0xBC0A577B7A5E9A55ULL, "73876871.ogg"},
	{0xBC0AAB70EFBCA2EEULL, "37218940.ogg"},
	{0xBC0ACA8B454FDB24ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.dura.ani"},
	{0xBC0B0F286E73FF7CULL, "llac_ht_0_hightown_a2_-65537_4.dds"},
	{0xBC0B5F83E182A054ULL, "345827452.ogg"},
	{0xBC0B617B3A0E5797ULL, "vxw_bonepitmining_a2_7t_0_0.vlm"},
	{0xBC0BA242500AC251ULL, "gen00fl_anders6119105.cl"},
	{0xBC0BD673501DB28DULL, "vxw_encounter2_5v_0_0.vlm"},
	{0xBC0C6830420532B5ULL, "vxht_chantry_2v0_0_0.vlm"},
	{0xBC0C8F63D017242DULL, "nat300ip_item_chest.utp"},
	{0xBC0CB4D92FE6D05BULL, "vxlt_0_lowtown_a2_n_byh_0_0.vlm"},
	{0xBC0CC6418D737C82ULL, "503537588.ogg"},
	{0xBC0D47E227C7326BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\4sc_basic.phy"},
	{0xBC0DBBBEDD37B904ULL, "zz_t_trailer_carver.cut"},
	{0xBC0DF29F6AE9F520ULL, "6065287_m.fxe"},
	{0xBC0E904D54F40CA2ULL, "gen00fl_varric6076454.cl"},
	{0xBC0F67E933C7B961ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_c.impa.ani"},
	{0xBC0FA1DB9FD3A27AULL, "6141781_m.fxe"},
	{0xBC0FAB239685F601ULL, "95286889.ogg"},
	{0xBC105C7EA9E3819BULL, "hlac_lt_0_lowtown_docks_a3_-196609_0z.dds"},
	{0xBC10C78B45544186ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.dura.gad"},
	{0xBC10E9C4B8BC6783ULL, "vxw_bonepitdestroyed_a3_17a_0_0.vlm"},
	{0xBC114DA5AE9FE64AULL, "1041612021.ogg"},
	{0xBC1189A432F497D9ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em06.msh"},
	{0xBC1193A432F4A8C7ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em06.mmh"},
	{0xBC11B3158D0307DAULL, "6171588_m.fxe"},
	{0xBC11FA45F6793602ULL, "vxlt_0_lowtown_a3_n_c_ep2_0_0.vlm"},
	{0xBC1217E0C3705535ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas_sa.dds"},
	{0xBC129388A51C0163ULL, "follower_barks6169904.cl"},
	{0xBC1319BFE12F21BDULL, "vxht_0_hightown_b5y_0_0.vlm"},
	{0xBC131B21BC7DE13BULL, "art\\levels\\races\\elf\\props\\_textures\\prp_d_ink_sa.dds"},
	{0xBC133B992EE34460ULL, "lgt250_way_dalish_conflict6147140.cl"},
	{0xBC1358E7861A074EULL, "vrc321_varric6094307.cl"},
	{0xBC13C87E99E0776BULL, "380391520.ogg"},
	{0xBC144CAB3C80D5F8ULL, "vxw_woundedcoast_1dh_0_0.vlm"},
	{0xBC148E41E943059AULL, "vxdr_0_deeproads_9ch_0_0.vlm"},
	{0xBC14C40004189E31ULL, "hlac_ca_caverns1_24425_3z.dds"},
	{0xBC150FC4475F1231ULL, "6041849_f.fxe"},
	{0xBC15220629617EC8ULL, "isa211_isabela6081985.cl"},
	{0xBC153E3863D84AAEULL, "qun111_ketojan6040666.cl"},
	{0xBC15B481E2C3432EULL, "mag101tr_aom_firstfight1.ncs"},
	{0xBC15DE9F7ECFEA03ULL, "6031362_m.fxe"},
	{0xBC16243D99CBFBAEULL, "qun231cr_hideout_qunari.utc"},
	{0xBC16469BAF2F76DCULL, "llac_lt_0_lowtown_docks_a3_65536_0.dds"},
	{0xBC1672E933CDEA03ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_c.impa.gad"},
	{0xBC16B51F830A0CFFULL, "vxlt_darktown_585_0_1.vlm"},
	{0xBC1700966ED72BD8ULL, "vxca_caverns2_8b4_0_0.vlm"},
	{0xBC170BD49A96B70FULL, "vxlt_warehouse_n_146_0_0.vlm"},
	{0xBC170EA4D2D52ECCULL, "vxlt_hangedman_n_1qr_0_0.vlm"},
	{0xBC17C28B455A5865ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.dura.evt"},
	{0xBC17F8848443C66DULL, "hlac_lt_0_lowtown_n_-65535_6x.dds"},
	{0xBC186A3EAD53BBA9ULL, "123803251.ogg"},
	{0xBC188FFC9B4FB248ULL, "seb321pt_faith.ncs"},
	{0xBC18A19EEB3FA489ULL, "gen00fl_varric6093573.cl"},
	{0xBC18BD7D9E07FC22ULL, "fen210_hadriana6110517.cl"},
	{0xBC18CC4FA1E240CBULL, "hlac_lt_0_lowtown_docks_a3_0_4x.dds"},
	{0xBC18CD87DF42EA00ULL, "gen_im_acc_rng_act1_05.uti"},
	{0xBC197118C3ED4A1AULL, "hlac_dr_0_deeproads_17038_1x.dds"},
	{0xBC1A6C59E83A9756ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medc.msh"},
	{0xBC1A7659E83AA854ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medc.mmh"},
	{0xBC1A7A59E83AAF67ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medc.mao"},
	{0xBC1B01C3CD129A57ULL, "413303555.ogg"},
	{0xBC1B2FC308C85F9AULL, "war_staminadrain.dds"},
	{0xBC1B31DD5DA128CFULL, "m_one121ar_warehouse.bnk"},
	{0xBC1B37D1679DEE6DULL, "one140cr_charade6134590.cl"},
	{0xBC1BAD02CED1915EULL, "6120902_m.fxe"},
	{0xBC1BE7FFE91C1BA6ULL, "art\\vfx\\combat\\vfx_d_invulner_exit_c.impa.evt"},
	{0xBC1BE8291036F851ULL, "mcr370_negotiation6080218.cl"},
	{0xBC1C1E9D9E26A4E2ULL, "283577019.ogg"},
	{0xBC1C891FC803024EULL, "6207812_m.fxe"},
	{0xBC1C99E75FE1E0E1ULL, "vxlt_hangedman_28h_0_0.vlm"},
	{0xBC1C9FA4E8DBCF8CULL, "vxht_0_hightown_a3_b1a_0_0.vlm"},
	{0xBC1D29B8E072380BULL, "vxlt_undercityencounter_r3_0_0.vlm"},
	{0xBC1D4FE933D3CDE8ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_c.impa.evt"},
	{0xBC1DC0F6465D451CULL, "499754635.ogg"},
	{0xBC1DF7A1D446D27BULL, "542172345.ogg"},
	{0xBC1DF8C8E233D143ULL, "plt_cod_itm_shield_aveline.plo"},
	{0xBC1E757476B1FA08ULL, "802726821.ogg"},
	{0xBC1EC75AB4A50820ULL, "nat380_nathaniel6090037.cl"},
	{0xBC1F295FE0CEFF13ULL, "plt_vis100pt_offer.plo"},
	{0xBC1F33E1B8B8ED0FULL, "vxdr_primevaltomb_2sq_0_1.vlm"},
	{0xBC1F4A3C8F2D9869ULL, "vxht_keep_n_6af_0_0.vlm"},
	{0xBC20BCB9A63E77F6ULL, "6097859_m.fxe"},
	{0xBC212EE3EB3C7EE2ULL, "art\\vfx\\cinematic\\vfx_merrill_reflect.impa.gad"},
	{0xBC21FEFFE920AEF1ULL, "art\\vfx\\combat\\vfx_d_invulner_exit_c.impa.gad"},
	{0xBC22117C0017C5DBULL, "6096048_m.fxe"},
	{0xBC224321391EF383ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.dura.evt"},
	{0xBC227F406201C477ULL, "6109696_m.fxe"},
	{0xBC22F3BBADA038E3ULL, "w_bonepitmining_a2_30_0.msh"},
	{0xBC23001EDB391E1AULL, "vxdr_0_deeproads_9di_0_0.vlm"},
	{0xBC2301BBADA0503DULL, "w_bonepitmining_a2_30_0.mmh"},
	{0xBC2305BBADA0574EULL, "w_bonepitmining_a2_30_0.mao"},
	{0xBC23437520907157ULL, "llac_lt_0_lowtown_docks_131073_0.dds"},
	{0xBC2365A6D1427BCCULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun_3.mao"},
	{0xBC23FE98139A5B3AULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_004a_0s.dds"},
	{0xBC240B92C337A3F7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upper.mmh"},
	{0xBC240F92C337AA04ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upper.mao"},
	{0xBC2419E347B3DD38ULL, "6055154_m.fxe"},
	{0xBC242192C337C969ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_upper.msh"},
	{0xBC242611F1666375ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_zevr.mmh"},
	{0xBC242811F16666EBULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_zevr.msh"},
	{0xBC2506E396D1A040ULL, "vxlt_0_lowtown_docks_a3_ctl_0_0.vlm"},
	{0xBC25691DB47177D8ULL, "vxht_0_hightown_bf1_0_0.vlm"},
	{0xBC257DB18C8CCF8FULL, "vxlt_warehouse_n_191_0_0.vlm"},
	{0xBC2585B45AC43C73ULL, "vxht_brothel_n_6e7_0_0.vlm"},
	{0xBC25FF652F6F2CDEULL, "vxht_0_hightown_a3_ccl_0_0.vlm"},
	{0xBC2608689B4422DEULL, "191333161.ogg"},
	{0xBC264E2E29CA8FF3ULL, "338986211.ogg"},
	{0xBC266E0F8B25F926ULL, "vxlt_gallowscourtyard_4wg_0_0.vlm"},
	{0xBC26A4A388332209ULL, "23799786.ogg"},
	{0xBC26C5C46C135E49ULL, "vxca_cave_18y_0_0.vlm"},
	{0xBC26E798FC0017FAULL, "lgt000cr_hate_gen_thug.utc"},
	{0xBC2700BEC60B6EC3ULL, "212589893.ogg"},
	{0xBC2751E3EB4125D9ULL, "art\\vfx\\cinematic\\vfx_merrill_reflect.impa.evt"},
	{0xBC27A163F88BDA52ULL, "119515152.ogg"},
	{0xBC27B56B2C315BFEULL, "6095808_m.fxe"},
	{0xBC2819DE8B9D2F5EULL, "717500121.ogg"},
	{0xBC28BCB22B607C15ULL, "6167657_f.fxe"},
	{0xBC28D9133F5796C7ULL, "art\\characters\\weapons\\greatswords\\w_gsw_003a.phy"},
	{0xBC28EDCB553FB04FULL, "6094724_f.fxe"},
	{0xBC28FFFFE926CE9FULL, "art\\vfx\\combat\\vfx_d_invulner_exit_c.impa.ani"},
	{0xBC291C213924D004ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.dura.gad"},
	{0xBC2952FC96E0BAF6ULL, "6144459_f.fxe"},
	{0xBC29B3538F1204DAULL, "vxca_caverns2_7u4_0_0.vlm"},
	{0xBC29DE0E06AAA83DULL, "hlac_ht_brothel_n_3367_0y.dds"},
	{0xBC2A7B378DC44D89ULL, "hlac_lt_0_lowtown_a2_n_65538_4z.dds"},
	{0xBC2B0BC451D7F961ULL, "vxlt_hangedman_25o_0_0.vlm"},
	{0xBC2B39B7F3880120ULL, "6164706_m.fxe"},
	{0xBC2B7CB54F128213ULL, "6048286_f.fxe"},
	{0xBC2B8B5919956DAEULL, "213056858.ogg"},
	{0xBC2B9572286F936AULL, "vxht_0_hightown_n_cba_0_0.vlm"},
	{0xBC2C35E1D52142DFULL, "774759011.ogg"},
	{0xBC2C76D80FF184FCULL, "6129607_f.fxe"},
	{0xBC2CA15E025949BDULL, "297069086.ogg"},
	{0xBC2D00F002EF4EFEULL, "vfx_temp1.mao"},
	{0xBC2D583B385496AFULL, "art\\characters\\weapons\\longswords\\w_lsw_018a.mao"},
	{0xBC2D643B3854AB4CULL, "art\\characters\\weapons\\longswords\\w_lsw_018a.mmh"},
	{0xBC2D6A3B3854B53EULL, "art\\characters\\weapons\\longswords\\w_lsw_018a.msh"},
	{0xBC2D88E48FD5F12CULL, "vxht_mansion2b_1bn_0_0.vlm"},
	{0xBC2DFD0C138723B5ULL, "character_character_p248.psh.fxo.dx11"},
	{0xBC2E306C77DF9B0FULL, "plt_cod_hst_kirk_arch_1a.plo"},
	{0xBC2F03253A4BDAF1ULL, "hlac_lt_gallowstemplar_f_2383_10y.dds"},
	{0xBC2F2125FB3AD7F7ULL, "dae000_adriano.bnk"},
	{0xBC2FCBC2CB9607C0ULL, "186681460.ogg"},
	{0xBC302B21392B08E2ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.dura.ani"},
	{0xBC3054A9A83879D8ULL, "6139159_m.fxe"},
	{0xBC30BE1EC5AF5631ULL, "gen00fl_sebastian6188630.cl"},
	{0xBC31457C85BA4EF6ULL, "mer320_hubert6096168.cl"},
	{0xBC31791D90BFA489ULL, "gen_im_cft_reg_k2.uti"},
	{0xBC31835A66BB834DULL, "203766360.ogg"},
	{0xBC3185755520067EULL, "vxht_chantry_2e6_0_0.vlm"},
	{0xBC320325FB3CDA0FULL, "dae000_adriano.cnv"},
	{0xBC3248B9E88F6135ULL, "gen00fl_anders6128343.cl"},
	{0xBC3259803BDEC21FULL, "6118597_m.fxe"},
	{0xBC32C76BF519D679ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_1x.dds"},
	{0xBC32E839D8039C5DULL, "art\\levels\\ztools\\default\\generic\\9ff_blocker.phy"},
	{0xBC331C6BAD489020ULL, "195529081.ogg"},
	{0xBC337102F85C75FFULL, "vxlt_warehouse_n_15u_0_0.vlm"},
	{0xBC3374ACF1ED9297ULL, "follower_banter6172763.cl"},
	{0xBC33B662C2C98216ULL, "vxht_0_hightown_a3_c89_0_0.vlm"},
	{0xBC340655D9D6EE4EULL, "llac_lt_hangedman_-1_14.dds"},
	{0xBC343FD4D4446D80ULL, "hlac_ht_bartrand_n_3367_1y.dds"},
	{0xBC346358118A63ADULL, "art\\levels\\races\\elf\\props\\_textures\\prp_aravell_n.dds"},
	{0xBC3471422165455EULL, "vxht_0_hightown_a3_cdm_0_0.vlm"},
	{0xBC34947A132CA548ULL, "vxlt_hangedman_25j_0_0.vlm"},
	{0xBC3559E3EB4D70C8ULL, "art\\vfx\\cinematic\\vfx_merrill_reflect.impa.ani"},
	{0xBC3697702F81C6ADULL, "gen_im_cft_rec_arn_03.uti"},
	{0xBC36B7744E20EFFFULL, "hlac_ca_caverns1_24424_6x.dds"},
	{0xBC36D9F63F877BBBULL, "vxlt_gallowstemplar_f_3r3_0_0.vlm"},
	{0xBC36E57FEFEDC62AULL, "328198849.ogg"},
	{0xBC36ED2D71E4E4E7ULL, "6024553_m.fxe"},
	{0xBC3704502B77828BULL, "vxht_chantry_n_1yg_0_0.vlm"},
	{0xBC374245FA44EB78ULL, "6113291_f.fxe"},
	{0xBC38909354E39BCCULL, "293077185.ogg"},
	{0xBC38FAAA38A0BF83ULL, "106193274.ogg"},
	{0xBC393748E8847380ULL, "328741076.ogg"},
	{0xBC396C579141A9B5ULL, "mcr371_orsino6083330.cl"},
	{0xBC397DA143CE11E1ULL, "vxlt_hangedman_22n_0_0.vlm"},
	{0xBC39B2BC0FD6546FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinsarc.mao"},
	{0xBC39BEBC0FD6680CULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinsarc.mmh"},
	{0xBC39C4BC0FD672FEULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03p1ruinsarc.msh"},
	{0xBC39CEBE958BA55EULL, "6069379_m.fxe"},
	{0xBC39EE9241089A93ULL, "6043386_f.fxe"},
	{0xBC39F34C1C3D6382ULL, "gen00fl_carver6108915.cl"},
	{0xBC3A2F95C0A18923ULL, "art\\characters\\creatures\\ragedemon\\animation\\casting.ani"},
	{0xBC3A342DE0697C18ULL, "vxlt_gallowstemplar_f_5qw_0_0.vlm"},
	{0xBC3AEC078558AB52ULL, "792060187.ogg"},
	{0xBC3BBB4F10AA0480ULL, "6034209_m.fxe"},
	{0xBC3BD2D174E22911ULL, "vxw_encounter2_fg_0_0.vlm"},
	{0xBC3BED4DA20DD4DAULL, "plt_mrl200pt_hunteramulets.plo"},
	{0xBC3C068F3065A5DAULL, "hlac_dr_0_deeproads_17039_2z.dds"},
	{0xBC3C493312AFB53CULL, "61620706.ogg"},
	{0xBC3C5708441A4754ULL, "em_ilen.mor"},
	{0xBC3C78D6B0FA6A46ULL, "llac_lt_0_lowtown_a2_n_c_-65537_3.dds"},
	{0xBC3CC331D68EC3DFULL, "344713265.ogg"},
	{0xBC3CD9661DCD312FULL, "6052170_m.fxe"},
	{0xBC3CE3BFE44028FDULL, "vxdr_primevalentrance_2oh_0_2.vlm"},
	{0xBC3D40968152604DULL, "228698918.ogg"},
	{0xBC3DA192989D5137ULL, "nat380_temmerin6067387.cl"},
	{0xBC3E7CA4331B6FFCULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em06.phy"},
	{0xBC3EB66008098295ULL, "6083964_f.fxe"},
	{0xBC3EE7BDA8142309ULL, "vxlt_0_lowtown_a2_n_hid_0_0.vlm"},
	{0xBC3F1FCD4D111D47ULL, "blood_conj_07_l3.mao"},
	{0xBC3FCB369529ADF4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp2a.phy"},
	{0xBC3FF77D0EF3103BULL, "vxlt_0_lowtown_docks_n_b8z_0_0.vlm"},
	{0xBC4063BE08152038ULL, "vxdr_0_deeproads_ckd_0_0.vlm"},
	{0xBC409813186039C5ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_zbm_0n.dds"},
	{0xBC40A8146936B44AULL, "fen210_fenris6108744.cl"},
	{0xBC40BD8DE413E043ULL, "6080195_m.fxe"},
	{0xBC40CB5D2DD4DA9FULL, "6117497_m.fxe"},
	{0xBC4126C9044FF76CULL, "w_bonepitdestroyed_a3_88_0.msh"},
	{0xBC412CC90450015EULL, "w_bonepitdestroyed_a3_88_0.mmh"},
	{0xBC4138C9045015F5ULL, "w_bonepitdestroyed_a3_88_0.mao"},
	{0xBC41573775393F62ULL, "vxlt_gallowstemplar_f_3nj_0_0.vlm"},
	{0xBC41851698E751AAULL, "6127976_m.fxe"},
	{0xBC419CA977DF2328ULL, "6125419_m.fxe"},
	{0xBC41A879F4EFE37CULL, "hlac_lt_0_lowtown_a2_n_c_131073_1z.dds"},
	{0xBC41BA3F32776FC2ULL, "6119196_f.fxe"},
	{0xBC41D31A4893FA31ULL, "dae000_hate_follower_react6103227.cl"},
	{0xBC42426E5E5C756BULL, "vxca_cave_27i_0_0.vlm"},
	{0xBC42696E3EB06D73ULL, "vxht_brothel_n_6k9_0_0.vlm"},
	{0xBC42A43CB8E88785ULL, "vxw_woundedcoast_1dq_0_0.vlm"},
	{0xBC42BF2EE78613B4ULL, "6016310_f.fxe"},
	{0xBC438FD33043002CULL, "vxlt_0_lowtown_a2_n_gyz_0_0.vlm"},
	{0xBC439FDD9A2EE185ULL, "sib131cr_slaver_lord.utc"},
	{0xBC43AAFCDE6A2F5AULL, "vxlt_0_lowtown_a2_n_c_e3b_0_0.vlm"},
	{0xBC43DD39D811FCC9ULL, "art\\levels\\ztools\\default\\generic\\9ff_blocker.mao"},
	{0xBC43E939D8121162ULL, "art\\levels\\ztools\\default\\generic\\9ff_blocker.mmh"},
	{0xBC43EB39D8121418ULL, "art\\levels\\ztools\\default\\generic\\9ff_blocker.msh"},
	{0xBC43ED2D9A2D938EULL, "vxht_mansion_n_6iy_0_0.vlm"},
	{0xBC443FAA770E942BULL, "6201269_f.fxe"},
	{0xBC447DDF57E88C60ULL, "hla_ht_0_hightown_0_0x.dds"},
	{0xBC454B4294A21F2DULL, "92950123.ogg"},
	{0xBC454BE64D75B1EDULL, "hlac_lt_darktown_-1_7z.dds"},
	{0xBC461FEEAA4A61E3ULL, "69622563.ogg"},
	{0xBC4620AA602A1F59ULL, "art\\vfx\\cinematic\\vfx_chest_pierce_charge_c.mmh"},
	{0xBC4677B9EB488C19ULL, "vxlt_gallowstemplar_5qy_0_0.vlm"},
	{0xBC467866A1B80D2FULL, "vxlt_0_lowtown_a3_n_c_eux_0_0.vlm"},
	{0xBC46970D72FE35DAULL, "vxca_caverns2_7o6_0_0.vlm"},
	{0xBC46AFAD40274153ULL, "6106921_m.fxe"},
	{0xBC46C9E7ACC2136DULL, "ts_generic_02.gda"},
	{0xBC4725B2F976663BULL, "gen00fl_merrill6144983.cl"},
	{0xBC473CA948023F75ULL, "6084522_m.fxe"},
	{0xBC478B3B35D1ABE1ULL, "796113732.ogg"},
	{0xBC484795C0ADEF1AULL, "art\\characters\\creatures\\ragedemon\\animation\\casting.evt"},
	{0xBC48DC31A8442285ULL, "one120_athenril6113285.cl"},
	{0xBC490B86D50E3F40ULL, "175082818.ogg"},
	{0xBC4979122D50CE01ULL, "zz_one_debug6069463.cl"},
	{0xBC49A8369531C6F1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp2a.msh"},
	{0xBC49B6369531DE0CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp2a.mao"},
	{0xBC49C2369531F2AFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksharp2a.mmh"},
	{0xBC49EBE6A067FC88ULL, "vxw_encounter3_83_0_0.vlm"},
	{0xBC49F5B63384235FULL, "dae000_merchant_mage_hf6164458.cl"},
	{0xBC4A1A1B71B88F71ULL, "zz_seb_debug6183189.cl"},
	{0xBC4A44AE66D84191ULL, "vxw_encounter2_af_0_0.vlm"},
	{0xBC4A6C9E73C2222CULL, "vxht_mansion2b_1xl_0_0.vlm"},
	{0xBC4B24D0110D687CULL, "llac_w_encounter3_-131072_0.dds"},
	{0xBC4B34A7D76C6B38ULL, "vxht_chantry_2cf_0_0.vlm"},
	{0xBC4B6E36A59CD843ULL, "156681819.ogg"},
	{0xBC4BC084E5FA73D1ULL, "vxca_cavern1_haunted_jw2_0_0.vlm"},
	{0xBC4CB7CAA5F8F6A3ULL, "vxlt_gallowstemplar_f_437_0_0.vlm"},
	{0xBC4D77DDC3290CD0ULL, "ts_darkspawn_arm_table.gda"},
	{0xBC4DA5F53B13B462ULL, "6127409_f.fxe"},
	{0xBC4DA6B973582DE2ULL, "vxlt_gallowstemplar_n_39g_0_0.vlm"},
	{0xBC4DDBD1C79E3C56ULL, "vxw_bonepitmining_a2_gy_0_0.vlm"},
	{0xBC4E0190A23B7335ULL, "genip_armoire.utp"},
	{0xBC4E0B963F6DECA0ULL, "ht_0_hightown_a3_30_0_ll.mmh"},
	{0xBC4E141F045937F8ULL, "vxw_woundedcoast_1pd_0_0.vlm"},
	{0xBC4E611ACBCEB9A3ULL, "6117000_m.fxe"},
	{0xBC4EFC1C340ABACFULL, "framebuffereffects_dof_p0.psh.fxo.dx11"},
	{0xBC4F105FE0F7F8B0ULL, "plt_vis100pt_offer.nss"},
	{0xBC4F1FAD4218DB3FULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialf.msh"},
	{0xBC4F25AD4218E5E1ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialf.mmh"},
	{0xBC4F31AD4218F982ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialf.mao"},
	{0xBC4F4695C0B40C0DULL, "art\\characters\\creatures\\ragedemon\\animation\\casting.gad"},
	{0xBC4F4E31E2A06A12ULL, "vxlt_warehouse_n_oa_0_0.vlm"},
	{0xBC4FC4849CE4B394ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwright_2.mao"},
	{0xBC50DB4B30A685F3ULL, "vxht_brothel_n_6t8_0_0.vlm"},
	{0xBC50E7484C9FEC8CULL, "6154709_m.fxe"},
	{0xBC511619AADEA005ULL, "vxw_woundedcoast_1op_0_0.vlm"},
	{0xBC51313E6377D14EULL, "hlac_w_bonepit_458757_0x.dds"},
	{0xBC5135F25FC34F43ULL, "6204165_f.fxe"},
	{0xBC5140319D476F91ULL, "dae000_gamlen6139542.cl"},
	{0xBC525F0A8C23AC0EULL, "vxht_mansion_n_6lz_0_0.vlm"},
	{0xBC5282C13A798E23ULL, "hlac_lt_gallowstemplar_2383_0y.dds"},
	{0xBC52E8D861DDEAB5ULL, "6108103_m.fxe"},
	{0xBC539EDB78FC5025ULL, "vxw_bonepitdestroyed_a3_270_0_0.vlm"},
	{0xBC5410C449D0DD7EULL, "dae000_loyal_mage6165225.cl"},
	{0xBC547A5E33031CDDULL, "6171532_m.fxe"},
	{0xBC54850427E761C1ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masb.phy"},
	{0xBC54EBC25BA5C9ABULL, "6027619_m.fxe"},
	{0xBC553B26DD599F06ULL, "797098479.ogg"},
	{0xBC553D0FC1826390ULL, "vxht_keep_n_67a_0_0.vlm"},
	{0xBC556C2328F08BD3ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.impa.ani"},
	{0xBC55D4083A3E50D3ULL, "vxlt_0_lowtown_h7s_0_0.vlm"},
	{0xBC56153A99EFB3E3ULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.dura.ani"},
	{0xBC56216D0C6CB68EULL, "824584490.ogg"},
	{0xBC566E40AFAE9D00ULL, "plt_zz_qcrpt_debug.plo"},
	{0xBC568456DDEAB40CULL, "w_blightlands_20_0.rml"},
	{0xBC575FD44A67CF7EULL, "616737770.ogg"},
	{0xBC576BB0C240CD1CULL, "gen00fl_anders6110066.cl"},
	{0xBC578C5B217EE27CULL, "153997019.ogg"},
	{0xBC57920E3BC38FD2ULL, "mag291ip_to_wryme.ncs"},
	{0xBC579E9B82649F08ULL, "one300_nuncio6076784.cl"},
	{0xBC57D806555FBDF9ULL, "character_character_p271.psh.fxo.dx11"},
	{0xBC584F23B1D7A5D4ULL, "vxlt_undercitydungeon_193_0_0.vlm"},
	{0xBC585DC3925E1508ULL, "vxw_encounter3_50_0_0.vlm"},
	{0xBC5866FBA6400C6FULL, "vxlt_warehouse_13t_0_0.vlm"},
	{0xBC588F0344251A85ULL, "79335752.ogg"},
	{0xBC58B68B58CE5A11ULL, "vxw_encounter2_le_0_0.vlm"},
	{0xBC58D2A9D1466836ULL, "hlac_lt_gallowstemplar_2384_1z.dds"},
	{0xBC58DE7B65B83AACULL, "vxht_mansion2b_1sm_0_0.vlm"},
	{0xBC5900EBD999E06EULL, "events.gda"},
	{0xBC592654064538D1ULL, "469786755.ogg"},
	{0xBC59272D63B0F286ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_medium.evt"},
	{0xBC595D7B5477E662ULL, "gen00fl_dog6141104.cl"},
	{0xBC595F9F44E96ECEULL, "vxht_0_hightown_a2_9wg_0_0.vlm"},
	{0xBC598C0A80474AF2ULL, "hlac_lt_0_lowtown_a2_n_c_-3_7y.dds"},
	{0xBC5992191FE9260DULL, "139253990.ogg"},
	{0xBC59D3EFF1214495ULL, "one121_trap6088456.cl"},
	{0xBC59DD9B71827915ULL, "vxlt_0_lowtown_a2_n_c_e26_0_0.vlm"},
	{0xBC5A1AC251583C79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_meredith.ani"},
	{0xBC5A384C8EB3F844ULL, "692794454.ogg"},
	{0xBC5A60E963A0CB99ULL, "hlac_lt_0_lowtown_docks_-1_2y.dds"},
	{0xBC5B29A797EF0F23ULL, "vxlt_gallowstemplar_f_486_0_0.vlm"},
	{0xBC5B6F53984EF579ULL, "vxw_bonepit_20q_0_0.vlm"},
	{0xBC5B733F87D9B679ULL, "6208173_m.fxe"},
	{0xBC5B8877C8B785FDULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_zbm_0n.dds"},
	{0xBC5B9ECA530DB7B0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16man.msh"},
	{0xBC5BACCA530DCFCAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16man.mmh"},
	{0xBC5BB0CA530DD601ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16man.mao"},
	{0xBC5C498179EE4784ULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.mmh"},
	{0xBC5D58114A446556ULL, "gencr_trap_explosives_4.utc"},
	{0xBC5DF019999DBBFEULL, "ser270_templar6127414.cl"},
	{0xBC5E68D07CD382AAULL, "6129776_m.fxe"},
	{0xBC5E8D3959CE62B2ULL, "vxlt_0_lowtown_a2_n_gpo_0_0.vlm"},
	{0xBC5EBA283092B86CULL, "6027771_m.fxe"},
	{0xBC5EC7BFE949EA1FULL, "hlac_w_encounter2_5_0z.dds"},
	{0xBC5F54BDFB92EF52ULL, "vxlt_0_lowtown_h7v_0_0.vlm"},
	{0xBC5FC5D564F007DAULL, "hlac_dr_slavers_3215_1z.dds"},
	{0xBC5FCFBC270AE051ULL, "421246802.ogg"},
	{0xBC5FE44567BDCDC6ULL, "qcr000_arishok6097549.cl"},
	{0xBC603E2D63B738D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_medium.gad"},
	{0xBC607FD2BF3319C9ULL, "900398394.ogg"},
	{0xBC608F7745450CA4ULL, "lt_0_lowtown_docks_a3_241_0.msh"},
	{0xBC60A17745452B0DULL, "lt_0_lowtown_docks_a3_241_0.mao"},
	{0xBC60A57745453246ULL, "lt_0_lowtown_docks_a3_241_0.mmh"},
	{0xBC60AE0152A61B82ULL, "513082629.ogg"},
	{0xBC60F5C2515E1BBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_meredith.gad"},
	{0xBC6114841B70C22EULL, "w_sundermount_209_0.msh"},
	{0xBC611B7BE6DD8C9FULL, "dae011_wench_mob6142281.cl"},
	{0xBC6122841B70DA7FULL, "w_sundermount_209_0.mao"},
	{0xBC612E841B70EE1CULL, "w_sundermount_209_0.mmh"},
	{0xBC613762E082DEF5ULL, "hlac_lt_0_lowtown_a2_131072_6x.dds"},
	{0xBC6206B88A401FECULL, "205934616.ogg"},
	{0xBC62642328FB080AULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.impa.evt"},
	{0xBC62E563696F0169ULL, "6107379_m.fxe"},
	{0xBC631EFF40851D0AULL, "pro000_cut2d_thefirstyear6015294.cl"},
	{0xBC632D3A99FA66DAULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.dura.evt"},
	{0xBC63A5D9C26B14B5ULL, "70015295.ogg"},
	{0xBC63AEECB3787C10ULL, "vxht_keep_n_62n_0_0.vlm"},
	{0xBC63C9AFE0EDE037ULL, "6137666_m.fxe"},
	{0xBC64187B20B506F8ULL, "401237193.ogg"},
	{0xBC648C432174B1C1ULL, "6129285_m.fxe"},
	{0xBC64AE3D72091E13ULL, "450743909.ogg"},
	{0xBC64BEAC92D51254ULL, "vxlt_0_lowtown_ep1_0_1.vlm"},
	{0xBC64D0DCCEC1B5F5ULL, "6156845_m.fxe"},
	{0xBC64E1395070F227ULL, "vxlt_gallowstemplar_n_2om_0_0.vlm"},
	{0xBC653F4375E8E8D0ULL, "6016448_f.fxe"},
	{0xBC6544FAD5077283ULL, "dae000ip_crowd_em4_talk15.utp"},
	{0xBC657D128A6EA85BULL, "870547347.ogg"},
	{0xBC6589C4B64BC598ULL, "vxlt_gallowstemplar_f_5br_0_0.vlm"},
	{0xBC65AD9B14F3F4CFULL, "6068492_m.fxe"},
	{0xBC65D91463F1B951ULL, "610229285.ogg"},
	{0xBC65EA0A4E519A03ULL, "qun160_vashothlead6055581.cl"},
	{0xBC66204D19F378B6ULL, "ht_0_hightown_a2_33_0.rml"},
	{0xBC663F2D63BBA67FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_laugh_medium.ani"},
	{0xBC6659E495AE1B3DULL, "llac_lt_darktown_-1_76.dds"},
	{0xBC6689D4AEE2731CULL, "6179054_m.fxe"},
	{0xBC6734BA085E248DULL, "gen00fl_merrill6116058.cl"},
	{0xBC6749376CBE7663ULL, "vxdr_0_deeproads_9ig_0_0.vlm"},
	{0xBC6766FD777FAA79ULL, "6034711_f.fxe"},
	{0xBC679A1834C0BF9CULL, "vxw_bonepitmining_a2_j8_0_0.vlm"},
	{0xBC67C90A5C0F6A89ULL, "vxht_0_hightown_a3_bbq_0_0.vlm"},
	{0xBC67E8D5FAB6D0B3ULL, "vxlt_warehouse_n_12h_0_0.vlm"},
	{0xBC6812C251646BA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\rushingstrikesyncenter_meredith.evt"},
	{0xBC681861BB589C38ULL, "vxht_chantry_2yd_0_0.vlm"},
	{0xBC687F09C75CBB36ULL, "vxca_cavern1_haunted_fmk_0_0.vlm"},
	{0xBC68A12044F41429ULL, "6051212_m.fxe"},
	{0xBC6943232900EF7DULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.impa.gad"},
	{0xBC69CD8395B6107BULL, "llac_lt_undercityencounter_-1_14.dds"},
	{0xBC69E27357AFE826ULL, "133275601.ogg"},
	{0xBC6A2C3A9A0083CDULL, "art\\vfx\\combat\\vfx_glb_arw_lit_c.dura.gad"},
	{0xBC6A95EF81F6CDF5ULL, "6110555_m.fxe"},
	{0xBC6AC5DC02487157ULL, "hlac_dr_dungeon_3217_6z.dds"},
	{0xBC6B0D4294E5E675ULL, "character_character_p17.psh.fxo.dx11.assembly"},
	{0xBC6B731E595B1065ULL, "46383427.ogg"},
	{0xBC6BD463DC33A973ULL, "hlac_ht_mansion_n_3367_0x.dds"},
	{0xBC6C067003059BFBULL, "1035690355.ogg"},
	{0xBC6C31EBC68C9B12ULL, "vxlt_warehouse_n_yc_0_0.vlm"},
	{0xBC6CE3A9F17EACECULL, "hlac_ca_caverns1_24424_17y.dds"},
	{0xBC6CF659C9B366ECULL, "663572629.ogg"},
	{0xBC6CFF164BC47B32ULL, "vxlt_0_lowtown_a2_n_gmh_0_0.vlm"},
	{0xBC6D73C04A834B82ULL, "aip_raiderassassin_base.gda"},
	{0xBC6D7F36625EBD14ULL, "vxlt_gallowscourtyard_n_5a5_0_0.vlm"},
	{0xBC6D8D8B1A31BB22ULL, "6083380_m.fxe"},
	{0xBC6DCFDEF4E93E4CULL, "753745250.ogg"},
	{0xBC6DEDDA3F233486ULL, "6038837_f.fxe"},
	{0xBC6DF0E9BEE39622ULL, "6144207_m.fxe"},
	{0xBC6E6E5097181B8FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\swq.anb"},
	{0xBC6E7ED572A6B3C4ULL, "vxlt_gallowstemplar_5cr_0_0.vlm"},
	{0xBC6EE56A062549ACULL, "vxlt_0_lowtown_a2_n_ghy_0_0.vlm"},
	{0xBC6F1DAA604D1F4EULL, "art\\vfx\\cinematic\\vfx_chest_pierce_charge_c.anb"},
	{0xBC6F30A7FABE2226ULL, "794343271.ogg"},
	{0xBC6F70C94F7DAF71ULL, "311363859.ogg"},
	{0xBC6FC798AF0F8426ULL, "and100_anders6095831.cl"},
	{0xBC701A6EE86499ADULL, "pro000_friedrich6050178.cl"},
	{0xBC7056252EE2BBA6ULL, "vxlt_0_lowtown_a2_fef_0_0.vlm"},
	{0xBC70711E4EACF8D5ULL, "6072682_m.fxe"},
	{0xBC7073F871B7BBEEULL, "vxdr_dungeon_2ie_0_0.vlm"},
	{0xBC70A4C4A6CB5E8FULL, "6148145_m.fxe"},
	{0xBC70DFA027BFF212ULL, "vxlt_darktown_4uk_0_0.vlm"},
	{0xBC70EA7AF0CF794FULL, "326083034.ogg"},
	{0xBC7178CAB7646951ULL, "39402672.ogg"},
	{0xBC71868A60CF8C0CULL, "vxca_caverns1_mines_dpd_0_0.vlm"},
	{0xBC7190502331500CULL, "4049616.ogg"},
	{0xBC71BF81E311B4CFULL, "mag101tr_aom_firstfight1.utt"},
	{0xBC726EB0336491A6ULL, "504539637.ogg"},
	{0xBC72724BC33B1D48ULL, "dae011_drunk1_sta6144145.cl"},
	{0xBC73474FFCA6813DULL, "vxdr_primevaltomb_6ve_0_0.vlm"},
	{0xBC73F053D380DB3FULL, "mag110_arianni6049716.cl"},
	{0xBC7404FC8693F2D6ULL, "follower_banter6166729.cl"},
	{0xBC7442C4DF756B4EULL, "hlac_lt_0_lowtown_131071_6x.dds"},
	{0xBC745586B88210A8ULL, "hlac_ht_0_hightown_n_-65535_0y.dds"},
	{0xBC747FEBEFD1CE63ULL, "vxlt_0_lowtown_n_fhg_0_0.vlm"},
	{0xBC7566BCC2471F5BULL, "w_woundedcoast_127_0.mmh"},
	{0xBC7572BCC24733F8ULL, "w_woundedcoast_127_0.mao"},
	{0xBC7584BCC2475245ULL, "w_woundedcoast_127_0.msh"},
	{0xBC75B03FA1DA8569ULL, "dae000ip_crowd_ff7_talk18.utp"},
	{0xBC7663EB90944D5AULL, "vxlt_gamlenshouse_n_iu_0_0.vlm"},
	{0xBC767559E889054FULL, "art\\characters\\playercharacter\\humanmale\\hlf_medc.phy"},
	{0xBC7754608474F6EFULL, "6043257_m.fxe"},
	{0xBC77AEA86EE4D324ULL, "19371580.ogg"},
	{0xBC77BA2BF6A34356ULL, "vxw_encounter2_camp_pc_0_1.vlm"},
	{0xBC77DD611455DD3BULL, "939262527.ogg"},
	{0xBC7808A0FC04E9AFULL, "hlac_ht_mansion2b_818_23x.dds"},
	{0xBC787D0CFE69BA1BULL, "6196732_f.fxe"},
	{0xBC78A38D473E618BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_1.ani"},
	{0xBC78C478650B5EB1ULL, "110558140.ogg"},
	{0xBC7983DEDF4BBA97ULL, "zz_dae_debug6029346.cl"},
	{0xBC799AD5C0856F7BULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanl.ani"},
	{0xBC79C0F4C95A6153ULL, "art\\characters\\creatures\\mubari\\animation\\cwkfr.ani"},
	{0xBC79C4F5F33CA432ULL, "gen_im_trs_cot_arc_lbw_r1.uti"},
	{0xBC7A8755C3E72478ULL, "mag200cr_feynriel.utc"},
	{0xBC7ABFD2C546E20CULL, "vxca_caverns2_83e_0_2.vlm"},
	{0xBC7ACECB4D0A2AE0ULL, "970416292.ogg"},
	{0xBC7AD1BD68313F14ULL, "322027066.ogg"},
	{0xBC7AF049E3728D26ULL, "374536712.ogg"},
	{0xBC7B4718FC26E76DULL, "gen00fl_aveline6140795.cl"},
	{0xBC7B4D6C0841271BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_trappad.mmh"},
	{0xBC7B596C08413BB8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_trappad.mao"},
	{0xBC7B6B6C08415A05ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\trap_trappad.msh"},
	{0xBC7B70F33DBA93B2ULL, "vxlt_0_lowtown_a2_n_gji_0_0.vlm"},
	{0xBC7BA279ECB0BF32ULL, "zz_followers_banter6158094.cl"},
	{0xBC7BCA75160DE091ULL, "qcr220_grey_wardens6135959.cl"},
	{0xBC7BE2EB54F1FBFEULL, "6174411_m.fxe"},
	{0xBC7C8216EE41B5F9ULL, "6045016_f.fxe"},
	{0xBC7C850E3BE295B7ULL, "mag291ip_to_wryme.utp"},
	{0xBC7C8E3F5DF15458ULL, "6117506_m.fxe"},
	{0xBC7CB333A3C91B83ULL, "141638749.ogg"},
	{0xBC7CF0B2649CCC44ULL, "vxlt_gallowstemplar_5ls_0_0.vlm"},
	{0xBC7D1729DC874A1FULL, "6077016_m.fxe"},
	{0xBC7D41848DCC6209ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon_mid.ani"},
	{0xBC7D796DDBE99424ULL, "vxht_0_hightown_a3_c6y_0_0.vlm"},
	{0xBC7D977D4147A0DEULL, "t3_eey_bl1.tnt"},
	{0xBC7DE12CD4DF0B35ULL, "lt_0_lowtown_docks_n_99_0.msh"},
	{0xBC7DE32CD4DF0EABULL, "lt_0_lowtown_docks_n_99_0.mmh"},
	{0xBC7DEF2CD4DF2348ULL, "lt_0_lowtown_docks_n_99_0.mao"},
	{0xBC7E57CBF756B53CULL, "art\\characters\\creatures\\spider\\animation\\blendtree\\spider_pcoverride.anb"},
	{0xBC7E84BDA062EE83ULL, "6034233_m.fxe"},
	{0xBC7E87169AB96E43ULL, "ene_clo.mao"},
	{0xBC7E9D9114356158ULL, "dae000ip_crowd_em4_talk9.utp"},
	{0xBC7EDBCB9C52F424ULL, "6068697_m.fxe"},
	{0xBC7F994904A2C78CULL, "vxht_mansion2_210_0_0.vlm"},
	{0xBC803FDD8C3962F7ULL, "vxlt_0_lowtown_docks_n_b5i_0_0.vlm"},
	{0xBC8086723FA4541BULL, "gen_im_wep_war_two_act2_06.uti"},
	{0xBC80950335A28FC1ULL, "zz_ser_debug6091659.cl"},
	{0xBC80D880D7B44C1FULL, "1030336205.ogg"},
	{0xBC8106896F8053A2ULL, "537845882.ogg"},
	{0xBC812CB8C1C3C4F1ULL, "6141056_m.fxe"},
	{0xBC813B189717D308ULL, "6174552_m.fxe"},
	{0xBC815C31EA5264A8ULL, "gen_im_trs_bks_chant.uti"},
	{0xBC816967327322AAULL, "672459782.ogg"},
	{0xBC816DF5B5D47883ULL, "vxca_caverns1_hlz_0_0.vlm"},
	{0xBC817F0CE4BFC401ULL, "885604433.ogg"},
	{0xBC81939848E0AF03ULL, "790828979.ogg"},
	{0xBC81EDEBB2EA9F37ULL, "vxlt_0_lowtown_docks_a3_btn_0_0.vlm"},
	{0xBC821461DC79FB08ULL, "vxlt_0_lowtown_a2_eu3_0_0.vlm"},
	{0xBC826D7E9A37F698ULL, "vxlt_gallowstemplar_f_5hp_0_0.vlm"},
	{0xBC8271F1E2A5B684ULL, "6043288_f.fxe"},
	{0xBC82D9775B6AB71FULL, "vxlt_gallowscourtyard_3vz_0_0.vlm"},
	{0xBC833A12138CB0D2ULL, "42974183.ogg"},
	{0xBC836989A81C14FFULL, "175225197.ogg"},
	{0xBC840C319ACC8C78ULL, "vrc321_varric6084348.cl"},
	{0xBC841ABF1ECBBE0AULL, "hlac_lt_0_lowtown_a2_n_c_65537_1y.dds"},
	{0xBC8428DEB5965FE0ULL, "mrl150_merrill6093843.cl"},
	{0xBC842CF150AAA763ULL, "vxdr_0_deeproads_9gi_0_0.vlm"},
	{0xBC844C817A0F9F7BULL, "art\\vfx\\combat\\vfx_glb_arw_spi_c.anb"},
	{0xBC845C848DD2AE8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon_mid.gad"},
	{0xBC855B909DE73719ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_exit.ani"},
	{0xBC85B2AF308FD7A2ULL, "vxlt_gallowstemplar_f_6s4_0_0.vlm"},
	{0xBC85BA2E516C8806ULL, "731937624.ogg"},
	{0xBC863254553A1F72ULL, "6075349_m.fxe"},
	{0xBC8638B36ED9EEECULL, "gen00fl_isabela6090430.cl"},
	{0xBC8641903C3AA41FULL, "hlac_lt_hangedman_-1_5y.dds"},
	{0xBC868B8D474A7602ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_1.evt"},
	{0xBC86B8F4C964DD8AULL, "art\\characters\\creatures\\mubari\\animation\\cwkfr.evt"},
	{0xBC86E42E68BF890BULL, "gen_ip_sack.utp"},
	{0xBC86F5B0C232CE64ULL, "gen00fl_fenris6116505.cl"},
	{0xBC87A2D5C091BA72ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanl.evt"},
	{0xBC884A7425B7BF21ULL, "919938855.ogg"},
	{0xBC8895A771C360EAULL, "216307526.ogg"},
	{0xBC8908579E4777FDULL, "882111305.ogg"},
	{0xBC8933E5D06F8D35ULL, "340998390.ogg"},
	{0xBC894F0A5B0A1DE4ULL, "vxw_bonepitmining_a2_5m_0_0.vlm"},
	{0xBC89B9153BD2DE69ULL, "dae000ip_crowd_ff4_talk6.utp"},
	{0xBC89EA015162D345ULL, "zz_fenpt_debug.ncs"},
	{0xBC8AC050177955CAULL, "pro000_aveline_intro6109368.cl"},
	{0xBC8AEBEDCBA28242ULL, "vxht_keep_n_61q_0_0.vlm"},
	{0xBC8B2191A5ED8BBFULL, "vxlt_gallowscourtyard_n_70s_0_0.vlm"},
	{0xBC8B24186AD16552ULL, "858703218.ogg"},
	{0xBC8B59848DD8C8F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_beckon_mid.evt"},
	{0xBC8BA0125BD8E74BULL, "vxca_caverns1_mines_gk9_0_0.vlm"},
	{0xBC8BFEDDF97B74DAULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolem.anb"},
	{0xBC8C36909DED175BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_exit.gad"},
	{0xBC8C65181414E3ABULL, "llac_lt_gallowscourtyard_n_-65535_1.dds"},
	{0xBC8C8041B0BCEB06ULL, "6084859_m.fxe"},
	{0xBC8D20CCF40681F5ULL, "llac_ht_keep_3732_3.dds"},
	{0xBC8D71E6D9128435ULL, "6167687_m.fxe"},
	{0xBC8D8A8D47509315ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_head_smash_1.gad"},
	{0xBC8D97F4C96AC4FDULL, "art\\characters\\creatures\\mubari\\animation\\cwkfr.gad"},
	{0xBC8D9AC64FEE6A13ULL, "gen_im_trs_dwv_med_hlm_r1.uti"},
	{0xBC8DC119AEBCEF34ULL, "ss_shield.dds"},
	{0xBC8DC1D5C0965A05ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_leanl.gad"},
	{0xBC8DD19F02B20B37ULL, "hlac_lt_backalley_n_0_2z.dds"},
	{0xBC8E4D04BCB85664ULL, "vxca_caverns2prc_78v_0_2.vlm"},
	{0xBC8E920E0E2CC86AULL, "6092496_m.fxe"},
	{0xBC8EF41BA87C187EULL, "art\\vfx\\combat\\vfx_glb_stun_c.impa.ani"},
	{0xBC8F149DCE978D47ULL, "403470564.ogg"},
	{0xBC8F15030943146AULL, "vxlt_0_lowtown_docks_b93_0_0.vlm"},
	{0xBC8F2A727CCB9DBCULL, "nat380_nathaniel6066498.cl"},
	{0xBC8F7A23FD70F2B4ULL, "vxlt_0_lowtown_a2_n_c_exm_0_0.vlm"},
	{0xBC900D3B1957AF1FULL, "vxlt_0_lowtown_a2_fi7_0_0.vlm"},
	{0xBC9064314D535354ULL, "vxca_cavern1_haunted_e8m_0_0.vlm"},
	{0xBC90863ECE701388ULL, "vxlt_0_lowtown_a2_ex0_0_0.vlm"},
	{0xBC9093289FC851D3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p_wkf.ani"},
	{0xBC909600CE27673BULL, "hlac_lt_0_lowtown_docks_n_-65539_0y.dds"},
	{0xBC909B232373EEBBULL, "isa211_isabela6069634.cl"},
	{0xBC90C876B79C1083ULL, "vxdr_primevaltomb_86e_0_0.vlm"},
	{0xBC91495EB62149B5ULL, "ach_kill_style.dds"},
	{0xBC9197F46E701C1BULL, "1050291522.ogg"},
	{0xBC92AE8ACB844656ULL, "638702947.ogg"},
	{0xBC92D8C854AA5DE1ULL, "6179600_m.fxe"},
	{0xBC92D9241EDE5C69ULL, "vxht_keep_n_69n_0_0.vlm"},
	{0xBC9325C1ED4500C4ULL, "6048259_m.fxe"},
	{0xBC932E715CD4981FULL, "728951278.ogg"},
	{0xBC9348B83D66559EULL, "293425096.ogg"},
	{0xBC934FE0EDB79D63ULL, "345596526.ogg"},
	{0xBC9353909DF36740ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\cower_exit.evt"},
	{0xBC93BFB8E5B83E03ULL, "fen210_hadriana6110569.cl"},
	{0xBC93D1776CEF39CAULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehanga.mao"},
	{0xBC93D5776CEF4019ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehanga.mmh"},
	{0xBC93DF776CEF5107ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_cheesehanga.msh"},
	{0xBC940D63555F9E64ULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.mmh"},
	{0xBC947115B7C48617ULL, "373991681.ogg"},
	{0xBC94803D0E3E5314ULL, "hlac_w_sundermount_0_7x.dds"},
	{0xBC94F9B7C3C57235ULL, "6088892_m.fxe"},
	{0xBC95416261D61951ULL, "mrl000_marethari6142166.cl"},
	{0xBC9566ABFD76DBBAULL, "6046887_f.fxe"},
	{0xBC9583FF0C89BFF5ULL, "lt_shop_probe_6l.dds"},
	{0xBC95B6CF3F9E1EBBULL, "mer120_hubert6077586.cl"},
	{0xBC97EF35E27FE009ULL, "hlac_lt_backalley_g_1_6x.dds"},
	{0xBC97F23560A602AAULL, "vxlt_0_lowtown_docks_a3_czi_0_1.vlm"},
	{0xBC9824E429A88B3FULL, "character_character_p95.psh.fxo.dx11"},
	{0xBC983963B8B2CCA5ULL, "191251842.ogg"},
	{0xBC9880DDFF0D9E1FULL, "zz_vrc_debug6200188.cl"},
	{0xBC98904A2C5F258BULL, "271969335.ogg"},
	{0xBC98FD9B166CD00FULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_act_c.impa.gad"},
	{0xBC996568E203938AULL, "vxlt_gallowstemplar_f_3t0_0_0.vlm"},
	{0xBC996B1CAE4FBC78ULL, "hlac_ht_mansion2b_818_17x.dds"},
	{0xBC99D46C4888FD44ULL, "vxlt_gallowstemplar_5rq_0_0.vlm"},
	{0xBC9ADCF323079942ULL, "gen00fl_aveline6122000.cl"},
	{0xBC9AF30F74FF8626ULL, "lgt310_mis_burn_dump.cnv"},
	{0xBC9B4240929F9ACFULL, "mag291_demon_marethari6028936.cl"},
	{0xBC9BF91E08E89547ULL, "6169900_m.fxe"},
	{0xBC9C481C1F8E1628ULL, "6108146_m.fxe"},
	{0xBC9C771E8580DB78ULL, "6055583_m.fxe"},
	{0xBC9CE043F08CFC39ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv1.ani"},
	{0xBC9CFD52B277BAA9ULL, "565123223.ogg"},
	{0xBC9D0C1BA8887E4FULL, "art\\vfx\\combat\\vfx_glb_stun_c.impa.evt"},
	{0xBC9D3C697C5A1862ULL, "hlac_lt_0_lowtown_-1_4y.dds"},
	{0xBC9D4C91DD0E1453ULL, "6129089_m.fxe"},
	{0xBC9DBA8E5765F0E3ULL, "lgt220_shp_mag_guard6149880.cl"},
	{0xBC9DF50F7501BFE6ULL, "lgt310_mis_burn_dump.bnk"},
	{0xBC9E5C8EE8A8593EULL, "hlac_w_encounter2_camp_65537_0x.dds"},
	{0xBC9EA1C6796B0D9FULL, "mcr371_alain6087133.cl"},
	{0xBC9F129B16715FBCULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_act_c.impa.evt"},
	{0xBC9FB0DE3CF4B81BULL, "dae000_bartrand6031152.cl"},
	{0xBCA01D4D1A24EC6AULL, "ht_0_hightown_a2_33_0.anb"},
	{0xBCA0A65C38F8C208ULL, "292020140.ogg"},
	{0xBCA0F4A16264B6A8ULL, "6166829_m.fxe"},
	{0xBCA10EED20D847DAULL, "hlac_lt_0_lowtown_a2_n_131071_1x.dds"},
	{0xBCA14B0110D474E9ULL, "vxht_keep_n_64m_0_0.vlm"},
	{0xBCA17B3966143DCFULL, "mag291_vincento6027693.cl"},
	{0xBCA1979EDF3B1944ULL, "6043359_m.fxe"},
	{0xBCA2F787F9395983ULL, "vxlt_0_lowtown_docks_bav_0_0.vlm"},
	{0xBCA377F447B8818EULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.mmh"},
	{0xBCA3B4F06C0A04C0ULL, "896813144.ogg"},
	{0xBCA3BB43F092DC7BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv1.gad"},
	{0xBCA3F96430295C67ULL, "vxw_encounter2_4k_0_0.vlm"},
	{0xBCA40D1BA88E9EF8ULL, "art\\vfx\\combat\\vfx_glb_stun_c.impa.gad"},
	{0xBCA4122F224847FCULL, "vxw_bonepitmining_a2_h2_0_0.vlm"},
	{0xBCA45972937A5ADCULL, "6054373_m.fxe"},
	{0xBCA46A289FD8B57DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p_wkf.gad"},
	{0xBCA473C2D29D0240ULL, "lt_0_lowtown_docks_n_30_0_lc.mmh"},
	{0xBCA48CA407608AD5ULL, "96602680.ogg"},
	{0xBCA4E1E11B14C4D9ULL, "vxht_mansion2_3cb_0_0.vlm"},
	{0xBCA57BDF968A378AULL, "238606395.ogg"},
	{0xBCA6488232A319C6ULL, "hlac_lt_0_lowtown_-65535_2x.dds"},
	{0xBCA6A9AEA490E9D9ULL, "vxlt_gallowscourtyard_n_7en_0_0.vlm"},
	{0xBCA6B954ED4CDF2DULL, "6175752_m.fxe"},
	{0xBCA7072D0174B13EULL, "vxlt_gallowstemplar_f_5r4_0_0.vlm"},
	{0xBCA74B617CFA95C2ULL, "qun131cr_warrens_thug5.utc"},
	{0xBCA74D071B29142FULL, "hlac_dr_slavers_3207_7z.dds"},
	{0xBCA75A58EC26CD4FULL, "vxdr_primevaltomb_7xx_0_0.vlm"},
	{0xBCA77D1C7415191DULL, "443114264.ogg"},
	{0xBCA7D745D3F9AC0AULL, "vxlt_gallowstemplar_f_3s1_0_0.vlm"},
	{0xBCA81A9C7A93F591ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_movement_bow.anb"},
	{0xBCA87FD61CF7FACAULL, "vxlt_0_lowtown_a3_n_c_e3c_0_0.vlm"},
	{0xBCA89CB4B02F0BEDULL, "character_character_p31.psh.fxo.dx11.assembly"},
	{0xBCA8E1DC6DCF4EBEULL, "6131183_f.fxe"},
	{0xBCA8E79595C7697FULL, "mag311_keran6054237.cl"},
	{0xBCA93A71C5200ABFULL, "111009003.ogg"},
	{0xBCA983F46ABD4E87ULL, "fx_waterpour01.mao"},
	{0xBCA9B621FEFF0DB4ULL, "dre000_dougal6131179.cl"},
	{0xBCA9D843F0977960ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv1.evt"},
	{0xBCAA2452E4B7A868ULL, "359180033.ogg"},
	{0xBCAAD980451C78B6ULL, "llac_ca_caverns2_5802_56.dds"},
	{0xBCAADE79C4CFB900ULL, "36501423.ogg"},
	{0xBCAB2F9271DAF436ULL, "vxlt_0_lowtown_n_bye_0_0.vlm"},
	{0xBCAB394D94117544ULL, "mer320_hubert6053200.cl"},
	{0xBCAB9F858FEE7CEDULL, "vxlt_undercityserialkiller_1j7_0_1.vlm"},
	{0xBCABAE48180630C6ULL, "6171063_m.fxe"},
	{0xBCAC2F468C40619EULL, "art\\vfx\\creatures\\fxc_arch_vc_p.cess.evt"},
	{0xBCAC7025C7812A1EULL, "gen00fl_carver6102694.cl"},
	{0xBCAD1A9B167DAAD5ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_act_c.impa.ani"},
	{0xBCAD4E9E628C8C06ULL, "vxlt_0_lowtown_a2_eqh_0_1.vlm"},
	{0xBCAE49E3E5241D3BULL, "lt_backalley_g_3_0.msh"},
	{0xBCAE5B40AFF8F28FULL, "plt_zz_qcrpt_debug.nss"},
	{0xBCAE5BE3E5243B86ULL, "lt_backalley_g_3_0.mao"},
	{0xBCAE67E3E5245025ULL, "lt_backalley_g_3_0.mmh"},
	{0xBCAED48CD93C28D5ULL, "hlac_ca_caverns1_mines_24423_4z.dds"},
	{0xBCAF37CB0A887F1BULL, "hlac_ca_caverns2_5802_24z.dds"},
	{0xBCAF39BE2D913336ULL, "vxlt_0_lowtown_a2_n_c_e1t_0_0.vlm"},
	{0xBCAF999E66A7ECD9ULL, "6132762_m.fxe"},
	{0xBCAFA3BF49A54F00ULL, "dae000_unhappy_mage6165333.cl"},
	{0xBCAFBC8238968EE1ULL, "6177400_m.fxe"},
	{0xBCAFCEC526CD3994ULL, "587350370.ogg"},
	{0xBCB013AFC7CC5056ULL, "6179768_m.fxe"},
	{0xBCB079F398745910ULL, "6134871_m.fxe"},
	{0xBCB163953A3C6B1FULL, "hlac_ht_mansion2_818_53y.dds"},
	{0xBCB16964EB2F7203ULL, "vxlt_0_lowtown_docks_bdy_0_0.vlm"},
	{0xBCB196F447C4F38DULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.anb"},
	{0xBCB1DEE1B8AA763CULL, "mag291_demon_marethari6028843.cl"},
	{0xBCB23420BBDA4DCEULL, "802969424.ogg"},
	{0xBCB246468C44F4C9ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.cess.gad"},
	{0xBCB2822106D7C6C8ULL, "vxlt_foundry_n_12y_0_0.vlm"},
	{0xBCB2BE9D4C63A967ULL, "vxw_blightlands_cq_0_0.vlm"},
	{0xBCB30CE0913EADFAULL, "vxlt_gallowstemplar_2mu_0_0.vlm"},
	{0xBCB32A829906A608ULL, "692457924.ogg"},
	{0xBCB39763DF2C60E4ULL, "zz_isa_debug6133979.cl"},
	{0xBCB3D3B1ED066902ULL, "6084356_m.fxe"},
	{0xBCB4E43327E7C212ULL, "6163604_m.fxe"},
	{0xBCB4F173F76A0959ULL, "17512825.ogg"},
	{0xBCB51B8B96870259ULL, "vxlt_gallowscourtyard_n_7hm_0_0.vlm"},
	{0xBCB58FCF0C51A8ECULL, "qcr000_aveline6041068.cl"},
	{0xBCB648B53C68600DULL, "1029810846.ogg"},
	{0xBCB68A04283A34FDULL, "art\\characters\\playercharacter\\humanmale\\hlh_masb.mao"},
	{0xBCB68E04283A3B36ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masb.mmh"},
	{0xBCB69804283A4C34ULL, "art\\characters\\playercharacter\\humanmale\\hlh_masb.msh"},
	{0xBCB69B439A885C33ULL, "vxht_0_hightown_a2_9zi_0_0.vlm"},
	{0xBCB6C1E9FCFD99A7ULL, "art\\vfx\\combat\\vfx_sg_slam_act_c.impa.gad"},
	{0xBCB711E67380C3D1ULL, "954338935.ogg"},
	{0xBCB7363BD4C86396ULL, "one280_merin6105435.cl"},
	{0xBCB761616C4704ABULL, "6172489_m.fxe"},
	{0xBCB7B7B2265A85EBULL, "417153646.ogg"},
	{0xBCB81CBE34095556ULL, "lt_0_lowtown_n_52_0_ll.mmh"},
	{0xBCB82C4B25D94101ULL, "vxlt_undercitydungeon_20v_0_0.vlm"},
	{0xBCB85A314994DF9FULL, "6153545_m.fxe"},
	{0xBCB8C0CB8BC7D588ULL, "unlockables.gda"},
	{0xBCB9049515F25942ULL, "hlac_w_bonepit_196608_0x.dds"},
	{0xBCB90F61988EC291ULL, "6111192_m.fxe"},
	{0xBCB917468C4AC317ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.cess.ani"},
	{0xBCB9C29B9CDEB051ULL, "vxlt_darktown_60y_0_0.vlm"},
	{0xBCBAE6F574B82546ULL, "6034870_m.fxe"},
	{0xBCBB494D8B6D36F4ULL, "gen00fl_aveline6140143.cl"},
	{0xBCBB64E3E52ED5F2ULL, "lt_backalley_g_3_0.anb"},
	{0xBCBBB9C823359CD4ULL, "vxca_cavern1_haunted_e7p_0_0.vlm"},
	{0xBCBBC0A0B90091F3ULL, "6139075_m.fxe"},
	{0xBCBC166FD0AF9442ULL, "mer120_hubert6077170.cl"},
	{0xBCBC80BA67D15027ULL, "6163832_m.fxe"},
	{0xBCBC8486E3B67213ULL, "260048029.ogg"},
	{0xBCBD0788087074AEULL, "6139948_m.fxe"},
	{0xBCBD10635582A85BULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.anb"},
	{0xBCBD8E5F3608DDC3ULL, "art\\characters\\playercharacter\\children\\animation\\blendtree\\child.anb"},
	{0xBCBDB6E9FD03A5B4ULL, "art\\vfx\\combat\\vfx_sg_slam_act_c.impa.evt"},
	{0xBCBE437BC68AFEA1ULL, "dae000_mother6127179.cl"},
	{0xBCBE44A25F2E49ADULL, "hlac_lt_0_lowtown_docks_n_0_3y.dds"},
	{0xBCBF50A406F78A40ULL, "isa311_isabela6092151.cl"},
	{0xBCBFAAD1749F5F47ULL, "hlac_lt_0_lowtown_docks_a3_-4_2y.dds"},
	{0xBCC00641312B28A1ULL, "vxht_chantry_n_2lm_0_0.vlm"},
	{0xBCC08AF46AD039DDULL, "fx_waterpour01.dds"},
	{0xBCC0992D5C763A97ULL, "6204104_m.fxe"},
	{0xBCC0EC00D0CCA0A3ULL, "hightown_market_guard_sta6169638.cl"},
	{0xBCC0FC7A24874D1EULL, "6095170_m.fxe"},
	{0xBCC1307A3E59C1E7ULL, "vxw_blightlands_dp_0_0.vlm"},
	{0xBCC1B97DD081C488ULL, "vxdr_primevalentrance_336_0_0.vlm"},
	{0xBCC1D20283CB92EAULL, "6101430_f.fxe"},
	{0xBCC2085937F4C44BULL, "6083090-mages_saved.cut"},
	{0xBCC21C685DD14848ULL, "520848223.ogg"},
	{0xBCC34AC4E1DAAC1EULL, "hlac_ht_mansion2b_818_16y.dds"},
	{0xBCC3B004AF7F1F1DULL, "llac_lt_gallowstemplar_n_2394_0.dds"},
	{0xBCC3D04216CFB156ULL, "595732306.ogg"},
	{0xBCC4C17B77A7414DULL, "6018567_m.fxe"},
	{0xBCC4E38590561DDAULL, "6172963_m.fxe"},
	{0xBCC4F1DB2BDD1629ULL, "hlac_lt_0_lowtown_a2_n_-2_6x.dds"},
	{0xBCC52D1952A80543ULL, "441516499.ogg"},
	{0xBCC5574333884DE7ULL, "838436577.ogg"},
	{0xBCC626BE96AD73BCULL, "mcr000pt_general.ncs"},
	{0xBCC6C18E555D1EA4ULL, "6083643_m.fxe"},
	{0xBCC6F33CD52E2AC6ULL, "zz_mrl_debug6185307.cl"},
	{0xBCC724ECFA0EF9A4ULL, "6144142_m.fxe"},
	{0xBCC730FF42998C6BULL, "mag101_grace6034224.cl"},
	{0xBCC757382C58462BULL, "6089901_m.fxe"},
	{0xBCC76D5A887D7835ULL, "hlac_lt_undercityserialkiller_-1_49z.dds"},
	{0xBCC8833F73DAADEDULL, "vxlt_undercityserialkiller_1d5_0_1.vlm"},
	{0xBCC94C799945481EULL, "318782661.ogg"},
	{0xBCC95DB4FC6162A8ULL, "1073657193.ogg"},
	{0xBCC9A6DF28C87FFFULL, "mcr310_argument6072297.cl"},
	{0xBCC9B9B3F782AE1FULL, "art\\vfx\\combat\\vfx_d_qun_direct_cast_c.impa.gad"},
	{0xBCC9BFABC9A68D91ULL, "hlac_lt_0_lowtown_a2_n_196607_1y.dds"},
	{0xBCC9F5AA6FEECB2FULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_hvyo_0d.dds"},
	{0xBCCA384C6000CC7AULL, "vxca_caverns2_8r6_0_0.vlm"},
	{0xBCCA504E1EA9D958ULL, "dre000_dougal6131203.cl"},
	{0xBCCA85504AB844FBULL, "hlac_w_encounter2_196606_0y.dds"},
	{0xBCCADFB38A1AE803ULL, "zz_spy130_darktown_res_1.cnv"},
	{0xBCCB1B4B4D961779ULL, "vxw_blightlands_11d_0_0.vlm"},
	{0xBCCB2B9BE81D1DDFULL, "pro000_guard6013446.cl"},
	{0xBCCB7B26345BFDECULL, "55256175.ogg"},
	{0xBCCB7C73B15B0B82ULL, "vxlt_darktown_6jd_0_0.vlm"},
	{0xBCCBAEE9FD0FD5EDULL, "art\\vfx\\combat\\vfx_sg_slam_act_c.impa.ani"},
	{0xBCCC54F58F233B48ULL, "hlac_w_sundermount_-393215_0y.dds"},
	{0xBCCC7566B5D6BB23ULL, "216568067.ogg"},
	{0xBCCCE4FD5ECD30EDULL, "726891862.ogg"},
	{0xBCCD613A1F3F03CDULL, "art\\levels\\races\\natural\\highlands\\props\\prp_craggyrock.mmh"},
	{0xBCCD653A1F3F0A1EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_craggyrock.mao"},
	{0xBCCD733A1F3F2253ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_craggyrock.msh"},
	{0xBCCD8A6BDC15E2C9ULL, "vxlt_0_lowtown_a2_n_c_hz3_0_0.vlm"},
	{0xBCCDCDEB2DF4C31DULL, "hlac_lt_0_lowtown_a2_n_c_-262144_0y.dds"},
	{0xBCCDD49D6FDB424CULL, "and230cr_templar2.utc"},
	{0xBCCE40CE284C50A9ULL, "vxht_0_hightown_n_c75_0_0.vlm"},
	{0xBCCE9BB743A5582BULL, "308976932.ogg"},
	{0xBCCEED567E865C5CULL, "qcr000_arishok6042493.cl"},
	{0xBCCFD8FA63A83B39ULL, "6099238_m.fxe"},
	{0xBCCFF35F62EB5EAEULL, "6068808_m.fxe"},
	{0xBCD000C11A5437CBULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv3.ani"},
	{0xBCD010B725BAD7EDULL, "vxlt_0_lowtown_a3_n_c_db6_0_0.vlm"},
	{0xBCD08C8BDEB68FC4ULL, "vxht_0_hightown_a2_4an_0_0.vlm"},
	{0xBCD0AEB3F788BA0CULL, "art\\vfx\\combat\\vfx_d_qun_direct_cast_c.impa.evt"},
	{0xBCD0BB8D2B7F2BBCULL, "6000729_m.fxe"},
	{0xBCD0BD5CF8AFAED6ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf_vx.evt"},
	{0xBCD165B0A1E6BA5DULL, "6177844_m.fxe"},
	{0xBCD1685A7715A0A2ULL, "art\\characters\\creatures\\spider\\animation\\cwkf.gad"},
	{0xBCD1E7D12EC4C3D3ULL, "190527067.ogg"},
	{0xBCD2444404B32D9FULL, "vxw_bonepitdestroyed_a3_2a0_0_0.vlm"},
	{0xBCD25CC3D756FABEULL, "vxlt_gallowstemplar_f_5c7_0_0.vlm"},
	{0xBCD28F6B9A4338FFULL, "vxlt_backalley_n_1y0_0_0.vlm"},
	{0xBCD2AA87EE1A1629ULL, "art\\vfx\\combat\\vfx_c_mer_grab_stab_c.mmh"},
	{0xBCD2C837F51C7776ULL, "vxlt_hangedman_16v_0_0.vlm"},
	{0xBCD2CFF3F1AFE93BULL, "998553688.ogg"},
	{0xBCD300C464EA2779ULL, "removetalent.ncs"},
	{0xBCD3706FF7EA55E8ULL, "vxlt_undercitydungeon_1nv_0_0.vlm"},
	{0xBCD392A37D59E7FCULL, "6109228_m.fxe"},
	{0xBCD3C42DF24DA3BCULL, "6178585_m.fxe"},
	{0xBCD4415379ECD142ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12.phy"},
	{0xBCD4C4E658158505ULL, "gen00fl_varric6152689.cl"},
	{0xBCD4DF11425A26BFULL, "vxca_caverns1_mines_jti_0_0.vlm"},
	{0xBCD521311CE0EFCEULL, "vxlt_backalley_n_1z0_0_0.vlm"},
	{0xBCD5295A50FF2979ULL, "1052299197.ogg"},
	{0xBCD571CBB8413C16ULL, "dm_bulf.mor"},
	{0xBCD5A04DF973E2D9ULL, "237716700.ogg"},
	{0xBCD5BB7257B93A5BULL, "qun221_petrice6075984.cl"},
	{0xBCD5C360D512A624ULL, "vxht_keep_4an_0_0.vlm"},
	{0xBCD5C4832D3ADE0BULL, "6065341_m.fxe"},
	{0xBCD5E1B0FD791004ULL, "creeper_bloodwound.dds"},
	{0xBCD5E8CE910A900BULL, "vxht_chantry_n_10p_0_0.vlm"},
	{0xBCD65A2778B356ACULL, "pro000_athenril6049462.cl"},
	{0xBCD7222C17DC5DAAULL, "6165633_m.fxe"},
	{0xBCD75863F43EA4CEULL, "art\\characters\\weapons\\quivers\\w_qui_dq02a.mmh"},
	{0xBCD76463F43EB965ULL, "art\\characters\\weapons\\quivers\\w_qui_dq02a.mao"},
	{0xBCD77263F43ED0BCULL, "art\\characters\\weapons\\quivers\\w_qui_dq02a.msh"},
	{0xBCD78A8428039804ULL, "lgt160im_wk_avern_messages.uti"},
	{0xBCD7B45CF8B5BE61ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf_vx.gad"},
	{0xBCD7F24405BCCE91ULL, "vxht_brothel_n_6ii_0_0.vlm"},
	{0xBCD827A8608400D7ULL, "dae000ip_crowd_lm3_beg.utp"},
	{0xBCD863F9572F5BB1ULL, "vxw_bonepit_25j_0_0.vlm"},
	{0xBCD88B5A771BFA99ULL, "art\\characters\\creatures\\spider\\animation\\cwkf.evt"},
	{0xBCD89E6B68A627C5ULL, "hlac_w_sundermount_393210_0z.dds"},
	{0xBCD941A90A83FD05ULL, "hlac_ht_mansion_n_3374_1y.dds"},
	{0xBCD953F1C01D77FAULL, "lt_gallowscourtyard_36_0_ll.mmh"},
	{0xBCD98D283F8C2FF9ULL, "vxw_blightlands_14k_0_0.vlm"},
	{0xBCD9C062174B00DEULL, "llac_lt_undercitydungeon_-1_3.dds"},
	{0xBCD9FAB64E3033A4ULL, "668324713.ogg"},
	{0xBCDA65B0620D7330ULL, "910659419.ogg"},
	{0xBCDADAED6A5A2870ULL, "6080233_m.fxe"},
	{0xBCDB14D3912A5F10ULL, "lgt220cr_mag_tmplr_captain.utc"},
	{0xBCDB4FF6F7A49A60ULL, "219103781.ogg"},
	{0xBCDB58DC6C890437ULL, "6116739_m.fxe"},
	{0xBCDB6D49AFAC70E9ULL, "360227135.ogg"},
	{0xBCDB772213E14387ULL, "zz_vault_debug6200376.cl"},
	{0xBCDB86AD76A11B75ULL, "vxdr_slavers_23b_0_0.vlm"},
	{0xBCDBA2609E6C4962ULL, "6065803_m.fxe"},
	{0xBCDBF085C6D49F1FULL, "lgt321cr_alistair_king.utc"},
	{0xBCDBFC48CE0BFB49ULL, "vxlt_0_lowtown_a2_n_c_hu2_0_0.vlm"},
	{0xBCDC8D235F0F9F83ULL, "vxca_cavern1_haunted_i1i_0_0.vlm"},
	{0xBCDD155218B9E759ULL, "6155646_m.fxe"},
	{0xBCDD6DB5A4181F7BULL, "vxca_caverns1_mines_k6j_0_0.vlm"},
	{0xBCDD8407C5199C98ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_6z.dds"},
	{0xBCDD91D4A48D2043ULL, "66438639.ogg"},
	{0xBCDDB55CF8BA2B0FULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf_vx.ani"},
	{0xBCDDBD50CBDFB4FCULL, "504469839.ogg"},
	{0xBCDDD4842A075FE7ULL, "171573814.ogg"},
	{0xBCDDE5729B83CD11ULL, "6207929_m.fxe"},
	{0xBCDDE8C11A604C42ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv3.evt"},
	{0xBCDE14342245F2E7ULL, "vxw_blightlands_rr_0_0.vlm"},
	{0xBCDE37A436A5D6B7ULL, "hlac_lt_0_lowtown_a2_65534_4y.dds"},
	{0xBCDE4355374CD576ULL, "770110095.ogg"},
	{0xBCDE5A3A1F4D6A8AULL, "art\\levels\\races\\natural\\highlands\\props\\prp_craggyrock.phy"},
	{0xBCDE96B3F794CF65ULL, "art\\vfx\\combat\\vfx_d_qun_direct_cast_c.impa.ani"},
	{0xBCDEB27A66F9DF40ULL, "vxht_playershouse_n_1gc_0_0.vlm"},
	{0xBCDF7CF8493AFA1FULL, "6087015_m.fxe"},
	{0xBCDFA436F6FAB7DEULL, "vxht_mansion2_1a1_0_0.vlm"},
	{0xBCDFEAD015DAC670ULL, "lt_0_lowtown_docks_a3_n_167_0.msh"},
	{0xBCDFF8D015DADD8AULL, "lt_0_lowtown_docks_a3_n_167_0.mmh"},
	{0xBCDFFCD015DAE4C1ULL, "lt_0_lowtown_docks_a3_n_167_0.mao"},
	{0xBCE007739A80B0FBULL, "gen00fl_anders6113116.cl"},
	{0xBCE007F1A0FE2C41ULL, "506069854.ogg"},
	{0xBCE043491FA852A9ULL, "mag291_demon_marethari6028927.cl"},
	{0xBCE056FA82626BBCULL, "hlac_lt_0_lowtown_a2_n_c_-131072_7x.dds"},
	{0xBCE0799FC665F748ULL, "hla_ht_0_hightown_n_-196608_0x.dds"},
	{0xBCE0CEC7E19128F2ULL, "character_character_p269.psh.fxo.dx11.assembly"},
	{0xBCE14B4DE6DD86C1ULL, "vxht_keep_69y_0_0.vlm"},
	{0xBCE1990A3B6DA4A7ULL, "vxlt_0_lowtown_n_enr_0_0.vlm"},
	{0xBCE1B6551E06A155ULL, "279356948.ogg"},
	{0xBCE1E24CE9E06E68ULL, "vxlt_undercitydungeon_1qw_0_0.vlm"},
	{0xBCE204806F50007CULL, "6102328_m.fxe"},
	{0xBCE223A510D82B09ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basicx.phy"},
	{0xBCE2357DF877F549ULL, "81309880.ogg"},
	{0xBCE23D18179C7DF4ULL, "207051384.ogg"},
	{0xBCE2981A47931041ULL, "gen00fl_carver6104842.cl"},
	{0xBCE2C9872F37A10FULL, "dae000cr_grand_cleric6178493.cl"},
	{0xBCE310FE631F0C59ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordb_d.dds"},
	{0xBCE31184C07F9BD6ULL, "hlac_lt_0_lowtown_a2_n_1_3x.dds"},
	{0xBCE3457419E37E85ULL, "vxlt_darktown_5r8_0_0.vlm"},
	{0xBCE38D0C88118EF9ULL, "hlac_ca_cavern1_haunted_24424_1z.dds"},
	{0xBCE3E7C11A64B655ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backrightlegattackv3.gad"},
	{0xBCE5345D7F8458D8ULL, "6118671_m.fxe"},
	{0xBCE5682E441B7818ULL, "vxlt_gallowstemplar_n_5yw_0_0.vlm"},
	{0xBCE58894B581D773ULL, "vxw_sundermount_2g1_0_0.vlm"},
	{0xBCE5935A77269288ULL, "art\\characters\\creatures\\spider\\animation\\cwkf.ani"},
	{0xBCE5ACBC740AA5D5ULL, "dra_cyc.mao"},
	{0xBCE5BDE336878627ULL, "hlac_ca_caverns2prc_5802_62y.dds"},
	{0xBCE5BDFD091D5825ULL, "6078463_f.fxe"},
	{0xBCE5EB72C12286F1ULL, "fen310_varania6118636.cl"},
	{0xBCE5FE16EAB4E414ULL, "6054012_m.fxe"},
	{0xBCE60A456731F5FCULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.dura.ani"},
	{0xBCE64C0B9EF16544ULL, "hm_gen_bloodmage.mor"},
	{0xBCE6573A89774865ULL, "377944771.ogg"},
	{0xBCE670E66B322CC3ULL, "6126689_f.fxe"},
	{0xBCE6D5D649257431ULL, "vxw_bonepit_20i_0_0.vlm"},
	{0xBCE6D79BCEBC5612ULL, "1006648191.ogg"},
	{0xBCE778E829533FD9ULL, "static_lava_p0.vsh.fxo"},
	{0xBCE790683A9780FDULL, "qcr000_arishok6109092.cl"},
	{0xBCE79B550CC3B8A0ULL, "vxlt_0_lowtown_a3_n_c_fid_0_0.vlm"},
	{0xBCE7B7F23C36AEDFULL, "6071146_m.fxe"},
	{0xBCE7EF282D4A2411ULL, "hlac_lt_0_lowtown_a2_n_-3_1z.dds"},
	{0xBCE84BCF53CA98A5ULL, "dae111ar_gamlens_house_nt.are"},
	{0xBCE8C1C525749AD4ULL, "vxlt_undercitydungeon_1o9_0_0.vlm"},
	{0xBCE93167833DDA32ULL, "llac_dr_primevalentrance_1724_4.dds"},
	{0xBCE960C6A8058E0CULL, "gen_im_trs_rdr_shd_wdn_r0.uti"},
	{0xBCE971F3C9AE5EE3ULL, "atl_startmenu_dxt5_dat.xml"},
	{0xBCE9A35EC7E08884ULL, "vxlt_0_lowtown_docks_8ug_0_0.vlm"},
	{0xBCE9B15073B68104ULL, "vxlt_0_lowtown_docks_a3_cth_0_0.vlm"},
	{0xBCEA6E25C00213C9ULL, "vxlt_0_lowtown_a2_n_c_hp1_0_0.vlm"},
	{0xBCEA72C1CCE4DFBBULL, "art\\characters\\playercharacter\\humanfemale\\arm_medc.msh"},
	{0xBCEA84C1CCE4FE06ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medc.mao"},
	{0xBCEA90C1CCE512A5ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medc.mmh"},
	{0xBCEB354939BCE1FEULL, "blood_w_s.dds"},
	{0xBCEB395BC52D693EULL, "mrl000_marethari_act3.stg"},
	{0xBCEC366C9A603F13ULL, "vxdr_primevaltomb_7p4_0_0.vlm"},
	{0xBCEC509EF8FCE150ULL, "6080611_m.fxe"},
	{0xBCEC74B30C82A50BULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_dsmla_0n.dds"},
	{0xBCEC8611143C0B67ULL, "vxw_blightlands_wm_0_0.vlm"},
	{0xBCECD632ED9D87E3ULL, "6113040_m.fxe"},
	{0xBCED110C7C09E51EULL, "dae200_saemus6091690.cl"},
	{0xBCED13E56E4D1BD5ULL, "vxlt_darktown_4o6_0_0.vlm"},
	{0xBCED174567382A3EULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.dura.gad"},
	{0xBCED1E52B8CE90FCULL, "lgt220_mag_tmplr_templar6140347.cl"},
	{0xBCED23FE154373B5ULL, "t3_cor_h03.tnt"},
	{0xBCED2D9D6003AB73ULL, "zz_anderscu2.cut"},
	{0xBCED707B3F5A47D7ULL, "458986162.ogg"},
	{0xBCEDBD61A41BE63FULL, "cry_val_cycle.mao"},
	{0xBCEDC843B9C78861ULL, "6119105_m.fxe"},
	{0xBCEE67E3E55A342EULL, "lt_backalley_g_3_0.rml"},
	{0xBCEEA410783ECA2AULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured.gad"},
	{0xBCEEB6FFEA066FB8ULL, "gen00fl_isabela6088255.cl"},
	{0xBCF03AA53E85885CULL, "559554939.ogg"},
	{0xBCF05429DBD686E8ULL, "vxlt_undercitydungeon_1tt_0_0.vlm"},
	{0xBCF102FFDA97FCF5ULL, "569728912.ogg"},
	{0xBCF191AD608E4665ULL, "hlac_lt_0_lowtown_a2_1_6y.dds"},
	{0xBCF1CA51493AE84FULL, "plt_zz_lgtpt_debug.plo"},
	{0xBCF1CAE66D4404FAULL, "mer320_hubert6053686.cl"},
	{0xBCF2B301F9BC296BULL, "vxw_sundermount_37a_0_1.vlm"},
	{0xBCF3F245673E0A5DULL, "art\\vfx\\combat\\vfx_ha_bd_bspew_aoe_p.dura.evt"},
	{0xBCF3FA71A777EFF3ULL, "vxw_sundermount_2h0_0_0.vlm"},
	{0xBCF45F1DDA030EDEULL, "art\\characters\\playercharacter\\humanfemale\\mt_ulm_basa.mao"},
	{0xBCF4B7FEBF224249ULL, "796400515.ogg"},
	{0xBCF50762F72F9BBAULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_basic_2.mao"},
	{0xBCF540E3E56010D2ULL, "lt_backalley_g_3_0.phy"},
	{0xBCF587107844B7C1ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured.evt"},
	{0xBCF678337C1E0108ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comd_0t.dds"},
	{0xBCF680E209E98347ULL, "vxdr_primevalentrance_2q3_0_1.vlm"},
	{0xBCF7A5F7C43A2AF1ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoirea.msh"},
	{0xBCF7B3F7C43A420CULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoirea.mao"},
	{0xBCF7BFF7C43A56AFULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoirea.mmh"},
	{0xBCF7D0FAFFB25CACULL, "6141102_f.fxe"},
	{0xBCF7E963955E5D9EULL, "gen00fl_aveline6137491.cl"},
	{0xBCF7F0CD1647D6C8ULL, "6034872_m.fxe"},
	{0xBCF8021B4F1645A6ULL, "vxw_encounter1_dc_0_0.vlm"},
	{0xBCF80793D81DEF83ULL, "vxw_bonepitdestroyed_a3_1gq_0_0.vlm"},
	{0xBCF82D1C94825C04ULL, "367694694.ogg"},
	{0xBCF8E08C5E2D21C5ULL, "6059329_f.fxe"},
	{0xBCF90AEA7E6DC0E5ULL, "mag370_meredith6171561.cl"},
	{0xBCF928537A0BC295ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12.mmh"},
	{0xBCF94A9DE5F3015FULL, "32289469.ogg"},
	{0xBCF94CCF53D90DD2ULL, "dae111ar_gamlens_house_nt.lst"},
	{0xBCF99170554A8319ULL, "vxht_0_hightown_b9a_0_0.vlm"},
	{0xBCFA01BE6130C126ULL, "698930326.ogg"},
	{0xBCFA69E3D27E4CE3ULL, "477183029.ogg"},
	{0xBCFA9A851996EF2AULL, "hlac_ht_mansion2_818_64z.dds"},
	{0xBCFB188990167619ULL, "hla_ht_0_hightown_a2_131073_0z.dds"},
	{0xBCFB59A9EC1D57E8ULL, "follower_banter6155736.cl"},
	{0xBCFB6A7A5BE25DF8ULL, "vxw_woundedcoast_14x_0_0.vlm"},
	{0xBCFB84169A4F3150ULL, "6020095_m.fxe"},
	{0xBCFB96344AE61040ULL, "vxht_playershouse_n_1qa_0_0.vlm"},
	{0xBCFBA787EE3D165EULL, "art\\vfx\\combat\\vfx_c_mer_grab_stab_c.anb"},
	{0xBCFBC65A3B3CE319ULL, "vxca_cavern1_haunted_gkb_0_1.vlm"},
	{0xBCFC174BC1F22E01ULL, "vxlt_0_lowtown_n_gjr_0_0.vlm"},
	{0xBCFCA6BB0997771EULL, "llac_w_encounter1_65541_0.dds"},
	{0xBCFCA8C094C0C150ULL, "50766181.ogg"},
	{0xBCFCC3B60BB99B21ULL, "455733530.ogg"},
	{0xBCFD0DA762CAABC8ULL, "407755105.ogg"},
	{0xBCFD1B4F8DD0CF51ULL, "ttx_roof_int_b_1.mmh"},
	{0xBCFD354F8DD0FB0FULL, "ttx_roof_int_b_1.msh"},
	{0xBCFDBCC8F024FB08ULL, "864691301.ogg"},
	{0xBCFDC34B91891E74ULL, "855320690.ogg"},
	{0xBCFEA5A86943AD67ULL, "6071515_m.fxe"},
	{0xBCFECD1A2E2C10C6ULL, "6060700_f.fxe"},
	{0xBCFEE4ECC69C0CA2ULL, "vxlt_0_lowtown_a2_d6y_0_0.vlm"},
	{0xBCFEE874D32C2147ULL, "6118193_m.fxe"},
	{0xBCFEF230C7644DC3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chantry_statues_sa.dds"},
	{0xBCFEF93095F3E616ULL, "hlac_dr_slavers_3215_8y.dds"},
	{0xBCFF122DB4AC06DCULL, "6069631_f.fxe"},
	{0xBCFF1FF32E01E2ADULL, "vxht_mansion2_24v_0_0.vlm"},
	{0xBD002AA7E8713745ULL, "248043590.ogg"},
	{0xBD0046A5A8167419ULL, "hlac_w_woundedcoast_196608_0z.dds"},
	{0xBD00487278CD872AULL, "49307445.ogg"},
	{0xBD00A307D68D9580ULL, "llac_ca_caverns1_mines_24421_1.dds"},
	{0xBD0124DEEBB241EBULL, "vxw_sundermount_32b_0_1.vlm"},
	{0xBD024BE82807A918ULL, "vxlt_gallowstemplar_n_5wu_0_0.vlm"},
	{0xBD026C4E996E0873ULL, "vxw_sundermount_2m3_0_0.vlm"},
	{0xBD027F10784F3410ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured.ani"},
	{0xBD0280823AA1F713ULL, "hlac_lt_backalley_g_0_9x.dds"},
	{0xBD0351FC655DECA4ULL, "hlac_ht_mansion2b_818_8z.dds"},
	{0xBD03C5B7B603FFA7ULL, "vxlt_gallowstemplar_n_26f_0_0.vlm"},
	{0xBD04416455DB02EAULL, "vxht_0_hightown_n_c5p_0_0.vlm"},
	{0xBD0446F2C08F8683ULL, "hlac_lt_0_lowtown_docks_n_-196610_1y.dds"},
	{0xBD04575E474CD483ULL, "vxdr_primevaltomb_8nm_0_0.vlm"},
	{0xBD04DC1D15C825CBULL, "vxw_bonepitmining_a2_ef_0_0.vlm"},
	{0xBD05FDF6038471E3ULL, "6080771_m.fxe"},
	{0xBD0673F8410C5E26ULL, "vxw_encounter1_id_0_0.vlm"},
	{0xBD095E36D779BEAFULL, "6129269_m.fxe"},
	{0xBD09832411B73E06ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomec_d.dds"},
	{0xBD09A44892EF55D1ULL, "6024573_m.fxe"},
	{0xBD0A1D0921D7996EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_left.evt"},
	{0xBD0AE0D3C742AC1FULL, "6048762_m.fxe"},
	{0xBD0B085FDAB6C2DBULL, "gen00fl_isabela6079070.cl"},
	{0xBD0BBE76BD474461ULL, "127732139.ogg"},
	{0xBD0C8A1078830437ULL, "follower_banter6143316.cl"},
	{0xBD0D34844E0463F9ULL, "vxca_caverns1_dqc_0_0.vlm"},
	{0xBD0D5CADB2588FE5ULL, "vxlt_0_lowtown_a2_n_c_eom_0_0.vlm"},
	{0xBD0DC07211D439B9ULL, "fen110_fenris6118071.cl"},
	{0xBD0E0CA510FD3F2EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basicx.mmh"},
	{0xBD0E18A510FD53C5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basicx.mao"},
	{0xBD0E26A510FD6B1CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basicx.msh"},
	{0xBD0E83DCAC2B17B0ULL, "w_bonepitdestroyed_a3_30_0.mao"},
	{0xBD0E87DCAC2B1EC3ULL, "w_bonepitdestroyed_a3_30_0.mmh"},
	{0xBD0E95DCAC2B361DULL, "w_bonepitdestroyed_a3_30_0.msh"},
	{0xBD0E9B1844080C4AULL, "118009573.ogg"},
	{0xBD0EC2736C211BE6ULL, "678723533.ogg"},
	{0xBD0EDCC6F2E25D68ULL, "6180308_m.fxe"},
	{0xBD0F0C93EA8CFC55ULL, "qun131_taskmaster6177186.cl"},
	{0xBD0F4754BBFD19CEULL, "hlac_lt_0_lowtown_docks_a3_65534_0z.dds"},
	{0xBD10140921DBF519ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_left.gad"},
	{0xBD10BDC519FDC198ULL, "vxlt_gallowstemplar_n_5jr_0_0.vlm"},
	{0xBD116FD088BAA351ULL, "413147564.ogg"},
	{0xBD11805684269B2EULL, "hlac_ht_chantry_n_2534_15z.dds"},
	{0xBD11EBD89D484307ULL, "qasoundmaterials.arl"},
	{0xBD1229443DFCE426ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_imp_c.impa.evt"},
	{0xBD129235BD88A11EULL, "6098766_m.fxe"},
	{0xBD12B34147D11B6AULL, "vxht_0_hightown_n_c2q_0_0.vlm"},
	{0xBD12C93B3942ED03ULL, "vxdr_primevaltomb_8il_0_0.vlm"},
	{0xBD13289BD7FD7929ULL, "268190353.ogg"},
	{0xBD13AC26880493BAULL, "dae000_arishok6068547.cl"},
	{0xBD13C26ADF643D9BULL, "6131152_m.fxe"},
	{0xBD13CF80961310B5ULL, "follower_banter6136710.cl"},
	{0xBD13F1CD672F1923ULL, "vxht_0_hightown_n_bu8_0_0.vlm"},
	{0xBD13F34052869A79ULL, "6065930_m.fxe"},
	{0xBD14463FAA8717C6ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_glv_cas_0d.dds"},
	{0xBD14958E3390EBDEULL, "gen00fl_fenris6110880.cl"},
	{0xBD14B7B9D19F2351ULL, "hlac_lt_0_lowtown_a2_-65534_0x.dds"},
	{0xBD14D78B60A5B800ULL, "hlac_w_bonepitmining_a2_0_6z.dds"},
	{0xBD14EB4DBC0A2083ULL, "vxw_bonepitdestroyed_a3_1ms_0_0.vlm"},
	{0xBD15543ECA4DDAD8ULL, "vxca_caverns2prc_4ob_0_0.vlm"},
	{0xBD15A5D3BB01B795ULL, "vxlt_0_lowtown_a2_n_c_es9_0_0.vlm"},
	{0xBD15CB80ED9BB893ULL, "art\\vfx\\combat\\vfx_glb_sleep_c.impa.ani"},
	{0xBD1688D22D3665C1ULL, "6079679_f.fxe"},
	{0xBD16A1D450629EBBULL, "6099175_m.fxe"},
	{0xBD16A756EE2A9F26ULL, "ob_star.mao"},
	{0xBD16B2BB0276993DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appb_0d.dds"},
	{0xBD16D2CB28776834ULL, "6000800_m.fxe"},
	{0xBD17250921E23167ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wander_left.ani"},
	{0xBD175D925DEAF29AULL, "6115281_m.fxe"},
	{0xBD179D1C28E0DF54ULL, "vxca_caverns2prc_a0j_0_0.vlm"},
	{0xBD1828A04B017F76ULL, "832974553.ogg"},
	{0xBD183D5324E7C12BULL, "6045431_m.fxe"},
	{0xBD1840443E017771ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_imp_c.impa.gad"},
	{0xBD18791541BE0367ULL, "gen00fl_isabela6100011.cl"},
	{0xBD18AF99F2C43644ULL, "6043148_m.fxe"},
	{0xBD19797484732FAEULL, "vxlt_darktown_5jo_0_0.vlm"},
	{0xBD1A5807B3738259ULL, "vxw_encounter3_db_0_0.vlm"},
	{0xBD1A9119FAD78297ULL, "6060315_m.fxe"},
	{0xBD1A9FA2EEBBF93AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_lean_forward.gad"},
	{0xBD1AA384160A9596ULL, "6077911_f.fxe"},
	{0xBD1AEE608BC7E4EBULL, "vxca_cave_2bf_0_0.vlm"},
	{0xBD1B1486181505DCULL, "seb000pt_midstage.ncs"},
	{0xBD1B1AE732DA1027ULL, "zz_mag_debug6054606.cl"},
	{0xBD1B2F27E43D44E3ULL, "60902226.ogg"},
	{0xBD1B6463166E07A1ULL, "6206276_m.fxe"},
	{0xBD1BF2C7D7E2B491ULL, "6139508_f.fxe"},
	{0xBD1BF78A911B8D6BULL, "static_static_p4.vsh.fxo.dx11.assembly"},
	{0xBD1CA538F7C08D67ULL, "620249239.ogg"},
	{0xBD1D17CEDC714856ULL, "dae000_male_templar_mob6174373.cl"},
	{0xBD1D2E42F7676B18ULL, "vxdr_0_deeproads_adc_0_0.vlm"},
	{0xBD1D40EBAE3DFEF8ULL, "hlac_dr_slavers_3216_6x.dds"},
	{0xBD1D562719B3052EULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.cess.evt"},
	{0xBD1D78FE78AA72D9ULL, "gen00fl_isabela_init6077691.cl"},
	{0xBD1F41443E07971FULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_imp_c.impa.ani"},
	{0xBD1FC2329480EEF7ULL, "vxlt_0_lowtown_feg_0_0.vlm"},
	{0xBD1FC4A9C632866EULL, "vxlt_gallowstemplar_3qp_0_0.vlm"},
	{0xBD1FEE94F2FCBBF5ULL, "218587587.ogg"},
	{0xBD20611197DD71E1ULL, "llac_lt_0_lowtown_docks_n_131071_5.dds"},
	{0xBD20883AEB028139ULL, "35680140.ogg"},
	{0xBD20BE2ECA265274ULL, "597146633.ogg"},
	{0xBD20FE9C48D9EF0BULL, "ter_brickround01_n.dds"},
	{0xBD21A2A2EEC21D51ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_lean_forward.evt"},
	{0xBD21BFD6F9B456CBULL, "vxw_bonepitmining_a2_ox_0_0.vlm"},
	{0xBD2221B258340AEFULL, "mag170_recruit026094514.cl"},
	{0xBD2252193D6804D9ULL, "vxlt_gallowstemplar_n_46v_0_0.vlm"},
	{0xBD2257974D388F5BULL, "454968016.ogg"},
	{0xBD228AFC33090995ULL, "331917092.ogg"},
	{0xBD22A97ECFB4D144ULL, "hlac_lt_gamlenshouse_n_-1_1y.dds"},
	{0xBD22DFEB0A2D467BULL, "zz_vrc_debug6064269.cl"},
	{0xBD22F0A09443B111ULL, "vxw_encounter2_6w_0_0.vlm"},
	{0xBD22F74C6A47580DULL, "hlac_lt_0_lowtown_a2_65534_5x.dds"},
	{0xBD2379E89357EBB0ULL, "mag101_grace6034085.cl"},
	{0xBD23BE8714C4C1E2ULL, "vxlt_0_lowtown_fh5_0_0.vlm"},
	{0xBD23BF9BD1ECF214ULL, "dae000cr_pol.utc"},
	{0xBD23C380EDA7E8CAULL, "art\\vfx\\combat\\vfx_glb_sleep_c.impa.evt"},
	{0xBD244D2719B914D9ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.cess.gad"},
	{0xBD24584C9BDB4A5DULL, "dae000_gamlen6088141.cl"},
	{0xBD24AF0F0F86B225ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_nud_0s.dds"},
	{0xBD24E5F08E3C6564ULL, "vxht_0_hightown_c15_0_0.vlm"},
	{0xBD2589A2999D3CA6ULL, "176057620.ogg"},
	{0xBD2604B46C3D8909ULL, "hlac_ca_caverns1_mines_24422_4y.dds"},
	{0xBD2605BA4D3C51B5ULL, "llac_ca_caverns2_5802_68.dds"},
	{0xBD26348A9A15ABE4ULL, "gen00fl_isabela6085780.cl"},
	{0xBD2652ABA0C39D62ULL, "vxlt_gallowstemplar_n_3rv_0_0.vlm"},
	{0xBD2687DCAC3FB0E0ULL, "w_bonepitdestroyed_a3_30_0.rml"},
	{0xBD2711C99629113FULL, "lt_0_lowtown_docks_124_0.msh"},
	{0xBD2717C996291BE1ULL, "lt_0_lowtown_docks_124_0.mmh"},
	{0xBD2723C996292F82ULL, "lt_0_lowtown_docks_124_0.mao"},
	{0xBD27796702AFBC12ULL, "6170095_m.fxe"},
	{0xBD28A87A183F5A5DULL, "hlac_ht_bartrand_n_2995_7x.dds"},
	{0xBD28C9E4A5699AD9ULL, "vxw_encounter3_aa_0_0.vlm"},
	{0xBD2994CE265848C8ULL, "lt_0_lowtown_docks_a3_n_5_0.rml"},
	{0xBD29A1FFFA014DF2ULL, "art\\vfx\\materials\\chunky\\vfx_c_rockarmor.mao"},
	{0xBD29BEA30308A031ULL, "6163643_m.fxe"},
	{0xBD29C20BD84A0F85ULL, "vxw_woundedcoast_14a_0_0.vlm"},
	{0xBD2A25C64EBB4A7DULL, "6030613_m.fxe"},
	{0xBD2AA280EDADCF3DULL, "art\\vfx\\combat\\vfx_glb_sleep_c.impa.gad"},
	{0xBD2AD236B5FE0A80ULL, "vrc221_bartrand6043019.cl"},
	{0xBD2AEFCDCF9E655FULL, "vxlt_0_lowtown_a3_n_c_fgi_0_0.vlm"},
	{0xBD2AFB7EEB320880ULL, "432501771.ogg"},
	{0xBD2B5E2719BF5027ULL, "art\\vfx\\combat\\vfx_glb_cold_slow_c.cess.ani"},
	{0xBD2B941A0E390B23ULL, "vxht_keep_6l0_0_0.vlm"},
	{0xBD2BD7358E2F45C7ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_left_hand.gad"},
	{0xBD2C1172B52C7E6BULL, "dae200_saemus6142809.cl"},
	{0xBD2CDB4EBE569CBFULL, "hlac_w_encounter3_0_2x.dds"},
	{0xBD2DAA480B5E11BCULL, "592687454.ogg"},
	{0xBD2E69699ADF9919ULL, "vxdr_primevaltomb_6up_0_0.vlm"},
	{0xBD2EBAA2EECCD020ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_d_lean_forward.ani"},
	{0xBD2EFA860ED2EA54ULL, "6167472_f.fxe"},
	{0xBD2F68F7C9F0845CULL, "vxdr_primevaltomb_7jy_0_0.vlm"},
	{0xBD307AD3FC41025AULL, "qcr220cr_qunari_hunter.utc"},
	{0xBD30910030002165ULL, "follower_banter6161145.cl"},
	{0xBD3120B706998F0AULL, "hlac_lt_0_lowtown_docks_a3_n_196606_0x.dds"},
	{0xBD31265417D6A197ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.impa.gad"},
	{0xBD31363B1B7D07F1ULL, "6205170_m.fxe"},
	{0xBD31517DD56A696AULL, "vxht_brothel_n_59m_0_0.vlm"},
	{0xBD3169C1CD20D352ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medc.phy"},
	{0xBD318F002002F005ULL, "vrc321tr_spooky_scenes_1.utt"},
	{0xBD31BC7B1EB84C5AULL, "follower_banter6163245.cl"},
	{0xBD31FC0817C1305EULL, "vxlt_gallowscourtyard_4uy_0_0.vlm"},
	{0xBD3249749BCA0247ULL, "428711208.ogg"},
	{0xBD32CC358E3551D4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_left_hand.evt"},
	{0xBD32DF1ED47B6732ULL, "491783376.ogg"},
	{0xBD3395588E5A48C9ULL, "vxlt_ruinedhouse_n_ib_0_0.vlm"},
	{0xBD33CEED6134B0C9ULL, "vxht_0_hightown_a3_9zn_0_0.vlm"},
	{0xBD3430908773D60DULL, "595887146.ogg"},
	{0xBD34DBFD3EDCE9ACULL, "6180915_m.fxe"},
	{0xBD35979F60D5B1F6ULL, "art\\characters\\creatures\\spider\\animation\\rn_lnl.ani"},
	{0xBD35D6C896931F67ULL, "964612081.ogg"},
	{0xBD361E3F41A6024FULL, "hlac_lt_gallowstemplar_n_2383_4z.dds"},
	{0xBD366C673865C38EULL, "follower_banter6143122.cl"},
	{0xBD36784F567635C9ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg106a.phy"},
	{0xBD36EE00380746B0ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\3ca_basic_1.mao"},
	{0xBD36FFABE0E7CE6AULL, "vxw_woundedcoast_n5_0_0.vlm"},
	{0xBD3755A7D012ADDCULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopeningmage.ani"},
	{0xBD383B5417DCE424ULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.impa.evt"},
	{0xBD38C30E66785090ULL, "vxw_encounter2_bf_0_0.vlm"},
	{0xBD39A18CAABF9E2AULL, "vxlt_gallowstemplar_f_4qp_0_0.vlm"},
	{0xBD39FC62700693E8ULL, "static_lava_p1.psh.fxo.dx11.assembly"},
	{0xBD3A05F7002F23A3ULL, "vxht_keep_6i3_0_0.vlm"},
	{0xBD3A7A1C7F3F9361ULL, "xx_extstone.dds"},
	{0xBD3BE95149795534ULL, "plt_zz_lgtpt_debug.nss"},
	{0xBD3C074F01C698DEULL, "491076283.ogg"},
	{0xBD3C623135D08C95ULL, "dae000ip_crowd_mf3_talk2.utp"},
	{0xBD3CDB468CD5B199ULL, "vxdr_primevaltomb_6xw_0_0.vlm"},
	{0xBD3D27A0BF773D83ULL, "328709255.ogg"},
	{0xBD3D62A7D0172F1EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopeningmage.gad"},
	{0xBD3D8CC0CDCBE017ULL, "vxlt_0_lowtown_a2_cmj_0_0.vlm"},
	{0xBD3DB42C23D2ACA7ULL, "582612193.ogg"},
	{0xBD3E2FA32C318C89ULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.mmh"},
	{0xBD3EDB3E81AED4E6ULL, "6098199_m.fxe"},
	{0xBD3EDC51FF2E0B2FULL, "em_amb_tranquil.mor"},
	{0xBD3F0CF762E3273EULL, "lowtown_male_beggar_sta6151029.cl"},
	{0xBD3F116922A3B2EAULL, "vxht_0_hightown_a2_c7e_0_0.vlm"},
	{0xBD3F3AB896818145ULL, "6093773_m.fxe"},
	{0xBD3F7878C0BA64D0ULL, "cd_mag131_taraneh.wbd"},
	{0xBD3FC4358E3FCE0DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_left_hand.ani"},
	{0xBD4033EC068084B4ULL, "vxht_brothel_n_5m6_0_1.vlm"},
	{0xBD405FAE24E81320ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_5x.dds"},
	{0xBD408281D1EAA1BAULL, "vxht_mansion_n_3ho_0_0.vlm"},
	{0xBD40CED311839A94ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforwardright.ani"},
	{0xBD41341228A231D3ULL, "vxlt_0_lowtown_a2_n_c_hjq_0_0.vlm"},
	{0xBD41813D5B00609DULL, "6171110_m.fxe"},
	{0xBD41EC786E897A49ULL, "gen_im_npc_elf_arm_cht_dalb.uti"},
	{0xBD429F9F60E049E7ULL, "art\\characters\\creatures\\spider\\animation\\rn_lnl.evt"},
	{0xBD42AC8476508DC5ULL, "149450497.ogg"},
	{0xBD42D92DB2C6D347ULL, "fx_circle.dds"},
	{0xBD433AAC720DDB2EULL, "201514793.ogg"},
	{0xBD43C5375CD1A476ULL, "6170227-pridedrain-d.cut"},
	{0xBD44341D5BE3182EULL, "follower_banter6143276.cl"},
	{0xBD45535417E7971DULL, "art\\vfx\\combat\\vfx_v_fir_single_proj_c.impa.ani"},
	{0xBD455FEE39D09130ULL, "zz_dre_debug6054630.cl"},
	{0xBD4586726D36D1CEULL, "vxdr_primevaltomb_3m3_0_1.vlm"},
	{0xBD4652480CB80589ULL, "654855316.ogg"},
	{0xBD46BBCBC044BBBFULL, "257049622.ogg"},
	{0xBD472093E1C8EE9EULL, "hlac_dr_slavers_3216_7y.dds"},
	{0xBD4734EB586E6910ULL, "vxw_encounter2_mg_0_0.vlm"},
	{0xBD479A4461EEEA43ULL, "106571696.ogg"},
	{0xBD47CBD31189B4F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\walkforwardright.gad"},
	{0xBD483C980B39367CULL, "6098795_m.fxe"},
	{0xBD487A8D73897566ULL, "art\\characters\\creatures\\mubari\\animation\\brk_ex.evt"},
	{0xBD48C60F264AE97DULL, "556493333.ogg"},
	{0xBD48DECFB2F1F9A9ULL, "6180297_m.fxe"},
	{0xBD49814DED140C29ULL, "hlac_lt_undercityencounter_-1_24y.dds"},
	{0xBD49C09F60E6A030ULL, "art\\characters\\creatures\\spider\\animation\\rn_lnl.gad"},
	{0xBD4A79618BD07F2DULL, "mag101tr_aom_way_out_2.utt"},
	{0xBD4AB4C7E9F6D052ULL, "vxlt_0_lowtown_a2_n_c_hjt_0_0.vlm"},
	{0xBD4B70B7EB411D18ULL, "vxdr_0_deeproads_d46_0_0.vlm"},
	{0xBD4B7EA76DD5FCEFULL, "fen310_varania6118559.cl"},
	{0xBD4B9F6704A5BBD5ULL, "llac_w_sundermount_-131073_0.dds"},
	{0xBD4C17C1EE0AF18DULL, "6164354_m.fxe"},
	{0xBD4C57F150F816CAULL, "vxw_bonepit_8n_0_0.vlm"},
	{0xBD4C63BD872273F3ULL, "6165250_m.fxe"},
	{0xBD4C72C25A4774C8ULL, "gen00fl_merrill6130040.cl"},
	{0xBD4C90AF011B4F03ULL, "vxdr_primevaltomb_8up_0_0.vlm"},
	{0xBD4C90B93CF6D5B8ULL, "gen_im_arm_glv_med_set_high.uti"},
	{0xBD4CB85FAA7E6420ULL, "vxlt_0_lowtown_a3_n_c_f4o_0_0.vlm"},
	{0xBD4CF2C68FAFFD76ULL, "vxw_encounter2_camp_pr_0_0.vlm"},
	{0xBD4CF4ABD3D24BCDULL, "192924912.ogg"},
	{0xBD4D43E10EB6C0C6ULL, "138361107.ogg"},
	{0xBD4D7DAD16DE20A3ULL, "6148875_m.fxe"},
	{0xBD4DC91F55344AECULL, "llac_lt_0_lowtown_docks_n_1_0.dds"},
	{0xBD4DCA69782E04A2ULL, "vxlt_gallowstemplar_n_6q6_0_0.vlm"},
	{0xBD4E0D4BC18E5C8CULL, "follower_banter6164703.cl"},
	{0xBD4F2605A0259EE1ULL, "6121093_f.fxe"},
	{0xBD4F5F452F72B34BULL, "6121926_m.fxe"},
	{0xBD4F7E5099CEE6E3ULL, "art\\characters\\creatures\\mubari\\animation\\pickup.evt"},
	{0xBD4F918D738FBAB1ULL, "art\\characters\\creatures\\mubari\\animation\\brk_ex.gad"},
	{0xBD4FB2C36F1D4406ULL, "llac_lt_0_lowtown_65534_0.dds"},
	{0xBD4FD65F3840ADB9ULL, "195525620.ogg"},
	{0xBD501CA10239596CULL, "t3_wer_b02.tnt"},
	{0xBD503C387C1FBD6EULL, "465244322.ogg"},
	{0xBD513CE6D8DC88B9ULL, "vxlt_hangedman_n_10h_0_0.vlm"},
	{0xBD516F7A38E9E099ULL, "6129484_f.fxe"},
	{0xBD5171AC7FD147A6ULL, "vxw_bonepitdestroyed_a3_26t_0_0.vlm"},
	{0xBD5179E98D299B80ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_meda_d.dds"},
	{0xBD518E328CAC3E42ULL, "vxw_encounter2_fh_0_0.vlm"},
	{0xBD51C21BB60EB62EULL, "229605815.ogg"},
	{0xBD52088418C05A74ULL, "hlac_w_sundermount_65530_0z.dds"},
	{0xBD5262D735B5F757ULL, "vxlt_backalley_g_1ea_0_0.vlm"},
	{0xBD52858495B75BD3ULL, "hlac_lt_0_lowtown_-65537_5z.dds"},
	{0xBD52B4F9AB3CFC65ULL, "6049752_m.fxe"},
	{0xBD532433F9823932ULL, "vxht_0_hightown_a2_7j2_0_0.vlm"},
	{0xBD53321D01CFE96DULL, "223410036.ogg"},
	{0xBD53767B544F415DULL, "vxw_woundedcoast_71_0_0.vlm"},
	{0xBD537EE24093DC34ULL, "vxlt_hangedman_n_211_0_0.vlm"},
	{0xBD5431CC2954E4ECULL, "vxht_mansion2_3eq_0_0.vlm"},
	{0xBD544D7446A0332DULL, "9247072.ogg"},
	{0xBD546570B1971856ULL, "character_character_p188.psh.fxo.dx11.assembly"},
	{0xBD549A0F2A6AA003ULL, "6026991_m.fxe"},
	{0xBD556DA3A2D71E9FULL, "vxdr_primevaltomb_8zg_0_2.vlm"},
	{0xBD55A6C84A648190ULL, "vxw_encounter2_hd_0_0.vlm"},
	{0xBD55D631748120CAULL, "vxlt_0_lowtown_a3_n_c_eow_0_0.vlm"},
	{0xBD55E1B1071C7F94ULL, "50358085.ogg"},
	{0xBD5630846D0C743EULL, "322321140.ogg"},
	{0xBD56575099D4C364ULL, "art\\characters\\creatures\\mubari\\animation\\pickup.gad"},
	{0xBD5666C890652D27ULL, "438024138.ogg"},
	{0xBD56928D7395DB5FULL, "art\\characters\\creatures\\mubari\\animation\\brk_ex.ani"},
	{0xBD569967624428C4ULL, "281430541.ogg"},
	{0xBD56C6BDB7E50A9EULL, "1010359047.ogg"},
	{0xBD56EAB5FD566AAEULL, "dae000_epic_merchant_36179497.cl"},
	{0xBD576B3A776CBD22ULL, "vxht_0_hightown_a3_buc_0_0.vlm"},
	{0xBD57EB328E9080C6ULL, "narrative_varric6172495.cl"},
	{0xBD587EBE5E6841B3ULL, "llac_ht_mansion_n_3367_0.dds"},
	{0xBD589DA95C77CC92ULL, "vxca_caverns2prc_8kq_0_1.vlm"},
	{0xBD58B1D3EE0CAB5FULL, "w_sundermount_137_0.rml"},
	{0xBD58CB75959FE724ULL, "vxht_0_hightown_a3_cbd_0_1.vlm"},
	{0xBD58FF813C89469BULL, "226664089.ogg"},
	{0xBD595E380326C102ULL, "6149600_m.fxe"},
	{0xBD5A2F815070DF77ULL, "zz_mer_debug6180856.cl"},
	{0xBD5A383A6F383925ULL, "follower_banter6163969.cl"},
	{0xBD5A4BEDFD79088CULL, "rog_notattack.dds"},
	{0xBD5A654BB42B213BULL, "365014604.ogg"},
	{0xBD5AC5FF2D978E10ULL, "105189097.ogg"},
	{0xBD5AD7299DBE27E0ULL, "cnvprev_human_male_rogue.utc"},
	{0xBD5AD78A0DED892EULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar1_.ani"},
	{0xBD5ADC6D713C40BBULL, "555610946.ogg"},
	{0xBD5B09B57F6EA2BEULL, "6173410_m.fxe"},
	{0xBD5BC63E71EED122ULL, "6117184_m.fxe"},
	{0xBD5C21BE50EA585CULL, "6091961_m.fxe"},
	{0xBD5CD943B61BC6E2ULL, "dae000_keep_male_servant_mo6160897.cl"},
	{0xBD5D528574CE6387ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.impa.gad"},
	{0xBD5D5663F23FA310ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_0y.dds"},
	{0xBD5D665099DAFB42ULL, "art\\characters\\creatures\\mubari\\animation\\pickup.ani"},
	{0xBD5D68E9E875A54EULL, "vxlt_0_lowtown_docks_a3_bv5_0_0.vlm"},
	{0xBD5D8ACEEF9656C6ULL, "94230987.ogg"},
	{0xBD5DF9B9484070F6ULL, "6097381_m.fxe"},
	{0xBD5E66135500C2EEULL, "static_static_p21.psh.fxo"},
	{0xBD5E7B3D0099856FULL, "nb.cut"},
	{0xBD5ED85B75AA0C88ULL, "vxlt_0_lowtown_bz5_0_0.vlm"},
	{0xBD5FC2BA7CFFAB31ULL, "1069280636.ogg"},
	{0xBD5FD0D3EE12FEF7ULL, "w_sundermount_137_0.phy"},
	{0xBD5FEAA881DE57BFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffold_1.mao"},
	{0xBD60000F7EA256C2ULL, "vxw_encounter2_ai_0_0.vlm"},
	{0xBD60A426B5D52EDAULL, "tbe100ip_shop_codex_5.utp"},
	{0xBD60DC4E5CE52084ULL, "hlac_ht_keep_n_4519_5z.dds"},
	{0xBD61128444D8B546ULL, "vxht_keep_62s_0_0.vlm"},
	{0xBD6136D97A47DE7AULL, "w_woundedcoast_154_0.phy"},
	{0xBD6148A9708C4D36ULL, "hlac_lt_0_lowtown_a2_1_11x.dds"},
	{0xBD629F635EB3EBB1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodboards.msh"},
	{0xBD62AD635EB403CCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodboards.mao"},
	{0xBD62B9635EB4186FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodboards.mmh"},
	{0xBD62C5178734CD48ULL, "783785459.ogg"},
	{0xBD62C92F0654C674ULL, "692057461.ogg"},
	{0xBD62F23BBD86C91BULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.cess.ani"},
	{0xBD630F9307543477ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_hed_flma_0n.dds"},
	{0xBD63614F569CFCEEULL, "art\\characters\\weapons\\dagger\\w_dgr_dg106a.mmh"},
	{0xBD636A79094C22EDULL, "dog_sl_02.mao"},
	{0xBD637B4F569D28DCULL, "art\\characters\\weapons\\dagger\\w_dgr_dg106a.msh"},
	{0xBD63C4E2AC46836BULL, "527466018.ogg"},
	{0xBD63CDBB50511A70ULL, "w_bonepit_147_0.msh"},
	{0xBD63DBBB5051318AULL, "w_bonepit_147_0.mmh"},
	{0xBD63DFBB505138C1ULL, "w_bonepit_147_0.mao"},
	{0xBD64324A8EC6B619ULL, "1036642573.ogg"},
	{0xBD64478574D46F94ULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.impa.evt"},
	{0xBD645406155BC39FULL, "1014732575.ogg"},
	{0xBD645458A1878DD5ULL, "81149239.ogg"},
	{0xBD647802802A5F81ULL, "6174153_m.fxe"},
	{0xBD648942F33BF17BULL, "300107525.ogg"},
	{0xBD64990646540301ULL, "lt_0_lowtown_a2_n_34_0_lc.mmh"},
	{0xBD65365EC290F966ULL, "524383034.ogg"},
	{0xBD656DC98A0F681FULL, "6092307_m.fxe"},
	{0xBD6571EA769A8396ULL, "dae021_brothel_onenter6196721.cl"},
	{0xBD658214F15B346EULL, "vxht_0_hightown_a2_atf_0_0.vlm"},
	{0xBD65BB055808654DULL, "llac_lt_gallowstemplar_n_7198_1.dds"},
	{0xBD65D213176B056BULL, "6081291_m.fxe"},
	{0xBD66096C3D2E20B4ULL, "392605865.ogg"},
	{0xBD663BD2E8D9B1A9ULL, "339298997.ogg"},
	{0xBD66CD40470D97BAULL, "vxlt_hangedman_n_1qt_0_0.vlm"},
	{0xBD672CA32C548CBEULL, "art\\vfx\\combat\\vfx_q_ari_buf_fr_lp_c.anb"},
	{0xBD681173F52C672FULL, "character_character_p112.psh.fxo.dx11.assembly"},
	{0xBD6851618B49D7A2ULL, "vxlt_0_lowtown_a2_n_hdd_0_0.vlm"},
	{0xBD69045CC7ED081EULL, "hm_vis_lowtown2.mor"},
	{0xBD69228040788A1BULL, "6021222_m.fxe"},
	{0xBD694B3333534973ULL, "325013977.ogg"},
	{0xBD695F2FBB9F3B83ULL, "lgt311_shw_mer_agatha6156491.cl"},
	{0xBD69622B30251251ULL, "w_blightlands_215_0.mmh"},
	{0xBD6962618BEB08A0ULL, "mag101tr_aom_way_out_2.ncs"},
	{0xBD696E2B302526F2ULL, "w_blightlands_215_0.mao"},
	{0xBD6978072F715F1DULL, "9191614.ogg"},
	{0xBD697C2B30253E0FULL, "w_blightlands_215_0.msh"},
	{0xBD69AEE2CF6322DDULL, "plt_cod_mgc_circle_heirarch.plo"},
	{0xBD69D680739C2E76ULL, "vxw_encounter2_camp_ft_0_0.vlm"},
	{0xBD69E0CE59F7B4C9ULL, "w_blightlands_54_0.rml"},
	{0xBD6A1B680284D230ULL, "6188660_f.fxe"},
	{0xBD6A2070D7B74A17ULL, "vxlt_gallowscourtyard_6x5_0_0.vlm"},
	{0xBD6A8C2E8499EB07ULL, "6072333_f.fxe"},
	{0xBD6AE544DB415A4FULL, "vxlt_warehouse_n_e6_0_0.vlm"},
	{0xBD6B01F6771071B6ULL, "vxlt_0_lowtown_a2_n_c_era_0_0.vlm"},
	{0xBD6C31044D23741DULL, "638391760.ogg"},
	{0xBD6C853D3DFD83E9ULL, "289105716.ogg"},
	{0xBD6C8E4D6CE0F51BULL, "6125266_f.fxe"},
	{0xBD6CD7F046F6DB6EULL, "static_static_p50.psh.fxo"},
	{0xBD6CF95A041420FDULL, "vxlt_0_lowtown_docks_n_cta_0_0.vlm"},
	{0xBD6D37A4802C42BBULL, "vxlt_0_lowtown_n_df3_0_0.vlm"},
	{0xBD6D97942AB20A79ULL, "dae150_chandan6148695.cl"},
	{0xBD6D9B81DA0E77A3ULL, "gen_im_cft_rec_arn_05.uti"},
	{0xBD6DFEBB10E63DCDULL, "rog_reverseattack.dds"},
	{0xBD6E1CC036601A42ULL, "hlac_w_sundermount_-262141_0z.dds"},
	{0xBD6E647217E84902ULL, "231982218.ogg"},
	{0xBD6E71EC70986F42ULL, "vxw_encounter2_lj_0_0.vlm"},
	{0xBD6ED4ABD3E15A64ULL, "mer220_hubert6088151.cl"},
	{0xBD6F07118200B24EULL, "6061784_m.fxe"},
	{0xBD6F3710E82E8066ULL, "vxca_caverns2_7uh_0_0.vlm"},
	{0xBD6FB08A0DFFA368ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar1_.gad"},
	{0xBD6FB74178F79B98ULL, "48653398.ogg"},
	{0xBD6FD2EC29FC07F8ULL, "gen00fl_anders6112921.cl"},
	{0xBD6FF6EFD1C7BA07ULL, "352577.ogg"},
	{0xBD6FFA3BBD916112ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.cess.evt"},
	{0xBD70479B8D2389E1ULL, "mag170_thrask6020532.cl"},
	{0xBD70A591AD3B1666ULL, "llac_w_encounter3_-327677_0.dds"},
	{0xBD70A792FABC08FFULL, "vxlt_hangedman_n_34h_0_0.vlm"},
	{0xBD70F1B1810E5928ULL, "809658068.ogg"},
	{0xBD71DD6978E143CBULL, "1072691319.ogg"},
	{0xBD71E35BFBB883DDULL, "114252200.ogg"},
	{0xBD723F8574E09FCDULL, "art\\vfx\\combat\\vfx_c_buff_friend_loop_c.impa.ani"},
	{0xBD72B9EB586D51CAULL, "vxlt_0_lowtown_a3_n_c_euq_0_0.vlm"},
	{0xBD7313832D9DC3FEULL, "202697336.ogg"},
	{0xBD73D4CE59FFF4A9ULL, "w_blightlands_54_0.mao"},
	{0xBD73DD840D53A8FCULL, "872481101.ogg"},
	{0xBD73E0CE5A000942ULL, "w_blightlands_54_0.mmh"},
	{0xBD73E2CE5A000CF8ULL, "w_blightlands_54_0.msh"},
	{0xBD73F4D58906281EULL, "6085931_m.fxe"},
	{0xBD7443F009611DEBULL, "6084191_m.fxe"},
	{0xBD756478CF5D7428ULL, "6206927_m.fxe"},
	{0xBD75D305D5E213E3ULL, "llac_w_woundedcoast_-65534_0.dds"},
	{0xBD75E2F2AE161508ULL, "vxht_0_hightown_a2_a7y_0_0.vlm"},
	{0xBD75E8A8C44E6E46ULL, "98931167.ogg"},
	{0xBD760FEFCF0A179EULL, "art\\vfx\\cinematic\\vfx_mth_arv_qunari_c.impa.evt"},
	{0xBD76265B75312807ULL, "dae000ip_crowd_em2_talk8.utp"},
	{0xBD76421677160C36ULL, "lgt110_gxa_coa_meeting6117506.cl"},
	{0xBD765F22C3D44969ULL, "vxw_encounter2_de_0_0.vlm"},
	{0xBD76CE25AF0AF146ULL, "gen00fl_anders6125187.cl"},
	{0xBD77160229D8DDF8ULL, "242254036.ogg"},
	{0xBD77193BBD97B4A5ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.cess.gad"},
	{0xBD77922C50FA4F19ULL, "vxlt_0_lowtown_n_f27_0_0.vlm"},
	{0xBD7795A5F4F6169DULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bottlewine_n.dds"},
	{0xBD77A24883BEFEDCULL, "vxdr_primevaltomb_7sr_0_0.vlm"},
	{0xBD781AE44282CE12ULL, "one220_yevhen6104153.cl"},
	{0xBD78E9B1AA26697AULL, "w_woundedcoast_37_0.rml"},
	{0xBD78EF3A4E1FDB12ULL, "static_static_p30.psh.fxo.dx11.assembly"},
	{0xBD795733BCEBC8FCULL, "517627611.ogg"},
	{0xBD7973D369068A36ULL, "vxlt_0_lowtown_a2_n_c_ewb_0_0.vlm"},
	{0xBD79C9D09B92C2E3ULL, "vxca_caverns2prc_9q2_0_0.vlm"},
	{0xBD7A4710AFDE7ADCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv4.ani"},
	{0xBD7A6422F4E4F333ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ww_basic_2.mao"},
	{0xBD7A6EB3135AA1F6ULL, "126217706.ogg"},
	{0xBD7AFB85F07A93D3ULL, "vxlt_0_lowtown_a2_n_c_hvu_0_0.vlm"},
	{0xBD7B002A5ED70D9BULL, "6122366_f.fxe"},
	{0xBD7B637E4F81CE4DULL, "vrc221_hugin6042562.cl"},
	{0xBD7BC46F77766045ULL, "6119831_m.fxe"},
	{0xBD7BCEA322E1C9ABULL, "201117771.ogg"},
	{0xBD7BF9F38658E91AULL, "6156548_f.fxe"},
	{0xBD7C01E0133CA4E7ULL, "6072711_f.fxe"},
	{0xBD7C1DE25222386BULL, "lgt200ip_eau_tome05.utp"},
	{0xBD7C26EFCF0EAAC9ULL, "art\\vfx\\cinematic\\vfx_mth_arv_qunari_c.impa.gad"},
	{0xBD7C681CD9D58CABULL, "vxlt_gallowstemplar_f_2rv_0_0.vlm"},
	{0xBD7C7F53E6A34BFFULL, "barr_02_f.mao"},
	{0xBD7E0C6067C989EFULL, "zz_dae_debug6047593.cl"},
	{0xBD7EAA538A7682D4ULL, "vxw_woundedcoast_1cq_0_0.vlm"},
	{0xBD7EC33F735E4C6BULL, "gen_im_trs_prf_sta_phy_r0.uti"},
	{0xBD7EE0D464D18AEEULL, "hlac_lt_gallowstemplar_n_2391_4x.dds"},
	{0xBD7F116FC2D13AEFULL, "204604182.ogg"},
	{0xBD7F1D6D0D81E431ULL, "vxlt_backalley_n_1x9_0_0.vlm"},
	{0xBD7F20EAA2F5BB5FULL, "vxlt_0_lowtown_haq_0_0.vlm"},
	{0xBD7F25193E19EB6DULL, "vxht_0_hightown_aur_0_0.vlm"},
	{0xBD7F38FC9AB648EAULL, "vxw_woundedcoast_w2_0_0.vlm"},
	{0xBD7F601012A7D2B6ULL, "atl_background_dxt5_dat.xml"},
	{0xBD7F77D246BE6734ULL, "ave121_jeven6105702.cl"},
	{0xBD7FA128C0814C89ULL, "vxht_0_hightown_a2_c9y_0_1.vlm"},
	{0xBD7FE261DE890FC8ULL, "438065202.ogg"},
	{0xBD8031B04D87C198ULL, "40685609.ogg"},
	{0xBD805410AFE2FC1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv4.gad"},
	{0xBD809429B2754CCBULL, "hf_one_har_guard3.mor"},
	{0xBD811F4AB4489375ULL, "vxht_playershouse_n_1pr_0_1.vlm"},
	{0xBD812BC84A636A4AULL, "vxlt_0_lowtown_a3_n_c_exr_0_0.vlm"},
	{0xBD81658FED668693ULL, "dae000_bodhan6146841.cl"},
	{0xBD81A2E1C4B73FE4ULL, "539690121.ogg"},
	{0xBD81AAEC8B214AC3ULL, "vxht_0_hightown_a2m_0_0.vlm"},
	{0xBD8265CED547656EULL, "vxht_0_hightown_a2_ajx_0_0.vlm"},
	{0xBD82698D35B98633ULL, "gen00fl_bethany6089994.cl"},
	{0xBD82D5252386563DULL, "346376475.ogg"},
	{0xBD82F7EFCF147917ULL, "art\\vfx\\cinematic\\vfx_mth_arv_qunari_c.impa.ani"},
	{0xBD84027E9564BFC3ULL, "llac_ca_cavern1_haunted_24424_14.dds"},
	{0xBD841E51996C2B43ULL, "one101_orders_followers6115899.cl"},
	{0xBD8454CFA00C2D88ULL, "vxht_0_hightown_a2_a2v_0_0.vlm"},
	{0xBD848EC923923C89ULL, "324331382.ogg"},
	{0xBD84AEE8167290DDULL, "vxlt_backalley_g_1jz_0_0.vlm"},
	{0xBD84B56A2C72E4C0ULL, "character_character_p65.psh.fxo.dx11"},
	{0xBD84BAB4B48FF5B5ULL, "vxlt_shop_d5_0_0.vlm"},
	{0xBD84D0FFB5CA61E9ULL, "vxw_encounter2_od_0_0.vlm"},
	{0xBD84F8BEF682D4C5ULL, "vxdr_primevalentrance_14l_0_0.vlm"},
	{0xBD850D9FB8FAA93DULL, "vxlt_0_lowtown_docks_n_b9n_0_1.vlm"},
	{0xBD85C603E8A7F4C7ULL, "dre000_varric6027786.cl"},
	{0xBD85FE4B15631411ULL, "6005734_m.fxe"},
	{0xBD86040942F06799ULL, "vxlt_0_lowtown_n_f56_0_0.vlm"},
	{0xBD864E78C0F61F92ULL, "cd_mag131_taraneh.bnk"},
	{0xBD8658135CFFD8DDULL, "vxca_caverns1_e56_0_0.vlm"},
	{0xBD869800204B3FC8ULL, "vrc321tr_spooky_scenes_1.ncs"},
	{0xBD871FD71C573D4DULL, "mag120_macha6080646.cl"},
	{0xBD872F10AFE8DC3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv4.evt"},
	{0xBD875DE149303A20ULL, "520001440.ogg"},
	{0xBD8771955039644FULL, "genip_pp_drk_sustained.utp"},
	{0xBD884F1D3816D932ULL, "760002067.ogg"},
	{0xBD88DD65A5FA2F05ULL, "342688240.ogg"},
	{0xBD88DEFA5E9696C5ULL, "vxht_0_hightown_a2_b1q_0_0.vlm"},
	{0xBD890EF25B30E298ULL, "dae000_flemeth6001287.cl"},
	{0xBD89784242A26690ULL, "llac_w_bonepit_65542_0.dds"},
	{0xBD8993D07EB7845BULL, "749713574.ogg"},
	{0xBD899670CC44D0FFULL, "llac_ht_playershouse_n_855_15.dds"},
	{0xBD89A7AFC9820560ULL, "913348268.ogg"},
	{0xBD8A22E6D3227A1EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothd_1.mao"},
	{0xBD8A2E7A54A00F94ULL, "vxlt_gallowstemplar_f_654_0_0.vlm"},
	{0xBD8AD6DB85C356ADULL, "gen00fl_merrill6122365.cl"},
	{0xBD8B27C3EBF61041ULL, "vxlt_0_lowtown_a2_e3d_0_0.vlm"},
	{0xBD8B389B3A234B59ULL, "llac_w_encounter3_-196610_0.dds"},
	{0xBD8B55A65484A042ULL, "vxw_encounter2_rt_0_0.vlm"},
	{0xBD8B5E488AF7B4E0ULL, "vxht_mansion2b_38b_0_0.vlm"},
	{0xBD8B9413E17D2BDEULL, "vxlt_gallowscourtyard_3v7_0_0.vlm"},
	{0xBD8BAEBAA9871C1DULL, "vxht_mansion2_1sp_0_0.vlm"},
	{0xBD8BF5D9C1559430ULL, "lgt110ip_at_gang_hideout_lt.utp"},
	{0xBD8C139428564397ULL, "vxdr_primevaltomb_4xk_0_0.vlm"},
	{0xBD8C30E248A0CAADULL, "519474236.ogg"},
	{0xBD8D1C307C6C9B54ULL, "vxw_woundedcoast_1nv_0_0.vlm"},
	{0xBD8D638DFC937B00ULL, "and111_refugee_barks6111537.cl"},
	{0xBD8D7CC1526D0D57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_r.gad"},
	{0xBD8DB9526EE87E8CULL, "vxlt_gallowscourtyard_7bu_0_0.vlm"},
	{0xBD8DD31F730AC0D6ULL, "ca_caverns1_26035_lc.mmh"},
	{0xBD8E1B99E52E6EA8ULL, "vxw_bonepitdestroyed_a3_4x_0_0.vlm"},
	{0xBD8E2A862B54DB3BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_rdy_crnf.ani"},
	{0xBD8E8B0D8BE5B363ULL, "llac_ht_chantry_n_2534_25.dds"},
	{0xBD8EE0230663B45BULL, "403551456.ogg"},
	{0xBD8EEF6592F6CAA7ULL, "vxlt_0_lowtown_n_e2f_0_0.vlm"},
	{0xBD8F047DE387CD81ULL, "gen00fl_aveline6116727.cl"},
	{0xBD8F35272FB5D95CULL, "325891529.ogg"},
	{0xBD8F376EB86EFD87ULL, "706002880.ogg"},
	{0xBD8F44F0EFC5E08DULL, "327804836.ogg"},
	{0xBD8F47A40DA3DE66ULL, "6125372_m.fxe"},
	{0xBD8F4BBAD1FBB081ULL, "gen00fl_fenris6126682.cl"},
	{0xBD8F9294A19ED7ADULL, "vxht_mansion2_2bl_0_0.vlm"},
	{0xBD8FBCA80B28D4BBULL, "gen00fl_merrill6150991.cl"},
	{0xBD8FC0411F47B1CAULL, "6087601_m.fxe"},
	{0xBD8FE245F0613AB0ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_3y.dds"},
	{0xBD8FEF9F6516A2B3ULL, "vxca_cavern1_haunted_gik_0_0.vlm"},
	{0xBD90239FC3F98CCAULL, "6140980_m.fxe"},
	{0xBD904FB9DCED6734ULL, "hlac_ht_keep_n_3732_3z.dds"},
	{0xBD90BFA700BFB084ULL, "bedsingle_sa.dds"},
	{0xBD90E9602B6E6D31ULL, "vxlt_0_lowtown_docks_a3_aql_0_0.vlm"},
	{0xBD9132AE3F451F22ULL, "vxht_0_hightown_a3_bao_0_0.vlm"},
	{0xBD9171F7AF96302AULL, "ch2.psh.fxo"},
	{0xBD9189D18DEFF595ULL, "hair2.psh.fxo"},
	{0xBD91ED8655708528ULL, "zz_approval_debug6003834.cl"},
	{0xBD925976B4786591ULL, "gen00fl_isabela6089530.cl"},
	{0xBD927F28D15D0A24ULL, "hlac_lt_backalley_n_65537_3x.dds"},
	{0xBD9320C50868A95DULL, "vxlt_backalley_g_1wy_0_0.vlm"},
	{0xBD93217AB0B8AAE3ULL, "6128019_m.fxe"},
	{0xBD9342DCA7C07A69ULL, "vxw_encounter2_jk_0_0.vlm"},
	{0xBD937F7CAAF0C1BDULL, "vxlt_0_lowtown_docks_n_b6o_0_1.vlm"},
	{0xBD93A53D856A019CULL, "vxca_caverns1_i1s_0_0.vlm"},
	{0xBD9449B403F548CBULL, "442960390.ogg"},
	{0xBD94A6EDCFDB9B0CULL, "hlac_w_bonepit_0_6z.dds"},
	{0xBD950B0C9183BC56ULL, "vxlt_0_lowtown_a3_n_c_fhu_0_0.vlm"},
	{0xBD951B48270BA06FULL, "gen00fl_isabela_init6175713.cl"},
	{0xBD9548397B50BB17ULL, "lowtown_male_cityguard_mob6150519.cl"},
	{0xBD95CF915748903CULL, "vxw_bonepit_zb_0_0.vlm"},
	{0xBD963ADBB123A3CFULL, "vxlt_warehouse_n_r3_0_0.vlm"},
	{0xBD974BF13DB42919ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandcrackl_1.mao"},
	{0xBD976736644D4AF3ULL, "one280_emrys6104252.cl"},
	{0xBD977CFB192F5233ULL, "hlac_ht_0_hightown_65538_0y.dds"},
	{0xBD97D56D54173B7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_dismiss.ani"},
	{0xBD97F53529A5E706ULL, "vxlt_0_lowtown_a2_n_h26_0_0.vlm"},
	{0xBD982FB4414AE001ULL, "528965718.ogg"},
	{0xBD98C25C3E998356ULL, "follower_banter6136643.cl"},
	{0xBD98D41398AC4F75ULL, "6109730_f.fxe"},
	{0xBD9905CAEFAA1516ULL, "975736384.ogg"},
	{0xBD9A0650F14A756FULL, "hlac_lt_0_lowtown_docks_n_-65539_4y.dds"},
	{0xBD9A273CF6591F44ULL, "670002597.ogg"},
	{0xBD9A77908B4D8492ULL, "70257991.ogg"},
	{0xBD9A8801546AD752ULL, "vxht_chantry_n_35a_0_1.vlm"},
	{0xBD9AA76A6E1FE2FCULL, "hlac_w_sundermount_3_0x.dds"},
	{0xBD9AD090D8BA1830ULL, "hlac_lt_0_lowtown_a2_n_c_131071_1x.dds"},
	{0xBD9B0443E5AD61BCULL, "6082064_m.fxe"},
	{0xBD9B17B0603D6FC4ULL, "235446826.ogg"},
	{0xBD9B74644C3C7EBFULL, "6086799_m.fxe"},
	{0xBD9B78908919270CULL, "hlac_lt_undercityencounter_-1_41y.dds"},
	{0xBD9BE7F754C0EE29ULL, "mer210_smith.bnk"},
	{0xBD9C2B2F60DE970CULL, "vxlt_gallowscourtyard_7gj_0_0.vlm"},
	{0xBD9C58CC351A5ED9ULL, "vxw_encounter3_9i_0_0.vlm"},
	{0xBD9CF1ABD3E728C4ULL, "682411648.ogg"},
	{0xBD9D1386305CC541ULL, "vxht_keep_6ft_0_0.vlm"},
	{0xBD9D177D8E41F16CULL, "hlac_lt_0_lowtown_docks_a3_-65539_2z.dds"},
	{0xBD9D23A57AB0C078ULL, "vxca_caverns1_mines_in7_0_0.vlm"},
	{0xBD9D611D3E2069F0ULL, "dre000_dougal6142199.cl"},
	{0xBD9E00696ABBC0A3ULL, "6043024_m.fxe"},
	{0xBD9E04719394F02DULL, "vxht_mansion2_2ok_0_0.vlm"},
	{0xBD9E0DEF8A882112ULL, "hlac_lt_0_lowtown_a2_0_5x.dds"},
	{0xBD9ECA689181C972ULL, "fen310_party_comment6116862.cl"},
	{0xBD9EE26D541D6FBEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_dismiss.gad"},
	{0xBD9F2EE0789F6FAFULL, "art\\levels\\races\\proxy\\props\\door\\animation\\dooropening.gad"},
	{0xBD9FC9F754C4A469ULL, "mer210_smith.cnv"},
	{0xBDA00938C803F644ULL, "art\\characters\\playercharacter\\humanmale\\cth_coma.msh"},
	{0xBDA01B38C80414ADULL, "art\\characters\\playercharacter\\humanmale\\cth_coma.mao"},
	{0xBDA01F38C8041BE6ULL, "art\\characters\\playercharacter\\humanmale\\cth_coma.mmh"},
	{0xBDA0C91D2F5C4CECULL, "gen_im_ammo_quiver_01.uti"},
	{0xBDA0EAB54F55A1C1ULL, "gen_im_acc_blt_act1_05.uti"},
	{0xBDA106E9FE667F5CULL, "51795537.ogg"},
	{0xBDA1A0E22CED36DAULL, "6088064_m.fxe"},
	{0xBDA23BD789A851AEULL, "pro000_opening_pc6001193.cl"},
	{0xBDA251862B65C6C5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_rdy_crnf.gad"},
	{0xBDA254F36D3F0441ULL, "vxht_0_hightown_n_bub_0_0.vlm"},
	{0xBDA2A9C1527FB5DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_r.ani"},
	{0xBDA33C3BDF474049ULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_aoe_p.mmh"},
	{0xBDA33E0A32072907ULL, "lgt000_invis_plc.utp"},
	{0xBDA376ED3044FBF6ULL, "vxca_cave_1on_0_1.vlm"},
	{0xBDA39FA367B0168FULL, "zz_approval_debug6003071.cl"},
	{0xBDA45B9DA839E2CEULL, "381836175.ogg"},
	{0xBDA4ACB8A319BC4FULL, "vxlt_warehouse_n_y2_0_0.vlm"},
	{0xBDA4ED3E4726D7A1ULL, "hlac_ht_chantry_2534_45y.dds"},
	{0xBDA51F6771750C63ULL, "vxca_caverns2prc_9f7_0_0.vlm"},
	{0xBDA5BD6D54234FDDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_dismiss.evt"},
	{0xBDA5C8277268EF6CULL, "704042841.ogg"},
	{0xBDA5E42262D10603ULL, "253215009.ogg"},
	{0xBDA65D936ECB0336ULL, "llac_lt_0_lowtown_docks_a3_-196611_0.dds"},
	{0xBDA671BBBBDFB39EULL, "hlac_lt_gamlenshouse_n_-1_8x.dds"},
	{0xBDA756B4C55F5B44ULL, "6080552_m.fxe"},
	{0xBDA77CE6BA371B99ULL, "95911332.ogg"},
	{0xBDA77D829CF084C3ULL, "834153351.ogg"},
	{0xBDA7A5D3EE506B4FULL, "w_sundermount_137_0.mao"},
	{0xBDA7AE8BFFB985F1ULL, "76491639.ogg"},
	{0xBDA7B1D3EE507FECULL, "w_sundermount_137_0.mmh"},
	{0xBDA7B7D3EE5089DEULL, "w_sundermount_137_0.msh"},
	{0xBDA800AF2C0CD49AULL, "6149086_m.fxe"},
	{0xBDA830F3CF59E419ULL, "6144518_m.fxe"},
	{0xBDA8D9A807E30401ULL, "377734648.ogg"},
	{0xBDA920C464E0A589ULL, "6129551_m.fxe"},
	{0xBDA9A912D5F901DDULL, "6111383_m.fxe"},
	{0xBDA9C8AC20F04914ULL, "vxlt_0_lowtown_a2_gw7_0_0.vlm"},
	{0xBDA9F109916D4BA2ULL, "133884481.ogg"},
	{0xBDAA1638C80C60B1ULL, "art\\characters\\playercharacter\\humanmale\\cth_coma.phy"},
	{0xBDAACE34E7E42CFCULL, "202371395.ogg"},
	{0xBDAAD4BFF4951909ULL, "vxlt_0_lowtown_docks_b4k_0_0.vlm"},
	{0xBDAAFA3EB86738BBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_basic.mmh"},
	{0xBDAB063EB8674D58ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_basic.mao"},
	{0xBDAB183EB8676BA5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_basic.msh"},
	{0xBDAB659C9A5BD8A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_grand_reveal.ani"},
	{0xBDAB85632252DDC1ULL, "vxht_keep_6ck_0_0.vlm"},
	{0xBDAB8ADE73C4EADAULL, "mag100cr_templar4.utc"},
	{0xBDAB95826CA6D8F8ULL, "vxca_caverns1_mines_ik4_0_0.vlm"},
	{0xBDABC3830D8925B8ULL, "vxca_cave_19r_0_1.vlm"},
	{0xBDAC3F829CE942E9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\coastwoodb_d.dds"},
	{0xBDAC4FADB5B21F57ULL, "mag311_huon6042691.cl"},
	{0xBDACE0DC51F0A9EBULL, "6080568_m.fxe"},
	{0xBDACF035966AF989ULL, "6137080_m.fxe"},
	{0xBDAD14F813D5CED4ULL, "gen00fl_merrill6113645.cl"},
	{0xBDAE1288AC64A27DULL, "lt_0_lowtown_a2_27_0_ll.mmh"},
	{0xBDAE4F9964578EECULL, "145226491.ogg"},
	{0xBDAE9C0EF84B86D7ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf01.phy"},
	{0xBDAF5B596FCBD2A0ULL, "vxlt_0_lowtown_h7t_0_0.vlm"},
	{0xBDAF9F4D0A5D6F80ULL, "ch1armtnt_ds2y.psh.fxo"},
	{0xBDAFF91ABFAE81CCULL, "hlac_lt_0_lowtown_n_65536_9x.dds"},
	{0xBDB0047EEC54DA5DULL, "vxlt_backalley_g_1yw_0_0.vlm"},
	{0xBDB026968BACAB69ULL, "vxw_encounter2_pi_0_0.vlm"},
	{0xBDB063B817422CF2ULL, "gencr_qunari_thrall_mage.utc"},
	{0xBDB1200F2BF6F4FEULL, "220886185.ogg"},
	{0xBDB12386D2A18315ULL, "495147997.ogg"},
	{0xBDB15A39E26C5205ULL, "6098233_m.fxe"},
	{0xBDB1786320F82AD9ULL, "409600785.ogg"},
	{0xBDB18BD40CD94EE1ULL, "6111090_f.fxe"},
	{0xBDB1B74B9EDFB91FULL, "6131601_f.fxe"},
	{0xBDB1D03ABB140E5FULL, "mcr370_negotiation6080274.cl"},
	{0xBDB22268E1F5107DULL, "601205748.ogg"},
	{0xBDB3E530496775F4ULL, "hlac_ht_keep_n_3733_0x.dds"},
	{0xBDB3E8C6C33DE8F8ULL, "6116286_m.fxe"},
	{0xBDB443275AD2E72FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2junk.mao"},
	{0xBDB44BE078B1FD75ULL, "art\\levels\\races\\proxy\\props\\door\\animation\\dooropening.ani"},
	{0xBDB44F275AD2FBCCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2junk.mmh"},
	{0xBDB455275AD305BEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn2junk.msh"},
	{0xBDB5346B0C727ED4ULL, "ser221_quentin6128858.cl"},
	{0xBDB54FB1F51A7A58ULL, "lt_0_lowtown_docks_n_12_0_lc.mmh"},
	{0xBDB57025ED5B434CULL, "6084725_m.fxe"},
	{0xBDB5ADF140146B4FULL, "follower_banter6167215.cl"},
	{0xBDB5C19B31721656ULL, "hm_noble_gen_5.mor"},
	{0xBDB64DAD6F4CC653ULL, "6100053_m.fxe"},
	{0xBDB656882D26110CULL, "vxlt_0_lowtown_docks_bam_0_0.vlm"},
	{0xBDB6A7CA9C47D161ULL, "vxlt_0_lowtown_docks_cwg_0_0.vlm"},
	{0xBDB6F1E09E77396CULL, "vxlt_warehouse_be_0_0.vlm"},
	{0xBDB75AC17CD306EEULL, "6110066_m.fxe"},
	{0xBDB7C26400A1842BULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.evt"},
	{0xBDB83A8912E66194ULL, "vxlt_0_lowtown_a2_gj4_0_0.vlm"},
	{0xBDB84D9C9A663A17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_grand_reveal.evt"},
	{0xBDB88406D76C44D5ULL, "hlac_lt_0_lowtown_a2_-3_0z.dds"},
	{0xBDB8B79BC9B08D1BULL, "vxw_sundermount_38q_0_0.vlm"},
	{0xBDB97FFFFE0EED8BULL, "llac_lt_0_lowtown_docks_a3_131071_5.dds"},
	{0xBDB9CB0253D59547ULL, "vxlt_gallowstemplar_359_0_0.vlm"},
	{0xBDB9CE0B6C5A0B20ULL, "hlac_dr_0_deeproads_19793_0y.dds"},
	{0xBDB9D71DC5AA3763ULL, "zz_dae_debug6075039.cl"},
	{0xBDBA3EC8B8E6CCA1ULL, "vxlt_hangedman_n_24w_0_0.vlm"},
	{0xBDBA8CDEC0F68C82ULL, "hlac_lt_backalley_g_65536_1x.dds"},
	{0xBDBAE67136954B92ULL, "vxht_mansion2_1wg_0_0.vlm"},
	{0xBDBAE82B7781212DULL, "vxht_mansion2_2ui_0_0.vlm"},
	{0xBDBAEB6AAEF40024ULL, "vxlt_hangedman_25f_0_0.vlm"},
	{0xBDBB45E078FBBA99ULL, "vxlt_warehouse_n_10l_0_0.vlm"},
	{0xBDBB6C602B13C137ULL, "37053425.ogg"},
	{0xBDBB916440A4A1AEULL, "6106810_m.fxe"},
	{0xBDBBD37D9FEC9EA3ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_1z.dds"},
	{0xBDBC1D9C3730B06AULL, "vxw_bonepit_1o4_0_0.vlm"},
	{0xBDBC21B7DB2330E8ULL, "107797779.ogg"},
	{0xBDBC2FD97A947DA3ULL, "w_woundedcoast_154_0.msh"},
	{0xBDBC3DD97A9495FDULL, "w_woundedcoast_154_0.mmh"},
	{0xBDBC41D97A949C0EULL, "w_woundedcoast_154_0.mao"},
	{0xBDBD19EA498533CEULL, "250799736.ogg"},
	{0xBDBD1FB32CFED743ULL, "6104004_m.fxe"},
	{0xBDBDABCEEF47511AULL, "654790425.ogg"},
	{0xBDBE388BA17149B4ULL, "714894254.ogg"},
	{0xBDBE3D70E6DF1CB2ULL, "vxlt_gallowstemplar_5ct_0_0.vlm"},
	{0xBDBE3FF1B454C9F8ULL, "vxlt_0_lowtown_a2_n_e18_0_0.vlm"},
	{0xBDBE7118608ED783ULL, "vxw_woundedcoast_nx_0_0.vlm"},
	{0xBDBEB31B5BAD7B06ULL, "vxht_mansion2_24i_0_0.vlm"},
	{0xBDBECCB2DCA043E7ULL, "hlac_ht_0_hightown_n_-196608_1y.dds"},
	{0xBDBEDB6400A7CDECULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.gad"},
	{0xBDBEE5385D1685B8ULL, "vxca_caverns1_mines_dpp_0_0.vlm"},
	{0xBDBF3AD8486D8EDBULL, "6051877_f.fxe"},
	{0xBDBF4E9C9A6C5AA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_grand_reveal.gad"},
	{0xBDBF67E80D9F46E2ULL, "786013847.ogg"},
	{0xBDBFC5FFD9FCC2DAULL, "vxdr_0_deeproads_euz_0_0.vlm"},
	{0xBDBFF948D333067EULL, "dae000_bodhan6146785.cl"},
	{0xBDBFFAB8595FE9CEULL, "27178261.ogg"},
	{0xBDC0C9B1AA63E831ULL, "w_woundedcoast_37_0.mmh"},
	{0xBDC0D5B1AA63FCD2ULL, "w_woundedcoast_37_0.mao"},
	{0xBDC0DB311FF719EAULL, "llac_w_blightlands_131075_0.dds"},
	{0xBDC0E3B1AA6414EFULL, "w_woundedcoast_37_0.msh"},
	{0xBDC117E548921D4CULL, "vxlt_ruinedhouse_n_ac_0_0.vlm"},
	{0xBDC127F95B2FBE3DULL, "zz_followers_banter6157753.cl"},
	{0xBDC1353E8A794424ULL, "ef_progenitor_3.mor"},
	{0xBDC1B196CCE7940CULL, "llac_lt_undercityserialkiller_-1_11.dds"},
	{0xBDC1B9D522B9BF31ULL, "hlac_ht_0_hightown_a2_-131073_1y.dds"},
	{0xBDC2032155613D63ULL, "vxca_caverns2prc_9h9_0_0.vlm"},
	{0xBDC230990BCEFC72ULL, "6079562_f.fxe"},
	{0xBDC2B7E8367E1FCDULL, "zz_and_debug6134166.cl"},
	{0xBDC2B909B8978275ULL, "436586959.ogg"},
	{0xBDC2BB22820C9B37ULL, "vxht_bartrand_n_6jv_0_0.vlm"},
	{0xBDC360FDCF30A2A1ULL, "vxlt_0_lowtown_a2_n_c_hte_0_0.vlm"},
	{0xBDC422AE58CB45A6ULL, "6153346_m.fxe"},
	{0xBDC47EF2382B859DULL, "isa300pt_find_isabela.ncs"},
	{0xBDC49D7C3B340E19ULL, "6161252_m.fxe"},
	{0xBDC4D1DF04D96B04ULL, "151621336.ogg"},
	{0xBDC53939D29D7472ULL, "6101386_f.fxe"},
	{0xBDC562405A6205F5ULL, "544558572.ogg"},
	{0xBDC5A084A9EB9415ULL, "806956371.ogg"},
	{0xBDC5D6AE9C3C69BEULL, "103705415.ogg"},
	{0xBDC5DA6400ADEAEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.ani"},
	{0xBDC60F6A7F3644D9ULL, "vxlt_0_lowtown_a2_eqj_0_0.vlm"},
	{0xBDC63A9B19988589ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_hands_clasped.anb"},
	{0xBDC64BBE9A2933D7ULL, "6041530_m.fxe"},
	{0xBDC665DEDC03540CULL, "vxht_0_hightown_n_cb7_0_0.vlm"},
	{0xBDC6673E6F34268AULL, "6170424_m.fxe"},
	{0xBDC6BBB3AB96FA7EULL, "vxlt_shop_g3_0_0.vlm"},
	{0xBDC6E7F9B168A788ULL, "hlac_lt_0_lowtown_a2_n_c_0_6x.dds"},
	{0xBDC725CF52334934ULL, "hlac_w_encounter3_-196604_0x.dds"},
	{0xBDC75E69F7E865EDULL, "vxht_0_hightown_alm_0_0.vlm"},
	{0xBDC89153924740C1ULL, "lt_0_lowtown_docks_a3_21_0_ll.mmh"},
	{0xBDC8999115910927ULL, "vxht_keep_n_7a5_0_0.vlm"},
	{0xBDC8ED643934610BULL, "qun111_ketojan6059223.cl"},
	{0xBDC990F87CB43DCFULL, "6030436_f.fxe"},
	{0xBDC9A7F2C1B3A26CULL, "vxw_encounter3_bq_0_0.vlm"},
	{0xBDC9D994CF7009E4ULL, "6143473_m.fxe"},
	{0xBDCA537218012E42ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_1.gad"},
	{0xBDCB989E96E44A05ULL, "mag100st_grace_outside.stg"},
	{0xBDCBBF3C2D09C1A6ULL, "qcr000_arishok6060006.cl"},
	{0xBDCC2576ED4F367DULL, "vxlt_0_lowtown_a2_n_gjr_0_0.vlm"},
	{0xBDCC393BDF6A407EULL, "art\\vfx\\combat\\vfx_itm_tar_bomb_aoe_p.anb"},
	{0xBDCCAF4DD8D53532ULL, "vxlt_gallowstemplar_5lu_0_0.vlm"},
	{0xBDCCC6B2AD4EE51BULL, "vxlt_undercityencounter_102_0_0.vlm"},
	{0xBDCD05A302022C39ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_c.impa.ani"},
	{0xBDCD24F84DA39386ULL, "vxht_mansion2_21j_0_0.vlm"},
	{0xBDCD82463FB1335BULL, "ht_0_hightown_a3_52_0_lc.mmh"},
	{0xBDCEB27E19C10984ULL, "vxht_playershouse_n_1tj_0_0.vlm"},
	{0xBDD0036A128AC34EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v2.evt"},
	{0xBDD00B02727AF8A2ULL, "vxca_cavern1_haunted_im8_0_0.vlm"},
	{0xBDD00DDBFC3BFCD6ULL, "554105628.ogg"},
	{0xBDD02C7E62898D7EULL, "vxht_mansion2_29j_0_0.vlm"},
	{0xBDD05E711B7ABC2FULL, "vxht_brothel_n_3fc_0_0.vlm"},
	{0xBDD0A9A95EDC0F0EULL, "w_bonepitdestroyed_a3_11_0.msh"},
	{0xBDD0B7A95EDC275FULL, "w_bonepitdestroyed_a3_11_0.mao"},
	{0xBDD0C3A95EDC3BFCULL, "w_bonepitdestroyed_a3_11_0.mmh"},
	{0xBDD0C9EB399AB2C2ULL, "vxht_brothel_n_6cd_0_0.vlm"},
	{0xBDD123E27393C30DULL, "6053688_f.fxe"},
	{0xBDD1767218078839ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_1.evt"},
	{0xBDD227F2DDA9AF5CULL, "gen00fl_aveline6121986.cl"},
	{0xBDD292A59473035DULL, "1070986031.ogg"},
	{0xBDD2EE94F63BF5E1ULL, "392090028.ogg"},
	{0xBDD302C265B5818EULL, "hlac_lt_0_lowtown_a3_n_c_131070_2y.dds"},
	{0xBDD3135E14624BC7ULL, "hlac_lt_0_lowtown_a2_65534_5z.dds"},
	{0xBDD325CEF7581926ULL, "480822536.ogg"},
	{0xBDD3A65C7769648CULL, "6108032_m.fxe"},
	{0xBDD3C58DAF358456ULL, "6109402_m.fxe"},
	{0xBDD3E0A302080C7BULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_c.impa.gad"},
	{0xBDD3EF5E71F74B38ULL, "hlac_lt_0_lowtown_docks_n_131070_4y.dds"},
	{0xBDD3F218B2916A04ULL, "mage_forcefield.dds"},
	{0xBDD3F9913AAE69A4ULL, "520976988.ogg"},
	{0xBDD436F6ECC153A8ULL, "6114067_f.fxe"},
	{0xBDD475758784929EULL, "6084179_f.fxe"},
	{0xBDD4D3659CB72CD2ULL, "6190053_m.fxe"},
	{0xBDD4D870CC6D505BULL, "423524555.ogg"},
	{0xBDD4E7B6CC5C4156ULL, "10977709.ogg"},
	{0xBDD594AD38F6D4F7ULL, "6129342_m.fxe"},
	{0xBDD61CCF390547F0ULL, "hlac_lt_0_lowtown_a3_n_c_65534_6x.dds"},
	{0xBDD626CE62BBB898ULL, "vxht_0_hightown_a3_c10_0_0.vlm"},
	{0xBDD62CF74A2A9990ULL, "6018841_m.fxe"},
	{0xBDD6A4115B3A27B0ULL, "399324969.ogg"},
	{0xBDD6AAD4E7979C43ULL, "6176012_m.fxe"},
	{0xBDD6DB9CBB0FDE49ULL, "vxlt_hangedman_n_2o2_0_0.vlm"},
	{0xBDD6FA6A1290D2F9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v2.gad"},
	{0xBDD731B8317662FFULL, "art\\levels\\races\\design\\prp_prpbodybaga.msh"},
	{0xBDD737B831766CA1ULL, "art\\levels\\races\\design\\prp_prpbodybaga.mmh"},
	{0xBDD747305D97833EULL, "638846322.ogg"},
	{0xBDD76C2695C03724ULL, "w_bonepitmining_a2_158_0.mmh"},
	{0xBDD7702695C03E77ULL, "w_bonepitmining_a2_158_0.mao"},
	{0xBDD7822695C05CC6ULL, "w_bonepitmining_a2_158_0.msh"},
	{0xBDD7B2E226A5857AULL, "vxdr_primevaltomb_6j2_0_0.vlm"},
	{0xBDD7F3E1D0CF0FFCULL, "268206329.ogg"},
	{0xBDD80B65228EFB29ULL, "hightown_male_noble1_sta6170794.cl"},
	{0xBDD887732EE79988ULL, "vxlt_undercityencounter_xv_0_0.vlm"},
	{0xBDD89FA38C8309B1ULL, "6016491_m.fxe"},
	{0xBDD8CF527CD53F89ULL, "6162446_m.fxe"},
	{0xBDD8DB949EE0465AULL, "vxlt_warehouse_108_0_0.vlm"},
	{0xBDD9C9F10D9C35DDULL, "vxlt_undercityserialkiller_22y_0_0.vlm"},
	{0xBDD9FDA3020CA960ULL, "art\\vfx\\combat\\vfx_m_sta_atk_range_c.impa.evt"},
	{0xBDDA1FE1ADD5BDF8ULL, "684227788.ogg"},
	{0xBDDA51938A9C6200ULL, "ht_0_hightown_17_0.anb"},
	{0xBDDB212ACACB4DB2ULL, "vxlt_gallowstemplar_5iv_0_0.vlm"},
	{0xBDDB4884C0886A64ULL, "mag170_thrask6054068.cl"},
	{0xBDDB884C3B832085ULL, "520310716.ogg"},
	{0xBDDBB5776E178414ULL, "gen00fl_fenris6126472.cl"},
	{0xBDDC1C6735176641ULL, "vxht_0_hightown_n_baf_0_0.vlm"},
	{0xBDDC4485C8C98F29ULL, "z_dust4.dds"},
	{0xBDDCA36989B30871ULL, "vxlt_hangedman_n_10p_0_0.vlm"},
	{0xBDDD052829195E3DULL, "dae000_merchant_hm_mob6174229.cl"},
	{0xBDDD0B6A12955B47ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_trip_v2.ani"},
	{0xBDDD331DE0174E56ULL, "792232366.ogg"},
	{0xBDDE007BD56718A1ULL, "hlac_lt_0_lowtown_docks_a3_-2_5x.dds"},
	{0xBDDE23423B4B92FBULL, "vxht_mansion2_2kq_0_0.vlm"},
	{0xBDDE73A42706EEF5ULL, "vxlt_darktown_2wd_0_0.vlm"},
	{0xBDDE7CDF64711122ULL, "vxca_cavern1_haunted_ij9_0_0.vlm"},
	{0xBDDE7D3C89450FBCULL, "6107806_m.fxe"},
	{0xBDDE7E7218122028ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_1.ani"},
	{0xBDDEBB3B19CBB957ULL, "vxw_bonepitmining_a2_21n_0_0.vlm"},
	{0xBDDED89F09D7DDB5ULL, "prop1.vsh.fxo"},
	{0xBDDEF1E060D0204DULL, "825450898.ogg"},
	{0xBDDF9EDC65F8CC37ULL, "vxht_bartrand_n_6pp_0_0.vlm"},
	{0xBDDFC4A3C0906C93ULL, "vxlt_darktown_4vw_0_0.vlm"},
	{0xBDE00994D948D0FDULL, "pro000_meeran6050093.cl"},
	{0xBDE03A65F36D8B85ULL, "cd_fen210_dying_magister.wbd"},
	{0xBDE06AE62DE35F38ULL, "vxlt_shop_ek_0_0.vlm"},
	{0xBDE0E507E132804DULL, "dre000_dougal6167434.cl"},
	{0xBDE13A46B0DB0739ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian.msh"},
	{0xBDE14446B0DB1827ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian.mmh"},
	{0xBDE147C755FC8261ULL, "l_fen110ar_abandoned_house.bnk"},
	{0xBDE14846B0DB1F74ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian.mao"},
	{0xBDE1573BC4D253F4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\collapse_to_kneel.ani"},
	{0xBDE1F2162ACEEC6BULL, "vxlt_undercityencounter_18j_0_0.vlm"},
	{0xBDE24A89DE9229CEULL, "hlac_lt_0_lowtown_a2_n_c_65537_5y.dds"},
	{0xBDE2EE1EBAFD7F4FULL, "mag160_feynriel6033389.cl"},
	{0xBDE36DB46A870F43ULL, "43513647.ogg"},
	{0xBDE3F459185C3113ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_onyxstatuette_n.dds"},
	{0xBDE40393702F5218ULL, "llac_lt_0_lowtown_65536_8.dds"},
	{0xBDE419888492A486ULL, "pro000_meeran.cnv"},
	{0xBDE48799044EEA40ULL, "365900065.ogg"},
	{0xBDE4906B739313EDULL, "vxlt_undercitydungeon_259_0_0.vlm"},
	{0xBDE4B1EB7B42ACECULL, "704482481.ogg"},
	{0xBDE4D85AD2B76BDBULL, "dae200cr_viscount6087210.cl"},
	{0xBDE53981899B49B6ULL, "hlac_lt_0_lowtown_docks_a3_-131075_1x.dds"},
	{0xBDE5A99D456FDE67ULL, "llac_lt_undercityserialkiller_-1_49.dds"},
	{0xBDE5B1567592ADDDULL, "w_blightlands_120_0.phy"},
	{0xBDE5B1B524A6095DULL, "vxw_bonepitmining_a2_27f_0_0.vlm"},
	{0xBDE5B5FC22A0D39DULL, "423141145.ogg"},
	{0xBDE5D7B81EF8EF9DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masd_0t.dds"},
	{0xBDE624BF189B9DFAULL, "vxdr_primevaltomb_6q3_0_0.vlm"},
	{0xBDE638C217A85858ULL, "qcr000_aveline6054376.cl"},
	{0xBDE69517FA063591ULL, "art\\vfx\\creatures\\fxc_arch_smite_p.anb"},
	{0xBDE696E2B66D6C36ULL, "hlac_lt_undercityencounter_-1_6x.dds"},
	{0xBDE71B888494DE46ULL, "pro000_meeran.bnk"},
	{0xBDE7543BC4D6BA56ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\collapse_to_kneel.gad"},
	{0xBDE8293E5FDAB766ULL, "cc_str400d.dds"},
	{0xBDE8ADF1ACA2ED4BULL, "vxca_caverns1_ii8_0_0.vlm"},
	{0xBDE8D84C6D5CF7F3ULL, "gen00fl_dog6146955.cl"},
	{0xBDE9165A466F458DULL, "497262458.ogg"},
	{0xBDE95D284770F235ULL, "556346082.ogg"},
	{0xBDEA51B814160527ULL, "82230657.ogg"},
	{0xBDEA5BBDE3EE645CULL, "llac_w_encounter1_3_0.dds"},
	{0xBDEAC61CC43EEDABULL, "dae000_oldmangossip6174899.cl"},
	{0xBDEB3EC456E5C4D3ULL, "dae000_meredith6166289.cl"},
	{0xBDEB7D1465619658ULL, "1032351908.ogg"},
	{0xBDEBE2086B3F426AULL, "vxca_caverns1_ea1_0_0.vlm"},
	{0xBDEC1B4EA39C7E2AULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_hammer.mao"},
	{0xBDEC1F4EA39C8579ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_hammer.mmh"},
	{0xBDEC294EA39C9667ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_hammer.msh"},
	{0xBDECF2B2DF2C77CAULL, "hlac_ht_mansion2b_818_3y.dds"},
	{0xBDED422AFF66ED2FULL, "vxht_brothel_n_3hm_0_0.vlm"},
	{0xBDEDBC8DE27148AAULL, "lgt000ip_store_epic2_1.utp"},
	{0xBDEE530795EB2840ULL, "8225326.ogg"},
	{0xBDEE6F3BC4DD06B5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\collapse_to_kneel.evt"},
	{0xBDEE7FF976B11D0AULL, "564953165.ogg"},
	{0xBDEF3799823FB167ULL, "vxlt_warehouse_n_q2_0_0.vlm"},
	{0xBDEFC3A95EF6EC4FULL, "w_bonepitdestroyed_a3_11_0.rml"},
	{0xBDEFEC2392FC570CULL, "vxca_cavern1_haunted_i1r_0_0.vlm"},
	{0xBDEFEC6205EE0100ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack.phy"},
	{0xBDF049BB0DB4E784ULL, "6140928_m.fxe"},
	{0xBDF07302E411BB03ULL, "342924403.ogg"},
	{0xBDF129A510E76041ULL, "llac_ht_mansion2b_818_50.dds"},
	{0xBDF16763FE89CAFFULL, "6083158_m.fxe"},
	{0xBDF16AA1D5D73C99ULL, "mag170_thrask6020529.cl"},
	{0xBDF17189ACEBE794ULL, "hlac_ca_caverns1_mines_24424_5y.dds"},
	{0xBDF21B3EB8A373F0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_basic.phy"},
	{0xBDF2210970A926DFULL, "6108057_m.fxe"},
	{0xBDF25687152B83D6ULL, "513670279.ogg"},
	{0xBDF2AFD232A67F5FULL, "vxlt_0_lowtown_h9y_0_0.vlm"},
	{0xBDF2CF6E937F4324ULL, "984128967.ogg"},
	{0xBDF3A697D5EDCA0AULL, "fen210_hadriana6110526.cl"},
	{0xBDF3B617FA10F80AULL, "art\\vfx\\creatures\\fxc_arch_smite_p.mmh"},
	{0xBDF3F672C88A7818ULL, "vxlt_hangedman_n_4r_0_0.vlm"},
	{0xBDF3F9110F710A34ULL, "303509574.ogg"},
	{0xBDF43A3AE1D7BC3DULL, "45555946.ogg"},
	{0xBDF47C8DF6D51BAEULL, "6085477_f.fxe"},
	{0xBDF4BF829959CD2BULL, "dae000_mother6135361.cl"},
	{0xBDF4DEC39DFA195AULL, "hlac_lt_undercityserialkiller_-1_22z.dds"},
	{0xBDF540E1FAB13B7CULL, "333344952.ogg"},
	{0xBDF55CFCA2E74537ULL, "plt_genpt_app_anders.plo"},
	{0xBDF5FC0D7BF931CCULL, "gen00fl_dog6167083.cl"},
	{0xBDF60816C3FD8640ULL, "plt_dae021pt_brothel.nss"},
	{0xBDF66949576A3690ULL, "6042477_m.fxe"},
	{0xBDF67A504EF81C15ULL, "one351ip_to_sundermount.utp"},
	{0xBDF67D0EF88907CCULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf01.mmh"},
	{0xBDF6830EF88911BEULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf01.msh"},
	{0xBDF6A65675A10E49ULL, "w_blightlands_120_0.mao"},
	{0xBDF6B25675A122E2ULL, "w_blightlands_120_0.mmh"},
	{0xBDF6B45675A12598ULL, "w_blightlands_120_0.msh"},
	{0xBDF6FF26AD977E3EULL, "qun230cr_guild_contact6182710.cl"},
	{0xBDF7B45606469C04ULL, "dae021_sister_lorena_sta6150090.cl"},
	{0xBDF7CB483F5084D0ULL, "6096436_m.fxe"},
	{0xBDF8F66DDEA5A91CULL, "ser221_gamlen6134828.cl"},
	{0xBDFA2053E46B69D9ULL, "vxht_0_hightown_n_b9f_0_0.vlm"},
	{0xBDFA26BF47962C9AULL, "vxlt_backalley_n_1y9_0_0.vlm"},
	{0xBDFB21A5AC43CA13ULL, "72053305.ogg"},
	{0xBDFB40A9AA39CD91ULL, "408118487.ogg"},
	{0xBDFB55B7FD093F03ULL, "vxw_bonepit_1vr_0_0.vlm"},
	{0xBDFBA09C54EE6E17ULL, "6187913_m.fxe"},
	{0xBDFBF0E82E82BB8EULL, "one100_party_comments6133460.cl"},
	{0xBDFC0D14A1A93368ULL, "hlac_w_bonepitdestroyed_a3_0_6z.dds"},
	{0xBDFC2010095FE5F9ULL, "6071254_m.fxe"},
	{0xBDFC293013D81756ULL, "vxw_encounter2_h_0_0.vlm"},
	{0xBDFC2F84A992C31FULL, "vxca_caverns1_mines_eaq_0_0.vlm"},
	{0xBDFC42AF4C1E30C7ULL, "vxht_0_hightown_a2q_0_0.vlm"},
	{0xBDFC547655237C4BULL, "mrl210_merrill6166866.cl"},
	{0xBDFC61A83F957613ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_1y.dds"},
	{0xBDFCE3024469824DULL, "600894811.ogg"},
	{0xBDFCF4B71A148E8CULL, "vxw_bonepitdestroyed_a3_1bu_0_0.vlm"},
	{0xBDFDA33DC19E13BCULL, "vxlt_warehouse_n_105_0_0.vlm"},
	{0xBDFDD0163E1C6808ULL, "66667627.ogg"},
	{0xBDFE31C5F75F4D2AULL, "6086026_m.fxe"},
	{0xBDFE7EEA16DB2A39ULL, "hlac_lt_undercityencounter_-1_50z.dds"},
	{0xBDFED5D00EBB1D6BULL, "vxlt_undercityencounter_12l_0_0.vlm"},
	{0xBDFF4DBBEC7CE2E9ULL, "ave100_donnic6106124.cl"},
	{0xBDFFDBB954F49511ULL, "dae300_alistair_king6133537.cl"},
	{0xBDFFEC348A8AAEEDULL, "6108873_m.fxe"},
	{0xBE0113B34CD757D9ULL, "342866974.ogg"},
	{0xBE01B946F25DCB2AULL, "vxlt_gallowstemplar_n_4sr_0_0.vlm"},
	{0xBE024E2DBC604DC7ULL, "6022564_m.fxe"},
	{0xBE025BBB205569F0ULL, "art\\vfx\\combat\\vfx_f_mer_emove_disapp_p.mmh"},
	{0xBE02956F08923A5DULL, "vxw_bonepitmining_a2_29d_0_0.vlm"},
	{0xBE02AA862B1A14C5ULL, "vxdr_primevaltomb_6j9_0_0.vlm"},
	{0xBE02E6C99D70AF55ULL, "182393160.ogg"},
	{0xBE033A41C9E06CDEULL, "534734130.ogg"},
	{0xBE037F9DD68C8D99ULL, "686768165.ogg"},
	{0xBE038982B1B22E9FULL, "280266986.ogg"},
	{0xBE042CC71F772ED6ULL, "vxlt_hangedman_n_1xc_0_0.vlm"},
	{0xBE04A8447BFD7230ULL, "6150385_m.fxe"},
	{0xBE051A8F56B5913FULL, "vxw_encounter2_camp_qq_0_1.vlm"},
	{0xBE052514E176FCA8ULL, "mrl150_hunter6071275.cl"},
	{0xBE052BBA8E6D6B35ULL, "188730639.ogg"},
	{0xBE053C43022B028CULL, "6154060_m.fxe"},
	{0xBE05B486733E0E59ULL, "vxdr_primevalentrance_37g_0_0.vlm"},
	{0xBE05B9B97001BB5DULL, "gen00fl_anders6112984.cl"},
	{0xBE05C4A876E28C1FULL, "776665413.ogg"},
	{0xBE05EC25DABF27AEULL, "16052118.ogg"},
	{0xBE063094B2E85CCDULL, "one100dg_lia6013840.cl"},
	{0xBE06B02995220CB4ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_022a_0s.dds"},
	{0xBE06DA007E681BEAULL, "6103699_m.fxe"},
	{0xBE0700ABE3F5F3A9ULL, "dae210_gate_guard6085041.cl"},
	{0xBE0739E68BAFC7FCULL, "w_woundedcoast_66_0.mao"},
	{0xBE073DD3176900B5ULL, "6108824_m.fxe"},
	{0xBE0745E68BAFDB9FULL, "w_woundedcoast_66_0.mmh"},
	{0xBE074BE68BAFE641ULL, "w_woundedcoast_66_0.msh"},
	{0xBE0802D0BF5C9958ULL, "vxlt_warehouse_n_15n_0_0.vlm"},
	{0xBE082A0EBC6B7D5BULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\rangedattackv1.evt"},
	{0xBE084C40A0D27A52ULL, "6160869_m.fxe"},
	{0xBE08565ACF5435A5ULL, "87561741.ogg"},
	{0xBE0892DFB79D8803ULL, "vxca_caverns2_4x5_0_0.vlm"},
	{0xBE08EB7E499532CBULL, "6160767_m.fxe"},
	{0xBE092D2C76F2719EULL, "art\\vfx\\combat\\vfx_w_gn2_stun_c.impa.ani"},
	{0xBE093AD814639A0DULL, "zz_approval_debug6003110.cl"},
	{0xBE096AE6CFE37B1BULL, "958505900.ogg"},
	{0xBE096C64E5D7FA59ULL, "918953066.ogg"},
	{0xBE096CEC331FB927ULL, "gen00fl_isabela6078976.cl"},
	{0xBE096D54CE4E0643ULL, "vxlt_gallowstemplar_3s3_0_0.vlm"},
	{0xBE098C84E00E899EULL, "vxdr_0_deeproads_es8_0_0.vlm"},
	{0xBE09E84DBCA08167ULL, "564512387.ogg"},
	{0xBE0A1393AC84A62EULL, "gen00fl_varric6076519.cl"},
	{0xBE0A58480E756408ULL, "54125906.ogg"},
	{0xBE0AF6BBFFC14841ULL, "zz_followers_banter6157256.cl"},
	{0xBE0B652BA0A0E364ULL, "hlac_ht_mansion2_818_34y.dds"},
	{0xBE0BA0F25309EE9EULL, "gen00fl_anders6125994.cl"},
	{0xBE0BD1F03A3E76B0ULL, "vxlt_warehouse_n_z8_0_0.vlm"},
	{0xBE0BE5804543109DULL, "follower_banter6160780.cl"},
	{0xBE0CC2E1E864204BULL, "689646844.ogg"},
	{0xBE0D2D46B1002C5CULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvian.phy"},
	{0xBE0E19CF946A0947ULL, "98116064.ogg"},
	{0xBE0E48BB3904BF59ULL, "vxlt_0_lowtown_a2_ehm_0_0.vlm"},
	{0xBE0F230EBC718F9CULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\rangedattackv1.gad"},
	{0xBE0F4EA8D26912D9ULL, "80254682.ogg"},
	{0xBE102E1D6C68810FULL, "200703838.ogg"},
	{0xBE107EBB2061E20FULL, "art\\vfx\\combat\\vfx_f_mer_emove_disapp_p.anb"},
	{0xBE10DDEB2FBA2138ULL, "vxlt_gallowscourtyard_78o_0_0.vlm"},
	{0xBE1157910CD51E9FULL, "172808863.ogg"},
	{0xBE116F0093E3C010ULL, "vxw_encounter2_7i_0_0.vlm"},
	{0xBE121F675FB8EB68ULL, "llac_lt_0_lowtown_n_-65535_3.dds"},
	{0xBE1241481B53F496ULL, "57059486.ogg"},
	{0xBE126B3EE5A3AFCEULL, "87578814.ogg"},
	{0xBE12D58C9814A870ULL, "6218599_m.fxe"},
	{0xBE135E1D4DDCE31FULL, "lt_warehouse_0_0.msh"},
	{0xBE13641D4DDCEDC1ULL, "lt_warehouse_0_0.mmh"},
	{0xBE136EE1A3C454F9ULL, "sib131im_will.uti"},
	{0xBE13701D4DDD0262ULL, "lt_warehouse_0_0.mao"},
	{0xBE14263C16D499A0ULL, "ht_0_hightown_a2_26_0.anb"},
	{0xBE14458CF0910A2FULL, "603469854.ogg"},
	{0xBE14561247CAA162ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magb.msh"},
	{0xBE14581247CAA418ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magb.mmh"},
	{0xBE14641247CAB8BBULL, "art\\characters\\playercharacter\\humanfemale\\rob_magb.mao"},
	{0xBE150962060D4E95ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack.msh"},
	{0xBE150B62060D520BULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack.mmh"},
	{0xBE151762060D66A8ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack.mao"},
	{0xBE15B112601C1FCAULL, "6092513_m.fxe"},
	{0xBE16220EBC77AC9AULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\rangedattackv1.ani"},
	{0xBE16BB4E7E593DD9ULL, "ser120_ghyslain6105521.cl"},
	{0xBE16C95124D2D3E9ULL, "mag310cr_controlled_2.utc"},
	{0xBE17452C76FED86FULL, "art\\vfx\\combat\\vfx_w_gn2_stun_c.impa.evt"},
	{0xBE176A305A9FEFEBULL, "vxlt_0_lowtown_docks_a3_d2v_0_0.vlm"},
	{0xBE178FF7E9FA01DAULL, "392547529.ogg"},
	{0xBE17B65E7B398C0CULL, "vxdr_primevaltomb_8nv_0_0.vlm"},
	{0xBE17F3D5B55407FBULL, "art\\characters\\weapons\\shortbows\\w_sbw_001a.mmh"},
	{0xBE17FA00EF09F79BULL, "lt_0_lowtown_docks_179_0.mmh"},
	{0xBE17FFD5B5541B98ULL, "art\\characters\\weapons\\shortbows\\w_sbw_001a.mao"},
	{0xBE180600EF0A0C38ULL, "lt_0_lowtown_docks_179_0.mao"},
	{0xBE1811D5B5543AE5ULL, "art\\characters\\weapons\\shortbows\\w_sbw_001a.msh"},
	{0xBE181800EF0A2A85ULL, "lt_0_lowtown_docks_179_0.msh"},
	{0xBE18501F9FF0BAFCULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_loop.ani"},
	{0xBE18566031E313EAULL, "character_character_p14.psh.fxo.dx11.assembly"},
	{0xBE185D5B8FC15642ULL, "hlac_lt_0_lowtown_a2_n_c_65534_3z.dds"},
	{0xBE1880D17E15460BULL, "gen00fl_merrill6106237.cl"},
	{0xBE1919D4C3F2E9A6ULL, "vxdr_primevalentrance_26z_0_0.vlm"},
	{0xBE191D73AEA677F4ULL, "6019885_f.fxe"},
	{0xBE19D870FE00BF8CULL, "vxw_bonepitdestroyed_a3_1hk_0_0.vlm"},
	{0xBE1A0341757095C4ULL, "523230632.ogg"},
	{0xBE1A6604F6D3769FULL, "913859017.ogg"},
	{0xBE1A87F65BF10A28ULL, "277165886.ogg"},
	{0xBE1A8F055BB3C79DULL, "357219653.ogg"},
	{0xBE1AD93B43516702ULL, "6075837_m.fxe"},
	{0xBE1ADFB726B58F25ULL, "art\\vfx\\materials\\chunky\\vfx_c_falloff_invert.mao"},
	{0xBE1B30938AD3C128ULL, "ht_0_hightown_17_0.rml"},
	{0xBE1CFDC4C5EC0078ULL, "278807200.ogg"},
	{0xBE1D10A9DFBDED63ULL, "qcr000_arishok6039304.cl"},
	{0xBE1D146FC8324444ULL, "vxlt_warehouse_n_12k_0_0.vlm"},
	{0xBE1D462C77034518ULL, "art\\vfx\\combat\\vfx_w_gn2_stun_c.impa.gad"},
	{0xBE1D7ADDF5ACD13BULL, "vxlt_0_lowtown_n_gcg_0_0.vlm"},
	{0xBE1DF83362FD7189ULL, "6201241_f.fxe"},
	{0xBE1E08E2857031D9ULL, "hlac_ht_keep_n_4519_14y.dds"},
	{0xBE1E37F5620932E4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pota_d.dds"},
	{0xBE1E5F1247D304A3ULL, "art\\characters\\playercharacter\\humanfemale\\rob_magb.phy"},
	{0xBE1E970EEAD37A6BULL, "llac_ca_caverns1_24424_11.dds"},
	{0xBE1F4B46D952231EULL, "t3_rgb.tnt"},
	{0xBE1F5D1F9FF6EF3EULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_loop.gad"},
	{0xBE1FE0DD85D9D890ULL, "vxw_encounter2_2v_0_0.vlm"},
	{0xBE1FE2C7F42E0829ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantlrg.mao"},
	{0xBE1FEEC7F42E1CC2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantlrg.mmh"},
	{0xBE1FF0C7F42E2078ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennantlrg.msh"},
	{0xBE2024191D93B37CULL, "hlac_lt_0_lowtown_a2_n_c_0_1y.dds"},
	{0xBE209B904B9A135DULL, "lla_ht_0_hightown_a3_131072_0.dds"},
	{0xBE20D39BF40AB2B2ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_7x.dds"},
	{0xBE211465F3A4E22FULL, "cd_fen210_dying_magister.bnk"},
	{0xBE21505223E9940DULL, "6065322_f.fxe"},
	{0xBE21FD14A9E8FA14ULL, "6171133_m.fxe"},
	{0xBE225730FF364400ULL, "600166074.ogg"},
	{0xBE226A1BF01DCCECULL, "55675604.ogg"},
	{0xBE2271816590B053ULL, "229634850.ogg"},
	{0xBE2280BF59C9A785ULL, "vrc221_bartrand.bnk"},
	{0xBE2286F3C7694B53ULL, "vxw_woundedcoast_1gn_0_0.vlm"},
	{0xBE22CD49F18A5086ULL, "gen00fl_isabela6015682.cl"},
	{0xBE22DE971A390BF6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw21_basic.msh"},
	{0xBE22E37058B27D01ULL, "qun220cr_javaris6071502.cl"},
	{0xBE22E8971A391CF4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw21_basic.mmh"},
	{0xBE22EC971A392307ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw21_basic.mao"},
	{0xBE2322AE4DDBBE8CULL, "vxlt_undercitydungeon_1c4_0_1.vlm"},
	{0xBE2356D15C0B96C8ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.tn90l.evt"},
	{0xBE23A7DFA7792530ULL, "cs_pro000_cs_opening.wbd"},
	{0xBE2422EF5212384AULL, "6097613_m.fxe"},
	{0xBE249DEE85CA32CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_to_drink.gad"},
	{0xBE2531A38CE8056BULL, "6040537_m.fxe"},
	{0xBE2562BF59CBAABDULL, "vrc221_bartrand.cnv"},
	{0xBE258AD28BE820FAULL, "462451798.ogg"},
	{0xBE25BD84515E0CFAULL, "gen00fl_anders6125947.cl"},
	{0xBE25C0559B5F7A3BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv3.ani"},
	{0xBE25CCD21159A359ULL, "vxca_cave_1p1_0_0.vlm"},
	{0xBE25EE03F7ECE97EULL, "llac_ca_caverns2_5802_12.dds"},
	{0xBE25F7AE197C4898ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comd_0d.dds"},
	{0xBE260360F9B0901FULL, "gen00fl_fenris6113877.cl"},
	{0xBE2628B2E1CC6856ULL, "73434940.ogg"},
	{0xBE26381F9FFCCF5DULL, "art\\characters\\creatures\\mubari\\animation\\tailwag_loop.evt"},
	{0xBE26BBBC848966D1ULL, "155902704.ogg"},
	{0xBE26D1879F96863FULL, "vxht_0_hightown_n_bcv_0_0.vlm"},
	{0xBE27702BFD3F6D3CULL, "hlac_lt_0_lowtown_docks_n_-131074_2z.dds"},
	{0xBE27E647872F4A49ULL, "653090543.ogg"},
	{0xBE27FDF47A5EE713ULL, "vxlt_darktown_4iz_0_0.vlm"},
	{0xBE2855C756389C13ULL, "l_fen110ar_abandoned_house.wbd"},
	{0xBE289030FB1C22A9ULL, "ave331_jeven6138720.cl"},
	{0xBE290BFA17F1B99FULL, "w_encounter2_camp_28_0.rml"},
	{0xBE293C633952250FULL, "vxdr_0_deeproads_f0s_0_0.vlm"},
	{0xBE29A27C1120699FULL, "static_staticreliefmap_p0.gsh.fxo.dx11.assembly"},
	{0xBE2A57BE889A448CULL, "gen00fl_anders6125505.cl"},
	{0xBE2A78C78C4510BFULL, "cotbedroll_d.dds"},
	{0xBE2ABCDD5C4C23A5ULL, "146247892.ogg"},
	{0xBE2ACC40819BF8BFULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\malemage1_.gad"},
	{0xBE2B29740CB32F5FULL, "mrl000_marethari6142146.cl"},
	{0xBE2B2D600CC19F72ULL, "477503684.ogg"},
	{0xBE2B746B2112CA1FULL, "llac_lt_0_lowtown_n_65535_7.dds"},
	{0xBE2C07A988BDE9D2ULL, "vxw_woundedcoast_1gk_0_0.vlm"},
	{0xBE2C781DDD6233BBULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v2.evt"},
	{0xBE2D462EEFFB255BULL, "vxlt_shop_61_0_0.vlm"},
	{0xBE2D56727D2C2EC4ULL, "867323731.ogg"},
	{0xBE2D915F26D3E78BULL, "6151275_m.fxe"},
	{0xBE2DC200EF09D50AULL, "art\\vfx\\combat\\vfx_glb_restore_c.anb"},
	{0xBE2E36C04AE30551ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun_3.mao"},
	{0xBE2EA3AA65A3C653ULL, "6083679_m.fxe"},
	{0xBE2ED7553FBAEEA7ULL, "llac_lt_0_lowtown_a2_n_c_-3_7.dds"},
	{0xBE2F8D7D4E434778ULL, "vxlt_0_lowtown_docks_a3_n_cxk_0_0.vlm"},
	{0xBE30193D3FB04D99ULL, "705394542.ogg"},
	{0xBE302AFA17F80D37ULL, "w_encounter2_camp_28_0.phy"},
	{0xBE3046E68BD2E2DCULL, "w_woundedcoast_66_0.anb"},
	{0xBE30489B57DAE30BULL, "6164976_m.fxe"},
	{0xBE30BA928B16256DULL, "face_face_p11.psh.fxo"},
	{0xBE30D044F951AE36ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.dura.evt"},
	{0xBE31272BBB017B79ULL, "6153976_m.fxe"},
	{0xBE31AB2AF7955388ULL, "vxht_0_hightown_a2_avj_0_0.vlm"},
	{0xBE31D5D356DC029AULL, "hlac_lt_0_lowtown_a2_n_1_8y.dds"},
	{0xBE31F4AA28B189C6ULL, "gen00fl_anders6125868.cl"},
	{0xBE3209C31278C8B1ULL, "rdr111dg_assistant6016641.cl"},
	{0xBE321397A0D45413ULL, "130918416.ogg"},
	{0xBE326A93807C5C8DULL, "one200dg_solivitus6048191.cl"},
	{0xBE32711DDD6693FCULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v2.gad"},
	{0xBE328B87BC157531ULL, "lt_0_lowtown_docks_n_204_0.msh"},
	{0xBE329987BC158D4CULL, "lt_0_lowtown_docks_n_204_0.mao"},
	{0xBE329AF5BD3D1D0CULL, "6086763_m.fxe"},
	{0xBE32A587BC15A1EFULL, "lt_0_lowtown_docks_n_204_0.mmh"},
	{0xBE32C8559B6A1232ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv3.evt"},
	{0xBE32E495BAA5B6B1ULL, "6136685_m.fxe"},
	{0xBE32F81D6E4C41EBULL, "143461871.ogg"},
	{0xBE340116C432DB03ULL, "plt_dae021pt_brothel.plo"},
	{0xBE345DAD047FA78DULL, "vxht_0_hightown_a3_caq_0_0.vlm"},
	{0xBE347E4B568FC184ULL, "dae000_bodhan6129155.cl"},
	{0xBE3487BE05DEC0EFULL, "vxht_0_hightown_a3_c3p_0_1.vlm"},
	{0xBE34CCB31CC487B0ULL, "6106635_f.fxe"},
	{0xBE35233F940686CAULL, "785555306.ogg"},
	{0xBE35540ADD1D64C3ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_tmpb_eye.mao"},
	{0xBE3599526B1FFC92ULL, "6111173_m.fxe"},
	{0xBE35DBAEE84DA209ULL, "vxw_woundedcoast_ic_0_0.vlm"},
	{0xBE361A197A243B53ULL, "mag101_grace6034215.cl"},
	{0xBE3725744D7FA603ULL, "eye_tints.gda"},
	{0xBE37452479BCAE7EULL, "hlac_lt_gallowscourtyard_n_131071_2x.dds"},
	{0xBE377D00BA852283ULL, "6126651_m.fxe"},
	{0xBE37C744F957BDC1ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.dura.gad"},
	{0xBE381FA9695816B9ULL, "6161425_m.fxe"},
	{0xBE383E55FB4E0EB1ULL, "6089613_m.fxe"},
	{0xBE386E67AA99643BULL, "art\\vfx\\combat\\vfx_g_dual_att_left_c.mmh"},
	{0xBE389DB2EE5EF33CULL, "ht_mansion2b_mmi.gff"},
	{0xBE3930DD725B1881ULL, "hlac_lt_0_lowtown_docks_n_-65537_0y.dds"},
	{0xBE39441D4DFD86EAULL, "lt_warehouse_0_0.rml"},
	{0xBE39701DDD6CB0FAULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tw_v2.ani"},
	{0xBE3978EE85DC4FB0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_drunk_to_drink.ani"},
	{0xBE39947DF4DCE81DULL, "hlac_lt_0_lowtown_n_65533_0y.dds"},
	{0xBE39D8EBEAFF4DAFULL, "hairalpha_hair_p2.vsh.fxo.dx11"},
	{0xBE39E7559B7065C5ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\meleeattackv3.gad"},
	{0xBE3AAD9D77E3F5F7ULL, "dae000_mother6121001.cl"},
	{0xBE3AB008E2CA97BCULL, "hlac_lt_0_lowtown_docks_a3_n_-65540_1x.dds"},
	{0xBE3AC753C9F4703EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_grenade_toss.evt"},
	{0xBE3ACE06EE994350ULL, "hlac_lt_0_lowtown_a2_-131073_0y.dds"},
	{0xBE3AD18E1A52A184ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxa_sa.dds"},
	{0xBE3BF2345AFA6F6FULL, "6206159_m.fxe"},
	{0xBE3C11233B3FF8E0ULL, "vxca_caverns1_mines_i9f_0_0.vlm"},
	{0xBE3C5B317295C40BULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchwall_sa.dds"},
	{0xBE3C63207DD952F6ULL, "6038683_m.fxe"},
	{0xBE3CAB294D53B36AULL, "vrc321ip_proj_vase_1.utp"},
	{0xBE3CC49769C60990ULL, "vxw_encounter2_8t_0_0.vlm"},
	{0xBE3CF72D527E63C5ULL, "mag121_eau_banter6121092.cl"},
	{0xBE3D852956A49C04ULL, "373166891.ogg"},
	{0xBE3E800C935F0BC0ULL, "vxht_bartrand_n_737_0_0.vlm"},
	{0xBE3EAB473FEB931CULL, "hla_ca_cave_2194_0x.dds"},
	{0xBE3EC844F95DDE6FULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.dura.ani"},
	{0xBE3F651D4E022A3EULL, "lt_warehouse_0_0.phy"},
	{0xBE3FA94081AE1905ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\malemage1_.ani"},
	{0xBE3FAF3D7F385DAEULL, "vxlt_0_lowtown_a2_n_e7x_0_0.vlm"},
	{0xBE401D07E98B6C08ULL, "vxht_0_hightown_a2_aqk_0_0.vlm"},
	{0xBE4063631A9EE23AULL, "art\\characters\\weapons\\spears\\textures\\w_qui_qun_0s.dds"},
	{0xBE40A8F3903D5ACFULL, "6030601_f.fxe"},
	{0xBE40A9FA0F11555FULL, "vxca_cavern1_haunted_edn_0_0.vlm"},
	{0xBE40C65709930E98ULL, "805248267.ogg"},
	{0xBE40D64B738C1AA7ULL, "llac_lt_0_lowtown_n_-131072_6.dds"},
	{0xBE41230FB5C0B3E0ULL, "hlac_w_encounter2_camp_393214_0x.dds"},
	{0xBE4125BC312D9BC7ULL, "407119324.ogg"},
	{0xBE41315073371339ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_0x.dds"},
	{0xBE4177EC620AB5D3ULL, "732316575.ogg"},
	{0xBE41CE51711DDE16ULL, "gen00fl_varric6118743.cl"},
	{0xBE41DC904FDBD971ULL, "vxw_bonepitdestroyed_a3_1wz_0_1.vlm"},
	{0xBE41DE53C9FAB669ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_grenade_toss.gad"},
	{0xBE4204F08A57AD68ULL, "vxw_woundedcoast_171_0_0.vlm"},
	{0xBE42768EB45EA707ULL, "74676317.ogg"},
	{0xBE4389B61D429F4EULL, "1035731473.ogg"},
	{0xBE438F08B6D7B983ULL, "vxw_bonepit_1oy_0_0.vlm"},
	{0xBE43BB65F23A517EULL, "vxht_mansion2_2ab_0_0.vlm"},
	{0xBE43C79339C7C7E2ULL, "llac_ht_chantry_2534_34.dds"},
	{0xBE44146D3ABBD4B5ULL, "6059408_f.fxe"},
	{0xBE455E1B602CABF9ULL, "6110366_m.fxe"},
	{0xBE458B20074E2363ULL, "vxht_keep_n_3xv_0_0.vlm"},
	{0xBE458F67AAA426A0ULL, "art\\vfx\\combat\\vfx_g_dual_att_left_c.anb"},
	{0xBE45C592B755E257ULL, "hlac_lt_0_lowtown_docks_n_196609_0y.dds"},
	{0xBE46418C1302DDFCULL, "306299979.ogg"},
	{0xBE4691865B4E2F39ULL, "6166525_m.fxe"},
	{0xBE46971B64ADCBF7ULL, "hlac_lt_0_lowtown_docks_n_-2_5x.dds"},
	{0xBE46AA795C1719C0ULL, "zz_mageslate_trans.cut"},
	{0xBE471BB3C11E17A3ULL, "zz_dre_debug6204377.cl"},
	{0xBE471DDE9E03DEC2ULL, "vxw_encounter2_1y_0_0.vlm"},
	{0xBE4779C48D047CF6ULL, "vxca_caverns2prc_62p_0_1.vlm"},
	{0xBE47A607FAC72A59ULL, "llac_ht_keep_n_3736_2.dds"},
	{0xBE47C1D2436A365EULL, "488997629.ogg"},
	{0xBE47CA739F824B8EULL, "227381894.ogg"},
	{0xBE47DD6E2092DD69ULL, "6151707_m.fxe"},
	{0xBE483E245889C898ULL, "281053906.ogg"},
	{0xBE48AF53CA0084B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_grenade_toss.ani"},
	{0xBE48CA23C013F575ULL, "hla_ht_0_hightown_n_-131074_0y.dds"},
	{0xBE48D01D560E1AE2ULL, "w_encounter3_1_0.msh"},
	{0xBE48D21D560E1D98ULL, "w_encounter3_1_0.mmh"},
	{0xBE48D997385206A2ULL, "6164348_m.fxe"},
	{0xBE48DE1D560E323BULL, "w_encounter3_1_0.mao"},
	{0xBE49BCC7A2FEA48BULL, "zz_dre_debug6056949.cl"},
	{0xBE4A119BF32F245FULL, "mag360_thrask6043910.cl"},
	{0xBE4A48F0719F3EC1ULL, "6173862_m.fxe"},
	{0xBE4A6DDB0FF01EAFULL, "zz_fol_debug6098095.cl"},
	{0xBE4A895EDED7D748ULL, "one300_zevran6075774.cl"},
	{0xBE4AD1B7270AE94CULL, "vxw_woundedcoast_1jr_0_0.vlm"},
	{0xBE4B89D086005690ULL, "vxw_blightlands_on_0_0.vlm"},
	{0xBE4B983578270F4AULL, "13351551.ogg"},
	{0xBE4BD21D561053ABULL, "w_encounter3_1_0.rml"},
	{0xBE4C8C165A5196D7ULL, "llac_lt_gallowstemplar_n_2394_2.dds"},
	{0xBE4CBF6222D250C8ULL, "mcr310_argument6182476.cl"},
	{0xBE4CE0CE7071C63CULL, "mcr370_meredith6083099.cl"},
	{0xBE4D6DD37796BC00ULL, "art\\characters\\playercharacter\\humanfemale\\arm_meda.phy"},
	{0xBE4DC2CFFB20B943ULL, "6176203_m.fxe"},
	{0xBE4E209EA9C3789CULL, "vxlt_0_lowtown_eki_0_0.vlm"},
	{0xBE4E7ADBA41A7A8DULL, "follower_banter6173339.cl"},
	{0xBE4E8DC81A76665EULL, "vxw_bonepitdestroyed_a3_29u_0_0.vlm"},
	{0xBE4F25BCCF4A7006ULL, "vxht_mansion2_2bs_0_0.vlm"},
	{0xBE4F431CC896511EULL, "841408560.ogg"},
	{0xBE4F8E4E9A2A416FULL, "cod_hst_kirk_arch_1a_4.utp"},
	{0xBE4FD6B1C0847A29ULL, "6187814_m.fxe"},
	{0xBE4FE59959CFBD1BULL, "monster2.cut"},
	{0xBE5015A9E600CCD1ULL, "hlac_lt_0_lowtown_-131072_1z.dds"},
	{0xBE50854B92DF345BULL, "combat_dw_r2_a1.gda"},
	{0xBE50D24F831A29EDULL, "6144763_m.fxe"},
	{0xBE512268E73BECD9ULL, "vxca_cave_1a6_0_0.vlm"},
	{0xBE51C515BAAA4EE3ULL, "fen110_fenris6118082.cl"},
	{0xBE51C9A7811CDE8DULL, "6115941_m.fxe"},
	{0xBE51D260FFC0E8E6ULL, "dae000_mother6099244.cl"},
	{0xBE51D2EE1917E220ULL, "vxlt_backalley_n_1pm_0_1.vlm"},
	{0xBE522D42E43069FEULL, "vxht_mansion2_2jc_0_0.vlm"},
	{0xBE52350179D6ED61ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\aveline_swordshield_idle.gad"},
	{0xBE524540A1AD66B4ULL, "938906971.ogg"},
	{0xBE52BF68CC39D309ULL, "vxw_woundedcoast_ca_0_0.vlm"},
	{0xBE530DB4571F0399ULL, "vxht_keep_n_3y7_0_0.vlm"},
	{0xBE532403B38BA092ULL, "883164916.ogg"},
	{0xBE532566D860E2B8ULL, "6047377_m.fxe"},
	{0xBE53B2BEE1D35D7BULL, "484499283.ogg"},
	{0xBE54A14D3B0D954DULL, "vxht_mansion_n_4ba_0_0.vlm"},
	{0xBE54C75407E0B282ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_meda_0d.dds"},
	{0xBE54EC46A80AD899ULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.mmh"},
	{0xBE55053C170BF8C8ULL, "ht_0_hightown_a2_26_0.rml"},
	{0xBE553824E12E1871ULL, "combat_smcn_range.gda"},
	{0xBE55648191354272ULL, "qun221_petrice6076173.cl"},
	{0xBE557DD6425CB5CAULL, "lt_0_lowtown_docks_n_246_0.mao"},
	{0xBE5581D6425CBC19ULL, "lt_0_lowtown_docks_n_246_0.mmh"},
	{0xBE558BD6425CCD07ULL, "lt_0_lowtown_docks_n_246_0.msh"},
	{0xBE55B170E2739D80ULL, "356935258.ogg"},
	{0xBE55C2B69173C554ULL, "hlac_lt_0_lowtown_a2_n_c_1_1z.dds"},
	{0xBE55C500EF2B2CFDULL, "art\\vfx\\combat\\vfx_glb_restore_c.mmh"},
	{0xBE55CA0313AA0E78ULL, "6172842_m.fxe"},
	{0xBE55E317BA3E2BC2ULL, "vxw_blightlands_fw_0_0.vlm"},
	{0xBE56874ADA27681CULL, "225835423.ogg"},
	{0xBE568A4916B8DE24ULL, "hlac_lt_0_lowtown_docks_a3_131070_2y.dds"},
	{0xBE56F57715E7B103ULL, "vxht_0_hightown_a2_ca2_0_0.vlm"},
	{0xBE573B0B481C69DEULL, "mcr370_negotiation6080163.cl"},
	{0xBE5795CCC688ABA9ULL, "water_drop.mao"},
	{0xBE57AD3540A07359ULL, "vxlt_0_lowtown_eh9_0_0.vlm"},
	{0xBE57C376804E03C4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_paintings_small_da.dds"},
	{0xBE5823EE0FF87BB5ULL, "6144734_m.fxe"},
	{0xBE58360179DB5A0FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\aveline_swordshield_idle.ani"},
	{0xBE58D3276E68D07CULL, "297866657.ogg"},
	{0xBE5967B62025165DULL, "1017218160.ogg"},
	{0xBE59D66349935209ULL, "6174935_m.fxe"},
	{0xBE59DBC66451CF46ULL, "747898986.ogg"},
	{0xBE5A26475821C737ULL, "gen00fl_anders6119381.cl"},
	{0xBE5A6F636E150712ULL, "6207944_m.fxe"},
	{0xBE5A7695DE770F18ULL, "507167709.ogg"},
	{0xBE5AE224D1C59C5FULL, "gen_im_cns_cft_wrn_07.uti"},
	{0xBE5C4D8AFC4C6E54ULL, "6070599_m.fxe"},
	{0xBE5D7CF1E335D169ULL, "vxw_encounter2_4u_0_0.vlm"},
	{0xBE5D9799C1408886ULL, "vxht_mansion2_2ot_0_0.vlm"},
	{0xBE5E1553AFDF04D2ULL, "827770575.ogg"},
	{0xBE5E1793F8AC1E3BULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocklarge.mmh"},
	{0xBE5E2393F8AC32D8ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocklarge.mao"},
	{0xBE5E357DDCC3B9B9ULL, "internal_soundsettings.gda"},
	{0xBE5E3593F8AC5125ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocklarge.msh"},
	{0xBE5EC7F6EF313E21ULL, "hlac_lt_0_lowtown_a2_n_c_65536_1y.dds"},
	{0xBE5F02BA34CF81A0ULL, "art\\levels\\races\\proxy\\props\\treasury\\prp_zzchestmetal_0.pwk"},
	{0xBE5F97A75F8778ADULL, "vxw_bonepitdestroyed_a3_1s3_0_0.vlm"},
	{0xBE5FC9DDE1ACE098ULL, "mer320_hubert6053697.cl"},
	{0xBE603B3F18E9DA79ULL, "vxht_chantry_20v_0_0.vlm"},
	{0xBE6097ADE91045FCULL, "vxca_caverns2_7mz_0_0.vlm"},
	{0xBE61213408C2F1C8ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb07a.msh"},
	{0xBE613F3408C324F2ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb07a.mmh"},
	{0xBE6143A5D9A196B4ULL, "6188601_m.fxe"},
	{0xBE62101A909873B7ULL, "40391150.ogg"},
	{0xBE62507CD8424643ULL, "hlac_ca_caverns1_mines_24426_10x.dds"},
	{0xBE62986B922F7071ULL, "hlac_lt_0_lowtown_docks_n_0_4x.dds"},
	{0xBE62A00F26DEA55FULL, "hlac_lt_0_lowtown_n_-1_6y.dds"},
	{0xBE62EC6C691869F6ULL, "zz_followers_banter6158982.cl"},
	{0xBE630D3301FAA5DBULL, "vxlt_0_lowtown_a3_n_c_e7z_0_0.vlm"},
	{0xBE63289EF1041043ULL, "gen00fl_anders6111132.cl"},
	{0xBE642E1E4B90F096ULL, "llac_lt_undercitydungeon_-1_18.dds"},
	{0xBE659E4434F14E4BULL, "hlac_lt_0_lowtown_a2_131070_4z.dds"},
	{0xBE65B9D3A27F492BULL, "vxht_0_hightown_a2_5tc_0_0.vlm"},
	{0xBE65F5DC3E94B66AULL, "50248049.ogg"},
	{0xBE67731255150417ULL, "vxht_chantry_n_20o_0_0.vlm"},
	{0xBE679C9AFF69B164ULL, "plt_lgt000pt_ultra_3_06.plo"},
	{0xBE67B5710AF71A4CULL, "vxw_woundedcoast_1pp_0_0.vlm"},
	{0xBE67D55857DCB609ULL, "dae000_paivel6134537.cl"},
	{0xBE682CCB3F3FC06AULL, "6070580_m.fxe"},
	{0xBE683221A9472E4CULL, "6132824_m.fxe"},
	{0xBE686A92F487240BULL, "art\\vfx\\combat\\vfx_glb_restore_health_c.impa.ani"},
	{0xBE689655FB1FF57CULL, "337673900.ogg"},
	{0xBE68991E55C02CA2ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_009a_0s.dds"},
	{0xBE691F6EB1856C92ULL, "tbe100im_antivan_helm.uti"},
	{0xBE6944CDF33DE9EBULL, "234839493.ogg"},
	{0xBE69484FB09146EEULL, "ss_castillon.bnk"},
	{0xBE694B05F0727E9AULL, "355921399.ogg"},
	{0xBE696C44BF7B586BULL, "mcr371_hostile_merrill6082168.cl"},
	{0xBE69D43C860FACB1ULL, "6077601_m.fxe"},
	{0xBE69E8BDC3FA6CCAULL, "gen00fl_aveline6139750.cl"},
	{0xBE6A2582537F19AEULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lf.ani"},
	{0xBE6A5993910C7F0BULL, "gen00fl_aveline6006068.cl"},
	{0xBE6A7DDFA7B4E0F2ULL, "cs_pro000_cs_opening.bnk"},
	{0xBE6AAB2892930493ULL, "llac_ht_chantry_n_2534_18.dds"},
	{0xBE6AF398CC93C6AFULL, "678887728.ogg"},
	{0xBE6B82B2F55B8034ULL, "6056731_m.fxe"},
	{0xBE6BC6DD6803CFC4ULL, "vxht_0_hightown_a2_9zj_0_0.vlm"},
	{0xBE6C422AFF701E69ULL, "vxw_blightlands_co_0_0.vlm"},
	{0xBE6C7E20677806B1ULL, "138741388.ogg"},
	{0xBE6C991E32E45723ULL, "hlac_ca_cavern1_haunted_24424_1x.dds"},
	{0xBE6CBB12687A8BF7ULL, "llac_ht_0_hightown_a3_131073_0.dds"},
	{0xBE6D66C8422BFD46ULL, "vxlt_warehouse_w0_0_0.vlm"},
	{0xBE6D6DD896D19EE3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wi_sit_grnd.ani"},
	{0xBE6F4FBD62A30E2BULL, "490015961.ogg"},
	{0xBE6F883C84DBEC28ULL, "static_static_p3.psh.fxo"},
	{0xBE6FA322B0260409ULL, "vxw_woundedcoast_ug_0_0.vlm"},
	{0xBE6FA4CCC69D5757ULL, "water_drop.dds"},
	{0xBE6FB2C1453FD228ULL, "gen00fl_carver6104449.cl"},
	{0xBE6FBC818A8CE296ULL, "follower_banter6163469.cl"},
	{0xBE6FF336F5C7D137ULL, "hlac_lt_gallowstemplar_n_2390_1y.dds"},
	{0xBE7027F227D8D19DULL, "vxlt_0_lowtown_a2_fj4_0_0.vlm"},
	{0xBE703CD1F2A76D81ULL, "plt_cod_itm_cuirass_winds.plo"},
	{0xBE7045E68C08C56CULL, "w_woundedcoast_66_0.rml"},
	{0xBE704E1515DC1787ULL, "hlac_lt_gallowscourtyard_n_65535_0x.dds"},
	{0xBE70679F7912E712ULL, "lt_0_lowtown_docks_a3_28_0_ll.mmh"},
	{0xBE70A2670C57B751ULL, "144092678.ogg"},
	{0xBE70D2A4F4EBC2DBULL, "6104182_f.fxe"},
	{0xBE70D69F10B5D40FULL, "hlac_dr_slavers_3214_10x.dds"},
	{0xBE70FB39D81CED5BULL, "hlac_lt_0_lowtown_a2_n_-3_1x.dds"},
	{0xBE714F81E2AB67B3ULL, "dre000_bodahn6033414.cl"},
	{0xBE71A2A765C4EB19ULL, "6207815_m.fxe"},
	{0xBE71C763A935A99DULL, "llac_dr_primevaltomb_659_9.dds"},
	{0xBE72595FD470CC2DULL, "hlac_ht_0_hightown_a2_-65536_2z.dds"},
	{0xBE7264B79E6BE16BULL, "vxlt_undercityencounter_1jd_0_0.vlm"},
	{0xBE728AD377B60995ULL, "art\\characters\\playercharacter\\humanfemale\\arm_meda.msh"},
	{0xBE728CD377B60D0BULL, "art\\characters\\playercharacter\\humanfemale\\arm_meda.mmh"},
	{0xBE7298D377B621A8ULL, "art\\characters\\playercharacter\\humanfemale\\arm_meda.mao"},
	{0xBE72C6D19E2A5CC2ULL, "vxw_blightlands_lq_0_0.vlm"},
	{0xBE72DD1A54BA4DB3ULL, "609908421.ogg"},
	{0xBE72DFFBEC414DE1ULL, "267124223.ogg"},
	{0xBE72EC37420594D1ULL, "hf_nat_were_support.mor"},
	{0xBE736AE84FD74502ULL, "vxw_bonepitmining_a2_262_0_0.vlm"},
	{0xBE73A1B2E96B5C92ULL, "205568956.ogg"},
	{0xBE73FC9B2134C99AULL, "vxlt_0_lowtown_docks_a3_br5_0_0.vlm"},
	{0xBE740DA9EC99FF4AULL, "120817694.ogg"},
	{0xBE747D1839511960ULL, "vxlt_0_lowtown_docks_n_bg8_0_0.vlm"},
	{0xBE75212F174497F8ULL, "vxdr_primevaltomb_6qp_0_1.vlm"},
	{0xBE755F558D7664ECULL, "vxlt_0_lowtown_a2_ewi_0_0.vlm"},
	{0xBE75A1E7AA4E31E2ULL, "llac_lt_0_lowtown_a2_65535_3.dds"},
	{0xBE75BCBBF99439B8ULL, "6131156_f.fxe"},
	{0xBE765292F4933882ULL, "art\\vfx\\combat\\vfx_glb_restore_health_c.impa.evt"},
	{0xBE76605B47CEA13FULL, "vxw_encounter3_7a_0_0.vlm"},
	{0xBE769EA83135D8EAULL, "6075680_m.fxe"},
	{0xBE76FFFA1833C68FULL, "w_encounter2_camp_28_0.mao"},
	{0xBE770BFA1833DB2CULL, "w_encounter2_camp_28_0.mmh"},
	{0xBE7711FA1833E51EULL, "w_encounter2_camp_28_0.msh"},
	{0xBE77518C9ACCEAB3ULL, "llac_dr_dungeon_3216_5.dds"},
	{0xBE776F839FF01A9EULL, "vxlt_darktown_52w_0_0.vlm"},
	{0xBE77C8DAFAB12D82ULL, "l_dae211ar_tavern_nt.bnk"},
	{0xBE781D82538B49FFULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lf.evt"},
	{0xBE7846197805C531ULL, "6078701_m.fxe"},
	{0xBE789E9CB507527AULL, "hlac_lt_0_lowtown_a2_n_c_-1_6x.dds"},
	{0xBE79317A12402BC3ULL, "gen00fl_isabela6079041.cl"},
	{0xBE79C98F9ACE82CCULL, "gen00fl_isabela6099349.cl"},
	{0xBE79F07056BBF25AULL, "hlac_ht_keep_n_3732_8y.dds"},
	{0xBE79F48FE75198B4ULL, "6053431_m.fxe"},
	{0xBE7A502950C1057DULL, "723706512.ogg"},
	{0xBE7A85D896DC51DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wi_sit_grnd.evt"},
	{0xBE7AA225852FE6FAULL, "mag101_alain6032266.cl"},
	{0xBE7AB407F16636E9ULL, "vxw_blightlands_dn_0_0.vlm"},
	{0xBE7ABB1978D2617BULL, "fxe_sp_vol_blend.mao"},
	{0xBE7AD0B856B92A58ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srda_0s.dds"},
	{0xBE7ADEA912912140ULL, "6045036_m.fxe"},
	{0xBE7AF63024FC0B1AULL, "6106911_m.fxe"},
	{0xBE7B588EF34A1E75ULL, "1031208576.ogg"},
	{0xBE7BD8A5342215C6ULL, "vxlt_warehouse_x1_0_0.vlm"},
	{0xBE7BE4BA1B7FCC07ULL, "182011936.ogg"},
	{0xBE7BFD0F07A0FFD8ULL, "mcr371_scared_mage_f6087717.cl"},
	{0xBE7C19999D585874ULL, "vxlt_0_lowtown_docks_n_cs9_0_0.vlm"},
	{0xBE7C9038604AEF40ULL, "tbe100ip_shop_codex_7.utp"},
	{0xBE7D35FC59C7F796ULL, "gen00fl_sebastian6179600.cl"},
	{0xBE7D5192F4995595ULL, "art\\vfx\\combat\\vfx_glb_restore_health_c.impa.gad"},
	{0xBE7D67170920DF50ULL, "vxca_caverns1_gkj_0_1.vlm"},
	{0xBE7DE7971A85C6EFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw21_basic.phy"},
	{0xBE7DE946A82DD88EULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.anb"},
	{0xBE7DFEB860A1C936ULL, "6083751_cleric_youngmother.cut"},
	{0xBE7E14FFA21C1C89ULL, "vxw_woundedcoast_rf_0_0.vlm"},
	{0xBE7E29B5BDD2EE35ULL, "vxlt_0_lowtown_a2_byi_0_0.vlm"},
	{0xBE7EFE82539133E8ULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_lf.gad"},
	{0xBE7F3B962A8990D6ULL, "gen00fl_bethany6109231.cl"},
	{0xBE7F6FA1FB1462D6ULL, "vxca_cavern1_haunted_ebq_0_0.vlm"},
	{0xBE800D861DE88EA5ULL, "6093629_m.fxe"},
	{0xBE8010D38C66971AULL, "6065003_m.fxe"},
	{0xBE80194FDAE63138ULL, "vxw_bonepitdestroyed_a3_18o_0_0.vlm"},
	{0xBE804B558787D896ULL, "714503093.ogg"},
	{0xBE812F46629DB183ULL, "737797937.ogg"},
	{0xBE8138AE90207542ULL, "vxw_blightlands_qr_0_0.vlm"},
	{0xBE8184D896E26ECDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wi_sit_grnd.gad"},
	{0xBE81BC3F84849F5DULL, "qcr220_orsino6061805.cl"},
	{0xBE82F178BBD7C337ULL, "1030466506.ogg"},
	{0xBE83AB45F6860F28ULL, "vxw_bonepitdestroyed_a3_oi_0_0.vlm"},
	{0xBE83F18FC1C5E53CULL, "vxlt_0_lowtown_gay_0_0.vlm"},
	{0xBE84103A3DBD077DULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_9z.dds"},
	{0xBE849AABCDF55EBCULL, "sse_ss_corpse_1.gda"},
	{0xBE84DF986C60EA9FULL, "vxlt_0_lowtown_a2_hey_0_0.vlm"},
	{0xBE84EF5C8D9C2B76ULL, "zz_mag_debug6051372.cl"},
	{0xBE860B59AD1911FBULL, "vxht_0_hightown_n_cay_0_0.vlm"},
	{0xBE86390FDC213E16ULL, "585467636.ogg"},
	{0xBE863CF397EB4C7AULL, "6081427_m.fxe"},
	{0xBE865C48ED158DE4ULL, "mag121_viveka6082422.cl"},
	{0xBE870CE6C344C5F1ULL, "vxlt_hangedman_n_254_0_0.vlm"},
	{0xBE8735876137AD4EULL, "hlac_ca_caverns1_24421_0x.dds"},
	{0xBE87AE3CCA881A2CULL, "mcr371_orsino6083364.cl"},
	{0xBE87D0DA11E041CFULL, "llac_ca_caverns1_mines_24425_6.dds"},
	{0xBE87D6D154D02F1CULL, "6161648_m.fxe"},
	{0xBE87DD302FA5FA6CULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingc.msh"},
	{0xBE87E3302FA6045EULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingc.mmh"},
	{0xBE87EF302FA618F5ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingc.mao"},
	{0xBE881AF18805B150ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_deadfish_n.dds"},
	{0xBE881FA7BDB6D01BULL, "hlac_lt_warehouse_0_5y.dds"},
	{0xBE891FDAA1BDD2DFULL, "hlac_lt_0_lowtown_a3_n_c_65537_4x.dds"},
	{0xBE8976A83100D526ULL, "vxca_caverns2prc_8cl_0_0.vlm"},
	{0xBE89AA717F76585CULL, "6201786_m.fxe"},
	{0xBE8A680BA41966C4ULL, "872233308.ogg"},
	{0xBE8A8D70AA766C07ULL, "vxlt_0_lowtown_docks_a3_b99_0_0.vlm"},
	{0xBE8AED3E3569C22DULL, "vxw_bonepitdestroyed_a3_1b0_0_0.vlm"},
	{0xBE8B1207AA3FD1ECULL, "w_encounter3_155_0.msh"},
	{0xBE8B1807AA3FDBDEULL, "w_encounter3_155_0.mmh"},
	{0xBE8B2407AA3FF075ULL, "w_encounter3_155_0.mao"},
	{0xBE8B4AFB3C96E32CULL, "vxht_keep_3ya_0_0.vlm"},
	{0xBE8B747C41FA07ABULL, "479141519.ogg"},
	{0xBE8B9DDD0996ABD4ULL, "6173063_m.fxe"},
	{0xBE8BC416CDDD0A47ULL, "w_encounter3_41_0.rml"},
	{0xBE8C0A01C95CD8B9ULL, "986411440.ogg"},
	{0xBE8C4A043789979FULL, "vxca_caverns2prc_8a3_0_0.vlm"},
	{0xBE8C8C850EBF58ECULL, "vxht_0_hightown_n_b1w_0_0.vlm"},
	{0xBE8D236FE2A39E79ULL, "hlac_ca_caverns2_5802_34y.dds"},
	{0xBE8D2C3D755B35D0ULL, "6179941_f.fxe"},
	{0xBE8E1125333278D5ULL, "hlac_lt_0_lowtown_a2_n_c_-2_0x.dds"},
	{0xBE8E1258C8F89035ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.mmh"},
	{0xBE8E358BD9D71797ULL, "qun110_petrice6037086.cl"},
	{0xBE8E5B18AF0D3B77ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_5_darren.gad"},
	{0xBE8F1B813187ED3EULL, "311735781.ogg"},
	{0xBE8F6063B41790B7ULL, "6164697_m.fxe"},
	{0xBE8F63995340DC3DULL, "125610242.ogg"},
	{0xBE8F8433AD12F9CDULL, "6115457_m.fxe"},
	{0xBE8FAA8B82168DC2ULL, "vxw_blightlands_zs_0_0.vlm"},
	{0xBE900A4F421E3AC0ULL, "zz_dae_debug6029866.cl"},
	{0xBE90717B3AFA0DDAULL, "517210699.ogg"},
	{0xBE9096BB11B34702ULL, "vxlt_0_lowtown_a2_n_ffp_0_0.vlm"},
	{0xBE9101FD4491C75BULL, "mag321cr_dulci.utc"},
	{0xBE91E656E3133A61ULL, "514512119.ogg"},
	{0xBE9351755E57031FULL, "vxlt_0_lowtown_a2_hbx_0_0.vlm"},
	{0xBE93AF9D1BCED67CULL, "684478229.ogg"},
	{0xBE93EDC4BD544B38ULL, "hlac_dr_0_deeproads_5841_3y.dds"},
	{0xBE943E5C7FCF2D97ULL, "6214963_m.fxe"},
	{0xBE946A374482DAA7ULL, "6133352_m.fxe"},
	{0xBE948D491C156588ULL, "vxht_brothel_n_6g2_0_0.vlm"},
	{0xBE9505AD97EEBD3BULL, "519827414.ogg"},
	{0xBE9576322F15A90EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10a.msh"},
	{0xBE957F63772CD78BULL, "6092591_f.fxe"},
	{0xBE9590322F15D5FCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10a.mmh"},
	{0xBE95C055E9FACE55ULL, "art\\characters\\creatures\\dragonling\\mesh05.msh"},
	{0xBE95C255E9FAD1CBULL, "art\\characters\\creatures\\dragonling\\mesh05.mmh"},
	{0xBE95E09A4FBF902FULL, "vxlt_gallowstemplar_f_6sd_0_1.vlm"},
	{0xBE95F559EBD93AF0ULL, "6058863_m.fxe"},
	{0xBE961D14D311F82AULL, "gen_im_wep_rog_dua_act1_04.uti"},
	{0xBE9661D90AAB55EAULL, "vxw_bonepitmining_a2_1k6_0_0.vlm"},
	{0xBE96A0E1A8A00B9EULL, "mag311_nyssa6042366.cl"},
	{0xBE96BE778FB9A16AULL, "gen_im_trs_krk_shd_met_r1.uti"},
	{0xBE97237108B0E6E7ULL, "hlac_ht_0_hightown_65536_1x.dds"},
	{0xBE97548F3358ED92ULL, "vxht_keep_4nq_0_0.vlm"},
	{0xBE9797C1D55267E9ULL, "vxw_blightlands_rl_0_0.vlm"},
	{0xBE979BEB44136E48ULL, "hlac_w_sundermount_327681_0x.dds"},
	{0xBE97AC15F4584C83ULL, "hlac_ht_keep_n_3732_18y.dds"},
	{0xBE97EF01B33C1829ULL, "gen00fl_varric6081206.cl"},
	{0xBE987ED0C6F596DEULL, "6127086_m.fxe"},
	{0xBE98B9AB7F97CE1CULL, "follower_banter6162091.cl"},
	{0xBE9937F65FCF23E7ULL, "hlac_ht_0_hightown_a3_131071_1y.dds"},
	{0xBE995F8C2756F57EULL, "genip_exploding_barrel_tiny.utp"},
	{0xBE99BFAE5AA6B6E2ULL, "dae200cr_viscount6084555.cl"},
	{0xBE9ABB635E2FB3A0ULL, "vxlt_0_lowtown_a2_n_c_hjv_0_0.vlm"},
	{0xBE9AEE8B509CAD7BULL, "vxlt_gallowstemplar_438_0_0.vlm"},
	{0xBE9B3509C977E079ULL, "6004377_m.fxe"},
	{0xBE9B3AEA659B554AULL, "6097402_m.fxe"},
	{0xBE9B886484849038ULL, "hlac_lt_undercityencounter_-1_39z.dds"},
	{0xBE9BD1DAAB7EDCA5ULL, "646298491.ogg"},
	{0xBE9C0557AE5D4B8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandeddeselectedexit.evt"},
	{0xBE9C0F58C904C8C2ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_loop_c.anb"},
	{0xBE9CB05BCE170470ULL, "zz_seb_debug6215777.cl"},
	{0xBE9CB2DAFAD024A0ULL, "l_dae211ar_tavern_nt.wbd"},
	{0xBE9CF30F52828B3BULL, "6103386_m.fxe"},
	{0xBE9DCEEC7686B322ULL, "vxht_brothel_n_5zp_0_0.vlm"},
	{0xBE9E60387347EC2AULL, "617247357.ogg"},
	{0xBE9ED95C8B42285BULL, "360822305.ogg"},
	{0xBE9EFD9B158CF4B9ULL, "dae011_drunk1_sta6149668.cl"},
	{0xBE9F13EFD09A571EULL, "6090058_m.fxe"},
	{0xBE9FE886CE97A450ULL, "767057925.ogg"},
	{0xBEA0D4C38ADB0029ULL, "vxdr_primevaltomb_2u4_0_0.vlm"},
	{0xBEA0EDAD49579B3CULL, "mag131_keran6086169.cl"},
	{0xBEA11190830CBE98ULL, "dre000_cut_sibling_dies6036293.cl"},
	{0xBEA121F514134450ULL, "mag160cr_slaver.utc"},
	{0xBEA1871BF1A59534ULL, "572915443.ogg"},
	{0xBEA20094A411DDE2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int_floor_r.dds"},
	{0xBEA20B79E4F4CEA3ULL, "vxht_keep_n_60c_0_0.vlm"},
	{0xBEA28818AF1E30FDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_5_darren.ani"},
	{0xBEA2BB448AE51758ULL, "6172763_m.fxe"},
	{0xBEA31E57AE63954CULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandeddeselectedexit.gad"},
	{0xBEA3399DCE3765CCULL, "vxlt_darktown_5c6_0_0.vlm"},
	{0xBEA3C416CDF19C64ULL, "w_encounter3_41_0.mmh"},
	{0xBEA3C816CDF1A2B7ULL, "w_encounter3_41_0.mao"},
	{0xBEA3DA16CDF1C106ULL, "w_encounter3_41_0.msh"},
	{0xBEA4466A19691156ULL, "6068335_f.fxe"},
	{0xBEA44CEAC7F607ECULL, "art\\characters\\playercharacter\\humanmale\\rob_ands.msh"},
	{0xBEA452EAC7F611DEULL, "art\\characters\\playercharacter\\humanmale\\rob_ands.mmh"},
	{0xBEA45EEAC7F62675ULL, "art\\characters\\playercharacter\\humanmale\\rob_ands.mao"},
	{0xBEA4B19AFF9D82ABULL, "plt_lgt000pt_ultra_3_06.nss"},
	{0xBEA536BD51092549ULL, "6135819_m.fxe"},
	{0xBEA53893F8E85970ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocklarge.phy"},
	{0xBEA58B694FA6BBE2ULL, "927735043.ogg"},
	{0xBEA5A39F9F3EF4F1ULL, "gen_im_trs_bst_spd_fang_01.uti"},
	{0xBEA5AEE5968EBFADULL, "vxdr_primevaltomb_8n3_0_0.vlm"},
	{0xBEA5E69D418F3437ULL, "572721643.ogg"},
	{0xBEA62DD86A468AB1ULL, "gen00fl_anders6173769.cl"},
	{0xBEA65C2CA90809D5ULL, "vxht_bartrand_n_5an_0_0.vlm"},
	{0xBEA67AB15CA3F0DEULL, "1053936197.ogg"},
	{0xBEA6A376153A9BF5ULL, "121238006.ogg"},
	{0xBEA6AF9767C17B12ULL, "460247726.ogg"},
	{0xBEA6EE3A224CBF59ULL, "plt_mer300pt_mine_massacre.nss"},
	{0xBEA718A24AA44B3EULL, "fires_n.dds"},
	{0xBEA765D3B8439CB2ULL, "gen00fl_aveline6122743.cl"},
	{0xBEA8788932165182ULL, "seb221st_flora_1.stg"},
	{0xBEA8A93A3E7B7166ULL, "qun100cr_qunari_warrior3.utc"},
	{0xBEA8AB24E4F4B0C4ULL, "6026977_f.fxe"},
	{0xBEA90D6607FAE893ULL, "one100dg_kelder6055621.cl"},
	{0xBEA9AF8C60D681B9ULL, "6048961_f.fxe"},
	{0xBEA9B169E8857D92ULL, "6139724_m.fxe"},
	{0xBEAA0FD70CEF3A4BULL, "6078124_f.fxe"},
	{0xBEAA1D57AE69B24AULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandeddeselectedexit.ani"},
	{0xBEAA5326DBFF2537ULL, "995249113.ogg"},
	{0xBEAB03B6554CB3B9ULL, "gen00fl_bethany6102763.cl"},
	{0xBEAB827FE857D2A8ULL, "270476631.ogg"},
	{0xBEAB99FD2306BDDBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_basic.mmh"},
	{0xBEABA5FD2306D278ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_basic.mao"},
	{0xBEABB7FD2306F0C5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_basic.msh"},
	{0xBEABB9341BCF6BE4ULL, "mag370_meredith6092193.cl"},
	{0xBEACC422C0CE7655ULL, "llac_ht_chantry_2534_6.dds"},
	{0xBEACC957AD1F512BULL, "hlac_dr_primevalentrance_1726_1z.dds"},
	{0xBEACFAC72D1A396FULL, "189200320.ogg"},
	{0xBEAD5EA51AB6E563ULL, "lt_0_lowtown_docks_a3_233_0.msh"},
	{0xBEAD6CA51AB6FCBDULL, "lt_0_lowtown_docks_a3_233_0.mmh"},
	{0xBEAD70A51AB703CEULL, "lt_0_lowtown_docks_a3_233_0.mao"},
	{0xBEAD95D0776C1934ULL, "147457631.ogg"},
	{0xBEADD53A2252B332ULL, "plt_mer300pt_mine_massacre.plo"},
	{0xBEADF8299A0121B9ULL, "vrc221_hugin6041116.cl"},
	{0xBEAEC87666187EACULL, "868548588.ogg"},
	{0xBEAECBCC23C201EDULL, "25077723.ogg"},
	{0xBEAEFC9EB01D6C80ULL, "vxht_mansion_n_43f_0_1.vlm"},
	{0xBEAFAD0FC2DB2EE8ULL, "804332383.ogg"},
	{0xBEB010D8DAE69AB0ULL, "6208240_m.fxe"},
	{0xBEB080778F36F3D7ULL, "6116234_m.fxe"},
	{0xBEB136EC7DF8DCD6ULL, "177199615.ogg"},
	{0xBEB175C276615D74ULL, "238274558.ogg"},
	{0xBEB1956B556A6D3BULL, "6096927_m.fxe"},
	{0xBEB22E4D62BC04D1ULL, "693520284.ogg"},
	{0xBEB2432D338BFF2EULL, "mrl000_marethari6129212.cl"},
	{0xBEB25F5B0B327D38ULL, "483767683.ogg"},
	{0xBEB303CEEE1AD9C7ULL, "6084692_m.fxe"},
	{0xBEB361AA765CF97EULL, "6095794_m.fxe"},
	{0xBEB3C0F0F0A3612EULL, "w_encounter2_camp_37_0.msh"},
	{0xBEB3CA78D527F281ULL, "hlac_dr_primevaltomb_658_0x.dds"},
	{0xBEB3CEF0F0A3797FULL, "w_encounter2_camp_37_0.mao"},
	{0xBEB3D2A75F20A8DFULL, "qcr270_fake_meredith6066322.cl"},
	{0xBEB3DAF0F0A38D1CULL, "w_encounter2_camp_37_0.mmh"},
	{0xBEB46036BB60E9DDULL, "character_character_p284.psh.fxo.dx11.assembly"},
	{0xBEB48CCCD38BA33AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sackclose_n.dds"},
	{0xBEB4ABF379C91655ULL, "hlac_dr_0_deeproads_17639_1z.dds"},
	{0xBEB4B2DF2969528DULL, "hlac_lt_gallowstemplar_n_2390_0x.dds"},
	{0xBEB4CE099AFE2255ULL, "vxht_bartrand_n_5dm_0_0.vlm"},
	{0xBEB4EC780FED9CD9ULL, "6101210_m.fxe"},
	{0xBEB50172854D2D90ULL, "305101598.ogg"},
	{0xBEB50501A6542535ULL, "vxw_bonepit_4a_0_0.vlm"},
	{0xBEB5E68B0092F5F5ULL, "static_rendertodepth_p2.psh.fxo"},
	{0xBEB616A71F5B53D9ULL, "6168484_m.fxe"},
	{0xBEB6EB16609005A6ULL, "llac_lt_0_lowtown_docks_n_-65537_4.dds"},
	{0xBEB73A0B25CDDADAULL, "hm_default.mor"},
	{0xBEB7A072E7943EFFULL, "702903488.ogg"},
	{0xBEB7C7FF53C0CE77ULL, "14347518.ogg"},
	{0xBEB82583C3E5DB1CULL, "ts_qunari_02.gda"},
	{0xBEB84880BDB96D8FULL, "genip_cin_qun_relic.utp"},
	{0xBEB8DC9094A0DB34ULL, "214611232.ogg"},
	{0xBEB9590F10FC12C1ULL, "art\\vfx\\combat\\vfx_gib_qun_lbrt.ani"},
	{0xBEB9A59105428DE1ULL, "13920499.ogg"},
	{0xBEB9ADB83F2A99BDULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersmall.mao"},
	{0xBEB9B1B83F2AA0F6ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersmall.mmh"},
	{0xBEB9BBB83F2AB1F4ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_braziersmall.msh"},
	{0xBEBA7458EA8CD368ULL, "character_character_p104.psh.fxo"},
	{0xBEBABC0D50163B3AULL, "vxca_cavern1_haunted_e6y_0_0.vlm"},
	{0xBEBACE0688793D40ULL, "419803551.ogg"},
	{0xBEBB57BDD653575DULL, "gen00fl_fenris6127036.cl"},
	{0xBEBB6F322F366CC7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye10a.phy"},
	{0xBEBB82BF13D213E1ULL, "290180263.ogg"},
	{0xBEBBBFBC5532BDAEULL, "23863679.ogg"},
	{0xBEBC3E34090FCEEDULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb07a.phy"},
	{0xBEBCCDF025D60830ULL, "216014987.ogg"},
	{0xBEBCD275D7821A5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\sw.anb"},
	{0xBEBCE2A48C53B437ULL, "llac_w_bonepit_5_0.dds"},
	{0xBEBD0FCC68A05A84ULL, "one140cr_charade6137035.cl"},
	{0xBEBD48591266D9C2ULL, "hlac_ca_caverns1_24426_0y.dds"},
	{0xBEBDA4565F0AD3F2ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.cess.gad"},
	{0xBEBDA4A70FCDE48DULL, "mag360_thrask6043563.cl"},
	{0xBEBDE08DC099980DULL, "264688678.ogg"},
	{0xBEBDF20A06832DC0ULL, "489573581.ogg"},
	{0xBEBE00BF93599020ULL, "179083620.ogg"},
	{0xBEBEE04921658769ULL, "dae100_viscount6118501.cl"},
	{0xBEBF1BDF8C53E20FULL, "vxlt_gallowscourtyard_n_62l_0_0.vlm"},
	{0xBEBFE2DFF1F7AF08ULL, "vxht_brothel_n_6v1_0_0.vlm"},
	{0xBEC0640F11024463ULL, "art\\vfx\\combat\\vfx_gib_qun_lbrt.gad"},
	{0xBEC0D144B7EF9523ULL, "vxht_mansion_n_6k6_0_0.vlm"},
	{0xBEC111607A628335ULL, "935247494.ogg"},
	{0xBEC12B4B9A606F39ULL, "121727966.ogg"},
	{0xBEC1B76FE08D9F6AULL, "vxw_bonepitmining_a2_1z1_0_0.vlm"},
	{0xBEC264CE136325DAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_keeplogo_sa.dds"},
	{0xBEC2B21B26C9ABA9ULL, "vxlt_0_lowtown_a2_n_c_eqg_0_0.vlm"},
	{0xBEC2C1E734F2E0F5ULL, "lgt210_eau_symp_templar6144453.cl"},
	{0xBEC2E8F9ABDB913CULL, "mag321_comtesse_de_launcet6064912.cl"},
	{0xBEC33FE68CF43AD5ULL, "vxht_bartrand_n_5ol_0_0.vlm"},
	{0xBEC36E88705514A9ULL, "vxlt_warehouse_1bm_0_0.vlm"},
	{0xBEC374AABCAC4D1AULL, "art\\levels\\races\\elf\\props\\_textures\\prp_basket_n.dds"},
	{0xBEC397B7124A23FBULL, "dae000_hate_enemies6167564.cl"},
	{0xBEC40639E3100DDFULL, "km_tbe_urchin.mor"},
	{0xBEC4441DBB8B37ACULL, "vxca_caverns2prc_4tk_0_0.vlm"},
	{0xBEC44902FD5F94BAULL, "zz_followers_banter6159633.cl"},
	{0xBEC4A7565F10F729ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.cess.evt"},
	{0xBEC51148C1E2E63EULL, "dae300_vashothlead6062779.cl"},
	{0xBEC5670ABD88DF56ULL, "484866590.ogg"},
	{0xBEC5BD2498B9E3B7ULL, "fen110_fenris6096432.cl"},
	{0xBEC6099258C78E63ULL, "503017548.ogg"},
	{0xBEC641ED54A0F003ULL, "follower_banter6143402.cl"},
	{0xBEC64E0E55951722ULL, "558997519.ogg"},
	{0xBEC6707878F2D2EDULL, "vxdr_primevaltomb_7tn_0_0.vlm"},
	{0xBEC67CC4879387C7ULL, "vxht_0_hightown_atg_0_0.vlm"},
	{0xBEC6856667266404ULL, "243233760.ogg"},
	{0xBEC6BCAF6490EDA6ULL, "hlac_lt_gallowstemplar_f_2391_3y.dds"},
	{0xBEC6E4B06A04101BULL, "1641205.ogg"},
	{0xBEC7090DAC62E268ULL, "hlac_lt_0_lowtown_docks_n_131072_1x.dds"},
	{0xBEC73D31BA30B746ULL, "vxlt_blackemporium_140_0_0.vlm"},
	{0xBEC79A86793E5C3CULL, "vxca_caverns2_89p_0_1.vlm"},
	{0xBEC7BA302FDBA2F9ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_paintingc.phy"},
	{0xBEC9D0791BB8D112ULL, "hlac_lt_undercityserialkiller_-1_13z.dds"},
	{0xBEC9D0F2AC41F71AULL, "899693235.ogg"},
	{0xBECA04DE505BF251ULL, "hlac_ht_chantry_2534_37x.dds"},
	{0xBECA61EE485F385AULL, "434887837.ogg"},
	{0xBECB0F02C62FEDA2ULL, "m_dre180ar_ruins_entrance.wbd"},
	{0xBECB1CD15AA88638ULL, "6021226_f.fxe"},
	{0xBECB427E73721B39ULL, "625255251.ogg"},
	{0xBECC3E583BA77B69ULL, "112590591.ogg"},
	{0xBECCC3B8836CA5CBULL, "mag100_cullen6086359.cl"},
	{0xBECCE26A1DE5C028ULL, "hlac_lt_0_lowtown_docks_a3_196607_2x.dds"},
	{0xBECD4349A1873A26ULL, "vxht_0_hightown_a2_b09_0_0.vlm"},
	{0xBECE062EB2F7B466ULL, "zz_one_debug6079315.cl"},
	{0xBECE0E471DA27AC0ULL, "637849911.ogg"},
	{0xBECE53D907C07659ULL, "vxht_mansion_n_6jw_0_0.vlm"},
	{0xBECE8F34A419AF4CULL, "vxlt_darktown_5r5_0_0.vlm"},
	{0xBECF4321A9E5ADA3ULL, "vxht_mansion_n_6n9_0_0.vlm"},
	{0xBECF49D34935E531ULL, "6207903_m.fxe"},
	{0xBED00932445B8C2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_face_subtle.ani"},
	{0xBED016D435A7FE1FULL, "hla_ht_0_hightown_n_-1_0z.dds"},
	{0xBED06B9FDCAC2CA6ULL, "362851137.ogg"},
	{0xBED12B14D1C62356ULL, "551394975.ogg"},
	{0xBED1B54D89141F69ULL, "gen_im_trs_gen_blt_r0.uti"},
	{0xBED1D018947B8F0CULL, "vxw_woundedcoast_nc_0_0.vlm"},
	{0xBED28F565F1D0B98ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s2_c.cess.ani"},
	{0xBED2DAF0F0BE3E6FULL, "w_encounter2_camp_37_0.rml"},
	{0xBED2EB321719762DULL, "hlac_w_encounter2_262146_0x.dds"},
	{0xBED3776CFD349C05ULL, "6018596_m.fxe"},
	{0xBED43B0646A92EA5ULL, "6176670_m.fxe"},
	{0xBED46BAF3071706FULL, "308407544.ogg"},
	{0xBED485A3C8B29FC2ULL, "z_dust_little.dds"},
	{0xBED4BCFF82904E2AULL, "ico_boots_med_3.dds"},
	{0xBED4E2556AE8EB6DULL, "vxdr_primevaltomb_7qm_0_0.vlm"},
	{0xBED5315B1EF820CDULL, "hlac_ca_cavern1_haunted_24424_8y.dds"},
	{0xBED53E86D83F54E1ULL, "vxw_bonepit_26o_0_0.vlm"},
	{0xBED5A8A614093EC1ULL, "6095775_m.fxe"},
	{0xBED5FAB15593620EULL, "694015953.ogg"},
	{0xBED612CCA754E8AEULL, "vxlt_0_lowtown_docks_n_azz_0_0.vlm"},
	{0xBED660ABA80781E4ULL, "w_woundedcoast_180_0.msh"},
	{0xBED672ABA807A04DULL, "w_woundedcoast_180_0.mao"},
	{0xBED676ABA807A686ULL, "w_woundedcoast_180_0.mmh"},
	{0xBED6DD4E61DEF5D8ULL, "hlac_lt_darktown_-1_65y.dds"},
	{0xBED6FEE5908FAED1ULL, "zz_followers_banter6157946.cl"},
	{0xBED73CD8A6463242ULL, "401051005.ogg"},
	{0xBED76545CBADD299ULL, "mag311_emile_de_launcet6185104.cl"},
	{0xBED76DD59FDD13EFULL, "479307832.ogg"},
	{0xBED81E372A19673CULL, "432533424.ogg"},
	{0xBED8B9BEC356F43DULL, "518673418.ogg"},
	{0xBED8F85C1AC4858BULL, "6139533_f.fxe"},
	{0xBED905D0CE729897ULL, "vxlt_0_lowtown_ex7_0_0.vlm"},
	{0xBED9343BE9619633ULL, "6175848_m.fxe"},
	{0xBED9609AD4B50757ULL, "mcr310_argument6074493.cl"},
	{0xBED9B9F0F0C424E7ULL, "w_encounter2_camp_37_0.phy"},
	{0xBED9E1D5D637738CULL, "vxht_0_hightown_a2_c73_0_0.vlm"},
	{0xBEDA821CFE1CCE68ULL, "6130981_m.fxe"},
	{0xBEDA955A46DB5ED3ULL, "zz_followers_banter6159234.cl"},
	{0xBEDADC95C392C3D6ULL, "hlac_w_sundermount_65530_0x.dds"},
	{0xBEDBABC035FB821DULL, "vxht_keep_42v_0_0.vlm"},
	{0xBEDCC5B5F9B68ED9ULL, "vxht_mansion_n_6mv_0_0.vlm"},
	{0xBEDD0E143703E3E2ULL, "plt_cod_hst_woundedcoast.plo"},
	{0xBEDD26CE8F10ACCBULL, "hlac_lt_0_lowtown_a3_n_c_131072_4x.dds"},
	{0xBEDD2FE4D7CA40B1ULL, "vxht_chantry_27o_0_0.vlm"},
	{0xBEDD84E14C6E32E1ULL, "6067752_m.fxe"},
	{0xBEDD9C60CB66238CULL, "hlac_ca_caverns2_5802_5x.dds"},
	{0xBEDDDE6A064C0E0FULL, "dae000_merchant_weapons_dm6164395.cl"},
	{0xBEDDEAE6DD80236CULL, "555235232.ogg"},
	{0xBEDDF1324467A0A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_face_subtle.evt"},
	{0xBEDE0145A185AE74ULL, "449120822.ogg"},
	{0xBEDE02D574CA9118ULL, "zz_followers_banter6158364.cl"},
	{0xBEDE5224735E001BULL, "594525124.ogg"},
	{0xBEDEB3A02A1DEC8DULL, "qcr000_arishok6042519.cl"},
	{0xBEDEC4CCE2708499ULL, "hlac_ht_0_hightown_a3_131072_0y.dds"},
	{0xBEDEC73C999400C8ULL, "vxw_bonepit_26j_0_0.vlm"},
	{0xBEDF01EBEB00ADC6ULL, "vxca_cave_2ga_0_1.vlm"},
	{0xBEDF6387BF731069ULL, "vxdr_0_deeproads_crt_0_2.vlm"},
	{0xBEDF673B0C1B1559ULL, "sib131ip_chest_will.utp"},
	{0xBEDF77B922C62EDFULL, "88168361.ogg"},
	{0xBEE06DABA80FEC51ULL, "w_woundedcoast_180_0.phy"},
	{0xBEE0F78186FBAB00ULL, "hlac_lt_0_lowtown_docks_65534_2y.dds"},
	{0xBEE1305F6C5614FCULL, "art\\cinematics\\deathblows\\meatgolem\\cs_pc_meatwarr.ani"},
	{0xBEE19AA015EFE735ULL, "vxca_cavern1_haunted_i22_0_0.vlm"},
	{0xBEE1DCB0A340CC72ULL, "vxht_bartrand_n_65b_0_0.vlm"},
	{0xBEE2D1401BFABFEBULL, "hlac_lt_hangedman_-1_1y.dds"},
	{0xBEE34C53B59B97B4ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.impa.ani"},
	{0xBEE3A9B7D479581BULL, "ser270_emeric6149279.cl"},
	{0xBEE3B063CA356D61ULL, "vxw_bonepit_21n_0_0.vlm"},
	{0xBEE4025671300EBBULL, "em_mrl351cr_fakemaleelf.mor"},
	{0xBEE429EAC82BB079ULL, "art\\characters\\playercharacter\\humanmale\\rob_ands.phy"},
	{0xBEE43F1459951B68ULL, "941617433.ogg"},
	{0xBEE4BA35FB743716ULL, "937744233.ogg"},
	{0xBEE4D37F7A1E5EBAULL, "art\\characters\\playercharacter\\humanmale\\bdy_medb.mmh"},
	{0xBEE4D77F7A1E65F1ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medb.mao"},
	{0xBEE4E57F7A1E7D00ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medb.msh"},
	{0xBEE4F032446DBDB5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_face_subtle.gad"},
	{0xBEE61B7AC77F35B9ULL, "vxw_woundedcoast_1rc_0_0.vlm"},
	{0xBEE65A09B3BBD744ULL, "6121558_m.fxe"},
	{0xBEE69DF9809720E4ULL, "gen_im_cns_flu_fenris_02.uti"},
	{0xBEE6F74AD79DE484ULL, "gen00fl_varric6133764.cl"},
	{0xBEE6FE7D923AAC06ULL, "vxlt_0_lowtown_docks_bw5_0_0.vlm"},
	{0xBEE777ADC068B117ULL, "vxlt_0_lowtown_es6_0_0.vlm"},
	{0xBEE7F07A38A19F99ULL, "one350_variel6218666.cl"},
	{0xBEE7F8ECF8D1AF9BULL, "gen00fl_merrill6127276.cl"},
	{0xBEE826D02F348522ULL, "gen00fl_aveline6129030.cl"},
	{0xBEE83D5F6C5C493EULL, "art\\cinematics\\deathblows\\meatgolem\\cs_pc_meatwarr.gad"},
	{0xBEE89FAA89B2447FULL, "art\\vfx\\combat\\vfx_f_mer_creep_act_c.mmh"},
	{0xBEE8B5C56806CCCAULL, "dae000_merchant_mage_hf6151497.cl"},
	{0xBEE8F25350A5BBD0ULL, "vxlt_0_lowtown_a2_gyy_0_0.vlm"},
	{0xBEE9620B5616431DULL, "vxw_encounter1_h8_0_0.vlm"},
	{0xBEE97AF18329A48BULL, "gen00fl_sebastian6174648.cl"},
	{0xBEEA0A057380953EULL, "180577370.ogg"},
	{0xBEEA4953B5A1B116ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.impa.gad"},
	{0xBEEB060DD0DCC64CULL, "vrc221_hugin.cnv"},
	{0xBEEB130BD806C271ULL, "174921681.ogg"},
	{0xBEEB2CE8BCECF8B5ULL, "ef_dalish_apprentice1.mor"},
	{0xBEEB419BEBC8095BULL, "vxca_cavern1_haunted_i99_0_0.vlm"},
	{0xBEEB86422DB81C23ULL, "6206191_m.fxe"},
	{0xBEEB8EAB5035EA17ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_lp.gad"},
	{0xBEEC00202397052CULL, "hlac_lt_gallowstemplar_2394_6x.dds"},
	{0xBEED39198B8A1948ULL, "vxw_bonepit_21k_0_0.vlm"},
	{0xBEED9EFBFF156921ULL, "87556949.ogg"},
	{0xBEEDE80DD0DEC97CULL, "vrc221_hugin.bnk"},
	{0xBEEE7C3C60FABB83ULL, "hlac_ht_chantry_2534_7y.dds"},
	{0xBEEE884CDBEA99A9ULL, "6017496_m.fxe"},
	{0xBEEE8AE1C616554BULL, "6136113_m.fxe"},
	{0xBEEE99EA4A613AECULL, "6102676_m.fxe"},
	{0xBEEEA62867280351ULL, "6078988_f.fxe"},
	{0xBEEECEAC000EEE2CULL, "996458221.ogg"},
	{0xBEEF07C70474EFF2ULL, "276425305.ogg"},
	{0xBEEF1B28C4229B69ULL, "6173743_m.fxe"},
	{0xBEEF1E941A7C8619ULL, "vxlt_darktown_6hi_0_0.vlm"},
	{0xBEEF59B4130F1C6AULL, "lgt160_wk_avern_bandits6112752.cl"},
	{0xBEEFBF91C7FB10EEULL, "vxlt_darktown_47q_0_0.vlm"},
	{0xBEF074D33F8C52DFULL, "and100_anders6096015.cl"},
	{0xBEF0D9F5AAEB7F19ULL, "678672654.ogg"},
	{0xBEF0DC4B90631CFBULL, "mag291ip_barrel.utp"},
	{0xBEF0FCAC7E3F9E88ULL, "685709517.ogg"},
	{0xBEF1259D4B78ACC3ULL, "ss_merrillcu.cut"},
	{0xBEF16453B5A7FE75ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.impa.evt"},
	{0xBEF18624E7CFC1B8ULL, "vxdr_primevaltomb_86j_0_0.vlm"},
	{0xBEF18DF76C1CBC32ULL, "llac_w_encounter2_-65534_0.dds"},
	{0xBEF23B6EE58B075BULL, "vxht_keep_n_4n7_0_0.vlm"},
	{0xBEF24345168C51FFULL, "mag311_keran6041697.cl"},
	{0xBEF271FA2A3858F2ULL, "mer320_hubert6082002.cl"},
	{0xBEF2F6B3B0D5C780ULL, "vxlt_undercityencounter_11n_0_0.vlm"},
	{0xBEF2FCAC94676117ULL, "vxlt_0_lowtown_h1y_0_0.vlm"},
	{0xBEF31F20B5BB28DBULL, "6136816_m.fxe"},
	{0xBEF330393739EAE6ULL, "6098850_m.fxe"},
	{0xBEF3613E13226CCFULL, "vxht_mansion2_1w6_0_0.vlm"},
	{0xBEF375E96CD1B08AULL, "vxht_chantry_2tl_0_1.vlm"},
	{0xBEF3B1F7F9FCCEABULL, "class_abilities_sebastian.gda"},
	{0xBEF3BAFD2344AB10ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0rw_basic.phy"},
	{0xBEF3F46F2FF95EB9ULL, "character_character_p80.psh.fxo.dx11"},
	{0xBEF444C5E310C43DULL, "ave200_donnic6110739.cl"},
	{0xBEF4A7D11D089E4BULL, "285910460.ogg"},
	{0xBEF4C34DC7CFE713ULL, "6077135_m.fxe"},
	{0xBEF4F48635797A4BULL, "vxlt_hangedman_n_1jt_0_0.vlm"},
	{0xBEF53F2F25E02239ULL, "6176568_m.fxe"},
	{0xBEF599BCC4C13258ULL, "vxht_0_hightown_a2_90d_0_0.vlm"},
	{0xBEF5E98AB25EC997ULL, "vxlt_0_lowtown_en1_0_0.vlm"},
	{0xBEF64324E53CA2CDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stocks_n.dds"},
	{0xBEF64FAEEE8F3F46ULL, "vxht_mansion_n_61y_0_0.vlm"},
	{0xBEF6C58FBA23A48CULL, "vxht_0_hightown_a2_c91_0_0.vlm"},
	{0xBEF7531A1858DF0FULL, "6049431_m.fxe"},
	{0xBEF7ADBCB3256419ULL, "950209042.ogg"},
	{0xBEF7B4DF7C283B07ULL, "ser191_emeric6120098.cl"},
	{0xBEF7B7EAC76818D9ULL, "dae210ar_docks.lst"},
	{0xBEF7B941A904D518ULL, "5015674.ogg"},
	{0xBEF7DAF5FE757FC7ULL, "256507709.ogg"},
	{0xBEF87B288C3751D7ULL, "w_blightlands_198_0.mmh"},
	{0xBEF87F288C3758E4ULL, "w_blightlands_198_0.mao"},
	{0xBEF891288C377749ULL, "w_blightlands_198_0.msh"},
	{0xBEF8AA8283B48868ULL, "vxlt_foundry_n_yc_0_0.vlm"},
	{0xBEF984B985E34B1CULL, "art\\characters\\playercharacter\\humanmale\\boo_medj.mao"},
	{0xBEF990B985E35FBFULL, "art\\characters\\playercharacter\\humanmale\\boo_medj.mmh"},
	{0xBEF996B985E36A61ULL, "art\\characters\\playercharacter\\humanmale\\boo_medj.msh"},
	{0xBEF9DB3FFFF9CFA9ULL, "w_encounter3_76_0.mao"},
	{0xBEF9E73FFFF9E442ULL, "w_encounter3_76_0.mmh"},
	{0xBEF9E93FFFF9E7F8ULL, "w_encounter3_76_0.msh"},
	{0xBEF9F2FE26B36A53ULL, "dae000_bodhan6129082.cl"},
	{0xBEFA3FE112D2404DULL, "l_qcr221ar_keep.bnk"},
	{0xBEFAB6925E8458A2ULL, "92448269.ogg"},
	{0xBEFAF39604F6915FULL, "vxlt_0_lowtown_a2_n_c_hny_0_0.vlm"},
	{0xBEFB1E639200CB1EULL, "lowtown_male_beggar_sta6156756.cl"},
	{0xBEFBA989ADD5D65DULL, "follower_banter6173120.cl"},
	{0xBEFBBDA64C56C0DBULL, "6133451_m.fxe"},
	{0xBEFBFCE07FAFEBF7ULL, "vxlt_0_lowtown_a3_n_c_gcl_0_0.vlm"},
	{0xBEFC0FB6D51D6B15ULL, "w_blightlands_99_0.msh"},
	{0xBEFC11B6D51D6E8BULL, "w_blightlands_99_0.mmh"},
	{0xBEFC1DB6D51D8328ULL, "w_blightlands_99_0.mao"},
	{0xBEFC4FC487845401ULL, "ave331cr_insurgent_guard4.utc"},
	{0xBEFC7A67ED77B99FULL, "gen00fl_aveline6137566.cl"},
	{0xBEFCFD52A0AB95C8ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4sd_basic_2.mao"},
	{0xBEFD12B672D869FFULL, "ts_templars_arm_props.gda"},
	{0xBEFDE3CD8679A675ULL, "459789207.ogg"},
	{0xBEFE21E08936E4E4ULL, "6079039_m.fxe"},
	{0xBEFE30F177C5836FULL, "water_pan_add.dds"},
	{0xBEFE3954539D7943ULL, "6120819_m.fxe"},
	{0xBEFE861A77EA8084ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_min_c.anb"},
	{0xBEFEAFEAC76E29B0ULL, "dae210ar_docks.ncs"},
	{0xBEFEBD69560AE970ULL, "hlac_lt_0_lowtown_n_65536_6y.dds"},
	{0xBEFF046AC2DF1B29ULL, "fen110_fenris6100276.cl"},
	{0xBEFFBBAB5046DF9DULL, "art\\characters\\creatures\\pridedemon\\animation\\a_lp.ani"},
	{0xBF001F901E975C47ULL, "6115868_m.fxe"},
	{0xBF008754573E0B9CULL, "dae300st_orsino.stg"},
	{0xBF015F006F35FA14ULL, "6068550_m.fxe"},
	{0xBF016E89865D7997ULL, "vxlt_0_lowtown_h6x_0_0.vlm"},
	{0xBF0190FDA7B1415BULL, "6131916_m.fxe"},
	{0xBF01D31B0518854FULL, "vxht_mansion2_1z1_0_0.vlm"},
	{0xBF01F14B9071B31EULL, "mag291ip_barrel.ncs"},
	{0xBF01F1B1678DB35BULL, "vxlt_0_lowtown_a3_n_c_ehw_0_0.vlm"},
	{0xBF020DCE8B28C35BULL, "219736450.ogg"},
	{0xBF02866E9A89FDD5ULL, "373551980.ogg"},
	{0xBF032A797043EC84ULL, "vxht_bartrand_n_733_0_0.vlm"},
	{0xBF036FB98EAD70EDULL, "follower_barks6164484.cl"},
	{0xBF03A3F4925CE9AEULL, "hlac_w_woundedcoast_-131064_0y.dds"},
	{0xBF03B536296D0AB8ULL, "vxlt_0_lowtown_docks_bay_0_0.vlm"},
	{0xBF03D1DC40FF33ECULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_up_c.impa.ani"},
	{0xBF03D2042A8165CEULL, "dre000_dougal6129824.cl"},
	{0xBF03E2377626DD06ULL, "vxlt_0_lowtown_docks_ba7_0_0.vlm"},
	{0xBF0409E3FC472D96ULL, "lgt331ar_hybris.are"},
	{0xBF045B67A454E217ULL, "vxlt_0_lowtown_ei0_0_0.vlm"},
	{0xBF048227A96347DFULL, "6078534_m.fxe"},
	{0xBF04AC1E77A3BBA9ULL, "gen00fl_merrill6119587.cl"},
	{0xBF05C097701267AFULL, "darktown_fem_elf_sta6174517.cl"},
	{0xBF05C323813B44D7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_beams.msh"},
	{0xBF05D523813B633AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_beams.mao"},
	{0xBF05D923813B6A49ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0dd_beams.mmh"},
	{0xBF05DF4CB571BA36ULL, "6064923_m.fxe"},
	{0xBF05E46161E3D9FEULL, "176294421.ogg"},
	{0xBF063DCB80947793ULL, "6100299_m.fxe"},
	{0xBF0691D5768AA1FFULL, "272167859.ogg"},
	{0xBF06CFA4018B323FULL, "projdecal_projdecal_p0.vsh.fxo"},
	{0xBF072F60A8F9D517ULL, "vxca_cavern1_haunted_ehr_0_0.vlm"},
	{0xBF073E0D4391B20BULL, "6108088_m.fxe"},
	{0xBF07462600AF974FULL, "481800060.ogg"},
	{0xBF077B8D6EC9A568ULL, "gen00fl_anders6119748.cl"},
	{0xBF07948B6D9C8D44ULL, "gen00fl_isabela6101813.cl"},
	{0xBF07AE5E73CB2620ULL, "255681984.ogg"},
	{0xBF07CDEA4AEE0F73ULL, "mag121_eau_banter6107379.cl"},
	{0xBF07E96E08A55388ULL, "6053606_m.fxe"},
	{0xBF081D074BD65E5BULL, "594681321.ogg"},
	{0xBF08B0EAC7767F9EULL, "dae210ar_docks.are"},
	{0xBF08F3B4DC81EC9CULL, "one220st_yevhen.stg"},
	{0xBF09C3D4C2C60BBAULL, "llac_dr_dungeon_3214_6.dds"},
	{0xBF09DEDC4103B52EULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_up_c.impa.gad"},
	{0xBF09F27F7A3DAFC5ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medb.phy"},
	{0xBF0B7A0571274978ULL, "plt_prm_cod_lion_shield.plo"},
	{0xBF0B8087CFA168C1ULL, "115155094.ogg"},
	{0xBF0B83C7E20451D3ULL, "vxlt_0_lowtown_docks_a3_n_d50_0_0.vlm"},
	{0xBF0C6E9E03616211ULL, "mag100_cullen6140932.cl"},
	{0xBF0CA51A77F6F287ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_min_c.mmh"},
	{0xBF0CA678992D088BULL, "404231243.ogg"},
	{0xBF0CC7D7A14E81FCULL, "196397922.ogg"},
	{0xBF0CD430541A1AADULL, "zz_ave_debug6207522.cl"},
	{0xBF0D92DAF6C96497ULL, "42865459.ogg"},
	{0xBF0DF34519FC86EBULL, "hlac_ht_mansion2_818_57y.dds"},
	{0xBF0DFE5194149A8DULL, "dae000_oldmangossip6174769.cl"},
	{0xBF0E41E587B67D71ULL, "598661563.ogg"},
	{0xBF0E6348E3B46BA8ULL, "gen00fl_dog6165742.cl"},
	{0xBF0E88F6E2515734ULL, "lowtown_male_urchin_sta6164231.cl"},
	{0xBF0EB327BA89CDF7ULL, "573998873.ogg"},
	{0xBF0EB4B830F5F68CULL, "vxw_bonepit_1vi_0_0.vlm"},
	{0xBF0ECF60BC1BB740ULL, "vxlt_gallowscourtyard_n_73i_0_0.vlm"},
	{0xBF0EE19A206F86DFULL, "dae000_gamlen6147224.cl"},
	{0xBF0EF1F49673BF7DULL, "vxlt_gallowstemplar_5co_0_0.vlm"},
	{0xBF0EF84884A5A8D4ULL, "542246973.ogg"},
	{0xBF0F198644312ECEULL, "gen00fl_fenris6126456.cl"},
	{0xBF0F6168985C4683ULL, "vxw_bonepitmining_a2_1nx_0_0.vlm"},
	{0xBF107C6F119ED9E1ULL, "796469998.ogg"},
	{0xBF10C3A4056AAFB7ULL, "breadoven_sa.dds"},
	{0xBF10D9DC4109CB0DULL, "art\\vfx\\combat\\vfx_m_sta_atk_r_up_c.impa.evt"},
	{0xBF11A0AA89D54ABCULL, "art\\vfx\\combat\\vfx_f_mer_creep_act_c.anb"},
	{0xBF121D02C66C0858ULL, "m_dre180ar_ruins_entrance.bnk"},
	{0xBF126BD9E8334530ULL, "hlac_lt_0_lowtown_a2_-65535_6x.dds"},
	{0xBF139825A368528AULL, "6097055_m.fxe"},
	{0xBF13F18E25042006ULL, "vxht_0_hightown_n_cbm_0_0.vlm"},
	{0xBF1443EB439735C7ULL, "1041288878.ogg"},
	{0xBF14938C73183907ULL, "6115271_m.fxe"},
	{0xBF14E2C1786CBEB5ULL, "hlac_w_encounter3_-131074_0y.dds"},
	{0xBF15047DA358F052ULL, "vxlt_0_lowtown_docks_a3_n_d55_0_0.vlm"},
	{0xBF15069A24BA50A4ULL, "zz_ave_debug6056796.cl"},
	{0xBF150BCC6280404CULL, "vxw_woundedcoast_14l_0_0.vlm"},
	{0xBF15C169A35DE484ULL, "272432138.ogg"},
	{0xBF15D27A7E91D73EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkb.evt"},
	{0xBF161F832BB33EABULL, "vxlt_gallowstemplar_43h_0_0.vlm"},
	{0xBF162FF46689974AULL, "696206445.ogg"},
	{0xBF164C3E23058E77ULL, "hlac_lt_undercitydungeon_-1_31z.dds"},
	{0xBF17AF3DE4FC3394ULL, "vxht_0_hightown_a3_bzx_0_0.vlm"},
	{0xBF17D6154B1E6EEFULL, "fxa_pow_smokes.mao"},
	{0xBF1830A03ECAC010ULL, "140733994.ogg"},
	{0xBF1872587C2D233AULL, "llac_lt_0_lowtown_a3_n_c_-131072_6.dds"},
	{0xBF18CC607A310626ULL, "vxdr_primevaltomb_3i6_0_0.vlm"},
	{0xBF18FB114A3E3D27ULL, "vxca_caverns2_7jt_0_0.vlm"},
	{0xBF191DED81F5ED66ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgred_0.pwk"},
	{0xBF1923CB87BFD096ULL, "hlac_ht_keep_n_3732_3x.dds"},
	{0xBF19508BC743596AULL, "hlac_ht_chantry_n_2534_19z.dds"},
	{0xBF19D5D14BB1B139ULL, "621312983.ogg"},
	{0xBF19F5A4D3FA6A53ULL, "vxlt_0_lowtown_docks_a3_n_d03_0_0.vlm"},
	{0xBF1A09694E4A098CULL, "vxw_woundedcoast_wd_0_0.vlm"},
	{0xBF1A49349C3C6D50ULL, "374412179.ogg"},
	{0xBF1A84B8E552AA5DULL, "147281613.ogg"},
	{0xBF1A97374BDEFB3EULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_maka_0t.dds"},
	{0xBF1B57E93043DFDCULL, "6059334_m.fxe"},
	{0xBF1B7B7C23E12F8DULL, "vxht_0_hightown_a3_c1a_0_0.vlm"},
	{0xBF1BADA7D856F633ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clenchfists.evt"},
	{0xBF1BFFC61AFA96D4ULL, "101119412.ogg"},
	{0xBF1CE97A7E981D69ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkb.gad"},
	{0xBF1D0A5B183077D0ULL, "587458216.ogg"},
	{0xBF1D551C96071FC4ULL, "56552843.ogg"},
	{0xBF1D7475DB96004CULL, "gen00fl_anders6096610.cl"},
	{0xBF1DCFAE7CD39F77ULL, "vxht_brothel_n_57o_0_1.vlm"},
	{0xBF1DEDB55FCD1BBEULL, "llac_lt_0_lowtown_a2_1_6.dds"},
	{0xBF1DF921740AF44DULL, "vxdr_primevaltomb_56z_0_0.vlm"},
	{0xBF1ECFC3FBB9762AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_bw_teleport.gad"},
	{0xBF1EFDCE40793EB8ULL, "vxw_bonepitdestroyed_a3_1sb_0_0.vlm"},
	{0xBF1F03F8B86F3613ULL, "6031295_f.fxe"},
	{0xBF1F6A6E700A7F36ULL, "6036771_m.fxe"},
	{0xBF2066A8ADC87C6EULL, "hf_amb_merchant_mage.mor"},
	{0xBF2197BD37C652D8ULL, "vxca_cavern1_haunted_gkn_0_0.vlm"},
	{0xBF21BF1B00DE31F6ULL, "vxw_sundermount_35o_0_0.vlm"},
	{0xBF21CBC524A6C7CAULL, "vxlt_darktown_624_0_0.vlm"},
	{0xBF220313D838FC40ULL, "hlac_lt_0_lowtown_docks_0_4x.dds"},
	{0xBF22A6A7D85D0974ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clenchfists.gad"},
	{0xBF23765A954F08D2ULL, "vxlt_0_lowtown_docks_a3_n_d06_0_0.vlm"},
	{0xBF239B9F0CF6EBFEULL, "vxca_cavern1_haunted_i76_0_0.vlm"},
	{0xBF23ADEF8BA3F5EEULL, "vxdr_0_deeproads_gfu_0_0.vlm"},
	{0xBF23BA7A7E9DEBB7ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkb.ani"},
	{0xBF246AE5CDC0A21DULL, "363541573.ogg"},
	{0xBF2491601DA9572BULL, "vxlt_gallowstemplar_46k_0_0.vlm"},
	{0xBF24C97F8FB16A6EULL, "6070113_m.fxe"},
	{0xBF25B2C3FBBF63C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_bw_teleport.evt"},
	{0xBF2635E112F76A77ULL, "l_qcr221ar_keep.wbd"},
	{0xBF26477AC62E4358ULL, "mag110_samson6018590.cl"},
	{0xBF27527755923577ULL, "vxlt_0_lowtown_a3_n_c_gpi_0_0.vlm"},
	{0xBF2809688F395B60ULL, "hlac_ht_mansion_n_3372_0y.dds"},
	{0xBF28218457261EDDULL, "6116409_m.fxe"},
	{0xBF284D2DC632074EULL, "ave100_donnic6105999.cl"},
	{0xBF289564F8D94E14ULL, "vxlt_0_lowtown_a2_n_heq_0_0.vlm"},
	{0xBF2897577BCC5830ULL, "gen00fl_merrill6151009.cl"},
	{0xBF28D14501DE0301ULL, "gen_im_acc_blt_act2_04.uti"},
	{0xBF2910E3FC665661ULL, "lgt331ar_hybris.lst"},
	{0xBF29496CA0CDE6F5ULL, "gen00fl_fenris6099830.cl"},
	{0xBF2991E6DD9E8889ULL, "vxht_playershouse_n_1rl_0_1.vlm"},
	{0xBF29AA5B83F67BE1ULL, "hlac_lt_0_lowtown_a2_n_c_65535_3z.dds"},
	{0xBF29B5A7D8634152ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clenchfists.ani"},
	{0xBF29CBDA82B4D620ULL, "6215548_spooky_3.cut"},
	{0xBF2A5538153EB201ULL, "6150343_f.fxe"},
	{0xBF2A67A9E96178ABULL, "fen110_party_comment6095249.cl"},
	{0xBF2AADD39C7F05CEULL, "6096067_m.fxe"},
	{0xBF2AF596D243E99DULL, "430861535.ogg"},
	{0xBF2B4DA8968025ECULL, "lgt250_way_dalish_conflict6147122.cl"},
	{0xBF2BD00030C8C4A3ULL, "gen00fl_anders6119187.cl"},
	{0xBF2BEA717DC1F65AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1c.mmh"},
	{0xBF2BEE717DC1FC91ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1c.mao"},
	{0xBF2BFC717DC214A0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1c.msh"},
	{0xBF2CFCB8A5188FCEULL, "hlac_dr_primevaltomb_659_8x.dds"},
	{0xBF2CFDC122B02EF2ULL, "dae100_bran6175868.cl"},
	{0xBF2D0A5824C3F72DULL, "vxht_0_hightown_n_c5c_0_0.vlm"},
	{0xBF2D162B6942DD68ULL, "116482446.ogg"},
	{0xBF2D2E015F86E9AFULL, "296964870.ogg"},
	{0xBF2D73D72DFC9A65ULL, "6016644_m.fxe"},
	{0xBF2E0927A7DF15E0ULL, "23529607.ogg"},
	{0xBF2E869D10F4581CULL, "352764805.ogg"},
	{0xBF2EAAD1CD861AD7ULL, "514115417.ogg"},
	{0xBF2ED73FABDEACB5ULL, "zz_isa_debug6133235.cl"},
	{0xBF2FAD94632CABCDULL, "26976661.ogg"},
	{0xBF2FB3D971256861ULL, "656537964.ogg"},
	{0xBF2FC794E35AAD1AULL, "art\\characters\\playercharacter\\elfmale\\em_rob_appd.mmh"},
	{0xBF2FCB94E35AB451ULL, "art\\characters\\playercharacter\\elfmale\\em_rob_appd.mao"},
	{0xBF2FD994E35ACC60ULL, "art\\characters\\playercharacter\\elfmale\\em_rob_appd.msh"},
	{0xBF3008E3FC6C6758ULL, "lgt331ar_hybris.ncs"},
	{0xBF306E9FEB04B9C1ULL, "vxht_keep_n_609_0_0.vlm"},
	{0xBF307AFCEB9C1E86ULL, "ser210_witness6168169.cl"},
	{0xBF3084B13D8F2F91ULL, "6150216_m.fxe"},
	{0xBF30B0E023D8EFC7ULL, "mag311_emile_de_launcet6043495.cl"},
	{0xBF31ED6C0325039DULL, "485058210.ogg"},
	{0xBF3284F69F181EB5ULL, "vxlt_0_lowtown_d66_0_0.vlm"},
	{0xBF32AAC3FBC9E010ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_bw_teleport.ani"},
	{0xBF32BF744F156278ULL, "6099701_m.fxe"},
	{0xBF32D60017B6A592ULL, "vxca_caverns1_mines_e7v_0_0.vlm"},
	{0xBF32D87C73213F6AULL, "nat380_sibling6067992.cl"},
	{0xBF32E338719D89CDULL, "vxca_cavern1_haunted_iz2_0_0.vlm"},
	{0xBF32FCB06F7E6E49ULL, "hlac_ht_chantry_n_2534_52x.dds"},
	{0xBF332EC4B3B9E38CULL, "148452990.ogg"},
	{0xBF33774BB183B1E0ULL, "vxlt_gallowstemplar_n_35a_0_0.vlm"},
	{0xBF33A4014C397BA4ULL, "595109825.ogg"},
	{0xBF33A5435B1189BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_right.ani"},
	{0xBF3455039F9C52D8ULL, "282799810.ogg"},
	{0xBF34C31EF362D5E6ULL, "6167481_f.fxe"},
	{0xBF34CFF85AFF858EULL, "vxlt_foundry_n_xb_0_0.vlm"},
	{0xBF35C45447884DF7ULL, "vxlt_0_lowtown_a3_n_c_gwh_0_0.vlm"},
	{0xBF35E2AE729CE313ULL, "890879538.ogg"},
	{0xBF35F754A5A7AD5EULL, "vxht_0_hightown_n_a2o_0_0.vlm"},
	{0xBF3608DE2C6B2CA2ULL, "468938897.ogg"},
	{0xBF36193D756052E8ULL, "dae000_gamlen6124143.cl"},
	{0xBF3621000E3B60DAULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonb.msh"},
	{0xBF362F000E3B7813ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonb.mao"},
	{0xBF3633000E3B7F20ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemonb.mmh"},
	{0xBF36389EFF8C565FULL, "6136669_m.fxe"},
	{0xBF3667CAEE605F69ULL, "mag170_thrask6167398.cl"},
	{0xBF37BA89B07D5371ULL, "vxw_woundedcoast_1fg_0_0.vlm"},
	{0xBF37C41F9AD89056ULL, "178327776.ogg"},
	{0xBF37D3E257C70EB2ULL, "6114398_m.fxe"},
	{0xBF38300ED0B9EC54ULL, "llac_lt_0_lowtown_a2_-196609_3.dds"},
	{0xBF386AF351599FE9ULL, "w_blightlands_124_0.phy"},
	{0xBF3913528A66FBB2ULL, "zz_approval_debug6003612.cl"},
	{0xBF393C16A4254FB2ULL, "hlac_w_encounter2_camp_65535_0z.dds"},
	{0xBF3960FAA416C6BFULL, "37747191.ogg"},
	{0xBF397C00CA56CF6AULL, "vxlt_warehouse_n_19i_0_1.vlm"},
	{0xBF39816790EF451CULL, "dae000_epic_merchant_26179391.cl"},
	{0xBF39A3E3987D10B1ULL, "plt_cod_mgc_school_blood.plo"},
	{0xBF39DD37EB9B5B09ULL, "vxw_woundedcoast_3q_0_0.vlm"},
	{0xBF39EB2F252B379BULL, "isa311_isabela6092507.cl"},
	{0xBF3A065332D49D33ULL, "669889698.ogg"},
	{0xBF3A478B6C5608FDULL, "vxlt_gallowstemplar_5rl_0_0.vlm"},
	{0xBF3A913801989CFBULL, "character_character_p101.psh.fxo.dx11"},
	{0xBF3AB2435B17BEFEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_right.gad"},
	{0xBF3AF794214F3F2AULL, "6067874_m.fxe"},
	{0xBF3B286895339C34ULL, "6162810_m.fxe"},
	{0xBF3BCD103B7FA99AULL, "vxlt_gallowstemplar_6s2_0_1.vlm"},
	{0xBF3BD68F288AB3ECULL, "follower_barks6153621.cl"},
	{0xBF3C60DF00058A6EULL, "mag170_hugh6080723.cl"},
	{0xBF3CB941F439DF56ULL, "hlac_ht_keep_n_3733_0z.dds"},
	{0xBF3D01AAB6E1662DULL, "6131150_m.fxe"},
	{0xBF3D4AD18F4350B8ULL, "vxca_cavern1_haunted_i1f_0_0.vlm"},
	{0xBF3E1729A3D056C0ULL, "llac_w_encounter3_-4_0.dds"},
	{0xBF3E1B77DC94FC18ULL, "fen201ip_blood_sacrafice.utp"},
	{0xBF3E607162E255FCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sworda_sa.dds"},
	{0xBF3ED3058E35FF9DULL, "vxlt_hangedman_21g_0_0.vlm"},
	{0xBF3ED835984E0242ULL, "384826540.ogg"},
	{0xBF3F07E81B5288A1ULL, "hlac_lt_0_lowtown_-196608_3z.dds"},
	{0xBF3F22041612A5F0ULL, "llac_lt_gallowscourtyard_-1_2.dds"},
	{0xBF3F322F8B951F67ULL, "gen00fl_varric6033717.cl"},
	{0xBF3F43DA62364BD6ULL, "mer240cr_miner2.utc"},
	{0xBF3FB991D1CE6402ULL, "6118620_m.fxe"},
	{0xBF3FEDCE5514F7BCULL, "vxlt_darktown_6c1_0_0.vlm"},
	{0xBF4054546A7041B8ULL, "w_bonepitmining_a2_184_0.phy"},
	{0xBF4073D296C98BD2ULL, "6104878_m.fxe"},
	{0xBF40FFDFF198A6FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_enter.ani"},
	{0xBF4147DF6AF57DE7ULL, "512607141.ogg"},
	{0xBF414BA468518E09ULL, "1007452303.ogg"},
	{0xBF418D435B1D9E1DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_point_right.evt"},
	{0xBF4196CC5A70416BULL, "gen00fl_fenris6126899.cl"},
	{0xBF41A998DB65719EULL, "vxlt_darktown_5te_0_0.vlm"},
	{0xBF41BB707A80DEB9ULL, "gen00fl_aveline6137577.cl"},
	{0xBF425A6235302624ULL, "vxdr_primevalentrance_2yu_0_0.vlm"},
	{0xBF425B00B834E7FDULL, "stoney_d.dds"},
	{0xBF429CFFE97A7C5BULL, "dae150_chandan6148666.cl"},
	{0xBF42ADA8B150397FULL, "w_bonepit_29_0.msh"},
	{0xBF42B3A8B1504321ULL, "w_bonepit_29_0.mmh"},
	{0xBF42BFA8B15057C2ULL, "w_bonepit_29_0.mao"},
	{0xBF432CE6BEC50BDFULL, "vxlt_0_lowtown_a2_n_c_hud_0_0.vlm"},
	{0xBF44ABCA8DC0C8ABULL, "ctplt_dot.mao"},
	{0xBF44EDC044531434ULL, "6109219_m.fxe"},
	{0xBF450BA5201A00ADULL, "sst_ss_player_female.gda"},
	{0xBF45AE1C612DCA50ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_grand.mao"},
	{0xBF45B21C612DD163ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_grand.mmh"},
	{0xBF45C01C612DE8BDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_grand.msh"},
	{0xBF45EDE29823B9B1ULL, "vxdr_0_deeproads_cwe_0_0.vlm"},
	{0xBF462C66A2736BF1ULL, "vxw_woundedcoast_1af_0_0.vlm"},
	{0xBF4645BF49BD2732ULL, "6113298_m.fxe"},
	{0xBF468E1B408E7DB6ULL, "vxlt_darktown_5j7_0_0.vlm"},
	{0xBF46A5CD8088BCB7ULL, "hlac_lt_warehouse_-65536_0x.dds"},
	{0xBF46A6FD5FA04189ULL, "6041406_m.fxe"},
	{0xBF46BC126362946CULL, "w_woundedcoast_6_0_ll.mmh"},
	{0xBF47BD3374CD3745ULL, "6176003_m.fxe"},
	{0xBF47C578DA52255CULL, "22982911.ogg"},
	{0xBF47F6BD1BDD9C9FULL, "266642508.ogg"},
	{0xBF4860C505CDF00DULL, "6071154_f.fxe"},
	{0xBF487C8E9A714E5FULL, "vxlt_hangedman_n_20u_0_0.vlm"},
	{0xBF498DB799B0B1EDULL, "w_bonepitdestroyed_a3_241_0.msh"},
	{0xBF499BB799B0C920ULL, "w_bonepitdestroyed_a3_241_0.mao"},
	{0xBF499FB799B0D073ULL, "w_bonepitdestroyed_a3_241_0.mmh"},
	{0xBF49C2E7138F3389ULL, "w_woundedcoast_74_0_ll.mmh"},
	{0xBF4A1550354BC22AULL, "ef_progenitor_1.mor"},
	{0xBF4A5365165B8838ULL, "vxw_bonepitdestroyed_a3_1ba_0_0.vlm"},
	{0xBF4ACB088605F3B6ULL, "lowtown_argument6169372.cl"},
	{0xBF4B068E989EE1F7ULL, "mag321_comtesse_de_launcet6043020.cl"},
	{0xBF4B2066FA407078ULL, "6080167_m.fxe"},
	{0xBF4B66733B4AD3D4ULL, "hlac_lt_0_lowtown_docks_a3_131071_0x.dds"},
	{0xBF4B6794CBAAE7DBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww44_cellchunkxlo.msh"},
	{0xBF4B7994CBAB0626ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww44_cellchunkxlo.mao"},
	{0xBF4B8594CBAB1AC5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww44_cellchunkxlo.mmh"},
	{0xBF4B9C762E9F759AULL, "hlac_lt_undercityencounter_-1_39x.dds"},
	{0xBF4BE6D96105D9FDULL, "51683169.ogg"},
	{0xBF4C07C9E504599BULL, "vxht_mansion2b_1y0_0_0.vlm"},
	{0xBF4C08EB2A9E40EBULL, "llac_lt_0_lowtown_n_-65539_0.dds"},
	{0xBF4C14D56ECC760CULL, "671830904.ogg"},
	{0xBF4C1AC590949524ULL, "hla_ht_0_hightown_a3_131073_1y.dds"},
	{0xBF4C4DD0E7742085ULL, "6095461_m.fxe"},
	{0xBF4C76324093D285ULL, "486039609.ogg"},
	{0xBF4D1E3112A56438ULL, "lt_0_lowtown_docks_a3_2_0.rml"},
	{0xBF4D4C94BB1A1C13ULL, "6101560_m.fxe"},
	{0xBF4DC33F29A542EAULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.anb"},
	{0xBF4DDF019362E617ULL, "one351ip_worldmap.utp"},
	{0xBF4DF82818B25AE7ULL, "6142051_m.fxe"},
	{0xBF4E1A7916AAB0D1ULL, "vxlt_gallowstemplar_4sj_0_0.vlm"},
	{0xBF4E4367C42FE2F2ULL, "52691041.ogg"},
	{0xBF4F17DFF1A50CCFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_enter.evt"},
	{0xBF4F346ABDFF8041ULL, "vxht_mansion_n_6kl_0_0.vlm"},
	{0xBF4F9A245A7C750FULL, "gen_im_fla_varric_02.uti"},
	{0xBF4FB7090141005AULL, "dae000_gamlen6131077.cl"},
	{0xBF4FF9E0FD05B296ULL, "hlac_w_encounter3_-196604_0z.dds"},
	{0xBF4FFC2FB121C1ECULL, "w_encounter3_34_0_lc.mmh"},
	{0xBF507B835C54CE1DULL, "w_woundedcoast_23_0.rml"},
	{0xBF5092E6DFAECE6EULL, "zz_mag_debug6053415.cl"},
	{0xBF50966D46E6F2C9ULL, "vxlt_0_lowtown_docks_n_b6c_0_1.vlm"},
	{0xBF50B5162F65FC35ULL, "mag160_feynriel6033394.cl"},
	{0xBF51827D1420E7B1ULL, "121741081.ogg"},
	{0xBF519EC3B0BB245FULL, "vxlt_0_lowtown_a2_n_c_hpg_0_0.vlm"},
	{0xBF51B19086C3C4B3ULL, "gen00fl_aveline6140014.cl"},
	{0xBF51F6F702469D5DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrga.phy"},
	{0xBF520039D22B7389ULL, "gen00fl_merrill6109788.cl"},
	{0xBF53F778C6D8C343ULL, "6132842_f.fxe"},
	{0xBF53FB8E8135D4F6ULL, "w_blightlands_229_0.msh"},
	{0xBF54058E8135E5F4ULL, "w_blightlands_229_0.mmh"},
	{0xBF54098E8135EC07ULL, "w_blightlands_229_0.mao"},
	{0xBF54182EEEA747FAULL, "ico_crossbow_1.dds"},
	{0xBF542F6CBFD11B24ULL, "hlac_lt_0_lowtown_docks_a3_-65538_9x.dds"},
	{0xBF543B42F7A442AEULL, "hlac_lt_0_lowtown_docks_n_131068_0z.dds"},
	{0xBF5452F36463AB37ULL, "42714573.ogg"},
	{0xBF545C9040948933ULL, "778774379.ogg"},
	{0xBF54902976D30088ULL, "vxw_sundermount_2a8_0_0.vlm"},
	{0xBF549E4394698471ULL, "vxw_woundedcoast_1le_0_0.vlm"},
	{0xBF54E694E379FE25ULL, "art\\characters\\playercharacter\\elfmale\\em_rob_appd.phy"},
	{0xBF54FFF832849636ULL, "vxlt_darktown_5o8_0_0.vlm"},
	{0xBF55877078E64FAAULL, "6153258_m.fxe"},
	{0xBF5610174F2A45BFULL, "ser211ip_to_gamlens.utp"},
	{0xBF56143903CDD1F6ULL, "vxca_caverns1_gcu_0_0.vlm"},
	{0xBF5618DFF1AB2D78ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_enter.gad"},
	{0xBF5623C2B775CAC9ULL, "vxht_store_bz_0_0.vlm"},
	{0xBF5692B3BD78A76DULL, "gen00fl_merrill6129852.cl"},
	{0xBF56D07DAC7B94DEULL, "vxht_0_hightown_a2_be2_0_0.vlm"},
	{0xBF56EE08EAC4710CULL, "vxw_bonepit_1ob_0_0.vlm"},
	{0xBF572B45504239FDULL, "vxlt_gallowstemplar_5xj_0_0.vlm"},
	{0xBF5836C36742713FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\barkburn.mao"},
	{0xBF58EF5562615946ULL, "cd_mag200_feynriel.bnk"},
	{0xBF58F243550B7306ULL, "vxw_bonepitdestroyed_a3_1g0_0_0.vlm"},
	{0xBF5905A91895F1FCULL, "vxdr_0_deeproads_7tx_0_1.vlm"},
	{0xBF5955B4E11530D7ULL, "985130533.ogg"},
	{0xBF59B7D7999B8B5CULL, "vxlt_undercityencounter_112_0_0.vlm"},
	{0xBF59F481FD3B3F34ULL, "gen00fl_carver6180607.cl"},
	{0xBF5A0FC6CF98133DULL, "432254306.ogg"},
	{0xBF5A1634DDD0698EULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.dura.evt"},
	{0xBF5A3D170C3BEEC8ULL, "6049469_m.fxe"},
	{0xBF5A50EF055B1C8BULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_arm_cas_0d.dds"},
	{0xBF5A58B75DA71973ULL, "6068001_m.fxe"},
	{0xBF5A8D158EE74D1FULL, "373478395.ogg"},
	{0xBF5AA3BED612D4A4ULL, "6084224_m.fxe"},
	{0xBF5ACFD8D63492E7ULL, "vxht_mansion2_1u0_0_0.vlm"},
	{0xBF5AE3E1871F5A42ULL, "hlac_w_encounter3_-327684_0x.dds"},
	{0xBF5B34397C8F851EULL, "vxlt_gallowscourtyard_n_768_0_0.vlm"},
	{0xBF5B625F97E1009CULL, "vxlt_0_lowtown_a3_n_c_hpy_0_0.vlm"},
	{0xBF5B99B986363904ULL, "art\\characters\\playercharacter\\humanmale\\boo_medj.phy"},
	{0xBF5BF5BD1E7C4F15ULL, "llac_lt_darktown_-1_87.dds"},
	{0xBF5C19841CBB6804ULL, "nat380_sibling6103564.cl"},
	{0xBF5C67BFCDC1AD29ULL, "gen00fl_varric6105429.cl"},
	{0xBF5CC9A4A85371BAULL, "hlac_lt_gallowsprison_n_-1_0y.dds"},
	{0xBF5DA647AFF598C1ULL, "vxht_mansion_n_6nc_0_0.vlm"},
	{0xBF5DE4A8FE4AE2ACULL, "hlac_lt_0_lowtown_a2_n_1_11y.dds"},
	{0xBF5E4405AF061201ULL, "110943423.ogg"},
	{0xBF5EA0C4F6787DEEULL, "vxlt_warehouse_n_cq_0_0.vlm"},
	{0xBF5EB74BDA84817BULL, "vxht_0_hightown_n_c4v_0_0.vlm"},
	{0xBF5EDE5713F84A9EULL, "follower_banter6156212.cl"},
	{0xBF5F4F54A7B069A8ULL, "6116750_m.fxe"},
	{0xBF5FC62B51325E03ULL, "467735520.ogg"},
	{0xBF5FF35CF4372D3FULL, "lla_ht_0_hightown_-65537_0.dds"},
	{0xBF60449612484D9DULL, "6201835_m.fxe"},
	{0xBF60952C189C8EECULL, "vxlt_0_lowtown_byh_0_0.vlm"},
	{0xBF60C2E0820BB5E6ULL, "912044617.ogg"},
	{0xBF60EF04E7555964ULL, "6117291_f.fxe"},
	{0xBF610D34DDD67939ULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.dura.gad"},
	{0xBF6167E8B74DF4E1ULL, "gen00fl_anders6125551.cl"},
	{0xBF61D0F4D3AC6BCAULL, "199632147.ogg"},
	{0xBF626047D05F36D7ULL, "6119756_m.fxe"},
	{0xBF62712146DE6DC4ULL, "6058962_m.fxe"},
	{0xBF6297C37BF98EBDULL, "331295060.ogg"},
	{0xBF62D68A7F41178BULL, "hlac_dr_primevaltomb_658_0z.dds"},
	{0xBF62EBF70254FDC9ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrga.mao"},
	{0xBF62F7F702551262ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrga.mmh"},
	{0xBF62F9F702551518ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintinglrga.msh"},
	{0xBF635EEED5711DFCULL, "llac_lt_0_lowtown_a2_n_c_-65536_0.dds"},
	{0xBF6367FDA3AC3D1CULL, "vxca_cavern1_haunted_e52_0_0.vlm"},
	{0xBF6371D5247AAEB6ULL, "vxlt_darktown_5p9_0_0.vlm"},
	{0xBF6383164589B639ULL, "6081675_m.fxe"},
	{0xBF6390FC2553B22EULL, "combat_deathblow_03.gda"},
	{0xBF63A4D183A96B80ULL, "al_carver_talents.gda"},
	{0xBF63F94D6ADC682AULL, "6156158_m.fxe"},
	{0xBF6453F3517EB30EULL, "w_blightlands_124_0.mmh"},
	{0xBF645FF3517EC7A5ULL, "w_blightlands_124_0.mao"},
	{0xBF646DF3517EDFFCULL, "w_blightlands_124_0.msh"},
	{0xBF64CEF0D3844647ULL, "hlac_lt_gallowstemplar_n_2390_0z.dds"},
	{0xBF65150C778085B8ULL, "vxdr_primevaltomb_8nb_0_0.vlm"},
	{0xBF652AB0FC992360ULL, "vxht_0_hightown_c72_0_1.vlm"},
	{0xBF6593861EA4571EULL, "1019428601.ogg"},
	{0xBF665A0D8AB7EA9FULL, "166623933.ogg"},
	{0xBF667613512B447AULL, "1024745542.ogg"},
	{0xBF66B9E637E019F5ULL, "follower_banter6155670.cl"},
	{0xBF6775B843AA7A5EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_treebark_d.dds"},
	{0xBF67DA5CF18CA72DULL, "vxht_0_hightown_a2_c7v_0_0.vlm"},
	{0xBF67DACA2B8D3FC3ULL, "vxdr_primevaltomb_7bz_0_0.vlm"},
	{0xBF67F8E501B5F8FCULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_wng_hdra_0d.dds"},
	{0xBF680F124EDC4F08ULL, "vxht_chantry_2uq_0_1.vlm"},
	{0xBF681E34DDDCB487ULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.dura.ani"},
	{0xBF68412272FBD892ULL, "fen210_orana6109699.cl"},
	{0xBF6893A8B170DC4AULL, "w_bonepit_29_0.rml"},
	{0xBF69D7068E69B6DBULL, "632448114.ogg"},
	{0xBF69EF25E2E9319CULL, "dae000_mother6035929.cl"},
	{0xBF6A2F687C834423ULL, "dae000_arishok6067725.cl"},
	{0xBF6AC272A5423EC9ULL, "t1_mue_gr3.tnt"},
	{0xBF6AC96EEB57EEEDULL, "590093032.ogg"},
	{0xBF6B0C37E2FD7F47ULL, "fen210_orana6109708.cl"},
	{0xBF6B3413B3916541ULL, "136509829.ogg"},
	{0xBF6C2BD62AF85FD7ULL, "follower_banter6169741.cl"},
	{0xBF6C7387E6C206C3ULL, "w_bonepitdestroyed_a3_60_0.mao"},
	{0xBF6C7787E6C20D10ULL, "w_bonepitdestroyed_a3_60_0.mmh"},
	{0xBF6C8587E6C2252AULL, "w_bonepitdestroyed_a3_60_0.msh"},
	{0xBF6CD30B3C6AAD32ULL, "vxlt_warehouse_n_12m_0_0.vlm"},
	{0xBF6D05EFEE39AD5CULL, "w_woundedcoast.anb"},
	{0xBF6D11EFEE39C1CEULL, "w_woundedcoast.arl"},
	{0xBF6D1AC343F40204ULL, "6205603_m.fxe"},
	{0xBF6D2928CC7A99FBULL, "vxht_0_hightown_n_c1i_0_0.vlm"},
	{0xBF6E39E9DBBCA78DULL, "hlac_w_encounter2_camp_131076_0y.dds"},
	{0xBF6E417B84EB2FC8ULL, "zz_dae_debug6029523.cl"},
	{0xBF6E958EC656AAD4ULL, "vxlt_0_lowtown_docks_a3_n_cvm_0_0.vlm"},
	{0xBF6E99577DAEE014ULL, "6122534_f.fxe"},
	{0xBF6EB95EF6CB7BACULL, "vxlt_warehouse_n_15z_0_0.vlm"},
	{0xBF6EBB24566D17EFULL, "vxht_mansion2b_3g8_0_0.vlm"},
	{0xBF6ECE64C67678C5ULL, "lt_0_lowtown_38_0.rml"},
	{0xBF6F60C3E72D0F0CULL, "6067195_m.fxe"},
	{0xBF6FA50A547A796BULL, "vxht_mansion2b_1dk_0_0.vlm"},
	{0xBF6FC8CE61D17576ULL, "6075774_m.fxe"},
	{0xBF700A33618C4705ULL, "533637204.ogg"},
	{0xBF709D197FFACA80ULL, "vxlt_0_lowtown_docks_a3_n_br8_0_0.vlm"},
	{0xBF70AE59BDADBF35ULL, "combat_2h_r1_a3.gda"},
	{0xBF70FFAB02839B7DULL, "vxht_0_hightown_a2_4as_0_0.vlm"},
	{0xBF7117EA07E7A318ULL, "6147756_m.fxe"},
	{0xBF7143796E060331ULL, "vxdr_0_deeproads_cfz_0_0.vlm"},
	{0xBF7177123B950F04ULL, "dae000_light_tombs_boss6179468.cl"},
	{0xBF7177975EA5A5DDULL, "dre000_bodahn6031285.cl"},
	{0xBF71ACDA77BABA58ULL, "vxlt_0_lowtown_a3_n_c_eod_0_1.vlm"},
	{0xBF729EC5252AC2B3ULL, "180863117.ogg"},
	{0xBF736ED9E1E51BDBULL, "sib111_mother6098816.cl"},
	{0xBF742C51D7AA8D39ULL, "194689356.ogg"},
	{0xBF7430578BFD6BB1ULL, "vxdr_0_deeproads_fa2_0_0.vlm"},
	{0xBF74408C984EEF29ULL, "gen_im_trs_gen_statue_04.uti"},
	{0xBF74E8E4C3457D4DULL, "vxlt_0_lowtown_a3_n_c_e8a_0_0.vlm"},
	{0xBF7558A76D4CA8ECULL, "193553338.ogg"},
	{0xBF75BA4C21C5A56DULL, "lt_gallowscourtyard_n_36_0.rml"},
	{0xBF75D5FD38F7A406ULL, "vxw_bonepitdestroyed_a3_1m2_0_0.vlm"},
	{0xBF764C39E382BFADULL, "vxht_0_hightown_a2_c2u_0_0.vlm"},
	{0xBF76594AAFA280D0ULL, "llac_lt_0_lowtown_docks_n_-65538_9.dds"},
	{0xBF7699BF2C96E8CBULL, "dae300_alistair_king6133809.cl"},
	{0xBF76C63F29C84CDDULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s3_c.mmh"},
	{0xBF76DE367C3B7785ULL, "gen_im_cns_cft_wrn_05.uti"},
	{0xBF773CB05444150EULL, "vxlt_0_lowtown_eki_0_2.vlm"},
	{0xBF774172F14785E2ULL, "genip_book_2.utp"},
	{0xBF77441DB91DCEB8ULL, "vxlt_0_lowtown_docks_b9q_0_0.vlm"},
	{0xBF77605C3E8CBEE2ULL, "follower_barks6173399.cl"},
	{0xBF77B392BA20C3E7ULL, "vxht_mansion2_1c2_0_0.vlm"},
	{0xBF78FF3112CA6AF0ULL, "lt_0_lowtown_docks_a3_2_0.anb"},
	{0xBF7912F6041CD226ULL, "vfx_bird_raven03.dds"},
	{0xBF7921BD3D1C18E4ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_1y.dds"},
	{0xBF79CD6A38CFCCC6ULL, "404985745.ogg"},
	{0xBF7A2D8BCC8A21C7ULL, "6175071_m.fxe"},
	{0xBF7A6B17E0457186ULL, "552493370.ogg"},
	{0xBF7A6F447DB88B1FULL, "vxlt_0_lowtown_a2_h2h_0_0.vlm"},
	{0xBF7A7D15902BED13ULL, "809290027.ogg"},
	{0xBF7AB586B5307F22ULL, "vxw_sundermount_1zt_0_0.vlm"},
	{0xBF7AE91956B500EAULL, "t3_wdg_w03.tnt"},
	{0xBF7B7866BF97E9F4ULL, "621358584.ogg"},
	{0xBF7BB3EB78DF68FFULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardleft.gad"},
	{0xBF7BCE320B92AFB6ULL, "146095094.ogg"},
	{0xBF7C03F073C4464EULL, "hlac_w_encounter3_-327683_0y.dds"},
	{0xBF7C0E44FCF6CD20ULL, "vxw_woundedcoast_1gi_0_0.vlm"},
	{0xBF7C21E741158C81ULL, "vxlt_undercityencounter_11c_0_0.vlm"},
	{0xBF7C649EC06C2998ULL, "vxw_encounter2_camp_8t_0_0.vlm"},
	{0xBF7CF147B65F2989ULL, "vxlt_gamlenshouse_n_n3_0_0.vlm"},
	{0xBF7D537ED6E0F7CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_weightshift_small.gad"},
	{0xBF7D70BCE099C5B7ULL, "vxht_0_hightown_a3_a2p_0_0.vlm"},
	{0xBF7DAC98B757213AULL, "vxw_woundedcoast_1dz_0_0.vlm"},
	{0xBF7DCCCDD6645824ULL, "447438343.ogg"},
	{0xBF7DCF1CBA760833ULL, "vxca_caverns1_mines_gkp_0_1.vlm"},
	{0xBF7E48A5F62B988FULL, "6132791_m.fxe"},
	{0xBF7E786C8770D109ULL, "t3_arm_irn.tnt"},
	{0xBF7ECC4B5586BD7BULL, "426695649.ogg"},
	{0xBF7EF50757883451ULL, "6162297_m.fxe"},
	{0xBF7FE29B143EA327ULL, "146760582.ogg"},
	{0xBF7FE5C04CB54A08ULL, "vxw_sundermount_2p7_0_0.vlm"},
	{0xBF803B400CF19E9BULL, "6166911_f.fxe"},
	{0xBF8057A7A4730731ULL, "ico_locked_root.dds"},
	{0xBF81A8EB78E3C1ECULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardleft.evt"},
	{0xBF81B3DCB008CB0EULL, "dr_dungeon_3207_ll.mmh"},
	{0xBF81DF86A3810D0BULL, "vxlt_warehouse_n_8m_0_0.vlm"},
	{0xBF821EEE31244CEAULL, "vrc221_varric6044178.cl"},
	{0xBF827961796E8EEEULL, "art\\levels\\4cinematics\\ferelden\\_textures\\w_small_dagger_0_sa.dds"},
	{0xBF82B6A19F4B42CEULL, "613315397.ogg"},
	{0xBF82CE0C23DD9A39ULL, "99961410.ogg"},
	{0xBF82FC5AAD47947CULL, "vxca_cavern1_haunted_e96_0_0.vlm"},
	{0xBF8384ADB47344C6ULL, "vxw_sundermount_301_0_1.vlm"},
	{0xBF83C7197B6B40CEULL, "art\\vfx\\combat\\vfx_ha_bd_slam_act_c.impa.evt"},
	{0xBF84367ED6E6E561ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_weightshift_small.evt"},
	{0xBF84524F7EC83E9FULL, "ht_brothel_n_3371_ll.mmh"},
	{0xBF847BF435060C87ULL, "gen_im_wep_mag_sta_act1_04.uti"},
	{0xBF84B63688400A61ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.down.ani"},
	{0xBF84BE16D578D82DULL, "vxht_0_hightown_a2_c9t_0_0.vlm"},
	{0xBF84E1264C7B9CCAULL, "hlac_w_bonepitdestroyed_a3_0_6x.dds"},
	{0xBF853AD8FB4C483DULL, "vxht_bartrand_n_6qe_0_0.vlm"},
	{0xBF8553EFD6F9294EULL, "vxht_bartrand_n_68m_0_0.vlm"},
	{0xBF85AE6F31BB571CULL, "llac_lt_0_lowtown_docks_n_-131076_0.dds"},
	{0xBF864365BDECD457ULL, "878531711.ogg"},
	{0xBF8650864786ED20ULL, "clf260_fell_orden6070485.cl"},
	{0xBF86A4355E613459ULL, "961822393.ogg"},
	{0xBF872B7A87EC39CDULL, "art\\levels\\4cinematics\\qunari\\prp_qunaricontro.msh"},
	{0xBF87397A87EC5100ULL, "art\\levels\\4cinematics\\qunari\\prp_qunaricontro.mao"},
	{0xBF873D7A87EC5853ULL, "art\\levels\\4cinematics\\qunari\\prp_qunaricontro.mmh"},
	{0xBF878A63AF145D5AULL, "isa321_velasco6117631.cl"},
	{0xBF878AFBC1ADF363ULL, "hlac_lt_undercityencounter_-1_50x.dds"},
	{0xBF8793E5AE2A1434ULL, "lt_0_lowtown_docks_5_0.rml"},
	{0xBF881397FA33DBA4ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb103a_0s.dds"},
	{0xBF884D89547D74EFULL, "vxht_0_hightown_n_9zr_0_0.vlm"},
	{0xBF88E380CF7294B5ULL, "6089507_m.fxe"},
	{0xBF88FC14DC560A7FULL, "dae300_alistair_king6133603.cl"},
	{0xBF8961724F3DAFD8ULL, "hlac_lt_blackemporium_1055_11x.dds"},
	{0xBF8995E4071555B4ULL, "vxlt_0_lowtown_a2_n_c_epd_0_1.vlm"},
	{0xBF8A39008FA680F0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_diagonal.mao"},
	{0xBF8A3D008FA68703ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_diagonal.mmh"},
	{0xBF8A4B008FA69F5DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_diagonal.msh"},
	{0xBF8A7506A097C6DDULL, "6056936_m.fxe"},
	{0xBF8ABE197B715079ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_act_c.impa.gad"},
	{0xBF8AD67BB2624218ULL, "vxw_encounter2_camp_3u_0_0.vlm"},
	{0xBF8AE992FBE516F6ULL, "vxlt_darktown_2qy_0_0.vlm"},
	{0xBF8B24EACDC6BDB2ULL, "dae000_madam_lusine6151478.cl"},
	{0xBF8B2832D7408872ULL, "64472094.ogg"},
	{0xBF8B4CD01D10B13FULL, "vxlt_0_lowtown_byg_0_0.vlm"},
	{0xBF8B97129E60D95AULL, "hla_ht_0_hightown_a2_-65534_0x.dds"},
	{0xBF8BC13688463B03ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.down.gad"},
	{0xBF8C3DC76AAC7A5DULL, "121738584.ogg"},
	{0xBF8C88C43866AA6BULL, "vxht_mansion2b_1rm_0_0.vlm"},
	{0xBF8CFC92AB304698ULL, "hlac_lt_gallowstemplar_f_2388_0x.dds"},
	{0xBF8D037F6BFC5434ULL, "vxca_caverns1_mines_ih1_0_0.vlm"},
	{0xBF8D9A34F2735F5EULL, "vxlt_0_lowtown_a2_he4_0_0.vlm"},
	{0xBF8DC73D21F23F15ULL, "vxdr_primevaltomb_4xp_0_1.vlm"},
	{0xBF8DE61773E8CA5BULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_ferelden_jimmy_t.dds"},
	{0xBF8DF62207557C30ULL, "mcr371cr_isabela_cut.utc"},
	{0xBF8DFB13445CFCF3ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_maj_c.anb"},
	{0xBF8E1742FE24036DULL, "vxlt_0_lowtown_eqh_0_2.vlm"},
	{0xBF8E546FCAEB212BULL, "vxca_cavern1_haunted_ei4_0_1.vlm"},
	{0xBF8E579D3EAB6288ULL, "vxw_sundermount_2u4_0_0.vlm"},
	{0xBF8F4804C260CA39ULL, "141385623.ogg"},
	{0xBF8F504BEAE64772ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_bloodsplater_da.dds"},
	{0xBF8F90EB78EFD645ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforwardleft.ani"},
	{0xBF8FF95A22B804FCULL, "lgt210pt_ingredients.ncs"},
	{0xBF9020F1B748BD0CULL, "312082790.ogg"},
	{0xBF902D0C5D35E414ULL, "hlac_lt_0_lowtown_a2_65537_7z.dds"},
	{0xBF90CF197B75D8C7ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_act_c.impa.ani"},
	{0xBF9145E63FBD3A61ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_qunari2_arishokresponse.gad"},
	{0xBF9156C9B879CDA4ULL, "docks_male_elven_worker_sta6175817.cl"},
	{0xBF9224BAE9600914ULL, "550776688.ogg"},
	{0xBF922E7ED6F314B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_weightshift_small.ani"},
	{0xBF923F646E81C36FULL, "hlac_ht_0_hightown_n_-2_0x.dds"},
	{0xBF925872102CE7F8ULL, "llac_lt_gallowscourtyard_n_-131074_0.dds"},
	{0xBF9261C017F5B2ADULL, "6021220_m.fxe"},
	{0xBF92B48962D0ADE6ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rblt.ani"},
	{0xBF93545897F3B8C2ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.dura.gad"},
	{0xBF9371D61F085BB9ULL, "vxw_woundedcoast_16o_0_0.vlm"},
	{0xBF93E8D50B627B34ULL, "698024739.ogg"},
	{0xBF94A775126557CDULL, "6001190_m.fxe"},
	{0xBF94FC7A111D20E1ULL, "lt_0_lowtown_docks_a3_n_4_0.rml"},
	{0xBF95A73CBF120B56ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgwht.phy"},
	{0xBF95AA1702E9C809ULL, "6129379_m.fxe"},
	{0xBF95E9729594D9A9ULL, "6066188_m.fxe"},
	{0xBF9661BB728B941CULL, "452182699.ogg"},
	{0xBF9663BBEAB57836ULL, "qun100_mage_party_comment6019104.cl"},
	{0xBF96741D5425372EULL, "hlac_w_woundedcoast_196610_0x.dds"},
	{0xBF973EB5B6246A83ULL, "hlac_ca_cavern1_haunted_24426_8y.dds"},
	{0xBF9746E63FC1A70FULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_qunari2_arishokresponse.ani"},
	{0xBF97E1805ED9ED19ULL, "431698622.ogg"},
	{0xBF9801122F0DFB0FULL, "6060161_f.fxe"},
	{0xBF981E14179933A0ULL, "vxlt_undercitydungeon_y2_0_0.vlm"},
	{0xBF983B6F781DCA4CULL, "art\\levels\\4cinematics\\ferelden\\_textures\\w_small_dagger_0_t.dds"},
	{0xBF9853685DC8CC9FULL, "6169360_m.fxe"},
	{0xBF986838BE50DFEEULL, "vxlt_warehouse_n_wu_0_0.vlm"},
	{0xBF989E179D894278ULL, "66168612.ogg"},
	{0xBF98AD1D2F0B6CA0ULL, "6121046_m.fxe"},
	{0xBF99EB259C38F454ULL, "vxlt_0_lowtown_docks_a3_n_cyp_0_0.vlm"},
	{0xBF99F1556298F3ECULL, "cd_mag200_feynriel.wbd"},
	{0xBF9A564372C99C33ULL, "vxht_bartrand_n_64x_0_0.vlm"},
	{0xBF9A76E12FEBA253ULL, "lgt110_gxa_coa_spy6117447.cl"},
	{0xBF9A775897FA12B9ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.dura.evt"},
	{0xBF9A90529B43523AULL, "vxw_woundedcoast_1jt_0_0.vlm"},
	{0xBF9A9EC5795F5303ULL, "6113220_m.fxe"},
	{0xBF9AB212E69B9D12ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ff_basic_1.mao"},
	{0xBF9AC5F67F13595FULL, "vxlt_ruinedhouse_n_i4_0_0.vlm"},
	{0xBF9AF3790A8B54CAULL, "llac_lt_0_lowtown_n_-65535_1.dds"},
	{0xBF9C0C11E46977DEULL, "vxlt_0_lowtown_a2_hb5_0_0.vlm"},
	{0xBF9C181344696B5CULL, "art\\vfx\\combat\\vfx_glb_nat_imp_maj_c.mmh"},
	{0xBF9C3D0E31FED313ULL, "6195873_m.fxe"},
	{0xBF9C41546ABE804DULL, "w_bonepitmining_a2_184_0.msh"},
	{0xBF9C4F546ABE9780ULL, "w_bonepitmining_a2_184_0.mao"},
	{0xBF9C53546ABE9ED3ULL, "w_bonepitmining_a2_184_0.mmh"},
	{0xBF9CEB5934F9F0CDULL, "6077458_m.fxe"},
	{0xBF9CF4BA8422B278ULL, "763919281.ogg"},
	{0xBF9D11CEE158CB8BULL, "vxca_caverns2_8ic_0_1.vlm"},
	{0xBF9D40AE7848AD33ULL, "6152173_m.fxe"},
	{0xBF9D741F0C214CE0ULL, "darktown_fem_lowlife_mob6176168.cl"},
	{0xBF9EC30447F8621FULL, "6041981_m.fxe"},
	{0xBF9EC5020798FA41ULL, "gen00fl_isabela6075278.cl"},
	{0xBF9EF004AC61745CULL, "fx_ice_splat1.mao"},
	{0xBF9F09818C12BA06ULL, "vxdr_primevaltomb_8z0_0_0.vlm"},
	{0xBFA0209F1B67D46AULL, "mcr371_hostile_anders6082129.cl"},
	{0xBFA08E3BD383F12CULL, "vxlt_gallowstemplar_n_6s3_0_1.vlm"},
	{0xBFA0903CBF1BEBF7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgwht.msh"},
	{0xBFA096D9C42F3C16ULL, "t3_shd_com.tnt"},
	{0xBFA0A23CBF1C0A5AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgwht.mao"},
	{0xBFA0A63CBF1C1169ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelgwht.mmh"},
	{0xBFA0C8B3EBF76FC4ULL, "llac_lt_gallowstemplar_2388_3.dds"},
	{0xBFA11A42C5C90307ULL, "hlac_lt_0_lowtown_a3_n_c_-1_0z.dds"},
	{0xBFA1CD64C6A1D0EDULL, "lt_0_lowtown_38_0.anb"},
	{0xBFA1E3B310FE7439ULL, "vxw_woundedcoast_11n_0_0.vlm"},
	{0xBFA208316A6C1DB4ULL, "6044615_m.fxe"},
	{0xBFA3722C42DBB092ULL, "zz_lgt_2_debug6174313.cl"},
	{0xBFA42072523DDCB7ULL, "302717964.ogg"},
	{0xBFA4332F49E76C93ULL, "6069867_f.fxe"},
	{0xBFA536A9413A1B07ULL, "6149054_m.fxe"},
	{0xBFA59FC4F50A204FULL, "hlac_ca_caverns2_5802_20z.dds"},
	{0xBFA5C73AB35EC5B5ULL, "6087707_m.fxe"},
	{0xBFA6E8D7887D1FC6ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.anb"},
	{0xBFA77F589804AAA8ULL, "art\\vfx\\combat\\vfx_m_bld_wound_lp_c.dura.ani"},
	{0xBFA79D8962E2E2E0ULL, "art\\vfx\\combat\\vfx_gib_hm_rob_rblt.gad"},
	{0xBFA7A1EF76613070ULL, "6151456_m.fxe"},
	{0xBFA7BD3040651CC3ULL, "hlac_lt_0_lowtown_65537_0z.dds"},
	{0xBFA80566317EAC0EULL, "hlac_ht_0_hightown_a3_-131074_1x.dds"},
	{0xBFA868E153F70A02ULL, "74339336.ogg"},
	{0xBFA889635CDFAD78ULL, "397763320.ogg"},
	{0xBFA8C8D33C471042ULL, "isa211_isabela6072606.cl"},
	{0xBFA903ACE50244F0ULL, "follower_banter6151369.cl"},
	{0xBFA9F5823E80931CULL, "531467363.ogg"},
	{0xBFAACC0210AE7EC7ULL, "520407034.ogg"},
	{0xBFAB0A6A7ED90F7BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masa_0s.dds"},
	{0xBFAB3601DC74D037ULL, "hlac_lt_0_lowtown_a3_n_c_1_4y.dds"},
	{0xBFABDDA93C87E8A8ULL, "1023458333.ogg"},
	{0xBFABEE4632825133ULL, "gen00fl_isabela6110300.cl"},
	{0xBFAC4FA72EB37FB6ULL, "255731120.ogg"},
	{0xBFAC520CB8A9836EULL, "22651040.ogg"},
	{0xBFAC6E3D5A97C648ULL, "zz_approval_debug6003463.cl"},
	{0xBFAC77835CA3263DULL, "w_woundedcoast_23_0.mao"},
	{0xBFAC7B835CA32D76ULL, "w_woundedcoast_23_0.mmh"},
	{0xBFAC85835CA33E74ULL, "w_woundedcoast_23_0.msh"},
	{0xBFAC9DAE9649D4D0ULL, "vxlt_gallowsprison_n_6na_0_0.vlm"},
	{0xBFAD04EFEE6F8FECULL, "w_woundedcoast.rml"},
	{0xBFAD7B5E7E08D286ULL, "vxdr_primevaltomb_8u1_0_0.vlm"},
	{0xBFADC7CB8B277A5FULL, "6111190_m.fxe"},
	{0xBFAE55B6D575872AULL, "912097510.ogg"},
	{0xBFAE88EA71D81520ULL, "llac_ht_brothel_n_3373_1.dds"},
	{0xBFAF60648A8556E7ULL, "hlac_ht_keep_n_3732_13x.dds"},
	{0xBFAF6CEE96676C60ULL, "lgt260_way_blood_mage6148394.cl"},
	{0xBFAF82A21BFF27EFULL, "25544614.ogg"},
	{0xBFAF910AEFC89296ULL, "6152330_m.fxe"},
	{0xBFAF9AB6B1862164ULL, "vxht_chantry_n_2ld_0_0.vlm"},
	{0xBFAFCDDDE4A81B0FULL, "926100247.ogg"},
	{0xBFAFE5A7157E7D67ULL, "mag330_walter6045568.cl"},
	{0xBFB04BDEBD8C117AULL, "art\\characters\\creatures\\gibgolem\\uni_gib.phy"},
	{0xBFB08E5AE3B1F586ULL, "6208555_f.fxe"},
	{0xBFB0B76BA8A88DDEULL, "gen00fl_varric6118772.cl"},
	{0xBFB2149F53E8A694ULL, "6053934_m.fxe"},
	{0xBFB3EBD78887AF71ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.mmh"},
	{0xBFB4775231184A14ULL, "vxca_caverns2prc_9r8_0_0.vlm"},
	{0xBFB48F1FEA0B07D6ULL, "788894479.ogg"},
	{0xBFB4B5711A50DA0EULL, "primitives_primitive_p6.vsh.fxo.dx11.assembly"},
	{0xBFB4D7D6C6125180ULL, "gen00fl_mrlact2_siblingtalk.stg"},
	{0xBFB544C183EFEFAAULL, "vxlt_warehouse_92_0_0.vlm"},
	{0xBFB5C443B426CAA7ULL, "6057739_m.fxe"},
	{0xBFB5F0A4DF378B93ULL, "vxht_bartrand_n_5v1_0_0.vlm"},
	{0xBFB5F19885F8AF33ULL, "6083143_m.fxe"},
	{0xBFB5F3EA78D9FE04ULL, "gen00_fl_merrill_houseact3.stg"},
	{0xBFB6565E07A1BD43ULL, "hlac_w_blightlands_0_7y.dds"},
	{0xBFB687130D545D52ULL, "6042039_m.fxe"},
	{0xBFB72D0427423EE9ULL, "vxlt_warehouse_n_t3_0_0.vlm"},
	{0xBFB766A2553972CCULL, "342755716.ogg"},
	{0xBFB7740C7F2F833AULL, "vxw_woundedcoast_1pv_0_0.vlm"},
	{0xBFB79AC950898101ULL, "gen00fl_anders6128321.cl"},
	{0xBFB7BA36291DF6EAULL, "nat380_sibling6207570.cl"},
	{0xBFB8098D7B0D53F6ULL, "800798891.ogg"},
	{0xBFB8662D544AEF19ULL, "459569742.ogg"},
	{0xBFB904BD206BD15FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofa.msh"},
	{0xBFB90ABD206BDB01ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofa.mmh"},
	{0xBFB916BD206BEFA2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofa.mao"},
	{0xBFB9654C6D2FB3C9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_twitch.ani"},
	{0xBFBA5CC9D94FDE2FULL, "862283725.ogg"},
	{0xBFBB6E1EF4911FC0ULL, "fen210_fenris6109433.cl"},
	{0xBFBB8578DC3C38B2ULL, "vxht_0_hightown_a2_9zl_0_0.vlm"},
	{0xBFBC1E5B6B7EA87EULL, "hlac_lt_0_lowtown_docks_n_-65538_9x.dds"},
	{0xBFBC91A897967634ULL, "vxlt_0_lowtown_a2_n_e83_0_0.vlm"},
	{0xBFBCC5893338A361ULL, "1004354754.ogg"},
	{0xBFBCF649CCAD8AB4ULL, "6034692_m.fxe"},
	{0xBFBD8A221FB778EEULL, "llac_ht_mansion2_818_33.dds"},
	{0xBFBDB41B4130A1A0ULL, "6134978_m.fxe"},
	{0xBFBDB6E5C8F5FA1CULL, "279953690.ogg"},
	{0xBFBE201A5980EAC0ULL, "vxht_keep_69t_0_0.vlm"},
	{0xBFBEE0413D598F32ULL, "ave000pt_gift_shield.ncs"},
	{0xBFBF24A02763DBADULL, "w_bonepit_2_0_ll.mmh"},
	{0xBFBF804C6D344D4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_twitch.gad"},
	{0xBFBFE0E2E46CEA17ULL, "6117726_f.fxe"},
	{0xBFC01AA7AADEAA0AULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_runf.gad"},
	{0xBFC0474C4D5DFA55ULL, "zz_followers_banter6157623.cl"},
	{0xBFC0600B5F166937ULL, "358066159.ogg"},
	{0xBFC09372E7F37309ULL, "vxw_bonepitmining_a2_1ug_0_0.vlm"},
	{0xBFC0F11E662A23A0ULL, "sib110_carver_tobrius.utc"},
	{0xBFC12D1BA814DE3DULL, "art\\characters\\weapons\\zzdao\\textures\\w_lsw_ls01a_0n.dds"},
	{0xBFC130B0F734029FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basic.msh"},
	{0xBFC136B0F7340D41ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basic.mmh"},
	{0xBFC142B0F73421E2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_basic.mao"},
	{0xBFC146C8F09EE782ULL, "55490357.ogg"},
	{0xBFC14AA1C1281C24ULL, "6149400_m.fxe"},
	{0xBFC1815311C8E0AFULL, "vxht_mansion2b_21g_0_0.vlm"},
	{0xBFC21855902D00B9ULL, "176708491.ogg"},
	{0xBFC23064C0D778F1ULL, "gen00fl_aveline6139962.cl"},
	{0xBFC25920B29D25ABULL, "ogre80.mao"},
	{0xBFC275B87079B0C0ULL, "6169583_m.fxe"},
	{0xBFC2A23A43B55BC8ULL, "vxlt_0_lowtown_docks_a3_aoc_0_0.vlm"},
	{0xBFC2E92F230E6294ULL, "vxca_caverns2prc_9y9_0_0.vlm"},
	{0xBFC3DE1515DD4DB4ULL, "vxca_caverns1_mines_j5m_0_0.vlm"},
	{0xBFC3DE794FC964EEULL, "6005030_m.fxe"},
	{0xBFC3FBDCBFCF6A45ULL, "335203096.ogg"},
	{0xBFC41DE67DC47DCBULL, "3835332.ogg"},
	{0xBFC441BC6845C61EULL, "mcr310_argument6072698.cl"},
	{0xBFC4493185527BEFULL, "937658649.ogg"},
	{0xBFC46281D12DA413ULL, "vxht_bartrand_n_5y0_0_0.vlm"},
	{0xBFC49A609E34FC56ULL, "mer200_ambush_leader6025028.cl"},
	{0xBFC4D4300118A0EFULL, "art\\vfx\\combat\\vfx_r_arc_front_max5_c.impa.gad"},
	{0xBFC538E3C176AFABULL, "6111132_m.fxe"},
	{0xBFC67D4C6D3A66B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_twitch.evt"},
	{0xBFC6AEB3582B4A8FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.mmh"},
	{0xBFC6FDA7AAE497A1ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_runf.evt"},
	{0xBFC749C601CE72A8ULL, "vxlt_0_lowtown_a2_n_gcl_0_0.vlm"},
	{0xBFC7CAE98CEE5770ULL, "966997921.ogg"},
	{0xBFC85EB2DBFC2BE4ULL, "ts_profane_prog_junk2.gda"},
	{0xBFC8E7562F8C3163ULL, "mrl351_phantomelves.bnk"},
	{0xBFC93499872C6720ULL, "407780778.ogg"},
	{0xBFC9C5434B4CB7D7ULL, "vxw_bonepitmining_a2_4a_0_0.vlm"},
	{0xBFC9EEC7F042BD91ULL, "602047066.ogg"},
	{0xBFCA1824779F3302ULL, "hlac_lt_0_lowtown_65533_3z.dds"},
	{0xBFCA352549B6C600ULL, "vxlt_0_lowtown_docks_a3_n_aqj_0_0.vlm"},
	{0xBFCA3D621E09DC63ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_0z.dds"},
	{0xBFCAA9C44CB429B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\ready.gad"},
	{0xBFCB31BF81EB975EULL, "hlac_w_bonepitmining_a2_6_0x.dds"},
	{0xBFCB55E833B03703ULL, "6035927_f.fxe"},
	{0xBFCB83650F57323CULL, "hlac_w_bonepitdestroyed_a3_262150_0x.dds"},
	{0xBFCBC9562F8E3393ULL, "mrl351_phantomelves.cnv"},
	{0xBFCBE930011EE29CULL, "art\\vfx\\combat\\vfx_r_arc_front_max5_c.impa.evt"},
	{0xBFCC6CCF53B3E532ULL, "gen00fl_sebastian6188723.cl"},
	{0xBFCC7E7095725264ULL, "vxht_chantry_n_2bj_0_0.vlm"},
	{0xBFCD19902CEC1946ULL, "vxht_keep_n_615_0_0.vlm"},
	{0xBFCD53976ED031BDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edgebase.mao"},
	{0xBFCD57976ED038F6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edgebase.mmh"},
	{0xBFCD61976ED049F4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ww_edgebase.msh"},
	{0xBFCDD8384E1BA7CBULL, "842088333.ogg"},
	{0xBFCE14AA9A21ED1CULL, "hlac_dr_0_deeproads_5842_10x.dds"},
	{0xBFCE1BCC34474669ULL, "ser120_ghyslain6120630.cl"},
	{0xBFCE98E1E87A505EULL, "dae321ip_at_keep_htown.utp"},
	{0xBFCF07095E2DEBD1ULL, "zz_vault_debug6178938.cl"},
	{0xBFCF8A1FBF0C959FULL, "398274925.ogg"},
	{0xBFCFBC7EB31E34A4ULL, "6142500_m.fxe"},
	{0xBFCFC14255959AACULL, "6042051_m.fxe"},
	{0xBFD0125BE8A14D7BULL, "vxca_caverns2_737_0_0.vlm"},
	{0xBFD0BEC44CB8B944ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\ready.evt"},
	{0xBFD157E03B7F0ECFULL, "llac_lt_0_lowtown_docks_a3_-65540_1.dds"},
	{0xBFD19FCFFB34BCD9ULL, "13903153.ogg"},
	{0xBFD1BD848BCBCADFULL, "dae000ip_crowd_ff8_talk17.utp"},
	{0xBFD1DD106B0D2FFCULL, "6037603_f.fxe"},
	{0xBFD1DF3141E084BEULL, "fen110_fenris6117992.cl"},
	{0xBFD20FEC52E1617EULL, "6202165_m.fxe"},
	{0xBFD28F84250C9F4FULL, "6016541_m.fxe"},
	{0xBFD3503EAEE1786EULL, "815618164.ogg"},
	{0xBFD3BA1EDD8B237CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_1.gad"},
	{0xBFD3F5A7AAEF14F0ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_runf.ani"},
	{0xBFD4C984B21A5F57ULL, "art\\characters\\creatures\\shade\\c_shade.phy"},
	{0xBFD558F7088DF56EULL, "6170194_m.fxe"},
	{0xBFD5CA16F0B431F9ULL, "6139683_f.fxe"},
	{0xBFD72C48F8C7EDC8ULL, "106428438.ogg"},
	{0xBFD75863725F7852ULL, "134005452.ogg"},
	{0xBFD7D32E65C5B168ULL, "art\\vfx\\materials\\vfx_smoke_whispy_add.mao"},
	{0xBFD839C8C822D430ULL, "vxht_bartrand_n_5sy_0_0.vlm"},
	{0xBFD8F13001297AB5ULL, "art\\vfx\\combat\\vfx_r_arc_front_max5_c.impa.ani"},
	{0xBFD9DBBE414F186EULL, "6163532_m.fxe"},
	{0xBFDAA9EAD3DC4431ULL, "6104856_f.fxe"},
	{0xBFDAB91EDD91407AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_1.ani"},
	{0xBFDBA155E49C8E6EULL, "vfx_addcloud01.dds"},
	{0xBFDC0D2D3B5829C9ULL, "vxdr_0_deeproads_458_0_0.vlm"},
	{0xBFDC8E9155C7BA21ULL, "llac_lt_0_lowtown_a2_n_c_0_5.dds"},
	{0xBFDCCB16ED98BA5EULL, "1011715073.ogg"},
	{0xBFDD12357C0F9F66ULL, "llac_lt_hangedman_n_-1_13.dds"},
	{0xBFDD19ED75580513ULL, "6092202_m.fxe"},
	{0xBFDD1A93B02E3170ULL, "6129852_f.fxe"},
	{0xBFDD6AE027364875ULL, "195232287.ogg"},
	{0xBFDD7B78FEF552F4ULL, "follower_banter6163452.cl"},
	{0xBFDDEE9732878CB3ULL, "6050271_m.fxe"},
	{0xBFDE6CA4E1EFE3BBULL, "346451787.ogg"},
	{0xBFDECCDF9FA96F1FULL, "follower_banter6151302.cl"},
	{0xBFDED5D5A954A0FAULL, "vxlt_0_lowtown_a2_n_c_ghq_0_0.vlm"},
	{0xBFDED6C44CC51F3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\ready.ani"},
	{0xBFDEF64A7654E3EDULL, "6027755_f.fxe"},
	{0xBFDF59725465E1C0ULL, "6163783_m.fxe"},
	{0xBFDFC95339112B31ULL, "mag121_idunna6084899.cl"},
	{0xBFE02C565E041BD9ULL, "vxht_0_hightown_a3_9v2_0_0.vlm"},
	{0xBFE02D06F9066EFFULL, "lt_0_lowtown_a2_n_ll.mmh"},
	{0xBFE07AC0780EDF3FULL, "6018802_m.fxe"},
	{0xBFE0E187A06D3B84ULL, "character_character_p131.psh.fxo"},
	{0xBFE1C8C2199726F3ULL, "hlac_ht_chantry_n_2534_52z.dds"},
	{0xBFE286338E586E14ULL, "vxca_caverns1_hkl_0_0.vlm"},
	{0xBFE370A848F13D9CULL, "356821713.ogg"},
	{0xBFE375F758967841ULL, "285519436.ogg"},
	{0xBFE3A668C1613FA0ULL, "lt_0_lowtown_a2_n_33_0_lc.mmh"},
	{0xBFE416E22F043BA1ULL, "6037102_m.fxe"},
	{0xBFE4430151F35E71ULL, "ring_evra.dds"},
	{0xBFE450759CC076D1ULL, "vxht_0_hightown_c1l_0_0.vlm"},
	{0xBFE475F328C3ABDDULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd01.phy"},
	{0xBFE51379304BF417ULL, "vxca_caverns2prc_afs_0_0.vlm"},
	{0xBFE5469656B06ADCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_lp.ani"},
	{0xBFE63BDF437A3A9FULL, "25353050.ogg"},
	{0xBFE67CDC27A75B1BULL, "one100dg_kelder6055438.cl"},
	{0xBFE685AF423D1AA1ULL, "345014536.ogg"},
	{0xBFE6FA2EB0C7C7D9ULL, "vxht_brothel_n_6dj_0_0.vlm"},
	{0xBFE72C0BEDD11E14ULL, "vxlt_0_lowtown_docks_btj_0_0.vlm"},
	{0xBFE7304F5E57DF69ULL, "vxlt_darktown_4y8_0_0.vlm"},
	{0xBFE809BD72E08151ULL, "llac_dr_dungeon_3215_8.dds"},
	{0xBFE85A20A7912B2BULL, "vxht_0_hightown_n_c19_0_0.vlm"},
	{0xBFE906BADB2A5D61ULL, "139750592.ogg"},
	{0xBFE91BB5E34B2F90ULL, "653579566.ogg"},
	{0xBFE9478BE6638544ULL, "hlac_lt_gallowstemplar_n_2394_0y.dds"},
	{0xBFE973E4ECCCC568ULL, "dae000_mother6108962.cl"},
	{0xBFE9A73BCC1121E1ULL, "6085749_m.fxe"},
	{0xBFE9E46D0575D514ULL, "vxlt_gallowscourtyard_7fa_0_0.vlm"},
	{0xBFEA10C060E65711ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_urn_n.dds"},
	{0xBFEA2FDC8786313FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandright_2.mao"},
	{0xBFEAE1283F9E7D03ULL, "gen_im_trs_gem_jade_03.uti"},
	{0xBFEB00ECF0826529ULL, "w_blightlands_159_0.mao"},
	{0xBFEB04B886498136ULL, "99753296.ogg"},
	{0xBFEB0CECF08279C2ULL, "w_blightlands_159_0.mmh"},
	{0xBFEB0EECF0827D78ULL, "w_blightlands_159_0.msh"},
	{0xBFEB539656B4EC1EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_lp.gad"},
	{0xBFEB7C30B9993F2BULL, "6018621_f.fxe"},
	{0xBFEB91031EC262EEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brklarge.msh"},
	{0xBFEB9F031EC27A3FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brklarge.mao"},
	{0xBFEBAB031EC28EDCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brklarge.mmh"},
	{0xBFEC707F05B5BD38ULL, "vxht_0_hightown_n_c52_0_0.vlm"},
	{0xBFED098D8FE1DE8EULL, "ss_coterie_assassin_m.wbd"},
	{0xBFEDEB4D6FFF67E9ULL, "288764628.ogg"},
	{0xBFEDF6A170ADC8ECULL, "675324386.ogg"},
	{0xBFEE06E8F36E1719ULL, "str_distor.mao"},
	{0xBFEE6F897F5F60A0ULL, "hlac_lt_0_lowtown_65536_5y.dds"},
	{0xBFEEAFB3584C9ECCULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.anb"},
	{0xBFEF1E6C4A369D74ULL, "ht_0_hightown_a2_8_0.phy"},
	{0xBFEF9B8DDF8CB131ULL, "949327696.ogg"},
	{0xBFEFF5AA1016518DULL, "hlac_lt_0_lowtown_a3_n_c_1_5x.dds"},
	{0xBFF057EBA798C0B4ULL, "592479097.ogg"},
	{0xBFF09877B2AEE272ULL, "672904556.ogg"},
	{0xBFF0A4D0D39451C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.presseddown.ani"},
	{0xBFF10746276EAB95ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.mmh"},
	{0xBFF210B7CE40BAD5ULL, "hlac_lt_gallowscourtyard_n_-65537_3z.dds"},
	{0xBFF22E9656BACC3DULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_lp.evt"},
	{0xBFF25AEEB46BCBD4ULL, "906989825.ogg"},
	{0xBFF2C2528EB68F51ULL, "vxht_0_hightown_c4c_0_0.vlm"},
	{0xBFF30DFF51C25342ULL, "vxlt_0_lowtown_a2_n_c_enx_0_1.vlm"},
	{0xBFF32E31E9784B26ULL, "6125545_m.fxe"},
	{0xBFF3C365E44F44A1ULL, "vxca_caverns1_mines_imw_0_0.vlm"},
	{0xBFF3DE5F729272B6ULL, "6068369_m.fxe"},
	{0xBFF45155CEA0875BULL, "w_blightlands_192_0.msh"},
	{0xBFF46355CEA0A5A6ULL, "w_blightlands_192_0.mao"},
	{0xBFF46915C6B7D48EULL, "6208237_m.fxe"},
	{0xBFF46F55CEA0BA45ULL, "w_blightlands_192_0.mmh"},
	{0xBFF50C47865314C4ULL, "6139883_m.fxe"},
	{0xBFF53DB0E15D7A17ULL, "w_bonepitdestroyed_a3_61_0.msh"},
	{0xBFF54FB0E15D997AULL, "w_bonepitdestroyed_a3_61_0.mao"},
	{0xBFF553B0E15D9F89ULL, "w_bonepitdestroyed_a3_61_0.mmh"},
	{0xBFF576F328D220E2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd01.mmh"},
	{0xBFF578F328D22398ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd01.msh"},
	{0xBFF5BF8400A7AB46ULL, "6119439_f.fxe"},
	{0xBFF5E6A6C7C6D5BDULL, "qun110_petrice6036425.cl"},
	{0xBFF69E1335AAD3FFULL, "vxlt_0_lowtown_a2_n_f9h_0_0.vlm"},
	{0xBFF700187573A704ULL, "zz_dae_debug6049954.cl"},
	{0xBFF70B44F0E7BA5BULL, "523146077.ogg"},
	{0xBFF71B647CFE0947ULL, "6096044_m.fxe"},
	{0xBFF75A5923549816ULL, "gen00fl_carver6104586.cl"},
	{0xBFF7704BECA43A4BULL, "vxca_caverns2prc_adm_0_0.vlm"},
	{0xBFF7A97AD36D773EULL, "vxw_encounter2_camp_46_0_0.vlm"},
	{0xBFF819F916286F22ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho01.phy"},
	{0xBFF84CFE65299C2DULL, "vxht_0_hightown_a2_cal_0_0.vlm"},
	{0xBFF8E3DA70B59DB4ULL, "6016036_m.fxe"},
	{0xBFF8F3A5CA8543DAULL, "dae000_mother6103077.cl"},
	{0xBFF8FB6C4A3EB671ULL, "ht_0_hightown_a2_8_0.msh"},
	{0xBFF9096C4A3ECD8CULL, "ht_0_hightown_a2_8_0.mao"},
	{0xBFF9156C4A3EE22FULL, "ht_0_hightown_a2_8_0.mmh"},
	{0xBFF941BEF008A5D0ULL, "hlac_ca_caverns1_mines_24424_9y.dds"},
	{0xBFF9C69CE4959A88ULL, "hlac_lt_0_lowtown_n_-262144_2z.dds"},
	{0xBFF9C84D103A8F81ULL, "vxlt_0_lowtown_docks_a3_n_br5_0_0.vlm"},
	{0xBFFA00FF643A3437ULL, "hlac_lt_0_lowtown_-3_2x.dds"},
	{0xBFFA539ADF04249CULL, "zz_dae_debug6047624.cl"},
	{0xBFFA613CC2558637ULL, "vxht_mansion2_2ek_0_0.vlm"},
	{0xBFFA63D7D42B072BULL, "962508337.ogg"},
	{0xBFFABF53B081ACC4ULL, "hlac_w_blightlands_65538_0x.dds"},
	{0xBFFB325CC4D95DA2ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornsml.msh"},
	{0xBFFB345CC4D96158ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornsml.mmh"},
	{0xBFFB405CC4D975FBULL, "art\\levels\\races\\natural\\blightlands\\foliage\\bsh_thornsml.mao"},
	{0xBFFB49B04CDE9146ULL, "zz_approval_debug6003793.cl"},
	{0xBFFC97A1E3EB0CE7ULL, "art\\vfx\\cinematic\\vfx_abom_trans_sw_p.anb"},
	{0xBFFCCEF3D97815F8ULL, "llac_w_encounter3_-131069_0.dds"},
	{0xBFFCD15A615CD2D8ULL, "seb_rogue_singletarget.ncs"},
	{0xBFFD784E32BE6D3AULL, "137460379.ogg"},
	{0xBFFD810F536CB32FULL, "dre000_dougal6131000.cl"},
	{0xBFFE4F9C7F517970ULL, "hlac_lt_0_lowtown_a2_65535_5z.dds"},
	{0xBFFE8CD0D3A06637ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.presseddown.evt"},
	{0xBFFEFFC388258B3DULL, "dae000_merchant_tailor_hm6164357.cl"},
	{0xBFFF0446277AE422ULL, "art\\vfx\\combat\\vfx_v_bl_single_proj_c.anb"},
	{0xBFFF0D9C2FE894B7ULL, "hlac_ca_caverns1_mines_24424_21z.dds"},
	{0xBFFF14A43CC0E672ULL, "character_character_p87.psh.fxo.dx11.assembly"},
	{0xBFFF69ED72449F14ULL, "vxca_caverns1_hur_0_0.vlm"},
	{0xBFFFF058DE9AC9B6ULL, "242905532.ogg"},
	{0xBFFFF9A5409FF183ULL, "6072346_f.fxe"},
	{0xC0002BF261643066ULL, "6088868_m.fxe"},
	{0xC000A5925FEAEB6FULL, "6138780_m.fxe"},
	{0xC000CAA2AC71323DULL, "vxlt_0_lowtown_a3_n_c_gos_0_1.vlm"},
	{0xC000CD6B53B29E18ULL, "654662674.ogg"},
	{0xC000D344F00D4E0DULL, "6020403_m.fxe"},
	{0xC0014E1206E86EFCULL, "vxlt_warehouse_15a_0_0.vlm"},
	{0xC001EE5CF4823DC9ULL, "vxca_caverns1_mines_i93_0_0.vlm"},
	{0xC0023542D6455D21ULL, "vxca_caverns1_mines_ijv_0_0.vlm"},
	{0xC0024559AA129907ULL, "594967027.ogg"},
	{0xC002E0CB344648C3ULL, "164642071.ogg"},
	{0xC0035D3614105C84ULL, "vrc321cr_housekeeper.utc"},
	{0xC003DDE894B3F8D9ULL, "vxht_brothel_n_6jh_0_0.vlm"},
	{0xC0047AB5088C6744ULL, "ca_caverns2prc_5802_ll.mmh"},
	{0xC00529825593E104ULL, "319224654.ogg"},
	{0xC0053592FE5939DAULL, "vxht_mansion2_3ew_0_0.vlm"},
	{0xC0058DD0D3A686C0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\trap_floortile.presseddown.gad"},
	{0xC005E228DE9A52CBULL, "vxca_caverns2prc_acl_0_0.vlm"},
	{0xC00710C1DF64C332ULL, "lgt000cr_crazed_loner.utc"},
	{0xC007649D1CA7B876ULL, "6102418_m.fxe"},
	{0xC00792E0D5C80E81ULL, "vxw_bonepitdestroyed_a3_kp_0_0.vlm"},
	{0xC007DE3446F51BD1ULL, "zz_mrl_debug6086520.cl"},
	{0xC00A661700DC00A5ULL, "vxlt_undercityencounter_1hd_0_0.vlm"},
	{0xC00B14BD186E6106ULL, "6170024_m.fxe"},
	{0xC00B2D4E8A55BA8BULL, "6077449_m.fxe"},
	{0xC00B44DEBDD8B0A3ULL, "art\\characters\\creatures\\gibgolem\\uni_gib.msh"},
	{0xC00B52DEBDD8C8FDULL, "art\\characters\\creatures\\gibgolem\\uni_gib.mmh"},
	{0xC00B81F245221512ULL, "vxca_caverns1_mines_ebe_0_0.vlm"},
	{0xC00BB35BF68686F2ULL, "6091861_f.fxe"},
	{0xC00BB77ECEDB8271ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.imp.gad"},
	{0xC00C31A481574372ULL, "19062491.ogg"},
	{0xC00C549769EB4F7BULL, "lt_0_lowtown_docks_100_0.mmh"},
	{0xC00C609769EB6318ULL, "lt_0_lowtown_docks_100_0.mao"},
	{0xC00C729769EB8265ULL, "lt_0_lowtown_docks_100_0.msh"},
	{0xC00C7B5CAD7ED0CEULL, "6084737_f.fxe"},
	{0xC00CC1E31BD125EBULL, "narrative_varric6172540.cl"},
	{0xC00D569A93BF75C3ULL, "vxdr_0_deeproads_gh6_0_0.vlm"},
	{0xC00D56EBFAAA6822ULL, "802051846.ogg"},
	{0xC00D91515032087EULL, "vxht_0_hightown_a2_c5w_0_0.vlm"},
	{0xC00DBABF3DA48F8AULL, "6036308_m.fxe"},
	{0xC00DBC750086B70EULL, "llac_w_sundermount_-131067_0.dds"},
	{0xC00DC823C26C6F47ULL, "vxdr_primevaltomb_7uc_0_0.vlm"},
	{0xC00E1D3B4A72F556ULL, "lowtown_fem_lowlife_mob6168972.cl"},
	{0xC00ECCDB577FCDEFULL, "ht_0_hightownexplosion_n_33_0_lc.mmh"},
	{0xC00EE53B8687646FULL, "vxca_caverns1_j27_0_1.vlm"},
	{0xC00F05B763AE4912ULL, "vxca_caverns2_5cl_0_0.vlm"},
	{0xC00F846BAABD0A5FULL, "gen00fl_sebastian6189666.cl"},
	{0xC010366AC8D0EF7EULL, "gen_im_arm_cht_nobe_t3.uti"},
	{0xC0106244FCF62F8CULL, "37630311.ogg"},
	{0xC01074C97225A5E0ULL, "gen00fl_carver6102667.cl"},
	{0xC0108A031EE172A7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_01brklarge.phy"},
	{0xC01125B13098F05CULL, "vxlt_0_lowtown_docks_a3_n_bfx_0_0.vlm"},
	{0xC0116C9854251876ULL, "433981426.ogg"},
	{0xC01229F0080D425EULL, "gen00fl_aveline6120778.cl"},
	{0xC012B87ECEE1A21FULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.imp.ani"},
	{0xC0130B4D9333E242ULL, "hlac_ht_0_hightown_a2_-131070_0x.dds"},
	{0xC013D26513272823ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_mattock_n.dds"},
	{0xC014328B18C0E71DULL, "one110_gustav6089823.cl"},
	{0xC0158B8F26111373ULL, "follower_banter6155742.cl"},
	{0xC015E87C14D624DEULL, "vxdr_primevaltomb_3j7_0_0.vlm"},
	{0xC0163F808748BEAEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1med.msh"},
	{0xC0164D808748D6FFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1med.mao"},
	{0xC01659808748EA9CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1med.mmh"},
	{0xC0168A6D97E86DD9ULL, "vxw_sundermount_20z_0_0.vlm"},
	{0xC0169834652B20D4ULL, "vxht_mansion2b_3ck_0_0.vlm"},
	{0xC0170E611219736CULL, "vxw_encounter2_ge_0_0.vlm"},
	{0xC0172BEC253B38FFULL, "565080287.ogg"},
	{0xC01742A6F5EE8836ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highwardrobe_d.dds"},
	{0xC0175382544DE9ACULL, "vfx_rendertoshadowmap_p0.psh.fxo.dx11"},
	{0xC017F4B6697286B7ULL, "ts_jump.ncs"},
	{0xC018489100B68722ULL, "pro000_merchant6051620.cl"},
	{0xC0187C0D7BB7C15BULL, "vxw_bonepit_1s5_0_0.vlm"},
	{0xC018A394CF5B91C6ULL, "vxlt_undercitydungeon_17c_0_0.vlm"},
	{0xC018FF8D900708BCULL, "ss_coterie_assassin_m.bnk"},
	{0xC0191650139BE596ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v2.ani"},
	{0xC019731305AF320EULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_loop.ani"},
	{0xC019AD99D1FEF606ULL, "vxht_brothel_n_59a_0_0.vlm"},
	{0xC019C71ECA0C22D0ULL, "gen00fl_anders6135997.cl"},
	{0xC019F3CF37182D92ULL, "vxca_caverns1_mines_egf_0_0.vlm"},
	{0xC019F72ECD5D986DULL, "25920661.ogg"},
	{0xC01AE3697BA847BCULL, "gen00fl_anders6113074.cl"},
	{0xC01BBC7D61727021ULL, "vxht_mansion2b_31d_0_0.vlm"},
	{0xC01BE1F148E95F0DULL, "art\\characters\\playercharacter\\humanmale\\rob_magb.mmh"},
	{0xC01BE5F148E9665EULL, "art\\characters\\playercharacter\\humanmale\\rob_magb.mao"},
	{0xC01BF3F148E97D93ULL, "art\\characters\\playercharacter\\humanmale\\rob_magb.msh"},
	{0xC01C329689F14480ULL, "gen00fl_isabela6088066.cl"},
	{0xC01C388495BE1DEFULL, "store_dae270cr_armor.utm"},
	{0xC01C9E84B257CBAFULL, "art\\characters\\creatures\\shade\\c_shade.mao"},
	{0xC01CAA84B257E04CULL, "art\\characters\\creatures\\shade\\c_shade.mmh"},
	{0xC01CB084B257EA3EULL, "art\\characters\\creatures\\shade\\c_shade.msh"},
	{0xC01D0FAC45506166ULL, "6086668_m.fxe"},
	{0xC01E00F916491475ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho01.mmh"},
	{0xC01E02F9164917EBULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho01.msh"},
	{0xC01E4B354B327E02ULL, "217250397.ogg"},
	{0xC01E4DD01852183AULL, "vxlt_darktown_5r3_0_0.vlm"},
	{0xC01ECD841878724CULL, "qun221_viscount6084739.cl"},
	{0xC01ED216D86E6EC9ULL, "vxca_caverns1_mines_i31_0_0.vlm"},
	{0xC01ED858CFDEDB71ULL, "vxw_woundedcoast_16w_0_0.vlm"},
	{0xC01F2EC690C288B8ULL, "vxw_woundedcoast_nw_0_0.vlm"},
	{0xC01F367225A2EC70ULL, "art\\levels\\4cinematics\\ferelden\\prp_killdagger.mao"},
	{0xC01F3A7225A2F283ULL, "art\\levels\\4cinematics\\ferelden\\prp_killdagger.mmh"},
	{0xC01F487225A30ADDULL, "art\\levels\\4cinematics\\ferelden\\prp_killdagger.msh"},
	{0xC01F5BC7DD7C5B06ULL, "vxw_woundedcoast_n9_0_0.vlm"},
	{0xC01F90A7C0C5345DULL, "1062122199.ogg"},
	{0xC0201AD28B4B25A9ULL, "1000612979.ogg"},
	{0xC0208D9E2C249EE8ULL, "dae150_chandan6205593.cl"},
	{0xC02094DF060317CFULL, "996366774.ogg"},
	{0xC020F340BC60F1C2ULL, "hlac_ca_caverns2prc_5802_9y.dds"},
	{0xC02194591426790DULL, "llac_dr_0_deeproads_5842_0.dds"},
	{0xC022C068CC48FE0CULL, "vxw_bonepitmining_a2_1nc_0_0.vlm"},
	{0xC02305C4FE6978AFULL, "vxlt_hangedman_w9_0_0.vlm"},
	{0xC023522D1EE55CB4ULL, "vxlt_warehouse_ad_0_0.vlm"},
	{0xC0235429DBEA44DAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_peak.mmh"},
	{0xC0235829DBEA4B11ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_peak.mao"},
	{0xC0236629DBEA6320ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ci_peak.msh"},
	{0xC02403C7EE11731EULL, "isa111_isabela6170376.cl"},
	{0xC02436EBFE6FEEEAULL, "6166922_m.fxe"},
	{0xC02494A1E40C59E8ULL, "art\\vfx\\cinematic\\vfx_abom_trans_sw_p.mmh"},
	{0xC02495932D772361ULL, "516539503.ogg"},
	{0xC024FC4A89DE8659ULL, "vxw_sundermount_25y_0_0.vlm"},
	{0xC0250A1157213954ULL, "vxht_mansion2b_3fh_0_0.vlm"},
	{0xC025216D9A6C5700ULL, "dae000_bartrand6033991.cl"},
	{0xC0256210E49D6C8EULL, "6179877_f.fxe"},
	{0xC025D28179711B25ULL, "fen110_fenris6118057.cl"},
	{0xC0261DCEC44F7600ULL, "116098485.ogg"},
	{0xC02657DCAE19E736ULL, "899147374.ogg"},
	{0xC02658890AED231CULL, "vxw_woundedcoast_1e3_0_0.vlm"},
	{0xC0266B1305B9AF5FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_loop.evt"},
	{0xC02672A5A3E0218FULL, "vxht_mansion2_25s_0_0.vlm"},
	{0xC02696073C99BC97ULL, "vxlt_0_lowtown_a2_n_c_hwz_0_0.vlm"},
	{0xC026A1376D47ACADULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_crumble.mmh"},
	{0xC026A1DCB0C53E5CULL, "713605419.ogg"},
	{0xC026A5376D47B3FEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_crumble.mao"},
	{0xC026B3376D47CB33ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_crumble.msh"},
	{0xC026DFF14D10CB62ULL, "one100_party_comments6133459.cl"},
	{0xC0271E5013A83087ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v2.evt"},
	{0xC0276D851F383F86ULL, "vxht_0_hightown_a2_c7i_0_0.vlm"},
	{0xC0281EB49BE53FA9ULL, "vxlt_undercitydungeon_xt_0_0.vlm"},
	{0xC02828BCFEA30EFFULL, "hlac_lt_0_lowtown_a2_-196607_0z.dds"},
	{0xC0287138B969C243ULL, "436640873.ogg"},
	{0xC028A8727ACA49A4ULL, "675880617.ogg"},
	{0xC028D957E1F2EFA7ULL, "99181054.ogg"},
	{0xC0291564558955D2ULL, "hlac_w_sundermount_-131069_0x.dds"},
	{0xC02921AF10F39841ULL, "6128866_m.fxe"},
	{0xC0295837C0807708ULL, "dev_res_fla.mao"},
	{0xC029A48CBBF96E79ULL, "hlac_ht_mansion2b_818_21x.dds"},
	{0xC029ADB9C3440C2FULL, "llac_lt_gallowstemplar_n_2383_11.dds"},
	{0xC029E735A45B05E9ULL, "gen_im_arm_cht_hvy_act3_01.uti"},
	{0xC02A012DF4FC5E70ULL, "vxlt_blackemporium_3s_0_1.vlm"},
	{0xC02AC9DAFCB29D08ULL, "vxlt_warehouse_n_aa_0_0.vlm"},
	{0xC02B41EE0BDF3D21ULL, "vxlt_gallowstemplar_6sd_0_0.vlm"},
	{0xC02B56E0924C3410ULL, "6147143_m.fxe"},
	{0xC02B6EA20A7841F5ULL, "6005644_m.fxe"},
	{0xC02B93080A1EBC40ULL, "6149358_m.fxe"},
	{0xC02B94309D6BEFECULL, "hlac_dr_primevaltomb_659_13x.dds"},
	{0xC02B944168331323ULL, "llac_w_bonepit_1_0.dds"},
	{0xC02BE5AF1845FF17ULL, "vxlt_hangedman_n_26k_0_0.vlm"},
	{0xC02C315C4404D873ULL, "vxht_keep_705_0_0.vlm"},
	{0xC02C4B004AC820E1ULL, "mcr310_argument6074459.cl"},
	{0xC02CDAF148F7C6CAULL, "art\\characters\\playercharacter\\humanmale\\rob_magb.phy"},
	{0xC02D17B467C1980EULL, "652761544.ogg"},
	{0xC02D2022A97CF174ULL, "gen00fl_sebastian6180523.cl"},
	{0xC02D47952BFF5A15ULL, "gen00fl_sebastian6174699.cl"},
	{0xC02D4A35C1D4F3F1ULL, "vxw_woundedcoast_11v_0_0.vlm"},
	{0xC02D4C1305BF9948ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_action_loop.gad"},
	{0xC02D601C5C8248EAULL, "vxlt_undercityserialkiller_1qh_0_0.vlm"},
	{0xC02DA0A382B8A138ULL, "vxw_woundedcoast_kt_0_0.vlm"},
	{0xC02DCA3ADABF9B94ULL, "vxw_bonepitdestroyed_a3_1wb_0_0.vlm"},
	{0xC02E0681369BC503ULL, "195090373.ogg"},
	{0xC02E0E0933D46055ULL, "774775752.ogg"},
	{0xC02E3F5013AE87D0ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rng_atk_v2.gad"},
	{0xC02F51FCC59A31C5ULL, "250364788.ogg"},
	{0xC03079EF50894583ULL, "vxlt_blackemporium_td_0_0.vlm"},
	{0xC030A5E348BA520BULL, "232175016.ogg"},
	{0xC030B6DFBDFE4044ULL, "mag110_samson6016163.cl"},
	{0xC030D439FB314794ULL, "6123387_m.fxe"},
	{0xC030F649327C89DCULL, "vxlt_0_lowtown_docks_a3_n_aq6_0_0.vlm"},
	{0xC032017571196414ULL, "6211934_m.fxe"},
	{0xC032913452D56E3DULL, "vxht_bartrand_n_65y_0_0.vlm"},
	{0xC03291B1CD39F404ULL, "559747418.ogg"},
	{0xC032B7083C799AF5ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg02.phy"},
	{0xC032FE081D849D4CULL, "523934707.ogg"},
	{0xC0332C2EEAA45699ULL, "201979962.ogg"},
	{0xC0339B10F9782A24ULL, "vxdr_primevaltomb_350_0_0.vlm"},
	{0xC034C8A63016EB9DULL, "hlac_ca_caverns1_mines_24423_9y.dds"},
	{0xC034CE2D0755433AULL, "w_woundedcoast_159_0.msh"},
	{0xC034DC2D07555B73ULL, "w_woundedcoast_159_0.mao"},
	{0xC034E02D07556180ULL, "w_woundedcoast_159_0.mmh"},
	{0xC03507E42E8FD517ULL, "vxlt_0_lowtown_a2_n_c_hru_0_0.vlm"},
	{0xC0352699935DBE1BULL, "vxlt_0_lowtown_docks_n_d2z_0_0.vlm"},
	{0xC035592813658490ULL, "vxlt_warehouse_n_91_0_0.vlm"},
	{0xC0358312ADA2ECDAULL, "6081409_m.fxe"},
	{0xC0360AE388EEC1C6ULL, "35022307.ogg"},
	{0xC036404D567431E0ULL, "con_ico_choice_special.dds"},
	{0xC0364D0D15AC8845ULL, "ts_apostates_arm_table.gda"},
	{0xC037283D1D84F467ULL, "mrl000_marethari6069807.cl"},
	{0xC0377BAD1C6E4855ULL, "vxca_caverns2_8dk_0_0.vlm"},
	{0xC0379A376D56146AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_crumble.phy"},
	{0xC037E14D88AE571AULL, "12084636.ogg"},
	{0xC037EFD973ACE956ULL, "vxw_woundedcoast_1cc_0_0.vlm"},
	{0xC0380A4B8CA12EBDULL, "6102951_f.fxe"},
	{0xC0380BB226AD346EULL, "vrc321_varric6072471.cl"},
	{0xC038608A92E120DCULL, "6071211_m.fxe"},
	{0xC03941E12EE98662ULL, "llac_ht_bartrand_n_3382_1.dds"},
	{0xC039A50122EA94F5ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firexl.mmh"},
	{0xC039F63A4845B510ULL, "vxht_0_hightown_a3_ccr_0_0.vlm"},
	{0xC03A0C20D0A93143ULL, "gen00fl_merrill6108635.cl"},
	{0xC03AB49C848F477CULL, "llac_ht_playershouse_n_855_9.dds"},
	{0xC03AC78E3FA6814FULL, "928222727.ogg"},
	{0xC03AFAFB0450DEF0ULL, "971248555.ogg"},
	{0xC03B0727BF867CD8ULL, "dae000_gamlen6145115.cl"},
	{0xC03B4DDE1C8ADED9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\shielddefenceq_override.anb"},
	{0xC03B549274D6EA87ULL, "602447213.ogg"},
	{0xC03B7F810810AF9AULL, "6087124_m.fxe"},
	{0xC03B8F3EF2AAEC01ULL, "lt_0_lowtown_docks_18_0_lc.mmh"},
	{0xC03BA87C377149D9ULL, "883617444.ogg"},
	{0xC03BAA3ADC072AC5ULL, "weapon_weapon_p0.psh.fxo.dx11"},
	{0xC03C174CD723CC75ULL, "6124703_m.fxe"},
	{0xC03C23DC78E8B9CBULL, "vxlt_undercitydungeon_1c6_0_0.vlm"},
	{0xC03C3C17CCB5B414ULL, "vxw_bonepitdestroyed_a3_1xc_0_0.vlm"},
	{0xC03D1C97BAFA0299ULL, "387362.ogg"},
	{0xC03E8AF879E68282ULL, "t3_shk_b02.tnt"},
	{0xC03F00BCC590B2A5ULL, "dae200cr_viscount6087164.cl"},
	{0xC03F69EEDC19AC50ULL, "al_rogue_s03.gda"},
	{0xC03F95BD66DBCE86ULL, "913578238.ogg"},
	{0xC03FA1F5C34405BFULL, "146671194.ogg"},
	{0xC03FB2FF70F249F9ULL, "230415466.ogg"},
	{0xC03FD63F3860C61CULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv1.ani"},
	{0xC03FDD6CD9C7AAC5ULL, "ser221_gamlen6135172.cl"},
	{0xC040353DF767727CULL, "zz_lgt_2_debug6211870.cl"},
	{0xC04047493AC1E233ULL, "hlac_lt_0_lowtown_a2_n_65538_4x.dds"},
	{0xC0404A904364CF6FULL, "vxlt_gallowstemplar_f_3s8_0_0.vlm"},
	{0xC0405CF299856019ULL, "vxht_bartrand_n_6rp_0_0.vlm"},
	{0xC040C22B2DB9EBAFULL, "6053779_m.fxe"},
	{0xC04148D29B66448CULL, "1040404294.ogg"},
	{0xC0415C80C8964B5CULL, "vxht_bartrand_n_5wy_0_0.vlm"},
	{0xC041AA60EEFA4F0AULL, "gen00fl_anders6113069.cl"},
	{0xC041C605168D95BFULL, "vxw_encounter2_gj_0_0.vlm"},
	{0xC041FC859B66E8EDULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12_tool.mmh"},
	{0xC04200859B66EF3EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12_tool.mao"},
	{0xC0420E859B670773ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12_tool.msh"},
	{0xC0432610EC64ABD3ULL, "llac_w_bonepitmining_a2_5_0.dds"},
	{0xC0433C42EED9541CULL, "vxw_woundedcoast_1k1_0_0.vlm"},
	{0xC04393763DC93A62ULL, "6149011_m.fxe"},
	{0xC043F4EF9F99055AULL, "6084709_m.fxe"},
	{0xC0447867149BBDD7ULL, "vxlt_0_lowtown_n_er6_0_0.vlm"},
	{0xC0448BFC4FC4CDAEULL, "lt_0_lowtown_44_0_ll.mmh"},
	{0xC0451FA023CCD690ULL, "vxlt_gallowscourtyard_n_4lg_0_0.vlm"},
	{0xC0452E2DDFEF8252ULL, "6087267_m.fxe"},
	{0xC0454147C6B50672ULL, "gen00fl_isabela6079679.cl"},
	{0xC045445E583196B5ULL, "mag131_keran6121104.cl"},
	{0xC04552C53FE6D0D4ULL, "561795500.ogg"},
	{0xC04554329DFE1652ULL, "llac_lt_blackemporium_1055_20.dds"},
	{0xC0455EEACB79B5CFULL, "hlac_lt_0_lowtown_docks_a3_-4_5x.dds"},
	{0xC045CABC41FAC7A0ULL, "dae000cr_quintus.utc"},
	{0xC046E33F3866FB5EULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv1.gad"},
	{0xC047137027A2100DULL, "static1_ds1.psh.fxo"},
	{0xC0476089C720D086ULL, "6016445_m.fxe"},
	{0xC0476506AFAA1042ULL, "mer210_coterie_contact6035284.cl"},
	{0xC0477CF8A8B87ED0ULL, "llac_dr_0_deeproads_17038_5.dds"},
	{0xC04781B5721C6F21ULL, "vxca_caverns1_mines_j21_0_0.vlm"},
	{0xC0478B0F56A79DBBULL, "44450570.ogg"},
	{0xC04816AE74ACC635ULL, "757772165.ogg"},
	{0xC048496FFE93FDC3ULL, "vxlt_warehouse_n_e2_0_0.vlm"},
	{0xC04868173A3BCD90ULL, "vxht_0_hightown_a3_cds_0_0.vlm"},
	{0xC048890878F2D21AULL, "character_character_p263.psh.fxo.dx11"},
	{0xC048C968FC323017ULL, "vxlt_hangedman_n_28e_0_0.vlm"},
	{0xC048CBC471A730BAULL, "741500278.ogg"},
	{0xC048DAE91BEC2C2DULL, "hlac_lt_0_lowtown_n_-65536_2y.dds"},
	{0xC04926F917BF6414ULL, "1050258131.ogg"},
	{0xC049648C11F0DA0EULL, "vxdr_0_deeproads_d39_0_0.vlm"},
	{0xC04982150CB4487FULL, "character_character_p73.psh.fxo.dx11.assembly"},
	{0xC049DCF3231496E3ULL, "6151503_m.fxe"},
	{0xC04ABD74333AB4AAULL, "vxht_keep_n_619_0_0.vlm"},
	{0xC04B2A11BB64A5F5ULL, "mag170_hugh6080748.cl"},
	{0xC04B4353C3B7DEABULL, "vxca_caverns2_73g_0_0.vlm"},
	{0xC04BB23F2381966AULL, "6207430_m.fxe"},
	{0xC04BBA44002E1386ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_cast_aoe_c.impa.ani"},
	{0xC04BC1DFFAD7A0F0ULL, "m_pro000ar_blightlands_real.bnk"},
	{0xC04CF76155AD1C1FULL, "follower_banter6171582.cl"},
	{0xC04D56A33759899FULL, "272857349.ogg"},
	{0xC04D6173DF583BFDULL, "hlac_lt_undercityserialkiller_-1_44y.dds"},
	{0xC04DBA05DFE52699ULL, "390947238.ogg"},
	{0xC04DBE3F386CDB7DULL, "art\\characters\\creatures\\corpse\\animation\\combat\\appearv1.evt"},
	{0xC04DC28FCC438411ULL, "lgt220_shp_mag_guard6149518.cl"},
	{0xC04DFB0AA6AE1C6CULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.impa.ani"},
	{0xC04E9022AE39CB9BULL, "vxlt_0_lowtown_a2_gch_0_0.vlm"},
	{0xC04EA7C3F4C7A634ULL, "vxlt_warehouse_pk_0_0.vlm"},
	{0xC04EF671D0025992ULL, "vxht_0_hightown_a2_bab_0_0.vlm"},
	{0xC04EFA474EA83355ULL, "hlac_ca_caverns1_mines_24425_10y.dds"},
	{0xC04F176230C5D000ULL, "290541391.ogg"},
	{0xC04F7154FFF4BCF6ULL, "t1_skn_d02.tnt"},
	{0xC04FCB8DC56012A3ULL, "vxw_encounter2_5h_0_0.vlm"},
	{0xC04FDD1F57354BCAULL, "hlac_lt_0_lowtown_docks_n_131072_1z.dds"},
	{0xC0500B8C48A3A5DAULL, "mag110st_arianni.stg"},
	{0xC0502CAA4C3427CEULL, "vxlt_0_lowtown_docks_n_9nn_0_0.vlm"},
	{0xC0504BC673E63C72ULL, "vxlt_gallowstemplar_4yk_0_0.vlm"},
	{0xC0505538ED002640ULL, "302175499.ogg"},
	{0xC05089E106427FBAULL, "dre000_dougal6129730.cl"},
	{0xC0519020FC96E3CFULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.impa.gad"},
	{0xC051A21B2C548849ULL, "plt_zz_tbept_debug.plo"},
	{0xC051B38BD919666EULL, "856273098.ogg"},
	{0xC0527D6123EB7592ULL, "llac_ht_0_hightown_n_65535_0.dds"},
	{0xC052DAEF06068539ULL, "vrc321_party_barks6095782.cl"},
	{0xC052EA440691D657ULL, "vxlt_0_lowtown_n_eo1_0_0.vlm"},
	{0xC052F5859B754FAAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball12_tool.phy"},
	{0xC0534FD8E477B3E7ULL, "6102126_f.fxe"},
	{0xC05350ABEE800116ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_0y.dds"},
	{0xC05375976A2789B0ULL, "lt_0_lowtown_docks_100_0.phy"},
	{0xC054080AA6B29DAEULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.impa.gad"},
	{0xC0545F67005A7955ULL, "vxca_caverns2_8ni_0_0.vlm"},
	{0xC054B8FD8DF2EF93ULL, "669734822.ogg"},
	{0xC054D39357991A56ULL, "vxw_woundedcoast_1ie_0_0.vlm"},
	{0xC054ED9DBEECA715ULL, "plt_zz_qunpt_debug.nss"},
	{0xC055426FCAB937B6ULL, "hlac_ht_0_hightown_a2_-196608_0x.dds"},
	{0xC0558AEDF9C5D326ULL, "13320591.ogg"},
	{0xC055B67BC8B8298AULL, "hlac_lt_0_lowtown_docks_a3_196607_2z.dds"},
	{0xC055DAB814A1DCEEULL, "w_blightlands_207_0.msh"},
	{0xC055E8B814A1F43FULL, "w_blightlands_207_0.mao"},
	{0xC055F4B814A208DCULL, "w_blightlands_207_0.mmh"},
	{0xC0560081A601EAE0ULL, "vxht_0_hightown_av6_0_0.vlm"},
	{0xC05641FE1BCDA46EULL, "489786336.ogg"},
	{0xC057200E913E1461ULL, "vxlt_0_lowtown_a2_n_c_gz8_0_0.vlm"},
	{0xC057A45FD9A2CAB2ULL, "692705281.ogg"},
	{0xC057C0D8037C9334ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masa.phy"},
	{0xC05818AEC508551FULL, "vxw_sundermount_349_0_0.vlm"},
	{0xC0583FCF8F6EA887ULL, "6070985_m.fxe"},
	{0xC05877A9D5CEC702ULL, "llac_lt_0_lowtown_131072_3.dds"},
	{0xC058A520FC9D267CULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.impa.evt"},
	{0xC058F63A589AEAB8ULL, "vxw_woundedcoast_zs_0_0.vlm"},
	{0xC05940AE4AD48BA9ULL, "hlac_w_encounter1_-131069_0y.dds"},
	{0xC0598CA30A39C84DULL, "vxlt_0_lowtown_gpo_0_0.vlm"},
	{0xC059A244003A28F7ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_cast_aoe_c.impa.evt"},
	{0xC059DA0F379A3908ULL, "mag131_taraneh6085551.cl"},
	{0xC05A6B3A2CFE9171ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_cupwood_01_0_n.dds"},
	{0xC05A8E880D1B79F9ULL, "vxlt_shop_iz_0_0.vlm"},
	{0xC05ABB9C413B5ED0ULL, "art\\vfx\\models\\vfx_twist_chunky_single.mmh"},
	{0xC05AC99C413B76EAULL, "art\\vfx\\models\\vfx_twist_chunky_single.msh"},
	{0xC05B030AA6B8B38DULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.impa.evt"},
	{0xC05B3097F026BE65ULL, "304702786.ogg"},
	{0xC05BEB9D3B52153FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobc_0s.dds"},
	{0xC05BF848BB6F86B0ULL, "dae200cr_viscount6087422.cl"},
	{0xC05C406779F6C286ULL, "vxw_bonepit_1v3_0_0.vlm"},
	{0xC05C731A9C5B4B89ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_crown_d.dds"},
	{0xC05CAC773DE97B33ULL, "ht_0_hightown_n_28_0.anb"},
	{0xC05CD4800C1D13C6ULL, "vxdr_primevaltomb_6qm_0_2.vlm"},
	{0xC05CFF04C9966224ULL, "756709004.ogg"},
	{0xC05D19A0E6BDBEB4ULL, "vxlt_warehouse_uh_0_0.vlm"},
	{0xC05D45F6613FEBE2ULL, "701331665.ogg"},
	{0xC05D684EC1F87212ULL, "vxht_0_hightown_a2_bjc_0_0.vlm"},
	{0xC05D926528C7F021ULL, "496790138.ogg"},
	{0xC05EA1B897864D1CULL, "ser220ip_to_estate.utp"},
	{0xC05F215990784590ULL, "hlac_lt_darktown_-1_18y.dds"},
	{0xC05FA344003E9580ULL, "art\\vfx\\combat\\vfx_c_2nd_deb_cast_aoe_c.impa.gad"},
	{0xC05FBB8DE592117FULL, "vxht_mansion2_280_0_0.vlm"},
	{0xC0601FFCD2C5851CULL, "vxw_woundedcoast_1q7_0_0.vlm"},
	{0xC0604FAC30D5147DULL, "vxdr_primevalentrance_1k9_0_0.vlm"},
	{0xC0609E69987A44A3ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.cess.evt"},
	{0xC0619DD80384AC31ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masa.msh"},
	{0xC061ABD80384C44CULL, "art\\characters\\playercharacter\\humanmale\\hlf_masa.mao"},
	{0xC061B4C0FFEF9AF3ULL, "rdr111dg_assistant6053150.cl"},
	{0xC061B7D80384D8EFULL, "art\\characters\\playercharacter\\humanmale\\hlf_masa.mmh"},
	{0xC061DBF01669C516ULL, "dae000_hate_enemies6167516.cl"},
	{0xC062C7EE45F4C555ULL, "vxht_0_hightown_bbt_0_0.vlm"},
	{0xC062D09FF874AB7CULL, "vxw_woundedcoast_1g5_0_0.vlm"},
	{0xC062DA76332A3F2FULL, "dae000_merchant_generic_hm6172594.cl"},
	{0xC062EE57BDF9DD0DULL, "vxlt_0_lowtown_a2_f4v_0_0.vlm"},
	{0xC063B72FB075AB26ULL, "vxw_sundermount_31s_0_0.vlm"},
	{0xC063D93E6AE56ECFULL, "zz_rdr_debug6206284.cl"},
	{0xC064600C6A1BA7FFULL, "dae121ip_office_door.utp"},
	{0xC064A35949C2E368ULL, "mag321cr_nella.utc"},
	{0xC064CA67D6B8A93AULL, "vxw_woundedcoast_14j_0_0.vlm"},
	{0xC064CF012A45EE4DULL, "gen00fl_fenris6126769.cl"},
	{0xC064F712BAFA8F9EULL, "cr_sm_sp.mao"},
	{0xC0651E2D355B5BA4ULL, "vxw_bonepit_26f_0_0.vlm"},
	{0xC065AD20FCA7BD95ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.impa.ani"},
	{0xC065E6CD52799A84ULL, "vxht_keep_63r_0_0.vlm"},
	{0xC0663DF277502EE6ULL, "unlock_pressed.cur"},
	{0xC066869D122B332EULL, "872326329.ogg"},
	{0xC0668A8BB6FE6D9FULL, "vxw_sundermount_338_0_0.vlm"},
	{0xC06768174A910338ULL, "vxw_woundedcoast_wp_0_0.vlm"},
	{0xC067776998802124ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.cess.gad"},
	{0xC06791AEA297FD94ULL, "vxw_bonepitdestroyed_a3_1kf_0_0.vlm"},
	{0xC0688992077AA455ULL, "vxlt_gallowstemplar_2j7_0_0.vlm"},
	{0xC06890D0452DD000ULL, "264499521.ogg"},
	{0xC068EE8380A80F90ULL, "mag360cr_hostage_carver.utc"},
	{0xC06A4CB111F0D1FDULL, "mag320_bsc_banter6100758.cl"},
	{0xC06AF2A2894A823FULL, "6166637_m.fxe"},
	{0xC06AFAA884A7BB20ULL, "t3_wdg_dw3.tnt"},
	{0xC06B091ED8ECDF87ULL, "221144777.ogg"},
	{0xC06B46DAA26272BFULL, "hlac_dr_primevaltomb_658_4z.dds"},
	{0xC06B4FFA53725BD4ULL, "6152866_m.fxe"},
	{0xC06BA58D5CC676A5ULL, "hlac_lt_blackemporium_1055_7x.dds"},
	{0xC06BD8DCA694678BULL, "dae000_light_tombs_boss.bnk"},
	{0xC06C975DC64D0C13ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_heathaze1x8m.mmh"},
	{0xC06CAFC55F449F96ULL, "224687702.ogg"},
	{0xC06CCFCFF082BE51ULL, "dae000ip_crowd_lm3_talk2.utp"},
	{0xC06D02A3D3907823ULL, "vxw_blightlands_ga_0_0.vlm"},
	{0xC06D0FC8D28E679EULL, "hlac_lt_gallowstemplar_n_2394_9x.dds"},
	{0xC06D4AE0DC4A5E97ULL, "6113530_m.fxe"},
	{0xC06DC1CB28094C51ULL, "ser270_templar6122504.cl"},
	{0xC06DF5FF298D55E1ULL, "gen00fl_aveline6151674.cl"},
	{0xC06E7BC924D10C42ULL, "mag110_vincento6017808.cl"},
	{0xC06E866998865902ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.cess.ani"},
	{0xC06EC404EED3D3E8ULL, "mcr310_argument6074368.cl"},
	{0xC06F4BB2E0D31445ULL, "6120826_m.fxe"},
	{0xC06FB1EAD78C09B2ULL, "50337857.ogg"},
	{0xC06FDADCA698546BULL, "dae000_light_tombs_boss.cnv"},
	{0xC0708776D59B315CULL, "6162986_m.fxe"},
	{0xC070A3DFFAF689BAULL, "m_pro000ar_blightlands_real.wbd"},
	{0xC070E6117424FE2AULL, "mrl310_party_comments6143045.cl"},
	{0xC07139CB37EADDD5ULL, "vxht_0_hightown_bos_0_0.vlm"},
	{0xC0716034AFEFF58DULL, "vxlt_0_lowtown_a2_f1u_0_0.vlm"},
	{0xC07247B7206ADD82ULL, "vxlt_gallowstemplar_4vf_0_0.vlm"},
	{0xC0731A89CF062CACULL, "6077059_m.fxe"},
	{0xC0737CF60987EC20ULL, "vxlt_0_lowtown_docks_a3_n_d04_0_0.vlm"},
	{0xC073900A27517424ULL, "vxw_bonepit_21g_0_0.vlm"},
	{0xC0739F06D4764743ULL, "vxlt_warehouse_n_r7_0_0.vlm"},
	{0xC073C17B9E5FDBBFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_monique.gad"},
	{0xC073F5A32564F0BFULL, "w_bonepitdestroyed_a3_25_0_lc.mmh"},
	{0xC07411EFD670B658ULL, "550561989.ogg"},
	{0xC07429426CC1FA67ULL, "228975286.ogg"},
	{0xC0746023F2D9519CULL, "6161460_m.fxe"},
	{0xC074DE3FBC2593F1ULL, "dae000_gamlen6022570.cl"},
	{0xC075595A96248B66ULL, "vxca_caverns1_mines_jg9_0_0.vlm"},
	{0xC07591B4D3BB4EE0ULL, "6039246_m.fxe"},
	{0xC075FCBC2B25D8E3ULL, "plt_cod_cha_fenris1.plo"},
	{0xC076038B948E1614ULL, "vxw_bonepitdestroyed_a3_1lg_0_0.vlm"},
	{0xC076B971C3C864D1ULL, "6105698_m.fxe"},
	{0xC076DA8318193CB0ULL, "6095406_m.fxe"},
	{0xC076EC9DBF098C76ULL, "plt_zz_qunpt_debug.plo"},
	{0xC076FB6EF970BCD5ULL, "vxlt_gallowstemplar_2e6_0_0.vlm"},
	{0xC0770D945BA67C1FULL, "bodyimp_v.dds"},
	{0xC077C8E7E5F1FEF7ULL, "205174592.ogg"},
	{0xC0790FFCD1D37922ULL, "6121570_m.fxe"},
	{0xC07942132CC02429ULL, "zz_waldo2.cut"},
	{0xC07958CD505D45D3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2.phy"},
	{0xC079CA083CB5BD10ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg02.msh"},
	{0xC079CCDC97023280ULL, "vxlt_0_lowtown_a2_n_gvw_0_0.vlm"},
	{0xC079D8083CB5D52AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hg02.mmh"},
	{0xC079E852A3002940ULL, "vxht_keep_6fy_0_0.vlm"},
	{0xC07A02B2B31B37E3ULL, "6173756_m.fxe"},
	{0xC07A13764AFA7EE9ULL, "2432918.ogg"},
	{0xC07A2745BB39709BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_reaction.evt"},
	{0xC07A37E3D138D815ULL, "hlac_ca_cavern1_haunted_24424_15y.dds"},
	{0xC07AABB996B7F75AULL, "vxw_encounter3_bw_0_0.vlm"},
	{0xC07AD2BC6108D7DFULL, "bind_cyc2.mao"},
	{0xC07B092B01129EBCULL, "t3_wdg_w05.tnt"},
	{0xC07B49CEFC1084C9ULL, "art\\vfx\\creatures\\fxc_drg_brth_l_c.mmh"},
	{0xC07BD83F621043D9ULL, "711090522.ogg"},
	{0xC07BE30A49B6BEFFULL, "vxlt_hangedman_vz_0_0.vlm"},
	{0xC07BFCFB5AFD10A2ULL, "224303846.ogg"},
	{0xC07C77C345E0F680ULL, "6096168_m.fxe"},
	{0xC07CB2C668166D89ULL, "6102831_m.fxe"},
	{0xC07CC34AC8BE3B67ULL, "6149869_m.fxe"},
	{0xC07D4EE23F423A34ULL, "art\\vfx\\combat\\vfx_g_spider_att_f2_c.anb"},
	{0xC07DB21119DC733CULL, "lowtown_male_urchin_sta6164200.cl"},
	{0xC07DC5D103627B77ULL, "563426599.ogg"},
	{0xC07DCAA80F96CFDAULL, "6040590_m.fxe"},
	{0xC07DDA929E72433DULL, "follower_banter6165435.cl"},
	{0xC07EC86BA7187253ULL, "6218660_m.fxe"},
	{0xC07EDD6CE6CEA505ULL, "llac_ht_keep_n_3732_19.dds"},
	{0xC07EE06F131711FCULL, "vxlt_undercitydungeon_yn_0_0.vlm"},
	{0xC07F500D94351D27ULL, "3516093.ogg"},
	{0xC07F675B2F6FC00DULL, "445109039.ogg"},
	{0xC0800777664A72C0ULL, "vxlt_gallowstemplar_n_45n_0_0.vlm"},
	{0xC0802890A67CD294ULL, "243803314.ogg"},
	{0xC0802EB7AC687507ULL, "gen00fl_aveline6129046.cl"},
	{0xC0812045BB3F83DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_reaction.gad"},
	{0xC08143B776EAA8C6ULL, "222125231.ogg"},
	{0xC08164B85C4156F2ULL, "vxht_0_hightown_cbj_0_0.vlm"},
	{0xC081B5831D83F418ULL, "mag100_cullen6148672.cl"},
	{0xC082085101780DDDULL, "157686116.ogg"},
	{0xC08210E3C66C5FC3ULL, "vxlt_warehouse_n_y6_0_0.vlm"},
	{0xC0824029853B084CULL, "119613557.ogg"},
	{0xC08260E6A53E38ECULL, "lla_ht_0_hightown_a2_-65534_0.dds"},
	{0xC082CA873665CB84ULL, "vxht_keep_69p_0_0.vlm"},
	{0xC082E32000724905ULL, "6126972_m.fxe"},
	{0xC083C0FD437B4E20ULL, "zz_start_mcr.utp"},
	{0xC084C1A51C2AA2DAULL, "dre000_varric6035157.cl"},
	{0xC084F7A47B939F2CULL, "art\\vfx\\models\\vfx_rockarmor_rh.mmh"},
	{0xC084F9C5529B103FULL, "dae200cr_viscount6087179.cl"},
	{0xC084FDA47B93A91EULL, "art\\vfx\\models\\vfx_rockarmor_rh.msh"},
	{0xC0850E4EAE2F2811ULL, "355383417.ogg"},
	{0xC085872BFC7E5C41ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_ian.gad"},
	{0xC085CFD912F9645AULL, "vxlt_gallowscourtyard_5gr_0_0.vlm"},
	{0xC0862CC346FF0318ULL, "145561855.ogg"},
	{0xC08634E926159D6CULL, "6171296_m.fxe"},
	{0xC0868ADD99F95DE9ULL, "6030428_m.fxe"},
	{0xC0871F45BB43EDDAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_reaction.ani"},
	{0xC088047050B7A8B9ULL, "hlac_ht_chantry_2534_39z.dds"},
	{0xC08858E58C2F9A58ULL, "follower_banter6160813.cl"},
	{0xC0885A2F94F641C0ULL, "vxht_keep_6cf_0_0.vlm"},
	{0xC0889E7B9E71FC05ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_monique.ani"},
	{0xC088C9773E0EE7EFULL, "ht_0_hightown_n_28_0.rml"},
	{0xC088E9ABF9B01504ULL, "vxlt_gallowstemplar_n_4ty_0_0.vlm"},
	{0xC08904C2FE191EE0ULL, "317297052.ogg"},
	{0xC0892CE599784BB8ULL, "798781639.ogg"},
	{0xC0893D70EE427ACAULL, "hlac_w_bonepit_393220_0y.dds"},
	{0xC08968C132E367D2ULL, "6161541_m.fxe"},
	{0xC08995D18264C5DCULL, "vxht_bartrand_n_5nr_0_0.vlm"},
	{0xC089B9E1A76B0217ULL, "mer320_hubert6082034.cl"},
	{0xC089CB411466A9CDULL, "vxht_chantry_n_30t_0_0.vlm"},
	{0xC089CC77F4422E7BULL, "vxlt_0_lowtown_docks_a3_czm_0_0.vlm"},
	{0xC089E65DB77CA923ULL, "vxw_blightlands_mc_0_0.vlm"},
	{0xC08A1DFC57FE04BBULL, "one300cr_crow_3.utc"},
	{0xC08A21005B0E2D09ULL, "lgt310_mag_melson_guards.bnk"},
	{0xC08A40F89CB23B9BULL, "6116181_f.fxe"},
	{0xC08A54E73BACD77FULL, "vxlt_hangedman_se_0_0.vlm"},
	{0xC08AC5303C962106ULL, "art\\vfx\\creatures\\vfx_c_corpse_c.cess.evt"},
	{0xC08AFAE6FC28A970ULL, "808331799.ogg"},
	{0xC08B2DE23F4E3FF7ULL, "art\\vfx\\combat\\vfx_g_spider_att_f2_c.mmh"},
	{0xC08B3D9AC4748D69ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junke_d.dds"},
	{0xC08BB8620EC9A22CULL, "hlac_ht_mansion_n_2995_5z.dds"},
	{0xC08C449F3807BFF6ULL, "qun211cr_elf_warrior6072377.cl"},
	{0xC08C71A1E7A50E2DULL, "6142318_m.fxe"},
	{0xC08C882BFC847CEFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_ian.ani"},
	{0xC08D138656ECE17EULL, "hightown_fem_noble1_sta6174598.cl"},
	{0xC08DB99635B623E7ULL, "454189952.ogg"},
	{0xC08E03005B11E349ULL, "lgt310_mag_melson_guards.cnv"},
	{0xC08E1D851BD70ED5ULL, "vxht_0_hightown_buq_0_0.vlm"},
	{0xC08EAD86E170F82AULL, "695719917.ogg"},
	{0xC08EC47B9ACAE972ULL, "hlac_lt_0_lowtown_docks_-131074_7z.dds"},
	{0xC08ECD87D5C8B7ABULL, "6201266_m.fxe"},
	{0xC08F0DCFBFCE26CAULL, "zz_followers_banter6157785.cl"},
	{0xC08F1B84CC776D00ULL, "vxht_mansion2b_1vn_0_0.vlm"},
	{0xC08FADCC7D812F9BULL, "art\\characters\\playercharacter\\humanfemale\\boo_wic.msh"},
	{0xC08FBFCC7D814EE6ULL, "art\\characters\\playercharacter\\humanfemale\\boo_wic.mao"},
	{0xC08FCBCC7D816285ULL, "art\\characters\\playercharacter\\humanfemale\\boo_wic.mmh"},
	{0xC091101F5A3CE302ULL, "vxca_caverns1_gkh_0_1.vlm"},
	{0xC0917B6ED162DDFCULL, "zz_followers_banter6157383.cl"},
	{0xC0918B898A4AEA72ULL, "688762229.ogg"},
	{0xC09190456B65D5BFULL, "6075651_f.fxe"},
	{0xC0919426B34E9AE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ledgedropland.evt"},
	{0xC091B76010FCDE95ULL, "vxca_caverns1_ijc_0_0.vlm"},
	{0xC091DC303C9C6751ULL, "art\\vfx\\creatures\\vfx_c_corpse_c.cess.gad"},
	{0xC0924492D08B6F45ULL, "w_sundermount_113_0.rml"},
	{0xC0926D8996D25A20ULL, "6171498_m.fxe"},
	{0xC09428ECE437BF66ULL, "6041741_f.fxe"},
	{0xC0942AEB525D7371ULL, "346808602.ogg"},
	{0xC094320607F03C92ULL, "vxlt_0_lowtown_a2_hfm_0_0.vlm"},
	{0xC0946D4667253C11ULL, "6065938_m.fxe"},
	{0xC09483E91AC440CEULL, "6102582_m.fxe"},
	{0xC094CCEE9023EA0BULL, "art\\vfx\\cinematic\\vfx_anders_healing_c.impa.ani"},
	{0xC094EF211DA0B82EULL, "vxlt_gallowstemplar_3on_0_0.vlm"},
	{0xC095101996FFDEF5ULL, "gen00fl_aveline6114871.cl"},
	{0xC095BADF07188F23ULL, "vxlt_0_lowtown_docks_b9o_0_1.vlm"},
	{0xC095DAD9F5155238ULL, "vxw_bonepit_1jy_0_0.vlm"},
	{0xC096220EE7D30EA3ULL, "dre000_bodahn6031324.cl"},
	{0xC0964778D5AA2D2AULL, "vxlt_undercitydungeon_17o_0_0.vlm"},
	{0xC0964A19BF07E7ECULL, "vxht_brothel_n_6bi_0_0.vlm"},
	{0xC0965DF2F02CA190ULL, "vxdr_0_deeproads_d2x_0_0.vlm"},
	{0xC0966D37DE594AFEULL, "vxca_cavern1_haunted_cjw_0_0.vlm"},
	{0xC096B0967AEE6380ULL, "vxlt_0_lowtown_a2_n_ghu_0_0.vlm"},
	{0xC0973AEA0D44A5B9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardc.mmh"},
	{0xC09744EA0D44B6A7ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardc.msh"},
	{0xC0976AFF4E2777C1ULL, "vxw_sundermount_23g_0_0.vlm"},
	{0xC097DD303CA0D4FFULL, "art\\vfx\\creatures\\vfx_c_corpse_c.cess.ani"},
	{0xC098112C04025F6FULL, "881178899.ogg"},
	{0xC0984BF9D1833AD0ULL, "103398046.ogg"},
	{0xC098583AA972C1A3ULL, "vxw_blightlands_vb_0_0.vlm"},
	{0xC098AB26B354E031ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ledgedropland.gad"},
	{0xC098DD834449D9B0ULL, "lt_0_lowtown_a2_42_0_ll.mmh"},
	{0xC099ACA228A77FDBULL, "vxca_cavern1_haunted_jfo_0_0.vlm"},
	{0xC09B6F44C1A95FF5ULL, "6165533_m.fxe"},
	{0xC09BCBDD39705877ULL, "6115755_f.fxe"},
	{0xC09BFBB9FA853FDAULL, "lt_0_lowtown_docks_n_216_0.mmh"},
	{0xC09BFFB9FA854611ULL, "lt_0_lowtown_docks_n_216_0.mao"},
	{0xC09C0DB9FA855E20ULL, "lt_0_lowtown_docks_n_216_0.msh"},
	{0xC09CCDCF5B30CB75ULL, "518761296.ogg"},
	{0xC09D2D548946F942ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_flat.msh"},
	{0xC09D2F548946FCF8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_flat.mmh"},
	{0xC09D30563F55B5DDULL, "6128390_m.fxe"},
	{0xC09D3B548947109BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_flat.mao"},
	{0xC09D3D85B528AC6FULL, "47467254.ogg"},
	{0xC09DB6C873DDC233ULL, "hlac_ht_mansion_n_3364_0y.dds"},
	{0xC09DF02691CA591FULL, "86603170.ogg"},
	{0xC09F081CDBEACE8CULL, "vxlt_0_lowtown_a2_n_c_f59_0_0.vlm"},
	{0xC09F290029399FCCULL, "dae000_worthy6166593.cl"},
	{0xC09F930FD8ED8BD2ULL, "vxht_mansion2b_26b_0_1.vlm"},
	{0xC09FAC26B35B00DFULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ledgedropland.ani"},
	{0xC09FDC16186946D6ULL, "6100760_m.fxe"},
	{0xC0A011BBBA314547ULL, "hlac_lt_0_lowtown_a3_n_c_1_5z.dds"},
	{0xC0A0293D02F2F715ULL, "vxca_caverns1_imb_0_0.vlm"},
	{0xC0A0812B14D449B9ULL, "mer141_jansen6031827.cl"},
	{0xC0A0A111FE732821ULL, "6097892_m.fxe"},
	{0xC0A0A27FFC3024D3ULL, "gen00fl_sebastian6174824.cl"},
	{0xC0A0E7A1B65C678DULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_noise_n.dds"},
	{0xC0A11A17FB3729D6ULL, "6083317_m.fxe"},
	{0xC0A1E3CE02B95565ULL, "ht_0_hightown_17_0_lc.mmh"},
	{0xC0A262CE86290CDEULL, "532949105.ogg"},
	{0xC0A2A3E2F9E65512ULL, "vxlt_0_lowtown_a2_hcn_0_0.vlm"},
	{0xC0A2B4EE902FFE82ULL, "art\\vfx\\cinematic\\vfx_anders_healing_c.impa.evt"},
	{0xC0A2C51EA0CC504EULL, "mag291_vincento6027661.cl"},
	{0xC0A2F44437505C94ULL, "vxlt_0_lowtown_docks_b7w_0_0.vlm"},
	{0xC0A35165BB8D253DULL, "168574781.ogg"},
	{0xC0A37C61812A780FULL, "character_character_p251.psh.fxo.dx11"},
	{0xC0A3C4D0F003C02FULL, "mcr370_templar26087644.cl"},
	{0xC0A41AACE10D2A05ULL, "6129555_m.fxe"},
	{0xC0A446CEFC3384FEULL, "art\\vfx\\creatures\\fxc_drg_brth_l_c.anb"},
	{0xC0A44CB6E70B6AB8ULL, "vxw_bonepit_1ov_0_0.vlm"},
	{0xC0A4F596DD30D1BFULL, "nat380_nathaniel6092522.cl"},
	{0xC0A53DE978E272C0ULL, "vxht_keep_6id_0_0.vlm"},
	{0xC0A5525391904321ULL, "vxlt_0_lowtown_a2_n_efh_0_0.vlm"},
	{0xC0A5738EEA9A91F1ULL, "6075276_m.fxe"},
	{0xC0A5AA9847D35BB4ULL, "t1_mub_wh3.tnt"},
	{0xC0A5DCDC401D9041ULL, "vxw_sundermount_24f_0_0.vlm"},
	{0xC0A5EC3275DDEB96ULL, "755945357.ogg"},
	{0xC0A5F93F918D4316ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfmid.msh"},
	{0xC0A6033F918D5414ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfmid.mmh"},
	{0xC0A6073F918D5B27ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfmid.mao"},
	{0xC0A61B7847D70074ULL, "gen00fl_sebastian6179912.cl"},
	{0xC0A72B727353F2D6ULL, "442167400.ogg"},
	{0xC0A74AD9EC748E1BULL, "6085344_f.fxe"},
	{0xC0A7BC22490E1FBCULL, "251260369.ogg"},
	{0xC0A7D45EAFFA9CECULL, "vxw_blightlands_9d_0_0.vlm"},
	{0xC0A8056EBB829434ULL, "hlac_lt_gamlenshouse_n_-1_0z.dds"},
	{0xC0A81063D417F893ULL, "one120_athenril6113265.cl"},
	{0xC0A83F0176B03863ULL, "432807179.ogg"},
	{0xC0A8A833E4F971E8ULL, "gen00fl_fenris6126704.cl"},
	{0xC0A8D7FD439A91D1ULL, "zz_start_mcr.ncs"},
	{0xC0A94B79956D21DAULL, "hightown_fem_courtesan_mob6168680.cl"},
	{0xC0A9B3EE90361B95ULL, "art\\vfx\\cinematic\\vfx_anders_healing_c.impa.gad"},
	{0xC0AA21E51E24ECFCULL, "778824484.ogg"},
	{0xC0AA30A774A4FCF4ULL, "l_env_wal_hightown_night_dist_oneshots.wbd"},
	{0xC0AA3DBA2B6670F7ULL, "6110855_f.fxe"},
	{0xC0AA53BFE8150779ULL, "703885020.ogg"},
	{0xC0AAE21AE8BE7D19ULL, "101589430.ogg"},
	{0xC0AB74BD54090B67ULL, "zz_qcr_debug6054935.cl"},
	{0xC0AC116561092393ULL, "1071980236.ogg"},
	{0xC0AC75CFB5C515FEULL, "vxht_0_hightown_a2_cbb_0_0.vlm"},
	{0xC0ACC5178E450874ULL, "vxht_mansion2b_1pm_0_1.vlm"},
	{0xC0ACD183AB54046DULL, "w_encounter2_camp_18_0.mmh"},
	{0xC0ACD583AB540ABEULL, "w_encounter2_camp_18_0.mao"},
	{0xC0ACE383AB5422F3ULL, "w_encounter2_camp_18_0.msh"},
	{0xC0AD0C68AC02C859ULL, "1006202781.ogg"},
	{0xC0AD0EA210206992ULL, "6150277_m.fxe"},
	{0xC0AD6AC16F9770D3ULL, "6075731_m.fxe"},
	{0xC0ADA8E3417D6FE4ULL, "454323508.ogg"},
	{0xC0ADA8ECD99F4494ULL, "vxdr_primevaltomb_8vf_0_0.vlm"},
	{0xC0ADBB478F7D74B3ULL, "hlac_lt_backalley_g_1_6z.dds"},
	{0xC0ADE8CA0611D9BBULL, "w_blightlands_178_0.msh"},
	{0xC0ADFACA0611F806ULL, "w_blightlands_178_0.mao"},
	{0xC0AE06CA06120CA5ULL, "w_blightlands_178_0.mmh"},
	{0xC0AE1D8C032C003DULL, "6048392_f.fxe"},
	{0xC0AE4916FB58CE75ULL, "vxht_mansion_n_4za_0_0.vlm"},
	{0xC0AE7F2E5B4C17DAULL, "hlac_ht_mansion2b_818_17z.dds"},
	{0xC0AEC7CF4E92B82FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v2.gad"},
	{0xC0AEF3044C5174E7ULL, "hlac_ca_cavern1_haunted_24426_3x.dds"},
	{0xC0AF159E94A1B8F7ULL, "w_bonepit_4_0_ll.mmh"},
	{0xC0AF49B44E5DE26AULL, "630969894.ogg"},
	{0xC0AF854CC6C45E82ULL, "framebuffer.gda"},
	{0xC0AF882C68666415ULL, "gen00fl_aveline6197207.cl"},
	{0xC0AF9944200862F0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberbasexa.mao"},
	{0xC0AF9D4420086903ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberbasexa.mmh"},
	{0xC0AFAB442008815DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberbasexa.msh"},
	{0xC0AFB7095349AFB8ULL, "llac_w_encounter2_camp_-65534_0.dds"},
	{0xC0AFB719D8F20079ULL, "follower_barks6154973.cl"},
	{0xC0AFE93A701FE125ULL, "117209724.ogg"},
	{0xC0AFF50F0780B273ULL, "934626777.ogg"},
	{0xC0B04ECE248DB02AULL, "llac_lt_0_lowtown_a2_n_-3_4.dds"},
	{0xC0B115BFEBDC6D92ULL, "vxlt_0_lowtown_a2_hho_0_0.vlm"},
	{0xC0B17079F9493615ULL, "336209167.ogg"},
	{0xC0B1707CD166A665ULL, "605469357.ogg"},
	{0xC0B17DA072AEA70CULL, "llac_lt_darktown_-1_28.dds"},
	{0xC0B195E289BA2591ULL, "vfx0_alt.psh.fxo"},
	{0xC0B1FC2C8D07EA49ULL, "gui_guipixelshader_p7.psh.fxo.dx11"},
	{0xC0B227FFC875E9F9ULL, "zz_followers_banter6157977.cl"},
	{0xC0B2BAFD43A2B496ULL, "zz_start_mcr.cnv"},
	{0xC0B2D128C0840B77ULL, "zz_followers_banter6159450.cl"},
	{0xC0B32DD3A2F418ECULL, "vxht_brothel_n_6ho_0_0.vlm"},
	{0xC0B33BDC62D4829FULL, "6155433_m.fxe"},
	{0xC0B394505EDA9480ULL, "vxlt_0_lowtown_a2_n_gbs_0_0.vlm"},
	{0xC0B3AA256767CDEBULL, "6055580_m.fxe"},
	{0xC0B4236B8D327276ULL, "vxlt_0_lowtown_docks_n_cp2_0_0.vlm"},
	{0xC0B43F6DE5A1F810ULL, "hightown_male_cityguard_mob6170057.cl"},
	{0xC0B46DEF2D5FA362ULL, "484683795.ogg"},
	{0xC0B47DDA5805ACAAULL, "hlac_lt_0_lowtown_-262144_2y.dds"},
	{0xC0B52AE95C4E6982ULL, "6118219_m.fxe"},
	{0xC0B56EEBFD7B7B59ULL, "vxw_sundermount_2kg_0_0.vlm"},
	{0xC0B58528516C171CULL, "pro000cs_gallows_arrival.cut"},
	{0xC0B5DCCF4E98FADCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v2.evt"},
	{0xC0B60847A0DC259EULL, "6174468_m.fxe"},
	{0xC0B614FEEB32601BULL, "mag160_feynriel6016568.cl"},
	{0xC0B64AEB731E77F0ULL, "532928766.ogg"},
	{0xC0B6A36B362FDF3BULL, "376272093.ogg"},
	{0xC0B6D183AB5C58E6ULL, "w_encounter2_camp_18_0.rml"},
	{0xC0B7E02692D53486ULL, "vxht_0_hightown_a2_cas_0_0.vlm"},
	{0xC0B7EFB9CE6C3A15ULL, "vxlt_warehouse_q1_0_0.vlm"},
	{0xC0B813F56BA809A5ULL, "hlac_lt_0_lowtown_docks_-65537_2x.dds"},
	{0xC0B8183BB2CCD971ULL, "6169506_m.fxe"},
	{0xC0B82492D0AC086EULL, "w_sundermount_113_0.mmh"},
	{0xC0B83092D0AC1C05ULL, "w_sundermount_113_0.mao"},
	{0xC0B83E92D0AC345CULL, "w_sundermount_113_0.msh"},
	{0xC0B8A47E2581CA3FULL, "follower_banter6160997.cl"},
	{0xC0B91E1B3BBD99F5ULL, "t3_rob_cot.tnt"},
	{0xC0B9B17568166627ULL, "vxlt_warehouse_10y_0_0.vlm"},
	{0xC0BA5403171C8881ULL, "6099829_m.fxe"},
	{0xC0BA54E77C116A6DULL, "6115866_m.fxe"},
	{0xC0BA5E38766B2211ULL, "399628804.ogg"},
	{0xC0BA7AABEDB41F5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_drunk.anb"},
	{0xC0BAA06840C2B407ULL, "6141511_m.fxe"},
	{0xC0BB13C95C82CDE0ULL, "6043063_m.fxe"},
	{0xC0BB2C6D84342EBEULL, "char_ext_test_night.arl"},
	{0xC0BB43A0805E20F3ULL, "dae000_ineria6136897.cl"},
	{0xC0BB700B7FB690E9ULL, "appearance.ncs"},
	{0xC0BB8F87621CD15AULL, "vxht_0_hightown_n_bnx_0_0.vlm"},
	{0xC0BC5EC4AC7355EAULL, "hlac_lt_0_lowtown_docks_a3_-131074_6z.dds"},
	{0xC0BD306EAA680D55ULL, "6102488_m.fxe"},
	{0xC0BDC95191796865ULL, "one110_meeran6089487.cl"},
	{0xC0BDCA83AB626B2AULL, "w_encounter2_camp_18_0.phy"},
	{0xC0BE0CF4DCA65183ULL, "art\\vfx\\combat\\vfx_r_ibl_disappear_imp_p.impa.evt"},
	{0xC0BE2B77EC22AD9FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\varic_damage_reactions.anb"},
	{0xC0BEDA1969DB731EULL, "981996402.ogg"},
	{0xC0BF21E326C85B15ULL, "6084921_f.fxe"},
	{0xC0BF2D0E180F3637ULL, "vxht_0_hightown_n_cam_0_0.vlm"},
	{0xC0BF54EF5CA45857ULL, "vxw_bonepitmining_a2_cp_0_0.vlm"},
	{0xC0C06CE801D0D0E1ULL, "ht_0_hightown_a2_27_0.rml"},
	{0xC0C0762E77E13E7AULL, "mag211_marethari6021338.cl"},
	{0xC0C09E442016DE28ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberbasexa.phy"},
	{0xC0C0C8AD6E074E1CULL, "6057347_m.fxe"},
	{0xC0C0D82F4D8B1CFAULL, "823828971.ogg"},
	{0xC0C101BDC37C0A3FULL, "vxht_brothel_n_6bf_0_0.vlm"},
	{0xC0C11FA7585214F0ULL, "llac_lt_backalley_n_65536_0.dds"},
	{0xC0C12F300CCB61FBULL, "186476382.ogg"},
	{0xC0C1BA79BB6565DAULL, "6189675_m.fxe"},
	{0xC0C23F3E5C077A97ULL, "162416715.ogg"},
	{0xC0C25F41E08EC7B2ULL, "6065941_f.fxe"},
	{0xC0C2A7F5446D351CULL, "occlusionquery.psh.fxo"},
	{0xC0C2F327A8203E8AULL, "6061274_m.fxe"},
	{0xC0C32E4BBB99306EULL, "6101231_m.fxe"},
	{0xC0C335ADD02C18E5ULL, "vxca_caverns1_ij3_0_0.vlm"},
	{0xC0C3407EA8388F4DULL, "749995103.ogg"},
	{0xC0C36FC1BFBA4962ULL, "isa311_isabela6185244.cl"},
	{0xC0C384B0FC5932CBULL, "6142068_m.fxe"},
	{0xC0C3D9875F10AF21ULL, "vxlt_warehouse_v8_0_0.vlm"},
	{0xC0C3E0C8EF7193D9ULL, "vxw_sundermount_2lf_0_0.vlm"},
	{0xC0C3E4CF4EA545F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v2.ani"},
	{0xC0C4E5F4DCAC2E04ULL, "art\\vfx\\combat\\vfx_r_ibl_disappear_imp_p.impa.gad"},
	{0xC0C5759FDD2B54B3ULL, "aip_revenant_base.gda"},
	{0xC0C59F462923E438ULL, "vxca_caverns1_efr_0_0.vlm"},
	{0xC0C5DA58A6288DECULL, "vxw_sundermount_26y_0_0.vlm"},
	{0xC0C623343FEB45DFULL, "llac_ht_bartrand_n_3371_2.dds"},
	{0xC0C66196C0625295ULL, "vxlt_warehouse_z0_0_0.vlm"},
	{0xC0C686263FF70E2CULL, "822307475.ogg"},
	{0xC0C75209BC31A5D2ULL, "vxca_caverns1_mines_i2q_0_0.vlm"},
	{0xC0C7834F224833E4ULL, "vxlt_backalley_n_203_0_0.vlm"},
	{0xC0C794D911694539ULL, "vxlt_0_lowtown_docks_a3_n_czx_0_0.vlm"},
	{0xC0C7D3D123EA1903ULL, "124174336.ogg"},
	{0xC0C852C481145580ULL, "6151171_m.fxe"},
	{0xC0C8C6C46E0782EDULL, "6110766_m.fxe"},
	{0xC0C8E52FCA7578D3ULL, "hlac_dr_primevaltomb_664_5z.dds"},
	{0xC0C8E6B1941401D3ULL, "vxht_mansion2b_3bg_0_0.vlm"},
	{0xC0C9D9E7498E8674ULL, "mag320cr_nobleman_2.utc"},
	{0xC0C9F78972BC018DULL, "follower_banter6153997.cl"},
	{0xC0CA01645412E9DAULL, "vxht_0_hightown_n_biy_0_0.vlm"},
	{0xC0CA9C46EE36F42BULL, "hlac_lt_0_lowtown_docks_-2_6z.dds"},
	{0xC0CAB3C8E0B2DF46ULL, "vxlt_0_lowtown_a3_n_c_hje_0_0.vlm"},
	{0xC0CAC3F3B7626B7AULL, "736405658.ogg"},
	{0xC0CB0A00B21B5100ULL, "6164874_m.fxe"},
	{0xC0CBF4F4DCB266E2ULL, "art\\vfx\\combat\\vfx_r_ibl_disappear_imp_p.impa.ani"},
	{0xC0CC8D180418C5ADULL, "mag170_hugh6080757.cl"},
	{0xC0CCA3DEF21CFF53ULL, "llac_lt_0_lowtown_-1_0.dds"},
	{0xC0CD8348B5CE3746ULL, "vxw_encounter2_r0_0_0.vlm"},
	{0xC0CD95F7864E0E7CULL, "hairalpha_hair_p5.psh.fxo.dx11"},
	{0xC0CDA41D7B3E2F6DULL, "6147667_m.fxe"},
	{0xC0CDC6CC4E9A70D7ULL, "vxw_bonepitmining_a2_ns_0_0.vlm"},
	{0xC0CE0202B435B27FULL, "vxw_encounter2_qy_0_1.vlm"},
	{0xC0CEBD7435C90D76ULL, "30172058.ogg"},
	{0xC0CF1B569E9F06C8ULL, "lgt260_eau_tmp_asst_harbor6143631.cl"},
	{0xC0CF28D9A604D289ULL, "584430573.ogg"},
	{0xC0D01CCBCAB9247AULL, "hlac_ca_caverns1_24424_27x.dds"},
	{0xC0D0A7EA67728CA1ULL, "vxlt_0_lowtown_a2_n_eqm_0_0.vlm"},
	{0xC0D0FCDC8611DD1EULL, "vxht_mansion2_326_0_0.vlm"},
	{0xC0D110A7EF9223FDULL, "6173361_m.fxe"},
	{0xC0D15F72C336345BULL, "gen00fl_dog6167712.cl"},
	{0xC0D17F58CC212857ULL, "867179786.ogg"},
	{0xC0D18DB635B0143BULL, "1035913755.ogg"},
	{0xC0D197C9409EE802ULL, "vxlt_0_lowtown_docks_a3_d0l_0_1.vlm"},
	{0xC0D1A78AC2223165ULL, "vxca_caverns1_im2_0_0.vlm"},
	{0xC0D1B06ABE416CC7ULL, "llac_dr_0_deeproads_5842_2.dds"},
	{0xC0D20EA3403A9A82ULL, "6112019_m.fxe"},
	{0xC0D2698FB94C52BCULL, "gen_im_arm_cht_come.uti"},
	{0xC0D2822BE820485DULL, "zz_followers_banter6157825.cl"},
	{0xC0D2CA5001887437ULL, "170471920.ogg"},
	{0xC0D36FADD73BB954ULL, "366938933.ogg"},
	{0xC0D3BB5B8E295664ULL, "vxht_bartrand_n_4hu_0_0.vlm"},
	{0xC0D3EC84CB0C95F1ULL, "mag121_idunna6083959.cl"},
	{0xC0D40FDD820ED972ULL, "vxlt_0_lowtown_n_ex9_0_0.vlm"},
	{0xC0D42E8ACA69C16AULL, "6085166_m.fxe"},
	{0xC0D4409F88B93303ULL, "872222954.ogg"},
	{0xC0D4F950445B0D73ULL, "497931534.ogg"},
	{0xC0D55E79737FAFDCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_collapse_help_up_bethany.ani"},
	{0xC0D607869F5B4173ULL, "6109285_m.fxe"},
	{0xC0D61EA774CA188EULL, "l_env_wal_hightown_night_dist_oneshots.bnk"},
	{0xC0D6A4CC7DBD2332ULL, "art\\characters\\playercharacter\\humanfemale\\boo_wic.phy"},
	{0xC0D6AF681598487EULL, "489115558.ogg"},
	{0xC0D6E3948F436DBBULL, "genip_node_i3.utp"},
	{0xC0D75878EA6923EDULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic.mmh"},
	{0xC0D75C78EA692A3EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic.mao"},
	{0xC0D76A78EA694273ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic.msh"},
	{0xC0D7B583ABDB6D35ULL, "891726172.ogg"},
	{0xC0D817045535BF13ULL, "zz_ser_debug6154334.cl"},
	{0xC0D84A4B73C9A471ULL, "vxw_bonepitmining_a2_220_0_1.vlm"},
	{0xC0D8BDE577B55EB3ULL, "vxlt_0_lowtown_docks_n_9rc_0_0.vlm"},
	{0xC0D9F6D583FBFCF2ULL, "ser191_emeric6108136.cl"},
	{0xC0D9FB1BE605E838ULL, "842287131.ogg"},
	{0xC0DA0A7258DA39E1ULL, "vxht_chantry_26l_0_0.vlm"},
	{0xC0DAFA4173320485ULL, "vxht_0_hightown_atm_0_0.vlm"},
	{0xC0DB5ED05A10AB00ULL, "ico_helm_heavy_3.dds"},
	{0xC0DB6B797384311EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_collapse_help_up_bethany.gad"},
	{0xC0DB95716A5B01C0ULL, "plt_lgt000pt_civ_spider.nss"},
	{0xC0DBD591F2312620ULL, "llac_ca_caverns2_5802_50.dds"},
	{0xC0DC51CD50B1B748ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_66x2.mmh"},
	{0xC0DC6AF9C1FB6871ULL, "grass_dp.dds"},
	{0xC0DC704E83091C79ULL, "vxlt_0_lowtown_docks_bdz_0_1.vlm"},
	{0xC0DCAC0116027189ULL, "hlac_lt_0_lowtown_docks_a3_0_1y.dds"},
	{0xC0DCB13DF006D9AAULL, "hlac_lt_0_lowtown_a2_n_c_-131073_2z.dds"},
	{0xC0DCB65B696B91CFULL, "zz_vrc_debug6200175.cl"},
	{0xC0DCF4FEC2E1765FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkc_sa.dds"},
	{0xC0DDCA80DD934EB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_right_walk.gad"},
	{0xC0DE2EB3CB6FFDC1ULL, "vxw_encounter2_52_0_0.vlm"},
	{0xC0DE7F4494DED001ULL, "vxht_mansion2b_1bo_0_0.vlm"},
	{0xC0DE9A6047A182A5ULL, "86809734.ogg"},
	{0xC0E0C482D35DC4D9ULL, "vxw_sundermount_2zd_0_0.vlm"},
	{0xC0E152BE0578D19EULL, "llac_lt_0_lowtown_a2_n_1_5.dds"},
	{0xC0E1F645B498AAB6ULL, "vxlt_warehouse_av_0_0.vlm"},
	{0xC0E24679738A113DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sister_collapse_help_up_bethany.evt"},
	{0xC0E28C7F7B618E0EULL, "vxht_0_hightown_a3_c70_0_0.vlm"},
	{0xC0E28D08DD4EB7B5ULL, "6064911_m.fxe"},
	{0xC0E37253DE0EACC0ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_1z.dds"},
	{0xC0E37F72CD8A302BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ai_transition.evt"},
	{0xC0E3BAA36CE7B2B4ULL, "mag131_taraneh6088703.cl"},
	{0xC0E3DF80DD97DE44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_right_walk.evt"},
	{0xC0E428359267F255ULL, "vxdr_primevalentrance_188_0_0.vlm"},
	{0xC0E49AF5BFC46DBCULL, "hlac_dr_dungeon_3216_2z.dds"},
	{0xC0E4B304DFE6DC6BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_int_atlas2_n.dds"},
	{0xC0E4BBF8B52504B1ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.cess.ani"},
	{0xC0E5D5DA8F90E777ULL, "6187702_m.fxe"},
	{0xC0E6268E12CA1F94ULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_1x.dds"},
	{0xC0E66B3D94362D19ULL, "vxht_mansion2b_2v4_0_0.vlm"},
	{0xC0E70F9D701E892DULL, "6076173_f.fxe"},
	{0xC0E71CBE2E59906EULL, "vxht_mansion2_1qq_0_0.vlm"},
	{0xC0E77B5E711D2D28ULL, "hlac_w_bonepitdestroyed_a3_327684_0x.dds"},
	{0xC0E7C0AC85E89232ULL, "580847895.ogg"},
	{0xC0E7DC2F375BA970ULL, "hlac_ht_keep_4519_15z.dds"},
	{0xC0E85178EA778AAAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic.phy"},
	{0xC0E908A54EA526DAULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.anb"},
	{0xC0E95E5C89C1C0CEULL, "vxht_0_hightown_a3_ani_0_0.vlm"},
	{0xC0E96E078D9B7465ULL, "vxlt_warehouse_qa_0_0.vlm"},
	{0xC0E9BD04B941ABEFULL, "219353855.ogg"},
	{0xC0E9CEC693EF49DFULL, "6070212_f.fxe"},
	{0xC0E9EA7651D90262ULL, "vxca_caverns1_mines_hue_0_0.vlm"},
	{0xC0EA4E0943676A74ULL, "762838228.ogg"},
	{0xC0EA5293EB906FC5ULL, "hlac_ht_0_hightown_a2_2_0x.dds"},
	{0xC0EA8C4EDE06A6FDULL, "mag311_nyssa.bnk"},
	{0xC0EA8DADED1D0ED1ULL, "905505661.ogg"},
	{0xC0EA9872CD9079ECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ai_transition.gad"},
	{0xC0EA9F66304935C8ULL, "hf_mer_cot_lt.mor"},
	{0xC0EAB757BF2CE9DFULL, "6165409_m.fxe"},
	{0xC0EABCD27D8E3192ULL, "ser221_gascard6134528.cl"},
	{0xC0EAC2AFD2FE67A5ULL, "vxht_mansion2b_1hq_0_0.vlm"},
	{0xC0EADB52E49D1147ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgtp_0d.dds"},
	{0xC0EB1FAC52C243A6ULL, "1033621231.ogg"},
	{0xC0EB6B4C998D63C0ULL, "lla_ht_0_hightown_a3_131073_0.dds"},
	{0xC0EB6E6923932F07ULL, "gen00fl_isabela6099476.cl"},
	{0xC0EBC6F8B52B3653ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.cess.gad"},
	{0xC0EBF0714667525DULL, "gen00fl_merrill6121760.cl"},
	{0xC0EC37054FB6B1C5ULL, "6122738_m.fxe"},
	{0xC0EC3E5E9EFF6D88ULL, "zz_followers_banter6159345.cl"},
	{0xC0ED08C28925856CULL, "6126376_f.fxe"},
	{0xC0EDC2D7C87E254AULL, "vxht_mansion2b_21n_0_0.vlm"},
	{0xC0EDE6C83A176C81ULL, "and100_anders6114236.cl"},
	{0xC0EDFA2171B0D67FULL, "wm_ico_ambush.dds"},
	{0xC0EE6E4EDE0A5C05ULL, "mag311_nyssa.cnv"},
	{0xC0EE88F74E7A7DA8ULL, "zz_fen_debug6134396.cl"},
	{0xC0EEE4A7738DD0B5ULL, "490456276.ogg"},
	{0xC0EEF024C1D6D2CBULL, "vxw_woundedcoast_1ep_0_0.vlm"},
	{0xC0EF262E22A5E664ULL, "character_character_p258.psh.fxo.dx11"},
	{0xC0F0393CAEA72AA1ULL, "6051999_m.fxe"},
	{0xC0F091FCAA9CB03FULL, "vxw_sundermount_26v_0_0.vlm"},
	{0xC0F122EB9345FC8FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\_textures\\prp_cot_n.dds"},
	{0xC0F14AD638B11D8DULL, "115405391.ogg"},
	{0xC0F180619FE4CEFDULL, "95015209.ogg"},
	{0xC0F19772CD9696EAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ai_transition.ani"},
	{0xC0F1C3F8B52F9CB8ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.cess.evt"},
	{0xC0F1F780DDA4443DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_right_walk.ani"},
	{0xC0F2D98EA3F5D513ULL, "332131145.ogg"},
	{0xC0F2DE2BE8F09EAFULL, "vxlt_0_lowtown_docks_b66_0_1.vlm"},
	{0xC0F32B1FF7ABCAB6ULL, "w_encounter3_52_0.msh"},
	{0xC0F3351FF7ABDBB4ULL, "w_encounter3_52_0.mmh"},
	{0xC0F3391FF7ABE2C7ULL, "w_encounter3_52_0.mao"},
	{0xC0F374E525DE392AULL, "71804628.ogg"},
	{0xC0F45F038C1375D0ULL, "6104832_m.fxe"},
	{0xC0F4752B2383289DULL, "6083817_m.fxe"},
	{0xC0F49991E1701E0AULL, "361983619.ogg"},
	{0xC0F50BB99BDDDFA9ULL, "vxlt_0_lowtown_docks_a3_n_d6w_0_0.vlm"},
	{0xC0F50F00CF928AE7ULL, "vxw_bonepitdestroyed_a3_26i_0_0.vlm"},
	{0xC0F50FB45166ADADULL, "570758391.ogg"},
	{0xC0F58E9B204FA8EEULL, "vxht_mansion2_1tr_0_0.vlm"},
	{0xC0F5DCA8B53D30DCULL, "vxht_mansion2b_1v2_0_0.vlm"},
	{0xC0F6997D4418F0D4ULL, "vxlt_gallowstemplar_f_3rh_0_0.vlm"},
	{0xC0F6DD2AAF2D7B12ULL, "vrc321_varric6154762.cl"},
	{0xC0F7D0397BB7D94EULL, "vxht_0_hightown_a3_akj_0_0.vlm"},
	{0xC0F8920D87AA94B0ULL, "6151218_m.fxe"},
	{0xC0F9283FCFC07BF3ULL, "dae021_sister_samea_sta6142113.cl"},
	{0xC0F93BEA0D9771A6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashshardc.phy"},
	{0xC0F94AD4139169FAULL, "qun211_party_comment6072266.cl"},
	{0xC0F9592135E9A6B7ULL, "lowtown_male_alienage1_sta6162757.cl"},
	{0xC0F97B482504779DULL, "6294080.ogg"},
	{0xC0FA0BD6B48A8B02ULL, "vxht_0_hightown_c1c_0_0.vlm"},
	{0xC0FA8DAB634AED9DULL, "w_bonepitmining_a2_215_0.mao"},
	{0xC0FA91AB634AF4D6ULL, "w_bonepitmining_a2_215_0.mmh"},
	{0xC0FA9BAB634B05D4ULL, "w_bonepitmining_a2_215_0.msh"},
	{0xC0FAA1F13B873938ULL, "vxlt_0_lowtown_docks_a3_cs1_0_0.vlm"},
	{0xC0FAC3C98A153E6AULL, "6102704_m.fxe"},
	{0xC0FB62FE78CB0101ULL, "vxht_mansion2b_1xm_0_0.vlm"},
	{0xC0FB8B3B23E278ADULL, "908506272.ogg"},
	{0xC0FBE502DC520F9FULL, "hlac_lt_0_lowtown_a2_n_65535_2y.dds"},
	{0xC0FC182A97E77CB5ULL, "6036703_m.fxe"},
	{0xC0FC252AE3049E99ULL, "vxht_bartrand_n_63c_0_0.vlm"},
	{0xC0FC5600863ED852ULL, "al_rogue_s01.gda"},
	{0xC0FCAC7E185ED3DAULL, "vxw_encounter3_7h_0_0.vlm"},
	{0xC0FD951A3A546322ULL, "6020431_m.fxe"},
	{0xC0FDA2D8AD827AE9ULL, "6030679_m.fxe"},
	{0xC0FED9FF9884DBB6ULL, "vxlt_warehouse_ox_0_0.vlm"},
	{0xC0FF19FED4BAB17EULL, "dae000ip_crowd_em3_talk11.utp"},
	{0xC0FF3202CBA634D4ULL, "48201861.ogg"},
	{0xC0FFDD6BEF8F91E5ULL, "gen00fl_anders6119170.cl"},
	{0xC100476669A1C8A8ULL, "vxht_mansion_n_4x2_0_0.vlm"},
	{0xC1004D37235469DFULL, "zz_followers_banter6159528.cl"},
	{0xC1004F85B151540CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_2.pwk"},
	{0xC1006548214D8515ULL, "mer240_jansen6050389.cl"},
	{0xC102023F91DBB10FULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfmid.phy"},
	{0xC1024BC101E8F086ULL, "261720307.ogg"},
	{0xC1028C2599B38695ULL, "269499158.ogg"},
	{0xC102B3604FD11F24ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channel.msh"},
	{0xC102B575B921263EULL, "vxlt_0_lowtown_a3_n_c_gmk_0_0.vlm"},
	{0xC102C5604FD13D8DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channel.mao"},
	{0xC102C9604FD144C6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_channel.mmh"},
	{0xC102CF7F2306D511ULL, "vxht_0_hightown_n_50n_0_0.vlm"},
	{0xC102F37A0CE8D140ULL, "zz_mag_debug6054584.cl"},
	{0xC102FCAC3DE23DCAULL, "follower_banter6161306.cl"},
	{0xC10380DDC188A367ULL, "vxw_bonepitdestroyed_a3_21h_0_0.vlm"},
	{0xC103E7A3E08A2AA8ULL, "vxlt_gallowsprison_n_6m0_0_0.vlm"},
	{0xC103EAFBDA51E76CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_r_open.ani"},
	{0xC1047C2245F22AB6ULL, "hlac_ca_caverns1_24426_9z.dds"},
	{0xC1050E14FBAA9B29ULL, "getvaultflag.ncs"},
	{0xC10537D72BA7AB5CULL, "703198235.ogg"},
	{0xC1054647A0ADA295ULL, "vxca_caverns1_i2k_0_0.vlm"},
	{0xC1054F17BD8FB15DULL, "vxw_encounter2_p4_0_1.vlm"},
	{0xC1056268558537E3ULL, "6095845_m.fxe"},
	{0xC105E8FCB578677EULL, "70632727.ogg"},
	{0xC106A2A4AC679F02ULL, "292560361.ogg"},
	{0xC106A86C555E6B65ULL, "6116214_m.fxe"},
	{0xC106CD39EA143D99ULL, "gen_im_trs_gem_jade_01.uti"},
	{0xC106D891B15AE4C7ULL, "107948936.ogg"},
	{0xC10718F8164B3AD2ULL, "mcr310_loyalist_mages6076863.cl"},
	{0xC10748DAEFDB9FC6ULL, "348004045.ogg"},
	{0xC1076F61E451D4A1ULL, "w_woundedcoast_52_0.rml"},
	{0xC107940CF6E949F9ULL, "vxca_cavern1_haunted_i1j_0_1.vlm"},
	{0xC107EE72DAD0E9F2ULL, "qun200_party_comment6095407.cl"},
	{0xC1081972E7FB740BULL, "515918446.ogg"},
	{0xC1083C72328D4A3CULL, "561456517.ogg"},
	{0xC1083D43609A9C02ULL, "39441410.ogg"},
	{0xC1085E10D4C107DDULL, "176133518.ogg"},
	{0xC108CD247F9778F7ULL, "gen_im_wep_mag_sta_act2_01.uti"},
	{0xC108E8C7E1F5873EULL, "lla_ht_0_hightown_n_-196608_0.dds"},
	{0xC1091F38A9DDEB1AULL, "lgt210_eau_symp_templar6154721.cl"},
	{0xC10935A67C0B56EAULL, "6107804_m.fxe"},
	{0xC1094FBAFB438022ULL, "6070236_m.fxe"},
	{0xC109AEBD9122AEA7ULL, "227327433.ogg"},
	{0xC109D4DB6AC11981ULL, "vxht_mansion2b_1sl_0_0.vlm"},
	{0xC109F7FBDA5668AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_r_open.gad"},
	{0xC10A181816F24579ULL, "vxca_cavern1_haunted_iys_0_0.vlm"},
	{0xC10A1C0161489009ULL, "hlac_ca_caverns1_mines_24423_0z.dds"},
	{0xC10ABC73C45FAE48ULL, "llac_lt_undercitydungeon_-1_5.dds"},
	{0xC10C12B9259680BBULL, "vxw_bonepitdestroyed_a3_lg_0_0.vlm"},
	{0xC10C34F66D5FB362ULL, "llac_dr_primevaltomb_657_0.dds"},
	{0xC10C5B7FE87011F0ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown2.mmh"},
	{0xC10C697FE870290AULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_lowtown2.msh"},
	{0xC10CB119EC0A1693ULL, "6097879_f.fxe"},
	{0xC10CD31A60187819ULL, "llac_dr_0_deeproads_5839_2.dds"},
	{0xC10D4BDC8A7AF436ULL, "vxlt_warehouse_py_0_0.vlm"},
	{0xC10D7C6BF0F95639ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02brgdebris_1.mao"},
	{0xC10E50457DCF7809ULL, "6152359_m.fxe"},
	{0xC10E54267AB45879ULL, "fen210_orana6133040.cl"},
	{0xC10F8B870C1ABAFEULL, "133391907.ogg"},
	{0xC10FFE00DE7D6959ULL, "239431366.ogg"},
	{0xC110F2FBDA5C7E8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\fingers_r_open.evt"},
	{0xC1110BA54EC67D8DULL, "art\\vfx\\combat\\vfx_c_rw_vulnerable_c.mmh"},
	{0xC11168E1720F46E8ULL, "6056619_f.fxe"},
	{0xC111AC1532DF5FD1ULL, "aip_mercenarywarrior2w_base.gda"},
	{0xC1127255043BD9EEULL, "vxht_mansion2_1bt_0_0.vlm"},
	{0xC112A1D77FA0362DULL, "78672548.ogg"},
	{0xC113D6C865CF20FCULL, "508909232.ogg"},
	{0xC1146CA6D417987DULL, "vxlt_backalley_n_1e6_0_0.vlm"},
	{0xC1149ADC83F7293DULL, "vxht_0_hightown_a3_a6f_0_0.vlm"},
	{0xC114D390C136F95FULL, "hlac_w_woundedcoast_65544_0x.dds"},
	{0xC114D5A2E8A7244EULL, "llac_lt_gallowstemplar_2391_7.dds"},
	{0xC1150967EA1E80EDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagkeep_n.dds"},
	{0xC1151D48DEC2045DULL, "179309385.ogg"},
	{0xC1154078DA861054ULL, "llac_lt_0_lowtown_docks_a3_262142_0.dds"},
	{0xC11713A80BB376F6ULL, "qun160_party_comment_16061230.cl"},
	{0xC11797E009B0D76BULL, "ser121_jethann6107116.cl"},
	{0xC117CA48C947435BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowerb.mmh"},
	{0xC117D648C94757F8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowerb.mao"},
	{0xC117E848C9477645ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowerb.msh"},
	{0xC11846B85CB73201ULL, "vxht_mansion2b_1vc_0_0.vlm"},
	{0xC1184983BD8C9441ULL, "vxw_blightlands_m9_0_0.vlm"},
	{0xC118660E125A4806ULL, "hlac_ht_mansion2b_818_8x.dds"},
	{0xC11876407BA6020CULL, "6147891_m.fxe"},
	{0xC118B77A35DD929FULL, "615188109.ogg"},
	{0xC11985E7E2FCC7B5ULL, "vrc321_party_barks6096318.cl"},
	{0xC1198E716A905683ULL, "plt_lgt000pt_civ_spider.plo"},
	{0xC11A4FBCA3B92943ULL, "hlac_w_sundermount_-327678_0z.dds"},
	{0xC11A96AC02F1DD91ULL, "vxw_encounter1_bb_0_0.vlm"},
	{0xC11AEFB4D0FA2B6BULL, "6133763_m.fxe"},
	{0xC11B95C3737DAD69ULL, "vrc321_varric6087085.cl"},
	{0xC11BB1DAE4DEB471ULL, "206533912.ogg"},
	{0xC11C3970B7704A04ULL, "6071282_m.fxe"},
	{0xC11C5053ACADA9ABULL, "ser221_party_comments6122878.cl"},
	{0xC11CF909A0D908ABULL, "vxlt_0_lowtown_n_bxk_0_0.vlm"},
	{0xC11D492B423CA47FULL, "136042146.ogg"},
	{0xC11D4F6349662B44ULL, "hlac_w_bonepitmining_a2_262150_0x.dds"},
	{0xC11E09F7B5FB7E43ULL, "112336524.ogg"},
	{0xC11E963EB30B7AF0ULL, "dae300_bran6143458.cl"},
	{0xC11F3363941B13E8ULL, "6042635_m.fxe"},
	{0xC11F434096A47ADBULL, "vxlt_backalley_g_1ue_0_0.vlm"},
	{0xC11F89563912CD14ULL, "vxw_woundedcoast_mn_0_0.vlm"},
	{0xC11FDBCA0B2E2881ULL, "vxlt_0_lowtown_a2_n_ghx_0_0.vlm"},
	{0xC12032B15C0E8F19ULL, "vxht_mansion2b_2b8_0_0.vlm"},
	{0xC1208217A7D07E19ULL, "llac_lt_0_lowtown_n_65533_3.dds"},
	{0xC12084AA2F946EB8ULL, "hlac_lt_0_lowtown_a2_n_c_131071_2y.dds"},
	{0xC121C34979D9C8FFULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallgreen.phy"},
	{0xC1221D1558136544ULL, "w_blightlands_148_0.mmh"},
	{0xC122211558136B97ULL, "w_blightlands_148_0.mao"},
	{0xC12223267BE2BB82ULL, "follower_banter6143287.cl"},
	{0xC12226777402AF3FULL, "genip_exploding_crate_lg_rd.utp"},
	{0xC122331558138AE6ULL, "w_blightlands_148_0.msh"},
	{0xC1224CD60DBAD0ECULL, "448529464.ogg"},
	{0xC122531127ABEC50ULL, "6217129_m.fxe"},
	{0xC122E5DD0297E0AEULL, "vxht_0_hightown_a2_aui_0_0.vlm"},
	{0xC1237941388EAA81ULL, "vxdr_primevalentrance_23s_0_1.vlm"},
	{0xC1238334FF715842ULL, "hlac_ca_caverns2prc_5802_58y.dds"},
	{0xC1240F0FAAFB40C8ULL, "hlac_lt_0_lowtown_docks_n_-3_3z.dds"},
	{0xC12423141A1E2681ULL, "vxca_caverns2_5a8_0_1.vlm"},
	{0xC124A3BC78BDC4E4ULL, "primitives_primitive_p2.vsh.fxo.dx11"},
	{0xC124E2955CEC1489ULL, "vxht_chantry_33j_0_0.vlm"},
	{0xC125565A85CC0985ULL, "gen_im_npc_wep_two_2hma.uti"},
	{0xC1257D75FF5E38ACULL, "qun200_party_comment6095364.cl"},
	{0xC125A594877B1F1AULL, "vxht_keep_n_67j_0_1.vlm"},
	{0xC125DE8281DD60C3ULL, "mag101_grace6053087.cl"},
	{0xC125E1A4F563E567ULL, "6118571_m.fxe"},
	{0xC125E8646FA6DC72ULL, "vxht_0_hightown_a2_577_0_0.vlm"},
	{0xC126631C627AB9DAULL, "vxlt_0_lowtown_a2_ewo_0_0.vlm"},
	{0xC126BB60AF82ACC1ULL, "vxw_blightlands_v8_0_0.vlm"},
	{0xC126C06DA57B8909ULL, "848666115.ogg"},
	{0xC126E81D6D9C1A8CULL, "6140991_m.fxe"},
	{0xC126F37FFF637D83ULL, "111893783.ogg"},
	{0xC12743C6A29D7151ULL, "6086191_m.fxe"},
	{0xC127B05E90A434F4ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_act_c.mmh"},
	{0xC127E927862B1E98ULL, "818640676.ogg"},
	{0xC12823D03683C286ULL, "gen00fl_dog6170610.cl"},
	{0xC1283384606119EEULL, "300550238.ogg"},
	{0xC1287C3B91C1C451ULL, "vfx_vfx_p8.vsh.fxo.dx11"},
	{0xC12882B8A8315805ULL, "llac_lt_gallowstemplar_n_2382_1.dds"},
	{0xC128B79889AF34CBULL, "vxw_woundedcoast_1qt_0_0.vlm"},
	{0xC128B959B94C95C8ULL, "6171894_m.fxe"},
	{0xC1290888F4E7F611ULL, "vxw_encounter1_ga_0_0.vlm"},
	{0xC1292CBDD16EC3D3ULL, "6149418_m.fxe"},
	{0xC1297681568C5D55ULL, "one100dg_vanard6173966.cl"},
	{0xC129D7AF730BC799ULL, "one100dg_nabil6059783.cl"},
	{0xC129F694288CDBFBULL, "gen00fl_aveline6139704.cl"},
	{0xC12A030CD0457D7AULL, "llac_dr_0_deeproads_5845_15.dds"},
	{0xC12A2311549E836DULL, "mcr310_argument6072273.cl"},
	{0xC12A248612EA75CDULL, "vxht_brothel_n_32u_0_0.vlm"},
	{0xC12AA2C1334B45ECULL, "863745898.ogg"},
	{0xC12AE1579E019B69ULL, "hlac_ht_chantry_2534_46x.dds"},
	{0xC12B257CC896FA5FULL, "993538868.ogg"},
	{0xC12B3C97921FE32BULL, "hlac_lt_0_lowtown_docks_a3_n_-262143_0z.dds"},
	{0xC12B6162DD5ADA58ULL, "185047203.ogg"},
	{0xC12C111660A34504ULL, "892224221.ogg"},
	{0xC12C59C75DE0E5B4ULL, "vxht_keep_n_6ag_0_0.vlm"},
	{0xC12C88FEB367C6A0ULL, "56670932.ogg"},
	{0xC12CA21486E3AD88ULL, "mrl210_merrill6116389.cl"},
	{0xC12CC5A1405E439EULL, "6105396_m.fxe"},
	{0xC12CF8B340933727ULL, "gencr_virvle_mage_f.utc"},
	{0xC12D0B592DC5E8D5ULL, "vxca_caverns2_83z_0_0.vlm"},
	{0xC12D2F192F5B00FBULL, "lgt000cr_mage_turn_in_4.utc"},
	{0xC12DB51D889A935BULL, "vxlt_backalley_g_1rd_0_0.vlm"},
	{0xC12E92C9C32B3056ULL, "508856527.ogg"},
	{0xC12EAED4F296142DULL, "vxht_playershouse_n_1ra_0_0.vlm"},
	{0xC12EC10D490F74A2ULL, "dre000_dougal6129815.cl"},
	{0xC12EC8C3CD9813EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v3.ani"},
	{0xC12F3036C655F671ULL, "6169717_m.fxe"},
	{0xC12F3925DE6B0762ULL, "564332676.ogg"},
	{0xC12F560EE8280AEEULL, "vxht_mansion2_1hv_0_0.vlm"},
	{0xC12F646D0A6D75CCULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_mora_n.dds"},
	{0xC12F8ACFF38C1222ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.cess.gad"},
	{0xC12F8CFA7F4D0E1BULL, "gen00fl_sebastian6174046.cl"},
	{0xC12FEBDED1BA4454ULL, "vxca_caverns2_96q_0_0.vlm"},
	{0xC12FF401A71AEE4AULL, "1058552439.ogg"},
	{0xC13057B07C1D120BULL, "792140363.ogg"},
	{0xC1307AF25C4981BFULL, "1071044749.ogg"},
	{0xC130E8EF27FED95AULL, "454695841.ogg"},
	{0xC1314B413D1330FBULL, "vxdr_primevaltomb_2xi_0_0.vlm"},
	{0xC1315060B803C97DULL, "vxlt_backalley_n_1k4_0_0.vlm"},
	{0xC13157B9F48DF92EULL, "vxht_0_hightown_a2_apj_0_0.vlm"},
	{0xC131DB6101AD1EE2ULL, "ach_soc_investigate.dds"},
	{0xC131FB013B775DD3ULL, "203830412.ogg"},
	{0xC13246205C29C915ULL, "llac_w_blightlands_131069_0.dds"},
	{0xC13247A542CA7F72ULL, "vxlt_0_lowtown_docks_aov_0_0.vlm"},
	{0xC13254FD5B3A5A5AULL, "hlac_dr_slavers_3216_6z.dds"},
	{0xC132964227D62466ULL, "438223460.ogg"},
	{0xC132C44979E83D04ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallgreen.mmh"},
	{0xC132C84979E84457ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallgreen.mao"},
	{0xC132D0EBB484DD63ULL, "6082597_m.fxe"},
	{0xC132DA4979E862A6ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzsmallgreen.msh"},
	{0xC132E5BB9E47A3A5ULL, "464829907.ogg"},
	{0xC132ECA35C5ECD23ULL, "vxca_caverns1_fn2_0_0.vlm"},
	{0xC13354724EE22D09ULL, "vxht_chantry_36i_0_0.vlm"},
	{0xC1336E589D022AE4ULL, "614140604.ogg"},
	{0xC133F5DC8BFF59C7ULL, "266738685.ogg"},
	{0xC1347609B2AF0A34ULL, "158472090.ogg"},
	{0xC1354DB969D22704ULL, "6045335_m.fxe"},
	{0xC135B68B317C3B4FULL, "6095598_f.fxe"},
	{0xC135BCDB235F99D6ULL, "qcr220_grey_wardens6076651.cl"},
	{0xC135E1AAEB56A906ULL, "character_character_p175.psh.fxo.dx11.assembly"},
	{0xC13617A1DE5393F1ULL, "vxlt_0_lowtown_docks_a3_n_ctb_0_0.vlm"},
	{0xC13638CDD7BCA7F5ULL, "gen00fl_aveline6112716.cl"},
	{0xC1365963BF29A848ULL, "6120759_m.fxe"},
	{0xC136798B1F4B43C0ULL, "vxht_mansion_n_60p_0_0.vlm"},
	{0xC136ADCFF3926C19ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.cess.evt"},
	{0xC1377C5F4ABB7E74ULL, "22623076.ogg"},
	{0xC13801262C7A821EULL, "vxw_encounter3_85_0_0.vlm"},
	{0xC1382823ED6A95B3ULL, "vxht_0_hightown_a3_avx_0_0.vlm"},
	{0xC138498F9642F4FFULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.impa.gad"},
	{0xC138D274CEEE82DAULL, "6189424_m.fxe"},
	{0xC138E8995119BA6FULL, "6120110_m.fxe"},
	{0xC138EBB741426EBAULL, "dae000_gamlen6048481.cl"},
	{0xC139AFA439A6BEB3ULL, "gen_im_trs_gen_crackedstaff.uti"},
	{0xC13A5E298CF4FBA8ULL, "6063854_m.fxe"},
	{0xC13AACEA3314652EULL, "483246078.ogg"},
	{0xC13ACF24D4A91F29ULL, "dae000_male_commoner2_mob6174441.cl"},
	{0xC13AEA13005FAF87ULL, "vxht_0_hightown_a3_bgl_0_0.vlm"},
	{0xC13AF881FE0C3AC9ULL, "vxca_caverns1_mines_cln_0_0.vlm"},
	{0xC13AFA2243FBF439ULL, "680381795.ogg"},
	{0xC13B5033AA3FFBA0ULL, "vxht_mansion2b_32p_0_0.vlm"},
	{0xC13BC0C3CDA2903FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v3.evt"},
	{0xC13BE9EEAFEFC99EULL, "vxca_caverns1_mines_j2c_0_1.vlm"},
	{0xC13C6D101CFEFE14ULL, "vxw_woundedcoast_gl_0_0.vlm"},
	{0xC13D0A9FACACFC40ULL, "mag311_nyssa6064845.cl"},
	{0xC13D904BE3C3ADDAULL, "vxht_0_hightown_n_b1q_0_0.vlm"},
	{0xC13DBB29A260A6A7ULL, "742068750.ogg"},
	{0xC13DE93CB0E6BAF7ULL, "vxw_bonepitdestroyed_a3_ls_0_0.vlm"},
	{0xC13E3505AB60E7DBULL, "864583199.ogg"},
	{0xC13E3D2AB589F34FULL, "782893028.ogg"},
	{0xC13E3E8F96474DECULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.impa.evt"},
	{0xC13F0681F82E8F51ULL, "vxca_caverns2_74p_0_0.vlm"},
	{0xC13F953E1826FDA5ULL, "llac_ca_caverns1_24421_1.dds"},
	{0xC13F9AC1733A3553ULL, "11660616.ogg"},
	{0xC13FC9E79FA34BB7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_basic_1.mao"},
	{0xC14093354E97B728ULL, "gen00fl_aveline6139859.cl"},
	{0xC140E712AC6AFBAFULL, "6071094_m.fxe"},
	{0xC1413F36BDE39375ULL, "674759157.ogg"},
	{0xC141548753ADAAF0ULL, "character_character_p160.psh.fxo.dx11.assembly"},
	{0xC14197808EBCCDE1ULL, "6137655_m.fxe"},
	{0xC141E6CECB3E52F7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofc.mmh"},
	{0xC141EACECB3E5904ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofc.mao"},
	{0xC141FCCECB3E7869ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_keeproofc.msh"},
	{0xC14233460B5BEF2BULL, "ht_brothel_n_3367_lc.mmh"},
	{0xC14234BDC68798AFULL, "dae200_cullen6140330.cl"},
	{0xC142776315A3301AULL, "6084547_m.fxe"},
	{0xC1429348054C7D9FULL, "vxlt_0_lowtown_a2_n_c_gll_0_0.vlm"},
	{0xC142A1C3CDA87A28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_die_v3.gad"},
	{0xC142CAEFE17A31E7ULL, "character_character_p141.psh.fxo.dx11"},
	{0xC143B5CFF39D0408ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hm_c.cess.ani"},
	{0xC1445D6A6AA657E5ULL, "vxw_encounter3_6p_0_0.vlm"},
	{0xC14463F8FF2C3557ULL, "6141138_f.fxe"},
	{0xC1446D3995C6828FULL, "6082229_m.fxe"},
	{0xC144B281117D8C18ULL, "hlac_lt_0_lowtown_-2_4z.dds"},
	{0xC144B3B4E50E5850ULL, "vxlt_0_lowtown_a2_bxg_0_0.vlm"},
	{0xC1451BB79864D28BULL, "set_ai_behavior.ncs"},
	{0xC1454A13B7140800ULL, "52504398.ogg"},
	{0xC145776E5597A828ULL, "6036746_m.fxe"},
	{0xC145AC54C0C4B73CULL, "vxht_mansion2b_2lc_0_0.vlm"},
	{0xC145C54354C26659ULL, "1066098553.ogg"},
	{0xC145C8A253C690E8ULL, "6149097_m.fxe"},
	{0xC145F90729CD918CULL, "fen210_hadriana6152632.cl"},
	{0xC14673031E709A9EULL, "vxw_encounter3_56_0_0.vlm"},
	{0xC1476161E487A090ULL, "w_woundedcoast_52_0.msh"},
	{0xC1476F61E487B8AAULL, "w_woundedcoast_52_0.mmh"},
	{0xC1477361E487BFE1ULL, "w_woundedcoast_52_0.mao"},
	{0xC1478CE21D5AA0C1ULL, "and311ip_store_lirene.utp"},
	{0xC147DB1D27CF6A6CULL, "vxw_sundermount_2cv_0_0.vlm"},
	{0xC148351FF7F41BB7ULL, "w_encounter3_52_0.rml"},
	{0xC14850281DC03F33ULL, "vxht_0_hightown_cfc_0_0.vlm"},
	{0xC1485154747E7CCFULL, "vxlt_0_lowtown_docks_9a1_0_0.vlm"},
	{0xC148AC191C17C43AULL, "385867829.ogg"},
	{0xC148F92ADF7CB6A4ULL, "vxlt_0_lowtown_docks_a3_aoo_0_0.vlm"},
	{0xC14914BED6EFB428ULL, "190535427.ogg"},
	{0xC1493A4E5C3AC6CDULL, "dae300_petrice6097898.cl"},
	{0xC149CC168997FA50ULL, "art\\vfx\\combat\\vfx_c_rw_s1_s2_trans_c.mmh"},
	{0xC149D7A932B45E78ULL, "890589043.ogg"},
	{0xC149DCA839928DEBULL, "6115778_m.fxe"},
	{0xC14A588893050393ULL, "552621295.ogg"},
	{0xC14A6A0EE5A3A6A4ULL, "hlac_w_sundermount_131067_0y.dds"},
	{0xC14BE0F9A799E157ULL, "6061352_m.fxe"},
	{0xC14C0EB19DA3F106ULL, "mag291_ntbanter6056740.cl"},
	{0xC14C268F96536245ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.impa.ani"},
	{0xC14CC3A5EF12FFFAULL, "6101370_m.fxe"},
	{0xC14D049C71DB2284ULL, "vxht_keep_6cb_0_0.vlm"},
	{0xC14D3B7091BFCD66ULL, "6171459_f.fxe"},
	{0xC14D47DCD7BF4915ULL, "mag110_arianni6166721.cl"},
	{0xC14DC7EDE18D3542ULL, "sse_ss_fereldan_f_1.gda"},
	{0xC14DE89A0048CD0DULL, "278291640.ogg"},
	{0xC14E01597AB59BEBULL, "01_glow.mao"},
	{0xC14F341FF7FA38AFULL, "w_encounter3_52_0.phy"},
	{0xC14F4749B51EBD4AULL, "vxlt_hangedman_w0_0_0.vlm"},
	{0xC14F6B3678E9A891ULL, "art\\characters\\creatures\\dragon\\animation\\die_lp_v2.ani"},
	{0xC14FB35E90C58BABULL, "art\\vfx\\combat\\vfx_r_brd_ssong_act_c.anb"},
	{0xC14FDB72EC75184AULL, "fo_smok.mao"},
	{0xC14FE3F8DBB3834DULL, "6149873_f.fxe"},
	{0xC14FE4B35D16BF30ULL, "dr_slavers_3207_ll.mmh"},
	{0xC150E344BC2D6EFDULL, "vxlt_0_lowtown_docks_n_b5z_0_1.vlm"},
	{0xC151746E7DAF54EAULL, "maoinit.erf.s"},
	{0xC15176574B6DFE2BULL, "118900561.ogg"},
	{0xC1524609AEBD6184ULL, "vxht_0_hightown_n_bph_0_0.vlm"},
	{0xC152A1A2C1D47DEEULL, "dae000ip_crowd_em4_talk7.utp"},
	{0xC152B3C4CC518A39ULL, "vxht_mansion2b_3ep_0_0.vlm"},
	{0xC152C45D9EF757C4ULL, "6110842_m.fxe"},
	{0xC152F92871C6FDFEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medf_0s.dds"},
	{0xC152F946DFFDD608ULL, "309795867.ogg"},
	{0xC1533D0A5BBD23B4ULL, "225118317.ogg"},
	{0xC153CDF224DBE8B4ULL, "lt_0_lowtown_docks_a3_mmap.dds"},
	{0xC153D3473C299C55ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl01.phy"},
	{0xC153EDBADEB0C8DBULL, "mcr310_argument6127685.cl"},
	{0xC1543A7F45BCA968ULL, "6142197_m.fxe"},
	{0xC1553AE7F7E9D05AULL, "and100_anders6111667.cl"},
	{0xC155889E5BC88AB3ULL, "103870774.ogg"},
	{0xC1559DE6BD730D96ULL, "6188650_m.fxe"},
	{0xC1564CFA19C582ECULL, "vxw_sundermount_2dw_0_0.vlm"},
	{0xC156763678EFDA33ULL, "art\\characters\\creatures\\dragon\\animation\\die_lp_v2.gad"},
	{0xC156AF4782C43AF2ULL, "6110754_m.fxe"},
	{0xC156B5DA7715174AULL, "dae120cr_hf_guard_1.utc"},
	{0xC156D7ACDF891A91ULL, "qun110_petrice6018564.cl"},
	{0xC156EF1689A2C06FULL, "art\\vfx\\combat\\vfx_c_rw_s1_s2_trans_c.anb"},
	{0xC157DF010661E3C4ULL, "6099150_m.fxe"},
	{0xC1583CC244F47C8DULL, "6075816_m.fxe"},
	{0xC158A38C65DFBC71ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_loop.ani"},
	{0xC1598BF14C68568FULL, "vxht_keep_5hk_0_0.vlm"},
	{0xC159D51953C05A29ULL, "199025528.ogg"},
	{0xC159F969E1B80093ULL, "dre000_dougal6128877.cl"},
	{0xC15A29AB2A6B5904ULL, "123694511.ogg"},
	{0xC15A3F510505CA7CULL, "vxlt_0_lowtown_docks_a3_n_d0h_0_0.vlm"},
	{0xC15B059AB5341D5EULL, "art\\characters\\creatures\\qunari\\textures\\qm_eye_stn_0d.dds"},
	{0xC15B2C7082BE561FULL, "dae200_saemus6141206.cl"},
	{0xC15B4E6A577850DAULL, "608595463.ogg"},
	{0xC15B6C286CBCA927ULL, "gen00fl_aveline6114851.cl"},
	{0xC15BB55A8A4DA6C6ULL, "6116158_m.fxe"},
	{0xC15BD0CD74B32060ULL, "hla_ht_0_hightown_a3_1_0x.dds"},
	{0xC15BD0E88FDF9E78ULL, "hlac_lt_0_lowtown_a3_n_c_65534_5y.dds"},
	{0xC15C08B6A4F9C1BAULL, "zz_otl_debug6060583.cl"},
	{0xC15C6DF5D58D3494ULL, "vxw_bonepit_1qd_0_0.vlm"},
	{0xC15C77491F09AF67ULL, "gen_im_wep_mag_sta_start.uti"},
	{0xC15DC01F49931F73ULL, "vxlt_undercitydungeon_24h_0_0.vlm"},
	{0xC15DFB86AE3813F8ULL, "6095073_m.fxe"},
	{0xC15E39B9170AE960ULL, "6115896_m.fxe"},
	{0xC15E3C7B5DD99AE3ULL, "vxdr_primevaltomb_5a7_0_2.vlm"},
	{0xC15E75B74E336AF6ULL, "vxlt_0_lowtown_a2_n_bxn_0_0.vlm"},
	{0xC15EA48C0F27B53BULL, "vxw_encounter1_j9_0_1.vlm"},
	{0xC15EB4A044C87A42ULL, "cod_ltr_ggt_note1.utp"},
	{0xC15F50BA3E4FF83FULL, "lt_0_lowtown_docks_a3_224_0.msh"},
	{0xC15F56BA3E5002E1ULL, "lt_0_lowtown_docks_a3_224_0.mmh"},
	{0xC15F62BA3E501682ULL, "lt_0_lowtown_docks_a3_224_0.mao"},
	{0xC15F7701E938AE9FULL, "vxlt_0_lowtown_a2_n_c_gbn_0_0.vlm"},
	{0xC15FAB4C0B42EE86ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.dura.evt"},
	{0xC15FAE8C65E5ED13ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_loop.gad"},
	{0xC1604AEC68E27314ULL, "dae000cr_grand_cleric6148082.cl"},
	{0xC1605BC8A8569E7BULL, "vxht_playershouse_n_1sh_0_0.vlm"},
	{0xC160B3B09E22A605ULL, "hlac_w_sundermount_393215_0y.dds"},
	{0xC160D34B4B1AC9DEULL, "vxw_sundermount_34t_0_0.vlm"},
	{0xC1611B74769092CBULL, "300334626.ogg"},
	{0xC16136078D046241ULL, "883793573.ogg"},
	{0xC16140D0F468EF52ULL, "6076384_m.fxe"},
	{0xC16165CBD5498F8BULL, "vxca_caverns1_mines_e86_0_0.vlm"},
	{0xC1618633CAA0ED98ULL, "llac_ht_mansion2_818_35.dds"},
	{0xC1619687CF37CF13ULL, "6019113_m.fxe"},
	{0xC161B7C47696681EULL, "vxlt_0_lowtown_a2_er4_0_0.vlm"},
	{0xC161E323B0335E02ULL, "hlac_w_woundedcoast_0_7x.dds"},
	{0xC162897212EDEDC4ULL, "vxlt_0_lowtown_a2_do0_0_0.vlm"},
	{0xC1628B64C080FEDAULL, "vxht_keep_n_6ia_0_0.vlm"},
	{0xC162B33A9911B6D9ULL, "ss_encampment.dds"},
	{0xC162B485593D5ECCULL, "dae000ip_crowd_ff5_talk2.utp"},
	{0xC1631C02FFB056ECULL, "6028939_f.fxe"},
	{0xC1636F5B6C11A69FULL, "follower_banter6156185.cl"},
	{0xC164BED70BBB9B6CULL, "vxw_sundermount_2it_0_0.vlm"},
	{0xC1654894811FE0D6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowb_sa.dds"},
	{0xC165752F50EFEFB0ULL, "166943223.ogg"},
	{0xC165B5BF09F12BD0ULL, "ave310_party_comment6134850.cl"},
	{0xC165FB14A9E33AE6ULL, "575700767.ogg"},
	{0xC16628C8925E85BEULL, "6152341_m.fxe"},
	{0xC16639C69CEA2ECDULL, "hlac_lt_undercityserialkiller_-1_52z.dds"},
	{0xC16668211648510FULL, "409348692.ogg"},
	{0xC1667B30913EDD46ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_brazier_d.dds"},
	{0xC166AB8C65EC0778ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_loop.evt"},
	{0xC166C24C0B4934D1ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.dura.gad"},
	{0xC166E7B450C0E970ULL, "vxlt_hangedman_n_1yf_0_0.vlm"},
	{0xC167078218092DFCULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyo_0n.dds"},
	{0xC167436A122395F4ULL, "isa111_isabela6175108.cl"},
	{0xC1680609DFF28AD4ULL, "vxlt_gallowstemplar_n_3nd_0_0.vlm"},
	{0xC16898D4A76D3325ULL, "hlac_ca_caverns1_mines_24426_11y.dds"},
	{0xC168E748382A1717ULL, "421831471.ogg"},
	{0xC16900F34B08FE07ULL, "meleering4.gda"},
	{0xC1690CD6116D65ADULL, "24112993.ogg"},
	{0xC16937EDAC018A63ULL, "lgt110_arl_bann_teagan6115611.cl"},
	{0xC1699B1EDD3F24F2ULL, "6087814_m.fxe"},
	{0xC169BB666DF3C907ULL, "hlac_lt_0_lowtown_docks_65532_1y.dds"},
	{0xC169D058B38E334AULL, "follower_banter6172940.cl"},
	{0xC16A2841E4EE6528ULL, "61431618.ogg"},
	{0xC16A5B6FFF039FEBULL, "vxdr_0_deeproads_9ek_0_0.vlm"},
	{0xC16ADFD2C7834D14ULL, "vxw_bonepit_1ze_0_0.vlm"},
	{0xC16AE6862C1CF6C4ULL, "34536805.ogg"},
	{0xC16B66959F33B5A1ULL, "vxlt_house_ez_0_0.vlm"},
	{0xC16BE69E98864D5FULL, "316899289.ogg"},
	{0xC16BF3C4E45D93E7ULL, "106946643.ogg"},
	{0xC16BF5E594EC3CC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_reaction.gad"},
	{0xC16CC34C0B4DA27FULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.dura.ani"},
	{0xC16D1EEF794AF05FULL, "27252755.ogg"},
	{0xC16D81D626E9B1E2ULL, "6149089_f.fxe"},
	{0xC16D8D27D5111AB3ULL, "pro000_cut2d_thefirstyear6005313.cl"},
	{0xC16D9B410995CD42ULL, "zz_mcr_debug6098525.cl"},
	{0xC16DC0B5999A84F2ULL, "417376.ogg"},
	{0xC16ECDA59A4CB6FBULL, "vxht_playershouse_n_1vk_0_0.vlm"},
	{0xC16F0A1800F4F846ULL, "6016140_m.fxe"},
	{0xC16F29C392A99284ULL, "vxht_0_hightown_n_bfn_0_0.vlm"},
	{0xC16F45283D10E25EULL, "vxw_sundermount_33u_0_0.vlm"},
	{0xC16F532471256E43ULL, "ts_circleofmagi_acc_table.gda"},
	{0xC16F907F24A41380ULL, "6173135_m.fxe"},
	{0xC16FCAFC62DFEBF6ULL, "mcr310_argument6074375.cl"},
	{0xC16FCCC02F6823A2ULL, "one100dg_nabil6054972.cl"},
	{0xC1701B84D5F9C71DULL, "vxht_keep_n_5ay_0_0.vlm"},
	{0xC1701F16C88FF7B8ULL, "vxw_bonepitmining_a2_1nw_0_0.vlm"},
	{0xC17090C0D6DFBC84ULL, "6149466_m.fxe"},
	{0xC170DEB8A9351AC6ULL, "6105604_m.fxe"},
	{0xC17128D678A9D8F8ULL, "775983328.ogg"},
	{0xC1714BE2720978C6ULL, "6043804_f.fxe"},
	{0xC1716699F68B6831ULL, "vxht_mansion2_2d2_0_0.vlm"},
	{0xC17188EF1637AE60ULL, "6147631_m.fxe"},
	{0xC17265BC28801393ULL, "vxca_caverns2_8f7_0_0.vlm"},
	{0xC1729286B1DAA66BULL, "6060276_m.fxe"},
	{0xC17292C12C438CBFULL, "vxw_sundermount_2cy_0_0.vlm"},
	{0xC172B7B104A55B11ULL, "hlac_w_blightlands_196609_0y.dds"},
	{0xC17318E594F296B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_reaction.evt"},
	{0xC17330B3FDB1B3ECULL, "vxw_sundermount_2ru_0_0.vlm"},
	{0xC17341EE3E2B55EEULL, "vxht_bartrand_n_5s7_0_0.vlm"},
	{0xC17345FF0CE41C74ULL, "vxw_bonepit_br_0_0.vlm"},
	{0xC1736B7283693124ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_punch_into_palm.ani"},
	{0xC173FDA48D25AD12ULL, "vxht_keep_n_5zs_0_0.vlm"},
	{0xC17417B06E5FBDD4ULL, "hairalpha_rendertoshadowmap_p0.vsh.fxo.dx11.assembly"},
	{0xC175599142B701F0ULL, "vxlt_hangedman_n_1vg_0_0.vlm"},
	{0xC175FE97C894CAB2ULL, "vxlt_gallowscourtyard_n_6y0_0_0.vlm"},
	{0xC1763483E4D76014ULL, "vxw_woundedcoast_sp_0_0.vlm"},
	{0xC176CE53506E2D4AULL, "w_encounter2_21_0.mao"},
	{0xC176D253506E3399ULL, "w_encounter2_21_0.mmh"},
	{0xC176D2EF8214B77AULL, "art\\vfx\\textures\\vfx_smoke_puffy_2_cs.dds"},
	{0xC176DC53506E4487ULL, "w_encounter2_21_0.msh"},
	{0xC176E4F53BEE7D44ULL, "zz_qun_debug6036785.cl"},
	{0xC177135CE728D96FULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.cess.gad"},
	{0xC177A74A74BFC267ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_002a_n.dds"},
	{0xC177C2A0E5FA7876ULL, "hlac_lt_0_lowtown_docks_a3_131070_3x.dds"},
	{0xC177F4DBF8F34FC2ULL, "809776454.ogg"},
	{0xC178B3A750556710ULL, "255506650.ogg"},
	{0xC178BDA650293B4CULL, "static_static_p45.psh.fxo.dx11"},
	{0xC178BE0353994895ULL, "vxht_keep_n_6j7_0_0.vlm"},
	{0xC178ED82AA1C162EULL, "905584746.ogg"},
	{0xC1795DACD38AEEEAULL, "6198192_m.fxe"},
	{0xC1796872836D9786ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_punch_into_palm.gad"},
	{0xC17973EB3F47020FULL, "vxht_0_hightown_c70_0_0.vlm"},
	{0xC17A0BFC921C82CCULL, "follower_banter6156030.cl"},
	{0xC17A63A6D2DABB93ULL, "743677771.ogg"},
	{0xC17AD7170504CE07ULL, "llac_lt_gallowstemplar_n_7198_3.dds"},
	{0xC17AF3813038A190ULL, "308196740.ogg"},
	{0xC17AF525ABB77A0AULL, "ser221ip_door_chase1.ncs"},
	{0xC17B0336B29EFFD5ULL, "473115136.ogg"},
	{0xC17B386FB7948A54ULL, "113469505.ogg"},
	{0xC17BA8F589417F0DULL, "229830317.ogg"},
	{0xC17CB6D5FDB5B051ULL, "hlac_lt_0_lowtown_n_-4_0z.dds"},
	{0xC17D6704E8FD246EULL, "mrl310_party_comments6142785.cl"},
	{0xC17D9BA0849FAB04ULL, "vxht_0_hightown_n_bao_0_0.vlm"},
	{0xC17DB3924ACE75E0ULL, "hlac_lt_backalley_n_65537_4y.dds"},
	{0xC17DCCF6D7EE8AA9ULL, "vxlt_gallowstemplar_f_435_0_0.vlm"},
	{0xC17DD8168E278769ULL, "6072303_m.fxe"},
	{0xC17E0AAEDA98BB53ULL, "gen00fl_aveline6099716.cl"},
	{0xC17E0AC0711F6CC2ULL, "w_bonepitmining_a2_143_0.msh"},
	{0xC17E0CC0711F7078ULL, "w_bonepitmining_a2_143_0.mmh"},
	{0xC17E18C0711F841BULL, "w_bonepitmining_a2_143_0.mao"},
	{0xC17E285CE72F1B1CULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.cess.evt"},
	{0xC17E87A1A838714FULL, "hlac_lt_warehouse_0_1y.dds"},
	{0xC17E8D61C7EFDF9DULL, "vxht_keep_n_5jx_0_0.vlm"},
	{0xC17E9B7E5A82991EULL, "vxlt_0_lowtown_a2_ex6_0_0.vlm"},
	{0xC17EA6AD2236BC20ULL, "hlac_lt_0_lowtown_131071_2z.dds"},
	{0xC17EB3473E426C53ULL, "628975343.ogg"},
	{0xC17EE5AAABCD8802ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue2_e.dds"},
	{0xC17EE690424BB575ULL, "llac_ca_cavern1_haunted_24424_12.dds"},
	{0xC17F5D78BC6FE5F3ULL, "rdr111dg_martin6054538.cl"},
	{0xC17F87A3411992F4ULL, "hlac_ht_brothel_n_3374_0x.dds"},
	{0xC17FD876E88180B1ULL, "vxht_mansion2_2a1_0_0.vlm"},
	{0xC1801E0B8D66E6EFULL, "lt_0_lowtown_a3_n_c_29_0_ll.mmh"},
	{0xC18020E594FD2EA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_reaction.ani"},
	{0xC18063728373AE65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_punch_into_palm.evt"},
	{0xC180643B569CE599ULL, "plt_isa000pt_endstage.nss"},
	{0xC180796966885995ULL, "gen00fl_varric6133684.cl"},
	{0xC180BECEF3E89EBCULL, "ss_varric.bnk"},
	{0xC18106120B06DB37ULL, "449249133.ogg"},
	{0xC1810E4726F91410ULL, "llac_lt_hangedman_n_-1_15.dds"},
	{0xC1813F156F9034EBULL, "266864065.ogg"},
	{0xC18154F98D66E74AULL, "6187299_m.fxe"},
	{0xC1815C883FCF9DF7ULL, "739183194.ogg"},
	{0xC181A290EFA7CC6CULL, "vxw_sundermount_2wj_0_0.vlm"},
	{0xC181ADCE49093099ULL, "ca_cavern1_haunted_probe_jzd.dds"},
	{0xC181B7DBFEDA34F4ULL, "vxw_bonepit_ms_0_0.vlm"},
	{0xC181F122D6D04456ULL, "fen310_fenris6118201.cl"},
	{0xC182012BC5D9C65FULL, "vxca_caverns1_mines_iiu_0_0.vlm"},
	{0xC182BC1B9495EDD5ULL, "vxlt_undercityserialkiller_yh_0_0.vlm"},
	{0xC182E3D4EB136B00ULL, "fen110_fenris6117921.cl"},
	{0xC1838EF18DF3B169ULL, "vxlt_gallowscourtyard_4wd_0_0.vlm"},
	{0xC183A1F61D51BE80ULL, "follower_banter6136353.cl"},
	{0xC183BEE0029BD4C9ULL, "follower_banter6170473.cl"},
	{0xC184BDEF4A1C2E2CULL, "6070233_f.fxe"},
	{0xC1856006BEEC06DFULL, "character_character_p55.psh.fxo.dx11.assembly"},
	{0xC185B090C36A8E1EULL, "hlac_lt_0_lowtown_a3_n_c_65534_2z.dds"},
	{0xC185EA238606F512ULL, "lowtown_fem_alienage1_sta6162584.cl"},
	{0xC186559B3BC3C96AULL, "vxlt_gallowstemplar_n_4na_0_0.vlm"},
	{0xC186562541FAB0EFULL, "qun160_vashothlead6053860.cl"},
	{0xC186AB8FE32837BAULL, "art\\characters\\playercharacter\\humanfemale\\boo_cas.mao"},
	{0xC1874B3B56A2DA72ULL, "plt_isa000pt_endstage.plo"},
	{0xC18752E192E5051AULL, "vxht_mansion2_289_0_0.vlm"},
	{0xC187673CAB16A87FULL, "vxht_0_hightown_c3t_0_0.vlm"},
	{0xC18768065BE825E1ULL, "wireframe.psh.fxo"},
	{0xC187B10F4C47D769ULL, "l_dae330ar_undercity_dt.bnk"},
	{0xC1881D838FB3989FULL, "282557996.ogg"},
	{0xC188A6C17AA02979ULL, "vxht_keep_5k4_0_0.vlm"},
	{0xC188AF8F5AD08C79ULL, "6142848_m.fxe"},
	{0xC1894C30BAB936D7ULL, "390992032.ogg"},
	{0xC18999320C86A18CULL, "994906344.ogg"},
	{0xC18A0EF3B34070ACULL, "375191976.ogg"},
	{0xC18A5CD00F673BEAULL, "genip_pp_gle_standing.utp"},
	{0xC18B305CE739B335ULL, "art\\vfx\\creatures\\fxc_arch_db_prj_p.cess.ani"},
	{0xC18B9406A920D0CAULL, "6059149_m.fxe"},
	{0xC18B966CE14E3C29ULL, "6044900_m.fxe"},
	{0xC18C17D564F50D1BULL, "vxca_caverns2prc_4k9_0_0.vlm"},
	{0xC18C362F5BA423CEULL, "vxht_0_hightown_n_cf1_0_0.vlm"},
	{0xC18C5E0A5FF58BD8ULL, "6202699_m.fxe"},
	{0xC18CC17382AD24B4ULL, "vxht_0_hightown_a2_621_0_0.vlm"},
	{0xC18D47A728E869C3ULL, "hlac_w_bonepitdestroyed_a3_131077_0x.dds"},
	{0xC18D61F069B506C6ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_5.dds"},
	{0xC18DE89E2D9B2003ULL, "vxlt_gallowstemplar_f_4u1_0_0.vlm"},
	{0xC18E4A53DA779931ULL, "vxht_mansion2_2j0_0_0.vlm"},
	{0xC18ECACCE343F096ULL, "vxlt_0_lowtown_a2_n_c_gkp_0_0.vlm"},
	{0xC18EF48B634BEA16ULL, "683247264.ogg"},
	{0xC190146DE19DE4ECULL, "vxw_sundermount_2xk_0_0.vlm"},
	{0xC1902AB0BDCA1F6DULL, "vxht_bartrand_n_5rs_0_0.vlm"},
	{0xC1902BA688F0F3E9ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallobigbird.phy"},
	{0xC190D4318C5EFF6FULL, "mrl000_marethari6129185.cl"},
	{0xC190F10D58A34D64ULL, "vxlt_hangedman_n_2dn_0_0.vlm"},
	{0xC1911B12C508428CULL, "gen00fl_varric6107279.cl"},
	{0xC191CB8C1B68E66CULL, "docks_scene_dockworkers6176486.cl"},
	{0xC192CC7AD0E4C44EULL, "rdr111dg_martin6054083.cl"},
	{0xC1932FCC3C1246ACULL, "6075133_f.fxe"},
	{0xC193A8A5E1BD85D6ULL, "vxlt_0_lowtown_a2_n_c_by3_0_0.vlm"},
	{0xC193CC0087894F18ULL, "vxca_cave_2jy_0_0.vlm"},
	{0xC193DA334E513CE7ULL, "nat320_delilah6065941.cl"},
	{0xC194291E04FFD425ULL, "vxlt_0_lowtown_a2_n_bz3_0_0.vlm"},
	{0xC194DB491BA18F20ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_8y.dds"},
	{0xC194ED3D91B455EEULL, "w_sundermount_150_0.phy"},
	{0xC19501C421965713ULL, "vxlt_gallowstemplar_2mx_0_0.vlm"},
	{0xC19595DE069D8F57ULL, "6122101_m.fxe"},
	{0xC1959ABFEA4FCB4DULL, "w_encounter1_55_0.mmh"},
	{0xC1959EBFEA4FD19EULL, "w_encounter1_55_0.mao"},
	{0xC195ACBFEA4FE9D3ULL, "w_encounter1_55_0.msh"},
	{0xC1961BF76C5445C3ULL, "cod_mgc_spirits.utp"},
	{0xC1963DE71CB6F67AULL, "genip_cin_dead_meredith.utp"},
	{0xC1964BFB2EA6F527ULL, "zz_dae_debug6045825.cl"},
	{0xC196DE28645A131BULL, "vxca_cavern1_haunted_j22_0_1.vlm"},
	{0xC1974AD09EF839FDULL, "6139877_m.fxe"},
	{0xC197AF59DEBF8261ULL, "hlac_lt_foundry_n_320_13x.dds"},
	{0xC197D26BD2AAF410ULL, "zz_dre_debug6168238.cl"},
	{0xC197D4DBABD24700ULL, "320002429.ogg"},
	{0xC197D900457AA5B7ULL, "vxht_0_hightown_n_c4b_0_0.vlm"},
	{0xC19810BA1B87C2C6ULL, "vxw_blightlands_f3_0_0.vlm"},
	{0xC1984AEF9B953CDCULL, "one140cr_charade6137602.cl"},
	{0xC198F43979B9D794ULL, "hlac_lt_gallowstemplar_2394_5y.dds"},
	{0xC19A0E1E86B8C964ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_die_v1.ani"},
	{0xC19A26F6E48B22C2ULL, "1009361424.ogg"},
	{0xC19A56B42C7EF5E4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cartempty_01_0_sa.dds"},
	{0xC19A8DF073D64553ULL, "gen00fl_anders6125687.cl"},
	{0xC19ABF8E8329870AULL, "dae000_bartrand6030642.cl"},
	{0xC19AE6473C65BF70ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl01.msh"},
	{0xC19AE8EE2BB0E86DULL, "vrc321_party_barks6096307.cl"},
	{0xC19AF4473C65D68AULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl01.mmh"},
	{0xC19B4E67BDCD6A6FULL, "6116317_m.fxe"},
	{0xC19B70668544357DULL, "zz_followers_banter6157670.cl"},
	{0xC19D30A0FD936463ULL, "zz_mcr_debug.cnv"},
	{0xC19D3B151E59991AULL, "465115202.ogg"},
	{0xC19D422E1AA41E6CULL, "nat380_nathaniel6066410.cl"},
	{0xC19DBED0FBCD8143ULL, "1027122181.ogg"},
	{0xC19DC66C2DE3B4D2ULL, "cod_hst_rivain.utp"},
	{0xC19DE8580225D63FULL, "vxw_sundermount_2rz_0_0.vlm"},
	{0xC19E4F6D5D50624FULL, "260305153.ogg"},
	{0xC19E5DA3860BE77DULL, "6095817_f.fxe"},
	{0xC19E912A75AA454BULL, "follower_barks6153390.cl"},
	{0xC19E9C8DAFC037EDULL, "vxht_bartrand_n_5ur_0_0.vlm"},
	{0xC19F5C123F5E2492ULL, "hlac_lt_warehouse_n_65537_2z.dds"},
	{0xC19F6BBB8E3D5BFCULL, "vxdr_primevaltomb_7e5_0_1.vlm"},
	{0xC1A03BE11CA37110ULL, "467776456.ogg"},
	{0xC1A09F64ED5D0464ULL, "zz_seb_debug6215764.cl"},
	{0xC1A0C92DB953026BULL, "gen00fl_sebastian6180434.cl"},
	{0xC1A10B1E86BEE2C6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_die_v1.gad"},
	{0xC1A141EF006DC38EULL, "159191747.ogg"},
	{0xC1A224F219F762E5ULL, "vxht_0_hightown_n_b8u_0_0.vlm"},
	{0xC1A27C4F96D2651EULL, "plt_cod_bks_chantry_hist_4.plo"},
	{0xC1A2C424A428214BULL, "358412071.ogg"},
	{0xC1A2C48F39C5B936ULL, "vxht_keep_n_6zp_0_0.vlm"},
	{0xC1A30F2F0A0288BFULL, "hlac_lt_gallowscourtyard_n_-131073_0x.dds"},
	{0xC1A33F0843682DAEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thin.phy"},
	{0xC1A38DA3BE602EBEULL, "200843438.ogg"},
	{0xC1A3A573FDB7283FULL, "715752228.ogg"},
	{0xC1A3C1E21152787DULL, "464016435.ogg"},
	{0xC1A406A52D839388ULL, "ico_junk.dds"},
	{0xC1A4B9C625C8A807ULL, "6089105_m.fxe"},
	{0xC1A4C10A3EC53575ULL, "aip_shade_base.gda"},
	{0xC1A4CEE53DF503E4ULL, "6033561_m.fxe"},
	{0xC1A561672AB69D02ULL, "qun200_party_comment6095344.cl"},
	{0xC1A5EE5DB07F8AE7ULL, "655633858.ogg"},
	{0xC1A67D9754276B1DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basic.phy"},
	{0xC1A6A00E332164FCULL, "555401127.ogg"},
	{0xC1A6C59F7F2AA7D3ULL, "vxlt_gallowstemplar_n_31z_0_0.vlm"},
	{0xC1A77B5169D4EE4EULL, "256468205.ogg"},
	{0xC1A7933A7E596586ULL, "hlac_lt_backalley_n_65537_3z.dds"},
	{0xC1A8061E86C4F8A5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_die_v1.evt"},
	{0xC1A886B0FE09E9B0ULL, "251205261.ogg"},
	{0xC1A92E6EDE876EF9ULL, "vxlt_0_lowtown_docks_bbm_0_0.vlm"},
	{0xC1A92F41EE669F07ULL, "247459615.ogg"},
	{0xC1A97BF34EBF23ADULL, "dae000_madam_lusine6151909.cl"},
	{0xC1A9AEF2AB98A4A5ULL, "mag101cr_apostate.utc"},
	{0xC1A9B0D056EBDA31ULL, "gen00fl_fenris6110805.cl"},
	{0xC1A9E68A906859B8ULL, "vxw_bonepitmining_a2_1zs_0_0.vlm"},
	{0xC1AA0362EEE4E361ULL, "dae000_sandal6148235.cl"},
	{0xC1AA134D80CDACD5ULL, "vxdr_primevaltomb_6yp_0_0.vlm"},
	{0xC1AA1843C9E3EC14ULL, "art\\vfx\\combat\\vfx_f_and_fade_c.mmh"},
	{0xC1AA1A48FCB27F27ULL, "6066317_m.fxe"},
	{0xC1AA41AA4FB2C97CULL, "204053576.ogg"},
	{0xC1AA7488FF94AC77ULL, "315898047.ogg"},
	{0xC1AB4CCF2C29D88CULL, "vxlt_undercityserialkiller_1g8_0_0.vlm"},
	{0xC1AB4ECFAFA8A308ULL, "llac_lt_0_lowtown_a2_n_1_3.dds"},
	{0xC1AB969C47C030EDULL, "vxlt_undercitydungeon_1d4_0_1.vlm"},
	{0xC1AB9906F37F2DB3ULL, "6154441_m.fxe"},
	{0xC1ABBDBDD2D0EAE0ULL, "vxdr_primevaltomb_7p3_0_0.vlm"},
	{0xC1AC09F3E3BBF22EULL, "llac_ca_caverns2_5802_67.dds"},
	{0xC1AC13EFA284A9C8ULL, "6016142_m.fxe"},
	{0xC1AC1C221BFE2883ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_enter.ani"},
	{0xC1AC5A34F41BEEBFULL, "vxw_sundermount_2we_0_0.vlm"},
	{0xC1ACBEACB840AC10ULL, "vxlt_0_lowtown_a2_fid_0_0.vlm"},
	{0xC1AD0E6AA1B6506DULL, "vxht_bartrand_n_5xq_0_0.vlm"},
	{0xC1AD4344B73602AEULL, "ttx_wall_base_1.mmh"},
	{0xC1AD4E1CB3F89306ULL, "gen_im_trs_gen_junk3_04.uti"},
	{0xC1AD5D44B7362E9CULL, "ttx_wall_base_1.msh"},
	{0xC1AD6BE382374501ULL, "seb201_lady_harimann6177429.cl"},
	{0xC1AD6FD231176530ULL, "6126991_m.fxe"},
	{0xC1ADB90FB8BBDB1BULL, "6164727_m.fxe"},
	{0xC1AE356DC101099CULL, "6084530_m.fxe"},
	{0xC1AE454C21BE1467ULL, "vxdr_primevalentrance_23d_0_0.vlm"},
	{0xC1AE60B6493A9C55ULL, "hlac_ca_cave_2194_13z.dds"},
	{0xC1AE99923420FE67ULL, "6166440_m.fxe"},
	{0xC1AEBC2551B70283ULL, "art\\vfx\\combat\\vfx_c_shadow_creature_c.dura.ani"},
	{0xC1AF48DD557EAC4BULL, "t3_arm_rev.tnt"},
	{0xC1AF88E693F742E5ULL, "695464058.ogg"},
	{0xC1AF8B3DF11DD7CEULL, "vxlt_gallowstemplar_45s_0_0.vlm"},
	{0xC1AFCEBD5C944A1EULL, "vxdr_primevalentrance_330_0_0.vlm"},
	{0xC1AFEBF4F4EABDDDULL, "gen00fl_aveline6139535.cl"},
	{0xC1B062D7C6FAE2D0ULL, "vxlt_0_lowtown_a3_n_c_ems_0_0.vlm"},
	{0xC1B0C49F52C96627ULL, "dae000_party_night_comment6201644.cl"},
	{0xC1B106D28FAB261AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\chargenidle.gad"},
	{0xC1B1144EEE2CA9F4ULL, "cod_hst_kirk_arch_1a_10.utp"},
	{0xC1B1366C2BBBD1B6ULL, "vxht_keep_n_6uq_0_0.vlm"},
	{0xC1B148D2A18BAD93ULL, "440650075.ogg"},
	{0xC1B158D9C7D235C5ULL, "hlac_ht_chantry_2534_38y.dds"},
	{0xC1B1B40F78C18723ULL, "vxlt_foundry_n_z1_0_0.vlm"},
	{0xC1B1EC3AE8FFCB1AULL, "vxlt_hangedman_sl_0_0.vlm"},
	{0xC1B22D867C5A4566ULL, "fen200cr_tevinter_magister.utc"},
	{0xC1B2B09E66CC36A3ULL, "hlac_ht_mansion2b_818_21z.dds"},
	{0xC1B2ECD9A6F4792DULL, "vxdr_slavers_1t1_0_1.vlm"},
	{0xC1B2F918D8F8F101ULL, "llac_lt_backalley_g_1_2.dds"},
	{0xC1B303194AA0BF0FULL, "163991882.ogg"},
	{0xC1B32B9EC3682E44ULL, "454919965.ogg"},
	{0xC1B33930FB82C484ULL, "qun221st_petrice_postvarn.stg"},
	{0xC1B3765D3EABDF3FULL, "follower_banter6164038.cl"},
	{0xC1B42ECA08F6FC12ULL, "vxdr_primevalentrance_1a0_0_0.vlm"},
	{0xC1B45112A4B6E7B5ULL, "354803039.ogg"},
	{0xC1B4D8EACA9DFAB6ULL, "vxlt_house_g7_0_0.vlm"},
	{0xC1B4E0672BC1F2C5ULL, "6085745_m.fxe"},
	{0xC1B4ED8D8642752BULL, "6017506_m.fxe"},
	{0xC1B4F473FF73F3C6ULL, "vxw_blightlands_l5_0_0.vlm"},
	{0xC1B58A590F4B48DCULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_014a_0d.dds"},
	{0xC1B5A85499F6E859ULL, "hlac_ht_playershouse_n_855_17x.dds"},
	{0xC1B5F52DBAB7C9C8ULL, "fen110_fenris6101957.cl"},
	{0xC1B601931AC1F6F2ULL, "18056189.ogg"},
	{0xC1B6042CA94FC014ULL, "6154051_m.fxe"},
	{0xC1B60CA582ADBBD6ULL, "6060390_m.fxe"},
	{0xC1B649696DCD24B7ULL, "vrc321st_haunted_chair.stg"},
	{0xC1B65EC68CB1E143ULL, "w_encounter3_175_0.mao"},
	{0xC1B662C68CB1E790ULL, "w_encounter3_175_0.mmh"},
	{0xC1B66A9ECC6096CAULL, "714234245.ogg"},
	{0xC1B670C68CB1FFAAULL, "w_encounter3_175_0.msh"},
	{0xC1B6A3DC04CC695FULL, "plt_cod_hst_kirkwall.plo"},
	{0xC1B6BD2B209701B0ULL, "gen00fl_aveline6120835.cl"},
	{0xC1B6EAC4A7E60FA8ULL, "811516929.ogg"},
	{0xC1B765865CD54B4AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspike.mao"},
	{0xC1B769865CD55199ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspike.mmh"},
	{0xC1B772975435CB89ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basic.mao"},
	{0xC1B773865CD56287ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspike.msh"},
	{0xC1B77E975435E022ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basic.mmh"},
	{0xC1B780975435E3D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_basic.msh"},
	{0xC1B84D493FFDBA28ULL, "outside_day_rig.msh"},
	{0xC1B85867825E7238ULL, "vxw_bonepitmining_a2_1wp_0_0.vlm"},
	{0xC1B8F320522D07B1ULL, "278531131.ogg"},
	{0xC1B934221C08DC7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_enter.evt"},
	{0xC1B943483C9EA8B8ULL, "llac_ca_cavern1_haunted_24425_7.dds"},
	{0xC1B98F69FE90EC6BULL, "vxlt_gallowstemplar_n_6s1_0_0.vlm"},
	{0xC1B99C6FB4E897B2ULL, "vxca_caverns1_gk8_0_1.vlm"},
	{0xC1B9D2198E887F33ULL, "954918555.ogg"},
	{0xC1BA289EA4FB89FBULL, "w_bonepitmining_a2_25_0.mmh"},
	{0xC1BA349EA4FB9D98ULL, "w_bonepitmining_a2_25_0.mao"},
	{0xC1BA469EA4FBBCE5ULL, "w_bonepitmining_a2_25_0.msh"},
	{0xC1BBB86160E1F6B1ULL, "199202958.ogg"},
	{0xC1BBB8F0E14FD4F3ULL, "dae000_flemeth6000898.cl"},
	{0xC1BBCC4CAE39F96CULL, "6121258_f.fxe"},
	{0xC1BBD42551C1B67AULL, "art\\vfx\\combat\\vfx_c_shadow_creature_c.dura.evt"},
	{0xC1BBE69476B23EAAULL, "vxlt_0_lowtown_a3_n_c_h2a_0_0.vlm"},
	{0xC1BBF46067A64FF6ULL, "6207153_m.fxe"},
	{0xC1BC0D9286FEEA9FULL, "6115279_m.fxe"},
	{0xC1BC14A68916070EULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallobigbird.mmh"},
	{0xC1BC20A689161BA5ULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallobigbird.mao"},
	{0xC1BC2EA6891633FCULL, "art\\levels\\races\\ferelden\\props\\gallows\\prp_gallobigbird.msh"},
	{0xC1BC73AE8FFA0719ULL, "6143007_m.fxe"},
	{0xC1BC77FCB072F4AAULL, "633069212.ogg"},
	{0xC1BC8F696920C528ULL, "6036531_m.fxe"},
	{0xC1BCEA20D94E5BC5ULL, "m_qun131ar_undercity_warrens.wbd"},
	{0xC1BD4533FC380F17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_interact_low.gad"},
	{0xC1BDE23B6106E874ULL, "968672145.ogg"},
	{0xC1BE4C7A396553EFULL, "310485530.ogg"},
	{0xC1BE6846683E6979ULL, "vxht_keep_n_4bi_0_0.vlm"},
	{0xC1BE90C4BEAC65F4ULL, "6107452_f.fxe"},
	{0xC1BEA99973E9CBEDULL, "150812567.ogg"},
	{0xC1BEC4D0DA43B219ULL, "565119860.ogg"},
	{0xC1BF127B4CE9F0BFULL, "vxht_chantry_1yl_0_0.vlm"},
	{0xC1BF24F6C935DC49ULL, "hlac_lt_0_lowtown_docks_n_-65537_3x.dds"},
	{0xC1BFF8A6EA2E614BULL, "597043674.ogg"},
	{0xC1C033221C0EF96DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_enter.gad"},
	{0xC1C080372659A450ULL, "gen_im_trs_rdr_two_2hm_r1.uti"},
	{0xC1C0DBA94184B27DULL, "66713565.ogg"},
	{0xC1C0E9F6DA684C00ULL, "hlac_lt_0_lowtown_a2_-131073_1z.dds"},
	{0xC1C0EEE7CD186C5AULL, "ico_gloves_light_2.dds"},
	{0xC1C1273D02979F6AULL, "dae021_sister_samea_sta6150180.cl"},
	{0xC1C1B63B974B6AC0ULL, "gen00fl_aveline6139842.cl"},
	{0xC1C1F137DB804F3DULL, "vxht_0_hightown_a3_ajz_0_0.vlm"},
	{0xC1C1F53DAA14F78FULL, "vxlt_0_lowtown_a2_i4c_0_0.vlm"},
	{0xC1C1FE23A8FBA66BULL, "dae000_epic_merchant_16179427.cl"},
	{0xC1C217C8CA721FA1ULL, "hlac_lt_0_lowtown_docks_a3_-3_7y.dds"},
	{0xC1C24BBB729B9221ULL, "qun110tr_amb_test.ncs"},
	{0xC1C2549EDFAB16A7ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.impa.gad"},
	{0xC1C2BF66CDC6526DULL, "557918006.ogg"},
	{0xC1C2D32551C7D36DULL, "art\\vfx\\combat\\vfx_c_shadow_creature_c.dura.gad"},
	{0xC1C313A0FDB40278ULL, "zz_mcr_debug.ncs"},
	{0xC1C361817CF9600DULL, "llac_lt_0_lowtown_a2_n_131073_2.dds"},
	{0xC1C36650F16A0C46ULL, "vxw_blightlands_q6_0_0.vlm"},
	{0xC1C3D4DA525A397BULL, "bl_spa_dot.mao"},
	{0xC1C427937749FD55ULL, "art\\vfx\\materials\\vfx_flare01_add.mao"},
	{0xC1C43B6633A5DC65ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potb_n.dds"},
	{0xC1C45A33FC3E51A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_interact_low.evt"},
	{0xC1C48A3731804460ULL, "mag101_alain6032206.cl"},
	{0xC1C4D729FC6A19A5ULL, "659537413.ogg"},
	{0xC1C511FE91884464ULL, "926771080.ogg"},
	{0xC1C5279645C8E1CAULL, "gen00fl_merrill6108261.cl"},
	{0xC1C5F9D5D1742894ULL, "vxca_caverns1_egm_0_0.vlm"},
	{0xC1C621D28FBDB000ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\idle\\chargenidle.ani"},
	{0xC1C65660FC5F048DULL, "vxht_0_hightown_n_ame_0_0.vlm"},
	{0xC1C68B5DB96E189FULL, "vxht_mansion2_1b8_0_1.vlm"},
	{0xC1C6AEFC9B501DD0ULL, "lgt110_orz_harr_carta6109254.cl"},
	{0xC1C6B046193EBE3DULL, "392297823.ogg"},
	{0xC1C72C003B3B2C51ULL, "hlac_dr_0_deeproads_5842_3y.dds"},
	{0xC1C76EEF7117A44DULL, "6166680_m.fxe"},
	{0xC1C7B949AF4D56A4ULL, "ser121_jethann6175426.cl"},
	{0xC1C7CD2717A5E48BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_flsa_0s.dds"},
	{0xC1C895F58294887DULL, "6124242_m.fxe"},
	{0xC1C898785BC0F0F8ULL, "vxdr_primevaltomb_7bu_0_0.vlm"},
	{0xC1C9499EDFB122B4ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.impa.evt"},
	{0xC1C98B2F57ADD3F1ULL, "122277538.ogg"},
	{0xC1C9A02708987D36ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_enter.ani"},
	{0xC1C9FA3E21ADE1D9ULL, "datetimelocalization.gda"},
	{0xC1CA50EC9F671BB3ULL, "vxht_0_hightown_c5j_0_0.vlm"},
	{0xC1CB2329C6D0BCF7ULL, "lt_gallowscourtyard_31_0.anb"},
	{0xC1CBACD935A0321DULL, "nat380_nathaniel_banter6067297.cl"},
	{0xC1CBCE729F07CAEFULL, "6153592_m.fxe"},
	{0xC1CC5C88A94B5270ULL, "529587707.ogg"},
	{0xC1CC7F3AFBD6055DULL, "eyedebug.psh.fxo"},
	{0xC1CD11AC8F39F65CULL, "follower_banter6169797.cl"},
	{0xC1CD5A90EA33A12FULL, "lt_0_lowtown_docks_a3_162_0.mao"},
	{0xC1CD6690EA33B5CCULL, "lt_0_lowtown_docks_a3_162_0.mmh"},
	{0xC1CD6C90EA33BFBEULL, "lt_0_lowtown_docks_a3_162_0.msh"},
	{0xC1CD770731B073E9ULL, "vxlt_0_lowtown_a2_fee_0_0.vlm"},
	{0xC1CDAFD51CE7F973ULL, "6122238_f.fxe"},
	{0xC1CE870EB68770B2ULL, "pro000_colonel_ewald6039947.cl"},
	{0xC1CE8F0F4C839FBBULL, "l_dae330ar_undercity_dt.wbd"},
	{0xC1CEE9F9DD8826CAULL, "668478719.ogg"},
	{0xC1CF169487069235ULL, "darktown_male_sickly_sta6175311.cl"},
	{0xC1CFBA8EDCD3F162ULL, "6112042_m.fxe"},
	{0xC1CFBC938C2D2CDDULL, "6173358_m.fxe"},
	{0xC1D02247F648525DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pause.evt"},
	{0xC1D04DCEEF4896E8ULL, "17486035.ogg"},
	{0xC1D07D2135F4414CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_out.gad"},
	{0xC1D08BE6D87C1004ULL, "367522805.ogg"},
	{0xC1D17233FC49049DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_interact_low.ani"},
	{0xC1D1FC1F8BB56FE7ULL, "vxht_0_hightown_a2_by3_0_0.vlm"},
	{0xC1D2193C0BD5F8BDULL, "vxht_0_hightown_cbq_0_0.vlm"},
	{0xC1D229BA58C28C44ULL, "ser270_emeric6149371.cl"},
	{0xC1D25D892227EA6FULL, "hlac_w_blightlands_131074_0x.dds"},
	{0xC1D27CB85517B65CULL, "6054756_f.fxe"},
	{0xC1D2C152CAED4F3AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl.phy"},
	{0xC1D2D25350BC92F2ULL, "w_encounter2_21_0.rml"},
	{0xC1D3051D02C01D67ULL, "6087037_m.fxe"},
	{0xC1D31B43CA06F6CBULL, "art\\vfx\\combat\\vfx_f_and_fade_c.anb"},
	{0xC1D34E823995977FULL, "779782182.ogg"},
	{0xC1D46722CDBF1900ULL, "430071144.ogg"},
	{0xC1D48EB1F1F0ABADULL, "hm_gamlen_h2s2.mor"},
	{0xC1D49F08EBDDB0DBULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyn.mmh"},
	{0xC1D4AB08EBDDC578ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyn.mao"},
	{0xC1D4BD08EBDDE3C5ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyn.msh"},
	{0xC1D51381C1C130F8ULL, "6095222_m.fxe"},
	{0xC1D51A9253C9746AULL, "zz_dae_debug6047639.cl"},
	{0xC1D5EC3D91EBEAB1ULL, "w_sundermount_150_0.mmh"},
	{0xC1D5F83D91EBFF52ULL, "w_sundermount_150_0.mao"},
	{0xC1D6063D91EC176FULL, "w_sundermount_150_0.msh"},
	{0xC1D625E81F302E35ULL, "6085614_f.fxe"},
	{0xC1D6531EA7BCE81EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_reaction.ani"},
	{0xC1D703452A303C55ULL, "gen_im_trs_gen_two_2hm_r2.uti"},
	{0xC1D70825AC05F91FULL, "ser221ip_door_chase1.utp"},
	{0xC1D7419EDFBD52EDULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.impa.ani"},
	{0xC1D77C2135FA5E4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_right_out.ani"},
	{0xC1D7A82708A4C827ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_enter.evt"},
	{0xC1D7F3D6275C2DC0ULL, "137321656.ogg"},
	{0xC1D7F3F30F2BB596ULL, "zz_dae_debug6029288.cl"},
	{0xC1D85BABD5FCBB3FULL, "dae000_quintus6154706.cl"},
	{0xC1D86401779899EDULL, "vxht_bartrand_n_5in_0_0.vlm"},
	{0xC1D8D35350C0FF86ULL, "w_encounter2_21_0.phy"},
	{0xC1D9A1BDEEA10FD8ULL, "vxlt_0_lowtown_a2_n_f4p_0_0.vlm"},
	{0xC1D9C3A206039F87ULL, "gen00fl_merrill6162248.cl"},
	{0xC1D9F49485978BB9ULL, "65764366.ogg"},
	{0xC1DA0C4FF43C7185ULL, "vxdr_primevaltomb_7ln_0_0.vlm"},
	{0xC1DA93EB22A7429DULL, "vxw_bonepit_27h_0_0.vlm"},
	{0xC1DAA7862540B679ULL, "t3_cth_nb4.tnt"},
	{0xC1DB3289705F2D1CULL, "mer221_sabin6047677.cl"},
	{0xC1DB52D11BA5616EULL, "6087220_f.fxe"},
	{0xC1DB72FA52A8C24DULL, "vxlt_darktown_518_0_1.vlm"},
	{0xC1DC3C0908E9573EULL, "6034922_m.fxe"},
	{0xC1DC8C03019E1B36ULL, "vxht_keep_n_6ft_0_0.vlm"},
	{0xC1DC9911C7C7CCFDULL, "art\\characters\\playercharacter\\elffemale\\arm_lgte.phy"},
	{0xC1DC9F7203D458C7ULL, "490564508.ogg"},
	{0xC1DCD32636B7D4CAULL, "genip_exploding_barrel_whit.utp"},
	{0xC1DCF112BBFA7E39ULL, "gen_im_delete_me_07.uti"},
	{0xC1DD87DAEF99EF2BULL, "vxlt_warehouse_1da_0_0.vlm"},
	{0xC1DDC92708A96C70ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_2hands_on_hip_enter.gad"},
	{0xC1DE1A477D2809F1ULL, "vxht_mansion2b_3eh_0_0.vlm"},
	{0xC1DE23E875760380ULL, "vxlt_backalley_g_1oi_0_1.vlm"},
	{0xC1DE3E5EA38918E4ULL, "vxlt_darktown_5gr_0_0.vlm"},
	{0xC1DE3FF77C575392ULL, "dae000_flemeth6000927.cl"},
	{0xC1DE53F19F0ACA0EULL, "6110404_m.fxe"},
	{0xC1DEA83C6A6C5207ULL, "142364649.ogg"},
	{0xC1DF122C7A1AF359ULL, "970224496.ogg"},
	{0xC1DF17882935FE61ULL, "573870401.ogg"},
	{0xC1DF3746286B28C3ULL, "vxht_mansion2_1z5_0_0.vlm"},
	{0xC1E0BD6CB49A6721ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headmid.gad"},
	{0xC1E0D3FBAAB65AB2ULL, "hlac_ca_caverns1_24424_12x.dds"},
	{0xC1E1142DE4AFE088ULL, "ave200_aveline_bar6116603.cl"},
	{0xC1E1180059DD4A3DULL, "ht_bartrand_n_3368.rml"},
	{0xC1E1342553AD80B7ULL, "da2terrain_terrain_p1.vsh.fxo.dx11"},
	{0xC1E135B5E528AD4DULL, "384178950.ogg"},
	{0xC1E158E94F469D17ULL, "6196955_m.fxe"},
	{0xC1E20CA5BBF832ABULL, "ico_helm_light_5.dds"},
	{0xC1E288A5B49E8F0FULL, "vxht_brothel_n_5gf_0_0.vlm"},
	{0xC1E33E08439E1071ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thin.mmh"},
	{0xC1E34A08439E2412ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thin.mao"},
	{0xC1E35808439E3C2FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thin.msh"},
	{0xC1E362DF6461DA94ULL, "6086730_m.fxe"},
	{0xC1E3D2116320EA12ULL, "15596712.ogg"},
	{0xC1E3F4B5C8E6F3FCULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_bdy_gib_0d.dds"},
	{0xC1E46B1EA7C94EEFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_reaction.evt"},
	{0xC1E47568E5CB6630ULL, "llac_lt_0_lowtown_a2_n_131072_2.dds"},
	{0xC1E4BC10460C01E9ULL, "hlac_lt_0_lowtown_a2_n_c_65536_2x.dds"},
	{0xC1E5308A942D1D5DULL, "680045413.ogg"},
	{0xC1E57C323FAD21F8ULL, "vxdr_primevaltomb_7xs_0_0.vlm"},
	{0xC1E582AA2AEFC054ULL, "gen00fl_sebastian6174872.cl"},
	{0xC1E585AB253A07F9ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse1naked_orsinoadefiance.gad"},
	{0xC1E5C7E7E5A4EA77ULL, "vis121_saemus.bnk"},
	{0xC1E60C115E75F26FULL, "6079696_m.fxe"},
	{0xC1E6143D26F64974ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_headache.ani"},
	{0xC1E630A15878B217ULL, "gen00fl_fenris6112113.cl"},
	{0xC1E6FFB4EE5A1B24ULL, "vxlt_0_lowtown_a2_n_c_gz1_0_0.vlm"},
	{0xC1E72178571911DBULL, "vxdr_0_deeproads_bd7_0_0.vlm"},
	{0xC1E76E80823D4253ULL, "art\\cinematics\\_crit path\\followerintros\\cs_drunk2_isa_intro.ani"},
	{0xC1E780775911C061ULL, "lgt310_mis_burn_rival_thug6152781.cl"},
	{0xC1E7BE6CB4A087CFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\fh_ovr_headmid.ani"},
	{0xC1E7D2AAF90E1328ULL, "hlac_w_encounter2_65535_0z.dds"},
	{0xC1E7FEDBA86EC832ULL, "gen00fl_merrill6119630.cl"},
	{0xC1E81BF3032C3A22ULL, "719694318.ogg"},
	{0xC1E83416BC8598EEULL, "6165000_m.fxe"},
	{0xC1E85296BFF5346AULL, "hlac_ht_bartrand_n_3369_0z.dds"},
	{0xC1E88F158CBFEE41ULL, "6088830_m.fxe"},
	{0xC1E8A9E7E5A6EC8FULL, "vis121_saemus.cnv"},
	{0xC1E8F260B8CC372FULL, "llac_ht_playershouse_n_855_20.dds"},
	{0xC1E8F816F67A2F70ULL, "743017512.ogg"},
	{0xC1EA0B5E5F8AD7D2ULL, "6130099_m.fxe"},
	{0xC1EA4ADAE5279C6BULL, "vxht_keep_3ib_0_0.vlm"},
	{0xC1EA63A8CAC6BD1AULL, "vxht_0_hightown_a2_5eq_0_0.vlm"},
	{0xC1EA6C1EA7CDBB98ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\kiss_aggressive_reaction.gad"},
	{0xC1EA9FAF6C03FBC8ULL, "6138721_m.fxe"},
	{0xC1EAA9AE59EC52E4ULL, "6097116_m.fxe"},
	{0xC1EAF48FFC83318CULL, "6127066_m.fxe"},
	{0xC1EAFDDFF39433B6ULL, "vxht_keep_n_6au_0_0.vlm"},
	{0xC1EB27B4AD021179ULL, "6118191_m.fxe"},
	{0xC1EB96AB253E9047ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse1naked_orsinoadefiance.ani"},
	{0xC1EBB0F31FD764C9ULL, "art\\characters\\creatures\\highdragon\\cn_wng_hdr.mao"},
	{0xC1EBBD77975B3472ULL, "mcr370_guardsman6086980.cl"},
	{0xC1EC113D26FAAFD6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_headache.gad"},
	{0xC1EC5BA8DE6457CCULL, "6064055_m.fxe"},
	{0xC1EC637EFB145565ULL, "lgt250_way_dalish_conflict6147172.cl"},
	{0xC1EC93CC24EB9A86ULL, "vxlt_0_lowtown_a2_n_c_f5c_0_0.vlm"},
	{0xC1ECD80170A54793ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.cess.evt"},
	{0xC1ED2EB24EED52F1ULL, "hlac_lt_0_lowtown_n_-65536_5x.dds"},
	{0xC1ED52D9A006B146ULL, "6122830_m.fxe"},
	{0xC1ED65C66C170355ULL, "991085221.ogg"},
	{0xC1ED80B3BF13BF06ULL, "422185330.ogg"},
	{0xC1ED9A11C7D64102ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgte.mmh"},
	{0xC1ED9C11C7D644B8ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgte.msh"},
	{0xC1EE14FBD387DA38ULL, "vxht_playershouse_n_1r0_0_0.vlm"},
	{0xC1EE5966A0E61538ULL, "zz_followers_banter6158715.cl"},
	{0xC1EEF5B5A39C2A90ULL, "vxw_bonepitdestroyed_a3_2ac_0_0.vlm"},
	{0xC1F0DF87A2E68C7CULL, "hlac_lt_0_lowtown_docks_a3_-65538_1x.dds"},
	{0xC1F0E02F511E81F0ULL, "vxht_mansion_n_6s9_0_0.vlm"},
	{0xC1F0F3909CCE9BFFULL, "206534591.ogg"},
	{0xC1F18F5AC476B1A6ULL, "hlac_w_bonepit_131073_0y.dds"},
	{0xC1F19056C5667471ULL, "hm_icarver_h2s2.mor"},
	{0xC1F21AF6CE7AF610ULL, "147083443.ogg"},
	{0xC1F24FD3AC65134BULL, "6149850_m.fxe"},
	{0xC1F285C9123943FCULL, "one121ip_at_lowtown.utp"},
	{0xC1F2D10170A9A7D4ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.cess.gad"},
	{0xC1F32C3D2700FC35ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_headache.evt"},
	{0xC1F388420FDFB167ULL, "gen00fl_aveline6139322.cl"},
	{0xC1F3CEE7A316CFC1ULL, "m_one130ar_booker_dungeon.wbd"},
	{0xC1F3EE0F31A33A78ULL, "vxdr_primevaltomb_7up_0_0.vlm"},
	{0xC1F47068B54FD39FULL, "zz_followers_banter6158268.cl"},
	{0xC1F4F571A02B4C53ULL, "mag100_cullen6082276.cl"},
	{0xC1F53E71DB32C86CULL, "dae011_drunk2_sta6149693.cl"},
	{0xC1F56010C0BA8BD4ULL, "gen00fl_fenris6126810.cl"},
	{0xC1F681EFE5946BABULL, "art\\characters\\creatures\\arh_abm_rev\\c_arh_body.mao"},
	{0xC1F6A8A1CA7F6F8EULL, "mer141_jansen.cnv"},
	{0xC1F6F009D828A285ULL, "vxdr_primevaltomb_7bl_0_0.vlm"},
	{0xC1F6F9104607F1E2ULL, "dae300_cullen6140585.cl"},
	{0xC1F70608FD11BF77ULL, "vxlt_0_lowtown_docks_n_cwj_0_0.vlm"},
	{0xC1F7260592C8CB02ULL, "ht_bartrand_n_3381.rml"},
	{0xC1F786F457190765ULL, "hlac_lt_0_lowtown_docks_n_-196610_2z.dds"},
	{0xC1F82029C6F7A50BULL, "lt_gallowscourtyard_31_0.rml"},
	{0xC1F868FBE9177DBDULL, "87820169.ogg"},
	{0xC1F87D8813A5F8AAULL, "vxw_encounter2_60_0_0.vlm"},
	{0xC1F898C0166E5651ULL, "6118595_m.fxe"},
	{0xC1F8CEF23A653DB4ULL, "154916293.ogg"},
	{0xC1F8EBE9060DA998ULL, "w_woundedcoast_44_0.rml"},
	{0xC1F9CAA1CA81DFFEULL, "mer141_jansen.bnk"},
	{0xC1F9E00170AFDFB2ULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.cess.ani"},
	{0xC1F9E6455972FFF5ULL, "vxlt_0_lowtown_a3_n_c_e1n_0_0.vlm"},
	{0xC1FA344F47F9F226ULL, "vxdr_primevalentrance_2og_0_2.vlm"},
	{0xC1FA59A06A8DCB42ULL, "6123428_m.fxe"},
	{0xC1FA59BD900F2DB9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medb_0s.dds"},
	{0xC1FA7A4E0EB7F42DULL, "lt_0_lowtown_docks_n_3_0_ll.mmh"},
	{0xC1FAA0536D4DE20EULL, "blood_cycle_sm.dds"},
	{0xC1FAA46ED1472D0CULL, "6063911_m.fxe"},
	{0xC1FABAEC09BF867FULL, "6084366_m.fxe"},
	{0xC1FB0C9B1312D615ULL, "dae300_vashothlead6062815.cl"},
	{0xC1FB23F7FC302484ULL, "vxht_mansion_n_60t_0_0.vlm"},
	{0xC1FB3062A22B7D79ULL, "w_encounter2_38_0.msh"},
	{0xC1FB3A62A22B8E67ULL, "w_encounter2_38_0.mmh"},
	{0xC1FB3E62A22B94B4ULL, "w_encounter2_38_0.mao"},
	{0xC1FB4580824DA5FDULL, "art\\cinematics\\_crit path\\followerintros\\cs_drunk2_isa_intro.gad"},
	{0xC1FB781B53F66BB2ULL, "6071280_m.fxe"},
	{0xC1FC09ECC1F2DE4CULL, "vxht_0_hightown_a2v_0_0.vlm"},
	{0xC1FC38CC60D44336ULL, "crt_high_dragon.wbd"},
	{0xC1FC56213C847DD9ULL, "vxlt_gallowstemplar_f_5az_0_0.vlm"},
	{0xC1FCD3C8EF9AFC37ULL, "6121157_m.fxe"},
	{0xC1FCE2253E337BF7ULL, "583652163.ogg"},
	{0xC1FD085335AA4DB0ULL, "vxht_0_hightown_a2_bi8_0_0.vlm"},
	{0xC1FD8E7B97F7401CULL, "6036939_m.fxe"},
	{0xC1FDC420D985B26FULL, "m_qun131ar_undercity_warrens.bnk"},
	{0xC1FDD19811DED35CULL, "vxca_caverns2_8gs_0_0.vlm"},
	{0xC1FDD19D0E9AD755ULL, "vxdr_primevaltomb_342_0_0.vlm"},
	{0xC1FE3E7ABD6C04F9ULL, "vxw_bonepitdestroyed_a3_1jd_0_0.vlm"},
	{0xC1FEDDBFDA90D334ULL, "art\\vfx\\creatures\\fxc_withfn_c.cess.ani"},
	{0xC1FF38132B0E6845ULL, "zz_mer_debug6062743.cl"},
	{0xC1FF492A93779973ULL, "166204112.ogg"},
	{0xC1FFA982084A8C8FULL, "6098719_f.fxe"},
	{0xC2009EFFA894843EULL, "vxca_cave_2e2_0_0.vlm"},
	{0xC200A762E2DB1588ULL, "6108274_m.fxe"},
	{0xC201049D21AC1DB5ULL, "gen00fl_aveline6106388.cl"},
	{0xC201424A00CA7B39ULL, "617570429.ogg"},
	{0xC201E01E9DCA302DULL, "pro000st_cavril_store.ncs"},
	{0xC202227CDFBCA410ULL, "979079548.ogg"},
	{0xC2025A8A6D629B64ULL, "896579529.ogg"},
	{0xC202BECC656DB4D6ULL, "6033968_transitionfixed.cut"},
	{0xC203C3CB7F55149BULL, "hlac_lt_0_lowtown_a2_-65534_0z.dds"},
	{0xC203F72D984D2D3BULL, "dae000ip_crowd_em3_talk9.utp"},
	{0xC20425C11603D8F0ULL, "1015229765.ogg"},
	{0xC2050972B7C0561EULL, "330690722.ogg"},
	{0xC205958D9547C43DULL, "dae000_maren6148419.cl"},
	{0xC205DABFDA96EC96ULL, "art\\vfx\\creatures\\fxc_withfn_c.cess.gad"},
	{0xC206051DBFE85C8DULL, "6118414_m.fxe"},
	{0xC2063CF749295E6EULL, "240388935.ogg"},
	{0xC206EF65059C112AULL, "vxw_encounter2_11_0_0.vlm"},
	{0xC20735B6BDE5614BULL, "llac_ht_mansion2b_818_52.dds"},
	{0xC2074E43B6953D93ULL, "one130st_booker.stg"},
	{0xC2075F76D55D10B9ULL, "vxlt_gallowstemplar_f_3or_0_0.vlm"},
	{0xC207722069AE34C2ULL, "6076601_m.fxe"},
	{0xC207B73A3EC61C6CULL, "mag310cr_scion_1.utc"},
	{0xC207DCD7FA24E5EDULL, "826372777.ogg"},
	{0xC2082606ADEF5E65ULL, "319444141.ogg"},
	{0xC208289EA53DAA88ULL, "w_bonepitmining_a2_25_0.rml"},
	{0xC208CB7D5C83E3C2ULL, "6128528_m.fxe"},
	{0xC208D8C228B866FBULL, "6017436_m.fxe"},
	{0xC208E421FF48ACFBULL, "1052001978.ogg"},
	{0xC2093F5AB97BA431ULL, "vxlt_0_lowtown_docks_bup_0_0.vlm"},
	{0xC20949520AE0B776ULL, "6139118_f.fxe"},
	{0xC20954BB72D7A434ULL, "qun110tr_amb_test.utt"},
	{0xC2095C520B289FA6ULL, "pro000_meeran6050104.cl"},
	{0xC209AE93AE0289E4ULL, "62291006.ogg"},
	{0xC20A0F5F5036B772ULL, "385712806.ogg"},
	{0xC20A72240F2D6004ULL, "6121086_m.fxe"},
	{0xC20A77277CB86321ULL, "6103357_m.fxe"},
	{0xC20A791F6946DD35ULL, "hlac_lt_0_lowtown_a2_n_131070_1x.dds"},
	{0xC20A8CDCFE4D7243ULL, "vxht_mansion2_1k6_0_0.vlm"},
	{0xC20B3C778140D458ULL, "333827107.ogg"},
	{0xC20B853D2981E09FULL, "493800533.ogg"},
	{0xC20BAB06DD562722ULL, "physics2.psh.fxo"},
	{0xC20BB207530A0B0DULL, "vxdr_primevalentrance_2gb_0_0.vlm"},
	{0xC20CF5BFDA9D39F5ULL, "art\\vfx\\creatures\\fxc_withfn_c.cess.evt"},
	{0xC20D436C5DA81DC0ULL, "vxht_keep_n_604_0_0.vlm"},
	{0xC20DAF70F9D55975ULL, "art\\vfx\\textures\\vfx_wispy_cs.dds"},
	{0xC20E2C306772CFAFULL, "125907577.ogg"},
	{0xC20E5B3844ECF2F9ULL, "butterfly02.dds"},
	{0xC20E643C46C9DE34ULL, "6074487_m.fxe"},
	{0xC20E6935A6961DDFULL, "6139563_m.fxe"},
	{0xC20F543A9AE28A06ULL, "461496495.ogg"},
	{0xC21069C1EE0D97E3ULL, "mag320pt_showdown.ncs"},
	{0xC210A6753CD23851ULL, "ave110tr_approach_donnic.utt"},
	{0xC210F0E75127B2CBULL, "534748918.ogg"},
	{0xC210F73739B67612ULL, "6065915_m.fxe"},
	{0xC2111081FB8A71E3ULL, "hlac_ht_chantry_2534_39x.dds"},
	{0xC2112D777E214FBBULL, "character_character_p82.psh.fxo.dx11"},
	{0xC2115E86B6CA830EULL, "vxht_mansion_n_5ar_0_0.vlm"},
	{0xC2117D2AFFA5F515ULL, "zz_ser_debug6078148.cl"},
	{0xC21198F2EDA0E3DAULL, "vxlt_0_lowtown_byn_0_0.vlm"},
	{0xC212244C16783981ULL, "467497655.ogg"},
	{0xC21262B0D7495C62ULL, "one100_party_comments6112612.cl"},
	{0xC212FA9D8CBE47A9ULL, "6126966_f.fxe"},
	{0xC213AE0755088BFBULL, "223511180.ogg"},
	{0xC213CB7B5F78AC09ULL, "hlac_lt_0_lowtown_a2_-3_4z.dds"},
	{0xC213EC214506459FULL, "w_bonepitmining_a2_241_0.msh"},
	{0xC213F22145065041ULL, "w_bonepitmining_a2_241_0.mmh"},
	{0xC213FE21450664E2ULL, "w_bonepitmining_a2_241_0.mao"},
	{0xC21417005A08A255ULL, "ht_bartrand_n_3368.anb"},
	{0xC214250DB5D6A080ULL, "412737103.ogg"},
	{0xC21476FAB1DE750DULL, "6117514_m.fxe"},
	{0xC214A4D3216DAB3DULL, "6156030_m.fxe"},
	{0xC2150E968E211AAEULL, "mcr370_meredith6081086.cl"},
	{0xC2167CD23E7D33C2ULL, "651596.ogg"},
	{0xC21758A5306A8920ULL, "vxca_caverns1_mines_i2s_0_0.vlm"},
	{0xC217848BC5F4FA17ULL, "hlac_ht_bartrand_n_2995_7z.dds"},
	{0xC217C5830681FA1AULL, "vxht_mansion2_2v3_0_0.vlm"},
	{0xC217E9E906285775ULL, "w_woundedcoast_44_0.msh"},
	{0xC217EBE906285AEBULL, "w_woundedcoast_44_0.mmh"},
	{0xC217F7E906286E88ULL, "w_woundedcoast_44_0.mao"},
	{0xC2181421527D2C85ULL, "73149618.ogg"},
	{0xC21832EF84A16F7EULL, "6106177_m.fxe"},
	{0xC218362E1F0C7314ULL, "vxw_bonepit_16q_0_0.vlm"},
	{0xC218F0E7A33625C3ULL, "m_one130ar_booker_dungeon.bnk"},
	{0xC21902D9307BCD3EULL, "6156856_f.fxe"},
	{0xC21939DB2070AED9ULL, "vxlt_gallowstemplar_f_5ox_0_0.vlm"},
	{0xC2194B325ED22CC9ULL, "698042042.ogg"},
	{0xC2196A865D281D86ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspike.phy"},
	{0xC219AB3CFEBE6AA0ULL, "hlac_lt_0_lowtown_docks_a3_65534_4x.dds"},
	{0xC21AABFC496F37DCULL, "vxht_0_hightown_bfu_0_0.vlm"},
	{0xC21AB551F5CB045CULL, "vxca_caverns2_8iq_0_0.vlm"},
	{0xC21BE43EC6E7EC00ULL, "hlac_ht_chantry_n_2534_11x.dds"},
	{0xC21BEE62B2501528ULL, "pro000_meeran6050076.cl"},
	{0xC21BF47363603BB6ULL, "llac_lt_0_lowtown_a2_131072_5.dds"},
	{0xC21BF6962C16AD73ULL, "vxw_encounter2_camp_p3_0_0.vlm"},
	{0xC21C3F975BC71BD2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple_0.pwk"},
	{0xC21C5C727337CA7CULL, "w_sundermount_167_0.mao"},
	{0xC21C68727337DE1FULL, "w_sundermount_167_0.mmh"},
	{0xC21C6E727337E8C1ULL, "w_sundermount_167_0.msh"},
	{0xC21CC008EC1B9E10ULL, "art\\characters\\playercharacter\\dwarfmale\\hlf_hvyn.phy"},
	{0xC21D827C547981DBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_basic.mmh"},
	{0xC21D8E7C54799678ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_basic.mao"},
	{0xC21DA07C5479B4C5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_basic.msh"},
	{0xC21DC15D70DC5EE3ULL, "mrl251ip_loot_a.utp"},
	{0xC21E3A46AD7F6993ULL, "6114064_f.fxe"},
	{0xC21F1F856632690CULL, "llac_lt_0_lowtown_a2_n_-65536_2.dds"},
	{0xC21F4702454CAEE5ULL, "6103322_m.fxe"},
	{0xC21F4905FB832BAEULL, "dae000_bodhan_letter6144306.cl"},
	{0xC21F8E5007045266ULL, "6174183_m.fxe"},
	{0xC21FAC043134EEF9ULL, "zz_one_debug6073983.cl"},
	{0xC220219D36EB80C9ULL, "6119754_m.fxe"},
	{0xC22074E64BB81683ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_001a_0n.dds"},
	{0xC2208CBDC2DE69B8ULL, "tbe200im_fate_gloves.uti"},
	{0xC2209AF05B20D790ULL, "6155742_m.fxe"},
	{0xC220FDD5B1170E29ULL, "6138760_m.fxe"},
	{0xC22245A0AE0AEC05ULL, "vxdr_primevaltomb_7ui_0_0.vlm"},
	{0xC222B0F278AA5446ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_masa_0d.dds"},
	{0xC222CB9593F217A7ULL, "vxca_cavern1_haunted_gkl_0_1.vlm"},
	{0xC222E8D7A3D48D8DULL, "6112614_m.fxe"},
	{0xC2238C4F49D0CB21ULL, "vxht_keep_n_623_0_0.vlm"},
	{0xC2240F81A07AFC90ULL, "lt_0_lowtown_docks_171_0.mao"},
	{0xC2241381A07B03A3ULL, "lt_0_lowtown_docks_171_0.mmh"},
	{0xC2242181A07B1BFDULL, "lt_0_lowtown_docks_171_0.msh"},
	{0xC2244330B94941B9ULL, "vxlt_gallowstemplar_f_3qp_0_0.vlm"},
	{0xC2244B7956BB0846ULL, "417415512.ogg"},
	{0xC2246E7A493C5211ULL, "6151383_m.fxe"},
	{0xC2247C03D1663E29ULL, "159246918.ogg"},
	{0xC2248375BE088B7FULL, "6075159_m.fxe"},
	{0xC224AC92F32FEDFBULL, "hlac_ht_0_hightown_a3_-65536_1x.dds"},
	{0xC2252BC74265DB0FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_end.gad"},
	{0xC225602D89ABFFB0ULL, "vxlt_0_lowtown_a2_hd9_0_0.vlm"},
	{0xC225A63D7E6180CCULL, "art\\vfx\\combat\\vfx_r_var_activation1_c.impa.ani"},
	{0xC226408CD1D3675EULL, "genip_bottle.utp"},
	{0xC2270D1A7C05243AULL, "257548159.ogg"},
	{0xC2272362A250A19CULL, "w_encounter2_38_0.phy"},
	{0xC22784A8DF5A8E01ULL, "dae300_cullen6140661.cl"},
	{0xC2285ECC60F9BEA4ULL, "crt_high_dragon.bnk"},
	{0xC228880742F90861ULL, "mag170_hugh6206267.cl"},
	{0xC228B3641B793A70ULL, "lt_0_lowtown_docks_203_0.mao"},
	{0xC228B7641B794083ULL, "lt_0_lowtown_docks_203_0.mmh"},
	{0xC228C5641B7958DDULL, "lt_0_lowtown_docks_203_0.msh"},
	{0xC2293E26E9234C5FULL, "dre000_stroud6036740.cl"},
	{0xC229BC451CAF18A2ULL, "character_character_p7.psh.fxo.dx11"},
	{0xC22A68731E0CC5F3ULL, "vxw_encounter2_camp_k2_0_0.vlm"},
	{0xC22AA82E9B8E6E74ULL, "vxht_keep_n_3he_0_0.vlm"},
	{0xC22ADA4890DC4DFCULL, "6092098_m.fxe"},
	{0xC22AFBFAAB955011ULL, "one101_orders_followers6115783.cl"},
	{0xC22B40C7426A6ABCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_end.evt"},
	{0xC22B702FBE0EEB5AULL, "351201129.ogg"},
	{0xC22B88B389B738C4ULL, "mag370_meredith6065007.cl"},
	{0xC22C093730A2E440ULL, "vxht_mansion_n_6ka_0_0.vlm"},
	{0xC22C1611FE489B21ULL, "579402114.ogg"},
	{0xC22CACA400F6051EULL, "vxlt_gallowstemplar_33j_0_0.vlm"},
	{0xC22CB33D7E67B50EULL, "art\\vfx\\combat\\vfx_r_var_activation1_c.impa.gad"},
	{0xC22D3F25F0493D23ULL, "397886077.ogg"},
	{0xC22D57272721572FULL, "737519186.ogg"},
	{0xC22D6B881BE8C69DULL, "6078792_m.fxe"},
	{0xC22D8C334E5134BFULL, "zz_one_debug6058986.cl"},
	{0xC22D8F76C7A64966ULL, "6032866_f.fxe"},
	{0xC22DA6AC9D85DCA4ULL, "6153692_m.fxe"},
	{0xC22E3A62A256E7C4ULL, "w_encounter2_38_0.rml"},
	{0xC22E937A28E19949ULL, "6114054_m.fxe"},
	{0xC22EA43D10417151ULL, "6087206_m.fxe"},
	{0xC22EBA52CB3BA263ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl.msh"},
	{0xC22EC6FD63B646E6ULL, "6031990_m.fxe"},
	{0xC22EC852CB3BB9BDULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl.mmh"},
	{0xC22ECC52CB3BC0CEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl.mao"},
	{0xC22ED21BBCB29040ULL, "6112138_m.fxe"},
	{0xC22F23A63F985A2EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_pick_up_from_ground.ani"},
	{0xC22F29DE5A50221FULL, "731123880.ogg"},
	{0xC230520D1ABBF2C2ULL, "vxw_encounter1_bm_0_0.vlm"},
	{0xC2305994E7191DE8ULL, "dae000_pc_home_ambients6208192.cl"},
	{0xC2305DE8CE8CD824ULL, "6166191_m.fxe"},
	{0xC2305F1D6F01627BULL, "vxdr_0_deeproads_eqc_0_0.vlm"},
	{0xC230A2CCFC0E409EULL, "hlac_ht_chantry_2534_50x.dds"},
	{0xC230D9806654FD96ULL, "hlac_lt_gamlenshouse_n_-1_0x.dds"},
	{0xC231230592FA3EFEULL, "ht_bartrand_n_3381.anb"},
	{0xC231DF3375C2630BULL, "66727887.ogg"},
	{0xC232985805C28F2AULL, "vxw_blightlands_l9_0_0.vlm"},
	{0xC232AE3D7E6C18EDULL, "art\\vfx\\combat\\vfx_r_var_activation1_c.impa.evt"},
	{0xC2332317D7A815EEULL, "combat_dw_r1_a1.gda"},
	{0xC233665B2B00E9B3ULL, "864365702.ogg"},
	{0xC23372D088CB25C9ULL, "601240505.ogg"},
	{0xC23399C94F773279ULL, "vxca_caverns2_4x6_0_1.vlm"},
	{0xC234851D5671B260ULL, "vxlt_0_lowtown_a2_n_c_ep0_0_1.vlm"},
	{0xC234A0E61416465CULL, "art\\levels\\races\\dwarf\\props\\lyrium\\0ww_broken.mao"},
	{0xC234ACE614165AFFULL, "art\\levels\\races\\dwarf\\props\\lyrium\\0ww_broken.mmh"},
	{0xC234B2E6141664A1ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\0ww_broken.msh"},
	{0xC234DFBC9C7EEEC2ULL, "follower_banter6155604.cl"},
	{0xC234F125384718AEULL, "6169350_m.fxe"},
	{0xC2350C43125186F7ULL, "6044616_m.fxe"},
	{0xC235667E248C3F71ULL, "6101218_m.fxe"},
	{0xC235876C07698D35ULL, "vxlt_blackemporium_u5_0_0.vlm"},
	{0xC235C17527CB41ACULL, "862567721.ogg"},
	{0xC236562A2ED934DDULL, "vxlt_0_lowtown_hak_0_0.vlm"},
	{0xC23663EFBFA31E66ULL, "art\\vfx\\materials\\vfx_flash_blend.mao"},
	{0xC236A29B97CFE998ULL, "follower_banter6161058.cl"},
	{0xC23731F2769CC346ULL, "fen110_anso6173730.cl"},
	{0xC2374533D3B1F420ULL, "6076012_m.fxe"},
	{0xC23754DD588F8891ULL, "vxlt_hangedman_10k_0_0.vlm"},
	{0xC2375DAC959C03C8ULL, "pro000_cut_innocent_blood.cnv"},
	{0xC237F7143BAA00AAULL, "6084262_m.fxe"},
	{0xC238DA501002DE73ULL, "vxw_encounter2_camp_f5_0_0.vlm"},
	{0xC238E171F85F3F98ULL, "551752636.ogg"},
	{0xC23948C74276B5D5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_end.ani"},
	{0xC239B18F4DD6F457ULL, "art\\levels\\races\\design\\_textures\\prp_chestc_d.dds"},
	{0xC239C7F09CD77585ULL, "llac_lt_0_lowtown_-1_6.dds"},
	{0xC23A5DE264AD0E62ULL, "vxlt_0_lowtown_a2_n_d1w_0_0.vlm"},
	{0xC23A79EB224DB4C1ULL, "6106192_m.fxe"},
	{0xC23A7B142298FCC0ULL, "vxht_mansion_n_6nn_0_0.vlm"},
	{0xC23A7FAC959E7330ULL, "pro000_cut_innocent_blood.bnk"},
	{0xC23B7487FF033113ULL, "lt_gallowstemplar_n_mmap_ph.dds"},
	{0xC23B93044B511F4CULL, "vxlt_gallowstemplar_3j7_0_0.vlm"},
	{0xC23C206DCC44043BULL, "815661635.ogg"},
	{0xC23CB41D8CACCC8EULL, "vxht_mansion_n_5pu_0_0.vlm"},
	{0xC23CEA0C78A4EA29ULL, "abi_m_stone_fist.bnk"},
	{0xC23D1BA63FA48A7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_pick_up_from_ground.evt"},
	{0xC23D56B50FDF3708ULL, "m_one241ar_carta_cave.wbd"},
	{0xC23DC3DA96D481C5ULL, "hlac_ca_cavern1_haunted_24421_0z.dds"},
	{0xC23E47C898DBE5FDULL, "w_sundermount_225_0.mao"},
	{0xC23E4BC898DBEC36ULL, "w_sundermount_225_0.mmh"},
	{0xC23E55C898DBFD34ULL, "w_sundermount_225_0.msh"},
	{0xC23E68C77BC21AF4ULL, "hlac_lt_0_lowtown_n_1_3x.dds"},
	{0xC23EC3EA0CB20B42ULL, "vxw_encounter1_gn_0_0.vlm"},
	{0xC23EF999D8C2D054ULL, "112028033.ogg"},
	{0xC23EFAC0F0C1502FULL, "vxdr_primevaltomb_4hs_0_0.vlm"},
	{0xC23F28EF874D5497ULL, "w_encounter3_174_0.msh"},
	{0xC23F3AEF874D73FAULL, "w_encounter3_174_0.mao"},
	{0xC23F3EEF874D7A09ULL, "w_encounter3_174_0.mmh"},
	{0xC23F437B6749B2E6ULL, "463966170.ogg"},
	{0xC23F5137A8CC309AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_medj_0d.dds"},
	{0xC23FACD63F69331DULL, "6067121_m.fxe"},
	{0xC23FDC39EA79B093ULL, "vxht_0_hightown_a2_b1w_0_0.vlm"},
	{0xC2403C4D09D515B6ULL, "zz_qun_debug6036826.cl"},
	{0xC240BCB8DE73E3BAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffold.mmh"},
	{0xC240C0B8DE73EAF1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffold.mao"},
	{0xC240CEB8DE740200ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffold.msh"},
	{0xC24193F560B2971AULL, "6057572_m.fxe"},
	{0xC241A0E5DBCCC605ULL, "6041402_m.fxe"},
	{0xC241A3E33ACC9C7EULL, "6167698_f.fxe"},
	{0xC241A59C6927E19CULL, "gen00fl_anders6125303.cl"},
	{0xC241A82C174A46ABULL, "6128356_m.fxe"},
	{0xC2422620EDC1214CULL, "6091874_f.fxe"},
	{0xC242303B717B4F66ULL, "vxw_encounter3_74_0_0.vlm"},
	{0xC2426A46F16366DCULL, "dae000_meredith6166275.cl"},
	{0xC242B2A5D7D98F74ULL, "mag210_arianni6024090.cl"},
	{0xC2432138AADB1E9BULL, "gen00fl_varric6033726.cl"},
	{0xC243488857E00439ULL, "mag101_grace6035870.cl"},
	{0xC243FCA63FAA7468ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_pick_up_from_ground.gad"},
	{0xC2441686B226923FULL, "6140080_m.fxe"},
	{0xC244A89B6F93CDAEULL, "vxw_blightlands_w8_0_0.vlm"},
	{0xC24572AC9D7D7FDDULL, "follower_banter6171122.cl"},
	{0xC2458321B7F821CCULL, "6175982_m.fxe"},
	{0xC2460F058B6D7742ULL, "6129818_f.fxe"},
	{0xC246D48E9FB649DDULL, "6173149_m.fxe"},
	{0xC2472A5286A363D5ULL, "277408298.ogg"},
	{0xC2473EDFA9C5220FULL, "one000_ilen6036939.cl"},
	{0xC2474C2D01F8F6F3ULL, "vxw_encounter2_camp_a4_0_0.vlm"},
	{0xC24768C849F30DC1ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_sledge.phy"},
	{0xC247FB0913D544C8ULL, "mag321_comtesse_de_launcet6064934.cl"},
	{0xC2482A339FA75028ULL, "gen00fl_anders6111263.cl"},
	{0xC2488E7288D1FF00ULL, "gen00fl_bethany6101444.cl"},
	{0xC249175E63B06694ULL, "687960602.ogg"},
	{0xC24920CF785B9065ULL, "vxht_keep_n_63l_0_0.vlm"},
	{0xC24999F52B5FCABEULL, "lt_gallowstemplar_n_7198.rml"},
	{0xC249A2CDA1FA5E16ULL, "additivering02.mao"},
	{0xC249A9B5705F7F6BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gear_sa.dds"},
	{0xC24A59535B9429EAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_tmpa_0d.dds"},
	{0xC24AC8A014B851ACULL, "vxlt_0_lowtown_a2_n_fiw_0_0.vlm"},
	{0xC24ADDAF58B6A199ULL, "store_dae200cr_hubert.utm"},
	{0xC24BD02FEBD0712BULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_follower_m_.ani"},
	{0xC24BE494C7BBFA28ULL, "vxlt_0_lowtown_a3_n_c_e31_0_0.vlm"},
	{0xC24C6505D839F7FDULL, "229567848.ogg"},
	{0xC24C70C8BB6B442EULL, "hlac_lt_gallowstemplar_f_2391_6z.dds"},
	{0xC24C72F551DAB10DULL, "240396878.ogg"},
	{0xC24CB535420ADF2BULL, "vxlt_0_lowtown_docks_a3_n_d0z_0_1.vlm"},
	{0xC24CD14304FBD2BBULL, "art\\vfx\\combat\\vfx_gib_hlk_p_v1.anb"},
	{0xC24CD322631226C0ULL, "266032670.ogg"},
	{0xC24CED9B01BC921EULL, "vxlt_0_lowtown_bz3_0_0.vlm"},
	{0xC24D094D0FD64041ULL, "vxht_chantry_149_0_0.vlm"},
	{0xC24D24A81AE60AE8ULL, "997556131.ogg"},
	{0xC24D42D752ED937BULL, "vxdr_0_deeproads_ege_0_0.vlm"},
	{0xC24D4E51F351A482ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_4x.dds"},
	{0xC24DA7F6AFAB0E76ULL, "vxht_0_hightown_5wg_0_0.vlm"},
	{0xC24DDCC2407B5F68ULL, "gen00fl_anders6125260.cl"},
	{0xC24DEB4ACF50CC2DULL, "w_blightlands_228_0.mmh"},
	{0xC24DEF4ACF50D37EULL, "w_blightlands_228_0.mao"},
	{0xC24DFD4ACF50EAB3ULL, "w_blightlands_228_0.msh"},
	{0xC24E37694E3886A9ULL, "vxht_keep_6c3_0_0.vlm"},
	{0xC24EC2F19D653D1CULL, "ser221st_gascard_final.stg"},
	{0xC2500A76AF283180ULL, "vxdr_0_deeproads_9cb_0_0.vlm"},
	{0xC25181B266F4D959ULL, "sib100_father_contact6109257.cl"},
	{0xC251C846A90CF6D1ULL, "6078114_m.fxe"},
	{0xC251CFA7F5C329BFULL, "w_woundedcoast_222_0.msh"},
	{0xC251D5A7F5C33461ULL, "w_woundedcoast_222_0.mmh"},
	{0xC251E11C1CE13E41ULL, "vxlt_gallowstemplar_f_5qb_0_0.vlm"},
	{0xC251E1A7F5C34802ULL, "w_woundedcoast_222_0.mao"},
	{0xC251F0EC80261DC7ULL, "888923591.ogg"},
	{0xC25231D99AF7CDCDULL, "6078104_f.fxe"},
	{0xC2528E1C76BD0ED0ULL, "vxca_cavern1_haunted_j4j_0_0.vlm"},
	{0xC252C87BC8941189ULL, "follower_friendship.dds"},
	{0xC252EB1E6AE66564ULL, "gen00fl_anders6121167.cl"},
	{0xC252F620856B58F4ULL, "vxlt_0_lowtown_gbx_0_0.vlm"},
	{0xC25330AB856F5AD6ULL, "vxlt_0_lowtown_docks_a3_n_cyb_0_0.vlm"},
	{0xC2533F8426AF52EEULL, "character_character_p177.psh.fxo.dx11"},
	{0xC25342500DDC2CE9ULL, "196275267.ogg"},
	{0xC253440280C2DC95ULL, "qcr220_orsino6061791.cl"},
	{0xC2534F77F3241FD6ULL, "387556073.ogg"},
	{0xC25368886CEE9155ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bazaar_n.dds"},
	{0xC25378D68B1AC5A9ULL, "vxht_0_hightown_n_bp9_0_0.vlm"},
	{0xC253AFC9F8D7B11DULL, "hlac_lt_undercitydungeon_-1_10y.dds"},
	{0xC253E953157389B8ULL, "1002281819.ogg"},
	{0xC25415DDD0914013ULL, "6172588_m.fxe"},
	{0xC2542E930B06AFE5ULL, "mag130st_danzig.stg"},
	{0xC2557E8030809D5DULL, "6163287_m.fxe"},
	{0xC255815DC3AECB10ULL, "ht_mansion2b_probe_1ng.dds"},
	{0xC256043DBF1A85EDULL, "llac_lt_0_lowtown_a2_-65535_1.dds"},
	{0xC256342B984A7900ULL, "gen00fl_aveline6122095.cl"},
	{0xC2569C7ECEA8EE28ULL, "910407223.ogg"},
	{0xC256B76A7AE5A0B2ULL, "302623576.ogg"},
	{0xC256EF98C12E3EF1ULL, "420860315.ogg"},
	{0xC25703A0EA3C5A0EULL, "gen00fl_aveline6139299.cl"},
	{0xC2571EC1A1AC5E4DULL, "dre000_bodahn6028162.cl"},
	{0xC257338EF670DE6DULL, "356430714.ogg"},
	{0xC2576E39C05D1973ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_door_male.evt"},
	{0xC257908BC49AB812ULL, "ico_movetopoint.dds"},
	{0xC257DCCE247358BDULL, "385706992.ogg"},
	{0xC258B05DD0492B2DULL, "6170979_m.fxe"},
	{0xC258D7E3DAA4D9C5ULL, "vxht_0_hightown_a3_b8w_0_0.vlm"},
	{0xC258F1BE97E82980ULL, "large_sa.dds"},
	{0xC2593EC8BE61CC9BULL, "vxca_caverns2prc_9fo_0_0.vlm"},
	{0xC25949DF3FE867E5ULL, "vxlt_0_lowtown_bxr_0_0.vlm"},
	{0xC25961B494014721ULL, "810215600.ogg"},
	{0xC259AEE6459D9D17ULL, "6037333_m.fxe"},
	{0xC259EBDC4F044765ULL, "lt_0_lowtown_docks_a3_n_24_0_ll.mmh"},
	{0xC25AAC23DAFE84B7ULL, "429665937.ogg"},
	{0xC25ACFABBE520CB1ULL, "art\\cinematics\\act1\\cs_fp_smirk.ani"},
	{0xC25B03053CD186D3ULL, "6166893_m.fxe"},
	{0xC25B2ED758595333ULL, "nat380_sibling6103583.cl"},
	{0xC25BA6C551EEC9D9ULL, "vxca_caverns1_j2a_0_1.vlm"},
	{0xC25C49F4F2CD77C2ULL, "vxdr_primevaltomb_9cv_0_0.vlm"},
	{0xC25CBE163217E1A2ULL, "vxht_chantry_35g_0_0.vlm"},
	{0xC25CEA994646D5C3ULL, "vxht_bartrand_n_5qx_0_0.vlm"},
	{0xC25D277DB325C261ULL, "59351594.ogg"},
	{0xC25D96067E446730ULL, "ht_mansion_n_3382.rml"},
	{0xC25DEE43AEBBD43BULL, "seb321_leliana6197140.cl"},
	{0xC25E6739C0632BB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_door_male.gad"},
	{0xC25E7C53A11E4A00ULL, "vxdr_0_deeproads_9dc_0_0.vlm"},
	{0xC25EC372734E043CULL, "6004408_m.fxe"},
	{0xC25EE1199DEBF2F6ULL, "aip_bloodmage_base.gda"},
	{0xC25F2D87E147A308ULL, "6133889_m.fxe"},
	{0xC25F9794EAA59FB8ULL, "composedof.vsh.fxo"},
	{0xC260B72FEBE2A2B5ULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_follower_m_.gad"},
	{0xC260DCAC86B364DEULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_ex.evt"},
	{0xC260EEEE5CCDF458ULL, "dae000_keep_bureaucrat_2_st6165389.cl"},
	{0xC261DAABBE583E53ULL, "art\\cinematics\\act1\\cs_fp_smirk.gad"},
	{0xC262052A8312160BULL, "llac_lt_backalley_g_1_0.dds"},
	{0xC262849D334CEE90ULL, "vxw_bonepitdestroyed_a3_29k_0_0.vlm"},
	{0xC2633528F079EE14ULL, "vxlt_backalley_g_1w4_0_0.vlm"},
	{0xC2634AF7694B61ACULL, "27034026.ogg"},
	{0xC26394DA43DA1669ULL, "vxw_encounter1_ig_0_0.vlm"},
	{0xC26445125B0843B5ULL, "hightown_male_noble_gossip6174051.cl"},
	{0xC2647639C067B092ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\listen_door_male.ani"},
	{0xC264A76BDD8A94B6ULL, "hlac_lt_undercityserialkiller_-1_6z.dds"},
	{0xC2656A4C43AE7717ULL, "vxca_caverns1_mines_iok_0_0.vlm"},
	{0xC265A37C54B76F10ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_basic.phy"},
	{0xC265A7753D1A79A4ULL, "ave110tr_approach_donnic.ncs"},
	{0xC266AA1901A50839ULL, "6116598_f.fxe"},
	{0xC266DA26654DDA0AULL, "dre180cs_return_surface.cut"},
	{0xC266E8B165481CD2ULL, "380336481.ogg"},
	{0xC267524F800DF048ULL, "hlac_lt_0_lowtown_a2_65536_0x.dds"},
	{0xC267F3AC86B9AA09ULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_ex.gad"},
	{0xC26815E1CC49896DULL, "hlac_lt_0_lowtown_a2_-1_0x.dds"},
	{0xC2682709EFEC42B6ULL, "hlac_ca_caverns2prc_5802_0z.dds"},
	{0xC2695759E6B98E1AULL, "llac_ca_cavern1_haunted_24425_5.dds"},
	{0xC2695E84EB01D9EDULL, "6020560_m.fxe"},
	{0xC2698CCBF693411EULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody2_.ani"},
	{0xC269AB993AEF054AULL, "6162588_m.fxe"},
	{0xC26A07D4DC7A60ACULL, "dae000cr_grand_cleric6147685.cl"},
	{0xC26A3F27A575BC8CULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_vael.msh"},
	{0xC26A4527A575C77EULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_vael.mmh"},
	{0xC26A63DAE7737E46ULL, "gen_im_cns_flu_isabela_04.uti"},
	{0xC26A873B7697BC4AULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_rob_appd_0n.dds"},
	{0xC26A8DE24AB801D4ULL, "dae000_merchant_rare_em6177949.cl"},
	{0xC26B1B233224B7A9ULL, "vxht_keep_6i1_0_0.vlm"},
	{0xC26B52A406606539ULL, "ave_act3_insurgent.mor"},
	{0xC26BB3082431A200ULL, "vxlt_0_lowtown_n_fec_0_0.vlm"},
	{0xC26C704364540CB1ULL, "6068447_m.fxe"},
	{0xC26D531D5FD62F76ULL, "hla_ht_0_hightown_a3_131073_0x.dds"},
	{0xC26D8409DB00C2D3ULL, "6125468_f.fxe"},
	{0xC26D9BDFECE055ABULL, "6034357_f.fxe"},
	{0xC26DF108734E94F3ULL, "hlac_lt_0_lowtown_docks_n_-65537_3z.dds"},
	{0xC26E74E453AE041EULL, "qcr220cr_bethany.utc"},
	{0xC26E8E5AE8A09459ULL, "and100_anders6095908.cl"},
	{0xC26EB1F67A834F23ULL, "vxlt_gallowstemplar_n_5y8_0_0.vlm"},
	{0xC26EC4AC86BF7957ULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_ex.ani"},
	{0xC26F5F41BEAE7FCFULL, "214524533.ogg"},
	{0xC26F71B1D803AFA0ULL, "vxht_mansion2_2os_0_1.vlm"},
	{0xC26F9F951841047CULL, "vxlt_0_lowtown_docks_a3_n_cp7_0_0.vlm"},
	{0xC26FA89F8FCA3210ULL, "lgt000ip_store_epic2_3.utp"},
	{0xC2702810687FAA7AULL, "6129479_m.fxe"},
	{0xC270486B425FD0CFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_scrubbucket_n.dds"},
	{0xC2704F701BEF968AULL, "hlac_w_bonepitdestroyed_a3_327684_0z.dds"},
	{0xC270D2A2896BCD3BULL, "6101069_m.fxe"},
	{0xC270F67A25430710ULL, "vxw_bonepitdestroyed_a3_24h_0_0.vlm"},
	{0xC27191518BC54846ULL, "197753684.ogg"},
	{0xC271CB4B808B314DULL, "6092561_m.fxe"},
	{0xC2726A0F2C1EB31EULL, "vxht_0_hightown_a2_a2p_0_0.vlm"},
	{0xC2728DA37ED8DD10ULL, "190670238.ogg"},
	{0xC272B025B4CB4F90ULL, "hlac_lt_hangedman_n_-1_9y.dds"},
	{0xC272C5FAF7C077AEULL, "gen00fl_fenris6127965.cl"},
	{0xC27357D3FAD4A097ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\blendtree\\trooper_overlay.anb"},
	{0xC2737C69AD349448ULL, "358611109.ogg"},
	{0xC2737D93271453C7ULL, "llac_lt_0_lowtown_a2_n_131073_0.dds"},
	{0xC273A7869FADD98AULL, "6125592_m.fxe"},
	{0xC2744780AE246125ULL, "6126954_m.fxe"},
	{0xC2745251AFEE3B2FULL, "701543766.ogg"},
	{0xC27467159860CA18ULL, "898136791.ogg"},
	{0xC2751925ABF0F6EDULL, "6070543_m.fxe"},
	{0xC2755787B4896F04ULL, "hlac_lt_0_lowtown_a2_n_c_-65535_1x.dds"},
	{0xC2757ADAFF873A8EULL, "w_woundedcoast_217_0.phy"},
	{0xC276007F6E119A49ULL, "6099586_m.fxe"},
	{0xC276B1682B6575D5ULL, "63876771.ogg"},
	{0xC276F96234C4AEA7ULL, "vxht_0_hightown_a3_7fz_0_0.vlm"},
	{0xC277886AFE7091FEULL, "vxlt_blackemporium_v3_0_0.vlm"},
	{0xC27852C779C207A3ULL, "vxht_0_hightown_a3_cao_0_0.vlm"},
	{0xC27856F46D84CEAAULL, "qun160_party_comment_26061244.cl"},
	{0xC2797BD75FE0F04CULL, "llac_ht_store_187_1.dds"},
	{0xC279A8BEE79E3A8DULL, "hlac_ht_chantry_2534_30y.dds"},
	{0xC279BC2F6D2FF002ULL, "hlac_dr_slavers_3217_2y.dds"},
	{0xC27B45CFE873D62BULL, "6175105_m.fxe"},
	{0xC27BE5474D1B69EAULL, "vxlt_gallowstemplar_n_43r_0_0.vlm"},
	{0xC27C029460A3751FULL, "991723550.ogg"},
	{0xC27C44769328E22EULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_andraste.msh"},
	{0xC27C52769328FA7FULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_andraste.mao"},
	{0xC27C5E7693290E1CULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_andraste.mmh"},
	{0xC27C9BF60AAE40CAULL, "6013320_f.fxe"},
	{0xC27CC0E59697D55DULL, "cnvprev_human_female_mage.utc"},
	{0xC27CEAA9164FD51EULL, "projdecal_projdecal_p1.vsh.fxo"},
	{0xC27D0A406A3C8BD5ULL, "1035298960.ogg"},
	{0xC27D1D18D371798DULL, "6118605_m.fxe"},
	{0xC27D86BC078D33C0ULL, "dae000_bartrand6027455.cl"},
	{0xC27D9C933AB82FDBULL, "vxlt_0_lowtown_a2_n_epa_0_1.vlm"},
	{0xC27DA5CBF6A41498ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody2_.gad"},
	{0xC27DAED0F7DE55F5ULL, "dra_lit.mao"},
	{0xC27E9971242473D1ULL, "6153547_m.fxe"},
	{0xC27ECED9ECFC4C3EULL, "llac_ca_cavern1_haunted_24424_28.dds"},
	{0xC27EEB8362D65494ULL, "vxw_bonepitmining_a2_1jo_0_0.vlm"},
	{0xC27EF0F1D064F090ULL, "l_and311ar_andersshop.wbd"},
	{0xC27F20994A1CF2B7ULL, "gen00fl_merrill6129920.cl"},
	{0xC27F5DB0D5B40EC2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\cuddle_sitting_female_loop.gad"},
	{0xC27F7172738C6B64ULL, "w_sundermount_167_0.phy"},
	{0xC27F750B71F6776CULL, "art\\levels\\races\\design\\prp_gasbarrel.phy"},
	{0xC27FC86F5A0F2513ULL, "llac_ca_cavern1_haunted_24424_1.dds"},
	{0xC28001160F57C722ULL, "gen_im_trs_drk_two_2hm_r0.uti"},
	{0xC2802376AEF45F2EULL, "1020758045.ogg"},
	{0xC28033CBAF1A5A76ULL, "6028167_m.fxe"},
	{0xC28043A1F5D93DB4ULL, "hf_bethany_h2s3.mor"},
	{0xC2805DB7D9E81F4AULL, "vxlt_gallowstemplar_479_0_0.vlm"},
	{0xC280B5BAD4A34640ULL, "hm_circlemage01.mor"},
	{0xC280F7BDBB40C3BBULL, "vxlt_0_lowtown_docks_n_c6m_0_0.vlm"},
	{0xC2814DC4FF79A587ULL, "vxht_brothel_n_5hc_0_0.vlm"},
	{0xC28170E7F6E49DB4ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_whippingpost_dp.dds"},
	{0xC28296F52B8F8B62ULL, "lt_gallowstemplar_n_7198.anb"},
	{0xC2829BB7091BF761ULL, "hlac_lt_0_lowtown_a2_n_0_3x.dds"},
	{0xC282FC28F25F213DULL, "vxlt_0_lowtown_docks_aom_0_0.vlm"},
	{0xC283175B48DDB4B9ULL, "one120_athenril6113278.cl"},
	{0xC2833D040E0EAE9FULL, "6092128_f.fxe"},
	{0xC2836EAD928796B4ULL, "vxw_encounter2_dd_0_0.vlm"},
	{0xC2839B6E75CBB705ULL, "6171309_m.fxe"},
	{0xC283FE76F633C3BBULL, "mag110_vincento6036918.cl"},
	{0xC28437FF7BC738EAULL, "vxht_keep_6ez_0_0.vlm"},
	{0xC2847CB5101B7ACAULL, "m_one241ar_carta_cave.bnk"},
	{0xC284C80C78E2667BULL, "abi_m_stone_fist.wbd"},
	{0xC285D93858120D2FULL, "6167601_m.fxe"},
	{0xC285EC088C5E6FB9ULL, "hlac_ht_chantry_n_2534_8y.dds"},
	{0xC286417897DFA338ULL, "6031336_m.fxe"},
	{0xC28721BD5A2D5F16ULL, "lt_0_lowtown_docks_a3_40_0.rml"},
	{0xC28730621CCDA748ULL, "gen_im_arm_cht_med_act2_02.uti"},
	{0xC28777FA97963F99ULL, "vxlt_gallowstemplar_4uh_0_0.vlm"},
	{0xC28779A0FC40B3C1ULL, "1052702396.ogg"},
	{0xC287834152E14D14ULL, "dae000_arishok6068574.cl"},
	{0xC287DB725D9BBD3DULL, "809328274.ogg"},
	{0xC287E34189791A5EULL, "60015417.ogg"},
	{0xC287E755B14FB8F6ULL, "lgt360cr_mag_orlanna.utc"},
	{0xC2881FA160C0EB7EULL, "239174178.ogg"},
	{0xC288636C5A2A9634ULL, "465011355.ogg"},
	{0xC28863701DFE9F61ULL, "279126938.ogg"},
	{0xC2887391853D4658ULL, "6149369_m.fxe"},
	{0xC28896EBE8E5DB65ULL, "mer210_coterie_contact6035278.cl"},
	{0xC2889A8BFB2C485FULL, "6133733_f.fxe"},
	{0xC2893564E2B04419ULL, "668477696.ogg"},
	{0xC28997067E69A398ULL, "ht_mansion_n_3382.anb"},
	{0xC28A1DC30C799660ULL, "434954797.ogg"},
	{0xC28A6CF1299C80F5ULL, "140090478.ogg"},
	{0xC28ADCA12AC56135ULL, "6075614_m.fxe"},
	{0xC28B824963BF95CBULL, "397147972.ogg"},
	{0xC28B8EF5C567920DULL, "6117705_m.fxe"},
	{0xC28C67A58ADB7687ULL, "vxlt_0_lowtown_docks_b4u_0_0.vlm"},
	{0xC28C9AF04F8ABD43ULL, "6207740_f.fxe"},
	{0xC28D0B4E161A8C51ULL, "6156647_m.fxe"},
	{0xC28D37D6AB0B8EF0ULL, "vrc221_varric6043599.cl"},
	{0xC28D3B55AA0F1099ULL, "vxht_0_hightown_cad_0_0.vlm"},
	{0xC28D8BCD3D5895A8ULL, "vxlt_0_lowtown_a2_bz9_0_0.vlm"},
	{0xC28E2EA3EE98F3ECULL, "747505549.ogg"},
	{0xC28E6EA540F4CE79ULL, "6101823_m.fxe"},
	{0xC28E6ED5A0C42CFCULL, "dae000cr_grand_cleric6183973.cl"},
	{0xC28EBCAEDE3F7490ULL, "follower_banter6149193.cl"},
	{0xC28F090697BCF7F3ULL, "6167083_m.fxe"},
	{0xC28F2D0F676AE60FULL, "vxht_0_hightown_a2_c6s_0_0.vlm"},
	{0xC290061632192042ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls109.mao"},
	{0xC2904C4F85A7679CULL, "gen_im_trs_rdr_one_mac_r1.uti"},
	{0xC290758733067334ULL, "83986321.ogg"},
	{0xC290BDB50520CEAAULL, "hlac_lt_0_lowtown_n_-131073_0x.dds"},
	{0xC2911D393ED0D2BBULL, "6111640_m.fxe"},
	{0xC291379D80BCF6B8ULL, "vxlt_blackemporium_tk_0_0.vlm"},
	{0xC2914822D140FF53ULL, "6163906_m.fxe"},
	{0xC2919E4C19E55F30ULL, "6149787_m.fxe"},
	{0xC291AEE10004C71FULL, "6097228_f.fxe"},
	{0xC291FBEF0BD4CBA6ULL, "vxlt_0_lowtown_a2_n_byq_0_0.vlm"},
	{0xC291FE393FBC088DULL, "6076909_m.fxe"},
	{0xC292D2D1F6248F20ULL, "6126704_m.fxe"},
	{0xC29388B0D5C500A8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\ambient\\compressed\\cuddle_sitting_female_loop.ani"},
	{0xC293AE3487A80588ULL, "127750273.ogg"},
	{0xC293B6869A1A5B6FULL, "aip_dalishwarrior_base.gda"},
	{0xC294251E66D670B3ULL, "llac_w_sundermount_196606_0.dds"},
	{0xC2963B56BE2DFF2FULL, "vxlt_backalley_g_1nu_0_1.vlm"},
	{0xC29660F71DF8E29AULL, "883875210.ogg"},
	{0xC297D5B32200BDD0ULL, "custic001.dds"},
	{0xC297F4834366D610ULL, "vxlt_0_lowtown_csg_0_0.vlm"},
	{0xC2980402C1749538ULL, "vxlt_0_lowtown_docks_n_b9o_0_1.vlm"},
	{0xC29816DA60389C31ULL, "vxw_bonepitdestroyed_a3_1os_0_0.vlm"},
	{0xC2981FF33F0D9BB8ULL, "hlac_lt_0_lowtown_a2_-65535_5y.dds"},
	{0xC2984F53D418F579ULL, "mag311_nyssa6051147.cl"},
	{0xC298CC95DE7C7BD9ULL, "ts_generic_wpn_table.gda"},
	{0xC2990DF5087E780EULL, "llac_w_encounter3_-327682_0.dds"},
	{0xC2993718B851DA0BULL, "vxca_cavern1_haunted_gk8_0_1.vlm"},
	{0xC2997C0E7AF08220ULL, "6118169_m.fxe"},
	{0xC299B8804FBE3A0AULL, "vxca_caverns1_mines_e8s_0_0.vlm"},
	{0xC299B89703CA0FB2ULL, "dae000ip_crowd_ff5_talk4.utp"},
	{0xC299CD0D972A329AULL, "6155919_m.fxe"},
	{0xC29A8D15140E6632ULL, "6004385_m.fxe"},
	{0xC29ABCC4C2D2A8F5ULL, "nat380_nathaniel6071282.cl"},
	{0xC29B19760471D989ULL, "lt_undercitydungeon_3567_lc.mmh"},
	{0xC29B6911D276562EULL, "856933309.ogg"},
	{0xC29BC8241C925B28ULL, "vxlt_warehouse_n_11r_0_0.vlm"},
	{0xC29CC2F34018DA20ULL, "250193855.ogg"},
	{0xC29CED2C4D9834AAULL, "gen00fl_varric6063845.cl"},
	{0xC29CF9EA18FC8F83ULL, "hlac_lt_0_lowtown_a2_n_-1_0y.dds"},
	{0xC29D31AAEC3F78C7ULL, "18439110.ogg"},
	{0xC29D7B82CA8CBDE3ULL, "hlac_ca_cavern1_haunted_24421_3y.dds"},
	{0xC29DB138EBF1EB82ULL, "6046952_m.fxe"},
	{0xC29DB26133E48B77ULL, "6049707_m.fxe"},
	{0xC29E917027740BFAULL, "zz_followers_banter6159552.cl"},
	{0xC29EA58D53205D1CULL, "1017291584.ogg"},
	{0xC29EA934E51A93EBULL, "hlac_lt_gallowscourtyard_-1_2y.dds"},
	{0xC29F463FB8BB073CULL, "dae000_dalish_apprentices6147515.cl"},
	{0xC29F4BF4F9011860ULL, "vxlt_darktown_4vp_0_0.vlm"},
	{0xC2A0187873846B74ULL, "vxht_chantry_n_34u_0_0.vlm"},
	{0xC2A0516723C8F197ULL, "gen00fl_dog6167644.cl"},
	{0xC2A052677673C7B4ULL, "vxw_encounter2_jj_0_0.vlm"},
	{0xC2A08017DC76F310ULL, "hlac_ht_mansion2_818_3y.dds"},
	{0xC2A0933044893FBCULL, "mag101_aombanter6052752.cl"},
	{0xC2A15898300C811BULL, "249272076.ogg"},
	{0xC2A1C5B0583D243AULL, "hlac_blank_exterior_0_0z.dds"},
	{0xC2A1F1AC5149BE28ULL, "hlac_w_encounter3_-196611_0z.dds"},
	{0xC2A1F1F6E2E62D84ULL, "104194692.ogg"},
	{0xC2A21138A3419CA4ULL, "vxw_sundermount_30c_0_1.vlm"},
	{0xC2A23D769349A5E7ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_andraste.phy"},
	{0xC2A25CFC64E79C2BULL, "672599274.ogg"},
	{0xC2A2995DAA6F73D8ULL, "zz_fol_debug6098069.cl"},
	{0xC2A2CBB6E890FCA9ULL, "6062122_m.fxe"},
	{0xC2A3085C14F60D93ULL, "194576414.ogg"},
	{0xC2A4344F4D36843BULL, "dre000_dougal6142204.cl"},
	{0xC2A4E05E7DC77E59ULL, "vxht_bartrand_n_3za_0_0.vlm"},
	{0xC2A53AE4821478CEULL, "vfx_rendertoshadowmap_p2.vsh.fxo.dx11.assembly"},
	{0xC2A5BC3823CA8E51ULL, "6087017_m.fxe"},
	{0xC2A5C9A53AAB983FULL, "llac_lt_darktown_-1_52.dds"},
	{0xC2A645CCEF98BC1DULL, "651390162.ogg"},
	{0xC2A675DFB36AADB8ULL, "vxlt_0_lowtown_docks_n_b6n_0_1.vlm"},
	{0xC2A6CADDBA257E01ULL, "llac_w_encounter1_-131071_0.dds"},
	{0xC2A73CA0B0FC25F5ULL, "vxlt_0_lowtown_a3_n_c_emr_0_0.vlm"},
	{0xC2A78A21A03679D9ULL, "vxlt_gallowstemplar_n_5iz_0_0.vlm"},
	{0xC2A79BA50698CBFEULL, "6120421_m.fxe"},
	{0xC2A79F382DC2E6EEULL, "llac_lt_gallowstemplar_7198_7.dds"},
	{0xC2A8188EB92A7F37ULL, "vxca_caverns2_7z4_0_0.vlm"},
	{0xC2A81D6D728C18B8ULL, "dae000_arishok6068426.cl"},
	{0xC2A8417E1D4F027FULL, "one100_party_comments6112673.cl"},
	{0xC2A88AD4019AA97AULL, "959382550.ogg"},
	{0xC2A8ECCC3377AD4DULL, "149234879.ogg"},
	{0xC2A90A3A655C40CEULL, "6177898_m.fxe"},
	{0xC2A96DC84A45E0FDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_sledge.mao"},
	{0xC2A971C84A45E736ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_sledge.mmh"},
	{0xC2A97BC84A45F834ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_sledge.msh"},
	{0xC2A9E00ACB00D5CBULL, "6117448_m.fxe"},
	{0xC2A9E511BE89F8DDULL, "vxlt_0_lowtown_h9c_0_0.vlm"},
	{0xC2AA1CBF13BE3E7CULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.dura.ani"},
	{0xC2AA5F4AB4142BB6ULL, "prp_str_bark_d.dds"},
	{0xC2AAC56D2E55E936ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtd.msh"},
	{0xC2AACF6D2E55FA34ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtd.mmh"},
	{0xC2AB1C27A5AD1819ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_vael.phy"},
	{0xC2AB288341061DCCULL, "6083478_m.fxe"},
	{0xC2AB90154A17F3ABULL, "vxdr_0_deeproads_eq3_0_0.vlm"},
	{0xC2ABE1595D9A607CULL, "6084509_m.fxe"},
	{0xC2ABEA5543433DDCULL, "gen00fl_isabela6075291.cl"},
	{0xC2AC28C388BA4AA3ULL, "809151937.ogg"},
	{0xC2ACE572736A94F1ULL, "6133066_m.fxe"},
	{0xC2AD052295D43348ULL, "6170655_m.fxe"},
	{0xC2AD2359D937BD2EULL, "vxlt_darktown_44d_0_0.vlm"},
	{0xC2AD8AD34AAA6877ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2c.msh"},
	{0xC2AD9CD34AAA86DAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2c.mao"},
	{0xC2ADA0D34AAA8DE9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2c.mmh"},
	{0xC2ADF2380DA72837ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2whitelyrium.msh"},
	{0xC2ADFD78A081348BULL, "combat_lc_r1_a4.gda"},
	{0xC2AE04380DA7469AULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2whitelyrium.mao"},
	{0xC2AE08380DA74DA9ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_2whitelyrium.mmh"},
	{0xC2AE0A72F7CA1C1BULL, "245876838.ogg"},
	{0xC2AE928E829EFFE9ULL, "vxw_bonepitdestroyed_a3_26w_0_0.vlm"},
	{0xC2AF9E74DAD13EA9ULL, "hlac_lt_0_lowtown_a2_n_c_65535_6y.dds"},
	{0xC2AFB68BDA10C020ULL, "6128504_m.fxe"},
	{0xC2B03CF59EA4F361ULL, "hlac_lt_foundry_n_320_0x.dds"},
	{0xC2B03EC58EC4FFC7ULL, "370583928.ogg"},
	{0xC2B084C46F146F59ULL, "mcr310_argument6127584.cl"},
	{0xC2B0A773188C90AAULL, "vxlt_0_lowtown_a2_n_cmt_0_0.vlm"},
	{0xC2B129BF13C472BEULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.dura.gad"},
	{0xC2B13D93DA871529ULL, "6069022_m.fxe"},
	{0xC2B16A337BFB8976ULL, "vxht_mansion_n_4fb_0_0.vlm"},
	{0xC2B1BBC3B8D07152ULL, "hlac_lt_0_lowtown_n_131069_0x.dds"},
	{0xC2B1DF5439E86395ULL, "vxca_caverns1_ck6_0_0.vlm"},
	{0xC2B279B264ECAB4DULL, "955354293.ogg"},
	{0xC2B29377390F0CB9ULL, "vxlt_gallowstemplar_n_3gr_0_0.vlm"},
	{0xC2B29C2163FD1FA1ULL, "6080837_f.fxe"},
	{0xC2B2A5D71DB201D3ULL, "6039946_m.fxe"},
	{0xC2B305112EE2C6FEULL, "dae000_hate_follower_react6103235.cl"},
	{0xC2B3688BFBF3B6A1ULL, "586627638.ogg"},
	{0xC2B3E3238677130CULL, "216871115.ogg"},
	{0xC2B3E7FDE29A1EE8ULL, "hlac_lt_0_lowtown_a2_-131072_1y.dds"},
	{0xC2B41D131EE74393ULL, "vxht_brothel_n_4hv_0_0.vlm"},
	{0xC2B47B67D57CA36CULL, "490035766.ogg"},
	{0xC2B4845E2491CB4FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masd_0d.dds"},
	{0xC2B501EEBB492174ULL, "mag170cr_hugh.utc"},
	{0xC2B54CD552543CDFULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medg.msh"},
	{0xC2B552D552544681ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medg.mmh"},
	{0xC2B5AF254F2CC8F9ULL, "vxlt_0_lowtown_deg_0_0.vlm"},
	{0xC2B5B07D3D340D38ULL, "character_character_p94.psh.fxo"},
	{0xC2B629BF49CF9751ULL, "rdr111dg_martin6057072.cl"},
	{0xC2B679DAFFBED051ULL, "w_woundedcoast_217_0.mmh"},
	{0xC2B685DAFFBEE4F2ULL, "w_woundedcoast_217_0.mao"},
	{0xC2B693DAFFBEFC0FULL, "w_woundedcoast_217_0.msh"},
	{0xC2B78A063C392CA1ULL, "fake_im_arm_hlm_med_champ.uti"},
	{0xC2B7940C44AAFE83ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupc.mao"},
	{0xC2B7980C44AB05D0ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupc.mmh"},
	{0xC2B7A60C44AB1DEAULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupc.msh"},
	{0xC2B804BF13CA52DDULL, "art\\vfx\\combat\\vfx_q_ari_bstorm_act_c.dura.evt"},
	{0xC2B86B72B8B43005ULL, "standard_targeting_pressed.cur"},
	{0xC2B917AB0E37FF77ULL, "6054868_m.fxe"},
	{0xC2B91ABDD9F4BDD6ULL, "lt_0_lowtown_docks_n_255_0.mao"},
	{0xC2B926BDD9F4D275ULL, "lt_0_lowtown_docks_n_255_0.mmh"},
	{0xC2B928BDD9F4D5EBULL, "lt_0_lowtown_docks_n_255_0.msh"},
	{0xC2B997882E06D4FBULL, "3786168.ogg"},
	{0xC2B9AAC476C10496ULL, "6096778_f.fxe"},
	{0xC2BA3058A77036B1ULL, "35742480.ogg"},
	{0xC2BAC3F2730FC13EULL, "qcr220_orsino6041815.cl"},
	{0xC2BADE2FE83C052EULL, "gen_im_trs_gen_arc_lbw_r1.uti"},
	{0xC2BB122BF8FB1595ULL, "vxca_cavern1_haunted_huv_0_0.vlm"},
	{0xC2BB4C1258075B3EULL, "zz_seb_debug6183197.cl"},
	{0xC2BB8DDE423C9149ULL, "314352960.ogg"},
	{0xC2BC9EB5D6CCAA25ULL, "hlac_w_blightlands_0_4z.dds"},
	{0xC2BCFC02958B6845ULL, "96252076.ogg"},
	{0xC2BD36215A5FF8B4ULL, "vxw_encounter2_ph_0_0.vlm"},
	{0xC2BD8EB3A80735A4ULL, "719514317.ogg"},
	{0xC2BDC62570DB0279ULL, "645449980.ogg"},
	{0xC2BDDF1A6DC71FB1ULL, "6147224_f.fxe"},
	{0xC2BE7ECA0370A3DBULL, "6042917_m.fxe"},
	{0xC2BEB20E6B233B14ULL, "hlac_lt_0_lowtown_docks_n_-65538_1z.dds"},
	{0xC2BF1B5A14B16974ULL, "6054470_m.fxe"},
	{0xC2BF2244B70DB9BDULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hlf_meda_0n.dds"},
	{0xC2BFD251979BD9D7ULL, "vxlt_0_lowtown_h5e_0_0.vlm"},
	{0xC2C028A51E9E5643ULL, "6067992_m.fxe"},
	{0xC2C04C0B722DC837ULL, "art\\levels\\races\\design\\prp_gasbarrel.mmh"},
	{0xC2C0500B722DCF44ULL, "art\\levels\\races\\design\\prp_gasbarrel.mao"},
	{0xC2C051312BDE7C15ULL, "vxca_caverns1_cn5_0_0.vlm"},
	{0xC2C06157D9913ECFULL, "611842456.ogg"},
	{0xC2C0620B722DEDA9ULL, "art\\levels\\races\\design\\prp_gasbarrel.msh"},
	{0xC2C0C8AAB4AF3989ULL, "72971216.ogg"},
	{0xC2C0CADD42343037ULL, "plt_lgt120pt_night_lies.nss"},
	{0xC2C14778E22A04E9ULL, "704702162.ogg"},
	{0xC2C1EAF9D44C98BEULL, "lla_ht_0_hightown_a3_65534_0.dds"},
	{0xC2C217C829386445ULL, "default_unspecified.dds"},
	{0xC2C239017FC45997ULL, "441274223.ogg"},
	{0xC2C24BDCD7CFF74CULL, "w_sundermount_12_0.msh"},
	{0xC2C251DCD7D0013EULL, "w_sundermount_12_0.mmh"},
	{0xC2C25DDCD7D015D5ULL, "w_sundermount_12_0.mao"},
	{0xC2C32D9428AFC30AULL, "241609332.ogg"},
	{0xC2C4225A2F2A25B8ULL, "character_character_p43.psh.fxo"},
	{0xC2C42901AD5F636DULL, "58855796.ogg"},
	{0xC2C4A5AA0EFB772BULL, "vxlt_warehouse_14q_0_0.vlm"},
	{0xC2C4BB58CBE671D7ULL, "vxht_chantry_n_2zm_0_0.vlm"},
	{0xC2C4C77C1C4DBEB4ULL, "seb201_lady_harimann6177581.cl"},
	{0xC2C4FA342178B57BULL, "hlac_lt_0_lowtown_a2_131070_5y.dds"},
	{0xC2C52D9BA6B3DC25ULL, "6162464_m.fxe"},
	{0xC2C52FF08EB1569DULL, "dae100st_mother.stg"},
	{0xC2C54D1C1265605AULL, "vxht_brothel_n_61v_0_0.vlm"},
	{0xC2C5AF6A23D837B6ULL, "art\\levels\\ztools\\default\\tilesets\\setext\\_textures\\xx_extwood.dds"},
	{0xC2C5D3BE7E9EB1A1ULL, "234183619.ogg"},
	{0xC2C6C6F1D0A25F52ULL, "l_and311ar_andersshop.bnk"},
	{0xC2C73A815BAED72AULL, "997975758.ogg"},
	{0xC2C73AEA56288686ULL, "gen00fl_sebastian6180680.cl"},
	{0xC2C75D6878826667ULL, "art\\vfx\\environment\\vfx_fire_small_p.dura.gad"},
	{0xC2C77B2EEBE5BC22ULL, "mag291_demon_marethari6028917.cl"},
	{0xC2C81227E71DC85AULL, "vxw_encounter2_gc_0_0.vlm"},
	{0xC2C845439433E659ULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_en.ani"},
	{0xC2C91D432B7FFD6FULL, "969150609.ogg"},
	{0xC2C9279711DAFFE8ULL, "356528115.ogg"},
	{0xC2C9529C1754907AULL, "llac_lt_gallowstemplar_f_2394_10.dds"},
	{0xC2C95F4FD7A538EDULL, "fen210_orana6109733.cl"},
	{0xC2C9B5A9C06D925DULL, "vxlt_0_lowtown_el9_0_0.vlm"},
	{0xC2C9C92C5756C5F1ULL, "6139666_m.fxe"},
	{0xC2C9D557E03898BCULL, "vxlt_gallowstemplar_4u1_0_0.vlm"},
	{0xC2C9EB1DF6466D16ULL, "6179045_m.fxe"},
	{0xC2CA87C920B81AACULL, "hlac_lt_gallowscourtyard_n_-65536_1z.dds"},
	{0xC2CB07BD2BBA8CCEULL, "vxht_0_hightown_a2_c80_0_0.vlm"},
	{0xC2CB2C74F7B32A0DULL, "hlac_lt_0_lowtown_docks_65533_2y.dds"},
	{0xC2CB4DFA252072F0ULL, "hlac_lt_0_lowtown_a3_n_c_0_5z.dds"},
	{0xC2CB95A2C363A724ULL, "110182076.ogg"},
	{0xC2CBC81120573EB6ULL, "114970479.ogg"},
	{0xC2CCC014D85E424AULL, "312892388.ogg"},
	{0xC2CD00A42C13FEF4ULL, "lowtown_male_beggar_sta6151081.cl"},
	{0xC2CD228875860ECCULL, "hlac_lt_0_lowtown_a2_n_c_-1_0x.dds"},
	{0xC2CD471453AA64CCULL, "550201667.ogg"},
	{0xC2CD505890E1E944ULL, "6099529_m.fxe"},
	{0xC2CD52687886BF74ULL, "art\\vfx\\environment\\vfx_fire_small_p.dura.evt"},
	{0xC2CD96D7D877B5A8ULL, "vxht_0_hightown_a2_9uw_0_0.vlm"},
	{0xC2CDDA2DA8C5A8C9ULL, "llac_ht_keep_3731_0.dds"},
	{0xC2CE4CE6A8D2451AULL, "107930365.ogg"},
	{0xC2CE7FC5D7AC9086ULL, "lt_0_lowtown_docks_n_32_0_lc.mmh"},
	{0xC2CEB155EA9553DEULL, "hlac_ht_bartrand_n_3364_1z.dds"},
	{0xC2CF20439439C59BULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_en.gad"},
	{0xC2CF3A9F978EEEFCULL, "6110572_m.fxe"},
	{0xC2CF491EFA8EBE8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryrb.msh"},
	{0xC2CF4F1EFA8EC97EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryrb.mmh"},
	{0xC2CF5B1EFA8EDD15ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryrb.mao"},
	{0xC2CFCB4B423E9703ULL, "vxlt_0_lowtown_n_h7h_0_0.vlm"},
	{0xC2CFDDBA4E618756ULL, "vxht_mansion2b_3cy_0_0.vlm"},
	{0xC2D0001973071C0FULL, "598858996.ogg"},
	{0xC2D004C37BA03614ULL, "42164144.ogg"},
	{0xC2D01FB332AA0A9AULL, "one000_ilen6060500.cl"},
	{0xC2D03384E0423C5CULL, "162436174.ogg"},
	{0xC2D04147073DA8D6ULL, "qcr000_arishok6166991.cl"},
	{0xC2D0A5AF8F110FDFULL, "6208179_m.fxe"},
	{0xC2D0C0CA5B072405ULL, "170277052.ogg"},
	{0xC2D179AAD2DDD84BULL, "6087279_m.fxe"},
	{0xC2D189191CAE7660ULL, "mcr370_meredith6085974.cl"},
	{0xC2D1CCF1BA47C65EULL, "lt_0_lowtown_28_0.rml"},
	{0xC2D1EDD93A8CFE84ULL, "vxht_keep_n_600_0_0.vlm"},
	{0xC2D22EC132164627ULL, "498542548.ogg"},
	{0xC2D27F722A6FEE03ULL, "w_bonepitmining_a2_26_0_ll.mmh"},
	{0xC2D2943721203E38ULL, "character_character_p32.psh.fxo"},
	{0xC2D296D5629FF66EULL, "hlac_lt_0_lowtown_docks_0_0z.dds"},
	{0xC2D2C1356F0F2EE8ULL, "6185484_m.fxe"},
	{0xC2D2DA801D7337EBULL, "6099535_m.fxe"},
	{0xC2D30D580FF0046FULL, "hlac_lt_gallowscourtyard_131071_2z.dds"},
	{0xC2D37D0198093EE5ULL, "250401573.ogg"},
	{0xC2D42C37C9FD9A5EULL, "6126834_m.fxe"},
	{0xC2D44721C448B288ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp05.msh"},
	{0xC2D45CB504A9F878ULL, "vxlt_warehouse_n_r8_0_0.vlm"},
	{0xC2D46521C448E5B2ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp05.mmh"},
	{0xC2D4BB883DD9C243ULL, "qun131_warrens_thug6019130.cl"},
	{0xC2D54CCFDBB10395ULL, "6132766_m.fxe"},
	{0xC2D63D4394401580ULL, "art\\characters\\creatures\\spider\\animation\\tal_ovlm_en.evt"},
	{0xC2D6CF28D05932A3ULL, "vxw_bonepitmining_a2_1wn_0_1.vlm"},
	{0xC2D76EC4BCBCD41BULL, "6107294_m.fxe"},
	{0xC2D7825C0743BBC7ULL, "vxht_bartrand_n_5qd_0_0.vlm"},
	{0xC2D78849F38D0F54ULL, "hlac_dr_primevaltomb_659_10y.dds"},
	{0xC2D7F94866A5BA45ULL, "gen00fl_sebastian6188667.cl"},
	{0xC2D845DDFD6CB0D9ULL, "vxlt_0_lowtown_docks_n_bf2_0_0.vlm"},
	{0xC2D8859AF50C5F7CULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_qiv_cb.mao"},
	{0xC2D8D023CBA905A8ULL, "hlac_ht_playershouse_n_855_8y.dds"},
	{0xC2D96865EA8052A0ULL, "gen00fl_fenris6103352.cl"},
	{0xC2D98E87EE4572C0ULL, "vxca_caverns2_6ha_0_1.vlm"},
	{0xC2D9BC17609F0E35ULL, "hlac_lt_0_lowtown_docks_a3_-3_8x.dds"},
	{0xC2DA1CAAA64F1231ULL, "187052879.ogg"},
	{0xC2DA54509526B4AAULL, "57027487.ogg"},
	{0xC2DA548183051ED2ULL, "vxlt_0_lowtown_n_dbb_0_0.vlm"},
	{0xC2DA7690410178E8ULL, "gen00fl_aveline6139086.cl"},
	{0xC2DB27E39A400F5FULL, "al_anders_s01.gda"},
	{0xC2DB2C7A171E51B7ULL, "art\\characters\\creatures\\stonegolem\\animation\\turnl90.gad"},
	{0xC2DB314752941101ULL, "mcr370_assignments6081728.cl"},
	{0xC2DB4A687892EEADULL, "art\\vfx\\environment\\vfx_fire_small_p.dura.ani"},
	{0xC2DBB4D5F9FE77DAULL, "6077826_m.fxe"},
	{0xC2DC39335CB6909BULL, "vxlt_0_lowtown_a3_n_c_ewz_0_0.vlm"},
	{0xC2DD9C126CB06C57ULL, "gen00fl_anders6111411.cl"},
	{0xC2DE29FA7CB0D67DULL, "vxdr_0_deeproads_f52_0_0.vlm"},
	{0xC2DE4F9740579FD6ULL, "vxht_mansion2b_3fz_0_0.vlm"},
	{0xC2DE555EBC9AC625ULL, "6031315_m.fxe"},
	{0xC2DEC8C9A474967EULL, "6042546_f.fxe"},
	{0xC2DF759372FB4013ULL, "6096497_m.fxe"},
	{0xC2DF80FAE8CB9AD0ULL, "qun111_ketojan_freed.stg"},
	{0xC2DF91F093C858D1ULL, "6043660_m.fxe"},
	{0xC2DFF82D481615EAULL, "98442386.ogg"},
	{0xC2E01E1981D74B1EULL, "mrl310_merrill6144602.cl"},
	{0xC2E036E3078E49EFULL, "vxlt_0_lowtown_docks_a3_d38_0_1.vlm"},
	{0xC2E0F37832CC6931ULL, "w_bonepitmining_a2_197_0.msh"},
	{0xC2E1017832CC814CULL, "w_bonepitmining_a2_197_0.mao"},
	{0xC2E10D7832CC95EFULL, "w_bonepitmining_a2_197_0.mmh"},
	{0xC2E1417A1722E144ULL, "art\\characters\\creatures\\stonegolem\\animation\\turnl90.evt"},
	{0xC2E153D5527983FEULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medg.phy"},
	{0xC2E1B144546C3A9AULL, "6075701_m.fxe"},
	{0xC2E20B2A6ABAB31EULL, "6038603_m.fxe"},
	{0xC2E27FF7FCCD0346ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.dura.ani"},
	{0xC2E2A45B962010C7ULL, "docks_male_human_worker_sta6175981.cl"},
	{0xC2E2CE91F6A010F8ULL, "vxlt_warehouse_n_y9_0_0.vlm"},
	{0xC2E3594B6EC39AA5ULL, "769149311.ogg"},
	{0xC2E35DA1F9179DE5ULL, "vxca_caverns1_ckf_0_0.vlm"},
	{0xC2E3EBEE0A06DAABULL, "6172272_m.fxe"},
	{0xC2E44098EA1A589CULL, "hlac_lt_0_lowtown_a2_n_1_10z.dds"},
	{0xC2E4815A2A22DA00ULL, "hlac_dr_primevaltomb_661_4y.dds"},
	{0xC2E49B8B9C34DFB6ULL, "root_rog_archery.dds"},
	{0xC2E4CF8D412ED8BFULL, "vxlt_gallowstemplar_f_5jq_0_0.vlm"},
	{0xC2E5B18B33128B63ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.cess.evt"},
	{0xC2E5F438F939D447ULL, "vxht_bartrand_n_5tg_0_0.vlm"},
	{0xC2E60B73C78F59A7ULL, "vxht_mansion2_3e6_0_0.vlm"},
	{0xC2E67D4E9FE51F3BULL, "6041390_m.fxe"},
	{0xC2E6B1F20DB4C6F6ULL, "142307643.ogg"},
	{0xC2E7064E14DE816AULL, "vxlt_0_lowtown_e19_0_0.vlm"},
	{0xC2E7571E6553387AULL, "138099664.ogg"},
	{0xC2E77C4657D367F8ULL, "vxlt_darktown_4u1_0_0.vlm"},
	{0xC2E7A6E415DAB259ULL, "seb221_flora_harimann6177283.cl"},
	{0xC2E7A787315E0F01ULL, "vxdr_0_deeproads_9dn_0_0.vlm"},
	{0xC2E7DBB380004AB4ULL, "6143485_f.fxe"},
	{0xC2E7E3109D2EFD72ULL, "6095048_m.fxe"},
	{0xC2E81C32E1E25478ULL, "6115377_m.fxe"},
	{0xC2E81E2DF6032A22ULL, "dae000tr_mail_notification.utt"},
	{0xC2E905C8EE7607DEULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.impa.ani"},
	{0xC2E93B4FA028AFE2ULL, "hlac_w_bonepit_0_2x.dds"},
	{0xC2E959D31E3C8B78ULL, "mag170_thrask6016013.cl"},
	{0xC2E9FBA2D5A63BFBULL, "371714426.ogg"},
	{0xC2EA0AA65D751AF9ULL, "6057144_m.fxe"},
	{0xC2EACD2B3E066E9FULL, "hlac_dr_slavers_3207_9x.dds"},
	{0xC2EB3518373585BDULL, "262734225.ogg"},
	{0xC2EB48EA9C8F733EULL, "6068264_m.fxe"},
	{0xC2EC0D8375980353ULL, "hlac_lt_0_lowtown_a2_n_c_131072_1z.dds"},
	{0xC2EC1EEAC664846FULL, "513395265.ogg"},
	{0xC2EC40BB22FBE796ULL, "199728477.ogg"},
	{0xC2EC4429169428C0ULL, "396257660.ogg"},
	{0xC2EC5AEB00E76EB9ULL, "vxlt_gallowstemplar_n_3sv_0_0.vlm"},
	{0xC2EC687D6C07A4DBULL, "hlac_lt_0_lowtown_docks_a3_-2_4y.dds"},
	{0xC2EC8A8B331867E4ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.cess.gad"},
	{0xC2ECA996E5841D4FULL, "vxdr_primevalentrance_1aa_0_0.vlm"},
	{0xC2ECB6A36C2C2A95ULL, "vxw_encounter2_c4_0_0.vlm"},
	{0xC2ED8C5D2962F776ULL, "darktown_male_lowlife_mob6174411.cl"},
	{0xC2EDFFE13737C491ULL, "310101500.ogg"},
	{0xC2EE5857C1D2AB99ULL, "vxlt_0_lowtown_n_ev4_0_0.vlm"},
	{0xC2EE7F5679A511AFULL, "6162776_m.fxe"},
	{0xC2EEE796B972A7B8ULL, "6034201_m.fxe"},
	{0xC2EEFE766D18D23DULL, "vxlt_0_lowtown_n_e24_0_0.vlm"},
	{0xC2EF47A063F7E3E3ULL, "vxlt_0_lowtown_n_erz_0_0.vlm"},
	{0xC2EF597A172F473DULL, "art\\characters\\creatures\\stonegolem\\animation\\turnl90.ani"},
	{0xC2EF6B1E4663BE3CULL, "594204480.ogg"},
	{0xC2EF93E5149B96EFULL, "mag160_varian_ilithis6018902.cl"},
	{0xC2EFA28D019D3C8CULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_dw_hammer_0_sa.dds"},
	{0xC2EFBE3A015F68EBULL, "6093735_m.fxe"},
	{0xC2EFE980BC360E0EULL, "follower_banter6163780.cl"},
	{0xC2F0336427564CE1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgta_0t.dds"},
	{0xC2F04576903A5402ULL, "w_woundedcoast_59_0.rml"},
	{0xC2F0521F1E974044ULL, "507859822.ogg"},
	{0xC2F067F7FCD917B7ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.dura.evt"},
	{0xC2F06CF3FA214550ULL, "916967260.ogg"},
	{0xC2F0B3A40D87C504ULL, "vxht_mansion_n_6ke_0_0.vlm"},
	{0xC2F0E8BF901A7ED7ULL, "vxca_cavern1_haunted_efx_0_0.vlm"},
	{0xC2F1D0F63737F48AULL, "vxlt_hangedman_20w_0_0.vlm"},
	{0xC2F246B0A3AE09BDULL, "raven_b_up.mao"},
	{0xC2F31AC1F0613BF9ULL, "vxw_bonepit_1pj_0_0.vlm"},
	{0xC2F3345E577D8F80ULL, "49372476.ogg"},
	{0xC2F3416A3324F13FULL, "vxlt_gallowstemplar_f_5mp_0_0.vlm"},
	{0xC2F374B605764F40ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_lowtown_gate_sa.dds"},
	{0xC2F3998B331E9FC2ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.cess.ani"},
	{0xC2F3A364B3766715ULL, "zz_ser_debug6085768.cl"},
	{0xC2F3C532FBCEFAC2ULL, "qun110_petrice6018640.cl"},
	{0xC2F4DE3BB4716701ULL, "vxlt_0_lowtown_n_fen_0_0.vlm"},
	{0xC2F4E3161B679740ULL, "584911992.ogg"},
	{0xC2F4F0D45C848E50ULL, "gen00fl_sebastian6188563.cl"},
	{0xC2F551332B396B1EULL, "6123662_m.fxe"},
	{0xC2F563C4AA987B5EULL, "6129619_m.fxe"},
	{0xC2F5AFF1F61AA3CAULL, "vxlt_0_lowtown_n_e1n_0_0.vlm"},
	{0xC2F60637456D1B8DULL, "ave121cr_jeven.utc"},
	{0xC2F61DC8EE80BAAFULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.impa.evt"},
	{0xC2F626048877E5DDULL, "mag110_arianni6016804.cl"},
	{0xC2F69021E124ECFFULL, "art\\characters\\playercharacter\\humanmale\\textures\\ulrichvon\\hm_arm_urva_0n.dds"},
	{0xC2F695C4A72EB64EULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialc.msh"},
	{0xC2F6A3C4A72ECD9FULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialc.mao"},
	{0xC2F6AFC4A72EE23CULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_vialc.mmh"},
	{0xC2F6E365D696F03EULL, "hlac_lt_gallowscourtyard_n_-2_3y.dds"},
	{0xC2F6EF72CE1B1BA4ULL, "hlac_ht_chantry_n_2534_31z.dds"},
	{0xC2F71BAC482ED9B6ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf4a.phy"},
	{0xC2F720052F25230FULL, "isa211_isabela6202068.cl"},
	{0xC2F750F14C7C8F34ULL, "vxw_blightlands_6a_0_0.vlm"},
	{0xC2F768F7FCDF3840ULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.dura.gad"},
	{0xC2F7A3145994E889ULL, "gen00fl_aveline6115487.cl"},
	{0xC2F7A585AED66203ULL, "6151363_m.fxe"},
	{0xC2F82C13B79A731FULL, "hlac_lt_0_lowtown_a2_196607_1x.dds"},
	{0xC2F84CAAC0570407ULL, "vxht_keep_n_42w_0_0.vlm"},
	{0xC2F8915EE8117B1EULL, "gen_im_npc_elf_arm_glv_lgte.uti"},
	{0xC2F8D87E752AA486ULL, "vxlt_undercityserialkiller_1gb_0_0.vlm"},
	{0xC2F8F8F823693E42ULL, "vxlt_0_lowtown_cfv_0_0.vlm"},
	{0xC2F913E58EAE53DCULL, "6153627_m.fxe"},
	{0xC2F9238FA0475EF6ULL, "788475687.ogg"},
	{0xC2F92AE37D782AEBULL, "6163787_f.fxe"},
	{0xC2F96F47B0F864C1ULL, "one140cr_charade6134608.cl"},
	{0xC2F9BF6A3B6F0E39ULL, "llac_ca_caverns1_mines_24424_18.dds"},
	{0xC2FA53EE7A0A3C83ULL, "gen00fl_isabela6096828.cl"},
	{0xC2FA7377E4B55D50ULL, "6176604_m.fxe"},
	{0xC2FB28805E224315ULL, "vxw_encounter2_n3_0_0.vlm"},
	{0xC2FB30E180460C93ULL, "936736908.ogg"},
	{0xC2FB358ED2ED5523ULL, "6175893_m.fxe"},
	{0xC2FBAD76726A1FBCULL, "6154204_m.fxe"},
	{0xC2FBC59AACDC21C4ULL, "vxht_0_hightown_a3_asx_0_0.vlm"},
	{0xC2FBCB21BFB54F5CULL, "6166102_m.fxe"},
	{0xC2FBF097A5A9AE9AULL, "llac_w_encounter2_65537_0.dds"},
	{0xC2FC51C740AA1C6AULL, "hlac_ca_caverns1_24425_11x.dds"},
	{0xC2FC95589F277C57ULL, "6125438_m.fxe"},
	{0xC2FCCD28181FCC50ULL, "73524764.ogg"},
	{0xC2FCF1336B9B2A2FULL, "6167476_m.fxe"},
	{0xC2FCF16C99A1BABBULL, "6163753_m.fxe"},
	{0xC2FCFD1F4F14D25FULL, "zz_one_debug6079330.cl"},
	{0xC2FD151C80933A41ULL, "vxlt_gallowstemplar_n_5yb_0_0.vlm"},
	{0xC2FD1B9450832173ULL, "zz_isaac_lowtownact3.cut"},
	{0xC2FD1EC8EE86DB58ULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_aoe_p.impa.gad"},
	{0xC2FD5112322DFD89ULL, "vxht_mansion2_3c2_0_0.vlm"},
	{0xC2FE3A9D092C6B42ULL, "6096210_m.fxe"},
	{0xC2FE569E8CDDF063ULL, "463186432.ogg"},
	{0xC2FEB2F5DBDE805FULL, "hlac_ca_cavern1_haunted_24425_11z.dds"},
	{0xC2FF055A69C44E12ULL, "vxlt_0_lowtown_docks_a3_bf9_0_0.vlm"},
	{0xC2FF2580FF7DDD84ULL, "vxht_mansion_n_6nj_0_0.vlm"},
	{0xC2FF63DB33BC3098ULL, "6208319_f.fxe"},
	{0xC300066DC182DA97ULL, "vxlt_0_lowtown_docks_n_aqn_0_0.vlm"},
	{0xC300330A5C6206D9ULL, "vxw_encounter2_camp_om_0_1.vlm"},
	{0xC3003398B0C66680ULL, "hlac_w_encounter2_65541_0y.dds"},
	{0xC30055EBC5DA0360ULL, "640291451.ogg"},
	{0xC300C33DD32FFCDEULL, "dae000_flemeth6000804.cl"},
	{0xC300CA526ED255E7ULL, "6130098_m.fxe"},
	{0xC30104AC48370757ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf4a.msh"},
	{0xC3010670EADF2C72ULL, "6045002_f.fxe"},
	{0xC3011AAC48372CC9ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hf4a.mmh"},
	{0xC3012F769048A257ULL, "w_woundedcoast_59_0.msh"},
	{0xC30141769048C0BAULL, "w_woundedcoast_59_0.mao"},
	{0xC30145769048C7C9ULL, "w_woundedcoast_59_0.mmh"},
	{0xC3017371D6F3CCA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_chest_out.ani"},
	{0xC3018C9EE2575479ULL, "vxw_bonepit_1ui_0_0.vlm"},
	{0xC30193611883D0F7ULL, "hlac_lt_0_lowtown_a2_n_-65539_0z.dds"},
	{0xC3021104C12EDC15ULL, "vxlt_undercitydungeon_1js_0_0.vlm"},
	{0xC30235B3C13418EBULL, "789237395.ogg"},
	{0xC30297DF38100311ULL, "824789184.ogg"},
	{0xC302B837538E713BULL, "gen00fl_anders6113093.cl"},
	{0xC3031BA9AB9C5F69ULL, "vxdr_primevaltomb_92v_0_0.vlm"},
	{0xC30342D9C1081588ULL, "vxht_keep_42g_0_0.vlm"},
	{0xC3039B74D34EFA59ULL, "vxlt_0_lowtown_docks_n_bu7_0_0.vlm"},
	{0xC303C3101D2F839EULL, "6128762_m.fxe"},
	{0xC303E5C6B913A9CAULL, "w_bonepitmining_a2_155_0.mao"},
	{0xC303E9C6B913B019ULL, "w_bonepitmining_a2_155_0.mmh"},
	{0xC303F3C6B913C107ULL, "w_bonepitmining_a2_155_0.msh"},
	{0xC30476BD1A8DAC1FULL, "m_dae230ar_undercity_dark.bnk"},
	{0xC304F91519E4EDD6ULL, "796891529.ogg"},
	{0xC3052BE0AB6DB3ACULL, "dae021_sister_samea_sta6150173.cl"},
	{0xC305923C128F787FULL, "gentr_talk.utt"},
	{0xC305B77537FC19C3ULL, "925494808.ogg"},
	{0xC305DE5DF11CF032ULL, "hlac_lt_0_lowtown_a3_n_c_131073_1y.dds"},
	{0xC305E575624C7F1FULL, "6040163_f.fxe"},
	{0xC305F5BE4D40978AULL, "339513428.ogg"},
	{0xC30613BFB6A0868DULL, "hlac_ca_cavern1_haunted_24421_8z.dds"},
	{0xC306612913EF69F8ULL, "hlac_dr_primevalentrance_1724_2z.dds"},
	{0xC306B5ED7FD1F2C1ULL, "vxht_0_hightown_a3_ca5_0_0.vlm"},
	{0xC306CE6D2EA4572FULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtd.phy"},
	{0xC306D9C18BBA73F1ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dslb_0s.dds"},
	{0xC3075CBDA2BB7548ULL, "467559530.ogg"},
	{0xC30785C280A46C5CULL, "6156527_m.fxe"},
	{0xC307EEBE87DBDCF0ULL, "zz_dre_debug6056953.cl"},
	{0xC3085BB4EBEBFC56ULL, "hlac_ht_brothel_n_3374_0z.dds"},
	{0xC308A59B9E985839ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg101a_0t.dds"},
	{0xC308BFA07D2023E3ULL, "6072316_m.fxe"},
	{0xC309099D360565D4ULL, "1014854825.ogg"},
	{0xC30928B2473040A8ULL, "follower_barks6156734.cl"},
	{0xC30931DCD80BCDF5ULL, "w_sundermount_12_0.rml"},
	{0xC30950DB3DE2ABBAULL, "vxlt_gallowstemplar_358_0_0.vlm"},
	{0xC30963C853011472ULL, "vxlt_0_lowtown_docks_a3_cvt_0_0.vlm"},
	{0xC309687D18B7065BULL, "vxlt_blackemporium_62_0_0.vlm"},
	{0xC30A7E49310E3746ULL, "153166013.ogg"},
	{0xC30B0DB452C7052FULL, "lgt160_wk_soph_hunter6119753.cl"},
	{0xC30B50F2E1CD68E4ULL, "vxht_mansion2b_31m_0_0.vlm"},
	{0xC30BADDD4273EA18ULL, "plt_lgt120pt_night_lies.plo"},
	{0xC30BC9F1BA793902ULL, "lt_0_lowtown_28_0.anb"},
	{0xC30C35A4600900BEULL, "6188621_m.fxe"},
	{0xC30C72A6BB9D0466ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.cess.ani"},
	{0xC30C7BFB1D99A89FULL, "gen00fl_merrill6106242.cl"},
	{0xC30C8384319F80D9ULL, "50663328.ogg"},
	{0xC30D435685FB58E0ULL, "6020550_m.fxe"},
	{0xC30D4CA14BAD9C12ULL, "6065809_cassandra_romantic.cut"},
	{0xC30D4EAAA174E16BULL, "vxca_caverns2prc_970_0_0.vlm"},
	{0xC30D53C0A70F9452ULL, "566551705.ogg"},
	{0xC30E1A1202B73418ULL, "vfx_sparkae_03.dds"},
	{0xC30E2B6F89DEF0CCULL, "vis121_saemus6167900.cl"},
	{0xC30E4CC71143DFDCULL, "m_dae150ar_sundermount.wbd"},
	{0xC30EFA4FFF15BC65ULL, "w_encounter3_mmap_ph.dds"},
	{0xC30F0FF3EA29F051ULL, "314786082.ogg"},
	{0xC30F8B71D700329AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_chest_out.evt"},
	{0xC310261EFAC61A19ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryrb.phy"},
	{0xC31082E1B324F495ULL, "vxlt_undercitydungeon_1mr_0_0.vlm"},
	{0xC31097B76AD4EEABULL, "dre000_dougal6129764.cl"},
	{0xC310B9F45ACFECA3ULL, "lgt310cr_shw_ors_inform_tmp.utc"},
	{0xC3115917E599929FULL, "247244857.ogg"},
	{0xC312136E80C614B6ULL, "gen00fl_aveline6116708.cl"},
	{0xC3121B05D5F99570ULL, "llac_lt_0_lowtown_docks_-196610_2.dds"},
	{0xC3127772D55953D2ULL, "vxca_caverns2prc_2ry_0_0.vlm"},
	{0xC31277FC371908E3ULL, "907683042.ogg"},
	{0xC312A92FF6FDBAA0ULL, "one200dg_solivitus.cnv"},
	{0xC313EB23908F3C5DULL, "vxht_chantry_n_2cd_0_0.vlm"},
	{0xC313EC3470857752ULL, "hlac_lt_0_lowtown_a2_-262143_0x.dds"},
	{0xC313EEC87B515F54ULL, "625618060.ogg"},
	{0xC3144856214791F1ULL, "vxht_0_hightown_a7x_0_0.vlm"},
	{0xC3147E066BBDBBA3ULL, "gen_im_arm_hlm_nobb.uti"},
	{0xC314BAA95CEAEBB3ULL, "gen00fl_carver6103614.cl"},
	{0xC314F1A99AA50B8AULL, "gen00fl_fenris6126496.cl"},
	{0xC31539A5C18F8C6BULL, "dae100_bran6097613.cl"},
	{0xC3155D564C927B54ULL, "vxca_caverns2prc_7u7_0_0.vlm"},
	{0xC31572855E819D70ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\arcanehorror.anb"},
	{0xC315842AE0BCD0A1ULL, "vxw_encounter2_h9_0_0.vlm"},
	{0xC3158B2FF6FFBDD8ULL, "one200dg_solivitus.bnk"},
	{0xC315BC385916D586ULL, "vxlt_undercityserialkiller_1qd_0_0.vlm"},
	{0xC315BDE8A48CF969ULL, "7571584.ogg"},
	{0xC315F11CCC149D09ULL, "6122037_m.fxe"},
	{0xC3160CB1A4B89CFEULL, "21164059.ogg"},
	{0xC3164FEF62882AFAULL, "6172880_m.fxe"},
	{0xC31650C786187283ULL, "6000692_m.fxe"},
	{0xC3168A71D7064F8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_chest_out.gad"},
	{0xC31697EE700F9518ULL, "vxlt_undercityencounter_1qv_0_0.vlm"},
	{0xC3171EFC52A0DD2CULL, "follower_barks6153760.cl"},
	{0xC317459EC53C26FAULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_cliffwall_n.dds"},
	{0xC317630672420BAAULL, "1016810955.ogg"},
	{0xC317827E4AF9E36FULL, "w_bonepitmining_a2_31_0_ll.mmh"},
	{0xC317D5A544F72CF2ULL, "vxlt_0_lowtown_docks_a3_cyu_0_0.vlm"},
	{0xC31807786761830BULL, "vxdr_slavers_133_0_0.vlm"},
	{0xC31836BD1513D802ULL, "26210991.ogg"},
	{0xC31892BC54694B35ULL, "79832280.ogg"},
	{0xC319ADEE97B4F519ULL, "vxlt_0_lowtown_n_ei9_0_0.vlm"},
	{0xC319C5F5E44F47A0ULL, "6172323_m.fxe"},
	{0xC319F7FF620AA166ULL, "vxlt_gallowstemplar_34w_0_0.vlm"},
	{0xC31A1E1778F77BF4ULL, "3p_200_mob(a-frame).stg"},
	{0xC31A5AA6BBA918D7ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.cess.evt"},
	{0xC31A952D3742C2B6ULL, "hlac_ht_mansion2_818_35x.dds"},
	{0xC31AD6638C3A35E4ULL, "vxlt_gallowstemplar_6sm_0_0.vlm"},
	{0xC31B2AA0B274C6CEULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v3_c.impa.evt"},
	{0xC31BDA243014574CULL, "mcr371_post_harvester6182080.cl"},
	{0xC31C9A5950E09009ULL, "6114432_m.fxe"},
	{0xC31DE6E449F29C36ULL, "gen00fl_sebastian6177909.cl"},
	{0xC31E320B033B5237ULL, "lowtown_fem_alienage1_sta6162572.cl"},
	{0xC31E333D81981324ULL, "6139432_m.fxe"},
	{0xC31E34F12B5B64E5ULL, "vxw_encounter2_cd_0_0.vlm"},
	{0xC31E73632016BEB9ULL, "llac_lt_0_lowtown_a2_n_65536_4.dds"},
	{0xC31E94AE8508A1AFULL, "framebuffereffects_twopassblurcocvertical_p0.psh.fxo.dx11"},
	{0xC31EADB28E89C464ULL, "vxlt_0_lowtown_docks_a3_n_d0a_0_1.vlm"},
	{0xC31ED72D7463D259ULL, "6180842_m.fxe"},
	{0xC31FB782A6303E90ULL, "6182711_m.fxe"},
	{0xC31FC06A83A7D91EULL, "vxht_0_hightown_a2_avl_0_0.vlm"},
	{0xC31FD7BA43E86B17ULL, "isa311_isabela6095335.cl"},
	{0xC32011D823D79AA9ULL, "fxa_cen_aoe_bb.mao"},
	{0xC32036AC8E546E5CULL, "6087517_m.fxe"},
	{0xC320F937A4CD22E0ULL, "vxlt_undercityserialkiller_1ik_0_1.vlm"},
	{0xC32113B41A3704DEULL, "ach_comp_armor.dds"},
	{0xC3212607EF578A0CULL, "vxlt_gallowstemplar_n_4n7_0_0.vlm"},
	{0xC3215BA6BBAF3960ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.cess.gad"},
	{0xC3216D69CF645433ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.dura.ani"},
	{0xC3216EEEA661D90DULL, "308216490.ogg"},
	{0xC321CA0C96079CECULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_door_female.ani"},
	{0xC32221A0B27AD679ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v3_c.impa.gad"},
	{0xC322A2755AE86E33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_action.evt"},
	{0xC322C2C2C32FBEFCULL, "vxht_0_hightown_n_9wz_0_0.vlm"},
	{0xC3236FEBE787EAFAULL, "vxht_0_hightown_a3_bgm_0_0.vlm"},
	{0xC3240404654F4FB7ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l_crnf.gad"},
	{0xC32497E001218A90ULL, "fx_noise01.dds"},
	{0xC324AD419D7B39FBULL, "art\\characters\\playercharacter\\humanfemale\\glv_masc.msh"},
	{0xC324BF419D7B5846ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masc.mao"},
	{0xC324CA0484D1CDDFULL, "art\\vfx\\combat\\vfx_f_and_fade_c.impa.gad"},
	{0xC324CB419D7B6CE5ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masc.mmh"},
	{0xC3250131CBCC74BCULL, "gen00fl_anders6111090.cl"},
	{0xC3251252C28D155EULL, "6162724_m.fxe"},
	{0xC325CBA63BC74720ULL, "vxht_0_hightown_n_cbj_0_1.vlm"},
	{0xC325FC7D33AE3328ULL, "lt_shop_probe_t3.dds"},
	{0xC3268A6B89F6C756ULL, "6081254_m.fxe"},
	{0xC327B0D61EA50E8CULL, "hlac_dr_dungeon_3217_10z.dds"},
	{0xC327CB49F4359847ULL, "vxht_keep_4n8_0_0.vlm"},
	{0xC327D70C960C1E2EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_door_female.gad"},
	{0xC32832A0B27F5EC7ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v3_c.impa.ani"},
	{0xC32869DC5400B9E6ULL, "vxlt_gallowstemplar_31x_0_0.vlm"},
	{0xC3286AB3567583C1ULL, "vxlt_gallowstemplar_n_5jg_0_0.vlm"},
	{0xC32879E6CF4973C3ULL, "qun120_arishok6050287.cl"},
	{0xC328A8A1EFB310BDULL, "vxht_keep_n_7ag_0_0.vlm"},
	{0xC328B2F3929AC0CBULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkback.evt"},
	{0xC32903252A095589ULL, "hlac_lt_undercitydungeon_-1_19z.dds"},
	{0xC3291D50560BD1C6ULL, "hlac_ca_caverns2_5802_13y.dds"},
	{0xC3299B755AEE8174ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_action.gad"},
	{0xC32A19046553DF44ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l_crnf.evt"},
	{0xC32A1B211D8BDD44ULL, "hlac_lt_0_lowtown_a2_65535_8y.dds"},
	{0xC32A710BF5F0FE04ULL, "782961599.ogg"},
	{0xC32A83326E8FCEC3ULL, "56568690.ogg"},
	{0xC32AD9901BD350B6ULL, "qun111_ketojan6040987.cl"},
	{0xC32AEC5FA2F5E651ULL, "mer140_bonepit_comments6077841.cl"},
	{0xC32AF299F906F484ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.mmh"},
	{0xC32AF65F439A270AULL, "zz_mrl_debug6109307.cl"},
	{0xC32B1884399FF96CULL, "hlac_lt_backalley_g_0_6z.dds"},
	{0xC32BB99EB5A71BEBULL, "vxht_keep_n_3zx_0_0.vlm"},
	{0xC32BBF0484D7D9CCULL, "art\\vfx\\combat\\vfx_f_and_fade_c.impa.evt"},
	{0xC32C47F0E25447EBULL, "6119877_m.fxe"},
	{0xC32C597BA2F7EBF7ULL, "150569704.ogg"},
	{0xC32C6A67213CED36ULL, "vxlt_0_lowtown_docks_b4i_0_1.vlm"},
	{0xC32C86DBFF95E038ULL, "101020334.ogg"},
	{0xC32CA6CE1D517D65ULL, "vxw_encounter2_nc_0_0.vlm"},
	{0xC32D81EFFFACBEC2ULL, "zz_mag_debug6053406.cl"},
	{0xC32D832669CFEFA7ULL, "hlac_lt_0_lowtown_a2_n_-131072_0x.dds"},
	{0xC32D8FD4F1CF5B8EULL, "gen_im_acc_amu_act3_05.uti"},
	{0xC32D930246D7817FULL, "vxht_0_hightown_a2_cd9_0_0.vlm"},
	{0xC32E3247759DF19EULL, "vxht_0_hightown_a2_aqm_0_0.vlm"},
	{0xC32E4F3EE0EECC17ULL, "ca_caverns1_mines_24423_ll.mmh"},
	{0xC32ED20C9612340DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\listen_door_female.evt"},
	{0xC32F06D5A08080D5ULL, "dre000_dougal6128880.cl"},
	{0xC32F15813F352016ULL, "517911950.ogg"},
	{0xC32F40257CDD0B2DULL, "vxlt_gallowstemplar_f_4uo_0_0.vlm"},
	{0xC32F6569CF70846AULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.dura.evt"},
	{0xC32FCBF392A1098CULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkback.gad"},
	{0xC3305CBD1AB2BB55ULL, "m_dae230ar_undercity_dark.wbd"},
	{0xC3306421C49742ADULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hp05.phy"},
	{0xC33078E48BBAF58EULL, "vxca_caverns1_jfw_0_0.vlm"},
	{0xC330AA755AF4B952ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_action.ani"},
	{0xC33138221789202CULL, "hlac_lt_gallowstemplar_n_7198_2x.dds"},
	{0xC331B4FF1628BD5EULL, "vxw_sundermount_37p_0_1.vlm"},
	{0xC3329A30DB0FCD69ULL, "994614343.ogg"},
	{0xC332DC2F5264B2E6ULL, "gen00fl_varric6118945.cl"},
	{0xC3344D9F3CB02DBDULL, "6084405_m.fxe"},
	{0xC3346E245357E77FULL, "vxca_cavern1_haunted_cn6_0_0.vlm"},
	{0xC334A858BB5B55CFULL, "ch2_d.psh.fxo"},
	{0xC334FD3B069CD131ULL, "177532192.ogg"},
	{0xC3350953726CFCC7ULL, "zz_mag_debug6051366.cl"},
	{0xC33580444F1F1EF5ULL, "follower_banter6136695.cl"},
	{0xC33586592A07036EULL, "hlac_w_woundedcoast_-131067_0x.dds"},
	{0xC335CAF392A5738AULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkback.ani"},
	{0xC336392DF64579FBULL, "dae000tr_mail_notification.ncs"},
	{0xC3364469CF766ADDULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.dura.gad"},
	{0xC33655FB0D7108C8ULL, "705928362.ogg"},
	{0xC336A3C12B1956EEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v1.evt"},
	{0xC336BB057E11F1DCULL, "vxht_keep_3c7_0_0.vlm"},
	{0xC3370E23FD2A0253ULL, "vxlt_0_lowtown_a2_n_c_db5_0_0.vlm"},
	{0xC337162AF27767E4ULL, "vxca_caverns1_mines_j28_0_0.vlm"},
	{0xC33753033336886EULL, "vxdr_0_deeproads_cum_0_1.vlm"},
	{0xC33756617A3DC411ULL, "vxlt_undercityencounter_1bo_0_1.vlm"},
	{0xC33788CA03BCEEA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_loop.gad"},
	{0xC337EFD80A39C7AEULL, "6121189_m.fxe"},
	{0xC3382B453EF8F030ULL, "vxlt_0_lowtown_docks_c50_0_0.vlm"},
	{0xC33831046560453DULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l_crnf.ani"},
	{0xC338A70484E23B25ULL, "art\\vfx\\combat\\vfx_f_and_fade_c.impa.ani"},
	{0xC338B717E4EE252EULL, "gen_im_wep_mag_sta_iconic.uti"},
	{0xC338EEF9748C6575ULL, "406878163.ogg"},
	{0xC3394E75D714820BULL, "vxht_0_hightown_a2_b7i_0_0.vlm"},
	{0xC339685C687701ADULL, "mag311_emile_de_launcet6171510.cl"},
	{0xC339CA35B64EEE41ULL, "hlac_ht_mansion_n_3365_0x.dds"},
	{0xC339D56F1B526361ULL, "mrl251ip_loot_c.utp"},
	{0xC33A1EE6569514D1ULL, "6097005_m.fxe"},
	{0xC33A49D46FB4767CULL, "hlac_lt_darktown_-1_86z.dds"},
	{0xC33A4AC711691756ULL, "m_dae150ar_sundermount.bnk"},
	{0xC33A5A1F0310728BULL, "6150269_m.fxe"},
	{0xC33AE1345422B20AULL, "492214255.ogg"},
	{0xC33B18AB0F4795E5ULL, "vxw_encounter2_ib_0_0.vlm"},
	{0xC33B5412AA2FB679ULL, "vxw_bonepit_1im_0_0.vlm"},
	{0xC33BC0D87C6BD2C5ULL, "lav_crust01a.dds"},
	{0xC33BC23F3E58C149ULL, "399303453.ogg"},
	{0xC33C19FCF6CCE708ULL, "400314798.ogg"},
	{0xC33C3810A39BB560ULL, "genip_node_g2.utp"},
	{0xC33C5438F37481E3ULL, "6162304_f.fxe"},
	{0xC33C9AC12B1DB299ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v1.gad"},
	{0xC33CA42467940A1EULL, "vxht_0_hightown_a2_aln_0_0.vlm"},
	{0xC33D7EE76FF6AC3EULL, "gen_im_cns_cft_pot_08.uti"},
	{0xC33E59CA03C2BDF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_loop.ani"},
	{0xC33E950933B4F998ULL, "82788706.ogg"},
	{0xC33EF33FD6B9D230ULL, "vxht_keep_n_5h7_0_0.vlm"},
	{0xC33F21BB93B6BE1BULL, "hlac_lt_0_lowtown_-131072_1x.dds"},
	{0xC33F3CA302A73DE7ULL, "vxht_keep_n_5k9_0_0.vlm"},
	{0xC33FD63344D4B089ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_reaction.ani"},
	{0xC3407D543B347228ULL, "qun220ip_at_chantry_n.utp"},
	{0xC34112D4904D3B0CULL, "ser270_emeric6149340.cl"},
	{0xC343279A82622DBCULL, "vxht_0_hightown_c95_0_0.vlm"},
	{0xC343388B9A0AC9A8ULL, "hlac_lt_darktown_-1_22x.dds"},
	{0xC343796DD3459576ULL, "6127143_m.fxe"},
	{0xC343ABC12B23EEE7ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_butt_atk_v1.ani"},
	{0xC343C8B350EE47A8ULL, "969929039.ogg"},
	{0xC343ECEE59711DA4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keep.msh"},
	{0xC343FEEE59713C0DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keep.mao"},
	{0xC34402EE59714346ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keep.mmh"},
	{0xC344A12116997ED6ULL, "fx_dot.dds"},
	{0xC344F4B982D9B54EULL, "6088297_m.fxe"},
	{0xC345F240CFF2DE55ULL, "vxw_encounter1_d5_0_1.vlm"},
	{0xC345F2DEC3FB5D71ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgta_0d.dds"},
	{0xC346E2132618B4B9ULL, "dae000_bartrand6107530.cl"},
	{0xC346F13344DAFD0BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_reaction.gad"},
	{0xC3474848C5EB0B2CULL, "hlac_lt_0_lowtown_a2_n_c_196607_0y.dds"},
	{0xC3474DE0940C3CDAULL, "vxht_brothel_n_6bo_0_0.vlm"},
	{0xC3475A23DD9773AEULL, "vxlt_warehouse_n_qg_0_0.vlm"},
	{0xC3476A676F3DABBDULL, "llac_lt_0_lowtown_a2_n_65535_7.dds"},
	{0xC347C052C90A9A8BULL, "vxht_0_hightown_a2_b8h_0_0.vlm"},
	{0xC347DB0DCD148E0BULL, "525311947.ogg"},
	{0xC347EADE39EC40D0ULL, "220041331.ogg"},
	{0xC3485B72199059A1ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.dura.ani"},
	{0xC34863821ABFD664ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\blendtree\\trooper_overlay.anb"},
	{0xC348B563B6472474ULL, "llac_ca_caverns1_24426_14.dds"},
	{0xC34912068A0678DCULL, "vxlt_undercityserialkiller_xv_0_0.vlm"},
	{0xC349EE29BE7E3590ULL, "vxht_0_hightown_n_a2q_0_0.vlm"},
	{0xC349F8447B2BBE68ULL, "6027829_m.fxe"},
	{0xC34A1ADF4AF8030BULL, "follower_banter6169738.cl"},
	{0xC34A1E3F985EA651ULL, "llac_lt_0_lowtown_docks_n_65534_1.dds"},
	{0xC34AECA2814312D7ULL, "plt_one111pt_find_gamlen.plo"},
	{0xC34B954A76EA8021ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanlean_1.mao"},
	{0xC34BC888362B473AULL, "vxht_0_hightown_a2p_0_0.vlm"},
	{0xC34C1EDD3B5DAF7BULL, "6096660_m.fxe"},
	{0xC34C26C9A35D7003ULL, "134736472.ogg"},
	{0xC34C7B00ACEF0A86ULL, "524726682.ogg"},
	{0xC34DEE3344E11770ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_medium_reaction.evt"},
	{0xC34E678FEBFC7A45ULL, "vxht_0_hightown_a3_bgf_0_0.vlm"},
	{0xC34EF9EE597B3B11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp_keep.phy"},
	{0xC34F667219968B43ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.dura.gad"},
	{0xC34FB11257114AF5ULL, "6152751_m.fxe"},
	{0xC34FD727F1E454EEULL, "vxht_0_hightown_a2_bpq_0_0.vlm"},
	{0xC34FD9993D52CEAEULL, "6155421_m.fxe"},
	{0xC350284CB9330C34ULL, "6113894_m.fxe"},
	{0xC35084A31A30FD36ULL, "vxw_encounter2_ow_0_0.vlm"},
	{0xC350F527DD8CD317ULL, "str_jaggedrock01_n.dds"},
	{0xC3513608466C6187ULL, "vxw_woundedcoast_uy_0_0.vlm"},
	{0xC351396A80CF7CA0ULL, "gen00fl_anders6113126.cl"},
	{0xC3518895CB2EE17BULL, "83573542.ogg"},
	{0xC3520FF8121C2D14ULL, "287188011.ogg"},
	{0xC3523CC09818FC8BULL, "vxht_0_hightown_n_ap8_0_0.vlm"},
	{0xC352B7C51A44CEE2ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkwall.phy"},
	{0xC352CF2AB0E30F44ULL, "hlac_lt_0_lowtown_a2_-131071_1x.dds"},
	{0xC352F599F9284C7BULL, "art\\vfx\\combat\\vfx_m_cre_glyph_paral_p.anb"},
	{0xC3535B96050E4456ULL, "6108621_m.fxe"},
	{0xC353699850ECDCFBULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\prp_ccliff02_n.dds"},
	{0xC3539D7EDAA600DDULL, "vxca_cave_134_0_0.vlm"},
	{0xC353FE82B75FFA78ULL, "gen00fl_fenris6128012.cl"},
	{0xC354357668BE3B2AULL, "dae000_flemeth6013273.cl"},
	{0xC354563EB1326513ULL, "6143171_m.fxe"},
	{0xC3547FF2CC6BA3A4ULL, "63369285.ogg"},
	{0xC354894E9ECCFCACULL, "art\\characters\\weapons\\maces\\w_mse_ms104a.msh"},
	{0xC3548F11AE7A15A4ULL, "964778650.ogg"},
	{0xC3548F4E9ECD069EULL, "art\\characters\\weapons\\maces\\w_mse_ms104a.mmh"},
	{0xC35499562A871F47ULL, "6079649_m.fxe"},
	{0xC3553DC83FF603DCULL, "6101479_m.fxe"},
	{0xC3556D59D023E325ULL, "6031126_m.fxe"},
	{0xC355A257148E91A2ULL, "6165174_m.fxe"},
	{0xC355CC00CF8D8C2EULL, "vxlt_warehouse_n_tx_0_0.vlm"},
	{0xC35627BD1605724EULL, "616640652.ogg"},
	{0xC3564372199C6E28ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.dura.evt"},
	{0xC3566C590C8D85ACULL, "6097260_m.fxe"},
	{0xC3569D4CF0395726ULL, "785213561.ogg"},
	{0xC356DA0CAE7699ECULL, "vxlt_gallowstemplar_n_5oy_0_0.vlm"},
	{0xC3571CB046BD5F66ULL, "vxlt_0_lowtown_by2_0_0.vlm"},
	{0xC35726E7DBE97178ULL, "follower_banter6173761.cl"},
	{0xC357514ECAAA9087ULL, "rdr111dg_harbormaster6055080.cl"},
	{0xC357B0E893C6D46CULL, "dae000_merchant_rare_em6177952.cl"},
	{0xC357F842274C7A88ULL, "vxlt_hangedman_24y_0_0.vlm"},
	{0xC35874B2BC122B6BULL, "hlac_lt_0_lowtown_a2_n_65535_6y.dds"},
	{0xC358D82584FEF1DAULL, "vxw_blightlands_9b_0_0.vlm"},
	{0xC359507F1A7CE1FFULL, "cd_dre000_varric.bnk"},
	{0xC3595E958731AB3DULL, "clf260_harley6093826.cl"},
	{0xC35A47161C6B073FULL, "6020610_m.fxe"},
	{0xC35A51050856A6B6ULL, "6177832_m.fxe"},
	{0xC35AA947A988F8ABULL, "6103580_m.fxe"},
	{0xC35B819A5FEE1727ULL, "vxw_encounter3_b6_0_0.vlm"},
	{0xC35BCD1C612454E6ULL, "t3_eey_gn2.tnt"},
	{0xC35BE62F18486DA4ULL, "w_bonepitdestroyed_a3_213_0.msh"},
	{0xC35BF82F18488C0DULL, "w_bonepitdestroyed_a3_213_0.mao"},
	{0xC35BF90A7216CBD8ULL, "vxht_0_hightown_c3o_0_0.vlm"},
	{0xC35BFC2F18489346ULL, "w_bonepitdestroyed_a3_213_0.mmh"},
	{0xC35C5D109CD5A7A1ULL, "hf_qfq_thug_leader.mor"},
	{0xC35CA0A3F1675DB7ULL, "hlac_dr_primevaltomb_658_11x.dds"},
	{0xC35CA82116AE1C04ULL, "fx_dot.mao"},
	{0xC35D20BDC07447E5ULL, "188441069.ogg"},
	{0xC35DAA88A303F32FULL, "vxlt_0_lowtown_h8z_0_0.vlm"},
	{0xC35DD32923FEEC3CULL, "seb321_leliana6199288.cl"},
	{0xC35E6C60B0A94736ULL, "vxht_0_hightown_a2_ap2_0_0.vlm"},
	{0xC35E78322B5711C5ULL, "vxht_keep_55j_0_0.vlm"},
	{0xC35E950F43D6A281ULL, "w_sundermount_122_0.phy"},
	{0xC35EF6800C2715B6ULL, "vxw_encounter2_jx_0_0.vlm"},
	{0xC35EFBA28153D54CULL, "plt_one111pt_find_gamlen.nss"},
	{0xC35F9941F15A0DA6ULL, "6179913_m.fxe"},
	{0xC35FAA19573A1D17ULL, "6208473_f.fxe"},
	{0xC35FC16EC7F50D45ULL, "6103986_m.fxe"},
	{0xC36026B28C2914A2ULL, "111577703.ogg"},
	{0xC36098E726212315ULL, "vxw_blightlands_77_0_0.vlm"},
	{0xC360AEAB2C8192D5ULL, "6161537_m.fxe"},
	{0xC360D2669428EDDAULL, "w_bonepitdestroyed_a3_108_0.phy"},
	{0xC3614C0D88573744ULL, "6093578_m.fxe"},
	{0xC3617F3F79E0EB37ULL, "qun160_vashothlead6053851.cl"},
	{0xC361D2AAC4CB6037ULL, "vxlt_0_lowtown_docks_a3_c3r_0_0.vlm"},
	{0xC361D327C2DB7506ULL, "787058414.ogg"},
	{0xC361F2B10798FC61ULL, "vxht_0_hightown_a3_bi8_0_0.vlm"},
	{0xC362031BE5EE6760ULL, "mag110st_cut_abomination.stg"},
	{0xC3623BD20B77C14EULL, "6141176_m.fxe"},
	{0xC362C7F0CDC31B63ULL, "art\\vfx\\combat\\vfx_q_ari_trip_act_c.anb"},
	{0xC362E0E24E676EE1ULL, "ch1emnotnt_vl.psh.fxo"},
	{0xC362E700E191CF40ULL, "gen00fl_anders6119454.cl"},
	{0xC363898C879C56FDULL, "6129450_m.fxe"},
	{0xC36402077F13E9FFULL, "art\\characters\\creatures\\dragon\\animation\\blendtree\\dragon_overlay.anb"},
	{0xC36448044FC7D4F8ULL, "dre000_varric6034642.cl"},
	{0xC36449D9EE07EDDFULL, "dre000_hireling16038663.cl"},
	{0xC364713FE14853DCULL, "6161275_m.fxe"},
	{0xC36523E2EE85A137ULL, "llac_lt_0_lowtown_a2_65536_10.dds"},
	{0xC3653A1752168715ULL, "w_bonepitdestroyed_a3_96_0.msh"},
	{0xC3653C1752168A8BULL, "w_bonepitdestroyed_a3_96_0.mmh"},
	{0xC365481752169F28ULL, "w_bonepitdestroyed_a3_96_0.mao"},
	{0xC3655DF7977D7804ULL, "6117906_m.fxe"},
	{0xC3659E0F43DCD09DULL, "w_sundermount_122_0.rml"},
	{0xC365EAC6B9667C06ULL, "w_bonepitmining_a2_155_0.phy"},
	{0xC3664C88E4F79258ULL, "184656097.ogg"},
	{0xC3664D4C48181378ULL, "ts_dwarven_arm_table.gda"},
	{0xC3666A1F19429308ULL, "vxlt_hangedman_21v_0_0.vlm"},
	{0xC36693B765EBDCEAULL, "148794223.ogg"},
	{0xC3672E6CBD5D0157ULL, "dae100_viscount6101875.cl"},
	{0xC3678DF9427FBEF7ULL, "744920777.ogg"},
	{0xC367A42C0CFF2428ULL, "hlac_lt_0_lowtown_a2_-65536_3z.dds"},
	{0xC367DCE2F3E81602ULL, "vxht_chantry_n_2v0_0_0.vlm"},
	{0xC368B8F30E611FBFULL, "6027710_m.fxe"},
	{0xC368FC5560ADC0F9ULL, "gen00fl_isabela6085682.cl"},
	{0xC369437E6CCD7686ULL, "hlac_lt_0_lowtown_docks_a3_-65538_9z.dds"},
	{0xC3697F56FF980D43ULL, "6094948_f.fxe"},
	{0xC36999A71D994FE9ULL, "408389722.ogg"},
	{0xC36A6AE7640CE458ULL, "vxht_0_hightown_c6l_0_0.vlm"},
	{0xC36A75139D8C4DEBULL, "449232966.ogg"},
	{0xC36A8C7CB1E91EB4ULL, "vxw_encounter2_4t_0_0.vlm"},
	{0xC36AB6FE32FCA936ULL, "hlac_ca_caverns2prc_5802_51z.dds"},
	{0xC36C11ABA5E39A13ULL, "553954383.ogg"},
	{0xC36CA4419DB8DF92ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masc.phy"},
	{0xC36E27477F2967E6ULL, "vxht_0_hightown_a2_a4n_0_0.vlm"},
	{0xC36E77BA94858A83ULL, "vxlt_0_lowtown_e1t_0_0.vlm"},
	{0xC36EA37BD2314834ULL, "hlac_lt_gallowstemplar_n_2394_1z.dds"},
	{0xC36F0146C4DF27F5ULL, "6146853_m.fxe"},
	{0xC36F3193202B448DULL, "vxw_sundermount_36y_0_0.vlm"},
	{0xC36F42E33D512134ULL, "631702834.ogg"},
	{0xC36F4BA3AC70DAABULL, "6096361_m.fxe"},
	{0xC36F5F3079B9D267ULL, "hlac_lt_0_lowtown_a2_-262144_4x.dds"},
	{0xC36F6C30806C42D1ULL, "dae000_flemeth6000953.cl"},
	{0xC36F94C5018C2B8CULL, "508321658.ogg"},
	{0xC36FA4F0CDCD698CULL, "art\\vfx\\combat\\vfx_q_ari_trip_act_c.mmh"},
	{0xC36FBA09EC686852ULL, "579449857.ogg"},
	{0xC36FE22C8F9FAA33ULL, "87747136.ogg"},
	{0xC37015D607141A8EULL, "vxdr_slavers_147_0_0.vlm"},
	{0xC3702B5BC7E7F699ULL, "277898952.ogg"},
	{0xC3705FB0CD80B0CBULL, "6144062_m.fxe"},
	{0xC37081D5B18C1B88ULL, "vxca_caverns1_mines_fm5_0_0.vlm"},
	{0xC37082D56A5942E8ULL, "plt_mag111pt_gamlen_pricia.nss"},
	{0xC370A6BCA3E07C8AULL, "hlac_w_encounter2_65535_0x.dds"},
	{0xC37205C11F5B08EFULL, "zz_and_debug6134094.cl"},
	{0xC37211F544B87675ULL, "6065307_m.fxe"},
	{0xC3725FF85E0FACD1ULL, "vxlt_0_lowtown_a2_n_c_ht5_0_0.vlm"},
	{0xC3728FEA7A21B5FFULL, "vxca_caverns2_7xj_0_0.vlm"},
	{0xC3729D14A02B7BEDULL, "6151488_m.fxe"},
	{0xC372A37769EE865AULL, "vxht_brothel_n_6uj_0_0.vlm"},
	{0xC372C738CA8CBB09ULL, "vxw_encounter3_d2_0_0.vlm"},
	{0xC372CFDD48432B7AULL, "lgt310_mag_melson6152206.cl"},
	{0xC372E31CD33E6C5CULL, "6166375_m.fxe"},
	{0xC3736377C0080597ULL, "gen_im_cns_tom_mag_archmage.uti"},
	{0xC373A61A16E35BD8ULL, "gen00fl_aveline6113657.cl"},
	{0xC373B1A08E279622ULL, "152875059.ogg"},
	{0xC3743FCBC0CADBC0ULL, "vxw_sundermount_23j_0_0.vlm"},
	{0xC374445BF16A4C0FULL, "vxw_sundermount_2zc_0_1.vlm"},
	{0xC374DE52F0D13797ULL, "30486037.ogg"},
	{0xC37506D67FCED923ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeloop.ani"},
	{0xC3751F6FA6CC643CULL, "6071692_f.fxe"},
	{0xC376089A012FBB17ULL, "6082117_m.fxe"},
	{0xC3763AB6C6C0A1E9ULL, "vxdr_primevalentrance_26y_0_0.vlm"},
	{0xC3764EBFE5DE2E82ULL, "vxht_chantry_n_2q1_0_0.vlm"},
	{0xC3768BD56A5DBDFBULL, "plt_mag111pt_gamlen_pricia.plo"},
	{0xC3768E3B153A4862ULL, "6020057_m.fxe"},
	{0xC377168486665D78ULL, "qun200_party_comment6095427.cl"},
	{0xC37792C51A63AB96ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkwall.mao"},
	{0xC3779EC51A63C035ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkwall.mmh"},
	{0xC377A0C51A63C3ABULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamwalkwall.msh"},
	{0xC377C16DB1BDEE71ULL, "vxlt_gallowstemplar_n_3sn_0_0.vlm"},
	{0xC377F55EDCF4F0EDULL, "vxca_caverns2_8kj_0_0.vlm"},
	{0xC377F9A47F773B9EULL, "6177429_m.fxe"},
	{0xC377FE2BE24AC36AULL, "6109998_m.fxe"},
	{0xC3781FBEFD863606ULL, "177429769.ogg"},
	{0xC3785EF3067A48A0ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastrock_da.dds"},
	{0xC378740367717636ULL, "905402109.ogg"},
	{0xC37930DEE0E08890ULL, "zz_vault_debug6177752.cl"},
	{0xC37951B5CE236BB4ULL, "vxw_blightlands_cn_0_0.vlm"},
	{0xC37A00E350E98E82ULL, "vxca_caverns2_8j7_0_1.vlm"},
	{0xC37A12861C812A1FULL, "6137590_m.fxe"},
	{0xC37A3F46D0D956D1ULL, "6160699_m.fxe"},
	{0xC37A9259C8D8E44FULL, "astest.ncs"},
	{0xC37AE3510F3F96ACULL, "w_sundermount_74_0.msh"},
	{0xC37AE9510F3FA09EULL, "w_sundermount_74_0.mmh"},
	{0xC37AF5510F3FB535ULL, "w_sundermount_74_0.mao"},
	{0xC37B128209F0FBA1ULL, "gen00fl_anders6125133.cl"},
	{0xC37B20A135A679BFULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg103a_0d.dds"},
	{0xC37B249ABE87F142ULL, "vxht_mansion_n_58a_0_0.vlm"},
	{0xC37B33E48A463F2AULL, "15007717.ogg"},
	{0xC37B53DCF81A3C26ULL, "mcr310_argument6074410.cl"},
	{0xC37BB21B881567ACULL, "6075417_m.fxe"},
	{0xC37BC339E9C67FB3ULL, "art\\levels\\ztools\\default\\generic\\0ri_basic.phy"},
	{0xC37BEB2E3FC8AF59ULL, "900152651.ogg"},
	{0xC37CA63F52DE6173ULL, "llac_ht_keep_3731_2.dds"},
	{0xC37CB34A240254ACULL, "art\\characters\\playercharacter\\elfmale\\arm_lgtp.msh"},
	{0xC37CB94A24025E9EULL, "art\\characters\\playercharacter\\elfmale\\arm_lgtp.mmh"},
	{0xC37CC54A24027335ULL, "art\\characters\\playercharacter\\elfmale\\arm_lgtp.mao"},
	{0xC37D55953B08E38FULL, "vxht_playershouse_n_18y_0_0.vlm"},
	{0xC37E1E373DFE6DA6ULL, "zz_vault_debug6201062.cl"},
	{0xC37E8144A137BAB1ULL, "vxw_bonepit_1pr_0_0.vlm"},
	{0xC37F19C13F156B4FULL, "w_blightlands_218_0.mao"},
	{0xC37F25C13F157FECULL, "w_blightlands_218_0.mmh"},
	{0xC37F2BC13F1589DEULL, "w_blightlands_218_0.msh"},
	{0xC37F30366CAFC30BULL, "1057671462.ogg"},
	{0xC37F5DAEDBD7AE80ULL, "hlac_ca_caverns1_mines_24424_8z.dds"},
	{0xC37F7C866CE28F44ULL, "vxht_0_hightown_c5i_0_0.vlm"},
	{0xC37F9CC5107488AFULL, "894724520.ogg"},
	{0xC37F9D11A770D027ULL, "dae400pt_post_game.ncs"},
	{0xC3802842BBCFD69BULL, "624866115.ogg"},
	{0xC3807E785B55128CULL, "6152606_m.fxe"},
	{0xC38109F8BC51F6CFULL, "62372049.ogg"},
	{0xC3811415EDA7711CULL, "qun131tr_party_ambient1.utt"},
	{0xC3813915BC82D389ULL, "vxw_encounter3_a1_0_0.vlm"},
	{0xC3815ED83269FBEEULL, "vxlt_hangedman_2cz_0_0.vlm"},
	{0xC38191B0B2EABC3FULL, "vxlt_gallowstemplar_n_5ov_0_0.vlm"},
	{0xC38206D20B825DFBULL, "vxlt_gallowscourtyard_n_74c_0_0.vlm"},
	{0xC382615660A24BCFULL, "llac_ht_mansion2b_818_12.dds"},
	{0xC3829CCAADBABCE3ULL, "6090851_m.fxe"},
	{0xC382A288A827EB1EULL, "vxht_brothel_n_6g4_0_0.vlm"},
	{0xC382B1A8B2C0F440ULL, "vxw_sundermount_24k_0_0.vlm"},
	{0xC382D3D80DAD690AULL, "ser270_templar6127498.cl"},
	{0xC3831ED67FDB3F1AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeloop.evt"},
	{0xC3837B4089272F09ULL, "932951420.ogg"},
	{0xC38465F074D51CDAULL, "214141407.ogg"},
	{0xC3847BFAB4C44051ULL, "102127060.ogg"},
	{0xC38501325DE4D242ULL, "6164065_m.fxe"},
	{0xC3850521F372CAD4ULL, "323299385.ogg"},
	{0xC3851BED741C4129ULL, "aip_qunariarcher_base.gda"},
	{0xC385358A96A9E89BULL, "6202357_m.fxe"},
	{0xC38581554753D9B3ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed_b.dura.ani"},
	{0xC385820A35DFCB52ULL, "dae000_arishok6067656.cl"},
	{0xC385E0AA2D9A7ED5ULL, "vxlt_0_lowtown_a2_n_c_gj4_0_0.vlm"},
	{0xC38634A2E1B78D93ULL, "182903091.ogg"},
	{0xC386367F1AA3A335ULL, "cd_dre000_varric.wbd"},
	{0xC386458A81F9B47AULL, "26227456.ogg"},
	{0xC386777F9C8EC1A4ULL, "hlac_w_encounter2_262145_0x.dds"},
	{0xC38727AF791784AEULL, "lgt311_shw_mer_mettin6156558.cl"},
	{0xC3875992E66076EAULL, "llac_ht_0_hightown_a3_-196607_0.dds"},
	{0xC38833CFBA4522DBULL, "vxw_bonepitmining_a2_9b_0_0.vlm"},
	{0xC38848756227EEA0ULL, "follower_banter6164754.cl"},
	{0xC3887AE256E09E15ULL, "276889252.ogg"},
	{0xC3887D8E1256DE95ULL, "mag110_vincento6017658.cl"},
	{0xC3890E4C75D77D61ULL, "lowtown_fem_alienage1_sta6162606.cl"},
	{0xC389692F300140F8ULL, "231576453.ogg"},
	{0xC3899E9BB9BCB6EEULL, "vxht_0_hightown_a2_bdu_0_0.vlm"},
	{0xC389C2707C157799ULL, "art\\characters\\creatures\\gallowsslavestatue\\cn_gss.msh"},
	{0xC389CA90903195FAULL, "zz_vault_debug6203370.cl"},
	{0xC389CC707C158887ULL, "art\\characters\\creatures\\gallowsslavestatue\\cn_gss.mmh"},
	{0xC389D0707C158FD4ULL, "art\\characters\\creatures\\gallowsslavestatue\\cn_gss.mao"},
	{0xC389F499E7DA1481ULL, "908313939.ogg"},
	{0xC38A1DD67FE15C0DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\chargeloop.gad"},
	{0xC38A388D1FD2DAA5ULL, "540373696.ogg"},
	{0xC38A486683ABA397ULL, "vxlt_gallowstemplar_f_34q_0_0.vlm"},
	{0xC38AF37CC355A1FBULL, "gen_im_trs_drk_one_mac_r1.uti"},
	{0xC38B39A2797932E8ULL, "hlac_lt_0_lowtown_a2_-65537_0x.dds"},
	{0xC38B53926F7C186DULL, "183438.ogg"},
	{0xC38B9D89F6184C10ULL, "855577977.ogg"},
	{0xC38BBFC38D4F2E43ULL, "ser221_gamlen6152722.cl"},
	{0xC38C154D0417758DULL, "vxw_sundermount_38w_0_0.vlm"},
	{0xC38C62A4C870D384ULL, "mrl310_merrill6144811.cl"},
	{0xC38C65D9D474B964ULL, "vxlt_0_lowtown_docks_n_arx_0_0.vlm"},
	{0xC38CC7BBEFB4ECAEULL, "llac_lt_gallowscourtyard_-131072_3.dds"},
	{0xC38CCC8E45C61E8DULL, "hlac_w_sundermount_262146_0y.dds"},
	{0xC38CF321932DD331ULL, "vxw_bonepit_1uq_0_0.vlm"},
	{0xC38D149C4EF06328ULL, "fxe_water_smokes_add.dds"},
	{0xC38D8FAF7A0E561CULL, "6182382_m.fxe"},
	{0xC38DD520347066A8ULL, "853252669.ogg"},
	{0xC38DE57DBEE6B2B0ULL, "6111100_m.fxe"},
	{0xC38E1734801A966EULL, "6160627_m.fxe"},
	{0xC38E367725C4F25FULL, "gen_im_trs_drk_junk2_03.uti"},
	{0xC38E755ADD3F657FULL, "hlac_w_bonepitdestroyed_a3_131078_0y.dds"},
	{0xC38E8A765ED5CD83ULL, "art\\vfx\\combat\\vfx_q_ari_charge_en_c.impa.ani"},
	{0xC38EECB8C1A3F10EULL, "l_mcr371ar_gallows_prison_nt.wbd"},
	{0xC38F20592574BAC1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_thin.phy"},
	{0xC38F6862926F84B6ULL, "hla_ht_0_hightown_a3_131070_0y.dds"},
	{0xC3900242A2A994BDULL, "lt_0_lowtown_n_44_0.anb"},
	{0xC390038DA4E0D4BFULL, "vxlt_gallowstemplar_n_5bq_0_0.vlm"},
	{0xC3900A11B8CFBE7DULL, "follower_banter6141558.cl"},
	{0xC3901367CA32E38FULL, "23952963.ogg"},
	{0xC3904D1AB18964EAULL, "mcr370_negotiation6080141.cl"},
	{0xC390958A36A8691AULL, "6139908_m.fxe"},
	{0xC390BAB8C81F7DE2ULL, "6075554_m.fxe"},
	{0xC39120ACC6607009ULL, "bkl_cl.mao"},
	{0xC3912385A4B70CC0ULL, "vxw_sundermount_29h_0_0.vlm"},
	{0xC391996FA3F8FE20ULL, "hlac_ca_caverns1_24421_4z.dds"},
	{0xC391C3E9EF97A71DULL, "w_sundermount_131_0.phy"},
	{0xC391D4833857E1B5ULL, "6078859_m.fxe"},
	{0xC3921734E5505D65ULL, "vxw_blightlands_7g_0_0.vlm"},
	{0xC39220F6A7FD9BCDULL, "6147155_m.fxe"},
	{0xC3929D2BED28E125ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gibbet_sa.dds"},
	{0xC392C55B90538204ULL, "llac_lt_warehouse_n_-65535_1.dds"},
	{0xC39365ED53173817ULL, "vxlt_0_lowtown_docks_n_bb8_0_0.vlm"},
	{0xC3937955476009EAULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed_b.dura.evt"},
	{0xC393B742FCCF32DFULL, "vxca_caverns1_mines_ck9_0_0.vlm"},
	{0xC393C3CFDC985745ULL, "6149896_m.fxe"},
	{0xC393ECF32E394850ULL, "vxlt_backalley_g_1dy_0_0.vlm"},
	{0xC393F414400C28FBULL, "vxlt_0_lowtown_docks_a3_n_d47_0_0.vlm"},
	{0xC394912EA3D0F334ULL, "721093453.ogg"},
	{0xC394E6C911EB3BE4ULL, "vxlt_0_lowtown_a2_n_efa_0_0.vlm"},
	{0xC3952534F15B2070ULL, "hlac_lt_0_lowtown_n_-196609_0x.dds"},
	{0xC3954FE28E81894BULL, "vxht_mansion2b_3dy_0_0.vlm"},
	{0xC395593728345645ULL, "217591811.ogg"},
	{0xC395664E9F045839ULL, "art\\characters\\weapons\\maces\\w_mse_ms104a.phy"},
	{0xC395857E31C74045ULL, "cd_mag311_nyssa.wbd"},
	{0xC39588C8ADCC8453ULL, "6153686_m.fxe"},
	{0xC396848B6B949580ULL, "vxlt_warehouse_n_15v_0_0.vlm"},
	{0xC396E260439845DEULL, "dre000_varric6027774.cl"},
	{0xC396F36391640B4BULL, "6164604_m.fxe"},
	{0xC397237C0ACE9207ULL, "453666643.ogg"},
	{0xC3976097544DBFB7ULL, "attack.ncs"},
	{0xC397B496A4287FAAULL, "39439295.ogg"},
	{0xC397C9E8217387B8ULL, "lgt000cr_elf_turn_in_2.utc"},
	{0xC397EDB2CB151B55ULL, "w_sundermount_78_0_ll.mmh"},
	{0xC39815722EB27039ULL, "501706738.ogg"},
	{0xC39848E3C21894E8ULL, "vxht_brothel_n_21x_0_0.vlm"},
	{0xC398C4E9EF9DC7A9ULL, "w_sundermount_131_0.rml"},
	{0xC398D0665001E35BULL, "6151006_m.fxe"},
	{0xC39902E2E0FE44D0ULL, "6106363_f.fxe"},
	{0xC399319A6FA790C9ULL, "ave221ip_duty_roster.utp"},
	{0xC3995D281490C334ULL, "fen210_fenris6109428.cl"},
	{0xC399D8C6402FF1D4ULL, "zz_one_debug6051496.cl"},
	{0xC39A33A03B6307B2ULL, "art\\characters\\weapons\\quivers\\w_qui_dq101.mao"},
	{0xC39A442D025BAA59ULL, "219464393.ogg"},
	{0xC39A58554765F05DULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfiremed_b.dura.gad"},
	{0xC39AAAC6A4F6A3F0ULL, "vxw_woundedcoast_1bf_0_0.vlm"},
	{0xC39BA2765EE0817AULL, "art\\vfx\\combat\\vfx_q_ari_charge_en_c.impa.evt"},
	{0xC39BF9717AD6F6C5ULL, "79470264.ogg"},
	{0xC39C0E469D119EEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter.gad"},
	{0xC39C23AC9CDFC6A9ULL, "741159187.ogg"},
	{0xC39C6095BED30395ULL, "gen_im_cft_reg_b1.uti"},
	{0xC39D3C331D957D26ULL, "vxht_0_hightown_a3_b93_0_0.vlm"},
	{0xC39D8111455041B0ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_001a_0t.dds"},
	{0xC39E423D3F66F651ULL, "6081994_m.fxe"},
	{0xC39EABD03151CD4DULL, "6081984_f.fxe"},
	{0xC39F17F29CCDA90AULL, "llac_ht_0_hightown_a2_65534_0.dds"},
	{0xC39FCEACC922EF8CULL, "6146908_m.fxe"},
	{0xC3A01AF64779A561ULL, "511236993.ogg"},
	{0xC3A01B36DC4AE39BULL, "vxdr_0_deeproads_bow_0_1.vlm"},
	{0xC3A02463E7B5CD8DULL, "vxlt_undercityencounter_104_0_0.vlm"},
	{0xC3A0425CA758833EULL, "336459615.ogg"},
	{0xC3A04413EDFA4958ULL, "bw_expo.mao"},
	{0xC3A08911D74675E5ULL, "vxw_blightlands_8f_0_0.vlm"},
	{0xC3A0C46EBCD695D0ULL, "hlac_lt_gallowstemplar_2394_9y.dds"},
	{0xC3A0C5444257E642ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_reva_0d.dds"},
	{0xC3A106800330DD89ULL, "hlac_lt_0_lowtown_a3_n_c_131072_1y.dds"},
	{0xC3A11190D0E6F9C7ULL, "632614863.ogg"},
	{0xC3A1525C63F3947AULL, "mag110_arianni6049706.cl"},
	{0xC3A16A9754562535ULL, "attack.cur"},
	{0xC3A1AFD9CE7A4475ULL, "mage_mindblast.dds"},
	{0xC3A1BADBADA6D37DULL, "6072322_f.fxe"},
	{0xC3A2226C67AE0B01ULL, "lgt210_qun_blk_qunari6142318.cl"},
	{0xC3A2291FEEC54B5FULL, "vxca_caverns1_mines_cl8_0_0.vlm"},
	{0xC3A235ACCE8E6FC5ULL, "6142996_m.fxe"},
	{0xC3A2475A47C0E4ACULL, "6150679_m.fxe"},
	{0xC3A2A1765EE69E6DULL, "art\\vfx\\combat\\vfx_q_ari_charge_en_c.impa.gad"},
	{0xC3A2A2BAC1EB0F22ULL, "794459776.ogg"},
	{0xC3A2B8E9EFA60789ULL, "w_sundermount_131_0.mao"},
	{0xC3A2C4E9EFA61C22ULL, "w_sundermount_131_0.mmh"},
	{0xC3A2C6E9EFA61FD8ULL, "w_sundermount_131_0.msh"},
	{0xC3A2F1469D178B81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter.evt"},
	{0xC3A30969C26B8529ULL, "hightown_fem_noble2_mob6169484.cl"},
	{0xC3A341A0067B6DEDULL, "vxht_0_hightown_bwc_0_0.vlm"},
	{0xC3A358A603E15464ULL, "vxlt_0_lowtown_a2_n_ekf_0_0.vlm"},
	{0xC3A3CF4BF204AFFAULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v1.gad"},
	{0xC3A4006347A32023ULL, "dae150_chandan6148802.cl"},
	{0xC3A4440CAEAE248EULL, "art\\characters\\creatures\\strider\\animation\\attackbackcenter.evt"},
	{0xC3A446A4A6BB229CULL, "sib111_mother6098974.cl"},
	{0xC3A4471130DE1034ULL, "660934538.ogg"},
	{0xC3A488E0FBFFF462ULL, "zz_followers_banter6157572.cl"},
	{0xC3A4A74CA405B534ULL, "vxw_blightlands_rm_0_0.vlm"},
	{0xC3A58815DF2CC46BULL, "688240542.ogg"},
	{0xC3A66A0C6F353468ULL, "6161464_m.fxe"},
	{0xC3A7318216F43547ULL, "130180259.ogg"},
	{0xC3A76BF3C8C9B5B3ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\da_material.mao"},
	{0xC3A7923AA47CEFDBULL, "6098594_m.fxe"},
	{0xC3A7ECC92944ADFBULL, "484421929.ogg"},
	{0xC3A9765F87D721BEULL, "161254173.ogg"},
	{0xC3A9A5D29A11B837ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_ffollower_scaredmage.gad"},
	{0xC3A9B8044935BEC1ULL, "crt_arcane_horror.wbd"},
	{0xC3A9D24BF2092011ULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v1.evt"},
	{0xC3AA4FC7DB68F14CULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_5.pwk"},
	{0xC3AAA2058ED7840EULL, "vxca_cave_27h_0_0.vlm"},
	{0xC3AAB596F6EF51EAULL, "vxca_cave_1d2_0_0.vlm"},
	{0xC3AAD7451D4909A7ULL, "vxht_0_hightown_n_bi8_0_0.vlm"},
	{0xC3AB325B85FFBAD7ULL, "vxlt_0_lowtown_a2_n_c_hlk_0_0.vlm"},
	{0xC3AB3B0CAEB43439ULL, "art\\characters\\creatures\\strider\\animation\\attackbackcenter.gad"},
	{0xC3ABAE100F8B95A6ULL, "vxht_0_hightown_a3_b44_0_0.vlm"},
	{0xC3AC14438F75FCD6ULL, "6084723_m.fxe"},
	{0xC3AC4A31D7CAE248ULL, "191970933.ogg"},
	{0xC3AC6ABD42FBABB0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02b.mmh"},
	{0xC3AC78BD42FBC3CAULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02b.msh"},
	{0xC3ACB762E7F46CECULL, "6168172_f.fxe"},
	{0xC3ACC1F9A3A0C9EEULL, "hlac_ht_mansion2_818_68z.dds"},
	{0xC3ADA1F73D3BAFC7ULL, "vxw_bonepitdestroyed_a3_1wl_0_1.vlm"},
	{0xC3ADE35E09B6BF9EULL, "hlac_lt_0_lowtown_a2_65535_1x.dds"},
	{0xC3AEBB3688638198ULL, "vxht_mansion2b_1hh_0_0.vlm"},
	{0xC3AED8C3BFAF3BB6ULL, "w_sundermount_238_0.msh"},
	{0xC3AEE2C3BFAF4CB4ULL, "w_sundermount_238_0.mmh"},
	{0xC3AEE6C3BFAF53C7ULL, "w_sundermount_238_0.mao"},
	{0xC3AF386E7ABCB9CFULL, "vxlt_backalley_n_1jm_0_0.vlm"},
	{0xC3AF5ABE57F4A25BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart.mmh"},
	{0xC3AF66BE57F4B6F8ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart.mao"},
	{0xC3AF78BE57F4D545ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_2m4m_dart.msh"},
	{0xC3AFC3A05F476CBEULL, "6084214_f.fxe"},
	{0xC3AFDF032D4F5756ULL, "vxlt_gallowstemplar_f_3rz_0_0.vlm"},
	{0xC3B03096C3F9F631ULL, "293468607.ogg"},
	{0xC3B06D5F799B440BULL, "6036688_m.fxe"},
	{0xC3B094BECB11ADFEULL, "6037004_m.fxe"},
	{0xC3B0B89743A6D84EULL, "801305733.ogg"},
	{0xC3B0BD3410DA3B79ULL, "6150137_m.fxe"},
	{0xC3B0C018C693860EULL, "249461660.ogg"},
	{0xC3B0E9469D23BBD0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter.ani"},
	{0xC3B11BD3CE2251CDULL, "35996577.ogg"},
	{0xC3B2245190F1DA4AULL, "follower_barks6153263.cl"},
	{0xC3B24C0CAEBA6F87ULL, "art\\characters\\creatures\\strider\\animation\\attackbackcenter.ani"},
	{0xC3B3023262323118ULL, "684249771.ogg"},
	{0xC3B337D6D5F453B3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp22_flared2a_1.mao"},
	{0xC3B3F7C3D1582CB7ULL, "6105432_m.fxe"},
	{0xC3B5B5707C3A9CBCULL, "art\\characters\\creatures\\gallowsslavestatue\\cn_gss.phy"},
	{0xC3B5F509E22FDD36ULL, "vxw_blightlands_6s_0_0.vlm"},
	{0xC3B62FD160BB7A47ULL, "hlac_ca_cavern1_haunted_24421_8x.dds"},
	{0xC3B6872A8318DAC2ULL, "6013437_m.fxe"},
	{0xC3B6BABD3059B10FULL, "282911288.ogg"},
	{0xC3B71A9A6FC06740ULL, "ave221ip_duty_roster.ncs"},
	{0xC3B7960D50A04FD1ULL, "character_character_p80.psh.fxo.dx11.assembly"},
	{0xC3B79708E82A4C73ULL, "vxw_bonepitdestroyed_a3_4w_0_0.vlm"},
	{0xC3B7EA4BF21586E0ULL, "art\\characters\\creatures\\spider\\animation\\big_attk_v1.ani"},
	{0xC3B8A0B4FAF59501ULL, "25508847.ogg"},
	{0xC3B8AC41280AB1A3ULL, "lt_0_lowtown_docks_n_180_0.mao"},
	{0xC3B8B041280AB8F0ULL, "lt_0_lowtown_docks_n_180_0.mmh"},
	{0xC3B8BB1B5575C9D3ULL, "6143870_m.fxe"},
	{0xC3B8BE41280AD00AULL, "lt_0_lowtown_docks_n_180_0.msh"},
	{0xC3B910EFFE11E39BULL, "hlac_ht_chantry_2534_37z.dds"},
	{0xC3B927370A3CC939ULL, "gen00fl_carver6122532.cl"},
	{0xC3B99AD463BAA0A6ULL, "vxht_mansion_n_4vb_0_0.vlm"},
	{0xC3BA2ECB3BE591F8ULL, "gen00fl_anders6125826.cl"},
	{0xC3BA3C58046113E0ULL, "hlac_dr_0_deeproads_5845_6z.dds"},
	{0xC3BA8075BFB09C23ULL, "6114849_m.fxe"},
	{0xC3BAAEE2AF21C48BULL, "llac_w_woundedcoast_65540_0.dds"},
	{0xC3BAE2B8C1C91B3CULL, "l_mcr371ar_gallows_prison_nt.bnk"},
	{0xC3BB48BE6BF24A7CULL, "vxlt_undercityencounter_13i_0_1.vlm"},
	{0xC3BB628BB9B2BCF4ULL, "6086895_m.fxe"},
	{0xC3BBC0510F76F239ULL, "w_sundermount_74_0.phy"},
	{0xC3BC773BE72793F1ULL, "515316178.ogg"},
	{0xC3BC9479A81EAAA1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_inquisitionsoldier1_.ani"},
	{0xC3BCB8B45E4B5143ULL, "follower_banter6141497.cl"},
	{0xC3BCCB6694774003ULL, "w_bonepitdestroyed_a3_108_0.msh"},
	{0xC3BCD9669477585DULL, "w_bonepitdestroyed_a3_108_0.mmh"},
	{0xC3BCDD6694775F6EULL, "w_bonepitdestroyed_a3_108_0.mao"},
	{0xC3BCF20612F40CA5ULL, "6105485_m.fxe"},
	{0xC3BD2D137A599A18ULL, "vxht_mansion2b_1ci_0_0.vlm"},
	{0xC3BD4C370896B75AULL, "w_encounter2_13_0.mmh"},
	{0xC3BD50370896BD91ULL, "w_encounter2_13_0.mao"},
	{0xC3BD5E370896D5A0ULL, "w_encounter2_13_0.msh"},
	{0xC3BD6F5A3EF8BB76ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_0y.dds"},
	{0xC3BD7D114E2EF990ULL, "vxht_0_hightown_n_ajy_0_0.vlm"},
	{0xC3BD904A2439B039ULL, "art\\characters\\playercharacter\\elfmale\\arm_lgtp.phy"},
	{0xC3BDBC084B14C026ULL, "6067809_m.fxe"},
	{0xC3BDD2D29A22ADBDULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_ffollower_scaredmage.ani"},
	{0xC3BF393C7050B3C4ULL, "6082342_f.fxe"},
	{0xC3BF65A3057998BFULL, "sst_ss_fleshgolem_1.gda"},
	{0xC3C01CBFC3145D04ULL, "hlac_w_encounter2_327685_0y.dds"},
	{0xC3C03C5FE7CD8564ULL, "vxlt_0_lowtown_a2_n_eqd_0_0.vlm"},
	{0xC3C06852FFB4816EULL, "vxlt_0_lowtown_a2_hdw_0_0.vlm"},
	{0xC3C0798B8D17B812ULL, "341612180.ogg"},
	{0xC3C08D2ADBDBF076ULL, "mcr371_angry_mage_m6087728.cl"},
	{0xC3C0968A221A69E8ULL, "6156101_m.fxe"},
	{0xC3C09AB282D0235DULL, "vxlt_0_lowtown_a2_n_c_hib_0_0.vlm"},
	{0xC3C0AB2B25388836ULL, "525058838.ogg"},
	{0xC3C0AB3196656C49ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft1a.phy"},
	{0xC3C0E7280E5CA450ULL, "llac_lt_0_lowtown_docks_n_-65537_2.dds"},
	{0xC3C0FE1835BCA883ULL, "ht_0_hightown_50_0.rml"},
	{0xC3C19A0F442B28BDULL, "w_sundermount_122_0.mao"},
	{0xC3C19E0F442B2FF6ULL, "w_sundermount_122_0.mmh"},
	{0xC3C1A80F442B40F4ULL, "w_sundermount_122_0.msh"},
	{0xC3C1C74ADFA98139ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_left_out.ani"},
	{0xC3C1DAC08138500AULL, "6085694_m.fxe"},
	{0xC3C200CE49E8ED3DULL, "lgt311ip_at_bandit_alley_lt.utp"},
	{0xC3C203C64ADBF4ABULL, "695800106.ogg"},
	{0xC3C2327A6BC87EBBULL, "hlac_dr_primevaltomb_658_1y.dds"},
	{0xC3C2C35933F04BA5ULL, "mag320_mage_conspirator6061727.cl"},
	{0xC3C2C9510F7D2055ULL, "w_sundermount_74_0.rml"},
	{0xC3C2DA191EED2D73ULL, "gen00fl_fenris6126730.cl"},
	{0xC3C3406C9035801EULL, "dae000_quintus.cnv"},
	{0xC3C35661807C319AULL, "foodstorage_sa.dds"},
	{0xC3C39F79A824DC43ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot1_inquisitionsoldier1_.gad"},
	{0xC3C448F377E7569CULL, "vxlt_0_lowtown_docks_n_b0a_0_0.vlm"},
	{0xC3C5247CEB38B7F6ULL, "fen000ip_sword_container.utp"},
	{0xC3C53E643D29CECAULL, "docks_male_elven_worker_sta6176098.cl"},
	{0xC3C59D433A1BDD2CULL, "vxlt_hangedman_sj_0_0.vlm"},
	{0xC3C6005D7AD8ED70ULL, "vxw_woundedcoast_1se_0_0.vlm"},
	{0xC3C6226C9037832EULL, "dae000_quintus.bnk"},
	{0xC3C6725053C69740ULL, "6080761_m.fxe"},
	{0xC3C6A2CA1CBC9691ULL, "hlac_lt_0_lowtown_-131071_0x.dds"},
	{0xC3C6EFC055169A67ULL, "623225344.ogg"},
	{0xC3C73CD926948456ULL, "hlac_lt_0_lowtown_n_1_3z.dds"},
	{0xC3C74BEB407E9F8DULL, "zz_fen_debug6165678.cl"},
	{0xC3C76E66EEB10456ULL, "vxlt_0_lowtown_docks_n_b54_0_0.vlm"},
	{0xC3C7E4E412C1F316ULL, "vxlt_gallowstemplar_f_6t9_0_0.vlm"},
	{0xC3C85AC38830923CULL, "6149682_m.fxe"},
	{0xC3C8946B66E1E772ULL, "115248362.ogg"},
	{0xC3C8A24ADFAF617BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\hair_left_out.gad"},
	{0xC3C8C929409094E5ULL, "follower_barks6154991.cl"},
	{0xC3C98BE15CDE9D67ULL, "dae021_brothel_onenter6196717.cl"},
	{0xC3C9B9231C6290A3ULL, "6099074_m.fxe"},
	{0xC3C9EAE5BD00FFB8ULL, "6040634_f.fxe"},
	{0xC3C9F187F2431A39ULL, "framebuffereffects_verticalblur_p0.vsh.fxo.dx11"},
	{0xC3C9F794D110546EULL, "6172659_m.fxe"},
	{0xC3CA0342A2DB0EB5ULL, "lt_0_lowtown_n_44_0.rml"},
	{0xC3CB69A6D0D8FB9BULL, "mcr370_cut_arrival_gallows6078043.cl"},
	{0xC3CB9EF06C4FB298ULL, "vxht_mansion2b_1ff_0_0.vlm"},
	{0xC3CC1C285EA8EACFULL, "vxlt_backalley_n_1po_0_0.vlm"},
	{0xC3CC3B3D3B65E1DAULL, "455417192.ogg"},
	{0xC3CCEC08A8CF8026ULL, "836413049.ogg"},
	{0xC3CD0A7F85B2F388ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelg.msh"},
	{0xC3CD1BE44E92C551ULL, "gen00fl_varric6099548.cl"},
	{0xC3CD1C7F85B31219ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelg.mao"},
	{0xC3CD287F85B326B2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelg.mmh"},
	{0xC3CD92AD42AEC0EEULL, "follower_banter6161034.cl"},
	{0xC3CDA4EFA813ABD5ULL, "214024374.ogg"},
	{0xC3CDAB196246CC44ULL, "6085242_f.fxe"},
	{0xC3CDB800BDC216C5ULL, "lt_darktown_3567_ll.mmh"},
	{0xC3CEDA04495514C3ULL, "crt_arcane_horror.bnk"},
	{0xC3CEE12C92923A33ULL, "6020445_m.fxe"},
	{0xC3CF0C8F74C63BDDULL, "vxlt_0_lowtown_a2_n_c_hda_0_0.vlm"},
	{0xC3CF37FF95A92711ULL, "gencr_templar_captain.utc"},
	{0xC3CF92974598C712ULL, "6126316_f.fxe"},
	{0xC3CFAE0A2180491DULL, "6153734_m.fxe"},
	{0xC3CFCDFA0CB660F9ULL, "lgt150_nat_were_supporter6110957.cl"},
	{0xC3D0622E1A57C1FAULL, "18866205.ogg"},
	{0xC3D087215CAD5CC9ULL, "w_bonepitmining_a2_134_0.phy"},
	{0xC3D1076069BA5EA0ULL, "200553122.ogg"},
	{0xC3D138029A4ED408ULL, "6129019_m.fxe"},
	{0xC3D14C632A3C0F21ULL, "vxw_encounter2_74_0_0.vlm"},
	{0xC3D17412664FAE33ULL, "949451881.ogg"},
	{0xC3D18A486A686783ULL, "6137190_m.fxe"},
	{0xC3D1E388DB8B3710ULL, "w_bonepitmining_a2_72_0.mao"},
	{0xC3D1E788DB8B3E23ULL, "w_bonepitmining_a2_72_0.mmh"},
	{0xC3D1F588DB8B567DULL, "w_bonepitmining_a2_72_0.msh"},
	{0xC3D26C8C8D6FFDABULL, "68451791.ogg"},
	{0xC3D2B4606A457358ULL, "413229432.ogg"},
	{0xC3D2CE457FAB338DULL, "211914160.ogg"},
	{0xC3D2D9BE2DF6F16CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston_6.pwk"},
	{0xC3D320A616EA62A9ULL, "vxht_keep_n_60a_0_0.vlm"},
	{0xC3D327021AF9EEC5ULL, "vxlt_0_lowtown_docks_b6d_0_1.vlm"},
	{0xC3D3BFA34BB80C83ULL, "vxht_mansion_n_3og_0_0.vlm"},
	{0xC3D3D798EFCD81F8ULL, "gen00fl_varric6033565.cl"},
	{0xC3D3D9218EED52B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_exit.evt"},
	{0xC3D4EE60A582F61DULL, "hlac_ca_caverns1_mines_24425_13x.dds"},
	{0xC3D4F719DB826C7EULL, "vxlt_gallowstemplar_f_4sn_0_0.vlm"},
	{0xC3D5A951D205CB48ULL, "vxca_cavern1_haunted_idi_0_0.vlm"},
	{0xC3D5D8A9B12EA877ULL, "mag371_elsa6064702.cl"},
	{0xC3D65F7E31FE96EFULL, "cd_mag311_nyssa.bnk"},
	{0xC3D6EBA455DFDB64ULL, "10130661.ogg"},
	{0xC3D733CB115CD6E8ULL, "mag360_alain6044662.cl"},
	{0xC3D7AA81933CFC60ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1b.phy"},
	{0xC3D83A964C43C740ULL, "vxw_blightlands_g6_0_0.vlm"},
	{0xC3D947742FCA6BE8ULL, "6087131_m.fxe"},
	{0xC3D9B9C65382A6D2ULL, "gen_im_arm_cht_elf_comc_t1.uti"},
	{0xC3D9C084A3A716C1ULL, "136613139.ogg"},
	{0xC3D9ED173D3A329DULL, "6089464_m.fxe"},
	{0xC3DAD2218EF365F4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_exit.gad"},
	{0xC3DADAFE1378CCB1ULL, "gen00fl_aveline6139504.cl"},
	{0xC3DB3BBF96330CEBULL, "static1_punch_ds2.psh.fxo"},
	{0xC3DBB5354953CEA1ULL, "water_cycle3.dds"},
	{0xC3DC20B3E1266A4FULL, "436252517.ogg"},
	{0xC3DC2C7FFB3153D2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\tarp_dp.dds"},
	{0xC3DCAACE4A007509ULL, "gen00fl_anders6119206.cl"},
	{0xC3DD5309848852B3ULL, "6027545_m.fxe"},
	{0xC3DDEC04C7A22645ULL, "6135653_m.fxe"},
	{0xC3DE12B4B0093F28ULL, "vxlt_darktown_6bv_0_0.vlm"},
	{0xC3DE305C48A2E885ULL, "mer221_sabin6080110.cl"},
	{0xC3DE8BC10918B571ULL, "vxca_caverns2_9dv_0_0.vlm"},
	{0xC3DEBC39EA1AF128ULL, "art\\levels\\ztools\\default\\generic\\0ri_basic.mmh"},
	{0xC3DEBDEEB4C57921ULL, "mrl000cr_wraith.ncs"},
	{0xC3DEC839EA1B05CBULL, "art\\levels\\ztools\\default\\generic\\0ri_basic.mao"},
	{0xC3DEDA39EA1B2452ULL, "art\\levels\\ztools\\default\\generic\\0ri_basic.msh"},
	{0xC3DF15DCC847562BULL, "t3_dog_ob2.tnt"},
	{0xC3DFDB3DF49A7960ULL, "720667798.ogg"},
	{0xC3DFDC0F42C56556ULL, "zz_fen_debug6134343.cl"},
	{0xC3E051A4749C4E9FULL, "hlac_lt_0_lowtown_n_65534_1y.dds"},
	{0xC3E0B791FE2AF865ULL, "1062800215.ogg"},
	{0xC3E0FE7BE5260165ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomed_sa.dds"},
	{0xC3E0FF8B8A50D3EBULL, "231560501.ogg"},
	{0xC3E1623D93D7A958ULL, "ico_potion_offense.dds"},
	{0xC3E187ADA7CB263BULL, "6113524_f.fxe"},
	{0xC3E1BC3DF7843C65ULL, "hlac_ht_chantry_2534_4z.dds"},
	{0xC3E1E1218EF99DD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_exit.ani"},
	{0xC3E299564C9E2CD6ULL, "hlac_ht_0_hightown_n_65536_2y.dds"},
	{0xC3E2DA944B98C95AULL, "568829978.ogg"},
	{0xC3E2E68C69E2B532ULL, "t3_shl_blu.tnt"},
	{0xC3E36B3708B7BB65ULL, "w_encounter2_13_0.phy"},
	{0xC3E3D9FD0F565BEEULL, "790594206.ogg"},
	{0xC3E45638218283C1ULL, "6148381_m.fxe"},
	{0xC3E47F809CCA6303ULL, "lowtown_fem_lowlife_mob6169187.cl"},
	{0xC3E48DA6D3F68F22ULL, "vxw_bonepitmining_a2_o5_0_0.vlm"},
	{0xC3E4EB49CCAA2517ULL, "art\\vfx\\combat\\vfx_glb_ele_p_major_c.impa.gad"},
	{0xC3E53D497C474F40ULL, "hlac_lt_darktown_-1_19z.dds"},
	{0xC3E59DBD55B60E7AULL, "6101225_f.fxe"},
	{0xC3E60761AAC48871ULL, "vxlt_0_lowtown_n_deh_0_0.vlm"},
	{0xC3E60BF14426FD1CULL, "vxca_caverns2_950_0_0.vlm"},
	{0xC3E6AC5734FECC56ULL, "art\\vfx\\textures\\vfx_debris_blurred.dds"},
	{0xC3E6AC733E39DFC0ULL, "vxw_blightlands_h7_0_0.vlm"},
	{0xC3E6DEF53E4AB297ULL, "6096425_f.fxe"},
	{0xC3E74B4AC87E9FD8ULL, "character_character_p90.psh.fxo.dx11"},
	{0xC3E782B9C308A720ULL, "6114235_m.fxe"},
	{0xC3E82E903F36777AULL, "gen00fl_aveline6132661.cl"},
	{0xC3E8347BC00F5EFAULL, "vxca_caverns2_8bw_0_1.vlm"},
	{0xC3E84603BD248369ULL, "469777113.ogg"},
	{0xC3E848CC97C09850ULL, "6107381_m.fxe"},
	{0xC3E89003B186644DULL, "mag100_karras6034687.cl"},
	{0xC3E892E0AC0C4531ULL, "vxht_0_hightown_a2_cb1_0_0.vlm"},
	{0xC3E89388CEA94768ULL, "6189013_m.fxe"},
	{0xC3E8AA47E14EE62FULL, "23968852.ogg"},
	{0xC3E9307990BA5B0FULL, "mag311_emile_de_launcet6043811.cl"},
	{0xC3E93164BAF0AAD6ULL, "vxca_caverns2_96c_0_0.vlm"},
	{0xC3E94FB72CAD4BF4ULL, "qun211cr_elf_warrior6072348.cl"},
	{0xC3E998031653CEA7ULL, "vxht_0_hightown_n_9uw_0_0.vlm"},
	{0xC3E9A07672C08A5CULL, "6139487_m.fxe"},
	{0xC3E9D1FD06F93AA3ULL, "4704172.ogg"},
	{0xC3EA4C3708BDA571ULL, "w_encounter2_13_0.rml"},
	{0xC3EB3A2C7F563EBBULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.bl_atk_v2.evt"},
	{0xC3EB5B7B57430266ULL, "lgt320_shw_ors_marlein6154739.cl"},
	{0xC3EB9029BA14C9F1ULL, "vxca_cavern1_haunted_dpe_0_0.vlm"},
	{0xC3EC0049CCB067A4ULL, "art\\vfx\\combat\\vfx_glb_ele_p_major_c.impa.evt"},
	{0xC3EC71BD43319BCBULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye02b.phy"},
	{0xC3ECBE1CA521AB76ULL, "zz_ave_debug6134018.cl"},
	{0xC3ED7CE5A87CFB3FULL, "lt_0_lowtown_a2_32_0.rml"},
	{0xC3ED9431968C336EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft1a.mmh"},
	{0xC3EDA031968C4705ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft1a.mao"},
	{0xC3EDAE31968C5F5CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rcksoft1a.msh"},
	{0xC3EDEC8104CB4F45ULL, "llac_ca_cavern1_haunted_24424_7.dds"},
	{0xC3EEC5ECF4A2D6ABULL, "vxlt_gallowscourtyard_n_710_0_0.vlm"},
	{0xC3EED34DAA88A386ULL, "vxlt_gallowstemplar_f_4up_0_0.vlm"},
	{0xC3EEEA81016E1532ULL, "zz_qun_debug6028292.cl"},
	{0xC3EF7514A580B825ULL, "vxw_bonepit_li_0_0.vlm"},
	{0xC3EFDFBECABA7E3EULL, "6109239_m.fxe"},
	{0xC3EFF4FC64121B7DULL, "vxht_bartrand_n_64f_0_0.vlm"},
	{0xC3F02F51D9ECD52CULL, "lt_0_lowtown_a2_n_c_8_0.msh"},
	{0xC3F03551D9ECDF1EULL, "lt_0_lowtown_a2_n_c_8_0.mmh"},
	{0xC3F04151D9ECF3B5ULL, "lt_0_lowtown_a2_n_c_8_0.mao"},
	{0xC3F04EE5777E67CAULL, "hlac_lt_0_lowtown_131074_0y.dds"},
	{0xC3F1047A18B9D528ULL, "dae000ip_crowd_ff1_talk13.utp"},
	{0xC3F1255925C78DFDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_thin.mao"},
	{0xC3F1295925C79436ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_thin.mmh"},
	{0xC3F1335925C7A534ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_thin.msh"},
	{0xC3F1E670E9E52929ULL, "vxht_mansion_n_6k4_0_0.vlm"},
	{0xC3F20C2AD84D0BCAULL, "newcontentavailable.gfx"},
	{0xC3F237FD5904BFEAULL, "gen00fl_anders6125705.cl"},
	{0xC3F29E4F47A72102ULL, "gen00fl_bethany6112217.cl"},
	{0xC3F2AAD34858C941ULL, "405282729.ogg"},
	{0xC3F2D295EA6330E8ULL, "6057343_m.fxe"},
	{0xC3F3008BE1F4E500ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_001a_0d.dds"},
	{0xC3F36BAC3FCE3557ULL, "vxlt_0_lowtown_a2_n_c_hsf_0_0.vlm"},
	{0xC3F3948C78B9AB97ULL, "126146771.ogg"},
	{0xC3F3C854C8CD912AULL, "892098271.ogg"},
	{0xC3F4776E2E548410ULL, "zz_approval_debug6003711.cl"},
	{0xC3F4C7FD4C339789ULL, "vxw_encounter3_99_0_0.vlm"},
	{0xC3F4D8CD8F28FF0FULL, "hlac_lt_0_lowtown_docks_a3_-2_0y.dds"},
	{0xC3F51E50302FF840ULL, "vxw_blightlands_m4_0_0.vlm"},
	{0xC3F56F78512A6088ULL, "vxca_caverns1_mines_eib_0_0.vlm"},
	{0xC3F5797BE3D2FBD2ULL, "212921007.ogg"},
	{0xC3F581AB5EC2493CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_market_stall_0_d.dds"},
	{0xC3F651DC1340522CULL, "297503497.ogg"},
	{0xC3F6F4874231FC18ULL, "vxht_mansion2b_1we_0_0.vlm"},
	{0xC3F754F98074584FULL, "plt_genpt_app_bethany.plo"},
	{0xC3F75AC0F9CE0E82ULL, "6021120_m.fxe"},
	{0xC3F8DEE3FCD3BF5AULL, "vxw_sundermount_2cp_0_0.vlm"},
	{0xC3F91849CCBB1A9DULL, "art\\vfx\\combat\\vfx_glb_ele_p_major_c.impa.ani"},
	{0xC3FA3CE2F1AA2B08ULL, "hlac_dr_primevalentrance_1724_3y.dds"},
	{0xC3FA4B19ADD2D051ULL, "6013338_f.fxe"},
	{0xC3FA51A326EABA62ULL, "442066620.ogg"},
	{0xC3FA52ADF6FE4127ULL, "675970449.ogg"},
	{0xC3FA87788796026EULL, "vxlt_0_lowtown_docks_cno_0_0.vlm"},
	{0xC3FABB1D4E3CD8D7ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden7.mmh"},
	{0xC3FACE7EF4704165ULL, "746493804.ogg"},
	{0xC3FAD11D4E3CFE49ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden7.msh"},
	{0xC3FB42BFC1E19C1AULL, "vxca_caverns2_71h_0_0.vlm"},
	{0xC3FB982C05FD3863ULL, "578377750.ogg"},
	{0xC3FBE4FF6AD87807ULL, "lowtown_argument6169366.cl"},
	{0xC3FC0F71329077CAULL, "plt_cod_aow_spellcombo8.plo"},
	{0xC3FD3710E1AB3841ULL, "6033714_f.fxe"},
	{0xC3FD37C9E698EF2BULL, "vxlt_gallowscourtyard_n_743_0_0.vlm"},
	{0xC3FD57E6FCEF6CACULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mede.phy"},
	{0xC3FD70215CD423EEULL, "w_bonepitmining_a2_134_0.mmh"},
	{0xC3FD7C215CD43785ULL, "w_bonepitmining_a2_134_0.mao"},
	{0xC3FD88C9A6A1B707ULL, "vxlt_0_lowtown_a2_n_c_ffg_0_0.vlm"},
	{0xC3FD8A215CD44FDCULL, "w_bonepitmining_a2_134_0.msh"},
	{0xC3FDC781935DFCF5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1b.msh"},
	{0xC3FDC981935E006BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1b.mmh"},
	{0xC3FDD581935E1408ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_tablelong1b.mao"},
	{0xC3FDDE044156F477ULL, "mag291_vincento6166235.cl"},
	{0xC3FE519BBCB096BEULL, "6102339_m.fxe"},
	{0xC3FF5A1386951C6BULL, "287919245.ogg"},
	{0xC3FFB1CE2654C7DDULL, "vxlt_hangedman_n_20o_0_0.vlm"},
	{0xC3FFD98453596A69ULL, "lgt311ar_gang_hideout.are"},
	{0xC4002BFD28139BDDULL, "hlac_ca_cavern1_haunted_24424_16x.dds"},
	{0xC4002FD3C8EFFCDAULL, "gen00fl_anders6125815.cl"},
	{0xC40062EBFD59658BULL, "6147122_m.fxe"},
	{0xC40078F325922401ULL, "hlac_ht_0_hightown_-3_0z.dds"},
	{0xC4007BF48A329D0CULL, "798362470.ogg"},
	{0xC4013632B251B0F7ULL, "hlac_lt_0_lowtown_a3_n_c_-1_1y.dds"},
	{0xC4015EE4D305C598ULL, "tbe100_xenon_urchin6189744.cl"},
	{0xC4016C12CF414D2BULL, "docks_fem_sailor_mob6176316.cl"},
	{0xC40178AF4B76EBC8ULL, "hlac_lt_undercitydungeon_-1_27z.dds"},
	{0xC40178E1842833C0ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls108a_0s.dds"},
	{0xC4017F28F1155C7AULL, "6083209_m.fxe"},
	{0xC401B9E1357F282AULL, "76542288.ogg"},
	{0xC402125477AC6740ULL, "ser191_emeric6120565.cl"},
	{0xC4022760B1D8C19EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masd_0n.dds"},
	{0xC402725D5B5E246CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\gt1a_portculgate_2.pwk"},
	{0xC402775D48F66867ULL, "hlac_ht_brothel_n_2995_4y.dds"},
	{0xC4028CF682C3075CULL, "6102125_f.fxe"},
	{0xC402B7BB02B9312EULL, "hlac_dr_primevaltomb_11985_0z.dds"},
	{0xC4032AA5FDB0D144ULL, "144454163.ogg"},
	{0xC40339DA3E29B009ULL, "vxw_encounter3_28_0_0.vlm"},
	{0xC403692AC5CEC69EULL, "6154189_m.fxe"},
	{0xC403902D222610C0ULL, "vxw_blightlands_v5_0_0.vlm"},
	{0xC403ED38DD692297ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.gad"},
	{0xC404AD6117CA33B6ULL, "6083331_m.fxe"},
	{0xC404C48A45C4E48FULL, "mcr371_post_harvester6083690.cl"},
	{0xC404D09A16AB32E1ULL, "hlac_w_sundermount_-196608_0y.dds"},
	{0xC40577043C539DF0ULL, "vxca_caverns2prc_62f_0_1.vlm"},
	{0xC405EE167F632FFDULL, "dae000_mother6035944.cl"},
	{0xC4060BC8F2810F56ULL, "vxht_0_hightown_a1i_0_0.vlm"},
	{0xC40639F935AD46B3ULL, "131741466.ogg"},
	{0xC40685BC06B8D8D0ULL, "follower_banter6167175.cl"},
	{0xC406B4A724D312EBULL, "464247262.ogg"},
	{0xC40750C0EEC9D7DAULL, "vxw_sundermount_2dq_0_0.vlm"},
	{0xC40754840036EE67ULL, "vxw_sundermount_31n_0_0.vlm"},
	{0xC40830812F3727D9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_2_darren.gad"},
	{0xC408344EBC4850F3ULL, "vxca_caverns1_gi2_0_0.vlm"},
	{0xC4083865F5735732ULL, "208163646.ogg"},
	{0xC4091B1142EE3C23ULL, "186942733.ogg"},
	{0xC4091B2541CFCEACULL, "6136939_f.fxe"},
	{0xC4092D23C6E884F7ULL, "hlac_lt_warehouse_n_1_2z.dds"},
	{0xC409A03C5B2215F4ULL, "6138821_f.fxe"},
	{0xC40A895E6569FE4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_lean_forward.evt"},
	{0xC40B0238DD6F6524ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.evt"},
	{0xC40B985AAF3477F2ULL, "gen00fl_dog6167157.cl"},
	{0xC40C5640AB9809AFULL, "vxlt_0_lowtown_a2_n_c_er4_0_0.vlm"},
	{0xC40CF817C58FF80AULL, "vxlt_0_lowtown_n_f66_0_2.vlm"},
	{0xC40CFE408FEBA007ULL, "vxw_woundedcoast_6l_0_0.vlm"},
	{0xC40D7A699B143135ULL, "follower_banter6163433.cl"},
	{0xC40D8F6138411CCCULL, "vxdr_primevaltomb_7ud_0_0.vlm"},
	{0xC40E1C03D57CC435ULL, "llac_w_woundedcoast_-131066_0.dds"},
	{0xC40EBF7AE4402BA3ULL, "6108142_f.fxe"},
	{0xC40F250C1B22BA2BULL, "vxlt_0_lowtown_docks_a3_n_d4g_0_0.vlm"},
	{0xC40F41812F3D6327ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_2_darren.ani"},
	{0xC40F9597752F7DADULL, "165079120.ogg"},
	{0xC4108F693AFF3BD7ULL, "1060851288.ogg"},
	{0xC411805E65700DF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_lean_forward.gad"},
	{0xC411E5C654306AF9ULL, "vxdr_0_deeproads_bwu_0_1.vlm"},
	{0xC412FC3B545D3093ULL, "mag101_aombanter6052605.cl"},
	{0xC412FC5F6B309C84ULL, "lt_0_lowtown_docks_n_155_0.msh"},
	{0xC4130E5F6B30BBEDULL, "lt_0_lowtown_docks_n_155_0.mao"},
	{0xC413125F6B30C226ULL, "lt_0_lowtown_docks_n_155_0.mmh"},
	{0xC4138142642CC1B2ULL, "llac_lt_0_lowtown_docks_0_0.dds"},
	{0xC4139F2103A6CCA0ULL, "6093826_m.fxe"},
	{0xC413A7B3D5960D44ULL, "hlac_ht_keep_4519_18y.dds"},
	{0xC413C04740A72620ULL, "202815990.ogg"},
	{0xC413C07A70CBB908ULL, "vxlt_hangedman_2uj_0_0.vlm"},
	{0xC413F3F0C15C0399ULL, "vxlt_0_lowtown_a2_n_c_eqf_0_1.vlm"},
	{0xC414168C26999C37ULL, "6026951_m.fxe"},
	{0xC4142B454C6A9750ULL, "mrl000_marethari6129456.cl"},
	{0xC4146131A16C5AF2ULL, "dae000_gamlen6139543.cl"},
	{0xC414821379F372E3ULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.dura.ani"},
	{0xC414B7D78E671BA3ULL, "art\\levels\\ztools\\default\\generic\\9ff_pile.mao"},
	{0xC414BBD78E6722F0ULL, "art\\levels\\ztools\\default\\generic\\9ff_pile.mmh"},
	{0xC414C6B24F213467ULL, "hlac_w_encounter2_camp_-131070_0z.dds"},
	{0xC414C9D78E673A0AULL, "art\\levels\\ztools\\default\\generic\\9ff_pile.msh"},
	{0xC415203158EFE5BBULL, "forcefeedback_plot.gda"},
	{0xC4157D718DFDA6EBULL, "588814797.ogg"},
	{0xC415AFCE03163B26ULL, "zz_and_debug6134130.cl"},
	{0xC415C29DE0BFF05AULL, "vxw_sundermount_2ir_0_0.vlm"},
	{0xC415ECBBD1334CC2ULL, "1038753234.ogg"},
	{0xC416029D71A09E3AULL, "vxht_0_hightown_atf_0_0.vlm"},
	{0xC4163C6DE8D9412DULL, "6151393_f.fxe"},
	{0xC416598B816D9F6FULL, "llac_lt_undercityserialkiller_-1_38.dds"},
	{0xC4165DA611E17430ULL, "vxlt_backalley_n_1wi_0_0.vlm"},
	{0xC416B50BEF256096ULL, "6086453_m.fxe"},
	{0xC416DA2A106BE53CULL, "al_mage_t02.gda"},
	{0xC41753617239B764ULL, "6160742_m.fxe"},
	{0xC4176B326B82336EULL, "vxlt_0_lowtown_docks_cta_0_0.vlm"},
	{0xC417915E65749647ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_lean_forward.ani"},
	{0xC417CE15CAF3B1CCULL, "gen00fl_anders6112972.cl"},
	{0xC4181A38DD7A181DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanl.ani"},
	{0xC41840AD1C63B2BBULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft_rockwraith.evt"},
	{0xC418AA662E341FC2ULL, "dae000_elegant6166932.cl"},
	{0xC419179DD8AE6FB0ULL, "vxw_blightlands_111_0_0.vlm"},
	{0xC4192121016C79BCULL, "vxht_0_hightown_90d_0_0.vlm"},
	{0xC4195103C8A6798AULL, "6074193_m.fxe"},
	{0xC419D9100E4EBE21ULL, "vxw_blightlands_p2_0_0.vlm"},
	{0xC41A1ACAC5976941ULL, "6039514_f.fxe"},
	{0xC41A4C92151C8451ULL, "922599845.ogg"},
	{0xC41AA17F834CB4DBULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.cess.ani"},
	{0xC41AD209C7510FCEULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_leaf.dds"},
	{0xC41B1586251CCD8BULL, "182859257.ogg"},
	{0xC41B416704DDC16CULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_013a_0s.dds"},
	{0xC41B4B27E63F7134ULL, "lt_0_lowtown_n_25_0_ll.mmh"},
	{0xC41C3BB4FA78E034ULL, "hlac_w_bonepitmining_a2_393222_0y.dds"},
	{0xC41C8C33755E422BULL, "genip_blocker_ogre_charge.utp"},
	{0xC41CB3DF6046728CULL, "832658817.ogg"},
	{0xC41D02199FA75814ULL, "vxht_chantry_32t_0_0.vlm"},
	{0xC41D095F6B3907F1ULL, "lt_0_lowtown_docks_n_155_0.phy"},
	{0xC41D53F52DCDFF4FULL, "seb201_lady_harimann6177405.cl"},
	{0xC41D96BED9267647ULL, "6125600_f.fxe"},
	{0xC41E39AD1C6812FCULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft_rockwraith.gad"},
	{0xC41F05FC0B50D2E6ULL, "zz_dae_debug6029882.cl"},
	{0xC41F406DCED09DE1ULL, "vxdr_primevaltomb_84e_0_0.vlm"},
	{0xC41F896289A3696EULL, "hlac_w_woundedcoast_196613_0y.dds"},
	{0xC41FD70A5CE49EA2ULL, "vis121_saemus6085695.cl"},
	{0xC41FE4B50933897BULL, "rdr111dg_martin6014677.cl"},
	{0xC4201955F2AD7B3CULL, "mag311tr_keran.utt"},
	{0xC42030AB7853103FULL, "28632480.ogg"},
	{0xC4204557124F0A50ULL, "vxw_bonepitmining_a2_26p_0_0.vlm"},
	{0xC4208E514BB22109ULL, "hlac_lt_0_lowtown_a3_n_c_131069_0z.dds"},
	{0xC421858BB748E3EFULL, "6098309_m.fxe"},
	{0xC4219A1379FE25DAULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.dura.evt"},
	{0xC421C88DC8D165FCULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sledge_d.dds"},
	{0xC42228F220D0DBC8ULL, "lla_ht_0_hightown_n_-65535_0.dds"},
	{0xC422A66EEA41600DULL, "follower_banter6164014.cl"},
	{0xC422B0064FC4A9F5ULL, "783155901.ogg"},
	{0xC422CC0099CC70C7ULL, "vxlt_0_lowtown_docks_b7y_0_1.vlm"},
	{0xC4236D72C6DBBAC8ULL, "103721077.ogg"},
	{0xC423912F6416F00EULL, "ts_prog_jade.gda"},
	{0xC424347AD2B608DAULL, "vxw_sundermount_2rs_0_0.vlm"},
	{0xC4244584D9D0EB3AULL, "6042545_m.fxe"},
	{0xC424A3283FE313AEULL, "vxht_0_hightown_a3_cf3_0_0.vlm"},
	{0xC4251939F129506FULL, "llac_lt_0_lowtown_docks_a3_-65538_8.dds"},
	{0xC42538AD1C6E2FFAULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\walkleft_rockwraith.ani"},
	{0xC4253DA63765B9B4ULL, "art\\characters\\playercharacter\\elffemale\\hlf_meda.phy"},
	{0xC42561FB609725D8ULL, "6072276_m.fxe"},
	{0xC42574FBBA8A99AFULL, "718721959.ogg"},
	{0xC425941A7A9A4FECULL, "35250092.ogg"},
	{0xC425C39480029670ULL, "6197230_m.fxe"},
	{0xC425D2504A0BB246ULL, "6092377_m.fxe"},
	{0xC425DD0F5D784BEEULL, "vxlt_0_lowtown_docks_cqb_0_0.vlm"},
	{0xC425FF091190D345ULL, "zz_dae_debug6049979.cl"},
	{0xC42684CD922CB199ULL, "ca_caverns1_probe_jze.dds"},
	{0xC4268B2CCB8B867BULL, "c_dog_tata.dds"},
	{0xC427C2E0BA9C920AULL, "6079293_m.fxe"},
	{0xC427F6FDB6BA0BAEULL, "vrc221_bartrand6043080.cl"},
	{0xC4284AED0044D6A1ULL, "vxw_blightlands_u1_0_0.vlm"},
	{0xC4284C80CB69BD9FULL, "lgt150ip_rand_2.utp"},
	{0xC4286FE1582F7EC8ULL, "6129849_f.fxe"},
	{0xC42877ED9A17A82CULL, "gen00fl_aveline6132788.cl"},
	{0xC42899137A0442CDULL, "art\\vfx\\cinematic\\vfx_ogre_slobber_c.dura.gad"},
	{0xC428A97F8358FFD2ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.cess.evt"},
	{0xC429277F860183ADULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratelg.phy"},
	{0xC4294520278816AAULL, "plt_cod_bks_kirkwall_hist_2.plo"},
	{0xC4294947958DF773ULL, "6141506_m.fxe"},
	{0xC42954F27B7ABFCAULL, "gen00fl_merrill6107000.cl"},
	{0xC4295896B94E20B1ULL, "10881818.ogg"},
	{0xC429C69E3421F3BCULL, "144423370.ogg"},
	{0xC42AC5BDFC1C278AULL, "hlac_w_encounter3_-196611_0x.dds"},
	{0xC42B6C122F0CEDFFULL, "vxlt_0_lowtown_a2_n_c_emu_0_0.vlm"},
	{0xC42B768B579E0C2CULL, "vxht_mansion2_2v5_0_0.vlm"},
	{0xC42BA1F29DA35D34ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgtb_0s.dds"},
	{0xC42BC62AF63600CAULL, "gen00fl_aveline6116656.cl"},
	{0xC42BEA84537EC1AEULL, "lgt311ar_gang_hideout.lst"},
	{0xC42BF974A4073BD9ULL, "957468587.ogg"},
	{0xC42C3A307A518222ULL, "hlac_w_bonepit_65540_0z.dds"},
	{0xC42C48B1D218B87EULL, "941627209.ogg"},
	{0xC42D2A2C2D52C945ULL, "vxht_brothel_n_53y_0_1.vlm"},
	{0xC42D5BDD5A40C259ULL, "dr_primevalentrance_1725.rml"},
	{0xC42D9FE5A8B31AA3ULL, "lt_0_lowtown_a2_32_0.anb"},
	{0xC42DC4B12CAB9A58ULL, "gen00fl_fenris6106676.cl"},
	{0xC42E099FBCBE60BAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_basic.phy"},
	{0xC42EEF356DBBB1C1ULL, "w_encounter2_camp_57_0.phy"},
	{0xC42F19BBAFB237E1ULL, "hlac_lt_0_lowtown_-262143_0y.dds"},
	{0xC42F1AA6376DD2B1ULL, "art\\characters\\playercharacter\\elffemale\\hlf_meda.msh"},
	{0xC42F28A6376DEACCULL, "art\\characters\\playercharacter\\elffemale\\hlf_meda.mao"},
	{0xC42F34A6376DFF6FULL, "art\\characters\\playercharacter\\elffemale\\hlf_meda.mmh"},
	{0xC42F8C990241FCD9ULL, "1035015305.ogg"},
	{0xC42F982E429E4270ULL, "w_sundermount_149_0.mao"},
	{0xC42F9C2E429E4883ULL, "w_sundermount_149_0.mmh"},
	{0xC42FA4050878417FULL, "ht_0_hightown_water_15346.msh"},
	{0xC42FAA0508784B21ULL, "ht_0_hightown_water_15346.mmh"},
	{0xC42FAA2E429E60DDULL, "w_sundermount_149_0.msh"},
	{0xC42FB60508785FC2ULL, "ht_0_hightown_water_15346.mao"},
	{0xC42FB7B61C272260ULL, "15624909.ogg"},
	{0xC42FC87F835F5365ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.cess.gad"},
	{0xC430055320E72A19ULL, "dae000_pickpocket_scene6175074.cl"},
	{0xC430FBB34AE36AE5ULL, "vxw_blightlands_nl_0_0.vlm"},
	{0xC4310C51DA2430B9ULL, "lt_0_lowtown_a2_n_c_8_0.phy"},
	{0xC431663401D05788ULL, "195461114.ogg"},
	{0xC431A0D0C6C25300ULL, "sib111_mother6098826.cl"},
	{0xC43230C29BB33862ULL, "hlac_ca_caverns2_5802_1z.dds"},
	{0xC4323AC64A462CCFULL, "dae311ip_at_lowtown.utp"},
	{0xC432A657C4AC215AULL, "vxw_sundermount_2wl_0_0.vlm"},
	{0xC4332D6CDA96CB4EULL, "6118093_m.fxe"},
	{0xC433B2D93ADE6233ULL, "lowtown_male_cityguard_mob6150713.cl"},
	{0xC434967E4F96F002ULL, "seb221_random_servants6187835.cl"},
	{0xC434A9D3BF1643F9ULL, "llac_lt_undercitydungeon_-1_44.dds"},
	{0xC434F8356DC02CDDULL, "w_encounter2_camp_57_0.rml"},
	{0xC4353E79C024983DULL, "6125551_m.fxe"},
	{0xC435A2F40D7B5880ULL, "vxlt_backalley_n_1y3_0_0.vlm"},
	{0xC435DBF75BFB77DEULL, "and110_refugee6111608.cl"},
	{0xC435EAECB9DA4F51ULL, "215608586.ogg"},
	{0xC43609F21F82AEBFULL, "6136214_m.fxe"},
	{0xC43614D8B29997CAULL, "vxca_cavern1_haunted_fm7_0_0.vlm"},
	{0xC43658E40969F3EAULL, "vxht_mansion_n_6hr_0_0.vlm"},
	{0xC4367DA3FA5B3A42ULL, "6066280_m.fxe"},
	{0xC4372358734A6DC4ULL, "one100dg_nabil6049136.cl"},
	{0xC437802746516F4FULL, "vxlt_0_lowtown_docks_a3_bfh_0_0.vlm"},
	{0xC437B6B206D70791ULL, "mrl210_merrill6102138.cl"},
	{0xC437BB09617758DCULL, "380501216.ogg"},
	{0xC437D0322A31CBEFULL, "6104170_m.fxe"},
	{0xC438201DC83C4E14ULL, "zz_followers_banter6159600.cl"},
	{0xC438380B40F51B1EULL, "follower_banter6141858.cl"},
	{0xC4383D014FC51F83ULL, "664126934.ogg"},
	{0xC43888DD09B29ABFULL, "vxht_0_hightown_a3_c31_0_0.vlm"},
	{0xC438EA389DAFBC84ULL, "vxw_sundermount_23n_0_0.vlm"},
	{0xC439138D137F5BFCULL, "6154110_m.fxe"},
	{0xC4397FD2F9FFD4C3ULL, "vxht_0_hightown_a2_bv2_0_0.vlm"},
	{0xC43AA3FFDED2380DULL, "6172757_m.fxe"},
	{0xC43AE45CC6E0BDE5ULL, "craterr_halff.dds"},
	{0xC43B24EBA982B532ULL, "dre000_varric6029462.cl"},
	{0xC43B8D29ABB0D7A1ULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef_lashesm1.mao"},
	{0xC43C1EFC58B37509ULL, "341309643.ogg"},
	{0xC43C3913537EB042ULL, "6217927_m.fxe"},
	{0xC43C5D69153BA931ULL, "381592681.ogg"},
	{0xC43C7A89A92E63F5ULL, "296077343.ogg"},
	{0xC43CEBC6560808A2ULL, "art\\characters\\creatures\\ragedemon\\animation\\exit.gad"},
	{0xC43D200E046A4716ULL, "mrl350ip_grave_a.utp"},
	{0xC43D30E1661F82C4ULL, "gen00fl_anders6119502.cl"},
	{0xC43D7129D42B0C50ULL, "vxlt_0_lowtown_a2_n_ff2_0_0.vlm"},
	{0xC43E17BCAB0F9766ULL, "dae000_sketch6179004.cl"},
	{0xC43E2EE6FD26BE77ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mede.mmh"},
	{0xC43E30A11099F6EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext_override.gad"},
	{0xC43E44E6FD26E3E9ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_mede.msh"},
	{0xC43E557DA595E373ULL, "hlac_lt_gallowstemplar_2390_0y.dds"},
	{0xC43EAF20621E9738ULL, "pro000_colonel_ewald6039967.cl"},
	{0xC43ECED99A2BA28DULL, "6208638_m.fxe"},
	{0xC43EEB751A32CBF2ULL, "qcr220_grey_wardens6135958.cl"},
	{0xC43F6D903CD98365ULL, "vxw_blightlands_sk_0_0.vlm"},
	{0xC43F9C1DEF7C0D8EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupa.mmh"},
	{0xC43FA81DEF7C2225ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupa.mao"},
	{0xC43FB3A211000284ULL, "hm_lgt_urn_barker.mor"},
	{0xC43FB61DEF7C3A7CULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockgroupa.msh"},
	{0xC43FC0C75085E3C2ULL, "art\\vfx\\materials\\decal\\vfx_projdecal_blob_shadow.mao"},
	{0xC4401B3A6D09573FULL, "mer320_hubert6090240.cl"},
	{0xC4409D2E42ACBDA8ULL, "w_sundermount_149_0.phy"},
	{0xC440E840909C5586ULL, "mag121_viveka.cnv"},
	{0xC4411834B6A239DAULL, "vxw_sundermount_2xm_0_0.vlm"},
	{0xC4415EC825227DECULL, "6084004_m.fxe"},
	{0xC441834DFEC350D0ULL, "864784832.ogg"},
	{0xC441B33CE7C3BDB0ULL, "vxlt_backalley_n_1xt_0_0.vlm"},
	{0xC441E48A01D3B309ULL, "269766892.ogg"},
	{0xC44221F1DEA0C8F7ULL, "hla_ht_0_hightown_a2_-1_0y.dds"},
	{0xC44256A817811DC8ULL, "mrl000_marethari6205684.cl"},
	{0xC4425877A16705EDULL, "hlac_lt_0_lowtown_a3_n_c_-2_4y.dds"},
	{0xC442F2A58B0048E5ULL, "553862747.ogg"},
	{0xC44336B8DCFCDB53ULL, "6077583_f.fxe"},
	{0xC44371B7CF6FC134ULL, "6016039_f.fxe"},
	{0xC4437DA79956E580ULL, "one140cr_charade6187708.cl"},
	{0xC443B056B21AB0BDULL, "6122451_m.fxe"},
	{0xC443B7CC81E43B19ULL, "follower_banter6136325.cl"},
	{0xC443EA40909E8F46ULL, "mag121_viveka.bnk"},
	{0xC443EFD88FF741A2ULL, "art\\vfx\\combat\\vfx_r_brd_shout_act_c.impa.gad"},
	{0xC44493881212367BULL, "lt_0_lowtown_docks_a3_200_0.mmh"},
	{0xC4449F8812124A18ULL, "lt_0_lowtown_docks_a3_200_0.mao"},
	{0xC444B18812126965ULL, "lt_0_lowtown_docks_a3_200_0.msh"},
	{0xC44513A1109FE381ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext_override.evt"},
	{0xC4455406BF681858ULL, "vxht_0_hightown_n_c78_0_0.vlm"},
	{0xC44593CFF2CF3D97ULL, "gen_im_trs_qun_sta_cld_r2.uti"},
	{0xC446029CC3B226AEULL, "749155597.ogg"},
	{0xC4462A963873DB9CULL, "docks_fem_sailor_mob6176429.cl"},
	{0xC446FAB9FBA8B33FULL, "vxht_0_hightown_a3_c40_0_0.vlm"},
	{0xC447034B497CD025ULL, "dae000_meredith6166306.cl"},
	{0xC4473E73154E74FEULL, "6116716_m.fxe"},
	{0xC4475C158FA5D504ULL, "vxw_sundermount_24o_0_0.vlm"},
	{0xC4478AD34B8938C3ULL, "lowtown_argument6169414.cl"},
	{0xC44803A84F4C24EBULL, "391642525.ogg"},
	{0xC4482BA347138ECAULL, "6169978_m.fxe"},
	{0xC448514A1132B6C5ULL, "237529137.ogg"},
	{0xC448CE57991F910FULL, "vxlt_0_lowtown_a2_n_c_ep2_0_0.vlm"},
	{0xC448D0C8EED710BBULL, "6175839_m.fxe"},
	{0xC448F4147EB803E6ULL, "6219235_m.fxe"},
	{0xC449BAC744717E8BULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.cess.evt"},
	{0xC44A344F625143F6ULL, "vxw_bonepit_8r_0_0.vlm"},
	{0xC44AB4528B17E392ULL, "gen00fl_bethany6127147.cl"},
	{0xC44B12D88FFD9B99ULL, "art\\vfx\\combat\\vfx_r_brd_shout_act_c.impa.evt"},
	{0xC44B43DF0775F0FBULL, "hlac_w_encounter2_262143_0y.dds"},
	{0xC44B57AF54599990ULL, "dev_res_cyc.mao"},
	{0xC44B9B49D8AC5B98ULL, "llac_lt_gallowstemplar_7198_1.dds"},
	{0xC44C9A5BB1FBB08BULL, "275906358.ogg"},
	{0xC44CC0C69A20B746ULL, "genip_jug.utp"},
	{0xC44CE65B7FA8B0F0ULL, "6043100_m.fxe"},
	{0xC44D5993F4D4968BULL, "279169061.ogg"},
	{0xC44DDF6D2ECF9BE5ULL, "vxw_blightlands_tj_0_0.vlm"},
	{0xC44E38B47216A86DULL, "6076937_f.fxe"},
	{0xC44E6C46938EE1E5ULL, "mcr310_argument6072310.cl"},
	{0xC44EDE8664A8A82CULL, "art\\levels\\races\\elf\\props\\_textures\\prp_fire_n.dds"},
	{0xC44F01AD41115BE5ULL, "vxw_sundermount_2gw_0_0.vlm"},
	{0xC44F8FA204B2F73AULL, "hlac_lt_0_lowtown_a3_n_c_-131071_0y.dds"},
	{0xC44FB2D3523D490CULL, "hlac_lt_0_lowtown_a3_n_c_131070_5z.dds"},
	{0xC45010CEA0F404F5ULL, "939747017.ogg"},
	{0xC4505A269DA9467DULL, "6094901_m.fxe"},
	{0xC4506764DE392182ULL, "zz_one_debug6099955.cl"},
	{0xC450D3C74477C84CULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.cess.gad"},
	{0xC45116C65618FA88ULL, "art\\characters\\creatures\\ragedemon\\animation\\exit.ani"},
	{0xC4515FD5D6B4FFCEULL, "301138307.ogg"},
	{0xC451816327B73BB2ULL, "6051274_f.fxe"},
	{0xC4518E8F53CC23F0ULL, "pro000_merchant.cnv"},
	{0xC451A36A03D633AFULL, "ss_dalish_f_1.wbd"},
	{0xC4521A052133E9CAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winspike.mao"},
	{0xC4521E052133F019ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winspike.mmh"},
	{0xC452280521340107ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winspike.msh"},
	{0xC4528F84FB84A289ULL, "vxht_0_hightown_n_b96_0_0.vlm"},
	{0xC4530BA110AC13D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ext_override.ani"},
	{0xC45333C1C32F7564ULL, "6172590_m.fxe"},
	{0xC4538A5771D7CD32ULL, "mag311_keran6041588.cl"},
	{0xC45395D500FAD040ULL, "6128009_m.fxe"},
	{0xC4539D791D0D9EA4ULL, "155649990.ogg"},
	{0xC453B28F85DA4DF6ULL, "vxlt_0_lowtown_a3_n_c_eok_0_0.vlm"},
	{0xC453D1963924F85DULL, "dae000ip_crowd_ff8_talk15.utp"},
	{0xC453D38DFC2C0337ULL, "6030664_m.fxe"},
	{0xC4544453B437B0F9ULL, "466153550.ogg"},
	{0xC45453566D1E410FULL, "vxlt_0_lowtown_a2_n_c_f4u_0_0.vlm"},
	{0xC454B08F53CE9348ULL, "pro000_merchant.bnk"},
	{0xC455052D22BE63A9ULL, "145512979.ogg"},
	{0xC455574A737A546EULL, "6096746_m.fxe"},
	{0xC455596B8E38CA1CULL, "vxlt_0_lowtown_fgz_0_0.vlm"},
	{0xC455C2D78E9EC50BULL, "art\\levels\\ztools\\default\\generic\\9ff_pile.phy"},
	{0xC455C54C3A6ABA1CULL, "hlac_lt_0_lowtown_a2_n_-196609_1y.dds"},
	{0xC455CDF2819BED84ULL, "vxw_sundermount_29l_0_0.vlm"},
	{0xC4564CB0900A4BDBULL, "6121548_m.fxe"},
	{0xC4568FB15591B91DULL, "vxdr_0_deeproads_crz_0_0.vlm"},
	{0xC457106A39E42C1BULL, "vxca_cavern1_haunted_jgp_0_0.vlm"},
	{0xC4577A21D338BAB0ULL, "6201254_m.fxe"},
	{0xC457C428257BEC94ULL, "genip_cup.utp"},
	{0xC457D2C7447DE54AULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.cess.ani"},
	{0xC457E723D1FC37B0ULL, "6135704_m.fxe"},
	{0xC4581AD890083388ULL, "art\\vfx\\combat\\vfx_r_brd_shout_act_c.impa.ani"},
	{0xC458E48BE15FC559ULL, "6074338_m.fxe"},
	{0xC4591133D58526A8ULL, "6000714_m.fxe"},
	{0xC4598D95D89D0DFCULL, "hlac_w_encounter3_-262141_0x.dds"},
	{0xC4599CA8E4256896ULL, "w_bonepitdestroyed_a3_47_0.rml"},
	{0xC45A8C27FE33F983ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha05.mmh"},
	{0xC45A9A27FE3411DDULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha05.msh"},
	{0xC45AFB0B1F8EAE64ULL, "vxlt_house_es_0_0.vlm"},
	{0xC45B30D6D91CBC30ULL, "vxca_caverns1_dq6_0_0.vlm"},
	{0xC45BB058351F1C86ULL, "vxht_mansion_n_57m_0_1.vlm"},
	{0xC45BCF20D88AC000ULL, "400136653.ogg"},
	{0xC45C31206CE38E66ULL, "6063849_m.fxe"},
	{0xC45C784200A11E75ULL, "vxht_brothel_n_3of_0_0.vlm"},
	{0xC45CCDD03DC7E325ULL, "6030693_m.fxe"},
	{0xC45CF8B32D48244EULL, "hla_ht_0_hightown_a2_-131071_0x.dds"},
	{0xC45D5353F4C9D9C0ULL, "vfx_rendertoshadowmap_p0.vsh.fxo.dx11.assembly"},
	{0xC45D5BF52774EFBBULL, "6101241_m.fxe"},
	{0xC45D738A33077465ULL, "vxw_sundermount_2hv_0_0.vlm"},
	{0xC45EB7C638CF8477ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_sml.mmh"},
	{0xC45F3548EDFEF28DULL, "hlac_lt_0_lowtown_docks_n_65537_0z.dds"},
	{0xC45F4237381DCBA3ULL, "6083953_f.fxe"},
	{0xC45F69265CF58FD2ULL, "docks_male_human_worker_sta6176006.cl"},
	{0xC45F8922E6D1B71EULL, "vxw_sundermount_2o0_0_0.vlm"},
	{0xC45FE3E5316748B3ULL, "vxlt_0_lowtown_a2_n_c_ejh_0_0.vlm"},
	{0xC460084E5FC5208EULL, "6089650_m.fxe"},
	{0xC4609E57B1358A18ULL, "460644255.ogg"},
	{0xC460AC4C30239222ULL, "mag371_elsa6171595.cl"},
	{0xC460B74F78F9A1F4ULL, "mer221_sabin6044759.cl"},
	{0xC4611B0E1A5B19BEULL, "mcr310_argument6127664.cl"},
	{0xC46143EDE82881A5ULL, "hlac_w_encounter2_camp_327683_0x.dds"},
	{0xC461A95163CF9969ULL, "151964429.ogg"},
	{0xC461E6AEBBC06E54ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.cess.ani"},
	{0xC46237C0A3544958ULL, "vxht_0_hightown_n_c96_0_0.vlm"},
	{0xC462AC69171B66E0ULL, "mainmenu.gfx"},
	{0xC462F1EDBB11ACD2ULL, "6136000_m.fxe"},
	{0xC46329988340B6E4ULL, "6089818_m.fxe"},
	{0xC463A385E1F84AE8ULL, "hlac_lt_undercitydungeon_-1_44y.dds"},
	{0xC464031F4677C72EULL, "170308115.ogg"},
	{0xC464B80E5F170C6CULL, "6110051_m.fxe"},
	{0xC464BCF387A216CBULL, "vxlt_gallowstemplar_n_34m_0_0.vlm"},
	{0xC465018E4787D19DULL, "vxdr_0_deeproads_coy_0_0.vlm"},
	{0xC4651D28C15569F6ULL, "w_bonepitmining_a2_19_0_lc.mmh"},
	{0xC4653385F6E54FFFULL, "vxlt_0_lowtown_a2_n_c_eqq_0_0.vlm"},
	{0xC46554B4C46100ECULL, "hlac_lt_darktown_-1_17x.dds"},
	{0xC4655DA9E1AD9AA1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_bda_0s.dds"},
	{0xC465877B2704A4F0ULL, "386113550.ogg"},
	{0xC465A20BCDFDC52FULL, "gen00fl_aveline6115496.cl"},
	{0xC465EE4C7554052FULL, "vxlt_0_lowtown_a2_n_c_hoz_0_0.vlm"},
	{0xC4665F6D0C2DBE1CULL, "vxlt_gallowstemplar_n_3r8_0_0.vlm"},
	{0xC4675668D355DDD9ULL, "6079038_m.fxe"},
	{0xC4675CD3A167C0FBULL, "vxht_mansion2_3bg_0_0.vlm"},
	{0xC4678C06D615B05AULL, "6131881_m.fxe"},
	{0xC467A25564CF687FULL, "art\\characters\\creatures\\hurlock\\animation\\blendtree\\hrlk_combat_movement_sw.anb"},
	{0xC467CA48C386E0D4ULL, "6000888_m.fxe"},
	{0xC4688A28F48BCEBDULL, "vxlt_0_lowtown_docks_a3_cyn_0_0.vlm"},
	{0xC468E3AEBBC687B6ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.cess.gad"},
	{0xC4694C2B151A6DCEULL, "vxht_0_hightown_n_c3s_0_1.vlm"},
	{0xC469563CE9E29D0AULL, "zz_followers_banter6157604.cl"},
	{0xC469856EB2FABA41ULL, "6116733_m.fxe"},
	{0xC469A081A7369E88ULL, "6085472_m.fxe"},
	{0xC469D39DE6A1E0FAULL, "vxht_brothel_n_5hb_0_0.vlm"},
	{0xC46A05924E1691F1ULL, "vxdr_primevaltomb_73s_0_0.vlm"},
	{0xC46A19447FC53D64ULL, "499337791.ogg"},
	{0xC46A1F759416FA40ULL, "character_character_p108.psh.fxo.dx11"},
	{0xC46A2AFFD7A5081CULL, "narrative_varric6172479.cl"},
	{0xC46A58006840676AULL, "one280_emrys6187702.cl"},
	{0xC46A5CAC773BDBCFULL, "dae000ip_crowd_mf6_talk1.utp"},
	{0xC46AE304294B343EULL, "llac_w_sundermount_327678_0.dds"},
	{0xC46B7A397870E5EEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_codexint_d.dds"},
	{0xC46B8D27FE426EA8ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha05.phy"},
	{0xC46B9A1889F4067EULL, "vxht_keep_67t_0_0.vlm"},
	{0xC46BD4026CEE3377ULL, "6100259_m.fxe"},
	{0xC46BE1E415AA8E26ULL, "708999246.ogg"},
	{0xC46CCAD66FF0E250ULL, "one110_meeran6089471.cl"},
	{0xC46CEC0151479A2FULL, "vxht_chantry_357_0_1.vlm"},
	{0xC46CF789144DEB66ULL, "484871789.ogg"},
	{0xC46CF78F490A308BULL, "6096318_m.fxe"},
	{0xC46D537F395D3836ULL, "6088060_m.fxe"},
	{0xC46D836AD0BEC4A0ULL, "6157087_m.fxe"},
	{0xC46DFAFFD8C7CF9EULL, "vxw_sundermount_2p1_0_0.vlm"},
	{0xC46E098BBEEFB316ULL, "6052147.cut"},
	{0xC46E54EACC1FC601ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_big.ani"},
	{0xC46E55C2235D6133ULL, "vxlt_0_lowtown_a2_n_c_eok_0_0.vlm"},
	{0xC46E76C6C3276ACCULL, "gen00fl_merrill6115962.cl"},
	{0xC46EFEAEBBCB2115ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.cess.evt"},
	{0xC46F0078A6EACACAULL, "633005179.ogg"},
	{0xC470182FDAF7A6F7ULL, "624055492.ogg"},
	{0xC4709A4D8F863501ULL, "gen00fl_aveline6140140.cl"},
	{0xC4714A306EB0083CULL, "vxht_playershouse_n_205_0_0.vlm"},
	{0xC4714E04DD917C48ULL, "6097075_m.fxe"},
	{0xC471C6CA2BE39020ULL, "787954507.ogg"},
	{0xC472071C4D1A55B3ULL, "6132496_m.fxe"},
	{0xC4723827D422CDEAULL, "858370638.ogg"},
	{0xC472A7AE2A80AF3EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree02_n.dds"},
	{0xC4734746C1D836C3ULL, "vxht_0_hightown_a2_bj6_0_0.vlm"},
	{0xC47356B8D2621370ULL, "vxw_woundedcoast_17i_0_0.vlm"},
	{0xC47358CE21C42EF5ULL, "mag121_idunna6092360.cl"},
	{0xC4739CCF94A17828ULL, "231092939.ogg"},
	{0xC4745FEACC2444A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_big.gad"},
	{0xC474E7E527B71B40ULL, "612559240.ogg"},
	{0xC474ED7E7203B1FAULL, "vxlt_0_lowtown_docks_b4t_0_0.vlm"},
	{0xC474F10521BE8A30ULL, "6060543_m.fxe"},
	{0xC475CE5AA1CC1B42ULL, "gen_im_npc_wep_dua_aria.uti"},
	{0xC475D31580348659ULL, "184616269.ogg"},
	{0xC475ED427CFEBE69ULL, "vxdr_primevaltomb_3i5_0_0.vlm"},
	{0xC47626BA85E42231ULL, "vxdr_primevaltomb_6j4_0_1.vlm"},
	{0xC4762C5C37BB1630ULL, "qun120cr_javaris6042982.cl"},
	{0xC47637286B791F4EULL, "vxlt_0_lowtown_a2_n_c_eup_0_0.vlm"},
	{0xC47693D6B2D2A796ULL, "766629749.ogg"},
	{0xC476ABCC03B8D8CCULL, "ts_templars_wpn_props.gda"},
	{0xC477D02D37AE84FDULL, "6068480_f.fxe"},
	{0xC477EDC4AD1A265DULL, "vxht_mansion2_1xp_0_1.vlm"},
	{0xC4780D87109743A9ULL, "dae000_player_generic_find6179059.cl"},
	{0xC4782AA23FA54F31ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_aoe_p.anb"},
	{0xC47831CF726C7B56ULL, "224754895.ogg"},
	{0xC4799576F0C6DEC1ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_ht_lightstreaks.mmh"},
	{0xC47A0BF57BEA1EFEULL, "vxht_keep_60u_0_0.vlm"},
	{0xC47A6436144065B6ULL, "6096851_f.fxe"},
	{0xC47AD9B900F88F54ULL, "739129997.ogg"},
	{0xC47AF4194AB09EA5ULL, "vxlt_0_lowtown_a3_n_c_ef7_0_0.vlm"},
	{0xC47B0EC93124C43CULL, "dae011_wench_mob6149861.cl"},
	{0xC47B205727855E53ULL, "llac_lt_0_lowtown_a2_n_c_-131073_3.dds"},
	{0xC47B3CEACC2A2788ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_cough_big.evt"},
	{0xC47B83103A28E48AULL, "280361004.ogg"},
	{0xC47B9D4843CB52C9ULL, "seb221_flora_harimann6177706.cl"},
	{0xC47BC388A3416AACULL, "6132990_m.fxe"},
	{0xC47BC55C2B5350B6ULL, "6083160_m.fxe"},
	{0xC47C1055F2FBCAD9ULL, "mag311tr_keran.ncs"},
	{0xC47C286AD7724AA1ULL, "6055167_f.fxe"},
	{0xC47C55EE443DDEB0ULL, "hm_gen_refugee1.mor"},
	{0xC47C6CDCCABDE81EULL, "vxw_sundermount_2u2_0_0.vlm"},
	{0xC47CA5C47E339BAFULL, "206911217.ogg"},
	{0xC47D783CAF06B664ULL, "zz_approval_debug6003975.cl"},
	{0xC47DC8A9FCA8EBFDULL, "6083457_m.fxe"},
	{0xC47DCD89DE9EEF4AULL, "6142202_f.fxe"},
	{0xC47DD533AFF2FD06ULL, "6036476_m.fxe"},
	{0xC47E08908C7069F5ULL, "mag170_hugh6094565.cl"},
	{0xC47E2F851E7C07D8ULL, "228748337.ogg"},
	{0xC47E5A2A627E1F10ULL, "6062088_m.fxe"},
	{0xC47E614E22C0B825ULL, "dre000_dougal6131200.cl"},
	{0xC47EBA14F121D0E2ULL, "60303734.ogg"},
	{0xC47F08265BBC9776ULL, "vxlt_0_lowtown_a3_n_c_exn_0_0.vlm"},
	{0xC47F14F322128EC7ULL, "vxca_caverns1_mines_eh2_0_0.vlm"},
	{0xC47F80D1A25F9D93ULL, "dre000_dougal6129755.cl"},
	{0xC47F94AF83769AADULL, "vxlt_0_lowtown_n_h25_0_0.vlm"},
	{0xC47FC3764B29A210ULL, "6055594_m.fxe"},
	{0xC47FDADEE73C0957ULL, "513242859.ogg"},
	{0xC47FEFA6D0D1A44EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutan.msh"},
	{0xC47FFDA6D0D1BB9FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutan.mao"},
	{0xC48009A6D0D1D03CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarpoutan.mmh"},
	{0xC480F2004992D397ULL, "129572894.ogg"},
	{0xC4816B0CA973DC42ULL, "dre000_hireling16038602.cl"},
	{0xC481BD6C06B51176ULL, "seb221_flora_harimann6177174.cl"},
	{0xC481C895C4582BF0ULL, "vxw_woundedcoast_12v_0_0.vlm"},
	{0xC481E5482B74029DULL, "vxdr_0_deeproads_caw_0_0.vlm"},
	{0xC48237CC9399DBCEULL, "vrc221_hugin6042804.cl"},
	{0xC4825E60901B7A6DULL, "6136814_m.fxe"},
	{0xC4825F0E8A786ADFULL, "propalpha1_dy.psh.fxo"},
	{0xC4827907ACF6B8B3ULL, "150227134.ogg"},
	{0xC48296669EDB0A37ULL, "vxlt_warehouse_145_0_0.vlm"},
	{0xC482F0A10F0DE828ULL, "gen00fl_fenris6126487.cl"},
	{0xC4830E594FFD405FULL, "vxht_0_hightown_av5_0_0.vlm"},
	{0xC4834DF7BC42F38EULL, "vxca_cave_2bg_0_0.vlm"},
	{0xC48372DF61AE2E76ULL, "6091387_m.fxe"},
	{0xC4838A3FC400E47EULL, "6074997_m.fxe"},
	{0xC483B028EF274F1AULL, "pro000_colonel_ewald6039976.cl"},
	{0xC483B3E903DDDA14ULL, "pro000_merchant6051569.cl"},
	{0xC483E0BC9C9D71C6ULL, "167018951.ogg"},
	{0xC483E9D3099E4FE0ULL, "vxht_bartrand_n_5y7_0_0.vlm"},
	{0xC4842525D9393299ULL, "hlac_lt_gallowstemplar_2390_1x.dds"},
	{0xC484BE365EEED0DCULL, "6179260_f.fxe"},
	{0xC484C6CF81D76289ULL, "vxlt_0_lowtown_ewd_0_0.vlm"},
	{0xC4854BA23FB011AAULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_aoe_p.mmh"},
	{0xC485DFCDC14BB6F7ULL, "vxht_0_hightown_b8h_0_0.vlm"},
	{0xC485FDCCA003B4E3ULL, "vxlt_0_lowtown_a3_n_c_fe7_0_0.vlm"},
	{0xC486648BBDE3B7B7ULL, "vxdr_0_deeproads_egq_0_0.vlm"},
	{0xC4866B82DA831AB0ULL, "6117667_m.fxe"},
	{0xC48701A0E1F06104ULL, "llac_ht_brothel_n_3369_0.dds"},
	{0xC4872D5E437FC79FULL, "vxht_0_hightown_a3_bkh_0_0.vlm"},
	{0xC4874CAE4389EE99ULL, "aip_spiderpoison_base.gda"},
	{0xC4878DCC4653EFE7ULL, "338032664.ogg"},
	{0xC487B7A5654B4BECULL, "gen00fl_anders6125909.cl"},
	{0xC4880AFC5DB0A378ULL, "581566160.ogg"},
	{0xC488104930866D7FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0cc_basic.msh"},
	{0xC488164930867721ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0cc_basic.mmh"},
	{0xC488224930868BC2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0cc_basic.mao"},
	{0xC4884DE4C1F73B34ULL, "t3_wdg_w27.tnt"},
	{0xC48851BED3BDF836ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_0x.dds"},
	{0xC488A853482EC940ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masa.phy"},
	{0xC48965F63CA6B725ULL, "vxlt_0_lowtown_a3_n_c_ei6_0_0.vlm"},
	{0xC4897477D860B999ULL, "23529703.ogg"},
	{0xC489EC0D59B937CAULL, "vxlt_0_lowtown_h8s_0_0.vlm"},
	{0xC48A029FBD0CB3E3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_basic.msh"},
	{0xC48A109FBD0CCB3DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_basic.mmh"},
	{0xC48A149FBD0CD24EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1wo_basic.mao"},
	{0xC48AC6BC183A8B75ULL, "vxlt_backalley_g_1dx_0_0.vlm"},
	{0xC48AFCE626B7126DULL, "6133453_m.fxe"},
	{0xC48AFE2AED67D688ULL, "vxw_encounter2_camp_r6_0_0.vlm"},
	{0xC48C05C145D62CDEULL, "52168178.ogg"},
	{0xC48C30861DC72917ULL, "hlac_lt_0_lowtown_n_-65535_7y.dds"},
	{0xC48C5E875148D610ULL, "vxht_brothel_n_4x1_0_0.vlm"},
	{0xC48CB3DCE3A0FC1EULL, "dae000_surprises6157192.cl"},
	{0xC48D07CDB92F0713ULL, "hlac_dr_0_deeproads_5849_9x.dds"},
	{0xC48D51964F468157ULL, "228892541.ogg"},
	{0xC48D947168F55B8CULL, "231731638.ogg"},
	{0xC48DF5DAF10F2764ULL, "383894089.ogg"},
	{0xC490AC01F61EA48FULL, "1000597603.ogg"},
	{0xC490BBE1DBB5ED15ULL, "follower_banter6171063.cl"},
	{0xC490D2D86EB64167ULL, "vxw_bonepit_2a8_0_0.vlm"},
	{0xC490F4356E0E84FDULL, "w_encounter2_camp_57_0.mao"},
	{0xC490F8356E0E8B36ULL, "w_encounter2_camp_57_0.mmh"},
	{0xC49102356E0E9C34ULL, "w_encounter2_camp_57_0.msh"},
	{0xC4910DF811202614ULL, "105065272.ogg"},
	{0xC49180745BE7307DULL, "6059968_m.fxe"},
	{0xC49180B1BCAC945EULL, "6094018_m.fxe"},
	{0xC49181B4B81B3737ULL, "gen_im_flw_sta_anders_01.uti"},
	{0xC491837BD03B5024ULL, "dae000ip_crowd_lm5_talk9.utp"},
	{0xC4919DE9B097ADDBULL, "art\\characters\\playercharacter\\humanmale\\hlf_masg.msh"},
	{0xC491AFE9B097CC26ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masg.mao"},
	{0xC491BBE9B097E0C5ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masg.mmh"},
	{0xC491CD6660D67D04ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga104a_0n.dds"},
	{0xC4921BB301FFD566ULL, "dae221st_dog_player.stg"},
	{0xC4925531CE94954CULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_3z.dds"},
	{0xC492BD6A040DF685ULL, "ss_dalish_f_1.bnk"},
	{0xC4930A51415C7BE7ULL, "hlac_ca_cavern1_haunted_24427_5y.dds"},
	{0xC49338AC73CD7B09ULL, "vxlt_0_lowtown_erk_0_0.vlm"},
	{0xC493BA91FC939907ULL, "t1_skn_h07.tnt"},
	{0xC49406156875A6EEULL, "hlac_lt_0_lowtown_a2_n_c_-65537_5x.dds"},
	{0xC494095B5FA5B2BBULL, "366837363.ogg"},
	{0xC494BD9D56475B91ULL, "vxw_sundermount_32q_0_0.vlm"},
	{0xC494CB41EB167045ULL, "vxht_brothel_n_5hi_0_0.vlm"},
	{0xC494D17E9106575DULL, "vxht_mansion2_1fn_0_1.vlm"},
	{0xC494DD5FCC793330ULL, "6112567_m.fxe"},
	{0xC4960F9271ADACD0ULL, "6202026_m.fxe"},
	{0xC496C99F7D946698ULL, "dae000ip_crowd_mf7_talk14.utp"},
	{0xC4976FF0C1C0EF9EULL, "hlac_ht_keep_4519_11x.dds"},
	{0xC497718C45E13F0EULL, "6075969_m.fxe"},
	{0xC497759F996C4EABULL, "amb_scene_dwarfmech.wbd"},
	{0xC497B2A4DEF147C2ULL, "qun231_party_comment6182831.cl"},
	{0xC497C93C6D973C1BULL, "nat320_delilah6065938.cl"},
	{0xC498168C3C8C6D5EULL, "211264738.ogg"},
	{0xC49865674234F1A5ULL, "vxht_keep_n_3ic_0_0.vlm"},
	{0xC49865CDB5C036DFULL, "814112220.ogg"},
	{0xC4988D82F057AC47ULL, "6047473_m.fxe"},
	{0xC498C87801247313ULL, "6139974_m.fxe"},
	{0xC499329ED62B240AULL, "vxdr_0_deeproads_bhg_0_1.vlm"},
	{0xC4997007DF5DEF08ULL, "vxw_encounter2_camp_m7_0_0.vlm"},
	{0xC49971CF673B8B76ULL, "vxlt_0_lowtown_gjr_0_0.vlm"},
	{0xC499945A47D6C9D0ULL, "zz_mcr_debug6091898.cl"},
	{0xC499BDE3A4B8433FULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_rack.msh"},
	{0xC499C3E3A4B84DE1ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_rack.mmh"},
	{0xC499CFCFC8FE57A4ULL, "6017439_m.fxe"},
	{0xC499CFE3A4B86182ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_rack.mao"},
	{0xC499E48614BCD9E1ULL, "vxht_brothel_n_5c4_0_0.vlm"},
	{0xC49A0DFB70720F50ULL, "zz_vispt_debug.ncs"},
	{0xC49A456BDC480D49ULL, "6035373_m.fxe"},
	{0xC49A9004E3D47827ULL, "6083077_m.fxe"},
	{0xC49ABA75841BB878ULL, "811580737.ogg"},
	{0xC49B13043489CFADULL, "character_rendertodepth_p7.psh.fxo.dx11.assembly"},
	{0xC49BA414EBD50206ULL, "ts_templars_wpn_table.gda"},
	{0xC49BE9904A364876ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_exit.ani"},
	{0xC49BFC57B157A049ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0dd_basic_2.mao"},
	{0xC49C93F69822A076ULL, "39643807.ogg"},
	{0xC49CF5C1A84883ECULL, "vxht_keep_73k_0_0.vlm"},
	{0xC49D480B60D95BC7ULL, "210010313.ogg"},
	{0xC49D52C10B9DBA70ULL, "6176309_m.fxe"},
	{0xC49E1A2769349E8EULL, "hlac_ht_mansion2b_818_7y.dds"},
	{0xC49E53047C92B156ULL, "893486358.ogg"},
	{0xC49F69558C049D6EULL, "art\\characters\\weapons\\shields\\w_shd_dssa.msh"},
	{0xC49F77558C04B4BFULL, "art\\characters\\weapons\\shields\\w_shd_dssa.mao"},
	{0xC49F83558C04C95CULL, "art\\characters\\weapons\\shields\\w_shd_dssa.mmh"},
	{0xC49FC2B82091FCE7ULL, "388324201.ogg"},
	{0xC4A01011DC7CD70EULL, "6078742_m.fxe"},
	{0xC4A0A8AC8138C90BULL, "543442425.ogg"},
	{0xC4A0E12A64C94123ULL, "vxw_bonepitdestroyed_a3_27f_0_0.vlm"},
	{0xC4A12672636CA908ULL, "638818433.ogg"},
	{0xC4A144842FD8A391ULL, "plt_genpt_saving_merrill.nss"},
	{0xC4A1684DEEAF4777ULL, "6074484_m.fxe"},
	{0xC4A18C9933B433CDULL, "lgt250_way_dalish_conflict6147143.cl"},
	{0xC4A1AC374B4C7C75ULL, "hlac_lt_warehouse_n_0_3x.dds"},
	{0xC4A1DFDDCDFB394FULL, "llac_dr_0_deeproads_5842_11.dds"},
	{0xC4A2463C863470B5ULL, "vxlt_0_lowtown_a2_csz_0_0.vlm"},
	{0xC4A2716A1891092BULL, "w_encounter2_43_0_lc.mmh"},
	{0xC4A28464F8B15025ULL, "hlac_ht_0_hightown_a3_65536_4x.dds"},
	{0xC4A2A41DFC20735DULL, "6114000_f.fxe"},
	{0xC4A2C62E01C07A59ULL, "295549016.ogg"},
	{0xC4A2D783F9B5467CULL, "vxlt_gallowstemplar_n_3p2_0_0.vlm"},
	{0xC4A2E07111C1E0F8ULL, "6070872_m.fxe"},
	{0xC4A3473868A8BF93ULL, "qun111_ketojan6040661.cl"},
	{0xC4A34F8DD37C2B5BULL, "479258633.ogg"},
	{0xC4A36D3BD61185C8ULL, "vxht_brothel_n_5c1_0_0.vlm"},
	{0xC4A37B5AA0472AABULL, "664020603.ogg"},
	{0xC4A46C5900AD66F4ULL, "6201827_m.fxe"},
	{0xC4A4AED30C849340ULL, "one100dg_kelder6055440.cl"},
	{0xC4A50A9F40802E68ULL, "189753827.ogg"},
	{0xC4A551B5224D41BDULL, "614112839.ogg"},
	{0xC4A587CBE000C261ULL, "249157827.ogg"},
	{0xC4A599CD27C14719ULL, "lt_warehouse_n_1_0_lc.mmh"},
	{0xC4A5C006AA945FF5ULL, "hlac_ca_caverns2prc_5802_41x.dds"},
	{0xC4A63D97CA52E1E7ULL, "690125970.ogg"},
	{0xC4A678B80BAF403FULL, "vxdr_primevaltomb_5ap_0_1.vlm"},
	{0xC4A6F8743A9DCFCDULL, "mrl150cr_revenant.utc"},
	{0xC4A70ED364332025ULL, "ico_longbow_2.dds"},
	{0xC4A77F74CAE987E3ULL, "llac_lt_0_lowtown_a2_n_65536_6.dds"},
	{0xC4A7AC4FB8A4F5FDULL, "50728050.ogg"},
	{0xC4A7E0A00AA3436DULL, "6139253_m.fxe"},
	{0xC4A7EF6D8E9ED153ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_0y.dds"},
	{0xC4A83D39DB8ECBD8ULL, "vxca_caverns2_71z_0_0.vlm"},
	{0xC4A8BF46FB3A20E7ULL, "6163182_m.fxe"},
	{0xC4A8F1904A40E067ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_exit.evt"},
	{0xC4A99199FE552589ULL, "vxca_cave_1af_0_0.vlm"},
	{0xC4A999EBDEB77E9DULL, "vxht_chantry_2sw_0_0.vlm"},
	{0xC4A9B557D114E966ULL, "vxht_brothel_n_6h5_0_0.vlm"},
	{0xC4A9E01BD3037A8FULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg103a_0t.dds"},
	{0xC4A9EFE76F63CD3DULL, "zz_endshot.cut"},
	{0xC4AA10FE07D00143ULL, "vxlt_0_lowtown_a2_hf8_0_0.vlm"},
	{0xC4AA3288D03ACD85ULL, "gen_im_trs_drk_junk2_01.uti"},
	{0xC4AA6335117EEE2EULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamdrain.dura.ani"},
	{0xC4AA6A9399BEF284ULL, "art\\vfx\\combat\\vfx_m_arc_element_act_c.anb"},
	{0xC4AAC9ABC8D0753FULL, "vxht_keep_79z_0_0.vlm"},
	{0xC4AB56E01B1EC084ULL, "vxw_blightlands_h3_0_0.vlm"},
	{0xC4AB6DBB29DD7D49ULL, "genlock_dagger.utc"},
	{0xC4AB7BC186E525A2ULL, "vxca_caverns1_iiu_0_0.vlm"},
	{0xC4ABA3A731AFCF17ULL, "zz_approval_debug6003765.cl"},
	{0xC4ABC2C2D3E63705ULL, "hlac_lt_gallowstemplar_2383_3z.dds"},
	{0xC4AD4EE7AC128828ULL, "hlac_lt_0_lowtown_a2_-65538_0y.dds"},
	{0xC4ADC553484E16D5ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masa.msh"},
	{0xC4ADC753484E1A4BULL, "art\\characters\\playercharacter\\humanfemale\\glv_masa.mmh"},
	{0xC4ADD353484E2EE8ULL, "art\\characters\\playercharacter\\humanfemale\\glv_masa.mao"},
	{0xC4AE8618B399413CULL, "hlac_lt_0_lowtown_a2_65536_7y.dds"},
	{0xC4AEA38CFB10ACCBULL, "vxca_caverns2_95s_0_0.vlm"},
	{0xC4AEB488A9F8DC1CULL, "follower_barks6169905.cl"},
	{0xC4AF323E97819535ULL, "6139204_m.fxe"},
	{0xC4AF69F8906D1130ULL, "vxlt_0_lowtown_ekg_0_2.vlm"},
	{0xC4AF8504CFAB490BULL, "hlac_ht_0_hightown_-3_0x.dds"},
	{0xC4AFB8BF7FD3D036ULL, "clf260_harley6069585.cl"},
	{0xC4B012904A4736B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_exit.gad"},
	{0xC4B09835702FE907ULL, "vxw_bonepitmining_a2_1rx_0_0.vlm"},
	{0xC4B0F0666871B5CFULL, "210763840.ogg"},
	{0xC4B115FAEE168BDDULL, "6113100_f.fxe"},
	{0xC4B12EE1F0C82818ULL, "character_character_p247.psh.fxo.dx11"},
	{0xC4B15171D1FEDD9BULL, "vxlt_backalley_g_1eu_0_0.vlm"},
	{0xC4B1643DBB41769AULL, "140327294.ogg"},
	{0xC4B16EBEB056E5D3ULL, "634094148.ogg"},
	{0xC4B19417971B0C05ULL, "1071462017.ogg"},
	{0xC4B19EB12E168009ULL, "hlac_lt_gallowstemplar_2393_0z.dds"},
	{0xC4B2EC943F2DCC96ULL, "mcr310_sibling6076295.cl"},
	{0xC4B30C5D83BA4E58ULL, "mag110_arianni6020491.cl"},
	{0xC4B36E1A5766EA56ULL, "112450177.ogg"},
	{0xC4B377450ACEC22FULL, "vxlt_hangedman_n_21v_0_0.vlm"},
	{0xC4B4097062B450B0ULL, "gen_im_acc_rng_act3_03.uti"},
	{0xC4B43F819D30773BULL, "pro000_merchant6051645.cl"},
	{0xC4B5174930AD669EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0cc_basic.phy"},
	{0xC4B62DDD844BF566ULL, "gen00fl_isabela6098209.cl"},
	{0xC4B653C1C34A2008ULL, "vxw_encounter2_camp_c5_0_0.vlm"},
	{0xC4B6EB07F226D086ULL, "vxlt_gallowstemplar_n_4wr_0_0.vlm"},
	{0xC4B7200312CCE47CULL, "6082243_f.fxe"},
	{0xC4B73C089C26CCDEULL, "vxht_0_hightown_a3_b66_0_0.vlm"},
	{0xC4B7F83F0C86CCA7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_urn_sa.dds"},
	{0xC4B8495F7B32A2DAULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_beta.mmh"},
	{0xC4B84D5F7B32A911ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_beta.mao"},
	{0xC4B85B35118B1E7FULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamdrain.dura.evt"},
	{0xC4B85B5F7B32C120ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_beta.msh"},
	{0xC4B8899399CB6487ULL, "art\\vfx\\combat\\vfx_m_arc_element_act_c.mmh"},
	{0xC4B8A9F74BD6C1DAULL, "158196400.ogg"},
	{0xC4B9158187A2A179ULL, "llac_w_bonepitdestroyed_a3_262144_0.dds"},
	{0xC4B91FB1AE017A9BULL, "hlac_lt_foundry_n_320_12y.dds"},
	{0xC4B962E27E6887D0ULL, "character_character_p78.psh.fxo.dx11.assembly"},
	{0xC4B9AAB9B4F7D0A4ULL, "w_encounter3_142_0.msh"},
	{0xC4B9BCB9B4F7EF0DULL, "w_encounter3_142_0.mao"},
	{0xC4B9C0B9B4F7F646ULL, "w_encounter3_142_0.mmh"},
	{0xC4B9C8BD0D14D904ULL, "vxw_blightlands_m0_0_0.vlm"},
	{0xC4BA4F597C52AF52ULL, "hlac_lt_0_lowtown_a2_n_-65537_1z.dds"},
	{0xC4BC22D2DDF1A771ULL, "dae000_madam_lusine6151400.cl"},
	{0xC4BC579F998B3879ULL, "amb_scene_dwarfmech.bnk"},
	{0xC4BC6B79F23D7C65ULL, "lgt210_magu_selby6142065.cl"},
	{0xC4BC8EA8E479CEE3ULL, "w_bonepitdestroyed_a3_47_0.msh"},
	{0xC4BC9CA8E479E63DULL, "w_bonepitdestroyed_a3_47_0.mmh"},
	{0xC4BCA0A8E479ED4EULL, "w_bonepitdestroyed_a3_47_0.mao"},
	{0xC4BCB13794CF2308ULL, "236131613.ogg"},
	{0xC4BDA71D3EBF6254ULL, "dra_ciga.mao"},
	{0xC4BDA847767A86F8ULL, "vxht_bartrand_n_5qw_0_0.vlm"},
	{0xC4BDC4939626DEADULL, "347986101.ogg"},
	{0xC4BDF26AB973B033ULL, "vxlt_gamlenshouse_n_3v_0_0.vlm"},
	{0xC4BE5E006B19A4B3ULL, "6930661.ogg"},
	{0xC4BF0068173107C5ULL, "mag320_orsino6071116.cl"},
	{0xC4BF37686B720BEDULL, "vxdr_primevaltomb_6wb_0_0.vlm"},
	{0xC4BF3C3511910868ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamdrain.dura.gad"},
	{0xC4BF43B3DA141BC1ULL, "ave200_party_comment6116363.cl"},
	{0xC4C035FB14A2D533ULL, "6125885_m.fxe"},
	{0xC4C0E9958219A6F8ULL, "fx_fade_skybox.dds"},
	{0xC4C105931E0C170DULL, "hlac_w_bonepitmining_a2_65541_0z.dds"},
	{0xC4C1269B96A5370EULL, "6161372_m.fxe"},
	{0xC4C1644023826DAEULL, "vxht_mansion2_1xf_0_0.vlm"},
	{0xC4C1E9FCCBAC0BC8ULL, "505047849.ogg"},
	{0xC4C2063E2256187FULL, "hlac_lt_0_lowtown_a2_131070_0z.dds"},
	{0xC4C2BD6EA54BC37FULL, "one140cr_charade6128526.cl"},
	{0xC4C33F882C57AFCBULL, "670427986.ogg"},
	{0xC4C3D0C2F2DB7E4AULL, "gen00fl_varric6093684.cl"},
	{0xC4C402F765A2B5F5ULL, "vxw_sundermount_2za_0_1.vlm"},
	{0xC4C43C8EF5F58C34ULL, "vxlt_warehouse_n_t2_0_0.vlm"},
	{0xC4C4B0B88B426925ULL, "6060375_m.fxe"},
	{0xC4C590AB51A08C75ULL, "gen00fl_aveline6140023.cl"},
	{0xC4C59D45EE41CB1EULL, "w_encounter2_camp_10_0.rml"},
	{0xC4C5C4E3A4DD8A5EULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_rack.phy"},
	{0xC4C639D02C61859CULL, "hlac_lt_gallowstemplar_f_2383_3x.dds"},
	{0xC4C672B07BE8E596ULL, "6116099_m.fxe"},
	{0xC4C67CC08ED2B9A4ULL, "vxht_0_hightown_a3_bx2_0_0.vlm"},
	{0xC4C67D4D537F1D83ULL, "da2terrain_rendertodepth_p2.vsh.fxo.dx11"},
	{0xC4C697928DDADC68ULL, "hlac_lt_0_lowtown_-131073_3z.dds"},
	{0xC4C76430A8ACC67FULL, "art\\levels\\races\\design\\prp_felandaris.msh"},
	{0xC4C76A30A8ACD021ULL, "art\\levels\\races\\design\\prp_felandaris.mmh"},
	{0xC4C7727C354B1B6CULL, "llac_ht_keep_n_4519_12.dds"},
	{0xC4C77630A8ACE4C2ULL, "art\\levels\\races\\design\\prp_felandaris.mao"},
	{0xC4C79F54B2D97653ULL, "lt_0_lowtown_docks_2_0_ll.mmh"},
	{0xC4C7E9262A519B46ULL, "lt_0_lowtown_a2_26_0_ll.mmh"},
	{0xC4C83A99FF0AF184ULL, "vxw_blightlands_v1_0_0.vlm"},
	{0xC4C849898DB42A31ULL, "vxht_playershouse_n_1i4_0_0.vlm"},
	{0xC4C86E5179A2CF9FULL, "lt_0_lowtown_docks_209_0.msh"},
	{0xC4C8745179A2DA41ULL, "lt_0_lowtown_docks_209_0.mmh"},
	{0xC4C8805179A2EEE2ULL, "lt_0_lowtown_docks_209_0.mao"},
	{0xC4C89C8855DC31CAULL, "vxht_chantry_n_2sw_0_0.vlm"},
	{0xC4C90C0292A3CBF2ULL, "vxlt_warehouse_100_0_0.vlm"},
	{0xC4C94B57DADE2B3AULL, "186967777.ogg"},
	{0xC4C9BA3EA9985BA0ULL, "vxht_0_hightown_a2_cat_0_1.vlm"},
	{0xC4CA2190E20BC916ULL, "6078081_m.fxe"},
	{0xC4CA5E4395818494ULL, "llac_dr_0_deeproads_5841_2.dds"},
	{0xC4CAE2CC52F86452ULL, "mag100_cullen6095143.cl"},
	{0xC4CB0742B1A1F623ULL, "1056085228.ogg"},
	{0xC4CB89885ACC3B6EULL, "vxht_keep_n_5gx_0_0.vlm"},
	{0xC4CBA2CA8EEB8CEFULL, "hlac_dr_primevaltomb_658_5y.dds"},
	{0xC4CBE6FA1D087B20ULL, "mer141_jansen6077787.cl"},
	{0xC4CC0C9D44DD330AULL, "hlac_lt_darktown_-1_22z.dds"},
	{0xC4CC1A2468709F78ULL, "vxht_bartrand_n_5tt_0_0.vlm"},
	{0xC4CC2721380AED8AULL, "mer120_hubert6059426.cl"},
	{0xC4CC37221B38910DULL, "fx_fireringfb.dds"},
	{0xC4CCB645EE4814E2ULL, "w_encounter2_camp_10_0.phy"},
	{0xC4CD27C7B9AFFB8CULL, "vxht_mansion_n_626_0_0.vlm"},
	{0xC4CD37BCEBC1A123ULL, "186002323.ogg"},
	{0xC4CD92B42C4B6EE1ULL, "6202265_m.fxe"},
	{0xC4CD986C59EFA8E2ULL, "81796928.ogg"},
	{0xC4CDA9455D68246DULL, "vxdr_primevaltomb_6za_0_0.vlm"},
	{0xC4CE03026AEAA110ULL, "hlac_ht_0_hightown_a2_-65534_1x.dds"},
	{0xC4CEED856DC03AAFULL, "vxht_bartrand_n_4fh_0_0.vlm"},
	{0xC4CEF7E1BF7E52B0ULL, "hlac_lt_gallowscourtyard_-131072_1z.dds"},
	{0xC4CF0FA1A0C69321ULL, "313431280.ogg"},
	{0xC4CF3B21E11AF832ULL, "vxht_0_hightown_c5o_0_0.vlm"},
	{0xC4CF9878889B4F8EULL, "6166272_m.fxe"},
	{0xC4CFD61D1578862EULL, "vxht_mansion2_1cg_0_0.vlm"},
	{0xC4D01CDA1781415EULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.mmh"},
	{0xC4D03386B91ED227ULL, "vxdr_dungeon_2hg_0_0.vlm"},
	{0xC4D0EB647CDA075AULL, "165855972.ogg"},
	{0xC4D121759B7BC6ACULL, "vxht_0_hightown_c6x_0_0.vlm"},
	{0xC4D174F5C62A54E2ULL, "vxdr_primevalentrance_1qp_0_0.vlm"},
	{0xC4D17631EB121C6EULL, "vxdr_slavers_1ax_0_0.vlm"},
	{0xC4D1DD274CADE82CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobb_0s.dds"},
	{0xC4D22C9E0656904DULL, "6182022_m.fxe"},
	{0xC4D2494F71FDD28DULL, "489464640.ogg"},
	{0xC4D253C252CF3205ULL, "109266472.ogg"},
	{0xC4D288F09956EA7FULL, "lgt110_gxa_coa_spy6117461.cl"},
	{0xC4D31EBEE7346773ULL, "448823336.ogg"},
	{0xC4D364E321773A96ULL, "vxht_0_hightown_a3_bdt_0_0.vlm"},
	{0xC4D3BA76DAA469ACULL, "character_character_p90.psh.fxo"},
	{0xC4D3F220BCF095B8ULL, "198247629.ogg"},
	{0xC4D4194ADDC061A3ULL, "ser200_aveline6201537.cl"},
	{0xC4D41AB74F9F2546ULL, "gen00fl_merrill6130114.cl"},
	{0xC4D4DAE97653453BULL, "6133033_m.fxe"},
	{0xC4D516FE89659067ULL, "vxht_keep_db_0_0.vlm"},
	{0xC4D5AAF334F1ABD1ULL, "mrl150_merrill6078335.cl"},
	{0xC4D5ED3B05E5BAD2ULL, "6080261_f.fxe"},
	{0xC4D5F75886E9FBC2ULL, "175066887.ogg"},
	{0xC4D60BA6603EF6DCULL, "hlac_lt_0_lowtown_a2_n_c_65538_1z.dds"},
	{0xC4D60E993DB777B1ULL, "589861541.ogg"},
	{0xC4D61A68E3F5DEB5ULL, "gen00fl_sebastian6179973.cl"},
	{0xC4D71E46C282A822ULL, "6099750_m.fxe"},
	{0xC4D75A300FDFA69CULL, "vxlt_0_lowtown_f0s_0_0.vlm"},
	{0xC4D7E2ECCF633887ULL, "vxht_playershouse_n_1p9_0_0.vlm"},
	{0xC4D7ED1F93A64195ULL, "hlac_lt_0_lowtown_65538_4y.dds"},
	{0xC4D8EF123CA1A0CEULL, "vxw_bonepit_1x8_0_1.vlm"},
	{0xC4D91C9B0F2CE1EBULL, "420899285.ogg"},
	{0xC4D95A679F43C919ULL, "vxht_keep_n_73a_0_0.vlm"},
	{0xC4D994E9B0D55472ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masg.phy"},
	{0xC4D9E122EE307D0EULL, "zz_ave_debug6134025.cl"},
	{0xC4DA04966CF2AE97ULL, "6174304_m.fxe"},
	{0xC4DA56E2DFB22332ULL, "perception.ncs"},
	{0xC4DADAC50224CE1BULL, "11030501.ogg"},
	{0xC4DBC9D894646A95ULL, "6168480_m.fxe"},
	{0xC4DC0B655472B1B7ULL, "6116616_m.fxe"},
	{0xC4DC5C34FC3A1AC4ULL, "ser191_emeric6120473.cl"},
	{0xC4DCCB6184D85473ULL, "mcr310_sibling6076278.cl"},
	{0xC4DCD2FA39C67E7BULL, "vxw_encounter3_eg_0_0.vlm"},
	{0xC4DCFBDA178B931DULL, "art\\vfx\\cinematic\\vfx_flem_drag_breath_c.anb"},
	{0xC4DD539C8B4E4756ULL, "vxca_cavern1_haunted_drd_0_0.vlm"},
	{0xC4DD5F625FB6532FULL, "vxht_bartrand_n_4kk_0_0.vlm"},
	{0xC4DD744CCE84EAD0ULL, "88263577.ogg"},
	{0xC4DDBC2B902C071FULL, "royalarcher_decoy.dds"},
	{0xC4DE5B84300C797AULL, "plt_genpt_saving_merrill.plo"},
	{0xC4DE685F7B53A6E5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_beta.phy"},
	{0xC4DE82198E4FF06DULL, "gen00fl_sebastian6180519.cl"},
	{0xC4DED773EF11E56DULL, "gen00fl_varric6109216.cl"},
	{0xC4DFDB97EA404215ULL, "168769437.ogg"},
	{0xC4E0C83A1E52910CULL, "follower_banter6168909.cl"},
	{0xC4E0D283A77F3F7AULL, "vxht_keep_n_42v_0_0.vlm"},
	{0xC4E140DA986F0698ULL, "6110802_m.fxe"},
	{0xC4E1D6C0136D5316ULL, "vxht_0_hightown_a3_bcu_0_0.vlm"},
	{0xC4E22C53CC9A822CULL, "character_character_p47.psh.fxo"},
	{0xC4E2905A594D03EAULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikia.mao"},
	{0xC4E2945A594D0A39ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikia.mmh"},
	{0xC4E29E5A594D1B27ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikia.msh"},
	{0xC4E432C705372ECFULL, "one300_zevran6075858.cl"},
	{0xC4E442B25C574248ULL, "mag170_recruit026094515.cl"},
	{0xC4E4F8413CF1E56FULL, "1058351491.ogg"},
	{0xC4E50083FC040E1FULL, "hlac_lt_darktown_-1_72z.dds"},
	{0xC4E57D86CB604C29ULL, "w_sundermount_135_0.phy"},
	{0xC4E5804239C862CAULL, "vxht_chantry_n_2iq_0_0.vlm"},
	{0xC4E59A12322DA9A2ULL, "zz_one_debug6069462.cl"},
	{0xC4E5B2351804ED34ULL, "6104274_m.fxe"},
	{0xC4E654C9C1595107ULL, "vxht_playershouse_n_1w8_0_0.vlm"},
	{0xC4E656B638616B38ULL, "dae000_merchant_mage_hf6164459.cl"},
	{0xC4E6A62DCBE80B3AULL, "874141865.ogg"},
	{0xC4E6D2D59F5952FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_enter.evt"},
	{0xC4E6F96F9D2443A2ULL, "6047524_m.fxe"},
	{0xC4E760540091980BULL, "zz_mer_debug6032340.cl"},
	{0xC4E81B3C74CD41C7ULL, "qun221_viscount6070180.cl"},
	{0xC4E888A273C4B0DFULL, "tbe200im_forgotten_chest.uti"},
	{0xC4E8BAE08199CF0CULL, "6066325_m.fxe"},
	{0xC4E8E8D2CB39E3BEULL, "6125487_f.fxe"},
	{0xC4E98588520719FAULL, "hlac_ht_0_hightown_n_-65535_1x.dds"},
	{0xC4E9E409E6C8FC5FULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_fenr.msh"},
	{0xC4E9EA09E6C90601ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_fenr.mmh"},
	{0xC4EA25158D0418E0ULL, "art\\vfx\\materials\\chunky\\vfx_c_lightning.mao"},
	{0xC4EA269C43EC450BULL, "246320530.ogg"},
	{0xC4EA40F90165B3E3ULL, "dae200cr_solivitus.utc"},
	{0xC4EA7CF72FA5582EULL, "mcr310_argument6074447.cl"},
	{0xC4EB398CE9080F9AULL, "zz_fol_debug6057000.cl"},
	{0xC4EB818E9018942BULL, "art\\levels\\races\\proxy\\props\\pressure\\blendtree\\pressureplate.anb"},
	{0xC4EB8686CB64C725ULL, "w_sundermount_135_0.rml"},
	{0xC4EBA6A05C2E444CULL, "6143514_m.fxe"},
	{0xC4EBB059FD31EB06ULL, "dae011cr_gossip_1.utc"},
	{0xC4EBC4E38A17F37FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masd_0t.dds"},
	{0xC4EBFFD8F2E23885ULL, "vxht_bartrand_n_5qn_0_0.vlm"},
	{0xC4EC06AAC292F199ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm01.mmh"},
	{0xC4EC10AAC2930287ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm01.msh"},
	{0xC4EC4FA09E4F4CE3ULL, "static_rendertodepth_p0.vsh.fxo.dx11.assembly"},
	{0xC4EC7180AFD1FFBCULL, "hlac_lt_0_lowtown_n_0_9y.dds"},
	{0xC4ECFABA3200C6A2ULL, "t3_wdg_dw1.tnt"},
	{0xC4ED1FC3F266CFFBULL, "vxht_keep_6as_0_0.vlm"},
	{0xC4ED5FDF28803716ULL, "zz_followers_banter6157763.cl"},
	{0xC4EDE9D59F5F9829ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_enter.gad"},
	{0xC4EE02735E339EAFULL, "6060467_m.fxe"},
	{0xC4EE58AFAA1685E2ULL, "vxdr_primevalentrance_1kr_0_0.vlm"},
	{0xC4EEB7332FE9AA4EULL, "ht_mansion_n_3371.anb"},
	{0xC4EEC3BA5B777981ULL, "t3_dps_b02.tnt"},
	{0xC4EF080E55725C33ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basicnb.phy"},
	{0xC4EF6CFD5EA7C79EULL, "mag311_keran6041575.cl"},
	{0xC4EF76385F582811ULL, "vxlt_darktown_2wh_0_0.vlm"},
	{0xC4EFC2753EAFB022ULL, "820617208.ogg"},
	{0xC4F0658B7B7AB372ULL, "6174598_m.fxe"},
	{0xC4F0783DB7D18D8AULL, "hlac_lt_0_lowtown_a2_-65536_3x.dds"},
	{0xC4F0907A246C1FF4ULL, "hlac_ca_caverns2_5802_6y.dds"},
	{0xC4F09E30BE909AACULL, "character_character_p36.psh.fxo"},
	{0xC4F0AC9EC44CE97BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_loop.ani"},
	{0xC4F0C50AB6E74973ULL, "hlac_ht_brothel_n_3382_1y.dds"},
	{0xC4F0D22BB9F48126ULL, "hlac_lt_darktown_-1_59x.dds"},
	{0xC4F110A9130FD9B7ULL, "prm000im_isaac_helm.uti"},
	{0xC4F19145EE66F196ULL, "w_encounter2_camp_10_0.mao"},
	{0xC4F19D45EE670635ULL, "w_encounter2_camp_10_0.mmh"},
	{0xC4F19F45EE6709ABULL, "w_encounter2_camp_10_0.msh"},
	{0xC4F211583601FE1BULL, "aip_ogre_base.gda"},
	{0xC4F23FAF2CABD272ULL, "sib131ar_exit_estate2.utp"},
	{0xC4F33E645FEC9B3AULL, "827952087.ogg"},
	{0xC4F38FBF4C7B9AEAULL, "vxlt_undercitydungeon_1im_0_0.vlm"},
	{0xC4F3F21F2BBE7B4AULL, "vxht_chantry_n_2dr_0_0.vlm"},
	{0xC4F40200A66ED922ULL, "6093550_m.fxe"},
	{0xC4F43848BDBCC5F4ULL, "6036548_m.fxe"},
	{0xC4F43D0113CA4834ULL, "485625907.ogg"},
	{0xC4F458C359641ED8ULL, "6117418_m.fxe"},
	{0xC4F4BAD59F656777ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_right_enter.ani"},
	{0xC4F5BADCBD1ACBFDULL, "6180002_m.fxe"},
	{0xC4F60A2CC98B94F1ULL, "168162795.ogg"},
	{0xC4F6683CDF04662AULL, "vxlt_gallowscourtyard_n_470_0_0.vlm"},
	{0xC4F6E25A4C8DAD53ULL, "lgt110_urn_con_leader6133024.cl"},
	{0xC4F6F97DB6B41EB7ULL, "6126691_m.fxe"},
	{0xC4F7343BC64DD8BCULL, "727877457.ogg"},
	{0xC4F7778D7D03B196ULL, "hlac_lt_gallowstemplar_n_2394_1x.dds"},
	{0xC4F7B75890F3FCA7ULL, "gen00fl_aveline_mother_died.stg"},
	{0xC4F7CD2BA07DF9F4ULL, "vxht_mansion2b_2ul_0_0.vlm"},
	{0xC4F7EB03A9CB9625ULL, "6105446_f.fxe"},
	{0xC4F9567792A7C30CULL, "lowtown_fem_alien_elf_2_mob6166363.cl"},
	{0xC4F9FA84AA8579F3ULL, "6085256_f.fxe"},
	{0xC4FA462C49274E1DULL, "hlac_lt_0_lowtown_a2_n_65536_7y.dds"},
	{0xC4FA6D6227259E8BULL, "452482606.ogg"},
	{0xC4FA71B5E4D85105ULL, "vxht_bartrand_n_5tm_0_0.vlm"},
	{0xC4FBC21454E6AAD9ULL, "gen00fl_bethany6102899.cl"},
	{0xC4FD616A763C4325ULL, "llac_lt_warehouse_0_4.dds"},
	{0xC4FDA48EABB78AD8ULL, "gen00fl_player6207780.cl"},
	{0xC4FE7403F75177FAULL, "art\\characters\\creatures\\pridedemon\\animation\\a_end.gad"},
	{0xC4FEE8E15FDCE904ULL, "702114300.ogg"},
	{0xC4FEF810D0324055ULL, "hlac_ht_0_hightown_n_-65534_1x.dds"},
	{0xC4FF153BFD5652F4ULL, "vxht_bartrand_n_654_0_0.vlm"},
	{0xC4FF1DFD2F094077ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junka_n.dds"},
	{0xC4FFD2B0E6FC21E5ULL, "hlac_ht_keep_4519_5x.dds"},
	{0xC5002E7135421D94ULL, "cntr_rock_f.mao"},
	{0xC500AD522221ECB9ULL, "6150331_m.fxe"},
	{0xC500C205671FB21EULL, "187685184.ogg"},
	{0xC500D30E213F277BULL, "vxht_0_hightown_n_byl_0_0.vlm"},
	{0xC500E8F9824CE6EFULL, "138721137.ogg"},
	{0xC502019C3E71B36AULL, "vxlt_undercitydungeon_1ln_0_0.vlm"},
	{0xC5020B0A532E75BCULL, "qun160_vashothlead6055580.cl"},
	{0xC502398178E2C5CDULL, "lgt220_shp_mag_guard6149907.cl"},
	{0xC50244A2C6BC69C2ULL, "vxdr_slavers_23a_0_0.vlm"},
	{0xC50281C13B7F110DULL, "vxht_mansion2b_1c8_0_0.vlm"},
	{0xC502A6D99EFB1FF2ULL, "6149349_m.fxe"},
	{0xC502E83A4F996AD1ULL, "vxlt_0_lowtown_docks_a3_n_d3p_0_0.vlm"},
	{0xC50302C1F77B59E7ULL, "hlac_ht_0_hightown_a3_-131072_0x.dds"},
	{0xC5032892AD490908ULL, "6097384_f.fxe"},
	{0xC503456DCF3D6DF7ULL, "ser120_ghyslain6104803.cl"},
	{0xC503EB787979C08BULL, "dae000cr_merchant_armor.utc"},
	{0xC503EF04D77B8D8FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch03.gad"},
	{0xC5042E93F2DDE4AEULL, "6155308_m.fxe"},
	{0xC5044418921F406CULL, "vxw_encounter1_fg_0_0.vlm"},
	{0xC5046369C0828CA3ULL, "hlac_lt_0_lowtown_docks_n_-262144_3y.dds"},
	{0xC504AEE7D835AB64ULL, "vxht_keep_n_675_0_0.vlm"},
	{0xC504BEEE5F03B8E5ULL, "6109377_m.fxe"},
	{0xC504D39EC45DD405ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_aside_right_loop.gad"},
	{0xC5054119CFEEE3D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_right_1.ani"},
	{0xC5063F0892741274ULL, "vxht_mansion2b_2nm_0_0.vlm"},
	{0xC5067A484D69D99BULL, "6111235_m.fxe"},
	{0xC506C56791DB6BF6ULL, "6143721_m.fxe"},
	{0xC506EDACFEFD3493ULL, "vxca_caverns1_j20_0_2.vlm"},
	{0xC506F5482EAFF0DBULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wb.evt"},
	{0xC507C2A835E0A5D3ULL, "831829918.ogg"},
	{0xC507C2AE0D21FC66ULL, "6040882_m.fxe"},
	{0xC507DBF1B6D3B7C6ULL, "682062785.ogg"},
	{0xC50825E27414E2C8ULL, "mrl310_merrill6113518.cl"},
	{0xC508944DEF0F63C7ULL, "dae000cr_grand_cleric6147701.cl"},
	{0xC5090E65F04F6B9FULL, "337938415.ogg"},
	{0xC50931519A841B5FULL, "hlac_dr_primevalentrance_1726_5z.dds"},
	{0xC50946E6E785721AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hlfbraizerxa.mmh"},
	{0xC5094AE6E7857951ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hlfbraizerxa.mao"},
	{0xC50958E6E7859160ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_hlfbraizerxa.msh"},
	{0xC50962E3AAE673DDULL, "6104218_f.fxe"},
	{0xC509796AEF430C27ULL, "vxht_0_hightown_a2_a4s_0_0.vlm"},
	{0xC5099C1D2A843FE8ULL, "6122808_f.fxe"},
	{0xC50A0404D7801D3CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch03.evt"},
	{0xC50A166A24102571ULL, "6113067_f.fxe"},
	{0xC50B4C19CFF36273ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_gesture_right_1.gad"},
	{0xC50BF4797BCD9205ULL, "288241859.ogg"},
	{0xC50CE02C30397E8CULL, "344865932.ogg"},
	{0xC50D0F0B38B3119DULL, "docks_male_overseer_mob6176547.cl"},
	{0xC50D85058B1CA63AULL, "92630822.ogg"},
	{0xC50DA75575157A2BULL, "916597862.ogg"},
	{0xC50DEE482EB6031CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wb.gad"},
	{0xC50E53BFA358B8B1ULL, "575712903.ogg"},
	{0xC50EB61148A13424ULL, "vxht_0_hightown_a3_bc7_0_0.vlm"},
	{0xC50F5153D85C7018ULL, "mag110_arianni6049717.cl"},
	{0xC50F515A9819E647ULL, "hlac_lt_0_lowtown_docks_n_65537_0x.dds"},
	{0xC510CDDF3AFA906FULL, "dae100_bran6118156.cl"},
	{0xC510E54180A60460ULL, "576023258.ogg"},
	{0xC5116686CB85604EULL, "w_sundermount_135_0.mmh"},
	{0xC5117286CB8574E5ULL, "w_sundermount_135_0.mao"},
	{0xC5118086CB858C3CULL, "w_sundermount_135_0.msh"},
	{0xC512184ABFCDC313ULL, "narrative_varric6180097.cl"},
	{0xC5128F03F7624EE0ULL, "art\\characters\\creatures\\pridedemon\\animation\\a_end.ani"},
	{0xC512B5F5841558ECULL, "vxw_encounter1_kd_0_0.vlm"},
	{0xC51330FB46A4E1D9ULL, "vxw_encounter2_camp_7t_0_0.vlm"},
	{0xC5140DB4244B9C4AULL, "hlac_lt_0_lowtown_a2_-65537_0z.dds"},
	{0xC514B0E5846A2AF4ULL, "vxht_mansion2b_2kr_0_0.vlm"},
	{0xC514ED482EBC201AULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wb.ani"},
	{0xC51514ABF2AB0B3AULL, "vxdr_primevaltomb_7le_0_0.vlm"},
	{0xC5151A7B6F3FFBD6ULL, "w_bonepitmining_a2_208_0.msh"},
	{0xC515247B6F400CD4ULL, "w_bonepitmining_a2_208_0.mmh"},
	{0xC515287B6F4013E7ULL, "w_bonepitmining_a2_208_0.mao"},
	{0xC51550E137A55A94ULL, "gen00fl_aveline6114054.cl"},
	{0xC515C7577C820A17ULL, "261797704.ogg"},
	{0xC515E0E7656A2B10ULL, "gen00fl_fenris6107147.cl"},
	{0xC515EB09E6EE437EULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_fenr.phy"},
	{0xC515FB941259A80AULL, "6129097_m.fxe"},
	{0xC51639F3CFDFB95CULL, "vxlt_0_lowtown_docks_a3_brc_0_0.vlm"},
	{0xC516EB835F93A3D7ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_06a_0n.dds"},
	{0xC5171F1FFC504D75ULL, "vxw_bonepitmining_a2_26q_0_0.vlm"},
	{0xC5173BC5066D0389ULL, "vxlt_undercitydungeon_1is_0_1.vlm"},
	{0xC51741A0DDF716ADULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_21roofplain.mmh"},
	{0xC51745A0DDF71DFEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_21roofplain.mao"},
	{0xC51753A0DDF73533ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_21roofplain.msh"},
	{0xC5178F4D88510C94ULL, "one110_pryce.cnv"},
	{0xC517EE12C1C4084FULL, "hlac_lt_0_lowtown_docks_a3_-1_0x.dds"},
	{0xC5180C04D78C6855ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch03.ani"},
	{0xC51887C8AA7D8D55ULL, "l_seb222ar_ruin.bnk"},
	{0xC518DF1A88C5229EULL, "vxlt_warehouse_n_ag_0_0.vlm"},
	{0xC519B2F24D088E04ULL, "6117021_m.fxe"},
	{0xC51A0C902D075F8EULL, "dae000_merchant_hm_mob.cnv"},
	{0xC51A492E804D9F3FULL, "llac_lt_0_lowtown_docks_a3_n_-65538_9.dds"},
	{0xC51B2C74F43F469BULL, "6121067_m.fxe"},
	{0xC51B714D8854C2D4ULL, "one110_pryce.bnk"},
	{0xC51BB7BF63B5B0D5ULL, "94890881.ogg"},
	{0xC51BB7C4E7E6223CULL, "vxht_keep_n_738_0_0.vlm"},
	{0xC51BD4FD40D0EA00ULL, "zz_followers_banter6158526.cl"},
	{0xC51C08187B9A6DC1ULL, "vxlt_0_lowtown_a3_n_c_d1w_0_0.vlm"},
	{0xC51C18EA8F0E6E78ULL, "mcr370_assignments6176740.cl"},
	{0xC51CADC198D27432ULL, "and100_anders6096038.cl"},
	{0xC51CEC01F174D640ULL, "vxlt_foundry_n_zf_0_0.vlm"},
	{0xC51D0901B064E4EBULL, "gen00fl_anders6125492.cl"},
	{0xC51D2E902D09CFFEULL, "dae000_merchant_hm_mob.bnk"},
	{0xC51D68F35EC63492ULL, "6015084_m.fxe"},
	{0xC51DAA014E1E58F9ULL, "256908009.ogg"},
	{0xC51DAD66DAEA3DFDULL, "345967591.ogg"},
	{0xC51DC9F1B3142877ULL, "dae000_arishok6067660.cl"},
	{0xC51DE694575F7E27ULL, "6000794_f.fxe"},
	{0xC51E17F6488E7399ULL, "hlac_lt_hangedman_-1_25y.dds"},
	{0xC51E8EEC0F0CABF1ULL, "llac_dr_primevaltomb_661_4.dds"},
	{0xC51EBED24F0933A5ULL, "6135749_m.fxe"},
	{0xC51F8E388FF0979BULL, "6168169_m.fxe"},
	{0xC51F95C009A390D3ULL, "mag110_vincento6017810.cl"},
	{0xC51FBE798A87F4D6ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be04a.msh"},
	{0xC51FC8798A8805D4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be04a.mmh"},
	{0xC51FE8130E4624B3ULL, "hlac_ca_caverns1_mines_24423_0x.dds"},
	{0xC5206CD61AAB2887ULL, "mcr310_argument6127568.cl"},
	{0xC52089DFA49D4A3FULL, "vxw_encounter1_ai_0_0.vlm"},
	{0xC5217DC9D82C4A6DULL, "681909446.ogg"},
	{0xC521BA33301508B2ULL, "ht_mansion_n_3371.rml"},
	{0xC5221A89F34EA145ULL, "qun110_petrice6042141.cl"},
	{0xC5223382729B7DD4ULL, "677091136.ogg"},
	{0xC5228E061BEE93D1ULL, "6080993_m.fxe"},
	{0xC522A90C6A6F01F2ULL, "hlac_dr_0_deeproads_5842_14z.dds"},
	{0xC5237E2E71139B71ULL, "llac_w_bonepit_0_3.dds"},
	{0xC523E1BE182233C2ULL, "fen310_varania6175041.cl"},
	{0xC523F86BD9489722ULL, "552015264.ogg"},
	{0xC5242F974BA2CCBBULL, "vxlt_warehouse_n_14j_0_0.vlm"},
	{0xC52457D628A997B9ULL, "follower_banter6160832.cl"},
	{0xC524A0B78B37E0CCULL, "hlac_ca_caverns1_24424_21x.dds"},
	{0xC524F2D283A7A082ULL, "hlac_lt_gallowscourtyard_-65538_2x.dds"},
	{0xC525807C9A9B1530ULL, "6140960_m.fxe"},
	{0xC525ADA1F8631C09ULL, "vxlt_undercitydungeon_1lp_0_1.vlm"},
	{0xC525D4D28EAA123FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_loop.gad"},
	{0xC5263A9300F0510AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comc_0t.dds"},
	{0xC5264AA9E77968B3ULL, "6075858_m.fxe"},
	{0xC5264B20E6F655DCULL, "6057101_m.fxe"},
	{0xC526A9A772C87C63ULL, "vxca_caverns2_7vt_0_0.vlm"},
	{0xC52722019A2A5281ULL, "hlac_lt_gallowstemplar_n_2383_6z.dds"},
	{0xC5279D22798195DFULL, "832180555.ogg"},
	{0xC527CAA9F52C9DD6ULL, "t1_mub_wh1.tnt"},
	{0xC5280F2371419146ULL, "qun221_viscount6070792.cl"},
	{0xC5283AA0DE057E6AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_21roofplain.phy"},
	{0xC5283EF943E14279ULL, "vxlt_blackemporium_vx_0_0.vlm"},
	{0xC528942F758BCEDEULL, "hlac_w_sundermount_65538_0x.dds"},
	{0xC5296C75D794F2A4ULL, "hlac_ca_cavern1_haunted_24422_6x.dds"},
	{0xC529721844313C8EULL, "6133120_m.fxe"},
	{0xC529745BC37BBC49ULL, "character_character_p68.psh.fxo"},
	{0xC529B636B819D906ULL, "6146911_m.fxe"},
	{0xC529C3D4EF473F57ULL, "24205770.ogg"},
	{0xC52AFC756E487CA1ULL, "vxlt_warehouse_n_sr_0_0.vlm"},
	{0xC52B0FF167EB2EE8ULL, "vxlt_undercityencounter_11f_0_0.vlm"},
	{0xC52B61FC1A3C6EF2ULL, "861813294.ogg"},
	{0xC52BA6D9AEE0E6A6ULL, "art\\vfx\\materials\\decal\\vfx_pd_blood_darkspawn.mao"},
	{0xC52C4F9D34F42032ULL, "gen00fl_merrill6121857.cl"},
	{0xC52C62E43718924BULL, "vxlt_0_lowtown_fh8_0_0.vlm"},
	{0xC52D382A80AE63A6ULL, "aip_fleshgolem_base.gda"},
	{0xC52D577E7BF4BF1CULL, "534755951.ogg"},
	{0xC52D688CF0746E6AULL, "270974607.ogg"},
	{0xC52DD6E9EDDDFC85ULL, "vxht_keep_4i3_0_0.vlm"},
	{0xC52DD85DF7CF8354ULL, "891013819.ogg"},
	{0xC52E0CF68E171F9FULL, "vxlt_gallowscourtyard_7fv_0_0.vlm"},
	{0xC52E864B33F07018ULL, "16147104.ogg"},
	{0xC52E8D46CEFEFAA0ULL, "25469048.ogg"},
	{0xC52E8D63C6D82A09ULL, "34734909.ogg"},
	{0xC52EB5E88F84A240ULL, "vxlt_gallowstemplar_f_5qo_0_0.vlm"},
	{0xC52EFBBC969362BFULL, "vxw_encounter1_fh_0_0.vlm"},
	{0xC52FBDC8C8CBB6E1ULL, "qcr000_aveline6041074.cl"},
	{0xC52FC01742E8910EULL, "lt_0_lowtown_27_0_lc.mmh"},
	{0xC52FCC339CD5F1B0ULL, "gen00fl_isabela6102485.cl"},
	{0xC52FE12FA7C11D52ULL, "6041089_m.fxe"},
	{0xC5302A087BE60FFFULL, "vxlt_warehouse_18i_0_0.vlm"},
	{0xC530465F5E801C73ULL, "6087312_m.fxe"},
	{0xC530DC1E914F571AULL, "isa311_isabela6094450.cl"},
	{0xC5318D46ADEFFF2EULL, "6087487_m.fxe"},
	{0xC531D4071E700395ULL, "6127243_m.fxe"},
	{0xC531FA50FE3CCC13ULL, "zz_seb_debug6187886.cl"},
	{0xC5324A9A2AD9A8F4ULL, "ser191_emeric6104732.cl"},
	{0xC5327BF4F4EC7922ULL, "6206881_f.fxe"},
	{0xC5328027A774FE2AULL, "6140863_m.fxe"},
	{0xC5338D2185546702ULL, "85320734.ogg"},
	{0xC533B73D22A52EB6ULL, "vxw_bonepitmining_a2_1xb_0_1.vlm"},
	{0xC5342C3BC2995C99ULL, "vxca_caverns2_7w1_0_0.vlm"},
	{0xC53431BF6EE5DF66ULL, "6049428_f.fxe"},
	{0xC5343929ACB0B305ULL, "vxht_bartrand_n_5hi_0_0.vlm"},
	{0xC5347D2A7014513AULL, "sse_ss_qunari_arishok.gda"},
	{0xC534C9ADF898EFA0ULL, "dae000_hate_enemies6167594.cl"},
	{0xC53512FA808E99C3ULL, "97806887.ogg"},
	{0xC535149CA83F07BFULL, "gencr_city_guard_ws.utc"},
	{0xC5352312CCC30A87ULL, "vxlt_undercityserialkiller_19a_0_0.vlm"},
	{0xC5354271E77B018FULL, "vxlt_undercityserialkiller_1ii_0_0.vlm"},
	{0xC53562BED2CEBAC1ULL, "971997986.ogg"},
	{0xC53619D0B17AEE74ULL, "6059235_m.fxe"},
	{0xC53694B2EA8A9899ULL, "hlac_ht_brothel_n_3382_0x.dds"},
	{0xC5369F5BC67CDA11ULL, "108556109.ogg"},
	{0xC536C593E8E34BDFULL, "208206957.ogg"},
	{0xC536E55A0D85D0D7ULL, "947203051.ogg"},
	{0xC5375206A4C4BD1AULL, "6124726_m.fxe"},
	{0xC537ADB0E3E79E55ULL, "324668387.ogg"},
	{0xC537C66B487C2335ULL, "aip_cartathug_base.gda"},
	{0xC537CBBC1F02537DULL, "6000971_m.fxe"},
	{0xC538569150780EB5ULL, "vxht_0_hightown_9zk_0_0.vlm"},
	{0xC538A4037759BA81ULL, "vxlt_0_lowtown_docks_a3_br2_0_0.vlm"},
	{0xC538DECB77A6FBD9ULL, "vxca_caverns1_j4z_0_0.vlm"},
	{0xC5393FD0D357809DULL, "328441775.ogg"},
	{0xC5395DEE5C26E2C5ULL, "vxw_woundedcoast_fv_0_0.vlm"},
	{0xC5396E52603E9521ULL, "vxlt_warehouse_n_vq_0_0.vlm"},
	{0xC53994BC6F53FF86ULL, "159459854.ogg"},
	{0xC53995AA746E2482ULL, "6160890_m.fxe"},
	{0xC539A363D1613224ULL, "isa300im_office_key.uti"},
	{0xC539BBE12D949CFAULL, "vxw_woundedcoast_ux_0_0.vlm"},
	{0xC539D81DA5847B15ULL, "hlac_lt_darktown_-1_79y.dds"},
	{0xC53A02156C3B7D5EULL, "gen_im_trs_bst_spd_silk_01.uti"},
	{0xC53A0C0339AA49EBULL, "dae311st_velasco_warehouse.stg"},
	{0xC53A442483E55151ULL, "286712050.ogg"},
	{0xC53AA9B9EDC7C836ULL, "gen00fl_anders6128342.cl"},
	{0xC53AAC4A224ED837ULL, "isa211_qunari_ambush.bnk"},
	{0xC53AB1D28EBC3285ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_loop.ani"},
	{0xC53B24ECA0D050B6ULL, "6042246_m.fxe"},
	{0xC53B70AB243E6530ULL, "6033169_m.fxe"},
	{0xC53C49350357730DULL, "vxht_mansion2b_1w4_0_0.vlm"},
	{0xC53C6FE38D6C7A6EULL, "b_gr_smo_01.mao"},
	{0xC53CA603E62ABC82ULL, "265682120.ogg"},
	{0xC53D6D9988897B3FULL, "vxw_encounter1_kk_0_0.vlm"},
	{0xC53D847E53A505AEULL, "vrc221_hugin6042563.cl"},
	{0xC53E8E4A22528E4FULL, "isa211_qunari_ambush.cnv"},
	{0xC53EF51D55BA7259ULL, "gen00fl_aveline6115281.cl"},
	{0xC53EFAD6646860B8ULL, "vxlt_0_lowtown_n_h2d_0_0.vlm"},
	{0xC53FE2001C553DA4ULL, "173980493.ogg"},
	{0xC5402D606BECC148ULL, "zz_dae_debug6047592.cl"},
	{0xC541CD57967A10B6ULL, "mcr371_orsino6083331.cl"},
	{0xC5429E18B48F7519ULL, "vxca_caverns2_7l0_0_0.vlm"},
	{0xC543789E95B9899CULL, "zz_one_debug6051526.cl"},
	{0xC5439276FAE5E99FULL, "vxca_caverns1_hue_0_0.vlm"},
	{0xC543B44ED9711A0FULL, "vxlt_undercityserialkiller_1nh_0_0.vlm"},
	{0xC5441014EE0827B4ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_wipe_eyes.ani"},
	{0xC54473CB996FCF6EULL, "vxw_bonepit_20j_0_0.vlm"},
	{0xC544955A599FD626ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikia.phy"},
	{0xC545451A428C909BULL, "123769469.ogg"},
	{0xC545616C41E63382ULL, "6025094_m.fxe"},
	{0xC5458603AA83128CULL, "vxca_cave_1dd_0_0.vlm"},
	{0xC54586866A1A6C17ULL, "vxht_mansion_n_2b5_0_0.vlm"},
	{0xC545DA6113FFAF1DULL, "hlac_lt_0_lowtown_docks_-65539_6z.dds"},
	{0xC54609ADBB1A28BEULL, "follower_banter6163894.cl"},
	{0xC54625C0C5C9F8C2ULL, "gut_geo.mao"},
	{0xC54648A139577A5BULL, "511269634.ogg"},
	{0xC54750A8699D1459ULL, "vxca_caverns1_j1y_0_0.vlm"},
	{0xC5475CDB1DC68A6BULL, "hlac_dr_slavers_3207_5x.dds"},
	{0xC5477EE39BB7485DULL, "901251063.ogg"},
	{0xC547F31C972093ABULL, "vxlt_0_lowtown_a2_e10_0_0.vlm"},
	{0xC548078766643D02ULL, "6165790_m.fxe"},
	{0xC5480BE2214CE369ULL, "hlac_lt_0_lowtown_docks_a3_-3_4x.dds"},
	{0xC5484F9B6770226CULL, "585770998.ogg"},
	{0xC548978852EE0156ULL, "59806149.ogg"},
	{0xC548A199742C6BC3ULL, "6119235_m.fxe"},
	{0xC548B33CA8D919A4ULL, "6080211_m.fxe"},
	{0xC548B824ACBBECFDULL, "6202154_f.fxe"},
	{0xC548BBAA8243F62CULL, "llac_w_bonepitdestroyed_a3_262145_0.dds"},
	{0xC549184A01E407DDULL, "276811442.ogg"},
	{0xC5491F66E0E4CCAEULL, "317770692.ogg"},
	{0xC54A51AD5B49FB0EULL, "vxlt_undercityencounter_18k_0_0.vlm"},
	{0xC54A63AA713C33DAULL, "w_bonepitmining_a2_167_0.mmh"},
	{0xC54A67AA713C3A11ULL, "w_bonepitmining_a2_167_0.mao"},
	{0xC54A75AA713C5220ULL, "w_bonepitmining_a2_167_0.msh"},
	{0xC54AA1A42BF97BF1ULL, "6175770_m.fxe"},
	{0xC54ABDD7F6025D24ULL, "plt_ssf000pt_main.nss"},
	{0xC54AEB27BB4D548FULL, "l_mcr370ar_gallows_nt.wbd"},
	{0xC54B0D14EE0E4116ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_wipe_eyes.gad"},
	{0xC54B13AFACBBA003ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v2.evt"},
	{0xC54BC6C68C94A93CULL, "6169635_m.fxe"},
	{0xC54CAAA7FBB14B9CULL, "tow_rock.mao"},
	{0xC54CD243D398ABEFULL, "vxht_0_hightown_a3_c5w_0_0.vlm"},
	{0xC54CE606709B41BEULL, "hm_gencr_raider.mor"},
	{0xC54D0DC25FD240FFULL, "vxlt_warehouse_12k_0_0.vlm"},
	{0xC54D75E9E09849D5ULL, "257906334.ogg"},
	{0xC54D8E657BD70DA4ULL, "vxca_cavern1_haunted_iod_0_0.vlm"},
	{0xC54DA1A40BC879ADULL, "isa311_isabela6092152.cl"},
	{0xC54E07AAC2E5BD86ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_dm01.phy"},
	{0xC54E2536C28101ECULL, "500789357.ogg"},
	{0xC54EC5297A50141EULL, "zz_fen_debug6134372.cl"},
	{0xC54EF2099CE66190ULL, "vxw_encounter2_dq_0_1.vlm"},
	{0xC54F233B56A84DD2ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg106a_0d.dds"},
	{0xC54FAEDBC78F5FDBULL, "hlac_lt_0_lowtown_-131071_0z.dds"},
	{0xC5500A5E1E827D23ULL, "6103269_m.fxe"},
	{0xC5509C818904BAD5ULL, "6181756_f.fxe"},
	{0xC550B514EBA53518ULL, "vxht_keep_30n_0_0.vlm"},
	{0xC5510FF5A6858D99ULL, "vxca_caverns2_7a7_0_0.vlm"},
	{0xC5516F47025D0DBAULL, "6083047_m.fxe"},
	{0xC551943EACBB9C74ULL, "6068341_m.fxe"},
	{0xC551ECAFACC17C84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v2.gad"},
	{0xC551ED5DB1E37EB9ULL, "438007768.ogg"},
	{0xC552010E55C6CDA8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basicnb.mmh"},
	{0xC5520D0E55C6E24BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basicnb.mao"},
	{0xC5521F0E55C700D2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0dd_basicnb.msh"},
	{0xC5522814EE148E75ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_wipe_eyes.evt"},
	{0xC5527CB0C76E2249ULL, "gen00fl_anders6110065.cl"},
	{0xC552B7E7A9558E7EULL, "vxht_keep_6gd_0_0.vlm"},
	{0xC553AD044EE62959ULL, "671781405.ogg"},
	{0xC553E6F4FE1B6D10ULL, "pssmstatic.psh.fxo"},
	{0xC553FD15BC9385EAULL, "6050295_f.fxe"},
	{0xC5543E720F9478BFULL, "310973907.ogg"},
	{0xC5552C6A91A067AEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowb_n.dds"},
	{0xC5556895FBB0D4E9ULL, "193803290.ogg"},
	{0xC555951B349BB1BCULL, "6086203_f.fxe"},
	{0xC555EC2A8499D944ULL, "llac_ht_0_hightown_a3_131071_1.dds"},
	{0xC55641A8401313C5ULL, "vxw_woundedcoast_xt_0_0.vlm"},
	{0xC55690673091DDD1ULL, "6046883_m.fxe"},
	{0xC556EB3BD9D0208EULL, "follower_banter6172269.cl"},
	{0xC557137666228443ULL, "6114535_m.fxe"},
	{0xC5573C482B2ED7C8ULL, "gen00fl_isabela_init6175712.cl"},
	{0xC5576179434DA1D2ULL, "405662171.ogg"},
	{0xC557B418D67E5C65ULL, "887401416.ogg"},
	{0xC55803F214DD0FABULL, "vxw_encounter3_e7_0_0.vlm"},
	{0xC5587A33F7F77DF8ULL, "hlac_w_sundermount_-2_0x.dds"},
	{0xC558FBAFACC7B562ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v2.ani"},
	{0xC5593A7D0683BFB7ULL, "llac_lt_0_lowtown_a2_n_c_131071_4.dds"},
	{0xC5598A9BFB9BB0B0ULL, "w_blightlands_2_0.rml"},
	{0xC559D7A63C9E1799ULL, "genip_node_i1.utp"},
	{0xC55A1C1FAEE1D66CULL, "mrl350ip_grave_c.utp"},
	{0xC55A5CC030D5D75EULL, "crt_dragonlings.bnk"},
	{0xC55AD5A877F0A19BULL, "vxlt_0_lowtown_a2_n_fe8_0_0.vlm"},
	{0xC55B204ED5052F2DULL, "vxht_mansion_n_62s_0_0.vlm"},
	{0xC55B3909ACAACB00ULL, "6084241_f.fxe"},
	{0xC55B3E37454B2FABULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.impa.evt"},
	{0xC55B8D19EBB8963BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pleading.evt"},
	{0xC55BA450C6475422ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.mmh"},
	{0xC55BC9051F500F56ULL, "vxdr_primevaltomb_75n_0_0.vlm"},
	{0xC55C12083F69C6CFULL, "vxdr_0_deeproads_f4o_0_0.vlm"},
	{0xC55C1DA2A997AEA5ULL, "6116452_m.fxe"},
	{0xC55CAC8AE597F044ULL, "dae000ip_crowd_ff7_talk11.utp"},
	{0xC55CE3493B188D57ULL, "vxca_caverns1_mines_j26_0_1.vlm"},
	{0xC55D4DFCAC7985BAULL, "vxdr_primevaltomb_7ub_0_0.vlm"},
	{0xC55D93C97446BF71ULL, "vxlt_darktown_472_0_0.vlm"},
	{0xC55DBF31FA76EB6AULL, "dre000_stroud6036695.cl"},
	{0xC55DCA7250556DD7ULL, "hlac_ca_caverns1_mines_24425_13z.dds"},
	{0xC55E376917A8078DULL, "lt_ruinedhouse_n_3567_ll.mmh"},
	{0xC55E517DFAFD2A6CULL, "203438468.ogg"},
	{0xC55E66A595E25DBBULL, "lgt260ip_at_tomecave.utp"},
	{0xC55E6EBA8AA880A2ULL, "1018300096.ogg"},
	{0xC55E7C3B107895A3ULL, "6104369_m.fxe"},
	{0xC55EA2B96EC05D1EULL, "6024546_f.fxe"},
	{0xC55EE030845FBED7ULL, "235432578.ogg"},
	{0xC55F12C2F0F9725FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1_tool.phy"},
	{0xC55F1AA97EABD978ULL, "6118330_m.fxe"},
	{0xC55F614EC486803DULL, "vxca_caverns2_82t_0_1.vlm"},
	{0xC55FADB5053E20DDULL, "hlac_lt_0_lowtown_a3_n_c_-196609_1y.dds"},
	{0xC5602870786FBA49ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookstand_n.dds"},
	{0xC560800562D9F288ULL, "one100_solivitus.cnv"},
	{0xC560C1823F70A6C5ULL, "700106443.ogg"},
	{0xC5613B46A190FF86ULL, "42324728.ogg"},
	{0xC5616F65526AED54ULL, "6177860_f.fxe"},
	{0xC5618619EBBCF67CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pleading.gad"},
	{0xC5619A8F0940AEFEULL, "578567815.ogg"},
	{0xC56210B3C385E309ULL, "727568161.ogg"},
	{0xC56212FCFA07298EULL, "lt_0_lowtown_mmap.dds"},
	{0xC5621356E7F008DAULL, "mcr370_meredith6085886.cl"},
	{0xC56248AD37C20826ULL, "vxlt_0_lowtown_docks_bf2_0_0.vlm"},
	{0xC56257374551796CULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.impa.gad"},
	{0xC5625C18D99B0BB3ULL, "gen00fl_anders6111235.cl"},
	{0xC562CF02B1F6FD04ULL, "vxca_caverns1_fna_0_0.vlm"},
	{0xC562D05CFAD13D04ULL, "zz_dae_debug6029488.cl"},
	{0xC5631543DBC52D96ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose1left.anb"},
	{0xC5632049D2752E2DULL, "809680705.ogg"},
	{0xC563B1243167581EULL, "6125466_m.fxe"},
	{0xC563C13F7F688A01ULL, "6129348_m.fxe"},
	{0xC564000859716985ULL, "674919573.ogg"},
	{0xC5641461FCF62438ULL, "vxlt_gallowstemplar_n_430_0_0.vlm"},
	{0xC5641C418256541DULL, "vxht_0_hightown_5y2_0_0.vlm"},
	{0xC5643619EA8C696AULL, "6026217_m.fxe"},
	{0xC5644EAA1BA32328ULL, "hlac_ht_mansion2b_818_43y.dds"},
	{0xC564945C7CDDBCD9ULL, "hm_mer_smith.mor"},
	{0xC5649F388B759ABDULL, "vxlt_gamlenshouse_n_5z_0_0.vlm"},
	{0xC564A20562DE15F0ULL, "one100_solivitus.bnk"},
	{0xC564B38532092C45ULL, "vxw_woundedcoast_us_0_0.vlm"},
	{0xC56510252FE4F0FEULL, "942608700.ogg"},
	{0xC5654D2B86C618C0ULL, "274541195.ogg"},
	{0xC5657B80ACC843D4ULL, "wm_tbe_city_day.map"},
	{0xC565A416BDD53726ULL, "fen121ar_mansion_nt.are"},
	{0xC56620742F491260ULL, "581202534.ogg"},
	{0xC566439FB7A8B197ULL, "ave200_donnic6111055.cl"},
	{0xC567030C4A81014BULL, "hlac_lt_0_lowtown_-65536_5y.dds"},
	{0xC56735673F362C0EULL, "vxlt_undercityencounter_12m_0_0.vlm"},
	{0xC5678C6362FE89A3ULL, "art\\characters\\creatures\\dragon\\animation\\eletr.ani"},
	{0xC567AD88CCD1475EULL, "vxlt_gallowsprison_n_64q_0_0.vlm"},
	{0xC567C0C2C44155BBULL, "6126475_m.fxe"},
	{0xC5681DBE687F39C0ULL, "vxdr_0_deeproads_gfz_0_1.vlm"},
	{0xC56825E82EACB9E4ULL, "vxlt_gallowstemplar_3ir_0_0.vlm"},
	{0xC5686ECE96DBEAA1ULL, "6051645_m.fxe"},
	{0xC5688519EBC3137AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pleading.ani"},
	{0xC5688952A8EAE323ULL, "vxdr_primevalentrance_27l_0_0.vlm"},
	{0xC5691B17BD66EC87ULL, "6142558_m.fxe"},
	{0xC56956374557966AULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.impa.ani"},
	{0xC5699A04E985EC5DULL, "lt_0_lowtown_docks_a3_n_water_16558.mao"},
	{0xC5699E04E985F296ULL, "lt_0_lowtown_docks_a3_n_water_16558.mmh"},
	{0xC569A804E9860394ULL, "lt_0_lowtown_docks_a3_n_water_16558.msh"},
	{0xC56A721F5FC33EA4ULL, "vxca_cavern1_haunted_iaj_0_0.vlm"},
	{0xC56B38BEF49785ECULL, "6132756_m.fxe"},
	{0xC56B6C2B3AE946B1ULL, "vxlt_blackemporium_im_0_0.vlm"},
	{0xC56BC149C2B387D5ULL, "plt_cod_ltr_raiders_wendell.plo"},
	{0xC56BE403D60BAC9FULL, "vxht_keep_n_4mr_0_0.vlm"},
	{0xC56BEC23172610E1ULL, "24026104.ogg"},
	{0xC56C05A6663CD7F1ULL, "vxlt_darktown_421_0_0.vlm"},
	{0xC56C941B785BC586ULL, "vxht_keep_6ev_0_0.vlm"},
	{0xC56CA5E66CF188EAULL, "bad_targeted.mao"},
	{0xC56CBED7F61F451FULL, "plt_ssf000pt_main.plo"},
	{0xC56CE0E7B10D4F89ULL, "635943801.ogg"},
	{0xC56CFDDA2212B146ULL, "zz_dae_debug6206731.cl"},
	{0xC56D86DF4132030EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keepthin.msh"},
	{0xC56D889C7850A000ULL, "6071209_f.fxe"},
	{0xC56D94DF41321B5FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keepthin.mao"},
	{0xC56DA0DF41322FFCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keepthin.mmh"},
	{0xC56DC0D0081E0B44ULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_c.mmh"},
	{0xC56F1B71C9FA2B1FULL, "w_bonepit_33_0.rml"},
	{0xC56F474328CFC4DCULL, "zz_one_debug6077401.cl"},
	{0xC56F7034F547872CULL, "6093555_f.fxe"},
	{0xC56F908C331ADF5CULL, "gen_im_trs_uni_bks_01.uti"},
	{0xC56FF83CAEA5FF1BULL, "540361503.ogg"},
	{0xC57013C2F107E764ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1_tool.mmh"},
	{0xC57017C2F107EDB7ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1_tool.mao"},
	{0xC57029C2F1080C06ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1_tool.msh"},
	{0xC5702CF76D935116ULL, "vxca_caverns2prc_9l4_0_0.vlm"},
	{0xC57033A66F31D02DULL, "936453627.ogg"},
	{0xC57082AA715D37E5ULL, "w_bonepitmining_a2_167_0.phy"},
	{0xC5708B86802FC7E1ULL, "423696298.ogg"},
	{0xC570C0480D47BE5BULL, "6149641_m.fxe"},
	{0xC570DC725C9462F4ULL, "qun221_petrice6075985.cl"},
	{0xC57121A4C8270AC7ULL, "hlac_w_bonepitmining_a2_65541_0x.dds"},
	{0xC571CF6EC10D76CEULL, "t3_hey_br2.tnt"},
	{0xC571F2C46360E61FULL, "vxw_bonepitdestroyed_a3_1wu_0_0.vlm"},
	{0xC5722487436FF033ULL, "6119096_f.fxe"},
	{0xC572D872043293B6ULL, "6205946_m.fxe"},
	{0xC5730A692D5A57DEULL, "kl.cut"},
	{0xC573256223FF44C5ULL, "vxw_woundedcoast_rr_0_0.vlm"},
	{0xC5735DBC9AD6F722ULL, "vxca_caverns2_91s_0_1.vlm"},
	{0xC573ABEE9033C56CULL, "vxlt_hangedman_1xr_0_0.vlm"},
	{0xC573E826D46EC716ULL, "vxht_keep_5dj_0_0.vlm"},
	{0xC57446F91064CE03ULL, "art\\characters\\creatures\\strider\\animation\\walkfront.ani"},
	{0xC574712451C8D559ULL, "vxht_mansion_n_57o_0_0.vlm"},
	{0xC57480F2B17401F3ULL, "6161434_m.fxe"},
	{0xC574B55E37A309F3ULL, "vxlt_0_lowtown_docks_n_cup_0_0.vlm"},
	{0xC574CA2063F15210ULL, "vxw_sundermount_34j_0_0.vlm"},
	{0xC5751A5581F241E6ULL, "707474254.ogg"},
	{0xC5752E8D901F066BULL, "6095070_m.fxe"},
	{0xC575435EA3819B9EULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_p.impa.evt"},
	{0xC575672D62370AA6ULL, "hlac_ca_caverns2_5802_61x.dds"},
	{0xC576EF3949531CC0ULL, "vxlt_gallowstemplar_f_5hh_0_0.vlm"},
	{0xC57704B4A9D82E02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_up.gad"},
	{0xC5770BC532B8342FULL, "6153471_m.fxe"},
	{0xC5773C7F50991B78ULL, "6135616_m.fxe"},
	{0xC5782A151DC89CE6ULL, "vxw_sundermount_2w0_0_0.vlm"},
	{0xC5787727F1585D06ULL, "6118292_f.fxe"},
	{0xC57899EB128AB955ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_0x.dds"},
	{0xC579145E7127761DULL, "6098216_f.fxe"},
	{0xC579C21F34A66C8FULL, "zz_start_dre6080512.cl"},
	{0xC579DA05FE6547D1ULL, "125658765.ogg"},
	{0xC57A2BDA604EFEE9ULL, "6162723_m.fxe"},
	{0xC57A77835832F071ULL, "vxlt_darktown_490_0_0.vlm"},
	{0xC57B167D3082D4AAULL, "plt_lgt220pt_mag_meeting.plo"},
	{0xC57B55D849E69787ULL, "hlac_lt_undercityserialkiller_-1_52x.dds"},
	{0xC57B5A5EA3862EC9ULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_p.impa.gad"},
	{0xC57BC7798AD662CFULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be04a.phy"},
	{0xC57BFA0D9D95619AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comc_0d.dds"},
	{0xC57C3483231348B8ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_groundfog.mmh"},
	{0xC57CA36363110C8DULL, "art\\characters\\creatures\\dragon\\animation\\eletr.gad"},
	{0xC57D22DD16BB0D56ULL, "gen_im_trs_slv_lgt_cht_r1.uti"},
	{0xC57D234E7B7B9030ULL, "6053875_m.fxe"},
	{0xC57D3EA95F703665ULL, "6051670_m.fxe"},
	{0xC57E27B4A9DE88F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_up.evt"},
	{0xC57EB91A4C6FEBE2ULL, "gen00fl_carver6104843.cl"},
	{0xC57EBAEAFED83A3EULL, "40392675.ogg"},
	{0xC57EFA0B534F14D4ULL, "6132815_m.fxe"},
	{0xC57F54FFECBCFC7DULL, "vxht_0_hightown_a3_aig_0_0.vlm"},
	{0xC57F9F9A635BD12CULL, "6076569_f.fxe"},
	{0xC57FC9C78B1AFE8EULL, "vxlt_0_lowtown_docks_a3_d2w_0_0.vlm"},
	{0xC580090AACFC0910ULL, "6035209_f.fxe"},
	{0xC580288EBEDB0124ULL, "vxca_caverns2prc_9zp_0_0.vlm"},
	{0xC5805E04EDE3CFA8ULL, "6177827_m.fxe"},
	{0xC580AA9BDE30D6D0ULL, "6034432_m.fxe"},
	{0xC58132D7FCECB34EULL, "mag110_vincento6051752.cl"},
	{0xC58159E8879078F8ULL, "framebuffereffects.mat"},
	{0xC5815EF9106F81FAULL, "art\\characters\\creatures\\strider\\animation\\walkfront.evt"},
	{0xC5817FD8B0BBB6BFULL, "vxlt_hangedman_1vs_0_0.vlm"},
	{0xC58189B37407AF99ULL, "gen00fl_isabela6090433.cl"},
	{0xC581AFFF6D4E5EC4ULL, "vxlt_0_lowtown_a2_n_c_fi5_0_0.vlm"},
	{0xC581E7CA186A1FC5ULL, "6123373_m.fxe"},
	{0xC5822B5EA38BFD17ULL, "art\\vfx\\combat\\vfx_m_frc_forcecrush_p.impa.ani"},
	{0xC5823C4AADA11F36ULL, "342907034.ogg"},
	{0xC58259DC2B826D9EULL, "hlac_w_sundermount_-2_1y.dds"},
	{0xC582C010DAFE5C15ULL, "6127222_f.fxe"},
	{0xC582CA758F9D0DF9ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_reva_0s.dds"},
	{0xC5833BFD55E76A90ULL, "vxw_sundermount_33k_0_0.vlm"},
	{0xC583CD6B93D68BC1ULL, "vxdr_primevaltomb_913_0_0.vlm"},
	{0xC584A350C66A57C9ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.anb"},
	{0xC584A47CA82D86BBULL, "6128675_m.fxe"},
	{0xC584BFDA51340C62ULL, "w_blightlands_153_0.msh"},
	{0xC584C1DA51340F18ULL, "w_blightlands_153_0.mmh"},
	{0xC584CDDA513423BBULL, "w_blightlands_153_0.mao"},
	{0xC584DC90FA75E7ECULL, "vxht_mansion_n_5zq_0_0.vlm"},
	{0xC5855FD05BD9A869ULL, "dae000ip_crowd_ff6_talk13.utp"},
	{0xC585821866580259ULL, "184390758.ogg"},
	{0xC585A074A67C045BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_meda_0n.dds"},
	{0xC5863A1A8CB60277ULL, "287471583.ogg"},
	{0xC5865CCC1F3476A8ULL, "hlac_lt_backalley_n_65536_4z.dds"},
	{0xC586F7E4CD6AFBE6ULL, "244454427.ogg"},
	{0xC5872A1B595ADC29ULL, "weapon_rendertodepth_p0.vsh.fxo.dx11"},
	{0xC58822BEBC4F4367ULL, "6131101_m.fxe"},
	{0xC58850DBDF1C2DE5ULL, "133938998.ogg"},
	{0xC5885DF910759EEDULL, "art\\characters\\creatures\\strider\\animation\\walkfront.gad"},
	{0xC5889D75BB725684ULL, "dae000_bodhan6146871.cl"},
	{0xC588A02FA3CB04BEULL, "vxca_caverns2_8oy_0_0.vlm"},
	{0xC588A5A3E9986FD3ULL, "zz_approval_debug6001820.cl"},
	{0xC588BB5CA7200743ULL, "gen00fl_fenris6113947.cl"},
	{0xC58977D55C47F686ULL, "vxht_keep_6kx_0_0.vlm"},
	{0xC5899C8A4550D709ULL, "890815840.ogg"},
	{0xC589D52D016FCED1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8man.mmh"},
	{0xC589E12D016FE372ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8man.mao"},
	{0xC589EF2D016FFA8FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8man.msh"},
	{0xC58A6C54CD961A9CULL, "hlac_lt_darktown_-1_25y.dds"},
	{0xC58A7982BEF7360FULL, "vxht_bartrand_n_701_0_0.vlm"},
	{0xC58AAB16BDF45F11ULL, "fen121ar_mansion_nt.lst"},
	{0xC58AC3C4237B88D0ULL, "91007478.ogg"},
	{0xC58B0527BB836565ULL, "l_mcr370ar_gallows_nt.bnk"},
	{0xC58B82682204624DULL, "al_rogue_t05.gda"},
	{0xC58BED9F933F948FULL, "vxlt_undercityserialkiller_1ww_0_0.vlm"},
	{0xC58BFADFDB4F786FULL, "vxlt_0_lowtown_a2_n_df3_0_0.vlm"},
	{0xC58C13449FD44FC9ULL, "316268717.ogg"},
	{0xC58C1A1AE796F329ULL, "llac_ht_0_hightown_a3_-65537_2.dds"},
	{0xC58C2FB4A9EAD3E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_up.ani"},
	{0xC58CA3A91631E93AULL, "llac_ht_chantry_n_2534_51.dds"},
	{0xC58CCAD9A438439AULL, "vxlt_0_lowtown_a2_n_c_eqx_0_0.vlm"},
	{0xC58D61ED093BA820ULL, "353512454.ogg"},
	{0xC58D8746EDC3AC9AULL, "vxlt_warehouse_1br_0_0.vlm"},
	{0xC58E3A31E5F08DABULL, "hm_guard1.mor"},
	{0xC58EC8DA513C6FA3ULL, "w_blightlands_153_0.phy"},
	{0xC58EE4321D45F2F8ULL, "213576457.ogg"},
	{0xC58EFF6837C50B09ULL, "6050408_m.fxe"},
	{0xC58F059FF209F8F7ULL, "vxdr_0_deeproads_coy_0_2.vlm"},
	{0xC58F06B76B79AC14ULL, "133533433.ogg"},
	{0xC58F36DF5F77B16AULL, "c_deepstkra_0t.dds"},
	{0xC58F4EB0B611BD99ULL, "vxlt_gamlenshouse_n_bw_0_0.vlm"},
	{0xC58F95545C4E02D9ULL, "6170777_m.fxe"},
	{0xC58FBE6A53A8C7CEULL, "isa211_isabela6202108.cl"},
	{0xC58FF8D07BA812E5ULL, "357166203.ogg"},
	{0xC590039EE11BC5D7ULL, "character_character_p156.psh.fxo.dx11.assembly"},
	{0xC5902CFC05A5DA68ULL, "vxlt_undercityencounter_1la_0_0.vlm"},
	{0xC59072CA07315F4AULL, "6087708_m.fxe"},
	{0xC5908FA8741FF66CULL, "vxlt_hangedman_1bp_0_0.vlm"},
	{0xC590F3464DE9C793ULL, "dr_0_deeproads_5839_ll.mmh"},
	{0xC590FA27CA22ECFFULL, "plt_vrc321pt_haunted.nss"},
	{0xC5913E986B79EACEULL, "6043177_m.fxe"},
	{0xC59145AF94DC1C7BULL, "vxht_0_hightown_n_b7f_0_0.vlm"},
	{0xC591AC87CB1DFF02ULL, "qun221_viscount6069026.cl"},
	{0xC591C316BDFAA6A8ULL, "fen121ar_mansion_nt.ncs"},
	{0xC591C48750E9CE72ULL, "373333857.ogg"},
	{0xC5923018CFB8225DULL, "hlac_w_bonepit_5_0z.dds"},
	{0xC59265E4DEBEB930ULL, "ser221_party_comments6163595.cl"},
	{0xC592F44B1D0A67FDULL, "mag291_torpor6026948.cl"},
	{0xC59306F335C57BF9ULL, "6041588_f.fxe"},
	{0xC5937FDF4152C6C7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1pp_keepthin.phy"},
	{0xC594118CFCE3EBC1ULL, "vxw_encounter2_camp_8a_0_0.vlm"},
	{0xC5941C9A51B8FEADULL, "mag131_taraneh6088739.cl"},
	{0xC595620B2D259AE5ULL, "6179352_m.fxe"},
	{0xC595DCD0202AF7BFULL, "bd_smok.mao"},
	{0xC5961058762BF103ULL, "397169886.ogg"},
	{0xC59614F3FF13AE41ULL, "follower_banter6166711.cl"},
	{0xC596ACD9EA6D22DAULL, "759621471.ogg"},
	{0xC596C3D00841153BULL, "art\\vfx\\combat\\vfx_m_spi_dispelmagic_c.anb"},
	{0xC5977E2BFC0010ACULL, "563848017.ogg"},
	{0xC597860563093FC7ULL, "one100_solivitus.stg"},
	{0xC5981ECBDCE9A229ULL, "6054743_m.fxe"},
	{0xC598FB787778765CULL, "6127606_m.fxe"},
	{0xC59931F308C318C4ULL, "12705478.ogg"},
	{0xC5997464A417E407ULL, "vxdr_0_deeproads_ew1_0_0.vlm"},
	{0xC599C5226A9EDD68ULL, "gen00fl_aveline6116647.cl"},
	{0xC59A0E41209B9E19ULL, "6088290_m.fxe"},
	{0xC59A2EC0310B6D64ULL, "crt_dragonlings.wbd"},
	{0xC59A382EA771D0C8ULL, "w_blightlands_117_0.mmh"},
	{0xC59A442EA771E56BULL, "w_blightlands_117_0.mao"},
	{0xC59A562EA77203F2ULL, "w_blightlands_117_0.msh"},
	{0xC59B3F1F7CBBD1FBULL, "vxht_mansion_n_66y_0_0.vlm"},
	{0xC59CC5F8A3BD74B7ULL, "6187948_m.fxe"},
	{0xC59D485B39432F9FULL, "vxw_bonepitdestroyed_a3_1fv_0_0.vlm"},
	{0xC59D59DFDB56557FULL, "804832855.ogg"},
	{0xC59D76CC8D27A8E6ULL, "715314964.ogg"},
	{0xC59DC21C156B5157ULL, "562890915.ogg"},
	{0xC59DF9B02B127C7AULL, "vfx_distortionmask_p4.vsh.fxo.dx11.assembly"},
	{0xC59E2E8C325E4EAFULL, "6087781_m.fxe"},
	{0xC59E3136A01C761AULL, "616835948.ogg"},
	{0xC59E9B7D17DA2E14ULL, "951540923.ogg"},
	{0xC59E9C82481CEB67ULL, "86855328.ogg"},
	{0xC59EB8B8618A1863ULL, "art\\characters\\creatures\\desiredemon\\animation\\p_tw_v1.ani"},
	{0xC59EF2AAC10906A5ULL, "hlac_lt_0_lowtown_a2_n_131073_0z.dds"},
	{0xC5A096F688EB7FD0ULL, "6142496_m.fxe"},
	{0xC5A10979E73DD393ULL, "t3_shk_h01.tnt"},
	{0xC5A240A7AD6E7D3FULL, "hlac_lt_undercitydungeon_-1_51x.dds"},
	{0xC5A2F267E43B9AEEULL, "w_woundedcoast_209_0.msh"},
	{0xC5A30067E43BB23FULL, "w_woundedcoast_209_0.mao"},
	{0xC5A30C67E43BC6DCULL, "w_woundedcoast_209_0.mmh"},
	{0xC5A349E5A59C4D5AULL, "434616824.ogg"},
	{0xC5A40C0ED85DEF9BULL, "hlac_lt_0_lowtown_a2_n_0_2y.dds"},
	{0xC5A46F290CE5551FULL, "6111346_f.fxe"},
	{0xC5A4ED27CA337F10ULL, "plt_vrc321pt_haunted.plo"},
	{0xC5A53EE890EDB299ULL, "zz_otl_debug6060601.cl"},
	{0xC5A586ED33E27570ULL, "6030475_f.fxe"},
	{0xC5A5D6379BA44618ULL, "dae000_orana6149075.cl"},
	{0xC5A5DF635FA08474ULL, "418243048.ogg"},
	{0xC5A622FCDB9B8027ULL, "vxht_brothel_n_617_0_0.vlm"},
	{0xC5A654F2C924E419ULL, "plt_cod_ltr_conspiracy_cold.plo"},
	{0xC5A710E070615A31ULL, "484425448.ogg"},
	{0xC5A7599CD238022BULL, "vxht_0_hightown_n_bj1_0_0.vlm"},
	{0xC5A78C04BEBFD77AULL, "6053569_m.fxe"},
	{0xC5A7BD10A43444E8ULL, "genip_pp_drk_one_shot.utp"},
	{0xC5A7C4195D94127AULL, "6024563_m.fxe"},
	{0xC5A84372FDB77090ULL, "llac_w_bonepitdestroyed_a3_327686_0.dds"},
	{0xC5A84B0DACDFF4BEULL, "zz_rdrpt_debug.ncs"},
	{0xC5A84D7465237628ULL, "gen_im_trs_tmp_hvy_glv_r1.uti"},
	{0xC5A850E8973BAFBEULL, "707935162.ogg"},
	{0xC5A8A798F96D7A18ULL, "cod_bks_kirkwall_hist_3.utp"},
	{0xC5A908E707592275ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.anb"},
	{0xC5A9719C32ECA818ULL, "vxdr_0_deeproads_9ih_0_0.vlm"},
	{0xC5A9AEBAF2ABD551ULL, "ser120_ghyslain6104218.cl"},
	{0xC5AA190EC2B68179ULL, "6018787_m.fxe"},
	{0xC5AA1EAF0B4BFDF6ULL, "6154729_m.fxe"},
	{0xC5AAD8D19B71FA7BULL, "w_woundedcoast_86_0.msh"},
	{0xC5AAEAD19B7218C6ULL, "w_woundedcoast_86_0.mao"},
	{0xC5AAF6D19B722D65ULL, "w_woundedcoast_86_0.mmh"},
	{0xC5AB1F7D30AC07A1ULL, "plt_lgt220pt_mag_meeting.nss"},
	{0xC5AB910D1F4F8518ULL, "306602614.ogg"},
	{0xC5ABD0B86194CB5AULL, "art\\characters\\creatures\\desiredemon\\animation\\p_tw_v1.evt"},
	{0xC5AC15E6810B1539ULL, "6208216_m.fxe"},
	{0xC5AC89ACF77C24A9ULL, "vxw_blightlands_hb_0_0.vlm"},
	{0xC5ACEFE1BDD7CB69ULL, "mag291_demon_marethari6028840.cl"},
	{0xC5AD207CCDE10947ULL, "follower_banter6173063.cl"},
	{0xC5AD3B98CD9CA083ULL, "6171035_m.fxe"},
	{0xC5ADAEC48640C26FULL, "291452475.ogg"},
	{0xC5ADAF9A9C472916ULL, "vxht_keep_5xn_0_0.vlm"},
	{0xC5ADCE5DAD31A988ULL, "vxlt_0_lowtown_docks_c0c_0_0.vlm"},
	{0xC5ADD8076158FF97ULL, "zz_vrc_debug6093919.cl"},
	{0xC5AE1863470D23BEULL, "z_cloud03.1.dds"},
	{0xC5AE63A8B8BF6ECBULL, "vxlt_0_lowtown_f53_0_0.vlm"},
	{0xC5AE74CA6E27112DULL, "6202355_m.fxe"},
	{0xC5AE789BFBE3D22DULL, "w_blightlands_2_0.msh"},
	{0xC5AE869BFBE3EA60ULL, "w_blightlands_2_0.mao"},
	{0xC5AE8A9BFBE3F0B3ULL, "w_blightlands_2_0.mmh"},
	{0xC5AE9CA97EDC3071ULL, "337787585.ogg"},
	{0xC5AECB937177BDD9ULL, "6033996_f.fxe"},
	{0xC5AF0877449A3B9BULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.cess.evt"},
	{0xC5AF698DAFE64F55ULL, "68678581.ogg"},
	{0xC5AFDA4E430B58E4ULL, "1038061777.ogg"},
	{0xC5B05F38E87B3464ULL, "vxlt_gallowstemplar_3rw_0_0.vlm"},
	{0xC5B10464766A1F8AULL, "docks_male_elven_worker_sta6175785.cl"},
	{0xC5B10EDB104AF4AAULL, "6155310_f.fxe"},
	{0xC5B2B56535538C8DULL, "llac_ca_caverns1_mines_24424_5.dds"},
	{0xC5B2CFB8619AE84DULL, "art\\characters\\creatures\\desiredemon\\animation\\p_tw_v1.gad"},
	{0xC5B340EFAFAC3299ULL, "vxlt_0_lowtown_a2_n_c_ehl_0_0.vlm"},
	{0xC5B39613C7062AE1ULL, "lt_gallowscourtyard_59_0.anb"},
	{0xC5B405DD02856494ULL, "475726092.ogg"},
	{0xC5B40B76386B0000ULL, "dae000_pc_home_ambients6208668.cl"},
	{0xC5B515820D13806EULL, "gen_im_acc_rng_act3_05.uti"},
	{0xC5B545AAB6DBFF36ULL, "ave200_donnic6111997.cl"},
	{0xC5B57C95BBCE477BULL, "ser121_jethann6104441.cl"},
	{0xC5B5DF325B64A75CULL, "6129806_m.fxe"},
	{0xC5B6017744A04EDCULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.cess.gad"},
	{0xC5B602550DA29730ULL, "664400660.ogg"},
	{0xC5B680EAFC9DB0E8ULL, "vxw_bonepitdestroyed_a3_ku_0_0.vlm"},
	{0xC5B6F41A4E28C612ULL, "dae000_hate_follower_react6103226.cl"},
	{0xC5B81BBB7ADC30B3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandright.phy"},
	{0xC5B8D09DDE2BD45AULL, "mcr310_argument6072711.cl"},
	{0xC5B8EFCFFF6985DBULL, "vxht_chantry_n_25x_0_0.vlm"},
	{0xC5B9536C14384DC2ULL, "860514998.ogg"},
	{0xC5B96928F9212A4AULL, "6106880_f.fxe"},
	{0xC5B96CA15CB0B69DULL, "vxht_mansion_n_4a8_0_0.vlm"},
	{0xC5B9B2E4D94DCB3FULL, "gen00fl_isabela6103968.cl"},
	{0xC5B9B5153CB8277FULL, "6092220_m.fxe"},
	{0xC5BA02ABF8969D48ULL, "hlac_ht_bartrand_n_3366_0y.dds"},
	{0xC5BA2C151D2F609FULL, "vxw_bonepitdestroyed_a3_1lp_0_0.vlm"},
	{0xC5BA9ADE068F4888ULL, "follower_banter6156179.cl"},
	{0xC5BAFB89E9723D29ULL, "vxw_blightlands_ma_0_0.vlm"},
	{0xC5BB086FC1776779ULL, "vxlt_undercityserialkiller_1t8_0_0.vlm"},
	{0xC5BB288676788237ULL, "vxlt_gallowscourtyard_n_74w_0_0.vlm"},
	{0xC5BB6762DC543318ULL, "one120_athenril6088292.cl"},
	{0xC5BBE9B466D3DFE1ULL, "6167508_m.fxe"},
	{0xC5BC007744A4B8DAULL, "art\\vfx\\combat\\vfx_glb_resurrect_crust_c.cess.ani"},
	{0xC5BCAB16C2736534ULL, "llac_ca_caverns1_mines_24424_20.dds"},
	{0xC5BCE6A7601D29ADULL, "6209255_m.fxe"},
	{0xC5BD0F71CA3C380FULL, "w_bonepit_33_0.mao"},
	{0xC5BD1B71CA3C4CACULL, "w_bonepit_33_0.mmh"},
	{0xC5BD2171CA3C569EULL, "w_bonepit_33_0.msh"},
	{0xC5BD5E2745D73409ULL, "vxlt_0_lowtown_a3_n_c_hsy_0_0.vlm"},
	{0xC5BD646BE06CE262ULL, "6182300_m.fxe"},
	{0xC5BD689B463A2409ULL, "vxht_brothel_n_633_0_0.vlm"},
	{0xC5BDB5E569E90C23ULL, "llac_lt_undercitydungeon_-1_46.dds"},
	{0xC5BE41513244EB65ULL, "710644982.ogg"},
	{0xC5BE6D1F4DC70495ULL, "hlac_ca_caverns2_5802_41y.dds"},
	{0xC5BE75765946B3A4ULL, "mrl210_merrill6166867.cl"},
	{0xC5BED0FC6F00997DULL, "zz_dae_debug6029177.cl"},
	{0xC5BF69A657A54BAFULL, "hlac_ca_caverns2prc_5802_65x.dds"},
	{0xC5BF78AA2AE945A6ULL, "dae110cr_human_m_01.utc"},
	{0xC5BFD7A879E00E6EULL, "vxlt_backalley_n_1r8_0_0.vlm"},
	{0xC5C00B6E79D318BEULL, "mcr310st_argument.stg"},
	{0xC5C0D811B3436369ULL, "vxw_sundermount_31p_0_0.vlm"},
	{0xC5C0E97C2CA1FE73ULL, "6057331_m.fxe"},
	{0xC5C18CA1A984A852ULL, "6060798_m.fxe"},
	{0xC5C1B219C6CFBCD0ULL, "97619244.ogg"},
	{0xC5C298C4EFE3A22AULL, "hm_dead_2.mor"},
	{0xC5C2CF9918FBFFF6ULL, "6081719_m.fxe"},
	{0xC5C36515DE2434BCULL, "qun160_vashothlead6053887.cl"},
	{0xC5C3C1B298BCE273ULL, "w_sundermount_205_0.phy"},
	{0xC5C46663BE76A33AULL, "w_encounter2_camp_43_0.rml"},
	{0xC5C4C9FB79FA2D87ULL, "vxdr_0_deeproads_ef2_0_0.vlm"},
	{0xC5C4EFEAAEE52E29ULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.mmh"},
	{0xC5C515038CC9DCBCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerbtall.mao"},
	{0xC5C521038CC9F15FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerbtall.mmh"},
	{0xC5C527038CC9FB01ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerbtall.msh"},
	{0xC5C575673D37EC73ULL, "follower_banter6143125.cl"},
	{0xC5C58E5CA5598829ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev4.gad"},
	{0xC5C59B8676C9A0DDULL, "6208699_f.fxe"},
	{0xC5C63ED8428E8FCEULL, "w_woundedcoast_76_0.msh"},
	{0xC5C64502312C85A2ULL, "gen_im_trs_dwv_junk3_03.uti"},
	{0xC5C64CD8428EA71FULL, "w_woundedcoast_76_0.mao"},
	{0xC5C6555616D8D918ULL, "vxdr_0_deeproads_9gf_0_0.vlm"},
	{0xC5C658D8428EBBBCULL, "w_woundedcoast_76_0.mmh"},
	{0xC5C72D9EF0844847ULL, "qun160_party_comment_16061228.cl"},
	{0xC5C745B5C8D0F55EULL, "613259220.ogg"},
	{0xC5C79710E6739F2DULL, "6042516_f.fxe"},
	{0xC5C822681C404396ULL, "follower_banter6172966.cl"},
	{0xC5C826F22EAE3FFFULL, "6097120_m.fxe"},
	{0xC5C83F6D4771993FULL, "333096908.ogg"},
	{0xC5C84FF2C941C22EULL, "plt_cod_ltr_conspiracy_cold.nss"},
	{0xC5C85194B70B942EULL, "one100dg_lia6013841.cl"},
	{0xC5C8AED2183D340BULL, "vxlt_0_lowtown_a2_n_c_gca_0_0.vlm"},
	{0xC5C8CE465B24E7A4ULL, "water_water_p1.vsh.fxo.dx11.assembly"},
	{0xC5C8F1DE01F69403ULL, "llac_lt_gallowstemplar_f_2384_5.dds"},
	{0xC5C921ABE8192B0AULL, "dae210_gate_guard6085040.cl"},
	{0xC5C9EB6EC93AF942ULL, "mag371_elsa6041520.cl"},
	{0xC5C9FE5683B3B089ULL, "qcr000_arishok6042490.cl"},
	{0xC5CA0DC208D6C0C3ULL, "vxdr_primevalentrance_1ae_0_0.vlm"},
	{0xC5CAA01F6C00E3FFULL, "vxlt_undercitydungeon_xb_0_0.vlm"},
	{0xC5CABB77F0517A51ULL, "vxht_mansion2b_314_0_0.vlm"},
	{0xC5CAD50597F170ECULL, "869243365.ogg"},
	{0xC5CAEF13B4D2DBADULL, "hlac_ca_cavern1_haunted_24424_20y.dds"},
	{0xC5CB0E752AF7FDE3ULL, "vxdr_primevaltomb_71b_0_1.vlm"},
	{0xC5CB132653842E13ULL, "6070451-merril-intro.cut"},
	{0xC5CB4763BE7C8D2EULL, "w_encounter2_camp_43_0.phy"},
	{0xC5CB7CE7C9B8C3A6ULL, "6068426_m.fxe"},
	{0xC5CBBA7A9CA7203CULL, "da2terrain_terrain_tools_p0.psh.fxo.dx11.assembly"},
	{0xC5CC1DA6CAEED37EULL, "569555597.ogg"},
	{0xC5CC5F5CA55F5777ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev4.ani"},
	{0xC5CCA38C56A86EDAULL, "509296164.ogg"},
	{0xC5CCE633B26446A6ULL, "6053132_m.fxe"},
	{0xC5CCF7551000657BULL, "6052776_m.fxe"},
	{0xC5CD9533DBD5C445ULL, "gen00fl_aveline6106919.cl"},
	{0xC5CE5F6C28A84806ULL, "640069935.ogg"},
	{0xC5CEE581DD20319FULL, "hlac_lt_gallowstemplar_f_2390_0z.dds"},
	{0xC5CF626A11730E06ULL, "vxlt_0_lowtown_e15_0_0.vlm"},
	{0xC5D041A198A74F4DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexwild_0.pwk"},
	{0xC5D0902EF4C0C576ULL, "og_flsh_03.dds"},
	{0xC5D0BB61A56FA323ULL, "hlac_ca_caverns1_24424_5y.dds"},
	{0xC5D0E444438F943DULL, "vxdr_primevaltomb_545_0_1.vlm"},
	{0xC5D141760AF21876ULL, "6084419_m.fxe"},
	{0xC5D1631300D1C52BULL, "mrl000_marethari6129224.cl"},
	{0xC5D1AD4D6DE6EB9BULL, "hlac_lt_foundry_n_320_1y.dds"},
	{0xC5D209E7077C28B6ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.mmh"},
	{0xC5D23A32BB267BDDULL, "hlac_ca_caverns1_mines_24425_6z.dds"},
	{0xC5D295C41344ACBBULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_f_c.impa.ani"},
	{0xC5D320EFE7D6EA5BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_major_v1.ani"},
	{0xC5D38C0A99EC3118ULL, "vxlt_0_lowtown_n_fff_0_0.vlm"},
	{0xC5D3B3802AB86327ULL, "vxht_0_hightown_a2_ana_0_0.vlm"},
	{0xC5D3D544A9F1FCCFULL, "81355986.ogg"},
	{0xC5D3E1896A04D8FFULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratequest_dp.dds"},
	{0xC5D3E1C382C693CFULL, "plt_prm000pt_frozen_boots.plo"},
	{0xC5D40181F6B6BABDULL, "6136696_m.fxe"},
	{0xC5D47816971F31D6ULL, "6109709_m.fxe"},
	{0xC5D4946073EDC456ULL, "da2terrain_terrain_tools_p0.vsh.fxo.dx11.assembly"},
	{0xC5D50478965FEFD8ULL, "ht_0_hightown_a3_61_0.rml"},
	{0xC5D587ACDA0F8B36ULL, "llac_lt_0_lowtown_docks_65534_1.dds"},
	{0xC5D62E134443778BULL, "hlac_lt_gallowstemplar_n_2383_6x.dds"},
	{0xC5D68BE5CB49A07DULL, "124386218.ogg"},
	{0xC5D7BBCC19D250BEULL, "118546203.ogg"},
	{0xC5D8AAF0CDFB8E9FULL, "vxlt_0_lowtown_docks_b0a_0_0.vlm"},
	{0xC5D8C1B185862488ULL, "343232182.ogg"},
	{0xC5D8CB957CCC405FULL, "vxdr_primevaltomb_7p0_0_0.vlm"},
	{0xC5D8CE3CDA341697ULL, "974112883.ogg"},
	{0xC5D8F4A9B5CDFE29ULL, "69678993.ogg"},
	{0xC5D94B6D7FB34159ULL, "382063124.ogg"},
	{0xC5D961FC989AE615ULL, "llac_ht_mansion2_818_49.dds"},
	{0xC5D9860A3AEF60A2ULL, "vxlt_0_lowtown_ekg_0_0.vlm"},
	{0xC5D9889D237AE392ULL, "90863259.ogg"},
	{0xC5D9A4B53D58FD91ULL, "677043484.ogg"},
	{0xC5D9CC894F753C57ULL, "1041451540.ogg"},
	{0xC5DA40E89ABE4751ULL, "vxlt_gallowstemplar_6s4_0_0.vlm"},
	{0xC5DAAD1E756715E6ULL, "hlac_ca_caverns1_24423_6x.dds"},
	{0xC5DAC44ECEE6F00EULL, "vxlt_undercityencounter_1je_0_0.vlm"},
	{0xC5DAC91F44F0751EULL, "6125615_m.fxe"},
	{0xC5DAF3DA132B9A39ULL, "art\\vfx\\creatures\\fxc_ashwraith_c.mmh"},
	{0xC5DB00A40414A847ULL, "dae000_bartrand6206145.cl"},
	{0xC5DBC6DC14DA59A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_chair_loop.ani"},
	{0xC5DC6FACD01C3055ULL, "hlac_w_woundedcoast_65538_0y.dds"},
	{0xC5DDD2EA18438144ULL, "zz_appearance.ncs"},
	{0xC5DE2C3B6DABF294ULL, "hlac_lt_gallowstemplar_n_7198_1y.dds"},
	{0xC5DE3E330BFEB3A0ULL, "lgt110_gxa_vgk_silver_order6117030.cl"},
	{0xC5DE6D8DCB5D0F2AULL, "6087471_f.fxe"},
	{0xC5DF009D5DFD20B1ULL, "llac_lt_0_lowtown_a3_n_c_0_4.dds"},
	{0xC5DF6B5410EA07E5ULL, "vxlt_gallowscourtyard_n_3w0_0_0.vlm"},
	{0xC5DF87AA9F718EAEULL, "gen00fl_fenris6126584.cl"},
	{0xC5DF9DC4134F44B2ULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_f_c.impa.evt"},
	{0xC5E013CF7CF0A8C2ULL, "t3_hlk_h03.tnt"},
	{0xC5E098A963EB886CULL, "6075242_m.fxe"},
	{0xC5E0CB4B83C2E891ULL, "hlac_w_encounter3_0_10y.dds"},
	{0xC5E0CC2209F01E78ULL, "hlac_ca_caverns2prc_5802_11x.dds"},
	{0xC5E128EFE7E33552ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_major_v1.evt"},
	{0xC5E1966ECE59B704ULL, "vxlt_foundry_n_zb_0_0.vlm"},
	{0xC5E1C8F91BB36F57ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he5a.phy"},
	{0xC5E228A43D74E7FEULL, "vxca_caverns1_i9y_0_0.vlm"},
	{0xC5E38AB9F27556C0ULL, "vxht_0_hightown_a3_ca8_0_0.vlm"},
	{0xC5E3B40AC422683DULL, "297289932.ogg"},
	{0xC5E452A1F62C9995ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile.mao"},
	{0xC5E458D842A7B90FULL, "w_woundedcoast_76_0.rml"},
	{0xC5E4F3B60226DF93ULL, "dae000_epic_merchant_36179490.cl"},
	{0xC5E5196D14C7CF04ULL, "gen00fl_isabela6085573.cl"},
	{0xC5E53B6EEDF9658EULL, "pro000_friedrich6050179.cl"},
	{0xC5E55A4256A1D866ULL, "6035129_m.fxe"},
	{0xC5E59560324FDBF8ULL, "515367935.ogg"},
	{0xC5E5E273E8B4D9FAULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masf.phy"},
	{0xC5E60EA28DC9F27AULL, "vxlt_0_lowtown_a2_n_c_fff_0_0.vlm"},
	{0xC5E63432936162CBULL, "narrative_varric6172492.cl"},
	{0xC5E67F3819F5F89AULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.anb"},
	{0xC5E6A7B9490B8D3AULL, "6170514_m.fxe"},
	{0xC5E6B298D74CB0ECULL, "zz_hawke_disappear.cut"},
	{0xC5E6BCC413559845ULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_f_c.impa.gad"},
	{0xC5E6DA99C39FAAACULL, "vxht_0_hightown_a2_c19_0_0.vlm"},
	{0xC5E75751038A4C56ULL, "qun210cr_guard6072087.cl"},
	{0xC5E75F77DD8B631FULL, "6095571_m.fxe"},
	{0xC5E762A3E968A739ULL, "art\\vfx\\materials\\vfx_small_dot.mao"},
	{0xC5E7754588B06435ULL, "6127294_m.fxe"},
	{0xC5E7972EF4D3AFA4ULL, "og_flsh_03.mao"},
	{0xC5E7B35296EA84EFULL, "hlac_w_woundedcoast_131073_0y.dds"},
	{0xC5E7E265751615AFULL, "vxlt_darktown_5jb_0_0.vlm"},
	{0xC5E80D212A825E1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_loop.gad"},
	{0xC5E8344737100D3EULL, "hlac_lt_0_lowtown_65535_0y.dds"},
	{0xC5E847EFE7E988E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_major_v1.gad"},
	{0xC5E86C64C08F3DAEULL, "vxlt_0_lowtown_a2_n_eoq_0_1.vlm"},
	{0xC5E8C773E569C9E7ULL, "mcr371_idunna6087283.cl"},
	{0xC5E8CBF4D02FB15FULL, "vxw_woundedcoast_1bd_0_1.vlm"},
	{0xC5EAB130FF3F5A0AULL, "ave331_jeven6138721.cl"},
	{0xC5EAE6BBBB10D65DULL, "ht_0_hightown_n_44_0.rml"},
	{0xC5EB2B1DE2AB0A7EULL, "hlac_w_blightlands_262142_0x.dds"},
	{0xC5EB49F2216CAB04ULL, "209246821.ogg"},
	{0xC5EB86897B98DA31ULL, "montage_subs_demohack6217102.cl"},
	{0xC5EB9DBAA96CF592ULL, "vxlt_0_lowtown_a2_n_h2b_0_0.vlm"},
	{0xC5EBDCED74105159ULL, "vxw_encounter2_camp_le_0_0.vlm"},
	{0xC5EBF07A2B1B2EABULL, "dae000_arishok6075159.cl"},
	{0xC5EBF6D19BA9C46EULL, "w_woundedcoast_86_0.rml"},
	{0xC5EC01687349F5FAULL, "92331783.ogg"},
	{0xC5EC2BF858FD0FF6ULL, "1030603841.ogg"},
	{0xC5ECA82CD204D320ULL, "164862256.ogg"},
	{0xC5ED388A5959243FULL, "vxw_encounter2_camp_5t_0_0.vlm"},
	{0xC5ED515D69898671ULL, "gen00fl_carver6103946.cl"},
	{0xC5ED6716B0140290ULL, "hlac_lt_ruinedhouse_n_-1_0x.dds"},
	{0xC5ED7233A08543CFULL, "297698239.ogg"},
	{0xC5ED733311165871ULL, "121974340.ogg"},
	{0xC5ED7713C7376E71ULL, "lt_gallowscourtyard_59_0.rml"},
	{0xC5EDBEF1642A4640ULL, "gen00fl_sebastian6174076.cl"},
	{0xC5EDECEAAF082E5EULL, "art\\vfx\\combat\\vfx_r_brd_csong_tag_loop_c.anb"},
	{0xC5EE4BF9A2C32503ULL, "zz_qun_debug6088657.cl"},
	{0xC5EE6FAE960D3C89ULL, "llac_lt_0_lowtown_docks_a3_n_-65537_4.dds"},
	{0xC5EF14C40A24EC7BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_varric.phy"},
	{0xC5EFA5CA146C2CBBULL, "vxdr_0_deeproads_cuo_0_2.vlm"},
	{0xC5F0084BC04FCF84ULL, "vxlt_foundry_n_uc_0_0.vlm"},
	{0xC5F07C32ED517430ULL, "gen00fl_dog6115121.cl"},
	{0xC5F07F8DCE4772B3ULL, "6133510_m.fxe"},
	{0xC5F0A52C929E8C8FULL, "6048285_m.fxe"},
	{0xC5F0B5911932DF27ULL, "chest_pressed.cur"},
	{0xC5F0C53504CA5073ULL, "vxlt_0_lowtown_a2_n_gcc_0_0.vlm"},
	{0xC5F0C6EED7F355EFULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.impa.gad"},
	{0xC5F0DDDC14ECDC8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_chair_loop.gad"},
	{0xC5F1D17662408F1FULL, "vxdr_primevaltomb_8vq_0_0.vlm"},
	{0xC5F1E830BB5DDCB9ULL, "214623593.ogg"},
	{0xC5F2386390492B35ULL, "sst_ss_midclass_thugs_m_1.gda"},
	{0xC5F2670B4CA7A0B8ULL, "6096898_m.fxe"},
	{0xC5F2B5F671BF06EAULL, "art\\characters\\creatures\\mubari\\animation\\relax.gad"},
	{0xC5F2C6E3C684CAD9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.ani"},
	{0xC5F33D562C1DCFECULL, "6169532_m.fxe"},
	{0xC5F360556C698304ULL, "vxlt_gallowstemplar_f_5qk_0_0.vlm"},
	{0xC5F40EF41DD3202AULL, "mag110_vincento.cnv"},
	{0xC5F459F497F3EED2ULL, "1055751895.ogg"},
	{0xC5F4EC750BEF3417ULL, "420512687.ogg"},
	{0xC5F537D82CACB5CBULL, "t3_hey_gn2.tnt"},
	{0xC5F53FAFFBB08B6BULL, "484393046.ogg"},
	{0xC5F54870F3A6DE99ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_galocelldoor.msh"},
	{0xC5F55081554C0710ULL, "zz_mer_debug6180857.cl"},
	{0xC5F55270F3A6EF87ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_galocelldoor.mmh"},
	{0xC5F55670F3A6F6D4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_galocelldoor.mao"},
	{0xC5F5626F08CF8BB0ULL, "mag360_samson6042059.cl"},
	{0xC5F56FE0F8691C40ULL, "6041959_m.fxe"},
	{0xC5F5840EAFC3E31EULL, "mag291_torpor6025993.cl"},
	{0xC5F584197713DB7AULL, "vxw_woundedcoast_6m_0_0.vlm"},
	{0xC5F58B94F05932F4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_ric_0d.dds"},
	{0xC5F5961E56800779ULL, "619252655.ogg"},
	{0xC5F5C5118188A3C3ULL, "6080973_m.fxe"},
	{0xC5F5D3096E422783ULL, "hlac_lt_0_lowtown_a2_n_-196608_1y.dds"},
	{0xC5F6456F1783E432ULL, "llac_dr_slavers_3216_2.dds"},
	{0xC5F65442670C2E2FULL, "vxlt_darktown_5om_0_0.vlm"},
	{0xC5F680B000FB7951ULL, "vxca_caverns1_mines_j2a_0_0.vlm"},
	{0xC5F68FDEE2A208D5ULL, "cs_mcr370cs_meredith_idol.bnk"},
	{0xC5F6A6C36F21B0E4ULL, "vxw_encounter2_m3_0_0.vlm"},
	{0xC5F6DEF674355FCAULL, "gen_im_fla_isabela_01.uti"},
	{0xC5F6E617FFB05E6BULL, "6151626_m.fxe"},
	{0xC5F6EC78AEFACE41ULL, "vxdr_primevalentrance_276_0_0.vlm"},
	{0xC5F7197EC35AD6C4ULL, "hlac_lt_0_lowtown_a3_n_c_0_8y.dds"},
	{0xC5F7DBEED7F9979CULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.impa.evt"},
	{0xC5F7F0F41DD6D652ULL, "mag110_vincento.bnk"},
	{0xC5F8C9EB1D4DCAA9ULL, "vxca_caverns1_mines_i1j_0_1.vlm"},
	{0xC5F8CC15A09A5940ULL, "hlac_lt_hangedman_n_-1_8z.dds"},
	{0xC5F90F0372DBED2BULL, "character_character_p2.vsh.fxo.dx11.assembly"},
	{0xC5F998F671C4F381ULL, "art\\characters\\creatures\\mubari\\animation\\relax.evt"},
	{0xC5F9A1E3C68AAA1BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.gad"},
	{0xC5F9B6858186004EULL, "sky.vsh.fxo"},
	{0xC5FA1853AE32DF64ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_2_ian.gad"},
	{0xC5FA4ECA660669D9ULL, "vxw_encounter2_camp_gd_0_0.vlm"},
	{0xC5FA5DA5EE8F0475ULL, "969411647.ogg"},
	{0xC5FA6FC0B32B778CULL, "553859766.ogg"},
	{0xC5FB753864C504A4ULL, "6136330_m.fxe"},
	{0xC5FBBB5EE08B8296ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_ex.ani"},
	{0xC5FBDBBEECCE46E9ULL, "hlac_ht_store_187_4x.dds"},
	{0xC5FC6528625FC134ULL, "411751140.ogg"},
	{0xC5FCE5ABD8B1DCB1ULL, "mer220_hubert6088152.cl"},
	{0xC5FD28212A94E800ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_on_ground_loop.ani"},
	{0xC5FDD7085CDA4B8FULL, "fx_fireringfb_blend.mao"},
	{0xC5FE2041C05A22F0ULL, "lt_0_lowtown_docks_a3_n_184_0.mao"},
	{0xC5FE2441C05A2903ULL, "lt_0_lowtown_docks_a3_n_184_0.mmh"},
	{0xC5FE3241C05A415DULL, "lt_0_lowtown_docks_a3_n_184_0.msh"},
	{0xC5FE72489EE30AC6ULL, "vxlt_0_lowtown_f1f_0_0.vlm"},
	{0xC5FEFB1F301F8576ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.dura.ani"},
	{0xC5FF17098494A50FULL, "6043385_m.fxe"},
	{0xC5FF1751AC485116ULL, "820774165.ogg"},
	{0xC5FFFDDFF60D1158ULL, "vxlt_0_lowtown_a2_n_c_emn_0_0.vlm"},
	{0xC6006E544FAF02F7ULL, "6097872_m.fxe"},
	{0xC6008F46624D61BAULL, "hlac_w_encounter2_camp_131073_0y.dds"},
	{0xC600AAF17D54EA3DULL, "vxdr_0_deeproads_eui_0_0.vlm"},
	{0xC600BEE3C690FA00ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_twitch.evt"},
	{0xC600F2B71DE4CB04ULL, "6122444_f.fxe"},
	{0xC6012753AE391742ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_2_ian.ani"},
	{0xC6012B033F5AD3E7ULL, "hlac_w_encounter2_camp_196607_0x.dds"},
	{0xC601DE4C6359065AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_visa_s.dds"},
	{0xC60207A00B097644ULL, "hlac_lt_0_lowtown_a2_n_c_-3_5y.dds"},
	{0xC6022B04ADB3D608ULL, "vxw_bonepit_27y_0_0.vlm"},
	{0xC6024E6300CC2D68ULL, "vxlt_0_lowtown_docks_a3_n_aqb_0_0.vlm"},
	{0xC6025A5DC25E40BFULL, "vxlt_foundry_n_104_0_0.vlm"},
	{0xC6027928564B0905ULL, "6125133_m.fxe"},
	{0xC602F0DA134CE72EULL, "art\\vfx\\creatures\\fxc_ashwraith_c.anb"},
	{0xC603D12254A3B350ULL, "1034071736.ogg"},
	{0xC603F5F66909F3FAULL, "vxw_woundedcoast_3n_0_0.vlm"},
	{0xC6040323BF931865ULL, "dae120ar_hightown_nt.are"},
	{0xC6041F7B3CA7B8C4ULL, "486309387.ogg"},
	{0xC604C61F590246AFULL, "vxlt_darktown_5pl_0_0.vlm"},
	{0xC604E3EED8042FB5ULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.impa.ani"},
	{0xC60518A06117C964ULL, "vxw_encounter2_h0_0_0.vlm"},
	{0xC6053FD622BE64C1ULL, "6093694_m.fxe"},
	{0xC60556AFE8DF6BCEULL, "6027720_m.fxe"},
	{0xC60582A34C3BBF4FULL, "168778716.ogg"},
	{0xC605A8BE16AFD700ULL, "955659641.ogg"},
	{0xC605AE7452141CF5ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.mmh"},
	{0xC605DE9E99CAE4F0ULL, "6111562_m.fxe"},
	{0xC605F2E8E11E274FULL, "hlac_ht_mansion_n_3371_1x.dds"},
	{0xC6063C706A9C8E71ULL, "vxlt_warehouse_n_w6_0_0.vlm"},
	{0xC606474F6C491C10ULL, "w_bonepit_245_0.mao"},
	{0xC6064B4F6C492323ULL, "w_bonepit_245_0.mmh"},
	{0xC606594F6C493B7DULL, "w_bonepit_245_0.msh"},
	{0xC6065F3C14F57AC5ULL, "vxlt_0_lowtown_docks_a3_cp1_0_0.vlm"},
	{0xC60677E01939C205ULL, "gen00fl_merrill6127551.cl"},
	{0xC606EF74ED8F5960ULL, "w_sundermount_185_0.phy"},
	{0xC6070CCACC1C7CCDULL, "6069794_m.fxe"},
	{0xC6072A84953E0F48ULL, "mag120_macha6080591.cl"},
	{0xC6074E72D3134948ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava_n.dds"},
	{0xC60790F671D123D0ULL, "art\\characters\\creatures\\mubari\\animation\\relax.ani"},
	{0xC607B286A06B8160ULL, "fen120ip_mansion_chest_2.utp"},
	{0xC6085E8F12036AF4ULL, "449201787.ogg"},
	{0xC608C0A757FC8259ULL, "vxw_encounter2_camp_bk_0_0.vlm"},
	{0xC608C833D31CCA92ULL, "6144557_m.fxe"},
	{0xC608F8D01C11D257ULL, "6171130_m.fxe"},
	{0xC60907E6FB575039ULL, "hm_rdr_guard_sleeping.mor"},
	{0xC60989E6685121B7ULL, "858665722.ogg"},
	{0xC609C35EE097CD87ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_ex.evt"},
	{0xC609E71556BB1D24ULL, "6131230_m.fxe"},
	{0xC60A1CBFAF74DA41ULL, "532670558.ogg"},
	{0xC60A214CBD447E60ULL, "596162766.ogg"},
	{0xC60A24070B857F04ULL, "6057192_m.fxe"},
	{0xC60AABCA3D7B0E1EULL, "hlac_ca_caverns2prc_5802_10y.dds"},
	{0xC60B37BA0AFA6FBEULL, "6117950_f.fxe"},
	{0xC60B4663BEB26FF1ULL, "w_encounter2_camp_43_0.mmh"},
	{0xC60B5263BEB28392ULL, "w_encounter2_camp_43_0.mao"},
	{0xC60B6063BEB29BAFULL, "w_encounter2_camp_43_0.msh"},
	{0xC60B63D251B4194FULL, "6083254_f.fxe"},
	{0xC60C031F302A1D67ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.dura.evt"},
	{0xC60C745F36D77E87ULL, "6155550_m.fxe"},
	{0xC60C76A76FF2ADABULL, "vxht_0_hightown_n_b76_0_0.vlm"},
	{0xC60CF64D44EE684CULL, "gen00fl_fenris6126425.cl"},
	{0xC60D0074049DA8FEULL, "isa000ip_demo_gate.utp"},
	{0xC60D0A9CE2F51869ULL, "hlac_w_sundermount_-262142_0x.dds"},
	{0xC60D417D9F529A65ULL, "904157627.ogg"},
	{0xC60DA8F0D382EA00ULL, "1059458622.ogg"},
	{0xC60DEE3C181CB81FULL, "hlac_ht_0_hightown_a3_-65538_0y.dds"},
	{0xC60E2817D6445C4DULL, "docks_male_laborer_mob6175397.cl"},
	{0xC60EF3C812FDAE2CULL, "vxca_caverns2_71n_0_0.vlm"},
	{0xC60F0423BF9D214BULL, "dae120ar_hightown_nt.ncs"},
	{0xC60F82381A19034DULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.mmh"},
	{0xC60FA5442E086CF9ULL, "hlac_ht_mansion_n_3370_1y.dds"},
	{0xC60FF6F8EDE7C60EULL, "gen00fl_isabela6075240.cl"},
	{0xC6102110100A00ECULL, "6163732_m.fxe"},
	{0xC6106340983466A8ULL, "mag291_demon_marethari6028937.cl"},
	{0xC610B182D41C7A42ULL, "ts_generic_arm_table.gda"},
	{0xC610C6B6912E57D2ULL, "379612853.ogg"},
	{0xC610E45EE09E24D0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_ex.gad"},
	{0xC6111E45DC1EB152ULL, "637365194.ogg"},
	{0xC611CC44306DD708ULL, "422378868.ogg"},
	{0xC611FD14883818AAULL, "lt_gallowscourtyard_n_51_0_ll.mmh"},
	{0xC61218C2ECD34E53ULL, "103275223.ogg"},
	{0xC612199677262D78ULL, "dae000_flemeth6013363.cl"},
	{0xC612204A9A2AF4BCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_exit.ani"},
	{0xC612581FBCBB40C2ULL, "vxw_encounter2_or_0_1.vlm"},
	{0xC61296206BA5508CULL, "mag120cr_templar_1.utc"},
	{0xC612AB74521EA182ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s3_c.anb"},
	{0xC612BABF29976692ULL, "vxca_caverns1_mines_dpj_0_0.vlm"},
	{0xC613241F303073B0ULL, "art\\vfx\\environment\\vfx_trap_bolt_p.dura.gad"},
	{0xC6137BF6F6AA9B8DULL, "vxlt_0_lowtown_a2_glo_0_0.vlm"},
	{0xC6140B7AE7C24DC6ULL, "llac_ht_0_hightown_-2_0.dds"},
	{0xC6145B252A44A9E7ULL, "gen00fl_isabela6085749.cl"},
	{0xC6147813BDEA7355ULL, "6097218_m.fxe"},
	{0xC6148664C22B4920ULL, "475083746.ogg"},
	{0xC614AE4D5C92A6F1ULL, "vxlt_warehouse_n_z5_0_0.vlm"},
	{0xC614B78E12884884ULL, "hlac_lt_0_lowtown_a3_n_c_131071_4z.dds"},
	{0xC6150423BFA18C92ULL, "dae120ar_hightown_nt.lst"},
	{0xC61506CAC8F7B245ULL, "356500350.ogg"},
	{0xC61517EEBFD31E62ULL, "6131412_m.fxe"},
	{0xC61574E92D39F7B2ULL, "gen_im_trs_qun_hvy_cht_r0.uti"},
	{0xC615F54E9351140FULL, "hlac_w_sundermount_-327683_0x.dds"},
	{0xC616074464DF32E2ULL, "653053307.ogg"},
	{0xC616235BF83EA06AULL, "w_bonepit_26_0_lc.mmh"},
	{0xC6164FB07571075AULL, "vxlt_gallowstemplar_n_5bx_0_0.vlm"},
	{0xC6169B877865C896ULL, "6081801_m.fxe"},
	{0xC6169C396BA909A8ULL, "mag291_vincento6027692.cl"},
	{0xC616BB9AA96D981AULL, "w_encounter3_90_0.mmh"},
	{0xC616BF9AA96D9F51ULL, "w_encounter3_90_0.mao"},
	{0xC616CD9AA96DB760ULL, "w_encounter3_90_0.msh"},
	{0xC616F56C6D2F3B7BULL, "zz_followers_banter6158985.cl"},
	{0xC61717C353A4D2B1ULL, "one100dg_vanard6047283.cl"},
	{0xC6173A10C512E312ULL, "6141657_m.fxe"},
	{0xC617D3C427C749AEULL, "tbe100_xenon6190988.cl"},
	{0xC6180E167E7BE431ULL, "hlac_w_encounter1_-131072_0y.dds"},
	{0xC61812C40861B04AULL, "vxht_bartrand_n_4pc_0_0.vlm"},
	{0xC618226E54CE3782ULL, "613847683.ogg"},
	{0xC618EDFA80FC24BAULL, "974335573.ogg"},
	{0xC6192D4A9A3129FEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_exit.gad"},
	{0xC61A84ECC17DFBD5ULL, "hlac_w_woundedcoast_-196602_0y.dds"},
	{0xC61AB37702B0C786ULL, "vxht_mansion_n_62l_0_0.vlm"},
	{0xC61AFBC40A49FC1AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_varric.msh"},
	{0xC61B006DEF116E36ULL, "hlac_ht_chantry_n_2534_25y.dds"},
	{0xC61B0DC40A4A1B60ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_varric.mmh"},
	{0xC61B0EB562BC4807ULL, "follower_banter6141589.cl"},
	{0xC61B14BB7B30A228ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandright.mmh"},
	{0xC61B20BB7B30B6CBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandright.mao"},
	{0xC61B32BB7B30D552ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandright.msh"},
	{0xC61BD48262991BB9ULL, "loginconnecting.gfx"},
	{0xC61BD736757FAC5AULL, "hlac_ht_chantry_2534_53y.dds"},
	{0xC61CD6B783570884ULL, "230489512.ogg"},
	{0xC61CD9C95E188DB0ULL, "zz_dre_debug6092012.cl"},
	{0xC61CF9AC34D481F0ULL, "149068132.ogg"},
	{0xC61D8A31FB1B4A8CULL, "docks_male_human_worker_sta6175957.cl"},
	{0xC61DE5BBBB3C2E75ULL, "ht_0_hightown_n_44_0.anb"},
	{0xC61E00C3830500B4ULL, "plt_prm000pt_frozen_boots.nss"},
	{0xC61E5B1F95A9485BULL, "gen_im_trs_gen_two_2ha_r0.uti"},
	{0xC61EA6514F65E3E4ULL, "1006194816.ogg"},
	{0xC61EB53360D14146ULL, "nat380_sibling6068001.cl"},
	{0xC61EC667A3224B31ULL, "hlac_ht_0_hightown_a2_-196609_0x.dds"},
	{0xC61F4AD7D320D477ULL, "mcr310_argument6087312.cl"},
	{0xC61FB3F389812424ULL, "941668453.ogg"},
	{0xC61FEFA590AF9AFDULL, "vxht_0_hightown_c5t_0_0.vlm"},
	{0xC61FFB77CACCAE99ULL, "6144402_m.fxe"},
	{0xC61FFE865A410875ULL, "zz_approval_debug6003837.cl"},
	{0xC620060B30CB9572ULL, "vxca_caverns1_mines_jgm_0_0.vlm"},
	{0xC620084A9A37091DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_exit.evt"},
	{0xC6203F519E4906FCULL, "one101_orders_followers6115898.cl"},
	{0xC6213B70F3CC03BCULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_galocelldoor.phy"},
	{0xC62168C66879B0E3ULL, "6067810_f.fxe"},
	{0xC621D0CFEDA642B5ULL, "665066853.ogg"},
	{0xC621D84B1BDDE15FULL, "plt_cod_itm_cold-blooded.plo"},
	{0xC621EDD3E8A0B40DULL, "vxlt_0_lowtown_a2_gcn_0_0.vlm"},
	{0xC6220E979267E9EBULL, "623089371.ogg"},
	{0xC622B5DEE2C7845FULL, "cs_mcr370cs_meredith_idol.wbd"},
	{0xC6231F526A7B303FULL, "gen00fl_carver6108905.cl"},
	{0xC623223C8771AE9CULL, "vxlt_0_lowtown_a3_n_c_gjl_0_0.vlm"},
	{0xC6235DCE23C2EDB4ULL, "6100255_f.fxe"},
	{0xC623D75426D16D37ULL, "6117911_f.fxe"},
	{0xC6240942AB79C487ULL, "dae300_bran6143091.cl"},
	{0xC62428E1B15C7FF0ULL, "dae000_denier6162685.cl"},
	{0xC6250EB9274388D1ULL, "vxlt_0_lowtown_a2_eoq_0_1.vlm"},
	{0xC6253B768B56567CULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_005a_0n.dds"},
	{0xC62580C7981FA959ULL, "428639470.ogg"},
	{0xC625BAB2990FA0E8ULL, "w_sundermount_205_0.mmh"},
	{0xC625C6B2990FB48BULL, "w_sundermount_205_0.mao"},
	{0xC625D8B2990FD312ULL, "w_sundermount_205_0.msh"},
	{0xC625EF142C79E414ULL, "lit_bas.mao"},
	{0xC6265C2D947E25F6ULL, "69135585.ogg"},
	{0xC6269BC9C83CB0FDULL, "gen00fl_carver6097218.cl"},
	{0xC6269C07C845FCC0ULL, "hlac_ht_mansion2_818_2z.dds"},
	{0xC626D1EA6FFC766AULL, "vxlt_0_lowtown_a2_n_csj_0_0.vlm"},
	{0xC62709DE7D36BA1DULL, "vxlt_gallowscourtyard_73c_0_0.vlm"},
	{0xC62758A9496CBF32ULL, "mag291st_torpor.stg"},
	{0xC627D068481902F6ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.impa.evt"},
	{0xC6283A7212874916ULL, "vxht_brothel_n_5lb_0_0.vlm"},
	{0xC628AE6CBB8E3EC7ULL, "424514461.ogg"},
	{0xC628E2AA52D28F0CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha10a.phy"},
	{0xC628FC2DA39F35BCULL, "6161613_m.fxe"},
	{0xC629978D0C7640B3ULL, "hlac_lt_0_lowtown_a2_-3_4x.dds"},
	{0xC629A9F91BF0F04CULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he5a.mmh"},
	{0xC629AFF91BF0FA3EULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he5a.msh"},
	{0xC62A2880B239488DULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medd_t.dds"},
	{0xC62AA08C6013867CULL, "1070532590.ogg"},
	{0xC62B44D9526A9331ULL, "vrc221_varric6043862.cl"},
	{0xC62B668AA4A22ADFULL, "fxa_evl_smoke_add.mao"},
	{0xC62D0C74EDB059F5ULL, "w_sundermount_185_0.msh"},
	{0xC62D0E74EDB05D6BULL, "w_sundermount_185_0.mmh"},
	{0xC62D1A74EDB07108ULL, "w_sundermount_185_0.mao"},
	{0xC62DBD4141F322B5ULL, "hlac_w_bonepitmining_a2_0_12x.dds"},
	{0xC62DC768481D5E81ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.impa.gad"},
	{0xC62E70B43D09DE71ULL, "vxca_caverns2prc_62k_0_1.vlm"},
	{0xC62EED9A6D7E8345ULL, "vxw_woundedcoast_3e_0_0.vlm"},
	{0xC62F13B6AFE9DD1BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\originals\\appearfromopen.evt"},
	{0xC62F291DAD60618EULL, "l_dae121ar_keep.wbd"},
	{0xC62F4F858893F90CULL, "mag160_varian_ilithis6016493.cl"},
	{0xC6300E0F0BB7DEA5ULL, "mcr371_scared_mage_f6087714.cl"},
	{0xC630AAD5C2F37DACULL, "6165608_m.fxe"},
	{0xC630F58FC463BFEDULL, "hlac_ca_caverns1_mines_24426_7x.dds"},
	{0xC63118DB95D9F2B1ULL, "21538985.ogg"},
	{0xC6316939802D96B0ULL, "lowtown_male_cityguard_mob6150518.cl"},
	{0xC631857C0DCB5287ULL, "717293701.ogg"},
	{0xC631F308DC39C5DEULL, "hlac_lt_0_lowtown_docks_-131073_2x.dds"},
	{0xC6322BCABF295D6FULL, "vxca_caverns1_mines_e8z_0_0.vlm"},
	{0xC632A7007CCC367EULL, "vxlt_hangedman_n_1y8_0_0.vlm"},
	{0xC633C286E3F264E2ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw104a.mao"},
	{0xC633C9EF67BC0509ULL, "6082115_m.fxe"},
	{0xC634406CFA644012ULL, "vxlt_shop_eq_0_0.vlm"},
	{0xC63446F2129038B7ULL, "hlac_lt_0_lowtown_n_-3_7x.dds"},
	{0xC634C86848237F2FULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.impa.ani"},
	{0xC634FA6437EBA54BULL, "mcr310_sibling6076403.cl"},
	{0xC635233B097DDCEFULL, "795138121.ogg"},
	{0xC63557E76B351704ULL, "hlac_ca_caverns1_24424_14x.dds"},
	{0xC63563B8CE9E3437ULL, "6214938_m.fxe"},
	{0xC6357BBB6F2CD29DULL, "vxlt_gallowscourtyard_74b_0_0.vlm"},
	{0xC63599435B960EE7ULL, "vxlt_0_lowtown_a2_n_byl_0_0.vlm"},
	{0xC635BD0594F21931ULL, "387142912.ogg"},
	{0xC6360CB6AFEFF05CULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\originals\\appearfromopen.gad"},
	{0xC6360F24E5364B54ULL, "w_encounter2_30_0.phy"},
	{0xC6367FA863914091ULL, "vxht_mansion2b_1z7_0_0.vlm"},
	{0xC636BE5082D5596CULL, "6093678_f.fxe"},
	{0xC6379500BEE3EEE8ULL, "6155922_m.fxe"},
	{0xC63886A69B441CD8ULL, "vxht_playershouse_n_1q9_0_0.vlm"},
	{0xC638A59DCA2AF27BULL, "vxlt_0_lowtown_n_gob_0_1.vlm"},
	{0xC638A5E61FADC74FULL, "zz_followers_banter6158295.cl"},
	{0xC639165AD952D510ULL, "6172877_m.fxe"},
	{0xC63920FA3BD241BCULL, "32787041.ogg"},
	{0xC639366B51BD3FDDULL, "vxca_caverns1_mines_iio_0_0.vlm"},
	{0xC639C553BDE57358ULL, "vxlt_0_lowtown_a2_n_c_eqj_0_0.vlm"},
	{0xC639E9924A257F3BULL, "zz_goto.ncs"},
	{0xC63A34289CE22D48ULL, "6035206_m.fxe"},
	{0xC63AAE6D41CA2827ULL, "llac_lt_hangedman_-1_11.dds"},
	{0xC63B46B1D7C516E0ULL, "dae000ip_crowd_lm2_talk13.utp"},
	{0xC63B552FD0EA7C49ULL, "vxht_bartrand_n_5vc_0_0.vlm"},
	{0xC63B64511B7805CEULL, "721484543.ogg"},
	{0xC63BE2791BBE123BULL, "hlac_lt_gallowstemplar_2382_0z.dds"},
	{0xC63C0BB6AFF45A5AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\originals\\appearfromopen.ani"},
	{0xC63CB1914D8081A0ULL, "1018052687.ogg"},
	{0xC63CE154545C6A6BULL, "hlac_lt_0_lowtown_n_65534_5y.dds"},
	{0xC63D0624E53C5AFCULL, "w_encounter2_30_0.rml"},
	{0xC63D378891A21E37ULL, "6119460_m.fxe"},
	{0xC63EA21266ECE37EULL, "dae000_gen_turn_in_2.cnv"},
	{0xC63EF6F181F756F9ULL, "6136461_m.fxe"},
	{0xC640198EFF94AF55ULL, "zz_roguedodge.cut"},
	{0xC640DB73E9017923ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masf.msh"},
	{0xC640E973E901917DULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masf.mmh"},
	{0xC640EC24E5401751ULL, "w_encounter2_30_0.msh"},
	{0xC640FA24E5402F6CULL, "w_encounter2_30_0.mao"},
	{0xC6410624E540430FULL, "w_encounter2_30_0.mmh"},
	{0xC6416F5A08406479ULL, "hlac_lt_backalley_g_1_0z.dds"},
	{0xC641841266EEE58EULL, "dae000_gen_turn_in_2.bnk"},
	{0xC64186DA22760A66ULL, "34064986.ogg"},
	{0xC641B639B8F5ECDDULL, "6171073_m.fxe"},
	{0xC6423E1DF466B72DULL, "vxht_mansion_n_6bc_0_0.vlm"},
	{0xC64358C78029C58AULL, "598877309.ogg"},
	{0xC6438E696013B533ULL, "331434821.ogg"},
	{0xC644CF2415B50FDBULL, "hlac_lt_0_lowtown_a2_n_65536_12y.dds"},
	{0xC645766FB418C818ULL, "vxca_caverns2prc_4zm_0_0.vlm"},
	{0xC645E08492A18A78ULL, "hla_ht_0_hightown_-65535_0y.dds"},
	{0xC6468AF114869623ULL, "6098778_m.fxe"},
	{0xC646C6E3E3CE0171ULL, "hlac_w_encounter2_camp_327679_0y.dds"},
	{0xC64799986FE31100ULL, "vxdr_0_deeproads_fh8_0_0.vlm"},
	{0xC648E20A64090E2CULL, "6101721_m.fxe"},
	{0xC648E79C75E4AA06ULL, "zz_followers_banter6158313.cl"},
	{0xC6493EC7857E8CECULL, "6070335_m.fxe"},
	{0xC6495E4BC6BC1E39ULL, "follower_banter6164700.cl"},
	{0xC6498A82AB9A732CULL, "hlac_lt_0_lowtown_a2_n_131071_7x.dds"},
	{0xC649C70CC2E094C9ULL, "vxht_bartrand_n_5yb_0_0.vlm"},
	{0xC649D8A484522436ULL, "121526186.ogg"},
	{0xC64A921BFD6E8A28ULL, "hurlock_mace_shield.utc"},
	{0xC64B457BF1CA955EULL, "gen00fl_dog6145229.cl"},
	{0xC64B66DA081E54D5ULL, "vxht_0_hightown_a3_a7q_0_0.vlm"},
	{0xC64B7F5CD49451E0ULL, "6052555_m.fxe"},
	{0xC64BB1FA7F982A0AULL, "226027050.ogg"},
	{0xC64C04980F307AE1ULL, "6036722-takehimandtry.cut"},
	{0xC64C5E6BB0DC5A06ULL, "fen000pt_main.ncs"},
	{0xC64CAB006C1EE396ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1rootbulb.mao"},
	{0xC64CB7006C1EF835ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1rootbulb.mmh"},
	{0xC64CB88546D17E2AULL, "319639694.ogg"},
	{0xC64CB9006C1EFBABULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1rootbulb.msh"},
	{0xC64CF0851074E920ULL, "llac_lt_0_lowtown_a2_131072_3.dds"},
	{0xC64D0F39AC9EDA63ULL, "vxdr_primevaltomb_7bk_0_1.vlm"},
	{0xC64D68CE73ED6F79ULL, "6131561_m.fxe"},
	{0xC64E9C2F33FA7615ULL, "zz_tbe100_barber6191011.cl"},
	{0xC64EB486719E6E2CULL, "152225450.ogg"},
	{0xC64EEA4E630EF9AFULL, "417042661.ogg"},
	{0xC64F019306D97E59ULL, "one100dg_vanard6176015.cl"},
	{0xC64F6BA438AD45CAULL, "hlac_lt_0_lowtown_-131073_3x.dds"},
	{0xC64F79C689055EE9ULL, "gen00fl_aveline6105468.cl"},
	{0xC64FC6B8C6AAA567ULL, "follower_banter6155715.cl"},
	{0xC6502816AAEC055DULL, "6174373_m.fxe"},
	{0xC650A94A0777DDFEULL, "ht_brothel_n.arl"},
	{0xC650B50417C01B02ULL, "gen00fl_fenris6127976.cl"},
	{0xC650F2EBE0710E5EULL, "vxlt_0_lowtown_docks_bn2_0_0.vlm"},
	{0xC6513EAD4D0DF2DBULL, "495831856.ogg"},
	{0xC65216A56E8C35F7ULL, "vxlt_0_lowtown_docks_a3_bg7_0_0.vlm"},
	{0xC652370B82A6B5A2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02a.msh"},
	{0xC652390B82A6B958ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02a.mmh"},
	{0xC65257F2F9A1CA55ULL, "555045276.ogg"},
	{0xC65271A6FE50C3EFULL, "gen_im_wep_war_one_act3_03.uti"},
	{0xC652C24564C2E146ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\left_hand_gesture_self.ani"},
	{0xC652EEC6C24AF477ULL, "gen00fl_bethany6121313.cl"},
	{0xC652F4487C6D8709ULL, "one300dg_solivitus6048303.cl"},
	{0xC653B4AAF33BCE37ULL, "vxlt_gallowstemplar_n_4qi_0_0.vlm"},
	{0xC654514E206F4D51ULL, "vxlt_0_lowtown_a2_n_ef8_0_0.vlm"},
	{0xC65569E0A415A0B0ULL, "llac_lt_undercitydungeon_-1_34.dds"},
	{0xC655A22AFD38F247ULL, "zz_fol_debug6103809.cl"},
	{0xC6561A2535A970DDULL, "vxca_caverns1_mines_icm_0_0.vlm"},
	{0xC656229087255165ULL, "dre000_cut_sibling_dies6036290.cl"},
	{0xC656259B1724E3CEULL, "6140489_m.fxe"},
	{0xC6562D943FFB4483ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_exit.evt"},
	{0xC65675A64C05B4E4ULL, "263762118.ogg"},
	{0xC656791D799236EEULL, "w_blightlands_34_0.msh"},
	{0xC656871D79924E3FULL, "w_blightlands_34_0.mao"},
	{0xC656931D799262DCULL, "w_blightlands_34_0.mmh"},
	{0xC656C7FFEF909605ULL, "gen00fl_isabela6088256.cl"},
	{0xC657515427B9BD1EULL, "llac_ht_mansion2b_818_6.dds"},
	{0xC65764E7500CDC3EULL, "208979807.ogg"},
	{0xC657DF2C55017928ULL, "459903008.ogg"},
	{0xC6580501A5053FAAULL, "vxw_bonepitdestroyed_a3_237_0_0.vlm"},
	{0xC65899DCDB58007FULL, "w_sundermount_95_0.msh"},
	{0xC6589FDCDB580A21ULL, "w_sundermount_95_0.mmh"},
	{0xC658ABDCDB581EC2ULL, "w_sundermount_95_0.mao"},
	{0xC658D4E0A4F48D19ULL, "634525707.ogg"},
	{0xC659B26740ED3815ULL, "169872824.ogg"},
	{0xC65B1F1DAD858BBCULL, "l_dae121ar_keep.bnk"},
	{0xC65B26403B20D2D4ULL, "gen00fl_dog6167025.cl"},
	{0xC65B3639A4E84E12ULL, "702401039.ogg"},
	{0xC65B66DC9476368EULL, "6147817_m.fxe"},
	{0xC65B8DECFB8B58B0ULL, "vxht_chantry_2t7_0_0.vlm"},
	{0xC65BBC4ED45AA1ACULL, "follower_banter6163405.cl"},
	{0xC65BD95FD7E4D3E2ULL, "art\\vfx\\models\\vfx_rain_fire.mmh"},
	{0xC65BDB5FD7E4D698ULL, "art\\vfx\\models\\vfx_rain_fire.msh"},
	{0xC65C4D19ABCD3189ULL, "vxht_brothel_n_6j8_0_0.vlm"},
	{0xC65CB85782DC92E3ULL, "vxca_caverns2prc_ab3_0_0.vlm"},
	{0xC65CD5EFABF8ADB5ULL, "llac_lt_gallowstemplar_f_2384_3.dds"},
	{0xC65D069440012104ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_exit.gad"},
	{0xC65D2CE2CE4290A7ULL, "vxht_0_hightown_c3m_0_1.vlm"},
	{0xC65D400B82B0CCE3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw02a.phy"},
	{0xC65D6DE77A1EE7BCULL, "hlac_w_woundedcoast_327687_0x.dds"},
	{0xC65D9F3F96F7B2CBULL, "951960438.ogg"},
	{0xC65DAC294BF8EE47ULL, "6043154_m.fxe"},
	{0xC65ED1E889A7E725ULL, "hlac_lt_0_lowtown_docks_n_131071_5x.dds"},
	{0xC65EF4228DB59625ULL, "dae000_bartrand6107556.cl"},
	{0xC65F13D855B5F6A2ULL, "gen00fl_dog6122239.cl"},
	{0xC65FC8355EF44293ULL, "6129107_f.fxe"},
	{0xC65FD85A6AAD51D0ULL, "hlac_ht_mansion_n_3382_1y.dds"},
	{0xC65FEF880A533F46ULL, "dae020cr_chantry2.utc"},
	{0xC6600880031C0505ULL, "94697482.ogg"},
	{0xC6600B4F483477C1ULL, "gen00fl_anders6113046.cl"},
	{0xC66019F08443B380ULL, "hlac_w_encounter2_327680_0z.dds"},
	{0xC6603B159A8E2446ULL, "108108365.ogg"},
	{0xC660AA4564CEF5B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\left_hand_gesture_self.evt"},
	{0xC660F254C3A02170ULL, "lt_0_lowtown_docks_a3_103_0.mao"},
	{0xC660F654C3A02783ULL, "lt_0_lowtown_docks_a3_103_0.mmh"},
	{0xC6610454C3A03FDDULL, "lt_0_lowtown_docks_a3_103_0.msh"},
	{0xC6614C21D136743EULL, "vxca_caverns2_80h_0_0.vlm"},
	{0xC661500BC6F57827ULL, "6043086_m.fxe"},
	{0xC661D3C8CD811E73ULL, "art\\vfx\\creatures\\fxc_footstep_r_c.impa.evt"},
	{0xC661F154B1F842B9ULL, "art\\vfx\\combat\\vfx_gib_hlk_p_v3.anb"},
	{0xC661F8F6C6903D8AULL, "ch1notnt_s2.psh.fxo"},
	{0xC6624962160E1CABULL, "237084296.ogg"},
	{0xC662BD7C3C9D0E30ULL, "6140638_m.fxe"},
	{0xC662D176DF6E8047ULL, "llac_ca_caverns1_mines_24424_7.dds"},
	{0xC662FF82A2512D78ULL, "652071748.ogg"},
	{0xC663AECD519B069BULL, "gen00fl_aveline6122160.cl"},
	{0xC663B1DFC1B4179FULL, "vxw_woundedcoast_my_0_0.vlm"},
	{0xC66401CF8E546129ULL, "vxca_caverns1_fn0_0_0.vlm"},
	{0xC6640F69854BFD6FULL, "6202775_m.fxe"},
	{0xC6641594400759E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hugging_human_male_exit.ani"},
	{0xC664B371FC3B0212ULL, "vxlt_gallowscourtyard_78u_0_0.vlm"},
	{0xC664CBA9D0668AF6ULL, "6108796_f.fxe"},
	{0xC6658EC5B55AAFB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_magestaff.ani"},
	{0xC665FC182F345318ULL, "vxdr_primevaltomb_53p_0_0.vlm"},
	{0xC66605BEF1CA700BULL, "vxlt_gallowscourtyard_n_6zx_0_0.vlm"},
	{0xC6661666F622916AULL, "835897631.ogg"},
	{0xC6662D04F8CB6389ULL, "meleering2.gda"},
	{0xC66645DF0829CAFFULL, "art\\characters\\creatures\\ragedemon\\animation\\blendtree\\ragedemon_overlay.anb"},
	{0xC666462B56F2ABDBULL, "vxht_chantry_n_2yt_0_0.vlm"},
	{0xC6666D47790C373EULL, "hlac_w_woundedcoast_-131069_0z.dds"},
	{0xC66691A6AA6CC87BULL, "llac_ht_mansion2b_818_41.dds"},
	{0xC6670B68B6B11671ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzstage_0.pwk"},
	{0xC6671DFB45446F5BULL, "vxca_cavern1_haunted_ehf_0_0.vlm"},
	{0xC66736572EE5D06FULL, "6143290_m.fxe"},
	{0xC667AB4564D51640ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\left_hand_gesture_self.gad"},
	{0xC667C789294119EAULL, "store_mcr371_sandal.utm"},
	{0xC6682A87D147A14EULL, "art\\vfx\\combat\\vfx_q_ari_chop_act_c.impa.ani"},
	{0xC66895974D1385E6ULL, "pro000_refugee_female6052085.cl"},
	{0xC668B56FA15C0CACULL, "6136640_m.fxe"},
	{0xC668CCC8CD8730B4ULL, "art\\vfx\\creatures\\fxc_footstep_r_c.impa.gad"},
	{0xC668CDD188D8A43EULL, "dae000_arishok6072830.cl"},
	{0xC66941C8958F096AULL, "zz_followers_banter6158469.cl"},
	{0xC669517FEB6FC895ULL, "6054581_m.fxe"},
	{0xC669B9AA5309E0D7ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha10a.mmh"},
	{0xC669CFAA530A0649ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha10a.msh"},
	{0xC669CFFB462F7644ULL, "640632975.ogg"},
	{0xC669D8B9866C4F0EULL, "6140917_m.fxe"},
	{0xC669FD85D18E2EA4ULL, "1020114166.ogg"},
	{0xC66A051C08559ED0ULL, "follower_banter6149232.cl"},
	{0xC66A4A4A17C0E635ULL, "sse_ss_templars_f_1.gda"},
	{0xC66A4B79E86E112CULL, "gen00fl_anders6113108.cl"},
	{0xC66B8796F6832DD9ULL, "6078006_m.fxe"},
	{0xC66B89B882E5BFC1ULL, "vxca_cavern1_haunted_k2t_0_0.vlm"},
	{0xC66BC2E7AB6BA19BULL, "hlac_lt_0_lowtown_n_-4_0x.dds"},
	{0xC66C0D308F0E5F23ULL, "6095333_m.fxe"},
	{0xC66C1F494489B602ULL, "t3_shl_lfi.tnt"},
	{0xC66C5C367FF63388ULL, "595388525.ogg"},
	{0xC66C8BC5B560C916ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_magestaff.gad"},
	{0xC66CB891E02C5E8AULL, "817436519.ogg"},
	{0xC66CE4DAEBAC1200ULL, "scaleform_sf_depth_p0.psh.fxo.dx11.assembly"},
	{0xC66D4CF0F980523BULL, "6057087_m.fxe"},
	{0xC66D99FA1353DE23ULL, "w_bonepitdestroyed_a3_116_0.mao"},
	{0xC66D9DFA1353E570ULL, "w_bonepitdestroyed_a3_116_0.mmh"},
	{0xC66DABFA1353FC8AULL, "w_bonepitdestroyed_a3_116_0.msh"},
	{0xC66DB17AB4ACECDAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat2a.mmh"},
	{0xC66DB57AB4ACF311ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat2a.mao"},
	{0xC66DC37AB4AD0B20ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_walkflat2a.msh"},
	{0xC66DDC3E90CFA68CULL, "llac_lt_gallowscourtyard_131069_0.dds"},
	{0xC66DFD62D10A5968ULL, "hlac_lt_0_lowtown_docks_a3_131072_0y.dds"},
	{0xC66EDBC8CD8BB592ULL, "art\\vfx\\creatures\\fxc_footstep_r_c.impa.ani"},
	{0xC66F5060AD216DA6ULL, "vxw_woundedcoast_h3_0_0.vlm"},
	{0xC66FB8DFBA232C43ULL, "vxlt_backalley_n_1dg_0_0.vlm"},
	{0xC66FD8ED61EEDADDULL, "570139987.ogg"},
	{0xC66FFAD0C53099EDULL, "6076435_m.fxe"},
	{0xC670F76914CB7462ULL, "vxlt_darktown_6k7_0_0.vlm"},
	{0xC671DED188414890ULL, "444889582.ogg"},
	{0xC671ED846AD16474ULL, "6066479_m.fxe"},
	{0xC671F754C3AE9CA8ULL, "lt_0_lowtown_docks_a3_103_0.phy"},
	{0xC671FBD62DD3F8E0ULL, "pro000_opening_pc.cnv"},
	{0xC67262FB3CF5B689ULL, "vxht_0_hightown_a3_9bf_0_0.vlm"},
	{0xC6727FDED394CA9CULL, "607671641.ogg"},
	{0xC672D881EDEFB951ULL, "vxlt_warehouse_vh_0_0.vlm"},
	{0xC67333ACC0EDB6F0ULL, "130394062.ogg"},
	{0xC673A6C5B5671675ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_magestaff.evt"},
	{0xC6748EF3579C73CAULL, "zz_one_debug6099771.cl"},
	{0xC674931D79AB602FULL, "w_blightlands_34_0.rml"},
	{0xC674A5160D25349AULL, "vxlt_warehouse_12b_0_0.vlm"},
	{0xC6754F309D8AFE0FULL, "6088783_f.fxe"},
	{0xC6756B736F7773A2ULL, "w_bonepitmining_a2_226_0.msh"},
	{0xC6756D736F777758ULL, "w_bonepitmining_a2_226_0.mmh"},
	{0xC67579736F778BFBULL, "w_bonepitmining_a2_226_0.mao"},
	{0xC675983C7F01716EULL, "6079019_m.fxe"},
	{0xC675D188C5B66B9DULL, "6125366_m.fxe"},
	{0xC675DDD62DD7AE18ULL, "pro000_opening_pc.bnk"},
	{0xC676071798C23C3FULL, "mcr371_scared_mage_f6087709.cl"},
	{0xC6762287D153D09FULL, "art\\vfx\\combat\\vfx_q_ari_chop_act_c.impa.evt"},
	{0xC67668E24535B1D2ULL, "hlac_lt_undercityencounter_-1_44x.dds"},
	{0xC677D7E565DA3557ULL, "712116470.ogg"},
	{0xC67805403B010542ULL, "526580839.ogg"},
	{0xC67826282DF486DEULL, "dae000_merchant_hm_mob6174228.cl"},
	{0xC67842CB56DDE234ULL, "gen00fl_varric6078124.cl"},
	{0xC678E380E8B40AECULL, "hlac_lt_hangedman_n_-1_6x.dds"},
	{0xC6796AC4C1D2B5C3ULL, "zz_and_debug6178591.cl"},
	{0xC679E486627CA7A6ULL, "691926294.ogg"},
	{0xC67AAD60E9735ADEULL, "vxw_bonepitdestroyed_a3_1w8_0_0.vlm"},
	{0xC67AC5608FD47F99ULL, "qun120cr_javaris6042839.cl"},
	{0xC67AE1BBAF85B91EULL, "hlac_lt_0_lowtown_a3_n_c_0_1x.dds"},
	{0xC67B0895B34F60FDULL, "hlac_lt_0_lowtown_a2_n_-65536_0y.dds"},
	{0xC67B275DCD93FB1BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_04tree.mmh"},
	{0xC67B28FE5C5F7A06ULL, "gen_im_trs_cot_arc_sbw_r0.uti"},
	{0xC67B335DCD940FB8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_04tree.mao"},
	{0xC67B455DCD942E05ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_04tree.msh"},
	{0xC67B6ABC2881DC3EULL, "hlac_ht_0_hightown_n_-65537_1z.dds"},
	{0xC67C030D698ED7B6ULL, "art\\characters\\creatures\\spider\\c_giaspidr_big.phy"},
	{0xC67C162CA531A62AULL, "vxlt_0_lowtown_f1j_0_0.vlm"},
	{0xC67C2A94DE25AB9EULL, "pro000_meeran6050092.cl"},
	{0xC67CC99DFA1F6CDEULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.dura.ani"},
	{0xC67CDE8E391B7103ULL, "gen00fl_fenris6110887.cl"},
	{0xC67CFFCD2F2C93BDULL, "vxlt_darktown_519_0_0.vlm"},
	{0xC67D0387D159BA88ULL, "art\\vfx\\combat\\vfx_q_ari_chop_act_c.impa.gad"},
	{0xC67D19C05E106368ULL, "188418633.ogg"},
	{0xC67D49877FBC2A6FULL, "921381959.ogg"},
	{0xC67DBE0CE16622E2ULL, "6136721_m.fxe"},
	{0xC67DC23D9F178626ULL, "vxw_woundedcoast_e4_0_0.vlm"},
	{0xC67E0057B60BF53BULL, "llac_lt_backalley_n_0_0.dds"},
	{0xC67E0631853F51BFULL, "301690635.ogg"},
	{0xC67E0DC185B8A2ACULL, "6079282_m.fxe"},
	{0xC67E268608C6D8BEULL, "vxht_0_hightown_b5n_0_0.vlm"},
	{0xC67E2ABCAC1944C3ULL, "vxlt_backalley_n_1of_0_0.vlm"},
	{0xC67E7FDCDB78A34AULL, "w_sundermount_95_0.rml"},
	{0xC67E8D3F4AA45005ULL, "ht_0_hightown_n_35_0.rml"},
	{0xC67EE2CBFB81F3AFULL, "810325318.ogg"},
	{0xC67F680BA4748F21ULL, "llac_dr_0_deeproads_17039_1.dds"},
	{0xC68014A3DA9D2DE1ULL, "6179521_m.fxe"},
	{0xC6805DFE763631F5ULL, "altar_n.dds"},
	{0xC6805E5A9AB2DCB8ULL, "6101199_f.fxe"},
	{0xC680C5A713F2CFEEULL, "mag360_thrask6043562.cl"},
	{0xC6818FE178E53C44ULL, "vxlt_gamlenshouse_n_6z_0_0.vlm"},
	{0xC681EF97DABEAC0FULL, "fen210_hadriana6110529.cl"},
	{0xC6824BC62923312EULL, "551141302.ogg"},
	{0xC682B643EAD1FAECULL, "6170163_m.fxe"},
	{0xC682BFADE6395A6CULL, "vrc221_bartrand6065396.cl"},
	{0xC68329E53ADEDCDBULL, "vxht_chantry_n_2ov_0_0.vlm"},
	{0xC6837C9679045D34ULL, "6023970_f.fxe"},
	{0xC6838E808AB8F6C9ULL, "vxht_bartrand_n_5mf_0_0.vlm"},
	{0xC683C1EF4B70343EULL, "6100276_m.fxe"},
	{0xC683FD87A6A2E9AEULL, "llac_dr_primevalentrance_1725_1.dds"},
	{0xC6848BB1B134C199ULL, "vxlt_0_lowtown_docks_a3_9re_0_0.vlm"},
	{0xC685EC0D69970557ULL, "art\\characters\\creatures\\spider\\c_giaspidr_big.msh"},
	{0xC686020D69972AC9ULL, "art\\characters\\creatures\\spider\\c_giaspidr_big.mmh"},
	{0xC6865986DF094ED7ULL, "llac_w_bonepit_196610_0.dds"},
	{0xC686DFB68A34F833ULL, "dae000_bodhan_letter6144257.cl"},
	{0xC6873B232E5CBB24ULL, "follower_banter6151251.cl"},
	{0xC6886A1546776B0AULL, "900227217.ogg"},
	{0xC6887D36FFBF5009ULL, "qun120cr_javaris6042729.cl"},
	{0xC6889FB86DA091C1ULL, "dre000_cut_bartrandbetrayal.bnk"},
	{0xC688FA76B94AF20BULL, "6043067_f.fxe"},
	{0xC6891F3DDB69735EULL, "vxw_bonepitdestroyed_a3_1x9_0_0.vlm"},
	{0xC689592F92BECE84ULL, "follower_banter6160686.cl"},
	{0xC689E19DFA2A1FAFULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.dura.evt"},
	{0xC68A1A172D5364D3ULL, "vxht_0_hightown_a3_c88_0_0.vlm"},
	{0xC68B365812C5EA49ULL, "follower_banter6166412.cl"},
	{0xC68BC1B86DA30131ULL, "dre000_cut_bartrandbetrayal.cnv"},
	{0xC68C015DC0DF745BULL, "dae000_bartrand6027520.cl"},
	{0xC68C341A910D9EA6ULL, "vxw_woundedcoast_b5_0_0.vlm"},
	{0xC68C9C999E0F5D43ULL, "vxlt_backalley_n_1ji_0_0.vlm"},
	{0xC68CBAFB905A58ECULL, "vxw_encounter2_camp_ou_0_0.vlm"},
	{0xC68D4D3B0727FB9DULL, "ach_kill_strider.dds"},
	{0xC68D8BA1DAB4183AULL, "mag170_thrask6020528.cl"},
	{0xC68DADE9EBD30BC4ULL, "226921112.ogg"},
	{0xC68DEE796A516164ULL, "6066624_m.fxe"},
	{0xC68E44198882DE5FULL, "mag311_keran6185078.cl"},
	{0xC68E989AE4FA39ABULL, "600545988.ogg"},
	{0xC68F98802171086EULL, "w_blightlands_230_0.mmh"},
	{0xC68FA48021711C05ULL, "w_blightlands_230_0.mao"},
	{0xC68FB005AE5F4482ULL, "hlac_ht_0_hightown_-196609_0y.dds"},
	{0xC68FB2802171345CULL, "w_blightlands_230_0.msh"},
	{0xC68FC0CC8692B97EULL, "hlac_ht_keep_3732_17x.dds"},
	{0xC6904E91CCAF9674ULL, "6097595_f.fxe"},
	{0xC6909E7C5CA6FEB4ULL, "vxlt_gallowscourtyard_4we_0_0.vlm"},
	{0xC690B34CA3F566D5ULL, "vxht_0_hightown_a3_by4_0_0.vlm"},
	{0xC690CECB8E13727BULL, "vxht_mansion_n_6ih_0_0.vlm"},
	{0xC690E29DFA304058ULL, "art\\vfx\\combat\\vfx_c_mer_rush_c.dura.gad"},
	{0xC6917451D513DE74ULL, "vxlt_gallowstemplar_n_2or_0_0.vlm"},
	{0xC691770274BCA3DCULL, "vxdr_0_deeproads_fbf_0_0.vlm"},
	{0xC6919DC6E1F53670ULL, "207571936.ogg"},
	{0xC691B514547F5C09ULL, "hlac_w_woundedcoast_-196601_0x.dds"},
	{0xC691D39841FE66C5ULL, "hlac_ht_0_hightown_a3_-131073_1z.dds"},
	{0xC691FF78717A8E96ULL, "art\\vfx\\materials\\vfx_twirl_add.mao"},
	{0xC6928C54152EA7DFULL, "follower_barks6153343.cl"},
	{0xC692B76F03E82AAFULL, "pin_cl.mao"},
	{0xC69313D578E323E5ULL, "zz_followers_banter6158367.cl"},
	{0xC693607F1EEA6128ULL, "6088071_m.fxe"},
	{0xC69392700DB10B85ULL, "vxca_caverns1_mines_eh8_0_0.vlm"},
	{0xC693B09C9024A14AULL, "dae000ip_crowd_ff7_talk17.utp"},
	{0xC694A90035DACFBEULL, "843000885.ogg"},
	{0xC696C40A5B0F2025ULL, "vxlt_0_lowtown_docks_a3_n_d0d_0_1.vlm"},
	{0xC697385064BA77A4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera.msh"},
	{0xC697492E9BFEE7E2ULL, "6168177_m.fxe"},
	{0xC6974A5064BA960DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera.mao"},
	{0xC6974E5064BA9D46ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera.mmh"},
	{0xC6978D246BF983FFULL, "vxlt_0_lowtown_docks_a3_n_d6a_0_0.vlm"},
	{0xC697B1C447F2822AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_exit_elf_male.gad"},
	{0xC698162EB1A2AA64ULL, "gen00fl_varric6093677.cl"},
	{0xC6982A708983C3B2ULL, "216844773.ogg"},
	{0xC698CBDF8BFC3FC5ULL, "6202269_m.fxe"},
	{0xC69938895F3A50F3ULL, "lgt320_shw_ors_black_op6156328.cl"},
	{0xC69A3A552130F9D8ULL, "6095657_m.fxe"},
	{0xC69ABB26FB39B346ULL, "729087366.ogg"},
	{0xC69AC9FD05B4265EULL, "6037615_m.fxe"},
	{0xC69ADEAFCCE2F84DULL, "6058930_m.fxe"},
	{0xC69AED80DDDC83FDULL, "6053540_m.fxe"},
	{0xC69B2CD88250716CULL, "vxw_encounter2_camp_jj_0_0.vlm"},
	{0xC69BF2432E6260DDULL, "03d_smokess.mao"},
	{0xC69C4BDA15B10B84ULL, "one140cr_charade6173699.cl"},
	{0xC69C74F43AF867D6ULL, "vxht_0_hightown_a2_8q0_0_0.vlm"},
	{0xC69CE63F9A1A3848ULL, "939261604.ogg"},
	{0xC69CF883405E13BFULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_0z.dds"},
	{0xC69D7953898C799FULL, "vxw_woundedcoast_ye_0_0.vlm"},
	{0xC69DBF0D01740FCDULL, "llac_lt_0_lowtown_docks_a3_n_-4_2.dds"},
	{0xC69DD5F857381FFAULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxa.phy"},
	{0xC69E618AFBABABC6ULL, "gen00fl_aveline6132742.cl"},
	{0xC69EB2F5551877F7ULL, "720471653.ogg"},
	{0xC69EB3D071567B86ULL, "lgt150_nat_were_supporter6110989.cl"},
	{0xC69EC533136255AFULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.dura.gad"},
	{0xC69F40A880098AFBULL, "vxht_mansion_n_6lk_0_0.vlm"},
	{0xC69F7A5255F47B26ULL, "571743906.ogg"},
	{0xC69FCA1B3D3E1EFAULL, "vxdr_primevaltomb_2t8_0_0.vlm"},
	{0xC69FED64CC7DE447ULL, "dae000_pickpocket_scene6175098.cl"},
	{0xC6A041F7914D7C53ULL, "plt_and000pt_midstage.plo"},
	{0xC6A07E880C5DE537ULL, "vxht_mansion2_3bs_0_0.vlm"},
	{0xC6A0967F5E2E7F1FULL, "vxw_bonepit_1qs_0_0.vlm"},
	{0xC6A0AFEA67ADE661ULL, "vxca_caverns1_e89_0_0.vlm"},
	{0xC6A0E749D5E13795ULL, "zz_rebenger_test.cut"},
	{0xC6A0ED8761AFA93CULL, "6195845_m.fxe"},
	{0xC6A0FB0C45587D93ULL, "468885576.ogg"},
	{0xC6A0FB81E1F26B79ULL, "125070664.ogg"},
	{0xC6A115540D69EC11ULL, "588857127.ogg"},
	{0xC6A1389CC20CFE5EULL, "vxlt_0_lowtown_a3_n_c_eqq_0_0.vlm"},
	{0xC6A15B13C0A6351EULL, "vxht_mansion2_1sg_0_0.vlm"},
	{0xC6A1715A1E909B97ULL, "lt_0_lowtown_docks_n_213_0.mmh"},
	{0xC6A1755A1E90A2A4ULL, "lt_0_lowtown_docks_n_213_0.mao"},
	{0xC6A1875A1E90C109ULL, "lt_0_lowtown_docks_n_213_0.msh"},
	{0xC6A18BE710402C66ULL, "llac_ht_bartrand_n_2995_4.dds"},
	{0xC6A1CF8446E96564ULL, "dae000cr_brothel_stumps.utc"},
	{0xC6A23BEE04A3825BULL, "6139149_m.fxe"},
	{0xC6A2455064C49511ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_quartera.phy"},
	{0xC6A2CD9A8A9B01A1ULL, "121753819.ogg"},
	{0xC6A2FC7C592BB978ULL, "176033277.ogg"},
	{0xC6A3D483C9374AB1ULL, "6108811_m.fxe"},
	{0xC6A46298F4BE4A9FULL, "33043600.ogg"},
	{0xC6A46DCC561000A1ULL, "6174544_m.fxe"},
	{0xC6A4A7603EB4A2ECULL, "hlac_lt_0_lowtown_a2_n_c_65538_0y.dds"},
	{0xC6A53F75BC09FE8CULL, "696181322.ogg"},
	{0xC6A56796A776C484ULL, "933202487.ogg"},
	{0xC6A5DA331368985CULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.dura.evt"},
	{0xC6A65B5CC2BD960FULL, "105059838.ogg"},
	{0xC6A66894836C6F96ULL, "t1_mue_wh2.tnt"},
	{0xC6A699908097C14EULL, "vxca_caverns1_mines_jfn_0_0.vlm"},
	{0xC6A69FC5E29724F1ULL, "130941597.ogg"},
	{0xC6A7215045841730ULL, "zz_vault_debug6203185.cl"},
	{0xC6A731F877E082C5ULL, "vxw_bonepitmining_a2_1xt_0_0.vlm"},
	{0xC6A763817FCB8CF7ULL, "ch1notnt_s2y.psh.fxo"},
	{0xC6A764722652CA49ULL, "577063298.ogg"},
	{0xC6A7B9722A1CFA4AULL, "dae000ip_crowd_ff4_talk9.utp"},
	{0xC6A7C908A29298EDULL, "538118832.ogg"},
	{0xC6A86B4C721DF090ULL, "hlac_dr_0_deeproads_5840_3z.dds"},
	{0xC6A87442ADD92BEAULL, "vxca_caverns1_fmv_0_0.vlm"},
	{0xC6A8789D444CA47FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f.gad"},
	{0xC6A8C3E995A409BCULL, "combat_st_r1_a1.gda"},
	{0xC6A8D3AD014D071AULL, "cd_mag320_mage_conspirator.bnk"},
	{0xC6A917D474F9CFA6ULL, "vxw_woundedcoast_t7_0_0.vlm"},
	{0xC6A9805381FB8E43ULL, "vxlt_backalley_n_1pk_0_0.vlm"},
	{0xC6A99EB5744689ECULL, "vxw_encounter2_camp_ek_0_0.vlm"},
	{0xC6AAA9F73BDF2569ULL, "333646879.ogg"},
	{0xC6AB12471D831AD8ULL, "zz_suhwan_animtest.cut"},
	{0xC6AB8CC44802EC10ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_exit_elf_male.ani"},
	{0xC6ABEB307B82921FULL, "vxw_woundedcoast_vd_0_0.vlm"},
	{0xC6ABEED1A92D0374ULL, "gen00fl_anders6125202.cl"},
	{0xC6ABEF5FA9C76A26ULL, "123554008.ogg"},
	{0xC6AC00B309B84AC6ULL, "vxlt_0_lowtown_n_gbq_0_0.vlm"},
	{0xC6AC050E0A47844EULL, "856710487.ogg"},
	{0xC6AC7B1EA87E4685ULL, "128237349.ogg"},
	{0xC6ACEE9637512B78ULL, "vxlt_0_lowtown_n_cif_0_0.vlm"},
	{0xC6AD895DFEDE88F4ULL, "zz_car_debug6122486.cl"},
	{0xC6ADD8467890BBACULL, "vxca_caverns2_7xe_0_0.vlm"},
	{0xC6ADE7C7DCC82C4DULL, "vxht_chantry_n_28e_0_0.vlm"},
	{0xC6AE2C0490C17DEDULL, "350671344.ogg"},
	{0xC6AE3D5246D8539CULL, "dae000cr_grand_cleric6147678.cl"},
	{0xC6AE5ABC58A8D4DCULL, "vxdr_0_deeproads_fhd_0_0.vlm"},
	{0xC6AE6D9D4450FD6CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f.evt"},
	{0xC6AED169ACE6284EULL, "hlac_ca_caverns2_5802_16z.dds"},
	{0xC6AF085C5024979FULL, "vxw_bonepit_1zr_0_0.vlm"},
	{0xC6AF0BE1A4A171E2ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.dura.gad"},
	{0xC6AF2CC1830D17ABULL, "317302058.ogg"},
	{0xC6AF5B0D1548D936ULL, "llac_lt_0_lowtown_docks_a3_196607_2.dds"},
	{0xC6AF77110959A034ULL, "hlac_lt_0_lowtown_a2_65535_9z.dds"},
	{0xC6AF78E9E89384CFULL, "hlac_ht_chantry_2534_33z.dds"},
	{0xC6AFA421686C70FBULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkd.mmh"},
	{0xC6AFAA79B40316DEULL, "vxlt_0_lowtown_a3_n_c_etr_0_0.vlm"},
	{0xC6AFB021686C8498ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkd.mao"},
	{0xC6AFBEB62E09F403ULL, "6107661_m.fxe"},
	{0xC6AFC221686CA3E5ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkd.msh"},
	{0xC6AFDE568C0E59BEULL, "6069803_m.fxe"},
	{0xC6AFF82D89EA9EA7ULL, "lgt260cr_eau_tmp_angelique.utc"},
	{0xC6B02C52DF39CBA7ULL, "art\\vfx\\materials\\vfx_d_blast_nodecay.mao"},
	{0xC6B039EF3E6E00D1ULL, "6021295_m.fxe"},
	{0xC6B0425964B6B5EFULL, "6091688_f.fxe"},
	{0xC6B0602F0DB29A3FULL, "aip_follower_aveline.gda"},
	{0xC6B11649F65B328BULL, "gen00fl_isabela6015685.cl"},
	{0xC6B18C3F4ACFA82DULL, "ht_0_hightown_n_35_0.anb"},
	{0xC6B1BC7152EA81A7ULL, "6176489_m.fxe"},
	{0xC6B2325D9F48A172ULL, "870216042.ogg"},
	{0xC6B28B7A509FEE2AULL, "fen210_hadriana6110395.cl"},
	{0xC6B29E053449F687ULL, "269765891.ogg"},
	{0xC6B328FD0C4EFB1DULL, "vxw_bonepitdestroyed_a3_19a_0_0.vlm"},
	{0xC6B36D9B5AE5D218ULL, "vxw_bonepit_lp_0_0.vlm"},
	{0xC6B3786EA6298F21ULL, "vxht_mansion2_1ar_0_0.vlm"},
	{0xC6B3E2331374E375ULL, "art\\vfx\\combat\\vfx_d_qun_ari_buff_c.dura.ani"},
	{0xC6B4C7CB4B618105ULL, "969898200.ogg"},
	{0xC6B51208FC92D56DULL, "cage_02.mao"},
	{0xC6B51301D854209FULL, "w_bonepit_4_0.rml"},
	{0xC6B52EE1A4A618D9ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.dura.evt"},
	{0xC6B562236FEE07C7ULL, "gen_im_trs_qun_horn_01.uti"},
	{0xC6B5C4898CE5728AULL, "6086967_m.fxe"},
	{0xC6B5E7CC7B77CC3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_in.gad"},
	{0xC6B60605918890D8ULL, "llac_ca_caverns2_5802_61.dds"},
	{0xC6B61B24AF440805ULL, "cd_mag130_danzig.wbd"},
	{0xC6B6C28EB931D245ULL, "555510031.ogg"},
	{0xC6B6C5E11ED9B313ULL, "character_character_p8.psh.fxo.dx11"},
	{0xC6B6DBED57161C0AULL, "343626080.ogg"},
	{0xC6B701C567B8DB93ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexb.phy"},
	{0xC6B73C111BE7D866ULL, "436913590.ogg"},
	{0xC6B7E22F310E60D1ULL, "vxht_keep_n_63a_0_1.vlm"},
	{0xC6B8C2E59196BBE8ULL, "vxht_mansion2b_1si_0_0.vlm"},
	{0xC6B90669D4FAF606ULL, "398426158.ogg"},
	{0xC6B91BD4695F2AC2ULL, "204468662.ogg"},
	{0xC6B93DA71EA44C65ULL, "6174517_m.fxe"},
	{0xC6B9D9E6F97D1AF6ULL, "268119524.ogg"},
	{0xC6BA1157BCD411B9ULL, "6164053_m.fxe"},
	{0xC6BA2B18DC9E132AULL, "196568007.ogg"},
	{0xC6BA46B06BE54582ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfsit_3.mao"},
	{0xC6BA5D0D6D78AA9FULL, "vxw_woundedcoast_sg_0_0.vlm"},
	{0xC6BA64BC74D4AABBULL, "w_woundedcoast_141_0.mmh"},
	{0xC6BA70BC74D4BF58ULL, "w_woundedcoast_141_0.mao"},
	{0xC6BA82BC74D4DDA5ULL, "w_woundedcoast_141_0.msh"},
	{0xC6BB483F24A6B6D3ULL, "gen_im_arm_cht_hvy_act0_02.uti"},
	{0xC6BB780FB407B004ULL, "vxlt_gallowstemplar_n_5wi_0_0.vlm"},
	{0xC6BC28A833940E09ULL, "qun221_viscount6070151.cl"},
	{0xC6BC559D445D11C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f.ani"},
	{0xC6BC59532A20541EULL, "444426622.ogg"},
	{0xC6BC912E298C9E45ULL, "hlac_lt_undercitydungeon_-1_8x.dds"},
	{0xC6BCD89569C5A672ULL, "llac_lt_blackemporium_1055_8.dds"},
	{0xC6BCD91317BD915BULL, "ave200_donnic6110621.cl"},
	{0xC6BCEACC7B7DF051ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_in.evt"},
	{0xC6BDFE19A13FD9F3ULL, "6177913_f.fxe"},
	{0xC6BECC9CC616EFC4ULL, "lla_ht_0_hightown_-1_0.dds"},
	{0xC6BED4FFDA42CD5AULL, "vxht_0_hightown_a3_c38_0_0.vlm"},
	{0xC6BF04705D8F58A2ULL, "qun220cr_javaris6071503.cl"},
	{0xC6BF3071372C04B0ULL, "473872645.ogg"},
	{0xC6C068C5B68B73FAULL, "vxht_playershouse_n_1p8_0_0.vlm"},
	{0xC6C0C174E7D4565AULL, "gen00fl_dog6167166.cl"},
	{0xC6C1D19A0B5BAB42ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyo_0t.dds"},
	{0xC6C1EA4B981FA7A1ULL, "vxht_mansion2_1dq_0_0.vlm"},
	{0xC6C26D36C9C3BCE6ULL, "ico_amulet_3.dds"},
	{0xC6C2BA979FD12D25ULL, "6123229_m.fxe"},
	{0xC6C2D9F6E668A341ULL, "lowtown_male_urchin_sta6164232.cl"},
	{0xC6C304236922F49DULL, "vxca_cavern1_haunted_inf_0_0.vlm"},
	{0xC6C336E1A4B263C8ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.dura.ani"},
	{0xC6C3CE8548FC09AEULL, "6165475_m.fxe"},
	{0xC6C4065A702776DCULL, "hlac_lt_gallowstemplar_n_2382_2x.dds"},
	{0xC6C415B25BCCB3C5ULL, "vxw_bonepitmining_a2_1rr_0_0.vlm"},
	{0xC6C432501CF06EA9ULL, "gen00fl_aveline6053275.cl"},
	{0xC6C476107BBCD510ULL, "6096960_m.fxe"},
	{0xC6C49330A5F859D7ULL, "one110_gustav6089784.cl"},
	{0xC6C534EE795B4A99ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_2.ani"},
	{0xC6C57CF9B570F992ULL, "vxw_bonepitdestroyed_a3_17t_0_0.vlm"},
	{0xC6C5D91FCC510005ULL, "hlac_ca_cavern1_haunted_24424_0y.dds"},
	{0xC6C5F4223C3AC0D2ULL, "vxlt_darktown_5ci_0_1.vlm"},
	{0xC6C6356A55F19788ULL, "vxlt_0_lowtown_docks_a3_d18_0_0.vlm"},
	{0xC6C6B0AE16915E68ULL, "zz_followers_banter6159297.cl"},
	{0xC6C6E8C98AE02622ULL, "genlock_emissary_def"},
	{0xC6C71BFE66F1F66BULL, "6113072_f.fxe"},
	{0xC6C720E2F066E433ULL, "vxca_caverns1_mines_gj2_0_0.vlm"},
	{0xC6C734C2838CD468ULL, "vxht_mansion2b_1vf_0_0.vlm"},
	{0xC6C736E5167FF4BCULL, "795120668.ogg"},
	{0xC6C73E9E3F2BAB0EULL, "166187852.ogg"},
	{0xC6C7CA519492E7D6ULL, "164983094.ogg"},
	{0xC6C7E31BEC313615ULL, "348017297.ogg"},
	{0xC6C7EA185679C904ULL, "w_bonepitdestroyed_a3_147_0.msh"},
	{0xC6C7FC185679E86DULL, "w_bonepitdestroyed_a3_147_0.mao"},
	{0xC6C800185679EEA6ULL, "w_bonepitdestroyed_a3_147_0.mmh"},
	{0xC6C825D4F3581F45ULL, "6114502_m.fxe"},
	{0xC6C83B1C8B2AC974ULL, "899864026.ogg"},
	{0xC6C855E9EE17862AULL, "hlac_w_sundermount_65534_0x.dds"},
	{0xC6C8E43DA507272CULL, "llac_w_bonepit_327686_0.dds"},
	{0xC6C992C492FF6F8EULL, "vxht_mansion2b_1ai_0_0.vlm"},
	{0xC6C9A51938BC8AFCULL, "6133926_m.fxe"},
	{0xC6C9E9ECA5FDC884ULL, "vxlt_gallowstemplar_n_5jn_0_0.vlm"},
	{0xC6CA02CC7B88A320ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_r_twist_in.ani"},
	{0xC6CA2EAD905936D8ULL, "739629581.ogg"},
	{0xC6CA3EE80D841A99ULL, "vxdr_0_deeproads_ess_0_0.vlm"},
	{0xC6CA6E07A78E5EEBULL, "vxlt_gallowstemplar_f_2oq_0_0.vlm"},
	{0xC6CA7AC06BCDC8D5ULL, "vxht_0_hightown_a3_be0_0_0.vlm"},
	{0xC6CA840B68DFB506ULL, "stoney2_s.dds"},
	{0xC6CAC1BF41B2AE45ULL, "vxdr_primevaltomb_2t3_0_0.vlm"},
	{0xC6CAE287831B863BULL, "6113084_m.fxe"},
	{0xC6CAEB28900CFF97ULL, "vxlt_0_lowtown_a3_n_c_gi2_0_0.vlm"},
	{0xC6CAFDCB69F53FC9ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robe.phy"},
	{0xC6CB1F71A9D7A16FULL, "851126312.ogg"},
	{0xC6CB6445B4AD9A89ULL, "hlac_w_encounter2_camp_196609_0x.dds"},
	{0xC6CBA8F884360185ULL, "396498309.ogg"},
	{0xC6CC0FEE79612ADBULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_2.gad"},
	{0xC6CC731F3C6888AEULL, "dre000_vampire6064165.cl"},
	{0xC6CC91BBAEF3BC4AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towere.mao"},
	{0xC6CC95BBAEF3C299ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towere.mmh"},
	{0xC6CC9FBBAEF3D387ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towere.msh"},
	{0xC6CCA98796487EC9ULL, "6095942_m.fxe"},
	{0xC6CCCD2407C7427DULL, "dae300_alistair_king6133788.cl"},
	{0xC6CD30B121F90B9DULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lt_lightstreaks.mmh"},
	{0xC6CD46DCCC38E5DAULL, "vxht_0_hightown_a3_c49_0_0.vlm"},
	{0xC6CD8CF384ED431FULL, "529316090.ogg"},
	{0xC6CDA438B1BA811CULL, "hlac_lt_0_lowtown_a2_n_c_196607_1z.dds"},
	{0xC6CDAFC9A933584AULL, "6087850_m.fxe"},
	{0xC6CDEDAD016C4FB8ULL, "cd_mag320_mage_conspirator.wbd"},
	{0xC6CDEF33AEAD130FULL, "dre000_dougal6131152.cl"},
	{0xC6CEB672BED226D7ULL, "hlac_lt_0_lowtown_docks_-65539_6x.dds"},
	{0xC6CED51EEFE8BB72ULL, "290022384.ogg"},
	{0xC6CEDAA2A8818C7AULL, "vxht_playershouse_n_1w9_0_0.vlm"},
	{0xC6CEF4E378D5A293ULL, "mag160_feynriel6033378.cl"},
	{0xC6CF531BA0C7D79FULL, "w_sundermount_143_0.msh"},
	{0xC6CF591BA0C7E241ULL, "w_sundermount_143_0.mmh"},
	{0xC6CF651BA0C7F6E2ULL, "w_sundermount_143_0.mao"},
	{0xC6CF8F9E07864D72ULL, "mag100_cullen6140933.cl"},
	{0xC6CF8FF86D098014ULL, "vxlt_0_lowtown_e2j_0_0.vlm"},
	{0xC6CF9DEE3F293D8BULL, "gen_im_arm_cht_elf_comc_t4.uti"},
	{0xC6CFAB727D128809ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant.phy"},
	{0xC6CFEA24AE542FB0ULL, "558199964.ogg"},
	{0xC6D028F7917675F0ULL, "plt_and000pt_midstage.nss"},
	{0xC6D04F1D69545C71ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.dura.ani"},
	{0xC6D0C4CC0022D601ULL, "vxdr_0_deeproads_fh5_0_0.vlm"},
	{0xC6D10364454AB020ULL, "61873322.ogg"},
	{0xC6D12ACC7AC7E8BCULL, "dae000_arishok6067809.cl"},
	{0xC6D12D112C8955AAULL, "ser121_jethann6105764.cl"},
	{0xC6D12FA65F2C1D26ULL, "6128890_m.fxe"},
	{0xC6D1699FBC57CE14ULL, "899996433.ogg"},
	{0xC6D16E9AE186C7B2ULL, "vxlt_0_lowtown_a2_n_c_fi3_0_0.vlm"},
	{0xC6D176005B190D1DULL, "vxca_cavern1_haunted_ike_0_0.vlm"},
	{0xC6D229A422D79202ULL, "426673882.ogg"},
	{0xC6D2524E40714C64ULL, "vis110_commoner36069056.cl"},
	{0xC6D2E476E41BCD08ULL, "713857722.ogg"},
	{0xC6D3BB1A599BFC63ULL, "189493029.ogg"},
	{0xC6D3E8DB648543BEULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_coma_0n.dds"},
	{0xC6D3EED6A7671212ULL, "vxw_bonepitdestroyed_a3_18u_0_0.vlm"},
	{0xC6D4565978BAAC3FULL, "vxw_encounter2_camp_ed_0_0.vlm"},
	{0xC6D496AB22E5C300ULL, "6048358_m.fxe"},
	{0xC6D5ABDFE9A35FE6ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_5y.dds"},
	{0xC6D5D1E88296BDE3ULL, "vxlt_warehouse_13p_0_0.vlm"},
	{0xC6D62FFD8FD18AFBULL, "486399055.ogg"},
	{0xC6D6890DD056FE7DULL, "mcr370_cut_player_rules.bnk"},
	{0xC6D6B69D3A2EF4DFULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_masf_0n.dds"},
	{0xC6D6F3DB8D2CC990ULL, "ht_0_hightown_n_16_0.rml"},
	{0xC6D749997D18694CULL, "vxht_bartrand_n_5qc_0_0.vlm"},
	{0xC6D75A1D695A8D13ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.dura.gad"},
	{0xC6D793D5B177A960ULL, "6114806_m.fxe"},
	{0xC6D7946940267841ULL, "hf_comtesse_dulci.mor"},
	{0xC6D7ACCC317D1718ULL, "seb300im_low_dagger.uti"},
	{0xC6D7ED0177BA5529ULL, "hlac_w_sundermount_-262145_0y.dds"},
	{0xC6D862AC3EA3FE56ULL, "vfx_woodbits02.dds"},
	{0xC6D971F9792C1027ULL, "m_lgt231ar_tome_final.wbd"},
	{0xC6D9B4C4C609DF27ULL, "vxw_sundermount_2c1_0_0.vlm"},
	{0xC6DA019228AFCD96ULL, "416981407.ogg"},
	{0xC6DA6B0DD05AB385ULL, "mcr370_cut_player_rules.cnv"},
	{0xC6DA86920063C134ULL, "vxlt_0_lowtown_a2_fed_0_0.vlm"},
	{0xC6DB3D969EE9F2EAULL, "493550743.ogg"},
	{0xC6DBD251A6F8E966ULL, "one121ar_warehouse.are"},
	{0xC6DC676DFA4E3A01ULL, "6152368_m.fxe"},
	{0xC6DD9D5CD247C251ULL, "hlac_dr_primevalentrance_1725_0z.dds"},
	{0xC6DDA48D05A1DFFDULL, "gen00fl_anders6125322.cl"},
	{0xC6DDBF10200D007DULL, "42589203.ogg"},
	{0xC6DDFC048F782603ULL, "6187801_m.fxe"},
	{0xC6DE17EEB8C76D64ULL, "vxw_blightlands_a4_0_0.vlm"},
	{0xC6DE5416243E383EULL, "vxdr_primevaltomb_6vr_0_0.vlm"},
	{0xC6DE571D6960A778ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.dura.evt"},
	{0xC6DF020AB175DB42ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_killdagger_t.dds"},
	{0xC6DF8C947B22F2B8ULL, "6111006_f.fxe"},
	{0xC6DFD257E5922460ULL, "655651916.ogg"},
	{0xC6E003A6C866445BULL, "6118430_m.fxe"},
	{0xC6E051070226A3C8ULL, "lowtown_male_alienage1_sta6162727.cl"},
	{0xC6E068D0972493D4ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\prp_zzhumaneye_d.dds"},
	{0xC6E0890D7D63F76FULL, "745515141.ogg"},
	{0xC6E0E4745D6C4ACCULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkforward.ani"},
	{0xC6E11CFD06CBB836ULL, "gen00fl_bethany6101006.cl"},
	{0xC6E16C85F8A7A18FULL, "mag101_grace6034048.cl"},
	{0xC6E17B092B8194FDULL, "isa120_hayder6068784.cl"},
	{0xC6E1D7B3281804CFULL, "hlac_w_sundermount_-327680_0y.dds"},
	{0xC6E1EBA0C10F17DEULL, "186210099.ogg"},
	{0xC6E25410DE256A65ULL, "hlac_w_encounter2_131075_0z.dds"},
	{0xC6E2C93BAAB71B29ULL, "vxlt_foundry_n_z3_0_0.vlm"},
	{0xC6E301CE13415A0CULL, "vxht_0_hightown_a3_9pa_0_0.vlm"},
	{0xC6E30B314EF25472ULL, "803770393.ogg"},
	{0xC6E325DC71C0EDC5ULL, "30954826.ogg"},
	{0xC6E334CCE468CF9CULL, "6144098_m.fxe"},
	{0xC6E353DD31602A29ULL, "457528105.ogg"},
	{0xC6E372B08B0DA397ULL, "6104821_f.fxe"},
	{0xC6E3C8B92740AE79ULL, "vxca_caverns2prc_78w_0_2.vlm"},
	{0xC6E3CC609D0B0750ULL, "6041043_m.fxe"},
	{0xC6E3CEDEF423EE57ULL, "vxlt_0_lowtown_a2_n_ehl_0_0.vlm"},
	{0xC6E40A4A9902703BULL, "6080735_m.fxe"},
	{0xC6E41B295F8E0FD4ULL, "vxlt_0_lowtown_a2_n_go4_0_0.vlm"},
	{0xC6E423A45C0C7FC6ULL, "vxca_caverns2prc_7vh_0_0.vlm"},
	{0xC6E4ADC37DE9D796ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_action.ani"},
	{0xC6E4C08AF16C4C38ULL, "vxht_mansion_n_6t4_0_0.vlm"},
	{0xC6E52768571D523BULL, "600711370.ogg"},
	{0xC6E5C099E3368A37ULL, "dae121ip_at_hightown.ncs"},
	{0xC6E605B2A36DC1E0ULL, "6113906_m.fxe"},
	{0xC6E6060257D53108ULL, "vxlt_0_lowtown_docks_a3_aln_0_0.vlm"},
	{0xC6E6C7CB290CE8A2ULL, "vxlt_0_lowtown_a3_n_c_hpw_0_0.vlm"},
	{0xC6E6CC830E299DE0ULL, "6099033_m.fxe"},
	{0xC6E7372954A0BA4BULL, "205590338.ogg"},
	{0xC6E747CC37A33E18ULL, "vxca_caverns2prc_9f8_0_0.vlm"},
	{0xC6E7CEE31F63AE05ULL, "768330019.ogg"},
	{0xC6E7F1745D727F0EULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkforward.gad"},
	{0xC6E826A1B7FFF7A7ULL, "vxw_sundermount_2d0_0_0.vlm"},
	{0xC6E88F5FCBF4F676ULL, "906746280.ogg"},
	{0xC6E962C8E00D2EFBULL, "gen00fl_merrill6107054.cl"},
	{0xC6E9F64E7DBF3D31ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfour.mmh"},
	{0xC6E9F958B5D970CEULL, "fedorkiw.ncs"},
	{0xC6EA024E7DBF51D2ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfour.mao"},
	{0xC6EA104E7DBF69EFULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushfour.msh"},
	{0xC6EA35787209CCFAULL, "w_bonepitmining_a2_62_0.msh"},
	{0xC6EA43787209E433ULL, "w_bonepitmining_a2_62_0.mao"},
	{0xC6EA47787209EB40ULL, "w_bonepitmining_a2_62_0.mmh"},
	{0xC6EA7F4DB11CD45AULL, "hf_guard1.mor"},
	{0xC6EAC558FEFE3E5DULL, "vxlt_0_lowtown_a2_n_ejh_0_0.vlm"},
	{0xC6EAF19A27D014D0ULL, "112247807.ogg"},
	{0xC6EB4B9EE0210F86ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.anb"},
	{0xC6EB5F3ED20F96C7ULL, "gen00fl_fenris6110670.cl"},
	{0xC6EB6069BB000345ULL, "vxht_playershouse_n_1p3_0_0.vlm"},
	{0xC6EBB83D0CA2D008ULL, "6059309_m.fxe"},
	{0xC6EC3184051EA3CAULL, "sib131ar_exit_estate.utp"},
	{0xC6EC89CBAABD85E4ULL, "vxw_blightlands_j5_0_0.vlm"},
	{0xC6ECFBE236373A2FULL, "vxca_caverns1_mines_ihv_0_0.vlm"},
	{0xC6ED0B53621C3782ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_weightshift.gad"},
	{0xC6ED35D6A12CE40FULL, "vxht_keep_62v_0_0.vlm"},
	{0xC6ED4C1A30F44246ULL, "6140190_f.fxe"},
	{0xC6EDD9EEAAA02ED9ULL, "lt_0_lowtown_docks_a3_n_28_0_ll.mmh"},
	{0xC6EDF23A3FC927F4ULL, "mag211_marethari6025347.cl"},
	{0xC6EE71F109E922B3ULL, "6101519_m.fxe"},
	{0xC6EE74053955FB4DULL, "dae000_gamlen6048505.cl"},
	{0xC6EEBCF19B08ED93ULL, "hlac_lt_0_lowtown_docks_-3_6y.dds"},
	{0xC6F03C4396FEF3BCULL, "6097152_m.fxe"},
	{0xC6F061A0160A6F16ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sluce_n.dds"},
	{0xC6F0CEB072F83811ULL, "vxlt_0_lowtown_a2_gyt_0_0.vlm"},
	{0xC6F1155C7C11A259ULL, "w_blightlands_161_0.msh"},
	{0xC6F11F5C7C11B347ULL, "w_blightlands_161_0.mmh"},
	{0xC6F1235C7C11B994ULL, "w_blightlands_161_0.mao"},
	{0xC6F13F32E36A6132ULL, "vxlt_0_lowtown_a2_n_c_e2e_0_0.vlm"},
	{0xC6F14868DC245FF4ULL, "vxlt_0_lowtown_n_exo_0_0.vlm"},
	{0xC6F19EA8D51CD43CULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar1_.gad"},
	{0xC6F1EDEAED228116ULL, "6156176_m.fxe"},
	{0xC6F240BBE61A06D7ULL, "vxlt_0_lowtown_a2_n_emo_0_0.vlm"},
	{0xC6F2B5C37DF62287ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_action.evt"},
	{0xC6F3ACF0BFBC5ED5ULL, "vxht_mansion2_2ie_0_0.vlm"},
	{0xC6F41A25CAF75414ULL, "w_encounter3_mmap.dds"},
	{0xC6F41D2FBAF284B6ULL, "6110624_m.fxe"},
	{0xC6F42E5362229279ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_weightshift.evt"},
	{0xC6F43E496000355BULL, "lowtown_male_beggar_sta6156760.cl"},
	{0xC6F46E51036A4FA6ULL, "vxw_bonepit_1m2_0_0.vlm"},
	{0xC6F4CED63FAB984CULL, "128138739.ogg"},
	{0xC6F4E85B68E1B90EULL, "vxht_mansion2b_1rl_0_0.vlm"},
	{0xC6F51F3E69DDB411ULL, "hlac_lt_0_lowtown_a3_n_c_-131073_2z.dds"},
	{0xC6F5391BA0E87B6AULL, "w_sundermount_143_0.rml"},
	{0xC6F5683560EE724EULL, "vxlt_gallowstemplar_n_4o2_0_0.vlm"},
	{0xC6F5947EE3666419ULL, "vxdr_0_deeproads_ebp_0_0.vlm"},
	{0xC6F6987EA9F61027ULL, "vxw_sundermount_2i3_0_0.vlm"},
	{0xC6F6F524AF7B5EAFULL, "cd_mag130_danzig.bnk"},
	{0xC6F70E51644D0F00ULL, "follower_banter6163867.cl"},
	{0xC6F74CCEB3774E73ULL, "mag330_walter6171551.cl"},
	{0xC6F7E6CB6A1C06EEULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robe.mmh"},
	{0xC6F7F2CB6A1C1A85ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robe.mao"},
	{0xC6F800CB6A1C32DCULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robe.msh"},
	{0xC6F813C2A5FF28B5ULL, "sst_ss_qunari_m_3.gda"},
	{0xC6F845E1A5878A0FULL, "follower_banter6141884.cl"},
	{0xC6F8836A6D124FFEULL, "gen00fl_merrill6115995.cl"},
	{0xC6F89DA8D522F13AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\extraanims_templar1_.ani"},
	{0xC6F8BC773FD7DD3BULL, "6076367_m.fxe"},
	{0xC6F8CEF85784BF23ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxa.msh"},
	{0xC6F8DCF85784D77DULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxa.mmh"},
	{0xC6F8E0F85784DD8EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_genbenchxa.mao"},
	{0xC6F929EF7EAFAFA0ULL, "w_woundedcoast_152_0.phy"},
	{0xC6F94E9EE02D5231ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.mmh"},
	{0xC6F9D246ACF61BC5ULL, "vxht_playershouse_n_1w2_0_0.vlm"},
	{0xC6F9D6C37DFC79D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\handshake_large_action.gad"},
	{0xC6F9EE90C90447D5ULL, "hlac_lt_undercityencounter_-1_10z.dds"},
	{0xC6FA181E6FF8C858ULL, "vxdr_0_deeproads_ev1_0_0.vlm"},
	{0xC6FB15C84D83D582ULL, "art\\characters\\weapons\\longswords\\w_lsw_008a.phy"},
	{0xC6FB94727D379C2EULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant.mmh"},
	{0xC6FBA0727D37B0C5ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant.mao"},
	{0xC6FBAE727D37C81CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potplant.msh"},
	{0xC6FBBD2841DBE020ULL, "6131599_m.fxe"},
	{0xC6FCE115B8418D7EULL, "vxlt_hangedman_n_176_0_0.vlm"},
	{0xC6FD06BDD5F69FE6ULL, "6078801_m.fxe"},
	{0xC6FD52F31D49334AULL, "qcr220_orsino6061772.cl"},
	{0xC6FD6B92E756222AULL, "lgt310_mag_melson.cnv"},
	{0xC6FE07C4346753ECULL, "vxdr_primevalentrance_2e9_0_0.vlm"},
	{0xC6FE5C3FE52ECBA7ULL, "6051327_f.fxe"},
	{0xC6FED46971209264ULL, "6079350_m.fxe"},
	{0xC6FF0091B0D3703DULL, "vxlt_darktown_5b0_0_0.vlm"},
	{0xC6FF4F39C3C946EBULL, "6115374_m.fxe"},
	{0xC6FF6C40229E9B89ULL, "vxw_sundermount_2f0_0_0.vlm"},
	{0xC6FFD4088A1E783BULL, "lowtown_argument6169375.cl"},
	{0xC6FFF7DB8B58228EULL, "gen00fl_merrill6122364.cl"},
	{0xC70017FFB4D6B824ULL, "6204859_m.fxe"},
	{0xC7005BFB3A485367ULL, "712562679.ogg"},
	{0xC70065AB6557218EULL, "298893418.ogg"},
	{0xC700D951A7181151ULL, "one121ar_warehouse.lst"},
	{0xC700FC9FF8A8744FULL, "hlac_lt_0_lowtown_docks_65533_1z.dds"},
	{0xC70118915196AE72ULL, "15568362.ogg"},
	{0xC70144F89FD0B52EULL, "vxlt_gallowstemplar_2xr_0_0.vlm"},
	{0xC7014D92E759D852ULL, "lgt310_mag_melson.bnk"},
	{0xC701532F8FB856C0ULL, "gen00fl_varric6033716.cl"},
	{0xC70223FF2CAF644DULL, "vxlt_0_lowtown_a3_n_c_f59_0_0.vlm"},
	{0xC7023653622EDD68ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_weightshift.ani"},
	{0xC702D162AD85D39FULL, "524032247.ogg"},
	{0xC702F527BCB27100ULL, "gen_im_trs_drk_two_2hm_r2.uti"},
	{0xC7030701D8962D8FULL, "w_bonepit_4_0.mao"},
	{0xC7031301D896422CULL, "w_bonepit_4_0.mmh"},
	{0xC7031901D8964C1EULL, "w_bonepit_4_0.msh"},
	{0xC7033204A84B6F46ULL, "hlac_lt_0_lowtown_a3_n_c_65536_9y.dds"},
	{0xC7038DE440191EAEULL, "vxca_caverns1_mines_edc_0_0.vlm"},
	{0xC703ABA05AF2ED6DULL, "hlac_lt_0_lowtown_a2_131072_5y.dds"},
	{0xC703F4DB8D53B9F8ULL, "ht_0_hightown_n_16_0.anb"},
	{0xC7045E7431D29FBFULL, "hlac_lt_0_lowtown_docks_65535_5x.dds"},
	{0xC704BBD642E63358ULL, "vxlt_warehouse_18r_0_0.vlm"},
	{0xC70565121B3B16E4ULL, "hlac_ht_mansion2b_818_66x.dds"},
	{0xC7058A84BB572E9EULL, "y_crate_02.dds"},
	{0xC705B1CBC5D95519ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_left_3.mao"},
	{0xC705E3D9270026B7ULL, "vxlt_0_lowtown_a2_e1t_0_0.vlm"},
	{0xC705E5ED518EF3F2ULL, "6112129_m.fxe"},
	{0xC70625952734B26FULL, "llac_lt_darktown_-1_41.dds"},
	{0xC70667D67D15D6AEULL, "vxdr_primevalentrance_231_0_0.vlm"},
	{0xC7067966427C2C83ULL, "vxw_woundedcoast_10k_0_0.vlm"},
	{0xC706EC9E2B82E18EULL, "90359776.ogg"},
	{0xC7070AD2DF15B3F0ULL, "llac_lt_backalley_g_0_13.dds"},
	{0xC707345C5F0C4347ULL, "vxlt_backalley_n_1ju_0_0.vlm"},
	{0xC70741154AC500BDULL, "6162129_m.fxe"},
	{0xC707497E5F8BB3A0ULL, "155591950.ogg"},
	{0xC70765698B9690DAULL, "gen00fl_fenris6113845.cl"},
	{0xC707A93C6B188933ULL, "592828218.ogg"},
	{0xC707BA5279F0F725ULL, "vxw_bonepitmining_a2_214_0_0.vlm"},
	{0xC707F151A71E58E8ULL, "one121ar_warehouse.ncs"},
	{0xC707F4A367BD4AC7ULL, "179219932.ogg"},
	{0xC708B705275F129BULL, "mag130_danzig6167374.cl"},
	{0xC708D8127F519F87ULL, "469063465.ogg"},
	{0xC708EF49C2B1B322ULL, "zz_vrc_debug6200158.cl"},
	{0xC7090ED96A2C6625ULL, "hlac_lt_0_lowtown_a2_n_-131072_8z.dds"},
	{0xC7094BC750EF69BEULL, "6105705_m.fxe"},
	{0xC7099B04FEF2A2FFULL, "vxht_keep_635_0_0.vlm"},
	{0xC709C3E306DD01A2ULL, "cs_pro000cs_flemeth_fake.wbd"},
	{0xC70A225F5A15731FULL, "vxca_caverns1_egz_0_0.vlm"},
	{0xC70B11D0A963CF04ULL, "hlac_lt_0_lowtown_n_65538_1z.dds"},
	{0xC70B3A5233AA9F66ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.anb"},
	{0xC70B475BAB65499DULL, "671833467.ogg"},
	{0xC70BAA03D7FE500DULL, "vxdr_0_deeproads_9g7_0_0.vlm"},
	{0xC70BB9F92BAE6F3CULL, "gen_im_trs_dal_arc_lbw_r0.uti"},
	{0xC70C03847B177706ULL, "hlac_ht_chantry_n_2534_31x.dds"},
	{0xC70C3C1118C58D82ULL, "vxca_caverns1_mines_j2n_0_0.vlm"},
	{0xC70CB949039C0621ULL, "6100356_m.fxe"},
	{0xC70CBE06BA1BF653ULL, "gen00fl_carver6108888.cl"},
	{0xC70CF24E0633A3FEULL, "lt_0_lowtown_11_0.rml"},
	{0xC70D027AA61F6770ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_clothrug.msh"},
	{0xC70D107AA61F7E8AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_clothrug.mmh"},
	{0xC70D147AA61F85C1ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_clothrug.mao"},
	{0xC70D2DFDFDDDACC5ULL, "l_dae310ar_lowtown.wbd"},
	{0xC70D897D95F3189CULL, "zz_approval_debug6003733.cl"},
	{0xC70DDE1D1494B409ULL, "vxw_sundermount_2k7_0_0.vlm"},
	{0xC70E21B97F1B1F1BULL, "vxht_0_hightown_a2_5i8_0_0.vlm"},
	{0xC70E27AD9C98421DULL, "vxlt_gallowscourtyard_7cs_0_0.vlm"},
	{0xC70E2F7449063E94ULL, "hlac_w_encounter3_-131069_0z.dds"},
	{0xC70EA86C36383816ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_andraste_n.dds"},
	{0xC70EB240F3F90FB1ULL, "framebuffereffects_composedof_p0.vsh.fxo"},
	{0xC70F30DDCA06E00AULL, "hlac_lt_backalley_n_65536_4x.dds"},
	{0xC70FD5236687F393ULL, "vxht_0_hightown_a3_b8q_0_0.vlm"},
	{0xC7108A0B9D5B13A9ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_come_0n.dds"},
	{0xC710F8CB72E9E257ULL, "639452969.ogg"},
	{0xC71148CE1A822054ULL, "hlac_lt_darktown_-1_14y.dds"},
	{0xC71193CD36D5D9F1ULL, "hlac_ht_chantry_n_2534_0y.dds"},
	{0xC711A9DC113C91A1ULL, "548843456.ogg"},
	{0xC711B39357694939ULL, "6069605_f.fxe"},
	{0xC71285032C785E05ULL, "lla_ht_0_hightown_n_65535_0.dds"},
	{0xC712E7D87373AC27ULL, "6110733_m.fxe"},
	{0xC7135C1A4953960BULL, "vxlt_gallowscourtyard_n_66t_0_0.vlm"},
	{0xC71412F59BC185DEULL, "follower_banter6171570.cl"},
	{0xC71423EA2BCB5A4AULL, "vxlt_darktown_5jk_0_0.vlm"},
	{0xC7146D48F1E61031ULL, "mag121_viveka6082421.cl"},
	{0xC714D4AFBE317DE9ULL, "6148396_m.fxe"},
	{0xC714E45269A412C3ULL, "vxlt_0_lowtown_docks_a3_bfl_0_0.vlm"},
	{0xC715B2F23CBB193DULL, "6167229_m.fxe"},
	{0xC715B8D1ECBCE140ULL, "zz_ave_debug6175051.cl"},
	{0xC715E74656D920BAULL, "217730442.ogg"},
	{0xC715F4AEA4BCA2B7ULL, "vxw_encounter3_es_0_0.vlm"},
	{0xC716CE80F4776EFFULL, "968165910.ogg"},
	{0xC7170F94DDB345F9ULL, "515593625.ogg"},
	{0xC71719AF1BB9D4CEULL, "643532683.ogg"},
	{0xC7177D6C1387C68CULL, "300671508.ogg"},
	{0xC717DF62809FCF64ULL, "vxw_blightlands_u8_0_0.vlm"},
	{0xC7183D5233B52E11ULL, "art\\vfx\\environment\\vfx_hnd_r_mug_c.mmh"},
	{0xC71855CD4F5C76E7ULL, "one300_zevran6075843.cl"},
	{0xC7185C9727068CB3ULL, "45721480.ogg"},
	{0xC718943C4C0B8B9FULL, "vxca_caverns1_ebe_0_0.vlm"},
	{0xC718E0B85B11A80DULL, "vxlt_0_lowtown_n_ff7_0_0.vlm"},
	{0xC718FAC5680B9A08ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexb.mmh"},
	{0xC71906C5680BAEABULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexb.mao"},
	{0xC71918C5680BCD32ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_sandpilexb.msh"},
	{0xC719408929F9CAF8ULL, "vxlt_0_lowtown_a2_hc4_0_0.vlm"},
	{0xC71A55926D035689ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb106a_0s.dds"},
	{0xC71A7BF97963B85DULL, "m_lgt231ar_tome_final.bnk"},
	{0xC71A9A8C0EDE43CDULL, "vxlt_hangedman_n_1w1_0_0.vlm"},
	{0xC71AA809CA08DDF9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\blendtree\\gateguardian_movement.anb"},
	{0xC71AD69790D3D6CEULL, "gen00fl_sebastian6180150.cl"},
	{0xC71AD94E483576F8ULL, "hlac_lt_0_lowtown_n_-196609_3y.dds"},
	{0xC71B0907BCE4F487ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chab_d.dds"},
	{0xC71B753AC0EBC55AULL, "hlac_dr_primevalentrance_1724_2x.dds"},
	{0xC71BF8C52C5FCA14ULL, "war_devour.dds"},
	{0xC71C4FFA068ACC89ULL, "vxw_sundermount_2l6_0_0.vlm"},
	{0xC71C998A8E8E5A9DULL, "vxlt_gallowscourtyard_7dr_0_0.vlm"},
	{0xC71D65A224D3249BULL, "515790655.ogg"},
	{0xC71D9ED7B8BA23FFULL, "dae021_fem_worshipper_sta6150361.cl"},
	{0xC71DDD3F89617AFEULL, "qcr220_orsino6061804.cl"},
	{0xC71DE96460D6528EULL, "art\\vfx\\materials\\vfx_smoke_puffy_vista.mao"},
	{0xC71E246F9FE1AC11ULL, "6020545_f.fxe"},
	{0xC71E46EF7ECEFD35ULL, "w_woundedcoast_152_0.msh"},
	{0xC71E4700587E0C13ULL, "vxht_0_hightown_a3_b7p_0_0.vlm"},
	{0xC71E48EF7ECF00ABULL, "w_woundedcoast_152_0.mmh"},
	{0xC71E54EF7ECF1548ULL, "w_woundedcoast_152_0.mao"},
	{0xC71E87FEB944AE38ULL, "vxht_mansion_n_6h0_0_0.vlm"},
	{0xC71E88BFD1E4E48AULL, "gen00fl_varric6105428.cl"},
	{0xC71E91296F27629AULL, "docks_male_human_worker_sta6175968.cl"},
	{0xC71EF3B804957B26ULL, "vxca_cavern1_haunted_jfv_0_0.vlm"},
	{0xC71F31D098212F12ULL, "249515501.ogg"},
	{0xC71F82EA36E6E3CCULL, "vxht_bartrand_n_5hd_0_0.vlm"},
	{0xC71F90A581317E4EULL, "6024573_f.fxe"},
	{0xC72071DCF59026B6ULL, "qun120_arishok6051676.cl"},
	{0xC7207509DFA8B88FULL, "6109003_m.fxe"},
	{0xC720F0C84DA46636ULL, "art\\characters\\weapons\\longswords\\w_lsw_008a.mao"},
	{0xC720F55EB2E2301AULL, "482574130.ogg"},
	{0xC720FCC84DA47AD5ULL, "art\\characters\\weapons\\longswords\\w_lsw_008a.mmh"},
	{0xC720FEC84DA47E4BULL, "art\\characters\\weapons\\longswords\\w_lsw_008a.msh"},
	{0xC7214D65D1E98E88ULL, "llac_ht_mansion2b_818_0.dds"},
	{0xC7216598547A120FULL, "6034699_m.fxe"},
	{0xC7219F9026D26458ULL, "vxlt_warehouse_12p_0_0.vlm"},
	{0xC721C71852DBD37CULL, "494548105.ogg"},
	{0xC72218F0DE6FE573ULL, "626674408.ogg"},
	{0xC72295C71DC172CAULL, "vxlt_darktown_5od_0_0.vlm"},
	{0xC72295CA9A89D4DBULL, "prm000im_fade_sword.uti"},
	{0xC7235FBE3FFCD287ULL, "211799031.ogg"},
	{0xC724188785FA16CFULL, "vxlt_0_lowtown_a2_n_h23_0_0.vlm"},
	{0xC72449B87B699AEAULL, "fen210_dying_magister6108600.cl"},
	{0xC7255CAA1C987D3CULL, "vrc321_party_barks6096263.cl"},
	{0xC7255F87CD4F64B3ULL, "w_bonepitmining_a2_107_0.phy"},
	{0xC7258367C825A729ULL, "gen00fl_aveline6140131.cl"},
	{0xC725AC50D755BEE0ULL, "hlac_w_woundedcoast_196614_0z.dds"},
	{0xC725CE5539B0E14EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandeddeselectedexit.evt"},
	{0xC725D1F6A494C583ULL, "6043280_m.fxe"},
	{0xC725D42E6A3482B5ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_7z.dds"},
	{0xC725EB8E6E3C0976ULL, "9586842.ogg"},
	{0xC725F98287FC7B1FULL, "hlac_ht_0_hightown_1_0y.dds"},
	{0xC72619EA8C1B6A13ULL, "qcr221ar_keep.ncs"},
	{0xC726207E2EC96562ULL, "dae011cr_bar_wench.utc"},
	{0xC72633F962C09750ULL, "hlac_ht_0_hightown_a2_65536_2x.dds"},
	{0xC72641785D5CF437ULL, "vxht_keep_6ag_0_0.vlm"},
	{0xC726E05EA8A59E80ULL, "hlac_lt_gallowstemplar_2394_8z.dds"},
	{0xC726FF129BE76565ULL, "295429109.ogg"},
	{0xC7270440DB9BBE5FULL, "hlac_lt_0_lowtown_docks_-2_2z.dds"},
	{0xC7271B55557A14CEULL, "19115502.ogg"},
	{0xC7273B5DED4D0CD8ULL, "lgt321cr_teagan.utc"},
	{0xC727745CB573E77BULL, "vxht_brothel_n_60d_0_0.vlm"},
	{0xC727ACE42CCD9971ULL, "zz_mer_debug6181057.cl"},
	{0xC72844DDEBF655BBULL, "6160617_m.fxe"},
	{0xC728893C50D32A94ULL, "mag320_orsino6070444.cl"},
	{0xC728BF6838DB763FULL, "vxdr_primevalentrance_2e6_0_0.vlm"},
	{0xC72A012D0E463482ULL, "hlac_lt_0_lowtown_1_5y.dds"},
	{0xC72A3EBC750CD8E1ULL, "188656351.ogg"},
	{0xC72A5E6B4176E622ULL, "6013301_m.fxe"},
	{0xC72AC1D6F880E509ULL, "vxw_sundermount_2q5_0_0.vlm"},
	{0xC72AEC8D1B33F592ULL, "440365891.ogg"},
	{0xC72B0B678084731DULL, "vxlt_gallowscourtyard_7iq_0_0.vlm"},
	{0xC72B64BFF57ACCE5ULL, "vxlt_gallowscourtyard_4vc_0_0.vlm"},
	{0xC72B955ABDA5A21DULL, "zz_fol_debug6062218.cl"},
	{0xC72B9D129BFEEA3FULL, "vxht_0_hightown_n_at0_0_0.vlm"},
	{0xC72C538253A0D2CFULL, "llac_w_woundedcoast_327686_0.dds"},
	{0xC72CB2E4697787C2ULL, "6169844_m.fxe"},
	{0xC72CB53CF103E6F3ULL, "193902313.ogg"},
	{0xC72CC55539B6F0F9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandeddeselectedexit.gad"},
	{0xC72D39EA8C21BF1AULL, "qcr221ar_keep.lst"},
	{0xC72E6CBEBFE93710ULL, "6080178_m.fxe"},
	{0xC72E8F76C55C45B5ULL, "zz_mag_debug6089444.cl"},
	{0xC72F2771999ABA0BULL, "6201571_f.fxe"},
	{0xC72F4256ACD4FB2DULL, "6136739_m.fxe"},
	{0xC72FB98824E7022DULL, "6111237_m.fxe"},
	{0xC73074785A28A8C7ULL, "lt_0_lowtown_docks_n_16_0.rml"},
	{0xC730D513F083DD6FULL, "vxht_0_hightown_a2_7g0_0_0.vlm"},
	{0xC73168FAD7AFC590ULL, "gen_im_trs_qun_hvy_cht_r2.uti"},
	{0xC732165C23A71DC1ULL, "face0_bssrdf_1.psh.fxo"},
	{0xC732D65539BB7947ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\warriortwohandeddeselectedexit.ani"},
	{0xC732DF24C626900EULL, "6139508_m.fxe"},
	{0xC733097239DD703FULL, "6113862_m.fxe"},
	{0xC7334DA0C0D7B991ULL, "hlac_lt_house_-1_0y.dds"},
	{0xC7338C99CEAA8AB9ULL, "hlac_lt_0_lowtown_docks_-3_7x.dds"},
	{0xC733D042A4994C3FULL, "6099171_m.fxe"},
	{0xC733E06BF19B2B61ULL, "t1_mue_te1.tnt"},
	{0xC734105963BF6ACAULL, "732638582.ogg"},
	{0xC73537D81A1BFAF0ULL, "vxca_caverns1_mines_cjy_0_0.vlm"},
	{0xC7356F64EF0FCB67ULL, "552040249.ogg"},
	{0xC73577F62FF7BC9FULL, "vxca_caverns1_ehg_0_0.vlm"},
	{0xC735A8C9B8D37EFDULL, "vxht_0_hightown_a2_c43_0_0.vlm"},
	{0xC735EDD8AFF74CA1ULL, "166881322.ogg"},
	{0xC736085240CC3571ULL, "422846292.ogg"},
	{0xC736911FE42969F7ULL, "259075069.ogg"},
	{0xC736B6BADDEC6E3BULL, "6165917_m.fxe"},
	{0xC7372D0771BA4AD8ULL, "6171794_f.fxe"},
	{0xC7379B70AC97E462ULL, "art\\characters\\creatures\\pridedemon\\animation\\die_lp.gad"},
	{0xC738036F16CE8681ULL, "6000809_m.fxe"},
	{0xC7380BA232E84F90ULL, "hf_barmaid.mor"},
	{0xC738B15B40DDD1F7ULL, "6061402_m.fxe"},
	{0xC73933B3EA76FD89ULL, "vxw_sundermount_2z4_0_0.vlm"},
	{0xC7396E96E0D63313ULL, "fo_rock.mao"},
	{0xC739B16788905A49ULL, "hlac_w_woundedcoast_262152_0z.dds"},
	{0xC739CA34209FEE31ULL, "mag370_meredith6092190.cl"},
	{0xC739F4C28C354BB7ULL, "vxht_0_hightown_n_byx_0_0.vlm"},
	{0xC73A5D0B8FB32B0FULL, "gen00fl_isabela6097355.cl"},
	{0xC73A924365C9260BULL, "6110792_f.fxe"},
	{0xC73A93CD7D0C5611ULL, "blood_conj_05.mao"},
	{0xC73A980E438BBC54ULL, "dae000_gamlen6162138.cl"},
	{0xC73ACDDA28149BEEULL, "vxht_chantry_27l_0_0.vlm"},
	{0xC73AFA8DC2E2F7D6ULL, "static_static_p34.psh.fxo.dx11.assembly"},
	{0xC73B28E1B5B2A152ULL, "lt_0_lowtown_docks_a3_n_9_0_ll.mmh"},
	{0xC73BFFEFBDD72AA0ULL, "499687171.ogg"},
	{0xC73C7375C61998C1ULL, "one300_zevran6070910.cl"},
	{0xC73C92457B37967CULL, "538571865.ogg"},
	{0xC73CD1310B5F46D7ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_trapdoorlid.phy"},
	{0xC73D129729651051ULL, "vxlt_warehouse_8z_0_0.vlm"},
	{0xC73D4AAC4F213FABULL, "vxlt_undercityencounter_uv_0_0.vlm"},
	{0xC73D5F6429A67192ULL, "vxlt_gallowscourtyard_7gd_0_0.vlm"},
	{0xC73DBBB025C87B46ULL, "vxca_caverns2prc_8q6_0_0.vlm"},
	{0xC73EC2A043A6A82EULL, "vxlt_hangedman_2bm_0_0.vlm"},
	{0xC73ECF36D706EA33ULL, "hlac_lt_undercitydungeon_-1_19x.dds"},
	{0xC7404C2D385C7413ULL, "mrl000_marethari6129215.cl"},
	{0xC74097C8738A8912ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_rug_n.dds"},
	{0xC740FBB6CA99C6E2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutd.msh"},
	{0xC740FDB6CA99C998ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutd.mmh"},
	{0xC74104949A187707ULL, "6118838_m.fxe"},
	{0xC74109B6CA99DE3BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutd.mao"},
	{0xC74156D7E367A357ULL, "vxht_mansion2b_3gp_0_0.vlm"},
	{0xC7419033DBC429EBULL, "hlac_lt_darktown_-1_76z.dds"},
	{0xC741A7AE508B2122ULL, "w_bonepitmining_a2_124_0.phy"},
	{0xC741BF7BB500D214ULL, "6001332_f.fxe"},
	{0xC741D09BEA2C44D8ULL, "6202215_m.fxe"},
	{0xC7421C7C90668D22ULL, "11342878.ogg"},
	{0xC7429ED6E3FEA3E5ULL, "gen_im_arm_boo_hvy_start.uti"},
	{0xC742C48C047D4FABULL, "art\\vfx\\creatures\\fxc_ogre_stmp_c.impa.ani"},
	{0xC743A2306506A22FULL, "w_bonepitmining_a2_88_0.mao"},
	{0xC743AB659E7AD34BULL, "6094796_m.fxe"},
	{0xC743AE306506B6CCULL, "w_bonepitmining_a2_88_0.mmh"},
	{0xC743B4306506C0BEULL, "w_bonepitmining_a2_88_0.msh"},
	{0xC743B8C0DC324E6BULL, "6163788_m.fxe"},
	{0xC7441AA6AAC9977DULL, "vxht_0_hightown_a2_c32_0_0.vlm"},
	{0xC7445AA6D7F96E6DULL, "gen00fl_dog6172082.cl"},
	{0xC744B8F67BC0669EULL, "hlac_lt_0_lowtown_n_-196609_4z.dds"},
	{0xC7456C59B2153359ULL, "6092217_m.fxe"},
	{0xC7457F286D45CE96ULL, "llac_ca_caverns1_mines_24424_22.dds"},
	{0xC745BF0DDA5ECC57ULL, "mag210_arianni6024543.cl"},
	{0xC745F10349E9C0E3ULL, "gen00fl_merrill6108486.cl"},
	{0xC745F13D80389F21ULL, "vxca_caverns2prc_6ng_0_1.vlm"},
	{0xC746182D57B81086ULL, "mag121_eau_banter6121093.cl"},
	{0xC746EF4E066516A2ULL, "lt_0_lowtown_11_0.anb"},
	{0xC74724B65A298F5AULL, "gen00fl_bethany6102762.cl"},
	{0xC74793DC81BCABF8ULL, "vxht_0_hightown_a2_b21_0_0.vlm"},
	{0xC7479C52933B4E8FULL, "lgt260tr_convo_trigger.ncs"},
	{0xC747C76CBC5EFE38ULL, "6134602_m.fxe"},
	{0xC747D1691EC6575AULL, "890118348.ogg"},
	{0xC747F5E24B9704FBULL, "llac_lt_0_lowtown_a2_n_-1_7.dds"},
	{0xC74817C3DF1A351BULL, "zz_firestaff2_v4.cut"},
	{0xC748A08FB85FD545ULL, "vxlt_undercityserialkiller_19k_0_0.vlm"},
	{0xC749275BF9E1FB04ULL, "6190998_m.fxe"},
	{0xC7493FB71A0AB46EULL, "vxht_chantry_20m_0_0.vlm"},
	{0xC749E086F5C97D1BULL, "375982706.ogg"},
	{0xC74A2186BE03D4AEULL, "dae150_harshal.cnv"},
	{0xC74AC59472953A5BULL, "gen00fl_merrill6123728.cl"},
	{0xC74AC6437481B3A5ULL, "gen_im_trs_chn_amu_r2.uti"},
	{0xC74AEDEE5485A821ULL, "mag100_cullen6096064.cl"},
	{0xC74B6CBBEECDFCBBULL, "qun100_mage_party_comment6019103.cl"},
	{0xC74BC670ACA8D648ULL, "art\\characters\\creatures\\pridedemon\\animation\\die_lp.ani"},
	{0xC74CBD2E110BF4DBULL, "vxlt_0_lowtown_a2_n_c_hkj_0_0.vlm"},
	{0xC74CE5354D795C1FULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubew.msh"},
	{0xC74CEB354D7966C1ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubew.mmh"},
	{0xC74CF7354D797B62ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubew.mao"},
	{0xC74CF758AF89BEECULL, "hlac_lt_gallowstemplar_n_2384_0x.dds"},
	{0xC74D4386BE06441EULL, "dae150_harshal.bnk"},
	{0xC74D7B2B7C122162ULL, "32499001.ogg"},
	{0xC74DF9887D4CD8DAULL, "vxht_keep_73m_0_0.vlm"},
	{0xC74E03CA90E185AFULL, "l_dae220ar_hightown.wbd"},
	{0xC74E07FDFE15036FULL, "l_dae310ar_lowtown.bnk"},
	{0xC74E43AB3E231ED5ULL, "558858796.ogg"},
	{0xC74F3C01CB5D2337ULL, "dae000_hate_follower_react6103210.cl"},
	{0xC74FCD885AA0EA6EULL, "vxdr_0_deeproads_f68_0_0.vlm"},
	{0xC74FD647634CEF4BULL, "hlac_ht_mansion_n_3365_0z.dds"},
	{0xC750A12F86AE0DBAULL, "dae000_bartrand6029006.cl"},
	{0xC750AC8C04896422ULL, "art\\vfx\\creatures\\fxc_ogre_stmp_c.impa.evt"},
	{0xC750AF20F55BDB20ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_8z.dds"},
	{0xC750D1E307191C58ULL, "cs_pro000cs_flemeth_fake.bnk"},
	{0xC750FD4B119EF4CFULL, "dae000ip_crowd_ff3_talk12.utp"},
	{0xC7519443FCDF73EBULL, "combat_range.gda"},
	{0xC751B38455C99CC6ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.dura.ani"},
	{0xC751C83134B78F6DULL, "em_gen_dalish_warrior.mor"},
	{0xC7520D7B71E2089DULL, "vxw_bonepitdestroyed_a3_1pv_0_0.vlm"},
	{0xC75238EA8C40D99DULL, "qcr221ar_keep.are"},
	{0xC7530F5D010A67CBULL, "6021248_m.fxe"},
	{0xC7534A04E3C37F8CULL, "747599135.ogg"},
	{0xC7543C11C0D78370ULL, "mag291_vincento6027722.cl"},
	{0xC754627A27EB675AULL, "592514952.ogg"},
	{0xC754D71260F16D7BULL, "mag210_arianni6024987.cl"},
	{0xC754DD7EE364A674ULL, "character_character_p7.psh.fxo.dx11.assembly"},
	{0xC754F43A78613FB5ULL, "630644714.ogg"},
	{0xC7551C33FA252B1AULL, "zz_rogue_femaile.cut"},
	{0xC7557440BCEF7BAEULL, "vxht_0_hightown_a2_bhy_0_0.vlm"},
	{0xC755AFA339A084B9ULL, "vxht_bartrand_n_5sl_0_0.vlm"},
	{0xC755B7B8210B8388ULL, "6146913_m.fxe"},
	{0xC7563C504FCA9CD8ULL, "hlac_w_encounter1_65538_0x.dds"},
	{0xC756425EBD917F81ULL, "hlac_ht_mansion2_818_12x.dds"},
	{0xC756D4B0FE2C9500ULL, "hlac_lt_gallowstemplar_f_2391_2x.dds"},
	{0xC756D87C70217CB7ULL, "vxht_0_hightown_n_boz_0_0.vlm"},
	{0xC756FAE30F8A1E77ULL, "qcr220_orsino6041830.cl"},
	{0xC75735DF4AB84AF9ULL, "llac_ca_caverns1_24423_3.dds"},
	{0xC7578ABCF5EDEB92ULL, "vxlt_0_lowtown_docks_bac_0_0.vlm"},
	{0xC757AB8C048F8135ULL, "art\\vfx\\creatures\\fxc_ogre_stmp_c.impa.gad"},
	{0xC757BEEEB501341FULL, "6218628_m.fxe"},
	{0xC757FE3CE1F7B4DBULL, "follower_banter6156224.cl"},
	{0xC7582EB72B3BF5F8ULL, "hlac_lt_0_lowtown_docks_a3_n_131072_1y.dds"},
	{0xC7584CC2C73C6EF7ULL, "801781480.ogg"},
	{0xC758A792707CAFCEULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.mmh"},
	{0xC758C1E8C8F58AEEULL, "850149742.ogg"},
	{0xC758FD5052226CBAULL, "w_bonepit_82_0.msh"},
	{0xC7590B50522284F3ULL, "w_bonepit_82_0.mao"},
	{0xC7590F5052228B00ULL, "w_bonepit_82_0.mmh"},
	{0xC7592DD2E14CF76FULL, "401772756.ogg"},
	{0xC759484AA9F7C34FULL, "vxht_mansion2_21o_0_0.vlm"},
	{0xC759542A6EE188C7ULL, "dae011_conspiracy_mob6149592.cl"},
	{0xC7599278E6AD65E2ULL, "1017763377.ogg"},
	{0xC759B757E4C4E5D8ULL, "476368798.ogg"},
	{0xC759B9BB6DD2E1B7ULL, "vxlt_0_lowtown_el9_0_2.vlm"},
	{0xC759D610E291E027ULL, "6045963_m.fxe"},
	{0xC75A9D3747DDD878ULL, "mag101_alain6053012.cl"},
	{0xC75ABF677EF073FCULL, "6140861_m.fxe"},
	{0xC75B9C6950091A8AULL, "802722456.ogg"},
	{0xC75C5EB7A7ADBD9FULL, "vxw_bonepit_16f_0_0.vlm"},
	{0xC75D0E5660EE0C66ULL, "6070157_m.fxe"},
	{0xC75D4428CF8E01CFULL, "ht_mansion_n_3367_ll.mmh"},
	{0xC75DAE37FF3738A6ULL, "6001034_m.fxe"},
	{0xC75E5FBB9CF4FDCDULL, "293471652.ogg"},
	{0xC75E73A9C0D713F6ULL, "6154000_m.fxe"},
	{0xC75ED47DA5FD7B6FULL, "follower_barks6153853.cl"},
	{0xC75EF0022F48A8EFULL, "6083805_f.fxe"},
	{0xC75F51503E5DD9FAULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv2.gad"},
	{0xC75F5321A734C944ULL, "116360700.ogg"},
	{0xC75F8F1A94E64520ULL, "6024969_m.fxe"},
	{0xC75F9B8455D5B137ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.dura.evt"},
	{0xC75FB0FFB15F1F5EULL, "6196727_m.fxe"},
	{0xC760034870A2991BULL, "art\\characters\\playercharacter\\humanmale\\arm_medc.phy"},
	{0xC760E71CC9D56C84ULL, "dae000_oldmangossip6174898.cl"},
	{0xC760F9AF11CA9DD7ULL, "817941502.ogg"},
	{0xC761A1156BE1B2CEULL, "835353357.ogg"},
	{0xC762060F77078EFEULL, "prop1_s2y.psh.fxo"},
	{0xC762328820974B34ULL, "vxw_encounter1_da_0_0.vlm"},
	{0xC762A5FC5F772BACULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masa_0s.dds"},
	{0xC762C014DB7DAF3CULL, "34182521.ogg"},
	{0xC762C0605AA8C2F5ULL, "hlac_lt_0_lowtown_a2_n_-4_0x.dds"},
	{0xC7635754750BAC53ULL, "6136328_m.fxe"},
	{0xC76362D2CC9B5771ULL, "vxht_mansion2_1t1_0_0.vlm"},
	{0xC76384189F000EC4ULL, "character_character_p41.psh.fxo.dx11.assembly"},
	{0xC764BBE00F9F5126ULL, "266312239.ogg"},
	{0xC764BFB690A677AAULL, "hlac_lt_0_lowtown_docks_131071_6y.dds"},
	{0xC76516B899172802ULL, "lgt311_shw_mer_mettin6156561.cl"},
	{0xC76527ED996F1FE2ULL, "vxw_bonepitmining_a2_1wk_0_1.vlm"},
	{0xC76554503E624A11ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv2.evt"},
	{0xC7658717AD08CE97ULL, "6109111_m.fxe"},
	{0xC765C69270876ECDULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s3_c.anb"},
	{0xC765DF1FE8291FD0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8mb.msh"},
	{0xC765ED1FE82937EAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8mb.mmh"},
	{0xC765F11FE8293E21ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp8mb.mao"},
	{0xC765FC99E7E40412ULL, "vxlt_0_lowtown_docks_bdl_0_0.vlm"},
	{0xC7669C8455DBD1C0ULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_hos_c.dura.gad"},
	{0xC766B46E2D7BF3ACULL, "vxlt_0_lowtown_a2_n_c_emz_0_0.vlm"},
	{0xC766B599FB004D2BULL, "art\\vfx\\textures\\zz_gradient_alpha.dds"},
	{0xC76782AE50ABB1D6ULL, "w_bonepitmining_a2_124_0.mao"},
	{0xC7678EAE50ABC675ULL, "w_bonepitmining_a2_124_0.mmh"},
	{0xC76790AE50ABC9EBULL, "w_bonepitmining_a2_124_0.msh"},
	{0xC7679A13A4CB2E92ULL, "vxht_0_hightown_n_cb9_0_0.vlm"},
	{0xC767FA3E3B2CE028ULL, "vxlt_gallowstemplar_4tv_0_0.vlm"},
	{0xC7683E42CD09E6BBULL, "lgt260cr_way_merc_1.utc"},
	{0xC768915E1CEE19E3ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf.evt"},
	{0xC76902888C003062ULL, "6068847_f.fxe"},
	{0xC76965C89BBE8847ULL, "warrior_5.cut"},
	{0xC7697D7676137715ULL, "gen_im_wep_rog_dua_default.uti"},
	{0xC769D0198EC65A0EULL, "hlac_dr_0_deeproads_5845_2x.dds"},
	{0xC76A0CAF38396182ULL, "vxlt_0_lowtown_a2_n_ef7_0_0.vlm"},
	{0xC76A9D8DC5F26FE1ULL, "vxlt_undercityserialkiller_184_0_0.vlm"},
	{0xC76AA69D4594AD84ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic_1.mao"},
	{0xC76AE0829EF04C04ULL, "dae000_mother6135360.cl"},
	{0xC76B1B448EC667B7ULL, "hlac_lt_backalley_g_0_10x.dds"},
	{0xC76BE57F430C80DEULL, "vxlt_0_lowtown_df9_0_0.vlm"},
	{0xC76BE58A377F2EFAULL, "dae000cr_grand_cleric6187940.cl"},
	{0xC76C32FDF5F247E2ULL, "vxdr_0_deeproads_f5y_0_0.vlm"},
	{0xC76C4674DD75B8ACULL, "6129086_m.fxe"},
	{0xC76C758E30AF0CCCULL, "follower_banter6154195.cl"},
	{0xC76C79E33913B1FEULL, "155581724.ogg"},
	{0xC76CA004ABB23BC4ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v3.ani"},
	{0xC76D069626187949ULL, "one110_tree_minder6133733.cl"},
	{0xC76D2EC330AFFE0AULL, "6013841_f.fxe"},
	{0xC76D8DD88D4754BFULL, "730069682.ogg"},
	{0xC76DEF9D87126392ULL, "mag370_meredith6185097.cl"},
	{0xC76E80BE0812E1BFULL, "w_bonepitdestroyed_a3_50_0.msh"},
	{0xC76E86BE0812EC61ULL, "w_bonepitdestroyed_a3_50_0.mmh"},
	{0xC76E92BE08130002ULL, "w_bonepitdestroyed_a3_50_0.mao"},
	{0xC76F42DCE7D40271ULL, "hlac_ca_cavern1_haunted_24424_27x.dds"},
	{0xC76F4F76D96EF7CEULL, "vxlt_gallowscourtyard_n_7gk_0_0.vlm"},
	{0xC76F6A5E1CF3F664ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf.gad"},
	{0xC76F6E58FC29735BULL, "707229974.ogg"},
	{0xC76F7F8671FEB3D9ULL, "ser120_ghyslain6118431.cl"},
	{0xC76FCA5D5BDB64E2ULL, "hlac_lt_gallowstemplar_7198_2y.dds"},
	{0xC76FF2B18B2E12E1ULL, "446048872.ogg"},
	{0xC77015E2483A4733ULL, "40895810.ogg"},
	{0xC7705A7EC0AC1A9DULL, "vxca_cavern1_haunted_i2z_0_0.vlm"},
	{0xC770A465128D63B4ULL, "vxw_encounter1_if_0_0.vlm"},
	{0xC770AECC666919AAULL, "6165443_f.fxe"},
	{0xC7716C0DB355D9C5ULL, "vxw_bonepitmining_a2_16f_0_0.vlm"},
	{0xC77177785A6044DBULL, "lt_0_lowtown_docks_n_16_0.anb"},
	{0xC771905419ABB437ULL, "llac_ca_caverns1_24427_1.dds"},
	{0xC771980BB783240BULL, "253470763.ogg"},
	{0xC771E600F4BF1FE0ULL, "dre000_dougal6131139.cl"},
	{0xC77256B6435CF7B6ULL, "store_and111cr_lirene.utm"},
	{0xC772E9F25892C1C3ULL, "gen00fl_anders6125993.cl"},
	{0xC7736C503E6EB0E0ULL, "art\\characters\\creatures\\ragedemon\\animation\\combat\\sneakattackv2.ani"},
	{0xC7739D04ABB85526ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v3.gad"},
	{0xC773A4645757EF60ULL, "vxht_brothel_n_4hq_0_0.vlm"},
	{0xC773FC41391A95A5ULL, "vxht_mansion2b_2ok_0_0.vlm"},
	{0xC774BB50CCFAF217ULL, "377325154.ogg"},
	{0xC774D237EE632BB2ULL, "6076230_m.fxe"},
	{0xC7763C4143F00E53ULL, "zz_ser_debug6078577.cl"},
	{0xC7765B02D1B28EE8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_woodpiles.msh"},
	{0xC77667FE72D350A2ULL, "vxdr_primevaltomb_7bn_0_1.vlm"},
	{0xC7766A349AEDC7D9ULL, "hlac_dr_slavers_3214_12x.dds"},
	{0xC7766D02D1B2AD79ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_woodpiles.mao"},
	{0xC7767902D1B2C112ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_woodpiles.mmh"},
	{0xC776795E1CFA2E42ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wkf.ani"},
	{0xC7768E2B17A2AA90ULL, "follower_banter6155871.cl"},
	{0xC776A218E04611A6ULL, "220880150.ogg"},
	{0xC776C1DAE0832C44ULL, "isa111_isabela6067100.cl"},
	{0xC77737FD127C85ADULL, "37094637.ogg"},
	{0xC7773C10F8F61083ULL, "116345030.ogg"},
	{0xC777629E344B896AULL, "6132664_m.fxe"},
	{0xC7779E4FCB32128FULL, "696537658.ogg"},
	{0xC777FE8C8F22E935ULL, "dae000_sandal6148240.cl"},
	{0xC778B102D48E9845ULL, "6162604_m.fxe"},
	{0xC778B646BFEA5677ULL, "6104277_m.fxe"},
	{0xC778EC354D9EA33EULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_qa1cubew.phy"},
	{0xC778ED222385FBC9ULL, "narrative_varric6124299.cl"},
	{0xC7790A194A21D9E3ULL, "sst_ss_grace.gda"},
	{0xC7792F16EF6DDA60ULL, "vxdr_primevaltomb_6j8_0_0.vlm"},
	{0xC7794BD6AC0F488CULL, "mrl150_merrill6197589.cl"},
	{0xC779519383981798ULL, "vrc221_varric6065955.cl"},
	{0xC7798AF46363E9F1ULL, "ser210_witness6168151.cl"},
	{0xC779AEF744DD8BE9ULL, "6060454_m.fxe"},
	{0xC77A0D517FECF99BULL, "119479927.ogg"},
	{0xC77A15333E7EE9AAULL, "zz_mcr_debug6080793.cl"},
	{0xC77A9804ABBE6B05ULL, "art\\characters\\creatures\\mubari\\animation\\p_tw_v3.evt"},
	{0xC77BD41AADA0E47DULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs103a_0t.dds"},
	{0xC77C0A7E91D53D79ULL, "qun200_party_comment6090409.cl"},
	{0xC77CCA68E43D5F8EULL, "476155141.ogg"},
	{0xC77DA4F790D26B2CULL, "gen00fl_anders6125570.cl"},
	{0xC77DEA2CFF652804ULL, "6091705_m.fxe"},
	{0xC77E34F5B314F0B5ULL, "dae000cr_ilen.utc"},
	{0xC77EE6AF326434EDULL, "6134320_m.fxe"},
	{0xC77F26C792C4752DULL, "gen00fl_carver6104451.cl"},
	{0xC77F5068E2E05237ULL, "vxdr_primevalentrance_3cu_0_0.vlm"},
	{0xC77F98A09D665AE8ULL, "737902912.ogg"},
	{0xC780280C0C38B311ULL, "6147663_m.fxe"},
	{0xC780468CB0878871ULL, "vxht_mansion2_1b7_0_0.vlm"},
	{0xC7809E33722E9C8AULL, "hlac_ht_playershouse_n_855_20y.dds"},
	{0xC780AD86FD68872DULL, "mer320_hubert6032192.cl"},
	{0xC781BF51D38BC7E0ULL, "dae000_orana6149044.cl"},
	{0xC781DA44F85E3062ULL, "follower_banter6172272.cl"},
	{0xC781E61D7BBB6A40ULL, "art\\levels\\races\\proxy\\props\\combatrings\\prp_30ring_uhm.mao"},
	{0xC781FA3D8B401F7AULL, "vxdr_0_deeproads_ew0_0_0.vlm"},
	{0xC78235C3E5142343ULL, "28070264.ogg"},
	{0xC783556D8050BF89ULL, "6061822_m.fxe"},
	{0xC7836B177FF636D8ULL, "vrc221_bartrand6043288.cl"},
	{0xC78394816D26C2D8ULL, "one280_merin6110000.cl"},
	{0xC783971B72600348ULL, "6136095_m.fxe"},
	{0xC784A6310B9CB32FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_trapdoorlid.mao"},
	{0xC784B2310B9CC7CCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_trapdoorlid.mmh"},
	{0xC784B8310B9CD1BEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_trapdoorlid.msh"},
	{0xC784CCE96CE02B32ULL, "hlac_w_bonepitdestroyed_a3_0_1y.dds"},
	{0xC7851E834F0A22E2ULL, "6137494_m.fxe"},
	{0xC7852041E21691E2ULL, "gen00fl_anders6116105.cl"},
	{0xC785561550693535ULL, "vxht_keep_n_8b1_0_0.vlm"},
	{0xC7856EC5FF57599CULL, "171154890.ogg"},
	{0xC78589BF7C975FF1ULL, "vxlt_0_lowtown_docks_a3_n_bup_0_0.vlm"},
	{0xC78661CF8E4C6D80ULL, "6076250_m.fxe"},
	{0xC786862C54707604ULL, "6175003_m.fxe"},
	{0xC78722DFC684B0C5ULL, "6167504_m.fxe"},
	{0xC7875664757A6453ULL, "ch1notnt_nolit.psh.fxo"},
	{0xC787B384CD7C5518ULL, "vxht_keep_n_3xy_0_0.vlm"},
	{0xC787F013C6BBE91FULL, "qcr220_grey_wardens6060398.cl"},
	{0xC7885887CDA3D628ULL, "w_bonepitmining_a2_107_0.mmh"},
	{0xC7885CB3A4C027E7ULL, "497251469.ogg"},
	{0xC7886487CDA3EACBULL, "w_bonepitmining_a2_107_0.mao"},
	{0xC7887687CDA40952ULL, "w_bonepitmining_a2_107_0.msh"},
	{0xC7888DBF895E2D67ULL, "vxht_mansion_n_4gx_0_0.vlm"},
	{0xC7892B769FD271EFULL, "hlac_ht_brothel_n_2995_3x.dds"},
	{0xC789C5ABE9713E93ULL, "gen_im_arm_cht_noba_t6.uti"},
	{0xC78ABB27C36A0649ULL, "6054013_m.fxe"},
	{0xC78B0E56C18E5CC9ULL, "vxdr_primevaltomb_6uq_0_1.vlm"},
	{0xC78B4AF111D1089AULL, "w_bonepitdestroyed_a3_25_0.rml"},
	{0xC78B5830B688BE94ULL, "vxlt_0_lowtown_eqw_0_0.vlm"},
	{0xC78BD4EF39BA6A9DULL, "vxw_bonepitdestroyed_a3_1dr_0_0.vlm"},
	{0xC78C8501C6AFF407ULL, "static_static_p88.psh.fxo.dx11"},
	{0xC78CDA708E1ABC85ULL, "424953619.ogg"},
	{0xC78CEA4870C95CBAULL, "art\\characters\\playercharacter\\humanmale\\arm_medc.msh"},
	{0xC78CF84870C974F3ULL, "art\\characters\\playercharacter\\humanmale\\arm_medc.mao"},
	{0xC78CFC4870C97B00ULL, "art\\characters\\playercharacter\\humanmale\\arm_medc.mmh"},
	{0xC78D1D8095793D2DULL, "lowtown_fem_urchin_mob6166251.cl"},
	{0xC78D63D5326B80FEULL, "isa211_isabela6072563.cl"},
	{0xC78D6D16C7D56736ULL, "6189671_m.fxe"},
	{0xC78D8EF4FFCCB5FCULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01brgdebris.mao"},
	{0xC78D9AF4FFCCC99FULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01brgdebris.mmh"},
	{0xC78DA0F4FFCCD441ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_01brgdebris.msh"},
	{0xC78DC68456E825FFULL, "gen00fl_anders6125948.cl"},
	{0xC78E368A597CF091ULL, "6087822_m.fxe"},
	{0xC78E8BA13FDE0E52ULL, "vxlt_0_lowtown_a2_n_d2n_0_0.vlm"},
	{0xC78EBAFC5278FB6EULL, "lt_0_lowtown_docks_137_0.phy"},
	{0xC78F0CB02D0F52D8ULL, "llac_ht_0_hightown_n_-131071_0.dds"},
	{0xC78F1DCA91194885ULL, "l_dae220ar_hightown.bnk"},
	{0xC78F63248BE02E7CULL, "6101440_m.fxe"},
	{0xC78F96FADD378DD5ULL, "108647902.ogg"},
	{0xC78FB6E1050773D7ULL, "31146408.ogg"},
	{0xC78FB9300B08850DULL, "6115332_m.fxe"},
	{0xC78FF25691EB28EBULL, "6149353_m.fxe"},
	{0xC79039EBAF861065ULL, "hlac_lt_0_lowtown_a2_n_-1_3z.dds"},
	{0xC790DC05DFDAEA69ULL, "zz_qun_debug6028266.cl"},
	{0xC790DFFB1D06C6A5ULL, "vxht_mansion2b_2ei_0_0.vlm"},
	{0xC7914C5D1A23220EULL, "hlac_ht_mansion2_818_48x.dds"},
	{0xC79208F864561BC8ULL, "6131195_m.fxe"},
	{0xC7925DC74C441C8AULL, "vxht_chantry_28q_0_0.vlm"},
	{0xC7934FD5110F29A8ULL, "vxlt_gallowstemplar_4eu_0_0.vlm"},
	{0xC793642363DFF9E4ULL, "hlac_lt_0_lowtown_docks_131070_4z.dds"},
	{0xC79366BE0831D18AULL, "w_bonepitdestroyed_a3_50_0.rml"},
	{0xC79453669258751CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_ave.mao"},
	{0xC7945F66925889BFULL, "art\\characters\\playercharacter\\humanfemale\\bdy_ave.mmh"},
	{0xC794656692589461ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_ave.msh"},
	{0xC7947DA072E15DB6ULL, "175304095.ogg"},
	{0xC7958323359B145CULL, "follower_banter6141801.cl"},
	{0xC795F5E27244BDB7ULL, "553112868.ogg"},
	{0xC796567B27E69176ULL, "llac_lt_0_lowtown_a2_n_0_4.dds"},
	{0xC796A3874D8E7081ULL, "hlac_ca_caverns2_5802_22z.dds"},
	{0xC796E280625E3A11ULL, "llac_dr_0_deeproads_5838_3.dds"},
	{0xC79729DDA874B01BULL, "gen00fl_varric6076040.cl"},
	{0xC7973D1006D57E84ULL, "gen_im_trs_cot_arc_sbw_r2.uti"},
	{0xC797A98AB0BBE63CULL, "pro000_opening6207391.cl"},
	{0xC797C5EBD37EF889ULL, "6174306_m.fxe"},
	{0xC79868E308AFE444ULL, "hlac_w_sundermount_5_0z.dds"},
	{0xC798C97DC57E9506ULL, "6167662_m.fxe"},
	{0xC798F4C4E015360FULL, "6051760_m.fxe"},
	{0xC79917C783282B5AULL, "hlac_lt_0_lowtown_a2_65535_2y.dds"},
	{0xC7992A0DB49AE4E6ULL, "art\\vfx\\combat\\vfx_m_cre_heal_act_c.impa.ani"},
	{0xC7995146ABDF2C84ULL, "mag110_arianni6016883.cl"},
	{0xC79A7EECEE83067FULL, "6089784_f.fxe"},
	{0xC79AAF5F5E29B54EULL, "zz_mcr_debug6080869.cl"},
	{0xC79ADFD0F12C82A7ULL, "ash_02_l3.mao"},
	{0xC79B4FE499F23A8CULL, "dae000_madam_lusine6156851.cl"},
	{0xC79B886030966D5EULL, "llac_w_blightlands_0_9.dds"},
	{0xC79B8B91498B1D06ULL, "hlac_w_encounter2_262145_0z.dds"},
	{0xC79BC9A2DF41FE38ULL, "6125176_m.fxe"},
	{0xC79BCC5BD3BB6AF7ULL, "472982786.ogg"},
	{0xC79CA35BCDA7AF17ULL, "vxlt_warehouse_n_un_0_0.vlm"},
	{0xC79CDDB832508040ULL, "vxlt_gallowscourtyard_471_0_0.vlm"},
	{0xC79CEA288E248D48ULL, "6179069_m.fxe"},
	{0xC79D0711D20AF8C1ULL, "vxca_cave_2am_0_0.vlm"},
	{0xC79D4A2D31041B82ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredround_sa.dds"},
	{0xC79D4E1863EA3E72ULL, "vxht_0_hightown_n_bcw_0_0.vlm"},
	{0xC79DA38E41A9EFB5ULL, "vxht_bartrand_n_4nr_0_0.vlm"},
	{0xC79E051E17C36395ULL, "6093489_m.fxe"},
	{0xC79E7BDA482AE5AAULL, "vxca_caverns2prc_8k8_0_0.vlm"},
	{0xC79E91A65AFB166AULL, "751782079.ogg"},
	{0xC79EB510D23EBE84ULL, "fen110_fenris6096204.cl"},
	{0xC79F0854E1FB88DCULL, "206610652.ogg"},
	{0xC79F6E6A006EE695ULL, "6042995_m.fxe"},
	{0xC79FB405DD7D9965ULL, "llac_lt_0_lowtown_n_0_9.dds"},
	{0xC7A092CC82A2F524ULL, "isa211_party_reaction6202341.cl"},
	{0xC7A13584592B9BB3ULL, "vxlt_gallowscourtyard_n_7ax_0_0.vlm"},
	{0xC7A1B945D123FF44ULL, "696922714.ogg"},
	{0xC7A1E241948CF0A8ULL, "hlac_lt_0_lowtown_a3_n_c_65536_0x.dds"},
	{0xC7A20958416B1882ULL, "755227176.ogg"},
	{0xC7A2227949EA89C4ULL, "vxlt_0_lowtown_docks_a3_97o_0_0.vlm"},
	{0xC7A24E5B78F1518DULL, "one350_variel6085176.cl"},
	{0xC7A2C158B48D8241ULL, "arena2.arl"},
	{0xC7A38BD7E86E14EEULL, "vxdr_0_deeproads_cwf_0_0.vlm"},
	{0xC7A450BD1BD2A477ULL, "386932259.ogg"},
	{0xC7A459B00B3EF968ULL, "mcr310_argument6087591.cl"},
	{0xC7A47BBD97F7CEAFULL, "vxlt_0_lowtown_docks_n_azw_0_0.vlm"},
	{0xC7A4DF29E0A17F53ULL, "qun110_petrice6018658.cl"},
	{0xC7A512844AE033FCULL, "art\\levels\\races\\elf\\props\\_textures\\prp_post_d.dds"},
	{0xC7A57910B1032B16ULL, "vxw_woundedcoast_5t_0_0.vlm"},
	{0xC7A585098D126563ULL, "mcr370_meredith6083591.cl"},
	{0xC7A5D4F3B80B5F6BULL, "art\\characters\\weapons\\shields\\w_shd_sea.mao"},
	{0xC7A6AB60BE8700D0ULL, "w_bonepit_92_0.mao"},
	{0xC7A6AF60BE8707E3ULL, "w_bonepit_92_0.mmh"},
	{0xC7A6BD60BE871F3DULL, "w_bonepit_92_0.msh"},
	{0xC7A7120DB4A6F957ULL, "art\\vfx\\combat\\vfx_m_cre_heal_act_c.impa.evt"},
	{0xC7A74A8A178211DBULL, "vxht_bartrand_n_4ey_0_0.vlm"},
	{0xC7A7560921C14CABULL, "39550696.ogg"},
	{0xC7A7E16F205DD66FULL, "396541816.ogg"},
	{0xC7A865267E898C38ULL, "88681119.ogg"},
	{0xC7A8667B22855668ULL, "lgt260_eau_tmp_roderick6143724.cl"},
	{0xC7A9056B1FDE3103ULL, "6092494_f.fxe"},
	{0xC7A907FC6CD01ECAULL, "mag100tr_alain_cower.ncs"},
	{0xC7A961FAEB556941ULL, "art\\characters\\weapons\\maces\\w_mce_mc103.mao"},
	{0xC7A967F3ABB79BA9ULL, "vxht_0_hightown_a3_cam_0_0.vlm"},
	{0xC7A9E1AB9EBF063BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basicx.msh"},
	{0xC7A9F3AB9EBF2486ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basicx.mao"},
	{0xC7A9FFAB9EBF3925ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basicx.mmh"},
	{0xC7AA93997C98D536ULL, "dae300_cullen6140602.cl"},
	{0xC7AABB577A7CA504ULL, "6136356_m.fxe"},
	{0xC7AB1538BF9DC797ULL, "vxlt_warehouse_n_xi_0_0.vlm"},
	{0xC7AB349CCE43C2A1ULL, "art\\levels\\ztools\\helpers\\character\\_textures\\mt_hed_dwfa_0d.dds"},
	{0xC7AC4FA0C9E78F35ULL, "595321239.ogg"},
	{0xC7AC9AC8D4D28192ULL, "vxw_bonepitdestroyed_a3_1gd_0_0.vlm"},
	{0xC7ACCCED13142F14ULL, "vxlt_0_lowtown_docks_b6q_0_1.vlm"},
	{0xC7ACFCA5CE9C155FULL, "dae000_mother6103078.cl"},
	{0xC7AD1F96FB2B921DULL, "vxht_chantry_2to_0_1.vlm"},
	{0xC7AD229A548A7F04ULL, "vxlt_hangedman_n_4n_0_0.vlm"},
	{0xC7AD4FD4612268FAULL, "vxdr_0_deeproads_ef3_0_0.vlm"},
	{0xC7AE130DB4AD19E0ULL, "art\\vfx\\combat\\vfx_m_cre_heal_act_c.impa.gad"},
	{0xC7AEAD1B49E2EF03ULL, "vxlt_foundry_n_w6_0_0.vlm"},
	{0xC7AEB8786AEDBF55ULL, "129271485.ogg"},
	{0xC7AF485C05E0AC89ULL, "6202148-everyoneshutup.cut"},
	{0xC7AF6A9636331A76ULL, "character_character_p100.psh.fxo.dx11.assembly"},
	{0xC7AF87AB98831BCFULL, "hlac_lt_foundry_n_320_16y.dds"},
	{0xC7AFC2ED20D11E97ULL, "vxlt_darktown_4ta_0_0.vlm"},
	{0xC7B06592C990EED1ULL, "mcr310_argument6096572.cl"},
	{0xC7B0A76E26867B94ULL, "6027779_m.fxe"},
	{0xC7B0ABD28C39DE1AULL, "526888559.ogg"},
	{0xC7B140D2D9A4E67AULL, "fxc_lof_root.dds"},
	{0xC7B19DB9E4C81385ULL, "211907344.ogg"},
	{0xC7B1D21E724481C1ULL, "hlac_lt_0_lowtown_a3_n_c_-1_7y.dds"},
	{0xC7B243EBCC0DFB30ULL, "vxlt_0_lowtown_a2_d9b_0_0.vlm"},
	{0xC7B2CC1D56875845ULL, "vis121_saemus6101172.cl"},
	{0xC7B32222FC22D49FULL, "6140951_f.fxe"},
	{0xC7B328ED28CDD3E1ULL, "6031391_m.fxe"},
	{0xC7B3340A744108AEULL, "6063837_f.fxe"},
	{0xC7B3394994D410AAULL, "6086804_m.fxe"},
	{0xC7B340854F58F843ULL, "dae000_arishok6078350.cl"},
	{0xC7B35C567EE7AF8FULL, "138195868.ogg"},
	{0xC7B453DE1CED0F50ULL, "qun231cr_qunari3.utc"},
	{0xC7B45771E70D33CDULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hcliffblend_n.dds"},
	{0xC7B47A0BB89C5111ULL, "vxlt_gallowsprison_n_6nl_0_0.vlm"},
	{0xC7B4D8F5BFC89A79ULL, "234933627.ogg"},
	{0xC7B4DD53EB30CE86ULL, "6127886_m.fxe"},
	{0xC7B4F92BD5F6E323ULL, "isa311_isabela6092381.cl"},
	{0xC7B547DF6723955AULL, "vxw_encounter1_fa_0_0.vlm"},
	{0xC7B57923C555FC46ULL, "hlac_ht_mansion2b_818_66z.dds"},
	{0xC7B59E005F988245ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_blueglowworms_e.dds"},
	{0xC7B60F17BC5EADA5ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_big_p.anb"},
	{0xC7B62A06970C4641ULL, "vxht_0_hightown_a3_c3n_0_1.vlm"},
	{0xC7B667CCD0D468CDULL, "6073775_m.fxe"},
	{0xC7B66D0CC8405266ULL, "148472033.ogg"},
	{0xC7B6D98CECACDC32ULL, "dae010_arguing_male.utc"},
	{0xC7B7351B447E1DF0ULL, "lt_gallowstemplar_f_probe_2qy.dds"},
	{0xC7B7B736EB34B75FULL, "1061197459.ogg"},
	{0xC7B83622938604C9ULL, "vxlt_gallowstemplar_51i_0_0.vlm"},
	{0xC7B887ACC6C1EA09ULL, "252842369.ogg"},
	{0xC7B8DD473143ACB8ULL, "78849404.ogg"},
	{0xC7B8EA9344016692ULL, "vxdr_primevaltomb_8nx_0_0.vlm"},
	{0xC7B91209CB383C75ULL, "one220_yevhen6152627.cl"},
	{0xC7B98325D9197FCEULL, "vxht_keep_6v7_0_0.vlm"},
	{0xC7B9BD2D189EDAA9ULL, "6109016_m.fxe"},
	{0xC7B9C82676864CD5ULL, "vxlt_0_lowtown_docks_aph_0_0.vlm"},
	{0xC7B9F38EF11A0F28ULL, "vxlt_0_lowtown_docks_n_97o_0_0.vlm"},
	{0xC7BA318CE6321029ULL, "mcr371ar_gallows_prison_nt.lst"},
	{0xC7BB0CA5C6C89A12ULL, "vxw_bonepitdestroyed_a3_1he_0_0.vlm"},
	{0xC7BB827E1799B6F4ULL, "vxlt_0_lowtown_n_e2y_0_0.vlm"},
	{0xC7BBA21951524D34ULL, "vxw_bonepitdestroyed_a3_26v_0_0.vlm"},
	{0xC7BC03A4049D2E0BULL, "6074435_m.fxe"},
	{0xC7BC887D4791A502ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_101a_d.dds"},
	{0xC7BD02FD0F0BB829ULL, "l_mag161ar_cave.wbd"},
	{0xC7BDF9DD69053C74ULL, "art\\vfx\\materials\\vfx_lightning_arc_fb_add.mao"},
	{0xC7BDF9F9B108E669ULL, "mag321_comtesse_de_launcet6064911.cl"},
	{0xC7BE515C2199713AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1b.mmh"},
	{0xC7BE555C21997871ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1b.mao"},
	{0xC7BE635C21998F80ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1b.msh"},
	{0xC7BE84B046FAC17DULL, "vxca_cavern1_haunted_ck3_0_0.vlm"},
	{0xC7BF1218B3C4261CULL, "367354082.ogg"},
	{0xC7BF2478194C779AULL, "vxlt_0_lowtown_docks_a3_n_d6h_0_0.vlm"},
	{0xC7BF666BA92F05C8ULL, "582150922.ogg"},
	{0xC7BF80BE1E35944FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv3.gad"},
	{0xC7BF9D6D1FB5ED33ULL, "vxca_caverns1_mines_j4m_0_0.vlm"},
	{0xC7BFAC0034891B97ULL, "71144217.ogg"},
	{0xC7BFE867D917A1C5ULL, "hlac_dr_slavers_3214_5z.dds"},
	{0xC7BFFEE4347D5D37ULL, "hlac_lt_gallowscourtyard_-65537_0z.dds"},
	{0xC7C06F91CC5A45EEULL, "vxdr_0_deeproads_cix_0_0.vlm"},
	{0xC7C0A2F0D00AA2F7ULL, "gen00fl_fenris6112045.cl"},
	{0xC7C0F9EF515D2177ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbed_sa.dds"},
	{0xC7C1098CE637EB60ULL, "mcr371ar_gallows_prison_nt.ncs"},
	{0xC7C1DAA717BFD026ULL, "18653733.ogg"},
	{0xC7C217FF793A4B3AULL, "hlac_lt_0_lowtown_a2_-131072_0x.dds"},
	{0xC7C2CF1658CA3100ULL, "vxht_mansion2_283_0_0.vlm"},
	{0xC7C32D6565989907ULL, "gen00fl_aveline6116710.cl"},
	{0xC7C3633EB82EB957ULL, "dae000_denier6162507.cl"},
	{0xC7C3B9BC5919ADDAULL, "vxw_encounter1_kb_0_0.vlm"},
	{0xC7C3D4849E925456ULL, "dae000ip_crowd_mf7_talk1.utp"},
	{0xC7C3E684F440F3A4ULL, "vxw_woundedcoast_81_0_0.vlm"},
	{0xC7C41E78FA47F6DCULL, "gui_guipixelshader_p0.psh.fxo.dx11"},
	{0xC7C46BE69639EE32ULL, "15398183.ogg"},
	{0xC7C47BB0F2DC8717ULL, "6085880_m.fxe"},
	{0xC7C4FA9ACDFD1A4CULL, "hlac_lt_0_lowtown_a2_n_c_65537_2z.dds"},
	{0xC7C52C19A8FA2222ULL, "728446522.ogg"},
	{0xC7C5FA97F2317523ULL, "gen00fl_aveline6139890.cl"},
	{0xC7C677D680200B05ULL, "10353593.ogg"},
	{0xC7C695BE1E3BD6FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv3.evt"},
	{0xC7C6B0C8C754242DULL, "vxht_keep_5xp_0_1.vlm"},
	{0xC7C6BB90F1D101ABULL, "686640485.ogg"},
	{0xC7C79C3DB605AB41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_mer_tw.ani"},
	{0xC7C82E1B4E89DBF8ULL, "438030213.ogg"},
	{0xC7C8B58F46982994ULL, "623572427.ogg"},
	{0xC7C92797F6F13D35ULL, "hlac_lt_gallowstemplar_f_2383_11z.dds"},
	{0xC7C99DC20A0DC5E3ULL, "745090413.ogg"},
	{0xC7C9C71A91FC41BAULL, "6110703_m.fxe"},
	{0xC7C9E066B95A822FULL, "vxlt_0_lowtown_a2_n_etx_0_0.vlm"},
	{0xC7CA0EA94C7ACCE4ULL, "art\\vfx\\combat\\vfx_str_appearance_c.impa.ani"},
	{0xC7CA200A6593A5BCULL, "gut_link_bl.mao"},
	{0xC7CA7B6BB3132CA3ULL, "hlac_lt_backalley_g_1_0x.dds"},
	{0xC7CB2A8CE64077CEULL, "mcr371ar_gallows_prison_nt.are"},
	{0xC7CB303E6FC1B876ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_mash_0n.dds"},
	{0xC7CB52A4FBED91C3ULL, "hlac_ht_0_hightown_a3_-196608_1y.dds"},
	{0xC7CB7A7DFA3BDF15ULL, "vxw_encounter1_e7_0_0.vlm"},
	{0xC7CBC6127CA89F76ULL, "w_blightlands_113_0.msh"},
	{0xC7CBD0127CA8B074ULL, "w_blightlands_113_0.mmh"},
	{0xC7CBD4127CA8B687ULL, "w_blightlands_113_0.mao"},
	{0xC7CCB25370C965CBULL, "t3_psp_qun.tnt"},
	{0xC7CCE743A60C5E10ULL, "dre000_stroud6036679.cl"},
	{0xC7CD00003BB8D9EFULL, "c_movement.wbd"},
	{0xC7CD2A39205145FDULL, "vxht_keep_n_70d_0_0.vlm"},
	{0xC7CD6397A1F096B1ULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.anb"},
	{0xC7CD669F87D16FFBULL, "132592489.ogg"},
	{0xC7CD8A6ADD6A3A4EULL, "800062701.ogg"},
	{0xC7CD96550B42901AULL, "vxlt_0_lowtown_docks_a3_n_d1i_0_0.vlm"},
	{0xC7CE9180E4346EB0ULL, "560116230.ogg"},
	{0xC7CEA73DB60BDCE3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_mer_tw.gad"},
	{0xC7CEE16EBE505E6EULL, "vxdr_0_deeproads_cfy_0_0.vlm"},
	{0xC7CF5CB7C6C220AEULL, "982487314.ogg"},
	{0xC7CF869665870008ULL, "y_crate_04.dds"},
	{0xC7CFB9FC52B09031ULL, "lt_0_lowtown_docks_137_0.mmh"},
	{0xC7CFC5FC52B0A4D2ULL, "lt_0_lowtown_docks_137_0.mao"},
	{0xC7CFD3FC52B0BCEFULL, "lt_0_lowtown_docks_137_0.msh"},
	{0xC7D0067B3244D90CULL, "6118135_f.fxe"},
	{0xC7D081846C0008FEULL, "vxlt_undercitydungeon_1np_0_0.vlm"},
	{0xC7D09EF2E8DDEE03ULL, "zz_followers_banter6157896.cl"},
	{0xC7D0BA355516DD3DULL, "dae320st_merch_armor.utm"},
	{0xC7D10BA94C80E646ULL, "art\\vfx\\combat\\vfx_str_appearance_c.impa.gad"},
	{0xC7D11944B4159559ULL, "hlac_w_encounter2_131073_0x.dds"},
	{0xC7D12AF16A17D2EDULL, "vxht_bartrand_n_6pb_0_0.vlm"},
	{0xC7D1BFB65D4AF47EULL, "627548657.ogg"},
	{0xC7D1CE4CDC47C6EEULL, "vxdr_0_deeproads_fa1_0_0.vlm"},
	{0xC7D25861E6370C24ULL, "vxw_woundedcoast_56_0_0.vlm"},
	{0xC7D27802D2011E0DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_woodpiles.phy"},
	{0xC7D27B968D195E70ULL, "6067366_m.fxe"},
	{0xC7D28B49D8AFD551ULL, "vxht_keep_n_62c_0_0.vlm"},
	{0xC7D2A0CA946C0ECBULL, "6163692_m.fxe"},
	{0xC7D2D7E789262283ULL, "gen00fl_aveline6122151.cl"},
	{0xC7D2DF6BAFC8B8E5ULL, "6148473_m.fxe"},
	{0xC7D2E592FBB5C782ULL, "llac_dr_dungeon_3217_9.dds"},
	{0xC7D32AF1120E8851ULL, "w_bonepitdestroyed_a3_25_0.mmh"},
	{0xC7D336F1120E9CF2ULL, "w_bonepitdestroyed_a3_25_0.mao"},
	{0xC7D344F1120EB40FULL, "w_bonepitdestroyed_a3_25_0.msh"},
	{0xC7D39DBE1E466E15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv3.ani"},
	{0xC7D3A01C45B4D1D1ULL, "hlac_dr_dungeon_3214_3z.dds"},
	{0xC7D51590FEC936D8ULL, "follower_banter6155592.cl"},
	{0xC7D5364124CCA067ULL, "312942967.ogg"},
	{0xC7D5843DB611BFC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_mer_tw.evt"},
	{0xC7D64A2E7F9F1725ULL, "6164813_m.fxe"},
	{0xC7D65923E2A4E2FEULL, "556190017.ogg"},
	{0xC7D6ABE05A727DD5ULL, "vxlt_0_lowtown_docks_azv_0_0.vlm"},
	{0xC7D775053E96582EULL, "vxht_0_hightown_a2_b5r_0_0.vlm"},
	{0xC7D806A94C86FC25ULL, "art\\vfx\\combat\\vfx_str_appearance_c.impa.evt"},
	{0xC7D847786596F845ULL, "vxdr_0_deeproads_ef8_0_0.vlm"},
	{0xC7D905601E3B69FBULL, "vxw_sundermount_21o_0_0.vlm"},
	{0xC7D9216BDDD95EB9ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st02a_0t.dds"},
	{0xC7D9B02D8C6C706AULL, "6032694.ogg"},
	{0xC7D9EC5AEC31F795ULL, "vxw_encounter1_j6_0_0.vlm"},
	{0xC7DA8497A1FB592AULL, "art\\vfx\\combat\\vfx_r_gn1_scloud_proj_p.mmh"},
	{0xC7DAC96EC82A4FEFULL, "vxht_0_hightown_n_aki_0_0.vlm"},
	{0xC7DAF7573DEF8D25ULL, "one120_athenril.bnk"},
	{0xC7DB1883F6B36817ULL, "vxlt_darktown_4eb_0_0.vlm"},
	{0xC7DB490F2B9E9FCFULL, "vxht_mansion2_2bv_0_0.vlm"},
	{0xC7DBDDD112E216CDULL, "vrc321_varric6086796.cl"},
	{0xC7DBFC8B327A40A0ULL, "ts_prog_onyx.gda"},
	{0xC7DD2D58E7E139F6ULL, "vxht_mansion2b_24q_0_0.vlm"},
	{0xC7DDBDFFEE55A4F4ULL, "dae150_chandan6148667.cl"},
	{0xC7DDD9573DF1905DULL, "one120_athenril.cnv"},
	{0xC7DE9BE6E47F4353ULL, "zz_mag_debug6053412.cl"},
	{0xC7DEB99AB70B3FCEULL, "vxlt_warehouse_cd_0_0.vlm"},
	{0xC7DF0D9DD55CD548ULL, "mrl150_merrill6077676.cl"},
	{0xC7DF1017BC81B4E6ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_big_p.mmh"},
	{0xC7DF39E4C2E80F2CULL, "6076870_m.fxe"},
	{0xC7DFB4DA4E09433AULL, "one300_murder_followers6086898.cl"},
	{0xC7E030EBE7075521ULL, "isa311_isabela6091841.cl"},
	{0xC7E0CA3ED82D24A4ULL, "vxw_woundedcoast_27_0_0.vlm"},
	{0xC7E11625EA770471ULL, "vxht_bartrand_n_5st_0_0.vlm"},
	{0xC7E14C80DD567528ULL, "vxht_0_hightown_c8r_0_0.vlm"},
	{0xC7E1D4FD0F2A860BULL, "l_mag161ar_cave.bnk"},
	{0xC7E20ECA3AAC9A4AULL, "plt_one100pt_magistrate.plo"},
	{0xC7E233618FA2921EULL, "vxlt_hangedman_n_w1_0_0.vlm"},
	{0xC7E283FA6C0EB253ULL, "ht_store_mmap_ph.dds"},
	{0xC7E2982E666D41E6ULL, "6075046_m.fxe"},
	{0xC7E2AB258AB9FFFFULL, "mag101_alain6032269.cl"},
	{0xC7E2C67D0E57E481ULL, "gen00fl_varric6109209.cl"},
	{0xC7E2FFC6807F1156ULL, "6064053_m.fxe"},
	{0xC7E3474BBAB52701ULL, "356953109.ogg"},
	{0xC7E3490356087613ULL, "zz_rdr_debug6066640.cl"},
	{0xC7E357A52FC14868ULL, "qalodi.arl"},
	{0xC7E3705C21B8C245ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala1b.phy"},
	{0xC7E3B0E14C8B3B12ULL, "6164238_m.fxe"},
	{0xC7E44CCDA6D768FDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_bethany.anb"},
	{0xC7E488E72F062B37ULL, "134948875.ogg"},
	{0xC7E4B8EF6BBF06F8ULL, "ser221_gamlen6131881.cl"},
	{0xC7E4CB5EED663E55ULL, "lgt320_mis_resp_swordhunter.bnk"},
	{0xC7E4F430A4D937EDULL, "6110075_f.fxe"},
	{0xC7E50B19BD12AA27ULL, "fen310_fenris6118237.cl"},
	{0xC7E5D1E2DB57A038ULL, "mage_vitalforce.dds"},
	{0xC7E683EE6646C9B5ULL, "vxw_sundermount_306_0_1.vlm"},
	{0xC7E7D06EE7F48071ULL, "6079030_m.fxe"},
	{0xC7E7ED5EED68ADADULL, "lgt320_mis_resp_swordhunter.cnv"},
	{0xC7E804B77F693F2AULL, "332397784.ogg"},
	{0xC7E870E7B0C42CF0ULL, "6191225_m.fxe"},
	{0xC7E87D7932A912F3ULL, "hlac_w_woundedcoast_262152_0x.dds"},
	{0xC7E9B342835A985CULL, "zz_mcr_debug6071535.cl"},
	{0xC7E9B3CC55A911ADULL, "art\\characters\\creatures\\spider\\c_corspidr.mmh"},
	{0xC7E9B7CC55A918FEULL, "art\\characters\\creatures\\spider\\c_corspidr.mao"},
	{0xC7E9C5CC55A93033ULL, "art\\characters\\creatures\\spider\\c_corspidr.msh"},
	{0xC7EA2A84218BEB51ULL, "nat380_sibling6103567.cl"},
	{0xC7EA7A5ECB726D8EULL, "mag311_emile_de_launcet6043823.cl"},
	{0xC7EA9FE7938B3E35ULL, "door.cur"},
	{0xC7EB856863F16D22ULL, "vxw_woundedcoast_1q9_0_0.vlm"},
	{0xC7EB9DAAC394F0F7ULL, "zz_followers_banter6159018.cl"},
	{0xC7EBB3B8FF718DFEULL, "llac_lt_0_lowtown_a2_-65537_0.dds"},
	{0xC7EC033EBE30AF71ULL, "6168000_m.fxe"},
	{0xC7ECC4D8C754D1E3ULL, "966737492.ogg"},
	{0xC7ED2B77A901584EULL, "vxlt_warehouse_de_0_0.vlm"},
	{0xC7ED3083E971EE3BULL, "hlac_ca_cavern1_haunted_24423_11x.dds"},
	{0xC7ED686FAEE4F700ULL, "vxlt_hangedman_sf_0_0.vlm"},
	{0xC7EDA7B74727386CULL, "159631658.ogg"},
	{0xC7EDB0BFFB6F5457ULL, "zz_amb_test6105069.cl"},
	{0xC7EE2CFCB95E8EA4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_walkway.mmh"},
	{0xC7EE30FCB95E95F7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_walkway.mao"},
	{0xC7EE42FCB95EB446ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_walkway.msh"},
	{0xC7EED1A901FC7B84ULL, "6103495_m.fxe"},
	{0xC7EEED2158A04DD2ULL, "6205410_m.fxe"},
	{0xC7EF1CD10EA283E9ULL, "vxlt_0_lowtown_a2_n_byr_0_0.vlm"},
	{0xC7EF22750F1DB07CULL, "dae200cr_viscount6087220.cl"},
	{0xC7EF5F5071058F70ULL, "6069166_m.fxe"},
	{0xC7EF961D9DDE23A4ULL, "907627704.ogg"},
	{0xC7F0D8AB9EFAF9D2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_basicx.phy"},
	{0xC7F0F9C6E1CA77B0ULL, "zz_one_debug6075018.cl"},
	{0xC7F1E00286E88E4EULL, "6027508_m.fxe"},
	{0xC7F1FF245ABE76C4ULL, "hlac_w_encounter2_327682_0x.dds"},
	{0xC7F30E5FF49B8F3FULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_mist_light.mmh"},
	{0xC7F36AA183D92438ULL, "mag210_arianni6024128.cl"},
	{0xC7F3DD8903FD2125ULL, "6125155_m.fxe"},
	{0xC7F3F5FD53510C15ULL, "262323001.ogg"},
	{0xC7F40796117C6E82ULL, "hla_ht_0_hightown_a3_65538_1y.dds"},
	{0xC7F40C9CF1036846ULL, "abilityusetype.gda"},
	{0xC7F49EBC35A92CD0ULL, "ave200_aveline_bar6115512.cl"},
	{0xC7F4A5E14957EDACULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxa_0.pwk"},
	{0xC7F4F59644BC9243ULL, "255497223.ogg"},
	{0xC7F5B9FACC91D990ULL, "537429781.ogg"},
	{0xC7F6637931A2DB9FULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\set_warmwindow_e.dds"},
	{0xC7F6686692AB6304ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_ave.phy"},
	{0xC7F7144F10B492A6ULL, "crt_aoe_fire.mao"},
	{0xC7F74ED3D0EBC985ULL, "vrc221ar_mansion_real.are"},
	{0xC7F779508961BF53ULL, "dre000_stroud6036713.cl"},
	{0xC7F7CA471A330449ULL, "vxca_caverns2_8fe_0_0.vlm"},
	{0xC7F82CC90F8AD0CFULL, "vxht_mansion2_2hp_0_0.vlm"},
	{0xC7F8657E96013E73ULL, "255561113.ogg"},
	{0xC7F882C8E46450BBULL, "869364645.ogg"},
	{0xC7F8CF952CFC5886ULL, "vxlt_undercitydungeon_1wc_0_0.vlm"},
	{0xC7F9493D21A82BC0ULL, "mrl000_marethari6069806.cl"},
	{0xC7F98A1A9E802C0FULL, "vxlt_warehouse_w5_0_0.vlm"},
	{0xC7FA64115916FC1BULL, "6116986_m.fxe"},
	{0xC7FA7D1CD90AD0CEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_die_v1.ani"},
	{0xC7FA86C588C91425ULL, "6119318_m.fxe"},
	{0xC7FAACCC55B7796AULL, "art\\characters\\creatures\\spider\\c_corspidr.phy"},
	{0xC7FBA7877A612945ULL, "390010574.ogg"},
	{0xC7FC47DC2227F820ULL, "clf360cr_evets.utc"},
	{0xC7FC5CEBBD2F675BULL, "254978459.ogg"},
	{0xC7FCDB80B350011FULL, "vxw_bonepitmining_a2_1vd_0_0.vlm"},
	{0xC7FCF8CBB96B1965ULL, "vxw_encounter1_eg_0_0.vlm"},
	{0xC7FD2CDE6F560EACULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody8_.ani"},
	{0xC7FD6D30CC50C5FAULL, "vxht_chantry_36v_0_0.vlm"},
	{0xC7FD90847D1C4D8EULL, "dae000_paivel6134464.cl"},
	{0xC7FDCB76E600D03AULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maga.phy"},
	{0xC7FE029FD0590914ULL, "36478815.ogg"},
	{0xC7FE2F75AEFBCF85ULL, "art\\vfx\\combat\\vfx_glb_ele_e_additive_c.anb"},
	{0xC7FE303AC142A628ULL, "vxht_0_hightown_c2p_0_0.vlm"},
	{0xC7FF0B9673531058ULL, "zz_avept_debug.ncs"},
	{0xC7FF2C19BB4E68DEULL, "vxca_cavern1_haunted_jsu_0_0.vlm"},
	{0xC7FF50740F140241ULL, "6173044_m.fxe"},
	{0xC7FFFA58D7D18207ULL, "llac_w_encounter2_262144_0.dds"},
	{0xC800A5CF2B3301A1ULL, "812955812.ogg"},
	{0xC800E5B6C093B3E5ULL, "ave200_donnic6110696.cl"},
	{0xC800E6B5AB042F51ULL, "6165645_m.fxe"},
	{0xC800FA7707A5AA62ULL, "vxdr_primevalentrance_fv_0_0.vlm"},
	{0xC80118CEE78723F9ULL, "gen_im_cns_flu_anders_01.uti"},
	{0xC8014FD3D0F4206BULL, "vrc221ar_mansion_real.ncs"},
	{0xC801A1624AA00A4EULL, "mcr310_loyalist_templars6076909.cl"},
	{0xC801ADDC3F846ED9ULL, "art\\vfx\\materials\\vfx_streak01_add.mao"},
	{0xC801B1D1B85AD0FAULL, "llac_lt_gallowstemplar_f_2392_1.dds"},
	{0xC8022340DF57859FULL, "6149450_m.fxe"},
	{0xC8026288BCD7D819ULL, "art\\characters\\weapons\\boulder\\textures\\w_boulder_d.dds"},
	{0xC802AAE4C76EB802ULL, "hlac_ht_0_hightown_a2_196607_0x.dds"},
	{0xC80327C9AF80A12BULL, "hlac_ca_cavern1_haunted_24425_7y.dds"},
	{0xC8037AFFEE780477ULL, "hlac_lt_warehouse_1_1z.dds"},
	{0xC803E7D550A6586CULL, "vxw_bonepit_244_0_0.vlm"},
	{0xC804129F7C7B6774ULL, "dae010_drunk_m1.utc"},
	{0xC8041AFC6D1CEADBULL, "mag100tr_alain_cower.utt"},
	{0xC80420231E6FBD3DULL, "6066878_m.fxe"},
	{0xC80439DE6F5C43EEULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody8_.gad"},
	{0xC8054E7067AAA48BULL, "hlac_ht_mansion2_818_12z.dds"},
	{0xC805A7B4B32345A5ULL, "361321364.ogg"},
	{0xC805CBDE9F9C1E43ULL, "genip_node_a3.utp"},
	{0xC805E6481BFD3ACBULL, "6095525_f.fxe"},
	{0xC805EAB020B4E49AULL, "653028590.ogg"},
	{0xC806DD394D76FD0AULL, "llac_lt_0_lowtown_a2_n_c_-196609_4.dds"},
	{0xC8071118FF7BEFF3ULL, "270285881.ogg"},
	{0xC8071FD99AE92F2EULL, "follower_banter6170256.cl"},
	{0xC8074F5922F58633ULL, "vfx_swirlcloud.dds"},
	{0xC807A748307201C1ULL, "genip_pp_cav_one_shot.utp"},
	{0xC807CBFB814BE95AULL, "vxlt_hangedman_1vz_0_0.vlm"},
	{0xC807FBF79076448FULL, "vxlt_warehouse_x4_0_0.vlm"},
	{0xC8084FD3D0FA3EB2ULL, "vrc221ar_mansion_real.lst"},
	{0xC808751CD917001FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_die_v1.evt"},
	{0xC808C9AF5725B5E3ULL, "p_armoire_01s.dds"},
	{0xC808D7840DAE5796ULL, "lt_0_lowtown_docks_21_0_ll.mmh"},
	{0xC808DA619350BFD4ULL, "vxlt_0_lowtown_n_cmv_0_0.vlm"},
	{0xC8092D3C2A8CC294ULL, "lgt110_gxa_vgk_silver_order6117049.cl"},
	{0xC809504EC20EDA0DULL, "llac_ca_caverns1_24421_10.dds"},
	{0xC809B25A0C9BE9C9ULL, "5207133.ogg"},
	{0xC809C67590D2A942ULL, "follower_banter6170128.cl"},
	{0xC80A4B9C94AC192EULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.impa.evt"},
	{0xC80A58003092459EULL, "777674727.ogg"},
	{0xC80A6F6F7DDF0192ULL, "hm_varian.mor"},
	{0xC80A87152B64924AULL, "dae021_keep_servant.utc"},
	{0xC80AA390F13D2F5AULL, "gen00fl_isabela6089540.cl"},
	{0xC80AEE0151B56356ULL, "elet.mao"},
	{0xC80B4D5DA546199FULL, "vxw_bonepitmining_a2_1sg_0_0.vlm"},
	{0xC80B5E05EB1C8BC3ULL, "875279423.ogg"},
	{0xC80B66898472CB8CULL, "vxw_woundedcoast_15s_0_0.vlm"},
	{0xC80B6AA8AB6131E5ULL, "vxw_encounter1_jf_0_0.vlm"},
	{0xC80BDA95B314A73AULL, "dae000_unhappy_mage6165344.cl"},
	{0xC80C191CDD7C9040ULL, "vxw_bonepitdestroyed_a3_271_0_0.vlm"},
	{0xC80D1A003BEEE9C5ULL, "c_movement.bnk"},
	{0xC80D6E3222BCD5D4ULL, "isa211_isabela6202278.cl"},
	{0xC80D9231CCECA05DULL, "hlac_ht_0_hightown_65533_1y.dds"},
	{0xC80E4AEF0AC846B8ULL, "265320693.ogg"},
	{0xC80EF4C22B397006ULL, "gen00fl_sebastian6177303.cl"},
	{0xC80F1217C3B7B98CULL, "dre000_vampire6034298.cl"},
	{0xC80F561CD91CEA08ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_die_v1.gad"},
	{0xC80F781D73242D57ULL, "waypointtypes.gda"},
	{0xC80FD1AD06F7261BULL, "vxw_bonepitmining_a2_de_0_0.vlm"},
	{0xC80FFF6A1FBB6E8EULL, "genip_trap.ncs"},
	{0xC8102A780C877562ULL, "pro000_friedrich6050162.cl"},
	{0xC81070C0563D4FF7ULL, "1010735992.ogg"},
	{0xC810A976641EFFBCULL, "27026847.ogg"},
	{0xC810F5FD34254D7AULL, "596261162.ogg"},
	{0xC81117CA3AD41A41ULL, "plt_one100pt_magistrate.nss"},
	{0xC811429C94B228D9ULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.impa.gad"},
	{0xC81154AB279D3829ULL, "vxdr_0_deeproads_ax2_0_0.vlm"},
	{0xC811B3382CDF3DEFULL, "vxdr_primevalentrance_2zk_0_0.vlm"},
	{0xC811CDB6E725BAA5ULL, "qcr220st_hunter.stg"},
	{0xC8124EB63B650C4CULL, "6070879_f.fxe"},
	{0xC812851A7E20ABA4ULL, "6035854_f.fxe"},
	{0xC812CA9A25556166ULL, "vxht_0_hightown_a3_c4e_0_0.vlm"},
	{0xC812D4A96B769CB1ULL, "229164449.ogg"},
	{0xC812E643D05F5AD7ULL, "vxlt_gallowstemplar_f_3sp_0_0.vlm"},
	{0xC81350848627ED70ULL, "618565278.ogg"},
	{0xC81413E8FF6ACF91ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightc_d.dds"},
	{0xC814269C996F2F46ULL, "6143233_m.fxe"},
	{0xC814473690DC067FULL, "813223860.ogg"},
	{0xC814A1864C57DC8DULL, "clf260_fell_orden6070486.cl"},
	{0xC8150EE68FB6068DULL, "combat_smcn_r4_a1.gda"},
	{0xC8154005D0C12D14ULL, "vxca_cave_18x_0_0.vlm"},
	{0xC8154E3DD5610354ULL, "6037381_m.fxe"},
	{0xC815559DB7DD31A7ULL, "ts_coterie_wpn_table.gda"},
	{0xC8157571B73A2B93ULL, "vxht_keep_55l_0_0.vlm"},
	{0xC8162A3133FC6FFCULL, "lt_0_lowtown_docks_166_0.mao"},
	{0xC816363133FC839FULL, "lt_0_lowtown_docks_166_0.mmh"},
	{0xC8163C3133FC8E41ULL, "lt_0_lowtown_docks_166_0.msh"},
	{0xC816938A563E3F72ULL, "vxlt_0_lowtown_a2_ewg_0_0.vlm"},
	{0xC817007F8AB72E90ULL, "306593806.ogg"},
	{0xC8176A7F6CB54525ULL, "6113518_m.fxe"},
	{0xC8179E189C8B1BAAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_basic.mao"},
	{0xC817A2189C8B22F9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_basic.mmh"},
	{0xC817AC189C8B33E7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_basic.msh"},
	{0xC817C575E0C20EF9ULL, "gen00fl_anders6096613.cl"},
	{0xC818539C94B86427ULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.impa.ani"},
	{0xC818810E7EE3A1CEULL, "vxlt_warehouse_wx_0_0.vlm"},
	{0xC818AA4CCF23684AULL, "gen00fl_anders6125653.cl"},
	{0xC818B1C65E729329ULL, "6085965_m.fxe"},
	{0xC819005C321E64CAULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_maga_0n.dds"},
	{0xC81952E7496BC7DAULL, "281981342.ogg"},
	{0xC819BE6A9838856FULL, "hlac_ca_cavern1_haunted_24427_2x.dds"},
	{0xC819D8667668E40CULL, "vxw_woundedcoast_10p_0_0.vlm"},
	{0xC81A581AE4BD5838ULL, "474220805.ogg"},
	{0xC81AC3FBFA3DD81DULL, "vxca_cavern1_haunted_j76_0_0.vlm"},
	{0xC81AFF0F281C37FFULL, "gen00fl_anders6116160.cl"},
	{0xC81B5F2C9B550DAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_lean_forward.ani"},
	{0xC81B61DCEE4F5259ULL, "vxlt_darktown_2ro_0_0.vlm"},
	{0xC81CAEF8506708A0ULL, "125707499.ogg"},
	{0xC81CE90942B224B2ULL, "gen00fl_sebastian6179785.cl"},
	{0xC81CFE38CD3BF5E2ULL, "dae011_corff_sta6152736.cl"},
	{0xC81D203F6203961AULL, "6053688_m.fxe"},
	{0xC81D6B44F9079525ULL, "llac_ca_caverns1_mines_24426_3.dds"},
	{0xC81D73924BC74881ULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_c.anb"},
	{0xC81DC1A4C619558CULL, "74191378.ogg"},
	{0xC81DFDEA0A7C599FULL, "153745866.ogg"},
	{0xC81E4389F8ED3E9BULL, "vxw_bonepitmining_a2_cd_0_0.vlm"},
	{0xC81EEE3121F997A0ULL, "1050055299.ogg"},
	{0xC81EF9201778AEC2ULL, "847818714.ogg"},
	{0xC81F2E660D91684CULL, "one100dg_kelder6055620.cl"},
	{0xC81F56874140ED0EULL, "art\\vfx\\combat\\vfx_glb_ele_c_deathblow_c.mmh"},
	{0xC81FA0100CCFDDE3ULL, "hlac_dr_slavers_3214_6y.dds"},
	{0xC81FA4EF386013E5ULL, "lgt111cr_alistair_drunk.utc"},
	{0xC81FFCC6434AFD6EULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownexit.ani"},
	{0xC8202F61882F3683ULL, "131603618.ogg"},
	{0xC820B246039AF986ULL, "dae010_prostitute_f1.utc"},
	{0xC820D6A0B64A9936ULL, "vxw_encounter1_ds_0_0.vlm"},
	{0xC820FB676D4F6E98ULL, "hlac_lt_gallowscourtyard_n_-2_2x.dds"},
	{0xC8212407CAD45653ULL, "hlac_lt_0_lowtown_2_0z.dds"},
	{0xC822CB80CD5DB3E0ULL, "6134310_m.fxe"},
	{0xC822F5188ED206F7ULL, "hlac_ht_0_hightown_-131069_0y.dds"},
	{0xC82302BF81F8CA6FULL, "vxht_0_hightown_n_apt_0_0.vlm"},
	{0xC8242EC24130EA33ULL, "vxca_caverns1_ij5_0_0.vlm"},
	{0xC8245BDF39F128BBULL, "mcr310_argument6127577.cl"},
	{0xC824AFB565381A5AULL, "vxlt_hangedman_1xt_0_0.vlm"},
	{0xC824FB76374D7688ULL, "dae000_merchant_generic_hm6172595.cl"},
	{0xC8254E7B331BA5D0ULL, "6036763_f.fxe"},
	{0xC825C924815F00C6ULL, "485835094.ogg"},
	{0xC825EF5417DFCF54ULL, "15667823.ogg"},
	{0xC826060EF60F0EB7ULL, "934018819.ogg"},
	{0xC8263075AF1D23C6ULL, "art\\vfx\\combat\\vfx_glb_ele_e_additive_c.mmh"},
	{0xC8263C9B9D9B2304ULL, "176028101.ogg"},
	{0xC8268BBF4B7AEA43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post.evt"},
	{0xC826F2EB70D9BA4EULL, "vxlt_warehouse_xy_0_0.vlm"},
	{0xC8270834F150D23AULL, "vxht_bartrand_n_5qe_0_0.vlm"},
	{0xC8272BF0CF2F007CULL, "pro000_colonel_ewald6005403.cl"},
	{0xC8274232E9D7544AULL, "dae021cr_female_worshipper.utc"},
	{0xC827654B37121402ULL, "6043634_m.fxe"},
	{0xC827663BDE623AF2ULL, "vxht_keep_n_6xn_0_0.vlm"},
	{0xC82773B123768E9EULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_att_c.impa.evt"},
	{0xC8277813EA75F40EULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv4.ani"},
	{0xC8287400BF3B01A9ULL, "llac_w_bonepitmining_a2_0_12.dds"},
	{0xC82902C7E0998DE4ULL, "vxlt_warehouse_n_vx_0_0.vlm"},
	{0xC82935D8EC33F09DULL, "vxca_cavern1_haunted_j25_0_0.vlm"},
	{0xC82941B7189B8C37ULL, "267509633.ogg"},
	{0xC829572C9B613DFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_lean_forward.evt"},
	{0xC829D251C1569138ULL, "732534747.ogg"},
	{0xC829D3B9E0456AD9ULL, "vxlt_darktown_2un_0_0.vlm"},
	{0xC82A171C8E1B43AAULL, "hlac_lt_gallowstemplar_2384_0y.dds"},
	{0xC82A1AE52E82EB98ULL, "497934268.ogg"},
	{0xC82A8AEFB226105EULL, "follower_banter6156155.cl"},
	{0xC82ABF2EFB4FE08DULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden1.mmh"},
	{0xC82AD12EFB4FFF13ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden1.msh"},
	{0xC82AE0DFB29D6262ULL, "gen00fl_sebastian6179895.cl"},
	{0xC82B54924BD351FAULL, "art\\vfx\\combat\\vfx_w_wns_shield_bash_c.mmh"},
	{0xC82B86B51423097AULL, "6167187_m.fxe"},
	{0xC82C21C8E0A25A28ULL, "850427901.ogg"},
	{0xC82C4523E2D04E7DULL, "84122509.ogg"},
	{0xC82C9C79F8C8B50DULL, "vxca_caverns2prc_9fi_0_0.vlm"},
	{0xC82CB566EAE3571BULL, "vxw_bonepitmining_a2_ng_0_0.vlm"},
	{0xC82CF4C6435579BFULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownexit.evt"},
	{0xC82D2240F1B89F18ULL, "vxlt_hangedman_n_32a_0_0.vlm"},
	{0xC82D64BF4B80C6C4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post.gad"},
	{0xC82D7587414D5F0DULL, "art\\vfx\\combat\\vfx_glb_ele_c_deathblow_c.anb"},
	{0xC82D81810036AC58ULL, "6027692_m.fxe"},
	{0xC82D8AB1237B21C9ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_att_c.impa.gad"},
	{0xC82DB562681CC2D6ULL, "6172884_m.fxe"},
	{0xC82DBA641AFDF597ULL, "follower_banter6155805.cl"},
	{0xC82E403103CBD2D3ULL, "plt_cod_cha_anders3.plo"},
	{0xC82E84DE9467A8ECULL, "lt_0_lowtown_docks_n_137_0.msh"},
	{0xC82E8ADE9467B2DEULL, "lt_0_lowtown_docks_n_137_0.mmh"},
	{0xC82E96DE9467C775ULL, "lt_0_lowtown_docks_n_137_0.mao"},
	{0xC82EC32780B41963ULL, "vxlt_0_lowtown_docks_a3_n_d2i_0_0.vlm"},
	{0xC82F400B72997018ULL, "643979446.ogg"},
	{0xC82F487DA840B1B6ULL, "vxw_encounter1_it_0_0.vlm"},
	{0xC82FBDA33199D7E3ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn02.mmh"},
	{0xC82FCBA33199EF3DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn02.msh"},
	{0xC82FE058434C6BB3ULL, "vxht_mansion2_31v_0_0.vlm"},
	{0xC82FFAE59F37A691ULL, "vxlt_0_lowtown_n_fg0_0_0.vlm"},
	{0xC8301302A02D950DULL, "dae300_alistair_king6134868.cl"},
	{0xC83021886F06E85AULL, "4393300.ogg"},
	{0xC830279C24B4B5D8ULL, "6167596_m.fxe"},
	{0xC830382C9B6727E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_lean_forward.gad"},
	{0xC8303B7BA462389DULL, "6043893_m.fxe"},
	{0xC8307BCE84ADA40AULL, "6059387_m.fxe"},
	{0xC830D3DCBB9B5621ULL, "vxw_bonepitdestroyed_a3_245_0_0.vlm"},
	{0xC83100511B2596C6ULL, "6093845_f.fxe"},
	{0xC8312131677E2A0DULL, "zz_followers_banter6158631.cl"},
	{0xC8316FF909363CAEULL, "vxdr_primevaltomb_3gp_0_0.vlm"},
	{0xC831749C73EEE2EFULL, "vxht_0_hightown_n_auw_0_0.vlm"},
	{0xC83211AC1EC3CF7BULL, "416004950.ogg"},
	{0xC83224CB04AEF3F8ULL, "dae300_alistair_king6133556.cl"},
	{0xC8322978F11E9903ULL, "art\\vfx\\environment\\vfx_blt_fire_burn_wood_p.dura.evt"},
	{0xC83236504D8AAA2FULL, "6103060_f.fxe"},
	{0xC8323DB04FB4EBD6ULL, "136012722.ogg"},
	{0xC832A09F332702B3ULL, "vxca_caverns1_im4_0_0.vlm"},
	{0xC832AA6CA60800F6ULL, "gen00fl_fenris6099831.cl"},
	{0xC832B5C2C23CEEB4ULL, "gencr_city_guard_ws_f.utc"},
	{0xC8330FA43D2A10C9ULL, "mi_gla_smo.mao"},
	{0xC83373BF4B854BA2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post.ani"},
	{0xC833BF24715099B5ULL, "6164174_m.fxe"},
	{0xC833D5C6435B63A8ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockdownexit.gad"},
	{0xC833F51A561D7456ULL, "vxht_bartrand_n_54j_0_0.vlm"},
	{0xC8344981E99585C7ULL, "mag110_vincento6016325.cl"},
	{0xC8345BB12380F017ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_att_c.impa.ani"},
	{0xC8347013EA80715FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv4.evt"},
	{0xC834BAFF54412EFFULL, "vxht_0_hightown_n_9bj_0_0.vlm"},
	{0xC834CC15F02FCFA8ULL, "gen00fl_merrill6119597.cl"},
	{0xC835602AAFBAEC38ULL, "one101cr_kelder.utc"},
	{0xC8357A11E346EABAULL, "vxht_bartrand_n_5tf_0_0.vlm"},
	{0xC835A6B3734FD721ULL, "t3_dey_gy1.tnt"},
	{0xC835DBE1449A07AAULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_conj_c.anb"},
	{0xC835EE642F268F06ULL, "6042574_m.fxe"},
	{0xC8365A62F8AFB5B3ULL, "hlac_lt_0_lowtown_a3_n_c_131069_0x.dds"},
	{0xC8366D4EE1C0D473ULL, "832595908.ogg"},
	{0xC83694FD59711BEDULL, "vxlt_0_lowtown_a2_n_fg8_0_0.vlm"},
	{0xC837278F2DB87499ULL, "follower_barks6153622.cl"},
	{0xC8374AF0F18A6C05ULL, "335036119.ogg"},
	{0xC8376D8C1BBD891EULL, "vxht_mansion_n_5c8_0_0.vlm"},
	{0xC837ECC94E2D4564ULL, "llac_lt_gallowscourtyard_-131073_0.dds"},
	{0xC837F39BEE32B634ULL, "art\\vfx\\environment\\vfx_fire_small_p.anb"},
	{0xC8384596D23B8359ULL, "vxlt_darktown_2xm_0_0.vlm"},
	{0xC8384B22B42C0996ULL, "hlac_lt_0_lowtown_a2_65535_9x.dds"},
	{0xC838536FC3EAE54CULL, "vxlt_0_lowtown_docks_b6i_0_1.vlm"},
	{0xC8390278F1247584ULL, "art\\vfx\\environment\\vfx_blt_fire_burn_wood_p.dura.gad"},
	{0xC83903635DB02E5BULL, "161975428.ogg"},
	{0xC8392A514614979BULL, "llac_lt_0_lowtown_docks_n_65534_3.dds"},
	{0xC839668F32E4A928ULL, "fen200ip_slaver_chest_1.utp"},
	{0xC839763362BC6A43ULL, "vxdr_0_deeproads_f4k_0_0.vlm"},
	{0xC83A2ADFBE644E5EULL, "189059039.ogg"},
	{0xC83A36E7152CE149ULL, "vxlt_gallowstemplar_5bp_0_0.vlm"},
	{0xC83A983431C940A6ULL, "32986200.ogg"},
	{0xC83B5113EA865B48ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv4.gad"},
	{0xC83BEF06CEB33510ULL, "544690554.ogg"},
	{0xC83C0DE9F56FCF67ULL, "zz_dae_debug6029290.cl"},
	{0xC83C10123989125EULL, "hlac_lt_0_lowtown_docks_n_-4_6z.dds"},
	{0xC83C19D0759CB7E4ULL, "hlac_w_blightlands_-131071_0y.dds"},
	{0xC83C69E590467925ULL, "follower_banter6160810.cl"},
	{0xC83D4401810DD2AEULL, "vxht_brothel_n_4gr_0_0.vlm"},
	{0xC83E11133FE5001FULL, "plt_mrl300pt_n_finstage.nss"},
	{0xC83E14D2594963DBULL, "6043554_m.fxe"},
	{0xC83E36CD5287DA7CULL, "follower_banter6161348.cl"},
	{0xC83ECAA2952AFF3CULL, "hlac_w_woundedcoast_-65531_0x.dds"},
	{0xC83FB23F63C042BAULL, "6076062_f.fxe"},
	{0xC8400397D4017EC9ULL, "vxca_caverns2_8yh_0_0.vlm"},
	{0xC8401178F12AAE62ULL, "art\\vfx\\environment\\vfx_blt_fire_burn_wood_p.dura.ani"},
	{0xC8404B3F78A5FEEFULL, "hlac_w_bonepitdestroyed_a3_262144_0z.dds"},
	{0xC8408587C877F1D4ULL, "6207560_m.fxe"},
	{0xC840BEA331A84C08ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn02.phy"},
	{0xC8415E6EC0122BD2ULL, "801932650.ogg"},
	{0xC84191CA75A4F4E8ULL, "23647457.ogg"},
	{0xC84268591204E65AULL, "character_character_p63.psh.fxo.dx11"},
	{0xC8429D7C71B8831EULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_idle.ani"},
	{0xC842A6547981CBAAULL, "6088759_f.fxe"},
	{0xC842D2C013AF8B63ULL, "946433648.ogg"},
	{0xC842FC6D29D10000ULL, "6146785_f.fxe"},
	{0xC8430D9BE421F620ULL, "sky.mat.dx11"},
	{0xC84378350DFADF46ULL, "hlac_lt_0_lowtown_docks_131070_4x.dds"},
	{0xC843B251C376C0F9ULL, "ab_cl.mao"},
	{0xC843CEC6C6643240ULL, "348868719.ogg"},
	{0xC8440D0D51960079ULL, "424002535.ogg"},
	{0xC8444948E3169C87ULL, "vxlt_0_lowtown_a2_e21_0_0.vlm"},
	{0xC844FF7B6304E588ULL, "lowtown_fem_prostitute_sta6152336.cl"},
	{0xC84554F51EF3C27EULL, "6109433_m.fxe"},
	{0xC84587F8BB29A110ULL, "llac_ht_bartrand_n_2995_2.dds"},
	{0xC845AF98F7A7958BULL, "hlac_ca_caverns2_5802_22x.dds"},
	{0xC845D29BEE3EBBF7ULL, "art\\vfx\\environment\\vfx_fire_small_p.mmh"},
	{0xC8461A84A1B1613DULL, "vxht_mansion2_3ed_0_0.vlm"},
	{0xC8465D9D2E80A731ULL, "llac_lt_undercityserialkiller_-1_36.dds"},
	{0xC84719E13181F350ULL, "vxht_0_hightown_9bb_0_0.vlm"},
	{0xC8471CE462CA1313ULL, "follower_barks6153758.cl"},
	{0xC8473D569F29E691ULL, "zz_fen_debug.ncs"},
	{0xC8473FE778C6647AULL, "6169758_m.fxe"},
	{0xC8475420D9F495D2ULL, "ss_athenril.bnk"},
	{0xC847B61A8AE3A4B1ULL, "panning1.cur"},
	{0xC847D3DDCABFB434ULL, "one100dg_kelder6060556.cl"},
	{0xC84805D2D2999E36ULL, "zz_vaultpt_debug.ncs"},
	{0xC8481FFEA79D2DE5ULL, "ave200_donnic6112135.cl"},
	{0xC848A8C40722F9C9ULL, "vxlt_gallowstemplar_5ow_0_0.vlm"},
	{0xC84942BCCE2AD843ULL, "lgt160pt_wk_avernus.ncs"},
	{0xC849BF1EF9620F2DULL, "fen210_fenris6109430.cl"},
	{0xC84A126A27F85087ULL, "843674881.ogg"},
	{0xC84A671BFEA045B3ULL, "vxlt_warehouse_qg_0_0.vlm"},
	{0xC84AE2BAB99AAC41ULL, "gen00fl_aveline6140041.cl"},
	{0xC84B37E63E3BA05CULL, "w_woundedcoast_48_0.mao"},
	{0xC84B43E63E3BB4FFULL, "w_woundedcoast_48_0.mmh"},
	{0xC84B49E63E3BBEA1ULL, "w_woundedcoast_48_0.msh"},
	{0xC84B54DC34B68E5EULL, "521076360.ogg"},
	{0xC84B880BB39A9135ULL, "224612996.ogg"},
	{0xC84BA19ADE8E6F4CULL, "225130148.ogg"},
	{0xC84BF190DAF52B27ULL, "vxlt_gallowstemplar_f_5j9_0_0.vlm"},
	{0xC84C60578B234E6AULL, "hlac_lt_gallowstemplar_n_2388_2x.dds"},
	{0xC84D2F2F1FB24A16ULL, "w_sundermount_201_0.msh"},
	{0xC84D392F1FB25B14ULL, "w_sundermount_201_0.mmh"},
	{0xC84D3D2F1FB26227ULL, "w_sundermount_201_0.mao"},
	{0xC84DE252F8D18287ULL, "vxlt_gallowstemplar_n_419_0_0.vlm"},
	{0xC84E39100BE8E217ULL, "qun131_taskmaster6039545.cl"},
	{0xC84F661F3D71B52EULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef01.phy"},
	{0xC84F7FBC1FFBAD90ULL, "gen00fl_sebastian6179996.cl"},
	{0xC84F901D058596C0ULL, "470789292.ogg"},
	{0xC84FB1FB60228FB1ULL, "hlac_ht_brothel_n_3367_1z.dds"},
	{0xC8500600D4F49241ULL, "ht_keep_n_4023_ll.mmh"},
	{0xC8501E274FCDBA0FULL, "hf_ibethany_h2s1.mor"},
	{0xC850522A81C22E36ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_posting_n.dds"},
	{0xC85060D1CFA83C92ULL, "hlac_ht_0_hightown_a2_1_1y.dds"},
	{0xC850A77FD3F461E1ULL, "6118071_f.fxe"},
	{0xC850B57C71C4E9EFULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_idle.evt"},
	{0xC850D2A709751187ULL, "caustic_big.mao"},
	{0xC850EB3A4E6475C4ULL, "art\\characters\\playercharacter\\qunarimale\\hed_stn.msh"},
	{0xC850FD3A4E64942DULL, "art\\characters\\playercharacter\\qunarimale\\hed_stn.mao"},
	{0xC851013A4E649B66ULL, "art\\characters\\playercharacter\\qunarimale\\hed_stn.mmh"},
	{0xC8511823E47FB897ULL, "6094455_m.fxe"},
	{0xC85204133FF59330ULL, "plt_mrl300pt_n_finstage.plo"},
	{0xC85220569F33BD56ULL, "zz_fen_debug.cnv"},
	{0xC852BAB6986C48CBULL, "zz_trap_core.ncs"},
	{0xC852CDF8412F2482ULL, "vxlt_warehouse_8u_0_0.vlm"},
	{0xC852F03914CD91C7ULL, "6075375_m.fxe"},
	{0xC8530F1CA412E000ULL, "6049656_m.fxe"},
	{0xC8536C2707A90D1BULL, "lt_warehouse_n_water_1849.mmh"},
	{0xC853782707A921B8ULL, "lt_warehouse_n_water_1849.mao"},
	{0xC8538A2707A94005ULL, "lt_warehouse_n_water_1849.msh"},
	{0xC853C2401FD65A5EULL, "6125273_m.fxe"},
	{0xC853C6D210E9DAFEULL, "6102533_m.fxe"},
	{0xC853DDFFE01B9935ULL, "rdr111dg_martin6053577.cl"},
	{0xC853E4B0885C0F90ULL, "hlac_lt_undercityencounter_-1_30y.dds"},
	{0xC853ECB1D28B398FULL, "6162310_m.fxe"},
	{0xC8547F5EA555189EULL, "vxlt_hangedman_21p_0_0.vlm"},
	{0xC8551D4B9E3C5831ULL, "llac_lt_0_lowtown_docks_a3_-65538_6.dds"},
	{0xC855332647FFB607ULL, "vxca_caverns1_ikj_0_0.vlm"},
	{0xC8556A63ABA2E545ULL, "647404228.ogg"},
	{0xC855BCC7EF9856A1ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.impa.ani"},
	{0xC85654ABB552E4F7ULL, "438106814.ogg"},
	{0xC85667570D58B5EAULL, "528527210.ogg"},
	{0xC856ABCB4207F1EAULL, "6170730_m.fxe"},
	{0xC856B67C71C95698ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_idle.gad"},
	{0xC856EE63E613A3F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedenter.evt"},
	{0xC857315439CE0CEAULL, "684482486.ogg"},
	{0xC857B96C5861499FULL, "6076100_f.fxe"},
	{0xC857CAE795E76EF9ULL, "vxlt_0_lowtown_e8i_0_0.vlm"},
	{0xC857F6A5F9E67C5FULL, "hlac_w_woundedcoast_-65528_0y.dds"},
	{0xC8582AA23F0C82D8ULL, "395976744.ogg"},
	{0xC85885C6F7380691ULL, "hlac_ht_0_hightown_n_65535_3x.dds"},
	{0xC8588ACE775586E1ULL, "364080538.ogg"},
	{0xC858D8F8F0965E33ULL, "vxlt_warehouse_zf_0_0.vlm"},
	{0xC859C476E64F2363ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maga.msh"},
	{0xC859D276E64F3ABDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_maga.mmh"},
	{0xC85ACF630A88DA91ULL, "llac_lt_0_lowtown_n_-3_1.dds"},
	{0xC85AEFC437909767ULL, "ss_shadow_m_1.bnk"},
	{0xC85AF83A4E6CE031ULL, "art\\characters\\playercharacter\\qunarimale\\hed_stn.phy"},
	{0xC85BA4C61CCD7CC4ULL, "hlac_lt_hangedman_n_-1_24z.dds"},
	{0xC85C3158455D693FULL, "dae200_bran6142981.cl"},
	{0xC85C5D4ED35F877FULL, "6076056_f.fxe"},
	{0xC85C9DA309689C27ULL, "157247474.ogg"},
	{0xC85CB7FCD425233FULL, "734797941.ogg"},
	{0xC85CC4E36D8CACB1ULL, "zz_mer_debug6073693.cl"},
	{0xC85CC7C7EF9E8843ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.impa.gad"},
	{0xC85CE563E617FF81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedenter.gad"},
	{0xC85EC4844C4F8F6BULL, "vxw_bonepitmining_a2_h1_0_0.vlm"},
	{0xC85EDEE144BD119DULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_conj_c.mmh"},
	{0xC85EE0D2D39DAFB6ULL, "hlac_ca_caverns1_mines_24421_3z.dds"},
	{0xC85F68F560F3C81BULL, "w_bonepitdestroyed_a3_85_0.mmh"},
	{0xC85F74F560F3DCB8ULL, "w_bonepitdestroyed_a3_85_0.mao"},
	{0xC85F86F560F3FB05ULL, "w_bonepitdestroyed_a3_85_0.msh"},
	{0xC85FC710F20791BCULL, "479893578.ogg"},
	{0xC860528CD21662E0ULL, "llac_lt_0_lowtown_a2_n_0_2.dds"},
	{0xC860608F3023F278ULL, "isa111_isabela6067085.cl"},
	{0xC860FD9A4ADEAB6CULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v2.gad"},
	{0xC8612D02C702CD87ULL, "vxlt_0_lowtown_a2_e83_0_0.vlm"},
	{0xC8614310FCB21B88ULL, "vxdr_0_deeproads_cxm_0_0.vlm"},
	{0xC8615A26D09D15FEULL, "349074216.ogg"},
	{0xC86174AAF452458CULL, "plt_cod_hst_calendar.plo"},
	{0xC861901A226415BAULL, "455740975.ogg"},
	{0xC861B41577FD06C7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage2.gad"},
	{0xC861D0F463C8DB83ULL, "6140783_m.fxe"},
	{0xC86220C93726D740ULL, "6116977_f.fxe"},
	{0xC86267FA7F40BF89ULL, "pro000_athenril.bnk"},
	{0xC862A04095CA4D9AULL, "mag210_arianni6024522.cl"},
	{0xC862D057BE41B927ULL, "gen00fl_dog6165761.cl"},
	{0xC863A4C7EFA46B28ULL, "art\\vfx\\combat\\vfx_w_fns_fenris1_loop_c.impa.evt"},
	{0xC863A50339F5CE87ULL, "vxca_caverns1_ine_0_0.vlm"},
	{0xC863AB4BAAA4EA37ULL, "mer210_coterie_contact6036776.cl"},
	{0xC863B10C1061A412ULL, "qun111_ketojan6019084.cl"},
	{0xC863E663E61E202FULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\warriortwohandedselectedenter.ani"},
	{0xC86423A15CB2957AULL, "6163558_m.fxe"},
	{0xC86451C4D6F695AEULL, "hla_ht_0_hightown_n_2_0y.dds"},
	{0xC864AA1118F4E0DBULL, "6148732_m.fxe"},
	{0xC864AB13320ADDA4ULL, "hlac_lt_0_lowtown_a3_n_c_65535_8y.dds"},
	{0xC864E41E864DA3D5ULL, "hlac_lt_0_lowtown_a2_n_196607_0z.dds"},
	{0xC8652AFA3C79D2D9ULL, "vxlt_undercityencounter_1is_0_0.vlm"},
	{0xC86591EF045F177BULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.dura.evt"},
	{0xC865C4D6F0F697B2ULL, "6175001_m.fxe"},
	{0xC865D76D2C9CC1EAULL, "836499818.ogg"},
	{0xC866247B0931326AULL, "vxlt_undercityserialkiller_1c3_0_1.vlm"},
	{0xC8663CC487DD8779ULL, "vxlt_0_lowtown_e3h_0_0.vlm"},
	{0xC86649FA7F4475C9ULL, "pro000_athenril.cnv"},
	{0xC866694874AD3190ULL, "ico_amulet_5.dds"},
	{0xC8666A12D101F0FEULL, "vxw_woundedcoast_1b8_0_0.vlm"},
	{0xC8677AA4EECFDFF2ULL, "6119663_m.fxe"},
	{0xC867CA03A0E449A8ULL, "432217373.ogg"},
	{0xC867FC9A4AE4C86AULL, "art\\characters\\creatures\\ragedemon\\animation\\com_v2.ani"},
	{0xC8685C53A71E26E1ULL, "6201483_f.fxe"},
	{0xC868999856F01C2EULL, "vxht_brothel_n_4vu_0_0.vlm"},
	{0xC868A915780312D4ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage2.evt"},
	{0xC86956B3380570ECULL, "lowtown_male_alienage1_sta6162785.cl"},
	{0xC869C426C2432FE2ULL, "one100dg_nabil6033122.cl"},
	{0xC869E718DFD15FF6ULL, "w_bonepitmining_a2_45_0.rml"},
	{0xC869EF487B43844BULL, "fen110_fenris6118024.cl"},
	{0xC869FBE81C22DE45ULL, "art\\levels\\races\\proxy\\props\\_textures\\char_test_d.dds"},
	{0xC86A1F64352BE478ULL, "6041833_meredith_order.cut"},
	{0xC86AE400D8E7A072ULL, "qun221_petrice_cleric.stg"},
	{0xC86B4FD8FD87FBC9ULL, "vxht_0_hightown_byq_0_0.vlm"},
	{0xC86B72647A5515ACULL, "vxlt_warehouse_18f_0_0.vlm"},
	{0xC86B9AB48DE16232ULL, "6096607_m.fxe"},
	{0xC86BB410E1DD596CULL, "vxlt_0_lowtown_a3_n_c_i0p_0_0.vlm"},
	{0xC86BC2F01B0133E5ULL, "dae000_junar6147929.cl"},
	{0xC86BDE2A9949C1D2ULL, "vxlt_foundry_n_123_0_0.vlm"},
	{0xC86C0996881CB6AEULL, "vxht_keep_n_5xi_0_0.vlm"},
	{0xC86C0CF90EF32B08ULL, "382498053.ogg"},
	{0xC86C105474376B02ULL, "w_sundermount_216_0.msh"},
	{0xC86C125474376EB8ULL, "w_sundermount_216_0.mmh"},
	{0xC86C1E547437835BULL, "w_sundermount_216_0.mao"},
	{0xC86C58C7C8712557ULL, "vxht_0_hightown_a2_a87_0_0.vlm"},
	{0xC86C8AEF046529BCULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.dura.gad"},
	{0xC86C9AD1A746B061ULL, "781599235.ogg"},
	{0xC86CA19C93321AF5ULL, "hlac_lt_0_lowtown_a2_n_-65535_4x.dds"},
	{0xC86CC7D5DD14740DULL, "430888109.ogg"},
	{0xC86D1DBE10498A60ULL, "6103226_m.fxe"},
	{0xC86D4CAF1259671AULL, "6065322_m.fxe"},
	{0xC86D73BFBB4EC511ULL, "rdr111st_martin.stg"},
	{0xC86D9734A098FF5CULL, "hlac_ht_0_hightown_a2_-131074_1y.dds"},
	{0xC86E39968DAAE446ULL, "mag131_taraneh6088684.cl"},
	{0xC86E3CE11C878A39ULL, "vxlt_hangedman_262_0_0.vlm"},
	{0xC86E58EFA23798FBULL, "6029006_m.fxe"},
	{0xC86E61DE949D5179ULL, "lt_0_lowtown_docks_n_137_0.phy"},
	{0xC86E690EAF6A478CULL, "6170516_m.fxe"},
	{0xC86E9B65FEE2E1DEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraitsb_sa.dds"},
	{0xC86EA6A82B083B7BULL, "169193352.ogg"},
	{0xC86EE665DAA949B4ULL, "275904823.ogg"},
	{0xC86EF4767BAD0543ULL, "follower_banter6136675.cl"},
	{0xC86F575CBFC10984ULL, "557535011.ogg"},
	{0xC86F6F8326E30F85ULL, "hlac_ca_caverns2_5802_33x.dds"},
	{0xC86F892547DBA62AULL, "gen_im_cns_pot_level_01.uti"},
	{0xC86F8C6068F5FDD9ULL, "835951646.ogg"},
	{0xC86F943BF1A30786ULL, "6067774_f.fxe"},
	{0xC86FB4EDEEA83408ULL, "vxdr_0_deeproads_cuj_0_0.vlm"},
	{0xC86FD85EDCDA16FDULL, "one130_thug6131595.cl"},
	{0xC86FFCD4B47B459FULL, "l_env_wal_lowtowndocks_callout.bnk"},
	{0xC8713C3DAB8AA4C5ULL, "6067110_m.fxe"},
	{0xC87258BB95B16FEFULL, "static_static_p28.psh.fxo"},
	{0xC872E3D45EFC3754ULL, "6133580_m.fxe"},
	{0xC872F218D3F2215BULL, "vxlt_hangedman_n_2fo_0_0.vlm"},
	{0xC87321CFEE8EDBCFULL, "vxlt_0_lowtown_docks_bw0_0_0.vlm"},
	{0xC87370715BEE854EULL, "vxw_encounter3_4i_0_0.vlm"},
	{0xC87389EF046B46BAULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.dura.ani"},
	{0xC874026CD6029C81ULL, "6187708_m.fxe"},
	{0xC87421DACC1039E2ULL, "vxlt_0_lowtown_a2_ded_0_0.vlm"},
	{0xC87440FA5EF54D55ULL, "vxht_0_hightown_n_c6n_0_0.vlm"},
	{0xC87444E63E5EBB3CULL, "w_woundedcoast_48_0.anb"},
	{0xC8748C1FD09522A3ULL, "vxlt_undercityencounter_1b4_0_0.vlm"},
	{0xC8748FAB6D5A1523ULL, "6174803_m.fxe"},
	{0xC874C127FF6C5290ULL, "501114932.ogg"},
	{0xC874DBEFC2F8097EULL, "vxw_woundedcoast_1m9_0_0.vlm"},
	{0xC875A115780D8F0DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage2.ani"},
	{0xC8761954743FCF43ULL, "w_sundermount_216_0.phy"},
	{0xC87679C5F75AF625ULL, "956808320.ogg"},
	{0xC876864510CED7AEULL, "hlac_lt_darktown_-1_56y.dds"},
	{0xC87693A33F24BC01ULL, "vxlt_hangedman_sk_0_0.vlm"},
	{0xC87693F33B52D016ULL, "lt_0_lowtown_a2_n_c_mmap_ph.dds"},
	{0xC8770CB1FCBEF083ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.cess.evt"},
	{0xC8775122DE42AEBAULL, "vxht_keep_4n9_0_0.vlm"},
	{0xC877AC484D76C437ULL, "6090345_m.fxe"},
	{0xC87800614DC20FDFULL, "qun120cr_javaris6043984.cl"},
	{0xC87825D391B95EDAULL, "dae300_alistair_king6133565.cl"},
	{0xC8784A4B162B926CULL, "6064554_m.fxe"},
	{0xC8787DF43B45E025ULL, "540887197.ogg"},
	{0xC87A45CAB9133500ULL, "hm_cgn_008.mor"},
	{0xC87A45CAB9133502ULL, "hm_cgn_008.mop"},
	{0xC87A4631E7059B36ULL, "vxw_bonepitdestroyed_a3_26d_0_0.vlm"},
	{0xC87A73E61D5FFD7AULL, "130363181.ogg"},
	{0xC87A963CFC4422E8ULL, "286491994.ogg"},
	{0xC87AA3189CDFA1E6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ff_basic.phy"},
	{0xC87B872B38041F54ULL, "zz_vault_debug6200364.cl"},
	{0xC87BFAAFA733D0D0ULL, "vxlt_0_lowtown_n_enx_0_1.vlm"},
	{0xC87D0382A39588D7ULL, "768570055.ogg"},
	{0xC87DA0EE2880194CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left.ani"},
	{0xC87DA549C249A603ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chanterb_sa.dds"},
	{0xC87DE5B1FCC4CD04ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.cess.gad"},
	{0xC87DEFB96089DFCFULL, "vxht_keep_5id_0_0.vlm"},
	{0xC87E0E7E582EAFC5ULL, "mag330_walter6044532.cl"},
	{0xC87E2B5014B07D5BULL, "hlac_lt_0_lowtown_a3_n_c_-131073_2x.dds"},
	{0xC87E3974596CD7E4ULL, "t1_mub_ta3.tnt"},
	{0xC87E68FC2F6BC8E4ULL, "44490656.ogg"},
	{0xC87EBBC48B66493FULL, "1066657105.ogg"},
	{0xC87EDC3097E73C4CULL, "vxca_caverns1_mines_eh5_0_0.vlm"},
	{0xC87F336D7B3D94FBULL, "anim_ambient.gda"},
	{0xC880251CEE05A684ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3b.msh"},
	{0xC88030082431D368ULL, "llac_lt_0_lowtown_docks_a3_-131074_8.dds"},
	{0xC880371CEE05C5EDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3b.mao"},
	{0xC8803B1CEE05CC26ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala3b.mmh"},
	{0xC88163F5C5E839DBULL, "vxlt_hangedman_n_2mn_0_0.vlm"},
	{0xC881762A4F398212ULL, "vxw_woundedcoast_kn_0_0.vlm"},
	{0xC881E24E4DE49DCEULL, "vxw_encounter3_1j_0_0.vlm"},
	{0xC881F452C55A5FA3ULL, "103621190.ogg"},
	{0xC8822586BFEEF43EULL, "vxlt_darktown_6og_0_0.vlm"},
	{0xC882A45A3B325B5AULL, "hlac_ht_keep_4519_12y.dds"},
	{0xC882B2D750EB65D5ULL, "vxht_0_hightown_n_c3m_0_0.vlm"},
	{0xC8830B355E5878C2ULL, "148894639.ogg"},
	{0xC88343F2A9C4EA2FULL, "6075692_m.fxe"},
	{0xC8834A419573EF34ULL, "vxdr_primevalentrance_26x_0_0.vlm"},
	{0xC8834E65AAA91F30ULL, "6049739_m.fxe"},
	{0xC8838EB7893F9A21ULL, "dae021pt_amb_chantry.ncs"},
	{0xC8840E40CAB7CB36ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.dura.ani"},
	{0xC8849E1FD931307CULL, "657046365.ogg"},
	{0xC884ADEE28864D8EULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left.gad"},
	{0xC884F4B1FCCB05E2ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.cess.ani"},
	{0xC88518E436890C36ULL, "6084316_m.fxe"},
	{0xC885237CB7AEF1E0ULL, "12980552.ogg"},
	{0xC8854656BB72E903ULL, "715143541.ogg"},
	{0xC88568FA3947BBDDULL, "lgt110tr_coa_meet_start.ncs"},
	{0xC8859B5314535EEDULL, "195958264.ogg"},
	{0xC886CC0F3A41DA07ULL, "vxht_0_hightown_n_be4_0_0.vlm"},
	{0xC886E5C437B5C09DULL, "ss_shadow_m_1.wbd"},
	{0xC88779F6F2F4C1E6ULL, "4578279.ogg"},
	{0xC8878E9A8011F329ULL, "6143270_m.fxe"},
	{0xC88817FE200DCCA0ULL, "115181674.ogg"},
	{0xC88823B4663D98E8ULL, "vxdr_0_deeproads_9cj_0_0.vlm"},
	{0xC88843CE09ACEB38ULL, "zz_mcr_debug6088346.cl"},
	{0xC888561E5E4146ACULL, "vxlt_warehouse_12d_0_0.vlm"},
	{0xC888B79115FDE173ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_9.pwk"},
	{0xC888B80ED8FBB3B6ULL, "vxw_bonepitdestroyed_a3_21e_0_0.vlm"},
	{0xC8897342FCC83659ULL, "947896655.ogg"},
	{0xC889AC37D06AC542ULL, "6077286_m.fxe"},
	{0xC889C7F852BFF2FAULL, "6042703_m.fxe"},
	{0xC889FA021702C6B2ULL, "6136354_m.fxe"},
	{0xC88A0DE7ECE7A70EULL, "835956554.ogg"},
	{0xC88A2FEB2A20075BULL, "6092291_m.fxe"},
	{0xC88A3044DE0A603EULL, "535108112.ogg"},
	{0xC88A6C8C9929E950ULL, "vxlt_0_lowtown_n_ekg_0_1.vlm"},
	{0xC88AA7692FC64021ULL, "476846035.ogg"},
	{0xC88AA8EE288AB16DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_gesture_left.evt"},
	{0xC88AA970B1033D6FULL, "6191248_m.fxe"},
	{0xC88AB05470935189ULL, "vxca_cavern1_haunted_k2n_0_2.vlm"},
	{0xC88AE77DE75D0D60ULL, "one100dg_lia6049677.cl"},
	{0xC88AF7F710FBFCE7ULL, "vxca_caverns2prc_9ub_0_0.vlm"},
	{0xC88B0CB5BE949E15ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_robb_0s.dds"},
	{0xC88B31EC24FBE24DULL, "6110810_m.fxe"},
	{0xC88B804D5F670A7FULL, "dae120ip_at_brothel.utp"},
	{0xC88BA348F27B3F2CULL, "lt_0_lowtown_docks_n_6_0.rml"},
	{0xC88CB6D2D39F010AULL, "zz_dae_debug6029121.cl"},
	{0xC88D5E38CB965C87ULL, "81848398.ogg"},
	{0xC88D6104DDC72A7BULL, "659075353.ogg"},
	{0xC88DE65BC1ACCD6AULL, "llac_ca_caverns1_24422_5.dds"},
	{0xC88E7AF53D6BC92CULL, "llac_lt_0_lowtown_65535_3.dds"},
	{0xC88EAD44DF4AA05CULL, "vxht_mansion2_2gf_0_0.vlm"},
	{0xC88ED4F89F9CBDB1ULL, "vxht_keep_6g7_0_0.vlm"},
	{0xC88F651F3DA797F1ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef01.mmh"},
	{0xC88F7F1F3DA7C3AFULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef01.msh"},
	{0xC88FC78E42E86854ULL, "6139380_m.fxe"},
	{0xC8909732C63E1E2EULL, "vxlt_0_lowtown_docks_c4a_0_0.vlm"},
	{0xC890B1FB24A72F95ULL, "6014248_reunionhappy.cut"},
	{0xC890F74018F3ABAAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wmda_0n.dds"},
	{0xC8911E44974C8623ULL, "mag200_feynriel6023886.cl"},
	{0xC89124B442E17E55ULL, "vxht_0_hightown_n_c8l_0_0.vlm"},
	{0xC89148EB6F3A6C25ULL, "vxlt_hangedman_rt_0_0.vlm"},
	{0xC8918E4E83ACF10DULL, "363373466.ogg"},
	{0xC891C5947F271A61ULL, "6105401_m.fxe"},
	{0xC891C9CC3E5B342BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_dismissive.ani"},
	{0xC891D60C02D8D6F3ULL, "art\\characters\\creatures\\strider\\animation\\leap.ani"},
	{0xC89215D109707C47ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.anb"},
	{0xC8921640CAC41627ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.dura.evt"},
	{0xC893208574D58B45ULL, "1048799069.ogg"},
	{0xC89327B2E537A831ULL, "unit_sphere.mmh"},
	{0xC89333B2E537BCD2ULL, "unit_sphere.mao"},
	{0xC89341B2E537D4EFULL, "unit_sphere.msh"},
	{0xC89362EA6347216AULL, "6163437_m.fxe"},
	{0xC8937D45D6695466ULL, "vxht_keep_731_0_0.vlm"},
	{0xC893D236B8AC5A09ULL, "6048802_f.fxe"},
	{0xC89419C695E634A2ULL, "493758352.ogg"},
	{0xC894289B3A56DB65ULL, "6097608_f.fxe"},
	{0xC8944774A999A682ULL, "332737108.ogg"},
	{0xC8952E509E3B0B0BULL, "lgt220_mag_tmplr_templar6140553.cl"},
	{0xC89627651E47C3B5ULL, "gen00fl_dog6114925.cl"},
	{0xC8962781820E2AB6ULL, "hlac_ca_caverns1_24426_12z.dds"},
	{0xC89695915833B168ULL, "vxdr_0_deeproads_9dk_0_0.vlm"},
	{0xC8983740CAC8BA70ULL, "art\\vfx\\combat\\vfx_r_arc_burst_smo_aoe_p.dura.gad"},
	{0xC89877BBDBB79ACBULL, "gen00fl_aveline6053286.cl"},
	{0xC898A1C81C161FDBULL, "6097191_m.fxe"},
	{0xC899FF687497C758ULL, "dre180cs_bartrandbetrayal.cut"},
	{0xC89A004F3268D64FULL, "953889113.ogg"},
	{0xC89A01D381699821ULL, "6060177_m.fxe"},
	{0xC89A20C38D00B3DEULL, "dae000_merchant_tailor_hm6164356.cl"},
	{0xC89A7408C2DF90CFULL, "hlac_lt_0_lowtown_a2_n_0_6y.dds"},
	{0xC89AB1AE11804AEDULL, "qun211cr_elf_warrior6072350.cl"},
	{0xC89AE14EC1085B75ULL, "681754741.ogg"},
	{0xC89AF9C02048129DULL, "gen_im_trs_prf_rocks_04.uti"},
	{0xC89AFA0C769A812BULL, "vxw_encounter2_3j_0_0.vlm"},
	{0xC89B3FDD8632DDBAULL, "6126734_f.fxe"},
	{0xC89BE2D4B4A054D5ULL, "l_env_wal_lowtowndocks_callout.wbd"},
	{0xC89CB8CE1242430DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hangbar_d.dds"},
	{0xC89CD81389F27CC7ULL, "zz_ave_debug6134000.cl"},
	{0xC89CEAFA5DF282DAULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.dura.gad"},
	{0xC89D419A81F7824CULL, "705239633.ogg"},
	{0xC89DE90FB02C315BULL, "sib131_slave_a6165271.cl"},
	{0xC89DED974500DC0AULL, "vxlt_0_lowtown_a2_gmo_0_0.vlm"},
	{0xC89E43C6EA1DD6A3ULL, "hla_ht_0_hightown_a3_-131070_0y.dds"},
	{0xC89E4685EA09B911ULL, "119033884.ogg"},
	{0xC89E9921DD7434C4ULL, "6101264_m.fxe"},
	{0xC89EB0E4AAD39363ULL, "pet_03.mao"},
	{0xC89ECC8E557E6953ULL, "hlac_ht_mansion2b_818_20y.dds"},
	{0xC89ECE0C02E3532AULL, "art\\characters\\creatures\\strider\\animation\\leap.evt"},
	{0xC89F090FB83436AEULL, "vxlt_0_lowtown_docks_c1b_0_0.vlm"},
	{0xC89F0EE9CCEE8BA6ULL, "6083986_m.fxe"},
	{0xC89F38506599518CULL, "320702877.ogg"},
	{0xC89FB1CC3E6748A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_dismissive.evt"},
	{0xC89FF24C63BCA69FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose3.anb"},
	{0xC8A014E081584FCFULL, "vxht_0_hightown_n_cbh_0_0.vlm"},
	{0xC8A088A9EEF7F784ULL, "fen110_party_comment6095248.cl"},
	{0xC8A09C547EB16393ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_tall.phy"},
	{0xC8A0D26513AE13ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_loop.ani"},
	{0xC8A0E25EF04E34C9ULL, "dae200_cullen6140286.cl"},
	{0xC8A0E4491A5229EBULL, "412801650.ogg"},
	{0xC8A1419578689EC2ULL, "363507464.ogg"},
	{0xC8A15A371CE06ADFULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rock_002_drelief.dds"},
	{0xC8A161A0808AE9E1ULL, "6161195_m.fxe"},
	{0xC8A169F52B46E3CEULL, "vxht_mansion_n_611_0_0.vlm"},
	{0xC8A1917EA20AF9AEULL, "lt_0_lowtown_docks_n_22_0.rml"},
	{0xC8A1A6720D3794DEULL, "vxlt_backalley_g_1jm_0_0.vlm"},
	{0xC8A229208DC388FAULL, "art\\characters\\creatures\\qunari\\textures\\qm_har_stn_0n.dds"},
	{0xC8A24ABB610226FDULL, "83057152.ogg"},
	{0xC8A278BEB30B8AC8ULL, "6127970_m.fxe"},
	{0xC8A294FF68C14379ULL, "isa211_isabela6202215.cl"},
	{0xC8A3AFC91E2E0B07ULL, "vxht_0_hightown_n_bk6_0_0.vlm"},
	{0xC8A3B05825001EBFULL, "vxca_caverns1_ghe_0_0.vlm"},
	{0xC8A3CC45DB470968ULL, "vxlt_0_lowtown_n_fek_0_0.vlm"},
	{0xC8A3EDFA5DF8A6F1ULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.dura.evt"},
	{0xC8A3FC0BCE8CA214ULL, "24429081.ogg"},
	{0xC8A447C178FC4A70ULL, "6118468_m.fxe"},
	{0xC8A49B57A3449D06ULL, "6125656_f.fxe"},
	{0xC8A4FB13B37CDFD5ULL, "715793841.ogg"},
	{0xC8A5458A463DB0D3ULL, "736737220.ogg"},
	{0xC8A5567EDBAB2933ULL, "vxw_encounter3_6v_0_0.vlm"},
	{0xC8A5574232F20377ULL, "vxht_bartrand_n_68h_0_0.vlm"},
	{0xC8A573404C3BF68EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_ready.ani"},
	{0xC8A5AD0C02E9399DULL, "art\\characters\\creatures\\strider\\animation\\leap.gad"},
	{0xC8A60DA982088BE8ULL, "lt_0_lowtown_a2_53_0.anb"},
	{0xC8A656E757C3C19FULL, "1010220690.ogg"},
	{0xC8A6AAC4DC14787FULL, "static_static_p6.psh.fxo.dx11.assembly"},
	{0xC8A6B0CC3E6D65B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\talk_dismissive.gad"},
	{0xC8A6B504A93D19E9ULL, "684644368.ogg"},
	{0xC8A6DF6513B2952EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_loop.gad"},
	{0xC8A79F6B8BAFA9D1ULL, "vxw_encounter2_bk_0_0.vlm"},
	{0xC8A7A65BEB615400ULL, "6206156_m.fxe"},
	{0xC8A7D0ED91BE17ADULL, "vxw_woundedcoast_105_0_0.vlm"},
	{0xC8A7DED07986FC79ULL, "462664801.ogg"},
	{0xC8A826915709B235ULL, "vxlt_darktown_5ii_0_0.vlm"},
	{0xC8A830907B1A2763ULL, "lla_ht_0_hightown_a2_-131072_0.dds"},
	{0xC8A84542778933A3ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_cassan_epilogue.ani"},
	{0xC8A89EA2E870F7B8ULL, "135235133.ogg"},
	{0xC8A9382F2000B80FULL, "w_sundermount_201_0.phy"},
	{0xC8A94DA112C12360ULL, "516799999.ogg"},
	{0xC8AAAF1B26C6191FULL, "vxht_chantry_2t5_0_1.vlm"},
	{0xC8AABA05EC97D4A5ULL, "6204959_m.fxe"},
	{0xC8AAC52CBC7B72BCULL, "10432138.ogg"},
	{0xC8AB2782B8D08888ULL, "hlac_w_bonepitmining_a2_0_4x.dds"},
	{0xC8AB86039122BBC3ULL, "llac_lt_0_lowtown_a3_n_c_-65536_8.dds"},
	{0xC8AB90FEC336D15CULL, "vxht_mansion2_2md_0_0.vlm"},
	{0xC8ABB8B28388EEB1ULL, "vxht_keep_6m5_0_0.vlm"},
	{0xC8AC4210F3C51F1AULL, "mag101_grace6034209.cl"},
	{0xC8AC99E26ABF002CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brrckbrd_n.dds"},
	{0xC8AC9CE623ED0E41ULL, "vxw_encounter1_f7_0_1.vlm"},
	{0xC8ACCBC1251BCB92ULL, "vxw_woundedcoast_zi_0_0.vlm"},
	{0xC8AD0CC6151D65B7ULL, "vxht_mansion_n_6cb_0_0.vlm"},
	{0xC8AD161AF1051649ULL, "hlac_lt_gallowstemplar_n_2383_3y.dds"},
	{0xC8AD8669316BA714ULL, "47879566.ogg"},
	{0xC8ADEE39FD0B14A0ULL, "atl_itemupgrad_dxt5_dat.xml"},
	{0xC8ADF0429D4EC32AULL, "6076238_carver_intro.cut"},
	{0xC8AE9633F4C83CB8ULL, "s_blood4.mao"},
	{0xC8AE9E9E30F52135ULL, "dae150_chandan6205590.cl"},
	{0xC8AF1B8BE80D8530ULL, "seb221cr_brett_harimann.utc"},
	{0xC8AF4BC30F355A4BULL, "222504055.ogg"},
	{0xC8AF6A8FC479FEE1ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be02a.phy"},
	{0xC8AFE8ECF9BDD112ULL, "vxw_bonepit_1vg_0_0.vlm"},
	{0xC8B009461ECED996ULL, "336155515.ogg"},
	{0xC8B0184EFF2DAD5EULL, "vxlt_backalley_g_1wn_0_0.vlm"},
	{0xC8B0408D2A67D6CFULL, "546149457.ogg"},
	{0xC8B0F7AAB399A55FULL, "vxht_bartrand_n_5y4_0_0.vlm"},
	{0xC8B105FA5E0359C0ULL, "art\\vfx\\creatures\\fxc_shale_ice_r_c.dura.ani"},
	{0xC8B135520F6E89A2ULL, "479268316.ogg"},
	{0xC8B1363D685D0DF8ULL, "lt_gallowscourtyard_n_probe_7az.dds"},
	{0xC8B19A707B12B87CULL, "39668172.ogg"},
	{0xC8B19F0C0BDC10C9ULL, "6180437_m.fxe"},
	{0xC8B1CC37B0A957EFULL, "6041538_m.fxe"},
	{0xC8B21CEA4B6EBF17ULL, "441601991.ogg"},
	{0xC8B26B404C4673DFULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_ready.evt"},
	{0xC8B2A1D4C348B553ULL, "mcr371_post_harvester6077600.cl"},
	{0xC8B2CC5955E362AFULL, "6163362_m.fxe"},
	{0xC8B40DF623B95047ULL, "vxdr_0_deeproads_f4w_0_0.vlm"},
	{0xC8B441C229561DA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackright.ani"},
	{0xC8B543E63E9650CCULL, "w_woundedcoast_48_0.rml"},
	{0xC8B572F1F1616334ULL, "vxht_0_hightown_a2_bg2_0_0.vlm"},
	{0xC8B611487DA5C251ULL, "vxw_encounter2_mj_0_0.vlm"},
	{0xC8B6986E48FFCAB5ULL, "vxlt_darktown_5rh_0_0.vlm"},
	{0xC8B6A36BC4AE2CAFULL, "336443107.ogg"},
	{0xC8B77B62823B2F3FULL, "6083486_m.fxe"},
	{0xC8B7BA798C8CBC87ULL, "gencr_decoy.utc"},
	{0xC8B7DDC65A86B22BULL, "vxw_encounter2_9l_0_0.vlm"},
	{0xC8B7DEC126ED0A3FULL, "6144307_m.fxe"},
	{0xC8B7EECD1A49BDA3ULL, "6036742_f.fxe"},
	{0xC8B841A78EAAC975ULL, "934822009.ogg"},
	{0xC8B8509FD15EABA3ULL, "zz_seb_debug6189574.cl"},
	{0xC8B9263BD8CCE94DULL, "vxht_bartrand_n_4ld_0_0.vlm"},
	{0xC8B94BBE6EAA4CC5ULL, "6055043_m.fxe"},
	{0xC8B94C404C4C5DC8ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_ready.gad"},
	{0xC8B9619EE434CE20ULL, "261748866.ogg"},
	{0xC8B96C606C29CDC7ULL, "llac_ca_caverns1_24421_12.dds"},
	{0xC8B9805E06EE016CULL, "vxlt_0_lowtown_docks_a3_d3v_0_0.vlm"},
	{0xC8B9C22E9967DFBBULL, "hlac_lt_0_lowtown_a2_-196609_0x.dds"},
	{0xC8B9F608EBECC895ULL, "mrl310_party_comments6143075.cl"},
	{0xC8B9FEE9ACD310FEULL, "6098968_m.fxe"},
	{0xC8BA4EA48DBA6363ULL, "174404421.ogg"},
	{0xC8BA89639671D3E1ULL, "6030406_m.fxe"},
	{0xC8BAA43212482AECULL, "qcr000_arishok6060101.cl"},
	{0xC8BAB6CED4930006ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.impa.evt"},
	{0xC8BB058FF7234713ULL, "hlac_ht_chantry_2534_15z.dds"},
	{0xC8BB12D109937C48ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.mmh"},
	{0xC8BB2AF056FCEBF4ULL, "zz_mag_debug6054614.cl"},
	{0xC8BB3D9E1711E412ULL, "vxw_woundedcoast_wj_0_0.vlm"},
	{0xC8BB4CC2295C4F43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackright.gad"},
	{0xC8BB90AB3A101EBDULL, "vxw_encounter3_bd_0_0.vlm"},
	{0xC8BC04AD40C3B194ULL, "vxlt_0_lowtown_a2_n_c_hz2_0_0.vlm"},
	{0xC8BC0FD41DB7B3C8ULL, "hlac_ht_0_hightown_a2_131072_0y.dds"},
	{0xC8BC5E3B66BCD24CULL, "hlac_w_blightlands_2_0y.dds"},
	{0xC8BC759E0ACB0920ULL, "vxlt_gallowstemplar_f_46z_0_1.vlm"},
	{0xC8BC98AB797D53E3ULL, "hlac_lt_0_lowtown_docks_-3_7z.dds"},
	{0xC8BCCEF77BE4470CULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_mag_0n.dds"},
	{0xC8BD524AC25DBEDEULL, "479671031.ogg"},
	{0xC8BD5C42779BB68DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_cassan_epilogue.gad"},
	{0xC8BEA561591DD6BAULL, "qun211cr_braggart_ally_m2.utc"},
	{0xC8BEC7DFC2CEC291ULL, "mag110_samson6016160.cl"},
	{0xC8BEEDC2BDBB1146ULL, "670359154.ogg"},
	{0xC8BF28DE845BAC76ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_gesturemaker.evt"},
	{0xC8BF5C7896F7BBC8ULL, "6129770_m.fxe"},
	{0xC8BF80786AA5EF27ULL, "vxlt_gallowstemplar_f_521_0_0.vlm"},
	{0xC8BF94395A1AB021ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackleft.ani"},
	{0xC8C051FCB7AB023FULL, "6091823_m.fxe"},
	{0xC8C071BF5C284CCFULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.impa.gad"},
	{0xC8C0BD1FE4FCC15CULL, "follower_banter6143379.cl"},
	{0xC8C0F4C048602894ULL, "hlac_w_bonepitmining_a2_393216_0z.dds"},
	{0xC8C1292728E49F4DULL, "hlac_ca_caverns1_mines_24422_5z.dds"},
	{0xC8C1CDCED4994651ULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.impa.gad"},
	{0xC8C1EC9C0C368AC6ULL, "vxw_bonepitdestroyed_a3_8b_0_0.vlm"},
	{0xC8C229C229623228ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkbackright.evt"},
	{0xC8C2939ED7348AD7ULL, "239619917.ogg"},
	{0xC8C2E26FBC504498ULL, "follower_banter6141755.cl"},
	{0xC8C36829513012F4ULL, "cd_mag360_alain.bnk"},
	{0xC8C3F47B14A5E07DULL, "vxlt_0_lowtown_a2_n_c_eou_0_0.vlm"},
	{0xC8C435BE3B2546E5ULL, "6127046_f.fxe"},
	{0xC8C46A2B61FD25EEULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\gra_roots_d.dds"},
	{0xC8C483256F9BDAD1ULL, "vxw_encounter2_hi_0_0.vlm"},
	{0xC8C4B2F306FE3479ULL, "w_encounter1_57_0.msh"},
	{0xC8C4BCF306FE4567ULL, "w_encounter1_57_0.mmh"},
	{0xC8C4C0F306FE4BB4ULL, "w_encounter1_57_0.mao"},
	{0xC8C4C99ACD4C6B34ULL, "93071118.ogg"},
	{0xC8C50F979CA2038BULL, "vxht_0_hightown_a3_bun_0_0.vlm"},
	{0xC8C51FDE84600801ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_gesturemaker.gad"},
	{0xC8C54CE64D8C0013ULL, "dae200cr_saemus.utc"},
	{0xC8C59E2D02CF06B8ULL, "1035007413.ogg"},
	{0xC8C5B4AE290A3138ULL, "6037690_f.fxe"},
	{0xC8C6275719811DC3ULL, "follower_banter6156215.cl"},
	{0xC8C62CF48F9FDD83ULL, "gen00fl_dog6107347.cl"},
	{0xC8C63027054AACC6ULL, "256446933.ogg"},
	{0xC8C69F395A20E1C3ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackleft.gad"},
	{0xC8C73CD328FEC227ULL, "6110038_m.fxe"},
	{0xC8C76AC2DB1414B3ULL, "hlac_lt_gallowstemplar_2393_0x.dds"},
	{0xC8C76B6144F784E9ULL, "mrl310_party_comments6142771.cl"},
	{0xC8C786BF5C2E8F7CULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.impa.evt"},
	{0xC8C7CECED49DB3FFULL, "art\\vfx\\combat\\vfx_c_mer_fightrecover_c.impa.ani"},
	{0xC8C7F23AF8E419ECULL, "vxlt_0_lowtown_docks_a3_d4w_0_0.vlm"},
	{0xC8C8154758688CCFULL, "hlac_lt_foundry_n_320_5y.dds"},
	{0xC8C8A343CB37B1A4ULL, "llac_ht_chantry_2534_21.dds"},
	{0xC8C8C66D8E0DF9BDULL, "hf_lgt_sophia_dryden.mor"},
	{0xC8C94FB99842C815ULL, "art\\vfx\\textures\\rockarmor\\vfx_rockarmor_0d.dds"},
	{0xC8C957F4DE8307CFULL, "zz_isa_debug6175886.cl"},
	{0xC8C994992DFB50EBULL, "800936197.ogg"},
	{0xC8C9B4E95F800462ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_win.phy"},
	{0xC8C9EB63A21936E2ULL, "w_sundermount_66_0_ll.mmh"},
	{0xC8C9FC24E2B99DE5ULL, "6135684_m.fxe"},
	{0xC8CA16EE5B590518ULL, "gen00fl_anders6062183.cl"},
	{0xC8CA28A3F927E95DULL, "gen_im_arm_cht_hvy_set_high.uti"},
	{0xC8CA356CD0666642ULL, "vxlt_gallowsprison_n_6nc_0_0.vlm"},
	{0xC8CA63C35E5EC4FCULL, "vxht_0_hightown_a3_c8c_0_0.vlm"},
	{0xC8CA768A32B9CA14ULL, "vxlt_0_lowtown_a2_n_c_hu3_0_0.vlm"},
	{0xC8CA8B9CB221941AULL, "6018573_m.fxe"},
	{0xC8CA9E9373E99250ULL, "6076742_m.fxe"},
	{0xC8CAC78CBCB74905ULL, "17150276.ogg"},
	{0xC8CADD7B121A97CEULL, "vxht_keep_n_6ma_0_0.vlm"},
	{0xC8CB69E97AE866D4ULL, "art\\vfx\\cinematic\\vfx_lightbeam02_p.mmh"},
	{0xC8CB9D0A357A3D48ULL, "hlac_lt_gallowstemplar_n_2391_2x.dds"},
	{0xC8CBCEA9718A13DEULL, "vxw_bonepitmining_a2_1j5_0_0.vlm"},
	{0xC8CBD918E0241243ULL, "w_bonepitmining_a2_45_0.msh"},
	{0xC8CBE718E024299DULL, "w_bonepitmining_a2_45_0.mmh"},
	{0xC8CBEB18E02430AEULL, "w_bonepitmining_a2_45_0.mao"},
	{0xC8CC20DE846628AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_gesturemaker.ani"},
	{0xC8CC31B14D2A095FULL, "one280_iwan6097270.cl"},
	{0xC8CD3151BAB02C14ULL, "ht_bartrand_n_2995.rml"},
	{0xC8CD59FA398557A0ULL, "lgt110tr_coa_meet_start.utt"},
	{0xC8CD7AD5A3BAE1BEULL, "hlac_w_sundermount_-262139_0y.dds"},
	{0xC8CD7C395A26C4A8ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkbackleft.evt"},
	{0xC8CDEB84CCA26E4CULL, "lgt000im_ultra_toe.uti"},
	{0xC8CE8D8BAB6234B0ULL, "vxlt_gallowscourtyard_7a0_0_0.vlm"},
	{0xC8CEC3C2334E7FC8ULL, "103422598.ogg"},
	{0xC8CEF2C63C36376BULL, "ss_fereldan_f_1.wbd"},
	{0xC8CF471AACEA0AF5ULL, "6153792_m.fxe"},
	{0xC8CFF0BC20ADB0F8ULL, "673745297.ogg"},
	{0xC8D0768DEA56DD78ULL, "vxlt_backalley_n_1dh_0_0.vlm"},
	{0xC8D0C389BA617104ULL, "vxw_bonepitdestroyed_a3_275_0_0.vlm"},
	{0xC8D0CAEBE2CF8235ULL, "6110670_m.fxe"},
	{0xC8D1E5195F17411FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appb_0n.dds"},
	{0xC8D2619D0E1AA17DULL, "154698642.ogg"},
	{0xC8D29C4515D711E7ULL, "6198693_f.fxe"},
	{0xC8D2BB072F7EAE8EULL, "vxdr_0_deeproads_9ej_0_0.vlm"},
	{0xC8D331E4E9F0C3E1ULL, "dae000cr_grand_cleric6173663.cl"},
	{0xC8D48E7B393F7AFBULL, "991189581.ogg"},
	{0xC8D48EBF5C392695ULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.impa.ani"},
	{0xC8D5070C4B741CC2ULL, "vxht_mansion_n_4zb_0_0.vlm"},
	{0xC8D5DFF662B693F8ULL, "dae000_sandal6147787.cl"},
	{0xC8D62D20D5860CFCULL, "gen00fl_merrill6108634.cl"},
	{0xC8D6371C7AF8D7ABULL, "vxw_sundermount_2b6_0_0.vlm"},
	{0xC8D6A31405B881B0ULL, "vxht_keep_5ka_0_0.vlm"},
	{0xC8D7532E46C3693EULL, "static_static_p69.psh.fxo.dx11"},
	{0xC8D832F871D45049ULL, "hla_ht_0_hightown_n_65535_0x.dds"},
	{0xC8D90523980F1978ULL, "6150491_m.fxe"},
	{0xC8D93E318C2395CAULL, "vxlt_0_lowtown_a2_n_enw_0_0.vlm"},
	{0xC8D94F580410B04EULL, "vxht_keep_n_6hb_0_0.vlm"},
	{0xC8D95083EB697F48ULL, "art\\vfx\\textures\\vfx_blood.dds"},
	{0xC8D9D4CDFE83D102ULL, "vxht_keep_n_67c_0_0.vlm"},
	{0xC8DA0FB7824B0E2BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkfront.evt"},
	{0xC8DAE6EE6B702086ULL, "hlac_x_narrativeroom_0_0y.dds"},
	{0xC8DAF115F58457B6ULL, "lt_0_lowtown_docks_a3_n_155_0.phy"},
	{0xC8DBA68CF28DAD72ULL, "6042905_m.fxe"},
	{0xC8DBF17FF72B1927ULL, "ser120_ghyslain6105000.cl"},
	{0xC8DBFBC6C3C0B5A6ULL, "hlac_ht_0_hightown_a3_65537_0x.dds"},
	{0xC8DC45E54A7F5641ULL, "6098901_m.fxe"},
	{0xC8DC6CEB5CA6EB01ULL, "vxht_mansion2_2v4_0_0.vlm"},
	{0xC8DC827FE10069A8ULL, "hlac_lt_undercityserialkiller_-1_28z.dds"},
	{0xC8DCD9E2AF94468EULL, "llac_lt_0_lowtown_a2_n_65533_2.dds"},
	{0xC8DD2D7CEBEC4F57ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperleft.mmh"},
	{0xC8DD317CEBEC5664ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperleft.mao"},
	{0xC8DD437CEBEC74C9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperleft.msh"},
	{0xC8DD48CBC945B8E7ULL, "vxlt_gallowscourtyard_7e7_0_0.vlm"},
	{0xC8DD8F7F0CB2D7C7ULL, "dae000_bartrand6206880.cl"},
	{0xC8DDB5A71FFF1F54ULL, "mcr371_hostile_fenris6082226.cl"},
	{0xC8DE34C3DD64EF46ULL, "vxw_sundermount_379_0_0.vlm"},
	{0xC8DEC9DB4F852481ULL, "72237905.ogg"},
	{0xC8DECDFE31F5084EULL, "hlac_ht_playershouse_n_855_24y.dds"},
	{0xC8DEE86ADC4CF5F8ULL, "vxlt_backalley_n_1oi_0_0.vlm"},
	{0xC8DF90BF2188C7BDULL, "vxht_0_hightown_n_bxm_0_0.vlm"},
	{0xC8DFB4AA0A04EBD4ULL, "253895134.ogg"},
	{0xC8DFD09649082FACULL, "dae000ip_crowd_mf7_talk3.utp"},
	{0xC8DFE3767A8D2F34ULL, "234577060.ogg"},
	{0xC8E02D37E8924B20ULL, "fen210_orana6109709.cl"},
	{0xC8E041F0F58B1323ULL, "llac_ca_caverns1_24423_1.dds"},
	{0xC8E0492EC7254812ULL, "ser311cr_quentin.ncs"},
	{0xC8E128B7825157ECULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkfront.gad"},
	{0xC8E208F05549F0E7ULL, "nat380_nathaniel6066709.cl"},
	{0xC8E22D3E7EC3F240ULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_spr01.phy"},
	{0xC8E2B2AFA16A7F32ULL, "6096016_m.fxe"},
	{0xC8E2BB23BA58F8CEULL, "vxlt_warehouse_9v_0_0.vlm"},
	{0xC8E347BE434B7561ULL, "dm_amb_commoner3.mor"},
	{0xC8E382A3072BCA4DULL, "gentr_teammove.utt"},
	{0xC8E3C10F6C758EADULL, "6165599_m.fxe"},
	{0xC8E438020B6223BFULL, "vxlt_0_lowtown_docks_a3_d3y_0_0.vlm"},
	{0xC8E45F97CACE6267ULL, "character_character_p228.psh.fxo.dx11.assembly"},
	{0xC8E4A5C0A9BB226DULL, "hlac_lt_undercityencounter_-1_25z.dds"},
	{0xC8E4CF196BEF074FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_coma_0t.dds"},
	{0xC8E4DA15F58C8557ULL, "lt_0_lowtown_docks_a3_n_155_0.msh"},
	{0xC8E4E5CB852967E4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02gravecairn_1.mao"},
	{0xC8E4EC15F58CA3BAULL, "lt_0_lowtown_docks_a3_n_155_0.mao"},
	{0xC8E4F015F58CAAC9ULL, "lt_0_lowtown_docks_a3_n_155_0.mmh"},
	{0xC8E5188FB8A7E421ULL, "hlac_lt_0_lowtown_docks_n_0_2z.dds"},
	{0xC8E5238BD0410D3DULL, "ser221_gascard6165651.cl"},
	{0xC8E558728B0F335CULL, "vxht_mansion2_2yx_0_0.vlm"},
	{0xC8E560605FD8E5CEULL, "vxlt_0_lowtown_a2_emm_0_0.vlm"},
	{0xC8E59EF2C3F0286CULL, "dae000cr_denier.utc"},
	{0xC8E6024CA725A6FCULL, "6208288_m.fxe"},
	{0xC8E60CA9823E6DA0ULL, "lt_0_lowtown_a2_53_0.rml"},
	{0xC8E6DC735374299AULL, "mcr310_cut_chantry_blows_up6087572.cl"},
	{0xC8E72C0284128CE6ULL, "mcr370_assignments6176719.cl"},
	{0xC8E7477D424AF5FCULL, "vxht_0_hightown_a3_c6a_0_0.vlm"},
	{0xC8E75893661D5FB1ULL, "6076286_m.fxe"},
	{0xC8E7A62DB3FC5FADULL, "918890972.ogg"},
	{0xC8E7C134F606C8CEULL, "vxht_keep_n_6cc_0_0.vlm"},
	{0xC8E827B7825774EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\walkfront.ani"},
	{0xC8E846AAF079E982ULL, "vxht_keep_n_62l_0_0.vlm"},
	{0xC8E8491162406711ULL, "332432471.ogg"},
	{0xC8E8C05449284F9FULL, "981976355.ogg"},
	{0xC8E8E1E5B572DF7FULL, "6131000_m.fxe"},
	{0xC8E9851E5A277788ULL, "zz_pro_debug6061830.cl"},
	{0xC8E9D8B2F330B836ULL, "364051579.ogg"},
	{0xC8EA7AC51E4EC800ULL, "vxht_0_hightown_c3w_0_0.vlm"},
	{0xC8EB2F6568756468ULL, "vxht_keep_3ye_0_0.vlm"},
	{0xC8EC70BE2495094DULL, "dae000ip_crowd_mf6_talk3.utp"},
	{0xC8EC8502C24762D4ULL, "one300_nuncio6085034.cl"},
	{0xC8ED34D50B815BA9ULL, "6162327_m.fxe"},
	{0xC8ED5C9AF7A35747ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1_0.pwk"},
	{0xC8ED5CD81DED3113ULL, "328064905.ogg"},
	{0xC8EE0123A17D3277ULL, "w_bonepit_236_0.msh"},
	{0xC8EE1323A17D50DAULL, "w_bonepit_236_0.mao"},
	{0xC8EE1723A17D57E9ULL, "w_bonepit_236_0.mmh"},
	{0xC8EE24437F1012E0ULL, "25037748.ogg"},
	{0xC8EE410CBB9F1695ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0sc_basic_2.mao"},
	{0xC8EE8FE95F9EE116ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_win.mao"},
	{0xC8EE9BE95F9EF5B5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_win.mmh"},
	{0xC8EE9DE95F9EF92BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_win.msh"},
	{0xC8EEFBEF0AE360DAULL, "vrc321_party_barks6095783.cl"},
	{0xC8EF080ED5C93B97ULL, "hlac_ca_cavern1_haunted_24424_16z.dds"},
	{0xC8EF2BB07415522CULL, "6088776_m.fxe"},
	{0xC8EF4388E97F1476ULL, "fen310_fenris6118692.cl"},
	{0xC8F0652E72844D0BULL, "vxht_0_hightown_a3_bjc_0_0.vlm"},
	{0xC8F0A5F30723589CULL, "w_encounter1_57_0.phy"},
	{0xC8F0DEEF19260B26ULL, "gen00fl_aveline6131394.cl"},
	{0xC8F1B0D68B94B50BULL, "shd_dot.mao"},
	{0xC8F28FF24B9DA243ULL, "follower_banter6148717.cl"},
	{0xC8F2A9DEFD583C3FULL, "vxlt_0_lowtown_docks_a3_d4x_0_0.vlm"},
	{0xC8F2E88A31C2EFD4ULL, "hm_gen_act2.mor"},
	{0xC8F33899B47220D2ULL, "6141231_m.fxe"},
	{0xC8F38772A355BB4FULL, "vxw_bonepitdestroyed_a3_1h4_0_0.vlm"},
	{0xC8F3A7150281DAB4ULL, "vxca_caverns2prc_8cn_0_0.vlm"},
	{0xC8F3CC9B1878B639ULL, "1026008457.ogg"},
	{0xC8F46CE97B0B708BULL, "art\\vfx\\cinematic\\vfx_lightbeam02_p.anb"},
	{0xC8F47D68E4AEF541ULL, "vxlt_undercityencounter_1eo_0_0.vlm"},
	{0xC8F4A637F41C8C30ULL, "ss_coterie_thug_f.bnk"},
	{0xC8F4BC97B1ED4EDDULL, "vxht_0_hightown_a2_6pb_0_0.vlm"},
	{0xC8F4D4C63C56D339ULL, "ss_fereldan_f_1.bnk"},
	{0xC8F4E02F6F0A5258ULL, "vxlt_gallowstemplar_f_438_0_0.vlm"},
	{0xC8F510E08D9228E9ULL, "dae000_male_templar_mob6174358.cl"},
	{0xC8F527B40B6CD61DULL, "6083978_m.fxe"},
	{0xC8F5649AE43179C9ULL, "zz_dae_debug6047627.cl"},
	{0xC8F642B1D15FDAD5ULL, "dae000_madam_lusine6162389.cl"},
	{0xC8F67E9C01BD4594ULL, "hlac_lt_0_lowtown_a2_n_131071_4y.dds"},
	{0xC8F6893CC38592CEULL, "6073790_m.fxe"},
	{0xC8F71246DB739373ULL, "vxht_0_hightown_a2_9ux_0_0.vlm"},
	{0xC8F72AC1699FC730ULL, "vxlt_0_lowtown_docks_bb8_0_0.vlm"},
	{0xC8F75042739D482CULL, "lgt110_gxa_vgk_silver_order6117052.cl"},
	{0xC8F7D4A3D1934092ULL, "6110783_m.fxe"},
	{0xC8F8266B236E654EULL, "945450872.ogg"},
	{0xC8F8C40660CEDB74ULL, "zz_followers_banter6158517.cl"},
	{0xC8F8DCA544EF8CE6ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.cess.ani"},
	{0xC8F8ECA21044E080ULL, "vxht_0_hightown_c6t_0_0.vlm"},
	{0xC8F8F379C3BCDF16ULL, "vxlt_undercityserialkiller_17b_0_0.vlm"},
	{0xC8F934065351CA06ULL, "6105550_m.fxe"},
	{0xC8F95251BAD59F14ULL, "ht_bartrand_n_2995.anb"},
	{0xC8F95C25ECDC0D0AULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.impa.gad"},
	{0xC8F9B97ACAD3B2E7ULL, "lt_0_lowtown_32_0.rml"},
	{0xC8F9BD5DD7E0D6EEULL, "6106995_m.fxe"},
	{0xC8F9C9BE2FA67817ULL, "vxht_mansion2_1vm_0_0.vlm"},
	{0xC8F9F6A8158F86A1ULL, "vxlt_0_lowtown_docks_9zr_0_0.vlm"},
	{0xC8FA5D1520633569ULL, "vxht_0_hightown_a3_b90_0_0.vlm"},
	{0xC8FA764127C86A4FULL, "6041475_m.fxe"},
	{0xC8FA83385155F99CULL, "mag170_hugh6170934.cl"},
	{0xC8FBA223187A73ECULL, "character_character_p216.psh.fxo.dx11.assembly"},
	{0xC8FBA6B1FD777429ULL, "6167027_m.fxe"},
	{0xC8FBC37C91A669BFULL, "6174880_f.fxe"},
	{0xC8FC87B5FA4235EBULL, "6033661_m.fxe"},
	{0xC8FCD32D9FA19A62ULL, "6124729_f.fxe"},
	{0xC8FD1ABB1FFABE01ULL, "vxlt_0_lowtown_docks_b51_0_1.vlm"},
	{0xC8FD9D8D660B6AACULL, "6083676_m.fxe"},
	{0xC8FDF41848D9C74AULL, "gen00fl_merrill6108629.cl"},
	{0xC8FE2B9A6FDF6DB7ULL, "hlac_ca_caverns1_mines_24421_11x.dds"},
	{0xC8FEA1C9249E7834ULL, "6167332_m.fxe"},
	{0xC8FEBDBDB767F4DDULL, "plt_lgt120pt_bhelen_ht.plo"},
	{0xC8FECE1F66BE8F05ULL, "vxlt_backalley_n_1dq_0_0.vlm"},
	{0xC8FED70B647A658BULL, "vxht_0_hightown_a3_bab_0_0.vlm"},
	{0xC8FF318A7BBA61A5ULL, "vxlt_gallowstemplar_n_2lj_0_0.vlm"},
	{0xC8FF41C8C33FE41EULL, "hlac_w_encounter2_-65532_0x.dds"},
	{0xC8FF6ADC2C6A464CULL, "59919658.ogg"},
	{0xC8FF77133BD2D957ULL, "958679348.ogg"},
	{0xC8FFBE25FDC27820ULL, "character_character_p14.psh.fxo"},
	{0xC9003F25ECE1FAA1ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.impa.evt"},
	{0xC900811BA6F60881ULL, "cod_hst_sundermount.utp"},
	{0xC900815BC309F313ULL, "zz_one_debug6099963.cl"},
	{0xC900AD4221C9B69FULL, "lt_0_lowtown_docks_a3_109_0.msh"},
	{0xC900B34221C9C141ULL, "lt_0_lowtown_docks_a3_109_0.mmh"},
	{0xC900BA7394415C30ULL, "dae000ip_crowd_ff3_talk4.utp"},
	{0xC900BF4221C9D5E2ULL, "lt_0_lowtown_docks_a3_109_0.mao"},
	{0xC901C960E1646972ULL, "vxlt_0_lowtown_byf_0_0.vlm"},
	{0xC901F1FF7E26F22EULL, "22278728.ogg"},
	{0xC90224D76FEFB98DULL, "llac_dr_primevaltomb_11985_3.dds"},
	{0xC90231224CE1B433ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_modal_activate.evt"},
	{0xC90295547F042208ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_tall.mmh"},
	{0xC902A1547F0436ABULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_tall.mao"},
	{0xC902B3547F045532ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp_tall.msh"},
	{0xC902EF45D6A50DC1ULL, "vxlt_undercityencounter_1bn_0_0.vlm"},
	{0xC9031E068AFC3433ULL, "vxht_0_hightown_n_b8s_0_0.vlm"},
	{0xC903603B8000EEF8ULL, "one100dg_nabil.stg"},
	{0xC9038229516622A6ULL, "cd_mag360_alain.wbd"},
	{0xC9044C0C696CE7ACULL, "zz_fen_debug6166027.cl"},
	{0xC90463A224FB947CULL, "hlac_w_bonepitmining_a2_0_3y.dds"},
	{0xC9049285556A2BCEULL, "6078988_m.fxe"},
	{0xC90526BB816D4F93ULL, "227300557.ogg"},
	{0xC905453A33159D0CULL, "hlac_lt_gallowsprison_n_-1_2y.dds"},
	{0xC905C0BDB76E180AULL, "plt_lgt120pt_bhelen_ht.nss"},
	{0xC9063C2EC74601E4ULL, "ser311cr_quentin.utc"},
	{0xC9066DB116CDAC9EULL, "mag320_bsc_banter6100759.cl"},
	{0xC906C4A544FBA157ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.cess.evt"},
	{0xC906E0D65FBBC9DBULL, "437366471.ogg"},
	{0xC9074A3E7EE33FD5ULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_spr01.msh"},
	{0xC9074C3E7EE3434BULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_spr01.mmh"},
	{0xC907583E7EE357E8ULL, "art\\characters\\weapons\\unique\\gallowsgateguardian\\w_ggg_spr01.mao"},
	{0xC907900939C6EDFEULL, "hlac_lt_0_lowtown_131072_1x.dds"},
	{0xC90818A17CD174D3ULL, "911829614.ogg"},
	{0xC9081B7F132D4A6DULL, "vxca_caverns1_mines_k6l_0_0.vlm"},
	{0xC908CEBD78588F27ULL, "503686074.ogg"},
	{0xC908E0A08686F744ULL, "6148641_m.fxe"},
	{0xC9092A224CE7C774ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_modal_activate.gad"},
	{0xC90943E7FC00A1DFULL, "and100_anders6111668.cl"},
	{0xC909DE4AD9B1061AULL, "vxht_keep_6f3_0_0.vlm"},
	{0xC90A3E01B22F3F78ULL, "vxlt_backalley_n_1pd_0_0.vlm"},
	{0xC90A63AC957E86B4ULL, "cocdownsample.psh.fxo"},
	{0xC90A6DF66F507D56ULL, "t3_wdg_w25.tnt"},
	{0xC90A8DF9E3CC8CE3ULL, "lt_0_lowtown_docks_a3_n_132_0.mao"},
	{0xC90A91095DCF4F1DULL, "gen_im_cft_reg_g2.uti"},
	{0xC90A91F9E3CC9330ULL, "lt_0_lowtown_docks_a3_n_132_0.mmh"},
	{0xC90A9FF9E3CCAB4AULL, "lt_0_lowtown_docks_a3_n_132_0.msh"},
	{0xC90B356583CD5433ULL, "vxht_keep_n_6ja_0_0.vlm"},
	{0xC90BA2D08FDE463CULL, "335860669.ogg"},
	{0xC90BAFE4F3ED8903ULL, "hlac_lt_0_lowtown_a2_131071_5z.dds"},
	{0xC90C1E166A4C9EC4ULL, "lowtown_fem_urchin_mob6165918.cl"},
	{0xC90C5B259858F905ULL, "ser191_emeric6120452.cl"},
	{0xC90C93D8C6D87BD7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_kill.gad"},
	{0xC90CE2EB41306982ULL, "6135360_m.fxe"},
	{0xC90D3725ECEC77F0ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.impa.ani"},
	{0xC90D434201DB867BULL, "197897037.ogg"},
	{0xC90D817122B0A4F9ULL, "qcr220_grey_wardens6060411.cl"},
	{0xC90D84DA5F94FB18ULL, "6020558_m.fxe"},
	{0xC90D940E69A34202ULL, "mag101_alain6032198.cl"},
	{0xC90DBEC08ACBFFBFULL, "hlac_ht_mansion2_818_16z.dds"},
	{0xC90DC5A54501C1E0ULL, "art\\vfx\\cinematic\\vfx_blood_pool_p.cess.gad"},
	{0xC90EDE5E3159C10FULL, "isa211_isabela6081849.cl"},
	{0xC90F2E066C910306ULL, "vxlt_0_lowtown_a2_n_csf_0_0.vlm"},
	{0xC90F47A4A57DCDB5ULL, "800711473.ogg"},
	{0xC90FAE7B31E09199ULL, "vxht_mansion2_3gv_0_0.vlm"},
	{0xC90FC9DB4A733E19ULL, "401622231.ogg"},
	{0xC90FDACC70AE3B91ULL, "art\\vfx\\materials\\chunky\\vfx_c_potion_tint.mao"},
	{0xC91039224CEDFF52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_quick_modal_activate.ani"},
	{0xC9107464CDE8A365ULL, "115430175.ogg"},
	{0xC91076C91B81264CULL, "ave200_donnic6061414.cl"},
	{0xC910F60844A235FCULL, "qun110_petrice6037388.cl"},
	{0xC911724EA683747FULL, "682037027.ogg"},
	{0xC91175321E6C4F64ULL, "vxlt_gallowscourtyard_n_4r1_0_0.vlm"},
	{0xC9117648BB5CB699ULL, "vxlt_gallowscourtyard_n_7ib_0_0.vlm"},
	{0xC9123F98B1373376ULL, "609900284.ogg"},
	{0xC912735980ABFB30ULL, "vxw_bonepitdestroyed_a3_1v5_0_0.vlm"},
	{0xC912738FC4CE8B56ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be02a.mmh"},
	{0xC9127D8FC4CE9C54ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_be02a.msh"},
	{0xC912EC27F0A0EFA9ULL, "dae000ip_crowd_mf8_talk10.utp"},
	{0xC91302A749053F9BULL, "6103650_m.fxe"},
	{0xC9133F6A34D855F3ULL, "gen_im_npc_arm_hum_boo_gwda.uti"},
	{0xC91340CEEBDC9D8EULL, "786971628.ogg"},
	{0xC91344678F021021ULL, "al_sebastian_talents.gda"},
	{0xC913A8D8C6DEBE64ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_kill.evt"},
	{0xC913BDFDF8DE83F9ULL, "vxht_mansion_n_3vb_0_0.vlm"},
	{0xC913E33A0C1ED1FDULL, "one140cr_charade6142168.cl"},
	{0xC914096B8D456A17ULL, "102863348.ogg"},
	{0xC9141DF259DDE467ULL, "6208311_m.fxe"},
	{0xC9144E4A83B02918ULL, "dre000_varric6207941.cl"},
	{0xC9146C2689850D50ULL, "ht_0_hightown_n_33_0_lc.mmh"},
	{0xC91488F7FAF10DB6ULL, "770881043.ogg"},
	{0xC91492BAC2F821D1ULL, "mer320_hubert6053180.cl"},
	{0xC914A0DF51065E60ULL, "hlac_ca_cavern1_haunted_24422_1y.dds"},
	{0xC9153BABBB6A6BD4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodmetal_n.dds"},
	{0xC91574751A9BFAF4ULL, "461328703.ogg"},
	{0xC915A1A9D4764D46ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.cess.ani"},
	{0xC915ABD0124A1D64ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.anb"},
	{0xC915B59E7CA4554AULL, "6096738_f.fxe"},
	{0xC915C9E420D79E06ULL, "lgt110_arl_bann_teagan6115577.cl"},
	{0xC915F46980D59CE9ULL, "gen00fl_anders6113077.cl"},
	{0xC916581F07E61722ULL, "971829240.ogg"},
	{0xC916787C63CAE63AULL, "llac_lt_0_lowtown_a3_n_c_131071_6.dds"},
	{0xC916925E0A99AC7AULL, "dae000_porfiria6162633.cl"},
	{0xC916BBB6668858E9ULL, "vxht_mansion_n_4va_0_0.vlm"},
	{0xC917466DDF9589B3ULL, "vxlt_0_lowtown_a2_eoz_0_0.vlm"},
	{0xC91773795980FBBFULL, "518002906.ogg"},
	{0xC917A6A4C064BAEEULL, "zz_lgtar_debug.ncs"},
	{0xC917E87A04AAB48EULL, "lt_0_lowtown_a2_n_water_21475.msh"},
	{0xC917F67A04AACCDFULL, "lt_0_lowtown_a2_n_water_21475.mao"},
	{0xC918027A04AAE17CULL, "lt_0_lowtown_a2_n_water_21475.mmh"},
	{0xC918216FA3916675ULL, "bodyparts.gda"},
	{0xC91866D03625F15AULL, "w_encounter3_3_0.mmh"},
	{0xC9186AD03625F791ULL, "w_encounter3_3_0.mao"},
	{0xC91878D036260FA0ULL, "w_encounter3_3_0.msh"},
	{0xC918968221B75F84ULL, "382132681.ogg"},
	{0xC91907C01C15CA56ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_keepwall_n.dds"},
	{0xC919381EFEB7F3F3ULL, "vxw_bonepit_83_0_0.vlm"},
	{0xC919A3C90D3E10BEULL, "ht_0_hightown_a3_18_0_ll.mmh"},
	{0xC919D19058E40477ULL, "hlac_lt_0_lowtown_-196609_0z.dds"},
	{0xC919E22C75922047ULL, "follower_banter6172244.cl"},
	{0xC91A8837F43D28FAULL, "ss_coterie_thug_f.wbd"},
	{0xC91B1FD9113A45E7ULL, "vxht_keep_n_4n3_0_0.vlm"},
	{0xC91B3341DC924D5AULL, "vxca_caverns1_jgp_0_0.vlm"},
	{0xC91BDB1CE85285BAULL, "hlac_ht_playershouse_n_855_32z.dds"},
	{0xC91BE0FEEFC0FEDAULL, "dae000_sandal6147796.cl"},
	{0xC91C7FBA5293FD71ULL, "hlac_lt_0_lowtown_n_65537_0y.dds"},
	{0xC91CFB2CCA6028BDULL, "gen00fl_anders6103667.cl"},
	{0xC91CFD124BBFB296ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_45_walk.ani"},
	{0xC91D1E2223717C22ULL, "6149367_m.fxe"},
	{0xC91D6339701B53EAULL, "15792857.ogg"},
	{0xC91DAB0524F0FE94ULL, "hlac_ht_0_hightown_a3_-65534_0z.dds"},
	{0xC91E892A597FB084ULL, "pro000_friedrich6051053.cl"},
	{0xC91E90546F87F856ULL, "6115787_f.fxe"},
	{0xC91F110FEDC58DC2ULL, "pro000_guard6052147.cl"},
	{0xC91F2596A9F09AE7ULL, "459241750.ogg"},
	{0xC91F8CF8FF573E5EULL, "vxw_bonepitmining_a2_29s_0_0.vlm"},
	{0xC91FAA802A8E8E0AULL, "6110123_m.fxe"},
	{0xC91FEE920D31035BULL, "llac_dr_0_deeproads_5838_1.dds"},
	{0xC92012FC5951B145ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_lightbeam_1.mao"},
	{0xC92019C32CF11190ULL, "vxca_caverns1_mines_dq7_0_0.vlm"},
	{0xC9213B9AB635A5CFULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\_textures\\bsh_wideleaf_dp.dds"},
	{0xC92147D04DEEE8AAULL, "6036695_f.fxe"},
	{0xC92149703AB8090EULL, "539333040.ogg"},
	{0xC921622E0EDD7EAFULL, "hlac_lt_0_lowtown_a2_131070_1y.dds"},
	{0xC921836CB20F8E71ULL, "zz_mcr_debug6071561.cl"},
	{0xC921C0D8C6EB245DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knee_to_groin_reaction_exit_kill.ani"},
	{0xC921C2F546D659C0ULL, "6033312_m.fxe"},
	{0xC921E814345B9271ULL, "6149276_m.fxe"},
	{0xC921EEDB8FC33CA4ULL, "ts_apostates_acc_table.gda"},
	{0xC9220F99F3081794ULL, "vxlt_ruinedhouse_n_ic_0_0.vlm"},
	{0xC922492EC5E27F94ULL, "vxht_0_hightown_a3_9zo_0_0.vlm"},
	{0xC9229B9AD32AF531ULL, "6150159_m.fxe"},
	{0xC922CAD01254DC67ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.mmh"},
	{0xC9230B2555EE2DA1ULL, "6153385_m.fxe"},
	{0xC92313359547F2A5ULL, "112241776.ogg"},
	{0xC92384F5A758CC41ULL, "6096478_f.fxe"},
	{0xC92389A9D48261B7ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.cess.evt"},
	{0xC923D3FC66F6BD7BULL, "mcr310_argument6074372.cl"},
	{0xC923EE63C274A869ULL, "vxca_caverns1_mines_do0_0_0.vlm"},
	{0xC92422BFA8D98C5EULL, "zz_followers_banter6158964.cl"},
	{0xC9242F9DFB2A2A80ULL, "gen_im_arm_cht_med_set_med.uti"},
	{0xC9247D731FED597EULL, "gen_im_acc_amu_act2_02.uti"},
	{0xC924AFFF9572ABC7ULL, "vxdr_0_deeproads_ck7_0_0.vlm"},
	{0xC924D79A3ED52A54ULL, "hlac_lt_hangedman_n_-1_5y.dds"},
	{0xC9251F550588C197ULL, "vxht_mansion2_1gn_0_0.vlm"},
	{0xC92572F828DC3439ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_qawarning.msh"},
	{0xC9257CF828DC4527ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_qawarning.mmh"},
	{0xC92580F828DC4C74ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_qawarning.mao"},
	{0xC925CA5BA11EDF38ULL, "358575371.ogg"},
	{0xC926C12A52B8FA55ULL, "6136638_m.fxe"},
	{0xC9273B06C2746E9AULL, "ser221_party_comments6163698.cl"},
	{0xC927476BA2DFE74FULL, "6148088_m.fxe"},
	{0xC927ECAE51929F6CULL, "hlac_dr_0_deeproads_17038_7x.dds"},
	{0xC9280AAAD27E8906ULL, "gen00fl_fenris6115362.cl"},
	{0xC928B860A94A6E3CULL, "gen00fl_anders6095493.cl"},
	{0xC9294F9EFA21779BULL, "zz_followers_banter6158493.cl"},
	{0xC92961DD1B25F434ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocksmall.phy"},
	{0xC929FF21DD3E925DULL, "ach_exp_caverns.dds"},
	{0xC92A5D000C36DF01ULL, "347750153.ogg"},
	{0xC92A8AA9D4888240ULL, "art\\vfx\\combat\\vfx_r_arc_bst_shot_proj_p.cess.gad"},
	{0xC92AB6533F5F5A0AULL, "hlac_lt_0_lowtown_a3_n_c_65536_0z.dds"},
	{0xC92AE87D08E92B55ULL, "vxca_caverns2_4ob_0_0.vlm"},
	{0xC92B05124BCBFD87ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_45_walk.evt"},
	{0xC92BE112C75E3B17ULL, "52874876.ogg"},
	{0xC92C33F524644320ULL, "184806968.ogg"},
	{0xC92C4AF36ADC43ECULL, "136373849.ogg"},
	{0xC92C9D41F6C99A9CULL, "vxlt_0_lowtown_docks_n_cyg_0_0.vlm"},
	{0xC92CE8307509438EULL, "ht_mansion2.arl"},
	{0xC92DBE8C31010124ULL, "301670829.ogg"},
	{0xC92E2603E6E6B72AULL, "gen_im_trs_krk_hgh_junk2_03.uti"},
	{0xC92F0ECCA620B834ULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.mmh"},
	{0xC92F80E964C9F6C7ULL, "vxht_chantry_2c5_0_0.vlm"},
	{0xC93022B04BA1C9D0ULL, "19341675.ogg"},
	{0xC93059F12651AAF1ULL, "6142376_m.fxe"},
	{0xC9307365733FF641ULL, "6179518_m.fxe"},
	{0xC9307A1C079F161DULL, "296300326.ogg"},
	{0xC930B50F1485E989ULL, "vxlt_0_lowtown_docks_n_bfb_0_0.vlm"},
	{0xC930CE1EA59CC533ULL, "mag291_vincento6027666.cl"},
	{0xC930F1667892877FULL, "6143157_m.fxe"},
	{0xC9310AAD4C1671DDULL, "6059010_m.fxe"},
	{0xC93179E47F27CD4DULL, "vxw_woundedcoast_1f3_0_0.vlm"},
	{0xC9317F81A4F41E62ULL, "6076029_m.fxe"},
	{0xC931F5B75A43FC10ULL, "hlac_lt_0_lowtown_docks_a3_-131075_7x.dds"},
	{0xC9320FD337E99436ULL, "vxlt_hangedman_2v1_0_0.vlm"},
	{0xC93212D9E04A8DF0ULL, "vxdr_primevaltomb_7ju_0_0.vlm"},
	{0xC93226124BD254D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_45_walk.gad"},
	{0xC9322B6ECE79C5E1ULL, "sse_ss_castillon.gda"},
	{0xC9333EDD1B2E0D31ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocksmall.msh"},
	{0xC9334CDD1B2E254CULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocksmall.mao"},
	{0xC933531887161E90ULL, "hlac_lt_undercityencounter_-1_0x.dds"},
	{0xC93358DD1B2E39EFULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrocksmall.mmh"},
	{0xC9337C7D393124B6ULL, "vxdr_primevalentrance_23i_0_0.vlm"},
	{0xC93388E266AD9547ULL, "6174872_m.fxe"},
	{0xC9340C78BB3CAEEDULL, "806589487.ogg"},
	{0xC934B72A3CF666CCULL, "6170633_m.fxe"},
	{0xC935AE038D834BFAULL, "w_bonepitmining_a2_34_0.rml"},
	{0xC935B48C91962FA8ULL, "6084732_m.fxe"},
	{0xC935F291CD7AB673ULL, "289452476.ogg"},
	{0xC936241B3E6532A8ULL, "598620114.ogg"},
	{0xC936583B802C3ACBULL, "one100dg_nabil.bnk"},
	{0xC9369780B1584145ULL, "225397023.ogg"},
	{0xC93815D4320096C8ULL, "6165635_m.fxe"},
	{0xC9384A0771A141B9ULL, "combat_ws_r1_combo.gda"},
	{0xC9385123AB4E89D3ULL, "6142631_m.fxe"},
	{0xC938549B090B1473ULL, "1030682447.ogg"},
	{0xC93856AB7424F804ULL, "6054374_f.fxe"},
	{0xC9387DE12931F202ULL, "737815063.ogg"},
	{0xC938D9311331F4BAULL, "102056349.ogg"},
	{0xC9395A3B802E73ABULL, "one100dg_nabil.cnv"},
	{0xC9397734887559FDULL, "6106725_m.fxe"},
	{0xC939BC7ACB099BFBULL, "lt_0_lowtown_32_0.anb"},
	{0xC93A450A64ECB9BAULL, "mag370_meredith6038458.cl"},
	{0xC93A6563A44C9D47ULL, "482622061.ogg"},
	{0xC93A73C3E81B8B42ULL, "6164469_m.fxe"},
	{0xC93A88233DA08F42ULL, "qcr000_arishok6060161.cl"},
	{0xC93ACAD2783C43CAULL, "follower_banter6171044.cl"},
	{0xC93AF92C1A113827ULL, "vxlt_0_lowtown_a2_n_c_eqh_0_1.vlm"},
	{0xC93B0F1EE8BFB31CULL, "vxlt_0_lowtown_docks_n_ctd_0_0.vlm"},
	{0xC93C239341C5CB02ULL, "572706018.ogg"},
	{0xC93D216EDBB027EBULL, "hlac_dr_primevaltomb_657_2y.dds"},
	{0xC93D35B4D5DEA205ULL, "6085011_m.fxe"},
	{0xC93D56A485037193ULL, "vxht_brothel_n_5el_0_0.vlm"},
	{0xC93DAF5F97B03F0BULL, "vxca_cavern1_haunted_im5_0_0.vlm"},
	{0xC93DBEC2655EADDAULL, "vxw_encounter2_camp_os_0_0.vlm"},
	{0xC93DC8F0568E44B0ULL, "vxw_bonepitdestroyed_a3_1a0_0_0.vlm"},
	{0xC93DF3AA1B833675ULL, "vxht_0_hightown_9bc_0_0.vlm"},
	{0xC93DF6E450241AEAULL, "t3_nbl_c03.tnt"},
	{0xC93E4D92D3CE24D2ULL, "sky_sky2_p1.vsh.fxo.dx11.assembly"},
	{0xC93E808784914E06ULL, "hlac_ca_cavern1_haunted_24422_6z.dds"},
	{0xC93F1C10888308C5ULL, "gen_im_trs_tmp_med_cht_r1.uti"},
	{0xC93F26EC067C0209ULL, "vxlt_0_lowtown_docks_n_bca_0_0.vlm"},
	{0xC93F282AAD3CCE5CULL, "931019519.ogg"},
	{0xC93F72AA9541F1B7ULL, "rdr111dg_assistant6016636.cl"},
	{0xC93F90801285262AULL, "vxw_bonepitdestroyed_a3_8n_0_0.vlm"},
	{0xC93F91703B7407FCULL, "vxht_keep_4av_0_0.vlm"},
	{0xC93FB2AC2B9AC848ULL, "plt_cod_prc_invisible_nude.plo"},
	{0xC93FD539C6CAEC30ULL, "778592802.ogg"},
	{0xC93FEBC1711DE5CDULL, "vxw_woundedcoast_1a2_0_0.vlm"},
	{0xC940801170458755ULL, "715081906.ogg"},
	{0xC940BA2B698758DDULL, "6095912_f.fxe"},
	{0xC9410598C07DA312ULL, "6082978_m.fxe"},
	{0xC9413ADB3AB1F51DULL, "1006749963.ogg"},
	{0xC941B14F98F246BAULL, "vxht_bartrand_n_5pk_0_1.vlm"},
	{0xC941CBA89091511AULL, "6217963_m.fxe"},
	{0xC941EE5A2B273D36ULL, "vxdr_primevalentrance_26j_0_0.vlm"},
	{0xC94366D8F7C75253ULL, "vxht_0_hightown_n_ai7_0_0.vlm"},
	{0xC9438FD3EF9129FCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_broken.mao"},
	{0xC9439A23BEA962D9ULL, "ser270_emeric6112664.cl"},
	{0xC9439BD3EF913D9FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_broken.mmh"},
	{0xC943A1D3EF914841ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_broken.msh"},
	{0xC943BB327A18E9C7ULL, "hlac_lt_0_lowtown_a2_-196608_2y.dds"},
	{0xC943BEE719D86365ULL, "one101cr_smuggler.utc"},
	{0xC943C4F067981951ULL, "146881508.ogg"},
	{0xC943E49578C100D6ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medp.phy"},
	{0xC94402A33E624904ULL, "vxw_encounter2_camp_3i_0_0.vlm"},
	{0xC94403C7B2EFDCA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_exit.gad"},
	{0xC9446138A4A1C85DULL, "hlac_ca_caverns2_5802_42x.dds"},
	{0xC9447D4091CA9AE5ULL, "follower_banter6154219.cl"},
	{0xC944872D56E8E06AULL, "hf_qmq_minion8.mor"},
	{0xC9448935C3F80EE4ULL, "193306558.ogg"},
	{0xC9448A38D6E8D335ULL, "s_ballista_02.mao"},
	{0xC9448A5EE96CE915ULL, "6180846_m.fxe"},
	{0xC94566D0364CDF71ULL, "w_encounter3_3_0.rml"},
	{0xC945ADD7216DFC5CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_side_up.evt"},
	{0xC94655450C7C5902ULL, "6098278_m.fxe"},
	{0xC9470EE4113B97B7ULL, "pro000ip_corpse_1.utp"},
	{0xC9472763DE3CE43EULL, "6092089_m.fxe"},
	{0xC948108DCC531ACFULL, "vxdr_primevaltomb_8p7_0_0.vlm"},
	{0xC9496715DCA713D8ULL, "139083585.ogg"},
	{0xC94980FBDAB5CB9CULL, "vxlt_0_lowtown_docks_n_coe_0_0.vlm"},
	{0xC94993934879DC36ULL, "ht_0_hightown_n_17_0_lc.mmh"},
	{0xC949CB39C3C0F281ULL, "hlac_lt_0_lowtown_a2_65534_2y.dds"},
	{0xC949ECE8B61CBF3DULL, "268711723.ogg"},
	{0xC949F230D3F7186AULL, "hlac_ht_0_hightown_n_65533_3y.dds"},
	{0xC949FED24E71D545ULL, "ht_0_hightown_n_53_0_ll.mmh"},
	{0xC94A205343EDF6AEULL, "ave260ip_signal_fire3.utp"},
	{0xC94A6644E4BE026EULL, "hlac_lt_warehouse_65537_3x.dds"},
	{0xC94A7092BA0C1384ULL, "73050588.ogg"},
	{0xC94A7CDABA898A2EULL, "6024109_m.fxe"},
	{0xC94A98F9E402834BULL, "lt_0_lowtown_docks_a3_n_132_0.phy"},
	{0xC94ABFF4E7893960ULL, "664605621.ogg"},
	{0xC94AF8C7B2F5E8B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_exit.evt"},
	{0xC94B05A23243F966ULL, "782061752.ogg"},
	{0xC94B38F3E0A2FA11ULL, "mag200_feynriel6023985.cl"},
	{0xC94BAC40D749A1B1ULL, "hlac_lt_0_lowtown_a3_n_c_-2_3x.dds"},
	{0xC94BBB776B777F3CULL, "4281720.ogg"},
	{0xC94BC88176F98A13ULL, "vxht_brothel_n_5ho_0_0.vlm"},
	{0xC94C1D54A01B8566ULL, "802206259.ogg"},
	{0xC94C213C89A6578BULL, "vxca_cavern1_haunted_ij4_0_0.vlm"},
	{0xC94C309F5754C65AULL, "vxw_encounter2_camp_jl_0_0.vlm"},
	{0xC94C3ACD48845D30ULL, "vxw_bonepitdestroyed_a3_1j1_0_0.vlm"},
	{0xC94C5F5680AAADABULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.dura.ani"},
	{0xC94C633E38E2DDCEULL, "mcr370_assignments6081710.cl"},
	{0xC94C64A348B627C7ULL, "vxht_chantry_2y7_0_0.vlm"},
	{0xC94CF6C273EFE552ULL, "vxdr_primevalentrance_19x_0_0.vlm"},
	{0xC94D03CAC52EE6EEULL, "vxht_mansion2_238_0_0.vlm"},
	{0xC94D1036715D6EAEULL, "mcr310_party_banter6087814.cl"},
	{0xC94D7E2D9AA637C9ULL, "vxdr_primevaltomb_6ql_0_0.vlm"},
	{0xC94D9675094CFA52ULL, "28208026.ogg"},
	{0xC94D9B40C8B31968ULL, "6179872_m.fxe"},
	{0xC94DD0F05A4CE520ULL, "zz_qcrpt_debug.ncs"},
	{0xC94DFA8E94AA1B32ULL, "105995803.ogg"},
	{0xC94E11E271A92D74ULL, "gen_im_trs_rdr_med_cht_r1.uti"},
	{0xC94E78EE20221CFCULL, "dae000_orsino6085448.cl"},
	{0xC94ECD9578CAE177ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medp.msh"},
	{0xC94EDF9578CAFFDAULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medp.mao"},
	{0xC94EE39578CB06E9ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medp.mmh"},
	{0xC94F7C6771B8F4D1ULL, "hlac_ht_bartrand_n_2995_0y.dds"},
	{0xC94F9C23565206D7ULL, "6119257_m.fxe"},
	{0xC94FD2101A803B70ULL, "590492166.ogg"},
	{0xC95004A29924DA2DULL, "vxlt_foundry_n_wh_0_0.vlm"},
	{0xC9501E9F9EDC8DDEULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_pota_1.mao"},
	{0xC9503C96F85FCD66ULL, "vxlt_hangedman_27n_0_0.vlm"},
	{0xC9512750FDB74B0DULL, "dae151ip_at_sundermount_a.utp"},
	{0xC951670AFE2BB5EBULL, "dae150_chandan6148778.cl"},
	{0xC95186C51DC58303ULL, "t3_dey_gy3.tnt"},
	{0xC951997BFF764C1DULL, "6109715_m.fxe"},
	{0xC95235FEF00F3BB4ULL, "mag160_feynriel6016569.cl"},
	{0xC952678AC8217880ULL, "6092602_f.fxe"},
	{0xC952A782C7F838E9ULL, "hlac_lt_0_lowtown_docks_a3_n_65536_0z.dds"},
	{0xC952AA7215A2AB0EULL, "vxht_keep_3ic_0_0.vlm"},
	{0xC95339EA96B617F7ULL, "vxca_cavern1_haunted_j25_0_2.vlm"},
	{0xC9533DE080C8D913ULL, "llac_lt_0_lowtown_-65538_0.dds"},
	{0xC9535E84509870C6ULL, "one100dg_kelder6173953.cl"},
	{0xC95360E9BAB76A55ULL, "289110769.ogg"},
	{0xC9544160EDA40D9FULL, "6005404_m.fxe"},
	{0xC954B5BD89B493D9ULL, "6125224_f.fxe"},
	{0xC954EA533EA5F712ULL, "mag121_idunna6084898.cl"},
	{0xC9554DB5D9A9A39DULL, "303074613.ogg"},
	{0xC9561355E181A610ULL, "144431232.ogg"},
	{0xC9561A16D0A73D40ULL, "hlac_ht_0_hightown_a3_65535_3y.dds"},
	{0xC9562A85C1F3CEA8ULL, "vxht_0_hightown_9x8_0_0.vlm"},
	{0xC956A20D0FD17690ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_hed_snga_0t.dds"},
	{0xC957118E22CF44FDULL, "nat380_sibling6067979.cl"},
	{0xC9571F4142907153ULL, "gen00fl_merrill6121137.cl"},
	{0xC95786CC321CD415ULL, "1013370899.ogg"},
	{0xC9578C8F6F80749DULL, "lgt311_shw_mer_mettin6156487.cl"},
	{0xC95811CCA643C2EBULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.anb"},
	{0xC95837B76FABF6F7ULL, "356919101.ogg"},
	{0xC95862CE049B4801ULL, "lt_0_lowtown_docks_123_0.phy"},
	{0xC958C53780B9C538ULL, "vxw_woundedcoast_15g_0_0.vlm"},
	{0xC958D3D2A7EF4443ULL, "451718997.ogg"},
	{0xC958D51618F0ECD9ULL, "mag291_vincento6027659.cl"},
	{0xC958F0C7B30218EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_right_exit.ani"},
	{0xC959367766D1333FULL, "6176481_m.fxe"},
	{0xC9593CE28940480FULL, "vxht_keep_n_610_0_0.vlm"},
	{0xC95970BEFD807AAAULL, "mag101_grace6034032.cl"},
	{0xC959966D81130499ULL, "hlac_lt_0_lowtown_a3_n_c_65536_11y.dds"},
	{0xC95A03B1B64E6211ULL, "421333395.ogg"},
	{0xC95A475680B6C222ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.dura.evt"},
	{0xC95A68A4E0C77535ULL, "6108311_f.fxe"},
	{0xC95AA27C494ADEDAULL, "vxw_encounter2_camp_em_0_0.vlm"},
	{0xC95AD6803AAC4047ULL, "vxht_chantry_2r6_0_0.vlm"},
	{0xC95B379A422E92F3ULL, "crt_spblm_med.mao"},
	{0xC95B37E2693F0CA2ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_0y.dds"},
	{0xC95B3FA9EB71C1E2ULL, "vxlt_0_lowtown_a2_d5t_0_0.vlm"},
	{0xC95BD8A7E3B38AAAULL, "vxw_sundermount_375_0_0.vlm"},
	{0xC95C0AA5EA683309ULL, "vxlt_0_lowtown_docks_n_bug_0_0.vlm"},
	{0xC95C5138D6FD0413ULL, "s_ballista_02.dds"},
	{0xC95C52530194229AULL, "vxht_0_hightown_n_9bc_0_0.vlm"},
	{0xC95CE911DA6112B3ULL, "6024522_m.fxe"},
	{0xC95D2658626992F1ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge8.ani"},
	{0xC95D43BBAC0C12ECULL, "266224424.ogg"},
	{0xC95E170E30B7C443ULL, "6138725_f.fxe"},
	{0xC95E8F76779FF5B6ULL, "llac_lt_gallowscourtyard_-2_3.dds"},
	{0xC95E94BEB6B5C572ULL, "vxlt_undercityencounter_s9_0_0.vlm"},
	{0xC95EC54B9C4023FCULL, "6091392_m.fxe"},
	{0xC95EEBEA33D07504ULL, "6179912_f.fxe"},
	{0xC95F8C0AFAFA3D76ULL, "mag110_samson6049730.cl"},
	{0xC95FE32333551792ULL, "532017112.ogg"},
	{0xC960733477C24309ULL, "dae100_bran6106041.cl"},
	{0xC96076B9F22884EEULL, "vxlt_shop_gc_0_0.vlm"},
	{0xC9609611206A0BB7ULL, "vxht_brothel_n_60p_0_0.vlm"},
	{0xC960CB3A8168790CULL, "character_character_p25.psh.fxo.dx11"},
	{0xC960F9ABDD818BB2ULL, "zz_followers_banter6158331.cl"},
	{0xC961465680BCDF35ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.dura.gad"},
	{0xC9616E6CA1A236A1ULL, "vxlt_0_lowtown_a2_hew_0_0.vlm"},
	{0xC961A02BCB433105ULL, "gen00fl_varric6033661.cl"},
	{0xC961E59B1EAB5995ULL, "art\\vfx\\creatures\\fxc_rage_slam_c.mmh"},
	{0xC961FB5344024993ULL, "ave260ip_signal_fire3.ncs"},
	{0xC9622674885B956AULL, "6166711_m.fxe"},
	{0xC962498AC6F88735ULL, "ser191_emeric6104753.cl"},
	{0xC962C7F15AA0FA82ULL, "zz_and_debug6134178.cl"},
	{0xC96320F9FA5C70C3ULL, "hlac_ca_caverns1_24421_9z.dds"},
	{0xC9636448F6462A02ULL, "6140315_m.fxe"},
	{0xC963BC7A7A08B2DEULL, "hlac_w_encounter2_196612_0y.dds"},
	{0xC963E0A5A4401A96ULL, "67412756.ogg"},
	{0xC9643158626FC393ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge8.gad"},
	{0xC9661E6432AA3281ULL, "6109210_m.fxe"},
	{0xC96659B5181DC007ULL, "vxlt_0_lowtown_ewz_0_0.vlm"},
	{0xC96668287A59E6E3ULL, "15969694.ogg"},
	{0xC9666B11E4CD54F3ULL, "6101137_m.fxe"},
	{0xC96683AA283D2675ULL, "zz_mage1.cut"},
	{0xC96688C7E6C5D76FULL, "mage_rockarmor.dds"},
	{0xC966FC74A9A95AB0ULL, "vxdr_primevaltomb_8f2_0_0.vlm"},
	{0xC96700E3593921CEULL, "gen00fl_fenris6106692.cl"},
	{0xC967371472AFDDB8ULL, "vxw_woundedcoast_10d_0_0.vlm"},
	{0xC967BAA3AA32AF69ULL, "t1_skn_h09.tnt"},
	{0xC967C41370B2DCA6ULL, "516685418.ogg"},
	{0xC9688031A9FE1111ULL, "255121216.ogg"},
	{0xC968BD013002D957ULL, "w_bonepit_175_0.msh"},
	{0xC968C62A829F6E66ULL, "vxca_caverns1_mines_fn2_0_0.vlm"},
	{0xC968CF013002F7BAULL, "w_bonepit_175_0.mao"},
	{0xC968D3013002FEC9ULL, "w_bonepit_175_0.mmh"},
	{0xC969024E7A837CF4ULL, "hlac_w_woundedcoast_262146_0z.dds"},
	{0xC9690763B3CBC78CULL, "133781621.ogg"},
	{0xC9695BA68D73770FULL, "6069864_m.fxe"},
	{0xC969694060804462ULL, "vxht_mansion2_20y_0_0.vlm"},
	{0xC9699BE7D1D7440BULL, "6165197_m.fxe"},
	{0xC96A4A84D5A9A32AULL, "vxw_sundermount_326_0_0.vlm"},
	{0xC96A621753D62BF1ULL, "art\\characters\\creatures\\pridedemon\\animation\\die.ani"},
	{0xC96A69C8E5A2BA8DULL, "6106893_m.fxe"},
	{0xC96BB0E780AEE23EULL, "60497099.ogg"},
	{0xC96BC53B8245BE5AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratetarp_n.dds"},
	{0xC96C2C4471C74201ULL, "llac_ca_caverns1_24425_9.dds"},
	{0xC96C8261536DA316ULL, "vxlt_undercityserialkiller_1oj_0_0.vlm"},
	{0xC96D0864512CEAE6ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic_3.mao"},
	{0xC96D34B6D73648A7ULL, "669530125.ogg"},
	{0xC96D3C054D5013F7ULL, "277868786.ogg"},
	{0xC96D8E45A4F3D95AULL, "hlac_w_sundermount_-2_0z.dds"},
	{0xC96DB4ACEDF1D9DEULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2.evt"},
	{0xC96DE4A02F3B37D3ULL, "ts_chantry_prog_junk3.gda"},
	{0xC96E110E449A2425ULL, "fxe_water_color.dds"},
	{0xC96E6D8EB0D65AF4ULL, "llac_lt_darktown_-1_19.dds"},
	{0xC96EA6F39CE4F1ECULL, "follower_banter6173361.cl"},
	{0xC96EE3CDB5A7E702ULL, "mrl250_pol6101796.cl"},
	{0xC96F60F66C85B802ULL, "1025774491.ogg"},
	{0xC96F6BE64254246CULL, "953831261.ogg"},
	{0xC96FE04993984F21ULL, "vxlt_0_lowtown_a2_hbv_0_0.vlm"},
	{0xC96FE137D3540709ULL, "zz_followers_banter6158015.cl"},
	{0xC96FE29B1EB79222ULL, "art\\vfx\\creatures\\fxc_rage_slam_c.anb"},
	{0xC97059EA4316DA77ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_ogre_ogremage.gad"},
	{0xC970BB6DEFBA63E9ULL, "follower_banter6136668.cl"},
	{0xC9716D1753DC5C93ULL, "art\\characters\\creatures\\pridedemon\\animation\\die.gad"},
	{0xC971D3877898773AULL, "hlac_lt_undercitydungeon_-1_45x.dds"},
	{0xC972505AE9BDF8DDULL, "hlac_lt_blackemporium_1055_24y.dds"},
	{0xC9725DEC4E092F71ULL, "cod_shw_mer_notes_0.utp"},
	{0xC9731A87AD7EAA2AULL, "990103811.ogg"},
	{0xC9737751E02CAF46ULL, "712366496.ogg"},
	{0xC973D6BF5019873AULL, "6035169_m.fxe"},
	{0xC974C906ED3160C4ULL, "vxca_caverns1_mines_j4n_0_0.vlm"},
	{0xC974CB920A13D887ULL, "vxlt_0_lowtown_eru_0_0.vlm"},
	{0xC974CBACEDF81F09ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2.gad"},
	{0xC975A210E70606FFULL, "vxht_keep_n_60w_0_0.vlm"},
	{0xC9764DB06092CBEBULL, "6118333_m.fxe"},
	{0xC977B74D50A23E64ULL, "vxht_mansion2_1pt_0_0.vlm"},
	{0xC977F686BFA2445EULL, "vxdr_0_deeproads_bb0_0_0.vlm"},
	{0xC97820ECE4B529BBULL, "vxlt_0_lowtown_docks_n_9zs_0_0.vlm"},
	{0xC97856D6A8C54D21ULL, "923236288.ogg"},
	{0xC9785822EE43670BULL, "dae000cr_grand_cleric6148151.cl"},
	{0xC9786A1753E276F8ULL, "art\\characters\\creatures\\pridedemon\\animation\\die.evt"},
	{0xC978AC586B41B9C2ULL, "art\\vfx\\combat\\vfx_d_teleport_com_p.mmh"},
	{0xC978C0D2AEC3BF4FULL, "vxht_0_hightown_n_c5y_0_0.vlm"},
	{0xC979136A797A5C1EULL, "vxw_encounter2_camp_r0_0_0.vlm"},
	{0xC9792AAA07919665ULL, "art\\levels\\races\\proxy\\props\\_textures\\fence_sa.dds"},
	{0xC9799936CFC6D8CBULL, "6018618_f.fxe"},
	{0xC979B475390345EDULL, "mag110_arianni6206021.cl"},
	{0xC979C7B54464FCB2ULL, "dre000_varric6208043.cl"},
	{0xC97B620B9C896D39ULL, "6174984_m.fxe"},
	{0xC97B9CACEDFDEE57ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv2.ani"},
	{0xC97BA45DE5A7CF90ULL, "vxlt_gallowsprison_n_64o_0_0.vlm"},
	{0xC97BE95B66B71425ULL, "vxlt_0_lowtown_n_eok_0_0.vlm"},
	{0xC97C5AB2ED649375ULL, "6092550_m.fxe"},
	{0xC97C756EAF2899B4ULL, "hlac_lt_0_lowtown_a3_n_c_0_9z.dds"},
	{0xC97C7863B7006C10ULL, "mag170_thrask6054028.cl"},
	{0xC97CC53712B4DF5BULL, "101974612.ogg"},
	{0xC97D35EE65097E31ULL, "vxlt_hangedman_n_1vj_0_0.vlm"},
	{0xC97D8E038DC0CAB1ULL, "w_bonepitmining_a2_34_0.mmh"},
	{0xC97D926EC02ECC51ULL, "6092395_f.fxe"},
	{0xC97D9A038DC0DF52ULL, "w_bonepitmining_a2_34_0.mao"},
	{0xC97DA8038DC0F76FULL, "w_bonepitmining_a2_34_0.msh"},
	{0xC97DD9F0E5730C2CULL, "dae000_flemeth6000899.cl"},
	{0xC97E057316DDF50BULL, "the_lion_of_orlais.dds"},
	{0xC97E34BE950AB91BULL, "946276403.ogg"},
	{0xC97E753D871BD53DULL, "vxht_0_hightown_n_b1b_0_0.vlm"},
	{0xC97E843C3117717FULL, "hlac_lt_backalley_n_0_19z.dds"},
	{0xC97EA4E150C8F844ULL, "art\\vfx\\combat\\vfx_d_cast_aoe_c.impa.ani"},
	{0xC97ECD79A5DDD1C7ULL, "6068356_f.fxe"},
	{0xC97F203A8C20FE96ULL, "art\\characters\\playercharacter\\humanmale\\cth_cara.mao"},
	{0xC97FFF8F9703F3DBULL, "rdr110dg_guard_warehouse6016678.cl"},
	{0xC98006E1E10B6F2CULL, "6029391_m.fxe"},
	{0xC9804B5DB91B1951ULL, "vxw_encounter2_7d_0_0.vlm"},
	{0xC9805D017B4072BFULL, "vxht_0_hightown_ca3_0_1.vlm"},
	{0xC9810F6C70944CA6ULL, "867383740.ogg"},
	{0xC98117D90084632FULL, "6098965_m.fxe"},
	{0xC981279072ABD565ULL, "vxw_sundermount_30f_0_1.vlm"},
	{0xC9813AB327C08656ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_003a_s.dds"},
	{0xC981B1964EE5647BULL, "vxlt_foundry_n_xs_0_0.vlm"},
	{0xC982048FCF5F4ADEULL, "art\\vfx\\combat\\vfx_r_ibl_fore_act_c.impa.ani"},
	{0xC98217D5A084A581ULL, "vxht_0_hightown_c6y_0_0.vlm"},
	{0xC9826EB9B5DAD74AULL, "vxlt_house_gk_0_0.vlm"},
	{0xC982BB22D38DCAA3ULL, "dre000_vampire6034343.cl"},
	{0xC982EAD72C6C754CULL, "w_encounter2_16_0.phy"},
	{0xC982FCCD4BB20BB4ULL, "vxdr_primevaltomb_3i4_0_0.vlm"},
	{0xC98311120937AB9FULL, "vxlt_blackemporium_3q_0_0.vlm"},
	{0xC9831259A2814AFEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mede_0t.dds"},
	{0xC9833AE3DF277944ULL, "vxca_caverns1_mines_j3o_0_0.vlm"},
	{0xC9833D6EFC09F107ULL, "vxlt_0_lowtown_emt_0_0.vlm"},
	{0xC983461B286D355EULL, "llac_ht_mansion2_818_51.dds"},
	{0xC9836BBF016BBA78ULL, "hlac_lt_0_lowtown_docks_65534_1x.dds"},
	{0xC9845A00AE942322ULL, "vxht_store_cb_0_0.vlm"},
	{0xC98486EA4327CFFDULL, "art\\cinematics\\deathblows\\ogre\\cs_db_ogre_ogremage.ani"},
	{0xC984DB5BD5442B6BULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar2_mcr370_playerrules.ani"},
	{0xC9856FAEB7A631E5ULL, "vxw_encounter2_camp_pu_0_0.vlm"},
	{0xC9859C2D46025055ULL, "102602073.ogg"},
	{0xC985A1E150CF11A6ULL, "art\\vfx\\combat\\vfx_d_cast_aoe_c.impa.gad"},
	{0xC985A4A9404C0C44ULL, "6076427_m.fxe"},
	{0xC9869A18A737EA07ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_interact_low.gad"},
	{0xC986C6847B23DAE1ULL, "zz_followers_banter6159261.cl"},
	{0xC986FF0A1BED08F3ULL, "hla_ht_0_hightown_n_65535_0z.dds"},
	{0xC98726532464594FULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkf.gad"},
	{0xC98732AFA0B9D7CFULL, "vxht_0_hightown_n_c2x_0_0.vlm"},
	{0xC9873BE89D5F96B0ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1b.mao"},
	{0xC9873FE89D5F9DC3ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1b.mmh"},
	{0xC9874DE89D5FB51DULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\tre_roots1b.msh"},
	{0xC987638FF6D73934ULL, "hlac_lt_0_lowtown_a2_n_c_-3_4z.dds"},
	{0xC98785476B70749EULL, "vxw_encounter2_camp_m1_0_0.vlm"},
	{0xC98811714B442DFEULL, "gen00fl_merrill6121761.cl"},
	{0xC98821AA13B34701ULL, "vxw_woundedcoast_h9_0_1.vlm"},
	{0xC988D5E63FEB8C26ULL, "435600109.ogg"},
	{0xC988EA6D3AB2C01EULL, "6088693_m.fxe"},
	{0xC989C1D72C724E54ULL, "w_encounter2_16_0.rml"},
	{0xC989EAA052C0C13AULL, "vxht_bartrand_n_5of_0_1.vlm"},
	{0xC98A14F28E18D6E8ULL, "645535268.ogg"},
	{0xC98A6FFFDAFA5770ULL, "372216023.ogg"},
	{0xC98A707F4800D4DAULL, "and111st_lirene.stg"},
	{0xC98AFCA48CCB4B05ULL, "dae300_cullen6140579.cl"},
	{0xC98B17CD7F938CC6ULL, "vxlt_0_lowtown_n_epd_0_1.vlm"},
	{0xC98BC95EE004B01CULL, "vxlt_0_lowtown_a2_n_gmv_0_0.vlm"},
	{0xC98C26B147ABE1C6ULL, "vxw_bonepitdestroyed_a3_bp_0_0.vlm"},
	{0xC98C9CE150D52785ULL, "art\\vfx\\combat\\vfx_d_cast_aoe_c.impa.evt"},
	{0xC98CE9869B8D52D4ULL, "aip_undeadskelaracher_base.gda"},
	{0xC98CF99CE6575DADULL, "6037696_m.fxe"},
	{0xC98D8F18A73DF614ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_interact_low.evt"},
	{0xC98DBE762CE289A6ULL, "6153621_m.fxe"},
	{0xC98E3B53246A9BFCULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkf.evt"},
	{0xC98E70E51E42E0E3ULL, "w_bonepit_112_0.mao"},
	{0xC98E74E51E42E730ULL, "w_bonepit_112_0.mmh"},
	{0xC98E82E51E42FF4AULL, "w_bonepit_112_0.msh"},
	{0xC98E98DE186DAF3CULL, "523867454.ogg"},
	{0xC98F1C8FCF69FDAFULL, "art\\vfx\\combat\\vfx_r_ibl_fore_act_c.impa.evt"},
	{0xC98F3BCBDFC6D546ULL, "6069818_m.fxe"},
	{0xC98F8757A2824F0BULL, "vxdr_primevaltomb_2uo_0_1.vlm"},
	{0xC990237340DB7CFBULL, "vxlt_foundry_n_sr_0_0.vlm"},
	{0xC99030B50AB0CF13ULL, "vxdr_0_deeproads_cph_0_2.vlm"},
	{0xC9909224B287EAD8ULL, "6105880_m.fxe"},
	{0xC990A35D5DB3E638ULL, "vxca_caverns1_fm4_0_0.vlm"},
	{0xC990D05EAA6DB886ULL, "vxca_caverns1_fmz_0_0.vlm"},
	{0xC990D9760230FF1CULL, "502506648.ogg"},
	{0xC990F1A124D637C3ULL, "531312776.ogg"},
	{0xC991447A11B9C0ADULL, "zz_mag_debug6054587.cl"},
	{0xC991AF4BEE000987ULL, "vxlt_0_lowtown_ehw_0_0.vlm"},
	{0xC991B916F53D6EFBULL, "898506625.ogg"},
	{0xC99219FEB9FCBC5EULL, "541760171.ogg"},
	{0xC99267CFED64BC67ULL, "dm_amb_commoner1.mor"},
	{0xC99271E79C767F0DULL, "374705070.ogg"},
	{0xC992C761F7963AF0ULL, "llac_lt_gallowstemplar_n_2394_12.dds"},
	{0xC993316A447EFCEBULL, "6112133_m.fxe"},
	{0xC993B3EB7EA677F4ULL, "225834709.ogg"},
	{0xC993E18BA99C4A65ULL, "vxw_encounter2_camp_kt_0_0.vlm"},
	{0xC99423F92AE257E9ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_3z.dds"},
	{0xC9942F5991C734B6ULL, "ser221_quentin6123332.cl"},
	{0xC9944B82875EF364ULL, "526581953.ogg"},
	{0xC9945D21F43DC31AULL, "6071154_m.fxe"},
	{0xC994E16445CA5CCEULL, "6162297_f.fxe"},
	{0xC9959988F095F37CULL, "247773118.ogg"},
	{0xC995F7245D668D1EULL, "vxw_encounter2_camp_h2_0_0.vlm"},
	{0xC9961D8FCF701E58ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_act_c.impa.gad"},
	{0xC996B90A7C75B92EULL, "vxca_caverns1_dni_0_0.vlm"},
	{0xC996C27B3E17AB0AULL, "6110372_m.fxe"},
	{0xC996CA91E30C20EEULL, "marskman_speed.dds"},
	{0xC99757967E7BE758ULL, "6173799_m.fxe"},
	{0xC99782671EF7F961ULL, "285336730.ogg"},
	{0xC9979709B50A54ABULL, "398008357.ogg"},
	{0xC997CCF4AFDDEF9FULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.impa.gad"},
	{0xC998C25BD554A9F5ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar2_mcr370_playerrules.gad"},
	{0xC9992EB0FC89C53EULL, "dae000_bartrand6027589.cl"},
	{0xC9997A002E65AE79ULL, "dae000_worthy6166590.cl"},
	{0xC9998D8F9097A36AULL, "552444653.ogg"},
	{0xC99995A9C6C132DEULL, "946103722.ogg"},
	{0xC99A2363E2145019ULL, "dae000cr_grand_cleric6178459.cl"},
	{0xC99A4AF877CB52B9ULL, "789299645.ogg"},
	{0xC99A988E39A1FA46ULL, "vxw_bonepitdestroyed_a3_mq_0_0.vlm"},
	{0xC99B435324753315ULL, "art\\characters\\creatures\\ragedemon\\animation\\cwkf.ani"},
	{0xC99B8718A74A264DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_interact_low.ani"},
	{0xC99BB97A16A1BA26ULL, "6112882_m.fxe"},
	{0xC99C0EEAE2D91285ULL, "dae000_meredith6085346.cl"},
	{0xC99C30531ED8A226ULL, "6156521_m.fxe"},
	{0xC99CE62546C95942ULL, "503382494.ogg"},
	{0xC99D01849BCA6086ULL, "mag291_torpor6027050.cl"},
	{0xC99D17DF934AE47CULL, "w_bonepitmining_a2_12_0_ll.mmh"},
	{0xC99D5E07A216C10FULL, "6023974_m.fxe"},
	{0xC99D8273C74B7ED1ULL, "vxw_blightlands_am_0_0.vlm"},
	{0xC99DC4267CE52F99ULL, "vxlt_warehouse_so_0_0.vlm"},
	{0xC99DFAD83A7AFCBEULL, "hlac_lt_ruinedhouse_n_-1_4z.dds"},
	{0xC99DFD28E8B3EC36ULL, "one220_yevhen6098933.cl"},
	{0xC99E6C29FA46048DULL, "6144983_m.fxe"},
	{0xC99E71199C39F8F6ULL, "gen00fl_aveline6114870.cl"},
	{0xC99E73B7B0EB2AE9ULL, "hlac_ht_0_hightown_a2_0_1y.dds"},
	{0xC99E8A97B8AC4F91ULL, "hlac_lt_0_lowtown_docks_n_-131075_1y.dds"},
	{0xC99EC1F4AFE3FB8CULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.impa.evt"},
	{0xC99EDDD11CE2C009ULL, "vxlt_gallowstemplar_n_527_0_0.vlm"},
	{0xC99F7BAC617F3225ULL, "84581362.ogg"},
	{0xC99FB0F821875F24ULL, "lt_0_lowtown_a3_n_c_21_0_lc.mmh"},
	{0xC99FCE57FF96ECF0ULL, "vxca_cave_265_0_0.vlm"},
	{0xC9A0421033993D1DULL, "6049819_m.fxe"},
	{0xC9A0AEE67A12330DULL, "6103906_m.fxe"},
	{0xC9A0B27BA809C5B0ULL, "751062441.ogg"},
	{0xC9A10F848FAC7837ULL, "944816089.ogg"},
	{0xC9A1AB586B64BD69ULL, "art\\vfx\\combat\\vfx_d_teleport_com_p.anb"},
	{0xC9A1CB4E5ACA6F93ULL, "character_character_p196.psh.fxo.dx11.assembly"},
	{0xC9A1CC1314F2F16DULL, "39720866.ogg"},
	{0xC9A253689B9262E5ULL, "vxw_encounter2_camp_fs_0_0.vlm"},
	{0xC9A27BA17EFD674EULL, "952924519.ogg"},
	{0xC9A28176021772C9ULL, "640201064.ogg"},
	{0xC9A2857AEAC0356CULL, "890254655.ogg"},
	{0xC9A2F7C38CDE8EEDULL, "vxlt_hangedman_t6_0_0.vlm"},
	{0xC9A2FEDF718DFE4BULL, "6076730_m.fxe"},
	{0xC9A32F3C92CC2AD4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_buttress.phy"},
	{0xC9A3333498EA9354ULL, "art\\characters\\creatures\\pridedemon\\animation\\dblw.ani"},
	{0xC9A36AF80564B34AULL, "distortiongeom.vsh.fxo"},
	{0xC9A44735228A5DE2ULL, "hlac_lt_0_lowtown_a3_n_c_131070_1x.dds"},
	{0xC9A489D0F9C634E9ULL, "water_pan_add_blur2.dds"},
	{0xC9A4C6E72BAFC08FULL, "vxlt_undercitydungeon_17f_0_0.vlm"},
	{0xC9A4ED02AEF780D8ULL, "vxlt_0_lowtown_a2_e31_0_0.vlm"},
	{0xC9A5554D6A238E7EULL, "hlac_lt_gallowsprison_n_-1_9z.dds"},
	{0xC9A5C9C294249612ULL, "gen00fl_anders6112996.cl"},
	{0xC9A5CE271E9EF11EULL, "6086893_m.fxe"},
	{0xC9A649D51B460516ULL, "vxlt_undercityserialkiller_1sn_0_0.vlm"},
	{0xC9A6A4D3EFE5CAE4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_broken.phy"},
	{0xC9A6C8D37A8EAED4ULL, "vxca_caverns2_810_0_1.vlm"},
	{0xC9A6DE0DFD963C78ULL, "follower_banter6163148.cl"},
	{0xC9A76773066E7936ULL, "6077822_m.fxe"},
	{0xC9A8D6755238D18AULL, "236302772.ogg"},
	{0xC9A8FDEF7A409AC4ULL, "6126450_m.fxe"},
	{0xC9A93BFB5945C439ULL, "6015623_m.fxe"},
	{0xC9AA19E3E96061B3ULL, "vxht_keep_n_63j_0_0.vlm"},
	{0xC9AA303498F0ACB6ULL, "art\\characters\\creatures\\pridedemon\\animation\\dblw.gad"},
	{0xC9AB673447450AA5ULL, "hlac_lt_0_lowtown_a2_n_c_131070_3x.dds"},
	{0xC9AB93C746652A75ULL, "39908220.ogg"},
	{0xC9ABA0F48EFD62D1ULL, "vxw_encounter2_8y_0_0.vlm"},
	{0xC9ABF450B9419751ULL, "vxw_blightlands_jl_0_0.vlm"},
	{0xC9AC2187AC761AE0ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_hed_snga_0d.dds"},
	{0xC9AC36036EDB4819ULL, "vxlt_warehouse_tn_0_0.vlm"},
	{0xC9AC72B3C37ABB48ULL, "6088234_f.fxe"},
	{0xC9AC7433DE3821DCULL, "lt_0_lowtown_a2_n_19_0_lc.mmh"},
	{0xC9ACA9F4AFF010E5ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.impa.ani"},
	{0xC9AD4B6734F6AA1EULL, "hlac_lt_0_lowtown_docks_65534_0y.dds"},
	{0xC9AD7AE0F29E81D7ULL, "6088933_m.fxe"},
	{0xC9AE04C538CA1444ULL, "vxca_caverns1_iy3_0_0.vlm"},
	{0xC9AE0C3C92D5F6D1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_buttress.msh"},
	{0xC9AE1A3C92D60EECULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_buttress.mao"},
	{0xC9AE1CBD956F05A7ULL, "vxht_0_hightown_a3_c4x_0_0.vlm"},
	{0xC9AE263C92D6228FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_buttress.mmh"},
	{0xC9AE6ACD8C0DCC56ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_90_c.anb"},
	{0xC9AE8B40E992522CULL, "6047250_m.fxe"},
	{0xC9AF20C36C084B8DULL, "6104806_m.fxe"},
	{0xC9AF815EBFA15BBDULL, "vxlt_0_lowtown_a2_enw_0_0.vlm"},
	{0xC9B026021BCFA40BULL, "vxht_playershouse_n_1qo_0_1.vlm"},
	{0xC9B05D7BA312D976ULL, "art\\levels\\races\\proxy\\props\\_textures\\coins_sa.dds"},
	{0xC9B05EE9566BB12CULL, "825931318.ogg"},
	{0xC9B0A32E8B4B2DC4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_loop.ani"},
	{0xC9B0DE4F3D2E352DULL, "hf_gencr_city_guard.mor"},
	{0xC9B1454FC7FE6FD8ULL, "6111751_m.fxe"},
	{0xC9B18AB73925FFE1ULL, "gen00fl_varric6020610.cl"},
	{0xC9B1E9356C8F1CDAULL, "vxht_0_hightown_n_at9_0_0.vlm"},
	{0xC9B240E91A0AAD47ULL, "llac_dr_primevaltomb_11985_1.dds"},
	{0xC9B2EEF57DFC576EULL, "6112033_m.fxe"},
	{0xC9B340B1856155B3ULL, "6042856_m.fxe"},
	{0xC9B34DD8819C24CBULL, "art\\characters\\weapons\\shields\\textures\\w_shd_001a_0d.dds"},
	{0xC9B390D77B8C6F4FULL, "vxht_0_hightown_a2_c7l_0_0.vlm"},
	{0xC9B39CC46061EA2EULL, "vxca_caverns1_dpk_0_0.vlm"},
	{0xC9B3F4134C2EDC5BULL, "gen_im_trs_dem_blt_r2.uti"},
	{0xC9B49E1DA45C2659ULL, "6040618_m.fxe"},
	{0xC9B4AE39D399F1B2ULL, "vxht_0_hightown_3w1_0_0.vlm"},
	{0xC9B4BE6860132350ULL, "w_woundedcoast_140_0.msh"},
	{0xC9B4CA874F2108E9ULL, "6196720_m.fxe"},
	{0xC9B4CC6860133B6AULL, "w_woundedcoast_140_0.mmh"},
	{0xC9B4D068601341A1ULL, "w_woundedcoast_140_0.mao"},
	{0xC9B5099569585772ULL, "hlac_lt_warehouse_n_65536_1x.dds"},
	{0xC9B54DA99C4A6C61ULL, "vxw_bonepitmining_a2_gr_0_0.vlm"},
	{0xC9B62E19FC5E2321ULL, "6040712_f.fxe"},
	{0xC9B659223C7A2708ULL, "6162560_m.fxe"},
	{0xC9B65A983E979482ULL, "622923040.ogg"},
	{0xC9B672B84DE4AF37ULL, "6080206_f.fxe"},
	{0xC9B695735FB68EC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debuffareacast.ani"},
	{0xC9B6A0D299362C5BULL, "dae000_arishok6074062.cl"},
	{0xC9B6D2B29F1E07C5ULL, "ip_restrict_race.ncs"},
	{0xC9B76FCC6C36B344ULL, "6123750_m.fxe"},
	{0xC9B779E6448CF14CULL, "6118583_m.fxe"},
	{0xC9B7A02E8B514726ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_loop.gad"},
	{0xC9B84295DAE14DF2ULL, "vxw_encounter2_gk_0_0.vlm"},
	{0xC9B9C77D7A7EF3F7ULL, "art\\characters\\weapons\\dagger\\w_dgr_004a.phy"},
	{0xC9B9D93602419F61ULL, "6116625_m.fxe"},
	{0xC9BA3B0E8EEBE37AULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgtb_0s.dds"},
	{0xC9BA42F3B0ED3933ULL, "vxlt_0_lowtown_bxt_0_0.vlm"},
	{0xC9BA662DAB37AFD1ULL, "vxw_blightlands_oc_0_0.vlm"},
	{0xC9BAA7E060D16099ULL, "vxlt_warehouse_ym_0_0.vlm"},
	{0xC9BAE490B570D3C8ULL, "6083334_f.fxe"},
	{0xC9BB392D37397998ULL, "36787373.ogg"},
	{0xC9BB5E72BCD5CA68ULL, "gen00fl_bethany6101398.cl"},
	{0xC9BB67CE04EFCE3DULL, "lt_0_lowtown_docks_123_0.mao"},
	{0xC9BB6BCE04EFD576ULL, "lt_0_lowtown_docks_123_0.mmh"},
	{0xC9BB6DCD8C185BC1ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_l_90_c.mmh"},
	{0xC9BB75CE04EFE674ULL, "lt_0_lowtown_docks_123_0.msh"},
	{0xC9BBAC502C04D4A4ULL, "ser121_jethann6175347.cl"},
	{0xC9BC1A32304074D8ULL, "vxht_0_hightown_n_bu7_0_0.vlm"},
	{0xC9BC65A1E23F4F62ULL, "vxlt_0_lowtown_docks_n_9r6_0_0.vlm"},
	{0xC9BC6A909743F977ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_3.gad"},
	{0xC9BCB0735FBB284BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debuffareacast.gad"},
	{0xC9BCCA5753BAD829ULL, "gen00fl_fenris6127991.cl"},
	{0xC9BCD9A79B63D9ABULL, "vxht_0_hightown_a3_c8q_0_1.vlm"},
	{0xC9BCF7E9B6A87F58ULL, "vxlt_gallowstemplar_n_456_0_0.vlm"},
	{0xC9BD0A4B88124110ULL, "sse_ss_bartrand.gda"},
	{0xC9BD1A31AEAA9B35ULL, "zz_dae_debug6029094.cl"},
	{0xC9BD1B568D15F91AULL, "qun110_petrice6042290.cl"},
	{0xC9BD5ACCA379BE02ULL, "vxw_encounter2_bd_0_0.vlm"},
	{0xC9BDC454631F4AF6ULL, "w_bonepitdestroyed_a3_19_0.msh"},
	{0xC9BDCE54631F5BF4ULL, "w_bonepitdestroyed_a3_19_0.mmh"},
	{0xC9BDD254631F6207ULL, "w_bonepitdestroyed_a3_19_0.mao"},
	{0xC9BDED3B89894FF8ULL, "hlac_ht_mansion2_818_22z.dds"},
	{0xC9BDFE72985D18B6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave3statue.mao"},
	{0xC9BE0A72985D2D55ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave3statue.mmh"},
	{0xC9BE0C72985D30CBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave3statue.msh"},
	{0xC9BE9B2E8B575D05ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_persuade_loop.evt"},
	{0xC9BED65F5D9F1848ULL, "vxw_bonepitmining_a2_gw_0_0.vlm"},
	{0xC9BF900DA37C9690ULL, "102856404.ogg"},
	{0xC9BFB173037D0664ULL, "qun211cr_braggart_ally_m4.utc"},
	{0xC9BFB5DCB19E5C80ULL, "gen00fl_aveline6112733.cl"},
	{0xC9BFE1F5C7F7484FULL, "hlac_w_encounter2_camp_-131067_0z.dds"},
	{0xC9BFEA7D5FD8C7B5ULL, "vxlt_0_lowtown_a2_n_e1q_0_0.vlm"},
	{0xC9C02A88610FDC13ULL, "additivecrack01.mao"},
	{0xC9C0845A4DE3C894ULL, "llac_lt_0_lowtown_a2_65538_0.dds"},
	{0xC9C15F42573B58B0ULL, "486544665.ogg"},
	{0xC9C1714FA80D3EBFULL, "6191283_f.fxe"},
	{0xC9C17A7241E96A1EULL, "vxdr_slavers_19p_0_0.vlm"},
	{0xC9C202B46D8287CFULL, "vxht_0_hightown_a2_c2o_0_0.vlm"},
	{0xC9C23D52838ECFC3ULL, "hlac_lt_0_lowtown_-65537_3x.dds"},
	{0xC9C2C1D72CA21317ULL, "w_encounter2_16_0.mmh"},
	{0xC9C2C5D72CA21A24ULL, "w_encounter2_16_0.mao"},
	{0xC9C2C9175254870FULL, "6105726_m.fxe"},
	{0xC9C2D7D72CA23889ULL, "w_encounter2_16_0.msh"},
	{0xC9C310831BCA8FF7ULL, "446064010.ogg"},
	{0xC9C31B8EEA14D754ULL, "gen00fl_varric6087463.cl"},
	{0xC9C33DC9E19B57BCULL, "vxlt_warehouse_e0_0_0.vlm"},
	{0xC9C3515E71CA6FBCULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured.ani"},
	{0xC9C37F90974A3B04ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_3.evt"},
	{0xC9C3A93EEE2CABCDULL, "vxlt_darktown_5kk_0_0.vlm"},
	{0xC9C3AD735FC141B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\debuffareacast.evt"},
	{0xC9C3BF868E4084E1ULL, "vxw_bonepitmining_a2_bq_0_0.vlm"},
	{0xC9C3E34BCB41D9DCULL, "hlac_lt_0_lowtown_docks_n_131070_3x.dds"},
	{0xC9C501FDE38C9FB5ULL, "40860786.ogg"},
	{0xC9C50C1A945B083BULL, "vxlt_gallowstemplar_4xu_0_0.vlm"},
	{0xC9C555A5BB39FF9BULL, "ht_0_hightown_n_24_0.rml"},
	{0xC9C59E687CB7C422ULL, "zz_mcr_debug6080872.cl"},
	{0xC9C5D1749B959284ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_imp_c.anb"},
	{0xC9C5F8AF77E8A33AULL, "one100dg_nabil6059782.cl"},
	{0xC9C617942D69B694ULL, "gen00fl_aveline6139705.cl"},
	{0xC9C6345BCCECADA0ULL, "6039303_m.fxe"},
	{0xC9C65602C3564555ULL, "llac_ht_0_hightown_65536_3.dds"},
	{0xC9C6EDEC862CAB76ULL, "6170253_m.fxe"},
	{0xC9C70F8AD1F38E8EULL, "vxlt_0_lowtown_n_f4z_0_0.vlm"},
	{0xC9C724D17F792EDEULL, "plt_cod_bks_lite_tombs_1.nss"},
	{0xC9C764BA96998ACFULL, "6126699_m.fxe"},
	{0xC9C769429ECC08D6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile0a.mao"},
	{0xC9C775429ECC1D75ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile0a.mmh"},
	{0xC9C777429ECC20EBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile0a.msh"},
	{0xC9C8A4EFF935B15FULL, "mer110cr_mine_scavenger_1.utc"},
	{0xC9C8B68DE8655646ULL, "art\\vfx\\environment\\vfx_hnd_r_pickaxe_c.dura.evt"},
	{0xC9C8BE693D0E3DB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_exit.ani"},
	{0xC9C8D80A9D2DC851ULL, "vxw_blightlands_pb_0_0.vlm"},
	{0xC9C927C14B4E5601ULL, "125126736.ogg"},
	{0xC9C9FB6F138F65D0ULL, "lgt231tr_undead_rise.ncs"},
	{0xC9CA5E5E71D0A4FEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured.gad"},
	{0xC9CA5E9AD68AB2D7ULL, "6086733_m.fxe"},
	{0xC9CA841F907039CEULL, "vxw_encounter3_fn_0_1.vlm"},
	{0xC9CB28734A0A98EBULL, "gen00fl_varric6017827.cl"},
	{0xC9CBA035B1C06CFFULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_meda_s.dds"},
	{0xC9CBE80EA070D3F6ULL, "hlac_ht_0_hightown_n_-65536_0z.dds"},
	{0xC9CBFB3015D8EB5AULL, "qun211_braggart6075689.cl"},
	{0xC9CC13EEF2EBB9A4ULL, "hlac_lt_0_lowtown_-1_7z.dds"},
	{0xC9CC1D3C90261BF1ULL, "vxht_0_hightown_a2_bu1_0_0.vlm"},
	{0xC9CD06AC7A596DFCULL, "6083706_m.fxe"},
	{0xC9CE2740EBF2E591ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl2_c.impa.ani"},
	{0xC9CE2A3DEC0AEEE6ULL, "vxlt_foundry_n_10a_0_0.vlm"},
	{0xC9CE9ACC0C1FA53AULL, "vxca_caverns1_mines_eh3_0_0.vlm"},
	{0xC9CEA4F53C409FCCULL, "llac_lt_0_lowtown_a3_n_c_65538_0.dds"},
	{0xC9CF048F46D5C36EULL, "vxht_mansion2_2d1_0_0.vlm"},
	{0xC9CF3A8FBD5A46BDULL, "dre180_db_rock_mage.cut"},
	{0xC9CF4F28EC21BF05ULL, "vxlt_0_lowtown_docks_a3_bug_0_0.vlm"},
	{0xC9CFB6759195555BULL, "6146795_f.fxe"},
	{0xC9CFC9693D146F53ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_exit.gad"},
	{0xC9CFCD8DE86B9B91ULL, "art\\vfx\\environment\\vfx_hnd_r_pickaxe_c.dura.gad"},
	{0xC9D050604166FD39ULL, "gen_im_cft_rec_psn_05.uti"},
	{0xC9D097909754EEFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_3.ani"},
	{0xC9D0F33F7E0A5F3FULL, "qcr220_orsino6061340.cl"},
	{0xC9D10AA610365059ULL, "6027352_m.fxe"},
	{0xC9D11AED10EDE934ULL, "hor_dust.mao"},
	{0xC9D1395E71D6841DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured.evt"},
	{0xC9D18C5EA0368E21ULL, "693934788.ogg"},
	{0xC9D1F65696BED529ULL, "vxw_bonepitdestroyed_a3_27d_0_0.vlm"},
	{0xC9D21B1BE022C44DULL, "vxlt_darktown_5lj_0_0.vlm"},
	{0xC9D2316380369D61ULL, "vxw_bonepitmining_a2_ip_0_0.vlm"},
	{0xC9D2667A8706F112ULL, "vxw_bonepitmining_a2_1kn_0_0.vlm"},
	{0xC9D2A0474E6E3857ULL, "art\\vfx\\creatures\\fxc_ohurl_imp_p.anb"},
	{0xC9D3353D0920CB7CULL, "gen00fl_isabela6099525.cl"},
	{0xC9D3A5F6DA7748FCULL, "6146838_m.fxe"},
	{0xC9D3D420218B5328ULL, "107923741.ogg"},
	{0xC9D3ED6FD41366DBULL, "6119165_m.fxe"},
	{0xC9D3F0749BA20487ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_imp_c.mmh"},
	{0xC9D3F83873FB5CB0ULL, "vxht_0_hightown_n_ccm_0_1.vlm"},
	{0xC9D45BD5DA999175ULL, "hlac_lt_backalley_n_0_12y.dds"},
	{0xC9D495293F7F1340ULL, "6176734_m.fxe"},
	{0xC9D4CB48F08A8A39ULL, "527981759.ogg"},
	{0xC9D53240EBF91733ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl2_c.impa.gad"},
	{0xC9D54B4B28D5FAB9ULL, "vxht_keep_n_582_0_0.vlm"},
	{0xC9D54BB5BFB8D408ULL, "298116438.ogg"},
	{0xC9D57D516FB5A6E9ULL, "vxlt_0_lowtown_a2_n_hf9_0_0.vlm"},
	{0xC9D5B075E8F1C732ULL, "603897577.ogg"},
	{0xC9D5C6693D18D5B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\show_paper_exit.evt"},
	{0xC9D6CE8DE871BC3FULL, "art\\vfx\\environment\\vfx_hnd_r_pickaxe_c.dura.ani"},
	{0xC9D72FC952ACF6EEULL, "hlac_ht_mansion2_818_6x.dds"},
	{0xC9D732D5154DF1FEULL, "6126922_m.fxe"},
	{0xC9D749E78F23E0D1ULL, "vxw_blightlands_ua_0_0.vlm"},
	{0xC9D86446E0B4A9CAULL, "vxht_keep_n_1vc_0_0.vlm"},
	{0xC9D891D43F25EE4EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mede_0d.dds"},
	{0xC9D95A5C0EAC5DC4ULL, "vxca_caverns1_ij6_0_0.vlm"},
	{0xC9D97B3B657C7A74ULL, "6083108_m.fxe"},
	{0xC9D985EB99EB2B53ULL, "220480573.ogg"},
	{0xC9DA07CF2638AF0AULL, "vxlt_gallowscourtyard_72s_0_0.vlm"},
	{0xC9DA3E868765EF02ULL, "vxw_encounter2_hf_0_0.vlm"},
	{0xC9DA67912D9FF80DULL, "hf_mage_ella.mor"},
	{0xC9DB39A15770E5C1ULL, "mag370_meredith6046911.cl"},
	{0xC9DB61642A16924EULL, "w_sundermount_199_0.msh"},
	{0xC9DB6F642A16A99FULL, "w_sundermount_199_0.mao"},
	{0xC9DB7B642A16BE3CULL, "w_sundermount_199_0.mmh"},
	{0xC9DBEFF43AE813E8ULL, "lt_0_lowtown_docks_n_227_0.msh"},
	{0xC9DC01F43AE83279ULL, "lt_0_lowtown_docks_n_227_0.mao"},
	{0xC9DC0DF43AE84612ULL, "lt_0_lowtown_docks_n_227_0.mmh"},
	{0xC9DC2F40EBFF3098ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl2_c.impa.evt"},
	{0xC9DD70E7D1D76FE1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicechair.phy"},
	{0xC9DD75493DD71067ULL, "8973143.ogg"},
	{0xC9DD766C38CBDBEEULL, "vxht_mansion2_2a2_0_0.vlm"},
	{0xC9DDE9892208F25FULL, "567959460.ogg"},
	{0xC9DE078478626C17ULL, "493830880.ogg"},
	{0xC9DE7F25F5BD8BBAULL, "128254860.ogg"},
	{0xC9DEEC5B4FD39083ULL, "llac_lt_0_lowtown_1_1.dds"},
	{0xC9DF029158EFB241ULL, "6152809_m.fxe"},
	{0xC9DF2C13F1D8EC3CULL, "6066109_m.fxe"},
	{0xC9DF63B262CECC28ULL, "gen00fl_aveline6128400.cl"},
	{0xC9DF6DA1D2E97C95ULL, "follower_banter6167121.cl"},
	{0xC9E0E0C3F65DC536ULL, "fx_web01.mao"},
	{0xC9E10203D7E5D52DULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgtb_0s.dds"},
	{0xC9E1475D5C5EA22CULL, "vxlt_0_lowtown_docks_a3_n_d1o_0_0.vlm"},
	{0xC9E1758874E6E982ULL, "lgt111ar_rand_sewer_1_nt.are"},
	{0xC9E191D8A20ACFDBULL, "hlac_ht_0_hightown_n_65535_3z.dds"},
	{0xC9E1C99FB8DB89C6ULL, "334887496.ogg"},
	{0xC9E1D340BE32003AULL, "388200289.ogg"},
	{0xC9E20037FADA7320ULL, "hlac_ht_keep_n_3736_2y.dds"},
	{0xC9E24DD92D17372EULL, "zz_otl_debug6060662.cl"},
	{0xC9E27EA9EFADA9E6ULL, "6076494_m.fxe"},
	{0xC9E281315A33B4ADULL, "llac_lt_backalley_g_65537_5.dds"},
	{0xC9E281FCA4401757ULL, "6108251_m.fxe"},
	{0xC9E2A4D44B4F1ED7ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_ext_c.impa.gad"},
	{0xC9E2C93ED87C86C4ULL, "mag211_marethari6025798.cl"},
	{0xC9E353503F3B1944ULL, "static_static_p10.vsh.fxo.dx11"},
	{0xC9E3710E74D1A52BULL, "oil_f2.dds"},
	{0xC9E39ACCF1815506ULL, "mag121_idunna6084160.cl"},
	{0xC9E3DB74B55BA3DBULL, "llac_lt_0_lowtown_n_-3_3.dds"},
	{0xC9E3EF1E503441D4ULL, "vxlt_warehouse_n_141_0_0.vlm"},
	{0xC9E4C40D7B8FF16EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_amllcrstnew_sa.dds"},
	{0xC9E50BF210C7DF44ULL, "vxht_mansion2_31u_0_0.vlm"},
	{0xC9E5AF6BA0CE4928ULL, "306760381.ogg"},
	{0xC9E5F44A567C5A80ULL, "hlac_ht_mansion_n_3382_0z.dds"},
	{0xC9E63C5D9E54DEBEULL, "6027386_m.fxe"},
	{0xC9E66A7B5727E9A9ULL, "vxlt_0_lowtown_a2_f1i_0_0.vlm"},
	{0xC9E6E98AB6025278ULL, "gen_im_acc_blt_act3_11.uti"},
	{0xC9E76DB061E34DF7ULL, "vxht_mansion2_2f2_0_0.vlm"},
	{0xC9E76FC90622BE58ULL, "vxht_0_hightown_n_bf2_0_0.vlm"},
	{0xC9E779CDE684A7F6ULL, "6076898_m.fxe"},
	{0xC9E7A4586D0C022AULL, "273516972.ogg"},
	{0xC9E7CC3900A27644ULL, "vxca_caverns1_im7_0_0.vlm"},
	{0xC9E7CD6C0FD35E2FULL, "gen00fl_fenris6113961.cl"},
	{0xC9E7E573AAC2BD19ULL, "vxht_0_hightown_n_9h2_0_0.vlm"},
	{0xC9E863B9D64AF24FULL, "vxw_bonepit_1v6_0_0.vlm"},
	{0xC9E8E5E3EDEAEE49ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.dura.ani"},
	{0xC9E8F562CA0E03AEULL, "vxw_bonepit_2am_0_0.vlm"},
	{0xC9E900F674124CF1ULL, "vxht_0_hightown_a2_bc7_0_0.vlm"},
	{0xC9E9160AC818419CULL, "hlac_lt_0_lowtown_a3_n_c_65536_7y.dds"},
	{0xC9E9B9D44B556164ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_ext_c.impa.evt"},
	{0xC9E9F3EB0DBAD0DEULL, "dre000_dougal6130970.cl"},
	{0xC9EA1C8D14A489E0ULL, "85113462.ogg"},
	{0xC9EADC157945DAE0ULL, "zz_dre_debug6036249.cl"},
	{0xC9EB6C8874EF2F55ULL, "lgt111ar_rand_sewer_1_nt.lst"},
	{0xC9EBA83C38296801ULL, "6102101_m.fxe"},
	{0xC9EBC3B0D29A4C87ULL, "gen00fl_anders6128356.cl"},
	{0xC9EBC57EB5D4A059ULL, "dae300_cullen6140636.cl"},
	{0xC9EBCFCC4017F1F1ULL, "vxlt_0_lowtown_a2_n_ep3_0_1.vlm"},
	{0xC9EC6BE2091C0037ULL, "dae000ip_crowd_ff6_talk15.utp"},
	{0xC9EC70DEE1852D09ULL, "534828201.ogg"},
	{0xC9EC8F61A037E37FULL, "109486319.ogg"},
	{0xC9ED18210EDE33F6ULL, "6133071_f.fxe"},
	{0xC9EE051F87B2B458ULL, "6107481_f.fxe"},
	{0xC9EE0E1829130B45ULL, "vxlt_gallowscourtyard_6qt_0_0.vlm"},
	{0xC9EE1CFAAFBA3B72ULL, "one101_orders_followers6115782.cl"},
	{0xC9EE200EB015FB30ULL, "llac_w_encounter2_196606_0.dds"},
	{0xC9EE2275CA487278ULL, "258136968.ogg"},
	{0xC9EEC886A6BB3903ULL, "character_character_p118.psh.fxo"},
	{0xC9EF00E3EDEF87CBULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.dura.gad"},
	{0xC9EF690D5E3584F4ULL, "dae000_gamlen6131101.cl"},
	{0xC9F0479F22E3FF3EULL, "vxht_bartrand_n_6qr_0_0.vlm"},
	{0xC9F05B2DE809FC93ULL, "vxw_woundedcoast_fp_0_0.vlm"},
	{0xC9F0972BB246ED26ULL, "vxw_bonepitmining_a2_7y_0_0.vlm"},
	{0xC9F1BDBB9ABD8D41ULL, "vxht_mansion2b_1yr_0_0.vlm"},
	{0xC9F1DEFECCBA0DD5ULL, "204629048.ogg"},
	{0xC9F25D94A7AF8EBFULL, "539800091.ogg"},
	{0xC9F260FB422A5A54ULL, "vxlt_warehouse_n_196_0_0.vlm"},
	{0xC9F2648874F54044ULL, "lgt111ar_rand_sewer_1_nt.ncs"},
	{0xC9F2D345495DB204ULL, "vfx_distortionmask_p4.vsh.fxo.dx11"},
	{0xC9F30EE684A06534ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_bdaa_0e.dds"},
	{0xC9F38C500773D0F6ULL, "vxlt_0_lowtown_n_e1r_0_0.vlm"},
	{0xC9F3E768BD765193ULL, "435312809.ogg"},
	{0xC9F425B95E9CBC53ULL, "llac_ht_0_hightown_-131071_0.dds"},
	{0xC9F427F99D1AF832ULL, "710936769.ogg"},
	{0xC9F46EE53E5691BBULL, "vxht_0_hightown_a3_apz_0_0.vlm"},
	{0xC9F5AABC565A8501ULL, "ability_usetype.gda"},
	{0xC9F5BB96F7CA3B7EULL, "vxlt_foundry_n_ye_0_0.vlm"},
	{0xC9F5E033B6EC79DAULL, "hlac_ca_caverns2prc_5802_11z.dds"},
	{0xC9F5FDE3EDF5A130ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.dura.evt"},
	{0xC9F64545842FE394ULL, "gen00fl_varric6107293.cl"},
	{0xC9F78EA8EF42FBB2ULL, "dae011cr_drunk_2.utc"},
	{0xC9F7D1D44B61C75DULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_ext_c.impa.ani"},
	{0xC9F803A5DBF28D81ULL, "mag210_arianni6024093.cl"},
	{0xC9F81ECF3A66F544ULL, "736390750.ogg"},
	{0xC9F8EB7205AADE54ULL, "hlac_lt_gallowstemplar_n_2384_3y.dds"},
	{0xC9F94ED4F3181D1AULL, "vxlt_warehouse_5v_0_0.vlm"},
	{0xC9F9AB0D15489656ULL, "w_blightlands_220_0.mao"},
	{0xC9F9B70D1548AAF5ULL, "w_blightlands_220_0.mmh"},
	{0xC9F9B90D1548AE6BULL, "w_blightlands_220_0.msh"},
	{0xC9FA5F2EB86AA56DULL, "vxht_mansion2_2nx_0_0.vlm"},
	{0xC9FA80ADBC73E667ULL, "736095438.ogg"},
	{0xC9FA9910221620D2ULL, "6140584_m.fxe"},
	{0xC9FA9D474E8F8558ULL, "art\\vfx\\creatures\\fxc_ohurl_imp_p.mmh"},
	{0xC9FABD645E9CAC61ULL, "qun131_party_comment6117280.cl"},
	{0xC9FAC61CD9EC9A55ULL, "vxca_caverns1_e92_0_0.vlm"},
	{0xC9FAF9B4FAD206CCULL, "prm000pt_promo_core.ncs"},
	{0xC9FB4D7740249EC8ULL, "art\\toolset\\_textures\\iconrgby_sa.dds"},
	{0xC9FB658DFA01D814ULL, "287603064.ogg"},
	{0xC9FC39931D7C86DBULL, "gen00fl_sebastian6174323.cl"},
	{0xC9FC6CAB710D9CBBULL, "dae000ip_crowd_ff5_talk10.utp"},
	{0xC9FC72738C584F12ULL, "gen00fl_merrill6115986.cl"},
	{0xC9FCBCB106855C8BULL, "232801399.ogg"},
	{0xC9FD053DA973B9BCULL, "vxlt_warehouse_y4_0_0.vlm"},
	{0xC9FD4342940DB00DULL, "llac_ca_caverns1_mines_24421_8.dds"},
	{0xC9FDC28F380EAA08ULL, "6129684_m.fxe"},
	{0xC9FE0CCD3A5C28A9ULL, "vxht_0_hightown_a2_2xr_0_0.vlm"},
	{0xC9FE2934E5D820C0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1ww_basic_2.mao"},
	{0xC9FE3849E2C0BC3DULL, "512964404.ogg"},
	{0xC9FE38689AD43993ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_vfx_suca_0s.dds"},
	{0xC9FEB367244C4707ULL, "vxca_cave_19p_0_0.vlm"},
	{0xC9FEB97C14DA17BEULL, "vxht_bartrand_n_6zs_0_0.vlm"},
	{0xC9FECD0ADA001513ULL, "vxw_woundedcoast_cs_0_0.vlm"},
	{0xC9FEE64A22AA677FULL, "gen00fl_anders6104000.cl"},
	{0xC9FEF10A6FFD98CAULL, "cd_act1_rockwraith_intro_demo_rogue.wbd"},
	{0xC9FEF1C3F676DFB8ULL, "fx_web01.dds"},
	{0xC9FF43E423991F6AULL, "one000_ilen6057357.cl"},
	{0xC9FF92B5CC1BB944ULL, "vxlt_warehouse_qd_0_0.vlm"},
	{0xC9FFAB7BCDBE9C18ULL, "6177197_m.fxe"},
	{0xC9FFC819494A98A4ULL, "vxlt_backalley_g_1xh_0_1.vlm"},
	{0xCA001B2C9F9B9C9CULL, "6071145_m.fxe"},
	{0xCA005A642A35A207ULL, "w_sundermount_199_0.phy"},
	{0xCA0084BCE3505E8AULL, "lt_0_lowtown_docks_96_0.mao"},
	{0xCA0088BCE35065D9ULL, "lt_0_lowtown_docks_96_0.mmh"},
	{0xCA0092BCE35076C7ULL, "lt_0_lowtown_docks_96_0.msh"},
	{0xCA01501B0DE849B8ULL, "zz_vault_debug6178955.cl"},
	{0xCA016CC572B1D546ULL, "atl_shared_dxt5_00.dds"},
	{0xCA017CB2A94CBA43ULL, "vxlt_0_lowtown_a2_n_h27_0_0.vlm"},
	{0xCA01871A1DC663F0ULL, "6099561_m.fxe"},
	{0xCA019C7D7ABC604FULL, "art\\characters\\weapons\\dagger\\w_dgr_004a.mao"},
	{0xCA01A87D7ABC74ECULL, "art\\characters\\weapons\\dagger\\w_dgr_004a.mmh"},
	{0xCA01AE7D7ABC7EDEULL, "art\\characters\\weapons\\dagger\\w_dgr_004a.msh"},
	{0xCA021671442FABD2ULL, "docks_male_sailor_mob6176653.cl"},
	{0xCA021D88EFB74D55ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.mmh"},
	{0xCA021DB4E15334ABULL, "6076208_m.fxe"},
	{0xCA0252B0B6663A5AULL, "132641197.ogg"},
	{0xCA02855A2BCD7ABCULL, "6136172_m.fxe"},
	{0xCA029955A7E6EC57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_teleport.gad"},
	{0xCA02BA68EA1F7A4DULL, "6164237_m.fxe"},
	{0xCA0372D60E2D0B50ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02barge.mao"},
	{0xCA0376D60E2D1263ULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02barge.mmh"},
	{0xCA0384D60E2D29BDULL, "art\\levels\\races\\ferelden\\props\\docks\\barge\\prp_02barge.msh"},
	{0xCA038D94879C4A1DULL, "213825165.ogg"},
	{0xCA039B579744711FULL, "852316233.ogg"},
	{0xCA03FED05C272F7AULL, "158170527.ogg"},
	{0xCA045382EA0EEF58ULL, "vxht_0_hightown_n_bl0_0_0.vlm"},
	{0xCA045F0A723A462CULL, "gen00fl_merrill6128760.cl"},
	{0xCA04751191A34036ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocksmall.mao"},
	{0xCA04811191A354D5ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocksmall.mmh"},
	{0xCA04831191A3584BULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrocksmall.msh"},
	{0xCA0494F9DCE6315DULL, "ts_kirk_darktown_03.gda"},
	{0xCA0499E67E63DE39ULL, "hlac_lt_0_lowtown_docks_n_-65539_2y.dds"},
	{0xCA04C92D8EAEEE19ULL, "vxht_0_hightown_n_9v0_0_0.vlm"},
	{0xCA04F99C52A136D6ULL, "345496520.ogg"},
	{0xCA04FAADBDA24F6FULL, "zz_t_marketing_carver.cut"},
	{0xCA0553CACC5FCD0AULL, "528969285.ogg"},
	{0xCA059EE0ED909D34ULL, "6069926_m.fxe"},
	{0xCA05A15594A50B15ULL, "436198354.ogg"},
	{0xCA05EF31B0E1499EULL, "zz_one_debug6077959.cl"},
	{0xCA078DAD7172A226ULL, "6166373_m.fxe"},
	{0xCA078E23B3051CF3ULL, "vxlt_darktown_6by_0_0.vlm"},
	{0xCA081B20D4DC34C5ULL, "244563717.ogg"},
	{0xCA085148E90BE4E6ULL, "hlac_lt_0_lowtown_a2_n_c_65534_2y.dds"},
	{0xCA0855EFB5200C87ULL, "vxht_mansion_n_6j9_0_0.vlm"},
	{0xCA086B491BFC4216ULL, "453546847.ogg"},
	{0xCA08A863796FA4ECULL, "and100_anders6095914.cl"},
	{0xCA08CC030EAE256EULL, "vxht_mansion2_2p5_0_0.vlm"},
	{0xCA08E82B3E3D5343ULL, "hlac_dr_primevaltomb_659_1z.dds"},
	{0xCA094D91C95884F1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1a.msh"},
	{0xCA095B91C9589C0CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1a.mao"},
	{0xCA096791C958B0AFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1a.mmh"},
	{0xCA096B8F62B73F01ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1ww_basic_1.mao"},
	{0xCA09AE55A7ED2EE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_teleport.evt"},
	{0xCA09FA3CC19CDD9AULL, "mage_haste.dds"},
	{0xCA0A34C75D0D8F30ULL, "49435559.ogg"},
	{0xCA0AEDB639307481ULL, "one120_athenril6198428.cl"},
	{0xCA0B5D027868BFB3ULL, "sp_spark_01.mao"},
	{0xCA0BFD695D1B3343ULL, "6072330_m.fxe"},
	{0xCA0C2DEE4EDF5312ULL, "vxw_bonepitmining_a2_1wj_0_0.vlm"},
	{0xCA0C443220BB1130ULL, "vxht_0_hightown_a2_c5i_0_0.vlm"},
	{0xCA0CAD77B217D115ULL, "6020484_f.fxe"},
	{0xCA0D1810DF08ADC1ULL, "dre000_varric6035189.cl"},
	{0xCA0D1EF07C9480A1ULL, "600147540.ogg"},
	{0xCA0D300FF4968203ULL, "w_sundermount_14_0.mao"},
	{0xCA0D340FF4968950ULL, "w_sundermount_14_0.mmh"},
	{0xCA0D3EE7CBF62D93ULL, "vxw_woundedcoast_xr_0_0.vlm"},
	{0xCA0D420FF496A16AULL, "w_sundermount_14_0.msh"},
	{0xCA0D556B22A760C6ULL, "art\\vfx\\combat\\cn_hed_corm.mao"},
	{0xCA0D5F1159931567ULL, "786804059.ogg"},
	{0xCA0D721246B35158ULL, "748680737.ogg"},
	{0xCA0D8D2FA06510BDULL, "6068450_m.fxe"},
	{0xCA0E025C6E6CE3A1ULL, "406969386.ogg"},
	{0xCA0E0492BE11D1C4ULL, "vxlt_warehouse_ze_0_0.vlm"},
	{0xCA0EA1757EA9BE41ULL, "vxht_mansion2b_1wp_0_0.vlm"},
	{0xCA0EA5C63868C480ULL, "vxht_0_hightown_a2_c15_0_0.vlm"},
	{0xCA0F1A88EFC1D2E2ULL, "art\\vfx\\creatures\\vfx_c_hd_spit_p.anb"},
	{0xCA0F6AC97A16A038ULL, "vxlt_foundry_n_w9_0_0.vlm"},
	{0xCA0F97CAC6D07286ULL, "vxlt_foundry_n_ww_0_0.vlm"},
	{0xCA0FAE940D3368A2ULL, "hlac_lt_0_lowtown_a2_n_c_-1_4z.dds"},
	{0xCA100E8F2E52D7CFULL, "seb221_flora_harimann6177667.cl"},
	{0xCA10D1849AACAB42ULL, "w_blightlands_74_0.msh"},
	{0xCA10D3849AACAEF8ULL, "w_blightlands_74_0.mmh"},
	{0xCA10DF849AACC29BULL, "w_blightlands_74_0.mao"},
	{0xCA11097DC891B3F1ULL, "qcr270_fake_meredith6066355.cl"},
	{0xCA1186D623ED93AAULL, "6154039_m.fxe"},
	{0xCA1209D17FB8EC29ULL, "plt_cod_bks_lite_tombs_1.plo"},
	{0xCA121D54CC3B39B0ULL, "vxlt_0_lowtown_db3_0_0.vlm"},
	{0xCA1225A802157E19ULL, "283942805.ogg"},
	{0xCA124717A426FA1EULL, "511893139.ogg"},
	{0xCA12CE5463679BF7ULL, "w_bonepitdestroyed_a3_19_0.rml"},
	{0xCA12F8122FE538A3ULL, "6112685_m.fxe"},
	{0xCA1354634DFCD7CEULL, "lgt210_qun_blk_qunari6142320.cl"},
	{0xCA13957B239B61EEULL, "hlac_ht_0_hightown_131071_2z.dds"},
	{0xCA13F8399B001F77ULL, "dae000ip_crowd_mf8_talk16.utp"},
	{0xCA14EB82341A7496ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkright.ani"},
	{0xCA152A2D0A741AEEULL, "primitive_d.mao"},
	{0xCA15DF3302E1FDAFULL, "isa311_hawke_comments6117905.cl"},
	{0xCA162A66B8E0C4F4ULL, "6006068_m.fxe"},
	{0xCA16C7CCA7162507ULL, "vxht_mansion_n_6m8_0_0.vlm"},
	{0xCA16F29FB78E73ADULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swiss_n.dds"},
	{0xCA171FE28D29D85BULL, "6101471_f.fxe"},
	{0xCA17C655A7F994DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_teleport.ani"},
	{0xCA17D2DE33E77189ULL, "vxlt_0_lowtown_docks_n_b6r_0_0.vlm"},
	{0xCA17F26F13D177DDULL, "lgt231tr_undead_rise.utt"},
	{0xCA180F568E142EC2ULL, "193427493.ogg"},
	{0xCA188109ABD2EC83ULL, "320604101.ogg"},
	{0xCA1890655F7AD48EULL, "799241550.ogg"},
	{0xCA18DD3C2E5F1153ULL, "mag330_walter6045854.cl"},
	{0xCA18E7517B34CF88ULL, "61625464.ogg"},
	{0xCA191E9BE51D9738ULL, "hlac_ht_bartrand_n_3366_1z.dds"},
	{0xCA19222AC97A88CAULL, "dae200_bran6094793.cl"},
	{0xCA193FC1A5CF95AEULL, "dre000_bodahn6028163.cl"},
	{0xCA19862FBE25A9A0ULL, "mrl251ip_bosshorde_c.utp"},
	{0xCA1AB60F12B129B0ULL, "vxht_0_hightown_a2_c0v_0_0.vlm"},
	{0xCA1AC2DFFFDD9197ULL, "mag360_samson6054691.cl"},
	{0xCA1ACE31FCAE9A72ULL, "hlac_lt_gallowstemplar_f_2383_7z.dds"},
	{0xCA1ADA849AB50E83ULL, "w_blightlands_74_0.phy"},
	{0xCA1AF4AD89C53AEAULL, "6097150_m.fxe"},
	{0xCA1B4A0313178D6BULL, "vxca_caverns2_7my_0_0.vlm"},
	{0xCA1B703777ECCA4BULL, "6128406_m.fxe"},
	{0xCA1BB0C4BDEC4613ULL, "vxw_woundedcoast_uu_0_0.vlm"},
	{0xCA1C2A2007DBD028ULL, "vxlt_0_lowtown_n_byk_0_0.vlm"},
	{0xCA1D148EAF9BFEC8ULL, "6106953_m.fxe"},
	{0xCA1D2FA8D4E8A967ULL, "w_sundermount_119_0.rml"},
	{0xCA1EC0EC33E4C01BULL, "vxlt_0_lowtown_h5q_0_0.vlm"},
	{0xCA1EE4895CE97B71ULL, "hlac_lt_undercitydungeon_-1_7y.dds"},
	{0xCA1EE6C9B1155ADDULL, "84253157.ogg"},
	{0xCA1F21EACC56D8BAULL, "hlac_w_sundermount_262143_0y.dds"},
	{0xCA1F31084BA26FDDULL, "dae000_mother6099168.cl"},
	{0xCA1FD425F7D7A216ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardc.msh"},
	{0xCA1FDE25F7D7B314ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardc.mmh"},
	{0xCA20344BC3920F28ULL, "vxlt_0_lowtown_a2_n_c_e1z_0_0.vlm"},
	{0xCA2050D07B5D92CAULL, "142729104.ogg"},
	{0xCA206A162CB8F030ULL, "135789982.ogg"},
	{0xCA206ED8CB0C3D51ULL, "llac_lt_gallowstemplar_n_2383_4.dds"},
	{0xCA20A43AA98E4E42ULL, "vxlt_0_lowtown_a3_n_c_e1m_0_0.vlm"},
	{0xCA2144A610802EA6ULL, "6129249_f.fxe"},
	{0xCA21DDDA33E71ADDULL, "689056948.ogg"},
	{0xCA221DB9886DE3C5ULL, "686426485.ogg"},
	{0xCA223E69416761EDULL, "qun221_petrice6075977.cl"},
	{0xCA224A5866352A41ULL, "zz_dae_debug6029854.cl"},
	{0xCA227F570B0A3A4DULL, "436580625.ogg"},
	{0xCA22F3823426BF87ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkright.evt"},
	{0xCA2336E7B9E48D43ULL, "52042653.ogg"},
	{0xCA23486D13951889ULL, "6109113_m.fxe"},
	{0xCA237D6F12210613ULL, "vxdr_dungeon_1lu_0_1.vlm"},
	{0xCA23A351309424ADULL, "6116984_m.fxe"},
	{0xCA23E707D391B970ULL, "qun131_warrens_thug6038574.cl"},
	{0xCA242C63C50D0CF8ULL, "350901394.ogg"},
	{0xCA242EA8D4EEC67FULL, "w_sundermount_119_0.phy"},
	{0xCA24CEBEFF9F88F1ULL, "hla_ht_0_hightown_a3_131072_0x.dds"},
	{0xCA24D6C6DFD28CE0ULL, "dae021_male_worshipper_sta6150040.cl"},
	{0xCA24E6AA171F4AE9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2a.phy"},
	{0xCA259C6B3BDEFB08ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_altar_sa.dds"},
	{0xCA263CA8A01FDFB1ULL, "llac_lt_0_lowtown_a2_-196607_0.dds"},
	{0xCA270395FE213101ULL, "mrl310_merrill6119861.cl"},
	{0xCA271BB4B50A1CFDULL, "gen00fl_merrill6107139.cl"},
	{0xCA27F2B5622059D9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\originals\\attackv2.ani"},
	{0xCA2818F00DDD6CFEULL, "6162412_m.fxe"},
	{0xCA283A98DB544E43ULL, "vxht_mansion2_24h_0_0.vlm"},
	{0xCA286B94C59D7314ULL, "l_dae230ar_undercity_dt.wbd"},
	{0xCA2888E165C0C6FEULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzbodyhang_0.pwk"},
	{0xCA28981CEC33621BULL, "s_deca_cr2.mao"},
	{0xCA2965ED0362E7F0ULL, "lt_0_lowtown_docks_108_0.mao"},
	{0xCA2969ED0362EE03ULL, "lt_0_lowtown_docks_108_0.mmh"},
	{0xCA2977ED0363065DULL, "lt_0_lowtown_docks_108_0.msh"},
	{0xCA2A1482342D16D0ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkright.gad"},
	{0xCA2A22A1AFE25E93ULL, "vxw_woundedcoast_rt_0_0.vlm"},
	{0xCA2A970042E536B8ULL, "29767133.ogg"},
	{0xCA2AC455EBEE74E8ULL, "ser221_gascard6134250.cl"},
	{0xCA2AD056337BE333ULL, "hlac_ht_playershouse_n_855_16y.dds"},
	{0xCA2B19A3F8146109ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_wound.ani"},
	{0xCA2B27412AFCF9B6ULL, "vxca_cave_1ad_0_1.vlm"},
	{0xCA2CCF21CA3ED7FAULL, "vxlt_0_lowtown_a2_e20_0_0.vlm"},
	{0xCA2D7269F55A5B98ULL, "and230_mage_girl6135825.cl"},
	{0xCA2DC90890AD4AC9ULL, "fen110_fenris6097073.cl"},
	{0xCA2E28363C7907BBULL, "w_woundedcoast_25_0.rml"},
	{0xCA2E3C8AAC002FF1ULL, "vxlt_0_lowtown_a2_hd4_0_0.vlm"},
	{0xCA2E46B15EE2A90BULL, "1048103523.ogg"},
	{0xCA2EB3578A83D0C9ULL, "mcr370_cut_animate_guardian6086017.cl"},
	{0xCA2ECDB56226391BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\originals\\attackv2.gad"},
	{0xCA2F2B7BDF4E1649ULL, "gen_im_npc_arm_hum_boo_gwdc.uti"},
	{0xCA2F40403FAEB0F1ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_1z.dds"},
	{0xCA30A3CF5768A32AULL, "650636170.ogg"},
	{0xCA31422C83F5F5EFULL, "clf260_harley6069722.cl"},
	{0xCA31BA4A058B3109ULL, "6102013_m.fxe"},
	{0xCA31BFF69153F1CCULL, "6161510_m.fxe"},
	{0xCA31D4F2F9F43DCBULL, "871684180.ogg"},
	{0xCA3234A3F81AAD8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_wound.gad"},
	{0xCA32629F8744D43DULL, "651116636.ogg"},
	{0xCA326B5C6D819821ULL, "822274660.ogg"},
	{0xCA326B89CF6F2B6CULL, "vxca_cave_ib_0_0.vlm"},
	{0xCA330092B3B9751CULL, "vrc321_party_barks6095756.cl"},
	{0xCA34F5C838518650ULL, "vxlt_0_lowtown_a2_n_c_htx_0_0.vlm"},
	{0xCA352FA8D4FD3A84ULL, "w_sundermount_119_0.mmh"},
	{0xCA3533A8D4FD41D7ULL, "w_sundermount_119_0.mao"},
	{0xCA3545A8D4FD6026ULL, "w_sundermount_119_0.msh"},
	{0xCA3556F5D2893407ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_floor_d.dds"},
	{0xCA35C8C51E2FCD36ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls107a.msh"},
	{0xCA35D2C51E2FDE34ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls107a.mmh"},
	{0xCA35D6C51E2FE547ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls107a.mao"},
	{0xCA36661B47F0CB27ULL, "830323659.ogg"},
	{0xCA3684868EBE39DFULL, "one110_pryce6088412.cl"},
	{0xCA368ACCFFD3857EULL, "6167712_m.fxe"},
	{0xCA369E0260CC7A6DULL, "dae000_orana6149111.cl"},
	{0xCA36AC75CD4A66C3ULL, "vxht_mansion2_21k_0_0.vlm"},
	{0xCA36BEF487C6C800ULL, "vxca_caverns2_71b_0_0.vlm"},
	{0xCA36C39328C994A9ULL, "vxw_encounter2_camp_89_0_0.vlm"},
	{0xCA3727E4BC3FE56FULL, "pro000_captain_wright6004405.cl"},
	{0xCA377E4E87CFC272ULL, "vxht_brothel_n_6bg_0_0.vlm"},
	{0xCA37E5DBC1CC51E9ULL, "73116562.ogg"},
	{0xCA37F4C7EE7FAF98ULL, "vxht_mansion2b_2or_0_0.vlm"},
	{0xCA38781D71D3E06CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05a.phy"},
	{0xCA3931A3F820C7F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_wound.evt"},
	{0xCA3939234A923612ULL, "18960392.ogg"},
	{0xCA393F08F5B6AD06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_walk.ani"},
	{0xCA394D3C7108615FULL, "hlac_w_bonepitmining_a2_393219_0x.dds"},
	{0xCA3974B20C7A163DULL, "dae000_arishok6080508.cl"},
	{0xCA39DD722DCE5F05ULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.anb"},
	{0xCA3AD7E6EE290E6BULL, "dae011_corff_sta6142468.cl"},
	{0xCA3B11ACEA58BC88ULL, "vxlt_shop_dl_0_0.vlm"},
	{0xCA3B7B626F2CA44EULL, "6165163_m.fxe"},
	{0xCA3B7FCAA9DCC0AEULL, "vxlt_0_lowtown_a2_n_c_gnw_0_0.vlm"},
	{0xCA3B953C8DE699C9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_higharmoire_n.dds"},
	{0xCA3C27C11CEC2E90ULL, "vxw_encounter1_is_0_1.vlm"},
	{0xCA3C29BD25BA02A0ULL, "628292188.ogg"},
	{0xCA3CB71A67B12052ULL, "643152120.ogg"},
	{0xCA3CE8DFBAD7554CULL, "209326820.ogg"},
	{0xCA3DB677957A25BEULL, "mcr310_argument6075831.cl"},
	{0xCA3DB8FF2F27F17AULL, "vxca_caverns1_ikk_0_0.vlm"},
	{0xCA3F16B82A6CDB11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pointytalk.ani"},
	{0xCA3F3E7D7B198C7BULL, "128239415.ogg"},
	{0xCA400F7D0832A997ULL, "6105937_m.fxe"},
	{0xCA4039605F4030BFULL, "270826556.ogg"},
	{0xCA403C51A4B2A912ULL, "vxlt_0_lowtown_a2_n_c_f57_0_0.vlm"},
	{0xCA4075E7D22BF51DULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicechair.mao"},
	{0xCA4079E7D22BFC56ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicechair.mmh"},
	{0xCA4083E7D22C0D54ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicechair.msh"},
	{0xCA416077A766FFE7ULL, "vxw_bonepit_1s1_0_0.vlm"},
	{0xCA4171BB6FE160BCULL, "dae000_meredith6085235.cl"},
	{0xCA41C91DCB462683ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_1x.dds"},
	{0xCA42155A575A9DD0ULL, "zz_followers_banter6158391.cl"},
	{0xCA428FC25FDE28BFULL, "182472207.ogg"},
	{0xCA42E43E94778BB3ULL, "492097158.ogg"},
	{0xCA431124FDFCE240ULL, "6090841_m.fxe"},
	{0xCA4327CFAA6026CCULL, "w_woundedcoast_24_0_ll.mmh"},
	{0xCA4358A44919CFF0ULL, "dae011_gossips6143007.cl"},
	{0xCA436923B9E2141EULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.mmh"},
	{0xCA438C06768AD707ULL, "vxlt_undercitydungeon_18c_0_0.vlm"},
	{0xCA43CDAEE258077DULL, "follower_banter6149190.cl"},
	{0xCA43D399485A6088ULL, "100015722.ogg"},
	{0xCA44062DBF884C15ULL, "fen110_fenris6101954.cl"},
	{0xCA441AEA9652D59EULL, "853573641.ogg"},
	{0xCA443EAA03BE2C73ULL, "breadoven_d.dds"},
	{0xCA447CCCA86C0867ULL, "mag291_ambient_26218719.cl"},
	{0xCA44BE9273E13F8CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masb_0d.dds"},
	{0xCA44FFB07106620DULL, "6110862_f.fxe"},
	{0xCA4500D67F9CF2B3ULL, "follower_banter6167220.cl"},
	{0xCA4535701ABFAD29ULL, "vxw_encounter2_camp_38_0_0.vlm"},
	{0xCA45DE40E558ED0FULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_c.mmh"},
	{0xCA4621B82A730CB3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pointytalk.gad"},
	{0xCA46340FF4C64E13ULL, "w_sundermount_14_0.rml"},
	{0xCA4666A4E075C818ULL, "vxht_mansion2b_2hs_0_0.vlm"},
	{0xCA46CF0A703B14E0ULL, "cd_act1_rockwraith_intro_demo_rogue.bnk"},
	{0xCA472708F5C2C277ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_walk.evt"},
	{0xCA474CBE76FD09DFULL, "t1_mul_ro2.tnt"},
	{0xCA47816499EBD1F7ULL, "179598127.ogg"},
	{0xCA483C0AD1BA545EULL, "art\\characters\\creatures\\dragon\\animation\\rst_tw_v1.ani"},
	{0xCA48ADD046963EBDULL, "6054876_f.fxe"},
	{0xCA48D0542269F5A2ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stonetree_d.dds"},
	{0xCA48EF1EE24B047CULL, "lgt320pt_mis_resp_deman.ncs"},
	{0xCA49089378C27772ULL, "vxw_blightlands_9j_0_0.vlm"},
	{0xCA490D84D0A161D2ULL, "mag121_idunna6083958.cl"},
	{0xCA492B68743D941BULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden6.msh"},
	{0xCA493913DE872F80ULL, "gen_im_trs_dwv_junk3_01.uti"},
	{0xCA4946132A42A2D4ULL, "vxlt_backalley_n_1nv_0_0.vlm"},
	{0xCA494968743DC705ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden6.mmh"},
	{0xCA498CE3871420A2ULL, "seb201_lady_harimann6177428.cl"},
	{0xCA49B2DBAE2B08FAULL, "vxlt_0_lowtown_a2_e82_0_0.vlm"},
	{0xCA49D3E28659B6B6ULL, "w_bonepitmining_a2_3_0.rml"},
	{0xCA4A7797202E497BULL, "vxht_bartrand_n_3sd_0_0.vlm"},
	{0xCA4AC8F2BD112F15ULL, "hlac_ht_mansion2b_818_32x.dds"},
	{0xCA4ACDAE89DAD10FULL, "6170982_m.fxe"},
	{0xCA4BDEAA2C8370B8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_rust_d.dds"},
	{0xCA4C2ADC211E09FAULL, "vxca_caverns1_ind_0_0.vlm"},
	{0xCA4C3BE9E7A12521ULL, "vxht_brothel_n_4xl_0_0.vlm"},
	{0xCA4C64A5B278036BULL, "824006159.ogg"},
	{0xCA4C82DE15B7A9BBULL, "art\\characters\\creatures\\ragedemon\\animation\\ablts_v2.ani"},
	{0xCA4D1EB82A792618ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pointytalk.evt"},
	{0xCA4D2808F5C72F00ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_walk.gad"},
	{0xCA4D92F1415BE092ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_male.gad"},
	{0xCA4DFA7AA6318215ULL, "hla_ht_0_hightown_n_65536_0y.dds"},
	{0xCA4E5BA2D6DE6061ULL, "20837014.ogg"},
	{0xCA4EA0F808434D88ULL, "vxlt_gallowscourtyard_73r_0_0.vlm"},
	{0xCA4EA54EB251BC1DULL, "65455411.ogg"},
	{0xCA4EE859D73CC683ULL, "409044284.ogg"},
	{0xCA50549DEEF62C8BULL, "vxlt_0_lowtown_docks_b6k_0_0.vlm"},
	{0xCA50610D49BCCD21ULL, "dae300_bran6142928.cl"},
	{0xCA50CFAA17445E0EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2a.mmh"},
	{0xCA50DBAA174472A5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2a.mao"},
	{0xCA50E8049FA5AE32ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkback.gad"},
	{0xCA50E9AA17448AFCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar2a.msh"},
	{0xCA514823B9EE19DDULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.anb"},
	{0xCA51D5B6AC270BB1ULL, "hlac_lt_0_lowtown_a3_n_c_-65534_0x.dds"},
	{0xCA51FC8E371A0045ULL, "sw_p_glow.mao"},
	{0xCA52494D39EC19D9ULL, "hlac_w_bonepit_262146_0z.dds"},
	{0xCA52569167D9EA5CULL, "681675962.ogg"},
	{0xCA52976C28D097ADULL, "6208058_m.fxe"},
	{0xCA529C38F7A141F8ULL, "6055006_m.fxe"},
	{0xCA533A7B4028ADADULL, "vxdr_0_deeproads_alj_0_0.vlm"},
	{0xCA542D78AA666EFAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_painting_d.dds"},
	{0xCA545994C5C28FAEULL, "l_dae230ar_undercity_dt.bnk"},
	{0xCA546955632401FDULL, "ser221_gascard6128797.cl"},
	{0xCA5495F1416204C9ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_male.evt"},
	{0xCA552117EC653AA9ULL, "775530482.ogg"},
	{0xCA552AF9F87B6884ULL, "mcr370_assignments6176722.cl"},
	{0xCA553CA4F1777CD7ULL, "6016466_m.fxe"},
	{0xCA55540AD1C5072FULL, "art\\characters\\creatures\\dragon\\animation\\rst_tw_v1.evt"},
	{0xCA556639B07CA685ULL, "6072660_m.fxe"},
	{0xCA568073701FEC20ULL, "mer320_hubert6096171.cl"},
	{0xCA5682CCC3B4D3C3ULL, "gen00fl_fenris6126385.cl"},
	{0xCA56B7AA9C9728D1ULL, "318042353.ogg"},
	{0xCA56FE18B2F9D721ULL, "vxlt_0_lowtown_a2_h2f_0_0.vlm"},
	{0xCA581DB268C3C1A2ULL, "vxht_mansion2b_3d4_0_0.vlm"},
	{0xCA58B9F85E3965E9ULL, "llac_lt_0_lowtown_a2_n_-2_2.dds"},
	{0xCA5906B7E0E33B3CULL, "6143107_m.fxe"},
	{0xCA59280F6AED4BA1ULL, "6123391_m.fxe"},
	{0xCA598ADE15C241B2ULL, "art\\characters\\creatures\\ragedemon\\animation\\ablts_v2.evt"},
	{0xCA59A69B1222660FULL, "vxw_encounter2_r5_0_0.vlm"},
	{0xCA59E878BD5C0418ULL, "1062879085.ogg"},
	{0xCA59EC5976A6043EULL, "llac_lt_0_lowtown_docks_262143_0.dds"},
	{0xCA5A6A79990F8B11ULL, "vxca_cavern1_haunted_j4g_0_0.vlm"},
	{0xCA5AB6604548494BULL, "6116494_m.fxe"},
	{0xCA5AF48AE2A773F0ULL, "6119827_m.fxe"},
	{0xCA5B2BE85CBD672DULL, "hlac_ca_caverns1_mines_24424_16z.dds"},
	{0xCA5B4C24AFD52C21ULL, "523661745.ogg"},
	{0xCA5B933734935D7BULL, "vxlt_gallowstemplar_n_5ul_0_0.vlm"},
	{0xCA5BB9D50E99AACEULL, "vxw_encounter2_lf_0_0.vlm"},
	{0xCA5C550AD1CB27D8ULL, "art\\characters\\creatures\\dragon\\animation\\rst_tw_v1.gad"},
	{0xCA5CAACECAFC2AB6ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.anb"},
	{0xCA5CD55F80E5B574ULL, "378686947.ogg"},
	{0xCA5D43338E2788FBULL, "dae011_conspiracy_mob6149589.cl"},
	{0xCA5EA46D52C54300ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_walltypexa.phy"},
	{0xCA5F1A4FF747C5B5ULL, "ach_comp_level_1.dds"},
	{0xCA5F2CF55A77E457ULL, "lowtown_fem_urchin_mob6165959.cl"},
	{0xCA5F5CD457C1DBFEULL, "w_bonepit_50_0.rml"},
	{0xCA5F6143B2220327ULL, "736561643.ogg"},
	{0xCA5FD5C2DB7B6B34ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dssa_0d.dds"},
	{0xCA5FF5085DCD74DBULL, "vxlt_0_lowtown_docks_n_aqz_0_0.vlm"},
	{0xCA5FF7FAAED81455ULL, "73221709.ogg"},
	{0xCA606176C427FECFULL, "vxca_caverns1_ea8_0_0.vlm"},
	{0xCA60A9DE15C89545ULL, "art\\characters\\creatures\\ragedemon\\animation\\ablts_v2.gad"},
	{0xCA60E00B2F73FCFAULL, "vxht_0_hightown_n_bz4_0_0.vlm"},
	{0xCA610E15E9975A78ULL, "6052106_m.fxe"},
	{0xCA6153D4191B82FFULL, "qun211_party_comment6072269.cl"},
	{0xCA61913FFFA4A5DFULL, "6115860_m.fxe"},
	{0xCA61DE722DEFB346ULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.mmh"},
	{0xCA61F8E42594EF03ULL, "fxe_fluffs_p.mao"},
	{0xCA6214C2DCAE1102ULL, "575888832.ogg"},
	{0xCA627DF1416E1938ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_male.ani"},
	{0xCA627E593E0ED275ULL, "lt_0_lowtown_44_0_lc.mmh"},
	{0xCA62BEC31D60E191ULL, "gen_im_arm_cht_med_lore_01.uti"},
	{0xCA62D2D788E4784BULL, "vxlt_0_lowtown_a2_n_gj4_0_0.vlm"},
	{0xCA63D5C36D113A24ULL, "256906958.ogg"},
	{0xCA64D3049FB633D8ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\walkback.ani"},
	{0xCA64E5EFF9F2B400ULL, "llac_w_bonepitmining_a2_262149_0.dds"},
	{0xCA65046AE8748C53ULL, "688113855.ogg"},
	{0xCA652E230888CD67ULL, "bld_dri_big.mao"},
	{0xCA6556918BD2D30AULL, "hlac_lt_undercityserialkiller_-1_28x.dds"},
	{0xCA6611CA17CC2D0CULL, "gen_im_npc_dwf_arm_boo_wara.uti"},
	{0xCA66162528F724DAULL, "hlac_lt_0_lowtown_a3_n_c_0_2y.dds"},
	{0xCA66274D38D9D2D9ULL, "6136137_m.fxe"},
	{0xCA66496CBE526032ULL, "hf_apostate01.mor"},
	{0xCA66728ED2D5D0F7ULL, "fen110_fenris6118107.cl"},
	{0xCA66AFB3FD746C3BULL, "vxht_keep_5h7_0_0.vlm"},
	{0xCA6757C23FE9543FULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsarch.msh"},
	{0xCA675DC23FE95EE1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsarch.mmh"},
	{0xCA6769C23FE97282ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinsarch.mao"},
	{0xCA6799EC5CE36421ULL, "6128291_m.fxe"},
	{0xCA67E63FA498232CULL, "vxlt_gallowstemplar_n_4t1_0_0.vlm"},
	{0xCA6832C8A33CF25DULL, "632835660.ogg"},
	{0xCA68997C16056CE6ULL, "116376503.ogg"},
	{0xCA68A8F0D156EC1FULL, "follower_banter6161121.cl"},
	{0xCA68F3F59B1CB544ULL, "vxw_encounter3_3z_0_0.vlm"},
	{0xCA6918A3B8CFC1F6ULL, "vxdr_primevaltomb_4hf_0_0.vlm"},
	{0xCA692E53AC758CFEULL, "787658294.ogg"},
	{0xCA69ADCECB06BA21ULL, "art\\vfx\\cinematic\\vfx_meredith_power_cin_c.mmh"},
	{0xCA6A110DD2EB876DULL, "6020429_m.fxe"},
	{0xCA6AA3785DDA7399ULL, "vxlt_0_lowtown_a2_eib_0_0.vlm"},
	{0xCA6B3843E5EFCB98ULL, "hlac_w_bonepitdestroyed_a3_4_0x.dds"},
	{0xCA6B84B1EC2F7E88ULL, "vxlt_gallowscourtyard_79p_0_0.vlm"},
	{0xCA6B89BA8DEE37BBULL, "dae000_arishok6078431.cl"},
	{0xCA6BF3B40DA6054AULL, "genip_station_poisons.utp"},
	{0xCA6C6A3C714A2CD0ULL, "6119405_m.fxe"},
	{0xCA6C7FE05B59E392ULL, "mag360cr_hostage_varric.utc"},
	{0xCA6C9B08BE2604FBULL, "hlac_lt_0_lowtown_docks_n_-2_6y.dds"},
	{0xCA6D044CF4085699ULL, "vxht_keep_n_6gd_0_0.vlm"},
	{0xCA6D313583A2E0CFULL, "6187940_m.fxe"},
	{0xCA6DBE5DC778D50AULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.dura.gad"},
	{0xCA6DDF40E57A414CULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_c.anb"},
	{0xCA6E2B363CAEF087ULL, "w_woundedcoast_25_0.anb"},
	{0xCA6F00FC6725178FULL, "vxlt_undercitydungeon_1yt_0_0.vlm"},
	{0xCA6F0CF27726FFE3ULL, "qcr220_orsino6041812.cl"},
	{0xCA6F51E8216A157AULL, "vxht_0_hightown_n_be5_0_0.vlm"},
	{0xCA6F98936300CDBFULL, "fx_dirt_blast.dds"},
	{0xCA70252DE98216D5ULL, "ave200_aveline_bar6116600.cl"},
	{0xCA7041872FAB038FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.cess.gad"},
	{0xCA704AF4F2DCA0ABULL, "hf_zmother_h1s3.mor"},
	{0xCA70D744F14E3946ULL, "mer210_smith6051393.cl"},
	{0xCA70FC5E3F7B440BULL, "6164200_m.fxe"},
	{0xCA7162AC5FC9D935ULL, "6020472_m.fxe"},
	{0xCA7169890CF90F85ULL, "6092153_f.fxe"},
	{0xCA7205C715AEA80AULL, "216028797.ogg"},
	{0xCA72343489DA2EB2ULL, "blood_clu_l2.mao"},
	{0xCA72D06CF9D5AA35ULL, "llac_lt_0_lowtown_1_7.dds"},
	{0xCA72FFFA224289BDULL, "docks_scene_dockworkers6176478.cl"},
	{0xCA7325472BD06941ULL, "hlac_lt_0_lowtown_a2_131073_0y.dds"},
	{0xCA73264CCF0CC181ULL, "55068810.ogg"},
	{0xCA740A52176B0A93ULL, "6183827_m.fxe"},
	{0xCA745CFC97489179ULL, "follower_banter6156033.cl"},
	{0xCA74A15DC77EC2A1ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.dura.evt"},
	{0xCA74AA7FB29F9845ULL, "vxlt_0_lowtown_a2_e89_0_0.vlm"},
	{0xCA74CD6EDC2689A2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masc_0n.dds"},
	{0xCA74E075FED16ADFULL, "6147661_m.fxe"},
	{0xCA75290B2EE2763EULL, "hlac_lt_0_lowtown_docks_196607_3x.dds"},
	{0xCA753BC609EFB742ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.impa.gad"},
	{0xCA7566EA212E885CULL, "vxht_0_hightown_a2_c1i_0_0.vlm"},
	{0xCA75869F3B892E23ULL, "129546329.ogg"},
	{0xCA7602141224E98AULL, "vxca_cavern1_haunted_imp_0_0.vlm"},
	{0xCA7656872FAF933CULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.cess.evt"},
	{0xCA771B3CD7EB95C0ULL, "lt_0_lowtown_a3_n_c_33_0_ll.mmh"},
	{0xCA77991E0CE0C3EEULL, "839257967.ogg"},
	{0xCA77A378BF5F4112ULL, "mag160_feynriel6107430.cl"},
	{0xCA77D5C9A1C92D99ULL, "vxdr_dungeon_238_0_0.vlm"},
	{0xCA77DA3FC18764D6ULL, "gen_im_trs_tmp_one_axe_r1.uti"},
	{0xCA7825D0FE5CFF3BULL, "vxht_0_hightown_c0m_0_0.vlm"},
	{0xCA786A24EA52DED9ULL, "zz_isa_debug6133240.cl"},
	{0xCA7876F1187F8E7BULL, "vxlt_gallowstemplar_n_5cn_0_0.vlm"},
	{0xCA7882EAC4E19FEDULL, "6027529_m.fxe"},
	{0xCA789D8EF285DBCEULL, "vxw_encounter2_rx_0_0.vlm"},
	{0xCA78FC80FD5D0176ULL, "mag100_cullen6206255.cl"},
	{0xCA79467EED7468A0ULL, "6205089_m.fxe"},
	{0xCA794F1D720B3137ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05a.mmh"},
	{0xCA79651D720B56A9ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos05a.msh"},
	{0xCA7B35FC4257694DULL, "one350_variel6087039.cl"},
	{0xCA7B7629E5FE6F19ULL, "vxht_keep_n_6bk_0_0.vlm"},
	{0xCA7B95CC1A7E3428ULL, "gen00fl_bethany6005731.cl"},
	{0xCA7BD549427C795FULL, "vxht_0_hightown_bg9_0_0.vlm"},
	{0xCA7BD8BE59D13E2CULL, "386829624.ogg"},
	{0xCA7BDD25F826100FULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashshardc.phy"},
	{0xCA7BE7E88ED616E9ULL, "zz_followers_banter6159702.cl"},
	{0xCA7C5EC609F61139ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.impa.evt"},
	{0xCA7C76655A7F04ABULL, "6135277_m.fxe"},
	{0xCA7D115DDD78363AULL, "6077941_m.fxe"},
	{0xCA7D20F79AF18EBAULL, "mag210_arianni6024975.cl"},
	{0xCA7D6749F960ECAFULL, "art\\levels\\ztest\\chad\\textures\\prp_orlaischair_sa.dds"},
	{0xCA7D6FF105FE18BDULL, "6050916_m.fxe"},
	{0xCA7E4B0710F4C026ULL, "6139481_m.fxe"},
	{0xCA7E4B7F71A83482ULL, "llac_ht_0_hightown_-65537_0.dds"},
	{0xCA7E567A53EB95EAULL, "hlac_lt_darktown_-1_52y.dds"},
	{0xCA7EAB64CFA015FFULL, "hm_gamlen_h1s3.mor"},
	{0xCA7EC51E0245BE89ULL, "hlac_w_bonepitdestroyed_a3_458752_0x.dds"},
	{0xCA7F2E059C603633ULL, "vxw_encounter2_cb_0_0.vlm"},
	{0xCA7F70C94C07F4AEULL, "6153922_m.fxe"},
	{0xCA7F98CF546E6BEDULL, "zz_cassandrafront.cut"},
	{0xCA7F9ED2251C4862ULL, "vxht_keep_5gw_0_0.vlm"},
	{0xCA8053FE18B9A2ECULL, "vxw_sundermount_30k_0_1.vlm"},
	{0xCA807CE4912741FFULL, "6160860_m.fxe"},
	{0xCA810C2A7D6DC217ULL, "hlac_w_bonepit_5_0x.dds"},
	{0xCA8110E7E16F9EDBULL, "6080623_m.fxe"},
	{0xCA81222514D3D5F6ULL, "vxlt_gallowstemplar_5uw_0_1.vlm"},
	{0xCA81995DC7893FF0ULL, "art\\vfx\\combat\\vfx_w_def_dam_reduc_loop_c.dura.ani"},
	{0xCA820FD5EF1899D3ULL, "515504087.ogg"},
	{0xCA8229E5CE967BF8ULL, "6093256_f.fxe"},
	{0xCA8251A4302CFD01ULL, "lowtown_male_beggar_sta6151082.cl"},
	{0xCA825EFEA99EB5F2ULL, "222165864.ogg"},
	{0xCA82EEF44C15A85FULL, "6087820_m.fxe"},
	{0xCA830D86F21B04D4ULL, "vxlt_backalley_n_1zz_0_0.vlm"},
	{0xCA8367BFB4504B82ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_fenris_fenrisrom_alt.gad"},
	{0xCA836ED5552B5438ULL, "6104754_f.fxe"},
	{0xCA83C16D52E49095ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_walltypexa.msh"},
	{0xCA83C36D52E4940BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_walltypexa.mmh"},
	{0xCA83CF6D52E4A8A8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_walltypexa.mao"},
	{0xCA83E1B2169ED9DCULL, "hlac_dr_0_deeproads_5841_4x.dds"},
	{0xCA83EA54E68B4D52ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_drink.gad"},
	{0xCA8429D1B1B28C1DULL, "zz_mer_debug6020473.cl"},
	{0xCA845E872FBBDE55ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t1_c.cess.ani"},
	{0xCA8473F1041B020AULL, "vxca_cavern1_haunted_ijq_0_0.vlm"},
	{0xCA84C8836E37E1F7ULL, "mer240_jansen6050318.cl"},
	{0xCA84D2641777C912ULL, "llac_lt_0_lowtown_n_1_3.dds"},
	{0xCA851F574F66E517ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_lgtc_0t.dds"},
	{0xCA853672CBDEA2E6ULL, "dae000ip_crowd_mf2_talk11.utp"},
	{0xCA8559B17F30285BULL, "vxlt_0_lowtown_a2_n_eiw_0_0.vlm"},
	{0xCA8615F6A941BFF7ULL, "zz_approval_debug6003097.cl"},
	{0xCA8697ADF05317BBULL, "vxht_0_hightown_c7l_0_0.vlm"},
	{0xCA86C6AF4FC9B4ACULL, "6059396_m.fxe"},
	{0xCA86D5D43BD758B4ULL, "152403104.ogg"},
	{0xCA86D9EEFBF7C73FULL, "6040818_m.fxe"},
	{0xCA871458E19699BEULL, "254791017.ogg"},
	{0xCA8732110ABFFC54ULL, "6171054_m.fxe"},
	{0xCA879E0FBA36437DULL, "dae160st_saemus1.stg"},
	{0xCA883FBC4B72D9FFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\set_tarp_dp.dds"},
	{0xCA89381B21DDA8FDULL, "w_woundedcoast_1_0.rml"},
	{0xCA89506535480EB0ULL, "vxw_bonepit_1i8_0_0.vlm"},
	{0xCA8966C60A00A928ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s3_c.impa.ani"},
	{0xCA89ED54E68FBD89ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_drink.evt"},
	{0xCA89FC0AA024BC06ULL, "6153412_m.fxe"},
	{0xCA8A13FC2C39277BULL, "zz_t_traler_shot3.cut"},
	{0xCA8A8F4D891E7EADULL, "hlac_lt_darktown_-1_44y.dds"},
	{0xCA8A941EF92E2041ULL, "6177840_m.fxe"},
	{0xCA8AADA1412F1B61ULL, "vxca_caverns1_id1_0_0.vlm"},
	{0xCA8ABD9D55138FE2ULL, "6103668_m.fxe"},
	{0xCA8AD062B93AF6B3ULL, "vxlt_0_lowtown_a3_n_c_hf8_0_0.vlm"},
	{0xCA8AE319C237504AULL, "mag360_thrask6064091.cl"},
	{0xCA8AE3802B694743ULL, "290071446.ogg"},
	{0xCA8B010E2C5D042FULL, "mag100_karras6035295.cl"},
	{0xCA8B50D457E70276ULL, "w_bonepit_50_0.mao"},
	{0xCA8B5CD457E71615ULL, "w_bonepit_50_0.mmh"},
	{0xCA8B5ED457E7198BULL, "w_bonepit_50_0.msh"},
	{0xCA8BCE9213CA7A37ULL, "vxlt_0_lowtown_docks_b80_0_1.vlm"},
	{0xCA8CA5CEB20A8D4EULL, "6105698_f.fxe"},
	{0xCA8D04F47CAC2900ULL, "zz_qun_debug6036929.cl"},
	{0xCA8D111E082DF84EULL, "6168779_m.fxe"},
	{0xCA8D1545B3FFA732ULL, "6055630_f.fxe"},
	{0xCA8D298FE47FCA04ULL, "hlac_lt_backalley_n_0_7z.dds"},
	{0xCA8D3754350D0FB7ULL, "6129401_m.fxe"},
	{0xCA8D6F2B5AB0FF9BULL, "85082604.ogg"},
	{0xCA8D9FE28E564EB3ULL, "vxw_encounter2_ne_0_0.vlm"},
	{0xCA8DCC3F4EA380C0ULL, "175153763.ogg"},
	{0xCA8EA993631B9B51ULL, "fx_dirt_blast.mao"},
	{0xCA8EB0D2AB041D98ULL, "6070182_m.fxe"},
	{0xCA8EB529A672B00EULL, "vxdr_primevaltomb_911_0_1.vlm"},
	{0xCA8ECA04C53F06DEULL, "862802318.ogg"},
	{0xCA8FA03A7F413F1CULL, "follower_banter6155853.cl"},
	{0xCA906039C7CEE888ULL, "187764539.ogg"},
	{0xCA9185DC5A5C4D4FULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.mmh"},
	{0xCA91B45CB07141E6ULL, "vxw_encounter2_camp_j0_0_0.vlm"},
	{0xCA91C3CED39F90F0ULL, "llac_lt_0_lowtown_docks_a3_n_-131073_0.dds"},
	{0xCA91D1C51E7E3B2FULL, "art\\characters\\weapons\\longswords\\w_lsw_ls107a.phy"},
	{0xCA91F64B3CB591D6ULL, "character_character_p133.psh.fxo.dx11"},
	{0xCA9232E6DFA848EEULL, "vxlt_0_lowtown_docks_bbv_0_0.vlm"},
	{0xCA927153B2442884ULL, "ser221_party_comments6122879.cl"},
	{0xCA92E5CDF6111A8AULL, "vxca_cavern1_haunted_igr_0_0.vlm"},
	{0xCA9374097A44A58FULL, "248874491.ogg"},
	{0xCA939705BCD35498ULL, "541607234.ogg"},
	{0xCA93B6E1EC5F85B3ULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_ext_c.anb"},
	{0xCA941935E6222A8EULL, "vxht_keep_n_3zy_0_0.vlm"},
	{0xCA941BEE4584D123ULL, "lt_0_lowtown_docks_a3_n_209_0.msh"},
	{0xCA9429EE4584E97DULL, "lt_0_lowtown_docks_a3_n_209_0.mmh"},
	{0xCA942DEE4584EF8EULL, "lt_0_lowtown_docks_a3_n_209_0.mao"},
	{0xCA94D88EA05CD97CULL, "gen_im_npc_elf_arm_glv_dala.uti"},
	{0xCA94FDADEEF22AB7ULL, "69571351.ogg"},
	{0xCA95007B137867D1ULL, "mag311_nyssa6041959.cl"},
	{0xCA95138EE7F05816ULL, "hlac_ca_caverns2prc_5802_18y.dds"},
	{0xCA9582CF153BB1F0ULL, "6093530_m.fxe"},
	{0xCA96001B94CA8831ULL, "art\\vfx\\creatures\\fxc_succubus_c.cess.ani"},
	{0xCA9606BED0E52D82ULL, "vxw_encounter2_7k_0_0.vlm"},
	{0xCA97174BC1A715F9ULL, "hlac_w_sundermount_393211_0y.dds"},
	{0xCA9728363CD1F088ULL, "w_woundedcoast_25_0.mmh"},
	{0xCA9732DE1453C5D1ULL, "llac_lt_0_lowtown_-196609_0.dds"},
	{0xCA9734363CD2052BULL, "w_woundedcoast_25_0.mao"},
	{0xCA9746363CD223B2ULL, "w_woundedcoast_25_0.msh"},
	{0xCA9781ACC6900700ULL, "vxca_cavern1_haunted_cn5_0_0.vlm"},
	{0xCA9790E8A14B57E6ULL, "6165233_m.fxe"},
	{0xCA979F3A8372B7D6ULL, "478212936.ogg"},
	{0xCA97D0F9C8A88981ULL, "vxht_0_hightown_a2_c18_0_0.vlm"},
	{0xCA97D554E69BD2F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_drink.ani"},
	{0xCA97D619DE8ECDAFULL, "qun111_ketojan6040148.cl"},
	{0xCA97DA77B7160905ULL, "6035870_m.fxe"},
	{0xCA9807F44392A3ECULL, "llac_lt_undercityserialkiller_-1_0.dds"},
	{0xCA986C644E8D9792ULL, "6020541_m.fxe"},
	{0xCA9892BFB462F168ULL, "art\\cinematics\\act3\\romances\\fenris\\cs_fenris_fenrisrom_alt.ani"},
	{0xCA98CBA249700D97ULL, "vxdr_dungeon_1mo_0_0.vlm"},
	{0xCA991F7E332533E1ULL, "vxca_caverns1_io0_0_0.vlm"},
	{0xCA999E7EDAE07729ULL, "vxdr_primevalentrance_27n_0_0.vlm"},
	{0xCA99A329F6184D7DULL, "119166049.ogg"},
	{0xCA9ACDE1F977F083ULL, "vxht_0_hightown_arm_0_0.vlm"},
	{0xCA9B0263C8C59E00ULL, "810978050.ogg"},
	{0xCA9B274AD30E8CFCULL, "847617294.ogg"},
	{0xCA9B3DE46FCA542AULL, "dae000_arishok6067832.cl"},
	{0xCA9B8214AE7F2E8DULL, "zz_followers_banter6159618.cl"},
	{0xCA9BC10D00827328ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_3z.dds"},
	{0xCA9BE39D39442F1EULL, "vxlt_0_lowtown_docks_c0e_0_0.vlm"},
	{0xCA9C03DA722AA8A7ULL, "112145973.ogg"},
	{0xCA9C11BF804C6733ULL, "vxw_encounter2_id_0_0.vlm"},
	{0xCA9C5E090C0AEEB4ULL, "dr_dungeon_3218.rml"},
	{0xCA9CACD6539BA43CULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower1_isabellasreturn.ani"},
	{0xCA9CB05E14D6FDF5ULL, "6147164_m.fxe"},
	{0xCA9CCA02D354E1AFULL, "llac_lt_0_lowtown_docks_a3_n_-65540_0.dds"},
	{0xCA9CE52247ED313FULL, "169560831.ogg"},
	{0xCA9CFD5DE828C8A0ULL, "vxdr_primevalentrance_19z_0_0.vlm"},
	{0xCA9D0B1B94D0B9D3ULL, "art\\vfx\\creatures\\fxc_succubus_c.cess.gad"},
	{0xCA9D22AF9CFA3618ULL, "6075282_m.fxe"},
	{0xCA9D459F7444FCD5ULL, "6076808_m.fxe"},
	{0xCA9D9DFA7A6D253EULL, "vxht_bartrand_n_65n_0_0.vlm"},
	{0xCA9DCD17CEE83238ULL, "sib131_letter_react6127325.cl"},
	{0xCA9DDB7F91E66403ULL, "hlac_w_encounter2_camp_196610_0x.dds"},
	{0xCA9E25ED04A13F0FULL, "505808595.ogg"},
	{0xCA9E3C7FA734EADBULL, "gen00fl_sebastian6195814.cl"},
	{0xCA9E466FE4DD589AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_masf_0s.dds"},
	{0xCA9E5CD3374B83CEULL, "538756394.ogg"},
	{0xCA9E678DBA4F7264ULL, "vxw_bonepitdestroyed_a3_2a7_0_0.vlm"},
	{0xCA9ECB89F437C660ULL, "follower_banter6141764.cl"},
	{0xCA9EF10CDC1D9187ULL, "vxlt_0_lowtown_a3_n_c_hsg_0_0.vlm"},
	{0xCA9F2F0FB8C29BD5ULL, "dae000_serendipity6161356.cl"},
	{0xCA9F381807B73E01ULL, "ser221_gascard6128292.cl"},
	{0xCA9F50800EA8F5AAULL, "6149745_m.fxe"},
	{0xCAA03F8721785C60ULL, "6078185_f.fxe"},
	{0xCAA0D3E1EC6A411CULL, "art\\vfx\\combat\\vfx_glb_bld_ds_imp_ext_c.mmh"},
	{0xCAA0F7FB07C98BFFULL, "6149159_m.fxe"},
	{0xCAA1955D2D79C8FBULL, "darktown_male_lowlife_mob6174416.cl"},
	{0xCAA1C7000107767DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gong_sa.dds"},
	{0xCAA23D6B631C595BULL, "vxlt_0_lowtown_a2_n_ewi_0_0.vlm"},
	{0xCAA2A7210900837CULL, "vxlt_foundry_n_12m_0_0.vlm"},
	{0xCAA2CBABF5BCFD12ULL, "hlac_dr_0_deeproads_5845_19z.dds"},
	{0xCAA2D25B9C513447ULL, "vxht_mansion2_24t_0_0.vlm"},
	{0xCAA2DCBD862FF350ULL, "vxlt_0_lowtown_docks_a3_cy6_0_1.vlm"},
	{0xCAA2DDEA1EA87CE6ULL, "vxca_caverns2_6h7_0_1.vlm"},
	{0xCAA3081B94D52038ULL, "art\\vfx\\creatures\\fxc_succubus_c.cess.evt"},
	{0xCAA32937509A161EULL, "art\\vfx\\textures\\vfx_lightbeams.dds"},
	{0xCAA347ED98B6CCD1ULL, "176068841.ogg"},
	{0xCAA367049828B684ULL, "6110907_m.fxe"},
	{0xCAA36B72D05653A8ULL, "t3_shl_pur.tnt"},
	{0xCAA36FFCD1B4D3A5ULL, "gen00fl_carver6104536.cl"},
	{0xCAA3B9D653A1D97EULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower1_isabellasreturn.gad"},
	{0xCAA3FF08671DE3B2ULL, "6151450_m.fxe"},
	{0xCAA4789BC2DB4602ULL, "vxw_encounter2_2t_0_0.vlm"},
	{0xCAA483BF7F713834ULL, "llac_ca_cavern1_haunted_24422_0.dds"},
	{0xCAA49021AD3DA234ULL, "6061367_f.fxe"},
	{0xCAA4A8A726DB2DB7ULL, "191684349.ogg"},
	{0xCAA4C9E560401A4FULL, "dae000ip_crowd_ff4_talk13.utp"},
	{0xCAA52737E0F2EDB2ULL, "hlac_lt_hangedman_-1_19y.dds"},
	{0xCAA552F1D1F166C1ULL, "528641599.ogg"},
	{0xCAA61274785D3017ULL, "781176814.ogg"},
	{0xCAA63BB1861F037AULL, "vxdr_primevalentrance_3b7_0_0.vlm"},
	{0xCAA65EB064CE591DULL, "8529704.ogg"},
	{0xCAA67195E0ADECE0ULL, "cth_cm5.tnt"},
	{0xCAA78126017CF0B3ULL, "hlac_w_woundedcoast_-196601_0z.dds"},
	{0xCAA7E08BCBDA8D3FULL, "mag291_ntbanter6056620.cl"},
	{0xCAA89C3D5528DD0AULL, "vxca_caverns2prc_9xd_0_0.vlm"},
	{0xCAA8DBC7BD9843E9ULL, "6125274_m.fxe"},
	{0xCAA91D8EAF85D89DULL, "506301040.ogg"},
	{0xCAA93FBEEB6E0903ULL, "vxht_0_hightown_aul_0_0.vlm"},
	{0xCAA9543406E3A29BULL, "w_encounter3_158_0.mmh"},
	{0xCAA9603406E3B738ULL, "w_encounter3_158_0.mao"},
	{0xCAA9723406E3D585ULL, "w_encounter3_158_0.msh"},
	{0xCAA9FC5489AD3977ULL, "vxht_0_hightown_c0y_0_0.vlm"},
	{0xCAAA01D4609D203DULL, "gen00fl_sebastian6188560.cl"},
	{0xCAAAA6359014D2E1ULL, "dae000_serendipity6161411.cl"},
	{0xCAAAF468FAFE91E2ULL, "vxht_keep_5pr_0_0.vlm"},
	{0xCAAB73E323D8368FULL, "6108146_f.fxe"},
	{0xCAABE812B99AEDD8ULL, "hlac_ht_0_hightown_a2_65536_1y.dds"},
	{0xCAAC3981B79CFDF7ULL, "vxht_0_hightown_n_cf4_0_0.vlm"},
	{0xCAAC8985182D7664ULL, "hlac_lt_0_lowtown_a2_0_3x.dds"},
	{0xCAACC5E286AE1C03ULL, "w_bonepitmining_a2_3_0.msh"},
	{0xCAACD3E286AE345DULL, "w_bonepitmining_a2_3_0.mmh"},
	{0xCAACD7E286AE3B6EULL, "w_bonepitmining_a2_3_0.mao"},
	{0xCAAD32DAABCBEA2EULL, "6142560_m.fxe"},
	{0xCAAD3C466BB16866ULL, "hlac_ca_caverns1_24426_3x.dds"},
	{0xCAAD6B280176288EULL, "vxw_encounter2_camp_3b_0_1.vlm"},
	{0xCAADAEFCC73F133CULL, "art\\characters\\weapons\\staff\\textures\\w_stf_015a_0n.dds"},
	{0xCAAE8C006D9D4F5FULL, "65466327.ogg"},
	{0xCAAF8214371C0FDBULL, "w_sundermount_26_0.mmh"},
	{0xCAAF8E14371C2478ULL, "w_sundermount_26_0.mao"},
	{0xCAAFA014371C42C5ULL, "w_sundermount_26_0.msh"},
	{0xCAAFAF5FEC99C952ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv1.gad"},
	{0xCAAFCB9E423A2A7FULL, "hlac_w_bonepitmining_a2_2_0y.dds"},
	{0xCAB01174C54027B2ULL, "853797964.ogg"},
	{0xCAB02DC116A35512ULL, "hf_gen_coterie_archer.mor"},
	{0xCAB0402598F36AC6ULL, "6107887_f.fxe"},
	{0xCAB0AF48551271DBULL, "vxlt_0_lowtown_a2_n_exh_0_0.vlm"},
	{0xCAB1409460E0AF17ULL, "qun110_petrice6018531.cl"},
	{0xCAB147E6BED8B57AULL, "vxca_caverns1_i3c_0_0.vlm"},
	{0xCAB15537A7730DB9ULL, "6115619_m.fxe"},
	{0xCAB1A07DF67F1073ULL, "w_bonepitdestroyed_a3_37_0.anb"},
	{0xCAB1C8F567441108ULL, "hlac_ht_mansion2_818_23y.dds"},
	{0xCAB1E36B3603F4A0ULL, "6201456_m.fxe"},
	{0xCAB242B73946CC1BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_avgdesk.mmh"},
	{0xCAB24EB73946E0B8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_avgdesk.mao"},
	{0xCAB260B73946FF05ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_avgdesk.msh"},
	{0xCAB298E20172CD12ULL, "6075666_m.fxe"},
	{0xCAB2DB1EAE707887ULL, "llac_lt_0_lowtown_docks_a3_n_-4_0.dds"},
	{0xCAB2E1776F081446ULL, "vxlt_0_lowtown_a2_n_gni_0_0.vlm"},
	{0xCAB33DD4DF159302ULL, "vxw_blightlands_ab_0_0.vlm"},
	{0xCAB3804C6515A013ULL, "591387163.ogg"},
	{0xCAB3EB5BBE79CEC6ULL, "art\\characters\\creatures\\strider\\animation\\ready.ani"},
	{0xCAB3F48D8185CE1CULL, "zz_docks_oceans.cut"},
	{0xCAB4EB9C25AAAD5AULL, "vxw_bonepit_242_0_0.vlm"},
	{0xCAB50F6939065DABULL, "vxht_0_hightown_a3_cbn_0_0.vlm"},
	{0xCAB51348F612E8DBULL, "6046913_m.fxe"},
	{0xCAB52F71DAE38690ULL, "vxlt_0_lowtown_a3_n_c_eqo_0_0.vlm"},
	{0xCAB54782EC728429ULL, "art\\vfx\\textures\\potion\\vfx_potion_sa.dds"},
	{0xCAB59FCFB82D096CULL, "hlac_dr_0_deeproads_5849_6z.dds"},
	{0xCAB5B25FEC9E3989ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv1.evt"},
	{0xCAB5BD3239789547ULL, "art\\vfx\\combat\\vfx_q_ari_trip_act_c.impa.gad"},
	{0xCAB6454AC94AD247ULL, "vxlt_shop_b9_0_0.vlm"},
	{0xCAB660735EE2DE8CULL, "ht_0_hightown_a3_9_0.phy"},
	{0xCAB6AAF53688169DULL, "rdr111dg_harbormaster.bnk"},
	{0xCAB6B9495D0E8F14ULL, "llac_lt_0_lowtown_a2_n_131071_1.dds"},
	{0xCAB7125146C660A0ULL, "6112220_m.fxe"},
	{0xCAB7D35EAECE23F5ULL, "royalarcher_flankingsneak.dds"},
	{0xCAB83BDD1C2F652BULL, "198177534.ogg"},
	{0xCAB849A05877A7C4ULL, "vxht_0_hightown_n_b8p_0_0.vlm"},
	{0xCAB86E317BA351F7ULL, "vxht_0_hightown_c7x_0_0.vlm"},
	{0xCAB8932B64CC2161ULL, "mag130_danzig.bnk"},
	{0xCAB8AD3A4EF14343ULL, "vxht_mansion2_2br_0_0.vlm"},
	{0xCAB94FB9D558C207ULL, "6013443_m.fxe"},
	{0xCAB98462300D15BBULL, "760097723.ogg"},
	{0xCAB9AE853754B1D7ULL, "724654801.ogg"},
	{0xCABA3326EA6B86BCULL, "vxdr_dungeon_23a_0_0.vlm"},
	{0xCABA86DC5A7F538CULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.anb"},
	{0xCABA8CF5368BCCA5ULL, "rdr111dg_harbormaster.cnv"},
	{0xCABAB05DFBE71E62ULL, "mag370_meredith6044369.cl"},
	{0xCABAC36F2D5A5FBEULL, "hlac_lt_0_lowtown_65536_7y.dds"},
	{0xCABB824FF137DD6EULL, "llac_w_sundermount_65532_0.dds"},
	{0xCABB8AEE9A3250C5ULL, "vxlt_0_lowtown_epd_0_1.vlm"},
	{0xCABB9F06E2316E36ULL, "169295578.ogg"},
	{0xCABBB52B64CE90D1ULL, "mag130_danzig.cnv"},
	{0xCABBCD043611C99AULL, "zz_one_debug6073982.cl"},
	{0xCABBE7C34F4127BCULL, "6180906_m.fxe"},
	{0xCABC02D01F94E7ABULL, "vxlt_gallowstemplar_f_536_0_0.vlm"},
	{0xCABC371B22090015ULL, "w_woundedcoast_1_0.anb"},
	{0xCABC40B48E46A170ULL, "278679391.ogg"},
	{0xCABC656A30E1E276ULL, "gen00fl_bethany6101385.cl"},
	{0xCABCB232397EA154ULL, "art\\vfx\\combat\\vfx_q_ari_trip_act_c.impa.evt"},
	{0xCABCDBF4ACAD5D1BULL, "gen_im_wep_war_shd_act3_04.uti"},
	{0xCABCDCE07E224B43ULL, "596823156.ogg"},
	{0xCABCF7871B393464ULL, "llac_lt_shop_-1_4.dds"},
	{0xCABD13510475BD4FULL, "ave200_aveline_bar6114110.cl"},
	{0xCABD37735EE8B794ULL, "ht_0_hightown_a3_9_0.rml"},
	{0xCABD5C7BA0D89493ULL, "vxlt_0_lowtown_docks_a3_cp7_0_0.vlm"},
	{0xCABDABD982596C36ULL, "vxlt_0_lowtown_docks_c49_0_0.vlm"},
	{0xCABDC20F3F03E54AULL, "gen00fl_isabela6098199.cl"},
	{0xCABDD3872EA0C83CULL, "vxht_keep_n_6b2_0_0.vlm"},
	{0xCABDE75E622A65ABULL, "hlac_ht_0_hightown_65536_2y.dds"},
	{0xCABE6A94EBE9A035ULL, "dae000_pc_home_ambients6208191.cl"},
	{0xCABE859903C947F9ULL, "lgt200cr_tome_abomination.utc"},
	{0xCABEBCC1496A0A02ULL, "w_encounter2_camp_24_0.msh"},
	{0xCABEBEC1496A0DB8ULL, "w_encounter2_camp_24_0.mmh"},
	{0xCABECAC1496A225BULL, "w_encounter2_camp_24_0.mao"},
	{0xCABECBDB56E200A7ULL, "vxca_caverns1_mines_e6h_0_0.vlm"},
	{0xCABEF4E35EE37466ULL, "vxlt_0_lowtown_docks_cot_0_0.vlm"},
	{0xCABF18CA29251CA8ULL, "vxlt_0_lowtown_a2_n_c_exq_0_0.vlm"},
	{0xCABF55C9008F63F4ULL, "328330829.ogg"},
	{0xCABF56836CFC81E8ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_0x.dds"},
	{0xCABF698F3A8FC069ULL, "vxlt_0_lowtown_docks_bf1_0_0.vlm"},
	{0xCABFA37DF68B52EEULL, "w_bonepitdestroyed_a3_37_0.msh"},
	{0xCABFB17DF68B6A3FULL, "w_bonepitdestroyed_a3_37_0.mao"},
	{0xCABFBD7DF68B7EDCULL, "w_bonepitdestroyed_a3_37_0.mmh"},
	{0xCABFC71499692639ULL, "6110919_m.fxe"},
	{0xCAC060FF5148C7C4ULL, "vxht_keep_n_6jb_0_0.vlm"},
	{0xCAC102144E92D3D9ULL, "vxht_0_hightown_bc3_0_0.vlm"},
	{0xCAC1164468DC1B97ULL, "hlac_ca_caverns1_mines_24425_6x.dds"},
	{0xCAC15C55A6C77702ULL, "vxw_encounter2_8v_0_0.vlm"},
	{0xCAC1BEC1496C43CBULL, "w_encounter2_camp_24_0.rml"},
	{0xCAC1D35BBE85E337ULL, "art\\characters\\creatures\\strider\\animation\\ready.evt"},
	{0xCAC1ED61943852B7ULL, "6096007_f.fxe"},
	{0xCAC2659E8A95090CULL, "lowtown_male_alienage2_mob6166555.cl"},
	{0xCAC2910C5456BE67ULL, "vxca_cavern1_haunted_jfk_0_0.vlm"},
	{0xCAC298605C7911F3ULL, "lgt260_eau_tmp_roderick6143710.cl"},
	{0xCAC3351F7A2CA142ULL, "pro000_colonel_ewald6040565.cl"},
	{0xCAC3896B827167FBULL, "hlac_ht_0_hightown_-196610_0z.dds"},
	{0xCAC39A5FECAA4EF8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\idle\\twitchv1.ani"},
	{0xCAC3A5A8E43769A2ULL, "dae300_cullen6140660.cl"},
	{0xCAC3C29445F05B59ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazierchain.msh"},
	{0xCAC3CC9445F06C47ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazierchain.mmh"},
	{0xCAC3D09445F07294ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_brazierchain.mao"},
	{0xCAC41A2988680236ULL, "26158428.ogg"},
	{0xCAC435756A449BD4ULL, "hm_paxley.mor"},
	{0xCAC43D53C3538B8CULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear5.gad"},
	{0xCAC466920D9AF002ULL, "vxlt_0_lowtown_a3_n_c_bz3_0_0.vlm"},
	{0xCAC473D41A1A4E36ULL, "vxht_keep_n_5k4_0_0.vlm"},
	{0xCAC487A26169C9B2ULL, "vxca_caverns1_mines_j4h_0_0.vlm"},
	{0xCAC49B9CED674613ULL, "188852636.ogg"},
	{0xCAC4B727BB40EAC7ULL, "vxlt_shop_e8_0_0.vlm"},
	{0xCAC55BD3527B43C1ULL, "239255520.ogg"},
	{0xCAC57E24C6D350EAULL, "character_character_p49.psh.fxo.dx11"},
	{0xCAC59DFA92118F56ULL, "w_blightlands_55_0_ll.mmh"},
	{0xCAC5B5CECAC396ADULL, "w_blightlands_112_0.mmh"},
	{0xCAC5B9CECAC39DFEULL, "w_blightlands_112_0.mao"},
	{0xCAC5C7CECAC3B533ULL, "w_blightlands_112_0.msh"},
	{0xCAC5CDFA7B2E869EULL, "20392699.ogg"},
	{0xCAC5E8F1EFFD2795ULL, "ser221_gamlen6131965.cl"},
	{0xCAC66DF61BCA982CULL, "vxca_caverns1_mines_j1y_0_0.vlm"},
	{0xCAC67BF02DFD1F89ULL, "443790280.ogg"},
	{0xCAC6D13BBB29847CULL, "hlac_lt_darktown_-1_62x.dds"},
	{0xCAC747693C7571BAULL, "dae000_sandal6148209.cl"},
	{0xCAC79790B7E36305ULL, "one350_variel6085217.cl"},
	{0xCAC7C196C74EDA87ULL, "6018543_m.fxe"},
	{0xCAC803861E219B93ULL, "vxlt_0_lowtown_a2_n_c_ffk_0_0.vlm"},
	{0xCAC822AC94674A89ULL, "follower_banter6169794.cl"},
	{0xCAC89BD74674221BULL, "one200_gold_followers6120894.cl"},
	{0xCAC89FCCBECE931CULL, "hlac_lt_0_lowtown_docks_a3_n_65534_0y.dds"},
	{0xCAC8C5C149726E43ULL, "w_encounter2_camp_24_0.phy"},
	{0xCAC8D45BBE8C03C0ULL, "art\\characters\\creatures\\strider\\animation\\ready.gad"},
	{0xCAC8D73B257563A2ULL, "lgt220_shp_mag_guard6149869.cl"},
	{0xCAC9072015756C13ULL, "vxw_woundedcoast_9a_0_0.vlm"},
	{0xCAC932AD8F387C9AULL, "hlac_ht_bartrand_n_3366_1x.dds"},
	{0xCAC95A6FE346CA5BULL, "vxca_caverns1_mines_i1a_0_0.vlm"},
	{0xCAC980237FEFDDF3ULL, "vxht_brothel_n_3f7_0_0.vlm"},
	{0xCAC9AA3239891D8DULL, "art\\vfx\\combat\\vfx_q_ari_trip_act_c.impa.ani"},
	{0xCAC9AF673C59B4ADULL, "6208249_m.fxe"},
	{0xCAC9E3FF9351D51CULL, "vxlt_0_lowtown_docks_n_8uh_0_0.vlm"},
	{0xCACA0CDE73FE11DDULL, "vxlt_0_lowtown_a3_n_c_gx3_0_0.vlm"},
	{0xCACA2F6F8BADE922ULL, "100278503.ogg"},
	{0xCACA3C53C357F58AULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fear5.ani"},
	{0xCACA5440A623BE14ULL, "isa200cr_tevinter_mage.utc"},
	{0xCACA6602CCB7D20FULL, "vxlt_0_lowtown_a3_n_c_h2h_0_0.vlm"},
	{0xCACA754A81B92E45ULL, "s_ballista_04.dds"},
	{0xCACA84DF7C2404AFULL, "vxw_encounter2_43_0_0.vlm"},
	{0xCACB688AA78B1B10ULL, "6115368_m.fxe"},
	{0xCACBF1453F7734A4ULL, "vxca_caverns1_mines_gi8_0_0.vlm"},
	{0xCACC069ECB503260ULL, "372691592.ogg"},
	{0xCACC6A3186FC9BFBULL, "344649016.ogg"},
	{0xCACC7207AD10FD44ULL, "vxlt_0_lowtown_a2_eoy_0_0.vlm"},
	{0xCACCECEF5FAF6E79ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_002a_0d.dds"},
	{0xCACCF4BD30A6A894ULL, "plt_cod_cha_viscount.plo"},
	{0xCACD0E4B95EA9D34ULL, "hlac_lt_0_lowtown_a2_n_c_-1_3y.dds"},
	{0xCACD4D6BBC0C5C76ULL, "6117459_f.fxe"},
	{0xCACD5D0705438A55ULL, "dae000cr_tomwise.utc"},
	{0xCACD7D3E308F5012ULL, "plt_prm_cod_sten_sword.plo"},
	{0xCACDE79C81F6B0B4ULL, "vxw_sundermount_31q_0_0.vlm"},
	{0xCACE76274770E719ULL, "6149629_m.fxe"},
	{0xCACF5E71CD274651ULL, "plt_mer210pt_cavern_of_dead.plo"},
	{0xCACF6D44205ACC77ULL, "6165969_f.fxe"},
	{0xCACF73F14088EC59ULL, "vxht_0_hightown_bl2_0_0.vlm"},
	{0xCACF8E904D16B1F7ULL, "vxlt_gallowstemplar_45v_0_0.vlm"},
	{0xCACF994C43A0A1F7ULL, "gen00fl_aveline6139933.cl"},
	{0xCAD01EB2F46C583CULL, "6143316_m.fxe"},
	{0xCAD0218EC301C402ULL, "vxw_blightlands_ol_0_0.vlm"},
	{0xCAD06161AB944FCBULL, "qun160cr_rebel_leader6069289.cl"},
	{0xCAD0EB67BE1FA97EULL, "hlac_ca_caverns2prc_5802_31x.dds"},
	{0xCAD0ED81A63BC79BULL, "static_static_p44.psh.fxo.dx11.assembly"},
	{0xCAD19AFCCB187EA5ULL, "hlac_ht_0_hightown_a3_-196608_2z.dds"},
	{0xCAD1EEE346C27D08ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamb.mmh"},
	{0xCAD1FAE346C291ABULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamb.mao"},
	{0xCAD20CE346C2B032ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamb.msh"},
	{0xCAD20FE135021E96ULL, "1026572625.ogg"},
	{0xCAD261F85102BA2FULL, "6082070_m.fxe"},
	{0xCAD2FF3FA4DBA1E7ULL, "183168985.ogg"},
	{0xCAD373F1D96B62E5ULL, "671144646.ogg"},
	{0xCAD37B132E9FA951ULL, "553608860.ogg"},
	{0xCAD3D678D098D078ULL, "vxlt_gallowstemplar_n_46z_0_1.vlm"},
	{0xCAD4374633E13B63ULL, "gen00fl_dog6109239.cl"},
	{0xCAD46A87302F5EC3ULL, "848106583.ogg"},
	{0xCAD488EA652C1779ULL, "118817252.ogg"},
	{0xCAD501DE0F0A22D6ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_fireflys10x10.mmh"},
	{0xCAD51BB335DAF35EULL, "6110418_f.fxe"},
	{0xCAD553661473DC04ULL, "412063576.ogg"},
	{0xCAD5776CA69F7BDCULL, "382352407.ogg"},
	{0xCAD590F432DD7443ULL, "vxht_mansion2_2ht_0_0.vlm"},
	{0xCAD59BDD939E7ADBULL, "496197926.ogg"},
	{0xCAD5A372ED59D580ULL, "vxca_caverns2_7xi_0_0.vlm"},
	{0xCAD5F137D574E49FULL, "70096256.ogg"},
	{0xCAD60CEEFF60A0E4ULL, "vxdr_primevaltomb_9e6_0_0.vlm"},
	{0xCAD63B33A477D375ULL, "gen00fl_anders6111260.cl"},
	{0xCAD64389758C8249ULL, "mer221_sabin6047674.cl"},
	{0xCAD6862F5ED59738ULL, "6085054_m.fxe"},
	{0xCAD77D57AF3AA2F0ULL, "6123659_m.fxe"},
	{0xCAD87CD42A860CAFULL, "zz_and_debug6178573.cl"},
	{0xCAD89D2A614566F2ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1ww_basic_2.mao"},
	{0xCAD8B880A3F5F387ULL, "vxlt_0_lowtown_a3_n_c_hoc_0_0.vlm"},
	{0xCAD996822F4F2DEAULL, "vxlt_0_lowtown_n_glo_0_0.vlm"},
	{0xCAD9C3E47889963BULL, "6109127_f.fxe"},
	{0xCADA74266E0EACBAULL, "gen00fl_anders6096524.cl"},
	{0xCADA8593A4F0663EULL, "mag371ip_meredith_door.ncs"},
	{0xCADAAD9BC137B656ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_windownb.mao"},
	{0xCADAB6366E6DEBEEULL, "37136369.ogg"},
	{0xCADAB99BC137CAF5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_windownb.mmh"},
	{0xCADABB9BC137CE6BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_windownb.msh"},
	{0xCADAC016975B8E88ULL, "vxw_bonepitdestroyed_a3_19p_0_0.vlm"},
	{0xCADAE79C97108667ULL, "6095553_m.fxe"},
	{0xCADAF58DBB881F4EULL, "vxht_brothel_n_6ij_0_0.vlm"},
	{0xCADB6D2AF6FBC14DULL, "vxlt_0_lowtown_a2_n_e3c_0_0.vlm"},
	{0xCADB9F9BB9EA6CD6ULL, "496772696.ogg"},
	{0xCADBEDC2454BE1B5ULL, "gen00fl_anders6125263.cl"},
	{0xCADC4DE365CCEC17ULL, "hlac_lt_0_lowtown_a2_-1_1y.dds"},
	{0xCADC5A311A823DA7ULL, "6059679_f.fxe"},
	{0xCADC62822BF6FC55ULL, "character_character_p11.psh.fxo"},
	{0xCADC91A74A28413DULL, "vxca_caverns1_cjq_0_0.vlm"},
	{0xCADCFF621C81B1D3ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.dura.evt"},
	{0xCADD07B97B1F519CULL, "vxlt_gallowstemplar_5cj_0_0.vlm"},
	{0xCADD2B826655310CULL, "994055130.ogg"},
	{0xCADDA9D39CEFF015ULL, "6162365_strong_boy.cut"},
	{0xCADDBD7DF6A47C2FULL, "w_bonepitdestroyed_a3_37_0.rml"},
	{0xCADE6761BA5B9282ULL, "96330582.ogg"},
	{0xCADE6D4490D75BB1ULL, "hlac_lt_gallowscourtyard_-131071_0y.dds"},
	{0xCADE936BB4F7DC82ULL, "vxw_blightlands_pm_0_0.vlm"},
	{0xCADF5CD019650344ULL, "41352052.ogg"},
	{0xCADFE75CEB22C27FULL, "200929407.ogg"},
	{0xCAE02F6149E644BBULL, "hlac_lt_0_lowtown_docks_a3_-65539_5x.dds"},
	{0xCAE07FE198D64449ULL, "106857200.ogg"},
	{0xCAE0B6F90D113EAAULL, "vxht_mansion2b_1iu_0_0.vlm"},
	{0xCAE0F4A099BA81F1ULL, "6043560_f.fxe"},
	{0xCAE0F7653F5CCD03ULL, "344410423.ogg"},
	{0xCAE10073B7C8AB27ULL, "143664364.ogg"},
	{0xCAE1305761AB2F33ULL, "hlac_w_encounter2_camp_196609_0z.dds"},
	{0xCAE20D531EF9259BULL, "art\\characters\\weapons\\shields\\textures\\w_shd_001a_0t.dds"},
	{0xCAE22148AD914592ULL, "6163396_m.fxe"},
	{0xCAE372977FB0D56CULL, "w_bonepitdestroyed_a3_161_0.msh"},
	{0xCAE378977FB0DF5EULL, "w_bonepitdestroyed_a3_161_0.mmh"},
	{0xCAE384977FB0F3F5ULL, "w_bonepitdestroyed_a3_161_0.mao"},
	{0xCAE3AA8687945B9AULL, "322173497.ogg"},
	{0xCAE3DAA74C098038ULL, "vxdr_primevaltomb_544_0_1.vlm"},
	{0xCAE3DF2808E4D9C6ULL, "6082216_f.fxe"},
	{0xCAE3F8621C87C414ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.dura.gad"},
	{0xCAE459A9464966E9ULL, "vxlt_0_lowtown_a3_n_c_ehm_0_0.vlm"},
	{0xCAE47382C5675C41ULL, "781426564.ogg"},
	{0xCAE488DC2C9C60B8ULL, "187636247.ogg"},
	{0xCAE4F2EEB3CCAD3AULL, "hlac_ht_0_hightown_a2_65533_2y.dds"},
	{0xCAE50D244D3DD792ULL, "vxlt_blackemporium_tq_0_0.vlm"},
	{0xCAE5341B222C001DULL, "w_woundedcoast_1_0.mao"},
	{0xCAE5381B222C0756ULL, "w_woundedcoast_1_0.mmh"},
	{0xCAE5421B222C1854ULL, "w_woundedcoast_1_0.msh"},
	{0xCAE596A9F0016E3BULL, "577209337.ogg"},
	{0xCAE5B3C12EAE802AULL, "art\\vfx\\combat\\vfx_m_ent_conj_c.anb"},
	{0xCAE5C445011C36DAULL, "sky_sky2_p1.psh.fxo"},
	{0xCAE5EAD9F9619D13ULL, "vxw_woundedcoast_3c_0_0.vlm"},
	{0xCAE61E36E30EC037ULL, "vxca_caverns1_mines_icm_0_2.vlm"},
	{0xCAE622F03BB32942ULL, "537538210.ogg"},
	{0xCAE6448DECB95274ULL, "vxlt_warehouse_10f_0_0.vlm"},
	{0xCAE7C166DBB41B47ULL, "vxca_cave_1tb_0_1.vlm"},
	{0xCAE7FD346204FB02ULL, "99373399.ogg"},
	{0xCAE80910993F071EULL, "6107163_m.fxe"},
	{0xCAE84E4A81D1E9EFULL, "s_ballista_04.mao"},
	{0xCAE85704841457C1ULL, "hf_cgn_004.mop"},
	{0xCAE9130B04CBA865ULL, "hlac_lt_0_lowtown_a2_n_-196608_2z.dds"},
	{0xCAE93AF712AA8FA6ULL, "394009878.ogg"},
	{0xCAEA0E29AF517DAAULL, "6004226_m.fxe"},
	{0xCAEA29F14E6840BCULL, "sum_conj_01.mao"},
	{0xCAEA6E60FF076628ULL, "vxlt_0_lowtown_a2_n_c_ewl_0_0.vlm"},
	{0xCAEA8E2F92BD974CULL, "6092313_m.fxe"},
	{0xCAEA99A1C41D24E1ULL, "6122239_m.fxe"},
	{0xCAEA9C023C0B4663ULL, "26171154.ogg"},
	{0xCAEAC45DC2DB1F54ULL, "438988774.ogg"},
	{0xCAEB07621C8DFCF2ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.dura.ani"},
	{0xCAEB374FC2A131D9ULL, "vxht_mansion_n_4yl_0_0.vlm"},
	{0xCAEB8EBA687971A7ULL, "vxlt_hangedman_27s_0_0.vlm"},
	{0xCAEB9902E5E8A3CAULL, "649170346.ogg"},
	{0xCAEBBD04287F8215ULL, "876743781.ogg"},
	{0xCAECC7300E82BF2AULL, "vxca_cave_2gm_0_1.vlm"},
	{0xCAECFAF9E26AF6AFULL, "gen_im_arm_hlm_hvy_set_high.uti"},
	{0xCAED0548A6EDF502ULL, "vxw_blightlands_un_0_0.vlm"},
	{0xCAED0D5FA7191DB2ULL, "mer140_bonepit_comments6077840.cl"},
	{0xCAED15A5ACB4CB0CULL, "art\\characters\\weapons\\greataxes\\w_gax_005a.msh"},
	{0xCAED1BA5ACB4D5FEULL, "art\\characters\\weapons\\greataxes\\w_gax_005a.mmh"},
	{0xCAED27A5ACB4E995ULL, "art\\characters\\weapons\\greataxes\\w_gax_005a.mao"},
	{0xCAEF91D3B48A9695ULL, "vxw_bonepitdestroyed_a3_4a_0_0.vlm"},
	{0xCAF08F71E0608919ULL, "dae011_drunk2_sta6149690.cl"},
	{0xCAF0C62A6432682CULL, "6146980_m.fxe"},
	{0xCAF1D6602555E656ULL, "hlac_w_woundedcoast_262146_0x.dds"},
	{0xCAF207359E357BB7ULL, "797732394.ogg"},
	{0xCAF2CC85EEDCFD03ULL, "77387955.ogg"},
	{0xCAF3104956552EB3ULL, "vxw_blightlands_7a_0_0.vlm"},
	{0xCAF3B5513B948693ULL, "zz_qun_debug6086853.cl"},
	{0xCAF410EBA8BF92E7ULL, "6089804_m.fxe"},
	{0xCAF48F4862882466ULL, "vxdr_primevalentrance_2ec_0_0.vlm"},
	{0xCAF54E0112D97509ULL, "58220079.ogg"},
	{0xCAF55C626CE48B90ULL, "vxlt_0_lowtown_docks_b02_0_0.vlm"},
	{0xCAF5AB6FB66EE6C3ULL, "mrl310_merrill6154558.cl"},
	{0xCAF5CECC532F0938ULL, "ht_0_hightown_a2_44_0.rml"},
	{0xCAF5D8A667CF9091ULL, "424224076.ogg"},
	{0xCAF67BA800716964ULL, "6103679_m.fxe"},
	{0xCAF705C866CBDA3FULL, "character_character_p282.psh.fxo.dx11.assembly"},
	{0xCAF737735F1A3057ULL, "ht_0_hightown_a3_9_0.mmh"},
	{0xCAF73B735F1A3764ULL, "ht_0_hightown_a3_9_0.mao"},
	{0xCAF741CF9832E5D9ULL, "lgtcr_dalish_archer.utc"},
	{0xCAF74D735F1A55C9ULL, "ht_0_hightown_a3_9_0.msh"},
	{0xCAF767AEA2D564DCULL, "vxht_0_hightown_a2_cdf_0_0.vlm"},
	{0xCAF76C561F3FA195ULL, "73267167.ogg"},
	{0xCAF7812EE28D5D72ULL, "vxw_sundermount_2dy_0_0.vlm"},
	{0xCAF7EFAD34B8FAD5ULL, "gen00fl_fenris6114535.cl"},
	{0xCAF81552FA0EFCDCULL, "sse_ss_coterie_assassin_f.gda"},
	{0xCAF8420CC7F780ACULL, "hlac_ht_0_hightown_-65536_1x.dds"},
	{0xCAF89DF4733CABA3ULL, "dae000_meredith6085241.cl"},
	{0xCAF8DFE00830A0AAULL, "follower_banter6170472.cl"},
	{0xCAF9000C84B3AFCCULL, "6097013_m.fxe"},
	{0xCAF90B04CBA76A62ULL, "lt_0_lowtown_docks_a3.rml"},
	{0xCAF95470E14620D2ULL, "hlac_lt_gallowstemplar_n_2388_5y.dds"},
	{0xCAF963B739830750ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_avgdesk.phy"},
	{0xCAF9FBF6B38762FAULL, "w_blightlands_128_0.mmh"},
	{0xCAF9FFF6B3876931ULL, "w_blightlands_128_0.mao"},
	{0xCAFA0DF6B3878140ULL, "w_blightlands_128_0.msh"},
	{0xCAFA20E5FCA9CF69ULL, "gen00fl_anders6109319.cl"},
	{0xCAFA5691879BBBC3ULL, "art\\characters\\playercharacter\\humanmale\\gor_viscount.mao"},
	{0xCAFAB804EC99435EULL, "212730407.ogg"},
	{0xCAFABF4B40741B28ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvianbroke.msh"},
	{0xCAFAD14B407439B9ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvianbroke.mao"},
	{0xCAFADD4B40744E52ULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvianbroke.mmh"},
	{0xCAFAF60F113761CFULL, "40362831.ogg"},
	{0xCAFB054878700841ULL, "hlac_ca_cavern1_haunted_24424_9z.dds"},
	{0xCAFB25E08428339BULL, "ch2_ds1.psh.fxo"},
	{0xCAFBADCA304B45DBULL, "vxca_cavern1_haunted_k2t_0_2.vlm"},
	{0xCAFBE982D41961B3ULL, "prty_fenris_armor_1.dds"},
	{0xCAFBF078040016ACULL, "vxlt_0_lowtown_docks_a3_n_d0i_0_1.vlm"},
	{0xCAFC8678904536B5ULL, "745026498.ogg"},
	{0xCAFCD60AC7608B49ULL, "182748928.ogg"},
	{0xCAFD1A159AA617B7ULL, "6177943_m.fxe"},
	{0xCAFD4EF947065635ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_act_c.mmh"},
	{0xCAFD54732CE72DEFULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_1y.dds"},
	{0xCAFD6D5E8D185A3EULL, "1016215150.ogg"},
	{0xCAFD9AB2F0FD6FAAULL, "vxht_mansion2b_1gw_0_0.vlm"},
	{0xCAFE0E2EC82CF1E6ULL, "lgt110_gxa_vgk_consprcy.ncs"},
	{0xCAFE297AC1F341BEULL, "art\\vfx\\creatures\\fxc_lotf_c.cess.evt"},
	{0xCAFE8214375FE468ULL, "w_sundermount_26_0.rml"},
	{0xCAFF8983CFAAE1DBULL, "zz_prmpt_debug6198254.cl"},
	{0xCAFFFF87E7A80FEDULL, "6149643_m.fxe"},
	{0xCB000DCC8FCD786DULL, "6181834_m.fxe"},
	{0xCB0013BB544985D2ULL, "673593416.ogg"},
	{0xCB001EF8A1036C7BULL, "hlac_lt_0_lowtown_n_131071_6z.dds"},
	{0xCB003A39E638CAD6ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.anb"},
	{0xCB003CDF70E1F743ULL, "w_bonepit_214_0.msh"},
	{0xCB004ADF70E20E9DULL, "w_bonepit_214_0.mmh"},
	{0xCB004EDF70E215AEULL, "w_bonepit_214_0.mao"},
	{0xCB0077AD6F0D7A91ULL, "fen110_anso6094995.cl"},
	{0xCB019CE40898FBFFULL, "498519236.ogg"},
	{0xCB025BCDAF1FECBDULL, "hlac_lt_0_lowtown_docks_a3_n_131071_5y.dds"},
	{0xCB0262B3CBB41571ULL, "art\\vfx\\combat\\vfx_gib_qunc_2.mao"},
	{0xCB03190A36A42577ULL, "6094305_m.fxe"},
	{0xCB0328EE7BA1A70BULL, "344418654.ogg"},
	{0xCB036C5710D789F2ULL, "hlac_lt_0_lowtown_a2_-65536_4y.dds"},
	{0xCB03797BF2191F22ULL, "6118307_m.fxe"},
	{0xCB03DFB078B66D4DULL, "409055709.ogg"},
	{0xCB0429191BB7154EULL, "6168528_m.fxe"},
	{0xCB049CB897E1663DULL, "mrl310_merrill6144207.cl"},
	{0xCB049CBE98D0A2C9ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4.ani"},
	{0xCB04A8EA5DF61592ULL, "mrl150_merrill6093913.cl"},
	{0xCB04C244657611F3ULL, "dae000_serendipity6161431.cl"},
	{0xCB04ED84F26781E4ULL, "6104779_m.fxe"},
	{0xCB05407AC1F987E9ULL, "art\\vfx\\creatures\\fxc_lotf_c.cess.gad"},
	{0xCB05498059FB0316ULL, "hlac_lt_0_lowtown_a3_n_c_0_9x.dds"},
	{0xCB05ACF6C86ABEF6ULL, "6092504_m.fxe"},
	{0xCB06102FBC89CB3BULL, "47928335.ogg"},
	{0xCB06320DB5BD87BDULL, "vxht_brothel_n_61e_0_0.vlm"},
	{0xCB06FB99D4E0F0CCULL, "vxlt_backalley_n_1jr_0_0.vlm"},
	{0xCB06FCF669AB8C0FULL, "vxw_encounter2_69_0_0.vlm"},
	{0xCB0788C502B18C91ULL, "6047410_f.fxe"},
	{0xCB07ADE7580559DAULL, "ca_cavern1_haunted_24422.rml"},
	{0xCB08254D28BF0190ULL, "gen00fl_isabela6079291.cl"},
	{0xCB085D5051019B1CULL, "vxlt_gallowstemplar_5ri_0_0.vlm"},
	{0xCB087B04A3ED4758ULL, "follower_barks6153290.cl"},
	{0xCB08EF01DC3F580FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_loop.gad"},
	{0xCB08F4F499EABDDEULL, "6082026_m.fxe"},
	{0xCB08F7198A75EFBDULL, "vxw_encounter2_gp_0_0.vlm"},
	{0xCB0903DF2E33E774ULL, "art\\vfx\\combat\\vfx_gib_cora_rtlb.gad"},
	{0xCB0915DC44E76BC2ULL, "6114497_m.fxe"},
	{0xCB0934C930D5B37AULL, "712703916.ogg"},
	{0xCB093B6508614E59ULL, "vxht_0_hightown_bx6_0_0.vlm"},
	{0xCB0A0BEA4670144EULL, "lgt000im_ultra_grimoire.uti"},
	{0xCB0A1EEEE1FD9AACULL, "89930124.ogg"},
	{0xCB0A8443D40E149CULL, "6108054_m.fxe"},
	{0xCB0AA240B09410F1ULL, "hlac_lt_0_lowtown_a2_n_65537_2y.dds"},
	{0xCB0AB7BE98D53C4BULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4.gad"},
	{0xCB0B0D55A4695335ULL, "llac_dr_slavers_3217_8.dds"},
	{0xCB0B15D146027D05ULL, "6051924_m.fxe"},
	{0xCB0B4BF947128EC2ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_act_c.anb"},
	{0xCB0B4F2CCA20883FULL, "follower_banter6143293.cl"},
	{0xCB0B6DE2941C6914ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v2.ani"},
	{0xCB0BE4991DA2A627ULL, "vxlt_gallowsprison_n_5r1_0_0.vlm"},
	{0xCB0C117AC1FF5637ULL, "art\\vfx\\creatures\\fxc_lotf_c.cess.ani"},
	{0xCB0C1FB4391ACB44ULL, "6113061_m.fxe"},
	{0xCB0C9B8F6A1FB5A9ULL, "6086212_m.fxe"},
	{0xCB0CAC84AF205101ULL, "fx_flesh2.mao"},
	{0xCB0CEF39C4089C3CULL, "zz_rogueslate_trans.cut"},
	{0xCB0D3D39E6435A41ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.mmh"},
	{0xCB0D3F74374331EAULL, "gen_im_trs_krk_hgh_junk1_02.uti"},
	{0xCB0D472255737B16ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtc.msh"},
	{0xCB0D512255738C14ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtc.mmh"},
	{0xCB0D552255739327ULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtc.mao"},
	{0xCB0D8A5D06FBAD0CULL, "vxlt_0_lowtown_a3_n_c_db3_0_0.vlm"},
	{0xCB0DF646822C1527ULL, "gen00fl_bethany6102871.cl"},
	{0xCB0DFDB4B10EBACAULL, "hlac_ht_keep_n_3733_1y.dds"},
	{0xCB0E7164D99E286DULL, "6142743_m.fxe"},
	{0xCB0EB6C12ED18A1DULL, "art\\vfx\\combat\\vfx_m_ent_conj_c.mmh"},
	{0xCB0F1B6307C1A3BDULL, "6071207_m.fxe"},
	{0xCB0F3EAC8C982675ULL, "lt_0_lowtown_docks_152_0.phy"},
	{0xCB0F42805CBD383CULL, "one220_yevhen6103118.cl"},
	{0xCB0F6CD052588137ULL, "vxht_0_hightown_n_c3v_0_1.vlm"},
	{0xCB0FBD15A0F38566ULL, "6019206_m.fxe"},
	{0xCB1012DF2E3A1F52ULL, "art\\vfx\\combat\\vfx_gib_cora_rtlb.ani"},
	{0xCB1012DF2E3A1F59ULL, "art\\vfx\\combat\\vfx_gib_cora_rtlb.anb"},
	{0xCB1037A1A1A9A296ULL, "hlac_lt_0_lowtown_a2_n_c_-3_4x.dds"},
	{0xCB10AE073915D081ULL, "6162312_m.fxe"},
	{0xCB10D7B8CFC9531BULL, "vxca_caverns2prc_abk_0_0.vlm"},
	{0xCB1138D6F961C15DULL, "473852963.ogg"},
	{0xCB114793927D0D43ULL, "6152692_m.fxe"},
	{0xCB115804315D9EE7ULL, "6140585_m.fxe"},
	{0xCB11B4BE98DB55B0ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv4.evt"},
	{0xCB11E9E6E7DE95D9ULL, "vxlt_gallowstemplar_5q9_0_0.vlm"},
	{0xCB122BA8045C21C2ULL, "atl_fonttest_dxt1_00.dds"},
	{0xCB12488575FEB4E0ULL, "423479395.ogg"},
	{0xCB124C48CDC88B98ULL, "hlac_lt_0_lowtown_65535_1x.dds"},
	{0xCB126AE294228376ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v2.gad"},
	{0xCB12D0D369058E3CULL, "hlac_lt_gallowstemplar_n_2391_5y.dds"},
	{0xCB13118C6036B3AFULL, "vxw_blightlands_w5_0_0.vlm"},
	{0xCB131A85246722ACULL, "dae000_mother6091978.cl"},
	{0xCB137E9B962A8FADULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masf_0t.dds"},
	{0xCB1385CBDEB222DFULL, "297377330.ogg"},
	{0xCB13936E1F67DC83ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_mad_man.ani"},
	{0xCB1442EBDB869B8CULL, "275840381.ogg"},
	{0xCB14889C1A51F4B5ULL, "6132487_m.fxe"},
	{0xCB14F5C55A0EA15EULL, "193424782.ogg"},
	{0xCB15C303667C7A35ULL, "6142714_m.fxe"},
	{0xCB167160E1192F35ULL, "vxlt_house_di_0_0.vlm"},
	{0xCB16791ED4355167ULL, "522890120.ogg"},
	{0xCB170D637F489962ULL, "vxht_brothel_n_5am_0_0.vlm"},
	{0xCB1720A11672A179ULL, "vxw_woundedcoast_1bs_0_0.vlm"},
	{0xCB173F2794CBCBF7ULL, "256129042.ogg"},
	{0xCB17F610CD6D7F98ULL, "vxlt_gallowstemplar_f_2ds_0_0.vlm"},
	{0xCB18184173D45E21ULL, "hf_noble_gen_1.mor"},
	{0xCB186FA3A2EEE5B9ULL, "vxw_bonepit_252_0_0.vlm"},
	{0xCB1985E29428CFD5ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v2.evt"},
	{0xCB198DB47EA664D0ULL, "6080387_m.fxe"},
	{0xCB19F6ED20CEDA5EULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_hvyo_0n.dds"},
	{0xCB1A1D8E35EBC866ULL, "vxlt_undercityencounter_z3_0_1.vlm"},
	{0xCB1A349B66D4D273ULL, "sky_ca_cavern1_haunted.mao"},
	{0xCB1A8CF0BFF405E0ULL, "art\\characters\\playercharacter\\humanfemale\\glv_nuda.phy"},
	{0xCB1ACA3855BC660CULL, "hlac_ht_mansion2b_818_0z.dds"},
	{0xCB1B41994EF9CE50ULL, "gen00fl_merrill6129921.cl"},
	{0xCB1B4D782D3ADBB3ULL, "vxht_brothel_n_6tx_0_0.vlm"},
	{0xCB1B7DBA509A5E27ULL, "llac_lt_gallowstemplar_2391_2.dds"},
	{0xCB1B9F93C4458F36ULL, "766492420.ogg"},
	{0xCB1C0150B0EB3F4CULL, "gen00fl_isabela6122915.cl"},
	{0xCB1C0F9ED1752F7FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_coma_0d.dds"},
	{0xCB1C1B4DEF50512FULL, "art\\vfx\\combat\\vfx_glb_pois_act_c.mmh"},
	{0xCB1C5F4F502A4241ULL, "vxdr_slavers_1sp_0_1.vlm"},
	{0xCB1C97A0A4693449ULL, "hlac_ca_caverns2_5802_27y.dds"},
	{0xCB1CB7D7B6E1AE5AULL, "vxlt_0_lowtown_a3_n_c_i0v_0_0.vlm"},
	{0xCB1CD5D4A087D459ULL, "clouds.vsh.fxo"},
	{0xCB1CF625E9196ACEULL, "884214801.ogg"},
	{0xCB1D0C01DC5032D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_ground_loop.ani"},
	{0xCB1D5DB4305CB0F5ULL, "vxlt_0_lowtown_a2_n_c_hyz_0_0.vlm"},
	{0xCB1D8F532ACD182CULL, "vxht_keep_6f5_0_0.vlm"},
	{0xCB1D9052A9F00B4AULL, "vxht_0_hightown_a2_aum_0_0.vlm"},
	{0xCB1DCE7A516ADF0CULL, "132783576.ogg"},
	{0xCB1E00806CF8658CULL, "6040840_m.fxe"},
	{0xCB1E042C01E4BCC8ULL, "6208385_m.fxe"},
	{0xCB1E114C55C473F5ULL, "one120_athenril6089429.cl"},
	{0xCB1E6084F1D0CBBBULL, "framebuffereffects_dof_p0.psh.fxo.dx11.assembly"},
	{0xCB1E79DFB23C8E00ULL, "gen00fl_dog6171163.cl"},
	{0xCB1EA53FDCCF5316ULL, "fen210_hadriana6110427.cl"},
	{0xCB1EBB27C6F5EC27ULL, "746374299.ogg"},
	{0xCB1F64F58998EF0CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sack_d.dds"},
	{0xCB1F707E1FD02BFAULL, "6166412_m.fxe"},
	{0xCB1F841F2C81F93EULL, "6155964_m.fxe"},
	{0xCB205D35D6055022ULL, "6112107_m.fxe"},
	{0xCB20856C1667229EULL, "6059233_m.fxe"},
	{0xCB20AB6E1F72907AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_mad_man.evt"},
	{0xCB217D041A96947FULL, "hlac_dr_slavers_3217_11x.dds"},
	{0xCB21B00F52C1B017ULL, "art\\vfx\\combat\\vfx_glb_ele_c_additive_c.impa.gad"},
	{0xCB21C42B4CEEFFF7ULL, "hlac_lt_gallowscourtyard_n_-131072_1x.dds"},
	{0xCB2231F8EFDFC7E1ULL, "art\\characters\\creatures\\dragon\\textures\\cn_bdy_drga_0n.dds"},
	{0xCB22D6C5B86FA6F2ULL, "vxw_sundermount_2wd_0_0.vlm"},
	{0xCB22D7527DB4224DULL, "lt_0_lowtown_a2_35_0_lc.mmh"},
	{0xCB22FF948427B35AULL, "zz_followers_banter6159480.cl"},
	{0xCB231110BDED188FULL, "6116245_m.fxe"},
	{0xCB233B25B246E0C3ULL, "629729749.ogg"},
	{0xCB23771B6782624FULL, "lt_0_lowtown_a3_n_c_18_0.rml"},
	{0xCB2394B2DE05D420ULL, "37238306.ogg"},
	{0xCB23C9F63AD5CE65ULL, "mag311ip_door.ncs"},
	{0xCB23DF53B8CD21CCULL, "vxlt_backalley_n_1pp_0_0.vlm"},
	{0xCB23E1FE03FB0A1AULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.anb"},
	{0xCB24008A7A0D9626ULL, "vxlt_gallowscourtyard_6k5_0_0.vlm"},
	{0xCB24340C77DBF18FULL, "vxw_blightlands_f6_0_0.vlm"},
	{0xCB2440570E132A1DULL, "252510398.ogg"},
	{0xCB24AA95679E2B99ULL, "vxht_0_hightown_a2_a74_0_0.vlm"},
	{0xCB252727E6EF6671ULL, "6112056_m.fxe"},
	{0xCB252B50217897F8ULL, "mag170_thrask6042856.cl"},
	{0xCB25410A34EDCC1CULL, "vxlt_gallowstemplar_5xo_0_0.vlm"},
	{0xCB257F40713EB1E2ULL, "vxht_brothel_n_5dn_0_0.vlm"},
	{0xCB25927E0868B9F9ULL, "vxw_woundedcoast_1mr_0_0.vlm"},
	{0xCB25F3EA4A0F01CEULL, "vxw_encounter2_6t_0_0.vlm"},
	{0xCB26999411559216ULL, "vxht_0_hightown_31g_0_0.vlm"},
	{0xCB276971CD71CEB6ULL, "plt_mer210pt_cavern_of_dead.nss"},
	{0xCB278E2570F3442DULL, "lt_undercityserialkiller_mmi.gff"},
	{0xCB27AA6E1F78AD6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_mad_man.gad"},
	{0xCB27BDF174DC8C1EULL, "1017839946.ogg"},
	{0xCB27D54AE53E6085ULL, "441451598.ogg"},
	{0xCB27E8DD172F0B2EULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robc.msh"},
	{0xCB27F6DD172F237FULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robc.mao"},
	{0xCB2802DD172F371CULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robc.mmh"},
	{0xCB28922CC767FC9CULL, "6169468_m.fxe"},
	{0xCB28BDB569A52C15ULL, "vxw_sundermount_358_0_0.vlm"},
	{0xCB28C50F52C7F2A4ULL, "art\\vfx\\combat\\vfx_glb_ele_c_additive_c.impa.evt"},
	{0xCB28E3FDB6662C3FULL, "hlac_w_encounter2_-65536_0y.dds"},
	{0xCB2918F782E4C6B2ULL, "vxca_caverns1_ij0_0_0.vlm"},
	{0xCB29FC7AF590515AULL, "vxca_caverns1_ghl_0_0.vlm"},
	{0xCB2A39656B15730CULL, "586171453.ogg"},
	{0xCB2AA683068D1078ULL, "vxdr_0_deeproads_fjz_0_0.vlm"},
	{0xCB2BCF912252C975ULL, "vxlt_0_lowtown_a2_n_c_hty_0_0.vlm"},
	{0xCB2BE03F6BCB6EB3ULL, "6096872_f.fxe"},
	{0xCB2CC924535C886AULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredround_n.dds"},
	{0xCB2CE123F6B6B8C4ULL, "6123329_m.fxe"},
	{0xCB2D74C7387A1CFFULL, "hlac_ht_chantry_n_2534_49z.dds"},
	{0xCB2DEE4FF57E6B45ULL, "hlac_lt_0_lowtown_docks_n_-65539_5x.dds"},
	{0xCB2DF2A5ACEC2699ULL, "art\\characters\\weapons\\greataxes\\w_gax_005a.phy"},
	{0xCB2E25E32A2EFE33ULL, "plt_cod_bks_lite_tombs_3.plo"},
	{0xCB2E41DEE4F91B7DULL, "zz_vault_debug6177751.cl"},
	{0xCB2E4A24813ACAF0ULL, "vxdr_primevaltomb_7zt_0_1.vlm"},
	{0xCB2EBAB246FEA926ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.anb"},
	{0xCB2ECB175E77BD56ULL, "619589572.ogg"},
	{0xCB2FEE986244E0E0ULL, "follower_banter6160949.cl"},
	{0xCB30855B7556AFAAULL, "vxlt_0_lowtown_a2_n_gne_0_0.vlm"},
	{0xCB30AE97A2BF14A8ULL, "gen00fl_isabela_init6175644.cl"},
	{0xCB30C0EBA1068C38ULL, "ser221_party_comments6209914.cl"},
	{0xCB316E4AB7F0CFD1ULL, "17302156.ogg"},
	{0xCB3186718F7EA4F1ULL, "gen00fl_merrill6111279.cl"},
	{0xCB328B0A410D134BULL, "xx_extceiling.dds"},
	{0xCB3298D7211246DAULL, "6106836_m.fxe"},
	{0xCB32C1B85446E4EFULL, "1068769026.ogg"},
	{0xCB32FC04CBD8C8B4ULL, "lt_0_lowtown_docks_a3.arl"},
	{0xCB333EA3E285B1B6ULL, "6131885_m.fxe"},
	{0xCB3391A133415520ULL, "lowtown_fem_alien_elf_2_mob6166334.cl"},
	{0xCB33C29104DA6ABDULL, "vxht_0_hightown_a2_an3_0_0.vlm"},
	{0xCB33CEB0A7310682ULL, "1033124596.ogg"},
	{0xCB34045AFA5ED279ULL, "vxw_woundedcoast_1hq_0_0.vlm"},
	{0xCB34539BA1B97D6CULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_aveline_mcr370_playerrules.gad"},
	{0xCB347053D8F5D01AULL, "mag311_nyssa6051146.cl"},
	{0xCB3490616F99601EULL, "6043673_f.fxe"},
	{0xCB34A782FA49A121ULL, "vxw_sundermount_347_0_0.vlm"},
	{0xCB34B36047C6D564ULL, "mer320_hubert6090366.cl"},
	{0xCB34BAF63AE427BCULL, "mag311ip_door.utp"},
	{0xCB34C2E5DCC7B203ULL, "vxdr_primevaltomb_77e_0_0.vlm"},
	{0xCB34D23FDEB7E44FULL, "aip_mercenaryarcher_base.gda"},
	{0xCB358F0C113DA8F2ULL, "vxht_chantry_1ym_0_0.vlm"},
	{0xCB359691C024F6AFULL, "6025385_m.fxe"},
	{0xCB35B1573626B14FULL, "126369172.ogg"},
	{0xCB35BEC727A4D588ULL, "vxlt_gallowscourtyard_7cb_0_0.vlm"},
	{0xCB35DD0F52D2A59DULL, "art\\vfx\\combat\\vfx_glb_ele_c_additive_c.impa.ani"},
	{0xCB3607698199730EULL, "769428839.ogg"},
	{0xCB364E91A4A24243ULL, "6125328_m.fxe"},
	{0xCB36D33779D5FF61ULL, "6108600_m.fxe"},
	{0xCB37042EFD38D94CULL, "267177873.ogg"},
	{0xCB370E4787E3C88FULL, "524843940.ogg"},
	{0xCB37CB18B7010BD9ULL, "971763091.ogg"},
	{0xCB39668812852CEEULL, "qunari_female_heads.gda"},
	{0xCB3A55D511113639ULL, "sky_ht_0_hightown_n.mao"},
	{0xCB3A9CBABC4BB09FULL, "6076064_m.fxe"},
	{0xCB3B32FC7F2003A4ULL, "one140cr_charade6166002.cl"},
	{0xCB3B529BA1BF9A6AULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_aveline_mcr370_playerrules.ani"},
	{0xCB3B5E7330C57417ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screena.phy"},
	{0xCB3B6B3568FCD1E9ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_108a_d.dds"},
	{0xCB3BFE11BC1FAB4CULL, "5127777.ogg"},
	{0xCB3BFFDA11A7D09EULL, "6068347_m.fxe"},
	{0xCB3C642924DF655EULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v4_c.impa.ani"},
	{0xCB3C8C41667031F7ULL, "9432458.ogg"},
	{0xCB3CB69C6E5536C9ULL, "gen00fl_anders6125300.cl"},
	{0xCB3CBDB2470AEBD1ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.mmh"},
	{0xCB3D1B8D1897FA1FULL, "gen00fl_varric6118821.cl"},
	{0xCB3D80D4B7B19611ULL, "6133457_m.fxe"},
	{0xCB3DADAFC35AE3B7ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpon.mmh"},
	{0xCB3DB1AFC35AEAC4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpon.mao"},
	{0xCB3DC3AFC35B0929ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlegrpon.msh"},
	{0xCB3DCBEFD417EA53ULL, "6174664_f.fxe"},
	{0xCB3DCCD31A8026E3ULL, "823126324.ogg"},
	{0xCB3DCF09D98889D4ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_ashfalling.mmh"},
	{0xCB3DDF4203E563F9ULL, "758002384.ogg"},
	{0xCB3E00CE3611DA94ULL, "737585399.ogg"},
	{0xCB3E8925CF92CCBAULL, "376742544.ogg"},
	{0xCB3F972495EE5DFFULL, "llac_ht_bartrand_n_3365_1.dds"},
	{0xCB3FC771BB10D6F1ULL, "vxdr_primevalentrance_1jb_0_0.vlm"},
	{0xCB3FD94BC3757E9BULL, "qun131_taskmaster6019253.cl"},
	{0xCB4013648CD5F035ULL, "follower_banter6169291.cl"},
	{0xCB404B0490FAC46BULL, "421150778.ogg"},
	{0xCB40554F52135FD4ULL, "dre000_dougal6142205.cl"},
	{0xCB40718DAAAC7CF8ULL, "6082583_m.fxe"},
	{0xCB40A9F0C0150675ULL, "art\\characters\\playercharacter\\humanfemale\\glv_nuda.msh"},
	{0xCB40ABF0C01509EBULL, "art\\characters\\playercharacter\\humanfemale\\glv_nuda.mmh"},
	{0xCB40B7F0C0151D88ULL, "art\\characters\\playercharacter\\humanfemale\\glv_nuda.mao"},
	{0xCB4117C65BC8228FULL, "vxw_blightlands_l0_0_0.vlm"},
	{0xCB416A53BC369CE9ULL, "vxlt_0_lowtown_docks_b58_0_0.vlm"},
	{0xCB41ACF7BA7E5BE1ULL, "lgt260_way_blood_mage6148366.cl"},
	{0xCB41B6DA01E441C7ULL, "vxht_mansion2_2oi_0_0.vlm"},
	{0xCB4207BCB86CF6DFULL, "111703545.ogg"},
	{0xCB42245617C9E2C1ULL, "6105438_m.fxe"},
	{0xCB427637EC54EAF9ULL, "vxw_woundedcoast_1sp_0_0.vlm"},
	{0xCB42F3D282207EABULL, "art\\levels\\races\\design\\_textures\\prp_herbs_sa.dds"},
	{0xCB43195FEC3FB9A1ULL, "vxw_sundermount_336_0_0.vlm"},
	{0xCB4334C2CEBDCA83ULL, "vxdr_primevaltomb_70d_0_0.vlm"},
	{0xCB439DC96B64065DULL, "6138733_m.fxe"},
	{0xCB43B872C95E901FULL, "821657741.ogg"},
	{0xCB43DC08EE6F4AFAULL, "6087007_m.fxe"},
	{0xCB43EA571EA4F1C5ULL, "61963224.ogg"},
	{0xCB4430A4199AEE08ULL, "vxlt_gallowscourtyard_7dc_0_0.vlm"},
	{0xCB44CE4ACC35C089ULL, "zz_mer_debug6017105.cl"},
	{0xCB44EDA3F3413F11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ec_transition.ani"},
	{0xCB45098E7D7A7B33ULL, "6088729_f.fxe"},
	{0xCB451C4DEF73586CULL, "art\\vfx\\combat\\vfx_glb_pois_act_c.anb"},
	{0xCB456B71C471A9F6ULL, "177267660.ogg"},
	{0xCB459D2DA682304CULL, "6041270-aftermath-intro2dup.cut"},
	{0xCB45FDF42771859BULL, "art\\vfx\\combat\\vfx_gib_hm_arm_lbrt.ani"},
	{0xCB468752D7FDD4BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley.mmh"},
	{0xCB468B52D7FDDBF1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley.mao"},
	{0xCB469952D7FDF300ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley.msh"},
	{0xCB47305F07395B61ULL, "106034005.ogg"},
	{0xCB478B4F01A29D1DULL, "vfx_distortionmask_p1.psh.fxo.dx11.assembly"},
	{0xCB47A260B73CE801ULL, "mrl310_merrill6120407.cl"},
	{0xCB47AFA0027ED195ULL, "289258548.ogg"},
	{0xCB481CB1B9859771ULL, "644882935.ogg"},
	{0xCB48A40B4C3364B9ULL, "t1_tat_bld.tnt"},
	{0xCB48BD861263C53AULL, "hlac_lt_0_lowtown_docks_n_65532_0z.dds"},
	{0xCB48C0F59D39E599ULL, "w_bonepitmining_a2_232_0.msh"},
	{0xCB48C8578A37545CULL, "vxlt_0_lowtown_docks_a3_cqs_0_0.vlm"},
	{0xCB48CA38F62509A7ULL, "6114420_f.fxe"},
	{0xCB48CAF59D39F687ULL, "w_bonepitmining_a2_232_0.mmh"},
	{0xCB48CEF59D39FDD4ULL, "w_bonepitmining_a2_232_0.mao"},
	{0xCB48D7E7531CADE4ULL, "335901638.ogg"},
	{0xCB497C2924EA182FULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v4_c.impa.evt"},
	{0xCB4984AC8A827480ULL, "504039470.ogg"},
	{0xCB49DD8B374442BBULL, "6149223_m.fxe"},
	{0xCB4A33DEAE062A58ULL, "vxht_keep_n_60l_0_0.vlm"},
	{0xCB4A88B47495F541ULL, "zz_qcr_debug6052122.cl"},
	{0xCB4BC4D5752EBEA5ULL, "gen00fl_anders6103920.cl"},
	{0xCB4BD3C2C354E87FULL, "353507297.ogg"},
	{0xCB4BF8A3F34770B3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ec_transition.gad"},
	{0xCB4C8103F4F1B312ULL, "vxlt_undercityserialkiller_1g6_0_0.vlm"},
	{0xCB4CBAB690EA3ECDULL, "vxlt_0_lowtown_docks_a3_n_cz4_0_0.vlm"},
	{0xCB4CE4FE041E14CDULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_proj_p.mmh"},
	{0xCB4DE1DD174FCEE7ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_robc.phy"},
	{0xCB4DEFDCBDFF7C70ULL, "vxlt_undercityencounter_141_0_0.vlm"},
	{0xCB4E0025652E0EFAULL, "hlac_ht_playershouse_n_855_9x.dds"},
	{0xCB4EBB81E07CE997ULL, "gen00fl_aveline6132756.cl"},
	{0xCB4EEDF28DBA7744ULL, "6120763_f.fxe"},
	{0xCB4F0B17C797C8FFULL, "vxw_blightlands_ht_0_0.vlm"},
	{0xCB4F515140542232ULL, "vxlt_warehouse_qb_0_0.vlm"},
	{0xCB4F559D9DF1CBDCULL, "mag211_marethari6025163.cl"},
	{0xCB4F56214FEFDFA6ULL, "vxlt_gallowscourtyard_6x8_0_0.vlm"},
	{0xCB4F89A34DBE3B0FULL, "vxw_blightlands_q3_0_0.vlm"},
	{0xCB4FAE5F09133EA2ULL, "vxlt_0_lowtown_docks_n_b0o_0_0.vlm"},
	{0xCB4FD94A95478F97ULL, "w_bonepitmining_a2_162_0.mmh"},
	{0xCB4FDD4A954796A4ULL, "w_bonepitmining_a2_162_0.mao"},
	{0xCB4FEF4A9547B509ULL, "w_bonepitmining_a2_162_0.msh"},
	{0xCB4FFD356C4CB76CULL, "58040382.ogg"},
	{0xCB507D2924F038D8ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v4_c.impa.gad"},
	{0xCB508C8445480407ULL, "vxw_sundermount_204_0_0.vlm"},
	{0xCB525A46DC171B9BULL, "vxlt_0_lowtown_a2_n_c_hsr_0_0.vlm"},
	{0xCB52F5A3F34D8A18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ec_transition.evt"},
	{0xCB540E1EECF7F1D6ULL, "6108768_m.fxe"},
	{0xCB5514FF1EA24316ULL, "fen110_fenris6117974.cl"},
	{0xCB5536055110AF09ULL, "hlac_lt_undercityencounter_-1_14z.dds"},
	{0xCB554172249F8618ULL, "vxlt_hangedman_rm_0_0.vlm"},
	{0xCB5557D8C2EFA38EULL, "83429352.ogg"},
	{0xCB5566B7245FB5A9ULL, "vxdr_primevalentrance_2vy_0_0.vlm"},
	{0xCB5567FB3B272DA1ULL, "6031154_f.fxe"},
	{0xCB5583EA5AF9688BULL, "gen00fl_sebastian6180687.cl"},
	{0xCB55AB172D092D92ULL, "hlac_w_bonepit_262149_0x.dds"},
	{0xCB55B3E7488329DBULL, "vxdr_primevalentrance_14s_0_1.vlm"},
	{0xCB5651AC8CD44890ULL, "lt_0_lowtown_docks_152_0.msh"},
	{0xCB565FAC8CD460AAULL, "lt_0_lowtown_docks_152_0.mmh"},
	{0xCB5663AC8CD467E1ULL, "lt_0_lowtown_docks_152_0.mao"},
	{0xCB56DC4B40C2AB4DULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluvianbroke.phy"},
	{0xCB573A347C2D6CDCULL, "vxlt_0_lowtown_docks_a3_ctp_0_0.vlm"},
	{0xCB5760F21EE0BAC7ULL, "707283682.ogg"},
	{0xCB584F68293A5B3BULL, "6142323_m.fxe"},
	{0xCB5872F0DE2AB438ULL, "498509542.ogg"},
	{0xCB59A112DEECAAFDULL, "6093684_f.fxe"},
	{0xCB59B3689E99861DULL, "primitives_primitive_p1.vsh.fxo.dx11"},
	{0xCB5A24F427827125ULL, "art\\vfx\\combat\\vfx_gib_hm_arm_lbrt.gad"},
	{0xCB5A3C0328D46665ULL, "vxw_sundermount_35h_0_0.vlm"},
	{0xCB5A648E9B9DC711ULL, "6139257_m.fxe"},
	{0xCB5A953CCD37066EULL, "928927980.ogg"},
	{0xCB5A9E5AE9325540ULL, "716158800.ogg"},
	{0xCB5A9EC8963B3788ULL, "vxdr_primevaltomb_80l_0_0.vlm"},
	{0xCB5B1BFEDAB6585FULL, "gen00fl_aveline.bnk"},
	{0xCB5B8920BA8B7856ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_prp1bluebase.msh"},
	{0xCB5B9320BA8B8954ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_prp1bluebase.mmh"},
	{0xCB5B9720BA8B9067ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_prp1bluebase.mao"},
	{0xCB5C7CD1F0E00F8BULL, "sst_ss_carta_thug_m.gda"},
	{0xCB5C9D96C2485BC6ULL, "hlac_lt_0_lowtown_a2_0_3z.dds"},
	{0xCB5D516AF13C13FAULL, "vxlt_0_lowtown_ert_0_0.vlm"},
	{0xCB5D7CF4B98DE17FULL, "vxw_blightlands_mw_0_0.vlm"},
	{0xCB5D8EA7E0DE92FBULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_harrowing_sa.dds"},
	{0xCB5DAC494FE0D7CBULL, "and110_refugee.bnk"},
	{0xCB5DC32E324A3AB2ULL, "vxlt_warehouse_zc_0_0.vlm"},
	{0xCB5DFDFEDAB85B67ULL, "gen00fl_aveline.cnv"},
	{0xCB5E01B3BBFFB8F5ULL, "hlac_lt_0_lowtown_docks_a3_196608_1y.dds"},
	{0xCB5E0CE32A57F7D0ULL, "plt_cod_bks_lite_tombs_3.nss"},
	{0xCB5E3A563C02A23CULL, "w_woundedcoast_167_0.mao"},
	{0xCB5E46563C02B6DFULL, "w_woundedcoast_167_0.mmh"},
	{0xCB5E4C563C02C081ULL, "w_woundedcoast_167_0.msh"},
	{0xCB5E9A93E5D072C7ULL, "vxht_mansion2_2uk_0_0.vlm"},
	{0xCB5EE16499E3C6C6ULL, "6067870_m.fxe"},
	{0xCB5F0BDC9FC77C9FULL, "6139466_m.fxe"},
	{0xCB5FD58CE1DB010FULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1pil.mao"},
	{0xCB5FE18CE1DB15ACULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1pil.mmh"},
	{0xCB5FE78CE1DB1F9EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1pil.msh"},
	{0xCB603EB38C4F03D0ULL, "vxlt_darktown_5gf_0_0.vlm"},
	{0xCB6068AF0C2CBB22ULL, "gen00fl_anders6116084.cl"},
	{0xCB607C29A57ADCF2ULL, "mcr370_varric6098337.cl"},
	{0xCB609BD796A29B1BULL, "royalarcher_sneakattack.dds"},
	{0xCB60AE494FE310ABULL, "and110_refugee.cnv"},
	{0xCB60BE7DC5A8ECA6ULL, "862470351.ogg"},
	{0xCB60D1DA95575E4BULL, "zz_followers_banter6157351.cl"},
	{0xCB6105A976B165E0ULL, "433075608.ogg"},
	{0xCB62048986419961ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage2_.gad"},
	{0xCB621109DD52A424ULL, "vxw_bonepitmining_a2_iy_0_0.vlm"},
	{0xCB63080B475894E3ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hb1a.mmh"},
	{0xCB63160B4758AC3DULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hb1a.msh"},
	{0xCB643C156F86047DULL, "6129339_m.fxe"},
	{0xCB6464AB4F37B279ULL, "264708417.ogg"},
	{0xCB646EBF0C7819EEULL, "vxca_caverns2_1xa_0_0.vlm"},
	{0xCB652A69C68EBC8AULL, "hightown_fem_noble2_mob6169485.cl"},
	{0xCB65387B89CDA5FAULL, "vxca_cave_1fx_0_0.vlm"},
	{0xCB66D0917592A023ULL, "557593874.ogg"},
	{0xCB6707F2B04084BCULL, "vxht_0_hightown_a2_a7m_0_0.vlm"},
	{0xCB6782D752E07260ULL, "717196025.ogg"},
	{0xCB67AFC772A87057ULL, "ave121_brennan6105921.cl"},
	{0xCB67B965EF51420DULL, "gen00fl_fenris6103351.cl"},
	{0xCB6805898646060FULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\mage2_.ani"},
	{0xCB68B657DEC459B2ULL, "918024987.ogg"},
	{0xCB68DFB366384574ULL, "512557234.ogg"},
	{0xCB68E3149A913ADEULL, "dae000ip_crowd_ff7_talk6.utp"},
	{0xCB68F9A98100F0D8ULL, "vxht_mansion_n_6k9_0_0.vlm"},
	{0xCB6913D4C46AE2F6ULL, "vxdr_primevaltomb_6qn_0_1.vlm"},
	{0xCB69502255C1E90FULL, "art\\characters\\playercharacter\\humanfemale\\arm_lgtc.phy"},
	{0xCB69825F90441E5FULL, "6117432_f.fxe"},
	{0xCB6A6B5C04837F19ULL, "vxlt_undercityencounter_v5_0_0.vlm"},
	{0xCB6A8424DBF2565AULL, "vxlt_0_lowtown_docks_a3_d3p_0_0.vlm"},
	{0xCB6B9E9EF4235EE1ULL, "vxlt_hangedman_n_1xh_0_0.vlm"},
	{0xCB6BA652D81D25C5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ss_trampulley.phy"},
	{0xCB6BC347E3322C7AULL, "vxlt_0_lowtown_emu_0_0.vlm"},
	{0xCB6BEED1AB83F9FFULL, "vxw_blightlands_vv_0_0.vlm"},
	{0xCB6C1DE05AEAA48DULL, "gen00fl_varric6075766.cl"},
	{0xCB6CBFEDFBD76437ULL, "300256359.ogg"},
	{0xCB6CD4AD4C2FADC7ULL, "mag291_torpor6022740.cl"},
	{0xCB6D0B98C55419C6ULL, "llac_lt_shop_-1_6.dds"},
	{0xCB6D0D8FEB974D1BULL, "vxlt_warehouse_196_0_0.vlm"},
	{0xCB6DF748A12A5FA6ULL, "character_character_p212.psh.fxo.dx11"},
	{0xCB6E80FD5007FACDULL, "6085877_m.fxe"},
	{0xCB6E875E2D798636ULL, "gen_im_trs_gem_cub_zirc_01.uti"},
	{0xCB6EE9207851795CULL, "6108915_m.fxe"},
	{0xCB6F2171DC0DFA3AULL, "art\\characters\\creatures\\dog\\textures\\c_dog_tata.dds"},
	{0xCB6F4A1B254D528AULL, "765916543.ogg"},
	{0xCB6F6E8D7E68ACC4ULL, "vxlt_0_lowtown_bxw_0_0.vlm"},
	{0xCB6FD43C5C55EB5CULL, "w_blightlands_42_0.rml"},
	{0xCB6FD4FA9DAE9EF5ULL, "gen00fl_anders6119448.cl"},
	{0xCB6FDBE2CCAD78C7ULL, "430533902.ogg"},
	{0xCB7025DAB0D9A042ULL, "vxca_cavern1_haunted_j4h_0_0.vlm"},
	{0xCB70F604E5275517ULL, "vxht_0_hightown_a2_bxl_0_0.vlm"},
	{0xCB70FA80A1628325ULL, "llac_lt_0_lowtown_n_131072_1.dds"},
	{0xCB7146278C75A232ULL, "6131969_f.fxe"},
	{0xCB714944A08ED926ULL, "vxdr_primevaltomb_4xf_0_0.vlm"},
	{0xCB7197FF46910F2DULL, "fx_dazz_halo.dds"},
	{0xCB71AE1F9F914FA7ULL, "gen00fl_fenris6108082.cl"},
	{0xCB71D0225A26EFC8ULL, "246351313.ogg"},
	{0xCB720A64F87754FBULL, "6000981_f.fxe"},
	{0xCB731D2C128D79ABULL, "hlac_lt_gallowscourtyard_131071_0x.dds"},
	{0xCB7383CCAAF09208ULL, "lt_warehouse_1_0_lc.mmh"},
	{0xCB73AA587BC3BE7AULL, "vxca_cave_1ay_0_0.vlm"},
	{0xCB73B4C7416BD4A7ULL, "art\\characters\\heraldry\\hld_healers.dds"},
	{0xCB73BA3C5C59A7B1ULL, "w_blightlands_42_0.msh"},
	{0xCB73C2FFFBA2181EULL, "vxca_caverns1_gj2_0_0.vlm"},
	{0xCB73C83C5C59BFCCULL, "w_blightlands_42_0.mao"},
	{0xCB73D43C5C59D46FULL, "w_blightlands_42_0.mmh"},
	{0xCB74090B47670908ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_hb1a.phy"},
	{0xCB752754B5780AC8ULL, "vxlt_hangedman_n_1xm_0_0.vlm"},
	{0xCB752AB0240D9679ULL, "docks_scene_dockworkers6176197.cl"},
	{0xCB75383AB2CEB709ULL, "mag360_samson6044869.cl"},
	{0xCB75478F76925436ULL, "6007717_m.fxe"},
	{0xCB75931455D99962ULL, "lgt150ip_ntb_wr_ing_chest.utp"},
	{0xCB75DC0EB5B2689FULL, "942634870.ogg"},
	{0xCB75F13AD9C60B03ULL, "1023531097.ogg"},
	{0xCB762B70C948D3BDULL, "859434.ogg"},
	{0xCB76D762020FEC2BULL, "vxht_0_hightown_n_5i3_0_0.vlm"},
	{0xCB776B8672F70958ULL, "vxht_mansion_n_6n6_0_0.vlm"},
	{0xCB7793CA12671078ULL, "vxlt_0_lowtown_a2_eqf_0_1.vlm"},
	{0xCB781BEB9E730D88ULL, "348090745.ogg"},
	{0xCB790CB8364F3F4DULL, "dae000ip_ambient_none.utp"},
	{0xCB79768C5DB054CEULL, "8942331.ogg"},
	{0xCB798A3AE02241A2ULL, "339038712.ogg"},
	{0xCB79A494FCEE9C08ULL, "vxw_bonepitdestroyed_a3_1pg_0_0.vlm"},
	{0xCB79CC81989ECFE3ULL, "vxlt_undercityencounter_yz_0_0.vlm"},
	{0xCB79D73203BA8AC5ULL, "vxlt_0_lowtown_ewp_0_0.vlm"},
	{0xCB79EA1FC0AB6616ULL, "ht_0_hightownexplosion_n_24_0.rml"},
	{0xCB7A917705A23C22ULL, "6118516_m.fxe"},
	{0xCB7AFA7468684310ULL, "hlac_dr_0_deeproads_5843_4z.dds"},
	{0xCB7AFD6114B772A2ULL, "398187505.ogg"},
	{0xCB7B0DDE67CC7D08ULL, "vxca_cavern1_haunted_i8m_0_0.vlm"},
	{0xCB7B12F1E7AFCD93ULL, "vxw_bonepitmining_a2_1rt_0_0.vlm"},
	{0xCB7B86A91DEA6D14ULL, "6088112_m.fxe"},
	{0xCB7BDB91E7C3C8C3ULL, "vxlt_0_lowtown_docks_bd9_0_0.vlm"},
	{0xCB7C2D6467AA85BAULL, "hlac_lt_0_lowtown_docks_a3_131072_1x.dds"},
	{0xCB7C631B25CF2261ULL, "6121505_f.fxe"},
	{0xCB7CBF4A28820500ULL, "989962920.ogg"},
	{0xCB7D0D02A16028B1ULL, "6202341_f.fxe"},
	{0xCB7D790BA4AAF343ULL, "vxht_0_hightown_n_cbl_0_0.vlm"},
	{0xCB7DE3CB22AFE932ULL, "6137227_m.fxe"},
	{0xCB7E3738D661E20BULL, "gen00fl_bethany6169915.cl"},
	{0xCB7F13230B2708CBULL, "1064804721.ogg"},
	{0xCB7FE9D067796AB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_exit.gad"},
	{0xCB8007FEDAD55216ULL, "gen00fl_aveline.utc"},
	{0xCB800CEFAD73DADEULL, "6082255_m.fxe"},
	{0xCB80EEC7D2C78954ULL, "31878045.ogg"},
	{0xCB81FCA6717FE8BBULL, "597895365.ogg"},
	{0xCB82337331012E6FULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screena.mao"},
	{0xCB823A5387320F95ULL, "779882633.ogg"},
	{0xCB823F733101420CULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screena.mmh"},
	{0xCB82457331014CFEULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_screena.msh"},
	{0xCB82709080A79539ULL, "6117613_m.fxe"},
	{0xCB827B988E0E8471ULL, "lgt311_shw_mer_mettin6156476.cl"},
	{0xCB834C60E1FC3CF5ULL, "hlac_ht_chantry_n_2534_42y.dds"},
	{0xCB8382EAC9F223B3ULL, "vxw_blightlands_ek_0_0.vlm"},
	{0xCB83ACA83CE1E014ULL, "fen310_fenris6144527.cl"},
	{0xCB841B2D078A50A0ULL, "vxdr_primevalentrance_1ij_0_0.vlm"},
	{0xCB8435E6928FDDE5ULL, "llac_w_encounter3_0_8.dds"},
	{0xCB84CE13B8C63368ULL, "gen00fl_isabela6075215.cl"},
	{0xCB84F585CE1E8BECULL, "t1_tat_grn.tnt"},
	{0xCB850D79943278CDULL, "76971331.ogg"},
	{0xCB85202AAEBB9A54ULL, "6019158_m.fxe"},
	{0xCB858F4E2BCF059EULL, "997030192.ogg"},
	{0xCB85E36B014C77F6ULL, "hlac_lt_0_lowtown_a3_n_c_65535_9x.dds"},
	{0xCB85FED0677DFA44ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_exit.evt"},
	{0xCB86307A096F5DBCULL, "6102485_m.fxe"},
	{0xCB86335E1B4AC0BDULL, "6071070_m.fxe"},
	{0xCB86A7A83626F3BAULL, "zz_followers_banter6158793.cl"},
	{0xCB87C86B6B1264F6ULL, "vxlt_0_lowtown_h8o_0_0.vlm"},
	{0xCB87FB9504EFAD1FULL, "zz_magewhite_in.cut"},
	{0xCB881671EEE4B488ULL, "vxw_bonepitdestroyed_a3_1ud_0_0.vlm"},
	{0xCB8816C4EEE563C7ULL, "6001353_f.fxe"},
	{0xCB88ABCC001ADAB7ULL, "6018746_f.fxe"},
	{0xCB88EBA09C95EEE8ULL, "hlac_lt_0_lowtown_n_-196607_0z.dds"},
	{0xCB88EF84A9C84807ULL, "794921755.ogg"},
	{0xCB8923C7FBEFAF03ULL, "hlac_lt_0_lowtown_a3_n_c_65537_7y.dds"},
	{0xCB893CADD16E6F40ULL, "310557072.ogg"},
	{0xCB89649B36CB061FULL, "rdr110dg_dockworker6015949.cl"},
	{0xCB8965154999AC6DULL, "vxht_0_hightown_n_c4p_0_0.vlm"},
	{0xCB8971424AAA1A7CULL, "mcr310_argument6092085.cl"},
	{0xCB897FD5A5F18129ULL, "dae000cr_grand_cleric6183970.cl"},
	{0xCB89FE27BBD0BBAFULL, "vxlt_darktown_43h_0_0.vlm"},
	{0xCB8A6BD2D63CEC23ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_roba_0s.dds"},
	{0xCB8A7BB2BDA1905DULL, "6060138_f.fxe"},
	{0xCB8AA24E65FBBA58ULL, "hm_gencr_city_guard.mor"},
	{0xCB8B457F3663F759ULL, "vxca_caverns2_829_0_0.vlm"},
	{0xCB8B9221499B6EA4ULL, "ave310_brennan6137242.cl"},
	{0xCB8BBAF99483197EULL, "370778921.ogg"},
	{0xCB8BC78A380082F5ULL, "516264904.ogg"},
	{0xCB8C0872CCE3D375ULL, "793335115.ogg"},
	{0xCB8C4536CF8D0AAEULL, "vxdr_0_deeproads_cel_0_0.vlm"},
	{0xCB8C8BB517A9FEB6ULL, "vxw_sundermount_31c_0_0.vlm"},
	{0xCB8CB28EDEC183C8ULL, "6100061_m.fxe"},
	{0xCB8CE28D6BF8D760ULL, "one100dg_vanard6048234.cl"},
	{0xCB8DD9BEC9138617ULL, "vxht_0_hightown_a2_bfn_0_0.vlm"},
	{0xCB8E1441D7D3BEE3ULL, "76830203.ogg"},
	{0xCB8E7F9AA2450D80ULL, "vxlt_0_lowtown_a2_n_c_emv_0_0.vlm"},
	{0xCB8E8D6D78F469BFULL, "qun200_party_comment6095376.cl"},
	{0xCB8E90E6CB3DB23BULL, "vxht_brothel_n_587_0_0.vlm"},
	{0xCB8ED0D6C78745E7ULL, "vxlt_undercityserialkiller_20a_0_0.vlm"},
	{0xCB8F1D97D96DE29AULL, "framebuffereffects_smallblur_p0.psh.fxo.dx11"},
	{0xCB8F4352A075A64CULL, "vxlt_0_lowtown_docks_a3_bfw_0_0.vlm"},
	{0xCB90301F8E423545ULL, "vxca_cave_1fs_0_0.vlm"},
	{0xCB90A8C1C85AE0F1ULL, "156136348.ogg"},
	{0xCB912BBB7910032FULL, "zz_qun_debug6049742.cl"},
	{0xCB91F4C7BBE83C33ULL, "vxw_blightlands_nj_0_0.vlm"},
	{0xCB920B0E99643BC8ULL, "vxlt_hangedman_n_1rk_0_0.vlm"},
	{0xCB92AB9DBA240552ULL, "139278231.ogg"},
	{0xCB92D2CAD2500B17ULL, "171606767.ogg"},
	{0xCB93B190F2957CFDULL, "vxht_keep_646_0_0.vlm"},
	{0xCB93B93A1B892622ULL, "6089085_m.fxe"},
	{0xCB9416D0678A603DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\sitting_ground_exit.ani"},
	{0xCB945930EF16BBD6ULL, "w_sundermount_33_0.rml"},
	{0xCB94D92C489EB088ULL, "6169990_m.fxe"},
	{0xCB94F6CAE801AB69ULL, "vxca_cavern1_haunted_j6e_0_0.vlm"},
	{0xCB9504E29CB5A7B2ULL, "247183247.ogg"},
	{0xCB9517C2333D4147ULL, "vxca_cave_18v_0_1.vlm"},
	{0xCB9556C61541E916ULL, "817658971.ogg"},
	{0xCB958FDBE244CBC9ULL, "52901187.ogg"},
	{0xCB95CFD6732C1B9EULL, "6202284_m.fxe"},
	{0xCB9604F4521B0E9BULL, "6090455_m.fxe"},
	{0xCB96A8F6C758130AULL, "6117940_m.fxe"},
	{0xCB96B5CCC74A5631ULL, "dae000_sandal6147771.cl"},
	{0xCB9716209EE455AEULL, "vxht_chantry_2aj_0_0.vlm"},
	{0xCB977530E98E6D22ULL, "6112716_m.fxe"},
	{0xCB97D6F23B8FC4EDULL, "vxht_0_hightown_n_c1o_0_0.vlm"},
	{0xCB97F1984BB8AE08ULL, "vxca_cavern1_haunted_i2k_0_0.vlm"},
	{0xCB985947EA16A5DAULL, "primitives_primitive_p4.vsh.fxo"},
	{0xCB989D6F3010D1CFULL, "6128396_m.fxe"},
	{0xCB98E6708111E6B4ULL, "lt_0_lowtown_docks_a3_ll.mmh"},
	{0xCB9931E1A71018DFULL, "hlac_lt_0_lowtown_a2_1_3x.dds"},
	{0xCB997AD399ABED43ULL, "342211709.ogg"},
	{0xCB99BD172857193DULL, "vxw_blightlands_9q_0_0.vlm"},
	{0xCB9A24B4C9540A50ULL, "hlac_ht_keep_3732_13z.dds"},
	{0xCB9B5977B04C6376ULL, "451079369.ogg"},
	{0xCB9BA43049B694E9ULL, "mag101_aombanter6052751.cl"},
	{0xCB9BB9F1ECF7A379ULL, "468914492.ogg"},
	{0xCB9BC950508BC491ULL, "vxlt_backalley_g_1dt_0_0.vlm"},
	{0xCB9BCB6A24DDFE2CULL, "6112102_f.fxe"},
	{0xCB9BDC7AD39DFFD3ULL, "dae000cr_grand_cleric6187925.cl"},
	{0xCB9BDCF7C735E012ULL, "llac_lt_0_lowtown_a2_n_-65535_3.dds"},
	{0xCB9BF630766620DDULL, "6117294_f.fxe"},
	{0xCB9C2ABCF85C9098ULL, "vxht_0_hightown_a2_5io_0_0.vlm"},
	{0xCB9C390B0EBC54D4ULL, "vxlt_backalley_n_1jh_0_2.vlm"},
	{0xCB9CA7DF12958AB2ULL, "vxca_caverns1_i28_0_0.vlm"},
	{0xCB9CE055CA25095AULL, "6149873_m.fxe"},
	{0xCB9D4CFA74C78CB9ULL, "hla_ht_0_hightown_a3_-131073_0z.dds"},
	{0xCB9D5FBCFDAD15E7ULL, "564088783.ogg"},
	{0xCB9D78F130D5493CULL, "6139920_m.fxe"},
	{0xCB9DCFB468ACCB09ULL, "qun221tr_viscount_ambient.ncs"},
	{0xCB9E93C8CF397EA8ULL, "vxht_bartrand_n_4f3_0_0.vlm"},
	{0xCB9EA1FC80384DC5ULL, "vxca_cave_1ar_0_0.vlm"},
	{0xCB9ED937586B4CD4ULL, "gen00fl_anders6113092.cl"},
	{0xCB9EE20A5D2D3F36ULL, "366735211.ogg"},
	{0xCB9F99FEA5DB5421ULL, "vrc321ip_spooky6_obj_8.utp"},
	{0xCBA066A4ADDE54B3ULL, "vxw_blightlands_sm_0_0.vlm"},
	{0xCBA08FBF03848E49ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_retrieval.ani"},
	{0xCBA0EDC56285D781ULL, "6031294_m.fxe"},
	{0xCBA0F4740189A6DEULL, "gen00fl_isabela6135698.cl"},
	{0xCBA139C8FDEA0713ULL, "art\\characters\\creatures\\mubari\\animation\\walkp.ani"},
	{0xCBA1586DB9515171ULL, "825288040.ogg"},
	{0xCBA1A66C2C2319CEULL, "20102818.ogg"},
	{0xCBA1B45F67A6D397ULL, "vxlt_gallowstemplar_5j6_0_0.vlm"},
	{0xCBA1B64401F4D3C6ULL, "68537869.ogg"},
	{0xCBA1FA5A1DC5B1CCULL, "art\\characters\\playercharacter\\humanmale\\arm_meda.msh"},
	{0xCBA2005A1DC5BBBEULL, "art\\characters\\playercharacter\\humanmale\\arm_meda.mmh"},
	{0xCBA20C5A1DC5D055ULL, "art\\characters\\playercharacter\\humanmale\\arm_meda.mao"},
	{0xCBA21017845E29F2ULL, "578137230.ogg"},
	{0xCBA2236DE48B957DULL, "vxht_keep_615_0_0.vlm"},
	{0xCBA2455D75553D39ULL, "vxlt_backalley_n_1rc_0_0.vlm"},
	{0xCBA25DA946E31D13ULL, "vxht_playershouse_n_1p5_0_0.vlm"},
	{0xCBA28723C7492031ULL, "vxw_woundedcoast_1bk_0_0.vlm"},
	{0xCBA33AA99F75ED8FULL, "gen00fl_fenris6126499.cl"},
	{0xCBA371C94058FC54ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab.phy"},
	{0xCBA378023A51A555ULL, "6067856_f.fxe"},
	{0xCBA3A8540DE512D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_carver.ani"},
	{0xCBA41293A5AE742FULL, "dae000_gamlen6005056.cl"},
	{0xCBA5168C1B50953DULL, "vxht_brothel_n_6hz_0_0.vlm"},
	{0xCBA52D110450CB93ULL, "dae000_brothel_gamlen6157080.cl"},
	{0xCBA554A2BA3FE32EULL, "hlac_w_bonepitdestroyed_a3_262147_0z.dds"},
	{0xCBA587FD90DA6E2EULL, "vxht_chantry_2zk_0_0.vlm"},
	{0xCBA5AFFB436602FEULL, "combat_st_r1_a3.gda"},
	{0xCBA6259DDA577199ULL, "vxca_caverns1_clk_0_0.vlm"},
	{0xCBA6AABF038927CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_retrieval.gad"},
	{0xCBA6AF35850BEC92ULL, "849877355.ogg"},
	{0xCBA6C50C276F0B0DULL, "mer340_bonepit_comments6083156.cl"},
	{0xCBA743A6A09274ACULL, "mist_fl_cl.mao"},
	{0xCBA7C180DCE69BAEULL, "art\\characters\\creatures\\mubari\\animation\\cwkbr.evt"},
	{0xCBA80FB21DA3552AULL, "845320471.ogg"},
	{0xCBA8C953192C40B4ULL, "gen_im_arm_boo_hvy_act1_01.uti"},
	{0xCBA8CA76360730E6ULL, "85655360.ogg"},
	{0xCBA8F76311170665ULL, "6067640_m.fxe"},
	{0xCBA90947F3EEC4E3ULL, "gen_im_acc_rng_mrl1.uti"},
	{0xCBA92466FEC59FE7ULL, "plt_cod_bks_chantry_hist_1.plo"},
	{0xCBA94488981E1FF5ULL, "mag110pt_wayward_son.ncs"},
	{0xCBA97F032D9BC4CEULL, "843513033.ogg"},
	{0xCBA9886E326D16FDULL, "zz_approval_debug6003712.cl"},
	{0xCBA98D44AE44C344ULL, "6208507_m.fxe"},
	{0xCBA9B3540DE99173ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_carver.gad"},
	{0xCBAA009130435FF2ULL, "189644457.ogg"},
	{0xCBAA0E07F4896C43ULL, "gen00fl_sebastian6178356.cl"},
	{0xCBAA40B41D5F05B2ULL, "vxw_encounter3_6s_0_0.vlm"},
	{0xCBAA84ED3EDD50A7ULL, "zz_pro_cutscenes6005835.cl"},
	{0xCBAA9F154A5DFE2BULL, "vxht_0_hightown_a3_c7q_0_0.vlm"},
	{0xCBAB7F2564493287ULL, "genip_cin_letter.utp"},
	{0xCBAB8B1956AE786DULL, "509809491.ogg"},
	{0xCBABAA1217FC260CULL, "334250131.ogg"},
	{0xCBABD5627863B50BULL, "mrl000_marethari6129336.cl"},
	{0xCBAC1E8899794C63ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgtp_0t.dds"},
	{0xCBAC76EC4EC0B196ULL, "ser121_jethann6216501.cl"},
	{0xCBACF91509765918ULL, "673855873.ogg"},
	{0xCBADA7BF038F4130ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_throw_retrieval.evt"},
	{0xCBADE2ED796F1FF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_action.ani"},
	{0xCBAE0EC7BD78FA5FULL, "hlac_ht_0_hightown_2_0x.dds"},
	{0xCBAE4EC94062C851ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab.msh"},
	{0xCBAE5CC94062E06CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab.mao"},
	{0xCBAE68C94062F40FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab.mmh"},
	{0xCBAE98523E2016D3ULL, "crate_could.mao"},
	{0xCBAEB880DCECAB59ULL, "art\\characters\\creatures\\mubari\\animation\\cwkbr.gad"},
	{0xCBAEB8B468BB17ECULL, "qun221tr_viscount_ambient.utt"},
	{0xCBAED60CB0781368ULL, "213285414.ogg"},
	{0xCBAED8819FD46D33ULL, "vxw_blightlands_tl_0_0.vlm"},
	{0xCBAF0EAEB2F90670ULL, "4641210.ogg"},
	{0xCBAF31C8FDF6374AULL, "art\\characters\\creatures\\mubari\\animation\\walkp.evt"},
	{0xCBAF32374FD4567BULL, "hlac_ht_mansion_n_3365_1y.dds"},
	{0xCBAFC4B12F712773ULL, "vxlt_0_lowtown_a2_n_fhk_0_0.vlm"},
	{0xCBAFFAC1B2162D33ULL, "vxw_sundermount_2gq_0_0.vlm"},
	{0xCBB064C88B50333CULL, "6068860_m.fxe"},
	{0xCBB07D4A297D1D9BULL, "671726369.ogg"},
	{0xCBB0B0540DEFAAD8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\injured_supported_carver.evt"},
	{0xCBB0CF8638D93593ULL, "vxht_playershouse_n_1w4_0_0.vlm"},
	{0xCBB0F900B93F38B1ULL, "vxw_woundedcoast_1mj_0_0.vlm"},
	{0xCBB10ADC19261D15ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.mmh"},
	{0xCBB1B88CEA02C54AULL, "6145104_f.fxe"},
	{0xCBB239630DA3D881ULL, "hlac_lt_warehouse_-65536_2x.dds"},
	{0xCBB26AFBFC35EE44ULL, "6128078_m.fxe"},
	{0xCBB29A53F2398665ULL, "gencr_raider_archer_f.utc"},
	{0xCBB30E12C2AA7C9FULL, "6128532_m.fxe"},
	{0xCBB336516B83E2A2ULL, "hightown_male_noble1_sta6170867.cl"},
	{0xCBB36C08C4C6FE08ULL, "vxw_bonepitdestroyed_a3_1dc_0_0.vlm"},
	{0xCBB388690D46ADBDULL, "vxht_brothel_n_6uy_0_0.vlm"},
	{0xCBB3B3F8AB157148ULL, "172329554.ogg"},
	{0xCBB3EB72FC783E8EULL, "149606550.ogg"},
	{0xCBB3FBFD49042FD1ULL, "vxlt_gallowscourtyard_n_4lj_0_0.vlm"},
	{0xCBB3FF7F43C21D8AULL, "881348132.ogg"},
	{0xCBB411EB911A33EBULL, "zz_lgtpt_debug.ncs"},
	{0xCBB4D165C07CDD17ULL, "gen00fl_varric6087585.cl"},
	{0xCBB4DB9EFE853C96ULL, "6129152_m.fxe"},
	{0xCBB58F525A44E0AAULL, "gen00fl_fenris6108023.cl"},
	{0xCBB5B73C97E85C09ULL, "art\\vfx\\materials\\vfx_glyph_paralysis.mao"},
	{0xCBB5C980DCF2E6A7ULL, "art\\characters\\creatures\\mubari\\animation\\cwkbr.ani"},
	{0xCBB610C8FDFC1DBDULL, "art\\characters\\creatures\\mubari\\animation\\walkp.gad"},
	{0xCBB66005EBDDF957ULL, "vxlt_darktown_4li_0_0.vlm"},
	{0xCBB6659ABBB360B7ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_4y.dds"},
	{0xCBB6C07D9EF5B343ULL, "t1_mue_te3.tnt"},
	{0xCBB6D266A5A26B51ULL, "957420972.ogg"},
	{0xCBB70A36E5AF759EULL, "vxlt_darktown_2s7_0_0.vlm"},
	{0xCBB7B54B3BDA0C68ULL, "cd_pro000_opening_pc.wbd"},
	{0xCBB7C3111E850A3DULL, "vxw_sundermount_26l_0_0.vlm"},
	{0xCBB7DAA704EC7AD3ULL, "gen_im_npc_wep_arc_sbwb.uti"},
	{0xCBB80EAD5F5CB841ULL, "577819017.ogg"},
	{0xCBB81FF0AFD1A6D5ULL, "hlac_w_encounter2_65534_0x.dds"},
	{0xCBB837968E54B441ULL, "6041477_m.fxe"},
	{0xCBB868593F14F784ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_coverplane.mmh"},
	{0xCBB86C593F14FED7ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_coverplane.mao"},
	{0xCBB87E593F151D26ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_coverplane.msh"},
	{0xCBB8F443ED097D16ULL, "gen00fl_carver6104890.cl"},
	{0xCBB93551EFA77920ULL, "hlac_ca_caverns2_5802_12x.dds"},
	{0xCBB93F5F486B090FULL, "zz_mrl_debug6109308.cl"},
	{0xCBB97ED8034AAE90ULL, "dae200_cullen6140300.cl"},
	{0xCBB9F94BB41B84F4ULL, "6125356_m.fxe"},
	{0xCBBA00EB17242223ULL, "vxlt_gallowsprison_n_62b_0_0.vlm"},
	{0xCBBA0C11D0E9C1A6ULL, "hlac_lt_0_lowtown_65533_2y.dds"},
	{0xCBBA2AC528757690ULL, "vxw_woundedcoast_y6_0_0.vlm"},
	{0xCBBA30792896EB93ULL, "root_rog_general1.dds"},
	{0xCBBA839BF3D29C7DULL, "706463883.ogg"},
	{0xCBBACEF4DE58666EULL, "6126364_m.fxe"},
	{0xCBBAD34AB9DB88B7ULL, "vxlt_foundry_n_xg_0_0.vlm"},
	{0xCBBAEAED7979B7E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_action.evt"},
	{0xCBBC39D18F0E88F5ULL, "mag360_thrask6053540.cl"},
	{0xCBBC9EBF1F5AC08CULL, "vxht_0_hightown_auw_0_0.vlm"},
	{0xCBBD12C649E1F998ULL, "ht_0_hightown_62_0.rml"},
	{0xCBBE44FB75DC0B16ULL, "cod_hst_kirk_gallows.utp"},
	{0xCBBE537A022D7228ULL, "lt_0_lowtown_n_30_0.rml"},
	{0xCBBE6C9EA40C45B3ULL, "vxw_sundermount_2hp_0_0.vlm"},
	{0xCBBE85312F270CA7ULL, "6175835_m.fxe"},
	{0xCBBF07DC193255A2ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose3_p.anb"},
	{0xCBBF0BDE2C173019ULL, "428177778.ogg"},
	{0xCBBF291759416E39ULL, "vxlt_backalley_n_1xa_0_0.vlm"},
	{0xCBBF613F24D6BB16ULL, "420052223.ogg"},
	{0xCBBF6441F88D03C7ULL, "vxlt_gallowscourtyard_7b2_0_0.vlm"},
	{0xCBBF6ADDAB355131ULL, "vxw_woundedcoast_1hi_0_0.vlm"},
	{0xCBBFA59BD5228AE5ULL, "631439887.ogg"},
	{0xCBBFB0C82829002DULL, "6131575_f.fxe"},
	{0xCBBFBBD333F5E308ULL, "vxdr_primevaltomb_8mg_0_0.vlm"},
	{0xCBBFE62249086A1CULL, "static_static_p52.psh.fxo.dx11"},
	{0xCBC010F0B5086778ULL, "379783915.ogg"},
	{0xCBC02438432A36DAULL, "art\\levels\\races\\design\\prp_rubble_5.pwk"},
	{0xCBC03F6FFD8DF143ULL, "dae100_viscount6086789.cl"},
	{0xCBC04F563C559024ULL, "w_woundedcoast_167_0.phy"},
	{0xCBC0EAD6F8503F72ULL, "llac_lt_0_lowtown_docks_-4_7.dds"},
	{0xCBC1252F573594EBULL, "gen00fl_varric6118942.cl"},
	{0xCBC1414B607BE353ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_ornate.phy"},
	{0xCBC1542753F69F99ULL, "vxht_brothel_n_62p_0_0.vlm"},
	{0xCBC20BED79800E30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_action.gad"},
	{0xCBC269A91BE10899ULL, "hlac_lt_0_lowtown_a2_n_-2_0z.dds"},
	{0xCBC26BB774C69F2EULL, "vxht_chantry_2pm_0_0.vlm"},
	{0xCBC279C0D4B24643ULL, "vxw_bonepitdestroyed_a3_1g1_0_0.vlm"},
	{0xCBC2B62A1C8EEF4EULL, "929182522.ogg"},
	{0xCBC330E3B8972D75ULL, "6150007_m.fxe"},
	{0xCBC3611E359EF429ULL, "295904510.ogg"},
	{0xCBC36B0637E5CB7FULL, "hlac_lt_0_lowtown_-65536_1y.dds"},
	{0xCBC3B1CD3214F062ULL, "vxht_0_hightown_a2_c4x_0_0.vlm"},
	{0xCBC3F68647BB4C77ULL, "gen00fl_fenris6113932.cl"},
	{0xCBC506C15226B287ULL, "6164969_m.fxe"},
	{0xCBC53CACDEBA6661ULL, "794541702.ogg"},
	{0xCBC550527737F9E0ULL, "6017844_m.fxe"},
	{0xCBC555D2AAF2255EULL, "68800396.ogg"},
	{0xCBC5A2133F7ABA7BULL, "llac_lt_0_lowtown_a2_-196608_1.dds"},
	{0xCBC5AAC2F864F278ULL, "vxdr_primevalentrance_30j_0_1.vlm"},
	{0xCBC65B97F25A7FAAULL, "vxlt_0_lowtown_a2_gy3_0_0.vlm"},
	{0xCBC66A1F4092BC77ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookmisc_n.dds"},
	{0xCBC6C315E6F3D662ULL, "544245965.ogg"},
	{0xCBC70A95518FB93EULL, "6207373_m.fxe"},
	{0xCBC7A28062DD6006ULL, "264406350.ogg"},
	{0xCBC8120BA4514AA4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03b.msh"},
	{0xCBC8280BA4517046ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03b.mmh"},
	{0xCBC8470E6A1D6F80ULL, "vxlt_0_lowtown_a2_n_c_eqr_0_0.vlm"},
	{0xCBC848A000CBCE16ULL, "548039278.ogg"},
	{0xCBC880BA9C1B6A77ULL, "169111260.ogg"},
	{0xCBC89CA21A6B8F10ULL, "vxw_woundedcoast_v7_0_0.vlm"},
	{0xCBC90B56D2FD4C08ULL, "vxlt_0_lowtown_gm9_0_0.vlm"},
	{0xCBC929B54792B9F6ULL, "mag100_cullen6093732.cl"},
	{0xCBCA0C234B71F3CBULL, "carpetold_dp.dds"},
	{0xCBCA28CEE6BB8013ULL, "llac_w_encounter3_-262144_0.dds"},
	{0xCBCA7E83AE104BD5ULL, "6141247_m.fxe"},
	{0xCBCAB7FAF7164F4BULL, "6959053.ogg"},
	{0xCBCB15652A6D11E0ULL, "6051554_m.fxe"},
	{0xCBCB98A4349E5CB3ULL, "132244689.ogg"},
	{0xCBCBC85CE8AE61DDULL, "vxlt_darktown_4id_0_0.vlm"},
	{0xCBCBE9D5E2A12CFCULL, "vxht_0_hightown_n_ava_0_0.vlm"},
	{0xCBCCA96E7FFDB39BULL, "hlac_dr_primevalentrance_1725_0x.dds"},
	{0xCBCCDE7B96025E33ULL, "vxw_sundermount_2ms_0_0.vlm"},
	{0xCBCD010499EE56A2ULL, "989990545.ogg"},
	{0xCBCD0E5723C35856ULL, "vxht_0_hightown_b85_0_0.vlm"},
	{0xCBCD115D30EA389FULL, "hightown_market_guard_sta6169685.cl"},
	{0xCBCD3077D768B8E4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_reaction.ani"},
	{0xCBCDB0911F778010ULL, "lgt310_mis_burn_murder1.cnv"},
	{0xCBCDDCBA9D2B69B1ULL, "vxw_woundedcoast_1sh_0_0.vlm"},
	{0xCBCDF547159C0653ULL, "pro000_captain_wright6004377.cl"},
	{0xCBCEA73227A8C8F8ULL, "zz_one_debug6099931.cl"},
	{0xCBCEB6D5FCCB4720ULL, "903332222.ogg"},
	{0xCBCEE36610DBFB1EULL, "genip_exploding_invis.utp"},
	{0xCBCF5C749EF89F60ULL, "vxht_0_hightown_a3_b8v_0_0.vlm"},
	{0xCBCF6319B0652712ULL, "hlac_ht_keep_3739_1x.dds"},
	{0xCBCFA198C067087FULL, "art\\characters\\playercharacter\\qunarimale\\har_mag.mao"},
	{0xCBD0EB9DC6A85EC3ULL, "vxw_bonepitdestroyed_a3_1h0_0_0.vlm"},
	{0xCBD0F50A772DF739ULL, "vxlt_0_lowtown_docks_a3_aq4_0_0.vlm"},
	{0xCBD119CAAD38ED4FULL, "6032266_m.fxe"},
	{0xCBD1D2911F7BA368ULL, "lgt310_mis_burn_murder1.bnk"},
	{0xCBD1E1C4AC6C55B0ULL, "vxdr_primevaltomb_26i_0_1.vlm"},
	{0xCBD1E8B798AE31A3ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.impa.evt"},
	{0xCBD2F4C3760A828AULL, "6077321_m.fxe"},
	{0xCBD3115CBEF8224DULL, "dae000ip_crowd_ff3_talk10.utp"},
	{0xCBD3155301016917ULL, "vxlt_0_lowtown_a2_hgk_0_0.vlm"},
	{0xCBD31F0BA45B6811ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk03b.phy"},
	{0xCBD3425F3425B35CULL, "ico_shortbow_1.dds"},
	{0xCBD387B3F0B751AEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_debuff.evt"},
	{0xCBD396E92415D3B4ULL, "582780796.ogg"},
	{0xCBD3E0C7A7B3BED4ULL, "hlac_dr_0_deeproads_17038_4y.dds"},
	{0xCBD42D77D76ED246ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_reaction.gad"},
	{0xCBD4E11EB96D7522ULL, "vxlt_gallowstemplar_n_3x4_0_0.vlm"},
	{0xCBD4FE889E12C7CBULL, "gen_im_wep_war_one_act2_08.uti"},
	{0xCBD5497CB2B88466ULL, "zz_flemeth_transform.cut"},
	{0xCBD55844C84BC8CEULL, "6080129_m.fxe"},
	{0xCBD5D266752D4D47ULL, "vxdr_primevaltomb_869_0_0.vlm"},
	{0xCBD62451E3FED5ABULL, "hlac_ht_keep_n_3732_10y.dds"},
	{0xCBD6C4AB51252064ULL, "vxca_caverns1_mines_efw_0_0.vlm"},
	{0xCBD70E7F0C61A790ULL, "vxw_woundedcoast_s4_0_0.vlm"},
	{0xCBD79AA8D30F5FE2ULL, "665097036.ogg"},
	{0xCBD7E743C5797F03ULL, "134885022.ogg"},
	{0xCBD8C1B798B40E24ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.impa.gad"},
	{0xCBD8D66C64088B2CULL, "6153915_m.fxe"},
	{0xCBD9374D71526542ULL, "art\\cinematics\\props\\prp_bookinquisbinside.mao"},
	{0xCBD950D52D5EE784ULL, "hlac_w_bonepitdestroyed_a3_0_3y.dds"},
	{0xCBD9DB0C010CEBCEULL, "6093547_m.fxe"},
	{0xCBDA00A0742B1940ULL, "hlac_lt_undercityencounter_-1_31z.dds"},
	{0xCBDA081F7F1A97F0ULL, "vxlt_0_lowtown_docks_a3_n_brh_0_0.vlm"},
	{0xCBDA7EB3F0BD6159ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_debuff.gad"},
	{0xCBDA9FF4790B436EULL, "6089878_m.fxe"},
	{0xCBDAEDFF896F6B5DULL, "5262785.ogg"},
	{0xCBDB132AC44A7C0AULL, "1032625748.ogg"},
	{0xCBDB2877D774E825ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_reaction.evt"},
	{0xCBDC1F39F18215EEULL, "vxw_bonepit_1pq_0_0.vlm"},
	{0xCBDC2CF27CD0107CULL, "dae120ip_at_keep.utp"},
	{0xCBDC67478750BF86ULL, "qcr220_grey_wardens6060443.cl"},
	{0xCBDCC2F2A563C48BULL, "6085840_m.fxe"},
	{0xCBDD252637D54F46ULL, "43350744.ogg"},
	{0xCBDD8B4B70519105ULL, "vrc221_bartrand6044862.cl"},
	{0xCBDDB618E0903700ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs108.phy"},
	{0xCBDDCE5190EEB7E0ULL, "vxht_0_hightown_a3_b7w_0_0.vlm"},
	{0xCBDDFE7EC2B89BE5ULL, "6125762_m.fxe"},
	{0xCBDE024453509B41ULL, "vxdr_0_deeproads_9gs_0_0.vlm"},
	{0xCBDE31A21D57B715ULL, "6141931_m.fxe"},
	{0xCBDE34FAA9CB84FFULL, "6149861_m.fxe"},
	{0xCBDE3FE468F2A2CDULL, "t_table_05.dds"},
	{0xCBDE84E9558EC5C2ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_shd_reva_0d.dds"},
	{0xCBDF704297679976ULL, "507112772.ogg"},
	{0xCBDFD0B798BA4602ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_lp_c.impa.ani"},
	{0xCBE038D8FA34491AULL, "plt_mrl100pt_n_postquest.nss"},
	{0xCBE0BED71F64B222ULL, "vxlt_0_lowtown_fgt_0_0.vlm"},
	{0xCBE12F9DF53D5DB5ULL, "vxca_caverns2_6ng_0_0.vlm"},
	{0xCBE18FB3F0C39CA7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_ability_debuff.ani"},
	{0xCBE1B9927E6B33E8ULL, "gen00fl_varric6081652.cl"},
	{0xCBE1E6B414175622ULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_bridgelegs.msh"},
	{0xCBE1E8B4141759D8ULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_bridgelegs.mmh"},
	{0xCBE1F4B414176E7BULL, "art\\levels\\races\\natural\\blightlands\\vista\\prp_bridgelegs.mao"},
	{0xCBE23C21BA9585CAULL, "579748068.ogg"},
	{0xCBE25D42C2A89BF3ULL, "vxdr_primevaltomb_6po_0_2.vlm"},
	{0xCBE299C7FF39BE53ULL, "hlac_lt_0_lowtown_docks_n_65534_3x.dds"},
	{0xCBE2C08B8AAFCC6CULL, "vxht_mansion_n_4bd_0_0.vlm"},
	{0xCBE2D75A1DFD0D59ULL, "art\\characters\\playercharacter\\humanmale\\arm_meda.phy"},
	{0xCBE2E452A47251F4ULL, "lgt160_ntb_elf_conflict6163327.cl"},
	{0xCBE303132AF5905AULL, "dae000_bartrand6107531.cl"},
	{0xCBE35CA04B218FF6ULL, "vxht_chantry_n_2im_0_0.vlm"},
	{0xCBE362BDF10B740FULL, "6114067_m.fxe"},
	{0xCBE3818B1BD69C8AULL, "lgt000ip_corpse_static.utp"},
	{0xCBE3FD9182D0BB13ULL, "seb221_ruxton_harimann6177191.cl"},
	{0xCBE410FE821781E2ULL, "vfx_base_seb.gda"},
	{0xCBE51AF9AB5B71F6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medc_0n.dds"},
	{0xCBE59A42061B79A6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree02_dp.dds"},
	{0xCBE5C2D6302A073FULL, "vxlt_gallowsprison_n_6ph_0_0.vlm"},
	{0xCBE5FF0438ACAF8AULL, "dae000_gamlen6048416.cl"},
	{0xCBE61AF2F171476EULL, "mcr370_negotiation6078735.cl"},
	{0xCBE6E179CFC092E1ULL, "vxca_cavern1_haunted_ebz_0_0.vlm"},
	{0xCBE739400B74827AULL, "vxca_cave_19q_0_0.vlm"},
	{0xCBE7A6B2BA9B140DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_stairs_d.dds"},
	{0xCBE7B9F0A4E8E5B4ULL, "follower_banter6171891.cl"},
	{0xCBE8235CFCCC8516ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_5.evt"},
	{0xCBE869B743811C61ULL, "vxw_woundedcoast_1c9_0_1.vlm"},
	{0xCBE8A096BD640E83ULL, "6170476_m.fxe"},
	{0xCBE954FC93DB56E9ULL, "vxw_woundedcoast_b6_0_0.vlm"},
	{0xCBE975300FB336EEULL, "6147457_m.fxe"},
	{0xCBEA2E35997C4F0BULL, "320163002.ogg"},
	{0xCBEA5DE9436B0327ULL, "isa111_isabela6068962.cl"},
	{0xCBEAB36DC8F9641FULL, "ser210ip_to_dungeon.utp"},
	{0xCBEB0B0F87B366F7ULL, "vxht_0_hightown_a2_c85_0_0.vlm"},
	{0xCBEB45A3842A6D3DULL, "220057074.ogg"},
	{0xCBEB462093B60165ULL, "gen00fl_varric6081981.cl"},
	{0xCBEB946427CA200AULL, "lt_0_lowtown_n_5_0.rml"},
	{0xCBEBFD7C8CD11066ULL, "vxw_bonepit_29m_0_0.vlm"},
	{0xCBED0AA5A10CDCF2ULL, "hlac_w_blightlands_262140_0z.dds"},
	{0xCBED0D5C0933BBACULL, "6110305_m.fxe"},
	{0xCBED49276AE630CCULL, "hlac_w_sundermount_-262143_0z.dds"},
	{0xCBEDBA702C0BAFA5ULL, "vxht_chantry_n_2z7_0_0.vlm"},
	{0xCBEDC5F35FD34EA0ULL, "vxlt_undercitydungeon_1if_0_1.vlm"},
	{0xCBEDCB83162A241CULL, "6102895_f.fxe"},
	{0xCBEDCF6B64466EC6ULL, "vxca_caverns2_7gc_0_1.vlm"},
	{0xCBEDDF09A412A162ULL, "6153856_m.fxe"},
	{0xCBEDE49FC9BAC886ULL, "6064930_m.fxe"},
	{0xCBEE52CACDE90EEFULL, "ch2alphemmultnt.psh.fxo"},
	{0xCBEEAD3DA77E9BECULL, "6143446_m.fxe"},
	{0xCBEEC1AF80EB7885ULL, "lgt320_mis_resp_swordhunter6150334.cl"},
	{0xCBEF056411737519ULL, "w_woundedcoast_93_0.msh"},
	{0xCBEF0F6411738607ULL, "w_woundedcoast_93_0.mmh"},
	{0xCBEF136411738D54ULL, "w_woundedcoast_93_0.mao"},
	{0xCBEF1A5CFCD294A1ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_5.gad"},
	{0xCBEF30B4115ACAA2ULL, "vxlt_0_lowtown_fhu_0_0.vlm"},
	{0xCBEF4838088D8AF0ULL, "crt_stone_golem.bnk"},
	{0xCBEF6EC9FE915119ULL, "vxw_encounter2_kh_0_0.vlm"},
	{0xCBF022BFE4A1E929ULL, "lt_gallowstemplar.arl"},
	{0xCBF04A07C64C7A9EULL, "vxca_caverns1_mines_j1y_0_2.vlm"},
	{0xCBF09475AB37BDBFULL, "vxht_mansion_n_4hi_0_0.vlm"},
	{0xCBF0955FA79DA853ULL, "6039397_m.fxe"},
	{0xCBF0DBC89C4847FAULL, "vxht_mansion_n_6j8_0_0.vlm"},
	{0xCBF1C4D89D59A622ULL, "vxlt_gallowstemplar_n_3r6_0_0.vlm"},
	{0xCBF2518A0FC0821FULL, "6134949_f.fxe"},
	{0xCBF2EBB87A9FBCFEULL, "269584790.ogg"},
	{0xCBF32A83D3DB44C1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk02.phy"},
	{0xCBF39A425E5D31BEULL, "gencr_coterie_thug_2h.utc"},
	{0xCBF3A2CA0648901FULL, "llac_lt_0_lowtown_a2_n_c_65538_2.dds"},
	{0xCBF47F958C0CD3DEULL, "clf260_harley6093827.cl"},
	{0xCBF4A59147971A3DULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rockwraith_s1.anb"},
	{0xCBF55356C1B6AB61ULL, "vxca_cavern1_haunted_egy_0_0.vlm"},
	{0xCBF572F7AE4B0D32ULL, "6116378_m.fxe"},
	{0xCBF5C9238559A945ULL, "6036480_m.fxe"},
	{0xCBF5FFAB547AEA2CULL, "6118513_f.fxe"},
	{0xCBF61B5CFCD8B54FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_5.ani"},
	{0xCBF64B30EF696E23ULL, "w_sundermount_33_0.msh"},
	{0xCBF65930EF69867DULL, "w_sundermount_33_0.mmh"},
	{0xCBF65D30EF698C8EULL, "w_sundermount_33_0.mao"},
	{0xCBF6880E2B129B35ULL, "hlac_lt_0_lowtown_n_-1_7x.dds"},
	{0xCBF788A36FF50A4EULL, "gen00fl_aveline6106396.cl"},
	{0xCBF7CFA39E758F83ULL, "sse_ss_crazy_m_1.gda"},
	{0xCBF810CE65A7D947ULL, "vxht_0_hightown_n_cbp_0_0.vlm"},
	{0xCBF837B24323E5F8ULL, "734601515.ogg"},
	{0xCBF861FEFFCA8F7CULL, "w_bonepit_252_0.mao"},
	{0xCBF86DFEFFCAA31FULL, "w_bonepit_252_0.mmh"},
	{0xCBF873FEFFCAADC1ULL, "w_bonepit_252_0.msh"},
	{0xCBF90C9002FB212BULL, "pro000tr_gallows_ewald.utt"},
	{0xCBF982503797E2C4ULL, "vxca_caverns1_ckc_0_0.vlm"},
	{0xCBF98346FBCE4508ULL, "vxdr_primevaltomb_8yc_0_0.vlm"},
	{0xCBF98B7DFA0816BDULL, "combat_ws_r1_a3.gda"},
	{0xCBF9CB63ECA44A50ULL, "6151162_m.fxe"},
	{0xCBF9D81C53808640ULL, "fer_ter_grass_003_diff.dds"},
	{0xCBFAC870D3D7AEBCULL, "723877989.ogg"},
	{0xCBFAD7B20AFBF2F2ULL, "mrl210_merrill6102139.cl"},
	{0xCBFB2542EF54E20DULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_5x.dds"},
	{0xCBFB4B1BCA13B3A9ULL, "6156197_m.fxe"},
	{0xCBFB8F63319C296DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand.mmh"},
	{0xCBFB9363319C2FBEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand.mao"},
	{0xCBFBA163319C47F3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand.msh"},
	{0xCBFBB990E1C209DFULL, "one100_solivitus6048157.cl"},
	{0xCBFC0922A5C86D27ULL, "vxca_caverns1_jg1_0_0.vlm"},
	{0xCBFC53DDA2CE802EULL, "default_rogue_female.utc"},
	{0xCBFC7E7EA026BB94ULL, "isa311_isabela6092230.cl"},
	{0xCBFCA7CC2D34A2C0ULL, "vxlt_undercityencounter_1ta_0_0.vlm"},
	{0xCBFD37E2F9B31266ULL, "dae300_alistair_king6133583.cl"},
	{0xCBFD4E1EC72E99FFULL, "ser120cr_shade_cut3.utc"},
	{0xCBFE3E7035D42979ULL, "vxw_woundedcoast_12c_0_0.vlm"},
	{0xCBFE5C107EDD5EABULL, "art\\cinematics\\follower plots\\sebastian\\cs_player_sebastianintro.ani"},
	{0xCBFE76A5A24BC7ABULL, "dae000_pickpocket_scene.bnk"},
	{0xCBFE7BC3B424E45CULL, "471716433.ogg"},
	{0xCBFEDB4B3C164F2AULL, "cd_pro000_opening_pc.bnk"},
	{0xCBFEFCFB183D0903ULL, "vxdr_primevaltomb_7qs_0_0.vlm"},
	{0xCBFF1A0B61E52744ULL, "6101285_f.fxe"},
	{0xCBFF4DA58E3E607AULL, "vxht_mansion_n_6m9_0_0.vlm"},
	{0xCBFF8F5600E322C4ULL, "gen00fl_isabela_init6110227.cl"},
	{0xCBFF9DCF4442EF13ULL, "vxlt_undercityserialkiller_19m_0_0.vlm"},
	{0xCBFFC3E1805BDE40ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_4x.dds"},
	{0xCC005288944FB01FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_180.gad"},
	{0xCC00659A95198E22ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.mmh"},
	{0xCC007CE0B09B7559ULL, "dae021_sister_samea_sta6150170.cl"},
	{0xCC009C98C79C1496ULL, "162332634.ogg"},
	{0xCC00B3A46DCEF66DULL, "495844924.ogg"},
	{0xCC011AE43DE1A159ULL, "6139965_m.fxe"},
	{0xCC01413E1ABBB2C7ULL, "vxht_keep_5zy_0_0.vlm"},
	{0xCC0178A5A24E008BULL, "dae000_pickpocket_scene.cnv"},
	{0xCC021D5B49E58D15ULL, "hlac_lt_blackemporium_1055_9z.dds"},
	{0xCC02336FC8FD26A4ULL, "follower_banter6170777.cl"},
	{0xCC025376C810F60BULL, "6048850_m.fxe"},
	{0xCC027D2A2228FB0AULL, "6161071_m.fxe"},
	{0xCC02D318E0AF8495ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs108.msh"},
	{0xCC02D518E0AF880BULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs108.mmh"},
	{0xCC02E118E0AF9CA8ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs108.mao"},
	{0xCC034A0FCDD6CA47ULL, "vxw_bonepitdestroyed_a3_1s1_0_0.vlm"},
	{0xCC041A6DDFE4C009ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_flare01.dds"},
	{0xCC042C197FF61355ULL, "vxlt_0_lowtown_docks_a3_n_9r6_0_0.vlm"},
	{0xCC043B0EA74BDB77ULL, "6174560_m.fxe"},
	{0xCC045C46C60F8CACULL, "hla_ht_0_hightown_n_65534_0z.dds"},
	{0xCC04A115E1DE819FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1b.msh"},
	{0xCC04A715E1DE8C41ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1b.mmh"},
	{0xCC04B315E1DEA0E2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1b.mao"},
	{0xCC04DF0A4E68C902ULL, "gen00fl_fenris6126751.cl"},
	{0xCC05DE1E999F324EULL, "art\\levels\\races\\proxy\\props\\_textures\\coins_n.dds"},
	{0xCC066382A06B9FB6ULL, "hlac_lt_gallowstemplar_2391_3z.dds"},
	{0xCC06FC42804570A2ULL, "727170534.ogg"},
	{0xCC06FD7AABC22DF1ULL, "6197143_m.fxe"},
	{0xCC07028F4E344F00ULL, "6116318_m.fxe"},
	{0xCC0723060A51BF5BULL, "6122794_m.fxe"},
	{0xCC07378F9923F476ULL, "6172291_m.fxe"},
	{0xCC0747889455BC0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_180.evt"},
	{0xCC07964B511F6791ULL, "6028362_m.fxe"},
	{0xCC07F8ECFAD77731ULL, "lt_0_lowtown_docks_a3_237_0.mmh"},
	{0xCC0804ECFAD78BD2ULL, "lt_0_lowtown_docks_a3_237_0.mao"},
	{0xCC08083BCCB010B2ULL, "vxht_0_hightown_n_b8v_0_0.vlm"},
	{0xCC0812ECFAD7A3EFULL, "lt_0_lowtown_docks_a3_237_0.msh"},
	{0xCC084D73896BBB59ULL, "vxdr_primevaltomb_6x7_0_0.vlm"},
	{0xCC0890FF0BB98D43ULL, "vxdr_primevaltomb_8z1_0_0.vlm"},
	{0xCC08CD8B4A10D57EULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_002a_0d.dds"},
	{0xCC08F78B57727180ULL, "gen00fl_merrill6123750.cl"},
	{0xCC0907B440CBFD1DULL, "hlac_lt_0_lowtown_a2_n_-3_5z.dds"},
	{0xCC09296747C6B7E8ULL, "vxdr_primevalentrance_2l7_0_1.vlm"},
	{0xCC095B842D41DD60ULL, "and230_mage_girl6135814.cl"},
	{0xCC09736E7CB3CC34ULL, "338728629.ogg"},
	{0xCC0A19900309AADAULL, "pro000tr_gallows_ewald.ncs"},
	{0xCC0A9E2A0FF7E0A5ULL, "vxht_chantry_n_2p5_0_0.vlm"},
	{0xCC0AF4EE679E2CAAULL, "47639359.ogg"},
	{0xCC0BBC59AD2B4937ULL, "w_encounter3_187_0.mmh"},
	{0xCC0BC059AD2B5044ULL, "w_encounter3_187_0.mao"},
	{0xCC0BD259AD2B6EA9ULL, "w_encounter3_187_0.msh"},
	{0xCC0C342906844DD9ULL, "6160857_m.fxe"},
	{0xCC0C5283E27D8219ULL, "vxw_encounter2_qn_0_0.vlm"},
	{0xCC0C886331AA902AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand.phy"},
	{0xCC0CCF26FEE8AFECULL, "addpotions.ncs"},
	{0xCC0CFF6990D58931ULL, "hlac_lt_0_lowtown_a2_131072_2x.dds"},
	{0xCC0CFFDD44342F2CULL, "vxlt_warehouse_5p_0_0.vlm"},
	{0xCC0D2F3E11E07E62ULL, "6149554_m.fxe"},
	{0xCC0D500324678A4BULL, "lgt310_mag_melson6152128.cl"},
	{0xCC0DF00C3AE3B789ULL, "635963735.ogg"},
	{0xCC0E2A30B12326B3ULL, "vfx_duststream.bnk"},
	{0xCC0E384CBC32732AULL, "art\\levels\\ztools\\default\\generic\\2ca_basic.mao"},
	{0xCC0E3C4CBC327A79ULL, "art\\levels\\ztools\\default\\generic\\2ca_basic.mmh"},
	{0xCC0E464CBC328B67ULL, "art\\levels\\ztools\\default\\generic\\2ca_basic.msh"},
	{0xCC0E851D04B07ECEULL, "6118595_f.fxe"},
	{0xCC0E85E49466B39DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1m.mao"},
	{0xCC0E89E49466BAD6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1m.mmh"},
	{0xCC0E93E49466CBD4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1m.msh"},
	{0xCC0EA982D51A3EF1ULL, "art\\vfx\\combat\\vfx_w_gn1_build_up_c.impa.ani"},
	{0xCC0EDDD14C691D53ULL, "311639664.ogg"},
	{0xCC0EFD9406E963A4ULL, "967064455.ogg"},
	{0xCC0FDDE0049E0318ULL, "41606924.ogg"},
	{0xCC1005E282C85067ULL, "vxdr_0_deeproads_bd3_0_0.vlm"},
	{0xCC10C826423F8BF1ULL, "vxlt_backalley_n_1le_0_0.vlm"},
	{0xCC10EF2C3F864078ULL, "354707080.ogg"},
	{0xCC114C6C89967685ULL, "818972260.ogg"},
	{0xCC11D4274465709DULL, "159633011.ogg"},
	{0xCC12732231056790ULL, "dae000cr_grand_cleric6181922.cl"},
	{0xCC1343107EEF9035ULL, "art\\cinematics\\follower plots\\sebastian\\cs_player_sebastianintro.gad"},
	{0xCC13FBDDB2905D84ULL, "vxlt_undercityencounter_10y_0_0.vlm"},
	{0xCC1424753AC547D0ULL, "568022270.ogg"},
	{0xCC142A3808AC73BAULL, "crt_stone_golem.wbd"},
	{0xCC152F889461D165ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\left_180.ani"},
	{0xCC15A96C8B1A0CF6ULL, "6175391_m.fxe"},
	{0xCC15B482D5206F93ULL, "art\\vfx\\combat\\vfx_w_gn1_build_up_c.impa.gad"},
	{0xCC15C314CE3A59AFULL, "vxlt_warehouse_n_qj_0_0.vlm"},
	{0xCC160A09862951B6ULL, "vxw_bonepit_2a5_0_0.vlm"},
	{0xCC162B24354065F9ULL, "mcr371_post_harvester6182083.cl"},
	{0xCC16997566F8DE0DULL, "follower_banter6164757.cl"},
	{0xCC16BCE22A7AA141ULL, "mer210_seldrek6078792.cl"},
	{0xCC16C8790B0D3B74ULL, "w_encounter2_camp_12_0.phy"},
	{0xCC16C997698FEC01ULL, "hlac_lt_0_lowtown_a3_n_c_1_2y.dds"},
	{0xCC179D6AD8E247E3ULL, "gen00fl_isabela6196856.cl"},
	{0xCC17AB9E1845B27EULL, "vxca_caverns2_978_0_0.vlm"},
	{0xCC183D081D6093F5ULL, "vxlt_hangedman_n_2cz_0_0.vlm"},
	{0xCC188A8E7F6E6894ULL, "vxca_caverns1_mines_gkc_0_1.vlm"},
	{0xCC1897B1F7EC50B2ULL, "hlac_lt_0_lowtown_a2_n_c_65538_5x.dds"},
	{0xCC191A03DAAD60A6ULL, "6088679_m.fxe"},
	{0xCC19486B655EDEA5ULL, "6131889_f.fxe"},
	{0xCC195209693EA203ULL, "gen_im_trs_qun_lgt_glv_r1.uti"},
	{0xCC19CD7582FD609AULL, "6077659_m.fxe"},
	{0xCC1A2CB51B29F65DULL, "6068563_f.fxe"},
	{0xCC1A5FDDCA532BB3ULL, "hlac_w_woundedcoast_262151_0y.dds"},
	{0xCC1A8D479E4B2224ULL, "vxca_caverns1_id8_0_0.vlm"},
	{0xCC1AE6BD23B64AE2ULL, "506759600.ogg"},
	{0xCC1AEDC8B1222BF1ULL, "gen00fl_anders6125407.cl"},
	{0xCC1B222A19C05A79ULL, "vxw_woundedcoast_18a_0_0.vlm"},
	{0xCC1B9CB7A8EAA05CULL, "vxdr_primevaltomb_2pr_0_0.vlm"},
	{0xCC1BD36CA0BCD745ULL, "vxht_mansion_n_6j3_0_0.vlm"},
	{0xCC1BDCC67713C30EULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_ex.ani"},
	{0xCC1C313ACD78F4D7ULL, "w_encounter3_95_0.mmh"},
	{0xCC1C353ACD78FBE4ULL, "w_encounter3_95_0.mao"},
	{0xCC1C473ACD791A49ULL, "w_encounter3_95_0.msh"},
	{0xCC1C67C32A8000F0ULL, "gen00fl_merrill6119647.cl"},
	{0xCC1CA9A7D50E308AULL, "6156576_m.fxe"},
	{0xCC1CB182D52689F8ULL, "art\\vfx\\combat\\vfx_w_gn1_build_up_c.impa.evt"},
	{0xCC1CBB82B624938EULL, "fort_sp_ring.mao"},
	{0xCC1D50E5804A33D3ULL, "386328733.ogg"},
	{0xCC1D82F5227BF2EBULL, "vxw_woundedcoast_1g6_0_0.vlm"},
	{0xCC1DA0026BBCF9FAULL, "vxlt_undercitydungeon_15a_0_0.vlm"},
	{0xCC1DA7AF32E8F2A8ULL, "vxht_bartrand_n_72p_0_0.vlm"},
	{0xCC1DBF790B134A1CULL, "w_encounter2_camp_12_0.rml"},
	{0xCC1DC7B475E035E2ULL, "character_character_p117.psh.fxo.dx11.assembly"},
	{0xCC1E458F7EE911B8ULL, "llac_lt_undercityencounter_-1_35.dds"},
	{0xCC1E51E5B16026EAULL, "6070603_m.fxe"},
	{0xCC1EF5C581606E14ULL, "ave121_brennan6105887.cl"},
	{0xCC1F3A033435A471ULL, "vxlt_backalley_n_1wd_0_0.vlm"},
	{0xCC20A5790B155471ULL, "w_encounter2_camp_12_0.msh"},
	{0xCC20B3790B156B8CULL, "w_encounter2_camp_12_0.mao"},
	{0xCC20BF790B15802FULL, "w_encounter2_camp_12_0.mmh"},
	{0xCC2158BD1F7A526DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basicthin.mmh"},
	{0xCC215CBD1F7A58BEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basicthin.mao"},
	{0xCC216ABD1F7A70F3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basicthin.msh"},
	{0xCC2184AA3DAA2EF2ULL, "vxlt_gallowstemplar_f_5vt_0_0.vlm"},
	{0xCC218B336034983EULL, "vxlt_gallowstemplar_n_2j7_0_0.vlm"},
	{0xCC21E403F338B51BULL, "vxlt_hangedman_n_2xq_0_0.vlm"},
	{0xCC220E4B4D78CDB2ULL, "vxdr_primevalentrance_31v_0_0.vlm"},
	{0xCC2292C8173A6361ULL, "isa311_isabela6175480.cl"},
	{0xCC23A9038DBCBF83ULL, "6107806_f.fxe"},
	{0xCC23AE0302F484FDULL, "vxca_caverns1_mines_j3r_0_0.vlm"},
	{0xCC2434F1C030722FULL, "vxlt_warehouse_n_tu_0_0.vlm"},
	{0xCC243A4B60D054C8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_ornate.mmh"},
	{0xCC24464B60D0696BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_ornate.mao"},
	{0xCC24521A0752DB88ULL, "6170501_m.fxe"},
	{0xCC24584B60D087F2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_ornate.msh"},
	{0xCC24951EAB54DC8AULL, "hlac_lt_0_lowtown_a2_n_c_-196608_3x.dds"},
	{0xCC24A2D192209D0BULL, "gen00fl_anders6142558.cl"},
	{0xCC24BBDFB7325D21ULL, "795644709.ogg"},
	{0xCC24EC636C167E03ULL, "6016839_m.fxe"},
	{0xCC2526376F8F0BCAULL, "vxlt_0_lowtown_a3_n_c_byr_0_0.vlm"},
	{0xCC252F4C7AB45802ULL, "lowtown_fem_alienage1_sta6162607.cl"},
	{0xCC2541391A5B6657ULL, "vxlt_warehouse_n_8i_0_0.vlm"},
	{0xCC255DBBFD8B2EDCULL, "llac_w_woundedcoast_-65533_0.dds"},
	{0xCC2585192C009270ULL, "815002985.ogg"},
	{0xCC25A295E0F9FAAAULL, "965614551.ogg"},
	{0xCC261D7B0A3BCAFEULL, "vxca_caverns2_989_0_0.vlm"},
	{0xCC262D875FE0A8DFULL, "best_served_dae000_meredith.stg"},
	{0xCC262F4990A1312BULL, "6171826_m.fxe"},
	{0xCC264946FDF9289FULL, "genip_cin_book_inquisition.utp"},
	{0xCC267A154D4A578BULL, "48175590.ogg"},
	{0xCC26A23AC91EA366ULL, "llac_lt_gallowscourtyard_-65537_3.dds"},
	{0xCC26D0F48D742BFAULL, "6165985_m.fxe"},
	{0xCC26DA33AE74FF75ULL, "lt_0_lowtown_docks_144_0.msh"},
	{0xCC26DC33AE7502EBULL, "lt_0_lowtown_docks_144_0.mmh"},
	{0xCC26E833AE751688ULL, "lt_0_lowtown_docks_144_0.mao"},
	{0xCC26FAB80F88C3E4ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rockblend_d.dds"},
	{0xCC27127E4D574409ULL, "6105077_f.fxe"},
	{0xCC272738D9E79978ULL, "vxca_caverns1_mines_k2q_0_0.vlm"},
	{0xCC273B909F2DEEEBULL, "478360225.ogg"},
	{0xCC273B954C6A8749ULL, "6105833_f.fxe"},
	{0xCC27530B932C0E51ULL, "mag320_orsino6060231.cl"},
	{0xCC27699D91C0FC30ULL, "702735888.ogg"},
	{0xCC276BFF29110D69ULL, "166195548.ogg"},
	{0xCC27C30BD220FC88ULL, "gen00fl_aveline6115497.cl"},
	{0xCC27C551204A985BULL, "756114667.ogg"},
	{0xCC27DB18CE66C6B7ULL, "gen00fl_aveline6053606.cl"},
	{0xCC282D5B7EF825C9ULL, "hlac_lt_gallowscourtyard_n_-131071_2z.dds"},
	{0xCC287C33A9087E83ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_enter.ani"},
	{0xCC28B51916496352ULL, "104139743.ogg"},
	{0xCC28D49357412D14ULL, "vxlt_0_lowtown_docks_n_bgl_0_0.vlm"},
	{0xCC291833BA69C538ULL, "vxca_caverns1_mines_cmt_0_0.vlm"},
	{0xCC29649A953C91C9ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.anb"},
	{0xCC296B94B7AAD4FDULL, "mcr370_meredith6147460.cl"},
	{0xCC2A454992B2EFC5ULL, "vxht_mansion_n_6m2_0_0.vlm"},
	{0xCC2A996F8E0A80B7ULL, "zz_isa_debug6175240.cl"},
	{0xCC2AEF249FA7CB76ULL, "vxlt_0_lowtown_a2_n_c_h2h_0_0.vlm"},
	{0xCC2C11DF5DB3127AULL, "vxlt_undercitydungeon_18b_0_0.vlm"},
	{0xCC2C81C4B5CDD546ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a_2.mao"},
	{0xCC2CBE3D2EA69397ULL, "vxht_0_hightown_a2_b9c_0_0.vlm"},
	{0xCC2CCE7D900577E6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\idle.ani"},
	{0xCC2CFDFE0318063EULL, "vxht_0_hightown_a3_ajm_0_0.vlm"},
	{0xCC2D57D12A43A196ULL, "llac_ca_cavern1_haunted_24422_2.dds"},
	{0xCC2DA60915ACAC26ULL, "6127600_m.fxe"},
	{0xCC2DB18B0FD9F4EBULL, "vxlt_darktown_59w_0_0.vlm"},
	{0xCC2E5069D29FACDEULL, "vxlt_0_lowtown_a2_coy_0_0.vlm"},
	{0xCC2E580677BDAD23ULL, "lgt330ip_shw_ors_to_spl_dpt.utp"},
	{0xCC2E9FA6A3B913C7ULL, "vxw_bonepitdestroyed_a3_1b2_0_0.vlm"},
	{0xCC2EB7B6FDA924FFULL, "6042030_m.fxe"},
	{0xCC2EBAC88A4AD6AAULL, "art\\vfx\\materials\\chunky\\vfx_c_spi_hand.mao"},
	{0xCC2EDFF894F176ACULL, "787106948.ogg"},
	{0xCC2EE316BC1FA571ULL, "hlac_lt_darktown_-1_43x.dds"},
	{0xCC2F4952C6BD4F51ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_reduction.ani"},
	{0xCC2FB5C677242A48ULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_ex.gad"},
	{0xCC2FB91EA7B000F3ULL, "236441492.ogg"},
	{0xCC2FF6872FA04772ULL, "vxlt_gallowstemplar_f_5yu_0_0.vlm"},
	{0xCC3065D7ECF8B871ULL, "vxw_blightlands_143_0_0.vlm"},
	{0xCC30766D0B564C55ULL, "hlac_lt_0_lowtown_a3_n_c_-1_8x.dds"},
	{0xCC309F70A4E1974FULL, "559440130.ogg"},
	{0xCC30A815E203C8BEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_table1b.phy"},
	{0xCC30C687705FB61CULL, "6819145.ogg"},
	{0xCC30C8201AF8689FULL, "sse_ss_midclass_thugs_f_1.gda"},
	{0xCC312C92DEAF1AC2ULL, "452530887.ogg"},
	{0xCC3149A212120846ULL, "402747963.ogg"},
	{0xCC3170052A261598ULL, "vxlt_0_lowtown_n_eru_0_0.vlm"},
	{0xCC31736B096E76CAULL, "zz_dog_debug6127915.cl"},
	{0xCC31C2CFBF3526D8ULL, "6121210_f.fxe"},
	{0xCC31CC274733AA3CULL, "vxw_encounter2_k1_0_0.vlm"},
	{0xCC31EC6F1A7536B7ULL, "844572795.ogg"},
	{0xCC320EC50E81FF99ULL, "6170799_m.fxe"},
	{0xCC3215EB104A4C36ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02b.phy"},
	{0xCC3239973A457D55ULL, "vxlt_0_lowtown_docks_b5m_0_1.vlm"},
	{0xCC323A3575AAC134ULL, "gen00fl_anders6119681.cl"},
	{0xCC324579A68A289AULL, "6169800_m.fxe"},
	{0xCC3251BD1F88B92AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basicthin.phy"},
	{0xCC328A97677C8D65ULL, "642545359.ogg"},
	{0xCC32F47EBD8920A8ULL, "mag101_alain6052994.cl"},
	{0xCC32FFE7F48228A1ULL, "6041496_f.fxe"},
	{0xCC330A74D111E28DULL, "gen00fl_fenris6126989.cl"},
	{0xCC33600E87120DC7ULL, "hlac_dr_dungeon_3214_8y.dds"},
	{0xCC339D8D6026DE8CULL, "hm_qun_varnell.mor"},
	{0xCC340505A2AFEE1AULL, "1013581667.ogg"},
	{0xCC340C4693677131ULL, "mag100_thrask6031979.cl"},
	{0xCC34599F69DBA9C4ULL, "vxw_encounter2_ca_0_0.vlm"},
	{0xCC346AF57A3667D7ULL, "6045036_f.fxe"},
	{0xCC3497EE0C5D6049ULL, "dae000_flemeth6001334.cl"},
	{0xCC355452C6C1CDF3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_reduction.gad"},
	{0xCC358870139703DBULL, "qcr220_grey_wardens6207463.cl"},
	{0xCC359433A913327AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_enter.evt"},
	{0xCC35E2C91121040AULL, "character_character_p120.psh.fxo.dx11"},
	{0xCC3630F5790CDF5EULL, "zz_fol_debug6056991.cl"},
	{0xCC36AE06E06105CCULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardright.ani"},
	{0xCC36CE94B6CADECEULL, "dae000ip_crowd_mf1_talk3.utp"},
	{0xCC36E29392CE6418ULL, "zz_spy133cr_disciple.utc"},
	{0xCC3725B93A143D22ULL, "art\\vfx\\textures\\vfx_branch.dds"},
	{0xCC3795E8903C5889ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrlb_0n.dds"},
	{0xCC37AAD2FE9BB919ULL, "follower_banter6136444.cl"},
	{0xCC37BBCB742D78A3ULL, "6128774_m.fxe"},
	{0xCC37D5743CB086B1ULL, "mer210_smith6091994.cl"},
	{0xCC37EFFC6A403CFEULL, "233584980.ogg"},
	{0xCC37F3B19A186617ULL, "vxca_caverns2_830_0_0.vlm"},
	{0xCC384A262C67B4A8ULL, "follower_banter6149315.cl"},
	{0xCC3892120CA1CFACULL, "401341216.ogg"},
	{0xCC38B928F33FD39FULL, "pro000_colonel_ewald6039979.cl"},
	{0xCC38FEF332E00B39ULL, "vxlt_0_lowtown_docks_a3_bse_0_0.vlm"},
	{0xCC392EEEC32D2902ULL, "6066618_f.fxe"},
	{0xCC3992C43D1E596BULL, "vxlt_undercitydungeon_ym_0_0.vlm"},
	{0xCC39ADEBC1C12AD6ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef03.phy"},
	{0xCC39C3D5A02BE6BDULL, "vxw_sundermount_2cc_0_0.vlm"},
	{0xCC3A35874197356CULL, "gen_im_npc_wep_sta_genc.uti"},
	{0xCC3AB18C214CE833ULL, "910001998.ogg"},
	{0xCC3B1F891DB4660EULL, "155220117.ogg"},
	{0xCC3B20432B7CF2B5ULL, "447426378.ogg"},
	{0xCC3B36C25C83F84EULL, "255851206.ogg"},
	{0xCC3B47F4D0CC8A55ULL, "gen00fl_anders6125735.cl"},
	{0xCC3B5688C6BBE58DULL, "6172638_m.fxe"},
	{0xCC3B7A1276B7BE79ULL, "lt_0_lowtown_a2_n_45_0_lc.mmh"},
	{0xCC3B94666FE8A2EBULL, "6033316_f.fxe"},
	{0xCC3BFEEB105279D7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02b.msh"},
	{0xCC3C14EB10529F49ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02b.mmh"},
	{0xCC3C5152C6C7E758ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_reduction.evt"},
	{0xCC3C55A011DF1243ULL, "1034321331.ogg"},
	{0xCC3C63D4957AFCB9ULL, "ser270_emeric6149343.cl"},
	{0xCC3C9333A9194F6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_2_enter.gad"},
	{0xCC3C9DD62FF36C82ULL, "vxlt_darktown_4f2_0_0.vlm"},
	{0xCC3D0E9B6A3DD7F3ULL, "6150328_m.fxe"},
	{0xCC3D2602336DAFBEULL, "vxht_0_hightown_cbf_0_0.vlm"},
	{0xCC3DBB06E0673A0EULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardright.gad"},
	{0xCC3E117D1FE47936ULL, "dae000ip_crowd_ff2_talk14.utp"},
	{0xCC3E9A0A13DE621BULL, "372529746.ogg"},
	{0xCC3E9A1B16E0EDEEULL, "vxca_cavern1_haunted_co8_0_0.vlm"},
	{0xCC3E9BC9FF9FAEE6ULL, "6143498_f.fxe"},
	{0xCC3EE144545938F6ULL, "follower_banter6136694.cl"},
	{0xCC3EE3DE023F1A3FULL, "572230563.ogg"},
	{0xCC3EECF0CF8CF333ULL, "6042742_m.fxe"},
	{0xCC3F1FF908DF4722ULL, "68598813.ogg"},
	{0xCC3FE1E21C1C2E18ULL, "vxlt_0_lowtown_n_eor_0_0.vlm"},
	{0xCC40D7E0C38DEE6AULL, "lgt200cr_gang_1_ht.utc"},
	{0xCC4161EC9C3AB78FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_signal_to_wait_look_behind.gad"},
	{0xCC41894BF712B8B0ULL, "lt_0_lowtown_docks_a3_n_224_0.msh"},
	{0xCC41974BF712D0CAULL, "lt_0_lowtown_docks_a3_n_224_0.mmh"},
	{0xCC419B4BF712D701ULL, "lt_0_lowtown_docks_a3_n_224_0.mao"},
	{0xCC41B77D9017ACE0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\partypicker\\idle.gad"},
	{0xCC41EF8C4E85C905ULL, "6072848_m.fxe"},
	{0xCC42A3685E9B1FCAULL, "llac_w_bonepitdestroyed_a3_65539_0.dds"},
	{0xCC42CB7C5BD1C244ULL, "vxw_encounter2_nf_0_0.vlm"},
	{0xCC432A786032F940ULL, "686545545.ogg"},
	{0xCC43B606E06B9DEDULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforwardright.evt"},
	{0xCC43BF1C21CC38BCULL, "art\\levels\\races\\proxy\\props\\religious\\prp_zzaltar.mao"},
	{0xCC43CB1C21CC4D5FULL, "art\\levels\\races\\proxy\\props\\religious\\prp_zzaltar.mmh"},
	{0xCC43D11C21CC5701ULL, "art\\levels\\races\\proxy\\props\\religious\\prp_zzaltar.msh"},
	{0xCC43DA4010501952ULL, "llac_lt_gallowscourtyard_n_131071_0.dds"},
	{0xCC440662DEF054F0ULL, "vxht_bartrand_n_5wu_0_0.vlm"},
	{0xCC444C1E8B298A09ULL, "vxw_encounter2_camp_l5_0_0.vlm"},
	{0xCC449099FE8399CEULL, "6087206_f.fxe"},
	{0xCC4496EBC1CB0B77ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef03.msh"},
	{0xCC44ACEBC1CB30E9ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef03.mmh"},
	{0xCC450113A1707534ULL, "6146809_m.fxe"},
	{0xCC4555029205A4F0ULL, "6080699_m.fxe"},
	{0xCC4555D8FA897A8DULL, "plt_mrl100pt_n_postquest.plo"},
	{0xCC459F91D8E39CC3ULL, "art\\characters\\creatures\\dragon\\animation\\cwfl.ani"},
	{0xCC4630D5FC56A06EULL, "vxht_0_hightown_a2_cb2_0_0.vlm"},
	{0xCC46518D89ACFF5BULL, "6088677_f.fxe"},
	{0xCC46C51B012A6F6BULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fl.evt"},
	{0xCC4776EC9C3F473CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_signal_to_wait_look_behind.evt"},
	{0xCC47887A6D0CE0D4ULL, "dr_0_deeproads_5838_ll.mmh"},
	{0xCC48151C2A50716CULL, "214002151.ogg"},
	{0xCC482A9517CEEB4AULL, "hlac_lt_0_lowtown_a2_n_c_-196609_0z.dds"},
	{0xCC4835B29221FF3DULL, "vxw_sundermount_2db_0_0.vlm"},
	{0xCC4864111D340D41ULL, "vxlt_gallowsprison_n_628_0_0.vlm"},
	{0xCC491689DF8929CAULL, "isa120_hayder.cnv"},
	{0xCC4955DC2F1BD91AULL, "6161369_m.fxe"},
	{0xCC49C7AEC0D8F4D0ULL, "art\\levels\\4cinematics\\qunari\\prp_qunarirelic.mao"},
	{0xCC49CBAEC0D8FBE3ULL, "art\\levels\\4cinematics\\qunari\\prp_qunarirelic.mmh"},
	{0xCC49D9AEC0D9133DULL, "art\\levels\\4cinematics\\qunari\\prp_qunarirelic.msh"},
	{0xCC4A0777F8C974FEULL, "hlac_lt_backalley_n_65537_6y.dds"},
	{0xCC4A20116A307EB1ULL, "dae000_mother6099173.cl"},
	{0xCC4A39F7E11E8A59ULL, "vxw_bonepit_kx_0_0.vlm"},
	{0xCC4A89C2F998DD26ULL, "6129800_m.fxe"},
	{0xCC4A93179F8BC41FULL, "vxlt_0_lowtown_ha1_0_0.vlm"},
	{0xCC4A95DF0C91C790ULL, "36460678.ogg"},
	{0xCC4AF97874C43E9AULL, "vxw_encounter2_5u_0_0.vlm"},
	{0xCC4AFAD815CDFBA0ULL, "dae200cr_sandal.utc"},
	{0xCC4BB0479697B737ULL, "ht_0_hightown_n_51_0.anb"},
	{0xCC4BB147BC62B2EDULL, "580426440.ogg"},
	{0xCC4BF889DF8B2CF2ULL, "isa120_hayder.bnk"},
	{0xCC4C2B62F4858599ULL, "6117518_m.fxe"},
	{0xCC4C481178A01BA0ULL, "666573289.ogg"},
	{0xCC4C58486E8F383BULL, "mage_dispelmagic.dds"},
	{0xCC4CA8C2CD4D99A6ULL, "vxht_chantry_n_20b_0_0.vlm"},
	{0xCC4CA9C97CF4D51EULL, "628314033.ogg"},
	{0xCC4CB79B078B587AULL, "gen00fl_isabela6079693.cl"},
	{0xCC4CDE1B012F052CULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fl.gad"},
	{0xCC4D3ADC30957156ULL, "mag101_grace.cnv"},
	{0xCC4D4AB436388556ULL, "821653850.ogg"},
	{0xCC4D6F49DE5B7204ULL, "qun211_braggart6075481.cl"},
	{0xCC4DA9AE6193E349ULL, "17061271.ogg"},
	{0xCC4E2AF4EFE6BA29ULL, "vxca_cave_2j4_0_0.vlm"},
	{0xCC4E3534D2227BB8ULL, "vxht_0_hightown_a6n_0_0.vlm"},
	{0xCC4E47EB37E8F33BULL, "co_06.mao"},
	{0xCC4E8800ADDB5536ULL, "hlac_lt_gallowstemplar_f_7198_2z.dds"},
	{0xCC4E9D7C1F61D77DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_generalwood_d.dds"},
	{0xCC4F9E0A2C3608BFULL, "444305416.ogg"},
	{0xCC4FABB572EC5A7EULL, "vxlt_0_lowtown_a3_n_c_deh_0_0.vlm"},
	{0xCC4FC5CFFE609BD4ULL, "dae021_male_worshipper_sta6150039.cl"},
	{0xCC503CDC3097AA36ULL, "mag101_grace.bnk"},
	{0xCC50C8A6C824BC7DULL, "6057034_m.fxe"},
	{0xCC513D594DC7DAC4ULL, "vxw_encounter2_ig_0_0.vlm"},
	{0xCC516AD323C6B1C5ULL, "mag170_thrask6016010.cl"},
	{0xCC518F33063807E9ULL, "vxw_bonepit_1m1_0_0.vlm"},
	{0xCC520C47505E871AULL, "951063198.ogg"},
	{0xCC522967CEE0B990ULL, "vxca_caverns1_eh4_0_0.vlm"},
	{0xCC52BDFB7D1FA289ULL, "vxw_encounter2_camp_g4_0_0.vlm"},
	{0xCC5301E898F49519ULL, "dae000_merchant_rare_em6177951.cl"},
	{0xCC53743F52FA70F3ULL, "641466549.ogg"},
	{0xCC53B791D8F002BAULL, "art\\characters\\creatures\\dragon\\animation\\cwfl.evt"},
	{0xCC53DC7E745A67C3ULL, "zz_vault_debug6200383.cl"},
	{0xCC53DD1B0135222AULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_fl.ani"},
	{0xCC543225A961DAEEULL, "hlac_lt_0_lowtown_a2_n_-262144_3y.dds"},
	{0xCC5501D1E5F1BB13ULL, "ico_staff_4.dds"},
	{0xCC553383D42E1E36ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk02.mmh"},
	{0xCC553D83D42E2F34ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hk02.msh"},
	{0xCC554830B15F5C11ULL, "vfx_duststream.wbd"},
	{0xCC557EEC9C4B9255ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_signal_to_wait_look_behind.ani"},
	{0xCC55B1046481C032ULL, "gen00pt_money.ncs"},
	{0xCC55F61DA132E0B4ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_act_c.anb"},
	{0xCC56A78F841817BDULL, "vxw_sundermount_2ia_0_0.vlm"},
	{0xCC56BA35491959BAULL, "vxlt_backalley_n_1jt_0_0.vlm"},
	{0xCC56EC6DF95C1832ULL, "isa311_isabela6092302.cl"},
	{0xCC5709836227A1C5ULL, "vxlt_undercitydungeon_18i_0_0.vlm"},
	{0xCC57187C765046F5ULL, "vxlt_darktown_5gg_0_0.vlm"},
	{0xCC57EC2023E4F24CULL, "dae021_torelin_mob6147909.cl"},
	{0xCC582A69CCC6CCBCULL, "461560899.ogg"},
	{0xCC584B76E84D6F8FULL, "6072819_m.fxe"},
	{0xCC586D4DED25DC64ULL, "vxca_caverns2prc_8a4_0_0.vlm"},
	{0xCC58E99002662BA1ULL, "vxlt_0_lowtown_a2_n_c_gxf_0_0.vlm"},
	{0xCC58FBFF61BEBDC1ULL, "109781032.ogg"},
	{0xCC59005E3A45EF33ULL, "6208183_f.fxe"},
	{0xCC594652C2E713E7ULL, "dae000_hired_prostitute.bnk"},
	{0xCC59B691D8F46CADULL, "art\\characters\\creatures\\dragon\\animation\\cwfl.gad"},
	{0xCC5A4706CEFD18A7ULL, "vxlt_0_lowtown_docks_coi_0_0.vlm"},
	{0xCC5A6CFB432E8E9AULL, "6111186_m.fxe"},
	{0xCC5A9D3FE67B9E19ULL, "6112618_m.fxe"},
	{0xCC5AA30B6B489EFBULL, "llac_lt_0_lowtown_docks_a3_n_196609_0.dds"},
	{0xCC5AEAA6E37BDAEEULL, "m_one300ar_crow_camp.wbd"},
	{0xCC5B1B62D2B95C87ULL, "6178388_m.fxe"},
	{0xCC5B4C1E058290F2ULL, "vxlt_gallowstemplar_f_5jp_0_0.vlm"},
	{0xCC5B85FA4A802723ULL, "6044072_m.fxe"},
	{0xCC5B87E2953149F0ULL, "dae300_alistair_king6134975.cl"},
	{0xCC5BB5617B315011ULL, "hlac_dr_dungeon_3217_1y.dds"},
	{0xCC5BC3E442F566CAULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue_e.dds"},
	{0xCC5C22C78BCE0B61ULL, "vxlt_undercitydungeon_134_0_0.vlm"},
	{0xCC5C5D52C2E9700CULL, "dae000_hired_prostitute.cut"},
	{0xCC5CA22DDA9BF5BDULL, "smoke_ciga_f.dds"},
	{0xCC5E3EC613C1185BULL, "wood.dds"},
	{0xCC5E417C1436F0F4ULL, "hlac_w_bonepit_2_0z.dds"},
	{0xCC5E4D05E6B578CEULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_002a_0t.dds"},
	{0xCC5F7D398E181216ULL, "vxlt_0_lowtown_a3_n_c_fh5_0_0.vlm"},
	{0xCC5FD2B251FF90A7ULL, "203021353.ogg"},
	{0xCC5FE08C5AF32C89ULL, "694411910.ogg"},
	{0xCC605A7225C3A2DCULL, "w_sundermount_189_0.mao"},
	{0xCC60667225C3B77FULL, "w_sundermount_189_0.mmh"},
	{0xCC606C7225C3C121ULL, "w_sundermount_189_0.msh"},
	{0xCC614574B7BCFD8BULL, "hlac_lt_warehouse_-65536_2z.dds"},
	{0xCC61D681F4F4F07EULL, "character_character_p56.psh.fxo.dx11"},
	{0xCC623A60D9806B78ULL, "vxlt_0_lowtown_a2_n_h28_0_0.vlm"},
	{0xCC62513E3E33D281ULL, "ico_mm_plotgiver.dds"},
	{0xCC63BE52063E20D0ULL, "lgt150_nat_were_supporter6111001.cl"},
	{0xCC63C7294F0FB01FULL, "387781575.ogg"},
	{0xCC63D51DA13EE677ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_act_c.mmh"},
	{0xCC63DBB6309CCDB3ULL, "477760792.ogg"},
	{0xCC63E9CA8CEB4647ULL, "w_bonepit_24_0_ll.mmh"},
	{0xCC643121DDEDB6D7ULL, "zz_ave_debug6064631.cl"},
	{0xCC649E46418CDE8FULL, "llac_lt_gallowstemplar_f_2388_5.dds"},
	{0xCC64AA6054320BBBULL, "hlac_ht_mansion2_818_13y.dds"},
	{0xCC64C1BF83EC54BBULL, "clf260_harley6069582.cl"},
	{0xCC6544CB162D5935ULL, "mag360_alain6044661.cl"},
	{0xCC6558304D702850ULL, "vxca_caverns1_mines_imj_0_0.vlm"},
	{0xCC65822F8F387033ULL, "6149160_m.fxe"},
	{0xCC659494F1555001ULL, "hlac_lt_0_lowtown_-3_0x.dds"},
	{0xCC65EDF5830B8920ULL, "plt_qun200pt_missing_qunari.nss"},
	{0xCC660ADF36531D40ULL, "6147437_m.fxe"},
	{0xCC672CC322E53AA2ULL, "hlac_ca_caverns1_24424_25z.dds"},
	{0xCC67449D6614CBF2ULL, "isa300im_documents.uti"},
	{0xCC6860BCEDC03AABULL, "mag120_macha6080617.cl"},
	{0xCC68C4462311CB02ULL, "6162489_m.fxe"},
	{0xCC690D7C5BB53678ULL, "172660846.ogg"},
	{0xCC69A712BF0367FFULL, "hlac_lt_0_lowtown_docks_n_-2_3z.dds"},
	{0xCC6A079296DEDD42ULL, "follower_banter6153967.cl"},
	{0xCC6A56E58640C945ULL, "w_blightlands_21_0.rml"},
	{0xCC6ACED1BD61BB66ULL, "6114476_f.fxe"},
	{0xCC6AFE2D9728C844ULL, "6103404_player_stern.cut"},
	{0xCC6B002F345F21C6ULL, "vxlt_0_lowtown_n_fiy_0_0.vlm"},
	{0xCC6B22D824F18273ULL, "hlac_lt_0_lowtown_docks_a3_-4_6y.dds"},
	{0xCC6B3C36592ACBC9ULL, "6088931_m.fxe"},
	{0xCC6C0345B70E9721ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelopen_0.pwk"},
	{0xCC6C1C048924A10AULL, "gen00fl_dog6167687.cl"},
	{0xCC6CD6464C10D3C7ULL, "hlac_ca_cavern1_haunted_24423_1y.dds"},
	{0xCC6CE3F773FE4DA5ULL, "vxlt_warehouse_1cv_0_0.vlm"},
	{0xCC6D00740A6F5C34ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev1.gad"},
	{0xCC6D278D7A15EF0DULL, "cod_mgc_chant_laws.utp"},
	{0xCC6E3AF281D3F1BCULL, "hlac_lt_0_lowtown_a2_n_c_1_9z.dds"},
	{0xCC6EB29FA5C37A97ULL, "1034216514.ogg"},
	{0xCC6EB91072AFCBFFULL, "zz_vault_debug6200358.cl"},
	{0xCC6EBB2C8BE36682ULL, "6149193_f.fxe"},
	{0xCC6EC098D6AFF8D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_wobble.ani"},
	{0xCC6EFFBE8D64505DULL, "zz_dre_debug6056950.cl"},
	{0xCC6F77CEC56FC45EULL, "vfx_chunkyparticle_p3.psh.fxo.dx11.assembly"},
	{0xCC6FAEBAEABE1D96ULL, "vxca_caverns1_i7c_0_0.vlm"},
	{0xCC6FAED67565C6A1ULL, "hlac_lt_0_lowtown_n_-65536_3z.dds"},
	{0xCC7091946E0B1523ULL, "vxca_caverns2prc_9fw_0_0.vlm"},
	{0xCC70DC245146F156ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv2.ani"},
	{0xCC70DEB13AED1E7BULL, "vxw_bonepitdestroyed_a3_258_0_0.vlm"},
	{0xCC7107B86C129587ULL, "795255341.ogg"},
	{0xCC7185D7282FC1EAULL, "81026287.ogg"},
	{0xCC723EF2080D51D5ULL, "hlac_w_encounter3_-196610_0x.dds"},
	{0xCC7292350C432312ULL, "zz_followers_banter6157815.cl"},
	{0xCC72B8FE01950DA6ULL, "tal_sha_03.mao"},
	{0xCC730733EE2CAC86ULL, "535672211.ogg"},
	{0xCC730F740A73E112ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\posev1.ani"},
	{0xCC737A2C47029C82ULL, "6034038_m.fxe"},
	{0xCC73CA0D3F6640D0ULL, "vxca_caverns1_mines_ijk_0_0.vlm"},
	{0xCC73CEEF08441D9FULL, "534593079.ogg"},
	{0xCC7469D56C4F2847ULL, "639731105.ogg"},
	{0xCC74F901AB40D9BEULL, "lt_warehouse_2_0_ll.mmh"},
	{0xCC7549D39540C2A5ULL, "312776678.ogg"},
	{0xCC7552BA4B80F30DULL, "qa_footstep.arl"},
	{0xCC7564734DAFA23AULL, "6175845_m.fxe"},
	{0xCC75A97F8D31DD17ULL, "vxht_brothel_n_4aa_0_0.vlm"},
	{0xCC75AFF747E89205ULL, "59345775.ogg"},
	{0xCC75B9AE2B7F53EAULL, "vxca_cave_2cp_0_0.vlm"},
	{0xCC75BD98D6B61236ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_wobble.gad"},
	{0xCC766D04F56A2821ULL, "art\\levels\\ztools\\default\\generic\\4sd_basic.phy"},
	{0xCC772AC0B2E949A7ULL, "vxlt_0_lowtown_docks_cuk_0_0.vlm"},
	{0xCC772EAA25E9555EULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_chab_d.dds"},
	{0xCC77313BA28BC238ULL, "llac_w_bonepit_262145_0.dds"},
	{0xCC776B4921E4FC3AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_daggers.gad"},
	{0xCC7780D234E89D07ULL, "1013365503.ogg"},
	{0xCC77AD4796BCED4BULL, "ht_0_hightown_n_51_0.rml"},
	{0xCC77D073D2A987CDULL, "hlac_lt_0_lowtown_docks_a3_0_0z.dds"},
	{0xCC786519165088FBULL, "vxlt_hangedman_1rc_0_0.vlm"},
	{0xCC78CDD4C1E2CE52ULL, "6121053_m.fxe"},
	{0xCC7910294696BA90ULL, "vxw_bonepit_165_0_0.vlm"},
	{0xCC7920C1745142A7ULL, "hlac_dr_dungeon_3214_10z.dds"},
	{0xCC799E985208C9B2ULL, "6055418_f.fxe"},
	{0xCC7A2C0864140251ULL, "vxw_encounter1_jk_0_1.vlm"},
	{0xCC7A7BC862A7D6EAULL, "6080491_m.fxe"},
	{0xCC7AA5440CC33DC3ULL, "6107012_m.fxe"},
	{0xCC7AB0ACA1D904C4ULL, "gen00fl_varric6102795.cl"},
	{0xCC7AFEB15767EAFDULL, "176482625.ogg"},
	{0xCC7B3315F4B48FB1ULL, "vxw_woundedcoast_17d_0_0.vlm"},
	{0xCC7BD898D6BAAB95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_wobble.evt"},
	{0xCC7C149A07633369ULL, "vxca_cavern1_haunted_jfu_0_0.vlm"},
	{0xCC7C1A38C145875AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_loop.gad"},
	{0xCC7C9480C1E49BC6ULL, "dae321st_dog_closet.stg"},
	{0xCC7D1B37A8BD5ADDULL, "6170430_m.fxe"},
	{0xCC7D6340F3B0F5B4ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_flare04.mao"},
	{0xCC7DA37E58494501ULL, "mag110_arianni6206018.cl"},
	{0xCC7DE42451518947ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv2.evt"},
	{0xCC7E6E4921EB2051ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_daggers.evt"},
	{0xCC7E8907ED89F2C3ULL, "mag210_arianni6031714.cl"},
	{0xCC7ED421D6911843ULL, "976688902.ogg"},
	{0xCC7ED7224258AE4BULL, "ht_0_hightown_n_20_0.anb"},
	{0xCC7EF2A193BB3096ULL, "6179863_m.fxe"},
	{0xCC7EF4F9DD0CD89BULL, "6031168_m.fxe"},
	{0xCC7F04B33671B961ULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_act_c.anb"},
	{0xCC80637F5CCFE02AULL, "346047679.ogg"},
	{0xCC80A080A1A73EBCULL, "lowtown_fem_lowlife_mob6169186.cl"},
	{0xCC80E963D779B1AEULL, "vxlt_gallowstemplar_n_685_0_0.vlm"},
	{0xCC812C3B120AAEDCULL, "mag131_keran6170941.cl"},
	{0xCC813D24B76F477EULL, "6201285_m.fxe"},
	{0xCC81B41A45FE0059ULL, "605810634.ogg"},
	{0xCC81D962B422655CULL, "6198690_f.fxe"},
	{0xCC81DA8970853695ULL, "follower_banter6169888.cl"},
	{0xCC8201ACBE0E7AB0ULL, "hlac_lt_0_lowtown_docks_-131075_5y.dds"},
	{0xCC8312D95AB91622ULL, "907954653.ogg"},
	{0xCC8331C1795ADDB5ULL, "6002007_m.fxe"},
	{0xCC8346537D28B86BULL, "433027024.ogg"},
	{0xCC834F7D3628FDB5ULL, "245042844.ogg"},
	{0xCC83994B8B246DB7ULL, "vxht_bartrand_n_3sp_0_0.vlm"},
	{0xCC83BDCE53B4D0E3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_basic_1.mao"},
	{0xCC83EBE50AD20EAAULL, "6018803_m.fxe"},
	{0xCC844D3943CF2427ULL, "vxht_0_hightown_n_asy_0_0.vlm"},
	{0xCC8464DA63EACD90ULL, "vxw_woundedcoast_78_0_0.vlm"},
	{0xCC8487D0A99A6DC6ULL, "885343292.ogg"},
	{0xCC84EB0568197125ULL, "6034837_m.fxe"},
	{0xCC8505245157DF90ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthattackv2.gad"},
	{0xCC8507A5F401F21FULL, "w_sundermount_87_0.rml"},
	{0xCC851E043567E209ULL, "vxlt_0_lowtown_a3_n_c_gon_0_0.vlm"},
	{0xCC853C52C30C3D1DULL, "dae000_hired_prostitute.wbd"},
	{0xCC857F113494EAFDULL, "vxht_mansion2_31h_0_0.vlm"},
	{0xCC859395E911D74FULL, "cyc_air_big.dds"},
	{0xCC86C47C4FF35419ULL, "mcr370cr_slave_statue.utc"},
	{0xCC86D6F60846A17BULL, "vxlt_hangedman_1yb_0_0.vlm"},
	{0xCC86E0A6E3A1041CULL, "m_one300ar_crow_camp.bnk"},
	{0xCC8776E5C345CCFAULL, "vxlt_0_lowtown_a3_n_c_hsf_0_0.vlm"},
	{0xCC877FA7C5F1B289ULL, "6067077_m.fxe"},
	{0xCC878AC542412184ULL, "vxlt_undercityencounter_1ha_0_0.vlm"},
	{0xCC881FF03D16FCC9ULL, "6086731_m.fxe"},
	{0xCC882763537E0720ULL, "zz_followers_banter6158997.cl"},
	{0xCC88365C84B8B46BULL, "gen00fl_anders6197682.cl"},
	{0xCC88E3455D232E43ULL, "vxw_woundedcoast_n8_0_0.vlm"},
	{0xCC8A3115A37FEDF7ULL, "6062825_m.fxe"},
	{0xCC8A7F69A614AA62ULL, "864750242.ogg"},
	{0xCC8B2C7E39C6B03BULL, "hlac_lt_0_lowtown_docks_a3_n_0_5z.dds"},
	{0xCC8B864921F5D320ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_daggers.ani"},
	{0xCC8BB9587A10EB26ULL, "6126572_f.fxe"},
	{0xCC8BC15588D9CA1AULL, "lightning.mao"},
	{0xCC8BE5B3367C0EDAULL, "art\\vfx\\combat\\vfx_ha_hd_mblast_act_c.mmh"},
	{0xCC8BFF2941F16CD4ULL, "vxlt_gallowscourtyard_n_62k_0_0.vlm"},
	{0xCC8C26A5F40845B7ULL, "w_sundermount_87_0.phy"},
	{0xCC8C290DD2ACDC92ULL, "lit_tep_dust.mao"},
	{0xCC8C46023FECDE57ULL, "dae011_badpoet_sta6146913.cl"},
	{0xCC8CABF04F1A93E1ULL, "489921293.ogg"},
	{0xCC8D2ED946101290ULL, "t3_wdg_gr3.tnt"},
	{0xCC8D35F8F77B0D7BULL, "vxlt_gallowsprison_n_5sf_0_0.vlm"},
	{0xCC8D5147CA9353D0ULL, "vxht_mansion2b_31y_0_0.vlm"},
	{0xCC8DF368AE7F6EC2ULL, "847298648.ogg"},
	{0xCC8E7F59441460D7ULL, "ht_bartrand_n_3371_ll.mmh"},
	{0xCC8EB052414E3C4CULL, "follower_barks6154152.cl"},
	{0xCC8ED392A192BEBDULL, "nat380_temmerin6067385.cl"},
	{0xCC8F3F9EF64890CDULL, "gen00fl_merrill6150988.cl"},
	{0xCC8FD127593B323EULL, "vxlt_0_lowtown_n_ex5_0_0.vlm"},
	{0xCC8FF43C17CA4265ULL, "284223284.ogg"},
	{0xCC900C2F3D6276F6ULL, "vxlt_0_lowtown_a2_n_c_hoo_0_0.vlm"},
	{0xCC903538C1565E40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_loop.ani"},
	{0xCC9036E58661626EULL, "w_blightlands_21_0.mmh"},
	{0xCC9042E586617605ULL, "w_blightlands_21_0.mao"},
	{0xCC9047C1E2485862ULL, "gen_im_npc_arm_hum_glv_gwdc.uti"},
	{0xCC9050E586618E5CULL, "w_blightlands_21_0.msh"},
	{0xCC9094EA0CEE8D19ULL, "6165164_m.fxe"},
	{0xCC90F5029EDF12C3ULL, "vxht_0_hightown_a2_c7h_0_0.vlm"},
	{0xCC90F91F0EE331E8ULL, "isa120_hayder6074984.cl"},
	{0xCC915315CF853241ULL, "vxw_bonepit_pa_0_0.vlm"},
	{0xCC9245486D3BAD60ULL, "525276100.ogg"},
	{0xCC92844EFCBA9307ULL, "6115473_m.fxe"},
	{0xCC92B7A47F388303ULL, "447285175.ogg"},
	{0xCC92BDCB04B782C2ULL, "6136295_m.fxe"},
	{0xCC933D90EF1E2F80ULL, "vxlt_warehouse_18j_0_0.vlm"},
	{0xCC938FE1275DFA89ULL, "vxlt_0_lowtown_a3_n_c_gbm_0_0.vlm"},
	{0xCC93F45A8052B87FULL, "qun131ip_treasure3.utp"},
	{0xCC946A6D1244D2C1ULL, "zz_followers_banter6158172.cl"},
	{0xCC94B4EB9F8981B7ULL, "vxlt_gallowstemplar_n_5ux_0_0.vlm"},
	{0xCC9548D2FA3CB9FBULL, "vxlt_hangedman_1de_0_0.vlm"},
	{0xCC95C58912677A0AULL, "6148193_m.fxe"},
	{0xCC9634296EA0AF6EULL, "hla_ht_0_hightown_n_1_0x.dds"},
	{0xCC965BC4BF5D24F4ULL, "929899668.ogg"},
	{0xCC9678CCEC12A455ULL, "463668141.ogg"},
	{0xCC976A23FD7E17F8ULL, "vxht_mansion2_21d_0_0.vlm"},
	{0xCC97CA9C13B480CAULL, "fen210_hadriana6174810.cl"},
	{0xCC97CBF33914E332ULL, "mrl150_merrill6078334.cl"},
	{0xCC97D0D1C96C9EC7ULL, "gen00fl_aveline6137669.cl"},
	{0xCC988475D80F4206ULL, "sse_ss_harlan.gda"},
	{0xCC98C6C7673170BFULL, "mi_gla_ex.mao"},
	{0xCC9939F70311D48BULL, "149338023.ogg"},
	{0xCC99AC466934E301ULL, "dae000_bartrand6064026.cl"},
	{0xCC99D3711968A0D2ULL, "756683094.ogg"},
	{0xCC99DB8737168FB9ULL, "6043710_m.fxe"},
	{0xCC99FC2EDC177EF0ULL, "mag311_nyssa_st.stg"},
	{0xCC9A12EA97D396B3ULL, "gen00fl_sebastian6177911.cl"},
	{0xCC9B7198B6767235ULL, "llac_w_encounter2_-65533_0.dds"},
	{0xCC9BE1471F858FF2ULL, "ser221_party_comments6209897.cl"},
	{0xCC9CAB029652AD25ULL, "follower_banter6141671.cl"},
	{0xCC9CD6B8750020D0ULL, "vxlt_gallowstemplar_6sy_0_0.vlm"},
	{0xCC9D1DBA1BCFC552ULL, "dae000_bartrand6027592.cl"},
	{0xCC9DBB6203DDE3D4ULL, "7702268.ogg"},
	{0xCC9E0549F38767F0ULL, "6176200_m.fxe"},
	{0xCC9E2B8BB1DE6898ULL, "llac_w_encounter2_camp_196613_0.dds"},
	{0xCC9E43044B314ABEULL, "vxlt_0_lowtown_n_eu6_0_0.vlm"},
	{0xCC9EB77C5007D037ULL, "mcr370cr_slave_statue.ncs"},
	{0xCC9F2E36896953DAULL, "6078104_m.fxe"},
	{0xCC9F57A4ABE877C9ULL, "sib111_mother6098977.cl"},
	{0xCC9FF24243F6164AULL, "6165031_m.fxe"},
	{0xCCA005D4EFE8C4FDULL, "vxlt_warehouse_qy_0_0.vlm"},
	{0xCCA1091E90340850ULL, "6173126_m.fxe"},
	{0xCCA1C5D5A8CACC47ULL, "6080146_m.fxe"},
	{0xCCA201BE19541309ULL, "vxlt_0_lowtown_a3_n_c_gyl_0_0.vlm"},
	{0xCCA2A5E75691B621ULL, "gen00fl_aveline6132794.cl"},
	{0xCCA2E6F5833F2BE3ULL, "plt_qun200pt_missing_qunari.plo"},
	{0xCCA38B320090C87DULL, "145290150.ogg"},
	{0xCCA3A1F9AA60EB01ULL, "dae110cr_refugee_f_01.utc"},
	{0xCCA46361A9DDE389ULL, "6069277_m.fxe"},
	{0xCCA47920107E9E6DULL, "696049070.ogg"},
	{0xCCA5237B24D405C4ULL, "vxht_brothel_n_6et_0_0.vlm"},
	{0xCCA52F18D3AB9258ULL, "6116444_m.fxe"},
	{0xCCA5AB56A6ECE702ULL, "zz_one_debug6025835.cl"},
	{0xCCA6204256547985ULL, "war_stun.dds"},
	{0xCCA6D1A4A954BBECULL, "llac_dr_dungeon_3217_7.dds"},
	{0xCCA712FCDCE222B9ULL, "6092583_m.fxe"},
	{0xCCA760F031D7DD45ULL, "6125335_m.fxe"},
	{0xCCA83BE323D0A244ULL, "vxw_blightlands_7b_0_0.vlm"},
	{0xCCA8484FF7BC3508ULL, "vxw_bonepit_1ra_0_0.vlm"},
	{0xCCA8E6B2941785F8ULL, "pro000_cut2d_voyage6004217.cl"},
	{0xCCA919909B82DC6BULL, "6094536_m.fxe"},
	{0xCCA97D7B4007FF91ULL, "gen_im_trs_qun_sta_fir_r0.uti"},
	{0xCCA9D6473D6508E7ULL, "929240410.ogg"},
	{0xCCAA3908EB0BDC9FULL, "hlac_ca_caverns1_mines_24426_14x.dds"},
	{0xCCAA9A5588F438CCULL, "lightning.dds"},
	{0xCCAAD422427DE397ULL, "ht_0_hightown_n_20_0.rml"},
	{0xCCAB250C289744C0ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagviscount.phy"},
	{0xCCAB4154996FD5C7ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_5z.dds"},
	{0xCCABF440F6E40408ULL, "6143724_f.fxe"},
	{0xCCAC31E925BC261BULL, "6025121_m.fxe"},
	{0xCCAC4618524D0F25ULL, "effects.gda"},
	{0xCCACB02904035415ULL, "vxlt_0_lowtown_docks_n_cuf_0_0.vlm"},
	{0xCCACB833B2FE51BAULL, "155260980.ogg"},
	{0xCCAD4EFC1021B4EBULL, "zz_dae_debug6029885.cl"},
	{0xCCAD5D1B9FB5505EULL, "atl_loginnewac_dxt5_00.dds"},
	{0xCCAD831F89182BC7ULL, "mer320_hubert6051892.cl"},
	{0xCCADA2DFA0FBA113ULL, "gen00fl_anders6125548.cl"},
	{0xCCAE56914EAA1D35ULL, "vxca_caverns2_9a1_0_0.vlm"},
	{0xCCAE77B1E1DEDD7DULL, "vxlt_warehouse_zx_0_0.vlm"},
	{0xCCAE8DB79C7E3C3BULL, "crt_sf_01.mao"},
	{0xCCAEAF261138F975ULL, "671232501.ogg"},
	{0xCCAEB9ED9CCB227DULL, "follower_banter6168036.cl"},
	{0xCCAEF3D0909370C0ULL, "6107239_m.fxe"},
	{0xCCAEFCE52B447497ULL, "30428389.ogg"},
	{0xCCAF04192B143E32ULL, "vxht_keep_n_63o_0_0.vlm"},
	{0xCCAF1B70345AA767ULL, "llac_ht_chantry_n_2534_29.dds"},
	{0xCCAF5572E63EEEB3ULL, "plt_mag320pt_otl_main.plo"},
	{0xCCAFB097B7D0B2D5ULL, "str_cyc2.mao"},
	{0xCCAFB68348F17555ULL, "dae000_mother6121032.cl"},
	{0xCCB0214AD30A6080ULL, "vxlt_warehouse_12h_0_0.vlm"},
	{0xCCB0EA6CE5750CACULL, "vxht_keep_n_60x_0_0.vlm"},
	{0xCCB0F2805894D199ULL, "hlac_lt_0_lowtown_docks_a3_-4_7x.dds"},
	{0xCCB0F6CA7BB9CF48ULL, "llac_ht_0_hightown_n_-65537_3.dds"},
	{0xCCB115A41580FDE1ULL, "zz_followers_banter6157281.cl"},
	{0xCCB11B8F0B35E9E1ULL, "vxca_cavern1_haunted_e8t_0_0.vlm"},
	{0xCCB15B97EF0E3331ULL, "dae000_templar_knight6165173.cl"},
	{0xCCB198A58375B2B7ULL, "vxlt_gallowstemplar_n_5cz_0_0.vlm"},
	{0xCCB1BCA2F7357658ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc1a.phy"},
	{0xCCB1BD20E9E52A65ULL, "lt_gallowstemplar_2388.rml"},
	{0xCCB23193B4DD14D4ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_dwfa_0d.dds"},
	{0xCCB23BEBC22703EAULL, "fx_smoketemp.dds"},
	{0xCCB26E89C7BA5C45ULL, "vxlt_0_lowtown_a3_n_c_hsm_0_0.vlm"},
	{0xCCB26FB264BCF871ULL, "6092077_m.fxe"},
	{0xCCB28BC824A3F54EULL, "gen00fl_varric6080129.cl"},
	{0xCCB3D533998DFDCCULL, "vxdr_0_deeproads_f4p_0_0.vlm"},
	{0xCCB43BD13DBC357EULL, "aip_mabari_base.gda"},
	{0xCCB47A416357596DULL, "vxlt_0_lowtown_docks_a3_czk_0_0.vlm"},
	{0xCCB47D189CA547ABULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse2cloth_orsinesdefiance.ani"},
	{0xCCB513E6DA79B29CULL, "6161356_m.fxe"},
	{0xCCB59D47260E1D10ULL, "vxlt_backalley_g_1js_0_0.vlm"},
	{0xCCB5C57508EE1352ULL, "333681248.ogg"},
	{0xCCB5CB675C264709ULL, "hlac_w_sundermount_262138_0x.dds"},
	{0xCCB5F25E10BA93FDULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose1.anb"},
	{0xCCB6301B17BF86ABULL, "686175776.ogg"},
	{0xCCB63FAC7286004AULL, "vxlt_darktown_43a_0_0.vlm"},
	{0xCCB66F8FA42D3527ULL, "combat_ws_r1_a1.gda"},
	{0xCCB685C5D74BF702ULL, "vxlt_0_lowtown_docks_a3_cy4_0_1.vlm"},
	{0xCCB6ADC015C6BAC4ULL, "vxw_blightlands_8c_0_0.vlm"},
	{0xCCB70F021C4940ACULL, "6134901_m.fxe"},
	{0xCCB788A8D68F59A8ULL, "vxlt_darktown_5j9_0_0.vlm"},
	{0xCCB7B410B13A422DULL, "w_bonepitdestroyed_a3_18_0.mmh"},
	{0xCCB7B810B13A497EULL, "w_bonepitdestroyed_a3_18_0.mao"},
	{0xCCB7C610B13A60B3ULL, "w_bonepitdestroyed_a3_18_0.msh"},
	{0xCCB82AA1CD18EE56ULL, "745500829.ogg"},
	{0xCCB84655B2C14171ULL, "vxlt_0_lowtown_ekj_0_2.vlm"},
	{0xCCB931DE35BB47E8ULL, "ht_keep_mmi.gff"},
	{0xCCB960BCBAC883C9ULL, "6095879_f.fxe"},
	{0xCCB9F0E65066F202ULL, "vxlt_warehouse_n_93_0_0.vlm"},
	{0xCCBA0F2AFB06E2CFULL, "gen00fl_aveline6116659.cl"},
	{0xCCBA60C296D0E582ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandleft.phy"},
	{0xCCBA8410F3CF6A1FULL, "vxca_cave_1oz_0_0.vlm"},
	{0xCCBAA305763E0273ULL, "vxht_mansion2b_213_0_0.vlm"},
	{0xCCBAA432C5F84EAFULL, "6041528_f.fxe"},
	{0xCCBAA444CC8A95C8ULL, "vxca_cavern1_haunted_e8q_0_0.vlm"},
	{0xCCBABE6A6424C898ULL, "vxw_bonepitmining_a2_n0_0_0.vlm"},
	{0xCCBAC95B171ED1B6ULL, "vxw_sundermount_2z6_0_1.vlm"},
	{0xCCBB53BFA651D016ULL, "407757949.ogg"},
	{0xCCBC0F82BCEA20C5ULL, "gen00fl_fenris6128011.cl"},
	{0xCCBE1701CC46FFB5ULL, "105161560.ogg"},
	{0xCCBE21FF2F3C881FULL, "vxlt_0_lowtown_h99_0_0.vlm"},
	{0xCCBE5EFE00987C5FULL, "51138348.ogg"},
	{0xCCBE66D72367B9F6ULL, "one000_ilen6036925.cl"},
	{0xCCBE8DF885472282ULL, "vxht_0_hightown_a3_ccp_0_0.vlm"},
	{0xCCBE98D07D95C827ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgtp_0s.dds"},
	{0xCCBEBF1703BED870ULL, "vxht_mansion2b_1b2_0_0.vlm"},
	{0xCCBEC7A86EFCB75AULL, "6092561_f.fxe"},
	{0xCCBEC9669F8BBC26ULL, "hlac_ca_caverns1_24427_0z.dds"},
	{0xCCBF2D28F2A115B2ULL, "vxlt_0_lowtown_bxq_0_0.vlm"},
	{0xCCBFA21B031F00D8ULL, "vxw_sundermount_243_0_0.vlm"},
	{0xCCBFB6873F8E9CA0ULL, "151973965.ogg"},
	{0xCCBFC27488F1433CULL, "mag311_nyssa6042221.cl"},
	{0xCCC044C04391B3DAULL, "6161574_m.fxe"},
	{0xCCC0DCB2F34FAB0EULL, "vfx_chunkyparticle_p3.vsh.fxo"},
	{0xCCC0E066B9B074C5ULL, "vxlt_0_lowtown_a3_n_c_hxl_0_0.vlm"},
	{0xCCC101956D40CFACULL, "plt_clf200pt_cliff_raiders.plo"},
	{0xCCC10276C13FC0ECULL, "vxca_cavern1_haunted_iz7_0_0.vlm"},
	{0xCCC10EDD4D33F09FULL, "vxlt_0_lowtown_eva_0_0.vlm"},
	{0xCCC13E598B1E2EFAULL, "vxlt_0_lowtown_a3_n_c_hob_0_0.vlm"},
	{0xCCC159258B20B1D9ULL, "hlac_ht_chantry_2534_40z.dds"},
	{0xCCC18B41FEADD5B6ULL, "6076291_f.fxe"},
	{0xCCC1B410B14296A6ULL, "w_bonepitdestroyed_a3_18_0.rml"},
	{0xCCC1EC5E96B1DA01ULL, "1073268891.ogg"},
	{0xCCC2370FC355EEE4ULL, "vxlt_0_lowtown_docks_95j_0_0.vlm"},
	{0xCCC2516EE68FBE68ULL, "zz_night.cut"},
	{0xCCC2AC2DF36AC31BULL, "hlac_dr_dungeon_3214_3x.dds"},
	{0xCCC36F72261843C4ULL, "w_sundermount_189_0.phy"},
	{0xCCC3F670479A692BULL, "515524474.ogg"},
	{0xCCC3FAA0490C1C4BULL, "gen00fl_varric6191291.cl"},
	{0xCCC41D0B3FA4EBEFULL, "vxlt_gallowstemplar_f_6su_0_0.vlm"},
	{0xCCC433B3C7AEB721ULL, "w_bonepitmining_a2_55_0.rml"},
	{0xCCC46253DC7502E5ULL, "mag110_arianni6049714.cl"},
	{0xCCC46CC5F3088E99ULL, "llac_lt_gallowstemplar_2383_5.dds"},
	{0xCCC4EC20B0503FE0ULL, "vxht_0_hightown_a3_bgg_0_0.vlm"},
	{0xCCC510EC365D308BULL, "580609322.ogg"},
	{0xCCC52C47E3D7768BULL, "545547687.ogg"},
	{0xCCC5A002BD3A5087ULL, "vxlt_0_lowtown_a2_n_c_bxm_0_0.vlm"},
	{0xCCC5C7E4F99D95C3ULL, "vxw_bonepit_1v2_0_0.vlm"},
	{0xCCC5FA85C8857228ULL, "vxlt_darktown_5o6_0_0.vlm"},
	{0xCCC61D9A92932060ULL, "793517246.ogg"},
	{0xCCC64A0E11EC5C02ULL, "gen00fl_dog6117684.cl"},
	{0xCCC68494F2E34852ULL, "238443509.ogg"},
	{0xCCC6AA87B9AB74DCULL, "vxht_chantry_n_2t5_0_1.vlm"},
	{0xCCC6B29CFAD95AEAULL, "6104297_f.fxe"},
	{0xCCC6FB9845ECD46EULL, "zz_dae_debug6029419.cl"},
	{0xCCC7B9B20C4D57CCULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_lsw_reva_0s.dds"},
	{0xCCC7D8A4D9DF2994ULL, "rdr111dg_martin6057141.cl"},
	{0xCCC7E6956D46C0E3ULL, "plt_clf200pt_cliff_raiders.nss"},
	{0xCCC83668367D8FA3ULL, "vxlt_0_lowtown_a3_n_c_efa_0_0.vlm"},
	{0xCCC83F74D72B972AULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_c.anb"},
	{0xCCC86E185640F519ULL, "vxw_blightlands_bl_0_0.vlm"},
	{0xCCC885AA52B9BCA5ULL, "hlac_lt_0_lowtown_-65537_2y.dds"},
	{0xCCC8D4173467E2A6ULL, "hlac_lt_gallowscourtyard_n_-65538_4y.dds"},
	{0xCCC8E2B1FFB03F03ULL, "lgt210_eau_symp_templar6144734.cl"},
	{0xCCC920DA618C7D1FULL, "lt_0_lowtown_docks_102_0.msh"},
	{0xCCC926DA618C87C1ULL, "lt_0_lowtown_docks_102_0.mmh"},
	{0xCCC932DA618C9C62ULL, "lt_0_lowtown_docks_102_0.mao"},
	{0xCCC964189CB77935ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse2cloth_orsinesdefiance.gad"},
	{0xCCC96C41D98C8DEAULL, "6036902_m.fxe"},
	{0xCCC9BA9312C5D554ULL, "follower_banter6141795.cl"},
	{0xCCC9D6440CE15224ULL, "479889018.ogg"},
	{0xCCCA09E8F59F8CA9ULL, "l_qun230ar_cavern.bnk"},
	{0xCCCA9DAE6121A9DBULL, "lgt220_shp_mag_guard6149497.cl"},
	{0xCCCAE53038A800C6ULL, "6110574_m.fxe"},
	{0xCCCB1A89975D9441ULL, "vxw_bonepit_le_0_0.vlm"},
	{0xCCCBFF610480D0DAULL, "6153579_m.fxe"},
	{0xCCCC60BCAFE0796BULL, "dae000_sketch6179003.cl"},
	{0xCCCC724ED03F5D60ULL, "rdr111dg_harbormaster6055081.cl"},
	{0xCCCCAE24C28DF8AEULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_chab_t.dds"},
	{0xCCCCE10186BE1D83ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.anb"},
	{0xCCCCE31C0BCB246AULL, "6096950_f.fxe"},
	{0xCCCD45B1E6987A41ULL, "685330185.ogg"},
	{0xCCCD743AECCA4FD0ULL, "6135982_m.fxe"},
	{0xCCCDA1429C4D7F88ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.anb"},
	{0xCCCE06044F65E4C8ULL, "25605152.ogg"},
	{0xCCCE13F7F5151958ULL, "vxw_sundermount_290_0_0.vlm"},
	{0xCCCE18AB6D70C5EBULL, "docks_fem_sailor_mob6176293.cl"},
	{0xCCCE59A12903F71AULL, "llac_lt_undercityencounter_-1_37.dds"},
	{0xCCCFBF531A9BB649ULL, "hlac_lt_0_lowtown_a2_65534_1x.dds"},
	{0xCCCFFCBA0C407562ULL, "vxlt_0_lowtown_a2_n_c_e25_0_0.vlm"},
	{0xCCD0420C28B69255ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagviscount.msh"},
	{0xCCD0440C28B695CBULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagviscount.mmh"},
	{0xCCD0500C28B6AA68ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_flagviscount.mao"},
	{0xCCD067A81C51A015ULL, "mrl000_marethari6205687.cl"},
	{0xCCD0ADA63E64DE66ULL, "follower_banter6155886.cl"},
	{0xCCD0BCC0931348C7ULL, "vxht_0_hightown_n_c5a_0_0.vlm"},
	{0xCCD102245A502F45ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_012a_0d.dds"},
	{0xCCD1A2699F070C4DULL, "6075669_m.fxe"},
	{0xCCD1D0089E55504CULL, "35721640.ogg"},
	{0xCCD21DF0AE99849CULL, "lgt320_shw_ors_marlein6156308.cl"},
	{0xCCD26150BB760E7AULL, "6036344_m.fxe"},
	{0xCCD2831183051ABEULL, "667146004.ogg"},
	{0xCCD292E4D56B84F0ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_diagonal_1.mao"},
	{0xCCD2D1A106C6D3A0ULL, "49204715.ogg"},
	{0xCCD2E90907DB3C23ULL, "6001887_f.fxe"},
	{0xCCD2FBA5F443FF0FULL, "w_sundermount_87_0.mao"},
	{0xCCD307A5F44413ACULL, "w_sundermount_87_0.mmh"},
	{0xCCD30DA5F4441D9EULL, "w_sundermount_87_0.msh"},
	{0xCCD356A4838780BBULL, "6039731_m.fxe"},
	{0xCCD3A7B997A72BBAULL, "hlac_lt_0_lowtown_a3_n_c_65535_2y.dds"},
	{0xCCD43172E624F8A5ULL, "vxlt_0_lowtown_a2_n_c_hl1_0_0.vlm"},
	{0xCCD46B87D54CF6E1ULL, "vxlt_0_lowtown_a3_n_c_hr6_0_0.vlm"},
	{0xCCD46C62BA7B8AA8ULL, "vxlt_darktown_5p7_0_0.vlm"},
	{0xCCD49ADEBE8A86AEULL, "mrl150_merrill6093841.cl"},
	{0xCCD4FCC2EFAAD760ULL, "vxht_keep_55k_0_0.vlm"},
	{0xCCD5029AB532816CULL, "6201248_f.fxe"},
	{0xCCD59108F0888AFAULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3_tool.phy"},
	{0xCCD5AAD35A72AA67ULL, "dae000_orsino6085474.cl"},
	{0xCCD5B6F1A5CAF3ABULL, "hlac_ca_cavern1_haunted_24426_0y.dds"},
	{0xCCD5C09502C7062AULL, "hlac_ca_caverns1_24426_8y.dds"},
	{0xCCD5C809DF5ADB1CULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright180.ani"},
	{0xCCD6019CD81A83FCULL, "6142647_m.fxe"},
	{0xCCD643BDB52839D3ULL, "vxlt_gallowscourtyard_n_516_0_0.vlm"},
	{0xCCD64538D5E10807ULL, "hlac_ca_caverns1_mines_24422_5x.dds"},
	{0xCCD6A02D02E02551ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_quna_0n.dds"},
	{0xCCD6F96D2910DE73ULL, "hlac_lt_gallowscourtyard_n_-131071_2x.dds"},
	{0xCCD716DC044F2217ULL, "vxlt_0_lowtown_a2_h9u_0_0.vlm"},
	{0xCCD76F274AF9AABEULL, "441013252.ogg"},
	{0xCCD7CAB63B41D773ULL, "llac_w_woundedcoast_196613_0.dds"},
	{0xCCD8325D0C9FF6A6ULL, "vxlt_0_lowtown_h1t_0_0.vlm"},
	{0xCCD85EB0144932ABULL, "rdr111dg_martin6054185.cl"},
	{0xCCD862B9CF2F78F9ULL, "lt_0_lowtown_docks_a3_n_21_0_lc.mmh"},
	{0xCCD87204F5BCFB5DULL, "art\\levels\\ztools\\default\\generic\\4sd_basic.mao"},
	{0xCCD87604F5BD0196ULL, "art\\levels\\ztools\\default\\generic\\4sd_basic.mmh"},
	{0xCCD88004F5BD1294ULL, "art\\levels\\ztools\\default\\generic\\4sd_basic.msh"},
	{0xCCD8AF7C1D390035ULL, "170955668.ogg"},
	{0xCCD934A4A6DF6785ULL, "vxlt_shop_e2_0_0.vlm"},
	{0xCCD9631804D9BD60ULL, "838350571.ogg"},
	{0xCCD9711C85C6F32BULL, "vxht_mansion_n_5a3_0_0.vlm"},
	{0xCCDA3557F4C49D54ULL, "gen00fl_anders6202864.cl"},
	{0xCCDA4D0FCCFAA056ULL, "zz_hawkewarrior_slashing.cut"},
	{0xCCDA4FC1707C6D9CULL, "vxht_0_hightown_a2_9zr_0_0.vlm"},
	{0xCCDA8418F6DA1A10ULL, "6148663_m.fxe"},
	{0xCCDAB20186CA0B0FULL, "art\\vfx\\combat\\vfx_gib_hlk_p.mao"},
	{0xCCDABE0186CA1FACULL, "art\\vfx\\combat\\vfx_gib_hlk_p.mmh"},
	{0xCCDAC40186CA299EULL, "art\\vfx\\combat\\vfx_gib_hlk_p.msh"},
	{0xCCDAD8DD4DCD447FULL, "lgt310_mag_melson6152209.cl"},
	{0xCCDAE6988EBA9F40ULL, "art\\vfx\\combat\\vfx_gib_ogre_1.mao"},
	{0xCCDB2A8793524372ULL, "vxlt_gallowstemplar_n_5qu_0_0.vlm"},
	{0xCCDBDEA4B057CD95ULL, "582831402.ogg"},
	{0xCCDC7125E1183361ULL, "vxca_cavern1_haunted_e7i_0_0.vlm"},
	{0xCCDCB7228977BCFDULL, "53431924.ogg"},
	{0xCCDCC3ED497769F6ULL, "hairalpha_hair_p1.vsh.fxo"},
	{0xCCDCD30F832B8E53ULL, "rdr111dg_assistant6098249.cl"},
	{0xCCDCD509DF61105EULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright180.gad"},
	{0xCCDD0B275C1806C4ULL, "716478759.ogg"},
	{0xCCDD4A25F2C44907ULL, "pro000_aveline_intro6109330.cl"},
	{0xCCDD9527D2DD33F6ULL, "vxlt_hangedman_n_21c_0_0.vlm"},
	{0xCCDDB75578340D06ULL, "llac_ht_chantry_2534_23.dds"},
	{0xCCDDFA962E2BDB9AULL, "6076909_f.fxe"},
	{0xCCDE9FC90DFB891BULL, "dae351ip_at_sundermount_a.utp"},
	{0xCCDEE3A96286045EULL, "6072087_m.fxe"},
	{0xCCDF3A0E6F520E3DULL, "one000_ilen6059259.cl"},
	{0xCCDF3C72E667E850ULL, "plt_mag320pt_otl_main.nss"},
	{0xCCDF6368C5F5253CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\_textures\\prx_lower_d.dds"},
	{0xCCDFC473767CD852ULL, "vxlt_gallowscourtyard_n_513_0_0.vlm"},
	{0xCCE007709C88B7A6ULL, "zz_prmpt_debug6200929.cl"},
	{0xCCE03BC296F17636ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandleft.mao"},
	{0xCCE047C296F18AD5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandleft.mmh"},
	{0xCCE049C296F18E4BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandleft.msh"},
	{0xCCE10F897C619267ULL, "312696321.ogg"},
	{0xCCE129F3C41AD4E8ULL, "6095495_m.fxe"},
	{0xCCE1D05C6AAD6A7CULL, "red25.dds"},
	{0xCCE217F9E2CE5078ULL, "6098752_f.fxe"},
	{0xCCE2826BB8C32A9CULL, "vxlt_backalley_n_1e3_0_0.vlm"},
	{0xCCE28978E4AAF63EULL, "hlac_lt_0_lowtown_docks_a3_-131075_3z.dds"},
	{0xCCE2E288CDA0D99BULL, "vxlt_0_lowtown_docks_a3_n_d21_0_0.vlm"},
	{0xCCE34B15772C27ADULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_lgta_0s.dds"},
	{0xCCE36E9FE1A0EFE0ULL, "vxht_keep_58h_0_0.vlm"},
	{0xCCE38032FDEA0F2BULL, "201816977.ogg"},
	{0xCCE3B009DF66F07DULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright180.evt"},
	{0xCCE3C856C4F146AEULL, "darktown_fem_lowlife_mob6174146.cl"},
	{0xCCE3E6DA117D18BEULL, "hlac_lt_undercityencounter_-1_4z.dds"},
	{0xCCE52433FF7101ABULL, "vxlt_hangedman_1w0_0_0.vlm"},
	{0xCCE584C0C1F03B10ULL, "6039457_m.fxe"},
	{0xCCE5F9DBA26CDF48ULL, "vxca_cavern1_haunted_e7l_0_0.vlm"},
	{0xCCE65AE1B64FB2EDULL, "6118663_m.fxe"},
	{0xCCE6A439FE960F26ULL, "vxlt_0_lowtown_h6u_0_0.vlm"},
	{0xCCE791031CB990C5ULL, "w_sundermount_41_0.anb"},
	{0xCCE7B66726743486ULL, "hlac_w_encounter2_camp_262142_0y.dds"},
	{0xCCE7CEC6CF73C594ULL, "vxlt_gallowstemplar_n_355_0_0.vlm"},
	{0xCCE8772FC9270073ULL, "6208225_m.fxe"},
	{0xCCE92107CE0D78D6ULL, "clf221_jalen6070387.cl"},
	{0xCCE9AC7164366347ULL, "6172773_m.fxe"},
	{0xCCEA14ADD9A121F0ULL, "vxht_mansion2b_1s1_0_0.vlm"},
	{0xCCEA1AF38184E665ULL, "485309482.ogg"},
	{0xCCEA8BF58079D0B9ULL, "6139525_m.fxe"},
	{0xCCEAC0250F76BEE3ULL, "vxlt_0_lowtown_a2_n_c_d5t_0_0.vlm"},
	{0xCCEBBA1AC5B3E33FULL, "vxca_cavern1_haunted_iz8_0_0.vlm"},
	{0xCCEC1BB5C0845A5FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgta_0t.dds"},
	{0xCCEC210868A38EB0ULL, "vxlt_undercityserialkiller_1zo_0_0.vlm"},
	{0xCCEC35FD8F92BE45ULL, "vxlt_0_lowtown_a3_n_c_hoi_0_0.vlm"},
	{0xCCEC7B9B036E0AA2ULL, "llac_w_encounter2_camp_196606_0.dds"},
	{0xCCECE8E2A4734100ULL, "zz_seb_debug.ncs"},
	{0xCCEE08EF19F6D83EULL, "vxlt_0_lowtown_docks_aoz_0_0.vlm"},
	{0xCCEE64A6F6F8EEC3ULL, "6179431_m.fxe"},
	{0xCCEE795FA14AFB6CULL, "hlac_lt_undercityserialkiller_-1_33z.dds"},
	{0xCCEEF88C09523895ULL, "vxlt_0_lowtown_a2_n_gwi_0_0.vlm"},
	{0xCCEF5AFD04F6885CULL, "vxdr_0_deeproads_evu_0_0.vlm"},
	{0xCCF02C5933FEA86FULL, "520773937.ogg"},
	{0xCCF08FFC6464550EULL, "294599309.ogg"},
	{0xCCF10A0664F3E301ULL, "ef_dera.mor"},
	{0xCCF10FE3F4A4B6BDULL, "plt_fen000pt_gift_book.plo"},
	{0xCCF12D4CB861A74DULL, "vxca_caverns2_942_0_0.vlm"},
	{0xCCF14274D74EA11DULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_c.mmh"},
	{0xCCF14F41B93927E1ULL, "vxlt_0_lowtown_a3_n_c_hd4_0_0.vlm"},
	{0xCCF1D81EF6BB1F9CULL, "121258445.ogg"},
	{0xCCF31DC3F8AE2873ULL, "dm_dougal.mor"},
	{0xCCF322280426EB5AULL, "6087029_m.fxe"},
	{0xCCF37AAE45A04E24ULL, "vxw_encounter2_j_0_0.vlm"},
	{0xCCF38BFF0C5FD923ULL, "vxlt_0_lowtown_a3_n_c_ewb_0_0.vlm"},
	{0xCCF4267390BD1393ULL, "6112924_m.fxe"},
	{0xCCF4A26947E4BC05ULL, "ch2armtnt_ds2y.psh.fxo"},
	{0xCCF4E4F0BAEA631EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_x_neg.ani"},
	{0xCCF4E87D90516FC7ULL, "ss_bartrand.bnk"},
	{0xCCF4FCBB87F47DCFULL, "230881551.ogg"},
	{0xCCF502F7383E7F3AULL, "156326436.ogg"},
	{0xCCF527DA61B1C43EULL, "lt_0_lowtown_docks_102_0.phy"},
	{0xCCF5821ECA86C8ECULL, "6118759_f.fxe"},
	{0xCCF5A0429C6ED023ULL, "art\\vfx\\combat\\vfx_w_off_penetra_loop_c.mmh"},
	{0xCCF6D6579A92953BULL, "mcr371_orsino6083336.cl"},
	{0xCCF6FAFD2D03404BULL, "hlac_lt_gallowstemplar_2383_8y.dds"},
	{0xCCF712E3F4A927EAULL, "plt_fen000pt_gift_book.nss"},
	{0xCCF71D23195D10FFULL, "lt_0_lowtown_docks_a3_n_164_0.msh"},
	{0xCCF72323195D1AA1ULL, "lt_0_lowtown_docks_a3_n_164_0.mmh"},
	{0xCCF72F23195D2F42ULL, "lt_0_lowtown_docks_a3_n_164_0.mao"},
	{0xCCF7494784FAF93FULL, "hlac_dr_0_deeproads_5839_2z.dds"},
	{0xCCF77962FF5769B9ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_lgte_0n.dds"},
	{0xCCF7F3F28E716E58ULL, "vxw_encounter2_5g_0_0.vlm"},
	{0xCCF80E41773E7472ULL, "vxlt_gallowstemplar_n_5ow_0_0.vlm"},
	{0xCCF8FF0F39A532B0ULL, "lt_0_lowtown_docks_a3_17_0_ll.mmh"},
	{0xCCF9C8768531178AULL, "vxlt_0_lowtown_a2_n_c_fh8_0_0.vlm"},
	{0xCCFA58E4E87D82A7ULL, "6128679_m.fxe"},
	{0xCCFA5AAC8625D80BULL, "art\\characters\\playercharacter\\humanmale\\textures\\arm_bdaa_0n.dds"},
	{0xCCFA5AD5E1A84EDBULL, "6149379_m.fxe"},
	{0xCCFA6C631960F2D9ULL, "hlac_lt_0_lowtown_docks_a3_-2_6y.dds"},
	{0xCCFA9A1320FFB392ULL, "plt_cod_dal_vallaslin.plo"},
	{0xCCFAD7F77A8DD3C8ULL, "vxlt_0_lowtown_a3_n_c_hd1_0_0.vlm"},
	{0xCCFAF537CCF3A87DULL, "vxw_encounter3_6h_0_0.vlm"},
	{0xCCFB382597E51DDDULL, "gen00fl_varric6093645.cl"},
	{0xCCFB59BDFDF9F7FBULL, "vxca_caverns1_j5o_0_0.vlm"},
	{0xCCFBBC476FC26357ULL, "zz_dre_debug6069194.cl"},
	{0xCCFBDA0D4801C657ULL, "vxdr_primevaltomb_9e8_0_1.vlm"},
	{0xCCFC2C5BDD55D134ULL, "vxlt_0_lowtown_a2_n_bys_0_0.vlm"},
	{0xCCFCC4E8713E685DULL, "z_dust8.dds"},
	{0xCCFCE41DAB06DCB3ULL, "270608626.ogg"},
	{0xCCFCFBF9264B8B10ULL, "fen210_hadriana6110018.cl"},
	{0xCCFD0F09620D75DEULL, "218731132.ogg"},
	{0xCCFD27C4DAA64EC2ULL, "zz_rogueslate_baddy.cut"},
	{0xCCFD407148F98B68ULL, "w_encounter2_44_0.msh"},
	{0xCCFD527148F9A9F9ULL, "w_encounter2_44_0.mao"},
	{0xCCFD5E7148F9BD92ULL, "w_encounter2_44_0.mmh"},
	{0xCCFD6A68FB485115ULL, "vxlt_0_lowtown_a2_n_glh_0_0.vlm"},
	{0xCCFD81FF3B8F12A4ULL, "sst_ss_slavers_f_1.gda"},
	{0xCCFDF5E2A481CACBULL, "zz_seb_debug.cnv"},
	{0xCCFEED7BA62F6D33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_reading.ani"},
	{0xCCFF55ACDF33A915ULL, "242625059.ogg"},
	{0xCCFF66259CAF5B9CULL, "vxlt_backalley_n_1k1_0_0.vlm"},
	{0xCCFF8F58C175F7C3ULL, "vxw_bonepit_1j6_0_0.vlm"},
	{0xCCFFE8A2D37AB30DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chantry_statues_n.dds"},
	{0xCD001B38342792F8ULL, "hlac_lt_hangedman_n_-1_18y.dds"},
	{0xCD0084590F96E316ULL, "143261159.ogg"},
	{0xCD011F93AB4F69A9ULL, "vxw_sundermount_38k_0_0.vlm"},
	{0xCD012AD04B2584CDULL, "6163399_m.fxe"},
	{0xCD014BDB875113CDULL, "771354569.ogg"},
	{0xCD01A788B36E742FULL, "95299106.ogg"},
	{0xCD01E59F21C559E1ULL, "gen00fl_anders6125977.cl"},
	{0xCD0229F527029A48ULL, "6156866_m.fxe"},
	{0xCD02B3C729A6B98EULL, "llac_ca_caverns1_mines_24424_13.dds"},
	{0xCD02D1CD0176228DULL, "419720945.ogg"},
	{0xCD03082312E7B46CULL, "vxlt_0_lowtown_a3_n_c_e8d_0_0.vlm"},
	{0xCD03EA1E5A33A80CULL, "vxdr_primevaltomb_7tk_0_0.vlm"},
	{0xCD0425B3C7E48310ULL, "w_bonepitmining_a2_55_0.msh"},
	{0xCD0433B3C7E49B2AULL, "w_bonepitmining_a2_55_0.mmh"},
	{0xCD0437B3C7E4A261ULL, "w_bonepitmining_a2_55_0.mao"},
	{0xCD04B43D3D3E32AAULL, "270290503.ogg"},
	{0xCD04D2B343D6F08CULL, "6077943_m.fxe"},
	{0xCD050AADD744B1A6ULL, "497477057.ogg"},
	{0xCD05285A1B996AEEULL, "zz_approval_debug6003559.cl"},
	{0xCD05478CA9F12489ULL, "962291039.ogg"},
	{0xCD0587CD1E99EE2AULL, "vxlt_0_lowtown_n_fgw_0_0.vlm"},
	{0xCD06801E69348CF2ULL, "vxlt_gallowstemplar_n_5bp_0_0.vlm"},
	{0xCD06EF72CA31B01CULL, "533079832.ogg"},
	{0xCD075DCD0FE69BBFULL, "vxw_bonepit_cv_0_0.vlm"},
	{0xCD07ACE52077AA19ULL, "6165335_m.fxe"},
	{0xCD07B786320EDD5FULL, "6028806_f.fxe"},
	{0xCD07FAAE032E6C75ULL, "6167091_f.fxe"},
	{0xCD082063B780CF35ULL, "vfx_chunkyparticle_p2.vsh.fxo.dx11"},
	{0xCD08B9F57AED5497ULL, "vxdr_primevaltomb_34x_0_0.vlm"},
	{0xCD08FDF0BAFB3698ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_x_neg.gad"},
	{0xCD0951AB1689EFAAULL, "447576784.ogg"},
	{0xCD095B5C697DAD7FULL, "mag110_arianni6049709.cl"},
	{0xCD096714BEE9C0FDULL, "vxw_encounter3_3g_0_0.vlm"},
	{0xCD09CC809F77CF56ULL, "698759963.ogg"},
	{0xCD0A82DCE121E4DDULL, "623472393.ogg"},
	{0xCD0AC30BC6AA7D73ULL, "vxlt_0_lowtown_docks_n_9bs_0_0.vlm"},
	{0xCD0ACB759EE34E3CULL, "vxw_blightlands_b5_0_0.vlm"},
	{0xCD0B8CC55FDBF8C7ULL, "vxht_0_hightown_a2_c7t_0_0.vlm"},
	{0xCD0B9D409D0D8971ULL, "1025052156.ogg"},
	{0xCD0BCDD5E8A34EFEULL, "vxca_cavern1_haunted_dq9_0_0.vlm"},
	{0xCD0C8D08EC280809ULL, "912381168.ogg"},
	{0xCD0C97526D1D533AULL, "llac_lt_0_lowtown_docks_a3_n_131071_1.dds"},
	{0xCD0CB03B39943391ULL, "292448495.ogg"},
	{0xCD0CBA41D24596BAULL, "lt_0_lowtown_docks_a3_n_204_0.msh"},
	{0xCD0CBF086309670BULL, "follower_banter6170633.cl"},
	{0xCD0CC841D245AEF3ULL, "lt_0_lowtown_docks_a3_n_204_0.mao"},
	{0xCD0CCC41D245B500ULL, "lt_0_lowtown_docks_a3_n_204_0.mmh"},
	{0xCD0CE57BA63B9D6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_reading.evt"},
	{0xCD0D18CBCA70F3ACULL, "ave121_brennan6105898.cl"},
	{0xCD0D61C7199A236CULL, "gen00fl_aveline6129029.cl"},
	{0xCD0D9747D6B34A04ULL, "626816997.ogg"},
	{0xCD0DA9A2F783B4EDULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc1a.msh"},
	{0xCD0DB1FC12B56757ULL, "25228925.ogg"},
	{0xCD0DBBA2F783D373ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc1a.mmh"},
	{0xCD0E12983E2548FEULL, "hlac_lt_0_lowtown_docks_a3_131070_1z.dds"},
	{0xCD0E63C71C0DB1AAULL, "6166636_m.fxe"},
	{0xCD0E96461FD5DDFAULL, "bder_proj_02.mao"},
	{0xCD0F7D221A8721B4ULL, "6081731_m.fxe"},
	{0xCD0FD388CFB7023EULL, "vxht_chantry_2t8_0_1.vlm"},
	{0xCD101FC138954A57ULL, "lt_0_lowtown_docks_159_0.msh"},
	{0xCD1031C1389568BAULL, "lt_0_lowtown_docks_159_0.mao"},
	{0xCD1035C138956FC9ULL, "lt_0_lowtown_docks_159_0.mmh"},
	{0xCD1039FA47ACD72FULL, "6042290_m.fxe"},
	{0xCD103F16A0E6DA00ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_arms_crossed.anb"},
	{0xCD105A5DF245BF25ULL, "llac_w_sundermount_-327685_0.dds"},
	{0xCD10789788DEFD5CULL, "hlac_lt_0_lowtown_a2_n_c_131071_5x.dds"},
	{0xCD107C031CDC7264ULL, "w_sundermount_41_0.msh"},
	{0xCD108E031CDC90CDULL, "w_sundermount_41_0.mao"},
	{0xCD1092031CDC9706ULL, "w_sundermount_41_0.mmh"},
	{0xCD10F721A6738DD7ULL, "cnvprev_human_male_archer2.utc"},
	{0xCD1101A113DE6B75ULL, "gen00fl_fenris6126484.cl"},
	{0xCD1121C236C2BFEAULL, "277914291.ogg"},
	{0xCD11AFB2552394B8ULL, "6102913_m.fxe"},
	{0xCD11BFB24768584AULL, "hlac_lt_0_lowtown_n_-196607_0x.dds"},
	{0xCD11C9A922512F99ULL, "vxw_sundermount_2jt_0_0.vlm"},
	{0xCD11E7E8F5DD08FBULL, "l_qun230ar_cavern.wbd"},
	{0xCD11F379E877D8C5ULL, "295307478.ogg"},
	{0xCD125BFB4C29C08CULL, "vxdr_primevaltomb_7qh_0_0.vlm"},
	{0xCD128B5E855A8A08ULL, "459383969.ogg"},
	{0xCD129E6D29F6951CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_basic.mao"},
	{0xCD12AA6D29F6A9BFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_basic.mmh"},
	{0xCD12B06D29F6B461ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_basic.msh"},
	{0xCD13448E2F3F8208ULL, "642779009.ogg"},
	{0xCD139328DA8D197FULL, "llac_lt_0_lowtown_a2_196606_0.dds"},
	{0xCD13A133691492E0ULL, "396486040.ogg"},
	{0xCD13C47BA64183DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\anders_reading.gad"},
	{0xCD1430996E7450CAULL, "697801249.ogg"},
	{0xCD1437A1BAB633B9ULL, "6114479_f.fxe"},
	{0xCD1493DA29767FCBULL, "6148111_f.fxe"},
	{0xCD14A0A69B6E750BULL, "hlac_lt_0_lowtown_-3_0z.dds"},
	{0xCD14D519570BD37DULL, "6072783_m.fxe"},
	{0xCD14D79CEB932B51ULL, "plt_dae010pt_amb_tavern.nss"},
	{0xCD14F1A62F0D11EBULL, "vxlt_0_lowtown_docks_a3_n_d0k_0_0.vlm"},
	{0xCD150B8BC83C6509ULL, "mcr310_argument6182488.cl"},
	{0xCD152B089CA1D3D8ULL, "vxw_blightlands_gn_0_0.vlm"},
	{0xCD154A6F78419ADDULL, "1023721979.ogg"},
	{0xCD15A42F04D34D44ULL, "hlac_lt_undercitydungeon_-1_34z.dds"},
	{0xCD15CD045729A956ULL, "w_bonepitmining_a2_103_0.msh"},
	{0xCD15CFAA01DCB43FULL, "vxw_bonepit_nq_0_0.vlm"},
	{0xCD15D7045729BA54ULL, "w_bonepitmining_a2_103_0.mmh"},
	{0xCD15DB045729C167ULL, "w_bonepitmining_a2_103_0.mao"},
	{0xCD15ED5F3645BBF9ULL, "632251501.ogg"},
	{0xCD15F3C514C9F18BULL, "vxlt_undercityserialkiller_15q_0_0.vlm"},
	{0xCD161C56EC06DA5FULL, "mcr370_meredith6085889.cl"},
	{0xCD16A4A0588D6BB8ULL, "gen00fl_sebastian6180813.cl"},
	{0xCD16C4F30A7D8212ULL, "hlac_lt_0_lowtown_n_-2_3z.dds"},
	{0xCD173A3C678B9419ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbroke.msh"},
	{0xCD17443C678BA507ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbroke.mmh"},
	{0xCD17483C678BAC54ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbroke.mao"},
	{0xCD175D3CDE224424ULL, "zz_mrl_debug6185301.cl"},
	{0xCD17FFD2FA1B87B6ULL, "character_character_p49.psh.fxo"},
	{0xCD1842C9AA62A608ULL, "6059994_m.fxe"},
	{0xCD1861FF352B96A4ULL, "vxlt_hangedman_n_1rg_0_0.vlm"},
	{0xCD18FDE9B9AF016CULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm04.msh"},
	{0xCD1903E9B9AF0B5EULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm04.mmh"},
	{0xCD1908AF0D8062ABULL, "plt_cod_prc_screaming_box.plo"},
	{0xCD193F469072AC6CULL, "110359492.ogg"},
	{0xCD198CF25238FE10ULL, "vxw_bonepitmining_a2_1v7_0_0.vlm"},
	{0xCD1A4025E9062983ULL, "hlac_ht_mansion2_818_41z.dds"},
	{0xCD1A48AD283AF1AAULL, "dae000_mother6136921.cl"},
	{0xCD1A89031CE4DCD1ULL, "w_sundermount_41_0.phy"},
	{0xCD1AF5B12D2E0F02ULL, "vxdr_0_deeproads_d2z_0_0.vlm"},
	{0xCD1B00E670C3F114ULL, "vxca_caverns1_mines_ckn_0_0.vlm"},
	{0xCD1B0F65CAF1011EULL, "235141493.ogg"},
	{0xCD1B43109A13BF49ULL, "mag100_karras6034702.cl"},
	{0xCD1B47625189D556ULL, "161880606.ogg"},
	{0xCD1C2AA6825344BBULL, "vxlt_0_lowtown_a2_n_c_elj_0_0.vlm"},
	{0xCD1C68C47F5DF481ULL, "vxlt_warehouse_18g_0_0.vlm"},
	{0xCD1C731E7BBD3F5EULL, "vxlt_0_lowtown_e20_0_0.vlm"},
	{0xCD1C781940D84497ULL, "6116850_m.fxe"},
	{0xCD1CDEAE5E11F78AULL, "6131907_f.fxe"},
	{0xCD1D55DCA24584A3ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_005a_d.dds"},
	{0xCD1E8268D8CE2A71ULL, "art\\vfx\\combat\\material__25.mao"},
	{0xCD1EE19C914655AAULL, "537208599.ogg"},
	{0xCD1F29A1631D56B2ULL, "mrl150_merrill6093526.cl"},
	{0xCD2076D9C88172E5ULL, "vxlt_0_lowtown_a2_n_gw9_0_0.vlm"},
	{0xCD20B0B53F171801ULL, "zz_rdr_debug6206303.cl"},
	{0xCD215C58BA5CFC85ULL, "t1_mul_br3.tnt"},
	{0xCD215E83DBA06FD4ULL, "6136169_m.fxe"},
	{0xCD2167A4127C07C3ULL, "vxht_0_hightown_a2_car_0_0.vlm"},
	{0xCD2192031CEB0ACDULL, "w_sundermount_41_0.rml"},
	{0xCD2224F7230CDC72ULL, "hlac_lt_backalley_n_65536_3y.dds"},
	{0xCD2236AB4BBF4DC3ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_5z.dds"},
	{0xCD22597481E21F5AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_rust_sa.dds"},
	{0xCD239CE58E97EC58ULL, "vxw_blightlands_ho_0_0.vlm"},
	{0xCD23A8721625E804ULL, "6201496_m.fxe"},
	{0xCD243FAE3A2020C0ULL, "646420626.ogg"},
	{0xCD244186F3D2CCBFULL, "vxw_bonepit_ip_0_0.vlm"},
	{0xCD2458431E10D6A9ULL, "408298309.ogg"},
	{0xCD247D18DDBE43ECULL, "gen00fl_anders6111234.cl"},
	{0xCD248C7AAB9C6F34ULL, "hlac_ca_caverns1_24424_22y.dds"},
	{0xCD24C0960D45FBD6ULL, "591222173.ogg"},
	{0xCD250778E262AD24ULL, "ser270_templar6137310.cl"},
	{0xCD2554B4FAB90926ULL, "vxht_chantry_n_2eq_0_0.vlm"},
	{0xCD257AC1F6B13E94ULL, "502472266.ogg"},
	{0xCD262D7260AD6101ULL, "qun221_petrice6075986.cl"},
	{0xCD26590C1F9A55E3ULL, "hla_ht_0_hightown_a3_-131073_0x.dds"},
	{0xCD2671AFEC11A036ULL, "character_character_p38.psh.fxo"},
	{0xCD26B4A69C58BE88ULL, "6056894_m.fxe"},
	{0xCD27387A996313A2ULL, "6028822_m.fxe"},
	{0xCD27DCC5711B0CF8ULL, "vxht_mansion2_2oz_0_0.vlm"},
	{0xCD27FECF442F1690ULL, "vxw_bonepitmining_a2_1s4_0_0.vlm"},
	{0xCD2842C61EFC6DC3ULL, "vxca_cavern1_haunted_cmk_0_0.vlm"},
	{0xCD28707F43C829C7ULL, "vxht_0_hightown_a2_c9v_0_0.vlm"},
	{0xCD28EEBB9C4C06FDULL, "vxht_mansion_n_630_0_0.vlm"},
	{0xCD29055DBA7C9FD1ULL, "fen210_tevinter_ambusher26108651.cl"},
	{0xCD296C4BBFB18C8CULL, "vxca_cave_236_0_0.vlm"},
	{0xCD29829E6F6AC2ACULL, "180843563.ogg"},
	{0xCD29C7A3A3983E72ULL, "717702020.ogg"},
	{0xCD2A2FA77E92522CULL, "gen00fl_anders6125255.cl"},
	{0xCD2AF28AEA2F2AC1ULL, "6187942_m.fxe"},
	{0xCD2BAD6C157FE339ULL, "vxht_chantry_206_0_0.vlm"},
	{0xCD2BBB8D9879E64DULL, "zz_undead.cut"},
	{0xCD2BF2523D92453CULL, "art\\characters\\creatures\\spider\\textures\\c_corspidra_0n.dds"},
	{0xCD2C80243549F00FULL, "hlac_w_encounter2_262147_0y.dds"},
	{0xCD2CD873A2EA3E22ULL, "675882231.ogg"},
	{0xCD2D0727474B7571ULL, "6126385_m.fxe"},
	{0xCD2D5A1E92FAB56BULL, "vxca_caverns1_mines_j22_0_1.vlm"},
	{0xCD2D81CD796AEC96ULL, "6152532_m.fxe"},
	{0xCD2DBFC7175BD6BFULL, "vxlt_0_lowtown_a3_n_c_e8k_0_0.vlm"},
	{0xCD2E99553471A21EULL, "dae270ip_armor_store.utp"},
	{0xCD2EAD63063D6099ULL, "vxw_sundermount_2pz_0_0.vlm"},
	{0xCD2EE8B6BA778B65ULL, "vxlt_0_lowtown_a2_n_gl8_0_0.vlm"},
	{0xCD2EFA47C4B8E3C9ULL, "6060743_f.fxe"},
	{0xCD2F43C7C40B6B1FULL, "6164059_m.fxe"},
	{0xCD2F4826DA9EA44CULL, "vxht_chantry_2c2_0_0.vlm"},
	{0xCD2F74CF02849550ULL, "lgt110_orz_harrowmont6172610.cl"},
	{0xCD2F775E3CB06D26ULL, "694706291.ogg"},
	{0xCD2F7FDF5C77C795ULL, "llac_ht_keep_3740_0.dds"},
	{0xCD2F8A51052DC959ULL, "llac_lt_warehouse_n_65536_0.dds"},
	{0xCD3034E5711B71AEULL, "hlac_lt_backalley_g_65536_5x.dds"},
	{0xCD3038E505EBE5BBULL, "6118203_m.fxe"},
	{0xCD308A08F0D52A23ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3_tool.msh"},
	{0xCD309808F0D5427DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3_tool.mmh"},
	{0xCD309C08F0D5488EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall16x3_tool.mao"},
	{0xCD314A855A0DECABULL, "6151468_m.fxe"},
	{0xCD31D8A7EE116AB1ULL, "vxlt_foundry_n_y6_0_0.vlm"},
	{0xCD3224CBC01A6421ULL, "lt_0_lowtown_a2_n_c_37_0.rml"},
	{0xCD3261023FE8012BULL, "dae000_gamlen6102418.cl"},
	{0xCD32619E928A5F43ULL, "dae200_saemus6142711.cl"},
	{0xCD326283E2507CB2ULL, "6164447_m.fxe"},
	{0xCD3383CE182DC690ULL, "vxw_bonepitmining_a2_29m_0_0.vlm"},
	{0xCD33B69EFD1D1395ULL, "6033564_m.fxe"},
	{0xCD33EFAA679DD261ULL, "6101392_f.fxe"},
	{0xCD34048B6547B864ULL, "mer120_hubert.cnv"},
	{0xCD3432B68B9E6E72ULL, "174432894.ogg"},
	{0xCD3446C00DF1C75BULL, "959432526.ogg"},
	{0xCD3446E4E2210B5FULL, "dae200cr_viscount6073936.cl"},
	{0xCD347A4B27C188D6ULL, "6162283_m.fxe"},
	{0xCD3545B91917C7A4ULL, "vxlt_hangedman_n_1de_0_0.vlm"},
	{0xCD3561297C63C3B1ULL, "genip_trap_bolt.utp"},
	{0xCD357E879F3389FDULL, "vxht_keep_n_6yb_0_0.vlm"},
	{0xCD358C04274BA086ULL, "483144284.ogg"},
	{0xCD35CCD143DF3F3FULL, "894238391.ogg"},
	{0xCD35E76C5432A915ULL, "76721729.ogg"},
	{0xCD363595BFF6F69AULL, "follower_banter6163568.cl"},
	{0xCD369CD0567CFAFBULL, "gen00fl_varric6099159.cl"},
	{0xCD370890C2B9835FULL, "mrl350cr_wraith.utc"},
	{0xCD37268B654A27C4ULL, "mer120_hubert.bnk"},
	{0xCD373A2D1292376BULL, "art\\characters\\playercharacter\\qunarimale\\hlm_mag.mao"},
	{0xCD37C0DA9524D616ULL, "6088402_m.fxe"},
	{0xCD37D748FB94F605ULL, "842096004.ogg"},
	{0xCD37DE28B1A7A50CULL, "vxca_cave_247_0_0.vlm"},
	{0xCD37FDD1CDB7EC93ULL, "gen_im_trs_prf_rocks_02.uti"},
	{0xCD3810447640C79DULL, "767664519.ogg"},
	{0xCD385479E6C375E1ULL, "vxca_caverns2prc_a0s_0_0.vlm"},
	{0xCD389CFDB84CECC3ULL, "hlac_dr_primevaltomb_664_3x.dds"},
	{0xCD38AE84976D9744ULL, "vxw_blightlands_eh_0_0.vlm"},
	{0xCD38E3CD46186A6BULL, "vrc321ip_at_hightown.utp"},
	{0xCD38F57A98E3C8B6ULL, "hlac_lt_0_lowtown_n_131070_6x.dds"},
	{0xCD39125D2C703F7AULL, "vxw_sundermount_205_0_0.vlm"},
	{0xCD394C7E634A2581ULL, "vxlt_warehouse_12e_0_0.vlm"},
	{0xCD39BAF394296A27ULL, "vxht_keep_n_5ic_0_0.vlm"},
	{0xCD39D06293A7F9B5ULL, "vxdr_primevaltomb_746_0_0.vlm"},
	{0xCD39D9CEF578860CULL, "6155625_m.fxe"},
	{0xCD39DE210CCC4671ULL, "874932639.ogg"},
	{0xCD3A765A309E500AULL, "vxlt_backalley_g_1kn_0_0.vlm"},
	{0xCD3A8D0B80127981ULL, "dae000_orana6149128.cl"},
	{0xCD3B3DA310144CA4ULL, "hlac_ca_cave_2193_1y.dds"},
	{0xCD3B6483E232347BULL, "vxht_0_hightown_n_a80_0_0.vlm"},
	{0xCD3BD05F67FC1DE4ULL, "vxht_brothel_n_4xe_0_0.vlm"},
	{0xCD3BDD091596F2A2ULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_imp_p.mmh"},
	{0xCD3C0C4B7EE8C86BULL, "llac_w_encounter2_-131067_0.dds"},
	{0xCD3C3CF57894A600ULL, "6072764_m.fxe"},
	{0xCD3CAA41A903C0AFULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2_loop.gad"},
	{0xCD3D11471BF6B5AFULL, "278707246.ogg"},
	{0xCD3D9000618F3F94ULL, "hlac_lt_0_lowtown_a2_n_c_0_2x.dds"},
	{0xCD3DADEB2DA77801ULL, "378007059.ogg"},
	{0xCD3DC010C2B4FC31ULL, "dae011_corff_sta6163841.cl"},
	{0xCD3E88252CB7FE79ULL, "gen00fl_isabela6088836.cl"},
	{0xCD3EAAC1F7E1FE3BULL, "6117303_m.fxe"},
	{0xCD3EB2EAB1DE88A8ULL, "gen_im_trs_gen_sta_spi_r0.uti"},
	{0xCD3F0C45064F7831ULL, "77100385.ogg"},
	{0xCD3F2F78EEA381F9ULL, "6075377_m.fxe"},
	{0xCD404A84E0078331ULL, "vxlt_foundry_n_t5_0_0.vlm"},
	{0xCD40809F72841D58ULL, "vxw_blightlands_vm_0_0.vlm"},
	{0xCD4097718B011456ULL, "6108540_m.fxe"},
	{0xCD412FB0516FAA7CULL, "character_character_p142.psh.fxo.dx11.assembly"},
	{0xCD41916D80CABDF7ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.cess.gad"},
	{0xCD41C2D08DA26E76ULL, "lt_0_lowtown_docks_n_25_0_lc.mmh"},
	{0xCD41DD2FA81821C8ULL, "vxca_caverns2prc_a0v_0_0.vlm"},
	{0xCD42049984EF9E4AULL, "vxlt_warehouse_n_qc_0_0.vlm"},
	{0xCD4205D0225AC788ULL, "17992525.ogg"},
	{0xCD4228AC22B497E2ULL, "nat300ip_explosive_barrel.utp"},
	{0xCD42341CDE4EBBACULL, "6106395_m.fxe"},
	{0xCD423CB4D0937050ULL, "vxlt_0_lowtown_a2_n_eqp_0_0.vlm"},
	{0xCD42A06B6495DE7AULL, "art\\vfx\\combat\\vfx_str_g_pound_r_leg_c.anb"},
	{0xCD4311CC047D5B2EULL, "465638013.ogg"},
	{0xCD4321372C6901BAULL, "6113993_m.fxe"},
	{0xCD432D3C67B0B93CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbroke.phy"},
	{0xCD4354D9B1F1FC25ULL, "303242791.ogg"},
	{0xCD436A2956150622ULL, "761539948.ogg"},
	{0xCD438D45716D6819ULL, "6175903_m.fxe"},
	{0xCD43BF41A90A035CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2_loop.evt"},
	{0xCD4412724A04D444ULL, "gui_guipixelshader_p2.psh.fxo.dx11.assembly"},
	{0xCD4431BF8B9A6992ULL, "genip_cave_blocker.utp"},
	{0xCD44691546697B22ULL, "follower_banter6170591.cl"},
	{0xCD44E33605223428ULL, "754132240.ogg"},
	{0xCD4632B7871AEE96ULL, "6083502_m.fxe"},
	{0xCD4673FA7C4A6330ULL, "hlac_lt_0_lowtown_docks_n_-4_2x.dds"},
	{0xCD4696162BCCC996ULL, "6144081_m.fxe"},
	{0xCD46A4086D979133ULL, "6146953_f.fxe"},
	{0xCD46CE10048A872CULL, "129512161.ogg"},
	{0xCD46F3A6E6EF4C78ULL, "hlac_lt_darktown_-1_9y.dds"},
	{0xCD4720618963AFC4ULL, "vxw_blightlands_ni_0_0.vlm"},
	{0xCD4792683F3CE012ULL, "588174384.ogg"},
	{0xCD47E50F484F7E5BULL, "zz_fen_debug6134344.cl"},
	{0xCD480A20D8024FEFULL, "282869460.ogg"},
	{0xCD4856140B2099AFULL, "6201288_m.fxe"},
	{0xCD48A66D80D0FF84ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.cess.evt"},
	{0xCD48E0AE87108EB2ULL, "22377524.ogg"},
	{0xCD48E8372294688AULL, "vxlt_backalley_g_1po_0_0.vlm"},
	{0xCD49230B0A8E0689ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_action.ani"},
	{0xCD4940EBABD04BB2ULL, "vxca_caverns1_cke_0_0.vlm"},
	{0xCD49A954BBBA612DULL, "vxdr_primevaltomb_7o3_0_0.vlm"},
	{0xCD49E145FA95F107ULL, "6034295_m.fxe"},
	{0xCD49E1CEF7A8B896ULL, "vxca_cavern1_haunted_eb1_0_0.vlm"},
	{0xCD4AFEF6601B7183ULL, "gencr_darkspawn_hurlock.utc"},
	{0xCD4BB026306736E8ULL, "snow_little.mao"},
	{0xCD4CA35F69801440ULL, "vxlt_0_lowtown_a2_n_bz2_0_0.vlm"},
	{0xCD4D881AF59CDCC5ULL, "pro000_friedrich6051072.cl"},
	{0xCD4D90C879504CCDULL, "vxlt_gallowstemplar_3ek_0_0.vlm"},
	{0xCD4FBFEF1EF55526ULL, "37089913.ogg"},
	{0xCD4FE06E44B8D1B0ULL, "6103349_m.fxe"},
	{0xCD4FEED68BB020ADULL, "follower_banter6172871.cl"},
	{0xCD50256058EF1C17ULL, "6071858_m.fxe"},
	{0xCD503E0B0A94530BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_action.gad"},
	{0xCD50D7D41B2C153BULL, "210479587.ogg"},
	{0xCD50DB6C913888B4ULL, "377901733.ogg"},
	{0xCD50EB99213BB94FULL, "6163281_m.fxe"},
	{0xCD513ABE0A691F05ULL, "ts_raiders_arm_table.gda"},
	{0xCD51C741A9164E75ULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_die_v2_loop.ani"},
	{0xCD51EE9CEBC7003AULL, "plt_dae010pt_amb_tavern.plo"},
	{0xCD51FF800C4954BBULL, "262553701.ogg"},
	{0xCD520695155AEDEEULL, "6070485_m.fxe"},
	{0xCD52288753763C79ULL, "dae000_fem_priest_mob6174184.cl"},
	{0xCD5244DAE4DF1006ULL, "static_rendertodepth_p1.psh.fxo.dx11"},
	{0xCD526AE47E3084A1ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_004a_0d.dds"},
	{0xCD52DC4C6485059DULL, "vxdr_0_deeproads_bhd_0_1.vlm"},
	{0xCD533BF1CDD04507ULL, "vxlt_gallowstemplar_f_5o6_0_0.vlm"},
	{0xCD5345861BB06610ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_booktop.msh"},
	{0xCD5353861BB07E2AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_booktop.mmh"},
	{0xCD5357861BB08561ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_booktop.mao"},
	{0xCD536EBBE16606C5ULL, "sum_pro_03.mao"},
	{0xCD53886B7B1C2327ULL, "6028438_m.fxe"},
	{0xCD53CC1E4ACFCFF5ULL, "hlac_w_encounter1_-65535_0y.dds"},
	{0xCD53E42D55015D21ULL, "940714602.ogg"},
	{0xCD5401C6935FD5A5ULL, "vxlt_warehouse_13f_0_0.vlm"},
	{0xCD54177ACE37A699ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_darren.ani"},
	{0xCD54303C3B0EA055ULL, "mrl150_merrill6071871.cl"},
	{0xCD5434EF879491CBULL, "vxlt_warehouse_n_120_0_0.vlm"},
	{0xCD543C28919D8E69ULL, "vxw_bonepitmining_a2_25l_0_0.vlm"},
	{0xCD5442773D56CF72ULL, "925006356.ogg"},
	{0xCD5445D5FB691D26ULL, "999435913.ogg"},
	{0xCD55020E95B5FE9DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_noba_0n.dds"},
	{0xCD55348A92AB8470ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupwood.mao"},
	{0xCD55388A92AB8A83ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupwood.mmh"},
	{0xCD55468A92ABA2DDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupwood.msh"},
	{0xCD55923E7B59C844ULL, "vxw_blightlands_sn_0_0.vlm"},
	{0xCD55B7F7D738B35CULL, "164925908.ogg"},
	{0xCD55BE6D80DBB37DULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_ele_c.cess.ani"},
	{0xCD56E4266EFCA46DULL, "149880218.ogg"},
	{0xCD572E54515D589CULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\4sd_basic_2.mao"},
	{0xCD5739EE32931C88ULL, "hlac_lt_0_lowtown_a2_-131073_3x.dds"},
	{0xCD573B0B0A9A6D70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\hug_friend_action.evt"},
	{0xCD575C3C2DBA49DBULL, "rog_shadow_obscured.dds"},
	{0xCD579E5CA8C0BA90ULL, "530438223.ogg"},
	{0xCD58BCBF7C44B37DULL, "vxht_0_hightown_n_b8m_0_0.vlm"},
	{0xCD58DAE9B9E4A9F9ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm04.phy"},
	{0xCD595D7149481A8DULL, "w_encounter2_44_0.phy"},
	{0xCD59688F5962779FULL, "vxca_cave_18o_0_0.vlm"},
	{0xCD59B0D6D8F50B97ULL, "w_bonepit_36_0_lc.mmh"},
	{0xCD59B79AB9ECA964ULL, "985223513.ogg"},
	{0xCD59F0B61A8EB3FDULL, "196858174.ogg"},
	{0xCD59FF46E1669E02ULL, "ttx_roof_side_a_1.mmh"},
	{0xCD5A0146E166A1B8ULL, "ttx_roof_side_a_1.msh"},
	{0xCD5A11164486F559ULL, "6137484_m.fxe"},
	{0xCD5A4FEF3174543DULL, "hlac_lt_backalley_n_0_11x.dds"},
	{0xCD5ADDA180D2EFD4ULL, "follower_banter6160956.cl"},
	{0xCD5AF1B9E6C24842ULL, "vxht_keep_n_5x4_0_1.vlm"},
	{0xCD5AF27ACE3D86DBULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_1_darren.gad"},
	{0xCD5B049428A5B950ULL, "and130tr_worldmap.utt"},
	{0xCD5B64B3DEF0F762ULL, "ave200_party_comment6116362.cl"},
	{0xCD5BC0146BB33C67ULL, "6051489_m.fxe"},
	{0xCD5BF90AAC285BEDULL, "vxdr_primevaltomb_2xo_0_0.vlm"},
	{0xCD5C31F937716B75ULL, "vxca_caverns1_mines_imk_0_0.vlm"},
	{0xCD5C5122B0B5C48DULL, "6043973_f.fxe"},
	{0xCD5CD228E1D4193DULL, "310269385.ogg"},
	{0xCD5CD343578466EEULL, "llac_lt_0_lowtown_a3_n_c_65537_1.dds"},
	{0xCD5D40ACCED54A2BULL, "llac_lt_blackemporium_1055_5.dds"},
	{0xCD5DA36B59EF3585ULL, "hlac_lt_0_lowtown_docks_n_-262144_0z.dds"},
	{0xCD5DC6DB5FF30DDBULL, "94218580.ogg"},
	{0xCD5DD212C8D383D1ULL, "vxw_bonepitdestroyed_a3_2an_0_0.vlm"},
	{0xCD5DD6A5AC410F05ULL, "390773015.ogg"},
	{0xCD5DEBFEBE152DBFULL, "909527670.ogg"},
	{0xCD5E0F473C1352E4ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintingsmlc_1.mao"},
	{0xCD5E20D264D438BBULL, "one220_dera6104098.cl"},
	{0xCD5E524B36AEEA30ULL, "6104249_m.fxe"},
	{0xCD5E86E8508EEB1FULL, "vxca_caverns2prc_62d_0_0.vlm"},
	{0xCD5EA6ED8D1955ECULL, "vxw_bonepit_6v_0_0.vlm"},
	{0xCD5EDA0FF69D0675ULL, "dae000ip_crowd_mf2_talk7.utp"},
	{0xCD5FBBB92B2DB41BULL, "art\\levels\\races\\design\\prp_gasbarrel_12.pwk"},
	{0xCD5FDAA861D518FFULL, "6048744_m.fxe"},
	{0xCD604205D1A8A3ACULL, "6163909_m.fxe"},
	{0xCD605A506E7E41E6ULL, "mrl150_merrill6150740.cl"},
	{0xCD605E71494E3B39ULL, "w_encounter2_44_0.rml"},
	{0xCD6094B775FB796CULL, "6101964_f.fxe"},
	{0xCD60A91F4338BB3BULL, "6090815_m.fxe"},
	{0xCD60F1BB03E2243EULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.mmh"},
	{0xCD61146837D88EFFULL, "79469259.ogg"},
	{0xCD613C1AE66A4288ULL, "hlac_w_woundedcoast_-65529_0y.dds"},
	{0xCD618244BFB9812FULL, "6066911_m.fxe"},
	{0xCD6263B75470074BULL, "gen00fl_merrill6130113.cl"},
	{0xCD639607CA0F264AULL, "vxlt_darktown_4wu_0_0.vlm"},
	{0xCD63A94CF2E5035EULL, "plt_cod_itm_overseer.plo"},
	{0xCD63DD2E3821B4D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_enter.ani"},
	{0xCD64041B6D4FE0C4ULL, "vxw_blightlands_to_0_0.vlm"},
	{0xCD6463F73C21BF3BULL, "vxlt_0_lowtown_a2_n_c_eua_0_0.vlm"},
	{0xCD64DC0915B9F649ULL, "art\\vfx\\combat\\vfx_m_pri_wntgrsp_imp_p.anb"},
	{0xCD64E2CBAF863E4FULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_c.mmh"},
	{0xCD64EE8AAD02B7E0ULL, "llac_w_woundedcoast_7_0.dds"},
	{0xCD6503CBC04585B1ULL, "lt_0_lowtown_a2_n_c_37_0.anb"},
	{0xCD65265B7F91A0C4ULL, "vxw_sundermount_2gr_0_0.vlm"},
	{0xCD653580056A0B9EULL, "6177748_m.fxe"},
	{0xCD6546B1DE89BCD8ULL, "vxw_encounter2_p5_0_1.vlm"},
	{0xCD65497D5F4C37F6ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.impa.evt"},
	{0xCD65568385AD2540ULL, "one280_emrys6120875.cl"},
	{0xCD65AA71260ADBAEULL, "vxlt_0_lowtown_a2_n_hcq_0_0.vlm"},
	{0xCD6611EF3BB9FE30ULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_c.mmh"},
	{0xCD6641A094338AE0ULL, "ts_apostates_wpn_props.gda"},
	{0xCD66AD9C7B138460ULL, "6092152_f.fxe"},
	{0xCD6750C7EDD26564ULL, "mrl250_pol6101501.cl"},
	{0xCD67E75E14638884ULL, "vxht_mansion2b_1ht_0_0.vlm"},
	{0xCD6878614D829295ULL, "vxlt_0_lowtown_docks_n_c4i_0_0.vlm"},
	{0xCD687FD73EA7F970ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_turnip_d.dds"},
	{0xCD688FD1DA3D9137ULL, "vxht_0_hightown_a3_c75_0_0.vlm"},
	{0xCD68D8C9A39629E2ULL, "669545616.ogg"},
	{0xCD68F3EEC5EC7CD3ULL, "414720119.ogg"},
	{0xCD68F87EE0B47550ULL, "6152160_f.fxe"},
	{0xCD68FB6DC38AE24CULL, "104427.ogg"},
	{0xCD6A25DC6DD0C4DEULL, "llac_w_encounter2_2_0.dds"},
	{0xCD6A6AE79E1E746DULL, "vxdr_primevaltomb_2wn_0_0.vlm"},
	{0xCD6A713161EC2105ULL, "vxw_bonepitdestroyed_a3_177_0_0.vlm"},
	{0xCD6AA3D6296783F5ULL, "vxca_caverns1_mines_ijj_0_0.vlm"},
	{0xCD6AA8A55C37DE08ULL, "115581735.ogg"},
	{0xCD6B248DDA3C2DD1ULL, "gen00fl_fenris6110799.cl"},
	{0xCD6B407D5F509381ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.impa.gad"},
	{0xCD6B6F2A011EEA2EULL, "424966094.ogg"},
	{0xCD6B72BD33A3B884ULL, "6004335_f.fxe"},
	{0xCD6B735CA11B2FBEULL, "6129163_m.fxe"},
	{0xCD6BA36B64B8E82DULL, "art\\vfx\\combat\\vfx_str_g_pound_r_leg_c.mmh"},
	{0xCD6C2ECEB7D5C1CAULL, "vxdr_0_deeproads_g7v_0_0.vlm"},
	{0xCD6C8CE73BCFEAEDULL, "zz_mag_debug6054604.cl"},
	{0xCD6CA00F829F4428ULL, "6064042_m.fxe"},
	{0xCD6CC90B5ADFA738ULL, "llac_w_woundedcoast_-262141_0.dds"},
	{0xCD6CE526D0DE08FDULL, "vxlt_0_lowtown_a2_eod_0_0.vlm"},
	{0xCD6CED1BF4820933ULL, "486819641.ogg"},
	{0xCD6D3F003557FD77ULL, "vis121_saemus6100094.cl"},
	{0xCD6DFA7F94BA747DULL, "573189815.ogg"},
	{0xCD6E1AB0775519EFULL, "hlac_ht_mansion2_818_17y.dds"},
	{0xCD6E7B1BABFA8A0FULL, "6116750_f.fxe"},
	{0xCD6E7BB9C9247DF0ULL, "515756463.ogg"},
	{0xCD6E859DB2866E0CULL, "mrl000_marethari6142118.cl"},
	{0xCD6EB4261E5B2382ULL, "45989298.ogg"},
	{0xCD6ED0BB03EE29FDULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.anb"},
	{0xCD6F215A584F4D91ULL, "hla_ht_0_hightown_131073_0z.dds"},
	{0xCD6F4F8C59F28996ULL, "542175778.ogg"},
	{0xCD6FE9FD5DAE10ABULL, "gen00fl_aveline6122797.cl"},
	{0xCD7076120B676730ULL, "w_woundedcoast_149_0.mao"},
	{0xCD707A120B676E43ULL, "w_woundedcoast_149_0.mmh"},
	{0xCD7088120B67859DULL, "w_woundedcoast_149_0.msh"},
	{0xCD708F31B6F0548BULL, "qcr000_arishok6059311.cl"},
	{0xCD70E52E382C4CC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_enter.evt"},
	{0xCD7126D57A1A38DEULL, "llac_lt_0_lowtown_docks_n_65532_0.dds"},
	{0xCD714F88755A5B75ULL, "163425039.ogg"},
	{0xCD71D6045778174FULL, "w_bonepitmining_a2_103_0.phy"},
	{0xCD71DCF8D1EFD4ACULL, "hlac_w_sundermount_327682_0y.dds"},
	{0xCD72417D5F56B42FULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.impa.ani"},
	{0xCD7253B26127C495ULL, "mag170_recruit026094516.cl"},
	{0xCD7265FA6F400C8EULL, "gen00fl_anders6119315.cl"},
	{0xCD729C3C216ADBC8ULL, "w_bonepitdestroyed_a3_158_0.msh"},
	{0xCD72AE3C216AFA59ULL, "w_bonepitdestroyed_a3_158_0.mao"},
	{0xCD72BA3C216B0EF2ULL, "w_bonepitdestroyed_a3_158_0.mmh"},
	{0xCD72E977303922AFULL, "vxht_mansion2_1ri_0_0.vlm"},
	{0xCD7334EF3BC4C44FULL, "art\\vfx\\combat\\vfx_m_sta_nat_close_c.anb"},
	{0xCD737D482079ABC1ULL, "vxlt_0_lowtown_a3_n_c_el9_0_0.vlm"},
	{0xCD740ADC3411CDC1ULL, "vxht_keep_n_3io_0_0.vlm"},
	{0xCD743981453F59B4ULL, "l_fen201ar_slavers_dungeon.wbd"},
	{0xCD74496AD8BA6457ULL, "mrl150_merrill6202024.cl"},
	{0xCD74B36D2A498304ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc22_basic.phy"},
	{0xCD75B140D3BAE2CFULL, "character_character_p37.psh.fxo.dx11.assembly"},
	{0xCD75E3AB0E322C2FULL, "zz_vrc_debug6200247.cl"},
	{0xCD761FC936521969ULL, "dae011_wench_mob6149862.cl"},
	{0xCD76593B0659A104ULL, "vxht_mansion2b_1cu_0_0.vlm"},
	{0xCD768A2E09C7CE89ULL, "6141870_m.fxe"},
	{0xCD76DED0F0CE7DE6ULL, "6179028_f.fxe"},
	{0xCD770C7147BED959ULL, "vxht_chantry_n_33a_0_0.vlm"},
	{0xCD7727539C522DF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_tl180_lfoot.gad"},
	{0xCD7774C70E000303ULL, "24476312.ogg"},
	{0xCD78062E3832A310ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_left_enter.gad"},
	{0xCD783D23DCFA9034ULL, "mag291_vincento6027111.cl"},
	{0xCD789E213BB39887ULL, "singleroofa_s.dds"},
	{0xCD78BD7200A5DC4CULL, "xx_extshingle.dds"},
	{0xCD78E30E53E23985ULL, "vxw_bonepitdestroyed_a3_186_0_0.vlm"},
	{0xCD790956F9E0715AULL, "art\\vfx\\materials\\chunky\\vfx_tube_blend_slow.mao"},
	{0xCD796E88186DF3BDULL, "dae000_flemeth6013258.cl"},
	{0xCD7994C796773866ULL, "6143453_m.fxe"},
	{0xCD79A51B5C774D97ULL, "hlac_w_sundermount_-327675_0y.dds"},
	{0xCD79A797B74795F8ULL, "qcr000_arishok6109125.cl"},
	{0xCD7A392C0A5FAC3BULL, "vxlt_gamlenshouse_n_1x_0_0.vlm"},
	{0xCD7A6F18B07DBB3DULL, "6016583_m.fxe"},
	{0xCD7A7157D132B23EULL, "289546268.ogg"},
	{0xCD7AADE75EA6EA9EULL, "vxdr_0_deeproads_con_0_0.vlm"},
	{0xCD7AD421947D8204ULL, "154959933.ogg"},
	{0xCD7B000E53D364FFULL, "w_bonepit_144_0.msh"},
	{0xCD7B060E53D36EA1ULL, "w_bonepit_144_0.mmh"},
	{0xCD7B120E53D38342ULL, "w_bonepit_144_0.mao"},
	{0xCD7B33DE332B1859ULL, "hlac_lt_0_lowtown_a2_n_c_65536_4z.dds"},
	{0xCD7B419B8A0FAFC0ULL, "plt_lgt110pt_harrow_escape.plo"},
	{0xCD7B68D94161907BULL, "clf260_harley6089580.cl"},
	{0xCD7C9E1A85EBA0A6ULL, "168997947.ogg"},
	{0xCD7CB9CE26FE48F6ULL, "mag121_idunna6092361.cl"},
	{0xCD7CF2EB427FB11CULL, "gen_im_trs_dal_junk1_03.uti"},
	{0xCD7D2EAB0E8D37B3ULL, "6104848_m.fxe"},
	{0xCD7D36EED5955585ULL, "6085886_m.fxe"},
	{0xCD7D38539C56B647ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_tl180_lfoot.ani"},
	{0xCD7D5A827FA0D47BULL, "vxca_caverns1_jff_0_0.vlm"},
	{0xCD7D6A7E2911B80DULL, "hlac_ht_0_hightown_n_1_1y.dds"},
	{0xCD7DF0B8EDBCA0C4ULL, "6064702_f.fxe"},
	{0xCD7E677A1EC01E63ULL, "lt_0_lowtown_docks_a3_12_0.rml"},
	{0xCD7E6D0E14A84908ULL, "vxlt_0_lowtown_docks_a3_n_czd_0_1.vlm"},
	{0xCD7E81AD3B7773D8ULL, "886297444.ogg"},
	{0xCD7E849E3A72B263ULL, "fen110_fenris6118129.cl"},
	{0xCD7E8B64C4B46EF3ULL, "hlac_lt_0_lowtown_a2_65534_1z.dds"},
	{0xCD7F9269C3CFF6B6ULL, "vxdr_0_deeproads_cy0_0_0.vlm"},
	{0xCD7FF39428C4B80DULL, "and130tr_worldmap.ncs"},
	{0xCD81001E1B9FD452ULL, "6174789_m.fxe"},
	{0xCD8154549399B951ULL, "762457626.ogg"},
	{0xCD815B54222F3B2FULL, "vxht_mansion2_1uh_0_0.vlm"},
	{0xCD817B120B75E368ULL, "w_woundedcoast_149_0.phy"},
	{0xCD819FBA1E19A62CULL, "6039949_m.fxe"},
	{0xCD820A15637DD1C4ULL, "vxw_sundermount_2mp_0_0.vlm"},
	{0xCD8255B64AFC68B1ULL, "hlac_lt_0_lowtown_docks_n_-131071_0z.dds"},
	{0xCD830345A7E9AE6CULL, "ht_chantry_2534_lc.mmh"},
	{0xCD8336329771A201ULL, "llac_lt_0_lowtown_a2_-131073_0.dds"},
	{0xCD83385CD0C781E1ULL, "655808395.ogg"},
	{0xCD83EEAE00006BF6ULL, "118701891.ogg"},
	{0xCD84183ADE1412B2ULL, "vxw_encounter2_cg_0_0.vlm"},
	{0xCD842B10B49496F5ULL, "vxht_mansion2b_31x_0_0.vlm"},
	{0xCD84CB17F84FB984ULL, "vxht_mansion2b_1fz_0_0.vlm"},
	{0xCD8532FD1ED2B224ULL, "vxca_cavern1_haunted_egp_0_0.vlm"},
	{0xCD8583A5B469E70EULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.cess.ani"},
	{0xCD858D40FDBB59B8ULL, "237339500.ogg"},
	{0xCD859C197208740DULL, "6128974_m.fxe"},
	{0xCD85B9B25D2C5FC2ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyo_0n.dds"},
	{0xCD85CF733CAE2274ULL, "6155841_m.fxe"},
	{0xCD85D7918B413CD5ULL, "gen_im_trs_crc_sta_02.uti"},
	{0xCD86072A362D5F9FULL, "6051280_f.fxe"},
	{0xCD86283DE8FA37A6ULL, "llac_lt_0_lowtown_a2_n_65534_1.dds"},
	{0xCD863DF6937A6292ULL, "follower_banter6160639.cl"},
	{0xCD86C0380CFC2B33ULL, "gen00fl_fenris6105610.cl"},
	{0xCD874F9FFCEAB4A7ULL, "vxw_bonepit_29p_0_0.vlm"},
	{0xCD87699091AA83F6ULL, "mag170_hugh6094564.cl"},
	{0xCD87E7519EB38F06ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent_override.ani"},
	{0xCD888A81EAD85C40ULL, "lgt300cr_gang_4_lt.utc"},
	{0xCD88BC2A682A0B69ULL, "519858344.ogg"},
	{0xCD88F437758604CBULL, "590288599.ogg"},
	{0xCD89D5F352DA5EB7ULL, "zz_mcr_debug6088422.cl"},
	{0xCD89EBD970722EDDULL, "327724664.ogg"},
	{0xCD89F6B70B10EF17ULL, "538355606.ogg"},
	{0xCD8AC6AE872B01CFULL, "gen00fl_anders6119705.cl"},
	{0xCD8AEB832A6659F6ULL, "vxlt_hangedman_n_2uo_0_0.vlm"},
	{0xCD8B2D875E441F29ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_gate_guardian_sa.dds"},
	{0xCD8B82A1B224F12EULL, "lt_0_lowtown_a2_water_21225.mmh"},
	{0xCD8B8EA1B22505C5ULL, "lt_0_lowtown_a2_water_21225.mao"},
	{0xCD8B9CA1B2251D1CULL, "lt_0_lowtown_a2_water_21225.msh"},
	{0xCD8C4E97811203DAULL, "166209680.ogg"},
	{0xCD8CD1300176D2F5ULL, "6149073_f.fxe"},
	{0xCD8D6D7D77F85AADULL, "1003876730.ogg"},
	{0xCD8D7C536FCB7B7DULL, "hlac_lt_gallowstemplar_f_2394_7y.dds"},
	{0xCD8DE3CBAFA9448CULL, "art\\vfx\\combat\\vfx_c_ogr_stomp_c.anb"},
	{0xCD8E0446B5C60F36ULL, "vxdr_0_deeproads_cv1_0_0.vlm"},
	{0xCD8E363930DF4954ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l90_p.ani"},
	{0xCD8E812415CFDDB4ULL, "w_encounter3_42_0.rml"},
	{0xCD8EBFE2D01359A3ULL, "6088202_f.fxe"},
	{0xCD8EEB9E9FF16FCAULL, "vxlt_darktown_4dp_0_0.vlm"},
	{0xCD8F5D79CF829200ULL, "269325896.ogg"},
	{0xCD8F916327C7F50AULL, "vxlt_undercityserialkiller_25o_0_0.vlm"},
	{0xCD8FCD31142553AFULL, "vxht_mansion2_1xk_0_0.vlm"},
	{0xCD9014A3558F5993ULL, "1007264265.ogg"},
	{0xCD901BAC22F6A2BFULL, "nat300ip_explosive_barrel.ncs"},
	{0xCD90887ED0982C41ULL, "279543391.ogg"},
	{0xCD90A1EE9FCBC577ULL, "dae000_pc_home_ambients6208478.cl"},
	{0xCD913BAAB700189DULL, "128898195.ogg"},
	{0xCD9163DBCECDFB2EULL, "dae000cr_grand_cleric6173679.cl"},
	{0xCD9175AD3E28BDF3ULL, "vxlt_gallowstemplar_4ty_0_0.vlm"},
	{0xCD91C41BA7FF1CE1ULL, "hlac_lt_0_lowtown_n_-65535_1y.dds"},
	{0xCD91E3C5EB9E74D7ULL, "hlac_lt_0_lowtown_a2_n_-3_5x.dds"},
	{0xCD927275C4B52E31ULL, "43855684.ogg"},
	{0xCD9273371AB2FF23ULL, "gen00fl_aveline6107068.cl"},
	{0xCD927BA5B474645FULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.cess.evt"},
	{0xCD9294631E694EABULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf_p_lfoot.evt"},
	{0xCD929624727F2ED2ULL, "11310789.ogg"},
	{0xCD92AA3A4C294981ULL, "230356109.ogg"},
	{0xCD92D2E3960FA65DULL, "6071183_f.fxe"},
	{0xCD9327211ECC2BBDULL, "245097341.ogg"},
	{0xCD93B0815F0163F5ULL, "vxlt_gallowstemplar_6sx_0_0.vlm"},
	{0xCD93F9F7F672F795ULL, "zz_followers_banter6159720.cl"},
	{0xCD9420A47DBA5C84ULL, "430703896.ogg"},
	{0xCD943080EB28C4F1ULL, "6126315_f.fxe"},
	{0xCD9456926DAAACB2ULL, "dae300_alistair_king6174591.cl"},
	{0xCD949CCE5CE9BE6FULL, "hlac_lt_0_lowtown_docks_a3_n_0_1z.dds"},
	{0xCD94FB370577A7FAULL, "genip_door_highin_normal.utp"},
	{0xCD95333930E562B6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_p_l90_p.gad"},
	{0xCD9580940CFB6338ULL, "vxdr_primevaltomb_77j_0_0.vlm"},
	{0xCD95CF519EBFA477ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent_override.evt"},
	{0xCD963A7AC771F977ULL, "vxlt_0_lowtown_a2_n_c_euu_0_0.vlm"},
	{0xCD96845E2FC2636CULL, "gen00fl_isabela6088840.cl"},
	{0xCD96DFAE0A977644ULL, "768011869.ogg"},
	{0xCD9753E653C1FF7FULL, "623826075.ogg"},
	{0xCD9770539CA9BFDAULL, "gen_im_trs_gen_statue_03.uti"},
	{0xCD97CA20D5F78E0DULL, "llac_w_encounter1_-65536_0.dds"},
	{0xCD987FD0AE6815DAULL, "hlac_lt_0_lowtown_docks_65534_1z.dds"},
	{0xCD9888E9AB083912ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comb_0d.dds"},
	{0xCD98BC7B1B8829F3ULL, "dae000_flemeth6001000.cl"},
	{0xCD995CA5B47A4E48ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_s_p.cess.gad"},
	{0xCD996AF88C557B90ULL, "vxw_sundermount_31d_0_1.vlm"},
	{0xCD997B6F0961E1EDULL, "vxht_playershouse_n_1vm_0_0.vlm"},
	{0xCD999C99A464A1EBULL, "vxlt_0_lowtown_a2_n_ekl_0_1.vlm"},
	{0xCD99AD631E6F986CULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf_p_lfoot.gad"},
	{0xCD99EE8720BB215FULL, "w_bonepitdestroyed_a3_28_0.rml"},
	{0xCD99F6AF0CB1CCE5ULL, "vxlt_0_lowtown_docks_n_c49_0_0.vlm"},
	{0xCD9A31CE4F0F2566ULL, "6016024_f.fxe"},
	{0xCD9A6E6ABFB99D47ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs105a_0d.dds"},
	{0xCD9AE3C7EB3BA817ULL, "mag211_marethari.bnk"},
	{0xCD9B430CF36CEB75ULL, "6142173_f.fxe"},
	{0xCD9BD0519EC41100ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_ent_override.gad"},
	{0xCD9BFAE930BA73CDULL, "6040011_f.fxe"},
	{0xCD9BFF66D378EDC7ULL, "vxht_0_hightown_a2_can_0_0.vlm"},
	{0xCD9C110AE93889DFULL, "vxlt_undercityencounter_143_0_1.vlm"},
	{0xCD9C1AF64F476979ULL, "6020425_m.fxe"},
	{0xCD9C7623A7BC27B6ULL, "vxdr_0_deeproads_cs2_0_0.vlm"},
	{0xCD9CA0F97E6B2139ULL, "982117320.ogg"},
	{0xCD9CAA888059F7F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_action.evt"},
	{0xCD9CD0565A3DC847ULL, "6105457_f.fxe"},
	{0xCD9CDCB616FE469DULL, "dae300_alistair_king6133810.cl"},
	{0xCD9D14D0672BDD64ULL, "6156740_f.fxe"},
	{0xCD9D16D093BF932AULL, "6048680_m.fxe"},
	{0xCD9E38F5DB5DABE3ULL, "433723766.ogg"},
	{0xCD9E3F0E061B6C2FULL, "vxht_mansion2_1cj_0_0.vlm"},
	{0xCD9E4000B9458ACCULL, "llac_lt_0_lowtown_a3_n_c_-65539_0.dds"},
	{0xCD9EC5C7EB3F5E2FULL, "mag211_marethari.cnv"},
	{0xCD9EC668FA37E061ULL, "6139873_m.fxe"},
	{0xCD9F51E239B0BA36ULL, "lgt311_shw_mer_mettin6156536.cl"},
	{0xCD9F5D7FCE4FE5C7ULL, "fen310_fenris6118686.cl"},
	{0xCD9F61B148591961ULL, "vxw_woundedcoast_1ch_0_0.vlm"},
	{0xCD9FE58CE13895EBULL, "mag360_thrask6043554.cl"},
	{0xCDA027814564764EULL, "l_fen201ar_slavers_dungeon.bnk"},
	{0xCDA031B53BF4674CULL, "lt_0_lowtown_a2_n_3_0.rml"},
	{0xCDA06D2B2216C0A0ULL, "6112788_f.fxe"},
	{0xCDA0AC631E75B56AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_wf_p_lfoot.ani"},
	{0xCDA0EB20F21C6047ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupa_d.dds"},
	{0xCDA0FBF4C20043B2ULL, "vxw_encounter2_ia_0_0.vlm"},
	{0xCDA102FF510661C7ULL, "llac_w_encounter3_-262139_0.dds"},
	{0xCDA1401F6D18131EULL, "897542667.ogg"},
	{0xCDA1A8C167B5F61AULL, "character_character_p58.psh.fxo.dx11.assembly"},
	{0xCDA2A188805E5381ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_action.gad"},
	{0xCDA2DA88DFF953C7ULL, "vxca_cavern1_haunted_cmw_0_0.vlm"},
	{0xCDA308F57C94DBECULL, "6140974_f.fxe"},
	{0xCDA31A67BDD646F1ULL, "6028924_m.fxe"},
	{0xCDA43D1D4C7FFB9DULL, "llac_lt_0_lowtown_docks_a3_n_65534_0.dds"},
	{0xCDA4A9804A5677ABULL, "61014810.ogg"},
	{0xCDA4E09B45760D56ULL, "191900367.ogg"},
	{0xCDA536261E1A5314ULL, "hlac_ht_0_hightown_-65536_2y.dds"},
	{0xCDA545F4682F392BULL, "vxdr_primevalentrance_22o_0_0.vlm"},
	{0xCDA569CA7304297FULL, "follower_banter6136477.cl"},
	{0xCDA5F2542199B4AAULL, "crust_webs3.dds"},
	{0xCDA69AD69C45EC08ULL, "gen00fl_fenris6112075.cl"},
	{0xCDA69D7647355CD9ULL, "66823321.ogg"},
	{0xCDA70ED41E0AFAACULL, "hlac_w_woundedcoast_0_9x.dds"},
	{0xCDA72A5F1AD3E231ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_004a_0t.dds"},
	{0xCDA74C83CD837965ULL, "51001868.ogg"},
	{0xCDA7AD3E72F9DB4DULL, "414136440.ogg"},
	{0xCDA7BE4736613468ULL, "vxlt_0_lowtown_docks_a3_d2p_0_1.vlm"},
	{0xCDA8166A5CDD837EULL, "vxlt_0_lowtown_n_gjq_0_0.vlm"},
	{0xCDA8641FBE8267EEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p_tw_v1.ani"},
	{0xCDA8EA6709ADC548ULL, "vxw_woundedcoast_1cm_0_0.vlm"},
	{0xCDA97DE1AB5136F9ULL, "82448210.ogg"},
	{0xCDA9A2888064742FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_sparring_action.ani"},
	{0xCDA9C588AF263149ULL, "follower_barks6169906.cl"},
	{0xCDAA4AEF74E00E89ULL, "6029058_f.fxe"},
	{0xCDAA6181596499FDULL, "zz_mer_debug6180854.cl"},
	{0xCDAAB5C66E58E91BULL, "vxht_mansion2b_30u_0_0.vlm"},
	{0xCDAAD81336D045FAULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\varric_crossbow_bowoverlay.anb"},
	{0xCDAAFEA9F8952EFBULL, "follower_barks6153611.cl"},
	{0xCDAAFEEB09DA73D5ULL, "art\\characters\\playercharacter\\humanfemale\\cth_betb.phy"},
	{0xCDAB1A076FACDA4BULL, "vxlt_backalley_n_1ep_0_0.vlm"},
	{0xCDABCF5883DDA0CAULL, "vxlt_darktown_4nr_0_0.vlm"},
	{0xCDABEA524E15D640ULL, "llac_lt_0_lowtown_a2_1_11.dds"},
	{0xCDABF3732D834C84ULL, "qun221_petrice6077341.cl"},
	{0xCDABF509936E3918ULL, "6134509_m.fxe"},
	{0xCDABF6187CABEF0AULL, "6028523_m.fxe"},
	{0xCDAC3BD38CBE2640ULL, "gen_im_npc_arm_hum_glv_gwda.uti"},
	{0xCDACED762CEB1EA6ULL, "164407329.ogg"},
	{0xCDACF9C18FC1519FULL, "82538425.ogg"},
	{0xCDAD09D3EF776F98ULL, "vxca_caverns2_75f_0_0.vlm"},
	{0xCDAD2BB19F741622ULL, "415056542.ogg"},
	{0xCDAD5FE21C184B5FULL, "vxlt_0_lowtown_a2_n_c_ghx_0_0.vlm"},
	{0xCDADA2AAF28C9020ULL, "hlac_ht_brothel_n_3371_2y.dds"},
	{0xCDADD38E3A4F31E1ULL, "vxw_woundedcoast_1no_0_0.vlm"},
	{0xCDAE87A8189D96A2ULL, "vxlt_0_lowtown_a3_n_c_gjb_0_0.vlm"},
	{0xCDAF056DCE0652F2ULL, "vxht_0_hightown_a3_c30_0_0.vlm"},
	{0xCDAF456EA05F62F8ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_2z.dds"},
	{0xCDAF6931CD875A02ULL, "344299485.ogg"},
	{0xCDB0A8CA7426D5BBULL, "ch1alphnotnt.psh.fxo"},
	{0xCDB0BDD9DF160B47ULL, "489382438.ogg"},
	{0xCDB0D8743A391922ULL, "llac_lt_gallowsprison_n_-1_9.dds"},
	{0xCDB1042A8035951CULL, "vxlt_darktown_6ky_0_0.vlm"},
	{0xCDB162B6F6C8808EULL, "70951439.ogg"},
	{0xCDB17F30801F3397ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_brazier_n.dds"},
	{0xCDB1B1CA6188C107ULL, "473776207.ogg"},
	{0xCDB1C8B808017CACULL, "gen00fl_aveline6115435.cl"},
	{0xCDB2525522D81B42ULL, "gui_passthrough_p0.vsh.fxo.dx11"},
	{0xCDB28D6FA261C741ULL, "325251504.ogg"},
	{0xCDB2B56C5FAE4646ULL, "vxlt_warehouse_n_13f_0_0.vlm"},
	{0xCDB2C73A45DE187CULL, "vxlt_0_lowtown_docks_a3_aow_0_0.vlm"},
	{0xCDB3AE2BD9B96740ULL, "6080840_m.fxe"},
	{0xCDB4422CD581E2C8ULL, "llac_ht_mansion2_818_57.dds"},
	{0xCDB493C6E34F8167ULL, "6134968_m.fxe"},
	{0xCDB4BC338025CBDCULL, "gen_im_wep_mag_sta_act3_05.uti"},
	{0xCDB5B8EA69098B36ULL, "vxht_chantry_2d1_0_0.vlm"},
	{0xCDB628E8B4F00856ULL, "vxlt_0_lowtown_docks_a3_bql_0_0.vlm"},
	{0xCDB720E7C318F4A6ULL, "vxlt_hangedman_n_1vp_0_1.vlm"},
	{0xCDB75C43FBA3DDC8ULL, "vxw_woundedcoast_1nj_0_0.vlm"},
	{0xCDB764D2DB06C02DULL, "vxca_cave_23s_0_0.vlm"},
	{0xCDB76B1236B32FC3ULL, "6126800_m.fxe"},
	{0xCDB8687DD8AB663CULL, "hlac_w_encounter2_camp_65537_0z.dds"},
	{0xCDB9010C0695A5D2ULL, "gen_im_acc_blt_act2_03.uti"},
	{0xCDB93B9B9918FFB4ULL, "mag100_thrask6032007.cl"},
	{0xCDB9917F99A4FB34ULL, "vxw_sundermount_32i_0_1.vlm"},
	{0xCDBA4D5AAA15AAE3ULL, "383007763.ogg"},
	{0xCDBA727EA70D4B3DULL, "vxlt_gallowstemplar_f_5mj_0_0.vlm"},
	{0xCDBB22C7EA079D2FULL, "vxht_mansion2_1it_0_0.vlm"},
	{0xCDBB8C99410E5B3CULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_c.impa.ani"},
	{0xCDBBB15F27F8C63FULL, "lt_0_lowtown_docks_a3_n_2_0_ll.mmh"},
	{0xCDBBB2902A4487ADULL, "499226100.ogg"},
	{0xCDBBB49FFE5F918BULL, "6181906_m.fxe"},
	{0xCDBBDB34480BC365ULL, "727555859.ogg"},
	{0xCDBC3D1FBE92CE28ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p_tw_v1.gad"},
	{0xCDBC456B2C454A61ULL, "vxw_woundedcoast_1in_0_0.vlm"},
	{0xCDBCCD478A6E6222ULL, "199845914.ogg"},
	{0xCDBCE90FA332DADEULL, "322205866.ogg"},
	{0xCDBD34B86B0848FBULL, "ht_0_hightown_a3_26_0_ll.mmh"},
	{0xCDBD4425CD651444ULL, "335415841.ogg"},
	{0xCDBD774ABFFC6B72ULL, "vxht_0_hightown_a3_c41_0_0.vlm"},
	{0xCDBDBB5F1FE1CE07ULL, "gen_im_cft_rec_arn_01.uti"},
	{0xCDBDE33A1B95CC46ULL, "mag121_idunna6084267.cl"},
	{0xCDBEE07FB46EB237ULL, "mag131_taraneh6094530.cl"},
	{0xCDBF0334C0A3B7C4ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_imp_c.impa.ani"},
	{0xCDBF44342998372DULL, "6090457_m.fxe"},
	{0xCDBF4A7A1EF783DFULL, "lt_0_lowtown_docks_a3_12_0.anb"},
	{0xCDBFE1638B147499ULL, "fxa_web_00.mao"},
	{0xCDC030E827C971BBULL, "genip_crate_3.utp"},
	{0xCDC0F6DEF2540ABDULL, "151043018.ogg"},
	{0xCDC1274951A45EC6ULL, "vxlt_warehouse_n_18g_0_0.vlm"},
	{0xCDC168C2E9A95D5FULL, "145682267.ogg"},
	{0xCDC1F29E2BE3BA86ULL, "vxlt_0_lowtown_n_glc_0_0.vlm"},
	{0xCDC299994114907EULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_c.impa.gad"},
	{0xCDC31AADFD69DF0DULL, "dae000_hate_enemies6167597.cl"},
	{0xCDC325E390ADE545ULL, "ico_staff_2.dds"},
	{0xCDC33EAD79E1DEDEULL, "vxca_cave_1o7_0_0.vlm"},
	{0xCDC3980AB2B1283BULL, "6033288.cut"},
	{0xCDC3F451DEE37E69ULL, "llac_ht_playershouse_n_855_0.dds"},
	{0xCDC4221F94C9BEA6ULL, "779168949.ogg"},
	{0xCDC42ED04A2D6483ULL, "vxlt_undercityencounter_191_0_0.vlm"},
	{0xCDC441C2BE62D383ULL, "320691921.ogg"},
	{0xCDC48C79F7757626ULL, "lgt210_magu_selby6142064.cl"},
	{0xCDC49AC5A6E620D6ULL, "vxlt_0_lowtown_docks_a3_bvm_0_0.vlm"},
	{0xCDC4B90CD1504E09ULL, "6170463_m.fxe"},
	{0xCDC4C9D3103B3D27ULL, "364012522.ogg"},
	{0xCDC511F691BB663FULL, "art\\vfx\\combat\\vfx_m_arc_modal_act_c.impa.gad"},
	{0xCDC549DB10F37BF9ULL, "2651858.ogg"},
	{0xCDC56A3C5423CD0DULL, "6027680_m.fxe"},
	{0xCDC5CE20ED99F648ULL, "vxw_woundedcoast_1ik_0_0.vlm"},
	{0xCDC60034C0A9D126ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_imp_c.impa.gad"},
	{0xCDC6DC9BB54EFCDEULL, "vxlt_0_lowtown_h9t_0_0.vlm"},
	{0xCDC76C16800B6C54ULL, "lgt110_gxa_coa_meeting6117500.cl"},
	{0xCDC7DA0FEBE0D0C4ULL, "lgt000cr_hate_extremist.utc"},
	{0xCDC7FDC153990B4BULL, "vxlt_backalley_n_1kr_0_0.vlm"},
	{0xCDC8105778A1D68DULL, "6167584_m.fxe"},
	{0xCDC826316631E698ULL, "vxlt_0_lowtown_a3_n_c_fe8_0_0.vlm"},
	{0xCDC86280B9E719D1ULL, "6149452_m.fxe"},
	{0xCDC8812416015677ULL, "w_encounter3_42_0.mmh"},
	{0xCDC8852416015C84ULL, "w_encounter3_42_0.mao"},
	{0xCDC8972416017BE9ULL, "w_encounter3_42_0.msh"},
	{0xCDC90A208B33A1EEULL, "vxlt_blackemporium_in_0_0.vlm"},
	{0xCDC97499411A6F9DULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_c.impa.evt"},
	{0xCDCA5449385063EFULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\dissapear.gad"},
	{0xCDCA6273574D280BULL, "vxht_chantry_35j_0_0.vlm"},
	{0xCDCA7328CE25085FULL, "qun221_viscount6205726.cl"},
	{0xCDCA80CE83C48F26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture_2.evt"},
	{0xCDCB8D7E58757A93ULL, "vrc221_hugin6042564.cl"},
	{0xCDCC06F691C1722CULL, "art\\vfx\\combat\\vfx_m_arc_modal_act_c.impa.evt"},
	{0xCDCC083AB8CFDE7FULL, "378368789.ogg"},
	{0xCDCC139CF8E2B171ULL, "vxdr_0_deeproads_f9z_0_0.vlm"},
	{0xCDCC40920F3E8FFBULL, "6069183_m.fxe"},
	{0xCDCCFB34C0AFE705ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_imp_c.impa.evt"},
	{0xCDCD15A76C515CCDULL, "vxht_0_hightown_a2_atq_0_0.vlm"},
	{0xCDCDA5E4AC2BA41BULL, "841799823.ogg"},
	{0xCDCDF5FA7CED715DULL, "999078247.ogg"},
	{0xCDCF10C6D328DBA5ULL, "vxlt_hangedman_n_2v1_0_0.vlm"},
	{0xCDCF4C94DF5E6F28ULL, "vxlt_0_lowtown_a2_n_e21_0_0.vlm"},
	{0xCDCF786A4D1A1EEDULL, "694048871.ogg"},
	{0xCDCFC26469617CBDULL, "lit_tep_pla.mao"},
	{0xCDCFE9D24FB28EE2ULL, "ave121_jeven6105700.cl"},
	{0xCDD097CE83C92271ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture_2.gad"},
	{0xCDD0E0F77C9FAE22ULL, "fx_twirl01_blend.mao"},
	{0xCDD0E9511A629D77ULL, "876130930.ogg"},
	{0xCDD160FA588447D1ULL, "vxw_bonepitdestroyed_a3_29f_0_0.vlm"},
	{0xCDD169493856A59CULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\dissapear.evt"},
	{0xCDD1995D092F120CULL, "w_encounter1_12_0.msh"},
	{0xCDD19F5D092F1CFEULL, "w_encounter1_12_0.mmh"},
	{0xCDD1AB5D092F3095ULL, "w_encounter1_12_0.mao"},
	{0xCDD20443ECF9F026ULL, "vxlt_0_lowtown_a2_n_c_hks_0_0.vlm"},
	{0xCDD2E50459A6C003ULL, "dae000_arishok6081800.cl"},
	{0xCDD3014D3685AB5AULL, "hlac_ht_mansion2_818_22x.dds"},
	{0xCDD3108405F56640ULL, "hlac_w_bonepit_196608_0z.dds"},
	{0xCDD31A5B5B9FF787ULL, "vxlt_gallowstemplar_n_5v5_0_0.vlm"},
	{0xCDD37935A134F6ABULL, "vxht_0_hightown_n_a6b_0_0.vlm"},
	{0xCDD37F1ABDC43C14ULL, "hlac_ca_caverns2prc_5802_39z.dds"},
	{0xCDD39571E201EE31ULL, "6177897_m.fxe"},
	{0xCDD3A9FAAC0B7530ULL, "vxlt_gallowstemplar_f_4sp_0_0.vlm"},
	{0xCDD3B566622714E7ULL, "zz_followers_banter6159090.cl"},
	{0xCDD418DB5536BD04ULL, "6001202_f.fxe"},
	{0xCDD42E9B8A5BB84FULL, "plt_lgt110pt_harrow_escape.nss"},
	{0xCDD449C479F1916FULL, "4620933.ogg"},
	{0xCDD44D0DBAD83404ULL, "gen_im_trs_dal_hvy_hlm_r1.uti"},
	{0xCDD490FEAEFCCD1EULL, "gen00fl_fenris6126607.cl"},
	{0xCDD4AE3AE98539F5ULL, "w_bonepitmining_a2_69_0.msh"},
	{0xCDD4B03AE9853D6BULL, "w_bonepitmining_a2_69_0.mmh"},
	{0xCDD4BC3AE9855108ULL, "w_bonepitmining_a2_69_0.mao"},
	{0xCDD4E66A70F0C96FULL, "1038691274.ogg"},
	{0xCDD5552A9BB83C90ULL, "vxw_bonepitmining_a2_178_0_0.vlm"},
	{0xCDD575F08AE34FD7ULL, "lt_0_lowtown_a2_32_0_ll.mmh"},
	{0xCDD5E8C99E700E39ULL, "6188688_m.fxe"},
	{0xCDD602410B96F419ULL, "hlac_w_bonepitmining_a2_262147_0y.dds"},
	{0xCDD61AD9FC78A273ULL, "6146895_m.fxe"},
	{0xCDD636248D6C64C3ULL, "6135909_m.fxe"},
	{0xCDD663ADC08BEEEBULL, "763359536.ogg"},
	{0xCDD66F9E458F23CBULL, "vxlt_backalley_n_1vu_0_0.vlm"},
	{0xCDD69730BC06D1A0ULL, "hm_lgt_ant_ass.mor"},
	{0xCDD6CBCC7B776385ULL, "art\\vfx\\creatures\\fxc_wisp_c.mmh"},
	{0xCDD6D45DABDD3251ULL, "6142552_m.fxe"},
	{0xCDD6F14C296550B9ULL, "llac_lt_undercityserialkiller_-1_18.dds"},
	{0xCDD78185373BAEC4ULL, "qcr000_cut2d_thechampion6025139.cl"},
	{0xCDD798CE83CF421FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture_2.ani"},
	{0xCDD844BF0D0C4B98ULL, "hm_ave_jeven.mor"},
	{0xCDD85F0B9CD4C37CULL, "vxlt_gallowstemplar_46y_0_1.vlm"},
	{0xCDD875CCAFB140ADULL, "llac_ca_caverns2_5802_59.dds"},
	{0xCDD87CD87CCD2774ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_exit.ani"},
	{0xCDD89D0C001CB591ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_floordirt_d.dds"},
	{0xCDD8A29F174E10C3ULL, "6129607_m.fxe"},
	{0xCDD8E7EA16EC3870ULL, "6088317_f.fxe"},
	{0xCDD975032EBF5D04ULL, "follower_banter6143241.cl"},
	{0xCDD97EABB5282C7AULL, "hlac_ht_mansion2b_818_42x.dds"},
	{0xCDD9BDB240F2D0CCULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar1_mcr370_playerflees.ani"},
	{0xCDD9D11CE91D6FBFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bucket_0.pwk"},
	{0xCDD9EEF691CD8685ULL, "art\\vfx\\combat\\vfx_m_arc_modal_act_c.impa.ani"},
	{0xCDDA291F5D1A0C08ULL, "6092421_m.fxe"},
	{0xCDDA5C83149EFE6DULL, "153980634.ogg"},
	{0xCDDB118720F2F3C3ULL, "w_bonepitdestroyed_a3_28_0.anb"},
	{0xCDDB3C499EB75AEBULL, "llac_ht_0_hightown_a3_-65536_1.dds"},
	{0xCDDC06D697A26216ULL, "vxlt_undercitydungeon_19a_0_0.vlm"},
	{0xCDDC94E6FD689766ULL, "6020488_m.fxe"},
	{0xCDDCFB6008549950ULL, "vxlt_house_eg_0_0.vlm"},
	{0xCDDD6FD9868FAFD0ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordc.mao"},
	{0xCDDD73D9868FB6E3ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordc.mmh"},
	{0xCDDD81D9868FCE3DULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_swordc.msh"},
	{0xCDDDC4E213C971BAULL, "lt_0_lowtown_a2_n_c_21_0_ll.mmh"},
	{0xCDDDE02C1D9ED7AFULL, "hlac_lt_backalley_n_0_18y.dds"},
	{0xCDDE1305A8523395ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw8sm.msh"},
	{0xCDDE1505A852370BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw8sm.mmh"},
	{0xCDDE2105A8524BA8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedraw8sm.mao"},
	{0xCDDE36B9D4DEA478ULL, "vxht_0_hightown_n_cbc_0_0.vlm"},
	{0xCDDE3D5C3D453C9DULL, "qun120cr_javaris6042981.cl"},
	{0xCDDE714938613DB5ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\dissapear.ani"},
	{0xCDDE79D87CD18DD6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_exit.gad"},
	{0xCDDF121631F7ABB2ULL, "vxlt_0_lowtown_docks_a3_cyd_0_1.vlm"},
	{0xCDDF85F72934F206ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanr.ani"},
	{0xCDDF963A3588E7A7ULL, "543762563.ogg"},
	{0xCDDFB2D80969208AULL, "445084638.ogg"},
	{0xCDDFD2D74A7A6051ULL, "vxw_bonepitdestroyed_a3_24e_0_0.vlm"},
	{0xCDE0361BFF8DD712ULL, "6034231_f.fxe"},
	{0xCDE0CAB240F9050EULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar1_mcr370_playerflees.gad"},
	{0xCDE1095AF2D3FF34ULL, "w_blightlands_169_0.phy"},
	{0xCDE128009FE81506ULL, "hlac_lt_0_lowtown_-1_7x.dds"},
	{0xCDE135A065624C59ULL, "1039800729.ogg"},
	{0xCDE160451A86460CULL, "6207454_m.fxe"},
	{0xCDE1AB81068F44B0ULL, "gen_im_wep_rog_dua_act3_04.uti"},
	{0xCDE22358B5DDF31EULL, "6094775_m.fxe"},
	{0xCDE243572D7B1DABULL, "6095110_f.fxe"},
	{0xCDE31A816C53BF2EULL, "vxca_caverns2prc_9u7_0_0.vlm"},
	{0xCDE31BAC057911CBULL, "349339185.ogg"},
	{0xCDE32F5E415E0E8DULL, "vxw_bonepitmining_a2_dc_0_0.vlm"},
	{0xCDE3C8CC7B81E952ULL, "art\\vfx\\creatures\\fxc_wisp_c.anb"},
	{0xCDE3EACBCC4131D5ULL, "vxw_encounter3_fp_0_0.vlm"},
	{0xCDE3F39602441D70ULL, "845350913.ogg"},
	{0xCDE4257037F92003ULL, "561866683.ogg"},
	{0xCDE48CB6EE6EBAF3ULL, "6143795_m.fxe"},
	{0xCDE54D482FFF5A15ULL, "gen00fl_isabela_init6175711.cl"},
	{0xCDE57FD4E3FD98C6ULL, "268874913.ogg"},
	{0xCDE594D87CD7DA35ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_right_exit.evt"},
	{0xCDE5C5B892EB6AD5ULL, "vxht_keep_6ye_0_0.vlm"},
	{0xCDE6A51CD3951C34ULL, "hlac_lt_0_lowtown_docks_n_-196611_2x.dds"},
	{0xCDE6DBA6210ECB0DULL, "character_character_p158.psh.fxo.dx11.assembly"},
	{0xCDE7158DBF095A56ULL, "hlac_w_bonepit_2_0x.dds"},
	{0xCDE728AE063C6AE9ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanl.ani"},
	{0xCDE7ABE159DBFF6CULL, "vxlt_0_lowtown_gpj_0_0.vlm"},
	{0xCDE7BB5A0EAB5C52ULL, "757520215.ogg"},
	{0xCDE87434BC616881ULL, "gen00fl_aveline6122089.cl"},
	{0xCDE8E28720FEE14FULL, "w_bonepitdestroyed_a3_28_0.mao"},
	{0xCDE8EE8720FEF5ECULL, "w_bonepitdestroyed_a3_28_0.mmh"},
	{0xCDE8F48720FEFFDEULL, "w_bonepitdestroyed_a3_28_0.msh"},
	{0xCDE927555FE93294ULL, "vxca_caverns1_i1j_0_1.vlm"},
	{0xCDE9F961503D8DCDULL, "vxht_0_hightown_a2_ajo_0_0.vlm"},
	{0xCDEA1B3735538D19ULL, "hlac_lt_warehouse_0_7y.dds"},
	{0xCDEA4FE360CBCA05ULL, "vxdr_primevaltomb_863_0_0.vlm"},
	{0xCDEAB24299B3615FULL, "gen00fl_sebastian6198192.cl"},
	{0xCDEAC14726D2F57CULL, "12750159.ogg"},
	{0xCDEAE65AF2DC1831ULL, "w_blightlands_169_0.msh"},
	{0xCDEAF45AF2DC304CULL, "w_blightlands_169_0.mao"},
	{0xCDEB005AF2DC44EFULL, "w_blightlands_169_0.mmh"},
	{0xCDEB8F8A1DDE370CULL, "482524843.ogg"},
	{0xCDEBA82D1FC3535AULL, "6081984_m.fxe"},
	{0xCDECF3D7927423ACULL, "6117977_m.fxe"},
	{0xCDED0477709B52D1ULL, "207427876.ogg"},
	{0xCDED6DF729410777ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanr.evt"},
	{0xCDEE43AE0642B76BULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanl.gad"},
	{0xCDEECA83F749C418ULL, "297825689.ogg"},
	{0xCDEF095F34E0EE98ULL, "mag100_karras6034529.cl"},
	{0xCDEF278BE18D61AAULL, "68451804.ogg"},
	{0xCDEF805E30E1ED36ULL, "vxht_chantry_2p5_0_0.vlm"},
	{0xCDEFC96C851B4648ULL, "6202151-whatisitnow.cut"},
	{0xCDF0B179649FF1E7ULL, "337734721.ogg"},
	{0xCDF107528051B51BULL, "w_sundermount_110_0.msh"},
	{0xCDF108363BE6F9E2ULL, "art\\vfx\\cinematic\\vfx_bld_dir_dsk_p.mmh"},
	{0xCDF119528051D466ULL, "w_sundermount_110_0.mao"},
	{0xCDF125528051E805ULL, "w_sundermount_110_0.mmh"},
	{0xCDF16CEF7CB6A1FEULL, "zz_fol_debug6072539.cl"},
	{0xCDF17530D6A43BE1ULL, "llac_dr_primevaltomb_663_2.dds"},
	{0xCDF211EB0A1696F0ULL, "art\\characters\\playercharacter\\humanfemale\\cth_betb.msh"},
	{0xCDF21626F3C6E2C3ULL, "zz_qun_debug6028228.cl"},
	{0xCDF21FEB0A16AE0AULL, "art\\characters\\playercharacter\\humanfemale\\cth_betb.mmh"},
	{0xCDF223EB0A16B541ULL, "art\\characters\\playercharacter\\humanfemale\\cth_betb.mao"},
	{0xCDF335B0A563E7D5ULL, "vxca_caverns1_ih9_0_0.vlm"},
	{0xCDF33D886E4F0583ULL, "qcr000_varric6065794.cl"},
	{0xCDF355EF597DAC54ULL, "vxht_mansion2_25t_0_0.vlm"},
	{0xCDF36EF729457400ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_rf_leanr.gad"},
	{0xCDF37CBB419177A8ULL, "vxca_caverns1_mines_j5q_0_0.vlm"},
	{0xCDF3A6BEAD16B53DULL, "lt_0_lowtown_docks_a3_223_0.mao"},
	{0xCDF3AABEAD16BC76ULL, "lt_0_lowtown_docks_a3_223_0.mmh"},
	{0xCDF3B4BEAD16CD74ULL, "lt_0_lowtown_docks_a3_223_0.msh"},
	{0xCDF3BE32A31C325EULL, "894853997.ogg"},
	{0xCDF40CCB2B0B0FC1ULL, "532304554.ogg"},
	{0xCDF4379584E18355ULL, "vxht_keep_6bd_0_0.vlm"},
	{0xCDF46EC528ED247CULL, "w_blightlands_95_0.mao"},
	{0xCDF47AC528ED381FULL, "w_blightlands_95_0.mmh"},
	{0xCDF480C528ED42C1ULL, "w_blightlands_95_0.msh"},
	{0xCDF4A77797D584B9ULL, "lowtown_fem_alien_elf_2_mob6166360.cl"},
	{0xCDF4E216990C6EB2ULL, "vxht_brothel_n_6er_0_0.vlm"},
	{0xCDF4E89D12D068D6ULL, "isa211_isabela6202166.cl"},
	{0xCDF540AE0648D0D0ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\movement\\mov_rf_injured_leanl.evt"},
	{0xCDF5466D7483BF5FULL, "lt_0_lowtown_a2_n_19_0_ll.mmh"},
	{0xCDF5989ADD273835ULL, "vxdr_primevaltomb_7u9_0_0.vlm"},
	{0xCDF5E942EA0661B5ULL, "26235953.ogg"},
	{0xCDF6F171BB46D546ULL, "and121_karl6093480.cl"},
	{0xCDF7481F68368095ULL, "351349778.ogg"},
	{0xCDF76344FDC641A2ULL, "6119292_m.fxe"},
	{0xCDF77B9F126D9979ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_isab_1.mao"},
	{0xCDF7A2A0DE9A5A4FULL, "298059242.ogg"},
	{0xCDF7E91C7EE27FE4ULL, "372116314.ogg"},
	{0xCDF7FA7E98090B32ULL, "vxw_blightlands_7d_0_0.vlm"},
	{0xCDF80B378F75228CULL, "hlac_lt_0_lowtown_a3_n_c_-131071_2y.dds"},
	{0xCDF81E77C79FCC6FULL, "and100st_anders_intro2.stg"},
	{0xCDF8606401040197ULL, "zz_mag_debug6027868.cl"},
	{0xCDF88B7A5AB765ABULL, "vxca_caverns1_jf6_0_0.vlm"},
	{0xCDF8BD7038DD5F4BULL, "769475810.ogg"},
	{0xCDF8D6C6DBB2BD72ULL, "6033307_m.fxe"},
	{0xCDFC284525962FE6ULL, "pro000_meeran6048840.cl"},
	{0xCDFC43193658C55EULL, "hlac_w_encounter3_-65540_0x.dds"},
	{0xCDFC61026E452B95ULL, "llac_ht_0_hightown_n_65532_0.dds"},
	{0xCDFC7EF602F66DACULL, "ht_0_hightown_29_0_lc.mmh"},
	{0xCDFC947702A2BAC0ULL, "ser200_aveline6201468.cl"},
	{0xCDFCDDF5BD741D60ULL, "vxht_brothel_n_5ek_0_0.vlm"},
	{0xCDFD1C0A585BCAE9ULL, "qun160_vashothlead6055583.cl"},
	{0xCDFD5F8BDBA9A23CULL, "fen110_fenris6100239.cl"},
	{0xCDFDAAB0AA5B3ADEULL, "6148940_m.fxe"},
	{0xCDFE658D11B65384ULL, "hlac_ca_caverns1_mines_24424_24z.dds"},
	{0xCDFE9D858BF7B5F4ULL, "mag110_vincento6017662.cl"},
	{0xCDFEA6DB4CD14DF9ULL, "zz_isa_debug6175291.cl"},
	{0xCDFECC0BA60CA0BBULL, "one130_thug6131575.cl"},
	{0xCDFF01A58ECA981CULL, "dae000_maren6148441.cl"},
	{0xCDFF79AF4C9C3A69ULL, "art\\vfx\\combat\\vfx_m_bld_sacrifice_imp_c.impa.ani"},
	{0xCE00020D9534FD63ULL, "6151284_m.fxe"},
	{0xCE001318254A3F8DULL, "vxw_bonepitmining_a2_na_0_0.vlm"},
	{0xCE009A2E05BEF207ULL, "54396066.ogg"},
	{0xCE00BFC7D6C1FEACULL, "hlac_lt_gallowstemplar_n_2394_8y.dds"},
	{0xCE00C0B463F4BB46ULL, "400778945.ogg"},
	{0xCE00F0EC35567752ULL, "6111592_f.fxe"},
	{0xCE0104D98002BD89ULL, "6016802_m.fxe"},
	{0xCE0209F2EC2A0636ULL, "ht_bartrand_n_3370.rml"},
	{0xCE02CDF2981E79B4ULL, "6179788_m.fxe"},
	{0xCE0383A0A3B01C17ULL, "6061231_m.fxe"},
	{0xCE038E6B4B2E6EF6ULL, "vxlt_darktown_5av_0_0.vlm"},
	{0xCE0393CF0DC666BAULL, "vxdr_0_deeproads_f4v_0_0.vlm"},
	{0xCE03AA8955438DE1ULL, "vxlt_0_lowtown_docks_a3_n_cy8_0_1.vlm"},
	{0xCE03F23E352EB0BEULL, "15791874.ogg"},
	{0xCE046BCF251BDD84ULL, "hlac_w_encounter3_-393214_0z.dds"},
	{0xCE04C77E4C23E8D7ULL, "vxca_caverns1_mines_e82_0_0.vlm"},
	{0xCE054CDBBF7DFED9ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_csand02_n.dds"},
	{0xCE05A8529349EDC6ULL, "qun120_arishok6050271.cl"},
	{0xCE06151508FC57E1ULL, "gen00fl_justice.utc"},
	{0xCE06503AAAA10E86ULL, "420310923.ogg"},
	{0xCE066C5B89FF23B2ULL, "vxw_blightlands_8e_0_0.vlm"},
	{0xCE0694AF4CA286EBULL, "art\\vfx\\combat\\vfx_m_bld_sacrifice_imp_c.impa.gad"},
	{0xCE06D0E0689E2F51ULL, "vxlt_0_lowtown_csj_0_0.vlm"},
	{0xCE074CDE14641705ULL, "vxht_keep_5wt_0_0.vlm"},
	{0xCE075D5E37167AAEULL, "vxlt_0_lowtown_docks_bxd_0_0.vlm"},
	{0xCE07654DAB946889ULL, "518904857.ogg"},
	{0xCE077A586CB38802ULL, "follower_banter6143104.cl"},
	{0xCE0806F5A3D5148AULL, "plt_zz_genpt_musical_mabari.plo"},
	{0xCE082003F7BFC7ECULL, "hlac_lt_gallowstemplar_n_2394_10x.dds"},
	{0xCE08A25B1DEC8396ULL, "art\\levels\\ztools\\default\\generic\\prp_joe_1.mao"},
	{0xCE090F7B9CAC2738ULL, "vxdr_primevaltomb_7ob_0_0.vlm"},
	{0xCE091289D127BC00ULL, "vxlt_0_lowtown_docks_a3_n_d1c_0_0.vlm"},
	{0xCE091360055EA6C3ULL, "hlac_dr_dungeon_3215_5y.dds"},
	{0xCE09DA8F767166B1ULL, "6172902_m.fxe"},
	{0xCE09F4F3BE984A42ULL, "875110369.ogg"},
	{0xCE0A15776637DFF4ULL, "vxw_encounter3_76_0_0.vlm"},
	{0xCE0A548E703CC1D4ULL, "6195837_f.fxe"},
	{0xCE0A8CD00FAED5A1ULL, "78763372.ogg"},
	{0xCE0AB4321B6DE393ULL, "hlac_ht_keep_n_4519_2x.dds"},
	{0xCE0AC1DE2CF800BEULL, "sky_ht_store.mao"},
	{0xCE0AC7F47AD4A24DULL, "6111611_m.fxe"},
	{0xCE0AF186A65E9326ULL, "bind_dis.mao"},
	{0xCE0B2F48159FF22BULL, "6137239_m.fxe"},
	{0xCE0B4FD2AF6A35E0ULL, "vxht_brothel_n_5hh_0_0.vlm"},
	{0xCE0B54961763BFCEULL, "6110000_m.fxe"},
	{0xCE0BA4BA5E2DE8C5ULL, "950998218.ogg"},
	{0xCE0C005A81836643ULL, "hlac_w_sundermount_65533_0x.dds"},
	{0xCE0C8E4B51465605ULL, "vxht_0_hightown_n_cbz_0_0.vlm"},
	{0xCE0D91AF4CA8A050ULL, "art\\vfx\\combat\\vfx_m_bld_sacrifice_imp_c.impa.evt"},
	{0xCE0DD8A9792358CAULL, "w_blightlands_123_0.mao"},
	{0xCE0DDCA979235F19ULL, "w_blightlands_123_0.mmh"},
	{0xCE0DE6A979237007ULL, "w_blightlands_123_0.msh"},
	{0xCE0E918543C51E57ULL, "60266395.ogg"},
	{0xCE0EF7338EDCEF78ULL, "998615587.ogg"},
	{0xCE0F4719B584328EULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_lp.ani"},
	{0xCE0F5A7A48BD7872ULL, "vxlt_0_lowtown_a2_ffk_0_0.vlm"},
	{0xCE0FB64E1245B6D3ULL, "llac_ht_playershouse_n_855_11.dds"},
	{0xCE1037769C14F8CEULL, "6013338_m.fxe"},
	{0xCE1050D1A6775184ULL, "ave_mod_item_acq.ncs"},
	{0xCE1067557ABC02DFULL, "83019017.ogg"},
	{0xCE1068C8C72C6CCFULL, "hlac_lt_gallowscourtyard_131069_1z.dds"},
	{0xCE10DA80009DC381ULL, "llac_lt_0_lowtown_docks_131070_1.dds"},
	{0xCE111B4F68CDB455ULL, "vxht_keep_6hb_0_0.vlm"},
	{0xCE111B5F907E94A6ULL, "vxlt_hangedman_n_26f_0_0.vlm"},
	{0xCE1130EF4DB6FD5AULL, "710723552.ogg"},
	{0xCE11437145021C6EULL, "vxlt_blackemporium_vc_0_0.vlm"},
	{0xCE1181605FC3BB3AULL, "6042224_m.fxe"},
	{0xCE11F3F17C883ED0ULL, "hlac_lt_0_lowtown_131072_5z.dds"},
	{0xCE1200483D248776ULL, "vxlt_darktown_5jw_0_0.vlm"},
	{0xCE120F6E99448EE1ULL, "gen_im_arm_boo_lgt_act2_03.uti"},
	{0xCE124ABE40B81577ULL, "follower_banter6169778.cl"},
	{0xCE1263855E5021BFULL, "vxlt_0_lowtown_gpe_0_0.vlm"},
	{0xCE12664473D5EF48ULL, "106217276.ogg"},
	{0xCE132BBC66228511ULL, "vxw_bonepitmining_a2_5t_0_0.vlm"},
	{0xCE1341454DDA8CF3ULL, "vxht_0_hightown_n_c9x_0_1.vlm"},
	{0xCE13513D9B2BA240ULL, "vxdr_0_deeproads_cfv_0_1.vlm"},
	{0xCE137974890210EDULL, "6209956_m.fxe"},
	{0xCE140BBFE35B4D49ULL, "mag320pt_otl_main.ncs"},
	{0xCE14C7ADED8B40A9ULL, "555390154.ogg"},
	{0xCE14D6C391B9D2DEULL, "vxca_cavern1_haunted_edc_0_0.vlm"},
	{0xCE14DDF4A20B1EA5ULL, "6148811_m.fxe"},
	{0xCE154E534D265022ULL, "215058366.ogg"},
	{0xCE15A543DE1ABB5AULL, "mcr370_varric6098306.cl"},
	{0xCE15BEBB065A2F85ULL, "vxht_keep_5zs_0_0.vlm"},
	{0xCE15D5B82E916E4CULL, "zz_seb_debug6215782.cl"},
	{0xCE165CBCA5AFF355ULL, "vxht_0_hightown_n_bgh_0_0.vlm"},
	{0xCE166699FD8635E9ULL, "mrl150_merrill6071930.cl"},
	{0xCE1682B4056D60A9ULL, "art\\characters\\creatures\\mubari\\animation\\amb_wk_sus.ani"},
	{0xCE16AB60092344C2ULL, "dae200_bran6094872.cl"},
	{0xCE16AE75C042D9D1ULL, "dae000_bodhan6146872.cl"},
	{0xCE16B06C9C7181A2ULL, "342537521.ogg"},
	{0xCE16E72B1AB0526EULL, "hlac_lt_0_lowtown_docks_131071_2y.dds"},
	{0xCE16F03E95AD1362ULL, "one120_athenril6113301.cl"},
	{0xCE17D81E98A1867EULL, "llac_lt_foundry_n_320_7.dds"},
	{0xCE17DE82AFA7418BULL, "6177194_m.fxe"},
	{0xCE17EA7F6408E0ECULL, "vxht_chantry_n_30q_0_0.vlm"},
	{0xCE180D486DB2191DULL, "follower_banter6163222.cl"},
	{0xCE184DE94343D2FEULL, "763374262.ogg"},
	{0xCE1892EEC99F32E4ULL, "6025359_m.fxe"},
	{0xCE18B3920D338161ULL, "911068660.ogg"},
	{0xCE18BD64870A86FCULL, "7093753.ogg"},
	{0xCE18CCD08D87AF61ULL, "hlac_ht_0_hightown_a3_1_1y.dds"},
	{0xCE1914C62537D110ULL, "gen00fl_bethany6028806.cl"},
	{0xCE1925C7919758F1ULL, "one300_murder_followers.bnk"},
	{0xCE197F5D096C9BB5ULL, "w_encounter1_12_0.rml"},
	{0xCE19DFE58D47CC20ULL, "static_static_p3.vsh.fxo.dx11.assembly"},
	{0xCE19E93790B9A8CFULL, "6077852_m.fxe"},
	{0xCE19FF7FF5A4C13FULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcratepile.phy"},
	{0xCE1A07363C09FC89ULL, "art\\vfx\\cinematic\\vfx_bld_dir_dsk_p.anb"},
	{0xCE1ADB305DE15B2FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_lgta_0d.dds"},
	{0xCE1B7953192F889EULL, "target_creature.cur"},
	{0xCE1BEEDAFEDD0CBCULL, "vxlt_warehouse_n_au_0_0.vlm"},
	{0xCE1C028535E6978CULL, "cs_pro000cs_voyage.bnk"},
	{0xCE1C07C791995B01ULL, "one300_murder_followers.cnv"},
	{0xCE1C3EF42E5719C1ULL, "872715207.ogg"},
	{0xCE1C3F19B58EAFDFULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_lp.evt"},
	{0xCE1CAE2E676D3DE4ULL, "gamestopgobig.dds"},
	{0xCE1CD078AAA44602ULL, "302887015.ogg"},
	{0xCE1CE73B7386256FULL, "vxca_cavern1_haunted_ijx_0_0.vlm"},
	{0xCE1D10F6BE8FD03FULL, "hlac_lt_0_lowtown_-3_4z.dds"},
	{0xCE1D55DE5FC94170ULL, "vxht_0_hightown_bfy_0_0.vlm"},
	{0xCE1D739DD03AE5C8ULL, "397952906.ogg"},
	{0xCE1D84C7D51B67F9ULL, "one351ip_charred_corpse.utp"},
	{0xCE1D9DB40573AD2BULL, "art\\characters\\creatures\\mubari\\animation\\amb_wk_sus.gad"},
	{0xCE1DAAB378E48B3AULL, "gen00fl_isabela6090432.cl"},
	{0xCE1DBE705F042B84ULL, "279127802.ogg"},
	{0xCE1DE0B010FE74CCULL, "fx_twirl_air.dds"},
	{0xCE1EDC90C6D76072ULL, "6164592_f.fxe"},
	{0xCE1F1695C916A2FAULL, "921784269.ogg"},
	{0xCE1F1748F440A1BBULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb4a.msh"},
	{0xCE1F1B9D48E15618ULL, "mrl250cr_varterral.ncs"},
	{0xCE1F349816382E96ULL, "mcr370cr_corpse_2.utc"},
	{0xCE1F3548F440D4A5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb4a.mmh"},
	{0xCE1F4D9DE73B3E4AULL, "vxht_0_hightown_a3_cf7_0_0.vlm"},
	{0xCE1F844195261A2DULL, "art\\characters\\creatures\\arh_abm_rev\\c_arh_head.mao"},
	{0xCE1F98BDB407AC27ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_templdesk_d.dds"},
	{0xCE201BAD4535B702ULL, "llac_ht_0_hightown_a2_0_1.dds"},
	{0xCE2081BD339AA96BULL, "6120921_m.fxe"},
	{0xCE2167529E9A78AEULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_flare02.mao"},
	{0xCE2245B42EB8129BULL, "dae000ip_crowd_em3_talk18.utp"},
	{0xCE22490D4A56A62EULL, "vxca_caverns2_8h1_0_1.vlm"},
	{0xCE22643BFC333D36ULL, "vxlt_0_lowtown_a2_ghl_0_0.vlm"},
	{0xCE22BCD57FBBB436ULL, "6024541_m.fxe"},
	{0xCE230200280F3AF1ULL, "vxht_0_hightown_n_7j2_0_0.vlm"},
	{0xCE230AEF913A1C92ULL, "vxlt_0_lowtown_e1a_0_0.vlm"},
	{0xCE232019B59499C8ULL, "art\\characters\\creatures\\dragon\\animation\\tal_ovlm_lp.gad"},
	{0xCE239AB405781390ULL, "art\\characters\\creatures\\mubari\\animation\\amb_wk_sus.evt"},
	{0xCE23B335DF7F5127ULL, "6119688_m.fxe"},
	{0xCE245E4F4056468AULL, "zz_followers_banter6159321.cl"},
	{0xCE248106FA72721BULL, "199479963.ogg"},
	{0xCE2489435D15E81EULL, "563515542.ogg"},
	{0xCE24FB070F2DD19EULL, "125666763.ogg"},
	{0xCE2539826041A639ULL, "da2_prc_seb_it-it.cre"},
	{0xCE2581AD65FA5E68ULL, "vxlt_hangedman_sn_0_0.vlm"},
	{0xCE261F3375E74F5CULL, "ch2armtnt_ds1y.psh.fxo"},
	{0xCE266F9A9C610776ULL, "gen00fl_fenris6126364.cl"},
	{0xCE27BCACA908AC5BULL, "6137635_m.fxe"},
	{0xCE27D470C4675B25ULL, "439925201.ogg"},
	{0xCE27D6226F6F5FB5ULL, "gen00fl_aveline6116644.cl"},
	{0xCE27E46B91DE9BF9ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_bdy_hvyn.mao"},
	{0xCE2868B19877F3D5ULL, "6154151_m.fxe"},
	{0xCE28E544A40574C8ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_5z.dds"},
	{0xCE2995E143DF8FBDULL, "176664168.ogg"},
	{0xCE29CE1291B26A04ULL, "6104308_m.fxe"},
	{0xCE29E1E09363DCA6ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack02.ani"},
	{0xCE29FB633EDC4C05ULL, "hlac_ca_caverns1_24424_6z.dds"},
	{0xCE2A29C5DCB0A7E9ULL, "lt_0_lowtown_a2_n_c_ll.mmh"},
	{0xCE2A2E2CB50F17FAULL, "6182270_m.fxe"},
	{0xCE2A3B547AA6267FULL, "zz_rogue3.cut"},
	{0xCE2A51EA42484BEBULL, "vxlt_0_lowtown_docks_a3_n_cp4_0_0.vlm"},
	{0xCE2ADBC67715264CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_exit.ani"},
	{0xCE2B007FF5B33644ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcratepile.mmh"},
	{0xCE2B047FF5B33C97ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcratepile.mao"},
	{0xCE2B09B75EA09EE8ULL, "hlac_ht_0_hightown_n_65536_1z.dds"},
	{0xCE2B167FF5B35BE6ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcratepile.msh"},
	{0xCE2B2B416A9EE864ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookofinquisition_sa.dds"},
	{0xCE2B6640BAEBC6B9ULL, "699820582.ogg"},
	{0xCE2BDF0B42E6E53DULL, "vxlt_gallowstemplar_n_5qn_0_0.vlm"},
	{0xCE2C750E3EB0FAF4ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzlargebrown_0.pwk"},
	{0xCE2D414E99501135ULL, "6119880_m.fxe"},
	{0xCE2DFF19746AC5A6ULL, "vxlt_hangedman_n_28h_0_0.vlm"},
	{0xCE2E1A09FB0F7B6CULL, "lgt200cr_tome_pride.utc"},
	{0xCE2EE4022110B876ULL, "vxlt_darktown_5py_0_0.vlm"},
	{0xCE2EF2F1AEAADE3BULL, "6197134_m.fxe"},
	{0xCE2F7B73BBB76F07ULL, "6169856_m.fxe"},
	{0xCE3019CD34A64254ULL, "hlac_ca_cave_2194_1x.dds"},
	{0xCE302C1C10F10447ULL, "80736737.ogg"},
	{0xCE30595065FCE1AAULL, "vxlt_warehouse_n_13j_0_0.vlm"},
	{0xCE30E2AD8754405EULL, "vxht_chantry_n_2cs_0_0.vlm"},
	{0xCE3125528087CC0EULL, "w_sundermount_110_0.rml"},
	{0xCE31727EE7C4272AULL, "lgt260_eau_tmp_flowergirl6143870.cl"},
	{0xCE3187122E4A3D28ULL, "6127530_m.fxe"},
	{0xCE31E8C6771B5A8EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_exit.gad"},
	{0xCE322142591DFE2AULL, "105186164.ogg"},
	{0xCE324244418AC70BULL, "llac_lt_0_lowtown_a2_-131073_2.dds"},
	{0xCE329BDEBE185E8EULL, "w_woundedcoast_128_0.msh"},
	{0xCE329DD5727EDE05ULL, "157108160.ogg"},
	{0xCE32A9DEBE1876DFULL, "w_woundedcoast_128_0.mao"},
	{0xCE32B5DEBE188B7CULL, "w_woundedcoast_128_0.mmh"},
	{0xCE33072D9B34C7FFULL, "vxht_0_hightown_a3_c1k_0_0.vlm"},
	{0xCE333D4A52576817ULL, "hlac_ca_caverns2_5802_42z.dds"},
	{0xCE3350BE63ED3672ULL, "combat_deathblow_10.gda"},
	{0xCE346AEDEC3FE45CULL, "221361157.ogg"},
	{0xCE34837143156422ULL, "hlac_lt_0_lowtown_docks_a3_65534_1y.dds"},
	{0xCE34B15A2CA34594ULL, "6116974_m.fxe"},
	{0xCE34BE1DB4986DC6ULL, "6154116_m.fxe"},
	{0xCE34D2279946933FULL, "w_encounter3_67_0.phy"},
	{0xCE34EC8E617020C9ULL, "lgt220_shp_mag_guard6149882.cl"},
	{0xCE352B477D6CCBC6ULL, "6034932_f.fxe"},
	{0xCE354A5ADBAB9281ULL, "dae200cr_viscount6087212.cl"},
	{0xCE35A6093646D56CULL, "vxca_cavern1_haunted_j5e_0_0.vlm"},
	{0xCE35B8F6476EF525ULL, "337809223.ogg"},
	{0xCE35EBDE9A5F702AULL, "hlac_lt_undercityserialkiller_-1_7y.dds"},
	{0xCE35F904FAD6DAA4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_tall.msh"},
	{0xCE360B04FAD6F90DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_tall.mao"},
	{0xCE360F04FAD70046ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_tall.mmh"},
	{0xCE366A761C1A1DF3ULL, "6018558_m.fxe"},
	{0xCE36B08BE5500EA4ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_meredith_meredithsintro.ani"},
	{0xCE36BC730B3EE582ULL, "hlac_ht_chantry_n_2534_14y.dds"},
	{0xCE36BED8505C11CFULL, "vxlt_gallowstemplar_f_46x_0_0.vlm"},
	{0xCE36C9E0936E3E17ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack02.evt"},
	{0xCE3773FEE5B5FFE4ULL, "777776975.ogg"},
	{0xCE37E3C6771FBE6DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\commander\\comm_exit.evt"},
	{0xCE3818A9F018317BULL, "atl_shared_dxt1_dat.xml"},
	{0xCE3853EE4B866446ULL, "6070441_m.fxe"},
	{0xCE386E38C82FA597ULL, "170192379.ogg"},
	{0xCE39167DBA94B375ULL, "vxlt_0_lowtown_a2_n_eqq_0_0.vlm"},
	{0xCE396B7CACC11E2AULL, "hlac_ca_caverns2prc_5802_1y.dds"},
	{0xCE39936B2C76EE3FULL, "zz_one_debug6099945.cl"},
	{0xCE3A03EBB64B57B4ULL, "366824023.ogg"},
	{0xCE3A8DA23521E31DULL, "261078784.ogg"},
	{0xCE3AAD06C36CB8F4ULL, "art\\effects\\cloudshadows\\w_bonepitdestroyed_a3\\cloud1.dds"},
	{0xCE3C06F2EC5B79EAULL, "ht_bartrand_n_3370.anb"},
	{0xCE3C74C02DE9575AULL, "mag131_taraneh6088679.cl"},
	{0xCE3C8B41A7C4ADA1ULL, "6207671_m.fxe"},
	{0xCE3CAD8BE5547506ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_meredith_meredithsintro.gad"},
	{0xCE3CC32954345242ULL, "155083479.ogg"},
	{0xCE3CDA061E2186B2ULL, "291992381.ogg"},
	{0xCE3D1C50D42EB424ULL, "dae000cr_marethari.utc"},
	{0xCE3DCAE093745EA0ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_attack02.gad"},
	{0xCE3E29F671105E72ULL, "vxht_keep_73e_0_0.vlm"},
	{0xCE3ECB2D57F2FA2AULL, "vxlt_warehouse_n_18k_0_0.vlm"},
	{0xCE3F47F5E01F6E36ULL, "vxlt_0_lowtown_a2_gbn_0_0.vlm"},
	{0xCE3FE04AA7699DF9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls101a_0s.dds"},
	{0xCE40D5B4856B0AB6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v3.evt"},
	{0xCE40E25886449D47ULL, "516731356.ogg"},
	{0xCE410604FAE0F811ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_tall.phy"},
	{0xCE410BD149FDEB2BULL, "6136874_m.fxe"},
	{0xCE410F87C494006FULL, "vxca_caverns2prc_9xm_0_0.vlm"},
	{0xCE411CC8AC57E7C1ULL, "hlac_ht_mansion2b_818_29x.dds"},
	{0xCE427612C99C26B5ULL, "w_bonepit_89_0.msh"},
	{0xCE427812C99C2A2BULL, "w_bonepit_89_0.mmh"},
	{0xCE428412C99C3EC8ULL, "w_bonepit_89_0.mao"},
	{0xCE42940584E14475ULL, "vrc321_housekeeper6087329.cl"},
	{0xCE42BD79906CD49BULL, "vxlt_0_lowtown_a2_n_ehx_0_0.vlm"},
	{0xCE42D7BB996799C4ULL, "6127414_f.fxe"},
	{0xCE4338DC930D80CCULL, "291218641.ogg"},
	{0xCE434AC224010A3BULL, "w_sundermount_124_0.rml"},
	{0xCE436814C12E0712ULL, "776258669.ogg"},
	{0xCE447D55FCAF2015ULL, "hlac_lt_0_lowtown_65537_7y.dds"},
	{0xCE44E38FF5ACC7B8ULL, "follower_banter6148464.cl"},
	{0xCE45245EE8A921A5ULL, "6017492_m.fxe"},
	{0xCE4588325A142E30ULL, "320813597.ogg"},
	{0xCE45D32799550844ULL, "w_encounter3_67_0.mmh"},
	{0xCE45D72799550E97ULL, "w_encounter3_67_0.mao"},
	{0xCE45E92799552DE6ULL, "w_encounter3_67_0.msh"},
	{0xCE45F89920DC0C35ULL, "6005828_m.fxe"},
	{0xCE462F94CBD44103ULL, "vxlt_undercityencounter_1n8_0_0.vlm"},
	{0xCE463A1DDD4E6D2BULL, "141093274.ogg"},
	{0xCE46BE39A04E4C74ULL, "hlac_lt_0_lowtown_a3_n_c_131073_2z.dds"},
	{0xCE4713C1298711CDULL, "6075290_m.fxe"},
	{0xCE47998251B01991ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_1z.dds"},
	{0xCE47CCB485711A41ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v3.gad"},
	{0xCE47E32F2FFEA32EULL, "6117027_m.fxe"},
	{0xCE49352CD9357E0EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_large.msh"},
	{0xCE49432CD935965FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_large.mao"},
	{0xCE494F2CD935AAFCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_large.mmh"},
	{0xCE499DF5481EEF3DULL, "6166363_m.fxe"},
	{0xCE49D039CBE8BA16ULL, "mag110_arianni6016904.cl"},
	{0xCE4A51C224073493ULL, "w_sundermount_124_0.phy"},
	{0xCE4A6735FA3B7D5BULL, "137139555.ogg"},
	{0xCE4A7F00174AB95FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_001_0t.dds"},
	{0xCE4A97B303DC2261ULL, "gen00fl_merrill6144981.cl"},
	{0xCE4AA9DE1224BC09ULL, "6161591_m.fxe"},
	{0xCE4AC2E7977854EDULL, "hlac_lt_0_lowtown_a2_n_c_65535_7z.dds"},
	{0xCE4B5A05A5287627ULL, "vxlt_0_lowtown_docks_a3_d31_0_0.vlm"},
	{0xCE4BD6AB92970F3BULL, "6199334_m.fxe"},
	{0xCE4C6D046A052669ULL, "216440644.ogg"},
	{0xCE4CAFD10F17D52FULL, "6185230_m.fxe"},
	{0xCE4CCAAB4F5625FAULL, "qun111_ketojan6041049.cl"},
	{0xCE4E46E3C31FDEA7ULL, "vxlt_0_lowtown_docks_a3_api_0_0.vlm"},
	{0xCE4E5F31BA344242ULL, "vxw_sundermount_31f_0_1.vlm"},
	{0xCE4E8B10810ACA47ULL, "gen00fl_aveline6114868.cl"},
	{0xCE4ECDB485773AEFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_atk_v3.ani"},
	{0xCE4F48332999CBE0ULL, "770459574.ogg"},
	{0xCE4F5BC2A5B05472ULL, "6072314_f.fxe"},
	{0xCE4F956967E4F94FULL, "hlac_w_encounter2_262144_0x.dds"},
	{0xCE4F95BE27532616ULL, "vxlt_undercitydungeon_1ai_0_0.vlm"},
	{0xCE4F96AEF0E4F52EULL, "vxlt_0_lowtown_docks_bcy_0_0.vlm"},
	{0xCE4FE741B666B710ULL, "1044023752.ogg"},
	{0xCE500B8C06C7C24EULL, "hlac_ht_0_hightown_n_65534_1z.dds"},
	{0xCE5156B1C5AD7F60ULL, "477803740.ogg"},
	{0xCE5175CDA3347406ULL, "vxdr_primevaltomb_7t1_0_0.vlm"},
	{0xCE51BC1EA19673FDULL, "vxca_caverns1_cpc_0_0.vlm"},
	{0xCE529B4129D6C65BULL, "art\\characters\\creatures\\spider\\animation\\hv_dmg.evt"},
	{0xCE52BD41A2ACB2D3ULL, "260029154.ogg"},
	{0xCE52F0D34A9986A8ULL, "vxlt_warehouse_ax_0_0.vlm"},
	{0xCE52F878C2874650ULL, "40643467.ogg"},
	{0xCE5394DC49E52030ULL, "6130127_m.fxe"},
	{0xCE53AEAD2EF8C16DULL, "hlac_w_blightlands_0_2x.dds"},
	{0xCE53E1438D8D1A69ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.cess.ani"},
	{0xCE53F873F4A6B14EULL, "gen00fl_varric6109217.cl"},
	{0xCE5469C767725928ULL, "isa211_isabela6071616.cl"},
	{0xCE54F3E202EA9137ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_slave_freize_n.dds"},
	{0xCE556837E99551B5ULL, "art\\vfx\\cinematic\\vfx_meredith_smokey.mmh"},
	{0xCE557D476CBC2602ULL, "659251868.ogg"},
	{0xCE5589124F632629ULL, "vxca_cavern1_haunted_i1i_0_2.vlm"},
	{0xCE55E73605F5EDB4ULL, "llac_lt_0_lowtown_docks_a3_n_65535_4.dds"},
	{0xCE5607D3D8CBBB36ULL, "t3_hey_bk1.tnt"},
	{0xCE561B49A4926DC3ULL, "vxlt_0_lowtown_a2_n_c_f5b_0_0.vlm"},
	{0xCE5655605660ECB6ULL, "563741149.ogg"},
	{0xCE56B5FD64309AC3ULL, "mag311_keran6041572.cl"},
	{0xCE574F0E732D0F44ULL, "hlac_w_encounter3_-262147_0z.dds"},
	{0xCE5761364D2BC262ULL, "t3_wdg_d00.tnt"},
	{0xCE57A2C4B61C4341ULL, "178470095.ogg"},
	{0xCE57E48E2C4F6AD1ULL, "57339334.ogg"},
	{0xCE58302245FD4870ULL, "6095925_m.fxe"},
	{0xCE58C80AED450BACULL, "6161064_m.fxe"},
	{0xCE591BBB041AD489ULL, "6166491_m.fxe"},
	{0xCE593A8382774645ULL, "6109756_f.fxe"},
	{0xCE5953C70ACBFAA8ULL, "one300_zevran6075859.cl"},
	{0xCE59944129DCD89CULL, "art\\characters\\creatures\\spider\\animation\\hv_dmg.gad"},
	{0xCE5A64D81D58FA53ULL, "zz_approval_debug6003116.cl"},
	{0xCE5A6885842F09A3ULL, "fx_grass.mao"},
	{0xCE5A9EF00BF5AE56ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredidies.ani"},
	{0xCE5ABCF117CD1E72ULL, "6176740_f.fxe"},
	{0xCE5AC256D614A029ULL, "6119192_m.fxe"},
	{0xCE5AC78D6C45A488ULL, "lt_0_lowtown_docks_water_16556.mmh"},
	{0xCE5AD38D6C45B92BULL, "lt_0_lowtown_docks_water_16556.mao"},
	{0xCE5AE58D6C45D7B2ULL, "lt_0_lowtown_docks_water_16556.msh"},
	{0xCE5AFC438D9366EBULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.cess.gad"},
	{0xCE5B36A62812D7DEULL, "lgt160pt_cod_avern_lett_1.utp"},
	{0xCE5B3C9548666D72ULL, "6043564_m.fxe"},
	{0xCE5C0C85361C8CBEULL, "cs_pro000cs_voyage.wbd"},
	{0xCE5C4F804E372AE3ULL, "follower_banter6160786.cl"},
	{0xCE5CB343DD933E84ULL, "qun200_party_comment6095382.cl"},
	{0xCE5D269AEAD5EBECULL, "751234710.ogg"},
	{0xCE5D3C3C7A620DA0ULL, "qun221_viscount6070181.cl"},
	{0xCE5DBB2A4C51AB22ULL, "549593584.ogg"},
	{0xCE5DDDCB9079DD03ULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_bdy_ogra_0d.dds"},
	{0xCE5E1B337E1E27F9ULL, "6116576_f.fxe"},
	{0xCE5E965F66C99EA7ULL, "919051147.ogg"},
	{0xCE5EAA80C5E050ACULL, "vxw_encounter2_5s_0_0.vlm"},
	{0xCE5F26534031FECDULL, "534891353.ogg"},
	{0xCE5F49EE7C73483BULL, "vxlt_undercityserialkiller_1c7_0_0.vlm"},
	{0xCE5F8FC39CE56B5FULL, "hlac_ca_cavern1_haunted_24425_3y.dds"},
	{0xCE605DAD3ABAF7BFULL, "vxca_cavern1_haunted_j5j_0_0.vlm"},
	{0xCE60934129E2F59AULL, "art\\characters\\creatures\\spider\\animation\\hv_dmg.ani"},
	{0xCE60F273828E88EEULL, "6136732_m.fxe"},
	{0xCE61A4DDAB89CEF0ULL, "lt_0_lowtown_docks_a3_208_0.mao"},
	{0xCE61A8DDAB89D503ULL, "lt_0_lowtown_docks_a3_208_0.mmh"},
	{0xCE61B6DDAB89ED5DULL, "lt_0_lowtown_docks_a3_208_0.msh"},
	{0xCE61F9438D998050ULL, "art\\vfx\\combat\\vfx_glb_arw_nat_c.cess.evt"},
	{0xCE626DFA94790264ULL, "6031141_m.fxe"},
	{0xCE62BCE37A09E812ULL, "6114813_m.fxe"},
	{0xCE636537E9A18A42ULL, "art\\vfx\\cinematic\\vfx_meredith_smokey.anb"},
	{0xCE63910F3C5E6E88ULL, "148781821.ogg"},
	{0xCE63C92240C565E1ULL, "mcr370_meredith6085986.cl"},
	{0xCE64C51E01FD859FULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_ext_c.mmh"},
	{0xCE655053ABB47B9BULL, "6060548_m.fxe"},
	{0xCE660E48F47C9552ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hb4a.phy"},
	{0xCE6669A67E42DE82ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_chantdoor_d.dds"},
	{0xCE66D8FDEA664054ULL, "dre000_dougal6129839.cl"},
	{0xCE67516307EB7510ULL, "hlac_ht_mansion2_818_33x.dds"},
	{0xCE6761208A80BFFBULL, "vxca_cavern1_haunted_gkh_0_1.vlm"},
	{0xCE6785F6F80E30B9ULL, "6121700_m.fxe"},
	{0xCE679BF1482EFB43ULL, "654272052.ogg"},
	{0xCE67D8E874F5066EULL, "6074030_m.fxe"},
	{0xCE68AE2BDFCE5523ULL, "cod_hst_kirk_arch_1a_0.utp"},
	{0xCE694FDBF7C469F7ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_act_c.anb"},
	{0xCE69B01CC67A7EA6ULL, "hlac_dr_slavers_3217_1x.dds"},
	{0xCE69CF3FD988346BULL, "6055005_m.fxe"},
	{0xCE6A04C9D3DF4284ULL, "6043104_m.fxe"},
	{0xCE6A4CF95A7D0F30ULL, "vxht_keep_67j_0_0.vlm"},
	{0xCE6A63D4B2CFECDFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainopen.msh"},
	{0xCE6A69D4B2CFF681ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainopen.mmh"},
	{0xCE6A75D4B2D00B22ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainopen.mao"},
	{0xCE6A8EAE9C849EFAULL, "391732649.ogg"},
	{0xCE6B35AE35C8AE63ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\run.ani"},
	{0xCE6B4EE7B9D77D45ULL, "793156103.ogg"},
	{0xCE6B63A3504084E7ULL, "1071870092.ogg"},
	{0xCE6B6A91FDEA2D48ULL, "primitives_primitive_p2.psh.fxo.dx11.assembly"},
	{0xCE6B74B30F57BA80ULL, "797567014.ogg"},
	{0xCE6BC7DC5A5B351BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_shush.ani"},
	{0xCE6BF1FAC7AF60E8ULL, "6100075_f.fxe"},
	{0xCE6C79780B3F5716ULL, "vxlt_undercitydungeon_1ok_0_0.vlm"},
	{0xCE6C80C7660D641FULL, "6164961_m.fxe"},
	{0xCE6C8F3310CFA30DULL, "lgt110_gxa_vgk_silver_order6117033.cl"},
	{0xCE6CAE8D722AC280ULL, "vxht_keep_636_0_0.vlm"},
	{0xCE6CCC2782BC40FAULL, "pro000_athenril6049460.cl"},
	{0xCE6CD418CDA8EF9DULL, "6103689_m.fxe"},
	{0xCE6D27EF521C2159ULL, "pro000_merchant6051571.cl"},
	{0xCE6D6996A2571C37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_stun.gad"},
	{0xCE6DA7D0863AB780ULL, "6107532_m.fxe"},
	{0xCE6DBB980F7C0B43ULL, "qun211_party_comment.bnk"},
	{0xCE6DDD06B22A81C3ULL, "hlac_ht_keep_n_3734_0z.dds"},
	{0xCE6E5BFC8277D612ULL, "vxht_mansion_n_628_0_0.vlm"},
	{0xCE6E749157154AC0ULL, "574236344.ogg"},
	{0xCE6EC7F00C069C90ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredidies.gad"},
	{0xCE6F08F5A3D258EFULL, "6173675_m.fxe"},
	{0xCE6F2E2CD95641C7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3pp_large.phy"},
	{0xCE6F9BB50597F3B8ULL, "271583170.ogg"},
	{0xCE6FC0F6C32987B9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grainsack_d.dds"},
	{0xCE6FD48D2E85B7A8ULL, "vxlt_warehouse_ov_0_0.vlm"},
	{0xCE6FDDA979762B06ULL, "w_blightlands_123_0.phy"},
	{0xCE6FE983F16A39ADULL, "6088288_f.fxe"},
	{0xCE6FF0A9B81E5D0EULL, "dae000_bartrand6031384.cl"},
	{0xCE70D5B432BED724ULL, "mag291_vincento6027339.cl"},
	{0xCE70DB83839D5811ULL, "104573100.ogg"},
	{0xCE719D980F7FC173ULL, "qun211_party_comment.cnv"},
	{0xCE71E1EF43FB82F7ULL, "llac_dr_dungeon_3216_9.dds"},
	{0xCE72103939F12026ULL, "vxht_0_hightown_a3_9u4_0_0.vlm"},
	{0xCE722574D1D8570AULL, "and230_mage_girl6109095.cl"},
	{0xCE7277858443B559ULL, "fx_grass.dds"},
	{0xCE72D1794C9EEFACULL, "6119297_f.fxe"},
	{0xCE72E636549D14C4ULL, "art\\characters\\creatures\\mubari\\animation\\com_sniff.ani"},
	{0xCE732A6D19985251ULL, "gen00fl_isabela6085570.cl"},
	{0xCE7342636C92082EULL, "vxht_keep_n_4nf_0_0.vlm"},
	{0xCE7369C7BAA52318ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_sf01a_0t.dds"},
	{0xCE737E96A25BABC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_stun.evt"},
	{0xCE7380B22926339AULL, "vxht_0_hightown_a3_car_0_0.vlm"},
	{0xCE738D73E09D9802ULL, "vxw_bonepitdestroyed_a3_2aa_0_0.vlm"},
	{0xCE743198B976E504ULL, "and100_anders6095837.cl"},
	{0xCE7435D87F7FC23AULL, "vxdr_0_deeproads_ck6_0_0.vlm"},
	{0xCE743C521461339CULL, "405275245.ogg"},
	{0xCE74CA74EBEB27DFULL, "gen00fl_dog6167169.cl"},
	{0xCE74FEFB645AC957ULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm02.phy"},
	{0xCE76F4E19C9CB79FULL, "w_bonepitmining_a2_27_0.msh"},
	{0xCE76FAE19C9CC241ULL, "w_bonepitmining_a2_27_0.mmh"},
	{0xCE7706E19C9CD6E2ULL, "w_bonepitmining_a2_27_0.mao"},
	{0xCE7743E92AECC698ULL, "vxca_caverns2_7kx_0_0.vlm"},
	{0xCE77D93583BFD52EULL, "gen00fl_merrill6109631.cl"},
	{0xCE77F0A0C202BE6DULL, "vxdr_0_deeproads_egc_0_0.vlm"},
	{0xCE783C26D1854B1AULL, "502287499.ogg"},
	{0xCE78411CCB7E4CEBULL, "6052105_m.fxe"},
	{0xCE784DAE35D3615AULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\run.evt"},
	{0xCE78672420DFD39CULL, "594119340.ogg"},
	{0xCE789B37C083DCE3ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingc_n.dds"},
	{0xCE789F5ECF22CE27ULL, "vxht_mansion_n_6n4_0_1.vlm"},
	{0xCE78BED64C7327B0ULL, "vxht_keep_60k_0_0.vlm"},
	{0xCE78CFDC5A65CD12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_shush.evt"},
	{0xCE78EF54A92B19EAULL, "6207382_m.fxe"},
	{0xCE790686D269A0A5ULL, "98383084.ogg"},
	{0xCE795123F9E741AEULL, "vxw_bonepitdestroyed_a3_1zo_0_0.vlm"},
	{0xCE79B6484B7EB26BULL, "character_character_p170.psh.fxo.dx11.assembly"},
	{0xCE79E33654A32E26ULL, "art\\characters\\creatures\\mubari\\animation\\com_sniff.gad"},
	{0xCE7A1837D1D8C758ULL, "vxlt_gallowstemplar_f_5t4_0_0.vlm"},
	{0xCE7ABA6D64C522AEULL, "vxca_cavern1_haunted_i6i_0_0.vlm"},
	{0xCE7AC284D468BF05ULL, "vxdr_primevaltomb_8d9_0_0.vlm"},
	{0xCE7B1857DACACF3FULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.mmh"},
	{0xCE7B27F591CAD516ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal.msh"},
	{0xCE7B31F591CAE614ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal.mmh"},
	{0xCE7B35F591CAED27ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal.mao"},
	{0xCE7B989B5104CF06ULL, "one200dg_solivitus6060354.cl"},
	{0xCE7BBCB0477179A8ULL, "sf_xy16icf32.vsh.fxo"},
	{0xCE7BE196D410B62CULL, "vxw_blightlands_gz_0_0.vlm"},
	{0xCE7C0CF57BD984C9ULL, "889556440.ogg"},
	{0xCE7C1DD60FA88871ULL, "lgt260_eau_tmp_conrad.bnk"},
	{0xCE7C3BB8F1F463CBULL, "vxca_cavern1_haunted_gi3_0_0.vlm"},
	{0xCE7C97EFDC2C58F4ULL, "6139850_m.fxe"},
	{0xCE7C9A601FA8582FULL, "llac_lt_backalley_g_0_5.dds"},
	{0xCE7CEBD06890E479ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flma_0d.dds"},
	{0xCE7DB54DF4A42FA0ULL, "dae000cr_grand_cleric6147700.cl"},
	{0xCE7DB5DEE9B2BC9DULL, "zz_dae_debug6029344.cl"},
	{0xCE7DB8774EA64918ULL, "6119621_m.fxe"},
	{0xCE7DF1CDF1757B31ULL, "6083398_m.fxe"},
	{0xCE7E466A207BD028ULL, "vxlt_warehouse_pw_0_0.vlm"},
	{0xCE7EAFFEA0A9C465ULL, "647510679.ogg"},
	{0xCE7EFFD60FAA8A81ULL, "lgt260_eau_tmp_conrad.cnv"},
	{0xCE7F06C24ED70D3AULL, "vxht_chantry_2c4_0_0.vlm"},
	{0xCE7F4CAE35D97E4DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\run.gad"},
	{0xCE7F60BD1E684699ULL, "dae000ip_crowd_ff5_talk12.utp"},
	{0xCE7F65E03081651DULL, "vxlt_hangedman_n_1th_0_0.vlm"},
	{0xCE7F829DD437DB5AULL, "594383707.ogg"},
	{0xCE7FEEDC5A6C20A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_shush.gad"},
	{0xCE804ED37F09442FULL, "6000883_m.fxe"},
	{0xCE80DE3654A94405ULL, "art\\characters\\creatures\\mubari\\animation\\com_sniff.evt"},
	{0xCE80E90F524A4941ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_act_c.anb"},
	{0xCE812FFEBD19A71AULL, "793274093.ogg"},
	{0xCE819696A26811BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_stun.ani"},
	{0xCE81B75D85A575BBULL, "plt_cod_hst_mag_glitterdust.plo"},
	{0xCE82C55D6EA751FFULL, "6153196_m.fxe"},
	{0xCE83CFA2DF0C2E48ULL, "6075743_m.fxe"},
	{0xCE8429EA949894C5ULL, "mcr370_assignments6176743.cl"},
	{0xCE843D5905DD8E24ULL, "ser221_gamlen6133137.cl"},
	{0xCE847D0C77E5F64DULL, "hlac_lt_0_lowtown_a2_n_c_-2_3y.dds"},
	{0xCE84A3EB020138D7ULL, "844260090.ogg"},
	{0xCE84B04B5F837C32ULL, "dae000_bodhan6129157.cl"},
	{0xCE84FCE844899A7EULL, "6110609_f.fxe"},
	{0xCE85803E206AD716ULL, "241022478.ogg"},
	{0xCE85D11A9ED30840ULL, "dae300_cullen6135349.cl"},
	{0xCE86AAE996113CE2ULL, "vxlt_undercityserialkiller_1ap_0_0.vlm"},
	{0xCE86DEB2FD06D112ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_dsaxa_0t.dds"},
	{0xCE86E618B2C64398ULL, "436324888.ogg"},
	{0xCE87BCC71339C39AULL, "33228235.ogg"},
	{0xCE8813E098582A5CULL, "hlac_lt_0_lowtown_a2_-65535_2x.dds"},
	{0xCE8828AF26632740ULL, "hlac_w_sundermount_-262141_0x.dds"},
	{0xCE88394AC56442CCULL, "narrative_varric6180096.cl"},
	{0xCE886D200BA60032ULL, "vxw_blightlands_en_0_0.vlm"},
	{0xCE895AB7406BA9B2ULL, "rmtst03_probe_3895.mtx"},
	{0xCE89F7656C0152DCULL, "llac_ht_0_hightown_a3_131070_0.dds"},
	{0xCE8A5FA7F3AD5DC2ULL, "mag360_thrask6053534.cl"},
	{0xCE8AC95A699284C9ULL, "vxht_0_hightown_a3_bg2_0_0.vlm"},
	{0xCE8ACB36889B759AULL, "6208638_f.fxe"},
	{0xCE8B47DE2E1DA136ULL, "qun110_petrice6108848.cl"},
	{0xCE8B5AF5837C8D68ULL, "vxht_mansion2_2v1_0_0.vlm"},
	{0xCE8BB14EDA49A957ULL, "6036451_m.fxe"},
	{0xCE8BC57E1FDE75A0ULL, "aip_ragedemon_base.gda"},
	{0xCE8BFD5242D34488ULL, "mag160_varian_ilithis6018882.cl"},
	{0xCE8C17ACB83739FFULL, "617351734.ogg"},
	{0xCE8C2C25BAC931CCULL, "43273948.ogg"},
	{0xCE8D39B347AFC880ULL, "6098306_m.fxe"},
	{0xCE8DC61E02208CDCULL, "art\\vfx\\combat\\vfx_r_gn1_trip_act_ext_c.anb"},
	{0xCE8DD7BD22777D9DULL, "vxlt_hangedman_n_1qg_0_0.vlm"},
	{0xCE8DEECB452C0F45ULL, "vxca_caverns1_fo7_0_0.vlm"},
	{0xCE8E02F4E86D0590ULL, "244982748.ogg"},
	{0xCE8ECA0F525651BAULL, "art\\vfx\\combat\\vfx_ha_bd_slam_act_c.mmh"},
	{0xCE8ED94ABAA80D92ULL, "6093200_m.fxe"},
	{0xCE8F0082615CD53CULL, "526281031.ogg"},
	{0xCE8F416DEAB4352BULL, "vxht_0_hightown_n_awq_0_0.vlm"},
	{0xCE8F4655A211524DULL, "hlac_w_sundermount_-262149_0y.dds"},
	{0xCE8F5D73A0541B16ULL, "839815134.ogg"},
	{0xCE8F9E7F3595044EULL, "vxlt_gallowstemplar_f_5ut_0_1.vlm"},
	{0xCE8FA43A03EA9F2AULL, "mer210_coterie_contact6036757.cl"},
	{0xCE8FAAE9DD284FE4ULL, "6025148_m.fxe"},
	{0xCE908562DC33FEA9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mase_0d.dds"},
	{0xCE90F8AF3D83539CULL, "446758370.ogg"},
	{0xCE9183FD7FA54E8EULL, "18471740.ogg"},
	{0xCE9199C0A39ACFBEULL, "docks_male_elven_worker_sta6175824.cl"},
	{0xCE923DBD61678101ULL, "vxlt_0_lowtown_docks_a3_n_d1n_0_0.vlm"},
	{0xCE9248E2668E9078ULL, "vxw_sundermount_2zh_0_1.vlm"},
	{0xCE924CDBF7E769F8ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_act_c.mmh"},
	{0xCE92D825D5B42A6DULL, "hlac_lt_0_lowtown_a3_n_c_196606_0z.dds"},
	{0xCE939691ED591A3FULL, "356475243.ogg"},
	{0xCE93D7707F9B0376ULL, "llac_ht_chantry_2534_30.dds"},
	{0xCE93E78EA19A7D77ULL, "496278807.ogg"},
	{0xCE941F0F99F54756ULL, "dr_primevaltomb_661_lc.mmh"},
	{0xCE94912FAF435333ULL, "hlac_w_bonepitdestroyed_a3_458752_0z.dds"},
	{0xCE94B6C00F385C8CULL, "mag110_vincento6017811.cl"},
	{0xCE94CD28718A3A7AULL, "sse_ss_circlemages_m_2.gda"},
	{0xCE95512E5D725333ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf_facem1.mao"},
	{0xCE966AD4B2F533FEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_grainopen.phy"},
	{0xCE96C4CD8DED1CF8ULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunarifort_n.dds"},
	{0xCE96DC2B6F158CE5ULL, "vxlt_gallowsprison_n_6my_0_0.vlm"},
	{0xCE96DEFCFD9C18B2ULL, "vxw_blightlands_no_0_0.vlm"},
	{0xCE9761D915A68637ULL, "vxht_mansion_n_5aw_0_0.vlm"},
	{0xCE97A9DF88654D23ULL, "6125478_m.fxe"},
	{0xCE98C550B7FCE72CULL, "vxw_blightlands_mx_0_0.vlm"},
	{0xCE993778FD94A22AULL, "344916791.ogg"},
	{0xCE995FECE272C99FULL, "400491009.ogg"},
	{0xCE998CE02DB0B1EFULL, "gen00fl_aveline6137711.cl"},
	{0xCE99BE41346627E5ULL, "316745012.ogg"},
	{0xCE9A5AFF28BF72B9ULL, "783623123.ogg"},
	{0xCE9A9C3132927A18ULL, "6113021_m.fxe"},
	{0xCE9B69C5A0BCE957ULL, "gen_im_npc_arm_hum_boo_krka.uti"},
	{0xCE9B9FDB18E58C55ULL, "zz_fol_debug6098097.cl"},
	{0xCE9BEA7C32C33E3AULL, "vxht_chantry_2y6_0_0.vlm"},
	{0xCE9CDAE19CBD5B6AULL, "w_bonepitmining_a2_27_0.rml"},
	{0xCE9F812A2315BC7CULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_2_ian.gad"},
	{0xCE9FA800E937FBB7ULL, "zz_slate1_baddy.cut"},
	{0xCEA055C58819C251ULL, "hlac_ht_0_hightown_n_2_1x.dds"},
	{0xCEA07CAC4117D501ULL, "hlac_ht_chantry_2534_31z.dds"},
	{0xCEA0F57E621FB5DFULL, "152062704.ogg"},
	{0xCEA181F2A050530AULL, "art\\vfx\\combat\\vfx_c_rw_bladestab_c.anb"},
	{0xCEA182B0C3C9E08CULL, "seb321_leliana6197862.cl"},
	{0xCEA1B0C809543081ULL, "6129999_m.fxe"},
	{0xCEA1B85C135A6D7DULL, "6177708_f.fxe"},
	{0xCEA248DD35C37699ULL, "214487801.ogg"},
	{0xCEA2DB481F731EC2ULL, "hlac_lt_0_lowtown_docks_0_1y.dds"},
	{0xCEA319FE928E502CULL, "547831099.ogg"},
	{0xCEA321287E1A2408ULL, "hlac_lt_0_lowtown_a3_n_c_-196608_3y.dds"},
	{0xCEA3433A43E22601ULL, "mag211_marethari6025344.cl"},
	{0xCEA34D3EFB7C4118ULL, "608077888.ogg"},
	{0xCEA36AC220F8E4E6ULL, "qcr000_aveline6054374.cl"},
	{0xCEA39BBE3AB27901ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_bdy_snga_0i.dds"},
	{0xCEA3C1677490328DULL, "hlac_lt_0_lowtown_a3_n_c_-262144_1y.dds"},
	{0xCEA41957DAEDD67CULL, "art\\vfx\\creatures\\fxc_shale_fire_l_c.anb"},
	{0xCEA4AD3970798CF5ULL, "mag291_vincento6027691.cl"},
	{0xCEA4BCF3C3A21473ULL, "vxdr_primevaltomb_1mu_0_0.vlm"},
	{0xCEA50D0FF07EB3B3ULL, "6078123_m.fxe"},
	{0xCEA52D915DE72297ULL, "hlac_ca_caverns1_mines_24426_6y.dds"},
	{0xCEA550D9EF923132ULL, "vxw_blightlands_sh_0_0.vlm"},
	{0xCEA5784D5AE71AF2ULL, "vxw_encounter1_fi_0_0.vlm"},
	{0xCEA5992CC55588E1ULL, "character_character_p35.psh.fxo.dx11.assembly"},
	{0xCEA5A28A77DA61D2ULL, "6039019_m.fxe"},
	{0xCEA5A2E84B4BF978ULL, "246363820.ogg"},
	{0xCEA5E7BAF38426F5ULL, "702522447.ogg"},
	{0xCEA64A4C76ADF0D9ULL, "gen00fl_dog6146957.cl"},
	{0xCEA6802A231BD97AULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_2_ian.ani"},
	{0xCEA697CD644AF941ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_imp_aoe_p.anb"},
	{0xCEA6E3D98CECC30BULL, "6070181_m.fxe"},
	{0xCEA7372DA9F2FFACULL, "vxw_blightlands_vy_0_0.vlm"},
	{0xCEA82D8ED2169A0AULL, "dae000_mother6028514.cl"},
	{0xCEA90630F48492C0ULL, "art\\vfx\\creatures\\vfx_c_drag_spit_c.mmh"},
	{0xCEA9211E6C9BD9C4ULL, "934937571.ogg"},
	{0xCEA9418E4A6BCF33ULL, "vxlt_gallowscourtyard_4lg_0_0.vlm"},
	{0xCEAA0E19789DE300ULL, "darktown_fem_elf_sta.cnv"},
	{0xCEAA43478BC635A3ULL, "526717436.ogg"},
	{0xCEAA5C5924B956BAULL, "vxht_chantry_2r7_0_0.vlm"},
	{0xCEAB80BCE65D9268ULL, "zz_fol_debug6097811.cl"},
	{0xCEABC64F7F628D74ULL, "1001597061.ogg"},
	{0xCEABEF79D20A1515ULL, "6140298_f.fxe"},
	{0xCEAC3FBF615C689CULL, "vxdr_0_deeproads_f57_0_0.vlm"},
	{0xCEAC482379FEF436ULL, "vxlt_gallowsprison_n_6cp_0_0.vlm"},
	{0xCEAC4AC22459F308ULL, "w_sundermount_124_0.mmh"},
	{0xCEAC56C2245A07ABULL, "w_sundermount_124_0.mao"},
	{0xCEAC68C2245A2632ULL, "w_sundermount_124_0.msh"},
	{0xCEACBCC63CEF3480ULL, "794708247.ogg"},
	{0xCEACFE3365A2234BULL, "nat380_sibling6068006.cl"},
	{0xCEAD8C9E1045B76AULL, "vxht_mansion2b_1ma_0_0.vlm"},
	{0xCEAD9CD8D8E47FCAULL, "vxht_mansion2_1ua_0_0.vlm"},
	{0xCEADF01978A19938ULL, "darktown_fem_elf_sta.bnk"},
	{0xCEAE2AA65DC52FB7ULL, "art\\vfx\\textures\\vfx_lightning_arc.dds"},
	{0xCEAE5EAFE220DEB3ULL, "724414964.ogg"},
	{0xCEAEE095699AD4A4ULL, "hlac_lt_0_lowtown_n_1_9z.dds"},
	{0xCEAEF69496665E2FULL, "682310205.ogg"},
	{0xCEAF0F2C3D9A0286ULL, "6101024_m.fxe"},
	{0xCEAF6532D8261656ULL, "vxdr_0_deeproads_f0f_0_0.vlm"},
	{0xCEAFB6E52E44A356ULL, "dae000_gen_turn_in_36198690.cl"},
	{0xCEAFC64F6723021DULL, "vxht_mansion_n_3qw_0_0.vlm"},
	{0xCEAFCAC6731CCA3FULL, "6048502_m.fxe"},
	{0xCEAFDFD2D44FA84CULL, "6082588_f.fxe"},
	{0xCEAFE27F207AA860ULL, "vxw_woundedcoast_fw_0_0.vlm"},
	{0xCEB00D1D67D06DD6ULL, "320148527.ogg"},
	{0xCEB07E67DF680C91ULL, "6135889_f.fxe"},
	{0xCEB0D5F1A9A44475ULL, "hlac_ht_0_hightown_a2_-131073_0z.dds"},
	{0xCEB10B76041BFA4CULL, "vxht_mansion2_21p_0_0.vlm"},
	{0xCEB1C5843880D4AFULL, "llac_ca_caverns2_5802_26.dds"},
	{0xCEB1D57DF7C5533AULL, "cod_hst_nevarra.utp"},
	{0xCEB1FEAC9854F7C8ULL, "6035318_m.fxe"},
	{0xCEB308ACFA224614ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_c.anb"},
	{0xCEB317CF88628CFCULL, "lgt310_shw_ors_informant6156349.cl"},
	{0xCEB31897C1BFA3AEULL, "vxw_bonepitdestroyed_a3_1nc_0_0.vlm"},
	{0xCEB33F91072F080DULL, "6096503_m.fxe"},
	{0xCEB357282B47AC95ULL, "604384558.ogg"},
	{0xCEB38C2A221255A6ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.dura.ani"},
	{0xCEB3C2B6E18849B2ULL, "vxw_blightlands_ti_0_0.vlm"},
	{0xCEB3DFDD5372F7F9ULL, "264281923.ogg"},
	{0xCEB3EA2A4CDD3372ULL, "vxw_encounter1_kj_0_0.vlm"},
	{0xCEB478CD645701BAULL, "art\\vfx\\combat\\vfx_r_shd_decoy_imp_aoe_p.mmh"},
	{0xCEB47B51995061BCULL, "357964385.ogg"},
	{0xCEB47CDDC80DB9D0ULL, "hlac_lt_backalley_g_65536_0y.dds"},
	{0xCEB4C4A1FD8B6137ULL, "601744467.ogg"},
	{0xCEB4E4F6F3CA09B2ULL, "vxw_sundermount_2gt_0_0.vlm"},
	{0xCEB555B67EE2DB8BULL, "6075281_m.fxe"},
	{0xCEB558E196868E6EULL, "hlac_ca_caverns1_24423_9y.dds"},
	{0xCEB569B4B854A790ULL, "w_encounter3_150_0.mao"},
	{0xCEB56DB4B854AEA3ULL, "w_encounter3_150_0.mmh"},
	{0xCEB57BB4B854C6FDULL, "w_encounter3_150_0.msh"},
	{0xCEB5BFB61C188574ULL, "21651917.ogg"},
	{0xCEB64B6D89958892ULL, "6026993_f.fxe"},
	{0xCEB6734B0CC457C1ULL, "hlac_w_woundedcoast_262150_0x.dds"},
	{0xCEB6D5C03BA8C2B3ULL, "gen00fl_anders6128338.cl"},
	{0xCEB6E2AC024A13F8ULL, "vxht_0_hightown_n_c5r_0_0.vlm"},
	{0xCEB6E930F4909EDFULL, "art\\vfx\\creatures\\vfx_c_drag_spit_c.anb"},
	{0xCEB72C6F16BA9245ULL, "91294323.ogg"},
	{0xCEB868CDE4C28395ULL, "vxht_mansion_n_4f5_0_0.vlm"},
	{0xCEB87121039D0273ULL, "lgt220_shp_mag_guard6175906.cl"},
	{0xCEB8E3D12BC9EAC5ULL, "149999978.ogg"},
	{0xCEB943227AD0FEC0ULL, "qun120cr_javaris6064586.cl"},
	{0xCEB95144D8DF1741ULL, "513861953.ogg"},
	{0xCEBA45566B1E3D03ULL, "vxht_bartrand_n_5rm_0_0.vlm"},
	{0xCEBB0B4F335395CAULL, "6208353_f.fxe"},
	{0xCEBB665BDE54792DULL, "vxca_cave_2ia_0_0.vlm"},
	{0xCEBB7F54F5CC7F6BULL, "cd_fen110_fenris.wbd"},
	{0xCEBB99375068FA06ULL, "122556911.ogg"},
	{0xCEBC0EB5CADA984AULL, "vxht_mansion2_1xb_0_0.vlm"},
	{0xCEBC5DB6D97BA271ULL, "follower_banter6155490.cl"},
	{0xCEBCDFFB64984A4CULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm02.mmh"},
	{0xCEBCE5FB6498543EULL, "art\\characters\\playercharacter\\humanmale\\mt_aet_hm02.msh"},
	{0xCEBD351BB41A51BBULL, "30531955.ogg"},
	{0xCEBDE1B35DF51916ULL, "817522067.ogg"},
	{0xCEBE3CA36512E2BFULL, "6043402_m.fxe"},
	{0xCEBE51AFC645C0CCULL, "6085488_f.fxe"},
	{0xCEBE545C1270C0E0ULL, "vxw_woundedcoast_ct_0_0.vlm"},
	{0xCEBE7615E35189E9ULL, "qun160_vashothlead6053884.cl"},
	{0xCEBE9EF0923E54F0ULL, "6149383_m.fxe"},
	{0xCEBED16FB720E1AEULL, "623085936.ogg"},
	{0xCEBF18EF9D849D65ULL, "ach_soc_gift.dds"},
	{0xCEBF98B87745B08DULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunk.msh"},
	{0xCEBFA6B87745C8C0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunk.mao"},
	{0xCEBFAAB87745CF13ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunk.mmh"},
	{0xCEBFE691AAB6E88BULL, "llac_lt_0_lowtown_docks_131070_3.dds"},
	{0xCEC03CCEA1582322ULL, "49002216.ogg"},
	{0xCEC0599A73DC2BECULL, "6032288_f.fxe"},
	{0xCEC05A07AEE8140AULL, "character_rendertodepth_p2.vsh.fxo.dx11.assembly"},
	{0xCEC0742A221CB717ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.dura.evt"},
	{0xCEC07DAAABE2C09CULL, "hlac_lt_gamlenshouse_n_-1_8z.dds"},
	{0xCEC08676B5071D71ULL, "457650024.ogg"},
	{0xCEC09E318FF008C0ULL, "362741373.ogg"},
	{0xCEC0DF3A6A8F9D2FULL, "283090249.ogg"},
	{0xCEC0DF72C581BD10ULL, "6119643_m.fxe"},
	{0xCEC0E7ACFA2E4BD7ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_c.mmh"},
	{0xCEC17E413F0F5EB6ULL, "6013392_m.fxe"},
	{0xCEC191485048872DULL, "mer110cr_miner1.utc"},
	{0xCEC1A031AA5400F3ULL, "vxht_0_hightown_a2_ccm_0_1.vlm"},
	{0xCEC261F662B0E612ULL, "6126409_m.fxe"},
	{0xCEC2C2A4115D458EULL, "isa311_isabela6092153.cl"},
	{0xCEC311F4B5A13162ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_std_beg_lp.gad"},
	{0xCEC356D3E5C02232ULL, "vxw_sundermount_2hu_0_0.vlm"},
	{0xCEC3777203B168EAULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_stn_0d.dds"},
	{0xCEC419D56AE034E9ULL, "genlock_sbow_dagger_shield.utc"},
	{0xCEC4D22CC2E8EBA2ULL, "6167598_m.fxe"},
	{0xCEC525122A2D6D3EULL, "vxlt_0_lowtown_docks_a3_cvx_0_0.vlm"},
	{0xCEC55488F4402C78ULL, "vxht_0_hightown_n_c2s_0_0.vlm"},
	{0xCEC5B1C4F0216D76ULL, "llac_lt_0_lowtown_a2_65535_7.dds"},
	{0xCEC60A63C34A7860ULL, "245445840.ogg"},
	{0xCEC625482E580033ULL, "vxlt_gallowscourtyard_4bi_0_0.vlm"},
	{0xCEC765454329A85EULL, "vxca_caverns2prc_aeo_0_0.vlm"},
	{0xCEC7752A2222D7A0ULL, "art\\vfx\\combat\\vfx_d_teleport_act_c.dura.gad"},
	{0xCEC79DB0CD02EE2AULL, "gen00fl_anders6110064.cl"},
	{0xCEC812D9D1337B18ULL, "hlac_lt_ruinedhouse_n_-1_3y.dds"},
	{0xCEC8342043DF7CA4ULL, "vxca_caverns2prc_a0z_0_0.vlm"},
	{0xCEC8622C5DB662F9ULL, "vxca_caverns2_97s_0_0.vlm"},
	{0xCEC89BC8175AE4CEULL, "hlac_lt_gallowscourtyard_n_65539_0y.dds"},
	{0xCEC8A90213B04348ULL, "vrc221_bartrand6043136.cl"},
	{0xCEC8B7335D145583ULL, "vxht_bartrand_n_5ul_0_0.vlm"},
	{0xCEC934F4B5A5D859ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_std_beg_lp.evt"},
	{0xCEC984F2A071AAFDULL, "art\\vfx\\combat\\vfx_c_rw_bladestab_c.mmh"},
	{0xCECA2BB3DF1D2722ULL, "llac_dr_slavers_3214_4.dds"},
	{0xCECA8092BCD0B0CAULL, "vxht_mansion2_1cc_0_0.vlm"},
	{0xCECAD94DDC10EB98ULL, "lt_warehouse_4_0_ll.mmh"},
	{0xCECCC6390466D960ULL, "vxw_woundedcoast_xu_0_0.vlm"},
	{0xCECDE9B57F793266ULL, "6125848_m.fxe"},
	{0xCECE10B6118A80D5ULL, "hlac_lt_foundry_n_320_22x.dds"},
	{0xCECE9FA6DBB949FDULL, "hlac_w_encounter2_camp_1_0z.dds"},
	{0xCECEC86A141D2314ULL, "6040871_m.fxe"},
	{0xCECF205F879F0EDEULL, "vxlt_0_lowtown_a2_n_c_hnt_0_0.vlm"},
	{0xCECF5BCA159D2B59ULL, "vxdr_0_deeproads_8wc_0_1.vlm"},
	{0xCECF93552801A739ULL, "zz_area_signage4.utp"},
	{0xCECFFFA2112682B0ULL, "hlac_ht_bartrand_n_3373_2y.dds"},
	{0xCED0516CE34571F1ULL, "900622068.ogg"},
	{0xCED065A6BB2025FEULL, "6132813_m.fxe"},
	{0xCED0AD6F2351F790ULL, "797536085.ogg"},
	{0xCED13C5F940FFFDBULL, "vxlt_warehouse_n_dy_0_0.vlm"},
	{0xCED165D796AB473EULL, "dae200_saemus6142789.cl"},
	{0xCED19B904E1EE865ULL, "gen_im_trs_chn_dua_dag_r2.uti"},
	{0xCED1B2D863153F1FULL, "6143085_m.fxe"},
	{0xCED1C8B0D7B63AB2ULL, "vxw_sundermount_2mv_0_0.vlm"},
	{0xCED260170CE1D0C1ULL, "hlac_lt_gallowscourtyard_n_-131072_3x.dds"},
	{0xCED2620BA179F5F3ULL, "6110430_m.fxe"},
	{0xCED3263F8D7A6BA3ULL, "qcr220_orsino6061803.cl"},
	{0xCED34A9FEC3E3CEAULL, "6160810_m.fxe"},
	{0xCED35193B7114165ULL, "6179983_m.fxe"},
	{0xCED36FB15AF0865CULL, "vxht_keep_63j_0_0.vlm"},
	{0xCED396EF1C2385BEULL, "vxlt_0_lowtown_docks_a3_cyy_0_0.vlm"},
	{0xCED39A29733FE71FULL, "docks_male_human_worker_sta6175967.cl"},
	{0xCED39E6BE7131770ULL, "367435802.ogg"},
	{0xCED42529C135A5ECULL, "6149035_m.fxe"},
	{0xCED46378FC06A134ULL, "gen00fl_aveline6152535.cl"},
	{0xCED57021249ACDFAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_basic.mmh"},
	{0xCED57421249AD431ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_basic.mao"},
	{0xCED58221249AEC40ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_basic.msh"},
	{0xCED5A50B5A01E33EULL, "vxca_caverns2_7u0_0_0.vlm"},
	{0xCED5DD3C9B410354ULL, "6165709_m.fxe"},
	{0xCED60740B02DF635ULL, "vxlt_warehouse_n_rq_0_0.vlm"},
	{0xCED60A89FE682496ULL, "zz_vrc_debug6200208.cl"},
	{0xCED60EA383315D9FULL, "170406948.ogg"},
	{0xCED67BC7EB63583BULL, "243228046.ogg"},
	{0xCED68FB7B6456CB0ULL, "art\\vfx\\environment\\vfx_fadelightbeams.mmh"},
	{0xCED6D4094FAC7B79ULL, "vxca_caverns2_98r_0_0.vlm"},
	{0xCED6EC9D88BBDA95ULL, "gen00fl_sebastian6179779.cl"},
	{0xCED7161A739E3EA3ULL, "vxlt_backalley_g_1bl_0_0.vlm"},
	{0xCED729104F0A6E03ULL, "vxht_bartrand_n_5xo_0_0.vlm"},
	{0xCED730F59219430FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal.phy"},
	{0xCED73CF4B5B22348ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_std_beg_lp.ani"},
	{0xCED74C3F4BED708CULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f_shrug.ani"},
	{0xCED77B7BFB2EAFB4ULL, "gen00fl_anders6107664.cl"},
	{0xCED78BDF59A967EEULL, "5679687.ogg"},
	{0xCED7DBEB585B8203ULL, "6103085_m.fxe"},
	{0xCED80A016C0F534BULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_aveline_victim.ani"},
	{0xCED81C15C3BCDF31ULL, "148219413.ogg"},
	{0xCED83A8C1B064D23ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbr.evt"},
	{0xCED8855382B6EA0DULL, "6074378_m.fxe"},
	{0xCED91D7FA8D71B45ULL, "vxlt_0_lowtown_a2_n_c_bxg_0_0.vlm"},
	{0xCED944BECA0210B2ULL, "art\\characters\\creatures\\strider\\animation\\turnleft90.gad"},
	{0xCED9623B08C1902CULL, "162480113.ogg"},
	{0xCED9A027A8C0D7CAULL, "hlac_ht_brothel_n_3374_1y.dds"},
	{0xCEDA609DBD98F047ULL, "6208430_m.fxe"},
	{0xCEDA7229741327D6ULL, "6070107_f.fxe"},
	{0xCEDA7DD468637DD0ULL, "448776251.ogg"},
	{0xCEDB3815F65CF1E0ULL, "vxw_woundedcoast_ur_0_0.vlm"},
	{0xCEDC3E9118251E6EULL, "hlac_lt_gallowstemplar_2384_4y.dds"},
	{0xCEDCA7670FFF40D2ULL, "ts_apostates_acc_props.gda"},
	{0xCEDCC0CC92323F98ULL, "168840447.ogg"},
	{0xCEDD4F8EA5E2F7A8ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he7a.msh"},
	{0xCEDD5A38068B56F1ULL, "gen00fl_varric6072965.cl"},
	{0xCEDD6D8EA5E32AD2ULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he7a.mmh"},
	{0xCEDD7517ED37066BULL, "vxht_0_hightown_n_9wy_0_0.vlm"},
	{0xCEDDA5454EB0F441ULL, "657077636.ogg"},
	{0xCEDDC32CA96DEE10ULL, "6113043_m.fxe"},
	{0xCEDDEC3928795067ULL, "hm_rdr_dockworker.mor"},
	{0xCEDE593F4BF3A5CEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\pose_f_shrug.gad"},
	{0xCEDE62FE500DE24EULL, "528349919.ogg"},
	{0xCEDEA172A21EEAFBULL, "vxca_cave_2gi_0_0.vlm"},
	{0xCEDEE9221E46FFCDULL, "zz_vault_debug6200374.cl"},
	{0xCEDF138C1B0C29A4ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbr.gad"},
	{0xCEE009506481EA2EULL, "vxlt_0_lowtown_docks_b5s_0_0.vlm"},
	{0xCEE047BECA0834E9ULL, "art\\characters\\creatures\\strider\\animation\\turnleft90.evt"},
	{0xCEE07DB89C815DA5ULL, "hlac_lt_0_lowtown_0_2x.dds"},
	{0xCEE16154F5ED1B39ULL, "cd_fen110_fenris.bnk"},
	{0xCEE17FAE3089C83BULL, "239847880.ogg"},
	{0xCEE2078E2773E096ULL, "gen00fl_merrill6127309.cl"},
	{0xCEE22A050E4DBAFCULL, "vxht_chantry_23r_0_0.vlm"},
	{0xCEE26F59CAEEF436ULL, "vxht_mansion_n_4vr_0_0.vlm"},
	{0xCEE27CC991D2EA63ULL, "75262689.ogg"},
	{0xCEE28CDC3E68F434ULL, "vxlt_0_lowtown_a2_n_hf8_0_0.vlm"},
	{0xCEE2B792B0EE746EULL, "gen_im_wep_rog_dua_act3_02.uti"},
	{0xCEE3F2C581D97A78ULL, "vxht_brothel_n_59o_0_0.vlm"},
	{0xCEE4791DA2240EB5ULL, "vxlt_warehouse_n_yp_0_0.vlm"},
	{0xCEE4A6F157066029ULL, "gen00fl_aveline6132839.cl"},
	{0xCEE4CE5CF6B54269ULL, "qun160_vashothlead6064680.cl"},
	{0xCEE4DC705A018854ULL, "w_encounter3_6_0.rml"},
	{0xCEE53A3D7EB1C585ULL, "vxht_0_hightown_n_c5k_0_0.vlm"},
	{0xCEE575F556EA5A79ULL, "6191291_m.fxe"},
	{0xCEE6228C1B126182ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbr.ani"},
	{0xCEE625AC6489E6BDULL, "w_woundedcoast_225_0.mao"},
	{0xCEE629AC6489EDF6ULL, "w_woundedcoast_225_0.mmh"},
	{0xCEE633AC6489FEF4ULL, "w_woundedcoast_225_0.msh"},
	{0xCEE6844C792FB1A4ULL, "follower_banter6169458.cl"},
	{0xCEE71C5B704E5C02ULL, "vxw_bonepitdestroyed_a3_29i_0_0.vlm"},
	{0xCEE798BD9B925359ULL, "6083585_m.fxe"},
	{0xCEE7A467CC48DE8AULL, "gen00fl_aveline6140130.cl"},
	{0xCEE85901BC30318FULL, "gen00fl_varric6081208.cl"},
	{0xCEE8DF38EEA963DCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_enter.ani"},
	{0xCEE926D2035765B5ULL, "vxlt_darktown_4ux_0_0.vlm"},
	{0xCEE9A0F38D56DF78ULL, "vxw_woundedcoast_n7_0_0.vlm"},
	{0xCEE9E71BA3F1BDE5ULL, "vxht_mansion_n_4fe_0_0.vlm"},
	{0xCEEA35DB952562B2ULL, "hlac_ht_mansion2b_818_59x.dds"},
	{0xCEEA706263D95FB5ULL, "hlac_ht_keep_3731_3y.dds"},
	{0xCEEAA16BE85328A0ULL, "228248239.ogg"},
	{0xCEEAA698F6F8936DULL, "87166102.ogg"},
	{0xCEEAF3E6D837B687ULL, "vxlt_0_lowtown_docks_n_b9m_0_0.vlm"},
	{0xCEEAF5256CAE1C50ULL, "art\\levels\\races\\design\\_textures\\prp_bigtreasure_dp.dds"},
	{0xCEEBDE643D325408ULL, "vxlt_blackemporium_ul_0_0.vlm"},
	{0xCEEBF1016C1FD1D5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_aveline_victim.gad"},
	{0xCEEC38E8223F080FULL, "follower_banter6141523.cl"},
	{0xCEEC51103E8D4012ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs05a_0d.dds"},
	{0xCEECA89B03B859F7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\trap_1dwarven_6.pwk"},
	{0xCEECB73743275F49ULL, "191440495.ogg"},
	{0xCEED2FBECA129658ULL, "art\\characters\\creatures\\strider\\animation\\turnleft90.ani"},
	{0xCEED4A242E179149ULL, "130139510.ogg"},
	{0xCEEDE92DA814BCB9ULL, "6150072_m.fxe"},
	{0xCEEEEC38EEADE51EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_enter.gad"},
	{0xCEEF4769CBEA34E6ULL, "6077808_m.fxe"},
	{0xCEEF86FE1F9C1BCFULL, "vxht_0_hightown_n_bwz_0_0.vlm"},
	{0xCEF00123C428BB53ULL, "vxht_0_hightown_a3_bxq_0_0.vlm"},
	{0xCEF0231B014087A4ULL, "6182715_m.fxe"},
	{0xCEF04F1848BD4A74ULL, "vxlt_gallowstemplar_31z_0_0.vlm"},
	{0xCEF07B0C866F93FCULL, "dae200_saemus6091696.cl"},
	{0xCEF0E30B002B7859ULL, "static_static_p22.psh.fxo"},
	{0xCEF1B0787194E67DULL, "vxw_encounter2_iz_0_0.vlm"},
	{0xCEF1B2B0CF12C3F8ULL, "vxht_0_hightown_a2_c7g_0_0.vlm"},
	{0xCEF1DCD97422E3CAULL, "307858021.ogg"},
	{0xCEF20062512CC481ULL, "hlac_lt_0_lowtown_docks_65533_3z.dds"},
	{0xCEF2ACA258CC6B41ULL, "6132876_m.fxe"},
	{0xCEF2F3FD2E1C33D6ULL, "43268502.ogg"},
	{0xCEF4E4492A35F2A0ULL, "6074472_m.fxe"},
	{0xCEF55D344464BB74ULL, "pro000_aveline6047252.cl"},
	{0xCEF58E3862447482ULL, "vxw_bonepitdestroyed_a3_24j_0_0.vlm"},
	{0xCEF5BD02BA7F5010ULL, "zz_lgt_debug6114297.cl"},
	{0xCEF5C738EEB3C53DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_enter.evt"},
	{0xCEF5F8070BC3404DULL, "6139279_m.fxe"},
	{0xCEF634E677AB65D8ULL, "mer320_hubert6053688.cl"},
	{0xCEF638E19639EB0DULL, "l_dae251ar_mount_caverns.bnk"},
	{0xCEF6A8C91C02355AULL, "dre000_dougal6129749.cl"},
	{0xCEF6AB298C7E5631ULL, "6047453_m.fxe"},
	{0xCEF78AC89A5FEB6FULL, "zz_followers_banter6158466.cl"},
	{0xCEF7CA7609FED49CULL, "412692181.ogg"},
	{0xCEF82D6C032216DBULL, "hla_ht_0_hightown_131073_0x.dds"},
	{0xCEF8804B59D7BEE6ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_110a_n.dds"},
	{0xCEF8D4A690554AAEULL, "llac_lt_0_lowtown_1_8.dds"},
	{0xCEF965C3CA2DCF07ULL, "vxlt_0_lowtown_docks_n_b6l_0_0.vlm"},
	{0xCEF981D9857D59D2ULL, "vxht_0_hightown_a3_bxt_0_0.vlm"},
	{0xCEF9BC92722C90B3ULL, "bink_bink_p0.psh.fxo"},
	{0xCEF9C042DC7682D4ULL, "206430362.ogg"},
	{0xCEFA426563689197ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v2_c.anb"},
	{0xCEFA68AB463C58EFULL, "6095948_m.fxe"},
	{0xCEFA8F2124BA1E05ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_basic.phy"},
	{0xCEFB059203074005ULL, "dae000_mother6006245.cl"},
	{0xCEFB1A7DB669B5E1ULL, "vxlt_0_lowtown_a2_n_c_by8_0_0.vlm"},
	{0xCEFB53639FC40C7CULL, "gen00fl_aveline6137497.cl"},
	{0xCEFB626B09975C4DULL, "ss_docks1.cut"},
	{0xCEFBE219B1BAD4A3ULL, "438929408.ogg"},
	{0xCEFC4C7337988B74ULL, "131826760.ogg"},
	{0xCEFC4EB543B81F5FULL, "6028360_m.fxe"},
	{0xCEFCB2C0BAB2D0FCULL, "6189018_f.fxe"},
	{0xCEFCC65BEF9D5FC1ULL, "vxca_cavern1_haunted_jg2_0_0.vlm"},
	{0xCEFCD8E82688F27CULL, "6014675_f.fxe"},
	{0xCEFD99E9943C22DBULL, "6112119_m.fxe"},
	{0xCEFE2229A2F4CEDFULL, "vrc221_hugin6041118.cl"},
	{0xCEFE4BB3D507E910ULL, "mag100_cullen6095112.cl"},
	{0xCEFF328CEB1104DBULL, "6042081_m.fxe"},
	{0xCEFF45E3F8776BF9ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepileb.msh"},
	{0xCEFF4FE3F8777CE7ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepileb.mmh"},
	{0xCEFF53E3F8778334ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepileb.mao"},
	{0xCEFF54E7F22190D9ULL, "static_static_p53.psh.fxo"},
	{0xCEFFB44D25F59D82ULL, "6116473_m.fxe"},
	{0xCF00248DC108DC78ULL, "vxht_0_hightown_a2_c2d_0_0.vlm"},
	{0xCF003BFB5E364CF3ULL, "6071173_m.fxe"},
	{0xCF004639D0F6B723ULL, "dre000_varric6030283.cl"},
	{0xCF00CF55026DDB98ULL, "llac_lt_0_lowtown_a3_n_c_65537_7.dds"},
	{0xCF01BF3E2224A6BCULL, "7469868.ogg"},
	{0xCF01D40F1DBB4BB1ULL, "288058569.ogg"},
	{0xCF02269F6F9191ACULL, "403756878.ogg"},
	{0xCF025D4E267F3F9CULL, "6076162_m.fxe"},
	{0xCF02BDC470FC3556ULL, "dae000_flemeth6001348.cl"},
	{0xCF03BADF44D8C466ULL, "mrl310_merrill6144589.cl"},
	{0xCF03BB6A7FEA9500ULL, "436815175.ogg"},
	{0xCF04C6939AF7E175ULL, "121887763.ogg"},
	{0xCF04FC96B17E8E9DULL, "ts_dwarven_acc_table.gda"},
	{0xCF050C7C7CD361BAULL, "913755664.ogg"},
	{0xCF05AEEB2AEAF40BULL, "6208543_f.fxe"},
	{0xCF05B26B1B408C10ULL, "494602829.ogg"},
	{0xCF05BD7504A7D2E5ULL, "llac_w_sundermount_327682_0.dds"},
	{0xCF05F70625B27329ULL, "282449257.ogg"},
	{0xCF06AB0BA1D63526ULL, "vxlt_hangedman_n_2mw_0_0.vlm"},
	{0xCF07749764822D45ULL, "1067664500.ogg"},
	{0xCF0786A665AA7422ULL, "hlac_lt_0_lowtown_a2_n_c_65537_6x.dds"},
	{0xCF078EDF5B7E10EFULL, "vxht_brothel_n_5k2_0_0.vlm"},
	{0xCF07F3B677737252ULL, "vxht_0_hightown_a3_bwu_0_0.vlm"},
	{0xCF08402E7F15C492ULL, "319780349.ogg"},
	{0xCF0933E0F1C49458ULL, "gen00fl_anders6196734.cl"},
	{0xCF09407A38D6FFBDULL, "hlac_ht_chantry_n_2534_41x.dds"},
	{0xCF0952F15A2714C9ULL, "830874685.ogg"},
	{0xCF09530FF2DDCB6DULL, "6140235_m.fxe"},
	{0xCF09DC43A2C6AC4DULL, "vxlt_0_lowtown_docks_a3_d3b_0_1.vlm"},
	{0xCF0C5138192153A4ULL, "follower_banter6172313.cl"},
	{0xCF0CA4F9C5C29F8CULL, "art\\characters\\heraldry\\hld_red_iron_mercs.dds"},
	{0xCF0CC1BDDF4923A3ULL, "gen00fl_fenris6127030.cl"},
	{0xCF0CCC83C8B763C2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackl_1.mao"},
	{0xCF0CE4DDA814EC53ULL, "vxht_0_hightown_a3_brs_0_0.vlm"},
	{0xCF0D41CC71941532ULL, "one140cr_charade6137037.cl"},
	{0xCF0DD195278F7907ULL, "6106632_f.fxe"},
	{0xCF0DEE5695B4C0F7ULL, "character_character_p192.psh.fxo.dx11.assembly"},
	{0xCF0ECDA237EE0237ULL, "vxw_encounter2_camp_6y_0_0.vlm"},
	{0xCF0ED79AEE44D52BULL, "49244172.ogg"},
	{0xCF0F6299E0A73E27ULL, "40098989.ogg"},
	{0xCF0F7AEA78C22E9BULL, "llac_lt_gallowstemplar_n_2383_6.dds"},
	{0xCF0FBD7C3C164E9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_diepose2.ani"},
	{0xCF0FE06FD36B4986ULL, "one100_solivitus6048118.cl"},
	{0xCF102D93B7A34C16ULL, "72624902.ogg"},
	{0xCF10E9072229F9CBULL, "497829424.ogg"},
	{0xCF11115A256E094BULL, "hlac_ca_cavern1_haunted_24424_9x.dds"},
	{0xCF1129AC08D088A6ULL, "6120835_f.fxe"},
	{0xCF1168179D9DD118ULL, "mcr371_scared_mage_f6087708.cl"},
	{0xCF11EFE9415775B5ULL, "one100dg_vanard6059701.cl"},
	{0xCF126E7605184EE2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wo_basic_1.mao"},
	{0xCF1280EB61C1E22EULL, "one220_yevhen6135849.cl"},
	{0xCF135320B5205104ULL, "stop_combat_mode.ncs"},
	{0xCF1430E634331C26ULL, "one100dg_lia6055164.cl"},
	{0xCF151CE893CC4DA6ULL, "vxlt_hangedman_n_2hx_0_0.vlm"},
	{0xCF15AA9BDA3AF758ULL, "319911219.ogg"},
	{0xCF15CC2B0C6EDB02ULL, "705548120.ogg"},
	{0xCF160D6883F8B719ULL, "640368303.ogg"},
	{0xCF16659369698AD2ULL, "vxht_0_hightown_a3_brv_0_0.vlm"},
	{0xCF1719C4FE45ABF6ULL, "vxca_caverns1_mines_ihc_0_0.vlm"},
	{0xCF17886A75C3ED49ULL, "llac_lt_0_lowtown_a3_n_c_65536_3.dds"},
	{0xCF18C02C7931ED1CULL, "pridedrain-h.cut"},
	{0xCF18D8364B7D9647ULL, "nat380_temmerin6067376.cl"},
	{0xCF195BEC42DC22E8ULL, "hlac_ht_chantry_2534_28y.dds"},
	{0xCF1962A095ADB5BAULL, "lt_undercityserialkiller_mmap_ph.dds"},
	{0xCF199A8FC60E3BE7ULL, "hlac_ht_0_hightown_-196608_1x.dds"},
	{0xCF19E6B38AD9C8FEULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_enter.evt"},
	{0xCF19F5F8B1316126ULL, "781621371.ogg"},
	{0xCF19FF84C777243FULL, "hlac_ht_0_hightown_-65535_1z.dds"},
	{0xCF1A215705DF3605ULL, "411123649.ogg"},
	{0xCF1A24F692D945F2ULL, "6090832_m.fxe"},
	{0xCF1A2CD9003A7BC6ULL, "gen_im_wep_war_one_act1_04.uti"},
	{0xCF1A4DC35CB5A748ULL, "one300cr_crow_4.utc"},
	{0xCF1B38BD45991308ULL, "vxht_mansion2b_269_0_0.vlm"},
	{0xCF1B8DEB765030C7ULL, "lt_0_lowtown_docks_a3_n_39_0_ll.mmh"},
	{0xCF1B9A7C1B8D332FULL, "vxw_bonepitdestroyed_a3_261_0_0.vlm"},
	{0xCF1BA62019A1C8D2ULL, "6016657_f.fxe"},
	{0xCF1C136FCD2EE6E9ULL, "6064528_m.fxe"},
	{0xCF1C5469D67B5286ULL, "gen00fl_aveline6107085.cl"},
	{0xCF1CA9581D19CBDEULL, "vxlt_hangedman_n_20x_0_0.vlm"},
	{0xCF1D33B78DE482A3ULL, "6122729_m.fxe"},
	{0xCF1D3F7F29E41AB7ULL, "vxw_encounter2_camp_1x_0_0.vlm"},
	{0xCF1D5D4926B0E2B0ULL, "dae011_gossips6149760.cl"},
	{0xCF1D99F5B8BD2E4EULL, "vxw_encounter1_ba_0_0.vlm"},
	{0xCF1D9CE818319ADAULL, "art\\vfx\\models\\vfx_vol_light_narrow_6.mmh"},
	{0xCF1DAEE81831B920ULL, "art\\vfx\\models\\vfx_vol_light_narrow_6.msh"},
	{0xCF1DDC705A314D17ULL, "w_encounter3_6_0.mmh"},
	{0xCF1DE0705A315424ULL, "w_encounter3_6_0.mao"},
	{0xCF1DF2705A317289ULL, "w_encounter3_6_0.msh"},
	{0xCF1DFD21D846D963ULL, "6102756_m.fxe"},
	{0xCF1E471F0BED393BULL, "1014839074.ogg"},
	{0xCF1FA8F773547B1AULL, "art\\vfx\\combat\\vfx_gib_chk.mmh"},
	{0xCF1FADE2BF7F9A93ULL, "vxca_caverns1_ikf_0_0.vlm"},
	{0xCF1FBAF773549A60ULL, "art\\vfx\\combat\\vfx_gib_chk.msh"},
	{0xCF203301B1BC3138ULL, "282074818.ogg"},
	{0xCF208290FF68EC50ULL, "194668301.ogg"},
	{0xCF20B1291D65E64AULL, "dae000ip_ambient_chair_s.utp"},
	{0xCF20C8C2B0F96043ULL, "6111280_m.fxe"},
	{0xCF20F57DAA20B2C8ULL, "follower_barks6153852.cl"},
	{0xCF20FDB38AE00E29ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_enter.gad"},
	{0xCF211CA89101A1F6ULL, "6131129_f.fxe"},
	{0xCF211F3DAF13ACC5ULL, "hlac_w_encounter3_0_2z.dds"},
	{0xCF2159BF13CF0CEBULL, "art\\vfx\\combat\\vfx_r_var_range_v5_c.mmh"},
	{0xCF219CC3EA002F2FULL, "vxlt_0_lowtown_a2_f53_0_0.vlm"},
	{0xCF2255A35B842BBEULL, "vxlt_gamlenshouse_n_iq_0_0.vlm"},
	{0xCF232EE19660C837ULL, "l_dae251ar_mount_caverns.wbd"},
	{0xCF233F65638B9198ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v2_c.mmh"},
	{0xCF23E47C3C273A25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\sibling_diepose2.gad"},
	{0xCF24710BC723E45DULL, "hf_gencr_merc_archer.mor"},
	{0xCF248021D299F3CAULL, "323415882.ogg"},
	{0xCF24FF372E03D8CAULL, "6068097_m.fxe"},
	{0xCF25D378609ED6A8ULL, "vxlt_house_g9_0_0.vlm"},
	{0xCF26272EB6732CB1ULL, "gen00fl_varric6093674.cl"},
	{0xCF26412D541FC58EULL, "505794235.ogg"},
	{0xCF2650BEBE7D39EAULL, "store_one300cr_ilen.utm"},
	{0xCF266B1DF266FC2CULL, "711239498.ogg"},
	{0xCF266F3B0205E154ULL, "vxht_keep_5hl_0_0.vlm"},
	{0xCF2679721EF3A5D0ULL, "6171443_m.fxe"},
	{0xCF26DDF70D9947CDULL, "dae000ip_crowd_ff4_talk11.utp"},
	{0xCF27CEB38AE5DD77ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\grab_action_enter.ani"},
	{0xCF287AE006750B8DULL, "6105896_m.fxe"},
	{0xCF2896D384CF5E72ULL, "6095732_m.fxe"},
	{0xCF2926EC766B421EULL, "vxlt_shop_dj_0_0.vlm"},
	{0xCF2929280887AC9FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_small.msh"},
	{0xCF292F280887B741ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_small.mmh"},
	{0xCF293B280887CBE2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_small.mao"},
	{0xCF296DA6E376D555ULL, "gen00fl_merrill6122224.cl"},
	{0xCF297F6129BAE3A6ULL, "gen_im_trs_krk_one_lsw_r0.uti"},
	{0xCF29AA9A378F2B88ULL, "vxht_mansion2b_236_0_0.vlm"},
	{0xCF29C183AC1E0AB8ULL, "follower_banter6161061.cl"},
	{0xCF29C8978C011D53ULL, "vxht_0_hightown_a3_bdu_0_0.vlm"},
	{0xCF29D95A6FADC433ULL, "45079970.ogg"},
	{0xCF2A044E34B726B6ULL, "character_character_p21.psh.fxo.dx11.assembly"},
	{0xCF2B24B50EB54243ULL, "vrc321_varric6084477.cl"},
	{0xCF2B33EA7A9C1B0DULL, "isa211_isabela6164481.cl"},
	{0xCF2B43227EB8E897ULL, "170785269.ogg"},
	{0xCF2B8E35F2CF3D3EULL, "hlac_dr_0_deeproads_5843_0x.dds"},
	{0xCF2BC24338D8C727ULL, "zz_followers_banter6157636.cl"},
	{0xCF2C0BD2AAB346CEULL, "vxw_encounter1_gb_0_0.vlm"},
	{0xCF2C22182C3AD78FULL, "dae000ip_crowd_lm5_talk15.utp"},
	{0xCF2C4EA5671A5FECULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_trqa_0d.dds"},
	{0xCF2CAA8B007C8DCBULL, "gen00fl_aveline6132745.cl"},
	{0xCF2CCFD4861531ECULL, "mag371_elsa6041148.cl"},
	{0xCF2D192EF7527357ULL, "llac_lt_0_lowtown_docks_a3_n_65534_2.dds"},
	{0xCF2D385320D0CE27ULL, "hf_noble_gen_3.mor"},
	{0xCF2E3ABF13D96250ULL, "art\\vfx\\combat\\vfx_r_var_range_v5_c.anb"},
	{0xCF2EED541A0BEFB8ULL, "follower_barks6153342.cl"},
	{0xCF2F2F96DFDFF567ULL, "6111969_m.fxe"},
	{0xCF2F4157A5752024ULL, "vxw_woundedcoast_1ca_0_0.vlm"},
	{0xCF2F64015FFA66C8ULL, "art\\vfx\\cinematic\\vfx_orge_fire_imp.anb"},
	{0xCF2F717FC0A415B9ULL, "vxht_mansion2_23c_0_0.vlm"},
	{0xCF2FD906CFDC9B37ULL, "qun111_ketojan6039153.cl"},
	{0xCF3004A4514501F9ULL, "fen121ip_to_hightown.utp"},
	{0xCF3010556DF90D26ULL, "art\\characters\\weapons\\staff\\standard_24.mao"},
	{0xCF306C110CD0AFADULL, "6111318_m.fxe"},
	{0xCF3084F8B69BB543ULL, "553221940.ogg"},
	{0xCF31071A8CB2EC76ULL, "lt_0_lowtown_docks_a3_35_0.rml"},
	{0xCF318C403BA7F7F6ULL, "water.mat"},
	{0xCF31DA44472559F8ULL, "vxw_woundedcoast_w0_0_0.vlm"},
	{0xCF324F6F2871BCDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_dualwield.ani"},
	{0xCF325F8DD1411F50ULL, "vxlt_gallowstemplar_3rc_0_0.vlm"},
	{0xCF32694A7FC46340ULL, "hlac_lt_0_lowtown_a2_n_c_65534_3x.dds"},
	{0xCF32D6B9DB3A15B3ULL, "6043657_m.fxe"},
	{0xCF32EFE5E4A87BCEULL, "21787238.ogg"},
	{0xCF34277688BA9280ULL, "6176158_m.fxe"},
	{0xCF3439EB1A527595ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_hvyn_0n.dds"},
	{0xCF35091F61CFA765ULL, "hlac_lt_0_lowtown_docks_a3_n_196608_0y.dds"},
	{0xCF351BCD5303A478ULL, "200479451.ogg"},
	{0xCF357D782AB260CBULL, "6093822_f.fxe"},
	{0xCF35C63E01162650ULL, "103897860.ogg"},
	{0xCF36274D68089DBFULL, "conversation_camera_setting.gda"},
	{0xCF36A306F389277CULL, "6166634_m.fxe"},
	{0xCF36D8AE65307BFCULL, "l_mag291ar_gallows_templar_fd.wbd"},
	{0xCF379B58B46A5144ULL, "t3_arm_ibk.tnt"},
	{0xCF37EB71820854E0ULL, "6126533_m.fxe"},
	{0xCF38131793F72216ULL, "429867532.ogg"},
	{0xCF392F0F10F18C66ULL, "mcr371_scared_mage_f6087715.cl"},
	{0xCF393745A561D205ULL, "6054380_f.fxe"},
	{0xCF396C8EA63187CDULL, "art\\characters\\playercharacter\\elffemale\\ef_har_he7a.phy"},
	{0xCF39CAB16CA3E87FULL, "hlac_dr_primevaltomb_665_1y.dds"},
	{0xCF39DDA61302849AULL, "ht_0_hightown_a2_37_0.rml"},
	{0xCF3A2A7CB5F72CA5ULL, "llac_lt_0_lowtown_n_65536_2.dds"},
	{0xCF3A69CA6DF3E33CULL, "732177618.ogg"},
	{0xCF3A735A16563EB4ULL, "78151558.ogg"},
	{0xCF3A956FAF31206EULL, "llac_ht_chantry_n_2534_4.dds"},
	{0xCF3B0A03B1C694F2ULL, "3128116.ogg"},
	{0xCF3BD50D4A0521D2ULL, "6139441_m.fxe"},
	{0xCF3BFB70EFD184AAULL, "hlac_ca_caverns2prc_5802_50y.dds"},
	{0xCF3C2AF2BFD7F3CDULL, "hlac_lt_0_lowtown_a3_n_c_131072_0z.dds"},
	{0xCF3C4E9EF6191420ULL, "qun160_party_comment_16061229.cl"},
	{0xCF3CFB8E01285FF7ULL, "127085619.ogg"},
	{0xCF3D14353C6E7731ULL, "vxca_caverns1_mines_cjt_0_0.vlm"},
	{0xCF3DB334976B38A4ULL, "vxw_woundedcoast_1nf_0_0.vlm"},
	{0xCF3DCC321AB31A13ULL, "hlac_w_blightlands_131072_0z.dds"},
	{0xCF3E93220B0C2560ULL, "6095778_m.fxe"},
	{0xCF403C2FC63265E7ULL, "vxdr_primevaltomb_4it_0_0.vlm"},
	{0xCF40576F287E07D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_dualwield.evt"},
	{0xCF40653D870AA78BULL, "mag311_emile_de_launcet6064719.cl"},
	{0xCF4113155426F839ULL, "t1_tat_dbl.tnt"},
	{0xCF4136890220F5CAULL, "llac_w_bonepitdestroyed_a3_327680_0.dds"},
	{0xCF417E1777B8EF4CULL, "vxht_mansion2_2on_0_0.vlm"},
	{0xCF41943D066B3530ULL, "lt_0_lowtown_docks_a3_9_0_lc.mmh"},
	{0xCF41B71D82022487ULL, "6118711_m.fxe"},
	{0xCF41FA1C6F21A0CCULL, "89344330.ogg"},
	{0xCF422FF21976F458ULL, "vxlt_gallowstemplar_n_5j2_0_0.vlm"},
	{0xCF42B5451F8078A5ULL, "mag311_keran6041695.cl"},
	{0xCF42E46EED9DB3B9ULL, "vxlt_shop_g8_0_0.vlm"},
	{0xCF4350FA72F4F460ULL, "6108048_f.fxe"},
	{0xCF442074C5BC8AEDULL, "hlac_lt_warehouse_n_0_0y.dds"},
	{0xCF4486423B436990ULL, "440552835.ogg"},
	{0xCF4498B84D02F725ULL, "gen_im_trs_krk_hvy_cht_r1.uti"},
	{0xCF44DCE84D210332ULL, "follower_banter6171606.cl"},
	{0xCF455A7F9F963200ULL, "vxht_keep_6f9_0_0.vlm"},
	{0xCF456637BC707BB0ULL, "337561825.ogg"},
	{0xCF457012BC2D4116ULL, "hlac_lt_0_lowtown_docks_-131074_6y.dds"},
	{0xCF457C4210E546C3ULL, "vxlt_undercityserialkiller_1ma_0_0.vlm"},
	{0xCF457CF6A7F17A30ULL, "vxht_keep_62d_0_1.vlm"},
	{0xCF45969A48A1117DULL, "vxht_brothel_n_6ei_0_0.vlm"},
	{0xCF465734F4E48CD4ULL, "vxht_0_hightown_c77_0_0.vlm"},
	{0xCF4660AA2FA6D050ULL, "6082129_m.fxe"},
	{0xCF46796C85293CB7ULL, "fen200cr_orana.utc"},
	{0xCF46A15C777384F9ULL, "6133747_m.fxe"},
	{0xCF46A7EE29DFCDF6ULL, "vxlt_0_lowtown_a2_n_e8a_0_0.vlm"},
	{0xCF46D21F7DF1B4E9ULL, "lt_0_lowtown_docks_215_0.mao"},
	{0xCF46DE1F7DF1C882ULL, "lt_0_lowtown_docks_215_0.mmh"},
	{0xCF46E01F7DF1CC38ULL, "lt_0_lowtown_docks_215_0.msh"},
	{0xCF472ED3D49BDB71ULL, "llac_dr_0_deeproads_5845_12.dds"},
	{0xCF4764FD37BB48DEULL, "lt_undercitydungeon.arl"},
	{0xCF47766F28845B65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_dualwield.gad"},
	{0xCF479B88A4137F13ULL, "plt_and100pt_recruited.plo"},
	{0xCF47A8EA5AA7A9DEULL, "80841911.ogg"},
	{0xCF48344425046E6DULL, "hlac_ca_caverns2prc_5802_42y.dds"},
	{0xCF485E1D664A0CCCULL, "follower_banter6143270.cl"},
	{0xCF48EBF77C823F79ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerctall.msh"},
	{0xCF48F5F77C825067ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerctall.mmh"},
	{0xCF48F9F77C8256B4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerctall.mao"},
	{0xCF49A8DD3A894FEFULL, "6167633_f.fxe"},
	{0xCF4A66AB34D5D7BEULL, "lgt000pt_tombs_third.ncs"},
	{0xCF4A9776151170C0ULL, "200235938.ogg"},
	{0xCF4B2814864BEB9EULL, "vxlt_gallowscourtyard_74u_0_0.vlm"},
	{0xCF4B90D9D411AE9EULL, "6100768_m.fxe"},
	{0xCF4C17C6BBE05A88ULL, "one111ip_ggt_crate_4.utp"},
	{0xCF4C251189615124ULL, "vxw_woundedcoast_1ig_0_0.vlm"},
	{0xCF4C62819BEEE654ULL, "6105655_m.fxe"},
	{0xCF4CF68CC9A779E7ULL, "gen00fl_fenris6108103.cl"},
	{0xCF4D1389B6CBA2A3ULL, "follower_banter6173126.cl"},
	{0xCF4D8690979EF728ULL, "vxht_keep_n_6fz_0_0.vlm"},
	{0xCF4DB80DB514A569ULL, "vxw_bonepitmining_a2_7z_0_0.vlm"},
	{0xCF4DCC22720E66DBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_potb_sa.dds"},
	{0xCF4DD5D868A0BD7AULL, "216750553.ogg"},
	{0xCF4DFE5A5E1C6835ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_keep.phy"},
	{0xCF4E00660074F2EAULL, "609783218.ogg"},
	{0xCF4F2E281444E1A8ULL, "797020773.ogg"},
	{0xCF4F3983FE5BFF1BULL, "706781.ogg"},
	{0xCF4F61B35AE33EB3ULL, "dae011_corff_sta6149649.cl"},
	{0xCF4F6E6ACBD3348FULL, "fen110_fenris6100278.cl"},
	{0xCF4F88BDEB30FA0BULL, "hlac_ht_chantry_2534_31x.dds"},
	{0xCF4FABD13CA852E3ULL, "812996796.ogg"},
	{0xCF506AF9F64222DAULL, "6035288_m.fxe"},
	{0xCF5071CEA58DA7B2ULL, "6013279_m.fxe"},
	{0xCF516AC879DE9730ULL, "vxht_keep_6gz_0_0.vlm"},
	{0xCF5176CEED808201ULL, "hlac_w_bonepitmining_a2_65536_0x.dds"},
	{0xCF51BEE0320E1B37ULL, "991634610.ogg"},
	{0xCF5234513AE057FFULL, "genip_pressure_plate.ncs"},
	{0xCF52926725C8E43EULL, "hlac_ht_0_hightown_a2_-196608_2z.dds"},
	{0xCF52A8F2D396EB52ULL, "6150354_m.fxe"},
	{0xCF52BD64A3AED863ULL, "sky_ht_0_hightown_a3.mao"},
	{0xCF52EF0B6BB4E8B2ULL, "335103663.ogg"},
	{0xCF52F7B4776E517FULL, "char2.vsh.fxo"},
	{0xCF52FD05E3FE21CAULL, "zz_qun_debug6028267.cl"},
	{0xCF531A3C8D148FD9ULL, "invisibletoai.ncs"},
	{0xCF537B3D6CEED985ULL, "479409947.ogg"},
	{0xCF538B0AE2D7395DULL, "dre000_sibling_has_blight6036505.cl"},
	{0xCF53C8AF0E8CE1B1ULL, "vxca_caverns2_97k_0_0.vlm"},
	{0xCF546DFAB33B0C08ULL, "lt_0_lowtown_n_38_0.anb"},
	{0xCF54C1AE1B61E0B5ULL, "zz_followers_banter6159294.cl"},
	{0xCF54CF2B65F485E0ULL, "6128733_m.fxe"},
	{0xCF54F98C94907826ULL, "qcr000cs_thechampion.cut"},
	{0xCF54FCA165BDE823ULL, "vxw_bonepitmining_a2_21b_0_0.vlm"},
	{0xCF550F3C92465F2CULL, "6139429_m.fxe"},
	{0xCF55161E8097D50FULL, "gen00fl_merrill6119589.cl"},
	{0xCF55302808ACF3BEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1ci_small.phy"},
	{0xCF554A5586CAFB21ULL, "vxht_0_hightown_a3_blg_0_0.vlm"},
	{0xCF558C497FCCC815ULL, "ss_t_deeproads_2.cut"},
	{0xCF55E08CCA2553FCULL, "gen00fl_anders6114497.cl"},
	{0xCF55FF8CEE1FF800ULL, "676249141.ogg"},
	{0xCF564919CFD0F686ULL, "618513394.ogg"},
	{0xCF567A1B116D47F9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\til_ruins01_sa.dds"},
	{0xCF56A0819BD455FAULL, "183959878.ogg"},
	{0xCF57076387B81141ULL, "vxlt_backalley_g_1e7_0_0.vlm"},
	{0xCF573D3A27FBCC3EULL, "hlac_lt_0_lowtown_docks_-196610_5y.dds"},
	{0xCF57532442F1394DULL, "vxlt_darktown_47f_0_0.vlm"},
	{0xCF57BE9FA8B1FD21ULL, "vxlt_0_lowtown_a2_n_hho_0_0.vlm"},
	{0xCF57D9CAC5C747C9ULL, "lgt220_shp_mag_guard6165457.cl"},
	{0xCF57EAD876EA335BULL, "vxlt_darktown_4oh_0_0.vlm"},
	{0xCF57FD2DBF7A00C7ULL, "ach_comp_ruler.dds"},
	{0xCF586301601D6A63ULL, "art\\vfx\\cinematic\\vfx_orge_fire_imp.mmh"},
	{0xCF591117A498C2E0ULL, "mer220_hubert6041219.cl"},
	{0xCF5971D7D9B076B3ULL, "vxlt_0_lowtown_ff2_0_0.vlm"},
	{0xCF59F18145ABE5E7ULL, "606388666.ogg"},
	{0xCF5A02B6C607C71EULL, "6105468_m.fxe"},
	{0xCF5A76029DBD804DULL, "356205349.ogg"},
	{0xCF5A7DD2DE4F7940ULL, "vxdr_primevaltomb_53h_0_0.vlm"},
	{0xCF5A7E00439F0690ULL, "mattest_2.cut"},
	{0xCF5A9E814887BB9AULL, "vxht_0_hightown_a3_c1b_0_0.vlm"},
	{0xCF5AEDEF3A355929ULL, "378938524.ogg"},
	{0xCF5B85BC6DC73DCFULL, "vxlt_0_lowtown_e10_0_0.vlm"},
	{0xCF5C2F4519E8CB7EULL, "209016503.ogg"},
	{0xCF5CD1C158AFFC51ULL, "lt_backalley_g_1_0_lc.mmh"},
	{0xCF5CED729C71B486ULL, "follower_banter6163274.cl"},
	{0xCF5D607C13BEAF4BULL, "vxlt_gallowscourtyard_n_6ym_0_0.vlm"},
	{0xCF5DD38229CAD4E0ULL, "llac_ht_chantry_2534_36.dds"},
	{0xCF5DD9615464D74CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasec_sa.dds"},
	{0xCF5DF1E02A6FEAB3ULL, "vxw_encounter1_ea_0_0.vlm"},
	{0xCF5E61D15BA5204CULL, "vxht_mansion2_2ul_0_0.vlm"},
	{0xCF5E9DEFA64ECCCFULL, "1002281658.ogg"},
	{0xCF5F43864D24DBECULL, "gen00fl_fenris6126450.cl"},
	{0xCF5F9C9980B159BBULL, "dae300_cullen6140605.cl"},
	{0xCF600E013A2E60E3ULL, "6082269_m.fxe"},
	{0xCF600EFDF8521A3DULL, "vxlt_0_lowtown_a2_ffp_0_0.vlm"},
	{0xCF604D06F1A9A8F8ULL, "vxw_bonepit_1j9_0_0.vlm"},
	{0xCF606FD54E42230FULL, "hlac_lt_0_lowtown_a2_n_65538_0x.dds"},
	{0xCF60880673E983AAULL, "gen00fl_isabela6096812.cl"},
	{0xCF60A3A25CDB7D96ULL, "719971440.ogg"},
	{0xCF62F519464F52B1ULL, "480810992.ogg"},
	{0xCF632CAFAE73B86DULL, "6119167_m.fxe"},
	{0xCF634431C126C77EULL, "art\\characters\\weapons\\staff\\textures\\w_stf_002a_0d.dds"},
	{0xCF635984520B6CE7ULL, "106281676.ogg"},
	{0xCF635FDB63A081A1ULL, "191020134.ogg"},
	{0xCF6388BE794A6FF9ULL, "gen_im_arm_cht_comc_t4.uti"},
	{0xCF63B51B1DACDC72ULL, "vxw_encounter2_camp_5u_0_0.vlm"},
	{0xCF63D6AE65576676ULL, "l_mag291ar_gallows_templar_fd.bnk"},
	{0xCF64B7FE7A4A9D0EULL, "art\\characters\\creatures\\stonegolem\\animation\\quake.ani"},
	{0xCF656F5B7D163CEEULL, "one350_variel6085177.cl"},
	{0xCF6647FBF37D43ABULL, "6052567_m.fxe"},
	{0xCF6659C65AB98DB4ULL, "740105164.ogg"},
	{0xCF66876CC775C0D8ULL, "562380952.ogg"},
	{0xCF66F52120932494ULL, "art\\vfx\\environment\\vfx_fire_torch_p.mmh"},
	{0xCF66F648794BABD4ULL, "845929638.ogg"},
	{0xCF6712AC7B95C58DULL, "81838759.ogg"},
	{0xCF67989661DA78A8ULL, "w_woundedcoast_205_0.mmh"},
	{0xCF67A49661DA8D4BULL, "w_woundedcoast_205_0.mao"},
	{0xCF67B69661DAABD2ULL, "w_woundedcoast_205_0.msh"},
	{0xCF68B43CBCF0111AULL, "vxlt_0_lowtown_a3_n_c_ekj_0_1.vlm"},
	{0xCF696247DCE72538ULL, "llac_ht_mansion2_818_40.dds"},
	{0xCF699BEA81017834ULL, "dae021_brothel_onenter6196727.cl"},
	{0xCF699E70669B2221ULL, "438975601.ogg"},
	{0xCF69A140713F3966ULL, "lgt260pt_way_blood_mage.ncs"},
	{0xCF6A9BC79900D669ULL, "vxw_bonepitmining_a2_9x_0_0.vlm"},
	{0xCF6ACFF2F1304781ULL, "6135261_f.fxe"},
	{0xCF6AD60309A54830ULL, "6017677_m.fxe"},
	{0xCF6B0216FE0E43B3ULL, "hlac_lt_undercityencounter_-1_14x.dds"},
	{0xCF6B4E20621AD308ULL, "art\\cinematics\\props\\prp_barge03.mao"},
	{0xCF6C5AA5CD5026B8ULL, "vxca_cave_2d3_0_0.vlm"},
	{0xCF6C7CC41ED4064EULL, "6046883_f.fxe"},
	{0xCF6C8D0D0477A66FULL, "348767323.ogg"},
	{0xCF6D7446B48A5BA9ULL, "316370830.ogg"},
	{0xCF6D8EC47AF21B4FULL, "6207736_m.fxe"},
	{0xCF6E620706F72615ULL, "lowtown_male_alienage1_sta6162724.cl"},
	{0xCF6ECFAF85350BA1ULL, "art\\characters\\weapons\\spears\\textures\\w_qui_qun_0d.dds"},
	{0xCF6EF2C001BBD6DEULL, "vxlt_ruinedhouse_n_i9_0_0.vlm"},
	{0xCF6F2726E1A64773ULL, "6127404_f.fxe"},
	{0xCF6F792E7E678596ULL, "hlac_lt_0_lowtown_docks_n_-196611_2z.dds"},
	{0xCF6FB28378D1171BULL, "vxlt_gallowstemplar_n_37h_0_0.vlm"},
	{0xCF7043C73CE33ECBULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.dura.evt"},
	{0xCF70FCABE171C304ULL, "6040563_m.fxe"},
	{0xCF71AFFE7A551A5FULL, "art\\characters\\creatures\\stonegolem\\animation\\quake.evt"},
	{0xCF71D4DE6CA8411BULL, "vxlt_gallowstemplar_f_3sd_0_0.vlm"},
	{0xCF721F1C1FF0529AULL, "hlac_lt_0_lowtown_docks_-131073_1y.dds"},
	{0xCF724B2E158A839AULL, "w_bonepitdestroyed_a3_224_0.msh"},
	{0xCF72592E158A9BD3ULL, "w_bonepitdestroyed_a3_224_0.mao"},
	{0xCF725D2E158AA2E0ULL, "w_bonepitdestroyed_a3_224_0.mmh"},
	{0xCF7366637670F666ULL, "6116074_f.fxe"},
	{0xCF73BAA61333C1A6ULL, "ht_0_hightown_a2_37_0.anb"},
	{0xCF7436DE26DD6A4DULL, "vxlt_darktown_49d_0_0.vlm"},
	{0xCF744F165EFF25D3ULL, "pro000_colonel_ewald6040553.cl"},
	{0xCF74D75B1846184BULL, "vxdr_0_deeproads_f5t_0_0.vlm"},
	{0xCF74E79E5AB037BDULL, "vxht_keep_4h8_0_0.vlm"},
	{0xCF757E8EDB467613ULL, "484511640.ogg"},
	{0xCF7583DF0E1A8247ULL, "vxlt_blackemporium_t8_0_0.vlm"},
	{0xCF75AFE19883E5AFULL, "llac_w_encounter2_camp_-131072_0.dds"},
	{0xCF75D92DBB9943D1ULL, "hlac_lt_0_lowtown_a2_-2_3x.dds"},
	{0xCF76EB63152739D0ULL, "546409251.ogg"},
	{0xCF76F0E518737C83ULL, "6176180_m.fxe"},
	{0xCF775CC73CE9878CULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.dura.gad"},
	{0xCF778288A43C78B0ULL, "plt_and100pt_recruited.nss"},
	{0xCF77E844BE853761ULL, "vxlt_0_lowtown_docks_n_b5n_0_1.vlm"},
	{0xCF7890FE7A5B0448ULL, "art\\characters\\creatures\\stonegolem\\animation\\quake.gad"},
	{0xCF78D6EE86B1F8F2ULL, "vxlt_foundry_n_105_0_0.vlm"},
	{0xCF792024B106EAEEULL, "107279231.ogg"},
	{0xCF7929DF982579C2ULL, "49252163.ogg"},
	{0xCF797B87E5CF1EE7ULL, "6108263_f.fxe"},
	{0xCF7982F4D591C3AEULL, "gen_im_flw_dua_isabela_02.uti"},
	{0xCF7AB5F2F51F5B8BULL, "387492413.ogg"},
	{0xCF7B1754367DB220ULL, "llac_lt_0_lowtown_n_196606_0.dds"},
	{0xCF7B512CF67B417DULL, "nat380_nathaniel6068341.cl"},
	{0xCF7BA66C1E1C950BULL, "535018353.ogg"},
	{0xCF7BB914CE805876ULL, "dm_traditionalist.mor"},
	{0xCF7CC995E2B26771ULL, "llac_ca_caverns2_5802_28.dds"},
	{0xCF7CDA5FBD002D05ULL, "llac_ht_playershouse_n_855_17.dds"},
	{0xCF7CE3CCD9494060ULL, "vxlt_0_lowtown_docks_a3_cp0_0_0.vlm"},
	{0xCF7CF60EA2041A31ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_hold_loop.ani"},
	{0xCF7D2ECC4D8E1AFBULL, "hlac_lt_0_lowtown_a2_n_c_-65539_0x.dds"},
	{0xCF7D375B49BB3B03ULL, "vxht_mansion2b_1t3_0_0.vlm"},
	{0xCF7D3A754D2E9240ULL, "6055110_m.fxe"},
	{0xCF7D3D8878232594ULL, "vxlt_0_lowtown_a3_n_c_gnx_0_0.vlm"},
	{0xCF7D5BC73CEDF18AULL, "art\\vfx\\combat\\vfx_w_jug_lineback_loop_c.dura.ani"},
	{0xCF7D73608D9C8D5EULL, "hlac_lt_hangedman_-1_12x.dds"},
	{0xCF7E1ECE1F25B329ULL, "art\\characters\\creatures\\stonegolem\\animation\\bdmg.ani"},
	{0xCF7E63FE075E47CFULL, "165971789.ogg"},
	{0xCF7EDF1859F0357BULL, "vxlt_0_lowtown_a3_n_c_i1d_0_0.vlm"},
	{0xCF7F0F482C2D5C1DULL, "gen00fl_aveline6101697.cl"},
	{0xCF7F43B5D8BDA8C3ULL, "vxlt_undercityserialkiller_1qe_0_0.vlm"},
	{0xCF7F5DCAC37CC51AULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.anb"},
	{0xCF7FDF5DD5BC0501ULL, "918679311.ogg"},
	{0xCF7FE49D47C5F09DULL, "lgt260_way_blood_mage.bnk"},
	{0xCF80A95A6F578D65ULL, "615032440.ogg"},
	{0xCF80D6E5E2CBE8E3ULL, "hlac_w_encounter3_0_1y.dds"},
	{0xCF80F241DD3E0EA4ULL, "231223072.ogg"},
	{0xCF814B78E17D2AF3ULL, "dae011_badpoet_sta6145206.cl"},
	{0xCF81A246014A69F1ULL, "follower_banter6169318.cl"},
	{0xCF81AADA9D1B7C34ULL, "llac_w_encounter2_65535_0.dds"},
	{0xCF81ABDC0BEBDD86ULL, "467347865.ogg"},
	{0xCF826A117D5B80ADULL, "vxlt_undercityencounter_13m_0_0.vlm"},
	{0xCF82C685DD52809EULL, "6030463_m.fxe"},
	{0xCF82FB10626E8068ULL, "mag160_feynriel6020425.cl"},
	{0xCF8324346B9181FBULL, "vxw_sundermount_309_0_0.vlm"},
	{0xCF8328E36B97BDC5ULL, "6109724_m.fxe"},
	{0xCF83294D142B954CULL, "612720116.ogg"},
	{0xCF8356E447956632ULL, "616121216.ogg"},
	{0xCF83C69D47C9A6A5ULL, "lgt260_way_blood_mage.cnv"},
	{0xCF83E0D61597654CULL, "294190177.ogg"},
	{0xCF84010EA20A4BD3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_hold_loop.gad"},
	{0xCF843958BBB469DAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cratetarp_dp.dds"},
	{0xCF8490EBAADDC6F8ULL, "6126308_m.fxe"},
	{0xCF84F76C710FAEAEULL, "dae100_viscount6103306.cl"},
	{0xCF8539CE1F2BFFABULL, "art\\characters\\creatures\\stonegolem\\animation\\bdmg.gad"},
	{0xCF855F51691DFF6DULL, "follower_banter6163864.cl"},
	{0xCF8597F6A0DC421AULL, "vxlt_0_lowtown_a3_n_c_eqd_0_1.vlm"},
	{0xCF863118DA726DABULL, "675579305.ogg"},
	{0xCF86358710237E05ULL, "6138742_m.fxe"},
	{0xCF863EEFC209B71BULL, "llac_lt_0_lowtown_-196609_2.dds"},
	{0xCF8665AF647AA9C2ULL, "dae000_gen_player_turn_in6177835.cl"},
	{0xCF8690B256D4364EULL, "follower_barks6153937.cl"},
	{0xCF87B4FD7EBDF2F8ULL, "99181242.ogg"},
	{0xCF881174A0C050DDULL, "hlac_lt_blackemporium_1055_4y.dds"},
	{0xCF88760801E9BDC8ULL, "hm_alistair_warden.mor"},
	{0xCF8934D54CFFD163ULL, "em_one_jelden.mor"},
	{0xCF893803F30DB00DULL, "dre000_varric6027784.cl"},
	{0xCF89B273581BB37EULL, "hlac_w_blightlands_196607_0x.dds"},
	{0xCF89D049B4BC88F7ULL, "6154702_m.fxe"},
	{0xCF8A5BA0DCE0472CULL, "1067270636.ogg"},
	{0xCF8A5D2F2E1F5F41ULL, "isa311_isabela6092505.cl"},
	{0xCF8A9461FE92136EULL, "hlac_lt_undercityencounter_-1_35x.dds"},
	{0xCF8A959AFB37398FULL, "vxlt_0_lowtown_f1c_0_0.vlm"},
	{0xCF8AF513B620A237ULL, "1057238289.ogg"},
	{0xCF8B36CE1F306610ULL, "art\\characters\\creatures\\stonegolem\\animation\\bdmg.evt"},
	{0xCF8B697822639442ULL, "6110826_m.fxe"},
	{0xCF8CC9381044D1D3ULL, "vxca_caverns2prc_9p1_0_0.vlm"},
	{0xCF8CD0B72501832FULL, "hlac_ht_chantry_n_2534_48y.dds"},
	{0xCF8CE9B83A0F9230ULL, "llac_lt_0_lowtown_docks_a3_65533_1.dds"},
	{0xCF8DF51E890A0E1FULL, "6090297_m.fxe"},
	{0xCF8E0BCDE1F1FBB4ULL, "vxca_caverns1_hi6_0_0.vlm"},
	{0xCF8E1610E5430970ULL, "leave_clo.mao"},
	{0xCF8E19B1622970C8ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chaintop.msh"},
	{0xCF8E2BB162298F59ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chaintop.mao"},
	{0xCF8E35FD0B38B455ULL, "134713242.ogg"},
	{0xCF8E37B16229A3F2ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chaintop.mmh"},
	{0xCF8E7DE67A1272D2ULL, "6076616_m.fxe"},
	{0xCF8E909D34FB2F90ULL, "lt_0_lowtown_docks_a3_252_0.msh"},
	{0xCF8E9E9D34FB47AAULL, "lt_0_lowtown_docks_a3_252_0.mmh"},
	{0xCF8EA29D34FB4EE1ULL, "lt_0_lowtown_docks_a3_252_0.mao"},
	{0xCF8F00DB902A4B73ULL, "gen00fl_merrill6122363.cl"},
	{0xCF8F44F3208916C7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_doorcrest_dp.dds"},
	{0xCF8F4CEC84394161ULL, "vxlt_0_lowtown_docks_a3_n_cvt_0_0.vlm"},
	{0xCF8F5F4964DB5DF4ULL, "lowtown_male_beggar_sta6156761.cl"},
	{0xCF8F605A99F6F103ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv3.ani"},
	{0xCF8FADD69A513EE0ULL, "llac_lt_0_lowtown_a2_65535_1.dds"},
	{0xCF8FF82120B62F4BULL, "art\\vfx\\environment\\vfx_fire_torch_p.anb"},
	{0xCF9001C1699AF7D0ULL, "6081206_f.fxe"},
	{0xCF907032F71213BBULL, "449443058.ogg"},
	{0xCF90EB6D34144F85ULL, "6062373_m.fxe"},
	{0xCF91118F3AE8F3AEULL, "llac_dr_slavers_3217_7.dds"},
	{0xCF911E9E34B777F0ULL, "1026132520.ogg"},
	{0xCF91E57083748BDFULL, "gen00fl_aveline6137579.cl"},
	{0xCF9347330251A973ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_kd_exit.ani"},
	{0xCF934A0641E716A0ULL, "hlac_lt_0_lowtown_a3_n_c_65536_8x.dds"},
	{0xCF9367E5654A98E0ULL, "304800843.ogg"},
	{0xCF93CA4F24D8CB20ULL, "6191269_m.fxe"},
	{0xCF93CBECC623DF96ULL, "vxdr_0_deeproads_bp3_0_1.vlm"},
	{0xCF93F2BDD4666B1EULL, "llac_ca_caverns1_mines_24423_9.dds"},
	{0xCF9433ABD3D2CC21ULL, "6086999_m.fxe"},
	{0xCF946CFAB370EEC0ULL, "lt_0_lowtown_n_38_0.rml"},
	{0xCF9492BBD56B0852ULL, "6105764_f.fxe"},
	{0xCF953745E03F9ADEULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.dura.ani"},
	{0xCF95484B536C5036ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkfl.ani"},
	{0xCF95A6D664F37604ULL, "268689971.ogg"},
	{0xCF95FF7A96BA5EDAULL, "vxw_encounter2_q0_0_1.vlm"},
	{0xCF96115A5E5A3E50ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_keep.msh"},
	{0xCF961F5A5E5A566AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_keep.mmh"},
	{0xCF962126DEA35CAFULL, "6004414_m.fxe"},
	{0xCF96235A5E5A5CA1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rr_keep.mao"},
	{0xCF9680D43B9536E1ULL, "mag330_walter6044824.cl"},
	{0xCF96D477B3EA7E74ULL, "hi_dust.mao"},
	{0xCF96E1556BF4423BULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha12a.msh"},
	{0xCF96FF556BF47525ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha12a.mmh"},
	{0xCF971A7C56F06C92ULL, "gen_im_acc_blt_act1_02.uti"},
	{0xCF97334B4F2FD9FBULL, "vxlt_undercityencounter_18z_0_0.vlm"},
	{0xCF97CE87C7794E80ULL, "vxht_0_hightown_n_9bi_0_0.vlm"},
	{0xCF9811FAE7B3E4B3ULL, "6106324_f.fxe"},
	{0xCF98180B513EEF29ULL, "vxdr_primevaltomb_34g_0_1.vlm"},
	{0xCF983131C6065482ULL, "922679683.ogg"},
	{0xCF985296D0170CE6ULL, "dre000_carver6037610.cl"},
	{0xCF985C86E92C7233ULL, "follower_banter6136709.cl"},
	{0xCF98814BE3695FCAULL, "hlac_lt_0_lowtown_n_1_2y.dds"},
	{0xCF9956651E61DD0EULL, "6147646_m.fxe"},
	{0xCF996885608F45B6ULL, "vxlt_0_lowtown_docks_n_b6p_0_1.vlm"},
	{0xCF997A1F796B43F1ULL, "dae000_mother6127188.cl"},
	{0xCF998142E034F5D7ULL, "6157189_m.fxe"},
	{0xCF9ADBF6F93FD075ULL, "6099155_m.fxe"},
	{0xCF9AFE81F4A1566DULL, "6121946_m.fxe"},
	{0xCF9B5BAF49B9D7D3ULL, "206291674.ogg"},
	{0xCF9BB326876FD6AFULL, "vxca_caverns2_5uk_0_0.vlm"},
	{0xCF9C16BF8547A2CBULL, "498965338.ogg"},
	{0xCF9C1AD66CCD1416ULL, "911907520.ogg"},
	{0xCF9C592FEDD4A93AULL, "dabloom.mao"},
	{0xCF9C785A9A01A4FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv3.evt"},
	{0xCF9CB2317E7EA6F3ULL, "character_character_p181.psh.fxo.dx11"},
	{0xCF9CB5F181B078D5ULL, "vxca_cave_163_0_0.vlm"},
	{0xCF9CBDB0C06AAFC9ULL, "hlac_lt_0_lowtown_a3_n_c_1_1x.dds"},
	{0xCF9D55038DCE1C62ULL, "gen00fl_anders6119306.cl"},
	{0xCF9E882398B1E304ULL, "452213245.ogg"},
	{0xCF9EF8F6D6123701ULL, "vxht_0_hightown_n_91m_0_0.vlm"},
	{0xCFA01299BD131E35ULL, "6065703_m.fxe"},
	{0xCFA03F33025C25AAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_kd_exit.evt"},
	{0xCFA065D957FF5F12ULL, "vrc221_varric6043863.cl"},
	{0xCFA10C73FB45E98BULL, "hlac_lt_0_lowtown_docks_65533_3x.dds"},
	{0xCFA11A639A78F4B9ULL, "gen_im_flw_sta_bethany_01.uti"},
	{0xCFA182F17C0F9CAAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_exit.gad"},
	{0xCFA2053237D039DCULL, "6150994_m.fxe"},
	{0xCFA24F45E04A4DAFULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.dura.evt"},
	{0xCFA28319F15E7BA5ULL, "560943479.ogg"},
	{0xCFA3504B53789B27ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkfl.evt"},
	{0xCFA3626354122805ULL, "152172634.ogg"},
	{0xCFA3775A9A07C1EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\attackv3.gad"},
	{0xCFA409D1F18DD0ADULL, "zz_ave_debug6175052.cl"},
	{0xCFA41B8A773872F3ULL, "gen_im_trs_dwv_junk5_04.uti"},
	{0xCFA46AD436BE1F3CULL, "vxht_mansion2_1su_0_0.vlm"},
	{0xCFA52830CEAEF04CULL, "hlac_lt_0_lowtown_docks_a3_-131073_1x.dds"},
	{0xCFA5722B31D248E0ULL, "vxw_woundedcoast_3d_0_0.vlm"},
	{0xCFA57766D036E0BBULL, "6015944_m.fxe"},
	{0xCFA60BF87E8A04E9ULL, "dae000_bartrand6031121.cl"},
	{0xCFA665D7940F464CULL, "pro000_opening_pc6001195.cl"},
	{0xCFA66C09D5D75F9CULL, "641059175.ogg"},
	{0xCFA6A0D0D4E44C9FULL, "6048303_f.fxe"},
	{0xCFA6D6C95F04F74FULL, "vxht_mansion_n_62i_0_0.vlm"},
	{0xCFA6F7A08E836F68ULL, "docks_male_elven_worker_sta6176113.cl"},
	{0xCFA71E3302620C1DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_kd_exit.gad"},
	{0xCFA773857CBBF01AULL, "llac_w_bonepitdestroyed_a3_262150_0.dds"},
	{0xCFA7840C2059CCADULL, "6111109_m.fxe"},
	{0xCFA7AB981B1295D4ULL, "209504112.ogg"},
	{0xCFA7EF560A00F840ULL, "vxw_bonepit_lh_0_0.vlm"},
	{0xCFA82E899814259AULL, "410209022.ogg"},
	{0xCFA860CAC39FCFCDULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.mmh"},
	{0xCFA865F17C158A41ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_exit.evt"},
	{0xCFA86C8667A9358EULL, "639088871.ogg"},
	{0xCFA86E1BAE16509EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkright.evt"},
	{0xCFA87FB5867C090FULL, "ht_0_hightown_18_0.rml"},
	{0xCFA883B4DEECBC33ULL, "621055567.ogg"},
	{0xCFA8BBDFA21A5BEBULL, "565269697.ogg"},
	{0xCFA95045E0506E58ULL, "art\\vfx\\environment\\vfx_trap_gas_barrel_c.dura.gad"},
	{0xCFA9714B537D3F70ULL, "art\\characters\\creatures\\stonegolem\\animation\\cwkfl.gad"},
	{0xCFA9F081553E7F5DULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\blendtree\\trooper_movement.anb"},
	{0xCFAA8DA9C462FF5AULL, "vxw_bonepit_iy_0_0.vlm"},
	{0xCFABF89818B4BDDAULL, "6169376_m.fxe"},
	{0xCFAC3E0D73542163ULL, "character_character_p167.psh.fxo.dx11.assembly"},
	{0xCFACB49B7A9788EDULL, "vxlt_gallowscourtyard_n_74e_0_0.vlm"},
	{0xCFACD1BFD6B5C68FULL, "gen00fl_varric6105427.cl"},
	{0xCFAD37E18DD74A60ULL, "6140267_m.fxe"},
	{0xCFADD0CDC4F77615ULL, "vxdr_0_deeproads_cft_0_2.vlm"},
	{0xCFADF48E95EFF292ULL, "6134301_m.fxe"},
	{0xCFAE59ABCEEF7972ULL, "fen310_varania6118642.cl"},
	{0xCFAE851BAE1AE3C9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkright.gad"},
	{0xCFAEBDDA888EBCD1ULL, "6118832_m.fxe"},
	{0xCFAF3E75F4231C1BULL, "440974281.ogg"},
	{0xCFAF6BEF6A3C19D4ULL, "vxht_brothel_n_5ga_0_0.vlm"},
	{0xCFAF881AF6AD0DEEULL, "llac_lt_0_lowtown_a3_n_c_-196609_1.dds"},
	{0xCFAF96AD271B1B0AULL, "hm_guard.mor"},
	{0xCFB057A30AC2D86EULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_0z.dds"},
	{0xCFB07CE3BF176A85ULL, "vxht_0_hightown_a2_cad_0_0.vlm"},
	{0xCFB0B8FAEB6D6AD0ULL, "603982369.ogg"},
	{0xCFB0F310B462A377ULL, "6084432_m.fxe"},
	{0xCFB159F70E702914ULL, "6027490_m.fxe"},
	{0xCFB17228A85E2BFBULL, "mer320_hubert6051889.cl"},
	{0xCFB1DD6A90CC36D9ULL, "vxht_mansion2b_2vt_0_0.vlm"},
	{0xCFB21F4BF6D82BDBULL, "w_blightlands_38_0.rml"},
	{0xCFB23AC84BA63F5CULL, "vxlt_gallowstemplar_n_45r_0_0.vlm"},
	{0xCFB2798D2834A955ULL, "6111080_m.fxe"},
	{0xCFB3601518983E91ULL, "hlac_lt_0_lowtown_n_65533_1z.dds"},
	{0xCFB3DDC90973084AULL, "hlac_ht_0_hightown_n_65536_1x.dds"},
	{0xCFB3DED901008307ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st105a_0d.dds"},
	{0xCFB40BE9E7EC095AULL, "vxlt_0_lowtown_a3_n_c_e8b_0_0.vlm"},
	{0xCFB442A9E9C81D49ULL, "qcr000_arishok6039306.cl"},
	{0xCFB457D21D4A8E55ULL, "75005017.ogg"},
	{0xCFB46B774110B38AULL, "564377172.ogg"},
	{0xCFB4D94B82B8EC1BULL, "6180394_m.fxe"},
	{0xCFB5561BAE20B217ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\walkright.ani"},
	{0xCFB55DF17C200690ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_exit.ani"},
	{0xCFB602A37352575DULL, "260252196.ogg"},
	{0xCFB6423AC57B7C85ULL, "mcr370_negotiation6080276.cl"},
	{0xCFB66CE779C0CFAAULL, "730360124.ogg"},
	{0xCFB67C36EA96F8B7ULL, "vxca_caverns1_jfr_0_0.vlm"},
	{0xCFB69525CD088211ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.impa.ani"},
	{0xCFB6A554550ACF94ULL, "6110059_m.fxe"},
	{0xCFB6CD7E3C29E97DULL, "dre000_cut_varric_intro.bnk"},
	{0xCFB7D520F3E645E9ULL, "zz_fen_debug6134368.cl"},
	{0xCFB82180BEDEFDDAULL, "128162235.ogg"},
	{0xCFB88841FA591E94ULL, "tal_tau_cl.mao"},
	{0xCFB8C3AC5DCB6ACEULL, "art\\characters\\weapons\\staff\\textures\\w_stf_002a_0t.dds"},
	{0xCFB8CCD67D6E9248ULL, "802018726.ogg"},
	{0xCFB94C8BC8EE1E3CULL, "6156224_m.fxe"},
	{0xCFB95721D6CDE629ULL, "dae000_junar6147859.cl"},
	{0xCFB9C91C43BEA27BULL, "lt_0_lowtown_docks_186_0.msh"},
	{0xCFB9DB1C43BEC0C6ULL, "lt_0_lowtown_docks_186_0.mao"},
	{0xCFB9E71C43BED565ULL, "lt_0_lowtown_docks_186_0.mmh"},
	{0xCFBAA2835A1B8AE1ULL, "vxlt_0_lowtown_docks_a3_n_cyi_0_0.vlm"},
	{0xCFBAAF7E3C2D9E85ULL, "dre000_cut_varric_intro.cnv"},
	{0xCFBABE65D35EB5C8ULL, "l_mcr310ar_lowtown_nt.wbd"},
	{0xCFBBA03442C832FAULL, "vxlt_gallowstemplar_n_5v4_0_0.vlm"},
	{0xCFBD1B0FE9854E89ULL, "lt_0_lowtown_a2_n_c_32_0_lc.mmh"},
	{0xCFBD2FB77A84D551ULL, "6117932_m.fxe"},
	{0xCFBD4AC7E7AEB19BULL, "6201626_m.fxe"},
	{0xCFBDA025CD0EB3B3ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.impa.gad"},
	{0xCFBDA41C499CD56BULL, "dre000_varric6034618.cl"},
	{0xCFBDC4DD9F583FCAULL, "696817438.ogg"},
	{0xCFBDFCBF3E1D76DEULL, "hlac_lt_gallowstemplar_f_2384_2x.dds"},
	{0xCFBE073D174CF534ULL, "vxlt_0_lowtown_docks_bc3_0_0.vlm"},
	{0xCFBE0AC148BDAD9CULL, "vxht_bartrand_n_64c_0_0.vlm"},
	{0xCFBEA1DDAF2153EDULL, "vxlt_0_lowtown_docks_a3_n_d41_0_0.vlm"},
	{0xCFBFAE00D54085ABULL, "6164459_m.fxe"},
	{0xCFBFE42134410AB9ULL, "vxht_mansion2_2ai_0_0.vlm"},
	{0xCFBFF966FA5A83C4ULL, "zz_vault_debug6203325.cl"},
	{0xCFC08002D97E3528ULL, "8890870.ogg"},
	{0xCFC08C449B7E851BULL, "135775450.ogg"},
	{0xCFC14CEDDEB53D80ULL, "llac_lt_gallowstemplar_2383_0.dds"},
	{0xCFC1ABCFB53CC8F3ULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.dura.ani"},
	{0xCFC1F3DFFB2C6790ULL, "gen00fl_anders6119227.cl"},
	{0xCFC293DF52485A37ULL, "da2_prc_seb_fr-fr.cre"},
	{0xCFC2E197679B7223ULL, "dre000_bodahn6031283.cl"},
	{0xCFC309FC116F37C2ULL, "mcr370_negotiation6078726.cl"},
	{0xCFC32C8AEE819126ULL, "444974697.ogg"},
	{0xCFC34B2874AF049BULL, "6185094_m.fxe"},
	{0xCFC3A5D660256A6DULL, "w_woundedcoast_20_0.mmh"},
	{0xCFC3A9D6602570BEULL, "w_woundedcoast_20_0.mao"},
	{0xCFC3B7D6602588F3ULL, "w_woundedcoast_20_0.msh"},
	{0xCFC3E2231C3BBD58ULL, "6173104_m.fxe"},
	{0xCFC4104F1363F321ULL, "6127001_m.fxe"},
	{0xCFC42B391B7036C8ULL, "vxlt_0_lowtown_docks_a3_n_cyl_0_0.vlm"},
	{0xCFC4424841DB7EF5ULL, "zz_isa_debug6133208.cl"},
	{0xCFC452CC0041D7ACULL, "hlac_ht_chantry_2534_51y.dds"},
	{0xCFC48648C2FAD48BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_dssta_0s.dds"},
	{0xCFC495B8C6DF220DULL, "6111601_f.fxe"},
	{0xCFC49D25CD14CD18ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.impa.evt"},
	{0xCFC4F630802B6CDFULL, "zz_approval_debug6003523.cl"},
	{0xCFC5625D39BC70F9ULL, "gen00fl_varric6080454.cl"},
	{0xCFC63457716EA64DULL, "vxlt_warehouse_n_cf_0_0.vlm"},
	{0xCFC6547C1FDCA5F3ULL, "llac_lt_0_lowtown_a3_n_c_65536_1.dds"},
	{0xCFC690ABD81D33D3ULL, "vxca_caverns2prc_9l5_0_0.vlm"},
	{0xCFC6971DBD51E2BDULL, "6089525_m.fxe"},
	{0xCFC69CE2DC909B77ULL, "vxca_caverns1_mines_jfk_0_0.vlm"},
	{0xCFC830D12B5783DAULL, "tow_01.mao"},
	{0xCFC834397E827981ULL, "hlac_dr_primevaltomb_658_13x.dds"},
	{0xCFC8D89458054080ULL, "6206887_f.fxe"},
	{0xCFC8F5BBEC30DC42ULL, "dae000_bartrand6206924.cl"},
	{0xCFC8FD62E18736D1ULL, "vxw_sundermount_35e_0_1.vlm"},
	{0xCFC988027C3F1CE0ULL, "pc_immortal.ncs"},
	{0xCFC9E1E6488CC7E0ULL, "vxca_caverns1_e6z_0_0.vlm"},
	{0xCFCA11DDF3DE7414ULL, "llac_lt_0_lowtown_docks_a3_65534_6.dds"},
	{0xCFCA3D30AFEF5D63ULL, "vxht_0_hightown_a3_c5s_0_0.vlm"},
	{0xCFCA4C2C951AE576ULL, "vxlt_warehouse_o8_0_0.vlm"},
	{0xCFCAE28C92C7BC0CULL, "6113991_m.fxe"},
	{0xCFCB245EB9C8EFCBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen.ani"},
	{0xCFCB89F190FC8B79ULL, "6072403_f.fxe"},
	{0xCFCBA95CFD83980CULL, "6099082_m.fxe"},
	{0xCFCBBB306D6AE58FULL, "6036531_f.fxe"},
	{0xCFCBE888A902034CULL, "dae000_worthy6166618.cl"},
	{0xCFCC2CE0345EBD22ULL, "lgt160cr_wk_avern_bandit_f.utc"},
	{0xCFCC4FA94E284AD4ULL, "vxht_brothel_n_5mg_0_0.vlm"},
	{0xCFCC791A09430DB4ULL, "vxlt_0_lowtown_docks_bf0_0_0.vlm"},
	{0xCFCC97E06A400290ULL, "gen_im_npc_wep_shd_gend.uti"},
	{0xCFCCAB657F00DD7BULL, "vxlt_0_lowtown_docks_a3_d2f_0_0.vlm"},
	{0xCFCDA5D6602DBEE6ULL, "w_woundedcoast_20_0.rml"},
	{0xCFCDB36D0AE4FFCEULL, "zz_mag_debug6024435.cl"},
	{0xCFCDBB119541EA47ULL, "6125698_m.fxe"},
	{0xCFCE55FE26372339ULL, "vxht_mansion2_2jh_0_0.vlm"},
	{0xCFCE8167B74483D8ULL, "mag110_vincento6016315.cl"},
	{0xCFCE85B32FD5F701ULL, "vxht_keep_6f4_0_0.vlm"},
	{0xCFCE85E7FB7FA302ULL, "vxht_0_hightown_n_a2s_0_0.vlm"},
	{0xCFCEA3CFB547452AULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.dura.evt"},
	{0xCFCF28079C270B21ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_oldmanwalkf_1_ian.gad"},
	{0xCFCF87551945D3A8ULL, "vxht_keep_n_69s_0_0.vlm"},
	{0xCFCF924B4B20B5D6ULL, "hlac_lt_0_lowtown_a3_n_c_131073_2x.dds"},
	{0xCFD0261A16910BA4ULL, "w_encounter1_30_0.msh"},
	{0xCFD0293C37712284ULL, "qcr000_arishok6060000.cl"},
	{0xCFD0381A16912A0DULL, "w_encounter1_30_0.mao"},
	{0xCFD03C1A16913146ULL, "w_encounter1_30_0.mmh"},
	{0xCFD09F66D260D607ULL, "zz_area_signage2.utp"},
	{0xCFD0A593FC82E2DBULL, "hlac_lt_0_lowtown_a2_n_c_-262144_1x.dds"},
	{0xCFD0F1510498B159ULL, "359551827.ogg"},
	{0xCFD1F6D530BC04AFULL, "vxca_caverns1_mines_edn_0_0.vlm"},
	{0xCFD1F6DBAB722B87ULL, "6089497_m.fxe"},
	{0xCFD2AE20974DBEFBULL, "hlac_lt_0_lowtown_n_131072_1z.dds"},
	{0xCFD30389DA59C15CULL, "qun120_arishok6052907.cl"},
	{0xCFD33958BE22CC21ULL, "flame.mao"},
	{0xCFD33A767FB570B2ULL, "vrc221_bartrand6042865.cl"},
	{0xCFD3A0B02E3D211FULL, "94849250.ogg"},
	{0xCFD3FC4129B7BBD5ULL, "qun110_petrice6018599.cl"},
	{0xCFD44C56EE5731F7ULL, "6110489_m.fxe"},
	{0xCFD47FDCAFA1E3DDULL, "hlac_lt_0_lowtown_docks_-65537_1y.dds"},
	{0xCFD50DF4A466AA90ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_6x.dds"},
	{0xCFD52F22EADB37CEULL, "lt_0_lowtown_docks_176_0.msh"},
	{0xCFD53D22EADB4F1FULL, "lt_0_lowtown_docks_176_0.mao"},
	{0xCFD54922EADB63BCULL, "lt_0_lowtown_docks_176_0.mmh"},
	{0xCFD579E1719E8EBFULL, "lgt120_orz_bln_official6176125.cl"},
	{0xCFD582CFB54D2B9DULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.dura.gad"},
	{0xCFD58A4513884E39ULL, "vxca_cavern1_haunted_iy3_0_0.vlm"},
	{0xCFD5DB8E39AD3CF3ULL, "gen_im_trs_gen_med_boo_r1.uti"},
	{0xCFD5E9AC47D6CF44ULL, "663464629.ogg"},
	{0xCFD629079C2D2BCFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_oldmanwalkf_1_ian.ani"},
	{0xCFD6388BF9DA05ADULL, "6174067_m.fxe"},
	{0xCFD63DB2A431FAB6ULL, "one220_yevhen6105680.cl"},
	{0xCFD663E8904458F0ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_109a_n.dds"},
	{0xCFD68874BC0E9646ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_exit.evt"},
	{0xCFD68EC829F85A96ULL, "71215884.ogg"},
	{0xCFD6DB01B0CB72B8ULL, "one100dg_vanard6017490.cl"},
	{0xCFD72B0FA010A5CDULL, "hlac_w_encounter1_-65538_0z.dds"},
	{0xCFD7713A5E3E3F5CULL, "vxht_keep_6ix_0_0.vlm"},
	{0xCFD7CD84B5FA872CULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_sylvesta.ani"},
	{0xCFD8B07C80420D9EULL, "vxca_cavern1_haunted_inq_0_0.vlm"},
	{0xCFD8C94F022FC025ULL, "static_static_p48.psh.fxo"},
	{0xCFD8D30565D7C1AEULL, "lt_0_lowtown_docks_204_0.msh"},
	{0xCFD8D6F0A13FAB3CULL, "fen110_fenris6097049.cl"},
	{0xCFD8E029DBF83EE5ULL, "6127034_m.fxe"},
	{0xCFD8E10565D7D9FFULL, "lt_0_lowtown_docks_204_0.mao"},
	{0xCFD8ED0565D7ED9CULL, "lt_0_lowtown_docks_204_0.mmh"},
	{0xCFD9012ED217F3CFULL, "6116083_m.fxe"},
	{0xCFD90C5EB9D50442ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen.evt"},
	{0xCFD91882F1F6FE95ULL, "6080452_m.fxe"},
	{0xCFD96E14E54A3125ULL, "dae300_alistair_king6133601.cl"},
	{0xCFDA075EB921DCBDULL, "hlac_ht_bartrand_n_3382_0z.dds"},
	{0xCFDA103BA077CD37ULL, "art\\levels\\races\\design\\_textures\\prp_rubble_d.dds"},
	{0xCFDA111ADDE81571ULL, "6074059_m.fxe"},
	{0xCFDA145E0A9FFEE0ULL, "vxlt_gamlenshouse_n_k9_0_0.vlm"},
	{0xCFDA6CA6D736259DULL, "6098949_m.fxe"},
	{0xCFDA9394BB5CB891ULL, "6151000_m.fxe"},
	{0xCFDACC3CA69E67F9ULL, "vxht_0_hightown_akd_0_0.vlm"},
	{0xCFDB4F5000E0FA50ULL, "hlac_ca_caverns2prc_5802_35z.dds"},
	{0xCFDB722608A318ABULL, "hlac_lt_0_lowtown_a2_n_-1_8y.dds"},
	{0xCFDCB0284DEEDB57ULL, "vxht_0_hightown_a2_b7m_0_0.vlm"},
	{0xCFDCB2A79B207E5CULL, "vxht_0_hightown_n_bfv_0_0.vlm"},
	{0xCFDD330166AE8059ULL, "vxht_mansion2b_2iy_0_0.vlm"},
	{0xCFDD865596E15C49ULL, "one220_yevhen6099027.cl"},
	{0xCFDD9F74BC14DB91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_exit.gad"},
	{0xCFDDD8556C3035D2ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha12a.phy"},
	{0xCFDE0D380CB3F69FULL, "964454225.ogg"},
	{0xCFDE6B989E07CF63ULL, "vxdr_0_deeproads_dqv_0_0.vlm"},
	{0xCFDEDA84B600BC6EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_sylvesta.gad"},
	{0xCFDF0B5EB9D96E55ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosedtoopen.gad"},
	{0xCFDF9FBEFBCA9A02ULL, "6091996_m.fxe"},
	{0xCFDFBADEE4C48819ULL, "vxlt_gallowscourtyard_4xi_0_0.vlm"},
	{0xCFE01B5410AACE7EULL, "qcr220_grey_wardens6135918.cl"},
	{0xCFE04664A184C7C7ULL, "mag120st_macha.stg"},
	{0xCFE0B441376BB31BULL, "876468828.ogg"},
	{0xCFE0C24B391E5948ULL, "hlac_lt_0_lowtown_a2_n_c_-65537_3x.dds"},
	{0xCFE0DB850DFDF0B0ULL, "vxca_caverns1_i9g_0_0.vlm"},
	{0xCFE13C1A169FA40DULL, "w_encounter1_30_0.rml"},
	{0xCFE14CC68E9F530DULL, "777631028.ogg"},
	{0xCFE1BC52BC8D3CAEULL, "vxw_bonepit_1hy_0_0.vlm"},
	{0xCFE1CB8B72743BAAULL, "qun131tr_spiders.utt"},
	{0xCFE1E3D85CF03DDEULL, "character_character_p265.psh.fxo.dx11.assembly"},
	{0xCFE212034EC69B9CULL, "gen00fl_merrill6108487.cl"},
	{0xCFE21B7B82578515ULL, "6172595_m.fxe"},
	{0xCFE35FC76BCDD341ULL, "vxw_bonepitmining_a2_218_0_0.vlm"},
	{0xCFE37394E3AE7EC3ULL, "pro000_meeran6050095.cl"},
	{0xCFE37AD7A13E13BDULL, "6087725_m.fxe"},
	{0xCFE4636A3474E892ULL, "6164317_m.fxe"},
	{0xCFE4A074BC1AFC3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\anders_intro_exit.ani"},
	{0xCFE4DB70C412DE18ULL, "hlac_lt_0_lowtown_65536_6x.dds"},
	{0xCFE57139568BFA2DULL, "6031852_m.fxe"},
	{0xCFE64E180C3E7071ULL, "mag291_demon_marethari6027940.cl"},
	{0xCFE6CBD20FC9DCBFULL, "gen00fl_merrillact2_fadei.stg"},
	{0xCFE6CD4DCABEBCDCULL, "6027560_m.fxe"},
	{0xCFE7284A379C47E7ULL, "qun211cr_braggart_ally_f.utc"},
	{0xCFE72E73C3258F68ULL, "6110544_f.fxe"},
	{0xCFE7BAF41885A43DULL, "vxw_encounter1_as_0_0.vlm"},
	{0xCFE882F7CFDE2DF1ULL, "6079359_m.fxe"},
	{0xCFE8F715C4B47EC9ULL, "fen110_fenris6118080.cl"},
	{0xCFEA881082F7258CULL, "vxht_bartrand_n_5xt_0_0.vlm"},
	{0xCFEAC05E22D295C5ULL, "340352075.ogg"},
	{0xCFEB0A780B9B1338ULL, "6162245_m.fxe"},
	{0xCFEB22053FE4F3D7ULL, "vxht_0_hightown_a2_b8l_0_0.vlm"},
	{0xCFEB24848D1696DCULL, "vxht_0_hightown_n_baw_0_0.vlm"},
	{0xCFEB2678F0A89696ULL, "6154750_m.fxe"},
	{0xCFEB39B80A235439ULL, "vxht_mansion2_2pn_0_0.vlm"},
	{0xCFEB5D01D039FED0ULL, "dae000_hate_follower_react6103211.cl"},
	{0xCFEB8B9896ACA5F1ULL, "6101175_m.fxe"},
	{0xCFEB92363472C2B3ULL, "hlac_dr_slavers_3214_13y.dds"},
	{0xCFEBA4DE58A498D9ULL, "vxht_mansion2b_2bx_0_0.vlm"},
	{0xCFEBA9F996C43CEEULL, "26031039.ogg"},
	{0xCFEC3EFF7E7A5E9BULL, "genip_flower.utp"},
	{0xCFEC413FB52F983BULL, "zz_isa_debug6133233.cl"},
	{0xCFEC85CB9569A4D6ULL, "vxht_bartrand_n_685_0_0.vlm"},
	{0xCFECAF1DAA5CF444ULL, "149613165.ogg"},
	{0xCFECB5C7A61A5CF1ULL, "6168109_m.fxe"},
	{0xCFEDA90DCF49F877ULL, "t3_wdg_tr3.tnt"},
	{0xCFEDB4EDF32CD79FULL, "1065866953.ogg"},
	{0xCFEDC0796FD05662ULL, "6078756_f.fxe"},
	{0xCFEDE39F3AEAA0BBULL, "6120899_m.fxe"},
	{0xCFEE119BEDC0B282ULL, "6096896_m.fxe"},
	{0xCFEE5B5859C973BEULL, "llac_w_blightlands_0_12.dds"},
	{0xCFEEA9E7692EEEB8ULL, "132144621.ogg"},
	{0xCFEEDA958F5E5C87ULL, "6087697_m.fxe"},
	{0xCFEEDBE38E113430ULL, "6171814_m.fxe"},
	{0xCFEF8423AC290D31ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_14gallowbird.mmh"},
	{0xCFEF9023AC2921D2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_14gallowbird.mao"},
	{0xCFEF9E23AC2939EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_14gallowbird.msh"},
	{0xCFF02E2FAE83552EULL, "vxw_bonepit_1mz_0_0.vlm"},
	{0xCFF030BA4E22B684ULL, "vxca_caverns2prc_4ap_0_0.vlm"},
	{0xCFF08A9F245BEDC8ULL, "mcr371_hostile_anders6082127.cl"},
	{0xCFF0911DBA824BD1ULL, "rune_armor_lucky.dds"},
	{0xCFF09D7611B01897ULL, "471711923.ogg"},
	{0xCFF0B5E6209DEA69ULL, "150256345.ogg"},
	{0xCFF0F81265CE4814ULL, "mag210_arianni6024986.cl"},
	{0xCFF1023A71CB8FF8ULL, "docks_scene_dockworkers6176204.cl"},
	{0xCFF1693414FCB434ULL, "vxlt_0_lowtown_a3_n_c_ehl_0_0.vlm"},
	{0xCFF189EE4750EFCDULL, "vxlt_warehouse_n_pc_0_0.vlm"},
	{0xCFF1B347FA77F39BULL, "vxlt_gallowstemplar_n_3jc_0_0.vlm"},
	{0xCFF1CBF9507EF057ULL, "2200436.ogg"},
	{0xCFF1D4BBBD6536F4ULL, "vxht_0_hightown_a2_asy_0_0.vlm"},
	{0xCFF1F491177EA61BULL, "359075532.ogg"},
	{0xCFF25834865E4ABAULL, "vxht_mansion2_24u_0_0.vlm"},
	{0xCFF260F26651FEE0ULL, "vxw_bonepitmining_a2_1iv_0_0.vlm"},
	{0xCFF287EE8BADD9DCULL, "6075105_m.fxe"},
	{0xCFF28A5FA2EB2A4FULL, "hlac_ca_cavern1_haunted_24423_6x.dds"},
	{0xCFF29CED585A76CEULL, "llac_ht_keep_3736_0.dds"},
	{0xCFF3DD8B029024BEULL, "6205599_m.fxe"},
	{0xCFF4C9C186C0219DULL, "405463111.ogg"},
	{0xCFF580FA1BE09749ULL, "6122214_f.fxe"},
	{0xCFF584EE9009AD0CULL, "pro000st_nina_saved.stg"},
	{0xCFF5A2EFCC408A69ULL, "hlac_lt_0_lowtown_n_-65536_6y.dds"},
	{0xCFF5C644901C662FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgtb_0n.dds"},
	{0xCFF5F4FBB2A3AFDBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomeb_d.dds"},
	{0xCFF5F8F90D0C44ADULL, "vxlt_undercityencounter_1ke_0_0.vlm"},
	{0xCFF6055D9F26D3FDULL, "vxw_blightlands_ss_0_0.vlm"},
	{0xCFF6192865F01DB7ULL, "vxdr_slavers_142_0_0.vlm"},
	{0xCFF62CD10A7BBCBDULL, "vxw_encounter1_fr_0_0.vlm"},
	{0xCFF7279E6EDA9EB8ULL, "6108586_m.fxe"},
	{0xCFF73153F2253CCCULL, "hlac_w_encounter2_65537_0x.dds"},
	{0xCFF7D80BDB7C86CCULL, "vxht_0_hightown_n_cbw_0_0.vlm"},
	{0xCFF835FEFAC6FB32ULL, "6167994_m.fxe"},
	{0xCFF846CBD8DBC912ULL, "6103233_m.fxe"},
	{0xCFF8AAB8F95F6AD0ULL, "nat380tr_final_battle.utt"},
	{0xCFF955D70944C7A6ULL, "vxw_woundedcoast_1hs_0_1.vlm"},
	{0xCFF996617F0CAF5CULL, "vxht_0_hightown_n_blt_0_0.vlm"},
	{0xCFF9A68B72888E93ULL, "qun131tr_spiders.ncs"},
	{0xCFF9E6AD7726A952ULL, "6172310_m.fxe"},
	{0xCFFA10103506C6ABULL, "vxlt_0_lowtown_a3_n_c_i14_0_0.vlm"},
	{0xCFFA8CA206574796ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_right.evt"},
	{0xCFFAEF76D9650E0DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxa.msh"},
	{0xCFFAFD76D9652640ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxa.mao"},
	{0xCFFB0176D9652C93ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxa.mmh"},
	{0xCFFB169364D8BE98ULL, "vxlt_warehouse_19a_0_0.vlm"},
	{0xCFFB24282883805CULL, "7524208.ogg"},
	{0xCFFB9EFD65561B2FULL, "gen_im_cft_reg_e.uti"},
	{0xCFFBA655D698C814ULL, "343096748.ogg"},
	{0xCFFBD79529BACA81ULL, "vxca_cave_1b2_0_1.vlm"},
	{0xCFFBE8A16ED6F2ACULL, "hlac_lt_0_lowtown_a2_65533_3z.dds"},
	{0xCFFC20183678EAF8ULL, "6107231_m.fxe"},
	{0xCFFC6A4C48F3F4DFULL, "vxht_0_hightown_a3_b8u_0_0.vlm"},
	{0xCFFCC9403F39F566ULL, "aip_dalishassassin_base.gda"},
	{0xCFFCCE053A64D87AULL, "lt_0_lowtown_docks_a3_44_0.rml"},
	{0xCFFD1FDAA2DF475EULL, "6059668_m.fxe"},
	{0xCFFD3E5DEBEE8315ULL, "vxlt_warehouse_n_11k_0_0.vlm"},
	{0xCFFDB83AAF2E36AAULL, "6020611_m.fxe"},
	{0xCFFDE608FDB20B38ULL, "856788170.ogg"},
	{0xCFFDEFB0EF084886ULL, "art\\characters\\creatures\\mubari\\animation\\tn90r.evt"},
	{0xCFFE4D28C29365D4ULL, "atl_worldmap_dxt1_dat.xml"},
	{0xCFFE5257F2177F7CULL, "lt_darktown_mmap_ph.dds"},
	{0xCFFE552C46A8132BULL, "vxw_sundermount_30i_0_0.vlm"},
	{0xCFFE6B1784882FE0ULL, "6197199_m.fxe"},
	{0xCFFE8FABB03F125CULL, "vxht_0_hightown_n_9b5_0_0.vlm"},
	{0xCFFF051B2536D5E6ULL, "6141910_m.fxe"},
	{0xCFFF3229E6924243ULL, "6170582_m.fxe"},
	{0xCFFF3DC8D267ACBDULL, "mag170_thrask6024961.cl"},
	{0xCFFFB89A2A1A44B2ULL, "6147530_m.fxe"},
	{0xCFFFC8F85632E5D0ULL, "gen00fl_dog6167252.cl"},
	{0xCFFFDEEE9F5B932EULL, "lgt260_way_blood_mage6148396.cl"},
	{0xCFFFFBCB3947084DULL, "vxlt_warehouse_n_wb_0_0.vlm"},
	{0xD00010774C8083ADULL, "vxlt_0_lowtown_docks_a3_n_bt7_0_0.vlm"},
	{0xD00082E09799A70BULL, "hlac_w_bonepitmining_a2_65536_0z.dds"},
	{0xD000A68FFC8D012BULL, "173686677.ogg"},
	{0xD000CFA6BD4AA071ULL, "577590079.ogg"},
	{0xD000DFDBE96A97B9ULL, "vxca_cavern1_haunted_ih0_0_0.vlm"},
	{0xD0015E9D93EDE51EULL, "gen00fl_aveline6115503.cl"},
	{0xD00183A2065D5721ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_right.gad"},
	{0xD001E465D39AF88AULL, "l_mcr310ar_lowtown_nt.bnk"},
	{0xD001ED3B1A2ED868ULL, "435004022.ogg"},
	{0xD00223AF2361D4E6ULL, "mrl350pt_pridedemon.ncs"},
	{0xD0022E2CBFED9925ULL, "hla_ht_0_hightown_-131074_0x.dds"},
	{0xD0024F67F4863D3EULL, "6208299_m.fxe"},
	{0xD0027B4F977E921DULL, "vxlt_hangedman_n_225_0_0.vlm"},
	{0xD003114D43AB77CCULL, "vxw_bonepitdestroyed_a3_1s6_0_0.vlm"},
	{0xD003732BC258CB47ULL, "dae000_flemeth6013289.cl"},
	{0xD004134F06D10BD3ULL, "vxht_keep_5xo_0_0.vlm"},
	{0xD0045DAB577D0E41ULL, "llac_lt_0_lowtown_docks_n_-196610_4.dds"},
	{0xD004773A911CEC7DULL, "vxw_blightlands_tr_0_0.vlm"},
	{0xD0049EADFC71D53DULL, "vxw_encounter1_kq_0_0.vlm"},
	{0xD004C72B80119CA4ULL, "274591547.ogg"},
	{0xD00506B0EF0E8ED1ULL, "art\\characters\\creatures\\mubari\\animation\\tn90r.gad"},
	{0xD005191157D99C80ULL, "6109035_m.fxe"},
	{0xD0052C9B53E0215CULL, "6169055_m.fxe"},
	{0xD005997AA35EAC7DULL, "vxw_sundermount_2go_0_0.vlm"},
	{0xD005E959E954DC75ULL, "6165159_m.fxe"},
	{0xD005FD1341D8E151ULL, "vxlt_0_lowtown_a2_h26_0_0.vlm"},
	{0xD00605236910CA50ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_e01_0t.dds"},
	{0xD0071181496D3760ULL, "6119131_m.fxe"},
	{0xD0074D8A79464B56ULL, "6078166_m.fxe"},
	{0xD00798FB6234EB58ULL, "vxht_0_hightown_a3_c1p_0_0.vlm"},
	{0xD007C7B3FB3AE026ULL, "vxw_woundedcoast_1sr_0_1.vlm"},
	{0xD007D44A13D137C3ULL, "6060441_m.fxe"},
	{0xD007E170BF383A37ULL, "qun221_petrice6083736.cl"},
	{0xD00884A2066377CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_right.ani"},
	{0xD008E40DF2F48286ULL, "rock_04.mao"},
	{0xD00974151B54F00EULL, "wpnstand_04.mao"},
	{0xD00A0923EBBF0A29ULL, "w_woundedcoast_218_0.phy"},
	{0xD00A25E14DB4AA1DULL, "6155691_m.fxe"},
	{0xD00A478D638DA1C3ULL, "vxht_0_hightown_a3_c3t_0_1.vlm"},
	{0xD00A8188980262C2ULL, "art\\vfx\\combat\\vfx_str_att_back_r_leg_c.mmh"},
	{0xD00A9BCAC3586BA1ULL, "vxlt_gallowscourtyard_7fx_0_0.vlm"},
	{0xD00AA08676D9DC7AULL, "ser120_ghyslain6118430.cl"},
	{0xD00B07B0EF12FC7FULL, "art\\characters\\creatures\\mubari\\animation\\tn90r.ani"},
	{0xD00BAC18BB1B547FULL, "follower_banter6172792.cl"},
	{0xD00BB03ADDE49B95ULL, "vxlt_warehouse_n_16j_0_0.vlm"},
	{0xD00BC4007CF83CABULL, "629263854.ogg"},
	{0xD00CE0D1F5139452ULL, "vxdr_dungeon_1lp_0_1.vlm"},
	{0xD00D9404C825AA52ULL, "vxht_keep_5xj_0_0.vlm"},
	{0xD00E1C4FFF2C7627ULL, "vxht_chantry_2ld_0_0.vlm"},
	{0xD00E1F1638B6E7BBULL, "mag160_feynriel6033392.cl"},
	{0xD00E6DA82B3D20CDULL, "vxlt_warehouse_n_za_0_0.vlm"},
	{0xD00F0AF25D6F9D7CULL, "gen00fl_anders6125992.cl"},
	{0xD00F32504B0E3D9DULL, "zz_vault_debug6203186.cl"},
	{0xD00F7A3F84754DC0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wi_edgebase_1.mao"},
	{0xD00F953B1969EE72ULL, "264252968.ogg"},
	{0xD00FEB9DDDBED49BULL, "114391877.ogg"},
	{0xD0101051D85CB74DULL, "dae000_orana6149047.cl"},
	{0xD0103BF93EB17D84ULL, "fxa_cen_smoke.dds"},
	{0xD01143825C1D93ECULL, "8719836.ogg"},
	{0xD011690A7005A74EULL, "6142216_m.fxe"},
	{0xD012364B352BA429ULL, "120201335.ogg"},
	{0xD012D0678B689901ULL, "vxht_chantry_n_2td_0_1.vlm"},
	{0xD01370A60B36FD7FULL, "gen00fl_aveline6129068.cl"},
	{0xD013BF1B39205ABBULL, "vxca_caverns2_78w_0_1.vlm"},
	{0xD013DA065623CB7FULL, "246782181.ogg"},
	{0xD0140B579554C4FDULL, "vxw_sundermount_2hn_0_0.vlm"},
	{0xD0143D00A54E013AULL, "vxlt_shop_e9_0_0.vlm"},
	{0xD01453C013BBACC9ULL, "gen_im_trs_dem_lgt_hlm_r1.uti"},
	{0xD014C23171A2519EULL, "hlac_w_bonepitdestroyed_a3_65539_0z.dds"},
	{0xD014C32E74D4350DULL, "vxca_cavern1_haunted_eak_0_0.vlm"},
	{0xD0150E64D212B020ULL, "dae000_pickpocket_scene6175099.cl"},
	{0xD015835E3B634FE0ULL, "6114631_m.fxe"},
	{0xD01603939A2B9D2FULL, "210204181.ogg"},
	{0xD0166DCD558BAF76ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.cess.ani"},
	{0xD0167247D77FDE66ULL, "vxlt_shop_cw_0_0.vlm"},
	{0xD016FC9F2BDEBBFAULL, "gen00fl_anders6128377.cl"},
	{0xD0182B7E96B2181AULL, "qun200_party_comment6090408.cl"},
	{0xD019201EAC1A6C0BULL, "vxca_caverns1_iix_0_0.vlm"},
	{0xD01949C7D0F2FFE5ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.mmh"},
	{0xD019765987020A9AULL, "llac_ht_mansion2_818_42.dds"},
	{0xD01983456FB2FE0CULL, "ht_keep_4023.rml"},
	{0xD019B0BE591CBD18ULL, "gen_im_arm_boo_med_act2_01.uti"},
	{0xD019E67F8CE254B3ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_flare03.dds"},
	{0xD01A9E980AA2056AULL, "vxca_caverns2_8k9_0_0.vlm"},
	{0xD01AC33DCF2AA418ULL, "one220_yevhen6103363.cl"},
	{0xD01B1CAA62300680ULL, "ht_0_hightown_a2_53_0.anb"},
	{0xD01B1F4BF73114A8ULL, "w_blightlands_38_0.mmh"},
	{0xD01B2B4BF731294BULL, "w_blightlands_38_0.mao"},
	{0xD01B3D4BF73147D2ULL, "w_blightlands_38_0.msh"},
	{0xD01B53E9D0AD0BD1ULL, "gen00fl_fenris6116573.cl"},
	{0xD01B61A479D15C3CULL, "llac_lt_0_lowtown_docks_a3_n_262143_0.dds"},
	{0xD01B9B137F91050CULL, "vxht_0_hightown_cf9_0_1.vlm"},
	{0xD01C057F14C6AE54ULL, "6129917_f.fxe"},
	{0xD01C165A952358EFULL, "hlac_w_bonepit_393222_0z.dds"},
	{0xD01CB772AEA69DBBULL, "6218974_m.fxe"},
	{0xD01D99B8F97E698DULL, "nat380tr_final_battle.ncs"},
	{0xD01EE477CFF7DD24ULL, "hlac_lt_undercityencounter_-1_42x.dds"},
	{0xD01F334869FCB866ULL, "676953259.ogg"},
	{0xD01F4CAE864202B4ULL, "72511794.ogg"},
	{0xD01F7F7F29114F82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up_thoughtful.gad"},
	{0xD0200FDD55959255ULL, "vxlt_0_lowtown_a3_n_c_ht8_0_0.vlm"},
	{0xD020214B302735A1ULL, "vxca_caverns1_huk_0_0.vlm"},
	{0xD0206DAA21C5D269ULL, "vrc321_party_barks6096260.cl"},
	{0xD02142AF3B9E036FULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfrun_1.mao"},
	{0xD02230A20D0C4E39ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m.mmh"},
	{0xD0227D34874ADD7DULL, "vxw_sundermount_2mm_0_0.vlm"},
	{0xD023409A9BB9D0B5ULL, "126528360.ogg"},
	{0xD02375CD55964767ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.cess.evt"},
	{0xD02386B4A0B6CA39ULL, "6100247_f.fxe"},
	{0xD0239ABA13D8FB62ULL, "528843839.ogg"},
	{0xD023BB6D4C7FD585ULL, "hlac_lt_backalley_g_1_1y.dds"},
	{0xD023D406B3B75B1EULL, "vxlt_gallowscourtyard_7cd_0_0.vlm"},
	{0xD02423E64131057BULL, "follower_banter6155676.cl"},
	{0xD02474C8FBB45062ULL, "llac_lt_gallowstemplar_2384_1.dds"},
	{0xD02475B5B066362DULL, "178847703.ogg"},
	{0xD024E424C975F6E6ULL, "vxlt_shop_fx_0_0.vlm"},
	{0xD025C571966911C5ULL, "vxlt_0_lowtown_docks_n_bfv_0_0.vlm"},
	{0xD025FB9AEB0567DFULL, "vxca_caverns1_mines_dnu_0_0.vlm"},
	{0xD026A27F2917AA79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up_thoughtful.evt"},
	{0xD026CDD76C9FE0E5ULL, "vxdr_primevaltomb_5bv_0_0.vlm"},
	{0xD026E395CEE75817ULL, "70680208.ogg"},
	{0xD026FFEA818D3651ULL, "hlac_ca_cavern1_haunted_24424_11y.dds"},
	{0xD02705F50B2F7EA7ULL, "708162933.ogg"},
	{0xD0270C9A586D0D05ULL, "6083605_f.fxe"},
	{0xD02746C7D0FF37B2ULL, "art\\vfx\\combat\\vfx_r_gn3_markerlight_c.anb"},
	{0xD0279CA684400A77ULL, "hlac_w_bonepitdestroyed_a3_262145_0y.dds"},
	{0xD027A5ECC2CCCEFBULL, "vxlt_undercityencounter_1jt_0_0.vlm"},
	{0xD0285D969961EE62ULL, "185143893.ogg"},
	{0xD0288E1601958ECAULL, "127668095.ogg"},
	{0xD0289A34DFB7935DULL, "vxht_0_hightown_b02_0_0.vlm"},
	{0xD0292DD13CF95080ULL, "atl_partypicke_dxt5_01.dds"},
	{0xD0293956BB426275ULL, "vxlt_gallowstemplar_3rb_0_0.vlm"},
	{0xD029B1B038355D25ULL, "vxlt_0_lowtown_a2_n_ev4_0_0.vlm"},
	{0xD029BC4BA04F5A09ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_loop.gad"},
	{0xD02A27845BC56DD8ULL, "gen00fl_anders6125949.cl"},
	{0xD02A96CD559C9DB0ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.cess.gad"},
	{0xD02ABBA128EB312AULL, "965846301.ogg"},
	{0xD02B94F3A3249662ULL, "vfx_distortionmask_p0.psh.fxo.dx11"},
	{0xD02BAE06E2887128ULL, "6059778_m.fxe"},
	{0xD02BCE6E9B999425ULL, "llac_lt_0_lowtown_a2_n_c_0_9.dds"},
	{0xD02C7F33C4BB3BBBULL, "one110_pryce_st.stg"},
	{0xD02CC610D70F41D1ULL, "fen110_fenris6096207.cl"},
	{0xD02CCD816810EF18ULL, "6178453_m.fxe"},
	{0xD02D1BA48B3BA529ULL, "466917240.ogg"},
	{0xD02D33B0E9A2826BULL, "gen00fl_varric6033559.cl"},
	{0xD02DDF988BA3C9F6ULL, "811034483.ogg"},
	{0xD02DE900EC8D940DULL, "zz_gregtest2.cut"},
	{0xD02E167549C64631ULL, "fx_sb_tear.dds"},
	{0xD02E66E4198DC14CULL, "vxw_bonepitdestroyed_a3_1b5_0_0.vlm"},
	{0xD02E74E17A8059E0ULL, "8844765.ogg"},
	{0xD02E9164C1392D5CULL, "6068522_f.fxe"},
	{0xD02ECE35F066D756ULL, "650786022.ogg"},
	{0xD02FBDB415A18318ULL, "218371628.ogg"},
	{0xD02FCEC94700F4A7ULL, "lt_0_lowtown_a3_n_c_29_0.rml"},
	{0xD0303FD23BA3B9CEULL, "vxdr_primevalentrance_21w_0_0.vlm"},
	{0xD030812E47D1E7A9ULL, "hlac_lt_0_lowtown_docks_-196609_1x.dds"},
	{0xD0308D4BA0552957ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_1_loop.ani"},
	{0xD030D4718F9DA493ULL, "vxlt_0_lowtown_ewv_0_0.vlm"},
	{0xD0313B1A420FA474ULL, "172133270.ogg"},
	{0xD0316CB3EBE2F654ULL, "vxht_brothel_n_52n_0_0.vlm"},
	{0xD0321E4104612C12ULL, "569085652.ogg"},
	{0xD03232A62DA1E6A2ULL, "6021216_m.fxe"},
	{0xD0329472FF15B7F6ULL, "vxlt_blackemporium_139_0_0.vlm"},
	{0xD032EC489B9E150BULL, "mag110cr_sailor_4.utc"},
	{0xD032FFE236F5E7BEULL, "544551576.ogg"},
	{0xD03306CB5C99EDF6ULL, "hlac_lt_0_lowtown_docks_a3_-131074_2z.dds"},
	{0xD0330811673E9887ULL, "dae120st_merch_armor.utm"},
	{0xD0330B2ADDBF782FULL, "6154210_m.fxe"},
	{0xD033808898256669ULL, "art\\vfx\\combat\\vfx_str_att_back_r_leg_c.anb"},
	{0xD033ADB28FB57259ULL, "84524257.ogg"},
	{0xD034374E885F2A45ULL, "vxlt_0_lowtown_docks_n_bcu_0_0.vlm"},
	{0xD03454AB11DF8332ULL, "341904120.ogg"},
	{0xD03462C2FB37CE9BULL, "6201477_m.fxe"},
	{0xD0347C2AF8788B65ULL, "6035935_m.fxe"},
	{0xD034AA7F2923F568ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_up_thoughtful.ani"},
	{0xD035F223EBE41E4EULL, "w_woundedcoast_218_0.mmh"},
	{0xD035FE23EBE432E5ULL, "w_woundedcoast_218_0.mao"},
	{0xD0360C23EBE44A3CULL, "w_woundedcoast_218_0.msh"},
	{0xD036777B2755D8B5ULL, "lgt260_eau_tmp_roderick6143727.cl"},
	{0xD0371938AA0F919CULL, "344574784.ogg"},
	{0xD0374F92230DB761ULL, "vxht_mansion2b_3cr_0_0.vlm"},
	{0xD037A65E14AE66EDULL, "mrl310_merrill6120357.cl"},
	{0xD037B65BEC0B63CDULL, "mer320_hubert6089961.cl"},
	{0xD037F9A44FD028BCULL, "llac_lt_0_lowtown_a2_1_4.dds"},
	{0xD0380EAC370887E0ULL, "824433527.ogg"},
	{0xD0382405660C798FULL, "vxlt_0_lowtown_n_gbt_0_0.vlm"},
	{0xD0385FB10B98037EULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\ready.ani"},
	{0xD038ADB396081023ULL, "ico_armor_light_2.dds"},
	{0xD038BA72D6782551ULL, "6208365_f.fxe"},
	{0xD03966BF514AC2E2ULL, "ss_isaachawke3.cut"},
	{0xD039A1CDBB45A8DDULL, "w_woundedcoast_131_0.phy"},
	{0xD039C8BAD6684B1EULL, "one300_zevran6075400.cl"},
	{0xD03A264164D2D6E2ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_fire01a.mao"},
	{0xD03A58270E5AD94EULL, "vxlt_hangedman_10h_0_0.vlm"},
	{0xD03A9A4320207960ULL, "vxw_bonepitmining_a2_1rs_0_0.vlm"},
	{0xD03B3EB1906F1B4AULL, "one100dg_kelder6055598.cl"},
	{0xD03C610D4B184BF2ULL, "vxw_encounter2_camp_jd_0_0.vlm"},
	{0xD03CC1177C9C8669ULL, "vxlt_undercityencounter_z0_0_1.vlm"},
	{0xD03CC4A0B7572AB1ULL, "gen00fl_isabela.utc"},
	{0xD03CED08FA1F7B24ULL, "6079567_m.fxe"},
	{0xD03DB55C62D4D290ULL, "vxlt_0_lowtown_a2_goj_0_1.vlm"},
	{0xD03DD5BDFA50F392ULL, "llac_w_bonepitmining_a2_0_0.dds"},
	{0xD03E063DBACD4E92ULL, "vxlt_undercitydungeon_1l6_0_0.vlm"},
	{0xD03E435020AD4085ULL, "6162068_m.fxe"},
	{0xD03EA02714EEAFB8ULL, "zz_ave_debug6064142.cl"},
	{0xD03EC1A1906CB047ULL, "6115356_m.fxe"},
	{0xD03EE5558777D9A5ULL, "hlac_dr_primevaltomb_664_2y.dds"},
	{0xD03EEC63849EE235ULL, "dae000ip_crowd_mf5_talk13.utp"},
	{0xD03F2D0838EA34DDULL, "and341tr_spider_ambush.ncs"},
	{0xD03F4493C7264FE6ULL, "38578665.ogg"},
	{0xD03F464E8193BD13ULL, "vxlt_0_lowtown_ery_0_0.vlm"},
	{0xD03F5EFE50D95AF7ULL, "w_bonepitdestroyed_a3_250_0.msh"},
	{0xD03F70FE50D9795AULL, "w_bonepitdestroyed_a3_250_0.mao"},
	{0xD03F74FE50D98069ULL, "w_bonepitdestroyed_a3_250_0.mmh"},
	{0xD03F9FE9CC837055ULL, "764336105.ogg"},
	{0xD03FAB74CF7693D5ULL, "6039957_m.fxe"},
	{0xD03FBE6E38646885ULL, "10851648.ogg"},
	{0xD040D7F17737C4B8ULL, "qcr220_orsino6041909.cl"},
	{0xD040D847E4626348ULL, "vxht_mansion2b_3cw_0_0.vlm"},
	{0xD040E2C1AF508B8AULL, "gen00fl_isabela6106182.cl"},
	{0xD04101BB9CE5C7B5ULL, "6075579_m.fxe"},
	{0xD04109E3E7029334ULL, "6068504_f.fxe"},
	{0xD0419F800422D7B5ULL, "vxht_mansion2_1zs_0_0.vlm"},
	{0xD041D20E97E182CEULL, "236022946.ogg"},
	{0xD041E037A126F740ULL, "vxlt_0_lowtown_a3_n_c_ei7_0_0.vlm"},
	{0xD041FD39AB071906ULL, "vxlt_undercityencounter_13r_0_0.vlm"},
	{0xD0433A22A0543766ULL, "317224430.ogg"},
	{0xD0439090B651F8A2ULL, "ser270_emeric6113635.cl"},
	{0xD043B988A2E291F8ULL, "ser221_gascard6122446.cl"},
	{0xD0445288D0733A6DULL, "6046911_m.fxe"},
	{0xD044970BEB1239ECULL, "hightown_market_guard_sta6169583.cl"},
	{0xD044F00ED93BB47FULL, "6034471_m.fxe"},
	{0xD04513F7BB6916F2ULL, "vxht_keep_n_538_0_0.vlm"},
	{0xD04591255EF1A3A0ULL, "hlac_ht_mansion2b_818_61y.dds"},
	{0xD0459305380D2420ULL, "289678688.ogg"},
	{0xD045A58F0AD7ADF3ULL, "vxlt_0_lowtown_n_byd_0_0.vlm"},
	{0xD045D32072C77D51ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_min_c.impa.ani"},
	{0xD04677B10BA4694FULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\ready.evt"},
	{0xD0467B46FA13BEC4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprails.msh"},
	{0xD0468D46FA13DD2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprails.mao"},
	{0xD0469146FA13E466ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprails.mmh"},
	{0xD04732187E6761B2ULL, "zz_dae_debug6049956.cl"},
	{0xD047631AEB31FEEDULL, "vxlt_gallowstemplar_f_45o_0_0.vlm"},
	{0xD047A64441DE16F1ULL, "364062798.ogg"},
	{0xD047DC00D24277CAULL, "vxw_bonepitdestroyed_a3_268_0_0.vlm"},
	{0xD047DC5D5A176EABULL, "vxlt_0_lowtown_docks_a3_d26_0_0.vlm"},
	{0xD047F0770B4C8252ULL, "hm_qmq_minion7.mor"},
	{0xD0483629D6B0DC0DULL, "dae000ip_crowd_lm5_talk17.utp"},
	{0xD048B0ED24A410BDULL, "vxlt_hangedman_1vi_0_0.vlm"},
	{0xD048B408224DBD1EULL, "vxdr_primevaltomb_80j_0_0.vlm"},
	{0xD048D6ABC4D075F4ULL, "6129473_m.fxe"},
	{0xD04913D6C17564F7ULL, "vxht_keep_688_0_0.vlm"},
	{0xD049288B7AB2BECCULL, "vxdr_primevaltomb_8n6_0_0.vlm"},
	{0xD049AFBAC68DECF3ULL, "vxlt_0_lowtown_a2_n_c_e1u_0_0.vlm"},
	{0xD04A0916C345BF57ULL, "vxw_sundermount_340_0_1.vlm"},
	{0xD04A66CFD766B812ULL, "hidden_cursor.cur"},
	{0xD04A76EDF7603302ULL, "vxca_caverns2prc_8ol_0_0.vlm"},
	{0xD04A96CDBB540949ULL, "w_woundedcoast_131_0.mao"},
	{0xD04A9FFCBBE70CA6ULL, "vxca_caverns1_mines_iof_0_0.vlm"},
	{0xD04AA2CDBB541DE2ULL, "w_woundedcoast_131_0.mmh"},
	{0xD04AA4CDBB542098ULL, "w_woundedcoast_131_0.msh"},
	{0xD04ADE9A7B603679ULL, "vrc321ip_spooky3_obj_1.utp"},
	{0xD04B26832A05F462ULL, "6101961_m.fxe"},
	{0xD04B4AA3E5178A71ULL, "vxht_mansion2_2aq_0_0.vlm"},
	{0xD04B4F77A3CA10A0ULL, "hlac_lt_0_lowtown_docks_a3_-196610_3z.dds"},
	{0xD04B89C26A836873ULL, "hlac_lt_0_lowtown_a3_n_c_1_1z.dds"},
	{0xD04BB3B055D2AAA4ULL, "zz_approval_debug6003795.cl"},
	{0xD04BDE2072CBFBF3ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_min_c.impa.gad"},
	{0xD04C4A192F4E47A3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_redflower.msh"},
	{0xD04C50B23973AB62ULL, "6168915_m.fxe"},
	{0xD04C58192F4E5FFDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_redflower.mmh"},
	{0xD04C6179859B43ADULL, "dae021_sister_alema_mob6143685.cl"},
	{0xD04C722FB238A34AULL, "6084349_m.fxe"},
	{0xD04CA17627B0E059ULL, "pro000_athenril6048765.cl"},
	{0xD04CC77BD3E9C448ULL, "zz_and_debug6176462.cl"},
	{0xD04D78B10BAA89F8ULL, "art\\characters\\creatures\\gibgolem\\animation\\armless\\ready.gad"},
	{0xD04D8C3552479A97ULL, "647247508.ogg"},
	{0xD04DB595C5B89D23ULL, "zz_mag_debug6039838.cl"},
	{0xD04DB82B7389D593ULL, "vxlt_0_lowtown_emx_0_0.vlm"},
	{0xD04DD2D448A6B751ULL, "6147915_m.fxe"},
	{0xD04E023A60B9BF27ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.dura.gad"},
	{0xD04E506DCFCF2754ULL, "vxht_brothel_n_58l_0_0.vlm"},
	{0xD04E79DE8AE9EA34ULL, "vxlt_0_lowtown_docks_b59_0_0.vlm"},
	{0xD04E839BFD969285ULL, "mag360_thrask6043912.cl"},
	{0xD04E88308B42392EULL, "hlac_lt_gallowscourtyard_n_-65538_1z.dds"},
	{0xD04EB30F5C6220D5ULL, "dre000_dougal6131002.cl"},
	{0xD04FA1B62AF771C1ULL, "vxht_mansion2b_2og_0_0.vlm"},
	{0xD04FAF2D9D006178ULL, "vxdr_primevaltomb_2ow_0_0.vlm"},
	{0xD0508846FA1C2931ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_steprails.phy"},
	{0xD050DFA8A6B531C0ULL, "vxht_keep_n_3ib_0_0.vlm"},
	{0xD051530A85530B50ULL, "vxw_encounter2_26_0_0.vlm"},
	{0xD05252DDED9AC607ULL, "dae000_surprises6157087.cl"},
	{0xD052D309C30E5DD9ULL, "hlac_lt_0_lowtown_a2_n_65536_10y.dds"},
	{0xD052DB2072D21558ULL, "art\\vfx\\combat\\vfx_glb_nat_imp_min_c.impa.evt"},
	{0xD052E535439A1406ULL, "478587593.ogg"},
	{0xD0531CCAC692FF9BULL, "6140633_m.fxe"},
	{0xD05380E5145C16F6ULL, "402136107.ogg"},
	{0xD054079B13A33CC5ULL, "vxht_keep_7a0_0_0.vlm"},
	{0xD054568D314BE2DAULL, "w_blightlands_151_0.mmh"},
	{0xD0545A8D314BE911ULL, "w_blightlands_151_0.mao"},
	{0xD0545CB6261E9AEAULL, "vxht_0_hightown_bp1_0_0.vlm"},
	{0xD054688D314C0120ULL, "w_blightlands_151_0.msh"},
	{0xD054B6D9E1E548F4ULL, "vxht_brothel_n_6b1_0_0.vlm"},
	{0xD054BF4D17A97B92ULL, "llac_lt_0_lowtown_docks_n_131070_3.dds"},
	{0xD054E1DD8A5D0CEAULL, "ser270_templar6112577.cl"},
	{0xD054F73A60BFCB34ULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.dura.evt"},
	{0xD0554B3A4C3E00AEULL, "vxw_bonepit_10q_0_0.vlm"},
	{0xD0559F3C0C1B6793ULL, "vxca_cave_1at_0_0.vlm"},
	{0xD055C16EE80C4EB3ULL, "6156289_m.fxe"},
	{0xD055F997F66EFB9CULL, "narrative_flemeth.cnv"},
	{0xD0560A15A6957B47ULL, "pro000ip_corpse_loot.ncs"},
	{0xD0564DDDC438904AULL, "vxw_bonepitdestroyed_a3_219_0_0.vlm"},
	{0xD0579A686CA8D74CULL, "vxdr_primevaltomb_8i7_0_0.vlm"},
	{0xD057E8C689C9D521ULL, "ico_boots_med_4.dds"},
	{0xD057FA51EBAC55A0ULL, "vxht_0_hightown_a3_bww_0_0.vlm"},
	{0xD0585ACB2A2F76C3ULL, "vxlt_gallowstemplar_f_4uq_0_0.vlm"},
	{0xD058AFE5BCFDFA5FULL, "656508591.ogg"},
	{0xD0591B97F6716BECULL, "narrative_flemeth.bnk"},
	{0xD059464346FB7223ULL, "vxlt_0_lowtown_a2_n_f50_0_0.vlm"},
	{0xD0599FFCD4CFD6BEULL, "vfx_dust_02.dds"},
	{0xD059B858CBF1125BULL, "vxht_mansion2_1vy_0_0.vlm"},
	{0xD059E3117A32DA6FULL, "7880453.ogg"},
	{0xD05A45CBEB10350BULL, "vxlt_gallowscourtyard_62c_0_0.vlm"},
	{0xD05A98BBDFDAD224ULL, "gen00fl_aveline6053287.cl"},
	{0xD05B4AD960F3D0B9ULL, "scaleform_sf_main_p1.vsh.fxo"},
	{0xD05B54E263858DB2ULL, "follower_banter6156003.cl"},
	{0xD05B87962577CE6CULL, "mag110_ws_banter6214710.cl"},
	{0xD05B9CDDBA60282DULL, "vxht_bartrand_n_5r3_0_0.vlm"},
	{0xD05BA1153C658B47ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_f_c.anb"},
	{0xD05BA4EEE5FA9087ULL, "vxw_encounter2_camp_25_0_0.vlm"},
	{0xD05BA50B5283CABCULL, "mcr370_negotiation6080165.cl"},
	{0xD05BC3F0D4E5CA90ULL, "gen00fl_fenris6112044.cl"},
	{0xD05BCB88A30CC19BULL, "llac_ht_chantry_2534_48.dds"},
	{0xD05BE11E8F8B278AULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtd.mao"},
	{0xD05BE51E8F8B2ED9ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtd.mmh"},
	{0xD05BEF1E8F8B3FC7ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtd.msh"},
	{0xD05BFBAA626765A8ULL, "ht_0_hightown_a2_53_0.rml"},
	{0xD05C10591D0870D4ULL, "vxht_0_hightown_a2_c6t_0_0.vlm"},
	{0xD05C4209B29C158BULL, "496488042.ogg"},
	{0xD05C682D23C66276ULL, "60186379.ogg"},
	{0xD05C75FADF55E6F6ULL, "hlac_ca_cave_2193_0x.dds"},
	{0xD05C9D38D90E572AULL, "595626542.ogg"},
	{0xD05CB3109AEB7E8DULL, "195278371.ogg"},
	{0xD05CF01EF0F77D33ULL, "dr_0_deeproads_19793.rml"},
	{0xD05D52F7A595269BULL, "216887715.ogg"},
	{0xD05D64AC97FCF0D0ULL, "305307594.ogg"},
	{0xD05D7C45AA1797BBULL, "vxlt_gallowscourtyard_n_7gz_0_0.vlm"},
	{0xD05E1AD2DE51DB96ULL, "vxlt_0_lowtown_a2_n_c_bys_0_0.vlm"},
	{0xD05EDAA0B77436ECULL, "gen00fl_isabela.cnv"},
	{0xD05F1924569BE675ULL, "lt_0_lowtown_docks_a3_244_0.msh"},
	{0xD05F1B24569BE9EBULL, "lt_0_lowtown_docks_a3_244_0.mmh"},
	{0xD05F2724569BFD88ULL, "lt_0_lowtown_docks_a3_244_0.mao"},
	{0xD05F2AE652755DC8ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_enco_0d.dds"},
	{0xD05F86563C12AD29ULL, "art\\characters\\creatures\\mubari\\animation\\amb_feed.ani"},
	{0xD05FAE8B0308CF2BULL, "6016696_f.fxe"},
	{0xD05FB0304A906259ULL, "6103678_m.fxe"},
	{0xD05FE20D17140571ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.anb"},
	{0xD0600DBE664A078EULL, "gen_im_wep_war_one_act3_06.uti"},
	{0xD0604C4260405CECULL, "6112062_m.fxe"},
	{0xD060CB768F331911ULL, "vxht_chantry_n_27l_0_0.vlm"},
	{0xD060E659F4F74550ULL, "seb221_flora_harimann6177727.cl"},
	{0xD061532E7E8A843CULL, "gen00fl_bethany6101523.cl"},
	{0xD061BA5083F6DC30ULL, "vxht_mansion_n_3v7_0_0.vlm"},
	{0xD061BC4156A56764ULL, "240102818.ogg"},
	{0xD061BCA0B776391CULL, "gen00fl_isabela.bnk"},
	{0xD061C1693E56239EULL, "6060398_m.fxe"},
	{0xD0621AC2CFCF3176ULL, "6149261_f.fxe"},
	{0xD062EF3A60CBFB6DULL, "art\\vfx\\combat\\vfx_m_bld_sac_proj_c.dura.ani"},
	{0xD06324FC35E612B7ULL, "6106712_m.fxe"},
	{0xD0637C75E6FF9696ULL, "6071136_m.fxe"},
	{0xD0641CA7FC5A39B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_6.gad"},
	{0xD0644CA954A08FF1ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube_tool.msh"},
	{0xD0645AA954A0A70CULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube_tool.mao"},
	{0xD06466A954A0BBAFULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_8m_tube_tool.mmh"},
	{0xD064778F046A390BULL, "gen00fl_varric6021109.cl"},
	{0xD0649C8B21B50163ULL, "one110_gustav6089825.cl"},
	{0xD064D083F83D68AEULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb104a_0n.dds"},
	{0xD064D3EA2F25BC3AULL, "6139529_f.fxe"},
	{0xD06500E464883633ULL, "6016586_f.fxe"},
	{0xD06594A7089041BDULL, "vxlt_hangedman_1xg_0_0.vlm"},
	{0xD065BF38D40A73BEULL, "6071860_m.fxe"},
	{0xD065E3B5875BA54DULL, "vxca_cave_gx_0_1.vlm"},
	{0xD0660DF500340090ULL, "sto_cl_wh.mao"},
	{0xD0667862157EF904ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stonetree_sa.dds"},
	{0xD06691FEA6DA10BAULL, "lt_0_lowtown_38_0_lc.mmh"},
	{0xD0669AF77F45BD21ULL, "vxlt_0_lowtown_a3_n_c_en3_0_0.vlm"},
	{0xD066A1563C18F9ABULL, "art\\characters\\creatures\\mubari\\animation\\amb_feed.gad"},
	{0xD06778653BFCF4A3ULL, "vxlt_gallowstemplar_f_2p8_0_0.vlm"},
	{0xD067B9D48901219EULL, "vxw_bonepitdestroyed_a3_1pa_0_0.vlm"},
	{0xD0682A35BDE72ADBULL, "vxht_mansion2_1yx_0_0.vlm"},
	{0xD0686C835C892CCEULL, "vxht_mansion_n_58m_0_0.vlm"},
	{0xD068ACDC04A7105CULL, "vxht_mansion_n_60l_0_0.vlm"},
	{0xD068D436B1BA349EULL, "247724574.ogg"},
	{0xD068DA4AB238827AULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedenter.gad"},
	{0xD068F5F0124D85C7ULL, "6095993_m.fxe"},
	{0xD069BCA435963358ULL, "318779755.ogg"},
	{0xD06A0EBAAC5640ADULL, "vxht_bartrand_n_5u2_0_0.vlm"},
	{0xD06B1FA7FC605DE9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_6.evt"},
	{0xD06B5E30F98E1403ULL, "gen_im_trs_cot_rng_r1.uti"},
	{0xD06BCE63A6BDF822ULL, "zz_t_5.cut"},
	{0xD06C85700EE3A2C1ULL, "vxht_mansion2b_2ee_0_0.vlm"},
	{0xD06C93584A389061ULL, "6152203_m.fxe"},
	{0xD06C99C2E6E5D904ULL, "vxw_bonepit_ll_0_0.vlm"},
	{0xD06C9E563C1D6010ULL, "art\\characters\\creatures\\mubari\\animation\\amb_feed.evt"},
	{0xD06CB91BD6AF86F7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\pose\\pose_a.gad"},
	{0xD06CD6A0B1CF8DA8ULL, "349020372.ogg"},
	{0xD06CF516DA052135ULL, "vxht_mansion2_1kp_0_0.vlm"},
	{0xD06D1B9DB86B0D04ULL, "hlac_lt_0_lowtown_a2_n_c_65538_3x.dds"},
	{0xD06D2E8AD285BCCCULL, "hlac_lt_0_lowtown_docks_n_262142_0x.dds"},
	{0xD06D370CABFE3D39ULL, "6070099_m.fxe"},
	{0xD06D648C7E580B56ULL, "art\\characters\\creatures\\stonegolem\\animation\\blendtree\\stonegolem.anb"},
	{0xD06D6699F823B1D3ULL, "914517882.ogg"},
	{0xD06D7E90A2245AE6ULL, "6142146_m.fxe"},
	{0xD06DE1497AB38F7BULL, "6080173_m.fxe"},
	{0xD06E030D17207AEAULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t4_c.mmh"},
	{0xD06E2130354DFB69ULL, "248085008.ogg"},
	{0xD06E229C962F6090ULL, "6121899_m.fxe"},
	{0xD06E29A33F7C14DAULL, "144938902.ogg"},
	{0xD06E2A85BCD377FDULL, "llac_lt_undercitydungeon_-1_22.dds"},
	{0xD06E3067F5444CBBULL, "hlac_lt_backalley_g_0_1x.dds"},
	{0xD06E8B30F640E010ULL, "6125781_m.fxe"},
	{0xD06EA68874DA6FBBULL, "vxht_0_hightown_n_dw_0_0.vlm"},
	{0xD06ED5B79A6E2BB3ULL, "6140336_m.fxe"},
	{0xD06ED8323EDC2F68ULL, "vxca_caverns2_71j_0_0.vlm"},
	{0xD06EDD4AB23CF291ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedenter.evt"},
	{0xD06F59FAC365E78BULL, "gen00fl_varric6078709.cl"},
	{0xD06F5E9A26097AE2ULL, "hlac_ht_mansion2b_818_4x.dds"},
	{0xD07056D77149E160ULL, "vxht_playershouse_n_1w3_0_0.vlm"},
	{0xD0709020260AA3F3ULL, "gen00fl_carver6206212.cl"},
	{0xD070A4C4CDAE4D37ULL, "vxw_encounter3_fk_0_1.vlm"},
	{0xD070B9B3E8DE7A10ULL, "dae000_madam_lusine6151397.cl"},
	{0xD0715AFDF3853AF3ULL, "dae300_cullen6135973.cl"},
	{0xD0715FD583FA6702ULL, "vxht_0_hightown_a2_b67_0_0.vlm"},
	{0xD071765BBD49FC1EULL, "art\\levels\\races\\proxy\\props\\door\\door_1.mao"},
	{0xD0719F6A635BE373ULL, "230989497.ogg"},
	{0xD071AB45F50F41A5ULL, "vxlt_gallowstemplar_f_358_0_0.vlm"},
	{0xD07268BCA36B8ABBULL, "6021118_m.fxe"},
	{0xD0726B28F2B8A16FULL, "vxlt_0_lowtown_a2_n_c_cmo_0_0.vlm"},
	{0xD0733AD945B4B42FULL, "83519270.ogg"},
	{0xD074084DB83F6711ULL, "gen00fl_aveline6139232.cl"},
	{0xD07431C48C87DBC5ULL, "gen_im_acc_blt_lore_02.uti"},
	{0xD07458B78CFCB078ULL, "vxw_bonepitmining_a2_1z3_0_0.vlm"},
	{0xD074CB66D032CA87ULL, "6092581_m.fxe"},
	{0xD074DE0BCF9886A0ULL, "vxht_0_hightown_a3_biu_0_0.vlm"},
	{0xD074EFD077226740ULL, "hlac_w_encounter3_-327684_0z.dds"},
	{0xD075A63215CCCDD4ULL, "6042059_m.fxe"},
	{0xD0760903E4DC456EULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v2.evt"},
	{0xD0761807A8B5D02CULL, "gen00fl_aveline6116690.cl"},
	{0xD0762BB17AF73A1EULL, "vxw_bonepitdestroyed_a3_1ub_0_0.vlm"},
	{0xD07655AC99F26A60ULL, "6214498_m.fxe"},
	{0xD07686683B81A10FULL, "6124752_m.fxe"},
	{0xD076A03ABAF9D3F1ULL, "vxht_mansion2_2pv_0_0.vlm"},
	{0xD076BA5B00367139ULL, "556695052.ogg"},
	{0xD076DFB05EAADF41ULL, "721218954.ogg"},
	{0xD077404B289B9E8BULL, "hlac_dr_primevaltomb_658_13z.dds"},
	{0xD07751D14C833A97ULL, "pro000_friedrich6050200.cl"},
	{0xD07807A7FC6ABF58ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_6.ani"},
	{0xD078131945D4A4A6ULL, "274632301.ogg"},
	{0xD0781F9027AC1A51ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexint.msh"},
	{0xD0782BA987684790ULL, "206236199.ogg"},
	{0xD0782D9027AC326CULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexint.mao"},
	{0xD078399027AC460FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexint.mmh"},
	{0xD07880979E4C592DULL, "vxht_bartrand_n_5x1_0_0.vlm"},
	{0xD07887A290F9BA7CULL, "927926348.ogg"},
	{0xD07898E4048DB0DEULL, "279277196.ogg"},
	{0xD078F41300F4A1D4ULL, "vxht_0_hightown_a2_c8z_0_0.vlm"},
	{0xD078F5F796002CD9ULL, "gen00fl_anders6125573.cl"},
	{0xD079D72F6E847BA1ULL, "vxw_bonepitdestroyed_a3_1fx_0_0.vlm"},
	{0xD079DA1DDE90252FULL, "421046583.ogg"},
	{0xD079DD629A44473CULL, "gen00fl_varric6129059.cl"},
	{0xD07A0F289198ABCEULL, "6149125_m.fxe"},
	{0xD07A295DA13BA6ECULL, "830926792.ogg"},
	{0xD07A70DC20589849ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v3.ani"},
	{0xD07A758D316CE6E5ULL, "w_blightlands_151_0.phy"},
	{0xD07B169315223892ULL, "675090874.ogg"},
	{0xD07B208370E99188ULL, "212484652.ogg"},
	{0xD07B2A844A54D41EULL, "480957489.ogg"},
	{0xD07B32E83CDA9D01ULL, "6061361_m.fxe"},
	{0xD07B37A6298B5CA8ULL, "vxlt_gallowstemplar_3p5_0_0.vlm"},
	{0xD07B497A0528A1CFULL, "379952512.ogg"},
	{0xD07BDF457BED1ADBULL, "vxlt_0_lowtown_docks_a3_n_bs8_0_0.vlm"},
	{0xD07BF7F391417DBFULL, "742509671.ogg"},
	{0xD07C0003E4E0A119ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v2.gad"},
	{0xD07C15EC2FDB78CAULL, "smoketest.ncc"},
	{0xD07C22E6D89EF553ULL, "749000126.ogg"},
	{0xD07C2E97C5324977ULL, "ms_dis.mao"},
	{0xD07C63C992792FE5ULL, "hlac_lt_0_lowtown_a3_n_c_65537_4z.dds"},
	{0xD07C995D27B6BDC8ULL, "316789989.ogg"},
	{0xD07CAE0B355511D9ULL, "6147783_f.fxe"},
	{0xD07CF54AB2495960ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\roguedualwieldselectedenter.ani"},
	{0xD07D749FACFD80D9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_loop.gad"},
	{0xD07DFBD6D30843CFULL, "6127551_m.fxe"},
	{0xD07E110F82F887ECULL, "vxht_mansion2b_2jy_0_0.vlm"},
	{0xD07E59A804A44FEBULL, "138597163.ogg"},
	{0xD07E9222B272CB22ULL, "gen00fl_sebastian6180521.cl"},
	{0xD07E9E267D112BD3ULL, "vxht_brothel_n_5lc_0_0.vlm"},
	{0xD07EB6D60492E800ULL, "287102066.ogg"},
	{0xD07EE77D1334C022ULL, "gen00fl_varric6109208.cl"},
	{0xD07F28E43EA93E62ULL, "6147871_m.fxe"},
	{0xD07F6E854D187BB9ULL, "vxht_keep_5ty_0_0.vlm"},
	{0xD07FC4E07CD26F45ULL, "897474208.ogg"},
	{0xD07FF170D9B9DF66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\exit.ani"},
	{0xD0800AB7ADB532C6ULL, "vxw_bonepitmining_a2_266_0_0.vlm"},
	{0xD0800B426C871221ULL, "w_encounter1_5_0.rml"},
	{0xD0800C70B7C79274ULL, "vxht_brothel_n_6u4_0_0.vlm"},
	{0xD0808BDC205D31CBULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v3.gad"},
	{0xD080E61BD6C07C7DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\pose\\pose_a.ani"},
	{0xD081216C7CDB4E69ULL, "vxlt_gallowscourtyard_6k6_0_0.vlm"},
	{0xD08132E5293FFAB7ULL, "6163586_m.fxe"},
	{0xD08141828BE81CD1ULL, "885517780.ogg"},
	{0xD0814ED58177D371ULL, "6052095_m.fxe"},
	{0xD081857258C01FEAULL, "6042040_m.fxe"},
	{0xD081E8DFC6F3ADF2ULL, "mag110_samson6016161.cl"},
	{0xD0823241B180216DULL, "vxlt_gallowscourtyard_54h_0_0.vlm"},
	{0xD0825B554033B75DULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pew.phy"},
	{0xD0826898CF5B8EF9ULL, "303129663.ogg"},
	{0xD082D2E3672E066CULL, "134505075.ogg"},
	{0xD0831103E4E6DD67ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v2.ani"},
	{0xD083C9E402A6226FULL, "vxlt_0_lowtown_a2_n_c_fe6_0_0.vlm"},
	{0xD08431A20D5F1A26ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast15_4m.phy"},
	{0xD084859FAD03BC27ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_loop.ani"},
	{0xD0849E153C888B48ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_f_c.mmh"},
	{0xD084D4B9266FF733ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basic.phy"},
	{0xD0850DEFA1D35BDBULL, "vxht_mansion2_1gz_0_0.vlm"},
	{0xD085A33D47CF565EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_forward.ani"},
	{0xD08606C0C4839195ULL, "vxlt_0_lowtown_a3_n_c_ex9_0_0.vlm"},
	{0xD08606D7D7CB339EULL, "vxca_cavern1_haunted_i2m_0_0.vlm"},
	{0xD0863EFE36F9BFC1ULL, "vxca_cavern1_haunted_ile_0_1.vlm"},
	{0xD08672FCF4F6CD24ULL, "6051801_f.fxe"},
	{0xD0871E083927D0A0ULL, "and341tr_spider_ambush.utt"},
	{0xD08731BA7A4245ACULL, "vxht_0_hightown_a3_cat_0_0.vlm"},
	{0xD0877E0065B7E3F6ULL, "68907264.ogg"},
	{0xD08788DC20634B30ULL, "art\\characters\\creatures\\hurlock\\animation\\crossbow\\hurlock_cb_rng_atk_v3.evt"},
	{0xD087B2FF64D16513ULL, "290797348.ogg"},
	{0xD087B87872E8B9FEULL, "669772784.ogg"},
	{0xD087DCAEC4550394ULL, "hlac_lt_0_lowtown_a2_n_-131073_2z.dds"},
	{0xD0888105838EC44EULL, "6142025_m.fxe"},
	{0xD088BC08EC1AFA6EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_exit.evt"},
	{0xD088E0BC4430950AULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.cess.gad"},
	{0xD088E40BB50FD4BCULL, "vxlt_0_lowtown_a2_fj1_0_0.vlm"},
	{0xD089D435E1DDE63DULL, "hlac_ca_caverns1_24425_7z.dds"},
	{0xD08A31044DB0C29DULL, "6166950_m.fxe"},
	{0xD08AE3DB511A8497ULL, "947481961.ogg"},
	{0xD08AF7C2FC4A5D8EULL, "one220_yevhen6135782.cl"},
	{0xD08AFA241385BCA4ULL, "vxlt_0_lowtown_a2_n_c_by1_0_0.vlm"},
	{0xD08B029B8B578C16ULL, "6040785_m.fxe"},
	{0xD08BBDE19362B389ULL, "hlac_lt_0_lowtown_a3_n_c_-1_4x.dds"},
	{0xD08BDA7EADE2E995ULL, "6136133_m.fxe"},
	{0xD08C059B38768EB7ULL, "hlac_lt_0_lowtown_n_0_5z.dds"},
	{0xD08C7A975365AF8DULL, "hlac_lt_gallowscourtyard_n_-65535_0y.dds"},
	{0xD08CAF063BCB300BULL, "vxca_caverns1_i1p_0_0.vlm"},
	{0xD08CEC857FA5F087ULL, "hlac_lt_0_lowtown_docks_a3_0_0x.dds"},
	{0xD08CEF9BC9FF92B1ULL, "llac_dr_dungeon_3214_1.dds"},
	{0xD08D9AC1309F56E2ULL, "6140971_m.fxe"},
	{0xD08DD970D9C5F3D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\exit.evt"},
	{0xD08EB308EC1F5619ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_exit.gad"},
	{0xD08EBD0EE510309EULL, "98380723.ogg"},
	{0xD08F2A9582A1EA67ULL, "vxht_chantry_n_2yp_0_0.vlm"},
	{0xD08F52C93567D1F7ULL, "vxlt_gallowscourtyard_n_7gn_0_0.vlm"},
	{0xD08FC3BC443682A1ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.cess.evt"},
	{0xD090062EF487388DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flaggallows_sa.dds"},
	{0xD0906759B9149242ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whorebanner.msh"},
	{0xD0906959B91495F8ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whorebanner.mmh"},
	{0xD0906B1B5FB993EDULL, "6172515_m.fxe"},
	{0xD0907559B914A99BULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_whorebanner.mao"},
	{0xD090C1161685D465ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_5_c.mmh"},
	{0xD090FB078B06E432ULL, "39921521.ogg"},
	{0xD0912FB60F07B8CCULL, "6176737_f.fxe"},
	{0xD0913CB2EBF1583AULL, "vxht_mansion2_2oh_0_0.vlm"},
	{0xD09197394AF6748CULL, "hm_castillon.mor"},
	{0xD0920A9D107CEBEDULL, "vxht_mansion2_3ba_0_0.vlm"},
	{0xD0922338AD784EA1ULL, "zz_applyeffect.ncs"},
	{0xD092BB3D47DA092FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_forward.evt"},
	{0xD0935055404217C9ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pew.mao"},
	{0xD0935C5540422C62ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pew.mmh"},
	{0xD0935E5540422F18ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_pew.msh"},
	{0xD0937C4F4C83F60EULL, "zz_dae_debug6029864.cl"},
	{0xD09381837242AB2EULL, "vxca_caverns1_e7w_0_0.vlm"},
	{0xD093B7DE6C0343CFULL, "abi_generic_warrior.wbd"},
	{0xD093BDE263A1A268ULL, "art\\vfx\\textures\\vfx_flare01a.dds"},
	{0xD093D53D904EAC87ULL, "6076125_m.fxe"},
	{0xD0940639BB70ECE2ULL, "6139730_f.fxe"},
	{0xD0947AE9C9B62D51ULL, "57722271.ogg"},
	{0xD094DA70D9CC1460ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\exit.gad"},
	{0xD094E6C407CD81AAULL, "782353305.ogg"},
	{0xD095109BDA6454ADULL, "6197258_f.fxe"},
	{0xD0956A7BE461FC40ULL, "223772528.ogg"},
	{0xD0957566CDB6F7F4ULL, "w_bonepit_49_0.rml"},
	{0xD095BC2E25FB7E80ULL, "6179256_m.fxe"},
	{0xD095C2C49471AD81ULL, "6155985_m.fxe"},
	{0xD095C408EC259267ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_action_exit.ani"},
	{0xD096DA5EDA0E0A6DULL, "6013319_m.fxe"},
	{0xD0979682B3D86477ULL, "hlac_w_blightlands_-65536_0z.dds"},
	{0xD0979F58898766BFULL, "6149369_f.fxe"},
	{0xD0985037099A37D8ULL, "858864841.ogg"},
	{0xD098F2FC60FF5DC6ULL, "llac_dr_dungeon_3216_11.dds"},
	{0xD099BC3D47E029D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_forward.gad"},
	{0xD099DCF8B069A7B4ULL, "llac_lt_undercityserialkiller_-1_20.dds"},
	{0xD09A0681F679A8F6ULL, "face_face_p3.psh.fxo.dx11"},
	{0xD09A887B5728646BULL, "hlac_ht_bartrand_n_3370_1z.dds"},
	{0xD09AA1549FC69AE5ULL, "vxlt_0_lowtown_a3_n_c_erk_0_0.vlm"},
	{0xD09ACC84BCCB19E2ULL, "541738941.ogg"},
	{0xD09AD664E6509DEFULL, "681890867.ogg"},
	{0xD09ADB20EC427AE4ULL, "6075685_f.fxe"},
	{0xD09B1C2A1EE58FDCULL, "ave331_post_fight6138799.cl"},
	{0xD09B37E3A6116622ULL, "1057976221.ogg"},
	{0xD09B9E1F8366AA31ULL, "llac_lt_0_lowtown_a3_n_c_-131072_1.dds"},
	{0xD09C10BB93406029ULL, "character_character_p110.psh.fxo.dx11"},
	{0xD09C4B36B4A124EDULL, "6131122_f.fxe"},
	{0xD09C5E98DD250BD5ULL, "367221134.ogg"},
	{0xD09CBBBC4440FFF0ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_proj_p.cess.ani"},
	{0xD09CF59A54EC1593ULL, "isa100pt_fools.ncs"},
	{0xD09D3DC2300A520BULL, "gen00fl_sebastian6177304.cl"},
	{0xD09E54CBF106908DULL, "6101795_m.fxe"},
	{0xD09E63F190AB3558ULL, "654424982.ogg"},
	{0xD09E7DFE9433442EULL, "mer221_sabin6047518.cl"},
	{0xD09EBE1616920C32ULL, "art\\vfx\\combat\\vfx_r_var_rfire_act_5_c.anb"},
	{0xD09F068675A8EBFAULL, "6165651_m.fxe"},
	{0xD09F0DFA0C9B25F0ULL, "vxlt_backalley_n_1yc_0_0.vlm"},
	{0xD09FE25BD87C537EULL, "zz_seb_debug6215775.cl"},
	{0xD09FEEE5ED58AB82ULL, "6172079_m.fxe"},
	{0xD0A04A351DAB7BAEULL, "vxlt_backalley_g_1u4_0_0.vlm"},
	{0xD0A0E4026A5969E2ULL, "dre180ip_crypt_treasure.utp"},
	{0xD0A20655B6B4F8B4ULL, "vxca_cavern1_haunted_j6d_0_0.vlm"},
	{0xD0A206AEDDED4203ULL, "vxht_0_hightown_bdx_0_0.vlm"},
	{0xD0A210048A490E83ULL, "vxca_caverns2_8kt_0_0.vlm"},
	{0xD0A2115CE2A0691AULL, "vxlt_0_lowtown_a2_f4u_0_0.vlm"},
	{0xD0A23712585E6F97ULL, "6129755_m.fxe"},
	{0xD0A29BA2F2E87E3AULL, "6118010_f.fxe"},
	{0xD0A2B62E4CDD5042ULL, "801328721.ogg"},
	{0xD0A2C66518105D14ULL, "lt_0_lowtown_a2_n_32_0_ll.mmh"},
	{0xD0A30AB3C19F91A1ULL, "vxdr_primevaltomb_8ez_0_0.vlm"},
	{0xD0A319E519071022ULL, "620879129.ogg"},
	{0xD0A345298A3FDDE2ULL, "vxca_caverns1_mines_j48_0_0.vlm"},
	{0xD0A345F5A360E7C3ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepiled.msh"},
	{0xD0A353F5A360FF1DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepiled.mmh"},
	{0xD0A357F5A361062EULL, "art\\levels\\races\\natural\\highlands\\props\\prp_bonepiled.mao"},
	{0xD0A37195C428B1CAULL, "art\\levels\\ztools\\helpers\\character\\prp_humanrun_1.mao"},
	{0xD0A3E8C5D12D0EE4ULL, "112182074.ogg"},
	{0xD0A428B8B214D6FBULL, "w_encounter3_117_0.phy"},
	{0xD0A4B71592C1071DULL, "vxw_woundedcoast_1di_0_0.vlm"},
	{0xD0A4DB599B9BC3B7ULL, "228531039.ogg"},
	{0xD0A56655EAB1CB7AULL, "748875672.ogg"},
	{0xD0A5C7CE931693D8ULL, "hlac_w_encounter2_camp_-65538_0z.dds"},
	{0xD0A64BDAAEA44F4EULL, "296049918.ogg"},
	{0xD0A6974B426F1019ULL, "253222063.ogg"},
	{0xD0A6ED42FC07F238ULL, "vxlt_undercityencounter_1n7_0_0.vlm"},
	{0xD0A6F9B9EB9F52DCULL, "304369137.ogg"},
	{0xD0A78C5D1306F430ULL, "llac_lt_0_lowtown_n_65538_1.dds"},
	{0xD0A7AFBF9DAEFE47ULL, "test_gate.stg"},
	{0xD0A7FE2FFE3424F6ULL, "vxht_keep_n_6f4_0_0.vlm"},
	{0xD0A8553DBBA40544ULL, "vxlt_0_lowtown_a2_fhg_0_0.vlm"},
	{0xD0A8C8B3876CAA3FULL, "vxht_mansion2b_2jv_0_0.vlm"},
	{0xD0A939C87D5D7337ULL, "art\\levels\\races\\proxy\\props\\pressure\\blendtree\\floortrap.anb"},
	{0xD0A97F08439C64C4ULL, "mag121_viveka6083388.cl"},
	{0xD0A9ECBE74918063ULL, "lt_0_lowtown_docks_n_170_0.msh"},
	{0xD0A9FABE749197BDULL, "lt_0_lowtown_docks_n_170_0.mmh"},
	{0xD0A9FEBE74919ECEULL, "lt_0_lowtown_docks_n_170_0.mao"},
	{0xD0AA6FD85BDD9EF5ULL, "6172858_m.fxe"},
	{0xD0AA8CB5E4918360ULL, "dae200_cullen6140506.cl"},
	{0xD0AAFF9243D11155ULL, "6117965_m.fxe"},
	{0xD0AB237706259FD0ULL, "6034825_m.fxe"},
	{0xD0AB2F31AC3DC8C9ULL, "71336249.ogg"},
	{0xD0AB9DFA2AC582E3ULL, "vxca_caverns1_mines_k2o_0_1.vlm"},
	{0xD0AC0FF40BD3C915ULL, "dae000ip_crowd_mf1_talk17.utp"},
	{0xD0AC1A69E25B3D2EULL, "6091703_m.fxe"},
	{0xD0AC9A5147616733ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd01a.phy"},
	{0xD0AC9B7542B412D6ULL, "vxht_mansion2_25f_0_0.vlm"},
	{0xD0AD1D374E358427ULL, "ave331_jeven6138739.cl"},
	{0xD0ADCC3F44A89D04ULL, "zz_mrl150ip_stage.utp"},
	{0xD0AE150664E7D981ULL, "zz_followers_banter6158514.cl"},
	{0xD0AE206CCFDD893AULL, "vxht_mansion2_2uj_0_0.vlm"},
	{0xD0AE62097AF43E30ULL, "mag360st_alain.stg"},
	{0xD0AEBC120FA1942EULL, "vxlt_backalley_g_1r5_0_0.vlm"},
	{0xD0AECB9543A0085CULL, "l_lgt331ar_hybris.wbd"},
	{0xD0AEE4CFB30940CEULL, "w_woundedcoast_27_0_lc.mmh"},
	{0xD0AEE533F23C3C8BULL, "gen_im_acc_blt_act2_06.uti"},
	{0xD0AF88924B1937FAULL, "dae011_corff_sta6149606.cl"},
	{0xD0AFC9DEA5F9A1BBULL, "gen00fl_sebastian6189510.cl"},
	{0xD0AFD7437E86C557ULL, "hlac_lt_0_lowtown_docks_n_-196610_5y.dds"},
	{0xD0B04CEEA566EEBEULL, "w_encounter2_5_0.rml"},
	{0xD0B08339D496819AULL, "vxlt_0_lowtown_a2_f1v_0_0.vlm"},
	{0xD0B08CF2C9957755ULL, "vxht_brothel_n_6fx_0_0.vlm"},
	{0xD0B094CD36B76DEEULL, "vxlt_0_lowtown_n_h5g_0_0.vlm"},
	{0xD0B11B1CE0BA2EBFULL, "ht_0_hightown_a3_53_0.rml"},
	{0xD0B12E4DC5B3BD31ULL, "233069822.ogg"},
	{0xD0B132CE58091075ULL, "mcr370ip_to_templar_hall.utp"},
	{0xD0B14EEAA8B4D499ULL, "vxw_encounter1_ce_0_0.vlm"},
	{0xD0B194950CDC0941ULL, "vxca_cavern1_haunted_icx_0_1.vlm"},
	{0xD0B20C845CBECA91ULL, "666747591.ogg"},
	{0xD0B27AC5E14AC3EFULL, "zz_fol_debug6072543.cl"},
	{0xD0B28637C3964A93ULL, "339310901.ogg"},
	{0xD0B2C9BB7B527C74ULL, "character_character_p98.psh.fxo"},
	{0xD0B30D53BE3BC0A8ULL, "vxw_encounter2_dx_0_0.vlm"},
	{0xD0B391FF70B363FEULL, "vxw_bonepitdestroyed_a3_ks_0_0.vlm"},
	{0xD0B3C0F3C0642C4AULL, "gen_im_cft_rec_wrn_06.uti"},
	{0xD0B3F1085EBB1BBEULL, "vxdr_0_deeproads_ejs_0_0.vlm"},
	{0xD0B40C763C1FACD5ULL, "dae000_merchant_generic_hm6172596.cl"},
	{0xD0B420292EABB57BULL, "22611625.ogg"},
	{0xD0B4434656299444ULL, "vxlt_0_lowtown_docks_a3_n_d16_0_1.vlm"},
	{0xD0B464E72D8E7219ULL, "6095263_m.fxe"},
	{0xD0B4699ECD417E3DULL, "vxw_encounter2_camp_5n_0_0.vlm"},
	{0xD0B5091FE0998E0CULL, "follower_banter6151453.cl"},
	{0xD0B539C5033CDAA1ULL, "vxlt_0_lowtown_docks_b0o_0_0.vlm"},
	{0xD0B55EF0EB215422ULL, "927049621.ogg"},
	{0xD0B563915999AD7EULL, "6077879_m.fxe"},
	{0xD0B5B7A80C8899C5ULL, "dre000_dougal6129788.cl"},
	{0xD0B6078718FDD341ULL, "hlac_ht_0_hightown_-196607_0x.dds"},
	{0xD0B613E6949135D7ULL, "hlac_w_bonepitdestroyed_a3_0_13y.dds"},
	{0xD0B640F112D68936ULL, "vxca_cavern1_haunted_i32_0_0.vlm"},
	{0xD0B641E5094E7364ULL, "745473776.ogg"},
	{0xD0B665EEA56B8482ULL, "w_encounter2_5_0.phy"},
	{0xD0B6700CF02A3D76ULL, "vxht_keep_n_6a5_0_0.vlm"},
	{0xD0B6B1FAF5C683E8ULL, "6002631_m.fxe"},
	{0xD0B731E091B5604AULL, "dae000_male_templar_mob6174359.cl"},
	{0xD0B761DCBF1A1D22ULL, "6095884_m.fxe"},
	{0xD0B7850E83B2CBE5ULL, "vxlt_0_lowtown_a3_n_c_exi_0_0.vlm"},
	{0xD0B7C5176B3526CCULL, "character_character_p69.psh.fxo.dx11"},
	{0xD0B913C8B57C6D8BULL, "qun160_vashothlead6064651.cl"},
	{0xD0BA5C85CAEC0F07ULL, "hlac_ca_caverns1_mines_24423_3y.dds"},
	{0xD0BB9CC4B9381BB9ULL, "hlac_lt_0_lowtown_docks_n_-65537_5z.dds"},
	{0xD0BBFC0E1566FDFBULL, "728978734.ogg"},
	{0xD0BCC79A07B2DE8BULL, "6133091_m.fxe"},
	{0xD0BD1C2032D339F0ULL, "6135641_m.fxe"},
	{0xD0BD4494FC8F9871ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_tmpa_0s.dds"},
	{0xD0BDDCA0FE882D36ULL, "lgt310_mis_burn_rival_thug6152775.cl"},
	{0xD0BDF63D493D0E01ULL, "vxlt_warehouse_5q_0_0.vlm"},
	{0xD0BE052E4E23FBB0ULL, "mag110_arianni6015767.cl"},
	{0xD0BE2AE61B8AD8F4ULL, "350014676.ogg"},
	{0xD0BE4427E1084E4DULL, "605490511.ogg"},
	{0xD0BE7D59FE0728BAULL, "lt_warehouse_n_lc.mmh"},
	{0xD0BEE61E8FDFADC6ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtd.phy"},
	{0xD0BEFC564DC38286ULL, "vxw_bonepitdestroyed_a3_nd_0_0.vlm"},
	{0xD0BF25207CB39AE0ULL, "vxlt_undercityserialkiller_19j_0_0.vlm"},
	{0xD0BF7F8EA6C15643ULL, "866423798.ogg"},
	{0xD0BF819CEC4B35FAULL, "6142168_m.fxe"},
	{0xD0BFC0C79AAAED19ULL, "vxw_encounter1_hd_0_0.vlm"},
	{0xD0BFFD426CBCDE10ULL, "w_encounter1_5_0.msh"},
	{0xD0C0047FF458256DULL, "llac_w_woundedcoast_262150_0.dds"},
	{0xD0C00B426CBCF62AULL, "w_encounter1_5_0.mmh"},
	{0xD0C00F426CBCFD61ULL, "w_encounter1_5_0.mao"},
	{0xD0C02F9DE0205900ULL, "one3510cr_dead_fighter_3.utc"},
	{0xD0C0E603F73AB8C7ULL, "hlac_ht_0_hightown_a3_-131071_0y.dds"},
	{0xD0C144F60A3C6609ULL, "6061233_m.fxe"},
	{0xD0C1BE2B11B40B9FULL, "6134847_m.fxe"},
	{0xD0C1D4ECE140FB68ULL, "hlac_ca_caverns2_5802_9x.dds"},
	{0xD0C202003AD14189ULL, "gen00fl_anders6119185.cl"},
	{0xD0C2AE374D67FEC5ULL, "mag101_alain6053011.cl"},
	{0xD0C2DD15F50052F5ULL, "gen00fl_merrill6119594.cl"},
	{0xD0C37825A53267BDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45basic.phy"},
	{0xD0C3AA32DCFAFEF6ULL, "lfx_temp01.dds"},
	{0xD0C3ABA1F532F321ULL, "vxlt_0_lowtown_docks_b7n_0_0.vlm"},
	{0xD0C3D56E4B8FC5FEULL, "6078979_m.fxe"},
	{0xD0C3F0D04BADF0ACULL, "hlac_dr_primevalentrance_1724_4x.dds"},
	{0xD0C4196EF9FE3669ULL, "hlac_lt_0_lowtown_docks_196609_0x.dds"},
	{0xD0C4617B9BCC219EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopening.ani"},
	{0xD0C4E7CC9BDE4B01ULL, "6122154_m.fxe"},
	{0xD0C4F15C19FF0643ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue2_sa.dds"},
	{0xD0C5144340C9C9AFULL, "art\\vfx\\creatures\\fxc_shade_c.cess.gad"},
	{0xD0C523D7E7BC9C68ULL, "6005731_m.fxe"},
	{0xD0C6748A79EB9B22ULL, "vxlt_0_lowtown_docks_n_ctj_0_0.vlm"},
	{0xD0C6BD55CF2C63B9ULL, "llac_lt_darktown_-1_21.dds"},
	{0xD0C6DD8B4878E364ULL, "ca_caverns1_mines_24423_lc.mmh"},
	{0xD0C7538C00094712ULL, "llac_lt_0_lowtown_-4_0.dds"},
	{0xD0C7A5D23D13D2BCULL, "nat380_temmerin6120946.cl"},
	{0xD0C8AFE089F2D21AULL, "48920904.ogg"},
	{0xD0C8EE46390D35DAULL, "vxht_chantry_n_30w_0_0.vlm"},
	{0xD0C8F2094A7A4C67ULL, "vxht_chantry_n_2el_0_0.vlm"},
	{0xD0C90302AB8ACE2FULL, "hlac_lt_0_lowtown_docks_n_-2_2y.dds"},
	{0xD0C9FB4BD404E827ULL, "ser221st_gamlen_follower.stg"},
	{0xD0CA455B7BD142FBULL, "hlac_ht_0_hightown_a2_-196610_0x.dds"},
	{0xD0CB46FBD546DF86ULL, "915833692.ogg"},
	{0xD0CB945114DE8E2FULL, "mag100cr_wilmod_demon.utc"},
	{0xD0CB97790923DBB3ULL, "hlac_w_sundermount_262138_0z.dds"},
	{0xD0CC11B9E4CE0A48ULL, "241178300.ogg"},
	{0xD0CC2813D324B9B8ULL, "hlac_lt_0_lowtown_docks_n_-4_1y.dds"},
	{0xD0CC294340D00C5CULL, "art\\vfx\\creatures\\fxc_shade_c.cess.evt"},
	{0xD0CC8B3D7F693C36ULL, "dae000_gamlen6124141.cl"},
	{0xD0CCE66987B3DFBBULL, "vxlt_warehouse_cu_0_0.vlm"},
	{0xD0CE4C7C92F89F13ULL, "vxca_caverns2prc_61u_0_0.vlm"},
	{0xD0CE59DE1BC84CAAULL, "6155700_m.fxe"},
	{0xD0CE6A26A35C9169ULL, "vxdr_primevaltomb_4xe_0_0.vlm"},
	{0xD0CE752A747654A0ULL, "dae011_conspiracy_mob6149593.cl"},
	{0xD0CEA60A4297A5F0ULL, "295001704.ogg"},
	{0xD0CECE42D40F6754ULL, "mir_rip_01.mao"},
	{0xD0CF7566CDE86FB7ULL, "w_bonepit_49_0.mmh"},
	{0xD0CF7966CDE876C4ULL, "w_bonepit_49_0.mao"},
	{0xD0CF8B66CDE89529ULL, "w_bonepit_49_0.msh"},
	{0xD0CFD3C6126EB8DAULL, "gen00fl_anders6103903.cl"},
	{0xD0CFF10DA227F1A8ULL, "vxw_encounter2_jv_0_0.vlm"},
	{0xD0D00FB8B239E69AULL, "w_encounter3_117_0.msh"},
	{0xD0D01DB8B239FED3ULL, "w_encounter3_117_0.mao"},
	{0xD0D021B8B23A05E0ULL, "w_encounter3_117_0.mmh"},
	{0xD0D083FE08E7F64CULL, "vxht_0_hightown_n_c5f_0_0.vlm"},
	{0xD0D1702EBD5D6244ULL, "zz_one_debug6079313.cl"},
	{0xD0D1FD011F8DE49CULL, "dae000_arishok6067643.cl"},
	{0xD0D2C16C88E1E866ULL, "6062990_m.fxe"},
	{0xD0D2C4A5B976C71FULL, "6135274_f.fxe"},
	{0xD0D2D0AC2C9FF113ULL, "vxht_mansion_n_6j5_0_0.vlm"},
	{0xD0D2E9A4A009736FULL, "vxlt_backalley_g_1kg_0_0.vlm"},
	{0xD0D2F28DEB2C5CC7ULL, "vxlt_gallowstemplar_f_4um_0_0.vlm"},
	{0xD0D30480C203381BULL, "gen00fl_merrill6124048.cl"},
	{0xD0D3195E2E17EA15ULL, "90299794.ogg"},
	{0xD0D3488F3295503AULL, "follower_barks6153623.cl"},
	{0xD0D379BFBF5FF1FAULL, "vxlt_0_lowtown_docks_n_b9l_0_0.vlm"},
	{0xD0D3A846A5C432ECULL, "6089124_m.fxe"},
	{0xD0D3AED5CCB47BE5ULL, "597599218.ogg"},
	{0xD0D46D25A540C829ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45basic.mao"},
	{0xD0D47925A540DCC2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45basic.mmh"},
	{0xD0D47B25A540E078ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww_45basic.msh"},
	{0xD0D48C0DBCCF0F44ULL, "6048466_m.fxe"},
	{0xD0D4D1DE6C3B06A5ULL, "abi_generic_warrior.bnk"},
	{0xD0D4E6676BE1B3A2ULL, "vxlt_0_lowtown_docks_n_cok_0_0.vlm"},
	{0xD0D58FD09552A740ULL, "scaleform_sf_solid_p0.psh.fxo.dx11"},
	{0xD0D5A64B8732F98FULL, "6174568_m.fxe"},
	{0xD0D5DE602607348FULL, "6165562_m.fxe"},
	{0xD0D5F7DF9B7C46B9ULL, "isa211_isabela6069979.cl"},
	{0xD0D78B2A6058432EULL, "hlac_w_woundedcoast_0_3x.dds"},
	{0xD0D7DB167C0061ECULL, "6156354_m.fxe"},
	{0xD0D7E09A7303FDD8ULL, "6047367_f.fxe"},
	{0xD0D8023537BCC104ULL, "6031260_m.fxe"},
	{0xD0D82FC487CD3255ULL, "200288618.ogg"},
	{0xD0D87A7B9BDCF518ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopening.gad"},
	{0xD0D97CBDF6D5A1E5ULL, "728617918.ogg"},
	{0xD0D980CD7EA27334ULL, "vxw_woundedcoast_h1_0_0.vlm"},
	{0xD0D9AB53EB9F0F8BULL, "6139691_m.fxe"},
	{0xD0D9BD87219B5746ULL, "ring2.dds"},
	{0xD0D9EF1A4D42045DULL, "6098783_m.fxe"},
	{0xD0DA07CF56AA3526ULL, "fx_firetwister1.dds"},
	{0xD0DA176EFE43983DULL, "vxca_caverns1_j2l_0_0.vlm"},
	{0xD0DA314340DC5775ULL, "art\\vfx\\creatures\\fxc_shade_c.cess.ani"},
	{0xD0DAC99543C53FD6ULL, "l_lgt331ar_hybris.bnk"},
	{0xD0DADEE828F32A40ULL, "348505186.ogg"},
	{0xD0DB2A69BF8C4386ULL, "6126932_f.fxe"},
	{0xD0DC171425423648ULL, "6076969_f.fxe"},
	{0xD0DC40EEA58C1536ULL, "w_encounter2_5_0.mao"},
	{0xD0DC4406C68CF31FULL, "ser221_party_comments6163697.cl"},
	{0xD0DC4CEEA58C29D5ULL, "w_encounter2_5_0.mmh"},
	{0xD0DC4EEEA58C2D4BULL, "w_encounter2_5_0.msh"},
	{0xD0DC5D20D2EBD2DDULL, "6109049_m.fxe"},
	{0xD0DC929ADF1A565DULL, "mcr310_argument6074491.cl"},
	{0xD0DDE21273B29CFFULL, "vxht_0_hightown_n_amo_0_0.vlm"},
	{0xD0DE061999A398EDULL, "vxht_keep_6uy_0_0.vlm"},
	{0xD0DE7E895A99691DULL, "vxw_woundedcoast_1pe_0_0.vlm"},
	{0xD0DE7F08A8C16B49ULL, "arch_blood_sh.mao"},
	{0xD0DE83B00E6BF9D9ULL, "hlac_lt_0_lowtown_docks_a3_-3_2z.dds"},
	{0xD0DE91815A1A9518ULL, "llac_ht_chantry_n_2534_2.dds"},
	{0xD0DE96F412485C3AULL, "zz_prmpt_debug6198154.cl"},
	{0xD0DF6869113BE819ULL, "47822170.ogg"},
	{0xD0E01BC0ECF87068ULL, "mcr371st_orsino.stg"},
	{0xD0E01D31B0B61D55ULL, "vxw_sundermount_22h_0_0.vlm"},
	{0xD0E05F59CBC2A69EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down_less.ani"},
	{0xD0E0640475B9C6ECULL, "art\\vfx\\cinematic\\vfx_spark_imp_p.impa.ani"},
	{0xD0E0C0FC83CAD3B3ULL, "vxca_caverns1_jlp_0_0.vlm"},
	{0xD0E0E0B50781A560ULL, "323269005.ogg"},
	{0xD0E1068EF3DE0A51ULL, "wos_03.mao"},
	{0xD0E1324D34D850D6ULL, "6092315_m.fxe"},
	{0xD0E142891E960993ULL, "vxht_mansion_n_6m4_0_0.vlm"},
	{0xD0E16B7E0045254DULL, "6126406_m.fxe"},
	{0xD0E1EB9CB1560A7AULL, "vxlt_0_lowtown_docks_n_b6m_0_0.vlm"},
	{0xD0E2106A10B31035ULL, "dae000_merchant_weapons_dm6164397.cl"},
	{0xD0E27CAD51332C30ULL, "zz_vault_debug6203206.cl"},
	{0xD0E289FBD3B5E9FCULL, "6172288_m.fxe"},
	{0xD0E2DF91C542DAE6ULL, "vxca_cavern1_haunted_iip_0_0.vlm"},
	{0xD0E2F53015ECF1AFULL, "vxdr_primevaltomb_3is_0_0.vlm"},
	{0xD0E32B0027727666ULL, "ht_0_hightown_a3_43_0_ll.mmh"},
	{0xD0E3A67D8928AB51ULL, "vxw_sundermount_34g_0_0.vlm"},
	{0xD0E4356A0A303B8DULL, "vxca_caverns2prc_abm_0_0.vlm"},
	{0xD0E47A2FFF5F5AE1ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga101a_0n.dds"},
	{0xD0E4A0C30DC6197AULL, "970441544.ogg"},
	{0xD0E4BCB09198098EULL, "254977293.ogg"},
	{0xD0E53A8C57BBDD69ULL, "pro000_meeran6050088.cl"},
	{0xD0E58F002639BB9BULL, "vxlt_darktown_4u7_0_1.vlm"},
	{0xD0E5FF0D0F75B3C3ULL, "301356982.ogg"},
	{0xD0E668C328AB0333ULL, "vxw_encounter2_camp_ps_0_0.vlm"},
	{0xD0E6710475BE482EULL, "art\\vfx\\cinematic\\vfx_spark_imp_p.impa.gad"},
	{0xD0E6A9D00B4B2A5AULL, "vxca_cavern1_haunted_j5c_0_0.vlm"},
	{0xD0E6DEBA4C83E285ULL, "vxht_playershouse_n_1ri_0_0.vlm"},
	{0xD0E77C10400005B6ULL, "hlac_lt_undercityserialkiller_-1_12y.dds"},
	{0xD0E785AC7604CB9FULL, "947450206.ogg"},
	{0xD0E789CBA8251225ULL, "125023602.ogg"},
	{0xD0E7988687E0CA9EULL, "vxdr_primevaltomb_8ye_0_0.vlm"},
	{0xD0E7A9694D0B5D41ULL, "vxlt_0_lowtown_a2_n_f5j_0_0.vlm"},
	{0xD0E7CDB926C468A8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basic.mmh"},
	{0xD0E7D9B926C47D4BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basic.mao"},
	{0xD0E7EBB926C49BD2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ff_basic.msh"},
	{0xD0E7F09E77A99F60ULL, "vxw_bonepitmining_a2_16g_0_0.vlm"},
	{0xD0E7F2AA70988BB4ULL, "vxw_woundedcoast_e6_0_0.vlm"},
	{0xD0E810F5D75233BFULL, "6113080_m.fxe"},
	{0xD0E8AA767B993647ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.impa.gad"},
	{0xD0E8D49CA05CBCD5ULL, "dae000_bartrand6027827.cl"},
	{0xD0E8EB46A24B773AULL, "fx_st_cloud.mao"},
	{0xD0E9CDEEB62AA38FULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_c.mmh"},
	{0xD0EAB56475135B24ULL, "6114925_m.fxe"},
	{0xD0EB9AF75F3CB17BULL, "dae000ip_crowd_lm4_talk7.utp"},
	{0xD0EC2719943867BBULL, "vxht_mansion2_33x_0_0.vlm"},
	{0xD0EC37FA612A1FD5ULL, "672405555.ogg"},
	{0xD0EC597683DA2EC3ULL, "361684427.ogg"},
	{0xD0EC72EDEFE718EEULL, "vxht_keep_6g4_0_0.vlm"},
	{0xD0ECD4C7861422A8ULL, "vxw_encounter2_pt_0_0.vlm"},
	{0xD0ED021868FC26F6ULL, "288904652.ogg"},
	{0xD0ED2B5FFA3D2CA8ULL, "hlac_lt_0_lowtown_docks_a3_-65537_4z.dds"},
	{0xD0ED6C0475C45E0DULL, "art\\vfx\\cinematic\\vfx_spark_imp_p.impa.evt"},
	{0xD0EDDBE1269DA848ULL, "601205370.ogg"},
	{0xD0EE0E22291AC7AAULL, "narrative_varric6124298.cl"},
	{0xD0EE7759CBCF0D6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down_less.evt"},
	{0xD0EE8F0EA2AC35D5ULL, "vxw_sundermount_27w_0_0.vlm"},
	{0xD0EF1EFB0CF58F4EULL, "47624717.ogg"},
	{0xD0EF23218679B18FULL, "mag291_torpor6026933.cl"},
	{0xD0EF9F767B9F4254ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.impa.evt"},
	{0xD0F079501D8DF2A3ULL, "1067827897.ogg"},
	{0xD0F0D81ACB008BACULL, "1026443112.ogg"},
	{0xD0F124B1010A7DFAULL, "lt_0_lowtown_docks_a3_n_99_0.mmh"},
	{0xD0F128B1010A8431ULL, "lt_0_lowtown_docks_a3_n_99_0.mao"},
	{0xD0F136B1010A9C40ULL, "lt_0_lowtown_docks_a3_n_99_0.msh"},
	{0xD0F13E1CE0F04E23ULL, "ht_0_hightown_a3_53_0.anb"},
	{0xD0F1485FBC0932F0ULL, "hlac_ht_playershouse_n_855_0y.dds"},
	{0xD0F1F9F85A4B955FULL, "vxht_0_hightown_a3_bgd_0_0.vlm"},
	{0xD0F2185A7B1EC3D1ULL, "vxw_sundermount_33f_0_0.vlm"},
	{0xD0F23D1102486FE6ULL, "6099195_f.fxe"},
	{0xD0F2CC8CB817E483ULL, "sse_ss_taraneh.gda"},
	{0xD0F2CCDC751C7506ULL, "6127275_m.fxe"},
	{0xD0F4433A65377F6CULL, "gen00fl_aveline6122802.cl"},
	{0xD0F467D9C332853CULL, "vxht_keep_42s_0_0.vlm"},
	{0xD0F47859CBD37A18ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_look_down_less.gad"},
	{0xD0F496EBB22846ADULL, "hlac_lt_0_lowtown_docks_n_-196610_4x.dds"},
	{0xD0F4DAA01AA11BB3ULL, "vxw_encounter2_camp_kr_0_0.vlm"},
	{0xD0F585255DD0C12FULL, "104805203.ogg"},
	{0xD0F5C94BCDA4C6A7ULL, "6071490_m.fxe"},
	{0xD0F5DB8B420CDFC1ULL, "vxlt_gallowstemplar_f_2pb_0_0.vlm"},
	{0xD0F66487628EA434ULL, "vxw_woundedcoast_b7_0_0.vlm"},
	{0xD0F790B07FA3975DULL, "113860279.ogg"},
	{0xD0F81FEF49331A19ULL, "vxht_mansion_n_5cs_0_0.vlm"},
	{0xD0F83C005D96293BULL, "vxlt_warehouse_rv_0_0.vlm"},
	{0xD0F8EF70FE620337ULL, "montage_subs_demohack6217153.cl"},
	{0xD0F9178F546143F6ULL, "fen210_hadriana6110535.cl"},
	{0xD0FA9641019C246EULL, "106314922.ogg"},
	{0xD0FAD38EC832B72DULL, "hlac_w_bonepitdestroyed_a3_0_12x.dds"},
	{0xD0FAD98014D5526DULL, "hf_isa_letter_thug.mor"},
	{0xD0FB2658EA08A650ULL, "6530745.ogg"},
	{0xD0FB6449F7F27EC4ULL, "one300_zevran6076502.cl"},
	{0xD0FB7CBCB7EEB37DULL, "dae000_pc_home_ambients6208317.cl"},
	{0xD0FB8EB41B7E211FULL, "360243834.ogg"},
	{0xD0FBFEC1FC2F4205ULL, "6016806_m.fxe"},
	{0xD0FC37F99293AA11ULL, "105875715.ogg"},
	{0xD0FC57268C6F6FDFULL, "plt_cod_aow_spellcombo1.plo"},
	{0xD0FC97767BA9BE8DULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.impa.ani"},
	{0xD0FCAACCF574CCADULL, "6101979_m.fxe"},
	{0xD0FCC26C953E4B88ULL, "fx_blood01.mao"},
	{0xD0FD00EB94A24E55ULL, "vxw_sundermount_28v_0_0.vlm"},
	{0xD0FD673E00B7FE0AULL, "zz_lgt_2_debug6211872.cl"},
	{0xD0FDD849FE5092B3ULL, "vxca_caverns2prc_8ek_0_0.vlm"},
	{0xD0FE1959265BB5F8ULL, "69464612.ogg"},
	{0xD0FE3859E771494AULL, "vxca_caverns1_mines_edg_0_0.vlm"},
	{0xD0FE64EF19CDF7BEULL, "6143729_m.fxe"},
	{0xD0FF1019699D2BAEULL, "60983048.ogg"},
	{0xD0FF2E2B4E8596F2ULL, "zz_qcr_debug6053502.cl"},
	{0xD0FFABB1CD8587CCULL, "6090842_f.fxe"},
	{0xD0FFC34BA92F0BE6ULL, "vxca_cavern1_haunted_icr_0_0.vlm"},
	{0xD100071E8BF3F771ULL, "vxlt_0_lowtown_docks_bb5_0_0.vlm"},
	{0xD1006168868BC109ULL, "dae000_arishok6067727.cl"},
	{0xD10147916187B3F2ULL, "ser191_emeric6107883.cl"},
	{0xD101550E2A6F1EFCULL, "6147183_m.fxe"},
	{0xD1015FCB09B3641FULL, "lt_0_lowtown_docks_a3_202_0.msh"},
	{0xD1016390681A7CA1ULL, "zz_eric_0124.cut"},
	{0xD10165CB09B36EC1ULL, "lt_0_lowtown_docks_a3_202_0.mmh"},
	{0xD10171CB09B38362ULL, "lt_0_lowtown_docks_a3_202_0.mao"},
	{0xD1020A9A99A62CDFULL, "vxht_keep_55h_0_0.vlm"},
	{0xD102978C2DC1D019ULL, "887765107.ogg"},
	{0xD102A8E0EB9738EEULL, "540725083.ogg"},
	{0xD102AA66AE516399ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_ladder.msh"},
	{0xD102B466AE517487ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_ladder.mmh"},
	{0xD102B866AE517BD4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_ladder.mao"},
	{0xD102BDE382479B00ULL, "qun211cr_henchmen2.cnv"},
	{0xD102F8858995D230ULL, "mag100_cullen6140962.cl"},
	{0xD1034C7D0C973433ULL, "vxw_encounter2_camp_fu_0_0.vlm"},
	{0xD10367B235E0ABBBULL, "460283891.ogg"},
	{0xD104B37B8EF41916ULL, "zz_dae_debug6029521.cl"},
	{0xD105A851DB5BB13AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_010a_0n.dds"},
	{0xD105CF24645659A6ULL, "1051084441.ogg"},
	{0xD106506FB6D83473ULL, "6001189_m.fxe"},
	{0xD106540E066EE2EEULL, "lowtown_fem_cityguard_mob6164820.cl"},
	{0xD106837167EA5B63ULL, "6050091_m.fxe"},
	{0xD1068AA47E0BD804ULL, "vxlt_0_lowtown_a3_n_c_ei3_0_0.vlm"},
	{0xD1069FE3824B5138ULL, "qun211cr_henchmen2.bnk"},
	{0xD106ADDD4F8C41BBULL, "vxlt_warehouse_wi_0_0.vlm"},
	{0xD106BFE4E0D49119ULL, "vxlt_hangedman_244_0_0.vlm"},
	{0xD106E8A24C350E41ULL, "4091195.ogg"},
	{0xD1074F2623C72452ULL, "550786249.ogg"},
	{0xD1076EBA221B508CULL, "6068789_m.fxe"},
	{0xD10780C3A8DBA96DULL, "vxw_encounter3_ea_0_0.vlm"},
	{0xD107A7E58FB7B6F6ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v4.evt"},
	{0xD107AF2D82F9C963ULL, "vxlt_hangedman_20z_0_0.vlm"},
	{0xD107C446A265E5ECULL, "fx_st_cloud.dds"},
	{0xD1084C616F7FDB53ULL, "735299523.ogg"},
	{0xD10906B6C5CBADA6ULL, "ave200_donnic6110697.cl"},
	{0xD109514A1B1EF6B7ULL, "vxca_cave_2bb_0_0.vlm"},
	{0xD1097DCBC9C1581EULL, "435637136.ogg"},
	{0xD10ABED4A1523BDBULL, "art\\vfx\\combat\\vfx_gib_hlk_p_1.mao"},
	{0xD10B09EA1FBD7FE7ULL, "48855380.ogg"},
	{0xD10B5402D5B0A64CULL, "vxht_0_hightown_a2_c7s_0_0.vlm"},
	{0xD10BD6050AFC7F01ULL, "6035324_m.fxe"},
	{0xD10CB0A2CA6BDE6EULL, "plt_vrc221pt_family_matter.plo"},
	{0xD10D9EE58FBC1281ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v4.gad"},
	{0xD10DF46D2A3514FCULL, "10567258.ogg"},
	{0xD10E0493275BE43BULL, "one120_athenril6088899.cl"},
	{0xD10E06C2EE0ABB93ULL, "vxlt_undercitydungeon_18o_0_0.vlm"},
	{0xD10E67AEBE79950CULL, "hlac_ht_chantry_n_2534_39z.dds"},
	{0xD10E787599D4B072ULL, "6126311_m.fxe"},
	{0xD10E7A8EABD9EDDCULL, "995272017.ogg"},
	{0xD10E84C7099DDC7CULL, "584580284.ogg"},
	{0xD10EEA1AE29A1A3AULL, "vxlt_gallowscourtyard_7b3_0_0.vlm"},
	{0xD10F111137DB2403ULL, "402722079.ogg"},
	{0xD10F2EEAF3691E12ULL, "t3_wdg_gr1.tnt"},
	{0xD10F5BB7D48993A5ULL, "hlac_dr_dungeon_3215_6z.dds"},
	{0xD10F935147B5D8A8ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd01a.mmh"},
	{0xD10FB15147B60BD2ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd01a.msh"},
	{0xD10FE8DB35ADCE17ULL, "genip_exploding_barrel_red.utp"},
	{0xD1106B1F170504F0ULL, "6067991_m.fxe"},
	{0xD110F0D744DA19ABULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_abma_0t.dds"},
	{0xD110F6458E4A4382ULL, "vxht_brothel_n_4x3_0_0.vlm"},
	{0xD110FE305E810F73ULL, "zz_ave_debug6207524.cl"},
	{0xD1110812CC2FE284ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vault_n.dds"},
	{0xD11149D6415686CFULL, "zz_followers_banter6157269.cl"},
	{0xD111ADDC9008397AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic.phy"},
	{0xD111CEEEB64BF7CCULL, "art\\vfx\\combat\\vfx_m_sta_atk_ele_c.anb"},
	{0xD1121F0DD16BCF4FULL, "289817259.ogg"},
	{0xD11248B250AE5325ULL, "hlac_w_sundermount_65533_1y.dds"},
	{0xD113136C6B1D6B10ULL, "vxlt_0_lowtown_a3_n_c_fgz_0_0.vlm"},
	{0xD1131D79F7EB5E44ULL, "vxw_encounter1_eb_0_0.vlm"},
	{0xD1134841467AD534ULL, "vxw_woundedcoast_t5_0_0.vlm"},
	{0xD1134BFEDA58DE9FULL, "409498094.ogg"},
	{0xD113518A361C5334ULL, "130943187.ogg"},
	{0xD1138D27BE9AF558ULL, "gen_im_trs_qun_amu_r0.uti"},
	{0xD1141E2D5F3FA065ULL, "gen_im_arm_cht_comb_t3.uti"},
	{0xD114929A0C21DE42ULL, "6144386_m.fxe"},
	{0xD11493D13986C5C4ULL, "6039714_m.fxe"},
	{0xD1149FE58FC2332FULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v4.ani"},
	{0xD114F424B2BB8EE7ULL, "512850648.ogg"},
	{0xD1151FBA41825A3BULL, "vxlt_warehouse_xh_0_0.vlm"},
	{0xD11540F615AAF1F1ULL, "vxht_keep_n_4b1_0_0.vlm"},
	{0xD1158A15839A1284ULL, "vxht_keep_n_3if_0_0.vlm"},
	{0xD11610D401540AADULL, "hm_amb_keep_guard2.mor"},
	{0xD116485B9DB0F362ULL, "vxca_caverns1_ima_0_0.vlm"},
	{0xD116746286E54E7BULL, "vxht_mansion2b_1ax_0_0.vlm"},
	{0xD1172ECA71309822ULL, "vxlt_0_lowtown_a2_n_gmx_0_0.vlm"},
	{0xD11750DB01FB8761ULL, "vxlt_gallowstemplar_f_3rq_0_0.vlm"},
	{0xD117A9863B8AFEFFULL, "vxht_0_hightown_n_aqk_0_0.vlm"},
	{0xD117B834FED9DFA9ULL, "hlac_lt_0_lowtown_n_-65535_2x.dds"},
	{0xD117CD8D617BFAEDULL, "vxht_keep_6au_0_0.vlm"},
	{0xD1192F03B61A7A29ULL, "393891673.ogg"},
	{0xD119714565AC0C89ULL, "vxht_0_hightown_bcc_0_0.vlm"},
	{0xD1197F6612BD76F9ULL, "one100dg_kelder6055623.cl"},
	{0xD119C5DFC7A6BECCULL, "vxht_0_hightown_a2_c2p_0_0.vlm"},
	{0xD11A02435EBB5F27ULL, "1025979855.ogg"},
	{0xD11A936C9556F9EAULL, "fx_blood01.dds"},
	{0xD11AD8A077F4830FULL, "t3_wpn_elm.tnt"},
	{0xD11ADA6E75DFD02AULL, "vxlt_0_lowtown_a2_n_ffx_0_0.vlm"},
	{0xD11B03049B51EE38ULL, "vxht_bartrand_n_5rb_0_0.vlm"},
	{0xD11B21897BB161D3ULL, "follower_banter6153991.cl"},
	{0xD11B4FB6A3002B71ULL, "vxw_bonepitdestroyed_a3_1v8_0_0.vlm"},
	{0xD11B5D72B99B1BB8ULL, "llac_w_blightlands_3_0.dds"},
	{0xD11B622FF64C34FCULL, "vxlt_gallowscourtyard_70s_0_0.vlm"},
	{0xD11BB87459A2F582ULL, "vxlt_0_lowtown_a2_h29_0_0.vlm"},
	{0xD11BCAD1022393EDULL, "6096460_m.fxe"},
	{0xD11CF58DE88B956CULL, "vxht_mansion2b_25t_0_0.vlm"},
	{0xD11D98E261508E71ULL, "22386916.ogg"},
	{0xD11DE27DD943A28BULL, "12822642.ogg"},
	{0xD11E2276DE2EBBB5ULL, "6066171_m.fxe"},
	{0xD11E539E74764BF2ULL, "6172505_m.fxe"},
	{0xD11E8EE298FD262DULL, "vxht_mansion2b_1tm_0_0.vlm"},
	{0xD11EEAFD0D73958AULL, "gen00fl_fenris6115296.cl"},
	{0xD11F3E2FB6F5DAD0ULL, "vxlt_darktown_48n_0_0.vlm"},
	{0xD11F49786B143219ULL, "vxca_caverns1_mines_fmx_0_0.vlm"},
	{0xD11F69560CB2BAEBULL, "6104941_m.fxe"},
	{0xD11FD50B4C4FC751ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_enter.ani"},
	{0xD11FFA606E7EADAFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1a.mao"},
	{0xD12006606E7EC24CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1a.mmh"},
	{0xD1200C606E7ECC3EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paper1a.msh"},
	{0xD12017FD0DAAB5A0ULL, "vxca_caverns2prc_9l2_0_0.vlm"},
	{0xD120728571CDA3C3ULL, "vxlt_gallowstemplar_n_4ws_0_0.vlm"},
	{0xD120A1A2CA7C6DFDULL, "plt_vrc221pt_family_matter.nss"},
	{0xD120D990C3503348ULL, "vxlt_gallowstemplar_f_3rt_0_0.vlm"},
	{0xD1218F56E9E176C4ULL, "vxw_encounter1_jc_0_0.vlm"},
	{0xD121CE80F7186BFAULL, "llac_lt_undercitydungeon_-1_14.dds"},
	{0xD12239E52CF44CF5ULL, "483140625.ogg"},
	{0xD1223CF0D45C54A9ULL, "pro000_colonel_ewald6005400.cl"},
	{0xD1228E665F911655ULL, "6153697_m.fxe"},
	{0xD122A5349B793615ULL, "vxlt_darktown_6la_0_0.vlm"},
	{0xD123576D1B909430ULL, "6105956_m.fxe"},
	{0xD12375861F156399ULL, "vxht_mansion_n_5rp_0_0.vlm"},
	{0xD123B2D307A10A71ULL, "vxht_keep_n_4i0_0_0.vlm"},
	{0xD123D4E11D967CB7ULL, "709992184.ogg"},
	{0xD124A829AC2D7C2FULL, "42738096.ogg"},
	{0xD124C91F8E655CCAULL, "vxht_0_hightown_a2_byb_0_0.vlm"},
	{0xD12545A2909811D5ULL, "6166194_m.fxe"},
	{0xD12547337DE42542ULL, "6104203_m.fxe"},
	{0xD125583F5F3A63BAULL, "vxdr_primevaltomb_868_0_0.vlm"},
	{0xD1256C801110ADC4ULL, "mag320_orsino6071128.cl"},
	{0xD125B3FC6B0AD5F8ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wo_basic_1.mao"},
	{0xD125D951AD734630ULL, "gen_im_trs_rdr_one_lsw_r0.uti"},
	{0xD125E00B4C5445F3ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_enter.gad"},
	{0xD1262307B5F7705FULL, "6134326_m.fxe"},
	{0xD126D493F1BBEBF0ULL, "770704061.ogg"},
	{0xD1274160B5CFB450ULL, "gen00fl_anders6096463.cl"},
	{0xD12747ED727E5348ULL, "vxlt_backalley_n_1vs_0_1.vlm"},
	{0xD12784813F0A54A6ULL, "vxlt_0_lowtown_a2_n_ewp_0_0.vlm"},
	{0xD127E344A9614865ULL, "lgt220cr_shp_mag_devrona.utc"},
	{0xD128326EA0F2DA6AULL, "vxht_keep_3fz_0_0.vlm"},
	{0xD128C04EB5F3090EULL, "ave150tr_approach_ambush.utt"},
	{0xD128FEA7443ADC37ULL, "gen00fl_merrill6145144.cl"},
	{0xD1292BB0660863DEULL, "hlac_lt_backalley_n_1_6x.dds"},
	{0xD12974E18D4806B8ULL, "vxht_bartrand_n_5uc_0_0.vlm"},
	{0xD129D40CE8424D7CULL, "vxlt_gallowscourtyard_75p_0_0.vlm"},
	{0xD12AAE242EBB1DAEULL, "442628574.ogg"},
	{0xD12AD83D60DDA91CULL, "6207703_f.fxe"},
	{0xD12B446B4323C2D2ULL, "gen00fl_aveline6139484.cl"},
	{0xD12B688ADA931465ULL, "875532960.ogg"},
	{0xD12C3F449C27AEDCULL, "mag200_feynriel6023887.cl"},
	{0xD12C6E4ED48951EEULL, "llac_lt_0_lowtown_a3_n_c_-65536_1.dds"},
	{0xD12CC57B666C6472ULL, "6175405_m.fxe"},
	{0xD12CDD0B4C5A5F58ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_right_90_enter.evt"},
	{0xD12D05A813BA1733ULL, "mag360_thrask6042373.cl"},
	{0xD12D173C8336EF40ULL, "hlac_ht_0_hightown_a2_-131070_0z.dds"},
	{0xD12D364C09BBD9D1ULL, "6119978_m.fxe"},
	{0xD12DCC84A63D92DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmtap.ani"},
	{0xD12E0ED8692A3662ULL, "vxlt_warehouse_q2_0_0.vlm"},
	{0xD12E16D99591525BULL, "vxht_bartrand_n_5o2_0_0.vlm"},
	{0xD12EA3794958B037ULL, "738971572.ogg"},
	{0xD12EE28131E5328CULL, "zz_vault_debug6203314.cl"},
	{0xD12F92A87CA54B02ULL, "hm_taraneh.mor"},
	{0xD13009377AE9CA91ULL, "406517242.ogg"},
	{0xD130592CCD93A782ULL, "fadeclouds.mao"},
	{0xD13078D831B8BDC6ULL, "gen00fl_carver6097321.cl"},
	{0xD130BC515CFF8DB0ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgte_0d.dds"},
	{0xD130EF46DCF7E428ULL, "dre000_hireling26038681.cl"},
	{0xD13111026685B262ULL, "908736520.ogg"},
	{0xD131A441E7DA72AEULL, "6096012_m.fxe"},
	{0xD131EE076AC8B7CBULL, "6039003_m.fxe"},
	{0xD13211158EFFEDD9ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v4.ani"},
	{0xD132DC6DF3DD9B4AULL, "follower_banter6136669.cl"},
	{0xD132E8D6904CD0C0ULL, "gen00fl_aveline6132847.cl"},
	{0xD133A46D22C7272EULL, "gen00fl_anders6093462.cl"},
	{0xD133C4FD130836C5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_smelt_n.dds"},
	{0xD133C795B0713BB0ULL, "6096826_f.fxe"},
	{0xD133FAC66195B7A4ULL, "hlac_w_woundedcoast_327688_0y.dds"},
	{0xD134051BFB6ADB29ULL, "vxw_bonepit_p9_0_0.vlm"},
	{0xD1344FECFAB312E0ULL, "6172547_m.fxe"},
	{0xD134F984A0A07F9FULL, "6110913_m.fxe"},
	{0xD13510CC8E25878EULL, "6020545_m.fxe"},
	{0xD135D045CD52D27BULL, "hlac_ht_mansion_n_2995_2x.dds"},
	{0xD135F65E31006D26ULL, "vxlt_0_lowtown_a2_n_exq_0_0.vlm"},
	{0xD13636601C04F5B9ULL, "hlac_dr_primevaltomb_658_3y.dds"},
	{0xD136494D7C2CA1E7ULL, "fen110_anso6095020.cl"},
	{0xD13652E10C79D09BULL, "w_bonepitmining_a2_137_0.msh"},
	{0xD13654FF49983D89ULL, "vxht_0_hightown_bia_0_0.vlm"},
	{0xD1365B87BE677A28ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_masc_0s.dds"},
	{0xD13664E10C79EFE6ULL, "w_bonepitmining_a2_137_0.mao"},
	{0xD13670E10C7A0385ULL, "w_bonepitmining_a2_137_0.mmh"},
	{0xD13688AD4BE6F00CULL, "gen_im_trs_tmp_hvy_hlm_r1.uti"},
	{0xD136C95B56DDD972ULL, "988904365.ogg"},
	{0xD136FAF9A6929DE7ULL, "zz_one_debug6099769.cl"},
	{0xD13740566A7CDBD4ULL, "vxht_0_hightown_a2_c3n_0_1.vlm"},
	{0xD137EA040AA50879ULL, "376847008.ogg"},
	{0xD1383768AF11D955ULL, "6077329_f.fxe"},
	{0xD138CFCC8D20EC53ULL, "60791112.ogg"},
	{0xD138D8501E78A842ULL, "950217910.ogg"},
	{0xD138EC158F05CD1BULL, "art\\characters\\creatures\\ragedemon\\animation\\p_tw_v4.gad"},
	{0xD13A0B8F12B1801FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_loop.gad"},
	{0xD13A4450191B749DULL, "fen310_party_comment6116856.cl"},
	{0xD13A7922F2669F64ULL, "dae000cr_grand_cleric6148150.cl"},
	{0xD13BD484A649DDD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmtap.evt"},
	{0xD13C6C26C36B07DBULL, "hlac_lt_0_lowtown_n_65533_1x.dds"},
	{0xD13C71F31E914AACULL, "tv_cl_blend.mao"},
	{0xD13C80B55B204EE2ULL, "vxlt_warehouse_z3_0_0.vlm"},
	{0xD13C8EF8594A6E39ULL, "403387954.ogg"},
	{0xD13CEF623754B810ULL, "378400723.ogg"},
	{0xD13D663264015556ULL, "27638400.ogg"},
	{0xD13D92F0937D18D2ULL, "6013424_m.fxe"},
	{0xD13DB2001B3BD4F3ULL, "hlac_lt_0_lowtown_docks_a3_-1_3y.dds"},
	{0xD13E49DCC16EDC95ULL, "223725802.ogg"},
	{0xD13E8784393A07B2ULL, "6075385_m.fxe"},
	{0xD13EF8E7DF72761CULL, "vxht_0_hightown_a3_alm_0_0.vlm"},
	{0xD13F52927FD6D047ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs08a_0t.dds"},
	{0xD13F8046051B7A48ULL, "hlac_lt_0_lowtown_docks_a3_n_-131071_0x.dds"},
	{0xD13F8A6B8B93E10FULL, "6054046_m.fxe"},
	{0xD1400F5112852D74ULL, "vxw_sundermount_2w2_0_0.vlm"},
	{0xD14040B3F6F563A1ULL, "vxw_woundedcoast_mw_0_0.vlm"},
	{0xD1406B4125FE358DULL, "vxlt_warehouse_190_0_0.vlm"},
	{0xD14079B0267D604AULL, "dae370ar_gallows.ncs"},
	{0xD140B34EB6078537ULL, "ave150tr_approach_ambush.ncs"},
	{0xD140DFAC597C5C04ULL, "vxht_keep_n_3xe_0_0.vlm"},
	{0xD141935D0DEA56ACULL, "hlac_lt_0_lowtown_a2_65535_0y.dds"},
	{0xD142F384A6503165ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_palmtap.gad"},
	{0xD144ACAD344ABA66ULL, "6118348_exitdoor1.cut"},
	{0xD144BAC95A719CE7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgta_0t.dds"},
	{0xD144FCE7B532EF6AULL, "vxdr_primevaltomb_6m7_0_0.vlm"},
	{0xD14535BA32A983CFULL, "art\\characters\\creatures\\buzzard\\animation\\blendtree\\buzzardperched.anb"},
	{0xD1462236D9705454ULL, "isa111_cut_isabela_intro.cnv"},
	{0xD1466D3F9186CD41ULL, "tbe200im_sliver_bow.uti"},
	{0xD14691B02681F4DBULL, "dae370ar_gallows.lst"},
	{0xD1472A676024A70BULL, "vxlt_0_lowtown_ekj_0_0.vlm"},
	{0xD14749D3ED8A28C7ULL, "hlac_w_woundedcoast_65536_0y.dds"},
	{0xD14796A74FB4EFBAULL, "vxht_0_hightown_n_cbq_0_0.vlm"},
	{0xD147BDDA53935C3FULL, "one300_murder_followers6086897.cl"},
	{0xD147D9ECC62691BEULL, "6092180_m.fxe"},
	{0xD147DADE1FFFDD4EULL, "244440945.ogg"},
	{0xD1488D3E231C9F0AULL, "917255032.ogg"},
	{0xD1490436D9725694ULL, "isa111_cut_isabela_intro.bnk"},
	{0xD149098F2DDA9CD8ULL, "vxw_bonepitdestroyed_a3_27i_0_0.vlm"},
	{0xD149E5B37EAF7C09ULL, "6154739_m.fxe"},
	{0xD14A9070CEF201DFULL, "76102281.ogg"},
	{0xD14A9769F2BBE974ULL, "80892232.ogg"},
	{0xD14BD6D395C5B028ULL, "follower_banter6164552.cl"},
	{0xD14CF278E09E85ECULL, "598370592.ogg"},
	{0xD14D6AC4D1688E9CULL, "vxht_0_hightown_a3_aib_0_0.vlm"},
	{0xD14DB7589848A5BAULL, "qun221_viscount6073894.cl"},
	{0xD14DF51DAB9AAB68ULL, "llac_lt_gallowstemplar_2383_11.dds"},
	{0xD14E055A7521205BULL, "lt_0_lowtown_docks_a3_n_139_0.mmh"},
	{0xD14E115A752134F8ULL, "lt_0_lowtown_docks_a3_n_139_0.mao"},
	{0xD14E235A75215345ULL, "lt_0_lowtown_docks_a3_n_139_0.msh"},
	{0xD14EA185BF34C5A6ULL, "vxca_caverns1_mines_i1x_0_0.vlm"},
	{0xD14ECC0929EABDE0ULL, "vxdr_0_deeproads_ef9_0_0.vlm"},
	{0xD14EE88F12C3A165ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_hand_on_hip_loop.ani"},
	{0xD14F6C1AA27637EEULL, "vxht_mansion_n_32m_0_0.vlm"},
	{0xD14FFA62E4EBFA1EULL, "736300887.ogg"},
	{0xD1500AA67D2B1741ULL, "vxlt_hangedman_lz_0_0.vlm"},
	{0xD1501962FBF6469FULL, "vxlt_0_lowtown_docks_n_bmy_0_0.vlm"},
	{0xD15076147985225DULL, "27873354.ogg"},
	{0xD150903848323C56ULL, "vxht_0_hightown_a3_an1_0_0.vlm"},
	{0xD150A33C8F9CDAE1ULL, "vxca_caverns2_96h_0_0.vlm"},
	{0xD150C71704C8C93AULL, "vxht_keep_5zx_0_0.vlm"},
	{0xD150D1267CB6C1B8ULL, "6125990_m.fxe"},
	{0xD1516D6804BDD85EULL, "hlac_lt_0_lowtown_docks_n_-1_1x.dds"},
	{0xD151D7CA7E4CBEFCULL, "6040970_m.fxe"},
	{0xD1529F50CB05A0F3ULL, "185269336.ogg"},
	{0xD152CFE8B7E3E0BAULL, "vxw_bonepitdestroyed_a3_1s0_0_0.vlm"},
	{0xD152E94D411A62EDULL, "6067734_m.fxe"},
	{0xD15367ACA0FCA8B1ULL, "6119524_f.fxe"},
	{0xD153719FD639D45CULL, "zz_seb_debug6189575.cl"},
	{0xD153733F66D85FBAULL, "llac_ca_caverns1_mines_24421_10.dds"},
	{0xD153B293B7254130ULL, "4507392.ogg"},
	{0xD1553E6315F59813ULL, "gen_im_npc_elf_arm_cht_enca.uti"},
	{0xD155444B09EEE2EAULL, "797394434.ogg"},
	{0xD15575822B9A9332ULL, "llac_lt_0_lowtown_a2_n_c_131072_2.dds"},
	{0xD155A1A6DE297CF4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tn90r.ani"},
	{0xD155E670DE5A7FF0ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\letterback_d.dds"},
	{0xD1562DF5488243DEULL, "vxca_caverns1_mines_iix_0_0.vlm"},
	{0xD1566A39DC92D56FULL, "gen00fl_merrill6109786.cl"},
	{0xD156D04804409FBBULL, "llac_dr_0_deeproads_5842_6.dds"},
	{0xD156F9BD2D51F0D4ULL, "6166421_m.fxe"},
	{0xD157961477B5D4BAULL, "follower_banter6170098.cl"},
	{0xD157A5377184136FULL, "hlac_dr_0_deeproads_5839_3y.dds"},
	{0xD157CC0AF2312494ULL, "atl_dialoguepo_dxt5_00.dds"},
	{0xD157CC53F12C2F7DULL, "6016958_m.fxe"},
	{0xD1580B5689E7C430ULL, "gen00fl_carver6102339.cl"},
	{0xD1583F0CD9B82707ULL, "vxlt_0_lowtown_a3_n_c_ghq_0_0.vlm"},
	{0xD1584E64FCFF7001ULL, "w_bonepitmining_a2_29_0_lc.mmh"},
	{0xD159131920732F91ULL, "ss_chamber.dds"},
	{0xD15914521DA90BD7ULL, "mag320_orsino6038562.cl"},
	{0xD159BCB3C7DDE833ULL, "gen00fl_merrill6129854.cl"},
	{0xD159CC510A3136DCULL, "llac_lt_undercitydungeon_-1_1.dds"},
	{0xD15A0CF9FD1BEB06ULL, "gen00fl_fenris6110871.cl"},
	{0xD15A2BF250F186C8ULL, "vxca_caverns2_96m_0_0.vlm"},
	{0xD15A2F084AF76FD0ULL, "1050445931.ogg"},
	{0xD15A6DC93A44FE37ULL, "1040607144.ogg"},
	{0xD15B69299ED9350FULL, "resource_state.gda"},
	{0xD15B9EA6DE2DE356ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_tn90r.gad"},
	{0xD15BE037DBDFA79BULL, "6121751_m.fxe"},
	{0xD15BEFC3000A708BULL, "mer221_sabin6080570.cl"},
	{0xD15C44D3B332EEF6ULL, "llac_dr_primevaltomb_657_4.dds"},
	{0xD15C6998B94DDFD1ULL, "6126730_m.fxe"},
	{0xD15C86F28E6835AFULL, "hlac_lt_0_lowtown_n_131071_2z.dds"},
	{0xD15C94918A95FB52ULL, "ht_0_hightownexplosion_n_34_0_lc.mmh"},
	{0xD15CBCFECD255DFAULL, "hlac_lt_0_lowtown_a3_n_c_65538_2z.dds"},
	{0xD15CDD87F91E7549ULL, "vxw_woundedcoast_62_0_0.vlm"},
	{0xD15D222D9B45D14DULL, "vxht_mansion2_1te_0_0.vlm"},
	{0xD15D246DDAE194A1ULL, "vxw_woundedcoast_gu_0_0.vlm"},
	{0xD15D990E8B59497DULL, "dae000_ineria6136868.cl"},
	{0xD15E28D88BA5451AULL, "converse.cur"},
	{0xD15E3D6FB1D622EFULL, "dre000_varric6027755.cl"},
	{0xD15E9EBDD4EB5534ULL, "vxw_bonepit_1m0_0_0.vlm"},
	{0xD15F0023B6E1042FULL, "vxlt_0_lowtown_docks_c4l_0_0.vlm"},
	{0xD15F5723DF86A063ULL, "6135790_m.fxe"},
	{0xD15F8114A0B84F20ULL, "74086113.ogg"},
	{0xD15F955C198AD3E4ULL, "590769301.ogg"},
	{0xD15FEEA12E9F9FBAULL, "zz_qa_open.utp"},
	{0xD15FF4AA30D47FB5ULL, "hlac_lt_0_lowtown_docks_n_-65540_2y.dds"},
	{0xD1602A3C38EFF64EULL, "llac_ca_caverns1_24423_8.dds"},
	{0xD16096E6DAB27F9AULL, "287580220.ogg"},
	{0xD160E3F69D53CE6CULL, "fen310_fenris6118702.cl"},
	{0xD160EB8768D78D7CULL, "6180970_f.fxe"},
	{0xD161A63B2898CFF7ULL, "90517294.ogg"},
	{0xD161E0A1991F206AULL, "vxdr_primevaltomb_6w1_0_0.vlm"},
	{0xD161E0CCEA7DEB81ULL, "vxht_keep_n_60y_0_0.vlm"},
	{0xD161E1F7093D048EULL, "6169192_m.fxe"},
	{0xD1623378B0ECAB6CULL, "hlac_lt_gallowstemplar_f_2394_12z.dds"},
	{0xD1629647420405ECULL, "llac_lt_backalley_n_65537_5.dds"},
	{0xD16389075CCEA5F1ULL, "vxw_bonepitdestroyed_a3_1o3_0_0.vlm"},
	{0xD163C26768AD049DULL, "9864486.ogg"},
	{0xD16494F4F8623AC1ULL, "171337159.ogg"},
	{0xD164A8019126AE5DULL, "454250351.ogg"},
	{0xD165C81E4EA1E62FULL, "vxht_keep_n_5ka_0_0.vlm"},
	{0xD1660FE3C0936C7AULL, "vxht_0_hightown_7iw_0_0.vlm"},
	{0xD166362E8BD7180DULL, "vxca_caverns2prc_a1d_0_0.vlm"},
	{0xD166C14E919BB3B8ULL, "906561171.ogg"},
	{0xD16708C8F546BEA9ULL, "6077348_m.fxe"},
	{0xD1673351A1E1B001ULL, "class_abilities_aveline.gda"},
	{0xD16744893E74C110ULL, "vxlt_0_lowtown_hab_0_0.vlm"},
	{0xD167E214DBB1F1E9ULL, "590738360.ogg"},
	{0xD1680F2DA564574FULL, "vxlt_undercitydungeon_1id_0_0.vlm"},
	{0xD1685D5E072C0750ULL, "683595259.ogg"},
	{0xD168E395AC8BCEECULL, "851032894.ogg"},
	{0xD1696BC9539D7613ULL, "vxlt_0_lowtown_a3_n_c_hsk_0_0.vlm"},
	{0xD169927685921573ULL, "plt_and000pt_main.nss"},
	{0xD16A29C0D959B526ULL, "vxlt_0_lowtown_docks_n_be4_0_0.vlm"},
	{0xD16A3D4214EC9449ULL, "llac_lt_0_lowtown_a3_n_c_-196608_3.dds"},
	{0xD16A5917FEF7CE89ULL, "vxlt_gallowstemplar_5qi_0_0.vlm"},
	{0xD16A7DD4A2AB6879ULL, "static_static_p19.psh.fxo.dx11.assembly"},
	{0xD16A939503FC768CULL, "hlac_lt_0_lowtown_a2_-131072_6x.dds"},
	{0xD16AAF6450E0860CULL, "t3_wpn_hid.tnt"},
	{0xD16B2010B05E8CCDULL, "264422016.ogg"},
	{0xD16B4F64EB148DC9ULL, "vxw_woundedcoast_31_0_0.vlm"},
	{0xD16B6095351267E2ULL, "prp_str_tentical_n.dds"},
	{0xD16BA6BBB1D77DEAULL, "340690925.ogg"},
	{0xD16C41098E5D1FB0ULL, "202459806.ogg"},
	{0xD16C649D52B26ECEULL, "t1_mub_or3.tnt"},
	{0xD16C72B03E183499ULL, "6021109_m.fxe"},
	{0xD16CA2CE3B1D4702ULL, "hlac_lt_0_lowtown_docks_a3_-196611_0z.dds"},
	{0xD16CA6DC9054D8A3ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic.msh"},
	{0xD16CAE6507F4C95CULL, "vxht_mansion_n_6nr_0_0.vlm"},
	{0xD16CB4DC9054F0FDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic.mmh"},
	{0xD16CB8DC9054F70EULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0dd_basic.mao"},
	{0xD16D73F22C1E6D56ULL, "vxht_0_hightown_a3_ap3_0_0.vlm"},
	{0xD16DE58D1B720E96ULL, "6108045_m.fxe"},
	{0xD16E4F8999A3CDACULL, "vxht_0_hightown_a3_c1d_0_0.vlm"},
	{0xD16E517E43A13E0EULL, "mrl150_dalish_guard6068733.cl"},
	{0xD16EBDA6B399E8C4ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_lit_p.mmh"},
	{0xD16F0CF3B5575863ULL, "rog_chancetostagger.dds"},
	{0xD16F39B0F1EBD002ULL, "art\\vfx\\materials\\vfx_smoke_puffy_glow.mao"},
	{0xD16F52E4EECD9E82ULL, "dae000cr_grand_cleric6173662.cl"},
	{0xD16F8C8CE6669BCEULL, "6033624_m.fxe"},
	{0xD16FB4B15CE84047ULL, "vxlt_gallowstemplar_n_4xr_0_0.vlm"},
	{0xD1704B1FA3B7486FULL, "817784529.ogg"},
	{0xD170B4B25F7AD633ULL, "art\\characters\\playercharacter\\children\\textures\\km_cth_boya_0n.dds"},
	{0xD170EB8C46E24B18ULL, "vxlt_0_lowtown_docks_a3_n_d2f_0_0.vlm"},
	{0xD170F09E6208568DULL, "character_rendertoshadowmap_p0.vsh.fxo"},
	{0xD171087A7E559689ULL, "6084409_m.fxe"},
	{0xD1711CEC06ADF2BAULL, "vxca_cavern1_haunted_i14_0_0.vlm"},
	{0xD171B32528EE0EF6ULL, "lt_0_lowtown_a3_n_c_52_0_lc.mmh"},
	{0xD1720D9D82D414CBULL, "vxht_mansion2_20t_0_0.vlm"},
	{0xD17232351D45E35DULL, "follower_banter6171085.cl"},
	{0xD172A2B026A74C18ULL, "dae370ar_gallows.are"},
	{0xD17311AF2C6E74DEULL, "vxca_caverns1_mines_icz_0_0.vlm"},
	{0xD1731382D348407CULL, "ss_fereldan_m_1.wbd"},
	{0xD173FC9687F5DC3BULL, "hlac_lt_0_lowtown_-3_1y.dds"},
	{0xD1744E2553D724F7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbar.msh"},
	{0xD174602553D7435AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbar.mao"},
	{0xD174642553D74A69ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbar.mmh"},
	{0xD1747144B9134432ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_sa.dds"},
	{0xD17482956FB990BCULL, "6110265_f.fxe"},
	{0xD17522C6BDA45807ULL, "vxlt_0_lowtown_a3_n_c_gbs_0_0.vlm"},
	{0xD17571EE65B3854CULL, "llac_lt_ruinedhouse_n_-1_2.dds"},
	{0xD175A1983CBB3367ULL, "vxlt_0_lowtown_a2_n_c_hkn_0_0.vlm"},
	{0xD1763FA219DDF893ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_hed_0e.dds"},
	{0xD1764D57DD3ABD5BULL, "283262134.ogg"},
	{0xD177611A1F4BE302ULL, "vxw_sundermount_3ag_0_0.vlm"},
	{0xD1776617A8232D3BULL, "art\\vfx\\cinematic\\vfx_bld_dir_p.impa.evt"},
	{0xD1776ABB0C1C235BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_l.ani"},
	{0xD178727E872DBF67ULL, "341973152.ogg"},
	{0xD178812E4EE67805ULL, "6087080_m.fxe"},
	{0xD1788DD03F77DCE4ULL, "gen00fl_dog6170616.cl"},
	{0xD178991E9B5FE190ULL, "18889460.ogg"},
	{0xD1789B9DCB4FCDA6ULL, "vxlt_0_lowtown_docks_n_bb5_0_0.vlm"},
	{0xD179B58D60E44B22ULL, "llac_lt_warehouse_n_-65536_0.dds"},
	{0xD179BD31A80D989FULL, "qun131_party_comment6117264.cl"},
	{0xD179BDB214A2B238ULL, "llac_lt_0_lowtown_docks_a3_-3_5.dds"},
	{0xD17A05E77F32024DULL, "vxht_mansion2_1bc_0_0.vlm"},
	{0xD17A0D287E60C39EULL, "gen00fl_bethany6097587.cl"},
	{0xD17BE5869290CDE7ULL, "gen00fl_carver6102947.cl"},
	{0xD17BEFBE4069FED5ULL, "hlac_w_encounter3_0_11z.dds"},
	{0xD17CFE26166B1FFEULL, "6075856_f.fxe"},
	{0xD17D49E10CB5C432ULL, "w_bonepitmining_a2_137_0.phy"},
	{0xD17D5F17A8278D7CULL, "art\\vfx\\cinematic\\vfx_bld_dir_p.impa.gad"},
	{0xD17E08D77269BBC3ULL, "6037603_m.fxe"},
	{0xD17E257F8DC62A3AULL, "vxw_bonepitdestroyed_a3_1b3_0_0.vlm"},
	{0xD17E270CC5A56F5EULL, "vxht_mansion2b_35k_0_0.vlm"},
	{0xD17E6D89F52AD465ULL, "pro000_opening_pc6001219.cl"},
	{0xD17E7FDDFA50F8BCULL, "6149501_m.fxe"},
	{0xD17F5628D3EA8851ULL, "gen00fl_carver6108950.cl"},
	{0xD17F720C06B9C200ULL, "357601042.ogg"},
	{0xD17FFE9CBB6BDDD7ULL, "6080761_f.fxe"},
	{0xD17FFEFA884135C1ULL, "gen00fl_sebastian6174044.cl"},
	{0xD18129509CE7EBA8ULL, "hlac_lt_0_lowtown_n_65536_5x.dds"},
	{0xD1818A2D25F4CDD0ULL, "6109804_m.fxe"},
	{0xD182CAA031ABFA8FULL, "gen00fl_carver6097263.cl"},
	{0xD182D60D4C785239ULL, "one300_zevran6076916.cl"},
	{0xD183032A25EB724EULL, "w_bonepitmining_a2_75_0.msh"},
	{0xD183112A25EB899FULL, "w_bonepitmining_a2_75_0.mao"},
	{0xD1831D2A25EB9E3CULL, "w_bonepitmining_a2_75_0.mmh"},
	{0xD18381A84EDF2419ULL, "hlac_lt_0_lowtown_docks_a3_-1_2x.dds"},
	{0xD1840020C326D1D4ULL, "mer220_hubert6041206.cl"},
	{0xD1840346D65674F6ULL, "hlac_lt_0_lowtown_a2_n_-65537_0y.dds"},
	{0xD18407E8ACDA9AAFULL, "6132676_m.fxe"},
	{0xD1845E17A82DAA7AULL, "art\\vfx\\cinematic\\vfx_bld_dir_p.impa.ani"},
	{0xD1850F79D94D88F8ULL, "arial12.mao"},
	{0xD1863BCBF9A43100ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_lgte_0t.dds"},
	{0xD18681A25EB7BB51ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_0z.dds"},
	{0xD1871BB7C291749CULL, "hlac_lt_gallowstemplar_n_2394_9z.dds"},
	{0xD187A1F99806A191ULL, "dae100cr_bran.utc"},
	{0xD187B1B04735F5B7ULL, "zz_one_debug6051509.cl"},
	{0xD187EB6B2679F1D3ULL, "6073755_m.fxe"},
	{0xD18862782CE7C1F1ULL, "mag360_samson6042045.cl"},
	{0xD1887A04B0C3DE21ULL, "vxw_woundedcoast_vj_0_0.vlm"},
	{0xD188955785FD96AAULL, "llac_ca_cave_2194_1.dds"},
	{0xD1889B21D3F3D079ULL, "llac_lt_0_lowtown_n_131070_3.dds"},
	{0xD1892424D61680B4ULL, "vxca_cavern1_haunted_jft_0_0.vlm"},
	{0xD1895E6D342589B4ULL, "289721605.ogg"},
	{0xD18A7E632FCA257EULL, "vxht_mansion2b_1bl_0_0.vlm"},
	{0xD18AFA11235496FFULL, "art\\vfx\\cinematic\\vfx_mage_punch_c.mmh"},
	{0xD18AFA574C59FFC9ULL, "vxlt_0_lowtown_docks_a3_coh_0_0.vlm"},
	{0xD18B55685BDBC46DULL, "w_blightlands_147_0.msh"},
	{0xD18B63685BDBDBA0ULL, "w_blightlands_147_0.mao"},
	{0xD18B67685BDBE2F3ULL, "w_blightlands_147_0.mmh"},
	{0xD18B7FA648506AF3ULL, "6075977_m.fxe"},
	{0xD18BE46C012A45E8ULL, "vxdr_primevaltomb_50x_0_1.vlm"},
	{0xD18C06DE0C5625FCULL, "vxca_cavern1_haunted_ixy_0_0.vlm"},
	{0xD18C5B6A970C9AE9ULL, "w_woundedcoast_135_0.phy"},
	{0xD18C91BB0C2EC1E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_l.gad"},
	{0xD18CE9AB5F3E3F36ULL, "6079658_f.fxe"},
	{0xD18D868184FF14BFULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_l_5p.gad"},
	{0xD18DAA60EFD1F258ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_enco_0t.dds"},
	{0xD18DF5F0D390E6B1ULL, "6093811_f.fxe"},
	{0xD18E34957796EA02ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs104a.msh"},
	{0xD18E36957796EDB8ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs104a.mmh"},
	{0xD18E42957797025BULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs104a.mao"},
	{0xD18E685FC2466E72ULL, "hlac_lt_gallowscourtyard_-65536_0x.dds"},
	{0xD18EAE85C417E7DEULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb107a_0e.dds"},
	{0xD18EBD3544CD2463ULL, "64515510.ogg"},
	{0xD18F83E38C8A1E1CULL, "mrl150_merrill6199327.cl"},
	{0xD18F9318BCBF5BBAULL, "1055789168.ogg"},
	{0xD190818BEA5BF98DULL, "6099241_m.fxe"},
	{0xD190D0BDE7413971ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.inactive.ani"},
	{0xD1913AC6AAFD0579ULL, "one220_yevhen6102820.cl"},
	{0xD191CF0EC80FEF4FULL, "547282492.ogg"},
	{0xD1923A88ED7EDD56ULL, "w_encounter3_101_0.phy"},
	{0xD192538CD9ED5D59ULL, "hlac_ht_store_187_2z.dds"},
	{0xD19267691DB2DCE2ULL, "w_sundermount_73_0.rml"},
	{0xD19287DCBB61DF79ULL, "6093566_m.fxe"},
	{0xD192D3EFB7B09583ULL, "follower_banter6156152.cl"},
	{0xD19319EECC8BBE60ULL, "mag291_torpor6026998.cl"},
	{0xD19373C3B7E75F77ULL, "vxw_encounter3_4l_0_0.vlm"},
	{0xD19373CFC1993748ULL, "6140865_m.fxe"},
	{0xD1937976C91E43D1ULL, "ht_0_hightown_a2_30_0_ll.mmh"},
	{0xD19400A0749EF83BULL, "gen_im_cft_reg_h1.uti"},
	{0xD1947B81850520ACULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_l_5p.evt"},
	{0xD1957F57AF3BFEA6ULL, "vxlt_0_lowtown_docks_n_bt7_0_0.vlm"},
	{0xD195D435433B31E1ULL, "6219276_m.fxe"},
	{0xD195E8BA0CDA4406ULL, "vxht_mansion2b_1mm_0_0.vlm"},
	{0xD196265A755F0D90ULL, "lt_0_lowtown_docks_a3_n_139_0.phy"},
	{0xD196583A212894ABULL, "6104372_m.fxe"},
	{0xD196B85F628A0C3AULL, "6106689_m.fxe"},
	{0xD196EBE1A2B9F6A1ULL, "vxw_woundedcoast_si_0_0.vlm"},
	{0xD1971468FA7C8E8EULL, "6147663_f.fxe"},
	{0xD197C0A6B3BCF2BBULL, "art\\vfx\\combat\\vfx_m_amg_estorm_lit_p.anb"},
	{0xD197CA8EBF52936CULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\harvester_pcoverride.anb"},
	{0xD197CB91CD5D93EDULL, "6173682_m.fxe"},
	{0xD197CDB94A38BD92ULL, "hlac_ht_mansion2_818_39x.dds"},
	{0xD197DBBDE7476A13ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_ext_trap.inactive.gad"},
	{0xD1983D95779F4E43ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs104a.phy"},
	{0xD19852B3B5C4C11FULL, "vxlt_0_lowtown_docks_n_bvd_0_0.vlm"},
	{0xD19872FEBBB09E90ULL, "mcr310_argument6127600.cl"},
	{0xD19939E934E6F3CBULL, "170354007.ogg"},
	{0xD1995AAB7087FAE8ULL, "dae000_arishok6067849.cl"},
	{0xD19968691DB8FC96ULL, "w_sundermount_73_0.phy"},
	{0xD19972CD8E887D41ULL, "6105864_f.fxe"},
	{0xD199C0B83846C841ULL, "llac_lt_0_lowtown_0_2.dds"},
	{0xD19A2B22DD9D48A8ULL, "vxw_encounter2_4h_0_0.vlm"},
	{0xD19A3481BD1AC067ULL, "llac_ht_0_hightown_a3_1_0.dds"},
	{0xD19A3E7975C2F48CULL, "fx_impact012.dds"},
	{0xD19A4FE74AF5DD16ULL, "vxht_mansion2b_3c9_0_0.vlm"},
	{0xD19A798B887A3809ULL, "mcr310_party_banter6087774.cl"},
	{0xD19B20CC6B2F8511ULL, "6118642_f.fxe"},
	{0xD19B71372B89A9E1ULL, "6050127_m.fxe"},
	{0xD19BC6A4494D9B4CULL, "vxht_0_hightown_a2_cai_0_0.vlm"},
	{0xD19C055728D82767ULL, "6096224_m.fxe"},
	{0xD19C6F96F0AD552AULL, "720150957.ogg"},
	{0xD19C913FD7696050ULL, "vxlt_warehouse_n_so_0_0.vlm"},
	{0xD19CE2DC06C24FA6ULL, "6156468_m.fxe"},
	{0xD19D040FB06B1917ULL, "6096168_f.fxe"},
	{0xD19D2388ED88BDF7ULL, "w_encounter3_101_0.msh"},
	{0xD19D3588ED88DC5AULL, "w_encounter3_101_0.mao"},
	{0xD19D3988ED88E369ULL, "w_encounter3_101_0.mmh"},
	{0xD19D6514498B0681ULL, "441248240.ogg"},
	{0xD19D9C537D4B1962ULL, "mag170_hugh6148618.cl"},
	{0xD19EC0E9E4627229ULL, "hlac_ht_0_hightown_a3_1_2x.dds"},
	{0xD19F3EF088570DCBULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robb_0t.dds"},
	{0xD19F6F928D473916ULL, "follower_banter6154027.cl"},
	{0xD1A01182D36F2AF6ULL, "ss_fereldan_m_1.bnk"},
	{0xD1A0422855222711ULL, "6174878_m.fxe"},
	{0xD1A08D54277ECD0EULL, "vxca_caverns2_7gz_0_0.vlm"},
	{0xD1A0DC38D02951B6ULL, "41512603.ogg"},
	{0xD1A12F0D2CC2C233ULL, "vxlt_0_lowtown_docks_d2y_0_0.vlm"},
	{0xD1A1A6C0A000A929ULL, "vxca_caverns2prc_9fu_0_0.vlm"},
	{0xD1A1BE36933E89BEULL, "6154721_m.fxe"},
	{0xD1A1E5A0A9DD77F7ULL, "vxw_encounter3_1o_0_0.vlm"},
	{0xD1A219C5A360174BULL, "hlac_lt_undercityserialkiller_-1_51y.dds"},
	{0xD1A2638185113505ULL, "art\\characters\\creatures\\mubari\\animation\\dg_tn_90_l_5p.ani"},
	{0xD1A273F7E4DCA7BCULL, "gen00fl_varric6076070.cl"},
	{0xD1A274687F76F0C8ULL, "6054078_m.fxe"},
	{0xD1A3039354A8512CULL, "zz_followers_banter6159240.cl"},
	{0xD1A32649675FE808ULL, "plt_lgt210pt_frame_follow.nss"},
	{0xD1A3333D1B75D813ULL, "vxlt_0_lowtown_a3_n_c_hoo_0_0.vlm"},
	{0xD1A351691DC12A37ULL, "w_sundermount_73_0.msh"},
	{0xD1A35506F5231749ULL, "llac_ht_mansion2b_818_61.dds"},
	{0xD1A363691DC1489AULL, "w_sundermount_73_0.mao"},
	{0xD1A367691DC14FA9ULL, "w_sundermount_73_0.mmh"},
	{0xD1A3D226E4572D68ULL, "fx_c_smoke.mao"},
	{0xD1A3D668311408D5ULL, "gen00fl_dog6115063.cl"},
	{0xD1A3F562B6DEC022ULL, "485938465.ogg"},
	{0xD1A40CFE4BCBC8AFULL, "886685787.ogg"},
	{0xD1A42079D968567AULL, "arial12.dds"},
	{0xD1A464E71907B91EULL, "6125350_m.fxe"},
	{0xD1A4713C1971C9FDULL, "6149081_m.fxe"},
	{0xD1A53614A15F5F08ULL, "134485151.ogg"},
	{0xD1A54DAF6B2DAF35ULL, "6021134_m.fxe"},
	{0xD1A5DC5E337A124EULL, "det_smo_th.mao"},
	{0xD1A60267D9F382E1ULL, "hla_ht_0_hightown_a3_65535_0z.dds"},
	{0xD1A6151207E0B4D4ULL, "352727400.ogg"},
	{0xD1A635B625882470ULL, "455600893.ogg"},
	{0xD1A639F04118F7F8ULL, "vxlt_undercityserialkiller_1mn_0_0.vlm"},
	{0xD1A64125907BD74CULL, "llac_lt_0_lowtown_docks_n_-65535_0.dds"},
	{0xD1A6FD27CC3FAD8CULL, "vxca_caverns2_8ln_0_0.vlm"},
	{0xD1A7CC007C09DF8BULL, "aip_dragon_base.gda"},
	{0xD1A7D415EDF22129ULL, "plt_cod_hst_nevarra.plo"},
	{0xD1A811D39815E4FFULL, "6041080_m.fxe"},
	{0xD1A8137A12D3D688ULL, "vxlt_0_lowtown_n_h8f_0_0.vlm"},
	{0xD1A85BF2F2650D36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_5.evt"},
	{0xD1A8E50CA171E4A5ULL, "6125977_m.fxe"},
	{0xD1A8F05BF9D795A8ULL, "vxw_blightlands_cr_0_0.vlm"},
	{0xD1A8F31C38500C78ULL, "ser120cr_market_guard.utc"},
	{0xD1A90AD9ECF41AD0ULL, "6027552_f.fxe"},
	{0xD1A9A04593D5ECF1ULL, "vxdr_primevaltomb_8r3_0_0.vlm"},
	{0xD1A9B3679BB1662AULL, "dae000_epic_merchant_26179393.cl"},
	{0xD1AA0D82023DFC97ULL, "vxht_0_hightown_a3_bty_0_0.vlm"},
	{0xD1AA2F496766161BULL, "plt_lgt210pt_frame_follow.plo"},
	{0xD1AA5746EC2B7455ULL, "vxw_sundermount_2tz_0_0.vlm"},
	{0xD1AB09644F39D66AULL, "vxlt_0_lowtown_n_f0r_0_0.vlm"},
	{0xD1AB1D31F34E1935ULL, "vxca_caverns1_mines_e7e_0_0.vlm"},
	{0xD1ACCFE0DCBC6420ULL, "vxht_brothel_n_5cy_0_0.vlm"},
	{0xD1ACDE2875B283EDULL, "gen00fl_aveline6114853.cl"},
	{0xD1AE4000CCDA8EE3ULL, "vxht_0_hightown_a2_7g4_0_0.vlm"},
	{0xD1AE5CA751CC1D37ULL, "6122800_m.fxe"},
	{0xD1AE943D53D24211ULL, "hlac_ca_caverns1_mines_24423_8z.dds"},
	{0xD1AEADEB3121D020ULL, "77258906.ogg"},
	{0xD1AED672F3E8FB73ULL, "gen00fl_fenris6099829.cl"},
	{0xD1AEEA3A857CBA07ULL, "vxlt_0_lowtown_a3_n_c_gvo_0_0.vlm"},
	{0xD1AEF72EC22D4CFEULL, "vxlt_0_lowtown_a2_n_gvy_0_0.vlm"},
	{0xD1AF3BB0F496A197ULL, "190987568.ogg"},
	{0xD1AF52F2F26B1CC1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_5.gad"},
	{0xD1AFC33946E88044ULL, "sst_ss_justice.gda"},
	{0xD1AFEA066AC34374ULL, "104623992.ogg"},
	{0xD1B013B9194D3459ULL, "6108474_f.fxe"},
	{0xD1B0D76D78671EDBULL, "lt_0_lowtown_docks_a3_194_0.mmh"},
	{0xD1B0E36D78673378ULL, "lt_0_lowtown_docks_a3_194_0.mao"},
	{0xD1B0F4848AD3B727ULL, "139766995.ogg"},
	{0xD1B0F56D786751C5ULL, "lt_0_lowtown_docks_a3_194_0.msh"},
	{0xD1B1895C85F83368ULL, "80604538.ogg"},
	{0xD1B1C46B479982A7ULL, "vxlt_shop_cj_0_0.vlm"},
	{0xD1B2024B63342E0FULL, "lgt200pt_tomes.ncs"},
	{0xD1B210D317FD6FA0ULL, "lgt200cr_gracious_gillian.utc"},
	{0xD1B2837678A29391ULL, "6149647_m.fxe"},
	{0xD1B33AA92A07024FULL, "hlac_ht_0_hightown_n_1_2z.dds"},
	{0xD1B3AF5FD8F05F96ULL, "6097225_f.fxe"},
	{0xD1B3CD17F0A03CE6ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.anb"},
	{0xD1B3FB1123779D3CULL, "art\\vfx\\cinematic\\vfx_mage_punch_c.anb"},
	{0xD1B5216BB30FEEBCULL, "gen00fl_varric6118774.cl"},
	{0xD1B52F316EC5C10CULL, "6164184_m.fxe"},
	{0xD1B57EB75CB55C44ULL, "6179053_f.fxe"},
	{0xD1B5E7BF12E4E65EULL, "ave310_party_comment6134852.cl"},
	{0xD1B64026EE841D18ULL, "ht_keep_n_3733.rml"},
	{0xD1B653AAD74F6B0BULL, "gen00fl_fenris6115365.cl"},
	{0xD1B653F2F2713D6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_5.ani"},
	{0xD1B68FB86D7A9EA5ULL, "character_character_p136.psh.fxo.dx11.assembly"},
	{0xD1B756E99367FD25ULL, "6122877_m.fxe"},
	{0xD1B76238EBCDAE28ULL, "vxw_blightlands_ds_0_0.vlm"},
	{0xD1B77424B1FF0BDAULL, "mrl000_marethari6129209.cl"},
	{0xD1B7B50B57098B92ULL, "6178468_m.fxe"},
	{0xD1B80BAEDCE87025ULL, "hlac_w_encounter2_5_0x.dds"},
	{0xD1B81C2CCF3B515EULL, "gen00fl_anders6103666.cl"},
	{0xD1B8446A9731AE0EULL, "w_woundedcoast_135_0.mmh"},
	{0xD1B8506A9731C2A5ULL, "w_woundedcoast_135_0.mao"},
	{0xD1B85E6A9731DAFCULL, "w_woundedcoast_135_0.msh"},
	{0xD1B8960432084046ULL, "art\\cinematics\\deathblows\\rockgolem\\db_rock_rockmage.ani"},
	{0xD1B8C923DE218CD5ULL, "vxw_sundermount_2yy_0_0.vlm"},
	{0xD1B8D4579D8BB4CEULL, "mrl310_merrill6120415.cl"},
	{0xD1B90F916A211875ULL, "753300520.ogg"},
	{0xD1B938020B0D406AULL, "313143898.ogg"},
	{0xD1B9F6174BE3AC04ULL, "gen00fl_isabela6099709.cl"},
	{0xD1BA4B36AB441403ULL, "6071692_m.fxe"},
	{0xD1BA9EBF44F8D4D4ULL, "gen_im_trs_gen_brokensword.uti"},
	{0xD1BAA2CABF5038DAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_diagonal.mmh"},
	{0xD1BAA6CABF503F11ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_diagonal.mao"},
	{0xD1BAB4CABF505720ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_diagonal.msh"},
	{0xD1BAFCD66DAFE5CCULL, "gen_im_npc_dwf_arm_cht_lgta.uti"},
	{0xD1BB556E2C4D27ADULL, "393497121.ogg"},
	{0xD1BB7CE632D0EC3CULL, "genip_trap_web.utp"},
	{0xD1BC48A8B1DB74D8ULL, "vxw_encounter2_camp_35_0_0.vlm"},
	{0xD1BC65E3E0AB3762ULL, "6147729_m.fxe"},
	{0xD1BCC614B9FC258DULL, "mcr370_varric6099735.cl"},
	{0xD1BCCD782BB47CBBULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\appear.evt"},
	{0xD1BCD3A0C803E274ULL, "443560240.ogg"},
	{0xD1BCFF9626F2C391ULL, "vxlt_0_lowtown_docks_a3_n_d2r_0_1.vlm"},
	{0xD1BD14AE85899DADULL, "211035795.ogg"},
	{0xD1BD3C81837BAAEBULL, "fen110_fenris6118051.cl"},
	{0xD1BD3DE4686092CCULL, "follower_barks6153759.cl"},
	{0xD1BDC065097AE920ULL, "6103323_m.fxe"},
	{0xD1BE22AE1C88E850ULL, "vxlt_darktown_4qu_0_0.vlm"},
	{0xD1BE5B303BDD3534ULL, "6147899_m.fxe"},
	{0xD1BEEC7B8BEE05C8ULL, "mag131tr_eau_demons1.utt"},
	{0xD1BEF359C683672EULL, "w_encounter2_12_0_ll.mmh"},
	{0xD1BFC93FAA550C38ULL, "167344395.ogg"},
	{0xD1C0179606A2912EULL, "gen00fl_anders6125965.cl"},
	{0xD1C03648398F9B27ULL, "vxlt_shop_fe_0_0.vlm"},
	{0xD1C03E6D9071DCADULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh2a.mmh"},
	{0xD1C0506D9071FB33ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh2a.msh"},
	{0xD1C065F44616C3EFULL, "bld_mac_oil.mao"},
	{0xD1C0BCD14AA4E2FAULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure.mmh"},
	{0xD1C0C0D14AA4E931ULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure.mao"},
	{0xD1C0CED14AA50140ULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure.msh"},
	{0xD1C0D017F0AACB91ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.mmh"},
	{0xD1C0D0671D3FD5D8ULL, "llac_w_sundermount_131071_0.dds"},
	{0xD1C16D47873FBDF7ULL, "vxht_mansion_n_614_0_0.vlm"},
	{0xD1C19D90B547403AULL, "300086313.ogg"},
	{0xD1C1C6D91B51DE0CULL, "vxht_0_hightown_a2_c3v_0_1.vlm"},
	{0xD1C289A4A19CDA48ULL, "505291585.ogg"},
	{0xD1C2C6782BB8DCFCULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\appear.gad"},
	{0xD1C2D0165FCFF466ULL, "vxca_caverns2prc_98h_0_0.vlm"},
	{0xD1C2E4A7EFC87B96ULL, "214412185.ogg"},
	{0xD1C2F658A37689AEULL, "hlac_lt_0_lowtown_docks_a3_n_-3_0x.dds"},
	{0xD1C359108F564E82ULL, "47079168.ogg"},
	{0xD1C35DCD5C5AA87FULL, "692623409.ogg"},
	{0xD1C39616B35D6407ULL, "550122936.ogg"},
	{0xD1C3A367754000A9ULL, "55867561.ogg"},
	{0xD1C40BFFC91C2B07ULL, "hlac_lt_0_lowtown_a2_-3_7y.dds"},
	{0xD1C41243CF7F43A6ULL, "984398973.ogg"},
	{0xD1C43A3593FF4336ULL, "hlac_ht_mansion2_818_60z.dds"},
	{0xD1C44F6FC6CD487EULL, "6133586_m.fxe"},
	{0xD1C459804D5BBE5AULL, "hlac_lt_0_lowtown_a3_n_c_-196608_2x.dds"},
	{0xD1C4655E7C640F88ULL, "dre000_carver6037591.cl"},
	{0xD1C4743A3FD6EA28ULL, "zz_dae_debug6029088.cl"},
	{0xD1C47934C0731099ULL, "290024110.ogg"},
	{0xD1C4974337F64384ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_hands_on_face.anb"},
	{0xD1C4ADF7C58394DEULL, "6099982_f.fxe"},
	{0xD1C4D97B693629A3ULL, "93437924.ogg"},
	{0xD1C4F91956F67CF6ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcrate.msh"},
	{0xD1C5031956F68DF4ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcrate.mmh"},
	{0xD1C5071956F69407ULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcrate.mao"},
	{0xD1C52CF2E36A5E4AULL, "152309804.ogg"},
	{0xD1C54538A6F8E1F9ULL, "qun211_braggart6075697.cl"},
	{0xD1C5FE4D03927515ULL, "715804152.ogg"},
	{0xD1C64BD4AD271873ULL, "6111165_m.fxe"},
	{0xD1C66DBDAE4C33B2ULL, "hlac_dr_0_deeproads_5841_0z.dds"},
	{0xD1C675ABBE42E0C2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45basic.phy"},
	{0xD1C6CBCFF5205A33ULL, "framebuffereffects_diffusiondof_p10.psh.fxo.dx11"},
	{0xD1C77C26D4B9503BULL, "vxht_keep_n_6hs_0_0.vlm"},
	{0xD1C80DB84317B3E3ULL, "463559928.ogg"},
	{0xD1C8178CCE88F1C8ULL, "6101444_m.fxe"},
	{0xD1C82D0B03154571ULL, "hlac_lt_undercityserialkiller_-1_45z.dds"},
	{0xD1C83A3682FAFCADULL, "46578548.ogg"},
	{0xD1C881450ED05E5AULL, "814651333.ogg"},
	{0xD1C8DA0C4455F5FDULL, "llac_ht_keep_n_4519_4.dds"},
	{0xD1C8E0FD92E4A43EULL, "6084411_m.fxe"},
	{0xD1C9C5782BBEF9FAULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\appear.ani"},
	{0xD1CA716F5389E3DDULL, "hlac_lt_0_lowtown_65537_4x.dds"},
	{0xD1CA7D7E2358F399ULL, "48229794.ogg"},
	{0xD1CBBBE75B700F74ULL, "6088303_m.fxe"},
	{0xD1CC0A3EBFF1AAFFULL, "dae000_tomwise6167314.cl"},
	{0xD1CC6206C35DC0EEULL, "vxlt_hangedman_n_1yh_0_0.vlm"},
	{0xD1CCDC6637A9CFFFULL, "748573474.ogg"},
	{0xD1CCF772FD396559ULL, "6106674_f.fxe"},
	{0xD1CD12E4415031FCULL, "vxca_caverns2prc_9pj_0_0.vlm"},
	{0xD1CD270B9741825BULL, "175741512.ogg"},
	{0xD1CD7F04321A7540ULL, "art\\cinematics\\deathblows\\rockgolem\\db_rock_rockmage.gad"},
	{0xD1CE6293CBA4C061ULL, "vxlt_0_lowtown_fe1_0_0.vlm"},
	{0xD1CE7FB5306C1CACULL, "6036795_m.fxe"},
	{0xD1CE9D9A46733C70ULL, "161558291.ogg"},
	{0xD1CEC3AA6E2FD79DULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_local_qunari_stairs.anb"},
	{0xD1CF3F4F2B2F3290ULL, "199563050.ogg"},
	{0xD1CF48F0C45CE1FAULL, "w_bonepitdestroyed_a3_43_0.rml"},
	{0xD1CF8B9B43A45964ULL, "vxca_caverns1_mines_dqc_0_0.vlm"},
	{0xD1CF8D7685E8BFDCULL, "plt_and000pt_main.plo"},
	{0xD1CFB02DD4B2A606ULL, "vxht_mansion2b_1qq_0_0.vlm"},
	{0xD1CFD5056ADA30C8ULL, "gen_im_cft_rec_wrn_04.uti"},
	{0xD1D054FFBA25FE37ULL, "vxlt_undercityencounter_18n_0_0.vlm"},
	{0xD1D06CB91E848303ULL, "vxlt_undercityserialkiller_zw_0_0.vlm"},
	{0xD1D0993AAAAE3464ULL, "da2terrain_terrain_p2.vsh.fxo.dx11.assembly"},
	{0xD1D0F6B857A39BE1ULL, "vxlt_gallowstemplar_n_3qp_0_0.vlm"},
	{0xD1D1376D9080446AULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hh2a.phy"},
	{0xD1D18B418FA53C76ULL, "6018911_m.fxe"},
	{0xD1D18F2420DA710EULL, "leave_01_l2.mao"},
	{0xD1D1C087236BA5FDULL, "162000981.ogg"},
	{0xD1D1E5E373217892ULL, "zz_mer_debug6073692.cl"},
	{0xD1D1E845AF97745BULL, "w_bonepit_119_0.mmh"},
	{0xD1D1F445AF9788F8ULL, "w_bonepit_119_0.mao"},
	{0xD1D20645AF97A745ULL, "w_bonepit_119_0.msh"},
	{0xD1D21F7AE5EE82CAULL, "vxht_0_hightown_a2_b5v_0_0.vlm"},
	{0xD1D22F872BF654B5ULL, "vxdr_dungeon_1lr_0_0.vlm"},
	{0xD1D2984FBF66169BULL, "follower_banter6170008.cl"},
	{0xD1D31645E49E4E04ULL, "hlac_w_bonepitdestroyed_a3_65536_0y.dds"},
	{0xD1D3445194E7F71AULL, "mer220_hubert6040834.cl"},
	{0xD1D3EF6F6D38FE91ULL, "6017870_m.fxe"},
	{0xD1D445F2CFB9DF28ULL, "vxw_blightlands_rq_0_0.vlm"},
	{0xD1D49181BD80A985ULL, "vxlt_undercityserialkiller_1mw_0_0.vlm"},
	{0xD1D61B98F80298C3ULL, "vxht_keep_6ew_0_0.vlm"},
	{0xD1D6EB4C9A0140D2ULL, "zz_isa_debug6133191.cl"},
	{0xD1D714B7E92FAC83ULL, "1005224894.ogg"},
	{0xD1D8035427AE023EULL, "6153896_m.fxe"},
	{0xD1D81617DB14055FULL, "457202098.ogg"},
	{0xD1D82577B29EADA0ULL, "vxht_brothel_n_5ld_0_0.vlm"},
	{0xD1D83A5D873A2F85ULL, "qun200_party_comment6090449.cl"},
	{0xD1D89F78B340B19DULL, "185055506.ogg"},
	{0xD1D9224719F10237ULL, "914147304.ogg"},
	{0xD1D98BFD82D41409ULL, "1896832.ogg"},
	{0xD1D9B482206E1691ULL, "6051560_m.fxe"},
	{0xD1DA7F6E18F847C8ULL, "vxlt_gallowstemplar_n_3qu_0_0.vlm"},
	{0xD1DAA4C6C9CC0038ULL, "gen00fl_aveline6139154.cl"},
	{0xD1DAAB65EDCE15BEULL, "vxht_mansion2_6u_0_0.vlm"},
	{0xD1DAD370CE258510ULL, "vxlt_0_lowtown_h9j_0_0.vlm"},
	{0xD1DAD3E3B553D96EULL, "vxlt_hangedman_n_1vi_0_0.vlm"},
	{0xD1DB1936762DE393ULL, "mcr310_party_banter6087813.cl"},
	{0xD1DB20C0B7A27080ULL, "111527708.ogg"},
	{0xD1DB30DB9CD1687AULL, "vxht_mansion_n_5v3_0_0.vlm"},
	{0xD1DB7103CB5E2B29ULL, "hlac_lt_hangedman_-1_8x.dds"},
	{0xD1DBA370DD084E19ULL, "llac_lt_0_lowtown_131072_4.dds"},
	{0xD1DBC75507F7EE3EULL, "vxca_caverns1_mines_jga_0_0.vlm"},
	{0xD1DC3AC3B13228DBULL, "668162862.ogg"},
	{0xD1DC8E5FAC5E4A68ULL, "ave200_aveline_bar6115013.cl"},
	{0xD1DD61D0FB8C0D26ULL, "6144254_m.fxe"},
	{0xD1DDA3ED7C12E258ULL, "lgt210ip_ingredient.ncs"},
	{0xD1DDEAA6C0FDCDD2ULL, "vxdr_primevaltomb_3lx_0_1.vlm"},
	{0xD1DE220AC6A8BE86ULL, "vxht_mansion2b_1tr_0_0.vlm"},
	{0xD1DED10B29D15E68ULL, "6149432_f.fxe"},
	{0xD1DED6A0452D1E57ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_goon1.gad"},
	{0xD1DF73A93E975D53ULL, "ser211ip_altar.ncs"},
	{0xD1DF79FFFB9BBED7ULL, "zz_dae_debug6049960.cl"},
	{0xD1DF7F8F9A542498ULL, "6079084_m.fxe"},
	{0xD1E0016408F159F8ULL, "vxlt_undercityserialkiller_1qj_0_0.vlm"},
	{0xD1E0358BEF3227D2ULL, "hlac_lt_gallowsprison_n_-1_7x.dds"},
	{0xD1E0BB45E7BC430CULL, "vxht_0_hightown_bwf_0_0.vlm"},
	{0xD1E0C1CABF713CE5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_diagonal.phy"},
	{0xD1E0C49B94180F8CULL, "vxca_caverns2_8xb_0_0.vlm"},
	{0xD1E15CA859B6741CULL, "lt_0_lowtown_docks_a3_n_31_0_lc.mmh"},
	{0xD1E19718E5C074F9ULL, "162604033.ogg"},
	{0xD1E20D1AF96C4185ULL, "llac_w_encounter2_camp_131075_0.dds"},
	{0xD1E22126EEA923D0ULL, "ht_keep_n_3733.anb"},
	{0xD1E23770497044A7ULL, "vxlt_gallowscourtyard_76n_0_0.vlm"},
	{0xD1E2584ADC529270ULL, "6053956_m.fxe"},
	{0xD1E29BE9159D9BFAULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_107a_n.dds"},
	{0xD1E2B7CFC1AFF7A8ULL, "vxw_blightlands_wn_0_0.vlm"},
	{0xD1E2EC990EE339E8ULL, "641710340.ogg"},
	{0xD1E38BADC05088CDULL, "vrc221_varric6065967.cl"},
	{0xD1E3B1190839605AULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.anb"},
	{0xD1E44C33574563ADULL, "nat320_delilah6065943.cl"},
	{0xD1E4A50AE4364543ULL, "vxca_caverns1_j23_0_1.vlm"},
	{0xD1E4F12E3052A8C5ULL, "hlac_lt_0_lowtown_docks_n_-65537_4y.dds"},
	{0xD1E50174748AF206ULL, "art\\levels\\races\\darkspawn\\props\\_textures\\prp_ogreskull_sa.dds"},
	{0xD1E53A9AC154D411ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgtp_0s.dds"},
	{0xD1E5443CA7AA9CC1ULL, "145252479.ogg"},
	{0xD1E549319524B177ULL, "gen00fl_sebastian6178381.cl"},
	{0xD1E5DBD14AC43305ULL, "art\\levels\\races\\proxy\\props\\pressure\\prp_pressure.phy"},
	{0xD1E6A7FC5490549BULL, "79961537.ogg"},
	{0xD1E6B1338F56BF21ULL, "vxca_caverns1_egt_0_0.vlm"},
	{0xD1E73688A22D4C80ULL, "terrain_toolset_rl.psh.fxo"},
	{0xD1E7574D69F779FEULL, "283922439.ogg"},
	{0xD1E7A7A1B2EE553FULL, "gen00fl_merrill6090268.cl"},
	{0xD1E7B9F84F60F3EBULL, "dae000_surprises6162511.cl"},
	{0xD1E7BDFB379CCFEDULL, "zz_dae_debug6045827.cl"},
	{0xD1E8123E02B06A60ULL, "611770367.ogg"},
	{0xD1E8F965B84BFEBBULL, "932841572.ogg"},
	{0xD1E9054031F76930ULL, "6120131_m.fxe"},
	{0xD1E913AD5634FADBULL, "weapon_rendertodepth_p0.psh.fxo.dx11.assembly"},
	{0xD1E91FABAB8F9F4CULL, "plt_lgt000pt_ultra_4_03.plo"},
	{0xD1E9F69E253C62FCULL, "vxca_caverns2prc_9zh_0_0.vlm"},
	{0xD1EA31FA6D97F856ULL, "296930632.ogg"},
	{0xD1EB50ABBE61BE76ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45basic.mao"},
	{0xD1EB5CABBE61D215ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45basic.mmh"},
	{0xD1EB5EABBE61D58BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww12_45basic.msh"},
	{0xD1EC0293676E3F1FULL, "dre180ip_deep_roads_chest_4.utp"},
	{0xD1EC44E0ABE649AAULL, "hlac_ht_chantry_n_2534_26x.dds"},
	{0xD1ECF1A1312EA37DULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_4x.dds"},
	{0xD1EDFAC070BDABC9ULL, "vxht_playershouse_n_1lc_0_0.vlm"},
	{0xD1EE056A755B2D1CULL, "vxht_0_hightown_c5q_0_0.vlm"},
	{0xD1EE170E316711B5ULL, "vxlt_0_lowtown_n_epc_0_1.vlm"},
	{0xD1EE4D9F18AB75FEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_activ_c.mmh"},
	{0xD1EEC84348AC52B2ULL, "gen00fl_aveline6106936.cl"},
	{0xD1EEDD21EE5744E2ULL, "vxht_keep_n_6j4_0_0.vlm"},
	{0xD1EF2E821F65B700ULL, "133089185.ogg"},
	{0xD1EFC2DA007ED9B1ULL, "vxca_caverns2_3i2_0_1.vlm"},
	{0xD1F004ABAB958F83ULL, "plt_lgt000pt_ultra_4_03.nss"},
	{0xD1F007F825B9EB08ULL, "480448255.ogg"},
	{0xD1F0202E6D086A6AULL, "886059774.ogg"},
	{0xD1F027C717AB43E9ULL, "240551316.ogg"},
	{0xD1F03233141B4BCBULL, "549917963.ogg"},
	{0xD1F08BEB77A9E2E0ULL, "284693974.ogg"},
	{0xD1F0AC0D86870901ULL, "hlac_lt_0_lowtown_a2_n_c_-65534_0y.dds"},
	{0xD1F0BD2137332A93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkbackright.ani"},
	{0xD1F0F27A54184900ULL, "385635242.ogg"},
	{0xD1F1C3010E4C7621ULL, "hlac_ca_cavern1_haunted_24424_21z.dds"},
	{0xD1F2FF52DBEEC9C3ULL, "vxht_keep_6ky_0_0.vlm"},
	{0xD1F39BD4972E6F5CULL, "qun221_petrice_intercept.stg"},
	{0xD1F3A2CDE10D937BULL, "gen00fl_aveline6112710.cl"},
	{0xD1F3CECC26F133BEULL, "104058406.ogg"},
	{0xD1F3DFA8B0D03AC5ULL, "6208668_f.fxe"},
	{0xD1F403A0453FC6DDULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_goon1.ani"},
	{0xD1F46D33A9C78A08ULL, "mrl000_marethari6129438.cl"},
	{0xD1F4A38E30C32FE9ULL, "76384768.ogg"},
	{0xD1F52310814CD7A1ULL, "vxca_caverns1_ebk_0_0.vlm"},
	{0xD1F52E56C142FC00ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_roba_0s.dds"},
	{0xD1F5FC085EA29B35ULL, "963923328.ogg"},
	{0xD1F6EB9DEAEB93D2ULL, "vxw_woundedcoast_8w_0_0.vlm"},
	{0xD1F6F1D5C9DBA777ULL, "vxlt_0_lowtown_a2_eji_0_0.vlm"},
	{0xD1F7C4643EAD4A4EULL, "6117276_m.fxe"},
	{0xD1F7DB55FBCCEABFULL, "29802578.ogg"},
	{0xD1F813D92FDC6E32ULL, "vxlt_0_lowtown_a2_fha_0_0.vlm"},
	{0xD1F8660F7FB398ADULL, "253604100.ogg"},
	{0xD1F8CDE76AD0EAD4ULL, "vxlt_0_lowtown_a2_n_c_ef7_0_0.vlm"},
	{0xD1F92FBE700BC639ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.impa.ani"},
	{0xD1F952AA60098A77ULL, "vxht_keep_n_6hg_0_0.vlm"},
	{0xD1F961DEA0F2BF82ULL, "vxw_sundermount_34h_0_0.vlm"},
	{0xD1F967C8F54F84A5ULL, "6062126_m.fxe"},
	{0xD1F9E027890D8A03ULL, "ev_wis_sm_dark.mao"},
	{0xD1FB0A70ED542749ULL, "6125380_m.fxe"},
	{0xD1FB38B688BACF48ULL, "278668748.ogg"},
	{0xD1FB5BF7B122A609ULL, "116860982.ogg"},
	{0xD1FB5EACF57BDE9CULL, "llac_lt_0_lowtown_a2_n_1_7.dds"},
	{0xD1FBD621A40CD203ULL, "dae000ip_crowd_mf2_talk9.utp"},
	{0xD1FBDCC51B8D0DD3ULL, "vxw_woundedcoast_5u_0_0.vlm"},
	{0xD1FBFF95F569F2B0ULL, "940462474.ogg"},
	{0xD1FC21784C80D319ULL, "vxht_mansion_n_59a_0_0.vlm"},
	{0xD1FC2C9F18B77ABDULL, "art\\vfx\\combat\\vfx_w_gn1_battle_activ_c.anb"},
	{0xD1FCA0138740921DULL, "hlac_w_blightlands_1_0y.dds"},
	{0xD1FCBEA3719E071FULL, "943372342.ogg"},
	{0xD1FCCC6A0F0208D5ULL, "hf_amb_loyal_mage.mor"},
	{0xD1FD0EDD1C5559EAULL, "follower_banter6163753.cl"},
	{0xD1FDC3F13AE6E5FFULL, "mer320_hubert6089682.cl"},
	{0xD1FDF22334664944ULL, "hlac_w_blightlands_327678_0z.dds"},
	{0xD1FE0D0A2EE96C7DULL, "vxlt_gallowscourtyard_n_4uy_0_0.vlm"},
	{0xD1FEB1BB635AB2ADULL, "lgt110_gxa_coa_spy6117522.cl"},
	{0xD1FF1059C0B6AC1EULL, "6108629_m.fxe"},
	{0xD1FFE752A4DE0664ULL, "follower_banter6143393.cl"},
	{0xD1FFE7BFAED8B157ULL, "6139808_m.fxe"},
	{0xD2000ABE7011A67BULL, "art\\vfx\\environment\\vfx_sunder_fogwall.impa.gad"},
	{0xD2005ED1DD4ABFEBULL, "733848301.ogg"},
	{0xD200CF3BF3F457BAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sidel.phy"},
	{0xD200F6F44B05B4FBULL, "llac_lt_gallowstemplar_n_7198_7.dds"},
	{0xD20155F4FDE08A23ULL, "gen00fl_aveline6139533.cl"},
	{0xD20156193CC57DD4ULL, "6092340_m.fxe"},
	{0xD201CC64D993E256ULL, "vrc221tr_second_corpses.ncs"},
	{0xD201D9E80C0AB7D6ULL, "vxht_0_hightown_n_ccj_0_0.vlm"},
	{0xD2021590D9FEA507ULL, "m_mag131ar_bloodmages.wbd"},
	{0xD2021E8F9983D6F5ULL, "gen00fl_aveline6116745.cl"},
	{0xD202341B0461E351ULL, "llac_lt_0_lowtown_a2_-65535_5.dds"},
	{0xD202BE761651D2F7ULL, "vxlt_warehouse_9s_0_0.vlm"},
	{0xD20320752E32EA97ULL, "vxht_mansion_n_70b_0_0.vlm"},
	{0xD2036429BF45EEDAULL, "6073775_f.fxe"},
	{0xD20378EF7627CFC8ULL, "vxlt_0_lowtown_n_g92_0_0.vlm"},
	{0xD203865674BF6F9FULL, "mcr371_hostile_merrill6082142.cl"},
	{0xD203A159CDDF31DBULL, "vxlt_0_lowtown_a3_n_c_el9_0_2.vlm"},
	{0xD203AEC56BC6AFE9ULL, "gencr_slaver_2w_f.utc"},
	{0xD20463B3556EEC7FULL, "art\\characters\\weapons\\staff\\w_stf_007a.msh"},
	{0xD20469B3556EF621ULL, "art\\characters\\weapons\\staff\\w_stf_007a.mmh"},
	{0xD20475B3556F0AC2ULL, "art\\characters\\weapons\\staff\\w_stf_007a.mao"},
	{0xD204843F8B5708A6ULL, "6037589_m.fxe"},
	{0xD204D0EBC64916D0ULL, "lt_0_lowtown_docks_n_24_0.rml"},
	{0xD2055D7ADCE1AC52ULL, "vxw_woundedcoast_5p_0_0.vlm"},
	{0xD20563B2BBD1BFF7ULL, "vxlt_0_lowtown_a2_emh_0_0.vlm"},
	{0xD20594213745413DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\walkbackright.gad"},
	{0xD205AD7775A102E1ULL, "6095013_f.fxe"},
	{0xD205CD19E9F701B7ULL, "vxlt_0_lowtown_docks_a3_d2r_0_0.vlm"},
	{0xD205E84300B639CBULL, "6118162_m.fxe"},
	{0xD205FA5F399C9A4CULL, "332131850.ogg"},
	{0xD20627BE70164360ULL, "art\\vfx\\environment\\vfx_sunder_fogwall.impa.evt"},
	{0xD20710A7811B551DULL, "follower_banter6161268.cl"},
	{0xD2078F0909B27F59ULL, "dae000ip_crowd_lm4_talk5.utp"},
	{0xD207A0F6F95CED66ULL, "hlac_ca_caverns1_mines_24421_9x.dds"},
	{0xD207BD715A693C29ULL, "vxlt_0_lowtown_a3_n_c_eib_0_0.vlm"},
	{0xD207C48751FFA2F7ULL, "vxht_keep_n_6cf_0_0.vlm"},
	{0xD207CC21EE3962B8ULL, "and100_anders6095990.cl"},
	{0xD207D3BB92E8D802ULL, "vxw_sundermount_33i_0_0.vlm"},
	{0xD207D9A5E7459D25ULL, "6069026_m.fxe"},
	{0xD208074508B2B808ULL, "6089417_f.fxe"},
	{0xD2093874E1FD0D59ULL, "187365120.ogg"},
	{0xD2094E6305D0B950ULL, "art\\characters\\creatures\\spider\\animation\\blendtree\\spider_movement.anb"},
	{0xD2095203ED7ECF37ULL, "pro000_soldier1.bnk"},
	{0xD20A4EA20D832653ULL, "vxw_woundedcoast_2t_0_0.vlm"},
	{0xD20AC7C920AD34F9ULL, "ave200_donnic6061417.cl"},
	{0xD20AE8E13AAE2FB9ULL, "lgt110_gxa_coa_spy6117445.cl"},
	{0xD20B0796130C0C1BULL, "6044645_m.fxe"},
	{0xD20B91436A2D0688ULL, "vxht_0_hightown_9p0_0_0.vlm"},
	{0xD20BB419085AB70DULL, "art\\vfx\\combat\\vfx_c_buff_loop_c.mmh"},
	{0xD20C1C0FCEEBB7BDULL, "w_encounter2_camp_48_0.phy"},
	{0xD20C52C045399387ULL, "dae000ip_ambient_sitcudlf_s.utp"},
	{0xD20CA17A7B637498ULL, "hlac_lt_0_lowtown_docks_a3_-131075_4y.dds"},
	{0xD20CBF3842E18BC1ULL, "485736794.ogg"},
	{0xD20D0ED201535CF1ULL, "vxw_woundedcoast_l4_0_0.vlm"},
	{0xD20D11771BE0B40DULL, "7301071.ogg"},
	{0xD20D3403ED82854FULL, "pro000_soldier1.cnv"},
	{0xD20D84F324D0EE76ULL, "vxw_encounter2_d6_0_0.vlm"},
	{0xD20DFEDBA8359718ULL, "hlac_lt_undercityencounter_-1_3y.dds"},
	{0xD20E076E0754B74FULL, "158838240.ogg"},
	{0xD20E1CE3B9B10E38ULL, "dae000_arishok6081672.cl"},
	{0xD20E8F64DCDB60DBULL, "999730442.ogg"},
	{0xD20E952A1F3B721AULL, "art\\levels\\ztools\\default\\generic\\3ww_basic.mmh"},
	{0xD20E992A1F3B7951ULL, "art\\levels\\ztools\\default\\generic\\3ww_basic.mao"},
	{0xD20EA72A1F3B9160ULL, "art\\levels\\ztools\\default\\generic\\3ww_basic.msh"},
	{0xD20EDC6035B3121EULL, "563937357.ogg"},
	{0xD20EF22EE443E4F8ULL, "w_bonepitmining_a2_43_0.rml"},
	{0xD20F36B4CCA2364AULL, "vxdr_primevaltomb_3iz_0_0.vlm"},
	{0xD2101144E1E89A26ULL, "6104895_f.fxe"},
	{0xD210178D48B230B0ULL, "isa111_isabela6075617.cl"},
	{0xD210AC85F552A1DBULL, "gen_im_trs_dal_hvy_boo_r1.uti"},
	{0xD210B1B771FC3ED8ULL, "vxdr_primevalentrance_27c_0_0.vlm"},
	{0xD211DA15F3F0861CULL, "933654866.ogg"},
	{0xD21206CA653908A1ULL, "vxca_caverns1_ehi_0_0.vlm"},
	{0xD2131D0FCEF1D849ULL, "w_encounter2_camp_48_0.rml"},
	{0xD213BEF46028158FULL, "vis121_saemus6118468.cl"},
	{0xD213CF57CED7C4D2ULL, "vxw_woundedcoast_2q_0_0.vlm"},
	{0xD214428A2148FA20ULL, "water0.vsh.fxo"},
	{0xD214FC373A746B2EULL, "mag101_alain6032204.cl"},
	{0xD215208F430F4171ULL, "vxht_0_hightown_a2_c5d_0_0.vlm"},
	{0xD21525147011F7A2ULL, "hlac_lt_0_lowtown_docks_a3_-131076_0z.dds"},
	{0xD2155C59D141F3DFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_galowint.msh"},
	{0xD2156259D141FD81ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_galowint.mmh"},
	{0xD2156E59D1421222ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_galowint.mao"},
	{0xD2159566B13ED149ULL, "vxdr_primevaltomb_2sf_0_0.vlm"},
	{0xD215B1A14EBD1BD4ULL, "vxlt_0_lowtown_a2_n_c_ep5_0_0.vlm"},
	{0xD21617F8A0F71DF5ULL, "vxlt_darktown_48o_0_0.vlm"},
	{0xD2162EBFE7B95ADEULL, "vxlt_0_lowtown_a2_n_eop_0_0.vlm"},
	{0xD2167C93802C3DD2ULL, "377840685.ogg"},
	{0xD216D312994657CEULL, "6165645_f.fxe"},
	{0xD216E29D9474C9CDULL, "llac_w_bonepitmining_a2_458752_0.dds"},
	{0xD21728F0C49A60B1ULL, "w_bonepitdestroyed_a3_43_0.mmh"},
	{0xD21734F0C49A7552ULL, "w_bonepitdestroyed_a3_43_0.mao"},
	{0xD21742F0C49A8D6FULL, "w_bonepitdestroyed_a3_43_0.msh"},
	{0xD21747821B9A8456ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chantrydoor.msh"},
	{0xD21751821B9A9554ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chantrydoor.mmh"},
	{0xD21755821B9A9C67ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_chantrydoor.mao"},
	{0xD2179F5EA734AE84ULL, "gen00fl_aveline6121977.cl"},
	{0xD217B5F5BC2B4657ULL, "and230tr_dissent_anders.utt"},
	{0xD217C2407409F3CAULL, "6167492_f.fxe"},
	{0xD217E52492526F3EULL, "hlac_ht_mansion2_818_37z.dds"},
	{0xD218FD7424EA6CADULL, "hlac_ht_chantry_2534_49y.dds"},
	{0xD2190953BF054CF3ULL, "llac_lt_0_lowtown_a3_n_c_-196608_1.dds"},
	{0xD2194273CE3FBAF6ULL, "zz_mag_debug6054599.cl"},
	{0xD219BBC9D759E1C0ULL, "vxca_caverns1_mines_j1w_0_2.vlm"},
	{0xD219FC6576CC7069ULL, "fx_aura_circle.mao"},
	{0xD21A0945AFD56190ULL, "w_bonepit_119_0.phy"},
	{0xD21A7DB1B93BAD80ULL, "dae000_arishok6067852.cl"},
	{0xD21ADB7B8C3C47F5ULL, "mag131tr_eau_demons1.ncs"},
	{0xD21B159ECF8B322CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_ctn180r.gad"},
	{0xD21B208F9BDF1C74ULL, "rdr110dg_guard_warehouse6016679.cl"},
	{0xD21B38DACC7D7323ULL, "905837081.ogg"},
	{0xD21B5C634978E669ULL, "vxlt_0_lowtown_ex9_0_0.vlm"},
	{0xD21B6CABEF5E6B05ULL, "hlac_lt_0_lowtown_a2_-196607_0x.dds"},
	{0xD21B6CB0BB493876ULL, "vxht_0_hightown_a2_auq_0_0.vlm"},
	{0xD21BF413A4A2DD1EULL, "6106429_m.fxe"},
	{0xD21BF6D016C706F6ULL, "vxw_encounter2_o7_0_0.vlm"},
	{0xD21C16D93E944C63ULL, "nat380_nathaniel_banter6067291.cl"},
	{0xD21C356B795255E0ULL, "259846864.ogg"},
	{0xD21C377F71AA06DBULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5_tool.phy"},
	{0xD21C7A0338E7D721ULL, "17665501.ogg"},
	{0xD21CF0EB6ED2E9B2ULL, "6016687_m.fxe"},
	{0xD21D110FCEFA1829ULL, "w_encounter2_camp_48_0.mao"},
	{0xD21D1D0FCEFA2CC2ULL, "w_encounter2_camp_48_0.mmh"},
	{0xD21D1F0FCEFA3078ULL, "w_encounter2_camp_48_0.msh"},
	{0xD21D41E8FCB97B6DULL, "dr_primevaltomb_665.rml"},
	{0xD21D6B686DAC2BCCULL, "qun221_petrice6077455.cl"},
	{0xD21D72CE68F9C589ULL, "vxht_0_hightown_b9q_0_0.vlm"},
	{0xD21E334BBF7B1A8BULL, "gen00fl_merrill6150920.cl"},
	{0xD21EE1DE1C622C6CULL, "717539538.ogg"},
	{0xD21EFCE86DF93DC4ULL, "6140227_m.fxe"},
	{0xD21FA496945700CEULL, "gen00fl_isabela6088064.cl"},
	{0xD21FB6F0972F03D1ULL, "361143894.ogg"},
	{0xD2200219574337EFULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzcrate.phy"},
	{0xD2210A7CEA321776ULL, "vfx_vfx_p10.vsh.fxo.dx11"},
	{0xD22136A022BBCBD4ULL, "vxlt_0_lowtown_a2_n_c_f4r_0_0.vlm"},
	{0xD22169CE5B92F0C4ULL, "vxlt_hangedman_16t_0_0.vlm"},
	{0xD22182C8ECD26DC5ULL, "vxw_encounter3_v_0_0.vlm"},
	{0xD221983EED82F3F0ULL, "hlac_lt_0_lowtown_a2_n_c_65533_0x.dds"},
	{0xD222149ECF914F2AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_ctn180r.ani"},
	{0xD2224813B6134479ULL, "dr_slavers_mmi.gff"},
	{0xD222B10A5B3C1116ULL, "llac_lt_undercityserialkiller_-1_22.dds"},
	{0xD2232A3FB5195F06ULL, "6042233_m.fxe"},
	{0xD2233616C298FC24ULL, "addmoney.ncs"},
	{0xD2234717F1C5EC47ULL, "vxlt_0_lowtown_n_h27_0_0.vlm"},
	{0xD223DB107EC45206ULL, "one350_variel6076801.cl"},
	{0xD2249FB4D25EBBADULL, "267372114.ogg"},
	{0xD224B946BEB81769ULL, "646000056.ogg"},
	{0xD224E1FFA62A7D1CULL, "zz_isaac_test2.cut"},
	{0xD2254ECA5DCE4054ULL, "6180912_f.fxe"},
	{0xD22584925CBD8DC8ULL, "zz_dae_debug6047637.cl"},
	{0xD225C043C88A2F3CULL, "rog_brand.dds"},
	{0xD225DE102FF14B04ULL, "322023329.ogg"},
	{0xD225F77A86A9AC7CULL, "6020613_m.fxe"},
	{0xD22640E55D62C060ULL, "llac_dr_primevaltomb_657_2.dds"},
	{0xD22670C597F298C9ULL, "hlac_ht_chantry_2534_34y.dds"},
	{0xD226D764D9B31167ULL, "vrc221tr_second_corpses.utt"},
	{0xD2276E984210D8FBULL, "vxlt_0_lowtown_a2_gn3_0_0.vlm"},
	{0xD227C23438960DC9ULL, "vxht_playershouse_n_1pg_0_0.vlm"},
	{0xD227C2D4C8DD810CULL, "mcr371_post_harvester6077601.cl"},
	{0xD227F6146CA9AA3CULL, "hlac_lt_0_lowtown_a2_n_c_-65537_4y.dds"},
	{0xD227FB112339C2E4ULL, "vxht_0_hightown_a3_bzh_0_0.vlm"},
	{0xD2285923CA83E1ECULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.cess.ani"},
	{0xD2285FD77980713EULL, "84534219.ogg"},
	{0xD2286B5AE6D1B110ULL, "ser121_jethann6106929.cl"},
	{0xD2287E21A10D3A04ULL, "565897288.ogg"},
	{0xD228936DA4D2134EULL, "dae000_bartrand6033993.cl"},
	{0xD2289E8AF7CBCF39ULL, "vxht_0_hightown_a2_cbi_0_0.vlm"},
	{0xD228A469E2D2C709ULL, "x_partypicker_3567.rml"},
	{0xD229545367065B24ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingd_n.dds"},
	{0xD2295E5C9D74B001ULL, "302577067.ogg"},
	{0xD2296FB64A628AABULL, "vxlt_0_lowtown_a2_el9_0_2.vlm"},
	{0xD2297C9A366FE4CAULL, "vxht_brothel_n_4bu_0_0.vlm"},
	{0xD229CE403B6EFEE9ULL, "vxlt_0_lowtown_es8_0_0.vlm"},
	{0xD229CF7135984B14ULL, "vxht_bartrand_n_5vb_0_0.vlm"},
	{0xD229D1C3BEBD979AULL, "llac_lt_0_lowtown_docks_a3_-3_7.dds"},
	{0xD229DE8DAD3F50F6ULL, "vxht_0_hightown_a2_apr_0_0.vlm"},
	{0xD229EBCD676F72DFULL, "vxht_brothel_n_5eh_0_0.vlm"},
	{0xD22A68AD08BD1F76ULL, "vxw_encounter2_j8_0_0.vlm"},
	{0xD22AEEB732438ED2ULL, "llac_ca_cavern1_haunted_24424_24.dds"},
	{0xD22B0D583E4551FEULL, "1070424354.ogg"},
	{0xD22B1FEFC5F486ACULL, "hlac_ht_keep_4519_10y.dds"},
	{0xD22BB6D63A95F3F3ULL, "hlac_ht_mansion_n_3371_2y.dds"},
	{0xD22C034D96701F3CULL, "588861480.ogg"},
	{0xD22CADE7C4C673DFULL, "6205213_m.fxe"},
	{0xD22CD3C50BCF1A47ULL, "mag121_idunna6092357.cl"},
	{0xD22D05810173D2FBULL, "mag100_cullen6206252.cl"},
	{0xD22D21E8F2721DD4ULL, "hlac_dr_0_deeproads_5843_8z.dds"},
	{0xD22D50119D53AAEFULL, "20016387.ogg"},
	{0xD22D8E0FD0F0F19DULL, "vxca_caverns1_mines_in6_0_0.vlm"},
	{0xD22DDFFF4C3196AFULL, "vxw_sundermount_316_0_0.vlm"},
	{0xD22DEF1F9580F5DEULL, "hlac_lt_0_lowtown_65534_4z.dds"},
	{0xD22DF02EE45E91D5ULL, "w_bonepitmining_a2_43_0.msh"},
	{0xD22DF22EE45E954BULL, "w_bonepitmining_a2_43_0.mmh"},
	{0xD22DFE2EE45EA9E8ULL, "w_bonepitmining_a2_43_0.mao"},
	{0xD22E029775A6E8CCULL, "hlac_ht_mansion2b_818_44x.dds"},
	{0xD22E604868ADEA11ULL, "vxlt_0_lowtown_a2_n_c_hyv_0_0.vlm"},
	{0xD22E6623CA88632EULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.cess.gad"},
	{0xD22E80830EBC2B80ULL, "6016663_m.fxe"},
	{0xD22E937880A1260FULL, "941185638.ogg"},
	{0xD22EB2202D513897ULL, "gen_im_wep_war_two_act3_01.uti"},
	{0xD22EB4766F1677BAULL, "vxht_0_hightown_n_c2a_0_0.vlm"},
	{0xD22F3A2D42FDD3FDULL, "zz_dog_debug6127864.cl"},
	{0xD22F46EE6397681DULL, "hlac_ht_keep_4519_6y.dds"},
	{0xD22F91B3A9A62913ULL, "vxlt_0_lowtown_docks_bdx_0_1.vlm"},
	{0xD22FE978A5F14A9AULL, "zz_vault_debug6203309.cl"},
	{0xD22FF3829126B1A2ULL, "848708274.ogg"},
	{0xD2308B96628BD688ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraits_sa.dds"},
	{0xD230D1EBC66E5338ULL, "lt_0_lowtown_docks_n_24_0.anb"},
	{0xD231E75A7B8D6495ULL, "6160853_m.fxe"},
	{0xD2325137757350ACULL, "6136327_m.fxe"},
	{0xD2328C4BD1639FDFULL, "hlac_ht_keep_n_3732_14y.dds"},
	{0xD233418DFA1DF33BULL, "116596670.ogg"},
	{0xD233454E81EF58A6ULL, "mrl250cr_spider_cv.utc"},
	{0xD2334B5D565279F1ULL, "449199135.ogg"},
	{0xD2337A952BC73721ULL, "llac_lt_0_lowtown_a2_n_-3_3.dds"},
	{0xD233B42A1F5AC325ULL, "art\\levels\\ztools\\default\\generic\\3ww_basic.phy"},
	{0xD2340A84667B8E5EULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.impa.evt"},
	{0xD2341E51BBD572F1ULL, "vxca_caverns2_8rp_0_1.vlm"},
	{0xD234596A367BCBF5ULL, "6087813_m.fxe"},
	{0xD234C9A99F075359ULL, "105219276.ogg"},
	{0xD23515CBAF96C277ULL, "zz_mer_debug6181060.cl"},
	{0xD2353E60EB5ED887ULL, "6121702_m.fxe"},
	{0xD235533F0F6DAEE9ULL, "vxlt_0_lowtown_h1w_0_0.vlm"},
	{0xD2356123CA8E790DULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.cess.evt"},
	{0xD2359B003342891AULL, "dae000_worthy6166591.cl"},
	{0xD235A1C90495CBA6ULL, "6149885_f.fxe"},
	{0xD235DE59B10299D8ULL, "129877512.ogg"},
	{0xD23634112A8C2649ULL, "vxht_playershouse_n_1wf_0_0.vlm"},
	{0xD237500473E81B00ULL, "t3_grl_g00.tnt"},
	{0xD2378938045AD8F2ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_4y.dds"},
	{0xD237EE772865FD4AULL, "vxht_brothel_n_4gv_0_0.vlm"},
	{0xD238147BA6CF7A8DULL, "918875461.ogg"},
	{0xD238414E278E6394ULL, "vxht_bartrand_n_5yc_0_0.vlm"},
	{0xD2384F99B66C3470ULL, "combat_st_r1_combo.gda"},
	{0xD2385DAA59658B5FULL, "vxht_brothel_n_5hk_0_0.vlm"},
	{0xD238A15025D55140ULL, "11073161.ogg"},
	{0xD2393164DE61AF9DULL, "6127232_m.fxe"},
	{0xD23941317A8F90ABULL, "mer120_hubert6018444.cl"},
	{0xD2394FEE33D50260ULL, "731807370.ogg"},
	{0xD2397F79A9DC9EC3ULL, "hlac_ht_brothel_n_3373_0z.dds"},
	{0xD239D5772D21DF7AULL, "lt_0_lowtown_docks_a3_n_102_0.phy"},
	{0xD239F1EE6FE30B93ULL, "commandtypes.gda"},
	{0xD23A47177FAB8FE0ULL, "6122766_m.fxe"},
	{0xD23A98CDB3DCC280ULL, "hlac_lt_backalley_g_65536_1z.dds"},
	{0xD23AC75E63ED9905ULL, "28923351.ogg"},
	{0xD23B0B421F95C40EULL, "hf_corthwaite.mor"},
	{0xD23B21846681D389ULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.impa.gad"},
	{0xD23BFFECC2E70A1DULL, "vxca_caverns1_mines_ik5_0_0.vlm"},
	{0xD23C571E9F3B2FCBULL, "6140130_m.fxe"},
	{0xD23CE22E1DBE19D2ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twra_0s.dds"},
	{0xD23DED1CA1287D78ULL, "58855146.ogg"},
	{0xD23DFFA040D3434CULL, "720984482.ogg"},
	{0xD23E2168C368CA49ULL, "gen00fl_merrill6121759.cl"},
	{0xD23E7ED595232C3CULL, "vxht_mansion_n_598_0_0.vlm"},
	{0xD23E9938EE353D6AULL, "vxdr_primevaltomb_9co_0_1.vlm"},
	{0xD23ED77487D67747ULL, "6099368_m.fxe"},
	{0xD23F1A33EF025B36ULL, "gen00fl_fenris6126706.cl"},
	{0xD23FB051E2371B7BULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_abma_0d.dds"},
	{0xD23FEBD6FA115CABULL, "matdef_dh_specshft.dds"},
	{0xD23FF754BA79B020ULL, "lt_0_lowtown_a3_n_c_36_0_lc.mmh"},
	{0xD23FF884225ED88FULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_lrgtileweb.mao"},
	{0xD2400484225EED2CULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_lrgtileweb.mmh"},
	{0xD2400A84225EF71EULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_lrgtileweb.msh"},
	{0xD24028CFEA9AF23CULL, "518250320.ogg"},
	{0xD24059376D837D15ULL, "6165753_m.fxe"},
	{0xD240D553125C81D1ULL, "356331086.ogg"},
	{0xD2416359D1673AFEULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_galowint.phy"},
	{0xD241F2846687A2D7ULL, "art\\vfx\\creatures\\vfx_c_rage_crust_c.impa.ani"},
	{0xD2420007D2EA3A17ULL, "6086005_m.fxe"},
	{0xD242C3778A9F6419ULL, "vxca_caverns2prc_8bl_0_0.vlm"},
	{0xD243082FE6ADC40CULL, "mag360cr_samson.utc"},
	{0xD2431F90DA364D3DULL, "m_mag131ar_bloodmages.bnk"},
	{0xD24324F9D5241B99ULL, "genip_crate_1.utp"},
	{0xD243C51C0163C769ULL, "vxlt_0_lowtown_h6v_0_0.vlm"},
	{0xD243FA4218D3D82DULL, "gen00fl_aveline6139320.cl"},
	{0xD2442AC7CD22CBFAULL, "vxw_encounter2_camp_24_0_0.vlm"},
	{0xD2443F5E6CA0877FULL, "229399391.ogg"},
	{0xD244423CFA431CA1ULL, "vxw_woundedcoast_7e_0_0.vlm"},
	{0xD2445CFB37DC6BFCULL, "vxht_keep_5x4_0_0.vlm"},
	{0xD244819A7E3F4AC1ULL, "2p_100.stg"},
	{0xD2448D42BF1F55CDULL, "6113055_m.fxe"},
	{0xD244A8D6640AE4E3ULL, "hlac_lt_0_lowtown_docks_n_-65537_5x.dds"},
	{0xD244D982A02AC890ULL, "vxht_chantry_n_2cm_0_0.vlm"},
	{0xD2461460F961B0E8ULL, "gen00fl_anders6113067.cl"},
	{0xD24692BBDF440DD2ULL, "6151227_m.fxe"},
	{0xD246AC4985ED928AULL, "llac_w_blightlands_65538_0.dds"},
	{0xD246AFC9D3252A3EULL, "108073474.ogg"},
	{0xD2474D0EEA4E914CULL, "ts_dalish_arm_props.gda"},
	{0xD24782EC03EBA547ULL, "vxht_0_hightown_n_bwb_0_0.vlm"},
	{0xD247C83D6BC9CEB6ULL, "vxca_cavern1_haunted_jff_0_0.vlm"},
	{0xD2481E7F71CF177AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5_tool.msh"},
	{0xD2482C7F71CF2EB3ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5_tool.mao"},
	{0xD248307F71CF35C0ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box_5x_5_tool.mmh"},
	{0xD24837AA98E60599ULL, "vxw_encounter2_camp_mz_0_0.vlm"},
	{0xD24860A38A53AFC7ULL, "vxlt_gallowstemplar_n_43c_0_0.vlm"},
	{0xD24907C4E870851DULL, "1044799111.ogg"},
	{0xD24949F8104B0952ULL, "6164942_m.fxe"},
	{0xD249650D1C63AB04ULL, "gen_im_trs_krk_hvy_boo_r1.uti"},
	{0xD249F3719DCDC471ULL, "vxw_bonepit_1p2_0_0.vlm"},
	{0xD249F7FEFD07A75BULL, "6017111_m.fxe"},
	{0xD24A5DF0D297DD26ULL, "vxht_0_hightown_a2_65f_0_0.vlm"},
	{0xD24AA8FE8C1364CAULL, "hlac_ca_caverns2_5802_9z.dds"},
	{0xD24AAF0DADCB0F9FULL, "hlac_lt_0_lowtown_a2_n_-131072_7y.dds"},
	{0xD24ABD5030D70AF1ULL, "ser121_jethann6175344.cl"},
	{0xD24AD273CB1C1002ULL, "6135997_m.fxe"},
	{0xD24B603650B49B08ULL, "fen110_fenris6101948.cl"},
	{0xD24B6B0E9EAB1930ULL, "6116332_m.fxe"},
	{0xD24BBDA159D9E45EULL, "vxca_caverns1_mines_i6i_0_0.vlm"},
	{0xD24BD822251F03CAULL, "386380262.ogg"},
	{0xD24C7CFC052D60B6ULL, "57621007.ogg"},
	{0xD24CFBFEEA1D6DA1ULL, "zz_oceans_11_day_v1.cut"},
	{0xD24D31B3B10C04D7ULL, "gen_im_arm_cht_elf_coma_t6.uti"},
	{0xD24DABB73E02DA82ULL, "gen00fl_varric6020611.cl"},
	{0xD24DB0C337725313ULL, "6040021_m.fxe"},
	{0xD24DC88680F81E36ULL, "l_qun210ar_fq_random_enc.wbd"},
	{0xD24E06DCB66F4CEDULL, "gen00fl_aveline6112730.cl"},
	{0xD24EAA5C0A94FB1DULL, "vxlt_0_lowtown_a2_n_e19_0_0.vlm"},
	{0xD24EE5AB6B1BB1B6ULL, "hlac_lt_0_lowtown_a2_n_-65534_0x.dds"},
	{0xD24F2CCAAB2B5C57ULL, "vxlt_0_lowtown_a2_n_c_gmk_0_0.vlm"},
	{0xD24F478ABD6F8B0AULL, "vxlt_hangedman_n_1mp_0_0.vlm"},
	{0xD24FC96779FF2CE3ULL, "llac_lt_darktown_-1_23.dds"},
	{0xD250F42B8AD07C41ULL, "766676940.ogg"},
	{0xD25147A2D5E34A3EULL, "6136209_f.fxe"},
	{0xD25172D4F49C54A7ULL, "vxw_bonepitdestroyed_a3_16n_0_0.vlm"},
	{0xD2519ABA203E1FC3ULL, "w_bonepitdestroyed_a3_236_0.msh"},
	{0xD251A8BA203E371DULL, "w_bonepitdestroyed_a3_236_0.mmh"},
	{0xD251ACBA203E3E2EULL, "w_bonepitdestroyed_a3_236_0.mao"},
	{0xD251DF516D21AEDCULL, "dae000_sabina6161506.cl"},
	{0xD25221189F3BCFF3ULL, "llac_ht_mansion2b_818_63.dds"},
	{0xD252B07F1247A3A0ULL, "dae000_bartrand6206881.cl"},
	{0xD253432CF150F5FFULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02b.msh"},
	{0xD253492CF150FFA1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02b.mmh"},
	{0xD25350A7F91C0C64ULL, "vxht_0_hightown_a3_bko_0_0.vlm"},
	{0xD253AFE7519DD014ULL, "follower_banter6136371.cl"},
	{0xD253FE162121CB12ULL, "genip_cage.utp"},
	{0xD254002A3513510CULL, "118095329.ogg"},
	{0xD2541AEBFF43FF9FULL, "vxlt_0_lowtown_docks_n_b7k_0_0.vlm"},
	{0xD25426E58437BFDEULL, "6080571_f.fxe"},
	{0xD2546C284E422F0BULL, "vxlt_0_lowtown_a3_n_c_hpz_0_0.vlm"},
	{0xD2548F2DF03D4F7EULL, "6064004_m.fxe"},
	{0xD255496B247DBD22ULL, "vxlt_0_lowtown_docks_n_8uf_0_0.vlm"},
	{0xD255BB3FD8ADD17AULL, "6099317_f.fxe"},
	{0xD255DA8F0D1605CAULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrow.mao"},
	{0xD255DE8F0D160C19ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrow.mmh"},
	{0xD255E88F0D161D07ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafrow.msh"},
	{0xD257200027A477E9ULL, "vxht_chantry_n_2tm_0_0.vlm"},
	{0xD25728DA5955E8CBULL, "hlac_ht_mansion2b_818_29z.dds"},
	{0xD2578A2F1E5F5E24ULL, "258509742.ogg"},
	{0xD258CB86BC63428EULL, "608897325.ogg"},
	{0xD258FA63D165436DULL, "hlac_ht_keep_n_3739_1x.dds"},
	{0xD25969AD884FA70EULL, "896710379.ogg"},
	{0xD259D8FB5D0F1871ULL, "6099443_m.fxe"},
	{0xD25A13F3FA1FEED3ULL, "vxca_caverns1_e8s_0_0.vlm"},
	{0xD25A1725CCD8E5AEULL, "vxht_chantry_30t_0_0.vlm"},
	{0xD25A2830E688485CULL, "lt_0_lowtown_docks_148_0.mao"},
	{0xD25A3430E6885CFFULL, "lt_0_lowtown_docks_148_0.mmh"},
	{0xD25A3A30E68866A1ULL, "lt_0_lowtown_docks_148_0.msh"},
	{0xD25A5BA4A19D064EULL, "tooltips.gfx"},
	{0xD25AB492CFB5DBF2ULL, "vxlt_0_lowtown_docks_a3_d3x_0_0.vlm"},
	{0xD25AD453A081B668ULL, "490053915.ogg"},
	{0xD25B129E49E90F3AULL, "vxht_0_hightown_a2_c7u_0_0.vlm"},
	{0xD25B38CEB58B75BBULL, "262107021.ogg"},
	{0xD25B6BFB635856F0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_ric_0d.dds"},
	{0xD25BBA7C08E1C4B4ULL, "6216144_m.fxe"},
	{0xD25BC6A96D2056B3ULL, "526533898.ogg"},
	{0xD25BD322B74A8B50ULL, "hlac_w_woundedcoast_0_6y.dds"},
	{0xD25C15CEF0CB19F1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4so_basic_2.mao"},
	{0xD25C576542854313ULL, "dae000_hate_follower_react6103197.cl"},
	{0xD25C663278F0E13AULL, "6125888_m.fxe"},
	{0xD25C71C65A0D6B8BULL, "vxca_cavern1_haunted_e51_0_0.vlm"},
	{0xD25C73111C9B7E32ULL, "514647784.ogg"},
	{0xD25CC83BF442AAE3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sidel.msh"},
	{0xD25CD63BF442C23DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sidel.mmh"},
	{0xD25CDA3BF442C94EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_sidel.mao"},
	{0xD25DB967AF65A38AULL, "vxlt_hangedman_n_1jq_0_0.vlm"},
	{0xD25DF0F61D00132AULL, "551877511.ogg"},
	{0xD25E1299563EA205ULL, "mrl351_pridedemon6165559.cl"},
	{0xD25E5C0D1B645332ULL, "vxlt_0_lowtown_ff7_0_0.vlm"},
	{0xD25EA85EA08F6CA5ULL, "ser191_emeric6120427.cl"},
	{0xD25ECAF5BC676C06ULL, "and230tr_dissent_anders.ncs"},
	{0xD25EE31CB3E0C6A7ULL, "dr_primevaltomb_12048_ll.mmh"},
	{0xD25F4F3FD2D83E39ULL, "910390014.ogg"},
	{0xD25F8AC9931F82C5ULL, "hlac_dr_primevaltomb_658_4x.dds"},
	{0xD25FD4347052A150ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxb.mao"},
	{0xD25FD8347052A863ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxb.mmh"},
	{0xD25FE07586699C0DULL, "isa311_isabela6092228.cl"},
	{0xD25FE6347052BFBDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxb.msh"},
	{0xD2600290A6730FBDULL, "vxht_keep_n_5u5_0_0.vlm"},
	{0xD2601D108728F00FULL, "590068357.ogg"},
	{0xD2603332F7B08C73ULL, "ht_0_hightown_a2_29_0.anb"},
	{0xD260838348ADE7B4ULL, "822016858.ogg"},
	{0xD260852A941AAF26ULL, "llac_lt_0_lowtown_docks_a3_196608_1.dds"},
	{0xD260B08EDFA142C5ULL, "vxw_encounter2_camp_70_0_0.vlm"},
	{0xD260BCA4A084F409ULL, "vxdr_primevaltomb_6xg_0_0.vlm"},
	{0xD260E8F04BA10B03ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_reaction.evt"},
	{0xD2610D9EBADF0B6AULL, "173788457.ogg"},
	{0xD2614BB36050DBB4ULL, "vxw_bonepitmining_a2_25m_0_0.vlm"},
	{0xD26176AA3916DD8AULL, "6134815_m.fxe"},
	{0xD261C284EB1224E4ULL, "vxht_0_hightown_a3_bfl_0_0.vlm"},
	{0xD2620DA9329A6425ULL, "hlac_lt_0_lowtown_a2_n_c_131072_6y.dds"},
	{0xD262DC156C23C732ULL, "vxw_encounter1_ed_0_0.vlm"},
	{0xD262EBA9010B42A6ULL, "6141723_m.fxe"},
	{0xD263440DFE4846CAULL, "vxht_brothel_n_4vq_0_0.vlm"},
	{0xD2634C2FA0D5DB4EULL, "gen00fl_isabela6135610.cl"},
	{0xD26394A9BB748D52ULL, "vxca_caverns1_e8v_0_0.vlm"},
	{0xD2640E1178938F31ULL, "718181302.ogg"},
	{0xD2648F3CA33CA7F2ULL, "dae_monster_aoe_instant.ncs"},
	{0xD264AD6A2BF53D48ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_cth_comx_0t.dds"},
	{0xD2651B647CD23699ULL, "vxw_encounter2_camp_cx_0_0.vlm"},
	{0xD26660A53A4C5DBBULL, "vxht_keep_n_61l_0_0.vlm"},
	{0xD26682920DC79019ULL, "6142211_m.fxe"},
	{0xD266942A0EE399C1ULL, "cd_mcr371_argument.bnk"},
	{0xD266A638967719E3ULL, "vxht_0_hightown_n_all_0_0.vlm"},
	{0xD266B40C2F74E14AULL, "vxht_keep_n_5xm_0_0.vlm"},
	{0xD266CB5512DEB750ULL, "vxlt_warehouse_n_9q_0_0.vlm"},
	{0xD266E56A3CCD57B5ULL, "vxca_caverns1_mines_ehh_0_0.vlm"},
	{0xD267558398C9539DULL, "vxca_caverns1_mines_iz2_0_0.vlm"},
	{0xD267A60336263D8EULL, "dae100_alistair_drunk6135104.cl"},
	{0xD267C1F04BA6E784ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_reaction.gad"},
	{0xD2681571EB8859F5ULL, "vxlt_warehouse_n_ba_0_0.vlm"},
	{0xD2683BD4800F6396ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_app_p.impa.ani"},
	{0xD2686D250C4B39A2ULL, "vxlt_gallowstemplar_5cd_0_0.vlm"},
	{0xD269266FC1ABF472ULL, "vxlt_0_lowtown_docks_a3_d4y_0_0.vlm"},
	{0xD269847B3BDF27BAULL, "vxht_0_hightown_a2_c2v_0_0.vlm"},
	{0xD26992D76702C485ULL, "mcr370_cut_epilogue6207415.cl"},
	{0xD269DD93507A46ADULL, "vxht_keep_36y_0_0.vlm"},
	{0xD26A859AE9C645AAULL, "zz_dae_debug6047626.cl"},
	{0xD26AF6FB0CCF9145ULL, "screenspace.psh.fxo"},
	{0xD26C10848F178D57ULL, "vxlt_0_lowtown_a2_n_c_gce_0_0.vlm"},
	{0xD26C1A4DB3DDEA31ULL, "gr_sp_01.mao"},
	{0xD26C4557D26ABA86ULL, "ts_dalish_arm_table.gda"},
	{0xD26C6275B0D9E931ULL, "459567726.ogg"},
	{0xD26CBB8ABA9B2AC0ULL, "235341844.ogg"},
	{0xD26D54150372E45CULL, "170693767.ogg"},
	{0xD26D650C2B64C2E6ULL, "6147394_m.fxe"},
	{0xD26DF10F2666697BULL, "blue50.dds"},
	{0xD26E1CA2ECE5FB5DULL, "guardian_dmgreduction.dds"},
	{0xD26E386FEA304A69ULL, "character_character_p118.psh.fxo.dx11.assembly"},
	{0xD26E3D3100670A17ULL, "mag121_idunna6084219.cl"},
	{0xD26EA4716F816842ULL, "vxdr_0_deeproads_crs_0_2.vlm"},
	{0xD26EB8894D6FACC2ULL, "6085827_m.fxe"},
	{0xD26ED0F04BAD2062ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_reaction.ani"},
	{0xD26F2DCF06131FB4ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v1.ani"},
	{0xD26F397E1FB90BB7ULL, "vxlt_undercityencounter_1gc_0_0.vlm"},
	{0xD26F69C7EC8470A0ULL, "follower_banter6163580.cl"},
	{0xD26F716AF0083CABULL, "hlac_ht_0_hightown_a3_-131074_0z.dds"},
	{0xD26F9E0CF0EFA35CULL, "458694758.ogg"},
	{0xD26FB4F403801B64ULL, "1015527263.ogg"},
	{0xD26FC7B7F1F49D1EULL, "26178431.ogg"},
	{0xD270647BBAB4237DULL, "hlac_ht_keep_3731_0x.dds"},
	{0xD2706A017A66A2B8ULL, "ave321cr_guardsman1.utc"},
	{0xD2706E6A34F8B3FBULL, "gen00fl_bethany6101382.cl"},
	{0xD2708E9CEEDBAD9EULL, "157999150.ogg"},
	{0xD27107336A82E005ULL, "6081965_m.fxe"},
	{0xD2711F2527C213BAULL, "85992306.ogg"},
	{0xD2714DF25E19DFB2ULL, "vxw_encounter1_je_0_0.vlm"},
	{0xD27166FF45876223ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword.evt"},
	{0xD271AEE72764F18BULL, "6164309_m.fxe"},
	{0xD271C5C453957369ULL, "151692986.ogg"},
	{0xD271D64678DB5B7EULL, "6165383_m.fxe"},
	{0xD271F04360491B32ULL, "vxdr_primevaltomb_9dc_0_0.vlm"},
	{0xD27208C1EF66C594ULL, "vxht_bartrand_n_5mg_0_0.vlm"},
	{0xD272193C8FD117C6ULL, "6076685_m.fxe"},
	{0xD2730C50AD17B607ULL, "vxca_cave_2l8_0_0.vlm"},
	{0xD273A5BC252FAA74ULL, "6219215_m.fxe"},
	{0xD27431CEE328D9F6ULL, "dae000_bartrand6031173.cl"},
	{0xD2755A2B35B14CD2ULL, "299740216.ogg"},
	{0xD275D892ACAEB453ULL, "llac_w_encounter1_65540_0.dds"},
	{0xD275FF71A50F960AULL, "hlac_lt_0_lowtown_docks_a3_-65537_4x.dds"},
	{0xD2762ACF06193916ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v1.gad"},
	{0xD27643D4801BAE87ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_app_p.impa.evt"},
	{0xD276937DD68C4ACBULL, "465090474.ogg"},
	{0xD2769B09391E7B61ULL, "mag170_hugh6080771.cl"},
	{0xD276A25BC8A072CCULL, "zz_one_debug6099962.cl"},
	{0xD276BCB540FB4971ULL, "6093643_m.fxe"},
	{0xD276D8528B4F0CFFULL, "hlac_lt_0_lowtown_docks_n_-131073_2x.dds"},
	{0xD276E0559E313719ULL, "834004707.ogg"},
	{0xD2772352456898D6ULL, "vxht_brothel_n_5jt_0_0.vlm"},
	{0xD27760928E4E2D35ULL, "6115112_m.fxe"},
	{0xD2778C4CBE9EEDECULL, "vxlt_0_lowtown_n_eqx_0_0.vlm"},
	{0xD2779151E800874CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12man.msh"},
	{0xD2779751E800913EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12man.mmh"},
	{0xD277A351E800A5D5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp12man.mao"},
	{0xD277F6582DD5403AULL, "vxht_0_hightown_a2_c9w_0_0.vlm"},
	{0xD278076A9B3E076CULL, "6172587_m.fxe"},
	{0xD2783FFF458D3EA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword.gad"},
	{0xD2787B5F013BF357ULL, "mag320_orsino6071169.cl"},
	{0xD278C3A5FE053ADEULL, "vxlt_0_lowtown_docks_a3_n_d32_0_1.vlm"},
	{0xD27942BE0D7E2274ULL, "6050104_m.fxe"},
	{0xD2794DA546656C6AULL, "vxlt_0_lowtown_a2_n_c_i0w_0_0.vlm"},
	{0xD279EE86811D99A4ULL, "l_qun210ar_fq_random_enc.bnk"},
	{0xD27AD037CD04755DULL, "qun111_ketojan6150162.cl"},
	{0xD27B465239760FE1ULL, "6187894_m.fxe"},
	{0xD27BE142253E6666ULL, "plt_cod_itm_enasalin.plo"},
	{0xD27C5C0A9E10282DULL, "6140631_m.fxe"},
	{0xD27CC09A7160A543ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4sd_basic.mao"},
	{0xD27CC49A7160AB90ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4sd_basic.mmh"},
	{0xD27CD29A7160C3AAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4sd_basic.msh"},
	{0xD27CE346DCAF91F8ULL, "lt_0_lowtown_a2_n_c_25_0.rml"},
	{0xD27CF1AD9BA306A7ULL, "vxw_bonepit_n9_0_0.vlm"},
	{0xD27D45CF061F8675ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v1.evt"},
	{0xD27D4E3C442C7FCEULL, "6112705_m.fxe"},
	{0xD27D64D4802205D0ULL, "art\\vfx\\combat\\vfx_r_dwp_reversal_app_p.impa.gad"},
	{0xD27D6CE6AB18EA6FULL, "hlac_lt_0_lowtown_-3_5y.dds"},
	{0xD27DA67394339331ULL, "plt_prm_cod_air_ring.plo"},
	{0xD27DFE3357455C65ULL, "zz_one_debug6058984.cl"},
	{0xD27E31B5355C7F27ULL, "vxca_cavern1_haunted_iim_0_0.vlm"},
	{0xD27F4EFF45937682ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword.ani"},
	{0xD27F68FFB757CA1BULL, "lt_warehouse_1_0.rml"},
	{0xD27F891C1BA240D1ULL, "888167773.ogg"},
	{0xD2804A2CF177EF1EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear02b.phy"},
	{0xD2806606B7C4A8B9ULL, "780097828.ogg"},
	{0xD280CD901DA8AC0BULL, "6095868_f.fxe"},
	{0xD28134411D7177DBULL, "dae000ip_crowd_ff1_talk14.utp"},
	{0xD2816736DEC5E6F6ULL, "vxw_blightlands_63_0_0.vlm"},
	{0xD2817BB33EE58958ULL, "hf_cassandra.mor"},
	{0xD28205B55104D02FULL, "vxca_caverns1_dpf_0_0.vlm"},
	{0xD2821D3B8A40101AULL, "fen210_hadriana6110597.cl"},
	{0xD2826C8C99DBB9BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_self_2.evt"},
	{0xD282EE60581C66FBULL, "character_character_p273.psh.fxo.dx11"},
	{0xD2831EF88D50A299ULL, "769137282.ogg"},
	{0xD283A0B7B6CE314DULL, "gen00fl_aveline6129044.cl"},
	{0xD283B2CBF86FBC5EULL, "dae000_gamlen6124365.cl"},
	{0xD283C92420B988B5ULL, "vxca_caverns1_mines_ebv_0_0.vlm"},
	{0xD284164EEDA94D20ULL, "274097634.ogg"},
	{0xD2843AF482579AC3ULL, "vxht_mansion_n_62m_0_0.vlm"},
	{0xD284400B0BFAC420ULL, "6172789_m.fxe"},
	{0xD2844D76518BFB82ULL, "mag321_comtesse_de_launcet6061395.cl"},
	{0xD284556EAE0FC33CULL, "gen00fl_fenris6116585.cl"},
	{0xD28465247EE86FF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_action.ani"},
	{0xD2852F4AF8934556ULL, "plt_cod_prc_books.plo"},
	{0xD285D33C065FE992ULL, "gen00fl_aveline6132885.cl"},
	{0xD285FAF758851AF6ULL, "zz_fen_debug6134394.cl"},
	{0xD28624033DCD3FAEULL, "race_base.gda"},
	{0xD2866FFFB75DF573ULL, "lt_warehouse_1_0.phy"},
	{0xD28694A8793C904FULL, "6056619_m.fxe"},
	{0xD286BD44977EDCFFULL, "vxw_sundermount_36s_0_0.vlm"},
	{0xD286E67EBAAFC8FAULL, "dae300_cullen6140637.cl"},
	{0xD2870028F65E74E6ULL, "6104262_f.fxe"},
	{0xD287A37ECB625ABBULL, "hla_ht_0_hightown_a3_-131071_0z.dds"},
	{0xD287BECCC84C8276ULL, "590298515.ogg"},
	{0xD2894F0A1A1C93CEULL, "vxw_sundermount_37q_0_0.vlm"},
	{0xD289743FA9173931ULL, "6121568_m.fxe"},
	{0xD289838C99E1FFE9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_self_2.gad"},
	{0xD289C96DD778DBB6ULL, "m_rai300ar_encounter.wbd"},
	{0xD28A4E78DB473DF6ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls105.mao"},
	{0xD28AEAE149CBEA41ULL, "w_sundermount_92_0.phy"},
	{0xD28AFD99099B74C7ULL, "621782470.ogg"},
	{0xD28B41A86D9648CAULL, "vxlt_0_lowtown_docks_c4e_0_0.vlm"},
	{0xD28B46A7E117CDC1ULL, "6177949_m.fxe"},
	{0xD28B638A8D991F27ULL, "vxw_bonepit_i8_0_0.vlm"},
	{0xD28BEF58B848DF9AULL, "602560146.ogg"},
	{0xD28BFE5F0D947FC5ULL, "dae000_worthy6166644.cl"},
	{0xD28C5032F7D5F82FULL, "ht_0_hightown_a2_29_0.rml"},
	{0xD28CA6D464ACCBF7ULL, "6020086_f.fxe"},
	{0xD28CA9C69E25E96CULL, "6027366_m.fxe"},
	{0xD28D9E95EF63D34CULL, "555192032.ogg"},
	{0xD28DEBE1B1D2041DULL, "mer320_hubert6082036.cl"},
	{0xD28DFEA44298B74BULL, "gen_im_acc_blt_act1_07.uti"},
	{0xD28E0A44309C9891ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier4_arishokresponse.ani"},
	{0xD28E1AE397FA67F9ULL, "137709237.ogg"},
	{0xD28E9ACE53E2BC88ULL, "art\\vfx\\textures\\vfx_electric_tile.dds"},
	{0xD28F070B37A6CFDEULL, "static_static_p14.psh.fxo"},
	{0xD28F1B2F90ECEDE2ULL, "seb200im_fade_1h.uti"},
	{0xD28F241838611B08ULL, "vxht_mansion_n_3hm_0_0.vlm"},
	{0xD28FCADF6AC94C72ULL, "663331273.ogg"},
	{0xD29014674BE83F23ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_exit.ani"},
	{0xD290548C99E7CE37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_generic_self_2.ani"},
	{0xD290BEFC1292E6A7ULL, "325842337.ogg"},
	{0xD290F3E149D0655DULL, "w_sundermount_92_0.rml"},
	{0xD29144D10C9B3A3DULL, "vxdr_slavers_1b2_0_0.vlm"},
	{0xD2916D247EF307E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_action.evt"},
	{0xD29182731526E25DULL, "mag131cr_keran_civ.utc"},
	{0xD291BE0CB663403AULL, "ico_2h_axe_2.dds"},
	{0xD291E093EB455A16ULL, "vxlt_0_lowtown_docks_n_b5t_0_0.vlm"},
	{0xD291E82709C7C86EULL, "one280_emrys6104273.cl"},
	{0xD292371A0C037201ULL, "qcr220_meredith6060314.cl"},
	{0xD2923B0112AFA135ULL, "vxca_caverns1_mines_egu_0_0.vlm"},
	{0xD292BB0E37F763F6ULL, "198828262.ogg"},
	{0xD292E8CF9FFE6E6AULL, "69445660.ogg"},
	{0xD2936B08C16AA375ULL, "vxlt_warehouse_n_sn_0_0.vlm"},
	{0xD2939338B3CF4641ULL, "gen00fl_varric6033724.cl"},
	{0xD293C2BBE22D8322ULL, "vxlt_gallowstemplar_5rg_0_0.vlm"},
	{0xD293D213D11CF7BFULL, "vxlt_0_lowtown_n_ett_0_0.vlm"},
	{0xD29420E48B16AD70ULL, "vxdr_0_deeproads_fbj_0_0.vlm"},
	{0xD29448339E5C8B86ULL, "gen00fl_varric6080428.cl"},
	{0xD29452B88360AEE1ULL, "667404099.ogg"},
	{0xD294CE772D6E7EA3ULL, "lt_0_lowtown_docks_a3_n_102_0.msh"},
	{0xD294D5461110F09FULL, "798602726.ogg"},
	{0xD294DC772D6E96FDULL, "lt_0_lowtown_docks_a3_n_102_0.mmh"},
	{0xD294E0772D6E9D0EULL, "lt_0_lowtown_docks_a3_n_102_0.mao"},
	{0xD29506856B32A8CFULL, "6053519_m.fxe"},
	{0xD295154430A2CA33ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier4_arishokresponse.gad"},
	{0xD295C9B95A2634F3ULL, "art\\vfx\\creatures\\vfx_c_hd_fswipe_c.anb"},
	{0xD296476BF9986DABULL, "gen00fl_anders6119176.cl"},
	{0xD296DE6463798702ULL, "qun131_party_comment6117281.cl"},
	{0xD2973B44D3563055ULL, "vxlt_foundry_n_vc_0_0.vlm"},
	{0xD297A02E698120CEULL, "6028400_m.fxe"},
	{0xD297B100F09DC4EEULL, "llac_ht_0_hightown_a2_65538_0.dds"},
	{0xD297C0E70C12AC4EULL, "vxw_sundermount_32r_0_0.vlm"},
	{0xD297D449B7FB6FBCULL, "220618762.ogg"},
	{0xD2983728C59C97F7ULL, "vxht_keep_n_61x_0_0.vlm"},
	{0xD2988E247EF95E30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\aveline_drop_shield_action.gad"},
	{0xD2998E8AC913914DULL, "304351988.ogg"},
	{0xD299B3855F8C614AULL, "vxlt_0_lowtown_docks_c1f_0_0.vlm"},
	{0xD299CA046E4B71BDULL, "vxht_keep_n_5i1_0_0.vlm"},
	{0xD29A0C8D1EF1FA57ULL, "hlac_ht_chantry_2534_3y.dds"},
	{0xD29A934E7F6FE3D5ULL, "vxlt_gallowstemplar_f_5rk_0_0.vlm"},
	{0xD29ACFEE48194663ULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.anb"},
	{0xD29B156F1948B027ULL, "vxca_cavern1_haunted_ico_0_0.vlm"},
	{0xD29B5E09D513305CULL, "vxlt_0_lowtown_a2_n_c_df2_0_0.vlm"},
	{0xD29B8A0517534612ULL, "6122020_f.fxe"},
	{0xD29B945CF62676C4ULL, "vxw_encounter2_camp_pp_0_0.vlm"},
	{0xD29C147893A38BF2ULL, "ave121_brennan6105830.cl"},
	{0xD29C7E6E49897658ULL, "6202029_f.fxe"},
	{0xD29CBD1ED983B560ULL, "trainingsteps.gda"},
	{0xD29CF9B2756B8BBDULL, "423189813.ogg"},
	{0xD29D4CEC9CF3843BULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.cess.ani"},
	{0xD29D6533D14D969BULL, "416834122.ogg"},
	{0xD29D78E8299CE85EULL, "static_static_p47.psh.fxo"},
	{0xD29E2C674BF4A51AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_exit.evt"},
	{0xD29E2DF4C8245720ULL, "569508250.ogg"},
	{0xD29E89D271FB7532ULL, "gen00fl_aveline6131901.cl"},
	{0xD29E98DB193C6C6DULL, "vxht_mansion_n_6cp_0_0.vlm"},
	{0xD29EC45836C379C1ULL, "151429612.ogg"},
	{0xD29F681484BB63DCULL, "ca_caverns1_mines_24426_ll.mmh"},
	{0xD29F72AA4D21A991ULL, "106008590.ogg"},
	{0xD29FBC7FB36D9B1CULL, "hlac_lt_0_lowtown_a3_n_c_65534_2x.dds"},
	{0xD29FDA879FAB5481ULL, "hlac_lt_0_lowtown_n_-3_5x.dds"},
	{0xD2A00834DBD632AFULL, "6163443_m.fxe"},
	{0xD2A012823FE51F62ULL, "6096346_m.fxe"},
	{0xD2A03D9596116DD8ULL, "950353713.ogg"},
	{0xD2A1B6C673DA77C0ULL, "llac_lt_0_lowtown_docks_a3_n_-196607_0.dds"},
	{0xD2A1C26DC4D138FBULL, "vxht_0_hightown_c3i_0_1.vlm"},
	{0xD2A1DCE5B360BBF5ULL, "vxlt_warehouse_n_vm_0_0.vlm"},
	{0xD2A1E87C202EC071ULL, "6047336_m.fxe"},
	{0xD2A223EF8904D094ULL, "6077738_m.fxe"},
	{0xD2A243F0C313103FULL, "vxlt_0_lowtown_n_eqw_0_0.vlm"},
	{0xD2A261487B33D57DULL, "vxht_bartrand_n_32e_0_0.vlm"},
	{0xD2A29C02BEFE2331ULL, "6070019_m.fxe"},
	{0xD2A29F901D276A2BULL, "vxlt_0_lowtown_a2_gnc_0_0.vlm"},
	{0xD2A2A8CAAD9D911DULL, "221868697.ogg"},
	{0xD2A3261E068432FAULL, "hlac_w_bonepitdestroyed_a3_393221_0z.dds"},
	{0xD2A3ACC348D9D000ULL, "vxw_blightlands_gv_0_0.vlm"},
	{0xD2A3E6B95A32A35CULL, "art\\vfx\\creatures\\vfx_c_hd_fswipe_c.mmh"},
	{0xD2A49506BA73CE2CULL, "6024077_m.fxe"},
	{0xD2A49508C78C1A77ULL, "691672000.ogg"},
	{0xD2A4BC9069CE68BEULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.cess.evt"},
	{0xD2A4DFE6E0F53097ULL, "vxlt_0_lowtown_docks_bgh_0_0.vlm"},
	{0xD2A52B674BFAC20DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_exit.gad"},
	{0xD2A53E3D79B63DFEULL, "mag100_thrask6031967.cl"},
	{0xD2A57D39A32B8828ULL, "673055709.ogg"},
	{0xD2A5A1DB61FA8E97ULL, "mag200_feynriel6023972.cl"},
	{0xD2A5B66BDB3B2540ULL, "671305905.ogg"},
	{0xD2A5F9F2EAD31F7CULL, "llac_ca_caverns2_5802_10.dds"},
	{0xD2A62D79A5D27177ULL, "lt_backalley_g_2_0_ll.mmh"},
	{0xD2A668298793244AULL, "vxdr_slavers_1ji_0_0.vlm"},
	{0xD2A673CFF9196719ULL, "hlac_lt_warehouse_n_0_9z.dds"},
	{0xD2A68D52B9C4A3ABULL, "mcr370_negotiation6080249.cl"},
	{0xD2A730815F178E24ULL, "vxlt_gallowstemplar_f_3rx_0_0.vlm"},
	{0xD2A762C0B333B61FULL, "vxca_caverns1_mines_cpd_0_0.vlm"},
	{0xD2A7ACEE4823948CULL, "art\\vfx\\combat\\vfx_spd_web_imp_c.mmh"},
	{0xD2A7F7787AFC2F83ULL, "132294860.ogg"},
	{0xD2A83CB95699B9F6ULL, "6144315_m.fxe"},
	{0xD2A875CC7843EDBAULL, "lgt210_eau_symp_templar6144467.cl"},
	{0xD2A8BFA2506A21E6ULL, "w_encounter2_camp_30_0_ll.mmh"},
	{0xD2A93910E3E58962ULL, "dre000_varric6035188.cl"},
	{0xD2A9C446DCD64AB0ULL, "lt_0_lowtown_a2_n_c_25_0.anb"},
	{0xD2AA0639E81C8F44ULL, "vxw_encounter2_camp_kq_0_0.vlm"},
	{0xD2AA485FF1977867ULL, "vxlt_hangedman_n_2mj_0_0.vlm"},
	{0xD2AA54EC9CFE1C32ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.cess.evt"},
	{0xD2AA67F5AC175A3DULL, "571294483.ogg"},
	{0xD2AACC0C1304E68AULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlm_meda_0n.dds"},
	{0xD2ABB83FF37B3C35ULL, "hlac_ht_chantry_n_2534_1z.dds"},
	{0xD2ABD39069D4AEE9ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.cess.gad"},
	{0xD2AC01DEED222063ULL, "vxdr_primevalentrance_2dq_0_0.vlm"},
	{0xD2ACBBF154710BCFULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_2_ian.gad"},
	{0xD2ACEB8D86082683ULL, "6051584_m.fxe"},
	{0xD2ACF9325292939BULL, "526060093.ogg"},
	{0xD2AD2C93799D3408ULL, "gen00fl_isabela6089267.cl"},
	{0xD2AD9766AFB0AA07ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robb_s.dds"},
	{0xD2ADA22A0F1FB473ULL, "cd_mcr371_argument.wbd"},
	{0xD2ADCDD830EBE449ULL, "vxw_bonepitmining_a2_162_0_0.vlm"},
	{0xD2AEDF8950459463ULL, "vxht_0_hightown_n_aus_0_0.vlm"},
	{0xD2AEF06B471EE143ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinswall.mao"},
	{0xD2AEF46B471EE790ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinswall.mmh"},
	{0xD2AF026B471EFFAAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinswall.msh"},
	{0xD2B03F6C1AA13267ULL, "345076417.ogg"},
	{0xD2B04E082E8322A4ULL, "vxlt_0_lowtown_docks_n_b81_0_0.vlm"},
	{0xD2B07DA4F4DFC3B2ULL, "hlac_lt_gallowscourtyard_-65535_0y.dds"},
	{0xD2B0A675C619B422ULL, "vxlt_gallowstemplar_5xa_0_0.vlm"},
	{0xD2B0C4763ED121D1ULL, "vxlt_gallowscourtyard_n_59t_0_0.vlm"},
	{0xD2B0C47D1701785BULL, "vxca_caverns2_8c4_0_0.vlm"},
	{0xD2B0EC0550B17ED2ULL, "867708060.ogg"},
	{0xD2B1274AC949183EULL, "vxht_mansion2_3es_0_0.vlm"},
	{0xD2B1302CDA539AD3ULL, "vxht_brothel_n_47j_0_1.vlm"},
	{0xD2B173EC9D046FC5ULL, "art\\vfx\\combat\\vfx_m_wpn_phys_crust_c.cess.gad"},
	{0xD2B18081AC55AFF7ULL, "zz_ave_debug6056762.cl"},
	{0xD2B180EEF1986427ULL, "vxdr_primevaltomb_5n0_0_0.vlm"},
	{0xD2B1BF21DF447E0DULL, "739215187.ogg"},
	{0xD2B21EA03ACFE880ULL, "vxw_blightlands_hw_0_0.vlm"},
	{0xD2B220879121EA22ULL, "469112777.ogg"},
	{0xD2B273E03AB2C058ULL, "vxht_0_hightown_n_amt_0_0.vlm"},
	{0xD2B27BBEBDCF29FAULL, "549574590.ogg"},
	{0xD2B2A49069DA7D37ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_wpn_c.cess.ani"},
	{0xD2B3B27B8162028FULL, "317839448.ogg"},
	{0xD2B3D54A15D6E116ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.impa.evt"},
	{0xD2B4050DA8A24EF1ULL, "vxht_0_hightown_a2_cbq_0_0.vlm"},
	{0xD2B48D0E1F12E014ULL, "231972493.ogg"},
	{0xD2B4955E34B1BA12ULL, "fen210_orana.cnv"},
	{0xD2B4E4F9981CC0DCULL, "vxw_sundermount_24w_0_0.vlm"},
	{0xD2B50E9A659E0AAFULL, "cs_pro000cs_openingrogue.wbd"},
	{0xD2B537A9481B4820ULL, "426644055.ogg"},
	{0xD2B5481CC223FD95ULL, "hlac_dr_0_deeproads_5842_2z.dds"},
	{0xD2B56ED6AD841108ULL, "vxlt_hangedman_n_1qv_0_0.vlm"},
	{0xD2B5C26533BC7B2CULL, "vxlt_0_lowtown_a2_f4s_0_0.vlm"},
	{0xD2B5C5D1DDFC965BULL, "6046838_m.fxe"},
	{0xD2B5C80CCCE3F1AFULL, "mag311_keran6041514.cl"},
	{0xD2B5EF6DD79E5724ULL, "m_rai300ar_encounter.bnk"},
	{0xD2B6403A0807C20DULL, "vxlt_0_lowtown_docks_a3_n_d27_0_0.vlm"},
	{0xD2B65365B3F019CCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01b.msh"},
	{0xD2B65965B3F023BEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01b.mmh"},
	{0xD2B6642F993EC88EULL, "mer220_hubert6041222.cl"},
	{0xD2B667AC61A7A860ULL, "739419544.ogg"},
	{0xD2B69D062C44DCCBULL, "vxht_0_hightown_c52_0_0.vlm"},
	{0xD2B7324DFE4CBFC8ULL, "821177587.ogg"},
	{0xD2B751136CA2E47CULL, "hlac_lt_gallowscourtyard_n_131071_2z.dds"},
	{0xD2B77708635C14D5ULL, "vxlt_gallowstemplar_f_5xi_0_0.vlm"},
	{0xD2B77B751A17A803ULL, "sst_ss_raider_f_1.gda"},
	{0xD2B831640B8E44CAULL, "6015740_f.fxe"},
	{0xD2B83B2E40D901ECULL, "art\\characters\\creatures\\highdragon\\animation\\blendtree\\highdragon_movement.anb"},
	{0xD2B87816DA12A7C4ULL, "vxw_encounter2_camp_fv_0_0.vlm"},
	{0xD2B8B75E34B5DD6AULL, "fen210_orana.bnk"},
	{0xD2B8BE71D13E56BEULL, "710572628.ogg"},
	{0xD2B90D9FB54D506BULL, "llac_ht_mansion2b_818_27.dds"},
	{0xD2B9238BD92C67C5ULL, "6088328_m.fxe"},
	{0xD2B965E43301A31DULL, "vxlt_ruinedhouse_n_1w_0_1.vlm"},
	{0xD2B9672330A6A2E5ULL, "llac_ht_bartrand_n_3371_0.dds"},
	{0xD2B9765054F4280DULL, "hlac_w_bonepit_196609_0x.dds"},
	{0xD2B9B77740D2055EULL, "6169677_m.fxe"},
	{0xD2BA77530EDCF5FBULL, "532348509.ogg"},
	{0xD2BACC4A15DCF0A1ULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.impa.gad"},
	{0xD2BAF4695AF3DA4FULL, "6149097_f.fxe"},
	{0xD2BB351799D937B3ULL, "vxw_sundermount_35n_0_0.vlm"},
	{0xD2BB7C94FD289D6DULL, "vxht_mansion_n_6in_0_0.vlm"},
	{0xD2BBB294C18A52E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield"},
	{0xD2BC41E600CD06CAULL, "gen00fl_anders6109318.cl"},
	{0xD2BC8C04A8042625ULL, "follower_barks6153293.cl"},
	{0xD2BCC65A3BF8548AULL, "art\\characters\\creatures\\shade\\animation\\blendtree\\shade.anb"},
	{0xD2BCEA09E9F3B2F4ULL, "dae000_orsino6081390.cl"},
	{0xD2BD375A6C3347DDULL, "640680885.ogg"},
	{0xD2BD3D30E6DCE944ULL, "lt_0_lowtown_docks_148_0.phy"},
	{0xD2BD9007F2E93225ULL, "416206758.ogg"},
	{0xD2BDB8626798337AULL, "w_bonepit_122_0.phy"},
	{0xD2BDC0F8F24FD862ULL, "725448338.ogg"},
	{0xD2BDCB9A71984EABULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\4sd_basic.phy"},
	{0xD2BDDBC3315342CFULL, "25773166.ogg"},
	{0xD2BEB2136FC9F4C6ULL, "155352977.ogg"},
	{0xD2BEBFE520793B24ULL, "vxlt_0_lowtown_docks_n_b56_0_0.vlm"},
	{0xD2BEFC7059B1B184ULL, "hlac_ht_0_hightown_a2_-131072_0x.dds"},
	{0xD2BF23D47272BC53ULL, "gencr_carta_thug.utc"},
	{0xD2BF4B9E4AE852EAULL, "llac_lt_0_lowtown_n_131071_5.dds"},
	{0xD2BFF2CBE38E7CA7ULL, "vxdr_primevaltomb_5c3_0_0.vlm"},
	{0xD2C03493DE713740ULL, "vxw_bonepitmining_a2_215_0_0.vlm"},
	{0xD2C06BFFB78F66E7ULL, "lt_warehouse_1_0.anb"},
	{0xD2C0907D2CC60100ULL, "vxw_blightlands_mt_0_0.vlm"},
	{0xD2C09BB40F948A32ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_tmpa_0d.dds"},
	{0xD2C0C2ABBF10C770ULL, "651450939.ogg"},
	{0xD2C0D8F15481E595ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_suspicious_2_ian.ani"},
	{0xD2C154C32FBA8972ULL, "124336184.ogg"},
	{0xD2C17E0F215FF8CCULL, "hlac_ht_playershouse_n_855_23z.dds"},
	{0xD2C1CD4A15E3114FULL, "art\\vfx\\combat\\vfx_glb_silence_crust_c.impa.ani"},
	{0xD2C227AA9052F4B6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ferelden_jimmy_n.dds"},
	{0xD2C25915671E8C74ULL, "942420140.ogg"},
	{0xD2C272C28F7BD6DBULL, "vxht_0_hightown_a2_bwy_0_0.vlm"},
	{0xD2C33ADCF3FC3F61ULL, "vxdr_primevaltomb_75e_0_0.vlm"},
	{0xD2C38C52C0581FF7ULL, "gen00fl_aveline6139472.cl"},
	{0xD2C398AD733265F2ULL, "fen110_anso6094994.cl"},
	{0xD2C3C5F1568CDA0FULL, "follower_banter6172824.cl"},
	{0xD2C3F4A716973006ULL, "hlac_lt_0_lowtown_n_1_9x.dds"},
	{0xD2C41929965ABB34ULL, "6108035_f.fxe"},
	{0xD2C425DFC3C461DDULL, "hlac_lt_0_lowtown_-131072_5z.dds"},
	{0xD2C43330296F250DULL, "static1_bl.psh.fxo"},
	{0xD2C4344225B293ACULL, "vxlt_0_lowtown_a2_f1p_0_0.vlm"},
	{0xD2C468E105A0CC25ULL, "6070892_f.fxe"},
	{0xD2C4D6F7507F1A78ULL, "6123656_f.fxe"},
	{0xD2C539CC16B770F8ULL, "lowtown_male_alienage1_sta6161634.cl"},
	{0xD2C56BF64618AACDULL, "sse_ss_danarius.gda"},
	{0xD2C5ACC51C118774ULL, "hlac_w_bonepitdestroyed_a3_0_2z.dds"},
	{0xD2C5B9BBEFDE7ECCULL, "ts_dalish_03.gda"},
	{0xD2C5BD081B371983ULL, "vxht_keep_n_3b8_0_0.vlm"},
	{0xD2C5E8E555522D55ULL, "vxlt_gallowstemplar_f_5ch_0_0.vlm"},
	{0xD2C629AD6E3BB299ULL, "mcr371_hostile_fenris6082218.cl"},
	{0xD2C64477DCC1D82EULL, "132027115.ogg"},
	{0xD2C6687D45182887ULL, "mag320tr_magetemplar.ncs"},
	{0xD2C6782264C457D1ULL, "316858293.ogg"},
	{0xD2C6E9F3CC08C044ULL, "vxw_encounter2_camp_aw_0_0.vlm"},
	{0xD2C751174FBFE125ULL, "w_sundermount_79_0_lc.mmh"},
	{0xD2C7AB5B0DBBA42AULL, "w_bonepitdestroyed_a3_5_0.mao"},
	{0xD2C7AF5B0DBBAB79ULL, "w_bonepitdestroyed_a3_5_0.mmh"},
	{0xD2C7B95B0DBBBC67ULL, "w_bonepitdestroyed_a3_5_0.msh"},
	{0xD2C7E175CAA77744ULL, "vxlt_backalley_g_1n1_0_2.vlm"},
	{0xD2C89E5A8D335DEBULL, "snow_01.dds"},
	{0xD2C8CB5D943E6419ULL, "900452973.ogg"},
	{0xD2C90A4D97F9F3B6ULL, "513879964.ogg"},
	{0xD2C966464CE9F2CFULL, "6142197_f.fxe"},
	{0xD2C98D116DCCBE47ULL, "framebuffereffects_composedof_p0.vsh.fxo.dx11"},
	{0xD2C9C5A7155C0D2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1lrg.msh"},
	{0xD2C9D3A7155C2560ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1lrg.mao"},
	{0xD2C9D4E1AE4414EAULL, "gen_im_fla_carver_01.uti"},
	{0xD2C9D7A7155C2BB3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning1lrg.mmh"},
	{0xD2C9EE71EF1EB5EDULL, "vxht_mansion_n_6lm_0_0.vlm"},
	{0xD2CA8652DF9272EDULL, "894170280.ogg"},
	{0xD2CB1EF6EC135FAEULL, "228040110.ogg"},
	{0xD2CB56CEA63396E8ULL, "vxlt_warehouse_18b_0_0.vlm"},
	{0xD2CB8D0CA93EFCC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v3.ani"},
	{0xD2CB9EE11EB4208DULL, "533514967.ogg"},
	{0xD2CBA4AC34885C4FULL, "llac_lt_0_lowtown_docks_n_0_6.dds"},
	{0xD2CC029F010220E7ULL, "hlac_lt_gallowscourtyard_n_65536_0y.dds"},
	{0xD2CC7F5CB9C258CBULL, "hlac_w_woundedcoast_262150_0z.dds"},
	{0xD2CD0B4FDB49E28BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_reaction.evt"},
	{0xD2CD197B0DEB76D4ULL, "w_blightlands_136_0.phy"},
	{0xD2CDA06ED44BC850ULL, "6101326_f.fxe"},
	{0xD2CDFE6B25B40E9BULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robb_0d.dds"},
	{0xD2CE26C70DC70A57ULL, "827742795.ogg"},
	{0xD2CE72137D7ECFD5ULL, "hlac_lt_house_-1_1z.dds"},
	{0xD2CE78A6DDA69787ULL, "one100dg_nabil6033099.cl"},
	{0xD2CF025A1EBC1980ULL, "vxw_blightlands_vu_0_0.vlm"},
	{0xD2CF066AC7B280C4ULL, "6202101_m.fxe"},
	{0xD2CF28B02F5A9499ULL, "6149122_m.fxe"},
	{0xD2CF643C2E7CE8CBULL, "hlac_lt_0_lowtown_docks_-196608_1y.dds"},
	{0xD2D011E00692CC27ULL, "213643424.ogg"},
	{0xD2D06A607F72C698ULL, "llac_lt_0_lowtown_a3_n_c_-65536_7.dds"},
	{0xD2D0B7180E7FBA73ULL, "mag170_hugh6080751.cl"},
	{0xD2D0F664EB380B6AULL, "1036368775.ogg"},
	{0xD2D172395AA17F34ULL, "hlac_w_encounter2_131074_0x.dds"},
	{0xD2D2287AD6BA134CULL, "vxht_playershouse_n_1rd_0_0.vlm"},
	{0xD2D2896C21FBA6BDULL, "926819406.ogg"},
	{0xD2D28A0CA9451626ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v3.gad"},
	{0xD2D2A3D211226647ULL, "6112115_m.fxe"},
	{0xD2D2A67BB8DCA6B3ULL, "vxca_caverns1_gir_0_0.vlm"},
	{0xD2D32737CBCBC5F8ULL, "vxw_encounter2_camp_2w_0_1.vlm"},
	{0xD2D338893976D87FULL, "75586325.ogg"},
	{0xD2D372B934C5864DULL, "741089228.ogg"},
	{0xD2D421390D9532F5ULL, "hlac_lt_hangedman_-1_4x.dds"},
	{0xD2D4244FDB502C4CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_reaction.gad"},
	{0xD2D43453B4753AA9ULL, "247614461.ogg"},
	{0xD2D5335723911305ULL, "6116856_m.fxe"},
	{0xD2D60D573DA98CA4ULL, "21019735.ogg"},
	{0xD2D637DABE084424ULL, "art\\characters\\playercharacter\\elfmale\\em_har_orsa.msh"},
	{0xD2D64DDABE0869C6ULL, "art\\characters\\playercharacter\\elfmale\\em_har_orsa.mmh"},
	{0xD2D67D76CA99E4A1ULL, "vxw_bonepitmining_a2_272_0_0.vlm"},
	{0xD2D6E99DA84D8FC7ULL, "zz_personality_debug6064398.cl"},
	{0xD2D7386A4E2D2089ULL, "vxdr_primevaltomb_5a7_0_0.vlm"},
	{0xD2D762050AA2CE63ULL, "llac_lt_0_lowtown_docks_a3_n_-65535_0.dds"},
	{0xD2D76AE447E3CC90ULL, "mag320_orsino6071216.cl"},
	{0xD2D76FDD72812D0BULL, "7995261.ogg"},
	{0xD2D7F67B0DF542D1ULL, "w_blightlands_136_0.msh"},
	{0xD2D8047B0DF55AECULL, "w_blightlands_136_0.mao"},
	{0xD2D8107B0DF56E8FULL, "w_blightlands_136_0.mmh"},
	{0xD2D82E49136D685AULL, "6110810_f.fxe"},
	{0xD2D87433C70DE5A2ULL, "w_bonepitmining_a2_84_0.msh"},
	{0xD2D87633C70DE958ULL, "w_bonepitmining_a2_84_0.mmh"},
	{0xD2D87E2E3DABA1D1ULL, "ser221st_gascard.stg"},
	{0xD2D88233C70DFDFBULL, "w_bonepitmining_a2_84_0.mao"},
	{0xD2D8DEC5AC8D9904ULL, "117597947.ogg"},
	{0xD2D8E2625F82B9ECULL, "380637630.ogg"},
	{0xD2D96A54E8914D21ULL, "vxw_bonepitmining_a2_1vj_0_0.vlm"},
	{0xD2D9850CA94B2C05ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_v3.evt"},
	{0xD2D9919275857CCFULL, "hlac_lt_0_lowtown_a3_n_c_-262144_2z.dds"},
	{0xD2DA1BA9FD6A463AULL, "art\\vfx\\combat\\vfx_glb_spi_imp_min_c.impa.gad"},
	{0xD2DA4A1D7B7FCE46ULL, "dae000_osric6162707.cl"},
	{0xD2DB234FDB56494AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_reaction.ani"},
	{0xD2DB5C1A5C003090ULL, "vxca_caverns2prc_aeq_0_0.vlm"},
	{0xD2DBC8BF30BA8BDCULL, "vis121_saemus6101337.cl"},
	{0xD2DC3CC734C6E49FULL, "790027394.ogg"},
	{0xD2DCC230158EE38FULL, "gen00fl_isabela6100042.cl"},
	{0xD2DCF056CA75FDDFULL, "vxw_woundedcoast_ft_0_0.vlm"},
	{0xD2DD6C34B6ABEDBFULL, "lt_0_lowtown_docks_26_0.rml"},
	{0xD2DDF6283CFB1F1AULL, "dae000_orana6149011.cl"},
	{0xD2DE9E6F03FDD98DULL, "6103757_m.fxe"},
	{0xD2DECC3552937BADULL, "hlac_ht_chantry_2534_2x.dds"},
	{0xD2DFCD5468C6B00CULL, "vxdr_primevaltomb_6ma_0_0.vlm"},
	{0xD2DFE85B73D0E0F5ULL, "zz_vrc_debug6200177.cl"},
	{0xD2DFF8ECE4898CBBULL, "w_encounter2_camp_29_0.msh"},
	{0xD2E00AECE489AB06ULL, "w_encounter2_camp_29_0.mao"},
	{0xD2E016ECE489BFA5ULL, "w_encounter2_camp_29_0.mmh"},
	{0xD2E043CF96B1FB0CULL, "602703358.ogg"},
	{0xD2E04EFD6504125BULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright30.evt"},
	{0xD2E0520DDC6C74F5ULL, "vxlt_darktown_4fq_0_0.vlm"},
	{0xD2E05712355745BDULL, "vrc321_varric6154756.cl"},
	{0xD2E068471DBF5380ULL, "6067300_m.fxe"},
	{0xD2E082E2ECB8E1A4ULL, "vxca_caverns2_96a_0_0.vlm"},
	{0xD2E0F66AA775BB3DULL, "hlac_dr_slavers_3216_12z.dds"},
	{0xD2E10417D98B3773ULL, "vxlt_darktown_5j6_0_0.vlm"},
	{0xD2E11EA9FD706A51ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_min_c.impa.evt"},
	{0xD2E1405AAFF69AECULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_exit.ani"},
	{0xD2E144DABE126191ULL, "art\\characters\\playercharacter\\elfmale\\em_har_orsa.phy"},
	{0xD2E1B202090AE401ULL, "6129106_m.fxe"},
	{0xD2E1FAD334494DCDULL, "6208573_m.fxe"},
	{0xD2E20CDCF585A4CCULL, "463787401.ogg"},
	{0xD2E2CC9F393E5E55ULL, "vxlt_gallowstemplar_f_5iv_0_0.vlm"},
	{0xD2E2DAFCC3132995ULL, "6168128_f.fxe"},
	{0xD2E2FD46E632AB56ULL, "6143439_m.fxe"},
	{0xD2E31734A0759710ULL, "vxlt_0_lowtown_a2_n_c_hnj_0_0.vlm"},
	{0xD2E35905434348BFULL, "vxw_bonepitmining_a2_p5_0_0.vlm"},
	{0xD2E3DDFAA7F28A8FULL, "892853006.ogg"},
	{0xD2E3F6F814B16FECULL, "vxw_bonepitmining_a2_a9_0_0.vlm"},
	{0xD2E40A5D2FF0CD92ULL, "qcr220_grey_wardens6135921.cl"},
	{0xD2E40F381F92CDB0ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_spi_c.mmh"},
	{0xD2E4335F21C2C2DEULL, "lt_0_lowtown_docks_a3_18_0_lc.mmh"},
	{0xD2E502A91B655B04ULL, "227643008.ogg"},
	{0xD2E558733D5322FEULL, "llac_ca_caverns1_mines_24423_12.dds"},
	{0xD2E569582FB39DA8ULL, "vxlt_warehouse_n_q1_0_0.vlm"},
	{0xD2E59BAEED282B0BULL, "hlac_w_encounter2_65539_0y.dds"},
	{0xD2E5AD9484FA94D7ULL, "dae000ip_crowd_ff2_talk11.utp"},
	{0xD2E5CEF373ACB655ULL, "6027961_m.fxe"},
	{0xD2E616FA234A6106ULL, "vxca_caverns2_8a3_0_0.vlm"},
	{0xD2E64E21ABE3F1FEULL, "489282898.ogg"},
	{0xD2E65D48EF0BD179ULL, "390200810.ogg"},
	{0xD2E67FB9722FDA70ULL, "llac_ca_caverns2prc_5802_31.dds"},
	{0xD2E70AB0C048BDBBULL, "dae000ip_crowd_ff2_talk6.utp"},
	{0xD2E714B52DFC7A26ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.cess.ani"},
	{0xD2E72375F2C8D849ULL, "llac_lt_0_lowtown_a3_n_c_-65537_3.dds"},
	{0xD2E747FD650A249CULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright30.gad"},
	{0xD2E74D5AAFFB1C2EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_exit.gad"},
	{0xD2E7700AD6CAD985ULL, "vxlt_gallowstemplar_f_4ug_0_0.vlm"},
	{0xD2E787242A141556ULL, "art\\vfx\\creatures\\fxc_drg_bfft_p.impa.ani"},
	{0xD2E78898E5CDC535ULL, "6160884_m.fxe"},
	{0xD2E7B75A8D4E38CDULL, "snow_01.mao"},
	{0xD2E868FFB7B0B3E8ULL, "lt_warehouse_1_0.mmh"},
	{0xD2E86C28B9DFD1CBULL, "hlac_lt_gallowscourtyard_n_-131072_3z.dds"},
	{0xD2E874FFB7B0C78BULL, "lt_warehouse_1_0.mao"},
	{0xD2E886FFB7B0E612ULL, "lt_warehouse_1_0.msh"},
	{0xD2E8DC60163AD41FULL, "837366130.ogg"},
	{0xD2E988D97D7458CDULL, "163335242.ogg"},
	{0xD2E9A3E4F98A339BULL, "6151338_m.fxe"},
	{0xD2E9F1ACE98B0749ULL, "dae000_mother6121061.cl"},
	{0xD2EA4DD0914174F8ULL, "678630327.ogg"},
	{0xD2EA79E32A2EE707ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga105a_0t.dds"},
	{0xD2EAAEE1D83752B8ULL, "6153595_m.fxe"},
	{0xD2EAB2B09CD2DC09ULL, "6169458_m.fxe"},
	{0xD2EAC23D8D2782A1ULL, "cod_hst_kirk_arch_1a_2.utp"},
	{0xD2EB6233BC6C165FULL, "vxw_woundedcoast_cw_0_0.vlm"},
	{0xD2EB70C0AAA23144ULL, "839240060.ogg"},
	{0xD2EB853FBD86043AULL, "vxht_0_hightown_a2_cao_0_0.vlm"},
	{0xD2EBBC6961C242B2ULL, "6069160_m.fxe"},
	{0xD2EC3B54028B2258ULL, "vxht_0_hightown_n_aqp_0_0.vlm"},
	{0xD2EC68472798079BULL, "369694152.ogg"},
	{0xD2EC83B784B89857ULL, "vxlt_gallowscourtyard_n_6pb_0_0.vlm"},
	{0xD2ECEFE14A1EBD7DULL, "w_sundermount_92_0.mao"},
	{0xD2ECF3E14A1EC3B6ULL, "w_sundermount_92_0.mmh"},
	{0xD2ECFDE14A1ED4B4ULL, "w_sundermount_92_0.msh"},
	{0xD2ED2B60AEE1A5C6ULL, "gen_im_trs_apo_lgt_hlm_r2.uti"},
	{0xD2ED3499D76D66D7ULL, "zz_followers_banter6158844.cl"},
	{0xD2EDA6742F4F10B1ULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.anb"},
	{0xD2EDB15EE17EE522ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_basic.phy"},
	{0xD2EDC859655E205BULL, "vxht_0_hightown_a2_bfz_0_0.vlm"},
	{0xD2EDCEE026CF3E4DULL, "6076212_f.fxe"},
	{0xD2EE198D92A14E2AULL, "vxlt_gallowstemplar_4sm_0_0.vlm"},
	{0xD2EE36A9FD7B1D20ULL, "art\\vfx\\combat\\vfx_glb_spi_imp_min_c.impa.ani"},
	{0xD2EE46FD6510419AULL, "art\\characters\\creatures\\highdragon\\animation\\exploration\\turnright30.ani"},
	{0xD2EE485AB001320DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_throne_exit.evt"},
	{0xD2EE4CBC40DFF6D0ULL, "946246458.ogg"},
	{0xD2EEAAE80F0A5AE5ULL, "1016428490.ogg"},
	{0xD2EECEAE8F014832ULL, "vxlt_0_lowtown_f49_0_0.vlm"},
	{0xD2EF14C7E71FF853ULL, "gen00fl_anders6119424.cl"},
	{0xD2EF1D8B7B283355ULL, "406792113.ogg"},
	{0xD2EF75F4CB814FF3ULL, "vxlt_darktown_5o9_0_0.vlm"},
	{0xD2EFFB6B47568AABULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinswall.phy"},
	{0xD2F06FF481C18582ULL, "one220_yevhen6135852.cl"},
	{0xD2F07E426E024B7BULL, "577296661.ogg"},
	{0xD2F0CF81B9462C3FULL, "ts_dalish_acc_table.gda"},
	{0xD2F132381F9D93CFULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_spi_c.anb"},
	{0xD2F13E7C2B3476D5ULL, "vxlt_gallowstemplar_f_5lu_0_0.vlm"},
	{0xD2F2577E9DBCFC1BULL, "qun221_viscount6070104.cl"},
	{0xD2F26061CA066A3AULL, "vxca_cavern1_haunted_cmv_0_0.vlm"},
	{0xD2F2A3CCC891AF7CULL, "gen00fl_fenris6126384.cl"},
	{0xD2F2C66F0D9D0F09ULL, "mcr310_argument6075809.cl"},
	{0xD2F39057826F8526ULL, "495181080.ogg"},
	{0xD2F398A9002B3380ULL, "314279921.ogg"},
	{0xD2F3FCB52E06DB97ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.cess.evt"},
	{0xD2F429A09B58CBD5ULL, "6101878_m.fxe"},
	{0xD2F48F092966FBF9ULL, "43137621.ogg"},
	{0xD2F48F242A1EAD47ULL, "art\\vfx\\creatures\\fxc_drg_bfft_p.impa.evt"},
	{0xD2F566125F400979ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrubble_da.dds"},
	{0xD2F5E1AE58DA5377ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_gesture_head.gad"},
	{0xD2F62896242AD8B7ULL, "6121507_f.fxe"},
	{0xD2F6289A65D5CD85ULL, "cs_pro000cs_openingrogue.bnk"},
	{0xD2F722D5C2C8D1C8ULL, "gen00fl_fenris6107169.cl"},
	{0xD2F73065B4277559ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw01b.phy"},
	{0xD2F769D6CBE0AFE2ULL, "vxca_cavern1_haunted_hii_0_0.vlm"},
	{0xD2F777DF2C0EAD64ULL, "vxlt_0_lowtown_docks_btz_0_0.vlm"},
	{0xD2F7A08A30B47D23ULL, "lt_0_lowtown_docks_a3_185_0.msh"},
	{0xD2F7AE8A30B4957DULL, "lt_0_lowtown_docks_a3_185_0.mmh"},
	{0xD2F7B28A30B49B8EULL, "lt_0_lowtown_docks_a3_185_0.mao"},
	{0xD2F8EC78EFF4B989ULL, "6173592_m.fxe"},
	{0xD2F916FB9C5CBCB2ULL, "gen00fl_anders6112901.cl"},
	{0xD2F9643393046394ULL, "dae011_conspiracy_mob6149588.cl"},
	{0xD2F9D410AE622EDFULL, "vxw_woundedcoast_xv_0_0.vlm"},
	{0xD2F9EC724FCD76CBULL, "498640015.ogg"},
	{0xD2FA304043B36464ULL, "vxlt_gallowscourtyard_7fq_0_0.vlm"},
	{0xD2FA93DF300D72A8ULL, "vxht_chantry_10p_0_0.vlm"},
	{0xD2FA97EF5E3548CAULL, "6131878_m.fxe"},
	{0xD2FAB3E0A63A15F0ULL, "6105692_f.fxe"},
	{0xD2FAC7742F59D32AULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.mmh"},
	{0xD2FAFDB52E0CFC20ULL, "art\\vfx\\combat\\vfx_m_bld_mgc_lp_c.cess.gad"},
	{0xD2FB275363EDC78AULL, "620149730.ogg"},
	{0xD2FBB0242A250390ULL, "art\\vfx\\creatures\\fxc_drg_bfft_p.impa.gad"},
	{0xD2FBEC3511FD3C19ULL, "gen_im_npc_elf_arm_hlm_dald.uti"},
	{0xD2FC11F6F924A9E5ULL, "vxw_bonepitmining_a2_24k_0_0.vlm"},
	{0xD2FC518BBC9B7370ULL, "llac_lt_0_lowtown_docks_n_-131073_1.dds"},
	{0xD2FCB10E4CB2E10CULL, "vxdr_primevaltomb_6wg_0_0.vlm"},
	{0xD2FCF22683944BBBULL, "framebuffereffects_twopassblurshadows_p0.psh.fxo.dx11.assembly"},
	{0xD2FDB4D41DF8CAD8ULL, "qun211_party_comment6072268.cl"},
	{0xD2FDE50643F6FE20ULL, "character_character_p131.psh.fxo.dx11"},
	{0xD2FDE7D1BD776873ULL, "vxlt_darktown_5p8_0_0.vlm"},
	{0xD2FEA4CFE003357AULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube_tool.msh"},
	{0xD2FEB2CFE0034CB3ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube_tool.mao"},
	{0xD2FEB6CFE00353C0ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_blast120_tube_tool.mmh"},
	{0xD2FED2B7435480C7ULL, "vxht_mansion_n_62q_0_0.vlm"},
	{0xD2FF193D811C0CC9ULL, "hlac_w_encounter2_camp_131077_0z.dds"},
	{0xD2FF1D5958272868ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexa.msh"},
	{0xD2FF2044F61F1B4BULL, "mer210_smith6051394.cl"},
	{0xD2FF2F59582746F9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexa.mao"},
	{0xD2FF3B5958275A92ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexa.mmh"},
	{0xD2FF433E0B96F9F5ULL, "hlac_ht_mansion2_818_52x.dds"},
	{0xD2FFCDC6D323B1B0ULL, "tbe100_mirror6200107.cl"},
	{0xD30041FC233E8528ULL, "gen00fl_varric6105743.cl"},
	{0xD300520EB10C2709ULL, "6129044_f.fxe"},
	{0xD3009D77BB086B3BULL, "vxca_caverns1_cmr_0_0.vlm"},
	{0xD300A6962BA63D28ULL, "6170149_m.fxe"},
	{0xD301666ED4A0A57CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\custom\\mw_slow_zone.ani"},
	{0xD301B0A0F675AAF0ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_p.mmh"},
	{0xD301BA5689C153FCULL, "6136217_m.fxe"},
	{0xD30225BB9D188C54ULL, "vxlt_undercityserialkiller_1in_0_0.vlm"},
	{0xD302C45D0E20F338ULL, "vxht_0_hightown_bdw_0_0.vlm"},
	{0xD30340A65B8B4FAFULL, "6095958_f.fxe"},
	{0xD3036B31AF391BECULL, "ser221_gascard6175529.cl"},
	{0xD3036F1137D6E05CULL, "dae000_hate_follower_react6103233.cl"},
	{0xD30378D7AF34DBB0ULL, "6148458_m.fxe"},
	{0xD303848A5DAE3F3FULL, "6117418_f.fxe"},
	{0xD303BE8E9D7F111CULL, "vxht_0_hightown_a2_c46_0_0.vlm"},
	{0xD303C8353D4BF7B5ULL, "229328494.ogg"},
	{0xD303F9637E9D6599ULL, "and100_anders6095917.cl"},
	{0xD304260A18D60F61ULL, "dre000_dougal6131125.cl"},
	{0xD30485D25026F75DULL, "94546485.ogg"},
	{0xD304BD252B018F70ULL, "atl_prccontent_dxt5_01.dds"},
	{0xD304F2857862C6A6ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_p.impa.ani"},
	{0xD304F420235BF377ULL, "mind_03_cycle.mao"},
	{0xD3051D824E3C1A08ULL, "one300_zevran6089720.cl"},
	{0xD3062439B2BBFF1EULL, "43010781.ogg"},
	{0xD306ABA6126878C6ULL, "vxht_0_hightown_c1g_0_0.vlm"},
	{0xD306DE1FDADBFA77ULL, "98777126.ogg"},
	{0xD307966A80BF0D09ULL, "6163658_m.fxe"},
	{0xD307F7C0192F4F3AULL, "6144841_m.fxe"},
	{0xD30815788DE993F9ULL, "639213492.ogg"},
	{0xD308177FB1F25665ULL, "hlac_w_woundedcoast_65544_0z.dds"},
	{0xD3083D84A19CBDD1ULL, "76454878.ogg"},
	{0xD30845EDA058475FULL, "vxw_woundedcoast_uq_0_0.vlm"},
	{0xD3085533D973B439ULL, "vxht_playershouse_n_1il_0_0.vlm"},
	{0xD308736ED4A6D9BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\custom\\mw_slow_zone.gad"},
	{0xD308F22C41F95042ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.mmh"},
	{0xD3093BF0F55E7BB7ULL, "614534266.ogg"},
	{0xD309DCD752765316ULL, "dae000_madam_lusine.cnv"},
	{0xD309FD6247BA550AULL, "hlac_lt_0_lowtown_n_65536_5z.dds"},
	{0xD30A0EAE58EB48FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_gesture_head.ani"},
	{0xD30A95C2C53C38ACULL, "27335947.ogg"},
	{0xD30B56FEA3AAC4E7ULL, "gen_im_npc_wep_dua_gend.uti"},
	{0xD30BB61260FB749CULL, "zz_seb_debug6183191.cl"},
	{0xD30BC4B7E8B28254ULL, "6071888_m.fxe"},
	{0xD30BDECF3C64EC5DULL, "250112394.ogg"},
	{0xD30C7EAF633C6AC6ULL, "572148599.ogg"},
	{0xD30C90A845F6D77FULL, "fxa_pois_drop.mao"},
	{0xD30D4D10CEE9E1E0ULL, "fen210_dying_magister6108578.cl"},
	{0xD30DDED7527A3FF6ULL, "dae000_madam_lusine.bnk"},
	{0xD30E036413BCCA68ULL, "711633531.ogg"},
	{0xD30E63CE77993668ULL, "33546755.ogg"},
	{0xD30EAE9C1925923FULL, "vxw_bonepitmining_a2_a6_0_0.vlm"},
	{0xD30F15DA0DA2296DULL, "hlac_lt_0_lowtown_a2_n_-65535_7y.dds"},
	{0xD30F17FB79947BEEULL, "6034279-coolestfollowertest.cut"},
	{0xD30F1D5C4FA0BEEAULL, "hlac_ht_keep_n_3743_1z.dds"},
	{0xD30F3DAC392BE517ULL, "t3_wdg_w19.tnt"},
	{0xD30F42B081595413ULL, "90464440.ogg"},
	{0xD30F487515E3786BULL, "144465403.ogg"},
	{0xD30F4E6ED4ACB9DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\custom\\mw_slow_zone.evt"},
	{0xD30F5987181840F1ULL, "gen_im_trs_rdr_junk1_03.uti"},
	{0xD30F88CDFBA9A5EBULL, "604079881.ogg"},
	{0xD30FA0E0CAE92F81ULL, "vxw_encounter2_5r_0_0.vlm"},
	{0xD30FD3A0F682230FULL, "art\\vfx\\combat\\vfx_m_frc_repwave_p.anb"},
	{0xD30FF30229E13632ULL, "gencr_corpse_onground.utc"},
	{0xD30FF63C43B302D3ULL, "586047139.ogg"},
	{0xD31027EBD70C576EULL, "316484985.ogg"},
	{0xD31060DA04FEBAFAULL, "303334196.ogg"},
	{0xD310B0B42F751B19ULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_p.mmh"},
	{0xD310ECCCD593C4D2ULL, "llac_lt_0_lowtown_a2_n_c_1_3.dds"},
	{0xD311A7337EC698A3ULL, "llac_lt_0_lowtown_n_131070_1.dds"},
	{0xD311B12A15C3C0FAULL, "6161439_m.fxe"},
	{0xD311CB1916D905BFULL, "964459114.ogg"},
	{0xD311DA85786D2817ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_p.impa.evt"},
	{0xD312306B8F75299CULL, "vxht_0_hightown_a2_c37_0_0.vlm"},
	{0xD312372329FDC7F8ULL, "6060433_m.fxe"},
	{0xD3130A6CDEB183E6ULL, "dae000_arishok6067696.cl"},
	{0xD313437302937FEDULL, "cd_mrl150_merrill.wbd"},
	{0xD31376B23DBBB94BULL, "6148857_m.fxe"},
	{0xD3138C5EE19F75D6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_basic.mao"},
	{0xD31390E2DD782FC7ULL, "6054288_m.fxe"},
	{0xD313985EE19F8A75ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_basic.mmh"},
	{0xD3139A5EE19F8DEBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0wi_basic.msh"},
	{0xD3139AD612E78840ULL, "zz_followers_banter6157794.cl"},
	{0xD313E75F3E6B6CBEULL, "vxlt_warehouse_bg_0_0.vlm"},
	{0xD314AC9D34E50658ULL, "463995585.ogg"},
	{0xD31505450AE10AB1ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.dura.ani"},
	{0xD3151D83045E9146ULL, "vxht_0_hightown_c4h_0_0.vlm"},
	{0xD315397C2541E662ULL, "seb201_lady_harimann6177583.cl"},
	{0xD31555DF063ED756ULL, "vxht_0_hightown_a2_c6f_0_0.vlm"},
	{0xD3155ABE3219C35EULL, "6131352_f.fxe"},
	{0xD315895467E0EBF5ULL, "14862817.ogg"},
	{0xD31649BB1FE11822ULL, "dr_slavers_3214.rml"},
	{0xD316C785BADEE9D2ULL, "6115414_m.fxe"},
	{0xD31791DD05AC749EULL, "359093920.ogg"},
	{0xD317F87DB3DB4B9CULL, "388462151.ogg"},
	{0xD3184A03A400108FULL, "vxw_bonepitdestroyed_a3_bu_0_0.vlm"},
	{0xD31857FD4D905121ULL, "6053311_m.fxe"},
	{0xD318B16267E4D2A3ULL, "w_bonepit_122_0.msh"},
	{0xD318BF6267E4EAFDULL, "w_bonepit_122_0.mmh"},
	{0xD318C36267E4F10EULL, "w_bonepit_122_0.mao"},
	{0xD318DB85787348A0ULL, "art\\vfx\\cinematic\\vfx_lyrium_c_p.impa.gad"},
	{0xD3190A1E592433A0ULL, "6167087_f.fxe"},
	{0xD319861C83EC2EB1ULL, "241631039.ogg"},
	{0xD3199734E6BB3698ULL, "6149282_m.fxe"},
	{0xD319AD3C2D4A4914ULL, "6127255_m.fxe"},
	{0xD319F51A6790C326ULL, "hlac_lt_0_lowtown_1_6x.dds"},
	{0xD31A4F1D035AA28CULL, "hlac_lt_undercitydungeon_-1_43x.dds"},
	{0xD31A6C852266F941ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls101a.phy"},
	{0xD31ADD8D236D58BFULL, "993955119.ogg"},
	{0xD31B557D456048DEULL, "mag320tr_magetemplar.utt"},
	{0xD31BB4DB84D17E86ULL, "60509363.ogg"},
	{0xD31C10450AE73C53ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.dura.gad"},
	{0xD31C1BB8115E2478ULL, "lgt320_mis_resp_swordhunter6150328.cl"},
	{0xD31C2C465B812A9AULL, "art\\vfx\\creatures\\fxc_bm_spit_p.impa.gad"},
	{0xD31C456CE4197C1FULL, "6092589_m.fxe"},
	{0xD31D20790B1BAABFULL, "vxw_bonepitmining_a2_l1_0_0.vlm"},
	{0xD31D6C1201292372ULL, "hlac_w_encounter2_camp_65541_0y.dds"},
	{0xD31D8F34B6E20D23ULL, "lt_0_lowtown_docks_26_0.anb"},
	{0xD31DA3EBA90F4A02ULL, "vxlt_0_lowtown_a2_n_c_gx9_0_0.vlm"},
	{0xD31E72A43509D8A2ULL, "lowtown_male_beggar_sta6151083.cl"},
	{0xD31ECA2C5B5A9546ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft.evt"},
	{0xD31EE35B1ECC081DULL, "504976749.ogg"},
	{0xD31EFDB4ED8E1955ULL, "99055564.ogg"},
	{0xD31F5493683BD01AULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_cbt_lari.mmh"},
	{0xD31F6251C14E518BULL, "vxw_woundedcoast_1e0_0_0.vlm"},
	{0xD31F6693683BEF60ULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_cbt_lari.msh"},
	{0xD32016ECE4BFA3AEULL, "w_encounter2_camp_29_0.rml"},
	{0xD320203B0600C62BULL, "278514349.ogg"},
	{0xD320AE3CD12D6916ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_isaa_0n.dds"},
	{0xD320F38026D6CEA3ULL, "qun221_petrice6083758.cl"},
	{0xD3211ABCCF5F4A74ULL, "6065759_m.fxe"},
	{0xD3211D7D311D1569ULL, "6106624_m.fxe"},
	{0xD321AF1264592F3BULL, "hightown_male_noble_gossip6174057.cl"},
	{0xD321C6D533C11991ULL, "6118436_m.fxe"},
	{0xD3220D450AEBA2B8ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.dura.evt"},
	{0xD3228D8ED55CFFFEULL, "dae000_templar_knight6165169.cl"},
	{0xD322CD7601F0F6E4ULL, "vxlt_0_lowtown_docks_bey_0_0.vlm"},
	{0xD322FB47AF6215E2ULL, "926824225.ogg"},
	{0xD323029580E838F6ULL, "vxlt_0_lowtown_a3_n_c_byn_0_0.vlm"},
	{0xD3232F465B874EB1ULL, "art\\vfx\\creatures\\fxc_bm_spit_p.impa.evt"},
	{0xD32336DCB421B346ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_aurora.mao"},
	{0xD32378E368943D6AULL, "lgt000ip_who_she_torn_page.utp"},
	{0xD323979B4BEEE207ULL, "847155676.ogg"},
	{0xD323AF5B0E0A09D2ULL, "w_bonepitdestroyed_a3_5_0.rml"},
	{0xD323BDE570554390ULL, "6082594_m.fxe"},
	{0xD3245EE036F77E88ULL, "lgt000cr_mage_turn_in_3.utc"},
	{0xD3245F8F94E01CB3ULL, "vxlt_darktown_2qx_0_0.vlm"},
	{0xD3245FFE616AC50EULL, "vxlt_warehouse_15c_0_0.vlm"},
	{0xD32493661DBE2470ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottlewine.mao"},
	{0xD32497661DBE2A83ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottlewine.mmh"},
	{0xD324A5661DBE42DDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bottlewine.msh"},
	{0xD3250D390F4A72D0ULL, "vxht_mansion2_1ao_0_0.vlm"},
	{0xD3254100EEC1DEB3ULL, "6061799_f.fxe"},
	{0xD3257ABFC002BA7EULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_drka_0t.dds"},
	{0xD3257D0E52DB8C0DULL, "hlac_lt_0_lowtown_a2_-2_4y.dds"},
	{0xD325E12C5B60DA91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft.gad"},
	{0xD3269D7161A7D5BEULL, "vxw_encounter3_bs_0_0.vlm"},
	{0xD326BBE095F6290FULL, "vxw_bonepitdestroyed_a3_mt_0_0.vlm"},
	{0xD326EFECE4C58052ULL, "w_encounter2_camp_29_0.phy"},
	{0xD326F966824C094BULL, "dae000ip_crowd_mf7_talk13.utp"},
	{0xD32708FFC3CF07D1ULL, "6013449_m.fxe"},
	{0xD327CAE44430A54AULL, "vxlt_backalley_g_1pm_0_2.vlm"},
	{0xD327CD46DBE57EBAULL, "art\\characters\\playercharacter\\humanmale\\glv_masb.phy"},
	{0xD327FB270459AED3ULL, "345777851.ogg"},
	{0xD327FE72D4A65FE2ULL, "vxca_caverns1_ck5_0_0.vlm"},
	{0xD3280911D8B14F18ULL, "6142382_m.fxe"},
	{0xD3287D4C0C117E73ULL, "555742564.ogg"},
	{0xD328CAC812BC274CULL, "659610267.ogg"},
	{0xD329BA806035AA25ULL, "238611473.ogg"},
	{0xD32ABCB1F06E6457ULL, "seb321_leliana6181749.cl"},
	{0xD32ADA65F3762D6EULL, "gen00fl_fenris6103350.cl"},
	{0xD32B50855A441190ULL, "vxlt_0_lowtown_a2_n_c_huw_0_0.vlm"},
	{0xD32B9255FD11C33FULL, "vxw_bonepitmining_a2_k0_0_0.vlm"},
	{0xD32BBB6A367251EDULL, "120032995.ogg"},
	{0xD32BE98280AC7E13ULL, "llac_lt_darktown_-1_56.dds"},
	{0xD32C539E5AB53960ULL, "lt_gallowstemplar_f_2388.rml"},
	{0xD32CE22C5B66FB3FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_trn_lft.ani"},
	{0xD32D02E9FB5619BBULL, "vxw_encounter2_fu_0_0.vlm"},
	{0xD32D703D5DD59829ULL, "386915276.ogg"},
	{0xD32D7B5272FAD5D4ULL, "vxlt_undercityserialkiller_1xm_0_0.vlm"},
	{0xD32D8A6C0CD24166ULL, "308230185.ogg"},
	{0xD32D96D4BC0E9568ULL, "hlac_lt_backalley_g_65537_1y.dds"},
	{0xD32DA70CECE71F15ULL, "49121835.ogg"},
	{0xD32E001E9785AE9CULL, "80642635.ogg"},
	{0xD32E1192B8E6CA49ULL, "vrc321_party_barks6095755.cl"},
	{0xD32E4E75F8F1FB18ULL, "vxht_mansion2b_3g3_0_0.vlm"},
	{0xD32E6F0138DFD743ULL, "327533945.ogg"},
	{0xD32F2525EFC7A869ULL, "vxlt_0_lowtown_a2_n_c_hkp_0_0.vlm"},
	{0xD32F3EE05F0F8FEEULL, "gen00fl_varric6075767.cl"},
	{0xD32F5E3D66B0C9CEULL, "one300st_zevran.stg"},
	{0xD32FE716250EE3ECULL, "6218911_m.fxe"},
	{0xD3307BF5BBBBDC27ULL, "6118614_m.fxe"},
	{0xD330F671440589D5ULL, "vxca_cave_1bn_0_1.vlm"},
	{0xD33147465B93B480ULL, "art\\vfx\\creatures\\fxc_bm_spit_p.impa.ani"},
	{0xD3314D9E8836571FULL, "zz_mrl_debug6184192.cl"},
	{0xD331F12C421C53E9ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.anb"},
	{0xD3321EEB932FBDDFULL, "hlac_ca_cavern1_haunted_24426_4y.dds"},
	{0xD3323998EA2B0856ULL, "vxht_0_hightown_a2_c8h_0_0.vlm"},
	{0xD33276BA3DCD4FFCULL, "vxht_brothel_n_540_0_0.vlm"},
	{0xD332944E54BED30AULL, "vxca_cavern1_haunted_hm5_0_0.vlm"},
	{0xD333280F4AEFF4BEULL, "vxht_mansion2_30l_0_0.vlm"},
	{0xD33347BDA20C3801ULL, "6124750_m.fxe"},
	{0xD3337F1601408B50ULL, "vxht_mansion2_1dl_0_0.vlm"},
	{0xD333BCA5928D0BA5ULL, "gen00fl_merrill6107096.cl"},
	{0xD334B42E24AEBABAULL, "nat380_nathaniel6066412.cl"},
	{0xD334C683B147FAC3ULL, "llac_w_encounter3_-327683_0.dds"},
	{0xD334CF1FA3B63A00ULL, "gen00fl_fenris6108083.cl"},
	{0xD33525A051EFE839ULL, "577184368.ogg"},
	{0xD3352DBD87EC418FULL, "vxw_bonepitdestroyed_a3_hw_0_0.vlm"},
	{0xD3361A8E7FEFD758ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_sword.phy"},
	{0xD336704FC69C7862ULL, "vxca_caverns1_cn6_0_0.vlm"},
	{0xD3367D245419A643ULL, "149556357.ogg"},
	{0xD3369F8DF4D72411ULL, "183322640.ogg"},
	{0xD33772F8C19681DEULL, "653451396.ogg"},
	{0xD337A04EFEA50B5BULL, "hlac_ca_caverns1_mines_24423_8x.dds"},
	{0xD3380E51EFE41FF8ULL, "ser221_party_comments6166189.cl"},
	{0xD338171438BE714AULL, "art\\vfx\\cinematic\\vfx_anders_romance_hand_c.anb"},
	{0xD338350D2D24D455ULL, "crust_webs.dds"},
	{0xD3384600F8B95289ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_04a_0s.dds"},
	{0xD33895CF44AD6A6FULL, "gen_im_trs_cot_med_boo_r1.uti"},
	{0xD338A8CE42D86FEEULL, "727336440.ogg"},
	{0xD338E9EEAEA319B6ULL, "art\\vfx\\combat\\vfx_r_brd_shout_imp_p.mmh"},
	{0xD33917EA158A80F3ULL, "6166699_m.fxe"},
	{0xD339ADB42F981B0EULL, "art\\vfx\\combat\\vfx_r_gn1_confuse_aoe_p.anb"},
	{0xD339C2624C3A2A10ULL, "vxlt_0_lowtown_a2_n_c_hpt_0_0.vlm"},
	{0xD33A10A20C9F9E87ULL, "follower_banter6026246.cl"},
	{0xD33A26E758AE3613ULL, "193367981.ogg"},
	{0xD33A4A084BED42A1ULL, "hlac_lt_0_lowtown_docks_-65539_4x.dds"},
	{0xD33AED603B5E58A5ULL, "dae000_arishok6070882.cl"},
	{0xD33B49D3CB0FAD81ULL, "vxw_blightlands_hz_0_0.vlm"},
	{0xD33B59180C9419A2ULL, "ser221_gascard6128293.cl"},
	{0xD33B64BEA8F40BAEULL, "6052759_m.fxe"},
	{0xD33C460BA53A828BULL, "vxw_woundedcoast_1k2_0_0.vlm"},
	{0xD33C56B4C848D86FULL, "26645817.ogg"},
	{0xD33C8BD0D5F95538ULL, "vxht_0_hightown_bps_0_0.vlm"},
	{0xD33CBDF855287269ULL, "1059990404.ogg"},
	{0xD33CDC3A2264596FULL, "ca_caverns1_24426.rml"},
	{0xD33CEC70EE15BC51ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\tn90l.ani"},
	{0xD33CFB93617960CCULL, "gen_im_cns_pot_xp_01.uti"},
	{0xD33D1C8527DE857AULL, "vxca_caverns1_k2p_0_2.vlm"},
	{0xD33DB222C4554A75ULL, "573400741.ogg"},
	{0xD33E74049D4274A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_3.ani"},
	{0xD33E77E651719348ULL, "706301942.ogg"},
	{0xD33EB445AA087015ULL, "255492248.ogg"},
	{0xD33EB8DBB3B37A51ULL, "hlac_ca_caverns1_mines_24427_2x.dds"},
	{0xD33EE9DD1CECC07CULL, "189282314.ogg"},
	{0xD33F3B766756EE07ULL, "hlac_w_sundermount_65541_0y.dds"},
	{0xD3403E92E3E6DFEDULL, "6168185_m.fxe"},
	{0xD3409FBB003D6EF3ULL, "6089484_m.fxe"},
	{0xD341938BD47BE5ADULL, "dae000cr_radha.utc"},
	{0xD3425573F3CA75B7ULL, "llac_w_sundermount_-327683_0.dds"},
	{0xD342F770EE1A3AF3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\tn90l.gad"},
	{0xD3441BA98077FF42ULL, "vxlt_0_lowtown_a2_n_c_hyy_0_0.vlm"},
	{0xD3447393685B2125ULL, "art\\toolset\\characters\\playercharacter\\humanmale\\mt_cbt_lari.phy"},
	{0xD344E06E832CC1A7ULL, "vxdr_primevaltomb_8ob_0_0.vlm"},
	{0xD344E831A4D5ED7BULL, "6196825_m.fxe"},
	{0xD34564257A1904E8ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg102a.msh"},
	{0xD3457F049D48A643ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_3.gad"},
	{0xD34582257A193712ULL, "art\\characters\\weapons\\dagger\\w_dgr_dg102a.mmh"},
	{0xD345A27FAF945874ULL, "art\\levels\\races\\proxy\\props\\pressure\\pressure.phy"},
	{0xD346792A7D9E09E4ULL, "vxht_keep_n_4hz_0_0.vlm"},
	{0xD346DBF5A91B8763ULL, "vxlt_0_lowtown_a3_n_c_ch9_0_0.vlm"},
	{0xD3473E50FA68D1D0ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_5y.dds"},
	{0xD347774395C73661ULL, "hlac_w_woundedcoast_-262142_0z.dds"},
	{0xD3479004916DB223ULL, "mag110_arianni6016802.cl"},
	{0xD347EF43A5B35855ULL, "vxht_chantry_n_2t8_0_1.vlm"},
	{0xD347F1B526B10DCDULL, "98661920.ogg"},
	{0xD34840687D1F3C42ULL, "pro000_meeran6048808.cl"},
	{0xD3487E84D8778965ULL, "6001293_m.fxe"},
	{0xD348C85CDF7110FDULL, "vxlt_0_lowtown_a2_fhz_0_0.vlm"},
	{0xD348ED9C625CDF82ULL, "780826830.ogg"},
	{0xD3494CD002C804DFULL, "6208352_f.fxe"},
	{0xD34956796BE0182AULL, "mag100_karras6052923.cl"},
	{0xD3495A3C818706C0ULL, "vxht_mansion2b_2eh_0_0.vlm"},
	{0xD349BBB0BD05C601ULL, "vxw_blightlands_my_0_0.vlm"},
	{0xD349E6A3DF424ABBULL, "vxw_encounter2_lw_0_0.vlm"},
	{0xD34A30C844E42DF9ULL, "6105859_f.fxe"},
	{0xD34A6606FDA547D2ULL, "vxlt_undercityserialkiller_12e_0_0.vlm"},
	{0xD34A9C7D29CFC675ULL, "one200dg_solivitus6066476.cl"},
	{0xD34A9D8549207EBEULL, "vxht_0_hightown_n_bxz_0_0.vlm"},
	{0xD34AAEBBEC114D9CULL, "hlac_ht_chantry_2534_50z.dds"},
	{0xD34AC3C909CC7102ULL, "w_bonepit_8_0.msh"},
	{0xD34AC5C909CC74B8ULL, "w_bonepit_8_0.mmh"},
	{0xD34AD1C909CC895BULL, "w_bonepit_8_0.mao"},
	{0xD34AFD2085B1266CULL, "784775873.ogg"},
	{0xD34B7595A7059FACULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench1.ani"},
	{0xD34C0B0F43D4C74FULL, "gen00fl_isabela6098196.cl"},
	{0xD34C3216FFAD826FULL, "dae011_gossips6149787.cl"},
	{0xD34C3AC2C0EC81B0ULL, "vxw_woundedcoast_126_0_0.vlm"},
	{0xD34C424231E1AD7DULL, "vxht_keep_70k_0_0.vlm"},
	{0xD34C5C049D4E8928ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_small_3.evt"},
	{0xD34CD551B233EF94ULL, "character_character_p52.psh.fxo.dx11.assembly"},
	{0xD34D619465BD8AB0ULL, "qun110_petrice6018530.cl"},
	{0xD34DAC642F9A766EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb12a.msh"},
	{0xD34DC5C909CEAACBULL, "w_bonepit_8_0.rml"},
	{0xD34DC6642F9AA25CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb12a.mmh"},
	{0xD34E2FF222342AA7ULL, "vxlt_0_lowtown_gc0_0_0.vlm"},
	{0xD34E8BFD5F1540EBULL, "119030977.ogg"},
	{0xD34EA54AC1D573BEULL, "vxlt_gallowscourtyard_5yh_0_0.vlm"},
	{0xD34EE69949C4798BULL, "hlac_lt_0_lowtown_n_-3_5z.dds"},
	{0xD34F2CEE6777EFFEULL, "472563702.ogg"},
	{0xD34F46BB2010D81EULL, "dr_slavers_3214.anb"},
	{0xD34F572E2E46C1D3ULL, "vxlt_undercityserialkiller_17c_0_0.vlm"},
	{0xD34F74F696E56273ULL, "6147805_m.fxe"},
	{0xD34F7F7FAF9C7171ULL, "art\\levels\\races\\proxy\\props\\pressure\\pressure.msh"},
	{0xD34F8D7FAF9C888CULL, "art\\levels\\races\\proxy\\props\\pressure\\pressure.mao"},
	{0xD34F9616F1DB72B7ULL, "vxht_mansion2b_1al_0_0.vlm"},
	{0xD34F997FAF9C9D2FULL, "art\\levels\\races\\proxy\\props\\pressure\\pressure.mmh"},
	{0xD34FCB1BA1981751ULL, "6080221_m.fxe"},
	{0xD3504E6405DF48ECULL, "452296260.ogg"},
	{0xD3505AAA4D9702F8ULL, "163806068.ogg"},
	{0xD35147046CD45C87ULL, "hlac_lt_0_lowtown_a3_n_c_131072_0x.dds"},
	{0xD3518F68B2A8FCD5ULL, "vxht_bartrand_n_3vq_0_0.vlm"},
	{0xD351AFFAABAC5AABULL, "fx_flamefb01.dds"},
	{0xD35222262489126DULL, "vxht_brothel_n_60b_0_0.vlm"},
	{0xD3528295A70BD4EEULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench1.gad"},
	{0xD3528D86726E17C2ULL, "vxlt_0_lowtown_a2_n_c_htz_0_0.vlm"},
	{0xD352FD07C804F33CULL, "art\\characters\\weapons\\shields\\textures\\w_shd_002a_n.dds"},
	{0xD3534D7302C97417ULL, "cd_mrl150_merrill.bnk"},
	{0xD35383CE0E06C143ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_3z.dds"},
	{0xD3550D6B8D364552ULL, "450438330.ogg"},
	{0xD3553A9EF8FC7FBBULL, "136489666.ogg"},
	{0xD3556416511A8FA0ULL, "vxw_woundedcoast_5r_0_0.vlm"},
	{0xD355BC548A4737E3ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamsml.anb"},
	{0xD355CDFEA184E2AAULL, "face_face_p3.psh.fxo"},
	{0xD356565AA87E7E7AULL, "art\\characters\\weapons\\longswords\\w_lsw_ls105a.mmh"},
	{0xD3565AA6FA3E35C4ULL, "vxlt_0_lowtown_docks_d2z_0_0.vlm"},
	{0xD356685AA87E9CC0ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls105a.msh"},
	{0xD356F061CA6DA1E5ULL, "6004193_m.fxe"},
	{0xD357B4EC11377B3DULL, "hlac_w_encounter2_camp_4_0y.dds"},
	{0xD357F773A95A7780ULL, "6086239_m.fxe"},
	{0xD3582D8DAEFBDE81ULL, "vxw_blightlands_vx_0_0.vlm"},
	{0xD3582DA70090F510ULL, "t3_wdg_w30.tnt"},
	{0xD35890578848D680ULL, "zz_endshot_v2.cut"},
	{0xD358D0E948DD1F54ULL, "vxlt_undercityserialkiller_1wp_0_0.vlm"},
	{0xD358D4AF91DF840DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wic_0n.dds"},
	{0xD358D7E3EF9B6052ULL, "vxlt_undercityserialkiller_17f_0_0.vlm"},
	{0xD35929C58926B38BULL, "vxw_woundedcoast_1q4_0_0.vlm"},
	{0xD3593076D4EB2E3CULL, "6170080_m.fxe"},
	{0xD35A5215C267979CULL, "494799390.ogg"},
	{0xD35A57020FE825F2ULL, "6016773_m.fxe"},
	{0xD35A6FDC3D5935C6ULL, "6091394_m.fxe"},
	{0xD35A7F430BD0C379ULL, "gen00fl_carver6108921.cl"},
	{0xD35B37F4FBA0796EULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyrium.mmh"},
	{0xD35B3A595875B78DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_rockpilexa.phy"},
	{0xD35B43F4FBA08D05ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyrium.mao"},
	{0xD35B505BD2DD654DULL, "dae000_bartrand6030621.cl"},
	{0xD35B51F4FBA0A55CULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyrium.msh"},
	{0xD35B88067B3BF21DULL, "level_buildtest_ext.arl"},
	{0xD35BDED3982A933FULL, "lt_0_lowtown_docks_n_189_0.msh"},
	{0xD35BE4D3982A9DE1ULL, "lt_0_lowtown_docks_n_189_0.mmh"},
	{0xD35BF0D3982AB182ULL, "lt_0_lowtown_docks_n_189_0.mao"},
	{0xD35C21A3E836A366ULL, "vxht_brothel_n_3v6_0_0.vlm"},
	{0xD35C424E3EB04C3FULL, "396117828.ogg"},
	{0xD35CA66ED97F98B0ULL, "994284826.ogg"},
	{0xD35D9CC95A39D5BDULL, "vxlt_0_lowtown_a3_n_c_i0e_0_0.vlm"},
	{0xD35DA89FA4CC937AULL, "vxlt_0_lowtown_a3_n_c_gbr_0_0.vlm"},
	{0xD35DE6D388DB7AF3ULL, "6140905_m.fxe"},
	{0xD35E016E73CD370AULL, "131897902.ogg"},
	{0xD35E423A28C0C724ULL, "vxlt_0_lowtown_fe8_0_0.vlm"},
	{0xD35EAF49A59B969BULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medi.mmh"},
	{0xD35ECD49A59BC985ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medi.msh"},
	{0xD35ED96D86A653F7ULL, "vxw_encounter2_fa_0_0.vlm"},
	{0xD360705F283F2F77ULL, "70878384.ogg"},
	{0xD360D65EB4BFA2A4ULL, "vxlt_gallowstemplar_n_3qy_0_0.vlm"},
	{0xD3611A1438E17B3DULL, "art\\vfx\\cinematic\\vfx_anders_romance_hand_c.mmh"},
	{0xD36154D52C37AB8EULL, "dae000_bartrand6031168.cl"},
	{0xD361F34DAE9FF805ULL, "fx_fireball01.dds"},
	{0xD362300EC2A090A5ULL, "993640132.ogg"},
	{0xD36231F6107E868BULL, "571231383.ogg"},
	{0xD36272FA66199863ULL, "80967445.ogg"},
	{0xD36299548A51860CULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamsml.mmh"},
	{0xD362AB08C70E10D5ULL, "vxlt_gallowstemplar_n_5pi_0_0.vlm"},
	{0xD362B679B0B6FE58ULL, "866206048.ogg"},
	{0xD362D5C206EA26E2ULL, "vxw_encounter2_c7_0_0.vlm"},
	{0xD362E879FFBDB642ULL, "mag100_karras6034516.cl"},
	{0xD3636409C3AD3226ULL, "590354974.ogg"},
	{0xD363D5F34310A820ULL, "vxw_woundedcoast_2s_0_0.vlm"},
	{0xD3643E16D2481131ULL, "vxlt_undercityencounter_113_0_0.vlm"},
	{0xD364FCED5B4B2686ULL, "character_character_p137.psh.fxo"},
	{0xD3650C8C9DB7EB3FULL, "w_woundedcoast_119_0.phy"},
	{0xD3663393A9D1AB88ULL, "dae000_gamlen6005057.cl"},
	{0xD366C26A3FB1F26DULL, "6039778_f.fxe"},
	{0xD366CA5DC32E7BBBULL, "vxw_encounter2_ri_0_0.vlm"},
	{0xD3674EBCB4503C81ULL, "177040006.ogg"},
	{0xD36798F6A640A8ECULL, "rdr111dg_martin6055006.cl"},
	{0xD367B46ADFD5298FULL, "628817362.ogg"},
	{0xD367E40C7F77B05EULL, "vxca_caverns1_mines_ck4_0_0.vlm"},
	{0xD368A2D4CC3EB254ULL, "sst_ss_caress.gda"},
	{0xD368CDC46A1C3B22ULL, "6208376_m.fxe"},
	{0xD368EC99B7A00A69ULL, "vxlt_0_lowtown_a2_n_c_hwt_0_0.vlm"},
	{0xD3691E7CA4D8B2B0ULL, "vxw_woundedcoast_184_0_0.vlm"},
	{0xD36967DFD0C61E67ULL, "vxw_woundedcoast_198_0_0.vlm"},
	{0xD3696A240380A632ULL, "908714128.ogg"},
	{0xD3696BA4EEDF7464ULL, "vxw_bonepitdestroyed_a3_1fq_0_0.vlm"},
	{0xD3699261FF0F1BC7ULL, "hlac_w_bonepit_196609_0z.dds"},
	{0xD369E60C2B93F66EULL, "mer340_bonepit_comments6083157.cl"},
	{0xD36A346B8D991659ULL, "65578026.ogg"},
	{0xD36A8D7C1470F360ULL, "738787613.ogg"},
	{0xD36B48A66C50599FULL, "mcr310_argument6075133.cl"},
	{0xD36B939FBFB2FF66ULL, "lgt250tr_crazedloner.utt"},
	{0xD36B9C9F5E0E28E8ULL, "ser221tr_chase_jump1.utt"},
	{0xD36C1A7C96C2ABFAULL, "vxlt_0_lowtown_a3_n_c_gys_0_0.vlm"},
	{0xD36C3DD034578B20ULL, "1068096800.ogg"},
	{0xD36C5368588F5549ULL, "vxht_mansion2_3gw_0_1.vlm"},
	{0xD36C6DBE1D7F4F6FULL, "6107157_m.fxe"},
	{0xD36C9820DA4DD84FULL, "305206243.ogg"},
	{0xD36D2769192786DFULL, "matdef_dc_specular.dds"},
	{0xD36D7889CBBCD1D8ULL, "mer341cr_jansen.utc"},
	{0xD36D97B67B065171ULL, "6208267_m.fxe"},
	{0xD36DA5ED43023B00ULL, "zz_pro_cutscenes6005834.cl"},
	{0xD36E3C41934C4CE9ULL, "vxlt_hangedman_n_26e_0_0.vlm"},
	{0xD36E6F9A5C634827ULL, "739931605.ogg"},
	{0xD36E8FD80763417DULL, "dae200_cullen6140303.cl"},
	{0xD36EA70A58867C0AULL, "vxca_caverns2prc_aci_0_0.vlm"},
	{0xD36F2FD397F1798EULL, "lt_0_lowtown_a2_n_c_25_0_ll.mmh"},
	{0xD36F5419AA946797ULL, "dae110ar_docks.ncs"},
	{0xD36F8D28A0283797ULL, "6100004_m.fxe"},
	{0xD36FDC4EF7B2BBB0ULL, "140594120.ogg"},
	{0xD3703B0EA19A569EULL, "6155847_m.fxe"},
	{0xD37053509624ECE0ULL, "6016739_m.fxe"},
	{0xD3705BC08EAEEABBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelred.msh"},
	{0xD3706DC08EAF0906ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelred.mao"},
	{0xD37079C08EAF1DA5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelred.mmh"},
	{0xD370814C4EA3618DULL, "819361490.ogg"},
	{0xD370BD1F97AD2406ULL, "t1_stb_wht.tnt"},
	{0xD371479EF8E03F62ULL, "vxw_encounter2_n0_0_0.vlm"},
	{0xD3715F5650E6A55DULL, "448982318.ogg"},
	{0xD371640F44A9C5D3ULL, "6172849_m.fxe"},
	{0xD3717EEC32490F8FULL, "lt_0_lowtown_docks_a3_n_194_0.mao"},
	{0xD3718AEC3249242CULL, "lt_0_lowtown_docks_a3_n_194_0.mmh"},
	{0xD37190EC32492E1EULL, "lt_0_lowtown_docks_a3_n_194_0.msh"},
	{0xD371A93E9A7144A4ULL, "dae000_brat.cnv"},
	{0xD371ACE542838E82ULL, "hlac_lt_gallowstemplar_f_2391_7y.dds"},
	{0xD371E4F6C3001225ULL, "llac_w_encounter3_-196603_0.dds"},
	{0xD372277437914E5EULL, "6116091_m.fxe"},
	{0xD372A5642FB98627ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb12a.phy"},
	{0xD3732CF67FCB08C4ULL, "607856929.ogg"},
	{0xD3734D14BAEEB43CULL, "6117821_m.fxe"},
	{0xD3738986528D48E2ULL, "6182083_m.fxe"},
	{0xD374367933A6407BULL, "vxht_keep_3xq_0_0.vlm"},
	{0xD374CB3E9A73B404ULL, "dae000_brat.bnk"},
	{0xD374E0B1A3E26678ULL, "vxlt_gallowstemplar_f_46s_0_0.vlm"},
	{0xD37573788096558CULL, "673849854.ogg"},
	{0xD375BDF6C4AE341BULL, "vxlt_warehouse_n_5z_0_0.vlm"},
	{0xD375D2A07F330DBBULL, "qun210cr_guard6072152.cl"},
	{0xD3760D8C9DC66044ULL, "w_woundedcoast_119_0.mmh"},
	{0xD376118C9DC66697ULL, "w_woundedcoast_119_0.mao"},
	{0xD376238C9DC685E6ULL, "w_woundedcoast_119_0.msh"},
	{0xD3762B0A623AF113ULL, "dae000_mother6199752.cl"},
	{0xD37643FBA7AD3D95ULL, "6106814_f.fxe"},
	{0xD3764936CB7FDC6BULL, "6116552_f.fxe"},
	{0xD3765419AA9A86E6ULL, "dae110ar_docks.lst"},
	{0xD376ED0E450BEBA0ULL, "one100dg_nabil6033115.cl"},
	{0xD376F79A6A9487F9ULL, "6126913_f.fxe"},
	{0xD377447B3CE76EEEULL, "209595376.ogg"},
	{0xD3775E08961A8348ULL, "l_qun210ar_javaris_encounter.wbd"},
	{0xD377C2524FB23325ULL, "223670450.ogg"},
	{0xD377C32DDB8DD0FDULL, "vxlt_0_lowtown_a3_n_c_df2_0_0.vlm"},
	{0xD37802E6271FE0CDULL, "826114573.ogg"},
	{0xD378D9817C0BFAB7ULL, "221037044.ogg"},
	{0xD379220E04FEC5CBULL, "315468040.ogg"},
	{0xD379F66CCA8FD693ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1light.phy"},
	{0xD37A07605B9E228DULL, "6164743_m.fxe"},
	{0xD37A27AFD39105A2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_lgtc_0t.dds"},
	{0xD37A5B74B4003E6FULL, "hlac_ht_0_hightown_-65535_0y.dds"},
	{0xD37A74410E3F9002ULL, "one100dg_vanard6056929.cl"},
	{0xD37A86BB083A5366ULL, "plt_vrc200pt_n_fm_start.plo"},
	{0xD37A90742FEB47D9ULL, "272142276.ogg"},
	{0xD37AEBADC7BDBC37ULL, "vxht_mansion2b_1ri_0_0.vlm"},
	{0xD37AF2E20FFE6414ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_alchemic.phy"},
	{0xD37AFA3A5CA75DCEULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_drka_0d.dds"},
	{0xD37B2ACFEAFDBE40ULL, "6173016_m.fxe"},
	{0xD37B755AA89DCE85ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls105a.phy"},
	{0xD37BBD276A9284F7ULL, "vxw_encounter2_lc_0_0.vlm"},
	{0xD37C039024C8B0D4ULL, "qun111_ketojan6040981.cl"},
	{0xD37C08DBF3419E8AULL, "383702063.ogg"},
	{0xD37C758522B9D2B6ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls101a.mmh"},
	{0xD37C7F8522B9E3B4ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls101a.msh"},
	{0xD37CE9FC15BAFFE2ULL, "vxlt_0_lowtown_a3_n_c_hcn_0_0.vlm"},
	{0xD37D32B54BA98B7BULL, "mag100_cullen6093735.cl"},
	{0xD37D846181D9D189ULL, "one200_gold_followers.bnk"},
	{0xD37E33C4FF79BDAEULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_pota.phy"},
	{0xD37ED817F69BB824ULL, "375937712.ogg"},
	{0xD37F09622588D77AULL, "ht_0_hightown_a3_20_0_lc.mmh"},
	{0xD37F111026189BE9ULL, "lt_0_lowtown_docks_a3_115_0.mao"},
	{0xD37F1D102618AF82ULL, "lt_0_lowtown_docks_a3_115_0.mmh"},
	{0xD37F1F102618B338ULL, "lt_0_lowtown_docks_a3_115_0.msh"},
	{0xD37F30288EEE4F4CULL, "gencr_qunari_warrior.utc"},
	{0xD37FD5EC369FDE53ULL, "6175949_m.fxe"},
	{0xD3803AA01B10A2F4ULL, "6053563_m.fxe"},
	{0xD3806319AAA2F4F1ULL, "dae110ar_docks.are"},
	{0xD380A3EB6D67167FULL, "422895579.ogg"},
	{0xD380B8FD5BA0F060ULL, "gen00fl_isabela6100029.cl"},
	{0xD3811FBDD6BD3D2EULL, "hlac_w_bonepit_393221_0z.dds"},
	{0xD381666181DD87C9ULL, "one200_gold_followers.cnv"},
	{0xD3826E37BC804350ULL, "dae000_katriela6162661.cl"},
	{0xD38292145C2D5CE2ULL, "765651730.ogg"},
	{0xD383A0595AD74CABULL, "724586393.ogg"},
	{0xD383A8DC2402D85CULL, "follower_banter6170319.cl"},
	{0xD383C646DC33D1E3ULL, "art\\characters\\playercharacter\\humanmale\\glv_masb.msh"},
	{0xD383D446DC33E93DULL, "art\\characters\\playercharacter\\humanmale\\glv_masb.mmh"},
	{0xD383D846DC33F04EULL, "art\\characters\\playercharacter\\humanmale\\glv_masb.mao"},
	{0xD38402C248810246ULL, "art\\vfx\\textures\\vfs_grenade_bottle.dds"},
	{0xD3842CE7F2097F77ULL, "ser221_gascard6127893.cl"},
	{0xD384322E4886BD05ULL, "vxlt_gallowstemplar_n_4xx_0_0.vlm"},
	{0xD3844545F62AAE6AULL, "vxca_caverns2prc_7v5_0_1.vlm"},
	{0xD3844F8A18B7142AULL, "vxht_0_hightown_c1k_0_0.vlm"},
	{0xD384B11EC24422B1ULL, "charhighlight.vsh.fxo"},
	{0xD384B85A14B15025ULL, "llac_lt_0_lowtown_n_65535_5.dds"},
	{0xD384CFE210067D11ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_alchemic.msh"},
	{0xD384DDE21006952CULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_alchemic.mao"},
	{0xD384E9E21006A9CFULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_alchemic.mmh"},
	{0xD38534C8DA534BFEULL, "vxdr_0_deeproads_9cl_0_0.vlm"},
	{0xD385A1B62ECDA4BEULL, "hlac_lt_0_lowtown_a3_n_c_-65535_2z.dds"},
	{0xD385E28C52F7018EULL, "vxw_woundedcoast_1g7_0_0.vlm"},
	{0xD3862AE93F3DE2EDULL, "fake_im_arm_glv_med_champ.uti"},
	{0xD3864F5ED2CBA564ULL, "vxw_bonepitdestroyed_a3_1lw_0_0.vlm"},
	{0xD386A2004DE2F0E2ULL, "vxlt_undercitydungeon_1mq_0_0.vlm"},
	{0xD386EE33BEFCA3E6ULL, "6097316_m.fxe"},
	{0xD38807508EB56396ULL, "plt_cod_hst_anderfels.plo"},
	{0xD38847FA146BAB1FULL, "vxht_0_hightown_ajq_0_0.vlm"},
	{0xD3887F00E2EB6C28ULL, "vxlt_0_lowtown_a2_n_c_f66_0_2.vlm"},
	{0xD388BAB03DCC1D65ULL, "vxlt_0_lowtown_n_ep3_0_1.vlm"},
	{0xD388E77A6244A36CULL, "art\\vfx\\cinematic\\vfx_barrel_explosion_p.impa.ani"},
	{0xD388ECFEDABCD288ULL, "follower_banter6143339.cl"},
	{0xD389312140ECDABBULL, "vxlt_darktown_364_0_0.vlm"},
	{0xD3895D8AB9B3D4B7ULL, "vxht_mansion2b_1uh_0_0.vlm"},
	{0xD389AF4590B2B89BULL, "w_encounter2_20_0_ll.mmh"},
	{0xD38A2AF9E2539A26ULL, "w_encounter1_17_0_ll.mmh"},
	{0xD38AB2E013E41572ULL, "22727521.ogg"},
	{0xD38B53B5749043A1ULL, "lt_0_lowtown_docks_a3_25_0.rml"},
	{0xD38B894645BF1B41ULL, "mrl000_marethari6069834.cl"},
	{0xD38B926133DD2E4AULL, "742025356.ogg"},
	{0xD38BF75D5001B536ULL, "6161570_m.fxe"},
	{0xD38C3685D022ABD5ULL, "zz_start_qcr6081936.cl"},
	{0xD38C5E1324DC709AULL, "ave321cr_guardsman3.utc"},
	{0xD38C9A752BA6539CULL, "vxca_cavern1_haunted_ck6_0_0.vlm"},
	{0xD38CD83E453033EFULL, "vxca_caverns1_mines_gk5_0_0.vlm"},
	{0xD38D007278A2E734ULL, "lgt210ip_selby_board.utp"},
	{0xD38D03EE7903CAA2ULL, "llac_w_bonepit_4_0.dds"},
	{0xD38D6280452FFEB7ULL, "art\\vfx\\models\\vfx_target_ring.mmh"},
	{0xD38D788045302429ULL, "art\\vfx\\models\\vfx_target_ring.msh"},
	{0xD38DAABCBEFF88A4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lute_sa.dds"},
	{0xD38DAB1B21B8781FULL, "543010509.ogg"},
	{0xD38DD0D5ADC89211ULL, "hlac_lt_0_lowtown_a2_n_c_-2_4x.dds"},
	{0xD38E009F9CF05A55ULL, "vxlt_gallowstemplar_n_5av_0_0.vlm"},
	{0xD38E5AE845AAE939ULL, "495658444.ogg"},
	{0xD38ECA34DA50687CULL, "508445147.ogg"},
	{0xD38EEF70154B0478ULL, "vxw_encounter2_camp_qj_0_1.vlm"},
	{0xD38EF47A624924AEULL, "art\\vfx\\cinematic\\vfx_barrel_explosion_p.impa.gad"},
	{0xD38F42E1799BFCDDULL, "qcr221_cut_qunari_defeated6205150.cl"},
	{0xD38F61D735CFB39BULL, "hlac_ht_0_hightown_n_2_1z.dds"},
	{0xD38F67DA50F2BBDCULL, "32968479.ogg"},
	{0xD38F89CBF22DF00FULL, "6177827_f.fxe"},
	{0xD390456DB86EF779ULL, "vxlt_0_lowtown_docks_a3_n_bfa_0_0.vlm"},
	{0xD3904FFA3B42B58FULL, "433627747.ogg"},
	{0xD3912095CBE78DC4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_torturerack_sa.dds"},
	{0xD391A68902D43F31ULL, "6174892_m.fxe"},
	{0xD392078E803E15EDULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_sword.msh"},
	{0xD392158E803E2D20ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_sword.mao"},
	{0xD392198E803E3473ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_sword.mmh"},
	{0xD3929F2941FA8A64ULL, "vxdr_primevaltomb_8es_0_0.vlm"},
	{0xD392C1670AAD2CAAULL, "vxht_0_hightown_c4d_0_0.vlm"},
	{0xD3930A2201C20472ULL, "vxca_cavern1_haunted_gke_0_1.vlm"},
	{0xD3932CE4C951D1D8ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_cth_comx_0d.dds"},
	{0xD3937ECBAE70F4CEULL, "6092395_m.fxe"},
	{0xD393A6A5CC49647EULL, "vxdr_0_deeproads_9dm_0_0.vlm"},
	{0xD394311077182E00ULL, "128130836.ogg"},
	{0xD3943FF6ECA9CDC1ULL, "mag121_idunna6084868.cl"},
	{0xD3945FFAA0A654F2ULL, "6105598_m.fxe"},
	{0xD39536AFA104DDA2ULL, "vxht_keep_5zq_0_1.vlm"},
	{0xD39578106E12E5E7ULL, "552734243.ogg"},
	{0xD3959E6FEFF1A0C7ULL, "218544157.ogg"},
	{0xD395A3961FBDD724ULL, "219188739.ogg"},
	{0xD395EAAB0F5024E7ULL, "6062779_m.fxe"},
	{0xD395EF7A624F3A8DULL, "art\\vfx\\cinematic\\vfx_barrel_explosion_p.impa.evt"},
	{0xD395EF879CE190F3ULL, "llac_lt_0_lowtown_a3_n_c_-65537_1.dds"},
	{0xD39611224055038EULL, "vxlt_darktown_59v_0_0.vlm"},
	{0xD396A2133518F70BULL, "zz_mer_debug6062745.cl"},
	{0xD396B9D70661C39FULL, "vxht_0_hightown_amp_0_0.vlm"},
	{0xD396EDDE21096528ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_chac_n.dds"},
	{0xD396F26520C337C2ULL, "6132771_f.fxe"},
	{0xD397531DFABF98C5ULL, "hlac_lt_gallowscourtyard_n_-131073_0z.dds"},
	{0xD39756A387E5B3D6ULL, "vxht_0_hightown_a2_ceo_0_0.vlm"},
	{0xD3975F6946FC2DCEULL, "qun221_petrice6075976.cl"},
	{0xD3977254C85D2311ULL, "137794051.ogg"},
	{0xD397DB9290E10C0AULL, "472237548.ogg"},
	{0xD3983552A88B5001ULL, "lgt160_ntb_elf_conflict6163324.cl"},
	{0xD3989F5222104221ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_c.anb"},
	{0xD39967E33DF5EACFULL, "t3_wpn_hrn.tnt"},
	{0xD399E081614EDD5BULL, "one100dg_vanard6173960.cl"},
	{0xD39A46D750DF2DBFULL, "mcr370_messenger_st.stg"},
	{0xD39AC0E239136183ULL, "6086203_m.fxe"},
	{0xD39AC49C2F040537ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med2_c.anb"},
	{0xD39AD922E5B90EEEULL, "705507363.ogg"},
	{0xD39B3FFE04FCD81CULL, "vxht_keep_n_70a_0_0.vlm"},
	{0xD39B4880BF82AAA5ULL, "nat380_nathaniel6070249.cl"},
	{0xD39BBB1D2324A24FULL, "vxlt_foundry_n_wr_0_0.vlm"},
	{0xD39BC9DB4939F043ULL, "379079534.ogg"},
	{0xD39C40EAAD93A944ULL, "gen_im_wep_war_one_act1_06.uti"},
	{0xD39CC5A1E246E268ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masb_0n.dds"},
	{0xD39CE05378E8A184ULL, "631103497.ogg"},
	{0xD39CE722C2C05DA2ULL, "vxdr_primevalentrance_1q0_0_0.vlm"},
	{0xD39D1DCC60CE42C2ULL, "hlac_lt_0_lowtown_docks_-131071_0x.dds"},
	{0xD39D861E4B73EC1FULL, "6118016_m.fxe"},
	{0xD39DB3B027F25A84ULL, "474667215.ogg"},
	{0xD39DC111EE9477B6ULL, "463539884.ogg"},
	{0xD39DCC7373E3BDF6ULL, "858377019.ogg"},
	{0xD39DE0A0576C9A20ULL, "6149061_f.fxe"},
	{0xD39E12EE36162933ULL, "vrc321_party_barks6096301.cl"},
	{0xD39E685D5E807DDCULL, "305165798.ogg"},
	{0xD39E7A4A768A7CC9ULL, "hlac_w_sundermount_-327682_0y.dds"},
	{0xD39EA40A326119AEULL, "mag100_karras6034927.cl"},
	{0xD39EB78FEBAEC2A5ULL, "character_character_p8.psh.fxo"},
	{0xD39EDA668FA9E343ULL, "zz_followers_banter6157676.cl"},
	{0xD39F1326CF3E55CFULL, "vxlt_gallowstemplar_f_5sz_0_0.vlm"},
	{0xD39F2F5A86B97C3EULL, "art\\vfx\\materials\\vfx_gaping_wound.mao"},
	{0xD39F7510B1976C6FULL, "609871798.ogg"},
	{0xD39F7D4B31E3075BULL, "hlac_w_bonepit_327682_0y.dds"},
	{0xD39F7FFEA25F798CULL, "mcr371_angry_mage_m6087025.cl"},
	{0xD39F8C10098889FBULL, "vxht_keep_3ir_0_0.vlm"},
	{0xD39FA62EB6B35CBAULL, "35673347.ogg"},
	{0xD39FF3760833A4B7ULL, "6002293_f.fxe"},
	{0xD3A0B919F391A9CDULL, "vxlt_hangedman_2zr_0_0.vlm"},
	{0xD3A181257A67940DULL, "art\\characters\\weapons\\dagger\\w_dgr_dg102a.phy"},
	{0xD3A1F25B6D99680EULL, "vxlt_undercitydungeon_yl_0_0.vlm"},
	{0xD3A2EA08964216AAULL, "fen110_fenris6097072.cl"},
	{0xD3A319E5D00ACF04ULL, "37395612.ogg"},
	{0xD3A32828DA5398AAULL, "vrc221pt_family_matter.ncs"},
	{0xD3A33843204A1805ULL, "vxlt_gallowstemplar_f_46j_0_0.vlm"},
	{0xD3A35BDE8EB6CB67ULL, "llac_ca_caverns1_mines_24424_16.dds"},
	{0xD3A38E5860DF67F2ULL, "223095722.ogg"},
	{0xD3A3CA5D863BA982ULL, "ave331_jeven.cnv"},
	{0xD3A3D164F7C25292ULL, "zz_seb_debug6215766.cl"},
	{0xD3A45C8801463D67ULL, "6069679_m.fxe"},
	{0xD3A4A74D5B9A2404ULL, "263716885.ogg"},
	{0xD3A4CE3A8397D364ULL, "vxlt_0_lowtown_docks_b0f_0_0.vlm"},
	{0xD3A4D21C5F01ACE5ULL, "vfx_lightning_f_06.dds"},
	{0xD3A4D77EF03388C2ULL, "llac_dr_primevaltomb_658_10.dds"},
	{0xD3A4E8BDF2D4E30AULL, "vxlt_undercitydungeon_1c3_0_0.vlm"},
	{0xD3A4EA7AE1B48B13ULL, "754299775.ogg"},
	{0xD3A5255116E8B67EULL, "prm000im_isaac_chest.uti"},
	{0xD3A52A0379E7BC6CULL, "681578667.ogg"},
	{0xD3A575DEF75F5D40ULL, "pro000_soldier_leader6040811.cl"},
	{0xD3A5D049A5D7D1D0ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_medi.phy"},
	{0xD3A676C88AF5929BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07a.msh"},
	{0xD3A68052221C4A9AULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_c.mmh"},
	{0xD3A694C88AF5C585ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07a.mmh"},
	{0xD3A69EAF843D7152ULL, "6082282_m.fxe"},
	{0xD3A706BF7137A58FULL, "560234661.ogg"},
	{0xD3A70FACC2FFC539ULL, "fx_cst_cloud_add.mao"},
	{0xD3A7AC5D863F5FBAULL, "ave331_jeven.bnk"},
	{0xD3A840D46DD848DEULL, "vxlt_gallowstemplar_n_3mg_0_0.vlm"},
	{0xD3A89E35E34EB23EULL, "vxw_encounter3_7l_0_0.vlm"},
	{0xD3A92A73764873B2ULL, "vxca_caverns2prc_7z8_0_0.vlm"},
	{0xD3A9A00EBEA50954ULL, "w_blightlands_59_0.rml"},
	{0xD3A9F1A4834495DFULL, "vxlt_0_lowtown_docks_a3_cp3_0_0.vlm"},
	{0xD3AA69A44EA2435DULL, "dae011_gossips6143004.cl"},
	{0xD3AA75748336CF70ULL, "ser200_aveline6201543.cl"},
	{0xD3AAA889CC7F9A61ULL, "ser120_ghyslain6119915.cl"},
	{0xD3AABFD6B7070D76ULL, "6110721_m.fxe"},
	{0xD3AADF248B36E64EULL, "386363104.ogg"},
	{0xD3AB023015D19F04ULL, "gen_im_npc_elf_arm_cht_lgte.uti"},
	{0xD3AB31E4745699E5ULL, "520096140.ogg"},
	{0xD3AB69167F4A7DBDULL, "vxlt_0_lowtown_docks_a3_d3c_0_0.vlm"},
	{0xD3ABF7FB3D6A049FULL, "6117116_m.fxe"},
	{0xD3AC33F63A8485E6ULL, "6075521_m.fxe"},
	{0xD3AC3EA2A8448E3BULL, "vxht_brothel_n_6cy_0_0.vlm"},
	{0xD3AC90FCD6EE8166ULL, "gen00fl_carver6104537.cl"},
	{0xD3ACCADFB70D7E6DULL, "gen00fl_dog6171160.cl"},
	{0xD3AD5282A32486C7ULL, "hf_qun_crazed_woman.mor"},
	{0xD3ADE54F2B34C573ULL, "hlac_w_encounter2_camp_131077_0x.dds"},
	{0xD3AE2CF407B852C4ULL, "31446361.ogg"},
	{0xD3AE3FE30AB8AA2EULL, "lt_0_lowtown_docks_a3_n_183_0.msh"},
	{0xD3AE4DE30AB8C27FULL, "lt_0_lowtown_docks_a3_n_183_0.mao"},
	{0xD3AE549844D1D09AULL, "vxw_sundermount_36z_0_0.vlm"},
	{0xD3AE59E30AB8D61CULL, "lt_0_lowtown_docks_a3_n_183_0.mmh"},
	{0xD3AEBFEF43DC5941ULL, "6030647_f.fxe"},
	{0xD3AEEF63F94A3D93ULL, "hlac_ht_brothel_n_2995_0y.dds"},
	{0xD3AEFC2D600E62D7ULL, "6041959_f.fxe"},
	{0xD3AF5F44BCED82C3ULL, "w_encounter3_127_0.msh"},
	{0xD3AF6D44BCED9A1DULL, "w_encounter3_127_0.mmh"},
	{0xD3AF7144BCEDA12EULL, "w_encounter3_127_0.mao"},
	{0xD3B011D988E1C2B1ULL, "112340448.ogg"},
	{0xD3B079DC294036DEULL, "fen310_fenris6118674.cl"},
	{0xD3B0D18A49C9EEFBULL, "vxdr_primevaltomb_50n_0_0.vlm"},
	{0xD3B0F2A7673BEC99ULL, "vxw_bonepit_274_0_0.vlm"},
	{0xD3B1DC22F6B98D84ULL, "vxht_0_hightown_n_c3i_0_1.vlm"},
	{0xD3B1E1F0096124E3ULL, "vxw_bonepit_23z_0_0.vlm"},
	{0xD3B1E79A56549617ULL, "vxw_woundedcoast_ok_0_0.vlm"},
	{0xD3B2EF7278C399EDULL, "lgt210ip_selby_board.ncs"},
	{0xD3B304A1F4CE71E3ULL, "w_encounter3_214_0.mao"},
	{0xD3B308A1F4CE7830ULL, "w_encounter3_214_0.mmh"},
	{0xD3B316A1F4CE904AULL, "w_encounter3_214_0.msh"},
	{0xD3B33C7CA03188B6ULL, "401978775.ogg"},
	{0xD3B34017758DEBE4ULL, "vxlt_0_lowtown_docks_b7g_0_0.vlm"},
	{0xD3B3853C3822CFD8ULL, "6000905_m.fxe"},
	{0xD3B39B452FAD70A0ULL, "vxca_caverns1_e8t_0_0.vlm"},
	{0xD3B3C2788850334DULL, "pro000_merchant6051654.cl"},
	{0xD3B3E07C84A72BCCULL, "ts_profane_01.gda"},
	{0xD3B40238AB4A4FC5ULL, "6196702_m.fxe"},
	{0xD3B486D0CCCB1EC4ULL, "6175407_m.fxe"},
	{0xD3B48979954FB3B7ULL, "580186586.ogg"},
	{0xD3B498AF008D9CDFULL, "vxlt_0_lowtown_a2_n_c_ffo_0_0.vlm"},
	{0xD3B640E95FFBDAA5ULL, "317097176.ogg"},
	{0xD3B6BDCCC02DEF57ULL, "vxlt_gallowscourtyard_n_66p_0_0.vlm"},
	{0xD3B6FEE6269EEFF2ULL, "971547306.ogg"},
	{0xD3B752C08EEADE52ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelred.phy"},
	{0xD3B766F0B6A9574EULL, "w_bonepit_242_0.msh"},
	{0xD3B76ABA6CDA7D4DULL, "vxht_0_hightown_a2_bpf_0_0.vlm"},
	{0xD3B774F0B6A96E9FULL, "w_bonepit_242_0.mao"},
	{0xD3B780F0B6A9833CULL, "w_bonepit_242_0.mmh"},
	{0xD3B79D02D7F21962ULL, "projdecal_projdecal_p1.vsh.fxo.dx11"},
	{0xD3B88F5E628BF30BULL, "605466945.ogg"},
	{0xD3B8B9391E35AFA4ULL, "6127973_f.fxe"},
	{0xD3B8B974CC8372DFULL, "vxw_bonepit_11s_0_0.vlm"},
	{0xD3B9869FBFF50A3FULL, "lgt250tr_crazedloner.ncs"},
	{0xD3B9D8C14FE69581ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_rnf.ani"},
	{0xD3B9DAF37140963DULL, "vxlt_0_lowtown_docks_a3_d4b_0_0.vlm"},
	{0xD3BA316F88FB7CE8ULL, "6068468_m.fxe"},
	{0xD3BA5D9D2D6595A2ULL, "cd_mrl210_merrill.wbd"},
	{0xD3BAEF2B2AFE03BEULL, "gen00fl_aveline6120837.cl"},
	{0xD3BAEF57FAD2DC6AULL, "fx_0034_diff.dds"},
	{0xD3BAFAAB3DBCF7FDULL, "qun221_viscount6084686.cl"},
	{0xD3BB197E3B9B2ECDULL, "hlac_w_bonepitdestroyed_a3_1_0z.dds"},
	{0xD3BB5234BFAEC8DBULL, "vxdr_primevaltomb_3hp_0_0.vlm"},
	{0xD3BCE5FD4C41342CULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_varric_.ani"},
	{0xD3BD07A963A55936ULL, "hlac_lt_warehouse_n_65537_3y.dds"},
	{0xD3BD199125252821ULL, "lgt110cr_urn_barker.utc"},
	{0xD3BDA99DE886BF5BULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hpath2_n.dds"},
	{0xD3BDB69E8FC2CAB9ULL, "lowtown_male_alienage2_mob6166556.cl"},
	{0xD3BE32B574BB6531ULL, "lt_0_lowtown_docks_a3_25_0.anb"},
	{0xD3BE32C4FFAFA071ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_pota.mmh"},
	{0xD3BE3DDB4D732ECEULL, "hlac_lt_0_lowtown_docks_-131076_0y.dds"},
	{0xD3BE3EC4FFAFB412ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_pota.mao"},
	{0xD3BE4CC4FFAFCC2FULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_pota.msh"},
	{0xD3BE84089656C60AULL, "l_qun210ar_javaris_encounter.bnk"},
	{0xD3BEDE84E9D76E85ULL, "327125759.ogg"},
	{0xD3BEF2D70342FA97ULL, "gen00fl_merrill6127548.cl"},
	{0xD3BEF9F7C2A26199ULL, "150671580.ogg"},
	{0xD3BF22F833738254ULL, "6061243_m.fxe"},
	{0xD3BF5140145AA0CAULL, "6043520_m.fxe"},
	{0xD3BFA082B8CAC170ULL, "859141785.ogg"},
	{0xD3BFD74CF2498F37ULL, "art\\characters\\weapons\\shields\\w_shd_dsla.msh"},
	{0xD3BFDB4247B6DA4CULL, "vxlt_undercityserialkiller_1mz_0_0.vlm"},
	{0xD3BFE3C14FEB1423ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_rnf.gad"},
	{0xD3BFE94CF249AD9AULL, "art\\characters\\weapons\\shields\\w_shd_dsla.mao"},
	{0xD3BFED4CF249B4A9ULL, "art\\characters\\weapons\\shields\\w_shd_dsla.mmh"},
	{0xD3C047F992DA1297ULL, "6096879_m.fxe"},
	{0xD3C05977484AAE97ULL, "vxw_woundedcoast_dj_0_0.vlm"},
	{0xD3C0648795E818EFULL, "6068373_m.fxe"},
	{0xD3C0691704F1C816ULL, "vxca_caverns2prc_78v_0_0.vlm"},
	{0xD3C096853652E6B1ULL, "mag291_demon_marethari6027841.cl"},
	{0xD3C0D812E1EC87FDULL, "6087372_m.fxe"},
	{0xD3C203B29D12CB79ULL, "255963654.ogg"},
	{0xD3C237C1284EC295ULL, "zz_and_debug6134096.cl"},
	{0xD3C267B7711984C5ULL, "vxlt_0_lowtown_a3_n_c_gve_0_0.vlm"},
	{0xD3C2C4604C9757B1ULL, "mer320_hubert6090365.cl"},
	{0xD3C2CCA9AE9229D2ULL, "and100_anders6096041.cl"},
	{0xD3C2E615BDF785D3ULL, "vxlt_undercityserialkiller_1ok_0_0.vlm"},
	{0xD3C327A485AE1007ULL, "zz_fol_debug6092332.cl"},
	{0xD3C3749C70BF51FAULL, "zz_dre_debug6037561.cl"},
	{0xD3C3C19C2F270538ULL, "art\\vfx\\combat\\vfx_r_dwp_front_med2_c.mmh"},
	{0xD3C3EA48530D76B5ULL, "486799321.ogg"},
	{0xD3C3F2FD4C47696EULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_varric_.gad"},
	{0xD3C47E35BFAEC140ULL, "vxlt_0_lowtown_a2_n_f5g_0_0.vlm"},
	{0xD3C4A0D9F03A982EULL, "848957531.ogg"},
	{0xD3C4B36519871176ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls102a_0s.dds"},
	{0xD3C5283BA1B1260EULL, "gen00fl_aveline6139840.cl"},
	{0xD3C567BB087A09A5ULL, "plt_vrc200pt_n_fm_start.nss"},
	{0xD3C5A5C61B3F1C59ULL, "lt_backalley_n_1_0_ll.mmh"},
	{0xD3C5A7B5F00705B3ULL, "vxlt_0_lowtown_a2_n_g60_0_0.vlm"},
	{0xD3C68086B4458741ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right_quick.ani"},
	{0xD3C68B9F5E5AB715ULL, "ser221tr_chase_jump1.ncs"},
	{0xD3C6A9D9F3140F61ULL, "hlac_ca_caverns1_mines_24425_4x.dds"},
	{0xD3C6AB301B3D577DULL, "one200dg_solivitus6048212.cl"},
	{0xD3C6C0C14FF0F708ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_rnf.evt"},
	{0xD3C6E471B5D688D5ULL, "llac_lt_0_lowtown_a2_65538_5.dds"},
	{0xD3C724D5000C476FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\revenant_movement.anb"},
	{0xD3C76F1D49F1EA36ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_down.evt"},
	{0xD3C78422351DFA7DULL, "dae000cr_grand_cleric6181921.cl"},
	{0xD3C85D19D0733C9CULL, "hlac_lt_0_lowtown_n_65536_2y.dds"},
	{0xD3C87B3FD116B51BULL, "w_blightlands_233_0.msh"},
	{0xD3C88D3FD116D466ULL, "w_blightlands_233_0.mao"},
	{0xD3C8949F0E897F13ULL, "dre000_varric6037004.cl"},
	{0xD3C8993FD116E805ULL, "w_blightlands_233_0.mmh"},
	{0xD3C8B13D1B727FC6ULL, "6083700_m.fxe"},
	{0xD3C931B07110BD63ULL, "273186819.ogg"},
	{0xD3C94BA1908B6C42ULL, "art\\characters\\creatures\\gibgolem\\animation\\blendtree\\gibgolem_movement.anb"},
	{0xD3C962931C8D1B0AULL, "dae000_pc_home_ambients6208341.cl"},
	{0xD3C9CA2899252904ULL, "llac_lt_0_lowtown_docks_a3_n_0_0.dds"},
	{0xD3CA9EE4A7B6F05DULL, "character_character_p107.psh.fxo.dx11.assembly"},
	{0xD3CAAD557130BF05ULL, "vxht_0_hightown_n_3w1_0_0.vlm"},
	{0xD3CAAFF0CF3D6673ULL, "gen00fl_isabela6097360.cl"},
	{0xD3CB005F9F02DEC7ULL, "14000914.ogg"},
	{0xD3CB05C2901C5955ULL, "6093885_m.fxe"},
	{0xD3CB1E219926EDE6ULL, "hm_qmq_patronm4.mor"},
	{0xD3CB385228BE019AULL, "vxw_sundermount_38t_0_0.vlm"},
	{0xD3CB8D7A361654CEULL, "63437616.ogg"},
	{0xD3CC3DF9281BD974ULL, "vxht_0_hightown_a3_c3f_0_0.vlm"},
	{0xD3CC66CB7F4C2452ULL, "vxlt_undercityserialkiller_1on_0_0.vlm"},
	{0xD3CD55BECE8CF01EULL, "llac_w_encounter2_camp_196611_0.dds"},
	{0xD3CD5A674C2DC091ULL, "vxlt_0_lowtown_docks_a3_n_d30_0_0.vlm"},
	{0xD3CD7B611C930791ULL, "hlac_lt_undercitydungeon_-1_11z.dds"},
	{0xD3CD8B86B44BB8E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right_quick.gad"},
	{0xD3CDD23E5C7BDD92ULL, "gen00fl_aveline6137590.cl"},
	{0xD3CDF954BD8FEA5AULL, "vxht_mansion2b_25r_0_0.vlm"},
	{0xD3CE661D49F7F9C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_down.gad"},
	{0xD3CECB543A40C717ULL, "vxw_woundedcoast_ae_0_0.vlm"},
	{0xD3CECD4A30E51B5FULL, "305999493.ogg"},
	{0xD3CEDDE60F5400BAULL, "dae000_flemeth6000765.cl"},
	{0xD3CEFD0F779A7AFEULL, "6209930_f.fxe"},
	{0xD3CF1E3CA7A3AD82ULL, "6134990_m.fxe"},
	{0xD3CF7103288AC1A4ULL, "lgt310_mag_melson6152129.cl"},
	{0xD3CF7BA25ED81209ULL, "vxht_chantry_n_331_0_0.vlm"},
	{0xD3D00FD2FC492424ULL, "357628814.ogg"},
	{0xD3D07262A4C3AC4DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\mt_har_flma_0n.dds"},
	{0xD3D187B680EE9CB6ULL, "6197724_m.fxe"},
	{0xD3D1D5641D4B272FULL, "175315399.ogg"},
	{0xD3D1D5FDE52AA995ULL, "ser221_party_comments6166209.cl"},
	{0xD3D1F4B1492167F2ULL, "qun211_braggart6075528.cl"},
	{0xD3D22184AF5EE7AEULL, "6131895_f.fxe"},
	{0xD3D254D42D6A983EULL, "vxlt_0_lowtown_a2_ewk_0_0.vlm"},
	{0xD3D28002DBCD9E5FULL, "vxw_woundedcoast_3g_0_0.vlm"},
	{0xD3D285716C024964ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sitting_table_clasped_2_loop.gad"},
	{0xD3D2A641525A41D9ULL, "516904936.ogg"},
	{0xD3D2B444FEAB849DULL, "784125596.ogg"},
	{0xD3D2C1357AD75FFCULL, "gen_im_wep_rog_dua_lore_01.uti"},
	{0xD3D4036CBB300DCAULL, "6033552_m.fxe"},
	{0xD3D43AEC4CCDD9FCULL, "53356408.ogg"},
	{0xD3D46886B4519BC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_right_quick.evt"},
	{0xD3D4CC93CABB3052ULL, "411401850.ogg"},
	{0xD3D51DC5D45E6813ULL, "vxht_0_hightown_a2_ca3_0_1.vlm"},
	{0xD3D5671D49FE1A6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\wrist_l_down.ani"},
	{0xD3D61B8DD9F16F0CULL, "467763973.ogg"},
	{0xD3D6AEDA1BE18381ULL, "hlac_lt_backalley_g_0_12x.dds"},
	{0xD3D6B805C0DF1F62ULL, "vxw_blightlands_74_0_0.vlm"},
	{0xD3D7060F9838D9C3ULL, "6092098_f.fxe"},
	{0xD3D7378631A8BFD1ULL, "gen00fl_isabela6201222.cl"},
	{0xD3D776048ADA37CDULL, "306229348.ogg"},
	{0xD3D789C57DC3E7BBULL, "vxlt_0_lowtown_a3_n_c_h24_0_0.vlm"},
	{0xD3D7AC0860F92DECULL, "plt_zz_carpt_debug.plo"},
	{0xD3D853FE7311BF4DULL, "plt_lgt210pt_lowdown.nss"},
	{0xD3D8540A0A5A98E4ULL, "6052902_m.fxe"},
	{0xD3D85EF07C97B422ULL, "277219631.ogg"},
	{0xD3D86AAE74E73872ULL, "sst_ss_player_male.gda"},
	{0xD3D87C8D1D738FF8ULL, "gen00fl_varric6118820.cl"},
	{0xD3D8C86247AD6CEDULL, "w_sundermount_121_0.mmh"},
	{0xD3D8CC6247AD733EULL, "w_sundermount_121_0.mao"},
	{0xD3D8DA6247AD8B73ULL, "w_sundermount_121_0.msh"},
	{0xD3D954A38BD15996ULL, "hlac_dr_0_deeproads_5842_15y.dds"},
	{0xD3D98BD5FB7254A1ULL, "6162572_m.fxe"},
	{0xD3D994716C088142ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\sitting_table_clasped_2_loop.ani"},
	{0xD3D9956B7EBD3266ULL, "static_static_p9.psh.fxo"},
	{0xD3D9F01BB86A3552ULL, "6122114_f.fxe"},
	{0xD3D9F3A3C944D99EULL, "vxlt_blackemporium_uj_0_0.vlm"},
	{0xD3DA4940D3464322ULL, "462583211.ogg"},
	{0xD3DA808DBB6A0BCDULL, "vxlt_hangedman_2fn_0_0.vlm"},
	{0xD3DAAFD61A11F1F4ULL, "vxht_0_hightown_a3_c4g_0_0.vlm"},
	{0xD3DAB45D866AC765ULL, "ave331_jeven.stg"},
	{0xD3DAB86647BFE049ULL, "dae120cr_amb_dwarf_3.utc"},
	{0xD3DAC807465ED94BULL, "vxht_0_hightown_a2_c3t_0_0.vlm"},
	{0xD3DAFB02F05382DAULL, "w_bonepitdestroyed_a3_149_0.msh"},
	{0xD3DB0902F0539A13ULL, "w_bonepitdestroyed_a3_149_0.mao"},
	{0xD3DB0D02F053A120ULL, "w_bonepitdestroyed_a3_149_0.mmh"},
	{0xD3DB1D4AD7E6726CULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgta.msh"},
	{0xD3DB234AD7E67C5EULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgta.mmh"},
	{0xD3DB2F4AD7E690F5ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgta.mao"},
	{0xD3DBEF6CCAE29508ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1light.mmh"},
	{0xD3DBF612D5435A1AULL, "follower_banter6166394.cl"},
	{0xD3DBFB6CCAE2A9ABULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1light.mao"},
	{0xD3DC0D6CCAE2C832ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_1light.msh"},
	{0xD3DC7198A313311EULL, "w_woundedcoast_28_0.rml"},
	{0xD3DE15263394C877ULL, "vxht_brothel_n_6cm_0_0.vlm"},
	{0xD3DE910860FF1E23ULL, "plt_zz_carpt_debug.nss"},
	{0xD3DF1883D83B8191ULL, "215193820.ogg"},
	{0xD3DF3533F7F04660ULL, "vxca_caverns1_ika_0_0.vlm"},
	{0xD3DF47690C5C599AULL, "vxlt_undercityencounter_107_0_0.vlm"},
	{0xD3DF730BB2250C85ULL, "w_woundedcoast_38_0.rml"},
	{0xD3DFBA4A5CBBF5D9ULL, "gen00fl_dog6114950.cl"},
	{0xD3E0222CE80F4795ULL, "6042003_m.fxe"},
	{0xD3E02A365F2F04F1ULL, "6103670_m.fxe"},
	{0xD3E0C74B5E009109ULL, "111799768.ogg"},
	{0xD3E0F4C4FF015D7EULL, "hlac_ca_cavern1_haunted_24422_3y.dds"},
	{0xD3E10578790AECF9ULL, "6031352_f.fxe"},
	{0xD3E10B4CA4272F76ULL, "302563771.ogg"},
	{0xD3E13E784F317B8EULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney_d.dds"},
	{0xD3E15F53918F3CFBULL, "vxw_bonepit_yl_0_0.vlm"},
	{0xD3E18D649ED9009EULL, "628495314.ogg"},
	{0xD3E1E178F004DFE1ULL, "hlac_lt_0_lowtown_a2_-1_7y.dds"},
	{0xD3E1E47CA16D1431ULL, "vxlt_0_lowtown_docks_a3_n_bre_0_0.vlm"},
	{0xD3E23331D51D014AULL, "gen00fl_anders6111092.cl"},
	{0xD3E2A00EBED4CE17ULL, "w_blightlands_59_0.mmh"},
	{0xD3E2A40EBED4D524ULL, "w_blightlands_59_0.mao"},
	{0xD3E2B60EBED4F389ULL, "w_blightlands_59_0.msh"},
	{0xD3E2C05142BCC6CDULL, "vxht_0_hightown_a2_bcc_0_0.vlm"},
	{0xD3E2C86247B5C166ULL, "w_sundermount_121_0.rml"},
	{0xD3E3744624FE0DF3ULL, "6092596_m.fxe"},
	{0xD3E39FD871662E4AULL, "vxw_bonepit_2ai_0_0.vlm"},
	{0xD3E3C0B4C516C166ULL, "6053287_f.fxe"},
	{0xD3E409F71B728FDBULL, "vxht_brothel_n_4ft_0_0.vlm"},
	{0xD3E583E1F971ECEEULL, "zz_vrc_debug6064271.cl"},
	{0xD3E58A8935F828A1ULL, "vxdr_primevaltomb_6nh_0_1.vlm"},
	{0xD3E59050C33F912FULL, "lowtown_male_lowlife2_mob6166754.cl"},
	{0xD3E59567D2D037B5ULL, "art\\characters\\creatures\\spider\\animation\\blendtree\\spider_monstrous.anb"},
	{0xD3E5F08662B608C7ULL, "vxht_mansion_n_6ba_0_0.vlm"},
	{0xD3E644887EE0D314ULL, "vxca_caverns2_8fd_0_0.vlm"},
	{0xD3E65D42E82F2C9EULL, "dae000cr_grand_cleric6147654.cl"},
	{0xD3E66BB670792C24ULL, "zz_vault_debug6203237.cl"},
	{0xD3E79B08F01B3016ULL, "vxht_chantry_n_2to_0_1.vlm"},
	{0xD3E7E325967F3797ULL, "6083622_m.fxe"},
	{0xD3E8757767B801F9ULL, "llac_lt_gallowstemplar_f_2390_0.dds"},
	{0xD3E980FE124018AEULL, "gen00fl_fenris6115384.cl"},
	{0xD3E9C16247BBD3AAULL, "w_sundermount_121_0.phy"},
	{0xD3E9D1F72AA84C1CULL, "m_mag161ar_cave.wbd"},
	{0xD3E9E09C7B4B9B94ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_ulm_drka.dds"},
	{0xD3EA1E4D1FD0015BULL, "661113773.ogg"},
	{0xD3EA38A08E9E9EB6ULL, "6048774_m.fxe"},
	{0xD3EA5768FE971770ULL, "llac_w_sundermount_-262143_0.dds"},
	{0xD3EA62A162DFB41DULL, "gen00fl_fenris6112111.cl"},
	{0xD3EA6886E8B5D79AULL, "6144983_f.fxe"},
	{0xD3EA6AFEFA754E92ULL, "ht_0_hightown_n_36_0.anb"},
	{0xD3EB2AC8DFB926B0ULL, "49921540.ogg"},
	{0xD3EB2F1076F35DE0ULL, "vxlt_0_lowtown_a2_e88_0_0.vlm"},
	{0xD3EB52F86A5EDFB2ULL, "vxw_encounter2_camp_pv_0_0.vlm"},
	{0xD3EB620ED1F6BC32ULL, "zz_arm_override_test.cut"},
	{0xD3EBB370209787FAULL, "648837323.ogg"},
	{0xD3EBC1514F42A3CAULL, "vxca_caverns1_mines_gkm_0_1.vlm"},
	{0xD3EC47214D0D0E87ULL, "hlac_w_encounter1_-65538_0x.dds"},
	{0xD3ECB5985BE40320ULL, "hlac_lt_0_lowtown_a2_0_4y.dds"},
	{0xD3ED0A2AAF522F48ULL, "6152772_m.fxe"},
	{0xD3ED6DC88B318632ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos07a.phy"},
	{0xD3EE51E7F51DAD53ULL, "6091874_m.fxe"},
	{0xD3EE5BB0C406A065ULL, "6162525_m.fxe"},
	{0xD3EECC39F2CEFD97ULL, "vxw_bonepit_1wm_0_0.vlm"},
	{0xD3EF8583D81B5B21ULL, "6088845_m.fxe"},
	{0xD3EFB2BFF65FA746ULL, "841188686.ogg"},
	{0xD3EFD1308385557BULL, "vxw_bonepit_do_0_0.vlm"},
	{0xD3F00A1270E30942ULL, "772876238.ogg"},
	{0xD3F09F4EE7E7D270ULL, "follower_barks6153430.cl"},
	{0xD3F0E369714AD99BULL, "gen00fl_varric6133682.cl"},
	{0xD3F120EF7A7FB684ULL, "vxht_mansion2b_2on_0_0.vlm"},
	{0xD3F16C04F3F39A18ULL, "vxlt_gallowstemplar_f_3t2_0_0.vlm"},
	{0xD3F290E7DB234AE4ULL, "lt_gallowscourtyard_25_0.rml"},
	{0xD3F2EA4458700A7BULL, "follower_banter6136693.cl"},
	{0xD3F325C193BE651CULL, "vxlt_0_lowtown_a2_n_c_e8e_0_0.vlm"},
	{0xD3F35004AFDE45BCULL, "vxw_bonepit_27m_0_0.vlm"},
	{0xD3F3B12640774B39ULL, "vxlt_blackemporium_v8_0_0.vlm"},
	{0xD3F3BDC3234134ECULL, "ss_slavers_f_1.bnk"},
	{0xD3F3BF0C2CEBDAB9ULL, "hlac_ht_brothel_n_2995_1x.dds"},
	{0xD3F496F0AE305C55ULL, "6152756_f.fxe"},
	{0xD3F4992D839C141DULL, "706782177.ogg"},
	{0xD3F4FF1EF72B5719ULL, "6072268_m.fxe"},
	{0xD3F61D8DA0253167ULL, "6178362_m.fxe"},
	{0xD3F636F6086F252CULL, "vxca_caverns2prc_8gz_0_0.vlm"},
	{0xD3F6550288755017ULL, "6086922_m.fxe"},
	{0xD3F6684A5ECB4671ULL, "864182842.ogg"},
	{0xD3F6DCB387E28159ULL, "hlac_w_woundedcoast_65542_0x.dds"},
	{0xD3F6EFBDF2881D5FULL, "vxca_caverns1_e6y_0_0.vlm"},
	{0xD3F8D210CB204702ULL, "gen00fl_fenris6126812.cl"},
	{0xD3F92FAE179393A3ULL, "622533105.ogg"},
	{0xD3F9B969B39DBA29ULL, "vxlt_backalley_g_1eo_0_0.vlm"},
	{0xD3F9C4D55C54F832ULL, "vxw_encounter2_camp_kw_0_0.vlm"},
	{0xD3F9C82A080E59F7ULL, "lt_0_lowtown_docks_98_0.mmh"},
	{0xD3F9CC2A080E6004ULL, "lt_0_lowtown_docks_98_0.mao"},
	{0xD3F9DE2A080E7F69ULL, "lt_0_lowtown_docks_98_0.msh"},
	{0xD3FA08D95D5E0D8EULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_partycall.phy"},
	{0xD3FA20EB10231097ULL, "vxw_woundedcoast_pf_0_0.vlm"},
	{0xD3FA81E91659FFF5ULL, "gen00fl_varric6103085.cl"},
	{0xD3FAF8E01780F977ULL, "vxht_brothel_n_6io_0_0.vlm"},
	{0xD3FB0B5460CBE15CULL, "320433708.ogg"},
	{0xD3FC4BC0F21E7801ULL, "6054044_m.fxe"},
	{0xD3FCAD8985CFE7D3ULL, "vxlt_undercityserialkiller_1so_0_0.vlm"},
	{0xD3FCCD8DB5FCB8E5ULL, "6167625_m.fxe"},
	{0xD3FD36D8C7326DE7ULL, "follower_banter6171076.cl"},
	{0xD3FD61622FDD80E3ULL, "1002065614.ogg"},
	{0xD3FD98B0BFF0A107ULL, "hlac_lt_gallowscourtyard_n_-65537_4y.dds"},
	{0xD3FDF760CA1173A1ULL, "6083945_m.fxe"},
	{0xD3FE68D08D452E65ULL, "6101077_f.fxe"},
	{0xD3FEC8CAA74555ADULL, "gen00fl_anders6125762.cl"},
	{0xD3FED59DD0533F63ULL, "1049592515.ogg"},
	{0xD3FF2A902E0D991DULL, "129851607.ogg"},
	{0xD3FFA40B26A8F7CDULL, "vxht_0_hightown_a2_bia_0_0.vlm"},
	{0xD3FFD8CB57D78967ULL, "532320932.ogg"},
	{0xD3FFDDE1E5E9B298ULL, "vxlt_gallowstemplar_f_3s3_0_0.vlm"},
	{0xD400251917937B6EULL, "vxw_woundedcoast_1bh_0_0.vlm"},
	{0xD400387456690838ULL, "92160851.ogg"},
	{0xD4004251537D0957ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_loop.gad"},
	{0xD4016B9D2DA1B058ULL, "cd_mrl210_merrill.bnk"},
	{0xD40277A6A97F2B8CULL, "6016691_f.fxe"},
	{0xD4028D175D840251ULL, "6136362_m.fxe"},
	{0xD402978B433A9DE3ULL, "368030571.ogg"},
	{0xD402D6C276D99EBDULL, "w_bonepitdestroyed_a3_57_0.mao"},
	{0xD402DAC276D9A5F6ULL, "w_bonepitdestroyed_a3_57_0.mmh"},
	{0xD402E4C276D9B6F4ULL, "w_bonepitdestroyed_a3_57_0.msh"},
	{0xD4031EECB6713794ULL, "vxht_0_hightown_bgn_0_0.vlm"},
	{0xD403244A0E4D1410ULL, "hlac_ht_0_hightown_n_-131072_1x.dds"},
	{0xD403454AAC938222ULL, "6033634_f.fxe"},
	{0xD403473E902BBCE0ULL, "hlac_w_bonepit_0_2z.dds"},
	{0xD4039EB21F201C2BULL, "dae000_gamlen6147219.cl"},
	{0xD403B3298D5094EDULL, "vxw_sundermount_21i_0_0.vlm"},
	{0xD404530BB243F2AEULL, "w_woundedcoast_38_0.mmh"},
	{0xD4045F0BB2440745ULL, "w_woundedcoast_38_0.mao"},
	{0xD4046D0BB2441E9CULL, "w_woundedcoast_38_0.msh"},
	{0xD404AE3EF5A33221ULL, "270571584.ogg"},
	{0xD4050FC1D91DCABEULL, "llac_lt_0_lowtown_a2_n_c_-65535_5.dds"},
	{0xD4062E3F47248652ULL, "vxlt_undercityserialkiller_1sj_0_0.vlm"},
	{0xD40711026719A738ULL, "uparrow.dds"},
	{0xD407FC38219882DCULL, "w_bonepitmining_a2_65_0.mao"},
	{0xD40808382198977FULL, "w_bonepitmining_a2_65_0.mmh"},
	{0xD4080E382198A121ULL, "w_bonepitmining_a2_65_0.msh"},
	{0xD4082B46A593D2A9ULL, "vxlt_backalley_g_1bn_0_0.vlm"},
	{0xD40836B24E4B10B2ULL, "vxw_encounter2_camp_fp_0_0.vlm"},
	{0xD4086598A3385796ULL, "w_woundedcoast_28_0.mao"},
	{0xD4087198A3386C35ULL, "w_woundedcoast_28_0.mmh"},
	{0xD4087398A3386FABULL, "w_woundedcoast_28_0.msh"},
	{0xD408B58E0A3ACCF9ULL, "vxht_bartrand_n_5nk_0_0.vlm"},
	{0xD4094C1F635BC390ULL, "vxlt_hangedman_n_2ei_0_0.vlm"},
	{0xD409A8834221E7E7ULL, "follower_barks6173411.cl"},
	{0xD409E08B3EABAB9CULL, "930698814.ogg"},
	{0xD40A18E1496948E0ULL, "hlac_lt_undercityencounter_-1_45y.dds"},
	{0xD40A333C58FA1A55ULL, "838052902.ogg"},
	{0xD40AE8AA38484B2BULL, "6090037_m.fxe"},
	{0xD40C54C037F2A9C4ULL, "w_encounter3_241_0.msh"},
	{0xD40C66C037F2C82DULL, "w_encounter3_241_0.mao"},
	{0xD40C6AC037F2CF66ULL, "w_encounter3_241_0.mmh"},
	{0xD40D0E8A3071383AULL, "w_sundermount_139_0.mmh"},
	{0xD40D128A30713F71ULL, "w_sundermount_139_0.mao"},
	{0xD40D1586E2928E07ULL, "6111558_m.fxe"},
	{0xD40D208A30715680ULL, "w_sundermount_139_0.msh"},
	{0xD40D6CC5CFBEC051ULL, "gen00fl_aveline6140777.cl"},
	{0xD40DA2BCC6D74B46ULL, "40812353.ogg"},
	{0xD40E04A95E6BE784ULL, "vxht_mansion2b_2el_0_0.vlm"},
	{0xD40E681262E40273ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maga.phy"},
	{0xD40E7128699DF741ULL, "6035845_m.fxe"},
	{0xD40E91BAB3B9502CULL, "hlac_lt_0_lowtown_a3_n_c_0_0y.dds"},
	{0xD40E96F6098993EEULL, "vxw_woundedcoast_1mi_0_0.vlm"},
	{0xD40F332D1A0598DDULL, "335743514.ogg"},
	{0xD40F3430B43B52CFULL, "mrl351ip_demon_statue.ncs"},
	{0xD40F88B1BB97ECF9ULL, "sse_ss_profanemob_1.gda"},
	{0xD40FD469E506CB45ULL, "70664154.ogg"},
	{0xD40FEE05772C6D4BULL, "follower_banter6163152.cl"},
	{0xD410618950CD94B8ULL, "vxht_keep_3f8_0_0.vlm"},
	{0xD410FEF44F7A1AD1ULL, "6131262_m.fxe"},
	{0xD4110BB74EA3F47DULL, "hlac_w_blightlands_-65539_0y.dds"},
	{0xD41111F60DFEA8AEULL, "lt_0_lowtown_docks_a3_104_0.msh"},
	{0xD4111FF60DFEC0FFULL, "lt_0_lowtown_docks_a3_104_0.mao"},
	{0xD4112BF60DFED49CULL, "lt_0_lowtown_docks_a3_104_0.mmh"},
	{0xD41185D9E3FFB42BULL, "ave200_aveline_bar6116181.cl"},
	{0xD411891095485C06ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeenter.ani"},
	{0xD411FEAFD654D3BCULL, "6087463_m.fxe"},
	{0xD4124D3E13DAD53CULL, "fen210_fenris6108737.cl"},
	{0xD412720BB25064ADULL, "w_woundedcoast_38_0.anb"},
	{0xD412D297168167A3ULL, "vxca_caverns2_8ld_0_1.vlm"},
	{0xD412ED509276AB54ULL, "llac_ht_bartrand_n_3373_3.dds"},
	{0xD412F78B4AA26588ULL, "one100dg_vanard6054288.cl"},
	{0xD413FF5106E796FDULL, "sst_ss_circlemages_f_1.gda"},
	{0xD41420B91AB7A1BBULL, "vxw_encounter2_6e_0_0.vlm"},
	{0xD4148D90DF004265ULL, "6126846_m.fxe"},
	{0xD4155D5B566A021DULL, "zz_followers_banter6159144.cl"},
	{0xD4156F51538FB1DDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_loop.ani"},
	{0xD415AA410FD76E65ULL, "vxht_mansion2b_24v_0_0.vlm"},
	{0xD415CFF72ACD8396ULL, "m_mag161ar_cave.bnk"},
	{0xD4166A014D651DB7ULL, "387751156.ogg"},
	{0xD416A88F40412932ULL, "vxw_encounter2_camp_aq_0_0.vlm"},
	{0xD417639603733810ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic.mao"},
	{0xD417679603733F23ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic.mmh"},
	{0xD41775263915DEA7ULL, "6043503_m.fxe"},
	{0xD41775960373577DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic.msh"},
	{0xD4178577A92317FEULL, "803760684.ogg"},
	{0xD419281B105E33AFULL, "6092242_m.fxe"},
	{0xD419805B41C8C56EULL, "445881372.ogg"},
	{0xD41A059CA4A9CA27ULL, "1011964467.ogg"},
	{0xD41A0D72B2D2A97FULL, "vxht_mansion2b_1c2_0_0.vlm"},
	{0xD41A69BD047B0F4BULL, "llac_lt_0_lowtown_docks_n_-196610_6.dds"},
	{0xD41A85909D2DE358ULL, "llac_w_sundermount_-131072_0.dds"},
	{0xD41AFA4AD81C1AF9ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_lgta.phy"},
	{0xD41B1B41259E3799ULL, "hlac_lt_foundry_n_320_10y.dds"},
	{0xD41C9F383570604EULL, "vxht_mansion2b_1l2_0_0.vlm"},
	{0xD41CDB75251300D3ULL, "llac_ht_0_hightown_65535_2.dds"},
	{0xD41D6DFEFAA0ADCEULL, "ht_0_hightown_n_36_0.rml"},
	{0xD41E2C65181C8A22ULL, "692652484.ogg"},
	{0xD41E7B5869A0AE9CULL, "vxlt_0_lowtown_a2_n_c_e7x_0_0.vlm"},
	{0xD41EE1A9992704C2ULL, "sst_ss_varric.gda"},
	{0xD41F3F23E7C5A2FDULL, "6156113_m.fxe"},
	{0xD41F444F98D4A0B5ULL, "vxlt_0_lowtown_docks_c3w_0_0.vlm"},
	{0xD41F711095547177ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeenter.evt"},
	{0xD41F775B5E15FB84ULL, "6043588_m.fxe"},
	{0xD41FB02ADBC4D6B5ULL, "6104855_m.fxe"},
	{0xD4200212CA7BA0FFULL, "599866541.ogg"},
	{0xD4203E5100773ED9ULL, "611652332.ogg"},
	{0xD420B3B344A95722ULL, "708300817.ogg"},
	{0xD420E4E4A07BBEC2ULL, "beredines_ring.dds"},
	{0xD420F7FA15ABC01FULL, "lgt150_nat_were_supporter6110959.cl"},
	{0xD4211DBDA0A731B0ULL, "6084875_m.fxe"},
	{0xD421399C5B7F6BFEULL, "vxca_caverns2prc_4ti_0_0.vlm"},
	{0xD4213C689B0C1C90ULL, "dae000_gamlen6139444.cl"},
	{0xD421447408778023ULL, "vxca_caverns2_8ke_0_1.vlm"},
	{0xD4218DCDA05C92E1ULL, "332506277.ogg"},
	{0xD4226DFC7EA23C5EULL, "hlac_ht_0_hightown_a3_65535_1y.dds"},
	{0xD42292960CADBA3BULL, "vxw_encounter2_1d_0_0.vlm"},
	{0xD4229E590304C108ULL, "static_static_p41.psh.fxo.dx11"},
	{0xD422F802E30BA36FULL, "vxw_bonepitmining_a2_ot_0_0.vlm"},
	{0xD4241368D32AF190ULL, "ser270_templar6122824.cl"},
	{0xD42432A8E9924169ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_k_bk.ani"},
	{0xD424C25FC64D3C10ULL, "vxht_bartrand_n_42o_0_0.vlm"},
	{0xD4252913BA88867BULL, "6068015_f.fxe"},
	{0xD42539991C78A69CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_loop.gad"},
	{0xD42572109558DE00ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\wallchargeenter.gad"},
	{0xD425BBD933384867ULL, "vxlt_gallowstemplar_6qd_0_0.vlm"},
	{0xD425EB1FEF44028CULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_bdy_sw.mao"},
	{0xD4260AA9BA2262FDULL, "mag371_elsa6064700.cl"},
	{0xD4265F4A1F8CADE8ULL, "767720250.ogg"},
	{0xD426887F97627FEDULL, "6072242_m.fxe"},
	{0xD4269615D262B995ULL, "404053009.ogg"},
	{0xD427724C5AFE8DF6ULL, "one120_athenril6089428.cl"},
	{0xD4278BEDD97C4F3AULL, "hlac_ht_chantry_2534_29x.dds"},
	{0xD427B536D0BBB6CDULL, "72146203.ogg"},
	{0xD42868960381B448ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0sc_basic.phy"},
	{0xD42893ADBAA75F97ULL, "vxw_bonepit_1ki_0_0.vlm"},
	{0xD428B12879DAB014ULL, "vxlt_gallowstemplar_5bq_0_0.vlm"},
	{0xD4295AC329468425ULL, "hla_ht_0_hightown_n_-1_0x.dds"},
	{0xD429B6E4D807622AULL, "llac_lt_0_lowtown_docks_a3_n_-131075_0.dds"},
	{0xD42A177C912476FDULL, "llac_w_encounter2_camp_262148_0.dds"},
	{0xD42ABFAC9011F2B9ULL, "llac_ca_caverns2prc_5802_27.dds"},
	{0xD42B171493EB3B45ULL, "521542510.ogg"},
	{0xD42B4DA8E9988DEBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_k_bk.gad"},
	{0xD42B7AAFED75C4EEULL, "vxw_woundedcoast_1sk_0_0.vlm"},
	{0xD42C012B74152A27ULL, "6164558_m.fxe"},
	{0xD42C38991C7EC39AULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_pray_loop.ani"},
	{0xD42C4785115DE0ACULL, "348692947.ogg"},
	{0xD42C4C0129DBC920ULL, "371139167.ogg"},
	{0xD42C80F4F25D7EB0ULL, "136838276.ogg"},
	{0xD42C84620D59A62BULL, "lgt160_wk_soph_sophia6119827.cl"},
	{0xD42CF045E6772ED2ULL, "557009299.ogg"},
	{0xD42D6A63ADBDA7C2ULL, "hlac_w_bonepitdestroyed_a3_65541_0y.dds"},
	{0xD42DB026F38ABA85ULL, "dre000_stroud6036742.cl"},
	{0xD42DC2C4EC723C2AULL, "6042105_m.fxe"},
	{0xD42DD791E95056DBULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_leaves01_sa.dds"},
	{0xD42E2130B455E372ULL, "mrl351ip_demon_statue.utp"},
	{0xD42F0EB154C05429ULL, "fx_crescent01.mao"},
	{0xD42FC7C6D5B952B9ULL, "167790452.ogg"},
	{0xD42FFBB4E662BE01ULL, "gen00fl_aveline6105446.cl"},
	{0xD430254590728B2AULL, "hlac_ca_caverns1_mines_24421_2y.dds"},
	{0xD43042FE735C185EULL, "plt_lgt210pt_lowdown.plo"},
	{0xD430777C197314ABULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_female_warrior.anb"},
	{0xD43157CF28E8073BULL, "vxw_blightlands_fj_0_0.vlm"},
	{0xD43169DFD501BBEFULL, "vxw_bonepitmining_a2_jw_0_0.vlm"},
	{0xD43218E17AECF9BAULL, "6077167_m.fxe"},
	{0xD4322D8A30908945ULL, "w_sundermount_139_0.phy"},
	{0xD43234510A0A8928ULL, "ave200_aveline_bar6114111.cl"},
	{0xD4324AA8E99EA750ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_k_bk.evt"},
	{0xD432C007409CF851ULL, "6164963_m.fxe"},
	{0xD4331B24F46A433CULL, "vxw_encounter2_camp_rs_0_1.vlm"},
	{0xD433343CB8435490ULL, "vxht_bartrand_n_47l_0_0.vlm"},
	{0xD4334A85818EB704ULL, "180687594.ogg"},
	{0xD4335BE5D36FE71EULL, "198986777.ogg"},
	{0xD43375815A6278A5ULL, "544300265.ogg"},
	{0xD433C7C32377291EULL, "ss_slavers_f_1.wbd"},
	{0xD433D4C0962A1924ULL, "dae000_arishok6081346.cl"},
	{0xD433F8897CF43886ULL, "hlac_lt_undercityencounter_-1_42z.dds"},
	{0xD4343EEC247593D5ULL, "6164087_f.fxe"},
	{0xD434661F26663AD1ULL, "hlac_ht_0_hightown_65533_0z.dds"},
	{0xD434A5FE8F8BCBA3ULL, "950075683.ogg"},
	{0xD435555363DED61AULL, "dae340st_confrontation.stg"},
	{0xD4357963A99086D0ULL, "645228915.ogg"},
	{0xD4357C965D1B45F1ULL, "vxdr_0_deeproads_by6_0_0.vlm"},
	{0xD43613B7E38E08A3ULL, "vxlt_0_lowtown_h5i_0_0.vlm"},
	{0xD4364C648433D6E2ULL, "468659369.ogg"},
	{0xD436632505C2E80DULL, "mer320_hubert6083769.cl"},
	{0xD43675077805D6AAULL, "nat380_temmerin.cnv"},
	{0xD4367FD2F70B32DEULL, "w_bonepitdestroyed_a3_10_0.rml"},
	{0xD436D947283649F8ULL, "vxht_keep_6ex_0_0.vlm"},
	{0xD436FE184F899996ULL, "91653545.ogg"},
	{0xD437058AAC9D7817ULL, "vxw_bonepit_1lh_0_0.vlm"},
	{0xD4370AF60E1F6C67ULL, "lt_0_lowtown_docks_a3_104_0.phy"},
	{0xD43723056BD0C894ULL, "vxlt_gallowstemplar_5ov_0_0.vlm"},
	{0xD4374AE04659492DULL, "vxca_caverns2prc_43g_0_0.vlm"},
	{0xD437707EE087350EULL, "hlac_w_bonepit_0_13x.dds"},
	{0xD437B4D2B2319F5BULL, "959638584.ogg"},
	{0xD438D75FFE251CCBULL, "cod_bks_kirkwall_hist_4.utp"},
	{0xD4390E8A30967351ULL, "w_sundermount_139_0.rml"},
	{0xD4391EF3262D0B5FULL, "llac_lt_gallowstemplar_2394_7.dds"},
	{0xD4397B151CEDB13FULL, "103079496.ogg"},
	{0xD43982F2195C914EULL, "vxht_mansion2b_1z4_0_0.vlm"},
	{0xD43A570778098CD2ULL, "nat380_temmerin.bnk"},
	{0xD43A745F16D698ABULL, "vxdr_primevaltomb_551_0_0.vlm"},
	{0xD43B07D95D95A351ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_partycall.mmh"},
	{0xD43B13D95D95B7F2ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_partycall.mao"},
	{0xD43B17D440B75D1BULL, "6116010_m.fxe"},
	{0xD43B21D95D95CF0FULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_partycall.msh"},
	{0xD43B3BD9615DF851ULL, "460351004.ogg"},
	{0xD43B97CC92C58BBBULL, "335012616.ogg"},
	{0xD43BBE13130237FFULL, "l_dae350ar_sundermount.bnk"},
	{0xD43C104F34401986ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_prostitute_2.ani"},
	{0xD43C19A463D90CE7ULL, "hlac_ht_mansion2b_818_50z.dds"},
	{0xD43C2A4A5E64DE7CULL, "hlac_lt_0_lowtown_docks_n_-65538_9z.dds"},
	{0xD43CDFCF5FDE3A39ULL, "973193319.ogg"},
	{0xD43CE6607E499994ULL, "vxht_0_hightown_bsr_0_0.vlm"},
	{0xD43D1496F1F5607AULL, "6081669_m.fxe"},
	{0xD43E0A64F8C15BA7ULL, "6148761_m.fxe"},
	{0xD43E414C043EF5B9ULL, "vxlt_undercityserialkiller_1ku_0_0.vlm"},
	{0xD43E6C0EEA791A99ULL, "6118371_m.fxe"},
	{0xD43EA25C9BF9C74DULL, "explo_shot.dds"},
	{0xD43ED5D29B197002ULL, "gen00fl_carver6102676.cl"},
	{0xD43F04C9CB5C430FULL, "vxlt_0_lowtown_a2_n_gnl_0_0.vlm"},
	{0xD43F3DE70523F084ULL, "965413796.ogg"},
	{0xD43F4D4C44E75539ULL, "w_sundermount_109_0.msh"},
	{0xD43F574C44E76627ULL, "w_sundermount_109_0.mmh"},
	{0xD43F5B4C44E76D74ULL, "w_sundermount_109_0.mao"},
	{0xD43F5E85AA6ABCDFULL, "lgt310_mis_burn_murder26152761.cl"},
	{0xD43F8257F1B05113ULL, "hlac_ca_cavern1_haunted_24427_1y.dds"},
	{0xD440109F21C6658FULL, "hlac_ca_caverns1_mines_24423_4x.dds"},
	{0xD4406ACA8D938343ULL, "94894920.ogg"},
	{0xD440E15D3501A590ULL, "444906764.ogg"},
	{0xD4412BC7883D2368ULL, "vxlt_0_lowtown_docks_a3_n_d1k_0_0.vlm"},
	{0xD44152446651CDE8ULL, "dae000cr_merchant_access.utc"},
	{0xD4416E8583EDEE92ULL, "6039525_m.fxe"},
	{0xD44170596DA921A1ULL, "6106127_m.fxe"},
	{0xD44177015A845EE0ULL, "gen_im_acc_rng_act2_04.uti"},
	{0xD441B5820D4D1A36ULL, "follower_banter6143171.cl"},
	{0xD441B801FB2A2660ULL, "plt_one300pt_herb_dragon.plo"},
	{0xD44210836FDC4A4EULL, "lgt000_chanter6166480.cl"},
	{0xD4422542A923480CULL, "llac_lt_darktown_-1_0.dds"},
	{0xD442E50208338524ULL, "llac_lt_0_lowtown_docks_n_-4_0.dds"},
	{0xD444DF2FE73AE78CULL, "arc.mao"},
	{0xD4458C4ABD02BF2BULL, "llac_lt_0_lowtown_docks_a3_-196610_0.dds"},
	{0xD445B9A2007FD3B3ULL, "6160990_m.fxe"},
	{0xD445F3D6BACE94E9ULL, "dae000ip_crowd_ff6_talk8.utp"},
	{0xD446248191730731ULL, "255039324.ogg"},
	{0xD4467F7D9926E0C5ULL, "6187898_m.fxe"},
	{0xD4471DB154D4FFD7ULL, "fx_crescent01.dds"},
	{0xD447BBA93C6B9397ULL, "133134710.ogg"},
	{0xD447ECFC763C784FULL, "6185484_f.fxe"},
	{0xD447F77AFDB0A00AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_basket_sa.dds"},
	{0xD4482F99CC041CB0ULL, "art\\levels\\4cinematics\\dwarf\\_textures\\dwarfchair_s.dds"},
	{0xD4484D325E300701ULL, "llac_lt_0_lowtown_a3_n_c_65535_4.dds"},
	{0xD4489DE2C429EFA0ULL, "774322928.ogg"},
	{0xD448C6CEA783DFFFULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chain.msh"},
	{0xD448CCCEA783E9A1ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chain.mmh"},
	{0xD448D8CEA783FE42ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_chain.mao"},
	{0xD4496D3D9BC1CD49ULL, "vxlt_gallowscourtyard_n_75n_0_0.vlm"},
	{0xD44990F40B3C9963ULL, "6114423_m.fxe"},
	{0xD449C32CD3BAAC40ULL, "vxlt_hangedman_ru_0_0.vlm"},
	{0xD449F84F344C2EF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_prostitute_2.evt"},
	{0xD44A0FBC48F9DAB8ULL, "vxdr_primevaltomb_6ws_0_0.vlm"},
	{0xD44A24136A856900ULL, "hlac_lt_0_lowtown_65533_3x.dds"},
	{0xD44A4964B1F46162ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_hvyo_0n.dds"},
	{0xD44A5662709200D2ULL, "961512310.ogg"},
	{0xD44A897A32DC8BF7ULL, "dae000_arishok6067747.cl"},
	{0xD44A8B17F40B5C29ULL, "6086843_m.fxe"},
	{0xD44B0CF4FD81FF1FULL, "6140035_m.fxe"},
	{0xD44B15FFFB08E52EULL, "vxlt_gallowscourtyard_6xp_0_0.vlm"},
	{0xD44B3DAE74D1CD5CULL, "hlac_w_bonepitmining_a2_6_0z.dds"},
	{0xD44B80DBF22A5579ULL, "6120840_m.fxe"},
	{0xD44BE2C00142D131ULL, "vxht_bartrand_n_5wx_0_0.vlm"},
	{0xD44BF576FFF2E074ULL, "gen00fl_varric6076427.cl"},
	{0xD44C3349EB93D054ULL, "725113388.ogg"},
	{0xD44C63FDDF971686ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_lower_d.dds"},
	{0xD44D204CA249D8FAULL, "vxca_cave_2k8_0_0.vlm"},
	{0xD44D6327D336FB87ULL, "mrl150_merrill6088666.cl"},
	{0xD44D743DFB9F26B1ULL, "hlac_lt_warehouse_n_0_7x.dds"},
	{0xD44D9968262D7833ULL, "vxlt_0_lowtown_docks_n_cp3_0_0.vlm"},
	{0xD44E3418B0D7D62FULL, "223864518.ogg"},
	{0xD44E3B890CD4383BULL, "vxw_blightlands_ll_0_0.vlm"},
	{0xD44E80D6C6E3F0D6ULL, "hlac_w_bonepitdestroyed_a3_0_2x.dds"},
	{0xD44F1F52D6D0A2DCULL, "vxlt_foundry_n_zz_0_0.vlm"},
	{0xD44F5B88EEA4F841ULL, "338085.ogg"},
	{0xD44FF94F34509B80ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\dg_prostitute_2.gad"},
	{0xD4501AA2A1A0E2BEULL, "plt_zz_drept_debug.nss"},
	{0xD4506074D2CA476CULL, "gen00fl_aveline6116733.cl"},
	{0xD4509CB2CE4C1AB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_injured_dead_loop.gad"},
	{0xD450F3A968E80D7EULL, "6054678_m.fxe"},
	{0xD4513BE015A7E4F9ULL, "vxlt_0_lowtown_a2_n_gy3_0_0.vlm"},
	{0xD4513E70118C81BBULL, "160746524.ogg"},
	{0xD451880D5AE70A31ULL, "hlac_ca_caverns2prc_5802_45x.dds"},
	{0xD452A9C8B3695BF5ULL, "lt_undercitydungeon_probe_1wa.dds"},
	{0xD452E328F2355844ULL, "262786092.ogg"},
	{0xD45304A3208C5732ULL, "hlac_lt_0_lowtown_a2_n_c_131071_1z.dds"},
	{0xD4531A8351184624ULL, "vxdr_primevaltomb_75l_0_0.vlm"},
	{0xD45336904B23F801ULL, "845670056.ogg"},
	{0xD45372653BEFAD0BULL, "6173550_m.fxe"},
	{0xD453D751C1CA4020ULL, "vxlt_undercityencounter_13u_0_1.vlm"},
	{0xD45406BF4FBCF994ULL, "vxlt_gallowstemplar_5ut_0_0.vlm"},
	{0xD454691997FDF477ULL, "vxw_encounter2_1p_0_0.vlm"},
	{0xD4556C2321E90868ULL, "6089750_f.fxe"},
	{0xD45590199716BA21ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2_loop.gad"},
	{0xD455E9E29619567BULL, "lt_0_lowtown_a3_n_c_14_0.rml"},
	{0xD45688CD3E8999A6ULL, "vxlt_0_lowtown_a3_n_c_bxq_0_0.vlm"},
	{0xD456D375AE04C470ULL, "vxlt_hangedman_s6_0_0.vlm"},
	{0xD4570D349D91232FULL, "vxht_chantry_2dd_0_0.vlm"},
	{0xD4573B22D2A5BDDEULL, "vxht_0_hightown_av8_0_0.vlm"},
	{0xD457544A4862A9D3ULL, "6083478_f.fxe"},
	{0xD45756E96CE5D4BFULL, "hlac_lt_0_lowtown_n_-3_1z.dds"},
	{0xD457804B51773E85ULL, "mag311_nyssa6051154.cl"},
	{0xD4579DB2CE523B5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\lying_injured_dead_loop.ani"},
	{0xD458204126691240ULL, "6098935_m.fxe"},
	{0xD4583D21F07696A3ULL, "371074335.ogg"},
	{0xD45881993AEFF338ULL, "vxdr_primevaltomb_6zp_0_0.vlm"},
	{0xD4589FCCE109D712ULL, "6129299_m.fxe"},
	{0xD458B23890D818AEULL, "vxlt_backalley_g_1vc_0_2.vlm"},
	{0xD459363DD825B65FULL, "vxw_bonepitmining_a2_1uq_0_0.vlm"},
	{0xD459CA1A6235CA94ULL, "vxht_0_hightown_byp_0_0.vlm"},
	{0xD45A464B0573E896ULL, "hlac_w_encounter2_131074_0z.dds"},
	{0xD45A583083F09401ULL, "6152566_m.fxe"},
	{0xD45AC0DDB2A7C759ULL, "vxlt_gamlenshouse_n_b7_0_0.vlm"},
	{0xD45B6281CB25E4F4ULL, "vxht_keep_733_0_0.vlm"},
	{0xD45B9229943FF17AULL, "vxca_cave_2l9_0_0.vlm"},
	{0xD45C0068EF8A2205ULL, "vxht_0_hightown_n_bwh_0_0.vlm"},
	{0xD45C24BEA790B72FULL, "vxlt_0_lowtown_a2_n_by4_0_0.vlm"},
	{0xD45C345E178A3273ULL, "store_lgt_epic2_2.utm"},
	{0xD45C3889C585EEA6ULL, "hlac_lt_0_lowtown_docks_a3_65533_3z.dds"},
	{0xD45C6B0A73A78BBCULL, "86654515.ogg"},
	{0xD45C9119971CDACFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_v2_loop.ani"},
	{0xD45CAD65FECA50BBULL, "vxw_blightlands_qo_0_0.vlm"},
	{0xD45CDE2075F22C85ULL, "vxlt_gallowstemplar_n_43i_0_0.vlm"},
	{0xD45CF0E2961F80D3ULL, "lt_0_lowtown_a3_n_c_14_0.phy"},
	{0xD45D73A7B66D9FBEULL, "6155393_m.fxe"},
	{0xD45DDB25B1C19B8AULL, "6144221_m.fxe"},
	{0xD45E04531F3E83EEULL, "208934088.ogg"},
	{0xD45F17864C967410ULL, "gen00fl_fenris6113933.cl"},
	{0xD45FA31D9D85E5C4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_quadpeak.mmh"},
	{0xD45FA71D9D85EC17ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_quadpeak.mao"},
	{0xD45FB91D9D860B66ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_quadpeak.msh"},
	{0xD45FBF500B6C39B7ULL, "897495997.ogg"},
	{0xD4606EB54F7A3683ULL, "6067800_m.fxe"},
	{0xD46072280DBA0D27ULL, "143253281.ogg"},
	{0xD4607E2EB69E5509ULL, "805748849.ogg"},
	{0xD460B262BA3AA43DULL, "455266351.ogg"},
	{0xD460E93974E06EDCULL, "vxlt_gallowstemplar_f_5qs_0_0.vlm"},
	{0xD460F217E0090296ULL, "vxlt_0_lowtown_n_gcn_0_0.vlm"},
	{0xD461D393158952C4ULL, "follower_banter6161043.cl"},
	{0xD4620B258B44CF3CULL, "6037580_f.fxe"},
	{0xD46253F5751C76D5ULL, "w_bonepit_212_0.msh"},
	{0xD46255F5751C7A4BULL, "w_bonepit_212_0.mmh"},
	{0xD4625712ACA32BC3ULL, "vxlt_undercitydungeon_1wb_0_0.vlm"},
	{0xD46261F5751C8EE8ULL, "w_bonepit_212_0.mao"},
	{0xD4627674E3806438ULL, "325004862.ogg"},
	{0xD4627FE74335EE7BULL, "6082676_m.fxe"},
	{0xD46373D2F7320C56ULL, "w_bonepitdestroyed_a3_10_0.mao"},
	{0xD4637FD2F73220F5ULL, "w_bonepitdestroyed_a3_10_0.mmh"},
	{0xD46381D2F732246BULL, "w_bonepitdestroyed_a3_10_0.msh"},
	{0xD463A2FA2833CE8AULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.cess.gad"},
	{0xD463C3842AD21940ULL, "657694504.ogg"},
	{0xD463CBEDA4A3ED69ULL, "vxlt_hangedman_n_2mt_0_0.vlm"},
	{0xD463D52D4D3B820DULL, "286526763.ogg"},
	{0xD463DD618613CF81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_reaction.ani"},
	{0xD463F82FE755C29EULL, "arc.dds"},
	{0xD4643DA94A0E5E47ULL, "6176658_m.fxe"},
	{0xD46466EE56618B49ULL, "6167647_m.fxe"},
	{0xD464973D9FAA7D81ULL, "6088456_m.fxe"},
	{0xD464AB4C776E6B84ULL, "static_static_p21.psh.fxo.dx11.assembly"},
	{0xD464E53F694F351BULL, "hlac_lt_0_lowtown_a2_-2_3z.dds"},
	{0xD464F9EF0BCE9981ULL, "589890879.ogg"},
	{0xD4656771AAEDD9B9ULL, "mag100_cullen6082274.cl"},
	{0xD4657F118F873BAFULL, "vxht_chantry_2ag_0_0.vlm"},
	{0xD4665A2D68511036ULL, "woodrope_d.dds"},
	{0xD466695374F6F8A8ULL, "6086194_m.fxe"},
	{0xD467038B41188D1EULL, "6139856_m.fxe"},
	{0xD4671DC548F5157FULL, "6205419_m.fxe"},
	{0xD4672A25DC0CADA3ULL, "vxca_cavern1_haunted_eft_0_0.vlm"},
	{0xD4676BF43EE3062AULL, "hlac_ca_caverns1_24426_13y.dds"},
	{0xD4682D2D0780C0CEULL, "hlac_lt_0_lowtown_n_65536_9z.dds"},
	{0xD468856731487E27ULL, "hlac_dr_slavers_3207_1x.dds"},
	{0xD468A4131328F935ULL, "l_dae350ar_sundermount.wbd"},
	{0xD468A8901737EEDBULL, "vxdr_primevaltomb_34l_0_0.vlm"},
	{0xD468CC77D2CCB0F3ULL, "dae000_madam_lusine6151626.cl"},
	{0xD46996E2DA213F39ULL, "vxlt_undercityserialkiller_1zz_0_0.vlm"},
	{0xD4699C9D3F28BE64ULL, "vxw_woundedcoast_yb_0_0.vlm"},
	{0xD469E86186184E23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_reaction.gad"},
	{0xD46A584F0804822CULL, "llac_lt_gallowscourtyard_n_-65537_4.dds"},
	{0xD46A85FA2839BC21ULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.cess.evt"},
	{0xD46B404C450C7A5CULL, "w_sundermount_109_0.phy"},
	{0xD46B42042203AF1CULL, "690092367.ogg"},
	{0xD46B59CAB564AE9FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wrba_0t.dds"},
	{0xD46B769B1DD5561BULL, "dae300_vashothlead6062813.cl"},
	{0xD46BF3489705D69CULL, "gen00fl_isabela6078982.cl"},
	{0xD46C0AB31208CAE6ULL, "isa311_castillon6092128.cl"},
	{0xD46CB9C913CAC3E4ULL, "vxw_bonepit_1qt_0_0.vlm"},
	{0xD46CBEEBC13A794BULL, "673004111.ogg"},
	{0xD46CE45EF66FB532ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_basic_1.mao"},
	{0xD46D348D6FC2A491ULL, "vxca_caverns1_mines_img_0_0.vlm"},
	{0xD46DA24498CCFB45ULL, "6053548_m.fxe"},
	{0xD46DCBCA2A11CFD0ULL, "seb221_flora_harimann6187799.cl"},
	{0xD46DE5516794B028ULL, "vxht_mansion2_1c1_0_0.vlm"},
	{0xD46DF5FF6CBA024AULL, "gen00fl_sebastian6178368.cl"},
	{0xD46E6A29ED21498EULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_act_c.impa.ani"},
	{0xD46EDFBA1DE0077AULL, "6110516_m.fxe"},
	{0xD46F978C0D15A6A5ULL, "369184327.ogg"},
	{0xD4702BEF3B062B87ULL, "vxlt_0_lowtown_docks_n_cvi_0_0.vlm"},
	{0xD47060B16754AB44ULL, "vxw_sundermount_35m_0_0.vlm"},
	{0xD47061126336C0E8ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maga.mmh"},
	{0xD4707F126336F312ULL, "art\\characters\\playercharacter\\humanmale\\bdy_maga.msh"},
	{0xD470823ECFDE287AULL, "ser221_quentin6123307.cl"},
	{0xD470C561861E3108ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_knife_throw_reaction.evt"},
	{0xD4720D1D837857C1ULL, "vxca_caverns2_7qo_0_0.vlm"},
	{0xD472574C4512BF84ULL, "w_sundermount_109_0.rml"},
	{0xD472833F8CFAF0C6ULL, "588288706.ogg"},
	{0xD473271BA4A7CE34ULL, "346258495.ogg"},
	{0xD4736D410CAAEDA1ULL, "fen210_orana6109715.cl"},
	{0xD4737AA7A3B6E982ULL, "hightown_male_noble_gossip6171299.cl"},
	{0xD473A323BA17720DULL, "vxht_mansion2b_3gb_0_0.vlm"},
	{0xD473CFBA32B82BC0ULL, "242481946.ogg"},
	{0xD473F0EE817D542FULL, "vxht_chantry_2zf_0_0.vlm"},
	{0xD4742C231606CD94ULL, "225159893.ogg"},
	{0xD47433288B29158EULL, "art\\characters\\playercharacter\\humanmale\\hm_hat_vis.msh"},
	{0xD47441288B292DDFULL, "art\\characters\\playercharacter\\humanmale\\hm_hat_vis.mao"},
	{0xD47448043D7D918FULL, "dae000_gamlen6048419.cl"},
	{0xD4744D288B29427CULL, "art\\characters\\playercharacter\\humanmale\\hm_hat_vis.mmh"},
	{0xD4744F51B6CC605AULL, "67724781.ogg"},
	{0xD47498C01FAF2D0AULL, "vxw_woundedcoast_1kw_0_0.vlm"},
	{0xD474FD0D0F79AE19ULL, "vxlt_0_lowtown_docks_a3_d1u_0_0.vlm"},
	{0xD475308722D49E3FULL, "vxht_chantry_n_23n_0_0.vlm"},
	{0xD475533AF9DA928CULL, "828302063.ogg"},
	{0xD47563A1630DDD90ULL, "mag311_emile_de_launcet6043692.cl"},
	{0xD4756CBAF60B70B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_left.ani"},
	{0xD4758FA23AEB2DFFULL, "6208319_m.fxe"},
	{0xD475B7B3FB9CEEE3ULL, "404380055.ogg"},
	{0xD476EAF15956BEEBULL, "llac_lt_gallowscourtyard_-65536_3.dds"},
	{0xD476EFFB02276C4CULL, "gen00fl_fenris6127963.cl"},
	{0xD47702460144C75DULL, "6208397_m.fxe"},
	{0xD4770A8AFCF08EA8ULL, "798286369.ogg"},
	{0xD47763A0EF69CC9BULL, "6074453_m.fxe"},
	{0xD4777DFA28443970ULL, "art\\vfx\\combat\\vfx_m_spi_frc_aoe_p.cess.ani"},
	{0xD477E7BEDF900398ULL, "540997648.ogg"},
	{0xD4780E7A311ED6E4ULL, "vxw_woundedcoast_vc_0_0.vlm"},
	{0xD4781088B85FB9A8ULL, "hm_qmq_minion1.mor"},
	{0xD47812F936DE0584ULL, "6119823_m.fxe"},
	{0xD47817F0A6BE6845ULL, "vxca_cave_2k3_0_0.vlm"},
	{0xD4781BA0625956ACULL, "llac_lt_gallowstemplar_f_2391_0.dds"},
	{0xD47825EDAD6727AAULL, "532565547.ogg"},
	{0xD4788CA9857DC4DBULL, "6041031_m.fxe"},
	{0xD4791EEDA125E1E6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_lgta_0n.dds"},
	{0xD479A2DBE4BDD464ULL, "485102743.ogg"},
	{0xD479B848B3B57637ULL, "genip_pp_lti_sustained.utp"},
	{0xD479E21D6A7D4A92ULL, "299215068.ogg"},
	{0xD47A09EC8E929C70ULL, "6046860_m.fxe"},
	{0xD47A9EEE6B7F52F9ULL, "hlac_lt_0_lowtown_docks_n_-2_4y.dds"},
	{0xD47AA61090DDBE0FULL, "609968027.ogg"},
	{0xD47ACF7A58CDB3FEULL, "543083632.ogg"},
	{0xD47B2BA605C0DC64ULL, "vxw_bonepit_1zu_0_0.vlm"},
	{0xD47B6229ED2BC6DFULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_act_c.impa.evt"},
	{0xD47B66669EEB7AC3ULL, "6110572_f.fxe"},
	{0xD47C6260E0FCF351ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v1_c.impa.ani"},
	{0xD47CD03C964C1898ULL, "lt_0_lowtown_docks_a3_19_0.rml"},
	{0xD47CE7601CF4BCCCULL, "208641788.ogg"},
	{0xD47CEDBA86EAA612ULL, "vxca_caverns1_mines_cmn_0_0.vlm"},
	{0xD47CFC495887D058ULL, "gen00fl_fenris6099877.cl"},
	{0xD47D43B5E5279954ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_loop.ani"},
	{0xD47D513D5F88208EULL, "vxlt_0_lowtown_docks_a3_n_d0j_0_0.vlm"},
	{0xD47D88E88DF1FA20ULL, "vxlt_undercityserialkiller_18y_0_0.vlm"},
	{0xD47D91EFEB6476B1ULL, "gen_im_cns_cft_arn_07.uti"},
	{0xD47DD671C66B9542ULL, "vxlt_darktown_5gd_0_0.vlm"},
	{0xD47FF3DD20DCEFA0ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.mmh"},
	{0xD480120C98247277ULL, "vxw_blightlands_lx_0_0.vlm"},
	{0xD48080600A880697ULL, "720394987.ogg"},
	{0xD480B15DEF33AB13ULL, "vxlt_gallowstemplar_n_5x7_0_0.vlm"},
	{0xD480F8BC11F2EF0EULL, "dae000_bartrand6027457.cl"},
	{0xD481553BBEB44A98ULL, "hlac_lt_0_lowtown_docks_-196610_4x.dds"},
	{0xD4817D43428F3167ULL, "6090271_m.fxe"},
	{0xD482152BB35BFA44ULL, "456678026.ogg"},
	{0xD4824329ED31B0C8ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_act_c.impa.gad"},
	{0xD4826D60E10171F3ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v1_c.impa.gad"},
	{0xD4828A306D801C9AULL, "6115762_m.fxe"},
	{0xD483360FD1A71B23ULL, "vxht_chantry_n_2ad_0_0.vlm"},
	{0xD48374BAF617BBA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_left.evt"},
	{0xD4840E6118684562ULL, "vxw_blightlands_s8_0_0.vlm"},
	{0xD4841E221B0AC5F7ULL, "866576597.ogg"},
	{0xD4842F930FF519D7ULL, "zz_otl_debug6060962.cl"},
	{0xD48440B5E52DB2B6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_loop.gad"},
	{0xD48452002551EE39ULL, "hlac_ca_cavern1_haunted_24427_0x.dds"},
	{0xD48469902AD81C7FULL, "176967496.ogg"},
	{0xD4846ED875A4F60AULL, "6164403_m.fxe"},
	{0xD484B36800FC4B52ULL, "268384421.ogg"},
	{0xD484B52643FA9FD1ULL, "hlac_lt_0_lowtown_65538_0y.dds"},
	{0xD484D610308F3696ULL, "ob_ciga.mao"},
	{0xD4857341E48BAEC5ULL, "mrl210_merrill6116452.cl"},
	{0xD48591A374C57F33ULL, "gen00fl_aveline6106391.cl"},
	{0xD4859E0FC268318CULL, "vxlt_0_lowtown_docks_a3_bez_0_0.vlm"},
	{0xD485BAEBC5FAA88BULL, "hlac_lt_backalley_g_0_12z.dds"},
	{0xD4862E77051B1F19ULL, "vxlt_darktown_2vc_0_0.vlm"},
	{0xD4869101260331A1ULL, "6206935_m.fxe"},
	{0xD4875A089576181FULL, "vxdr_primevaltomb_7zv_0_0.vlm"},
	{0xD487B2F67B1B8A5AULL, "w_woundedcoast_19_0.anb"},
	{0xD487D21586581A44ULL, "vxca_caverns1_giq_0_0.vlm"},
	{0xD487D748D05DDE30ULL, "hlac_lt_hangedman_-1_16z.dds"},
	{0xD487DB3294350F9BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.ani"},
	{0xD487E5E2428817ABULL, "vxw_bonepit_j1_0_0.vlm"},
	{0xD48846BDE13EAC4EULL, "vxca_caverns1_gkd_0_1.vlm"},
	{0xD48855587C3FDB72ULL, "one100dg_nabil6049134.cl"},
	{0xD488758E91F4A7F1ULL, "155686525.ogg"},
	{0xD489040399F46609ULL, "6051595_m.fxe"},
	{0xD48909484A04928CULL, "522527720.ogg"},
	{0xD48928A29C93A204ULL, "vxlt_0_lowtown_a2_n_f5c_0_0.vlm"},
	{0xD4896A60E1078B58ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v1_c.impa.evt"},
	{0xD48995BAF61C5FF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_back_run_left.gad"},
	{0xD489AF101490B51CULL, "403210853.ogg"},
	{0xD48A53A57F81C535ULL, "dre000_cut_sibling_dies.bnk"},
	{0xD48A5BB5E5324C15ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\follower_hf_loop.evt"},
	{0xD48A730FD9625DF7ULL, "400772120.ogg"},
	{0xD48AA2286E372EBAULL, "6101514_m.fxe"},
	{0xD48AB58E070E2CB6ULL, "hlac_ca_cavern1_haunted_24426_13z.dds"},
	{0xD48AF5415D470942ULL, "dae000_arishok6068576.cl"},
	{0xD48B75D9EDEBE307ULL, "one100dg_vanard6054634.cl"},
	{0xD48BE6FEAC509E94ULL, "218670664.ogg"},
	{0xD48C3ED04AC2B85CULL, "vxlt_gallowstemplar_f_5bv_0_0.vlm"},
	{0xD48C799596A06167ULL, "319635691.ogg"},
	{0xD48CE312DA0C9FA6ULL, "6107111_f.fxe"},
	{0xD48D05F2D5000329ULL, "638747073.ogg"},
	{0xD48D6E843E8ED88CULL, "hlac_ht_0_hightown_a2_65533_0y.dds"},
	{0xD48D70A1D86E744DULL, "vxlt_0_lowtown_docks_cqu_0_0.vlm"},
	{0xD48D75A57F84348DULL, "dre000_cut_sibling_dies.cnv"},
	{0xD48DA05605B3A511ULL, "gen00fl_isabela_init6110224.cl"},
	{0xD48DD6DD20E8FBBFULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.anb"},
	{0xD48E3B38BF63EA22ULL, "gen00fl_isabela6135629.cl"},
	{0xD48E62E16F14F072ULL, "gen00fl_anders6119500.cl"},
	{0xD48E94853E72877AULL, "rock_group_02.dds"},
	{0xD48F0307CD7C44F3ULL, "6149084_f.fxe"},
	{0xD4909C0668A56F7FULL, "789602562.ogg"},
	{0xD490FC0D5F76351AULL, "6110862_m.fxe"},
	{0xD491446FCDCF5CF1ULL, "follower_banter6170774.cl"},
	{0xD49157E163FB77A2ULL, "art\\vfx\\combat\\vfx_str_leap_act_c.mmh"},
	{0xD4917C7A039B5E0AULL, "vxw_woundedcoast_1qq_0_0.vlm"},
	{0xD4920A3FF951729CULL, "vxlt_0_lowtown_a2_n_c_eop_0_0.vlm"},
	{0xD492803E0A5E5DE2ULL, "vxw_blightlands_t9_0_0.vlm"},
	{0xD493187F26281033ULL, "403908270.ogg"},
	{0xD4936F76A235AE23ULL, "57188532.ogg"},
	{0xD49375CB1202B070ULL, "zz_lgt_1_debug6140403.cl"},
	{0xD49394793700885CULL, "235378602.ogg"},
	{0xD493C7BF855A3227ULL, "163902118.ogg"},
	{0xD49418BDC9B13F66ULL, "llac_w_woundedcoast_5_0.dds"},
	{0xD4941C10BB114BB1ULL, "vxht_bartrand_n_5ns_0_0.vlm"},
	{0xD494467DD0AAB78DULL, "6048962_m.fxe"},
	{0xD494A7E4B6CBC053ULL, "424819296.ogg"},
	{0xD494E0FC2D2BF239ULL, "495354524.ogg"},
	{0xD494E332943FA792ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.evt"},
	{0xD494E4C517032AAAULL, "ach_comp_spec.dds"},
	{0xD4956E68219BB662ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_broken.msh"},
	{0xD4957068219BB918ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_broken.mmh"},
	{0xD4957C68219BCDBBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_broken.mao"},
	{0xD497227AB566B4C8ULL, "hlac_lt_warehouse_65537_5x.dds"},
	{0xD497AC843212CCCDULL, "and230_mage_girl6135817.cl"},
	{0xD497F2183DA4D351ULL, "6143087_m.fxe"},
	{0xD4980FA667E85398ULL, "6148708_m.fxe"},
	{0xD49838FA02076912ULL, "hlac_ht_chantry_n_2534_21y.dds"},
	{0xD498474F15F9AEBBULL, "hlac_lt_darktown_-1_32z.dds"},
	{0xD49854E203542C15ULL, "179553995.ogg"},
	{0xD4985DE42E45E442ULL, "hlac_w_encounter2_-131068_0x.dds"},
	{0xD49873E0C50F7632ULL, "524114371.ogg"},
	{0xD498BD9AACFEC98FULL, "dae000_gamlen6121563.cl"},
	{0xD4994FFC31819573ULL, "vxlt_warehouse_oy_0_0.vlm"},
	{0xD499FFA2A1DEED09ULL, "plt_zz_drept_debug.plo"},
	{0xD49A0D942B68A845ULL, "llac_lt_darktown_-1_50.dds"},
	{0xD49A2C288B49D947ULL, "art\\characters\\playercharacter\\humanmale\\hm_hat_vis.phy"},
	{0xD49A6CA271DE2B20ULL, "vxlt_undercityserialkiller_12g_0_0.vlm"},
	{0xD49AA501FB762EEFULL, "plt_one300pt_herb_dragon.nss"},
	{0xD49B0DA0E8B59258ULL, "llac_dr_slavers_3217_1.dds"},
	{0xD49B17BE186EB5E4ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_3z.dds"},
	{0xD49B7BB1F0EB5C56ULL, "darktown_male_elf_sta6174559.cl"},
	{0xD49C02329445FB25ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf_injured_leanr.gad"},
	{0xD49CECF40C1C392DULL, "6049730_m.fxe"},
	{0xD49CF6398AF6183BULL, "gen00fl_varric6118793.cl"},
	{0xD49D64AF1B4536DFULL, "vxht_playershouse_n_1w0_0_0.vlm"},
	{0xD49D6E93F161DC23ULL, "llac_lt_0_lowtown_docks_-262143_0.dds"},
	{0xD49D9517D31FDC13ULL, "vxlt_gallowstemplar_n_5v9_0_0.vlm"},
	{0xD49DAC4277721682ULL, "vxlt_0_lowtown_a2_gjr_0_0.vlm"},
	{0xD49E2AA09D89F7E4ULL, "605610905.ogg"},
	{0xD49EA1E2D26D0624ULL, "hlac_lt_gallowstemplar_7198_1z.dds"},
	{0xD49FFB8DF3721837ULL, "art\\levels\\races\\proxy\\props\\door\\prp_door.phy"},
	{0xD4A1463A6B5EFF28ULL, "518715669.ogg"},
	{0xD4A184748F24C3C1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose1.ani"},
	{0xD4A189F340604833ULL, "hlac_lt_0_lowtown_a3_n_c_-1_4z.dds"},
	{0xD4A1F029370F0228ULL, "559864939.ogg"},
	{0xD4A1F9D26C523D6BULL, "vxht_0_hightown_a3_c6b_0_0.vlm"},
	{0xD4A20ABB9EA35C92ULL, "vxht_0_hightown_a6t_0_0.vlm"},
	{0xD4A27C3E8DE9A9ABULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyn_0s.dds"},
	{0xD4A2B85AC2A0D00DULL, "6043862_m.fxe"},
	{0xD4A3AF5DBF6438BBULL, "251577257.ogg"},
	{0xD4A3F9428E8CAFD2ULL, "542925142.ogg"},
	{0xD4A425AC88CC2374ULL, "478975389.ogg"},
	{0xD4A4B31597529D11ULL, "vxlt_gallowstemplar_6st_0_0.vlm"},
	{0xD4A5B793CE3E500FULL, "6110481_m.fxe"},
	{0xD4A5E34AC9670848ULL, "rock_small_diffuse.dds"},
	{0xD4A60F8F1092379BULL, "284817272.ogg"},
	{0xD4A61880E51E9888ULL, "lt_0_lowtown_docks_125_0.mmh"},
	{0xD4A62480E51EAD2BULL, "lt_0_lowtown_docks_125_0.mao"},
	{0xD4A629E52C4F8EE6ULL, "975335547.ogg"},
	{0xD4A63680E51ECBB2ULL, "lt_0_lowtown_docks_125_0.msh"},
	{0xD4A64A02BC7180D7ULL, "6156735_m.fxe"},
	{0xD4A6A771951370D2ULL, "gen00fl_merrill6111278.cl"},
	{0xD4A73EBD8D8B2B6BULL, "hlac_lt_0_lowtown_a2_n_-131072_3y.dds"},
	{0xD4A7774667AC3697ULL, "vxw_woundedcoast_4z_0_0.vlm"},
	{0xD4A784644BD61AFCULL, "w_encounter3_161_0.mao"},
	{0xD4A790644BD62E9FULL, "w_encounter3_161_0.mmh"},
	{0xD4A796644BD63941ULL, "w_encounter3_161_0.msh"},
	{0xD4A7A12F7778BF33ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs101a_0t.dds"},
	{0xD4A7C1D92377ADF3ULL, "vxlt_warehouse_px_0_0.vlm"},
	{0xD4A88F748F2AF563ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose1.gad"},
	{0xD4A8B13C96711F50ULL, "lt_0_lowtown_docks_a3_19_0.anb"},
	{0xD4A8BE4E204A2950ULL, "545283155.ogg"},
	{0xD4A8DE7F63D443A0ULL, "vxlt_undercityserialkiller_17d_0_0.vlm"},
	{0xD4A968E92BAFB630ULL, "6151389_f.fxe"},
	{0xD4A9ABB69536CDF6ULL, "vxlt_darktown_4nn_0_0.vlm"},
	{0xD4A9BA5789E13A1BULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs05a01.msh"},
	{0xD4A9D85789E16D05ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs05a01.mmh"},
	{0xD4AA0CD3949EAEE2ULL, "gen00fl_carver6103640.cl"},
	{0xD4AA2C3CE404BD89ULL, "523124062.ogg"},
	{0xD4AA872F1A0EFA1AULL, "6142850_m.fxe"},
	{0xD4AAD70ED0ED36C2ULL, "hlac_lt_0_lowtown_a3_n_c_-3_5z.dds"},
	{0xD4AADDB84EBB91AFULL, "584154592.ogg"},
	{0xD4AAE1C21652552DULL, "vxlt_0_lowtown_docks_n_b5j_0_1.vlm"},
	{0xD4AB8475CE24CEE6ULL, "dae000_gamlen6131356.cl"},
	{0xD4ABAA5FF243B58AULL, "610478545.ogg"},
	{0xD4ABCD740966A95EULL, "hm_qun_commoner.mor"},
	{0xD4ABD4914B50B9C1ULL, "vxca_caverns2_7ec_0_0.vlm"},
	{0xD4ABDC70056CB959ULL, "ser191_emeric6104766.cl"},
	{0xD4ABE7EC9B99B544ULL, "hlac_lt_0_lowtown_a2_-65536_6y.dds"},
	{0xD4AC17229F88A1E5ULL, "6016503_m.fxe"},
	{0xD4AC44513F5C4347ULL, "95209338.ogg"},
	{0xD4AC6125C86A6DFFULL, "pro000_colonel_ewald6040575.cl"},
	{0xD4ACA4E8BE5EDE23ULL, "6177572_m.fxe"},
	{0xD4ACEC28F694C8EEULL, "6115453_f.fxe"},
	{0xD4ACF33F413D73D7ULL, "6100130_m.fxe"},
	{0xD4AD191FDC035671ULL, "6013412_m.fxe"},
	{0xD4AD771BE615ACE1ULL, "700807106.ogg"},
	{0xD4AD90FC5CA9C77AULL, "follower_barks6153762.cl"},
	{0xD4AD950D382FBDA7ULL, "549671298.ogg"},
	{0xD4AE0FC1AFAC24C1ULL, "6116428_f.fxe"},
	{0xD4AE1E7630B12084ULL, "6080192_m.fxe"},
	{0xD4AE3E1BAADF66E7ULL, "vxlt_0_lowtown_a2_n_eis_0_0.vlm"},
	{0xD4AE541DA8D5723EULL, "hlac_ht_keep_n_3736_0y.dds"},
	{0xD4AEFCE6EF0B5F80ULL, "vxht_0_hightown_a3_cax_0_0.vlm"},
	{0xD4AF6C748F30D848ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\mother_kneelingpose1.evt"},
	{0xD4AF8B648AC280B1ULL, "481688983.ogg"},
	{0xD4AF969E4E5AD172ULL, "80278997.ogg"},
	{0xD4AF99C4E9440C7CULL, "6068965_m.fxe"},
	{0xD4AFB5F67B3CE10DULL, "w_woundedcoast_19_0.mmh"},
	{0xD4AFB9F67B3CE85EULL, "w_woundedcoast_19_0.mao"},
	{0xD4AFC7F67B3CFF93ULL, "w_woundedcoast_19_0.msh"},
	{0xD4AFDEAC9EC06AEBULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_lgtc_0n.dds"},
	{0xD4AFE206911A20FBULL, "720708784.ogg"},
	{0xD4AFF03ED9F7FABCULL, "189617855.ogg"},
	{0xD4AFF8BB8AC7E214ULL, "rdr111dg_martin6014337.cl"},
	{0xD4B08AF26C927BD9ULL, "826892482.ogg"},
	{0xD4B1850E234F331CULL, "mcr310_argument6127662.cl"},
	{0xD4B188C47713E833ULL, "mer210pt_cavern_of_dead.ncs"},
	{0xD4B195E7BC3F0CB6ULL, "ser221_gascard6133671.cl"},
	{0xD4B1E1EB740DEC9CULL, "wm_ico_cave.dds"},
	{0xD4B24987EF3B7E82ULL, "6174770_m.fxe"},
	{0xD4B2A13D69724384ULL, "6093534_m.fxe"},
	{0xD4B31A51A1201E42ULL, "dae000cr_merchant_rare.utc"},
	{0xD4B39F8BD4D76E5DULL, "vxht_0_hightown_a3_b8o_0_0.vlm"},
	{0xD4B4F319CD993E64ULL, "vxw_bonepit_1ny_0_0.vlm"},
	{0xD4B5A7442D91DE05ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_006a_0n.dds"},
	{0xD4B5ADBD5D3163DFULL, "6142745_m.fxe"},
	{0xD4B5CA2B64D0ADC1ULL, "672871066.ogg"},
	{0xD4B5E92359A24F17ULL, "vxw_woundedcoast_1u_0_0.vlm"},
	{0xD4B6A42D0F395FB5ULL, "hlac_lt_0_lowtown_n_65537_1z.dds"},
	{0xD4B754EF0BB22A8BULL, "enemy_stop_combat.ncs"},
	{0xD4B7ABCBE03AA1C0ULL, "vxlt_gallowstemplar_n_2lk_0_0.vlm"},
	{0xD4B7AC8FF21DFA82ULL, "6087479_f.fxe"},
	{0xD4B86ABEC942FCF4ULL, "28719340.ogg"},
	{0xD4B91592DFC26470ULL, "llac_w_sundermount_-65539_0.dds"},
	{0xD4B997408B555655ULL, "mag330_walter6045966.cl"},
	{0xD4BA192056CA6ED4ULL, "vxht_keep_62q_0_0.vlm"},
	{0xD4BA4BAD78CDC917ULL, "241551214.ogg"},
	{0xD4BA56E1641E7B49ULL, "art\\vfx\\combat\\vfx_str_leap_act_c.anb"},
	{0xD4BAB5F67B46E886ULL, "w_woundedcoast_19_0.rml"},
	{0xD4BB80F171E547A3ULL, "hlac_dr_slavers_3217_12y.dds"},
	{0xD4BBBFA2F0580FE8ULL, "6071259_m.fxe"},
	{0xD4BC2AFFD66228D3ULL, "dre000_sibling_has_blight6036459.cl"},
	{0xD4BC8FACD6EAE9ADULL, "46658428.ogg"},
	{0xD4BCB9F139C3B2A1ULL, "dae000_surprises6151920.cl"},
	{0xD4BCFCD678E49CDEULL, "one110_meeran6089473.cl"},
	{0xD4BD1960C5014F55ULL, "llac_lt_0_lowtown_a2_n_131071_4.dds"},
	{0xD4BD1DFF242C5C1BULL, "fen110_fenris6117973.cl"},
	{0xD4BD58606DA79129ULL, "6080298_m.fxe"},
	{0xD4BD5DA39384EF67ULL, "6082283_m.fxe"},
	{0xD4BDD377129D5EFAULL, "673159591.ogg"},
	{0xD4BDE197BAD22665ULL, "6049811_m.fxe"},
	{0xD4BE12BEDF54B7CBULL, "vfx_rendertoshadowmap_p3.vsh.fxo.dx11"},
	{0xD4BE5CE4C0D9F530ULL, "580867921.ogg"},
	{0xD4BEA19870C7F95DULL, "lt_hangedman_n_mmap.dds"},
	{0xD4BEB86D4467B5BEULL, "mrl210_merrill6166854.cl"},
	{0xD4BFAA58B0FB089AULL, "vxlt_gallowscourtyard_n_5xp_0_0.vlm"},
	{0xD4BFCE978E87872FULL, "574571718.ogg"},
	{0xD4BFE28B65A85BBAULL, "6048433_m.fxe"},
	{0xD4BFE9E29673F248ULL, "lt_0_lowtown_a3_n_c_14_0.mmh"},
	{0xD4BFF5E2967406EBULL, "lt_0_lowtown_a3_n_c_14_0.mao"},
	{0xD4C007E296742572ULL, "lt_0_lowtown_a3_n_c_14_0.msh"},
	{0xD4C08718DD697266ULL, "lgt110_gxa_vgk_silver_order6117003.cl"},
	{0xD4C1185243AF4AC4ULL, "follower_banter6170166.cl"},
	{0xD4C14CF87030E1FDULL, "mag291ar_gallows_templar_fd.lst"},
	{0xD4C1DCED7F5FFC7DULL, "vxlt_hangedman_16i_0_0.vlm"},
	{0xD4C3031F06EB197DULL, "178680142.ogg"},
	{0xD4C3AE028B855C9BULL, "qcr220_orsino6061797.cl"},
	{0xD4C3C413B605024BULL, "233254310.ogg"},
	{0xD4C3E538E140AE22ULL, "dae020cr_hm_noble1.utc"},
	{0xD4C403241FCEC7C1ULL, "360423924.ogg"},
	{0xD4C41937F87EE737ULL, "2891595.ogg"},
	{0xD4C47F5BE6324A0BULL, "21192514.ogg"},
	{0xD4C4B2D327023AEBULL, "art\\vfx\\combat\\vfx_g_dual_att_right_c.impa.ani"},
	{0xD4C4D0E986532A6AULL, "6015853_m.fxe"},
	{0xD4C4E4D90AA04B60ULL, "557592272.ogg"},
	{0xD4C509B7F82798BAULL, "vxlt_blackemporium_t9_0_0.vlm"},
	{0xD4C5A7EA904A4109ULL, "520483731.ogg"},
	{0xD4C5FD6319FC9491ULL, "6209952_m.fxe"},
	{0xD4C73EFF2A5975E6ULL, "vxlt_blackemporium_rw_0_0.vlm"},
	{0xD4C7E02CFBC72928ULL, "lt_0_lowtown_a3_n_c_5_0.msh"},
	{0xD4C7EF4CF9B9CC9CULL, "hlac_lt_0_lowtown_a2_65535_1z.dds"},
	{0xD4C7F22CFBC747B9ULL, "lt_0_lowtown_a3_n_c_5_0.mao"},
	{0xD4C7FE2CFBC75C52ULL, "lt_0_lowtown_a3_n_c_5_0.mmh"},
	{0xD4C806C5693C5800ULL, "6114051_m.fxe"},
	{0xD4C844F87036F2ECULL, "mag291ar_gallows_templar_fd.ncs"},
	{0xD4C860A780DEC7B0ULL, "hlac_lt_gallowstemplar_f_2384_6z.dds"},
	{0xD4C8A7D95FD1B65FULL, "6017872_m.fxe"},
	{0xD4C8BA0399FEEB90ULL, "qun221_petrice6083694.cl"},
	{0xD4C9C1B0BDE105A2ULL, "art\\vfx\\textures\\level_fx\\vfx_bird_01.dds"},
	{0xD4CA31F0113B9C2BULL, "vxca_caverns2prc_9ox_0_0.vlm"},
	{0xD4CAC4184E986721ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_sa.dds"},
	{0xD4CB21D58ECB97E7ULL, "vxlt_0_lowtown_a2_n_ewm_0_0.vlm"},
	{0xD4CB39C7727C9B4AULL, "zz_pro_debug6059754.cl"},
	{0xD4CBAEDD96BFAFFEULL, "vxlt_warehouse_n_15x_0_0.vlm"},
	{0xD4CBC42811ABD362ULL, "226262190.ogg"},
	{0xD4CBEDDE45BC1C8DULL, "52956503.ogg"},
	{0xD4CC0CAFF278B963ULL, "134384002.ogg"},
	{0xD4CC1398C5FBD44BULL, "hlac_ht_0_hightown_-196607_0z.dds"},
	{0xD4CC388EAAD8E360ULL, "708171071.ogg"},
	{0xD4CC4E2F1DCD9A88ULL, "vxw_woundedcoast_1dx_0_0.vlm"},
	{0xD4CD16EE2222B8F9ULL, "vxca_caverns2_8kw_0_1.vlm"},
	{0xD4CD1CE5B0F65D64ULL, "6151311_m.fxe"},
	{0xD4CD7328C02DB7A1ULL, "qaesnf1.arl"},
	{0xD4CDB4AEFED96157ULL, "6152278_m.fxe"},
	{0xD4CE0275BAC03B6FULL, "6108259_m.fxe"},
	{0xD4CE24DA0CDA3CB0ULL, "llac_lt_0_lowtown_a2_0_1.dds"},
	{0xD4CE50F5F434191EULL, "character_character_p108.psh.fxo.dx11.assembly"},
	{0xD4CF1AC7A05C3C8BULL, "vxw_bonepitdestroyed_a3_220_0_1.vlm"},
	{0xD4CFB78469421012ULL, "633881391.ogg"},
	{0xD4CFC0C6D86F21A4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_45walkway.mmh"},
	{0xD4CFC4C6D86F28F7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_45walkway.mao"},
	{0xD4CFD6C6D86F4746ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_45walkway.msh"},
	{0xD4D04D9156EEE9CAULL, "gen00fl_sebastian6195839.cl"},
	{0xD4D070342CB23050ULL, "960172980.ogg"},
	{0xD4D07E99E3BE05F8ULL, "123747916.ogg"},
	{0xD4D0A18B33C21575ULL, "art\\characters\\weapons\\shields\\w_shd_twra.msh"},
	{0xD4D0A38B33C218EBULL, "art\\characters\\weapons\\shields\\w_shd_twra.mmh"},
	{0xD4D0AF2CB7E85E46ULL, "vxlt_0_lowtown_hb1_0_0.vlm"},
	{0xD4D0AF8B33C22C88ULL, "art\\characters\\weapons\\shields\\w_shd_twra.mao"},
	{0xD4D0AF96A6A5399EULL, "6082372_m.fxe"},
	{0xD4D0ED405A0D1EA5ULL, "vxlt_darktown_4a4_0_0.vlm"},
	{0xD4D135CA3D7EF76DULL, "llac_ht_0_hightown_a2_-196609_0.dds"},
	{0xD4D19AD3270C9C62ULL, "art\\vfx\\combat\\vfx_g_dual_att_right_c.impa.evt"},
	{0xD4D257D5EA983021ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_act_c.anb"},
	{0xD4D2B9CB61010A4CULL, "vxlt_gallowstemplar_f_4uj_0_0.vlm"},
	{0xD4D2DE38D23263CBULL, "6151239_m.fxe"},
	{0xD4D36B4711A2EA74ULL, "qcr000_arishok6166997.cl"},
	{0xD4D3CBD3037894BAULL, "follower_banter6136445.cl"},
	{0xD4D416AA213EDA70ULL, "556047139.ogg"},
	{0xD4D452432974DDD5ULL, "716125892.ogg"},
	{0xD4D481EC9A2B0031ULL, "static_static_p60.psh.fxo.dx11"},
	{0xD4D4D3FB1FF3F09CULL, "ch2emmultnt.psh.fxo"},
	{0xD4D5C6955033E0D5ULL, "448727599.ogg"},
	{0xD4D69C3B152305BBULL, "hlac_dr_primevaltomb_658_12y.dds"},
	{0xD4D6A11F5B61B65CULL, "629750140.ogg"},
	{0xD4D6AB68BF31A653ULL, "307688144.ogg"},
	{0xD4D6D50173959786ULL, "fen110_fenris6096228.cl"},
	{0xD4D7DA5971BA31B8ULL, "dae000_bartrand6027499.cl"},
	{0xD4D815D18BEA8BC3ULL, "art\\levels\\races\\design\\_textures\\prp_chesta_n.dds"},
	{0xD4D8593B88411272ULL, "188931318.ogg"},
	{0xD4D88398C8B4FD00ULL, "llac_lt_0_lowtown_131072_1.dds"},
	{0xD4D884D49A57D85AULL, "ser270_emeric6149342.cl"},
	{0xD4D899D32712B975ULL, "art\\vfx\\combat\\vfx_g_dual_att_right_c.impa.gad"},
	{0xD4D8A1172E9445CBULL, "vxlt_0_lowtown_a2_n_c_e28_0_0.vlm"},
	{0xD4D8CCA3DC78A2B1ULL, "ach_comp_craft_all.dds"},
	{0xD4D915B8D7F10427ULL, "571561659.ogg"},
	{0xD4D92984B1EF6894ULL, "t3_dey_te3.tnt"},
	{0xD4D92D924BCE5779ULL, "vxlt_0_lowtown_docks_n_c52_0_0.vlm"},
	{0xD4D97AFC19196230ULL, "6103051_m.fxe"},
	{0xD4D993B280C1B067ULL, "vxlt_0_lowtown_a2_n_exl_0_0.vlm"},
	{0xD4DA1042A67850F7ULL, "12826310.ogg"},
	{0xD4DA1F7992651C58ULL, "hlac_lt_backalley_n_65537_7x.dds"},
	{0xD4DA64F68770F511ULL, "6097197_m.fxe"},
	{0xD4DAC00C0FC3B308ULL, "vxw_woundedcoast_1oy_0_0.vlm"},
	{0xD4DB2698FDA06409ULL, "694384610.ogg"},
	{0xD4DB6879BAC0EFBCULL, "6139713_m.fxe"},
	{0xD4DB8E311870B2E7ULL, "dae200_bran6094787.cl"},
	{0xD4DB9C97A57D1B33ULL, "fen_mod_item_acq.ncs"},
	{0xD4DD1A872D233957ULL, "683058911.ogg"},
	{0xD4DD30DA76E8E15DULL, "vxca_caverns1_mines_dno_0_0.vlm"},
	{0xD4DD3C505864544EULL, "6119289_m.fxe"},
	{0xD4DD7ECAB669B83EULL, "6166214_m.fxe"},
	{0xD4DE61DD337FBF83ULL, "vxht_bartrand_n_6jz_0_0.vlm"},
	{0xD4DEB1F93CFCB595ULL, "dre000_vampire6034312.cl"},
	{0xD4DECCF824C746EDULL, "static_static_p15.psh.fxo.dx11"},
	{0xD4DEF408717CEE9DULL, "hlac_lt_0_lowtown_a2_n_131072_4y.dds"},
	{0xD4DEFD610FB20959ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyo_0s.dds"},
	{0xD4DF2173989B521EULL, "6085272_m.fxe"},
	{0xD4DF731B0DE639F2ULL, "llac_ht_chantry_n_2534_33.dds"},
	{0xD4DFE00C9ED20F18ULL, "6135003_m.fxe"},
	{0xD4E018F19620EBABULL, "6017814_m.fxe"},
	{0xD4E038D5EAA4389AULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_act_c.mmh"},
	{0xD4E0454B658E8B66ULL, "899487225.ogg"},
	{0xD4E0CE127715BB5DULL, "gen00fl_anders6111413.cl"},
	{0xD4E1930C4B427579ULL, "llac_lt_gamlenshouse_n_-1_5.dds"},
	{0xD4E1A6E88F67C19CULL, "vxlt_gamlenshouse_n_3m_0_0.vlm"},
	{0xD4E1AB76C60F52EAULL, "hlac_lt_0_lowtown_a3_n_c_-65536_3x.dds"},
	{0xD4E1B7C6F66AFA4AULL, "art\\characters\\playercharacter\\humanmale\\hlf_masc.mao"},
	{0xD4E1BBC6F66B0099ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masc.mmh"},
	{0xD4E1C5C6F66B1187ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masc.msh"},
	{0xD4E23D6D8B6AB7EEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sandrune_n.dds"},
	{0xD4E29500BE077C27ULL, "808294167.ogg"},
	{0xD4E29F8D7ABB9E72ULL, "6116562_f.fxe"},
	{0xD4E2EE57B70B8C18ULL, "rdr111dg_martin6054579.cl"},
	{0xD4E485CF069D273DULL, "lgt110_orz_harrowmont6172613.cl"},
	{0xD4E4A19D34CB02F7ULL, "mcr310_party_banter6087798.cl"},
	{0xD4E4A22961FEB5DBULL, "rdr111dg_martin6055069.cl"},
	{0xD4E4B27943270B56ULL, "vxht_mansion2_1xo_0_1.vlm"},
	{0xD4E57DC6D3A6105EULL, "307216511.ogg"},
	{0xD4E5A075011AC976ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2tileweb.mao"},
	{0xD4E5AC75011ADD15ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2tileweb.mmh"},
	{0xD4E5AE75011AE08BULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_2tileweb.msh"},
	{0xD4E5B0A8DE218E5DULL, "6100073_m.fxe"},
	{0xD4E5D9F97AC7E379ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medc_0n.dds"},
	{0xD4E5E8C387AF437DULL, "hlac_lt_0_lowtown_docks_n_-65540_1x.dds"},
	{0xD4E609396AC3F0C5ULL, "fen110st_fenris_hates_mages.stg"},
	{0xD4E6165BA085CC0DULL, "docks_male_human_worker_sta6175983.cl"},
	{0xD4E665990C896357ULL, "6083428_m.fxe"},
	{0xD4E6D08DF3ADD18FULL, "art\\levels\\races\\proxy\\props\\door\\prp_door.mao"},
	{0xD4E6DC8DF3ADE62CULL, "art\\levels\\races\\proxy\\props\\door\\prp_door.mmh"},
	{0xD4E6E28DF3ADF01EULL, "art\\levels\\races\\proxy\\props\\door\\prp_door.msh"},
	{0xD4E7659631BDD451ULL, "vxht_0_hightown_a3_bz1_0_0.vlm"},
	{0xD4E79F6F3DC46FF9ULL, "vxlt_0_lowtown_docks_n_c01_0_0.vlm"},
	{0xD4E7BCCD7298075BULL, "483468150.ogg"},
	{0xD4E7FD065451A8C4ULL, "vxlt_darktown_2tx_0_0.vlm"},
	{0xD4E812B2F3F841CBULL, "6039307_f.fxe"},
	{0xD4E8135324694AEEULL, "hlac_lt_undercityserialkiller_-1_3y.dds"},
	{0xD4E931E901B9CB88ULL, "vxw_woundedcoast_1jv_0_0.vlm"},
	{0xD4E9691E42B48B34ULL, "576650250.ogg"},
	{0xD4E9E61D95FA39EFULL, "vxca_caverns1_ii4_0_0.vlm"},
	{0xD4EA5EE53AC17B2CULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzladder.msh"},
	{0xD4EA64E53AC1851EULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzladder.mmh"},
	{0xD4EA70E53AC199B5ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzladder.mao"},
	{0xD4EA80BB0ADF28D1ULL, "vxlt_gallowsprison_n_64b_0_0.vlm"},
	{0xD4EA8733ECC86715ULL, "hlac_ca_caverns1_mines_24423_12y.dds"},
	{0xD4EAF6B135FC26E6ULL, "gen00fl_fenris6106674.cl"},
	{0xD4EB348E14424413ULL, "vxlt_0_lowtown_docks_n_bcs_0_0.vlm"},
	{0xD4EB834687E73A02ULL, "vxht_0_hightown_5jd_0_0.vlm"},
	{0xD4EB92F136CAF8EEULL, "hlac_ca_caverns2prc_5802_5y.dds"},
	{0xD4EBB144084BBCDDULL, "aip_coteriemember_base.gda"},
	{0xD4EBBBE1BA876E02ULL, "6051048_m.fxe"},
	{0xD4EBDB5EB22EA084ULL, "6066414_f.fxe"},
	{0xD4EBFDA8426C73A1ULL, "fen310_fenris6144524.cl"},
	{0xD4EC32F826AEF05FULL, "vxlt_undercityserialkiller_19i_0_0.vlm"},
	{0xD4EC980E7BF6819AULL, "6042141_m.fxe"},
	{0xD4ECFC27265DF2AEULL, "qun120cr_javaris6044195.cl"},
	{0xD4ED02ECE042C32BULL, "vxw_bonepit_76_0_0.vlm"},
	{0xD4ED2EACACFFB61CULL, "968997039.ogg"},
	{0xD4ED55F870562BDAULL, "mag291ar_gallows_templar_fd.are"},
	{0xD4EDA0B14DBB4835ULL, "20873151.ogg"},
	{0xD4EDD0FA3DF94FA5ULL, "vxlt_darktown_4k2_0_0.vlm"},
	{0xD4EF22E89DD170BAULL, "dae000_merchant_rare_em6177950.cl"},
	{0xD4F05BC2C80E5060ULL, "vxlt_0_lowtown_ewq_0_0.vlm"},
	{0xD4F05CF226D03E55ULL, "6136335_m.fxe"},
	{0xD4F11D7F1E9C8CE9ULL, "art\\characters\\creatures\\dragon\\animation\\dblw_lp_v1.gad"},
	{0xD4F123121E6E2200ULL, "218017258.ogg"},
	{0xD4F1615A6B9AE509ULL, "771397021.ogg"},
	{0xD4F1E03DFAD0D06AULL, "lt_0_lowtown_docks_a3_167_0.mao"},
	{0xD4F1E43DFAD0D6B9ULL, "lt_0_lowtown_docks_a3_167_0.mmh"},
	{0xD4F1EE3DFAD0E7A7ULL, "lt_0_lowtown_docks_a3_167_0.msh"},
	{0xD4F1F6E4CC99DD65ULL, "hlac_ca_cavern1_haunted_24425_11x.dds"},
	{0xD4F209E5B2BD58E9ULL, "405722193.ogg"},
	{0xD4F22CF6F39D2380ULL, "105949668.ogg"},
	{0xD4F24878096F6AC8ULL, "fxc_dark_smoke.mao"},
	{0xD4F2A6FA54B3B371ULL, "dae000_light_tombs_boss6179452.cl"},
	{0xD4F32079AE448E4FULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_003a_0s.dds"},
	{0xD4F3329407EEB317ULL, "694852085.ogg"},
	{0xD4F370BBE7AD5BF3ULL, "6175325_m.fxe"},
	{0xD4F42285D017A6DDULL, "6105373_m.fxe"},
	{0xD4F469B7BD53510BULL, "615850648.ogg"},
	{0xD4F4F967ED478763ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb5.mmh"},
	{0xD4F50767ED479EBDULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb5.msh"},
	{0xD4F50D12EB604BD6ULL, "w_woundedcoast_201_0.msh"},
	{0xD4F51712EB605CD4ULL, "w_woundedcoast_201_0.mmh"},
	{0xD4F51B12EB6063E7ULL, "w_woundedcoast_201_0.mao"},
	{0xD4F576B51B5CC7D0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsfloor.msh"},
	{0xD4F584B51B5CDFEAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsfloor.mmh"},
	{0xD4F588B51B5CE621ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02ruinsfloor.mao"},
	{0xD4F5D77323B3ECD1ULL, "vxht_0_hightown_a3_bq0_0_0.vlm"},
	{0xD4F6474B7477BA1BULL, "hlac_w_encounter2_327684_0y.dds"},
	{0xD4F66B4969690089ULL, "84594816.ogg"},
	{0xD4F6A477F235A0D3ULL, "dae120cr_amb_dwarf_1.utc"},
	{0xD4F6DE3738A5DEE9ULL, "zz_followers_banter6158886.cl"},
	{0xD4F6E2AF86237346ULL, "lgt320_mis_resp_swordhunter6150335.cl"},
	{0xD4F7594408492C0BULL, "llac_lt_0_lowtown_a3_n_c_65535_6.dds"},
	{0xD4F7678B556AA647ULL, "6128064_m.fxe"},
	{0xD4F7EBB3E72E70B3ULL, "vxlt_0_lowtown_a2_n_bxo_0_0.vlm"},
	{0xD4F7EE7F1EA25B37ULL, "art\\characters\\creatures\\dragon\\animation\\dblw_lp_v1.ani"},
	{0xD4F80F4D4B599FFAULL, "6095612_f.fxe"},
	{0xD4F84969B8A86524ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden8.mmh"},
	{0xD4F85F69B8A88AC6ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_ferelden8.msh"},
	{0xD4F955A70DEA01CDULL, "vxht_keep_n_5ss_0_0.vlm"},
	{0xD4FA19A73FFEBC6FULL, "6162655_m.fxe"},
	{0xD4FAB0CC47D392A0ULL, "460405535.ogg"},
	{0xD4FAC576F03E059FULL, "6099979_m.fxe"},
	{0xD4FB6986F6A79925ULL, "dae000ip_crowd_lm4_talk14.utp"},
	{0xD4FBC3BBEDF0BB5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_roll_shoulders.gad"},
	{0xD4FBE70725F5FB20ULL, "244239611.ogg"},
	{0xD4FC11FE9BD76426ULL, "1015026834.ogg"},
	{0xD4FC45585886FD8BULL, "art\\characters\\creatures\\highdragon\\animation\\blendtree\\highdragon_overlay.anb"},
	{0xD4FC63A1415CC889ULL, "6139778_m.fxe"},
	{0xD4FDEC872E7F64B8ULL, "6134581_m.fxe"},
	{0xD4FF614BB9A1D80EULL, "6114817_m.fxe"},
	{0xD4FFF1338B235CA6ULL, "356921341.ogg"},
	{0xD50030A8328A2035ULL, "vxw_blightlands_13v_0_0.vlm"},
	{0xD5003D94EB30D8D1ULL, "6115479_m.fxe"},
	{0xD50062E7EDE9B6B1ULL, "vxht_mansion2b_1v3_0_0.vlm"},
	{0xD5014D0BFB18721DULL, "733252588.ogg"},
	{0xD5023804C3080674ULL, "734672650.ogg"},
	{0xD502C6BBEDF6DF71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_roll_shoulders.evt"},
	{0xD502E3F550B15E62ULL, "mcr370_cut_player_survives6098275.cl"},
	{0xD50307C4B46959F3ULL, "gen00fl_aveline6004170.cl"},
	{0xD5038E834B8C1D0EULL, "hlac_w_blightlands_-1_0y.dds"},
	{0xD503B3F358EB5331ULL, "361657739.ogg"},
	{0xD50478B3E9CC67EAULL, "vxdr_primevalentrance_2l6_0_2.vlm"},
	{0xD5052D3DDD975DBCULL, "dae000_flemeth6000802.cl"},
	{0xD505F712A6370843ULL, "6164983_m.fxe"},
	{0xD506014BD8155EC7ULL, "vxlt_ruinedhouse_n_h3_0_0.vlm"},
	{0xD506079E4FFAAC99ULL, "hlac_lt_0_lowtown_a2_n_0_0y.dds"},
	{0xD506724228B32C18ULL, "323944180.ogg"},
	{0xD50682246F093B9DULL, "static_depth.psh.fxo"},
	{0xD506B4B05295FAE0ULL, "vxca_cave_1fr_0_0.vlm"},
	{0xD506B7EB61C8AFB0ULL, "6113638_m.fxe"},
	{0xD507A82400285742ULL, "137931396.ogg"},
	{0xD507B1D793863E43ULL, "sf_glyph.vsh.fxo"},
	{0xD508418CD84F1581ULL, "6170889_m.fxe"},
	{0xD508A12B0C587846ULL, "zz_applyvfx.ncs"},
	{0xD508C5A65E37C204ULL, "hlac_lt_0_lowtown_a2_n_1_9x.dds"},
	{0xD50916B20A9B215FULL, "vxlt_undercityserialkiller_13k_0_0.vlm"},
	{0xD50995D53410E2E0ULL, "11981620.ogg"},
	{0xD509D39E0C11267CULL, "zz_prmpt_debug6198264.cl"},
	{0xD50A25382C17EB28ULL, "vxlt_0_lowtown_a2_fex_0_0.vlm"},
	{0xD50A2EEA9CC41588ULL, "static_static_p36.psh.fxo.dx11.assembly"},
	{0xD50A99644C2ABBE4ULL, "w_encounter3_161_0.phy"},
	{0xD50AB3A91E59B658ULL, "1057813401.ogg"},
	{0xD50AB9CD6C0BAA6BULL, "656028921.ogg"},
	{0xD50AE35DD5E623C2ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic.phy"},
	{0xD50AE588231729D1ULL, "6058871_m.fxe"},
	{0xD50B04DA9F865693ULL, "vxht_keep_7a6_0_0.vlm"},
	{0xD50B1A10778D13D8ULL, "zz_vault_debug6200359.cl"},
	{0xD50B30C6541171EFULL, "6110290_m.fxe"},
	{0xD50BBD7A8B982281ULL, "682197460.ogg"},
	{0xD50BF01F722829D6ULL, "dre000_sandal6033768.cl"},
	{0xD50CFF0C7974E242ULL, "615132299.ogg"},
	{0xD50D9F2D71E578B0ULL, "lowtown_male_lowlife1_sta6164954.cl"},
	{0xD50E4906093F9517ULL, "llac_ht_0_hightown_2_0.dds"},
	{0xD50E4FC064D28E6CULL, "6136705_m.fxe"},
	{0xD50E6D99B09F8D04ULL, "vxlt_hangedman_rq_0_0.vlm"},
	{0xD50F2686984587C0ULL, "971780008.ogg"},
	{0xD50FC126C7D36BA4ULL, "plt_lgt320pt_streets.plo"},
	{0xD50FD943E0D3E548ULL, "gen00fl_aveline6114094.cl"},
	{0xD50FDEBBEE019240ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_roll_shoulders.ani"},
	{0xD50FEE1E34B931AEULL, "44729624.ogg"},
	{0xD510167A3FFAA6DFULL, "6095510_f.fxe"},
	{0xD51043AA8CA78227ULL, "qun211tr_braggart_amb.ncs"},
	{0xD510E7F26EDA6AECULL, "gen00fl_isabela_init6075939.cl"},
	{0xD510FD7AD932CB8CULL, "dae000cr_grand_cleric6187924.cl"},
	{0xD51108475985A9FCULL, "599567102.ogg"},
	{0xD511166C2127BBCAULL, "282310433.ogg"},
	{0xD511D2F7E74A921FULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumg.msh"},
	{0xD511D8F7E74A9CC1ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumg.mmh"},
	{0xD511E4F7E74AB162ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriumg.mao"},
	{0xD512197432B7936EULL, "301286354.ogg"},
	{0xD51273D434ABAAE3ULL, "578475821.ogg"},
	{0xD512AE072CFB3E6CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic.phy"},
	{0xD512BB2D07A01DD1ULL, "vxht_0_hightown_a3_bk6_0_0.vlm"},
	{0xD512CEF857B0F653ULL, "912982068.ogg"},
	{0xD51325C1AE988C20ULL, "mag210_arianni6021202.cl"},
	{0xD513E52D63155763ULL, "864348350.ogg"},
	{0xD513FE82BEAE7D96ULL, "one100dg_vanard6054295.cl"},
	{0xD514FE7621A4F4DFULL, "vxw_bonepitmining_a2_16d_0_0.vlm"},
	{0xD515152C19D856F8ULL, "152982153.ogg"},
	{0xD515268D448C1360ULL, "vxca_cave_1as_0_0.vlm"},
	{0xD5153205D763838BULL, "vxdr_primevaltomb_910_0_1.vlm"},
	{0xD51656E0D18956F2ULL, "848341376.ogg"},
	{0xD5167863584EF68DULL, "zz_followers_banter6158994.cl"},
	{0xD5168A24EA248D93ULL, "vxlt_0_lowtown_docks_n_brt_0_0.vlm"},
	{0xD516DCE7589B5B5BULL, "hlac_lt_0_lowtown_a2_n_c_-2_4z.dds"},
	{0xD516E27976A050F0ULL, "136858836.ogg"},
	{0xD516FFCF0291FA1CULL, "256335208.ogg"},
	{0xD517AFC489EE4636ULL, "dae000_orsino6085415.cl"},
	{0xD517E1841D538547ULL, "lt_0_lowtown_n_20_0_ll.mmh"},
	{0xD517E80DDF87BF22ULL, "922214686.ogg"},
	{0xD517F769838FBC4BULL, "667348045.ogg"},
	{0xD51821D2ED1C5DF4ULL, "vxlt_hangedman_27l_0_0.vlm"},
	{0xD51872CAD72B1542ULL, "hla_ht_0_hightown_n_-131072_0x.dds"},
	{0xD5189D757DC1A5A3ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_dbl_c.impa.evt"},
	{0xD5196BB1197B40B6ULL, "vxw_bonepitmining_a2_7i_0_0.vlm"},
	{0xD519E910E4ED154BULL, "6069640_m.fxe"},
	{0xD51A080AB4291169ULL, "hlac_ht_chantry_n_2534_5z.dds"},
	{0xD51AE48FB1F8ABDDULL, "48798696.ogg"},
	{0xD51AF6194119A456ULL, "follower_banter6166978.cl"},
	{0xD51B501D347F9E87ULL, "vxdr_primevaltomb_6zr_0_1.vlm"},
	{0xD51B63A01068B80AULL, "133382073.ogg"},
	{0xD51C6D66F38507A0ULL, "vxlt_undercityserialkiller_1ol_0_0.vlm"},
	{0xD51D1C554E0711EAULL, "gen00fl_isabela6075293.cl"},
	{0xD51D1D3C3D9D850BULL, "art\\characters\\creatures\\mubari\\animation\\sit_p.ani"},
	{0xD51D4C3F4FE29D2FULL, "vxca_caverns2_817_0_1.vlm"},
	{0xD51DAF0FF26137D5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_upper_d.dds"},
	{0xD51E7521F0F5680EULL, "6047410_m.fxe"},
	{0xD51E8FA8A30636D5ULL, "vxlt_hangedman_n_sz_0_0.vlm"},
	{0xD51EE4F5A889C456ULL, "ave331_post_fight6138818.cl"},
	{0xD51F0A1F13B3B435ULL, "isa120_hayder6074987.cl"},
	{0xD51F76757DC78224ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_dbl_c.impa.gad"},
	{0xD51FE94D6D293A39ULL, "565754379.ogg"},
	{0xD5201840EA461C92ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medb_0n.dds"},
	{0xD520355AA506A043ULL, "ave331_jeven6138771.cl"},
	{0xD5206F240D519B84ULL, "art\\characters\\playercharacter\\humanmale\\bdy_magc.msh"},
	{0xD52085240D51C126ULL, "art\\characters\\playercharacter\\humanmale\\bdy_magc.mmh"},
	{0xD520923D02A72BE7ULL, "6025154_m.fxe"},
	{0xD521294C18378FA8ULL, "mag291cr_crowd_elf_4.utc"},
	{0xD52173C1D1F1E9A3ULL, "vxlt_0_lowtown_a2_n_c_hlg_0_0.vlm"},
	{0xD52186E387A8C67BULL, "hlac_lt_0_lowtown_n_-131072_4z.dds"},
	{0xD521AD4E10E32CB8ULL, "zz_lgt_1_debug6188678.cl"},
	{0xD521D35DD43463F9ULL, "plt_lgt320pt_mis_resp_deman.nss"},
	{0xD52256B5666ECE8EULL, "dae000ip_crowd_lm1_talk14.utp"},
	{0xD522FAF0A24B1125ULL, "lgt110_gxa_coa_spy6117463.cl"},
	{0xD522FD4BDA226C37ULL, "6039549_f.fxe"},
	{0xD5237D6A063175D1ULL, "735361543.ogg"},
	{0xD52397E782D1D2C9ULL, "hlac_dr_0_deeproads_5842_6z.dds"},
	{0xD523DBB0E32452E8ULL, "rog_pinningshot.dds"},
	{0xD523EBFF2355A948ULL, "hlac_ht_mansion2_818_0x.dds"},
	{0xD523FD2CFC15B94DULL, "lt_0_lowtown_a3_n_c_5_0.phy"},
	{0xD52484B758933EA4ULL, "gen00fl_merrill6130112.cl"},
	{0xD5251B8DA108ADA8ULL, "510330332.ogg"},
	{0xD525EA8B8FE1C76BULL, "864898541.ogg"},
	{0xD52685757DCDBA02ULL, "art\\vfx\\combat\\vfx_glb_phys_imp_dbl_c.impa.ani"},
	{0xD526E39CC6519667ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_stand_r.dds"},
	{0xD5271E661F3C3604ULL, "6124143_f.fxe"},
	{0xD527D1288A2A67E4ULL, "6042283_f.fxe"},
	{0xD527DAE21F6B536CULL, "6168117_m.fxe"},
	{0xD527F3D789D17B82ULL, "hlac_dr_primevaltomb_11985_1y.dds"},
	{0xD52873D99AB9F37FULL, "dae100_alistair_drunk6135150.cl"},
	{0xD528BA5DD43A57D2ULL, "plt_lgt320pt_mis_resp_deman.plo"},
	{0xD528CE27EDA62E2CULL, "306667268.ogg"},
	{0xD52900F3BC7E9BDCULL, "vxlt_gallowstemplar_n_5wq_0_0.vlm"},
	{0xD5294F1BA46E3772ULL, "834829525.ogg"},
	{0xD52992C16ED2C5A3ULL, "6094561_m.fxe"},
	{0xD529F492A66DE75EULL, "nat380_temmerin6067384.cl"},
	{0xD529FE2CFC1A26F9ULL, "lt_0_lowtown_a3_n_c_5_0.rml"},
	{0xD52A7C240D5A06F1ULL, "art\\characters\\playercharacter\\humanmale\\bdy_magc.phy"},
	{0xD52AC56972476A22ULL, "99263238.ogg"},
	{0xD52B00E16A42A079ULL, "6115354_m.fxe"},
	{0xD52B053C3DA99982ULL, "art\\characters\\creatures\\mubari\\animation\\sit_p.evt"},
	{0xD52B0B22F725DD2CULL, "zz_ave_debug6134023.cl"},
	{0xD52BEB76BC2078CAULL, "vxca_caverns1_icr_0_1.vlm"},
	{0xD52CCC6B10F23E98ULL, "hlac_lt_0_lowtown_-65535_6z.dds"},
	{0xD52DF3641DEC3BD4ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi01.phy"},
	{0xD52E9B3D5B146B9DULL, "vxht_0_hightown_aqv_0_0.vlm"},
	{0xD52EC33F31A73049ULL, "vxht_mansion2_3cr_0_0.vlm"},
	{0xD52FBE5DD6050176ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic.mao"},
	{0xD52FCA5DD6051515ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic.mmh"},
	{0xD52FCC5DD605188BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic.msh"},
	{0xD53074F6B7475271ULL, "gen_im_cft_reg_f1.uti"},
	{0xD530AB66775BFEF4ULL, "vxca_caverns1_mines_fn0_0_0.vlm"},
	{0xD532043C3DAFB695ULL, "art\\characters\\creatures\\mubari\\animation\\sit_p.gad"},
	{0xD53261877E2884F3ULL, "298817327.ogg"},
	{0xD5331D1AD5C263CDULL, "vxht_keep_n_5go_0_0.vlm"},
	{0xD5334546F384A020ULL, "w_sundermount_35_0.phy"},
	{0xD5338B973749F6F2ULL, "6116543_m.fxe"},
	{0xD5338DDA41B5ECF6ULL, "457964573.ogg"},
	{0xD533A8DCE583A899ULL, "hlac_lt_foundry_n_320_3y.dds"},
	{0xD533C6CF2B2F4E53ULL, "86423794.ogg"},
	{0xD5353EDA475C7668ULL, "pro000_captain_wright.cnv"},
	{0xD5356F03F77B8D15ULL, "482265202.ogg"},
	{0xD53602DF1D51AB0EULL, "lt_0_lowtown_a2_n_c_42_0_lc.mmh"},
	{0xD5364F6AFD6771B6ULL, "vxw_bonepitmining_a2_9k_0_0.vlm"},
	{0xD5367EC186AE3C96ULL, "gencr_profane_a.utc"},
	{0xD53840C93A7550CAULL, "dae011_wench_mob6149863.cl"},
	{0xD53860DA475EE5D0ULL, "pro000_captain_wright.bnk"},
	{0xD538D0641DF607D1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi01.msh"},
	{0xD538E0E437756F1EULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.dura.evt"},
	{0xD538EA641DF6338FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi01.mmh"},
	{0xD538EA7C3E7678CBULL, "lt_warehouse_5_0.anb"},
	{0xD53909C4660844CDULL, "vxdr_slavers_1ao_0_0.vlm"},
	{0xD53979855FF44C7FULL, "vxlt_0_lowtown_a2_gcd_0_0.vlm"},
	{0xD53986CC89BB50CDULL, "vxdr_primevalentrance_1j6_0_0.vlm"},
	{0xD539AF0A755365D5ULL, "hlac_lt_0_lowtown_docks_n_-131075_0z.dds"},
	{0xD539C94A69124181ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cantryanimationbase.gad"},
	{0xD53A037D35147FF8ULL, "6126784_m.fxe"},
	{0xD53A0C860F3F1950ULL, "vxht_0_hightown_a3_c3s_0_1.vlm"},
	{0xD53A2480E4AC3E67ULL, "zz_vrc_debug6200210.cl"},
	{0xD53A4890E4D4D0F5ULL, "345983654.ogg"},
	{0xD53A6446F38AF3D8ULL, "w_sundermount_35_0.rml"},
	{0xD53A7CE813EFC424ULL, "vxlt_0_lowtown_a3_n_c_gc3_0_0.vlm"},
	{0xD53AC6FD2FF79A4DULL, "vxlt_0_lowtown_docks_c5i_0_0.vlm"},
	{0xD53B729095C1557BULL, "mag170_hugh6094563.cl"},
	{0xD53BF86EFF540694ULL, "vxdr_primevaltomb_6qm_0_0.vlm"},
	{0xD53C077C19F3827DULL, "vxw_encounter2_camp_pm_0_0.vlm"},
	{0xD53C12D41BB6FB65ULL, "759008123.ogg"},
	{0xD53C194C74F6AFB7ULL, "zz_seb_debug6215750.cl"},
	{0xD53C228763CC2370ULL, "hlac_dr_0_deeproads_17039_0z.dds"},
	{0xD53C48462AC38988ULL, "vxdr_dungeon_1t2_0_1.vlm"},
	{0xD53CE7E9DDA163BCULL, "15258519.ogg"},
	{0xD53D8D440A52D22BULL, "vxca_caverns1_k6k_0_0.vlm"},
	{0xD53DFEFDD5C0A755ULL, "atl_controller_dxt5_00.dds"},
	{0xD53E0B08B862062BULL, "6103114_f.fxe"},
	{0xD53E36764A412B8CULL, "387250425.ogg"},
	{0xD53E48BE5C5C150BULL, "469211394.ogg"},
	{0xD53EF7E4377A0249ULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.dura.gad"},
	{0xD53F3ECDAA729BD2ULL, "zz_followers_banter6157990.cl"},
	{0xD53FDDBD4A9467CCULL, "6016765_f.fxe"},
	{0xD540CA4A6918622FULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cantryanimationbase.ani"},
	{0xD5410082F44B3331ULL, "455838299.ogg"},
	{0xD5411D0A90FB09D4ULL, "gen00fl_varric6104937.cl"},
	{0xD5415E648B575247ULL, "6147250_m.fxe"},
	{0xD5418EF7C7B87C4DULL, "vxht_keep_n_5hn_0_0.vlm"},
	{0xD54243AF8D798682ULL, "one3510cr_dead_fighter_1.utc"},
	{0xD542831304E38DDEULL, "gen_im_acc_rng_act2_02.uti"},
	{0xD542C2BAACB29D15ULL, "vxht_0_hightown_c8k_0_0.vlm"},
	{0xD5431034B337FF8DULL, "vxlt_gallowstemplar_n_37n_0_0.vlm"},
	{0xD54316471B30D20CULL, "pro000_captain_wright6004376.cl"},
	{0xD5434739DB91C58BULL, "105861482.ogg"},
	{0xD543BCC6F6BDCC86ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masc.phy"},
	{0xD54464304B4934FEULL, "691863647.ogg"},
	{0xD544A1AAD2FBF1E7ULL, "vxlt_0_lowtown_docks_n_a0m_0_0.vlm"},
	{0xD54516CD229E5DD7ULL, "vxlt_gallowscourtyard_722_0_0.vlm"},
	{0xD545328FA70F298DULL, "vxlt_gallowstemplar_f_3sb_0_0.vlm"},
	{0xD54555BB9573B473ULL, "mag100_cullen6093748.cl"},
	{0xD545C8E4377FD097ULL, "art\\vfx\\combat\\vfx_c_pride_fire_p.dura.ani"},
	{0xD54658CE76046B3FULL, "6171794_m.fxe"},
	{0xD5473FDA8F5F4298ULL, "hlac_w_sundermount_-4_0z.dds"},
	{0xD547B39D5DDAE88EULL, "434193710.ogg"},
	{0xD547DC59F69E2ED6ULL, "vxlt_undercityserialkiller_11t_0_0.vlm"},
	{0xD547EBB1B6ACE1DAULL, "one300_zevran6089102.cl"},
	{0xD5480A1B6F88C7D8ULL, "vxlt_undercityencounter_1q6_0_0.vlm"},
	{0xD5483EC4716B3B38ULL, "6103077_m.fxe"},
	{0xD54922FD4A988517ULL, "vxlt_0_lowtown_docks_a3_c14_0_0.vlm"},
	{0xD549615183C7A53AULL, "vxlt_undercityserialkiller_1qx_0_0.vlm"},
	{0xD549AE2C3240EFDBULL, "948957720.ogg"},
	{0xD549BA0CA4342182ULL, "6111937_m.fxe"},
	{0xD549C0191FF6818FULL, "hlac_lt_0_lowtown_a2_-3_0x.dds"},
	{0xD549EC32A2F8F5C2ULL, "character_rendertodepth_p4.psh.fxo.dx11.assembly"},
	{0xD549F71A2C92D3D0ULL, "6092228_f.fxe"},
	{0xD54A0B23B2B5989AULL, "dre000_anders6045606.cl"},
	{0xD54A4BA63D9B9493ULL, "hlac_w_encounter2_camp_262148_0z.dds"},
	{0xD54A79590BE99AFDULL, "vxw_encounter2_camp_kl_0_0.vlm"},
	{0xD54AAE0D2C9C9943ULL, "6166916_m.fxe"},
	{0xD54AF801AFF9B8D5ULL, "101514748.ogg"},
	{0xD54B2BCFC2899B46ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_3x.dds"},
	{0xD54BAC68C4B88B73ULL, "vxht_keep_n_6up_0_0.vlm"},
	{0xD54C3CB18BB64560ULL, "hlac_ca_caverns2_5802_1x.dds"},
	{0xD54C744639E60DB7ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfalll.mmh"},
	{0xD54C784639E614C4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfalll.mao"},
	{0xD54C8A4639E63329ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandfalll.msh"},
	{0xD54CB9E4312D675FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edgebase.msh"},
	{0xD54CBFE4312D7101ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edgebase.mmh"},
	{0xD54CCBE4312D85A2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edgebase.mao"},
	{0xD54CD626C8073DEBULL, "plt_lgt320pt_streets.nss"},
	{0xD54DB21F01B91C0FULL, "376187799.ogg"},
	{0xD54DB3105822FB6DULL, "6216286_m.fxe"},
	{0xD54F4EE7639AB473ULL, "vxlt_house_g6_0_0.vlm"},
	{0xD54F6116F62C8292ULL, "vxht_0_hightown_arx_0_0.vlm"},
	{0xD54FA344037806F7ULL, "character_character_p53.psh.fxo"},
	{0xD54FBAD8D554AF55ULL, "6131218_f.fxe"},
	{0xD550659DD25BBA5BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillare.mmh"},
	{0xD5507F92FFD1DDD9ULL, "682654792.ogg"},
	{0xD550839DD25BED45ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillare.msh"},
	{0xD5508449FE03C9DFULL, "vxdr_0_deeproads_ew9_0_0.vlm"},
	{0xD550BAE8CC50E3C3ULL, "38259725.ogg"},
	{0xD5511612EBAEB9CFULL, "w_woundedcoast_201_0.phy"},
	{0xD5526B1B90FE68A8ULL, "qun221_petrice_allied_after.stg"},
	{0xD552A241E07803B4ULL, "hlac_w_bonepit_262145_0z.dds"},
	{0xD552ECEA756DCD55ULL, "plt_cod_ltr_junk_propaganda.plo"},
	{0xD55338AE092ED148ULL, "dae111st_dog_player.stg"},
	{0xD5536E3AF6635433ULL, "6076738_m.fxe"},
	{0xD5536F12C1A08DD4ULL, "gen00fl_aveline6140789.cl"},
	{0xD5537423F849F4A7ULL, "6140202_m.fxe"},
	{0xD55385072D328F37ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic.mmh"},
	{0xD55389072D329644ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic.mao"},
	{0xD5539715657AEF94ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_left.gad"},
	{0xD5539B072D32B4A9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wi_basic.msh"},
	{0xD553DA256E073BFEULL, "6143523_m.fxe"},
	{0xD554CD9836526493ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_f.ani"},
	{0xD554E9BE99D4D1D6ULL, "27326088.ogg"},
	{0xD5550D479F4C88EAULL, "713449947.ogg"},
	{0xD555571D60FAB496ULL, "6086909_m.fxe"},
	{0xD55595A2470DEC31ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robc_0d.dds"},
	{0xD55634DABB5D69A0ULL, "vxlt_undercityserialkiller_1sh_0_0.vlm"},
	{0xD55665B752AAF511ULL, "6172779_m.fxe"},
	{0xD5568772C765D0DBULL, "hlac_lt_undercitydungeon_-1_11x.dds"},
	{0xD556A67CCB5F974CULL, "134938039.ogg"},
	{0xD556CB0208E5880EULL, "w_bonepit_95_0.msh"},
	{0xD556D5C272908183ULL, "gencr_qunari_archer.utc"},
	{0xD556D90208E5A05FULL, "w_bonepit_95_0.mao"},
	{0xD556E50208E5B4FCULL, "w_bonepit_95_0.mmh"},
	{0xD5585106FFBFCAA9ULL, "826091825.ogg"},
	{0xD5586246F3A3EDB5ULL, "w_sundermount_35_0.msh"},
	{0xD5586446F3A3F12BULL, "w_sundermount_35_0.mmh"},
	{0xD5587046F3A405C8ULL, "w_sundermount_35_0.mao"},
	{0xD558D831F01471D1ULL, "zz_dog_debug6134428.cl"},
	{0xD558EB35FDDFB37DULL, "vxw_encounter2_camp_fk_0_0.vlm"},
	{0xD5595F5ADA4CC4ABULL, "602713059.ogg"},
	{0xD5599A555CC8EC04ULL, "224465896.ogg"},
	{0xD559AB1F8DF7AC45ULL, "6075262_f.fxe"},
	{0xD559FF76BBD61DE7ULL, "hlac_w_woundedcoast_196615_0z.dds"},
	{0xD55A519613027A28ULL, "zz_followers_banter6158931.cl"},
	{0xD55A76343E605FA6ULL, "mrl150_merrill6070600.cl"},
	{0xD55A7DEE9E4DC561ULL, "6170113_m.fxe"},
	{0xD55AA61565812872ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_left.ani"},
	{0xD55AC34ED5104CCDULL, "rdr111dg_harbormaster6055082.cl"},
	{0xD55ADAA368D3EB1BULL, "zz_dae_debug6045450.cl"},
	{0xD55ADC84CF322D5FULL, "669270645.ogg"},
	{0xD55AE32C499976EDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_staffa_d.dds"},
	{0xD55AEC8714FE5EAAULL, "6083487_m.fxe"},
	{0xD55B1ADB2693D850ULL, "lt_0_lowtown_docks_a3_n_131_0.mao"},
	{0xD55B1EDB2693DF63ULL, "lt_0_lowtown_docks_a3_n_131_0.mmh"},
	{0xD55B2CDB2693F6BDULL, "lt_0_lowtown_docks_a3_n_131_0.msh"},
	{0xD55B3B3599CA4BA3ULL, "vxlt_0_lowtown_a2_n_c_hxk_0_0.vlm"},
	{0xD55B4FE5B9B039AAULL, "6144306_m.fxe"},
	{0xD55B9A15A102EBCAULL, "vxdr_0_deeproads_cvm_0_0.vlm"},
	{0xD55CC89591975903ULL, "clf260_harley6093820.cl"},
	{0xD55CF69DAA9085ADULL, "6111585_m.fxe"},
	{0xD55CF760B2D0CFFFULL, "6144138_m.fxe"},
	{0xD55D1171F7777BDEULL, "6051901_m.fxe"},
	{0xD55D2BCEB3E10037ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdgc_0s.dds"},
	{0xD55D6712F514D8E3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\prp_secretwheeldoor_d.dds"},
	{0xD55D7D80B0BE460DULL, "373624527.ogg"},
	{0xD55D92A2397C682AULL, "6103383_m.fxe"},
	{0xD55DBF4FD5D690CFULL, "6189013_f.fxe"},
	{0xD55DD2F3E8229B12ULL, "vxht_0_hightown_auy_0_0.vlm"},
	{0xD55E79206C989D7AULL, "975530119.ogg"},
	{0xD55E88804411CE1AULL, "vxht_mansion2b_1y5_0_0.vlm"},
	{0xD55ED17D26A2BAD3ULL, "337802295.ogg"},
	{0xD55F33294A9970ABULL, "follower_barks6154997.cl"},
	{0xD55F5370287246EDULL, "hlac_ht_keep_3736_3x.dds"},
	{0xD55FE1CD98568570ULL, "hlac_dr_slavers_3215_3z.dds"},
	{0xD55FE2B880B2574CULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2char.msh"},
	{0xD55FE8B880B2613EULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2char.mmh"},
	{0xD55FF4B880B275D5ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm2char.mao"},
	{0xD560491E22BFF64AULL, "218951590.ogg"},
	{0xD561559B3490503CULL, "585725439.ogg"},
	{0xD561A0BC0D13A88DULL, "6090314_m.fxe"},
	{0xD561A4072DDE695CULL, "vxlt_darktown_4t7_0_1.vlm"},
	{0xD56247904953C3A9ULL, "qun221_viscount6070128.cl"},
	{0xD562A6EC368D1307ULL, "gen00fl_varric6104415.cl"},
	{0xD562B6337D419058ULL, "72822063.ogg"},
	{0xD56306C7316ADBE6ULL, "hlac_ht_0_hightown_-131072_0x.dds"},
	{0xD5631B9DAF4C161FULL, "6150078_m.fxe"},
	{0xD5634F2BB78088F9ULL, "gen00fl_anders6202360.cl"},
	{0xD5636B853B26D816ULL, "924555839.ogg"},
	{0xD5637B352169303EULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.impa.evt"},
	{0xD56394071B95DA68ULL, "6151386_m.fxe"},
	{0xD563A5F988A7899DULL, "409930563.ogg"},
	{0xD564329FF29DE582ULL, "6133109_f.fxe"},
	{0xD5649A84464DC38EULL, "6205010_m.fxe"},
	{0xD564E77C3E9BAE17ULL, "lt_warehouse_5_0.rml"},
	{0xD56530AA8CEFA27EULL, "qun211tr_braggart_amb.utt"},
	{0xD56557F63D2B1555ULL, "vxht_mansion_n_5bt_0_0.vlm"},
	{0xD565D7BB573E9B27ULL, "dae000_flemeth6001350.cl"},
	{0xD565F61806A0B2F0ULL, "dae000_flemeth6000965.cl"},
	{0xD566271B5E486D7EULL, "dae000cr_grand_cleric6124709.cl"},
	{0xD5664D5E33585F01ULL, "vxdr_0_deeproads_ad6_0_0.vlm"},
	{0xD56651618D5E3F95ULL, "character_character_p39.psh.fxo.dx11.assembly"},
	{0xD56676246E958A21ULL, "6166338_m.fxe"},
	{0xD566E506B0DD83ADULL, "hlac_lt_0_lowtown_a2_n_131070_2y.dds"},
	{0xD5671C4C20BE7A70ULL, "640387068.ogg"},
	{0xD5673E76F1DC1B75ULL, "360012209.ogg"},
	{0xD5675D12EFD5CBFDULL, "vxw_encounter2_camp_aj_0_0.vlm"},
	{0xD5678ECB7EBF1694ULL, "art\\levels\\races\\primeval\\props\\relics\\prp_megatikib_1.mao"},
	{0xD5679B3FCB82DBE2ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appc_0d.dds"},
	{0xD567CBDCCD173C13ULL, "greenscreen_noeffects.sub.xml"},
	{0xD567D091E9378303ULL, "183210666.ogg"},
	{0xD567E6ADE1900E63ULL, "313379470.ogg"},
	{0xD56829BA90CA7009ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_countthrone_d.dds"},
	{0xD569A49836647B3DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_f.gad"},
	{0xD569AD128BC06423ULL, "vxlt_0_lowtown_a2_n_c_hsj_0_0.vlm"},
	{0xD56A24CA6158CC10ULL, "vxca_caverns1_mines_iif_0_0.vlm"},
	{0xD56A85BFE59DA986ULL, "gen00fl_merrill6130129.cl"},
	{0xD56A9235216F7669ULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.impa.gad"},
	{0xD56A960962BC332BULL, "6062058_m.fxe"},
	{0xD56B10D92189596BULL, "bder_proj_center.mao"},
	{0xD56B183CC6AE5508ULL, "6164197_m.fxe"},
	{0xD56B2C9F6B12E5D1ULL, "478929341.ogg"},
	{0xD56B3525FE1B62D1ULL, "hlac_ca_caverns2_5802_45y.dds"},
	{0xD56B681ADB446B8FULL, "6203488_m.fxe"},
	{0xD56BB4700E22F1AEULL, "137956971.ogg"},
	{0xD56BC3E0FB14ECE3ULL, "9711647.ogg"},
	{0xD56C6172CB8C60C4ULL, "vxlt_0_lowtown_a2_n_g92_0_0.vlm"},
	{0xD56CD4F88622E0C3ULL, "gen_im_trs_drk_one_axe_r2.uti"},
	{0xD56CEF667EC89188ULL, "vxlt_darktown_5r1_0_0.vlm"},
	{0xD56CF33D7DF68585ULL, "6125559_m.fxe"},
	{0xD56CFAA296247774ULL, "177500488.ogg"},
	{0xD56D1E963CA05E73ULL, "sib131ip_family_shield.utp"},
	{0xD56D795412D2D37FULL, "hlac_ca_cavern1_haunted_24424_5x.dds"},
	{0xD56DA5CB1B65C036ULL, "mag360_alain6044660.cl"},
	{0xD56DC9C9E29AF8A8ULL, "6087203_m.fxe"},
	{0xD56DF7075A180C18ULL, "656290783.ogg"},
	{0xD56DFC49A1856941ULL, "qun221_petrice6077852.cl"},
	{0xD56DFD312B76D874ULL, "89659227.ogg"},
	{0xD56E2B60D94E7DBDULL, "vxht_chantry_2l6_0_0.vlm"},
	{0xD56F05D2573E36CDULL, "hlac_lt_0_lowtown_docks_a3_n_196607_3y.dds"},
	{0xD56F792B629914AFULL, "one280_emrys6104304.cl"},
	{0xD56F7BEED2BCEDCCULL, "vxlt_gallowstemplar_n_4xu_0_0.vlm"},
	{0xD56FB33C4B67B197ULL, "733660657.ogg"},
	{0xD56FD4BDEB17CD34ULL, "6071022_m.fxe"},
	{0xD5703EDDBD7542BAULL, "lt_0_lowtown_docks_a3_188_0.msh"},
	{0xD5704CDDBD755AF3ULL, "lt_0_lowtown_docks_a3_188_0.mao"},
	{0xD57050DDBD756100ULL, "lt_0_lowtown_docks_a3_188_0.mmh"},
	{0xD570732EFF657A1DULL, "follower_banter6172329.cl"},
	{0xD571361673105D91ULL, "vxca_caverns1_mines_i7u_0_0.vlm"},
	{0xD57150EB89E1A28EULL, "6139257_f.fxe"},
	{0xD57161379FA73B41ULL, "462515628.ogg"},
	{0xD5716335217544B7ULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.impa.ani"},
	{0xD571B016ACDB9ED1ULL, "art\\vfx\\textures\\vfx_pp_fademask.dds"},
	{0xD571BB1EBCFF7C8EULL, "232721843.ogg"},
	{0xD571CC53FBED4B3DULL, "vxdr_0_deeproads_cm3_0_0.vlm"},
	{0xD571D156C9C1BB93ULL, "darktown_fem_lowlife_mob6174141.cl"},
	{0xD5729A04525C78FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_archerbow.evt"},
	{0xD5738DA8577E536EULL, "vxdr_primevaltomb_7ug_0_1.vlm"},
	{0xD5739269435F177AULL, "440462697.ogg"},
	{0xD57403DFB20DA6BEULL, "vxw_encounter2_gg_0_0.vlm"},
	{0xD574445BDBC82624ULL, "vxlt_0_lowtown_a3_n_c_gw7_0_0.vlm"},
	{0xD57489756D27CBBBULL, "vxht_chantry_12m_0_0.vlm"},
	{0xD574A9DB2B122B06ULL, "mag321_comtesse_de_launcet6064899.cl"},
	{0xD575542C8EE3F340ULL, "6126973_m.fxe"},
	{0xD575BBCD8BE3D251ULL, "vxca_caverns2prc_9wr_0_0.vlm"},
	{0xD575FFA6BE1078C3ULL, "220501410.ogg"},
	{0xD5761C59BF413506ULL, "hlac_ca_caverns2_5802_0y.dds"},
	{0xD576BA718A699B02ULL, "ico_poison_crow.dds"},
	{0xD576F85A88EAB3A3ULL, "vxdr_primevaltomb_6zn_0_1.vlm"},
	{0xD57701FF9A9AD4F3ULL, "vxht_keep_n_6fu_0_0.vlm"},
	{0xD5774818EAB2F23DULL, "gen00flst_varric_acttwob.stg"},
	{0xD577B71A652739E9ULL, "hlac_ca_cavern1_haunted_24424_24y.dds"},
	{0xD57879237A35AAA4ULL, "qun221_viscount6070794.cl"},
	{0xD5788FF5D9315CB0ULL, "6176172_m.fxe"},
	{0xD578B583EC37D643ULL, "vxlt_0_lowtown_a2_n_csg_0_0.vlm"},
	{0xD578C0E43152AE7EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edgebase.phy"},
	{0xD579577127597757ULL, "vxlt_0_lowtown_a2_n_hdm_0_0.vlm"},
	{0xD5797B43ED5C966CULL, "ht_0_hightown_a2_9_0.msh"},
	{0xD5798143ED5CA05EULL, "ht_0_hightown_a2_9_0.mmh"},
	{0xD5798D43ED5CB4F5ULL, "ht_0_hightown_a2_9_0.mao"},
	{0xD579B1045262BE29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_archerbow.gad"},
	{0xD579BCD6DB5041C9ULL, "vxca_caverns1_iz7_0_0.vlm"},
	{0xD579EEB90D4312A8ULL, "qcr000_arishok6059981.cl"},
	{0xD57A557FE659ABEFULL, "979894388.ogg"},
	{0xD57A632B2CECA339ULL, "6190986_m.fxe"},
	{0xD57A7FE904066824ULL, "art\\vfx\\combat\\vfx_gib_ogr_p.explode.gad"},
	{0xD57AAB4C133E1BBBULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_twitch.evt"},
	{0xD57B3DCFCDA1C15DULL, "hlac_lt_0_lowtown_docks_n_131071_6y.dds"},
	{0xD57B651A6FEC9E05ULL, "6122459_m.fxe"},
	{0xD57B6C3A27FDFF1AULL, "vxht_mansion2b_1w7_0_0.vlm"},
	{0xD57BACCD981BB67DULL, "18136929.ogg"},
	{0xD57BB180A6D493E9ULL, "lowtown_fem_lowlife_mob6169185.cl"},
	{0xD57C09F332B76BD0ULL, "hlac_w_encounter2_196613_0x.dds"},
	{0xD57C32A3C5F73499ULL, "6057096_m.fxe"},
	{0xD57C6A983AE4531FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_large.gad"},
	{0xD57CCB1DD7BEA3E3ULL, "hlac_ht_brothel_n_2995_1z.dds"},
	{0xD57CF9D304EB3E10ULL, "fx_c_vol_light01.mao"},
	{0xD57D13CCB8996F14ULL, "6075245_f.fxe"},
	{0xD57D73BD56AB9968ULL, "vxht_keep_6f1_0_0.vlm"},
	{0xD57D91C0D40A31ADULL, "168064211.ogg"},
	{0xD57DE6A89642A757ULL, "zz_isa_debug6175235.cl"},
	{0xD57E16CEE8036CDAULL, "6176570_m.fxe"},
	{0xD57E566577831DEAULL, "vxw_sundermount_37d_0_1.vlm"},
	{0xD57E6B23ECFDEEBAULL, "74888531.ogg"},
	{0xD57EB26E6DC191BCULL, "mcr310_argument6075099.cl"},
	{0xD57F1A734DB85544ULL, "6087345_f.fxe"},
	{0xD57F42612E47B6E4ULL, "673346546.ogg"},
	{0xD57F7CEAA8CB3C55ULL, "vxlt_hangedman_1qt_0_0.vlm"},
	{0xD57FBDDD02819D04ULL, "mcr310_argument6074416.cl"},
	{0xD57FCFF96D074F5CULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_aveline_mcr370_playerflees.ani"},
	{0xD58017252672B0E1ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pile_rags_0_sa.dds"},
	{0xD5803E30EDE363BDULL, "vxdr_0_deeproads_cj2_0_0.vlm"},
	{0xD580820452688D77ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_archerbow.ani"},
	{0xD580A44C13427BFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_twitch.gad"},
	{0xD5811659D68A1682ULL, "6139709_f.fxe"},
	{0xD581541628C0702EULL, "6164443_m.fxe"},
	{0xD5816D65571754C5ULL, "var_placeable_ambient.gda"},
	{0xD5818EE9040CA002ULL, "art\\vfx\\combat\\vfx_gib_ogr_p.explode.ani"},
	{0xD58193ADC5B0A223ULL, "vxlt_0_lowtown_n_elj_0_2.vlm"},
	{0xD581D90C45E044A6ULL, "vxlt_warehouse_n_uc_0_0.vlm"},
	{0xD581FBA15B813B90ULL, "6095752_m.fxe"},
	{0xD5821A9D3D1AC1B7ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_explo_c.anb"},
	{0xD58231908039E7CAULL, "art\\vfx\\materials\\vfx_noise_add.mao"},
	{0xD5823BB021174655ULL, "vxht_mansion_n_5hr_0_0.vlm"},
	{0xD5827BF6897065A8ULL, "6077011_m.fxe"},
	{0xD582C7F38250E170ULL, "vxlt_0_lowtown_docks_aqz_0_0.vlm"},
	{0xD582CDC2BF22A65BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_forward_5_steps.ani"},
	{0xD582EFEE878CB9C2ULL, "vxca_caverns1_mines_imh_0_0.vlm"},
	{0xD5835F983AEA5F0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_large.evt"},
	{0xD58378C33008745DULL, "gen00fl_merrill6119644.cl"},
	{0xD583BC9A33CF82A2ULL, "ser191_emeric6104730.cl"},
	{0xD5840024B0EE4C9DULL, "195427724.ogg"},
	{0xD58420FB03D7320FULL, "6177583_m.fxe"},
	{0xD5842117956CC055ULL, "980715740.ogg"},
	{0xD58446BE74CFE231ULL, "gen00fl_aveline6101727.cl"},
	{0xD5844AC9265F8BFFULL, "gen_im_trs_dal_hvy_glv_r2.uti"},
	{0xD585333FA24D97FDULL, "vxca_cave_220_0_0.vlm"},
	{0xD58573DC8C90ED73ULL, "vxht_keep_n_6at_0_0.vlm"},
	{0xD585DCF96D0BD09EULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_aveline_mcr370_playerflees.gad"},
	{0xD5861DF6353BF881ULL, "vxdr_0_deeproads_d1l_0_0.vlm"},
	{0xD5863BAE018ECA6EULL, "dae000_hate_enemies6167596.cl"},
	{0xD5863C9060C9568BULL, "gen_im_cft_reg_i2.uti"},
	{0xD5868F3CCB3707E1ULL, "mcr310_argument6173615.cl"},
	{0xD586CE6705851279ULL, "883759647.ogg"},
	{0xD5870860F4B45628ULL, "combat_ar_r1_a1.gda"},
	{0xD58776CA2FF697BDULL, "934717707.ogg"},
	{0xD587A34C134898FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\wall_lean_right_twitch.ani"},
	{0xD58897C358F194D4ULL, "636701689.ogg"},
	{0xD588C1C8C37D90C7ULL, "6202209_m.fxe"},
	{0xD5890152467A4AF9ULL, "follower_barks6154151.cl"},
	{0xD5893A1716E3681BULL, "213446722.ogg"},
	{0xD589D5506BE4B973ULL, "1072130706.ogg"},
	{0xD589ECB6E7C5F7C4ULL, "569780627.ogg"},
	{0xD58A5B6F78F10629ULL, "vxlt_0_lowtown_a2_n_f52_0_0.vlm"},
	{0xD58A82B6ECCE2043ULL, "lgt110_urn_guard6133140.cl"},
	{0xD58B0F1ABD3AAEBDULL, "vxht_chantry_2b4_0_0.vlm"},
	{0xD58B49180AC578A9ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_roba_0d.dds"},
	{0xD58B63318BAAF2F9ULL, "vxca_caverns2_8bi_0_0.vlm"},
	{0xD58B91FB167D85F2ULL, "hlac_w_blightlands_196605_0z.dds"},
	{0xD58BD3B9F6BD2854ULL, "gen00fl_anders6128344.cl"},
	{0xD58C9AF04C1CC798ULL, "6068495_m.fxe"},
	{0xD58CC8426979366AULL, "vxw_sundermount_32g_0_1.vlm"},
	{0xD58D691505305016ULL, "835505635.ogg"},
	{0xD58E21263F756AEFULL, "397371720.ogg"},
	{0xD58E2B3BE91C03BFULL, "hlac_lt_backalley_g_0_16z.dds"},
	{0xD58E8203AA8048CCULL, "vxlt_gallowstemplar_f_46g_0_0.vlm"},
	{0xD58EB00DDFD97C3DULL, "vxdr_0_deeproads_cg1_0_0.vlm"},
	{0xD58F27AF72B0DF3EULL, "mag100_cullen6086341.cl"},
	{0xD58FA0DA5F5221FCULL, "vxlt_undercityserialkiller_178_0_0.vlm"},
	{0xD58FDFEA75A16532ULL, "plt_cod_ltr_junk_propaganda.nss"},
	{0xD5900EC8B6B6F7D2ULL, "gen00fl_anders6125406.cl"},
	{0xD590197114AE4BC4ULL, "hlac_lt_0_lowtown_n_131073_1z.dds"},
	{0xD5904AE937D65D26ULL, "vxlt_warehouse_n_xd_0_0.vlm"},
	{0xD590D5C2BF2EF152ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_forward_5_steps.evt"},
	{0xD590EDD37B667E28ULL, "6078111_m.fxe"},
	{0xD590F08EFC66850AULL, "10786826.ogg"},
	{0xD59124C24FCE0865ULL, "vxht_0_hightown_c29_0_0.vlm"},
	{0xD5913DBE53E01C05ULL, "zz_fen_debug6215261.cl"},
	{0xD59147983AF67465ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_large.ani"},
	{0xD591606C35B234C6ULL, "vxht_chantry_n_2l6_0_0.vlm"},
	{0xD59161CB7982D242ULL, "vxca_caverns1_mines_iji_0_0.vlm"},
	{0xD591DFE805F86532ULL, "6093536_m.fxe"},
	{0xD5921347E953D44AULL, "lt_0_lowtown_docks_a3_n_32_0_ll.mmh"},
	{0xD59257FD65109CE0ULL, "vxht_mansion_n_6j2_0_0.vlm"},
	{0xD59302C60CD2840FULL, "ter_stoney_n.dds"},
	{0xD5938468ED46CA3BULL, "gen00fl_sebastian6179975.cl"},
	{0xD5938E4FDDC69603ULL, "679262616.ogg"},
	{0xD59438FABC95BEEFULL, "dae000ip_crowd_lm1_talk2.utp"},
	{0xD594694F7813844BULL, "vxlt_0_lowtown_a2_n_c_es7_0_0.vlm"},
	{0xD595B1150A1A2459ULL, "vxdr_0_deeproads_es3_0_0.vlm"},
	{0xD595DB213348B03DULL, "llac_ca_caverns1_24424_17.dds"},
	{0xD595EB2D52949E38ULL, "dae200_bran6094812.cl"},
	{0xD596742D6DFD73E0ULL, "vxlt_0_lowtown_a3_n_c_ep1_0_1.vlm"},
	{0xD596EE36FAC43163ULL, "t3_glk_b03.tnt"},
	{0xD5975F6BCFD97165ULL, "qun120_arishok6172057.cl"},
	{0xD5979A67AFFAFD12ULL, "vxht_0_hightown_aie_0_0.vlm"},
	{0xD597CBF99F92769FULL, "vxlt_darktown_4v3_0_0.vlm"},
	{0xD597F4C2BF3544E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\run_forward_5_steps.gad"},
	{0xD597F96ED9CA1D2FULL, "6216675_m.fxe"},
	{0xD59806732A54015FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_floor_tile_n.dds"},
	{0xD5980DB038AF56FBULL, "w_blightlands_79_0.msh"},
	{0xD59816C38B63C51FULL, "hlac_lt_0_lowtown_131071_3z.dds"},
	{0xD5981FB038AF7546ULL, "w_blightlands_79_0.mao"},
	{0xD598272A7D2E5061ULL, "6120798_m.fxe"},
	{0xD5982BB038AF89E5ULL, "w_blightlands_79_0.mmh"},
	{0xD599244843379351ULL, "hlac_lt_blackemporium_1055_25z.dds"},
	{0xD599CBB8D1453EF9ULL, "vxht_0_hightown_bfl_0_0.vlm"},
	{0xD599D35B69ED5419ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_a_c.impa.ani"},
	{0xD59A194552C1B06FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wrba_0d.dds"},
	{0xD59A7A3D716A1923ULL, "6106091_m.fxe"},
	{0xD59A8B0F8E9B64DCULL, "lt_0_lowtown_docks_n_37_0.rml"},
	{0xD59AE6835EAC5ED5ULL, "dae000_male_commoner1_mob6174448.cl"},
	{0xD59AF1072104F8FBULL, "24224331.ogg"},
	{0xD59B9CAC7D331CEEULL, "character_character_p245.psh.fxo.dx11"},
	{0xD59BC50B16ED771EULL, "lgt000im_ultra_talwain.uti"},
	{0xD59C397FC4CB4FD0ULL, "vxht_0_hightown_a2_bel_0_0.vlm"},
	{0xD59C5BFE6BE022CAULL, "dm_waxler.mor"},
	{0xD59D45255D40579AULL, "hlac_w_bonepitdestroyed_a3_0_9y.dds"},
	{0xD59DADDFDA738024ULL, "w_blightlands_199_0.msh"},
	{0xD59DBFDFDA739E8DULL, "w_blightlands_199_0.mao"},
	{0xD59DC3DFDA73A5C6ULL, "w_blightlands_199_0.mmh"},
	{0xD59E1133CA3AE0B1ULL, "6136924_m.fxe"},
	{0xD59F05EB4DC9F463ULL, "ser210_witness6168148.cl"},
	{0xD59F2561E160529CULL, "vxlt_0_lowtown_eot_0_1.vlm"},
	{0xD59F57DC5CFEF428ULL, "llac_ht_0_hightown_-65536_3.dds"},
	{0xD59F70D5AB4300DBULL, "dre000_dougal6128886.cl"},
	{0xD59F7A300CE7E522ULL, "6043872_m.fxe"},
	{0xD59F8BC457F24B09ULL, "dae000_loyal_mage6165220.cl"},
	{0xD5A07A1BC71DC5BBULL, "art\\characters\\playercharacter\\elfmale\\rob_orsa.mmh"},
	{0xD5A0861BC71DDA58ULL, "art\\characters\\playercharacter\\elfmale\\rob_orsa.mao"},
	{0xD5A0981BC71DF8A5ULL, "art\\characters\\playercharacter\\elfmale\\rob_orsa.msh"},
	{0xD5A0AE5B69F3345BULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_a_c.impa.gad"},
	{0xD5A0D200FEC79F1BULL, "zz_vault_debug6177609.cl"},
	{0xD5A0E1E6B02F5EF1ULL, "dre000_sibling_has_blight.stg"},
	{0xD5A116CD8978274AULL, "mag360cr_hostage_fenris.utc"},
	{0xD5A142AF5E252FAAULL, "402685405.ogg"},
	{0xD5A144157F2CC6A3ULL, "vxlt_warehouse_130_0_0.vlm"},
	{0xD5A1A61922825CE3ULL, "789429.ogg"},
	{0xD5A1B79F391CBCEFULL, "hlac_lt_darktown_-1_36z.dds"},
	{0xD5A238BDDB54BED7ULL, "vxht_0_hightown_a3_baf_0_0.vlm"},
	{0xD5A298B6737CDDC6ULL, "205959295.ogg"},
	{0xD5A2A2231B3E0B35ULL, "vxlt_0_lowtown_docks_a3_bu7_0_0.vlm"},
	{0xD5A2CF92CD477DFAULL, "character_character_p1.vsh.fxo.dx11"},
	{0xD5A2FDFD03A3F4BAULL, "140539167.ogg"},
	{0xD5A30F4F8851708CULL, "6112696_f.fxe"},
	{0xD5A36C9BDBABDD5AULL, "vxw_bonepitmining_a2_l8_0_0.vlm"},
	{0xD5A3731076A0C36AULL, "zz_approval_debug6003692.cl"},
	{0xD5A43965CA05EDC9ULL, "vxw_encounter3_dr_0_0.vlm"},
	{0xD5A4B18169F27F02ULL, "vxca_caverns1_mines_dq5_0_0.vlm"},
	{0xD5A5126DB1328B49ULL, "vxca_caverns1_ik4_0_0.vlm"},
	{0xD5A5A86101DD6487ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.dura.gad"},
	{0xD5A5BC5C6E5AF558ULL, "mag110_arianni6049708.cl"},
	{0xD5A6A416B5B95977ULL, "47530701.ogg"},
	{0xD5A6EC03323CF1EBULL, "463134190.ogg"},
	{0xD5A74C58445D4A78ULL, "hlac_lt_0_lowtown_a2_n_c_65533_3y.dds"},
	{0xD5A7C0DEBEB4C4E9ULL, "vxca_caverns1_mines_ioe_0_0.vlm"},
	{0xD5A7CB5B69F98440ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_a_c.impa.evt"},
	{0xD5A7F2538DE7FE37ULL, "vxlt_gallowstemplar_n_6s5_0_0.vlm"},
	{0xD5A81AF4D85F88CDULL, "628241398.ogg"},
	{0xD5A8BBC31210168AULL, "vxlt_0_lowtown_docks_a3_n_c1f_0_0.vlm"},
	{0xD5A8E1E1C3995804ULL, "525401526.ogg"},
	{0xD5A96B19A7BF2989ULL, "ser270_templar6127411.cl"},
	{0xD5A9DBC30A3FEA78ULL, "gencr_talvashoth_captain.utc"},
	{0xD5AAB07B965EBE46ULL, "vxlt_0_lowtown_a2_n_c_ekf_0_0.vlm"},
	{0xD5AB179D3D3DC1B8ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_explo_c.mmh"},
	{0xD5AB7D0631683150ULL, "966859568.ogg"},
	{0xD5ABAA29B8B39338ULL, "6082179_f.fxe"},
	{0xD5ABB4EE5ED01BEBULL, "6034541_m.fxe"},
	{0xD5AC337753FA7049ULL, "vxlt_gallowstemplar_n_46f_0_0.vlm"},
	{0xD5AC5F5FBDFA8F6DULL, "vxlt_foundry_n_xu_0_0.vlm"},
	{0xD5AC8FB37DF2748BULL, "statecnt_crowd.gda"},
	{0xD5AC9D6101E37094ULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.dura.evt"},
	{0xD5AD17399BF1A263ULL, "645211725.ogg"},
	{0xD5AD5814D0255167ULL, "247517162.ogg"},
	{0xD5AD582D9E9C64B7ULL, "vxht_keep_3xe_0_0.vlm"},
	{0xD5ADB11CFCA4C55AULL, "122042759.ogg"},
	{0xD5AE3572A8CB9A35ULL, "6197770_m.fxe"},
	{0xD5AEBF4B4801ADA5ULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.anb"},
	{0xD5AF23A6C62ADD66ULL, "hlac_lt_gallowstemplar_2391_9x.dds"},
	{0xD5AF355D9610038CULL, "58068803.ogg"},
	{0xD5AF7804C7B136D2ULL, "ico_boots_heavy_4.dds"},
	{0xD5AFB5F27122DF23ULL, "vxlt_warehouse_163_0_0.vlm"},
	{0xD5AFCB34FD640BCDULL, "war_dmgupgrade.dds"},
	{0xD5AFCCC9E544C94BULL, "llac_lt_0_lowtown_0_0.dds"},
	{0xD5AFFB7E940A19F4ULL, "follower_banter6166367.cl"},
	{0xD5B0836596B53B80ULL, "zz_mrl_debug6109312.cl"},
	{0xD5B0AA5A428F3660ULL, "vxlt_undercitydungeon_12j_0_0.vlm"},
	{0xD5B1223F0CAF7A7BULL, "gen_im_arm_cht_comb_t5.uti"},
	{0xD5B17BDFB1C3FC9CULL, "hlac_ht_keep_4519_11z.dds"},
	{0xD5B1D092750494D5ULL, "qun160_vashothlead6053923.cl"},
	{0xD5B1D70CFB8797F9ULL, "203765314.ogg"},
	{0xD5B24ACB35C6494CULL, "54735974.ogg"},
	{0xD5B254B018ACA6D4ULL, "553415227.ogg"},
	{0xD5B2554200E061C9ULL, "6056801_f.fxe"},
	{0xD5B2A8F02DC5386BULL, "vxca_caverns1_mines_e4r_0_0.vlm"},
	{0xD5B328F0784C046AULL, "6218852_m.fxe"},
	{0xD5B3CBBE3AE4BBE3ULL, "llac_ca_caverns2prc_5802_25.dds"},
	{0xD5B3EE859010B401ULL, "mag110_vincento6017661.cl"},
	{0xD5B436AAB800EF63ULL, "vxlt_gallowscourtyard_n_6z0_0_0.vlm"},
	{0xD5B4E6A6B24294A9ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_rec.mmh"},
	{0xD5B5270C9C829F35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2sml.msh"},
	{0xD5B5290C9C82A2ABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2sml.mmh"},
	{0xD5B5350C9C82B748ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning2sml.mao"},
	{0xD5B56EA34FB79B13ULL, "vxlt_0_lowtown_docks_n_b9a_0_0.vlm"},
	{0xD5B5D7E3DC19FF93ULL, "349498062.ogg"},
	{0xD5B682FACC23933CULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.impa.ani"},
	{0xD5B68CAC3689460FULL, "llac_ht_mansion_n_3371_1.dds"},
	{0xD5B77D556E67A1CEULL, "vxlt_gallowsprison_n_6no_0_0.vlm"},
	{0xD5B79C5584D3A5B1ULL, "6168776_m.fxe"},
	{0xD5B7AB4FFD3B3E1CULL, "plt_prm000pt_deep_blade_.nss"},
	{0xD5B7EAEAE380E084ULL, "isa311st_castillon.stg"},
	{0xD5B8C0C05C2C3CF8ULL, "6122743_f.fxe"},
	{0xD5B908FC198EDDB5ULL, "hlac_lt_undercitydungeon_-1_6z.dds"},
	{0xD5B923E6FC0E2C3EULL, "hightown_market_guard_sta6174745.cl"},
	{0xD5B93FDA2589A65DULL, "vxht_keep_55r_0_0.vlm"},
	{0xD5B9445E70A39E55ULL, "vxlt_hangedman_1mp_0_0.vlm"},
	{0xD5B95843ED923EF9ULL, "ht_0_hightown_a2_9_0.phy"},
	{0xD5B9E8AF3AEA8287ULL, "hlac_ca_cave_2194_14y.dds"},
	{0xD5B9FB5FF1C433AEULL, "vxlt_0_lowtown_exr_0_0.vlm"},
	{0xD5BA4E885FF68774ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic_4.mao"},
	{0xD5BA8CED76F62D64ULL, "vxw_bonepitmining_a2_1yb_0_0.vlm"},
	{0xD5BA956101EFA0CDULL, "art\\vfx\\combat\\vfx_d_aoe_imp_p.dura.ani"},
	{0xD5BB17261F76F6E7ULL, "zz_tbe100_barber6191008.cl"},
	{0xD5BBFF2391DAE577ULL, "dae021_sister_samea_sta6142092.cl"},
	{0xD5BC240EC5225C61ULL, "hlac_lt_0_lowtown_a3_n_c_-2_5z.dds"},
	{0xD5BC2CED96A58A9AULL, "hlac_lt_0_lowtown_docks_131073_0y.dds"},
	{0xD5BC4F51C624422BULL, "art\\vfx\\combat\\vfx_str_g_pound_l_leg_c.mmh"},
	{0xD5BC6D8D0CA86BB8ULL, "41459575.ogg"},
	{0xD5BD162D1C1EC9CAULL, "vxht_chantry_2pi_0_0.vlm"},
	{0xD5BD2CBBD45BFB61ULL, "ef_varania.mor"},
	{0xD5BD874CD4255485ULL, "follower_banter6151281.cl"},
	{0xD5BD8FFACC29C87EULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.impa.gad"},
	{0xD5BE29465C0A5B6DULL, "vxlt_gallowstemplar_f_5wl_0_0.vlm"},
	{0xD5BE65FC299AAC83ULL, "43653251.ogg"},
	{0xD5BE79D58E4720E1ULL, "6209850_f.fxe"},
	{0xD5BF3754B8671CEBULL, "llac_dr_dungeon_3207_3.dds"},
	{0xD5BF3BEB25B129A7ULL, "vxdr_primevalentrance_2su_0_1.vlm"},
	{0xD5BF3D0A5C7F024AULL, "qun160_vashothlead6055582.cl"},
	{0xD5BF41116FC54A92ULL, "dae000_mother6099172.cl"},
	{0xD5BFA489BEE3C5DAULL, "308861794.ogg"},
	{0xD5BFCF94687CDD9BULL, "w_encounter1_26_0.mmh"},
	{0xD5BFDB94687CF238ULL, "w_encounter1_26_0.mao"},
	{0xD5BFE3989AAE45D6ULL, "53317849.ogg"},
	{0xD5BFED94687D1085ULL, "w_encounter1_26_0.msh"},
	{0xD5C01D3673983172ULL, "gen_im_arm_cht_med_act0_02.uti"},
	{0xD5C01F4CDB8D1432ULL, "vxw_sundermount_35k_0_0.vlm"},
	{0xD5C05197DE2B2794ULL, "art\\vfx\\combat\\vfx_glb_confused_c.dura.ani"},
	{0xD5C06143ED986C15ULL, "ht_0_hightown_a2_9_0.rml"},
	{0xD5C0763A2784F6A8ULL, "6166461_m.fxe"},
	{0xD5C08905D1D4C059ULL, "157933751.ogg"},
	{0xD5C16340B619D53BULL, "6098240_m.fxe"},
	{0xD5C1B96AFB277F77ULL, "lowtown_male_lowlife2_mob6166765.cl"},
	{0xD5C23A51EC21058AULL, "266248046.ogg"},
	{0xD5C24EF4E975584EULL, "vxht_mansion_n_4zn_0_0.vlm"},
	{0xD5C2DE7A203096A5ULL, "44578090.ogg"},
	{0xD5C368D3DAF1A313ULL, "255475902.ogg"},
	{0xD5C42405C5028415ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3_0.pwk"},
	{0xD5C46AFACC2FA79DULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.impa.evt"},
	{0xD5C4C09FA2E74F5FULL, "549347472.ogg"},
	{0xD5C50AF632C8259AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbench_dp.dds"},
	{0xD5C5600779A5B38FULL, "6067699_m.fxe"},
	{0xD5C5C20F8DF9080BULL, "vxht_mansion2b_3d9_0_0.vlm"},
	{0xD5C5E65A9B837E5BULL, "vxdr_0_deeproads_el9_0_1.vlm"},
	{0xD5C5FF549EEFE259ULL, "298798945.ogg"},
	{0xD5C609ECB0AEB8E8ULL, "138667306.ogg"},
	{0xD5C66D64541DB6D4ULL, "vxlt_warehouse_w2_0_0.vlm"},
	{0xD5C67000D14F8D63ULL, "w_encounter2_camp_4_0.rml"},
	{0xD5C6AB90C73B7349ULL, "zz_fen_debug6165498.cl"},
	{0xD5C6E65CB67B9C35ULL, "6048720_m.fxe"},
	{0xD5C740FBD55337E9ULL, "6116096_m.fxe"},
	{0xD5C74E97DE3141F6ULL, "art\\vfx\\combat\\vfx_glb_confused_c.dura.gad"},
	{0xD5C76FEF126DD402ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm04.phy"},
	{0xD5C80C43C522D704ULL, "gen00fl_varric6033639.cl"},
	{0xD5C83979B17E96C9ULL, "vxht_0_hightown_n_bzk_0_0.vlm"},
	{0xD5C883200C7E7673ULL, "one280_emrys6120056.cl"},
	{0xD5C8FECA68EC45E4ULL, "vxw_bonepitmining_a2_1vc_0_0.vlm"},
	{0xD5C9ADD2DEEFAAEFULL, "6001336_m.fxe"},
	{0xD5C9BBB976AFDF63ULL, "246663958.ogg"},
	{0xD5CA3051C6304A90ULL, "art\\vfx\\combat\\vfx_str_g_pound_l_leg_c.anb"},
	{0xD5CA6E333AEF2C70ULL, "plt_cod_crt_desire_demon.plo"},
	{0xD5CA9FEC30058432ULL, "hlac_lt_0_lowtown_a2_-65535_6z.dds"},
	{0xD5CB4B520E23ABCCULL, "347247106.ogg"},
	{0xD5CB57EFD2E8BBE6ULL, "503641904.ogg"},
	{0xD5CBEC5FF7AA4F58ULL, "dre000_stroud6036734.cl"},
	{0xD5CC30DAE87413A8ULL, "6156642_m.fxe"},
	{0xD5CC6392139947E8ULL, "6033962_seldrek_intro.cut"},
	{0xD5CC8353E1ACFCA6ULL, "mag110_arianni6049715.cl"},
	{0xD5CD7700D155B7DBULL, "w_encounter2_camp_4_0.phy"},
	{0xD5CDB0E47CFC375BULL, "gen00fl_fenris6126657.cl"},
	{0xD5CDEFE21595CA93ULL, "235617591.ogg"},
	{0xD5CE02BBDC0160A7ULL, "ser211ip_leave.utp"},
	{0xD5CE0E13127E025AULL, "wm_city_night.map"},
	{0xD5CE4AC4BDFA266BULL, "dae000_bartrand6201359.cl"},
	{0xD5CE6997DE378E55ULL, "art\\vfx\\combat\\vfx_glb_confused_c.dura.evt"},
	{0xD5CE8E943E92CE65ULL, "vxca_caverns2_7m6_0_1.vlm"},
	{0xD5CEB18E9C9315E9ULL, "892942830.ogg"},
	{0xD5CEBD609E942FB7ULL, "hlac_ht_0_hightown_a2_-131071_1y.dds"},
	{0xD5CF1D7E3F7E1350ULL, "6125965_m.fxe"},
	{0xD5CF8E47B96264C5ULL, "vxlt_undercityencounter_1c5_0_0.vlm"},
	{0xD5CFB9F63BD7F069ULL, "vxht_mansion2b_33z_0_0.vlm"},
	{0xD5D0399172FE8FD1ULL, "123547186.ogg"},
	{0xD5D0A2EAECBC4009ULL, "279258589.ogg"},
	{0xD5D0F02DAB643FC0ULL, "6034072_m.fxe"},
	{0xD5D1AEAEDC391E63ULL, "vxlt_blackemporium_s4_0_1.vlm"},
	{0xD5D263F92CEA8B73ULL, "hlac_dr_0_deeproads_5842_6x.dds"},
	{0xD5D272297463B00EULL, "character_character_p210.psh.fxo.dx11.assembly"},
	{0xD5D2B6A511CAB848ULL, "map_day_1280x1024.dds"},
	{0xD5D303370AC31A50ULL, "vxlt_0_lowtown_a2_hej_0_0.vlm"},
	{0xD5D361CE4EF5504BULL, "face0_nolit.psh.fxo"},
	{0xD5D36DE3581E546AULL, "887952269.ogg"},
	{0xD5D384550C6BF274ULL, "87444067.ogg"},
	{0xD5D3B358FC613EABULL, "6086283_m.fxe"},
	{0xD5D40BDE5A16D862ULL, "127629706.ogg"},
	{0xD5D483C93CEAF5E4ULL, "vxw_bonepitmining_a2_228_0_0.vlm"},
	{0xD5D48EE1E98E1CAAULL, "hlac_w_bonepit_6_0x.dds"},
	{0xD5D4DF414613CF54ULL, "vxlt_warehouse_x3_0_0.vlm"},
	{0xD5D4F82EA6355B8BULL, "6119135_f.fxe"},
	{0xD5D5D8C6F68D8AB4ULL, "6179978_f.fxe"},
	{0xD5D5F1BFDC0666CBULL, "ser221_gascard6134457.cl"},
	{0xD5D6008B8D49671FULL, "rog_hatebomb.dds"},
	{0xD5D65CD5943D6695ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_low.phy"},
	{0xD5D65DC0348FAD5DULL, "mrl150_merrill6070502.cl"},
	{0xD5D6B3CB3AC12F6DULL, "gen00fl_anders6109364.cl"},
	{0xD5D6DF19D5B064AEULL, "vxlt_0_lowtown_ent_0_0.vlm"},
	{0xD5D70F8EF3F7D8CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cb_transition.gad"},
	{0xD5D74AD2232ED05CULL, "6176293_m.fxe"},
	{0xD5D790B0FA908332ULL, "vxca_caverns1_giw_0_0.vlm"},
	{0xD5D7C04B4824B4E6ULL, "art\\vfx\\creatures\\fxc_shale_ice_l_c.mmh"},
	{0xD5D7E7323CF77C63ULL, "292698645.ogg"},
	{0xD5D7F7E6B05EC61EULL, "dre000_sibling_has_blight.cnv"},
	{0xD5D8233D72ACDBD4ULL, "824538406.ogg"},
	{0xD5D82A21C427259EULL, "cod_ltr_junk_propaganda.utp"},
	{0xD5D83E7F10650587ULL, "mer221_sabin.bnk"},
	{0xD5D873D61A0E19BDULL, "463834986.ogg"},
	{0xD5D8ADC4747EAE37ULL, "vxht_keep_3if_0_0.vlm"},
	{0xD5D95DF20999C8B8ULL, "6074457_f.fxe"},
	{0xD5DAD9E6B060C92EULL, "dre000_sibling_has_blight.bnk"},
	{0xD5DADA6A41018A15ULL, "ave100_ambush_16106200.cl"},
	{0xD5DB607F106775FFULL, "mer221_sabin.cnv"},
	{0xD5DB79B1720FDEB4ULL, "vxdr_primevaltomb_4xd_0_0.vlm"},
	{0xD5DBF28FBF666BB8ULL, "6072904_m.fxe"},
	{0xD5DC005481C4CF0EULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_arishok_arishokresponse.ani"},
	{0xD5DC9395BB959648ULL, "337280083.ogg"},
	{0xD5DCB2AEE4F58575ULL, "249730912.ogg"},
	{0xD5DCC3BF5F729155ULL, "vxw_encounter2_camp_9j_0_0.vlm"},
	{0xD5DCE9BE0B431F27ULL, "328449672.ogg"},
	{0xD5DD2304D1168E61ULL, "llac_lt_gallowstemplar_2394_9.dds"},
	{0xD5DDEA65C3E89ED9ULL, "vxdr_0_deeproads_eh6_0_0.vlm"},
	{0xD5DDF28EF3FDC661ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cb_transition.evt"},
	{0xD5DEB157FDDD507AULL, "zz_ave_debug6174605.cl"},
	{0xD5DEFDC6610A2C50ULL, "w_bonepit_111_0.mao"},
	{0xD5DF01C6610A3363ULL, "w_bonepit_111_0.mmh"},
	{0xD5DF0FC6610A4ABDULL, "w_bonepit_111_0.msh"},
	{0xD5DF134F0CC8C930ULL, "365212614.ogg"},
	{0xD5DF19556E7733BEULL, "6076890_m.fxe"},
	{0xD5DF97B4EFBFFFC7ULL, "346224424.ogg"},
	{0xD5DFF182D1655515ULL, "124016104.ogg"},
	{0xD5E072EB6537B210ULL, "hlac_w_sundermount_131075_0z.dds"},
	{0xD5E1DB4D2A537F57ULL, "431177037.ogg"},
	{0xD5E1EF12A3199400ULL, "hlac_w_woundedcoast_0_7z.dds"},
	{0xD5E23364E49A0F09ULL, "lt_0_lowtown_a2_45_0.rml"},
	{0xD5E4D6469D97BDABULL, "mcr310_argument6072316.cl"},
	{0xD5E4E1C4F62415A8ULL, "6087012_m.fxe"},
	{0xD5E4ECD39C874924ULL, "hlac_lt_0_lowtown_131069_0y.dds"},
	{0xD5E51F95FA0D7FA7ULL, "6121913_m.fxe"},
	{0xD5E550F6C7A67D2EULL, "vxlt_0_lowtown_eiu_0_0.vlm"},
	{0xD5E574847A2B2585ULL, "6178496_m.fxe"},
	{0xD5E5F429B8BAF04BULL, "642611429.ogg"},
	{0xD5E63702EA208602ULL, "llac_ht_keep_n_3733_0.dds"},
	{0xD5E70F7EEB0884EEULL, "371928084.ogg"},
	{0xD5E79B1BC75A00F0ULL, "art\\characters\\playercharacter\\elfmale\\rob_orsa.phy"},
	{0xD5E7C6744B6E8663ULL, "w_encounter2_55_0.rml"},
	{0xD5E8558D9D085D5BULL, "qcr221_cut_diplomatic_sol6201288.cl"},
	{0xD5E89BBB1A115FD1ULL, "391730936.ogg"},
	{0xD5E8C5E4AD79665FULL, "697022553.ogg"},
	{0xD5E940B98A46B18BULL, "6074387_m.fxe"},
	{0xD5E99E5B9FDA36C1ULL, "vxlt_0_lowtown_a2_n_c_hx1_0_0.vlm"},
	{0xD5E9F33EB4C01337ULL, "vxdr_primevaltomb_50z_0_0.vlm"},
	{0xD5E9FC2C718EEE2FULL, "6129913_m.fxe"},
	{0xD5EB359C5168A9D5ULL, "vxw_encounter2_camp_4i_0_0.vlm"},
	{0xD5EB40BF7D067728ULL, "vxw_bonepitdestroyed_a3_26j_0_0.vlm"},
	{0xD5EB63C928FA81C6ULL, "win_beam.mao"},
	{0xD5EB889679C1E9F7ULL, "245632667.ogg"},
	{0xD5EB910A5AE096ACULL, "958668936.ogg"},
	{0xD5EB960196206A27ULL, "gen_im_cns_cft_arn_09.uti"},
	{0xD5EBEA8EF409F5B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cb_transition.ani"},
	{0xD5ED1AB66E423D71ULL, "dae220ip_at_chantry.utp"},
	{0xD5ED56EF128E7955ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm04.mmh"},
	{0xD5ED58EF128E7CCBULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm04.msh"},
	{0xD5EDAEC0256394D4ULL, "hlac_lt_0_lowtown_docks_n_-3_7z.dds"},
	{0xD5EE22E579DB5F6EULL, "hlac_ca_caverns2prc_5802_54y.dds"},
	{0xD5EE49875853171AULL, "dae000_fem_priest_mob6174185.cl"},
	{0xD5EE662829631D2FULL, "533365798.ogg"},
	{0xD5EECD744B74B0DBULL, "w_encounter2_55_0.phy"},
	{0xD5EF5AB63C34C932ULL, "gen_im_wep_war_one_act0_03.uti"},
	{0xD5EF67217735F17BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrotha_1.mao"},
	{0xD5EF82EF0080B0D7ULL, "vxca_cavern1_haunted_ij0_0_0.vlm"},
	{0xD5EFAA658C28FAD4ULL, "6143204_m.fxe"},
	{0xD5EFD95481D53648ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_arishok_arishokresponse.gad"},
	{0xD5EFE8DA5959AA8BULL, "mcr370_moira6087500.cl"},
	{0xD5F01744C03F35E2ULL, "gen00fl_aveline6139879.cl"},
	{0xD5F07FE2EC6AF915ULL, "vxw_bonepit_1wv_0_1.vlm"},
	{0xD5F0B79D3637965BULL, "vxlt_0_lowtown_a2_n_c_gjj_0_0.vlm"},
	{0xD5F17D129A4F0E32ULL, "art\\vfx\\combat\\vfx_c_profane_crust_c.dura.gad"},
	{0xD5F196DA0BC463E7ULL, "495499209.ogg"},
	{0xD5F1E490B9E47755ULL, "lt_gallowsprison_n_3567_lc.mmh"},
	{0xD5F2209658149F2AULL, "vxht_chantry_n_2v8_0_0.vlm"},
	{0xD5F245C06FC80706ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pota_sa.dds"},
	{0xD5F273A1E202B29CULL, "429848191.ogg"},
	{0xD5F28B17C0981B5FULL, "hlac_lt_0_lowtown_a2_-262144_3y.dds"},
	{0xD5F2C8EBFDCC6302ULL, "205644696.ogg"},
	{0xD5F2F8715D786BACULL, "vxlt_undercityencounter_101_0_0.vlm"},
	{0xD5F30960C7433FA1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl.phy"},
	{0xD5F33F985EFC1B3EULL, "vxht_brothel_n_6bk_0_0.vlm"},
	{0xD5F38F108760E049ULL, "vxht_0_hightown_n_bkh_0_0.vlm"},
	{0xD5F3969F154CB1D0ULL, "vxht_brothel_n_42o_0_0.vlm"},
	{0xD5F3E6616C213E05ULL, "6173596_m.fxe"},
	{0xD5F4BDE4FD0CCD7CULL, "vxlt_undercityserialkiller_1j3_0_0.vlm"},
	{0xD5F4C5D4CA2D5E51ULL, "one140cr_charade6134316.cl"},
	{0xD5F58FBD115624E6ULL, "vis121_saemus6101564.cl"},
	{0xD5F60B6CEF41585AULL, "gencr_templar_archer.utc"},
	{0xD5F62CB21ED0C6EFULL, "vxlt_hangedman_n_1jx_0_0.vlm"},
	{0xD5F747B2A74EAC51ULL, "hlac_w_sundermount_-327677_0y.dds"},
	{0xD5F79B5FE887C522ULL, "gen00fl_isabela6079073.cl"},
	{0xD5F7B2967C3CCA0BULL, "6079087_m.fxe"},
	{0xD5F81133565B55ABULL, "static_static_p79.psh.fxo.dx11"},
	{0xD5F880129A553269ULL, "art\\vfx\\combat\\vfx_c_profane_crust_c.dura.evt"},
	{0xD5F8E7F95882CDF4ULL, "494691267.ogg"},
	{0xD5F90270274A00CEULL, "dae000_cora6162458.cl"},
	{0xD5F95E00D17AC87AULL, "w_encounter2_camp_4_0.msh"},
	{0xD5F96C00D17ADFB3ULL, "w_encounter2_camp_4_0.mao"},
	{0xD5F97000D17AE6C0ULL, "w_encounter2_camp_4_0.mmh"},
	{0xD5F9B29C6EFC8FA8ULL, "vxw_bonepitdestroyed_a3_21k_0_0.vlm"},
	{0xD5FA00979136C728ULL, "vxw_sundermount_32u_0_1.vlm"},
	{0xD5FA0180D516592AULL, "6050384_m.fxe"},
	{0xD5FA077F40F6CCECULL, "gen_im_trs_slv_junk1_01.uti"},
	{0xD5FA6E034E548585ULL, "vxht_mansion_n_6bk_0_0.vlm"},
	{0xD5FAFF1832B9D566ULL, "qun111ip_trapdoor.utp"},
	{0xD5FB770677148F62ULL, "158575238.ogg"},
	{0xD5FC0A0CAAEC8610ULL, "197656404.ogg"},
	{0xD5FC2114C3CC8061ULL, "642370043.ogg"},
	{0xD5FC6976E1173797ULL, "6050918_m.fxe"},
	{0xD5FC85DE6FA032E1ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxb.phy"},
	{0xD5FCA89767891E6CULL, "vxca_cavern1_haunted_e41_0_0.vlm"},
	{0xD5FCDBD9F732AB5CULL, "vxht_0_hightown_a3_c14_0_0.vlm"},
	{0xD5FD07FD7CA67F6BULL, "698302588.ogg"},
	{0xD5FD0B5106455F90ULL, "hlac_w_sundermount_-131070_0x.dds"},
	{0xD5FDD1DBB68CC4F4ULL, "art\\effects\\cloudshadows\\w_bonepitmining_a2\\cloud1.dds"},
	{0xD5FDF5AEB9C01D19ULL, "lt_0_lowtown_docks_193_0.msh"},
	{0xD5FDFFAEB9C02E07ULL, "lt_0_lowtown_docks_193_0.mmh"},
	{0xD5FE03AEB9C03554ULL, "lt_0_lowtown_docks_193_0.mao"},
	{0xD5FE2CC11AE7B233ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_sidel_1.mao"},
	{0xD5FE6BE7D966324AULL, "vxlt_0_lowtown_a2_go1_0_0.vlm"},
	{0xD5FE8EDFCA41EBDCULL, "guardian_protectaura.dds"},
	{0xD5FEB9D6A41C157CULL, "6113861_m.fxe"},
	{0xD5FF80A70ED7F17CULL, "6099172_m.fxe"},
	{0xD5FF8C92B06BEAC6ULL, "rdr110dg_dockworker6015955.cl"},
	{0xD5FFA281667DF746ULL, "hlac_lt_0_lowtown_a2_n_c_-1_5y.dds"},
	{0xD6002DDA008CFA5FULL, "hlac_lt_gallowsprison_n_-1_13x.dds"},
	{0xD6003C1C22A93D02ULL, "vxlt_gallowsprison_n_64m_0_0.vlm"},
	{0xD600856458E12329ULL, "6179796_m.fxe"},
	{0xD600E829C193E205ULL, "vxlt_hangedman_n_140_0_0.vlm"},
	{0xD6014F02DC58A463ULL, "489407968.ogg"},
	{0xD6017A1ED9ABE527ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_medp_0t.dds"},
	{0xD602087C0742CA50ULL, "vxht_brothel_n_47l_0_0.vlm"},
	{0xD602C501F3A8EBC4ULL, "vxlt_0_lowtown_docks_n_cp0_0_0.vlm"},
	{0xD602D2F0D045C344ULL, "hlac_lt_warehouse_1_4z.dds"},
	{0xD6032FC1EF02E5FCULL, "vxlt_undercityserialkiller_1o0_0_0.vlm"},
	{0xD6038989BA11E6A0ULL, "6095969_m.fxe"},
	{0xD603BFF73D784761ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v2.ani"},
	{0xD6041FA468DDF782ULL, "6148135_m.fxe"},
	{0xD60440F6D8BD6FCEULL, "dae150_chandan6148654.cl"},
	{0xD604485B24816091ULL, "6079150_f.fxe"},
	{0xD604C9DD4FEED03EULL, "vxw_blightlands_9f_0_0.vlm"},
	{0xD604E8C6CD30541AULL, "gen00fl_merrill6115960.cl"},
	{0xD60568129A5F93D8ULL, "art\\vfx\\combat\\vfx_c_profane_crust_c.dura.ani"},
	{0xD6059E12D9A84F51ULL, "docks_fem_sailor_mob6176314.cl"},
	{0xD60625836C6CAFE7ULL, "hlac_lt_0_lowtown_a2_n_1_6z.dds"},
	{0xD6063DC4BD6F9377ULL, "follower_banter6169837.cl"},
	{0xD60655644A8C2AABULL, "154675748.ogg"},
	{0xD60677138C1558F4ULL, "character_character_p274.psh.fxo.dx11.assembly"},
	{0xD606F09468B918D0ULL, "w_encounter1_26_0.phy"},
	{0xD60728151C2EAED1ULL, "vxlt_0_lowtown_a3_n_c_ekl_0_0.vlm"},
	{0xD60761BD3E51C5EFULL, "hlac_lt_0_lowtown_n_-131071_1x.dds"},
	{0xD6077FE69A124D84ULL, "vxlt_0_lowtown_n_eho_0_0.vlm"},
	{0xD6082A7DCB72B347ULL, "442681686.ogg"},
	{0xD6083963A111746DULL, "6089867_f.fxe"},
	{0xD60857D74A8B5931ULL, "vxca_caverns2_8gr_0_0.vlm"},
	{0xD608848A2499CD1AULL, "6164223_m.fxe"},
	{0xD609DF23B9550B5EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_action.ani"},
	{0xD60A041DC9C21B7DULL, "hlac_w_encounter3_-327678_0x.dds"},
	{0xD60A62CEE4E1143AULL, "657436676.ogg"},
	{0xD60A6415856A437DULL, "gen00fl_anders6125285.cl"},
	{0xD60A878110EA05CDULL, "vxlt_backalley_n_1l1_0_0.vlm"},
	{0xD60ACAF73D7E7803ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v2.gad"},
	{0xD60C073BD8DC478DULL, "vis121_saemus6086668.cl"},
	{0xD60C2158364CA91AULL, "772579287.ogg"},
	{0xD60C274A0A60B712ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_masf_0n.dds"},
	{0xD60CB24E7901AA8DULL, "vxht_chantry_n_1yq_0_0.vlm"},
	{0xD60D334DFE2D0A44ULL, "376491778.ogg"},
	{0xD60D5E11D024B763ULL, "hlac_ca_cavern1_haunted_24427_0z.dds"},
	{0xD60D86D6D4E3939EULL, "673686327.ogg"},
	{0xD60D952008299D49ULL, "vxdr_primevalentrance_3bh_0_0.vlm"},
	{0xD60D9F9E68DF70ECULL, "6048133_f.fxe"},
	{0xD60DCF9468BEFF28ULL, "w_encounter1_26_0.rml"},
	{0xD60DEE5180A0874BULL, "123910140.ogg"},
	{0xD60E089836456676ULL, "ht_0_hightown_n_16_0_ll.mmh"},
	{0xD60EA4974B58A80CULL, "vxht_bartrand_n_6pg_0_0.vlm"},
	{0xD60EC8FCBD437BAFULL, "isa311_castillon6092080.cl"},
	{0xD60ED0932A420BDEULL, "718301286.ogg"},
	{0xD60EE9B7AB2FC7B4ULL, "6004167_m.fxe"},
	{0xD60EEB7FE19176ECULL, "vxlt_hangedman_n_25e_0_0.vlm"},
	{0xD60F90814A2EBD0FULL, "6154063_m.fxe"},
	{0xD60FBC4FFD85D147ULL, "plt_prm000pt_deep_blade_.plo"},
	{0xD6100123B59253EDULL, "6173956_m.fxe"},
	{0xD6100CE6487C761AULL, "llac_ht_0_hightown_a2_-131069_0.dds"},
	{0xD6101A841658F2A6ULL, "6075334_m.fxe"},
	{0xD610235242E84C3EULL, "vxht_brothel_n_6hm_0_0.vlm"},
	{0xD61045A0742BC7A8ULL, "6048765_m.fxe"},
	{0xD610E36C826AC3D5ULL, "vxca_caverns1_j3w_0_0.vlm"},
	{0xD61180CD3DD32BCAULL, "6020071_m.fxe"},
	{0xD61194AC04BDC889ULL, "vxht_bartrand_n_66l_0_0.vlm"},
	{0xD611A19EE0F8FE7CULL, "vxlt_undercityserialkiller_1d1_0_0.vlm"},
	{0xD611A7F73D845BE8ULL, "art\\characters\\creatures\\ragedemon\\animation\\atk_v2.evt"},
	{0xD61204724D3EE39AULL, "dae000cr_grand_cleric6187939.cl"},
	{0xD612166F321385D6ULL, "vxlt_undercityserialkiller_1wb_0_0.vlm"},
	{0xD6123FA6C0B611D8ULL, "354629735.ogg"},
	{0xD61290CD72A6ED2EULL, "lt_0_lowtown_docks_211_0.mmh"},
	{0xD6129CCD72A701C5ULL, "lt_0_lowtown_docks_211_0.mao"},
	{0xD612AACD72A7191CULL, "lt_0_lowtown_docks_211_0.msh"},
	{0xD6130AAB5AF687A5ULL, "weapon.mat"},
	{0xD6131925940CFE53ULL, "mag100_cullen6082319.cl"},
	{0xD6133BA24E18E3ECULL, "hlac_lt_gallowscourtyard_-131072_4y.dds"},
	{0xD6137D08D235B10DULL, "hlac_ht_0_hightown_a2_-131071_0x.dds"},
	{0xD613ABB9A468A60DULL, "129652932.ogg"},
	{0xD6144D2C94E93084ULL, "64872603.ogg"},
	{0xD61454EDE5D7BC1DULL, "vxca_cavern1_haunted_d5z_0_0.vlm"},
	{0xD6147BA1EE799D3DULL, "dae000ip_crowd_em3_talk7.utp"},
	{0xD615324236BC8913ULL, "6118135_m.fxe"},
	{0xD6153264E4C56729ULL, "lt_0_lowtown_a2_45_0.anb"},
	{0xD615CB4E2CCB47EBULL, "dre000_dougal6131206.cl"},
	{0xD6168862F9F8BC9FULL, "plt_cod_ltr_black_powder_1.nss"},
	{0xD6169C53E9A14973ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gong_n.dds"},
	{0xD616DF38581CA59AULL, "art\\characters\\creatures\\ragedemon\\animation\\blendtree\\ragedemon.anb"},
	{0xD616E8692B8FCC0DULL, "537692861.ogg"},
	{0xD616F6F4E90C2A8AULL, "vxht_mansion2b_34l_0_0.vlm"},
	{0xD616F723B95FBE2FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_action.evt"},
	{0xD617423FD631F9F0ULL, "ht_0_hightown_20_0_ll.mmh"},
	{0xD6177810DC766C67ULL, "6066343_f.fxe"},
	{0xD617C4A26F86F8EDULL, "vxdr_primevaltomb_26p_0_1.vlm"},
	{0xD6183E1C48AB89D1ULL, "fx_depthbias.mao"},
	{0xD618520369532099ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken4xa_1.mao"},
	{0xD618F95E02E01E4DULL, "vxlt_backalley_n_1w0_0_0.vlm"},
	{0xD619163D688C183DULL, "6068516_f.fxe"},
	{0xD619378F652752B9ULL, "vxlt_gallowstemplar_n_4uk_0_0.vlm"},
	{0xD619F52D5521A0CCULL, "w_blightlands_154_0.phy"},
	{0xD61A19CE0D7D9C04ULL, "zz_and_debug6134136.cl"},
	{0xD61A62A119188576ULL, "gen00fl_fenris6126485.cl"},
	{0xD61A8107FC856137ULL, "vxw_bonepit_yx_0_0.vlm"},
	{0xD61AAB7E5D016ADCULL, "vxlt_hangedman_n_4v_0_0.vlm"},
	{0xD61AB4744B99C17AULL, "w_encounter2_55_0.msh"},
	{0xD61AB4C1EEF05676ULL, "vxlt_gallowscourtyard_4w7_0_0.vlm"},
	{0xD61AC2744B99D8B3ULL, "w_encounter2_55_0.mao"},
	{0xD61AC55269B91CA9ULL, "mcr370tr_gate_guardian_aoe.ncs"},
	{0xD61AC6744B99DFC0ULL, "w_encounter2_55_0.mmh"},
	{0xD61AF3B0FC39D88AULL, "321442112.ogg"},
	{0xD61B5CADE2718A10ULL, "dae000_gamlen6005026.cl"},
	{0xD61BAD159ECC1DD7ULL, "6116977_m.fxe"},
	{0xD61BB3B80E99CA90ULL, "vxlt_0_lowtown_a2_n_eja_0_0.vlm"},
	{0xD61C0EBE87A5FF04ULL, "6103103_m.fxe"},
	{0xD61CADE67F79D5C8ULL, "515275571.ogg"},
	{0xD61CDD8E6DF2A012ULL, "dae000ip_crowd_mf7_talk16.utp"},
	{0xD61D6FD5947989B0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_low.msh"},
	{0xD61D72D8102BA0E7ULL, "vxht_mansion2b_25b_0_1.vlm"},
	{0xD61D7DD59479A1CAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_low.mmh"},
	{0xD61D81D59479A801ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3wo_low.mao"},
	{0xD61DF823B965DED8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_backhand_action.gad"},
	{0xD61E7300A7886C6DULL, "6176656_m.fxe"},
	{0xD61E82A83F63430EULL, "6028362_f.fxe"},
	{0xD61E952F34DE64BEULL, "vxht_brothel_n_6un_0_0.vlm"},
	{0xD61EE2D18ABECDD0ULL, "hlac_lt_gallowscourtyard_-65538_3y.dds"},
	{0xD61EEE7018020A61ULL, "6135636_f.fxe"},
	{0xD61EEEFB6E7ECAEBULL, "vxht_0_hightown_a2_bns_0_1.vlm"},
	{0xD61F0EBA2D867F3DULL, "6083361_m.fxe"},
	{0xD61F38E5EFB60A49ULL, "1003336779.ogg"},
	{0xD61F45E99EACAC81ULL, "vxht_0_hightown_a3_c1e_0_0.vlm"},
	{0xD61F4D377BBCB58FULL, "hlac_lt_0_lowtown_a2_n_c_-2_0z.dds"},
	{0xD61F57E39D3C0730ULL, "hlac_lt_0_lowtown_a2_-131072_9y.dds"},
	{0xD61FACF3F13E2F9EULL, "vxlt_0_lowtown_n_e30_0_0.vlm"},
	{0xD61FCE18E2EC0499ULL, "gen00fl_anders6111237.cl"},
	{0xD6200688F6B3E109ULL, "vxht_bartrand_n_63s_0_0.vlm"},
	{0xD62077CDADE7E659ULL, "vxlt_darktown_58b_0_0.vlm"},
	{0xD620825B8A708640ULL, "6082279_m.fxe"},
	{0xD620E5F5DD6F7091ULL, "506700418.ogg"},
	{0xD62115F1AB6EA14BULL, "dre000_cut_return_surface.bnk"},
	{0xD62125D90A803767ULL, "dae000_brat6156885.cl"},
	{0xD6212DC048B1C413ULL, "crt_rubble_golem.bnk"},
	{0xD621DD02088A19ABULL, "6174407_m.fxe"},
	{0xD622006D50910D8FULL, "504938442.ogg"},
	{0xD6226FFC15B680C4ULL, "zz_dae_debug6029884.cl"},
	{0xD6227866D539733AULL, "vxlt_gallowstemplar_f_4ul_0_0.vlm"},
	{0xD622F73056E0F012ULL, "6148711_m.fxe"},
	{0xD6230E55679A99CCULL, "vxlt_0_lowtown_n_h20_0_0.vlm"},
	{0xD62339DF36F65466ULL, "mcr310_argument6072294.cl"},
	{0xD623519432C9804EULL, "hurlock_axe.utc"},
	{0xD623A41F28B2A193ULL, "6117035_m.fxe"},
	{0xD62417F1AB70DA2BULL, "dre000_cut_return_surface.cnv"},
	{0xD62456B5139AF7A1ULL, "rdr111dg_martin6014675.cl"},
	{0xD6245C9EAC96F24CULL, "6101138_m.fxe"},
	{0xD62520F6446853A5ULL, "320658199.ogg"},
	{0xD625302CFB3654A1ULL, "767276097.ogg"},
	{0xD625323B821BBDB1ULL, "vxht_0_hightown_bft_0_0.vlm"},
	{0xD6253B912E778A31ULL, "vxca_caverns2_8ip_0_0.vlm"},
	{0xD625D847AA601A57ULL, "vxht_keep_5cr_0_0.vlm"},
	{0xD625F6F357B73A50ULL, "zz_mcr_debug6088423.cl"},
	{0xD62601D2009D8A9BULL, "6051980_m.fxe"},
	{0xD6262C45E7C1878EULL, "304513815.ogg"},
	{0xD6263A2A4BACCA49ULL, "vxw_encounter3_9y_0_0.vlm"},
	{0xD626BDC1C7590E16ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_023a_0t.dds"},
	{0xD626D4174B216F78ULL, "6099877_m.fxe"},
	{0xD6275A4D74608D38ULL, "76363234.ogg"},
	{0xD627751BB5643083ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_orsino_chant_ex.ani"},
	{0xD627C19D243C0F09ULL, "946514581.ogg"},
	{0xD62822A7342159ABULL, "638287821.ogg"},
	{0xD628545F9CAD59AAULL, "vxlt_0_lowtown_a2_n_c_ekj_0_0.vlm"},
	{0xD628F31A8C0E21E0ULL, "vxlt_0_lowtown_a3_n_c_hsl_0_0.vlm"},
	{0xD62A2EE1F002119FULL, "lt_0_lowtown_docks_a3_n_31_0_ll.mmh"},
	{0xD62A4D64941B1F39ULL, "character_rendertodepth_p5.vsh.fxo.dx11"},
	{0xD62A7B62FA094FB0ULL, "plt_cod_ltr_black_powder_1.plo"},
	{0xD62A809B799C1784ULL, "6104003_m.fxe"},
	{0xD62A9CF3D1E83D47ULL, "dae100_alistair_drunk6135161.cl"},
	{0xD62AB31B0F414AAFULL, "6127664_f.fxe"},
	{0xD62B60CB46A96506ULL, "gen00fl_anders6125824.cl"},
	{0xD62B81EE1F861118ULL, "562266955.ogg"},
	{0xD62B976DD1D91D2FULL, "592472763.ogg"},
	{0xD62BC2EEA4A6ED10ULL, "dae000_pc_home_ambients6208479.cl"},
	{0xD62BE80A1FED012AULL, "vxht_chantry_n_2b4_0_0.vlm"},
	{0xD62C09382F6A0565ULL, "vxlt_0_lowtown_n_e1u_0_0.vlm"},
	{0xD62C0C4ED512067FULL, "6098315-amanmageswouldlist.cut"},
	{0xD62C7B8D49F64817ULL, "510094628.ogg"},
	{0xD62C86DB356729CCULL, "84712336.ogg"},
	{0xD62C8E41B1B917C1ULL, "hlac_w_encounter1_0_1y.dds"},
	{0xD62CC4B1A9135F6EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextable.phy"},
	{0xD62CD65386CC588AULL, "6040914_m.fxe"},
	{0xD62DC70CFF68117EULL, "ser221st_gamlen.stg"},
	{0xD62E01E8ABB4E3A9ULL, "388036379.ogg"},
	{0xD62E221543267E53ULL, "gen00pt_class_race_gend.ncs"},
	{0xD62EB450985CC2B1ULL, "6168527_m.fxe"},
	{0xD62F161A84B2882FULL, "hlac_lt_0_lowtown_n_131072_5z.dds"},
	{0xD62F4F550C1751BBULL, "93212606.ogg"},
	{0xD62FA73B853CB1C1ULL, "vxlt_warehouse_13j_0_0.vlm"},
	{0xD63006E9D3FECF19ULL, "930481990.ogg"},
	{0xD630640B972B8437ULL, "pro000_cut2d_thefirstyear6005292.cl"},
	{0xD6308EB47AA1D35DULL, "358978669.ogg"},
	{0xD632040319FAAABCULL, "26364122.ogg"},
	{0xD6327EDD8F96FA12ULL, "vxlt_0_lowtown_n_gf1_0_0.vlm"},
	{0xD63297DC12FF1801ULL, "hlac_lt_0_lowtown_docks_n_-131074_8y.dds"},
	{0xD632E928683B454DULL, "544045469.ogg"},
	{0xD632EE2B670A1C88ULL, "hlac_w_encounter1_-131070_0z.dds"},
	{0xD63362F90A46025FULL, "6112634_m.fxe"},
	{0xD6338F98A8C9B4B0ULL, "hlac_lt_backalley_n_1_2z.dds"},
	{0xD633B868E6D9030EULL, "lt_warehouse.anb"},
	{0xD633CC68E6D92504ULL, "lt_warehouse.arl"},
	{0xD633E191CE50E68AULL, "897676441.ogg"},
	{0xD633EAE8CDCBBBABULL, "220031924.ogg"},
	{0xD6341CF3C72CC0DBULL, "240827363.ogg"},
	{0xD634AC073DA2E2C9ULL, "vxw_encounter3_2x_0_0.vlm"},
	{0xD634ACA25F8733ABULL, "hlac_lt_0_lowtown_docks_a3_n_0_3x.dds"},
	{0xD6360894A4D2EA79ULL, "character_character_p2.psh.fxo.dx11"},
	{0xD63635746ABE4B2CULL, "157955112.ogg"},
	{0xD637048F04DD05DCULL, "428656946.ogg"},
	{0xD6373B8BC490D0A7ULL, "vxlt_blackemporium_ci_0_0.vlm"},
	{0xD6373EA0ED5F5D6EULL, "6075966_f.fxe"},
	{0xD63764C1E0719237ULL, "vxw_bonepit_oz_0_0.vlm"},
	{0xD637847D07AD83ADULL, "510976782.ogg"},
	{0xD637847E6ED7A415ULL, "vxca_caverns2_4o3_0_1.vlm"},
	{0xD637A52AF28850A2ULL, "vxdr_0_deeproads_f59_0_0.vlm"},
	{0xD6382DE6C3C001ABULL, "vxht_brothel_n_345_0_0.vlm"},
	{0xD638A6E557B2CC78ULL, "gen00fl_sebastian6177811.cl"},
	{0xD6392E6485AD0B8AULL, "vxlt_0_lowtown_docks_a3_n_cox_0_0.vlm"},
	{0xD6393973EE5DF0ADULL, "mcr371_idunna6087281.cl"},
	{0xD639490B6CEA1000ULL, "che_firework.mao"},
	{0xD639810B51BD7679ULL, "gen_im_trs_crc_junk2_04.uti"},
	{0xD639A323E605993FULL, "vxlt_hangedman_n_25j_0_0.vlm"},
	{0xD639FB31CEC25CA5ULL, "vxht_0_hightown_a3_c0f_0_0.vlm"},
	{0xD63A06CA64A61759ULL, "mag170_thrask6016955.cl"},
	{0xD63A19D6C25A1233ULL, "6208320_m.fxe"},
	{0xD63B8A22EA454959ULL, "gen_im_cns_flu_aveline_02.uti"},
	{0xD63B8C1BB575016DULL, "art\\cinematics\\_crit path\\mage crisis\\mcr310cs_chantry_explosion\\cs_orsino_chant_ex.gad"},
	{0xD63C6C4012289B6DULL, "6135172_m.fxe"},
	{0xD63CD23CB90360F4ULL, "6163552_m.fxe"},
	{0xD63CEA42DAA01209ULL, "vxht_bartrand_n_69q_0_0.vlm"},
	{0xD63CF735B6DB47FCULL, "vxlt_undercityserialkiller_1s4_0_0.vlm"},
	{0xD63D11790E1750C7ULL, "vxw_woundedcoast_154_0_0.vlm"},
	{0xD63D2519BB42B0B6ULL, "vxlt_0_lowtown_a2_cfv_0_0.vlm"},
	{0xD63D40C76ABFE1F0ULL, "vxht_0_hightown_n_c7p_0_0.vlm"},
	{0xD63E00350958B30FULL, "llac_ht_chantry_2534_4.dds"},
	{0xD63E19187732CA41ULL, "vxlt_warehouse_16i_0_0.vlm"},
	{0xD63E56FBEB6170CDULL, "hla_ht_0_hightown_a2_65538_1y.dds"},
	{0xD63F029EEF0DF5D2ULL, "vxca_cave_1as_0_2.vlm"},
	{0xD63F1F445611605BULL, "dae000_merchant_generic_hm6172675.cl"},
	{0xD63F8D6B904BB1D3ULL, "47196169.ogg"},
	{0xD63FE28708F47946ULL, "character_character_p36.psh.fxo.dx11.assembly"},
	{0xD6402E10D17830D9ULL, "one100cr_guard.utc"},
	{0xD6406DC0A66B8E4AULL, "6042848_m.fxe"},
	{0xD64121A16BECD17BULL, "lt_0_lowtown_docks_n_165_0.mmh"},
	{0xD6412DA16BECE518ULL, "lt_0_lowtown_docks_n_165_0.mao"},
	{0xD6413FA16BED0465ULL, "lt_0_lowtown_docks_n_165_0.msh"},
	{0xD64142CCD4B681C1ULL, "art\\vfx\\creatures\\fxc_drg_brth_m_c.impa.ani"},
	{0xD641CACA77E17158ULL, "follower_banter6136476.cl"},
	{0xD641CAD06E41D067ULL, "follower_banter6169154.cl"},
	{0xD642D45507C26E75ULL, "w_bonepitdestroyed_a3_204_0.phy"},
	{0xD64367911ED68709ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_ric_0n.dds"},
	{0xD643A98BFFA1EB44ULL, "1033326833.ogg"},
	{0xD643C8AE09475BEEULL, "art\\characters\\creatures\\desiredemon\\animation\\k_bk.ani"},
	{0xD6440BBD3434DC84ULL, "6176015_m.fxe"},
	{0xD64466D66685F063ULL, "6037342_m.fxe"},
	{0xD6450998FA915506ULL, "gen00fl_varric6033567.cl"},
	{0xD6453459CF38632DULL, "53671382.ogg"},
	{0xD645BBDAC7DC8C39ULL, "vxht_0_hightown_n_65f_0_0.vlm"},
	{0xD645D4AF1DB2150EULL, "vxht_0_hightown_n_9wx_0_0.vlm"},
	{0xD645E805C253530FULL, "hlac_dr_0_deeproads_17639_1x.dds"},
	{0xD646984B4D04F958ULL, "6075986_f.fxe"},
	{0xD6469FC3B5B61A2BULL, "vxht_brothel_n_334_0_0.vlm"},
	{0xD646D4A9FA3EC5BBULL, "follower_banter6155733.cl"},
	{0xD6473019CA58225CULL, "6045098_m.fxe"},
	{0xD647457A286F206CULL, "vxht_bartrand_n_4la_0_0.vlm"},
	{0xD6474A2979C052CCULL, "vxht_0_hightown_n_bwe_0_0.vlm"},
	{0xD647F849F580A48BULL, "art\\characters\\playercharacter\\elfmale\\em_cth_comb.mao"},
	{0xD647FBA90C68039EULL, "1004643772.ogg"},
	{0xD64804C750EB1593ULL, "vxht_0_hightown_7iz_0_0.vlm"},
	{0xD6484DCCD4BCB363ULL, "art\\vfx\\creatures\\fxc_drg_brth_m_c.impa.gad"},
	{0xD64866D6B74E77B7ULL, "518514520.ogg"},
	{0xD6486D0EC0B87525ULL, "vxht_0_hightown_a3_c5e_0_0.vlm"},
	{0xD6488A3AD6F3CE57ULL, "hlac_lt_0_lowtown_a2_n_-1_4y.dds"},
	{0xD648AAB52F585D9FULL, "6114851_f.fxe"},
	{0xD649A8C06D71868EULL, "hlac_lt_0_lowtown_n_1_6y.dds"},
	{0xD64A51546D0DCE7BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_enter.evt"},
	{0xD64A787EB2F74277ULL, "hlac_lt_0_lowtown_n_-65537_0z.dds"},
	{0xD64AF695E46B31F3ULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.impa.ani"},
	{0xD64B8356000D6947ULL, "vxw_woundedcoast_107_0_0.vlm"},
	{0xD64B96234666B1E7ULL, "ser270_cut_aftermath6162269.cl"},
	{0xD64B96F6AD38C936ULL, "vxlt_0_lowtown_a2_cmw_0_0.vlm"},
	{0xD64BEBFD953E2C07ULL, "vxlt_gallowstemplar_5qw_0_0.vlm"},
	{0xD64C4ADEFB9433AAULL, "49165753.ogg"},
	{0xD64C915CCEFF5C6CULL, "defaultnormal.dds"},
	{0xD64CC141D6A0DCF9ULL, "vxht_0_hightown_a2_4n6_0_0.vlm"},
	{0xD64D3333855AD130ULL, "static_static_p70.psh.fxo.dx11"},
	{0xD64E5EC5E909483CULL, "232543128.ogg"},
	{0xD64EC54AF568BD7FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc01.msh"},
	{0xD64ECB4AF568C721ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc01.mmh"},
	{0xD64F2ACCD4C29648ULL, "art\\vfx\\creatures\\fxc_drg_brth_m_c.impa.evt"},
	{0xD64F40206C4E51A7ULL, "hlac_ht_keep_n_3732_6z.dds"},
	{0xD650C0AE0951D83FULL, "art\\characters\\creatures\\desiredemon\\animation\\k_bk.evt"},
	{0xD65124229514838FULL, "qcr220_meredith6076632.cl"},
	{0xD6514A546D13E0BCULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_enter.gad"},
	{0xD651A024FE0F0261ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.impa.ani"},
	{0xD651C27BC8C78678ULL, "dr_dungeon_3207.rml"},
	{0xD651D8AEF91BFD2EULL, "6201490_m.fxe"},
	{0xD651F6828B6BAFE3ULL, "549955125.ogg"},
	{0xD6524289B063DE4AULL, "6150714_m.fxe"},
	{0xD6529EFC1B51279AULL, "vxlt_0_lowtown_a2_gll_0_0.vlm"},
	{0xD652D579FC46582BULL, "lgt210_magu_selby6142063.cl"},
	{0xD652E7FB1654523FULL, "gen_im_trs_dal_med_glv_r1.uti"},
	{0xD653198EB0CA227EULL, "ht_brothel_n_3367.rml"},
	{0xD6533398E51CA9FFULL, "follower_banter6153973.cl"},
	{0xD653FFE4100FC42EULL, "vxlt_hangedman_n_16i_0_0.vlm"},
	{0xD65479579A7D6F78ULL, "255927677.ogg"},
	{0xD6547DF591A84C03ULL, "232412474.ogg"},
	{0xD654CD8140F5B939ULL, "bld_mac_bl.mao"},
	{0xD654F0F857FF2713ULL, "vxlt_0_lowtown_fag_0_0.vlm"},
	{0xD655057E94C8297EULL, "142141786.ogg"},
	{0xD6550E60C79612DDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl.mao"},
	{0xD6551260C7961916ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl.mmh"},
	{0xD6551C60C7962A14ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl.msh"},
	{0xD656121E69993C8AULL, "vxlt_0_lowtown_docks_a3_n_cuz_0_0.vlm"},
	{0xD6561A8986CE2EA3ULL, "vxlt_hangedman_n_2bk_0_0.vlm"},
	{0xD6561AE7F6AEF0CDULL, "vxlt_0_lowtown_docks_a3_aqx_0_0.vlm"},
	{0xD6561EDA1D566D02ULL, "ave200_donnic6110657.cl"},
	{0xD6564925BCD140F1ULL, "gen00fl_anders6125182.cl"},
	{0xD656C463A5DAC49EULL, "lgt000pt_ultra_4_03.ncs"},
	{0xD656CD59463F8FC7ULL, "6080924_m.fxe"},
	{0xD656F41B5BB2CEEEULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.impa.evt"},
	{0xD657800EF28D3C3EULL, "mag291_demon_marethari6027978.cl"},
	{0xD6579B80D994EDC6ULL, "vxht_0_hightown_a3_c6z_0_1.vlm"},
	{0xD657A1AE0957C228ULL, "art\\characters\\creatures\\desiredemon\\animation\\k_bk.gad"},
	{0xD657B2C0DB835F37ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgtp_0n.dds"},
	{0xD657EE95E475AE2AULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.impa.evt"},
	{0xD65849546D19FDBAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_enter.ani"},
	{0xD658AB24FE153303ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.impa.gad"},
	{0xD658E5B728BD18D7ULL, "gen_im_npc_wep_shd_rdra.uti"},
	{0xD65902E2B201221DULL, "vxdr_primevaltomb_7bt_0_0.vlm"},
	{0xD65905CC0EDB2CDCULL, "dae200cr_bran.utc"},
	{0xD6590B5C22F948D0ULL, "t3_smg.tnt"},
	{0xD6594BE79C035AB6ULL, "6027625_m.fxe"},
	{0xD65962830FAA1A0AULL, "6043541_f.fxe"},
	{0xD6599998A52DD274ULL, "vxw_encounter2_camp_j2_0_0.vlm"},
	{0xD659CC2D55573E97ULL, "w_blightlands_154_0.mmh"},
	{0xD659D02D555745A4ULL, "w_blightlands_154_0.mao"},
	{0xD659E22D55576409ULL, "w_blightlands_154_0.msh"},
	{0xD65A0C69FB1085FCULL, "650595408.ogg"},
	{0xD65A8D661544C0E6ULL, "6068800_m.fxe"},
	{0xD65AD1CC666E612DULL, "704292120.ogg"},
	{0xD65B06B1FC607C00ULL, "11638845.ogg"},
	{0xD65B517E2D25DFD0ULL, "vxdr_primevaltomb_3ji_0_0.vlm"},
	{0xD65BADE031D3E948ULL, "gen00fl_aveline6137710.cl"},
	{0xD65C00A6B2302393ULL, "533622961.ogg"},
	{0xD65C8E66D9C91386ULL, "mer210_coterie_contact6035363.cl"},
	{0xD65C913C40470756ULL, "mrl150_merrill6071870.cl"},
	{0xD65CEB1B5BB72A99ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.impa.gad"},
	{0xD65D0E2772CA83BCULL, "6096814_m.fxe"},
	{0xD65D59763307CAEBULL, "vxlt_foundry_n_12n_0_0.vlm"},
	{0xD65DBEB844015709ULL, "6068674_m.fxe"},
	{0xD65E5217DF380893ULL, "docks_male_laborer_mob6175391.cl"},
	{0xD65E62A587BEC8C9ULL, "vxlt_0_lowtown_a2_ety_0_0.vlm"},
	{0xD65EB70D0BF16453ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkb.ani"},
	{0xD65ECD95E47B949DULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.impa.gad"},
	{0xD65F3F64F2236561ULL, "dae000_bartrand6030618.cl"},
	{0xD65F430151BAC417ULL, "vxdr_0_deeproads_ex8_0_0.vlm"},
	{0xD65F8824FE1B16E8ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.impa.evt"},
	{0xD65F8ADE6FF4B81DULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxb.mao"},
	{0xD65F8EDE6FF4BF56ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxb.mmh"},
	{0xD65F98DE6FF4D054ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_boxb.msh"},
	{0xD65FD4730AB001C2ULL, "vxlt_0_lowtown_a2_n_eqr_0_0.vlm"},
	{0xD65FEFDEE87CAC3FULL, "art\\vfx\\combat\\vfx_glb_ground_dust_p.impa.gad"},
	{0xD6601B28938D63A1ULL, "292202947.ogg"},
	{0xD6601EAC4E89C59DULL, "6143425_m.fxe"},
	{0xD6607FC86AA9E1E2ULL, "6051260_m.fxe"},
	{0xD6609967EC5A6670ULL, "hlac_lt_undercityencounter_-1_28x.dds"},
	{0xD6609CBDEB631744ULL, "zz_approval_debug6003144.cl"},
	{0xD660D8686DEA3A73ULL, "gen00fl_anders6211902.cl"},
	{0xD66110D90D47401AULL, "vxlt_0_lowtown_a2_gcm_0_0.vlm"},
	{0xD661A1D6E992F475ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_comb_0d.dds"},
	{0xD66206CC31ABBB11ULL, "fen110ip_to_hideout.utp"},
	{0xD662BA8E53E683E0ULL, "vxlt_0_lowtown_a3_n_c_hoh_0_0.vlm"},
	{0xD6630837A4F4D2A6ULL, "dae000_orana6149077.cl"},
	{0xD6634BCA1A297043ULL, "6082243_m.fxe"},
	{0xD6636729131BEADEULL, "character_character_p7.psh.fxo"},
	{0xD6639660915F31DDULL, "follower_banter6171269.cl"},
	{0xD663F7BBBC198937ULL, "mcr370_cut_arrival_gallows.bnk"},
	{0xD663FC1B5BBD66E7ULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.impa.ani"},
	{0xD6648C3ACD04F76DULL, "vxht_0_hightown_a3_cb8_0_0.vlm"},
	{0xD6649CD792AD18F6ULL, "vxlt_0_lowtown_a2_fe6_0_0.vlm"},
	{0xD664F6C8BC1E4A65ULL, "6060487_m.fxe"},
	{0xD66506885CF277B6ULL, "6075442_m.fxe"},
	{0xD665EB5A9C02B72AULL, "ch1alpharmtnt.psh.fxo"},
	{0xD66631B78FC79381ULL, "6146800_m.fxe"},
	{0xD666BB68E7046272ULL, "lt_warehouse.rml"},
	{0xD666D4186C669914ULL, "242722841.ogg"},
	{0xD666E4DEE882B82CULL, "art\\vfx\\combat\\vfx_glb_ground_dust_p.impa.evt"},
	{0xD66723A514247F7AULL, "vxlt_0_lowtown_docks_n_cqi_0_0.vlm"},
	{0xD667D9BBBC1D3F4FULL, "mcr370_cut_arrival_gallows.cnv"},
	{0xD667DB931FF60E4AULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge1.gad"},
	{0xD667F15222C9BE6FULL, "6041695_m.fxe"},
	{0xD668295CC7A197CBULL, "6131592_m.fxe"},
	{0xD66838B2B248F2B9ULL, "vxlt_0_lowtown_docks_n_b5f_0_1.vlm"},
	{0xD668392CA5C47E74ULL, "vxdr_slavers_15n_0_0.vlm"},
	{0xD6684BC048EDFA71ULL, "crt_rubble_golem.wbd"},
	{0xD668C93CAE360590ULL, "art\\vfx\\models\\vfx_icicles01.msh"},
	{0xD668D73CAE361DAAULL, "art\\vfx\\models\\vfx_icicles01.mmh"},
	{0xD668F771092C45C9ULL, "vxw_encounter2_qo_0_1.vlm"},
	{0xD6692F0C95B6B047ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockback_v1.gad"},
	{0xD66A339DADFC3BBEULL, "s_po_splat_cl.mao"},
	{0xD66A9F1DF6153DA3ULL, "llac_lt_gamlenshouse_n_-1_7.dds"},
	{0xD66AD29636F883A8ULL, "6092300_m.fxe"},
	{0xD66BA61397B2D107ULL, "6180973_f.fxe"},
	{0xD66BAF0D0BFBE08AULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkb.evt"},
	{0xD66C9EC999E07F4DULL, "hlac_w_encounter2_camp_327684_0y.dds"},
	{0xD66CE4477012B068ULL, "vxlt_warehouse_5t_0_0.vlm"},
	{0xD66D060176FCDA26ULL, "6020466_m.fxe"},
	{0xD66D079200829AD4ULL, "372358264.ogg"},
	{0xD66DC3B1A94AF431ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextable.mmh"},
	{0xD66DCFB1A94B08D2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextable.mao"},
	{0xD66DD23E709ECB21ULL, "699559923.ogg"},
	{0xD66DDDB1A94B20EFULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextable.msh"},
	{0xD66DF1C517428F23ULL, "262016752.ogg"},
	{0xD66E66B74370D43EULL, "693877284.ogg"},
	{0xD66EA1C283CD067CULL, "824210276.ogg"},
	{0xD66F06045E839BBCULL, "dae000_arishok6081801.cl"},
	{0xD66F6880E2443B72ULL, "vxw_bonepit_xv_0_0.vlm"},
	{0xD670173E2196BC15ULL, "vxdr_primevaltomb_4h1_0_0.vlm"},
	{0xD670240C95BCBC54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockback_v1.evt"},
	{0xD67060A8FAD35007ULL, "pro000cr_soldier5.utc"},
	{0xD671027211315949ULL, "108950446.ogg"},
	{0xD67102FF8C6932A7ULL, "vxlt_blackemporium_we_0_0.vlm"},
	{0xD67111F7FE9C2C16ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_act_c.anb"},
	{0xD6720ACF52360AEEULL, "6111178_f.fxe"},
	{0xD6728E0D0C01C7FDULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkb.gad"},
	{0xD672C53FF1B1319DULL, "hlac_w_sundermount_-65533_0y.dds"},
	{0xD67301F0247D8681ULL, "741907618.ogg"},
	{0xD67329567CD13F71ULL, "vxht_keep_67g_0_0.vlm"},
	{0xD6733B428B0A57DDULL, "305584350.ogg"},
	{0xD67411C36CE2A491ULL, "hlac_lt_0_lowtown_a2_n_65536_6z.dds"},
	{0xD674371BD4AFFB29ULL, "hlac_ca_caverns2_5802_39z.dds"},
	{0xD67485A1BC2CA370ULL, "gen00fl_fenris6115377.cl"},
	{0xD6749447956DA989ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandlright_1.mao"},
	{0xD674C573840FF8CBULL, "147063617.ogg"},
	{0xD674CCDEE88ECC85ULL, "art\\vfx\\combat\\vfx_glb_ground_dust_p.impa.ani"},
	{0xD675090FFCD5BD46ULL, "gen_im_trs_gen_sta_ele_r0.uti"},
	{0xD67575E51A73966EULL, "ht_0_hightown_a3_mmap_ph.dds"},
	{0xD6758DB40C6DAC9BULL, "hlac_lt_0_lowtown_a3_n_c_-196609_0x.dds"},
	{0xD675A8E2914AB618ULL, "follower_banter6170575.cl"},
	{0xD6760D2773AA363DULL, "6143710_f.fxe"},
	{0xD6763ECD6574BCDBULL, "character_character_p189.psh.fxo.dx11"},
	{0xD6765F02EA3E19BEULL, "dae000_bartrand6016668.cl"},
	{0xD6769730738C0C98ULL, "6214490_m.fxe"},
	{0xD676ECEBE37838EDULL, "gen_im_arm_boo_hvy_act3_02.uti"},
	{0xD676F7AC838414C4ULL, "446672355.ogg"},
	{0xD67718DE0099AE74ULL, "vxht_0_hightown_n_aob_0_0.vlm"},
	{0xD677A9547E9E164DULL, "6129296_m.fxe"},
	{0xD677AB352524BF36ULL, "6169372_m.fxe"},
	{0xD677CF23B1B26FEDULL, "vxlt_gallowstemplar_n_5rk_0_0.vlm"},
	{0xD6780AF3760E99DBULL, "386650175.ogg"},
	{0xD6780C7AC2EEF43CULL, "797231481.ogg"},
	{0xD67810CC378891F0ULL, "vxht_0_hightown_a2_c1e_0_0.vlm"},
	{0xD6789962E5A4BFA6ULL, "one120_athenril6088290.cl"},
	{0xD678DAEA90CD11E8ULL, "follower_banter6154128.cl"},
	{0xD678F6C03FCBFAECULL, "gen00fl_anders6128339.cl"},
	{0xD6792400DC9DD226ULL, "6191982_m.fxe"},
	{0xD679447328EE9C28ULL, "6097000_m.fxe"},
	{0xD6797876D330A5C9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockend.phy"},
	{0xD679ACA717677861ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90pause.ani"},
	{0xD67A5E5C4222183EULL, "qun120cr_javaris6042980.cl"},
	{0xD67AB4B266602B96ULL, "mag170_recruit026094517.cl"},
	{0xD67B052326DC0BE7ULL, "w_bonepit_21_0_lc.mmh"},
	{0xD67B7E83990FDB81ULL, "cod_hst_dalish1.utp"},
	{0xD67BC179CA31D47EULL, "sound_movement_footsteps.gda"},
	{0xD67BC5561734ED97ULL, "6084241_m.fxe"},
	{0xD67BCC4AF58FB69EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc01.phy"},
	{0xD67C234F2269570CULL, "hlac_lt_0_lowtown_a3_n_c_65535_0y.dds"},
	{0xD67C7B2FFEF6DE04ULL, "412595545.ogg"},
	{0xD67CB69320082B30ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge1.ani"},
	{0xD67D1C0C95C7388DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\knockback_v1.ani"},
	{0xD67DDA5DD43A53F2ULL, "vxw_bonepit_cw_0_0.vlm"},
	{0xD67DDE142D094FF0ULL, "l_env_wal_gallows_ltcough.bnk"},
	{0xD67E06A43D42764FULL, "mag210_arianni6057325.cl"},
	{0xD67E65701C8659EFULL, "6163894_m.fxe"},
	{0xD67EB5A05E179205ULL, "gen00fl_sebastian6180810.cl"},
	{0xD67EC7E431B1894FULL, "6122808_m.fxe"},
	{0xD67F07484E3100AFULL, "seb221_flora_harimann6177708.cl"},
	{0xD67F14F7FEA86E81ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_act_c.mmh"},
	{0xD6801A5DE87E436CULL, "mcr371_orsino6083328.cl"},
	{0xD6809715E774C14AULL, "qun160_vashothlead6053885.cl"},
	{0xD680B7A7176DA903ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90pause.gad"},
	{0xD6812BCF286F64BFULL, "llac_lt_0_lowtown_docks_a3_-131074_3.dds"},
	{0xD6817113EB4AD044ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_012a_0n.dds"},
	{0xD6819B336EC757F1ULL, "vxht_keep_60f_0_0.vlm"},
	{0xD681FADA857A8F99ULL, "6135349_m.fxe"},
	{0xD68297EB2603849EULL, "634885370.ogg"},
	{0xD683452E932C1D58ULL, "hlac_w_encounter2_327681_0y.dds"},
	{0xD68359CA72B7FB72ULL, "6047380_m.fxe"},
	{0xD68390E481FA32CBULL, "mag200_feynriel6023969.cl"},
	{0xD68393907E0771FBULL, "181510496.ogg"},
	{0xD683961343DE7441ULL, "gencr_raider_ws.utc"},
	{0xD6842777E1F0AC44ULL, "199666698.ogg"},
	{0xD684A56341D80CE9ULL, "vxlt_0_lowtown_a2_n_ews_0_0.vlm"},
	{0xD685150CD5CCA1B0ULL, "117850372.ogg"},
	{0xD685251AEE608831ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_action.ani"},
	{0xD685306FCE67F673ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_rnf.ani"},
	{0xD6862C6440E112F1ULL, "mcr310_sibling6076401.cl"},
	{0xD6869322FBC3E9CAULL, "6208448_m.fxe"},
	{0xD686DA1F53A004F4ULL, "vxlt_0_lowtown_docks_cov_0_0.vlm"},
	{0xD68704D88F318930ULL, "6077315_f.fxe"},
	{0xD6873575A305E3A5ULL, "136199118.ogg"},
	{0xD68794A717738CE8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\pauseturnl90pause.evt"},
	{0xD687D935EB934F19ULL, "face_tsb_p0.psh.fxo.dx11"},
	{0xD68819B927E70AF7ULL, "vxlt_gallowscourtyard_n_4vb_0_0.vlm"},
	{0xD688B479001DED41ULL, "gen00fl_aveline6152536.cl"},
	{0xD688C10A3098A159ULL, "gen_im_trs_drk_one_axe_r0.uti"},
	{0xD688D9F834EED486ULL, "464253352.ogg"},
	{0xD689E75507FE9090ULL, "w_bonepitdestroyed_a3_204_0.msh"},
	{0xD689F55507FEA8AAULL, "w_bonepitdestroyed_a3_204_0.mmh"},
	{0xD689F95507FEAFE1ULL, "w_bonepitdestroyed_a3_204_0.mao"},
	{0xD68A05B670C7F4CAULL, "vxlt_gallowstemplar_n_652_0_0.vlm"},
	{0xD68A174A3F4BE5A2ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpa.msh"},
	{0xD68A194A3F4BE958ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpa.mmh"},
	{0xD68A599098E0076DULL, "mcr371_post_harvester6083689.cl"},
	{0xD68A71452A6711EBULL, "pro000_meeran6048847.cl"},
	{0xD68ACFBA5AE2D9C7ULL, "6046834_m.fxe"},
	{0xD68AD2BC506660E0ULL, "350313940.ogg"},
	{0xD68B8F9797F7456BULL, "6083086_m.fxe"},
	{0xD68BCC7BFF45FBC1ULL, "gen00fl_anders6107667.cl"},
	{0xD68BE8CF7BA7BC7BULL, "637766518.ogg"},
	{0xD68C0A2077D10121ULL, "211616273.ogg"},
	{0xD68C301AEE66B9D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_action.gad"},
	{0xD68C4C3AC6306C72ULL, "vxw_bonepit_np_0_0.vlm"},
	{0xD68C8C4634AE5155ULL, "452296751.ogg"},
	{0xD68D168EB0FB9522ULL, "ht_brothel_n_3367.anb"},
	{0xD68D291DDCC294E8ULL, "llac_lt_0_lowtown_65537_1.dds"},
	{0xD68D49E30A954FADULL, "hlac_lt_0_lowtown_a2_n_-1_5x.dds"},
	{0xD68D83FF5123C5F2ULL, "mag101_grace6034227.cl"},
	{0xD68E14E8E2DAC297ULL, "6071209_m.fxe"},
	{0xD68E1F4381561D92ULL, "dre180cr_rock_wraith.ncs"},
	{0xD68EB80DD3EB8B52ULL, "vxlt_gallowscourtyard_61w_0_0.vlm"},
	{0xD68F0BB5FD6F55B8ULL, "vis160cr_qunari.utc"},
	{0xD68F23DC42449841ULL, "199029054.ogg"},
	{0xD68F63ECCED1324BULL, "zz_slate_3.cut"},
	{0xD68FA66871ACEA89ULL, "621326867.ogg"},
	{0xD68FEA49831B5BC4ULL, "hlac_w_encounter3_-327681_0y.dds"},
	{0xD68FEBD941920D13ULL, "llac_ca_caverns2prc_5802_18.dds"},
	{0xD6901DDCEE06AAFCULL, "dae000_surprises6157194.cl"},
	{0xD6906CB77770A819ULL, "6130049_m.fxe"},
	{0xD690C8779CB2605AULL, "lowtown_fem_alien_elf_2_mob6166361.cl"},
	{0xD6912D1AED389ABFULL, "6070905_m.fxe"},
	{0xD691B8BC7501D86CULL, "263122170.ogg"},
	{0xD6921B4918990EC5ULL, "vxlt_0_lowtown_docks_n_cqb_0_0.vlm"},
	{0xD692286FCE7272AAULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_rnf.evt"},
	{0xD6922D1AEE6B2038ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\danzig_threaten_action.evt"},
	{0xD69267218EAF2F5CULL, "lt_0_lowtown_docks_a3_248_0.mao"},
	{0xD69273218EAF43FFULL, "lt_0_lowtown_docks_a3_248_0.mmh"},
	{0xD69279218EAF4DA1ULL, "lt_0_lowtown_docks_a3_248_0.msh"},
	{0xD692B17E811231EDULL, "vxdr_0_deeproads_aay_0_0.vlm"},
	{0xD692FDDBD41D4D3BULL, "rdr111dg_assistant6053161.cl"},
	{0xD6931348AECC92F5ULL, "vxht_0_hightown_a2_bem_0_0.vlm"},
	{0xD693174033CE2569ULL, "vxlt_0_lowtown_a2_n_exr_0_0.vlm"},
	{0xD6932BE2FBA2453FULL, "199521734.ogg"},
	{0xD6939198CC736FA5ULL, "1033979768.ogg"},
	{0xD693B524B6258FBCULL, "dae000_merchant_access_hf6162978.cl"},
	{0xD693EB2CFFFAC7DEULL, "6072865_m.fxe"},
	{0xD6941F73E0D07D80ULL, "s_ballista_05_l3.mao"},
	{0xD6948EA375662032ULL, "6119119_m.fxe"},
	{0xD6948EEF0910F036ULL, "6163172_m.fxe"},
	{0xD694B2DD959EA0EDULL, "vxlt_gallowstemplar_n_5hi_0_0.vlm"},
	{0xD694D99E74B9D6DDULL, "and121_karl6093896.cl"},
	{0xD694DFD7BD906557ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qm_exp_movement.anb"},
	{0xD694EB341873B90EULL, "6103237_m.fxe"},
	{0xD695204A3F55FCE3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_cashstumpa.phy"},
	{0xD69555DD277FD120ULL, "follower_banter6155646.cl"},
	{0xD695D0C1F26709BEULL, "930374912.ogg"},
	{0xD696C1514A054F73ULL, "gen00fl_dog6109323.cl"},
	{0xD697218F03767634ULL, "6106315_m.fxe"},
	{0xD6974342844D2CACULL, "707285049.ogg"},
	{0xD697BF2A06F49378ULL, "vxlt_undercitydungeon_1fn_0_0.vlm"},
	{0xD6986921944310FFULL, "vxht_mansion_n_46w_0_0.vlm"},
	{0xD698B3CD23F2069BULL, "497168723.ogg"},
	{0xD699076FCE78591DULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90l_rnf.gad"},
	{0xD699374A53CB2AE6ULL, "6136776_m.fxe"},
	{0xD69946F9614356F1ULL, "574865772.ogg"},
	{0xD699604EECFF714FULL, "dae221_viscount_chantry6154260.cl"},
	{0xD699B53930CC2D85ULL, "bl_spa_sh.mao"},
	{0xD699E77DFBBA308EULL, "gen_im_trs_gen_lgt_hlm_r2.uti"},
	{0xD699FD61D10C3A7BULL, "6135620_m.fxe"},
	{0xD69AB509921D13B7ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriume.mmh"},
	{0xD69AB909921D1AC4ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriume.mao"},
	{0xD69ACB09921D3929ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_newlyriume.msh"},
	{0xD69AED0BAAE97C54ULL, "one130_thug6131574.cl"},
	{0xD69B9CB0A3FCE02CULL, "qun211_braggart6095483.cl"},
	{0xD69BDA620572E6C9ULL, "and230st_base_post_terrors.stg"},
	{0xD69C41E77D4CE12EULL, "hlac_lt_0_lowtown_docks_a3_65535_3x.dds"},
	{0xD69C5300BCA7837BULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm06.mmh"},
	{0xD69C7100BCA7B665ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm06.msh"},
	{0xD69C7267E5B15B38ULL, "56080700.ogg"},
	{0xD69C78B8100AE576ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjr_strt.ani"},
	{0xD69CB4F212A97E10ULL, "hlac_w_bonepitmining_a2_131076_0x.dds"},
	{0xD69D09C5080AAFA0ULL, "dae000_keep_guard_2_sta6163815.cl"},
	{0xD69D92BD5214FE0BULL, "p_spit.mao"},
	{0xD69E057EADB4FB6AULL, "6117823_m.fxe"},
	{0xD69E1B8D4AD5ABE7ULL, "209332466.ogg"},
	{0xD69E424ACCA4ECEFULL, "lt_0_lowtown_a2_54_0.rml"},
	{0xD69E42B4055C2937ULL, "230591250.ogg"},
	{0xD69E79B1794BCB54ULL, "llac_lt_0_lowtown_docks_n_-131074_6.dds"},
	{0xD69EA4C4F0FF3AD9ULL, "art\\vfx\\textures\\vfx_blood_pool_grow.dds"},
	{0xD69EABCF142FFF27ULL, "6160730_m.fxe"},
	{0xD69EAEC77197F6E8ULL, "art\\characters\\weapons\\unique\\meredith\\w_gsw_014a.msh"},
	{0xD69EAFA49A10C218ULL, "mer210_seldrek6078809.cl"},
	{0xD69EC0C771981579ULL, "art\\characters\\weapons\\unique\\meredith\\w_gsw_014a.mao"},
	{0xD69ECCC771982912ULL, "art\\characters\\weapons\\unique\\meredith\\w_gsw_014a.mmh"},
	{0xD69ED7EA4B5C0D50ULL, "114156158.ogg"},
	{0xD69F2D831AF7AF62ULL, "qun131_taskmaster6039775.cl"},
	{0xD69F95717FF4E648ULL, "173758182.ogg"},
	{0xD69FA933BE6B19AEULL, "hlac_lt_0_lowtown_1_9y.dds"},
	{0xD69FEA7EBAABAA29ULL, "493298024.ogg"},
	{0xD6A05D5894CFF1A7ULL, "vxht_mansion2b_280_0_0.vlm"},
	{0xD6A0628A88454C58ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls107.mao"},
	{0xD6A07588D26A7CCAULL, "853047537.ogg"},
	{0xD6A0A669EDF95DDFULL, "6070498_f.fxe"},
	{0xD6A1654A2854115CULL, "mag121_idunna6084881.cl"},
	{0xD6A172599F2A683BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01b.msh"},
	{0xD6A190599F2A9B25ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01b.mmh"},
	{0xD6A2B8D8E6142BDEULL, "28315618.ogg"},
	{0xD6A2C0142D2838BAULL, "l_env_wal_gallows_ltcough.wbd"},
	{0xD6A3159A4E80EA00ULL, "6061355_m.fxe"},
	{0xD6A324BA8794B96DULL, "vxlt_gallowstemplar_n_5ch_0_0.vlm"},
	{0xD6A348716A6DB766ULL, "hlac_w_encounter2_262148_0y.dds"},
	{0xD6A35D110A69D18EULL, "6104337_m.fxe"},
	{0xD6A3FE11C8599B20ULL, "789471820.ogg"},
	{0xD6A43922792395EBULL, "321811830.ogg"},
	{0xD6A43C57DBAF8EF2ULL, "vxlt_0_lowtown_a3_n_c_e8j_0_0.vlm"},
	{0xD6A457F981EF4ECCULL, "hlac_lt_0_lowtown_docks_196608_0z.dds"},
	{0xD6A4770ACF3200A8ULL, "vxdr_primevalentrance_23g_0_0.vlm"},
	{0xD6A484CCD56ADE30ULL, "da2terrain_terrain_tools_p2.psh.fxo"},
	{0xD6A49D6E3A7A1B92ULL, "ts_raiders_acc_props.gda"},
	{0xD6A4C6B36D397335ULL, "hlac_lt_0_lowtown_a2_n_65537_3z.dds"},
	{0xD6A4D09BC7C39DBEULL, "vxw_sundermount_2ct_0_0.vlm"},
	{0xD6A530E4B8769C14ULL, "585796774.ogg"},
	{0xD6A559F19ECC9863ULL, "vxlt_0_lowtown_a2_gvc_0_0.vlm"},
	{0xD6A63106F8EAABF8ULL, "vxlt_undercitydungeon_1io_0_0.vlm"},
	{0xD6A66176D3576CEEULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockend.mmh"},
	{0xD6A66D76D3578085ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockend.mao"},
	{0xD6A67B76D35798DCULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockend.msh"},
	{0xD6A6EF5CFAD879CAULL, "qun160_vashothlead6064681.cl"},
	{0xD6A6FB1C8E545BD0ULL, "zz_followers_banter6157840.cl"},
	{0xD6A72D40D260DB9EULL, "825152647.ogg"},
	{0xD6A7BD292F3C4087ULL, "6142806_m.fxe"},
	{0xD6A7CBE2F9B2A4F2ULL, "305220039.ogg"},
	{0xD6A7DB5A48B5F0DDULL, "6045792_m.fxe"},
	{0xD6A814EDC26B76B3ULL, "25335832.ogg"},
	{0xD6A83E65E7195152ULL, "839655631.ogg"},
	{0xD6A905292F588677ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a.phy"},
	{0xD6A906283978CFC3ULL, "470973375.ogg"},
	{0xD6A980B810157D67ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjr_strt.evt"},
	{0xD6A9BA5F285FCEE7ULL, "993798684.ogg"},
	{0xD6AA666FD51F1C27ULL, "6150737_m.fxe"},
	{0xD6AAB1C9201906DFULL, "dre000_dougal6129746.cl"},
	{0xD6AAF3A970225637ULL, "vxw_bonepit_7r_0_0.vlm"},
	{0xD6AB1222C6032B61ULL, "llac_ca_caverns1_24422_2.dds"},
	{0xD6AB17EF771A308AULL, "lgt000cr_hate_tevinter.utc"},
	{0xD6ABC9F28ADB70BDULL, "795701572.ogg"},
	{0xD6ABE650128CD8ADULL, "gen00fl_carver6104598.cl"},
	{0xD6AC5B83BC452472ULL, "6176664_m.fxe"},
	{0xD6ACB03242EB1371ULL, "426707632.ogg"},
	{0xD6AD22A8DFD14DC4ULL, "6033998_m.fxe"},
	{0xD6AEC534FF490BE4ULL, "art\\characters\\creatures\\mubari\\c_mubari.mmh"},
	{0xD6AEDB34FF493086ULL, "art\\characters\\creatures\\mubari\\c_mubari.msh"},
	{0xD6AEE1206CE0B846ULL, "pro000_colonel_ewald6039965.cl"},
	{0xD6AF69B0E1B4A8B0ULL, "6151096_m.fxe"},
	{0xD6AF983DAAED855BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_action.ani"},
	{0xD6AFA1457F1B8F4CULL, "47514465.ogg"},
	{0xD6AFFE63805A22A5ULL, "vxlt_0_lowtown_a2_n_i4f_0_0.vlm"},
	{0xD6B001FBC1281E8FULL, "vxca_caverns2_7ap_0_1.vlm"},
	{0xD6B0235080E9EFF3ULL, "vxca_caverns1_cot_0_0.vlm"},
	{0xD6B05331494F6240ULL, "gen00fl_aveline6116625.cl"},
	{0xD6B08A4153F2DE4AULL, "vxht_0_hightown_a3_b3y_0_0.vlm"},
	{0xD6B0A1B8101BD3B0ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjr_strt.gad"},
	{0xD6B0EA6F2184DB59ULL, "vxlt_darktown_5nh_0_0.vlm"},
	{0xD6B143145F521E4CULL, "gen_im_trs_gen_junk1_04.uti"},
	{0xD6B145982948A6F7ULL, "hlac_lt_0_lowtown_a2_n_65534_2z.dds"},
	{0xD6B16196C3CE9046ULL, "6170068_m.fxe"},
	{0xD6B1633AE2981921ULL, "hlac_lt_darktown_-1_45z.dds"},
	{0xD6B17C4F1E5DDC0FULL, "548617181.ogg"},
	{0xD6B1B77F4739D85CULL, "216645954.ogg"},
	{0xD6B24ED312F50192ULL, "652707931.ogg"},
	{0xD6B264EDF1AA1D37ULL, "hlac_lt_gallowstemplar_n_2391_9z.dds"},
	{0xD6B2A8D195391F1AULL, "6081588_m.fxe"},
	{0xD6B2E3FED83BD56CULL, "vxlt_darktown_50s_0_0.vlm"},
	{0xD6B33B092007C34AULL, "6044293_m.fxe"},
	{0xD6B34278B9B9B63EULL, "vxw_sundermount_2du_0_0.vlm"},
	{0xD6B3903014F19B12ULL, "vxdr_primevaltomb_7qf_0_0.vlm"},
	{0xD6B412438176D764ULL, "dre180cr_rock_wraith.utc"},
	{0xD6B41657EF019A09ULL, "6086920_m.fxe"},
	{0xD6B494D602E72950ULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn_low.mmh"},
	{0xD6B4A2D602E7416AULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn_low.msh"},
	{0xD6B5042554ADBE10ULL, "vxlt_0_lowtown_docks_n_bv7_0_0.vlm"},
	{0xD6B54A2D13117485ULL, "hlac_lt_0_lowtown_a2_131070_0x.dds"},
	{0xD6B5857E3BB8347BULL, "vxlt_backalley_g_1i1_0_0.vlm"},
	{0xD6B595CCA1B939F0ULL, "381522355.ogg"},
	{0xD6B5C4258645FB57ULL, "6040872_m.fxe"},
	{0xD6B5F94B39C8CFDFULL, "gen00fl_fenris6126954.cl"},
	{0xD6B6289AC473D9DFULL, "face_face_p6.psh.fxo"},
	{0xD6B684CA45DD4FFBULL, "hlac_ht_mansion2b_818_28y.dds"},
	{0xD6B69E65F26BBB8EULL, "llac_w_encounter2_camp_-131068_0.dds"},
	{0xD6B6CCC579903002ULL, "vxca_caverns2prc_9dx_0_0.vlm"},
	{0xD6B716944994C134ULL, "mcr310_sibling6076293.cl"},
	{0xD6B787D2FAF59A52ULL, "hlac_w_encounter2_-65535_0y.dds"},
	{0xD6B7B181A7963361ULL, "pro000_merchant6051647.cl"},
	{0xD6B7C58A1F2017BDULL, "vxht_mansion_n_4hs_0_0.vlm"},
	{0xD6B7E4D4615A654AULL, "dae000_surprises6157189.cl"},
	{0xD6B7FC92C6D94BC8ULL, "6108238_m.fxe"},
	{0xD6B81B4C57315BE0ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graveb.phy"},
	{0xD6B941268FD98F5EULL, "192715236.ogg"},
	{0xD6B94D81013F8397ULL, "hlac_lt_0_lowtown_65537_4z.dds"},
	{0xD6B98489FD59300BULL, "qun110_petrice6042147.cl"},
	{0xD6B9BE02F3219CA9ULL, "480083668.ogg"},
	{0xD6B9C7CD04488C85ULL, "964926333.ogg"},
	{0xD6B9CF80463D3C51ULL, "547610444.ogg"},
	{0xD6BA330B15738D41ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_avea_0n.dds"},
	{0xD6BB23B847BDF984ULL, "6176204_m.fxe"},
	{0xD6BB31F49DBCB71CULL, "337270888.ogg"},
	{0xD6BB4A52E1A84B92ULL, "6090869_m.fxe"},
	{0xD6BB5D0D15205945ULL, "6069686_m.fxe"},
	{0xD6BB85B032327100ULL, "zz_spy130cr_darktown_res_2.utc"},
	{0xD6BC200E3FC4C9B2ULL, "vxlt_0_lowtown_a2_n_g94_0_0.vlm"},
	{0xD6BC24A8BEF04611ULL, "mrl351_pridedemon6165577.cl"},
	{0xD6BC74845744B4F4ULL, "vxdr_primevalentrance_2ea_0_0.vlm"},
	{0xD6BCE7B161621D7BULL, "gen_im_trs_dal_one_lsw_r2.uti"},
	{0xD6BD52BC2E48850BULL, "vxlt_0_lowtown_docks_n_b0b_0_0.vlm"},
	{0xD6BDA03DAAF9D052ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_action.evt"},
	{0xD6BE50F7722776F3ULL, "766287353.ogg"},
	{0xD6BE8F71D1110B93ULL, "hlac_lt_0_lowtown_a3_n_c_131071_3x.dds"},
	{0xD6BED809B6FFC666ULL, "llac_lt_0_lowtown_a2_65537_1.dds"},
	{0xD6BEFB4187FDD65CULL, "hlac_w_sundermount_-1_0y.dds"},
	{0xD6BF27F6D4498B62ULL, "hlac_lt_0_lowtown_a2_-131072_2z.dds"},
	{0xD6BF3A8A46F556BAULL, "vxlt_gallowstemplar_n_4xs_0_0.vlm"},
	{0xD6BF522EBA6828F4ULL, "vxht_0_hightown_n_atg_0_0.vlm"},
	{0xD6BF9922CECA046CULL, "vxw_woundedcoast_1f6_0_0.vlm"},
	{0xD6BFD24ABBA6C3F0ULL, "character_rendertodepth_p4.vsh.fxo.dx11.assembly"},
	{0xD6C0737E967969FEULL, "dae100_bran6118292.cl"},
	{0xD6C0F995E4A8C404ULL, "6095006_m.fxe"},
	{0xD6C155DBCA31EDECULL, "vxlt_darktown_55p_0_0.vlm"},
	{0xD6C1A2545791DAF4ULL, "follower_barks6172008.cl"},
	{0xD6C1B455ABAFCEBEULL, "vxw_sundermount_2iv_0_0.vlm"},
	{0xD6C20F551B8FB533ULL, "vxlt_gallowsprison_n_67m_0_0.vlm"},
	{0xD6C297E36ED3C4ACULL, "6141134_m.fxe"},
	{0xD6C2CF14D74757D5ULL, "hlac_w_bonepitmining_a2_327684_0y.dds"},
	{0xD6C3991A7C50A4C7ULL, "hlac_ht_0_hightown_a2_-131071_0z.dds"},
	{0xD6C3BEFBC699C8F4ULL, "6155290_m.fxe"},
	{0xD6C3E18E7ED41955ULL, "vxw_encounter2_camp_iz_0_0.vlm"},
	{0xD6C3EE092F442749ULL, "6170566_m.fxe"},
	{0xD6C405626F7BB6ABULL, "6198675_f.fxe"},
	{0xD6C4BF3DAB0023E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\slap_forehand_action.gad"},
	{0xD6C4FC9776BC1CC7ULL, "444302372.ogg"},
	{0xD6C50A176FA5B623ULL, "gen00fl_bethany6101272.cl"},
	{0xD6C6500FF6336B18ULL, "6162973_m.fxe"},
	{0xD6C65E694BDA4DEEULL, "llac_lt_0_lowtown_-65535_4.dds"},
	{0xD6C6808358D00BACULL, "111779958.ogg"},
	{0xD6C716ADA4F4103BULL, "zz_fol_debug6092329.cl"},
	{0xD6C768C1690114DAULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyo.mmh"},
	{0xD6C76CC169011B11ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyo.mao"},
	{0xD6C77AC169013320ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyo.msh"},
	{0xD6C7E6AEC87C2417ULL, "hlac_w_blightlands_0_3y.dds"},
	{0xD6C82DDE39295C28ULL, "vxht_0_hightown_a2_by0_0_0.vlm"},
	{0xD6C831C311DC7FF9ULL, "vxlt_undercityserialkiller_xo_0_0.vlm"},
	{0xD6C8A6614CC2939DULL, "363552106.ogg"},
	{0xD6C8B4CB5FD209E2ULL, "gen00fl_varric6078126.cl"},
	{0xD6C8C15974DFA6F4ULL, "one000_ilen6037069.cl"},
	{0xD6C8CCC55735FDBDULL, "mag120_macha6091374.cl"},
	{0xD6C9921119922A4EULL, "6079073_m.fxe"},
	{0xD6C9DCFFF4C45D75ULL, "vxlt_0_lowtown_a2_hek_0_0.vlm"},
	{0xD6CA126A365A0D3BULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn05.phy"},
	{0xD6CA25BC2037CFAAULL, "vxlt_0_lowtown_a2_n_c_hm5_0_0.vlm"},
	{0xD6CAA08088796712ULL, "vxca_cave_238_0_0.vlm"},
	{0xD6CAA08CADCF07C4ULL, "hlac_lt_backalley_n_65536_1y.dds"},
	{0xD6CAF0BD6ACD48E9ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_abma_0n.dds"},
	{0xD6CBEB12DC47E181ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.resetup.ani"},
	{0xD6CC6CA4043BB322ULL, "vxdr_primevalentrance_f6_0_0.vlm"},
	{0xD6CC838F7F7086F6ULL, "6070135_m.fxe"},
	{0xD6CCFFC12F848A62ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\leftlegram.gad"},
	{0xD6CD01366A3871B4ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur3_flemeth_fake.ani"},
	{0xD6CD5A40E637DA80ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_4z.dds"},
	{0xD6CD84A7458A0DA2ULL, "vxlt_gallowscourtyard_n_654_0_0.vlm"},
	{0xD6CE0AFFC0C01CECULL, "vxw_woundedcoast_1a7_0_0.vlm"},
	{0xD6CE4F1FE23C4C3FULL, "lt_gallowscourtyard_n_49_0.rml"},
	{0xD6CE8234E078DB3BULL, "926631686.ogg"},
	{0xD6CE8E03E94B2101ULL, "12674649.ogg"},
	{0xD6CEE0A2EB5C47B9ULL, "vxht_keep_6go_0_0.vlm"},
	{0xD6CF2204B5EA7D0AULL, "llac_w_bonepitdestroyed_a3_393218_0.dds"},
	{0xD6CF2FBD84FE8E96ULL, "vfx_blood01.dds"},
	{0xD6CF72704CBE0371ULL, "gen00fl_dog6167659.cl"},
	{0xD6CF86B4E1486AC7ULL, "6044883_m.fxe"},
	{0xD6CF9F8B7108816DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_basic.mmh"},
	{0xD6CFA38B710887BEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_basic.mao"},
	{0xD6CFB18B71089FF3ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_basic.msh"},
	{0xD6D026329DA5E73EULL, "vxw_sundermount_2rw_0_0.vlm"},
	{0xD6D037FA697DF3ABULL, "vxlt_0_lowtown_a2_n_c_coy_0_0.vlm"},
	{0xD6D05DAA89A5F44AULL, "942054219.ogg"},
	{0xD6D079128556B659ULL, "w_sundermount_224_0.msh"},
	{0xD6D083128556C747ULL, "w_sundermount_224_0.mmh"},
	{0xD6D087128556CD94ULL, "w_sundermount_224_0.mao"},
	{0xD6D0912142724F68ULL, "mer120_hubert6059428.cl"},
	{0xD6D0E3B9C8EA2896ULL, "mist_dark.mao"},
	{0xD6D112FF4D936718ULL, "zz_area_signage12.utp"},
	{0xD6D190A8A7713D1DULL, "6027598_m.fxe"},
	{0xD6D1F612DC4C6023ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.resetup.gad"},
	{0xD6D1FBADBF89DCBFULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grasslrg.msh"},
	{0xD6D1FF263743C1CBULL, "6051399_m.fxe"},
	{0xD6D201ADBF89E761ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grasslrg.mmh"},
	{0xD6D22314786CE575ULL, "fen210_fenris6108749.cl"},
	{0xD6D2536B70CA31D5ULL, "vxw_encounter2_camp_dy_0_0.vlm"},
	{0xD6D2B092E0E0D27CULL, "hm_cavril.mor"},
	{0xD6D322C12F893159ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\leftlegram.evt"},
	{0xD6D39886A70CA61AULL, "gen00fl_merrill_final.stg"},
	{0xD6D3ADF4587D3A70ULL, "100580213.ogg"},
	{0xD6D3C7AD0131703EULL, "follower_banter6172760.cl"},
	{0xD6D3FE366A3E8B16ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur3_flemeth_fake.gad"},
	{0xD6D40B33BFBE7405ULL, "weapon1_ds2y.psh.fxo"},
	{0xD6D4889875525D85ULL, "vxlt_undercitydungeon_1iv_0_0.vlm"},
	{0xD6D4E04CAAC57CC8ULL, "6106038_m.fxe"},
	{0xD6D577B95DA3160BULL, "764995090.ogg"},
	{0xD6D6805E6A8A7ADEULL, "combat_dw_r1_combo.gda"},
	{0xD6D69F6075F5AA96ULL, "zz_dae_debug6047590.cl"},
	{0xD6D6E84CEA293347ULL, "6172430_m.fxe"},
	{0xD6D6F1BCD9A60393ULL, "vxlt_0_lowtown_f2n_0_0.vlm"},
	{0xD6D6F9AC894A46FBULL, "6128639_m.fxe"},
	{0xD6D70682904793C8ULL, "527668097.ogg"},
	{0xD6D72947359D8FAEULL, "1070462639.ogg"},
	{0xD6D76694A241287BULL, "235670941.ogg"},
	{0xD6D79EF5BB391963ULL, "49033859.ogg"},
	{0xD6D7F6B9E18861C2ULL, "hlac_lt_0_lowtown_docks_-65538_2z.dds"},
	{0xD6D831338147BD34ULL, "hlac_lt_0_lowtown_a2_n_-196609_2x.dds"},
	{0xD6D84EDCE6BA75F5ULL, "vxlt_0_lowtown_a2_hbj_0_0.vlm"},
	{0xD6D8D312DC524308ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_trap.resetup.evt"},
	{0xD6D8DA20997FB14AULL, "912500935.ogg"},
	{0xD6D8E2BB83827C26ULL, "gen_im_trs_apo_lgt_boo_r1.uti"},
	{0xD6D9583E3B30E5F4ULL, "vxdr_primevalentrance_2og_0_0.vlm"},
	{0xD6DA2A92E8789C57ULL, "vxht_brothel_n_5kz_0_0.vlm"},
	{0xD6DA6E142AFB41FBULL, "gen_im_cns_tom_rog_bard.uti"},
	{0xD6DA769DD278A074ULL, "812663372.ogg"},
	{0xD6DA7E7C42D24066ULL, "vxw_bonepitmining_a2_ac_0_0.vlm"},
	{0xD6DAF56C71669F76ULL, "6075235_m.fxe"},
	{0xD6DAF9537CEF6F67ULL, "6018891_m.fxe"},
	{0xD6DB61F61B338D7FULL, "plt_cod_hst_mag_deepmush.plo"},
	{0xD6DB76538B4A6D16ULL, "hlac_w_bonepit_262145_0x.dds"},
	{0xD6DBF4E8A1D62C63ULL, "mag101_grace6034080.cl"},
	{0xD6DC7CDCB2B6356CULL, "vxw_woundedcoast_1l4_0_0.vlm"},
	{0xD6DE21D141D12FFDULL, "6075706_m.fxe"},
	{0xD6DE384C57525C75ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graveb.msh"},
	{0xD6DE3884F7C3E723ULL, "hlac_ht_brothel_n_3365_0y.dds"},
	{0xD6DE3A4C57525FEBULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graveb.mmh"},
	{0xD6DE464C57527388ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_graveb.mao"},
	{0xD6DE5643E8F205BCULL, "hlac_lt_warehouse_65537_2y.dds"},
	{0xD6DE980F8F9BFFBEULL, "vxw_sundermount_2wh_0_0.vlm"},
	{0xD6DF15B9FA882B8BULL, "vxca_caverns2_953_0_0.vlm"},
	{0xD6DF702A210A444CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_abrupt_stop.ani"},
	{0xD6DFF8392E86C492ULL, "6128700_m.fxe"},
	{0xD6E056A2118B48EFULL, "follower_banter6165415.cl"},
	{0xD6E0988B7116E82AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0pp_basic.phy"},
	{0xD6E0C5C85C2E29D2ULL, "hm_ignacio_strand.mor"},
	{0xD6E0DB15119A7DFBULL, "vxlt_backalley_g_1x2_0_0.vlm"},
	{0xD6E1210CF352EC5DULL, "vxht_brothel_n_5er_0_0.vlm"},
	{0xD6E12AC12F957C48ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\leftlegram.ani"},
	{0xD6E1550B7FE88402ULL, "6140157_m.fxe"},
	{0xD6E173114E07214FULL, "62056305.ogg"},
	{0xD6E19617D0522175ULL, "6101261_m.fxe"},
	{0xD6E20173F9772633ULL, "gen00fl_varric6109210.cl"},
	{0xD6E22B7C1C812E09ULL, "weapon_rendertodepth_p0.psh.fxo"},
	{0xD6E27D2D6BE420FCULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgtp_0d.dds"},
	{0xD6E299BCE9B17F70ULL, "gen_im_arm_cht_lgt_act3_02.uti"},
	{0xD6E37400BCE3BDB0ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_aet_dm06.phy"},
	{0xD6E3A4FCE97652D3ULL, "6045035_m.fxe"},
	{0xD6E3D10B4F0FECBEULL, "6104293_f.fxe"},
	{0xD6E41E0C19CD0D39ULL, "mcr371_hostile_anders.stg"},
	{0xD6E47C3968977DB4ULL, "art\\characters\\weapons\\zzdao\\textures\\w_lsw_ls01a_0s.dds"},
	{0xD6E4AA9C99102C29ULL, "678407010.ogg"},
	{0xD6E4D5D9E99DC37EULL, "6053235_m.fxe"},
	{0xD6E536B37AB45BA4ULL, "art\\vfx\\cinematic\\vfx_scaredmage.mmh"},
	{0xD6E5D7958C1F0174ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_zbm_0s.dds"},
	{0xD6E65768894D40C3ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderweba.msh"},
	{0xD6E66568894D581DULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderweba.mmh"},
	{0xD6E66968894D5F2EULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_spiderweba.mao"},
	{0xD6E67D2A2110788EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_abrupt_stop.gad"},
	{0xD6E78F8F50A577F4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_coffer.phy"},
	{0xD6E7B43C665C4A22ULL, "hlac_lt_0_lowtown_docks_a3_-131073_5z.dds"},
	{0xD6E7F154256DB33BULL, "llac_ca_caverns1_mines_24424_3.dds"},
	{0xD6E7F575441F31B0ULL, "644713537.ogg"},
	{0xD6E81A31C256079BULL, "hlac_ca_caverns1_mines_24425_5y.dds"},
	{0xD6E83A4E1B07015FULL, "ca_caverns1_mines_24426_lc.mmh"},
	{0xD6E869599F665BD2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_tet01b.phy"},
	{0xD6E8A5CD9CE6D3CEULL, "ts_dalish_01.gda"},
	{0xD6E8BFE6DF7365DEULL, "vxlt_darktown_4is_0_0.vlm"},
	{0xD6E8F05934C858E6ULL, "vxw_bonepitmining_a2_ld_0_0.vlm"},
	{0xD6E8F703AEDA1668ULL, "w_bonepitmining_a2_240_0.mmh"},
	{0xD6E90303AEDA2A0BULL, "w_bonepitmining_a2_240_0.mao"},
	{0xD6E91503AEDA4892ULL, "w_bonepitmining_a2_240_0.msh"},
	{0xD6E91963D5682E9AULL, "one100dg_vanard6048261.cl"},
	{0xD6EA377B2D368C05ULL, "follower_banter6163248.cl"},
	{0xD6EA432B31099E41ULL, "130624522.ogg"},
	{0xD6EA50C6D33E26BFULL, "vxw_woundedcoast_1f9_0_0.vlm"},
	{0xD6EA9DAADE9EF514ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyo.phy"},
	{0xD6EAFA9FB32DB6F3ULL, "vxw_encounter2_ov_0_0.vlm"},
	{0xD6EB1BFDC374B8CCULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushe.msh"},
	{0xD6EB21FDC374C2BEULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushe.mmh"},
	{0xD6EB2DFDC374D755ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_bushe.mao"},
	{0xD6EB8D3C7F32FD0DULL, "qun221_viscount6070182.cl"},
	{0xD6EBC45CCF4878B9ULL, "vxht_keep_6mm_0_0.vlm"},
	{0xD6EBCEF9928380A9ULL, "vxlt_0_lowtown_n_er8_0_0.vlm"},
	{0xD6EC2141507B1242ULL, "vxlt_house_jd_0_0.vlm"},
	{0xD6EC48CDB2E0A2F6ULL, "vxw_encounter1_d3_0_0.vlm"},
	{0xD6EC6200DCAE23FFULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn7plank.msh"},
	{0xD6EC6800DCAE2DA1ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn7plank.mmh"},
	{0xD6EC7400DCAE4242ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn7plank.mao"},
	{0xD6EC782A2114DC6DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_abrupt_stop.evt"},
	{0xD6ECC6E78939FB51ULL, "6115611_m.fxe"},
	{0xD6ED0ACD016EB85AULL, "vxht_chantry_n_28f_0_0.vlm"},
	{0xD6ED6AC7A00E4175ULL, "hlac_w_sundermount_327678_0y.dds"},
	{0xD6ED770EA74276F1ULL, "6115498_f.fxe"},
	{0xD6ED87C1692218E5ULL, "art\\characters\\playercharacter\\elfmale\\glv_hvyo.phy"},
	{0xD6ED8FABFC04D31BULL, "sst_ss_harlan.gda"},
	{0xD6EDC11B05F19077ULL, "inputmap.gda"},
	{0xD6EE2A120E8929D8ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb107a_0s.dds"},
	{0xD6EE2A8073B8E9DBULL, "362508832.ogg"},
	{0xD6EE31CF67083FCFULL, "llac_lt_darktown_-1_85.dds"},
	{0xD6EE5CBC4AF5C28DULL, "hlac_ca_caverns1_mines_24427_5y.dds"},
	{0xD6EF7D5022955D3BULL, "character_rendertodepth_p7.vsh.fxo.dx11"},
	{0xD6EF92E9E54904DDULL, "vxht_brothel_n_5hq_0_0.vlm"},
	{0xD6EF9CBCCC4143B5ULL, "vxdr_0_deeproads_cqo_0_0.vlm"},
	{0xD6F0379552A3EEDDULL, "6126874_f.fxe"},
	{0xD6F047562AEC16A1ULL, "vxlt_0_lowtown_e2s_0_0.vlm"},
	{0xD6F047C20457C6AAULL, "993207066.ogg"},
	{0xD6F05C4E5097465EULL, "hlac_ht_0_hightown_n_131073_0y.dds"},
	{0xD6F06298F5E150A3ULL, "6191785_m.fxe"},
	{0xD6F0DA292F95F2CFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a.mao"},
	{0xD6F0E6292F96076CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a.mmh"},
	{0xD6F0EC292F96115EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile1a.msh"},
	{0xD6F1402CB217D49BULL, "llac_lt_0_lowtown_a2_-65535_7.dds"},
	{0xD6F16C8F50AD90F1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_coffer.msh"},
	{0xD6F177C66BAAC0F6ULL, "lt_0_lowtown_a3_n_c_10_0.msh"},
	{0xD6F17A8F50ADA80CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_coffer.mao"},
	{0xD6F181C66BAAD1F4ULL, "lt_0_lowtown_a3_n_c_10_0.mmh"},
	{0xD6F185C66BAAD807ULL, "lt_0_lowtown_a3_n_c_10_0.mao"},
	{0xD6F1868F50ADBCAFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1cw_coffer.mmh"},
	{0xD6F24F717C593ACAULL, "vxca_cave_17z_0_1.vlm"},
	{0xD6F2542A2A96C85BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_containera_0.pwk"},
	{0xD6F2AA090BA591AFULL, "vxlt_0_lowtown_docks_brk_0_0.vlm"},
	{0xD6F2BFDD9AA4FA46ULL, "zz_dog_debug6204276.cl"},
	{0xD6F2D6FD690D767CULL, "mag311_keran6041573.cl"},
	{0xD6F372FC033F3583ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforward.evt"},
	{0xD6F376F4E37BDF24ULL, "6082679_m.fxe"},
	{0xD6F37F458C427EA4ULL, "llac_lt_0_lowtown_a2_n_c_65536_7.dds"},
	{0xD6F39B2807100522ULL, "692359345.ogg"},
	{0xD6F3B65065D085D7ULL, "vxlt_0_lowtown_docks_b6o_0_0.vlm"},
	{0xD6F3B70B07742BC7ULL, "vxlt_foundry_n_wj_0_0.vlm"},
	{0xD6F477137E06890EULL, "190310996.ogg"},
	{0xD6F47AAADEA70E11ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyo.msh"},
	{0xD6F488AADEA7262CULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyo.mao"},
	{0xD6F494AADEA73ACFULL, "art\\characters\\playercharacter\\dwarfmale\\glv_hvyo.mmh"},
	{0xD6F49BD6031D196BULL, "art\\characters\\playercharacter\\qunarimale\\uni_stn_low.phy"},
	{0xD6F4FABBD16096F0ULL, "hlac_lt_undercityserialkiller_-1_20z.dds"},
	{0xD6F568B940F2BA00ULL, "6110885_f.fxe"},
	{0xD6F5F96A367F1DDAULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn05.msh"},
	{0xD6F60B6A367F3C20ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hn05.mmh"},
	{0xD6F6292FC8BB7FABULL, "sib131_letter_react6127319.cl"},
	{0xD6F666B6DB4BF69FULL, "667463191.ogg"},
	{0xD6F68EA49676EBA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_gesture_left_hand.ani"},
	{0xD6F69DE89C1EB1B6ULL, "6027532_f.fxe"},
	{0xD6F7F44504921889ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_45_walk.ani"},
	{0xD6F8C2A3C5343F3FULL, "vxw_woundedcoast_1a8_0_0.vlm"},
	{0xD6F8E25D49A600F3ULL, "vxht_0_hightown_a2_ap3_0_0.vlm"},
	{0xD6F96C7CA523CF73ULL, "vxw_encounter2_jy_0_0.vlm"},
	{0xD6F96EC3688AA9A0ULL, "864313810.ogg"},
	{0xD6FA1190B92F6E70ULL, "vxht_0_hightown_a2_cdj_0_0.vlm"},
	{0xD6FA4BFC03451204ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforward.gad"},
	{0xD6FA7F7BFA9DC834ULL, "247328437.ogg"},
	{0xD6FAACCA70AAD5DDULL, "vxlt_0_lowtown_docks_b4g_0_0.vlm"},
	{0xD6FAB43CB9F0FF56ULL, "345218964.ogg"},
	{0xD6FABAA942F97411ULL, "one000_ilen6205645.cl"},
	{0xD6FABAAAA4D6BB76ULL, "vxw_encounter1_i4_0_0.vlm"},
	{0xD6FACBC771E6860DULL, "art\\characters\\weapons\\unique\\meredith\\w_gsw_014a.phy"},
	{0xD6FAFBF120C6C664ULL, "161982919.ogg"},
	{0xD6FB0773A26D94A2ULL, "vxlt_blackemporium_s1_0_1.vlm"},
	{0xD6FC26A00B91B274ULL, "668719779.ogg"},
	{0xD6FC75408B3F7865ULL, "hlac_dr_primevalentrance_1726_5x.dds"},
	{0xD6FCB7A95F08AF55ULL, "6110211_f.fxe"},
	{0xD6FCFC81CED7D148ULL, "vxlt_0_lowtown_a2_n_d3s_0_0.vlm"},
	{0xD6FD2A506ED3F2C4ULL, "6084537_f.fxe"},
	{0xD6FDAF7383A33DE8ULL, "543892434.ogg"},
	{0xD6FE99F4B9292E4CULL, "vxht_mansion_n_62v_0_0.vlm"},
	{0xD6FEEC145F4BEC7FULL, "gen00fl_bethany6102897.cl"},
	{0xD6FEFB5E07D462A2ULL, "zz_car_debug6122484.cl"},
	{0xD6FF05B3FF402307ULL, "vxht_0_hightown_b9o_0_0.vlm"},
	{0xD6FF0F450498650BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_45_walk.gad"},
	{0xD6FF408FEA5495DFULL, "hm_brett_harimann.mor"},
	{0xD6FFCF16B8722C75ULL, "vxdr_primevalentrance_2q8_0_0.vlm"},
	{0xD6FFE535F67F6B81ULL, "551557034.ogg"},
	{0xD7001C174267A3CFULL, "jumpattack.cut"},
	{0xD700774BDEFA7332ULL, "one100dg_kelder6055613.cl"},
	{0xD700B39C3CD04B7DULL, "84737620.ogg"},
	{0xD700F6D0E3445238ULL, "lt_0_lowtown_n_mmap_ph.dds"},
	{0xD7015AFC034B4AE2ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\walkforward.ani"},
	{0xD7016935040AE1EAULL, "vxlt_hangedman_5z_0_0.vlm"},
	{0xD701E0FBDDAD64D0ULL, "nat380_temmerin6120933.cl"},
	{0xD702187404C113F6ULL, "625173428.ogg"},
	{0xD702275E858B5125ULL, "6035280_f.fxe"},
	{0xD7024298BE476851ULL, "and100_anders6095834.cl"},
	{0xD70249CF8A58F552ULL, "mag160ip_at_to_ws_cave.utp"},
	{0xD70258895D0DCEB6ULL, "6075135_f.fxe"},
	{0xD702741A9B3C3D05ULL, "28921681.ogg"},
	{0xD70283A772406AE1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.ani"},
	{0xD70323C982E453EDULL, "dae000_arishok6074059.cl"},
	{0xD703B106B13878ADULL, "662717055.ogg"},
	{0xD703B8922579C417ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_action.gad"},
	{0xD703D4471540612FULL, "vxlt_gallowscourtyard_7eo_0_0.vlm"},
	{0xD70458878964755FULL, "prty_carver_armor_3_warden.dds"},
	{0xD704A6A49683519AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_gesture_left_hand.evt"},
	{0xD7050190B55E8004ULL, "hlac_lt_gallowscourtyard_-65535_2y.dds"},
	{0xD7053C95226A325DULL, "ts_beasts_01.gda"},
	{0xD70548545363D691ULL, "6041815_m.fxe"},
	{0xD7060C45049E7F70ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_right_45_walk.evt"},
	{0xD7066D2D5868DF4FULL, "726322315.ogg"},
	{0xD706D3BCA8BCFE71ULL, "1054029680.ogg"},
	{0xD706F9651A456444ULL, "ca_caverns1_mines_24424.rml"},
	{0xD7070475200D3ABCULL, "6080582_f.fxe"},
	{0xD70705F1AEBF797EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0pp_flared1a_1.mao"},
	{0xD70850BB8095931BULL, "vxlt_backalley_n_1pb_0_1.vlm"},
	{0xD708E4C080EE5FA2ULL, "w_blightlands_12_0.msh"},
	{0xD708E6C080EE6358ULL, "w_blightlands_12_0.mmh"},
	{0xD708F2C080EE77FBULL, "w_blightlands_12_0.mao"},
	{0xD7090EF59C12D8DCULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_attk_v1.ani"},
	{0xD7098EA772469B83ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.gad"},
	{0xD70A4EC467CD3236ULL, "741548362.ogg"},
	{0xD70ACD92258006A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_action.evt"},
	{0xD70B38CFE3DEAA71ULL, "hlac_lt_0_lowtown_a2_n_131070_5x.dds"},
	{0xD70B4EC7E6AA9710ULL, "gen_im_wep_war_one_act0_01.uti"},
	{0xD70BA5A496896E8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_gesture_left_hand.gad"},
	{0xD70BB2848AB680D1ULL, "gen_im_wep_war_two_act2_04.uti"},
	{0xD70BB63ED7E5AC31ULL, "716099838.ogg"},
	{0xD70C667D5EB0670FULL, "vxht_keep_n_36x_0_0.vlm"},
	{0xD70C74874F092F4AULL, "zz_followers_banter6158142.cl"},
	{0xD70CAC18552D93E2ULL, "llac_w_encounter1_-131069_0.dds"},
	{0xD70CB775CEBB723EULL, "w_bonepitdestroyed_a3_26_0_lc.mmh"},
	{0xD70CC4D1243D292BULL, "one200_gold_followers6117247.cl"},
	{0xD70CE6C080F24C6BULL, "w_blightlands_12_0.rml"},
	{0xD70D804509C9358AULL, "204141941.ogg"},
	{0xD70D9F6F2D187904ULL, "857591593.ogg"},
	{0xD70DCB6C3F083CD6ULL, "vxca_caverns1_mines_jf6_0_0.vlm"},
	{0xD70DEC23478F89CEULL, "vxlt_0_lowtown_a3_n_c_e1a_0_0.vlm"},
	{0xD70E39B37AD7659BULL, "art\\vfx\\cinematic\\vfx_scaredmage.anb"},
	{0xD70E9A5F61933FD4ULL, "6136648_m.fxe"},
	{0xD70EB6A4606BF7C4ULL, "6099356_m.fxe"},
	{0xD70EC63434631479ULL, "6104618_f.fxe"},
	{0xD70ED60ADA26E1FBULL, "hlac_dr_primevalentrance_1726_10z.dds"},
	{0xD70F1BF59C175A1EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_attk_v1.gad"},
	{0xD70F90610617466DULL, "dre000_cut2d_ragstoriches6025106.cl"},
	{0xD70FFDB10532C6ECULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03b.phy"},
	{0xD71054BFCF70C6C3ULL, "vis120_commoner16069034.cl"},
	{0xD7106BA7724C7F68ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.evt"},
	{0xD710AE37CF9BB336ULL, "follower_banter6170050.cl"},
	{0xD710BCC03448F2CCULL, "6039970_m.fxe"},
	{0xD710D24EADAE9AF3ULL, "lowtown_guard_chase6169516.cl"},
	{0xD710D3D08B21B6FDULL, "vxw_sundermount_35p_0_0.vlm"},
	{0xD710FD44CB923C6FULL, "hlac_ht_0_hightown_n_-131069_0x.dds"},
	{0xD7113D7ADD73669AULL, "6201591_m.fxe"},
	{0xD711573767A5CAB5ULL, "6140327_m.fxe"},
	{0xD71182B35F0676ECULL, "dae011_corff_sta6149648.cl"},
	{0xD711BE28C3B42B74ULL, "gen00fl_fenris6108098.cl"},
	{0xD711E320402E2F4DULL, "vxlt_0_lowtown_docks_a3_a0k_0_0.vlm"},
	{0xD7122DB8BCA6D8D7ULL, "mag131_taraneh6094545.cl"},
	{0xD71239C9424977BDULL, "879415137.ogg"},
	{0xD7125F0CDA064A2AULL, "vxlt_0_lowtown_a2_n_c_ht2_0_0.vlm"},
	{0xD712637DD495E90CULL, "6101516_m.fxe"},
	{0xD7135247298E1DB5ULL, "g_p_smok.mao"},
	{0xD713B938736B4DF9ULL, "qun111_ketojan6040663.cl"},
	{0xD714071B17EF45F5ULL, "357689623.ogg"},
	{0xD7146A038FE7BCD0ULL, "vxca_caverns1_j24_0_1.vlm"},
	{0xD7147618F7080038ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneya.phy"},
	{0xD7147B95E2D1B4D1ULL, "1010236215.ogg"},
	{0xD7147C077517821EULL, "art\\vfx\\combat\\vfx_w_jug_tremor_lar_aoe_p.mmh"},
	{0xD71495A8EBC6E331ULL, "6076574_m.fxe"},
	{0xD7149CCE263A9F9CULL, "987775058.ogg"},
	{0xD7155C3AA0A73196ULL, "character_rendertodepth_p0.vsh.fxo"},
	{0xD715D0D9D5B79B37ULL, "6102147_m.fxe"},
	{0xD715F6F59C1D3A3DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_attk_v1.evt"},
	{0xD715FD7351E1702EULL, "vxlt_0_lowtown_a2_n_c_hr0_0_0.vlm"},
	{0xD71607F773C8AA83ULL, "vxht_mansion2_2eg_0_0.vlm"},
	{0xD71608EE748BD637ULL, "6131189_m.fxe"},
	{0xD71667E705A205ADULL, "dae000_hate_enemies6167508.cl"},
	{0xD7166C8A0D429FA1ULL, "667898573.ogg"},
	{0xD71751607F84BFB5ULL, "vxht_chantry_2co_0_0.vlm"},
	{0xD717ADCE6B845C93ULL, "vxw_encounter2_camp_76_0_0.vlm"},
	{0xD717D185D6C2D2D4ULL, "vxlt_0_lowtown_a2_ep5_0_1.vlm"},
	{0xD717E592258AB99DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\feynriel_execution_sword_action.ani"},
	{0xD717F0EB922407FCULL, "wm_ico_playershome.dds"},
	{0xD7180BB1883FF94EULL, "6021205_m.fxe"},
	{0xD719705180F4A30EULL, "6042978_m.fxe"},
	{0xD7199DF7F6A962C9ULL, "lowtown_male_alienage2_mob6166699.cl"},
	{0xD719A221DC545BD7ULL, "static_static_p75.psh.fxo.dx11"},
	{0xD719F96D743F3C7CULL, "147337067.ogg"},
	{0xD71B040C19FC23D6ULL, "mcr371_hostile_anders.cnv"},
	{0xD71C37ED538F2C4BULL, "store_dae210_access.utm"},
	{0xD71D5184E03CCFF7ULL, "512_ext_test.arl"},
	{0xD71DABE1EB2583C9ULL, "hlac_w_sundermount_262147_0z.dds"},
	{0xD71DECEA8518C441ULL, "dae021_brothel_onenter6196724.cl"},
	{0xD71DED3A08BB812FULL, "mer210_coterie_contact6036758.cl"},
	{0xD71DF1C422094F9CULL, "lt_0_lowtown_37_0_ll.mmh"},
	{0xD71DF4729F93266AULL, "combat_ar_r1_a3.gda"},
	{0xD71DFF9FE187DB2FULL, "vxlt_0_lowtown_docks_bct_0_0.vlm"},
	{0xD71E060C19FE5CB6ULL, "mcr371_hostile_anders.bnk"},
	{0xD71E1421DEA931FBULL, "w_encounter3_17_0.msh"},
	{0xD71E2621DEA95046ULL, "w_encounter3_17_0.mao"},
	{0xD71E3221DEA964E5ULL, "w_encounter3_17_0.mmh"},
	{0xD71E66DC4096F89CULL, "6081599_m.fxe"},
	{0xD71ED1FEF87D74DBULL, "ss_coterie_thug_m.bnk"},
	{0xD71F410035C7AF60ULL, "418879234.ogg"},
	{0xD71F755917A25349ULL, "dae321cr_petrice.utc"},
	{0xD720455913B1A2BDULL, "vxlt_shop_f7_0_0.vlm"},
	{0xD720CC531739C215ULL, "6164008_m.fxe"},
	{0xD720E5C16057FE40ULL, "174090197.ogg"},
	{0xD720EA1E9E5A47B7ULL, "6179075_m.fxe"},
	{0xD721954AC27572E1ULL, "217803600.ogg"},
	{0xD721972E0CFEA255ULL, "vxlt_0_lowtown_a3_n_c_gz1_0_0.vlm"},
	{0xD7225B07752387DDULL, "art\\vfx\\combat\\vfx_w_jug_tremor_lar_aoe_p.anb"},
	{0xD722A0F1A1B1E7C3ULL, "461370701.ogg"},
	{0xD723139DCE6C1C5EULL, "gen_im_trs_qun_hvy_boo_r1.uti"},
	{0xD7238F56B33430EAULL, "6112972_m.fxe"},
	{0xD723EA3A989C2CC3ULL, "dae000ip_crowd_ff3_talk3.utp"},
	{0xD7240F48083E1D17ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgta.phy"},
	{0xD724C5142F13BA14ULL, "w_bonepitmining_a2_203_0.phy"},
	{0xD724F3F7CFE734E9ULL, "hlac_lt_0_lowtown_n_65537_5z.dds"},
	{0xD72532273821FCEFULL, "vxlt_0_lowtown_a3_n_c_hp6_0_0.vlm"},
	{0xD72535503F82FC1BULL, "367692682.ogg"},
	{0xD72591F44D580FE3ULL, "344161604.ogg"},
	{0xD7261FAB5D7A7513ULL, "vxw_encounter2_camp_29_0_0.vlm"},
	{0xD726202C3EDDD2D4ULL, "vxht_keep_n_617_0_0.vlm"},
	{0xD7267F05CEC07AC2ULL, "llac_lt_hangedman_n_-1_6.dds"},
	{0xD7271F583CC7B6CEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_door.phy"},
	{0xD72738BF61AD8826ULL, "vxht_mansion2_1me_0_0.vlm"},
	{0xD727C4ABD1985834ULL, "6170336_m.fxe"},
	{0xD727E2D1F8148640ULL, "6083184_m.fxe"},
	{0xD727F119DAD5B7FEULL, "lgtcr_dalish_warrior.utc"},
	{0xD7282BEE07D15D8AULL, "llac_ht_0_hightown_-65536_1.dds"},
	{0xD7282FBB0897D431ULL, "hlac_lt_0_lowtown_docks_a3_n_131071_4z.dds"},
	{0xD7284534AA2525FDULL, "vxca_caverns1_gil_0_0.vlm"},
	{0xD7286BEE182C33B4ULL, "vxlt_0_lowtown_ex8_0_0.vlm"},
	{0xD7288523ED0695FDULL, "309319938.ogg"},
	{0xD7296425B8358786ULL, "vxw_encounter3_g5_0_1.vlm"},
	{0xD729B988058B3709ULL, "hlac_lt_house_-1_5z.dds"},
	{0xD729F6E3D655028AULL, "6080558_m.fxe"},
	{0xD72A4B138EBAF6AAULL, "6040819_m.fxe"},
	{0xD72A57339238CF82ULL, "6083651_f.fxe"},
	{0xD72ABFCB2D32E28BULL, "56831387.ogg"},
	{0xD72AD2D8AEE006CBULL, "6134898_m.fxe"},
	{0xD72B2534459F80F4ULL, "lgt100cr_gang_2_lt.utc"},
	{0xD72B4DF5EBB00007ULL, "one280_iwan6096754.cl"},
	{0xD72B63B8522B4250ULL, "mrl150_cut_amulet_ritual.stg"},
	{0xD72B8596F9034147ULL, "art\\characters\\heraldry\\hld_dalish_2.dds"},
	{0xD72B8909A53D94A9ULL, "6110270_m.fxe"},
	{0xD72BA056734E08C8ULL, "vxht_0_hightown_a2_c3r_0_1.vlm"},
	{0xD72C0B2BD14E6CBAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf.gad"},
	{0xD72C6A679406B4C7ULL, "347372429.ogg"},
	{0xD72C717CD37DF3AFULL, "vxlt_0_lowtown_docks_bfw_0_0.vlm"},
	{0xD72DC695AC4926A5ULL, "vxca_caverns1_mines_e8h_0_0.vlm"},
	{0xD72E64A8BBC40409ULL, "gen00fl_merrill6129900.cl"},
	{0xD72E880AD81ACB27ULL, "gen00fl_dog6167693.cl"},
	{0xD72EA2142F1BD311ULL, "w_bonepitmining_a2_203_0.msh"},
	{0xD72EB0142F1BEB2CULL, "w_bonepitmining_a2_203_0.mao"},
	{0xD72EBB87E9CBB392ULL, "cod_bks_kirkwall_hist_1.utp"},
	{0xD72EBC142F1BFFCFULL, "w_bonepitmining_a2_203_0.mmh"},
	{0xD72EBD1685944CE1ULL, "lgt110_gxa_coa_meeting6117503.cl"},
	{0xD72F25C3E3C89DB0ULL, "413555544.ogg"},
	{0xD73015C3DB2F6F70ULL, "mag121_idunna6084179.cl"},
	{0xD73022F6F43DC542ULL, "591964923.ogg"},
	{0xD7304214AD50953FULL, "6005808_m.fxe"},
	{0xD730E6D848BA0C97ULL, "754362675.ogg"},
	{0xD730F1FFFA764454ULL, "hlac_lt_0_lowtown_a2_65536_4x.dds"},
	{0xD731AAAB7C0A7892ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_enter.gad"},
	{0xD731B3C225C9C6EBULL, "qcr000_aveline6054373.cl"},
	{0xD731CA3BC8A3DC0BULL, "6149281_m.fxe"},
	{0xD731D9CF9DEB66B9ULL, "vxht_mansion_n_326_0_0.vlm"},
	{0xD7321758FACACA3DULL, "6081728_f.fxe"},
	{0xD732B131A1943CCEULL, "adie_cl_f.mao"},
	{0xD7330E2BD15490D1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf.evt"},
	{0xD7334F9A6249CC39ULL, "650137508.ogg"},
	{0xD7338F9A7D0FFBFEULL, "mer120_hubert6059141.cl"},
	{0xD73411F26A3E7040ULL, "6041403_m.fxe"},
	{0xD7346B19F1BB0A71ULL, "llac_w_encounter2_camp_327683_0.dds"},
	{0xD734C5B0BC01A254ULL, "llac_lt_0_lowtown_a2_n_c_-131072_9.dds"},
	{0xD73609ED476AF6B9ULL, "6167863_m.fxe"},
	{0xD736DDCB0A224C34ULL, "vxlt_0_lowtown_es9_0_0.vlm"},
	{0xD736FAA15887C129ULL, "hlac_w_sundermount_-3_0z.dds"},
	{0xD7375CF6CCA44CB5ULL, "character_character_p194.psh.fxo.dx11"},
	{0xD73760F8AC45AEDCULL, "vxlt_hangedman_1uu_0_0.vlm"},
	{0xD737B8878223FF9BULL, "vxdr_primevalentrance_16l_0_0.vlm"},
	{0xD7385307D84246B0ULL, "hlac_lt_0_lowtown_65534_0x.dds"},
	{0xD738ADAB7C109CC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_enter.evt"},
	{0xD738C9E1CC7DAC01ULL, "295437915.ogg"},
	{0xD738E0B9F0566F22ULL, "6034535_m.fxe"},
	{0xD73944F3F6DFBC37ULL, "374036516.ogg"},
	{0xD73991BCEB2E14B5ULL, "zz_fol_debug6097812.cl"},
	{0xD739BA68B53A623DULL, "hlac_w_bonepitdestroyed_a3_393219_0z.dds"},
	{0xD739C73017E4F402ULL, "vxlt_0_lowtown_a3_n_c_eqm_0_0.vlm"},
	{0xD73A3BEAF3939B91ULL, "vxlt_gallowstemplar_3rn_0_0.vlm"},
	{0xD73A644F1997B129ULL, "dre000_carver6037574.cl"},
	{0xD73A8B9F19BBB360ULL, "6115569_f.fxe"},
	{0xD73B58532D26C4B8ULL, "gen00fl_fenris6107194.cl"},
	{0xD73BB9B0B071C2A2ULL, "540851795.ogg"},
	{0xD73BCB7E4321CEBEULL, "6127042_m.fxe"},
	{0xD73C53599E776C47ULL, "709162099.ogg"},
	{0xD73CCC9A4708B3A6ULL, "rog_teleport.dds"},
	{0xD73CD11D0A261B46ULL, "6155388_m.fxe"},
	{0xD73D3BA35FC9C0E7ULL, "aip_templarmeredith_plot.gda"},
	{0xD73D5B9E9C867693ULL, "vxca_caverns1_j48_0_0.vlm"},
	{0xD73F0D9F66E07C73ULL, "85675561.ogg"},
	{0xD73F1FA2FAD45881ULL, "6084438_m.fxe"},
	{0xD74002C883999F8DULL, "zz_followers_banter6157546.cl"},
	{0xD7402528E8F668CDULL, "939766827.ogg"},
	{0xD740262BD15F43A0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf.ani"},
	{0xD7411EBF58DB8033ULL, "dae000_unhappy_mage6165336.cl"},
	{0xD74150CABBB04AB1ULL, "239586064.ogg"},
	{0xD741AE7E5E0BFA4CULL, "vrc221_hugin6042565.cl"},
	{0xD741FD26290AED98ULL, "hlac_ht_mansion2_818_30y.dds"},
	{0xD7424F8A36CB5187ULL, "vxdr_primevaltomb_6xy_0_0.vlm"},
	{0xD74262C9DE20FC34ULL, "vxlt_0_lowtown_h1v_0_0.vlm"},
	{0xD742EF09F0B41791ULL, "6172511_m.fxe"},
	{0xD743027CBC9313CCULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_teag.phy"},
	{0xD7433F86719B9EDBULL, "695406539.ogg"},
	{0xD74351A7CF0F8319ULL, "follower_banter6160968.cl"},
	{0xD744282B701C4D7FULL, "vxca_caverns2prc_9fc_0_0.vlm"},
	{0xD744A06DA2496309ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_03.ani"},
	{0xD744C07D9180B6BCULL, "561347327.ogg"},
	{0xD744E05395E0500FULL, "6084732_f.fxe"},
	{0xD74584CA762C3B3BULL, "mag121_idunna6084008.cl"},
	{0xD745DF17FA787825ULL, "411255300.ogg"},
	{0xD745EADC319900D1ULL, "vxw_woundedcoast_s9_0_0.vlm"},
	{0xD7462B6AF2FE03E6ULL, "6106818_m.fxe"},
	{0xD7465C686743F24AULL, "63053669.ogg"},
	{0xD74681C66BF311F7ULL, "lt_0_lowtown_a3_n_c_10_0.rml"},
	{0xD74695AB7C1CB138ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_sleep_on_table_enter.ani"},
	{0xD74747A99D2D3699ULL, "llac_lt_undercityserialkiller_-1_9.dds"},
	{0xD747612C06ED1746ULL, "vxlt_hangedman_n_2w9_0_0.vlm"},
	{0xD747A3E0CC07160CULL, "6048431_m.fxe"},
	{0xD747C818A66738FBULL, "hlac_lt_gallowscourtyard_n_-131072_2y.dds"},
	{0xD7480BE232611BA4ULL, "face_face_p1.psh.fxo"},
	{0xD7485FD220883CDEULL, "oip_rock_01.mao"},
	{0xD7487DC43F83E83FULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha07a.mmh"},
	{0xD74883C43F83F2E1ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha07a.msh"},
	{0xD748B07B9A429048ULL, "fer_ter_grass_001.mao"},
	{0xD749682573C1EFC8ULL, "hlac_lt_0_lowtown_n_65535_0z.dds"},
	{0xD749702CD9E995A2ULL, "vxht_bartrand_n_64m_0_0.vlm"},
	{0xD749C4ABADAC9E75ULL, "hlac_lt_0_lowtown_a2_n_65535_3x.dds"},
	{0xD749FD996689B886ULL, "llac_lt_darktown_-1_79.dds"},
	{0xD749FDC742BBD4B0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\set_lowtowndoor_n.dds"},
	{0xD74A109FDC2DDA1BULL, "vxht_0_hightown_a2_b8x_0_0.vlm"},
	{0xD74A360D148718B3ULL, "266271623.ogg"},
	{0xD74A3C2AA92C0695ULL, "hlac_lt_0_lowtown_docks_n_-131074_7x.dds"},
	{0xD74A8FD557294C94ULL, "llac_w_woundedcoast_327685_0.dds"},
	{0xD74ACD08FCA4DC1AULL, "vxdr_0_deeproads_9g4_0_0.vlm"},
	{0xD74B01D1C975B6C8ULL, "6004220_m.fxe"},
	{0xD74B14972834AF37ULL, "rdr110dg_guard_warehouse6016731.cl"},
	{0xD74BBB6DA24FAF8BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_03.gad"},
	{0xD74BD548AA3CAEF1ULL, "vxht_keep_6zp_0_0.vlm"},
	{0xD74BDD967964E3F5ULL, "6148644_m.fxe"},
	{0xD74C80C66BF77BEFULL, "lt_0_lowtown_a3_n_c_10_0.phy"},
	{0xD74CBEFE2ACC0444ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_boardtop_1.mao"},
	{0xD74CE2F6125DDDC3ULL, "art\\vfx\\combat\\vfx_g_spider_fla_bst_c.anb"},
	{0xD74D34F2C14DFD90ULL, "hlac_lt_gallowstemplar_f_2388_2z.dds"},
	{0xD74D4A9BADF88D8FULL, "vxlt_warehouse_n_18b_0_0.vlm"},
	{0xD74D6AD72AE1694CULL, "mag120_macha6080641.cl"},
	{0xD74DB58E605B1D02ULL, "vxht_0_hightown_a3_c3q_0_1.vlm"},
	{0xD74E4B5908901A75ULL, "mcr370_assignments6081746.cl"},
	{0xD74EB4D63FB00596ULL, "ter_stoney2_s.dds"},
	{0xD74EBD44DD7B89DCULL, "hlac_ht_bartrand_n_3364_1x.dds"},
	{0xD74F34C5DC7355F3ULL, "character_character_p123.psh.fxo.dx11.assembly"},
	{0xD74F59639C63A84CULL, "890309922.ogg"},
	{0xD7501A2559E4F6A3ULL, "vxht_0_hightown_a3_7j2_0_0.vlm"},
	{0xD7506B067B2C96F3ULL, "vxw_blightlands_6r_0_0.vlm"},
	{0xD7508C82D9308C29ULL, "649592231.ogg"},
	{0xD750D4A6D01714B4ULL, "vxlt_0_lowtown_h6w_0_0.vlm"},
	{0xD750D4B1056A17B7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03b.mmh"},
	{0xD750EAB1056A3D29ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed03b.msh"},
	{0xD75107FE38D78317ULL, "135667663.ogg"},
	{0xD7513150C28E2934ULL, "287438275.ogg"},
	{0xD7519CB8051741DEULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hoo_lgtb_0s.dds"},
	{0xD751C87D3DCEE2ECULL, "vxlt_darktown_5kn_0_0.vlm"},
	{0xD7522F2984B5C2FCULL, "zz_fen_debug6134374.cl"},
	{0xD7525A18BC9953BAULL, "w_bonepitmining_a2_112_0.mmh"},
	{0xD7525E18BC995AF1ULL, "w_bonepitmining_a2_112_0.mao"},
	{0xD7526C18BC997200ULL, "w_bonepitmining_a2_112_0.msh"},
	{0xD752839D67E154B2ULL, "vxlt_0_lowtown_docks_n_cp6_0_0.vlm"},
	{0xD752B86DA255C9F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_03.evt"},
	{0xD752C44FDECE8F84ULL, "756093802.ogg"},
	{0xD752CCF5E7EBCE44ULL, "6082495_m.fxe"},
	{0xD752DBC46BD5615BULL, "vxdr_0_deeproads_fi8_0_0.vlm"},
	{0xD7532BD6BCF2C35BULL, "514079827.ogg"},
	{0xD753906C26244B0DULL, "335754077.ogg"},
	{0xD7539845C2B2FEB1ULL, "vxlt_undercityserialkiller_xw_0_0.vlm"},
	{0xD754BD6C1657AFC2ULL, "llac_lt_gallowstemplar_f_2394_3.dds"},
	{0xD754F9D8CC77EA8DULL, "6148378_m.fxe"},
	{0xD75510E614A68A37ULL, "1026299367.ogg"},
	{0xD755C6BEF3C631D2ULL, "vxht_mansion2_2b6_0_1.vlm"},
	{0xD755DCE1B779AE7FULL, "art\\characters\\weapons\\shields\\w_shd_kitb.msh"},
	{0xD755E2E1B779B821ULL, "art\\characters\\weapons\\shields\\w_shd_kitb.mmh"},
	{0xD755EEE1B779CCC2ULL, "art\\characters\\weapons\\shields\\w_shd_kitb.mao"},
	{0xD75662858A37EE8BULL, "120865449.ogg"},
	{0xD756A587E43DFCADULL, "6175665_f.fxe"},
	{0xD7571803EBB717A3ULL, "792885019.ogg"},
	{0xD757AF49E0F02A53ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.cess.evt"},
	{0xD757B66FEC80CDD6ULL, "dae000ip_crowd_mf8_talk6.utp"},
	{0xD757BE355CE46268ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gouda_d.dds"},
	{0xD757F0A5A71CE46DULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_lgta_0n.dds"},
	{0xD757FEB84D695513ULL, "fx_c_twister_blend.mao"},
	{0xD758007F6FA92E8CULL, "213293444.ogg"},
	{0xD75835371E52E1C9ULL, "vxht_mansion_n_6jg_0_0.vlm"},
	{0xD75872C82A7B48E9ULL, "841713085.ogg"},
	{0xD758E677B89BD707ULL, "vxw_bonepit_z5_0_0.vlm"},
	{0xD75958D135C75735ULL, "hlac_lt_0_lowtown_docks_-3_0y.dds"},
	{0xD759B2FA57168267ULL, "6127893_f.fxe"},
	{0xD759B583B446FB9FULL, "gen_im_trs_krk_arc_sbw_r1.uti"},
	{0xD75A0F722A1AEAD0ULL, "365487290.ogg"},
	{0xD75A47259C32C771ULL, "vxht_keep_6gw_0_0.vlm"},
	{0xD75A6249078BF5FBULL, "zz_dae_debug6029878.cl"},
	{0xD75A6FDC299F4775ULL, "w_bonepitdestroyed_a3_216_0.msh"},
	{0xD75A71DC299F4AEBULL, "w_bonepitdestroyed_a3_216_0.mmh"},
	{0xD75A7DDC299F5E88ULL, "w_bonepitdestroyed_a3_216_0.mao"},
	{0xD75AA76F370F4033ULL, "gen00fl_sebastian6179908.cl"},
	{0xD75ABFF61269DFECULL, "art\\vfx\\combat\\vfx_g_spider_fla_bst_c.mmh"},
	{0xD75B4335B0C76185ULL, "vxlt_0_lowtown_a3_n_c_h1w_0_0.vlm"},
	{0xD75BA08EB5AD748EULL, "prop1_punch.psh.fxo"},
	{0xD75BC5757C827DEFULL, "static_static_p2.psh.fxo.dx11"},
	{0xD75BEC226D09B801ULL, "vxlt_0_lowtown_a3_n_c_d66_0_0.vlm"},
	{0xD75C0345479FA1B8ULL, "vxht_bartrand_n_6ps_0_0.vlm"},
	{0xD75C192A2CA6D0EEULL, "5111161.ogg"},
	{0xD75C1BB6AC8E2FA8ULL, "zz_followers_banter6159288.cl"},
	{0xD75C92782FCE7E08ULL, "6066713_f.fxe"},
	{0xD75CB952EABF18BFULL, "885879954.ogg"},
	{0xD75CD5113690110BULL, "vxlt_0_lowtown_fhx_0_0.vlm"},
	{0xD75CF57CA78D4978ULL, "mrl000_marethari6142158.cl"},
	{0xD75D00577918FBE1ULL, "387900121.ogg"},
	{0xD75D30B0AE41F7E0ULL, "gen_im_trs_bst_drg_eye_03.uti"},
	{0xD75D462D93D5EE45ULL, "6082043_m.fxe"},
	{0xD75D4BBD541FCC82ULL, "523269863.ogg"},
	{0xD75DFBF7E47ACBD9ULL, "401480093.ogg"},
	{0xD75E4C6B86C5224CULL, "81107092.ogg"},
	{0xD75E515F7F7BA139ULL, "vxca_caverns1_i9r_0_0.vlm"},
	{0xD75E6BB85256A903ULL, "mrl150_cut_amulet_ritual.bnk"},
	{0xD75E8C024BDB0F23ULL, "vxht_0_hightown_a3_7g5_0_0.vlm"},
	{0xD75EA849E0F63C94ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.cess.gad"},
	{0xD75EB73523E537E5ULL, "6056859_m.fxe"},
	{0xD75EF99AFFFA5EEFULL, "vxlt_0_lowtown_a3_n_c_hl2_0_0.vlm"},
	{0xD75F3221DEE0FBEEULL, "w_encounter3_17_0.rml"},
	{0xD75FCDE66CA91309ULL, "llac_w_sundermount_-65534_0.dds"},
	{0xD75FE0D7DCF0E5E7ULL, "432998554.ogg"},
	{0xD7603A5A2FC4FB6CULL, "vxlt_darktown_5lo_0_0.vlm"},
	{0xD7604891106ABB9CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_deadfish_sa.dds"},
	{0xD76095A754020FAEULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench3.ani"},
	{0xD7609D24091D3B75ULL, "vxdr_primevaltomb_3ow_0_0.vlm"},
	{0xD760DEB3423F1A44ULL, "vxca_caverns1_i1i_0_2.vlm"},
	{0xD761046FD8D8B8BAULL, "llac_dr_dungeon_3217_12.dds"},
	{0xD7611AF937FA4AABULL, "qun160_vashothbandits6050370.cl"},
	{0xD761362BE2CFFF74ULL, "27350292.ogg"},
	{0xD7614DB85258AC33ULL, "mrl150_cut_amulet_ritual.cnv"},
	{0xD761596D77703FB1ULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_spiderweb2_da.dds"},
	{0xD761AC32EB994F94ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge14.ani"},
	{0xD761C755D7410F1CULL, "vxht_keep_643_0_0.vlm"},
	{0xD76220FE353B8B67ULL, "one220_yevhen6103289.cl"},
	{0xD76245611DA6364CULL, "589245023.ogg"},
	{0xD76248028C933347ULL, "6096527_m.fxe"},
	{0xD762768E6B30148EULL, "vxlt_0_lowtown_n_eoe_0_1.vlm"},
	{0xD762FBB7099F2184ULL, "6125459_m.fxe"},
	{0xD763CFAEB436FE0BULL, "1043450938.ogg"},
	{0xD7642F8AF657C534ULL, "vxht_chantry_n_2tl_0_0.vlm"},
	{0xD76578C6DEE35F2AULL, "dae000ip_crowd_mf7_drunk1.utp"},
	{0xD765B749E0FC7572ULL, "art\\vfx\\cinematic\\vfx_blood_smear_p.cess.ani"},
	{0xD765BC9CDD43D4CEULL, "638400593.ogg"},
	{0xD765EBE5078FBD27ULL, "6096431_m.fxe"},
	{0xD765F2B09F820CE1ULL, "602677197.ogg"},
	{0xD765F3D94256EA22ULL, "6015761_m.fxe"},
	{0xD766A7141048FA49ULL, "vxht_mansion_n_6mf_0_0.vlm"},
	{0xD766B8C538C55A01ULL, "vxlt_0_lowtown_a2_f4r_0_0.vlm"},
	{0xD766BFFEF8BB0BD9ULL, "ss_coterie_thug_m.wbd"},
	{0xD766FBAC83EBEBE8ULL, "127011086.ogg"},
	{0xD7671E583CFD9891ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_door.mmh"},
	{0xD7672A583CFDAD32ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_door.mao"},
	{0xD76738583CFDC54FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9dd_door.msh"},
	{0xD7675854AA91EF87ULL, "vxw_bonepit_e4_0_0.vlm"},
	{0xD767D59246DFA2A9ULL, "hlac_lt_0_lowtown_a2_-1_4x.dds"},
	{0xD767E139D16FA49CULL, "184725553.ogg"},
	{0xD768A78917DCC973ULL, "mcr371_emile_de_launcet.bnk"},
	{0xD768A932EB9F69F6ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge14.gad"},
	{0xD768CD72753832E7ULL, "zz_dae_debug6029517.cl"},
	{0xD7690B466A184ADBULL, "676938301.ogg"},
	{0xD769FEC26DA36799ULL, "dae000ip_crowd_ff2_talk4.utp"},
	{0xD76A4B0D3E0EC43FULL, "6112007_m.fxe"},
	{0xD76AE4480879D76FULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgta.mao"},
	{0xD76AF0480879EB0CULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgta.mmh"},
	{0xD76AF6480879F5FEULL, "art\\characters\\playercharacter\\dwarfmale\\boo_lgta.msh"},
	{0xD76B1FC856E0FA3CULL, "hlac_ht_mansion2_818_7y.dds"},
	{0xD76BAD041B45DCEDULL, "hlac_lt_0_lowtown_docks_-262144_1x.dds"},
	{0xD76BB80A85CC06C5ULL, "6085143_m.fxe"},
	{0xD76CC98917E0EBC3ULL, "mcr371_emile_de_launcet.cnv"},
	{0xD76D325E410D9DE9ULL, "6153028_m.fxe"},
	{0xD76D43520D64C35DULL, "567559545.ogg"},
	{0xD76D818A7B60052CULL, "32842142.ogg"},
	{0xD76E011D6BD3B9D1ULL, "follower_banter6163129.cl"},
	{0xD76E2807168B7BF1ULL, "917216805.ogg"},
	{0xD76E4C00CEC6B1EDULL, "dae000_orsino6081388.cl"},
	{0xD76EBFE22AEA258BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_angry.evt"},
	{0xD76EE1FE49D93F2DULL, "6083745_cleric_captainguard.cut"},
	{0xD76EFB29781B7CF8ULL, "docks_male_human_worker_sta6175966.cl"},
	{0xD76F3F725C3AD344ULL, "gen_im_trs_apo_lgt_hlm_r0.uti"},
	{0xD7700590D1098EE2ULL, "vxlt_gallowscourtyard_4v0_0_0.vlm"},
	{0xD7703932C937279CULL, "vxht_keep_610_0_0.vlm"},
	{0xD7706318F7563ECDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneya.msh"},
	{0xD7707118F7565600ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneya.mao"},
	{0xD7707518F7565D53ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneya.mmh"},
	{0xD770CC5D9E2CFC90ULL, "vxlt_0_lowtown_docks_n_b78_0_0.vlm"},
	{0xD77102F670B85AF0ULL, "rdr111dg_martin6054269.cl"},
	{0xD7715D4317BCCA1CULL, "llac_ht_mansion2b_818_4.dds"},
	{0xD7719BDCFE8586D4ULL, "qun120_arishok6051670.cl"},
	{0xD771AA30E14D4B54ULL, "vxlt_undercitydungeon_17a_0_0.vlm"},
	{0xD771AEBBF37A3E90ULL, "gen00fl_isabela6101479.cl"},
	{0xD771F1B24F7B5B6BULL, "6098931_f.fxe"},
	{0xD77283E404C48AE5ULL, "6042206_f.fxe"},
	{0xD772CCF0DBD302BAULL, "vxlt_0_lowtown_n_h26_0_0.vlm"},
	{0xD772E7A3BD592F99ULL, "vxht_mansion_n_4wf_0_0.vlm"},
	{0xD77338FC4CE0421DULL, "vxlt_warehouse_n_y8_0_0.vlm"},
	{0xD773436DC79FA893ULL, "6043019_m.fxe"},
	{0xD7735EA64000D556ULL, "vxht_keep_62c_0_0.vlm"},
	{0xD773E07D985D9E30ULL, "ca_caverns1_mines_24427_lc.mmh"},
	{0xD773F0672DBE906BULL, "hlac_w_sundermount_-131071_0x.dds"},
	{0xD774461CEB171D2EULL, "695636225.ogg"},
	{0xD774999A9B19690EULL, "art\\characters\\creatures\\qunari\\textures\\qm_hlm_magb_0d.dds"},
	{0xD77503AC981C7BDDULL, "vxht_keep_n_49d_0_0.vlm"},
	{0xD7752AA22ABB7281ULL, "vxlt_0_lowtown_a2_f1q_0_0.vlm"},
	{0xD7756EA7541429E8ULL, "art\\cinematics\\_crit path\\followerintros\\cs_isa_intro_hench3.gad"},
	{0xD775BC0F98237508ULL, "lowtown_male_alienage1_sta6162738.cl"},
	{0xD775D2BE96BBB35FULL, "gen00fl_anders6125508.cl"},
	{0xD775D8E22AF06F4CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_angry.gad"},
	{0xD7767812C9086D27ULL, "one300_zevran6077760.cl"},
	{0xD77704DFFF44FA7DULL, "gen00fl_anders6119224.cl"},
	{0xD777E6B0AE13991DULL, "vxlt_darktown_4u0_0_0.vlm"},
	{0xD777E84689440832ULL, "vxlt_0_lowtown_a2_n_gz9_0_0.vlm"},
	{0xD778149AF9679DFFULL, "161033380.ogg"},
	{0xD7782798E84A57AFULL, "218544957.ogg"},
	{0xD77841EA19BEA6E9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_arches.phy"},
	{0xD7792DA02EBB1AEEULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galolargehi.mmh"},
	{0xD77939A02EBB2E85ULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galolargehi.mao"},
	{0xD77947A02EBB46DCULL, "art\\levels\\races\\ferelden\\kirkgallows\\ext\\4co_gallows\\vista_galolargehi.msh"},
	{0xD7797DCB034B1ED6ULL, "vxlt_gallowstemplar_n_4yl_0_0.vlm"},
	{0xD77995FAB97C73D5ULL, "gen_im_trs_bst_drg_scale_01.uti"},
	{0xD779E620C0B704CCULL, "96791743.ogg"},
	{0xD779F420A9E5F902ULL, "vxlt_0_lowtown_docks_b00_0_0.vlm"},
	{0xD779F47A6309939CULL, "6115863_m.fxe"},
	{0xD779FE4886289BFDULL, "vxlt_darktown_2qf_0_0.vlm"},
	{0xD77B1B48DAAE19C3ULL, "526055637.ogg"},
	{0xD77B478F6F9EAC81ULL, "z_cloud_2.dds"},
	{0xD77B761AE6598261ULL, "292938068.ogg"},
	{0xD77B7CE2C147D6ACULL, "hlac_ht_mansion2_818_24x.dds"},
	{0xD77BA43B3303B669ULL, "6156328_m.fxe"},
	{0xD77C656BDC306079ULL, "vxlt_warehouse_n_153_0_0.vlm"},
	{0xD77C69D7B02FAE1FULL, "vxht_bartrand_n_5yt_0_0.vlm"},
	{0xD77C889240CBDCA8ULL, "341718110.ogg"},
	{0xD77CAAFC7DA8AB8CULL, "6148793_m.fxe"},
	{0xD77CC3EA9B0E2B7EULL, "vxdr_0_deeproads_fh6_0_0.vlm"},
	{0xD77CD7E22AF68C4AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_angry.ani"},
	{0xD77D0E3ABBDFA8A7ULL, "hlac_lt_0_lowtown_docks_n_65534_0y.dds"},
	{0xD77D3957C473A09EULL, "vxht_mansion_n_3hk_0_0.vlm"},
	{0xD77D78AE40E7F630ULL, "624225490.ogg"},
	{0xD77D87AE5D0272F4ULL, "hlac_lt_0_lowtown_a2_-131072_5y.dds"},
	{0xD77DA059D737F815ULL, "vxlt_0_lowtown_n_byr_0_0.vlm"},
	{0xD77DA5BA4EC75B9AULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.cess.gad"},
	{0xD77E00BCFAB3E045ULL, "6139672_f.fxe"},
	{0xD77E5034962E3B2CULL, "lt_0_lowtown_a2_n_46_0_lc.mmh"},
	{0xD77E969F8E23BD60ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm05.phy"},
	{0xD77FBBBB6FE78688ULL, "gen00fl_isabela6079567.cl"},
	{0xD7806A185795F051ULL, "ser120_ghyslain6104193.cl"},
	{0xD7807936690F443CULL, "hlac_w_woundedcoast_-131069_0x.dds"},
	{0xD780B8BF71991CEEULL, "599235582.ogg"},
	{0xD780C8A1D8CCC348ULL, "723315254.ogg"},
	{0xD780DAC29C848B33ULL, "6110987_f.fxe"},
	{0xD781A75AF69D30E1ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgtb_0d.dds"},
	{0xD7828A4ACA905179ULL, "narrative_varric6180095.cl"},
	{0xD782D97CBCC8B197ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_teag.mmh"},
	{0xD782EF7CBCC8D709ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_teag.msh"},
	{0xD7831A4C6A2C33F1ULL, "971762997.ogg"},
	{0xD7834CCA16CAF78FULL, "hlac_lt_gallowscourtyard_n_-65537_3x.dds"},
	{0xD783B991306B3AEAULL, "281081503.ogg"},
	{0xD783D02FEDA38604ULL, "ser130ar_emeric_encounter.are"},
	{0xD783F5A84F18CCB3ULL, "6176557_m.fxe"},
	{0xD78445F243D14DA4ULL, "art\\vfx\\combat\\vfx_str_appearance_c.anb"},
	{0xD78498DAD10F70B9ULL, "w_bonepitmining_a2_100_0.msh"},
	{0xD784A2DAD10F81A7ULL, "w_bonepitmining_a2_100_0.mmh"},
	{0xD784A6DAD10F88F4ULL, "w_bonepitmining_a2_100_0.mao"},
	{0xD784A8BA4ECD7FB1ULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.cess.evt"},
	{0xD784BA84CB5E64BFULL, "182194947.ogg"},
	{0xD785310B585DFDF2ULL, "822856760.ogg"},
	{0xD785311AFE5CCC98ULL, "289673744.ogg"},
	{0xD785D1E210F7DDFCULL, "6103381_m.fxe"},
	{0xD785DF0B93C2E125ULL, "vxlt_hangedman_20l_0_0.vlm"},
	{0xD786D7A2A6280664ULL, "271204853.ogg"},
	{0xD787D5C9BA14F452ULL, "llac_lt_0_lowtown_a2_n_c_-262144_4.dds"},
	{0xD787F792B97D8AAEULL, "lt_0_lowtown_a2_37_0.anb"},
	{0xD7883C4A54E58B66ULL, "6086189_m.fxe"},
	{0xD788560A640AD844ULL, "hlac_lt_0_lowtown_docks_a3_-65540_2z.dds"},
	{0xD788615AC6D87166ULL, "hlac_ht_0_hightown_a2_-65536_3z.dds"},
	{0xD7886878F34DCC3AULL, "6080437_m.fxe"},
	{0xD78890673CFC239DULL, "art\\characters\\weapons\\axes\\w_axe104a.mao"},
	{0xD788987A835A0D41ULL, "vxlt_0_lowtown_docks_a3_n_d2s_0_0.vlm"},
	{0xD788A95DBDF41DC0ULL, "vxca_caverns2_7en_0_0.vlm"},
	{0xD788C92CBCBA6946ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.anb"},
	{0xD788E01622A3D9F2ULL, "vxlt_0_lowtown_gpd_0_0.vlm"},
	{0xD78902E845084BCCULL, "zz_and_debug6134161.cl"},
	{0xD7891B1949987BEFULL, "vxca_caverns2prc_a6k_0_0.vlm"},
	{0xD7893CDD40FB085FULL, "gen_im_trs_gen_shd_met_r0.uti"},
	{0xD78ADBAAC2DFF175ULL, "blood_conj_01.mao"},
	{0xD78B784CECE60787ULL, "vxlt_0_lowtown_n_f58_0_0.vlm"},
	{0xD78BBB440C1BF77AULL, "hlac_w_encounter2_camp_262146_0y.dds"},
	{0xD78C82149FF7E2BBULL, "731498315.ogg"},
	{0xD78CCCF40DEFB181ULL, "vxht_0_hightown_a2_2xj_0_0.vlm"},
	{0xD78D444E65A97106ULL, "mcr370_meredith6085895.cl"},
	{0xD78D57BCCC8CD969ULL, "109080225.ogg"},
	{0xD78D7290A961BCF8ULL, "135948169.ogg"},
	{0xD78DACD2D9D08B30ULL, "6169900_f.fxe"},
	{0xD78E0CCE28F819B3ULL, "849263331.ogg"},
	{0xD78E2175A72459D7ULL, "6176734_f.fxe"},
	{0xD78E32375A6FDF16ULL, "lt_0_lowtown_a3_n_c_26_0_ll.mmh"},
	{0xD78E44452E52F3B6ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbench.msh"},
	{0xD78E4E452E5304B4ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbench.mmh"},
	{0xD78E52452E530BC7ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbench.mao"},
	{0xD78E76713DBC8A8EULL, "216502556.ogg"},
	{0xD78E9F901C322552ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_4y.dds"},
	{0xD78EC705D3D95F15ULL, "vxlt_0_lowtown_h88_0_0.vlm"},
	{0xD78ECB6B6F8B5D3AULL, "405915877.ogg"},
	{0xD78FA08EEBBBF270ULL, "89992515.ogg"},
	{0xD790646C4C3893AFULL, "6143131_m.fxe"},
	{0xD790E2044E2212A8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eaglehead_d.dds"},
	{0xD79264F243DDBFA7ULL, "art\\vfx\\combat\\vfx_str_appearance_c.mmh"},
	{0xD792B1FAC92A2A95ULL, "art\\vfx\\creatures\\fxc_drg_brth_c_c.anb"},
	{0xD792C0BA4ED9E580ULL, "art\\vfx\\cinematic\\vfx_frozen_no_ice_c.cess.ani"},
	{0xD793BCAE162A9E76ULL, "212816752.ogg"},
	{0xD79425965659775DULL, "vxw_woundedcoast_fn_0_0.vlm"},
	{0xD7942D93ECDF37FEULL, "hlac_lt_blackemporium_1055_17x.dds"},
	{0xD794EDAA9F88C61CULL, "hlac_lt_0_lowtown_a3_n_c_0_1z.dds"},
	{0xD794FE15D6600EECULL, "6099451_m.fxe"},
	{0xD7951024EBA5CCDBULL, "lt_0_lowtown_a2_n_c_10_0.mmh"},
	{0xD7951889F031C104ULL, "288250889.ogg"},
	{0xD7951C24EBA5E178ULL, "lt_0_lowtown_a2_n_c_10_0.mao"},
	{0xD7952E24EBA5FFC5ULL, "lt_0_lowtown_a2_n_c_10_0.msh"},
	{0xD79576AB1884E93CULL, "hlac_ht_0_hightown_n_-65537_1x.dds"},
	{0xD795CC2CBCC4F8F1ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.mmh"},
	{0xD795DECBF9942B61ULL, "6097904_m.fxe"},
	{0xD79612A0A24743AEULL, "616696974.ogg"},
	{0xD7966184E7374FD6ULL, "vxlt_gallowstemplar_n_4sn_0_0.vlm"},
	{0xD79684969E2B8D45ULL, "hlac_lt_undercitydungeon_-1_51z.dds"},
	{0xD79708C4EDF8BBBAULL, "vxht_0_hightown_n_bwc_0_0.vlm"},
	{0xD7978B62806CA052ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.mmh"},
	{0xD797E67C7460C63AULL, "vxlt_gallowstemplar_n_43b_0_0.vlm"},
	{0xD79806945576F668ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.anb"},
	{0xD7986C21DCE68D9EULL, "25208809.ogg"},
	{0xD7988FCDA4BAB7DEULL, "280625744.ogg"},
	{0xD79963DB262F363FULL, "vxlt_darktown_5qc_0_0.vlm"},
	{0xD799A539CDCAC15FULL, "hlac_w_woundedcoast_-131070_0y.dds"},
	{0xD79A0AFC0C187610ULL, "vxlt_0_lowtown_n_d5u_0_0.vlm"},
	{0xD79A7F61B0899CA0ULL, "9862845.ogg"},
	{0xD79AB0D3647DD421ULL, "vxlt_0_lowtown_h97_0_0.vlm"},
	{0xD79AB22A8CB8D26EULL, "vxca_caverns1_mines_cjw_0_0.vlm"},
	{0xD79ACF02DEEB1623ULL, "206295450.ogg"},
	{0xD79ADF3730C002AEULL, "390092720.ogg"},
	{0xD79B25A976E86DDFULL, "6148736_m.fxe"},
	{0xD79B522EAF7A961CULL, "6087823_m.fxe"},
	{0xD79B62381DF389F1ULL, "follower_banter6172310.cl"},
	{0xD79BD386AC10998CULL, "dae000_bodhan6129128.cl"},
	{0xD79BE9DBCAC2EAEEULL, "591022746.ogg"},
	{0xD79C417616B6A876ULL, "hlac_lt_0_lowtown_-196607_0y.dds"},
	{0xD79C8D6531980C1AULL, "802743879.ogg"},
	{0xD79D2F22D2D93C4CULL, "character_character_p104.psh.fxo.dx11"},
	{0xD79D3AE819B82FCEULL, "6095893_m.fxe"},
	{0xD79E3D3A933B7919ULL, "vxht_mansion_n_4dk_0_0.vlm"},
	{0xD79F335CA3F53A9BULL, "156271834.ogg"},
	{0xD79F417D407FE69FULL, "llac_ca_caverns2_5802_44.dds"},
	{0xD79FBDEBB6D20B3FULL, "6105880_f.fxe"},
	{0xD7A0123A246B4C3BULL, "plt_cod_itm_garahel.plo"},
	{0xD7A01ED257A3470AULL, "qcr000_arishok6057101.cl"},
	{0xD7A03840812EE5B2ULL, "6095004_m.fxe"},
	{0xD7A06E01D45290BDULL, "dae000_hate_follower_react6103212.cl"},
	{0xD7A098DFA471D175ULL, "w_encounter3_133_0.phy"},
	{0xD7A1696A2555A314ULL, "6139655_f.fxe"},
	{0xD7A1FF58070C2579ULL, "6118317_m.fxe"},
	{0xD7A24C4DA2406ED2ULL, "6147707_m.fxe"},
	{0xD7A26C5108FF5A79ULL, "zz_seb_debug6187884.cl"},
	{0xD7A29773484F8FDDULL, "vxw_woundedcoast_cm_0_0.vlm"},
	{0xD7A2D36A9C1E0D98ULL, "696161676.ogg"},
	{0xD7A2D72FEDBE42A6ULL, "ser130ar_emeric_encounter.ncs"},
	{0xD7A2DD817DA1662AULL, "mag291_vincento6027395.cl"},
	{0xD7A319AA54E31C5CULL, "vxw_encounter2_camp_8p_0_0.vlm"},
	{0xD7A384163996969EULL, "vxlt_hangedman_n_1qp_0_0.vlm"},
	{0xD7A41AE2F6EF0583ULL, "403379153.ogg"},
	{0xD7A42AEA19E3BA0EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_arches.mmh"},
	{0xD7A436EA19E3CEA5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_arches.mao"},
	{0xD7A444EA19E3E6FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_arches.msh"},
	{0xD7A4B39F8E44BDF5ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm05.msh"},
	{0xD7A4B59F8E44C16BULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm05.mmh"},
	{0xD7A513F3EF137EB4ULL, "hlac_lt_0_lowtown_docks_a3_-131073_2y.dds"},
	{0xD7A525A65022E491ULL, "vxca_caverns2_62o_0_0.vlm"},
	{0xD7A564510BD4C18DULL, "t3_hey_gy1.tnt"},
	{0xD7A585E32E6EB9CEULL, "plt_cod_ltr_enemies_macha.plo"},
	{0xD7A634B1908BEFC0ULL, "vxht_chantry_n_2z6_0_0.vlm"},
	{0xD7A66002909F41C2ULL, "6060123_m.fxe"},
	{0xD7A6FDCE428F3FE1ULL, "mag371_elsa6041154.cl"},
	{0xD7A7E370F38D53CAULL, "675276968.ogg"},
	{0xD7A7E4B47834DD83ULL, "6162989_m.fxe"},
	{0xD7A84D7F9F25979AULL, "6043973_m.fxe"},
	{0xD7A8700AB0378D55ULL, "hlac_lt_0_lowtown_a2_131071_2y.dds"},
	{0xD7A943B82A379A29ULL, "322442677.ogg"},
	{0xD7A9C8DCF8B7F96FULL, "6087644_m.fxe"},
	{0xD7A9CF2FEDC4539FULL, "ser130ar_emeric_encounter.lst"},
	{0xD7AA86C43FD6C184ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha07a.phy"},
	{0xD7AA9D69DB4C348BULL, "gen00fl_aveline6107082.cl"},
	{0xD7AABD67B0AB528FULL, "vxlt_0_lowtown_a3_n_c_ep3_0_0.vlm"},
	{0xD7AAC126E81F1661ULL, "hlac_ca_caverns1_mines_24426_0y.dds"},
	{0xD7AAC2F010A548A8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobd_0t.dds"},
	{0xD7AAEB25D7142D29ULL, "gen00fl_carver6102691.cl"},
	{0xD7AB09860D26071FULL, "hlac_ca_caverns1_24421_5z.dds"},
	{0xD7AB34C6890A7686ULL, "mcr371_alain6087130.cl"},
	{0xD7AC0F8E746B5521ULL, "vxlt_0_lowtown_eqn_0_0.vlm"},
	{0xD7AC4D6C4EE569BDULL, "810194105.ogg"},
	{0xD7AC698C45716131ULL, "6134456_m.fxe"},
	{0xD7AC7D7DA2F3D9D2ULL, "320019665.ogg"},
	{0xD7ACF48FFB36EE05ULL, "follower_banter6148467.cl"},
	{0xD7AD1A08AF3BBFC2ULL, "ht_store_mmap.dds"},
	{0xD7AD485E8D9AC943ULL, "dae000_merchant_generic_hm6172626.cl"},
	{0xD7AE4289364BE0BBULL, "hlac_lt_0_lowtown_n_-3_4y.dds"},
	{0xD7AE67CC4E968331ULL, "gen00fl_merrill6119657.cl"},
	{0xD7AE6A63182C1B46ULL, "6161329_m.fxe"},
	{0xD7AE8F8B75BFF5D4ULL, "art\\vfx\\combat\\vfx_glb_cold_explo_p.impa.ani"},
	{0xD7AEA3974353F771ULL, "vxlt_0_lowtown_e81_0_0.vlm"},
	{0xD7AFC1C2C23A4B42ULL, "221414316.ogg"},
	{0xD7B00A9A0216C567ULL, "6109747_f.fxe"},
	{0xD7B01B7ED229572EULL, "268446710.ogg"},
	{0xD7B05A728797F7ADULL, "dae000tr_base.ncs"},
	{0xD7B09F9C6565ABFCULL, "197819689.ogg"},
	{0xD7B109503A45A85DULL, "vxw_woundedcoast_xl_0_0.vlm"},
	{0xD7B11CB3BDBE5309ULL, "vxht_0_hightown_cat_0_0.vlm"},
	{0xD7B18B8746D934DCULL, "vxw_encounter2_camp_3q_0_0.vlm"},
	{0xD7B21024A68A79DEULL, "fen110_fenris6096431.cl"},
	{0xD7B2DB001A91030EULL, "hlac_w_bonepit_0_6x.dds"},
	{0xD7B301F1717A0197ULL, "hlac_lt_0_lowtown_-131072_5x.dds"},
	{0xD7B34F66B07E1404ULL, "978023865.ogg"},
	{0xD7B3AA9EA07BAAAAULL, "zz_one_debug6051524.cl"},
	{0xD7B3E2B8818DA0F4ULL, "vxw_bonepit_29o_0_0.vlm"},
	{0xD7B45C9CD656167DULL, "character_character_p97.psh.fxo.dx11"},
	{0xD7B488A924EC3D50ULL, "isa120tr_chantry_run.utt"},
	{0xD7B4AA7D3AC34BF5ULL, "vxht_brothel_n_4xp_0_0.vlm"},
	{0xD7B4D94F7CF3998CULL, "9666810.ogg"},
	{0xD7B4F3C867E369CAULL, "gen00fl_varric6105415.cl"},
	{0xD7B5456D23A4C59DULL, "6177295_m.fxe"},
	{0xD7B58C8B75C60F36ULL, "art\\vfx\\combat\\vfx_glb_cold_explo_p.impa.gad"},
	{0xD7B5D75F0E4A6CBDULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowc_sa.dds"},
	{0xD7B5EDC229A961FCULL, "1031670057.ogg"},
	{0xD7B5F151F4827E77ULL, "vxlt_0_lowtown_a2_fga_0_0.vlm"},
	{0xD7B62782C05C439EULL, "6088305_m.fxe"},
	{0xD7B6314B7808101FULL, "vxht_bartrand_n_5mp_0_0.vlm"},
	{0xD7B641B4706185B0ULL, "150066051.ogg"},
	{0xD7B6CFFD2FABBF62ULL, "w_encounter2_camp_42_0.msh"},
	{0xD7B6D1760C47F1E5ULL, "dae000ip_show_world_map.utp"},
	{0xD7B6D1FD2FABC218ULL, "w_encounter2_camp_42_0.mmh"},
	{0xD7B6DDFD2FABD6BBULL, "w_encounter2_camp_42_0.mao"},
	{0xD7B730323907E111ULL, "364573720.ogg"},
	{0xD7B7B94D5B958B15ULL, "6097231_f.fxe"},
	{0xD7B7E7A2B0F52235ULL, "269391115.ogg"},
	{0xD7B86252146923DCULL, "580986264.ogg"},
	{0xD7B95B9495B1582DULL, "6217515_m.fxe"},
	{0xD7B976E32E7F495DULL, "plt_cod_ltr_enemies_macha.nss"},
	{0xD7B9D1FD2FADF82BULL, "w_encounter2_camp_42_0.rml"},
	{0xD7BA36C883740B61ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\hf_hed_casa_0d.dds"},
	{0xD7BA5DA565EAE1EEULL, "341584946.ogg"},
	{0xD7BA6A507C2E85E0ULL, "ser221_quentin6123327.cl"},
	{0xD7BAB2475A6AB0C3ULL, "hlac_lt_0_lowtown_-2_1z.dds"},
	{0xD7BAF4C333E2356DULL, "zz_fol_debug6097809.cl"},
	{0xD7BAFA92B9A8E912ULL, "lt_0_lowtown_a2_37_0.rml"},
	{0xD7BB1282A7E406B2ULL, "dae000_mother6135362.cl"},
	{0xD7BB5B17C725E486ULL, "art\\characters\\playercharacter\\qunarimale\\qm_hlm_magb.mao"},
	{0xD7BBA78B75CAA895ULL, "art\\vfx\\combat\\vfx_glb_cold_explo_p.impa.evt"},
	{0xD7BBB2FAC94D31D6ULL, "art\\vfx\\creatures\\fxc_drg_brth_c_c.mmh"},
	{0xD7BC173FB68BDC96ULL, "mag101_aombanter6052776.cl"},
	{0xD7BC3F2B90ED1CBCULL, "965286600.ogg"},
	{0xD7BD1574354A0FF1ULL, "vxlt_0_lowtown_e30_0_0.vlm"},
	{0xD7BD22E7BB5DFD51ULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtczz.msh"},
	{0xD7BD3CE7BB5E290FULL, "art\\characters\\playercharacter\\humanmale\\boo_lgtczz.mmh"},
	{0xD7BD5081B4AF8499ULL, "6128978_m.fxe"},
	{0xD7BDD8697154DA89ULL, "vxht_0_hightown_a3_aqs_0_0.vlm"},
	{0xD7BDFF2F59EB2637ULL, "dae300_bran6143473.cl"},
	{0xD7BEE311EAF85679ULL, "6112117_m.fxe"},
	{0xD7BF7B2D2C3BC0DDULL, "vxw_woundedcoast_uk_0_0.vlm"},
	{0xD7BFF12676B6B518ULL, "6096458_m.fxe"},
	{0xD7C03FB058EB31EEULL, "6060541_m.fxe"},
	{0xD7C0455393089965ULL, "vxlt_0_lowtown_h8h_0_0.vlm"},
	{0xD7C08A62808FA3F9ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.anb"},
	{0xD7C0ABEF603403C9ULL, "380897147.ogg"},
	{0xD7C0D8FD2FB422A3ULL, "w_encounter2_camp_42_0.phy"},
	{0xD7C0E687A7647483ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.active.ani"},
	{0xD7C105945599F903ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.mmh"},
	{0xD7C105C4F41812D1ULL, "vxca_caverns1_eh9_0_0.vlm"},
	{0xD7C208921E0231A3ULL, "vxht_mansion2b_30m_0_0.vlm"},
	{0xD7C23AFC79664743ULL, "zz_dae_debug6029171.cl"},
	{0xD7C2537138EA4322ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdvista01.mmh"},
	{0xD7C269D76546D0DBULL, "art\\vfx\\cinematic\\vfx_chest_pierce_charge_c.impa.evt"},
	{0xD7C283CCCAB240DEULL, "vxca_caverns1_mines_hvl_0_0.vlm"},
	{0xD7C3085E3B49E676ULL, "hlac_lt_gallowstemplar_n_2388_6x.dds"},
	{0xD7C3186B747820C0ULL, "vxht_chantry_n_2p4_0_0.vlm"},
	{0xD7C35414F6303842ULL, "zz_mageslate_trans2.cut"},
	{0xD7C3AC10869F9620ULL, "gen00fl_aveline6114869.cl"},
	{0xD7C3B168B207B5FFULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\minor.gad"},
	{0xD7C3B3A3EBC6731EULL, "522843103.ogg"},
	{0xD7C43B3315A9143FULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf01.phy"},
	{0xD7C4EA022F1FAD1CULL, "vxlt_gallowscourtyard_n_7ac_0_0.vlm"},
	{0xD7C53A566C820EABULL, "lgt110tr_coa_meet_fight.ncs"},
	{0xD7C57678DF04C373ULL, "dae000ip_crowd_lm2_talk14.utp"},
	{0xD7C5B90361E1CDD2ULL, "lt_0_lowtown_n_45_0_ll.mmh"},
	{0xD7C5C6244992D145ULL, "6116197_m.fxe"},
	{0xD7C5DE7A9D40E0C3ULL, "6129324_m.fxe"},
	{0xD7C6C185929E965AULL, "mag160_varian_ilithis6016491.cl"},
	{0xD7C729F0DD3079E1ULL, "6064717_m.fxe"},
	{0xD7C7494D2E3E97C8ULL, "764312429.ogg"},
	{0xD7C7E8D2E7A049E9ULL, "vxca_caverns1_mines_cn0_0_0.vlm"},
	{0xD7C8F56B0869AB7FULL, "282484609.ogg"},
	{0xD7C962D7654CE31CULL, "art\\vfx\\cinematic\\vfx_chest_pierce_charge_c.impa.gad"},
	{0xD7C9A668B20C0EECULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\minor.evt"},
	{0xD7CA6A89A1103558ULL, "6090409_m.fxe"},
	{0xD7CA807754D8CB1EULL, "fxa_cen_splash.dds"},
	{0xD7CADC73C3244ADCULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredidol.ani"},
	{0xD7CB0092189F0174ULL, "6024566_f.fxe"},
	{0xD7CB0341A87DA53FULL, "779986711.ogg"},
	{0xD7CC4E9E269CDA87ULL, "mer140_bonepit_comments6062373.cl"},
	{0xD7CC77F3953E3C73ULL, "hlac_w_sundermount_262147_0x.dds"},
	{0xD7CD0600CC152961ULL, "dae000_pickpocket_scene6175104.cl"},
	{0xD7CD347E9ACA9C9FULL, "qun200_party_comment6090407.cl"},
	{0xD7CDACCA3E9F6229ULL, "569593048.ogg"},
	{0xD7CDF55BA2CA8993ULL, "follower_banter6170699.cl"},
	{0xD7CECC09680E8D0EULL, "mrl310_merrill6143812.cl"},
	{0xD7D0358FE8979787ULL, "hlac_w_bonepitdestroyed_a3_1_0x.dds"},
	{0xD7D061D76553001AULL, "art\\vfx\\cinematic\\vfx_chest_pierce_charge_c.impa.ani"},
	{0xD7D063F8C640D50BULL, "vxlt_0_lowtown_f0p_0_0.vlm"},
	{0xD7D0A2971BD9F11CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_forward.ani"},
	{0xD7D0D715E55366D4ULL, "llac_lt_0_lowtown_-3_5.dds"},
	{0xD7D0E973C328CC1EULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredidol.gad"},
	{0xD7D191E708129252ULL, "616161397.ogg"},
	{0xD7D19818B67E6776ULL, "mer120_hubert6059435.cl"},
	{0xD7D198766078471AULL, "6128974_f.fxe"},
	{0xD7D1AB24B1B3126BULL, "rune_weapon_nature.dds"},
	{0xD7D1BC66429EAC51ULL, "294211822.ogg"},
	{0xD7D210C3BC1B8CC5ULL, "mrl210_merrill6102156.cl"},
	{0xD7D303327866083BULL, "vxw_encounter2_pb_0_1.vlm"},
	{0xD7D36C6FB15A3AEDULL, "vxlt_undercityserialkiller_1ib_0_1.vlm"},
	{0xD7D40D8633FE72A2ULL, "hlac_ca_caverns2_5802_17y.dds"},
	{0xD7D41CC822C0735BULL, "733241094.ogg"},
	{0xD7D4A33971D58C40ULL, "107392224.ogg"},
	{0xD7D4FD87A775456DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dwarven_trap.active.gad"},
	{0xD7D53C3315B78944ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf01.mmh"},
	{0xD7D5523315B7AEE6ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf01.msh"},
	{0xD7D5B8A92A9D4D5CULL, "genip_crafting_node.ncs"},
	{0xD7D5D14E83FCC21DULL, "6077818_m.fxe"},
	{0xD7D624ED81A9349CULL, "668535212.ogg"},
	{0xD7D6CC8E31AF80CEULL, "ts_profane_03.gda"},
	{0xD7D77083C0B200B1ULL, "spit_shot.mao"},
	{0xD7D78E68B2182345ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\minor.ani"},
	{0xD7D7A7F55D407B2AULL, "hlac_lt_gallowstemplar_2391_2y.dds"},
	{0xD7D7AF971BE0265EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_forward.gad"},
	{0xD7D834FE2A50CC45ULL, "6133601_f.fxe"},
	{0xD7D84896FAEFCDA9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_006a_0d.dds"},
	{0xD7D8D47391012EE2ULL, "466378219.ogg"},
	{0xD7D90BF7974C27E1ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6.mmh"},
	{0xD7D93A1EDC59A651ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_basic_2.mao"},
	{0xD7D973321C6B8778ULL, "vxlt_0_lowtown_a2_n_csh_0_0.vlm"},
	{0xD7D977A9250B3C0DULL, "isa120tr_chantry_run.ncs"},
	{0xD7D9B3B7BA01A14FULL, "6116388_m.fxe"},
	{0xD7DA13B21D283E84ULL, "6125648_m.fxe"},
	{0xD7DA1E1D07363331ULL, "334708222.ogg"},
	{0xD7DA6D15A34FAF93ULL, "guiachievementunlocked.gfx"},
	{0xD7DAAABAACE96737ULL, "hlac_lt_0_lowtown_a2_-196609_4x.dds"},
	{0xD7DAB6FD50794359ULL, "46935648.ogg"},
	{0xD7DAD3C8C9825A83ULL, "vxht_0_hightown_n_caa_0_0.vlm"},
	{0xD7DB20A82E769B59ULL, "vxht_mansion2_3g6_0_0.vlm"},
	{0xD7DBD70B4CDE3ECEULL, "312144555.ogg"},
	{0xD7DBEDC7AF51A256ULL, "ss_morrigan.dds"},
	{0xD7DC957EE17A82D8ULL, "vxlt_0_lowtown_n_i4c_0_0.vlm"},
	{0xD7DCC41004944B9CULL, "w_bonepit_85_0.mao"},
	{0xD7DCD0100494603FULL, "w_bonepit_85_0.mmh"},
	{0xD7DCD61004946AE1ULL, "w_bonepit_85_0.msh"},
	{0xD7DD00BE75C50E3DULL, "vxw_bonepit_nk_0_0.vlm"},
	{0xD7DD2A2BB299801BULL, "gen00fl_isabela6075229.cl"},
	{0xD7DD3124EBE3BA10ULL, "lt_0_lowtown_a2_n_c_10_0.phy"},
	{0xD7DD991FFF88A3CFULL, "vxdr_primevaltomb_7t4_0_0.vlm"},
	{0xD7DDB04100B9C89CULL, "lt_0_lowtown_n_28_0_lc.mmh"},
	{0xD7DDC4F713ABE9CAULL, "vxca_caverns2prc_9u3_0_0.vlm"},
	{0xD7DE27C5E419B609ULL, "256360896.ogg"},
	{0xD7DE5957074CF85CULL, "6115100_m.fxe"},
	{0xD7DE67B55F236610ULL, "6043838_m.fxe"},
	{0xD7DE8A971BE6067DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_lean_forward.evt"},
	{0xD7DEC03FA5D828A5ULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.anb"},
	{0xD7DECB2B6DD79761ULL, "vfx_blt_lava_emitter.wbd"},
	{0xD7DEDE84EBECCD27ULL, "hlac_w_bonepitmining_a2_65542_0y.dds"},
	{0xD7DEF1821E813BD8ULL, "dae000_mother6127143.cl"},
	{0xD7DFC7B3C2DEDE33ULL, "hlac_w_encounter2_camp_131074_0y.dds"},
	{0xD7DFE5A9179CFC0EULL, "llac_lt_undercityserialkiller_-1_13.dds"},
	{0xD7E02C0B03A5DAEEULL, "6112565_m.fxe"},
	{0xD7E071696E1F640CULL, "vxlt_hangedman_t3_0_0.vlm"},
	{0xD7E07943339398EAULL, "zz_one_debug6077403.cl"},
	{0xD7E09A97DBDDD23CULL, "vxht_mansion_n_4d2_0_0.vlm"},
	{0xD7E142CF1ED88761ULL, "w_woundedcoast_179_0.phy"},
	{0xD7E182330D292E97ULL, "gen_im_arm_hlm_hvy_act1_03.uti"},
	{0xD7E231D740244398ULL, "301121017.ogg"},
	{0xD7E33DBDE6CAFB79ULL, "t3_wdg_w17.tnt"},
	{0xD7E33F566C9B1492ULL, "lgt110tr_coa_meet_fight.utt"},
	{0xD7E351EAF5A9FE9CULL, "vxca_caverns1_mines_j4v_0_0.vlm"},
	{0xD7E3C9D3A0642884ULL, "855941280.ogg"},
	{0xD7E41024EBE9A168ULL, "lt_0_lowtown_a2_n_c_10_0.rml"},
	{0xD7E4710965CE85E1ULL, "6165278_m.fxe"},
	{0xD7E49B3185A22ED2ULL, "fx_twist.dds"},
	{0xD7E56BC15FBD3A1EULL, "seb221_flora_harimann6177682.cl"},
	{0xD7E5B7C3D88AB64CULL, "vxht_mansion_n_6bf_0_0.vlm"},
	{0xD7E618ED5B6A8769ULL, "vxlt_0_lowtown_n_d6y_0_0.vlm"},
	{0xD7E61F46E73D3B30ULL, "docks_male_human_worker_sta6176044.cl"},
	{0xD7E6206D2515903AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala0a.msh"},
	{0xD7E62E6D2515A873ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala0a.mao"},
	{0xD7E6326D2515AE80ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala0a.mmh"},
	{0xD7E65B5EE3F8C260ULL, "zz_approval_debug.cnv"},
	{0xD7E6FA7041B7CFBAULL, "art\\vfx\\textures\\vfx_blood_dir02.dds"},
	{0xD7E723852A1647D8ULL, "db_cl.mao"},
	{0xD7E7ABDFA4ADF390ULL, "w_encounter3_133_0.msh"},
	{0xD7E7B9DFA4AE0BAAULL, "w_encounter3_133_0.mmh"},
	{0xD7E7BDDFA4AE12E1ULL, "w_encounter3_133_0.mao"},
	{0xD7E7EC866BCE13D1ULL, "vxw_bonepit_168_0_0.vlm"},
	{0xD7E7F20DFDFF3A84ULL, "clf260_harley6069688.cl"},
	{0xD7E87F0E2FA578A0ULL, "6180454_m.fxe"},
	{0xD7E8F7461F2BF9DAULL, "6040011_m.fxe"},
	{0xD7E94E7FCA25B450ULL, "hlac_lt_0_lowtown_n_-196608_3z.dds"},
	{0xD7EA4D452EA161AFULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_highbench.phy"},
	{0xD7EA81510DEF8DF4ULL, "qun210cr_guard6072081.cl"},
	{0xD7EA971F089B28F8ULL, "591892383.ogg"},
	{0xD7EAE670489DB3E9ULL, "922431541.ogg"},
	{0xD7EB3672A142968BULL, "follower_banter6163273.cl"},
	{0xD7EB5271390D46C9ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdvista01.anb"},
	{0xD7EB68A862884ED1ULL, "gen00fl_isabela6090842.cl"},
	{0xD7EBB7D106CCA03EULL, "vxlt_gallowstemplar_f_5yy_0_0.vlm"},
	{0xD7EC4FA5944A6DB6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_run.evt"},
	{0xD7EC5FBAB82B907AULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_hed_hdra_0s.dds"},
	{0xD7EC72704C3E05DBULL, "700116127.ogg"},
	{0xD7EC897A04122F1BULL, "vxlt_0_lowtown_a3_n_c_cmn_0_0.vlm"},
	{0xD7EC8C90912E2D2BULL, "dre000_cut_sibling_dies6036296.cl"},
	{0xD7ECD4FCDC43296BULL, "gen00fl_isabela6211931.cl"},
	{0xD7ECE07B9C41CF65ULL, "98009739.ogg"},
	{0xD7ED31FFF99972CBULL, "gen00fl_isabela6088250.cl"},
	{0xD7ED5E1971B0D9B0ULL, "vxca_caverns2prc_7qg_0_0.vlm"},
	{0xD7EE28A5ED24CBC8ULL, "w_woundedcoast_124_0.mmh"},
	{0xD7EE34A5ED24E06BULL, "w_woundedcoast_124_0.mao"},
	{0xD7EE46A5ED24FEF2ULL, "w_woundedcoast_124_0.msh"},
	{0xD7EFAD0EF0DE940DULL, "6174755_m.fxe"},
	{0xD7F0464B7178C6ABULL, "dae000_elegant6166908.cl"},
	{0xD7F05AA6E31DFC2AULL, "vxlt_hangedman_n_2d0_0_0.vlm"},
	{0xD7F09C29A4141730ULL, "6118193_f.fxe"},
	{0xD7F152E2C952CE99ULL, "6162688_m.fxe"},
	{0xD7F16720EEE574ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_across.ani"},
	{0xD7F194248A9B339EULL, "w_sundermount_129_0.rml"},
	{0xD7F199ECF07BEA44ULL, "vxht_mansion_n_4cc_0_0.vlm"},
	{0xD7F1C3C7E7A0171CULL, "vxca_caverns1_mines_j3w_0_0.vlm"},
	{0xD7F1D6503A2C8738ULL, "6135814_m.fxe"},
	{0xD7F2792037EFAEF7ULL, "tal_tau_fl.mao"},
	{0xD7F288D6A7CAE4C2ULL, "895692687.ogg"},
	{0xD7F2BC36468CD369ULL, "6156179_m.fxe"},
	{0xD7F2E0D0F7023724ULL, "vxlt_0_lowtown_a2_n_daq_0_0.vlm"},
	{0xD7F338EFA5B6DF39ULL, "131830506.ogg"},
	{0xD7F34137892226D1ULL, "vxw_woundedcoast_75_0_0.vlm"},
	{0xD7F346A594507D41ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_run.gad"},
	{0xD7F37C449EE6E3F8ULL, "32586732.ogg"},
	{0xD7F3802745C1CDAAULL, "gen_im_cft_rec_psn_02.uti"},
	{0xD7F3C1EDCD85E28EULL, "6151029_m.fxe"},
	{0xD7F450E183BAAFF2ULL, "hlac_lt_0_lowtown_a3_n_c_65534_6z.dds"},
	{0xD7F46C1A434AD10AULL, "w_woundedcoast_60_0.mao"},
	{0xD7F4701A434AD859ULL, "w_woundedcoast_60_0.mmh"},
	{0xD7F47A1A434AE947ULL, "w_woundedcoast_60_0.msh"},
	{0xD7F4A6736B23E0EBULL, "hlac_lt_0_lowtown_131071_7z.dds"},
	{0xD7F56BE378D7844FULL, "ave000im_wesley_shield.uti"},
	{0xD7F59D3D56D2DCBBULL, "195255035.ogg"},
	{0xD7F5ACC1B94D70B6ULL, "vxht_mansion_n_3r5_0_0.vlm"},
	{0xD7F5D42CD70C6A71ULL, "hlac_lt_0_lowtown_a2_n_-65536_1z.dds"},
	{0xD7F63C3ED1E3C963ULL, "hlac_lt_undercitydungeon_-1_52y.dds"},
	{0xD7F65E2297C025EBULL, "dae000_bartrand6107550.cl"},
	{0xD7F66463A4F160A9ULL, "gen00fl_aveline6137494.cl"},
	{0xD7F6D9D437D86655ULL, "hlac_dr_dungeon_3217_0z.dds"},
	{0xD7F796D6675EC52DULL, "vxlt_0_lowtown_a2_n_f2g_0_0.vlm"},
	{0xD7F7C8E92B374D24ULL, "hla_ht_0_hightown_n_-65534_0y.dds"},
	{0xD7F7DBA822019177ULL, "t1_mul_te2.tnt"},
	{0xD7F7EBAD887AD82FULL, "215098011.ogg"},
	{0xD7F83BD012CFAF1FULL, "6074404_m.fxe"},
	{0xD7F87420EEEBA9EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_across.gad"},
	{0xD7F8B5938F06888CULL, "one200dg_solivitus6048196.cl"},
	{0xD7F8B7B373D15CBBULL, "hlac_dr_0_deeproads_5849_1x.dds"},
	{0xD7F8B9566F9CB06FULL, "6037557_m.fxe"},
	{0xD7F8CD17A5DEA295ULL, "vxht_chantry_n_2d9_0_0.vlm"},
	{0xD7F917A3F45AFD39ULL, "zz_approval_debug6001822.cl"},
	{0xD7F95A5F61F1C0D3ULL, "pro000_meeran6048810.cl"},
	{0xD7F96F05376DC08CULL, "6076539_m.fxe"},
	{0xD7FA47A594569DEFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_right_run.ani"},
	{0xD7FA7E99C6087C4FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_trapdoor_d.dds"},
	{0xD7FB592E3D02825FULL, "vxlt_0_lowtown_a2_e29_0_0.vlm"},
	{0xD7FB8EF9570EC6C0ULL, "6208485_f.fxe"},
	{0xD7FC4D595A6EC6BAULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclocker.phy"},
	{0xD7FC6A3185B66D50ULL, "fx_twist.mao"},
	{0xD7FCBBBDB3C3BB15ULL, "6110476_m.fxe"},
	{0xD7FCE800F8D3CAC2ULL, "qun111im_ketojan_amulet.uti"},
	{0xD7FD1DD517B56DDBULL, "hlac_lt_0_lowtown_a2_n_65536_6x.dds"},
	{0xD7FD30E68BFCD4BFULL, "6107481_m.fxe"},
	{0xD7FD423137467F49ULL, "vxlt_0_lowtown_a3_n_c_hx8_0_0.vlm"},
	{0xD7FD6BD848EA7E1CULL, "216474109.ogg"},
	{0xD7FDEC75F85D3A2FULL, "zz_followers_banter6157690.cl"},
	{0xD7FE47530AB9387EULL, "820971090.ogg"},
	{0xD7FE6AC39717E42AULL, "6020538_m.fxe"},
	{0xD7FEB2E0795C2DB5ULL, "art\\vfx\\combat\\vfx_gib_corm_lbrt.anb"},
	{0xD7FEB2E0795C2DBEULL, "art\\vfx\\combat\\vfx_gib_corm_lbrt.ani"},
	{0xD7FF08689623951DULL, "lt_0_lowtown_n_2_0.rml"},
	{0xD7FF3301CF156368ULL, "gen00fl_sebastian6189025.cl"},
	{0xD7FF6F20EEF1BFCDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_across.evt"},
	{0xD7FF987D1756AA76ULL, "vxca_caverns2_5ab_0_0.vlm"},
	{0xD7FFC09735608875ULL, "vxca_caverns2_7ar_0_1.vlm"},
	{0xD7FFC4BFBB48E719ULL, "6167588_m.fxe"},
	{0xD8000B11B2E68314ULL, "vxht_0_hightown_a3_4as_0_0.vlm"},
	{0xD8005167AA0AAE7FULL, "267988371.ogg"},
	{0xD800E380FF8CB433ULL, "6095522_m.fxe"},
	{0xD8016547BF75DB31ULL, "484227307.ogg"},
	{0xD8029AA415AA6068ULL, "ch1notnt_dy.psh.fxo"},
	{0xD802DF66814EA7EEULL, "gen_im_npc_dwf_arm_cht_cxt1.uti"},
	{0xD8031B4D6506884FULL, "vxca_cave_23i_0_0.vlm"},
	{0xD803271AC8BEE1A9ULL, "31919705.ogg"},
	{0xD8033C7C9CBC7329ULL, "308007748.ogg"},
	{0xD8034398511CC7DBULL, "45957490.ogg"},
	{0xD803B6EE6BCD8334ULL, "6078447_m.fxe"},
	{0xD803CCB2876E02A5ULL, "vxlt_foundry_n_11y_0_0.vlm"},
	{0xD803CE325ED33510ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowerb_1.mao"},
	{0xD803ED2B6DF6ED63ULL, "vfx_blt_lava_emitter.bnk"},
	{0xD804735C097FD4E0ULL, "vxca_caverns1_fo6_0_0.vlm"},
	{0xD804B6E5368317A1ULL, "vxht_chantry_n_2c0_0_0.vlm"},
	{0xD8050CF79771645EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball6.phy"},
	{0xD8056D3BD214E691ULL, "6067730_m.fxe"},
	{0xD80587FDEDC1075CULL, "weapon1_ds1y.psh.fxo"},
	{0xD805BA8439C71387ULL, "6164213_m.fxe"},
	{0xD805C2EDC7941FA1ULL, "static1_fast.psh.fxo"},
	{0xD805DAB50AF183A4ULL, "47019807.ogg"},
	{0xD80620E3AD8736EAULL, "6139711_m.fxe"},
	{0xD8063CE0C84F8394ULL, "art\\vfx\\combat\\vfx_g_spider_fb_aoe_p.mmh"},
	{0xD806921C8D81F8A9ULL, "dae100_bran6106038.cl"},
	{0xD80698888E6496E9ULL, "vxlt_backalley_n_1ji_0_2.vlm"},
	{0xD80768535EBD929EULL, "594372930.ogg"},
	{0xD807C13FA5FB2FE6ULL, "art\\vfx\\combat\\vfx_m_blood_charge_c.mmh"},
	{0xD807DB5EAF1A9561ULL, "vxht_0_hightown_a2_7fy_0_0.vlm"},
	{0xD808F0A4E0114710ULL, "qun110_petrice6041266.cl"},
	{0xD80954C5B474923FULL, "357606369.ogg"},
	{0xD80B1F6A63643D4DULL, "vxdr_0_deeproads_cwq_0_0.vlm"},
	{0xD80BAEF20F0F2D7DULL, "6155505_m.fxe"},
	{0xD80BBB0990C58042ULL, "727280925.ogg"},
	{0xD80BC6C48965CE3BULL, "vxw_encounter1_bp_0_0.vlm"},
	{0xD80C0C04985D9CA0ULL, "6174219_m.fxe"},
	{0xD80C333EC4ABD573ULL, "vxw_blightlands_wa_0_0.vlm"},
	{0xD80C5D4B25A245B6ULL, "follower_banter6163858.cl"},
	{0xD80C89ECF51FBE24ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkback.ani"},
	{0xD80CD83C14731DB4ULL, "hlac_lt_darktown_-1_26x.dds"},
	{0xD80CF239903BFE51ULL, "233863091.ogg"},
	{0xD80D48F4130C98BEULL, "6175399_m.fxe"},
	{0xD80DCF0D0A7880B7ULL, "hlac_lt_0_lowtown_a2_n_c_1_5y.dds"},
	{0xD80DE5379BE4A486ULL, "6076959_m.fxe"},
	{0xD80DF0C0FA5F8958ULL, "341519999.ogg"},
	{0xD80E81A7E78F34C5ULL, "vxht_0_hightown_bcw_0_0.vlm"},
	{0xD80FD4BCCD4E9170ULL, "vxdr_primevaltomb_2uq_0_0.vlm"},
	{0xD8109CC4CBDCE5A9ULL, "zz_and_debug6178593.cl"},
	{0xD810D4F0D8FE5D7DULL, "gen00fl_fenris6112047.cl"},
	{0xD810DD7DEE90AF75ULL, "hlac_ht_0_hightown_-131071_1y.dds"},
	{0xD8117AEAC19721C6ULL, "619628954.ogg"},
	{0xD81286ECF5242486ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkback.gad"},
	{0xD812880D9C4E1E43ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medh_0t.dds"},
	{0xD812A3994A35113EULL, "ave200_donnic6111715.cl"},
	{0xD812CBE0796D0138ULL, "art\\vfx\\combat\\vfx_gib_corm_lbrt.gad"},
	{0xD81301D0C5C605B5ULL, "mcr371_hostile_fenris6082250.cl"},
	{0xD813115F5B890BC2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock6a.msh"},
	{0xD813135F5B890F78ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock6a.mmh"},
	{0xD81316C79F67CD7CULL, "332305521.ogg"},
	{0xD8131F5F5B89231BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock6a.mao"},
	{0xD81351D4091A56ECULL, "gen00fl_dog6170982.cl"},
	{0xD8139A083A22AEEBULL, "hlac_w_bonepitmining_a2_131077_0x.dds"},
	{0xD814108E4325A1E9ULL, "gen00fl_fenris6110885.cl"},
	{0xD8142A371244AFE5ULL, "616343464.ogg"},
	{0xD8145026295ABE64ULL, "6118107_m.fxe"},
	{0xD814926CAEF66392ULL, "hla_ht_0_hightown_a3_-196608_0x.dds"},
	{0xD814BBB995B4EB74ULL, "6133672_m.fxe"},
	{0xD81539968866BE36ULL, "vxca_cavern1_haunted_eho_0_0.vlm"},
	{0xD815BBC82B73E9ADULL, "476601151.ogg"},
	{0xD815C8709A4376DAULL, "6149446_m.fxe"},
	{0xD815F05432BB884DULL, "vxht_chantry_2os_0_0.vlm"},
	{0xD815F759F90FD73DULL, "seb221_flora_harimann6177724.cl"},
	{0xD815FD2F8794FE4AULL, "llac_lt_0_lowtown_65537_3.dds"},
	{0xD815FD76C4B0C376ULL, "ca_caverns1_mines_24422.rml"},
	{0xD8171C93C769ACA8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_woodmetal_r.dds"},
	{0xD8171E003B195DEBULL, "vxdr_primevalentrance_1t2_0_0.vlm"},
	{0xD817686E74FCD60AULL, "vxca_caverns2_95v_0_0.vlm"},
	{0xD817710933E3B12BULL, "6060012_m.fxe"},
	{0xD8181085A95574C4ULL, "876021165.ogg"},
	{0xD81869A34384C531ULL, "hlac_lt_0_lowtown_a2_n_-65535_0x.dds"},
	{0xD8187E22F4B79D33ULL, "vxlt_0_lowtown_a3_n_c_exo_0_0.vlm"},
	{0xD8192FE71095BBB1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mb.msh"},
	{0xD8193CE0B8CBA371ULL, "407252601.ogg"},
	{0xD8193DE71095D3CCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mb.mao"},
	{0xD81949E71095E86FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarp16mb.mmh"},
	{0xD8196448B5C98BD2ULL, "6147116_m.fxe"},
	{0xD81981ECF52A3B65ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\walkback.evt"},
	{0xD8199147555A55CDULL, "vxdr_0_deeproads_clp_0_0.vlm"},
	{0xD81A38A17B5BE6BBULL, "vxw_encounter1_gs_0_0.vlm"},
	{0xD81A3A03A0434CA0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaricommander_overlay.anb"},
	{0xD81A7FEC1B27E76BULL, "vxw_sundermount_344_0_1.vlm"},
	{0xD81A9353E4862914ULL, "zz_area_signage_20.utp"},
	{0xD81B3048310A37BEULL, "gen00fl_aveline6101696.cl"},
	{0xD81B43F6F6CEC006ULL, "6131892_m.fxe"},
	{0xD81B79D5A8988648ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ff_basic_1.mao"},
	{0xD81BC4224946217DULL, "dae320st_merch_tailor.utm"},
	{0xD81C514E468E9408ULL, "hlac_ht_0_hightown_a2_-65537_4z.dds"},
	{0xD81C56A5CECDB2DFULL, "hlac_lt_gallowscourtyard_65534_1y.dds"},
	{0xD81D1A5F5B916F03ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock6a.phy"},
	{0xD81D3538D088BD92ULL, "6128858_m.fxe"},
	{0xD81D6D1A436DD84EULL, "w_woundedcoast_60_0.anb"},
	{0xD81D88248AC05A16ULL, "w_sundermount_129_0.mao"},
	{0xD81D94248AC06EB5ULL, "w_sundermount_129_0.mmh"},
	{0xD81D96248AC0722BULL, "w_sundermount_129_0.msh"},
	{0xD81E2FC898F89EB8ULL, "dae000_gen_turn_in_26178482.cl"},
	{0xD81F11A8346FB7F3ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\jagged_n.dds"},
	{0xD81F43B2922F1D40ULL, "vxlt_hangedman_n_1vf_0_1.vlm"},
	{0xD81F516F12699ABAULL, "320935840.ogg"},
	{0xD81F7D776369C737ULL, "hm_carver_h1s2.mor"},
	{0xD81FCC736AB030AAULL, "hlac_lt_gallowstemplar_f_7198_3y.dds"},
	{0xD81FDE997ADF6B90ULL, "hlac_w_encounter1_3_0z.dds"},
	{0xD82001DCBDB0843DULL, "vxdr_primevaltomb_5ca_0_0.vlm"},
	{0xD820234B2C70EAB9ULL, "6067297_m.fxe"},
	{0xD8208F03F8814495ULL, "vxlt_0_lowtown_a2_n_c_exh_0_0.vlm"},
	{0xD821729D6CC790C3ULL, "framebuffereffects_gaussblur.csh.hlsl_p1.csh.fxo.dx11"},
	{0xD821CF8968F72094ULL, "hlac_w_bonepitmining_a2_0_0x.dds"},
	{0xD821F93048CE7C81ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_2y.dds"},
	{0xD822018A54A5CDF0ULL, "vxht_playershouse_n_1zp_0_0.vlm"},
	{0xD82240C0AC91CEFCULL, "6050386_m.fxe"},
	{0xD822784DDE7332B6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v5.ani"},
	{0xD822BB3658A463DFULL, "llac_lt_0_lowtown_docks_a3_131070_0.dds"},
	{0xD822C2031B50D6E4ULL, "6117007_m.fxe"},
	{0xD822CD5696D1BE4CULL, "366781553.ogg"},
	{0xD82344D9C2620371ULL, "6169865_m.fxe"},
	{0xD823BA64F2A498D4ULL, "vxlt_0_lowtown_a3_n_c_hjg_0_0.vlm"},
	{0xD823F17945D80FA2ULL, "dae000_cora6162461.cl"},
	{0xD824316D4610566BULL, "636882772.ogg"},
	{0xD824769866350342ULL, "hlac_lt_0_lowtown_a2_n_-262144_0x.dds"},
	{0xD824CD7825F5E334ULL, "6161217_m.fxe"},
	{0xD824E6DBD10FD87CULL, "6176482_m.fxe"},
	{0xD824EA839BA782A2ULL, "6041737_m.fxe"},
	{0xD82520F6C419C30DULL, "393623014.ogg"},
	{0xD82547C656BC0718ULL, "vxht_0_hightown_a3_cel_0_0.vlm"},
	{0xD8258880A56EAD15ULL, "aip_hurlockemissary_trooper.gda"},
	{0xD825CD065057DE07ULL, "6141090_m.fxe"},
	{0xD82622CF91D2BE51ULL, "vxca_caverns1_e52_0_0.vlm"},
	{0xD826554A8C7FFFD6ULL, "llac_ht_mansion2b_818_39.dds"},
	{0xD82678B4C3414537ULL, "6127982_m.fxe"},
	{0xD8267E2F32FC3AE2ULL, "isa311_isabela6092504.cl"},
	{0xD82689E4C7BFF0D4ULL, "vxw_encounter2_r2_0_0.vlm"},
	{0xD82758C6DA3CB2E9ULL, "hlac_lt_undercitydungeon_-1_2z.dds"},
	{0xD827A0120D46C312ULL, "206380986.ogg"},
	{0xD827A2A2AB11AEEFULL, "491422214.ogg"},
	{0xD8289E261106A1D7ULL, "173695101.ogg"},
	{0xD82979BBB712A073ULL, "6084545_f.fxe"},
	{0xD82991BF51335661ULL, "vxlt_0_lowtown_n_cmo_0_0.vlm"},
	{0xD829E61E86EF4DFCULL, "gen00fl_merrill6151096.cl"},
	{0xD829EFC47306816FULL, "vxw_woundedcoast_1qx_0_0.vlm"},
	{0xD82A5FB6CE4F2D9EULL, "art\\levels\\races\\proxy\\props\\door\\prp_door_3.pwk"},
	{0xD82A74B1C1925250ULL, "223241385.ogg"},
	{0xD82AEE43E2EE5EBCULL, "6024121_m.fxe"},
	{0xD82AF0292A8CC3E6ULL, "dre000_cut_sibling_dies6036314.cl"},
	{0xD82B1396A13DE7EAULL, "6079056_f.fxe"},
	{0xD82B3C0C475EE253ULL, "6169580_m.fxe"},
	{0xD82B6561CB7B65C5ULL, "vxht_0_hightown_biu_0_0.vlm"},
	{0xD82BBEE4217C0676ULL, "vxlt_0_lowtown_docks_9ra_0_0.vlm"},
	{0xD82BE21340D6CD4DULL, "455798696.ogg"},
	{0xD82C36ACEC832C3CULL, "vxlt_shop_dx_0_0.vlm"},
	{0xD82CE028B1B17058ULL, "6109699_m.fxe"},
	{0xD82DAA6FA35FD23AULL, "483531042.ogg"},
	{0xD82DC4ED61696B9BULL, "hlac_ca_caverns1_mines_24427_2z.dds"},
	{0xD82DD21BD281118BULL, "6083939_f.fxe"},
	{0xD82E0F3D2B97B71AULL, "272318744.ogg"},
	{0xD82E2B016FA61448ULL, "mag370_meredith6092175.cl"},
	{0xD82E86263F50C8E4ULL, "art\\vfx\\models\\vfx_gui_mov_point.msh"},
	{0xD82E9C263F50ED86ULL, "art\\vfx\\models\\vfx_gui_mov_point.mmh"},
	{0xD82F1C7488C4D154ULL, "33150266.ogg"},
	{0xD82F3FE0C8728E4BULL, "art\\vfx\\combat\\vfx_g_spider_fb_aoe_p.anb"},
	{0xD82F5692296D2578ULL, "6178527_m.fxe"},
	{0xD82F71D0BA7D27C4ULL, "92801497.ogg"},
	{0xD82F9DA1CDDE2505ULL, "mag360_samson6042030.cl"},
	{0xD82FABCB683BEDABULL, "623467263.ogg"},
	{0xD82FF54711CF1E7CULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_f_c.impa.ani"},
	{0xD8302082EBC2877BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_exit.evt"},
	{0xD830538D000B9C4BULL, "832760537.ogg"},
	{0xD83057F4BA1ED37BULL, "659119338.ogg"},
	{0xD8307367469BE670ULL, "vxht_playershouse_n_1qq_0_0.vlm"},
	{0xD830804DDE7F7DA7ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v5.evt"},
	{0xD830FC152501FE7EULL, "769932691.ogg"},
	{0xD830FC2AEE7C762BULL, "6032060_m.fxe"},
	{0xD8312110FAAFE51BULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_foam02.dds"},
	{0xD83143DC87677834ULL, "hlac_lt_0_lowtown_a2_-65536_7z.dds"},
	{0xD8316097CD9E053DULL, "vxdr_primevaltomb_878_0_0.vlm"},
	{0xD83162D02C1E02EBULL, "418937535.ogg"},
	{0xD8317CE810F12CE7ULL, "mag291_demon_marethari6028859.cl"},
	{0xD832A1D440721182ULL, "mag330_walter6044825.cl"},
	{0xD832D40E16A7B94DULL, "vxht_chantry_2eq_0_0.vlm"},
	{0xD832E7CC00369133ULL, "vxlt_0_lowtown_gjs_0_0.vlm"},
	{0xD833006CBD05FA89ULL, "6188590_m.fxe"},
	{0xD8331A7512880248ULL, "vxlt_0_lowtown_n_cmj_0_0.vlm"},
	{0xD8333F5517EBFBB4ULL, "6166317_m.fxe"},
	{0xD833426792D95CB5ULL, "6070792_m.fxe"},
	{0xD833F13C00D55CEBULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.impa.ani"},
	{0xD83471D67C66677AULL, "vxlt_gallowstemplar_5qv_0_0.vlm"},
	{0xD83508A09353F1B5ULL, "docks_male_elven_worker_sta6176110.cl"},
	{0xD83538A7D82E2282ULL, "6076256_m.fxe"},
	{0xD83595F2EA0A00D8ULL, "260206714.ogg"},
	{0xD835F7638EA3C3A1ULL, "vxca_cave_18a_0_0.vlm"},
	{0xD83683938CE8A426ULL, "vrc221_varric6065957.cl"},
	{0xD836916AC9853400ULL, "docks_male_elven_worker_sta6175794.cl"},
	{0xD83695DDD0A4B167ULL, "vxdr_primevaltomb_7y7_0_0.vlm"},
	{0xD836A14DDE8421F0ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\idle\\twitch_v5.gad"},
	{0xD837024711D552BEULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_f_c.impa.gad"},
	{0xD8371982EBC899BCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_exit.gad"},
	{0xD837657443695D74ULL, "vxca_caverns1_gh2_0_0.vlm"},
	{0xD8378AA95266CDEDULL, "hlac_w_encounter2_196609_0z.dds"},
	{0xD837B91557CCBDAEULL, "character_character_p92.psh.fxo.dx11"},
	{0xD8386D493988C3D4ULL, "6092167_f.fxe"},
	{0xD8389FA39A9D6E90ULL, "571302864.ogg"},
	{0xD838CF9E1064B34BULL, "fen110_anso6094979.cl"},
	{0xD838D0CE8020D74BULL, "vxht_keep_61s_0_0.vlm"},
	{0xD8393FEAD742E75EULL, "qun110_petrice6042243.cl"},
	{0xD839793048DD592FULL, "vxw_woundedcoast_hv_0_0.vlm"},
	{0xD83AB9A0D4913E64ULL, "mag101cr_grace.utc"},
	{0xD83B5205A3A788D8ULL, "6102799_m.fxe"},
	{0xD83BC6D652AEC1BAULL, "hlac_lt_backalley_g_65537_0x.dds"},
	{0xD83C34B0BE7AF5B4ULL, "vxlt_0_lowtown_a2_n_c_hkq_0_0.vlm"},
	{0xD83CB5F6C2807A33ULL, "mrl251ip_bosshorde_d.utp"},
	{0xD83D105CA6E44130ULL, "hlac_ca_caverns1_24422_2y.dds"},
	{0xD83D18440A5060A8ULL, "vxlt_0_lowtown_docks_c14_0_0.vlm"},
	{0xD83D18D4F1A4A0A1ULL, "6064924_m.fxe"},
	{0xD83D1994443A6762ULL, "vxlt_0_lowtown_ffg_0_0.vlm"},
	{0xD83D1C003F9F7C04ULL, "follower_banter6161142.cl"},
	{0xD83D9D4814B60877ULL, "vxw_encounter1_bd_0_0.vlm"},
	{0xD83DDD4711DB32DDULL, "art\\vfx\\combat\\vfx_m_sta_ice_close_f_c.impa.evt"},
	{0xD83E1882EBCEB6BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_exit.ani"},
	{0xD83EC2836576E402ULL, "vxw_woundedcoast_y4_0_0.vlm"},
	{0xD83F99C3F242CDFFULL, "spark.mao"},
	{0xD83FAFB8EE60CDF6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_sway_big.ani"},
	{0xD83FC5F80A6DBD3CULL, "vxlt_gallowscourtyard_73f_0_0.vlm"},
	{0xD83FF07FBC91F564ULL, "442231479.ogg"},
	{0xD840D93C00DFBE62ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.impa.evt"},
	{0xD840DC7EEE76164DULL, "381547830.ogg"},
	{0xD840E2506A61C901ULL, "6177581_m.fxe"},
	{0xD84180EA1FE9B200ULL, "zz_tbe100_barber.cnv"},
	{0xD841999C94F9DB62ULL, "vxw_encounter1_e4_0_0.vlm"},
	{0xD841E126A9A1E992ULL, "968554182.ogg"},
	{0xD84224F1D4DF4059ULL, "vxw_encounter3_g7_0_0.vlm"},
	{0xD8429E4756FB8755ULL, "vxw_bonepitdestroyed_a3_78_0_0.vlm"},
	{0xD842AF1D6B761B79ULL, "follower_banter6143273.cl"},
	{0xD8430A051596BF5FULL, "gen00fl_aveline6114293.cl"},
	{0xD84313EDDE89B025ULL, "vxdr_primevalentrance_24w_0_0.vlm"},
	{0xD8435C6EA72D7637ULL, "6129782_m.fxe"},
	{0xD843910583A710DDULL, "671523407.ogg"},
	{0xD844111A31A4F4A5ULL, "hlac_dr_slavers_3207_9z.dds"},
	{0xD84447CF1F2D0C9DULL, "w_woundedcoast_179_0.mao"},
	{0xD8444BCF1F2D13D6ULL, "w_woundedcoast_179_0.mmh"},
	{0xD84455CF1F2D24D4ULL, "w_woundedcoast_179_0.msh"},
	{0xD844896AC3B61AA6ULL, "643753767.ogg"},
	{0xD844C498079AB0BEULL, "6036425_m.fxe"},
	{0xD8451DD3553654FEULL, "6044972_m.fxe"},
	{0xD845490A5C1F9773ULL, "cs_mrl150cs_merrill_intro.bnk"},
	{0xD84593136E309065ULL, "108733425.ogg"},
	{0xD84596A752D9EF4EULL, "hlac_lt_0_lowtown_a2_n_-262143_0y.dds"},
	{0xD8463404DB7DBA95ULL, "268210925.ogg"},
	{0xD8464AB18BA1930CULL, "vxht_mansion2_2kr_0_0.vlm"},
	{0xD846FD33DC311F78ULL, "w_sundermount_130_0.phy"},
	{0xD8470876FE13DDC7ULL, "710565272.ogg"},
	{0xD847590A3A87793CULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_exit.ani"},
	{0xD847A5CAB3DA7574ULL, "437959133.ogg"},
	{0xD847D83C00E5DB75ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.impa.gad"},
	{0xD8481FC62B4B1353ULL, "6163780_m.fxe"},
	{0xD8484F2241C8B668ULL, "6142845_m.fxe"},
	{0xD848701D0CF9EEECULL, "vxca_caverns1_mines_ile_0_0.vlm"},
	{0xD848740D6E493961ULL, "6128025_f.fxe"},
	{0xD848B209E28DCEDBULL, "hlac_lt_0_lowtown_docks_-65539_5y.dds"},
	{0xD8494820DBA4507FULL, "hlac_lt_0_lowtown_docks_-65540_2x.dds"},
	{0xD849EDE82D249E3FULL, "lt_0_lowtown_docks_a3_n_17_0_ll.mmh"},
	{0xD84A1A78F1619B27ULL, "pro000st_nina_intro.stg"},
	{0xD84A36CC06CF7AB2ULL, "vxht_0_hightown_8q0_0_0.vlm"},
	{0xD84A4F805E581791ULL, "6173686_m.fxe"},
	{0xD84A8D036BFD8997ULL, "vxlt_0_lowtown_docks_n_b58_0_1.vlm"},
	{0xD84ABC0E2A26F78BULL, "256116426.ogg"},
	{0xD84BE49ACE638E15ULL, "vxlt_0_lowtown_a2_n_c_ewu_0_0.vlm"},
	{0xD84C0F2506AC20F7ULL, "vxw_encounter1_gg_0_0.vlm"},
	{0xD84C5FD79111D5BEULL, "984895092.ogg"},
	{0xD84CA6A880837768ULL, "follower_banter6163993.cl"},
	{0xD84CB7B8EE6B65E7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_sway_big.evt"},
	{0xD84CFC33DC358890ULL, "w_sundermount_130_0.rml"},
	{0xD84D3460576CFC82ULL, "vxw_woundedcoast_v5_0_0.vlm"},
	{0xD84D4B110E2C1E20ULL, "6060103_m.fxe"},
	{0xD84D8D3ECD8E3E81ULL, "995156099.ogg"},
	{0xD84DA3141867C6EFULL, "380634553.ogg"},
	{0xD84DDFE4D268A72BULL, "6038660_m.fxe"},
	{0xD84E0626F11FBA64ULL, "art\\characters\\creatures\\mubari\\animation\\brk_lp.ani"},
	{0xD84E0A66FF2B0611ULL, "zz_vault_debug6203326.cl"},
	{0xD84E105B16FCF622ULL, "vxlt_0_lowtown_docks_n_9fj_0_0.vlm"},
	{0xD84E37D4FC63D5BCULL, "vxlt_gallowscourtyard_74g_0_0.vlm"},
	{0xD84E58902D61973AULL, "vxht_mansion_n_62p_0_0.vlm"},
	{0xD84E660A3A8DAE7EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_exit.gad"},
	{0xD84EAA2C708A382EULL, "6147534_m.fxe"},
	{0xD84EF1AE3BEB8409ULL, "hlac_ht_0_hightown_n_65535_0y.dds"},
	{0xD84F145B118908B7ULL, "221980694.ogg"},
	{0xD84F4D0438A2D399ULL, "dre000_dougal6129821.cl"},
	{0xD84FB7D5F86437D7ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_varb.phy"},
	{0xD84FCD6DADA0EBEAULL, "288838223.ogg"},
	{0xD84FF74C0B5DB0C2ULL, "vxlt_gallowstemplar_3ra_0_0.vlm"},
	{0xD8505775F9B4B8CAULL, "mag160_feynriel6020395.cl"},
	{0xD850701A439936B2ULL, "w_woundedcoast_60_0.rml"},
	{0xD8512B8954CD6B99ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_vase_sa.dds"},
	{0xD8513799E52C0792ULL, "6162948_m.fxe"},
	{0xD8517002AB38F925ULL, "hlac_lt_0_lowtown_a2_196607_1z.dds"},
	{0xD85198BA8C7EE27FULL, "426806139.ogg"},
	{0xD851E8AFBA38EC65ULL, "6064977_m.fxe"},
	{0xD851F35F42F2D19AULL, "m_qun111ar_safehouse.bnk"},
	{0xD85242C40A35FC25ULL, "6167023_m.fxe"},
	{0xD8527B380C2022D2ULL, "gen00fl_varric6072964.cl"},
	{0xD8528EB53E1A020DULL, "hlac_lt_0_lowtown_a2_n_c_1_4x.dds"},
	{0xD852CCAACC42019BULL, "mcr310_argument6072679.cl"},
	{0xD852E9AD20DD2DBEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\trap_pressureston_0.pwk"},
	{0xD85383E101A6310CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_painting_sa.dds"},
	{0xD85394D619407D1CULL, "vxht_mansion2_31m_0_0.vlm"},
	{0xD853D8B8EE71BC30ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_sway_big.gad"},
	{0xD85405DD67086B82ULL, "llac_ht_mansion2_818_60.dds"},
	{0xD854D952F3E186F4ULL, "hlac_lt_0_lowtown_a2_-65537_4x.dds"},
	{0xD8550326F125D3C6ULL, "art\\characters\\creatures\\mubari\\animation\\brk_lp.gad"},
	{0xD8553B0F32906A2AULL, "genip_cin_table_leg.utp"},
	{0xD855410A3A938D9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\crouch_squat_exit.evt"},
	{0xD855CF9FF8E03312ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_l_2.gad"},
	{0xD8565CEA2CC98A2FULL, "vxw_woundedcoast_bp_0_0.vlm"},
	{0xD85665A1D29C3AA6ULL, "877382191.ogg"},
	{0xD856A3F6AC3D65A1ULL, "6036776_m.fxe"},
	{0xD856BAF8A15DAF45ULL, "vxht_0_hightown_bxr_0_0.vlm"},
	{0xD8570C785E9842BCULL, "seb000pt_exiled_prince_.ncs"},
	{0xD85719BCB27B966DULL, "gen00fl_isabela_init6076591.cl"},
	{0xD85720B07C2FC946ULL, "hlac_ca_caverns1_24424_24y.dds"},
	{0xD857C4516946285AULL, "6111611_f.fxe"},
	{0xD857C5ABA49E2596ULL, "vxlt_0_lowtown_docks_n_be5_0_1.vlm"},
	{0xD857E900DA968238ULL, "471091051.ogg"},
	{0xD857F913B52FE086ULL, "llac_lt_0_lowtown_docks_n_-4_2.dds"},
	{0xD8583DD06C909133ULL, "826467453.ogg"},
	{0xD85846595ABD19E3ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclocker.msh"},
	{0xD85854595ABD313DULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclocker.mmh"},
	{0xD85858595ABD384EULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclocker.mao"},
	{0xD85874CE26D31F1FULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_boda.msh"},
	{0xD8587ACE26D329C1ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_boda.mmh"},
	{0xD8588FB940438BDDULL, "6149434_m.fxe"},
	{0xD858DC8BA27C7E90ULL, "lt_0_lowtown_docks_n_132_0.mao"},
	{0xD858E08BA27C85A3ULL, "lt_0_lowtown_docks_n_132_0.mmh"},
	{0xD858EE8BA27C9DFDULL, "lt_0_lowtown_docks_n_132_0.msh"},
	{0xD859DE48F1D5DA7BULL, "gen00fl_dog6165747.cl"},
	{0xD85A1484FE280B93ULL, "lgt000ip_old_locked_crate.utp"},
	{0xD85A948652529C99ULL, "gen00fl_fenris6126453.cl"},
	{0xD85AA42FCFED5003ULL, "art\\vfx\\models\\vfx_target_cone.mao"},
	{0xD85AA82FCFED5750ULL, "art\\vfx\\models\\vfx_target_cone.mmh"},
	{0xD85AB62FCFED6F6AULL, "art\\vfx\\models\\vfx_target_cone.msh"},
	{0xD85AD82DB48A2F29ULL, "522211286.ogg"},
	{0xD85B158A5A483CF8ULL, "vxca_caverns1_fmt_0_0.vlm"},
	{0xD85B42034325C539ULL, "549574153.ogg"},
	{0xD85B5858438744ACULL, "vxlt_0_lowtown_docks_n_d2y_0_0.vlm"},
	{0xD85BECDC5E1C5107ULL, "art\\characters\\creatures\\desiredemon\\animation\\rangedaoe.gad"},
	{0xD85BFE26F12BE9A5ULL, "art\\characters\\creatures\\mubari\\animation\\brk_lp.evt"},
	{0xD85C3FE35581E1F0ULL, "6072333_m.fxe"},
	{0xD85C7591D17C28F6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2bwin.msh"},
	{0xD85C7F91D17C39F4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2bwin.mmh"},
	{0xD85C8391D17C4007ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2bwin.mao"},
	{0xD85CBEC9DCE817F2ULL, "spellconjuringspeeds.gda"},
	{0xD85CE6BC24309C3DULL, "6043181_m.fxe"},
	{0xD85D11F342B83D74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_repulsion_wave.ani"},
	{0xD85D6DC34EDADBCEULL, "vxca_cavern1_haunted_j4d_0_0.vlm"},
	{0xD85E3A3C3C41A5D1ULL, "qcr000_arishok6060003.cl"},
	{0xD85E3E472CDFA286ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_023a_0d.dds"},
	{0xD85E984045868E02ULL, "gen00fl_dog6167027.cl"},
	{0xD85EC4E56BC3739CULL, "hlac_lt_0_lowtown_n_-196609_4x.dds"},
	{0xD85EE075C72E80BCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala3a_1.mao"},
	{0xD85F5EC2342D8A64ULL, "gen00fl_sebastian6177305.cl"},
	{0xD86085C78916AA75ULL, "hlac_lt_0_lowtown_a2_-1_8x.dds"},
	{0xD860ADC3F069254DULL, "dae000_merchant_armor_hm.bnk"},
	{0xD86105452C1A2696ULL, "hlac_lt_0_lowtown_a2_n_-196609_2z.dds"},
	{0xD86125ED92E75F2BULL, "6123641_m.fxe"},
	{0xD86157308508B4B8ULL, "zz_approval_debug6003522.cl"},
	{0xD861B392EE481B28ULL, "6080996_m.fxe"},
	{0xD86250E289CD6079ULL, "gen00fl_isabela_init6076490.cl"},
	{0xD8625D4ADFD1A46EULL, "6147862_m.fxe"},
	{0xD862A17F79F5BB39ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutcn.msh"},
	{0xD862AB7F79F5CC27ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutcn.mmh"},
	{0xD862AF7F79F5D374ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarpoutcn.mao"},
	{0xD862D41B61E93B10ULL, "llac_lt_0_lowtown_a2_65537_7.dds"},
	{0xD862E1DC5E225D14ULL, "art\\characters\\creatures\\desiredemon\\animation\\rangedaoe.evt"},
	{0xD862E84306F81BAAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_rogue.gad"},
	{0xD8630EF342BCA3D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_repulsion_wave.gad"},
	{0xD8632DD64C0F2F55ULL, "vfx_cone_lightning_dot_01.mao"},
	{0xD863CFC3F06B94B5ULL, "dae000_merchant_armor_hm.cnv"},
	{0xD864561D5883CBFDULL, "dae150_harshal6148851.cl"},
	{0xD864F26582AD568CULL, "llac_w_blightlands_262146_0.dds"},
	{0xD8652C2682004356ULL, "vxht_mansion2_331_0_0.vlm"},
	{0xD8658E351774E180ULL, "568965108.ogg"},
	{0xD865B56410AF9F80ULL, "lt_0_lowtown_a2_25_0_lc.mmh"},
	{0xD866036AD87A16D9ULL, "hlac_lt_0_lowtown_a2_n_65538_2x.dds"},
	{0xD8668BE047E5F511ULL, "mcr370_negotiation6095519.cl"},
	{0xD866F3061A2734CAULL, "vxca_cavern1_haunted_i9l_0_0.vlm"},
	{0xD867D2DC4558CC16ULL, "vxlt_0_lowtown_a2_n_gof_0_0.vlm"},
	{0xD868F883F831272AULL, "773163779.ogg"},
	{0xD8690519E6CDBCBFULL, "6199330_m.fxe"},
	{0xD86905DA4A2C03BDULL, "and111_lirene6173808.cl"},
	{0xD8698BF1F805C520ULL, "934517151.ogg"},
	{0xD8699B09ACE596D7ULL, "128905905.ogg"},
	{0xD869B32C76715480ULL, "500983216.ogg"},
	{0xD869CB4306FE0941ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_rogue.evt"},
	{0xD869E18BA28AFAC8ULL, "lt_0_lowtown_docks_n_132_0.phy"},
	{0xD86A16247D8E6C92ULL, "dae000cr_grand_cleric6124712.cl"},
	{0xD86A29F342C2F035ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_repulsion_wave.evt"},
	{0xD86A5A7AF6C3C298ULL, "llac_lt_0_lowtown_-65535_2.dds"},
	{0xD86AB25DA9782BC7ULL, "mrl150cr_dalish_guard2.utc"},
	{0xD86ABA9FF8F26BB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_l_2.ani"},
	{0xD86B8AC35B3C0C96ULL, "gen_im_wep_war_one_act2_01.uti"},
	{0xD86BBF7F1D6F411AULL, "vxca_caverns2prc_9fj_0_0.vlm"},
	{0xD86BCDEE8C71165DULL, "vxlt_0_lowtown_a2_n_c_ffu_0_0.vlm"},
	{0xD86BE896B6D8CDD6ULL, "82489067.ogg"},
	{0xD86BE89AFA5086BBULL, "store_dae100_bodahn.utm"},
	{0xD86D60272525E7F4ULL, "370552770.ogg"},
	{0xD86D969200324F7FULL, "6777507.ogg"},
	{0xD86DB594E4249AD0ULL, "6072711_m.fxe"},
	{0xD86DD3C49BE654B1ULL, "6103195_m.fxe"},
	{0xD86EC7DA7179BEECULL, "seb201_lady_harimann6177431.cl"},
	{0xD86EE37ED3A7B762ULL, "hlac_lt_undercitydungeon_-1_47z.dds"},
	{0xD86EE5CA39C903CDULL, "357473066.ogg"},
	{0xD86EFC73D56EDA5CULL, "6149105_m.fxe"},
	{0xD86F4DEC3990BD0BULL, "gen00fl_anders6112924.cl"},
	{0xD86F91249CD9D062ULL, "6169911_f.fxe"},
	{0xD86F97CA84DD77ABULL, "6128741_m.fxe"},
	{0xD86FEEB45866EC5DULL, "vxw_bonepit_11i_0_0.vlm"},
	{0xD87030645F64DB4CULL, "397104949.ogg"},
	{0xD87058C21BD7D978ULL, "t3_rob_rev.tnt"},
	{0xD8707B4D43A32ADDULL, "hlac_ca_caverns1_mines_24423_11x.dds"},
	{0xD8708220819B4E99ULL, "6067658_f.fxe"},
	{0xD870B67D1DD8233DULL, "169582327.ogg"},
	{0xD870CF27D1C7BCEEULL, "6140962_m.fxe"},
	{0xD870D9DC5E2E8D4DULL, "art\\characters\\creatures\\desiredemon\\animation\\rangedaoe.ani"},
	{0xD871472F178F6DA5ULL, "vxht_0_hightown_n_a4r_0_0.vlm"},
	{0xD8725EF5BD617F9EULL, "gen00fl_dog6166486.cl"},
	{0xD8727E27CDFADE64ULL, "26503949.ogg"},
	{0xD872CD5676324DC4ULL, "prp_str_mirrorstatue_01n.dds"},
	{0xD872E5A102A0B1BBULL, "lgt310_mis_burn_rival_thug6152772.cl"},
	{0xD87340A410B5BB2FULL, "vxw_woundedcoast_tr_0_0.vlm"},
	{0xD8736D1A212690FFULL, "w_blightlands_104_0.phy"},
	{0xD873A832AC16E8BBULL, "vxw_bonepitdestroyed_a3_23b_0_0.vlm"},
	{0xD873B28196F6A92CULL, "dae000ip_crowd_mf8_talk4.utp"},
	{0xD873FB99A13F6C4AULL, "vxw_bonepitdestroyed_a3_1zk_0_0.vlm"},
	{0xD8741997D0AA08F5ULL, "400137185.ogg"},
	{0xD8742BAE832962DFULL, "lgt110_orz_harrowmont6108993.cl"},
	{0xD874C1E15E63522CULL, "vxlt_0_lowtown_a2_gck_0_0.vlm"},
	{0xD87509A11536D717ULL, "vxlt_gallowscourtyard_6wd_0_0.vlm"},
	{0xD8754E909A33DA04ULL, "6095921_m.fxe"},
	{0xD87556E9279C8C22ULL, "6174158_m.fxe"},
	{0xD8757F2696865556ULL, "6057338_m.fxe"},
	{0xD875A75C30D04255ULL, "98454089.ogg"},
	{0xD875A8B05838DB69ULL, "6129920_m.fxe"},
	{0xD875FC24865357B4ULL, "vxlt_0_lowtown_a2_n_c_hwu_0_0.vlm"},
	{0xD87613218B91B891ULL, "vxw_encounter2_89_0_0.vlm"},
	{0xD8761578A4634895ULL, "463008637.ogg"},
	{0xD8763C346049AFF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_loop.gad"},
	{0xD876C34307088590ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arishok_kill_rogue.ani"},
	{0xD8770D5F43121A38ULL, "m_qun111ar_safehouse.wbd"},
	{0xD87749AB5A15354CULL, "art\\characters\\weapons\\shields\\w_shd_amell.msh"},
	{0xD8774FAB5A153F3EULL, "art\\characters\\weapons\\shields\\w_shd_amell.mmh"},
	{0xD8775BAB5A1553D5ULL, "art\\characters\\weapons\\shields\\w_shd_amell.mao"},
	{0xD87863422395DF22ULL, "lgt000cr_hate_dalish.utc"},
	{0xD8787D180FB46A78ULL, "624956475.ogg"},
	{0xD879088A984124F1ULL, "vxht_mansion_n_3s9_0_0.vlm"},
	{0xD87964F6FCFC9E93ULL, "vxlt_0_lowtown_a3_n_c_gvc_0_0.vlm"},
	{0xD879F690A0878F3EULL, "hlac_ht_brothel_n_3371_0y.dds"},
	{0xD87A35CA48FD648AULL, "6128844_f.fxe"},
	{0xD87A5935F2F1F03BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_stone.mmh"},
	{0xD87A6535F2F204D8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_stone.mao"},
	{0xD87A7735F2F22325ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_stone.msh"},
	{0xD87ABD0EAEC31E80ULL, "vxht_0_hightown_n_aqh_0_0.vlm"},
	{0xD87B1C72EB6FBBA9ULL, "one300_zevran6089705.cl"},
	{0xD87B32099D17DA06ULL, "598366210.ogg"},
	{0xD87B4BCC61FF9A34ULL, "vxlt_hangedman_n_26d_0_0.vlm"},
	{0xD87C1324C88AA9CDULL, "hlac_lt_gallowscourtyard_-2_0x.dds"},
	{0xD87D1F52D9162EE5ULL, "one100dg_kelder6056891.cl"},
	{0xD87D77377EAE43E1ULL, "hlac_lt_0_lowtown_a2_131072_4z.dds"},
	{0xD87D94894D9E7546ULL, "415729314.ogg"},
	{0xD87DFCD66BE085EDULL, "and_templar_2.utc"},
	{0xD87E0152C1A8A326ULL, "vxdr_dungeon_1mb_0_0.vlm"},
	{0xD87E9B9F292C7015ULL, "mcr371_hostile_anders6082124.cl"},
	{0xD87EA9734B8BB14CULL, "hlac_ht_keep_3732_8x.dds"},
	{0xD87EC80FD0E6B775ULL, "6019087_m.fxe"},
	{0xD87EEF609A3B743FULL, "qun120cr_javaris6042837.cl"},
	{0xD87F091CC7B22C3BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree01_sa.dds"},
	{0xD87F2C35414ECF23ULL, "gen00fl_aveline6137588.cl"},
	{0xD87F4D69B9504BC9ULL, "vxht_brothel_n_6ay_0_0.vlm"},
	{0xD87FC727D6FB39D2ULL, "gen_im_trs_crc_amu_r0.uti"},
	{0xD88013EA3CC438FBULL, "gen_im_trs_dal_med_boo_r2.uti"},
	{0xD880B1A4E97E4BE2ULL, "combat_2h_r2_combo.gda"},
	{0xD880F469C67C6D99ULL, "1064450942.ogg"},
	{0xD881AFC563559152ULL, "6119384_m.fxe"},
	{0xD881DCCF10E9C6EAULL, "gen00fl_varric6126324.cl"},
	{0xD882057C3E247113ULL, "6172113_m.fxe"},
	{0xD8824B3170738A53ULL, "zz_followers_banter6158637.cl"},
	{0xD8828525CE8F51A9ULL, "vxlt_0_lowtown_a3_n_c_feu_0_0.vlm"},
	{0xD882AC8876EB34A7ULL, "356679177.ogg"},
	{0xD8834B9681EA0086ULL, "dae000_flemeth6013361.cl"},
	{0xD8837CFC8ADC5B90ULL, "691201463.ogg"},
	{0xD8837FC466A2385CULL, "6125395_f.fxe"},
	{0xD883E0B28CF77D08ULL, "cod_ltr_ggt_note3.utp"},
	{0xD883EFA71E58108CULL, "mag360_thrask6043564.cl"},
	{0xD8846E1A21350504ULL, "w_blightlands_104_0.mmh"},
	{0xD884721A21350C57ULL, "w_blightlands_104_0.mao"},
	{0xD8847BCE26F8663EULL, "art\\characters\\playercharacter\\dwarfmale\\mt_cbt_boda.phy"},
	{0xD884841A21352AA6ULL, "w_blightlands_104_0.msh"},
	{0xD884D5B1F65701BAULL, "gen_im_npc_wep_sta_bet_fake.uti"},
	{0xD8852474364222DEULL, "vxht_bartrand_n_5y9_0_0.vlm"},
	{0xD88527AF25F319CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_windowxa.msh"},
	{0xD8852DAF25F323BEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_windowxa.mmh"},
	{0xD88539AF25F33855ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_windowxa.mao"},
	{0xD8855957FAD21C71ULL, "366952226.ogg"},
	{0xD8856E492C3B091DULL, "dae011_gossips6149763.cl"},
	{0xD8857C715978EED1ULL, "vxw_bonepitmining_a2_27b_0_0.vlm"},
	{0xD885B1B0698C7B10ULL, "hlac_lt_0_lowtown_n_131070_0x.dds"},
	{0xD8867028F601A3A3ULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.cess.ani"},
	{0xD886AC4C69C9BB38ULL, "470024641.ogg"},
	{0xD886C2310EFE0C98ULL, "6108397_m.fxe"},
	{0xD886D006C87DEFC5ULL, "6019022_f.fxe"},
	{0xD8870103E6A4E059ULL, "154541146.ogg"},
	{0xD88762EBEEAC2B63ULL, "6072671_m.fxe"},
	{0xD887835503BFAE12ULL, "6043282_f.fxe"},
	{0xD887BF9F4DF7911BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.evt"},
	{0xD888ECA926C74436ULL, "ht_bartrand_n.arl"},
	{0xD888EDEE19C58CEDULL, "vxht_mansion_n_5ae_0_0.vlm"},
	{0xD888F38AFE159272ULL, "hlac_lt_0_lowtown_docks_a3_-65537_3y.dds"},
	{0xD8896D1BD6AFEE85ULL, "vxca_caverns1_fmm_0_0.vlm"},
	{0xD88B01F1052EDA35ULL, "vxca_caverns1_mines_dpy_0_0.vlm"},
	{0xD88B2734605BE798ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lying_injured_loop.ani"},
	{0xD88C7FE9FE63F62DULL, "zz_dae_debug6029292.cl"},
	{0xD88CABB44AA538C0ULL, "vxlt_0_lowtown_a2_n_c_hl0_0_0.vlm"},
	{0xD88CB976CED0CBE6ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_005a_0d.dds"},
	{0xD88CE61417A5236EULL, "454854862.ogg"},
	{0xD88CE7370A26B1EFULL, "qun120cr_javaris6042727.cl"},
	{0xD88D1952844A8FEAULL, "745368106.ogg"},
	{0xD88D670A5C5D7FD1ULL, "cs_mrl150cs_merrill_intro.wbd"},
	{0xD88DB48D28B39B82ULL, "214069983.ogg"},
	{0xD88DBF46AB466449ULL, "vxht_brothel_n_6jx_0_0.vlm"},
	{0xD88E077AC5716F22ULL, "6002585_m.fxe"},
	{0xD88E0BA391E8B2AFULL, "311450739.ogg"},
	{0xD88E3730297710A1ULL, "llac_lt_0_lowtown_docks_a3_262143_0.dds"},
	{0xD88EB89F4DFDA45CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.gad"},
	{0xD88EEB3226E83AA2ULL, "w_encounter3_30_0.msh"},
	{0xD88EED3226E83E58ULL, "w_encounter3_30_0.mmh"},
	{0xD88EF93226E852FBULL, "w_encounter3_30_0.mao"},
	{0xD88F031422A6C0EDULL, "6111223_f.fxe"},
	{0xD88F3340CA0D006EULL, "6092565_f.fxe"},
	{0xD88F6817D1D6D8B6ULL, "hm_refugee02.mor"},
	{0xD88FB62A8E82BEC7ULL, "gen00fl_anders6111257.cl"},
	{0xD89021A2554BA9D2ULL, "6114484_m.fxe"},
	{0xD890883247526C7AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3_0.pwk"},
	{0xD890F5A1E519C618ULL, "mag170_thrask6020526.cl"},
	{0xD890FE14C612946AULL, "6076807_m.fxe"},
	{0xD8913C72311B23C5ULL, "hlac_lt_0_lowtown_docks_a3_n_-196611_0x.dds"},
	{0xD891B03B753FDC36ULL, "6182476_m.fxe"},
	{0xD891BD5DE7DF9703ULL, "hlac_ht_chantry_2534_13x.dds"},
	{0xD891E5E81C0CD1FEULL, "gen_im_trs_dal_arc_lbw_r2.uti"},
	{0xD8921766C62DA9E0ULL, "w_encounter3_147_0.phy"},
	{0xD8929F1DDE13D426ULL, "402104968.ogg"},
	{0xD892B45996D815A7ULL, "vxw_bonepitmining_a2_kx_0_0.vlm"},
	{0xD892BEB3EFE02C60ULL, "6128922_m.fxe"},
	{0xD892ED3226EC276BULL, "w_encounter3_30_0.rml"},
	{0xD89453E01A32058BULL, "vxlt_undercityencounter_1bm_0_1.vlm"},
	{0xD8946B17797C452CULL, "llac_lt_hangedman_n_-1_8.dds"},
	{0xD894829FC65C441CULL, "112654256.ogg"},
	{0xD8948828F60E099AULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.cess.evt"},
	{0xD894B79F4E020E5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_enter.ani"},
	{0xD894EB67E69DF196ULL, "6151397_m.fxe"},
	{0xD89559B057B5C079ULL, "vxlt_0_lowtown_docks_a3_brz_0_0.vlm"},
	{0xD895B727A5DD51EDULL, "zz_combat_load_test.lst"},
	{0xD895FA58549A19F9ULL, "vxca_cavern1_haunted_gk3_0_0.vlm"},
	{0xD8964D36CD055413ULL, "6096611_m.fxe"},
	{0xD89693D143D3F4DAULL, "781037415.ogg"},
	{0xD896DB2C69666FD1ULL, "vxw_bonepitmining_a2_1s9_0_0.vlm"},
	{0xD896EBD16A7E9C3EULL, "vxlt_gallowstemplar_n_5qy_0_0.vlm"},
	{0xD897034585085287ULL, "zz_car_debug6122472.cl"},
	{0xD8976E452EAD85D8ULL, "vxdr_primevaltomb_530_0_0.vlm"},
	{0xD89798D5F8A1B8CCULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_varb.mmh"},
	{0xD8979ED5F8A1C2BEULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_varb.msh"},
	{0xD8984236B2406FDBULL, "578676897.ogg"},
	{0xD89845DDD3B3DBE2ULL, "one100dg_kelder6060554.cl"},
	{0xD8984875D0693187ULL, "300532767.ogg"},
	{0xD89857242938FD99ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_avgdesk_d.dds"},
	{0xD89865475E72480FULL, "pro000_aveline6177136.cl"},
	{0xD8987F668037FDBDULL, "hlac_lt_0_lowtown_docks_a3_n_65535_5x.dds"},
	{0xD898F796744062C2ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.mmh"},
	{0xD8990E2C5E55C63EULL, "vxlt_gallowstemplar_f_55u_0_0.vlm"},
	{0xD89935CF0141F503ULL, "hlac_lt_0_lowtown_a2_1_0y.dds"},
	{0xD8999D63EA0349CFULL, "439242753.ogg"},
	{0xD89A30590852626DULL, "w_sundermount_62_0_lc.mmh"},
	{0xD89A51BFC1E47719ULL, "vxht_0_hightown_n_4n1_0_0.vlm"},
	{0xD89A6A1E9CA1C088ULL, "vxht_0_hightown_a2r_0_0.vlm"},
	{0xD89A80957C0DFC3CULL, "51118240.ogg"},
	{0xD89AC392CC0E2074ULL, "nat380_nathaniel6066317.cl"},
	{0xD89ACA50A0969D40ULL, "vxht_mansion_n_6s8_0_1.vlm"},
	{0xD89B832B90D9824DULL, "835989531.ogg"},
	{0xD89B8728F614268DULL, "art\\vfx\\combat\\vfx_m_sta_crust_c.cess.gad"},
	{0xD89B8A3CD50B0984ULL, "dae000_arishok6082017.cl"},
	{0xD89BC7958663D117ULL, "test_dd.mao"},
	{0xD89C4B07C38CD286ULL, "6101503_m.fxe"},
	{0xD89CA97B2D9DA532ULL, "llac_lt_0_lowtown_a2_n_131072_0.dds"},
	{0xD89D977E1E698BA3ULL, "6161166_m.fxe"},
	{0xD89DEAE961D7B2C6ULL, "6170240_m.fxe"},
	{0xD89E201C8757E640ULL, "6148924_m.fxe"},
	{0xD89EA391CE351A08ULL, "vis121_saemus6099982.cl"},
	{0xD89F2CE8C467D82EULL, "6040565_m.fxe"},
	{0xD8A06BCBAE7AD949ULL, "vxw_sundermount_20j_0_0.vlm"},
	{0xD8A183A5AA07E4B7ULL, "361795527.ogg"},
	{0xD8A1EA33DC7DAA0DULL, "w_sundermount_130_0.msh"},
	{0xD8A1F833DC7DC240ULL, "w_sundermount_130_0.mao"},
	{0xD8A1FC33DC7DC893ULL, "w_sundermount_130_0.mmh"},
	{0xD8A29261C339D876ULL, "lt_0_lowtown_docks_a3_156_0.mao"},
	{0xD8A29E61C339EC15ULL, "lt_0_lowtown_docks_a3_156_0.mmh"},
	{0xD8A2A061C339EF8BULL, "lt_0_lowtown_docks_a3_156_0.msh"},
	{0xD8A2B3EB3A05B93EULL, "vxht_bartrand_n_3ys_0_0.vlm"},
	{0xD8A2E74ECB90082AULL, "vxlt_0_lowtown_docks_a3_n_d3w_0_0.vlm"},
	{0xD8A4A6B04831AFAEULL, "vxw_woundedcoast_13k_0_0.vlm"},
	{0xD8A4EF7F5D4EC93DULL, "vxdr_primevaltomb_8o0_0_0.vlm"},
	{0xD8A4FB77BF24F274ULL, "6064734_m.fxe"},
	{0xD8A51A4312888690ULL, "6126825_m.fxe"},
	{0xD8A538EEEB70E385ULL, "gen_im_trs_gen_shd_met_r2.uti"},
	{0xD8A57EB6376722F7ULL, "vxw_bonepitdestroyed_a3_23v_0_0.vlm"},
	{0xD8A5A850FA875695ULL, "hlac_dr_primevaltomb_659_6y.dds"},
	{0xD8A5CD59E52A2BE1ULL, "mcr310_confused_guard6077049.cl"},
	{0xD8A5CF3D281F6E39ULL, "s_aoe_cl.mao"},
	{0xD8A66AC7878BCB56ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_appa_0n.dds"},
	{0xD8A68E863076A83EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_stump01_d.dds"},
	{0xD8A72E83739146D7ULL, "vxca_caverns1_e4k_0_0.vlm"},
	{0xD8A748229BC2DB3AULL, "895873811.ogg"},
	{0xD8A7D5B583669646ULL, "6018807_m.fxe"},
	{0xD8A7F36C7642C178ULL, "hlac_ca_caverns2_5802_65z.dds"},
	{0xD8A84A72C26C7027ULL, "vxlt_0_lowtown_n_e80_0_0.vlm"},
	{0xD8A86B6804669507ULL, "137694836.ogg"},
	{0xD8A871D38A3E7C3AULL, "665335495.ogg"},
	{0xD8AA2C72397DF5FEULL, "6165775_f.fxe"},
	{0xD8AA9585920DB83AULL, "vxw_encounter3_fz_0_1.vlm"},
	{0xD8AAD717213B7C6EULL, "728785841.ogg"},
	{0xD8AAEFB151EA6324ULL, "6148667_m.fxe"},
	{0xD8AB18549286A2D8ULL, "vxw_bonepit_p4_0_0.vlm"},
	{0xD8AB449094F1993AULL, "gen00fl_aveline6140017.cl"},
	{0xD8AB63F79C39F64FULL, "668928042.ogg"},
	{0xD8AB77D5C69E1ABCULL, "lgt110_arl_bann_teagan6115629.cl"},
	{0xD8ABE3B9BAE23FE4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edge.msh"},
	{0xD8ABF5B9BAE25E4DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edge.mao"},
	{0xD8ABF9B9BAE26486ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edge.mmh"},
	{0xD8AC461F6D37B820ULL, "hlac_w_bonepit_65540_0x.dds"},
	{0xD8AC5475186E3C0AULL, "dae200cr_viscount6087222.cl"},
	{0xD8AC879ED5FBB3B5ULL, "vxlt_shop_eb_0_0.vlm"},
	{0xD8ACD1DE191D87EBULL, "192601891.ogg"},
	{0xD8ACF7139A72BB1AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04a.phy"},
	{0xD8ACFE5274BB514FULL, "1029989531.ogg"},
	{0xD8AD208F5878157BULL, "fen210_hadriana6110532.cl"},
	{0xD8AD38733FE779F3ULL, "43344160.ogg"},
	{0xD8AD41ABB396697FULL, "t3_abm_b01.tnt"},
	{0xD8ADC30D6917CE4AULL, "vxw_bonepitdestroyed_a3_1ng_0_0.vlm"},
	{0xD8ADFB787F7B965CULL, "778518112.ogg"},
	{0xD8AE038353F91D14ULL, "qun221_petrice6088220.cl"},
	{0xD8AE27FF29702FA9ULL, "ctplt_smk.mao"},
	{0xD8AE3F02FC56323DULL, "vxlt_0_lowtown_gcb_0_0.vlm"},
	{0xD8AEDDA8A070F1C9ULL, "vxw_sundermount_25i_0_0.vlm"},
	{0xD8AF33E954104885ULL, "515288934.ogg"},
	{0xD8AF962BB640E5B6ULL, "6139154_f.fxe"},
	{0xD8AFB9B7C5862318ULL, "hlac_lt_0_lowtown_a3_n_c_-65535_5y.dds"},
	{0xD8B032A3774F7FB6ULL, "zz_approval_debug6003072.cl"},
	{0xD8B04E2B4502AE12ULL, "176055891.ogg"},
	{0xD8B08E3111BC7561ULL, "192819421.ogg"},
	{0xD8B0FEDE15ADC635ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2bwin.phy"},
	{0xD8B11B94B5AAE8B1ULL, "525217045.ogg"},
	{0xD8B1AC9E85836EBEULL, "913111415.ogg"},
	{0xD8B1CD3FD8D98DEAULL, "244983213.ogg"},
	{0xD8B1DEC8E53EAFC8ULL, "607644573.ogg"},
	{0xD8B2C90EC62CCB05ULL, "6054369_m.fxe"},
	{0xD8B3186F273F4B04ULL, "vxht_chantry_n_2a7_0_0.vlm"},
	{0xD8B3390DFE33846BULL, "330080045.ogg"},
	{0xD8B33D61792746FEULL, "zz_followers_banter6158061.cl"},
	{0xD8B37FBAF6EDCA03ULL, "vxht_0_hightown_n_c4n_0_0.vlm"},
	{0xD8B3B5F47F696186ULL, "hlac_lt_gallowstemplar_7198_1x.dds"},
	{0xD8B3F7A6CBF46D8FULL, "vxlt_darktown_58e_0_1.vlm"},
	{0xD8B45F575C145EE5ULL, "darktown_fem_fer_urch_sta6174785.cl"},
	{0xD8B4CF40713D0E9DULL, "llac_ht_0_hightown_a2_-65537_1.dds"},
	{0xD8B54B23706E99AAULL, "763900708.ogg"},
	{0xD8B54BBAB88D45EEULL, "art\\characters\\heraldry\\hld_merchants.dds"},
	{0xD8B56FB36FDFCDC9ULL, "mrl251ip_at_sundermount.utp"},
	{0xD8B58C6FA2F7A880ULL, "qun111_ketojan6040004.cl"},
	{0xD8B5F0B9BAEAAA51ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_edge.phy"},
	{0xD8B6208ABB026462ULL, "6148845_m.fxe"},
	{0xD8B6C66B8963C9F2ULL, "vxht_0_hightown_n_c73_0_1.vlm"},
	{0xD8B6CACAF9285323ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_enter.evt"},
	{0xD8B6FCC58C0E88C7ULL, "qun111_ketojan6041037.cl"},
	{0xD8B71A7E7CADD741ULL, "208016237.ogg"},
	{0xD8B7475ADBB68A4FULL, "6092361_m.fxe"},
	{0xD8B75F7EC30A95ADULL, "vxht_mansion2_2e9_0_0.vlm"},
	{0xD8B77E91D1C8E3EFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1ww_galow2bwin.phy"},
	{0xD8B7B2D9596EEEEFULL, "hlac_lt_0_lowtown_n_-3_0y.dds"},
	{0xD8B8005AC07C28A1ULL, "143388019.ogg"},
	{0xD8B83466C64EAA75ULL, "w_encounter3_147_0.msh"},
	{0xD8B83666C64EADEBULL, "w_encounter3_147_0.mmh"},
	{0xD8B84266C64EC188ULL, "w_encounter3_147_0.mao"},
	{0xD8B93BF03DA2B953ULL, "515145219.ogg"},
	{0xD8B99C0C684298B9ULL, "mag121_viveka6156629.cl"},
	{0xD8BAB647CBD72578ULL, "6090430_m.fxe"},
	{0xD8BACC237D4F122AULL, "768113530.ogg"},
	{0xD8BACE4E2745397AULL, "gen00fl_sebastian6179947.cl"},
	{0xD8BAD18441352EC1ULL, "2128921.ogg"},
	{0xD8BBAF031CB52630ULL, "art\\characters\\weapons\\staff\\w_stf_016a.mao"},
	{0xD8BBB3031CB52D43ULL, "art\\characters\\weapons\\staff\\w_stf_016a.mmh"},
	{0xD8BBC1031CB5449DULL, "art\\characters\\weapons\\staff\\w_stf_016a.msh"},
	{0xD8BBC7CB3C9C8DC7ULL, "570131732.ogg"},
	{0xD8BBF3D3F775DA03ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_conj_c.anb"},
	{0xD8BBF5B93FEEA80FULL, "ts_coterie_acc_props.gda"},
	{0xD8BC15CA42365137ULL, "6165629_m.fxe"},
	{0xD8BC77E052B742C6ULL, "darkcloud01.mao"},
	{0xD8BCA5E39148C59DULL, "6170702_m.fxe"},
	{0xD8BD9F73D1531A1AULL, "hlac_lt_hangedman_-1_11y.dds"},
	{0xD8BDA3CAF92E2FA4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_enter.gad"},
	{0xD8BDDE8E61D66EA4ULL, "plt_mrl100pt_n_mourn.plo"},
	{0xD8BDF63AFCA044ECULL, "vxlt_0_lowtown_docks_a3_n_ct3_0_0.vlm"},
	{0xD8BDFCA0F3B83C62ULL, "dr_primevaltomb_658_lc.mmh"},
	{0xD8BE0BB49751C50FULL, "6085448_m.fxe"},
	{0xD8BE7A74A0EAB1E3ULL, "vxca_caverns2prc_9xi_0_0.vlm"},
	{0xD8BE8EEAF1BED1E8ULL, "vxw_encounter2_5w_0_0.vlm"},
	{0xD8BEA396969EFDA4ULL, "mag131_taraneh6088682.cl"},
	{0xD8BEE7408625A329ULL, "955768944.ogg"},
	{0xD8BF175FF8638E05ULL, "492228167.ogg"},
	{0xD8BF573AF849E64BULL, "221326836.ogg"},
	{0xD8BFBD20C4F9DFCCULL, "6076052_m.fxe"},
	{0xD8BFD8511ABA1721ULL, "vxlt_0_lowtown_a2_n_c_hd5_0_0.vlm"},
	{0xD8BFEF46B695CBDAULL, "vxlt_hangedman_n_25c_0_0.vlm"},
	{0xD8BFF69C6951E4EDULL, "vxlt_hangedman_1dc_0_0.vlm"},
	{0xD8C0423C692147C3ULL, "183051999.ogg"},
	{0xD8C17A35F32E2B70ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\4so_stone.phy"},
	{0xD8C18A6A2C1DE0AEULL, "vxw_woundedcoast_19m_0_0.vlm"},
	{0xD8C1F197E8E3E283ULL, "vxht_0_hightown_n_c1q_0_0.vlm"},
	{0xD8C1F69674636669ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.anb"},
	{0xD8C20546A1770C33ULL, "943951242.ogg"},
	{0xD8C225BC35C3897AULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em03.mmh"},
	{0xD8C237BC35C3A7C0ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em03.msh"},
	{0xD8C26ABA7F0EB875ULL, "hlac_lt_0_lowtown_a2_n_c_-262143_0x.dds"},
	{0xD8C2C027A6044EAAULL, "zz_combat_load_test.are"},
	{0xD8C2DD2C75C939D3ULL, "vxlt_undercitydungeon_1oj_0_0.vlm"},
	{0xD8C3256EDFE83D5EULL, "ss_carver.bnk"},
	{0xD8C3368B11A2A5E5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_nobb_0n.dds"},
	{0xD8C382FF2753697BULL, "6128011_m.fxe"},
	{0xD8C3E9455F4E99E5ULL, "725192782.ogg"},
	{0xD8C4B2CAF9346782ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\paranoid_meredith_enter.ani"},
	{0xD8C50ECD3D5124BFULL, "dae150_chandan6148600.cl"},
	{0xD8C5D15BB500AE2DULL, "vxht_mansion2_2n8_0_0.vlm"},
	{0xD8C5D874880E1841ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_beckon.ani"},
	{0xD8C604AF262A7559ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1ww_windowxa.phy"},
	{0xD8C60D7BBC42B64DULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_table_d.dds"},
	{0xD8C76AC4CDABE80EULL, "6135889_m.fxe"},
	{0xD8C7D31B8B3EE4DDULL, "hlac_lt_0_lowtown_a2_n_c_131070_0y.dds"},
	{0xD8C889176558AB65ULL, "vxlt_0_lowtown_a2_n_c_hoh_0_0.vlm"},
	{0xD8C9049C9A27125AULL, "toolset_vars.gda"},
	{0xD8C90AFCB19F3A9CULL, "vxw_encounter3_6m_0_0.vlm"},
	{0xD8C97ADB7D606830ULL, "333959000.ogg"},
	{0xD8C9D0D3F781DC2CULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_conj_c.mmh"},
	{0xD8C9D729F5458DE3ULL, "6035861_m.fxe"},
	{0xD8CA20CE772A8181ULL, "vxht_0_hightown_n_ama_0_0.vlm"},
	{0xD8CB185FC5DF3FD4ULL, "hlac_lt_undercitydungeon_-1_23z.dds"},
	{0xD8CB647FFEEF41D9ULL, "mag311ar_warehouse.lst"},
	{0xD8CBB0E3100CFF34ULL, "mag360_alain6044639.cl"},
	{0xD8CBB1EA1892421AULL, "6027474_m.fxe"},
	{0xD8CBE5169985FE3CULL, "vxw_bonepitdestroyed_a3_19d_0_0.vlm"},
	{0xD8CC0C53B5394C6FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0ww_basic_2.mao"},
	{0xD8CC46C4BA917BECULL, "286986768.ogg"},
	{0xD8CC71B1BCE6B13AULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_exit.gad"},
	{0xD8CCB4031CC3A268ULL, "art\\characters\\weapons\\staff\\w_stf_016a.phy"},
	{0xD8CCE374881449E3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_beckon.gad"},
	{0xD8CD007B60B736FCULL, "6045968_m.fxe"},
	{0xD8CDA348471598F6ULL, "zz_isa_debug6133209.cl"},
	{0xD8CDACD3A8FDEAC2ULL, "6113672_m.fxe"},
	{0xD8CDBC3FFDD02172ULL, "824542434.ogg"},
	{0xD8CDE5BEE2DAAA9AULL, "281765197.ogg"},
	{0xD8CE88C6E5DC5D35ULL, "vxlt_gallowscourtyard_77g_0_0.vlm"},
	{0xD8CEF8586427256AULL, "6165457_m.fxe"},
	{0xD8D0111BD61509B9ULL, "art\\characters\\creatures\\spider\\textures\\c_poispidra_0s.dds"},
	{0xD8D0EC039E35EB24ULL, "585151849.ogg"},
	{0xD8D0FC75B71C4D10ULL, "gen00fl_merrill6142085.cl"},
	{0xD8D179C91E10F216ULL, "hlac_lt_gallowstemplar_f_2383_6y.dds"},
	{0xD8D20DDADBB40E85ULL, "819901775.ogg"},
	{0xD8D22133B9141535ULL, "dre000_dougal6131150.cl"},
	{0xD8D25C7FFEF552B0ULL, "mag311ar_warehouse.ncs"},
	{0xD8D2DC80D3CB5F56ULL, "313621288.ogg"},
	{0xD8D3419D543D3FC5ULL, "dae000_madam_lusine6152148.cl"},
	{0xD8D3422F2F924292ULL, "220138688.ogg"},
	{0xD8D35B5E8681603BULL, "llac_ht_bartrand_n_3368_0.dds"},
	{0xD8D374B1BCECD551ULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_exit.evt"},
	{0xD8D3C074881A2CC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_beckon.evt"},
	{0xD8D44338A6F6C6ADULL, "vxht_mansion2_2k7_0_0.vlm"},
	{0xD8D580E98F7A1783ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medp_0d.dds"},
	{0xD8D5A337398B0270ULL, "hm_stroud.mor"},
	{0xD8D5AF3600C21791ULL, "6095999_m.fxe"},
	{0xD8D6123091257178ULL, "lt_0_lowtown_docks_a3_33_0.rml"},
	{0xD8D64FC744636B0FULL, "lgt110ar_gang_hideout_lt.are"},
	{0xD8D67313DE3E9A08ULL, "llac_w_bonepitmining_a2_65537_0.dds"},
	{0xD8D6C1260BE22702ULL, "vxca_caverns1_eh6_0_0.vlm"},
	{0xD8D6CEDE91FE8F56ULL, "gen_im_trs_dem_lgt_boo_r1.uti"},
	{0xD8D71F982C04D92AULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.dura.gad"},
	{0xD8D76C21C5F2F187ULL, "mag211_marethari6025311.cl"},
	{0xD8D77CD9A395531CULL, "vxw_encounter3_3b_0_0.vlm"},
	{0xD8D78B805077C560ULL, "723016694.ogg"},
	{0xD8D78C433FCF5F47ULL, "one120_athenril6089430.cl"},
	{0xD8D86B280C67AB7BULL, "459406300.ogg"},
	{0xD8D8BE70C4343A55ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_basic.phy"},
	{0xD8D9426AAE01DD38ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobd_0d.dds"},
	{0xD8DA53DA7F3C8ACDULL, "6090866_m.fxe"},
	{0xD8DA7A3EA7D59525ULL, "lgtcr_revenant_1.utc"},
	{0xD8DA8B939A21EE80ULL, "lgt110_urn_guard6133109.cl"},
	{0xD8DAB3C60BDA14CAULL, "vxlt_0_lowtown_docks_b5w_0_0.vlm"},
	{0xD8DACFBF2913F0D2ULL, "vxlt_undercitydungeon_1rh_0_0.vlm"},
	{0xD8DAD9F4E08C75ECULL, "vxlt_0_lowtown_docks_a3_n_cz1_0_0.vlm"},
	{0xD8DB1EC2FF6A9365ULL, "vxlt_0_lowtown_docks_a3_9k6_0_0.vlm"},
	{0xD8DB50777A46EF6DULL, "178516474.ogg"},
	{0xD8DB597DF051342EULL, "one100dg_lia6049675.cl"},
	{0xD8DBC03CF659F7BCULL, "vxca_cavern1_haunted_inc_0_0.vlm"},
	{0xD8DC4C40CAE3B889ULL, "165058482.ogg"},
	{0xD8DC500FEBBCD273ULL, "gen_im_trs_drk_shd_met_r1.uti"},
	{0xD8DC5260B9E8463DULL, "gen00fl_anders6096460.cl"},
	{0xD8DC5D7FFEFDA89EULL, "mag311ar_warehouse.are"},
	{0xD8DC7EF194568AB6ULL, "character_character_p89.psh.fxo.dx11"},
	{0xD8DE02982C0AC6C1ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.dura.evt"},
	{0xD8DE57718D457F88ULL, "lt_0_lowtown_docks_a3_225_0.mmh"},
	{0xD8DE63718D45942BULL, "lt_0_lowtown_docks_a3_225_0.mao"},
	{0xD8DE75718D45B2B2ULL, "lt_0_lowtown_docks_a3_225_0.msh"},
	{0xD8DE7B98AE1B3925ULL, "vxht_keep_n_55m_0_0.vlm"},
	{0xD8DF420404279347ULL, "6060702_m.fxe"},
	{0xD8DF9D080A7A6400ULL, "gen00fl_varric6081212.cl"},
	{0xD8E0399789AB381EULL, "static_static_p1.psh.fxo"},
	{0xD8E08CB1BCF78820ULL, "art\\characters\\creatures\\ragedemon\\animation\\kd_exit.ani"},
	{0xD8E0C85BA99D2E2AULL, "art\\characters\\playercharacter\\humanmale\\glv_magb.mao"},
	{0xD8E0CC5BA99D3579ULL, "art\\characters\\playercharacter\\humanmale\\glv_magb.mmh"},
	{0xD8E0D65BA99D4667ULL, "art\\characters\\playercharacter\\humanmale\\glv_magb.msh"},
	{0xD8E0EB54E64E6EACULL, "mag370_meredith6044356.cl"},
	{0xD8E13EFBF8ABD338ULL, "mag100_cullen6082347.cl"},
	{0xD8E176E54E2909EFULL, "314385724.ogg"},
	{0xD8E17E329788E25CULL, "vxht_0_hightown_n_aq4_0_0.vlm"},
	{0xD8E2AA77AA80EB1BULL, "vxlt_hangedman_n_c_0_0.vlm"},
	{0xD8E346A6873F65D3ULL, "vxca_caverns2prc_a56_0_0.vlm"},
	{0xD8E34D610EF71911ULL, "dae000_mother6099241.cl"},
	{0xD8E3D241CCAF473BULL, "6162493_m.fxe"},
	{0xD8E3D547E1EB655AULL, "365359629.ogg"},
	{0xD8E3E70512712631ULL, "6047714_m.fxe"},
	{0xD8E4271EEE5C072FULL, "675817863.ogg"},
	{0xD8E493C23BA42E15ULL, "vxlt_hangedman_n_214_0_0.vlm"},
	{0xD8E4DF8D24F647C8ULL, "6018805_m.fxe"},
	{0xD8E4DFC85A5F04D8ULL, "vxw_bonepit_l0_0_0.vlm"},
	{0xD8E50405B4B5AC8FULL, "gen_im_npc_hum_arm_cht_chna.uti"},
	{0xD8E516DBB2820FBBULL, "hlac_lt_backalley_g_0_13y.dds"},
	{0xD8E54D528B4D3706ULL, "nat380_nathaniel_banter.cnv"},
	{0xD8E5E7234C103A7AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_loop.gad"},
	{0xD8E600EEB3C31A9FULL, "llac_lt_0_lowtown_a2_n_c_-65537_6.dds"},
	{0xD8E6CA558F4C70A7ULL, "zz_followers_banter6159339.cl"},
	{0xD8E7420D548A8F22ULL, "vxht_0_hightown_a3_avm_0_0.vlm"},
	{0xD8E744BC35E2D985ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em03.phy"},
	{0xD8E748C74471D2CCULL, "lgt110ar_gang_hideout_lt.lst"},
	{0xD8E78B8CE6685E7AULL, "vxht_0_hightown_b9n_0_0.vlm"},
	{0xD8E7A639136CCC76ULL, "825085459.ogg"},
	{0xD8E84F528B4F70C6ULL, "nat380_nathaniel_banter.bnk"},
	{0xD8E89A09158AE034ULL, "mer221_sabin6048290.cl"},
	{0xD8E8C5B77132A34DULL, "6095766_m.fxe"},
	{0xD8E904B49DBCB169ULL, "6072862_m.fxe"},
	{0xD8E9142A0C551956ULL, "vxw_encounter3_12_0_0.vlm"},
	{0xD8E958700C7A60FDULL, "lt_0_lowtown_docks_11_0.rml"},
	{0xD8E9AD3C8A2ADEF4ULL, "6161518_m.fxe"},
	{0xD8E9DDC23BB6928AULL, "cloud_f.dds"},
	{0xD8E9ED3C57CB8F91ULL, "4638822.ogg"},
	{0xD8EA1137A1B3A69AULL, "6169189_m.fxe"},
	{0xD8EA2F3BD86AFF04ULL, "6142287_m.fxe"},
	{0xD8EAC19ED47688EBULL, "6034076_f.fxe"},
	{0xD8EAE3FE323D344CULL, "mag170_hugh6206273.cl"},
	{0xD8EAFA982C154310ULL, "art\\vfx\\combat\\vfx_d_teleport_dir_p.dura.ani"},
	{0xD8EBEA234C14AA91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_loop.evt"},
	{0xD8EC177237DA74CCULL, "6119984_m.fxe"},
	{0xD8EC1C06B8C6EBD8ULL, "vxlt_0_lowtown_n_f66_0_0.vlm"},
	{0xD8ECDDA9E0761E79ULL, "6103570_f.fxe"},
	{0xD8ED5AB95AB9B624ULL, "fighting80.mao"},
	{0xD8EDD4C8EA749D21ULL, "gen00fl_merrill6107056.cl"},
	{0xD8EE32C8FD6F120CULL, "vxht_0_hightown_n_caz_0_0.vlm"},
	{0xD8EE564FDF104950ULL, "aip_mercenarycaptain_base.gda"},
	{0xD8EE58A65F3672E7ULL, "828661910.ogg"},
	{0xD8EE5B0FB9205801ULL, "sib131_slave_a6165273.cl"},
	{0xD8EEEFC97FEA4193ULL, "gen00fl_carver6102662.cl"},
	{0xD8EF61E515B92558ULL, "6123230_f.fxe"},
	{0xD8EF7551D23DE02DULL, "285745310.ogg"},
	{0xD8EF7ACFBA3D5BECULL, "hlac_ca_caverns2prc_5802_22z.dds"},
	{0xD8EF9667A0E2DF7AULL, "ht_0_hightownexplosion_n_20_0.rml"},
	{0xD8EFA4F15DB13A90ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc104a_0t.dds"},
	{0xD8EFC2A5D5841881ULL, "hlac_dr_dungeon_3216_8y.dds"},
	{0xD8EFFEDFAFD2FC62ULL, "follower_banter6172849.cl"},
	{0xD8F0398E0DF30B28ULL, "w_bonepitdestroyed_a3_21_0_ll.mmh"},
	{0xD8F0BA9FE318C345ULL, "art\\vfx\\combat\\vfx_glb_ele_f_major_c.anb"},
	{0xD8F0D245D3111DD8ULL, "20276968.ogg"},
	{0xD8F117562D797C54ULL, "vxw_blightlands_f1_0_0.vlm"},
	{0xD8F204DF1B5091F0ULL, "vxdr_0_deeproads_evy_0_0.vlm"},
	{0xD8F3ACFF3AE610FEULL, "6069281_m.fxe"},
	{0xD8F4978F0B965350ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_large.mao"},
	{0xD8F49B8F0B965A63ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_large.mmh"},
	{0xD8F4A98F0B9671BDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_large.msh"},
	{0xD8F4E722FC2401B8ULL, "999619007.ogg"},
	{0xD8F4F92F2FD8528CULL, "41984422.ogg"},
	{0xD8F5A8E5F488D76BULL, "6076091_m.fxe"},
	{0xD8F5B3EA4680A7A2ULL, "vxht_0_hightown_a3_asn_0_0.vlm"},
	{0xD8F6005E0B180190ULL, "6162735_m.fxe"},
	{0xD8F60F5A5735FDFBULL, "w_bonepitdestroyed_a3_137_0.mmh"},
	{0xD8F610AF0892DC7AULL, "6143804_m.fxe"},
	{0xD8F61B5A57361198ULL, "w_bonepitdestroyed_a3_137_0.mao"},
	{0xD8F62D5A573630E5ULL, "w_bonepitdestroyed_a3_137_0.msh"},
	{0xD8F7048080F7991CULL, "vxht_mansion_n_635_0_0.vlm"},
	{0xD8F713E6DBC5B297ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p_tw_v1.gad"},
	{0xD8F71E2EB483E08DULL, "vxlt_darktown_4kz_0_0.vlm"},
	{0xD8F76D1F0A128F13ULL, "311735340.ogg"},
	{0xD8F7716DEE38AC2FULL, "ser221_gamlen6134825.cl"},
	{0xD8F7A686BBBCE63EULL, "161918954.ogg"},
	{0xD8F7F6ABDC1333ECULL, "llac_w_encounter2_262147_0.dds"},
	{0xD8F81F197C20F774ULL, "6166418_m.fxe"},
	{0xD8F85C7790E794F2ULL, "vxht_brothel_n_3ka_0_0.vlm"},
	{0xD8F8A263228FAFC5ULL, "hlac_lt_0_lowtown_docks_65535_5z.dds"},
	{0xD8F8F1B03FFB351DULL, "6016961_m.fxe"},
	{0xD8F911DE15EB9C50ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2bwin.msh"},
	{0xD8F91F37AAE5D8B2ULL, "vxht_mansion_n_4xx_0_0.vlm"},
	{0xD8F91FDE15EBB46AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2bwin.mmh"},
	{0xD8F923DE15EBBAA1ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galow2bwin.mao"},
	{0xD8F98E5B76E0DB87ULL, "follower_banter6169288.cl"},
	{0xD8F9B958414C451FULL, "zz_followers_banter6159564.cl"},
	{0xD8F9CCD497C01D20ULL, "vxca_cavern1_haunted_eb7_0_0.vlm"},
	{0xD8F9D525C2409B38ULL, "vxht_0_hightown_a3_a2s_0_0.vlm"},
	{0xD8F9EEEDCB90C111ULL, "hair0.vsh.fxo"},
	{0xD8FA02234C211160ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\lean_on_table_loop.ani"},
	{0xD8FA480345FC4642ULL, "mer141_jansen6077796.cl"},
	{0xD8FA648205418C7AULL, "6140801_m.fxe"},
	{0xD8FA83018B26C662ULL, "6013849_f.fxe"},
	{0xD8FAF38E620A40EBULL, "plt_mrl100pt_n_mourn.nss"},
	{0xD8FB00283207754CULL, "6150327_m.fxe"},
	{0xD8FB3F25746730D3ULL, "vxlt_darktown_37k_0_0.vlm"},
	{0xD8FB977D1D126298ULL, "qun231_varnell6078132.cl"},
	{0xD8FBD073161B8B8EULL, "hlac_ca_caverns2_5802_5z.dds"},
	{0xD8FBFCB7349F35F6ULL, "869698536.ogg"},
	{0xD8FCBF7521CE3AB2ULL, "vxlt_gallowsprison_n_6mz_0_0.vlm"},
	{0xD8FCCC37C2AB6C41ULL, "6138737_m.fxe"},
	{0xD8FCE65A8F0A6949ULL, "1013402363.ogg"},
	{0xD8FCFA04E296313BULL, "6031300_m.fxe"},
	{0xD8FD09262F5A87CBULL, "hlac_ca_caverns1_mines_24422_6y.dds"},
	{0xD8FD8642ADAADAA8ULL, "vxw_sundermount_31m_0_0.vlm"},
	{0xD8FDA7E783982802ULL, "vxw_bonepit_167_0_0.vlm"},
	{0xD8FDABDCCE472FFBULL, "427079697.ogg"},
	{0xD8FDD3C207AF3DD8ULL, "6128330_f.fxe"},
	{0xD8FDED67733C0B6AULL, "hla_ht_0_hightown_a3_-131072_0z.dds"},
	{0xD8FE1E0911B66517ULL, "vxca_caverns1_ihs_0_0.vlm"},
	{0xD8FE28D281D4039AULL, "hlac_lt_gallowstemplar_f_2384_1y.dds"},
	{0xD8FE93448797FEDBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic.phy"},
	{0xD8FEA135EC054AF6ULL, "vxdr_primevaltomb_3te_0_0.vlm"},
	{0xD8FEDA07B443A81DULL, "llac_w_blightlands_65536_1.dds"},
	{0xD9000CABDEB8B7BDULL, "6115956_m.fxe"},
	{0xD90059B73C0D08EDULL, "vxw_encounter2_camp_6k_0_0.vlm"},
	{0xD901690D2828B5E4ULL, "vxlt_gallowscourtyard_6pb_0_0.vlm"},
	{0xD901EFDE0CB28E8DULL, "223637886.ogg"},
	{0xD901F9A07213175DULL, "vxw_bonepitmining_a2_1xl_0_0.vlm"},
	{0xD90247ECE65FD98BULL, "93074521.ogg"},
	{0xD902AAC6E834F5C7ULL, "hlac_lt_0_lowtown_a2_n_c_1_4z.dds"},
	{0xD902C9DAFF5E7B87ULL, "gen_im_npc_wep_sta_mar.uti"},
	{0xD902D0665789111AULL, "6034720_m.fxe"},
	{0xD902F330914C2A30ULL, "lt_0_lowtown_docks_a3_33_0.anb"},
	{0xD903460894843461ULL, "lowtown_argument6169377.cl"},
	{0xD904A472A83FD99CULL, "art\\characters\\weapons\\longswords\\w_lsw_002a.mao"},
	{0xD904B072A83FEE3FULL, "art\\characters\\weapons\\longswords\\w_lsw_002a.mmh"},
	{0xD904B672A83FF8E1ULL, "art\\characters\\weapons\\longswords\\w_lsw_002a.msh"},
	{0xD904C52F9A1E120EULL, "gen00fl_varric6033714.cl"},
	{0xD904F9FB30075154ULL, "730803739.ogg"},
	{0xD90508850EC2FBF8ULL, "6085106_m.fxe"},
	{0xD90557252A123941ULL, "vxlt_0_lowtown_a2_n_ev8_0_0.vlm"},
	{0xD9057CA9DB563FD3ULL, "hlac_lt_hangedman_n_-1_23x.dds"},
	{0xD905909B6A391D01ULL, "dae000ip_crowd_mf6_talk7.utp"},
	{0xD9059C8F0BA4CE88ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0pp_large.phy"},
	{0xD9066DB4300A7B67ULL, "640382491.ogg"},
	{0xD906AEBFA00700F0ULL, "430899555.ogg"},
	{0xD906CE5482DDAD72ULL, "vxht_brothel_n_3vb_0_0.vlm"},
	{0xD90732E67570A5DBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2light.msh"},
	{0xD90744E67570C426ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2light.mao"},
	{0xD90750E67570D8C5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2light.mmh"},
	{0xD907849E2DD9CEEDULL, "fen210_hadriana6110556.cl"},
	{0xD90798E91D2B5613ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_wic_0d.dds"},
	{0xD90818F809199166ULL, "6060142_f.fxe"},
	{0xD9081CF7D15BC27AULL, "364231377.ogg"},
	{0xD9089BD0E9B75F56ULL, "vxht_mansion_n_61i_0_0.vlm"},
	{0xD908C8F50AD186B7ULL, "658188984.ogg"},
	{0xD908F0139AC10E43ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04a.msh"},
	{0xD908FE139AC1259DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed04a.mmh"},
	{0xD909686DF03CD2F5ULL, "hlac_ht_mansion2b_818_26y.dds"},
	{0xD909CA6CE5DDD1B6ULL, "6133545_f.fxe"},
	{0xD90A1BF2629CF1A9ULL, "gen00fl_anders6125991.cl"},
	{0xD90AA242380B8E32ULL, "261379819.ogg"},
	{0xD90B40E6DBD6A81DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_p_tw_v1.ani"},
	{0xD90C04564FBA5F41ULL, "ch1armtnt_dy.psh.fxo"},
	{0xD90C479280AD7C7EULL, "6106869_m.fxe"},
	{0xD90C7C07F7E015C0ULL, "vxlt_warehouse_13g_0_0.vlm"},
	{0xD90C9CF05FF2C5A2ULL, "zz_mag_debug6054616.cl"},
	{0xD90CA7D0BFF4B133ULL, "mag121_idunna6084013.cl"},
	{0xD90D1312DDFB6376ULL, "vxdr_primevaltomb_3if_0_0.vlm"},
	{0xD90D6DDA70D55547ULL, "distortionwater.vsh.fxo"},
	{0xD90ECB942E03216DULL, "vxw_encounter2_camp_1j_0_0.vlm"},
	{0xD90EE55D90797417ULL, "zz_followers_banter6158769.cl"},
	{0xD90F4A37135677AAULL, "root_mage_arcane.dds"},
	{0xD90F758F22BBCB9EULL, "hlac_dr_slavers_3215_7x.dds"},
	{0xD9103FB7F397F714ULL, "392890663.ogg"},
	{0xD9106B7D64092FDDULL, "vxw_bonepitmining_a2_1uk_0_0.vlm"},
	{0xD9106B80357CBF96ULL, "w_bonepitmining_a2_244_0.msh"},
	{0xD9107580357CD094ULL, "w_bonepitmining_a2_244_0.mmh"},
	{0xD9107980357CD7A7ULL, "w_bonepitmining_a2_244_0.mao"},
	{0xD9112BA4A8CE824FULL, "6154940_m.fxe"},
	{0xD91147FFE96C84BBULL, "rdr111dg_martin6053571.cl"},
	{0xD91152E7487A76A9ULL, "6170203_m.fxe"},
	{0xD9116ABF5DE66111ULL, "lt_0_lowtown_docks_a3_n_128_0.mmh"},
	{0xD9116F5574A41EBBULL, "hlac_lt_0_lowtown_a2_65538_5z.dds"},
	{0xD91176BF5DE675B2ULL, "lt_0_lowtown_docks_a3_n_128_0.mao"},
	{0xD91184BF5DE68DCFULL, "lt_0_lowtown_docks_a3_n_128_0.msh"},
	{0xD911ABC60B17B649ULL, "108280381.ogg"},
	{0xD9126F8C10AB58AAULL, "6095300_f.fxe"},
	{0xD912A4DBCD8A9112ULL, "vxw_bonepitdestroyed_a3_n0_0_0.vlm"},
	{0xD912CAC85670FF56ULL, "vxlt_0_lowtown_a2_ey1_0_0.vlm"},
	{0xD912D0FF21D3E0ADULL, "lgt110cr_orz_harrow_carta.utc"},
	{0xD9131417720F5984ULL, "vxw_encounter3_ft_0_1.vlm"},
	{0xD913167E2DE74869ULL, "hlac_lt_0_lowtown_a2_n_65537_7z.dds"},
	{0xD913667A1BD1600AULL, "zz_hawke_magecast.cut"},
	{0xD91380CAD142A252ULL, "884939812.ogg"},
	{0xD91420E4E819F8FEULL, "vxlt_hangedman_vw_0_0.vlm"},
	{0xD914211CA00A31A5ULL, "vxht_0_hightown_a2_b86_0_0.vlm"},
	{0xD9153B0B0D0A3AA6ULL, "707320508.ogg"},
	{0xD915DFEA24B66F26ULL, "26245683.ogg"},
	{0xD916120FFAD36865ULL, "vxlt_hangedman_n_21d_0_0.vlm"},
	{0xD9168F9300979727ULL, "6016798_m.fxe"},
	{0xD917288A32AB980AULL, "vxlt_0_lowtown_f56_0_0.vlm"},
	{0xD91801CDCF787313ULL, "997403289.ogg"},
	{0xD9187CF5A628E6BCULL, "follower_banter6171576.cl"},
	{0xD918B9F9371C70D8ULL, "vxca_caverns2prc_9fx_0_0.vlm"},
	{0xD918DBE076A37B46ULL, "vxca_caverns1_mines_j2j_0_0.vlm"},
	{0xD9191EE614E46FA2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetop.msh"},
	{0xD91920E614E47358ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetop.mmh"},
	{0xD9192A812A8A4AC1ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredacti.gad"},
	{0xD9192CE614E487FBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetop.mao"},
	{0xD9198C50D22A2DF0ULL, "vxlt_warehouse_128_0_0.vlm"},
	{0xD919BB9FE33BC986ULL, "art\\vfx\\combat\\vfx_glb_ele_f_major_c.mmh"},
	{0xD91A482FCC571B65ULL, "hlac_lt_0_lowtown_docks_-2_2x.dds"},
	{0xD91AE243B265B15BULL, "vxlt_0_lowtown_a2_n_hgt_0_0.vlm"},
	{0xD91B15ABFD173740ULL, "lt_0_lowtown_a3_n_c_42_0_ll.mmh"},
	{0xD91C57700CA5B815ULL, "lt_0_lowtown_docks_11_0.anb"},
	{0xD91C6CED035BC5D4ULL, "vxw_blightlands_q4_0_0.vlm"},
	{0xD91C9C1395DDBD4BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgta_0d.dds"},
	{0xD91D0A4A2CF0A665ULL, "6115964_f.fxe"},
	{0xD91D49F6A6F0C355ULL, "302244059.ogg"},
	{0xD91D699A7209A5DFULL, "vxlt_0_lowtown_ewr_0_0.vlm"},
	{0xD91E0B38D5BB1033ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_action.ani"},
	{0xD91E10C3042DA8E4ULL, "mer221_sabin6080571.cl"},
	{0xD91E1910862A5C8BULL, "118088984.ogg"},
	{0xD91E4FD4250C74E3ULL, "follower_banter6169819.cl"},
	{0xD91E6CA10C4C59BBULL, "vxdr_primevalentrance_21f_0_0.vlm"},
	{0xD91EA5DEFB400CB1ULL, "21240341.ogg"},
	{0xD91EDD5A55FF485DULL, "vxw_bonepitmining_a2_1rj_0_0.vlm"},
	{0xD91EECB8C004C3CBULL, "vxca_caverns1_iml_0_0.vlm"},
	{0xD91F33FF940BC68AULL, "zz_seb_debug6214527.cl"},
	{0xD91F52EBA0E107F3ULL, "cod_bks_lite_vials_0.utp"},
	{0xD91FD170C4705D70ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_basic.msh"},
	{0xD91FDF70C470748AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_basic.mmh"},
	{0xD91FE370C4707BC1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_basic.mao"},
	{0xD9202B812A906B6FULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_follow3_meredacti.ani"},
	{0xD9202E45F03D04F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\baseswordshield\\blendtree\\commander.anb"},
	{0xD921B53450759D78ULL, "hlac_lt_0_lowtown_docks_a3_196609_0x.dds"},
	{0xD921DE9566603DE4ULL, "6148872_m.fxe"},
	{0xD92262E99C840BCFULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.dura.gad"},
	{0xD92292C1DA10117EULL, "vxlt_hangedman_sh_0_0.vlm"},
	{0xD922CA7EC3A9BC8FULL, "6147892_m.fxe"},
	{0xD923104FB22BA0A5ULL, "mrl210_merrill_varterralpt2.stg"},
	{0xD923257F5A895559ULL, "w_encounter2_camp_56_0.msh"},
	{0xD9232F7F5A896647ULL, "w_encounter2_camp_56_0.mmh"},
	{0xD923337F5A896C94ULL, "w_encounter2_camp_56_0.mao"},
	{0xD9237230D41C2C1BULL, "hlac_ht_0_hightown_65533_0x.dds"},
	{0xD92389DDE08CD4A5ULL, "dae200_bran6094763.cl"},
	{0xD923CEF876EB9335ULL, "6096572_m.fxe"},
	{0xD925922F7B3730C0ULL, "386168620.ogg"},
	{0xD925AAFC30ADC644ULL, "6173966_m.fxe"},
	{0xD92607876D0F22EDULL, "vxca_caverns1_j5i_0_0.vlm"},
	{0xD9262337A7C0C133ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_ceilingcap.phy"},
	{0xD926D28460F0459CULL, "vxht_0_hightown_n_b8h_0_0.vlm"},
	{0xD926E3C729CF453CULL, "vxlt_gallowscourtyard_7cv_0_0.vlm"},
	{0xD9271A8CD9485711ULL, "vxlt_darktown_48c_0_0.vlm"},
	{0xD92721F9292C87F7ULL, "357823216.ogg"},
	{0xD9273A309694149FULL, "6115738_m.fxe"},
	{0xD9277B9E38E7B997ULL, "vrc321_varric6082695.cl"},
	{0xD9278EC1AEA351BDULL, "vxca_cavern1_haunted_j5p_0_0.vlm"},
	{0xD927E114D89AD6C2ULL, "6119601_m.fxe"},
	{0xD928130BE103C082ULL, "vxca_caverns1_j26_0_1.vlm"},
	{0xD92824F646FEF690ULL, "25853647.ogg"},
	{0xD9285EB2171BC7F5ULL, "t1_tat_org.tnt"},
	{0xD92884F5DF03793BULL, "vxlt_hangedman_10y_0_0.vlm"},
	{0xD9289F4B9DAB790DULL, "vxht_0_hightown_a3_c3j_0_1.vlm"},
	{0xD9294B878F6531CAULL, "6104565_f.fxe"},
	{0xD92963A9EF5CE92AULL, "qcr000_arishok6039307.cl"},
	{0xD92977E99C8A4E7CULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.dura.evt"},
	{0xD92987511C9F0AB5ULL, "vxht_mansion2_2o3_0_0.vlm"},
	{0xD929C24D1244819EULL, "vxht_keep_n_8bn_0_0.vlm"},
	{0xD92A3228D2F3D62DULL, "vxlt_hangedman_5i_0_0.vlm"},
	{0xD92A7A4487BD0F7AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic.msh"},
	{0xD92A8401182069B4ULL, "zz_dre_debug6037646.cl"},
	{0xD92A884487BD26B3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic.mao"},
	{0xD92A8C4487BD2DC0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_basic.mmh"},
	{0xD92AD5631DF38CFAULL, "vxdr_primevaltomb_6xx_0_0.vlm"},
	{0xD92B0E08C120BB5DULL, "6125483_m.fxe"},
	{0xD92B1989B495CF2AULL, "146183432.ogg"},
	{0xD92B3A8F1D4F4372ULL, "101222093.ogg"},
	{0xD92B889EDA6B4D61ULL, "296013969.ogg"},
	{0xD92BFB3C5B38E6C2ULL, "mag320_orsino6070446.cl"},
	{0xD92C0338D5C7406AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_action.evt"},
	{0xD92C2657943A6B57ULL, "6169539_m.fxe"},
	{0xD92CE7EA514EF4FBULL, "vxw_bonepitmining_a2_81_0_0.vlm"},
	{0xD92CFFDBD0A3FBD9ULL, "323156577.ogg"},
	{0xD92D37BAD58FB80DULL, "959740649.ogg"},
	{0xD92D43365C609EF4ULL, "hlac_lt_undercitydungeon_-1_40y.dds"},
	{0xD92D7C0F313B4336ULL, "w_sundermount_191_0.msh"},
	{0xD92D860F313B5434ULL, "w_sundermount_191_0.mmh"},
	{0xD92D8A0F313B5B47ULL, "w_sundermount_191_0.mao"},
	{0xD92DD73C47D992A3ULL, "6030624_m.fxe"},
	{0xD92DEB0F5EE09F77ULL, "47162922.ogg"},
	{0xD92DFD18B7203006ULL, "vxlt_hangedman_ti_0_0.vlm"},
	{0xD92E6C4F930D774AULL, "6198480_m.fxe"},
	{0xD92E9FBC6DC9072FULL, "blood_conj_07.mao"},
	{0xD92EB7A51D6D0401ULL, "6106004_m.fxe"},
	{0xD92F153032A18213ULL, "vxca_cave_2k5_0_0.vlm"},
	{0xD9302FBD186C4759ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit_2.ani"},
	{0xD930DE6D2217CE49ULL, "vxw_sundermount_2fp_0_0.vlm"},
	{0xD93188505DC4E0E5ULL, "vxht_bartrand_n_4fz_0_0.vlm"},
	{0xD931AFE9A7B32F4EULL, "llac_ht_store_187_3.dds"},
	{0xD9320F56D08B1356ULL, "vxht_keep_n_6m9_0_0.vlm"},
	{0xD932290023181E0BULL, "613022052.ogg"},
	{0xD9323BC0C497CB4AULL, "6168151_m.fxe"},
	{0xD93286DB934A49BBULL, "str_sp.mao"},
	{0xD9329661094FF906ULL, "gen00fl_fenris6113874.cl"},
	{0xD93299FA8BDA1EE3ULL, "hlac_lt_gallowstemplar_n_2383_9y.dds"},
	{0xD932B5E4E35D8885ULL, "6216369_m.fxe"},
	{0xD932E238D5CD26DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\chest_kick_action.gad"},
	{0xD932F0C9F8010861ULL, "6125790_f.fxe"},
	{0xD9331BFA04067736ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_ext_c.impa.ani"},
	{0xD933C287FC91505FULL, "vxca_cave_1fq_0_0.vlm"},
	{0xD933E83D8797A067ULL, "vxlt_warehouse_n_xy_0_0.vlm"},
	{0xD9341CD922A3DEC4ULL, "6176666_m.fxe"},
	{0xD9344F45DD9FAD7AULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_detonate.gad"},
	{0xD93551B3746AE05BULL, "vxht_0_hightown_a3_ben_0_0.vlm"},
	{0xD935860858470A7FULL, "137842169.ogg"},
	{0xD936141B963F25D7ULL, "vfx_chunkyparticle_p4.vsh.fxo.dx11.assembly"},
	{0xD9367FE99C94E595ULL, "art\\vfx\\cinematic\\vfx_dragon_transform_p.dura.ani"},
	{0xD936857B1A0A4DFEULL, "146462412.ogg"},
	{0xD9370ABD1872269BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit_2.gad"},
	{0xD9382F8D6D95FCD3ULL, "mrl310_party_comments6142629.cl"},
	{0xD93895F1E8CAF767ULL, "vxlt_darktown_4tq_0_0.vlm"},
	{0xD93946A3B81D788BULL, "240442050.ogg"},
	{0xD93960D9DE58C54CULL, "381877541.ogg"},
	{0xD93977B3E44AD27EULL, "65461477.ogg"},
	{0xD93986AB7DC13789ULL, "6084243_m.fxe"},
	{0xD939E527C7F54D93ULL, "6218589_m.fxe"},
	{0xD939EFE4F7B2E836ULL, "vxlt_hangedman_n_26v_0_0.vlm"},
	{0xD93A051303E5280AULL, "7809958.ogg"},
	{0xD93A3505630F8A82ULL, "vxlt_backalley_g_1jq_0_0.vlm"},
	{0xD93A5245DDA41D91ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_detonate.evt"},
	{0xD93AC2AE18BAA631ULL, "6013337_m.fxe"},
	{0xD93AF1E45B98D39EULL, "1022897191.ogg"},
	{0xD93BA4C76BD6D93DULL, "857404618.ogg"},
	{0xD93BB5252342AD42ULL, "569092436.ogg"},
	{0xD93BC37FA14479D9ULL, "6018633_m.fxe"},
	{0xD93CF285AB2025C4ULL, "vxdr_primevaltomb_5bs_0_0.vlm"},
	{0xD93D305A57723830ULL, "w_bonepitdestroyed_a3_137_0.phy"},
	{0xD93D8851E367393FULL, "ht_brothel_n_mmap_ph.dds"},
	{0xD93D97C4F29F0F85ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_avea_0n.dds"},
	{0xD93E27BD18787680ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit_2.evt"},
	{0xD93E3862BC10A9B9ULL, "vxlt_undercitydungeon_1kt_0_0.vlm"},
	{0xD93E77A18416D9F4ULL, "hlac_w_blightlands_262143_0y.dds"},
	{0xD93E9519B7072267ULL, "6140017_m.fxe"},
	{0xD93F504A140DE6C9ULL, "vxw_sundermount_2kw_0_0.vlm"},
	{0xD93F689F98DF501DULL, "6118692_m.fxe"},
	{0xD93F6D88F272C094ULL, "fen310_fenris6118694.cl"},
	{0xD93F9CC891C4D90FULL, "hlac_ca_cave_2194_13x.dds"},
	{0xD93FD6D04C69A075ULL, "gen00fl_aveline6132879.cl"},
	{0xD93FE80BF03FFA99ULL, "hightown_market_guard_sta6169580.cl"},
	{0xD93FFA2D4FBAF965ULL, "vxht_bartrand_n_4ky_0_0.vlm"},
	{0xD94025C1584AB41BULL, "1065128066.ogg"},
	{0xD94039BC0EEE7222ULL, "gen_im_cut_cross_bolt.uti"},
	{0xD94049F39550C67EULL, "6144979_m.fxe"},
	{0xD9404C98EACBB035ULL, "6139320_f.fxe"},
	{0xD94080D01E1C22E6ULL, "s_stand_03.dds"},
	{0xD94123FA0412C227ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_ext_c.impa.evt"},
	{0xD9412BE52A8B343DULL, "lgt260_eau_tmp_ending6171807.cl"},
	{0xD9416C509FC4127AULL, "vxw_bonepit_z4_0_0.vlm"},
	{0xD941832B6A810D0DULL, "6148150_m.fxe"},
	{0xD94218281BFD1F29ULL, "hlac_lt_0_lowtown_a3_n_c_-2_0y.dds"},
	{0xD9423ADA783491DFULL, "6121940_m.fxe"},
	{0xD942B500DA30BA14ULL, "dae000_flemeth6000796.cl"},
	{0xD942C7A2934D9D73ULL, "993238836.ogg"},
	{0xD9430AB393B19C31ULL, "587735734.ogg"},
	{0xD9432A9584DF389CULL, "6075646_m.fxe"},
	{0xD943C7810DBB763CULL, "vxlt_gallowscourtyard_7it_0_0.vlm"},
	{0xD943CD5BA9F1B466ULL, "art\\characters\\playercharacter\\humanmale\\glv_magb.phy"},
	{0xD9444B297FD3A242ULL, "183618623.ogg"},
	{0xD945246BFA55DEE0ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc104a_0d.dds"},
	{0xD9452CF2BE3B7AA0ULL, "713041549.ogg"},
	{0xD9461D319DF925AAULL, "art\\vfx\\combat\\vfx_m_bld_control_c.impa.gad"},
	{0xD94662E5AC97A98FULL, "mag360_thrask6043895.cl"},
	{0xD9468CF63AFD924CULL, "vxlt_0_lowtown_a3_n_c_h1z_0_0.vlm"},
	{0xD9470ACC9AF0A653ULL, "lt_0_lowtown_docks_a3_n_3_0_lc.mmh"},
	{0xD9473E98AED17BE5ULL, "6095387_f.fxe"},
	{0xD94744FA04176670ULL, "art\\vfx\\combat\\vfx_glb_ice_imp_ext_c.impa.gad"},
	{0xD947587233B79E6BULL, "hlac_lt_0_lowtown_a2_n_c_65536_10y.dds"},
	{0xD9475B2A307203C5ULL, "vxdr_primevaltomb_6ut_0_0.vlm"},
	{0xD947A4D858E1872CULL, "6084196_m.fxe"},
	{0xD948249BE449B82DULL, "gen00fl_dog6141145.cl"},
	{0xD9486A45DDB08460ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_detonate.ani"},
	{0xD9487FCF8A874E03ULL, "282583184.ogg"},
	{0xD948A6E25505A302ULL, "vxlt_backalley_g_1wr_0_0.vlm"},
	{0xD948F10DE4C41B5DULL, "mag210_arianni6024541.cl"},
	{0xD94907285AEFC04AULL, "6024090_f.fxe"},
	{0xD949587D0D4C52DFULL, "one100dg_vanard6173858.cl"},
	{0xD9495F89C03126CEULL, "45938768.ogg"},
	{0xD94A06979B360DA9ULL, "cd_isa321_velasco.wbd"},
	{0xD94ACD6CE4397930ULL, "mag360_samson6044880.cl"},
	{0xD94AD247D3978CCBULL, "700199023.ogg"},
	{0xD94B10FF21B25122ULL, "vxht_keep_5g7_0_0.vlm"},
	{0xD94B3A623265531BULL, "mcr310_argument6182473.cl"},
	{0xD94B4FCB2B443210ULL, "llac_lt_gallowstemplar_n_2394_9.dds"},
	{0xD94BEF7DE86030DBULL, "860885969.ogg"},
	{0xD94C09BE2DDF683AULL, "m_thi121ar_mansion.bnk"},
	{0xD94C360882FE85B5ULL, "vxw_bonepitdestroyed_a3_18f_0_0.vlm"},
	{0xD94C789203FA31D2ULL, "6027786_m.fxe"},
	{0xD94C9E547324CE27ULL, "llac_lt_0_lowtown_a2_n_65536_11.dds"},
	{0xD94D00319DFF1341ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.impa.evt"},
	{0xD94D83FEBFC9317DULL, "mcr310_argument6127603.cl"},
	{0xD94D8A6CD876EFFDULL, "vxlt_0_lowtown_a2_n_bxq_0_0.vlm"},
	{0xD94DC2270603FF49ULL, "vxw_sundermount_2lv_0_0.vlm"},
	{0xD94E12B1CCBAB81DULL, "vxht_bartrand_n_5qv_0_0.vlm"},
	{0xD94E7C3CB9701580ULL, "dae000_keep_maid_sta6162419.cl"},
	{0xD94E966FDCA8DCD1ULL, "plt_lgt000pt_tombs_final.plo"},
	{0xD94EBD6B1C2A9F4EULL, "isa120_hayder6170361.cl"},
	{0xD94F1CAA2BF8B7F5ULL, "6113064_m.fxe"},
	{0xD94F29E675AE4C72ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_2light.phy"},
	{0xD94F58CC532ED6FDULL, "hlac_lt_0_lowtown_a3_n_c_-65537_5z.dds"},
	{0xD94FCA6556AA6BB2ULL, "vxht_mansion_n_4ff_0_0.vlm"},
	{0xD9500605921619EDULL, "6121160_m.fxe"},
	{0xD9501206D6635DDBULL, "vxht_mansion2_2l8_0_0.vlm"},
	{0xD95014F4113CFCA5ULL, "one100dg_nabil6033141.cl"},
	{0xD950187F5AB02D7CULL, "w_encounter2_camp_56_0.phy"},
	{0xD950611181EB4DD5ULL, "mrl310_party_comments6143048.cl"},
	{0xD9509D60A55E6728ULL, "ts_gems_04.gda"},
	{0xD95116E977529D6CULL, "549866578.ogg"},
	{0xD951225A05AF290FULL, "hlac_lt_0_lowtown_docks_-65539_1y.dds"},
	{0xD9512D49D0F78BAEULL, "6111560_m.fxe"},
	{0xD9515621278A1984ULL, "vxlt_0_lowtown_a2_n_c_hl4_0_0.vlm"},
	{0xD9521B7CA19D9402ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgta.msh"},
	{0xD9521D7CA19D97B8ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgta.mmh"},
	{0xD952297CA19DAC5BULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgta.mao"},
	{0xD95258A515759EBBULL, "257392212.ogg"},
	{0xD9525A1930AF503AULL, "vxht_mansion_n_6hb_0_0.vlm"},
	{0xD952643D6C98E9F1ULL, "6112544_m.fxe"},
	{0xD9536EB3D0BD146AULL, "zz_dre_debug6204374.cl"},
	{0xD953A1917512B34FULL, "299003132.ogg"},
	{0xD9548D3B36B9FBB2ULL, "6095923_m.fxe"},
	{0xD954C2C5FC2DD2EDULL, "vxdr_primevalentrance_229_0_0.vlm"},
	{0xD955374EDA0D00CDULL, "dae000_gamlen6093303.cl"},
	{0xD9554EAA9826D34EULL, "w_bonepit_108_0.phy"},
	{0xD95566255AA460B6ULL, "vxht_bartrand_n_4ep_0_0.vlm"},
	{0xD9556B187A1147E6ULL, "llac_dr_primevaltomb_659_6.dds"},
	{0xD955A3714BE6F16EULL, "6128366_m.fxe"},
	{0xD955B28D1854AD60ULL, "art\\characters\\creatures\\strider\\textures\\c_stridera_0s.dds"},
	{0xD955CD0722681C45ULL, "vxdr_primevaltomb_6xs_0_0.vlm"},
	{0xD955DA6B72AC4C1EULL, "6076079_f.fxe"},
	{0xD9562E723BBAFD9EULL, "304877212.ogg"},
	{0xD9562F7F5AB4BFA4ULL, "w_encounter2_camp_56_0.rml"},
	{0xD95634B60933F4F7ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafglw.msh"},
	{0xD95646B60934135AULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafglw.mao"},
	{0xD9564AB609341A69ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\bsh_bigleafglw.mmh"},
	{0xD9568750F9F33ACFULL, "791807201.ogg"},
	{0xD95687DCEE2DB7BCULL, "lgt320im_mis_resp_sword_01.uti"},
	{0xD956D8962AA58F19ULL, "mag110_ws_banter6214713.cl"},
	{0xD95707ACC68E6D0BULL, "6095773_m.fxe"},
	{0xD9570B3F92DA6C91ULL, "dae000_madam_lusine6152229.cl"},
	{0xD9571EB9AF736C27ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\diev1.gad"},
	{0xD957237A91046DEAULL, "vxlt_0_lowtown_docks_a3_cz8_0_0.vlm"},
	{0xD9575D4795D10DCEULL, "928190249.ogg"},
	{0xD9575F7C76A064B7ULL, "qun110_petrice6018508.cl"},
	{0xD9576E11CB3CD27EULL, "mag291_vincento6027720.cl"},
	{0xD957B428429E9620ULL, "6119012_m.fxe"},
	{0xD957D6E94F0CD09FULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.cess.gad"},
	{0xD9580B25848B7193ULL, "hlac_w_woundedcoast_-196608_0y.dds"},
	{0xD95850B8046BB67DULL, "ico_helm_light_3.dds"},
	{0xD95862FF74C0FE5AULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpa.phy"},
	{0xD9586C8C69F7F798ULL, "85340723.ogg"},
	{0xD958ADCDD5A36AD6ULL, "lowtown_male_beggar_sta6151050.cl"},
	{0xD958F9FF4431B2F3ULL, "72560779.ogg"},
	{0xD959133A7755B645ULL, "docks_scene_dockworkers6176207.cl"},
	{0xD9593196CEE2CB6AULL, "vxca_caverns2_7tw_0_0.vlm"},
	{0xD959F8319E098F90ULL, "art\\vfx\\combat\\vfx_m_bld_control_c.impa.ani"},
	{0xD95AC04FE95042F0ULL, "63138007.ogg"},
	{0xD95B82D383F804DBULL, "w_bonepit_201_0.mmh"},
	{0xD95B8ED383F81978ULL, "w_bonepit_201_0.mao"},
	{0xD95B99E569807C92ULL, "361776738.ogg"},
	{0xD95BA0D383F837C5ULL, "w_bonepit_201_0.msh"},
	{0xD95BDF1C88B4607BULL, "vxlt_hangedman_4p_0_0.vlm"},
	{0xD95C247CA1A5F843ULL, "art\\characters\\playercharacter\\humanmale\\bdy_lgta.phy"},
	{0xD95C848EBEB0D09DULL, "vxht_bartrand_n_5tu_0_0.vlm"},
	{0xD95C970AA93E7AB6ULL, "zz_car_debug6127335.cl"},
	{0xD95CA32834DCB0AFULL, "6177396_m.fxe"},
	{0xD95D3A20D09E0E17ULL, "6156533_m.fxe"},
	{0xD95D48BDAC01AFFCULL, "vxw_encounter2_r_0_0.vlm"},
	{0xD95D561660C80DF2ULL, "llac_lt_gallowscourtyard_-1_0.dds"},
	{0xD95D5EF3067DBF63ULL, "vxca_caverns2prc_91r_0_0.vlm"},
	{0xD95D65A60434A4F6ULL, "56104136.ogg"},
	{0xD95E13B9AF797834ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\diev1.evt"},
	{0xD95E316454DB54B1ULL, "582684479.ogg"},
	{0xD95E36E63EC06B62ULL, "vxlt_0_lowtown_n_gkr_0_0.vlm"},
	{0xD95E7E114DC0EE2CULL, "vxdr_0_deeproads_9g2_0_0.vlm"},
	{0xD95EA2F271C5B934ULL, "537594870.ogg"},
	{0xD95ECBE94F12DC8CULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.cess.evt"},
	{0xD95EF743EA571932ULL, "art\\vfx\\combat\\vfx_glb_confused_c.cess.gad"},
	{0xD95F8DC9337A1A41ULL, "vxht_0_hightown_n_9zl_0_0.vlm"},
	{0xD95FE47516F1D6CDULL, "79764109.ogg"},
	{0xD95FFE76BD7ADC41ULL, "150539787.ogg"},
	{0xD960CC392DCB7FABULL, "815060564.ogg"},
	{0xD9616129C8D1BD26ULL, "vxca_caverns2_8p0_0_0.vlm"},
	{0xD96167E0D8B4EB63ULL, "844772177.ogg"},
	{0xD961EE3C8A4B4917ULL, "vxlt_0_lowtown_e0o_0_0.vlm"},
	{0xD962F0ED6069D2C0ULL, "559400825.ogg"},
	{0xD9636189F35B8746ULL, "865275201.ogg"},
	{0xD9638763C681DE89ULL, "llac_w_sundermount_0_0.dds"},
	{0xD96388B9945B2297ULL, "6146785_m.fxe"},
	{0xD963D25DB6E13C8DULL, "sto_fl_cl.mao"},
	{0xD963EF74C847C564ULL, "6048123_m.fxe"},
	{0xD963EFCEA89CB82EULL, "vxht_chantry_33p_0_1.vlm"},
	{0xD9646557DEDB553AULL, "807789308.ogg"},
	{0xD964D952996D59BBULL, "6174902_m.fxe"},
	{0xD9652031C238CD96ULL, "6115564_m.fxe"},
	{0xD965F8E09B74BAECULL, "vxht_mansion2b_3e9_0_0.vlm"},
	{0xD965FA43EA5D3D69ULL, "art\\vfx\\combat\\vfx_glb_confused_c.cess.evt"},
	{0xD966124D51475F6DULL, "vxht_0_hightown_n_a86_0_0.vlm"},
	{0xD96633E429CBB864ULL, "lgt110_arl_bann_teagan6115571.cl"},
	{0xD966B7753745F77CULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\flame.ani"},
	{0xD966B972A892C784ULL, "art\\characters\\weapons\\longswords\\w_lsw_002a.phy"},
	{0xD966E3CF47C544D9ULL, "82136649.ogg"},
	{0xD966F5D54AE37EA7ULL, "hlac_lt_gamlenshouse_n_-1_5z.dds"},
	{0xD967018C4BC2C46DULL, "mag101_grace6034051.cl"},
	{0xD96728E415839CA9ULL, "vxlt_0_lowtown_h5k_0_0.vlm"},
	{0xD967BB51DD306C29ULL, "6077241_m.fxe"},
	{0xD967C4AF5023B5D9ULL, "716955518.ogg"},
	{0xD967D6C191028D3DULL, "95919471.ogg"},
	{0xD968446D41DFA1F5ULL, "6080715_m.fxe"},
	{0xD968801F3538AF2FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cooker_sa.dds"},
	{0xD968979F625E14D6ULL, "6112027_f.fxe"},
	{0xD968B0B9908FE99FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\pp_male_warrior.anb"},
	{0xD968D6EE38611CACULL, "and100_anders6173566.cl"},
	{0xD969310B90441956ULL, "fxa_pois_spin.dds"},
	{0xD96968C2BFB190D7ULL, "gen_im_arm_hlm_hvy_act2_02.uti"},
	{0xD9697EFCA62BEBADULL, "6103340_m.fxe"},
	{0xD969CCED77CDEECEULL, "351959774.ogg"},
	{0xD969CD228976A7E4ULL, "233711814.ogg"},
	{0xD969D9B41A2B7DA5ULL, "316276754.ogg"},
	{0xD969EFD24C4F6898ULL, "gen_im_trs_qun_lgt_hlm_r2.uti"},
	{0xD969F90DEB61E6A6ULL, "359050873.ogg"},
	{0xD96A48842699E01FULL, "qun221_viscount6084734.cl"},
	{0xD96AA5E0E9F03049ULL, "vxw_sundermount_2zt_0_0.vlm"},
	{0xD96ABA8C6C4817AAULL, "vxht_mansion2b_313_0_0.vlm"},
	{0xD96AC994FF190BBCULL, "vxw_bonepitdestroyed_a3_1ps_0_0.vlm"},
	{0xD96AE81B7414FAE5ULL, "dae000_arishok6174621.cl"},
	{0xD96B2C53765656A4ULL, "cod_ltr_sk_starkhaven.utp"},
	{0xD96B7D24B615DD5FULL, "mrl000_marethari6129206.cl"},
	{0xD96B8C1BD495D5F7ULL, "vxht_mansion_n_52n_0_0.vlm"},
	{0xD96BB4C5D0D4462CULL, "vxlt_0_lowtown_n_ff2_0_0.vlm"},
	{0xD96C089E3A460CBBULL, "dae150_chandan6205596.cl"},
	{0xD96C0BB9AF85A86DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\diev1.ani"},
	{0xD96C2A07FEC692B9ULL, "lgt160_ntb_elf_conflict6163399.cl"},
	{0xD96C32DE69F6ECBCULL, "vxca_cavern1_haunted_i8y_0_0.vlm"},
	{0xD96CB3E94F1EF1E5ULL, "art\\vfx\\combat\\vfx_glb_purity_crust_hf_c.cess.ani"},
	{0xD96D25392601E219ULL, "6021202_m.fxe"},
	{0xD96D3124426D1B4AULL, "gen00fl_anders6119496.cl"},
	{0xD96D43B62176CD8BULL, "6128738_m.fxe"},
	{0xD96D798F6FE55FEBULL, "vxlt_0_lowtown_docks_a3_aot_0_0.vlm"},
	{0xD96DC475374C2BBEULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\flame.gad"},
	{0xD96E58F2E4233DAEULL, "vxht_mansion2b_331_0_0.vlm"},
	{0xD96E8E7102E6EA9CULL, "6034731_m.fxe"},
	{0xD96EB3260E4E61AFULL, "vxlt_backalley_g_1u9_0_0.vlm"},
	{0xD96EBB2A62736B32ULL, "pro000_friedrich6051051.cl"},
	{0xD96ED88B67D2CF12ULL, "vxht_bartrand_n_5rx_0_0.vlm"},
	{0xD96ED8979B54DB8BULL, "cd_isa321_velasco.bnk"},
	{0xD96EEAC8BF6F9CC2ULL, "vxw_encounter2_24_0_0.vlm"},
	{0xD96F0E2253CD95A0ULL, "vxlt_undercitydungeon_13y_0_0.vlm"},
	{0xD96F1F146284B6EAULL, "vxlt_0_lowtown_n_bxn_0_0.vlm"},
	{0xD96F39210F9BD242ULL, "382053447.ogg"},
	{0xD97052B7E0936B64ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse3cloth_orsinosdefiance.gad"},
	{0xD97056E242D0D0B6ULL, "6143963_m.fxe"},
	{0xD9708EA30D5442A2ULL, "hlac_ht_mansion2b_818_40z.dds"},
	{0xD97094D46FBE8F18ULL, "6016731_m.fxe"},
	{0xD9709CE9A3CE595AULL, "vxw_woundedcoast_1f0_0_0.vlm"},
	{0xD970C9B8F54F174DULL, "hlac_w_sundermount_327675_0z.dds"},
	{0xD970DB7873573AB4ULL, "vxlt_hangedman_n_2mu_0_0.vlm"},
	{0xD970FAF8EEC36D0CULL, "sst_ss_bethany.gda"},
	{0xD97108242DB42D13ULL, "175246132.ogg"},
	{0xD97123BE2DFEB0D8ULL, "m_thi121ar_mansion.wbd"},
	{0xD9716907ADE391D9ULL, "gen00fl_aveline6116693.cl"},
	{0xD9723F76D8F36781ULL, "484705726.ogg"},
	{0xD972AEB78355EFC5ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_d_ink_n.dds"},
	{0xD972E243EA679ED8ULL, "art\\vfx\\combat\\vfx_glb_confused_c.cess.ani"},
	{0xD97348554E6DEC82ULL, "character_character_p252.psh.fxo.dx11"},
	{0xD9735905F97D845CULL, "vxca_caverns2_586_0_1.vlm"},
	{0xD973857AECD24549ULL, "6040870_m.fxe"},
	{0xD9739B690E9D0297ULL, "6049656_f.fxe"},
	{0xD973FE25D40E42FAULL, "vxlt_0_lowtown_n_f59_0_0.vlm"},
	{0xD974634CF2450D64ULL, "hlac_lt_0_lowtown_docks_a3_65534_2z.dds"},
	{0xD974773CF4E4DE9AULL, "6105896_f.fxe"},
	{0xD9749F7537520BDDULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\flame.evt"},
	{0xD975208CC496BA14ULL, "addxp.ncs"},
	{0xD9754D294824B2F6ULL, "6126467_m.fxe"},
	{0xD9757F9DDF65BE96ULL, "mrl150_merrill6077674.cl"},
	{0xD975EB0ADE4E1017ULL, "6087503_m.fxe"},
	{0xD97611993C087F94ULL, "680842219.ogg"},
	{0xD9762D2ECF2684A9ULL, "6078341_m.fxe"},
	{0xD9764170152AD31DULL, "360373410.ogg"},
	{0xD976A495B23A6202ULL, "1012011992.ogg"},
	{0xD976C5A5A3FC5BB8ULL, "vxht_mansion2_2eh_0_0.vlm"},
	{0xD97704BD17425948ULL, "hlac_ht_0_hightown_a3_-65537_3z.dds"},
	{0xD97731C7F0C9CD6FULL, "6036459_m.fxe"},
	{0xD97758CAB129B287ULL, "63579146.ogg"},
	{0xD97761B7E099A342ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_corpse3cloth_orsinosdefiance.ani"},
	{0xD97794A3E3AE4F07ULL, "106789514.ogg"},
	{0xD977A506CB6888F8ULL, "ser221_party_comments6163696.cl"},
	{0xD977AE0762E9F7C8ULL, "art\\vfx\\models\\vfx_long_stick.msh"},
	{0xD977CC0762EA2AF2ULL, "art\\vfx\\models\\vfx_long_stick.mmh"},
	{0xD977F0D99822042DULL, "6104240_m.fxe"},
	{0xD9782DFAE205A300ULL, "678340855.ogg"},
	{0xD978937236844723ULL, "vxlt_gallowstemplar_n_38q_0_0.vlm"},
	{0xD97924F8378B2274ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_exit.ani"},
	{0xD97A3FFD16B4E4AAULL, "vxlt_gallowstemplar_6s3_0_0.vlm"},
	{0xD97B6DF378CBEDBFULL, "hlac_ca_caverns1_24423_10x.dds"},
	{0xD97C457744BBB17BULL, "vxlt_0_lowtown_n_ewm_0_0.vlm"},
	{0xD97CC6C41BE4BDADULL, "gen00fl_aveline6139768.cl"},
	{0xD97D250300447A2FULL, "vxlt_backalley_g_1r8_0_0.vlm"},
	{0xD97D4A6859C8E792ULL, "vxht_bartrand_n_5uy_0_0.vlm"},
	{0xD97D78DD22ABC756ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fc_transition.ani"},
	{0xD97DBB7DA0B539AAULL, "zz_approval_debug6003731.cl"},
	{0xD97DBC905BCE8F48ULL, "vxlt_warehouse_n_1b7_0_0.vlm"},
	{0xD97E3B8E006221A6ULL, "vxht_0_hightown_bu6_0_0.vlm"},
	{0xD97E904FD32D4C12ULL, "6079154_m.fxe"},
	{0xD97F0EC695C471DAULL, "vxw_woundedcoast_1a1_0_0.vlm"},
	{0xD97F21F8378F88D6ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_2hands_on_hip_exit.gad"},
	{0xD97F96BBE28706AAULL, "6070126_m.fxe"},
	{0xD98015E6E420EA8FULL, "lgt000cr_epic_merchant_4.utc"},
	{0xD98023DC8AEC3BB1ULL, "hlac_lt_backalley_n_0_16y.dds"},
	{0xD98077E9B57B5808ULL, "589675680.ogg"},
	{0xD98086548D2DFEE2ULL, "1034872191.ogg"},
	{0xD981380B9058B684ULL, "fxa_pois_spin.mao"},
	{0xD98180AFE5914A5CULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.dura.ani"},
	{0xD981A50D58AF6228ULL, "7187140.ogg"},
	{0xD981DFF9706294C0ULL, "qcr220_orsino6061789.cl"},
	{0xD98288F4981A5047ULL, "6079270_m.fxe"},
	{0xD982981424B08838ULL, "ser270_emeric6113505.cl"},
	{0xD9829E6EE37ECECAULL, "vxw_woundedcoast_by_0_0.vlm"},
	{0xD982C5477B334E56ULL, "vxw_sundermount_312_0_1.vlm"},
	{0xD982E909589A105CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphangb.mao"},
	{0xD982EC533CCC2B47ULL, "6198351_m.fxe"},
	{0xD982F509589A24FFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphangb.mmh"},
	{0xD982FB09589A2EA1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_hngtarphangb.msh"},
	{0xD98332DA5B615DBEULL, "w_encounter2_camp_38_0_ll.mmh"},
	{0xD98350E1EAF2F163ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage.evt"},
	{0xD98409C35DBF6F4BULL, "vxca_caverns1_i2e_0_0.vlm"},
	{0xD984244364317A9CULL, "81647787.ogg"},
	{0xD98459E1DAF8CB77ULL, "6108307_m.fxe"},
	{0xD984F9CC37D4C263ULL, "vxht_brothel_n_5k6_0_0.vlm"},
	{0xD985178BBF0C872AULL, "one280cr_emissary.utc"},
	{0xD98549B06B99A145ULL, "706842716.ogg"},
	{0xD985569ABD584DCEULL, "hlac_lt_0_lowtown_a2_n_c_-1_0z.dds"},
	{0xD98597E8970F376FULL, "follower_banter6173092.cl"},
	{0xD9859834146A976CULL, "dae000cr_chanter.utc"},
	{0xD986089D6E8F78A4ULL, "art\\vfx\\materials\\vfx_entropic_blend.mao"},
	{0xD986452D23749F56ULL, "6110512_m.fxe"},
	{0xD9865EB92C1349C6ULL, "cen_smk_smk.mao"},
	{0xD9868227CDA7EB6BULL, "dae000_gamlen6145110.cl"},
	{0xD986D968618CFF6FULL, "564318651.ogg"},
	{0xD98714D17FBEBC1DULL, "zz_followers_banter6157348.cl"},
	{0xD9872EF85866D657ULL, "fen110_fenris6096230.cl"},
	{0xD9878DAFE595CB9EULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.dura.gad"},
	{0xD9879ED505B21114ULL, "gen_im_wep_war_one_act2_03.uti"},
	{0xD987A1D27DBCFEF7ULL, "870851497.ogg"},
	{0xD987E2471BC66B1AULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_wkf.gad"},
	{0xD988CB50A0E09572ULL, "art\\vfx\\creatures\\fxc_ogre_db_a_c.mmh"},
	{0xD9891C37A81532A8ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_ceilingcap.mmh"},
	{0xD9891E067EEF75F6ULL, "vxlt_0_lowtown_docks_c4y_0_0.vlm"},
	{0xD9892837A815474BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_ceilingcap.mao"},
	{0xD9893A37A81565D2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ca_ceilingcap.msh"},
	{0xD9894F1C761BAAAEULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.cess.evt"},
	{0xD989F7F798DC0BA2ULL, "128446308.ogg"},
	{0xD98A0AA60B7B5753ULL, "908760715.ogg"},
	{0xD98A29288989C92DULL, "hlac_w_bonepitmining_a2_0_11y.dds"},
	{0xD98A29E1EAF8CDE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage.gad"},
	{0xD98A698080AAC82FULL, "vxw_bonepitmining_a2_1hv_0_0.vlm"},
	{0xD98A80DD22B65F47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fc_transition.evt"},
	{0xD98AE9A46B9B3036ULL, "w_sundermount_80_0_ll.mmh"},
	{0xD98B1D94EF627B84ULL, "vxlt_0_lowtown_a2_n_c_hx8_0_0.vlm"},
	{0xD98B73278823CDFDULL, "lt_0_lowtown_docks_a3_n_163_0.mao"},
	{0xD98B77278823D436ULL, "lt_0_lowtown_docks_a3_n_163_0.mmh"},
	{0xD98B81278823E534ULL, "lt_0_lowtown_docks_a3_n_163_0.msh"},
	{0xD98BBC454BBF0012ULL, "vxht_bartrand_n_5xz_0_0.vlm"},
	{0xD98BF1DC37B9C6A0ULL, "vxlt_undercitydungeon_19g_0_0.vlm"},
	{0xD98CF90B09DC21CDULL, "hlac_lt_0_lowtown_a2_65537_2x.dds"},
	{0xD98D5E6395271D32ULL, "334413332.ogg"},
	{0xD98D80A387BA8A5AULL, "vxw_woundedcoast_1l2_0_0.vlm"},
	{0xD98DA274566BDA6DULL, "dae200_saemus6141198.cl"},
	{0xD98DE073CE2FB328ULL, "hlac_lt_0_lowtown_docks_-131074_5z.dds"},
	{0xD98E0898D47D1F2AULL, "6075226_m.fxe"},
	{0xD98E5B0ABFDECFF7ULL, "gen00fl_isabela6085679.cl"},
	{0xD98E68AFE59BABBDULL, "art\\vfx\\combat\\vfx_m_cre_glyph_rep_tar_c.dura.evt"},
	{0xD98E926ADCC8D204ULL, "6198428_m.fxe"},
	{0xD98EA15463D195C0ULL, "2801746.ogg"},
	{0xD98EB10E9F7B1211ULL, "llac_lt_backalley_g_65537_1.dds"},
	{0xD98EE5471BCC8F31ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_wkf.evt"},
	{0xD98F41151D1080EEULL, "6175153_m.fxe"},
	{0xD9902E2E2013CD35ULL, "hlac_w_blightlands_131071_0y.dds"},
	{0xD990461C7621BA59ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.cess.gad"},
	{0xD99083ECE68CB9C5ULL, "vxlt_0_lowtown_n_f23_0_0.vlm"},
	{0xD990A4AA9E375F3DULL, "rdr111dg_assistant6016634.cl"},
	{0xD990B0849FE8DD3FULL, "vxht_mansion2b_3e6_0_0.vlm"},
	{0xD9911ACD864C4624ULL, "61824498.ogg"},
	{0xD99138E1EAFF05C2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\additive_damage.ani"},
	{0xD991519CC9DA0C5CULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.dura.ani"},
	{0xD991A1DD22BCB590ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_fc_transition.gad"},
	{0xD991EA1D932D5415ULL, "vxlt_hangedman_n_2u8_0_0.vlm"},
	{0xD99230E32C10E3F6ULL, "6128667_m.fxe"},
	{0xD9923E66E9282173ULL, "one100dg_elren6060992.cl"},
	{0xD9926E2CF8D720CEULL, "6104904_m.fxe"},
	{0xD9929E17A2B983E5ULL, "33883031.ogg"},
	{0xD993170DF168F890ULL, "471577654.ogg"},
	{0xD993199437DAFF0FULL, "242445170.ogg"},
	{0xD9944AC771AEB6E8ULL, "w_blightlands_182_0.msh"},
	{0xD9945CC771AED579ULL, "w_blightlands_182_0.mao"},
	{0xD99468C771AEE912ULL, "w_blightlands_182_0.mmh"},
	{0xD9946D9FFA16FAF8ULL, "6174938_m.fxe"},
	{0xD994C099895E09ACULL, "6115980_m.fxe"},
	{0xD994D1A383AF657EULL, "vxht_0_hightown_a2_5yi_0_0.vlm"},
	{0xD994F67EFD5FCD09ULL, "hlac_lt_0_lowtown_docks_n_-131075_4z.dds"},
	{0xD9951F78255C544BULL, "ht_chantry_n_2534_lc.mmh"},
	{0xD99533B26397BB46ULL, "vxlt_warehouse_vc_0_0.vlm"},
	{0xD9954DAA985CB511ULL, "w_bonepit_108_0.mmh"},
	{0xD99559AA985CC9B2ULL, "w_bonepit_108_0.mao"},
	{0xD99567AA985CE1CFULL, "w_bonepit_108_0.msh"},
	{0xD995A58F873B7E9AULL, "w_bonepitmining_a2_255_0.mmh"},
	{0xD995A98F873B85D1ULL, "w_bonepitmining_a2_255_0.mao"},
	{0xD995AC4DBF458716ULL, "hlac_lt_darktown_-1_26z.dds"},
	{0xD995B78F873B9DE0ULL, "w_bonepitmining_a2_255_0.msh"},
	{0xD9961E43BCA83FE2ULL, "319264429.ogg"},
	{0xD9961F2BD4FB553CULL, "vxw_bonepitdestroyed_a3_1cv_0_0.vlm"},
	{0xD9967FC47CF216AAULL, "vxca_caverns1_mines_j2f_0_0.vlm"},
	{0xD996CCBFF06D02BCULL, "gen00fl_merrill6108230.cl"},
	{0xD996DE40086BAC67ULL, "qcr000_arishok6057094.cl"},
	{0xD9971EADDE58D0F0ULL, "vxlt_undercitydungeon_x1_0_0.vlm"},
	{0xD9974FB6CA9C8FABULL, "ave200_donnic6110690.cl"},
	{0xD997571C7627F5A7ULL, "art\\vfx\\combat\\vfx_d_invulner_sustain_c.cess.ani"},
	{0xD9975E9CC9DE8D9EULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.dura.gad"},
	{0xD9977C9EFC14F538ULL, "6135228_f.fxe"},
	{0xD9979DBCBCCB8E1EULL, "dae000_pc_home_ambients6208316.cl"},
	{0xD9983F520E0241A9ULL, "vxca_cavern1_haunted_efv_0_0.vlm"},
	{0xD99855ADB0D792BCULL, "w_sundermount_66_0.mao"},
	{0xD998560E3727E225ULL, "353908342.ogg"},
	{0xD99861ADB0D7A75FULL, "w_sundermount_66_0.mmh"},
	{0xD99867ADB0D7B101ULL, "w_sundermount_66_0.msh"},
	{0xD9986EB9729D0F67ULL, "6119385_m.fxe"},
	{0xD998D5E5C8C7BD44ULL, "vxw_bonepit_8p_0_0.vlm"},
	{0xD998E452AC14851FULL, "260536067.ogg"},
	{0xD998E57D46D7F165ULL, "6104781_m.fxe"},
	{0xD998F1CF3AF03F5CULL, "gen00fl_anders6107682.cl"},
	{0xD99911D34623698CULL, "6080642_m.fxe"},
	{0xD999293128A7E27BULL, "vxlt_0_lowtown_n_emo_0_0.vlm"},
	{0xD9995DE16A42DCE0ULL, "6179515_m.fxe"},
	{0xD999EE97AE91401FULL, "m_dae351ar_mount_caverns.bnk"},
	{0xD99A1EF5111904DDULL, "58677646.ogg"},
	{0xD99A80C5A5F71FB1ULL, "427872045.ogg"},
	{0xD99A8A8704567197ULL, "vxlt_0_lowtown_a2_es2_0_0.vlm"},
	{0xD99B323777EDE78DULL, "hlac_lt_gallowstemplar_n_2383_2z.dds"},
	{0xD99B37D271430302ULL, "vxw_bonepitmining_a2_27m_0_0.vlm"},
	{0xD99B7AB0840C61C1ULL, "884862683.ogg"},
	{0xD99BB463D57134D8ULL, "vxlt_0_lowtown_a3_n_c_ewm_0_0.vlm"},
	{0xD99BC26197E72202ULL, "hlac_lt_darktown_-1_55x.dds"},
	{0xD99BCF2653FD8288ULL, "hlac_ca_caverns2_5802_64y.dds"},
	{0xD99BED7A47156ED9ULL, "vxlt_gallowstemplar_39c_0_0.vlm"},
	{0xD99C2AAFEB610EB7ULL, "art\\vfx\\textures\\vfx_gore.dds"},
	{0xD99C4A025ABCD760ULL, "vxlt_0_lowtown_docks_n_bfw_0_0.vlm"},
	{0xD99C5FF2F9D141B0ULL, "6086253_m.fxe"},
	{0xD99C8D3E41D68AECULL, "mcr370_assignments6081716.cl"},
	{0xD99CCF5A1B81201BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_sw_holy_smite.evt"},
	{0xD99CFD471BD8F500ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_tn90r_wkf.ani"},
	{0xD99D52AC187FDD8BULL, "hm_one_veld.mor"},
	{0xD99D7933E376A068ULL, "lt_0_lowtown_docks_208_0.mmh"},
	{0xD99D8533E376B40BULL, "lt_0_lowtown_docks_208_0.mao"},
	{0xD99D9733E376D292ULL, "lt_0_lowtown_docks_208_0.msh"},
	{0xD99DBCDC88FAAE43ULL, "dae000_meredith6085278.cl"},
	{0xD99E24B08F3A6B82ULL, "vxw_bonepitmining_a2_1v5_0_0.vlm"},
	{0xD99E399CC9E46DBDULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_ice_p.dura.evt"},
	{0xD99E905A557193AFULL, "art\\levels\\races\\elf\\props\\_textures\\prp_vase_n.dds"},
	{0xD99ECC9A0657154EULL, "vxlt_backalley_g_1dw_0_0.vlm"},
	{0xD99EF5C9D882D245ULL, "vxlt_0_lowtown_n_f52_0_0.vlm"},
	{0xD99EFBCB6180017AULL, "6076050_m.fxe"},
	{0xD99F2428CE54A09EULL, "zz_followers_banter6159453.cl"},
	{0xD99F4D3D07D1543BULL, "399438166.ogg"},
	{0xD99F8D511DE1C4CBULL, "6047482_m.fxe"},
	{0xD9A02D5A08589275ULL, "art\\vfx\\materials\\vfx_dot_blur.mao"},
	{0xD9A040DBD01AEB3DULL, "dae000_gamlen6101408.cl"},
	{0xD9A0931C34216971ULL, "6101957_m.fxe"},
	{0xD9A09E6379380605ULL, "6083624_m.fxe"},
	{0xD9A0D147CFC5A3BFULL, "6147463_m.fxe"},
	{0xD9A0EFE0773FD76AULL, "6088749_m.fxe"},
	{0xD9A11484BFD1B937ULL, "6106943_m.fxe"},
	{0xD9A1386A7509567DULL, "106666516.ogg"},
	{0xD9A179F4506EF557ULL, "w_encounter3_138_0.msh"},
	{0xD9A18BF4506F13BAULL, "w_encounter3_138_0.mao"},
	{0xD9A18FF4506F1AC9ULL, "w_encounter3_138_0.mmh"},
	{0xD9A1F8807BE3DCA1ULL, "685915645.ogg"},
	{0xD9A25425EC50C744ULL, "vxlt_0_lowtown_a3_n_c_eoi_0_0.vlm"},
	{0xD9A254B0EF374E44ULL, "gen00fl_varric6033558.cl"},
	{0xD9A2D42046DBED00ULL, "rdr111dg_martin6055073.cl"},
	{0xD9A2F64EE15EFB3CULL, "988940312.ogg"},
	{0xD9A3A3D38435F210ULL, "w_bonepit_201_0.phy"},
	{0xD9A3C85A1B87335CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_sw_holy_smite.gad"},
	{0xD9A3E00CC484C8B7ULL, "llac_w_encounter3_-196605_0.dds"},
	{0xD9A43E6B828B5281ULL, "hlac_ht_0_hightown_n_1_0z.dds"},
	{0xD9A447367846B154ULL, "719920488.ogg"},
	{0xD9A46D9CF9BBDC91ULL, "vxlt_warehouse_n_sb_0_0.vlm"},
	{0xD9A4CDF72E056D0BULL, "6068496_m.fxe"},
	{0xD9A4FEC1768DD680ULL, "hlac_lt_gallowscourtyard_-65538_2z.dds"},
	{0xD9A545AFB863F5A5ULL, "vxht_brothel_n_4y3_0_0.vlm"},
	{0xD9A55D9D3738EA28ULL, "328395280.ogg"},
	{0xD9A5B5957B44FD29ULL, "6166671_m.fxe"},
	{0xD9A61D056554AD49ULL, "185642522.ogg"},
	{0xD9A625804E2EF6CDULL, "vxht_mansion2_260_0_0.vlm"},
	{0xD9A65E00EDFE01D0ULL, "portraits.gfx"},
	{0xD9A66000EDFE056FULL, "portraits.gda"},
	{0xD9A664D84BCA5475ULL, "6042902_m.fxe"},
	{0xD9A6A16FDCF36536ULL, "plt_lgt000pt_tombs_final.nss"},
	{0xD9A6A2FAC66AE738ULL, "llac_w_woundedcoast_262146_0.dds"},
	{0xD9A6A383E66D8321ULL, "gen_im_npc_dwf_arm_hlm_lega.uti"},
	{0xD9A734B91347CCB5ULL, "vxlt_gallowscourtyard_7bh_0_0.vlm"},
	{0xD9A7CC19D6BE0926ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_right.evt"},
	{0xD9A7D941DE23F88AULL, "6096473_f.fxe"},
	{0xD9A818E22C65862BULL, "vxw_bonepitmining_a2_8a_0_0.vlm"},
	{0xD9A8238840E2B6CEULL, "6137528_m.fxe"},
	{0xD9A8A01E687204DCULL, "hlac_w_sundermount_65538_0z.dds"},
	{0xD9A8A28CBA1ACD3BULL, "vfx_steamsml.bnk"},
	{0xD9A8FACACE94F7F3ULL, "qun100_mage_party_comment6019121.cl"},
	{0xD9A9A74BA95A26FAULL, "311700877.ogg"},
	{0xD9A9C75A1B8B9D5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_sw_holy_smite.ani"},
	{0xD9A9DD0BCFC088E6ULL, "941566764.ogg"},
	{0xD9AA25932C38BFD4ULL, "one120_athenril6088898.cl"},
	{0xD9AA2C7CDAE9293DULL, "vxht_brothel_n_32e_0_0.vlm"},
	{0xD9AAB7AF8008F0A6ULL, "mrl000ar_house.ncs"},
	{0xD9AABBDF4CB2EFE0ULL, "vxlt_0_lowtown_docks_n_bct_0_0.vlm"},
	{0xD9AC1FA546487D63ULL, "345239458.ogg"},
	{0xD9AC5901CAE92CFCULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg102a_0n.dds"},
	{0xD9AC968D81308402ULL, "vxw_bonepitmining_a2_1s6_0_0.vlm"},
	{0xD9ACBD557737208BULL, "53276907.ogg"},
	{0xD9ACEA05537D3E53ULL, "1041352373.ogg"},
	{0xD9ACFD3E0D921FDEULL, "vxht_mansion2_1po_0_1.vlm"},
	{0xD9AD20BB6838DE9BULL, "w_bonepitdestroyed_a3_1_0_lc.mmh"},
	{0xD9AD554E2C540507ULL, "552240979.ogg"},
	{0xD9ADA159FC23BDD0ULL, "qun211cr_henchmen26088116.cl"},
	{0xD9ADD83E45F216D5ULL, "w_bonepitdestroyed_a3_150_0.phy"},
	{0xD9ADE319D6C29C71ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_right.gad"},
	{0xD9ADF405B961184AULL, "vxw_woundedcoast_qt_0_0.vlm"},
	{0xD9AE258C8D6720B5ULL, "10461395.ogg"},
	{0xD9AE9B268EB29570ULL, "hlac_ht_mansion2b_818_15z.dds"},
	{0xD9AF10406B2E1E85ULL, "6086924_m.fxe"},
	{0xD9AF2BA8E9DD4E27ULL, "vxht_mansion_n_33o_0_0.vlm"},
	{0xD9AF61BD6935EFEAULL, "6083649_m.fxe"},
	{0xD9AFB60B03EDE994ULL, "mag110_samson6049736.cl"},
	{0xD9B05A1FE5C74FB9ULL, "follower_banter6151450.cl"},
	{0xD9B09A09DDF0F88CULL, "mag100_karras6035324.cl"},
	{0xD9B0CA50A101E519ULL, "art\\vfx\\creatures\\fxc_ogre_db_a_c.anb"},
	{0xD9B101EAF798A113ULL, "356697490.ogg"},
	{0xD9B175EA3AA9C1E8ULL, "hlac_lt_0_lowtown_docks_-131075_6x.dds"},
	{0xD9B181DF42E7D26CULL, "6055621_f.fxe"},
	{0xD9B28CED7183D565ULL, "338093223.ogg"},
	{0xD9B2DF79EBB1F511ULL, "vxlt_warehouse_n_va_0_0.vlm"},
	{0xD9B302E5B8E7863CULL, "vxw_bonepitdestroyed_a3_1it_0_0.vlm"},
	{0xD9B3A9AF415751E9ULL, "vxlt_0_lowtown_a3_n_c_bxr_0_0.vlm"},
	{0xD9B40DCCEA88A703ULL, "237572092.ogg"},
	{0xD9B45BFF750F5083ULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpa.msh"},
	{0xD9B469FF750F68DDULL, "art\\characters\\playercharacter\\humanmale\\hlf_tmpa.mmh"},
	{0xD9B4735C9419A977ULL, "ser200_aveline6201519.cl"},
	{0xD9B4E419D6C8BC1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_right.ani"},
	{0xD9B5527356894067ULL, "6113585_m.fxe"},
	{0xD9B57AF68D3C70B6ULL, "133073933.ogg"},
	{0xD9B615AA8A76BF73ULL, "766482679.ogg"},
	{0xD9B66EA8FC209D36ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.dura.evt"},
	{0xD9B6B2260AA8064AULL, "mcr310_sibling6076307.cl"},
	{0xD9B78C803F632BDFULL, "6080627_m.fxe"},
	{0xD9B7D84B5CF7A0E2ULL, "6087761_f.fxe"},
	{0xD9B7D8C7B110CB7EULL, "6180687_m.fxe"},
	{0xD9B7F9AF115071B2ULL, "615906120.ogg"},
	{0xD9B81B8C552F3402ULL, "vxw_bonepitmining_a2_29o_0_0.vlm"},
	{0xD9B82F9A7CC0162EULL, "6165762_f.fxe"},
	{0xD9B8B622FA94A5AEULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb04a.phy"},
	{0xD9B91BB51A10AC98ULL, "67805949.ogg"},
	{0xD9B955778BA44AD7ULL, "vxlt_undercityserialkiller_15u_0_0.vlm"},
	{0xD9B963A57BB5D134ULL, "zz_followers_banter6158913.cl"},
	{0xD9B9AEAFE5A6509DULL, "6161601_m.fxe"},
	{0xD9BA17EE3239E196ULL, "hlac_lt_0_lowtown_a2_-65536_7x.dds"},
	{0xD9BA298678E144BEULL, "mindconj_distor2.mao"},
	{0xD9BA40CE1FD8DA9EULL, "6154303_m.fxe"},
	{0xD9BAC64E90246A8EULL, "zz_isa_debug6133216.cl"},
	{0xD9BBEBEE740B6FB7ULL, "vxht_mansion2b_2px_0_0.vlm"},
	{0xD9BC13720024B911ULL, "357494920.ogg"},
	{0xD9BC38F16C2F1376ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_005a_0t.dds"},
	{0xD9BC6D76415813D6ULL, "dae000_merchant_generic_hm6172597.cl"},
	{0xD9BC8E9AD651D86FULL, "579868134.ogg"},
	{0xD9BCF5201D845D94ULL, "6169604_m.fxe"},
	{0xD9BD30C3B1B213A3ULL, "dae000_sandal6147768.cl"},
	{0xD9BD65A8FC26ACC1ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.dura.gad"},
	{0xD9BD6C0CCF5816FAULL, "39201184.ogg"},
	{0xD9BD6E29B7BD1CCCULL, "308916601.ogg"},
	{0xD9BD90FF809B97AAULL, "6163756_m.fxe"},
	{0xD9BD9F8EEC8B3D33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\idle.evt"},
	{0xD9BDA2DD8FE81635ULL, "vxdr_primevaltomb_86c_0_0.vlm"},
	{0xD9BDD39A5B2C086AULL, "6086949_m.fxe"},
	{0xD9BDF590B546B151ULL, "422167254.ogg"},
	{0xD9BE3DE2C783D90CULL, "plt_zz_dogpt_debug.plo"},
	{0xD9BE7A7D42C19853ULL, "6070879_m.fxe"},
	{0xD9BE9B3B800D3653ULL, "437674268.ogg"},
	{0xD9BEB252B33FBD50ULL, "6098333_m.fxe"},
	{0xD9BEB4019DDDB8CBULL, "hm_nuncio_guard.mor"},
	{0xD9BEC1D836899FCBULL, "gen00fl_carver6097326.cl"},
	{0xD9BED8A811C24686ULL, "dre000_dougal6129789.cl"},
	{0xD9BEFEF65DCAD638ULL, "vxht_mansion2_2ze_0_0.vlm"},
	{0xD9BF37DFD03CF844ULL, "vxlt_0_lowtown_a3_n_c_euo_0_0.vlm"},
	{0xD9BF65D50F0BF25FULL, "vxht_mansion_n_6j1_0_0.vlm"},
	{0xD9BF8E2661D6E5BCULL, "371324224.ogg"},
	{0xD9BFBD1955F12B90ULL, "statecnt_f_anders.gda"},
	{0xD9BFD9E68714FF9EULL, "6123423_f.fxe"},
	{0xD9C029C0E669138CULL, "lt_0_lowtown_docks_n_39_0_ll.mmh"},
	{0xD9C062ADB0F8FA9CULL, "w_sundermount_66_0.anb"},
	{0xD9C09847A663AE34ULL, "zz_lgt_1_debug6114356.cl"},
	{0xD9C0EDF8624F878AULL, "103689610.ogg"},
	{0xD9C1256DF8AE7D4FULL, "follower_banter6136666.cl"},
	{0xD9C139D6951DC02DULL, "gen00fl_aveline6132844.cl"},
	{0xD9C1A00CEE733C11ULL, "vxca_cavern1_haunted_j23_0_2.vlm"},
	{0xD9C2434FB3BF1D35ULL, "28557886.ogg"},
	{0xD9C26482B9A2AE7AULL, "lt_0_lowtown_a3_n_c_11_0.anb"},
	{0xD9C28D3CA9C7EECBULL, "220554311.ogg"},
	{0xD9C2CDB74632E764ULL, "6176428_m.fxe"},
	{0xD9C3938F44456986ULL, "6128797_m.fxe"},
	{0xD9C39E71ABF9BFDEULL, "character_character_p223.psh.fxo.dx11"},
	{0xD9C3BF496A9DA383ULL, "vxca_caverns2_7z8_0_0.vlm"},
	{0xD9C3EDB34DC2D27DULL, "hlac_lt_0_lowtown_n_-65538_0y.dds"},
	{0xD9C3EDCAF8FB9832ULL, "1913299.ogg"},
	{0xD9C406E9D3357420ULL, "vxca_cavern1_haunted_e89_0_0.vlm"},
	{0xD9C466A8FC2CCD6FULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.dura.ani"},
	{0xD9C4988EEC915074ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\idle.gad"},
	{0xD9C4F5517A9E7519ULL, "vxht_chantry_n_35w_0_0.vlm"},
	{0xD9C5549E4C626BFEULL, "gen_im_arm_cht_lgt_act0_01.uti"},
	{0xD9C5704A4B88D67EULL, "6103946_f.fxe"},
	{0xD9C5D497AEB64F55ULL, "m_dae351ar_mount_caverns.wbd"},
	{0xD9C68B1C34328CD0ULL, "6103068_m.fxe"},
	{0xD9C6D1C1A2DFC3CDULL, "lt_0_lowtown_a2_44_0_lc.mmh"},
	{0xD9C6E99858E51471ULL, "6102829_f.fxe"},
	{0xD9C6F98A5D4391B2ULL, "vxlt_gallowsprison_n_67h_0_0.vlm"},
	{0xD9C7A04B94F17A90ULL, "391000768.ogg"},
	{0xD9C7BD2E32EA8E1DULL, "6069880_m.fxe"},
	{0xD9C7D0ABCDB0D519ULL, "6046159_m.fxe"},
	{0xD9C88A22F73721B1ULL, "dae000cr_grand_cleric6148153.cl"},
	{0xD9C8C0E1FA891B39ULL, "393987426.ogg"},
	{0xD9C919A536070F1BULL, "w_blightlands_22_0.msh"},
	{0xD9C92BA536072E66ULL, "w_blightlands_22_0.mao"},
	{0xD9C937A536074205ULL, "w_blightlands_22_0.mmh"},
	{0xD9C93DAAFAC038A2ULL, "art\\vfx\\combat\\vfx_m_sph_groupheal_c.mmh"},
	{0xD9CA19C618208060ULL, "llac_lt_darktown_-1_68.dds"},
	{0xD9CA212049E41C29ULL, "6087624_m.fxe"},
	{0xD9CA5095C2C219C9ULL, "vxca_caverns1_foc_0_0.vlm"},
	{0xD9CA7211D45465A9ULL, "19117107.ogg"},
	{0xD9CA8C2BD7F70260ULL, "vxht_keep_7a1_0_0.vlm"},
	{0xD9CAE5E7D145E43DULL, "14869315.ogg"},
	{0xD9CAE76D2712D796ULL, "6039998_m.fxe"},
	{0xD9CB070DB56768DBULL, "vxlt_0_lowtown_docks_bsx_0_0.vlm"},
	{0xD9CB1BC41FE9D13EULL, "vxlt_0_lowtown_a2_ffg_0_0.vlm"},
	{0xD9CB5C410DE70D83ULL, "6102485_f.fxe"},
	{0xD9CBA78EEC978852ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\idle.ani"},
	{0xD9CBCF3752A1AB86ULL, "mag101_alain6053010.cl"},
	{0xD9CC8F6B9BB8ECB3ULL, "210263151.ogg"},
	{0xD9CD14345AC6B38CULL, "6046972_m.fxe"},
	{0xD9CD6DC1E1E818F0ULL, "hlac_ht_keep_3732_7y.dds"},
	{0xD9CDA9BCC23310C4ULL, "vxlt_0_lowtown_a3_n_c_exl_0_0.vlm"},
	{0xD9CDCE61BF54731CULL, "vxht_keep_n_63q_0_0.vlm"},
	{0xD9CDD7B201020ADFULL, "vxht_mansion_n_6m0_0_0.vlm"},
	{0xD9CE208C075B48CFULL, "vxlt_undercityencounter_13w_0_0.vlm"},
	{0xD9CE515C35D4C81EULL, "558855420.ogg"},
	{0xD9CE53053D5F5F37ULL, "art\\vfx\\cinematic\\vfx_lightbeam01_p.dura.gad"},
	{0xD9CF6F68118814A2ULL, "773068706.ogg"},
	{0xD9CF959F5C87CAA0ULL, "plt_cod_itm_bloom.plo"},
	{0xD9CFB5F31C81DA2FULL, "vxw_bonepitmining_a2_205_0_0.vlm"},
	{0xD9D0230A1EB55E83ULL, "vxlt_0_lowtown_docks_a3_cwr_0_0.vlm"},
	{0xD9D048E051202C7DULL, "1020512033.ogg"},
	{0xD9D071F55BD13455ULL, "llac_lt_0_lowtown_docks_a3_65534_3.dds"},
	{0xD9D0D65F44B8B108ULL, "w_woundedcoast_223_0.mmh"},
	{0xD9D0E25F44B8C5ABULL, "w_woundedcoast_223_0.mao"},
	{0xD9D0F45F44B8E432ULL, "w_woundedcoast_223_0.msh"},
	{0xD9D1185962C960DDULL, "mrl000_marethari6129305.cl"},
	{0xD9D1474D6060E76BULL, "ser221_party_comments6111560.cl"},
	{0xD9D1A1AEF5182EEFULL, "vxlt_0_lowtown_docks_a3_n_coq_0_0.vlm"},
	{0xD9D227C184D151A2ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.impa.gad"},
	{0xD9D2421ED88A6283ULL, "gen00fl_anders6135992.cl"},
	{0xD9D243765C4F7FEAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebclosed.mao"},
	{0xD9D247765C4F8639ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebclosed.mmh"},
	{0xD9D251765C4F9727ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomebclosed.msh"},
	{0xD9D296210872126DULL, "lt_0_lowtown_docks_120_0.mmh"},
	{0xD9D29A21087218BEULL, "lt_0_lowtown_docks_120_0.mao"},
	{0xD9D2A821087230F3ULL, "lt_0_lowtown_docks_120_0.msh"},
	{0xD9D2E11B702C7C7EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\vst_channel_n.dds"},
	{0xD9D314E32A5B5F72ULL, "narrative_varric6172543.cl"},
	{0xD9D3672E6C948D99ULL, "vxht_chantry_n_32v_0_0.vlm"},
	{0xD9D3D3E59A5108EBULL, "follower_banter6160816.cl"},
	{0xD9D4099F2A2419EFULL, "6042280_f.fxe"},
	{0xD9D437A10E8E2545ULL, "vxlt_warehouse_n_10h_0_0.vlm"},
	{0xD9D44094C8494A4BULL, "dae000_madam_lusine6152173.cl"},
	{0xD9D468053D63EEC4ULL, "art\\vfx\\cinematic\\vfx_lightbeam01_p.dura.evt"},
	{0xD9D4957D48C4B371ULL, "6147847_m.fxe"},
	{0xD9D49E01106C229DULL, "6097073_f.fxe"},
	{0xD9D4A203EC117E39ULL, "6038837_m.fxe"},
	{0xD9D4B3D3CD590321ULL, "6035144_m.fxe"},
	{0xD9D52EBD5D262FA6ULL, "vxlt_0_lowtown_a2_hih_0_0.vlm"},
	{0xD9D5393D80D4320DULL, "6043546_f.fxe"},
	{0xD9D550981CD09A0CULL, "vxht_0_hightown_n_c1j_0_0.vlm"},
	{0xD9D6409CB11692CFULL, "lt_0_lowtown_docks_n_193_0.mao"},
	{0xD9D64C9CB116A76CULL, "lt_0_lowtown_docks_n_193_0.mmh"},
	{0xD9D6529CB116B15EULL, "lt_0_lowtown_docks_n_193_0.msh"},
	{0xD9D66026CDB607DEULL, "hlac_lt_0_lowtown_131070_5z.dds"},
	{0xD9D7323E500E1C0CULL, "6042859_m.fxe"},
	{0xD9D84BF954FC274AULL, "31283050.ogg"},
	{0xD9D86C1F3F8DF098ULL, "hlac_ht_keep_n_3736_1x.dds"},
	{0xD9D8AE46046E82B9ULL, "hlac_ht_mansion2_818_11y.dds"},
	{0xD9D8D76D40F98585ULL, "289883757.ogg"},
	{0xD9D94AC184D7AB99ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.impa.evt"},
	{0xD9D9A188DBBBEC44ULL, "vxlt_0_lowtown_gos_0_0.vlm"},
	{0xD9DA2FA2F85E3B17ULL, "vxdr_primevalentrance_397_0_0.vlm"},
	{0xD9DA76AA0C20FE41ULL, "655445194.ogg"},
	{0xD9DB24293414E391ULL, "42686012.ogg"},
	{0xD9DB823AD5478585ULL, "vxw_encounter1_j7_0_1.vlm"},
	{0xD9DB96AC38EF77BBULL, "6105490_m.fxe"},
	{0xD9DCE04BF3CC59FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_pose.ani"},
	{0xD9DD41B8826D9D39ULL, "hlac_ht_brothel_n_3366_0z.dds"},
	{0xD9DD89AEA63670ABULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medb_0n.dds"},
	{0xD9DDA4396DB3D08FULL, "90304818.ogg"},
	{0xD9DDAD649EB3F056ULL, "hlac_lt_0_lowtown_a2_-65537_4z.dds"},
	{0xD9DDB831EC8A8BDFULL, "vxlt_undercitydungeon_12i_0_0.vlm"},
	{0xD9DDF77186E14A9CULL, "vxlt_0_lowtown_bxo_0_0.vlm"},
	{0xD9DE0C7AF127CCDBULL, "mcr370cs_rules_surv_f.cut"},
	{0xD9DE3FD2908D605CULL, "6118352_m.fxe"},
	{0xD9DE822BACD7A8CFULL, "hlac_ca_caverns1_mines_24425_1y.dds"},
	{0xD9DE89FEB1A609ABULL, "ave200_donnic6112133.cl"},
	{0xD9DED9A3A52D4796ULL, "art\\characters\\creatures\\strider\\animation\\spit.evt"},
	{0xD9DEEE3738AF9E96ULL, "488909437.ogg"},
	{0xD9DF163F8A716D66ULL, "qun120cr_javaris6042878.cl"},
	{0xD9DF5DE600BBB04AULL, "t3_ogr_b02.tnt"},
	{0xD9DFA724BFDE7389ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunksingle.phy"},
	{0xD9DFC645CF68F68AULL, "896376982.ogg"},
	{0xD9DFC67D0CB5DFEEULL, "vxdr_primevaltomb_7sp_0_0.vlm"},
	{0xD9DFDBCBAB470F64ULL, "vxlt_0_lowtown_e2z_0_0.vlm"},
	{0xD9E009B0B8202691ULL, "790806061.ogg"},
	{0xD9E08F2F933DFB7CULL, "vxca_cavern1_haunted_ebk_0_0.vlm"},
	{0xD9E0F449B61B2FF5ULL, "834478498.ogg"},
	{0xD9E20278CA338427ULL, "6037089_m.fxe"},
	{0xD9E2716277905934ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hlf_meda_0s.dds"},
	{0xD9E280053D7054BDULL, "art\\vfx\\cinematic\\vfx_lightbeam01_p.dura.ani"},
	{0xD9E292CE1308AA39ULL, "vxw_woundedcoast_1b3_0_0.vlm"},
	{0xD9E2A9D0E6BA79E7ULL, "33891658.ogg"},
	{0xD9E2D48367024C47ULL, "554641307.ogg"},
	{0xD9E3075A3ABACBF1ULL, "6140947_m.fxe"},
	{0xD9E31EDA586EF118ULL, "one300_murder_followers6086896.cl"},
	{0xD9E36CFEA58C72F1ULL, "plt_cod_ltr_lother_miriam.plo"},
	{0xD9E38F210880792AULL, "lt_0_lowtown_docks_120_0.phy"},
	{0xD9E3A6CB2F409679ULL, "6128066_m.fxe"},
	{0xD9E3CB2387CD12A6ULL, "6127583_m.fxe"},
	{0xD9E45E70E6B25155ULL, "832216667.ogg"},
	{0xD9E4681D3164A753ULL, "vxlt_0_lowtown_docks_b7u_0_1.vlm"},
	{0xD9E482778136383AULL, "w_blightlands_216_0.mmh"},
	{0xD9E483233B6C91A9ULL, "qun211_braggart6075709.cl"},
	{0xD9E4867781363F71ULL, "w_blightlands_216_0.mao"},
	{0xD9E4947781365680ULL, "w_blightlands_216_0.msh"},
	{0xD9E4EF6AEB174387ULL, "477750496.ogg"},
	{0xD9E4F278AA09CF1FULL, "zz_vault_debug6203306.cl"},
	{0xD9E50BC65B3C59B5ULL, "vxht_keep_n_4at_0_0.vlm"},
	{0xD9E5A09DC935587CULL, "dae000_mother6136905.cl"},
	{0xD9E5AC2143BA3CDEULL, "lowtown_male_alienage1_sta6162754.cl"},
	{0xD9E5D0A3A5335721ULL, "art\\characters\\creatures\\strider\\animation\\spit.gad"},
	{0xD9E623025FDAE7ADULL, "hlac_lt_0_lowtown_a2_n_-196608_4x.dds"},
	{0xD9E6378AA2816354ULL, "432464351.ogg"},
	{0xD9E64DFF35491F66ULL, "490104103.ogg"},
	{0xD9E652C184E24388ULL, "art\\vfx\\combat\\vfx_m_bld_deathmagic_proj_c.impa.ani"},
	{0xD9E6AE414EB29BB9ULL, "zz_ser_debug6078575.cl"},
	{0xD9E6B8308383FBB8ULL, "6167264_m.fxe"},
	{0xD9E6F29C82C53B6CULL, "ht_0_hightownexplosion_n_17_0.rml"},
	{0xD9E7285A567AC7DAULL, "zz_followers_banter6159237.cl"},
	{0xD9E72B9DD7837AC2ULL, "hlac_lt_0_lowtown_docks_131071_1x.dds"},
	{0xD9E7CE99E46A72BBULL, "510724420.ogg"},
	{0xD9E7E2D90F4CA49BULL, "vxlt_0_lowtown_a2_n_c_gys_0_0.vlm"},
	{0xD9E7F9ACC19826F3ULL, "llac_w_encounter2_camp_-65537_0.dds"},
	{0xD9E81AC7387C4141ULL, "6141841_m.fxe"},
	{0xD9E9722B27AF88FDULL, "6143865_m.fxe"},
	{0xD9E9913F8CA15292ULL, "vxht_mansion2b_1m9_0_0.vlm"},
	{0xD9E9984FB3E59C38ULL, "vxlt_hangedman_58_0_0.vlm"},
	{0xD9E9D4CBA725B0DCULL, "vxca_caverns2_83f_0_1.vlm"},
	{0xD9EA20AEEA72562CULL, "6045630_m.fxe"},
	{0xD9EA716A15EB7736ULL, "dae000_merchant_weapons_dm6164396.cl"},
	{0xD9EA98D0FE9C01B9ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em05.msh"},
	{0xD9EAA2D0FE9C12A7ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em05.mmh"},
	{0xD9EAB0187E3703F9ULL, "234166416.ogg"},
	{0xD9EAC54358CEADB4ULL, "751747799.ogg"},
	{0xD9EB09F9C60170CAULL, "6099310_m.fxe"},
	{0xD9EB1BD0960BDEC6ULL, "6097631_m.fxe"},
	{0xD9EB6782B9C5B82DULL, "lt_0_lowtown_a3_n_c_11_0.mmh"},
	{0xD9EB6B82B9C5BF7EULL, "lt_0_lowtown_a3_n_c_11_0.mao"},
	{0xD9EB7982B9C5D6B3ULL, "lt_0_lowtown_a3_n_c_11_0.msh"},
	{0xD9EBBAB349FAB765ULL, "gen_im_arm_cht_elf_comb_t5.uti"},
	{0xD9EBBDD8789AC714ULL, "vxht_0_hightown_a3_a4m_0_0.vlm"},
	{0xD9EBE756DEEE19C9ULL, "637546765.ogg"},
	{0xD9ECB1AF828378DCULL, "6117252_m.fxe"},
	{0xD9ECD1A3A53977CFULL, "art\\characters\\creatures\\strider\\animation\\spit.ani"},
	{0xD9EE8568D9045FEFULL, "vxlt_0_lowtown_docks_a3_n_cus_0_0.vlm"},
	{0xD9EE872BD5A51DEFULL, "hlac_lt_backalley_g_0_17y.dds"},
	{0xD9EED5666C890412ULL, "llac_ht_0_hightown_-131074_0.dds"},
	{0xD9EF0E29E5E5E485ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_harrowing_n.dds"},
	{0xD9EF2728F1A94C54ULL, "one220_yevhen6098935.cl"},
	{0xD9EF52C750EA84EEULL, "331217450.ogg"},
	{0xD9EF6C3078193610ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloorb.msh"},
	{0xD9EF783736EBB179ULL, "833587009.ogg"},
	{0xD9EF7A3078194E2AULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloorb.mmh"},
	{0xD9EF7E3078195561ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamfloorb.mao"},
	{0xD9EF908CBA56B139ULL, "vfx_steamsml.wbd"},
	{0xD9F067C771FD460DULL, "w_blightlands_182_0.phy"},
	{0xD9F09C41BD220CFDULL, "aip_templarmeredith_base.gda"},
	{0xD9F09DA28A946D98ULL, "ser221st_gamlen_mother.stg"},
	{0xD9F104AB04FEC2B9ULL, "vxw_woundedcoast_1m2_0_0.vlm"},
	{0xD9F1359CA59523D6ULL, "dae000_bartrand6027826.cl"},
	{0xD9F180048FBC4D6DULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_htwood_d.dds"},
	{0xD9F1A645AE48E504ULL, "189754314.ogg"},
	{0xD9F1AEF39CC50932ULL, "932352262.ogg"},
	{0xD9F1BF1C61AE4098ULL, "vxca_caverns1_mines_e8q_0_0.vlm"},
	{0xD9F1EF2B30FDD7A9ULL, "6144248_m.fxe"},
	{0xD9F1F94BF3DEE078ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\kneel_defeated_pose.gad"},
	{0xD9F2366318B129DEULL, "dae300_bran6143508.cl"},
	{0xD9F23CAAFAE33C49ULL, "art\\vfx\\combat\\vfx_m_sph_groupheal_c.anb"},
	{0xD9F2F5B54432CBE2ULL, "character_character_p27.psh.fxo.dx11"},
	{0xD9F3301B95A9EEE6ULL, "vxca_cavern1_haunted_e4k_0_0.vlm"},
	{0xD9F3CD102E67C4E1ULL, "fen210_fenris6109412.cl"},
	{0xD9F42333F31ADFBBULL, "gen00fl_fenris6126701.cl"},
	{0xD9F42AE271B8821EULL, "vxw_woundedcoast_1pr_0_0.vlm"},
	{0xD9F4A55E9C73F5B1ULL, "38941091.ogg"},
	{0xD9F4AD15E7A602D5ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgtp_0n.dds"},
	{0xD9F4EB3E462E39F0ULL, "w_bonepitdestroyed_a3_150_0.msh"},
	{0xD9F4F93E462E510AULL, "w_bonepitdestroyed_a3_150_0.mmh"},
	{0xD9F4FD3E462E5841ULL, "w_bonepitdestroyed_a3_150_0.mao"},
	{0xD9F52A8178421A67ULL, "spider.dds"},
	{0xD9F56782B9CE0CA6ULL, "lt_0_lowtown_a3_n_c_11_0.rml"},
	{0xD9F5908F8FC7E303ULL, "6068860_f.fxe"},
	{0xD9F63E2A190D83AFULL, "mcr370_meredith6084424.cl"},
	{0xD9F7817F10AB154BULL, "vxht_keep_n_6j9_0_0.vlm"},
	{0xD9F790F96D658D51ULL, "hlac_ht_mansion2b_818_36x.dds"},
	{0xD9F7E169BA369AF2ULL, "static_static_p100.psh.fxo.dx11"},
	{0xD9F8B0F8479B7921ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_rf.ani"},
	{0xD9F8B522FACA8871ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb04a.mmh"},
	{0xD9F8CEE4671A4AF5ULL, "724449168.ogg"},
	{0xD9F8CF22FACAB42FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bb04a.msh"},
	{0xD9F8E8158F67736EULL, "llac_lt_0_lowtown_a2_n_c_65534_3.dds"},
	{0xD9F92A9A5E962E13ULL, "6005295_m.fxe"},
	{0xD9F93ADB71E5F55FULL, "llac_lt_0_lowtown_65534_5.dds"},
	{0xD9F9789A4A720580ULL, "art\\characters\\weapons\\axes\\w_axe_01a.mao"},
	{0xD9F9793FF0ABF331ULL, "fx_bloodsm.mao"},
	{0xD9F9A8E6F543514CULL, "llac_w_encounter3_-262143_0.dds"},
	{0xD9F9D358F2600500ULL, "413410672.ogg"},
	{0xD9F9EB40483CB73EULL, "vxht_keep_73i_0_0.vlm"},
	{0xD9FA6AADB12A80A4ULL, "w_sundermount_66_0.phy"},
	{0xD9FA75F1B8BF9DA2ULL, "isa111st_courtyard.stg"},
	{0xD9FA9D36F8CAAC77ULL, "follower_barks6153408.cl"},
	{0xD9FAAE833A4A930EULL, "126787685.ogg"},
	{0xD9FBE381B7FF7B7AULL, "5038088.ogg"},
	{0xD9FC2395F53C3CC3ULL, "hlac_w_bonepitdestroyed_a3_196609_0z.dds"},
	{0xD9FC23CA68EB19DAULL, "6118826_m.fxe"},
	{0xD9FC36D8B7C5D2ABULL, "6135656_f.fxe"},
	{0xD9FC49D74E65FCBBULL, "6111323_m.fxe"},
	{0xD9FC6082B9D41FEAULL, "lt_0_lowtown_a3_n_c_11_0.phy"},
	{0xD9FC96C69105ABEBULL, "hlac_lt_0_lowtown_n_131071_0x.dds"},
	{0xD9FCA353CF96D326ULL, "seb201_lady_harimann6177309.cl"},
	{0xD9FD07D34F2BBBC6ULL, "and100_anders6096016.cl"},
	{0xD9FD3BC5B2F30B89ULL, "vxht_keep_n_73q_0_0.vlm"},
	{0xD9FD775B78B990A1ULL, "hlac_lt_0_lowtown_a2_n_-3_7x.dds"},
	{0xD9FE009E726E05B0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphanga.msh"},
	{0xD9FE0E9E726E1DCAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphanga.mmh"},
	{0xD9FE129E726E2401ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarphanga.mao"},
	{0xD9FE91D4141902B3ULL, "966066594.ogg"},
	{0xD9FF6A4009152400ULL, "lgt260_eau_tmp_dockworker6143605.cl"},
	{0xD9FF7687F6F4DB39ULL, "vxw_woundedcoast_1h1_0_0.vlm"},
	{0xD9FF80C3C76A8AA0ULL, "vxlt_undercitydungeon_1ao_0_0.vlm"},
	{0xD9FFBBF847A1AAC3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_rf.gad"},
	{0xD9FFC2E2A8A0599EULL, "6108760_m.fxe"},
	{0xD9FFFF825443E30CULL, "87666043.ogg"},
	{0xDA006DB3AB0DF582ULL, "283309272.ogg"},
	{0xDA0117ED5D67CC37ULL, "hlac_lt_0_lowtown_docks_a3_-65539_1x.dds"},
	{0xDA01551B11A9D9F0ULL, "vxht_playershouse_n_1ps_0_1.vlm"},
	{0xDA0161ADB130902CULL, "w_sundermount_66_0.rml"},
	{0xDA021517672F40C8ULL, "hlac_ca_caverns1_24423_3y.dds"},
	{0xDA02CC4412D337DAULL, "vxlt_darktown_5kh_0_0.vlm"},
	{0xDA02DA57E61B7AD1ULL, "6147137_m.fxe"},
	{0xDA02DEDD3F9B10E7ULL, "hightown_fem_noble2_mob6169536.cl"},
	{0xDA040A1AB90D52A2ULL, "hlac_ht_playershouse_n_855_27x.dds"},
	{0xDA045C564A97651FULL, "vxlt_0_lowtown_n_gon_0_1.vlm"},
	{0xDA046F0BEFE94EA0ULL, "6084484_m.fxe"},
	{0xDA06267685B73529ULL, "follower_banter6136677.cl"},
	{0xDA0632721AB2D62EULL, "6031266_m.fxe"},
	{0xDA06474144002FF0ULL, "6115271_f.fxe"},
	{0xDA0685FE3652D35AULL, "dae000_bodhan6129081.cl"},
	{0xDA0698F847A78DA8ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_rf.evt"},
	{0xDA06D579E8424EDEULL, "281805763.ogg"},
	{0xDA0712D2A22A5301ULL, "dae000_arishok6074060.cl"},
	{0xDA07580878A53BEBULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.impa.ani"},
	{0xDA0766F46BB10715ULL, "hlac_dr_dungeon_3216_7x.dds"},
	{0xDA079935ECE622EBULL, "hlac_ca_caverns1_24421_1z.dds"},
	{0xDA088D535493A4AEULL, "344028256.ogg"},
	{0xDA092513AD5226A8ULL, "plt_cod_ltr_junk_bank_offer.nss"},
	{0xDA0937A5363D260EULL, "w_blightlands_22_0.rml"},
	{0xDA09678A375BA535ULL, "vxca_cavern1_haunted_i1o_0_0.vlm"},
	{0xDA0A95A75AE1325AULL, "6117926_m.fxe"},
	{0xDA0B68314BBB57E0ULL, "plt_vil00pt_reactivity.nss"},
	{0xDA0B9024C00387AEULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunksingle.mmh"},
	{0xDA0B9C24C0039C45ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunksingle.mao"},
	{0xDA0BAA24C003B39CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunksingle.msh"},
	{0xDA0C298BB98F3943ULL, "414490153.ogg"},
	{0xDA0C2B3497F3F94AULL, "vxca_caverns1_edz_0_0.vlm"},
	{0xDA0C2E0A0A09618FULL, "vxw_bonepitmining_a2_263_0_0.vlm"},
	{0xDA0C8C98F16FA964ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgta_0t.dds"},
	{0xDA0CC39CF06DFAF8ULL, "6051312_m.fxe"},
	{0xDA0DE864E8EAF3B9ULL, "vxw_woundedcoast_1s0_0_0.vlm"},
	{0xDA0E5BB39CD801A9ULL, "6161102_m.fxe"},
	{0xDA0F30F2EE63A5C6ULL, "502831782.ogg"},
	{0xDA0F51D9C81DB160ULL, "630179152.ogg"},
	{0xDA0F5612C29A21FDULL, "w_bonepit_143_0.mao"},
	{0xDA0F5A12C29A2836ULL, "w_bonepit_143_0.mmh"},
	{0xDA0F6412C29A3934ULL, "w_bonepit_143_0.msh"},
	{0xDA102E13AD5854BBULL, "plt_cod_ltr_junk_bank_offer.plo"},
	{0xDA109215B54EF51AULL, "6111601_m.fxe"},
	{0xDA10CCC7C2D731DBULL, "140817819.ogg"},
	{0xDA10D4B9C38A47DBULL, "character_character_p24.psh.fxo.dx11.assembly"},
	{0xDA1108318A7DF7F5ULL, "168493330.ogg"},
	{0xDA11341C80521BAAULL, "vxlt_0_lowtown_a2_n_el9_0_0.vlm"},
	{0xDA113E2104C9505AULL, "vxlt_darktown_5li_0_0.vlm"},
	{0xDA11655955FAF215ULL, "390790232.ogg"},
	{0xDA118076C92A8409ULL, "6183960_m.fxe"},
	{0xDA11F3EDA6DF2926ULL, "vxht_0_hightown_a2_b9n_0_0.vlm"},
	{0xDA11FE469E13113CULL, "one120_athenril6088313.cl"},
	{0xDA12B5F9255FB31FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_darren.gad"},
	{0xDA12D79669E656AAULL, "vxlt_warehouse_vo_0_0.vlm"},
	{0xDA1402AA62587B7EULL, "6170415_m.fxe"},
	{0xDA142C506749F641ULL, "lt_0_lowtown_a2_n_c_33_0_ll.mmh"},
	{0xDA14400878AF9D62ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.impa.evt"},
	{0xDA14EC1ED41A2E16ULL, "554531159.ogg"},
	{0xDA1510F3F2A5B0E3ULL, "596962439.ogg"},
	{0xDA151E31ACE92E78ULL, "qun131_party_comment6117265.cl"},
	{0xDA15690C32F44BE7ULL, "zz_qun_debug6028273.cl"},
	{0xDA15787E62393E8BULL, "mag330_walter6044534.cl"},
	{0xDA159C90CD3A912DULL, "hlac_ht_0_hightown_-65537_4y.dds"},
	{0xDA15E3A20B2E28A5ULL, "1066900054.ogg"},
	{0xDA1613C675091E7AULL, "6076627_m.fxe"},
	{0xDA16701DBE73DCB5ULL, "vxw_bonepitdestroyed_a3_1bh_0_0.vlm"},
	{0xDA167782F74E7D88ULL, "272790544.ogg"},
	{0xDA168BD0FEC126DCULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em05.phy"},
	{0xDA169ECC48C53084ULL, "gen00fl_fenris6127109.cl"},
	{0xDA16B661552808E4ULL, "vxlt_0_lowtown_ha6_0_0.vlm"},
	{0xDA16F893F148CD4BULL, "zz_amb_invis.utp"},
	{0xDA1761D9B8AFBD38ULL, "207200003.ogg"},
	{0xDA17919250A350FFULL, "dae011_corff_sta6149609.cl"},
	{0xDA18584E968295CFULL, "follower_banter6156206.cl"},
	{0xDA1941FE9BBDF5B6ULL, "6076717_m.fxe"},
	{0xDA194DB7B342F11CULL, "hlac_w_encounter2_-65532_0z.dds"},
	{0xDA1A004A24C39A92ULL, "ts_qunari_wpn_table.gda"},
	{0xDA1A3AFCF80832A8ULL, "1033240171.ogg"},
	{0xDA1A55B684E9BB09ULL, "243599368.ogg"},
	{0xDA1A5CF4C4C352A5ULL, "6006245_m.fxe"},
	{0xDA1ADFA597C72CEFULL, "hlac_lt_0_lowtown_a2_65538_1z.dds"},
	{0xDA1B2761754D8E8BULL, "6148388_f.fxe"},
	{0xDA1B3F0878B5BA75ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.impa.gad"},
	{0xDA1BB691226791DCULL, "vxdr_primevaltomb_3m1_0_1.vlm"},
	{0xDA1C0B133B177BCAULL, "33586375.ogg"},
	{0xDA1C0FD6406BB08AULL, "6041114_f.fxe"},
	{0xDA1C1FFA8D1E1162ULL, "gen00fl_sebastian6174045.cl"},
	{0xDA1C647DAB56BBA0ULL, "vxlt_undercitydungeon_1om_0_0.vlm"},
	{0xDA1C65301ECD04B8ULL, "qun211_braggart6075687.cl"},
	{0xDA1C7AA27BA389D0ULL, "vxlt_gallowstemplar_f_650_0_0.vlm"},
	{0xDA1C83AECA06E172ULL, "hlac_lt_0_lowtown_a2_65535_5x.dds"},
	{0xDA1C87C6A51BA8B9ULL, "6055443_m.fxe"},
	{0xDA1CA434AFEDE41BULL, "560665335.ogg"},
	{0xDA1CCD908ECE1A29ULL, "6166002_m.fxe"},
	{0xDA1D4F3E2A3A0D0BULL, "6053793_m.fxe"},
	{0xDA1DB1CB8D414EF7ULL, "vxca_cave_204_0_0.vlm"},
	{0xDA1DEE3EC8BC4B59ULL, "vxlt_gallowstemplar_3nj_0_0.vlm"},
	{0xDA1E41C7816788F8ULL, "vxlt_0_lowtown_docks_n_bw0_0_0.vlm"},
	{0xDA1E8221F87620F8ULL, "282349037.ogg"},
	{0xDA1F45444BC06272ULL, "vxht_mansion2b_2jw_0_0.vlm"},
	{0xDA1FB689C25B9A76ULL, "w_woundedcoast_120_0.msh"},
	{0xDA1FC089C25BAB74ULL, "w_woundedcoast_120_0.mmh"},
	{0xDA1FC489C25BB187ULL, "w_woundedcoast_120_0.mao"},
	{0xDA2090AF4017561DULL, "131253576.ogg"},
	{0xDA211E566524E320ULL, "vxlt_0_lowtown_a3_n_c_hng_0_0.vlm"},
	{0xDA212CB4E4670BD5ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_hammer_d.dds"},
	{0xDA213DA268BDCE3DULL, "hm_gamlen_h3s3.mor"},
	{0xDA2140914483A4D5ULL, "6092557_f.fxe"},
	{0xDA219FF128342425ULL, "vxw_bonepit_28m_0_0.vlm"},
	{0xDA21AFC463E0A501ULL, "aip_slavermage_base.gda"},
	{0xDA21D8F08C9C0F05ULL, "qcr000_arishok6060186.cl"},
	{0xDA21FDA2E3BDBC98ULL, "qun160_vashothlead6064593.cl"},
	{0xDA22B774CC30F6D6ULL, "art\\vfx\\models\\vfx_disk.mmh"},
	{0xDA22C174CC3107D4ULL, "art\\vfx\\models\\vfx_disk.msh"},
	{0xDA22FA24E34688C3ULL, "hlac_w_woundedcoast_262145_0x.dds"},
	{0xDA230C67D02A8038ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_nuda.phy"},
	{0xDA23E0FE27B91790ULL, "gen00fl_sebastian6189443.cl"},
	{0xDA23EDC13B166778ULL, "6178883_f.fxe"},
	{0xDA243E4CD0E95803ULL, "llac_lt_gallowscourtyard_-65538_1.dds"},
	{0xDA24775E9C5FF2C6ULL, "hlac_lt_0_lowtown_docks_a3_65534_2x.dds"},
	{0xDA24E046FE03CD02ULL, "6149470_f.fxe"},
	{0xDA259D856C5C1997ULL, "vxca_cave_1cf_0_0.vlm"},
	{0xDA2608649691FE3BULL, "w_bonepit_32_0.msh"},
	{0xDA26162E53AE221DULL, "mag110_arianni6015764.cl"},
	{0xDA261A6496921C86ULL, "w_bonepit_32_0.mao"},
	{0xDA26266496923125ULL, "w_bonepit_32_0.mmh"},
	{0xDA264F4F6008000EULL, "follower_banter6141919.cl"},
	{0xDA26B06064897426ULL, "hlac_ht_chantry_n_2534_17x.dds"},
	{0xDA2792F92571D465ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_panhandle_2_darren.ani"},
	{0xDA2801593157AE21ULL, "mag200_feynriel6030448.cl"},
	{0xDA282A0B1CF3D954ULL, "80173843.ogg"},
	{0xDA2870FC176AA914ULL, "hlac_ht_mansion2_818_27y.dds"},
	{0xDA2885A5F0BF7B8FULL, "235066845.ogg"},
	{0xDA28DCC743809CFEULL, "llac_ca_caverns1_mines_24422_0.dds"},
	{0xDA2932F951609DB1ULL, "llac_blank_exterior_0_0y.dds"},
	{0xDA295BC6F333D622ULL, "llac_lt_0_lowtown_a2_n_c_131070_4.dds"},
	{0xDA298BFBC6ECF12BULL, "438392893.ogg"},
	{0xDA29C1AD19FE5CF8ULL, "324919817.ogg"},
	{0xDA2A2B17DEFDAEF5ULL, "fx_streak02_add.mao"},
	{0xDA2A33813712F439ULL, "zz_vault_debug6203317.cl"},
	{0xDA2A6DBFF27C94BAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture.gad"},
	{0xDA2B0675C1E1DC7AULL, "948394137.ogg"},
	{0xDA2BEFDFF2120B9BULL, "79000580.ogg"},
	{0xDA2BFF2441E3F22CULL, "llac_lt_0_lowtown_n_65534_1.dds"},
	{0xDA2C23A87F376777ULL, "vxca_cave_257_0_0.vlm"},
	{0xDA2C6E6197F5C343ULL, "708725387.ogg"},
	{0xDA2C8869D9632ECAULL, "llac_ca_caverns2prc_5802_64.dds"},
	{0xDA2D35E93F6A045FULL, "6145224_f.fxe"},
	{0xDA2DEE5A960B7734ULL, "hlac_ht_mansion2b_818_47y.dds"},
	{0xDA2E16F43B6FFAFEULL, "vxht_0_hightown_c6z_0_0.vlm"},
	{0xDA2E314D45F6DD75ULL, "hlac_lt_0_lowtown_n_65534_0x.dds"},
	{0xDA2E65455E740A1FULL, "vxht_0_hightown_b7i_0_0.vlm"},
	{0xDA2EA39664C9FFA6ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.cess.evt"},
	{0xDA2EFDE01EEFDA0CULL, "art\\characters\\creatures\\rubblegolem\\textures\\c_rbg_0e.dds"},
	{0xDA2F83AB904BCB6EULL, "233900139.ogg"},
	{0xDA2FD5D8B10AF637ULL, "908107192.ogg"},
	{0xDA304D98CA44B378ULL, "vxdr_primevaltomb_5ak_0_1.vlm"},
	{0xDA30712331480003ULL, "6153628_m.fxe"},
	{0xDA30815FFF17D508ULL, "dae000_arishok6067737.cl"},
	{0xDA30903EED76AC7DULL, "vxlt_gallowsprison_n_6gc_0_0.vlm"},
	{0xDA310A77B71EDE2BULL, "400846496.ogg"},
	{0xDA3131C51F98DA35ULL, "vxdr_primevaltomb_8nk_0_0.vlm"},
	{0xDA313C1E4D4FEAABULL, "hlac_dr_0_deeproads_5845_3z.dds"},
	{0xDA3170BFF282B8D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture.evt"},
	{0xDA318AB86BE21DF3ULL, "169869220.ogg"},
	{0xDA3191435CEF4EBBULL, "hlac_lt_undercityserialkiller_-1_34x.dds"},
	{0xDA31A009A3B6C41DULL, "6195853_m.fxe"},
	{0xDA31B63CDC1972AEULL, "vxlt_0_lowtown_a2_n_ex9_0_0.vlm"},
	{0xDA326E7F2EC4BB5DULL, "6124718_m.fxe"},
	{0xDA3270F58EA0D7F0ULL, "hlac_dr_slavers_3216_4z.dds"},
	{0xDA328D8DFE82594AULL, "18709939.ogg"},
	{0xDA32DE9E760CAC79ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopec.msh"},
	{0xDA32E89E760CBD67ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopec.mmh"},
	{0xDA32EC9E760CC3B4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopec.mao"},
	{0xDA3324C5BD53EB88ULL, "lt_0_lowtown_33_0_ll.mmh"},
	{0xDA3332A85147AF5EULL, "ht_brothel_n_mmap.dds"},
	{0xDA3353D7A2600DB5ULL, "vxw_bonepitdestroyed_a3_1hv_0_0.vlm"},
	{0xDA343FB39F1A5EE4ULL, "6038595_m.fxe"},
	{0xDA3476A282300DFEULL, "126964411.ogg"},
	{0xDA34BA9664CE92F1ULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.cess.gad"},
	{0xDA351C346C6881C1ULL, "222743120.ogg"},
	{0xDA353E05B92337F9ULL, "hlac_lt_0_lowtown_a3_n_c_-1_2z.dds"},
	{0xDA354163ADC73FEEULL, "gen_im_npc_wep_shd_dala.uti"},
	{0xDA35886A5BC90735ULL, "115461818.ogg"},
	{0xDA35ECB7628A913EULL, "vxlt_hangedman_n_nv_0_0.vlm"},
	{0xDA3757543972DCE9ULL, "hlac_lt_darktown_-1_40y.dds"},
	{0xDA37F6F5B7160C3FULL, "lt_gallowstemplar_mmap.dds"},
	{0xDA3895A3FD186C3BULL, "6075982_m.fxe"},
	{0xDA38FF6C18C8CBD5ULL, "gen00fl_fenris6113963.cl"},
	{0xDA39018C45669F23ULL, "dae000_mother6002009.cl"},
	{0xDA390E0070679293ULL, "141642454.ogg"},
	{0xDA39B0B0170F7B58ULL, "301699749.ogg"},
	{0xDA39E15B91AE92CBULL, "art\\characters\\creatures\\ragedemon\\animation\\k_bk.ani"},
	{0xDA39FA59E00017E5ULL, "546820426.ogg"},
	{0xDA3A04EE28270E49ULL, "1052398470.ogg"},
	{0xDA3A831106071DC7ULL, "6075053_f.fxe"},
	{0xDA3A9E9B951D21E6ULL, "hlac_w_bonepit_327684_0z.dds"},
	{0xDA3B37E64CA7AA1CULL, "character_character_p205.psh.fxo.dx11.assembly"},
	{0xDA3B3D650AECE256ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_whistle.ani"},
	{0xDA3B77FEA5D6FB56ULL, "plt_cod_ltr_lother_miriam.nss"},
	{0xDA3B9176F9B60BB8ULL, "vxht_0_hightown_n_can_0_0.vlm"},
	{0xDA3B9B7732C7D2BAULL, "mag370_meredith6046947.cl"},
	{0xDA3BBB9664D4B29FULL, "art\\vfx\\combat\\vfx_m_frc_slowzone_p.cess.ani"},
	{0xDA3BE9FDFB750E1AULL, "w_woundedcoast_34_0.mmh"},
	{0xDA3BEDFDFB751551ULL, "w_woundedcoast_34_0.mao"},
	{0xDA3BFBFDFB752D60ULL, "w_woundedcoast_34_0.msh"},
	{0xDA3C0B537324A365ULL, "vxlt_darktown_5aq_0_0.vlm"},
	{0xDA3C0F8157235AF9ULL, "6140352_m.fxe"},
	{0xDA3C14FCDFDF0484ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_nuda.msh"},
	{0xDA3C26FCDFDF23EDULL, "art\\characters\\playercharacter\\dwarfmale\\arm_nuda.mao"},
	{0xDA3C2AFCDFDF2A26ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_nuda.mmh"},
	{0xDA3CD8B683EBFE4CULL, "vxw_woundedcoast_153_0_0.vlm"},
	{0xDA3D9883F8A1C615ULL, "hlac_ht_bartrand_n_2995_1z.dds"},
	{0xDA3DAB521C0F8657ULL, "llac_ht_0_hightown_a2_-65537_3.dds"},
	{0xDA3E4472EFA29425ULL, "hlac_lt_darktown_-1_72x.dds"},
	{0xDA3E88BFF28D6BA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_right_hand_gesture.ani"},
	{0xDA3E8FC38F232DD8ULL, "hlac_ht_chantry_n_2534_35z.dds"},
	{0xDA3EC21086DA743DULL, "261029597.ogg"},
	{0xDA3EC4E13AB91C67ULL, "171045801.ogg"},
	{0xDA3F54D4C2BA2C17ULL, "6139493_m.fxe"},
	{0xDA3FF06C9640200CULL, "339565764.ogg"},
	{0xDA4068BF9B883663ULL, "gen00fl_anders6125412.cl"},
	{0xDA40DB0D672B5EA2ULL, "dae000_gamlen6131103.cl"},
	{0xDA40EC0300734BD4ULL, "1021004293.ogg"},
	{0xDA40EE96C37F473FULL, "6121210_m.fxe"},
	{0xDA41723B1C055836ULL, "6141529_m.fxe"},
	{0xDA41764E5758D898ULL, "1066637016.ogg"},
	{0xDA41C5B494562635ULL, "vxw_bonepitdestroyed_a3_1mu_0_0.vlm"},
	{0xDA41E6257797A28CULL, "11557397.ogg"},
	{0xDA41F1F2796AF3BCULL, "vxca_cavern1_haunted_j20_0_0.vlm"},
	{0xDA43AC51DB3D9976ULL, "llac_w_sundermount_65533_1.dds"},
	{0xDA43AC93B7D18D70ULL, "vxlt_gallowscourtyard_n_73y_0_0.vlm"},
	{0xDA44DE6BB4F36128ULL, "dr_dungeon_3214.rml"},
	{0xDA44F1D1A2E152A1ULL, "zz_followers_banter6157559.cl"},
	{0xDA45121229DB3BE5ULL, "art\\levels\\races\\proxy\\props\\_textures\\tent_d.dds"},
	{0xDA4552C61E9AD9EDULL, "6206874_m.fxe"},
	{0xDA456EB4DED907C2ULL, "vxdr_0_deeproads_gbm_0_0.vlm"},
	{0xDA457B3883E2F989ULL, "581572212.ogg"},
	{0xDA4621FCDFE76FF1ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_nuda.phy"},
	{0xDA462F958B4530B6ULL, "dae100_alistair_drunk6135092.cl"},
	{0xDA464D0C504908D1ULL, "rdrip_store.utp"},
	{0xDA47850200AFEEB2ULL, "lowtown_male_urchin_sta6164189.cl"},
	{0xDA47C95B91BAA742ULL, "art\\characters\\creatures\\ragedemon\\animation\\k_bk.evt"},
	{0xDA47ED0D4FF298E0ULL, "6152804_m.fxe"},
	{0xDA4845650AF77A47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_whistle.evt"},
	{0xDA4861314BEEF9A3ULL, "plt_vil00pt_reactivity.plo"},
	{0xDA487F3BF92FCBDDULL, "716665211.ogg"},
	{0xDA48EE8F44107EB1ULL, "llac_w_bonepit_65536_0.dds"},
	{0xDA49066B4EC6F1E2ULL, "llac_lt_0_lowtown_-65537_6.dds"},
	{0xDA499A1B715820AEULL, "gen_im_arm_hlm_med_set_med.uti"},
	{0xDA49ACB68C922656ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_4.ani"},
	{0xDA49D7924927FB18ULL, "vxlt_hangedman_20u_0_0.vlm"},
	{0xDA4A79C474725978ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_vael_0n.dds"},
	{0xDA4A7D30651ABBE5ULL, "vxlt_darktown_5jp_0_0.vlm"},
	{0xDA4A86B139E9DBD0ULL, "vxlt_0_lowtown_a2_gj0_0_0.vlm"},
	{0xDA4A8DAD56BD529DULL, "zz_vault_debug6203205.cl"},
	{0xDA4ACFBE1ACDD42EULL, "lt_0_lowtown_docks_a3_111_0.mmh"},
	{0xDA4ADBBE1ACDE8C5ULL, "lt_0_lowtown_docks_a3_111_0.mao"},
	{0xDA4AE9BE1ACE001CULL, "lt_0_lowtown_docks_a3_111_0.msh"},
	{0xDA4B1FA8AA17B24FULL, "vxw_bonepitdestroyed_a3_na_0_0.vlm"},
	{0xDA4B481D055F2D4AULL, "dae021_chantry_initiate_mob.cnv"},
	{0xDA4B4A9375E216CCULL, "vxw_woundedcoast_100_0_0.vlm"},
	{0xDA4B4E492208DB47ULL, "hlac_lt_gallowstemplar_n_2383_2x.dds"},
	{0xDA4B5B8E333ABE1BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_glv_lgta_0t.dds"},
	{0xDA4BA6EA03B9DE19ULL, "590516960.ogg"},
	{0xDA4C67FE23C056F2ULL, "6126940_m.fxe"},
	{0xDA4CC07FA68DFF79ULL, "art\\characters\\creatures\\mubari\\animation\\dieloop_v2.ani"},
	{0xDA4CD779DE21092FULL, "llac_lt_0_lowtown_a2_-131072_7.dds"},
	{0xDA4D280042405AF9ULL, "plt_genpt_party_base.nss"},
	{0xDA4D39474D51AD34ULL, "llac_lt_0_lowtown_docks_-65538_0.dds"},
	{0xDA4D73F8D162DD7DULL, "vxlt_gallowsprison_n_6ma_0_0.vlm"},
	{0xDA4D8237F21F0C98ULL, "6148384_m.fxe"},
	{0xDA4D91D42830C9C0ULL, "641271023.ogg"},
	{0xDA4DC85B91BF1155ULL, "art\\characters\\creatures\\ragedemon\\animation\\k_bk.gad"},
	{0xDA4DE203044B62BAULL, "llac_lt_darktown_-1_17.dds"},
	{0xDA4DE91F03C903F3ULL, "fen210_fenris6109436.cl"},
	{0xDA4E10C6DC56DE2BULL, "283022701.ogg"},
	{0xDA4E157F03850B35ULL, "vxdr_primevaltomb_8di_0_0.vlm"},
	{0xDA4E211E3006B3B7ULL, "lowtown_male_lowlife1_sta6165040.cl"},
	{0xDA4E2A1D05613072ULL, "dae021_chantry_initiate_mob.bnk"},
	{0xDA4E56B95FC48925ULL, "w_woundedcoast_45_0_ll.mmh"},
	{0xDA4EA6CFC4BA4DCEULL, "6208365_m.fxe"},
	{0xDA4EB8D1E0F39984ULL, "vxlt_undercitydungeon_wx_0_0.vlm"},
	{0xDA4ED2F9C5674544ULL, "935154465.ogg"},
	{0xDA4F0DE6EF41D0B9ULL, "zz_mag_debug6053410.cl"},
	{0xDA4F14E70FE7D798ULL, "art\\characters\\creatures\\strider\\animation\\blendtree\\strider_movement.anb"},
	{0xDA4F66650AFDD090ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_whistle.gad"},
	{0xDA4FE11326683BEFULL, "58971449.ogg"},
	{0xDA4FF5AFF3962B4AULL, "vxw_bonepit_10u_0_0.vlm"},
	{0xDA4FFC7C999CCAB4ULL, "hlac_lt_backalley_n_65536_0z.dds"},
	{0xDA500B687180C35EULL, "vxca_caverns2_8h0_0_0.vlm"},
	{0xDA509F5C3BAF7532ULL, "vxca_caverns1_j2f_0_1.vlm"},
	{0xDA50A768EE440AA3ULL, "vxht_chantry_n_1yo_0_0.vlm"},
	{0xDA51A3E739FC830CULL, "gen_im_acc_amu_act3_07.uti"},
	{0xDA51BB3FDDEC9D7AULL, "dae021_sister_samea_sta6142110.cl"},
	{0xDA5216321C52C73AULL, "qcr000_arishok6060103.cl"},
	{0xDA5269EE69BCD3B4ULL, "lt_0_lowtown_docks_n_31_0_ll.mmh"},
	{0xDA5288EE03E77DA9ULL, "vxlt_0_lowtown_a2_n_c_hle_0_0.vlm"},
	{0xDA52E17F6CF534C7ULL, "llac_w_bonepitmining_a2_196614_0.dds"},
	{0xDA52F0924E1291DDULL, "isa211_isabela6069412.cl"},
	{0xDA5306C4E664D7C0ULL, "vxw_woundedcoast_1mf_0_1.vlm"},
	{0xDA532DDF6EA74453ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit.evt"},
	{0xDA534A7D2CA4778BULL, "hlac_ht_0_hightown_n_1_0x.dds"},
	{0xDA536C535F16C0EEULL, "387801427.ogg"},
	{0xDA539B7FA693DEBBULL, "art\\characters\\creatures\\mubari\\animation\\dieloop_v2.gad"},
	{0xDA53AD5DD8BD95A1ULL, "428720668.ogg"},
	{0xDA53CE139AEA7C3FULL, "vxlt_gallowstemplar_f_3j7_0_0.vlm"},
	{0xDA540F0042464ED2ULL, "plt_genpt_party_base.plo"},
	{0xDA542B1024A53B04ULL, "gen00fl_varric6105376.cl"},
	{0xDA547602B620695FULL, "mag321_comtesse_de_launcet6064715.cl"},
	{0xDA548D2A020CC0C1ULL, "ser270_cut_aftermath.bnk"},
	{0xDA54ACE1E0A340AFULL, "365842391.ogg"},
	{0xDA54C676A9F4B2A0ULL, "vxlt_darktown_37l_0_0.vlm"},
	{0xDA54D742462FB52DULL, "vxht_0_hightown_n_c40_0_0.vlm"},
	{0xDA5557CDA4167D5CULL, "6118525_m.fxe"},
	{0xDA563153B3D3C0CCULL, "559391318.ogg"},
	{0xDA56B4B68C9CBE47ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_4.evt"},
	{0xDA56CD7281E853C2ULL, "rdr111dg_martin6054546.cl"},
	{0xDA570ED576BCEA6BULL, "sse_ss_fenris.gda"},
	{0xDA574A1EE88C43C5ULL, "6127409_m.fxe"},
	{0xDA57932267BB9374ULL, "vxht_chantry_2by_0_0.vlm"},
	{0xDA57AC469F3AFB2EULL, "lgt000ip_ultra_tome.ncs"},
	{0xDA57AF2A020F3031ULL, "ser270_cut_aftermath.cnv"},
	{0xDA57B9BC28C2A94BULL, "183746073.ogg"},
	{0xDA5842A113AA5D31ULL, "977218755.ogg"},
	{0xDA58B861D8ACB2ABULL, "hlac_ht_mansion2_818_49z.dds"},
	{0xDA58F90E8A3E6C36ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2windo.msh"},
	{0xDA59030E8A3E7D34ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2windo.mmh"},
	{0xDA59070E8A3E8447ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2windo.mao"},
	{0xDA593B219DB5F6ECULL, "qun131_taskmaster6019205.cl"},
	{0xDA596A52E3DD1CB6ULL, "6137582_m.fxe"},
	{0xDA5A26DF6EAD5694ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit.gad"},
	{0xDA5A6420208893F2ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_player.gad"},
	{0xDA5AE1E87D349FC9ULL, "vxlt_gallowstemplar_f_5pe_0_0.vlm"},
	{0xDA5AE5428E1E6C6AULL, "zz_mcr_debug6071537.cl"},
	{0xDA5B04A60ADB171EULL, "vxlt_darktown_5r7_0_0.vlm"},
	{0xDA5B2241F7FC1F18ULL, "6137669_m.fxe"},
	{0xDA5B62652DBFF1C9ULL, "6110058_m.fxe"},
	{0xDA5C7C3235CAA9AFULL, "246140062.ogg"},
	{0xDA5C820441667427ULL, "vxht_mansion_n_3wc_0_0.vlm"},
	{0xDA5D35BBBC4E8862ULL, "6121457_m.fxe"},
	{0xDA5D5EF499AA0E69ULL, "gen00fl_dog6107345.cl"},
	{0xDA5DB28C0BF3351FULL, "vxht_brothel_n_4c3_0_0.vlm"},
	{0xDA5DD5B68CA31490ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_4.gad"},
	{0xDA5E19B81BA1DD97ULL, "mcr370_guardsman6087661.cl"},
	{0xDA5E678CE58C43CAULL, "vxw_bonepit_15v_0_0.vlm"},
	{0xDA5E8D4D438D035EULL, "llac_lt_0_lowtown_docks_-131074_7.dds"},
	{0xDA5EC3E5D4CEADE1ULL, "zz_facefx_test6088479.cl"},
	{0xDA5ED19E7631D09CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3ww_slopec.phy"},
	{0xDA5ED2A1F83D971AULL, "291917948.ogg"},
	{0xDA5EEE628EC975E9ULL, "536897700.ogg"},
	{0xDA5F03F7E3A4D740ULL, "zz_vault_debug6177613.cl"},
	{0xDA5F06FB51CAF2F4ULL, "zz_dae_debug6142949.cl"},
	{0xDA5F3DAEAE321044ULL, "sib131_letter_react.cnv"},
	{0xDA5F6DC79F0D1246ULL, "vxlt_warehouse_8q_0_0.vlm"},
	{0xDA5FDAA5A13CE2B6ULL, "zz_qcr_debug6073544.cl"},
	{0xDA60158B86D27A70ULL, "802220411.ogg"},
	{0xDA603A128D95C65EULL, "hlac_ca_caverns1_24421_11x.dds"},
	{0xDA60476A26D1D084ULL, "42181035.ogg"},
	{0xDA610FA744B7E3E5ULL, "sst_ss_shadow_m_1.gda"},
	{0xDA6135DF6EB38F72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_exit.ani"},
	{0xDA613997321373BEULL, "vxw_encounter1_fe_0_0.vlm"},
	{0xDA61E6F4A4A5422FULL, "vxdr_0_deeproads_but_0_0.vlm"},
	{0xDA624C138E14B9F4ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgta_0d.dds"},
	{0xDA630CDC6741E546ULL, "608547604.ogg"},
	{0xDA6353AC83BB801AULL, "vxlt_0_lowtown_a2_n_c_csh_0_0.vlm"},
	{0xDA635FAEAE3632A4ULL, "sib131_letter_react.bnk"},
	{0xDA63B64259218488ULL, "vxdr_primevaltomb_7lg_0_0.vlm"},
	{0xDA640BB2DE5170F0ULL, "vxlt_0_lowtown_docks_a3_n_d62_0_0.vlm"},
	{0xDA641407BAE52BACULL, "vxht_keep_n_370_0_0.vlm"},
	{0xDA645008DB6CC72FULL, "character_character_p286.psh.fxo.dx11"},
	{0xDA6472CF27FF8FDFULL, "vxlt_0_lowtown_a3_n_c_hxn_0_0.vlm"},
	{0xDA6478D2DEEC8756ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisa_1.mao"},
	{0xDA64E3F99ACC362BULL, "753711186.ogg"},
	{0xDA65151BDB39BAFBULL, "gen_im_npc_dwf_arm_cht_comx.uti"},
	{0xDA6528F3E7ACF6B6ULL, "242103286.ogg"},
	{0xDA6576E5E388BCE6ULL, "prty_anders_armor_2.dds"},
	{0xDA65E908D29D3927ULL, "vxlt_hangedman_n_1fl_0_0.vlm"},
	{0xDA66266496C8152EULL, "w_bonepit_32_0.rml"},
	{0xDA66AFD73CEE51E8ULL, "88667168.ogg"},
	{0xDA66CE7E15F2130CULL, "6198672_f.fxe"},
	{0xDA6711330BD76B55ULL, "isa311_hawke_comments6117907.cl"},
	{0xDA6769EC16679988ULL, "lgt360_mag_orlanna.cnv"},
	{0xDA6774D3F8250685ULL, "vxlt_0_lowtown_f9h_0_1.vlm"},
	{0xDA6782572A61A8A3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_right_hand.evt"},
	{0xDA67E9FDFB9A4931ULL, "w_woundedcoast_34_0.rml"},
	{0xDA67EB5094244CB9ULL, "dre000_stroud6036711.cl"},
	{0xDA67FD8FF8737BE4ULL, "llac_lt_backalley_n_1_2.dds"},
	{0xDA686A25F27C34AFULL, "dae000_mother6035924.cl"},
	{0xDA68731C57590BA6ULL, "hlac_lt_hangedman_n_-1_17z.dds"},
	{0xDA6907264486F4A9ULL, "6144459_m.fxe"},
	{0xDA69323D3AB0D996ULL, "vxca_caverns1_mines_j28_0_2.vlm"},
	{0xDA6957C3A9BEF09AULL, "hlac_lt_backalley_n_1_5y.dds"},
	{0xDA69912098F1A09AULL, "vxca_cavern1_haunted_jgu_0_0.vlm"},
	{0xDA6A2BA54C3A5E07ULL, "706959991.ogg"},
	{0xDA6A48309928B544ULL, "vxlt_darktown_6l4_0_0.vlm"},
	{0xDA6A669F3B6E74B8ULL, "gen00fl_isabela6101541.cl"},
	{0xDA6A907943515A98ULL, "6208698_f.fxe"},
	{0xDA6A9BD5E3476169ULL, "hlac_ht_0_hightown_0_2z.dds"},
	{0xDA6B370CD591407CULL, "hlac_w_blightlands_262142_0z.dds"},
	{0xDA6B8BEC166BBCF0ULL, "lgt360_mag_orlanna.bnk"},
	{0xDA6BFF74ECB16E45ULL, "vxht_keep_n_3y3_0_0.vlm"},
	{0xDA6C9BD9135CE4D6ULL, "lt_gallowstemplar_n_2391.rml"},
	{0xDA6CADFFE1EB10AAULL, "vxlt_blackemporium_15o_0_0.vlm"},
	{0xDA6CEF22556CF45EULL, "vxca_caverns2_8r2_0_0.vlm"},
	{0xDA6D14D0861EF517ULL, "8382668.ogg"},
	{0xDA6D185401C8814AULL, "762978276.ogg"},
	{0xDA6D91F9C9D9F4C8ULL, "6081985_f.fxe"},
	{0xDA6DBE4E53C3687EULL, "gen00fl_fenris6126575.cl"},
	{0xDA6E0C68F3EF7136ULL, "hlac_ht_0_hightown_a2_1_2x.dds"},
	{0xDA6E5B572A678524ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_right_hand.gad"},
	{0xDA6F4F20209ACB98ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_player.ani"},
	{0xDA6F8301BF62DDA3ULL, "6071710_f.fxe"},
	{0xDA6FAB7424098C3EULL, "vxw_encounter1_kf_0_0.vlm"},
	{0xDA6FC34239344808ULL, "6077114_m.fxe"},
	{0xDA6FF51BFAEBE7E8ULL, "w_woundedcoast_25_0_lc.mmh"},
	{0xDA7058D1969B5AAFULL, "vxdr_0_deeproads_bpw_0_0.vlm"},
	{0xDA70DAAE160DC5ECULL, "vxca_caverns2_936_0_0.vlm"},
	{0xDA7169D4F7F76D7BULL, "lgt110_urn_guard.bnk"},
	{0xDA727D8FD0478970ULL, "vxlt_0_lowtown_docks_a3_n_d13_0_0.vlm"},
	{0xDA72E4AC19F5A85FULL, "vxlt_0_lowtown_a3_n_c_hei_0_0.vlm"},
	{0xDA72F3D732AB9B17ULL, "6092602_m.fxe"},
	{0xDA732F729961FBBEULL, "6059152_m.fxe"},
	{0xDA73A8D437843C6CULL, "740736753.ogg"},
	{0xDA741921CB127D0DULL, "mcr371_angry_mage_m6087732.cl"},
	{0xDA74F016771B7799ULL, "fen210_fenris6109409.cl"},
	{0xDA74F5F4882846E0ULL, "vxlt_0_lowtown_docks_n_b9f_0_0.vlm"},
	{0xDA756A572A6DBD02ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_right_hand.ani"},
	{0xDA756BD4F7FB593BULL, "lgt110_urn_guard.cnv"},
	{0xDA756E1889A82079ULL, "hf_bethany.mor"},
	{0xDA7585162A159A3BULL, "hlac_lt_shop_-1_1z.dds"},
	{0xDA75B5EEA312D50BULL, "6202142_m.fxe"},
	{0xDA75EE2689DFFCCFULL, "37038145.ogg"},
	{0xDA75EE5BBB3E0EC9ULL, "dae000ip_crowd_mf1_talk8.utp"},
	{0xDA7600474B847758ULL, "gencr_city_guard_archer.utc"},
	{0xDA7604D68DF6EA25ULL, "590342808.ogg"},
	{0xDA76537166DD9554ULL, "pro000st_gamlen_mother.stg"},
	{0xDA765E9928E9E3C2ULL, "6151054_m.fxe"},
	{0xDA771621AD4FDE38ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_003a_d.dds"},
	{0xDA782731939B15BBULL, "hlac_ca_caverns1_mines_24421_12y.dds"},
	{0xDA78668226148C9FULL, "gen_im_fla_fenris_02.uti"},
	{0xDA7887316B1462F2ULL, "74709933.ogg"},
	{0xDA790EFDAEE9465CULL, "sf_xy16ic32.vsh.fxo"},
	{0xDA796B15D96754B5ULL, "vxdr_primevaltomb_8uv_0_0.vlm"},
	{0xDA79CCD59A58EEC7ULL, "vxdr_primevaltomb_779_0_0.vlm"},
	{0xDA7A70E54A35F113ULL, "art\\vfx\\combat\\vfx_m_spi_md_blast_c.anb"},
	{0xDA7A8431B7FB86BBULL, "zz_dae_debug6029092.cl"},
	{0xDA7AB04A4DE8F755ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.anb"},
	{0xDA7ABF89C2A8556FULL, "w_woundedcoast_120_0.phy"},
	{0xDA7AD0C7E03D1137ULL, "llac_w_sundermount_393210_0.dds"},
	{0xDA7BDB2105500A41ULL, "hlac_w_encounter2_camp_131075_0x.dds"},
	{0xDA7D2C5413900AE2ULL, "follower_barks6153883.cl"},
	{0xDA7EC555B5148B19ULL, "ser270_emeric6113582.cl"},
	{0xDA7ECBBAF34D34BDULL, "vxht_mansion_n_53n_0_1.vlm"},
	{0xDA7EF92B4269DA82ULL, "zz_vault_debug6200366.cl"},
	{0xDA7EF967D078BECDULL, "art\\characters\\playercharacter\\dwarfmale\\glv_nuda.msh"},
	{0xDA7F0767D078D600ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_nuda.mao"},
	{0xDA7F0B67D078DD53ULL, "art\\characters\\playercharacter\\dwarfmale\\glv_nuda.mmh"},
	{0xDA7F4C8B0803DE6CULL, "vxca_caverns2_947_0_0.vlm"},
	{0xDA7F508977FA88CEULL, "mag210_arianni6057317.cl"},
	{0xDA7F70876E8B532CULL, "vxca_caverns2prc_9fl_0_0.vlm"},
	{0xDA8012C6142875BEULL, "gen00fl_anders6125356.cl"},
	{0xDA804C29AC098037ULL, "ch2notnt_ds1y.psh.fxo"},
	{0xDA822B8187B8D8E1ULL, "crt_gate_guardian.wbd"},
	{0xDA827046EF58B02CULL, "hlac_w_sundermount_0_5z.dds"},
	{0xDA8283065B2EEBCCULL, "6018594_f.fxe"},
	{0xDA82D31982AEF9FBULL, "1014930801.ogg"},
	{0xDA832C7CCAF7E9DDULL, "vxht_chantry_36e_0_0.vlm"},
	{0xDA8367D17A1E5F60ULL, "vxlt_0_lowtown_docks_n_b6g_0_0.vlm"},
	{0xDA83956CB9FB36D5ULL, "112053328.ogg"},
	{0xDA83A63CE673C051ULL, "528959054.ogg"},
	{0xDA83B0A166C3F87FULL, "6020558_f.fxe"},
	{0xDA83F3CC22E75312ULL, "ave310_donnic6134712.cl"},
	{0xDA83FBC8F84575A7ULL, "art\\vfx\\creatures\\vfx_c_drag_tswipe_c.anb"},
	{0xDA845AC6D9FF2306ULL, "nhy.cut"},
	{0xDA849068C4F546DDULL, "200600676.ogg"},
	{0xDA8550D49435CCE6ULL, "plt_lgt160pt_wk_avernus.nss"},
	{0xDA85A7672D76BF7AULL, "6208228_m.fxe"},
	{0xDA866AD646EB52A8ULL, "zz_followers_banter6157268.cl"},
	{0xDA869F2AB6249948ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofnw.mmh"},
	{0xDA86AB2AB624ADEBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofnw.mao"},
	{0xDA86BD2AB624CC72ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0cc_45roofnw.msh"},
	{0xDA878DE54A40AD7CULL, "art\\vfx\\combat\\vfx_m_spi_md_blast_c.mmh"},
	{0xDA87C5D6EFB8DF70ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viala.msh"},
	{0xDA87D3D6EFB8F68AULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viala.mmh"},
	{0xDA87D7D6EFB8FDC1ULL, "art\\levels\\races\\mage\\prop\\arcane\\prp_viala.mao"},
	{0xDA87F42084480A4BULL, "vxht_keep_n_683_0_0.vlm"},
	{0xDA883059F0ED849BULL, "hlac_lt_blackemporium_1055_25x.dds"},
	{0xDA883EB28C4F0747ULL, "vxdr_primevaltomb_708_0_0.vlm"},
	{0xDA889780E617D5B9ULL, "gen00fl_aveline6132646.cl"},
	{0xDA88E922F9567F90ULL, "802504610.ogg"},
	{0xDA8A11AAEBDAB7B0ULL, "594231776.ogg"},
	{0xDA8A4208EA777601ULL, "ave310cr_brennan.utc"},
	{0xDA8A94E391B77349ULL, "mrl150_merrill6199324.cl"},
	{0xDA8AC7A4CCFE6354ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge3.gad"},
	{0xDA8B5928C46ACFEFULL, "fen110_fenris6117986.cl"},
	{0xDA8BC4C4FB943AC2ULL, "mer320_hubert6053750.cl"},
	{0xDA8BF98E894BDD75ULL, "llac_w_encounter2_65538_0.dds"},
	{0xDA8C55DF7D27FBA2ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.impa.gad"},
	{0xDA8C8FD437D3FC2DULL, "hlac_ht_0_hightown_a2_0_0z.dds"},
	{0xDA8CB1211F588EB2ULL, "vxdr_primevaltomb_5bu_0_0.vlm"},
	{0xDA8CC654508CA959ULL, "vxht_brothel_n_620_0_0.vlm"},
	{0xDA8D8A115A6E8036ULL, "ts_generic_05.gda"},
	{0xDA8DF852A9AE88B1ULL, "follower_banter6143390.cl"},
	{0xDA8E20779E6E3F1CULL, "mcr310_argument6075837.cl"},
	{0xDA8E2BF9199AD5CAULL, "vxca_caverns1_e7s_0_0.vlm"},
	{0xDA8F23FA2EAAF922ULL, "vxht_0_hightown_a2_c48_0_0.vlm"},
	{0xDA8F25C429835264ULL, "dae000_gamlen6149469.cl"},
	{0xDA8F537BE3EDBDD8ULL, "follower_banter6170113.cl"},
	{0xDA8FC865FDE1D95FULL, "vxlt_0_lowtown_a3_n_c_hok_0_0.vlm"},
	{0xDA8FCA3B6BFD4CF1ULL, "679892448.ogg"},
	{0xDA903EF0D2223F5EULL, "6162045_m.fxe"},
	{0xDA9088A91EF8B136ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\walkf.ani"},
	{0xDA90DEE5BCB5C8E3ULL, "hm_amb_commoner1.mor"},
	{0xDA91478F4F8F0D0DULL, "234520739.ogg"},
	{0xDA914B0B20920624ULL, "6119647_m.fxe"},
	{0xDA915A0F5432E0C9ULL, "6163324_m.fxe"},
	{0xDA9174D9B3CC2738ULL, "810679078.ogg"},
	{0xDA91997B4874651AULL, "hlac_lt_0_lowtown_docks_n_-1_2y.dds"},
	{0xDA91D6A4CD049B32ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge3.ani"},
	{0xDA921F454D9802EBULL, "dae000_merchant_weapons_dm.bnk"},
	{0xDA92C2CA15AE1ACDULL, "152053908.ogg"},
	{0xDA92CF3578BDAD66ULL, "hlac_lt_0_lowtown_docks_0_2x.dds"},
	{0xDA9328EB466B9DE9ULL, "6143353_m.fxe"},
	{0xDA936C7A7F28339EULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.dura.evt"},
	{0xDA9378DF7D2E5599ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.impa.evt"},
	{0xDA94858D4B729203ULL, "vxdr_primevaltomb_2wp_0_0.vlm"},
	{0xDA950A89883E2AA7ULL, "qun110tr_petrice_ambient.ncs"},
	{0xDA9529B503A9329BULL, "748937773.ogg"},
	{0xDA9571A6D7095F3CULL, "qun110_petrice6036422.cl"},
	{0xDA9621454D9BEECBULL, "dae000_merchant_weapons_dm.cnv"},
	{0xDA9665FD763E22CBULL, "vxht_keep_n_632_0_0.vlm"},
	{0xDA96820168F5F827ULL, "6142925_m.fxe"},
	{0xDA96880E77D8D5EFULL, "vxht_mansion2b_3du_0_0.vlm"},
	{0xDA96C6A0CCDEA155ULL, "399035659.ogg"},
	{0xDA971C119D9FCA02ULL, "ico_bomb_sticky.dds"},
	{0xDA9726E950C0EB0BULL, "549858907.ogg"},
	{0xDA9780842057AE16ULL, "vxht_0_hightown_b8u_0_0.vlm"},
	{0xDA97B260D6B1307CULL, "dae000_dalish_apprentices6147559.cl"},
	{0xDA984945BA0F34D9ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden2.mmh"},
	{0xDA984F532322A2D9ULL, "907740640.ogg"},
	{0xDA985345BA0F45C7ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden2.msh"},
	{0xDA9958F45976517AULL, "12463213.ogg"},
	{0xDA996D4FF9E6A4BCULL, "vxw_bonepit_1ru_0_0.vlm"},
	{0xDA99837A7F2CC6C9ULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.dura.gad"},
	{0xDA99A4436B046D31ULL, "llac_lt_0_lowtown_a3_n_c_131071_1.dds"},
	{0xDA9A65D494484C21ULL, "plt_lgt160pt_wk_avernus.plo"},
	{0xDA9ABE0C7375D1FAULL, "zz_fen_debug6166025.cl"},
	{0xDA9AF4AAC503A64AULL, "vxlt_backalley_g_1u0_0_0.vlm"},
	{0xDA9B65EC140ABCECULL, "dev_res_up2.mao"},
	{0xDA9B9BF0D9813896ULL, "lowtown_fem_alienage1_sta6162522.cl"},
	{0xDA9BA403F4C64DE4ULL, "vxlt_0_lowtown_eqg_0_0.vlm"},
	{0xDA9BC6AD9DACF1FCULL, "vxlt_ruinedhouse_n_2v_0_0.vlm"},
	{0xDA9BC7BDD4B56D56ULL, "dae200_cullen6140333.cl"},
	{0xDA9BFD41AD6929B1ULL, "vxdr_0_deeproads_fbg_0_0.vlm"},
	{0xDA9C8FB92F251626ULL, "6205151_m.fxe"},
	{0xDA9CCCC32269DEA9ULL, "822884504.ogg"},
	{0xDA9D7173473F4171ULL, "398368125.ogg"},
	{0xDA9D7DB620F9E688ULL, "vxdr_primevaltomb_7xc_0_0.vlm"},
	{0xDA9D84F7EA09FCE9ULL, "gen00fl_varric6076073.cl"},
	{0xDA9D95D720A111A2ULL, "vxht_0_hightown_a2_c39_0_0.vlm"},
	{0xDA9E3763E8D24A9BULL, "6051654_m.fxe"},
	{0xDA9E549359D612D9ULL, "zz_followers_banter6159243.cl"},
	{0xDA9E6BF3844F13DBULL, "lt_0_lowtown_docks_a3_n_6_0.rml"},
	{0xDA9E90A91F04FC27ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\walkf.evt"},
	{0xDA9EE756799B0578ULL, "mcr371_hostile_merrill6082143.cl"},
	{0xDA9EF1D8689C0F6BULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_spiderweb_sa.dds"},
	{0xDA9F890A1BD5A152ULL, "vxlt_0_lowtown_docks_a3_n_8pz_0_0.vlm"},
	{0xDA9F8B3D5C1708ECULL, "205125742.ogg"},
	{0xDA9FE7695CFD0FB4ULL, "272167108.ogg"},
	{0xDAA041DF5B6CE234ULL, "art\\vfx\\creatures\\fxc_bm_voml_c.mmh"},
	{0xDAA04A139458380DULL, "zz_ave_debug6134002.cl"},
	{0xDAA0547A7F329517ULL, "art\\vfx\\combat\\vfx_glb_stun_crust_c.dura.ani"},
	{0xDAA080DF7D38ED88ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_loop_c.impa.ani"},
	{0xDAA0D69E57EF98EFULL, "hlac_ht_keep_n_3731_1z.dds"},
	{0xDAA0E305523525C8ULL, "pri_fire.mao"},
	{0xDAA16F0BBBC9001AULL, "156576093.ogg"},
	{0xDAA1813DC941A81EULL, "vxw_woundedcoast_14n_0_0.vlm"},
	{0xDAA1BCBB39B9D6D9ULL, "6134957_m.fxe"},
	{0xDAA236A05DF80520ULL, "ht_brothel_n_3365.anb"},
	{0xDAA2B14A4E0A4A96ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_carck_p.mmh"},
	{0xDAA2E2837254479CULL, "vxw_encounter2_cy_0_0.vlm"},
	{0xDAA2FB71ED129B6AULL, "505827593.ogg"},
	{0xDAA31B89177D0CA5ULL, "496781543.ogg"},
	{0xDAA323ACDE2756BBULL, "704687182.ogg"},
	{0xDAA3B1CDAEF48FFFULL, "491459450.ogg"},
	{0xDAA3C4E4AFFD9B61ULL, "vxlt_backalley_n_1zc_0_0.vlm"},
	{0xDAA3D19AB54E6B35ULL, "6118959_m.fxe"},
	{0xDAA3EED162814A81ULL, "vxlt_undercityencounter_100_0_0.vlm"},
	{0xDAA44FC8CD594F8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt4open.msh"},
	{0xDAA455C8CD595A7EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt4open.mmh"},
	{0xDAA461C8CD596E15ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt4open.mao"},
	{0xDAA482949286966EULL, "599247200.ogg"},
	{0xDAA48E7D5BC5DD1EULL, "hlac_lt_undercityserialkiller_-1_24x.dds"},
	{0xDAA4A0CD0E2C9BD4ULL, "combat_in_r1_a1.gda"},
	{0xDAA4A971ADB026CBULL, "qun200_party_comment6182722.cl"},
	{0xDAA4B1A91F09A070ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\walkf.gad"},
	{0xDAA4B5C8B102A06DULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comc_0d.dds"},
	{0xDAA5AE1C641C067EULL, "6153504_m.fxe"},
	{0xDAA5D7ED285CD3C4ULL, "hlac_ht_keep_n_3732_9x.dds"},
	{0xDAA698D9138E578AULL, "lt_gallowstemplar_n_2391.anb"},
	{0xDAA705C9AF1A0070ULL, "dae421ar_player_home_nt.are"},
	{0xDAA709E13F8B0B5AULL, "lgt110_gxa_coa_spy6117444.cl"},
	{0xDAA74D8187D82EE3ULL, "crt_gate_guardian.bnk"},
	{0xDAA7538EE9E1A606ULL, "517280786.ogg"},
	{0xDAA777F8850EC090ULL, "6057739_f.fxe"},
	{0xDAA77C1B719F5E24ULL, "vxht_0_hightown_b21_0_0.vlm"},
	{0xDAA86558952301FEULL, "6166427_m.fxe"},
	{0xDAA8A9C8A818E9C2ULL, "760262755.ogg"},
	{0xDAA91359855DDFCEULL, "31745316.ogg"},
	{0xDAA963640286B43BULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarpb.mmh"},
	{0xDAA96F640286C8D8ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarpb.mao"},
	{0xDAA981640286E725ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarpb.msh"},
	{0xDAA9AA0E3478DA59ULL, "vxht_brothel_n_686_0_0.vlm"},
	{0xDAA9CB07D560E09CULL, "vxlt_undercitydungeon_1g9_0_0.vlm"},
	{0xDAAA042F0C7800BFULL, "vxca_caverns2_948_0_0.vlm"},
	{0xDAAA50F9D15DDDC1ULL, "hlac_lt_0_lowtown_a2_n_c_-2_2z.dds"},
	{0xDAAA80B5C378C68BULL, "6128860_m.fxe"},
	{0xDAAAC944CEB3527CULL, "vxca_cavern1_haunted_e8e_0_0.vlm"},
	{0xDAAAD35DF7C20788ULL, "856793570.ogg"},
	{0xDAAB11FCF243E843ULL, "art\\vfx\\textures\\vfx_d_curvedvert.dds"},
	{0xDAAB1CC9AF1E1072ULL, "dae421ar_player_home_nt.ncs"},
	{0xDAAB3A40AFFBE03AULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_lava2_n.dds"},
	{0xDAAB857FF8E5D360ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masb.phy"},
	{0xDAABE4FAADB373D9ULL, "vxht_0_hightown_a2_avg_0_0.vlm"},
	{0xDAABFE62DF657668ULL, "6115867_m.fxe"},
	{0xDAAC8303D4F846C2ULL, "hlac_lt_gallowstemplar_2384_7x.dds"},
	{0xDAACF8C8F86875A8ULL, "art\\vfx\\creatures\\vfx_c_drag_tswipe_c.mmh"},
	{0xDAAD006E0357BDBDULL, "208277775.ogg"},
	{0xDAAD49EFA7709844ULL, "909851284.ogg"},
	{0xDAAD606730C7056BULL, "hlac_dr_primevaltomb_659_9z.dds"},
	{0xDAADAD96992775B3ULL, "gen00fl_isabela6088063.cl"},
	{0xDAADCA5EB72D5A16ULL, "6153658_m.fxe"},
	{0xDAAE3B19E1A8825FULL, "vfx_ring02.dds"},
	{0xDAAE6939D9F6DCE6ULL, "499192992.ogg"},
	{0xDAAE780391145EB7ULL, "lt_0_lowtown_docks_243_0.mmh"},
	{0xDAAE7C03911465C4ULL, "lt_0_lowtown_docks_243_0.mao"},
	{0xDAAE8E0391148429ULL, "lt_0_lowtown_docks_243_0.msh"},
	{0xDAAEEC0F4664E652ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.mmh"},
	{0xDAAF4B8B0A20C526ULL, "hlac_lt_darktown_-1_6z.dds"},
	{0xDAAF5F9E37A8BEF2ULL, "77982735.ogg"},
	{0xDAAF6D395CA3117CULL, "dae000ip_crowd_ff8_talk18.utp"},
	{0xDAAF9707A55C2BC5ULL, "6152311_m.fxe"},
	{0xDAB009A15CB6979BULL, "6068464_f.fxe"},
	{0xDAB0B7AFFD3DAA6FULL, "vxdr_0_deeproads_782_0_0.vlm"},
	{0xDAB100A7191CA5E1ULL, "6120902_f.fxe"},
	{0xDAB114C9AF226E63ULL, "dae421ar_player_home_nt.lst"},
	{0xDAB17AFEFDB0C14CULL, "zz_dae_debug6029079.cl"},
	{0xDAB18ECC8C8565B0ULL, "hlac_dr_primevaltomb_9621_3y.dds"},
	{0xDAB1DF213E79CFE7ULL, "694681339.ogg"},
	{0xDAB1FEF7CF1FFBDCULL, "hlac_lt_0_lowtown_docks_-131073_2z.dds"},
	{0xDAB23417363DA287ULL, "one300dg_lia.bnk"},
	{0xDAB244E0413C6CB0ULL, "vxw_bonepitdestroyed_a3_km_0_0.vlm"},
	{0xDAB3300EC9C10009ULL, "vxlt_gamlenshouse_n_bg_0_0.vlm"},
	{0xDAB3A45772CBB369ULL, "vxlt_warehouse_n_dc_0_0.vlm"},
	{0xDAB3B7469F896CCBULL, "lgt000ip_ultra_tome.utp"},
	{0xDAB44ADC719CF2C7ULL, "dae000_bartrand6206967.cl"},
	{0xDAB47E8748DBBE53ULL, "vxlt_0_lowtown_a2_n_c_byr_0_0.vlm"},
	{0xDAB4A0EEEF31738AULL, "971652301.ogg"},
	{0xDAB4D0698D681234ULL, "vxca_caverns1_mines_iod_0_0.vlm"},
	{0xDAB5020E8A8CDA2FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_extrud2windo.phy"},
	{0xDAB52437C4BFE703ULL, "vxht_playershouse_n_1sp_0_0.vlm"},
	{0xDAB55617364012FFULL, "one300dg_lia.cnv"},
	{0xDAB588AD39A67A33ULL, "6121045_m.fxe"},
	{0xDAB5D9585A9A3015ULL, "6075939_m.fxe"},
	{0xDAB620A6FCEAD4E6ULL, "vxw_woundedcoast_1fl_0_0.vlm"},
	{0xDAB6BE1AB75892C1ULL, "vxlt_undercityencounter_wb_0_0.vlm"},
	{0xDAB6CF9489165DEEULL, "6121758_m.fxe"},
	{0xDAB6FF2E2F152408ULL, "hlac_lt_0_lowtown_docks_-4_1z.dds"},
	{0xDAB75B0AB2B39508ULL, "hlac_dr_primevaltomb_663_3z.dds"},
	{0xDAB76E1B2D0078A1ULL, "182234906.ogg"},
	{0xDAB7B09892C67D17ULL, "qun111_ketojan6040782.cl"},
	{0xDAB7D189CDDB4C77ULL, "6161214_m.fxe"},
	{0xDAB806C72A7E09BCULL, "hlac_lt_ruinedhouse_n_-1_4x.dds"},
	{0xDAB831DAD638E6FAULL, "gen00fl_fenris6106689.cl"},
	{0xDAB8E0FB91555AB1ULL, "vxdr_0_deeproads_fhe_0_0.vlm"},
	{0xDAB94B77DD5D983FULL, "916864531.ogg"},
	{0xDAB959932E88D46CULL, "6015868_m.fxe"},
	{0xDAB975316190794BULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_hed_0d.dds"},
	{0xDABA28FE1185F1C2ULL, "vxlt_warehouse_n_sm_0_0.vlm"},
	{0xDABAB35A65F99ECCULL, "mag210_arianni6152621.cl"},
	{0xDABABFF177DF969AULL, "216677867.ogg"},
	{0xDABAFD4B22DB4743ULL, "vxdr_primevaltomb_7t0_0_0.vlm"},
	{0xDABBA316FDC62155ULL, "6045860_m.fxe"},
	{0xDABBAD2FAE602F48ULL, "749583703.ogg"},
	{0xDABBD9B9E78DD6DEULL, "320351822.ogg"},
	{0xDABC090F767ABC81ULL, "6095808_f.fxe"},
	{0xDABC2129FC95420EULL, "68805579.ogg"},
	{0xDABC3C3BA9237296ULL, "6156558_m.fxe"},
	{0xDABC77DB9A7D4665ULL, "6084356_f.fxe"},
	{0xDABCF1318AF38A6DULL, "104751249.ogg"},
	{0xDABD60795688178BULL, "6085551_f.fxe"},
	{0xDABDB841D0EFF744ULL, "llac_w_bonepitdestroyed_a3_65542_0.dds"},
	{0xDABDD6811D7A2F38ULL, "framebuffereffects_ps3downsamplemsaa_p0.psh.fxo.dx11.assembly"},
	{0xDABDFF3D0A305CD2ULL, "vxlt_0_lowtown_a2_n_c_byw_0_0.vlm"},
	{0xDABE07212F1C9D38ULL, "hlac_lt_0_lowtown_65534_3y.dds"},
	{0xDABE16BF84AFF373ULL, "gen_im_npc_arm_hum_cht_rdrb.uti"},
	{0xDABE3F0C5385B826ULL, "lt_warehouse_n_mmap_ph.dds"},
	{0xDABE5324128D6AE0ULL, "hlac_lt_0_lowtown_a3_n_c_131070_1z.dds"},
	{0xDABED52F7248771FULL, "6144398_m.fxe"},
	{0xDABF40B4AC6A5B4EULL, "art\\vfx\\combat\\vfx_gib_hm_masc_rtl.anb"},
	{0xDABF72B5B00BB9CBULL, "gen_im_cut_arrow.uti"},
	{0xDABF7CCBC3EC839AULL, "1010617749.ogg"},
	{0xDAC0678A9DD61BDAULL, "449956087.ogg"},
	{0xDAC0AD71E33A1EDEULL, "vxca_caverns2_53p_0_0.vlm"},
	{0xDAC0D0364AC0A425ULL, "370912766.ogg"},
	{0xDAC14813F1F07A49ULL, "vxlt_gallowscourtyard_n_3x9_0_0.vlm"},
	{0xDAC20B47E48EF5FAULL, "697141318.ogg"},
	{0xDAC3072985B6CF5FULL, "6096466_m.fxe"},
	{0xDAC34A160E7DB07BULL, "993769415.ogg"},
	{0xDAC39614B6B5FF83ULL, "vxht_playershouse_n_1vs_0_0.vlm"},
	{0xDAC3D228310C521AULL, "vxlt_backalley_g_1ep_0_0.vlm"},
	{0xDAC42A2DAE6A5E43ULL, "llac_lt_backalley_n_0_13.dds"},
	{0xDAC4529A0DC3336EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\4sd_basic_2.mao"},
	{0xDAC45FD5558B8F24ULL, "vxht_0_hightown_b87_0_0.vlm"},
	{0xDAC49283EEE0ED66ULL, "vxw_woundedcoast_1am_0_0.vlm"},
	{0xDAC513FEA216D149ULL, "278845801.ogg"},
	{0xDAC54AD9C6CB37B3ULL, "fire_blue.mao"},
	{0xDAC57FF965EB3139ULL, "6113936_m.fxe"},
	{0xDAC637FE50C4493BULL, "501893245.ogg"},
	{0xDAC72027E33EEF3AULL, "pro000_cut2d_thefirstyear6005310.cl"},
	{0xDAC728395311441DULL, "ht_0_hightownexplosion_n_52_0.rml"},
	{0xDAC77D29A05F06ADULL, "6095985_f.fxe"},
	{0xDAC7AEE74C8B4E4FULL, "atl_prestartme_dxt5_01.dds"},
	{0xDAC7DBA3B4DA0071ULL, "6128343_m.fxe"},
	{0xDAC84C892C7E7D08ULL, "6101501_m.fxe"},
	{0xDAC944DF5B8FECEBULL, "art\\vfx\\creatures\\fxc_bm_voml_c.anb"},
	{0xDAC9895849681BBEULL, "6111659_f.fxe"},
	{0xDAC9B98132618E46ULL, "147590532.ogg"},
	{0xDACA9554CE09CD01ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_0z.dds"},
	{0xDACAD613B16ECB67ULL, "vxht_mansion2_1mx_0_0.vlm"},
	{0xDACAE261E4DF53BFULL, "44156957.ogg"},
	{0xDACAF3581475951DULL, "hlac_w_woundedcoast_-262144_0x.dds"},
	{0xDACB5B58D060CC04ULL, "416113587.ogg"},
	{0xDACBFF05A7CF50C4ULL, "ht_0_hightown_n_45_0.anb"},
	{0xDACC6ADC1B04460FULL, "467957508.ogg"},
	{0xDACC89F4E8E856F5ULL, "zz_isa_debug6175884.cl"},
	{0xDACD22473F2401D7ULL, "lt_0_lowtown_docks_n_224_0.msh"},
	{0xDACD32656AE8A617ULL, "hlac_ht_keep_n_3739_0y.dds"},
	{0xDACD34473F24203AULL, "lt_0_lowtown_docks_n_224_0.mao"},
	{0xDACD38473F242749ULL, "lt_0_lowtown_docks_n_224_0.mmh"},
	{0xDACD470C643E8F9FULL, "6141098_m.fxe"},
	{0xDACD8CAA0F59BCD3ULL, "w_bonepitdestroyed_a3_24_0_ll.mmh"},
	{0xDACDC2EDB090F296ULL, "473788144.ogg"},
	{0xDACE70E284BC27CDULL, "vxlt_0_lowtown_docks_bud_0_0.vlm"},
	{0xDACEFCCBD7FEEC10ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_maka_0t.dds"},
	{0xDACF2262DFD6B9E1ULL, "6069379_f.fxe"},
	{0xDACF8EA81A31B6A1ULL, "qun160_party_comment_16061235.cl"},
	{0xDAD0378A1E47A3F4ULL, "6068006_m.fxe"},
	{0xDAD0A16D3027F8BEULL, "vxlt_hangedman_1xp_0_0.vlm"},
	{0xDAD126D97B87EC12ULL, "hlac_lt_0_lowtown_docks_n_196608_1x.dds"},
	{0xDAD146A7EFED5F5FULL, "422195037.ogg"},
	{0xDAD1A27FF906D3F5ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masb.msh"},
	{0xDAD1A47FF906D76BULL, "art\\characters\\playercharacter\\humanfemale\\hlh_masb.mmh"},
	{0xDAD1B42371544334ULL, "vxca_caverns1_mines_iaj_0_0.vlm"},
	{0xDAD1BB57044CFDB6ULL, "vxw_woundedcoast_133_0_0.vlm"},
	{0xDAD220A2AA0C079CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_be_transition.ani"},
	{0xDAD25941B1225AA9ULL, "vxlt_warehouse_132_0_0.vlm"},
	{0xDAD2AF9AD775A4F5ULL, "dae000_oldmangossip6175753.cl"},
	{0xDAD2B16E83974318ULL, "6116736_m.fxe"},
	{0xDAD33E7F9A73E01BULL, "6177132_m.fxe"},
	{0xDAD3B796B2889762ULL, "6115879_f.fxe"},
	{0xDAD42560FE323335ULL, "gen00fl_anders6113064.cl"},
	{0xDAD4F44F92CA5541ULL, "6094959_f.fxe"},
	{0xDAD574B84C1AC616ULL, "vxlt_0_lowtown_a2_n_ekf_0_2.vlm"},
	{0xDAD5A4CAF2EBC78FULL, "vxw_encounter3_g0_0_1.vlm"},
	{0xDAD61EDBA4959BFCULL, "vxca_cavern1_haunted_e7x_0_0.vlm"},
	{0xDAD647F6B2357A7DULL, "zz_approval_debug6003095.cl"},
	{0xDAD6F61EF2057CC2ULL, "6178437_m.fxe"},
	{0xDAD75097CF3AC5C8ULL, "hlac_lt_undercityencounter_-1_33x.dds"},
	{0xDAD7B848FB4A94CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ready.gad"},
	{0xDAD7E3355F195AD4ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masb.phy"},
	{0xDAD7EB0F4687E9F9ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.anb"},
	{0xDAD7F5F196D929ECULL, "282525529.ogg"},
	{0xDAD8C75C8D74F26EULL, "ht_mansion_n_4152_ll.mmh"},
	{0xDAD92DA2AA123CDEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_be_transition.gad"},
	{0xDAD93ECBF6B72BCDULL, "vxht_keep_5k8_0_0.vlm"},
	{0xDAD97DF717A48C54ULL, "467183520.ogg"},
	{0xDAD982DBFBAFBE1AULL, "newmatobj1.mao"},
	{0xDAD9AD28BC5FD894ULL, "1031812037.ogg"},
	{0xDAD9B85FAC9BAF0AULL, "6113845_m.fxe"},
	{0xDADB1B99B609CDAFULL, "6170418_m.fxe"},
	{0xDADB301BB6D5F7E8ULL, "gen00fl_anders6119487.cl"},
	{0xDADB389AF848DC4FULL, "607810069.ogg"},
	{0xDADBA236228B1536ULL, "verticalblur.vsh.fxo"},
	{0xDADBFDB23AA40BD9ULL, "lla_ht_0_hightown_-196608_0.dds"},
	{0xDADC330E355271D5ULL, "mag100_karras6035297.cl"},
	{0xDADC7D31344A4DA7ULL, "788922296.ogg"},
	{0xDADC9612D329B15FULL, "gen00fl_varric6107274.cl"},
	{0xDADCFFEADBDAFF56ULL, "vxlt_0_lowtown_a3_n_c_hju_0_0.vlm"},
	{0xDADDB3BC2C86B91FULL, "6072891_m.fxe"},
	{0xDADE8B236F74BB67ULL, "vxlt_0_lowtown_docks_a3_n_bg8_0_0.vlm"},
	{0xDADE944B442D0103ULL, "sse_ss_slavers_f_1.gda"},
	{0xDADECD48FB50D77CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ready.evt"},
	{0xDADEDDDAC9868D66ULL, "744861466.ogg"},
	{0xDADEE9255026D196ULL, "qun160_vashothlead6053863.cl"},
	{0xDADF17ED679264DBULL, "vxht_0_hightown_n_c60_0_0.vlm"},
	{0xDADF4CEA8CA21AFDULL, "hlac_lt_0_lowtown_docks_-3_3x.dds"},
	{0xDADF53B9A3077CF2ULL, "llac_lt_backalley_n_65536_2.dds"},
	{0xDADFE71479DD7D22ULL, "gen_im_cft_rec_arn_08.uti"},
	{0xDAE008A2AA181CFDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_be_transition.evt"},
	{0xDAE05B5D8C722A46ULL, "qun200_party_comment6090448.cl"},
	{0xDAE07F7888592BC3ULL, "vxca_cave_23m_0_0.vlm"},
	{0xDAE08976CAD98064ULL, "llac_ca_caverns2_5802_32.dds"},
	{0xDAE0CB1EA3187329ULL, "vxlt_warehouse_161_0_0.vlm"},
	{0xDAE18E911C443B73ULL, "6101112_f.fxe"},
	{0xDAE19DD6CA3732CAULL, "vxlt_undercityserialkiller_200_0_0.vlm"},
	{0xDAE20351CCEF9EE8ULL, "qun_231_varnell_2.stg"},
	{0xDAE23DEDE6F8A507ULL, "zz_followers_banter6158541.cl"},
	{0xDAE24CBA1A7C65EEULL, "hlac_ca_caverns1_mines_24421_6y.dds"},
	{0xDAE2C0355F2326D1ULL, "art\\characters\\playercharacter\\humanfemale\\boo_masb.msh"},
	{0xDAE2CE355F233EECULL, "art\\characters\\playercharacter\\humanfemale\\boo_masb.mao"},
	{0xDAE2DA355F23528FULL, "art\\characters\\playercharacter\\humanfemale\\boo_masb.mmh"},
	{0xDAE315A05E2F6448ULL, "ht_brothel_n_3365.rml"},
	{0xDAE47E758D8DC20BULL, "615747018.ogg"},
	{0xDAE57DD58E8CC5EAULL, "6103825_f.fxe"},
	{0xDAE58176FCD9508DULL, "6091379_m.fxe"},
	{0xDAE6431EC2E868FCULL, "vxlt_undercitydungeon_1a3_0_0.vlm"},
	{0xDAE6749D5C61BEC6ULL, "vxht_keep_n_67g_0_0.vlm"},
	{0xDAE694A3A008743FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_reaction_loop.gad"},
	{0xDAE6B646264E9D15ULL, "898041412.ogg"},
	{0xDAE6EBE40920EA77ULL, "997467203.ogg"},
	{0xDAE6EE83C29303EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stumble_back.ani"},
	{0xDAE71624A620C6E8ULL, "vxht_0_hightown_a3_cap_0_0.vlm"},
	{0xDAE7B8E0AF5FE4DAULL, "294402957.ogg"},
	{0xDAE7CE006518F1E9ULL, "286416815.ogg"},
	{0xDAE7FE1670C2C24CULL, "1174405.ogg"},
	{0xDAE850692CEAD5A4ULL, "6125492_f.fxe"},
	{0xDAE92F92BB20FA40ULL, "gen00fl_aveline6053320.cl"},
	{0xDAE9F78988864AFEULL, "qun110tr_petrice_ambient.utt"},
	{0xDAEA3D9733737A1DULL, "gen00fl_varric6191288.cl"},
	{0xDAEB1EB712F689E4ULL, "character_character_p29.psh.fxo"},
	{0xDAEBD037BF76E5C2ULL, "art\\vfx\\models\\vfx_trap_radialblade.msh"},
	{0xDAEBD237BF76E978ULL, "art\\vfx\\models\\vfx_trap_radialblade.mmh"},
	{0xDAEBD548FB5B6E95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\ready.ani"},
	{0xDAEBDE37BF76FD1BULL, "art\\vfx\\models\\vfx_trap_radialblade.mao"},
	{0xDAED6BCB3AA41569ULL, "vxlt_warehouse_n_xg_0_0.vlm"},
	{0xDAED89CA59887D5BULL, "vxht_0_hightown_n_c33_0_0.vlm"},
	{0xDAEDC2D4DDDD9F8DULL, "6104738_m.fxe"},
	{0xDAEDD5B2584A31DDULL, "6092496_f.fxe"},
	{0xDAEDEFA5F325EA26ULL, "vxdr_primevaltomb_34u_0_0.vlm"},
	{0xDAEE6A53548576ADULL, "328325466.ogg"},
	{0xDAEE9F10E8392EB6ULL, "vxw_woundedcoast_195_0_0.vlm"},
	{0xDAEED12B021B98E9ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs106a.phy"},
	{0xDAEF8CA8FE2E6ACFULL, "vxht_bartrand_n_5i4_0_0.vlm"},
	{0xDAEFD615CF0F52E3ULL, "vxlt_0_lowtown_a2_n_c_cmk_0_0.vlm"},
	{0xDAEFE81AC4C336E6ULL, "vxw_woundedcoast_1rp_0_0.vlm"},
	{0xDAF0242B27876B6AULL, "hlac_ht_0_hightown_n_-65536_1y.dds"},
	{0xDAF0471435D97778ULL, "hlac_w_encounter1_65541_0x.dds"},
	{0xDAF061DF0AA3B3CBULL, "vxlt_undercitydungeon_14c_0_0.vlm"},
	{0xDAF0846402C2EF70ULL, "art\\levels\\races\\ferelden\\props\\docks\\prp_cratetarpb.phy"},
	{0xDAF1031BE777B574ULL, "hlac_ht_0_hightown_a2_-131074_2x.dds"},
	{0xDAF1F3E744982E8CULL, "vxlt_0_lowtown_docks_a3_czn_0_0.vlm"},
	{0xDAF1FC5C6D183F69ULL, "353409005.ogg"},
	{0xDAF212C160893032ULL, "vxlt_0_lowtown_a3_n_c_eoo_0_0.vlm"},
	{0xDAF2197F9BD76C69ULL, "vxlt_darktown_4to_0_0.vlm"},
	{0xDAF23F23407F22E4ULL, "w_bonepit_36_0_ll.mmh"},
	{0xDAF24529B3B9EEE1ULL, "vxlt_0_lowtown_docks_a3_bbb_0_0.vlm"},
	{0xDAF2CAA03B0FB60BULL, "718666900.ogg"},
	{0xDAF2D62B2D640963ULL, "lt_0_lowtown_docks_n_203_0.mao"},
	{0xDAF2DA2B2D640FB0ULL, "lt_0_lowtown_docks_n_203_0.mmh"},
	{0xDAF2DB6A6DC1C5BDULL, "hlac_ca_caverns2_5802_30y.dds"},
	{0xDAF2E82B2D6427CAULL, "lt_0_lowtown_docks_n_203_0.msh"},
	{0xDAF2EE54CA7A535BULL, "sse_ss_isabela.gda"},
	{0xDAF327919B042607ULL, "vxht_bartrand_n_66r_0_0.vlm"},
	{0xDAF37E487D48D94CULL, "vxlt_foundry_n_wm_0_0.vlm"},
	{0xDAF39B1406FA1412ULL, "dre000_sibling_has_blight6036535.cl"},
	{0xDAF3B860A855CEAFULL, "character_character_p19.psh.fxo.dx11.assembly"},
	{0xDAF3CCDAAA3E250DULL, "w_bonepit_14_0.mmh"},
	{0xDAF3D0DAAA3E2C5EULL, "w_bonepit_14_0.mao"},
	{0xDAF3DEDAAA3E4393ULL, "w_bonepit_14_0.msh"},
	{0xDAF3E683C29D803FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stumble_back.evt"},
	{0xDAF4359FB276F896ULL, "vfxchunky0_voltex.psh.fxo"},
	{0xDAF44B5AB6DE1064ULL, "vxht_chantry_n_2c9_0_0.vlm"},
	{0xDAF4D0D5E6D36FBFULL, "sse_ss_lowclass_thugs_f_1.gda"},
	{0xDAF4E67A4E57D746ULL, "vxht_keep_n_62h_0_0.vlm"},
	{0xDAF4FA48CC8187C9ULL, "vxht_bartrand_n_3nm_0_0.vlm"},
	{0xDAF514FAAEB62C13ULL, "mag131cr_bloodmage2.utc"},
	{0xDAF68A6F0F13AE58ULL, "vxlt_0_lowtown_docks_b5d_0_1.vlm"},
	{0xDAF71884E8AD7D84ULL, "134279590.ogg"},
	{0xDAF71D21AA2EEAC4ULL, "gen_im_trs_gen_sta_ele_r2.uti"},
	{0xDAF76D5FB8922F34ULL, "6110709_m.fxe"},
	{0xDAF7983ED687DF82ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_activ_c.mmh"},
	{0xDAF79D9D1826CC9AULL, "6060840_m.fxe"},
	{0xDAF80A2BD54C0DCBULL, "gen00fl_varric6033667.cl"},
	{0xDAF85C3A489C26D8ULL, "121966647.ogg"},
	{0xDAF875CE146F0C46ULL, "zz_mcr_debug6088344.cl"},
	{0xDAF8CE9AEE9727AFULL, "zz_dae_debug6047629.cl"},
	{0xDAF8DA5B38638F6BULL, "follower_banter6164760.cl"},
	{0xDAF910249D3C3BD3ULL, "279420973.ogg"},
	{0xDAF94D5979659408ULL, "hlac_lt_gallowscourtyard_-131073_0z.dds"},
	{0xDAF9672A24FABC90ULL, "llac_dr_primevaltomb_659_0.dds"},
	{0xDAF973E3E1FE43DAULL, "vxca_caverns1_mines_ilc_0_0.vlm"},
	{0xDAF9C6795A9E714DULL, "vxlt_0_lowtown_docks_bba_0_0.vlm"},
	{0xDAFA4F13FED23B29ULL, "ht_0_hightownexplosion_n_41_0.rml"},
	{0xDAFA7389EE244E45ULL, "fen110_fenris6096180.cl"},
	{0xDAFA8E19C6889E9AULL, "vxlt_0_lowtown_a2_n_eoe_0_1.vlm"},
	{0xDAFA98FAE9F9A703ULL, "llac_ht_brothel_n_3366_0.dds"},
	{0xDAFAC783C2A36A28ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stumble_back.gad"},
	{0xDAFB67827D3EA4D4ULL, "qun211_braggart_after_fight.stg"},
	{0xDAFB71A3A01A9485ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\isabela_post_sex_reaction_loop.ani"},
	{0xDAFB83416514E7F3ULL, "927037531.ogg"},
	{0xDAFBFBA74B7E95DBULL, "vxht_0_hightown_n_c82_0_0.vlm"},
	{0xDAFCB6BBF849EE10ULL, "vxlt_darktown_4im_0_0.vlm"},
	{0xDAFCCF1867B46A96ULL, "35570710.ogg"},
	{0xDAFCF926BCF764DFULL, "genip_corpse3.utp"},
	{0xDAFDA26CBED39FC4ULL, "and100_anders6206361.cl"},
	{0xDAFDBAC7F155600DULL, "follower_banter6163583.cl"},
	{0xDAFDC8B44BB4F7BEULL, "hlac_lt_gallowstemplar_f_2388_6x.dds"},
	{0xDAFE3665C5A3F3F0ULL, "hlac_w_sundermount_-393214_0z.dds"},
	{0xDAFE43616583DAA1ULL, "6101030_m.fxe"},
	{0xDAFECCDAAA482C86ULL, "w_bonepit_14_0.rml"},
	{0xDAFED5B562BECB51ULL, "833715387.ogg"},
	{0xDAFEFF2609125BC2ULL, "488205704.ogg"},
	{0xDB00849E527F48B2ULL, "vxlt_0_lowtown_a3_n_c_erh_0_0.vlm"},
	{0xDB0134D0DEFCD3E3ULL, "vxlt_0_lowtown_a2_n_c_fe2_0_0.vlm"},
	{0xDB01794786038FF1ULL, "6175136_m.fxe"},
	{0xDB0193FF5BBAA591ULL, "vxht_mansion2_1pm_0_0.vlm"},
	{0xDB01996E8CFA3E87ULL, "vxht_bartrand_n_63m_0_0.vlm"},
	{0xDB0202AF7B890FC5ULL, "hlac_ht_mansion2_818_16x.dds"},
	{0xDB024BF133FCFB9EULL, "1045889885.ogg"},
	{0xDB026AF20CE0172CULL, "vxlt_0_lowtown_a3_n_c_ewy_0_0.vlm"},
	{0xDB02F42A24D82A6AULL, "hlac_ca_caverns1_24425_2z.dds"},
	{0xDB0326D8A6D499FCULL, "vxlt_undercitydungeon_1o1_0_0.vlm"},
	{0xDB0333D1047EF56AULL, "6209978_f.fxe"},
	{0xDB034EC7FE46B5A7ULL, "gen00fl_isabela6106178.cl"},
	{0xDB03AE7E895B29AAULL, "620319896.ogg"},
	{0xDB0424BE2E20D50EULL, "cli600cr_ds_attacker.ncs"},
	{0xDB04A52A1E5E0856ULL, "fx_hand.mao"},
	{0xDB057689BAB42E07ULL, "hlac_lt_undercityencounter_-1_17y.dds"},
	{0xDB05A987DAF42798ULL, "llac_ht_brothel_n_3372_0.dds"},
	{0xDB05C555D8C7B8F1ULL, "lt_0_lowtown_docks_a3_n_177_0.mmh"},
	{0xDB05D155D8C7CC92ULL, "lt_0_lowtown_docks_a3_n_177_0.mao"},
	{0xDB05DF55D8C7E4AFULL, "lt_0_lowtown_docks_a3_n_177_0.msh"},
	{0xDB06055D34C89093ULL, "red.dds"},
	{0xDB06104182111172ULL, "hlac_ht_keep_4519_15x.dds"},
	{0xDB0615E8E0E69BC9ULL, "vxlt_gallowstemplar_n_5xe_0_0.vlm"},
	{0xDB06F722D33F910FULL, "rog_paralyzesneakattack.dds"},
	{0xDB071AE196912E53ULL, "mag291_ambient_16217557.cl"},
	{0xDB074DF752D4E3ECULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03b.msh"},
	{0xDB0753F752D4EDDEULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03b.mmh"},
	{0xDB081E690ED3D993ULL, "gen00fl_sebastian6180449.cl"},
	{0xDB089357393EDA06ULL, "llac_lt_0_lowtown_a2_n_c_65536_5.dds"},
	{0xDB08C33B2217FE07ULL, "vxht_mansion2b_2l4_0_0.vlm"},
	{0xDB08FFF6B87EB71AULL, "vxlt_0_lowtown_a2_n_epd_0_1.vlm"},
	{0xDB0A2904A69F5451ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemc_bdy_0s.dds"},
	{0xDB0A5E310543E5B0ULL, "mag121_idunna6084218.cl"},
	{0xDB0A6FCA0F1304D9ULL, "6125949_f.fxe"},
	{0xDB0AED0580E52B3AULL, "6103945_m.fxe"},
	{0xDB0AEE054647A643ULL, "gen_im_trs_gem_opal_03.uti"},
	{0xDB0B2D0D8FE34896ULL, "6141604_m.fxe"},
	{0xDB0B3308C1DC7845ULL, "6000892_m.fxe"},
	{0xDB0C1E05A8056924ULL, "ht_0_hightown_n_45_0.rml"},
	{0xDB0C653AFC4FA4B5ULL, "89444446.ogg"},
	{0xDB0C75B2DBB5F4B2ULL, "551653025.ogg"},
	{0xDB0C99EA5E9878D7ULL, "zz_dog_debug6127747.cl"},
	{0xDB0D6B3B4C5D0154ULL, "one110cr_coterie_thug_2.utc"},
	{0xDB0DCA448DC5AB10ULL, "w_bonepitdestroyed_a3_164_0.mao"},
	{0xDB0DCE448DC5B223ULL, "w_bonepitdestroyed_a3_164_0.mmh"},
	{0xDB0DDC448DC5CA7DULL, "w_bonepitdestroyed_a3_164_0.msh"},
	{0xDB0EFD3B3249B298ULL, "6074977_m.fxe"},
	{0xDB0F47F1204AC50BULL, "w_encounter2_37_0_ll.mmh"},
	{0xDB0F76730C8FF555ULL, "gen00fl_aveline6137472.cl"},
	{0xDB0F8FDDC0B67F77ULL, "vxdr_slavers_23h_0_0.vlm"},
	{0xDB0FA6ADD0F2EC63ULL, "vxlt_0_lowtown_a2_n_c_fh5_0_0.vlm"},
	{0xDB0FCCBB7978EF7CULL, "hlac_ht_keep_3732_17z.dds"},
	{0xDB1020B2BB556A75ULL, "gen00fl_anders6121199.cl"},
	{0xDB104488B579A791ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basic.msh"},
	{0xDB104E269D68AFD8ULL, "vxlt_0_lowtown_a2_n_c_csz_0_0.vlm"},
	{0xDB105288B579BFACULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basic.mao"},
	{0xDB1056EB9978C6D1ULL, "vxdr_primevaltomb_4hl_0_1.vlm"},
	{0xDB105E88B579D44FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basic.mmh"},
	{0xDB109A8B8E10B7EAULL, "mcr310_party_banter6087775.cl"},
	{0xDB1124AE5626C6B3ULL, "6033962_m.fxe"},
	{0xDB1155C71210455DULL, "161883121.ogg"},
	{0xDB11A7D32DC4B479ULL, "follower_banter6170510.cl"},
	{0xDB11E50B0CECA2BEULL, "hlac_ht_mansion2b_818_63y.dds"},
	{0xDB1280E630DE60A9ULL, "226553585.ogg"},
	{0xDB131BCF74493874ULL, "574926919.ogg"},
	{0xDB13546B65A4CCF5ULL, "vxlt_gallowstemplar_f_651_0_0.vlm"},
	{0xDB13AC5F2D4CA39CULL, "vxht_brothel_n_6el_0_0.vlm"},
	{0xDB13CCB7D6698FB6ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.impa.ani"},
	{0xDB143D6927217B38ULL, "vxht_0_hightown_n_c4a_0_0.vlm"},
	{0xDB15073DEFD80395ULL, "smoketemp.mao"},
	{0xDB150F1171AAA54DULL, "vxw_bonepitdestroyed_a3_1vl_0_0.vlm"},
	{0xDB1696C382BBFAA2ULL, "702763544.ogg"},
	{0xDB16B48579021C8AULL, "hlac_lt_0_lowtown_docks_-131074_5x.dds"},
	{0xDB16C4C72C49401CULL, "vxw_blightlands_7z_0_0.vlm"},
	{0xDB16CBB4D0A7256AULL, "vxlt_warehouse_14t_0_0.vlm"},
	{0xDB1725AEA1A80A13ULL, "6052747_m.fxe"},
	{0xDB1777841D559D72ULL, "follower_banter6163506.cl"},
	{0xDB17BD204A4DDB5BULL, "llac_lt_backalley_g_65537_3.dds"},
	{0xDB17D92258D2B8F7ULL, "6169962_m.fxe"},
	{0xDB17DB8B88529BF1ULL, "llac_lt_0_lowtown_a2_-131072_9.dds"},
	{0xDB17F4A0C49CB873ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_roba_0d.dds"},
	{0xDB17FFE0BCD23BA2ULL, "vxca_cavern1_haunted_ck8_0_0.vlm"},
	{0xDB1811A1A28E6146ULL, "llac_lt_backalley_n_1_0.dds"},
	{0xDB1882485EE5D6A3ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03b.mmh"},
	{0xDB1890485EE5EEFDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03b.msh"},
	{0xDB18C9D591972D3DULL, "167870716.ogg"},
	{0xDB195B59B0295684ULL, "plt_lgt000pt_ultra_5_04.plo"},
	{0xDB1964E709091EF3ULL, "gen00fl_sebastian6189077.cl"},
	{0xDB1A198103A54EBCULL, "vxw_bonepitmining_a2_2a9_0_0.vlm"},
	{0xDB1A99FBF2820F25ULL, "117957818.ogg"},
	{0xDB1ABA2B0240AC0EULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs106a.mmh"},
	{0xDB1AC62B0240C0A5ULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs106a.mao"},
	{0xDB1AD42B0240D8FCULL, "art\\characters\\weapons\\greatswords\\w_gsw_gs106a.msh"},
	{0xDB1B30923A338502ULL, "6138789_m.fxe"},
	{0xDB1B71CE5F0AE238ULL, "hlac_w_blightlands_-65540_0x.dds"},
	{0xDB1BE5CDE288A93FULL, "vxlt_gallowstemplar_n_3t1_0_0.vlm"},
	{0xDB1C03BB64870C83ULL, "6127120_m.fxe"},
	{0xDB1C5AB3E4225E57ULL, "hf_qmq_visguard4.mor"},
	{0xDB1CEA6CCD7AC330ULL, "vxht_bartrand_n_5z2_0_0.vlm"},
	{0xDB1D25740170A33FULL, "498758549.ogg"},
	{0xDB1D6592ADB4CCBCULL, "follower_banter6165432.cl"},
	{0xDB1D6858366B79B2ULL, "vxlt_0_lowtown_a3_n_c_exj_0_0.vlm"},
	{0xDB1D8AED3B0C0612ULL, "isa000ip_talisman_container.utp"},
	{0xDB1DB750CEBC2D15ULL, "vxht_keep_6h3_0_1.vlm"},
	{0xDB1DFB880E6ECCAAULL, "6101244_m.fxe"},
	{0xDB1E185F670CA615ULL, "vxca_caverns1_mines_ih4_0_0.vlm"},
	{0xDB1E5529E90DE472ULL, "6174921_m.fxe"},
	{0xDB1E7D2870E66F87ULL, "vxht_bartrand_n_69o_0_0.vlm"},
	{0xDB1EE87F18616716ULL, "combat_smce_r4_a1.gda"},
	{0xDB1F048D439B9D4BULL, "76112802.ogg"},
	{0xDB1F08E9E6A14A37ULL, "gen_im_trs_rdr_hvy_cht_r1.uti"},
	{0xDB1F4224D599A03AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh.gad"},
	{0xDB1F7664F4D9353EULL, "vxw_encounter2_camp_5y_0_0.vlm"},
	{0xDB1F973ED6A93029ULL, "art\\vfx\\combat\\vfx_w_bsk_berserk_activ_c.anb"},
	{0xDB1FEB324BA932B7ULL, "6067840_m.fxe"},
	{0xDB1FF1270E983D53ULL, "one280_emrys6104274.cl"},
	{0xDB2019E886955A2DULL, "vxht_0_hightown_a2_cf9_0_0.vlm"},
	{0xDB2019F90BC2BF9AULL, "235425452.ogg"},
	{0xDB204F4DACB04365ULL, "325336117.ogg"},
	{0xDB2088CFCF0457F5ULL, "zz_followers_banter6157788.cl"},
	{0xDB2115A6CDE8FC5DULL, "6101430_m.fxe"},
	{0xDB21793D63CD5982ULL, "llac_lt_0_lowtown_docks_a3_n_-2_3.dds"},
	{0xDB21D4B7D675DAA7ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.impa.evt"},
	{0xDB22DC4765D16606ULL, "vxht_0_hightown_n_c10_0_0.vlm"},
	{0xDB22F95D351A5306ULL, "6080657_m.fxe"},
	{0xDB22FF37C458329CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabs.mao"},
	{0xDB230B37C458473FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabs.mmh"},
	{0xDB231137C45851E1ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabs.msh"},
	{0xDB236780B95BA90AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1c.mao"},
	{0xDB236B80B95BB059ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1c.mmh"},
	{0xDB237580B95BC147ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1c.msh"},
	{0xDB23F41E57FCB9B5ULL, "845950183.ogg"},
	{0xDB23F687CDAF95A7ULL, "1029297713.ogg"},
	{0xDB24A0D059825549ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_srda_0n.dds"},
	{0xDB24BF102F286958ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzbedroll.phy"},
	{0xDB2512F300B33F35ULL, "vxw_woundedcoast_zz_0_0.vlm"},
	{0xDB2543AF960C69D6ULL, "vxht_brothel_n_6c0_0_0.vlm"},
	{0xDB256F5A8D40DA22ULL, "follower_banter6163876.cl"},
	{0xDB257A82D43A2006ULL, "941207540.ogg"},
	{0xDB258BD90EFEBC64ULL, "vxca_cave_18h_0_0.vlm"},
	{0xDB25A356F71A4F9FULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffy2_d.dds"},
	{0xDB25A6F506042F07ULL, "vxht_mansion2b_2b6_0_0.vlm"},
	{0xDB2632F5F52046EDULL, "vxca_cave_2mm_0_0.vlm"},
	{0xDB264524D59FC451ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh.evt"},
	{0xDB2649040129BA0EULL, "llac_lt_0_lowtown_n_2_0.dds"},
	{0xDB26A54553AB601BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_masa_0s.dds"},
	{0xDB26BBE540F8C45FULL, "darktown_male_fereldan_sta6175135.cl"},
	{0xDB26CA260C084633ULL, "387786004.ogg"},
	{0xDB26E795A0EABE8FULL, "art\\levels\\races\\dwarf\\props\\furniture\\prp_table_1.mao"},
	{0xDB273AF76D247ACCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_up_loop.ani"},
	{0xDB278904A81BD8ABULL, "77016862.ogg"},
	{0xDB27F5B7D67A7EF0ULL, "art\\vfx\\combat\\vfx_glb_att_fail_c.impa.gad"},
	{0xDB282970D0BEB7A5ULL, "hlac_lt_gallowstemplar_f_2390_0x.dds"},
	{0xDB2845FC715E2BA1ULL, "mcr310_argument6074370.cl"},
	{0xDB28E5748D4512A5ULL, "6067837_m.fxe"},
	{0xDB2933C52B925FE9ULL, "hlac_lt_0_lowtown_docks_262142_0z.dds"},
	{0xDB29480D17308BC2ULL, "dre000_varric6027796.cl"},
	{0xDB2983485EF44BC8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye03b.phy"},
	{0xDB29AED4337B7A55ULL, "zz_and_debug6178571.cl"},
	{0xDB29F4B12C3710BDULL, "dae030cr_male_elf.utc"},
	{0xDB2A0D52BDA18B61ULL, "vxlt_darktown_6i6_0_0.vlm"},
	{0xDB2A17B3D2E25204ULL, "vxlt_0_lowtown_docks_a3_n_d2z_0_0.vlm"},
	{0xDB2A57AAD47EC1BFULL, "vxlt_gallowstemplar_n_3s0_0_0.vlm"},
	{0xDB2AA7C849C89563ULL, "vxlt_undercitydungeon_1iq_0_1.vlm"},
	{0xDB2BCB020909FB28ULL, "vxlt_0_lowtown_a2_n_byo_0_0.vlm"},
	{0xDB2C0AE4692A83B2ULL, "6085518_f.fxe"},
	{0xDB2C298CFCCA61EFULL, "mcr371cr_idunna.utc"},
	{0xDB2C9216254091A8ULL, "hlac_ht_bartrand_n_3381_0z.dds"},
	{0xDB2CC8D1F7B9FC3FULL, "vxlt_undercityencounter_zw_0_0.vlm"},
	{0xDB2CE996319CEF11ULL, "vxht_mansion2_1ab_0_0.vlm"},
	{0xDB2D33F6D7A282B7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\teleport.gad"},
	{0xDB2D898D6B3B9684ULL, "6067995_m.fxe"},
	{0xDB2E47F76D2AAF0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\toast_drink_up_loop.gad"},
	{0xDB2E485C731587CFULL, "984615749.ogg"},
	{0xDB2E50069CF82B6FULL, "vxlt_gallowstemplar_f_3di_0_0.vlm"},
	{0xDB2E581A10E04DA2ULL, "qcr220_meredith6060315.cl"},
	{0xDB2E64240712501DULL, "hlac_lt_0_lowtown_a2_131071_1x.dds"},
	{0xDB2E67631E7CBFB7ULL, "vxca_caverns1_gka_0_1.vlm"},
	{0xDB2EB9EED68B615BULL, "vxw_woundedcoast_aq_0_0.vlm"},
	{0xDB2EC1BC9459E9C9ULL, "vxht_bartrand_n_3zq_0_0.vlm"},
	{0xDB2EC95DC83DEBAAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_killbag_n.dds"},
	{0xDB2ED6259EFD1578ULL, "qun231tr_hideout_comment.ncs"},
	{0xDB2EDDD6839ADFABULL, "6166875_m.fxe"},
	{0xDB2F1467924EE731ULL, "hlac_ht_chantry_n_2534_46y.dds"},
	{0xDB2F2CC1764F33DFULL, "6125690_m.fxe"},
	{0xDB2F642E6EA3C2C4ULL, "33387642.ogg"},
	{0xDB2F9B82CF35C513ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hpath2_darelief.dds"},
	{0xDB2FAB2F517D2C54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkback.ani"},
	{0xDB300BDF676FE6FBULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpa.mmh"},
	{0xDB3029DF677019E5ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpa.msh"},
	{0xDB3090191138D49CULL, "vxht_brothel_n_6kb_0_0.vlm"},
	{0xDB30A2CFF6FA304BULL, "vxca_caverns1_ck8_0_0.vlm"},
	{0xDB30B5403EF9D929ULL, "hlac_ca_caverns1_mines_24426_3x.dds"},
	{0xDB30D463A217A102ULL, "566749236.ogg"},
	{0xDB30EEABE428DCA6ULL, "dae000_quintus6154705.cl"},
	{0xDB32022FAA40DE7CULL, "vxca_caverns2prc_a0b_0_0.vlm"},
	{0xDB3211898E8FA779ULL, "74559649.ogg"},
	{0xDB3288D1275044B4ULL, "vxlt_0_lowtown_docks_n_aoc_0_0.vlm"},
	{0xDB329E4C83B96309ULL, "271539916.ogg"},
	{0xDB3329E078E2122AULL, "173403174.ogg"},
	{0xDB332A4D31603067ULL, "mag320_nobleman_26060836.cl"},
	{0xDB333496ACD3554EULL, "133134676.ogg"},
	{0xDB3348F6D7A71244ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\teleport.evt"},
	{0xDB335D24D5AA7720ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_rdy_flrsh.ani"},
	{0xDB3360EAE61DAEF3ULL, "476919232.ogg"},
	{0xDB3384CFF2A957B5ULL, "vxw_woundedcoast_wy_0_0.vlm"},
	{0xDB34031EF98C6820ULL, "dae000ip_crowd_lm4_beg.utp"},
	{0xDB34075486144A48ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_1z.dds"},
	{0xDB34768F6F023D9EULL, "art\\vfx\\creatures\\fxc_arc_shd_c.cess.ani"},
	{0xDB3498385DDE51BEULL, "vxlt_foundry_n_109_0_0.vlm"},
	{0xDB356D468FE98DB5ULL, "vxca_caverns2_4x3_0_0.vlm"},
	{0xDB366731F518AE1FULL, "6017817_f.fxe"},
	{0xDB3670FA49F7DE86ULL, "590153906.ogg"},
	{0xDB367EC1FA65E48BULL, "316798728.ogg"},
	{0xDB36A82F518345B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\walkback.gad"},
	{0xDB370173A3088D13ULL, "hlac_lt_backalley_n_0_0x.dds"},
	{0xDB375D78B302A4A7ULL, "hightown_male_noble_gossip6171380.cl"},
	{0xDB37BC67A1FBC877ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.impa.gad"},
	{0xDB37E938940DD310ULL, "394510543.ogg"},
	{0xDB3842F39887AC87ULL, "808794432.ogg"},
	{0xDB389A1EB994B42FULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkl.gad"},
	{0xDB38C7131E444F51ULL, "799492421.ogg"},
	{0xDB38DA7E31915EE3ULL, "844934609.ogg"},
	{0xDB38FAFBF4715F7BULL, "llac_lt_0_lowtown_docks_a3_n_-4_4.dds"},
	{0xDB3919A53BBEADE3ULL, "vxlt_undercitydungeon_1lr_0_1.vlm"},
	{0xDB392C2179D602DBULL, "plt_prm000pt_irons_belt.nss"},
	{0xDB39BF7FA3116F92ULL, "vxca_caverns2_8ag_0_0.vlm"},
	{0xDB3A09766D17E248ULL, "6147187_m.fxe"},
	{0xDB3A3DA43864AC7DULL, "dae000ip_crowd_mf6_drunk1.utp"},
	{0xDB3A49FBE57C2B4AULL, "hlac_lt_0_lowtown_docks_-131075_6z.dds"},
	{0xDB3A7B0EA3404337ULL, "6153264_m.fxe"},
	{0xDB3AD55946580642ULL, "803740509.ogg"},
	{0xDB3B019380125CEFULL, "hlac_lt_undercityserialkiller_-1_38x.dds"},
	{0xDB3B5DFE0A42BE9BULL, "w_sundermount_142_0.rml"},
	{0xDB3BD6B620D31FB5ULL, "gen00fl_sebastian6173746.cl"},
	{0xDB3BDC8DAD4FCCE0ULL, "6076293_f.fxe"},
	{0xDB3C16608AB3E960ULL, "follower_banter6163427.cl"},
	{0xDB3C2F5D34F667DAULL, "red.tnt"},
	{0xDB3C6628ECAB3AF7ULL, "tbe100ip_to_world_map.utp"},
	{0xDB3CA54758D88A67ULL, "6206061_m.fxe"},
	{0xDB3CF449A716FAE7ULL, "combat_st_mv_default.gda"},
	{0xDB3E0D22EAFCA224ULL, "llac_lt_0_lowtown_a2_n_c_65535_0.dds"},
	{0xDB3E6699DB1376E0ULL, "563216777.ogg"},
	{0xDB3ED167A2020A04ULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.impa.evt"},
	{0xDB3EEF669588753AULL, "6149855_m.fxe"},
	{0xDB3EF0C934995B14ULL, "mcr310_loyalist_templars6076895.cl"},
	{0xDB3F8599B288CBB3ULL, "hlac_lt_house_-1_5x.dds"},
	{0xDB3F9F3D454C1D7BULL, "902093628.ogg"},
	{0xDB3FAF1EB99AF6DCULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkl.evt"},
	{0xDB405E7D1F87BAABULL, "hlac_ht_0_hightown_-196608_2y.dds"},
	{0xDB40A64F1F07ADE1ULL, "6130997_m.fxe"},
	{0xDB40A9C2A2AEB22AULL, "324044461.ogg"},
	{0xDB41064AEE433A08ULL, "100029844.ogg"},
	{0xDB4160F6D7B3783DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\teleport.ani"},
	{0xDB41CCA5AFAAD757ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_lhand.gad"},
	{0xDB41D1776718678AULL, "308986383.ogg"},
	{0xDB42276979F89AD6ULL, "vxht_brothel_n_6i2_0_0.vlm"},
	{0xDB428E8F6F0EA46FULL, "art\\vfx\\creatures\\fxc_arc_shd_c.cess.evt"},
	{0xDB42D31F9741D2BAULL, "vxlt_0_lowtown_a2_n_csz_0_0.vlm"},
	{0xDB4327ADC6FAC0E8ULL, "lgt120_orz_bln_official6140190.cl"},
	{0xDB437550E4DC2325ULL, "6153293_m.fxe"},
	{0xDB4463AB2BF2A5F6ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01enter.ani"},
	{0xDB45038DB55FA9D0ULL, "6102126_m.fxe"},
	{0xDB4541F35E36FE05ULL, "6068458_m.fxe"},
	{0xDB45477B38D603DBULL, "lt_0_lowtown_docks_n_122_0.phy"},
	{0xDB457C0EDE6846F0ULL, "hlac_ht_mansion_n_3373_2x.dds"},
	{0xDB45CC2D8B50E4A1ULL, "demoral_06.mao"},
	{0xDB462C8D15E31EA5ULL, "6166590_m.fxe"},
	{0xDB464AB6FBE2D0C1ULL, "325279054.ogg"},
	{0xDB46ED259F11CE75ULL, "qun231tr_hideout_comment.utt"},
	{0xDB4711F66D7A6706ULL, "dae000_sandal6147785.cl"},
	{0xDB472AF7530A8C79ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear03b.phy"},
	{0xDB47CAF008662C39ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha01a.mmh"},
	{0xDB47D4F008663D27ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha01a.msh"},
	{0xDB480DEFA15CEF2DULL, "static_static_p26.psh.fxo"},
	{0xDB48548D87D966B6ULL, "41964035.ogg"},
	{0xDB488F8F6F131118ULL, "art\\vfx\\creatures\\fxc_arc_shd_c.cess.gad"},
	{0xDB48E1A5AFB119E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_lhand.evt"},
	{0xDB48ECEB95365BB7ULL, "6156164_m.fxe"},
	{0xDB48F87E8A30E871ULL, "859901542.ogg"},
	{0xDB49A5CE96F3D72EULL, "gen_im_arm_cht_lgt_act3_04.uti"},
	{0xDB49A94113082A2EULL, "w_bonepit_157_0.phy"},
	{0xDB49D0A750F3AB79ULL, "dae000_merchant_access_hf6162986.cl"},
	{0xDB4A7BCA54395D25ULL, "art\\vfx\\environment\\vfx_blade_grooves_p.mmh"},
	{0xDB4A89441666D44EULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_3x.dds"},
	{0xDB4AAFFC4064C843ULL, "da2terrain_rendertodepth_p0.vsh.fxo"},
	{0xDB4AFDBF7DC0525EULL, "921421687.ogg"},
	{0xDB4B33C080E45C6FULL, "vxlt_gallowstemplar_f_3nk_0_0.vlm"},
	{0xDB4BD6C3F8BD2106ULL, "gen00fl_varric6133861.cl"},
	{0xDB4BE967A20CBDFDULL, "art\\vfx\\combat\\vfx_glb_cold_die_c.impa.ani"},
	{0xDB4C481E9E413265ULL, "l_lgt251ar_tome_cave.wbd"},
	{0xDB4CAF24ED78C7E2ULL, "6204804_m.fxe"},
	{0xDB4CF051724F0309ULL, "dae000_sabina6161505.cl"},
	{0xDB4D32F75B0679C6ULL, "194541156.ogg"},
	{0xDB4DA68892A6C815ULL, "vxlt_0_lowtown_a2_n_e86_0_0.vlm"},
	{0xDB4DB71EB9A741F5ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_walkl.ani"},
	{0xDB4E225B78B0FC74ULL, "vxlt_gallowsprison_n_6a8_0_0.vlm"},
	{0xDB4E247141BEA502ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_imp_c.mmh"},
	{0xDB4E717801946410ULL, "sword_of_the_deep.dds"},
	{0xDB4EA10965CD7ECDULL, "gen00fl_aveline6101491.cl"},
	{0xDB4EBAD48F1997DBULL, "vxw_bonepit_1wy_0_0.vlm"},
	{0xDB4F247875EEED0CULL, "vxlt_gamlenshouse_n_mg_0_0.vlm"},
	{0xDB4F2AAF444A5E00ULL, "284673921.ogg"},
	{0xDB4F6C8B0B3C75B4ULL, "vxlt_0_lowtown_docks_n_aqa_0_0.vlm"},
	{0xDB4F96AD263BE065ULL, "vxca_caverns1_mines_ihd_0_0.vlm"},
	{0xDB4FFB83D89F0881ULL, "zz_prmpt_debug6198256.cl"},
	{0xDB5054917C6024E3ULL, "vxlt_backalley_g_1kc_0_0.vlm"},
	{0xDB506FD06909BC1CULL, "6131911_m.fxe"},
	{0xDB5081528641B299ULL, "llac_lt_gallowstemplar_2394_10.dds"},
	{0xDB50E9F1E1B0D76AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_left.mao"},
	{0xDB50EDF1E1B0DDB9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_left.mmh"},
	{0xDB50F7F1E1B0EEA7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_left.msh"},
	{0xDB513A367BC4634CULL, "mcr310_party_banter6087812.cl"},
	{0xDB514F3B53A51866ULL, "861790690.ogg"},
	{0xDB515EDFC392E9E2ULL, "hlac_dr_primevaltomb_9621_8z.dds"},
	{0xDB516BAB2BFD3DE7ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01enter.evt"},
	{0xDB51AD92B00B575DULL, "hlac_dr_primevalentrance_1725_4z.dds"},
	{0xDB51CE1078741374ULL, "1068660595.ogg"},
	{0xDB523ABE32AF822FULL, "991556167.ogg"},
	{0xDB52CC8BF7B7A4F9ULL, "527939082.ogg"},
	{0xDB52F61618C4FEA1ULL, "dialog.cur"},
	{0xDB52F8C035F88F94ULL, "6092205_m.fxe"},
	{0xDB538C3E3C2C97F1ULL, "vxlt_0_lowtown_a2_n_coy_0_0.vlm"},
	{0xDB54F2F510ED9F95ULL, "vxca_caverns1_mines_j5h_0_0.vlm"},
	{0xDB550FB69E9481ABULL, "6122732_m.fxe"},
	{0xDB55377F86EF722DULL, "vxca_cavern1_haunted_flw_0_0.vlm"},
	{0xDB554709DBEE02ECULL, "vxlt_hangedman_c_0_0.vlm"},
	{0xDB55F822139E6612ULL, "224856041.ogg"},
	{0xDB55FD825656009DULL, "vxht_0_hightown_a2_ba9_0_0.vlm"},
	{0xDB567059B05D28CBULL, "plt_lgt000pt_ultra_5_04.nss"},
	{0xDB5699E0D1DE47CFULL, "ser191_emeric6108777.cl"},
	{0xDB56A33986FDA092ULL, "vxca_caverns2_8ka_0_0.vlm"},
	{0xDB56F9A5AFBD7FDDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_left_down_lhand.ani"},
	{0xDB56FB43C14090E7ULL, "vxht_bartrand_n_5o6_0_0.vlm"},
	{0xDB5715934D501B6BULL, "974819074.ogg"},
	{0xDB574389664D8290ULL, "833460664.ogg"},
	{0xDB5819D1EDB1D026ULL, "364220784.ogg"},
	{0xDB5819E1E6BC6608ULL, "6092227_m.fxe"},
	{0xDB588CAB2C039430ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_cast01enter.gad"},
	{0xDB588F2658DD98F3ULL, "weapon1_nolit.psh.fxo"},
	{0xDB58BFBB8B0FEBE5ULL, "416830177.ogg"},
	{0xDB59574D31B46F1EULL, "gen00fl_isabela6079293.cl"},
	{0xDB598D61F9EE128FULL, "w_encounter1_mmap_ph.dds"},
	{0xDB59C828777ED231ULL, "236515005.ogg"},
	{0xDB59E288F1119132ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_dt_smoke_lrg.mmh"},
	{0xDB5A0F85AC6DAADBULL, "vxw_woundedcoast_pr_0_0.vlm"},
	{0xDB5A57ADC3B0D049ULL, "6077911_m.fxe"},
	{0xDB5AEE7D26EB2B52ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medb_0d.dds"},
	{0xDB5AF1CC98B60B89ULL, "6002784_m.fxe"},
	{0xDB5BD3803106D5CFULL, "vxht_mansion2b_1qt_0_0.vlm"},
	{0xDB5BE0766D37DEEFULL, "lgt310_mag_grigor6153028.cl"},
	{0xDB5CD3A8F63701B8ULL, "hlac_ca_caverns1_mines_24424_18y.dds"},
	{0xDB5CDAF854F5BFC4ULL, "dae000_surprises6162510.cl"},
	{0xDB5D48622B791FCDULL, "vxw_bonepitdestroyed_a3_1og_0_0.vlm"},
	{0xDB5EC66E6E563D63ULL, "vxlt_backalley_g_1pb_0_0.vlm"},
	{0xDB5EE0019644F90BULL, "91585379.ogg"},
	{0xDB5F03AA6F588386ULL, "391494767.ogg"},
	{0xDB5F47B7A06218DCULL, "6109744_f.fxe"},
	{0xDB6033EE3A39616CULL, "vrc321_party_barks6096300.cl"},
	{0xDB606CFD87BF199EULL, "6215886_m.fxe"},
	{0xDB60CB9EA4DD88EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_action.gad"},
	{0xDB60F0700B09CCF9ULL, "dae000_flemeth6013412.cl"},
	{0xDB6142893FB138D4ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_014a_0s.dds"},
	{0xDB6236F11F0B30C8ULL, "20139964.ogg"},
	{0xDB6260CDD56D689DULL, "6140284_m.fxe"},
	{0xDB62EFCBA79105A1ULL, "862397497.ogg"},
	{0xDB6375B8DACFE135ULL, "424046430.ogg"},
	{0xDB63FDAA4C6881EFULL, "517186656.ogg"},
	{0xDB64547FB93171A0ULL, "6069441_m.fxe"},
	{0xDB6468C19F687093ULL, "zz_mer_debug6045081.cl"},
	{0xDB64CA1A90232ED7ULL, "300768616.ogg"},
	{0xDB64CDE1E38EF380ULL, "279102609.ogg"},
	{0xDB650BC0CC97EC92ULL, "vxht_0_hightown_ban_0_0.vlm"},
	{0xDB65B0F24F4892E2ULL, "vxw_bonepitdestroyed_a3_4b_0_0.vlm"},
	{0xDB660719432C3F0FULL, "zz_isa_debug6133981.cl"},
	{0xDB664DCA2D406663ULL, "hlac_ht_brothel_n_3366_0x.dds"},
	{0xDB66C3FDB854ED1FULL, "lt_0_lowtown_37_0_lc.mmh"},
	{0xDB671C2DC7DD3BF2ULL, "gen00fl_sebastian6180437.cl"},
	{0xDB676962C07183FBULL, "pro000_meeran6050073.cl"},
	{0xDB678032F19B6C9CULL, "6139256_m.fxe"},
	{0xDB67A6D135FEFE16ULL, "vxlt_warehouse_n_14c_0_0.vlm"},
	{0xDB67AE9EA4E37581ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_action.evt"},
	{0xDB6839B9B2CCED96ULL, "24803274.ogg"},
	{0xDB68E1C09E0BE886ULL, "vxca_caverns2_79l_0_0.vlm"},
	{0xDB68E3280885554DULL, "243165147.ogg"},
	{0xDB69372179FF3974ULL, "plt_prm000pt_irons_belt.plo"},
	{0xDB6A039C36F13CD5ULL, "gen00fl_anders6114472.cl"},
	{0xDB6A455D22FCEE4FULL, "vxht_mansion2b_1tw_0_0.vlm"},
	{0xDB6AE10B43BC6F53ULL, "218736877.ogg"},
	{0xDB6B28269F0BEE11ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\blendtree\\floortrap.anb"},
	{0xDB6B7AF2D047C1A2ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.cess.gad"},
	{0xDB6B8EEADE271B03ULL, "ht_mansion2b_818_lc.mmh"},
	{0xDB6B90F2DAD90F13ULL, "dae230ip_at_to_sewers1.utp"},
	{0xDB6BE946B2F1E27EULL, "vxca_caverns2_71l_0_0.vlm"},
	{0xDB6BFEFEB16788EFULL, "6066379_m.fxe"},
	{0xDB6BFFF4A4EBF3BBULL, "w_bonepit_5_0.msh"},
	{0xDB6C0B415F4CE0D1ULL, "349555055.ogg"},
	{0xDB6C11F4A4EC1206ULL, "w_bonepit_5_0.mao"},
	{0xDB6C1DF4A4EC26A5ULL, "w_bonepit_5_0.mmh"},
	{0xDB6CC130200EE2B8ULL, "6119759_m.fxe"},
	{0xDB6CC70BB7CD58AFULL, "6153509_m.fxe"},
	{0xDB6CD515D81DE9ACULL, "hlac_lt_0_lowtown_a3_n_c_-196608_4x.dds"},
	{0xDB6E3B3373A3A4E0ULL, "vxht_mansion_n_5m8_0_0.vlm"},
	{0xDB6E76FF99A63150ULL, "6149869_f.fxe"},
	{0xDB6EAF1E3414B080ULL, "vxht_mansion2b_1h0_0_0.vlm"},
	{0xDB6EEF0D04DEC4B7ULL, "vxlt_0_lowtown_docks_a3_d0e_0_1.vlm"},
	{0xDB6FA25B56AEA9B8ULL, "one120_athenril6113277.cl"},
	{0xDB6FB8A175E29F6CULL, "qun221_petrice6077420.cl"},
	{0xDB6FDD21F3C38905ULL, "and100_anders6095993.cl"},
	{0xDB6FF9C7FB55043DULL, "vxht_mansion2b_2jl_0_0.vlm"},
	{0xDB70BAD48C1EAF1EULL, "912637460.ogg"},
	{0xDB712E7B38FB147AULL, "lt_0_lowtown_docks_n_122_0.msh"},
	{0xDB713C7B38FB2BB3ULL, "lt_0_lowtown_docks_n_122_0.mao"},
	{0xDB71407B38FB32C0ULL, "lt_0_lowtown_docks_n_122_0.mmh"},
	{0xDB71C82B21D2CD85ULL, "hlac_lt_backalley_n_0_19x.dds"},
	{0xDB71E45239D0F4B9ULL, "vxlt_foundry_n_yn_0_0.vlm"},
	{0xDB71FAD9C72CE2F1ULL, "vxw_woundedcoast_189_0_0.vlm"},
	{0xDB7255F5A790084EULL, "6126730_f.fxe"},
	{0xDB728A5E54A672AAULL, "vxht_keep_n_62d_0_0.vlm"},
	{0xDB729DF2D04E1B99ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.cess.evt"},
	{0xDB730E31F4F06D65ULL, "485031545.ogg"},
	{0xDB734B1160C8617AULL, "6146898_m.fxe"},
	{0xDB7352CB31E2D2A9ULL, "llac_w_woundedcoast_-196605_0.dds"},
	{0xDB73CCB33277BAB4ULL, "mag360_samson.cnv"},
	{0xDB73FF9DA2D861BCULL, "6117968_m.fxe"},
	{0xDB7463638E9256F3ULL, "6164208_m.fxe"},
	{0xDB748E25E4E20C11ULL, "186698135.ogg"},
	{0xDB75692D868E9492ULL, "llac_lt_0_lowtown_a2_n_-131072_1.dds"},
	{0xDB75701251A61A32ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.dura.gad"},
	{0xDB75A69EA4EFA5D0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v2_action.ani"},
	{0xDB75A8B7BD5E30C4ULL, "897859396.ogg"},
	{0xDB7618AE27F51696ULL, "vxlt_warehouse_n_19d_0_0.vlm"},
	{0xDB76237141DFF5A9ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_imp_c.anb"},
	{0xDB765BBFE0A1AFFAULL, "6065344_m.fxe"},
	{0xDB7688DD859DDF5EULL, "6078180_m.fxe"},
	{0xDB76B884BD4E6D95ULL, "782271972.ogg"},
	{0xDB76D5B26CD96976ULL, "gen00fl_aveline6128402.cl"},
	{0xDB770EFA34B8C9F9ULL, "6046836_m.fxe"},
	{0xDB772CDF67AC2130ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_tmpa.phy"},
	{0xDB7766630152F85FULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.cess.gad"},
	{0xDB77AEB3327B70F4ULL, "mag360_samson.bnk"},
	{0xDB77CE0572468E95ULL, "one300dg_lia6064523.cl"},
	{0xDB78288D4E3C571DULL, "isa111_isabela6075614.cl"},
	{0xDB782BE20B509A25ULL, "vxlt_0_lowtown_a2_n_c_gm1_0_0.vlm"},
	{0xDB7839E3C7A9C9D2ULL, "vxlt_gallowscourtyard_n_3tz_0_0.vlm"},
	{0xDB78AB5AB9B4E0A0ULL, "605472594.ogg"},
	{0xDB790A4158FF4E4BULL, "1049670309.ogg"},
	{0xDB7979F74B0C2DC6ULL, "zz_followers_banter6157965.cl"},
	{0xDB79A1667822F20BULL, "hlac_lt_0_lowtown_a3_n_c_-262144_0x.dds"},
	{0xDB7A22F9F5C7CF6AULL, "664757285.ogg"},
	{0xDB7A69B81C11AD5EULL, "vxdr_primevaltomb_6j6_0_0.vlm"},
	{0xDB7A6E3705E8BCD0ULL, "nat380_nathaniel_banter6067311.cl"},
	{0xDB7AF2AD09F4A6B0ULL, "hlac_ca_caverns2_5802_55y.dds"},
	{0xDB7B6328A0825971ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_critical_strike.ani"},
	{0xDB7C2FF8F3FB3926ULL, "ave310cr_bandit_f.utc"},
	{0xDB7C6335561B4C89ULL, "6116745_f.fxe"},
	{0xDB7C682C5B5E8455ULL, "gen00fl_varric6063848.cl"},
	{0xDB7C731251AC3E69ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.dura.evt"},
	{0xDB7CACF767A68D81ULL, "fx_lightningbolt01.dds"},
	{0xDB7D20FB260AC900ULL, "vxht_mansion2b_1c1_0_0.vlm"},
	{0xDB7D6BDEC39AB25AULL, "vxlt_0_lowtown_docks_a3_n_cq6_0_0.vlm"},
	{0xDB7D96F4F1E237BEULL, "w_woundedcoast_11_0.rml"},
	{0xDB7DE7467B63C358ULL, "746289888.ogg"},
	{0xDB7DEEDD750D8FAFULL, "hlac_lt_0_lowtown_n_-131072_0z.dds"},
	{0xDB7E05214E92DC31ULL, "6042788_m.fxe"},
	{0xDB7E5B630159044CULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.cess.evt"},
	{0xDB7E8B2D24EA4D22ULL, "vxlt_0_lowtown_a2_n_c_e8k_0_0.vlm"},
	{0xDB7E9D7CB98D2CC8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thinrock.msh"},
	{0xDB7EAF7CB98D4B59ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thinrock.mao"},
	{0xDB7EBB7CB98D5FF2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thinrock.mmh"},
	{0xDB7EE8CEEAD866A8ULL, "37126973.ogg"},
	{0xDB7EF5664D38A86FULL, "hlac_lt_shop_-1_5z.dds"},
	{0xDB7F24D651D60265ULL, "vxlt_0_lowtown_a2_n_e8f_0_0.vlm"},
	{0xDB7FA5F2D058B388ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_ogr_p.cess.ani"},
	{0xDB80534E056313A4ULL, "953285336.ogg"},
	{0xDB80562F2BC70D39ULL, "vxlt_foundry_n_tm_0_0.vlm"},
	{0xDB808589CB16831FULL, "hlac_dr_primevaltomb_664_7x.dds"},
	{0xDB809C15C79600C2ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_1y.dds"},
	{0xDB80AC102F76A7EDULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzbedroll.msh"},
	{0xDB80BA102F76BF20ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzbedroll.mao"},
	{0xDB80BE102F76C673ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zzbedroll.mmh"},
	{0xDB81BCEE52BE79FAULL, "6143798_m.fxe"},
	{0xDB81EF7AB0841D92ULL, "vxht_0_hightown_bwh_0_0.vlm"},
	{0xDB81F8D05CDFEA12ULL, "vxca_caverns2_8xl_0_0.vlm"},
	{0xDB826E28A0888A13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_critical_strike.gad"},
	{0xDB8286DBC815CDC3ULL, "gen00fl_isabela_init6077297.cl"},
	{0xDB831142EFBCB0F3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_paintings_symbol_da.dds"},
	{0xDB831F920FEAA6A0ULL, "6094548_m.fxe"},
	{0xDB83CFDDEE33B6B5ULL, "vxht_0_hightown_a2_ca4_0_0.vlm"},
	{0xDB8463134CFB48C2ULL, "vxlt_warehouse_n_9s_0_0.vlm"},
	{0xDB84A4E3B9D27F9EULL, "855509048.ogg"},
	{0xDB851437C4AB2084ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_slabs.phy"},
	{0xDB85153272D25833ULL, "vxht_mansion2b_21s_0_0.vlm"},
	{0xDB8579875FEB6F85ULL, "llac_ht_0_hightown_a2_-196608_1.dds"},
	{0xDB85EDB5DCBB4C3BULL, "vxw_sundermount_32c_0_0.vlm"},
	{0xDB863B8813B4DB76ULL, "vxw_sundermount_2zv_0_1.vlm"},
	{0xDB872BF4075091B2ULL, "hlac_lt_gallowscourtyard_-131072_1x.dds"},
	{0xDB8771CDBAF76AB6ULL, "zz_followers_banter6157275.cl"},
	{0xDB87A8BF422B6ADAULL, "222984124.ogg"},
	{0xDB87B5B5B7F0753BULL, "6102806_m.fxe"},
	{0xDB8830FCD5DD4C89ULL, "389319790.ogg"},
	{0xDB88B12179F45324ULL, "hlac_lt_gallowsprison_n_-1_1x.dds"},
	{0xDB88DB950E07C5DEULL, "vxdr_primevaltomb_6q7_0_0.vlm"},
	{0xDB8930EBB874DB7FULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastgrass_da.dds"},
	{0xDB89314400532E30ULL, "vxht_mansion2b_1br_0_0.vlm"},
	{0xDB895B1251B69FD8ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.dura.ani"},
	{0xDB896B28A08EA478ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_critical_strike.evt"},
	{0xDB8972158786F8C7ULL, "ht_mansion_n_mmi.gff"},
	{0xDB89A841133E0CF1ULL, "w_bonepit_157_0.mmh"},
	{0xDB89B441133E2092ULL, "w_bonepit_157_0.mao"},
	{0xDB89BC3879E30FF6ULL, "596746931.ogg"},
	{0xDB89C241133E38AFULL, "w_bonepit_157_0.msh"},
	{0xDB89DFF465BE9568ULL, "vis121_saemus6118469.cl"},
	{0xDB8A8C32C0F58248ULL, "6040974_m.fxe"},
	{0xDB8AD2B6A358CD0FULL, "6152605_m.fxe"},
	{0xDB8B13E0442117AEULL, "840559560.ogg"},
	{0xDB8B4363016365A5ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.cess.ani"},
	{0xDB8BCC3CE579CAF9ULL, "vxht_mansion2_22p_0_0.vlm"},
	{0xDB8BDDBBB590CADAULL, "vxlt_0_lowtown_docks_a3_n_cz7_0_0.vlm"},
	{0xDB8C221E9E76D50FULL, "l_lgt251ar_tome_cave.bnk"},
	{0xDB8C67BFC31FB267ULL, "pro000_cut_act_of_mercy6001031.cl"},
	{0xDB8C9751F824673AULL, "vxw_woundedcoast_155_0_0.vlm"},
	{0xDB8CA67AAEE43A0DULL, "901478099.ogg"},
	{0xDB8D2F43D7ED2F13ULL, "qun200pt_blackform2.ncs"},
	{0xDB8E2C9B5BDE13ABULL, "vxlt_0_lowtown_a3_n_c_fff_0_0.vlm"},
	{0xDB8E62501601C8C3ULL, "pro000_opening6178542.cl"},
	{0xDB8E821EFD9555BCULL, "170933672.ogg"},
	{0xDB8F1DC8D825E4D7ULL, "118277908.ogg"},
	{0xDB8F34FA1EE5833CULL, "hlac_lt_0_lowtown_docks_196607_3z.dds"},
	{0xDB8F50839FB1C9F8ULL, "667870390.ogg"},
	{0xDB8F5EC9755265C9ULL, "6155673_m.fxe"},
	{0xDB8FCB445E581603ULL, "vxlt_warehouse_116_0_0.vlm"},
	{0xDB8FF32865D7B24FULL, "vxdr_primevalentrance_2q8_0_2.vlm"},
	{0xDB900F7F97C3AE74ULL, "6180807_m.fxe"},
	{0xDB90182B1FB80823ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishock.anb"},
	{0xDB906157A27A3612ULL, "vxht_0_hightown_bpi_0_0.vlm"},
	{0xDB90696591394749ULL, "vxht_keep_7ad_0_0.vlm"},
	{0xDB909D8874F465C6ULL, "llac_ca_caverns2_5802_30.dds"},
	{0xDB909E58364F6A5AULL, "zz_zz.cut"},
	{0xDB91612749402040ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hat_noba_0n.dds"},
	{0xDB91A6B294A879CAULL, "132685571.ogg"},
	{0xDB91AAA1F42FE1C4ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_003a_0d.dds"},
	{0xDB91B4EE108B5A34ULL, "vxlt_0_lowtown_a2_n_ewr_0_0.vlm"},
	{0xDB91C44A4C96EB5CULL, "vxht_0_hightown_a3_8rr_0_0.vlm"},
	{0xDB92E1EC25828DC3ULL, "147844658.ogg"},
	{0xDB93035A27AEABBBULL, "lt_0_lowtown_docks_a3_n_113_0.mmh"},
	{0xDB930F5A27AEC058ULL, "lt_0_lowtown_docks_a3_n_113_0.mao"},
	{0xDB93215A27AEDEA5ULL, "lt_0_lowtown_docks_a3_n_113_0.msh"},
	{0xDB93D0FDBD669719ULL, "70589532.ogg"},
	{0xDB942E82530E5055ULL, "one300_zevran6089723.cl"},
	{0xDB945E66B25DCEBFULL, "vxht_0_hightown_a2_aoz_0_0.vlm"},
	{0xDB945F92CEB164BBULL, "vxw_sundermount_39b_0_0.vlm"},
	{0xDB94DDBAF79C140DULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockaxxl.msh"},
	{0xDB94E7DE489A945FULL, "798433922.ogg"},
	{0xDB94EBBAF79C2C40ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockaxxl.mao"},
	{0xDB94EFBAF79C3293ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_kcrockaxxl.mmh"},
	{0xDB95864A093D49D0ULL, "6085977_m.fxe"},
	{0xDB96B3627E5B0DCDULL, "6131099_m.fxe"},
	{0xDB970AD88A4FAF5FULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand_3.mao"},
	{0xDB97D0BC7CD22D46ULL, "isa211_isabela6069444.cl"},
	{0xDB980759B1263360ULL, "vxw_encounter3_w_0_0.vlm"},
	{0xDB9837F3C1CB2FEFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_5.pwk"},
	{0xDB98AB53BFA30E59ULL, "hm_gamlen_h1s1.mor"},
	{0xDB98C3BA4250EFC2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lute_d.dds"},
	{0xDB98DEEFF56EB2D6ULL, "defaultstate.sta.xml"},
	{0xDB990F670BD681FCULL, "vxw_woundedcoast_1cy_0_0.vlm"},
	{0xDB99BDBAD85F4991ULL, "hlac_w_encounter2_camp_0_0y.dds"},
	{0xDB99CB84E6A6AECBULL, "488029208.ogg"},
	{0xDB9A115848A37006ULL, "6080446_m.fxe"},
	{0xDB9A4080E6DE47D6ULL, "6110582_f.fxe"},
	{0xDB9A5CC06AAF3318ULL, "vxht_0_hightown_n_akb_0_0.vlm"},
	{0xDB9AD10A7C43307AULL, "gen00fl_merrill6128762.cl"},
	{0xDB9B092EEA1A7FBAULL, "vxw_woundedcoast_106_0_0.vlm"},
	{0xDB9B39168E43E39DULL, "vxlt_0_lowtown_a2_hf6_0_0.vlm"},
	{0xDB9B6EE708D67E22ULL, "vxlt_0_lowtown_a2_n_c_e2u_0_0.vlm"},
	{0xDB9C0620DB6D1143ULL, "1000473040.ogg"},
	{0xDB9C5F4095396F1AULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_containera_sa.dds"},
	{0xDB9C6752F4FD93E5ULL, "mer200cr_ambush_leader.utc"},
	{0xDB9CE4273A50E818ULL, "llac_lt_0_lowtown_a2_n_c_65534_5.dds"},
	{0xDB9D94047120E8C4ULL, "6095473_m.fxe"},
	{0xDB9DC26E2323453FULL, "134655959.ogg"},
	{0xDB9E3D21504E2E83ULL, "vxlt_warehouse_149_0_0.vlm"},
	{0xDB9E643DAB34B865ULL, "vrc221st_bartrand.stg"},
	{0xDB9EF58A6981ED03ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_lrg.anb"},
	{0xDB9F3832A0C03D30ULL, "vxlt_0_lowtown_a2_n_gyf_0_0.vlm"},
	{0xDB9FC8226ECD48CAULL, "370905318.ogg"},
	{0xDBA026CB028172B4ULL, "vxlt_0_lowtown_a2_n_exs_0_0.vlm"},
	{0xDBA078D8026E415FULL, "art\\toolset\\characters\\playercharacter\\dwarfmale\\mt_art_dm00.phy"},
	{0xDBA0C1C421CEC1E3ULL, "6082993_m.fxe"},
	{0xDBA0D2C6D655A237ULL, "352997891.ogg"},
	{0xDBA20220B6716668ULL, "1027044408.ogg"},
	{0xDBA2EA182AF30AA8ULL, "27188372.ogg"},
	{0xDBA350F62055B6C0ULL, "339406676.ogg"},
	{0xDBA45DFE0A9BA868ULL, "w_sundermount_142_0.mmh"},
	{0xDBA469FE0A9BBC0BULL, "w_sundermount_142_0.mao"},
	{0xDBA47BFE0A9BDA92ULL, "w_sundermount_142_0.msh"},
	{0xDBA5444AB2A29310ULL, "framebuffereffects_motionblurcombine_p0.psh.fxo"},
	{0xDBA5E396628FE4D3ULL, "525984187.ogg"},
	{0xDBA614FDE43F5F30ULL, "vxht_mansion2b_1xp_0_0.vlm"},
	{0xDBA6A36DA7DDA83EULL, "dae000_cora6162357.cl"},
	{0xDBA73AA9014FA05EULL, "608049124.ogg"},
	{0xDBA775EDB263055CULL, "6076359_m.fxe"},
	{0xDBA7ED499FA94049ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_1x.dds"},
	{0xDBA7F8C217C7FD9BULL, "205777249.ogg"},
	{0xDBA894AD505EADD2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanright.gad"},
	{0xDBA8964D61D3DB13ULL, "6137231_m.fxe"},
	{0xDBA89DCA0EEF6968ULL, "712350910.ogg"},
	{0xDBA8A3F280D3AD02ULL, "l_dae321ar_chantry.bnk"},
	{0xDBA8AFF6C965FBF9ULL, "vxht_mansion2_28v_0_0.vlm"},
	{0xDBA8CDE39918B5F9ULL, "6207746_f.fxe"},
	{0xDBA95CFF266C19FBULL, "783269080.ogg"},
	{0xDBA98AF4F2075E36ULL, "w_woundedcoast_11_0.mao"},
	{0xDBA98CE1C905606DULL, "vxlt_0_lowtown_a3_n_c_i1b_0_0.vlm"},
	{0xDBA996F4F20772D5ULL, "w_woundedcoast_11_0.mmh"},
	{0xDBA998F4F207764BULL, "w_woundedcoast_11_0.msh"},
	{0xDBA9CBF008B8F826ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha01a.phy"},
	{0xDBA9CF8899AA3CDAULL, "6126720_m.fxe"},
	{0xDBAAF30A4AAB58B6ULL, "vxlt_gallowstemplar_6qi_0_0.vlm"},
	{0xDBAB0ED20ECCC960ULL, "6105435_m.fxe"},
	{0xDBAB25F370F7D3D1ULL, "605998039.ogg"},
	{0xDBAB68C2FF82A02BULL, "art\\vfx\\combat\\vfx_itm_myt_favor_aoe_p.impa.ani"},
	{0xDBAC1DF4A5220AAEULL, "w_bonepit_5_0.rml"},
	{0xDBAC8645CEA5180FULL, "39932740.ogg"},
	{0xDBACC721F8CA08DBULL, "6111279_m.fxe"},
	{0xDBACD28A698DEF2CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_lavajet_lrg.mmh"},
	{0xDBAD35800BCF8DDAULL, "gen00fl_sebastian6174827.cl"},
	{0xDBAD502173AE092DULL, "vxca_caverns2_7ca_0_0.vlm"},
	{0xDBAD630D63CD29F8ULL, "102252309.ogg"},
	{0xDBAD7FB39CDD929FULL, "qun110_petrice6042132.cl"},
	{0xDBAE97AD50631E09ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanright.evt"},
	{0xDBAEBECA6D5B6842ULL, "27555325.ogg"},
	{0xDBAEE561FA39EB51ULL, "hlac_lt_blackemporium_1055_5z.dds"},
	{0xDBAF42596F311952ULL, "zz_vault_debug6203198.cl"},
	{0xDBAF8AC4B747D8D3ULL, "254539254.ogg"},
	{0xDBB07884C291B652ULL, "fen310_party_comment.cnv"},
	{0xDBB0A3A28DDA4C1CULL, "740216510.ogg"},
	{0xDBB1204E48BCC865ULL, "cli_army_ds_fortroof.gda"},
	{0xDBB14089539F261FULL, "170267389.ogg"},
	{0xDBB179D8027CB664ULL, "art\\toolset\\characters\\playercharacter\\dwarfmale\\mt_art_dm00.mmh"},
	{0xDBB18FD8027CDB06ULL, "art\\toolset\\characters\\playercharacter\\dwarfmale\\mt_art_dm00.msh"},
	{0xDBB1949F071B98E3ULL, "acl_1p.stg"},
	{0xDBB20D295973608EULL, "6118642_m.fxe"},
	{0xDBB23F1489A2F23EULL, "vxht_brothel_n_75c_0_0.vlm"},
	{0xDBB285E41BCE3792ULL, "35278508.ogg"},
	{0xDBB2C4CB961BB42BULL, "qun211_party_comment6072272.cl"},
	{0xDBB2EEF1E203A9A6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_left.phy"},
	{0xDBB34268C8FD962AULL, "gen00fl_merrill6121758.cl"},
	{0xDBB39A84C29425AAULL, "fen310_party_comment.bnk"},
	{0xDBB3B1D024789F55ULL, "w_bonepitdestroyed_a3_230_0.msh"},
	{0xDBB3B3D02478A2CBULL, "w_bonepitdestroyed_a3_230_0.mmh"},
	{0xDBB3BFD02478B768ULL, "w_bonepitdestroyed_a3_230_0.mao"},
	{0xDBB4222460138A3FULL, "llac_dr_slavers_3214_10.dds"},
	{0xDBB42E56829F093DULL, "vxlt_blackemporium_y9_0_0.vlm"},
	{0xDBB4670D19F5AE95ULL, "vxlt_warehouse_aa_0_0.vlm"},
	{0xDBB484BB6B0107D6ULL, "409904905.ogg"},
	{0xDBB486DAD63577B0ULL, "vxht_mansion2b_1sq_0_0.vlm"},
	{0xDBB4C1D7668727BAULL, "nat380_sibling6103580.cl"},
	{0xDBB5672BAFB1D5B7ULL, "vxlt_0_lowtown_n_ewy_0_0.vlm"},
	{0xDBB57741B4E41AC6ULL, "456820492.ogg"},
	{0xDBB58158B1718CBFULL, "vxlt_0_lowtown_n_do0_0_0.vlm"},
	{0xDBB5908B6DC143B9ULL, "hlac_ht_mansion2b_818_38z.dds"},
	{0xDBB59A5B427357E8ULL, "752782760.ogg"},
	{0xDBB647D8ABF894FCULL, "6075224_m.fxe"},
	{0xDBB6593ED5D08996ULL, "vxht_0_hightown_a2_9xb_0_0.vlm"},
	{0xDBB67830B7885AD9ULL, "6180621_f.fxe"},
	{0xDBB7004BA5502DD5ULL, "hlac_ht_mansion2_818_46y.dds"},
	{0xDBB72E854366028EULL, "6174878_f.fxe"},
	{0xDBB7B947D8A9FDC0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_treebark_sa.dds"},
	{0xDBB7C439680B8677ULL, "vxw_sundermount_32w_0_0.vlm"},
	{0xDBB7E6E0BFE3C824ULL, "597148499.ogg"},
	{0xDBB895A1674DE409ULL, "271237458.ogg"},
	{0xDBB8CFEA91ED0B7FULL, "one350_vinell6087490.cl"},
	{0xDBB8D1F290B5CB34ULL, "vxca_caverns1_mines_i1z_0_0.vlm"},
	{0xDBB8E365F846A253ULL, "gen00fl_fenris6103357.cl"},
	{0xDBB950C2FF8EB4A2ULL, "art\\vfx\\combat\\vfx_itm_myt_favor_aoe_p.impa.evt"},
	{0xDBB9BB72BC8A2625ULL, "qcr270ar_gallows.are"},
	{0xDBB9F05A003BB7F1ULL, "hlac_ht_0_hightown_-65537_3x.dds"},
	{0xDBBA29A59D2D1D00ULL, "lt_0_lowtown_a2_n_c_18_0.rml"},
	{0xDBBA996A3A96DCEBULL, "8753464.ogg"},
	{0xDBBB4C9D4F4752F3ULL, "vxw_encounter1_dr_0_0.vlm"},
	{0xDBBB754B35E7F497ULL, "vxht_brothel_n_6f2_0_0.vlm"},
	{0xDBBBE44C44CE0AC6ULL, "404301732.ogg"},
	{0xDBBC00E7572863A1ULL, "follower_banter6136372.cl"},
	{0xDBBC14802BB1F75CULL, "qun221_petrice6083759.cl"},
	{0xDBBC6743A9046C2FULL, "496423081.ogg"},
	{0xDBBC7FAD506F3378ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\runfrontleanright.ani"},
	{0xDBBCA6A784C67A76ULL, "lightning_stone.dds"},
	{0xDBBCA6BEC8363D32ULL, "vxca_cavern1_haunted_gh9_0_0.vlm"},
	{0xDBBCFDC011240E79ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2win.msh"},
	{0xDBBD07C011241F67ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2win.mmh"},
	{0xDBBD0BC0112425B4ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2win.mao"},
	{0xDBBDA1C837526B82ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_2.gad"},
	{0xDBBDD01269360AD4ULL, "hightown_male_noble_gossip6174056.cl"},
	{0xDBBE4A1763F60023ULL, "hlac_lt_0_lowtown_a3_n_c_-1_2x.dds"},
	{0xDBBEF6D108095115ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_2y.dds"},
	{0xDBBFA871207C1D66ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v4.ani"},
	{0xDBC028FB58D30CCEULL, "554084035.ogg"},
	{0xDBC04FC2FF94D1B5ULL, "art\\vfx\\combat\\vfx_itm_myt_favor_aoe_p.impa.gad"},
	{0xDBC07F064B69790EULL, "6117742_m.fxe"},
	{0xDBC1A85B80BF5B3DULL, "lt_0_lowtown_n_27_0_lc.mmh"},
	{0xDBC1DC0C22926C6EULL, "vxlt_undercityencounter_110_0_0.vlm"},
	{0xDBC20F3B171B230EULL, "dre000_bodahn6031275.cl"},
	{0xDBC224171F523A3EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch01.evt"},
	{0xDBC282D712F19F63ULL, "6053270_m.fxe"},
	{0xDBC3BC72BC927C0BULL, "qcr270ar_gallows.ncs"},
	{0xDBC41B2283D2661CULL, "vxw_blightlands_sv_0_0.vlm"},
	{0xDBC47577D51E883AULL, "vxca_caverns1_fmf_0_0.vlm"},
	{0xDBC4C4C83758C679ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_2.evt"},
	{0xDBC57C44887C95BAULL, "750329947.ogg"},
	{0xDBC5AD04AD3DD3E6ULL, "follower_barks6153292.cl"},
	{0xDBC5D0DCE53ABB7EULL, "vxht_0_hightown_a2_cd4_0_0.vlm"},
	{0xDBC6B3427D966DDAULL, "6128520_m.fxe"},
	{0xDBC6F0CBF992C364ULL, "llac_w_bonepitdestroyed_a3_327682_0.dds"},
	{0xDBC7246BA1BF20C8ULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\small_d.dds"},
	{0xDBC724F9A7822BBEULL, "vxca_caverns1_jgt_0_0.vlm"},
	{0xDBC791414D52FEB9ULL, "gen00fl_merrill6121135.cl"},
	{0xDBC849A82A50C5BEULL, "367328807.ogg"},
	{0xDBC88337F8AC9E8CULL, "vxht_playershouse_n_1sk_0_0.vlm"},
	{0xDBC8D66C3E99D870ULL, "llac_lt_gallowscourtyard_n_-131072_2.dds"},
	{0xDBC8F1B144F05490ULL, "docks_male_elven_worker_sta6175840.cl"},
	{0xDBC93B171F588069ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch01.gad"},
	{0xDBC9699CC54C4CB0ULL, "genip_ballista.utp"},
	{0xDBC9899437CF72C2ULL, "gen00fl_aveline6139707.cl"},
	{0xDBC9BE9949B970A6ULL, "hlac_lt_0_lowtown_docks_a3_-131074_8x.dds"},
	{0xDBC9CE73CE762F59ULL, "mag160_feynriel6204956.cl"},
	{0xDBC9DC617502C3DBULL, "ht_0_hightown_a2_25_0.rml"},
	{0xDBCABC72BC989B52ULL, "qcr270ar_gallows.lst"},
	{0xDBCB85E00925C79AULL, "hlac_w_blightlands_-65540_0z.dds"},
	{0xDBCBC98346451F40ULL, "follower_barks6173410.cl"},
	{0xDBCC3100AE08FF8AULL, "vxw_bonepitdestroyed_a3_99_0_0.vlm"},
	{0xDBCC3395BEE55E28ULL, "226660936.ogg"},
	{0xDBCCBEA8C27EA9FCULL, "static_static_p34.psh.fxo"},
	{0xDBCCFA48BB4CE175ULL, "vxlt_gallowstemplar_n_660_0_0.vlm"},
	{0xDBCD478EC7439DF8ULL, "hlac_w_encounter2_camp_327682_0z.dds"},
	{0xDBCD896117C1E0F5ULL, "hlac_lt_gallowscourtyard_-131071_1z.dds"},
	{0xDBCD8DF280F2A420ULL, "l_dae321ar_chantry.wbd"},
	{0xDBCD9656AE32091BULL, "lgt260_eau_tmp_asst_harbor6143634.cl"},
	{0xDBCDBAC256E99BA0ULL, "dae000_female_tranquil6165130.cl"},
	{0xDBCDFBC361F5D67CULL, "305584812.ogg"},
	{0xDBCE172F144F6A8DULL, "zz_fla.ncs"},
	{0xDBCE61ADD086DD55ULL, "ca_cavern1_haunted_26035_lc.mmh"},
	{0xDBCF5C215E5A71FCULL, "6041397_m.fxe"},
	{0xDBCF859FBBE4E073ULL, "mag311_nyssa6064840.cl"},
	{0xDBCFD3E2B1618477ULL, "llac_lt_gallowstemplar_f_2382_3.dds"},
	{0xDBD00C171F5E4EB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\twitch01.ani"},
	{0xDBD0112EB7801188ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemon_head.mao"},
	{0xDBD08B3837787AFBULL, "vxca_caverns2_7mi_0_0.vlm"},
	{0xDBD109BDBB77300CULL, "252645529.ogg"},
	{0xDBD120350B3907E3ULL, "215524464.ogg"},
	{0xDBD16EB0108453E8ULL, "qun120_arishok6051461.cl"},
	{0xDBD1C88D3189B0A0ULL, "lt_0_lowtown_docks_a3_n_104_0.phy"},
	{0xDBD20F057B4FA4A4ULL, "follower_banter6163153.cl"},
	{0xDBD27865481BC2CCULL, "dae000_hate_follower_react6103196.cl"},
	{0xDBD27AA38B638B06ULL, "6071518_m.fxe"},
	{0xDBD28CFF75C87E9CULL, "vxw_blightlands_tw_0_0.vlm"},
	{0xDBD2CCC837651168ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_clasped_1_gesture_2.ani"},
	{0xDBD36A861CD09A92ULL, "gen00fl_carver6005911.cl"},
	{0xDBD39AAD8069DCFAULL, "6018555_m.fxe"},
	{0xDBD3AF3F880A3E9CULL, "vxw_sundermount_2gj_0_0.vlm"},
	{0xDBD49171208E5260ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v4.gad"},
	{0xDBD53FD0A0C92D79ULL, "313855008.ogg"},
	{0xDBD55C333DFD204FULL, "6098785_m.fxe"},
	{0xDBD570421CDA745CULL, "vxht_mansion2b_1hl_0_0.vlm"},
	{0xDBD60BC5A58963B3ULL, "617081244.ogg"},
	{0xDBD60D58F8241696ULL, "llac_lt_0_lowtown_docks_-65538_2.dds"},
	{0xDBD66DC8E97E253BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windown.phy"},
	{0xDBD67AA1E388C910ULL, "vxca_caverns1_mines_hvr_0_0.vlm"},
	{0xDBD67E94EEC1490EULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.mmh"},
	{0xDBD6D31BD5BD6B5BULL, "332324808.ogg"},
	{0xDBD6EDE6A6721A30ULL, "325143796.ogg"},
	{0xDBD6F2113E8B467CULL, "dae100cr_qunari_warrior3.utc"},
	{0xDBD6F514EAA2B70CULL, "vxht_playershouse_n_1vh_0_0.vlm"},
	{0xDBD7C665102E3E82ULL, "mer320_hubert6089959.cl"},
	{0xDBD7CEBC88E355ABULL, "dae000_keep_maid_sta.bnk"},
	{0xDBD86C78F34DC446ULL, "vxw_encounter2_ma_0_0.vlm"},
	{0xDBD8D08E446F3416ULL, "hlac_lt_backalley_n_65536_0x.dds"},
	{0xDBD95FC0339EF4FAULL, "6052845_m.fxe"},
	{0xDBD9BA7CB9DA09EDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0pp_thinrock.phy"},
	{0xDBDAD0BC88E58E8BULL, "dae000_keep_maid_sta.cnv"},
	{0xDBDC87EAE7D57431ULL, "zz_ave_debug6207719.cl"},
	{0xDBDCD4C9D30A1FCEULL, "hlac_ca_caverns1_24424_21z.dds"},
	{0xDBDCE5B6B4CD6BF5ULL, "w_encounter2_12_0_lc.mmh"},
	{0xDBDCFFE40D4AC7A9ULL, "nat380_nathaniel6071039.cl"},
	{0xDBDD548D74ECFE2EULL, "one100dg_vanard6048236.cl"},
	{0xDBDDDD122DD143FBULL, "6145164_m.fxe"},
	{0xDBDDDD416E95FD0EULL, "50064795.ogg"},
	{0xDBDDE3F724E60309ULL, "ave310_brennan6137216.cl"},
	{0xDBDE5783CC3C6EF4ULL, "vxlt_shop_cu_0_0.vlm"},
	{0xDBDE9D93F7603B74ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0pp13_keepthin_1.mao"},
	{0xDBDED726C86289F0ULL, "fake_warrior_female.utc"},
	{0xDBDF5C62E332EFCBULL, "vxw_blightlands_79_0_0.vlm"},
	{0xDBDF5C89B17A87F0ULL, "vxht_0_hightown_n_bfz_0_0.vlm"},
	{0xDBDFBCA3EFD7F815ULL, "vxlt_warehouse_pn_0_0.vlm"},
	{0xDBDFD3D0FDA949BCULL, "6161099_m.fxe"},
	{0xDBE02C0E8AD54C14ULL, "w_sundermount_101_0.phy"},
	{0xDBE0A650E5CC18BBULL, "vxlt_darktown_2wz_0_0.vlm"},
	{0xDBE0DCBF4718B271ULL, "6103599_m.fxe"},
	{0xDBE2211C7A00571CULL, "vxw_sundermount_2hk_0_0.vlm"},
	{0xDBE27BBD09D0B122ULL, "cod_ltr_emeric_note.utp"},
	{0xDBE29BFA28D759A1ULL, "162934813.ogg"},
	{0xDBE2A82F5BA9DF02ULL, "follower_banter6156098.cl"},
	{0xDBE36962A17777AEULL, "t3_brt_b02.tnt"},
	{0xDBE3A9D957524BCEULL, "qun221_viscount_missing_qun.stg"},
	{0xDBE3C2D9D3D80FD0ULL, "vxht_0_hightown_a3_bfx_0_0.vlm"},
	{0xDBE3E21F0ED08CDCULL, "vxht_mansion2b_1cm_0_0.vlm"},
	{0xDBE435F06BB564BEULL, "605093063.ogg"},
	{0xDBE49D94EECDBB0DULL, "art\\vfx\\combat\\vfx_gui_pp_env_p.anb"},
	{0xDBE570315077965CULL, "6165929_m.fxe"},
	{0xDBE5F761EA8DE0A5ULL, "6208303_f.fxe"},
	{0xDBE62059CF4099F1ULL, "vxw_bonepitmining_a2_1x9_0_1.vlm"},
	{0xDBE657DAFF649B16ULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\open.ani"},
	{0xDBE662170BF9E4B0ULL, "follower_banter6141456.cl"},
	{0xDBE66A1C90D33F54ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_003a_0t.dds"},
	{0xDBE66CAA64E1CA61ULL, "67911168.ogg"},
	{0xDBE677D9377E2BCCULL, "zz_otl_debug6060664.cl"},
	{0xDBE67B2F28088A85ULL, "art\\characters\\creatures\\rubblegolem\\animation\\blendtree\\rubblegolem_s3.anb"},
	{0xDBE68B1913D3219AULL, "w_encounter1_61_0.msh"},
	{0xDBE6991913D339D3ULL, "w_encounter1_61_0.mao"},
	{0xDBE69D1913D340E0ULL, "w_encounter1_61_0.mmh"},
	{0xDBE6DE55E543DCC6ULL, "vxw_encounter2_hb_0_0.vlm"},
	{0xDBE7230E8ADB5BBCULL, "w_sundermount_101_0.rml"},
	{0xDBE754F7ADBA5C18ULL, "zz_and_debug6134143.cl"},
	{0xDBE8F0C01149329CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww_galow2win.phy"},
	{0xDBE92A7930032CEBULL, "9269948.ogg"},
	{0xDBE943947FDB041EULL, "6070364_m.fxe"},
	{0xDBE98264BC1C7023ULL, "llac_w_sundermount_393214_0.dds"},
	{0xDBE989FEA57EC4A8ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_008a_d.dds"},
	{0xDBE9B07A304E8C3EULL, "vxw_encounter2_camp_ow_0_0.vlm"},
	{0xDBE9B2BE268A2C95ULL, "478044116.ogg"},
	{0xDBE9B73E225547DFULL, "61803177.ogg"},
	{0xDBEA090E8ADD6511ULL, "w_sundermount_101_0.msh"},
	{0xDBEA170E8ADD7D2CULL, "w_sundermount_101_0.mao"},
	{0xDBEA230E8ADD91CFULL, "w_sundermount_101_0.mmh"},
	{0xDBEA32CEA41043E5ULL, "vxlt_0_lowtown_docks_n_apg_0_0.vlm"},
	{0xDBEA4244C2B3CD68ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf03.msh"},
	{0xDBEA6044C2B3FF92ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf03.mmh"},
	{0xDBEAB5D5EE2E4536ULL, "746098420.ogg"},
	{0xDBEB242D7950B87FULL, "6126714_m.fxe"},
	{0xDBEB4763A99C54F8ULL, "260393306.ogg"},
	{0xDBEC313A049C3601ULL, "dae000ip_crowd_ff3_talk14.utp"},
	{0xDBEC35AA89AADC88ULL, "follower_banter6169847.cl"},
	{0xDBEC35FC486292CCULL, "473108647.ogg"},
	{0xDBEC5400C3B1D65BULL, "10268385.ogg"},
	{0xDBEC939A789035FBULL, "829411868.ogg"},
	{0xDBECC960BE328774ULL, "vxlt_shop_fz_0_0.vlm"},
	{0xDBED14E4B505B687ULL, "6105986_m.fxe"},
	{0xDBED75AFB8E11B0EULL, "vxht_0_hightown_a3_c8a_0_0.vlm"},
	{0xDBEDB4CCCDBF03A9ULL, "gen00fl_fenris6126387.cl"},
	{0xDBEDCE3FD529084BULL, "vxw_blightlands_88_0_0.vlm"},
	{0xDBEE2E80E1CE1095ULL, "vxlt_warehouse_um_0_0.vlm"},
	{0xDBEE45ADEF9F623CULL, "6166199_m.fxe"},
	{0xDBEE5DEB182231BCULL, "dre000_dougal6130976.cl"},
	{0xDBEEAFD6C7721817ULL, "6127630_m.fxe"},
	{0xDBEEE9F13DB497CCULL, "6151272_m.fxe"},
	{0xDBEEEF7EC039E1FFULL, "dae300_cullen6140638.cl"},
	{0xDBEF4499F5161C26ULL, "376425124.ogg"},
	{0xDBEF67440E16F63DULL, "6140980_f.fxe"},
	{0xDBEFA7AA949F5528ULL, "vxht_0_hightown_5gy_0_0.vlm"},
	{0xDBF092F96BF66F9CULL, "vxw_sundermount_2mh_0_0.vlm"},
	{0xDBF0B39BB57FBFD3ULL, "vxht_mansion2b_3c8_0_0.vlm"},
	{0xDBF0CDE439D5498CULL, "91657385.ogg"},
	{0xDBF1CD69FE1CAB7FULL, "6118113_m.fxe"},
	{0xDBF2727F1DE4E20EULL, "zz_dae_debug6143038.cl"},
	{0xDBF2AE21D28A594DULL, "vxht_brothel_n_509_0_0.vlm"},
	{0xDBF2DC31CFE9F44FULL, "372634604.ogg"},
	{0xDBF377C2A1196135ULL, "hlac_ca_cavern1_haunted_24425_2x.dds"},
	{0xDBF3DD3B663BAEF8ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrb_0t.dds"},
	{0xDBF48684DF6F07B4ULL, "78327100.ogg"},
	{0xDBF4D06F3D6C9F8AULL, "31814610.ogg"},
	{0xDBF5827E692F10BDULL, "ser221ar_murderer_estate_nt.lst"},
	{0xDBF5B236E2FA3CA5ULL, "vxlt_gallowscourtyard_n_5gr_0_0.vlm"},
	{0xDBF5DB97CAA59E80ULL, "838584756.ogg"},
	{0xDBF6594E85B2514DULL, "mcr371_orsino6083305.cl"},
	{0xDBF6E58D31A8FE35ULL, "lt_0_lowtown_docks_a3_n_104_0.msh"},
	{0xDBF6E78D31A901ABULL, "lt_0_lowtown_docks_a3_n_104_0.mmh"},
	{0xDBF6F38D31A91648ULL, "lt_0_lowtown_docks_a3_n_104_0.mao"},
	{0xDBF70FB61ECB025CULL, "928604074.ogg"},
	{0xDBF79F61C7BBCD7BULL, "vxlt_0_lowtown_h28_0_0.vlm"},
	{0xDBF7A9B36E335FDCULL, "hlac_lt_0_lowtown_docks_-131075_1y.dds"},
	{0xDBF822572244A4BEULL, "vxw_encounter2_camp_jh_0_0.vlm"},
	{0xDBF827F339EB845AULL, "snow_t_01.mao"},
	{0xDBF85F597ECCBC90ULL, "hlac_w_sundermount_-5_0y.dds"},
	{0xDBF874D488DE4AA1ULL, "vxlt_0_lowtown_a2_n_ep2_0_0.vlm"},
	{0xDBF884EE6BEB720AULL, "6175957_m.fxe"},
	{0xDBF88618DBC30F64ULL, "character_character_p184.psh.fxo.dx11"},
	{0xDBF8D0837691A82EULL, "459823906.ogg"},
	{0xDBF8E9EC7E42F0BCULL, "vxca_cavern1_haunted_j2a_0_1.vlm"},
	{0xDBF91553DAACED09ULL, "llac_lt_0_lowtown_a2_-65536_2.dds"},
	{0xDBF964B3620AA602ULL, "vxlt_0_lowtown_docks_a3_d39_0_1.vlm"},
	{0xDBF9B7458E959FC2ULL, "gen00fl_varric6107291.cl"},
	{0xDBF9F0ECDCCDED1CULL, "vxlt_0_lowtown_a2_n_cw0_0_0.vlm"},
	{0xDBFB53CA091B7E51ULL, "hlac_lt_0_lowtown_docks_-65537_0z.dds"},
	{0xDBFB80DAFF773D50ULL, "art\\levels\\races\\proxy\\props\\armoire\\animation\\open.gad"},
	{0xDBFC112FE6A2BDDBULL, "vxw_bonepit_13u_0_0.vlm"},
	{0xDBFC7A7E693521ACULL, "ser221ar_murderer_estate_nt.ncs"},
	{0xDBFCB0C5EF09BF9CULL, "lowtown_male_cityguard_mob.cnv"},
	{0xDBFCD1AA8F5CE666ULL, "hf_bethany_h1s2.mor"},
	{0xDBFD25062ECF43D8ULL, "vxw_bonepitmining_a2_21m_0_0.vlm"},
	{0xDBFD4AD4EAA77471ULL, "vxlt_foundry_n_yv_0_0.vlm"},
	{0xDBFD77C71112AF67ULL, "6041202_m.fxe"},
	{0xDBFD7D649626DBBBULL, "follower_banter6169297.cl"},
	{0xDBFDDF71C9B0919AULL, "184744716.ogg"},
	{0xDBFE1F0C43AE9388ULL, "82940911.ogg"},
	{0xDBFE5A582CE85533ULL, "hlac_w_sundermount_262144_0y.dds"},
	{0xDBFF04D65DEC881CULL, "vxw_sundermount_2vi_0_0.vlm"},
	{0xDBFF73DF8DC1C9E8ULL, "hlac_ht_bartrand_n_3373_1x.dds"},
	{0xDBFFAFC480BD0434ULL, "vxlt_0_lowtown_docks_n_bt5_0_0.vlm"},
	{0xDBFFD2C5EF0C2FECULL, "lowtown_male_cityguard_mob.bnk"},
	{0xDBFFE35FC6279670ULL, "6090368_m.fxe"},
	{0xDC004BFCE893ECE3ULL, "dae000_epic_merchant_46179521.cl"},
	{0xDC006C696EF1F93DULL, "vxht_mansion2b_28b_0_0.vlm"},
	{0xDC01398DC6991BF0ULL, "vxht_0_hightown_n_9b9_0_0.vlm"},
	{0xDC014677040BDE81ULL, "gen00fl_varric6076424.cl"},
	{0xDC021158340C15F6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_1.evt"},
	{0xDC0254C8E9A335DAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windown.msh"},
	{0xDC0262C8E9A34D13ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windown.mao"},
	{0xDC0266C8E9A35420ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_windown.mmh"},
	{0xDC02A1DFC8E581EAULL, "dae000ip_crowd_ff2_talk18.utp"},
	{0xDC036B50D58906A9ULL, "hlac_lt_0_lowtown_a2_131072_1y.dds"},
	{0xDC038A56A01C27C8ULL, "dre000_sibling_has_blight6037035.cl"},
	{0xDC0474C913411C60ULL, "hm_clf_evets.mor"},
	{0xDC051BC46CE88871ULL, "302112732.ogg"},
	{0xDC059F5996D76D3BULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgtb.phy"},
	{0xDC062001981B7D27ULL, "6144057_m.fxe"},
	{0xDC069434143ABD3EULL, "vxw_encounter2_camp_ei_0_0.vlm"},
	{0xDC06A9A6712BEF78ULL, "mcr310_argument6075132.cl"},
	{0xDC06C0996258BF32ULL, "359845310.ogg"},
	{0xDC06E6B17AD46321ULL, "vxlt_0_lowtown_a2_n_eu1_0_0.vlm"},
	{0xDC07AD8C2F19DE70ULL, "6126803_m.fxe"},
	{0xDC07D0B761174AF3ULL, "19057054.ogg"},
	{0xDC08085834107181ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_1.gad"},
	{0xDC084D77EA565304ULL, "hlac_ht_0_hightown_n_65536_4y.dds"},
	{0xDC08916CF4C982D9ULL, "vxw_sundermount_2u9_0_0.vlm"},
	{0xDC097723224CFCD3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_basic.phy"},
	{0xDC098EA62AD2D38AULL, "w_sundermount_161_0.mao"},
	{0xDC0992A62AD2DAD9ULL, "w_sundermount_161_0.mmh"},
	{0xDC099CA62AD2EBC7ULL, "w_sundermount_161_0.msh"},
	{0xDC09DF617538ACA7ULL, "ht_0_hightown_a2_25_0.anb"},
	{0xDC0A2D19815011DBULL, "vxlt_0_lowtown_docks_a3_cx7_0_0.vlm"},
	{0xDC0A551C7427D6E2ULL, "dr_primevaltomb_mmap_ph.dds"},
	{0xDC0AB3BD2B52058DULL, "hlac_ca_caverns2_5802_26z.dds"},
	{0xDC0B0AD8DA9CFA0FULL, "vxlt_gallowstemplar_4vg_0_0.vlm"},
	{0xDC0B25C3FF1AD3F2ULL, "6041430_m.fxe"},
	{0xDC0B6875FC543A33ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_007a_0d.dds"},
	{0xDC0B73A74D1FCE01ULL, "fen210_hadriana6110549.cl"},
	{0xDC0BCCFED8052894ULL, "art\\vfx\\combat\\vfx_glb_ele_s_extreme_c.mmh"},
	{0xDC0BD215DC4E417BULL, "vxca_cavern1_haunted_ei5_0_0.vlm"},
	{0xDC0C037BD446E7B6ULL, "stop_combat.ncs"},
	{0xDC0C15B022CD33B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_reaction.ani"},
	{0xDC0C686BFF2EEC84ULL, "gen00fl_anders6119177.cl"},
	{0xDC0C7EF2D0546FBEULL, "929289971.ogg"},
	{0xDC0CD7B9C288202EULL, "202652832.ogg"},
	{0xDC0D1F06D83A2D1DULL, "llac_lt_0_lowtown_a2_n_c_-65536_5.dds"},
	{0xDC0E05D87E694020ULL, "vxlt_0_lowtown_a2_n_c_byu_0_0.vlm"},
	{0xDC0E59315AA941DEULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.dura.ani"},
	{0xDC0ECA88CE1330F4ULL, "vxlt_gallowscourtyard_76q_0_0.vlm"},
	{0xDC0EDE4660E811BDULL, "vxht_mansion2b_25a_0_0.vlm"},
	{0xDC0EDFEA11B9BB4DULL, "vxht_0_hightown_a2_c50_0_0.vlm"},
	{0xDC0F09583416922FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\drunk_gesture_right_1.ani"},
	{0xDC0F5C6279E2FB51ULL, "6206085_m.fxe"},
	{0xDC0F69EC754E7E49ULL, "6166972_m.fxe"},
	{0xDC0FC061A65805C1ULL, "292443748.ogg"},
	{0xDC0FD952AE768962ULL, "llac_lt_warehouse_65536_0.dds"},
	{0xDC10C4F564148EC1ULL, "597591283.ogg"},
	{0xDC10DC3FE29DC023ULL, "hlac_lt_0_lowtown_docks_n_-131073_1y.dds"},
	{0xDC115221C8F8B3AAULL, "lla_ht_0_hightown_n_-131070_0.dds"},
	{0xDC115E7936B5150EULL, "399962659.ogg"},
	{0xDC120961134780FCULL, "ss_fenris1.cut"},
	{0xDC12413787E3DD25ULL, "431174346.ogg"},
	{0xDC12EA602823E487ULL, "hlac_w_encounter3_-65539_0y.dds"},
	{0xDC13240E8B00980CULL, "w_sundermount_101_0.anb"},
	{0xDC134E76A429337FULL, "6075282_f.fxe"},
	{0xDC13606B5A825BF5ULL, "plt_lgt000pt_ultra_5_06.nss"},
	{0xDC139D4099501EFFULL, "vxw_bonepitmining_a2_4y_0_0.vlm"},
	{0xDC13D7B73E402FD4ULL, "6070436_m.fxe"},
	{0xDC144ABA055167C4ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.impa.ani"},
	{0xDC1466948433F5A7ULL, "6093638_m.fxe"},
	{0xDC14BC2E800254F3ULL, "vxw_bonepitdestroyed_a3_26e_0_0.vlm"},
	{0xDC14DAAD87AD7B80ULL, "20197191.ogg"},
	{0xDC151CF62C87EFB3ULL, "follower_banter6136356.cl"},
	{0xDC15311584D2AB74ULL, "69682056.ogg"},
	{0xDC1549A7C34854C6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_hunch_en.evt"},
	{0xDC15D7BDF5ECE95BULL, "vxw_woundedcoast_1a_0_0.vlm"},
	{0xDC15E15D54FAAA55ULL, "isa211_isabela6070872.cl"},
	{0xDC15F3EEF9676692ULL, "primitivetext.mao"},
	{0xDC162F061BEDD5CEULL, "vxw_bonepitmining_a2_5w_0_0.vlm"},
	{0xDC1652D2EFB09BDDULL, "162989970.ogg"},
	{0xDC168C7ECFD8AA30ULL, "6038612_m.fxe"},
	{0xDC16B242A1C215FFULL, "hlac_lt_warehouse_n_-65535_0z.dds"},
	{0xDC16D70AACD24661ULL, "vxdr_0_deeproads_f0m_0_0.vlm"},
	{0xDC16E6456224FFBBULL, "w_bonepit_133_0.mmh"},
	{0xDC16F24562251458ULL, "w_bonepit_133_0.mao"},
	{0xDC170445622532A5ULL, "w_bonepit_133_0.msh"},
	{0xDC1732BF965049B1ULL, "vxht_mansion2_22h_0_0.vlm"},
	{0xDC17F49BC0B81FC2ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyn_0t.dds"},
	{0xDC18087A34247BB5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candlecluster_sa.dds"},
	{0xDC184E4CFC62F6F3ULL, "hlac_ca_cavern1_haunted_24423_5y.dds"},
	{0xDC18C0818766C872ULL, "vxlt_0_lowtown_n_eqv_0_0.vlm"},
	{0xDC1A1DB022D97EA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_reaction.evt"},
	{0xDC1A4D5F5CEAB5C2ULL, "6144873_f.fxe"},
	{0xDC1A849D2EF37009ULL, "6112870_m.fxe"},
	{0xDC1ABE764E7550F3ULL, "vxlt_0_lowtown_a2_f5g_0_0.vlm"},
	{0xDC1AF1D2DF2E758FULL, "1061812349.ogg"},
	{0xDC1B211D3337E74BULL, "6147571_m.fxe"},
	{0xDC1B41DBF6C4E92AULL, "703552796.ogg"},
	{0xDC1B47BA05578126ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.impa.gad"},
	{0xDC1B71315AB3F4AFULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.dura.evt"},
	{0xDC1C60A7C34E9A11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_hunch_en.gad"},
	{0xDC1C8F9CC3D38A54ULL, "lt_0_lowtown_a2_21_0_lc.mmh"},
	{0xDC1C9B78800CD99EULL, "dae321st_velasco.stg"},
	{0xDC1CB44B4C29702DULL, "vxht_playershouse_n_1o2_0_0.vlm"},
	{0xDC1CDD4EFCBCE051ULL, "421544799.ogg"},
	{0xDC1D154FBC1FC25BULL, "6033988_m.fxe"},
	{0xDC1D3BB598A64B2BULL, "505308943.ogg"},
	{0xDC1D69C1B0368A4CULL, "dre000_bodahn6028165.cl"},
	{0xDC1D8F36BBA2D9E6ULL, "6110453_m.fxe"},
	{0xDC1DA992D6A8EEDCULL, "vxht_mansion2b_1wa_0_0.vlm"},
	{0xDC1DC5CC319BA993ULL, "lt_gallowstemplar_n_2387.rml"},
	{0xDC1E9F3CCF7B08ABULL, "vxlt_0_lowtown_a3_n_c_f4p_0_0.vlm"},
	{0xDC1EF7B68C7B1729ULL, "ave310_party_comment6134848.cl"},
	{0xDC1F204E49D47FF2ULL, "1010808480.ogg"},
	{0xDC1F29332E79C66FULL, "vxlt_0_lowtown_fg5_0_0.vlm"},
	{0xDC1F6A1DCD668315ULL, "vxlt_0_lowtown_eoy_0_1.vlm"},
	{0xDC1FCB369F2E6120ULL, "425888667.ogg"},
	{0xDC1FE962B26B3E81ULL, "llac_lt_0_lowtown_docks_n_-65538_4.dds"},
	{0xDC203EB022DE22F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\velasco_prostitute_reaction.gad"},
	{0xDC20591EA8BA4E97ULL, "585863800.ogg"},
	{0xDC205FC06E26F248ULL, "vxdr_0_deeproads_f0h_0_0.vlm"},
	{0xDC20B64B1BCBFDA9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock4a.phy"},
	{0xDC214A0EFAE76C89ULL, "6146787_m.fxe"},
	{0xDC215091A16D545CULL, "6175757_m.fxe"},
	{0xDC2163F8D02EDA50ULL, "gen_im_trs_dal_junk2_02.uti"},
	{0xDC216A1066669DBAULL, "dre000_dougal6131119.cl"},
	{0xDC218B7E69545A9AULL, "ser221ar_murderer_estate_nt.are"},
	{0xDC2242BA055D9705ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.impa.evt"},
	{0xDC2272315ABA1558ULL, "art\\vfx\\combat\\vfx_glb_sleep_crust_c.dura.gad"},
	{0xDC232E0B71F86D73ULL, "vxw_bonepitdestroyed_a3_21d_0_0.vlm"},
	{0xDC2361A7C354BABFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_hunch_en.ani"},
	{0xDC23ED72CF178DF5ULL, "180324775.ogg"},
	{0xDC247CF79AD154B8ULL, "vxca_caverns2_7z7_0_0.vlm"},
	{0xDC24FA8D40CB75D2ULL, "6105789_m.fxe"},
	{0xDC25636667243175ULL, "vxw_encounter3_g2_0_1.vlm"},
	{0xDC25F1904F22E6BBULL, "gen00fl_aveline6139083.cl"},
	{0xDC266BDCA8CF0925ULL, "vxht_chantry_28j_0_0.vlm"},
	{0xDC26827A9705661FULL, "lt_0_lowtown_n_28_0_ll.mmh"},
	{0xDC2698878A541942ULL, "lt_0_lowtown_docks_25_0_ll.mmh"},
	{0xDC26DA16F0BD66D9ULL, "gen_im_trs_gem_opal_01.uti"},
	{0xDC2729E4D6E6AFB0ULL, "vxlt_0_lowtown_docks_n_c5g_0_0.vlm"},
	{0xDC275135446983C2ULL, "vxht_0_hightown_a3_65g_0_0.vlm"},
	{0xDC278344C3D7F5C7ULL, "42035689.ogg"},
	{0xDC27920696E17557ULL, "223612005.ogg"},
	{0xDC27B676CA5495EBULL, "vxlt_gallowstemplar_f_6qc_0_0.vlm"},
	{0xDC27F1128255A00EULL, "873172671.ogg"},
	{0xDC28A87B84458666ULL, "6149521_m.fxe"},
	{0xDC28B2043C3B0440ULL, "6206228_m.fxe"},
	{0xDC28B5CFC03A727BULL, "vxca_cavern1_haunted_ec7_0_0.vlm"},
	{0xDC29A7EF2E39646BULL, "677419140.ogg"},
	{0xDC2A2587598F81C0ULL, "vxlt_gallowstemplar_f_359_0_0.vlm"},
	{0xDC2A5876E6647234ULL, "static_static_p74.psh.fxo.dx11"},
	{0xDC2B9C6A51621A5DULL, "hlac_dr_primevaltomb_659_5x.dds"},
	{0xDC2C15F766C73640ULL, "687127690.ogg"},
	{0xDC2C3796F45F2EC0ULL, "atl_fonttest_dxt1_02.dds"},
	{0xDC2C912B61C3BE65ULL, "hlac_w_bonepitmining_a2_393219_0z.dds"},
	{0xDC2CAD0A37318E93ULL, "mag100_karras6034920.cl"},
	{0xDC2CAEE8B4D15479ULL, "381748969.ogg"},
	{0xDC2CD1479044D9E4ULL, "qcr220_grey_wardens6060445.cl"},
	{0xDC2D93E2619232A7ULL, "hlac_lt_gallowstemplar_n_2394_4z.dds"},
	{0xDC2DBD24E5BC00B6ULL, "pro000_aveline6047232.cl"},
	{0xDC2DFE4F1FC301F1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_enter.ani"},
	{0xDC2F1F42B7AB57C0ULL, "dae011_drunk1_sta6144138.cl"},
	{0xDC2F3DA1F6EDCF6DULL, "6116647_m.fxe"},
	{0xDC3013E819B8D6C6ULL, "6027715_m.fxe"},
	{0xDC302560A6789F11ULL, "dae000cr_grand_cleric6187916.cl"},
	{0xDC302F34AC6EBA6FULL, "vxlt_gallowstemplar_n_3rw_0_0.vlm"},
	{0xDC30785DF4692CD0ULL, "oil_256.dds"},
	{0xDC3109F7B9D7BCAEULL, "332940135.ogg"},
	{0xDC314F671CEA0DF9ULL, "945830608.ogg"},
	{0xDC31844874D38164ULL, "gen00fl_sebastian6188660.cl"},
	{0xDC31865996FC7DDAULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgtb.msh"},
	{0xDC31985996FC9C20ULL, "art\\characters\\playercharacter\\humanmale\\hoo_lgtb.mmh"},
	{0xDC31CEA274DE78DBULL, "487048035.ogg"},
	{0xDC3236B316B5486AULL, "follower_banter6169856.cl"},
	{0xDC325E6DF008F775ULL, "140446028.ogg"},
	{0xDC32EBADA6CF9ADEULL, "330751249.ogg"},
	{0xDC331CA0615338FDULL, "primitives_primitive_p1.psh.fxo"},
	{0xDC33A48B0D19A224ULL, "vxlt_backalley_n_1zj_0_0.vlm"},
	{0xDC33AC6044DA5A09ULL, "6087501_m.fxe"},
	{0xDC33C6DEFC304CADULL, "pro000_soldier_leader6040812.cl"},
	{0xDC3418E79F20E18BULL, "vxlt_0_lowtown_docks_n_ctg_0_0.vlm"},
	{0xDC349DED2B3D13C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_major_v1.evt"},
	{0xDC34CFFED828334BULL, "art\\vfx\\combat\\vfx_glb_ele_s_extreme_c.anb"},
	{0xDC35094F1FC93293ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_enter.gad"},
	{0xDC35584FD34CE858ULL, "59058229.ogg"},
	{0xDC355F6B5A9F4056ULL, "plt_lgt000pt_ultra_5_06.plo"},
	{0xDC35725B6026B459ULL, "6051270_m.fxe"},
	{0xDC359BC1C8DCC830ULL, "vxlt_0_lowtown_docks_n_c0d_0_0.vlm"},
	{0xDC35B990B0C0858DULL, "vxdr_primevalentrance_2ok_0_1.vlm"},
	{0xDC36180F0BBA1A35ULL, "6137552_m.fxe"},
	{0xDC362B594574E458ULL, "hlac_w_encounter2_camp_262143_0x.dds"},
	{0xDC36338DEFCB6C53ULL, "gen_im_trs_rdr_two_2ha_r2.uti"},
	{0xDC364571F3F3753EULL, "hlac_lt_0_lowtown_n_131070_4z.dds"},
	{0xDC36587BF06D73F3ULL, "vxca_cavern1_haunted_k2o_0_1.vlm"},
	{0xDC36F0393BE896E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v3.evt"},
	{0xDC37FF2ED6ABE7FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_lose.gad"},
	{0xDC3849E87B111E6AULL, "vrc221tr_bartrand_stage_1.utt"},
	{0xDC38946D8E6E1BB3ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnl90pause.evt"},
	{0xDC38A563026D3958ULL, "290283895.ogg"},
	{0xDC38C1C42B12B5B4ULL, "613150725.ogg"},
	{0xDC3920CC65BF8B50ULL, "llac_ht_0_hightown_n_-65538_0.dds"},
	{0xDC395D6C81BE6499ULL, "6030445_f.fxe"},
	{0xDC3967585CAA34DDULL, "art\\characters\\weapons\\unique\\arishok\\w_gsw_002a.mao"},
	{0xDC396B585CAA3B16ULL, "art\\characters\\weapons\\unique\\arishok\\w_gsw_002a.mmh"},
	{0xDC3975585CAA4C14ULL, "art\\characters\\weapons\\unique\\arishok\\w_gsw_002a.msh"},
	{0xDC39930D67DCB215ULL, "vxlt_blackemporium_13n_0_0.vlm"},
	{0xDC39C8DC0B8571C5ULL, "271704352.ogg"},
	{0xDC3AAA6EDE6363A7ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_bdy_hlk_0n.dds"},
	{0xDC3AD32E416A8BC9ULL, "vxlt_0_lowtown_docks_n_b93_0_0.vlm"},
	{0xDC3AFB669486BEFCULL, "zz_followers_banter6157677.cl"},
	{0xDC3B6B8B6ED1EF0EULL, "vxdr_0_deeproads_dvz_0_0.vlm"},
	{0xDC3B76ED2B42F044ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_major_v1.gad"},
	{0xDC3BCA5D28A10A9BULL, "6116339_m.fxe"},
	{0xDC3C064F1FCF4CF8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_kd_enter.evt"},
	{0xDC3C0D797EAD8950ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em04.mmh"},
	{0xDC3C1B797EADA16AULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em04.msh"},
	{0xDC3C2050AAACC5E9ULL, "vxht_mansion2_1vc_0_0.vlm"},
	{0xDC3C6295A61D7E40ULL, "vxht_0_hightown_n_9za_0_0.vlm"},
	{0xDC3CBE5D2BA5498FULL, "seal_pls_ex.mao"},
	{0xDC3D5397B0E7D10EULL, "llac_lt_0_lowtown_a2_n_-65536_0.dds"},
	{0xDC3E022ED6B05811ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_lose.evt"},
	{0xDC3E07393BEEDC31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v3.gad"},
	{0xDC3E09BE2E862D5EULL, "337872994.ogg"},
	{0xDC3F3D7C18988CF5ULL, "6111173_f.fxe"},
	{0xDC3F497BEF97AA8DULL, "6049675_m.fxe"},
	{0xDC3F529540DF2838ULL, "vxlt_0_lowtown_docks_a3_czz_0_0.vlm"},
	{0xDC3F7F968D98FA86ULL, "vxlt_0_lowtown_docks_a3_cz4_0_0.vlm"},
	{0xDC4013A27E22F001ULL, "vxht_mansion2b_1w0_0_0.vlm"},
	{0xDC406B179492DA1DULL, "vxdr_primevaltomb_3a1_0_0.vlm"},
	{0xDC40F799A89FBEC4ULL, "character_character_p140.psh.fxo.dx11"},
	{0xDC413F6B203369DCULL, "52405365.ogg"},
	{0xDC4185ED2B477522ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_poison_major_v1.ani"},
	{0xDC428AC49116FA0BULL, "vxlt_0_lowtown_docks_n_cof_0_0.vlm"},
	{0xDC42A4F749670442ULL, "vxht_mansion2_1am_0_0.vlm"},
	{0xDC42C745101DD2FDULL, "vxlt_0_lowtown_a3_n_c_eot_0_0.vlm"},
	{0xDC430A5D5F69628AULL, "6137602_f.fxe"},
	{0xDC433CE3F181423AULL, "vxlt_foundry_n_wk_0_0.vlm"},
	{0xDC43BC6B6A59FFEFULL, "water_water_p0.vsh.fxo"},
	{0xDC43EAD0D18135D3ULL, "6076334_m.fxe"},
	{0xDC44FE652CC87412ULL, "6093183_m.fxe"},
	{0xDC4508393BF4FCDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_die_gup_v3.ani"},
	{0xDC45C17AA13B58AFULL, "dae000_bartrand6205980.cl"},
	{0xDC46391465A4B982ULL, "plt_seb120pt_duty_docks.plo"},
	{0xDC465F44C3025C8DULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hf03.phy"},
	{0xDC4689990B5417EBULL, "6178524_m.fxe"},
	{0xDC47320681108E45ULL, "6187771_m.fxe"},
	{0xDC4737AEECBDB543ULL, "follower_banter6149196.cl"},
	{0xDC479E8007A6472AULL, "vxw_encounter2_bl_0_0.vlm"},
	{0xDC47B052237866DDULL, "hlac_dr_dungeon_3214_7z.dds"},
	{0xDC4826E522C7B796ULL, "mag160_varian_ilithis6018901.cl"},
	{0xDC48B24DABD511F6ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar2_scardemage.ani"},
	{0xDC49450B3360A449ULL, "vxlt_0_lowtown_docks_n_b62_0_0.vlm"},
	{0xDC4A1A80FE6335FCULL, "ach_oth_mages.dds"},
	{0xDC4A2BBA592F3C6DULL, "6126666_m.fxe"},
	{0xDC4A497782B5E6E4ULL, "mer120_hubert6020653.cl"},
	{0xDC4A6FF9036C6B48ULL, "266287122.ogg"},
	{0xDC4A703717C8B028ULL, "6099235_m.fxe"},
	{0xDC4B3C4A8E1EE02AULL, "982555985.ogg"},
	{0xDC4B7DE8FEE115E9ULL, "bw_oil_loop.mao"},
	{0xDC4C0A9DADE25BA0ULL, "zz_personality_debug6064399.cl"},
	{0xDC4C1A2ED6BCBEE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_lose.ani"},
	{0xDC4C9F4B1BF111CEULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock4a.mmh"},
	{0xDC4CAB4B1BF12665ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock4a.mao"},
	{0xDC4CB94B1BF13DBCULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_rock4a.msh"},
	{0xDC4D10B0DA993AA5ULL, "ht_0_hightownexplosion_n_45_0.rml"},
	{0xDC4D1F93907A8BF2ULL, "6050093_f.fxe"},
	{0xDC4D276DDBA6A137ULL, "vxw_bonepitmining_a2_h5_0_0.vlm"},
	{0xDC4D3AA7CA69139EULL, "96143072.ogg"},
	{0xDC4DD23F87B1815EULL, "qun160_vashothlead6053852.cl"},
	{0xDC4DE4EF1010D1DAULL, "cd_gen00fl_isabela.bnk"},
	{0xDC4DEC949BD6E975ULL, "vxht_0_hightown_2xr_0_0.vlm"},
	{0xDC4DF33D006F17E0ULL, "6148087_m.fxe"},
	{0xDC4E08BE71199691ULL, "309836363.ogg"},
	{0xDC4E3F80DE4B0579ULL, "106139203.ogg"},
	{0xDC4E86ABEDF91E05ULL, "943169074.ogg"},
	{0xDC4F5DD4A8449D6AULL, "vxlt_gallowscourtyard_n_746_0_0.vlm"},
	{0xDC4F78200ACC0CFEULL, "610707964.ogg"},
	{0xDC5024E87B257153ULL, "vrc221tr_bartrand_stage_1.ncs"},
	{0xDC50638343B6F46BULL, "33515789.ogg"},
	{0xDC508CF14089D32FULL, "6170803_m.fxe"},
	{0xDC50BDBA7EF18B86ULL, "384036426.ogg"},
	{0xDC50F048A18060F0ULL, "736232546.ogg"},
	{0xDC50FBEC18283D1FULL, "hlac_dr_dungeon_3215_9y.dds"},
	{0xDC5172CA6D047927ULL, "vxw_woundedcoast_1fq_0_0.vlm"},
	{0xDC51836019D81659ULL, "da2_prc_seb_ru-ru.cre"},
	{0xDC522148BFE4CD89ULL, "hlac_dr_dungeon_3217_4z.dds"},
	{0xDC5224A95E6E56CDULL, "vxw_bonepit_1ue_0_0.vlm"},
	{0xDC52D0B47E604AB9ULL, "491021604.ogg"},
	{0xDC5326463F0779CCULL, "6129557_f.fxe"},
	{0xDC53938F21841E64ULL, "701685308.ogg"},
	{0xDC53CF69BF480CD7ULL, "hlac_w_woundedcoast_-262144_0z.dds"},
	{0xDC53E2A2E34635D3ULL, "mrl000_marethari6068857.cl"},
	{0xDC54008D0FFD0BA2ULL, "vxlt_undercitydungeon_1it_0_1.vlm"},
	{0xDC54313BA6834EF3ULL, "gen00fl_aveline6139847.cl"},
	{0xDC5440AAF8027647ULL, "532650730.ogg"},
	{0xDC5449D047A21B0FULL, "53497104.ogg"},
	{0xDC546A981FF17952ULL, "330425306.ogg"},
	{0xDC54C19B7431104AULL, "vxht_chantry_30p_0_0.vlm"},
	{0xDC54D8434F31A6D5ULL, "vxca_caverns2_807_0_0.vlm"},
	{0xDC54F9C22EB22FDBULL, "6137187_m.fxe"},
	{0xDC553DBD3B639716ULL, "vxht_0_hightown_a2_91m_0_0.vlm"},
	{0xDC55746B3F656102ULL, "static_static_p10.psh.fxo"},
	{0xDC55AB786C0004A0ULL, "w_bonepit_124_0.phy"},
	{0xDC55C45BEDCCD7D3ULL, "180027539.ogg"},
	{0xDC55E83B8F75712EULL, "vxca_caverns2_9fk_0_0.vlm"},
	{0xDC56098B22265CA6ULL, "abi_r_rev_attack.bnk"},
	{0xDC56105CF99C5FAAULL, "vxw_encounter2_mm_0_0.vlm"},
	{0xDC5626DCE0D50086ULL, "wm_ico_doorway.dds"},
	{0xDC56785461BEEA69ULL, "6167675_f.fxe"},
	{0xDC56B95B49C1F8F3ULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_1z.dds"},
	{0xDC56D33E6BC36E4CULL, "6143095_m.fxe"},
	{0xDC572207774CFB91ULL, "vxdr_primevalentrance_2nh_0_1.vlm"},
	{0xDC575ECD37770221ULL, "754173664.ogg"},
	{0xDC585F4E4FB91E71ULL, "6119308_f.fxe"},
	{0xDC58C73AAD4070C8ULL, "ave331cr_insurgent_m3.utc"},
	{0xDC594EB32F17CAC5ULL, "vxht_bartrand_n_6rt_0_0.vlm"},
	{0xDC595AF81F5DAD39ULL, "llac_lt_0_lowtown_docks_a3_n_131073_0.dds"},
	{0xDC597673B25C18F6ULL, "vxdr_0_deeproads_cvq_0_0.vlm"},
	{0xDC59B7A034EEFCB5ULL, "lt_0_lowtown_docks_a3_183_0.msh"},
	{0xDC59B9A034EF002BULL, "lt_0_lowtown_docks_a3_183_0.mmh"},
	{0xDC59C5A034EF14C8ULL, "lt_0_lowtown_docks_a3_183_0.mao"},
	{0xDC59CD843636042EULL, "and230_mage_girl6135816.cl"},
	{0xDC59D596079700F2ULL, "vxw_bonepitmining_a2_p4_0_0.vlm"},
	{0xDC59D79E949FA65AULL, "lowtown_male_alienage2_mob6166557.cl"},
	{0xDC5A3D8392CD9C19ULL, "lt_gallowstemplar_2384.rml"},
	{0xDC5A89593801A77FULL, "50099884.ogg"},
	{0xDC5B0ACBF7801BB6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_buff.ani"},
	{0xDC5B4530F6857386ULL, "6131005_m.fxe"},
	{0xDC5B53B4EE18FC0AULL, "vxca_caverns1_mines_hum_0_0.vlm"},
	{0xDC5B8EBD6DDBB557ULL, "mcr371_alain6087128.cl"},
	{0xDC5B92298B0FDD72ULL, "6174019_m.fxe"},
	{0xDC5BCDE4E97746F5ULL, "hlac_ht_0_hightown_a3_-65538_1x.dds"},
	{0xDC5CB5F17D0A697AULL, "dae000_mother6121248.cl"},
	{0xDC5CDB4DABE60030ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar2_scardemage.gad"},
	{0xDC5CE3B83701D2C2ULL, "hlac_w_encounter1_-131068_0y.dds"},
	{0xDC5DA23884FB3562ULL, "87174528.ogg"},
	{0xDC5E15D31BA2C2D5ULL, "hlac_lt_darktown_-1_83x.dds"},
	{0xDC5E1DF530064619ULL, "hlac_ca_cavern1_haunted_24423_4x.dds"},
	{0xDC5E28305A847847ULL, "mag291_vincento6027688.cl"},
	{0xDC5E6CEDCC38ECBDULL, "vxlt_gallowstemplar_f_2lj_0_0.vlm"},
	{0xDC5E70501C9B1F2BULL, "gen_im_trs_dal_two_2hs_r1.uti"},
	{0xDC5ECAE39834C0B6ULL, "vxlt_0_lowtown_a2_n_exa_0_0.vlm"},
	{0xDC5F24F2B84FBF4EULL, "art\\levels\\races\\natural\\highlands\\foliage\\plotspecific\\prp_flowerwild.msh"},
	{0xDC5F32F2B84FD69FULL, "art\\levels\\races\\natural\\highlands\\foliage\\plotspecific\\prp_flowerwild.mao"},
	{0xDC5F3EF2B84FEB3CULL, "art\\levels\\races\\natural\\highlands\\foliage\\plotspecific\\prp_flowerwild.mmh"},
	{0xDC5FB43632E2BF93ULL, "w_sundermount_29_0.anb"},
	{0xDC5FCDB17364A998ULL, "11440754.ogg"},
	{0xDC60D0EFC4B90910ULL, "primitives_primitive_p2.vsh.fxo"},
	{0xDC614B16DCCE686AULL, "vxlt_0_lowtown_docks_a3_n_d4b_0_0.vlm"},
	{0xDC6209826DC8EE4EULL, "vxht_mansion2b_2xt_0_1.vlm"},
	{0xDC62A35C95F65E0EULL, "vxht_keep_4at_0_0.vlm"},
	{0xDC62C25E07DA794AULL, "zz_followers_banter6157648.cl"},
	{0xDC62E9F6AB6E6999ULL, "rdr111dg_martin6055005.cl"},
	{0xDC63E648315B7982ULL, "static_static_p43.psh.fxo"},
	{0xDC6424B0DB0070E3ULL, "vxlt_undercityserialkiller_159_0_0.vlm"},
	{0xDC648239EB92782AULL, "vxw_encounter2_hn_0_0.vlm"},
	{0xDC64ED074C7E2007ULL, "mag131_keran6091393.cl"},
	{0xDC66895602B4E2CEULL, "w_encounter3_157_0.msh"},
	{0xDC6690D72F53CBF6ULL, "6065836_m.fxe"},
	{0xDC6696359E427F7DULL, "479718603.ogg"},
	{0xDC66975602B4FA1FULL, "w_encounter3_157_0.mao"},
	{0xDC66A35602B50EBCULL, "w_encounter3_157_0.mmh"},
	{0xDC66AC8D7FE50DFDULL, "6146871_m.fxe"},
	{0xDC670DF017427D73ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.dura.evt"},
	{0xDC6795CCB0F6DEAFULL, "3272762.ogg"},
	{0xDC67C3756FEE3E53ULL, "follower_banter6164751.cl"},
	{0xDC68362D632A00CAULL, "vxlt_blackemporium_8e_0_0.vlm"},
	{0xDC68B68E9D94F03DULL, "gen00fl_isabela6087671.cl"},
	{0xDC68C1BADCF8FCD3ULL, "323836897.ogg"},
	{0xDC6912CBF78C66A7ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_buff.evt"},
	{0xDC691F1ECD934C5BULL, "1059833860.ogg"},
	{0xDC6920D351D237E9ULL, "124587549.ogg"},
	{0xDC6966F4EDF8E258ULL, "vxht_mansion2b_2o2_0_0.vlm"},
	{0xDC6985C1EA3388C2ULL, "201379310.ogg"},
	{0xDC69EE1C3F451C8BULL, "6088836_m.fxe"},
	{0xDC6AD808C799B8EEULL, "vxlt_0_lowtown_docks_a3_n_bcy_0_0.vlm"},
	{0xDC6AF0956B08C381ULL, "mag101_grace6034068.cl"},
	{0xDC6B4175B6DAB7F3ULL, "water_pan2.mao"},
	{0xDC6B43FA034429A4ULL, "art\\vfx\\combat\\vfx_m_amg_conj_c.impa.ani"},
	{0xDC6B81F20EF544B4ULL, "hlac_lt_0_lowtown_a3_n_c_65536_4x.dds"},
	{0xDC6B97F13676878BULL, "303886496.ogg"},
	{0xDC6BBFA7F257C3FCULL, "llac_lt_0_lowtown_a2_-65537_2.dds"},
	{0xDC6BFBFF6F4F060DULL, "273808896.ogg"},
	{0xDC6C24C5757727FAULL, "279823326.ogg"},
	{0xDC6C702322A16E48ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_basic.mmh"},
	{0xDC6C7C2322A182EBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_basic.mao"},
	{0xDC6C8E2322A1A172ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3pp12_basic.msh"},
	{0xDC6CB73632ED4F0EULL, "w_sundermount_29_0.msh"},
	{0xDC6CC53632ED675FULL, "w_sundermount_29_0.mao"},
	{0xDC6CD13632ED7BFCULL, "w_sundermount_29_0.mmh"},
	{0xDC6D1DBFBB488848ULL, "6027717_m.fxe"},
	{0xDC6D6613F32AD65CULL, "493972073.ogg"},
	{0xDC6E06F017488FB4ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.dura.gad"},
	{0xDC6E1CDBE6001C7DULL, "vxlt_0_lowtown_a3_n_c_exq_0_0.vlm"},
	{0xDC6E568450F0AA27ULL, "vxw_woundedcoast_1ls_0_0.vlm"},
	{0xDC6E6B78731A2060ULL, "314353430.ogg"},
	{0xDC6F33CBF7910AF0ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_buff.gad"},
	{0xDC6F96641A702228ULL, "dae000_gen_turn_in_56178425.cl"},
	{0xDC6FCF0456CFE4CBULL, "vxw_blightlands_e3_0_0.vlm"},
	{0xDC7022A7FCF3D8D0ULL, "6135869_m.fxe"},
	{0xDC70295549C62D33ULL, "874121057.ogg"},
	{0xDC70E3DA095700DDULL, "6054493_m.fxe"},
	{0xDC710CC6989D04D8ULL, "hlac_ca_cavern1_haunted_24422_2x.dds"},
	{0xDC7117139CE2B4DAULL, "vxlt_house_fh_0_0.vlm"},
	{0xDC712A9123764872ULL, "and230_mage_girl6109118.cl"},
	{0xDC7140FA03489006ULL, "art\\vfx\\combat\\vfx_m_amg_conj_c.impa.gad"},
	{0xDC714B736B55B44CULL, "462672313.ogg"},
	{0xDC7180F27094B5FFULL, "456532941.ogg"},
	{0xDC71821B18AA6AC6ULL, "zz_vault_debug6178957.cl"},
	{0xDC71AAB7C0AA595DULL, "486451540.ogg"},
	{0xDC71F31F08885339ULL, "w_sundermount_67_0_ll.mmh"},
	{0xDC722AA53AB9CC31ULL, "127987418.ogg"},
	{0xDC72D059A2D25EABULL, "vxlt_0_lowtown_h2h_0_0.vlm"},
	{0xDC73FEEF1031CD78ULL, "cd_gen00fl_isabela.wbd"},
	{0xDC7415F0174D1492ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.dura.ani"},
	{0xDC74B6642847472BULL, "llac_lt_0_lowtown_docks_65533_3.dds"},
	{0xDC74C5D840F4717FULL, "hlac_lt_0_lowtown_a3_n_c_196607_1x.dds"},
	{0xDC74F4B6AA39893DULL, "6092513_f.fxe"},
	{0xDC75347E013B1B68ULL, "6075642_m.fxe"},
	{0xDC7572C71C5F598AULL, "6096944_m.fxe"},
	{0xDC757CB6754B6271ULL, "zz_vault_debug6203234.cl"},
	{0xDC75FEB3122591A6ULL, "vxlt_0_lowtown_a2_n_c_gjs_0_0.vlm"},
	{0xDC7603433880AE35ULL, "vxw_bonepitmining_a2_1zz_0_0.vlm"},
	{0xDC760C6FED8A736FULL, "vxlt_0_lowtown_n_g60_0_0.vlm"},
	{0xDC76204760235D09ULL, "static_static_p24.psh.fxo.dx11"},
	{0xDC76E15E758FDF4EULL, "6048969_m.fxe"},
	{0xDC783BFA034EA6E5ULL, "art\\vfx\\combat\\vfx_m_amg_conj_c.impa.evt"},
	{0xDC7840B9C6C0FE65ULL, "gen00fl_fenris6106648.cl"},
	{0xDC78E6FEC5011F09ULL, "vxlt_darktown_582_0_0.vlm"},
	{0xDC79705FB00EF157ULL, "lt_0_lowtown_a2_43_0_ll.mmh"},
	{0xDC79783414D49850ULL, "vxlt_0_lowtown_docks_btn_0_0.vlm"},
	{0xDC798860A73B0797ULL, "244618882.ogg"},
	{0xDC79D72D924A1ED6ULL, "hlac_ht_0_hightown_a2_-131074_2z.dds"},
	{0xDC7A14A7FF88F7FDULL, "vxlt_0_lowtown_gon_0_0.vlm"},
	{0xDC7A154921A443F3ULL, "6115895_m.fxe"},
	{0xDC7A76CC5E5E45EBULL, "837623066.ogg"},
	{0xDC7AA5617C4045FAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_gurney_sa.dds"},
	{0xDC7ABBD31AA6B24EULL, "561658483.ogg"},
	{0xDC7AC8786C1F5235ULL, "w_bonepit_124_0.msh"},
	{0xDC7ACA786C1F55ABULL, "w_bonepit_124_0.mmh"},
	{0xDC7ACECCD108D3D1ULL, "6111552_m.fxe"},
	{0xDC7AD6786C1F6A48ULL, "w_bonepit_124_0.mao"},
	{0xDC7ADA07826FC3DBULL, "art\\vfx\\combat\\vfx_glb_ele_s_deathblow_c.impa.ani"},
	{0xDC7B46BA373682FEULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrda_0n.dds"},
	{0xDC7BCB6F96FFEE3DULL, "zz_isa_debug6175242.cl"},
	{0xDC7C14797EE3796BULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_em04.phy"},
	{0xDC7C30952C794F50ULL, "vxlt_gallowscourtyard_7fe_0_0.vlm"},
	{0xDC7CAE5F0B78842DULL, "follower_banter6172978.cl"},
	{0xDC7D1C54476BF073ULL, "6201576_m.fxe"},
	{0xDC7D8956462F6D73ULL, "6135246_m.fxe"},
	{0xDC7DC8196D2796D8ULL, "711765179.ogg"},
	{0xDC7E05ED793F40F4ULL, "vxw_bonepitdestroyed_a3_16q_0_0.vlm"},
	{0xDC7E3653D7D1C504ULL, "6123365_m.fxe"},
	{0xDC7E40E148C5FD4BULL, "vxw_blightlands_n2_0_0.vlm"},
	{0xDC7EAD4EA7F0A7A0ULL, "261256127.ogg"},
	{0xDC7ECB1DB70E4208ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_lowdoor2_sa.dds"},
	{0xDC7F0445EACDF55CULL, "485785388.ogg"},
	{0xDC7FE3025A84FF95ULL, "zz_followers_banter6157872.cl"},
	{0xDC801F64768495A0ULL, "114832627.ogg"},
	{0xDC806E804A44765FULL, "190397149.ogg"},
	{0xDC809AB1AD95A7ECULL, "hlac_lt_0_lowtown_docks_-131073_3y.dds"},
	{0xDC812B17659F94F6ULL, "vxht_chantry_2dq_0_0.vlm"},
	{0xDC81423694C8772BULL, "vxlt_0_lowtown_h7k_0_0.vlm"},
	{0xDC818307F67FE671ULL, "largebits.mao"},
	{0xDC81CCE6DD76DB0AULL, "qun120_arishok6050284.cl"},
	{0xDC81EFC99054C451ULL, "vxlt_0_lowtown_a3_n_c_fgw_0_0.vlm"},
	{0xDC8240D737DEA393ULL, "174838314.ogg"},
	{0xDC82770C363E2EFBULL, "pro000_merchant6051278.cl"},
	{0xDC82AC357E9EE8E2ULL, "gen00fl_anders6119683.cl"},
	{0xDC82B6D3BA54EE00ULL, "hlac_ht_0_hightown_a2_196607_0z.dds"},
	{0xDC82F3221E0A4A29ULL, "hlac_lt_0_lowtown_a2_131069_0z.dds"},
	{0xDC835364DCFC25A1ULL, "vxlt_0_lowtown_docks_n_bsi_0_0.vlm"},
	{0xDC8475202A76C6B5ULL, "vxw_bonepitmining_a2_1wy_0_0.vlm"},
	{0xDC849A0D61833544ULL, "503213488.ogg"},
	{0xDC84CC0A883E6276ULL, "vxdr_0_deeproads_cet_0_0.vlm"},
	{0xDC84D3279665DE8BULL, "vxlt_0_lowtown_a2_n_gmu_0_0.vlm"},
	{0xDC852B2CDD794A72ULL, "vxw_bonepitmining_a2_a7_0_0.vlm"},
	{0xDC85382DED3FFDE1ULL, "461452434.ogg"},
	{0xDC855047282A8338ULL, "6038630_m.fxe"},
	{0xDC858A5B5DD00BE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\neck_shot_from_behind.ani"},
	{0xDC864AAED1E51358ULL, "vxht_mansion2b_2e0_0_0.vlm"},
	{0xDC868D0730852826ULL, "static_static_p90.psh.fxo.dx11"},
	{0xDC87033A31A7F61DULL, "6027114_m.fxe"},
	{0xDC871FED8971D6DEULL, "6170837_m.fxe"},
	{0xDC876E7176B218F3ULL, "658474522.ogg"},
	{0xDC8781AEC10C10B2ULL, "follower_banter6167065.cl"},
	{0xDC888684F17F107DULL, "vxlt_0_lowtown_gjm_0_0.vlm"},
	{0xDC88B111C581D064ULL, "569295396.ogg"},
	{0xDC88BC4715BBE881ULL, "qcr000_arishok6166994.cl"},
	{0xDC88D19956EE6BBBULL, "vxlt_gallowscourtyard_n_5yf_0_0.vlm"},
	{0xDC88E207827C0ED2ULL, "art\\vfx\\combat\\vfx_glb_ele_s_deathblow_c.impa.evt"},
	{0xDC88F82871C32E16ULL, "309864200.ogg"},
	{0xDC894F3C39219EB9ULL, "mag330_walter6045856.cl"},
	{0xDC897453F30D3A6BULL, "character_character_p19.psh.fxo.dx11"},
	{0xDC89DD9AD80A586EULL, "vxlt_backalley_n_1ps_0_1.vlm"},
	{0xDC89EA08829BA817ULL, "vxw_sundermount_2e7_0_0.vlm"},
	{0xDC8A8F7B35656D50ULL, "llac_ht_chantry_2534_50.dds"},
	{0xDC8ACBA9366BB1AAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_maso_0s.dds"},
	{0xDC8AE1EFA43DFE85ULL, "sf_d_xy16icf32.vsh.fxo"},
	{0xDC8B146ABAACB1EDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_5.pwk"},
	{0xDC8B2E559C5FBF4FULL, "w_bonepitdestroyed_a3_101_0.mao"},
	{0xDC8B3A3E34DCDB27ULL, "vxw_woundedcoast_1rm_0_0.vlm"},
	{0xDC8B3A559C5FD3ECULL, "w_bonepitdestroyed_a3_101_0.mmh"},
	{0xDC8B40559C5FDDDEULL, "w_bonepitdestroyed_a3_101_0.msh"},
	{0xDC8B71B1B194D5E2ULL, "6105943_m.fxe"},
	{0xDC8B854B48DB96ABULL, "vxw_woundedcoast_1ck_0_1.vlm"},
	{0xDC8BD13633082C4FULL, "w_sundermount_29_0.rml"},
	{0xDC8BE2FB842E7137ULL, "dre000_sibling_has_blight6036561.cl"},
	{0xDC8BEC1D2646B8CDULL, "vxw_bonepit_1ia_0_0.vlm"},
	{0xDC8C875B5DD62546ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\neck_shot_from_behind.gad"},
	{0xDC8CB2BE3ABC15CBULL, "vxw_blightlands_s5_0_0.vlm"},
	{0xDC8CEC3E418BDF87ULL, "vxlt_gallowstemplar_4uf_0_0.vlm"},
	{0xDC8D01DDEB0174A7ULL, "41442055.ogg"},
	{0xDC8D04697627B534ULL, "gen00fl_varric6133683.cl"},
	{0xDC8D4E1157D30C6DULL, "llac_lt_0_lowtown_a2_65535_8.dds"},
	{0xDC8E1149803FCE38ULL, "vxlt_gamlenshouse_n_jr_0_0.vlm"},
	{0xDC8E3E856EE3FE43ULL, "740173526.ogg"},
	{0xDC8E4E7FBC467D5EULL, "vfx_cloud_dark3.dds"},
	{0xDC8E7BD4E26823E2ULL, "6070576_f.fxe"},
	{0xDC8F0B445D4CE514ULL, "follower_banter6136692.cl"},
	{0xDC8F1E06638AC1CFULL, "vxca_caverns2_8k0_0_0.vlm"},
	{0xDC8F5C87471795BDULL, "hlac_ht_mansion2b_818_25x.dds"},
	{0xDC8F94A06C91CAB7ULL, "hlac_dr_primevaltomb_659_14z.dds"},
	{0xDC90010782826265ULL, "art\\vfx\\combat\\vfx_glb_ele_s_deathblow_c.impa.gad"},
	{0xDC900B8E786CD013ULL, "cd_dre000_stroud.bnk"},
	{0xDC90342318BC7D0FULL, "hlac_ht_0_hightown_n_-65534_1z.dds"},
	{0xDC904D3BA7D6FD21ULL, "6113975_f.fxe"},
	{0xDC9088F39D19E5DDULL, "102678414.ogg"},
	{0xDC90E09E8B964575ULL, "894347385.ogg"},
	{0xDC90F97215A51F87ULL, "vxht_0_hightown_a3_atr_0_0.vlm"},
	{0xDC912F3675871287ULL, "hlac_lt_gallowscourtyard_-2_0z.dds"},
	{0xDC9171DD4C85BF9DULL, "gen00fl_varric6152692.cl"},
	{0xDC918A8D44898D21ULL, "6177835_m.fxe"},
	{0xDC91917D75B1D981ULL, "dae200_saemus6141183.cl"},
	{0xDC91B61599481950ULL, "vxca_caverns1_huv_0_0.vlm"},
	{0xDC9252AB241B91D8ULL, "286156710.ogg"},
	{0xDC928E2855EAA56CULL, "6117818_m.fxe"},
	{0xDC92AF995404B087ULL, "vxht_0_hightown_caj_0_0.vlm"},
	{0xDC92BEF17E3FD9D0ULL, "957002102.ogg"},
	{0xDC92CEF86EBCA78FULL, "vxht_0_hightown_c1b_0_0.vlm"},
	{0xDC93803895F6318EULL, "vxlt_0_lowtown_docks_a3_n_d2q_0_1.vlm"},
	{0xDC93825B5DDC3B25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\neck_shot_from_behind.evt"},
	{0xDC939D09CF6F62F2ULL, "vxw_bonepitmining_a2_l0_0_0.vlm"},
	{0xDC93A216E5428E75ULL, "mag131ip_at_undercity.utp"},
	{0xDC93E0BB36E61E8DULL, "6206175_m.fxe"},
	{0xDC93F83F92ABFF3EULL, "6039610_f.fxe"},
	{0xDC940612D93A5505ULL, "lt_0_lowtown_a2_41_0.rml"},
	{0xDC9444A1C10344AAULL, "llac_lt_0_lowtown_docks_n_-131075_1.dds"},
	{0xDC95EF79AAD01031ULL, "429210412.ogg"},
	{0xDC960B8B225C444CULL, "abi_r_rev_attack.wbd"},
	{0xDC96B9F69A1A2353ULL, "6092313_f.fxe"},
	{0xDC975B3A49E282B0ULL, "seb321_leliana6180970.cl"},
	{0xDC982CE16557703DULL, "ts_templars_03.gda"},
	{0xDC98529C63188CB6ULL, "vxlt_gallowstemplar_f_34l_0_0.vlm"},
	{0xDC985BE57491C097ULL, "vxw_sundermount_2n6_0_0.vlm"},
	{0xDC989F810518D84AULL, "mrl351cr_fakeradha.utc"},
	{0xDC991A7DBB736AEEULL, "vxlt_0_lowtown_docks_a3_n_aq4_0_0.vlm"},
	{0xDC99CF527E1F5D26ULL, "6013356_m.fxe"},
	{0xDC99D4273758F001ULL, "llac_lt_0_lowtown_a2_n_c_65537_7.dds"},
	{0xDC9A50BFF6941F5DULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclecturn.phy"},
	{0xDC9A5A556877C927ULL, "hlac_lt_0_lowtown_a2_n_c_131069_0y.dds"},
	{0xDC9AD0FEA78D3B39ULL, "mcr371_angry_mage_m6087026.cl"},
	{0xDC9B249B2CB22E4BULL, "vxw_blightlands_t4_0_0.vlm"},
	{0xDC9BCDFDCD9A7510ULL, "6066317_f.fxe"},
	{0xDC9BD7521A6083D1ULL, "pro000_meeran6050101.cl"},
	{0xDC9BE973E80586E0ULL, "473422437.ogg"},
	{0xDC9C696DD9335A80ULL, "6115309_f.fxe"},
	{0xDC9C80F01DAADCF2ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkb.gad"},
	{0xDC9C8D0DD32C6C00ULL, "vxlt_0_lowtown_a2_gcg_0_0.vlm"},
	{0xDC9CEA32D0632242ULL, "6055079_m.fxe"},
	{0xDC9DBD4841CDA6BBULL, "lla_ht_0_hightown_a2_1_0.dds"},
	{0xDC9DD93E45610C5CULL, "dr_dungeon_mmap_ph.dds"},
	{0xDC9DE543157ABB92ULL, "vxdr_primevaltomb_6wi_0_0.vlm"},
	{0xDC9E2136B26494D9ULL, "6144583_m.fxe"},
	{0xDC9E51EBCB7B0A29ULL, "hf_gen_street_thug.mor"},
	{0xDC9E82807EFD0376ULL, "cod_bks_lite_tombs_last_2.utp"},
	{0xDC9E917C2FA088D0ULL, "hlac_lt_blackemporium_1055_13z.dds"},
	{0xDC9F6B4F079B3807ULL, "vxht_0_hightown_a3_aqm_0_0.vlm"},
	{0xDCA041A7B6D7A3B6ULL, "1036542849.ogg"},
	{0xDCA1CF2E56C18673ULL, "syncdeathlarge.gda"},
	{0xDCA20EE6C1657B72ULL, "vxw_bonepitmining_a2_k1_0_0.vlm"},
	{0xDCA297595A134BD5ULL, "388147783.ogg"},
	{0xDCA2A68E2EC31B46ULL, "vxw_encounter2_7o_0_0.vlm"},
	{0xDCA2EAAEBA4E8788ULL, "hlac_w_woundedcoast_-131071_0y.dds"},
	{0xDCA33292BE7B962AULL, "vrc321_party_barks6095754.cl"},
	{0xDCA404A4C7B2D2F5ULL, "513592163.ogg"},
	{0xDCA46980C4BC5866ULL, "nat380_nathaniel6070248.cl"},
	{0xDCA492BE70F06616ULL, "6089540_m.fxe"},
	{0xDCA5291097B11121ULL, "705015742.ogg"},
	{0xDCA52BD38C103BD3ULL, "6097013_f.fxe"},
	{0xDCA545BD0ACE5668ULL, "vxlt_0_lowtown_docks_a3_n_d0m_0_1.vlm"},
	{0xDCA54D83D1A72B02ULL, "252560141.ogg"},
	{0xDCA56BBE1992B3C6ULL, "art\\vfx\\cinematic\\vfx_meredith_dies_c.impa.evt"},
	{0xDCA6075D37E13721ULL, "darktown_male_lowlife_mob6174414.cl"},
	{0xDCA615A5EA0C5EBDULL, "ht_0_hightown_49_0.rml"},
	{0xDCA6652D05A857B7ULL, "126345109.ogg"},
	{0xDCA6683947013607ULL, "mcr371_hostile_aveline6082194.cl"},
	{0xDCA66BAF2CB5AAEDULL, "hlac_dr_dungeon_3215_0x.dds"},
	{0xDCA6CDC26687D917ULL, "vxw_sundermount_2s1_0_0.vlm"},
	{0xDCA720F08375AE5AULL, "gen00fl_anders6125684.cl"},
	{0xDCA79454E57C76B1ULL, "922169150.ogg"},
	{0xDCA7E89E09F39188ULL, "llac_lt_0_lowtown_a2_n_1_10.dds"},
	{0xDCA824FE61DAFA1AULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_act_c.anb"},
	{0xDCA82BCF5B7C1780ULL, "character_character_p265.psh.fxo.dx11"},
	{0xDCA9375BA4AAB76EULL, "docks_male_human_worker_sta6175982.cl"},
	{0xDCA93A1262DF3E8AULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\malemage2_.gad"},
	{0xDCA94E0A34CBEB3DULL, "953195759.ogg"},
	{0xDCA94E2932205ACFULL, "seb321_leliana6199285.cl"},
	{0xDCA99F10AEB373E4ULL, "69066126.ogg"},
	{0xDCAA217B74C1C1D1ULL, "atl_loginconne_dxt5_02.dds"},
	{0xDCAAC4CDB9FF6B74ULL, "vxca_cavern1_haunted_iir_0_0.vlm"},
	{0xDCAAFFD5E43B19DBULL, "vxw_bonepitmining_a2_1pr_0_0.vlm"},
	{0xDCAB45BFF6A27FC9ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclecturn.mao"},
	{0xDCAB51BFF6A29462ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclecturn.mmh"},
	{0xDCAB53BFF6A29718ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_magiclecturn.msh"},
	{0xDCAB5D77D610567BULL, "6060737_m.fxe"},
	{0xDCAB84B72AADEC43ULL, "vxlt_undercityencounter_13s_0_0.vlm"},
	{0xDCABB842BBAF88BEULL, "6070811_f.fxe"},
	{0xDCABE84EB94F9B49ULL, "hlac_ht_bartrand_n_2995_5z.dds"},
	{0xDCAC321465FB6149ULL, "plt_seb120pt_duty_docks.nss"},
	{0xDCAC57200770D412ULL, "vxdr_primevaltomb_6zj_0_0.vlm"},
	{0xDCAC82BE1998F911ULL, "art\\vfx\\cinematic\\vfx_meredith_dies_c.impa.gad"},
	{0xDCAC8C4BACD1A2C3ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p.ani"},
	{0xDCAC9313A45AAD59ULL, "6141683_m.fxe"},
	{0xDCAD24405D0BAE7AULL, "299175908.ogg"},
	{0xDCAD277E836B5E26ULL, "6098826_f.fxe"},
	{0xDCAD2F1516411A61ULL, "vxca_cavern1_haunted_j2c_0_2.vlm"},
	{0xDCAD6BD46B02CE03ULL, "gen00fl_sebastian6188566.cl"},
	{0xDCAD7A66E6B83F24ULL, "vxlt_0_lowtown_docks_a3_bqn_0_0.vlm"},
	{0xDCADC3C23FD0CA2FULL, "vxht_0_hightown_a2_c3i_0_1.vlm"},
	{0xDCAEBF6B9A884697ULL, "follower_banter6163530.cl"},
	{0xDCAEFCCA66FB1F15ULL, "vxht_keep_n_6fc_0_0.vlm"},
	{0xDCAFE3FE5F65C810ULL, "static_static_p38.psh.fxo"},
	{0xDCAFF44D7BDBA858ULL, "lowtown_argument.cnv"},
	{0xDCB1186B20B933C6ULL, "vxw_encounter2_2p_0_0.vlm"},
	{0xDCB14A8E457DE980ULL, "6075205_f.fxe"},
	{0xDCB169FA647D3E7DULL, "lgt110_gxa_vgk_conspirator36117092.cl"},
	{0xDCB16BF01DBD1498ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_cwkb.ani"},
	{0xDCB1ADCD697A7A3CULL, "vxlt_blackemporium_ze_0_0.vlm"},
	{0xDCB1B6DDB45349F7ULL, "6197589_m.fxe"},
	{0xDCB22FCB8F562886ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.dura.ani"},
	{0xDCB2795C5D90EFC0ULL, "hlac_ht_0_hightown_a2_-65536_4y.dds"},
	{0xDCB285FE183DE3B8ULL, "hlac_ht_0_hightown_a3_65535_0x.dds"},
	{0xDCB2BA244C04B1A3ULL, "gen00fl_dog6172079.cl"},
	{0xDCB2C109821D01D6ULL, "character_character_p120.psh.fxo"},
	{0xDCB34E84ACDF4A01ULL, "hlac_lt_gallowscourtyard_n_-65535_1z.dds"},
	{0xDCB34FCB6CAAB59CULL, "172377845.ogg"},
	{0xDCB3559395FE0DF5ULL, "6139724_f.fxe"},
	{0xDCB383BE199F19BFULL, "art\\vfx\\cinematic\\vfx_meredith_dies_c.impa.ani"},
	{0xDCB3B072D91F704BULL, "hlac_lt_0_lowtown_docks_a3_0_3y.dds"},
	{0xDCB3D64D7BDF5E60ULL, "lowtown_argument.bnk"},
	{0xDCB3DD85408B0D6CULL, "6051892_f.fxe"},
	{0xDCB3F305F0B98938ULL, "one110_meeran6089505.cl"},
	{0xDCB3FBCEE51DCA42ULL, "695915111.ogg"},
	{0xDCB4702DCA4ACC1BULL, "fen110_fenris6101952.cl"},
	{0xDCB47DBF2AC11566ULL, "vxlt_0_lowtown_docks_a3_n_cti_0_0.vlm"},
	{0xDCB4B7928A7C8D15ULL, "hlac_w_encounter2_camp_-65533_0z.dds"},
	{0xDCB4E12855F3762CULL, "rdr110st_dockworker.stg"},
	{0xDCB4F3B0AB041236ULL, "700768680.ogg"},
	{0xDCB53F9F587DF197ULL, "vxw_sundermount_2t0_0_0.vlm"},
	{0xDCB56627D012FB0AULL, "hlac_ht_bartrand_n_3381_0x.dds"},
	{0xDCB56A46A343938CULL, "vxdr_0_deeproads_egf_0_0.vlm"},
	{0xDCB572946B47B11DULL, "qun110_petrice6018533.cl"},
	{0xDCB69E468DC55C55ULL, "6040548_m.fxe"},
	{0xDCB6B68CCE03686BULL, "llac_dr_0_deeproads_5840_0.dds"},
	{0xDCB70B43C5EBD129ULL, "87771874.ogg"},
	{0xDCB74035B1E4C7D7ULL, "hlac_lt_0_lowtown_a2_131071_1z.dds"},
	{0xDCB7B616252E1E08ULL, "l_dae140ar_bone_pit.wbd"},
	{0xDCB81154F25AFBA0ULL, "one100dg_vanard6017476.cl"},
	{0xDCB81BDAFF294C0AULL, "558612892.ogg"},
	{0xDCB820C337CC544AULL, "sky_lt_backalley_n.mao"},
	{0xDCB8C082FCD4E12AULL, "hlac_lt_undercityserialkiller_-1_11x.dds"},
	{0xDCB8CD630FA06E30ULL, "w_bonepitdestroyed_a3_7_0.rml"},
	{0xDCB948734A1469E5ULL, "430950741.ogg"},
	{0xDCB95A6DB182322AULL, "6112676_f.fxe"},
	{0xDCB98D2D0A0D53B4ULL, "llac_lt_gallowstemplar_f_7198_4.dds"},
	{0xDCB9C8958FB4A9D8ULL, "54551448.ogg"},
	{0xDCBA31B47BA02EB5ULL, "vxht_playershouse_n_1r9_0_0.vlm"},
	{0xDCBA4BC10B69F31BULL, "dre000_varric6033318.cl"},
	{0xDCBA8F5B5825F085ULL, "vxlt_0_lowtown_n_glt_0_0.vlm"},
	{0xDCBAA44BACDE08BAULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p.evt"},
	{0xDCBACFCED56CF947ULL, "hlac_ca_caverns2_5802_26x.dds"},
	{0xDCBB39BB0766635BULL, "6171431_m.fxe"},
	{0xDCBB3C464B9745BAULL, "264617826.ogg"},
	{0xDCBB7E0D17C978CDULL, "hlac_lt_0_lowtown_docks_n_-65538_4x.dds"},
	{0xDCBBD8F1D4D2E30BULL, "6106194_m.fxe"},
	{0xDCBC1C012C6F485CULL, "hlac_lt_0_lowtown_docks_n_-4_6x.dds"},
	{0xDCBC3AE9A29ADFDEULL, "vxdr_primevalentrance_2q6_0_1.vlm"},
	{0xDCBD077E233F1CDAULL, "hlac_ca_caverns2_5802_65x.dds"},
	{0xDCBD151262EFA970ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\malemage2_.ani"},
	{0xDCBD6DE0541C99B2ULL, "158950987.ogg"},
	{0xDCBD6EA758F13795ULL, "vxht_keep_n_6ab_0_0.vlm"},
	{0xDCBDED289C35580DULL, "llac_lt_gallowstemplar_2383_9.dds"},
	{0xDCBF4F21D272710BULL, "hlac_w_encounter3_0_9y.dds"},
	{0xDCBFDDA43CF380C6ULL, "vxw_blightlands_af_0_0.vlm"},
	{0xDCBFE869E251FFBBULL, "dae000_cora6162444.cl"},
	{0xDCC017CB8F623DF7ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.dura.evt"},
	{0xDCC01DBFBC172CDBULL, "vxdr_primevaltomb_2q2_0_0.vlm"},
	{0xDCC077431A0431B6ULL, "hlac_dr_0_deeproads_5845_18y.dds"},
	{0xDCC083C12D4AA9C7ULL, "96793456.ogg"},
	{0xDCC0A34BACE272ADULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\mov_p.gad"},
	{0xDCC0B1BABA0713B0ULL, "774316637.ogg"},
	{0xDCC12C0F49699328ULL, "gen00fl_isabela6098197.cl"},
	{0xDCC14D981C9C89F0ULL, "6085249_m.fxe"},
	{0xDCC206B63DDB2493ULL, "6172984_m.fxe"},
	{0xDCC289DEBC9329E6ULL, "6121181_m.fxe"},
	{0xDCC3A0A20E23274CULL, "vxdr_0_deeproads_7ty_0_0.vlm"},
	{0xDCC3A9ECCD7027FEULL, "vxlt_0_lowtown_a2_n_c_emx_0_0.vlm"},
	{0xDCC4891AEDD96FB1ULL, "vxht_mansion2_2vd_0_0.vlm"},
	{0xDCC4B8D1D3A81350ULL, "6087037_f.fxe"},
	{0xDCC4F41DC49267E4ULL, "92888182.ogg"},
	{0xDCC59337CA7A2887ULL, "385339251.ogg"},
	{0xDCC5AABEE10C3535ULL, "isa311_isabela6092163.cl"},
	{0xDCC5BB84341BE22CULL, "leave_06.mao"},
	{0xDCC5E407D5D7D7D9ULL, "mag311_huon6059235.cl"},
	{0xDCC618CB8F66AA80ULL, "art\\vfx\\combat\\vfx_r_brd_csong_loop_c.dura.gad"},
	{0xDCC658D159F94E0CULL, "hlac_lt_0_lowtown_n_1_1z.dds"},
	{0xDCC6FC299F86897DULL, "6018574_m.fxe"},
	{0xDCC7706F895D29DDULL, "6070580_f.fxe"},
	{0xDCC7A8879DEB9C74ULL, "vxca_cavern1_haunted_icp_0_0.vlm"},
	{0xDCC7BABEDB1334ADULL, "463596879.ogg"},
	{0xDCC8797F6663F4D0ULL, "vxht_keep_n_4hn_0_0.vlm"},
	{0xDCC882CC79E781A8ULL, "zz_dre_debug6205403.cl"},
	{0xDCC899C642AABA81ULL, "qun221_petrice6083392.cl"},
	{0xDCC9037D9DE24114ULL, "6014614_m.fxe"},
	{0xDCC9A36987AD1EB4ULL, "6083042_f.fxe"},
	{0xDCC9F9FEA7BA7D84ULL, "384219348.ogg"},
	{0xDCCA078B624560E5ULL, "face_face_p6.psh.fxo.dx11"},
	{0xDCCA70CED60F1BB1ULL, "6072260_m.fxe"},
	{0xDCCA72A660BA29ECULL, "6049228_f.fxe"},
	{0xDCCAB2D5F9F2166DULL, "580636292.ogg"},
	{0xDCCB26F093C56EBBULL, "dae150_chandan6148648.cl"},
	{0xDCCB90E5955F819CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2a.mao"},
	{0xDCCB9CE5955F963FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2a.mmh"},
	{0xDCCBA2E5955FA0E1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_pennants2a.msh"},
	{0xDCCBDA4A89F0A98DULL, "6103410_m.fxe"},
	{0xDCCBE2ACAACE7272ULL, "gen00fl_varric6102797.cl"},
	{0xDCCC2AC5CBE33A75ULL, "vxht_0_hightown_a3_bkz_0_0.vlm"},
	{0xDCCCAA7EDF60D700ULL, "58690910.ogg"},
	{0xDCCDF279D22DDE73ULL, "vxdr_primevalentrance_23h_0_0.vlm"},
	{0xDCCDFC2504A564C6ULL, "vxw_encounter2_8r_0_0.vlm"},
	{0xDCCE4F812EE99946ULL, "vxw_blightlands_jg_0_0.vlm"},
	{0xDCCE883FF3B92E50ULL, "vxw_bonepitdestroyed_a3_1xg_0_0.vlm"},
	{0xDCCE8881829739B8ULL, "al_aveline_s01.gda"},
	{0xDCCEF5745C84638BULL, "llac_lt_0_lowtown_docks_n_-65538_6.dds"},
	{0xDCCFE2FF2F5D8BD7ULL, "vxlt_0_lowtown_elj_0_1.vlm"},
	{0xDCD01CCEE579068CULL, "162780578.ogg"},
	{0xDCD058B61A122283ULL, "6127609_m.fxe"},
	{0xDCD05BA58EEA92F9ULL, "vxht_0_hightown_c32_0_0.vlm"},
	{0xDCD0B2FD66C95B95ULL, "gen00fl_anders6125708.cl"},
	{0xDCD0DFD80B41B4FAULL, "6164973_m.fxe"},
	{0xDCD106616B78DDC6ULL, "vxlt_0_lowtown_a3_n_c_bz7_0_0.vlm"},
	{0xDCD127FE61FE04CDULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_act_c.mmh"},
	{0xDCD161790EAD4666ULL, "vxlt_0_lowtown_docks_a3_n_czk_0_0.vlm"},
	{0xDCD19B97855E1F7FULL, "hlac_ca_caverns1_24424_9y.dds"},
	{0xDCD1E94EADBE17EBULL, "6090433_m.fxe"},
	{0xDCD3412A5009811DULL, "nat380_sibling6068035.cl"},
	{0xDCD3F075A5765575ULL, "ts_dalish_wpn_table.gda"},
	{0xDCD427E13CC79C2CULL, "vxht_brothel_n_541_0_1.vlm"},
	{0xDCD45B29A63DBD9FULL, "mcr370_meredith6083623.cl"},
	{0xDCD4903EAC045440ULL, "431431365.ogg"},
	{0xDCD4DB8808EB4F6BULL, "qun230cr_guild_contact6068875.cl"},
	{0xDCD5D1C1A1BB5B9BULL, "vxht_0_hightown_a3_bpq_0_0.vlm"},
	{0xDCD5E24C7B53425DULL, "6075680_f.fxe"},
	{0xDCD5F0ACF9CEFFDAULL, "103241470.ogg"},
	{0xDCD6148520537C6BULL, "vxlt_gallowscourtyard_70p_0_0.vlm"},
	{0xDCD620E663AC1A51ULL, "vxlt_0_lowtown_hao_0_0.vlm"},
	{0xDCD69809E537231CULL, "t1_mub_pu3.tnt"},
	{0xDCD6EAB7795F2832ULL, "dre000_dougal6129767.cl"},
	{0xDCD7298E78A90671ULL, "cd_dre000_stroud.wbd"},
	{0xDCD7755A8FD85994ULL, "6019114_m.fxe"},
	{0xDCD783308228642CULL, "vxlt_backalley_g_1ev_0_0.vlm"},
	{0xDCD7EE8C22C5AC09ULL, "6169959_m.fxe"},
	{0xDCD83A8C6094FF75ULL, "crt_rubblegolem_rockwraith_fly.bnk"},
	{0xDCD857D63A1E0860ULL, "class_abilities_isabella.gda"},
	{0xDCD92277C9BB1274ULL, "hlac_ca_caverns2_5802_59y.dds"},
	{0xDCD925B6A2986D46ULL, "6067932_f.fxe"},
	{0xDCD944C17078B687ULL, "527378188.ogg"},
	{0xDCD976DBEB297631ULL, "185438266.ogg"},
	{0xDCD9BA13821C6811ULL, "ht_0_hightown_a3_mmi.gff"},
	{0xDCD9D279D83F7726ULL, "186758694.ogg"},
	{0xDCDA36FDEA631096ULL, "ser221_party_comments6166208.cl"},
	{0xDCDA3B371E18E6C2ULL, "t3_wdg_gb1.tnt"},
	{0xDCDA3B5B380ABF7DULL, "lt_gallowscourtyard_n_50_0.rml"},
	{0xDCDA66AB2CECF944ULL, "hlac_ht_mansion2b_818_18y.dds"},
	{0xDCDA9CA2BDD952F5ULL, "vxht_0_hightown_a3_bfy_0_0.vlm"},
	{0xDCDB35ECB01138D1ULL, "hlac_lt_0_lowtown_a3_n_c_1_9x.dds"},
	{0xDCDB49661DDF780EULL, "226703066.ogg"},
	{0xDCDB55EEFF44B925ULL, "ht_0_hightown_a2_38_0.rml"},
	{0xDCDB94493AA161DEULL, "804546753.ogg"},
	{0xDCDB9E6EB8683F79ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_019a_0t.dds"},
	{0xDCDBE1A2E964336FULL, "285209553.ogg"},
	{0xDCDC41B3A555E716ULL, "150823494.ogg"},
	{0xDCDC6456C423F6F3ULL, "vxdr_primevalentrance_26k_0_0.vlm"},
	{0xDCDCC15E20DFB1C6ULL, "vxw_blightlands_oh_0_0.vlm"},
	{0xDCDD03643F5CC3BCULL, "vxlt_blackemporium_ux_0_0.vlm"},
	{0xDCDD5C6417DE67D6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_loop.ani"},
	{0xDCDD7EF0A99DB681ULL, "vxlt_0_lowtown_a2_gxw_0_0.vlm"},
	{0xDCDDEED194521F82ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.impa.gad"},
	{0xDCDE282D2F4C1D3DULL, "543719787.ogg"},
	{0xDCDEB563CD0F45B2ULL, "6053950_m.fxe"},
	{0xDCDEC1B1EF23197FULL, "l_mag321ar_mansion.bnk"},
	{0xDCDECD8280E0AB79ULL, "vxht_0_hightown_c61_0_0.vlm"},
	{0xDCDED2900665331CULL, "ht_keep_3732.anb"},
	{0xDCDEE4F4E4DBD0C5ULL, "hlac_ht_mansion2b_818_39y.dds"},
	{0xDCDEF3C2D1AE934AULL, "6098263_m.fxe"},
	{0xDCE07B18262A5965ULL, "vxht_keep_n_6f3_0_0.vlm"},
	{0xDCE08F2DF44496DBULL, "ave200_aveline_bar6116606.cl"},
	{0xDCE0BC6A356481DBULL, "vxht_playershouse_n_1u2_0_0.vlm"},
	{0xDCE1FE2B1F6AF469ULL, "llac_lt_0_lowtown_n_1_4.dds"},
	{0xDCE23863424D5E2FULL, "6153130_m.fxe"},
	{0xDCE2669C2C5B71ACULL, "dre180cs_sibling_dies.cut"},
	{0xDCE34355D60F6560ULL, "863350478.ogg"},
	{0xDCE35321252DC2D2ULL, "vxlt_0_lowtown_docks_n_b8w_0_0.vlm"},
	{0xDCE392F67A876D3AULL, "6164601_m.fxe"},
	{0xDCE3D4F97936927CULL, "6155265_m.fxe"},
	{0xDCE4D4843436BD3EULL, "leave_06.dds"},
	{0xDCE511D194587A79ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.impa.evt"},
	{0xDCE545514A4E4949ULL, "hlac_lt_0_lowtown_0_8z.dds"},
	{0xDCE5AA7A37B7B390ULL, "dae000_arishok6067746.cl"},
	{0xDCE623D304C7A866ULL, "dae011cr_bartender.utc"},
	{0xDCE6D1DAA0B25E5EULL, "260008824.ogg"},
	{0xDCE7018004B81D4AULL, "vxw_bonepitmining_a2_75_0_0.vlm"},
	{0xDCE724F9D844B13EULL, "gen00fl_varric6079796.cl"},
	{0xDCE7AF2BA780AA33ULL, "gen00fl_aveline6122090.cl"},
	{0xDCE8444855CF3CD3ULL, "vxlt_0_lowtown_docks_n_b5u_0_0.vlm"},
	{0xDCE87FAEA544B7BEULL, "358725891.ogg"},
	{0xDCE8C7603BA83049ULL, "vxlt_0_lowtown_fj4_0_0.vlm"},
	{0xDCE8F1661C9000A4ULL, "769455631.ogg"},
	{0xDCE9820C6FD5DAF2ULL, "hlac_lt_0_lowtown_a3_n_c_0_5x.dds"},
	{0xDCE9B62E63A0607BULL, "vxca_caverns2prc_aa2_0_0.vlm"},
	{0xDCE9C820428F5959ULL, "hlac_lt_gallowscourtyard_-131074_0z.dds"},
	{0xDCE9D2CA5DAB53DCULL, "art\\characters\\playercharacter\\children\\mt_bas_km_lashesm1.mao"},
	{0xDCE9D87D23EB34ACULL, "158325037.ogg"},
	{0xDCEA1F47742279FAULL, "vxht_bartrand_n_63l_0_0.vlm"},
	{0xDCEA32AE6FE63789ULL, "92034210.ogg"},
	{0xDCEA646417E8FFC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_loop.evt"},
	{0xDCEACB5E4B8D85C4ULL, "item_properties.gda"},
	{0xDCEB0B5FADFFFA71ULL, "gen00fl_carver6102325.cl"},
	{0xDCEB333B12D5CA46ULL, "vxw_blightlands_pi_0_0.vlm"},
	{0xDCEB6BF9D7A55F50ULL, "vxw_bonepitdestroyed_a3_1fe_0_0.vlm"},
	{0xDCEBBA551813D80FULL, "75204414.ogg"},
	{0xDCEBC3F4CC54D267ULL, "6074466_m.fxe"},
	{0xDCEBC44D36415338ULL, "6041358_f.fxe"},
	{0xDCEC508B5F219600ULL, "hlac_lt_0_lowtown_docks_a3_65533_4y.dds"},
	{0xDCED3C6FFDFE5383ULL, "6129809_m.fxe"},
	{0xDCEDCA1EA235A056ULL, "art\\vfx\\models\\vfx_plane.mao"},
	{0xDCEDD31249D0E1E7ULL, "vxlt_gallowstemplar_2or_0_0.vlm"},
	{0xDCEDD61EA235B4F5ULL, "art\\vfx\\models\\vfx_plane.mmh"},
	{0xDCEDD81EA235B86BULL, "art\\vfx\\models\\vfx_plane.msh"},
	{0xDCEDE869EADEB17AULL, "vxlt_undercityencounter_1bx_0_1.vlm"},
	{0xDCEDEBE231037AEEULL, "659150484.ogg"},
	{0xDCEE509D207A94ACULL, "6133764_m.fxe"},
	{0xDCEE90BA2A73CD9BULL, "860971635.ogg"},
	{0xDCEE9E60A867F54FULL, "vxw_sundermount_33y_0_1.vlm"},
	{0xDCEEECF5182071E5ULL, "vxht_keep_n_6a2_0_0.vlm"},
	{0xDCEF0442300D115AULL, "555554696.ogg"},
	{0xDCEF8F0143C85879ULL, "6148231_m.fxe"},
	{0xDCEFE11FE8A0200BULL, "vxlt_0_lowtown_docks_n_c3r_0_0.vlm"},
	{0xDCEFE4931A59D511ULL, "follower_banter6161040.cl"},
	{0xDCF0AA40344376AFULL, "6156230_m.fxe"},
	{0xDCF149140940397AULL, "vxht_mansion2b_2l5_0_0.vlm"},
	{0xDCF17148A815C6FBULL, "vxlt_0_lowtown_a2_n_c_hj9_0_0.vlm"},
	{0xDCF178763D25A069ULL, "870608198.ogg"},
	{0xDCF1856417EF5610ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_loop.gad"},
	{0xDCF1BA060B5BA6EEULL, "vxht_bartrand_n_5np_0_0.vlm"},
	{0xDCF1C4FE1723DB52ULL, "vxlt_0_lowtown_docks_n_b5p_0_0.vlm"},
	{0xDCF1D1A4B62455F9ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cooker_n.dds"},
	{0xDCF2902C02C58BD6ULL, "fx_mfx_ad_wing.dds"},
	{0xDCF29B6607FB23FCULL, "lt_0_lowtown_a3_n_c_mmi.gff"},
	{0xDCF2D9F86FF7C623ULL, "vxlt_darktown_4pa_0_0.vlm"},
	{0xDCF2E6277E0F4514ULL, "6202189_m.fxe"},
	{0xDCF319D19464C568ULL, "art\\vfx\\combat\\vfx_q_ari_impale_act_c.impa.ani"},
	{0xDCF33CC6A2EA5FEDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruin.mmh"},
	{0xDCF340C6A2EA663EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruin.mao"},
	{0xDCF34EC6A2EA7E73ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruin.msh"},
	{0xDCF372FB32664FB0ULL, "dae000_arishok6072843.cl"},
	{0xDCF430671D9EFD8FULL, "6107085_m.fxe"},
	{0xDCF5405D66D51C27ULL, "dae000_meredith6092205.cl"},
	{0xDCF58B8A81B882C5ULL, "hlac_lt_0_lowtown_a2_65534_6y.dds"},
	{0xDCF591D973FEF0EBULL, "nat380_nathaniel6067269.cl"},
	{0xDCF6B5CCD6513632ULL, "925809111.ogg"},
	{0xDCF6D91BA6C7069AULL, "lt_0_lowtown_docks_n_105_0.phy"},
	{0xDCF70194F541EAB0ULL, "fen110_anso6094963.cl"},
	{0xDCF823EC67CFA4CFULL, "art\\levels\\races\\proxy\\props\\_textures\\foodstorage_n.dds"},
	{0xDCF8F99DBC24EE92ULL, "ht_playershouse_n_855_lc.mmh"},
	{0xDCF9F4D293718918ULL, "268880435.ogg"},
	{0xDCFAA9F89C2D375AULL, "hlac_dr_slavers_3215_4y.dds"},
	{0xDCFABB0FB8B1A6FFULL, "vxw_bonepitmining_a2_di_0_0.vlm"},
	{0xDCFB404BA42A6016ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_small.evt"},
	{0xDCFB9C5380F7A36EULL, "zz_mag_debug6051365.cl"},
	{0xDCFBC352A676C40CULL, "isa211_isabela6072598.cl"},
	{0xDCFBF8CA8C300E3BULL, "sib111_mother6098838.cl"},
	{0xDCFC02C6BB78644AULL, "6140328_m.fxe"},
	{0xDCFC76C41BF910EAULL, "131297751.ogg"},
	{0xDCFCADE36BEF5A0DULL, "6166987_m.fxe"},
	{0xDCFCB93B8D83DAEAULL, "810333424.ogg"},
	{0xDCFCBF2E6E8A3BA6ULL, "art\\levels\\races\\design\\trap_floorlever_6.pwk"},
	{0xDCFCD4199E7DFB92ULL, "6054263_m.fxe"},
	{0xDCFD64E72ADA65C6ULL, "6136211_m.fxe"},
	{0xDCFD8143BDFE6B02ULL, "seb321_leliana6197143.cl"},
	{0xDCFD844F5DC7A405ULL, "gen00fl_anders6125151.cl"},
	{0xDCFDA77E61EA8E57ULL, "vxlt_gallowscourtyard_62g_0_0.vlm"},
	{0xDCFDBA8D9A9D8424ULL, "176538925.ogg"},
	{0xDCFDE500352C4E63ULL, "499262018.ogg"},
	{0xDCFE00DE35BE70F9ULL, "6145131_m.fxe"},
	{0xDCFE0F569BF3F0D7ULL, "887113043.ogg"},
	{0xDCFEDC16256A61CAULL, "l_dae140ar_bone_pit.bnk"},
	{0xDCFF4541103F4857ULL, "gen00fl_bethany6101225.cl"},
	{0xDCFF5CFE29C897C5ULL, "vxht_mansion2b_2v0_0_0.vlm"},
	{0xDCFF630CB17A1B5CULL, "105624163.ogg"},
	{0xDCFFBAF0FB3651FAULL, "vxht_mansion2b_2i6_0_0.vlm"},
	{0xDD0074EEE1EF7867ULL, "zz_followers_banter6159717.cl"},
	{0xDD00C042B878F628ULL, "vxht_mansion_n_42p_0_0.vlm"},
	{0xDD010D0283025F04ULL, "vxht_0_hightown_n_9ze_0_0.vlm"},
	{0xDD0138B189A25BD5ULL, "gen_im_trs_cot_med_cht_r2.uti"},
	{0xDD013FFDEC1E3331ULL, "17154690.ogg"},
	{0xDD02374BA4306FA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_small.gad"},
	{0xDD023E785DBE6407ULL, "vxlt_0_lowtown_n_es5_0_0.vlm"},
	{0xDD02561C767FA16CULL, "vxlt_darktown_49a_0_0.vlm"},
	{0xDD039E84A7523F71ULL, "594072129.ogg"},
	{0xDD03DEDF5F2A7974ULL, "651472489.ogg"},
	{0xDD03E539E8A44E4AULL, "vxw_bonepitmining_a2_97_0_0.vlm"},
	{0xDD03F99E71222BEAULL, "art\\characters\\weapons\\longswords\\w_lsw_004a.mao"},
	{0xDD03FD9E71223239ULL, "art\\characters\\weapons\\longswords\\w_lsw_004a.mmh"},
	{0xDD04079E71224327ULL, "art\\characters\\weapons\\longswords\\w_lsw_004a.msh"},
	{0xDD0435C6A2F8C6AAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\3wo_ruin.phy"},
	{0xDD04608C60BA7AFFULL, "crt_rubblegolem_rockwraith_fly.wbd"},
	{0xDD04690E67F70D68ULL, "697169552.ogg"},
	{0xDD0488CAA9B7272BULL, "190385766.ogg"},
	{0xDD049187D707E1EFULL, "6109775_m.fxe"},
	{0xDD05CDAD0E5C13C6ULL, "l_env_wal_gallows_military_night.bnk"},
	{0xDD070301580EAAFAULL, "vxht_bartrand_n_69n_0_0.vlm"},
	{0xDD0742EEB23113D5ULL, "static_passthrough_p0.psh.fxo.dx11"},
	{0xDD0760DDD663AC56ULL, "6039976_m.fxe"},
	{0xDD07979B22B231B4ULL, "dae300_vashothlead6062812.cl"},
	{0xDD0896BB72761479ULL, "596824384.ogg"},
	{0xDD092CECAAA7BF7FULL, "vxw_bonepitmining_a2_ch_0_0.vlm"},
	{0xDD09384BA436904FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_weightshift_small.ani"},
	{0xDD0938C007BEA147ULL, "vxlt_undercityencounter_19m_0_0.vlm"},
	{0xDD0977659CBE3E04ULL, "follower_barks6153698.cl"},
	{0xDD09D5FD5B14F60FULL, "hlac_lt_0_lowtown_docks_a3_n_196607_0z.dds"},
	{0xDD0A97D8288E6DCEULL, "pri_ring.mao"},
	{0xDD0B103AA6351C3EULL, "6039401_f.fxe"},
	{0xDD0B1638BC8AE80DULL, "vxdr_primevalentrance_16j_0_0.vlm"},
	{0xDD0BA7B1EF49DAB5ULL, "l_mag321ar_mansion.wbd"},
	{0xDD0BC3EE3D3ABF36ULL, "mag110_cut_abomination.cnv"},
	{0xDD0C172C1D13FA27ULL, "6130054_m.fxe"},
	{0xDD0C195B53E0A6D7ULL, "vxlt_gallowscourtyard_69f_0_0.vlm"},
	{0xDD0C7A5F45570885ULL, "6183964_m.fxe"},
	{0xDD0CC89FB044D475ULL, "static_static_p109.psh.fxo.dx11"},
	{0xDD0D3A5B38361695ULL, "lt_gallowscourtyard_n_50_0.anb"},
	{0xDD0D7567E1A2ECF1ULL, "606830424.ogg"},
	{0xDD0D96B15F8B00E0ULL, "vxlt_0_lowtown_a3_n_c_gyy_0_0.vlm"},
	{0xDD0D9DF2EBE7DFB2ULL, "llac_lt_undercitydungeon_-1_36.dds"},
	{0xDD0DB8382D40C913ULL, "vxht_0_hightown_bxt_0_0.vlm"},
	{0xDD0DBB630FE88FADULL, "w_bonepitdestroyed_a3_7_0.msh"},
	{0xDD0DC9630FE8A7E0ULL, "w_bonepitdestroyed_a3_7_0.mao"},
	{0xDD0DCD630FE8AE33ULL, "w_bonepitdestroyed_a3_7_0.mmh"},
	{0xDD0E2CCDED2C6A7AULL, "vxht_mansion2b_2b7_0_0.vlm"},
	{0xDD0E50ACDF577042ULL, "6142116_f.fxe"},
	{0xDD0EC5EE3D3CF816ULL, "mag110_cut_abomination.bnk"},
	{0xDD0F3D089B043A2DULL, "778747656.ogg"},
	{0xDD0F724CAFC47167ULL, "59514629.ogg"},
	{0xDD0F886D5265DB43ULL, "602545847.ogg"},
	{0xDD0F91130D05F933ULL, "6148116_m.fxe"},
	{0xDD0FBDB253E3F723ULL, "vxlt_darktown_4zc_0_0.vlm"},
	{0xDD102266A39ECD6DULL, "cd_vrc221_housekeeper.bnk"},
	{0xDD102A06970792BFULL, "vxlt_darktown_47l_0_0.vlm"},
	{0xDD105EEEDD342DB1ULL, "art\\characters\\weapons\\staff\\w_stf_014a.mmh"},
	{0xDD106AEEDD344252ULL, "art\\characters\\weapons\\staff\\w_stf_014a.mao"},
	{0xDD1078EEDD345A6FULL, "art\\characters\\weapons\\staff\\w_stf_014a.msh"},
	{0xDD108E939B44937EULL, "6061245_m.fxe"},
	{0xDD10FEA1B11F39C0ULL, "mcr310_argument6072663.cl"},
	{0xDD112C9CD7C5BFB0ULL, "mrl150_dalish_guard6068674.cl"},
	{0xDD11765AC1280255ULL, "w_encounter3_121_0.msh"},
	{0xDD11785AC12805CBULL, "w_encounter3_121_0.mmh"},
	{0xDD11845AC1281A68ULL, "w_encounter3_121_0.mao"},
	{0xDD125CA40FC31937ULL, "296088840.ogg"},
	{0xDD12626D34C9E39CULL, "6047478_m.fxe"},
	{0xDD12C1ADE5C7887BULL, "82067044.ogg"},
	{0xDD13B94394A2C635ULL, "vxdr_0_deeproads_bwp_0_0.vlm"},
	{0xDD13CCBBF9EC39D5ULL, "vxw_bonepit_mv_0_0.vlm"},
	{0xDD13DDBEC95CDF31ULL, "6118379_m.fxe"},
	{0xDD1407CE524B2344ULL, "hlac_lt_0_lowtown_a2_n_c_0_8z.dds"},
	{0xDD147890AE86E519ULL, "ser221_gascard6133740.cl"},
	{0xDD14C0F30B51F085ULL, "hlac_dr_slavers_3217_11z.dds"},
	{0xDD1516EB78970945ULL, "vxht_bartrand_n_63g_0_0.vlm"},
	{0xDD153BD0450A4EA4ULL, "212984677.ogg"},
	{0xDD15A4CB8BDE981EULL, "hf_cgn_003.mop"},
	{0xDD15A6CF5F9AFC68ULL, "vxlt_0_lowtown_a2_f4w_0_0.vlm"},
	{0xDD160BB46C78B88AULL, "follower_banter6141494.cl"},
	{0xDD16B16464189FC4ULL, "6085119_m.fxe"},
	{0xDD16EA1074C93DB3ULL, "dae011ip_order_drink.utp"},
	{0xDD17428490A03F39ULL, "6208555_m.fxe"},
	{0xDD1754BD451457DBULL, "470498151.ogg"},
	{0xDD179EC99C9DD7FFULL, "vxw_bonepitmining_a2_nk_0_0.vlm"},
	{0xDD17DC0234E90960ULL, "25025549.ogg"},
	{0xDD1839527CF9FCC3ULL, "hlac_w_bonepit_1_0x.dds"},
	{0xDD18CB5F8B3561DEULL, "6153166_m.fxe"},
	{0xDD18ECC4168AC90DULL, "903346578.ogg"},
	{0xDD19A646107691ADULL, "w_woundedcoast_121_0.mmh"},
	{0xDD19AA46107698FEULL, "w_woundedcoast_121_0.mao"},
	{0xDD19B8461076B033ULL, "w_woundedcoast_121_0.msh"},
	{0xDD1A175D5D838562ULL, "ch1alphmultnt.psh.fxo"},
	{0xDD1A7089C603AD50ULL, "300754722.ogg"},
	{0xDD1B0FAE6A15D081ULL, "lgt220_shp_mag_guard6149495.cl"},
	{0xDD1B6AD90BA1E121ULL, "hlac_lt_0_lowtown_docks_n_-196610_3y.dds"},
	{0xDD1BA2F831E4381FULL, "dae000_hate_follower_react6167483.cl"},
	{0xDD1BDA7E67DF9337ULL, "art\\vfx\\combat\\vfx_f_mer_emove_app_p.impa.gad"},
	{0xDD1BE1A4BB7D7BF5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair1_sa.dds"},
	{0xDD1C9677FB4D3B7BULL, "994816051.ogg"},
	{0xDD1C9A4D5FF79430ULL, "981430249.ogg"},
	{0xDD1D090DE55AA48AULL, "6205106_m.fxe"},
	{0xDD1D62B02B4E2F48ULL, "261192690.ogg"},
	{0xDD1D913B7F68C29AULL, "w_bonepitmining_a2_7_0.rml"},
	{0xDD1E846102C616BCULL, "hlac_lt_undercityencounter_-1_34y.dds"},
	{0xDD1F223241AA9507ULL, "vxlt_0_lowtown_n_ei7_0_0.vlm"},
	{0xDD1FD190069CC8ACULL, "ht_keep_3732.rml"},
	{0xDD2010196B57C53FULL, "vxdr_0_deeproads_crq_0_1.vlm"},
	{0xDD20387585DFB8D0ULL, "591011625.ogg"},
	{0xDD204A723511B6AAULL, "vxw_encounter2_7c_0_0.vlm"},
	{0xDD20B8B629357905ULL, "hlac_ht_chantry_n_2534_49x.dds"},
	{0xDD2112EF5C396207ULL, "vxca_caverns2_7n1_0_1.vlm"},
	{0xDD217541BAF412EFULL, "6107975_m.fxe"},
	{0xDD21B8D7B3680AF7ULL, "dae221ip_at_keep_htown.utp"},
	{0xDD21DA214F61EE73ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_female.ani"},
	{0xDD21EF7E67E422C4ULL, "art\\vfx\\combat\\vfx_f_mer_emove_app_p.impa.evt"},
	{0xDD21F75456E0B27BULL, "512158474.ogg"},
	{0xDD224F9BBB52F7B1ULL, "6117279_m.fxe"},
	{0xDD22528612336DE6ULL, "95288457.ogg"},
	{0xDD22CF3FE73647B4ULL, "fen210_hadriana6110421.cl"},
	{0xDD23FD2E12A15494ULL, "hlac_w_encounter2_-65533_0y.dds"},
	{0xDD2418AC519114E8ULL, "vxlt_0_lowtown_a2_f1t_0_0.vlm"},
	{0xDD244942A434E5AFULL, "6071883_m.fxe"},
	{0xDD2472A4CDD01E39ULL, "follower_banter6163297.cl"},
	{0xDD2480C433A38306ULL, "vxht_playershouse_n_1v2_0_0.vlm"},
	{0xDD2482698C069D5EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_magc_0n.dds"},
	{0xDD24E55BF8A4E661ULL, "hlac_w_sundermount_131072_0z.dds"},
	{0xDD24EC1EC25F1C89ULL, "6207802_m.fxe"},
	{0xDD251EA083037767ULL, "vxht_0_hightown_a2_bky_0_0.vlm"},
	{0xDD2533BA7CEA46ACULL, "971247411.ogg"},
	{0xDD25B841636C3101ULL, "vxlt_0_lowtown_a2_gcj_0_0.vlm"},
	{0xDD25C27195CBA963ULL, "482627848.ogg"},
	{0xDD261C79EBAAD247ULL, "vxlt_undercityencounter_13o_0_0.vlm"},
	{0xDD26D3ED817272EBULL, "861194921.ogg"},
	{0xDD273D3C7D2B7A5EULL, "6156266_m.fxe"},
	{0xDD2756F323403CD3ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdmgfront.evt"},
	{0xDD283C2604D8A8EBULL, "834402699.ogg"},
	{0xDD287619B28F2FADULL, "811428096.ogg"},
	{0xDD2883D3CD6A55C3ULL, "vxw_bonepitdestroyed_a3_ne_0_0.vlm"},
	{0xDD28A9CA92C9C1E5ULL, "357992213.ogg"},
	{0xDD28D83B67B06428ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunklow_n.dds"},
	{0xDD296EBE88EC403BULL, "6175163_m.fxe"},
	{0xDD2A078AAF5A653EULL, "6083065_m.fxe"},
	{0xDD2A90A103B91BAAULL, "llac_lt_0_lowtown_a2_n_c_-131071_0.dds"},
	{0xDD2A99960F3FDAEBULL, "6162448_m.fxe"},
	{0xDD2A9F461084F96AULL, "w_woundedcoast_121_0.phy"},
	{0xDD2AF5026487D71EULL, "mer120_hubert6070236.cl"},
	{0xDD2AF57BB49BD2DCULL, "mag170_thrask6054077.cl"},
	{0xDD2B477432687B77ULL, "vrc321st_spooky_scene_2.stg"},
	{0xDD2B7AEAD750BD0AULL, "6208606_m.fxe"},
	{0xDD2C5A045AD5C014ULL, "qun150cr_merc.utc"},
	{0xDD2CA16C37D02823ULL, "vxlt_darktown_4le_0_0.vlm"},
	{0xDD2CF021FE8BFFC4ULL, "dae000ip_crowd_ff6_talk1.utp"},
	{0xDD2D0DC07AF3C3BFULL, "vxlt_darktown_49n_0_0.vlm"},
	{0xDD2D116591D78B45ULL, "284092772.ogg"},
	{0xDD2D96B113FCFB08ULL, "ef_nat_elf_dalish.mor"},
	{0xDD2DB5F966199F32ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\dr1c_default_2.pwk"},
	{0xDD2DC59927693A2AULL, "mag291_demon_marethari6029051.cl"},
	{0xDD2DD2751140AB61ULL, "6127086_f.fxe"},
	{0xDD2E4FF323464F14ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdmgfront.gad"},
	{0xDD2ED2214F6C6AAAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_female.evt"},
	{0xDD2EE4B4CDB0499BULL, "vxht_mansion_n_74z_0_0.vlm"},
	{0xDD2EF73022BAD3DAULL, "6208573_f.fxe"},
	{0xDD2F1AF3B23CA38CULL, "674958698.ogg"},
	{0xDD2F723C8A2333A4ULL, "6083529_f.fxe"},
	{0xDD2F789DE5321F28ULL, "vxlt_shop_88_0_0.vlm"},
	{0xDD30034077B7BFACULL, "6113858_m.fxe"},
	{0xDD30077E67F088BDULL, "art\\vfx\\combat\\vfx_f_mer_emove_app_p.impa.ani"},
	{0xDD301B07A349C783ULL, "243700524.ogg"},
	{0xDD3025C5D4E8AAEFULL, "hlac_ht_chantry_2534_19z.dds"},
	{0xDD304F1820AEFD72ULL, "llac_lt_0_lowtown_docks_-196610_0.dds"},
	{0xDD305051DD08335AULL, "vxca_caverns2_941_0_0.vlm"},
	{0xDD30A9C034854BA1ULL, "vxlt_0_lowtown_docks_n_b7e_0_0.vlm"},
	{0xDD3122CBAD9A012DULL, "vxlt_warehouse_11h_0_0.vlm"},
	{0xDD31289BF8FBBA7DULL, "6081749_m.fxe"},
	{0xDD3158E1451662B0ULL, "217907578.ogg"},
	{0xDD316B369CF9633AULL, "fa_rock_f.mao"},
	{0xDD316C6ECB5D08FFULL, "pro000_meeran6048832.cl"},
	{0xDD3186E2646E002DULL, "735363233.ogg"},
	{0xDD31B40C8F249B7BULL, "172930892.ogg"},
	{0xDD31FAA55C833A45ULL, "vxht_bartrand_n_69e_0_0.vlm"},
	{0xDD31FBACA18BC8A5ULL, "vxlt_0_lowtown_a2_gmt_0_0.vlm"},
	{0xDD320C7CF471AC48ULL, "rdr110dg_guard_warehouse6016701.cl"},
	{0xDD32144894B05FEDULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_heathaze.mmh"},
	{0xDD32728C6FC37F44ULL, "263730588.ogg"},
	{0xDD3299240F7A4AE6ULL, "6218830_m.fxe"},
	{0xDD32C3AEF608C92DULL, "6042097_f.fxe"},
	{0xDD3305B5C2A8844DULL, "lt_gallowstemplar_f_2385.rml"},
	{0xDD3331D4272A19D2ULL, "zz_followers_banter6157411.cl"},
	{0xDD333B1C2FDE507EULL, "lt_0_lowtown_a2_n_36_0.rml"},
	{0xDD334D57D1234E6AULL, "vxlt_gallowscourtyard_514_0_0.vlm"},
	{0xDD334F4B4C03A779ULL, "164716949.ogg"},
	{0xDD3367B9B0A85DD4ULL, "vxw_bonepitdestroyed_a3_239_0_0.vlm"},
	{0xDD34186D2CE5E9C9ULL, "vxht_mansion_n_5mm_0_0.vlm"},
	{0xDD3425702BB3B6F5ULL, "275410907.ogg"},
	{0xDD345A01B980786BULL, "vxw_bonepitdestroyed_a3_1wx_0_1.vlm"},
	{0xDD34EB273557C217ULL, "ch2armtnt_s2y.psh.fxo"},
	{0xDD35406C5E53D6D0ULL, "vxlt_0_lowtown_docks_b7s_0_0.vlm"},
	{0xDD355EF3234C87F2ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualdmgfront.ani"},
	{0xDD3590D3567670F0ULL, "vxht_mansion_n_6zr_0_0.vlm"},
	{0xDD35B1214F72511DULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_shrug_off_reaction_female.gad"},
	{0xDD3652F1FABFA79BULL, "ser221_gamlen6131963.cl"},
	{0xDD3662B3A4ACD771ULL, "art\\characters\\creatures\\qunari\\textures\\qm_hed_stn_0s.dds"},
	{0xDD36C945FD71F8DAULL, "hlac_lt_0_lowtown_docks_a3_196609_0z.dds"},
	{0xDD37EDB403D1D63FULL, "6217126_m.fxe"},
	{0xDD383F508CA2D737ULL, "6209852_f.fxe"},
	{0xDD3890998C660127ULL, "vxw_woundedcoast_16a_0_0.vlm"},
	{0xDD38B3322204F50EULL, "6118436_f.fxe"},
	{0xDD38B8ADE1553FF4ULL, "lgt000cr_sketch.utc"},
	{0xDD38EC48356D4A60ULL, "vxlt_0_lowtown_a3_n_c_gvd_0_0.vlm"},
	{0xDD390BC1D4702524ULL, "510398239.ogg"},
	{0xDD39148E931640ADULL, "256712013.ogg"},
	{0xDD3917398FD2F3D4ULL, "gen00fl_varric6118792.cl"},
	{0xDD39DD61D67E843AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_broken.mmh"},
	{0xDD39E161D67E8B71ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_broken.mao"},
	{0xDD39EF61D67EA280ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_broken.msh"},
	{0xDD3A40CC404DCBC4ULL, "vxht_mansion2b_21p_0_0.vlm"},
	{0xDD3A567D1C66E3ECULL, "hlac_lt_0_lowtown_docks_a3_131071_2z.dds"},
	{0xDD3ACB3D1540C45AULL, "6076212_m.fxe"},
	{0xDD3B134929C640A3ULL, "vxlt_darktown_4kd_0_0.vlm"},
	{0xDD3B7127E096E00FULL, "931561233.ogg"},
	{0xDD3B890D62098D55ULL, "66459801.ogg"},
	{0xDD3BE6E953FD1A21ULL, "6083336_m.fxe"},
	{0xDD3C1724F09D0D46ULL, "76708114.ogg"},
	{0xDD3C1866A3C3F697ULL, "cd_vrc221_housekeeper.wbd"},
	{0xDD3C21137A5FF7E9ULL, "542610243.ogg"},
	{0xDD3C2566C899C099ULL, "ht_0_hightown_n_35_0_ll.mmh"},
	{0xDD3D818843421C2AULL, "vxw_blightlands_aj_0_0.vlm"},
	{0xDD3D9BA5AC707FAEULL, "gen_im_trs_slv_sta_phy_r2.uti"},
	{0xDD3DC1C99C0FB113ULL, "262522594.ogg"},
	{0xDD3DD280AAF7CB4AULL, "lowtown_fem_lowlife_mob6169184.cl"},
	{0xDD3DD583BBA6208CULL, "779497559.ogg"},
	{0xDD3DE1B5A6B35F09ULL, "llac_w_bonepitdestroyed_a3_0_7.dds"},
	{0xDD3E17C28BF82F2BULL, "follower_bonuses.gda"},
	{0xDD3E9C9D18940CE3ULL, "hlac_ht_mansion2b_818_38x.dds"},
	{0xDD3ED9C8DE5B6195ULL, "llac_lt_0_lowtown_docks_n_-131074_3.dds"},
	{0xDD3EE95072675DC2ULL, "6167284_m.fxe"},
	{0xDD3F2E20064BE4BDULL, "6136955_m.fxe"},
	{0xDD3F6F6BDA981415ULL, "vxht_keep_n_64i_0_0.vlm"},
	{0xDD3F72D766F94837ULL, "vxca_caverns2_77o_0_0.vlm"},
	{0xDD3F94A89F9019ADULL, "vxlt_warehouse_14g_0_0.vlm"},
	{0xDD4088644D48EB7FULL, "681406657.ogg"},
	{0xDD408D8D22FDB645ULL, "gen00fl_varric6118823.cl"},
	{0xDD40A467330CCF3EULL, "gen00fl_dog6167647.cl"},
	{0xDD40E42145A57971ULL, "combat_2h_r1_combo.gda"},
	{0xDD413D26B88CDCF8ULL, "vxca_cave_23b_0_0.vlm"},
	{0xDD414007FE939C29ULL, "gen00fl_sebastian6178354.cl"},
	{0xDD41603F34FF0B26ULL, "vxht_0_hightown_a3_4c7_0_0.vlm"},
	{0xDD418B4482FDEC7AULL, "hlac_dr_dungeon_3207_2z.dds"},
	{0xDD41BE43792D8AABULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamloop_10s.dura.ani"},
	{0xDD41E54700F81CEFULL, "dae000_meredith6166278.cl"},
	{0xDD4269A4B3F07E4BULL, "6162365_m.fxe"},
	{0xDD4282C0C249088AULL, "289120178.ogg"},
	{0xDD430138BD666450ULL, "131645848.ogg"},
	{0xDD434040C55CD439ULL, "mcr310cr_templar_commander.utc"},
	{0xDD4372D40A5DED43ULL, "487119261.ogg"},
	{0xDD438731DDD84B11ULL, "gen_im_trs_gen_two_2ha_r2.uti"},
	{0xDD43AB39A0B12D22ULL, "vxlt_undercityserialkiller_1l7_0_0.vlm"},
	{0xDD44A007C849545DULL, "vxw_encounter3_95_0_0.vlm"},
	{0xDD45612BB00716F8ULL, "dae000_dalish_apprentices6147753.cl"},
	{0xDD4572E7115973CCULL, "gen00fl_sebastian6180000.cl"},
	{0xDD45A434ABFE5959ULL, "hlac_ca_caverns1_mines_24423_6x.dds"},
	{0xDD45A68416038E61ULL, "272885814.ogg"},
	{0xDD45E4E8695A768EULL, "hlac_dr_dungeon_3217_10x.dds"},
	{0xDD45F36F015B0FACULL, "6060271_m.fxe"},
	{0xDD45F6207BE223D1ULL, "6174046_f.fxe"},
	{0xDD460AED4EA5CC38ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grandright_3.mao"},
	{0xDD462B1CCC0D927FULL, "129046476.ogg"},
	{0xDD4676092368471BULL, "mag321_comtesse_de_launcet6064931.cl"},
	{0xDD46CFAD0E93AE6CULL, "l_env_wal_gallows_military_night.wbd"},
	{0xDD46F3BAE0032BA0ULL, "dae000cr_grand_cleric6153125.cl"},
	{0xDD46FEEA26013C01ULL, "625252210.ogg"},
	{0xDD471EF032EFDB83ULL, "gen00fl_aveline6105439.cl"},
	{0xDD47664202030AD4ULL, "700971754.ogg"},
	{0xDD47E9DDDC1E0765ULL, "lt_gallowstemplar_n_2385.rml"},
	{0xDD486ECF1478ADB6ULL, "6151137_m.fxe"},
	{0xDD48E038E172150BULL, "llac_lt_0_lowtown_a2_n_c_65537_5.dds"},
	{0xDD49907C7EB34A0FULL, "6053375_m.fxe"},
	{0xDD49A94C1F9278F4ULL, "vxca_cavern1_haunted_i0y_0_0.vlm"},
	{0xDD49AFCDF35CDE51ULL, "vxlt_0_lowtown_h9g_0_0.vlm"},
	{0xDD4A3A6EA5706B60ULL, "6116400_m.fxe"},
	{0xDD4B0BB405AB9634ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_3x.dds"},
	{0xDD4B57C95DCC4414ULL, "vxw_woundedcoast_63_0_0.vlm"},
	{0xDD4BA226F19BC852ULL, "vxlt_gallowstemplar_n_3n2_0_0.vlm"},
	{0xDD4BB174D7F80819ULL, "gen00fl_aveline6116730.cl"},
	{0xDD4BD5C085555273ULL, "1067459386.ogg"},
	{0xDD4BF365353834AAULL, "vxw_blightlands_jk_0_0.vlm"},
	{0xDD4C05C91AEB184EULL, "6126896_m.fxe"},
	{0xDD4C288C7C6858BBULL, "vxlt_0_lowtown_a2_n_gn0_0_0.vlm"},
	{0xDD4C2F1F0EA07B28ULL, "303616764.ogg"},
	{0xDD4C574669856045ULL, "810727273.ogg"},
	{0xDD4CA225A0DAEA23ULL, "gen00fl_varric6093643.cl"},
	{0xDD4DD6391D6DDCDFULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthactivate.gad"},
	{0xDD4DEC6DF93A5DFAULL, "w_woundedcoast_139_0.mmh"},
	{0xDD4DF06DF93A6431ULL, "w_woundedcoast_139_0.mao"},
	{0xDD4DFE6DF93A7C40ULL, "w_woundedcoast_139_0.msh"},
	{0xDD4EE67424542266ULL, "307896566.ogg"},
	{0xDD4FA64379399F22ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamloop_10s.dura.evt"},
	{0xDD4FD6B4FD7445D4ULL, "6155517_m.fxe"},
	{0xDD4FF654B4DA5057ULL, "6099698_m.fxe"},
	{0xDD50DB81A5E696CBULL, "6167065_m.fxe"},
	{0xDD511F69BEDB07AFULL, "57000180.ogg"},
	{0xDD5126E38DBD3DC7ULL, "store_dae310_store4.utm"},
	{0xDD51644C27602F40ULL, "mag320cr_crowd_3.utc"},
	{0xDD51C0CC4804FB47ULL, "6043806_m.fxe"},
	{0xDD51C83F1B7BDAE7ULL, "gen00fl_isabela_init6175708.cl"},
	{0xDD51D1BA93128B29ULL, "hlac_w_woundedcoast_131076_0x.dds"},
	{0xDD52CFE255EBB0F2ULL, "qcr220_qunari_hunter6041427.cl"},
	{0xDD52D21BA71559C3ULL, "lt_0_lowtown_docks_n_105_0.msh"},
	{0xDD52E01BA715711DULL, "lt_0_lowtown_docks_n_105_0.mmh"},
	{0xDD52E41BA715782EULL, "lt_0_lowtown_docks_n_105_0.mao"},
	{0xDD5304037D1F90ACULL, "763145990.ogg"},
	{0xDD5311E4BA3F6CDDULL, "vxw_encounter3_24_0_0.vlm"},
	{0xDD5420C3DE4D01B3ULL, "6076011_m.fxe"},
	{0xDD542BB4549E4B84ULL, "w_encounter3_20_0.anb"},
	{0xDD54316F73128340ULL, "717743151.ogg"},
	{0xDD545395C110DD6CULL, "vxlt_warehouse_n_cc_0_0.vlm"},
	{0xDD54B539127708C6ULL, "331134342.ogg"},
	{0xDD54CB391D73E8CCULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthactivate.evt"},
	{0xDD54D7BF7332600CULL, "6149894_f.fxe"},
	{0xDD5510F30F7C8EA7ULL, "ts_templars_01.gda"},
	{0xDD55C47C2799A346ULL, "189218922.ogg"},
	{0xDD5666FBCAE052C4ULL, "vxlt_undercityencounter_1bv_0_1.vlm"},
	{0xDD5669695FFFD0A7ULL, "6107386_m.fxe"},
	{0xDD56A543793FBC35ULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamloop_10s.dura.gad"},
	{0xDD56BCF2C4390375ULL, "630754194.ogg"},
	{0xDD56F41F83BC4F08ULL, "33744315.ogg"},
	{0xDD5778E4B0D4C454ULL, "vxlt_0_lowtown_f1d_0_0.vlm"},
	{0xDD5784EAA1248F99ULL, "gen00fl_sebastian6177913.cl"},
	{0xDD584ECFBA9F4EEBULL, "56826149.ogg"},
	{0xDD585358EFA85ECCULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he04.msh"},
	{0xDD585958EFA868BEULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he04.mmh"},
	{0xDD590AB4EC211069ULL, "llac_ht_keep_n_3739_1.dds"},
	{0xDD5928C5966B792DULL, "hlac_lt_0_lowtown_docks_n_0_6z.dds"},
	{0xDD595CC89B828F60ULL, "6058886_f.fxe"},
	{0xDD59864475556B77ULL, "vxht_mansion2b_1z1_0_0.vlm"},
	{0xDD59C9A64FC25C94ULL, "vxw_woundedcoast_30_0_0.vlm"},
	{0xDD59D89D5A0141BAULL, "695410722.ogg"},
	{0xDD5A1403E391E0D2ULL, "vxlt_gallowstemplar_n_3e3_0_0.vlm"},
	{0xDD5A6542272E4D2AULL, "vxw_blightlands_od_0_0.vlm"},
	{0xDD5AB460636853F3ULL, "hlac_ht_bartrand_n_2995_5x.dds"},
	{0xDD5AE7D859296543ULL, "805985910.ogg"},
	{0xDD5C5D4375E5EB6FULL, "atl_chargen_dxt5_02.dds"},
	{0xDD5C8064C9958B66ULL, "vxht_mansion2b_3ec_0_0.vlm"},
	{0xDD5C96BB9467D5A6ULL, "hlac_lt_0_lowtown_docks_a3_-196611_1y.dds"},
	{0xDD5CA9033D540717ULL, "pro000_friedrich6050183.cl"},
	{0xDD5D74FE7457E6B2ULL, "6123367_m.fxe"},
	{0xDD5DAE368C59ADF2ULL, "202854425.ogg"},
	{0xDD5DC9568ED9E8A3ULL, "vxlt_0_lowtown_docks_a3_czd_0_1.vlm"},
	{0xDD5DFF1075056466ULL, "dae011ip_order_drink.ncs"},
	{0xDD5E00DD3E3411EEULL, "vxht_mansion2b_1v0_0_0.vlm"},
	{0xDD5E2EE81917DB9EULL, "hlac_ht_mansion2b_818_11x.dds"},
	{0xDD5E384506FC395FULL, "6137631_m.fxe"},
	{0xDD5E3C7A2CDB3BC4ULL, "hlac_dr_slavers_3216_9y.dds"},
	{0xDD5E69F8D1824D5FULL, "17603925.ogg"},
	{0xDD5EDA58A6ED433DULL, "591769590.ogg"},
	{0xDD5F575EF158CA97ULL, "hlac_ca_caverns1_mines_24423_11z.dds"},
	{0xDD5FE4AC71B83132ULL, "mag370_meredith6046846.cl"},
	{0xDD5FE6D550E86800ULL, "gen00fl_aveline6120783.cl"},
	{0xDD601C9ED9714691ULL, "dae000_meredith6092575.cl"},
	{0xDD6078853DFAFBCCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_strt.gad"},
	{0xDD607AC898654333ULL, "gen00fl_anders6125218.cl"},
	{0xDD60DD53F09A8BBEULL, "vxw_bonepit_246_0_0.vlm"},
	{0xDD613208331F832EULL, "mer141ip_at_bone_pit.utp"},
	{0xDD6147CF73AC66C7ULL, "lgt200ip_eau_tome.ncs"},
	{0xDD614C9A5C220407ULL, "zz_approval_debug6003066.cl"},
	{0xDD61B3391D7E4A25ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperassasin\\stealthactivate.ani"},
	{0xDD6240B454AAAC99ULL, "w_encounter3_20_0.msh"},
	{0xDD624AB454AABD87ULL, "w_encounter3_20_0.mmh"},
	{0xDD624EB454AAC4D4ULL, "w_encounter3_20_0.mao"},
	{0xDD625A9656F86F0BULL, "hlac_lt_gallowscourtyard_n_-65535_1x.dds"},
	{0xDD629A0FC258C91AULL, "hlac_ht_0_hightown_a3_65535_0z.dds"},
	{0xDD62D3AA901A1D3FULL, "195227539.ogg"},
	{0xDD63940467E4BDEBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_boiler_n.dds"},
	{0xDD6419668FB7302AULL, "667960985.ogg"},
	{0xDD641C356F49D8B1ULL, "188291624.ogg"},
	{0xDD64412371119970ULL, "296909818.ogg"},
	{0xDD644690037F87E8ULL, "6151489_m.fxe"},
	{0xDD6469F35BFB8735ULL, "zz_tbe100_barber6196159.cl"},
	{0xDD6517AAA41034E2ULL, "vxlt_0_lowtown_a2_n_gwj_0_0.vlm"},
	{0xDD651F963D3EA442ULL, "586844148.ogg"},
	{0xDD65713B7FA64251ULL, "w_bonepitmining_a2_7_0.mmh"},
	{0xDD657D3B7FA656F2ULL, "w_bonepitmining_a2_7_0.mao"},
	{0xDD6583469899230FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_intimidate_ex.evt"},
	{0xDD658B3B7FA66E0FULL, "w_bonepitmining_a2_7_0.msh"},
	{0xDD65FE9E7174FE26ULL, "art\\characters\\weapons\\longswords\\w_lsw_004a.phy"},
	{0xDD66720342AC49A3ULL, "575289942.ogg"},
	{0xDD66C2E6531CAAD6ULL, "590645650.ogg"},
	{0xDD6777853E0118CAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_strt.ani"},
	{0xDD6859AD3F7B7ADBULL, "gen00fl_fenris6114533.cl"},
	{0xDD6882D4EC0D63FFULL, "dae000cr_grand_cleric6147688.cl"},
	{0xDD689854356ABCD0ULL, "w_bonepit_209_0.mao"},
	{0xDD689C54356AC3E3ULL, "w_bonepit_209_0.mmh"},
	{0xDD68AA54356ADB3DULL, "w_bonepit_209_0.msh"},
	{0xDD68D3321F73C5EAULL, "6134910_m.fxe"},
	{0xDD68D71F192465AAULL, "vxw_blightlands_pe_0_0.vlm"},
	{0xDD69750536FB6A3DULL, "vxlt_0_lowtown_n_eqm_0_0.vlm"},
	{0xDD6A3AAAA030BA3FULL, "312639407.ogg"},
	{0xDD6ACC387E55734FULL, "llac_lt_0_lowtown_a2_-196609_4.dds"},
	{0xDD6AE1C77D0DBF5DULL, "ave121_brennan6105923.cl"},
	{0xDD6AE30332EA911CULL, "730289421.ogg"},
	{0xDD6CA240832C582BULL, "vxlt_0_lowtown_a2_n_c_hji_0_0.vlm"},
	{0xDD6CCD12C86998E5ULL, "zz_rdr_debug6066660.cl"},
	{0xDD6D381C300FC322ULL, "lt_0_lowtown_a2_n_36_0.anb"},
	{0xDD6DA74F67831F6EULL, "1010758347.ogg"},
	{0xDD6DAACA825DE709ULL, "gen_im_trs_tmp_one_mac_r1.uti"},
	{0xDD6E9E608629322AULL, "6119681_m.fxe"},
	{0xDD6EBC49F07DA416ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_qu101a_0s.dds"},
	{0xDD6F52D5328D271DULL, "vxlt_backalley_n_1oh_0_0.vlm"},
	{0xDD6F5B65A4AEF283ULL, "vxlt_gallowstemplar_n_4qe_0_0.vlm"},
	{0xDD7069856A77074AULL, "840792314.ogg"},
	{0xDD7074DEF853FD3EULL, "12098116.ogg"},
	{0xDD70EDB999C74E65ULL, "vxht_keep_n_649_0_0.vlm"},
	{0xDD710B3CA9600EF0ULL, "6111283_m.fxe"},
	{0xDD7118AB1DDFD93EULL, "mag291_vincento6027342.cl"},
	{0xDD713542072C0094ULL, "205678752.ogg"},
	{0xDD71419C5550A3FBULL, "vxlt_warehouse_15r_0_0.vlm"},
	{0xDD7142716E62F440ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_femalestatue_sa.dds"},
	{0xDD717DC098861C83ULL, "vxlt_gallowstemplar_f_45q_0_0.vlm"},
	{0xDD71ACD6FCBEF3B8ULL, "hf_noblewoman_1.mor"},
	{0xDD71AD5773096792ULL, "gen00fl_fenris6126566.cl"},
	{0xDD7245DABC125101ULL, "ico_1h_sword_3.dds"},
	{0xDD7251FA40AFE26AULL, "qun110_petrice6042261.cl"},
	{0xDD730A5E167A9275ULL, "vxlt_gallowscourtyard_7fd_0_0.vlm"},
	{0xDD730B6DF959AE05ULL, "w_woundedcoast_139_0.phy"},
	{0xDD73695C638C06FFULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.dura.gad"},
	{0xDD73CA560E762F5DULL, "6143748_m.fxe"},
	{0xDD73D92B595709F5ULL, "hlac_ht_0_hightown_1_1x.dds"},
	{0xDD74356EE0D0E4CDULL, "680548422.ogg"},
	{0xDD747351D8F1F7B1ULL, "prm000im_sten_sword.uti"},
	{0xDD753E5E70B6F26BULL, "847623579.ogg"},
	{0xDD7580A5355EC3EDULL, "87535393.ogg"},
	{0xDD758B25874B60F4ULL, "61988622.ogg"},
	{0xDD75A95D3D4B6EABULL, "6137653_m.fxe"},
	{0xDD75A9CB57C74B88ULL, "vxht_bartrand_n_5qg_0_0.vlm"},
	{0xDD75B6182A26577CULL, "6002811_m.fxe"},
	{0xDD76485A29EF8D2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\ready.ani"},
	{0xDD7669C8E8FC2D66ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.cess.ani"},
	{0xDD7695768805416DULL, "hlac_lt_hangedman_-1_7y.dds"},
	{0xDD7748FC0B1A7E2AULL, "vxw_blightlands_uf_0_0.vlm"},
	{0xDD775CAE9AF6BBF9ULL, "6206878_m.fxe"},
	{0xDD776A20801C27DCULL, "vxht_keep_30j_0_0.vlm"},
	{0xDD7777CEDA8EBB67ULL, "1073289764.ogg"},
	{0xDD77E0E07C05C842ULL, "6129260_m.fxe"},
	{0xDD78C046113B8291ULL, "vxht_0_hightown_a2_bzl_0_0.vlm"},
	{0xDD790B17375478ADULL, "pro000cs_flemeth_real.cut"},
	{0xDD795375737F7333ULL, "llac_w_sundermount_0_2.dds"},
	{0xDD795E5C63905FECULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.dura.evt"},
	{0xDD797498B107CE94ULL, "vxlt_0_lowtown_docks_a3_coi_0_0.vlm"},
	{0xDD7981CB1D728C85ULL, "dae000_mother6127517.cl"},
	{0xDD79AB4D88E7F596ULL, "253465823.ogg"},
	{0xDD79BFB224B69B04ULL, "dae000_gamlen6147218.cl"},
	{0xDD79C5DA040541BAULL, "6148402_f.fxe"},
	{0xDD7A0B30323C3841ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_tarp_new_n.dds"},
	{0xDD7A50F56898887BULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_hugetreasure.msh"},
	{0xDD7A62F56898A6C6ULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_hugetreasure.mao"},
	{0xDD7A6EF56898BB65ULL, "art\\levels\\races\\primeval\\props\\plotspecific\\prp_hugetreasure.mmh"},
	{0xDD7A7B2B8F86F2DAULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.dura.gad"},
	{0xDD7AB1AFE2B17CA3ULL, "6129661_m.fxe"},
	{0xDD7AE99FBDD2DB6DULL, "vxht_mansion2b_1uz_0_0.vlm"},
	{0xDD7B05272B2E6793ULL, "qun120cr_javaris6044192.cl"},
	{0xDD7C3273B2AD4531ULL, "vxlt_0_lowtown_docks_alp_0_0.vlm"},
	{0xDD7C9CA16E1EB48EULL, "6128704_m.fxe"},
	{0xDD7CEF127BF296FEULL, "gen00fl_anders6111412.cl"},
	{0xDD7D3826239A2E40ULL, "zz_tbe100_barber6191009.cl"},
	{0xDD7D555A29F5C26EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\ready.gad"},
	{0xDD7DCD4296A50B03ULL, "vxlt_gallowstemplar_n_4vd_0_0.vlm"},
	{0xDD7E0960A147EAE1ULL, "298899846.ogg"},
	{0xDD7F0B39C584FFBFULL, "vxlt_warehouse_n_cl_0_0.vlm"},
	{0xDD7F1C25ACBCE6DFULL, "lgt300cr_jess_varvel.utc"},
	{0xDD7F32A32744851FULL, "470579522.ogg"},
	{0xDD7FA92C96F326ECULL, "vxlt_warehouse_n_pf_0_0.vlm"},
	{0xDD7FBA29360C9BA1ULL, "64187605.ogg"},
	{0xDD7FC29D39A62A90ULL, "mcr310_party_banter6087799.cl"},
	{0xDD7FC32966D9DE74ULL, "rdr111dg_martin6055068.cl"},
	{0xDD8017F187E166CDULL, "161445277.ogg"},
	{0xDD80318A206BE63AULL, "llac_lt_undercityencounter_-1_42.dds"},
	{0xDD810BB2DFCD8A9BULL, "vxca_caverns2prc_9g0_0_0.vlm"},
	{0xDD8129B5BE6FE928ULL, "vxca_caverns2prc_9ua_0_0.vlm"},
	{0xDD817E2B8F8D16F1ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.dura.evt"},
	{0xDD818CAB9EC0C8DEULL, "vxlt_0_lowtown_a2_n_c_ghu_0_0.vlm"},
	{0xDD81C8E2BE73C402ULL, "6139402_f.fxe"},
	{0xDD824A7F02FEE8AAULL, "llac_w_sundermount_-262145_0.dds"},
	{0xDD839ACFBC2F6839ULL, "one300_zevran6093489.cl"},
	{0xDD83DE6B7EAB9659ULL, "458647759.ogg"},
	{0xDD83E54EA6000182ULL, "vxlt_gallowscourtyard_648_0_0.vlm"},
	{0xDD841BA849BD6408ULL, "vxht_bartrand_n_5td_0_0.vlm"},
	{0xDD8427F51C1C6FFCULL, "6005911_m.fxe"},
	{0xDD84505A29FBD84DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\ready.evt"},
	{0xDD8451C8E90841D7ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.cess.evt"},
	{0xDD849C5BCF904ED5ULL, "763492295.ogg"},
	{0xDD84D383EBC0DF8FULL, "6177688_m.fxe"},
	{0xDD84E1AA123CF2EBULL, "6082272_m.fxe"},
	{0xDD84FD5F7E6917E5ULL, "hlac_ht_chantry_2534_12y.dds"},
	{0xDD85E5CAA24B8007ULL, "hlac_w_sundermount_327675_0x.dds"},
	{0xDD8604DB79A598CFULL, "hlac_lt_gallowstemplar_2395_1y.dds"},
	{0xDD86AEB84EE09454ULL, "dae000ip_exit_house.ncs"},
	{0xDD87465C639C7445ULL, "art\\vfx\\combat\\vfx_m_sta_phys_rng_proj_p.dura.ani"},
	{0xDD879BF82CAD23FCULL, "hlac_lt_0_lowtown_131072_1z.dds"},
	{0xDD87A6D9E9963304ULL, "ave200_aveline_bar6116180.cl"},
	{0xDD8804D5605E72C9ULL, "vxlt_hangedman_1jq_0_0.vlm"},
	{0xDD882A5E4ABCDAE8ULL, "hlac_w_encounter2_camp_2_0z.dds"},
	{0xDD8847F13894334AULL, "hlac_ht_bartrand_n_3373_1z.dds"},
	{0xDD887F5E28B4461EULL, "vxht_0_hightown_a2t_0_0.vlm"},
	{0xDD88AD379200BCF5ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wmda_0s.dds"},
	{0xDD89418BE505775AULL, "vxca_cavern1_haunted_j27_0_1.vlm"},
	{0xDD895979CC125A14ULL, "818788293.ogg"},
	{0xDD899EFCDE134507ULL, "hlac_w_bonepitmining_a2_458757_0y.dds"},
	{0xDD89EB72D84775FAULL, "924839566.ogg"},
	{0xDD8A248DDA84D67CULL, "311223912.ogg"},
	{0xDD8AA450A4A35DB1ULL, "vxlt_0_lowtown_docks_aqw_0_0.vlm"},
	{0xDD8AB6C89F6FEA9EULL, "gen00fl_anders6111166.cl"},
	{0xDD8B0496BAB28BABULL, "vxca_cavern1_haunted_e3w_0_0.vlm"},
	{0xDD8B2D9C92978D5CULL, "163972486.ogg"},
	{0xDD8B52C8E90E6260ULL, "art\\vfx\\combat\\vfx_glb_stealth_crust_c.cess.gad"},
	{0xDD8BF68B656FCFD9ULL, "6117006_m.fxe"},
	{0xDD8C0DA5AE7FED24ULL, "ht_bartrand_n_3371_lc.mmh"},
	{0xDD8C368F1679581DULL, "vxlt_backalley_n_1uf_0_0.vlm"},
	{0xDD8C75D097CB75E5ULL, "hlac_lt_0_lowtown_a2_1_3z.dds"},
	{0xDD8CB3CD9DF6D8DAULL, "vxlt_gallowstemplar_3eh_0_0.vlm"},
	{0xDD8D024CA06ED6B3ULL, "vxlt_0_lowtown_n_e1s_0_0.vlm"},
	{0xDD8D37019205886AULL, "263669851.ogg"},
	{0xDD8D4510E82C81CCULL, "128794784.ogg"},
	{0xDD8DAB7C9B3B46C8ULL, "61651287.ogg"},
	{0xDD8E1B0988E93F6CULL, "vxlt_warehouse_n_wg_0_0.vlm"},
	{0xDD8E6BD928901294ULL, "vxlt_gamlenshouse_n_km_0_0.vlm"},
	{0xDD8E8CDDDED4B9A8ULL, "zz_ser_debug6077995.cl"},
	{0xDD8E962B8F97C9C0ULL, "art\\vfx\\combat\\vfx_m_ent_horror_imp_c.dura.ani"},
	{0xDD8ECE3E6CA0C453ULL, "vxlt_blackemporium_4e_0_1.vlm"},
	{0xDD8F1EE16B67900EULL, "dae021_brothel_onenter6196714.cl"},
	{0xDD8F2C45FE448907ULL, "158644748.ogg"},
	{0xDD8F31D064528D7AULL, "376294629.ogg"},
	{0xDD8F9D3865AD270AULL, "qun211cr_braggart_ally_f3.utc"},
	{0xDD8FA0C2E3AFD628ULL, "dae000ip_crowd_em1_talk15.utp"},
	{0xDD8FA2F653E67681ULL, "vfx_glo_enviro.wbd"},
	{0xDD8FBB2EEDB18C7BULL, "283369723.ogg"},
	{0xDD8FE35A9EBE7EDEULL, "6153315_m.fxe"},
	{0xDD9017E31DD3A496ULL, "383180908.ogg"},
	{0xDD907927D0EDFF6EULL, "lt_0_lowtown_docks_41_0.rml"},
	{0xDD90810184B7759CULL, "lt_0_lowtown_docks_a3_n_203_0.mao"},
	{0xDD908D0184B78A3FULL, "lt_0_lowtown_docks_a3_n_203_0.mmh"},
	{0xDD90930184B794E1ULL, "lt_0_lowtown_docks_a3_n_203_0.msh"},
	{0xDD90CB7933529AFDULL, "522479031.ogg"},
	{0xDD9126E4528A7342ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_enter.gad"},
	{0xDD917585754949F6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wi_stala1a_1.mao"},
	{0xDD9182E5815FDA76ULL, "vxht_0_hightown_a3_b9c_0_0.vlm"},
	{0xDD91AE90D23FFD11ULL, "vxht_0_hightown_n_a73_0_0.vlm"},
	{0xDD922FF695B5129AULL, "936249593.ogg"},
	{0xDD93B6A67946C372ULL, "949494748.ogg"},
	{0xDD93D4595A4C6BB9ULL, "vxca_caverns2_973_0_0.vlm"},
	{0xDD93ED61367EEAFBULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_05skeleton.msh"},
	{0xDD93FF61367F0946ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_05skeleton.mao"},
	{0xDD940B61367F1DE5ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_05skeleton.mmh"},
	{0xDD941162F46701F3ULL, "hlac_lt_0_lowtown_0_8x.dds"},
	{0xDD953CBAAFE29154ULL, "hlac_lt_gallowstemplar_n_2391_6x.dds"},
	{0xDD954AB454D617E4ULL, "w_encounter3_20_0.rml"},
	{0xDD957EA850D41EAAULL, "hlac_ht_mansion2_818_61y.dds"},
	{0xDD95C893A6D72577ULL, "mag110_cut_abomination6037104.cl"},
	{0xDD95FB18830CA4D7ULL, "llac_lt_0_lowtown_a2_n_c_-65536_7.dds"},
	{0xDD965C5E2DFB0189ULL, "zz_pro_debug6008058.cl"},
	{0xDD9677816C4F7A93ULL, "68455841.ogg"},
	{0xDD96CD2D758DE1B9ULL, "671686995.ogg"},
	{0xDD96FAAFF141DD3DULL, "6038987_m.fxe"},
	{0xDD97A9F5CFC7B451ULL, "vxca_cavern1_haunted_edp_0_0.vlm"},
	{0xDD980E47864522DEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_lowtown_gate_d.dds"},
	{0xDD9812BFE0299E24ULL, "ser221_gascard6134456.cl"},
	{0xDD9849E45290CD39ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_enter.evt"},
	{0xDD98570F5143A501ULL, "zz_fen_debug6134346.cl"},
	{0xDD98FBFFFA5C2C2DULL, "mer221_sabin6048288.cl"},
	{0xDD993058EFDFBA59ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_he04.phy"},
	{0xDD995AE838D5934EULL, "40217789.ogg"},
	{0xDD99AB617B5B9647ULL, "vxlt_undercityencounter_1kg_0_0.vlm"},
	{0xDD9A192758CC41EAULL, "vxlt_0_lowtown_a2_e15_0_0.vlm"},
	{0xDD9A4DFA59D63521ULL, "6152157_m.fxe"},
	{0xDD9AA86C086F709DULL, "vxlt_backalley_n_1pe_0_0.vlm"},
	{0xDD9B416EBBC8050EULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firetarp_a.mmh"},
	{0xDD9B829F664D9523ULL, "vxlt_warehouse_n_8e_0_0.vlm"},
	{0xDD9C4EB109D372F3ULL, "650314781.ogg"},
	{0xDD9C8CE67ADF57ECULL, "vxlt_warehouse_n_zd_0_0.vlm"},
	{0xDD9CB1DCF9DB96CFULL, "vxht_keep_3fs_0_0.vlm"},
	{0xDD9CD1D4C12E91FAULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_22arches.mmh"},
	{0xDD9CD5D4C12E9831ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_22arches.mao"},
	{0xDD9CDDB61A862B14ULL, "vxlt_gamlenshouse_n_nr_0_0.vlm"},
	{0xDD9CE3D4C12EB040ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_22arches.msh"},
	{0xDD9CF651AD879AC3ULL, "6169665_m.fxe"},
	{0xDD9D12595D374D49ULL, "vxw_encounter2_camp_7d_0_0.vlm"},
	{0xDD9D21FE7F28F6A9ULL, "ps3umbradebug.psh.fxo"},
	{0xDD9DEA43E5A46795ULL, "gen00fl_aveline6114097.cl"},
	{0xDD9E29066BD8F8AAULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_016a_0d.dds"},
	{0xDD9E477131DC8D8DULL, "6122786_f.fxe"},
	{0xDD9E553790B4975EULL, "6156015_m.fxe"},
	{0xDD9EDF1E6BE896E2ULL, "vxlt_0_lowtown_a2_n_gcv_0_0.vlm"},
	{0xDD9FF4C27355F2F6ULL, "vxht_0_hightown_a3_b4d_0_0.vlm"},
	{0xDDA0865E7891309FULL, "vxht_bartrand_n_6se_0_0.vlm"},
	{0xDDA0DDD359B6A00AULL, "lowtown_argument6169417.cl"},
	{0xDDA0E554B3258E79ULL, "963357813.ogg"},
	{0xDDA0F8DBFBB8B586ULL, "338554903.ogg"},
	{0xDDA12037D939FE0CULL, "6104847_m.fxe"},
	{0xDDA152B6AE345765ULL, "hlac_ht_0_hightown_2_0z.dds"},
	{0xDDA1B77B8607086DULL, "hlac_ht_mansion2_818_51y.dds"},
	{0xDDA204C41BF20F40ULL, "vxlt_undercityencounter_rl_0_0.vlm"},
	{0xDDA2151CB6D88A87ULL, "hlac_lt_0_lowtown_a2_65537_2z.dds"},
	{0xDDA246364C428439ULL, "vxca_caverns2_982_0_0.vlm"},
	{0xDDA28B0D5C3C247EULL, "6163129_m.fxe"},
	{0xDDA2A56212F02504ULL, "lgt160_wk_soph_sophia6119826.cl"},
	{0xDDA2C2368B36CF1DULL, "b_gr_fire_len.mao"},
	{0xDDA2E15D555B9F42ULL, "6063078_m.fxe"},
	{0xDDA32F90F886F74DULL, "486946262.ogg"},
	{0xDDA335CE588B0E48ULL, "gen00fl_anders6119209.cl"},
	{0xDDA3ADDD76390D9FULL, "hlac_lt_0_lowtown_a2_n_c_131072_3x.dds"},
	{0xDDA3C2DFBACFD547ULL, "vxlt_0_lowtown_docks_a3_n_bt5_0_0.vlm"},
	{0xDDA40E9237400D98ULL, "hlac_ht_brothel_n_3371_1x.dds"},
	{0xDDA413EAF1A50E7AULL, "art\\characters\\playercharacter\\humanmale\\hoo_robc.phy"},
	{0xDDA4410A4B1F46CFULL, "6124406_m.fxe"},
	{0xDDA455B049D9F338ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow2base.phy"},
	{0xDDA47F871823B447ULL, "676661283.ogg"},
	{0xDDA532EB8F9AB0DBULL, "follower_banter6164508.cl"},
	{0xDDA551E4529B6528ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\sitting_chair_enter.ani"},
	{0xDDA56C928932A4A8ULL, "vxlt_0_lowtown_docks_bx2_0_0.vlm"},
	{0xDDA69E6E1DC2762EULL, "plt_cod_hst_dalish1.plo"},
	{0xDDA6B6B109EE4D24ULL, "vxdr_0_deeproads_f0d_0_0.vlm"},
	{0xDDA71F6504AA53E7ULL, "vxht_0_hightown_a2_b4p_0_0.vlm"},
	{0xDDA720B0A53E30D2ULL, "follower_banter6161272.cl"},
	{0xDDA734DA04A7D58DULL, "hlac_ht_mansion2b_818_31y.dds"},
	{0xDDA75CA3E8B76BCBULL, "657523221.ogg"},
	{0xDDA7D0F2A7E1A22FULL, "334858066.ogg"},
	{0xDDA7EE4F15FC38C1ULL, "6169300_m.fxe"},
	{0xDDA8A90B7AC5F605ULL, "92860826.ogg"},
	{0xDDA8E166CA4ADB7BULL, "6127090_m.fxe"},
	{0xDDAA29FCD688BDEAULL, "vxw_encounter3_e2_0_0.vlm"},
	{0xDDAA60D09B67493FULL, "vxlt_warehouse_n_pi_0_0.vlm"},
	{0xDDAAC290AA5D61B3ULL, "hlac_lt_0_lowtown_docks_n_-131075_4x.dds"},
	{0xDDAB84364F2D65C9ULL, "vxw_encounter2_camp_2k_0_0.vlm"},
	{0xDDAC36AB7242CB4BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_left.ani"},
	{0xDDAC6479BB63CAAEULL, "hlac_ht_chantry_n_2534_18y.dds"},
	{0xDDACBDBADD529D01ULL, "dae000_arishok6067868.cl"},
	{0xDDACE084B0E8825CULL, "1023020446.ogg"},
	{0xDDACF26E3CD2C4C3ULL, "zz_approval_debug6003714.cl"},
	{0xDDAE14B3740F1A39ULL, "6016445_f.fxe"},
	{0xDDAE6E601EEFBFCFULL, "llac_w_sundermount_-196609_0.dds"},
	{0xDDAF1DD5226A80E5ULL, "6152122_m.fxe"},
	{0xDDAF2DD8D1D1D1A3ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.impa.evt"},
	{0xDDAF6D18366BC6A4ULL, "lgt160_wk_soph_sophia6119772.cl"},
	{0xDDAF6F3020697D24ULL, "wm_ico_freemarches_roll.dds"},
	{0xDDAFB616D789A21FULL, "dae000ip_crowd_em3_talk14.utp"},
	{0xDDAFC6CC2E54AE42ULL, "llac_dr_slavers_3207_0.dds"},
	{0xDDB05A373C5B6E7DULL, "hlac_lt_0_lowtown_docks_n_-65539_6y.dds"},
	{0xDDB12ADFC3456C8BULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.dura.ani"},
	{0xDDB1564E896B1592ULL, "fx_glyph_warding.dds"},
	{0xDDB15E68E2D32825ULL, "6119196_m.fxe"},
	{0xDDB1754E4C82FF0DULL, "vxlt_0_lowtown_a3_n_c_cfi_0_0.vlm"},
	{0xDDB17E844D2F57C2ULL, "6174415_m.fxe"},
	{0xDDB1A300D4BCFED2ULL, "art\\characters\\creatures\\desiredemon\\animation\\die_lp_v1.gad"},
	{0xDDB1E459C530F1BAULL, "llac_ca_caverns2prc_5802_51.dds"},
	{0xDDB1FBD08607EE97ULL, "6143016_m.fxe"},
	{0xDDB209DA5D7CE2E4ULL, "mcr370_moira6087501.cl"},
	{0xDDB2B6E61D788BA8ULL, "6161428_m.fxe"},
	{0xDDB359473C108872ULL, "hlac_lt_0_lowtown_n_-196609_0z.dds"},
	{0xDDB4172360CD70F8ULL, "6105825_f.fxe"},
	{0xDDB41CE8EF24A062ULL, "588097119.ogg"},
	{0xDDB450391D98F37CULL, "vxlt_0_lowtown_docks_a3_n_cyx_0_0.vlm"},
	{0xDDB4FDB43D6FD5D5ULL, "402435814.ogg"},
	{0xDDB53BE3ACBB8E64ULL, "6019122_f.fxe"},
	{0xDDB54F2CAC1E9344ULL, "6031134_m.fxe"},
	{0xDDB5970751A6DD95ULL, "gen00fl_varric6076091.cl"},
	{0xDDB5C4F654077F83ULL, "vfx_glo_enviro.bnk"},
	{0xDDB5EDC9288B0BCAULL, "232385243.ogg"},
	{0xDDB606D8D1D7AE24ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.impa.gad"},
	{0xDDB67138FD794077ULL, "6036716_m.fxe"},
	{0xDDB69AD0F5FF3066ULL, "vxlt_house_f4_0_0.vlm"},
	{0xDDB6F02BEBC458F5ULL, "vxw_bonepitdestroyed_a3_1wg_0_0.vlm"},
	{0xDDB72E836B59BFBEULL, "combat_2h_r1_a4.gda"},
	{0xDDB7EF5BE6C5D4A2ULL, "939315466.ogg"},
	{0xDDB8504AC6C8E894ULL, "sp_module_start.ncs"},
	{0xDDB8693C8264B696ULL, "gen_im_wep_mag_sta_act2_04.uti"},
	{0xDDB8D2AD8D5D61BFULL, "vxlt_warehouse_n_wh_0_0.vlm"},
	{0xDDB90365CAE3DF1DULL, "gen00fl_varric6087587.cl"},
	{0xDDB9215E4C9682D6ULL, "454854206.ogg"},
	{0xDDB939F1BCBB81DEULL, "follower_banter6170498.cl"},
	{0xDDB96EB91EE70BC3ULL, "6016927_m.fxe"},
	{0xDDB9964F76615610ULL, "96414631.ogg"},
	{0xDDB9B6A1DC3BDF79ULL, "vxht_0_hightown_n_c7e_0_0.vlm"},
	{0xDDB9DA0B9173CBC3ULL, "6163465_m.fxe"},
	{0xDDB9EC31B928D9B2ULL, "310384645.ogg"},
	{0xDDBA1EAB724EDFC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_left.evt"},
	{0xDDBA3329D94EE89AULL, "228369015.ogg"},
	{0xDDBB72459246FF2BULL, "1066525992.ogg"},
	{0xDDBC307715DAA117ULL, "mag370_meredith6172395.cl"},
	{0xDDBCA0A25B7BE047ULL, "dae000_fenarel6133312.cl"},
	{0xDDBD15D8D1DDE602ULL, "art\\vfx\\combat\\vfx_v_bld_deathmagic_p.impa.ani"},
	{0xDDBD29ABFA601C3CULL, "6101875_m.fxe"},
	{0xDDBD30883169B069ULL, "6077087_m.fxe"},
	{0xDDBD8BA9696B17C1ULL, "darktown_male_elf_sta6174561.cl"},
	{0xDDBDBC015AF11F2FULL, "hlac_lt_0_lowtown_a2_n_2_1z.dds"},
	{0xDDBF12DFC3518102ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.dura.evt"},
	{0xDDBF5348506537F5ULL, "vxht_keep_n_4ho_0_0.vlm"},
	{0xDDBF7C12C35C1053ULL, "hlac_lt_undercitydungeon_-1_33x.dds"},
	{0xDDBFB22DC7CFE6C7ULL, "mrl000_marethari6069861.cl"},
	{0xDDBFB7D62CC4B525ULL, "vxlt_gallowstemplar_f_6sg_0_0.vlm"},
	{0xDDBFCAF88C5F7E47ULL, "340403795.ogg"},
	{0xDDC01DAB725349D5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_twist_look_left.gad"},
	{0xDDC03B259493AD76ULL, "458945122.ogg"},
	{0xDDC043ECF864DE9BULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_enter.ani"},
	{0xDDC046A632BE8BA1ULL, "character_character_p244.psh.fxo.dx11.assembly"},
	{0xDDC0879DA8B59FEAULL, "mag211_marethari6025161.cl"},
	{0xDDC130B60A65C849ULL, "zz_dre_debug6056969.cl"},
	{0xDDC18730EAD12FDCULL, "160582514.ogg"},
	{0xDDC1F0D4C14DE205ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0cc_22arches.phy"},
	{0xDDC220DA14FE292EULL, "1072172378.ogg"},
	{0xDDC2C0298E396749ULL, "vxca_caverns1_j4j_0_0.vlm"},
	{0xDDC2F66E3B9C8FBAULL, "6130086_m.fxe"},
	{0xDDC33B1C9BC6C7AFULL, "gen00fl_anders6128385.cl"},
	{0xDDC38FFA78356CACULL, "gen00fl_anders6119313.cl"},
	{0xDDC3EDC0687C27AEULL, "6099732_m.fxe"},
	{0xDDC4CF15BA334FE9ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_conj_c.impa.ani"},
	{0xDDC51BE9AEB260A9ULL, "art\\characters\\playercharacter\\elfmale\\mt_bas_em_lashesm1.mao"},
	{0xDDC547DC9A0B516BULL, "vxht_keep_5d3_0_0.vlm"},
	{0xDDC55BE766A4ACEBULL, "art\\cinematics\\props\\prp_barge04.mao"},
	{0xDDC56208DDBA7175ULL, "vxw_bonepitdestroyed_a3_1xf_0_0.vlm"},
	{0xDDC57B36A19A5C84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_exit.ani"},
	{0xDDC58E00D4CD8478ULL, "art\\characters\\creatures\\desiredemon\\animation\\die_lp_v1.ani"},
	{0xDDC5F2DCFE484465ULL, "vxlt_gallowstemplar_3oq_0_0.vlm"},
	{0xDDC611DFC3579E15ULL, "art\\vfx\\combat\\vfx_r_ibl_fore_loop_c.dura.gad"},
	{0xDDC66589AFFBEAC7ULL, "hlac_lt_gallowscourtyard_n_-131071_1y.dds"},
	{0xDDC6720DF3CDFDD7ULL, "mer110_jansen6032853.cl"},
	{0xDDC70D725157681FULL, "dae000cr_grand_cleric6187936.cl"},
	{0xDDC715AB172799D5ULL, "zz_vrc_debug6200245.cl"},
	{0xDDC723443665F3EDULL, "553583491.ogg"},
	{0xDDC7448A7F537A3FULL, "vxlt_warehouse_n_zk_0_0.vlm"},
	{0xDDC78707CD3EB88DULL, "34838621.ogg"},
	{0xDDC7B6C12DE3D5D5ULL, "qcr000_cut_arishok_response.bnk"},
	{0xDDC7CFDEFB2151B6ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_8z.dds"},
	{0xDDC7EE330011C44FULL, "vxdr_primevaltomb_6p9_0_0.vlm"},
	{0xDDC80875B212871EULL, "vxlt_gallowstemplar_3j5_0_0.vlm"},
	{0xDDC867E31A4949FEULL, "vxlt_warehouse_18d_0_0.vlm"},
	{0xDDC8A98F81D18CBEULL, "vxlt_0_lowtown_a3_n_c_i0r_0_0.vlm"},
	{0xDDC9148882F3958CULL, "vxlt_gallowstemplar_f_4ty_0_0.vlm"},
	{0xDDC957DA04B82CDFULL, "vxlt_0_lowtown_docks_n_bft_0_0.vlm"},
	{0xDDC9807076C00793ULL, "vxht_0_hightown_b9c_0_0.vlm"},
	{0xDDCA0581081DD195ULL, "gencr_slaver_mage.utc"},
	{0xDDCAB2180470DB44ULL, "532957496.ogg"},
	{0xDDCAD8C12DE6452DULL, "qcr000_cut_arishok_response.cnv"},
	{0xDDCB685E03DA0FAAULL, "character_character_p249.psh.fxo.dx11"},
	{0xDDCB8143DDFD8379ULL, "6060700_m.fxe"},
	{0xDDCB943FEDDAC1A0ULL, "zz_dae_debug6029800.cl"},
	{0xDDCBD27A913E5196ULL, "w_sundermount_146_0.msh"},
	{0xDDCBDC7A913E6294ULL, "w_sundermount_146_0.mmh"},
	{0xDDCBE07A913E69A7ULL, "w_sundermount_146_0.mao"},
	{0xDDCBEA15BA399C6BULL, "art\\vfx\\combat\\vfx_m_pri_tempest_conj_c.impa.gad"},
	{0xDDCC459AEF8A2F6DULL, "305245064.ogg"},
	{0xDDCC5E3A2A9C1B87ULL, "lgt160_ntb_elf_conflict6163371.cl"},
	{0xDDCC7836A1A076E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_exit.gad"},
	{0xDDCC921BE4ABCDB3ULL, "6128507_m.fxe"},
	{0xDDCCA09E55F57BDAULL, "follower_banter6163555.cl"},
	{0xDDCCA6455AE8D703ULL, "mrl000_marethari6129453.cl"},
	{0xDDCCDD4435D9E9A9ULL, "vxlt_gallowstemplar_64y_0_0.vlm"},
	{0xDDCCE13B228367EDULL, "875295531.ogg"},
	{0xDDCCFF8060716ABBULL, "530500493.ogg"},
	{0xDDCD022B517EE09CULL, "927824723.ogg"},
	{0xDDCD11F596A928B9ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.cess.ani"},
	{0xDDCD1C856A540167ULL, "vxlt_0_lowtown_docks_97o_0_0.vlm"},
	{0xDDCD3FE39C82BDA2ULL, "152238924.ogg"},
	{0xDDCD4BECF86F7692ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_enter.evt"},
	{0xDDCD4FAF42BCA1B5ULL, "6065291_m.fxe"},
	{0xDDCD5EB28B0B94EBULL, "tbe100_xenon6190969.cl"},
	{0xDDCDB2CCA1BB4999ULL, "vrc221_hugin6042801.cl"},
	{0xDDCDC4D81AF5BB5FULL, "6072244_m.fxe"},
	{0xDDCDFAFD1D539620ULL, "951285349.ogg"},
	{0xDDCE2A10BE0166EAULL, "vxht_0_hightown_n_by9_0_0.vlm"},
	{0xDDCE4222C6BF5925ULL, "6113796_m.fxe"},
	{0xDDCE9BB56D02B22DULL, "lt_gallowscourtyard_77_0_ll.mmh"},
	{0xDDCEA0967DA574E3ULL, "art\\levels\\races\\design\\_textures\\plc_ballista_01_0_d.dds"},
	{0xDDCF21AC6B9CA1B1ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_dustmots_med.mmh"},
	{0xDDCF6F2827EB9505ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\blendtree\\meredith_pcoverride.anb"},
	{0xDDCFB2B63221FD96ULL, "6042199_f.fxe"},
	{0xDDCFD13CD46DC6BDULL, "6168530_m.fxe"},
	{0xDDCFD2880CA3AFFFULL, "gen00fl_varric6074064.cl"},
	{0xDDD026A7C7639ADCULL, "vxdr_0_deeproads_9il_0_0.vlm"},
	{0xDDD0E8C50C079308ULL, "hlac_lt_backalley_g_65536_3x.dds"},
	{0xDDD10173AB134472ULL, "205869718.ogg"},
	{0xDDD14A5CE82E097AULL, "vxlt_gallowstemplar_f_2e6_0_0.vlm"},
	{0xDDD27336A1A4D9C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_exit.evt"},
	{0xDDD2E715BA3FB5D0ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_conj_c.impa.evt"},
	{0xDDD31EA0A203F825ULL, "smallblur.psh.fxo"},
	{0xDDD34D19BA2E0009ULL, "535429777.ogg"},
	{0xDDD38B71C1F5A997ULL, "hlac_ht_keep_3736_2y.dds"},
	{0xDDD3ECF596AF08FBULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.cess.gad"},
	{0xDDD46AECF875CA25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\thinking_enter.gad"},
	{0xDDD48E99EC3E5C70ULL, "hlac_lt_0_lowtown_docks_n_65535_0y.dds"},
	{0xDDD49ED9FDED1F5DULL, "vxlt_0_lowtown_ewh_0_0.vlm"},
	{0xDDD4F103F72F2CCBULL, "721848106.ogg"},
	{0xDDD5B8248B9D8768ULL, "lt_0_lowtown_docks_a3_108_0.mmh"},
	{0xDDD5C4248B9D9B0BULL, "lt_0_lowtown_docks_a3_108_0.mao"},
	{0xDDD5D6248B9DB992ULL, "lt_0_lowtown_docks_a3_108_0.msh"},
	{0xDDD611823180800BULL, "vxlt_gallowstemplar_5ci_0_0.vlm"},
	{0xDDD6443F7372AF78ULL, "gen00fl_sebastian6180417.cl"},
	{0xDDD6600FF207DCCFULL, "vxdr_primevaltomb_6w8_0_0.vlm"},
	{0xDDD6B41834A13BE1ULL, "431892194.ogg"},
	{0xDDD74F6D759F8D06ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgta_s.dds"},
	{0xDDD793909A9E3014ULL, "mag170_hugh6094562.cl"},
	{0xDDD7A79D437FEA0AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_rob_chaa_0n.dds"},
	{0xDDD7C9B6F6AE455FULL, "vxlt_0_lowtown_docks_n_bcw_0_0.vlm"},
	{0xDDD8E38608E5CCE4ULL, "one280_iwan6096908.cl"},
	{0xDDD8E5BE36155FD0ULL, "6105743_m.fxe"},
	{0xDDD8F8045090FBE1ULL, "77564695.ogg"},
	{0xDDD941683480BD67ULL, "hlac_ht_0_hightown_n_0_3z.dds"},
	{0xDDD95CEDCBEF3D28ULL, "622544522.ogg"},
	{0xDDD981ECDFDE622CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_blueceiling_n.dds"},
	{0xDDDA09F596B3A5E0ULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_proj_p.cess.evt"},
	{0xDDDA537393395BD6ULL, "character_character_p201.psh.fxo.dx11"},
	{0xDDDABEADEC4A9A9BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_diagonal.phy"},
	{0xDDDADF942D159691ULL, "vxlt_0_lowtown_a2_heg_0_0.vlm"},
	{0xDDDAE220F9A7159AULL, "132453440.ogg"},
	{0xDDDB2234C4765B69ULL, "vxdr_dungeon_1ma_0_0.vlm"},
	{0xDDDB679627539A3BULL, "6207814_f.fxe"},
	{0xDDDC7501D56B7A94ULL, "vxht_playershouse_n_1lb_0_0.vlm"},
	{0xDDDCBB842F0968F8ULL, "468687229.ogg"},
	{0xDDDD0CDDA7920327ULL, "869988289.ogg"},
	{0xDDDD110122B8B4B6ULL, "art\\vfx\\materials\\chunky\\vfx_c_fire_tile_scl_e.mao"},
	{0xDDDD31B4CA8D7385ULL, "6083380_f.fxe"},
	{0xDDDD88C44D272B02ULL, "w_encounter2_47_0.rml"},
	{0xDDDDDC35F8CB3381ULL, "763572535.ogg"},
	{0xDDDE977FE8408EE2ULL, "vxlt_0_lowtown_docks_n_c4j_0_0.vlm"},
	{0xDDDFAFDF7108A6FFULL, "6029459_m.fxe"},
	{0xDDE003AB88DB1B51ULL, "700813262.ogg"},
	{0xDDE07C026DF47658ULL, "ts_circleofmagi_arm_props.gda"},
	{0xDDE0D17A4C6858AEULL, "6093532_m.fxe"},
	{0xDDE0E4AE5A37D833ULL, "371019411.ogg"},
	{0xDDE1108F221719D9ULL, "ts_chantry_prog_junk1.gda"},
	{0xDDE159122441B0D3ULL, "dae000_brothel_orana6157214.cl"},
	{0xDDE19A021521F73DULL, "dae021_sister_alema_mob6144851.cl"},
	{0xDDE1FEC8F96C28BBULL, "vxlt_0_lowtown_a2_gyf_0_0.vlm"},
	{0xDDE204B719061E1EULL, "hlac_dr_slavers_3216_0x.dds"},
	{0xDDE2EFC464002392ULL, "gen00fl_aveline6122729.cl"},
	{0xDDE310B6EFE337DDULL, "vxlt_0_lowtown_erg_0_0.vlm"},
	{0xDDE345293E7C5BEDULL, "6164769_m.fxe"},
	{0xDDE346D90EA36EC0ULL, "dae000_brat6156884.cl"},
	{0xDDE3614A81021273ULL, "gen00fl_aveline6121899.cl"},
	{0xDDE38DD166AF7EE6ULL, "l_env_wal_lowtown_act_2_3_crowd_bed.bnk"},
	{0xDDE3A2B143C35364ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_lowerblank_d.dds"},
	{0xDDE4369EEB5E0767ULL, "vxdr_primevaltomb_3ht_0_0.vlm"},
	{0xDDE47DBAD166D0FDULL, "459694844.ogg"},
	{0xDDE489C44D2D4BB6ULL, "w_encounter2_47_0.phy"},
	{0xDDE4F65E1D82CDD1ULL, "vxca_cave_181_0_0.vlm"},
	{0xDDE50AD4C34BC155ULL, "6182710_m.fxe"},
	{0xDDE5132F9060679EULL, "6082296_f.fxe"},
	{0xDDE54A527971A75BULL, "dirty_cir.mao"},
	{0xDDE54B9CFE357AFEULL, "vxlt_warehouse_12f_0_0.vlm"},
	{0xDDE5AB878702EA8BULL, "199900007.ogg"},
	{0xDDE5C6BF40ED6EB4ULL, "vxca_caverns1_mines_j29_0_1.vlm"},
	{0xDDE5E53ABB9E5C38ULL, "vxdr_0_deeproads_eqt_0_0.vlm"},
	{0xDDE62CF7063EA53BULL, "zz_followers_banter6158538.cl"},
	{0xDDE6F29209232427ULL, "419269420.ogg"},
	{0xDDE7193FC0F03C4BULL, "1061463262.ogg"},
	{0xDDE730A3FE2C62BBULL, "6092612_f.fxe"},
	{0xDDE75A565A6B8689ULL, "lt_0_lowtown_docks_138_0.phy"},
	{0xDDE7706ABE6070E4ULL, "isa321_velasco6091874.cl"},
	{0xDDE7992253E39DA5ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_har_blda_0n.dds"},
	{0xDDE82F0E60504258ULL, "vxw_bonepitmining_a2_4b_0_0.vlm"},
	{0xDDE8BF34433141ABULL, "mrl150_merrill6070607.cl"},
	{0xDDE8C1B1576F8E48ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_ext_c.anb"},
	{0xDDE9366900463C79ULL, "hf_ibethany_h3s2.mor"},
	{0xDDE93A20B2A5BACFULL, "art\\characters\\creatures\\ashwraith\\animation\\blendtree\\ashwraith_overlay.anb"},
	{0xDDE951711F0BAF11ULL, "vxlt_0_lowtown_a2_hbf_0_0.vlm"},
	{0xDDE982C38E0DA2E6ULL, "vis121_saemus6101285.cl"},
	{0xDDE9C04A465BB73AULL, "497906177.ogg"},
	{0xDDEA425FC153E16CULL, "plt_mrl000pt_main.plo"},
	{0xDDEB0E7E7E6497A9ULL, "zz_vault_debug6200381.cl"},
	{0xDDEB66BD194D5C50ULL, "hightown_male_noble_gossip6171309.cl"},
	{0xDDEBAD17E7285672ULL, "vxht_chantry_n_23o_0_0.vlm"},
	{0xDDED0A61AB4FCBDCULL, "vxdr_0_deeproads_9gb_0_0.vlm"},
	{0xDDED2CBCC3F81D26ULL, "177404517.ogg"},
	{0xDDEDE36FBA46D1C1ULL, "hlac_lt_gallowscourtyard_n_-1_1y.dds"},
	{0xDDEE10EC1ED3F34CULL, "vxht_0_hightown_au7_0_0.vlm"},
	{0xDDEE72C44D357957ULL, "w_encounter2_47_0.msh"},
	{0xDDEE84C44D3597BAULL, "w_encounter2_47_0.mao"},
	{0xDDEE88C44D359EC9ULL, "w_encounter2_47_0.mmh"},
	{0xDDEE9C816B122DE0ULL, "vxca_cave_2k2_0_0.vlm"},
	{0xDDEF986DDD0E031EULL, "ser120_ghyslain6104800.cl"},
	{0xDDF023791952CC79ULL, "206574931.ogg"},
	{0xDDF0443E3E8585E2ULL, "897389534.ogg"},
	{0xDDF09D736FED9AEBULL, "vxht_keep_5u4_0_0.vlm"},
	{0xDDF0A4F9A99A5296ULL, "6147677_m.fxe"},
	{0xDDF0B5F3DB459986ULL, "vxlt_warehouse_11w_0_0.vlm"},
	{0xDDF0B79FB39CBAF5ULL, "vxw_bonepitdestroyed_a3_1kc_0_0.vlm"},
	{0xDDF1275FC159D1A3ULL, "plt_mrl000pt_main.nss"},
	{0xDDF1455CAD037E51ULL, "hlac_lt_0_lowtown_65537_3y.dds"},
	{0xDDF14D242831479DULL, "127725629.ogg"},
	{0xDDF17D7E023612F2ULL, "gen00fl_isabela6099329.cl"},
	{0xDDF1F576F0674588ULL, "gen_im_npc_wep_two_mera.uti"},
	{0xDDF21087250CE2FCULL, "59130479.ogg"},
	{0xDDF26D0AC99C7153ULL, "354287905.ogg"},
	{0xDDF2E881087C563AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_016a_0t.dds"},
	{0xDDF32AE1D4163550ULL, "200201263.ogg"},
	{0xDDF3850C8256801EULL, "6085396_f.fxe"},
	{0xDDF3F84CE17940EAULL, "6106384_m.fxe"},
	{0xDDF45603B9C7AE16ULL, "hlac_lt_0_lowtown_a3_n_c_65536_4z.dds"},
	{0xDDF482DC82381856ULL, "6089786_m.fxe"},
	{0xDDF486A6A9048F79ULL, "vxht_0_hightown_a2_c1p_0_0.vlm"},
	{0xDDF57B0EB1F22291ULL, "6202203_m.fxe"},
	{0xDDF5CBD1750B006FULL, "308605751.ogg"},
	{0xDDF675FC666769F4ULL, "6162384_m.fxe"},
	{0xDDF79A0381F257DFULL, "vxht_keep_7a2_0_0.vlm"},
	{0xDDF7F0F04905C796ULL, "vxht_mansion2_1gb_0_1.vlm"},
	{0xDDF84C11179E010EULL, "t1_tat_red.tnt"},
	{0xDDF87C29D1A73FCCULL, "fighting80b.mao"},
	{0xDDF8BB77BF2286E0ULL, "958488701.ogg"},
	{0xDDF8DA3494553F2FULL, "ts_qunari_prog_junk2.gda"},
	{0xDDF8FDF6C8F518D7ULL, "vxlt_0_lowtown_efi_0_0.vlm"},
	{0xDDF908BECF290B69ULL, "llac_lt_0_lowtown_a2_n_-65535_4.dds"},
	{0xDDFA0ECF19E4C1D0ULL, "vxw_woundedcoast_vw_0_0.vlm"},
	{0xDDFA3534241F305BULL, "w_bonepitdestroyed_a3_170_0.phy"},
	{0xDDFBD6E0DFC506AFULL, "620071762.ogg"},
	{0xDDFC1B839062290CULL, "vxlt_0_lowtown_docks_a3_n_d44_0_0.vlm"},
	{0xDDFC9567D4AE033CULL, "hlac_lt_0_lowtown_docks_a3_-131075_3x.dds"},
	{0xDDFCB4C6020079BEULL, "405236880.ogg"},
	{0xDDFCE99FE07B37DDULL, "hlac_lt_0_lowtown_0_1y.dds"},
	{0xDDFD2BFAEE86C750ULL, "vxw_bonepit_1za_0_0.vlm"},
	{0xDDFDF2C9018025BCULL, "hlac_lt_undercityencounter_-1_4x.dds"},
	{0xDDFE5869BC551D93ULL, "vxdr_primevaltomb_6ur_0_0.vlm"},
	{0xDDFE7DD4D70BA2B3ULL, "645500951.ogg"},
	{0xDDFE8BD914621396ULL, "6117033_f.fxe"},
	{0xDDFECA933580A10CULL, "gen00fl_fenris6113894.cl"},
	{0xDDFEF15F4A22BBF9ULL, "mag320_orsino.bnk"},
	{0xDDFEFACF8081DFB8ULL, "vxlt_warehouse_146_0_0.vlm"},
	{0xDDFF0CEAF1F1ADA3ULL, "art\\characters\\playercharacter\\humanmale\\hoo_robc.msh"},
	{0xDDFF16D69B906B16ULL, "6140777_m.fxe"},
	{0xDDFF1AEAF1F1C5FDULL, "art\\characters\\playercharacter\\humanmale\\hoo_robc.mmh"},
	{0xDDFF1EEAF1F1CC0EULL, "art\\characters\\playercharacter\\humanmale\\hoo_robc.mao"},
	{0xDDFF2754FAA11F9EULL, "6133674_m.fxe"},
	{0xDDFF27D0CD3BB206ULL, "vxlt_warehouse_14x_0_0.vlm"},
	{0xDDFF297CA592D375ULL, "vxw_bonepitdestroyed_a3_1lb_0_0.vlm"},
	{0xDDFF4F4628AF79C7ULL, "vxht_bartrand_n_5r1_0_0.vlm"},
	{0xDDFF691C1D1B5A3FULL, "hlac_ht_chantry_n_2534_50x.dds"},
	{0xDDFF70378C811743ULL, "vxht_mansion2b_1ml_0_0.vlm"},
	{0xDDFFAD700469AA16ULL, "471103813.ogg"},
	{0xDE0042B04A2831CDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow2base.msh"},
	{0xDE0050B04A284900ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow2base.mao"},
	{0xDE0054B04A285053ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_gallow2base.mmh"},
	{0xDE00DD1B673DA2DBULL, "vxht_chantry_31v_0_0.vlm"},
	{0xDE00E80422B31DF9ULL, "770410587.ogg"},
	{0xDE0167190762C98BULL, "vxlt_gallowstemplar_5rj_0_0.vlm"},
	{0xDE01BD0CDF6A719FULL, "llac_ca_cave_2194_8.dds"},
	{0xDE02135F4A252B59ULL, "mag320_orsino.cnv"},
	{0xDE023DBAC4A1BB8FULL, "dae000ip_crowd_em4_talk2.utp"},
	{0xDE028C432CB11157ULL, "ht_0_hightown_51_0_lc.mmh"},
	{0xDE028EA051CAC7B3ULL, "vxdr_primevaltomb_4hg_0_0.vlm"},
	{0xDE02C8F49F8A8D38ULL, "vxdr_0_deeproads_egr_0_0.vlm"},
	{0xDE03520A1B536B5CULL, "llac_ht_mansion2_818_53.dds"},
	{0xDE03581818E00BD6ULL, "6147905_m.fxe"},
	{0xDE0375A17F83892DULL, "6174535_m.fxe"},
	{0xDE03777D15153401ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.dura.ani"},
	{0xDE037F3B8F700C7AULL, "601918605.ogg"},
	{0xDE03986D44422F56ULL, "462862385.ogg"},
	{0xDE03E65A55E3764DULL, "vxdr_0_deeproads_bhe_0_0.vlm"},
	{0xDE040FA815ECA014ULL, "vxdr_primevaltomb_2sg_0_0.vlm"},
	{0xDE04344E02085CF9ULL, "239064966.ogg"},
	{0xDE047DEA837ECD65ULL, "6006072_m.fxe"},
	{0xDE04870654B21231ULL, "lt_0_lowtown_docks_n_174_0.mmh"},
	{0xDE04930654B226D2ULL, "lt_0_lowtown_docks_n_174_0.mao"},
	{0xDE04A10654B23EEFULL, "lt_0_lowtown_docks_n_174_0.msh"},
	{0xDE04DC21B6CD3349ULL, "647006132.ogg"},
	{0xDE05013FCC8C4AD7ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_n.dds"},
	{0xDE05074F1BAA0454ULL, "vxlt_0_lowtown_n_gbs_0_0.vlm"},
	{0xDE051901C95C8646ULL, "hlac_lt_0_lowtown_docks_a3_-131076_1y.dds"},
	{0xDE05626E62E4908CULL, "gen00fl_fenris6126469.cl"},
	{0xDE060C6AF017F51DULL, "vxlt_0_lowtown_docks_a3_bfb_0_0.vlm"},
	{0xDE0663F6965103AFULL, "6092163_m.fxe"},
	{0xDE06A3038621D6DDULL, "6212808_m.fxe"},
	{0xDE07149FEC1091A5ULL, "526034933.ogg"},
	{0xDE07A5ADEC715E3AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_diagonal.msh"},
	{0xDE07B3ADEC717673ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_diagonal.mao"},
	{0xDE07B7ADEC717C80ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_diagonal.mmh"},
	{0xDE080F1DD6903A3CULL, "600077105.ogg"},
	{0xDE081C3382679A46ULL, "dae000_elegant6166950.cl"},
	{0xDE083DD11D8E68DBULL, "gen00fl_isabela6106167.cl"},
	{0xDE0854C5CF79AB72ULL, "mag321cr_butler.utc"},
	{0xDE0859839C8E77BDULL, "vxlt_0_lowtown_n_e8b_0_0.vlm"},
	{0xDE0880AC0BDADA50ULL, "vxw_woundedcoast_st_0_0.vlm"},
	{0xDE088114BAD346C3ULL, "w_sundermount_58_0.mao"},
	{0xDE088514BAD34D10ULL, "w_sundermount_58_0.mmh"},
	{0xDE089314BAD3652AULL, "w_sundermount_58_0.msh"},
	{0xDE08A1F314E1EE7BULL, "6164598_m.fxe"},
	{0xDE09341C18CA6293ULL, "vxlt_warehouse_n_1cg_0_0.vlm"},
	{0xDE09827D1519B2A3ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.dura.gad"},
	{0xDE0A148CF7CE98C2ULL, "ser221_gamlen6135061.cl"},
	{0xDE0A1CF1C0E4BD9EULL, "dae000_arishok6067663.cl"},
	{0xDE0A432B2186E563ULL, "vxca_caverns1_mines_gk1_0_0.vlm"},
	{0xDE0A8F68317E9399ULL, "6098199_f.fxe"},
	{0xDE0AD511B098F96DULL, "lt_0_lowtown_docks_a3_220_0.mmh"},
	{0xDE0AD911B098FFBEULL, "lt_0_lowtown_docks_a3_220_0.mao"},
	{0xDE0AE711B09917F3ULL, "lt_0_lowtown_docks_a3_220_0.msh"},
	{0xDE0AF94669EDE11EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masa_0d.dds"},
	{0xDE0B686C092CA08FULL, "lowtown_fem_prostitute_sta6152351.cl"},
	{0xDE0BF1D9B020CD0EULL, "112741329.ogg"},
	{0xDE0C38AC90F7E8DAULL, "vxht_0_hightown_a2_atr_0_0.vlm"},
	{0xDE0C4E7ADE608D39ULL, "dae000cr_grand_cleric6187927.cl"},
	{0xDE0C6DDA1510C343ULL, "hlac_lt_0_lowtown_a3_n_c_1_11y.dds"},
	{0xDE0CCA46AE4B3613ULL, "vxdr_primevaltomb_6xu_0_0.vlm"},
	{0xDE0CEE614721CE67ULL, "vxht_brothel_n_4fp_0_0.vlm"},
	{0xDE0D8F81BB2A5F60ULL, "pro000_colonel_ewald6040602.cl"},
	{0xDE0D933A8FC929E4ULL, "lgt320_shw_ors_secret_thug6156370.cl"},
	{0xDE0DC1231AA59247ULL, "vxht_bartrand_n_5u0_0_0.vlm"},
	{0xDE0E1ED305271FCDULL, "llac_w_blightlands_131072_0.dds"},
	{0xDE0E4D3A2489E5A4ULL, "mag121_idunna6084261.cl"},
	{0xDE0F6464ACD325A4ULL, "954350811.ogg"},
	{0xDE0F747152C9CD01ULL, "vxdr_0_deeproads_9g3_0_0.vlm"},
	{0xDE0FBA6CF1A1334EULL, "878617553.ogg"},
	{0xDE0FE23177B5FB14ULL, "6036953_m.fxe"},
	{0xDE100E53973F6C13ULL, "6113991_f.fxe"},
	{0xDE104E0CBCEF1CEDULL, "t1_tat_yel.tnt"},
	{0xDE105F7D151F9588ULL, "art\\vfx\\combat\\vfx_c_ogr_wall_charge_c.dura.evt"},
	{0xDE106DF8AD3B3D3AULL, "6209871_f.fxe"},
	{0xDE107C02C5AE6E1CULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rocks_n.dds"},
	{0xDE10D1AAF0C36591ULL, "vxdr_primevalentrance_2q4_0_0.vlm"},
	{0xDE11669666244956ULL, "6087586_m.fxe"},
	{0xDE119D20A3077E10ULL, "dre000_bodahn6031307.cl"},
	{0xDE11C0B1579291E3ULL, "art\\vfx\\combat\\vfx_glb_lit_imp_ext_c.mmh"},
	{0xDE12004837EFADB3ULL, "445949165.ogg"},
	{0xDE12BC048D69E132ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_007a_0n.dds"},
	{0xDE12BEDEE600A897ULL, "dof.mao"},
	{0xDE131A2EEF994723ULL, "6085264_m.fxe"},
	{0xDE133AF1E09E8C06ULL, "600300287.ogg"},
	{0xDE1343565A909AAEULL, "lt_0_lowtown_docks_138_0.mmh"},
	{0xDE134F565A90AF45ULL, "lt_0_lowtown_docks_138_0.mao"},
	{0xDE135D565A90C69CULL, "lt_0_lowtown_docks_138_0.msh"},
	{0xDE1360AF47083A34ULL, "31847780.ogg"},
	{0xDE13D0E2B7CB765FULL, "art\\levels\\ztools\\default\\generic\\2ci_basic.msh"},
	{0xDE13D6E2B7CB8001ULL, "art\\levels\\ztools\\default\\generic\\2ci_basic.mmh"},
	{0xDE13E2E2B7CB94A2ULL, "art\\levels\\ztools\\default\\generic\\2ci_basic.mao"},
	{0xDE1444AE065F3F53ULL, "dae000_hate_enemies6167591.cl"},
	{0xDE1455939447D8E1ULL, "vxlt_0_lowtown_a2_d7q_0_0.vlm"},
	{0xDE14838D8D97E895ULL, "vxw_encounter2_da_0_0.vlm"},
	{0xDE148F8FD133A052ULL, "hlac_ht_keep_3743_0z.dds"},
	{0xDE14D5D388471C2FULL, "6097263_m.fxe"},
	{0xDE154ED81CE54D35ULL, "ser270_templar6127495.cl"},
	{0xDE1564E2929AEF6BULL, "6119333_f.fxe"},
	{0xDE163C759D43DC94ULL, "vxht_playershouse_n_1pf_0_0.vlm"},
	{0xDE167778C8E243E7ULL, "vxlt_undercityserialkiller_y2_0_0.vlm"},
	{0xDE16CCCB6F7BAAD3ULL, "vxca_caverns2prc_78w_0_0.vlm"},
	{0xDE1733784FC4CF8CULL, "6042024_m.fxe"},
	{0xDE174F2F788A300FULL, "475772627.ogg"},
	{0xDE1755E34B84FF96ULL, "lt_0_lowtown_docks_128_0.mao"},
	{0xDE1761E34B851435ULL, "lt_0_lowtown_docks_128_0.mmh"},
	{0xDE1763E34B8517ABULL, "lt_0_lowtown_docks_128_0.msh"},
	{0xDE17BED869246A93ULL, "gen_im_trs_dwv_med_boo_r2.uti"},
	{0xDE1804834E2791E8ULL, "236432888.ogg"},
	{0xDE184F181088099EULL, "6064563_m.fxe"},
	{0xDE184F43493A18E9ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_glv_hvyo_0n.dds"},
	{0xDE1854EDE94DA1EDULL, "vxlt_0_lowtown_a2_gz9_0_0.vlm"},
	{0xDE190EFB2D391186ULL, "gen00fl_merrill6106241.cl"},
	{0xDE1934E587C04E16ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l.evt"},
	{0xDE1996A2780C13E4ULL, "vxca_caverns1_mines_ii2_0_0.vlm"},
	{0xDE19D53B57093BF9ULL, "389863996.ogg"},
	{0xDE1A28FC1CE28A0CULL, "hlac_lt_gallowstemplar_n_2395_0z.dds"},
	{0xDE1A9435B481978CULL, "hf_amb_commoner1.mor"},
	{0xDE1AF8F22B903E93ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_chantrybasic.phy"},
	{0xDE1C33000C9BAAC7ULL, "vxht_bartrand_n_5x3_0_0.vlm"},
	{0xDE1D4EB8A1CAEB8BULL, "882477018.ogg"},
	{0xDE1DFF3AF29FC83CULL, "645146723.ogg"},
	{0xDE1EAB1F667D1FC1ULL, "650362232.ogg"},
	{0xDE1EADCB8936EA3CULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.dura.ani"},
	{0xDE1EC51C09998995ULL, "217459416.ogg"},
	{0xDE1F0F930C382D1FULL, "6051621_m.fxe"},
	{0xDE1F832929C70AA9ULL, "gen00fl_aveline6139405.cl"},
	{0xDE1F8F89AA6CBBD5ULL, "241580801.ogg"},
	{0xDE1F99C334E54FFEULL, "gen00fl_merrill6119645.cl"},
	{0xDE1F9DF3C4972A66ULL, "vxlt_gallowstemplar_f_3jb_0_0.vlm"},
	{0xDE202BE587C65DA1ULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l.gad"},
	{0xDE20371ECDD5FE5EULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.dura.ani"},
	{0xDE20CB81490261F5ULL, "art\\characters\\heraldry\\hld_raiders.dds"},
	{0xDE20F40C1BEEA4BBULL, "6122040_m.fxe"},
	{0xDE2107DA40B0DE28ULL, "vxlt_0_lowtown_a3_n_c_ehp_0_0.vlm"},
	{0xDE2134E4725A33C0ULL, "6082110_m.fxe"},
	{0xDE2151EC662B760BULL, "ico_1h_sword_1.dds"},
	{0xDE218A93D8734B03ULL, "vxlt_0_lowtown_a2_e1x_0_0.vlm"},
	{0xDE218E902CEA0252ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar2.gad"},
	{0xDE21C5049DD3CBA8ULL, "hlac_dr_0_deeproads_5842_16z.dds"},
	{0xDE22395FA44931ACULL, "69809275.ogg"},
	{0xDE2243359F3F8EA1ULL, "follower_banter6172259.cl"},
	{0xDE22586880DBD7CFULL, "vxht_keep_n_30j_0_0.vlm"},
	{0xDE236134301BBB93ULL, "vxw_bonepit_z9_0_0.vlm"},
	{0xDE248FD166E7188CULL, "l_env_wal_lowtown_act_2_3_crowd_bed.wbd"},
	{0xDE24A7DE4BCABD29ULL, "6075328_m.fxe"},
	{0xDE24AE528F39F514ULL, "vxht_playershouse_n_1wg_0_0.vlm"},
	{0xDE24E7B998D84D3EULL, "vxlt_0_lowtown_docks_a3_d4u_0_0.vlm"},
	{0xDE2506909EAAF5FFULL, "vxlt_0_lowtown_a3_n_c_fec_0_0.vlm"},
	{0xDE2506DC93953226ULL, "6109733_f.fxe"},
	{0xDE254ABA2146A5B3ULL, "18909529.ogg"},
	{0xDE25613C5D09C47CULL, "hlac_lt_gallowsprison_n_-1_9x.dds"},
	{0xDE258152F8CF1B20ULL, "746555798.ogg"},
	{0xDE25BACB893D1F7EULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.dura.gad"},
	{0xDE25C9692FBDBB96ULL, "53639697.ogg"},
	{0xDE25C9963616C326ULL, "260529488.ogg"},
	{0xDE25EF55D0E3CE7FULL, "731978185.ogg"},
	{0xDE261C34244440FAULL, "w_bonepitdestroyed_a3_170_0.msh"},
	{0xDE262A3424445833ULL, "w_bonepitdestroyed_a3_170_0.mao"},
	{0xDE262E3424445F40ULL, "w_bonepitdestroyed_a3_170_0.mmh"},
	{0xDE267F2705AB23F6ULL, "37341743.ogg"},
	{0xDE26B45AE887381BULL, "6051889_m.fxe"},
	{0xDE26E5C9DD15E089ULL, "vxlt_gallowscourtyard_n_7fs_0_0.vlm"},
	{0xDE272CE587CC7E4FULL, "art\\characters\\creatures\\stonegolem\\animation\\ctn180l.ani"},
	{0xDE2743A72F67D24BULL, "vxdr_0_deeproads_a9w_0_0.vlm"},
	{0xDE27DB7A918CBF8FULL, "w_sundermount_146_0.phy"},
	{0xDE285AF5D7EE64F8ULL, "vxca_cave_2cr_0_0.vlm"},
	{0xDE291C6674E419DAULL, "vxht_0_hightown_a2_ajl_0_0.vlm"},
	{0xDE292B2A2A2AB4BFULL, "art\\characters\\heraldry\\hld_antivan_crows.dds"},
	{0xDE29967BA217DFC0ULL, "363806336.ogg"},
	{0xDE2A24E478329EFCULL, "hlac_lt_backalley_g_65536_4y.dds"},
	{0xDE2A9E5F29C1D2C1ULL, "llac_lt_hangedman_n_-1_18.dds"},
	{0xDE2AA4DCFE91C347ULL, "vxht_bartrand_n_5c2_0_0.vlm"},
	{0xDE2B0E1C40194906ULL, "llac_lt_0_lowtown_a2_n_65535_8.dds"},
	{0xDE2B5DC0CBC24368ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrb_0d.dds"},
	{0xDE2C1A35485519F5ULL, "vxlt_0_lowtown_docks_b7r_0_0.vlm"},
	{0xDE2C1BF0A784BEE6ULL, "lgt110_gxa_coa_spy6117462.cl"},
	{0xDE2C3BD076C1F2DAULL, "48909358.ogg"},
	{0xDE2C49D4CACD75ABULL, "lgt310_mag_melson6152212.cl"},
	{0xDE2C95CB8942FE9DULL, "art\\vfx\\creatures\\fxc_shale_lit_l_c.dura.evt"},
	{0xDE2CDCAD5DA72EF2ULL, "narrative_varric_demo.cnv"},
	{0xDE2D3D87C65DD6B5ULL, "hlac_w_blightlands_327680_0x.dds"},
	{0xDE2D4F1ECDE0B12FULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.dura.evt"},
	{0xDE2D55756A441D5EULL, "vxlt_gallowscourtyard_n_6xo_0_0.vlm"},
	{0xDE2D7FA3EAB1CFD9ULL, "vxht_mansion2_1sl_0_0.vlm"},
	{0xDE2D81FB0EFA8080ULL, "6053008_m.fxe"},
	{0xDE2F463509B60AABULL, "hlac_lt_gallowscourtyard_n_-131074_0y.dds"},
	{0xDE2F6B56D745D006ULL, "dre000_dougal6130981.cl"},
	{0xDE2FB2E037F52FCAULL, "vxx_narrativeroom_3l_0_0.vlm"},
	{0xDE2FFEAD5DA99E4AULL, "narrative_varric_demo.bnk"},
	{0xDE30334D65213F38ULL, "718112471.ogg"},
	{0xDE30FD47E68B1FFCULL, "vxht_mansion2b_3cc_0_0.vlm"},
	{0xDE312B6C8D2E9254ULL, "959370188.ogg"},
	{0xDE31674771841995ULL, "vxw_encounter2_jo_0_0.vlm"},
	{0xDE31851CD9AECA5CULL, "762473139.ogg"},
	{0xDE3355F0C4B88861ULL, "gencr_slaver_ws.utc"},
	{0xDE338FD9458FD409ULL, "vxht_mansion2_3gf_0_0.vlm"},
	{0xDE33CA7D41D68CD3ULL, "397069997.ogg"},
	{0xDE341502A05D3CEBULL, "follower_banter6141677.cl"},
	{0xDE34501ECDE6D1D8ULL, "art\\vfx\\combat\\vfx_m_ent_entcloud_aoe_p.dura.gad"},
	{0xDE346CB24BCF5E89ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb101a_0t.dds"},
	{0xDE3495D2FE07EAA4ULL, "6105848_m.fxe"},
	{0xDE34A54A5480DBB9ULL, "hlac_lt_warehouse_-65536_1y.dds"},
	{0xDE34CF40C64CEA3EULL, "47337549.ogg"},
	{0xDE34DC6F1119F5A2ULL, "art\\characters\\playercharacter\\humanmale\\glv_medb.msh"},
	{0xDE34DE6F1119F958ULL, "art\\characters\\playercharacter\\humanmale\\glv_medb.mmh"},
	{0xDE34EA6F111A0DFBULL, "art\\characters\\playercharacter\\humanmale\\glv_medb.mao"},
	{0xDE3557A6CF0BF909ULL, "vxlt_gallowscourtyard_n_7ir_0_0.vlm"},
	{0xDE3579902CFA87F8ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar2.ani"},
	{0xDE36A185EEEA5533ULL, "6172320_m.fxe"},
	{0xDE36EA086C7DB71FULL, "llac_lt_0_lowtown_docks_a3_n_-65539_0.dds"},
	{0xDE3732F4D9ED919FULL, "6179387_m.fxe"},
	{0xDE3790C3828ACBF3ULL, "6027453_m.fxe"},
	{0xDE37F65F38E1AA57ULL, "89358585.ogg"},
	{0xDE38597E48642A92ULL, "hlac_lt_0_lowtown_docks_a3_-262144_3z.dds"},
	{0xDE38B9D4737CD776ULL, "610114391.ogg"},
	{0xDE392B39842A267AULL, "411331100.ogg"},
	{0xDE3937AB54597943ULL, "vxht_mansion2b_1qp_0_0.vlm"},
	{0xDE39DC10C624FEA0ULL, "gen_im_wep_mag_sta_act3_01.uti"},
	{0xDE3A676446104DC9ULL, "qun231cr_viscount_guard3.utc"},
	{0xDE3A9AF58374403CULL, "zz_fol_debug6056997.cl"},
	{0xDE3AE99851A659ABULL, "6036358.ogg"},
	{0xDE3B26EB52A6CF1CULL, "ser210_witness6168149.cl"},
	{0xDE3B5A8EE0F26320ULL, "6064154_m.fxe"},
	{0xDE3B61914ACD76E1ULL, "zz_start_dre.utp"},
	{0xDE3BE82F6D18B95CULL, "lyriumghost_bloodfrenzy.dds"},
	{0xDE3C6FA09B51EB66ULL, "gen_im_trs_dal_hvy_cht_r1.uti"},
	{0xDE3CF30103A47AB4ULL, "zz_vault_debug6177608.cl"},
	{0xDE3E838D504F9813ULL, "320733508.ogg"},
	{0xDE3F361D49F0CB9AULL, "vxw_bonepitmining_a2_nb_0_0.vlm"},
	{0xDE3F38D3899334CCULL, "766977330.ogg"},
	{0xDE3F6D2B620132D8ULL, "6044619_m.fxe"},
	{0xDE3FA7DE8ABE04F0ULL, "ico_belt_4.dds"},
	{0xDE3FC39C0BE069A9ULL, "35031043.ogg"},
	{0xDE3FD1617E640721ULL, "hlac_ht_chantry_2534_48z.dds"},
	{0xDE3FD932326C9C06ULL, "zz_one_debug6099933.cl"},
	{0xDE3FE3BFEE985D5BULL, "6034817_m.fxe"},
	{0xDE3FE56F11240CE3ULL, "art\\characters\\playercharacter\\humanmale\\glv_medb.phy"},
	{0xDE401011DC70E60DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_05bonetrack.mmh"},
	{0xDE401411DC70ED5EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_05bonetrack.mao"},
	{0xDE401C72F2F15220ULL, "6085436_m.fxe"},
	{0xDE402211DC710493ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_05bonetrack.msh"},
	{0xDE405DF245116A2CULL, "6172846_m.fxe"},
	{0xDE413FFE2B6542ABULL, "vxlt_darktown_535_0_0.vlm"},
	{0xDE415C5FB581B5E8ULL, "366699274.ogg"},
	{0xDE41C7733BBA322BULL, "hlac_lt_darktown_-1_34x.dds"},
	{0xDE41C9968D95E88EULL, "288211527.ogg"},
	{0xDE41CB998B5CBEA0ULL, "vxlt_0_lowtown_docks_n_b5r_0_0.vlm"},
	{0xDE41E8330ED01883ULL, "vxht_keep_6ak_0_0.vlm"},
	{0xDE421C6193F4BDBCULL, "character_character_p63.psh.fxo.dx11.assembly"},
	{0xDE42613EB4DFBD16ULL, "llac_lt_gallowstemplar_f_7198_6.dds"},
	{0xDE428514BB04C5D3ULL, "w_sundermount_58_0.rml"},
	{0xDE432C101B194668ULL, "266488284.ogg"},
	{0xDE433B7EE65DC3E9ULL, "205554783.ogg"},
	{0xDE43453B8BF04940ULL, "dae100cr_solivitus.utc"},
	{0xDE43B42F0BD247EEULL, "6051251_m.fxe"},
	{0xDE44B64C68E91434ULL, "108525313.ogg"},
	{0xDE4526737D1A13E9ULL, "vxlt_gallowstemplar_533_0_0.vlm"},
	{0xDE459433DF8262D2ULL, "6114008_m.fxe"},
	{0xDE45B9D078EA1BD5ULL, "store_lgt_epic1_3.utm"},
	{0xDE45C075A2610FCAULL, "vxlt_gallowscourtyard_6xt_0_0.vlm"},
	{0xDE46C4B6167FF62BULL, "6135659_m.fxe"},
	{0xDE46E22C0FC5743DULL, "w_bonepitdestroyed_a3_17_0_ll.mmh"},
	{0xDE4718C6CA4DF229ULL, "6102418_f.fxe"},
	{0xDE472B848FF9EFD0ULL, "hlac_w_encounter2_393214_0z.dds"},
	{0xDE47A988464F91C3ULL, "vxht_mansion2b_1ts_0_0.vlm"},
	{0xDE493DA58DD15A46ULL, "vxlt_darktown_4f6_0_0.vlm"},
	{0xDE497133101A084CULL, "6075978_m.fxe"},
	{0xDE498C14BB0AF02BULL, "w_sundermount_58_0.phy"},
	{0xDE4AC1A457E37E78ULL, "454473052.ogg"},
	{0xDE4AF73B35FDF1FCULL, "lt_0_lowtown_docks_n_143_0.mao"},
	{0xDE4B033B35FE059FULL, "lt_0_lowtown_docks_n_143_0.mmh"},
	{0xDE4B093B35FE1041ULL, "lt_0_lowtown_docks_n_143_0.msh"},
	{0xDE4B4FC3C0B78E18ULL, "gen00flst_fenris_gallows.stg"},
	{0xDE4BE50AE716B19DULL, "42288069.ogg"},
	{0xDE4CF081BA9EDC3CULL, "dae000_bartrand6064058.cl"},
	{0xDE4D7165EE1DEB4DULL, "lt_0_lowtown_docks_n_mmap_ph.dds"},
	{0xDE4DDCB9D01A569FULL, "vxht_bartrand_n_67i_0_0.vlm"},
	{0xDE4DF52221F371ABULL, "vxw_woundedcoast_1gv_0_0.vlm"},
	{0xDE4E02DDA693696FULL, "6088119_m.fxe"},
	{0xDE4E2085B3153C12ULL, "1068055195.ogg"},
	{0xDE4E4B0155704A95ULL, "vxw_encounter2_pm_0_0.vlm"},
	{0xDE4ECFCF370782ACULL, "6175946_m.fxe"},
	{0xDE4EF7AD4BA35BB1ULL, "163018509.ogg"},
	{0xDE4EF8FEACA03150ULL, "6041520_m.fxe"},
	{0xDE4F53AEAA2E7559ULL, "hlac_lt_0_lowtown_a2_-3_2x.dds"},
	{0xDE4FF3073DE05F5AULL, "6136376_m.fxe"},
	{0xDE505A1000C63103ULL, "vxht_keep_6jj_0_0.vlm"},
	{0xDE508635C3C975BFULL, "86189357.ogg"},
	{0xDE50931BEB6CE4ECULL, "6135731_m.fxe"},
	{0xDE509F9582B6A87CULL, "6126472_m.fxe"},
	{0xDE50D9445A994F80ULL, "dae000_merchant_armor_hm6164271.cl"},
	{0xDE510C89E2FD109AULL, "w_sundermount_153_0.mmh"},
	{0xDE511089E2FD17D1ULL, "w_sundermount_153_0.mao"},
	{0xDE511E89E2FD2FE0ULL, "w_sundermount_153_0.msh"},
	{0xDE51D106978DD8A6ULL, "20445052.ogg"},
	{0xDE51F391C5ACF051ULL, "vxlt_0_lowtown_a2_n_c_hng_0_0.vlm"},
	{0xDE52093A52575B34ULL, "vxlt_0_lowtown_a2_n_ekg_0_1.vlm"},
	{0xDE5213F43E9BA978ULL, "plt_cod_crt_sloth_demon.plo"},
	{0xDE5217464D9CB564ULL, "6069979_m.fxe"},
	{0xDE521C61B12603EDULL, "vxlt_0_lowtown_a2_gn5_0_0.vlm"},
	{0xDE5232635B450462ULL, "hlac_ht_0_hightown_n_65537_0x.dds"},
	{0xDE524ADF54F2E060ULL, "70116087.ogg"},
	{0xDE531E7A9CCF28D8ULL, "w_encounter3_68_0.phy"},
	{0xDE537F5FAA50CE31ULL, "fxc_puke_vol2.dds"},
	{0xDE53E998E33F6CE5ULL, "vxht_0_hightown_a2_auv_0_0.vlm"},
	{0xDE540610D1787B52ULL, "6113108_m.fxe"},
	{0xDE54286FF250CAECULL, "follower_banter6163449.cl"},
	{0xDE5473B83EBD3B65ULL, "vxw_encounter2_o0_0_0.vlm"},
	{0xDE54C89CFA446498ULL, "ave221_brennan6110475.cl"},
	{0xDE553604787E630DULL, "terrain3.vsh.fxo"},
	{0xDE555466339BCA94ULL, "6096490_m.fxe"},
	{0xDE559B7D3DA87003ULL, "vxht_0_hightown_n_byt_0_0.vlm"},
	{0xDE55F3096465EC13ULL, "vxlt_0_lowtown_n_f54_0_0.vlm"},
	{0xDE5655BACB481E4FULL, "6095495_f.fxe"},
	{0xDE565BC1114639ADULL, "335938806.ogg"},
	{0xDE56A1E7EB7E268BULL, "follower_banter6173764.cl"},
	{0xDE56B28754561685ULL, "vxca_cave_2ck_0_0.vlm"},
	{0xDE56B9644C5FFFF5ULL, "docks_male_elven_worker_sta6176095.cl"},
	{0xDE56C345B01CEEEAULL, "ht_0_hightownexplosion_n_33_0.rml"},
	{0xDE56C7D978576609ULL, "6179768_f.fxe"},
	{0xDE573FDAA989A178ULL, "6154266_m.fxe"},
	{0xDE57744E0F28FACCULL, "hlac_lt_0_lowtown_a3_n_c_-4_0z.dds"},
	{0xDE57AE052F43481CULL, "6065654_m.fxe"},
	{0xDE5821D61B5625F3ULL, "vxlt_0_lowtown_docks_c48_0_0.vlm"},
	{0xDE58E2514E2886ACULL, "gen00fl_dog6109322.cl"},
	{0xDE590117E784DBC8ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_slaterock_n.dds"},
	{0xDE59152F0A69C743ULL, "6078097_m.fxe"},
	{0xDE597789BC8F6BBCULL, "6119597_m.fxe"},
	{0xDE5984A3C9643E7DULL, "442171974.ogg"},
	{0xDE5A54714F54A8D0ULL, "vxdr_primevaltomb_5c9_0_1.vlm"},
	{0xDE5AAA0F06856F26ULL, "443956492.ogg"},
	{0xDE5B518EAB89C467ULL, "em_hireling_2.mor"},
	{0xDE5B74672A55DE6EULL, "243669602.ogg"},
	{0xDE5C0A1EC4CAC812ULL, "gen00fl_merrill6130095.cl"},
	{0xDE5C15DFD84C4FCCULL, "6169828_m.fxe"},
	{0xDE5C9990F76A7554ULL, "pro000_athenril6048731.cl"},
	{0xDE5CB66DF437A26BULL, "vxht_brothel_n_5f9_0_0.vlm"},
	{0xDE5CEB01AFBEC405ULL, "hlac_lt_0_lowtown_docks_n_-2_3x.dds"},
	{0xDE5D399825FCA6AFULL, "vxht_mansion_n_3fd_0_0.vlm"},
	{0xDE5D40033DE40E47ULL, "507298187.ogg"},
	{0xDE5D90885F09EF32ULL, "336261140.ogg"},
	{0xDE5D9FC42A4B78E7ULL, "72954970.ogg"},
	{0xDE5DFEC6830E4F7CULL, "mage_hexvunerable.dds"},
	{0xDE5E229D66DB37EAULL, "lt_gallowscourtyard_43_0_ll.mmh"},
	{0xDE5E2CA2E671A1D6ULL, "dae000_cora6162436.cl"},
	{0xDE5E7F6790C9E6AAULL, "362722017.ogg"},
	{0xDE5EE9D0B31707F3ULL, "vxht_keep_n_5k5_0_0.vlm"},
	{0xDE5F151A3F446B22ULL, "6090407_m.fxe"},
	{0xDE5F6B2F0B26F282ULL, "vxlt_0_lowtown_h9h_0_0.vlm"},
	{0xDE6034467CA99FCAULL, "829191798.ogg"},
	{0xDE6078C10692856EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masa_0t.dds"},
	{0xDE60A5614CA8508BULL, "6128926_f.fxe"},
	{0xDE60A64D0EC98A48ULL, "art\\levels\\races\\proxy\\props\\_textures\\bigboat_sa.dds"},
	{0xDE60C0B3A9A2B58AULL, "107986463.ogg"},
	{0xDE60EACBC108BA4BULL, "art\\vfx\\combat\\vfx_glb_ele_n_extreme_c.impa.evt"},
	{0xDE6116D7B5AE03EAULL, "qun221_petrice6185482.cl"},
	{0xDE61310753B63AB8ULL, "hlac_ca_caverns1_24423_2z.dds"},
	{0xDE6134EB71B0939EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_other.ani"},
	{0xDE619295BB396CB6ULL, "851579951.ogg"},
	{0xDE61D3CB8CDF559DULL, "522973351.ogg"},
	{0xDE6200DE70FE8F31ULL, "vxht_0_hightown_n_c9s_0_0.vlm"},
	{0xDE621424B903FF71ULL, "lt_0_lowtown_a2_38_0_ll.mmh"},
	{0xDE625B75D5358565ULL, "vxht_0_hightown_a2_apu_0_0.vlm"},
	{0xDE62C578CBB54EEAULL, "dae000_dalish_apprentices6147562.cl"},
	{0xDE62DACF5EB6C509ULL, "gen00fl_anders6173772.cl"},
	{0xDE62DDFA0A641DFEULL, "vxlt_undercitydungeon_xo_0_0.vlm"},
	{0xDE62E232299F4858ULL, "vxlt_0_lowtown_n_er5_0_0.vlm"},
	{0xDE6325F9CE25E404ULL, "6080974_m.fxe"},
	{0xDE63CEC044FE69CEULL, "vxht_chantry_n_27o_0_0.vlm"},
	{0xDE63F433A918B0CAULL, "dae000ip_crowd_ff6_talk7.utp"},
	{0xDE640D5A2F9E8883ULL, "vxht_0_hightown_n_bdw_0_0.vlm"},
	{0xDE64560AAEE5ECF1ULL, "hlac_ht_mansion2_818_19y.dds"},
	{0xDE64DAB25D254F76ULL, "vxlt_0_lowtown_docks_n_b60_0_1.vlm"},
	{0xDE64ED3DB135492CULL, "629600775.ogg"},
	{0xDE65B53ED407B20BULL, "6148387_m.fxe"},
	{0xDE66215F71BD8B46ULL, "vxlt_darktown_4h8_0_0.vlm"},
	{0xDE66AF30F8D6F004ULL, "vxlt_undercityencounter_rh_0_0.vlm"},
	{0xDE66B7DAEA6DBB50ULL, "vxca_caverns1_ebv_0_0.vlm"},
	{0xDE66E44077B14101ULL, "6146909_m.fxe"},
	{0xDE67460FA180D952ULL, "55872920.ogg"},
	{0xDE674C6E98ACA449ULL, "vxw_encounter2_camp_qv_0_0.vlm"},
	{0xDE6803CBC10F030CULL, "art\\vfx\\combat\\vfx_glb_ele_n_extreme_c.impa.gad"},
	{0xDE6833BF8E5210E1ULL, "clf260_harley6069580.cl"},
	{0xDE687C73DFB08B18ULL, "6173041_m.fxe"},
	{0xDE69438CD01CDFE9ULL, "one300_zevran6076497.cl"},
	{0xDE69829E52B03D95ULL, "1059099909.ogg"},
	{0xDE6A46C62E2D3E9EULL, "gen00fl_bethany6028804.cl"},
	{0xDE6A67D16D7DFBCCULL, "w_encounter2_31_0.msh"},
	{0xDE6A6DD16D7E05BEULL, "w_encounter2_31_0.mmh"},
	{0xDE6A79D16D7E1A55ULL, "w_encounter2_31_0.mao"},
	{0xDE6AFBBEAF0E8916ULL, "580421518.ogg"},
	{0xDE6B47914AF66E53ULL, "zz_start_dre.cnv"},
	{0xDE6B67944DAC0D41ULL, "mcr310_sibling6076290.cl"},
	{0xDE6B9187E825E66AULL, "618898279.ogg"},
	{0xDE6B91B14C128F0CULL, "spt_leafcards.psh.fxo"},
	{0xDE6C1AF3B0273556ULL, "178104070.ogg"},
	{0xDE6C44941732A514ULL, "mag131_taraneh6088725.cl"},
	{0xDE6C9309C57B3F91ULL, "6202074_m.fxe"},
	{0xDE6CF88C0A1BA65DULL, "vxca_cave_19b_0_0.vlm"},
	{0xDE6D785DD1D633C2ULL, "mag320cr_crowd_1.utc"},
	{0xDE6D834F66A2D97CULL, "6128672_m.fxe"},
	{0xDE6D9149122B3C84ULL, "611040170.ogg"},
	{0xDE6DE6E59CF2B10EULL, "hlac_lt_0_lowtown_a3_n_c_131070_5x.dds"},
	{0xDE6E02CBC1136D0AULL, "art\\vfx\\combat\\vfx_glb_ele_n_extreme_c.impa.ani"},
	{0xDE6E15E482C39BA5ULL, "vxlt_gallowscourtyard_72h_0_0.vlm"},
	{0xDE6ED74BA9992151ULL, "vxlt_0_lowtown_a2_n_c_hde_0_0.vlm"},
	{0xDE6EECF436438C34ULL, "vxlt_0_lowtown_a2_n_eqe_0_1.vlm"},
	{0xDE6F2A70FD27F735ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgte_0s.dds"},
	{0xDE6F4CEB71BCFA6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_other.evt"},
	{0xDE708CD8A843BBDFULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_022a_0d.dds"},
	{0xDE70CD52C72B9DE5ULL, "vxht_0_hightown_a2_akt_0_0.vlm"},
	{0xDE70DCD9FDA756C6ULL, "dre000_hireling16038660.cl"},
	{0xDE70F17DE7894209ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_mod_c.mmh"},
	{0xDE7103E21628630FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_rack_sa.dds"},
	{0xDE71372D4A62FD15ULL, "llac_lt_0_lowtown_-3_0.dds"},
	{0xDE71459257A52A45ULL, "69401825.ogg"},
	{0xDE71540F1B9560D8ULL, "vxlt_0_lowtown_n_eo2_0_0.vlm"},
	{0xDE71B268E4791F08ULL, "6070130_f.fxe"},
	{0xDE7216E5014819DDULL, "vxca_caverns2prc_61k_0_0.vlm"},
	{0xDE727F372194A103ULL, "vxht_0_hightown_n_bov_0_0.vlm"},
	{0xDE72DC58A31D1BB8ULL, "6055438_f.fxe"},
	{0xDE73038ADCDDC9B3ULL, "vxlt_undercitydungeon_1j4_0_1.vlm"},
	{0xDE7320D0824E631CULL, "vxlt_gallowstemplar_n_51k_0_0.vlm"},
	{0xDE73485735530173ULL, "vxdr_slavers_1j4_0_0.vlm"},
	{0xDE735057869820C6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\ready.ani"},
	{0xDE73B3B3BCBD2F14ULL, "6136125_m.fxe"},
	{0xDE73C0B25C2FEBC6ULL, "lt_0_lowtown_a2_27_0_lc.mmh"},
	{0xDE73C341C6787508ULL, "605183214.ogg"},
	{0xDE7407DBF36F8668ULL, "6136815_m.fxe"},
	{0xDE742147AB84D5FFULL, "dae000_arishok6068504.cl"},
	{0xDE74271BC5FDCA8BULL, "6145087_m.fxe"},
	{0xDE7481AFAD3EDA36ULL, "vxht_0_hightown_ap6_0_0.vlm"},
	{0xDE754DEB71C16718ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_left_hand_gesture_other.gad"},
	{0xDE7566A2D0368549ULL, "ht_0_hightown_a2_12_0_ll.mmh"},
	{0xDE75721728B41AFAULL, "vxlt_gallowstemplar_4ug_0_0.vlm"},
	{0xDE75B02D776F9F1DULL, "lowtown_male_lowlife1_sta6164957.cl"},
	{0xDE75BE4B8AA2BCC9ULL, "vxw_encounter2_camp_lu_0_0.vlm"},
	{0xDE762B89E31C61A5ULL, "w_sundermount_153_0.phy"},
	{0xDE76999AA554B394ULL, "gen00fl_fenris6126362.cl"},
	{0xDE77C4F7D1719D3BULL, "lgt260_eau_tmp_dockworker.bnk"},
	{0xDE77DBB97AE5C048ULL, "hlac_w_blightlands_262147_0y.dds"},
	{0xDE7855B859B0D3E7ULL, "460186628.ogg"},
	{0xDE788039C89E6D8DULL, "6025109_m.fxe"},
	{0xDE7888DBC0E6B2ACULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_brickround01_sa.dds"},
	{0xDE78D4D56D810588ULL, "6136107_m.fxe"},
	{0xDE78D703BFF844CDULL, "vxlt_blackemporium_vt_0_0.vlm"},
	{0xDE79B6A01F754D38ULL, "886829432.ogg"},
	{0xDE7A07FA4D7DFD47ULL, "vxht_mansion2b_1mp_0_0.vlm"},
	{0xDE7A2BB8C4922743ULL, "6161647_m.fxe"},
	{0xDE7A822A5C2241D8ULL, "159276628.ogg"},
	{0xDE7AC6F7D173D6FBULL, "lgt260_eau_tmp_dockworker.cnv"},
	{0xDE7AE5E60841AB17ULL, "574108956.ogg"},
	{0xDE7B335FFFDE9C98ULL, "827515871.ogg"},
	{0xDE7B35723B2CEBFAULL, "vxht_0_hightown_cak_0_0.vlm"},
	{0xDE7B8D8AC2901C32ULL, "vxht_mansion2b_24u_0_0.vlm"},
	{0xDE7C292E3772EA57ULL, "mag101_alain6053008.cl"},
	{0xDE7CA28010241ECDULL, "6116724_m.fxe"},
	{0xDE7CCF367D26210DULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhangred.mmh"},
	{0xDE7CD3367D26285EULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhangred.mao"},
	{0xDE7CD4D20A2A337CULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st01a_0n.dds"},
	{0xDE7CE1367D263F93ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzhangred.msh"},
	{0xDE7CF1F22BE2FD08ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_chantrybasic.mmh"},
	{0xDE7CF3CF1B66D730ULL, "gen00fl_carver6097316.cl"},
	{0xDE7CFDF22BE311ABULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_chantrybasic.mao"},
	{0xDE7D0FF22BE33032ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_chantrybasic.msh"},
	{0xDE7D2EAD15CF594EULL, "hlac_lt_0_lowtown_a2_n_c_65537_2x.dds"},
	{0xDE7D9E93C3DE4CE3ULL, "165868494.ogg"},
	{0xDE7E6F4A5F067CAAULL, "w_blightlands_246_0.mao"},
	{0xDE7E734A5F0683F9ULL, "w_blightlands_246_0.mmh"},
	{0xDE7E7D4A5F0694E7ULL, "w_blightlands_246_0.msh"},
	{0xDE7EA88FD8CCF46FULL, "character_character_p86.psh.fxo.dx11"},
	{0xDE7F056D7460EF72ULL, "475543480.ogg"},
	{0xDE80613DB5606CC2ULL, "w_sundermount_59_0.rml"},
	{0xDE808DC07CAD0DE7ULL, "llac_lt_undercityserialkiller_-1_16.dds"},
	{0xDE80960EFC62243CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball3.mmh"},
	{0xDE80B1F13731B218ULL, "342642676.ogg"},
	{0xDE81385786A43537ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\ready.evt"},
	{0xDE81759A53750325ULL, "ss_redcliffe.dds"},
	{0xDE81966345D5885BULL, "6050183_m.fxe"},
	{0xDE81D9806A6C317FULL, "mag291_demon_marethari6166173.cl"},
	{0xDE820CCCFA4732BDULL, "hlac_w_encounter3_-196608_0y.dds"},
	{0xDE8226067E4F782AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_left_down_loop.gad"},
	{0xDE82CAF0D9E01F32ULL, "6209379_m.fxe"},
	{0xDE82CB5F0627350AULL, "725075767.ogg"},
	{0xDE82D030D26618EEULL, "vxht_0_hightown_bfw_0_0.vlm"},
	{0xDE82ED097276C2BFULL, "vxht_mansion_n_4a3_0_1.vlm"},
	{0xDE82FCE24B90BF36ULL, "vxca_caverns2_9f3_0_0.vlm"},
	{0xDE839B94CE59EC50ULL, "vxca_caverns1_eht_0_0.vlm"},
	{0xDE83E949E307C0F8ULL, "gen_im_npc_wep_two_2hsb.uti"},
	{0xDE83FB75D968EAA6ULL, "6085262_f.fxe"},
	{0xDE83FCF8D50B32E1ULL, "6065000_m.fxe"},
	{0xDE84539E2809A384ULL, "t3_cth_nb1.tnt"},
	{0xDE84877556AF803BULL, "w_bonepitdestroyed_a3_51_0.rml"},
	{0xDE84C6866745C087ULL, "6114834_m.fxe"},
	{0xDE84F9A5BA457F6AULL, "6034386_m.fxe"},
	{0xDE851787A9468896ULL, "6189745_m.fxe"},
	{0xDE852B6393B1CDEDULL, "70204152.ogg"},
	{0xDE852CFD1A1D097BULL, "486099519.ogg"},
	{0xDE855D988BD803D2ULL, "773379262.ogg"},
	{0xDE857C12CDC71795ULL, "45940434.ogg"},
	{0xDE8615EAA7832695ULL, "6103595_m.fxe"},
	{0xDE868BB815F54C52ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_basket_n.dds"},
	{0xDE871FF509821D6FULL, "97985283.ogg"},
	{0xDE87623DB5668D76ULL, "w_sundermount_59_0.phy"},
	{0xDE87998767745EA2ULL, "vxca_caverns1_mines_jdh_0_0.vlm"},
	{0xDE87CD7A4459320DULL, "714624607.ogg"},
	{0xDE87CF907462E225ULL, "vxlt_gallowstemplar_n_6qe_0_0.vlm"},
	{0xDE87E620879A1419ULL, "6156885_m.fxe"},
	{0xDE880A2799F841F2ULL, "mcr370_cut_attack_begins6082061.cl"},
	{0xDE8810AD0548AEE3ULL, "follower_banter6172767.cl"},
	{0xDE88395786AA55C0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\ready.gad"},
	{0xDE8865D1478CE115ULL, "vxw_blightlands_6d_0_0.vlm"},
	{0xDE8873DF63953DDCULL, "6120984_m.fxe"},
	{0xDE889D95B6883FC3ULL, "6166747_m.fxe"},
	{0xDE88BEC4F76A917BULL, "gen_im_arm_cht_elf_comb_t3.uti"},
	{0xDE89A8EBA8B31C22ULL, "123390396.ogg"},
	{0xDE89BA85EECFE4DDULL, "gen00fl_varric6021110.cl"},
	{0xDE89CF69E74A55EEULL, "hlac_lt_gallowstemplar_2391_6y.dds"},
	{0xDE8A056B356B15E5ULL, "zz_one_debug6099947.cl"},
	{0xDE8B2AA3118E848FULL, "hlac_lt_undercityencounter_-1_10x.dds"},
	{0xDE8B34E09A713F13ULL, "vxlt_warehouse_n_10n_0_0.vlm"},
	{0xDE8B5D8C1A444D84ULL, "hlac_lt_0_lowtown_n_0_0z.dds"},
	{0xDE8B6A510EFB608CULL, "201870813.ogg"},
	{0xDE8BE644BA5BE822ULL, "6137054_m.fxe"},
	{0xDE8C05E88C6C0EC2ULL, "6112045_f.fxe"},
	{0xDE8C73AAF4D4AEABULL, "6054222_m.fxe"},
	{0xDE8CDC598AB1A8D3ULL, "106709657.ogg"},
	{0xDE8CDEC036DD70B8ULL, "mag131_taraneh6088677.cl"},
	{0xDE8D5D0612ED2450ULL, "lt_0_lowtown_docks_a3_n_112_0.msh"},
	{0xDE8D6B0612ED3C6AULL, "lt_0_lowtown_docks_a3_n_112_0.mmh"},
	{0xDE8D6F0612ED42A1ULL, "lt_0_lowtown_docks_a3_n_112_0.mao"},
	{0xDE8DAC7A62FEBDB5ULL, "476645568.ogg"},
	{0xDE8DD592B53CCCCBULL, "rdr110dg_dockworker6015952.cl"},
	{0xDE8E01ECC20109E8ULL, "zz_ave_debug6175060.cl"},
	{0xDE8E609BA670ED46ULL, "6066763_m.fxe"},
	{0xDE8E7BA7290597C2ULL, "vxht_0_hightown_a2_bzc_0_0.vlm"},
	{0xDE8F110D29EFC744ULL, "llac_lt_0_lowtown_a2_n_c_-1_6.dds"},
	{0xDE8F288F1496914CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1ww_basic_2.mao"},
	{0xDE8F895EF0A1B0C8ULL, "llac_lt_0_lowtown_docks_-131074_1.dds"},
	{0xDE8FBA8300199EA0ULL, "hlac_lt_0_lowtown_a2_n_c_-1_4x.dds"},
	{0xDE8FC2E6D53748DAULL, "hlac_lt_undercityserialkiller_-1_27y.dds"},
	{0xDE8FC3197456DA31ULL, "llac_lt_0_lowtown_docks_a3_n_131071_6.dds"},
	{0xDE9001A43FE0CAB2ULL, "384131604.ogg"},
	{0xDE9021AB084FF956ULL, "hlac_dr_0_deeproads_5841_1y.dds"},
	{0xDE904B617AF8ABAFULL, "vxca_cave_1ls_0_0.vlm"},
	{0xDE906A914B15C5A8ULL, "zz_start_dre.ncs"},
	{0xDE90DD43C7FCFE72ULL, "6032202_f.fxe"},
	{0xDE914B3DB56EBA17ULL, "w_sundermount_59_0.msh"},
	{0xDE9155CE8A80F113ULL, "lt_0_lowtown_docks_12_0_lc.mmh"},
	{0xDE915D3DB56ED97AULL, "w_sundermount_59_0.mao"},
	{0xDE91613DB56EDF89ULL, "w_sundermount_59_0.mmh"},
	{0xDE9169880FAAF98BULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg_b.dura.evt"},
	{0xDE91706173A19AA5ULL, "vxht_0_hightown_a3_bu0_0_0.vlm"},
	{0xDE918CFA42E72D67ULL, "vxdr_primevaltomb_34h_0_0.vlm"},
	{0xDE91E732B24E0B4BULL, "hlac_w_encounter2_camp_131075_0z.dds"},
	{0xDE92BA4FE3E0745DULL, "6136304_m.fxe"},
	{0xDE92EFCCA8D3A746ULL, "6107199_f.fxe"},
	{0xDE92FB89AF7C4BA9ULL, "413484311.ogg"},
	{0xDE933863593BD907ULL, "6113934_m.fxe"},
	{0xDE954C337147B64EULL, "vxlt_0_lowtown_n_eki_0_1.vlm"},
	{0xDE9552DE46E7250DULL, "349290149.ogg"},
	{0xDE955386B5CB3374ULL, "mcr371cr_bethany_cut.utc"},
	{0xDE95AF5DAC856623ULL, "vxht_0_hightown_a3_c53_0_0.vlm"},
	{0xDE95D62A4555130CULL, "841414691.ogg"},
	{0xDE95EFD93BD14A9EULL, "llac_ca_caverns2_5802_9.dds"},
	{0xDE95FA8A49868E17ULL, "lla_ht_0_hightown_a2_65534_0.dds"},
	{0xDE95FB3825352F9BULL, "vxdr_primevaltomb_2vv_0_1.vlm"},
	{0xDE9601067E5FE210ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_left_down_loop.ani"},
	{0xDE9605120F4BD1C5ULL, "vxht_0_hightown_a3_aoy_0_0.vlm"},
	{0xDE960E1199071F61ULL, "vxht_mansion2_2fd_0_0.vlm"},
	{0xDE961D6C3F1561B0ULL, "526580262.ogg"},
	{0xDE965D60532CFBE9ULL, "dre000_varric6027771.cl"},
	{0xDE96C24C37BA34D4ULL, "vxca_caverns2prc_8q4_0_0.vlm"},
	{0xDE96FFB93379C8E4ULL, "art\\characters\\weapons\\dagger\\w_dgr_109.mao"},
	{0xDE971FF18F4CA4C6ULL, "lgt150_nat_were_supporter6110945.cl"},
	{0xDE973A9A3A9A7B34ULL, "face0_ds1y.psh.fxo"},
	{0xDE982F5413C1DD1BULL, "6216789_m.fxe"},
	{0xDE987140F20ADCEBULL, "6041408_m.fxe"},
	{0xDE9882880FB1434CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg_b.dura.gad"},
	{0xDE989FE2AC1A545AULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em03.phy"},
	{0xDE98EF605A936FFFULL, "vxca_cavern1_haunted_fn7_0_0.vlm"},
	{0xDE99357EB06CEF39ULL, "hlac_lt_0_lowtown_a3_n_c_1_7z.dds"},
	{0xDE995EBE591816C6ULL, "zz_fen_debug6215260.cl"},
	{0xDE9A2CE27F7B6AD1ULL, "vxlt_0_lowtown_a2_n_c_huz_0_0.vlm"},
	{0xDE9A6DC56F598141ULL, "hlac_w_sundermount_262145_0x.dds"},
	{0xDE9A6FCBEA78BE53ULL, "255465855.ogg"},
	{0xDE9A7317D0B072D4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3.phy"},
	{0xDE9A9422EBF87909ULL, "vxw_encounter3_gg_0_0.vlm"},
	{0xDE9AD517DF827A89ULL, "vrc221_varric6044188.cl"},
	{0xDE9AD6E8AD177AC3ULL, "420939600.ogg"},
	{0xDE9B4EB4935A8691ULL, "dre000_varric6035178.cl"},
	{0xDE9B52FF01E8C8EAULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.anb"},
	{0xDE9B65D064DB3D1DULL, "6165633_f.fxe"},
	{0xDE9C48199599B314ULL, "vxw_bonepitmining_a2_163_0_0.vlm"},
	{0xDE9C9DB694CBFBAEULL, "738454337.ogg"},
	{0xDE9CC1EF32512078ULL, "6108443_m.fxe"},
	{0xDE9CE06EA15D3113ULL, "424318535.ogg"},
	{0xDE9D591C6A5AECA9ULL, "vxlt_gallowstemplar_f_3ix_0_0.vlm"},
	{0xDE9DE11C9C524AD2ULL, "6103310_m.fxe"},
	{0xDE9E6EB1686AC6B5ULL, "gen00fl_sebastian6189909.cl"},
	{0xDE9F0FF73A0D3B5FULL, "zz_lgt_3_debug6140431.cl"},
	{0xDE9F4EEB8E615961ULL, "dae000_keep_noble_sta6161528.cl"},
	{0xDE9F7C8A5B27AD51ULL, "vxlt_hangedman_n_20k_0_0.vlm"},
	{0xDE9F81880FB7604AULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_brazfirelrg_b.dura.ani"},
	{0xDE9F8C68E812CD7FULL, "mag100_thrask6032060.cl"},
	{0xDE9F96C75A5BCB48ULL, "vxht_mansion2_2fa_0_0.vlm"},
	{0xDEA03EF246E86E16ULL, "182707676.ogg"},
	{0xDEA062EFBF6254E3ULL, "vxw_bonepitmining_a2_op_0_0.vlm"},
	{0xDEA069BB2D28AA45ULL, "vxlt_gallowstemplar_4ul_0_0.vlm"},
	{0xDEA102E8CC4888F3ULL, "art\\vfx\\cinematic\\vfx_anders_intro_c.impa.ani"},
	{0xDEA17029CC648226ULL, "6005038_m.fxe"},
	{0xDEA176E5345CFB9EULL, "67006735.ogg"},
	{0xDEA1AFDD5AF47DC7ULL, "728825821.ogg"},
	{0xDEA1E68D02535606ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.impa.evt"},
	{0xDEA25688877EB525ULL, "vxlt_0_lowtown_a2_n_hd7_0_0.vlm"},
	{0xDEA281873F1605A2ULL, "6126352_f.fxe"},
	{0xDEA2FF8A4A627766ULL, "6112910_m.fxe"},
	{0xDEA35F4E6A641DDDULL, "vxlt_undercitydungeon_15r_0_0.vlm"},
	{0xDEA36FD2021DEEA1ULL, "314312898.ogg"},
	{0xDEA3A471DBC394DAULL, "w_bonepitmining_a2_41_0.rml"},
	{0xDEA42273FD9A5E6CULL, "gen00fl_varric6109211.cl"},
	{0xDEA43580BAFD8C41ULL, "6131183_m.fxe"},
	{0xDEA4DA4B0E963939ULL, "vxca_cavern1_haunted_do0_0_0.vlm"},
	{0xDEA4DABA6E378A69ULL, "vxht_0_hightown_a2_bwc_0_0.vlm"},
	{0xDEA5480AC6534AAFULL, "vxlt_0_lowtown_eny_0_0.vlm"},
	{0xDEA56A17D0BA6A8FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3.mmh"},
	{0xDEA5750EFC810807ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball3.phy"},
	{0xDEA5DADF36FE77C1ULL, "zz_followers_banter6157766.cl"},
	{0xDEA61F66FAC9775CULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_shapes.mao"},
	{0xDEA63B686CF3A12CULL, "6099532_m.fxe"},
	{0xDEA6A455887B67B0ULL, "6122511_m.fxe"},
	{0xDEA6B81830840D7CULL, "140540666.ogg"},
	{0xDEA6BC7FB002261BULL, "176037858.ogg"},
	{0xDEA6C61D5A9E082BULL, "vxlt_0_lowtown_docks_a3_n_cyj_0_1.vlm"},
	{0xDEA6D3F72FB37DF6ULL, "vxlt_shop_6x_0_0.vlm"},
	{0xDEA71200CB4A1953ULL, "148074737.ogg"},
	{0xDEA7569FFCDD56E9ULL, "vxlt_gallowstemplar_2ol_0_0.vlm"},
	{0xDEA7A30E89A05B21ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\unequipped_weapons_away.ani"},
	{0xDEA7E0080CE20CA6ULL, "370076584.ogg"},
	{0xDEA8134983943758ULL, "6049708_m.fxe"},
	{0xDEA83D70D03F1693ULL, "hlac_ca_caverns1_mines_24424_23x.dds"},
	{0xDEA8421A5A556A61ULL, "vxht_keep_57x_0_0.vlm"},
	{0xDEA89EBF71718351ULL, "vxlt_0_lowtown_a2_n_c_hpy_0_0.vlm"},
	{0xDEA8FD8D02599C51ULL, "art\\vfx\\creatures\\fxc_mor_gui_c.impa.gad"},
	{0xDEA91CFEBC80B1F0ULL, "vxw_sundermount_23z_0_0.vlm"},
	{0xDEA92937C819D622ULL, "6095519_f.fxe"},
	{0xDEA9A89E6879DB29ULL, "vxlt_gallowstemplar_n_38s_0_0.vlm"},
	{0xDEAA1A9F57CAC420ULL, "vxlt_gallowstemplar_n_3e1_0_0.vlm"},
	{0xDEAB3071C1FA5069ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv2.ani"},
	{0xDEAB44D16DB55759ULL, "w_encounter2_31_0.phy"},
	{0xDEAB6909487C067CULL, "lt_hangedman.arl"},
	{0xDEABB1F4DB8F0A5BULL, "gen00fl_anders6125733.cl"},
	{0xDEAC8708B2D3DDF8ULL, "art\\characters\\creatures\\ashwraith\\animation\\blendtree\\ashwraith_movement.anb"},
	{0xDEACE543E686F932ULL, "qun200_party_comment6095380.cl"},
	{0xDEAD65ED5BD4BB42ULL, "360757640.ogg"},
	{0xDEADFAE8CC53052AULL, "art\\vfx\\cinematic\\vfx_anders_intro_c.impa.evt"},
	{0xDEAEAE0E89A68CC3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\unequipped_weapons_away.gad"},
	{0xDEAED4CCB1586D63ULL, "vxw_bonepitmining_a2_js_0_0.vlm"},
	{0xDEAEFE8D025E09FFULL, "art\\vfx\\creatures\\fxc_mor_gui_c.impa.ani"},
	{0xDEAF0B7A9D1D676DULL, "w_encounter3_68_0.msh"},
	{0xDEAF197A9D1D7EA0ULL, "w_encounter3_68_0.mao"},
	{0xDEAF1D7A9D1D85F3ULL, "w_encounter3_68_0.mmh"},
	{0xDEB075A92419B4DFULL, "vxlt_0_lowtown_docks_n_b6d_0_0.vlm"},
	{0xDEB07E05001D00D4ULL, "832576261.ogg"},
	{0xDEB080FC1A870311ULL, "zz_dae_debug6029887.cl"},
	{0xDEB0F51F937DE70DULL, "mer320_hubert6051890.cl"},
	{0xDEB14DD16DB9D275ULL, "w_encounter2_31_0.rml"},
	{0xDEB18626DD9ABA7AULL, "andev_item_acquired.ncs"},
	{0xDEB1CD3B92FFA989ULL, "l_sib131ar_family_estate.wbd"},
	{0xDEB1D12B5C5A365DULL, "vxlt_undercitydungeon_18q_0_0.vlm"},
	{0xDEB203AE6ADC2FD2ULL, "ht_0_hightown_24_0.rml"},
	{0xDEB223EDA730D043ULL, "follower_banter6168030.cl"},
	{0xDEB244E088D81470ULL, "gen_im_arm_glv_hvy_start.uti"},
	{0xDEB24B71C2009CEBULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv2.gad"},
	{0xDEB295106D6C5601ULL, "276928283.ogg"},
	{0xDEB3417EC6FE85C2ULL, "hlac_lt_0_lowtown_a2_n_c_65534_6y.dds"},
	{0xDEB34C97602DA2E9ULL, "vxht_0_hightown_a2_bxb_0_0.vlm"},
	{0xDEB350EA39900D49ULL, "atl_load_town_dxt5_dat.xml"},
	{0xDEB364F4B552758BULL, "gen00fl_merrill6106527.cl"},
	{0xDEB39D442AEDB496ULL, "19568186.ogg"},
	{0xDEB3B9E7B849632FULL, "vxlt_0_lowtown_eix_0_0.vlm"},
	{0xDEB434588D54CB0AULL, "lgt160_ntb_elf_conflict.cnv"},
	{0xDEB4D3ECD9C66B11ULL, "war_righteousstrike.dds"},
	{0xDEB4D9E8CC58EB9DULL, "art\\vfx\\cinematic\\vfx_anders_intro_c.impa.gad"},
	{0xDEB5236952ACE7AAULL, "mrl150cr_demonstandin.utc"},
	{0xDEB607B610B6F2DAULL, "6062763_m.fxe"},
	{0xDEB693AD2C6F7A4AULL, "6142476_m.fxe"},
	{0xDEB6D0491B4A7D28ULL, "lt_0_lowtown_25_0_lc.mmh"},
	{0xDEB6F7F891FDFA96ULL, "vxht_bartrand_n_68u_0_0.vlm"},
	{0xDEB70ADA96AFF052ULL, "fx_glyph_paral.dds"},
	{0xDEB750EB0CC38E5CULL, "lgt210_eau_symp_templar6144740.cl"},
	{0xDEB755CC4A92E553ULL, "vxlt_0_lowtown_n_gco_0_0.vlm"},
	{0xDEB778793D11596CULL, "190225650.ogg"},
	{0xDEB788F6FA1B6227ULL, "propalpha1.psh.fxo"},
	{0xDEB798E76B01DB5CULL, "6131155_m.fxe"},
	{0xDEB7DEAC62F14DFEULL, "513662063.ogg"},
	{0xDEB808C219914768ULL, "6017762_f.fxe"},
	{0xDEB816588D588132ULL, "lgt160_ntb_elf_conflict.bnk"},
	{0xDEB8341D2DC74572ULL, "121510073.ogg"},
	{0xDEB9097E09DF023EULL, "hlac_lt_0_lowtown_a2_0_6y.dds"},
	{0xDEB9104096A8756FULL, "6174195_m.fxe"},
	{0xDEB93F076FD47364ULL, "632970816.ogg"},
	{0xDEB94871C206B650ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\backleftlegattackv2.evt"},
	{0xDEB9D13DFEE7E142ULL, "vxht_0_hightown_a2_ben_0_0.vlm"},
	{0xDEB9E41F06BC1B65ULL, "vxw_blightlands_64_0_0.vlm"},
	{0xDEBA1A0932BB559FULL, "107803981.ogg"},
	{0xDEBA31927A3CFBD6ULL, "qun160_vashothlead6053922.cl"},
	{0xDEBAA41589C01210ULL, "vxht_0_hightown_by4_0_0.vlm"},
	{0xDEBBDAA3325DD8B6ULL, "gen00fl_fenris6126359.cl"},
	{0xDEBBE602F7DAE11EULL, "sst_ss_circlemages_m_1.gda"},
	{0xDEBBF1E8BC578C47ULL, "627095518.ogg"},
	{0xDEBC601736530DE6ULL, "vxlt_gallowstemplar_n_3ga_0_0.vlm"},
	{0xDEBD6E91FFF595D1ULL, "206180903.ogg"},
	{0xDEBDB93319C50353ULL, "lgt110_gxa_vgk_silver_order6117035.cl"},
	{0xDEBDFE03ABF719DEULL, "6084468_m.fxe"},
	{0xDEBE6A7EF4AB2A68ULL, "vxdr_primevalentrance_2gc_0_0.vlm"},
	{0xDEBE7F5C5D6858DCULL, "6079017_m.fxe"},
	{0xDEBE820BCB9F7569ULL, "vxca_caverns2_8p3_0_0.vlm"},
	{0xDEBEA6412389890DULL, "lt_0_lowtown_docks_119_0.mmh"},
	{0xDEBEAA412389905EULL, "lt_0_lowtown_docks_119_0.mao"},
	{0xDEBEB8412389A793ULL, "lt_0_lowtown_docks_119_0.msh"},
	{0xDEBF0E3F5B5153F5ULL, "llac_lt_backalley_n_0_15.dds"},
	{0xDEBF78E358AC0277ULL, "art\\vfx\\environment\\vfx_bl_smoke_vista_s_p.mmh"},
	{0xDEC05C08F3B4FD4DULL, "568349453.ogg"},
	{0xDEC05D0B7E5BDECBULL, "hlac_lt_0_lowtown_a2_n_c_-2_2x.dds"},
	{0xDEC07F5AC9120866ULL, "llac_lt_0_lowtown_n_-262144_0.dds"},
	{0xDEC0A1A7889B3C7AULL, "gen00fl_anders6125257.cl"},
	{0xDEC0D6820BE783D2ULL, "vxlt_0_lowtown_n_gcj_0_0.vlm"},
	{0xDEC0D6946A751BC4ULL, "523591337.ogg"},
	{0xDEC1125315F83007ULL, "6076477_m.fxe"},
	{0xDEC1249A307959E4ULL, "6172994_m.fxe"},
	{0xDEC168803957E8CEULL, "607589234.ogg"},
	{0xDEC18AB94FE5A622ULL, "6109697_m.fxe"},
	{0xDEC1D3EFA043F3EAULL, "649970513.ogg"},
	{0xDEC2B22C33A487DCULL, "6168027_m.fxe"},
	{0xDEC2E599DE3D40F9ULL, "963796418.ogg"},
	{0xDEC31C0FA055B7D8ULL, "mag110cr_sailor_3.utc"},
	{0xDEC31F2250DFD22CULL, "6093732_m.fxe"},
	{0xDEC41F6DC7409E00ULL, "vxca_caverns2prc_9hn_0_0.vlm"},
	{0xDEC455FF020BD2DDULL, "art\\vfx\\combat\\vfx_r_dwp_bleed_loop_c.mmh"},
	{0xDEC46398C26A9FB2ULL, "and100_anders6095835.cl"},
	{0xDEC47C9D563DFE4DULL, "hlac_lt_foundry_n_320_21y.dds"},
	{0xDEC5908FA960F44EULL, "395026384.ogg"},
	{0xDEC5A15AC12A17C8ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\hf_hed_mera_0n.dds"},
	{0xDEC5B129D7BC2A8FULL, "477174038.ogg"},
	{0xDEC5E06F74B98C3BULL, "6135989_m.fxe"},
	{0xDEC5E60AE3D53A1FULL, "6128742_f.fxe"},
	{0xDEC60BEB0AA3BF28ULL, "106452651.ogg"},
	{0xDEC673ED0ED484E1ULL, "vxht_0_hightown_a3_ak9_0_0.vlm"},
	{0xDEC6AA7B9FEF2594ULL, "6100291_f.fxe"},
	{0xDEC6BAFA6C65E724ULL, "6095842_m.fxe"},
	{0xDEC7DC794589FD9FULL, "vxht_0_hightown_a3_b85_0_0.vlm"},
	{0xDEC7E80A4869135BULL, "mrl310_party_comments6185817.cl"},
	{0xDEC88481876CF1CCULL, "lgt220_shp_mag_guard6149900.cl"},
	{0xDEC959A2FC18DD8CULL, "gen00fl_carver6097424.cl"},
	{0xDEC962E27E7541B8ULL, "dae000ip_crowd_lm2_talk2.utp"},
	{0xDEC97140DD2C3602ULL, "vxht_chantry_n_2co_0_0.vlm"},
	{0xDECA61BE6748B8B0ULL, "vxlt_warehouse_n_15f_0_0.vlm"},
	{0xDECB0703F63F0A29ULL, "zz_one_debug6074007.cl"},
	{0xDECB2C7926EDE815ULL, "97650282.ogg"},
	{0xDECB302835429D78ULL, "hlac_ht_mansion_n_3373_1y.dds"},
	{0xDECB48400B7252A9ULL, "6163229_m.fxe"},
	{0xDECB5BA19301A8FFULL, "hlac_lt_backalley_g_1_4x.dds"},
	{0xDECB6A68224A8985ULL, "6117184_f.fxe"},
	{0xDECB97FD403C31DDULL, "347792168.ogg"},
	{0xDECBCA2881FFA69DULL, "dae000ip_crowd_em3_talk16.utp"},
	{0xDECBD44D2097ABBBULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_deadmeredith_e.dds"},
	{0xDECBFF2E9458CF82ULL, "vxdr_0_deeproads_cqn_0_2.vlm"},
	{0xDECC036B8E075715ULL, "6065941_m.fxe"},
	{0xDECC077249A171D8ULL, "hlac_lt_0_lowtown_a2_n_-131071_0z.dds"},
	{0xDECDA476AA840609ULL, "950131750.ogg"},
	{0xDECDBE09442CC1F3ULL, "vxlt_undercityencounter_1bu_0_1.vlm"},
	{0xDECEB61E3D05A5C6ULL, "heallight.mao"},
	{0xDECECF42D1BA0570ULL, "weapon_weapon_p5.psh.fxo.dx11.assembly"},
	{0xDECED9DA96C42ED0ULL, "fx_glyph_paral.mao"},
	{0xDECF9F412397F0CAULL, "lt_0_lowtown_docks_119_0.phy"},
	{0xDECFD5971D4DFC7DULL, "vxlt_0_lowtown_a2_n_fh5_0_0.vlm"},
	{0xDECFFC9641DBBEA2ULL, "6102797_m.fxe"},
	{0xDED0AFE23DE14F7DULL, "gen_im_trs_qun_sta_cld_r0.uti"},
	{0xDED0E604EACA0186ULL, "t3_eey_am1.tnt"},
	{0xDED14F3C03F8928EULL, "112745787.ogg"},
	{0xDED1BD41163F75B9ULL, "gen_im_wep_war_shd_act0_03.uti"},
	{0xDED1E58298B64B5EULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_lyriumtile_n.dds"},
	{0xDED1EA704FDDC3D8ULL, "6051767_m.fxe"},
	{0xDED1F3E7F4BA0B6EULL, "hlac_lt_gallowstemplar_f_7198_7y.dds"},
	{0xDED214B333FEBFD6ULL, "t3_wpn_slr.tnt"},
	{0xDED2F34EB1D1D22CULL, "lowtown_guard_chase6169517.cl"},
	{0xDED3022788E9D506ULL, "6106703_f.fxe"},
	{0xDED3043CF54EDC24ULL, "lt_0_lowtown_a2_32_0_lc.mmh"},
	{0xDED3205F5EA8A9B7ULL, "vxca_caverns2_853_0_1.vlm"},
	{0xDED3DFC5B07DFF96ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_3z.dds"},
	{0xDED41697F3E47CACULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_3z.dds"},
	{0xDED423438C92BBB2ULL, "6094502_m.fxe"},
	{0xDED47D4E2CDABB14ULL, "gen_im_wep_mag_sta_act2_06.uti"},
	{0xDED47D60EC9E55A3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_left_down.ani"},
	{0xDED4A43F5D799348ULL, "hlac_ca_caverns1_24426_6y.dds"},
	{0xDED4E5CA00CA9D61ULL, "vxht_0_hightown_a3_ap8_0_0.vlm"},
	{0xDED4F99A12CA1082ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackleft.gad"},
	{0xDED4FAE316DFEF8CULL, "593287590.ogg"},
	{0xDED576F5C92D169DULL, "plt_seb000pt_gift_bow.plo"},
	{0xDED59914F0B708E9ULL, "vxlt_gallowstemplar_5a9_0_0.vlm"},
	{0xDED62868ED5B497DULL, "6163618_m.fxe"},
	{0xDED62D1552CD8822ULL, "6090216_m.fxe"},
	{0xDED6415B11A582D5ULL, "vxca_caverns2prc_3i2_0_0.vlm"},
	{0xDED64BFF582D7B79ULL, "vxlt_0_lowtown_docks_8uf_0_0.vlm"},
	{0xDED64E563780161FULL, "vxht_0_hightown_a3_b74_0_0.vlm"},
	{0xDED659F1040373F9ULL, "vxlt_0_lowtown_docks_a3_cti_0_0.vlm"},
	{0xDED66CC6F76447FEULL, "hlac_lt_undercityencounter_-1_47y.dds"},
	{0xDED68C9A287927C4ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.anb"},
	{0xDED69F3B931E786BULL, "l_sib131ar_family_estate.bnk"},
	{0xDED7363837BFE8AAULL, "zz_followers_banter6159471.cl"},
	{0xDED780A7CA607BF7ULL, "vxlt_gallowsprison_n_6nj_0_0.vlm"},
	{0xDED7CDED71F88814ULL, "vxdr_0_deeproads_ewn_0_0.vlm"},
	{0xDED8A3910CB88B3BULL, "w_sundermount_4_0.mmh"},
	{0xDED8AF910CB89FD8ULL, "w_sundermount_4_0.mao"},
	{0xDED8C1910CB8BE25ULL, "w_sundermount_4_0.msh"},
	{0xDED916015B28138FULL, "52147587.ogg"},
	{0xDEDA40445AEC88F4ULL, "dae000_merchant_generic_hm6172674.cl"},
	{0xDEDB5C70032FD9E9ULL, "vxht_0_hightown_bu5_0_0.vlm"},
	{0xDEDC1C9A12D06B79ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackleft.evt"},
	{0xDEDC78B8EB950611ULL, "mcr370_templar26087008.cl"},
	{0xDEDC79F5C9333ACAULL, "plt_seb000pt_gift_bow.nss"},
	{0xDEDCE9C00840700EULL, "6209952_f.fxe"},
	{0xDEDD11C5B3F79326ULL, "6042765_m.fxe"},
	{0xDEDDD4212B5686AAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maga_0t.dds"},
	{0xDEDE6BB05FE046ACULL, "follower_banter6160974.cl"},
	{0xDEE042DABF1A8A9AULL, "6048962_f.fxe"},
	{0xDEE0548F50D53731ULL, "6135150_m.fxe"},
	{0xDEE0C40F107758B0ULL, "hlac_lt_0_lowtown_131070_1x.dds"},
	{0xDEE10299B478A7D4ULL, "216000285.ogg"},
	{0xDEE20E5CB54D6E8AULL, "gen00fl_fenris6113944.cl"},
	{0xDEE234CB5CBECAFBULL, "art\\characters\\creatures\\spider\\animation\\kd_lp.ani"},
	{0xDEE25247456E3277ULL, "vxht_mansion_n_4zk_0_0.vlm"},
	{0xDEE25F3526775792ULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_hed_reva_0d.dds"},
	{0xDEE29560ECAABB9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_left_down.evt"},
	{0xDEE2A740F637E75CULL, "vxca_caverns2prc_9q9_0_0.vlm"},
	{0xDEE30C0664856DC9ULL, "llac_lt_undercitydungeon_-1_8.dds"},
	{0xDEE33B6A4639F116ULL, "ave100_ambush_16106201.cl"},
	{0xDEE3AB9A2883E6C7ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrenzy_lp_c.mmh"},
	{0xDEE40AF1E2AD2169ULL, "vxlt_gallowstemplar_5j8_0_0.vlm"},
	{0xDEE4C03329762E9FULL, "vxht_0_hightown_a3_b27_0_0.vlm"},
	{0xDEE520215D9BF70DULL, "mag100_cullen6206085.cl"},
	{0xDEE5402278C04B3BULL, "gen00fl_aveline6116642.cl"},
	{0xDEE593E6E691F1F5ULL, "vxw_bonepit_1qa_0_0.vlm"},
	{0xDEE69D0B1B1B7E9BULL, "hlac_ht_mansion2b_818_36z.dds"},
	{0xDEE6B6F4393B45B7ULL, "lt_0_lowtown_docks_a3_143_0.mmh"},
	{0xDEE6BAF4393B4CC4ULL, "lt_0_lowtown_docks_a3_143_0.mao"},
	{0xDEE6BECD796F8C0DULL, "zz_mer_debug6073709.cl"},
	{0xDEE6CCF4393B6B29ULL, "lt_0_lowtown_docks_a3_143_0.msh"},
	{0xDEE70C61774E1E3FULL, "6047367_m.fxe"},
	{0xDEE7763DAFFA494AULL, "6179593_m.fxe"},
	{0xDEE7AD94D9668871ULL, "vxlt_0_lowtown_a3_n_c_d7q_0_0.vlm"},
	{0xDEE7B20B25884DFCULL, "6065936_f.fxe"},
	{0xDEE7F49B90FB295CULL, "320102444.ogg"},
	{0xDEE8007C38325339ULL, "hlac_lt_0_lowtown_-3_3y.dds"},
	{0xDEE885695B4DF2CAULL, "605895371.ogg"},
	{0xDEE8DA6244252F31ULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.anb"},
	{0xDEE8E7043C6E113DULL, "6172744_m.fxe"},
	{0xDEE902752FC8EBA0ULL, "and100_anders6095976.cl"},
	{0xDEE924799A0B5176ULL, "rdr111it_martin_cargo.uti"},
	{0xDEE94536786CCF14ULL, "388538902.ogg"},
	{0xDEE956CCAF7B9E1EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_com_v1.evt"},
	{0xDEE99460ECB0D88DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_left_down.gad"},
	{0xDEE9B85DE7AE8D79ULL, "6137714_m.fxe"},
	{0xDEEA249A12DCB668ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\turningattackleft.ani"},
	{0xDEEA64635E342318ULL, "vxht_chantry_28s_0_0.vlm"},
	{0xDEEA8471DBFF6091ULL, "w_bonepitmining_a2_41_0.mmh"},
	{0xDEEA9071DBFF7532ULL, "w_bonepitmining_a2_41_0.mao"},
	{0xDEEA9E71DBFF8D4FULL, "w_bonepitmining_a2_41_0.msh"},
	{0xDEEAB27CEFFE4B30ULL, "w_encounter3_201_0.msh"},
	{0xDEEAC07CEFFE634AULL, "w_encounter3_201_0.mmh"},
	{0xDEEAC45144E69F7AULL, "vxlt_0_lowtown_n_es4_0_0.vlm"},
	{0xDEEAC47CEFFE6981ULL, "w_encounter3_201_0.mao"},
	{0xDEEB8D3675B80945ULL, "hlac_dr_0_deeproads_5839_2x.dds"},
	{0xDEEBDB5FEC873A69ULL, "609132783.ogg"},
	{0xDEEC3097639DFB58ULL, "vxw_bonepitmining_a2_np_0_0.vlm"},
	{0xDEEC609F79889C5FULL, "vxht_0_hightown_7vs_0_0.vlm"},
	{0xDEEC7A45DA70CCCAULL, "ch2notnt_ds1.psh.fxo"},
	{0xDEED22FA9FDE604CULL, "vxdr_primevaltomb_7o6_0_0.vlm"},
	{0xDEED877557086908ULL, "w_bonepitdestroyed_a3_51_0.mmh"},
	{0xDEED937557087DABULL, "w_bonepitdestroyed_a3_51_0.mao"},
	{0xDEEDA57557089C32ULL, "w_bonepitdestroyed_a3_51_0.msh"},
	{0xDEEE3C290E8B33B1ULL, "6113126_f.fxe"},
	{0xDEEE9244D9694070ULL, "w_bonepit_37_0_lc.mmh"},
	{0xDEEEB1248C4B4CDEULL, "en_flare.mao"},
	{0xDEEED2E75FD09664ULL, "567708500.ogg"},
	{0xDEEF6DCCAF803149ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_com_v1.gad"},
	{0xDEEF78BD9CE71A53ULL, "vxca_caverns2prc_7ud_0_0.vlm"},
	{0xDEEF7D849674F4AFULL, "vxht_chantry_n_2e4_0_0.vlm"},
	{0xDEEFECEA98A990B6ULL, "6142542_m.fxe"},
	{0xDEF09AF544DFABC2ULL, "vxlt_0_lowtown_a2_heh_0_0.vlm"},
	{0xDEF0E89803E604F5ULL, "vxw_woundedcoast_vv_0_0.vlm"},
	{0xDEF103F77B338E92ULL, "6130006_m.fxe"},
	{0xDEF13F0E587695FAULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga106a_0n.dds"},
	{0xDEF1400FFFEFCF86ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.cess.evt"},
	{0xDEF15137A9C5B4ABULL, "dae000_orana6149070.cl"},
	{0xDEF1A445D4222ED6ULL, "6150321_m.fxe"},
	{0xDEF1C3AF279D5414ULL, "dre000cr_dougal_thugs.utc"},
	{0xDEF1C66E3E55BC69ULL, "plt_cod_prc_dowsingrods.plo"},
	{0xDEF1F6BBE31E66A8ULL, "902221138.ogg"},
	{0xDEF2B0442E9C129CULL, "w_sundermount_39_0.mao"},
	{0xDEF2BC442E9C273FULL, "w_sundermount_39_0.mmh"},
	{0xDEF2C2442E9C31E1ULL, "w_sundermount_39_0.msh"},
	{0xDEF327730338FE81ULL, "51348093.ogg"},
	{0xDEF42AB3644F07EEULL, "6061847_m.fxe"},
	{0xDEF498E2AC68A683ULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em03.msh"},
	{0xDEF4A6E2AC68BEDDULL, "art\\characters\\playercharacter\\elfmale\\mt_ant_em03.mmh"},
	{0xDEF5279FB99D16D1ULL, "hlac_w_sundermount_327685_0y.dds"},
	{0xDEF58151177D6850ULL, "vxlt_backalley_g_1w0_0_0.vlm"},
	{0xDEF59DC97680FD74ULL, "6036341_f.fxe"},
	{0xDEF5C0E9B52D0068ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlavaed.mmh"},
	{0xDEF5CCE9B52D140BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlavaed.mao"},
	{0xDEF5DEE9B52D3292ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlavaed.msh"},
	{0xDEF5FB62442FF1AAULL, "art\\vfx\\combat\\vfx_f_fen_lyrium_crust_c.mmh"},
	{0xDEF6278AFE2B6FE6ULL, "vxlt_gallowstemplar_n_3se_0_0.vlm"},
	{0xDEF63ECCAF85FF97ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_com_v1.ani"},
	{0xDEF65BCB5CCFB585ULL, "art\\characters\\creatures\\spider\\animation\\kd_lp.gad"},
	{0xDEF67B26FB2F5BE4ULL, "vxlt_0_lowtown_a2_n_eu8_0_0.vlm"},
	{0xDEF758E12E6429BDULL, "6175644_m.fxe"},
	{0xDEF7C59C0AC81DE0ULL, "6032269_m.fxe"},
	{0xDEF823511514583DULL, "246045811.ogg"},
	{0xDEF824B2D7FAC46FULL, "hm_bar_gossip1.mor"},
	{0xDEF82AC79E9B9DAAULL, "llac_ht_playershouse_n_855_36.dds"},
	{0xDEF8570FFFF615D1ULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.cess.gad"},
	{0xDEF8A3F5488077B2ULL, "148935574.ogg"},
	{0xDEF8A813C86A248CULL, "blood_imp_08.mao"},
	{0xDEF8E7AEB3B5E4E4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_bronzefloor.msh"},
	{0xDEF8F9AEB3B6034DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_bronzefloor.mao"},
	{0xDEF8FDAEB3B60986ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_bronzefloor.mmh"},
	{0xDEF9204EA2573DE7ULL, "hlac_ht_keep_n_4519_1y.dds"},
	{0xDEFA1363D451BE45ULL, "vxdr_0_deeproads_8u0_0_0.vlm"},
	{0xDEFA24D9204E8842ULL, "art\\levels\\ztools\\default\\generic\\0ow_basic.phy"},
	{0xDEFA4857A4F85161ULL, "mcr371_orsino6083334.cl"},
	{0xDEFA526882F03F26ULL, "8001944.ogg"},
	{0xDEFA951406A2184AULL, "ico_armor_medium_4.dds"},
	{0xDEFABF38B88346FAULL, "art\\vfx\\combat\\vfx_gib_hm_roba.msh"},
	{0xDEFAD138B8836540ULL, "art\\vfx\\combat\\vfx_gib_hm_roba.mmh"},
	{0xDEFB6FE62284664EULL, "105166434.ogg"},
	{0xDEFBA15CACF97095ULL, "vxw_encounter2_4q_0_0.vlm"},
	{0xDEFBC4AAC9736A3EULL, "character_character_p168.psh.fxo.dx11"},
	{0xDEFCC4F742ECE838ULL, "741266220.ogg"},
	{0xDEFD53FF680C2AEDULL, "w_woundedcoast_226_0.mmh"},
	{0xDEFD57FF680C313EULL, "w_woundedcoast_226_0.mao"},
	{0xDEFD65FF680C4973ULL, "w_woundedcoast_226_0.msh"},
	{0xDEFDD9634072904FULL, "670654689.ogg"},
	{0xDEFDF84464553D5AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_feynriel.gad"},
	{0xDEFDFC2D581EC485ULL, "dae200_bran6094811.cl"},
	{0xDEFE2A13FF5C65A2ULL, "6025304_m.fxe"},
	{0xDEFE30336E9790F1ULL, "nat380_sibling6068004.cl"},
	{0xDEFE576E648E6A80ULL, "llac_ht_0_hightown_-65537_2.dds"},
	{0xDEFE580FFFFA837FULL, "art\\vfx\\combat\\vfx_glb_com_disorient_s1_c.cess.ani"},
	{0xDEFE86EDA7EAB2F3ULL, "669106274.ogg"},
	{0xDEFEEE477A27B25DULL, "zz_dre_debug6069196.cl"},
	{0xDEFF0CD236D5C442ULL, "vxlt_0_lowtown_a2_hbi_0_0.vlm"},
	{0xDEFF36CC857C053DULL, "vxht_0_hightown_9uz_0_0.vlm"},
	{0xDEFF5A74F5DC1D75ULL, "vxw_woundedcoast_su_0_0.vlm"},
	{0xDEFFB760963C0D7EULL, "follower_banter6171268.cl"},
	{0xDEFFD378D0E7C4D7ULL, "vxca_cavern1_haunted_e55_0_0.vlm"},
	{0xDEFFD44117AE340EULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_cupwood_01_0_d.dds"},
	{0xDEFFF31FC13ECAF2ULL, "gen00fl_fenris6099880.cl"},
	{0xDF0009361CF08937ULL, "321482171.ogg"},
	{0xDF00A004AE1A328FULL, "vxht_chantry_n_2v1_0_0.vlm"},
	{0xDF00B3562C07ACF6ULL, "dr_primevaltomb_657.rml"},
	{0xDF0198762A4E1C43ULL, "vxlt_gallowstemplar_557_0_0.vlm"},
	{0xDF01C960BC9DB535ULL, "zz_kneelsister.cut"},
	{0xDF01D9C07B3E1433ULL, "gen00fl_merrill6128930.cl"},
	{0xDF029B5ABD5BA524ULL, "6085490_m.fxe"},
	{0xDF02D63EA042C577ULL, "vxlt_gallowsprison_n_6qg_0_0.vlm"},
	{0xDF02F4AEB3BE4F51ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ff_bronzefloor.phy"},
	{0xDF034EA68DB2CF3EULL, "6113095_f.fxe"},
	{0xDF035789842687D6ULL, "vxlt_0_lowtown_docks_b5z_0_1.vlm"},
	{0xDF0385BD800B9EA5ULL, "636691831.ogg"},
	{0xDF03A70473A31026ULL, "6149103_m.fxe"},
	{0xDF03F652A3989971ULL, "6207777_f.fxe"},
	{0xDF041EDA32DFC6DEULL, "gen_im_trs_dal_junk1_01.uti"},
	{0xDF045E49CDA3B4B3ULL, "6114236_m.fxe"},
	{0xDF049AF9EDFBB5C0ULL, "211995306.ogg"},
	{0xDF04DD13161ED76BULL, "ser221_gamlen6135282.cl"},
	{0xDF04FB44645B6171ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_feynriel.evt"},
	{0xDF055B25E2D5D2DAULL, "hlac_w_encounter1_65541_0z.dds"},
	{0xDF05A022AC0D1621ULL, "hlac_ht_chantry_2534_5y.dds"},
	{0xDF067A04D2D74989ULL, "hlac_w_encounter1_0_2x.dds"},
	{0xDF077D416A2186A3ULL, "hlac_ca_cavern1_haunted_24424_6y.dds"},
	{0xDF07A80B28D2D07AULL, "vxlt_0_lowtown_n_ei6_0_0.vlm"},
	{0xDF0839485F032DE5ULL, "285612274.ogg"},
	{0xDF0847EA938884F1ULL, "hlac_lt_gallowstemplar_f_2394_6z.dds"},
	{0xDF089B3A1EBD3B1EULL, "zz_lgt_3_debug6176147.cl"},
	{0xDF08E1089EB0F2D4ULL, "llac_lt_hangedman_-1_5.dds"},
	{0xDF09B5BFB0FF45BAULL, "isa111_isabela6068914.cl"},
	{0xDF09E0F90356219BULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.cess.evt"},
	{0xDF09F429BC56019AULL, "700538566.ogg"},
	{0xDF0A06B483CA914CULL, "vxdr_primevaltomb_7e4_0_0.vlm"},
	{0xDF0A10296EAFD625ULL, "vxca_cavern1_haunted_imk_0_0.vlm"},
	{0xDF0A19B8EA8C3DF0ULL, "180171284.ogg"},
	{0xDF0A6695C933BD95ULL, "vxw_blightlands_ck_0_0.vlm"},
	{0xDF0B11281C9E0166ULL, "439608332.ogg"},
	{0xDF0B150A5CC9BC90ULL, "static_static_p14.vsh.fxo.dx11"},
	{0xDF0B84ACD61BD801ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation5.ani"},
	{0xDF0B9DD7E0F1699EULL, "mcr310_argument6087311.cl"},
	{0xDF0BA07A3D8E69CDULL, "vxlt_undercityserialkiller_1t4_0_0.vlm"},
	{0xDF0BDF5D0F5A1CDBULL, "hlac_lt_0_lowtown_a2_n_-3_6y.dds"},
	{0xDF0BF623A57F096DULL, "6043808_m.fxe"},
	{0xDF0C3AE623291130ULL, "212272242.ogg"},
	{0xDF0C482A20C55113ULL, "dae100_bran6058937.cl"},
	{0xDF0CF8FCCC1627DCULL, "6021225_m.fxe"},
	{0xDF0D5D1EC2E821D1ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\ready.ani"},
	{0xDF0D73F38C0E28F1ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_6craggyrock.mmh"},
	{0xDF0D7CC6F4D6D1F9ULL, "dae000_madam_lusine6152197.cl"},
	{0xDF0D7FF38C0E3C92ULL, "art\\levels\\races\\natural\\highlands\\props\\prp_6craggyrock.mao"},
	{0xDF0D8DF38C0E54AFULL, "art\\levels\\races\\natural\\highlands\\props\\prp_6craggyrock.msh"},
	{0xDF0E4CFF681A91AAULL, "w_woundedcoast_226_0.phy"},
	{0xDF0EF74707CD1C04ULL, "gen00fl_anders6119898.cl"},
	{0xDF0F11E1A0104B0FULL, "vxht_chantry_n_2q0_0_0.vlm"},
	{0xDF0F4A73AA0E6C30ULL, "6125600_m.fxe"},
	{0xDF0F869531EFB217ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\til_hightown_main_n.dds"},
	{0xDF10C1C2A9E6150CULL, "art\\vfx\\combat\\vfx_glb_phys_imp_maj_c.impa.ani"},
	{0xDF10D9F9035C34DCULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.cess.gad"},
	{0xDF10F9E405DBEEC1ULL, "485913292.ogg"},
	{0xDF10FE6FF58F444AULL, "hlac_w_encounter2_camp_2_0x.dds"},
	{0xDF1137AB89D9CED0ULL, "vxw_bonepitmining_a2_1jk_0_0.vlm"},
	{0xDF113FF14D637850ULL, "w_bonepit_132_0.msh"},
	{0xDF114518FDD1201AULL, "hlac_ca_caverns1_24423_2x.dds"},
	{0xDF114DF14D63906AULL, "w_bonepit_132_0.mmh"},
	{0xDF1151F14D6396A1ULL, "w_bonepit_132_0.mao"},
	{0xDF115D2E247888A0ULL, "ht_playershouse_n_mmap.dds"},
	{0xDF1178FFC74FFC66ULL, "6111146_m.fxe"},
	{0xDF118FACD62056A3ULL, "art\\cinematics\\additional animation\\cs_genericmale_additionalanimation5.gad"},
	{0xDF119589D42F41DAULL, "lt_0_lowtown_a2_n_c_27_0.rml"},
	{0xDF11C179ACCAEDB1ULL, "and230_mage_girl6144066.cl"},
	{0xDF11C28EDD1AAD85ULL, "6089705_m.fxe"},
	{0xDF11CE892E211AC3ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magc.mmh"},
	{0xDF11DC892E21321DULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magc.msh"},
	{0xDF120D75A3C34622ULL, "6082220_m.fxe"},
	{0xDF12134464661440ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_killshot_feynriel.ani"},
	{0xDF1217EC69BEF764ULL, "lowtown_fem_lowlife_mob6169177.cl"},
	{0xDF1257C937F1F548ULL, "w_bonepit_99_0.mmh"},
	{0xDF125F327DDA29F0ULL, "404078238.ogg"},
	{0xDF1263C937F209EBULL, "w_bonepit_99_0.mao"},
	{0xDF126533A4198192ULL, "llac_w_blightlands_131074_1.dds"},
	{0xDF1275C937F22872ULL, "w_bonepit_99_0.msh"},
	{0xDF127CA7383748FCULL, "screentexrect.vsh.fxo"},
	{0xDF13503E6BF5AB7AULL, "t3_bow_hrn.tnt"},
	{0xDF135AA46A12BCD0ULL, "vrc321_party_barks6095778.cl"},
	{0xDF13681EC2ECA073ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\ready.gad"},
	{0xDF13AF5A4658E829ULL, "vxht_0_hightown_a2_a89_0_0.vlm"},
	{0xDF13F6F80EBC91AFULL, "vxlt_hangedman_n_19e_0_0.vlm"},
	{0xDF1421E7C32242B3ULL, "vxlt_0_lowtown_a2_n_c_e15_0_0.vlm"},
	{0xDF149C0BACC714A9ULL, "174733309.ogg"},
	{0xDF14B5F610E6E940ULL, "738285825.ogg"},
	{0xDF150A67BFEE0A81ULL, "vxht_keep_n_371_0_0.vlm"},
	{0xDF15A0A6C09F38C1ULL, "vxht_0_hightown_a3_c0j_0_0.vlm"},
	{0xDF15E7334E114414ULL, "1028004090.ogg"},
	{0xDF163A98A18BFF5AULL, "6178396_m.fxe"},
	{0xDF1646DBB15EADEAULL, "6153698_m.fxe"},
	{0xDF1675A36E8D3247ULL, "348836870.ogg"},
	{0xDF16D767B4AD5C71ULL, "6082195_m.fxe"},
	{0xDF16D8F903609EDAULL, "art\\vfx\\combat\\vfx_glb_fire_imp_ext_c.cess.ani"},
	{0xDF178903AFACFBF5ULL, "6048463_m.fxe"},
	{0xDF179E15879009A2ULL, "atl_startmenu2_dxt1_02.dds"},
	{0xDF17CEC2A9EC4A4EULL, "art\\vfx\\combat\\vfx_glb_phys_imp_maj_c.impa.gad"},
	{0xDF18820660A5EEA5ULL, "vxca_cavern1_haunted_ijj_0_0.vlm"},
	{0xDF189621C9F94609ULL, "statecnt_informational.gda"},
	{0xDF18D872BB29D615ULL, "vxw_blightlands_dj_0_0.vlm"},
	{0xDF193B9124EEAFB6ULL, "llac_dr_slavers_3214_0.dds"},
	{0xDF195C7369E11761ULL, "mer110_jansen6032899.cl"},
	{0xDF1A066871AB0876ULL, "hlac_w_blightlands_-65538_0y.dds"},
	{0xDF1A07568D47EF99ULL, "sp_module_item_acq.ncs"},
	{0xDF1A10B3CAB129A2ULL, "6107289_f.fxe"},
	{0xDF1A651EC2F2B9D8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\ready.evt"},
	{0xDF1AFEF5AFCAF546ULL, "380929722.ogg"},
	{0xDF1B66C4579CBAFEULL, "529074757.ogg"},
	{0xDF1B898BC4B76301ULL, "137772675.ogg"},
	{0xDF1B8F3CE8894652ULL, "zz_mrl_debug6185303.cl"},
	{0xDF1C170E62A82588ULL, "370788898.ogg"},
	{0xDF1C7EA8D01C6C82ULL, "6107534_m.fxe"},
	{0xDF1C94E76D9F5589ULL, "vxw_encounter3_8n_0_0.vlm"},
	{0xDF1C95E4D63ACF06ULL, "6103974_m.fxe"},
	{0xDF1CC86F3B32786CULL, "gen00fl_sebastian6179909.cl"},
	{0xDF1CFF42D167272BULL, "vrc321ip_ledger.ncs"},
	{0xDF1D30C4EC8B8621ULL, "467316411.ogg"},
	{0xDF1D5363807CBE2CULL, "6099705_m.fxe"},
	{0xDF1EC9C2A9F2602DULL, "art\\vfx\\combat\\vfx_glb_phys_imp_maj_c.impa.evt"},
	{0xDF1EFFD9206D65F6ULL, "art\\levels\\ztools\\default\\generic\\0ow_basic.mao"},
	{0xDF1F0BD9206D7995ULL, "art\\levels\\ztools\\default\\generic\\0ow_basic.mmh"},
	{0xDF1F0DD9206D7D0BULL, "art\\levels\\ztools\\default\\generic\\0ow_basic.msh"},
	{0xDF208228E07E54EDULL, "6164831_m.fxe"},
	{0xDF213E68040D6AB3ULL, "vxlt_0_lowtown_h8n_0_0.vlm"},
	{0xDF21651C90D0F56BULL, "art\\characters\\playercharacter\\children\\mt_bas_km_facem1.mao"},
	{0xDF21933B7D71BB80ULL, "6086734_m.fxe"},
	{0xDF22CF892E2F8FE8ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_magc.phy"},
	{0xDF23235363D4DFE2ULL, "6114871_m.fxe"},
	{0xDF235DCBC49FDDF7ULL, "dae210_gate_guard6099082.cl"},
	{0xDF239D73B116543EULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.impa.ani"},
	{0xDF2406DFDCF333F4ULL, "vxht_brothel_n_3v4_0_0.vlm"},
	{0xDF241283B2955141ULL, "vxht_0_hightown_a3_c5i_0_0.vlm"},
	{0xDF243F259EE43CB9ULL, "vxlt_0_lowtown_n_gjz_0_0.vlm"},
	{0xDF2441FE395EC2C0ULL, "one220_yevhen6103288.cl"},
	{0xDF24BC53005DA3DBULL, "6102954_m.fxe"},
	{0xDF24E5137130A7B8ULL, "vxlt_gallowstemplar_f_4yj_0_0.vlm"},
	{0xDF251C8B3C55AB35ULL, "6151105_m.fxe"},
	{0xDF253AA1F0728D53ULL, "1066236456.ogg"},
	{0xDF2561B1ADF29AA2ULL, "plt_prm_cod_lost_tome.plo"},
	{0xDF25779D16FD6794ULL, "w_bonepitmining_a2_108_0.phy"},
	{0xDF25FAE0A1A31475ULL, "6043563_m.fxe"},
	{0xDF261D3BB6E85C7AULL, "6134851_m.fxe"},
	{0xDF26A3910CFAACC8ULL, "w_sundermount_4_0.rml"},
	{0xDF26F1DC6C374184ULL, "6042368_f.fxe"},
	{0xDF2701B1B0A6B873ULL, "vxw_sundermount_31b_0_0.vlm"},
	{0xDF274A4FAD1FEE95ULL, "vxw_blightlands_ii_0_0.vlm"},
	{0xDF27594AAA449D4BULL, "6041512_m.fxe"},
	{0xDF27A206E15B9A37ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlargeb.msh"},
	{0xDF27B406E15BB89AULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlargeb.mao"},
	{0xDF27B806E15BBFA9ULL, "art\\levels\\races\\primeval\\props\\lyrium\\prp_lyriumlargeb.mmh"},
	{0xDF28A3926707C54EULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernrock_sa.dds"},
	{0xDF28CFB9B1C21E98ULL, "ss_isabela.bnk"},
	{0xDF291E1B28D33BA1ULL, "crust_webs4.dds"},
	{0xDF2A361E81A93100ULL, "vxlt_0_lowtown_a3_n_c_ewu_0_0.vlm"},
	{0xDF2A4572CC0D4FB0ULL, "662301145.ogg"},
	{0xDF2AE091B00661DCULL, "vxca_caverns2prc_9h2_0_0.vlm"},
	{0xDF2AEC6A1EA966C5ULL, "6163452_m.fxe"},
	{0xDF2AEE72795B6A40ULL, "zz_dae_debug6029516.cl"},
	{0xDF2B06C45F956E09ULL, "vxw_encounter3_5m_0_0.vlm"},
	{0xDF2B353BAF73D29FULL, "vxca_caverns1_mines_dr1_0_0.vlm"},
	{0xDF2B5AC50CDB9F0DULL, "dae000_keep_guard_2_sta6163816.cl"},
	{0xDF2BB7AC46839DC7ULL, "428596827.ogg"},
	{0xDF2C023F9051E059ULL, "6176429_m.fxe"},
	{0xDF2C119966A5EDB7ULL, "vxca_cavern1_haunted_e91_0_0.vlm"},
	{0xDF2DB037542B93D3ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_lichen.dds"},
	{0xDF2DD2AA56E82230ULL, "341731069.ogg"},
	{0xDF2DEFBEA4C6DFBDULL, "vxht_chantry_n_28u_0_0.vlm"},
	{0xDF2DF229DD4ED2F6ULL, "vxlt_gallowscourtyard_7ez_0_0.vlm"},
	{0xDF2E376AE287940DULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\blendtree\\trooper.anb"},
	{0xDF2E75CE72E2A13FULL, "6080083_m.fxe"},
	{0xDF2E87EC7F807FE2ULL, "vxht_mansion_n_4f6_0_0.vlm"},
	{0xDF2EFCDCB7444AE7ULL, "lt_0_lowtown_n_25_0_lc.mmh"},
	{0xDF2F24221C4A3AF3ULL, "187752560.ogg"},
	{0xDF2F2D2BD9A6C6D4ULL, "vxw_encounter2_camp_p1_0_1.vlm"},
	{0xDF2F549D17058091ULL, "w_bonepitmining_a2_108_0.msh"},
	{0xDF2F629D170598ACULL, "w_bonepitmining_a2_108_0.mao"},
	{0xDF2F6E9D1705AD4FULL, "w_bonepitmining_a2_108_0.mmh"},
	{0xDF2FB8E98CC822F8ULL, "6044620_m.fxe"},
	{0xDF2FC6525CD34174ULL, "vxw_bonepitdestroyed_a3_1x2_0_1.vlm"},
	{0xDF3031724CFE401EULL, "ht_0_hightown_17_0_ll.mmh"},
	{0xDF3066E773943201ULL, "vxca_caverns2prc_9fm_0_0.vlm"},
	{0xDF30B573B121070FULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.impa.evt"},
	{0xDF30F7FD4391B1D1ULL, "sky_dr_dungeon.mao"},
	{0xDF3128964A6F263EULL, "6076314_f.fxe"},
	{0xDF315BDFCCCED9C6ULL, "fx_fire01.dds"},
	{0xDF318A608DCFE1B7ULL, "ht_0_hightown_10_0.rml"},
	{0xDF319A2750CFF63AULL, "w_woundedcoast_214_0.mmh"},
	{0xDF319E2750CFFD71ULL, "w_woundedcoast_214_0.mao"},
	{0xDF31AC2750D01480ULL, "w_woundedcoast_214_0.msh"},
	{0xDF31D4C2BC2AB4D5ULL, "6079562_m.fxe"},
	{0xDF320EC52A20275FULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha13a.mmh"},
	{0xDF3214C52A203101ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha13a.msh"},
	{0xDF321EE2A10FD388ULL, "vxca_caverns2_8ea_0_0.vlm"},
	{0xDF324F9F5FF3E7E9ULL, "vxlt_0_lowtown_a2_hik_0_0.vlm"},
	{0xDF3259BC7F3E9590ULL, "vrc221_cassandra6040892.cl"},
	{0xDF32939BC7F9E43AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maga_0d.dds"},
	{0xDF32A53800C7321FULL, "mcr371_scared_mage_m6087700.cl"},
	{0xDF3315BBF3DD3FC4ULL, "628068789.ogg"},
	{0xDF33722F14872ADDULL, "6085166_f.fxe"},
	{0xDF3422766D189F62ULL, "lt_warehouse_n_0_0_lc.mmh"},
	{0xDF34708A2EC4B6B6ULL, "hlac_lt_0_lowtown_n_-2_2y.dds"},
	{0xDF3481161E94DDF7ULL, "6035160_m.fxe"},
	{0xDF34A53AD4B1E959ULL, "art\\vfx\\combat\\vfx_gui_target_ret_120_p.dura.ani"},
	{0xDF34B418B9E6EDD7ULL, "hlac_lt_0_lowtown_a3_n_c_65537_3y.dds"},
	{0xDF34BF82155DBD44ULL, "gen00fl_varric6076539.cl"},
	{0xDF34DD6382F92CD5ULL, "6101386_m.fxe"},
	{0xDF34E5ED326C4C8EULL, "463071429.ogg"},
	{0xDF353B9F8DA38955ULL, "6067389_m.fxe"},
	{0xDF3559F52D8EA0FAULL, "146156197.ogg"},
	{0xDF355CB9BC7FE637ULL, "6155532_m.fxe"},
	{0xDF35A21C29A6BDB1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action_male.ani"},
	{0xDF35B4562BA7CC2CULL, "6080163_m.fxe"},
	{0xDF35BC2C9F160715ULL, "vxw_blightlands_rh_0_0.vlm"},
	{0xDF35E46287E03CB6ULL, "zz_dog_debug6127924.cl"},
	{0xDF3636D88CD2CE59ULL, "lt_0_lowtown_docks_n_28_0_lc.mmh"},
	{0xDF3651CA0369FE32ULL, "fx_fire_crack.mao"},
	{0xDF369CE832CAC81AULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_manystrip_n.dds"},
	{0xDF37923654DAEC28ULL, "vxw_bonepit_1s2_0_0.vlm"},
	{0xDF37B673B12727B8ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.impa.gad"},
	{0xDF38B8E4EC887985ULL, "dae200cr_viscount6073934.cl"},
	{0xDF393AE791A93B1AULL, "528992434.ogg"},
	{0xDF3944B83B7F109DULL, "6160822_m.fxe"},
	{0xDF39C13578ADDA5DULL, "w_bonepitdestroyed_a3_30_0_lc.mmh"},
	{0xDF3A2D031EFB012FULL, "6132993_m.fxe"},
	{0xDF3A371D6468EF2CULL, "6093505_m.fxe"},
	{0xDF3A9047095518D6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_basic.msh"},
	{0xDF3A99F6D7494F05ULL, "vxlt_undercityencounter_13e_0_0.vlm"},
	{0xDF3A9A47095529D4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_basic.mmh"},
	{0xDF3A9E47095530E7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_basic.mao"},
	{0xDF3AE82270842D9EULL, "gen_im_wep_mag_sta_act3_07.uti"},
	{0xDF3B0442D1802D16ULL, "vrc321ip_ledger.utp"},
	{0xDF3B1DE409517B6AULL, "904273018.ogg"},
	{0xDF3B5634BE1F6512ULL, "6107350_m.fxe"},
	{0xDF3B803AD4B7C89BULL, "art\\vfx\\combat\\vfx_gui_target_ret_120_p.dura.gad"},
	{0xDF3BA3A0BE9C923FULL, "6208067_m.fxe"},
	{0xDF3BA76AA7E03B02ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_kneeling_faint.anb"},
	{0xDF3BAF3DD6CFD5E4ULL, "398267799.ogg"},
	{0xDF3BB3FE18297BECULL, "6098514_f.fxe"},
	{0xDF3C108B91917D82ULL, "6098300_m.fxe"},
	{0xDF3C3FFACFBCFC91ULL, "6044889_m.fxe"},
	{0xDF3C619B96BCF83DULL, "vxht_chantry_n_23t_0_0.vlm"},
	{0xDF3CABC7CC972B1EULL, "character_character_p23.psh.fxo.dx11"},
	{0xDF3CAD1C29ACEF53ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action_male.gad"},
	{0xDF3CF464FC2AF09DULL, "hlac_ht_mansion2_818_45x.dds"},
	{0xDF3D2A2071B19A4BULL, "pro000_colonel_ewald6039962.cl"},
	{0xDF3D4F2AC7F3B769ULL, "dae000_flemeth6013395.cl"},
	{0xDF3DE69DEBC57112ULL, "zz_batchexport1.arl"},
	{0xDF3EB8F22E7B980FULL, "89533067.ogg"},
	{0xDF3EFA47FB6EC9F1ULL, "llac_lt_undercitydungeon_-1_13.dds"},
	{0xDF3FB2BB8893C6ADULL, "6095651_f.fxe"},
	{0xDF3FE43D6B74CC6BULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_flem_flemeth_real.ani"},
	{0xDF3FFDD4533D5B81ULL, "dae000ip_crowd_ff4_talk15.utp"},
	{0xDF4016C9460108F1ULL, "vxht_0_hightown_an3_0_0.vlm"},
	{0xDF4030A5EA8CB7B1ULL, "dre000_varric6208028.cl"},
	{0xDF4061E47F9FBE41ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_0y.dds"},
	{0xDF40D1A82714201BULL, "mrl000_marethari6205681.cl"},
	{0xDF40E46232FB6304ULL, "art\\cinematics\\deathblows\\rockgolem\\db_pc_rockwarr.ani"},
	{0xDF41FB3D174C52B4ULL, "6110706_f.fxe"},
	{0xDF422DE66DEA4A44ULL, "vxlt_0_lowtown_n_e1p_0_0.vlm"},
	{0xDF4264A46AAC05B2ULL, "6080976_m.fxe"},
	{0xDF429D3AD4BE1880ULL, "art\\vfx\\combat\\vfx_gui_target_ret_120_p.dura.evt"},
	{0xDF42AA1C29B155B8ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\meredith_kills_bethany_action_male.evt"},
	{0xDF42D76776DB4DEDULL, "6104415_m.fxe"},
	{0xDF4365538208BB70ULL, "6059914_m.fxe"},
	{0xDF4379BE4C00BBEBULL, "plt_cod_hst_thedas_geogrphy.plo"},
	{0xDF442E7B29DFBDB3ULL, "6133737_f.fxe"},
	{0xDF4468CD1343771BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_ornate.mmh"},
	{0xDF4474CD13438BB8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_ornate.mao"},
	{0xDF44778F4D12E535ULL, "6190053_f.fxe"},
	{0xDF447AD3A489909FULL, "plt_cod_itm_fallen.plo"},
	{0xDF4486CD1343AA05ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_ornate.msh"},
	{0xDF44A5964806DEAFULL, "docks_fem_sailor_mob6176424.cl"},
	{0xDF44EAEA5EE414DDULL, "hlac_lt_0_lowtown_a2_n_c_-65536_1x.dds"},
	{0xDF44FA698A1C7E8CULL, "vxw_encounter3_ed_0_0.vlm"},
	{0xDF4564ADF5DFC928ULL, "6075854_f.fxe"},
	{0xDF4612D87DB62A60ULL, "vxlt_undercityencounter_1c4_0_0.vlm"},
	{0xDF462DD4662B474FULL, "dae000_surprises6157186.cl"},
	{0xDF46BBD82378EFE8ULL, "6028367_m.fxe"},
	{0xDF46CF344D5A9522ULL, "pro000_aveline6047250.cl"},
	{0xDF46D283693C6F92ULL, "614048115.ogg"},
	{0xDF47E16233017D66ULL, "art\\cinematics\\deathblows\\rockgolem\\db_pc_rockwarr.gad"},
	{0xDF47E9D4508B2848ULL, "849883996.ogg"},
	{0xDF481D82A9964CF6ULL, "vxlt_0_lowtown_docks_bca_0_0.vlm"},
	{0xDF4827114FEDE103ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_tiki_sa.dds"},
	{0xDF485669B3F0C38CULL, "6048758_m.fxe"},
	{0xDF48EF9B9B1F966BULL, "zz_tbe100_barber_flyer.cnv"},
	{0xDF491567FF74A686ULL, "6044962_f.fxe"},
	{0xDF4A50F2531A437CULL, "vxca_caverns2_96y_0_0.vlm"},
	{0xDF4A64AF06991C23ULL, "dae000ip_crowd_ff5_talk9.utp"},
	{0xDF4A79AD72C51E12ULL, "dae011_gossips6143016.cl"},
	{0xDF4AFF1F51B230D0ULL, "vxw_bonepitmining_a2_1vw_0_0.vlm"},
	{0xDF4B22236363269CULL, "280595032.ogg"},
	{0xDF4B7289D4607EE6ULL, "lt_0_lowtown_a2_n_c_27_0.anb"},
	{0xDF4BCE8E0A0ED366ULL, "vxlt_hangedman_9_0_0.vlm"},
	{0xDF4BE7227B1F9962ULL, "6029054_m.fxe"},
	{0xDF4BF6A740F3ACBEULL, "zz_approval_debug6003766.cl"},
	{0xDF4C05CE1586159FULL, "dae000_pickpocket_scene6175160.cl"},
	{0xDF4C15A84D3A2764ULL, "905562960.ogg"},
	{0xDF4C50FBAC552681ULL, "hlac_lt_0_lowtown_a2_-2_5z.dds"},
	{0xDF4C6AA830F1EF78ULL, "hlac_lt_0_lowtown_a2_n_c_65535_9y.dds"},
	{0xDF4C9FEA298E385BULL, "434581868.ogg"},
	{0xDF4D399C67DC2C0DULL, "vxca_cave_i6_0_1.vlm"},
	{0xDF4D86788182653DULL, "6134467_m.fxe"},
	{0xDF4DE14AFCD3C867ULL, "6119709_m.fxe"},
	{0xDF4E3D4616817391ULL, "6060721_f.fxe"},
	{0xDF4EE59C97BF24ACULL, "vxw_bonepitmining_a2_4v_0_0.vlm"},
	{0xDF4F72B2A093C72EULL, "hlac_lt_0_lowtown_docks_a3_-131074_7y.dds"},
	{0xDF4F7DB3DDFD569EULL, "mag100_cullen6095110.cl"},
	{0xDF4FF7FE2C6140DFULL, "vxlt_0_lowtown_fjb_0_0.vlm"},
	{0xDF5041CD5F6C09E2ULL, "gen00fl_aveline6122163.cl"},
	{0xDF5074BCB64B54D2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_viscdesk_sa.dds"},
	{0xDF50F756B105DF7CULL, "hlac_ca_caverns2prc_5802_31z.dds"},
	{0xDF51650FE11AAB88ULL, "41324642.ogg"},
	{0xDF5207CD113306CCULL, "262095769.ogg"},
	{0xDF52DD2F1AE453CEULL, "w_woundedcoast_mmap.dds"},
	{0xDF52F3F572A941ADULL, "hlac_lt_0_lowtown_n_-1_4y.dds"},
	{0xDF53213497F8FD86ULL, "llac_lt_0_lowtown_a2_n_c_65535_2.dds"},
	{0xDF5325D1B1D23084ULL, "597687525.ogg"},
	{0xDF5372C82D95DB3EULL, "mage_bloodmagic.dds"},
	{0xDF538E4B591C2BC4ULL, "dae000_meredith6166301.cl"},
	{0xDF53CB3D6B854AF5ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_flem_flemeth_real.gad"},
	{0xDF53FB11C4FACD2AULL, "vxlt_gallowscourtyard_n_4lm_0_0.vlm"},
	{0xDF5418D696171573ULL, "follower_banter6172877.cl"},
	{0xDF54BD62A9B15409ULL, "vxlt_0_lowtown_a2_eun_0_0.vlm"},
	{0xDF54C5442EEF0084ULL, "w_sundermount_39_0.phy"},
	{0xDF55473342BCD00CULL, "vxht_keep_6ap_0_0.vlm"},
	{0xDF554BC30211AF75ULL, "gen00fl_varric6093689.cl"},
	{0xDF559875B1DE9061ULL, "vxht_keep_5st_0_0.vlm"},
	{0xDF55FC1379BFCC5EULL, "dre000_vampire6034322.cl"},
	{0xDF565375F0860E53ULL, "qun231cr_viscount_guard1.utc"},
	{0xDF568F5F9B8C6576ULL, "vxlt_0_lowtown_docks_bfb_0_0.vlm"},
	{0xDF56B92750EF4745ULL, "w_woundedcoast_214_0.phy"},
	{0xDF5713A0DE85E1D2ULL, "hlac_lt_0_lowtown_a3_n_c_65535_5x.dds"},
	{0xDF5769F1BB279FF6ULL, "6149235_m.fxe"},
	{0xDF5790C72C64B275ULL, "6039711_m.fxe"},
	{0xDF582DCEE0E1D32FULL, "575041049.ogg"},
	{0xDF5833241E172A4DULL, "hlac_lt_0_lowtown_-131071_2z.dds"},
	{0xDF58448C5E00DA31ULL, "ave310tr_party_comment.utt"},
	{0xDF58714D530FD61FULL, "gen00fl_fenris6126428.cl"},
	{0xDF58C89D6C4F28E5ULL, "vxw_blightlands_i9_0_0.vlm"},
	{0xDF5948E67550EC26ULL, "sky_sky2_p1.psh.fxo.dx11"},
	{0xDF594E98AAD4E30DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scrubbrush.msh"},
	{0xDF595C98AAD4FB40ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scrubbrush.mao"},
	{0xDF595E6C56560A4EULL, "vxht_chantry_n_2bp_0_0.vlm"},
	{0xDF596098AAD50193ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_scrubbrush.mmh"},
	{0xDF5971201CBF51FFULL, "hlac_ca_caverns1_mines_24422_2y.dds"},
	{0xDF59C2545B4AC7FBULL, "6103724_m.fxe"},
	{0xDF5A7388214B42FBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_basket_0.pwk"},
	{0xDF5AD9E2EEC0CF61ULL, "vxht_0_hightown_n_cfb_0_0.vlm"},
	{0xDF5ADE78BF2ABF4CULL, "vxca_cave_23v_0_0.vlm"},
	{0xDF5B7F7DA8DE6D65ULL, "vxlt_0_lowtown_a3_n_c_byi_0_0.vlm"},
	{0xDF5BBC442EF5100CULL, "w_sundermount_39_0.rml"},
	{0xDF5C1D6302DD418EULL, "6172578_m.fxe"},
	{0xDF5C71F8FC0AE7D9ULL, "gen00fl_isabela6075245.cl"},
	{0xDF5C8494748CD5F3ULL, "6067282_m.fxe"},
	{0xDF5D48CAC1ABB6F1ULL, "vxlt_undercitydungeon_27g_0_0.vlm"},
	{0xDF5D857B5E891DF5ULL, "498300872.ogg"},
	{0xDF5D91A3D46EFE75ULL, "vxca_caverns1_ebw_0_0.vlm"},
	{0xDF5E86245C787EF3ULL, "sib131_slaver_guard.bnk"},
	{0xDF5EC14AB241ADA2ULL, "vxlt_gallowscourtyard_n_7dn_0_0.vlm"},
	{0xDF5F102418A44BD2ULL, "vxlt_0_lowtown_a3_n_c_glt_0_0.vlm"},
	{0xDF5FF3D899019E02ULL, "vxca_caverns2prc_aes_0_0.vlm"},
	{0xDF6056EC81786AEFULL, "mag100_cullen6093551.cl"},
	{0xDF60AC870AF1C5D7ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\horror.gad"},
	{0xDF616B7E53C1EE5BULL, "6187834_m.fxe"},
	{0xDF6176B1164D9095ULL, "6018637_m.fxe"},
	{0xDF619D45C7989188ULL, "350276234.ogg"},
	{0xDF61A0D29BCE1D26ULL, "hlac_lt_gallowstemplar_f_2391_4x.dds"},
	{0xDF61C073B4D9F2B1ULL, "vxlt_0_lowtown_docks_a3_ctq_0_0.vlm"},
	{0xDF6209E2FC446304ULL, "781489786.ogg"},
	{0xDF6269B4594920B6ULL, "gen00fl_aveline6115868.cl"},
	{0xDF62A6B3D0DC064FULL, "ave100cr_donnic.utc"},
	{0xDF62A8245C7CA143ULL, "sib131_slaver_guard.cnv"},
	{0xDF632F3F9BA76C89ULL, "vxlt_0_lowtown_a2_exm_0_0.vlm"},
	{0xDF634ECB9F5470B4ULL, "llac_ht_chantry_n_2534_40.dds"},
	{0xDF635CFC91A65141ULL, "6143227_m.fxe"},
	{0xDF63AD6472A87405ULL, "6101467_m.fxe"},
	{0xDF63D8E965461339ULL, "w_woundedcoast_202_0.msh"},
	{0xDF63E2E965462427ULL, "w_woundedcoast_202_0.mmh"},
	{0xDF63E6E965462B74ULL, "w_woundedcoast_202_0.mao"},
	{0xDF64029B7ED60EA3ULL, "303310436.ogg"},
	{0xDF64510A0E4C9FE1ULL, "hlac_w_encounter1_4_0z.dds"},
	{0xDF64B49C52F60092ULL, "hlac_lt_undercityserialkiller_-1_16y.dds"},
	{0xDF64F61642714607ULL, "vxca_caverns2_7f8_0_0.vlm"},
	{0xDF661353A9BAD912ULL, "vxw_bonepitdestroyed_a3_1xd_0_1.vlm"},
	{0xDF66235982318C9AULL, "art\\characters\\playercharacter\\elfmale\\boo_nuda.phy"},
	{0xDF664E38D95CBBB8ULL, "6166807_m.fxe"},
	{0xDF66917E1FB30BD4ULL, "6101413_m.fxe"},
	{0xDF673A7A5E454165ULL, "vxw_blightlands_r8_0_0.vlm"},
	{0xDF673BC3C982A895ULL, "6052907_m.fxe"},
	{0xDF67510A7250EB63ULL, "vxca_caverns1_ii0_0_0.vlm"},
	{0xDF67AEF2DD770482ULL, "vxlt_0_lowtown_a2_n_c_hnh_0_0.vlm"},
	{0xDF67C1870AF80864ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\horror.evt"},
	{0xDF67F3BAE00470DAULL, "art\\characters\\weapons\\shields\\textures\\w_shd_007a_s.dds"},
	{0xDF68DE90B64A77DAULL, "lgt250_way_dalish_conflict6147159.cl"},
	{0xDF68FA7D7195278CULL, "vxht_0_hightown_n_byo_0_0.vlm"},
	{0xDF694BBFE0B6E7E1ULL, "vxht_0_hightown_n_cca_0_0.vlm"},
	{0xDF695C90280C7C0AULL, "6095237_m.fxe"},
	{0xDF69E724480537F6ULL, "6041491_f.fxe"},
	{0xDF69F64706824B66ULL, "vxca_cavern1_haunted_dpn_0_0.vlm"},
	{0xDF6A31F649268832ULL, "6044812_f.fxe"},
	{0xDF6A82C7F837B392ULL, "mrl250_pol6101503.cl"},
	{0xDF6A8F3FF4D35A0EULL, "6175478_m.fxe"},
	{0xDF6AD3F3E32E8A39ULL, "vxca_caverns2prc_7qr_0_0.vlm"},
	{0xDF6B4697DC09D672ULL, "319870369.ogg"},
	{0xDF6BBA51AC6A748FULL, "one101_orders_followers6115895.cl"},
	{0xDF6BBAA7B3A1CF71ULL, "vxlt_undercitydungeon_20f_0_0.vlm"},
	{0xDF6C019992D2F463ULL, "t3_wdg_w28.tnt"},
	{0xDF6C5E3F3C8D734EULL, "267415234.ogg"},
	{0xDF6C6445115BF201ULL, "3450249.ogg"},
	{0xDF6C823186DD6543ULL, "vxca_caverns2_8k4_0_0.vlm"},
	{0xDF6CA9CCD0A6414DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_lgta_0n.dds"},
	{0xDF6CD569127030CCULL, "encounter.gda"},
	{0xDF6DC0AAC55A3FE1ULL, "ave200_donnic6111992.cl"},
	{0xDF6DE4E728F74B69ULL, "6153161_m.fxe"},
	{0xDF6E56588F09D26AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0dd_galoframe.mao"},
	{0xDF6E5A588F09D8B9ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0dd_galoframe.mmh"},
	{0xDF6E64588F09E9A7ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0dd_galoframe.msh"},
	{0xDF6EB75686D5E6A2ULL, "vxca_caverns1_mines_j4x_0_0.vlm"},
	{0xDF6F14CFA9ACE169ULL, "vxht_0_hightown_a2_b9m_0_0.vlm"},
	{0xDF6F4879D63465A1ULL, "6085931_f.fxe"},
	{0xDF6F7E329D887085ULL, "dre000_varric6207995.cl"},
	{0xDF6FDCFA212E8D20ULL, "424660028.ogg"},
	{0xDF70541CA5092CA0ULL, "vxca_caverns2prc_78p_0_0.vlm"},
	{0xDF7071389935F3C7ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.impa.gad"},
	{0xDF70C0628AEEEE40ULL, "6061827_m.fxe"},
	{0xDF7167C4AF6EF8C4ULL, "plt_lgt000pt_ultra_3_01.nss"},
	{0xDF7177914EB75217ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ac_transition.gad"},
	{0xDF71D7F7B2989976ULL, "vxw_bonepitdestroyed_a3_1zw_0_0.vlm"},
	{0xDF7267B1A7BB1694ULL, "6169288_m.fxe"},
	{0xDF726F5EF1FFE34FULL, "6208391_m.fxe"},
	{0xDF72AEAD2518D812ULL, "pro000ar_blightlands_fake.are"},
	{0xDF72B3CDE8A76B19ULL, "qaitest.arl"},
	{0xDF72DA6017DEC1E4ULL, "hlac_w_blightlands_65536_0z.dds"},
	{0xDF737EA1E9408101ULL, "llac_ca_cavern1_haunted_24423_0.dds"},
	{0xDF73B37CDC6BAFEEULL, "follower_banter6173060.cl"},
	{0xDF73E58AF365BEE0ULL, "hlac_lt_gallowsprison_n_-1_6y.dds"},
	{0xDF7448407DAAE852ULL, "1139574.ogg"},
	{0xDF748C1184BBB8D7ULL, "109102990.ogg"},
	{0xDF7547A88DE00CA2ULL, "hlac_ht_0_hightown_n_65534_0y.dds"},
	{0xDF756FABC02356B2ULL, "345085177.ogg"},
	{0xDF759D337CA0C07CULL, "vxht_0_hightown_a4p_0_0.vlm"},
	{0xDF75D9870B046E5DULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\horror.ani"},
	{0xDF762CE3857AF16DULL, "llac_w_sundermount_65539_0.dds"},
	{0xDF76461F9D23256CULL, "6017501_m.fxe"},
	{0xDF7683760E9E0649ULL, "dae100_bran6118288.cl"},
	{0xDF76FD31213D208FULL, "134750829.ogg"},
	{0xDF7725FFD4624145ULL, "6149089_m.fxe"},
	{0xDF776638993BFFD4ULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.impa.evt"},
	{0xDF777B6AE31FB35DULL, "mrl150_merrill6202026.cl"},
	{0xDF77F6318B9A6852ULL, "vfx_chunkyparticle_p3.vsh.fxo.dx11"},
	{0xDF780AF1C64DDC4BULL, "hlac_lt_0_lowtown_a3_n_c_131072_3y.dds"},
	{0xDF78648D819947ACULL, "953335875.ogg"},
	{0xDF78829024B4719FULL, "vfx_cyc_air_add.mao"},
	{0xDF788C914EBD94A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ac_transition.evt"},
	{0xDF78945DCEB076A2ULL, "dae000_bartrand6027523.cl"},
	{0xDF78BE9AA71CED64ULL, "vxlt_0_lowtown_docks_a3_n_cyq_0_1.vlm"},
	{0xDF790CDFD0368408ULL, "26202899.ogg"},
	{0xDF7973C0ED886F2DULL, "hlac_lt_0_lowtown_a3_n_c_65537_2x.dds"},
	{0xDF798986686ECE94ULL, "zz_approval_debug6003830.cl"},
	{0xDF79FB02F4D68421ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_exit.ani"},
	{0xDF7A755DB85B2F75ULL, "vxca_caverns1_ehu_0_0.vlm"},
	{0xDF7B5B7FDABBE015ULL, "plt_lgt310pt_kind_want.nss"},
	{0xDF7C35E824DA9BC8ULL, "837298504.ogg"},
	{0xDF7C9A5199758423ULL, "41258472.ogg"},
	{0xDF7CA5AD25211DC5ULL, "pro000ar_blightlands_fake.lst"},
	{0xDF7CD29CE41C95B3ULL, "vxht_chantry_n_2il_0_0.vlm"},
	{0xDF7D293378CBFF22ULL, "vxca_caverns1_mines_j3y_0_0.vlm"},
	{0xDF7D45AB348C3682ULL, "gen00fl_merrill6162254.cl"},
	{0xDF7DDC8B0971FB71ULL, "gen00fl_aveline6132747.cl"},
	{0xDF7E00C1588AABF1ULL, "6082105_f.fxe"},
	{0xDF7E3FF167A4CE0AULL, "character_character_p216.psh.fxo.dx11"},
	{0xDF7F0D386A50925DULL, "genip_corpse1.utp"},
	{0xDF7F11F4AF531F91ULL, "hlac_w_encounter3_-131075_0z.dds"},
	{0xDF7F4FCBF70E7771ULL, "weapon_weapon_p2.psh.fxo"},
	{0xDF7FA390AA35F265ULL, "892327994.ogg"},
	{0xDF809D682ABE8441ULL, "follower_banter6172963.cl"},
	{0xDF809D71B46190D0ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic.mao"},
	{0xDF80A171B46197E3ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic.mmh"},
	{0xDF80AF71B461AF3DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic.msh"},
	{0xDF80D0EDB8778CCFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_hands_behind_back.anb"},
	{0xDF80D155888079A6ULL, "vxht_0_hightown_a3_bys_0_0.vlm"},
	{0xDF810602F4DCB5C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_exit.gad"},
	{0xDF81791681DB2B0DULL, "vxlt_0_lowtown_a2_n_c_gmy_0_0.vlm"},
	{0xDF81A65E0063FC1CULL, "vxw_encounter1_ez_0_0.vlm"},
	{0xDF81BC9520399EA9ULL, "vxht_chantry_n_1ym_0_0.vlm"},
	{0xDF81D10A75B04C47ULL, "6111275_m.fxe"},
	{0xDF81FB80231CA1EEULL, "vxht_keep_3y3_0_0.vlm"},
	{0xDF82047D93A0173EULL, "gen_im_npc_elf_arm_glv_dalc.uti"},
	{0xDF826980C3D29A79ULL, "10419820.ogg"},
	{0xDF826B1D615289FCULL, "vxlt_gallowstemplar_n_3nl_0_0.vlm"},
	{0xDF8273CE15C18089ULL, "vxht_keep_n_6lu_0_0.vlm"},
	{0xDF827C038E89653EULL, "443698693.ogg"},
	{0xDF839801625B2CDCULL, "dre000_cut_arrivalprimeval.cnv"},
	{0xDF83BDAD25276554ULL, "pro000ar_blightlands_fake.ncs"},
	{0xDF83CE490B4357FCULL, "955657337.ogg"},
	{0xDF83D377AE8FF480ULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.mmh"},
	{0xDF840D1763E642A4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct1window.msh"},
	{0xDF841F1763E6610DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct1window.mao"},
	{0xDF84231763E66846ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct1window.mmh"},
	{0xDF845E3899467C0DULL, "art\\vfx\\cinematic\\vfx_fire_strip_p.impa.ani"},
	{0xDF84A71C870C108BULL, "vxlt_warehouse_n_12p_0_0.vlm"},
	{0xDF84CBD1772DA9D6ULL, "vxw_encounter1_b9_0_0.vlm"},
	{0xDF8592A3B47F261FULL, "dae011_drunk1_sta6149679.cl"},
	{0xDF85A4914EC8479DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_ac_transition.ani"},
	{0xDF86846AF8E18B7CULL, "6103288_m.fxe"},
	{0xDF86A7A3FE44359DULL, "6074064_m.fxe"},
	{0xDF86BA01625D9C2CULL, "dre000_cut_arrivalprimeval.bnk"},
	{0xDF86D0D45DF63AC4ULL, "59838689.ogg"},
	{0xDF87B6230C9AE28FULL, "49602400.ogg"},
	{0xDF87E302F4E298A8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\hands_on_face_exit.evt"},
	{0xDF8852DC2AED5482ULL, "6199285_m.fxe"},
	{0xDF888F434727BCBCULL, "648084442.ogg"},
	{0xDF88C4B45DEFAFBBULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbarout.mmh"},
	{0xDF88D0B45DEFC458ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbarout.mao"},
	{0xDF88E2B45DEFE2A5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarpbarout.msh"},
	{0xDF88F633E764D5F4ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_4x.dds"},
	{0xDF8927B5E78967BBULL, "vxht_mansion_n_47l_0_0.vlm"},
	{0xDF8A08AA28C69CC2ULL, "6135067_m.fxe"},
	{0xDF8A4137C13FA230ULL, "6047473_f.fxe"},
	{0xDF8A5EFADB59D233ULL, "6157084_m.fxe"},
	{0xDF8A9F319B9D9EFEULL, "art\\characters\\creatures\\shriek\\animation\\combat\\ready.evt"},
	{0xDF8AF11A81D1DE26ULL, "7331156.ogg"},
	{0xDF8B219B1E69E675ULL, "189593110.ogg"},
	{0xDF8B89CD137FB250ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_ornate.phy"},
	{0xDF8C13D1B1FF069AULL, "dre000_dougal6129756.cl"},
	{0xDF8C2C9E32B19C5AULL, "dae000_pc_home_ambients6208297.cl"},
	{0xDF8D7C0A2F05374AULL, "735504262.ogg"},
	{0xDF8DE964DD01E0B8ULL, "6173672_m.fxe"},
	{0xDF8E341CF0C4DA17ULL, "6107532_f.fxe"},
	{0xDF8E4C1C207A34AAULL, "hlac_lt_warehouse_n_65537_0x.dds"},
	{0xDF8E84D67189B609ULL, "vxlt_0_lowtown_a2_eir_0_0.vlm"},
	{0xDF8F0B9A993F24A7ULL, "685913151.ogg"},
	{0xDF8F18F408C5D437ULL, "hlac_lt_backalley_g_0_5x.dds"},
	{0xDF8F1A1763F06011ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_oct1window.phy"},
	{0xDF8F416A43F6BD5DULL, "vxht_brothel_n_6d6_0_0.vlm"},
	{0xDF8F70B710CCEDEAULL, "vxlt_undercityencounter_us_0_0.vlm"},
	{0xDF8FCBE9656B385CULL, "w_woundedcoast_202_0.phy"},
	{0xDF90B677AE9A4D9FULL, "art\\vfx\\combat\\vfx_m_cre_paralysis_c.anb"},
	{0xDF90E5AB07B79909ULL, "vxht_keep_n_6gt_0_0.vlm"},
	{0xDF91296FEF6D6626ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\additive_damage.ani"},
	{0xDF91A271B4700C08ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wo_basic.phy"},
	{0xDF91B6319BA3E429ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\ready.gad"},
	{0xDF91F71CCAFF87DEULL, "adder_of_antiva.dds"},
	{0xDF92067A5E663DD5ULL, "fen210_hadriana6110398.cl"},
	{0xDF929139F6997944ULL, "51915618.ogg"},
	{0xDF92ACAED01A6657ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v3.gad"},
	{0xDF93125E6071A2BAULL, "426148676.ogg"},
	{0xDF9368C4AF8BE0BFULL, "plt_lgt000pt_ultra_3_01.plo"},
	{0xDF93DEC9A1EE6597ULL, "vxlt_0_lowtown_docks_n_b4z_0_0.vlm"},
	{0xDF93F1B279FAB4EFULL, "6207708_f.fxe"},
	{0xDF9417C52A7300A4ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha13a.phy"},
	{0xDF94779681EBB15FULL, "ser121_jethann6175922.cl"},
	{0xDF9552106B194849ULL, "6188650_f.fxe"},
	{0xDF958F584DB8D0A7ULL, "cod_bks_chantry_hist_4.utp"},
	{0xDF95E43CEED2D742ULL, "examine.cur"},
	{0xDF95F4C36B54B7EAULL, "character_character_p89.psh.fxo"},
	{0xDF966F0BFC0719ACULL, "vxlt_undercityserialkiller_1j2_0_1.vlm"},
	{0xDF9676296797C55AULL, "vxlt_darktown_47e_0_0.vlm"},
	{0xDF96DDA05DA4A7FCULL, "6119683_m.fxe"},
	{0xDF979F9298BB40A0ULL, "vxca_caverns2_73i_0_1.vlm"},
	{0xDF9887319BA9B377ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\ready.ani"},
	{0xDF988A45A552F05FULL, "vxlt_hangedman_1t9_0_0.vlm"},
	{0xDF98D280D96ED002ULL, "w_sundermount_172_0.phy"},
	{0xDF9955244FBAB919ULL, "6090318_m.fxe"},
	{0xDF999FC78C926967ULL, "mer240_jansen6050417.cl"},
	{0xDF99C1AED020A8E4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v3.evt"},
	{0xDF99EA5D148BD3CDULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga103a_0t.dds"},
	{0xDF9A04D29F02849EULL, "vxdr_primevaltomb_7uf_0_0.vlm"},
	{0xDF9A271997824327ULL, "6121131_m.fxe"},
	{0xDF9AA13BE01E8E05ULL, "seb201_lady_harimann6177353.cl"},
	{0xDF9ACF9376E8755BULL, "6048397_m.fxe"},
	{0xDF9AD751D74FD2D3ULL, "454999351.ogg"},
	{0xDF9B1B2FAB46DB57ULL, "w_bonepitmining_a2_250_0.mmh"},
	{0xDF9B1F2FAB46E264ULL, "w_bonepitmining_a2_250_0.mao"},
	{0xDF9B312FAB4700C9ULL, "w_bonepitmining_a2_250_0.msh"},
	{0xDF9B4A43692882A4ULL, "seb321_leliana6180909.cl"},
	{0xDF9B58B213783B5EULL, "gencr_slaver_2h_f.utc"},
	{0xDF9B9D1EA3CBCAFAULL, "rock_jagged_big.msh"},
	{0xDF9BAF1EA3CBE940ULL, "rock_jagged_big.mmh"},
	{0xDF9BFDB8162F449EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medp_0d.dds"},
	{0xDF9C59F311C960ECULL, "vxht_keep_4bs_0_0.vlm"},
	{0xDF9CD9F54E8A0745ULL, "mag291st_demon_marethari.stg"},
	{0xDF9D15F3092AB75BULL, "vxlt_gallowstemplar_f_473_0_0.vlm"},
	{0xDF9D25A782EC6A70ULL, "atl_startmenu_dxt5_02.dds"},
	{0xDF9D2D8E887AE2F6ULL, "vxw_bonepitdestroyed_a3_1ez_0_0.vlm"},
	{0xDF9D5A7FDAD8C576ULL, "plt_lgt310pt_kind_want.plo"},
	{0xDF9D7190AE0FDFE1ULL, "89909460.ogg"},
	{0xDF9D743BE93957D9ULL, "6038594_m.fxe"},
	{0xDF9DB50F6C6CAAA6ULL, "vxht_0_hightown_a3_bsu_0_0.vlm"},
	{0xDF9DC53326F0AE86ULL, "hlac_lt_gallowsprison_n_-1_1z.dds"},
	{0xDF9E116FEF77C797ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\additive_damage.evt"},
	{0xDF9EB38EE6E5645CULL, "448800956.ogg"},
	{0xDF9FBDF3D77D0920ULL, "dae100_alistair_drunk6135160.cl"},
	{0xDF9FD7B44CAB0AC3ULL, "gen00fl_aveline6152549.cl"},
	{0xDFA029598E53ACA6ULL, "vxlt_0_lowtown_a2_n_hgm_0_0.vlm"},
	{0xDFA033AFD27F5066ULL, "t3_cth_nb3.tnt"},
	{0xDFA0BA879562376EULL, "670131914.ogg"},
	{0xDFA11F50B8AD0119ULL, "6126891_m.fxe"},
	{0xDFA132A122FB11F6ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_06a_0s.dds"},
	{0xDFA13E1BA41AFD63ULL, "w_bonepitmining_a2_190_0.mao"},
	{0xDFA1421BA41B03B0ULL, "w_bonepitmining_a2_190_0.mmh"},
	{0xDFA1501BA41B1BCAULL, "w_bonepitmining_a2_190_0.msh"},
	{0xDFA193A4129F4EEEULL, "dae000_bartrand6206146.cl"},
	{0xDFA1D8DF9D55DCFEULL, "vxht_keep_638_0_0.vlm"},
	{0xDFA1EDB6D25F95BAULL, "mag360_thrask6053516.cl"},
	{0xDFA21EE330A5A258ULL, "204482453.ogg"},
	{0xDFA22A4D161255EFULL, "hlac_w_blightlands_65539_0x.dds"},
	{0xDFA30FE4E909D433ULL, "6080274_m.fxe"},
	{0xDFA33A4AC32D2315ULL, "and110_refugee6214661.cl"},
	{0xDFA368F5F3330A4EULL, "565357742.ogg"},
	{0xDFA3ECF3747C49A4ULL, "vxht_mansion2_25d_0_0.vlm"},
	{0xDFA3F1423B26E456ULL, "weapon_rendertoshadowmap_p0.psh.fxo.dx11"},
	{0xDFA424430D089A88ULL, "fx_dis_ring.mao"},
	{0xDFA466A05D4AD06AULL, "character_character_p78.psh.fxo"},
	{0xDFA47D2632EF2B5AULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv2.gad"},
	{0xDFA48BBEB9880EC8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a.msh"},
	{0xDFA49DBEB9882D59ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a.mao"},
	{0xDFA4A9BEB98841F2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a.mmh"},
	{0xDFA4D8A7503D6137ULL, "vxca_caverns2prc_a49_0_0.vlm"},
	{0xDFA4EA4E9210F56BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ready.ani"},
	{0xDFA4F50AEBCD05A3ULL, "dre000_sibling_has_blight6036503.cl"},
	{0xDFA5126FEF7DE820ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\additive_damage.gad"},
	{0xDFA524534225EA35ULL, "vxht_keep_1vc_0_0.vlm"},
	{0xDFA668C6EC93F09FULL, "dae000_mother6103034.cl"},
	{0xDFA68BDECC433296ULL, "6075653_m.fxe"},
	{0xDFA70CB135FD8347ULL, "6035295_f.fxe"},
	{0xDFA7D9AED02D0EDDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v3.ani"},
	{0xDFA8F732FBCFB4E3ULL, "gen00fl_dog6115124.cl"},
	{0xDFA9CF4A629C720CULL, "art\\vfx\\combat\\vfx_gib_hm_robc.msh"},
	{0xDFA9D54A629C7CFEULL, "art\\vfx\\combat\\vfx_gib_hm_robc.mmh"},
	{0xDFAA6350C046C837ULL, "vxlt_warehouse_15f_0_0.vlm"},
	{0xDFAA7DE08D15FCE3ULL, "6087091_m.fxe"},
	{0xDFAA991648A6350AULL, "hlac_dr_0_deeproads_5842_16x.dds"},
	{0xDFAA9BEECA31A166ULL, "one100dg_lia6055159.cl"},
	{0xDFAB09C05C307174ULL, "396057859.ogg"},
	{0xDFAB802632F54F71ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv2.evt"},
	{0xDFAC252427E2EE5DULL, "vxht_brothel_n_6j4_0_0.vlm"},
	{0xDFAC26EC5E62C326ULL, "vxht_0_hightown_a3_bnv_0_0.vlm"},
	{0xDFAC47B8389A579AULL, "zz_seb_debug6215780.cl"},
	{0xDFAC67B7E394861BULL, "416968065.ogg"},
	{0xDFACE26B26FA0450ULL, "ch1alphem.psh.fxo"},
	{0xDFACE941B817E3F9ULL, "vxht_playershouse_n_1qd_0_0.vlm"},
	{0xDFAD458C5E491B84ULL, "ave310tr_party_comment.ncs"},
	{0xDFAD4FFD50071B33ULL, "zz_followers_banter6158523.cl"},
	{0xDFADAD0C8FC01F8AULL, "dae200_saemus6091694.cl"},
	{0xDFADBDB56A20E123ULL, "llac_lt_ruinedhouse_n_-1_9.dds"},
	{0xDFADCB8163CA47C3ULL, "zz_mer_debug6180852.cl"},
	{0xDFADFDE0A7F3E206ULL, "vxht_0_hightown_a3_7iz_0_0.vlm"},
	{0xDFAE139B7708D12EULL, "lt_0_lowtown_52_0_lc.mmh"},
	{0xDFAE5A73F3F2BC8EULL, "mcr371_idunna6087280.cl"},
	{0xDFAF45591D8DA0D3ULL, "one300_nuncio6075146.cl"},
	{0xDFAF976B5983EB2DULL, "vxlt_gallowstemplar_n_480_0_0.vlm"},
	{0xDFAFA5A84FBABC71ULL, "vxlt_0_lowtown_n_gjb_0_0.vlm"},
	{0xDFAFBBD6F82A8771ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\_textures\\bsh_thorn_n.dds"},
	{0xDFAFBE8428801488ULL, "follower_banter6155475.cl"},
	{0xDFAFD8CC14208292ULL, "vxht_bartrand_n_6pi_0_0.vlm"},
	{0xDFAFE20CF2E109A6ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier2_epilogue.ani"},
	{0xDFB01FB4AA595172ULL, "vxlt_hangedman_n_25k_0_0.vlm"},
	{0xDFB080061F260B73ULL, "vxw_bonepit_2a4_0_0.vlm"},
	{0xDFB1363B12FD59EEULL, "vxlt_undercityserialkiller_xt_0_0.vlm"},
	{0xDFB19FDD16D25460ULL, "dae300_vashothlead6062782.cl"},
	{0xDFB1CABFA9841871ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxb_d.dds"},
	{0xDFB1D24E921B56E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ready.evt"},
	{0xDFB1DD75F71E68A2ULL, "621870512.ogg"},
	{0xDFB1EF3DAADD54B7ULL, "302710565.ogg"},
	{0xDFB2030BD40DEACBULL, "714941490.ogg"},
	{0xDFB21DDF3BE751D7ULL, "279417838.ogg"},
	{0xDFB22DA24A468F13ULL, "zz_vault_debug6203359.cl"},
	{0xDFB2BFC7F42E2A15ULL, "762665553.ogg"},
	{0xDFB359E34B83F65AULL, "vxlt_darktown_49g_0_0.vlm"},
	{0xDFB3615211E8F601ULL, "vxlt_0_lowtown_a3_n_c_ewx_0_0.vlm"},
	{0xDFB3C15A4190D8FCULL, "6113083_m.fxe"},
	{0xDFB4A5D5991C5E4AULL, "vxlt_0_lowtown_exv_0_0.vlm"},
	{0xDFB527FA14EE4869ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench2f_arishokaddress.gad"},
	{0xDFB537EB72F1C182ULL, "vxlt_hangedman_n_20d_0_0.vlm"},
	{0xDFB56CCB6AA7AFA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bd_transition.ani"},
	{0xDFB56D647A8CDDD2ULL, "zz_followers_banter6159156.cl"},
	{0xDFB587125E04472FULL, "hlac_w_sundermount_65535_0z.dds"},
	{0xDFB587C67CB59C97ULL, "154841514.ogg"},
	{0xDFB5EC6174C556F1ULL, "6051278_m.fxe"},
	{0xDFB63C22BA294530ULL, "gen00fl_varric6033679.cl"},
	{0xDFB73D7C39258F5FULL, "6173950_m.fxe"},
	{0xDFB799137A05B885ULL, "vxw_bonepitdestroyed_a3_ns_0_0.vlm"},
	{0xDFB7A52D267D72E4ULL, "6029045_m.fxe"},
	{0xDFB7C2A47045A22BULL, "80552012.ogg"},
	{0xDFB8982633000240ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\slavedropv2.ani"},
	{0xDFB8C2CB01CAAAA8ULL, "vxca_cavern1_haunted_jfh_0_0.vlm"},
	{0xDFB8D14E922173F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\ready.gad"},
	{0xDFB8D374A7A5B5FFULL, "438036685.ogg"},
	{0xDFB8E243AF3540C6ULL, "6108809_m.fxe"},
	{0xDFB904BCAC08284BULL, "6060203_m.fxe"},
	{0xDFB93BB3637F5302ULL, "lt_0_lowtown_docks_a3_n_198_0.msh"},
	{0xDFB93DB3637F56B8ULL, "lt_0_lowtown_docks_a3_n_198_0.mmh"},
	{0xDFB949B3637F6B5BULL, "lt_0_lowtown_docks_a3_n_198_0.mao"},
	{0xDFB9AE5C74314083ULL, "6110265_m.fxe"},
	{0xDFBA06C1B8D4ECDEULL, "183305190.ogg"},
	{0xDFBA2B949D66B0CAULL, "dae000_arishok6068590.cl"},
	{0xDFBA471CB2DADEE6ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\tre_tree02_sa.dds"},
	{0xDFBB5B1EAA0DFC79ULL, "vxht_playershouse_n_1tk_0_0.vlm"},
	{0xDFBBC7C4BBE5D0A7ULL, "lowtown_male_beggar_sta6151048.cl"},
	{0xDFBBD723C537F772ULL, "61066304.ogg"},
	{0xDFBBF8FA14F416B7ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_hench2f_arishokaddress.ani"},
	{0xDFBC103B5DE7DF70ULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.mmh"},
	{0xDFBC9DB48238812AULL, "hlac_w_blightlands_196608_0y.dds"},
	{0xDFBCEF6F2176A78AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basic.mao"},
	{0xDFBCF36F2176AED9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basic.mmh"},
	{0xDFBCFD6F2176BFC7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1wi_basic.msh"},
	{0xDFBDB97BCFF3DCA5ULL, "1052833450.ogg"},
	{0xDFBE5A20893B9782ULL, "vxlt_0_lowtown_a2_n_c_hpv_0_0.vlm"},
	{0xDFBE67127B404298ULL, "mer210_coterie_contact6079350.cl"},
	{0xDFBEAD80D98F60B6ULL, "w_sundermount_172_0.mao"},
	{0xDFBEB980D98F7555ULL, "w_sundermount_172_0.mmh"},
	{0xDFBEBB80D98F78CBULL, "w_sundermount_172_0.msh"},
	{0xDFBEEDA03E0D4610ULL, "vxlt_0_lowtown_a2_h1v_0_0.vlm"},
	{0xDFBFCA1647952ED0ULL, "298411300.ogg"},
	{0xDFC02F600E5723F3ULL, "gen_im_cft_reg_i.uti"},
	{0xDFC09D82CD679A92ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v2.gad"},
	{0xDFC0B8AC05724F39ULL, "art\\vfx\\combat\\vfx_a_stealth_att_2_c.mmh"},
	{0xDFC1573CE2BCB7A1ULL, "6081279_m.fxe"},
	{0xDFC1861BC4A6879EULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v1.evt"},
	{0xDFC1EEEAB4E774D7ULL, "follower_banter6143213.cl"},
	{0xDFC2199BE8703C52ULL, "199318250.ogg"},
	{0xDFC21C59827FDFC3ULL, "art\\characters\\playercharacter\\elfmale\\boo_nuda.msh"},
	{0xDFC22A59827FF71DULL, "art\\characters\\playercharacter\\elfmale\\boo_nuda.mmh"},
	{0xDFC22E59827FFE2EULL, "art\\characters\\playercharacter\\elfmale\\boo_nuda.mao"},
	{0xDFC23B05B90B735DULL, "vxht_0_hightown_a3_9bj_0_0.vlm"},
	{0xDFC287BD9C32E081ULL, "mrl310_merrill-walkdownmnt.cut"},
	{0xDFC289E827877C45ULL, "hlac_lt_0_lowtown_a3_n_c_1_6y.dds"},
	{0xDFC29BAEB7346F4FULL, "1627551.ogg"},
	{0xDFC2C4043828AAD7ULL, "vxw_woundedcoast_1e4_0_0.vlm"},
	{0xDFC2E2D07128028AULL, "gen00fl_anders6096076.cl"},
	{0xDFC377B038C67C90ULL, "gen00fl_anders6125431.cl"},
	{0xDFC384CB6AB4159AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bd_transition.evt"},
	{0xDFC3CB0CF2F18BA0ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_epilogue\\cs_soldier2_epilogue.gad"},
	{0xDFC4652138A2D133ULL, "vxlt_0_lowtown_a2_n_c_h2i_0_0.vlm"},
	{0xDFC4C832006D8403ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_act_c.impa.evt"},
	{0xDFC515315B2188DFULL, "vxw_encounter3_t_0_0.vlm"},
	{0xDFC54DEDF2CE0B7BULL, "ht_keep_n_3734.rml"},
	{0xDFC5809FB82CA886ULL, "619566122.ogg"},
	{0xDFC5AF592B1BA7DFULL, "6176650_m.fxe"},
	{0xDFC6225926FFED95ULL, "230286789.ogg"},
	{0xDFC6266B8C62460EULL, "117570249.ogg"},
	{0xDFC6744AD59795B5ULL, "hlac_ht_keep_n_3732_15x.dds"},
	{0xDFC79D1BC4AB1AC9ULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v1.gad"},
	{0xDFC7A082CD6DBEC9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v2.evt"},
	{0xDFC8BE7DF54D2055ULL, "dae000_foreshadow6179560.cl"},
	{0xDFC9B71C3025CC91ULL, "363111994.ogg"},
	{0xDFCA333B5DF4578FULL, "art\\vfx\\combat\\vfx_c_pride_f_shield_c.anb"},
	{0xDFCA4D1AC43488C9ULL, "character_character_p105.psh.fxo"},
	{0xDFCA83CB6ABA328DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bd_transition.gad"},
	{0xDFCAA4A2ABA2A498ULL, "6180444_f.fxe"},
	{0xDFCADE84EFF1E1EDULL, "6090332_m.fxe"},
	{0xDFCAF58F99A1E489ULL, "mag311st_huon.stg"},
	{0xDFCB5D4FC23B19F3ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_deeproads_d.dds"},
	{0xDFCB95BF1DF6A864ULL, "6072557_m.fxe"},
	{0xDFCBA13200736084ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_act_c.impa.gad"},
	{0xDFCBFE06799916F2ULL, "6125127_m.fxe"},
	{0xDFCC7C5DD35B9D85ULL, "art\\vfx\\textures\\vfx_blood_sh_si.dds"},
	{0xDFCCE91D5777C2FCULL, "6080732_m.fxe"},
	{0xDFCCEF29D61C6D5DULL, "ch2_ds2y.psh.fxo"},
	{0xDFCD7A4452B417E1ULL, "hlac_w_blightlands_0_5y.dds"},
	{0xDFCDA598FC0E06D1ULL, "153664318.ogg"},
	{0xDFCE6E1BC4B0E917ULL, "art\\characters\\creatures\\dragon\\animation\\dmg_v1.ani"},
	{0xDFCE743C332E0C92ULL, "hlac_ht_0_hightown_65535_3z.dds"},
	{0xDFCE9FD1DFAB5B6FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\additive\\head_forward.gad"},
	{0xDFCF5A3EA19FD7EFULL, "173965947.ogg"},
	{0xDFCFC919D4B2D021ULL, "6084179_m.fxe"},
	{0xDFCFDEDADC61997CULL, "6071252_f.fxe"},
	{0xDFD01913D4E2A3B8ULL, "vxlt_gallowstemplar_n_4qj_0_0.vlm"},
	{0xDFD0234D281EB061ULL, "vxht_brothel_n_5gx_0_0.vlm"},
	{0xDFD05B588F5CA4A6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0dd_galoframe.phy"},
	{0xDFD09715CB9C725EULL, "vxht_0_hightown_a3_b8x_0_0.vlm"},
	{0xDFD0B8A949FC4A2FULL, "390968990.ogg"},
	{0xDFD133DC15F935FBULL, "hla_ht_0_hightown_1_0x.dds"},
	{0xDFD153B2B26C2F09ULL, "lt_0_lowtown_docks_n_195_0.mao"},
	{0xDFD15FB2B26C43A2ULL, "lt_0_lowtown_docks_n_195_0.mmh"},
	{0xDFD161B2B26C4758ULL, "lt_0_lowtown_docks_n_195_0.msh"},
	{0xDFD18900ED9F2C41ULL, "6129788_m.fxe"},
	{0xDFD1E55C855EE1D5ULL, "608821745.ogg"},
	{0xDFD24970C7BB626FULL, "6144361_m.fxe"},
	{0xDFD2B03200799962ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_act_c.impa.ani"},
	{0xDFD3598A26DD9C0BULL, "vxht_0_hightown_a2_9zq_0_0.vlm"},
	{0xDFD4A79D2F43B82DULL, "6114869_m.fxe"},
	{0xDFD4AB80A9C9B861ULL, "follower_banter6136714.cl"},
	{0xDFD4ECA14BDA2B71ULL, "ave321cr_guardswoman.utc"},
	{0xDFD4F83BB181715BULL, "vxlt_gallowscourtyard_6wp_0_0.vlm"},
	{0xDFD54177BEEF39A1ULL, "vxht_mansion2_3go_0_1.vlm"},
	{0xDFD554C96C96CE63ULL, "zz_dre_debug6092017.cl"},
	{0xDFD58882CD79D338ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\idle\\twitch_v2.ani"},
	{0xDFD5CDF80CCBF1AEULL, "llac_ht_0_hightown_65532_0.dds"},
	{0xDFD618CF49B1BF69ULL, "607496476.ogg"},
	{0xDFD648578AEEEDA8ULL, "gencr_virvle_mage_m.utc"},
	{0xDFD791D23EDED531ULL, "hlac_ht_0_hightown_a2_-1_1y.dds"},
	{0xDFD7C48B765C1DBAULL, "vxdr_primevaltomb_709_0_0.vlm"},
	{0xDFD833F5A9CF6DB8ULL, "306869970.ogg"},
	{0xDFD890241D3341F6ULL, "vxlt_0_lowtown_docks_b5k_0_0.vlm"},
	{0xDFD899227867D3FDULL, "art\\characters\\weapons\\greataxes\\w_gax_107.mao"},
	{0xDFD8F35A7BBEAF00ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_avea_0d.dds"},
	{0xDFD95EB88A4C80C5ULL, "6112042_f.fxe"},
	{0xDFD970A0159A214AULL, "vxca_cave_1bt_0_0.vlm"},
	{0xDFDA24521E9A8532ULL, "vxht_0_hightown_c88_0_0.vlm"},
	{0xDFDA9D5CF67DD1F4ULL, "6180808_f.fxe"},
	{0xDFDAB3EB641AD77DULL, "vxht_mansion2b_21m_0_0.vlm"},
	{0xDFDB8C346F24590BULL, "vxlt_backalley_n_1e0_0_0.vlm"},
	{0xDFDB96A1C1B516DDULL, "gen00fl_fenris6115374.cl"},
	{0xDFDBEA199CE2A62AULL, "zz_isa_debug6219323.cl"},
	{0xDFDC21F6D48DCF76ULL, "6136949_m.fxe"},
	{0xDFDC9075C2B063A9ULL, "718866593.ogg"},
	{0xDFDE009B4EEF10DAULL, "556250845.ogg"},
	{0xDFDE0B49BF1258AEULL, "vxca_caverns1_mines_hi5_0_0.vlm"},
	{0xDFDE42B0E0555474ULL, "one280_emrys6197976.cl"},
	{0xDFDE71E0E2866B88ULL, "1025851361.ogg"},
	{0xDFDEC0EB129CA93BULL, "6151341_m.fxe"},
	{0xDFDEFA41AA6C6522ULL, "vxht_mansion2_31c_0_0.vlm"},
	{0xDFDF39D807DC7340ULL, "6064639_m.fxe"},
	{0xDFDFA7BE1C14DBD7ULL, "vxw_woundedcoast_1k6_0_0.vlm"},
	{0xDFDFE505B4C93F97ULL, "hlac_lt_0_lowtown_docks_-262144_0y.dds"},
	{0xDFE06C4AED5F5E6AULL, "ser200_aveline6201534.cl"},
	{0xDFE072B69862CCBAULL, "dae000_bodhan_letter6144254.cl"},
	{0xDFE075D629B554D9ULL, "vxht_mansion_n_5i1_0_0.vlm"},
	{0xDFE11DC1CEDCAA4FULL, "6100075_m.fxe"},
	{0xDFE133C7836D2590ULL, "hlac_lt_gallowscourtyard_n_-2_0z.dds"},
	{0xDFE16DEB8E15BC84ULL, "6094728_m.fxe"},
	{0xDFE1DB01301F1322ULL, "654054622.ogg"},
	{0xDFE25BA0743FF95AULL, "hlac_w_encounter2_camp_327682_0x.dds"},
	{0xDFE2985EC6EC9203ULL, "ave200_aveline_bar6116174.cl"},
	{0xDFE2BCD1DFBC3535ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\additive\\head_forward.ani"},
	{0xDFE2CF6573A39ACAULL, "vxca_cavern1_haunted_eh3_0_0.vlm"},
	{0xDFE2DE7F7568D7D6ULL, "159001345.ogg"},
	{0xDFE2E6FFBCE9A0D0ULL, "gen_im_arm_boo_med_act3_02.uti"},
	{0xDFE33FBA5E1203ADULL, "vxht_keep_6a5_0_0.vlm"},
	{0xDFE344CCBF8229F6ULL, "vxdr_primevaltomb_8xz_0_0.vlm"},
	{0xDFE37BEC48EA9E10ULL, "364486797.ogg"},
	{0xDFE3A89895BBD6DFULL, "6135174_m.fxe"},
	{0xDFE3EA4FA19E3F74ULL, "6080986_f.fxe"},
	{0xDFE3F6AADFB27230ULL, "vxca_cavern1_haunted_gkf_0_0.vlm"},
	{0xDFE4E341EC15ECF5ULL, "6090835_m.fxe"},
	{0xDFE4FBD91D1B5D52ULL, "629825583.ogg"},
	{0xDFE4FCE23C806869ULL, "6153312_m.fxe"},
	{0xDFE555226B8E06EFULL, "vxca_caverns1_mines_hze_0_0.vlm"},
	{0xDFE677F3563B63EEULL, "mcr310_argument6074348.cl"},
	{0xDFE6BC3911D4CCE0ULL, "6104447_m.fxe"},
	{0xDFE70BE3BD7D1F09ULL, "513336062.ogg"},
	{0xDFE7825664FACA18ULL, "vxlt_0_lowtown_a2_gvl_0_0.vlm"},
	{0xDFE784877982C5F3ULL, "106422228.ogg"},
	{0xDFE801FEFCE5851BULL, "444990117.ogg"},
	{0xDFE81DDFDB6974C8ULL, "vxht_brothel_n_5jz_0_0.vlm"},
	{0xDFE86BAF9A841366ULL, "vxdr_primevaltomb_73x_0_0.vlm"},
	{0xDFE87BDC3BE9ABBDULL, "74078287.ogg"},
	{0xDFE8B5AC05939C2EULL, "art\\vfx\\combat\\vfx_a_stealth_att_2_c.anb"},
	{0xDFE94D96AE17CD3EULL, "6045979_m.fxe"},
	{0xDFE955F1C6FFC6C8ULL, "dae000cr_grand_cleric6124772.cl"},
	{0xDFE99EE6954781B7ULL, "1033156611.ogg"},
	{0xDFE9EBB0CFB55546ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightb_dp.dds"},
	{0xDFE9F130AAD5326DULL, "vxdr_primevaltomb_55g_0_0.vlm"},
	{0xDFEA581FD71D51D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_walk_to_bed.gad"},
	{0xDFEAAA0EB47C27CFULL, "gen00fl_anders6196821.cl"},
	{0xDFEABBA567AA3BD9ULL, "vxca_caverns1_mines_fm8_0_0.vlm"},
	{0xDFEBAE9A4ABB501FULL, "vxht_0_hightown_a3_aui_0_0.vlm"},
	{0xDFEBFB1E2B5CCAB0ULL, "407918621.ogg"},
	{0xDFEC097917614A34ULL, "vxlt_warehouse_n_ua_0_0.vlm"},
	{0xDFEC5B1A017EAB75ULL, "vxlt_backalley_g_1w1_0_0.vlm"},
	{0xDFEC5DA00F04A695ULL, "darktown_fem_elf_sta6174528.cl"},
	{0xDFED4F3E8F8DBB7FULL, "vxca_caverns2prc_agd_0_0.vlm"},
	{0xDFED55B983000489ULL, "gen00fl_anders6112980.cl"},
	{0xDFED5E5C51F49EA2ULL, "follower_banter6136647.cl"},
	{0xDFED66810C26AB88ULL, "553278804.ogg"},
	{0xDFED6C1E9C627DA2ULL, "vxht_mansion2_34l_0_0.vlm"},
	{0xDFED745F7EB51FAFULL, "pro000_opening6178524.cl"},
	{0xDFED7FBD658A2C72ULL, "mcr370_messenger6071828.cl"},
	{0xDFEDCF675210CB2BULL, "gen00fl_anders6125108.cl"},
	{0xDFEE3960DEC3E36DULL, "95334551.ogg"},
	{0xDFEED5CB93D08955ULL, "6099133_m.fxe"},
	{0xDFEEE3B10E0DF299ULL, "ave200_donnic6111989.cl"},
	{0xDFEEE684A6A894F5ULL, "203354264.ogg"},
	{0xDFEF29D2C79ED92EULL, "6095000_m.fxe"},
	{0xDFF09470FC81286DULL, "ht_0_hightownexplosion_n_12_0_ll.mmh"},
	{0xDFF0AAE1D1BF26B0ULL, "dae000_meredith6085338.cl"},
	{0xDFF0EF0751056390ULL, "6151287_m.fxe"},
	{0xDFF155A5B39772F8ULL, "hlac_ca_caverns1_24427_4x.dds"},
	{0xDFF16D1FD7239464ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_walk_to_bed.evt"},
	{0xDFF17D32B2D3E8EEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_medp_0t.dds"},
	{0xDFF18C854C76CBFEULL, "6133818_m.fxe"},
	{0xDFF1B19750081C2DULL, "vxht_keep_6j4_0_0.vlm"},
	{0xDFF1DBF5956DA25BULL, "vxlt_gallowscourtyard_6yr_0_0.vlm"},
	{0xDFF2056BB13C7C19ULL, "6162781_m.fxe"},
	{0xDFF23A9B96D47DF4ULL, "one300_nuncio6076780.cl"},
	{0xDFF2965D6DE9EE0EULL, "6021353_m.fxe"},
	{0xDFF3551EDE0C0575ULL, "6095735_m.fxe"},
	{0xDFF3597B565F6C4AULL, "6169484_m.fxe"},
	{0xDFF36EBF2E7680E9ULL, "6156012_m.fxe"},
	{0xDFF3C6FF5D841F6FULL, "vxca_caverns1_mines_hud_0_0.vlm"},
	{0xDFF3F9639C343620ULL, "vxlt_0_lowtown_n_gij_0_0.vlm"},
	{0xDFF409179C72D383ULL, "vxht_mansion_n_6ls_0_0.vlm"},
	{0xDFF4560F1785C63BULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsackwheat_0.pwk"},
	{0xDFF45CAB5E2EA789ULL, "vxlt_gallowstemplar_393_0_0.vlm"},
	{0xDFF4E3F603077ED3ULL, "116052390.ogg"},
	{0xDFF50486FC2E0E39ULL, "48451144.ogg"},
	{0xDFF529ED692A5615ULL, "lgt210_shp_kill_qunari36156879.cl"},
	{0xDFF5F43356F0E298ULL, "vxlt_0_lowtown_a2_gmm_0_0.vlm"},
	{0xDFF62030CED7E08FULL, "hlac_lt_0_lowtown_a2_n_196607_0x.dds"},
	{0xDFF63A76940509F1ULL, "vxca_caverns2prc_7qj_0_0.vlm"},
	{0xDFF6799FB3011C9CULL, "tables_sa.dds"},
	{0xDFF6B7BB8E591D6FULL, "vxca_caverns1_j2d_0_2.vlm"},
	{0xDFF7597747E7916BULL, "w_encounter2_5_0_lc.mmh"},
	{0xDFF7F40854B34815ULL, "308838333.ogg"},
	{0xDFF8218C0E6D3B35ULL, "913433208.ogg"},
	{0xDFF86FEE53108A0BULL, "vxlt_backalley_n_1k2_0_0.vlm"},
	{0xDFF87F6823552CFFULL, "79873216.ogg"},
	{0xDFF89DEC8ED38AA7ULL, "hlac_w_sundermount_2_0x.dds"},
	{0xDFF8C6E3FCF7A862ULL, "mrl000_marethari6154506.cl"},
	{0xDFF9810565C03F22ULL, "vxlt_warehouse_qr_0_0.vlm"},
	{0xDFF9985E65AB1958ULL, "vxlt_0_lowtown_a3_n_c_fex_0_0.vlm"},
	{0xDFFA3DB3FCCC4158ULL, "llac_lt_0_lowtown_a2_n_c_131073_1.dds"},
	{0xDFFA3E3D0B5D7334ULL, "6088255_m.fxe"},
	{0xDFFA7B56095762B4ULL, "vxlt_warehouse_n_xf_0_0.vlm"},
	{0xDFFA8E39E73F3544ULL, "w_bonepitmining_a2_145_0.msh"},
	{0xDFFAA039E73F53ADULL, "w_bonepitmining_a2_145_0.mao"},
	{0xDFFAA439E73F5AE6ULL, "w_bonepitmining_a2_145_0.mmh"},
	{0xDFFB2B37AFF10346ULL, "6099174_m.fxe"},
	{0xDFFB5D0D566E4B8BULL, "hlac_lt_0_lowtown_a2_-2_5x.dds"},
	{0xDFFB6E471D114275ULL, "215612707.ogg"},
	{0xDFFBB8D5D41CEF41ULL, "art\\characters\\creatures\\stonegolem\\animation\\die_v1.ani"},
	{0xDFFC18F2183829BBULL, "vxlt_0_lowtown_a3_n_c_e1p_0_0.vlm"},
	{0xDFFC2E6C5F90F714ULL, "193447398.ogg"},
	{0xDFFC8B7800010CD7ULL, "vxw_woundedcoast_1q0_0_0.vlm"},
	{0xDFFCC765B4EE4394ULL, "6112065_f.fxe"},
	{0xDFFCE63A02F41BAEULL, "vxlt_0_lowtown_docks_a3_n_bsi_0_0.vlm"},
	{0xDFFD79379D623D5FULL, "vxht_mansion_n_29x_0_0.vlm"},
	{0xDFFF2BAAC236924FULL, "6161449_m.fxe"},
	{0xDFFF7A909E74D34EULL, "mcr371_post_harvester6083688.cl"},
	{0xDFFF851FD72FFA5DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_action_walk_to_bed.ani"},
	{0xDFFF962CCCFEFB05ULL, "llac_w_sundermount_196613_0.dds"},
	{0xDFFFA8BEB9D4EBEDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile3a.phy"},
	{0xE0002C33E7A558DFULL, "lt_0_lowtown_docks_a3_n_142_0.msh"},
	{0xE0003233E7A56281ULL, "lt_0_lowtown_docks_a3_n_142_0.mmh"},
	{0xE0003E33E7A57722ULL, "lt_0_lowtown_docks_a3_n_142_0.mao"},
	{0xE0007748A3329499ULL, "6167481_m.fxe"},
	{0xE000B9F860467362ULL, "one300dg_solivitus6060366.cl"},
	{0xE00111F0E09E7DDBULL, "20636646.ogg"},
	{0xE0013FF69AB22F78ULL, "mag121_idunna6084113.cl"},
	{0xE0023E0BB0725CE1ULL, "one130_thug6131577.cl"},
	{0xE002C3D5D42320E3ULL, "art\\characters\\creatures\\stonegolem\\animation\\die_v1.gad"},
	{0xE00391833795BD50ULL, "w_encounter3_217_0.mao"},
	{0xE00395833795C463ULL, "w_encounter3_217_0.mmh"},
	{0xE003A3833795DBBDULL, "w_encounter3_217_0.msh"},
	{0xE003F8CB2D932391ULL, "374790209.ogg"},
	{0xE00432A8DEFB750FULL, "813583038.ogg"},
	{0xE00442F74DF05DDFULL, "vxw_encounter2_camp_72_0_0.vlm"},
	{0xE00463F80135779BULL, "zz_followers_banter6159726.cl"},
	{0xE004AC5385FB2271ULL, "vxca_caverns2prc_7ti_0_0.vlm"},
	{0xE004EA101A607B56ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_lbrt.anb"},
	{0xE005575C9615325AULL, "vxlt_warehouse_n_ce_0_0.vlm"},
	{0xE005C831E08DE74CULL, "art\\levels\\races\\qunari\\props\\_textures\\prp_qunaritent_n.dds"},
	{0xE005C91700388DF6ULL, "llac_lt_0_lowtown_docks_-3_8.dds"},
	{0xE005F59226EC2899ULL, "vxht_chantry_n_21f_0_0.vlm"},
	{0xE006A918EABAEF74ULL, "dre180cr_drake.utc"},
	{0xE0070173E0F3F739ULL, "6173769_m.fxe"},
	{0xE0075F9E1F4EF214ULL, "w_woundedcoast_31_0.rml"},
	{0xE007994AC9968020ULL, "6072377_m.fxe"},
	{0xE0079C6C9C9E5462ULL, "43436338.ogg"},
	{0xE007BFD4925E8643ULL, "w_bonepitdestroyed_a3_193_0.msh"},
	{0xE007C787D50A01B1ULL, "6119286_m.fxe"},
	{0xE007CDD4925E9D9DULL, "w_bonepitdestroyed_a3_193_0.mmh"},
	{0xE007D1D4925EA4AEULL, "w_bonepitdestroyed_a3_193_0.mao"},
	{0xE007F33C16477304ULL, "w_encounter3_186_0.msh"},
	{0xE008053C1647926DULL, "w_encounter3_186_0.mao"},
	{0xE008093C164798A6ULL, "w_encounter3_186_0.mmh"},
	{0xE008117473DB11F5ULL, "vxw_bonepitmining_a2_1jj_0_0.vlm"},
	{0xE00833F67A46A982ULL, "232243835.ogg"},
	{0xE0083F20FCDA5F4FULL, "zz_fen_debug6134366.cl"},
	{0xE0089CBF2643364AULL, "12470778.ogg"},
	{0xE0089FD9DCECD460ULL, "vxlt_0_lowtown_docks_n_cqc_0_0.vlm"},
	{0xE0096C0A9E2602C1ULL, "vxca_caverns1_mines_e8d_0_0.vlm"},
	{0xE009A0D5D42903C8ULL, "art\\characters\\creatures\\stonegolem\\animation\\die_v1.evt"},
	{0xE00ADCA5FD1410E4ULL, "6180901_f.fxe"},
	{0xE00B2295A4408133ULL, "594555695.ogg"},
	{0xE00BCAAD37B9D28AULL, "359062289.ogg"},
	{0xE00BCDAA3204A716ULL, "736082901.ogg"},
	{0xE00BCDC48047E976ULL, "595527614.ogg"},
	{0xE00BE5DEB1D304B7ULL, "vxw_woundedcoast_1g2_0_0.vlm"},
	{0xE00C0D1D00A467EDULL, "vxw_bonepit_yj_0_0.vlm"},
	{0xE00C18AA0C34D6CAULL, "6034023_m.fxe"},
	{0xE00C7F2040C9BEAAULL, "follower_banter6169975.cl"},
	{0xE00C8E3E902013E1ULL, "6161414_m.fxe"},
	{0xE00D179B22F8D3D8ULL, "6120009_m.fxe"},
	{0xE00D3710CF3CA02BULL, "vxht_mansion2b_32o_0_0.vlm"},
	{0xE00D385D467749F0ULL, "dae000_bodhan6146822.cl"},
	{0xE00D44765321E0BFULL, "character_character_p101.psh.fxo.dx11.assembly"},
	{0xE00D9354360877BEULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauserunf.evt"},
	{0xE00DA19F43144803ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_exit.ani"},
	{0xE00DA297AD3AEE37ULL, "vxlt_darktown_54p_0_0.vlm"},
	{0xE00DD4063A40D0D0ULL, "dae000_flemeth6001896.cl"},
	{0xE00ECCBB74B9BCD5ULL, "gen00fl_isabela6079564.cl"},
	{0xE00EE16587AA81B3ULL, "llac_lt_0_lowtown_a2_-65536_0.dds"},
	{0xE00F9A4928477410ULL, "6000794_m.fxe"},
	{0xE01031592209DD67ULL, "675076978.ogg"},
	{0xE0103D353A4B9D4AULL, "6163684_m.fxe"},
	{0xE010BE6A9A0B2AD6ULL, "vxlt_0_lowtown_f1v_0_0.vlm"},
	{0xE010DD1D80206720ULL, "vxlt_0_lowtown_n_gch_0_0.vlm"},
	{0xE010F469FA17B13BULL, "em_amb_tomwise.mor"},
	{0xE01154E5AF4CE045ULL, "hlac_lt_0_lowtown_-3_4x.dds"},
	{0xE0116C9E74ECA6D5ULL, "qun120cr_javaris6042440.cl"},
	{0xE011E1EACCAC3B6FULL, "1007887057.ogg"},
	{0xE0123F48C509A891ULL, "36576842.ogg"},
	{0xE013128CD375DF8AULL, "gen00fl_anders6114495.cl"},
	{0xE013A3C7C2C2A4EDULL, "242014291.ogg"},
	{0xE013D380EA231141ULL, "6133871_m.fxe"},
	{0xE013FFD307E86F55ULL, "25930992.ogg"},
	{0xE0144E43205EBE8EULL, "6131092_m.fxe"},
	{0xE0144FAD327B770FULL, "hlac_lt_0_lowtown_a2_n_-3_2y.dds"},
	{0xE0153002EEDDD219ULL, "hlac_lt_0_lowtown_n_-65535_4z.dds"},
	{0xE01657C906C38A65ULL, "hlac_ht_0_hightown_n_-196610_0y.dds"},
	{0xE0171A6D285E5C18ULL, "ip_restrict_follower_avelin.ncs"},
	{0xE0174424B410E561ULL, "6037865_m.fxe"},
	{0xE0197F969A719F8BULL, "454666129.ogg"},
	{0xE019C15392F7E0ACULL, "hlac_lt_0_lowtown_docks_a3_n_-131072_0y.dds"},
	{0xE019FF2807EC88D9ULL, "415593862.ogg"},
	{0xE01A28E39A09897FULL, "823773801.ogg"},
	{0xE01A35F1DD7EEE20ULL, "gen_im_trs_gem_snowobsid_02.uti"},
	{0xE01A7EF9F29A806DULL, "vxw_bonepit_di_0_0.vlm"},
	{0xE01A87FC057952C8ULL, "gen00fl_varric6093611.cl"},
	{0xE01A9B60E72F0675ULL, "6065915_f.fxe"},
	{0xE01AB99F431EFBFAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_exit.evt"},
	{0xE01AE0877069D285ULL, "166465265.ogg"},
	{0xE01B06241D629B97ULL, "vxlt_hangedman_n_t1_0_0.vlm"},
	{0xE01B34B812C53C0EULL, "6079150_m.fxe"},
	{0xE01B354DE722629EULL, "hlac_ca_caverns1_24427_5y.dds"},
	{0xE01B54D380811C52ULL, "6048824_m.fxe"},
	{0xE01B8D4E6A7A530BULL, "mcr370_meredith6085892.cl"},
	{0xE01BC6FDF5BDAB78ULL, "ave200_donnic6110493.cl"},
	{0xE01C02A07EAF35ACULL, "627181447.ogg"},
	{0xE01C14749F3106B7ULL, "vxlt_darktown_59s_0_0.vlm"},
	{0xE01CB567C5F5C3A5ULL, "vxlt_0_lowtown_docks_b65_0_0.vlm"},
	{0xE01CDD0233C30935ULL, "6052958_m.fxe"},
	{0xE01CE2F0EE04BAB8ULL, "vxht_mansion2_3bp_0_0.vlm"},
	{0xE01D221B6BC3CE58ULL, "6031291_m.fxe"},
	{0xE01D3D25177CAF07ULL, "568397885.ogg"},
	{0xE01D95776E43955CULL, "454062450.ogg"},
	{0xE01DE7C1B8726B12ULL, "6128937_m.fxe"},
	{0xE01E1A3EE3CED504ULL, "6145181_m.fxe"},
	{0xE01E3EB1E3107005ULL, "242048179.ogg"},
	{0xE01E9B94E43E108EULL, "6067719_m.fxe"},
	{0xE01E9EDBD57E266AULL, "w_encounter3_39_0.mao"},
	{0xE01EA2DBD57E2CB9ULL, "w_encounter3_39_0.mmh"},
	{0xE01EACDBD57E3DA7ULL, "w_encounter3_39_0.msh"},
	{0xE01F2738627DA5FEULL, "541391658.ogg"},
	{0xE0203E35B4E04464ULL, "vxlt_0_lowtown_docks_n_b7l_0_0.vlm"},
	{0xE020E626D88B69F2ULL, "follower_banter6170388.cl"},
	{0xE02107501821A48EULL, "gen00fl_carver6104599.cl"},
	{0xE02114C6E10C0490ULL, "6173438_m.fxe"},
	{0xE0215EA887C9BD83ULL, "1056884638.ogg"},
	{0xE02193221C3FE794ULL, "6068833_m.fxe"},
	{0xE021B89F432518EDULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_exit.gad"},
	{0xE021EA0AC1CC63E8ULL, "vxw_sundermount_36x_0_0.vlm"},
	{0xE022034697955D43ULL, "6064603_m.fxe"},
	{0xE02241905B3FB863ULL, "hlac_lt_0_lowtown_a3_n_c_1_7x.dds"},
	{0xE022598DDCE9BD00ULL, "682478326.ogg"},
	{0xE0228AB39359A60BULL, "llac_ca_cavern1_haunted_24423_2.dds"},
	{0xE022EE71C1F9A746ULL, "hlac_w_blightlands_65536_0x.dds"},
	{0xE023B996FDDE8129ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action.ani"},
	{0xE025D827134B3569ULL, "393737878.ogg"},
	{0xE027F5ADCC756B37ULL, "vxlt_undercitydungeon_yi_0_0.vlm"},
	{0xE028326C7C5DC294ULL, "6165745_m.fxe"},
	{0xE0291BB2C7593259ULL, "6206197_m.fxe"},
	{0xE029AA5E06399981ULL, "1051977003.ogg"},
	{0xE029AC9842BA21EBULL, "35076453.ogg"},
	{0xE029D4891ADA7C02ULL, "character_character_p3.psh.fxo"},
	{0xE029FB82800C66A2ULL, "lt_0_lowtown_a2_n_c_20_0_lc.mmh"},
	{0xE02A89B3FADB37CFULL, "659770276.ogg"},
	{0xE02AD496FDE4CDABULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action.gad"},
	{0xE02BAF1515646985ULL, "6092604_m.fxe"},
	{0xE02BC2188DBCC1A8ULL, "396458186.ogg"},
	{0xE02BD45621CECA34ULL, "747455125.ogg"},
	{0xE02C8766AB6F4757ULL, "art\\vfx\\materials\\vfx_blood_pool.mao"},
	{0xE02CA76535EA16B5ULL, "hightown_male_noble1_sta6170798.cl"},
	{0xE02D538E2D219908ULL, "6165767_m.fxe"},
	{0xE02DCB18F030CC41ULL, "428020311.ogg"},
	{0xE02E3642EB51AC6CULL, "hlac_w_sundermount_131066_0z.dds"},
	{0xE02E9D167CAC8ED3ULL, "6095856_m.fxe"},
	{0xE02EAD2DCC0F9504ULL, "149903932.ogg"},
	{0xE02F39A845275669ULL, "llac_ca_caverns2_5802_19.dds"},
	{0xE02F67263F694CFFULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.impa.gad"},
	{0xE02F67CB0E27CCD3ULL, "919704984.ogg"},
	{0xE02FDD59153A3100ULL, "lgt220_mag_tmplr_templar6140354.cl"},
	{0xE02FEDE533F79816ULL, "6136471_m.fxe"},
	{0xE0301B7A61F6118FULL, "mag160_feynriel6020925.cl"},
	{0xE030311AA50E78B5ULL, "vxlt_gallowstemplar_n_439_0_0.vlm"},
	{0xE030B07CF83A2785ULL, "vxht_bartrand_n_5x9_0_0.vlm"},
	{0xE030C22DE2385328ULL, "192210205.ogg"},
	{0xE030D196FDE93410ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\book_hand_off_action.evt"},
	{0xE030D503871A91A3ULL, "815043585.ogg"},
	{0xE03116DFC8792A1FULL, "469776673.ogg"},
	{0xE031809E0EA8D15EULL, "871835190.ogg"},
	{0xE033044606B237F5ULL, "vxlt_undercityserialkiller_221_0_0.vlm"},
	{0xE033107F00507E02ULL, "vxht_0_hightown_a3_9i4_0_0.vlm"},
	{0xE033809E1F746514ULL, "w_woundedcoast_31_0.anb"},
	{0xE033DBEB165E65CEULL, "6116590_m.fxe"},
	{0xE033F94EC56F5100ULL, "llac_lt_0_lowtown_docks_n_-262143_0.dds"},
	{0xE03492C2FF8C2B1CULL, "594361288.ogg"},
	{0xE034CC8F18DC2856ULL, "420683403.ogg"},
	{0xE034CF88F511FD48ULL, "hlac_lt_undercityserialkiller_-1_9y.dds"},
	{0xE0355C263F6DA5ECULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.impa.evt"},
	{0xE036036BB720ACACULL, "vxw_bonepitmining_a2_df_0_0.vlm"},
	{0xE037303C6BCEF299ULL, "672726917.ogg"},
	{0xE0376E3EE6C4D8C8ULL, "6113865_m.fxe"},
	{0xE037E61F28B2C316ULL, "vxht_0_hightown_a3_aps_0_0.vlm"},
	{0xE037E77F006F9C14ULL, "vxlt_gallowscourtyard_n_75o_0_0.vlm"},
	{0xE03821C0B7717124ULL, "vxht_keep_57q_0_0.vlm"},
	{0xE038350F5180B4C8ULL, "6099176_m.fxe"},
	{0xE0398C9BED84B653ULL, "529042317.ogg"},
	{0xE039B05F7109139BULL, "6132926_m.fxe"},
	{0xE039DC646B497AE6ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_barrelbroke_0.pwk"},
	{0xE03AD977C4B8E557ULL, "art\\characters\\playercharacter\\humanfemale\\arm_medj.phy"},
	{0xE03B8F01D92DB95EULL, "dae000_hate_follower_react6103213.cl"},
	{0xE03B91E01D3D364AULL, "ice_shard_pet_blend.mao"},
	{0xE03BFD0CA3A52118ULL, "hlac_ht_mansion2b_818_62x.dds"},
	{0xE03C385D50BB6F7DULL, "vxlt_hangedman_w3_0_0.vlm"},
	{0xE03C3908D5AD0BAFULL, "vxht_0_hightown_a3_9uq_0_0.vlm"},
	{0xE03C9E61FF54866CULL, "6175102_m.fxe"},
	{0xE03CAC5E23E62EA7ULL, "t3_shd_dsk.tnt"},
	{0xE03CE1961416C93AULL, "vxdr_primevaltomb_7o0_0_0.vlm"},
	{0xE03D033CED6BC7A5ULL, "370801628.ogg"},
	{0xE03D19D0B0CC569CULL, "139871737.ogg"},
	{0xE03D26A479AE2143ULL, "one280_emrys6104289.cl"},
	{0xE03DF63A53200746ULL, "vxlt_0_lowtown_a2_fib_0_0.vlm"},
	{0xE03E5FC225EDB096ULL, "6131571_m.fxe"},
	{0xE03E9EF234EE986DULL, "vxlt_0_lowtown_n_f4m_0_0.vlm"},
	{0xE03EA8607B7FC39BULL, "zz_dae_debug6047597.cl"},
	{0xE03F0A94A4470095ULL, "qun221_petrice_saemus_dead.stg"},
	{0xE03F1ED05DED945AULL, "vxlt_warehouse_n_wa_0_0.vlm"},
	{0xE03F4AE06B9439BFULL, "ave200_donnic6110643.cl"},
	{0xE03FD56F106F534EULL, "vxlt_0_lowtown_a2_n_c_htv_0_0.vlm"},
	{0xE04089A45DC0F717ULL, "hlac_dr_primevalentrance_1725_4x.dds"},
	{0xE040B476744EA7AAULL, "6071499_m.fxe"},
	{0xE040D63B5BD42386ULL, "mcr370_varric6098315.cl"},
	{0xE041435F2BAE304AULL, "6034671_f.fxe"},
	{0xE0415F9E1F806AD7ULL, "w_woundedcoast_31_0.mmh"},
	{0xE041639E1F8071E4ULL, "w_woundedcoast_31_0.mao"},
	{0xE041759E1F809049ULL, "w_woundedcoast_31_0.msh"},
	{0xE0419540AF1DA3E2ULL, "6103022_m.fxe"},
	{0xE0419E5188653E68ULL, "gencr_carta_assassin.utc"},
	{0xE041A91F5C54171DULL, "681709581.ogg"},
	{0xE041D8E83BB373F5ULL, "vxw_bonepitmining_a2_1vv_0_0.vlm"},
	{0xE042535919BA000CULL, "652293768.ogg"},
	{0xE0425AD42725AB6BULL, "905439137.ogg"},
	{0xE042DE34626A4E8AULL, "vxdr_primevalentrance_1qx_0_0.vlm"},
	{0xE04344263F79BA45ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.impa.ani"},
	{0xE0434DE31D5AB39EULL, "qcr220_orsino6041833.cl"},
	{0xE0435E345C8EDDA5ULL, "aip_qunarithrallmage_base.gda"},
	{0xE0437F6485A203B5ULL, "docks_male_elven_worker_sta6175788.cl"},
	{0xE0438DA84F7F9573ULL, "6089959_f.fxe"},
	{0xE044178BD29BE1DCULL, "995560341.ogg"},
	{0xE0443C977A4DC891ULL, "6105878_m.fxe"},
	{0xE0447548A916C52CULL, "vxw_bonepitmining_a2_cg_0_0.vlm"},
	{0xE044913CEFC8B722ULL, "follower_banter6156227.cl"},
	{0xE044C7F27FC18B4DULL, "llac_lt_0_lowtown_131069_0.dds"},
	{0xE045894E0320AAB7ULL, "6200070_m.fxe"},
	{0xE0459A24E2D24A3CULL, "hlac_dr_0_deeproads_5843_0z.dds"},
	{0xE04606A1E93258E5ULL, "mag170_thrask6020525.cl"},
	{0xE046125FC7645291ULL, "233350328.ogg"},
	{0xE046AFC1A9F8E85FULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb103a_0t.dds"},
	{0xE046F5AFC5B06035ULL, "vxht_keep_6n0_0_0.vlm"},
	{0xE047318FCB19FB39ULL, "llac_lt_hangedman_-1_22.dds"},
	{0xE047A66385441F68ULL, "plt_lgt000pt_ultra_2_01.plo"},
	{0xE047B6A33CF0A025ULL, "lt_gallowscourtyard_37_0.anb"},
	{0xE047BBD7300D77F9ULL, "mag120_macha6080642.cl"},
	{0xE047C8F06932DE22ULL, "437344809.ogg"},
	{0xE047E1B679FA7324ULL, "vxht_0_hightown_a3_an3_0_0.vlm"},
	{0xE048A0F03E46F108ULL, "679571922.ogg"},
	{0xE048EB35F0656F0DULL, "fen210_tevinter_ambusher6108307.cl"},
	{0xE048FD7BB886F129ULL, "6139118_m.fxe"},
	{0xE049000ED2749C20ULL, "vxca_caverns2prc_7uc_0_0.vlm"},
	{0xE04A3DF9B74DABCBULL, "vxw_encounter1_e9_0_0.vlm"},
	{0xE04A425C88A80741ULL, "6165623_m.fxe"},
	{0xE04AB7C3A1FB4A0CULL, "6130084_m.fxe"},
	{0xE04AF81570FD8C7EULL, "vxht_0_hightown_n_91n_0_0.vlm"},
	{0xE04AFA93C9F42709ULL, "6068335_m.fxe"},
	{0xE04B1087F5F3A269ULL, "vxdr_primevaltomb_34v_0_0.vlm"},
	{0xE04BF03B6A88D028ULL, "mag170_hugh6080798.cl"},
	{0xE04C2D919DE9A577ULL, "6078848_m.fxe"},
	{0xE04C523A4FC9B22BULL, "vxlt_0_lowtown_docks_a3_n_cwu_0_0.vlm"},
	{0xE04CCB4661BDBE28ULL, "vxlt_backalley_g_1rc_0_0.vlm"},
	{0xE04D0BDECDC020ADULL, "6071834_m.fxe"},
	{0xE04D1363D58E1F4FULL, "dae000_orsino6154303.cl"},
	{0xE04D402989E31729ULL, "zz_fen_debug6134377.cl"},
	{0xE04D90AD4FE3ACDAULL, "vxlt_warehouse_n_zb_0_0.vlm"},
	{0xE04DFA39D64FAEB4ULL, "mag110_arianni6016902.cl"},
	{0xE04E9618EAF88402ULL, "dre180cr_drake.ncs"},
	{0xE04EC6F009475469ULL, "mag291_ambient_26218753.cl"},
	{0xE04F6792CF516046ULL, "t3_rob_blo.tnt"},
	{0xE04FD05B279E7D0CULL, "155400397.ogg"},
	{0xE04FFAF85F26A05EULL, "gen00fl_dog6167250.cl"},
	{0xE051A80805B1A98DULL, "292065685.ogg"},
	{0xE051C934ED2AAC70ULL, "vxlt_foundry_n_zv_0_0.vlm"},
	{0xE051E34C7F41D15FULL, "ts_kirk_alienage_03.gda"},
	{0xE0523C915942AAD5ULL, "598096192.ogg"},
	{0xE0524177E212292EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_rdy_crnf.ani"},
	{0xE052759E53119BBBULL, "vxht_0_hightown_a2_9b9_0_0.vlm"},
	{0xE052DCDB8956ECE8ULL, "vxlt_undercityencounter_105_0_0.vlm"},
	{0xE052E7259B0CDDACULL, "vxw_bonepitmining_a2_nd_0_0.vlm"},
	{0xE053DDCB00ECFB2BULL, "6107100_m.fxe"},
	{0xE053DE3FDD27D8C1ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\mt_hed_dwfa_0d.dds"},
	{0xE054706842CB22A2ULL, "vxw_encounter3_6c_0_0.vlm"},
	{0xE054D0AA32CE8E5DULL, "gen00fl_isabela6096761.cl"},
	{0xE055CA8940B72297ULL, "93672341.ogg"},
	{0xE05653936BF08BA4ULL, "vxht_0_hightown_a3_ak0_0_0.vlm"},
	{0xE058AFD6A943C44BULL, "vxw_encounter1_j8_0_0.vlm"},
	{0xE058DB9BDD03DDB8ULL, "560429006.ogg"},
	{0xE058E233E450C15BULL, "90616959.ogg"},
	{0xE058E32DE6BED4AFULL, "vxlt_ruinedhouse_n_to_0_0.vlm"},
	{0xE0592DD2661773C3ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.dura.ani"},
	{0xE05973BEC533AA4FULL, "6083613_m.fxe"},
	{0xE05A43A4773A33C8ULL, "fireflipbook.mao"},
	{0xE05A49A63A3C1041ULL, "gen00fl_dog6117469.cl"},
	{0xE05AB03DB67DC29BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_both_hands.evt"},
	{0xE05ACECC8E4EE370ULL, "6142558_f.fxe"},
	{0xE05B7097CFE0B6EBULL, "78263306.ogg"},
	{0xE05B8F2DDBC8527EULL, "137265121.ogg"},
	{0xE05B94D8B7B46746ULL, "hf_one_crow_3.mor"},
	{0xE05BFA00C3E3082AULL, "hlac_ca_cavern1_haunted_24426_12y.dds"},
	{0xE05C049F701E6D15ULL, "6151314_m.fxe"},
	{0xE05C0B99DBAC3453ULL, "6082588_m.fxe"},
	{0xE05C191C4A95C470ULL, "lt_0_lowtown_a3_n_c_26_0.rml"},
	{0xE05C3B625D61C08EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc04.msh"},
	{0xE05C55625D61ED7CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc04.mmh"},
	{0xE05C88D478981AC0ULL, "715648036.ogg"},
	{0xE05D1AF058A2A69DULL, "806496615.ogg"},
	{0xE05D1B0572C082C9ULL, "seb221ar_harimann_estate.are"},
	{0xE05D7E1A84DB4CD6ULL, "vxlt_undercitydungeon_22c_0_0.vlm"},
	{0xE05D8FFD12D8EAAFULL, "139739528.ogg"},
	{0xE05E42C8ECCC9484ULL, "6080106_f.fxe"},
	{0xE05EA2E8A4FAE0A3ULL, "mage_coneofcold.dds"},
	{0xE05EC000834B7CA6ULL, "vxlt_0_lowtown_docks_bvb_0_0.vlm"},
	{0xE05F30AB3E016974ULL, "187717674.ogg"},
	{0xE05F64DD9BBC7C7FULL, "llac_lt_gallowstemplar_2388_4.dds"},
	{0xE05FC1EE46567F8AULL, "vxdr_primevalentrance_1kz_0_0.vlm"},
	{0xE05FC29B18BABC34ULL, "hlac_ht_mansion2b_818_19z.dds"},
	{0xE05FE58633B84C83ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masd.mmh"},
	{0xE05FF38633B864DDULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masd.msh"},
	{0xE0601A89103A136BULL, "6152357_m.fxe"},
	{0xE0603A51E1506393ULL, "dae000_orana6149041.cl"},
	{0xE0603B11DF20C4F0ULL, "vxlt_foundry_n_uw_0_0.vlm"},
	{0xE060A58FADCB8678ULL, "gen00fl_isabela6099345.cl"},
	{0xE060CD579791DEF2ULL, "vxw_woundedcoast_1f8_0_0.vlm"},
	{0xE060DC3BDA15E528ULL, "and121_karl6093859.cl"},
	{0xE0610EE102107D5EULL, "57216926.ogg"},
	{0xE0612939368B9F4DULL, "6072400_m.fxe"},
	{0xE06156071C54F0C1ULL, "dae000ip_crowd_lm5_talk13.utp"},
	{0xE061A93DB683D5DCULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_both_hands.gad"},
	{0xE061C920BED653CAULL, "w_woundedcoast_132_0.mao"},
	{0xE061CD20BED65A19ULL, "w_woundedcoast_132_0.mmh"},
	{0xE061D720BED66B07ULL, "w_woundedcoast_132_0.msh"},
	{0xE061FC041AD46049ULL, "vxca_caverns1_get_0_0.vlm"},
	{0xE062148D87D5BE8EULL, "6122777_m.fxe"},
	{0xE062E24534C13B22ULL, "vxw_encounter3_3l_0_0.vlm"},
	{0xE0637589F3173BF7ULL, "hlac_ca_caverns1_mines_24424_1z.dds"},
	{0xE063971310EBCAF6ULL, "vxw_woundedcoast_hc_0_0.vlm"},
	{0xE063C8C687D5C95FULL, "795477699.ogg"},
	{0xE06471347F296C30ULL, "6151870_m.fxe"},
	{0xE064C5705DE6A424ULL, "vxht_0_hightown_a3_ap1_0_0.vlm"},
	{0xE065075923977061ULL, "zz_mcr_debug6080859.cl"},
	{0xE065344A708317F5ULL, "gen00fl_isabela6135646.cl"},
	{0xE0656C72B92FFAD9ULL, "826090923.ogg"},
	{0xE065F6157BD090BFULL, "vxht_keep_49a_0_0.vlm"},
	{0xE0668D9C6717509BULL, "mer200im_pickaxe.uti"},
	{0xE066E47D62C9D8AEULL, "vxht_bartrand_n_5pn_0_0.vlm"},
	{0xE067172B032CA6B4ULL, "6152174_m.fxe"},
	{0xE0671A77E2244368ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_rdy_crnf.gad"},
	{0xE0673FA23AB428A6ULL, "6170693_m.fxe"},
	{0xE06745D26623D9BAULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.dura.evt"},
	{0xE06797951D902122ULL, "6171337_m.fxe"},
	{0xE067A83DB6883FDAULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\noble_gesture_both_hands.ani"},
	{0xE067C4049D69184BULL, "vxdr_primevaltomb_9a1_0_0.vlm"},
	{0xE067ED5A6CE26233ULL, "hlac_dr_dungeon_3217_4x.dds"},
	{0xE0680DF1CC29F8F9ULL, "mag320_noblewoman_1.bnk"},
	{0xE06846C1EAC27314ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_lbrt.anb"},
	{0xE0684AFDF99A1C6EULL, "vxca_caverns2prc_485_0_0.vlm"},
	{0xE068609799A4E083ULL, "follower_banter6136639.cl"},
	{0xE068957E9FA63278ULL, "qun200_party_comment6090406.cl"},
	{0xE068A22A7FE5B8FAULL, "vxdr_primevaltomb_53s_0_1.vlm"},
	{0xE068C26138CBAD4BULL, "6134249_m.fxe"},
	{0xE069138F79A3699BULL, "6089573_f.fxe"},
	{0xE06930195D270350ULL, "follower_banter6156104.cl"},
	{0xE069583F9782E889ULL, "qcr220_orsino6061801.cl"},
	{0xE069916F1212DD64ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunkhigh.msh"},
	{0xE069A36385610027ULL, "plt_lgt000pt_ultra_2_01.nss"},
	{0xE069A36F1212FBCDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunkhigh.mao"},
	{0xE069A76F12130206ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunkhigh.mmh"},
	{0xE069F866E2CD3378ULL, "llac_ht_0_hightown_n_131071_1.dds"},
	{0xE06A21576E2B4592ULL, "6102568_m.fxe"},
	{0xE06A7D76CDA24CD3ULL, "6085488_m.fxe"},
	{0xE06AA2568D603C85ULL, "1067270417.ogg"},
	{0xE06AF8487BB0E50CULL, "root_c_01.mao"},
	{0xE06B2FF1CC2C6859ULL, "mag320_noblewoman_1.cnv"},
	{0xE06B50795132DBDBULL, "mrl211ar_house.lst"},
	{0xE06BD62480CE9FF9ULL, "914233667.ogg"},
	{0xE06C9E5615CBF13CULL, "hlac_ht_0_hightown_a2_-196608_2x.dds"},
	{0xE06CE8B14B7BB94DULL, "vxht_keep_673_0_0.vlm"},
	{0xE06D22F50F9E8B67ULL, "qun210cr_guard6087761.cl"},
	{0xE06D3BD3F3F51F7FULL, "t1_skn_t03.tnt"},
	{0xE06D44D2662843ADULL, "art\\vfx\\environment\\vfx_trap_fire_jet_s_p.dura.gad"},
	{0xE06DE527988167B0ULL, "6077598_m.fxe"},
	{0xE06ED90DE4332074ULL, "hlac_w_encounter2_camp_65540_0x.dds"},
	{0xE06EE7B69DC2517EULL, "mag320_mage_conspirator6040910.cl"},
	{0xE06F0E67155951FCULL, "351553679.ogg"},
	{0xE06F3F348987F772ULL, "vxw_woundedcoast_1a9_0_0.vlm"},
	{0xE06FDCDDAB023FAFULL, "21377373.ogg"},
	{0xE07009075866FA11ULL, "6148239_m.fxe"},
	{0xE070E68633C6C1A8ULL, "art\\characters\\playercharacter\\humanfemale\\hlf_masd.phy"},
	{0xE070FB0BE32181EFULL, "6050327_m.fxe"},
	{0xE071198B1003BA12ULL, "228174547.ogg"},
	{0xE071D1BDB3590551ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4extrude.mmh"},
	{0xE071DDBDB35919F2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4extrude.mao"},
	{0xE071EBBDB359310FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2co_roof4extrude.msh"},
	{0xE071FB06E9047A72ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.dura.gad"},
	{0xE07208F002E1E376ULL, "vxw_woundedcoast_ed_0_0.vlm"},
	{0xE07209CDA91C6DE1ULL, "cd_6013395_flemeth_outro.wbd"},
	{0xE0735D7C86AC7F62ULL, "vxlt_warehouse_n_11h_0_0.vlm"},
	{0xE073860A68CFFAE4ULL, "vxca_caverns1_mines_cpc_0_0.vlm"},
	{0xE0739E6F121B47D1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bunkhigh.phy"},
	{0xE07484A641878BF6ULL, "llac_lt_gallowstemplar_2384_5.dds"},
	{0xE074B0555C90AF9FULL, "6148404_m.fxe"},
	{0xE074B94360F3E188ULL, "llac_w_encounter3_-327680_0.dds"},
	{0xE07609720F8639A2ULL, "6174037_m.fxe"},
	{0xE0760E328E497C2FULL, "hlac_lt_warehouse_n_-65535_1y.dds"},
	{0xE076694A04E3B912ULL, "gen00fl_isabela6015686.cl"},
	{0xE0772CCE31E7E223ULL, "6111983_m.fxe"},
	{0xE077555C5A4FF40BULL, "ch2_nolit.psh.fxo"},
	{0xE07779AA9B8234B0ULL, "art\\levels\\races\\design\\_textures\\trap_floorlever_n.dds"},
	{0xE0778F43FA5A7F68ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kitb_0s.dds"},
	{0xE0783719E74D3211ULL, "837617382.ogg"},
	{0xE07844C988791FCEULL, "dae000_arishok6074058.cl"},
	{0xE0785DDFBE3EFFE5ULL, "vxlt_warehouse_n_qx_0_0.vlm"},
	{0xE078FE06E90A9DA9ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.dura.evt"},
	{0xE0799C6AC2D9BF56ULL, "mcr310_argument6182449.cl"},
	{0xE079A2DBD5CAD812ULL, "w_encounter3_39_0.rml"},
	{0xE079D265B5F67E5FULL, "367642395.ogg"},
	{0xE079D6738D373950ULL, "6131101_f.fxe"},
	{0xE079F5C3A60979FDULL, "vxw_sundermount_37g_0_1.vlm"},
	{0xE07A16081D2E3A43ULL, "vxht_keep_3fw_0_0.vlm"},
	{0xE07A2FA31BA3F266ULL, "art\\characters\\creatures\\mubari\\animation\\die_v2.evt"},
	{0xE07A84D1CAC14C7EULL, "vxht_keep_6f7_0_0.vlm"},
	{0xE07AD523F46B1B0FULL, "mag101ip_at_mag100ar_enc.utp"},
	{0xE07B5A8E3D71D1CDULL, "vxht_keep_602_0_0.vlm"},
	{0xE07B6684B415A5AEULL, "hlac_ht_bartrand_n_3369_1y.dds"},
	{0xE07B771ED653D8A4ULL, "hlac_lt_0_lowtown_a3_n_c_-131071_1x.dds"},
	{0xE07BC48DD3839028ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hgrass_n.dds"},
	{0xE07BCA2211DA8844ULL, "2490494.ogg"},
	{0xE07C8940742CEA16ULL, "29970308.ogg"},
	{0xE07CC7B6DAEB0761ULL, "vxht_0_hightown_a2_c6m_0_0.vlm"},
	{0xE07CEBE4543EF4B9ULL, "304097583.ogg"},
	{0xE07D30953CE80513ULL, "6085916_m.fxe"},
	{0xE07D6EDBA48026A5ULL, "one300dg_solivitus.bnk"},
	{0xE07E27068CE708A2ULL, "vxlt_0_lowtown_a2_epb_0_0.vlm"},
	{0xE07E730785449288ULL, "hlac_w_bonepitmining_a2_131078_0z.dds"},
	{0xE07E7AE44A5D1291ULL, "6145139_m.fxe"},
	{0xE07EAC3A2AEAD412ULL, "zz_fen_debug6218577.cl"},
	{0xE07F73B0C293317AULL, "llac_ca_caverns2prc_5802_0.dds"},
	{0xE07FAFF532C48B3DULL, "mag110_arianni6016822.cl"},
	{0xE07FC81510DCA8B4ULL, "6083144_m.fxe"},
	{0xE0802D433E7B85C6ULL, "vxw_bonepitdestroyed_a3_2ae_0_0.vlm"},
	{0xE08039F173622A73ULL, "gen00fl_sebastian6174073.cl"},
	{0xE0803FE5AB8074F3ULL, "t_rouge_4_new_slwp.cut"},
	{0xE08050DBA48229DDULL, "one300dg_solivitus.cnv"},
	{0xE0807ACCF4D7FBF6ULL, "vxw_woundedcoast_be_0_0.vlm"},
	{0xE080B6CD5839B7A1ULL, "6040593_m.fxe"},
	{0xE080F12E65426CCCULL, "one300_zevran6076959.cl"},
	{0xE08127CC5C4C0C73ULL, "hlac_w_encounter1_-131067_0y.dds"},
	{0xE08146A31BAA37B1ULL, "art\\characters\\creatures\\mubari\\animation\\die_v2.gad"},
	{0xE0819548E6EAF3EAULL, "dae000_bodhan6146781.cl"},
	{0xE081A910AB752C67ULL, "qun160_vashothlead6055578.cl"},
	{0xE081B7A33D221A2DULL, "lt_gallowscourtyard_37_0.rml"},
	{0xE081BD8FE7D500DDULL, "mag210_arianni6024109.cl"},
	{0xE081CF5978A297E2ULL, "vxlt_warehouse_n_16i_0_0.vlm"},
	{0xE08234625D828447ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hc04.phy"},
	{0xE0823AF92EDF8DD4ULL, "554552626.ogg"},
	{0xE0823CA030E1D57DULL, "831274985.ogg"},
	{0xE08275435560CE86ULL, "vxlt_0_lowtown_a2_e19_0_0.vlm"},
	{0xE082AE77C4F651AFULL, "art\\characters\\playercharacter\\humanfemale\\arm_medj.mao"},
	{0xE082BA77C4F6664CULL, "art\\characters\\playercharacter\\humanfemale\\arm_medj.mmh"},
	{0xE082C077C4F6703EULL, "art\\characters\\playercharacter\\humanfemale\\arm_medj.msh"},
	{0xE082E215C2714F76ULL, "vxw_encounter2_camp_qt_0_1.vlm"},
	{0xE082E5D129D3A804ULL, "one200_gold_followers6117246.cl"},
	{0xE082EE433E8FADFEULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_c.impa.ani"},
	{0xE0830C0572E1381FULL, "seb221ar_harimann_estate.ncs"},
	{0xE083BCDDB645B362ULL, "gen00fl_varric6076043.cl"},
	{0xE0840F1F221BACDFULL, "w_bonepit_162_0.msh"},
	{0xE084151F221BB681ULL, "w_bonepit_162_0.mmh"},
	{0xE084211F221BCB22ULL, "w_bonepit_162_0.mao"},
	{0xE08487347F3059F5ULL, "375118170.ogg"},
	{0xE084E41BABA6E622ULL, "pri_air.mao"},
	{0xE08597B2B39ADC6CULL, "llac_lt_0_lowtown_-2_4.dds"},
	{0xE085CA0E0DCB144DULL, "follower_barks6153789.cl"},
	{0xE085F5A7BA58BF08ULL, "mcr371_hostile_anders6082138.cl"},
	{0xE086506C9C3FB348ULL, "vxht_0_hightown_a2_c6h_0_0.vlm"},
	{0xE0869D45F1059203ULL, "6004003_m.fxe"},
	{0xE086D9078412DECEULL, "llac_lt_0_lowtown_a3_n_c_65538_2.dds"},
	{0xE086E606E916B218ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.dura.ani"},
	{0xE08710A9975914EEULL, "601933224.ogg"},
	{0xE0877E098D4A28FDULL, "511037754.ogg"},
	{0xE087EB041D7DB0CAULL, "mag101_grace6034388.cl"},
	{0xE0881A1C4ABB00D8ULL, "lt_0_lowtown_a3_n_c_26_0.anb"},
	{0xE08847A31BB0585FULL, "art\\characters\\creatures\\mubari\\animation\\die_v2.ani"},
	{0xE08867A097FF927DULL, "vxw_sundermount_32d_0_1.vlm"},
	{0xE088E05FA102EB59ULL, "hlac_lt_gallowscourtyard_n_-65537_1x.dds"},
	{0xE0890BBE668CBA5DULL, "vxw_sundermount_2z8_0_0.vlm"},
	{0xE0892C0572E5DA0EULL, "seb221ar_harimann_estate.lst"},
	{0xE0896781BE8D5EC0ULL, "vxlt_darktown_4a5_0_0.vlm"},
	{0xE08987309EF2375BULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.mmh"},
	{0xE0899B5DA8D09D83ULL, "68041181.ogg"},
	{0xE089ABF1437777ABULL, "284308729.ogg"},
	{0xE089E20B9B8FFED5ULL, "vxht_mansion2b_2ui_0_0.vlm"},
	{0xE089EFD40FE04D0AULL, "fx_c_fresnel.mao"},
	{0xE08A1597592DC626ULL, "vxlt_0_lowtown_docks_bge_0_0.vlm"},
	{0xE08A3CA38E8E526AULL, "511210028.ogg"},
	{0xE08B12334AECC4CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_enter.gad"},
	{0xE08B916276D2C83FULL, "hlac_lt_0_lowtown_docks_n_-65537_1x.dds"},
	{0xE08BB367C0951066ULL, "mag110_vincento6016317.cl"},
	{0xE08C075EBFCCE998ULL, "follower_banter6143137.cl"},
	{0xE08C120049AC5316ULL, "qcr220_meredith6171321.cl"},
	{0xE08CC38B69BC0C13ULL, "249839871.ogg"},
	{0xE08DA4183AB4B4DFULL, "hlac_dr_0_deeproads_5845_7z.dds"},
	{0xE08DA535566EDFABULL, "439680005.ogg"},
	{0xE08DE3DDA05D3102ULL, "vxlt_hangedman_n_2ek_0_0.vlm"},
	{0xE08DFA4A26E9D959ULL, "vxlt_hangedman_2ug_0_0.vlm"},
	{0xE08DFBCC2D116E7AULL, "817865612.ogg"},
	{0xE08E499D8B3D5E43ULL, "dae000_mother6121005.cl"},
	{0xE08E57FB5EFDAE35ULL, "6217858_m.fxe"},
	{0xE08EE8B2611CC1F0ULL, "vxlt_gallowstemplar_f_5bz_0_0.vlm"},
	{0xE08FAD874D4F49A4ULL, "one200_gold_followers6117112.cl"},
	{0xE08FCE1D12ADC726ULL, "6112154_m.fxe"},
	{0xE0904AF8879E353AULL, "6094793_m.fxe"},
	{0xE090997BBF7CD049ULL, "mag121_idunna6083942.cl"},
	{0xE090C073FB00AFDBULL, "vxlt_gallowscourtyard_66o_0_0.vlm"},
	{0xE09106433E9C13CFULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_c.impa.evt"},
	{0xE0912A1BF86857D5ULL, "vxht_bartrand_n_651_0_0.vlm"},
	{0xE091714EC5E27EBFULL, "581581410.ogg"},
	{0xE09179AE6E451427ULL, "vxlt_gallowstemplar_n_3sx_0_0.vlm"},
	{0xE091BA36AC53021FULL, "6055449_m.fxe"},
	{0xE091EC81E222B332ULL, "vxlt_0_lowtown_n_e1v_0_0.vlm"},
	{0xE091F5334AF2B261ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_enter.evt"},
	{0xE092591786637838ULL, "vxw_encounter3_ep_0_0.vlm"},
	{0xE092C23193934931ULL, "6128970_m.fxe"},
	{0xE092C744651AF20CULL, "6140339_m.fxe"},
	{0xE0931122E14C8BC2ULL, "and111_lirene6090207.cl"},
	{0xE09375CE407842BDULL, "6042865_m.fxe"},
	{0xE0943EBE4B2D51D6ULL, "6145229_f.fxe"},
	{0xE0947047B6F2C54EULL, "gen_im_wep_rog_dua_act0_03.uti"},
	{0xE0950126E554ACDAULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_04ruinstower.phy"},
	{0xE095D64346240928ULL, "gen00fl_aveline6114803.cl"},
	{0xE096511B2A87AB9BULL, "6122435_m.fxe"},
	{0xE096B5B49EDDA084ULL, "flesh2.dds"},
	{0xE0972129E26E63E2ULL, "llac_lt_0_lowtown_docks_n_-262144_1.dds"},
	{0xE0972BCDA93BC3E3ULL, "cd_6013395_flemeth_outro.bnk"},
	{0xE097617951583318ULL, "mrl211ar_house.are"},
	{0xE097A8309EFEACC0ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_l_aoe_p.anb"},
	{0xE097C02A8E890C79ULL, "lgt110_gxa_vgk_silver_order6117028.cl"},
	{0xE09807433EA23478ULL, "art\\vfx\\combat\\vfx_m_sta_ele_close_c.impa.gad"},
	{0xE0983795432AF415ULL, "392751039.ogg"},
	{0xE09853E88D861755ULL, "vxht_mansion2b_2nh_0_0.vlm"},
	{0xE09911E6F57B08DDULL, "236844804.ogg"},
	{0xE0991373A5FF8999ULL, "307370281.ogg"},
	{0xE099384490AE51A8ULL, "dae210_gate_guard6085125.cl"},
	{0xE0994E88AE3B76AFULL, "llac_ht_0_hightown_a2_-131072_1.dds"},
	{0xE099891F411F0540ULL, "mag160_feynriel6020403.cl"},
	{0xE099A53BDF079823ULL, "vxlt_gallowstemplar_2pc_0_0.vlm"},
	{0xE099AB70BED73861ULL, "vxht_0_hightown_a2_c8c_0_0.vlm"},
	{0xE099FAE2C4F41B7BULL, "hlac_ht_0_hightown_a2_-65535_1x.dds"},
	{0xE09A802A5D11CD06ULL, "vxlt_gallowstemplar_n_472_0_0.vlm"},
	{0xE09A8C1BBF873FA8ULL, "mrl000_marethari6129401.cl"},
	{0xE09ACAFC2A37C566ULL, "hlac_lt_0_lowtown_a3_n_c_-3_4y.dds"},
	{0xE09B52216CD19035ULL, "vxlt_0_lowtown_e1r_0_0.vlm"},
	{0xE09C0A19E50BE92AULL, "6095018_f.fxe"},
	{0xE09C1640AABE72BBULL, "hla_ht_0_hightown_2_0y.dds"},
	{0xE09D5E86D8C42CF6ULL, "vxw_woundedcoast_tg_0_0.vlm"},
	{0xE09DF173A7EFDC9BULL, "hlac_lt_blackemporium_1055_5x.dds"},
	{0xE09F5C367DFB1E28ULL, "vxw_bonepitmining_a2_20n_0_0.vlm"},
	{0xE09FED334AFEE1B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_enter.ani"},
	{0xE0A115E86BF2A925ULL, "785908251.ogg"},
	{0xE0A12354CE4DB11CULL, "vxht_mansion_n_54c_0_0.vlm"},
	{0xE0A15EF2627CA4E9ULL, "hlac_lt_0_lowtown_docks_n_-196610_0x.dds"},
	{0xE0A165D2ECC7AB4CULL, "6092599_m.fxe"},
	{0xE0A183AB2E70B310ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bedroll_n.dds"},
	{0xE0A26B88298E58CAULL, "character_character_p197.psh.fxo.dx11.assembly"},
	{0xE0A2792363FFD984ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_enter.ani"},
	{0xE0A2A5E13F03C9B8ULL, "vxht_keep_6ad_0_0.vlm"},
	{0xE0A317CF080BFB64ULL, "dae011_corff_sta6142436.cl"},
	{0xE0A33426802BE448ULL, "vxht_0_hightown_a2_c8f_0_0.vlm"},
	{0xE0A33D3BF053E9E1ULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl_c.anb"},
	{0xE0A46B46DDF04917ULL, "584884860.ogg"},
	{0xE0A5448020E78AECULL, "6172275_m.fxe"},
	{0xE0A55A8D2985F0F1ULL, "350445499.ogg"},
	{0xE0A58B1C05382BE2ULL, "crt_mabari.bnk"},
	{0xE0A6139DAF38AE44ULL, "6119427_m.fxe"},
	{0xE0A625321F20A6F4ULL, "6087315_m.fxe"},
	{0xE0A64B3BA2798FC0ULL, "vxlt_darktown_4k3_0_0.vlm"},
	{0xE0A655629968C8E5ULL, "6046952_f.fxe"},
	{0xE0A6D1C62F1033EFULL, "gen_im_wep_war_one_lore_02.uti"},
	{0xE0A7C8FFA362ECFBULL, "6107070_m.fxe"},
	{0xE0A7F3C57009F177ULL, "531980157.ogg"},
	{0xE0A87EAEB082AA26ULL, "6064589_m.fxe"},
	{0xE0A8DE87C20F5CC3ULL, "6189018_m.fxe"},
	{0xE0A904AF2DE57E43ULL, "6014675_m.fxe"},
	{0xE0A96F4D6FC14FBBULL, "vxlt_0_lowtown_a3_n_c_eml_0_0.vlm"},
	{0xE0A976236405F3E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_enter.gad"},
	{0xE0A994825FBEDBBEULL, "gen_im_trs_bks_orlais.uti"},
	{0xE0AA6A0154640EBDULL, "dae310cr_elf_f_01.utc"},
	{0xE0AA9D143363283AULL, "vxca_cave_23p_0_0.vlm"},
	{0xE0AAC06AAEDD9A4EULL, "269595046.ogg"},
	{0xE0AAE099CFEFA1C1ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potc.phy"},
	{0xE0AAE752FBBF487CULL, "911313604.ogg"},
	{0xE0AB0103B16BD23CULL, "vxlt_0_lowtown_a3_n_c_f9h_0_0.vlm"},
	{0xE0ABAA354DD396BFULL, "6202029_m.fxe"},
	{0xE0ABD063CABA4576ULL, "vxw_woundedcoast_qh_0_0.vlm"},
	{0xE0AC2EE56E2F2CB6ULL, "6137479_f.fxe"},
	{0xE0AC7C12129B684AULL, "553409790.ogg"},
	{0xE0AC8F9188FA77D2ULL, "100977536.ogg"},
	{0xE0AC9F195ECBCC84ULL, "6217244_m.fxe"},
	{0xE0AD6626FA3E3650ULL, "807712733.ogg"},
	{0xE0ADBC645E4DC04EULL, "vxlt_0_lowtown_docks_a3_cy8_0_1.vlm"},
	{0xE0AE11E5CA0CD3FDULL, "6133138_m.fxe"},
	{0xE0AEC2476B7CE283ULL, "gen00fl_anders6119385.cl"},
	{0xE0AF1DCE1C98532EULL, "vxht_bartrand_n_5oc_0_0.vlm"},
	{0xE0AF7123640A56C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_enter.evt"},
	{0xE0B016B90D9317ACULL, "w_bonepitmining_a2_127_0.msh"},
	{0xE0B01CB90D93219EULL, "w_bonepitmining_a2_127_0.mmh"},
	{0xE0B01E3BF05E3F5AULL, "art\\vfx\\combat\\vfx_w_2hd_atk_cl_c.mmh"},
	{0xE0B028B90D933635ULL, "w_bonepitmining_a2_127_0.mao"},
	{0xE0B09B414AAF707EULL, "packages_base.gda"},
	{0xE0B0AA50FDFDC815ULL, "vxlt_darktown_4n9_0_0.vlm"},
	{0xE0B117BE30F9E238ULL, "vxht_keep_6je_0_0.vlm"},
	{0xE0B237EFF21C2DA2ULL, "vxht_0_hightown_n_b8f_0_0.vlm"},
	{0xE0B2BAA5370D7756ULL, "vxht_mansion_n_523_0_0.vlm"},
	{0xE0B2E1BEB90AFE3CULL, "em_way_dalish_hunter.mor"},
	{0xE0B2FD4A025C4A2DULL, "133398349.ogg"},
	{0xE0B32E8C24105EB0ULL, "hlac_ht_0_hightown_a2_65537_1z.dds"},
	{0xE0B3733003F837DBULL, "138300671.ogg"},
	{0xE0B4092124699280ULL, "fx_handgrab.mao"},
	{0xE0B41110BDD05238ULL, "888888404.ogg"},
	{0xE0B4803387836A13ULL, "hm_decimus.mor"},
	{0xE0B4BF0547E0274CULL, "dae000_gamlen6048502.cl"},
	{0xE0B576FDA9ED5A91ULL, "vxlt_0_lowtown_a2_n_c_gbp_0_0.vlm"},
	{0xE0B6862CBA95F386ULL, "vxht_0_hightown_n_by5_0_0.vlm"},
	{0xE0B6C05ABA288990ULL, "175919178.ogg"},
	{0xE0B705AC23F2958CULL, "7139927.ogg"},
	{0xE0B72C1DE52E2BA4ULL, "gen_im_wep_mag_sta_act1_03.uti"},
	{0xE0B7557721A63370ULL, "6219039_m.fxe"},
	{0xE0B791A33027CA9DULL, "vxca_caverns1_mines_k2p_0_0.vlm"},
	{0xE0B7A9DA8662ECAEULL, "503059345.ogg"},
	{0xE0B7B466C7D3FA4DULL, "254970602.ogg"},
	{0xE0B7CF6065426A53ULL, "vxca_cave_1oo_0_1.vlm"},
	{0xE0B7F08E3ED07A9FULL, "follower_banter6154198.cl"},
	{0xE0B7FD6FD31E9A37ULL, "632996767.ogg"},
	{0xE0B857C5E13C7342ULL, "hlac_lt_undercityserialkiller_-1_0x.dds"},
	{0xE0B89586FBB1F561ULL, "hlac_lt_0_lowtown_-131072_7x.dds"},
	{0xE0B8A574A0281E2FULL, "6133815_m.fxe"},
	{0xE0B8AD3BC2A6C0E2ULL, "6103251_m.fxe"},
	{0xE0B91F6E9F0AB0B6ULL, "vxlt_warehouse_n_xt_0_0.vlm"},
	{0xE0B98C3AAFC86DECULL, "6096137_m.fxe"},
	{0xE0B9A0737BBF0C72ULL, "6044639_m.fxe"},
	{0xE0BA4FD7F0D89A11ULL, "6041037_m.fxe"},
	{0xE0BA670B68E03C31ULL, "dae011_badpoet_sta6146928.cl"},
	{0xE0BAE39C28D10859ULL, "follower_banter6155982.cl"},
	{0xE0BB4238CC24ABA2ULL, "198527545.ogg"},
	{0xE0BBD763F39E2142ULL, "hlac_ca_caverns2prc_5802_6x.dds"},
	{0xE0BC32B18E5A185AULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_001a_0s.dds"},
	{0xE0BC89E28BF80A36ULL, "6034346_m.fxe"},
	{0xE0BCE892B779A7FCULL, "vxdr_primevaltomb_75t_0_0.vlm"},
	{0xE0BCF9D10180BB0AULL, "vxlt_warehouse_n_125_0_0.vlm"},
	{0xE0BDCC2280EED222ULL, "vxdr_primevalentrance_34e_0_0.vlm"},
	{0xE0BDEC0BE977E161ULL, "hlac_lt_warehouse_n_0_1z.dds"},
	{0xE0BDED700E50EA8BULL, "lt_0_lowtown_docks_a3_n_2_0.anb"},
	{0xE0BE28C9CE9075B5ULL, "6149492_m.fxe"},
	{0xE0BE72015373063EULL, "6121686_m.fxe"},
	{0xE0BE90D91E65F051ULL, "443308447.ogg"},
	{0xE0BFE0699A14D085ULL, "gen00fl_fenris6113848.cl"},
	{0xE0BFE7F55CC09429ULL, "6028167_f.fxe"},
	{0xE0C00DE03DE5C4DEULL, "449112209.ogg"},
	{0xE0C01D323B9DEE73ULL, "6074512_m.fxe"},
	{0xE0C01F0A13E53B82ULL, "6147556_m.fxe"},
	{0xE0C0B37DFAF5B9CAULL, "vxca_caverns2prc_af2_0_0.vlm"},
	{0xE0C0BCC8CB975B62ULL, "6208700_f.fxe"},
	{0xE0C133042741D006ULL, "790875024.ogg"},
	{0xE0C165484CF92B0BULL, "6177849_f.fxe"},
	{0xE0C1EBB3F1D3E79EULL, "dae000_madam_lusine6151395.cl"},
	{0xE0C1FFDB48BFC4E1ULL, "hlac_ca_caverns2prc_5802_43z.dds"},
	{0xE0C273ADB0791DF9ULL, "6119439_m.fxe"},
	{0xE0C2B1BB8B646BFDULL, "6163870_m.fxe"},
	{0xE0C322A1A926A313ULL, "roots.mao"},
	{0xE0C36DD3E07AA04BULL, "6119730_m.fxe"},
	{0xE0C3773232A5AE20ULL, "532199340.ogg"},
	{0xE0C3970DEA457256ULL, "18306185.ogg"},
	{0xE0C3A97F636860D5ULL, "vxht_mansion2b_2au_0_0.vlm"},
	{0xE0C3CE20BF292606ULL, "w_woundedcoast_132_0.phy"},
	{0xE0C40FA7A6FAB19DULL, "777483629.ogg"},
	{0xE0C41E7783B889B1ULL, "6020784_m.fxe"},
	{0xE0C43CE4A7A05BF8ULL, "6163497_m.fxe"},
	{0xE0C4A69F04084DA1ULL, "6189676_m.fxe"},
	{0xE0C4CB085FD239B8ULL, "vxht_0_hightown_n_bdx_0_0.vlm"},
	{0xE0C4DAEA37F7511DULL, "vxht_0_hightown_a3_b14_0_0.vlm"},
	{0xE0C51E982093E2E9ULL, "6036771_f.fxe"},
	{0xE0C5D076A6FD6857ULL, "hlac_ht_mansion2_818_45z.dds"},
	{0xE0C606EA9763B3B5ULL, "hlac_ca_cavern1_haunted_24426_5x.dds"},
	{0xE0C6605746B58322ULL, "vxlt_0_lowtown_a2_eoo_0_0.vlm"},
	{0xE0C6BB672508CD5CULL, "536921260.ogg"},
	{0xE0C6CC94C69E9779ULL, "one100dg_lia6013844.cl"},
	{0xE0C785AA931FE481ULL, "6162598_m.fxe"},
	{0xE0C7983CD9086B3BULL, "hlac_ht_brothel_n_3364_0z.dds"},
	{0xE0C7C9400DA27612ULL, "vrc321cr_golem_guard.ncs"},
	{0xE0C7D667755B1B36ULL, "6118822_m.fxe"},
	{0xE0C874A10C88A5D5ULL, "hlac_lt_gallowstemplar_2388_2x.dds"},
	{0xE0C896755B419E88ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_dq101a_0d.dds"},
	{0xE0C8992EE2AD01F0ULL, "llac_w_bonepitmining_a2_196608_0.dds"},
	{0xE0C8C12BDCE36622ULL, "vxht_keep_n_6eu_0_0.vlm"},
	{0xE0C8C490788FE0C6ULL, "6042310_m.fxe"},
	{0xE0C94D81909DB644ULL, "vxlt_0_lowtown_a2_n_c_e16_0_0.vlm"},
	{0xE0C955249386DBB7ULL, "hlac_lt_darktown_-1_10z.dds"},
	{0xE0C993BD3815EC6FULL, "355247359.ogg"},
	{0xE0C996A4E848B3F8ULL, "6050405_m.fxe"},
	{0xE0C9A241BADF860CULL, "vrc321pt_haunted.ncs"},
	{0xE0C9E00E7C39A932ULL, "6173775_m.fxe"},
	{0xE0C9FC1DEC4A0418ULL, "hlac_lt_0_lowtown_docks_131070_0z.dds"},
	{0xE0CA0EE7DAAF4916ULL, "vxht_0_hightown_c7e_0_0.vlm"},
	{0xE0CA751C05572200ULL, "crt_mabari.wbd"},
	{0xE0CAD6D8281B87B9ULL, "zz_approval_debug6003114.cl"},
	{0xE0CAD9C0A316B310ULL, "hlac_lt_0_lowtown_-262144_4y.dds"},
	{0xE0CB53E8DE804AA7ULL, "743238639.ogg"},
	{0xE0CBE2B05B62FDD6ULL, "mer341_earl6054244.cl"},
	{0xE0CC3CF266C0290AULL, "gen00fl_anders6125990.cl"},
	{0xE0CC9553D0305DF1ULL, "773804062.ogg"},
	{0xE0CC9FE38A8C0424ULL, "mag311_emile_de_launcet6043616.cl"},
	{0xE0CCBD7500A9569FULL, "mcr371_orsino6083475.cl"},
	{0xE0CCF8D9F2A468FEULL, "vxht_0_hightown_n_9bg_0_0.vlm"},
	{0xE0CDE3C10F7507FEULL, "993288654.ogg"},
	{0xE0CDEAA215626616ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_crumble.mao"},
	{0xE0CDEFD143E45735ULL, "art\\levels\\races\\ferelden\\props\\_textures\\ossuary_d.dds"},
	{0xE0CDF6A215627AB5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_crumble.mmh"},
	{0xE0CDF8A215627E2BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0cc_crumble.msh"},
	{0xE0CF0A4E7ADD38A6ULL, "415078766.ogg"},
	{0xE0CFC1EFAB36CC11ULL, "mag100_karras6035316.cl"},
	{0xE0CFF926BD0AF5ADULL, "vxht_bartrand_n_68k_0_1.vlm"},
	{0xE0D09A1EC4C5E187ULL, "hlac_lt_0_lowtown_docks_n_-65538_4z.dds"},
	{0xE0D168BAE9F1ED93ULL, "llac_ht_0_hightown_n_-131072_0.dds"},
	{0xE0D1DF443F90F979ULL, "lla_ht_0_hightown_a3_-65536_0.dds"},
	{0xE0D2301B7A6C9B57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_strt.gad"},
	{0xE0D26FCCDAEE0E6DULL, "6015779_m.fxe"},
	{0xE0D44076A1CF2B9BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_interact_low_elf_male.ani"},
	{0xE0D4488748E33DDFULL, "517839789.ogg"},
	{0xE0D4925595D51697ULL, "gen00fl_anders6121203.cl"},
	{0xE0D4955518161183ULL, "196689406.ogg"},
	{0xE0D52D8B18A90763ULL, "166482332.ogg"},
	{0xE0D5366F9A2CABB1ULL, "6167580_m.fxe"},
	{0xE0D5D926A0606E8EULL, "vxlt_0_lowtown_docks_a3_aou_0_0.vlm"},
	{0xE0D5F7878515FD01ULL, "6167698_m.fxe"},
	{0xE0D66A01D188DE44ULL, "vxlt_0_lowtown_h8m_0_0.vlm"},
	{0xE0D670CD1BFF849BULL, "mag131_taraneh6088751.cl"},
	{0xE0D6BEC7F2EAF485ULL, "vxht_mansion2b_1te_0_0.vlm"},
	{0xE0D6FCB9D01A713DULL, "6110018_m.fxe"},
	{0xE0D71B2A9BC20A6FULL, "661333540.ogg"},
	{0xE0D720758A2CED75ULL, "429534886.ogg"},
	{0xE0D7EA51A0908E73ULL, "and111_lirene6097120.cl"},
	{0xE0D8BEF03D2F0192ULL, "hlac_ht_0_hightown_n_-131074_1x.dds"},
	{0xE0D8DED5408CEE89ULL, "isa211_isabela6072566.cl"},
	{0xE0D8ECDD040E6761ULL, "qun120_arishok6051673.cl"},
	{0xE0D92025945DDF11ULL, "gen_im_trs_dem_blt_r0.uti"},
	{0xE0D966D3B3ECBEA0ULL, "122947762.ogg"},
	{0xE0D96FBA30EA6C01ULL, "ca_link_cl.mao"},
	{0xE0D97455187917D6ULL, "120861035.ogg"},
	{0xE0D9EEF3D3238151ULL, "6113571_f.fxe"},
	{0xE0DA1F45298B086BULL, "mag311_keran6041693.cl"},
	{0xE0DA47085BF607B9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkl.ani"},
	{0xE0DA56BB71B2577CULL, "1560325.ogg"},
	{0xE0DA87838FE0796FULL, "character_rendertoshadowmap_p0.vsh.fxo.dx11.assembly"},
	{0xE0DA9DE50E1A77A8ULL, "plt_cod_hst_qun_asit_tal_eb.plo"},
	{0xE0DAB7A88553F9B5ULL, "follower_banter6163990.cl"},
	{0xE0DB2B6C71FF11B3ULL, "hlac_lt_0_lowtown_a2_n_65538_3y.dds"},
	{0xE0DB45D0FB1189F7ULL, "vxlt_0_lowtown_a3_n_c_emx_0_0.vlm"},
	{0xE0DB9F444D04924AULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_hvyn_0s.dds"},
	{0xE0DBDF9A54CBAA3DULL, "6013320_m.fxe"},
	{0xE0DC05450259CDA8ULL, "294363726.ogg"},
	{0xE0DC1B71B8CCE449ULL, "gen_im_cft_reg_k.uti"},
	{0xE0DC77654F3EB0E1ULL, "llac_lt_0_lowtown_n_131071_2.dds"},
	{0xE0DC847A9A56A99BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_walls_d.dds"},
	{0xE0DD1C5A7CB003C4ULL, "611119100.ogg"},
	{0xE0DDE8CD5DE5698EULL, "one300_zevran6075840.cl"},
	{0xE0DE02AAAF0491F9ULL, "gen00fl_fenris6126581.cl"},
	{0xE0DF42257B555CE2ULL, "vxlt_0_lowtown_a3_n_c_er8_0_0.vlm"},
	{0xE0DFE130FBA04F23ULL, "6152069_m.fxe"},
	{0xE0DFF5C2E388CA2EULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa.mmh"},
	{0xE0E00FC2E388F61CULL, "art\\vfx\\combat\\vfx_gib_hm_tmpa.msh"},
	{0xE0E0518DC466D14BULL, "6113130_m.fxe"},
	{0xE0E083CACAE3143AULL, "dae000_bartrand6206906.cl"},
	{0xE0E0950C72C4D74AULL, "gen00fl_anders6207696.cl"},
	{0xE0E122085BFBE7FBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkl.gad"},
	{0xE0E12F46AC22A916ULL, "vxlt_gallowstemplar_45k_0_0.vlm"},
	{0xE0E17DDF8E5C1F51ULL, "832798015.ogg"},
	{0xE0E2764055CEE7FCULL, "297536000.ogg"},
	{0xE0E277AAC8FB9DDAULL, "18433971.ogg"},
	{0xE0E2C71F12C5D9E7ULL, "disabled_unlock_pressed.cur"},
	{0xE0E2D6C2F951591DULL, "hlac_ht_0_hightown_a3_0_1z.dds"},
	{0xE0E2FCB1ACE8791BULL, "4203637.ogg"},
	{0xE0E32523E0E9FC58ULL, "vxht_chantry_n_2zn_0_0.vlm"},
	{0xE0E3986A18A8FEB4ULL, "hlac_dr_slavers_3216_8z.dds"},
	{0xE0E3BF0C53295BADULL, "hlac_lt_0_lowtown_n_-65536_7z.dds"},
	{0xE0E40A7A04F9FC69ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_dsaxa_0s.dds"},
	{0xE0E419D4654E7CC0ULL, "598318756.ogg"},
	{0xE0E4870C5B3951D0ULL, "470815920.ogg"},
	{0xE0E4A64F5579562CULL, "zz_dae_debug6029862.cl"},
	{0xE0E4A9410E44C2DEULL, "vxht_keep_58f_0_0.vlm"},
	{0xE0E4FAEE83F0E707ULL, "6208557_m.fxe"},
	{0xE0E5277BC458D5C1ULL, "203349374.ogg"},
	{0xE0E52D4D599D9B54ULL, "vxw_bonepitdestroyed_a3_br_0_0.vlm"},
	{0xE0E5EB639B18EA94ULL, "169097345.ogg"},
	{0xE0E5EDDAD6350640ULL, "408358097.ogg"},
	{0xE0E62B301093A3F9ULL, "w_woundedcoast_109_0.msh"},
	{0xE0E635301093B4E7ULL, "w_woundedcoast_109_0.mmh"},
	{0xE0E639301093BB34ULL, "w_woundedcoast_109_0.mao"},
	{0xE0E6C4BD31E5560CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd3a.msh"},
	{0xE0E6CABD31E560FEULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd3a.mmh"},
	{0xE0E6EEC775D45780ULL, "17029667.ogg"},
	{0xE0E6F273564FA9AAULL, "374505772.ogg"},
	{0xE0E6FF53FFC46777ULL, "270420521.ogg"},
	{0xE0E738B215CB70A0ULL, "hlac_ca_caverns1_24424_25x.dds"},
	{0xE0E73F085C0084E0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_cwkl.evt"},
	{0xE0E75D1B7A7F43DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\crushing_prison_strt.ani"},
	{0xE0E7C8E44DCF58AAULL, "dae000_sabina6161485.cl"},
	{0xE0E839809949321EULL, "lt_0_lowtown_docks_a3_n_24_0_lc.mmh"},
	{0xE0E8459716956B7DULL, "hlac_lt_undercityserialkiller_-1_41z.dds"},
	{0xE0E86776A1E01725ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\crouch_squat_interact_low_elf_male.gad"},
	{0xE0E88E2CC759940EULL, "95239376.ogg"},
	{0xE0E89547C0D05EA9ULL, "49745596.ogg"},
	{0xE0E8B5036734ACF6ULL, "vxw_bonepit_1mb_0_0.vlm"},
	{0xE0E93F8255672DABULL, "vxca_caverns1_mines_i2l_0_0.vlm"},
	{0xE0E964E0C28F6317ULL, "6206887_m.fxe"},
	{0xE0E987ADEE78BA7FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkback.gad"},
	{0xE0E9EA700E7620D7ULL, "lt_0_lowtown_docks_a3_n_2_0.rml"},
	{0xE0EA530038512CE0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldbase_1.mao"},
	{0xE0EA5FDBB6D16F9BULL, "hlac_lt_0_lowtown_docks_-65537_0x.dds"},
	{0xE0EAED11327126C9ULL, "6091678_m.fxe"},
	{0xE0EB85BC57FA08B8ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_robe_s.dds"},
	{0xE0EBC8507019CC9FULL, "6148615_m.fxe"},
	{0xE0EBD695183F351FULL, "6180462_m.fxe"},
	{0xE0EC22DD4A26DA16ULL, "llac_ht_chantry_n_2534_42.dds"},
	{0xE0EC5ADD0140288BULL, "254563709.ogg"},
	{0xE0EC6DD3DEA83264ULL, "6139927_f.fxe"},
	{0xE0ECA52DFB35217AULL, "6147808_m.fxe"},
	{0xE0ECAB19F5C7CEBEULL, "vxlt_gallowsprison_n_6ab_0_0.vlm"},
	{0xE0ECFDEF78D97F72ULL, "364886364.ogg"},
	{0xE0ED201738462F6AULL, "vxw_encounter2_3o_0_0.vlm"},
	{0xE0ED3F3A17D773A7ULL, "572683587.ogg"},
	{0xE0EDB4026D4B7562ULL, "vxlt_0_lowtown_a3_n_c_eu9_0_0.vlm"},
	{0xE0EDB885808A1C6AULL, "gen_im_arm_cht_comc_t3.uti"},
	{0xE0EDBC400DC32FE4ULL, "vrc321cr_golem_guard.utc"},
	{0xE0EDE5431CB7EFD3ULL, "706830193.ogg"},
	{0xE0EE6692225B96F9ULL, "hlac_lt_gallowscourtyard_131070_1z.dds"},
	{0xE0EEFE3E299633D4ULL, "llac_w_sundermount_-4_0.dds"},
	{0xE0EF0C58530C3E49ULL, "vxlt_undercityencounter_1ic_0_0.vlm"},
	{0xE0EF3870E91021DEULL, "6133861_m.fxe"},
	{0xE0EF5460D3CEC446ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_action.ani"},
	{0xE0EF7CADEE7D136CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkback.evt"},
	{0xE0EFD31A0090DF79ULL, "vxht_bartrand_n_732_0_0.vlm"},
	{0xE0F00BF62E62A815ULL, "726359845.ogg"},
	{0xE0F04C5BE1706CDCULL, "zz_seb_debug6215773.cl"},
	{0xE0F0FA26E5A2FF03ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_04ruinstower.msh"},
	{0xE0F10826E5A3175DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_04ruinstower.mmh"},
	{0xE0F10C26E5A31E6EULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_04ruinstower.mao"},
	{0xE0F14B26620FC077ULL, "hlac_lt_gallowscourtyard_-2_1y.dds"},
	{0xE0F177A65E5EE7ACULL, "6174032_f.fxe"},
	{0xE0F195580D6CCEE8ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_hthedge.msh"},
	{0xE0F1B3580D6D0112ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_hthedge.mmh"},
	{0xE0F24C2834D633C5ULL, "vxht_keep_n_73e_0_0.vlm"},
	{0xE0F2A4CB2FC09C9DULL, "vxlt_0_lowtown_a2_n_h29_0_0.vlm"},
	{0xE0F3642C756AFAE8ULL, "424554996.ogg"},
	{0xE0F39F2A4B93B3D4ULL, "vxw_bonepitdestroyed_a3_ms_0_0.vlm"},
	{0xE0F3BC2ACE2C49C6ULL, "vxw_bonepitdestroyed_a3_29m_0_0.vlm"},
	{0xE0F3C01EC9CAA918ULL, "6177845_m.fxe"},
	{0xE0F3E17B30A6C43BULL, "lgt260_eau_tmp_roderick6143721.cl"},
	{0xE0F3FD4162A689C7ULL, "zz_mer_debug6189051.cl"},
	{0xE0F3FF0EBF56E629ULL, "mag291_torpor6025996.cl"},
	{0xE0F53A9096B714EEULL, "llac_ht_0_hightown_a2_-65535_0.dds"},
	{0xE0F635ED015B6C8EULL, "llac_lt_0_lowtown_docks_n_131073_0.dds"},
	{0xE0F69671BAA71981ULL, "916035672.ogg"},
	{0xE0F747BEA5AA6871ULL, "hlac_lt_0_lowtown_n_-1_3x.dds"},
	{0xE0F79F787B9E457CULL, "vxlt_0_lowtown_n_gc4_0_0.vlm"},
	{0xE0F83965BC611B39ULL, "363629857.ogg"},
	{0xE0F8E9703DD15F37ULL, "ser270_cut_aftermath6162294.cl"},
	{0xE0F8F63ADE69728CULL, "vxht_mansion_n_6ik_0_0.vlm"},
	{0xE0F9175640F20CBDULL, "718183670.ogg"},
	{0xE0F92F224351BAEBULL, "6201644_m.fxe"},
	{0xE0F95EEE24673F49ULL, "6096778_m.fxe"},
	{0xE0F96C52B6391423ULL, "6180017_m.fxe"},
	{0xE0F9E13C3DA30E31ULL, "vxdr_0_deeproads_evt_0_0.vlm"},
	{0xE0FA280704E4E492ULL, "hightown_fem_noble1_sta6169255.cl"},
	{0xE0FAC70FF2CF150AULL, "6174590_m.fxe"},
	{0xE0FBB4D1C48727B8ULL, "vxdr_primevalentrance_3cv_0_0.vlm"},
	{0xE0FC48D859F519E0ULL, "737395873.ogg"},
	{0xE0FD33425C298924ULL, "lgt310cr_mag_grigor.utc"},
	{0xE0FD3C60D3DAD8B7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_action.evt"},
	{0xE0FD64ADEE8927C5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\combatwalkback.ani"},
	{0xE0FDC611FF231F17ULL, "vxht_mansion_n_21x_0_0.vlm"},
	{0xE0FDD12744CA212AULL, "vxw_bonepitdestroyed_a3_2ai_0_0.vlm"},
	{0xE0FE71E28EAA3B1AULL, "gen00fl_isabela_init6076491.cl"},
	{0xE0FFBC93A42E7B43ULL, "vxht_keep_n_30n_0_0.vlm"},
	{0xE1003D8DFF089622ULL, "552863683.ogg"},
	{0xE100D43859EDF33DULL, "vxlt_hangedman_n_25p_0_0.vlm"},
	{0xE1016E8EF88CD5F2ULL, "vxlt_darktown_5la_0_0.vlm"},
	{0xE101758B3E96D274ULL, "306438019.ogg"},
	{0xE101B2A53D18E649ULL, "mag311_huon6042688.cl"},
	{0xE101FADB2BCB7BDBULL, "386723567.ogg"},
	{0xE1022E07C0226246ULL, "vxw_bonepitdestroyed_a3_24n_0_0.vlm"},
	{0xE1023AB5B29B691BULL, "27608820.ogg"},
	{0xE1025A37AD3C8835ULL, "281924296.ogg"},
	{0xE102ACC5EA3E7D95ULL, "zz_fol_debug6072541.cl"},
	{0xE102B344777C23CEULL, "6115611_f.fxe"},
	{0xE103548FBE6423B7ULL, "vxlt_0_lowtown_docks_a3_n_d0o_0_0.vlm"},
	{0xE1040AD6B963E499ULL, "hlac_ht_mansion2_818_58z.dds"},
	{0xE1043D60D3E0F940ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_deadly_strike_action.gad"},
	{0xE104757A59BEE51CULL, "vxca_caverns1_jlk_0_0.vlm"},
	{0xE10484F410907931ULL, "vxlt_warehouse_187_0_0.vlm"},
	{0xE104884C7F37FD6EULL, "vxlt_backalley_n_1ym_0_0.vlm"},
	{0xE104FDFEDCE90F7CULL, "seb321_leliana6181810.cl"},
	{0xE1054F20DC30FD2CULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dsdgc_0d.dds"},
	{0xE105651D00EBC7E8ULL, "gen00pt_specializations.ncs"},
	{0xE105A584A4D1119BULL, "mag120_macha6080594.cl"},
	{0xE105ACE112845A48ULL, "184014750.ogg"},
	{0xE10635C38A5BD952ULL, "291449117.ogg"},
	{0xE106CBBA5185A0D8ULL, "gen00fl_merrill6127211.cl"},
	{0xE106CD981E198B6DULL, "6111119_f.fxe"},
	{0xE107435FD57745A8ULL, "llac_ht_keep_n_3732_3.dds"},
	{0xE1074A54DF233CA0ULL, "vxdr_primevaltomb_2w6_0_1.vlm"},
	{0xE1075349D0473D39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug.ani"},
	{0xE1076817D05F8B0CULL, "vxht_mansion_n_6lh_0_0.vlm"},
	{0xE107E8C2557149F6ULL, "6119898_m.fxe"},
	{0xE107ED69DDDDF7ACULL, "6018423_m.fxe"},
	{0xE1081E065A25E8DBULL, "hlac_w_encounter3_-131075_0x.dds"},
	{0xE10836EB3097A5BEULL, "mrl310_merrill6128819.cl"},
	{0xE108674749CC80BCULL, "llac_w_blightlands_0_10.dds"},
	{0xE1099F47A3ECA7C9ULL, "6179045_f.fxe"},
	{0xE109A5A6E681E76CULL, "gen00fl_dog6172085.cl"},
	{0xE10A03D11C32606AULL, "vxw_encounter2_9i_0_0.vlm"},
	{0xE10A0792A94D2676ULL, "vxca_caverns1_edf_0_0.vlm"},
	{0xE10A1D40FE9F149CULL, "vxw_encounter2_camp_ph_0_0.vlm"},
	{0xE10A3FD9DB100266ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.impa.evt"},
	{0xE10BA89D491D346DULL, "vxht_keep_n_3zn_0_0.vlm"},
	{0xE10C00481BD0BCB9ULL, "dae100_bran6118345.cl"},
	{0xE10C2C8EDE911545ULL, "802835586.ogg"},
	{0xE10C571D64B51F64ULL, "vis121_saemus6101175.cl"},
	{0xE10C83436D1B68E7ULL, "6085835_m.fxe"},
	{0xE10C93855D865F8AULL, "dae000_arishok6078353.cl"},
	{0xE10CE599D04274FDULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potc.mao"},
	{0xE10CE999D0427B36ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potc.mmh"},
	{0xE10CF399D0428C34ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_potc.msh"},
	{0xE10D15D93ADE3A69ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_basket.phy"},
	{0xE10D67A9EAE36BD9ULL, "166212445.ogg"},
	{0xE10DD66851DA68D7ULL, "vxlt_0_lowtown_a2_n_ep5_0_1.vlm"},
	{0xE10DD814A423412EULL, "6202671_m.fxe"},
	{0xE10E2E49D04D1D7BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug.gad"},
	{0xE10E4373B8343445ULL, "hlac_ht_0_hightown_-65535_1x.dds"},
	{0xE10EBB81F41206C0ULL, "magev_item_acquired.ncs"},
	{0xE10EEA1B94173F90ULL, "6102755_m.fxe"},
	{0xE10F002E79D51569ULL, "art\\vfx\\environment\\vfx_fog_ground_p.mmh"},
	{0xE10F4EF233A40CDFULL, "vxlt_0_lowtown_a3_n_c_ghy_0_0.vlm"},
	{0xE10F5F2C2790ECCEULL, "vxlt_hangedman_n_2ys_0_0.vlm"},
	{0xE10F865A3A128416ULL, "zz_dre_debug6168218.cl"},
	{0xE10F997E06D340F1ULL, "character_character_p169.psh.fxo.dx11.assembly"},
	{0xE10FA7F6F7108A47ULL, "269780245.ogg"},
	{0xE10FE06BEA82EE72ULL, "vxlt_darktown_5qb_0_0.vlm"},
	{0xE110487FAF0F29BBULL, "6064085_m.fxe"},
	{0xE1104BF2BAFAE493ULL, "vxht_bartrand_n_6rr_0_0.vlm"},
	{0xE110A7E8F9399ACAULL, "vxlt_0_lowtown_a2_n_c_hr4_0_0.vlm"},
	{0xE11156D9DB1647B1ULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.impa.gad"},
	{0xE1116C8CCC293C43ULL, "gen00fl_varric6118739.cl"},
	{0xE1119184C2175147ULL, "vxht_keep_3he_0_0.vlm"},
	{0xE111949686BAE28BULL, "dae000_flemeth6013366.cl"},
	{0xE111CA4592373F56ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_4z.dds"},
	{0xE1120360BA906D75ULL, "6202104_m.fxe"},
	{0xE11308CFBB54CA03ULL, "v_cage_01.dds"},
	{0xE113502812409389ULL, "vxw_encounter2_kx_0_0.vlm"},
	{0xE113B43FB3F3CBD8ULL, "hlac_lt_0_lowtown_-1_3z.dds"},
	{0xE113D7A0C1485A45ULL, "w_sundermount_89_0_ll.mmh"},
	{0xE113DEE8E999038AULL, "vxlt_backalley_n_1eu_0_0.vlm"},
	{0xE11400C2578D3CCAULL, "tbe100_xenon6189678.cl"},
	{0xE1144B49D051BA60ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_shrug.evt"},
	{0xE114F70B4E53527FULL, "6630912.ogg"},
	{0xE1151C0708760795ULL, "hlac_ca_cavern1_haunted_24424_10z.dds"},
	{0xE116DFADE1AF4644ULL, "gen00fl_bethany6083988.cl"},
	{0xE1179CB104BB713EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ent.ani"},
	{0xE117E7C4E31C0A26ULL, "hlac_dr_primevaltomb_11985_2x.dds"},
	{0xE118441227ED311BULL, "vxlt_undercityserialkiller_15a_0_0.vlm"},
	{0xE11857D9DB1C685FULL, "art\\vfx\\combat\\vfx_gui_pp_mage_c.impa.ani"},
	{0xE1188F1DF0952D1CULL, "vxw_encounter2_camp_ki_0_0.vlm"},
	{0xE1189AE63E9A7D04ULL, "one100dg_lia6055162.cl"},
	{0xE118A7725A8E3EC1ULL, "sst_ss_lowclass_thugs_m_1.gda"},
	{0xE118D70B5BD4574AULL, "mcr370_negotiation6080167.cl"},
	{0xE1194CD658C0DA89ULL, "646059713.ogg"},
	{0xE119F45F4D1B6191ULL, "335353656.ogg"},
	{0xE11A08665CCCB073ULL, "vxdr_0_deeproads_cy1_0_0.vlm"},
	{0xE11A4AA1DF1096B0ULL, "912321398.ogg"},
	{0xE11A787BFF17BA17ULL, "hlac_dr_primevaltomb_659_5z.dds"},
	{0xE11ACD48F79534F4ULL, "lt_0_lowtown_a2_13_0.rml"},
	{0xE11B420387DE1D1EULL, "gen_im_trs_gem_snowobsid_04.uti"},
	{0xE11BD702294E4D14ULL, "vxlt_gallowscourtyard_50i_0_0.vlm"},
	{0xE11BF5B1B76BD7AEULL, "465915934.ogg"},
	{0xE11C197208A15F33ULL, "6178375_m.fxe"},
	{0xE11C1C32216665F6ULL, "character_character_p1.psh.fxo.dx11.assembly"},
	{0xE11C5921DEA49DABULL, "vxht_0_hightown_n_9x8_0_0.vlm"},
	{0xE11C9D2F615250E2ULL, "376077637.ogg"},
	{0xE11CB8907A901CD6ULL, "hlac_lt_0_lowtown_a2_n_c_131071_0y.dds"},
	{0xE11CEC0F1D2389C4ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_cast_c.anb"},
	{0xE11D3D809654FF14ULL, "vxht_mansion2_3cs_0_0.vlm"},
	{0xE11DDFFBEC04A234ULL, "371216594.ogg"},
	{0xE11DFDC7B031B682ULL, "6054832_m.fxe"},
	{0xE11EE2B13122B7C0ULL, "519510751.ogg"},
	{0xE11EE5A6A6E2E23BULL, "6171448_m.fxe"},
	{0xE11EE8386CBD4591ULL, "w_bonepit_210_0.mmh"},
	{0xE11EF4386CBD5A32ULL, "w_bonepit_210_0.mao"},
	{0xE11F02386CBD724FULL, "w_bonepit_210_0.msh"},
	{0xE11F405D7D8F70BAULL, "6163697_m.fxe"},
	{0xE11FB7763428D6A1ULL, "364756679.ogg"},
	{0xE1205679C33BC9ADULL, "6113631_m.fxe"},
	{0xE120A00CDCFD44F6ULL, "6005830_m.fxe"},
	{0xE120A1BB7C4E77F4ULL, "186140444.ogg"},
	{0xE1211A7E3751DD29ULL, "llac_ca_cavern1_haunted_24424_23.dds"},
	{0xE12165DDF564CE04ULL, "mer320_hubert6053693.cl"},
	{0xE12168ADF47CAA31ULL, "vxlt_warehouse_125_0_0.vlm"},
	{0xE1221BB4877227D9ULL, "vxw_encounter2_camp_h9_0_0.vlm"},
	{0xE122592FE6E6F7F3ULL, "76269841.ogg"},
	{0xE122882D01DCF856ULL, "lgt110_gxa_coa_spy6135798.cl"},
	{0xE122A22EC652CFF2ULL, "zz_one_debug6079311.cl"},
	{0xE122A68454221B4CULL, "6118060_m.fxe"},
	{0xE122C1C07AF35C86ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v1.evt"},
	{0xE12331451D3DC7B1ULL, "425973270.ogg"},
	{0xE123586F2ED6C79AULL, "vxca_caverns2prc_abn_0_0.vlm"},
	{0xE1237C02BF20649FULL, "227760090.ogg"},
	{0xE12388CC542B4F12ULL, "gen00fl_merrill6119656.cl"},
	{0xE123E5E2DB3B933DULL, "zz_followers_banter6158415.cl"},
	{0xE124203FBC15F59BULL, "mag101_aombanter6052771.cl"},
	{0xE124B4B104C6240FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ent.evt"},
	{0xE124C7047BAB5A78ULL, "vxdr_0_deeproads_6nv_0_0.vlm"},
	{0xE1253EE529CB1BF2ULL, "llac_lt_backalley_g_0_11.dds"},
	{0xE125E3B7617F7FCCULL, "vxlt_undercityencounter_13h_0_0.vlm"},
	{0xE12613FB7537EE5AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxe.msh"},
	{0xE12621FB75380593ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxe.mao"},
	{0xE12625FB75380CA0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_cavedrawxe.mmh"},
	{0xE12657115DA22E5DULL, "314990010.ogg"},
	{0xE1266C258CBC3451ULL, "6176418_m.fxe"},
	{0xE126D5F6C61F052BULL, "618937340.ogg"},
	{0xE126D622CD2031BDULL, "vrc221_bartrand6043485.cl"},
	{0xE12700FAE28B459CULL, "vxw_encounter2_camp_fn_0_0.vlm"},
	{0xE1271EEA1D15A32FULL, "6136643_m.fxe"},
	{0xE1275B55BBF3EA59ULL, "llac_lt_gallowstemplar_f_2391_9.dds"},
	{0xE1275EBC7B881064ULL, "follower_banter6173189.cl"},
	{0xE127A1BD321CB199ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hd3a.phy"},
	{0xE1280861E5178341ULL, "vxlt_gallowstemplar_2p9_0_0.vlm"},
	{0xE128CA94BA0A9865ULL, "vfx_chunkyparticle_p2.psh.fxo"},
	{0xE128EFD97A769ACCULL, "346912720.ogg"},
	{0xE1299B3CD9DB8CD1ULL, "dae000_arishok6082014.cl"},
	{0xE129A89ED20818F2ULL, "hlac_lt_0_lowtown_docks_-131074_2y.dds"},
	{0xE129D8C07AF9A2D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v1.gad"},
	{0xE12A0B0F1D2E48C7ULL, "art\\vfx\\combat\\vfx_m_arc_bolt_cast_c.mmh"},
	{0xE12A1935E8EE5402ULL, "562174582.ogg"},
	{0xE12A71BF7295F751ULL, "follower_banter6170262.cl"},
	{0xE12A8F179E6831CFULL, "6081619_m.fxe"},
	{0xE12AAF300AA6129CULL, "dre000_bodahn6031369.cl"},
	{0xE12B9059463FE227ULL, "gen00fl_isabela6135662.cl"},
	{0xE12BB5B104CC44B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ass_ent.gad"},
	{0xE12BB92B1BDC1AD6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_pull_back.evt"},
	{0xE12C42E60B7D1DCEULL, "vxlt_hangedman_n_2cu_0_0.vlm"},
	{0xE12CCF50480395E7ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.anb"},
	{0xE12D0C1677507F52ULL, "79135007.ogg"},
	{0xE12D578398D8FABBULL, "6174748_m.fxe"},
	{0xE12D6DDD99CE1EA0ULL, "gen00fl_varric6107312.cl"},
	{0xE12DBA952531AE80ULL, "dae000_mother6028527.cl"},
	{0xE12DCDCF3AA4B283ULL, "vxw_sundermount_21w_0_0.vlm"},
	{0xE12DFC9AE81022A3ULL, "mcr310_argument6074497.cl"},
	{0xE12E129529665676ULL, "198427366.ogg"},
	{0xE12E3888EF169181ULL, "vxca_cavern1_haunted_jgr_0_0.vlm"},
	{0xE12E55CF76964BC5ULL, "119466745.ogg"},
	{0xE12E6FD456D27E9FULL, "818919180.ogg"},
	{0xE12ECE2061A55E9CULL, "hlac_w_bonepitdestroyed_a3_65539_0x.dds"},
	{0xE12ED8584333E0ECULL, "vxlt_hangedman_2zw_0_0.vlm"},
	{0xE12F092B2620EB43ULL, "follower_banter6155874.cl"},
	{0xE12FD9C07AFE107FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_rng_atk_v1.ani"},
	{0xE13015C33977014EULL, "zz_fol_debug6097808.cl"},
	{0xE1308D9179684059ULL, "vxw_encounter2_camp_c8_0_0.vlm"},
	{0xE130C2A2CD85348AULL, "vxlt_backalley_n_1kw_0_0.vlm"},
	{0xE1317B1BC731B70FULL, "107463402.ogg"},
	{0xE131FE36B1980944ULL, "vxht_chantry_n_2io_0_0.vlm"},
	{0xE1320E6B3828B598ULL, "vxlt_gallowscourtyard_72q_0_0.vlm"},
	{0xE1326D6B5125156DULL, "hlac_lt_gallowscourtyard_-1_4y.dds"},
	{0xE132B02B1BE22A61ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_pull_back.gad"},
	{0xE1330DF522CC6829ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twra_0t.dds"},
	{0xE134399191823FB5ULL, "236563747.ogg"},
	{0xE134444F0044368FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_bloody_hands.gad"},
	{0xE1345400B1CA2979ULL, "276413336.ogg"},
	{0xE135063DB5F9200BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_chandelier_d.dds"},
	{0xE1353219F94E9338ULL, "662112778.ogg"},
	{0xE135DCE3E3C273CCULL, "16350731.ogg"},
	{0xE136301BC4926044ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1achar.msh"},
	{0xE136421BC4927EADULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1achar.mao"},
	{0xE136461BC49285E6ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_deadm1achar.mmh"},
	{0xE1364B6D97C6D5A6ULL, "61104286.ogg"},
	{0xE1368C63D12E0697ULL, "hlac_dr_dungeon_3214_7x.dds"},
	{0xE136FD2E79F6619EULL, "art\\vfx\\environment\\vfx_fog_ground_p.anb"},
	{0xE1371B25A6F450A4ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_6y.dds"},
	{0xE13745A90CDAB8C9ULL, "vxlt_undercityencounter_1rf_0_0.vlm"},
	{0xE1375FBC91197C08ULL, "6128662_f.fxe"},
	{0xE137939CE6A21E56ULL, "334720858.ogg"},
	{0xE1388CAC3F1733F4ULL, "251459028.ogg"},
	{0xE138AE54E81CC116ULL, "vxdr_primevaltomb_1cz_0_0.vlm"},
	{0xE138B12B1BE6970FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\shoulder_r_pull_back.ani"},
	{0xE138B82FACE1A783ULL, "520892230.ogg"},
	{0xE138FED93B034D8EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_basket.mmh"},
	{0xE1390AD93B036225ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_basket.mao"},
	{0xE13918D93B037A7CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_basket.msh"},
	{0xE13972F3EEA82614ULL, "677010276.ogg"},
	{0xE13984DCF7FBE47CULL, "dre000_cut_arrivaldeeproads6027613.cl"},
	{0xE13A1CB2C4CEBDF0ULL, "vxlt_gallowstemplar_n_5jz_0_0.vlm"},
	{0xE13A594F0048C63CULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_bloody_hands.evt"},
	{0xE13AA4890986565FULL, "vxlt_0_lowtown_a3_n_c_gyz_0_0.vlm"},
	{0xE13AF439AC5DF1EFULL, "203721406.ogg"},
	{0xE13B9A9D8EACC24EULL, "6099400_m.fxe"},
	{0xE13BBE9A17916AA1ULL, "6189427_m.fxe"},
	{0xE13BF498C099C53BULL, "w_sundermount_86_0.msh"},
	{0xE13C0698C099E386ULL, "w_sundermount_86_0.mao"},
	{0xE13C1298C099F825ULL, "w_sundermount_86_0.mmh"},
	{0xE13C32BCCB0D8F73ULL, "lt_0_lowtown_docks_103_0.phy"},
	{0xE13C764C6B84AEC2ULL, "6128807_m.fxe"},
	{0xE13CD5DE70E65974ULL, "gen00fl_aveline6120798.cl"},
	{0xE13CD7D79E182F9AULL, "pro000_opening_pc6001197.cl"},
	{0xE13D8F3CAA6164B5ULL, "hlac_w_encounter3_-262144_1y.dds"},
	{0xE13E2B6EEDA915E8ULL, "184146541.ogg"},
	{0xE13E55F73ABB71B0ULL, "6137683_m.fxe"},
	{0xE13E80B7EBB75D60ULL, "llac_lt_gallowstemplar_2384_3.dds"},
	{0xE13ECCDEB7960CBBULL, "6082122_m.fxe"},
	{0xE13EDD2E7F537B2BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile3a_1.mao"},
	{0xE13F347FBF7B4D0AULL, "vxlt_backalley_n_1vp_0_0.vlm"},
	{0xE13F4557D9BDAE6EULL, "llac_w_blightlands_65534_0.dds"},
	{0xE13F4A7717F9619AULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportout.gad"},
	{0xE13F7C49CA4B2DDCULL, "vxht_0_hightown_n_c3q_0_1.vlm"},
	{0xE13FA011A97C2704ULL, "vxlt_darktown_4z0_0_0.vlm"},
	{0xE13FC19D1A088DB7ULL, "114383070.ogg"},
	{0xE140014F8E4B87CEULL, "lt_0_lowtown_docks_n_191_0.mmh"},
	{0xE1400D4F8E4B9C65ULL, "lt_0_lowtown_docks_n_191_0.mao"},
	{0xE1401B4F8E4BB3BCULL, "lt_0_lowtown_docks_n_191_0.msh"},
	{0xE1416F9E2C31A760ULL, "mer140_bonepit_comments6062372.cl"},
	{0xE141957F41FCEE96ULL, "vxht_brothel_n_5j4_0_0.vlm"},
	{0xE14377D07845F5B1ULL, "6175514_m.fxe"},
	{0xE143BFFE4CF3D5A5ULL, "317106865.ogg"},
	{0xE143D3DA3AF310A8ULL, "6001094_m.fxe"},
	{0xE14403290BFC2157ULL, "dre000_cut2d_ragstoriches6025085.cl"},
	{0xE145D18FE3DFE357ULL, "gen00fl_isabela6090813.cl"},
	{0xE1464D7717FF85B1ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportout.evt"},
	{0xE14769FCB6581BA3ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockwall_2.mao"},
	{0xE147A00F92C1B53DULL, "vxht_bartrand_n_4qs_0_0.vlm"},
	{0xE147CA9A85A83B1FULL, "sib100_father_contact6109260.cl"},
	{0xE14850868EF997CEULL, "146910137.ogg"},
	{0xE148614F00551155ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_bloody_hands.ani"},
	{0xE1487285A080A7E3ULL, "vxca_cavern1_haunted_hm8_0_0.vlm"},
	{0xE14963519289475AULL, "hlac_ht_mansion2b_818_12y.dds"},
	{0xE14A0B848B3F6F59ULL, "dae000_paivel6134461.cl"},
	{0xE14A1413E51F6435ULL, "833978090.ogg"},
	{0xE14A2F577DEA4FECULL, "6024118_m.fxe"},
	{0xE14ACFF09AA3C24FULL, "dae100_viscount6102067.cl"},
	{0xE14B7C3DCFFB8F31ULL, "6035367_f.fxe"},
	{0xE14B9A921F837BBBULL, "vxca_cavern1_haunted_j4u_0_0.vlm"},
	{0xE14C208073F99E0DULL, "6000812_m.fxe"},
	{0xE14C2834099CF4F8ULL, "dae000_epic_merchant_4.cnv"},
	{0xE14C68451A9377E4ULL, "6176110_m.fxe"},
	{0xE14CB4C165460D43ULL, "seb221_flora_harimann6177685.cl"},
	{0xE14CE6B8F9E71625ULL, "llac_lt_0_lowtown_a2_n_c_65538_0.dds"},
	{0xE14D1DFC582BC745ULL, "vxca_caverns2_7w5_0_0.vlm"},
	{0xE14D7BEF2999817AULL, "vxca_caverns2_7f9_0_0.vlm"},
	{0xE14D8E7D258944DBULL, "260569548.ogg"},
	{0xE14DA3E116D4582FULL, "art\\vfx\\cinematic\\vfx_scaredmage.impa.gad"},
	{0xE14DD20E04EF5876ULL, "vxw_bonepit_10i_0_0.vlm"},
	{0xE14E11EE9B723F84ULL, "vxlt_darktown_4a1_0_0.vlm"},
	{0xE14EB1BF7CE30D3EULL, "6070504_m.fxe"},
	{0xE14F0B5331CD4823ULL, "vxdr_primevalentrance_16t_0_0.vlm"},
	{0xE14F691A616AF8C9ULL, "6026225_m.fxe"},
	{0xE1500A3409A0AA00ULL, "dae000_epic_merchant_4.bnk"},
	{0xE1502EE7020644F3ULL, "vxht_chantry_2d0_0_0.vlm"},
	{0xE15046FB7D4CC7EAULL, "gen00fl_sebastian6187766.cl"},
	{0xE1507695723089F1ULL, "482124631.ogg"},
	{0xE1508A12B815EDE5ULL, "721854516.ogg"},
	{0xE150B88E79FFC76DULL, "6121029_m.fxe"},
	{0xE151763FBFC4B8EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_ambient_other"},
	{0xE151A98C4E4DF1DCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_fall.ani"},
	{0xE1523B85EEE03558ULL, "102910283.ogg"},
	{0xE152625840E191CFULL, "6089884_m.fxe"},
	{0xE1526739B69184A4ULL, "6095756_m.fxe"},
	{0xE152A66BF619D1F6ULL, "6076519_m.fxe"},
	{0xE153263314A2D829ULL, "212200307.ogg"},
	{0xE153ADC753B0F3B3ULL, "llac_w_bonepitdestroyed_a3_0_5.dds"},
	{0xE153E11023425F9AULL, "6149679_m.fxe"},
	{0xE1542E5FBCABD200ULL, "mcr370st_negotiation.stg"},
	{0xE1546577180BEB80ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\teleportout.ani"},
	{0xE1546B43CA07EFE4ULL, "vxht_0_hightown_ajv_0_0.vlm"},
	{0xE154731DDA1A152FULL, "zz_dae_debug6075035.cl"},
	{0xE154B8E116DA9ADCULL, "art\\vfx\\cinematic\\vfx_scaredmage.impa.evt"},
	{0xE154CC504824E2E8ULL, "art\\vfx\\combat\\vfx_m_pri_tempest_aoe_p.mmh"},
	{0xE154DF92AA49BB54ULL, "hlac_w_encounter2_-65534_0x.dds"},
	{0xE15599268045AC6FULL, "vxlt_0_lowtown_a2_n_gm9_0_0.vlm"},
	{0xE155B07C4F6846F1ULL, "hlac_w_sundermount_-65537_0y.dds"},
	{0xE155CBFE53C086F9ULL, "hlac_lt_gallowscourtyard_n_-131072_0y.dds"},
	{0xE156019CA4CE012EULL, "931738148.ogg"},
	{0xE1566705EE63CFA8ULL, "zz_qun_debug6028269.cl"},
	{0xE156B3DECAD577D3ULL, "vxca_cave_18z_0_1.vlm"},
	{0xE1570450D1A69A31ULL, "6031702_m.fxe"},
	{0xE1570FE4C5186979ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_02stump.msh"},
	{0xE15719E4C5187A67ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_02stump.mmh"},
	{0xE1571DE4C51880B4ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_02stump.mao"},
	{0xE15742E7E3B0856AULL, "hlac_lt_0_lowtown_docks_a3_-131074_3y.dds"},
	{0xE1574FD0944E5C43ULL, "character_character_p141.psh.fxo.dx11.assembly"},
	{0xE15792150BE0820FULL, "llac_ht_0_hightown_65536_1.dds"},
	{0xE157B68C4E52731EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_fall.gad"},
	{0xE157B872A357C476ULL, "art\\vfx\\combat\\vfx_ha_hd_leap_act_c.anb"},
	{0xE1587E577298F34EULL, "6093600_m.fxe"},
	{0xE159155FF2A7114CULL, "6140652_m.fxe"},
	{0xE15984EF4B9BFC0BULL, "ave310_donnic6134801.cl"},
	{0xE159B4B167BB78A5ULL, "523802436.ogg"},
	{0xE159C933957B9E39ULL, "llac_w_encounter3_-327678_0.dds"},
	{0xE159E6A0D308357CULL, "gen_im_trs_drk_junk2_04.uti"},
	{0xE15A0C6F1179943BULL, "vxca_cavern1_haunted_j3t_0_0.vlm"},
	{0xE15A374EE3EDA4FFULL, "follower_banter6163408.cl"},
	{0xE15A8D17D1D9285FULL, "dre000_vampire6034295.cl"},
	{0xE15AB76855387647ULL, "900255781.ogg"},
	{0xE15BC3F96F9F9CE9ULL, "zz_followers_banter6157757.cl"},
	{0xE15BD237C201EB53ULL, "vxca_caverns2prc_62q_0_1.vlm"},
	{0xE15BD4B27E8D467DULL, "hlac_ht_0_hightown_n_-196608_0x.dds"},
	{0xE15C3A0E4CF56CC4ULL, "6172082_m.fxe"},
	{0xE15C43EAF6E570F6ULL, "vxw_bonepit_15j_0_0.vlm"},
	{0xE15C5A3ACA52D8CFULL, "6110544_m.fxe"},
	{0xE15C6FF7C8D75EA2ULL, "580996054.ogg"},
	{0xE15C83CB8D685804ULL, "vxlt_darktown_4l6_0_0.vlm"},
	{0xE15DB1E665FDA6DEULL, "6136677_m.fxe"},
	{0xE15E7F2039FA4A87ULL, "dae021_torelin_mob6150229.cl"},
	{0xE15E918C4E58533DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\punched_from_front_fall.evt"},
	{0xE15ED2C9D7E8C976ULL, "w_bonepit_2_0_lc.mmh"},
	{0xE15EFEB2C6EB6A03ULL, "genip_pp_lte_standing.utp"},
	{0xE15FDD1694077E19ULL, "6041741_m.fxe"},
	{0xE16002F3853AB724ULL, "vxht_brothel_n_5gq_0_0.vlm"},
	{0xE1600D4518683BCFULL, "568153771.ogg"},
	{0xE1605D738AABF398ULL, "hlac_lt_0_lowtown_n_131070_3y.dds"},
	{0xE160A0844761F17EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90_alt.evt"},
	{0xE161856D3A324E2CULL, "lla_ht_0_hightown_a3_-65537_0.dds"},
	{0xE161ABB57D3EC7FAULL, "gencr_mercenary_warrior_2w.utc"},
	{0xE161B6DA9BCF404BULL, "6081686_m.fxe"},
	{0xE161EB5B65897E9BULL, "mer320_hubert6079335.cl"},
	{0xE161F74853B222B2ULL, "art\\characters\\weapons\\boulder\\textures\\w_boulder_s.dds"},
	{0xE162A8EBBEDC9812ULL, "dae300_alistair_king6134968.cl"},
	{0xE162C0E116E6E5F5ULL, "art\\vfx\\cinematic\\vfx_scaredmage.impa.ani"},
	{0xE163462C72EAD7BCULL, "llac_ht_0_hightown_n_-196608_1.dds"},
	{0xE163AD4D7B92A9B4ULL, "hlac_w_encounter2_65538_0y.dds"},
	{0xE163E9EBF4C77DF6ULL, "hm_amb_old_gossip1.mor"},
	{0xE16410DB837EE19DULL, "1019772943.ogg"},
	{0xE16425848D6BB64DULL, "56744766.ogg"},
	{0xE1644AC14069E8E9ULL, "6071139_m.fxe"},
	{0xE164DAFD3B24901AULL, "vxlt_0_lowtown_a2_n_c_hsw_0_0.vlm"},
	{0xE1650639A650565CULL, "6164508_m.fxe"},
	{0xE165BB72A36406E1ULL, "art\\vfx\\combat\\vfx_ha_hd_leap_act_c.mmh"},
	{0xE165C54D38983418ULL, "mag370_meredith6039851.cl"},
	{0xE167325F4A09229FULL, "vxht_0_hightown_n_c0v_0_0.vlm"},
	{0xE1673649E94CBE89ULL, "vxht_brothel_n_6go_0_0.vlm"},
	{0xE167B784476836A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90_alt.gad"},
	{0xE1680725DEB3E027ULL, "ss_carta_thug_m.wbd"},
	{0xE1687A8D8C62504AULL, "vxht_0_hightown_a3_9ux_0_0.vlm"},
	{0xE168C679F801147FULL, "gen00fl_anders6113105.cl"},
	{0xE169F55BE428637EULL, "vxlt_warehouse_5r_0_0.vlm"},
	{0xE16A0B0815D2752DULL, "vxht_mansion_n_6c0_0_0.vlm"},
	{0xE16A31858E15992DULL, "6103249_m.fxe"},
	{0xE16A3344FF96A8DEULL, "lt_0_lowtown_a3_n_c_19_0.rml"},
	{0xE16AB91E6520D86DULL, "w_sundermount_242_0.msh"},
	{0xE16AC71E6520EFA0ULL, "w_sundermount_242_0.mao"},
	{0xE16ACB1E6520F6F3ULL, "w_sundermount_242_0.mmh"},
	{0xE16ACFF8CC9AD1E4ULL, "6080798_m.fxe"},
	{0xE16AE0F73F2DC640ULL, "gen00fl_isabela6109798.cl"},
	{0xE16AF5A87F5E7084ULL, "vxlt_darktown_4k7_0_0.vlm"},
	{0xE16B1BCAD2B82F2CULL, "qun100_mage_party_comment6019120.cl"},
	{0xE16B1E5A646CD0CEULL, "vxlt_darktown_5gh_0_0.vlm"},
	{0xE16BDFD95AE584A3ULL, "vxht_0_hightown_n_ap0_0_0.vlm"},
	{0xE16C8A4C4EE0759CULL, "45125795.ogg"},
	{0xE16DB4A57FCD1D41ULL, "zz_followers_banter6158910.cl"},
	{0xE16DDAD22BA10897ULL, "lt_0_lowtown_a2_n_c_water_21225.mmh"},
	{0xE16DDED22BA10FA4ULL, "lt_0_lowtown_a2_n_c_water_21225.mao"},
	{0xE16DF0D22BA12E09ULL, "lt_0_lowtown_a2_n_c_water_21225.msh"},
	{0xE16E612C183FADF8ULL, "6082118_m.fxe"},
	{0xE16E74D07730CFA4ULL, "vxht_brothel_n_5jv_0_0.vlm"},
	{0xE16E7B5F67868C8CULL, "pro000_meeran6048811.cl"},
	{0xE16E8884476E05F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\turn_left_90_alt.ani"},
	{0xE16F26825AF20316ULL, "957377435.ogg"},
	{0xE16FFD02CDEE4D1AULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dazed.gad"},
	{0xE17032943812778DULL, "llac_ht_keep_n_4519_17.dds"},
	{0xE1712BB19040D530ULL, "lgt120_orz_bln_official6140204.cl"},
	{0xE1717278FD41DB0FULL, "271191185.ogg"},
	{0xE17204CB148DB9BEULL, "543565518.ogg"},
	{0xE17263F8F8895105ULL, "6061794_m.fxe"},
	{0xE17272CC0C325E15ULL, "954394212.ogg"},
	{0xE17287E777C125A7ULL, "vxlt_0_lowtown_n_eo0_0_1.vlm"},
	{0xE172C0C666331C56ULL, "6153809_m.fxe"},
	{0xE172D6659BAFB6C8ULL, "character_character_p21.psh.fxo.dx11"},
	{0xE1730FA138D495AFULL, "vxw_woundedcoast_13f_0_0.vlm"},
	{0xE1731C6DEEBDF701ULL, "6042546_m.fxe"},
	{0xE173C44D125A8382ULL, "hlac_lt_gallowscourtyard_n_-65538_0y.dds"},
	{0xE173C8FCBFA84595ULL, "t3_wdg_tr1.tnt"},
	{0xE1741D90B1C01147ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_013a_0t.dds"},
	{0xE174920F364A0C66ULL, "gen00fl_anders6116163.cl"},
	{0xE174C2E5C1F0FC6AULL, "zz_dre_debug6054629.cl"},
	{0xE174E947AA7AFA41ULL, "zz_lgt_1_debug6114355.cl"},
	{0xE1750881933DFAEAULL, "6060605_f.fxe"},
	{0xE1753B91B343C27AULL, "hlac_dr_0_deeproads_5845_11z.dds"},
	{0xE1756F3C4755E92FULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb103a_0d.dds"},
	{0xE176966C3A5ABBAEULL, "character_character_p96.psh.fxo"},
	{0xE1770002CDF47131ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dazed.evt"},
	{0xE17797B66EFB7918ULL, "471818544.ogg"},
	{0xE177C2D5B9879FBBULL, "590599479.ogg"},
	{0xE1783F09FB446488ULL, "hlac_w_woundedcoast_0_5x.dds"},
	{0xE178CD4FFF4AD87BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordb_n.dds"},
	{0xE1794A7EFB4919AEULL, "6115384_m.fxe"},
	{0xE179546B74E6FE03ULL, "959458427.ogg"},
	{0xE17962D963DBE796ULL, "qun111_ketojan6019022.cl"},
	{0xE17989473D4FC969ULL, "dre000_sibling_has_blight6037050.cl"},
	{0xE1799071EEE887D3ULL, "hlac_dr_dungeon_3214_14z.dds"},
	{0xE17990BB061E44C4ULL, "vxlt_0_lowtown_a2_n_c_h2j_0_0.vlm"},
	{0xE17A30A74D3049BDULL, "gen00fl_merrill6145146.cl"},
	{0xE17A4234D74FCEE4ULL, "6013291_m.fxe"},
	{0xE17A7B1AFCAFE29EULL, "6180801_m.fxe"},
	{0xE17B3564F2B54522ULL, "hlac_w_blightlands_-131072_0z.dds"},
	{0xE17B518D63E2AB6FULL, "llac_ca_caverns1_24426_13.dds"},
	{0xE17B5785A6E23F55ULL, "6091861_m.fxe"},
	{0xE17B9C868FF3954CULL, "zz_spy132cr_booker.utc"},
	{0xE17BF0F1F7C1B6EAULL, "189635181.ogg"},
	{0xE17C1298C0CFDC2EULL, "w_sundermount_86_0.rml"},
	{0xE17C99C753CFC388ULL, "6149866_m.fxe"},
	{0xE17D0215534AF6E6ULL, "6059304_f.fxe"},
	{0xE17D2D7B90E9459AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_011a_0s.dds"},
	{0xE17D2F8D14E4697CULL, "gen00fl_anders6125325.cl"},
	{0xE17D7115AAD3B5F7ULL, "vxca_cavern1_haunted_j4a_0_0.vlm"},
	{0xE17DA50F701521DBULL, "vxlt_0_lowtown_eov_0_0.vlm"},
	{0xE17DB2A1E986647FULL, "354275067.ogg"},
	{0xE17EBB2D00E0EF43ULL, "nat380_nathaniel6068347.cl"},
	{0xE17F13E4F5A1D792ULL, "6121141_m.fxe"},
	{0xE17F39B63244C146ULL, "6144302_m.fxe"},
	{0xE17F40A72385D239ULL, "mag360_thrask6043567.cl"},
	{0xE17F5980FABA15FAULL, "548935228.ogg"},
	{0xE17F6FA7CC824639ULL, "263800590.ogg"},
	{0xE17F8508F566E71DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15a.phy"},
	{0xE17F8E464AA4C19AULL, "vxlt_warehouse_193_0_0.vlm"},
	{0xE17FD1EBC670FDAAULL, "vxw_bonepitdestroyed_a3_24b_0_0.vlm"},
	{0xE17FEF608A922FDEULL, "6087429_m.fxe"},
	{0xE180395B3ABBB440ULL, "dae210_gate_guard6209978.cl"},
	{0xE1803ED8C2A040A0ULL, "ogre_heads.gda"},
	{0xE1804715FE513E7BULL, "gen00fl_merrill6119592.cl"},
	{0xE181750E302C5FF7ULL, "6129597_m.fxe"},
	{0xE18193A98AA95598ULL, "t1_mub_pi3.tnt"},
	{0xE181C157092E6A02ULL, "6143299_m.fxe"},
	{0xE18220E26CDA0DDDULL, "842542028.ogg"},
	{0xE18244DCC0A3F51FULL, "mrl250pt_varterral.ncs"},
	{0xE1828E89251CC20FULL, "69924267.ogg"},
	{0xE182EB66097F2DFBULL, "art\\vfx\\combat\\vfx_c_single_tag_stun_c.impa.evt"},
	{0xE182EB98C0D5B8D2ULL, "w_sundermount_86_0.phy"},
	{0xE18302E4C53D8D9CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_02stump.phy"},
	{0xE1836A4E3464C817ULL, "dae000_bartrand6206633.cl"},
	{0xE183FF9CD7BB3F09ULL, "w_bonepitdestroyed_a3_48_0.rml"},
	{0xE1849DE748AB3AA1ULL, "lgt210_eau_symp_templar6144457.cl"},
	{0xE1850638ABE2FA60ULL, "vxht_0_hightown_bcv_0_0.vlm"},
	{0xE18508492C50D42EULL, "character_character_p41.psh.fxo"},
	{0xE1851802CE00D700ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_dazed.ani"},
	{0xE186816BA68BFDA1ULL, "656192846.ogg"},
	{0xE186C39AE06D1BC1ULL, "335625412.ogg"},
	{0xE186F7B60B5E4CD6ULL, "6110439_m.fxe"},
	{0xE18711A3C5A13890ULL, "vxlt_0_lowtown_a2_n_c_goj_0_0.vlm"},
	{0xE1876C8152AC088AULL, "424574332.ogg"},
	{0xE1876F82988C0661ULL, "327875766.ogg"},
	{0xE187ACF1C302B946ULL, "6091826_m.fxe"},
	{0xE187D9EDB27499ACULL, "m_one130ar_disciple_dungeon.bnk"},
	{0xE1887811CBA3A208ULL, "vxlt_backalley_n_1dx_0_0.vlm"},
	{0xE188799087314301ULL, "vxlt_backalley_g_1ew_0_0.vlm"},
	{0xE188C06574218CAEULL, "gen00fl_aveline6116713.cl"},
	{0xE1896A89160C87E0ULL, "hlac_ht_mansion2b_818_4z.dds"},
	{0xE189B1CD60A9480FULL, "follower_banter6161345.cl"},
	{0xE189CF6395315500ULL, "hlac_ht_brothel_n_3369_1y.dds"},
	{0xE189E4660985403CULL, "art\\vfx\\combat\\vfx_c_single_tag_stun_c.impa.gad"},
	{0xE189F65AC9DEA6F3ULL, "vxht_chantry_2p4_0_0.vlm"},
	{0xE18A7C0FBDFD33A2ULL, "sib131_slave_a6165272.cl"},
	{0xE18AAE50220F8EE3ULL, "fen310_party_comment6116850.cl"},
	{0xE18B00C1FF3A2B4CULL, "vxlt_undercityencounter_1nc_0_0.vlm"},
	{0xE18B1EB68ED5CA90ULL, "lt_0_lowtown_docks_n_115_0.msh"},
	{0xE18B2CB68ED5E2AAULL, "lt_0_lowtown_docks_n_115_0.mmh"},
	{0xE18B30B68ED5E9E1ULL, "lt_0_lowtown_docks_n_115_0.mao"},
	{0xE18B889581C479F7ULL, "zz_dre_debug6168288.cl"},
	{0xE18B90C339830C1EULL, "dae000_pc_home_ambients6208697.cl"},
	{0xE18B9234350CF533ULL, "6080112_f.fxe"},
	{0xE18C08572AF402F6ULL, "vxca_caverns1_e7o_0_0.vlm"},
	{0xE18C16EC620B3A5BULL, "vxlt_0_lowtown_eji_0_0.vlm"},
	{0xE18DB8F40A571E21ULL, "hlac_lt_0_lowtown_a2_n_131070_3z.dds"},
	{0xE18DC884A73B7FACULL, "art\\toolset\\_textures\\iconrgby_n.dds"},
	{0xE18E1E87310B7EFCULL, "hlac_lt_0_lowtown_docks_a3_131070_1x.dds"},
	{0xE18E27171A65D514ULL, "119721627.ogg"},
	{0xE18E8002FBA9FD8CULL, "vxw_sundermount_354_0_1.vlm"},
	{0xE18EF39CD7C532E9ULL, "w_bonepitdestroyed_a3_48_0.mao"},
	{0xE18EFF9CD7C54682ULL, "w_bonepitdestroyed_a3_48_0.mmh"},
	{0xE18F019CD7C54A38ULL, "w_bonepitdestroyed_a3_48_0.msh"},
	{0xE18F99731B0CA0FDULL, "dae000_fem_commoner1_mob6174489.cl"},
	{0xE18F9BCC05C9E78CULL, "vxht_brothel_n_60g_0_0.vlm"},
	{0xE18FA3E2C3F31D97ULL, "genip_quill_pen.utp"},
	{0xE18FF05D0A3564BFULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.dura.gad"},
	{0xE1908608F5755C22ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15a.mmh"},
	{0xE1908808F5755FD8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos15a.msh"},
	{0xE190B5E6AC73FCA3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.opentoclosed.ani"},
	{0xE190BDA28ACAF85BULL, "dae000_pc_home_ambients6208320.cl"},
	{0xE190BDEE5812202FULL, "zz_vault_debug6177771.cl"},
	{0xE190E366098B5D3AULL, "art\\vfx\\combat\\vfx_c_single_tag_stun_c.impa.ani"},
	{0xE191F751ADB1DA25ULL, "vxw_bonepit_177_0_0.vlm"},
	{0xE1920B3CF7BF84E1ULL, "w_sundermount_5_0.rml"},
	{0xE1922BCDE7EEA00FULL, "lt_0_lowtown_a2_n_c_53_0_ll.mmh"},
	{0xE192A1D0A14BCB6AULL, "6114295_m.fxe"},
	{0xE192B3A72EB3BC94ULL, "vxw_encounter3_ds_0_0.vlm"},
	{0xE192C62389718581ULL, "125073005.ogg"},
	{0xE192D860AF1E2536ULL, "82787671.ogg"},
	{0xE19378159DD912E0ULL, "vxht_0_hightown_blw_0_0.vlm"},
	{0xE1937A261E46ECAEULL, "character_character_p30.psh.fxo"},
	{0xE1938CAF15E05A14ULL, "vxca_caverns1_ik5_0_0.vlm"},
	{0xE193CE2F9EF03AF3ULL, "gen00fl_varric6033713.cl"},
	{0xE193D5A1B3E99DDEULL, "ht_keep_3734.rml"},
	{0xE194D9BFB0B61E41ULL, "hlac_lt_0_lowtown_docks_a3_n_196607_2z.dds"},
	{0xE195080E538DD24EULL, "6119902_m.fxe"},
	{0xE1951F6BB284457DULL, "dae000ip_crowd_lm4_talk9.utp"},
	{0xE19538EB50601855ULL, "vxlt_gallowstemplar_f_2mu_0_0.vlm"},
	{0xE1959B0CA96A5225ULL, "ave321tr_varric_check.utt"},
	{0xE195C6D2881D8009ULL, "6110256_m.fxe"},
	{0xE196040056A9D2C1ULL, "793076663.ogg"},
	{0xE196D93BBEBAC0A6ULL, "isa111_isabela6066769.cl"},
	{0xE196E55D0A3B70ACULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.dura.evt"},
	{0xE1971D4ACE8D2BA6ULL, "vxw_woundedcoast_of_0_0.vlm"},
	{0xE197A714C9684BA3ULL, "6050114_m.fxe"},
	{0xE1982FDA5C8783E5ULL, "one300_murder_followers6086895.cl"},
	{0xE1983CC2623D33F2ULL, "6151366_m.fxe"},
	{0xE1984DA938195F5FULL, "llac_lt_gallowstemplar_f_2393_1.dds"},
	{0xE19901F225C73C51ULL, "vxlt_0_lowtown_a2_n_eoc_0_0.vlm"},
	{0xE1992D8D2277B31BULL, "atl_loginconne_dxt5_00.dds"},
	{0xE199F0A6D515EE1FULL, "vxw_woundedcoast_i5_0_0.vlm"},
	{0xE19A168BD59CC448ULL, "vrc321ip_spooky_3_owner.utp"},
	{0xE19AADB8B8A83F14ULL, "vxlt_gallowstemplar_n_46g_0_0.vlm"},
	{0xE19BC49DE09CCB10ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothb.mao"},
	{0xE19BC89DE09CD223ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothb.mmh"},
	{0xE19BD69DE09CEA7DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothb.msh"},
	{0xE19C3C5BDA863578ULL, "hlac_dr_slavers_3217_5z.dds"},
	{0xE19C4532BFEC5D5AULL, "vxht_mansion2_1tf_0_0.vlm"},
	{0xE19C6A2AAD565EA4ULL, "ht_0_hightown_a3_8_0.msh"},
	{0xE19C7C2AAD567D0DULL, "ht_0_hightown_a3_8_0.mao"},
	{0xE19C802AAD568446ULL, "ht_0_hightown_a3_8_0.mmh"},
	{0xE19CEC1A0B962B9FULL, "6125631_f.fxe"},
	{0xE19D0467E42974B3ULL, "vxlt_darktown_5aw_0_0.vlm"},
	{0xE19D10D384B8E55CULL, "6137540_m.fxe"},
	{0xE19D8ABDB0FC7060ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01wisptree.phy"},
	{0xE19DE6497A2ADC4DULL, "402668995.ogg"},
	{0xE19DFCEB142A26CBULL, "scaleform_sf_main_p1.vsh.fxo.dx11.assembly"},
	{0xE19E2BBCCB604DE8ULL, "lt_0_lowtown_docks_103_0.mmh"},
	{0xE19E37BCCB60618BULL, "lt_0_lowtown_docks_103_0.mao"},
	{0xE19E49BCCB608012ULL, "lt_0_lowtown_docks_103_0.msh"},
	{0xE19E82E70236E06DULL, "615116045.ogg"},
	{0xE19ECDE6AC80629AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.opentoclosed.evt"},
	{0xE19ED433C1BD4597ULL, "261362678.ogg"},
	{0xE19EE3878111C7EEULL, "6130125_m.fxe"},
	{0xE19F388931623691ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_loop.ani"},
	{0xE19F670899610F02ULL, "lowtown_argument6169376.cl"},
	{0xE19F868B94E37194ULL, "vxlt_0_lowtown_a2_e8l_0_0.vlm"},
	{0xE1A0AC365D5A1692ULL, "984419913.ogg"},
	{0xE1A12550AE33346CULL, "w_encounter2_camp_13_0.msh"},
	{0xE1A12B50AE333E5EULL, "w_encounter2_camp_13_0.mmh"},
	{0xE1A13750AE3352F5ULL, "w_encounter2_camp_13_0.mao"},
	{0xE1A13B0C40CA0DFAULL, "vxlt_0_lowtown_a2_n_c_ehc_0_0.vlm"},
	{0xE1A211BD271B5BD1ULL, "hlac_lt_0_lowtown_docks_n_131071_7z.dds"},
	{0xE1A29D141478AE63ULL, "vxlt_0_lowtown_a3_n_c_hp2_0_0.vlm"},
	{0xE1A32C0AED3B93EDULL, "258137966.ogg"},
	{0xE1A34E185CCD3DE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_twitch.evt"},
	{0xE1A39A4E0B2FAC1BULL, "204512452.ogg"},
	{0xE1A3F0DB8FAD35D7ULL, "6169607_m.fxe"},
	{0xE1A411227990FB98ULL, "llac_lt_0_lowtown_a2_131073_2.dds"},
	{0xE1A43044FFC81B82ULL, "lt_0_lowtown_a3_n_c_19_0.anb"},
	{0xE1A43D628212BA57ULL, "vxlt_0_lowtown_docks_a3_n_co9_0_0.vlm"},
	{0xE1A4CD5D0A478505ULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.dura.ani"},
	{0xE1A5498817BD970BULL, "vxdr_0_deeproads_f54_0_0.vlm"},
	{0xE1A55BCBAF8FD308ULL, "vxlt_backalley_n_1jv_0_0.vlm"},
	{0xE1A578F028D3D96BULL, "art\\vfx\\combat\\vfx_r_modal_act_c.impa.ani"},
	{0xE1A5CCE6AC867F8DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_door_slab.opentoclosed.gad"},
	{0xE1A6280AB0AC1575ULL, "vxht_chantry_32q_0_0.vlm"},
	{0xE1A6438931686833ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_loop.gad"},
	{0xE1A6B48F8FACBE64ULL, "hlac_ht_chantry_n_2534_23y.dds"},
	{0xE1A6C83BE3E5BD96ULL, "llac_w_encounter3_-65537_0.dds"},
	{0xE1A773CF17BD54D1ULL, "vxlt_0_lowtown_a2_n_epb_0_0.vlm"},
	{0xE1A7772AAD607C11ULL, "ht_0_hightown_a3_8_0.phy"},
	{0xE1A77927AB0B5F33ULL, "6090240_m.fxe"},
	{0xE1A833D1FBF4C573ULL, "zz_ave_debug6175054.cl"},
	{0xE1A86283C70C069FULL, "vxw_woundedcoast_f4_0_0.vlm"},
	{0xE1A862F211FE714DULL, "vxlt_0_lowtown_e8e_0_0.vlm"},
	{0xE1A8D2939CE34200ULL, "vxlt_0_lowtown_a2_n_fe7_0_0.vlm"},
	{0xE1A91125DEEB885DULL, "ss_carta_thug_m.bnk"},
	{0xE1A9161A20EBAD79ULL, "llac_ca_cavern1_haunted_24426_5.dds"},
	{0xE1A94C0B0726D1EAULL, "hlac_w_encounter2_65536_0y.dds"},
	{0xE1A9CB229FC6A9D8ULL, "765540797.ogg"},
	{0xE1AA65185CD38331ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_twitch.gad"},
	{0xE1AA76FA06100564ULL, "gen00fl_fenris6110877.cl"},
	{0xE1AAA2B22043EAE3ULL, "vxca_cavern1_haunted_e5y_0_0.vlm"},
	{0xE1AB2BCB87F63A7EULL, "673574631.ogg"},
	{0xE1AB627A0C3A128CULL, "ef_gen_dalish_warrior.mor"},
	{0xE1AB7644D61F8D33ULL, "vxlt_darktown_5jv_0_0.vlm"},
	{0xE1AB9E19CB9D50CEULL, "fen310_fenris6118234.cl"},
	{0xE1ABD7B563926654ULL, "968964510.ogg"},
	{0xE1ACB852566B56AFULL, "dae000cr_grand_cleric6147675.cl"},
	{0xE1ACC99DE0AB4748ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothb.phy"},
	{0xE1ACFFC0F741DCCEULL, "476930045.ogg"},
	{0xE1AD4089316E8198ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_loop.evt"},
	{0xE1AD54B3CD487D3BULL, "6070528_f.fxe"},
	{0xE1ADB4FF34A8FB21ULL, "gen00fl_aveline6117643.cl"},
	{0xE1AE0A0ED146B686ULL, "6095637_m.fxe"},
	{0xE1AE6619608A96DAULL, "qun221_viscount6070477.cl"},
	{0xE1AE6F6FBAC9DC95ULL, "dre000_varric6027757.cl"},
	{0xE1AE91EA60FC4651ULL, "gen00fl_aveline6102076.cl"},
	{0xE1AED60947693031ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_lp.ani"},
	{0xE1AF4357C9532D4CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic.phy"},
	{0xE1AF6E5D3BD13352ULL, "vxw_bonepitdestroyed_a3_lj_0_0.vlm"},
	{0xE1AF893EDF888E29ULL, "gen00fl_anders6119398.cl"},
	{0xE1AFEEF62B5A992BULL, "vxca_cavern1_haunted_j21_0_2.vlm"},
	{0xE1B059DB686E3D77ULL, "405117041.ogg"},
	{0xE1B0DFE89D17C839ULL, "159854744.ogg"},
	{0xE1B117596D597E6DULL, "vxlt_hangedman_n_30v_0_0.vlm"},
	{0xE1B166185CD9A3DFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_twitch.ani"},
	{0xE1B1CA966AC99F81ULL, "536270998.ogg"},
	{0xE1B1E1F4F132DD77ULL, "6119319_f.fxe"},
	{0xE1B260F028DE3AE2ULL, "art\\vfx\\combat\\vfx_r_modal_act_c.impa.evt"},
	{0xE1B297B4EA87317BULL, "vxht_0_hightown_a2_a6m_0_0.vlm"},
	{0xE1B2B0F23D8BD5E5ULL, "ico_potion_elfroot.dds"},
	{0xE1B3DEC34198B837ULL, "552550302.ogg"},
	{0xE1B40104B2795CA6ULL, "vxw_woundedcoast_ah_0_0.vlm"},
	{0xE1B4C45853898AA6ULL, "dae200_bran6142982.cl"},
	{0xE1B53A38DD203068ULL, "vxht_mansion2b_1c9_0_0.vlm"},
	{0xE1B542EFBBD29CFEULL, "6099209_f.fxe"},
	{0xE1B552792D17B0C1ULL, "center.mao"},
	{0xE1B5E109476F61D3ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\d_lp.gad"},
	{0xE1B5E5AC09B36D51ULL, "vxlt_0_lowtown_a2_n_eua_0_0.vlm"},
	{0xE1B61B23CA4CE0ECULL, "vxlt_0_lowtown_a2_hde_0_0.vlm"},
	{0xE1B6487A9085A117ULL, "6179256_f.fxe"},
	{0xE1B6D460B9021F1FULL, "vxw_woundedcoast_c7_0_0.vlm"},
	{0xE1B6D4CF03F489CDULL, "vxlt_0_lowtown_e3d_0_0.vlm"},
	{0xE1B6F68E09FF1ADDULL, "116682827.ogg"},
	{0xE1B8589351D6A986ULL, "and121_karl6116255.cl"},
	{0xE1B8DB96BE6DBAE7ULL, "dae000_madam_lusine6152076.cl"},
	{0xE1B8F4884ED0E3AFULL, "zz_followers_banter6159393.cl"},
	{0xE1B9140EFC570A66ULL, "6172017_m.fxe"},
	{0xE1B956BD754FD353ULL, "gen_im_trs_qun_biscuit_03.uti"},
	{0xE1B95FF028E457F5ULL, "art\\vfx\\combat\\vfx_r_modal_act_c.impa.gad"},
	{0xE1B9E821C815A5B3ULL, "vxlt_darktown_5oy_0_0.vlm"},
	{0xE1BA03C8415091DCULL, "152846305.ogg"},
	{0xE1BA3F0105832F5EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_sidetable_0.pwk"},
	{0xE1BBB843E4DE101DULL, "266599947.ogg"},
	{0xE1BC8057614C58A6ULL, "darktown_fem_fer_urch_sta6174784.cl"},
	{0xE1BD7D27CD927261ULL, "516053092.ogg"},
	{0xE1BE675836AE87C2ULL, "hlac_ca_caverns2prc_5802_57x.dds"},
	{0xE1BEA58E04266E19ULL, "hlac_lt_0_lowtown_a2_2_0z.dds"},
	{0xE1BED38281B62C0AULL, "6155249_m.fxe"},
	{0xE1BEF282CED2BFD1ULL, "6034687_f.fxe"},
	{0xE1BF8FCA9AA03F59ULL, "el_lr_inside_day.mmh"},
	{0xE1BFAD5F76C5A19CULL, "6181831_m.fxe"},
	{0xE1BFFAE17E984F4BULL, "vxw_sundermount_35g_0_1.vlm"},
	{0xE1C012929B332425ULL, "vxlt_backalley_g_1nv_0_0.vlm"},
	{0xE1C032008DE47C87ULL, "6069419_m.fxe"},
	{0xE1C05809F9307E17ULL, "vxw_encounter2_camp_n1_0_0.vlm"},
	{0xE1C0617F5EE033FAULL, "6064888_m.fxe"},
	{0xE1C093F370A68004ULL, "440329259.ogg"},
	{0xE1C1524EEE4292C5ULL, "225745789.ogg"},
	{0xE1C17CFBDBA26511ULL, "w_blightlands_204_0.msh"},
	{0xE1C18AFBDBA27D2CULL, "w_blightlands_204_0.mao"},
	{0xE1C196FBDBA291CFULL, "w_blightlands_204_0.mmh"},
	{0xE1C1A7F14F869781ULL, "44099940.ogg"},
	{0xE1C1BB7E143A15CCULL, "6147125_m.fxe"},
	{0xE1C23F85937C0408ULL, "vxlt_backalley_n_1pt_0_0.vlm"},
	{0xE1C26099A1B4EE9FULL, "429030461.ogg"},
	{0xE1C26B5A2C78936DULL, "vxca_caverns1_j1w_0_2.vlm"},
	{0xE1C2871E343A5CBCULL, "fen110_fenris6096176.cl"},
	{0xE1C2B1033C6CF03FULL, "6000905_f.fxe"},
	{0xE1C39C9E3DC49530ULL, "vrc321_varric6082694.cl"},
	{0xE1C3A7BDB11D70F5ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01wisptree.msh"},
	{0xE1C3A9BDB11D746BULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01wisptree.mmh"},
	{0xE1C3ABB8AEEF6BAAULL, "art\\characters\\creatures\\spider\\animation\\tn90l.gad"},
	{0xE1C3B0598F9C0BCDULL, "197368425.ogg"},
	{0xE1C3B3AEC880FDAAULL, "vxlt_0_lowtown_cmo_0_0.vlm"},
	{0xE1C3B5BDB11D8808ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_01wisptree.mao"},
	{0xE1C41D762CE57AFAULL, "6139539_m.fxe"},
	{0xE1C44ECE8E171785ULL, "6129836_m.fxe"},
	{0xE1C489CFDEEE77C8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic_3.mao"},
	{0xE1C4C8084DD19DBBULL, "6111929_m.fxe"},
	{0xE1C501D2C08D3DB9ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_wkf.ani"},
	{0xE1C5259808227E14ULL, "vxht_0_hightown_n_bek_0_0.vlm"},
	{0xE1C73B1EDB7A9721ULL, "117368304.ogg"},
	{0xE1C75C9783CA555BULL, "darktown_fem_elf_sta6174513.cl"},
	{0xE1C76421377ADEE3ULL, "527554993.ogg"},
	{0xE1C7D14BAB3C66C1ULL, "hlac_lt_warehouse_n_-65536_0y.dds"},
	{0xE1C7F43B33ED9B72ULL, "framebuffereffects_radialblur_p0.psh.fxo.dx11.assembly"},
	{0xE1C8045F8BF3688EULL, "llac_w_bonepitdestroyed_a3_131076_0.dds"},
	{0xE1C8477689DA6C8AULL, "follower_banter6136676.cl"},
	{0xE1C859FEBA0BBE33ULL, "vxlt_darktown_5px_0_0.vlm"},
	{0xE1C8A63BC257F040ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\baseswordshield\\blendtree\\commander_overlay.anb"},
	{0xE1C8E3EDB2AC41DEULL, "m_one130ar_disciple_dungeon.wbd"},
	{0xE1C9222046466717ULL, "vxca_caverns1_jw1_0_0.vlm"},
	{0xE1C99C269B034558ULL, "vxlt_0_lowtown_a2_n_c_hlh_0_0.vlm"},
	{0xE1C99D79B69D0DF7ULL, "pro000_opening6178515.cl"},
	{0xE1C9F10AF67B35E2ULL, "582379930.ogg"},
	{0xE1CA41FE5DC72A1BULL, "hlac_lt_0_lowtown_a3_n_c_1_9z.dds"},
	{0xE1CA682EEA1D2399ULL, "vxlt_hangedman_n_27j_0_0.vlm"},
	{0xE1CA8EB8AEF55941ULL, "art\\characters\\creatures\\spider\\animation\\tn90l.evt"},
	{0xE1CA93811A73D812ULL, "37614478.ogg"},
	{0xE1CA9B2C707D3C92ULL, "vxht_mansion2_2er_0_0.vlm"},
	{0xE1CAACFCF60B1814ULL, "dae100_alistair_drunk6135119.cl"},
	{0xE1CB0F0342C8C857ULL, "gen00fl_varric6019887.cl"},
	{0xE1CB138B4D60ED4EULL, "6080091_m.fxe"},
	{0xE1CB247258A2C1BDULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_greyrubble_d.dds"},
	{0xE1CB6532643799BCULL, "601215263.ogg"},
	{0xE1CB9C246FB8F4EEULL, "148412516.ogg"},
	{0xE1CBD0ED40184A0DULL, "889863043.ogg"},
	{0xE1CBDCD2C0931DFBULL, "art\\characters\\creatures\\ragedemon\\animation\\p_wkf.gad"},
	{0xE1CBE402A70F386EULL, "dae000ip_crowd_lm2_beg.utp"},
	{0xE1CBFD0DF45C0A89ULL, "llac_lt_0_lowtown_-262144_4.dds"},
	{0xE1CC142B047A2313ULL, "vxht_bartrand_n_63a_0_0.vlm"},
	{0xE1CC53F40B0045EBULL, "vxlt_undercityserialkiller_1d2_0_0.vlm"},
	{0xE1CC5A0AA7A1A107ULL, "gen00fl_isabela6075269.cl"},
	{0xE1CCB2C06A70679BULL, "vxht_chantry_35z_0_0.vlm"},
	{0xE1CD1470C83E6997ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.anb"},
	{0xE1CD3920927789DAULL, "hlac_dr_slavers_3216_3y.dds"},
	{0xE1CD3F8965B174E0ULL, "vxht_0_hightown_bxs_0_0.vlm"},
	{0xE1CDB930BB7231ECULL, "734093217.ogg"},
	{0xE1CDB950E36290F9ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_slidel.gad"},
	{0xE1CDD90C73D15D7BULL, "hlac_w_encounter2_camp_393215_0x.dds"},
	{0xE1CDDADA174DB175ULL, "follower_banter6164850.cl"},
	{0xE1CDF8BACC9CD083ULL, "vxca_caverns1_icl_0_2.vlm"},
	{0xE1CE08B02414F018ULL, "vxca_cavern1_haunted_imr_0_0.vlm"},
	{0xE1CE0AB3E89C282BULL, "ts_playvfx.ncs"},
	{0xE1CE10F7C096519CULL, "6142746_m.fxe"},
	{0xE1CE122120FAE224ULL, "16638204.ogg"},
	{0xE1CE6843A6800E57ULL, "437163283.ogg"},
	{0xE1CE7D13962F1949ULL, "vxlt_0_lowtown_docks_n_cq6_0_0.vlm"},
	{0xE1CE846F8D293CA5ULL, "vxlt_backalley_g_1ku_0_0.vlm"},
	{0xE1CE9C52A3C492BCULL, "vxlt_ruinedhouse_n_6z_0_0.vlm"},
	{0xE1CF27F1F24F0204ULL, "dre000_hireling16038639.cl"},
	{0xE1CFD104D768182AULL, "457563774.ogg"},
	{0xE1D02D9AD3B684A5ULL, "vxca_caverns1_e41_0_0.vlm"},
	{0xE1D03DA175ECC463ULL, "llac_lt_hangedman_-1_20.dds"},
	{0xE1D05DB101F9114EULL, "vxlt_0_lowtown_docks_a3_n_d33_0_0.vlm"},
	{0xE1D06571EB3BA5CEULL, "241563993.ogg"},
	{0xE1D0B190ED4FC5EDULL, "191581262.ogg"},
	{0xE1D10D1E7010E77DULL, "llac_ca_cavern1_haunted_24427_0.dds"},
	{0xE1D10FE4FFCB0ED4ULL, "dae000_mother6108966.cl"},
	{0xE1D14298B41C6944ULL, "mag311_emile_de_launcet6062688.cl"},
	{0xE1D17B29676F707CULL, "hlac_w_woundedcoast_0_2y.dds"},
	{0xE1D1F9D2C097BAE0ULL, "art\\characters\\creatures\\ragedemon\\animation\\p_wkf.evt"},
	{0xE1D1FE638AA36FFDULL, "6036308_f.fxe"},
	{0xE1D236CDC861ACF6ULL, "llac_ht_mansion2_818_8.dds"},
	{0xE1D28E724159BF2EULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_bdy_ggg_0d.dds"},
	{0xE1D2FAA8F4A52685ULL, "plt_cod_itm_vigilance.plo"},
	{0xE1D2FD3CF7F703D0ULL, "w_sundermount_5_0.msh"},
	{0xE1D3088C4491B0DFULL, "gen_im_cft_reg_d1.uti"},
	{0xE1D30B3CF7F71BEAULL, "w_sundermount_5_0.mmh"},
	{0xE1D30F3CF7F72221ULL, "w_sundermount_5_0.mao"},
	{0xE1D33DF79997E293ULL, "vxht_mansion2b_2l8_0_0.vlm"},
	{0xE1D3870B8CFA8C12ULL, "324139071.ogg"},
	{0xE1D3B81A9CEE501FULL, "vxw_woundedcoast_u1_0_0.vlm"},
	{0xE1D3CA50E3671947ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_slidel.ani"},
	{0xE1D3CDDB39B918E6ULL, "hlac_ht_0_hightown_131071_0x.dds"},
	{0xE1D437BC590275ACULL, "vxca_cavern1_haunted_fn8_0_0.vlm"},
	{0xE1D48E39BA10DED0ULL, "6061346_m.fxe"},
	{0xE1D49F4D90F23386ULL, "6130095_m.fxe"},
	{0xE1D503711EEE576DULL, "fen110_fenris6100261.cl"},
	{0xE1D55393336F48AEULL, "vxlt_gallowstemplar_56g_0_0.vlm"},
	{0xE1D60CA687C4BF5AULL, "vxht_mansion2_1hb_0_0.vlm"},
	{0xE1D712FFE2B84730ULL, "667217672.ogg"},
	{0xE1D786B8AEFFD590ULL, "art\\characters\\creatures\\spider\\animation\\tn90l.ani"},
	{0xE1D78A0C37178340ULL, "zz_qun_debug6028272.cl"},
	{0xE1D7A919350EDF21ULL, "6070248_m.fxe"},
	{0xE1D867187C867509ULL, "76972415.ogg"},
	{0xE1D8B77763B2B457ULL, "vxlt_gallowscourtyard_n_3vy_0_0.vlm"},
	{0xE1D8D5E60A685173ULL, "6217456_m.fxe"},
	{0xE1D8DA0BDC133C19ULL, "vxlt_hangedman_n_22i_0_0.vlm"},
	{0xE1D8FD6AFCA08DD4ULL, "vxht_mansion2b_1gy_0_0.vlm"},
	{0xE1D9007B9D121BE9ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.dura.ani"},
	{0xE1D90D0962735512ULL, "vxht_mansion2_2ns_0_0.vlm"},
	{0xE1D94D49551AE0E7ULL, "245615080.ogg"},
	{0xE1D9815F57536098ULL, "hf_gen_circle_mage.mor"},
	{0xE1D9A0C92C8B6D62ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy_crnf.gad"},
	{0xE1D9BDF2791C27F3ULL, "vxlt_0_lowtown_a2_gbo_0_0.vlm"},
	{0xE1DA1E8A060312ECULL, "vxht_keep_n_57z_0_0.vlm"},
	{0xE1DB8B90BE6A5A96ULL, "vxdr_primevaltomb_34t_0_1.vlm"},
	{0xE1DC14766E167E8AULL, "6121173_f.fxe"},
	{0xE1DC4CE52F665CDEULL, "lgt260_eau_tmp_ending6171806.cl"},
	{0xE1DC6487DC511063ULL, "vxlt_0_lowtown_a3_n_c_hl6_0_0.vlm"},
	{0xE1DC7A8D160B0898ULL, "vxca_cavern1_haunted_ijs_0_0.vlm"},
	{0xE1DC9BBABE5A2B07ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_ctn180r.gad"},
	{0xE1DCCC062313AAF9ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobb.mmh"},
	{0xE1DCCE38AE4DB1EFULL, "138192982.ogg"},
	{0xE1DCD6062313BBE7ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobb.msh"},
	{0xE1DCF64C7F1F5525ULL, "vxlt_backalley_g_1pt_0_0.vlm"},
	{0xE1DD0505059D1FCFULL, "hlac_ht_mansion2b_818_32z.dds"},
	{0xE1DD4BA8764E860DULL, "891773141.ogg"},
	{0xE1DE0BC82921A074ULL, "587040378.ogg"},
	{0xE1DE944263512AE0ULL, "dae200cr_viscount6087242.cl"},
	{0xE1DEDD4C0E180EDBULL, "prop1_s1y.psh.fxo"},
	{0xE1DF953F073004D1ULL, "463206071.ogg"},
	{0xE1DFABCB686A3D79ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\heavydamage.ani"},
	{0xE1DFC3C92C901459ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy_crnf.evt"},
	{0xE1E01B7B9D18686BULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.dura.gad"},
	{0xE1E0384E26B93EB2ULL, "gen00fl_anders6111355.cl"},
	{0xE1E0D2C7CEC1033FULL, "vxlt_0_lowtown_a2_hdj_0_0.vlm"},
	{0xE1E0FE5B8C3E4EC6ULL, "933591871.ogg"},
	{0xE1E10250AE68DCF9ULL, "w_encounter2_camp_13_0.phy"},
	{0xE1E146E96A7F9236ULL, "mag360_alain.cnv"},
	{0xE1E14A1BF292034AULL, "319232732.ogg"},
	{0xE1E195B4DAA9F1F7ULL, "vxca_caverns1_mines_k2p_0_2.vlm"},
	{0xE1E1E12DB84008C3ULL, "445700739.ogg"},
	{0xE1E214915F360FBCULL, "357226694.ogg"},
	{0xE1E229F78EE4689FULL, "vxw_woundedcoast_r0_0_0.vlm"},
	{0xE1E24FDF1CE69CA9ULL, "fxe_lava_explo.mao"},
	{0xE1E270C4818A114AULL, "845551716.ogg"},
	{0xE1E273C75CFDABA6ULL, "vxw_bonepit_1xa_0_0.vlm"},
	{0xE1E2A10591B755FDULL, "vxlt_0_lowtown_n_e1m_0_0.vlm"},
	{0xE1E2C3CF46D74379ULL, "1019697354.ogg"},
	{0xE1E31BE06D020673ULL, "464542747.ogg"},
	{0xE1E3397BDDF2AD96ULL, "zz_and_debug6176460.cl"},
	{0xE1E34C0E32F303A3ULL, "hlac_lt_0_lowtown_docks_-65540_1y.dds"},
	{0xE1E3F61F1233E165ULL, "vxht_mansion2_1uz_0_0.vlm"},
	{0xE1E43C19E8431D55ULL, "zz_isa_debug6133176.cl"},
	{0xE1E4426BDD70ED0CULL, "hlac_lt_backalley_g_65537_6x.dds"},
	{0xE1E448E96A81CB16ULL, "mag360_alain.bnk"},
	{0xE1E4579ECBB86B1CULL, "character_character_p171.psh.fxo.dx11"},
	{0xE1E47A7AA251E8A5ULL, "6127117_m.fxe"},
	{0xE1E4B5D068F11CFEULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv1.evt"},
	{0xE1E4BB4CEA1CA420ULL, "476085992.ogg"},
	{0xE1E4C8C743BBC590ULL, "hm_lgt_den_rev2.mor"},
	{0xE1E4D5EF6BCE43FDULL, "art\\vfx\\materials\\chunky\\vfx_c_crater_half.mao"},
	{0xE1E4ED9C079F6F4BULL, "mag360_thrask6043914.cl"},
	{0xE1E4FB11546B708DULL, "716131992.ogg"},
	{0xE1E5255EF347B534ULL, "one300_zevran6075770.cl"},
	{0xE1E534FE376942F9ULL, "mag170_hugh6206270.cl"},
	{0xE1E5B97D12299AB8ULL, "one100dg_vanard6173859.cl"},
	{0xE1E5BD704A4288A8ULL, "ht_0_hightown_a2_4_0.rml"},
	{0xE1E643CF76CE9A92ULL, "hlac_ca_cavern1_haunted_24421_5z.dds"},
	{0xE1E686CB68701CBBULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\heavydamage.gad"},
	{0xE1E6E37AD9E47DC4ULL, "mag110_samson6018594.cl"},
	{0xE1E7187B9D1E81D0ULL, "art\\vfx\\combat\\vfx_glb_acid_proj_p.dura.evt"},
	{0xE1E730C7B6258DB5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int2_r.dds"},
	{0xE1E76791949DD282ULL, "6162342_m.fxe"},
	{0xE1E7AAB857BE9B6DULL, "vxht_brothel_n_34c_0_0.vlm"},
	{0xE1E802F2332B5E4FULL, "242027480.ogg"},
	{0xE1E80B419A7514ABULL, "vxlt_gallowstemplar_377_0_0.vlm"},
	{0xE1E80B50AE6F0A15ULL, "w_encounter2_camp_13_0.rml"},
	{0xE1E858E4CF3CA012ULL, "mrl000_marethari6117693.cl"},
	{0xE1E896ACC38D2596ULL, "hlac_ht_mansion2b_818_19x.dds"},
	{0xE1E8B426822F3185ULL, "6121570_f.fxe"},
	{0xE1E8BCD46C4C1B86ULL, "dae000_maren6148326.cl"},
	{0xE1E8F36A9C20E80BULL, "vxlt_0_lowtown_n_eod_0_1.vlm"},
	{0xE1E8F7E4E8665413ULL, "vxht_bartrand_n_69c_0_0.vlm"},
	{0xE1E911E9CDB2F2F5ULL, "6172502_m.fxe"},
	{0xE1E95DDDFF049A8BULL, "ht_0_hightown_a2_21_0.anb"},
	{0xE1E997944B53A946ULL, "vxht_0_hightown_n_a7x_0_0.vlm"},
	{0xE1EA99A6E313EE82ULL, "llac_lt_warehouse_n_0_4.dds"},
	{0xE1EAA40CA9B2A1E8ULL, "ave321tr_varric_check.ncs"},
	{0xE1EAE8C3972B7AF6ULL, "6087393_m.fxe"},
	{0xE1EAEC6A08012118ULL, "vxca_cavern1_haunted_igp_0_0.vlm"},
	{0xE1EB2C93FB1C9951ULL, "character_rendertodepth_p5.vsh.fxo.dx11.assembly"},
	{0xE1EB70D468ACDA28ULL, "505923408.ogg"},
	{0xE1EBCCD068F76229ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv1.gad"},
	{0xE1EC36343157B89FULL, "947399954.ogg"},
	{0xE1EC38E9281CACEDULL, "hlac_lt_0_lowtown_a2_n_65535_0y.dds"},
	{0xE1EC3AD230046705ULL, "gen00fl_bethany6098051.cl"},
	{0xE1ED4654A1C066A2ULL, "6152363_m.fxe"},
	{0xE1ED70CFB295504EULL, "302801527.ogg"},
	{0xE1ED7DBCAA3C43F4ULL, "850172001.ogg"},
	{0xE1ED83D0FB8E2F41ULL, "358572326.ogg"},
	{0xE1ED89850488ECECULL, "1036655293.ogg"},
	{0xE1EDA3CB68766CA0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\heavydamage.evt"},
	{0xE1EDCBC92C9C5F48ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy_crnf.ani"},
	{0xE1EDD7C5889AC533ULL, "724077439.ogg"},
	{0xE1EDE4007C153592ULL, "606498410.ogg"},
	{0xE1EEAFEB6512E83DULL, "647505875.ogg"},
	{0xE1EEBC889E5A3982ULL, "vxca_caverns1_mines_iid_0_0.vlm"},
	{0xE1EF01ACA4F89268ULL, "vxht_mansion2b_1w5_0_0.vlm"},
	{0xE1EF1A57C988CB17ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic.mmh"},
	{0xE1EF1E57C988D224ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic.mao"},
	{0xE1EF3057C988F089ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ff_basic.msh"},
	{0xE1EF36E968B77D1CULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_c.impa.ani"},
	{0xE1EF727715A106CDULL, "909910212.ogg"},
	{0xE1EFB2652C758AD4ULL, "gen00fl_dog6114922.cl"},
	{0xE1EFB33FBF955461ULL, "zz_isa_debug6133231.cl"},
	{0xE1EFC3D859AAAED2ULL, "aip_harvester_base.gda"},
	{0xE1EFE36307B93843ULL, "6036936_m.fxe"},
	{0xE1EFED5B65274502ULL, "6129718_m.fxe"},
	{0xE1F00FF4B616C190ULL, "6051327_m.fxe"},
	{0xE1F043F3CA50C495ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_4y.dds"},
	{0xE1F0624E4522B35CULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.dura.ani"},
	{0xE1F0AC3F91C07BCBULL, "6172782_m.fxe"},
	{0xE1F11CF635317F83ULL, "126743316.ogg"},
	{0xE1F126DEA4134A27ULL, "6030686_m.fxe"},
	{0xE1F188BABE6C674DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_ctn180r.ani"},
	{0xE1F18CC94A437A65ULL, "679183812.ogg"},
	{0xE1F1C03BDFB88C74ULL, "6160786_m.fxe"},
	{0xE1F267FC0429F9E5ULL, "vxht_mansion2_1xy_0_0.vlm"},
	{0xE1F2757333429CC0ULL, "vxht_0_hightown_a3_c0g_0_0.vlm"},
	{0xE1F29DD068FD3177ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\attackv1.ani"},
	{0xE1F2A1B418E78529ULL, "6144341_m.fxe"},
	{0xE1F2B936F5013D51ULL, "ser191_emeric6107121.cl"},
	{0xE1F31E8F98E03447ULL, "gen00fl_fenris6103322.cl"},
	{0xE1F33C626CD32954ULL, "176805283.ogg"},
	{0xE1F3807C7B7D4050ULL, "qun110_petrice6018509.cl"},
	{0xE1F3810F0D88ECD2ULL, "6196859_m.fxe"},
	{0xE1F3F4FD5E3C14F7ULL, "537861001.ogg"},
	{0xE1F46A1270340442ULL, "mag210_arianni6024984.cl"},
	{0xE1F528A8AFB9E606ULL, "6153193_m.fxe"},
	{0xE1F52D062172ED44ULL, "art\\characters\\playercharacter\\humanmale\\glv_medp.msh"},
	{0xE1F53F0621730BADULL, "art\\characters\\playercharacter\\humanmale\\glv_medp.mao"},
	{0xE1F54306217312E6ULL, "art\\characters\\playercharacter\\humanmale\\glv_medp.mmh"},
	{0xE1F61170C8616998ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.mmh"},
	{0xE1F61C9549B4B3EDULL, "vxht_brothel_n_33b_0_0.vlm"},
	{0xE1F643E968BDB25EULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_c.impa.gad"},
	{0xE1F66CF30EA671E3ULL, "lt_0_lowtown_docks_a3_n_245_0.msh"},
	{0xE1F66F4E4527349EULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.dura.gad"},
	{0xE1F67AF30EA6893DULL, "lt_0_lowtown_docks_a3_n_245_0.mmh"},
	{0xE1F67EF30EA6904EULL, "lt_0_lowtown_docks_a3_n_245_0.mao"},
	{0xE1F67FC770A0D483ULL, "llac_lt_0_lowtown_a3_n_c_-262144_0.dds"},
	{0xE1F6F1D6B5D422B1ULL, "hlac_w_bonepitdestroyed_a3_327680_0y.dds"},
	{0xE1F7A98AE0E28F6BULL, "vxlt_undercityserialkiller_1s7_0_0.vlm"},
	{0xE1F7A9D64DBFC204ULL, "zz_fen_debug6215239.cl"},
	{0xE1F7AD1F8F0589D6ULL, "hlac_w_encounter2_camp_65540_0z.dds"},
	{0xE1F7D9FE0D7797C0ULL, "vxlt_undercityencounter_140_0_1.vlm"},
	{0xE1F7F157911B41E6ULL, "369410124.ogg"},
	{0xE1F815EFF89FE618ULL, "art\\characters\\weapons\\quivers\\textures\\w_qui_dq101a_0t.dds"},
	{0xE1F8BD0290009FF8ULL, "67175844.ogg"},
	{0xE1F8BE9733FF9BE3ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.dura.evt"},
	{0xE1F8BFA2EF373F21ULL, "6169018_m.fxe"},
	{0xE1F8E3E2A13B51E2ULL, "blurring_alpha.mao"},
	{0xE1F9A56E563CB271ULL, "lgt210cr_eau_keran_father.utc"},
	{0xE1F9DA06630B8625ULL, "vxlt_backalley_g_1zr_0_0.vlm"},
	{0xE1FA42FF60F56FC1ULL, "vxht_0_hightown_n_apj_0_0.vlm"},
	{0xE1FA5EDF1CFB4857ULL, "fxe_lava_explo.dds"},
	{0xE1FA7CA5F1A8F336ULL, "vxlt_darktown_580_0_1.vlm"},
	{0xE1FB2375E22BCF1FULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_f_c.impa.gad"},
	{0xE1FB38958F5DB53BULL, "dae100_alistair_drunk6135095.cl"},
	{0xE1FB43D5C1D10F61ULL, "493999765.ogg"},
	{0xE1FB7A24473DFD4FULL, "gen00fl_anders6119499.cl"},
	{0xE1FBCE8B2BBF3149ULL, "one110_gustav6089827.cl"},
	{0xE1FBF62F87DC2133ULL, "522222026.ogg"},
	{0xE1FC8EBD51346DC0ULL, "plt_dae00pt_main_plot.nss"},
	{0xE1FD1EE968C3927DULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_c.impa.evt"},
	{0xE1FD4A4E452D14BDULL, "art\\vfx\\combat\\vfx_v_fir_imm_lp_c.dura.evt"},
	{0xE1FDA668F2C2B64DULL, "357070597.ogg"},
	{0xE1FE890B646BDE93ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\inactiveready.ani"},
	{0xE1FEE9473DE08C02ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb106a_0d.dds"},
	{0xE1FF16839E84274CULL, "vxht_bartrand_n_5r6_0_0.vlm"},
	{0xE1FF3A06217B57B1ULL, "art\\characters\\playercharacter\\humanmale\\glv_medp.phy"},
	{0xE1FF85BC0D8CB4F0ULL, "vxht_0_hightown_a3_c18_0_0.vlm"},
	{0xE1FF979734057864ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.dura.gad"},
	{0xE1FFFEEAA744EBE5ULL, "6021120_f.fxe"},
	{0xE200A2D96718E398ULL, "vxca_caverns2prc_9xf_0_0.vlm"},
	{0xE200A92A0951922EULL, "vxlt_gallowstemplar_59z_0_0.vlm"},
	{0xE200E7502538B540ULL, "vxht_0_hightown_a3_c5d_0_0.vlm"},
	{0xE20121F76CF87AF7ULL, "weapon_weapon_p0.vsh.fxo.dx11.assembly"},
	{0xE20148AAD3EA7DA3ULL, "122032667.ogg"},
	{0xE20214849767802CULL, "and100_anders6095958.cl"},
	{0xE2021875E231DB0CULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_f_c.impa.evt"},
	{0xE202310CA3613877ULL, "hlac_lt_0_lowtown_a2_65537_3y.dds"},
	{0xE20251DD44E4AEF0ULL, "421855063.ogg"},
	{0xE20285C26B1D718AULL, "6033400_m.fxe"},
	{0xE202A81318C2F873ULL, "6099030_m.fxe"},
	{0xE202D5BB6B30E539ULL, "ser221_gascard6173494.cl"},
	{0xE202E14C04273CBFULL, "843023111.ogg"},
	{0xE202E86DB3B2E38FULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.mmh"},
	{0xE202E87FABFA092BULL, "llac_lt_gallowscourtyard_n_-1_4.dds"},
	{0xE20333996C58AD48ULL, "ave310_donnic6134772.cl"},
	{0xE20359E0240C13FBULL, "6135967_m.fxe"},
	{0xE2035C64EB97790AULL, "lt_0_lowtown_docks_160_0.mao"},
	{0xE2036064EB978059ULL, "lt_0_lowtown_docks_160_0.mmh"},
	{0xE2036A64EB979147ULL, "lt_0_lowtown_docks_160_0.msh"},
	{0xE2037CF7DEAA62A5ULL, "hlac_ca_caverns1_mines_24426_14z.dds"},
	{0xE203DC45B3BB77A3ULL, "vxlt_0_lowtown_a3_n_c_cmv_0_0.vlm"},
	{0xE20408EEA9C287CCULL, "lgt260_way_blood_mage6148390.cl"},
	{0xE20452819E1A3083ULL, "6078332_m.fxe"},
	{0xE204D8E1B526F336ULL, "follower_banter6141887.cl"},
	{0xE204F79FFEB77D06ULL, "108528406.ogg"},
	{0xE20565347D8B7347ULL, "124350556.ogg"},
	{0xE2062764DDE48E6BULL, "6188637_m.fxe"},
	{0xE206A697340BB042ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_loop_c.dura.ani"},
	{0xE206B967A861AECAULL, "gen_im_arm_cht_hvy_act2_01.uti"},
	{0xE206FDFDD28DABF8ULL, "vxht_0_hightown_bp3_0_0.vlm"},
	{0xE2071893DAA21329ULL, "llac_lt_0_lowtown_a2_n_c_1_4.dds"},
	{0xE20719C6DB4F8254ULL, "hlac_w_woundedcoast_7_0z.dds"},
	{0xE207473E9213D754ULL, "hlac_w_sundermount_65539_0y.dds"},
	{0xE20771401D3B53ACULL, "pet_body.mao"},
	{0xE207D3F40F136136ULL, "vxht_mansion2_1vp_0_0.vlm"},
	{0xE207FABC2E79EE43ULL, "gen00fl_sebastian6179993.cl"},
	{0xE208B4DC52EB8841ULL, "vxht_0_hightown_n_aui_0_0.vlm"},
	{0xE209C95B156A82FCULL, "gen00fl_aveline6139183.cl"},
	{0xE209ED9BC96AFBAEULL, "6149381_m.fxe"},
	{0xE20A277C5F0A1467ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medd_0d.dds"},
	{0xE20A42F6DFA94537ULL, "vxlt_0_lowtown_docks_a3_n_cw1_0_0.vlm"},
	{0xE20A7D1E1E3E1FAAULL, "6149226_m.fxe"},
	{0xE20A8E4DD44FCE3AULL, "pro000_varric_cass6208370.cl"},
	{0xE20B21BAC50EBBD9ULL, "hlac_lt_0_lowtown_a2_n_65536_8x.dds"},
	{0xE20B9AEC0DEE0028ULL, "vxlt_blackemporium_87_0_0.vlm"},
	{0xE20BE81380F9CDD4ULL, "vxlt_0_lowtown_a2_n_gnk_0_0.vlm"},
	{0xE20C24458A9D1F60ULL, "zz_car_debug6122473.cl"},
	{0xE20C810B64780ECAULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\inactiveready.evt"},
	{0xE20C8A73DC0E08DEULL, "vxht_brothel_n_5hf_0_0.vlm"},
	{0xE20CB25143E4B30EULL, "lowtown_fem_prostitute_sta6152363.cl"},
	{0xE20CCC77BB3C2D07ULL, "6096821_f.fxe"},
	{0xE20D8860907A3FCCULL, "vxht_bartrand_n_5u7_0_0.vlm"},
	{0xE20D890871A596CDULL, "hla_ht_0_hightown_-131073_0y.dds"},
	{0xE20DC95E32DFF526ULL, "vxw_bonepit_1kd_0_0.vlm"},
	{0xE20E495EF23DA211ULL, "vfx_vfx_p0.vsh.fxo.dx11.assembly"},
	{0xE20EA4A3C643075FULL, "dae000ip_crowd_ff1_talk18.utp"},
	{0xE20F2FA11FBCCCCCULL, "gen00fl_sebastian6174869.cl"},
	{0xE20F4BB5E8162AE5ULL, "vxht_mansion2_1fg_0_0.vlm"},
	{0xE20F8CCE3DDE86C2ULL, "6044235_m.fxe"},
	{0xE20F93B785A905DCULL, "vxlt_0_lowtown_a2_n_fek_0_0.vlm"},
	{0xE2100075E23DF065ULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_f_c.impa.ani"},
	{0xE21040FCB4E231A2ULL, "prop1_nolit.psh.fxo"},
	{0xE2104FB28BD2F22AULL, "537311565.ogg"},
	{0xE210ECB377767804ULL, "sst_ss_danarius.gda"},
	{0xE2112F6D1B2155DBULL, "vxht_mansion_n_6o2_0_0.vlm"},
	{0xE211340622F0E599ULL, "vxht_bartrand_n_4se_0_0.vlm"},
	{0xE211AC8D8E0E58D7ULL, "6098935_f.fxe"},
	{0xE211D57754D1BFD8ULL, "vxlt_0_lowtown_a2_n_c_hse_0_0.vlm"},
	{0xE211DDD156E17BF0ULL, "gen00fl_fenris6116545.cl"},
	{0xE2126D856C3D66CDULL, "6173035_m.fxe"},
	{0xE213600B647DF53DULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\inactiveready.gad"},
	{0xE2144C4C534FA04BULL, "gen00fl_carver6104602.cl"},
	{0xE21456E4035A8ABDULL, "859348060.ogg"},
	{0xE2145F4A5FCAB8FBULL, "hlac_lt_0_lowtown_a2_-131071_2z.dds"},
	{0xE214D1F46B470501ULL, "vxw_bonepitmining_a2_21x_0_0.vlm"},
	{0xE2155ADDFF29D0D7ULL, "ht_0_hightown_a2_21_0.rml"},
	{0xE215C3324E462BF4ULL, "6081663_m.fxe"},
	{0xE215E29F07F27AE2ULL, "zz_followers_banter6158490.cl"},
	{0xE21681A9F5D63FCDULL, "243254201.ogg"},
	{0xE2172EB3D12EE119ULL, "gen00fl_merrill6129856.cl"},
	{0xE21781006E59F89AULL, "vxlt_hangedman_n_2vz_0_0.vlm"},
	{0xE21783F577B8EA62ULL, "6151083_m.fxe"},
	{0xE217F0C2B51BE928ULL, "6126581_f.fxe"},
	{0xE2185F78BB61142EULL, "6142281_m.fxe"},
	{0xE2189ABA7FDA9FA7ULL, "54829717.ogg"},
	{0xE2189F6ABF3170AEULL, "mcr310_argument6075790.cl"},
	{0xE218B4D3D19F5DB7ULL, "vxlt_0_lowtown_docks_a3_n_cp0_0_0.vlm"},
	{0xE218C63C34C22C51ULL, "591393519.ogg"},
	{0xE218EF9A418F8C19ULL, "6171459_m.fxe"},
	{0xE218F8DCA0F5B66EULL, "art\\vfx\\combat\\vfx_a_stealth_att_1_c.impa.evt"},
	{0xE2190C1D04D61F32ULL, "vxlt_0_lowtown_a2_n_c_e10_0_0.vlm"},
	{0xE219769BBCC4BCF4ULL, "571384189.ogg"},
	{0xE21AC2FBAA7982D1ULL, "vxlt_gallowstemplar_f_38s_0_0.vlm"},
	{0xE21AED332A4662E2ULL, "hlac_lt_0_lowtown_n_1_5x.dds"},
	{0xE21B201C2020AAA9ULL, "fen110_fenris6096409.cl"},
	{0xE21B308B70E3CC57ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06a.phy"},
	{0xE21BAA59150D835FULL, "vxca_cave_2k1_0_0.vlm"},
	{0xE21BC4C53379A347ULL, "gen00fl_varric6076013.cl"},
	{0xE21BD627D158E03CULL, "6052771_m.fxe"},
	{0xE21BEF453876FCF1ULL, "lt_0_lowtown_docks_a3_n_29_0_lc.mmh"},
	{0xE21BFA3D8270584CULL, "vxht_bartrand_n_5x4_0_0.vlm"},
	{0xE21C8640765A1053ULL, "100254452.ogg"},
	{0xE21D5560735B7470ULL, "vxlt_hangedman_n_4z_0_0.vlm"},
	{0xE21D631CC37EC8C2ULL, "vxlt_gallowscourtyard_n_5a3_0_0.vlm"},
	{0xE21DEB514DAAE0C6ULL, "6050380_m.fxe"},
	{0xE21E0AC281C10920ULL, "hlac_ht_keep_3739_0y.dds"},
	{0xE21E253A052A36F4ULL, "35100683.ogg"},
	{0xE21EEFDCA0FA1219ULL, "art\\vfx\\combat\\vfx_a_stealth_att_1_c.impa.gad"},
	{0xE21F222B9C25CA6BULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowa_sa.dds"},
	{0xE21F4941E78751EAULL, "hlac_lt_0_lowtown_docks_n_262143_0y.dds"},
	{0xE21F7FF49DCD45CAULL, "gen00fl_dog6107344.cl"},
	{0xE2204EA5E22D6B47ULL, "llac_ht_keep_n_4519_15.dds"},
	{0xE2207B69F431B2A3ULL, "vxdr_primevaltomb_39g_0_0.vlm"},
	{0xE220FC5B2457BCE3ULL, "758347054.ogg"},
	{0xE2216B224935239DULL, "hlac_ca_cavern1_haunted_24425_12y.dds"},
	{0xE221AA4A86488083ULL, "art\\vfx\\textures\\vfx_white.dds"},
	{0xE222930F2BB3D77EULL, "6128400_f.fxe"},
	{0xE222BAEBD3B44541ULL, "vxdr_primevalentrance_2ni_0_0.vlm"},
	{0xE22303D14AB478FAULL, "ht_0_hightown_a3_51_0_lc.mmh"},
	{0xE223074E57DC5923ULL, "gen00fl_fenris6126572.cl"},
	{0xE2232648A250B194ULL, "vxw_encounter3_2y_0_0.vlm"},
	{0xE2233CDF78400029ULL, "vxlt_0_lowtown_docks_n_b4e_0_1.vlm"},
	{0xE2239F788452E1A5ULL, "vxlt_undercityserialkiller_25t_0_0.vlm"},
	{0xE223D05B5BCB7FD3ULL, "6104949_m.fxe"},
	{0xE2240F165FE1B7FFULL, "lgt360cr_mag_bandit_leader.utc"},
	{0xE224431CE5D62A11ULL, "253191486.ogg"},
	{0xE22600DCA1004E67ULL, "art\\vfx\\combat\\vfx_a_stealth_att_1_c.impa.ani"},
	{0xE226289D45C14732ULL, "vxlt_0_lowtown_h8k_0_0.vlm"},
	{0xE2269BF67F0FD7B7ULL, "mrl210_merrill6166776.cl"},
	{0xE226A7F5E98A79D8ULL, "tce_wallsm_0.phy"},
	{0xE226E17EF52527DEULL, "293936583.ogg"},
	{0xE227276057E59CDBULL, "lgt110_gxa_vgk_silver_order6116986.cl"},
	{0xE2276B791A31EA6AULL, "75228134.ogg"},
	{0xE2283EC718F3A916ULL, "481894770.ogg"},
	{0xE22869843A1F1979ULL, "6116158_f.fxe"},
	{0xE228E34AAF58F842ULL, "hlac_ht_0_hightown_a2_-2_0x.dds"},
	{0xE22913EE3FE07CCCULL, "6106648_f.fxe"},
	{0xE2296357CCCC5FCEULL, "gen00fl_dog6165762.cl"},
	{0xE2298B171BF33F24ULL, "vxht_0_hightown_a2_c8j_0_0.vlm"},
	{0xE229B6BFEA70C0D8ULL, "165477156.ogg"},
	{0xE229D615F197003FULL, "gen00fl_isabela6096792.cl"},
	{0xE22AD16E661B06BCULL, "356002400.ogg"},
	{0xE22AE71444E6FFF1ULL, "w_bonepitdestroyed_a3.arl"},
	{0xE22AE96DB3D437CCULL, "art\\vfx\\environment\\vfx_blt_fire_wall_p.anb"},
	{0xE22B29CAB970943EULL, "vxw_bonepit_xz_0_0.vlm"},
	{0xE22B770127EA7374ULL, "cineanim_global_events.wbd"},
	{0xE22D09B005F7D437ULL, "mer240_jansen6081433.cl"},
	{0xE22D25AA97C99193ULL, "hlac_lt_0_lowtown_n_-131071_2y.dds"},
	{0xE22E47EF03802FA2ULL, "vxw_encounter2_cw_0_0.vlm"},
	{0xE22E7B33287D087AULL, "da2terrain_terrain_tools_p2.psh.fxo.dx11.assembly"},
	{0xE22EEB713AE28A26ULL, "hlac_w_bonepitdestroyed_a3_65540_0z.dds"},
	{0xE22F98EA9B51FA36ULL, "zz_combat_demo.are"},
	{0xE22FBABAB2BA1D39ULL, "vxw_bonepitdestroyed_a3_kx_0_0.vlm"},
	{0xE22FD4DA2DFCD56CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_adjust.ani"},
	{0xE230424889CDBDD2ULL, "fen110_fenris6118027.cl"},
	{0xE230A0D3A0379E85ULL, "dae300_alistair_king6133568.cl"},
	{0xE230B223B168E255ULL, "vxht_0_hightown_cb9_0_0.vlm"},
	{0xE23123AC1931AD83ULL, "6153761_m.fxe"},
	{0xE2316F2088D7BDC4ULL, "hlac_lt_0_lowtown_docks_a3_-65537_1y.dds"},
	{0xE231E00F7DEAAA05ULL, "699535129.ogg"},
	{0xE2321EDB7B310BACULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation1.ani"},
	{0xE23237621624B175ULL, "dae000ip_crowd_mf7_merchant.utp"},
	{0xE232878B6DF8B008ULL, "hlac_lt_0_lowtown_docks_a3_-65538_7z.dds"},
	{0xE232B8613EE503BBULL, "vxca_cavern1_haunted_jde_0_0.vlm"},
	{0xE232D9845E5EC071ULL, "one100dg_kelder6173956.cl"},
	{0xE233139E07E52DF1ULL, "art\\characters\\creatures\\mubari\\animation\\rundescend.ani"},
	{0xE23317D90AED329FULL, "vxca_cavern1_haunted_gkd_0_1.vlm"},
	{0xE233298AE4F5AAB6ULL, "vxht_mansion2_1gs_0_0.vlm"},
	{0xE2334BEB97B14C71ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\gibexplodeendpose.ani"},
	{0xE233B5ADC9443513ULL, "vrc221_varric6065961.cl"},
	{0xE2344E0D56A8F247ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfl.gad"},
	{0xE234A2BE31EBAC4AULL, "sho_rock.mao"},
	{0xE234C41596456AE9ULL, "hlac_ht_keep_n_3732_19x.dds"},
	{0xE234D3BAF7A2AF1EULL, "llac_ht_brothel_n_3370_0.dds"},
	{0xE234F36DB97EAFF6ULL, "gen00fl_dog6114919.cl"},
	{0xE2353073668CC8A2ULL, "vxlt_undercitydungeon_1g7_0_0.vlm"},
	{0xE235508351CB19ADULL, "70144052.ogg"},
	{0xE235AB0736007209ULL, "vxw_sundermount_2jd_0_0.vlm"},
	{0xE235E1DA2E0156AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_adjust.gad"},
	{0xE23630FC6193CB1DULL, "follower_banter6163169.cl"},
	{0xE2363FF601BC267DULL, "hlac_ht_0_hightown_a3_131071_0x.dds"},
	{0xE2373A21CF37686EULL, "mcr371_angry_mage_m6087733.cl"},
	{0xE237CCD5D73CC44AULL, "102548763.ogg"},
	{0xE2380C4E54F06BD9ULL, "vxdr_0_deeproads_cfs_0_1.vlm"},
	{0xE238100D35742606ULL, "vxht_mansion_n_6f0_0_0.vlm"},
	{0xE2381994A40F34DCULL, "plt_one100pt_herb_gland.nss"},
	{0xE238540A8F2D32FAULL, "6140650_m.fxe"},
	{0xE2386CBDB1058236ULL, "vxlt_0_lowtown_docks_a3_n_bv2_0_0.vlm"},
	{0xE2392BDB7B3740EEULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation1.gad"},
	{0xE2393B67B0A494CEULL, "1025429526.ogg"},
	{0xE23A1E9E07EB5E93ULL, "art\\characters\\creatures\\mubari\\animation\\rundescend.gad"},
	{0xE23A41DAF69B454FULL, "dae021_sister_lorena_sta6150068.cl"},
	{0xE23A56D17EC94253ULL, "564123093.ogg"},
	{0xE23A56DBDFA440A1ULL, "875407916.ogg"},
	{0xE23A56EB97B77D13ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\gibexplodeendpose.gad"},
	{0xE23A581E559AD930ULL, "129363237.ogg"},
	{0xE23A87BD5169C283ULL, "plt_dae00pt_main_plot.plo"},
	{0xE23AA3C2BE3100C5ULL, "583361331.ogg"},
	{0xE23AB9C97A1C57F7ULL, "vxw_encounter2_pv_0_1.vlm"},
	{0xE23AE870773EA4F5ULL, "hlac_lt_0_lowtown_docks_-131072_0z.dds"},
	{0xE23B430D56AEFE54ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfl.evt"},
	{0xE23B62B934EB9703ULL, "6089431_m.fxe"},
	{0xE23BBE41B0C66AF9ULL, "6016140_f.fxe"},
	{0xE23BDDAD1D4D778CULL, "6031747_m.fxe"},
	{0xE23BDF8CC4C605A9ULL, "447736990.ogg"},
	{0xE23C091B4FB69A03ULL, "mag320_mage_conspirator6040773.cl"},
	{0xE23C1B33B604C883ULL, "643876252.ogg"},
	{0xE23C28E839E027CAULL, "6126610_m.fxe"},
	{0xE23C75795F79C420ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeoc.phy"},
	{0xE23CB9CBF5764822ULL, "vxw_encounter2_np_0_0.vlm"},
	{0xE23CDAF342814DFEULL, "764927708.ogg"},
	{0xE23CF1495606AA4DULL, "6128938_m.fxe"},
	{0xE23D5B7D166444D3ULL, "6176737_m.fxe"},
	{0xE23DDDA72456525DULL, "6018817_m.fxe"},
	{0xE23E000C21DAEB79ULL, "6043804_m.fxe"},
	{0xE23EA8B19AD4C928ULL, "one100dg_kelder6055596.cl"},
	{0xE23F0983B5DC364AULL, "follower_banter6167955.cl"},
	{0xE23F678541250168ULL, "6042914_m.fxe"},
	{0xE23F89322ED9CB35ULL, "419272747.ogg"},
	{0xE23F95890B27C603ULL, "6156461_m.fxe"},
	{0xE23FCD06236829E6ULL, "art\\characters\\playercharacter\\humanfemale\\hoo_nobb.phy"},
	{0xE24061266D77499FULL, "vxca_caverns1_ec6_0_0.vlm"},
	{0xE24077B4F215CB96ULL, "dae111ar_tavern.are"},
	{0xE240B01C504E2784ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_nud_0t.dds"},
	{0xE240D2DE85D98B29ULL, "vxlt_darktown_4za_0_0.vlm"},
	{0xE2411B9E07F178F8ULL, "art\\characters\\creatures\\mubari\\animation\\rundescend.evt"},
	{0xE2414C4A831AA590ULL, "vxlt_gallowscourtyard_n_6xq_0_0.vlm"},
	{0xE2419B67D6EBC336ULL, "vxht_mansion2_1jt_0_0.vlm"},
	{0xE241D55A17D27C92ULL, "950259985.ogg"},
	{0xE2421D90F1E5F681ULL, "w_bonepitdestroyed_a3_118_0.phy"},
	{0xE2422136075538DFULL, "844765887.ogg"},
	{0xE242414D9F2A8B39ULL, "6048496_m.fxe"},
	{0xE242933A6D6651C1ULL, "gen00fl_aveline6116639.cl"},
	{0xE2432A13F493EFACULL, "242902049.ogg"},
	{0xE243941B08907C5CULL, "6208345_m.fxe"},
	{0xE243D03AB52C7621ULL, "loadhints_random.gda"},
	{0xE2441CE427F68A89ULL, "vxw_sundermount_2ok_0_0.vlm"},
	{0xE2446D9172FA192CULL, "875425953.ogg"},
	{0xE24474413B293809ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shortbed_sa.dds"},
	{0xE24524464AD6DF6BULL, "613593796.ogg"},
	{0xE245305E1BFCCCD9ULL, "hlac_ca_cave_2194_11x.dds"},
	{0xE2462EF01E8D3DE5ULL, "171666503.ogg"},
	{0xE24653F6863823A7ULL, "hlac_lt_0_lowtown_a2_n_c_-3_1z.dds"},
	{0xE2467DA48D545889ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.cess.ani"},
	{0xE24681EA276A3E86ULL, "vxht_mansion_n_6i1_0_0.vlm"},
	{0xE246F00B9F529436ULL, "gen00fl_isabela6097356.cl"},
	{0xE246F893CC21FD16ULL, "6042439_m.fxe"},
	{0xE2470D01BB8D93C6ULL, "one100dg_vanard6017492.cl"},
	{0xE247A36DADF6DB93ULL, "6016691_m.fxe"},
	{0xE247BEB347208DC4ULL, "hlac_ht_chantry_2534_52x.dds"},
	{0xE247E2F0A54D98C8ULL, "ave200_donnic6061274.cl"},
	{0xE24828003C2981B2ULL, "6161189_m.fxe"},
	{0xE2483B0D56B97A8DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfl.ani"},
	{0xE248A19F0F87761FULL, "hlac_w_encounter3_-327679_0y.dds"},
	{0xE248E255BE40B322ULL, "6000802_m.fxe"},
	{0xE2494FFBFE35F985ULL, "mag100_cullen6082344.cl"},
	{0xE249716BCB86A4E9ULL, "zz_vault_debug6203289.cl"},
	{0xE249CF4E94F69373ULL, "zz_isa_debug6133211.cl"},
	{0xE249DBB21858DEACULL, "vxht_chantry_n_2zz_0_0.vlm"},
	{0xE24A136280598574ULL, "llac_lt_0_lowtown_a2_-2_6.dds"},
	{0xE24AA020C55F7CB9ULL, "hlac_dr_primevaltomb_658_10y.dds"},
	{0xE24B2BA8E76C60A2ULL, "vxw_encounter2_iq_0_0.vlm"},
	{0xE24B76B3A1F30E48ULL, "vxlt_0_lowtown_a2_n_c_f4n_0_0.vlm"},
	{0xE24B824702B781AEULL, "lt_0_lowtown_docks_a3_238_0.mmh"},
	{0xE24B83D24772D543ULL, "763630295.ogg"},
	{0xE24B89C5B379D610ULL, "6045963_f.fxe"},
	{0xE24B8E4702B79645ULL, "lt_0_lowtown_docks_a3_238_0.mao"},
	{0xE24B9C4702B7AD9CULL, "lt_0_lowtown_docks_a3_238_0.msh"},
	{0xE24BE2F259CB098AULL, "follower_banter6148714.cl"},
	{0xE24C8D21CB87BE60ULL, "mag211_marethari6025310.cl"},
	{0xE24CAD4345642B20ULL, "one120_athenril6089431.cl"},
	{0xE24CDCE6EC6C9794ULL, "vxlt_0_lowtown_a2_etx_0_0.vlm"},
	{0xE24D5DB9E6CDB5DAULL, "vxht_keep_4bu_0_0.vlm"},
	{0xE24D76333174F38FULL, "906232490.ogg"},
	{0xE24D9170246F28BEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2.mmh"},
	{0xE24D98A48D5AA50BULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.cess.gad"},
	{0xE24DB0BDCD696835ULL, "925511833.ogg"},
	{0xE24DBE95391488CFULL, "787351163.ogg"},
	{0xE24DDA5118FD484AULL, "corpses_sa.dds"},
	{0xE24E46B2641F6D33ULL, "97800265.ogg"},
	{0xE24ECA69BFBBC4CEULL, "hlac_lt_0_lowtown_docks_a3_196607_3y.dds"},
	{0xE24F2EA6DE055292ULL, "830342909.ogg"},
	{0xE24F94D3F3ABE696ULL, "lt_0_lowtown_docks_a3_228_0.mao"},
	{0xE24F9D5FDFE36557ULL, "lt_0_lowtown_docks_218_0.mmh"},
	{0xE24FA0D3F3ABFB35ULL, "lt_0_lowtown_docks_a3_228_0.mmh"},
	{0xE24FA15FDFE36C64ULL, "lt_0_lowtown_docks_218_0.mao"},
	{0xE24FA2D3F3ABFEABULL, "lt_0_lowtown_docks_a3_228_0.msh"},
	{0xE24FB35FDFE38AC9ULL, "lt_0_lowtown_docks_218_0.msh"},
	{0xE24FBFA7AE96B39BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_window.phy"},
	{0xE24FE5B06039ACD2ULL, "zz_approval_debug6003797.cl"},
	{0xE250066F9450D210ULL, "698924163.ogg"},
	{0xE2500D44C8E1DBB6ULL, "vxht_mansion2_1mu_0_0.vlm"},
	{0xE25048093E0AB3F8ULL, "combat_ar_r5_combo.gda"},
	{0xE25060D823582E08ULL, "6103294_f.fxe"},
	{0xE250C9734BDD34F5ULL, "hlac_ca_caverns2prc_5802_60y.dds"},
	{0xE2511D2B1BA5DBA0ULL, "qcr000_cut_arishok_response6039425.cl"},
	{0xE2514481353CFC2FULL, "follower_banter6173117.cl"},
	{0xE251AC12402874DFULL, "752461538.ogg"},
	{0xE251E3E0F8190A26ULL, "plt_ave000pt_gift_shield.plo"},
	{0xE25208B0B90E2EBCULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy.ani"},
	{0xE252182FC194C348ULL, "or_db_dot.mao"},
	{0xE2522312A82177BDULL, "vxlt_darktown_5lz_0_0.vlm"},
	{0xE2524F14DAB78AA0ULL, "241422168.ogg"},
	{0xE252732FE87AB210ULL, "vxht_0_hightown_a3_4n6_0_0.vlm"},
	{0xE2528EC119ECA309ULL, "vxw_sundermount_2pj_0_0.vlm"},
	{0xE2528F6DE2C68D79ULL, "vxht_keep_63s_0_0.vlm"},
	{0xE252BC4DA5BA750DULL, "800193406.ogg"},
	{0xE252E094ECFC2A83ULL, "622043354.ogg"},
	{0xE253018AEB47EC27ULL, "w_woundedcoast_43_0.rml"},
	{0xE2537D524B125CE4ULL, "vxdr_primevaltomb_7zq_0_0.vlm"},
	{0xE2539581C670EB37ULL, "fen110_fenris6100353.cl"},
	{0xE25495A48D60BF70ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_t_p.cess.evt"},
	{0xE2549FEA9B712201ULL, "zz_combat_demo.lst"},
	{0xE2552FFD04F1C673ULL, "vxlt_gallowsprison_n_6lp_0_0.vlm"},
	{0xE255E42B82FF6DCDULL, "vxht_mansion_n_3yo_0_0.vlm"},
	{0xE256827863CC9212ULL, "299025061.ogg"},
	{0xE256845E608410C9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_p.ani"},
	{0xE256E814450C3C57ULL, "w_bonepitdestroyed_a3.rml"},
	{0xE257061550861424ULL, "hlac_w_woundedcoast_-131066_0y.dds"},
	{0xE2575665B1C924EFULL, "6081307_m.fxe"},
	{0xE2576501280F8F0EULL, "cineanim_global_events.bnk"},
	{0xE2582D59431346FDULL, "vxdr_primevaltomb_4hy_0_0.vlm"},
	{0xE258426C789D7084ULL, "1027056807.ogg"},
	{0xE258467318D7C803ULL, "6087631_m.fxe"},
	{0xE258F8ABE570FA21ULL, "vxht_bartrand_n_5yz_0_0.vlm"},
	{0xE258F91B5106D208ULL, "vxlt_0_lowtown_docks_a3_bun_0_0.vlm"},
	{0xE25915B0B91463FEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy.gad"},
	{0xE259A62D9A8D2928ULL, "6005403_m.fxe"},
	{0xE25A7EEF7D57DA0CULL, "6015833_m.fxe"},
	{0xE25A92865A13AF44ULL, "llac_lt_0_lowtown_a3_n_c_-3_7.dds"},
	{0xE25B258CD52FCB15ULL, "6051274_m.fxe"},
	{0xE25B5479AB377D6FULL, "825813435.ogg"},
	{0xE25B68C968303E01ULL, "sst_ss_varian_ilithis.gda"},
	{0xE25B97EA9B7733F8ULL, "zz_combat_demo.ncs"},
	{0xE25B9DD127EAC56CULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga102a_0t.dds"},
	{0xE25BA73DB0768175ULL, "lt_0_lowtown_docks_n_129_0.msh"},
	{0xE25BA93DB07684EBULL, "lt_0_lowtown_docks_n_129_0.mmh"},
	{0xE25BB53DB0769888ULL, "lt_0_lowtown_docks_n_129_0.mao"},
	{0xE25C05A50807027EULL, "lt_darktown_3567_lc.mmh"},
	{0xE25C9F5E6088AA4BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_p.gad"},
	{0xE25D2E521FA496A7ULL, "6139787_m.fxe"},
	{0xE25D9B4413B89BC8ULL, "6176486_m.fxe"},
	{0xE25DB69869C5BC29ULL, "vxlt_darktown_4lg_0_0.vlm"},
	{0xE25E48627AE1E086ULL, "70983850.ogg"},
	{0xE25E91CD7C41929DULL, "vxlt_0_lowtown_docks_ba8_0_0.vlm"},
	{0xE25EA9D382F6A25CULL, "vxht_mansion2b_2ov_0_0.vlm"},
	{0xE25F04350CBB9329ULL, "genlock_axe.utc"},
	{0xE25F1368B8F08A70ULL, "ht_0_hightown_a3_25_0.rml"},
	{0xE25F908AC1A59EF4ULL, "418053384.ogg"},
	{0xE25FC94A91D197ABULL, "dre000_varric6207944.cl"},
	{0xE25FF0B0B91A431DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rdy.evt"},
	{0xE25FFDA1B2DBE7BEULL, "vxlt_0_lowtown_a3_n_c_e8f_0_0.vlm"},
	{0xE2603C76609B1F35ULL, "6174764_m.fxe"},
	{0xE26094EF9A17903DULL, "vxlt_darktown_5qy_0_0.vlm"},
	{0xE261361E7210E740ULL, "6082987_m.fxe"},
	{0xE2613BDD4C9BF30EULL, "6173686_f.fxe"},
	{0xE2614FC352106BB2ULL, "6030438_m.fxe"},
	{0xE261745A4CB5406FULL, "6019121_f.fxe"},
	{0xE261750C37A86CA8ULL, "453235712.ogg"},
	{0xE26192795F9911B5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeoc.msh"},
	{0xE26194795F99152BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeoc.mmh"},
	{0xE261A0795F9929C8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railspikeoc.mao"},
	{0xE261D937F90F567EULL, "6116492_f.fxe"},
	{0xE26220B0D4A4591AULL, "ico_mm_store.dds"},
	{0xE262B2C29F4B649AULL, "gen00fl_isabela6090872.cl"},
	{0xE2630774E1B4400CULL, "ico_potion_wardencologne.dds"},
	{0xE263118B71214D4CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06a.mmh"},
	{0xE263178B7121573EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw06a.msh"},
	{0xE2635475DA7AF1DCULL, "lowtown_male_cityguard_mob6150491.cl"},
	{0xE2636B0C5F627CEAULL, "hlac_w_encounter3_-196612_0y.dds"},
	{0xE2636B13987D236EULL, "zz_ave_debug6134003.cl"},
	{0xE2637F2E6CFDC4A3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\prp_spruce2_sa.dds"},
	{0xE26397DABEBBF70BULL, "vxlt_0_lowtown_docks_a3_8pl_0_0.vlm"},
	{0xE2639C5E608EC3B0ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_l90_p.evt"},
	{0xE263A1D9F6E7DEF3ULL, "vxlt_gallowsprison_n_6is_0_0.vlm"},
	{0xE263AA383E62AB08ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_6x.dds"},
	{0xE263BF0AE3D9D11EULL, "vxht_bartrand_n_5qa_0_0.vlm"},
	{0xE2642AE944D95841ULL, "hla_ht_0_hightown_a2_65538_0z.dds"},
	{0xE2647FC0597A5EDDULL, "282028977.ogg"},
	{0xE2657EB4F234F461ULL, "dae111ar_tavern.lst"},
	{0xE2666D5EFED71A08ULL, "dae200_cullen6140289.cl"},
	{0xE266742495227B70ULL, "6049458_m.fxe"},
	{0xE26678286751AF38ULL, "hlac_lt_0_lowtown_131070_2y.dds"},
	{0xE266BF6BFC450FACULL, "vxht_chantry_n_2px_0_0.vlm"},
	{0xE2678A545D5B633EULL, "512694579.ogg"},
	{0xE267CCBA98BCA11FULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_3_sylvesta.gad"},
	{0xE267D15EB0286932ULL, "gen00fl_aveline6121975.cl"},
	{0xE2683EC5C1D9F461ULL, "llac_lt_0_lowtown_docks_-262144_1.dds"},
	{0xE268E8B9FEDF4544ULL, "6154279_m.fxe"},
	{0xE2697840421E7EB6ULL, "art\\vfx\\textures\\water_decal.dds"},
	{0xE2699799DA01C2CFULL, "709734612.ogg"},
	{0xE26A17AC4D8B13C4ULL, "mag110_arianni6016779.cl"},
	{0xE26A5E5491936AEDULL, "sp_module_ppicker_closed.ncs"},
	{0xE26A7869483FDB73ULL, "gen_im_trs_rdr_junk2_04.uti"},
	{0xE26A92353E73D730ULL, "gen00fl_aveline6139262.cl"},
	{0xE26B018AEB5C7E44ULL, "w_woundedcoast_43_0.mmh"},
	{0xE26B058AEB5C8497ULL, "w_woundedcoast_43_0.mao"},
	{0xE26B178AEB5CA3E6ULL, "w_woundedcoast_43_0.msh"},
	{0xE26B47C896FC7D5BULL, "atl_saveload_dxt5_dat.xml"},
	{0xE26B9B71FB741315ULL, "vxlt_0_lowtown_a2_fea_0_0.vlm"},
	{0xE26BB2D4A423D0D7ULL, "hlac_ht_0_hightown_a3_0_1x.dds"},
	{0xE26C283B27DEDC56ULL, "6119509_f.fxe"},
	{0xE26C4EEB08B16CACULL, "hlac_w_blightlands_327683_0x.dds"},
	{0xE26C564233C9DDAEULL, "6025025_m.fxe"},
	{0xE26C6C7BC37B6816ULL, "hlac_dr_slavers_3216_8x.dds"},
	{0xE26C76B4F23B0558ULL, "dae111ar_tavern.ncs"},
	{0xE26D3943EAB01B85ULL, "6020431_f.fxe"},
	{0xE26DAA3DD85A56F0ULL, "hlac_w_sundermount_-2_2z.dds"},
	{0xE26DD9D9F677CE89ULL, "5134457.ogg"},
	{0xE26EA1242B1ED59DULL, "vxht_0_hightown_n_cbb_0_0.vlm"},
	{0xE26EEC7FFACC8F64ULL, "592098904.ogg"},
	{0xE26F7327C6B2BE79ULL, "vxht_keep_69q_0_0.vlm"},
	{0xE27037D7258DF54EULL, "vxca_caverns1_mines_imd_0_0.vlm"},
	{0xE270BC64A8F708B5ULL, "vxdr_primevaltomb_354_0_1.vlm"},
	{0xE272165066BC6770ULL, "gen00fl_isabela6079668.cl"},
	{0xE27230E7D5CFE99EULL, "vxht_bartrand_n_5tb_0_0.vlm"},
	{0xE272536BBC607B4AULL, "gen00fl_varric6118776.cl"},
	{0xE2726007B96A9F2BULL, "dup4_cutoff.mao"},
	{0xE272DAE1506B6E48ULL, "dre000_sibling_has_blight6036551.cl"},
	{0xE273AA7F034B80D2ULL, "50521409.ogg"},
	{0xE2741DF8720F4EB7ULL, "t3_wdg_db3.tnt"},
	{0xE27438119902562EULL, "vxlt_gallowstemplar_5ar_0_0.vlm"},
	{0xE2754C6B9DA6FB1DULL, "w_bonepitdestroyed_a3_107_0.phy"},
	{0xE2758CAD5393FC0FULL, "915432126.ogg"},
	{0xE275924C89FD07F9ULL, "vxlt_gallowstemplar_n_4tx_0_0.vlm"},
	{0xE275A252D5B7E8BAULL, "vxlt_undercityencounter_13n_0_0.vlm"},
	{0xE275FE909541B63EULL, "6101073_m.fxe"},
	{0xE2762598D710369CULL, "448414603.ogg"},
	{0xE276715489DFCDB2ULL, "dm_yevhen.mor"},
	{0xE27673F33F82BED3ULL, "12882684.ogg"},
	{0xE2770E2161AC8774ULL, "vxw_bonepitmining_a2_pb_0_0.vlm"},
	{0xE277184E80B896E7ULL, "vxht_0_hightown_a3_b5z_0_0.vlm"},
	{0xE2777174FCC8EB25ULL, "6025996_m.fxe"},
	{0xE27772A3CD2E5F23ULL, "hlac_lt_gallowscourtyard_131070_1x.dds"},
	{0xE2777C2FD754B1A3ULL, "6187889_m.fxe"},
	{0xE2779FD1C0A7A838ULL, "vxlt_undercityencounter_u1_0_0.vlm"},
	{0xE278E73A9982D5B4ULL, "vxca_cavern1_haunted_icr_0_2.vlm"},
	{0xE2799DBD7EA68E8AULL, "6033677_m.fxe"},
	{0xE279B77EEB13221AULL, "6072133_m.fxe"},
	{0xE279CF37C352AACCULL, "vxht_mansion2b_1mw_0_0.vlm"},
	{0xE27A2C56B806D9EFULL, "797764160.ogg"},
	{0xE27B014AF377E29AULL, "ht_store.arl"},
	{0xE27B15D1D755B119ULL, "vxw_woundedcoast_1d5_0_0.vlm"},
	{0xE27B75D3B4A5F018ULL, "zz_followers_banter6158571.cl"},
	{0xE27B83D156EA3C9DULL, "pro000_friedrich6050202.cl"},
	{0xE27B896740B19FCCULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v1.ani"},
	{0xE27C4F08293EFB5CULL, "hlac_w_encounter3_-65540_0z.dds"},
	{0xE27C875D943EA803ULL, "hlac_ht_0_hightown_a3_-2_0x.dds"},
	{0xE27CA6A7AEBD773AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_window.msh"},
	{0xE27CA9BA98CEC265ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_3_3_sylvesta.ani"},
	{0xE27CB4A7AEBD8F73ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_window.mao"},
	{0xE27CB8A7AEBD9580ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww22_window.mmh"},
	{0xE27CF7943DA0E0F6ULL, "6121901_m.fxe"},
	{0xE27D521908EEB49DULL, "gen00fl_anders6119299.cl"},
	{0xE27D5F7420CCA90BULL, "855112636.ogg"},
	{0xE27DD1AACBBAEC43ULL, "591695897.ogg"},
	{0xE27DE2FD7980D9B2ULL, "dae000_merchant_dm_mob.cnv"},
	{0xE27DE6FD200AE85EULL, "m_dae221ar_brothel_nt.bnk"},
	{0xE27E1025EE7E4894ULL, "vxdr_primevalentrance_37v_0_0.vlm"},
	{0xE27EE7A54E27A1A7ULL, "gen00fl_anders6111129.cl"},
	{0xE27FA0337D493169ULL, "hlac_lt_0_lowtown_a2_n_c_65536_9y.dds"},
	{0xE27FD467F933476FULL, "hlac_ht_keep_n_4519_6x.dds"},
	{0xE27FE93FBDE7DD31ULL, "hlac_lt_hangedman_-1_0x.dds"},
	{0xE2803437ACC713CDULL, "118603776.ogg"},
	{0xE28104FD7983490AULL, "dae000_merchant_dm_mob.bnk"},
	{0xE281D7DFAA4015DAULL, "783434188.ogg"},
	{0xE2820E8A9F66E373ULL, "40810017.ogg"},
	{0xE2826E81BC1917B3ULL, "6051391_m.fxe"},
	{0xE28294F5E9D8B86DULL, "tce_wallsm_0.msh"},
	{0xE282966740B7D40EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v1.gad"},
	{0xE282A6F5E9D8D6F3ULL, "tce_wallsm_0.mmh"},
	{0xE282B55517A2B207ULL, "blue25.dds"},
	{0xE282E5D8F558B680ULL, "ca_cave_mmi.gff"},
	{0xE2835D48D73664D5ULL, "vxlt_0_lowtown_n_exj_0_0.vlm"},
	{0xE2842BA1C9197B9FULL, "vxw_bonepit_113_0_0.vlm"},
	{0xE284CBD1FD7E0B30ULL, "hlac_ht_chantry_2534_20y.dds"},
	{0xE286416B9DB55B89ULL, "w_bonepitdestroyed_a3_107_0.mao"},
	{0xE2864D6B9DB57022ULL, "w_bonepitdestroyed_a3_107_0.mmh"},
	{0xE2864F6B9DB573D8ULL, "w_bonepitdestroyed_a3_107_0.msh"},
	{0xE2865AEDD4D80E22ULL, "6201612_m.fxe"},
	{0xE28666B00FC395DAULL, "w_encounter2_camp_40_0.mmh"},
	{0xE2866AB00FC39C11ULL, "w_encounter2_camp_40_0.mao"},
	{0xE28678B00FC3B420ULL, "w_encounter2_camp_40_0.msh"},
	{0xE286B51E006ED6D9ULL, "vxca_caverns2_2rs_0_0.vlm"},
	{0xE28736CBF51F1EFDULL, "6061274_f.fxe"},
	{0xE2875A3C38356649ULL, "692905727.ogg"},
	{0xE287F083EBD9F00CULL, "vxw_encounter2_39_0_0.vlm"},
	{0xE28824BEEF1AD7B8ULL, "genip_door_gallowin_normal.utp"},
	{0xE288572845A02F4FULL, "vxdr_primevaltomb_8q1_0_1.vlm"},
	{0xE288916740BC37EDULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_attk_v1.evt"},
	{0xE288C568958CCE3EULL, "876368882.ogg"},
	{0xE288F31C6620EE16ULL, "6172912_m.fxe"},
	{0xE28907C1AF30B360ULL, "vxht_0_hightown_b9d_0_0.vlm"},
	{0xE2892188DDB3CBEFULL, "6110052_m.fxe"},
	{0xE2892ADF0BBF84CDULL, "6102567_m.fxe"},
	{0xE28987AEC94BC999ULL, "vxw_woundedcoast_1o4_0_0.vlm"},
	{0xE28990045D68DD10ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pulpit_d.dds"},
	{0xE289F138DBAA6649ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_single.phy"},
	{0xE28A663EAF7C1CA8ULL, "vxw_bonepitmining_a2_7g_0_0.vlm"},
	{0xE28A76F76B7101E3ULL, "vxw_woundedcoast_1kz_0_0.vlm"},
	{0xE28A965AF37ADC38ULL, "809907076.ogg"},
	{0xE28AB803F5AFE8D0ULL, "892238084.ogg"},
	{0xE28B1468B915C6D8ULL, "ht_0_hightown_a3_25_0.anb"},
	{0xE28B943C02E7D9E9ULL, "hlac_ht_chantry_2534_6x.dds"},
	{0xE28BB53A6F421BBAULL, "gen00fl_aveline6122800.cl"},
	{0xE28BC43D30E00FF3ULL, "mrl000_marethari6069803.cl"},
	{0xE28C6CFAD07358E8ULL, "vxlt_0_lowtown_a2_gxr_0_0.vlm"},
	{0xE28C8BD1F2254985ULL, "dae000_pc_home_ambients6208542.cl"},
	{0xE28C9B049486E332ULL, "vxlt_0_lowtown_a2_n_c_ei8_0_0.vlm"},
	{0xE28CACF6DFE4CC26ULL, "vxw_woundedcoast_4w_0_0.vlm"},
	{0xE28CB92503B6E2BBULL, "bowlingforferelden.ncs"},
	{0xE28CF968D6FFB53BULL, "vxlt_backalley_g_1df_0_0.vlm"},
	{0xE28D424EB342DC77ULL, "6088256_m.fxe"},
	{0xE28D4BEB05060449ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax101a_0d.dds"},
	{0xE28D955E503A9B29ULL, "413586453.ogg"},
	{0xE28DC6AF5379CB45ULL, "zz_cave_accolades_3_v1.cut"},
	{0xE28DFC0DC511F224ULL, "715166389.ogg"},
	{0xE28E54F1AC16C12FULL, "vxlt_0_lowtown_a2_ep2_0_1.vlm"},
	{0xE28E687024A67A59ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule_2x2.phy"},
	{0xE28E6C65471CE0E7ULL, "gen00fl_aveline6139832.cl"},
	{0xE28EA821B9ADEAA4ULL, "vxht_0_hightown_a2_cem_0_0.vlm"},
	{0xE28EFEF68291BEDDULL, "vxlt_0_lowtown_docks_n_bcm_0_0.vlm"},
	{0xE28F664A78E585C6ULL, "6219280_m.fxe"},
	{0xE28F66FDC636EC92ULL, "vxht_0_hightown_n_cat_0_0.vlm"},
	{0xE28F9AAA6AFA1CFAULL, "327678878.ogg"},
	{0xE29017718049AF0AULL, "llac_ht_keep_n_3732_1.dds"},
	{0xE2901815B840F688ULL, "hlac_dr_slavers_3217_4y.dds"},
	{0xE2901B6F16DFF1A4ULL, "6155389_f.fxe"},
	{0xE290534816CF356BULL, "6214707_m.fxe"},
	{0xE290A0F34BFD1AFAULL, "6175328_m.fxe"},
	{0xE290FA967BA5A7F2ULL, "274608968.ogg"},
	{0xE2911BCB7CEE872EULL, "vxlt_gallowstemplar_5wt_0_0.vlm"},
	{0xE2912A94A45B7A07ULL, "plt_one100pt_herb_gland.plo"},
	{0xE29193A7FA84B1AEULL, "56762813.ogg"},
	{0xE291B4CFD17BF403ULL, "llac_w_woundedcoast_0_4.dds"},
	{0xE291D989B01696FEULL, "invalid.cur"},
	{0xE29222C9BE5AC460ULL, "gen00fl_aveline6115417.cl"},
	{0xE2923B01F314CEB6ULL, "879870694.ogg"},
	{0xE2928A241EE37FE0ULL, "vxdr_primevaltomb_6dv_0_0.vlm"},
	{0xE293D9C1B0D19638ULL, "464755265.ogg"},
	{0xE293F9241EE526B6ULL, "vxht_bartrand_n_5p6_0_0.vlm"},
	{0xE2943A95D8814BE0ULL, "6207739_m.fxe"},
	{0xE2944CCE5E9B65A7ULL, "6126633_m.fxe"},
	{0xE29453D98A0949D6ULL, "hlac_lt_0_lowtown_a2_-65535_7y.dds"},
	{0xE294A5DBB084CBA8ULL, "565664006.ogg"},
	{0xE294CD4E6C98C671ULL, "vxlt_undercityencounter_1lt_0_0.vlm"},
	{0xE29522609CE1CFB3ULL, "846928978.ogg"},
	{0xE29586F5A162795BULL, "lgt120_orz_bln_official.bnk"},
	{0xE2959F29201369FFULL, "vxht_mansion2b_30y_0_0.vlm"},
	{0xE295BE36FDA5D410ULL, "follower_barks6153409.cl"},
	{0xE29641D9EDBEFCF1ULL, "843592785.ogg"},
	{0xE2966F96ECF33814ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_nud_0d.dds"},
	{0xE2969E8F88156104ULL, "546771860.ogg"},
	{0xE2971FDC6A8B7A43ULL, "619691241.ogg"},
	{0xE297C026CB8BE5DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_loop.gad"},
	{0xE297E40D8B87BC7CULL, "llac_lt_foundry_n_320_5.dds"},
	{0xE29830EEA2B120CEULL, "vxht_mansion2b_31w_0_0.vlm"},
	{0xE29849621F2B2635ULL, "cod_mgc_chant_schism.utp"},
	{0xE29852CAEB27D7CAULL, "6116486_m.fxe"},
	{0xE29871C0822A3AE6ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_imp_c.impa.evt"},
	{0xE298F10FAF8B315EULL, "mrl150_merrill6070451.cl"},
	{0xE29988F5A166651BULL, "lgt120_orz_bln_official.cnv"},
	{0xE2999175F31A9364ULL, "face2.psh.fxo"},
	{0xE29B018502441EE9ULL, "llac_w_bonepitmining_a2_0_7.dds"},
	{0xE29B1ED3D1DAE4A6ULL, "vxw_woundedcoast_1x_0_0.vlm"},
	{0xE29B41174C9CBB03ULL, "416014177.ogg"},
	{0xE29B946FC90A609DULL, "619676217.ogg"},
	{0xE29BB42BA538F4F7ULL, "6083356_m.fxe"},
	{0xE29C00AB968D2F30ULL, "vxlt_0_lowtown_a2_n_f4x_0_0.vlm"},
	{0xE29C515F2435F9B5ULL, "xx_extbrick.dds"},
	{0xE29C924A7E0941B4ULL, "6084116_m.fxe"},
	{0xE29CA5976E1C7149ULL, "fer_stone.mao"},
	{0xE29CC4E0F858C165ULL, "plt_ave000pt_gift_shield.nss"},
	{0xE29D19BFBAA34D65ULL, "vxlt_0_lowtown_a2_fe1_0_0.vlm"},
	{0xE29D7D7246C72C45ULL, "vxlt_0_lowtown_docks_a3_9ra_0_0.vlm"},
	{0xE29DB3BB351F6EBAULL, "t1_mub_pi1.tnt"},
	{0xE29DF79522B223A6ULL, "123131896.ogg"},
	{0xE29EDF173568F441ULL, "6100077_m.fxe"},
	{0xE29F2291BB812D3DULL, "vxca_cavern1_haunted_j2d_0_1.vlm"},
	{0xE29F807E387ECDABULL, "follower_banter6160993.cl"},
	{0xE29F88C082308031ULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_imp_c.impa.gad"},
	{0xE2A0279C381BED00ULL, "llac_ht_0_hightown_a2_0_3.dds"},
	{0xE2A0410D609943F1ULL, "954967973.ogg"},
	{0xE2A04630174A1981ULL, "6083044_m.fxe"},
	{0xE2A04B764B9B3802ULL, "vxlt_0_lowtown_a2_n_ejc_0_0.vlm"},
	{0xE2A04D1BF4DAD1D0ULL, "vxlt_gallowscourtyard_6zt_0_0.vlm"},
	{0xE2A06B46D34FC71BULL, "26295467.ogg"},
	{0xE2A1019AFD8964E0ULL, "qcr220_meredith6054754.cl"},
	{0xE2A1220AC7B72511ULL, "6076798_m.fxe"},
	{0xE2A138403FA211C2ULL, "6067764_m.fxe"},
	{0xE2A159225AC55CB7ULL, "seb321_leliana6180948.cl"},
	{0xE2A263B8378ED0F4ULL, "vxw_bonepitmining_a2_aa_0_0.vlm"},
	{0xE2A2716448F137A2ULL, "fex_bd_c_chm.mmh"},
	{0xE2A2736448F13B58ULL, "fex_bd_c_chm.msh"},
	{0xE2A2F0596388B399ULL, "gen_im_fla_merril_02.uti"},
	{0xE2A333D2A7072EA2ULL, "dae000_arishok6074061.cl"},
	{0xE2A396CDB18D87BCULL, "6115629_m.fxe"},
	{0xE2A411061209827FULL, "vxht_mansion2b_37x_0_0.vlm"},
	{0xE2A43F5B86EE4ADDULL, "6208825_m.fxe"},
	{0xE2A44A2F69BA8812ULL, "vxlt_gallowstemplar_435_0_0.vlm"},
	{0xE2A4E8CA7A2CD409ULL, "sw_bl_sh.mao"},
	{0xE2A4EE5CCB1BECC7ULL, "6165648_f.fxe"},
	{0xE2A52290F23A7CBDULL, "w_bonepitdestroyed_a3_118_0.mao"},
	{0xE2A52690F23A83F6ULL, "w_bonepitdestroyed_a3_118_0.mmh"},
	{0xE2A53090F23A94F4ULL, "w_bonepitdestroyed_a3_118_0.msh"},
	{0xE2A57D39B2B86FCEULL, "hlac_w_sundermount_-262143_0x.dds"},
	{0xE2A629ED0784500AULL, "6034046_f.fxe"},
	{0xE2A689C08236A0DFULL, "art\\vfx\\combat\\vfx_m_spi_deathsyphon_imp_c.impa.ani"},
	{0xE2A706CA12AE78F0ULL, "mag170_thrask6016026.cl"},
	{0xE2A71187587E213DULL, "vxw_woundedcoast_112_0_0.vlm"},
	{0xE2A7199B5031D510ULL, "hlac_lt_0_lowtown_n_-131073_2x.dds"},
	{0xE2A72C24B6E0A52DULL, "16684848.ogg"},
	{0xE2A736EC8FBC7F03ULL, "6037580_m.fxe"},
	{0xE2A7B65F4D1DEDCEULL, "vxlt_gallowstemplar_6sw_0_0.vlm"},
	{0xE2A7BE88F7FD5421ULL, "fen310_fenris6118697.cl"},
	{0xE2A7D0CD98D076BBULL, "230749723.ogg"},
	{0xE2A80DC0BF1287FFULL, "287697457.ogg"},
	{0xE2A83AED39CDF773ULL, "vxht_mansion_n_4gt_0_0.vlm"},
	{0xE2A83F635D6D0C6AULL, "hlac_lt_0_lowtown_a2_n_-65537_3x.dds"},
	{0xE2A898B1C6BB924EULL, "momnt_flare.mao"},
	{0xE2A8DE13C2C1CBDEULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf_bk.ani"},
	{0xE2A90B63D3774260ULL, "477112554.ogg"},
	{0xE2A92F72DC22440DULL, "vxht_0_hightown_a3_bpw_0_0.vlm"},
	{0xE2A9FF4AED66574EULL, "vxca_caverns1_mines_iyx_0_0.vlm"},
	{0xE2AA2722EDE37ED7ULL, "6101564_m.fxe"},
	{0xE2AA3D1D0D50B004ULL, "mag101_alain6032273.cl"},
	{0xE2AB0B44BBE9A431ULL, "hlac_ht_mansion2_818_56x.dds"},
	{0xE2AB66646621B693ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_visa_n.dds"},
	{0xE2ABB8D28E2B8B08ULL, "6153568_m.fxe"},
	{0xE2ABD778D4E213D3ULL, "6090842_m.fxe"},
	{0xE2ABDB26CB9CBCC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\wall_lean_left_loop.ani"},
	{0xE2ABE2B0667DEFDDULL, "vxlt_0_lowtown_docks_n_buk_0_0.vlm"},
	{0xE2ABF85B9DA84B9EULL, "vxht_bartrand_n_5hf_0_0.vlm"},
	{0xE2AC583E3EDC82A7ULL, "280138351.ogg"},
	{0xE2AC7C45A4F082F1ULL, "6176421_m.fxe"},
	{0xE2AC85B00FE499E5ULL, "w_encounter2_camp_40_0.phy"},
	{0xE2ACAC9906630FB1ULL, "mag210_arianni6024130.cl"},
	{0xE2AD72002F8152A0ULL, "81791754.ogg"},
	{0xE2ADC62444146C85ULL, "s_aoe_cyc.mao"},
	{0xE2ADF9FDCB03880EULL, "298276956.ogg"},
	{0xE2AEACFC7FC581D5ULL, "character_character_p17.psh.fxo.dx11"},
	{0xE2AEB0CB05474339ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_act_c.impa.ani"},
	{0xE2AF0EBD8ED865F2ULL, "gen00fl_isabela6110181.cl"},
	{0xE2AFA06A2A3A2122ULL, "dre000_cut2d_ragstoriches6025118.cl"},
	{0xE2AFCE393CF07461ULL, "hlac_lt_0_lowtown_a2_n_131073_1y.dds"},
	{0xE2AFD1B4AB10F090ULL, "6072078_m.fxe"},
	{0xE2AFEFB7B4F69B49ULL, "llac_ht_0_hightown_a2_-65536_4.dds"},
	{0xE2B067579BCB24C4ULL, "aip_follower_merril.gda"},
	{0xE2B079D71C57824BULL, "hlac_w_sundermount_262145_0z.dds"},
	{0xE2B09C0CC62B76ECULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinr.phy"},
	{0xE2B09C0FAF6B40A2ULL, "vxw_blightlands_8u_0_0.vlm"},
	{0xE2B0D5952984E974ULL, "vxw_bonepitmining_a2_lf_0_0.vlm"},
	{0xE2B11D204D1E7B5EULL, "vxw_woundedcoast_fy_0_0.vlm"},
	{0xE2B1303847AE598CULL, "dae000_arishok6068563.cl"},
	{0xE2B13CFF3DC14FCEULL, "llac_w_blightlands_-65540_0.dds"},
	{0xE2B1BECA744EECFAULL, "w_blightlands_197_0.msh"},
	{0xE2B1CCCA744F0433ULL, "w_blightlands_197_0.mao"},
	{0xE2B1D0CA744F0B40ULL, "w_blightlands_197_0.mmh"},
	{0xE2B1EB141F9CA6E9ULL, "llac_lt_0_lowtown_docks_n_131070_4.dds"},
	{0xE2B25CF8673F06D3ULL, "hlac_lt_warehouse_n_1_0x.dds"},
	{0xE2B277B059649317ULL, "art\\cinematics\\deathblows\\ogre\\cs_db_pc_ogrewarr.gad"},
	{0xE2B27BE3675A8F99ULL, "gen00fl_fenris6106697.cl"},
	{0xE2B2DE6CE9C39F9DULL, "mag360_samson6044883.cl"},
	{0xE2B33BDEB05F5DE1ULL, "gen00fl_sebastian6189512.cl"},
	{0xE2B366B00FEA83F1ULL, "w_encounter2_camp_40_0.rml"},
	{0xE2B3762FA9C9886CULL, "gen00fl_isabela6135616.cl"},
	{0xE2B3A98E0E962158ULL, "6087197_m.fxe"},
	{0xE2B3B649AF619155ULL, "mag200_feynriel6217342.cl"},
	{0xE2B40EC6669C8517ULL, "245032953.ogg"},
	{0xE2B41DD0640EEA21ULL, "hlac_lt_0_lowtown_a2_n_-1_2y.dds"},
	{0xE2B44678642EA4D2ULL, "691504401.ogg"},
	{0xE2B47EFDD4F02448ULL, "403842579.ogg"},
	{0xE2B4DD459F2E1319ULL, "vxw_woundedcoast_1p9_0_0.vlm"},
	{0xE2B4EB88ADA66506ULL, "ser221_gascard6122444.cl"},
	{0xE2B517234489896BULL, "lt_gallowscourtyard_n_45_0.rml"},
	{0xE2B553605E7616D4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1_tool.phy"},
	{0xE2B58BCB054D237BULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_act_c.impa.gad"},
	{0xE2B5A8C981B29783ULL, "6030480_m.fxe"},
	{0xE2B5F613C2CC7EAFULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf_bk.evt"},
	{0xE2B624AE40D1EFEEULL, "6017675_m.fxe"},
	{0xE2B677F029C4E162ULL, "dae000_hate_enemies6167512.cl"},
	{0xE2B69045EBB19C07ULL, "demo_rogue.utc"},
	{0xE2B6DA38DBD12D6EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_single.mmh"},
	{0xE2B6E638DBD14105ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_single.mao"},
	{0xE2B6F438DBD1595CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\4so_single.msh"},
	{0xE2B75A31173A56D1ULL, "w_sundermount_255_0.msh"},
	{0xE2B76831173A6EECULL, "w_sundermount_255_0.mao"},
	{0xE2B77431173A828FULL, "w_sundermount_255_0.mmh"},
	{0xE2B77EF24C02277DULL, "6139510_m.fxe"},
	{0xE2B7A14FCE185C8DULL, "vxht_0_hightown_a3_bov_0_0.vlm"},
	{0xE2B7AB7A6861410EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_forward.ani"},
	{0xE2B7DDD58A31C3B2ULL, "gen_im_wep_war_two_act3_08.uti"},
	{0xE2B7FAA0457C074FULL, "750092075.ogg"},
	{0xE2B901D516EA6EC4ULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_conj_c.mmh"},
	{0xE2BA1E2BFB1E9189ULL, "zz_followers_banter6157821.cl"},
	{0xE2BA4955C6D82C53ULL, "6095742_m.fxe"},
	{0xE2BA4FF806C2302BULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mage_scaredmage.ani"},
	{0xE2BAEF799A64A22BULL, "hlac_lt_0_lowtown_a2_65538_3x.dds"},
	{0xE2BB05DFB5702A39ULL, "14949530.ogg"},
	{0xE2BB4AFD3B71E9C3ULL, "follower_banter6155799.cl"},
	{0xE2BB65B9AC168C08ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsa_0d.dds"},
	{0xE2BBA8CB0551C060ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_act_c.impa.evt"},
	{0xE2BBD0B3D15FAC37ULL, "gen00fl_merrill6119660.cl"},
	{0xE2BBD48AD5264E54ULL, "ser191_emeric6104754.cl"},
	{0xE2BBF584A264BCA2ULL, "vxht_brothel_n_6kl_0_0.vlm"},
	{0xE2BBF965562729E0ULL, "272826049.ogg"},
	{0xE2BC24B9F4ADA9DAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_catacombs.mmh"},
	{0xE2BC28B9F4ADB011ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_catacombs.mao"},
	{0xE2BC34DFD3760B4AULL, "vxw_woundedcoast_19i_0_0.vlm"},
	{0xE2BC36B9F4ADC820ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_catacombs.msh"},
	{0xE2BCE657AA8DB6FCULL, "hlac_lt_0_lowtown_docks_a3_n_-3_1y.dds"},
	{0xE2BCF713C2D29F58ULL, "art\\characters\\creatures\\spider\\animation\\tal_dbuf_bk.gad"},
	{0xE2BD37D0177A2D90ULL, "vxht_bartrand_n_6s6_0_0.vlm"},
	{0xE2BD3FC065F56627ULL, "97529876.ogg"},
	{0xE2BD6BC19FD0B730ULL, "6088682_m.fxe"},
	{0xE2BD732F1FF80A3CULL, "vxw_woundedcoast_1dl_0_0.vlm"},
	{0xE2BDB8FD20407E64ULL, "m_dae221ar_brothel_nt.wbd"},
	{0xE2BDDFBAF4C5C960ULL, "vxdr_primevaltomb_6uu_0_0.vlm"},
	{0xE2BDEC4594C68EB6ULL, "127721958.ogg"},
	{0xE2BDFB4B7A480E3DULL, "393569523.ogg"},
	{0xE2BE628C6E73A59BULL, "vxca_caverns1_mines_gki_0_0.vlm"},
	{0xE2BECFEC624FEF77ULL, "vxlt_backalley_g_1dr_0_0.vlm"},
	{0xE2BF926003EA0B55ULL, "dae000_arishok6067734.cl"},
	{0xE2BFF2FA01BFC913ULL, "face_face_p2.psh.fxo"},
	{0xE2C030605E7FE2D1ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1_tool.msh"},
	{0xE2C03E605E7FFAECULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1_tool.mao"},
	{0xE2C04A605E800E8FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1_tool.mmh"},
	{0xE2C09E6653F673BFULL, "vxht_keep_n_5tx_0_0.vlm"},
	{0xE2C0B6CDDB2D83DBULL, "lowtown_male_beggar_sta6151057.cl"},
	{0xE2C13E291B447C2BULL, "cnvprev_human_female_rogue.utc"},
	{0xE2C163F21BB52324ULL, "w_encounter3_94_0.msh"},
	{0xE2C175F21BB5418DULL, "w_encounter3_94_0.mao"},
	{0xE2C179F21BB548C6ULL, "w_encounter3_94_0.mmh"},
	{0xE2C1839356543B0DULL, "6048290_m.fxe"},
	{0xE2C208887D21254CULL, "vxht_mansion2b_1th_0_0.vlm"},
	{0xE2C242FF30B3DAF4ULL, "6165654_f.fxe"},
	{0xE2C28DBAE26E77B7ULL, "301050664.ogg"},
	{0xE2C31F60892B49F8ULL, "static_static_p78.psh.fxo.dx11"},
	{0xE2C32D93F038468CULL, "vxw_sundermount_2be_0_0.vlm"},
	{0xE2C3EEDF4933BF82ULL, "mcr310_argument6127574.cl"},
	{0xE2C40C3DD758D522ULL, "mrl150_cut_merrill_intro6117750.cl"},
	{0xE2C4550061D066F7ULL, "120165297.ogg"},
	{0xE2C4A37A686BBE5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_forward.evt"},
	{0xE2C57F5B9DBFFFA5ULL, "vxlt_0_lowtown_docks_a3_n_cvi_0_1.vlm"},
	{0xE2C62E5E39286E9AULL, "174207625.ogg"},
	{0xE2C63328464345CFULL, "6154166_m.fxe"},
	{0xE2C6A4B05975889DULL, "art\\cinematics\\deathblows\\ogre\\cs_db_pc_ogrewarr.ani"},
	{0xE2C72230FE02F418ULL, "vxht_brothel_n_5k9_0_0.vlm"},
	{0xE2C75D882E46FDA9ULL, "620249934.ogg"},
	{0xE2C7B74671374D52ULL, "vxlt_0_lowtown_a2_n_gob_0_0.vlm"},
	{0xE2C7D8E5976CC330ULL, "881327933.ogg"},
	{0xE2C7EAA73AD2F066ULL, "hlac_lt_0_lowtown_docks_-65538_3y.dds"},
	{0xE2C80EE0AD3F88C4ULL, "873554031.ogg"},
	{0xE2C85C9AD326BE43ULL, "286492747.ogg"},
	{0xE2C86146E7028176ULL, "dre000_hireling26038683.cl"},
	{0xE2C8E63B7F99D16CULL, "vxlt_0_lowtown_docks_c5l_0_0.vlm"},
	{0xE2C94E0C060FCE51ULL, "mag291tr_nt_torpor.utt"},
	{0xE2CA02CB23A6A51DULL, "ht_0_hightown_a3_42_0_ll.mmh"},
	{0xE2CA6761945AD522ULL, "vxht_brothel_n_6tm_0_0.vlm"},
	{0xE2CB847A6871A848ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_gesture_forward.gad"},
	{0xE2CBA10D217D8182ULL, "vxlt_0_lowtown_a2_n_eun_0_0.vlm"},
	{0xE2CBAF6C7CEDFDFAULL, "dae000cr_grand_cleric6178465.cl"},
	{0xE2CBE50C11EE22BCULL, "vxw_woundedcoast_1om_0_0.vlm"},
	{0xE2CC5197E6BBE1E0ULL, "vxdr_primevaltomb_6xr_0_0.vlm"},
	{0xE2CC5882EF5CF02AULL, "vxht_0_hightown_n_ajs_0_0.vlm"},
	{0xE2CC97CB98432929ULL, "vxlt_warehouse_n_8g_0_0.vlm"},
	{0xE2CD4C10AE5D0B16ULL, "ave221_brennan6059778.cl"},
	{0xE2CDC10406EFC4C7ULL, "6140192_m.fxe"},
	{0xE2CE3BE2E8FF0D6BULL, "940637613.ogg"},
	{0xE2CEB373AA37DD21ULL, "246190414.ogg"},
	{0xE2CEF0F7AC89CC5BULL, "pro000_merchant6057566.cl"},
	{0xE2CF36F806D461B5ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mage_scaredmage.gad"},
	{0xE2CFAE36175AB36CULL, "vxht_keep_n_5hk_0_0.vlm"},
	{0xE2CFC2C0D4BA69BEULL, "738310747.ogg"},
	{0xE2CFD318E3552A15ULL, "mag170_thrask6015998.cl"},
	{0xE2CFF570484A538DULL, "6043390_m.fxe"},
	{0xE2D0226294B63B65ULL, "297822537.ogg"},
	{0xE2D023CA564B00BCULL, "523280950.ogg"},
	{0xE2D05BAA6C764762ULL, "and341ip_loot_corpse_1.utp"},
	{0xE2D0DEA5406AB3CAULL, "687898176.ogg"},
	{0xE2D0E4FAA4C4A7B5ULL, "6076384_f.fxe"},
	{0xE2D10A3A6D17FFABULL, "6177727_m.fxe"},
	{0xE2D1A804CF4DE586ULL, "llac_w_encounter2_131070_0.dds"},
	{0xE2D1C16AB8D746EDULL, "vxlt_backalley_g_1l8_0_0.vlm"},
	{0xE2D1C7D6454ABADAULL, "6144287_m.fxe"},
	{0xE2D1E5806F8C2A21ULL, "vxw_encounter2_qw_0_1.vlm"},
	{0xE2D220B50BF2569DULL, "vxlt_0_lowtown_docks_b90_0_0.vlm"},
	{0xE2D24515032FEA49ULL, "hlac_lt_0_lowtown_a2_-2_0y.dds"},
	{0xE2D284A8E3E64290ULL, "hlac_lt_0_lowtown_65535_3z.dds"},
	{0xE2D29462B1315D1FULL, "vxlt_gamlenshouse_n_kz_0_0.vlm"},
	{0xE2D2A5F735274D12ULL, "hlac_w_sundermount_-131076_0x.dds"},
	{0xE2D3173B47B5AFE9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\tre_stump01_d.dds"},
	{0xE2D32547ABB4BEF5ULL, "195660745.ogg"},
	{0xE2D357132712C066ULL, "vxht_mansion2_1q9_0_0.vlm"},
	{0xE2D363F76B80F098ULL, "6116619_m.fxe"},
	{0xE2D372B0A1390C40ULL, "246332790.ogg"},
	{0xE2D393D564589AD2ULL, "550437763.ogg"},
	{0xE2D4127DE499086FULL, "wos_01.mao"},
	{0xE2D438BE075EAAD6ULL, "vxw_bonepitdestroyed_a3_8r_0_0.vlm"},
	{0xE2D463D2DDE33491ULL, "vxlt_0_lowtown_a2_g94_0_0.vlm"},
	{0xE2D4A64B8A41D368ULL, "vxlt_0_lowtown_a2_gco_0_0.vlm"},
	{0xE2D4EF97BDEB43FEULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_stoney2_n.dds"},
	{0xE2D511BBBF3FBADCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosed.ani"},
	{0xE2D521CA0BAAE820ULL, "38303928.ogg"},
	{0xE2D5C94D2A6DF79AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicc.phy"},
	{0xE2D617E7CB53A06CULL, "vxdr_primevalentrance_1o0_0_0.vlm"},
	{0xE2D624F76178C714ULL, "zz_fen_debug6134392.cl"},
	{0xE2D62D8DA81C9755ULL, "gen_im_npc_arm_hum_boo_rdra.uti"},
	{0xE2D67A056B4EAD03ULL, "vxlt_foundry_n_xk_0_0.vlm"},
	{0xE2D6840D1B9A536CULL, "45450275.ogg"},
	{0xE2D6E39D68FF9A30ULL, "37821614.ogg"},
	{0xE2D7A829BFA6D8C6ULL, "hlac_lt_0_lowtown_docks_a3_-131073_4y.dds"},
	{0xE2D88CA6A9A25169ULL, "363755075.ogg"},
	{0xE2D8A90C8F6B9268ULL, "930427689.ogg"},
	{0xE2D8E482CA6F3AECULL, "dae000_bartrand.stg"},
	{0xE2D90FB0A3B57C59ULL, "dae380ar_deep_roads.lst"},
	{0xE2D9245B4A9F2646ULL, "vfx_distortionmask_p2.vsh.fxo.dx11"},
	{0xE2D9C024F0360804ULL, "vxdr_0_deeproads_dqe_0_0.vlm"},
	{0xE2DA56E903E43B3CULL, "vxw_woundedcoast_1jb_0_0.vlm"},
	{0xE2DAA5337642AD02ULL, "455188018.ogg"},
	{0xE2DAB989C8D2A026ULL, "dae011_corff_sta6149615.cl"},
	{0xE2DAC84851EAD778ULL, "194040037.ogg"},
	{0xE2DAD3B64D61EB78ULL, "vxht_keep_3fx_0_0.vlm"},
	{0xE2DAD4158A1D8BBCULL, "48693295.ogg"},
	{0xE2DB1EBBBF443C1EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\ddclosed.gad"},
	{0xE2DB85E70D2C562CULL, "gen00fl_sebastian6189076.cl"},
	{0xE2DB900B63E79C42ULL, "hlac_w_sundermount_-262144_0y.dds"},
	{0xE2DC1E76E0794CD6ULL, "6106289_m.fxe"},
	{0xE2DCE3F0B7D9E5B7ULL, "hlac_lt_hangedman_n_-1_1z.dds"},
	{0xE2DCFEF6F98782F4ULL, "hlac_w_blightlands_196606_0y.dds"},
	{0xE2DD112E28BC58BDULL, "6142202_m.fxe"},
	{0xE2DD7B94A1368F15ULL, "vxlt_gallowscourtyard_46s_0_0.vlm"},
	{0xE2DD7ED415D03EA0ULL, "6039810_f.fxe"},
	{0xE2DD920AF8C21B44ULL, "fen121cs_what_fenris_finds.cut"},
	{0xE2DD9C3C9232418AULL, "vxdr_0_deeproads_f5q_0_0.vlm"},
	{0xE2DDF5F0939D0C31ULL, "vxht_keep_63k_0_0.vlm"},
	{0xE2DDFBAF15F7E2E7ULL, "llac_lt_0_lowtown_a2_n_-65536_5.dds"},
	{0xE2DE3FE7F5BB0AE1ULL, "74765648.ogg"},
	{0xE2DEA0FB42A15FA1ULL, "6077825_m.fxe"},
	{0xE2DEF155F2257D68ULL, "vxca_caverns1_ics_0_2.vlm"},
	{0xE2DF2B519E46A6B9ULL, "6118292_m.fxe"},
	{0xE2DF50D2C0BD6EEEULL, "805394018.ogg"},
	{0xE2DF94EF6762B46BULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_hammer_sa.dds"},
	{0xE2DFA8BAC36455D1ULL, "vxlt_0_lowtown_docks_n_b75_0_0.vlm"},
	{0xE2E007B0A3BB8D30ULL, "dae380ar_deep_roads.ncs"},
	{0xE2E039B33740D35AULL, "6141187_m.fxe"},
	{0xE2E0630D45A994BFULL, "art\\levels\\races\\design\\_textures\\prp_bodybagcoins_n.dds"},
	{0xE2E0752C841CC6EEULL, "follower_banter6172247.cl"},
	{0xE2E1063FA327759FULL, "vxlt_gamlenshouse_n_ne_0_0.vlm"},
	{0xE2E1A8C0A5697ADFULL, "cn_hed_hlka_0t.dds"},
	{0xE2E1B4D4896B6EA0ULL, "gen00fl_merrill6127220.cl"},
	{0xE2E200F3B1C48267ULL, "follower_banter6143419.cl"},
	{0xE2E204D5170D78BBULL, "art\\vfx\\combat\\vfx_m_spi_walkbomb_conj_c.anb"},
	{0xE2E21BC4C778BB7FULL, "112760211.ogg"},
	{0xE2E243B9F4CEADE5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0dd_catacombs.phy"},
	{0xE2E28AC80FFD19FCULL, "party.mao"},
	{0xE2E2FFC406179E9BULL, "w_bonepitmining_a2_26_0.rml"},
	{0xE2E358E7F7F78966ULL, "vxht_0_hightown_a3_626_0_0.vlm"},
	{0xE2E35C4D98AA92DAULL, "447177211.ogg"},
	{0xE2E3CEA1244FCF46ULL, "vxht_brothel_n_4d3_0_0.vlm"},
	{0xE2E3D8CB76F6131EULL, "6177281_m.fxe"},
	{0xE2E4264B36C8D1B2ULL, "78575550.ogg"},
	{0xE2E44625ADA7AA11ULL, "ach_comp_gold.dds"},
	{0xE2E44C34AFB1DA02ULL, "vxca_caverns1_mines_i6u_0_0.vlm"},
	{0xE2E44DD268467CE2ULL, "hm_connor.mor"},
	{0xE2E59D699D10D661ULL, "6075702_m.fxe"},
	{0xE2E5DCD0FC8F349FULL, "one350_variel6085541.cl"},
	{0xE2E62F62486059C1ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_lp.ani"},
	{0xE2E640D710C966C0ULL, "vxca_caverns1_mines_e8i_0_0.vlm"},
	{0xE2E67C3B9C1D967EULL, "6088778_m.fxe"},
	{0xE2E69FE0B62F00F1ULL, "mag311_keran6041408.cl"},
	{0xE2E6EB5BCC53D214ULL, "6099459_m.fxe"},
	{0xE2E71E1D65DB7FF1ULL, "vxlt_0_lowtown_docks_a3_n_bf9_0_0.vlm"},
	{0xE2E80D9B065A9AA6ULL, "vxw_bonepitmining_a2_1of_0_0.vlm"},
	{0xE2E81416CEFDDDE8ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bucket.msh"},
	{0xE2E820B834F41D27ULL, "68195499.ogg"},
	{0xE2E82616CEFDFC79ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bucket.mao"},
	{0xE2E83216CEFE1012ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bucket.mmh"},
	{0xE2E843EC095E7903ULL, "vxlt_gallowstemplar_f_5wr_0_0.vlm"},
	{0xE2E8C0E577B88EE8ULL, "gen00fl_anders6125750.cl"},
	{0xE2E921A77FC81975ULL, "l_dae111ar_tavern_nt.bnk"},
	{0xE2E92FF746CD7C4EULL, "plt_car100pt_birthright.plo"},
	{0xE2E9313A6568ACBAULL, "vxlt_undercityencounter_1kf_0_0.vlm"},
	{0xE2E98E0C83C9F7B6ULL, "vis121_saemus6113530.cl"},
	{0xE2EA08B0A3C3E31EULL, "dae380ar_deep_roads.are"},
	{0xE2EA421E34E38F50ULL, "lowtown_male_lowlife1_sta6165041.cl"},
	{0xE2EB2CCC77DEEFE2ULL, "6135823_m.fxe"},
	{0xE2EC0C187689815DULL, "vxht_bartrand_n_332_0_0.vlm"},
	{0xE2EC9E0B5829C53FULL, "llac_w_bonepitmining_a2_393219_0.dds"},
	{0xE2ECD6E495E675E6ULL, "6065960_m.fxe"},
	{0xE2ED2C7375715D33ULL, "156716607.ogg"},
	{0xE2ED310C9539E6F1ULL, "231287004.ogg"},
	{0xE2ED3A6248668B63ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_lp.gad"},
	{0xE2ED48BA9616BBACULL, "hlac_ht_0_hightown_65537_1x.dds"},
	{0xE2EDF84D04BC00E6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv1.ani"},
	{0xE2EE2CDCDE2567ECULL, "585817747.ogg"},
	{0xE2EE4C3D369FDA1BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_humble_forward.ani"},
	{0xE2EE8AAECD525551ULL, "vxw_encounter2_p0_0_1.vlm"},
	{0xE2EE93ED3014E09DULL, "hlac_lt_undercitydungeon_-1_15z.dds"},
	{0xE2EF076422ADA09FULL, "6113035_m.fxe"},
	{0xE2EF4A2961FA18DBULL, "6082218_m.fxe"},
	{0xE2EF781C951D8E1FULL, "vxlt_gamlenshouse_n_ad_0_0.vlm"},
	{0xE2EF8D05CB195361ULL, "157760090.ogg"},
	{0xE2EF9E725AB1603AULL, "6076735_m.fxe"},
	{0xE2EFE9F4BFBB048FULL, "mag360_alain6044612.cl"},
	{0xE2F05B707A0460DCULL, "6005057_f.fxe"},
	{0xE2F07CC87860A498ULL, "gen00fl_merrill6161648.cl"},
	{0xE2F0D702BAFDB138ULL, "mag321_comtesse_de_launcet6064714.cl"},
	{0xE2F0E06A2C65CD9CULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuarylow.mao"},
	{0xE2F0EC6A2C65E23FULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuarylow.mmh"},
	{0xE2F0F26A2C65ECE1ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_ossuarylow.msh"},
	{0xE2F1730CC662C7B7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinr.mmh"},
	{0xE2F1770CC662CEC4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinr.mao"},
	{0xE2F1890CC662ED29ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ss_outinr.msh"},
	{0xE2F19877CB4E071BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhanga.mmh"},
	{0xE2F1A477CB4E1BB8ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhanga.mao"},
	{0xE2F1B677CB4E3A05ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_chainhanga.msh"},
	{0xE2F284F51E6DC8D4ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_1z.dds"},
	{0xE2F32F5D71F6432DULL, "6067666_m.fxe"},
	{0xE2F347681BA79D15ULL, "vxw_encounter1_dd_0_0.vlm"},
	{0xE2F366CFE9AF5E70ULL, "379599305.ogg"},
	{0xE2F370D60B35AB65ULL, "6077810_m.fxe"},
	{0xE2F380D6832C5C72ULL, "335577512.ogg"},
	{0xE2F4098EC190F660ULL, "409671098.ogg"},
	{0xE2F41762486C6E48ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_std_beg_lp.evt"},
	{0xE2F41F67BB800A4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_left.ani"},
	{0xE2F43123D81AB460ULL, "6177944_m.fxe"},
	{0xE2F45BE1FC9E16C2ULL, "6167608_m.fxe"},
	{0xE2F4705493A14102ULL, "635572381.ogg"},
	{0xE2F499EAF10526A4ULL, "dae000_meredith6085341.cl"},
	{0xE2F4DBC03E28248DULL, "6098201_f.fxe"},
	{0xE2F5633F2CAE4E1AULL, "vxlt_0_lowtown_docks_a3_n_d24_0_0.vlm"},
	{0xE2F5AE7E0CF8E6A4ULL, "mer130pt_when_and_where.ncs"},
	{0xE2F5C1CFB20E007EULL, "705293323.ogg"},
	{0xE2F5FC2BEA52F770ULL, "hlac_ca_caverns2prc_5802_13z.dds"},
	{0xE2F69866EAA39083ULL, "vxdr_0_deeproads_cfa_0_1.vlm"},
	{0xE2F6E531CBB0700DULL, "lt_0_lowtown_docks_a3_219_0.mmh"},
	{0xE2F6E931CBB0775EULL, "lt_0_lowtown_docks_a3_219_0.mao"},
	{0xE2F6F731CBB08E93ULL, "lt_0_lowtown_docks_a3_219_0.msh"},
	{0xE2F7706B23BAC190ULL, "6050092_m.fxe"},
	{0xE2F81FE8A07BDCB4ULL, "6110600_m.fxe"},
	{0xE2F85B89092E155AULL, "6164461_m.fxe"},
	{0xE2F87CD875C900F1ULL, "vxlt_0_lowtown_docks_a3_n_alp_0_0.vlm"},
	{0xE2F88B1DF0F4B300ULL, "hlac_lt_hangedman_n_-1_10y.dds"},
	{0xE2F8C0ADB48978D8ULL, "llac_w_bonepit_0_13.dds"},
	{0xE2F943A077760819ULL, "6032866_m.fxe"},
	{0xE2F94925F41AB430ULL, "64677857.ogg"},
	{0xE2F968440097B81FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_ric_0s.dds"},
	{0xE2FA1B3CAB120FF7ULL, "6152177_m.fxe"},
	{0xE2FA3AB8207EB930ULL, "mcr370_guardsman6087660.cl"},
	{0xE2FB1AFACA259EF8ULL, "6097844_m.fxe"},
	{0xE2FB543D36AA7212ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_humble_forward.evt"},
	{0xE2FB650FFE1954D3ULL, "328103761.ogg"},
	{0xE2FB91D45C33D88DULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue2_n.dds"},
	{0xE2FBE04D04C81557ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv1.evt"},
	{0xE2FC446E55AA3140ULL, "6094452_m.fxe"},
	{0xE2FC55444CDE5CA8ULL, "6117937_m.fxe"},
	{0xE2FCDEA9CE397357ULL, "qcr000_arishok6042474.cl"},
	{0xE2FD20F746DE0BDDULL, "plt_car100pt_birthright.nss"},
	{0xE2FD3CABD8BCE7FFULL, "21220532.ogg"},
	{0xE2FD70E3434929B7ULL, "dae000_arishok6072819.cl"},
	{0xE2FDEDE4638DFF1FULL, "lt_0_lowtown_docks_n_167_0.msh"},
	{0xE2FDF3E4638E09C1ULL, "lt_0_lowtown_docks_n_167_0.mmh"},
	{0xE2FDFFE4638E1E62ULL, "lt_0_lowtown_docks_n_167_0.mao"},
	{0xE2FE4E10EFE952FEULL, "primitives_primitive_p2.psh.fxo.dx11"},
	{0xE2FEE6BCC2546143ULL, "dae000_pc_home_ambients6208311.cl"},
	{0xE2FF1454CFD7984AULL, "603824522.ogg"},
	{0xE2FF17511128CD94ULL, "lt_0_lowtown_a3_n_c_37_0_ll.mmh"},
	{0xE2FF235D8C87A46BULL, "gen_im_arm_boo_lgt_act2_01.uti"},
	{0xE2FF74F8B00A16A0ULL, "vxw_bonepitdestroyed_a3_lh_0_0.vlm"},
	{0xE2FF7B8A5C3D5DFDULL, "llac_ca_cave_2194_11.dds"},
	{0xE2FFB2F405CF1577ULL, "zz_lgt_2_debug6140450.cl"},
	{0xE2FFE79025ADB2D9ULL, "vxca_caverns1_gj9_0_0.vlm"},
	{0xE3001F2DCCBC58AFULL, "vxlt_0_lowtown_a2_n_ex4_0_0.vlm"},
	{0xE30031C75ABED683ULL, "hlac_lt_gallowsprison_n_-1_10y.dds"},
	{0xE30040E8B6FCBAF3ULL, "vxht_0_hightown_57f_0_0.vlm"},
	{0xE301134C6A6A6ABAULL, "vxw_bonepitdestroyed_a3_iy_0_0.vlm"},
	{0xE301B281B5496A7DULL, "zz_ave_debug6056760.cl"},
	{0xE301B9450D9DB595ULL, "vxw_encounter1_ic_0_0.vlm"},
	{0xE3021767BB8C399FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_left.evt"},
	{0xE3023CDAF94B4171ULL, "6201534_m.fxe"},
	{0xE302733D36B0C5A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_humble_forward.gad"},
	{0xE302A9DCF80EBE71ULL, "6135204_m.fxe"},
	{0xE302E14D04CE35E0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\attackv1.gad"},
	{0xE303CF15B0400314ULL, "6093659_m.fxe"},
	{0xE303E6DF8152CBFCULL, "711558681.ogg"},
	{0xE3044A03C2C36BB0ULL, "cloth1.psh.fxo"},
	{0xE304675729E4EB0BULL, "40318207.ogg"},
	{0xE3052DCA7E20D903ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grindstone_sa.dds"},
	{0xE3057A15030E1F86ULL, "dr_primevalentrance_mmap_ph.dds"},
	{0xE3059AD9AA7C3279ULL, "follower_banter6170253.cl"},
	{0xE306100C539818F8ULL, "vxdr_primevaltomb_6p2_0_0.vlm"},
	{0xE306B112B125F9A1ULL, "161945450.ogg"},
	{0xE306EEB567BF1971ULL, "vxlt_0_lowtown_docks_a3_n_aqw_0_0.vlm"},
	{0xE307087DAAE07F60ULL, "vxlt_0_lowtown_n_f22_0_0.vlm"},
	{0xE3072A8593FCD43EULL, "mag100_cullen6140960.cl"},
	{0xE3072D15643C9572ULL, "vxht_mansion2b_3e7_0_0.vlm"},
	{0xE3073573FAB20BEAULL, "vxht_0_hightown_a3_cbk_0_0.vlm"},
	{0xE30838186F39FFC7ULL, "6107658_m.fxe"},
	{0xE308A9C70945FF50ULL, "hlac_ht_0_hightown_n_131072_0x.dds"},
	{0xE308DD7B995B0DB4ULL, "zz_dae_debug6029527.cl"},
	{0xE308F867BB922388ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\step_to_side_left.gad"},
	{0xE30940D19601AB1BULL, "hlac_w_bonepitdestroyed_a3_65537_0y.dds"},
	{0xE309B277BDCD20EAULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeblue.mao"},
	{0xE309B677BDCD2739ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeblue.mmh"},
	{0xE309C077BDCD3827ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeblue.msh"},
	{0xE309D5BA8FB203C9ULL, "6078081_f.fxe"},
	{0xE30A21D8E73516D9ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_gibbet.msh"},
	{0xE30A2BD8E73527C7ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_gibbet.mmh"},
	{0xE30A2FD8E7352E14ULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_gibbet.mao"},
	{0xE30A3F34F70B6735ULL, "dae000_merchant_generic_hm6172659.cl"},
	{0xE30B3A7C70F25F8EULL, "lt_0_lowtown_docks_a3_n_136_0.msh"},
	{0xE30B487C70F277DFULL, "lt_0_lowtown_docks_a3_n_136_0.mao"},
	{0xE30B547C70F28C7CULL, "lt_0_lowtown_docks_a3_n_136_0.mmh"},
	{0xE30B88DE80AFA8EDULL, "vxlt_backalley_g_1x4_0_0.vlm"},
	{0xE30BC11E8D371DA1ULL, "llac_ca_cave_2194_6.dds"},
	{0xE30BEC82CA9AA1D7ULL, "dae000_bartrand.bnk"},
	{0xE30C7B296225BC71ULL, "matdef_dw_normal.dds"},
	{0xE30D174E70341943ULL, "6180970_m.fxe"},
	{0xE30D33D2FAC2CFC7ULL, "lt_0_lowtown_docks_a3_5_0.rml"},
	{0xE30D448307A5CFB0ULL, "llac_lt_0_lowtown_a3_n_c_1_0.dds"},
	{0xE30D692D26D32614ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medb.phy"},
	{0xE30DB9C5E9A660B8ULL, "mag160_feynriel6037371.cl"},
	{0xE30E28DAE14095EAULL, "vxlt_gallowstemplar_f_533_0_0.vlm"},
	{0xE30E5959E4D64291ULL, "695204872.ogg"},
	{0xE30E863F78F8CEC5ULL, "vxdr_0_deeproads_ebt_0_0.vlm"},
	{0xE30ECE82CA9CA4EFULL, "dae000_bartrand.cnv"},
	{0xE30F635EEC1CA07CULL, "one130_thug6131592.cl"},
	{0xE3104771C8E3558AULL, "6112121_m.fxe"},
	{0xE31078021187FAE9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1_tool.phy"},
	{0xE31156535402801FULL, "m_qun100ar_exterior.bnk"},
	{0xE3118EF541ABF8DDULL, "art\\characters\\playercharacter\\humanfemale\\rob_appp.phy"},
	{0xE31190469FB27E75ULL, "art\\vfx\\models\\vfx_rockarmor_la.msh"},
	{0xE31192469FB281EBULL, "art\\vfx\\models\\vfx_rockarmor_la.mmh"},
	{0xE311AB79ABD1E7F0ULL, "gen00fl_fenris6128009.cl"},
	{0xE31368DFFC08246AULL, "mer210cr_coterie_thug_1.utc"},
	{0xE31380A4C8CB7596ULL, "116280145.ogg"},
	{0xE313ADAD8BBCD5D0ULL, "hlac_ca_caverns1_24425_13x.dds"},
	{0xE313DF3F89392FC9ULL, "946563820.ogg"},
	{0xE313ED61E32E1BE4ULL, "hlac_ca_caverns2_5802_14z.dds"},
	{0xE31481E9458E3178ULL, "vxdr_primevaltomb_6w3_0_0.vlm"},
	{0xE3148CE21EDB6A0DULL, "hlac_ht_chantry_2534_35z.dds"},
	{0xE31530A9A46F5D01ULL, "hlac_ca_caverns1_mines_24427_4z.dds"},
	{0xE3153BEF3EE7ED1EULL, "ser270_templar6112555.cl"},
	{0xE31547A77FED94FFULL, "l_dae111ar_tavern_nt.wbd"},
	{0xE3157A5A9CD697E0ULL, "vxlt_0_lowtown_n_f53_0_0.vlm"},
	{0xE3162784278C6E44ULL, "fx_ghost.mao"},
	{0xE316368DE2C5A69FULL, "vxw_bonepitmining_a2_1x6_0_0.vlm"},
	{0xE31668298F54611FULL, "804348954.ogg"},
	{0xE31724C081097E5CULL, "w_encounter2_camp_3_0.mao"},
	{0xE31730C0810992FFULL, "w_encounter2_camp_3_0.mmh"},
	{0xE31736C081099CA1ULL, "w_encounter2_camp_3_0.msh"},
	{0xE317462D26DB3F11ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medb.msh"},
	{0xE317542D26DB572CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medb.mao"},
	{0xE31754ECFDD8DAF1ULL, "vxht_bartrand_n_5w8_0_0.vlm"},
	{0xE317602D26DB6BCFULL, "art\\characters\\playercharacter\\humanfemale\\bdy_medb.mmh"},
	{0xE317DDE40E4A4501ULL, "6074997_f.fxe"},
	{0xE317F1D59726510DULL, "229913159.ogg"},
	{0xE318070B09787D21ULL, "mag110_samson6049735.cl"},
	{0xE318E940BD6A012EULL, "vxca_caverns2_83u_0_0.vlm"},
	{0xE31934E8B99AF84DULL, "6095026_f.fxe"},
	{0xE31951CCBB15773BULL, "vxw_bonepitmining_a2_26o_0_0.vlm"},
	{0xE319A7D196ED80D2ULL, "942383009.ogg"},
	{0xE319EA2A30136428ULL, "hlac_lt_gallowstemplar_f_2383_5z.dds"},
	{0xE31A22AB627BE43EULL, "tbe100im_artlathan_bow.uti"},
	{0xE31A785CDF5F846AULL, "vxlt_gallowstemplar_n_5r8_0_0.vlm"},
	{0xE31AC6BFD263A116ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_110a_0t.dds"},
	{0xE31B5A2101C9CB52ULL, "6053337_f.fxe"},
	{0xE31BFC950A6EFC92ULL, "6071690_f.fxe"},
	{0xE31C1EC02B09FB8FULL, "518498259.ogg"},
	{0xE31C46167FD4DE33ULL, "hlac_w_encounter1_0_2z.dds"},
	{0xE31C89C3A67C2A93ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_exit.evt"},
	{0xE31CAA0AEB0108F1ULL, "413378380.ogg"},
	{0xE31CAE0D123DEDB9ULL, "vxlt_0_lowtown_a2_n_gys_0_0.vlm"},
	{0xE31CB309F5192E1CULL, "206091510.ogg"},
	{0xE31D9614EC283146ULL, "vxht_brothel_n_4x7_0_0.vlm"},
	{0xE31E4D0A9DE70C83ULL, "vxht_chantry_2ta_0_1.vlm"},
	{0xE31E4F0C06580FA4ULL, "mag291tr_nt_torpor.ncs"},
	{0xE31E76485C2AD038ULL, "98767490.ogg"},
	{0xE31EABCAA44BCD44ULL, "llac_w_bonepitmining_a2_327682_0.dds"},
	{0xE31EB03D1742390FULL, "gen00fl_isabela6099528.cl"},
	{0xE31EE8A4C2AE69A8ULL, "6121922_f.fxe"},
	{0xE31F3B64FAF10846ULL, "6037599_f.fxe"},
	{0xE31FD34380A18D5FULL, "86282818.ogg"},
	{0xE320038DFADCC18CULL, "art\\characters\\playercharacter\\humanmale\\hlf_mash.msh"},
	{0xE320098DFADCCC7EULL, "art\\characters\\playercharacter\\humanmale\\hlf_mash.mmh"},
	{0xE320158DFADCE015ULL, "art\\characters\\playercharacter\\humanmale\\hlf_mash.mao"},
	{0xE3205736C9AA7CABULL, "6108609_f.fxe"},
	{0xE32173B6A0179856ULL, "vxlt_undercityencounter_17n_0_0.vlm"},
	{0xE3217A30CC925A64ULL, "and341ip_at_bonepit.utp"},
	{0xE321A1624F6A0577ULL, "6141744_m.fxe"},
	{0xE321B2F6F21FE299ULL, "649000576.ogg"},
	{0xE321F4FB4439C270ULL, "dre000_varric6034659.cl"},
	{0xE32282C3A6808AD4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_exit.gad"},
	{0xE32283F541BA5949ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appp.mao"},
	{0xE322898D215E8816ULL, "708684595.ogg"},
	{0xE3228FF541BA6DE2ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appp.mmh"},
	{0xE32291F541BA7098ULL, "art\\characters\\playercharacter\\humanfemale\\rob_appp.msh"},
	{0xE322CD7952D7B348ULL, "gen_im_arm_cht_hvy_act2_03.uti"},
	{0xE322F3C6378449F8ULL, "vxdr_primevaltomb_6z0_0_0.vlm"},
	{0xE32305EC4E3649FDULL, "vxca_cavern1_haunted_fm4_0_0.vlm"},
	{0xE32354FD17DAF768ULL, "gen00fl_fenris6115298.cl"},
	{0xE3237DAC2BF940F9ULL, "6135950_m.fxe"},
	{0xE3238AD855BD0FB9ULL, "vxht_mansion2b_1xe_0_0.vlm"},
	{0xE32402C4064F3B67ULL, "w_bonepitmining_a2_26_0.anb"},
	{0xE32422D04705E0A5ULL, "111572112.ogg"},
	{0xE3246EA9D841BCE7ULL, "vxht_0_hightown_a3_byn_0_0.vlm"},
	{0xE324767646E22CDBULL, "dae000_merchant_generic_hm6172590.cl"},
	{0xE324C5B5DAD6D765ULL, "vxw_encounter1_d4_0_0.vlm"},
	{0xE32598E17B7E0DB5ULL, "llac_w_woundedcoast_0_2.dds"},
	{0xE325A6F11DB60CC1ULL, "6086911_m.fxe"},
	{0xE32621B1D7E9BD4DULL, "root_war_wepshield.dds"},
	{0xE3266B46281A6391ULL, "zz_isaac_shot_one_v1.cut"},
	{0xE32702E58FE0155EULL, "161596527.ogg"},
	{0xE32819458D8C8121ULL, "6105396_f.fxe"},
	{0xE32836792397D8ADULL, "6125700_m.fxe"},
	{0xE3287AF5E2314D68ULL, "6098597_m.fxe"},
	{0xE328E5E78207E004ULL, "fx_greasepool.dds"},
	{0xE3294489021CF053ULL, "vxlt_0_lowtown_docks_n_c1d_0_0.vlm"},
	{0xE3295AC844B17720ULL, "dae000_bodhan_letter6144278.cl"},
	{0xE32965E483B3CFD2ULL, "6171603_m.fxe"},
	{0xE3297BDF880CFBC4ULL, "320803790.ogg"},
	{0xE329863F8C3680A6ULL, "qcr220_orsino6061343.cl"},
	{0xE32990551EE71E30ULL, "74854932.ogg"},
	{0xE32991C3A686C2B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_forward_1_exit.ani"},
	{0xE329BB028FEAB43FULL, "198769715.ogg"},
	{0xE329ED51D65D302FULL, "281065957.ogg"},
	{0xE32A28E62A46C099ULL, "vxlt_darktown_5wc_0_0.vlm"},
	{0xE32A4E7B25B4CF77ULL, "zz_otl_debug6060939.cl"},
	{0xE32AB41D8473E7A4ULL, "dae000_osric6162701.cl"},
	{0xE32B01D7B58F687CULL, "ach_oth_complete.dds"},
	{0xE32B59EE374D46DEULL, "6179922_m.fxe"},
	{0xE32B8853270DB92DULL, "gen00fl_anders6125645.cl"},
	{0xE32C37B483C28462ULL, "938004675.ogg"},
	{0xE32C48CA8A414254ULL, "genip_pp_hl_standing.utp"},
	{0xE32C4C0380816134ULL, "272012774.ogg"},
	{0xE32C65CF1C4FCEC8ULL, "vxlt_undercityserialkiller_1ir_0_0.vlm"},
	{0xE32D2311C1BF4384ULL, "vxlt_0_lowtown_a2_n_i4c_0_0.vlm"},
	{0xE32DA6DC4227557FULL, "vxlt_0_lowtown_a2_n_eht_0_0.vlm"},
	{0xE32DD2B05CEFB135ULL, "709136913.ogg"},
	{0xE32DFB237A4C4BEFULL, "vxdr_primevalentrance_15m_0_0.vlm"},
	{0xE32E2A05198764D9ULL, "vxca_caverns1_djf_0_0.vlm"},
	{0xE32E6C6118115932ULL, "vxlt_0_lowtown_a2_n_c_hok_0_0.vlm"},
	{0xE32E9DE366C1F0FBULL, "471435366.ogg"},
	{0xE32EE078F68939D7ULL, "6126973_f.fxe"},
	{0xE32FAA9CB15284C5ULL, "6027334_m.fxe"},
	{0xE32FC6C6A2D0873AULL, "6137568_m.fxe"},
	{0xE32FCC29325C453CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mc.mao"},
	{0xE32FD829325C59DFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mc.mmh"},
	{0xE32FDE29325C6381ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp16mc.msh"},
	{0xE3301C1CE9DB52AFULL, "396807651.ogg"},
	{0xE3301C70C6E411DCULL, "art\\vfx\\combat\\vfx_r_dwp_leftb_close1_c.impa.ani"},
	{0xE33046EBC0291526ULL, "vxw_bonepitmining_a2_1pk_0_0.vlm"},
	{0xE3311ED7775493C9ULL, "58790985.ogg"},
	{0xE3312BE3C9DC5E9EULL, "32986222.ogg"},
	{0xE331337C71132347ULL, "lt_0_lowtown_docks_a3_n_136_0.phy"},
	{0xE3314EF64CF13B61ULL, "vxlt_undercityserialkiller_1nv_0_0.vlm"},
	{0xE3315A84D37C4679ULL, "vxlt_gallowstemplar_n_45k_0_0.vlm"},
	{0xE331686497DE8A09ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh01.phy"},
	{0xE3319FCB1443A90BULL, "dae300_alistair_king6133553.cl"},
	{0xE331C24D2ABC4AC3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicc.msh"},
	{0xE331D04D2ABC621DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicc.mmh"},
	{0xE331D44D2ABC692EULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ff_basicc.mao"},
	{0xE331F11382557C3CULL, "hlac_ht_mansion2_818_37x.dds"},
	{0xE331FCB547B32839ULL, "vxht_mansion2b_1sd_0_0.vlm"},
	{0xE332222BCBBE75A3ULL, "llac_ca_cavern1_haunted_24426_7.dds"},
	{0xE3322B165C8F7A00ULL, "dre000_sandal6033771.cl"},
	{0xE3325BE87F2019CCULL, "397337127.ogg"},
	{0xE332654508295E31ULL, "728848383.ogg"},
	{0xE332C53EC3718ED2ULL, "vxlt_0_lowtown_docks_n_c1a_0_0.vlm"},
	{0xE332CCC4EB1B54ADULL, "vxw_encounter2_9z_0_0.vlm"},
	{0xE332DAE91721D8FFULL, "lla_ht_0_hightown_-2_0.dds"},
	{0xE3331DAA06556513ULL, "vxlt_gallowstemplar_n_2r6_0_0.vlm"},
	{0xE3333792CCCCEFE5ULL, "vxw_encounter1_i3_0_0.vlm"},
	{0xE333BE33231104A2ULL, "floatylayer.gfx"},
	{0xE333C02F4A008D38ULL, "hlac_ca_caverns1_24426_7z.dds"},
	{0xE333DBC700BBA5C4ULL, "sse_ss_livingarmor_1.gda"},
	{0xE33456B5581188E9ULL, "hlac_ca_cavern1_haunted_24426_9x.dds"},
	{0xE335702FC0CB4460ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.mmh"},
	{0xE33572374ED66AA9ULL, "cs_mcr370cs_attack_begins.bnk"},
	{0xE3362970C6E8931EULL, "art\\vfx\\combat\\vfx_r_dwp_leftb_close1_c.impa.gad"},
	{0xE33635869F01A83BULL, "vxw_bonepitmining_a2_28a_0_0.vlm"},
	{0xE336466DFE434928ULL, "follower_banter6136667.cl"},
	{0xE3364D8BBEB89CDBULL, "lt_0_lowtown_docks_n_208_0.mmh"},
	{0xE336598BBEB8B178ULL, "lt_0_lowtown_docks_n_208_0.mao"},
	{0xE336661C15C17B60ULL, "197007497.ogg"},
	{0xE3366B8BBEB8CFC5ULL, "lt_0_lowtown_docks_n_208_0.msh"},
	{0xE3368CDD7992D57EULL, "79059716.ogg"},
	{0xE3369E05E44F8CDDULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_noba_0n.dds"},
	{0xE33714D8E75BEEFCULL, "art\\levels\\races\\ferelden\\props\\torture\\prp_gibbet.phy"},
	{0xE337174235F5FD5FULL, "hlac_ht_keep_3733_0x.dds"},
	{0xE3379E972862E41BULL, "vxw_bonepitmining_a2_e6_0_0.vlm"},
	{0xE33811C42D394D87ULL, "68588695.ogg"},
	{0xE3387B33FB28775CULL, "w_encounter2_52_0.mao"},
	{0xE33886DEA285AB19ULL, "984388445.ogg"},
	{0xE3388733FB288BFFULL, "w_encounter2_52_0.mmh"},
	{0xE3388D33FB2895A1ULL, "w_encounter2_52_0.msh"},
	{0xE3399BE391AF2B82ULL, "vxlt_0_lowtown_docks_n_bv5_0_0.vlm"},
	{0xE339AD3D02281579ULL, "vxht_keep_6cc_0_0.vlm"},
	{0xE33A1260B5C01B42ULL, "vxlt_gallowstemplar_f_652_0_0.vlm"},
	{0xE33A269C64082EECULL, "vxw_bonepit_20x_0_0.vlm"},
	{0xE33AD7AC0E45E748ULL, "vxlt_undercityserialkiller_1ns_0_0.vlm"},
	{0xE33C18B9341D6DFFULL, "vxlt_0_lowtown_a2_n_emw_0_0.vlm"},
	{0xE33C2C35F1AFF2E2ULL, "422847520.ogg"},
	{0xE33C610211AD0E0EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1_tool.mmh"},
	{0xE33C6D0211AD22A5ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1_tool.mao"},
	{0xE33C7ACDBF0FEF3BULL, "zz_followers_banter6157272.cl"},
	{0xE33C7B0211AD3AFCULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1_tool.msh"},
	{0xE33D0470C6EE733DULL, "art\\vfx\\combat\\vfx_r_dwp_leftb_close1_c.impa.evt"},
	{0xE33D3C5354278F55ULL, "m_qun100ar_exterior.wbd"},
	{0xE33DAB22FCCBED92ULL, "dae000cr_grand_cleric6148152.cl"},
	{0xE33E50F4D0A72A67ULL, "ser221_party_comments6166210.cl"},
	{0xE33E594CF8432140ULL, "vxlt_hangedman_20m_0_0.vlm"},
	{0xE33EA710D7F70E73ULL, "6134313_m.fxe"},
	{0xE33EEEAA3F0A5479ULL, "vxht_0_hightown_n_bpi_0_0.vlm"},
	{0xE33EFAB3E7160BF3ULL, "280359587.ogg"},
	{0xE33F638CA3BC6C80ULL, "hlac_lt_darktown_-1_1y.dds"},
	{0xE3406E9239A940B9ULL, "vxht_mansion2b_1vk_0_0.vlm"},
	{0xE3406F3A0C0FB7D9ULL, "vxlt_0_lowtown_docks_a3_d15_0_0.vlm"},
	{0xE3407D18E2B8F066ULL, "117141371.ogg"},
	{0xE340B2C7D4B36BE2ULL, "vxw_bonepitmining_a2_248_0_0.vlm"},
	{0xE340E6725A8D0E2BULL, "hlac_ht_keep_n_3731_3x.dds"},
	{0xE340F11679C1EA6EULL, "mcr370_assignments6182068.cl"},
	{0xE341876EB7604FCAULL, "gen00fl_fenris6116587.cl"},
	{0xE341B379791D5AAEULL, "vxht_mansion_n_5m7_0_1.vlm"},
	{0xE3422B6E62022DF4ULL, "ht_0_hightown_n_28_0_lc.mmh"},
	{0xE3423F17D22FFC2CULL, "w_blightlands_223_0.msh"},
	{0xE3424517D230061EULL, "w_blightlands_223_0.mmh"},
	{0xE3425117D2301AB5ULL, "w_blightlands_223_0.mao"},
	{0xE342C46D670BB92FULL, "vxlt_0_lowtown_docks_n_bbp_0_0.vlm"},
	{0xE343532FC0D7507FULL, "art\\vfx\\combat\\vfx_r_gn2_accu_bl_loop_c.anb"},
	{0xE343C6ADDE14EFE7ULL, "vxlt_0_lowtown_a2_n_hgp_0_0.vlm"},
	{0xE343DB04183A5219ULL, "qcr220cr_noblewoman.utc"},
	{0xE3443116CF4C6D0DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bucket.phy"},
	{0xE344A15FF9D032BEULL, "vxlt_0_lowtown_gph_0_0.vlm"},
	{0xE34507AA10A215CEULL, "6093868_m.fxe"},
	{0xE3453B56D07F756AULL, "allabilities.ncs"},
	{0xE345B74C4A4F5D3DULL, "6149113_m.fxe"},
	{0xE34660ECAF14BFA0ULL, "mag110_samson6049817.cl"},
	{0xE346684D65F5B344ULL, "ser221_party_comments6111561.cl"},
	{0xE346AF788300B094ULL, "vxht_mansion_n_6jf_0_0.vlm"},
	{0xE346EBF06EAF156BULL, "6044623_m.fxe"},
	{0xE346F3DCF284469AULL, "315922321.ogg"},
	{0xE3474D16588A451DULL, "vxht_0_hightown_n_c5s_0_0.vlm"},
	{0xE34774AC6679121FULL, "gen00fl_fenris6128062.cl"},
	{0xE348140D6AD42E1CULL, "vxlt_darktown_2qc_0_0.vlm"},
	{0xE348AE46E3E01B1AULL, "6117759_m.fxe"},
	{0xE3495C02F85BE38FULL, "hlac_w_encounter2_65534_0z.dds"},
	{0xE34A37A2D2035D20ULL, "llac_dr_slavers_3214_6.dds"},
	{0xE34ABC31D2E1EAE0ULL, "vxlt_warehouse_n_10i_0_0.vlm"},
	{0xE34AF424B95334DBULL, "vxw_sundermount_226_0_0.vlm"},
	{0xE34B738D224C9ADBULL, "lt_0_lowtown_docks_212_0.msh"},
	{0xE34B858D224CB926ULL, "lt_0_lowtown_docks_212_0.mao"},
	{0xE34B918D224CCDC5ULL, "lt_0_lowtown_docks_212_0.mmh"},
	{0xE34BB2C6C3934CD1ULL, "681684462.ogg"},
	{0xE34BFFC406708868ULL, "w_bonepitmining_a2_26_0.mmh"},
	{0xE34C0BC406709C0BULL, "w_bonepitmining_a2_26_0.mao"},
	{0xE34C1DC40670BA92ULL, "w_bonepitmining_a2_26_0.msh"},
	{0xE34C50AE8ADE6EDCULL, "928749041.ogg"},
	{0xE34D067B5DE8A864ULL, "249369256.ogg"},
	{0xE34D869FC34E0497ULL, "vxw_bonepitdestroyed_a3_7r_0_0.vlm"},
	{0xE34DCE887CDECB42ULL, "6106173_m.fxe"},
	{0xE34E16F4B068CEB4ULL, "923917582.ogg"},
	{0xE34E394E5C1E75CAULL, "9589083.ogg"},
	{0xE34E641944320E97ULL, "mcr310_party_banter6200740.cl"},
	{0xE34ED51F12BC903AULL, "vxht_bartrand_n_5r0_0_0.vlm"},
	{0xE34F28E69B6C3270ULL, "follower_banter6166671.cl"},
	{0xE34F808E14FE4E22ULL, "vxw_blightlands_sx_0_0.vlm"},
	{0xE34F9C7454D1A57EULL, "gen00flst_bethany_carver2.stg"},
	{0xE34FCDDD98D5A57EULL, "llac_lt_0_lowtown_n_65537_4.dds"},
	{0xE3505244FF1488F1ULL, "mer210_smith6051396.cl"},
	{0xE3508166395696FDULL, "926948968.ogg"},
	{0xE350DEE74963C184ULL, "273910723.ogg"},
	{0xE3518F8C445978DCULL, "6164720_m.fxe"},
	{0xE3523001B1A4E844ULL, "vxlt_darktown_5at_0_0.vlm"},
	{0xE3523565743CE008ULL, "vxdr_0_deeproads_f4t_0_0.vlm"},
	{0xE352C58EE0A6661EULL, "fen110_fenris6118104.cl"},
	{0xE3536E5DA98EDAA9ULL, "hlac_w_blightlands_0_6x.dds"},
	{0xE353FCD230FEA7CFULL, "6149432_m.fxe"},
	{0xE35433DEB98F93E6ULL, "tbe100ip_free_weapons.utp"},
	{0xE3548C21A2E3B799ULL, "qun131_taskmaster6019206.cl"},
	{0xE354AAFEB73C8884ULL, "ave200_donnic6112132.cl"},
	{0xE355215574F6C914ULL, "vxht_mansion_n_6mg_0_0.vlm"},
	{0xE355BEF34A805D9DULL, "vxht_0_hightown_n_c2r_0_0.vlm"},
	{0xE35690F6E6144679ULL, "vxht_keep_6ia_0_0.vlm"},
	{0xE356EE6F15C9C342ULL, "347446930.ogg"},
	{0xE3571B41252A049FULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgtp.phy"},
	{0xE35766FAB90BC413ULL, "vxlt_gallowstemplar_4uj_0_0.vlm"},
	{0xE357A8B10BFF2362ULL, "6171904_m.fxe"},
	{0xE358794A30D83BE6ULL, "gen00fl_anders6104003.cl"},
	{0xE358A8B465328334ULL, "480173009.ogg"},
	{0xE358F40CE82DFCF3ULL, "vxlt_shop_87_0_0.vlm"},
	{0xE35958559374B6FEULL, "hlac_lt_gallowstemplar_n_2388_4z.dds"},
	{0xE3595893F24EE871ULL, "40600683.ogg"},
	{0xE359AF880FE69035ULL, "6130099_f.fxe"},
	{0xE35A110866C0B0DEULL, "gen00fl_fenris6105851.cl"},
	{0xE35B28D63E58B795ULL, "vxw_bonepitdestroyed_a3_21r_0_0.vlm"},
	{0xE35BCF6B8A021B9AULL, "mrl210_merrill6116400.cl"},
	{0xE35BD26422F68579ULL, "vxht_0_hightown_n_bfo_0_0.vlm"},
	{0xE35BEAA6FEF9C6A4ULL, "dae000ip_crowd_mf1_talk1.utp"},
	{0xE35BF728B5C038FCULL, "6169904_f.fxe"},
	{0xE35CA26FD81235C0ULL, "gen00fl_anders6113005.cl"},
	{0xE35D1F1B66C86105ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.mmh"},
	{0xE35D2A55CB849513ULL, "vxht_0_hightown_caf_0_0.vlm"},
	{0xE35D46FC04B2A8BAULL, "vxht_bartrand_n_5u1_0_0.vlm"},
	{0xE35D516498039E2EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh01.mmh"},
	{0xE35D6B649803CA1CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hh01.msh"},
	{0xE35D7D328DBC9A84ULL, "hlac_w_blightlands_-3_0y.dds"},
	{0xE35DAAABFBA8E778ULL, "vxlt_0_lowtown_n_f0p_0_0.vlm"},
	{0xE35DF26B06F466A2ULL, "vxw_blightlands_ty_0_0.vlm"},
	{0xE35ECF635C1E4599ULL, "lgt210_qun_blk_qunari6142325.cl"},
	{0xE35F14AB193626A2ULL, "vxw_sundermount_2gd_0_0.vlm"},
	{0xE35F1E149A1F56BAULL, "457368874.ogg"},
	{0xE35F265567F7975CULL, "combat_lmm_range.gda"},
	{0xE35F8E3DB7A75571ULL, "vxht_bartrand_n_5n3_0_0.vlm"},
	{0xE35FF98BF94250B3ULL, "278965921.ogg"},
	{0xE3609F95BDE03732ULL, "6116639_m.fxe"},
	{0xE360A1DEA39B00C4ULL, "vxlt_darktown_5ju_0_0.vlm"},
	{0xE360B46EA522DE57ULL, "vxlt_0_lowtown_docks_n_b0f_0_0.vlm"},
	{0xE360BED4ED373C58ULL, "117428919.ogg"},
	{0xE360C9161CC1873CULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.impa.ani"},
	{0xE360E08DFB141D19ULL, "art\\characters\\playercharacter\\humanmale\\hlf_mash.phy"},
	{0xE3612DFA7F256E69ULL, "llac_lt_0_lowtown_n_-131073_1.dds"},
	{0xE36139961EADB5D9ULL, "25848354.ogg"},
	{0xE361B05933F99855ULL, "6103128_m.fxe"},
	{0xE361B303C3927250ULL, "449286470.ogg"},
	{0xE361D594E6134090ULL, "ht_0_hightown_25_0_lc.mmh"},
	{0xE361DC20940CFA50ULL, "vxlt_0_lowtown_e2n_0_0.vlm"},
	{0xE363BBFD7DD56E72ULL, "355388477.ogg"},
	{0xE3640750F0376448ULL, "721511786.ogg"},
	{0xE36429DEADFC5A86ULL, "559846779.ogg"},
	{0xE364DE40687C513FULL, "vxw_bonepit_20w_0_0.vlm"},
	{0xE364E75DCC0C6839ULL, "vxlt_0_lowtown_a2_n_gbv_0_0.vlm"},
	{0xE365A70971250566ULL, "vrc221_varric6048383.cl"},
	{0xE365CA19EB81D2EEULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.dura.ani"},
	{0xE366A5DADCC00D72ULL, "235312147.ogg"},
	{0xE367BD3854EBACF1ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_3y.dds"},
	{0xE367D6161CC7BC7EULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.impa.gad"},
	{0xE3681C41253879A4ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgtp.mmh"},
	{0xE3682041253880F7ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgtp.mao"},
	{0xE368324125389F46ULL, "art\\characters\\playercharacter\\elffemale\\bdy_lgtp.msh"},
	{0xE3685823E2DC2ABAULL, "vxca_caverns1_gkp_0_1.vlm"},
	{0xE368D1B3479A263FULL, "lowtown_male_alienage1_sta6162788.cl"},
	{0xE36903688B7B684AULL, "vxht_mansion2b_335_0_0.vlm"},
	{0xE369411BBBA67A35ULL, "gen00fl_anders6119484.cl"},
	{0xE36A1A289A4A89ACULL, "vxlt_0_lowtown_a2_n_c_hxp_0_0.vlm"},
	{0xE36A1C1B66D2E6D2ULL, "art\\vfx\\cinematic\\vfx_mer_death_scorch_p.anb"},
	{0xE36A444114EC9DF9ULL, "vxht_0_hightown_n_ban_0_0.vlm"},
	{0xE36AAFA81F0CDF42ULL, "qun160_party_comment_16061234.cl"},
	{0xE36AB70E68C0BD05ULL, "6143010_m.fxe"},
	{0xE36AE89BED9BDF73ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_arrow_01_d.dds"},
	{0xE36B119982D8AA8CULL, "vxca_caverns1_flw_0_0.vlm"},
	{0xE36B65FB78CFC498ULL, "315265158.ogg"},
	{0xE36B9C32BD7AAD93ULL, "vxht_0_hightown_cdi_0_0.vlm"},
	{0xE36BB8D8F6A8C13AULL, "vxht_bartrand_n_5x2_0_0.vlm"},
	{0xE36C5256D2F37CCEULL, "pro000cr_colonel_ewald.utc"},
	{0xE36CED3E58C5C3A5ULL, "95545218.ogg"},
	{0xE36D3277C3BA3CF1ULL, "vxw_sundermount_2zm_0_1.vlm"},
	{0xE36D4F35CB139307ULL, "hlac_lt_0_lowtown_-131071_2x.dds"},
	{0xE36D63D308D1EF6FULL, "art\\characters\\creatures\\desiredemon\\animation\\d_strt.gad"},
	{0xE36E9A03243A062EULL, "vxlt_undercityserialkiller_yg_0_0.vlm"},
	{0xE36EB1161CCD9B9DULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose1_p.impa.evt"},
	{0xE36EDE5F5AB14FEBULL, "character_character_p112.psh.fxo.dx11"},
	{0xE36F13BB95911944ULL, "vxlt_darktown_5oz_0_0.vlm"},
	{0xE36F31C386D715F5ULL, "vrc321_varric6087089.cl"},
	{0xE36F5993976D933DULL, "ter_bigrock01_n.dds"},
	{0xE36F7233110FD256ULL, "isa311_hawke_comments6117906.cl"},
	{0xE370761DE0B5D20DULL, "6162802_m.fxe"},
	{0xE370A08DD9B63E71ULL, "710703580.ogg"},
	{0xE37133E34F250C66ULL, "hlac_ca_caverns1_mines_24424_17z.dds"},
	{0xE3714446598BA17AULL, "one300_zevran6076962.cl"},
	{0xE372500CA4D95B03ULL, "zz_approval_debug6133729.cl"},
	{0xE37268F6B712551EULL, "zz_approval_debug6003094.cl"},
	{0xE372C219EB8C4F3FULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.dura.evt"},
	{0xE37302F73EA8478EULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statueswordb.phy"},
	{0xE3730A23AED2DDB0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8ma.msh"},
	{0xE3731823AED2F5CAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8ma.mmh"},
	{0xE3731C23AED2FC01ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp8ma.mao"},
	{0xE37337A7B59F6215ULL, "genip_trap_switch.ncs"},
	{0xE37378CFCB5ED237ULL, "plt_dae020pt_amb_keep.nss"},
	{0xE373B44702E63528ULL, "hlac_ht_playershouse_n_855_3x.dds"},
	{0xE374396028DBAE4EULL, "vxca_caverns1_mines_i7v_0_0.vlm"},
	{0xE374522995E6ACEAULL, "287490352.ogg"},
	{0xE3745D2FD819AB9DULL, "vxht_brothel_n_59n_0_0.vlm"},
	{0xE37478D308D8311CULL, "art\\characters\\creatures\\desiredemon\\animation\\d_strt.evt"},
	{0xE374DFFE03FD143EULL, "vxht_chantry_n_30s_0_0.vlm"},
	{0xE375318EF488A586ULL, "vxlt_0_lowtown_a2_n_c_daq_0_0.vlm"},
	{0xE37566AF4B338CC6ULL, "367141552.ogg"},
	{0xE375B800A5AB52D3ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.dura.evt"},
	{0xE37663DD11811184ULL, "art\\characters\\creatures\\buzzard\\animation\\gliding.ani"},
	{0xE37684CCE5987492ULL, "vxht_0_hightown_n_c1d_0_0.vlm"},
	{0xE3768F07641288F6ULL, "6208479_m.fxe"},
	{0xE376DD20F555E9FDULL, "vxlt_0_lowtown_h8p_0_0.vlm"},
	{0xE3772E09F2FAE8E6ULL, "194700614.ogg"},
	{0xE37761FFDB44FA3FULL, "634946438.ogg"},
	{0xE37794713D285A42ULL, "hlac_lt_0_lowtown_a2_n_-65536_8z.dds"},
	{0xE3788C058C40A22CULL, "vxlt_0_lowtown_a2_n_c_hsq_0_0.vlm"},
	{0xE378FD5F3AAEE804ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_a.ani"},
	{0xE379025E87F9F080ULL, "vxht_0_hightown_a2_5w5_0_0.vlm"},
	{0xE3793F5F9930A650ULL, "arch_cloud.mao"},
	{0xE379884706BFB5E1ULL, "vxlt_undercityserialkiller_1wi_0_0.vlm"},
	{0xE379A319EB923928ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_aoe_p.dura.gad"},
	{0xE37A0B5DE397109DULL, "vxw_woundedcoast_n6_0_0.vlm"},
	{0xE37A39C0815E1F44ULL, "w_encounter2_camp_3_0.phy"},
	{0xE37A4B331A8DE250ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\hf_overlay.anb"},
	{0xE37BD2AB58437E53ULL, "830480936.ogg"},
	{0xE37BF559AD8C1632ULL, "vxlt_hangedman_n_21g_0_0.vlm"},
	{0xE37BF67DB91678DCULL, "vxlt_hangedman_ri_0_0.vlm"},
	{0xE37BF864FD2257A2ULL, "vxw_sundermount_2mf_0_0.vlm"},
	{0xE37C693B5796AFC4ULL, "1004755631.ogg"},
	{0xE37CB100A5B16514ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.dura.gad"},
	{0xE37D2F31B27354C5ULL, "qun131_party_comment6117266.cl"},
	{0xE37D468A01FC6213ULL, "follower_banter6141761.cl"},
	{0xE37D50374F13E6FBULL, "cs_mcr370cs_attack_begins.wbd"},
	{0xE37D60DD11872BE6ULL, "art\\characters\\creatures\\buzzard\\animation\\gliding.gad"},
	{0xE37E362207837EECULL, "gen00fl_aveline6114797.cl"},
	{0xE37EA87BDF1A4D84ULL, "6152695_m.fxe"},
	{0xE37FFA5F3AB50266ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\pose\\pose_a.gad"},
	{0xE380636F28527039ULL, "663410353.ogg"},
	{0xE380A205D81B5B0DULL, "plt_cod_crt_varterral.plo"},
	{0xE38130C081642ECCULL, "w_encounter2_camp_3_0.rml"},
	{0xE3817939DB9559BEULL, "6139214_m.fxe"},
	{0xE38180D308E2C935ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_strt.ani"},
	{0xE381CDDBC7E324DEULL, "vxw_bonepitdestroyed_a3_1xh_0_1.vlm"},
	{0xE38207C0EDA4073FULL, "6174906_m.fxe"},
	{0xE3821D1B2552F51DULL, "vxht_0_hightown_a2_c7f_0_0.vlm"},
	{0xE382491B94E7BB02ULL, "gen00fl_varric6206136.cl"},
	{0xE382776C716284D7ULL, "vxca_caverns2_957_0_0.vlm"},
	{0xE3828CDE070362BDULL, "llac_ht_mansion_n_3370_2.dds"},
	{0xE382BD12D1942AE2ULL, "llac_lt_0_lowtown_a2_131071_6.dds"},
	{0xE382CE91585FE33FULL, "6034215_m.fxe"},
	{0xE38345B06360CFB0ULL, "zz_getnearestinfo.ncs"},
	{0xE38358DFA14080D8ULL, "w_woundedcoast_116_0.phy"},
	{0xE3836BD1C923DCC9ULL, "hlac_ca_cavern1_haunted_24424_14z.dds"},
	{0xE383C000A5B79DF2ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.dura.ani"},
	{0xE38493615D6178C6ULL, "qun120cr_javaris6043987.cl"},
	{0xE3854298B19E79D0ULL, "6125428_m.fxe"},
	{0xE3859819CABB2B9BULL, "qun160_vashothlead6055567.cl"},
	{0xE385B14E75439E1FULL, "hlac_dr_slavers_3207_6y.dds"},
	{0xE385E28096BCE51CULL, "754569914.ogg"},
	{0xE3863D8CE73EA054ULL, "347173434.ogg"},
	{0xE3867316C4CC1135ULL, "hlac_lt_0_lowtown_a2_-262144_2x.dds"},
	{0xE386916E6DBB4D6BULL, "vxw_bonepitmining_a2_ib_0_0.vlm"},
	{0xE386CF1E063CB614ULL, "171792366.ogg"},
	{0xE386FAB627DE4DCAULL, "vxlt_gallowstemplar_6t6_0_0.vlm"},
	{0xE3874B27FE804387ULL, "vxlt_gallowscourtyard_5dz_0_0.vlm"},
	{0xE38792955890AA20ULL, "hlac_lt_0_lowtown_a2_n_c_65537_6z.dds"},
	{0xE387A2F4883866ABULL, "art\\vfx\\combat\\vfx_w_jug_tremor_aoe_p.impa.ani"},
	{0xE388F48F9B77AA77ULL, "ttx_arch_inner_b_2.mmh"},
	{0xE388F94E1AE2E4F3ULL, "6103225_m.fxe"},
	{0xE3890A8F9B77CFE9ULL, "ttx_arch_inner_b_2.msh"},
	{0xE3893177A39B9449ULL, "mcr310_argument6075834.cl"},
	{0xE38A080F891D7BF2ULL, "706334539.ogg"},
	{0xE38A2E49AB6D557FULL, "1060828990.ogg"},
	{0xE38A6A41EF187022ULL, "vxw_sundermount_2vg_0_0.vlm"},
	{0xE38A7C7A13098657ULL, "6034072_f.fxe"},
	{0xE38ACF7FB505ED22ULL, "gen00fl_sebastian6195817.cl"},
	{0xE38B2ED5316B0A7FULL, "vxht_mansion2b_3gh_0_0.vlm"},
	{0xE38B71C07C0AB318ULL, "6148977_m.fxe"},
	{0xE38C023D78109905ULL, "vxlt_0_lowtown_n_f0i_0_0.vlm"},
	{0xE38CECBB7E02CF4FULL, "dae000_meredith6085238.cl"},
	{0xE38D376C4697AF04ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_cas.msh"},
	{0xE38D496C4697CE6DULL, "art\\characters\\playercharacter\\humanfemale\\bdy_cas.mao"},
	{0xE38D4D6C4697D4A6ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_cas.mmh"},
	{0xE38DA0BA4CAA6795ULL, "zz_qcr_debug6055516.cl"},
	{0xE38DE66AACDCFE72ULL, "261925939.ogg"},
	{0xE38E4792F01E5B63ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.cess.ani"},
	{0xE38EF3B2008A1377ULL, "955080831.ogg"},
	{0xE38F1D67AB68B38EULL, "w_bonepit_116_0.msh"},
	{0xE38F2B67AB68CBDFULL, "w_bonepit_116_0.mao"},
	{0xE38F3767AB68E07CULL, "w_bonepit_116_0.mmh"},
	{0xE38F3F1F67A10882ULL, "6067306_m.fxe"},
	{0xE38F8A174C946D74ULL, "lgt310cr_shw_ors_informant.utc"},
	{0xE38FBAE4CA6C0616ULL, "pro000_captain_wright6004406.cl"},
	{0xE390160DFACFA48EULL, "vxlt_gallowstemplar_f_6qb_0_0.vlm"},
	{0xE3908D94EA318F8CULL, "6041094_m.fxe"},
	{0xE390B926ABBF86E3ULL, "vxlt_undercityencounter_13b_0_1.vlm"},
	{0xE391274D59386FFCULL, "514738307.ogg"},
	{0xE391406E4A55FBBFULL, "6031115_m.fxe"},
	{0xE3919982CA457325ULL, "lgt000cr_kirk_turn_in_1.utc"},
	{0xE39289A3D34CB531ULL, "hlac_w_woundedcoast_65539_0z.dds"},
	{0xE3929B87D63B08BEULL, "vxca_cavern1_haunted_j5g_0_0.vlm"},
	{0xE392C00D966C2865ULL, "gen_im_npc_elf_arm_cht_appd.uti"},
	{0xE3933F4E9BC0D3ACULL, "vxlt_warehouse_195_0_0.vlm"},
	{0xE393A7D71C96BA3EULL, "157114351.ogg"},
	{0xE394103982DBD5D5ULL, "642376948.ogg"},
	{0xE394596D92717DCFULL, "hlac_lt_blackemporium_1055_9x.dds"},
	{0xE39548990713675AULL, "genip_exploding_crate_sm_rd.utp"},
	{0xE3958AF488447B22ULL, "art\\vfx\\combat\\vfx_w_jug_tremor_aoe_p.impa.evt"},
	{0xE395E010E669BF41ULL, "llac_ht_0_hightown_1_1.dds"},
	{0xE395E3B990AB5E13ULL, "6182943_m.fxe"},
	{0xE396C33D2E4CBF25ULL, "vxht_chantry_354_0_0.vlm"},
	{0xE396FCB6FA9C9F06ULL, "1062148170.ogg"},
	{0xE39726E4DEDC14F5ULL, "vxlt_gallowscourtyard_6zu_0_0.vlm"},
	{0xE397446C46A01A71ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_cas.phy"},
	{0xE39790E7A7BA547EULL, "430218673.ogg"},
	{0xE397E592C549CC56ULL, "lt_ruinedhouse_n_3567_lc.mmh"},
	{0xE397FCBB7AA9EB8DULL, "575923178.ogg"},
	{0xE3983AEF6C80035EULL, "vxw_woundedcoast_6i_0_0.vlm"},
	{0xE39847F0B4BA3B4EULL, "115435944.ogg"},
	{0xE398A6BF608C3CECULL, "vxlt_backalley_n_1l4_0_0.vlm"},
	{0xE3994CDFBAFD7271ULL, "6043912_m.fxe"},
	{0xE3999BC39C9569E9ULL, "dae000_merchant_tailor_hm6164353.cl"},
	{0xE39A2877BF3CE1F6ULL, "lt_0_lowtown_docks_n_3_0_lc.mmh"},
	{0xE39A3E67DD79E06FULL, "153814898.ogg"},
	{0xE39AA011E5F1C7C9ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_eye_blu_0d.dds"},
	{0xE39ACD2AC96D9A33ULL, "lla_ht_0_hightown_65538_0.dds"},
	{0xE39B12189B69EEDDULL, "496491781.ogg"},
	{0xE39B46F07A2AF320ULL, "6167624_m.fxe"},
	{0xE39B5F92F0290E5AULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.cess.evt"},
	{0xE39B6A3E0CF65728ULL, "6140810_m.fxe"},
	{0xE39B9033FB7D1844ULL, "w_encounter2_52_0.phy"},
	{0xE39BAD8B58F0B965ULL, "208868063.ogg"},
	{0xE39BD3F81D519E6EULL, "dae200_bran6094731.cl"},
	{0xE39BF0ECCC9BD366ULL, "llac_w_sundermount_-393212_0.dds"},
	{0xE39C5BBC4D9C3F5AULL, "zz_delete_after_press_demo.cut"},
	{0xE39C65F7B1D2EEE5ULL, "zz_and_debug6134140.cl"},
	{0xE39C6EBF742B864CULL, "6000681_f.fxe"},
	{0xE39C7C3C303F8E46ULL, "class_abilities_bethany.gda"},
	{0xE39C89F4884A9835ULL, "art\\vfx\\combat\\vfx_w_jug_tremor_aoe_p.impa.gad"},
	{0xE39CE515D48F9C47ULL, "gen00fl_isabela6085763.cl"},
	{0xE39D01B9C03451F2ULL, "704487620.ogg"},
	{0xE39E3B80C0F99065ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernlava_d.dds"},
	{0xE39E7646FD93C95FULL, "vxlt_undercityencounter_1r0_0_0.vlm"},
	{0xE39E93D68DCAC73AULL, "follower_banter6167223.cl"},
	{0xE39EA90D4ED9EEEEULL, "llac_lt_0_lowtown_-131072_0.dds"},
	{0xE39EEAF410FE2315ULL, "784881441.ogg"},
	{0xE39F277EC68F7332ULL, "hm_kerod.mor"},
	{0xE39FED2F2DA4A600ULL, "517788480.ogg"},
	{0xE3A0199A7FB09E8FULL, "6078111_f.fxe"},
	{0xE3A06625D14C531AULL, "hlac_ca_caverns1_24421_12y.dds"},
	{0xE3A0B8C24F61A3B7ULL, "vxlt_gallowstemplar_f_3x5_0_0.vlm"},
	{0xE3A14B46BEC67E9BULL, "6129221_m.fxe"},
	{0xE3A16E57383EDA57ULL, "lowtown_male_lowlife1_sta6165035.cl"},
	{0xE3A1E347C001D8FBULL, "6094002_m.fxe"},
	{0xE3A1E9C95F068D83ULL, "vxlt_gallowstemplar_n_5ru_0_0.vlm"},
	{0xE3A25E92F02F2B4DULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.cess.gad"},
	{0xE3A28733FB8327CCULL, "w_encounter2_52_0.rml"},
	{0xE3A2CEC164DDE120ULL, "6088871_m.fxe"},
	{0xE3A324214812E148ULL, "zz_beggar_test6105727.cl"},
	{0xE3A336ED88566276ULL, "6044284_m.fxe"},
	{0xE3A3386A26DACF87ULL, "vxlt_warehouse_sa_0_0.vlm"},
	{0xE3A3999297AE2DB4ULL, "follower_banter6154021.cl"},
	{0xE3A3A8E0FC02460CULL, "36803040.ogg"},
	{0xE3A3E1A9A73DB95DULL, "gen00fl_aveline6104953.cl"},
	{0xE3A47DAFE0D24969ULL, "954741250.ogg"},
	{0xE3A5DD770273A732ULL, "vxht_bartrand_n_4kz_0_0.vlm"},
	{0xE3A5FC27BDCF7EAAULL, "vxlt_0_lowtown_a2_h21_0_0.vlm"},
	{0xE3A6ACCC5E761BDEULL, "vxw_woundedcoast_3j_0_0.vlm"},
	{0xE3A6DD70852C4887ULL, "mag311_emile_de_launcet6043806.cl"},
	{0xE3A7189C5282556CULL, "vxlt_backalley_n_1w5_0_0.vlm"},
	{0xE3A72636291BAD09ULL, "6206885_m.fxe"},
	{0xE3A7813126B6F8CBULL, "vxw_sundermount_356_0_0.vlm"},
	{0xE3A8148FC3146D71ULL, "vxlt_0_lowtown_a2_n_gyk_0_0.vlm"},
	{0xE3A841D80BA4E77AULL, "framebuffereffects_radialblur_p0.vsh.fxo"},
	{0xE3A892C727C6E4E0ULL, "w_encounter1_35_0.phy"},
	{0xE3A8C7394000C56AULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys.dura.gad"},
	{0xE3A912858752BE91ULL, "hlac_ht_keep_4519_7z.dds"},
	{0xE3A9D294194F50ECULL, "vxw_blightlands_11m_0_0.vlm"},
	{0xE3AAB164D6E5FCC2ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillarb.msh"},
	{0xE3AAB364D6E60078ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\prp_ivypillarb.mmh"},
	{0xE3AAEB43EDAE4784ULL, "vxca_caverns1_mines_e8m_0_0.vlm"},
	{0xE3AB210E24CFD047ULL, "2060995.ogg"},
	{0xE3AB7A5BDBD11193ULL, "dae000_bartrand6030627.cl"},
	{0xE3AB8CCAA454F671ULL, "130584824.ogg"},
	{0xE3ABB267D1313FA8ULL, "llac_w_sundermount_327683_0.dds"},
	{0xE3AC28940EFD35E4ULL, "vxlt_darktown_4i9_0_0.vlm"},
	{0xE3AC449090C4F5A1ULL, "dae000_party_night_comment6201620.cl"},
	{0xE3AC6FDFE3AF92FCULL, "vxw_bonepitdestroyed_a3_p0_0_0.vlm"},
	{0xE3AC7E01DC530AD8ULL, "hla_ht_0_hightown_n_65538_0z.dds"},
	{0xE3AC95C289622197ULL, "vxca_caverns1_gk0_0_0.vlm"},
	{0xE3AD7803782A766DULL, "vxht_mansion2b_2pj_0_0.vlm"},
	{0xE3AEB57618C69668ULL, "vxlt_0_lowtown_n_fh8_0_0.vlm"},
	{0xE3AFAA394006B201ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys.dura.evt"},
	{0xE3B017A8FDBE0C5CULL, "vxlt_gallowstemplar_n_2xs_0_0.vlm"},
	{0xE3B05BA650FCA603ULL, "vxlt_gallowstemplar_n_5ut_0_0.vlm"},
	{0xE3B0F7C87EAF99B7ULL, "hlac_lt_gallowstemplar_n_2384_7z.dds"},
	{0xE3B0FB74424BA2F2ULL, "llac_lt_gallowstemplar_n_2394_10.dds"},
	{0xE3B162C48FE043DDULL, "737354130.ogg"},
	{0xE3B16B602572F496ULL, "follower_banter6173158.cl"},
	{0xE3B1EA453CEAD7B4ULL, "vxht_0_hightown_a2_bwb_0_0.vlm"},
	{0xE3B21292DA3D9835ULL, "6148107_m.fxe"},
	{0xE3B21F63284E1BABULL, "882825040.ogg"},
	{0xE3B29C13965A855AULL, "zz_varric_arrows_v1.cut"},
	{0xE3B2CB6E03BF9069ULL, "6136670_m.fxe"},
	{0xE3B2F5AE75623E7DULL, "vxht_mansion_n_4zx_0_1.vlm"},
	{0xE3B340BA06B0EF5DULL, "6050295_m.fxe"},
	{0xE3B39EF63F874481ULL, "135008444.ogg"},
	{0xE3B401F73EDFDD51ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statueswordb.mmh"},
	{0xE3B40DF73EDFF1F2ULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statueswordb.mao"},
	{0xE3B41BF73EE0090FULL, "art\\levels\\races\\ferelden\\props\\chantry\\prp_statueswordb.msh"},
	{0xE3B4228F4B9ADFDEULL, "304021228.ogg"},
	{0xE3B4662ACB90A434ULL, "vxlt_gallowstemplar_2om_0_0.vlm"},
	{0xE3B46C518F3DF49CULL, "6117766_m.fxe"},
	{0xE3B4ABC06EEE29EDULL, "6000917_m.fxe"},
	{0xE3B5B32F3BC92FA5ULL, "6129818_m.fxe"},
	{0xE3B630BCCB4CDCBEULL, "365727075.ogg"},
	{0xE3B63D6EED9B7775ULL, "hla_ht_0_hightown_n_-131071_0z.dds"},
	{0xE3B64B1B78B6451CULL, "317556995.ogg"},
	{0xE3B6A1E2C1087BB1ULL, "vxht_keep_6fd_0_0.vlm"},
	{0xE3B73BFEF87EE204ULL, "mcr370_meredith6085941.cl"},
	{0xE3B7535A1A4E4886ULL, "6034474_m.fxe"},
	{0xE3B79ECEBDC6770CULL, "472108813.ogg"},
	{0xE3B7E8C7CB724E00ULL, "zz_isaachawke1.utc"},
	{0xE3B88B56BCAB079CULL, "t3_hey_bl2.tnt"},
	{0xE3B8A1A0C6C3749EULL, "ser121_jethann6104613.cl"},
	{0xE3B8AF992B5DD4B5ULL, "140589673.ogg"},
	{0xE3BB352EF22CCD5AULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb2.msh"},
	{0xE3BB472EF22CEBA0ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_fereldenb2.mmh"},
	{0xE3BB58FD62C8BA21ULL, "422712515.ogg"},
	{0xE3BB7A0B15C9C354ULL, "144492455.ogg"},
	{0xE3BB9FFF65D484FEULL, "vxw_encounter2_5q_0_0.vlm"},
	{0xE3BBEC4C5171371EULL, "gen00fl_aveline6139930.cl"},
	{0xE3BBF69266C67716ULL, "zz_dae_debug6047635.cl"},
	{0xE3BC46F3EBEE42C0ULL, "vxlt_foundry_n_11x_0_0.vlm"},
	{0xE3BC55A6706F34D4ULL, "mag211_marethari6025800.cl"},
	{0xE3BCC1078452C531ULL, "vxlt_gallowstemplar_n_45s_0_0.vlm"},
	{0xE3BCE795DC7AC5A3ULL, "6099231_f.fxe"},
	{0xE3BCF291217A71B6ULL, "llac_ht_mansion_n_3374_1.dds"},
	{0xE3BD1DED25463FEFULL, "6155679_m.fxe"},
	{0xE3BD21E855CA3FD0ULL, "campaign_base_en-us.cre"},
	{0xE3BDA2394012E250ULL, "art\\vfx\\environment\\level_fx\\bugs\\vfx_flys.dura.ani"},
	{0xE3BE4757E3483013ULL, "vxlt_0_lowtown_n_ev6_0_0.vlm"},
	{0xE3BE5BCFCB9E8C18ULL, "plt_dae020pt_amb_keep.plo"},
	{0xE3BE627B799F6F68ULL, "vxdr_0_deeproads_9g6_0_0.vlm"},
	{0xE3BE99FA5DD1EB63ULL, "vxlt_0_lowtown_docks_a3_n_c1k_0_0.vlm"},
	{0xE3BECFCC0290551AULL, "6033321_m.fxe"},
	{0xE3BEE0581DAEE26CULL, "llac_lt_gallowstemplar_n_2388_2.dds"},
	{0xE3BF2DB14B2F04C6ULL, "pro000cs_voyage.cut"},
	{0xE3BF37ABE8F9BEABULL, "dae000_quintus6154702.cl"},
	{0xE3BF65342029587DULL, "zz_isa_debug6133147.cl"},
	{0xE3C01C240AC70087ULL, "vxlt_warehouse_yc_0_0.vlm"},
	{0xE3C05C222EE0F034ULL, "vxht_0_hightown_a2_bxc_0_0.vlm"},
	{0xE3C1121DF834B0B6ULL, "6202218_f.fxe"},
	{0xE3C13E9808362022ULL, "hm_thug01.mor"},
	{0xE3C1571462DB9C22ULL, "l_mag131ar_bloodmages.wbd"},
	{0xE3C1B27F53E0890FULL, "hlac_lt_0_lowtown_a3_n_c_-1_8z.dds"},
	{0xE3C1D965360826F8ULL, "6019187_m.fxe"},
	{0xE3C223FC435F9AB2ULL, "35120210.ogg"},
	{0xE3C2472294F6248AULL, "llac_lt_0_lowtown_a2_n_c_1_10.dds"},
	{0xE3C275ADDF4B6FA9ULL, "370025446.ogg"},
	{0xE3C2BDEA422F4491ULL, "vxlt_0_lowtown_a2_n_hex_0_0.vlm"},
	{0xE3C2D76EE5F63B88ULL, "vxdr_0_deeproads_ck4_0_0.vlm"},
	{0xE3C315BED249D88AULL, "tables_d.dds"},
	{0xE3C31CE545D60C38ULL, "darktown_male_fereldan_sta6175134.cl"},
	{0xE3C33E5DECA53DC0ULL, "205318469.ogg"},
	{0xE3C3668A8EA06DCDULL, "hlac_w_encounter2_2_0y.dds"},
	{0xE3C379BE34A9D2DDULL, "6026217_f.fxe"},
	{0xE3C3EF81C6760B86ULL, "819562726.ogg"},
	{0xE3C3F99C871AF35BULL, "qcr000_arishok6060203.cl"},
	{0xE3C401738D490146ULL, "hlac_ca_caverns2_5802_14x.dds"},
	{0xE3C43CBB4E88820BULL, "hlac_ca_caverns1_mines_24427_4x.dds"},
	{0xE3C44FF9E6BF07D0ULL, "6198693_m.fxe"},
	{0xE3C459B7BB849944ULL, "707420460.ogg"},
	{0xE3C4A39FE0FDA86AULL, "zz_seb_debug6189577.cl"},
	{0xE3C4A8F3C8F65DC7ULL, "hlac_ht_chantry_2534_35x.dds"},
	{0xE3C4B423D2BF45E5ULL, "35198108.ogg"},
	{0xE3C4FF4454C90CC4ULL, "hlac_lt_0_lowtown_a2_65535_3x.dds"},
	{0xE3C5A0395202F716ULL, "lt_0_lowtown_docks_107_0.msh"},
	{0xE3C5AA3952030814ULL, "lt_0_lowtown_docks_107_0.mmh"},
	{0xE3C5AE3952030F27ULL, "lt_0_lowtown_docks_107_0.mao"},
	{0xE3C5CDC3385C07C5ULL, "art\\vfx\\creatures\\fxc_drg_brth_s_c.anb"},
	{0xE3C68108FB8C2434ULL, "mrl310_party_comments6143072.cl"},
	{0xE3C6C489DE7AFA29ULL, "vxht_0_hightown_a3_c51_0_0.vlm"},
	{0xE3C710CC289FA2FBULL, "gen00fl_bethany6005734.cl"},
	{0xE3C8444D33587106ULL, "836801811.ogg"},
	{0xE3C8A5AD08F7ECCDULL, "vxht_keep_n_8ap_0_0.vlm"},
	{0xE3C8B243011FA9B1ULL, "700041768.ogg"},
	{0xE3C8BE9ECC377643ULL, "583267822.ogg"},
	{0xE3C90608796FA029ULL, "w_bonepitdestroyed_a3_103_0.phy"},
	{0xE3C965D1BB063250ULL, "w_bonepit_34_0.rml"},
	{0xE3C978C0962ACD2AULL, "hlac_lt_0_lowtown_a3_n_c_-65537_0z.dds"},
	{0xE3C98DBAE71BE10CULL, "art\\characters\\creatures\\dragon\\animation\\ddie_lp.ani"},
	{0xE3C98DD3378B13BAULL, "vxht_mansion2b_1mq_0_0.vlm"},
	{0xE3CA552CEFE0D331ULL, "vxht_0_hightown_n_bpq_0_0.vlm"},
	{0xE3CA67608F84D8CDULL, "follower_banter6163424.cl"},
	{0xE3CA9EEA7F9F3CF3ULL, "454452524.ogg"},
	{0xE3CB1C8775309ADEULL, "vxca_caverns1_e6t_0_0.vlm"},
	{0xE3CB575CA8DCEF93ULL, "6118496_m.fxe"},
	{0xE3CB86FA2EA03AAAULL, "w_woundedcoast_240_0.mao"},
	{0xE3CB8AFA2EA041F9ULL, "w_woundedcoast_240_0.mmh"},
	{0xE3CB91969A52C2B4ULL, "vxca_caverns2_8p2_0_0.vlm"},
	{0xE3CB94FA2EA052E7ULL, "w_woundedcoast_240_0.msh"},
	{0xE3CBD514EA7FC071ULL, "vxht_mansion2b_1vs_0_0.vlm"},
	{0xE3CC0A6F9810BDB3ULL, "vxlt_blackemporium_138_0_0.vlm"},
	{0xE3CC9CFDDD938248ULL, "zz_slate1_comp.cut"},
	{0xE3CCB934D53E4893ULL, "vxlt_0_lowtown_n_es9_0_0.vlm"},
	{0xE3CD3F6034E8D703ULL, "vxlt_gallowstemplar_n_5cv_0_0.vlm"},
	{0xE3CDA858B54D8688ULL, "vxht_chantry_2c6_0_0.vlm"},
	{0xE3CDC9ED8255D109ULL, "follower_banter6160605.cl"},
	{0xE3CE31017B66B70DULL, "194332326.ogg"},
	{0xE3CEAFC727E7E575ULL, "w_encounter1_35_0.msh"},
	{0xE3CEB1C727E7E8EBULL, "w_encounter1_35_0.mmh"},
	{0xE3CEBB7D123E5B11ULL, "vxht_0_hightown_a3_blw_0_0.vlm"},
	{0xE3CEBDC727E7FC88ULL, "w_encounter1_35_0.mao"},
	{0xE3CF22FFC8F8AEEDULL, "6034815_m.fxe"},
	{0xE3CFA8B99E6BD53EULL, "132468617.ogg"},
	{0xE3CFB0E3467DE007ULL, "552809336.ogg"},
	{0xE3D086A5BE4E635EULL, "mag311_nyssa6050130.cl"},
	{0xE3D09ABAE722164EULL, "art\\characters\\creatures\\dragon\\animation\\ddie_lp.gad"},
	{0xE3D09EE3C4AE80B1ULL, "hlac_w_woundedcoast_-196603_0z.dds"},
	{0xE3D0CE285D994FC5ULL, "337657608.ogg"},
	{0xE3D10F61E24B03EBULL, "ts_dwarven_wpn_table.gda"},
	{0xE3D12DF1881C9B86ULL, "vxdr_0_deeproads_cue_0_1.vlm"},
	{0xE3D16E0537AF33E7ULL, "vxlt_hangedman_30o_0_0.vlm"},
	{0xE3D1D01D2D87BBCDULL, "vxlt_undercitydungeon_1n7_0_0.vlm"},
	{0xE3D24291CFB0CF16ULL, "al_rogue_t02.gda"},
	{0xE3D27155D56514FDULL, "vxht_chantry_n_2ir_0_0.vlm"},
	{0xE3D2E7742B2BEED6ULL, "llac_lt_0_lowtown_a2_-2_4.dds"},
	{0xE3D3404B05B22FEEULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbma.msh"},
	{0xE3D35A4B05B25BDCULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbma.mmh"},
	{0xE3D36A2698F6B6F1ULL, "vxlt_0_lowtown_a2_n_ghh_0_0.vlm"},
	{0xE3D36B5E4BA8F5BFULL, "art\\characters\\playercharacter\\humanfemale\\glv_bda.phy"},
	{0xE3D36C56EF4501EFULL, "vxlt_warehouse_n_1e5_0_0.vlm"},
	{0xE3D3B3394013C4AAULL, "6005809_m.fxe"},
	{0xE3D3FD3F0380D8BDULL, "108655930.ogg"},
	{0xE3D43A752DCE7E00ULL, "zz_followers_banter6157398.cl"},
	{0xE3D48A381DC3733FULL, "vxw_blightlands_11b_0_0.vlm"},
	{0xE3D54755A01C7E53ULL, "lt_0_lowtown_docks_a3_n_16_0_lc.mmh"},
	{0xE3D57C3334DABC06ULL, "vxw_encounter2_3c_0_0.vlm"},
	{0xE3D591FE64D53B83ULL, "20978546.ogg"},
	{0xE3D5C54F87887F02ULL, "434605235.ogg"},
	{0xE3D68D22EA2B5FA2ULL, "fen310_fenris6118205.cl"},
	{0xE3D701D9C5E867BBULL, "llac_w_woundedcoast_327682_0.dds"},
	{0xE3D71C33DAD8E4E0ULL, "genip_node_b1.utp"},
	{0xE3D7394D78F4B90FULL, "mag330_walter6065030.cl"},
	{0xE3D73DF630AB27ECULL, "follower_banter6136357.cl"},
	{0xE3D78FAD561794C1ULL, "6034922_f.fxe"},
	{0xE3D7D70353B18337ULL, "plt_cod_ltr_danzig_wayward.nss"},
	{0xE3D80F1FF36BEC59ULL, "gen00fl_aveline6114848.cl"},
	{0xE3D8862E3B8B305AULL, "one110_pryce6088806.cl"},
	{0xE3D8D7157404EA7EULL, "vxw_blightlands_gx_0_0.vlm"},
	{0xE3D8F797A35C9453ULL, "6202115_m.fxe"},
	{0xE3D926607C66867AULL, "293748610.ogg"},
	{0xE3D9BAB74284D8CCULL, "304386479.ogg"},
	{0xE3D9C431C5BF06C4ULL, "flesh.dds"},
	{0xE3DA05DECAF3ED9BULL, "hlac_lt_0_lowtown_-131072_6y.dds"},
	{0xE3DA0EC5813B1925ULL, "398161326.ogg"},
	{0xE3DAF2F1CC936C29ULL, "vxdr_0_deeproads_dq4_0_0.vlm"},
	{0xE3DC109ADCAFBEF6ULL, "dae000_oldmangossip6175752.cl"},
	{0xE3DC7C097CEC4439ULL, "isa200pt_night_terrors.ncs"},
	{0xE3DCD13146EC0016ULL, "6118088_m.fxe"},
	{0xE3DD44E0E8E5E022ULL, "153558687.ogg"},
	{0xE3DD8CBEF5E4A309ULL, "124482721.ogg"},
	{0xE3DED03C842E1973ULL, "6013252_m.fxe"},
	{0xE3DF45DFA18EBF6DULL, "w_woundedcoast_116_0.msh"},
	{0xE3DF53DFA18ED6A0ULL, "w_woundedcoast_116_0.mao"},
	{0xE3DF57DFA18EDDF3ULL, "w_woundedcoast_116_0.mmh"},
	{0xE3DF6499D221F32CULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgte.phy"},
	{0xE3DFE166281D7B6AULL, "vxht_0_hightown_a3_c9z_0_0.vlm"},
	{0xE3DFE5CAA3FBB1A4ULL, "qcr220_grey_wardens6207606.cl"},
	{0xE3DFED8EC14B6CF3ULL, "183197473.ogg"},
	{0xE3E0080D70165F0FULL, "194609996.ogg"},
	{0xE3E0595347B7A776ULL, "vxca_cave_vx_0_1.vlm"},
	{0xE3E0E332C75B2D7DULL, "vxht_chantry_n_2dq_0_0.vlm"},
	{0xE3E118B6909E4E8AULL, "ave310_party_comment6134849.cl"},
	{0xE3E19747D8B23E69ULL, "ef_jess_varvel.mor"},
	{0xE3E1A005E301F0A7ULL, "zz_seb_debug6214533.cl"},
	{0xE3E1F77996EAC531ULL, "vxht_keep_6iy_0_0.vlm"},
	{0xE3E251EADB09BC23ULL, "vxlt_0_lowtown_a2_n_c_gci_0_0.vlm"},
	{0xE3E25F1DE6B828F4ULL, "mag101_grace6034180.cl"},
	{0xE3E28405B10533E0ULL, "vxht_keep_n_3y2_0_0.vlm"},
	{0xE3E2AC0006010DDDULL, "zz_dae_debug6049962.cl"},
	{0xE3E2FB7211F4648EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_loop.ani"},
	{0xE3E2FC150FB98BBFULL, "vxw_blightlands_14m_0_0.vlm"},
	{0xE3E3217FA0978098ULL, "hlac_lt_0_lowtown_a2_0_7x.dds"},
	{0xE3E33F2ACECC5640ULL, "6044923_m.fxe"},
	{0xE3E35052B0BF0E8DULL, "hlac_w_sundermount_-196609_0z.dds"},
	{0xE3E3D440F41B729AULL, "hlac_ca_caverns1_24426_7x.dds"},
	{0xE3E4128B722C6D9EULL, "llac_lt_0_lowtown_-65536_3.dds"},
	{0xE3E433C38ECD5E2EULL, "hlac_lt_gallowstemplar_n_2392_1z.dds"},
	{0xE3E46C5E4BB76AC4ULL, "art\\characters\\playercharacter\\humanfemale\\glv_bda.mmh"},
	{0xE3E4825E4BB79066ULL, "art\\characters\\playercharacter\\humanfemale\\glv_bda.msh"},
	{0xE3E48DEAE472FA90ULL, "vxht_0_hightown_a3_b8f_0_0.vlm"},
	{0xE3E516FCA8BBA495ULL, "64014426.ogg"},
	{0xE3E53E0FFF67E949ULL, "6144811_m.fxe"},
	{0xE3E5B865D8B2984DULL, "6051615_f.fxe"},
	{0xE3E5CA9EA5780140ULL, "226496063.ogg"},
	{0xE3E5F1A755D0861AULL, "dae000_merchant_access_hf6162987.cl"},
	{0xE3E63753AB87F4FCULL, "vxw_bonepitdestroyed_a3_l4_0_0.vlm"},
	{0xE3E6D4133E0CB1B1ULL, "zz_mer_debug6062747.cl"},
	{0xE3E738E6D3CD0431ULL, "vxht_0_hightown_n_bfw_0_0.vlm"},
	{0xE3E73AC9D75C74ABULL, "6031397_m.fxe"},
	{0xE3E748F265FB02FEULL, "vxw_blightlands_hy_0_0.vlm"},
	{0xE3E76678BF5336D6ULL, "23485103.ogg"},
	{0xE3E7CC988C7D1779ULL, "vxht_mansion_n_60u_0_0.vlm"},
	{0xE3E7CFF73C90168DULL, "one120_athenril6088901.cl"},
	{0xE3E7DB06E072664EULL, "vxlt_house_ei_0_0.vlm"},
	{0xE3E83B168CC92093ULL, "6112696_m.fxe"},
	{0xE3E86BFAD1E32734ULL, "vxht_0_hightown_bu4_0_0.vlm"},
	{0xE3E891AB142CE213ULL, "hlac_ht_mansion2b_818_54z.dds"},
	{0xE3E9D9059F3B648CULL, "vxlt_foundry_n_xp_0_0.vlm"},
	{0xE3E9D96B6B40EED7ULL, "zz_warrior_shielddrop.cut"},
	{0xE3EA1620F8929A4DULL, "hlac_ht_0_hightown_n_65535_1z.dds"},
	{0xE3EA8C129939B788ULL, "vxht_chantry_2y4_0_0.vlm"},
	{0xE3EAE5344974D398ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_dsgsa_0t.dds"},
	{0xE3EB0E9922400183ULL, "vxdr_0_deeproads_d30_0_0.vlm"},
	{0xE3EB36D9383DC1CFULL, "vxlt_gallowstemplar_46g_0_0.vlm"},
	{0xE3EB430E38E21896ULL, "6117188_m.fxe"},
	{0xE3EB9119FB61410FULL, "vxlt_warehouse_8t_0_0.vlm"},
	{0xE3EB9F36F62A8C11ULL, "vxht_0_hightown_a3_bfu_0_0.vlm"},
	{0xE3EBB969557D55B7ULL, "vxht_0_hightown_a2_a6y_0_0.vlm"},
	{0xE3EBBC729BF44E14ULL, "vxlt_0_lowtown_a3_n_c_hx9_0_0.vlm"},
	{0xE3EBE41C2F1755F0ULL, "hlac_ht_bartrand_n_3367_0z.dds"},
	{0xE3EC684B0F8DE720ULL, "vxlt_0_lowtown_a2_n_gxe_0_0.vlm"},
	{0xE3EC8F1EA22E155FULL, "6206872_m.fxe"},
	{0xE3ED1D6C6FC42697ULL, "vxlt_0_lowtown_docks_a3_d04_0_0.vlm"},
	{0xE3ED47624A8FE865ULL, "664034420.ogg"},
	{0xE3ED80BB1EB8A7B2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bottle_d.dds"},
	{0xE3EDD0029A144C08ULL, "723631718.ogg"},
	{0xE3EE08334DD98DE1ULL, "gen00fl_aveline6139302.cl"},
	{0xE3EECEC3387F0E06ULL, "art\\vfx\\creatures\\fxc_drg_brth_s_c.mmh"},
	{0xE3EF03D945E9D1FCULL, "art\\vfx\\combat\\vfx_w_def_def_mode_act_c.impa.ani"},
	{0xE3EF459F819DF767ULL, "vxca_caverns1_hhz_0_0.vlm"},
	{0xE3F0007655A3284AULL, "864819226.ogg"},
	{0xE3F01B81F4B2886BULL, "art\\characters\\creatures\\mubari\\animation\\wkf.evt"},
	{0xE3F04DE07CE2E7F1ULL, "vxlt_0_lowtown_a2_n_gbn_0_0.vlm"},
	{0xE3F0B77147E9DA1DULL, "vxw_bonepit_1j8_0_0.vlm"},
	{0xE3F107E245513674ULL, "883544116.ogg"},
	{0xE3F11A7CB1606EB4ULL, "vxlt_gallowstemplar_5j9_0_0.vlm"},
	{0xE3F20941F12CF999ULL, "16110130.ogg"},
	{0xE3F25FED18C6ED06ULL, "vxw_encounter2_9e_0_0.vlm"},
	{0xE3F2ED9745A42CDFULL, "vxht_0_hightown_a3_8q0_0_0.vlm"},
	{0xE3F2FFC7D6691310ULL, "vxht_0_hightown_a3_b7g_0_0.vlm"},
	{0xE3F300CBC2B4D400ULL, "430777125.ogg"},
	{0xE3F3643D95986A7AULL, "qun210cr_javaris_guard.utc"},
	{0xE3F3A76C99358E01ULL, "one110_tree_minder6133768.cl"},
	{0xE3F3B83E0B682FC2ULL, "vxlt_gallowstemplar_n_663_0_0.vlm"},
	{0xE3F3D078684C889BULL, "lt_0_lowtown_docks_n_97_0.msh"},
	{0xE3F3E278684CA7E6ULL, "lt_0_lowtown_docks_n_97_0.mao"},
	{0xE3F3EE78684CBB85ULL, "lt_0_lowtown_docks_n_97_0.mmh"},
	{0xE3F40AC64E6C8C78ULL, "plt_cod_mgc_chant_templar.plo"},
	{0xE3F41E31468ECD82ULL, "fx_smokebillow_fb.dds"},
	{0xE3F486C5D9B6954FULL, "vxlt_0_lowtown_docks_as3_0_0.vlm"},
	{0xE3F4EF087994B44EULL, "w_bonepitdestroyed_a3_103_0.mmh"},
	{0xE3F4FB087994C8E5ULL, "w_bonepitdestroyed_a3_103_0.mao"},
	{0xE3F509087994E03CULL, "w_bonepitdestroyed_a3_103_0.msh"},
	{0xE3F564DE905A0B04ULL, "aip_stonegolem_statue.gda"},
	{0xE3F5AAC3C5C31CB1ULL, "vxht_0_hightown_n_bav_0_0.vlm"},
	{0xE3F5AEA9D6BC2BACULL, "llac_lt_0_lowtown_docks_n_196607_2.dds"},
	{0xE3F5BACF57F11B7EULL, "vxw_blightlands_mz_0_0.vlm"},
	{0xE3F610D945F0063EULL, "art\\vfx\\combat\\vfx_w_def_def_mode_act_c.impa.gad"},
	{0xE3F6125EAA99FC6BULL, "ser191_emeric6120421.cl"},
	{0xE3F61C229C3B6F8BULL, "mer141_jansen6031832.cl"},
	{0xE3F63481F4B71E2CULL, "art\\characters\\creatures\\mubari\\animation\\wkf.gad"},
	{0xE3F67331C93A0484ULL, "x_partypicker_mmap_ph.dds"},
	{0xE3F6960948E6C54BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_reversal.evt"},
	{0xE3F6D4721204CBC8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sebastian_crouch_loop.gad"},
	{0xE3F7B20E22E6613EULL, "986392263.ogg"},
	{0xE3F7D4949332DD7EULL, "6110523_m.fxe"},
	{0xE3F7FB6B5EF293B3ULL, "546844934.ogg"},
	{0xE3F818CEC9151834ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_4_darren.ani"},
	{0xE3F821BB287D3405ULL, "993149224.ogg"},
	{0xE3F83538B5C1A2EFULL, "vxdr_primevalentrance_1ks_0_0.vlm"},
	{0xE3F8394B05D13FA7ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_zbma.phy"},
	{0xE3F85F717E32F383ULL, "732205572.ogg"},
	{0xE3F8601628337658ULL, "mag291_vincento6027656.cl"},
	{0xE3F8F942B0629C67ULL, "hlac_w_bonepit_131077_0x.dds"},
	{0xE3F8F96F79E2D4F5ULL, "w_bonepitdestroyed_a3_187_0.phy"},
	{0xE3F998D791A094E6ULL, "hlac_w_woundedcoast_327681_0x.dds"},
	{0xE3F9D9BB9DF1931BULL, "577435167.ogg"},
	{0xE3FA1D89D7CEC768ULL, "435503301.ogg"},
	{0xE3FA2395F6B95234ULL, "vxht_0_hightown_a2_blg_0_0.vlm"},
	{0xE3FA73C64F47C474ULL, "vxht_chantry_n_288_0_0.vlm"},
	{0xE3FA905A65D8DD83ULL, "zz_followers_banter6158394.cl"},
	{0xE3FB2167F2CB1F30ULL, "dae011_corff_sta6164568.cl"},
	{0xE3FB7E1FAB75E79EULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_f_c.mmh"},
	{0xE3FB8E2F4309F040ULL, "vxdr_primevalentrance_24v_0_0.vlm"},
	{0xE3FBA2EC3D4B308CULL, "vxlt_gallowstemplar_f_5wi_0_0.vlm"},
	{0xE3FBE9026F54F36CULL, "dae000_orana6149116.cl"},
	{0xE3FC9D3CCC7F2BA6ULL, "rock_c_04.dds"},
	{0xE3FC9E9C3FD030C8ULL, "dae000_gamlen6005048.cl"},
	{0xE3FCC5200C09AC6AULL, "vxht_0_hightown_a3_c7t_0_0.vlm"},
	{0xE3FCEBD945F5E65DULL, "art\\vfx\\combat\\vfx_w_def_def_mode_act_c.impa.evt"},
	{0xE3FD0ED39EDF8987ULL, "vxlt_blackemporium_10a_0_0.vlm"},
	{0xE3FD3381F4BD3B2AULL, "art\\characters\\creatures\\mubari\\animation\\wkf.ani"},
	{0xE3FD7F4F8082C35CULL, "hlac_lt_hangedman_-1_12z.dds"},
	{0xE3FD8414DDAC1FF2ULL, "modelhier_core.fdl"},
	{0xE3FDAF0948ED0E0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_reversal.gad"},
	{0xE3FDDD6D30197A5CULL, "125011620.ogg"},
	{0xE3FE27CD03E5B307ULL, "vxw_encounter3_8p_0_0.vlm"},
	{0xE3FEA80CC41CA232ULL, "hlac_lt_0_lowtown_docks_n_-4_2z.dds"},
	{0xE3FECA911705F54CULL, "llac_lt_hangedman_n_-1_20.dds"},
	{0xE3FF15CEC91B3196ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_ground_pound_4_darren.gad"},
	{0xE3FF302E526C8512ULL, "vxca_caverns2prc_78p_0_2.vlm"},
	{0xE3FF35A4BEF5ED23ULL, "vxlt_0_lowtown_a2_n_c_gyk_0_0.vlm"},
	{0xE3FF4A54DEDBDB2AULL, "6117306_m.fxe"},
	{0xE3FF8B4BF16600FBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_gallowpodium.msh"},
	{0xE3FF9D4BF1661F46ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_gallowpodium.mao"},
	{0xE3FFA94BF16633E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_gallowpodium.mmh"},
	{0xE3FFDBFC9CC80381ULL, "vxw_bonepitmining_a2_4w_0_0.vlm"},
	{0xE4006D0C78C88B67ULL, "6086386_m.fxe"},
	{0xE40217994F1EEA22ULL, "170672391.ogg"},
	{0xE4027E4076715E01ULL, "ave331_jeven6138742.cl"},
	{0xE40283C2C6A239A9ULL, "6081719_f.fxe"},
	{0xE40303B9D5280204ULL, "vxlt_hangedman_20i_0_0.vlm"},
	{0xE40326A410B81B00ULL, "914867975.ogg"},
	{0xE4037B0450978E8FULL, "hlac_w_encounter3_-196610_0z.dds"},
	{0xE403A4F0DF0B519BULL, "lowtown_fem_alienage1_sta6162525.cl"},
	{0xE403AE0948F1780AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\bw_reversal.ani"},
	{0xE403D3CB2F796243ULL, "sky_sky2_p0.psh.fxo"},
	{0xE403ED0E9BBDF7A0ULL, "6162757_m.fxe"},
	{0xE4040036D87D4AABULL, "6206814_m.fxe"},
	{0xE404C1E4702ED089ULL, "822911431.ogg"},
	{0xE40517518D9C0303ULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_act_c.anb"},
	{0xE40543552527FEFCULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_bigrock01_n.dds"},
	{0xE405D1A72AC5DA50ULL, "6208311_f.fxe"},
	{0xE40601ADFED0AD46ULL, "vxw_sundermount_34l_0_0.vlm"},
	{0xE40602E461BB6683ULL, "237132496.ogg"},
	{0xE40669B76093CE5AULL, "hlac_ca_caverns1_24427_4z.dds"},
	{0xE4067DDA21B0C5B6ULL, "6076216_m.fxe"},
	{0xE4070D86B19C9078ULL, "593508381.ogg"},
	{0xE4076D5BF9AC20D3ULL, "art\\characters\\creatures\\pridedemon\\animation\\grab.ani"},
	{0xE407C2F74FDD0FCBULL, "zz_followers_banter6157962.cl"},
	{0xE407CBA052974FF3ULL, "181097937.ogg"},
	{0xE40855E99CF1F2F9ULL, "hlac_lt_0_lowtown_docks_n_65534_5x.dds"},
	{0xE40865146317B6D8ULL, "l_mag131ar_bloodmages.bnk"},
	{0xE4086ACB69FFE0AEULL, "6135643_m.fxe"},
	{0xE4086B028773FB73ULL, "7806099.ogg"},
	{0xE4086C6F51C11235ULL, "ser191_emeric6120557.cl"},
	{0xE408C6947EC59709ULL, "lt_0_lowtown_a2_5_0.rml"},
	{0xE40915EAE8EB9831ULL, "vxht_0_hightown_n_9b4_0_0.vlm"},
	{0xE4095D1FAB81ED5DULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_f_c.anb"},
	{0xE409BE624B01E97CULL, "hlac_w_blightlands_196607_0z.dds"},
	{0xE40A32162C25B2CDULL, "vxlt_darktown_388_0_0.vlm"},
	{0xE40A32C1C214E0E6ULL, "vxlt_0_lowtown_a2_n_d6f_0_0.vlm"},
	{0xE40A35740C645A35ULL, "363218365.ogg"},
	{0xE40A880DDBC4B6CEULL, "vxlt_warehouse_89_0_0.vlm"},
	{0xE40A91A47CF1A872ULL, "gen_im_trs_rdr_med_glv_r1.uti"},
	{0xE40AF839BFE52BF7ULL, "hlac_lt_gallowstemplar_2394_1z.dds"},
	{0xE40B0E969DA5CC1AULL, "535253548.ogg"},
	{0xE40B592B33F21D1CULL, "gen00fl_aveline6120831.cl"},
	{0xE40BC7112E6EA15CULL, "6040637_m.fxe"},
	{0xE40C2D3146A37920ULL, "fx_smokebillow_fb.mao"},
	{0xE40C35BC9FC93250ULL, "6173647_m.fxe"},
	{0xE40C3911B40A193BULL, "ave151cr_ambush_bandit_m.utc"},
	{0xE40C76DEBF286B0DULL, "vxht_brothel_n_5kh_0_0.vlm"},
	{0xE40C96E210239BCDULL, "6126910_m.fxe"},
	{0xE40C99A9F5DBCB87ULL, "vxw_encounter3_5s_0_0.vlm"},
	{0xE40CCAEA96DFCC7AULL, "ave221_brennan.cnv"},
	{0xE40D6671ECAF1C41ULL, "415184.ogg"},
	{0xE40DE09E98B677DFULL, "lowtown_male_alienage2_mob6166558.cl"},
	{0xE40E40934A385C68ULL, "follower_banter6163481.cl"},
	{0xE40E510CC19E5B6CULL, "994404921.ogg"},
	{0xE40EC164BE438FACULL, "6188606_f.fxe"},
	{0xE40FC05501933271ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_activ_c.impa.ani"},
	{0xE40FD7695BE84175ULL, "46807871.ogg"},
	{0xE40FEB31C9C2E84CULL, "dae000_keep_servant_1_sta6156865.cl"},
	{0xE40FECEA96E23BC2ULL, "ave221_brennan.bnk"},
	{0xE4108C41EC7F4038ULL, "vxw_sundermount_2bq_0_0.vlm"},
	{0xE410D5A32035D9EDULL, "6050181_m.fxe"},
	{0xE4115E5FA93FD104ULL, "vxlt_0_lowtown_docks_a3_d0z_0_1.vlm"},
	{0xE411C723F1598E3AULL, "vxht_mansion2b_1tn_0_0.vlm"},
	{0xE411E3203D09C1E4ULL, "follower_banter6160668.cl"},
	{0xE4120EE2DED5A072ULL, "vxca_caverns2_938_0_0.vlm"},
	{0xE41245EFDCFA2132ULL, "820130798.ogg"},
	{0xE4125EA9077BB419ULL, "vxca_caverns1_mines_iki_0_0.vlm"},
	{0xE4125EEB8DB41020ULL, "6167457_m.fxe"},
	{0xE4126C5D80955AF5ULL, "6144557_f.fxe"},
	{0xE4127D6D5F63A79EULL, "147605540.ogg"},
	{0xE412F4518DA8052CULL, "art\\vfx\\combat\\vfx_w_jug_dam_redu_act_c.mmh"},
	{0xE413253887C6F130ULL, "6036539_m.fxe"},
	{0xE413CC4C0FBEEEA8ULL, "11132554.ogg"},
	{0xE4142EF9E2CFF39BULL, "484171949.ogg"},
	{0xE41447A7ECCAF6C7ULL, "fxc_dirty_smbl.mao"},
	{0xE414567D7DAA0AC1ULL, "6156856_m.fxe"},
	{0xE414655BF9B69D0AULL, "art\\characters\\creatures\\pridedemon\\animation\\grab.evt"},
	{0xE4146833614E382BULL, "zz_one_debug6058982.cl"},
	{0xE4146F58B7194F4BULL, "zz_prmpt_debug6200975.cl"},
	{0xE414738AF0C6C5C6ULL, "vxw_sundermount_33m_0_0.vlm"},
	{0xE414E9D3BD9CC731ULL, "ico_2h_axe_5.dds"},
	{0xE415AB2A06D11B8CULL, "w_sundermount_156_0.phy"},
	{0xE415CB4CC5C3E923ULL, "43831073.ogg"},
	{0xE4167B3B69A719E2ULL, "gen_im_cns_cft_psn_02.uti"},
	{0xE416A95805700A5AULL, "art\\characters\\weapons\\staff\\textures\\w_stf_001a_0s.dds"},
	{0xE416CB5501996313ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_activ_c.impa.gad"},
	{0xE4178A18FD62A905ULL, "hlac_lt_0_lowtown_a3_n_c_-1_3y.dds"},
	{0xE417D1BE6152733DULL, "lt_0_lowtown_a2_n_c_19_0_ll.mmh"},
	{0xE418ACCD433E93BDULL, "6106880_m.fxe"},
	{0xE418D6F487E44250ULL, "cd_vrc321_housekeeper.bnk"},
	{0xE419263744DB5FCCULL, "mag101_alain6032202.cl"},
	{0xE4198B439408A735ULL, "hlac_ht_brothel_n_2995_6y.dds"},
	{0xE419EA67E613D3E1ULL, "hlac_lt_0_lowtown_a2_n_-131072_9y.dds"},
	{0xE41A05157A5BE2EDULL, "w_bonepitdestroyed_a3_54_0.mmh"},
	{0xE41A09157A5BE93EULL, "w_bonepitdestroyed_a3_54_0.mao"},
	{0xE41A17157A5C0173ULL, "w_bonepitdestroyed_a3_54_0.msh"},
	{0xE41A1B9030DF22C1ULL, "6110546_m.fxe"},
	{0xE41A3D0CDCEBC351ULL, "750634930.ogg"},
	{0xE41B2B606AA5337AULL, "lgt260_eau_tmp_roderick6143713.cl"},
	{0xE41B445BF9BC847DULL, "art\\characters\\creatures\\pridedemon\\animation\\grab.gad"},
	{0xE41B706BBF3DAB9BULL, "vxca_caverns1_iih_0_0.vlm"},
	{0xE41C29CA18A04D36ULL, "vxw_encounter3_ge_0_1.vlm"},
	{0xE41D72F261ED5F50ULL, "557558937.ogg"},
	{0xE41DA458DD8C38D0ULL, "vxlt_0_lowtown_eqs_0_0.vlm"},
	{0xE41DC855019F7D78ULL, "art\\vfx\\combat\\vfx_w_bsk_charge_activ_c.impa.evt"},
	{0xE41DEFB1C3A169CEULL, "dae000_arishok6067850.cl"},
	{0xE41DFEB177AB05EBULL, "hlac_lt_0_lowtown_n_-131072_2x.dds"},
	{0xE41E53D1BB4E53CDULL, "w_bonepit_34_0.msh"},
	{0xE41E5D9A521844B9ULL, "6121986_m.fxe"},
	{0xE41E61D1BB4E6B00ULL, "w_bonepit_34_0.mao"},
	{0xE41E65D1BB4E7253ULL, "w_bonepit_34_0.mmh"},
	{0xE41EBCEE85FCD29CULL, "174259206.ogg"},
	{0xE41F053A4FD61DABULL, "hlac_lt_shop_-1_7x.dds"},
	{0xE41F9625552963EAULL, "336981565.ogg"},
	{0xE4203ADADBC2FFFFULL, "gen00fl_fenris6106686.cl"},
	{0xE4203B99D25944F7ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgte.mmh"},
	{0xE4205199D2596A69ULL, "art\\characters\\playercharacter\\humanfemale\\glv_lgte.msh"},
	{0xE42080BFD0CBB8F2ULL, "vxca_caverns2_949_0_0.vlm"},
	{0xE420C5DDE773E74DULL, "6171507_m.fxe"},
	{0xE421359A6F44051AULL, "6043546_m.fxe"},
	{0xE421393F752B04A1ULL, "nat380_temmerin6067367.cl"},
	{0xE42148B22A32982AULL, "6090810_m.fxe"},
	{0xE4217961435A583BULL, "6164297_m.fxe"},
	{0xE421A9395251650FULL, "lt_0_lowtown_docks_107_0.phy"},
	{0xE421BE942C5FADC4ULL, "664784632.ogg"},
	{0xE421F4A0B39C2F93ULL, "mag131cr_bloodmage.utc"},
	{0xE422BA0353F13D18ULL, "plt_cod_ltr_danzig_wayward.plo"},
	{0xE4232C9899651E20ULL, "follower_banter6164044.cl"},
	{0xE4236D5E18786C6AULL, "6103616_f.fxe"},
	{0xE42374C561E6EFC2ULL, "hlac_w_encounter1_-65533_0x.dds"},
	{0xE4239DBFB86E4269ULL, "zz_followers_banter6158961.cl"},
	{0xE42405157A643766ULL, "w_bonepitdestroyed_a3_54_0.rml"},
	{0xE424271D9AADF7E2ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.cess.gad"},
	{0xE42461E07A84565EULL, "dae300_petrice6098083.cl"},
	{0xE424D94357846AA3ULL, "dae000_arishok6068474.cl"},
	{0xE424E1F028D219B5ULL, "vxw_woundedcoast_15n_0_0.vlm"},
	{0xE4251E894A9B7605ULL, "362523231.ogg"},
	{0xE4252F98C3F0C73FULL, "art\\characters\\creatures\\dragon\\animation\\kd_enter.gad"},
	{0xE4255900B31E1432ULL, "95625852.ogg"},
	{0xE4258629DA6075E7ULL, "vxht_mansion2_1hr_0_1.vlm"},
	{0xE425E61D8DA42619ULL, "hlac_w_encounter2_camp_65536_0x.dds"},
	{0xE4264FFBF40D84CFULL, "131950620.ogg"},
	{0xE42661D91CD12A1EULL, "llac_w_bonepitmining_a2_393221_0.dds"},
	{0xE427922A407E9778ULL, "714104007.ogg"},
	{0xE4279547805568E8ULL, "770724636.ogg"},
	{0xE4282F3B21D6B603ULL, "419079171.ogg"},
	{0xE42882C5E4C55B6AULL, "123187167.ogg"},
	{0xE428F2833A5F83E0ULL, "vxw_encounter2_2w_0_1.vlm"},
	{0xE4290AEF9674107CULL, "hlac_lt_0_lowtown_a3_n_c_131074_0y.dds"},
	{0xE429C1707C0FADC0ULL, "vxht_0_hightown_n_a4s_0_0.vlm"},
	{0xE429CD29FB40C7C5ULL, "103692819.ogg"},
	{0xE429D6451FDA44C4ULL, "19666972.ogg"},
	{0xE42A4A1D9AB29ED9ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.cess.evt"},
	{0xE42A6853EC8FEB1FULL, "vxdr_0_deeproads_epx_0_0.vlm"},
	{0xE42AC4D59D1D508AULL, "619075859.ogg"},
	{0xE42BC99AD0829EDFULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_ferelden_jimmy_sa.dds"},
	{0xE42BE97E07DABE9DULL, "hlac_dr_primevalentrance_1726_6y.dds"},
	{0xE42BEEECF2D47D78ULL, "vxlt_0_lowtown_a2_n_c_i0u_0_0.vlm"},
	{0xE42C2498C3F6D32CULL, "art\\characters\\creatures\\dragon\\animation\\kd_enter.evt"},
	{0xE42C4D6BA12B8059ULL, "6179913_f.fxe"},
	{0xE42C81AEF689B812ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_002a_0s.dds"},
	{0xE42CD52C34CDFF63ULL, "vxht_mansion2b_34a_0_0.vlm"},
	{0xE42D0EACB07BC909ULL, "follower_banter6171126.cl"},
	{0xE42E04B7B4DF40F1ULL, "darktown_male_fer_urch_sta6175543.cl"},
	{0xE42EEDBA6DEE849DULL, "vxca_caverns1_ea2_0_0.vlm"},
	{0xE4300641F93ADFF1ULL, "vxlt_0_lowtown_docks_n_c5j_0_0.vlm"},
	{0xE4307F626587955BULL, "gen00fl_isabela6135679.cl"},
	{0xE430AD6B899DF4AFULL, "915913443.ogg"},
	{0xE430D97F4AA93185ULL, "ht_0_hightown_a2_34_0_ll.mmh"},
	{0xE4312FF86C38FF04ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4_tool.msh"},
	{0xE43141F86C391E6DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4_tool.mao"},
	{0xE43145F86C3924A6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4_tool.mmh"},
	{0xE431F328A632DBFEULL, "qcr000_arishok6059309.cl"},
	{0xE4320101AD5543D3ULL, "hlac_lt_0_lowtown_131073_1z.dds"},
	{0xE4320FD438B3E156ULL, "zz_and_debug6178570.cl"},
	{0xE43237E3733C9573ULL, "hlac_ca_cavern1_haunted_24424_14x.dds"},
	{0xE43249B7E8EF020AULL, "hlac_lt_0_lowtown_a2_n_-196609_3y.dds"},
	{0xE432A42D4D6381C3ULL, "zz_dog_debug6127862.cl"},
	{0xE432BD7E8A430095ULL, "vxlt_0_lowtown_docks_9r6_0_0.vlm"},
	{0xE43353CD1AC83235ULL, "vxw_woundedcoast_10m_0_0.vlm"},
	{0xE433C6FC0C99B497ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_1z.dds"},
	{0xE433CB6F24E7578BULL, "282338539.ogg"},
	{0xE434A833D922B7C0ULL, "259376875.ogg"},
	{0xE434D85D3F9780C5ULL, "vxlt_gallowscourtyard_n_76w_0_0.vlm"},
	{0xE4353BC2C60B7F5CULL, "zz_followers_banter6159024.cl"},
	{0xE43588870744439BULL, "hm_hubert.mor"},
	{0xE4359BF7A96FCF65ULL, "mag210_arianni6024978.cl"},
	{0xE435BBA2744552E1ULL, "vxlt_gallowstemplar_f_4qg_0_0.vlm"},
	{0xE4360B63446F2275ULL, "vxlt_gallowscourtyard_63z_0_0.vlm"},
	{0xE43624EE30A7D1BEULL, "331538885.ogg"},
	{0xE436476F10CEAAAFULL, "236265939.ogg"},
	{0xE43650483FF7B187ULL, "vxlt_0_lowtown_a2_eup_0_0.vlm"},
	{0xE4370FEF98944027ULL, "412189262.ogg"},
	{0xE43750EB4D5E5494ULL, "dre000_bethany6037308.cl"},
	{0xE437696AE21ADF1AULL, "w_blightlands_170_0.phy"},
	{0xE43802B095D3A48BULL, "vxht_0_hightown_a3_aln_0_0.vlm"},
	{0xE438521D9ABEE9C8ULL, "art\\vfx\\creatures\\vfx_c_abom_crust_c.cess.ani"},
	{0xE4395E98CC3A1BA7ULL, "vxht_0_hightown_a3_5jd_0_0.vlm"},
	{0xE439F42B5B76251BULL, "439484591.ogg"},
	{0xE43A0C98C402E785ULL, "art\\characters\\creatures\\dragon\\animation\\kd_enter.ani"},
	{0xE43A7FBC7BAD3D22ULL, "1036127585.ogg"},
	{0xE43AC25E954D32BEULL, "853724391.ogg"},
	{0xE43ACCA16BD3CE63ULL, "gen00fl_fenris6112117.cl"},
	{0xE43ADCDF88784BB2ULL, "character_character_p287.psh.fxo.dx11.assembly"},
	{0xE43B3CF86C416A71ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat4x4_tool.phy"},
	{0xE43BD91F55C8243DULL, "196049778.ogg"},
	{0xE43C8111C291A146ULL, "6075639_m.fxe"},
	{0xE43D0226452B8D35ULL, "744514938.ogg"},
	{0xE43D29102542586FULL, "6166978_m.fxe"},
	{0xE43DB8F488032B1AULL, "cd_vrc321_housekeeper.wbd"},
	{0xE43DBD14FA4FB48CULL, "mag360cr_hostage_isabela.utc"},
	{0xE43E042C551FA7A9ULL, "km_cricket.mor"},
	{0xE43E08AE3103ACDCULL, "hlac_lt_gallowstemplar_f_2384_2z.dds"},
	{0xE43F2DD8286665A4ULL, "vxlt_undercityserialkiller_11v_0_0.vlm"},
	{0xE43F5F2A2EA73534ULL, "vxlt_0_lowtown_a2_hij_0_0.vlm"},
	{0xE4400C6F7A1EF710ULL, "w_bonepitdestroyed_a3_187_0.msh"},
	{0xE4401A6F7A1F0F2AULL, "w_bonepitdestroyed_a3_187_0.mmh"},
	{0xE4401E6F7A1F1661ULL, "w_bonepitdestroyed_a3_187_0.mao"},
	{0xE4402DFA88782811ULL, "vxca_caverns2prc_9w2_0_0.vlm"},
	{0xE4405E7D44A0F82DULL, "hlac_w_encounter2_262149_0y.dds"},
	{0xE440C35A049BCBA1ULL, "vxca_caverns2_8rq_0_0.vlm"},
	{0xE440CE2CE01036DCULL, "6140559_m.fxe"},
	{0xE4410DFA88758441ULL, "vxdr_dungeon_1t7_0_1.vlm"},
	{0xE44118DF9234308AULL, "fxa_pois_split_blur.dds"},
	{0xE441742C972F2AB3ULL, "vxht_keep_n_6f5_0_0.vlm"},
	{0xE4418B68CDCE782FULL, "gen00fl_merrill6121757.cl"},
	{0xE441DECBB8E68149ULL, "6170764_m.fxe"},
	{0xE4429071481034F0ULL, "6120768_m.fxe"},
	{0xE442E4EA970D87CDULL, "ave221_brennan.stg"},
	{0xE442F22108B08886ULL, "pro000_cut_act_of_mercy.cnv"},
	{0xE442F8A357CA048DULL, "6102573_f.fxe"},
	{0xE4439260CE165789ULL, "6023886_m.fxe"},
	{0xE443975C07127B74ULL, "lit_bt_cyc.mao"},
	{0xE443E0254052A429ULL, "hlac_lt_warehouse_n_0_4y.dds"},
	{0xE443E41D68661E5AULL, "393258385.ogg"},
	{0xE443E95AAF626F20ULL, "409254095.ogg"},
	{0xE4442D7F663B6B61ULL, "vxlt_gallowstemplar_f_4vf_0_0.vlm"},
	{0xE44433812C6F4D6BULL, "6000689_m.fxe"},
	{0xE444587346FBD616ULL, "qun221_petrice6077880.cl"},
	{0xE44463C149BE81BEULL, "16632338.ogg"},
	{0xE4446F2940E0D0DFULL, "311610474.ogg"},
	{0xE44478B1C9F359A6ULL, "vxdr_0_deeproads_ex5_0_0.vlm"},
	{0xE44494D44E8BDBCEULL, "hlac_w_encounter1_-65534_0y.dds"},
	{0xE444C22531EDCA07ULL, "vxlt_0_lowtown_a2_exs_0_0.vlm"},
	{0xE4450BDA11A5F5CAULL, "vxca_caverns1_mines_i14_0_0.vlm"},
	{0xE4459B72552B4796ULL, "147601948.ogg"},
	{0xE445F42108B2C246ULL, "pro000_cut_act_of_mercy.bnk"},
	{0xE445F43BD659C687ULL, "612458917.ogg"},
	{0xE4464A295E27F4BCULL, "vxca_caverns1_mines_in3_0_0.vlm"},
	{0xE44656A50C0B0F17ULL, "gen00fl_isabela6110155.cl"},
	{0xE4466A662A8608C4ULL, "772730349.ogg"},
	{0xE446748D87C9BD0BULL, "vxht_0_hightown_a3_aia_0_0.vlm"},
	{0xE4467C6B028BEBA7ULL, "t1_mul_br1.tnt"},
	{0xE446DE1BA1D9C531ULL, "800079672.ogg"},
	{0xE447539B4D62E7AAULL, "vrc221tr_huggin.ncs"},
	{0xE447824BF1A3A692ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9pp_gallowpodium.phy"},
	{0xE447A61CDADDF96EULL, "hlac_ht_mansion2_818_65y.dds"},
	{0xE447F3F49F9A87A9ULL, "hlac_ca_caverns2prc_5802_46y.dds"},
	{0xE44840312C14B219ULL, "1400659.ogg"},
	{0xE44867E991994247ULL, "dae000_mother6004115.cl"},
	{0xE448F093BCDC3264ULL, "6174804_m.fxe"},
	{0xE4495ABFA71370E6ULL, "70590331.ogg"},
	{0xE449D06F94C50144ULL, "hlac_lt_0_lowtown_n_-65537_5z.dds"},
	{0xE44A3B45F3053CD9ULL, "qcr220_grey_wardens6207436.cl"},
	{0xE44A908563B6DC20ULL, "6016502_m.fxe"},
	{0xE44AD6C6D48E2146ULL, "gen00fl_aveline6139156.cl"},
	{0xE44ADA48A3C1B61AULL, "lt_0_lowtown_docks_134_0.mmh"},
	{0xE44ADE48A3C1BD51ULL, "lt_0_lowtown_docks_134_0.mao"},
	{0xE44AE8939BA25448ULL, "6218708_m.fxe"},
	{0xE44AEC48A3C1D560ULL, "lt_0_lowtown_docks_134_0.msh"},
	{0xE44B8B3680F071F9ULL, "mcr310_party_banter6087811.cl"},
	{0xE44BECC2B3139FB7ULL, "vxlt_gallowstemplar_n_3p5_0_0.vlm"},
	{0xE44CB088D5BEA5E5ULL, "vxht_0_hightown_a2_btz_0_0.vlm"},
	{0xE44D9FB51A5C7E24ULL, "vxlt_undercityserialkiller_16w_0_0.vlm"},
	{0xE44E4B3D631FAE3AULL, "w_bonepitdestroyed_a3_26_0.mmh"},
	{0xE44E4F3D631FB571ULL, "w_bonepitdestroyed_a3_26_0.mao"},
	{0xE44E5D3D631FCC80ULL, "w_bonepitdestroyed_a3_26_0.msh"},
	{0xE44E741BBB8195D0ULL, "llac_lt_0_lowtown_docks_a3_n_-262144_2.dds"},
	{0xE44E9FD77A6E4091ULL, "vxca_caverns2prc_9r1_0_0.vlm"},
	{0xE44EF35C934C7702ULL, "ht_0_hightownexplosion_n_27_0_lc.mmh"},
	{0xE44EF5C290E49979ULL, "6051410_m.fxe"},
	{0xE44F0C9BD16EE063ULL, "vxlt_0_lowtown_docks_a3_n_cou_0_0.vlm"},
	{0xE44FE60989254333ULL, "vxht_keep_n_6a4_0_0.vlm"},
	{0xE4500E06955A3B7CULL, "follower_barks6164414.cl"},
	{0xE450A5BD06EAC501ULL, "llac_w_encounter2_65534_0.dds"},
	{0xE45130DCC0D48DB3ULL, "gen00fl_aveline6112736.cl"},
	{0xE451798606CE54D2ULL, "6148467_m.fxe"},
	{0xE451FC822DFE52ECULL, "6082183_m.fxe"},
	{0xE4534625CDBD0C5BULL, "6174029_m.fxe"},
	{0xE453A3CDD8BC2A73ULL, "6208544_m.fxe"},
	{0xE454959B0C740606ULL, "lt_0_lowtown_docks_a3_n_probe_ctt.dds"},
	{0xE45542B7B3E203D2ULL, "hlac_lt_gallowstemplar_2383_6z.dds"},
	{0xE455BDEAC272349FULL, "vxdr_0_deeproads_ece_0_0.vlm"},
	{0xE455C412514F19BFULL, "6075986_m.fxe"},
	{0xE455F392C64D229DULL, "vxlt_hangedman_m0_0_1.vlm"},
	{0xE45604DB266F8FFBULL, "dae021_brothel_onenter6196708.cl"},
	{0xE4562AB1F2C00C7AULL, "qun231_varnell6077283.cl"},
	{0xE4564FB1260C0349ULL, "6081801_f.fxe"},
	{0xE456822A07086D57ULL, "w_sundermount_156_0.mmh"},
	{0xE456862A07087464ULL, "w_sundermount_156_0.mao"},
	{0xE456982A070892C9ULL, "w_sundermount_156_0.msh"},
	{0xE45721E75C038C42ULL, "follower_banter6136373.cl"},
	{0xE45764F805A104B9ULL, "llac_lt_darktown_-1_9.dds"},
	{0xE45818E1BEA8E954ULL, "923802077.ogg"},
	{0xE45853073005CE94ULL, "art\\vfx\\combat\\vfx_r_var_rain_arrow_act_c.impa.ani"},
	{0xE458537D7635EAF6ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_7x.dds"},
	{0xE458909209E7A8E0ULL, "6000941_m.fxe"},
	{0xE45897A3C54963D4ULL, "follower_banter6161313.cl"},
	{0xE458B5E97E0E3D75ULL, "vxlt_0_lowtown_e2o_0_0.vlm"},
	{0xE45901E4FF98704EULL, "art\\characters\\playercharacter\\humanfemale\\glv_medb.phy"},
	{0xE45982D31F424667ULL, "hlac_lt_gallowstemplar_2382_2x.dds"},
	{0xE45A1A834B160EADULL, "follower_barks6173413.cl"},
	{0xE45A38A2BE447A38ULL, "818312632.ogg"},
	{0xE45A467E6F3C2F05ULL, "vxlt_0_lowtown_a2_n_c_i0l_0_0.vlm"},
	{0xE45A4D6927BC8125ULL, "371210540.ogg"},
	{0xE45A9BE62A8B97FFULL, "vxlt_gallowstemplar_n_37d_0_0.vlm"},
	{0xE45B84EE3F672219ULL, "vrc321_party_barks6096303.cl"},
	{0xE45B8711D52BC342ULL, "hlac_ht_0_hightown_n_131071_2y.dds"},
	{0xE45BB9CA4E31ABAEULL, "6067993_m.fxe"},
	{0xE45BBA89363F9368ULL, "6020428_m.fxe"},
	{0xE45C0BC25BBA8B0DULL, "dae000_female_tranquil6165133.cl"},
	{0xE45C1D46590BFC5BULL, "6070788_f.fxe"},
	{0xE45CF33D8EEA0C1EULL, "376954974.ogg"},
	{0xE45D7790CCD51512ULL, "642735511.ogg"},
	{0xE45DCE982BE9C680ULL, "vxw_bonepitmining_a2_dj_0_0.vlm"},
	{0xE45E0B68A693372AULL, "hlac_dr_dungeon_3207_8x.dds"},
	{0xE45E222E51C0C388ULL, "122403930.ogg"},
	{0xE45E4B815A798D12ULL, "6116662_m.fxe"},
	{0xE45F15C954B17950ULL, "combat_blood.gda"},
	{0xE45F5007300BE8F6ULL, "art\\vfx\\combat\\vfx_r_var_rain_arrow_act_c.impa.gad"},
	{0xE45FD16DA903F091ULL, "6113231_m.fxe"},
	{0xE4606E5F1FF46B6CULL, "6085083_m.fxe"},
	{0xE4608501C6883A6DULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyn.mmh"},
	{0xE4608901C68840BEULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyn.mao"},
	{0xE460935B96A10B8CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbtall.msh"},
	{0xE46094152BCAE54AULL, "vxlt_0_lowtown_docks_n_bex_0_0.vlm"},
	{0xE4609701C68858F3ULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyn.msh"},
	{0xE460995B96A1167EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbtall.mmh"},
	{0xE460A55B96A12A15ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerbtall.mao"},
	{0xE461172FF3950BEDULL, "hlac_lt_0_lowtown_docks_n_196607_3x.dds"},
	{0xE4619C7ECD41D153ULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.cess.ani"},
	{0xE4623C7B54488478ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyn_0s.dds"},
	{0xE46254E4F18AE805ULL, "346330396.ogg"},
	{0xE4626566DA75BB72ULL, "gen_im_cft_rec_pot_04.uti"},
	{0xE463DAADC0A67413ULL, "war_defbonus.dds"},
	{0xE463E22CD84CA9A6ULL, "follower_banter6143290.cl"},
	{0xE464BF8E0CF7199FULL, "hlac_lt_0_lowtown_docks_a3_-196610_2z.dds"},
	{0xE46560E06B02B2C9ULL, "vis121_saemus6099699.cl"},
	{0xE4658582F812ACA7ULL, "6128338_m.fxe"},
	{0xE4659453AE772DA3ULL, "45347013.ogg"},
	{0xE465BD6A2588A60BULL, "gen00fl_varric6076567.cl"},
	{0xE4661213B6B950F9ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_f_c.impa.ani"},
	{0xE4666B0730123555ULL, "art\\vfx\\combat\\vfx_r_var_rain_arrow_act_c.impa.evt"},
	{0xE46696EDB09320DCULL, "vxlt_0_lowtown_n_eov_0_0.vlm"},
	{0xE466A3EB2A92EAE3ULL, "mag100_thrask6034871.cl"},
	{0xE466DCC4496ADA67ULL, "128304953.ogg"},
	{0xE4671EEA9F5135DAULL, "vxlt_0_lowtown_a2_hdc_0_0.vlm"},
	{0xE4672A59D05300E1ULL, "850718338.ogg"},
	{0xE46752D2A7C48483ULL, "vxht_0_hightown_a3_c97_0_0.vlm"},
	{0xE46767032798ECF3ULL, "vxlt_0_lowtown_a3_n_c_eva_0_0.vlm"},
	{0xE468064470C811CFULL, "hightown_fem_courtesan_mob.bnk"},
	{0xE46832E32CA3E84DULL, "6034051_m.fxe"},
	{0xE468D1BEDC9F3DD3ULL, "ico_plot.dds"},
	{0xE468F25E3F7F9F45ULL, "6155916_m.fxe"},
	{0xE4693C53990EA1EDULL, "6201350_m.fxe"},
	{0xE469DA23890D0B4BULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.dura.ani"},
	{0xE46A1C417C0A6905ULL, "hlac_lt_0_lowtown_docks_n_-131073_2z.dds"},
	{0xE46A8C420AC7660CULL, "dae000_gamlen6145120.cl"},
	{0xE46AC7E804E71EC9ULL, "6141922_m.fxe"},
	{0xE46B284470CA8137ULL, "hightown_fem_courtesan_mob.cnv"},
	{0xE46B3301C12E3A4BULL, "389973642.ogg"},
	{0xE46B37E76A7E763AULL, "830867471.ogg"},
	{0xE46B3902C5A20A32ULL, "one100dg_elren6014718.cl"},
	{0xE46B6B8D61EB42EAULL, "301702168.ogg"},
	{0xE46C0A2454D22C8FULL, "dae000_tomwise6167264.cl"},
	{0xE46C40751DDFDF00ULL, "vxw_bonepitmining_a2_ck_0_0.vlm"},
	{0xE46C89CB7093E8B2ULL, "llac_w_sundermount_-65542_0.dds"},
	{0xE46CED13B6BF303BULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_f_c.impa.gad"},
	{0xE46D04903FEF98B1ULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.anb"},
	{0xE46D181F2E4736DCULL, "art\\vfx\\combat\\vfx_glb_restore_stamina_c.impa.ani"},
	{0xE46D884455D09990ULL, "mcr370_meredith6084414.cl"},
	{0xE46DC79288514073ULL, "mag100_thrask6032033.cl"},
	{0xE46DDBC17D81222BULL, "960032650.ogg"},
	{0xE46E104A4F863C79ULL, "6111277_m.fxe"},
	{0xE46E16A2A485B910ULL, "399643770.ogg"},
	{0xE46E947ECD4C4D8AULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.cess.evt"},
	{0xE46EA61852BD3BEEULL, "6105954_m.fxe"},
	{0xE46F05F21DC0FDCAULL, "vxlt_0_lowtown_docks_n_bby_0_0.vlm"},
	{0xE46F2B39968D4099ULL, "gen00fl_aveline6139319.cl"},
	{0xE46FF948A3E10725ULL, "lt_0_lowtown_docks_134_0.phy"},
	{0xE470109B7ACF2038ULL, "107598329.ogg"},
	{0xE470812E76C5C7B0ULL, "6068356_m.fxe"},
	{0xE47094AC8D4308ABULL, "486446638.ogg"},
	{0xE470AA9A9B3AE4D5ULL, "6045002_m.fxe"},
	{0xE470B510E119A741ULL, "vxlt_0_lowtown_a2_n_c_gc3_0_0.vlm"},
	{0xE47148D2B9DE6004ULL, "6176583_m.fxe"},
	{0xE4717D64B0A89770ULL, "6102773_m.fxe"},
	{0xE4717E01C696A12AULL, "art\\characters\\playercharacter\\elffemale\\boo_hvyn.phy"},
	{0xE471A1531D1783A0ULL, "187170006.ogg"},
	{0xE471D5804C625BDDULL, "368792781.ogg"},
	{0xE47297EFB8B5ED8FULL, "vxht_keep_n_67b_0_0.vlm"},
	{0xE472A48E895A4376ULL, "6143189_m.fxe"},
	{0xE472D4537DCB33C2ULL, "6166382_m.fxe"},
	{0xE472E2B44530FC11ULL, "vxlt_0_lowtown_docks_bql_0_0.vlm"},
	{0xE47312DB9D9DDD85ULL, "vxht_mansion_n_1pn_0_0.vlm"},
	{0xE473251F2E4BB81EULL, "art\\vfx\\combat\\vfx_glb_restore_stamina_c.impa.gad"},
	{0xE473331B3D539631ULL, "vxht_mansion_n_60m_0_0.vlm"},
	{0xE473C5AB3D1B2E57ULL, "vxlt_0_lowtown_f0t_0_0.vlm"},
	{0xE4740A13B6C58020ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_f_c.impa.evt"},
	{0xE47429F6BB921D1EULL, "6149064_m.fxe"},
	{0xE474509FA8501BB0ULL, "6098287_m.fxe"},
	{0xE474BE1D0D9F9021ULL, "vxlt_gallowstemplar_5um_0_0.vlm"},
	{0xE47508CAA289395CULL, "vxlt_0_lowtown_n_ehw_0_0.vlm"},
	{0xE475737ECD5234FDULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.cess.gad"},
	{0xE475A7CF620D37E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_reaction_left_shoulder.evt"},
	{0xE476843FBD324C37ULL, "dr_0_deeproads_5839.anb"},
	{0xE476A4C01E9A00CDULL, "6031351_m.fxe"},
	{0xE476B5F76766BF8DULL, "fonts_en-us.gfx"},
	{0xE477313DCDE63FECULL, "6032002_f.fxe"},
	{0xE4774BDB37F4987EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium_trim1.ani"},
	{0xE4775BA460A0DDE0ULL, "6035848_m.fxe"},
	{0xE4776FFEDAE75857ULL, "hlac_lt_undercitydungeon_-1_15x.dds"},
	{0xE477C22389191FC2ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.dura.evt"},
	{0xE477D18CBA90982DULL, "vxlt_foundry_n_x5_0_0.vlm"},
	{0xE477E97CEFCC0D7EULL, "hlac_w_encounter2_131071_0y.dds"},
	{0xE478061FABA0EF65ULL, "vxht_0_hightown_a2_bgg_0_0.vlm"},
	{0xE478747779059182ULL, "llac_ht_keep_3732_15.dds"},
	{0xE47883C013E49A17ULL, "6086239_f.fxe"},
	{0xE47939EA16632EE0ULL, "1034363374.ogg"},
	{0xE47962A297939C81ULL, "hlac_lt_0_lowtown_a2_n_c_1_3y.dds"},
	{0xE47979CC49F81036ULL, "633553841.ogg"},
	{0xE47A001F2E51983DULL, "art\\vfx\\combat\\vfx_glb_restore_stamina_c.impa.evt"},
	{0xE47A25903FFA5B2AULL, "art\\vfx\\combat\\vfx_glb_fire_die_c.mmh"},
	{0xE47A4086AB5C265CULL, "6043841_m.fxe"},
	{0xE47A4B3D6344E951ULL, "w_bonepitdestroyed_a3_26_0.rml"},
	{0xE47AA96FCE45AF87ULL, "6129272_m.fxe"},
	{0xE47ABC2547F57E5DULL, "vxlt_0_lowtown_f2p_0_0.vlm"},
	{0xE47AD8D7657F658BULL, "6126901_m.fxe"},
	{0xE47B0565CC4B550DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxc.mmh"},
	{0xE47B0965CC4B5C5EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxc.mao"},
	{0xE47B1765CC4B7393ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_bannerxc.msh"},
	{0xE47B4541485A71DCULL, "6107293_m.fxe"},
	{0xE47B46AA8E7D12D5ULL, "follower_banter6169844.cl"},
	{0xE47C245A78602EB4ULL, "vxht_0_hightown_a2_b9l_0_0.vlm"},
	{0xE47C39880FF43314ULL, "vxht_brothel_n_6jy_0_0.vlm"},
	{0xE47C4A70E705025DULL, "vxw_encounter3_n_0_0.vlm"},
	{0xE47CBECF62137D31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_reaction_left_shoulder.gad"},
	{0xE47D16A87705FB23ULL, "lt_0_lowtown_docks_n_211_0.msh"},
	{0xE47D24A87706137DULL, "lt_0_lowtown_docks_n_211_0.mmh"},
	{0xE47D28A87706198EULL, "lt_0_lowtown_docks_n_211_0.mao"},
	{0xE47DC123891D89D5ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.dura.gad"},
	{0xE47E173A77D1F8FEULL, "6059398_m.fxe"},
	{0xE47E95EADA25B0C9ULL, "6163405_m.fxe"},
	{0xE47F003AC4943E36ULL, "vxht_0_hightown_a2_bxq_0_0.vlm"},
	{0xE47FEF41A29EAFF0ULL, "6107873_m.fxe"},
	{0xE480119D260056BCULL, "vxca_caverns1_mines_iz7_0_0.vlm"},
	{0xE4805EBACB631F2CULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_arishok_isabellasreturn.gad"},
	{0xE480D61CFF520C35ULL, "6147025_m.fxe"},
	{0xE480EA6D8FF40F07ULL, "vxht_keep_n_6be_0_0.vlm"},
	{0xE480F6AA871748EFULL, "421769303.ogg"},
	{0xE48109CCAAAC060FULL, "vxht_keep_n_62m_0_0.vlm"},
	{0xE481922B7F87366FULL, "follower_banter6149211.cl"},
	{0xE482560CA5F440B5ULL, "hlac_w_encounter3_-131071_0z.dds"},
	{0xE482CBDA37AA7220ULL, "vxlt_0_lowtown_docks_n_c1c_0_0.vlm"},
	{0xE482CC332AAE5780ULL, "mrl150_merrill6071869.cl"},
	{0xE48357964CFA5C09ULL, "460078754.ogg"},
	{0xE4835C79C285F9C0ULL, "6062372_m.fxe"},
	{0xE4837CA06B21E58DULL, "953395007.ogg"},
	{0xE483830D4B8AE467ULL, "zz_ev23449.cut"},
	{0xE483BFCF62199DDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_e_reaction_left_shoulder.ani"},
	{0xE4840CF24ECBCB91ULL, "vxlt_gallowscourtyard_7fh_0_0.vlm"},
	{0xE484368C8BB0B583ULL, "vxht_0_hightown_a3_c79_0_0.vlm"},
	{0xE48437085DAF5EF9ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_8whitelyrium.msh"},
	{0xE48441085DAF6FE7ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_8whitelyrium.mmh"},
	{0xE48445085DAF7634ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_8whitelyrium.mao"},
	{0xE4849B754A13CEA0ULL, "lgt210_eau_keran_father6144901.cl"},
	{0xE484A84D6F93F812ULL, "vxlt_0_lowtown_docks_8pz_0_0.vlm"},
	{0xE484C92C6096EB56ULL, "gen00fl_varric6063849.cl"},
	{0xE48563DB3800FE4FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium_trim1.evt"},
	{0xE4857A37BC49AC49ULL, "gen00fl_isabela_init6120902.cl"},
	{0xE485C9ED44E80C12ULL, "dr_primevaltomb_659_ll.mmh"},
	{0xE4860E69C2408362ULL, "123848107.ogg"},
	{0xE4875DBACB693C2AULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_arishok_isabellasreturn.ani"},
	{0xE488113948D5D467ULL, "vfx0.vsh.fxo"},
	{0xE488158CD1F9152AULL, "lt_0_lowtown_docks_a3_n_41_0.rml"},
	{0xE48821F0D88E5F59ULL, "gen00fl_isabela6097362.cl"},
	{0xE488674B4246A291ULL, "vxca_caverns2prc_9n5_0_0.vlm"},
	{0xE488703A2D38BC8FULL, "6097000_f.fxe"},
	{0xE488949EB49E32BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mc.msh"},
	{0xE488A29EB49E4AF3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mc.mao"},
	{0xE488A69EB49E5100ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mc.mmh"},
	{0xE488E171315D7FE1ULL, "hlac_lt_0_lowtown_a2_n_-65535_6z.dds"},
	{0xE4899324DA3CABD2ULL, "6206913_f.fxe"},
	{0xE48A89AD937F8611ULL, "t3_dog_d01.tnt"},
	{0xE48A9D45CD8212A7ULL, "amb_scene_kidsplay.wbd"},
	{0xE48B4E2D56AED738ULL, "6107150_m.fxe"},
	{0xE48C220DA12438EBULL, "hlac_w_bonepitmining_a2_458752_0z.dds"},
	{0xE48C53F1E44116D1ULL, "mrl251ip_dead_harshal.utp"},
	{0xE48C64DB38071EF8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b_weightshift_medium_trim1.gad"},
	{0xE48C937532695C51ULL, "plt_lgt000pt_ultra_2_03.nss"},
	{0xE48D746043293CACULL, "6070876_m.fxe"},
	{0xE48EE60D1328A814ULL, "vxw_sundermount_20k_0_0.vlm"},
	{0xE48EFE67ABEEFD9EULL, "356974941.ogg"},
	{0xE48F0E22E32CC76AULL, "dre000_vampire6034340.cl"},
	{0xE48F57B9202BFEF3ULL, "llac_lt_0_lowtown_docks_a3_-65539_5.dds"},
	{0xE48FF2949ABD39E1ULL, "darktown_male_sickly_sta6175315.cl"},
	{0xE4900D47B3F69BBAULL, "sg_blur_dot.mao"},
	{0xE49023C4BECFAF7FULL, "812242734.ogg"},
	{0xE490314B0561DBB9ULL, "mag320_orsino6037936.cl"},
	{0xE49040976D181F45ULL, "710897038.ogg"},
	{0xE4908B3083503406ULL, "hlac_lt_0_lowtown_a3_n_c_-131071_1z.dds"},
	{0xE4919B56A4ECAA15ULL, "dre000_sibling_has_blight6037036.cl"},
	{0xE492176EF715612AULL, "vxw_sundermount_33a_0_0.vlm"},
	{0xE49242F0E5D6F456ULL, "vxht_mansion2b_1gk_0_0.vlm"},
	{0xE492625EA6DC6172ULL, "vxlt_gallowstemplar_n_3t0_0_0.vlm"},
	{0xE4927CF6428E5698ULL, "vxlt_backalley_g_1kl_0_0.vlm"},
	{0xE492DD778EA6EFAEULL, "follower_banter6136724.cl"},
	{0xE492DF8216427A54ULL, "follower_banter6143177.cl"},
	{0xE4932DD1EF884515ULL, "vxca_caverns2prc_8dj_0_0.vlm"},
	{0xE493626AE2693243ULL, "w_blightlands_170_0.msh"},
	{0xE4936C645ADA0247ULL, "hlac_w_sundermount_-196609_0x.dds"},
	{0xE4937063DA80C09DULL, "6075478_m.fxe"},
	{0xE493706AE269499DULL, "w_blightlands_170_0.mmh"},
	{0xE493746AE26950AEULL, "w_blightlands_170_0.mao"},
	{0xE493B59F289C4631ULL, "513926252.ogg"},
	{0xE493EB78E198F738ULL, "hlac_lt_undercityserialkiller_-1_8z.dds"},
	{0xE494035235003E86ULL, "53967501.ogg"},
	{0xE4951227142D090CULL, "one280_emrys6104275.cl"},
	{0xE495498490564744ULL, "vxlt_0_lowtown_a2_n_e8c_0_0.vlm"},
	{0xE4957EA1B0C50C5FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgtb_0d.dds"},
	{0xE495C2521D6587DDULL, "542289883.ogg"},
	{0xE495F6F46F80B555ULL, "386547338.ogg"},
	{0xE4961D53A0A72B12ULL, "6128483_m.fxe"},
	{0xE4963E387994DAFAULL, "6155766_m.fxe"},
	{0xE49656A3E9182299ULL, "vxlt_0_lowtown_docks_a3_n_cx6_0_0.vlm"},
	{0xE4965E820C4FBD28ULL, "refugeecharge.cut"},
	{0xE4966622B3D8CDACULL, "vxlt_undercitydungeon_1d1_0_1.vlm"},
	{0xE49683D08ED8B33EULL, "28113373.ogg"},
	{0xE496F5DA129B7940ULL, "vxht_keep_n_55l_0_0.vlm"},
	{0xE497433FB0162791ULL, "static_static_p42.psh.fxo.dx11.assembly"},
	{0xE49800479416DBE5ULL, "697205914.ogg"},
	{0xE4980612788A8ECDULL, "621030072.ogg"},
	{0xE498584E18A5E22AULL, "and100_anders6095888.cl"},
	{0xE4985FDB7E89101CULL, "gencr_mercenary_archer_f.utc"},
	{0xE4986DB116F7215FULL, "45903697.ogg"},
	{0xE49900E4FFCE5211ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medb.mmh"},
	{0xE4990CE4FFCE66B2ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medb.mao"},
	{0xE4991AE4FFCE7ECFULL, "art\\characters\\playercharacter\\humanfemale\\glv_medb.msh"},
	{0xE4993EEA0EEE28A3ULL, "863773427.ogg"},
	{0xE49953FAB343386DULL, "vis121_saemus6118471.cl"},
	{0xE499E2653947EDDEULL, "s_pool_dot.mao"},
	{0xE49A5B88F3A3474AULL, "vxlt_0_lowtown_docks_n_bqt_0_0.vlm"},
	{0xE49B5B9A3EA262A1ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_workstool_0.pwk"},
	{0xE49C7F8915B33B60ULL, "556498342.ogg"},
	{0xE49CC1A2FBA1D528ULL, "260485645.ogg"},
	{0xE49D79B02EC5C98BULL, "6174630_m.fxe"},
	{0xE49DA7A40BE6CEB1ULL, "6099245_m.fxe"},
	{0xE49E2F6D291EFC5FULL, "hlac_lt_darktown_-1_38x.dds"},
	{0xE49F2D7F506D77BFULL, "882906181.ogg"},
	{0xE49F34492AF89435ULL, "dae021cr_chantry_initiate.utc"},
	{0xE49FADA51ED0B1BCULL, "hlac_lt_0_lowtown_a3_n_c_-65535_2x.dds"},
	{0xE49FF11C97FC2A27ULL, "hlac_lt_hangedman_n_-1_20y.dds"},
	{0xE4A0025E8FD0666CULL, "vxlt_0_lowtown_docks_a3_a0n_0_0.vlm"},
	{0xE4A019436B3DECD2ULL, "vxdr_primevalentrance_176_0_0.vlm"},
	{0xE4A01E08D04D9F4EULL, "vxlt_0_lowtown_n_d1w_0_0.vlm"},
	{0xE4A092DAE996B661ULL, "vxca_caverns1_mines_jw2_0_0.vlm"},
	{0xE4A0D43B98D279F2ULL, "vxlt_gallowstemplar_n_3s1_0_0.vlm"},
	{0xE4A0ED5849777745ULL, "6168177_f.fxe"},
	{0xE4A0EED334846F18ULL, "vxlt_backalley_g_1pm_0_0.vlm"},
	{0xE4A13D23350F0BD1ULL, "321865682.ogg"},
	{0xE4A1B7391D6FC042ULL, "cr_sm.mao"},
	{0xE4A1EE41B74BB1E2ULL, "6129048_m.fxe"},
	{0xE4A1EE9D25DD5132ULL, "vxlt_darktown_5ar_0_0.vlm"},
	{0xE4A2813FBD57824BULL, "dr_0_deeproads_5839.rml"},
	{0xE4A2BE3BDAE5CD8AULL, "hlac_lt_gallowstemplar_f_2383_5x.dds"},
	{0xE4A3669B4DB166BBULL, "vrc221tr_huggin.utt"},
	{0xE4A38909D1BA7332ULL, "1058119207.ogg"},
	{0xE4A443C0C750FF9BULL, "vxlt_undercitydungeon_1ii_0_1.vlm"},
	{0xE4A47249DE6550B4ULL, "6070267_m.fxe"},
	{0xE4A47FD37E0CC4A9ULL, "6108796_m.fxe"},
	{0xE4A4DFD2885B44E2ULL, "360131813.ogg"},
	{0xE4A4E3CC8F51508BULL, "454132323.ogg"},
	{0xE4A568543E8A2829ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junkc_n.dds"},
	{0xE4A57161D7FE5BBEULL, "zz_followers_banter6159426.cl"},
	{0xE4A5868A110F0F59ULL, "vxht_brothel_n_5kf_0_2.vlm"},
	{0xE4A5A3CBCCBF8529ULL, "ico_boots_heavy_3.dds"},
	{0xE4A5B992C4567647ULL, "vxlt_0_lowtown_a2_n_fgz_0_0.vlm"},
	{0xE4A630A59007A03BULL, "650060664.ogg"},
	{0xE4A679D3EDDA5FB9ULL, "815755439.ogg"},
	{0xE4A67BCDBD9617D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_en.ani"},
	{0xE4A6D9548C1D5D5DULL, "547661178.ogg"},
	{0xE4A6F9366301017FULL, "gen_im_trs_dem_amu_r1.uti"},
	{0xE4A729860F6C4C34ULL, "6027542_m.fxe"},
	{0xE4A77AEAA533C8B4ULL, "6108483_m.fxe"},
	{0xE4A80EA81AAFCD2FULL, "6208675_f.fxe"},
	{0xE4A861707753451DULL, "hlac_w_sundermount_131074_0x.dds"},
	{0xE4A8B483A8D2E69CULL, "llac_w_sundermount_65534_0.dds"},
	{0xE4A97FD4183E26BAULL, "art\\vfx\\cinematic\\vfx_anders_childchest_c.anb"},
	{0xE4A98ABD4EEC3D8EULL, "144310393.ogg"},
	{0xE4A99DEEBA56A90BULL, "isa200im_letter.uti"},
	{0xE4A9B28CDA5CFC58ULL, "463362462.ogg"},
	{0xE4A9D3DB6C61909DULL, "mag200_feynriel6023970.cl"},
	{0xE4AA1B90AAEB6248ULL, "vxca_caverns1_mines_jw7_0_0.vlm"},
	{0xE4AB03B6F7762586ULL, "hlac_w_blightlands_65535_0y.dds"},
	{0xE4AB47689AC47F9DULL, "479198159.ogg"},
	{0xE4ABA285A22FDD6FULL, "6163506_m.fxe"},
	{0xE4ABDFCC82A99B98ULL, "lgt210_eau_symp_templar6144469.cl"},
	{0xE4AC5BEADED1F7A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hm_injured_movement.anb"},
	{0xE4AD188741DCFB1BULL, "6088780_m.fxe"},
	{0xE4AD70C32AEB28FDULL, "oil_dot_2.dds"},
	{0xE4AD78CDBD9C3136ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_en.gad"},
	{0xE4AE0E9D1C5C3F08ULL, "llac_lt_0_lowtown_-65536_5.dds"},
	{0xE4AE44E7F9B65C0EULL, "6164685_m.fxe"},
	{0xE4AE8BD70CD33CC4ULL, "hlac_ht_keep_4519_13x.dds"},
	{0xE4AE9DD0B01E0135ULL, "dae000_bartrand6207001.cl"},
	{0xE4AEF6DC9332E987ULL, "471737682.ogg"},
	{0xE4AEFF99C22595CFULL, "vxht_0_hightown_n_c15_0_0.vlm"},
	{0xE4AF53D4E298624CULL, "860832906.ogg"},
	{0xE4AFF7E11E54F8F2ULL, "llac_lt_0_lowtown_docks_a3_n_-131073_2.dds"},
	{0xE4AFFF39100B1957ULL, "6014617_m.fxe"},
	{0xE4B04BDDBDC61593ULL, "gen_im_arm_cht_lgt_set_med.uti"},
	{0xE4B0607A17D369B2ULL, "vxlt_darktown_5js_0_0.vlm"},
	{0xE4B12CEF4E7586CAULL, "156079709.ogg"},
	{0xE4B194770D442686ULL, "dae000ip_crowd_em2_talk16.utp"},
	{0xE4B22C6CE0A62DC9ULL, "960941819.ogg"},
	{0xE4B29F067084498CULL, "vxht_bartrand_n_3sg_0_0.vlm"},
	{0xE4B2C4216E09FF96ULL, "art\\vfx\\combat\\vfx_w_wns_atk_close_c.impa.ani"},
	{0xE4B2C628DEBDFE1AULL, "vxht_mansion2b_3ga_0_0.vlm"},
	{0xE4B2E426D05B6934ULL, "6075167_m.fxe"},
	{0xE4B33734E4FAB931ULL, "root_war_general1.dds"},
	{0xE4B388E3DA4D5B9CULL, "6148735_m.fxe"},
	{0xE4B38E4412901EF0ULL, "6082240_m.fxe"},
	{0xE4B393CDBDA0CA95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_weep_en.evt"},
	{0xE4B3AF33986F17DAULL, "950040570.ogg"},
	{0xE4B44C623A6C4D38ULL, "character_character_p177.psh.fxo.dx11.assembly"},
	{0xE4B548C992F3450CULL, "vxlt_gallowstemplar_n_5rn_0_0.vlm"},
	{0xE4B683D5875A46B6ULL, "477011626.ogg"},
	{0xE4B6FFBD9E770AFFULL, "ico_2h_sword_0.dds"},
	{0xE4B719827CBDE928ULL, "6076632_m.fxe"},
	{0xE4B729BB50D14FC0ULL, "hlac_lt_0_lowtown_docks_-131071_0z.dds"},
	{0xE4B8B0147953A6C9ULL, "1227935.ogg"},
	{0xE4B8BFF7E5F07198ULL, "vxlt_backalley_n_1vc_0_1.vlm"},
	{0xE4B8D5A20AD4739FULL, "254325176.ogg"},
	{0xE4B92A14ED1E896EULL, "130666901.ogg"},
	{0xE4B970684D1E0997ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.anb"},
	{0xE4B98D1CB14986AEULL, "blood.dds"},
	{0xE4B9C36DE4E6A387ULL, "887966393.ogg"},
	{0xE4BA11EB71F3FD22ULL, "vxlt_undercityencounter_13g_0_1.vlm"},
	{0xE4BAB60021A8D49BULL, "follower_banter6141578.cl"},
	{0xE4BB435CDDAD1763ULL, "art\\characters\\creatures\\mubari\\animation\\sit_loop.ani"},
	{0xE4BC4DC820A19BEEULL, "vxlt_hangedman_s8_0_0.vlm"},
	{0xE4BC62EDC931F23AULL, "fen110_anso6170065.cl"},
	{0xE4BC957443F70274ULL, "19222334.ogg"},
	{0xE4BC9CC02B4A807FULL, "hlac_ht_0_hightown_a3_-65536_3z.dds"},
	{0xE4BD0EB1BFC9AF53ULL, "6128103_m.fxe"},
	{0xE4BD9887BBB93DD6ULL, "vxht_mansion2b_1pn_0_0.vlm"},
	{0xE4BE1F021381971EULL, "163940508.ogg"},
	{0xE4BE298112945940ULL, "vxht_0_hightown_a2_b76_0_0.vlm"},
	{0xE4BE5353D9AFE531ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_2y.dds"},
	{0xE4BE8620BFA69AE0ULL, "vxht_brothel_n_359_0_0.vlm"},
	{0xE4BE9679CAD9F179ULL, "vxht_keep_n_601_0_0.vlm"},
	{0xE4BEB8B695D4F0E2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tlfrbroom_01_0_n.dds"},
	{0xE4BEE36DE03695C0ULL, "llac_dr_primevaltomb_658_12.dds"},
	{0xE4BF74573DB1DF19ULL, "si_wispy.mao"},
	{0xE4C023359DDD0B14ULL, "6079340_f.fxe"},
	{0xE4C06297EA28E4DEULL, "306086650.ogg"},
	{0xE4C07FB31B862A6FULL, "follower_banter6169859.cl"},
	{0xE4C0CC216E164A87ULL, "art\\vfx\\combat\\vfx_w_wns_atk_close_c.impa.evt"},
	{0xE4C24BD6A5E24017ULL, "follower_banter6172712.cl"},
	{0xE4C2F8CA869F391DULL, "6163148_m.fxe"},
	{0xE4C33C67A47CD766ULL, "6121088_m.fxe"},
	{0xE4C365D3E1F354E2ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.cess.gad"},
	{0xE4C36C87F763FBBDULL, "6117262_m.fxe"},
	{0xE4C3AAB39B75FB4DULL, "55067918.ogg"},
	{0xE4C3BAA684E95D8CULL, "vxlt_gallowstemplar_n_5uo_0_0.vlm"},
	{0xE4C3ED1954F4E5CAULL, "6096171_m.fxe"},
	{0xE4C42AC7BE661402ULL, "894839768.ogg"},
	{0xE4C4560574193A14ULL, "6168110_f.fxe"},
	{0xE4C45EA4891FF9B4ULL, "6106683_m.fxe"},
	{0xE4C4E3E71425E533ULL, "vxw_encounter2_camp_pb_0_1.vlm"},
	{0xE4C4E5447733AE4BULL, "918668513.ogg"},
	{0xE4C52D996A303456ULL, "vxlt_0_lowtown_a2_n_gny_0_0.vlm"},
	{0xE4C5487327138F10ULL, "6201658_m.fxe"},
	{0xE4C56608566E5BD6ULL, "w_woundedcoast_129_0.mao"},
	{0xE4C57208566E7075ULL, "w_woundedcoast_129_0.mmh"},
	{0xE4C57408566E73EBULL, "w_woundedcoast_129_0.msh"},
	{0xE4C66BF666284AA4ULL, "6162731_m.fxe"},
	{0xE4C66EBA7FF36BC3ULL, "llac_lt_undercityencounter_-1_9.dds"},
	{0xE4C6FE66D9D30EBAULL, "6065289_m.fxe"},
	{0xE4C7ED216E1CA1D0ULL, "art\\vfx\\combat\\vfx_w_wns_atk_close_c.impa.gad"},
	{0xE4C7F2C024159211ULL, "vxw_bonepitdestroyed_a3_1wk_0_0.vlm"},
	{0xE4C7FEBDC4576610ULL, "vxht_mansion2b_6s_0_0.vlm"},
	{0xE4C925E784F0E2B2ULL, "static_static_p21.psh.fxo.dx11"},
	{0xE4C988D3E1F7FBD9ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.cess.evt"},
	{0xE4C9AA75329D313AULL, "plt_lgt000pt_ultra_2_03.plo"},
	{0xE4C9B16C4760EC62ULL, "648831112.ogg"},
	{0xE4C9FC4125C92C1BULL, "6086980_m.fxe"},
	{0xE4CA496D823CFAADULL, "hlac_w_encounter2_camp_-65531_0x.dds"},
	{0xE4CA53AD78A2A73FULL, "mcr371_hostile_fenris6082216.cl"},
	{0xE4CAB7399F92C120ULL, "gen00flst_sebacttwoc.stg"},
	{0xE4CACAC366223A45ULL, "616310756.ogg"},
	{0xE4CB0746716A42FFULL, "6048227_m.fxe"},
	{0xE4CB1E1921F345B2ULL, "dae000_leonato6162493.cl"},
	{0xE4CB3AD8B72743DBULL, "479752360.ogg"},
	{0xE4CB7A3CE7681284ULL, "561942340.ogg"},
	{0xE4CBA745CDB9BADDULL, "amb_scene_kidsplay.bnk"},
	{0xE4CBAC4937049984ULL, "w_bonepit_48_0.msh"},
	{0xE4CBBE493704B8EDULL, "w_bonepit_48_0.mao"},
	{0xE4CBC2493704BF26ULL, "w_bonepit_48_0.mmh"},
	{0xE4CBF2B6566A0B68ULL, "44605641.ogg"},
	{0xE4CC60A20ACA258AULL, "vxht_mansion2_27p_0_0.vlm"},
	{0xE4CC64EFC8C62C7BULL, "6116568_m.fxe"},
	{0xE4CC9B5E048A71C0ULL, "vxht_0_hightown_a2_b87_0_0.vlm"},
	{0xE4CCC9654D47D179ULL, "dae000_hate_follower_react6103195.cl"},
	{0xE4CCE5B76422C449ULL, "vxlt_0_lowtown_n_f2g_0_0.vlm"},
	{0xE4CCE86470C9E676ULL, "hlac_lt_0_lowtown_n_-65539_0y.dds"},
	{0xE4CD14E5D2F067E2ULL, "hlac_ca_caverns2_5802_4y.dds"},
	{0xE4CE40A9453508B3ULL, "lt_gamlenshouse_n_mmi.gff"},
	{0xE4CE5641FF3CA695ULL, "751740287.ogg"},
	{0xE4CE5B1C5AEADFF8ULL, "664287748.ogg"},
	{0xE4CED1330D384ED3ULL, "6147434_m.fxe"},
	{0xE4CF5A5CDDBDE74DULL, "art\\characters\\creatures\\mubari\\animation\\sit_loop.gad"},
	{0xE4CF7F3E1D2B61CAULL, "fen210_fenris6108735.cl"},
	{0xE4CF94E5C809239BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww22_basic.msh"},
	{0xE4CF9CF5C228498DULL, "lit_up_01.mao"},
	{0xE4CFA6E5C80942E6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww22_basic.mao"},
	{0xE4CFB2E5C8095685ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww22_basic.mmh"},
	{0xE4CFCDD10227D765ULL, "l_lgt211ar_rand_sewer_1_nt.wbd"},
	{0xE4D055DC88B67111ULL, "vxht_keep_n_4hc_0_0.vlm"},
	{0xE4D0A58AC066EBEDULL, "hlac_lt_0_lowtown_n_65534_3y.dds"},
	{0xE4D175FEF4DE6A55ULL, "1062764854.ogg"},
	{0xE4D20337A4F36A85ULL, "304771119.ogg"},
	{0xE4D242C731C8F946ULL, "c_poispidra_0t.dds"},
	{0xE4D2799EE1D3C1DCULL, "6121248_m.fxe"},
	{0xE4D282D41861316DULL, "art\\vfx\\cinematic\\vfx_anders_childchest_c.mmh"},
	{0xE4D2DF2D71590A63ULL, "zz_mer_debug6039338.cl"},
	{0xE4D30275499FC83CULL, "305649605.ogg"},
	{0xE4D32AA60B0BC57CULL, "vxlt_0_lowtown_a2_n_gx9_0_0.vlm"},
	{0xE4D3510823FDB912ULL, "797322563.ogg"},
	{0xE4D384BF1748CDFDULL, "llac_lt_blackemporium_1055_3.dds"},
	{0xE4D3CA04472162F1ULL, "vxca_caverns2_8g2_0_0.vlm"},
	{0xE4D417BC003568A2ULL, "6044617_m.fxe"},
	{0xE4D45C20E359AFE7ULL, "llac_ca_caverns2prc_5802_29.dds"},
	{0xE4D4D9335E36E3BAULL, "vxlt_0_lowtown_docks_n_cs6_0_1.vlm"},
	{0xE4D511A1FB97D265ULL, "501547124.ogg"},
	{0xE4D54F135B87C753ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravee.phy"},
	{0xE4D5DCFADA8516E2ULL, "763915827.ogg"},
	{0xE4D5F30A7A004AB7ULL, "dae000cr_grand_cleric6148164.cl"},
	{0xE4D67FBC3B4D345DULL, "prm000im_isaac_gloves.uti"},
	{0xE4D6EAD8C82D5EE5ULL, "6062988_m.fxe"},
	{0xE4D74E376CDF39E5ULL, "6161609_m.fxe"},
	{0xE4D790D3E20446C8ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_p.cess.ani"},
	{0xE4D7FD1FB94C897AULL, "art\\vfx\\combat\\vfx_d_teleport_full_c.anb"},
	{0xE4D80B88D6AFA4D8ULL, "vxw_woundedcoast_19z_0_1.vlm"},
	{0xE4D8794B59F959C2ULL, "vxlt_0_lowtown_a2_n_hew_0_0.vlm"},
	{0xE4D8B40131BA7BBCULL, "vxlt_0_lowtown_docks_a3_n_cxo_0_0.vlm"},
	{0xE4D96D8B0D5A981DULL, "vxw_bonepitmining_a2_1k5_0_0.vlm"},
	{0xE4D9792363605B7FULL, "6043327_m.fxe"},
	{0xE4D9D1EA7B602320ULL, "485074910.ogg"},
	{0xE4DA26A83EA5C85DULL, "745355466.ogg"},
	{0xE4DA502397071699ULL, "143254585.ogg"},
	{0xE4DABAE3C9DF55F2ULL, "135976993.ogg"},
	{0xE4DABBDEEE0A9B0DULL, "961531354.ogg"},
	{0xE4DAD07E14AD0561ULL, "vxht_keep_62h_0_0.vlm"},
	{0xE4DAD27EFCC03E0AULL, "vxht_mansion2_20q_0_0.vlm"},
	{0xE4DAD29203E0E6EEULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_mother_victim_sis.ani"},
	{0xE4DB049916FA8E6BULL, "6016916_m.fxe"},
	{0xE4DB57945618DCC9ULL, "vxlt_0_lowtown_n_f5f_0_0.vlm"},
	{0xE4DB6FEB89874E6BULL, "6163494_m.fxe"},
	{0xE4DB8782222EF4B5ULL, "vxlt_foundry_n_w0_0_0.vlm"},
	{0xE4DB8D89275AACE7ULL, "sebcr_merc_archer_f.utc"},
	{0xE4DBAC1613559CB2ULL, "294065688.ogg"},
	{0xE4DBB17FAB24BC60ULL, "ave310_brennan6136876.cl"},
	{0xE4DC1A73EF5FF85EULL, "dm_sandal.mor"},
	{0xE4DC49E4A56DDF74ULL, "6101130_m.fxe"},
	{0xE4DCC249371332EDULL, "w_bonepit_48_0.rml"},
	{0xE4DCD658A167E7B0ULL, "260122847.ogg"},
	{0xE4DD3C9AD8A4356DULL, "hlac_lt_0_lowtown_a3_n_c_-2_1z.dds"},
	{0xE4DD5C449DD4B7F9ULL, "vxht_mansion_n_6kd_0_0.vlm"},
	{0xE4DDF2378B4E31E9ULL, "vxht_0_hightown_a3_byp_0_0.vlm"},
	{0xE4DE574C4797E73FULL, "dae000cr_grand_cleric6185501.cl"},
	{0xE4DEEBDFD752B79CULL, "vxlt_0_lowtown_a2_n_gc8_0_0.vlm"},
	{0xE4DF22DE1A1A14C5ULL, "gen00fl_fenris6126871.cl"},
	{0xE4DFF1DA4EFA8B41ULL, "489897079.ogg"},
	{0xE4E01997F340929DULL, "vxdr_primevaltomb_2ov_0_0.vlm"},
	{0xE4E040FC50338980ULL, "323631228.ogg"},
	{0xE4E059AFB86931C2ULL, "875741765.ogg"},
	{0xE4E098DB23E63DABULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationaoe.evt"},
	{0xE4E09E6068D58E8CULL, "vxlt_gallowstemplar_n_5cm_0_0.vlm"},
	{0xE4E0AE42B7C323B8ULL, "6053000_m.fxe"},
	{0xE4E0BC45DD870501ULL, "dae000_keep_maid_sta6162405.cl"},
	{0xE4E0C8D93CAA3A79ULL, "zz_otl_debug6060667.cl"},
	{0xE4E0F4301FF5E5B5ULL, "gen00fl_isabela6100040.cl"},
	{0xE4E11F88E06E5236ULL, "vxlt_0_lowtown_a2_hgv_0_0.vlm"},
	{0xE4E1688D0C57AD1DULL, "static1_d.psh.fxo"},
	{0xE4E17375F89BBB5CULL, "vxw_bonepitmining_a2_n4_0_0.vlm"},
	{0xE4E1F6B0D9FC85FDULL, "hlac_dr_primevaltomb_658_7y.dds"},
	{0xE4E1FEF173432DFDULL, "6080575_m.fxe"},
	{0xE4E22490817EC85DULL, "231992128.ogg"},
	{0xE4E26D684D410998ULL, "art\\vfx\\environment\\vfx_trap_fire_jet_v_p.mmh"},
	{0xE4E274622648584AULL, "6089741.cut"},
	{0xE4E2AAA0CD3DDBBCULL, "framebuffereffects_smallblur_p0.vsh.fxo"},
	{0xE4E2B666E4C499C1ULL, "vxlt_gallowscourtyard_n_6k5_0_0.vlm"},
	{0xE4E2B9357DBF92DDULL, "c_giaspidra_0t.dds"},
	{0xE4E30FEBD787BB33ULL, "6039300_m.fxe"},
	{0xE4E3331F29BE97F7ULL, "lt_0_lowtown_a2_37_0_lc.mmh"},
	{0xE4E383EDB6ED4BF3ULL, "198453580.ogg"},
	{0xE4E3C1123FBCF383ULL, "vrc321_varric6154750.cl"},
	{0xE4E3C8A47AC7B6F8ULL, "weapon_rendertoshadowmap_p0.vsh.fxo"},
	{0xE4E45933D601B148ULL, "vxht_keep_62m_0_0.vlm"},
	{0xE4E48DE42411263CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.impa.ani"},
	{0xE4E63BE1D3668028ULL, "w_woundedcoast_142_0.mmh"},
	{0xE4E647E1D36694CBULL, "w_woundedcoast_142_0.mao"},
	{0xE4E659E1D366B352ULL, "w_woundedcoast_142_0.msh"},
	{0xE4E6F9CBBC298B81ULL, "vxw_bonepitmining_a2_dg_0_0.vlm"},
	{0xE4E70F332BBC5B8BULL, "vxht_0_hightown_c5r_0_0.vlm"},
	{0xE4E7206D87208B26ULL, "qun200_party_comment6095375.cl"},
	{0xE4E78C57D3D0A394ULL, "vxlt_0_lowtown_a2_n_c_eoi_0_1.vlm"},
	{0xE4E7B1DB23EC876CULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationaoe.gad"},
	{0xE4E7CAB83A61B69BULL, "659203105.ogg"},
	{0xE4E7D6843B067913ULL, "and230_mage_girl6135811.cl"},
	{0xE4E821A4B0045052ULL, "6209890_m.fxe"},
	{0xE4E886CA32C3C9FBULL, "hla_ht_0_hightown_a3_0_0x.dds"},
	{0xE4E88F73E7C0B390ULL, "vxht_0_hightown_a3_bbt_0_0.vlm"},
	{0xE4E945CCEBD7CACCULL, "ico_maul_2.dds"},
	{0xE4E955B1D0D1A78CULL, "art\\characters\\weapons\\maces\\w_mce_001a.msh"},
	{0xE4E95BB1D0D1B27EULL, "art\\characters\\weapons\\maces\\w_mce_001a.mmh"},
	{0xE4E967B1D0D1C615ULL, "art\\characters\\weapons\\maces\\w_mce_001a.mao"},
	{0xE4E9EEBDA4C65459ULL, "vxdr_0_deeproads_cug_0_0.vlm"},
	{0xE4E9FA3D2C918A3BULL, "307631445.ogg"},
	{0xE4EB3BC7795D2240ULL, "6105930_m.fxe"},
	{0xE4EB9AE424175B7EULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.impa.gad"},
	{0xE4EBCD107CCDBD97ULL, "16443269.ogg"},
	{0xE4EBCE218FCAD079ULL, "vxht_mansion_n_6nk_0_0.vlm"},
	{0xE4EC1EA405580076ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_act_c.impa.evt"},
	{0xE4EC2239B2EB420AULL, "6160674_m.fxe"},
	{0xE4ECD333D1996C81ULL, "vxw_bonepit_1wj_0_1.vlm"},
	{0xE4ED6B2499E0C94FULL, "plt_prm_cod_irons_belt.plo"},
	{0xE4ED6DAF8AD461F2ULL, "character_character_p111.psh.fxo"},
	{0xE4ED9E3ED1E3B6D7ULL, "6041403_f.fxe"},
	{0xE4EDE4388AE190D3ULL, "vxht_0_hightown_b8t_0_0.vlm"},
	{0xE4EE36AB6A524784ULL, "6095594_m.fxe"},
	{0xE4EE9010766DACB6ULL, "1216925.ogg"},
	{0xE4EEA36E03127802ULL, "1034937238.ogg"},
	{0xE4EEA708E4B87F51ULL, "hlac_lt_0_lowtown_a2_n_c_131070_1z.dds"},
	{0xE4EEAB9203F14D28ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_mother_victim_sis.gad"},
	{0xE4EEB0DB23F2A46AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\devastationaoe.ani"},
	{0xE4EED5391FE83C3CULL, "da2terrain_terrain_p4.psh.fxo.dx11.assembly"},
	{0xE4EFF70C069D133CULL, "1021265179.ogg"},
	{0xE4F0AABE2E223FEDULL, "6187836_m.fxe"},
	{0xE4F0E1933FDD5D15ULL, "113110752.ogg"},
	{0xE4F17D84786027C1ULL, "37550065.ogg"},
	{0xE4F1A9034AFC1BE8ULL, "6149642_m.fxe"},
	{0xE4F1C9F70E673171ULL, "gen_im_acc_rng_lore_03.uti"},
	{0xE4F1D082581A48FCULL, "pro000_aveline_intro6109384.cl"},
	{0xE4F215A4055C5C01ULL, "art\\vfx\\combat\\vfx_c_2nd_buff_act_c.impa.gad"},
	{0xE4F275E4241D3A9DULL, "art\\vfx\\combat\\vfx_glb_com_brittle_s2_c.impa.evt"},
	{0xE4F2E8BB1BFE88C2ULL, "vxht_0_hightown_a3_bky_0_0.vlm"},
	{0xE4F32F06A103B789ULL, "vxht_0_hightown_n_5wg_0_0.vlm"},
	{0xE4F34856F9F7DB91ULL, "vxw_bonepitdestroyed_a3_1fh_0_0.vlm"},
	{0xE4F398B065F8F294ULL, "6202080_m.fxe"},
	{0xE4F4EF403A6DF642ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured.gad"},
	{0xE4F56BA8AE1FA401ULL, "vxw_bonepitmining_a2_cf_0_0.vlm"},
	{0xE4F5B9B97BBD4335ULL, "ht_mansion_n_4152_lc.mmh"},
	{0xE4F5EFE644E3ACB4ULL, "vxdr_primevaltomb_6r4_0_0.vlm"},
	{0xE4F60870972EA10BULL, "dae200_saemus6141202.cl"},
	{0xE4F6C15213A5636BULL, "crt_skeleton.wbd"},
	{0xE4F6D1796841816EULL, "vxht_mansion2_2gd_0_0.vlm"},
	{0xE4F70D98B826347BULL, "gen00fl_fenris6103339.cl"},
	{0xE4F7521B20A6F68EULL, "vxca_cavern1_haunted_iap_0_0.vlm"},
	{0xE4F764EE4C362F52ULL, "vxht_0_hightown_b8q_0_0.vlm"},
	{0xE4F79B565988205AULL, "genip_pp_dr_standing.utp"},
	{0xE4F7AFBD72B890F0ULL, "mcr371_alain6087129.cl"},
	{0xE4F7C68FB7E49A66ULL, "38637184.ogg"},
	{0xE4F8071FE461C018ULL, "6032261_m.fxe"},
	{0xE4F8609A96BC6CD9ULL, "vxdr_0_deeproads_crf_0_0.vlm"},
	{0xE4F91147DF1A2BB7ULL, "qun120_arishok_guard6054847.cl"},
	{0xE4F916A405627CAFULL, "art\\vfx\\combat\\vfx_c_2nd_buff_act_c.impa.ani"},
	{0xE4F93CFA601E6D42ULL, "209810040.ogg"},
	{0xE4F97F1EA050E9FDULL, "llac_lt_0_lowtown_docks_-1_1.dds"},
	{0xE4F99E2B283A96B4ULL, "mrl000_marethari6129423.cl"},
	{0xE4F9E95472A8FFCCULL, "genip_lever.utp"},
	{0xE4FA06E42D47CBE5ULL, "vxht_0_hightown_a2_b0n_0_0.vlm"},
	{0xE4FA0FBFB5995A09ULL, "6108621_f.fxe"},
	{0xE4FA3E8032C3012BULL, "mcr371cr_aveline_cut.utc"},
	{0xE4FACE66A6FE2A27ULL, "vxht_bartrand_n_5sb_0_0.vlm"},
	{0xE4FAD5F16F3A62E9ULL, "vxht_0_hightown_a3_bsv_0_0.vlm"},
	{0xE4FBF26DAC19DC17ULL, "234796899.ogg"},
	{0xE4FC0A0917A52DC4ULL, "vxlt_warehouse_a4_0_0.vlm"},
	{0xE4FC12403A745039ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured.evt"},
	{0xE4FC8858ADB89E8AULL, "hlac_ht_playershouse_n_855_3z.dds"},
	{0xE4FCD325D15E94A5ULL, "pro000_colonel_ewald6040577.cl"},
	{0xE4FCDA57605CFBC0ULL, "zz_dae_debug6047589.cl"},
	{0xE4FD652FD1405BCEULL, "vrc320ip_at_mansion.ncs"},
	{0xE4FD87133C2E9783ULL, "w_encounter3_242_0.msh"},
	{0xE4FD9170A3A9C7DAULL, "art\\vfx\\combat\\vfx_r_arc_sgrenade_act_c.anb"},
	{0xE4FD95133C2EAFDDULL, "w_encounter3_242_0.mmh"},
	{0xE4FD99133C2EB6EEULL, "w_encounter3_242_0.mao"},
	{0xE4FDA7641F40A575ULL, "dae000_gen_turn_in_56178426.cl"},
	{0xE4FDDFB39C95A094ULL, "6173089_m.fxe"},
	{0xE4FE21FF8B7B4D6CULL, "701863636.ogg"},
	{0xE4FE4B830C9F2EB9ULL, "dae230ar_undercity_darktown.lst"},
	{0xE4FE78A05E3CF865ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.mmh"},
	{0xE4FEF7ECF54DA3F9ULL, "572046496.ogg"},
	{0xE4FF0AF6B04B453AULL, "rdr111dg_martin6055004.cl"},
	{0xE4FF1F9B1D03AC71ULL, "vxca_caverns2_8t7_0_0.vlm"},
	{0xE4FF70654C7A91FAULL, "vxlt_gallowscourtyard_n_7im_0_0.vlm"},
	{0xE4FFCB1B1D7B4CCBULL, "zz_vault_debug6178950.cl"},
	{0xE500052EB8E00A69ULL, "6177832_f.fxe"},
	{0xE5002E0FA7D9554FULL, "fxa_cen_aoe_coulds.mao"},
	{0xE5004664EB4A2254ULL, "666081588.ogg"},
	{0xE501001FB96F932DULL, "art\\vfx\\combat\\vfx_d_teleport_full_c.mmh"},
	{0xE5015A980DF4A142ULL, "vxht_0_hightown_a3_bfz_0_0.vlm"},
	{0xE5015D841C5A0539ULL, "vxdr_0_deeproads_9cg_0_0.vlm"},
	{0xE501ACFB9530022AULL, "22266341.ogg"},
	{0xE501AF13EC3F3BA5ULL, "vxw_bonepitmining_a2_ex_0_0.vlm"},
	{0xE50211519AABE541ULL, "vxht_chantry_28f_0_0.vlm"},
	{0xE50238AA092778D8ULL, "vxlt_0_lowtown_a2_n_eho_0_0.vlm"},
	{0xE502B8BFBC922517ULL, "hlac_w_bonepit_196613_0y.dds"},
	{0xE502E2FB3E17C47AULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_skeleton.msh"},
	{0xE502F0FB3E17DBB3ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_skeleton.mao"},
	{0xE502F4FB3E17E2C0ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_skeleton.mmh"},
	{0xE5032CD3BDBC801FULL, "art\\levels\\races\\natural\\proxy\\foliage\\tre_zzpine.mmh"},
	{0xE50332D3BDBC8AC1ULL, "art\\levels\\races\\natural\\proxy\\foliage\\tre_zzpine.msh"},
	{0xE503DD85A015BC81ULL, "vxw_bonepitmining_a2_ne_0_0.vlm"},
	{0xE50443830CA38C90ULL, "dae230ar_undercity_darktown.ncs"},
	{0xE5045BD4EC9675BBULL, "vxw_bonepitmining_a2_5f_0_0.vlm"},
	{0xE504FE24CB6D755FULL, "61148395.ogg"},
	{0xE50537E55C93BF89ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerb.mao"},
	{0xE50543E55C93D422ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerb.mmh"},
	{0xE50545E55C93D7D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_towerb.msh"},
	{0xE5056A41880CFB5BULL, "vxlt_gallowstemplar_f_5r5_0_0.vlm"},
	{0xE505FF9CB8A33F44ULL, "6207771_m.fxe"},
	{0xE506894E1FB4BE82ULL, "gen_im_trs_rdr_junk1_04.uti"},
	{0xE506A29EB79C1766ULL, "466071091.ogg"},
	{0xE507980F1AD58F04ULL, "6098801_m.fxe"},
	{0xE507B2F4675437B0ULL, "vxht_0_hightown_n_c70_0_0.vlm"},
	{0xE5081E87B54C490EULL, "6135113_m.fxe"},
	{0xE5085B0893406817ULL, "zz_dae_debug6029315.cl"},
	{0xE50882140398576DULL, "vxw_sundermount_377_0_1.vlm"},
	{0xE508AA974DA65349ULL, "vxdr_primevaltomb_5aw_0_0.vlm"},
	{0xE508E46CEBF54673ULL, "75342576.ogg"},
	{0xE50914477DCA602CULL, "vxlt_0_lowtown_docks_a3_n_d22_0_0.vlm"},
	{0xE5091A403A7EE828ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured.ani"},
	{0xE50947CE61307B69ULL, "vxht_0_hightown_a3_bnu_0_0.vlm"},
	{0xE50A2C26B4F33501ULL, "6115800_m.fxe"},
	{0xE50AF41993B55621ULL, "vxht_0_hightown_b7g_0_0.vlm"},
	{0xE50B1CBA9D30CE82ULL, "dae000_arishok6078432.cl"},
	{0xE50B87BD7AB5A85DULL, "vxlt_hangedman_176_0_0.vlm"},
	{0xE50BBC08719FFFD4ULL, "vxlt_warehouse_n_13d_0_0.vlm"},
	{0xE50C38846F2013F3ULL, "558164750.ogg"},
	{0xE50C75A05E493032ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.anb"},
	{0xE50C763E7990EC2AULL, "6199331_m.fxe"},
	{0xE50C7B9FFDD7678DULL, "439376386.ogg"},
	{0xE50CB0188AED9BBCULL, "ss_wardrobe.cut"},
	{0xE50CCE475B796350ULL, "vxca_caverns2prc_ae1_0_0.vlm"},
	{0xE50D8FFF2D2083C1ULL, "fen110_fenris6117971.cl"},
	{0xE50DAC4C62A98B64ULL, "6129059_m.fxe"},
	{0xE50DE5D617B290C8ULL, "zz_pro_debug6203155.cl"},
	{0xE50F251E28D8ADA1ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_ian.ani"},
	{0xE50F257FFB890B49ULL, "llac_lt_darktown_-1_90.dds"},
	{0xE50F26F18BDAAB0FULL, "6077940_m.fxe"},
	{0xE50F44830CAD967EULL, "dae230ar_undercity_darktown.are"},
	{0xE50F5D61C96027D0ULL, "w_bonepit_3_0.rml"},
	{0xE50F5D6A8C80BD97ULL, "vxca_cave_t2_0_0.vlm"},
	{0xE50F5F31612593C9ULL, "6077133_m.fxe"},
	{0xE50FA7D1025D7A0FULL, "l_lgt211ar_rand_sewer_1_nt.bnk"},
	{0xE50FCF610E501DB9ULL, "vxdr_0_deeproads_9df_0_0.vlm"},
	{0xE5102C10DDE40C91ULL, "vxw_bonepitdestroyed_a3_1ln_0_0.vlm"},
	{0xE51077496A82E6D6ULL, "53513458.ogg"},
	{0xE510AA86FB1D9158ULL, "vxlt_0_lowtown_a2_n_eml_0_0.vlm"},
	{0xE511159AE370EC99ULL, "darktown_male_sickly_sta6175328.cl"},
	{0xE511AAB0E673D093ULL, "131294386.ogg"},
	{0xE511E8CCDE4C494FULL, "vxca_caverns2_73k_0_0.vlm"},
	{0xE512352557A310D4ULL, "vxht_0_hightown_a2_9bb_0_0.vlm"},
	{0xE512406B6E1FA9DDULL, "lt_0_lowtown_docks_a3_n_123_0.phy"},
	{0xE51243FD7BC7DB46ULL, "ts_dwarven_arm_props.gda"},
	{0xE513064FABD537B9ULL, "dae000_orsino.bnk"},
	{0xE513AD0D6131E1E9ULL, "hlac_lt_0_lowtown_a2_n_131070_6y.dds"},
	{0xE513F2FB542F8763ULL, "vxht_mansion2b_3dq_0_0.vlm"},
	{0xE51411AD46C25D95ULL, "vxdr_primevaltomb_3i1_0_0.vlm"},
	{0xE514200294798441ULL, "qcr220_orsino6061795.cl"},
	{0xE51433788E9BE12BULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\pose_a.evt"},
	{0xE5146D3182656CB0ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_vial_n.dds"},
	{0xE5147E21C78D7BF5ULL, "vxlt_0_lowtown_eqr_0_0.vlm"},
	{0xE51494913D82F16DULL, "696727233.ogg"},
	{0xE514DB3459382804ULL, "face_face_p5.psh.fxo.dx11"},
	{0xE5157FCD7E5914C3ULL, "394764721.ogg"},
	{0xE5160DB30AD47EF5ULL, "gen00fl_merrill6130069.cl"},
	{0xE516284FABD7A719ULL, "dae000_orsino.cnv"},
	{0xE516301E28DEDF43ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_ian.gad"},
	{0xE5166E348537BBE7ULL, "vxht_0_hightown_n_c37_0_0.vlm"},
	{0xE516746CE91731C4ULL, "dae000_arishok6067690.cl"},
	{0xE5168BE5C8451732ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0ww22_basic.phy"},
	{0xE516F3F0F58E6FEDULL, "vxw_sundermount_324_0_1.vlm"},
	{0xE5171D378F3D5DF5ULL, "436720997.ogg"},
	{0xE5179E3132EC56DCULL, "443514445.ogg"},
	{0xE51856EB6CB196F6ULL, "110067289.ogg"},
	{0xE5189E03A6E94D81ULL, "6110700_m.fxe"},
	{0xE51A1C4E984D03BEULL, "63444478.ogg"},
	{0xE51A2DA89880F64AULL, "zz_pro_debug6196169.cl"},
	{0xE51A2DE563961854ULL, "vxlt_warehouse_n_18e_0_0.vlm"},
	{0xE51AA3B0C6390D97ULL, "hlac_dr_dungeon_3215_1y.dds"},
	{0xE51B4C788EA22AECULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\pose_a.gad"},
	{0xE51BA031BD22821CULL, "vxlt_0_lowtown_a2_n_bxt_0_0.vlm"},
	{0xE51BA5A44A5060B6ULL, "6136372_m.fxe"},
	{0xE51BCFCCAF8209EBULL, "vxca_cave_18b_0_1.vlm"},
	{0xE51CA18F3DF240F6ULL, "seb221_flora_harimann6177664.cl"},
	{0xE51CA35213C5FF39ULL, "crt_skeleton.bnk"},
	{0xE51D0615916375B9ULL, "vxlt_0_lowtown_n_fef_0_0.vlm"},
	{0xE51DD4C4464B17A8ULL, "gen00fl_fenris6126397.cl"},
	{0xE51E7140441C3FDBULL, "gen_im_trs_gen_lgt_boo_r0.uti"},
	{0xE51EF5390E5472C9ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man3_.ani"},
	{0xE51F1BB0C920D7BCULL, "llac_lt_0_lowtown_a2_n_c_-65535_7.dds"},
	{0xE51F58AE86C59D14ULL, "vxw_sundermount_2fq_0_0.vlm"},
	{0xE51FD60B95AD7D5EULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3_0.pwk"},
	{0xE51FD7C8E764762AULL, "177533318.ogg"},
	{0xE51FDD53D3D3CF42ULL, "196146946.ogg"},
	{0xE520A3375D0BF66CULL, "146188985.ogg"},
	{0xE520D9B5709C6BD9ULL, "6029044_m.fxe"},
	{0xE52109F2539E71D4ULL, "w_woundedcoast_101_0.phy"},
	{0xE5211BFDAAF0A442ULL, "vxlt_shop_4f_0_0.vlm"},
	{0xE52163EE713DBECCULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dazed.ani"},
	{0xE5221B7021680622ULL, "qcr220_grey_wardens6207460.cl"},
	{0xE522466EE68857B9ULL, "hlac_lt_0_lowtown_n_-3_6y.dds"},
	{0xE5224B788EA847EAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\pose\\pose_a.ani"},
	{0xE5224DFB6BF92C5BULL, "vxlt_gallowstemplar_f_5x7_0_0.vlm"},
	{0xE5229B07200A8A48ULL, "rog_rainofarrows.dds"},
	{0xE5229C716294A985ULL, "static_passthrough_p0.vsh.fxo"},
	{0xE522EEC2CADD1A74ULL, "vxlt_hangedman_n_10y_0_0.vlm"},
	{0xE523178BE0173C3CULL, "860506252.ogg"},
	{0xE523356B6E2E0A49ULL, "lt_0_lowtown_docks_a3_n_123_0.mao"},
	{0xE523416B6E2E1EE2ULL, "lt_0_lowtown_docks_a3_n_123_0.mmh"},
	{0xE523436B6E2E2198ULL, "lt_0_lowtown_docks_a3_n_123_0.msh"},
	{0xE5242B62427E3267ULL, "878334156.ogg"},
	{0xE524512C8018572FULL, "vxca_caverns2_9ff_0_0.vlm"},
	{0xE524B11380AC475EULL, "vxdr_0_deeproads_ew4_0_0.vlm"},
	{0xE524E011772DD467ULL, "vxht_0_hightown_n_c86_0_0.vlm"},
	{0xE52510390E590C4BULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man3_.gad"},
	{0xE52552B658B6F97BULL, "416657633.ogg"},
	{0xE525553F9C16BBFAULL, "919774133.ogg"},
	{0xE52561AB21F717E3ULL, "mag291_vincento6027345.cl"},
	{0xE5259470A3CB1E8DULL, "art\\vfx\\combat\\vfx_r_arc_sgrenade_act_c.mmh"},
	{0xE5260E0FA9095418ULL, "625217731.ogg"},
	{0xE5263E77DC1DA9D9ULL, "dae000_madam_lusine6151624.cl"},
	{0xE5264D6316F4E2CBULL, "6084480_f.fxe"},
	{0xE5274EF188A85DE9ULL, "814735069.ogg"},
	{0xE527BA26F517C3B1ULL, "19802722.ogg"},
	{0xE527C340E9B6AAA3ULL, "gen00fl_isabela6015692.cl"},
	{0xE527E54A1CB34D19ULL, "6063849_f.fxe"},
	{0xE52803FB890B4CD0ULL, "dre000_sibling_has_blight6036560.cl"},
	{0xE52870EE7143F30EULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dazed.gad"},
	{0xE528A03BC772FC45ULL, "vxht_mansion2_3gr_0_0.vlm"},
	{0xE52933ED2D116522ULL, "background_defaults.gda"},
	{0xE529455DE4D1427DULL, "847714323.ogg"},
	{0xE52A32B1D1090319ULL, "art\\characters\\weapons\\maces\\w_mce_001a.phy"},
	{0xE52ACB01D593FE1DULL, "6015847_m.fxe"},
	{0xE52B2A3CD6EDE35DULL, "vxht_mansion_n_502_0_0.vlm"},
	{0xE52B73A4C11520CBULL, "vxlt_0_lowtown_docks_n_cuh_0_0.vlm"},
	{0xE52B97274772A16AULL, "vxht_mansion_n_5a6_0_0.vlm"},
	{0xE52BD2274F848FC2ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.cess.gad"},
	{0xE52BE6F253A83DD1ULL, "w_woundedcoast_101_0.msh"},
	{0xE52BEEBD69A7C302ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wrba_0t.dds"},
	{0xE52BF4F253A855ECULL, "w_woundedcoast_101_0.mao"},
	{0xE52C00F253A8698FULL, "w_woundedcoast_101_0.mmh"},
	{0xE52C3A32AFDF868FULL, "69667465.ogg"},
	{0xE52C5F825014ACCBULL, "vxht_mansion2b_1kv_0_0.vlm"},
	{0xE52C96FA8F5BB59BULL, "vxlt_0_lowtown_eux_0_0.vlm"},
	{0xE52CF36DF350C1BDULL, "6087850_f.fxe"},
	{0xE52CFAFBBC8FB137ULL, "vxlt_0_lowtown_n_ens_0_1.vlm"},
	{0xE52D04AAC2218525ULL, "vxw_bonepitmining_a2_je_0_0.vlm"},
	{0xE52D707CEECA9B78ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomea_d.dds"},
	{0xE52DA8EF61950B3FULL, "512182824.ogg"},
	{0xE52DCA8B78BBB594ULL, "vxw_sundermount_2kv_0_0.vlm"},
	{0xE52DFA5B0CEB122DULL, "740591029.ogg"},
	{0xE52E0465FDDB6E0CULL, "gen00fl_fenris6103356.cl"},
	{0xE52E1638364A1C9FULL, "637758890.ogg"},
	{0xE52E6BEE714856EDULL, "art\\characters\\creatures\\shriek\\animation\\combat\\dazed.evt"},
	{0xE52EB8721237D47BULL, "6056835_m.fxe"},
	{0xE52EDEEBD989D437ULL, "vxlt_0_lowtown_a2_n_c_deh_0_0.vlm"},
	{0xE52F4337EBE3EF43ULL, "6086171_m.fxe"},
	{0xE52F5794039689ACULL, "818123007.ogg"},
	{0xE52FC7E69AE266B7ULL, "6053090_m.fxe"},
	{0xE52FE30F99F495B9ULL, "smoke_ciga_f4.dds"},
	{0xE530A66B0B019ACEULL, "6178338_m.fxe"},
	{0xE530B60EB14E8F77ULL, "hlac_lt_0_lowtown_docks_n_-65538_5y.dds"},
	{0xE530BFD85DEF44DBULL, "vxlt_gallowstemplar_f_5c6_0_0.vlm"},
	{0xE530CD967A762190ULL, "116171058.ogg"},
	{0xE5312907574911BFULL, "hlac_ca_cavern1_haunted_24423_13z.dds"},
	{0xE531C770300374DDULL, "6017764_m.fxe"},
	{0xE5320895169C8AAFULL, "785771509.ogg"},
	{0xE5321E97322587DBULL, "hlac_ht_keep_4519_7x.dds"},
	{0xE53280B5017AAFB4ULL, "127655361.ogg"},
	{0xE5328FDFCD22E34EULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_fin_c.impa.evt"},
	{0xE532C2C49924F2FBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_ornate.msh"},
	{0xE532D4C499251146ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_ornate.mao"},
	{0xE532E0C4992525E5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_ornate.mmh"},
	{0xE532F5274F8AE9B9ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.cess.evt"},
	{0xE53346444EEE603FULL, "gen00fl_varric6076440.cl"},
	{0xE53396E7900EC721ULL, "hlac_lt_0_lowtown_docks_a3_n_-65535_0z.dds"},
	{0xE534A0FB3378D972ULL, "157558218.ogg"},
	{0xE5350B2B0D2B958AULL, "6128072_m.fxe"},
	{0xE53546CB8AD7DA65ULL, "384635556.ogg"},
	{0xE536F1FEAC6A16DAULL, "mcr371_angry_mage_m6087027.cl"},
	{0xE53735A5BC0448A4ULL, "246282783.ogg"},
	{0xE53736911E48FD88ULL, "695363091.ogg"},
	{0xE53771C314A62C5AULL, "920400799.ogg"},
	{0xE5383B0D4034D94DULL, "llac_lt_0_lowtown_65535_6.dds"},
	{0xE53848135BDC38C8ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravee.mmh"},
	{0xE53854135BDC4D6BULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravee.mao"},
	{0xE53866135BDC6BF2ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_gravee.msh"},
	{0xE5387809A01BD2BFULL, "art\\characters\\playercharacter\\dwarfmale\\boo_meda.msh"},
	{0xE5387E09A01BDD61ULL, "art\\characters\\playercharacter\\dwarfmale\\boo_meda.mmh"},
	{0xE538FE4AA071CE0EULL, "hlac_ht_mansion2b_818_0x.dds"},
	{0xE53986DFCD28F2F9ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_fin_c.impa.gad"},
	{0xE539C497B8244A6DULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_eye_cb.mao"},
	{0xE539E60D72610001ULL, "hlac_lt_gallowscourtyard_-196608_0z.dds"},
	{0xE53A09043968B9EAULL, "vxht_mansion_n_5j7_0_0.vlm"},
	{0xE53A0B1EBD40AF0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustain.evt"},
	{0xE53A1B4473D5DA64ULL, "vxht_bartrand_n_5vr_0_0.vlm"},
	{0xE53A33C75ED994D1ULL, "art\\levels\\races\\design\\prp_silverite.mmh"},
	{0xE53A3FC75ED9A972ULL, "art\\levels\\races\\design\\prp_silverite.mao"},
	{0xE53A4DC75ED9C08FULL, "art\\levels\\races\\design\\prp_silverite.msh"},
	{0xE53A758552C9B688ULL, "339031018.ogg"},
	{0xE53AA6E556310B92ULL, "vxdr_primevaltomb_2xd_0_0.vlm"},
	{0xE53AC029FEC02476ULL, "obliskbasecolor.dds"},
	{0xE53B0FD76F50157FULL, "243762429.ogg"},
	{0xE53B4EB8CEB8A708ULL, "hlac_lt_gallowscourtyard_-65538_0x.dds"},
	{0xE53B6CD8AE85C9B7ULL, "vxlt_0_lowtown_n_ekl_0_1.vlm"},
	{0xE53BC33AC1595048ULL, "mag360_samson6044866.cl"},
	{0xE53BF8839D607FB7ULL, "1041737196.ogg"},
	{0xE53C117CEA9D84E6ULL, "vxht_0_hightown_a2_bp9_0_0.vlm"},
	{0xE53C3C686AB1CE14ULL, "vxw_sundermount_2lw_0_0.vlm"},
	{0xE53C78B47ECC5DBBULL, "art\\vfx\\textures\\vfx_crsh_pris.dds"},
	{0xE53CBDDA21C86D22ULL, "6049706_m.fxe"},
	{0xE53CF8F82DEC19D9ULL, "vxht_chantry_2vd_0_0.vlm"},
	{0xE53CFDC3FF168A60ULL, "qun221_petrice6077983.cl"},
	{0xE53D3F34A95EEC4EULL, "vxlt_gallowstemplar_3rm_0_0.vlm"},
	{0xE53D57C53E388504ULL, "vxlt_0_lowtown_docks_n_cs8_0_1.vlm"},
	{0xE53D85AE1881DDE0ULL, "294579918.ogg"},
	{0xE53E09C91C7F9914ULL, "apr_base.gda"},
	{0xE53E35B3D7AF5DEAULL, "vxw_bonepitdestroyed_a3_25m_0_0.vlm"},
	{0xE53E9A08B81E447CULL, "forcefeedback_abilities.gda"},
	{0xE53EB5983AB1FF94ULL, "seb201_lady_harimann6177387.cl"},
	{0xE53EE2150C2A7FF2ULL, "433673185.ogg"},
	{0xE53EF7E7CA5792A0ULL, "zz_fol_debug6066200.cl"},
	{0xE53EFDD82C72670CULL, "gen00fl_aveline6114043.cl"},
	{0xE53EFE184B1A2F2BULL, "6090391_f.fxe"},
	{0xE53F97DFCD2D7B47ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_fin_c.impa.ani"},
	{0xE53FFD274F9581A8ULL, "art\\vfx\\combat\\vfx_r_gn3_hbomb_tar_loop_c.cess.ani"},
	{0xE540CD8046A82664ULL, "llac_dr_0_deeproads_5843_4.dds"},
	{0xE540F9CAECD4C408ULL, "vxht_mansion2_24w_0_0.vlm"},
	{0xE541021EBD46BEB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustain.gad"},
	{0xE5411C22A51C41D8ULL, "6122151_m.fxe"},
	{0xE541285D3CBC5FC2ULL, "darktown_male_lowlife_mob6174415.cl"},
	{0xE54157DE6EF991EEULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforward.ani"},
	{0xE541BCBEF84E411FULL, "698091112.ogg"},
	{0xE541BDD235B687BEULL, "ht_0_hightown_33_0.rml"},
	{0xE54221E5FB3750FCULL, "361845894.ogg"},
	{0xE5424FBD3477DF38ULL, "vxht_mansion_n_4ah_0_1.vlm"},
	{0xE542A51EF9C36398ULL, "llac_lt_0_lowtown_-131072_6.dds"},
	{0xE54338DCAB2932B7ULL, "6176752_f.fxe"},
	{0xE54384C80A3D077CULL, "llac_w_encounter2_camp_131070_0.dds"},
	{0xE5439CD3B3BA3CF6ULL, "vxlt_gallowscourtyard_6xh_0_0.vlm"},
	{0xE543AACCD316D8D9ULL, "llac_lt_hangedman_n_-1_1.dds"},
	{0xE543D726E952D37DULL, "25347804.ogg"},
	{0xE543F5FA1EB21B32ULL, "ser121_jethann6108806.cl"},
	{0xE543FF8A8BB3DC9AULL, "hlac_lt_undercityserialkiller_-1_8x.dds"},
	{0xE54480BC512A7669ULL, "500319472.ogg"},
	{0xE5448A75018BA029ULL, "gen_im_arm_cht_med_act1_02.uti"},
	{0xE544E05EBDD236A2ULL, "vxlt_gallowstemplar_f_56d_0_0.vlm"},
	{0xE54512D9CF885AFEULL, "6218661_m.fxe"},
	{0xE54525DF0DF1CA09ULL, "hlac_ca_caverns1_mines_24425_11z.dds"},
	{0xE5453029CB834B2FULL, "6059395_m.fxe"},
	{0xE5458FFB05F197E5ULL, "vxdr_primevaltomb_3ia_0_0.vlm"},
	{0xE5466786D8E156CDULL, "843665579.ogg"},
	{0xE5468E91C41F276AULL, "6052773_m.fxe"},
	{0xE546A5F7E944BD2CULL, "6060198_m.fxe"},
	{0xE546D13A444001B4ULL, "154028945.ogg"},
	{0xE54743622B06501DULL, "vxlt_0_lowtown_a2_n_c_f5l_0_0.vlm"},
	{0xE547D17222F53B11ULL, "vxdr_primevaltomb_8vo_0_0.vlm"},
	{0xE547F96CF8CC55A6ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.impa.ani"},
	{0xE5480806F7BF89D1ULL, "gen_im_wep_war_shd_act3_06.uti"},
	{0xE548131EBD4CFA07ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\devastator\\circlehighmage\\combat\\invulnerablesustain.ani"},
	{0xE5488D2165CBF2E4ULL, "vxht_bartrand_n_5ys_0_0.vlm"},
	{0xE548E6CAFA2D8AFEULL, "llac_lt_0_lowtown_n_-196609_0.dds"},
	{0xE548F65FFCE3313DULL, "6112733_m.fxe"},
	{0xE548F7DA4BB26A24ULL, "ser211cr_demon_mharen.utc"},
	{0xE5492C7C10448767ULL, "vxw_woundedcoast_dz_0_0.vlm"},
	{0xE54AE06B9F652230ULL, "follower_banter6163531.cl"},
	{0xE54B126AB61DDF43ULL, "hla_ht_0_hightown_65538_0x.dds"},
	{0xE54BDDCECC828CDDULL, "face_face_p8.psh.fxo"},
	{0xE54C04387ABD1A3CULL, "loginnewaccount.gfx"},
	{0xE54C303C2484EA83ULL, "6080582_m.fxe"},
	{0xE54CA5B8C6ABD3FFULL, "452875604.ogg"},
	{0xE54CA790C9A5766AULL, "vxw_bonepitdestroyed_a3_20n_0_0.vlm"},
	{0xE54D2960F83A7053ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_left.evt"},
	{0xE54D84529C308D09ULL, "vxlt_0_lowtown_a2_fdz_0_0.vlm"},
	{0xE54E4FDE6F040E3FULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforward.evt"},
	{0xE54E50D567C72BD0ULL, "vxlt_0_lowtown_a2_n_c_hz6_0_0.vlm"},
	{0xE54E5A2128F5BE2EULL, "6043685_m.fxe"},
	{0xE54E8C445B798919ULL, "hlac_lt_0_lowtown_a2_n_65535_9z.dds"},
	{0xE54E95D494271A53ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v1.evt"},
	{0xE54EA678ADE17814ULL, "6097224_m.fxe"},
	{0xE54ED67AB1EB2BCDULL, "vxw_woundedcoast_12o_0_0.vlm"},
	{0xE54F28579E91DB4BULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.dura.evt"},
	{0xE54F3ACFD1D3F35EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_chin.ani"},
	{0xE54F3D0F4E3A1675ULL, "gen00fl_isabela6098194.cl"},
	{0xE54F6BA7DECADC88ULL, "vxht_mansion2_21t_0_0.vlm"},
	{0xE54FD7CB8596D498ULL, "follower_banter6155628.cl"},
	{0xE54FD91D5656DF3AULL, "1054077200.ogg"},
	{0xE55006AA568E90DEULL, "vxdr_0_deeproads_ef7_0_0.vlm"},
	{0xE55055F2E170C5F2ULL, "6107664_m.fxe"},
	{0xE5511D1804FCEF44ULL, "6060998_m.fxe"},
	{0xE551939470248CBEULL, "qun110_petrice6018532.cl"},
	{0xE5522485BB584151ULL, "6113885_f.fxe"},
	{0xE552C86D40011289ULL, "6084723_f.fxe"},
	{0xE5542260F8408294ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_left.gad"},
	{0xE554E16CF8D6B717ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.impa.evt"},
	{0xE55530DE6F09F828ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkforward.gad"},
	{0xE555337CB38FDED2ULL, "hlac_lt_0_lowtown_a2_n_-65537_4y.dds"},
	{0xE5558ED4942D2C94ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v1.gad"},
	{0xE555A49656F99522ULL, "6103645_f.fxe"},
	{0xE5563EDA570A25EBULL, "664275719.ogg"},
	{0xE55641579E98240CULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.dura.gad"},
	{0xE5565567FD3FC2E7ULL, "6033492_m.fxe"},
	{0xE55693CC7EB804F5ULL, "zz_dre_debug6205400.cl"},
	{0xE556B3CC3C2C6723ULL, "6170128_m.fxe"},
	{0xE55783754D844665ULL, "6127513_m.fxe"},
	{0xE55787FB39820E75ULL, "llac_lt_0_lowtown_docks_-65535_0.dds"},
	{0xE55795B3C0641FEDULL, "vxw_bonepitmining_a2_87_0_0.vlm"},
	{0xE5579E59023A9FE7ULL, "vxw_woundedcoast_au_0_0.vlm"},
	{0xE55806B6763B19C6ULL, "6143674_f.fxe"},
	{0xE5582012C4A6EFB4ULL, "vxdr_primevaltomb_34w_0_0.vlm"},
	{0xE5585E681C5BBC94ULL, "hlac_w_encounter3_-327675_0z.dds"},
	{0xE5587EC0493005DAULL, "428684950.ogg"},
	{0xE5588F94225DE9F7ULL, "dre000_sibling_has_blight6036486.cl"},
	{0xE558FD0AA068A98EULL, "vxht_keep_5u5_0_0.vlm"},
	{0xE55920224E9DFF14ULL, "vxw_sundermount_2zu_0_0.vlm"},
	{0xE559702FD18ECD6BULL, "vrc320ip_at_mansion.utp"},
	{0xE559C05740542DECULL, "6144599_m.fxe"},
	{0xE559F76A8635A5F9ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls106a_0t.dds"},
	{0xE55A8CFB943D4CEAULL, "vxlt_gallowsprison_n_5s3_0_0.vlm"},
	{0xE55A91E6C93B4C3FULL, "6135750_m.fxe"},
	{0xE55ABCF7825CAC54ULL, "vxlt_0_lowtown_a3_n_c_ej6_0_0.vlm"},
	{0xE55B3160F846BB72ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_glance_left.ani"},
	{0xE55B369374ED8A23ULL, "6163546_m.fxe"},
	{0xE55B384EF7211DBFULL, "vxht_chantry_27q_0_0.vlm"},
	{0xE55B38EF5CB5CB89ULL, "vxht_0_hightown_n_8q0_0_0.vlm"},
	{0xE55BA046EF805A04ULL, "vxht_keep_n_55h_0_0.vlm"},
	{0xE55BC3B08533BD75ULL, "234896126.ogg"},
	{0xE55BC58F43962F04ULL, "6179436_m.fxe"},
	{0xE55BD8A22968F418ULL, "6132873_m.fxe"},
	{0xE55BE26CF8DCD7A0ULL, "art\\vfx\\combat\\vfx_str_g_pound_aoe_l_p.impa.gad"},
	{0xE55C40579E9C8E0AULL, "art\\vfx\\creatures\\fxc_sylvan_fire_c.dura.ani"},
	{0xE55C52CFD1DEA62FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_chin.evt"},
	{0xE55C9DD494336572ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_attk_v1.ani"},
	{0xE55CC2B259BD4450ULL, "vxlt_0_lowtown_a2_n_c_hu7_0_0.vlm"},
	{0xE55CDD7F57773FFCULL, "6070906_m.fxe"},
	{0xE55D15D4A8D6AEBAULL, "vxht_keep_36z_0_0.vlm"},
	{0xE55D34982E781B1AULL, "6140247_m.fxe"},
	{0xE55D34E1EBC51AABULL, "custic_light_f.dds"},
	{0xE55D498D4BDF911DULL, "character_character_p75.psh.fxo.dx11.assembly"},
	{0xE55DBFF11CE435C0ULL, "vxht_brothel_n_4y2_0_0.vlm"},
	{0xE55E6B235D8D2084ULL, "30980307.ogg"},
	{0xE55E6E08E48B8DE7ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabright.gad"},
	{0xE55E8BDEA9328600ULL, "vxca_caverns2_701_0_0.vlm"},
	{0xE55E97C55C74F473ULL, "6092573_f.fxe"},
	{0xE55F4A2AD46DECCBULL, "vxlt_0_lowtown_a2_n_gzd_0_0.vlm"},
	{0xE55F702077CAFE2BULL, "647426338.ogg"},
	{0xE55FC0A2F86B2DD8ULL, "vxht_mansion2b_3gs_0_0.vlm"},
	{0xE5603D742CDECF43ULL, "637473337.ogg"},
	{0xE5603E300F232214ULL, "lgt000im_ultra_waxlerhat.uti"},
	{0xE5608290669ED146ULL, "hlac_ht_mansion2b_818_41y.dds"},
	{0xE560E06164206910ULL, "hlac_lt_0_lowtown_n_65535_8z.dds"},
	{0xE56142C6524B09CBULL, "zz_area_signage15.utp"},
	{0xE56149DFB9409BE2ULL, "gen00fl_aveline6132856.cl"},
	{0xE5626919F396CB16ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_open_sa.dds"},
	{0xE56273B4E9DDC8E5ULL, "vxdr_primevaltomb_3go_0_0.vlm"},
	{0xE56353CFD1E4C6D8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_rub_chin.gad"},
	{0xE56356824A17EE52ULL, "188384525.ogg"},
	{0xE5635F9DFDC01A40ULL, "vxdr_0_deeproads_9c3_0_1.vlm"},
	{0xE5636DBC293F298BULL, "static_static_p107.psh.fxo.dx11"},
	{0xE563A8D7ACF175EBULL, "zz_ave_debug6064095.cl"},
	{0xE564167348EFADA2ULL, "6104745_f.fxe"},
	{0xE56446EFA8EFF9BBULL, "vxlt_0_lowtown_docks_b53_0_1.vlm"},
	{0xE5644B61C9A8494DULL, "w_bonepit_3_0.msh"},
	{0xE5645961C9A86080ULL, "w_bonepit_3_0.mao"},
	{0xE5645D61C9A867D3ULL, "w_bonepit_3_0.mmh"},
	{0xE5646308E48FE6F4ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabright.evt"},
	{0xE5647F09A04119DEULL, "art\\characters\\playercharacter\\dwarfmale\\boo_meda.phy"},
	{0xE56538B04A53AD4CULL, "35023544.ogg"},
	{0xE565961FF8496698ULL, "lgt160_ntb_elf_conflict6163341.cl"},
	{0xE565B899ECEBBB0AULL, "gen00fl_sebastian6195826.cl"},
	{0xE565EB6963478342ULL, "42159568.ogg"},
	{0xE56635D3BE110D64ULL, "art\\levels\\races\\natural\\proxy\\foliage\\tre_zzpine.phy"},
	{0xE566678D8FC8AE35ULL, "6148828_m.fxe"},
	{0xE566F0C07C1A628AULL, "hlac_lt_darktown_-1_27y.dds"},
	{0xE56715F9A9ED8E58ULL, "vxlt_gallowstemplar_f_35a_0_0.vlm"},
	{0xE5671BAD2C88E6B5ULL, "256967972.ogg"},
	{0xE5676EE7925EC20EULL, "vxht_keep_5x6_0_0.vlm"},
	{0xE567A4AE1181718CULL, "qun110_petrice6037685.cl"},
	{0xE567D89FE322512FULL, "649413365.ogg"},
	{0xE567EC00B4E382A7ULL, "6151165_m.fxe"},
	{0xE56839A86BAEA99EULL, "hlac_lt_0_lowtown_docks_n_65536_0z.dds"},
	{0xE568CE1375C09C49ULL, "character_character_p93.psh.fxo.dx11"},
	{0xE569478F10F21DFFULL, "445711474.ogg"},
	{0xE5696DD002D117B6ULL, "vxdr_primevaltomb_3hy_0_0.vlm"},
	{0xE569AA2BE917363FULL, "vxht_chantry_20p_0_0.vlm"},
	{0xE569B76CC1747912ULL, "vxht_playershouse_n_1se_0_0.vlm"},
	{0xE56A41DFF0850DEAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9mw_pulleya_1.mao"},
	{0xE56AA1EFFECCC5C6ULL, "31600583.ogg"},
	{0xE56AC47388A5215AULL, "lt_0_lowtown_docks_a3_31_0.rml"},
	{0xE56B2B973B6BBA79ULL, "960250647.ogg"},
	{0xE56B348F4BB35CD0ULL, "vxlt_0_lowtown_a2_n_c_hp4_0_0.vlm"},
	{0xE56BBA3495D75CCDULL, "vxw_woundedcoast_18m_0_0.vlm"},
	{0xE56C849A237F8868ULL, "hlac_ca_caverns1_24422_1x.dds"},
	{0xE56CC23B51A5412BULL, "741492024.ogg"},
	{0xE56E0435B4F87E63ULL, "ser221_party_comment_1.stg"},
	{0xE56E5D12E51585B4ULL, "467887390.ogg"},
	{0xE56E638EC7F91171ULL, "dae310cr_human_f_01.utc"},
	{0xE56F1FFB2E67CFA4ULL, "gen00fl_anders6116815.cl"},
	{0xE56F3B2CB1792EDFULL, "art\\levels\\ztools\\default\\generic\\2wp_basic.phy"},
	{0xE56F9CA61191CAAEULL, "mag291_ntbanter6056612.cl"},
	{0xE56FBC29AB195378ULL, "mcr370_meredith6083622.cl"},
	{0xE56FCE86DA5E9C3BULL, "gen00fl_bethany6109218.cl"},
	{0xE56FD037B4926635ULL, "mrl350ip_grave_f.utp"},
	{0xE570D55E79487921ULL, "785720303.ogg"},
	{0xE5716F43FB87BDC1ULL, "gen00fl_carver6104895.cl"},
	{0xE571BF5014F9FACEULL, "6137043_m.fxe"},
	{0xE571E9331AAE0DD9ULL, "llac_lt_gallowstemplar_f_2394_4.dds"},
	{0xE57230C2EE3B7B5FULL, "91128316.ogg"},
	{0xE5725B08E49C162DULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\grabright.ani"},
	{0xE5727435970926CAULL, "311841655.ogg"},
	{0xE572A6B23978BFC5ULL, "6068847_m.fxe"},
	{0xE572B546311F7C5AULL, "lava_color.dds"},
	{0xE572D142CAE37115ULL, "hlac_lt_0_lowtown_65538_1z.dds"},
	{0xE57301A841980B8FULL, "vxw_encounter2_hg_0_0.vlm"},
	{0xE57312DA84A759CCULL, "vxlt_0_lowtown_a2_n_c_eoq_0_1.vlm"},
	{0xE5733921E12BF3E7ULL, "6039692_m.fxe"},
	{0xE5734D48CCA57921ULL, "6177429_f.fxe"},
	{0xE57374408D133196ULL, "w_bonepitmining_a2_22_0.msh"},
	{0xE5737E408D134294ULL, "w_bonepitmining_a2_22_0.mmh"},
	{0xE57382408D1349A7ULL, "w_bonepitmining_a2_22_0.mao"},
	{0xE573A8C39F68E1C5ULL, "777814283.ogg"},
	{0xE573BD486D33A5F6ULL, "78982916.ogg"},
	{0xE573F7AC3A7B30C7ULL, "gen00fl_sebastian6177296.cl"},
	{0xE5740530498BDC25ULL, "6177592_m.fxe"},
	{0xE57408D4090CE321ULL, "vxht_brothel_n_4w1_0_0.vlm"},
	{0xE575100BB264E4D7ULL, "6160752_m.fxe"},
	{0xE575F4A962E70689ULL, "lgt110_orz_harrowmont.bnk"},
	{0xE576321AF5EBFF1CULL, "446595582.ogg"},
	{0xE576380D212B885DULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_7x.dds"},
	{0xE57734A440A14F31ULL, "vxht_mansion_n_6ns_0_0.vlm"},
	{0xE5774BFE0246C9CCULL, "vxlt_0_lowtown_n_f0d_0_0.vlm"},
	{0xE577A0B16648DD54ULL, "vxlt_0_lowtown_a3_n_c_ep4_0_0.vlm"},
	{0xE577A4444DCF231BULL, "vxca_cavern1_haunted_ee1_0_0.vlm"},
	{0xE577C3CE6C8936F9ULL, "art\\levels\\4cinematics\\ferelden\\prp_fereldencoin.msh"},
	{0xE577CDCE6C8947E7ULL, "art\\levels\\4cinematics\\ferelden\\prp_fereldencoin.mmh"},
	{0xE577D1CE6C894E34ULL, "art\\levels\\4cinematics\\ferelden\\prp_fereldencoin.mao"},
	{0xE5782949B36A9192ULL, "vxht_playershouse_n_1vf_0_0.vlm"},
	{0xE5782CC98AAED0DAULL, "vxdr_slavers_1al_0_0.vlm"},
	{0xE578A9D1AE61DCDAULL, "vxdr_primevalentrance_1j5_0_0.vlm"},
	{0xE5793294C184B855ULL, "s_smoke_01.dds"},
	{0xE5793CAC82882426ULL, "vxlt_gallowscourtyard_n_773_0_0.vlm"},
	{0xE5798D1FAA0B964DULL, "dae000_gamlen6005771.cl"},
	{0xE579ADAB9E7D5448ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyo_0t.dds"},
	{0xE579D6A962EABCC9ULL, "lgt110_orz_harrowmont.cnv"},
	{0xE579EA02549E265AULL, "vxlt_0_lowtown_docks_c5j_0_0.vlm"},
	{0xE57A0038B4A9DEFDULL, "vxlt_hangedman_n_2up_0_0.vlm"},
	{0xE57A929C45A28EF8ULL, "follower_banter6163287.cl"},
	{0xE57AB9C499629892ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1wo_ornate.phy"},
	{0xE57ABAD235E64862ULL, "ht_0_hightown_33_0.anb"},
	{0xE57AD3F22076916FULL, "vxlt_darktown_6k6_0_0.vlm"},
	{0xE57BB8B9ECE65A96ULL, "828776040.ogg"},
	{0xE57BE8126BBD95AAULL, "zz_seb_debug6183193.cl"},
	{0xE57C2B1CD44BF943ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_008a_0t.dds"},
	{0xE57C548DB64FA169ULL, "gen_im_trs_tmp_insignia_01.uti"},
	{0xE57C62F40A77A30EULL, "mag170_thrask6016921.cl"},
	{0xE57CA27F3066F0A1ULL, "dae020cr_merch_weapon.utc"},
	{0xE57CA9D50BAFC5E0ULL, "6075137_f.fxe"},
	{0xE57CBA42545F0ABDULL, "vxca_caverns2_6hi_0_0.vlm"},
	{0xE57DB4732A5BF7AFULL, "vxw_sundermount_2h5_0_1.vlm"},
	{0xE57DCABDF42E601DULL, "367066450.ogg"},
	{0xE57E597884EB8950ULL, "gen_im_cft_rec_pot_06.uti"},
	{0xE57F3682C8D5660EULL, "6135302_m.fxe"},
	{0xE57F625FAB0DFF4AULL, "vxlt_0_lowtown_n_d66_0_0.vlm"},
	{0xE58002885F514CECULL, "follower_banter6143168.cl"},
	{0xE5803C2CB187A3E4ULL, "art\\levels\\ztools\\default\\generic\\2wp_basic.mmh"},
	{0xE580402CB187AA37ULL, "art\\levels\\ztools\\default\\generic\\2wp_basic.mao"},
	{0xE580522CB187C886ULL, "art\\levels\\ztools\\default\\generic\\2wp_basic.msh"},
	{0xE58055BBC1F03E57ULL, "mrl250cr_thief_a.utc"},
	{0xE5807375D50F3069ULL, "6179925_m.fxe"},
	{0xE580A9675BDE4615ULL, "art\\levels\\races\\design\\prp_chesta_0.pwk"},
	{0xE580B1FCEC5F085AULL, "vxht_keep_n_5hm_0_0.vlm"},
	{0xE5815F02386784DBULL, "1020611044.ogg"},
	{0xE581BDA1D0B89387ULL, "202936931.ogg"},
	{0xE582042612881913ULL, "6119833_m.fxe"},
	{0xE5828100578DAC0CULL, "6099208_m.fxe"},
	{0xE582F3EFD81CE967ULL, "vxw_woundedcoast_pv_0_0.vlm"},
	{0xE583515E0F5E52DAULL, "272659577.ogg"},
	{0xE583B369D01710DCULL, "hlac_lt_gallowstemplar_f_2382_1y.dds"},
	{0xE5840015EE7E19A8ULL, "1002515015.ogg"},
	{0xE5845594CBB5B59AULL, "vxlt_gallowstemplar_f_3sa_0_0.vlm"},
	{0xE58460297F61E88AULL, "1065267531.ogg"},
	{0xE5846B722D22625FULL, "41201962.ogg"},
	{0xE584A53D1CD3CF08ULL, "follower_banter6170711.cl"},
	{0xE584BD834DD6A7B4ULL, "vxdr_primevaltomb_9e7_0_1.vlm"},
	{0xE584C4477632A864ULL, "6129248_m.fxe"},
	{0xE58531D88D81449EULL, "86995645.ogg"},
	{0xE5855C543015D6A8ULL, "vxht_chantry_n_2ys_0_0.vlm"},
	{0xE585E05059D3B924ULL, "212201954.ogg"},
	{0xE5868CE5F7C98B6DULL, "vxlt_hangedman_4v_0_0.vlm"},
	{0xE5880A5B143A75D8ULL, "6035321_m.fxe"},
	{0xE588883A21DD8F7AULL, "llac_lt_0_lowtown_a2_n_65537_2.dds"},
	{0xE588D0490EC964FEULL, "6084235_m.fxe"},
	{0xE588D95CE50B0F1EULL, "580670882.ogg"},
	{0xE58973106BF0B6BFULL, "dre000_dougal6131116.cl"},
	{0xE58994847D9F8F53ULL, "409080424.ogg"},
	{0xE58A9680EDA0652DULL, "zz_vrc_debug6200212.cl"},
	{0xE58A9F90B73B357CULL, "6043508_m.fxe"},
	{0xE58AC09A2C51A967ULL, "vxdr_dungeon_1sp_0_1.vlm"},
	{0xE58B03E5294FA22AULL, "vxlt_0_lowtown_a2_n_c_gb7_0_0.vlm"},
	{0xE58B23FA486145F7ULL, "6110603_m.fxe"},
	{0xE58B256A4C5E2735ULL, "424415258.ogg"},
	{0xE58B9C8CAAA32C6AULL, "art\\characters\\weapons\\staff\\textures\\w_stf_006a_0d.dds"},
	{0xE58BED1E20B87A37ULL, "hlac_dr_0_deeproads_17038_0z.dds"},
	{0xE58C0772EA66F073ULL, "gen00fl_anders6107672.cl"},
	{0xE58C1044C4C5C2E4ULL, "qun231cr_varnell_minion6.utc"},
	{0xE58C22917F4990BAULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12_tool.mmh"},
	{0xE58C26917F4997F1ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12_tool.mao"},
	{0xE58C34917F49AF00ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12_tool.msh"},
	{0xE58C4A605439E57EULL, "917290475.ogg"},
	{0xE58CC35992655007ULL, "vxlt_undercityencounter_x5_0_0.vlm"},
	{0xE58CF87B7C477727ULL, "6104422_m.fxe"},
	{0xE58D4B4C7DEC1D3DULL, "zz_seb_debug6215752.cl"},
	{0xE58D69498E990B22ULL, "hlac_lt_0_lowtown_n_65536_8y.dds"},
	{0xE58DAD3B9ED1B8A1ULL, "ps3downsamplemsaa.psh.fxo"},
	{0xE58DB24C66FCE0A2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinl_2.mao"},
	{0xE58DFC13EFFC2A0CULL, "llac_w_encounter1_65536_0.dds"},
	{0xE58E46FF02BA1815ULL, "mag321_comtesse_de_launcet6043408.cl"},
	{0xE58E540036B0A767ULL, "follower_banter6172802.cl"},
	{0xE58E6A1853B53BCDULL, "231503852.ogg"},
	{0xE58E718F6933CEA9ULL, "vxca_cave_2e8_0_1.vlm"},
	{0xE58FA7E6578249C4ULL, "vxlt_blackemporium_168_0_0.vlm"},
	{0xE5905AFA3FE58064ULL, "263007165.ogg"},
	{0xE5906FE0F003383CULL, "6096747_f.fxe"},
	{0xE59084AE4BD74A75ULL, "dae000ip_crowd_mf7_talk6.utp"},
	{0xE590B950FC736107ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobb_0t.dds"},
	{0xE591338A6CA1231CULL, "lt_0_lowtown_docks_a3_n_108_0.mao"},
	{0xE5913F8A6CA137BFULL, "lt_0_lowtown_docks_a3_n_108_0.mmh"},
	{0xE591458A6CA14261ULL, "lt_0_lowtown_docks_a3_n_108_0.msh"},
	{0xE59145CAE16F5284ULL, "vxht_0_hightown_a2_b83_0_0.vlm"},
	{0xE5914EF5A70A0A88ULL, "lt_0_lowtown_docks_a3_177_0.msh"},
	{0xE59160F5A70A2919ULL, "lt_0_lowtown_docks_a3_177_0.mao"},
	{0xE59161FB47C4BB23ULL, "hlac_lt_0_lowtown_docks_n_65534_5z.dds"},
	{0xE591646589F2D0D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_20_walk.ani"},
	{0xE5916CF5A70A3DB2ULL, "lt_0_lowtown_docks_a3_177_0.mmh"},
	{0xE591DBA3DF0B713FULL, "gen00fl_varric6088863.cl"},
	{0xE591E2A92C15C3C4ULL, "art\\vfx\\combat\\vfx_w_wns_right_b_close1_c.impa.ani"},
	{0xE592268E9E3534FBULL, "13780901.ogg"},
	{0xE5925565E1985282ULL, "w_bonepitmining_a2_13_0.msh"},
	{0xE5925765E1985638ULL, "w_bonepitmining_a2_13_0.mmh"},
	{0xE5925B6E6662E01BULL, "gen00fl_aveline6139829.cl"},
	{0xE5926365E1986ADBULL, "w_bonepitmining_a2_13_0.mao"},
	{0xE592987CD7249567ULL, "250576174.ogg"},
	{0xE592C0B5393CE628ULL, "sst_ss_merrill.gda"},
	{0xE5936F031F8EAF96ULL, "6029031_f.fxe"},
	{0xE593CE31220BEF28ULL, "vxht_chantry_n_2tp_0_0.vlm"},
	{0xE593EC2B69FA6575ULL, "saveload.gfx"},
	{0xE593FAE94AAE146EULL, "525798642.ogg"},
	{0xE5942056059D29ACULL, "1028578609.ogg"},
	{0xE5942429073F46DDULL, "vxw_encounter2_camp_l9_0_0.vlm"},
	{0xE5946DDDC9ED99C7ULL, "6173406_m.fxe"},
	{0xE594EAA96D0E90D5ULL, "1674144.ogg"},
	{0xE5952A4F80647D9DULL, "hlac_lt_0_lowtown_a2_n_65536_2z.dds"},
	{0xE5955765E19A8C4BULL, "w_bonepitmining_a2_13_0.rml"},
	{0xE5958AE552C9F303ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofr.msh"},
	{0xE59598E552CA0B5DULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofr.mmh"},
	{0xE5959CE552CA126EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3ca_gallowroofr.mao"},
	{0xE595AB52A466722CULL, "6206212_f.fxe"},
	{0xE596216E2A5460B6ULL, "w_blightlands_255_0.mao"},
	{0xE5962356A895FDDDULL, "w_bonepit_103_0.phy"},
	{0xE5962D6E2A547555ULL, "w_blightlands_255_0.mmh"},
	{0xE5962F6E2A5478CBULL, "w_blightlands_255_0.msh"},
	{0xE5970356D2B5D2E4ULL, "hlac_w_encounter3_-65534_0y.dds"},
	{0xE59723C05846950DULL, "vxlt_0_lowtown_a2_n_c_fhk_0_0.vlm"},
	{0xE59741390C723741ULL, "hlac_w_sundermount_-196612_0z.dds"},
	{0xE597523C53E9CDB3ULL, "780065081.ogg"},
	{0xE59770B953889EADULL, "gen00fl_merrill6130054.cl"},
	{0xE597849636C70343ULL, "vxlt_hangedman_th_0_0.vlm"},
	{0xE598DFA92C1BDD26ULL, "art\\vfx\\combat\\vfx_w_wns_right_b_close1_c.impa.gad"},
	{0xE5993EA8B306746CULL, "vxdr_0_deeproads_cwt_0_0.vlm"},
	{0xE5996482D1829779ULL, "tools_light_lg.mmh"},
	{0xE5996E82D182A867ULL, "tools_light_lg.msh"},
	{0xE5999D543797A865ULL, "vxht_brothel_n_4vj_0_0.vlm"},
	{0xE599C8536695E0C2ULL, "vxlt_gallowscourtyard_7fg_0_0.vlm"},
	{0xE599E60EE79FC41DULL, "360593065.ogg"},
	{0xE59A3CD3EC398E14ULL, "hlac_ht_0_hightown_65537_2y.dds"},
	{0xE59A565EBA7905E1ULL, "hla_ht_0_hightown_a2_-65536_0y.dds"},
	{0xE59BCA66927A28D3ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm02.phy"},
	{0xE59C551C81782AF6ULL, "800497204.ogg"},
	{0xE59CA28DDC14D8B9ULL, "6042574_f.fxe"},
	{0xE59D0D5268C442E5ULL, "vxht_keep_n_5xv_0_0.vlm"},
	{0xE59D5F435298AD82ULL, "555347728.ogg"},
	{0xE59DEB83B99CF897ULL, "mrl351cr_fakepol.utc"},
	{0xE59E5C6589FD4D0AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_20_walk.evt"},
	{0xE59EB38409548A22ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\exit.gad"},
	{0xE59EF462D0F5B755ULL, "358529450.ogg"},
	{0xE59F1394E303C871ULL, "pro000_refugee_male6052108.cl"},
	{0xE59F4FA1B5F0292DULL, "mcr310_argument6072660.cl"},
	{0xE59F51BD5E10A27CULL, "6033536_m.fxe"},
	{0xE59FDAA92C21F305ULL, "art\\vfx\\combat\\vfx_w_wns_right_b_close1_c.impa.evt"},
	{0xE59FEED61E74A58CULL, "345722186.ogg"},
	{0xE5A02406C1BCEDDEULL, "vxht_chantry_36r_0_0.vlm"},
	{0xE5A07000FCCBC1AAULL, "pro000_merchant6051839.cl"},
	{0xE5A1187CBEC90363ULL, "vrc221tr_bartrand_final_tlk.ncs"},
	{0xE5A14DEC48F4DC94ULL, "855649955.ogg"},
	{0xE5A1A38D6A4A5DA6ULL, "6021223_m.fxe"},
	{0xE5A21F4208550364ULL, "vxlt_gallowscourtyard_3x7_0_0.vlm"},
	{0xE5A2430A5D7427ADULL, "dre000_varric6036979.cl"},
	{0xE5A29605F9355F5DULL, "vxw_encounter2_camp_g8_0_0.vlm"},
	{0xE5A302F069BA1D6AULL, "vxw_encounter2_camp_68_0_0.vlm"},
	{0xE5A32B64AD34F1C3ULL, "682876345.ogg"},
	{0xE5A332CA2125317DULL, "6131356_m.fxe"},
	{0xE5A37A6E199DF52BULL, "511719384.ogg"},
	{0xE5A3808C5B1B8E46ULL, "6070122_m.fxe"},
	{0xE5A40F92825DBF6CULL, "vxht_chantry_2ov_0_0.vlm"},
	{0xE5A4A17388D65E66ULL, "lt_0_lowtown_docks_a3_31_0.anb"},
	{0xE5A4D3461E6DE84BULL, "550166811.ogg"},
	{0xE5A4D4EC33357DBBULL, "ht_0_hightownexplosion_n_54_0.rml"},
	{0xE5A5161945E65821ULL, "206935327.ogg"},
	{0xE5A53B658A03347DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_left_20_walk.gad"},
	{0xE5A558A14B82867EULL, "lt_0_lowtown_a3_n_c_42_0.rml"},
	{0xE5A5599A7E4219C9ULL, "vxlt_0_lowtown_glw_0_0.vlm"},
	{0xE5A599F306F0B3D2ULL, "235114495.ogg"},
	{0xE5A5D56A942A980CULL, "gen00fl_anders6113122.cl"},
	{0xE5A5D684095AE419ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\exit.evt"},
	{0xE5A64D15A12EA298ULL, "vxlt_undercityserialkiller_156_0_0.vlm"},
	{0xE5A6A9BF3DF59F33ULL, "t3_dog_d03.tnt"},
	{0xE5A6BC83C02C9D2FULL, "zz_vault_debug6203250.cl"},
	{0xE5A6CBA247D9A101ULL, "694434292.ogg"},
	{0xE5A6F340B3D0DC6DULL, "vxlt_0_lowtown_n_ewk_0_0.vlm"},
	{0xE5A6F54AD355709BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_weaponhanger.msh"},
	{0xE5A6FAEFBDCBF433ULL, "vxw_encounter2_d7_0_0.vlm"},
	{0xE5A6FE025FD8FA2BULL, "604532959.ogg"},
	{0xE5A7074AD3558FE6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_weaponhanger.mao"},
	{0xE5A7134AD355A385ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_weaponhanger.mmh"},
	{0xE5A71856A8A45E49ULL, "w_bonepit_103_0.mao"},
	{0xE5A72456A8A472E2ULL, "w_bonepit_103_0.mmh"},
	{0xE5A72656A8A47598ULL, "w_bonepit_103_0.msh"},
	{0xE5A7B085A4FC8CECULL, "vxdr_0_deeproads_clu_0_0.vlm"},
	{0xE5A7C387A470A182ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsacks.msh"},
	{0xE5A7C587A470A538ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsacks.mmh"},
	{0xE5A7D187A470B9DBULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsacks.mao"},
	{0xE5A8198FB1404FCDULL, "139182715.ogg"},
	{0xE5A82E6922B39CEDULL, "vxdr_0_deeproads_8sj_0_0.vlm"},
	{0xE5A8A1FCF1EF3BFAULL, "vxlt_warehouse_yb_0_0.vlm"},
	{0xE5A8E7DF01C5188EULL, "pro000_soldier_leader6040813.cl"},
	{0xE5A929F61E584C0FULL, "259699054.ogg"},
	{0xE5A9379F58EE99DCULL, "hlac_lt_backalley_n_1_6z.dds"},
	{0xE5A9412003C08915ULL, "llac_lt_hangedman_-1_0.dds"},
	{0xE5A94967C0B94D93ULL, "vxht_keep_n_73c_0_0.vlm"},
	{0xE5A95D50335594BFULL, "43742849.ogg"},
	{0xE5A9B7E1D5A593FAULL, "6058924_m.fxe"},
	{0xE5A9B95C3E2202B9ULL, "dre000_varric6207968.cl"},
	{0xE5A9E92B339CFB33ULL, "pro000_aveline6047248.cl"},
	{0xE5AA3375AE9BC26BULL, "6166776_m.fxe"},
	{0xE5AA34D925F7A5ECULL, "gen00fl_aveline6115479.cl"},
	{0xE5AA618AC8539769ULL, "6083331_f.fxe"},
	{0xE5AAB07BAB17C8F7ULL, "hlac_ht_mansion_n_2995_6x.dds"},
	{0xE5AABC5D4974417BULL, "ht_0_hightownexplosion_n_8_0.mmh"},
	{0xE5AAC85D49745518ULL, "ht_0_hightownexplosion_n_8_0.mao"},
	{0xE5AADA5D49747465ULL, "ht_0_hightownexplosion_n_8_0.msh"},
	{0xE5AAE572D1DC051EULL, "llac_lt_warehouse_65537_5.dds"},
	{0xE5AAE5D15DB41F4BULL, "hlac_lt_0_lowtown_docks_a3_n_196607_2x.dds"},
	{0xE5AB1744AC75AA41ULL, "6084214_m.fxe"},
	{0xE5AB2C9309096AFBULL, "6037621_m.fxe"},
	{0xE5AB4CA371C812C1ULL, "zz_dae_debug6045452.cl"},
	{0xE5AC24D79CD619FAULL, "lowtown_male_alienage2_mob6166522.cl"},
	{0xE5AC6690BAC7DA2EULL, "6123363_m.fxe"},
	{0xE5AC92B29400C51CULL, "sst_ss_isabela.gda"},
	{0xE5ACE2ABDF40E921ULL, "llac_lt_0_lowtown_docks_a3_n_-131075_7.dds"},
	{0xE5ACFEA6C8AEBD64ULL, "27935260.ogg"},
	{0xE5AD3342B08F8534ULL, "pro000_meeran6048969.cl"},
	{0xE5AD3F61B5929E4EULL, "191533974.ogg"},
	{0xE5AD58D80239D2A3ULL, "dae000ip_crowd_lm1_talk6.utp"},
	{0xE5AD6F0594231C69ULL, "vxlt_0_lowtown_docks_bsb_0_0.vlm"},
	{0xE5AD93CC30AFF6AEULL, "6211954_m.fxe"},
	{0xE5ADBDF2C6B54A83ULL, "3781587.ogg"},
	{0xE5ADBEE6DE5DD5E8ULL, "dae110ip_at_to_trench_a.utp"},
	{0xE5ADED4ADC7B7F1EULL, "382067080.ogg"},
	{0xE5AE649811BED5BBULL, "vxht_0_hightown_c41_0_0.vlm"},
	{0xE5AE9ABE7F03FE3EULL, "art\\characters\\creatures\\shade\\textures\\c_shade_0n.dds"},
	{0xE5AEF1D75B88CBCAULL, "hlac_lt_0_lowtown_a2_-65537_5y.dds"},
	{0xE5AF2FA4DFEBFEAFULL, "6146983_m.fxe"},
	{0xE5AF73DF10DA3973ULL, "llac_w_bonepit_262146_0.dds"},
	{0xE5B01FE239AE3BBBULL, "129941346.ogg"},
	{0xE5B03EA4D301A8B5ULL, "hlac_lt_gallowscourtyard_65534_0x.dds"},
	{0xE5B09990B363C0BAULL, "ser221_gascard6133741.cl"},
	{0xE5B0EDB758769C48ULL, "336935649.ogg"},
	{0xE5B0FD15E30363A1ULL, "hlac_w_bonepitmining_a2_0_10z.dds"},
	{0xE5B107E2EB2B77DDULL, "vxw_encounter2_camp_b7_0_0.vlm"},
	{0xE5B141917F68E1C5ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_ball12_tool.phy"},
	{0xE5B1506D878290DAULL, "hlac_dr_slavers_3217_5x.dds"},
	{0xE5B1629CD0DB847EULL, "dre000_varric6208010.cl"},
	{0xE5B174CD5BB035EAULL, "vxw_encounter2_camp_19_0_0.vlm"},
	{0xE5B1CC87A47905C3ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsacks.phy"},
	{0xE5B1F6A6279CA357ULL, "one220_yevhen6114124.cl"},
	{0xE5B2141055DE4FDFULL, "vxht_0_hightown_bcu_0_0.vlm"},
	{0xE5B22401DBEE7CC0ULL, "6170466_m.fxe"},
	{0xE5B242660875942DULL, "gen00fl_merrill6128275.cl"},
	{0xE5B269D31E48F140ULL, "hlac_w_encounter2_-131068_0z.dds"},
	{0xE5B27F5300BA4352ULL, "233776015.ogg"},
	{0xE5B2816F7453D7ECULL, "vxht_chantry_2hw_0_0.vlm"},
	{0xE5B29D8A1CC08513ULL, "hlac_lt_0_lowtown_a2_n_1_2z.dds"},
	{0xE5B2B1905247DD0FULL, "qun221_viscount6070126.cl"},
	{0xE5B2B59A860D3F7FULL, "gen00fl_aveline6139713.cl"},
	{0xE5B2CCF863540941ULL, "hlac_lt_0_lowtown_a2_65537_5y.dds"},
	{0xE5B2DB9D4ED72B9FULL, "w_bonepitdestroyed_a3_135_0.msh"},
	{0xE5B2DE8409657C08ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basetwohanded\\combat\\exit.ani"},
	{0xE5B2E19D4ED73641ULL, "w_bonepitdestroyed_a3_135_0.mmh"},
	{0xE5B2ED9D4ED74AE2ULL, "w_bonepitdestroyed_a3_135_0.mao"},
	{0xE5B35600FE1D3B3FULL, "6107702_m.fxe"},
	{0xE5B3CC071FDCDDE1ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_hed_hdra_0d.dds"},
	{0xE5B45BC21A0B91D0ULL, "vxlt_ruinedhouse_n_ig_0_0.vlm"},
	{0xE5B49556ECE5F9D0ULL, "vxht_0_hightown_a3_9zk_0_0.vlm"},
	{0xE5B4AB359DA9818EULL, "mnd_imp_smoke.mao"},
	{0xE5B4E2AD54443E33ULL, "vxht_0_hightown_a2_aup_0_0.vlm"},
	{0xE5B502373AD45652ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzcart_0.pwk"},
	{0xE5B51B45BEB40D57ULL, "6208485_m.fxe"},
	{0xE5B522BF9A95E0E8ULL, "gen_im_trs_crc_lgt_cht_r2.uti"},
	{0xE5B5251F269C4BECULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nbla.phy"},
	{0xE5B54C048A448683ULL, "w_encounter2_37_0.mao"},
	{0xE5B550048A448DD0ULL, "w_encounter2_37_0.mmh"},
	{0xE5B55E048A44A5EAULL, "w_encounter2_37_0.msh"},
	{0xE5B56CCCAFC20CB3ULL, "vxw_encounter2_o6_0_0.vlm"},
	{0xE5B6B5E712779A45ULL, "vxlt_warehouse_sk_0_0.vlm"},
	{0xE5B6CB1074A580B4ULL, "hlac_lt_0_lowtown_docks_a3_-65537_0z.dds"},
	{0xE5B6D71897EE8BC5ULL, "vxlt_0_lowtown_a2_eom_0_1.vlm"},
	{0xE5B6DD131BF3B20CULL, "6036482_f.fxe"},
	{0xE5B6F38FCA83A8CFULL, "mcr371_hostile_anders6082100.cl"},
	{0xE5B703F836BFCDF8ULL, "dae000_hate_follower_react6167482.cl"},
	{0xE5B75EF7207C9B87ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_lrdb_0n.dds"},
	{0xE5B7F0479CB3033EULL, "llac_w_woundedcoast_-131065_0.dds"},
	{0xE5B7F75666057EF2ULL, "6048385_m.fxe"},
	{0xE5B866976A690690ULL, "hlac_lt_0_lowtown_a2_n_-262144_1y.dds"},
	{0xE5B87B3F9A540636ULL, "lt_0_lowtown_docks_111_0.mao"},
	{0xE5B8873F9A541AD5ULL, "lt_0_lowtown_docks_111_0.mmh"},
	{0xE5B8893F9A541E4BULL, "lt_0_lowtown_docks_111_0.msh"},
	{0xE5B8CAD7276FE894ULL, "vxca_caverns1_fob_0_0.vlm"},
	{0xE5B8E1D947709962ULL, "vxht_0_hightown_c8h_0_0.vlm"},
	{0xE5B8FD8210C1A9EAULL, "vxca_caverns1_mines_j27_0_1.vlm"},
	{0xE5B9754B9836D170ULL, "6018543_f.fxe"},
	{0xE5B9B588480BC66EULL, "686339050.ogg"},
	{0xE5B9E5060FF3099AULL, "dae000_bodhan_letter6144302.cl"},
	{0xE5BA2384C167BC49ULL, "321884724.ogg"},
	{0xE5BA33BA9970737BULL, "138569218.ogg"},
	{0xE5BACC3645835433ULL, "hlac_ca_caverns1_mines_24423_7y.dds"},
	{0xE5BAE8103671861CULL, "hlac_ht_playershouse_n_855_4y.dds"},
	{0xE5BB39A86DDA696CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtb.msh"},
	{0xE5BB3FA86DDA735EULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtb.mmh"},
	{0xE5BB4BA86DDA87F5ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtb.mao"},
	{0xE5BB78DED01B13F3ULL, "6135897_m.fxe"},
	{0xE5BBB256C8C1D960ULL, "6053516_m.fxe"},
	{0xE5BBB76EB19335D1ULL, "ca_caverns2.arl"},
	{0xE5BC6B676649C35AULL, "203317521.ogg"},
	{0xE5BC75EF76898B16ULL, "942808492.ogg"},
	{0xE5BCCA84E48359E4ULL, "89323760.ogg"},
	{0xE5BCE8672B28B5AFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardianwep.mao"},
	{0xE5BCF4672B28CA4CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardianwep.mmh"},
	{0xE5BCFA672B28D43EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_guardianwep.msh"},
	{0xE5BD8D2FA1AC3172ULL, "hlac_dr_0_deeproads_5842_11y.dds"},
	{0xE5BE0A85A088E0D3ULL, "vxlt_warehouse_n_14b_0_0.vlm"},
	{0xE5BE43FD619F207BULL, "w_sundermount_25_0.rml"},
	{0xE5BE64AB4822A5C3ULL, "qun221_viscount6084680.cl"},
	{0xE5BED5A1E951A51DULL, "ser270_emeric6149321.cl"},
	{0xE5BF23C7F7EE38A8ULL, "vxht_chantry_n_2eo_0_0.vlm"},
	{0xE5BF8C88BD6E0B01ULL, "262589942.ogg"},
	{0xE5BF8DB7B4AD3657ULL, "vxlt_0_lowtown_a3_n_c_gjk_0_0.vlm"},
	{0xE5C058DD0D5C684FULL, "6153660_m.fxe"},
	{0xE5C085ED47D4685FULL, "vxht_0_hightown_blt_0_0.vlm"},
	{0xE5C093A4D2ACF9DAULL, "follower_banter6163296.cl"},
	{0xE5C0AB2B6B8E8255ULL, "one280_emrys6104306.cl"},
	{0xE5C1EC334734F11CULL, "lt_0_lowtown_46_0_ll.mmh"},
	{0xE5C21DB2EE26C06CULL, "117967401.ogg"},
	{0xE5C253161A9D787CULL, "6202808_m.fxe"},
	{0xE5C25F3DE6E7EA4AULL, "dae000_flemeth6000800.cl"},
	{0xE5C2689F161AB627ULL, "vxlt_0_lowtown_a2_n_d8u_0_0.vlm"},
	{0xE5C3548A463A56B3ULL, "vxht_0_hightown_a2_aps_0_0.vlm"},
	{0xE5C3DEA9A1B82533ULL, "vxw_encounter2_j9_0_0.vlm"},
	{0xE5C3F2F6BD1B2D67ULL, "623546812.ogg"},
	{0xE5C41A26FD93974BULL, "dre000_stroud6036744.cl"},
	{0xE5C43E1C4E220D2FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_lgtb_0t.dds"},
	{0xE5C484E6533D7FD2ULL, "vxht_mansion_n_2ny_0_1.vlm"},
	{0xE5C4C029948D6125ULL, "25061100.ogg"},
	{0xE5C4E2FDC0F043C0ULL, "hlac_lt_0_lowtown_a3_n_c_-3_5x.dds"},
	{0xE5C6263777D0FDD2ULL, "mag110_arianni6015776.cl"},
	{0xE5C67B968896F3FDULL, "ht_brothel_n_3364_ll.mmh"},
	{0xE5C699A49013CC4DULL, "zz_fol_debug6092330.cl"},
	{0xE5C6C722D0C76589ULL, "vxlt_hangedman_n_wj_0_0.vlm"},
	{0xE5C7059E1561B20AULL, "zz_prmpt_debug6198266.cl"},
	{0xE5C7377F3878CB84ULL, "829543078.ogg"},
	{0xE5C76C9FD4ACACE4ULL, "6060226_m.fxe"},
	{0xE5C78B3B61DD7F52ULL, "vxlt_warehouse_n_14g_0_0.vlm"},
	{0xE5C823B8745EDC11ULL, "vxw_woundedcoast_jf_0_0.vlm"},
	{0xE5C849B90F544968ULL, "405858242.ogg"},
	{0xE5C85D98462AE1AAULL, "hlac_ht_0_hightown_-65534_1x.dds"},
	{0xE5C86FE79F8CD295ULL, "827044075.ogg"},
	{0xE5C87E408D5B8297ULL, "w_bonepitmining_a2_22_0.rml"},
	{0xE5CA09F29BBBCE1EULL, "hlac_w_bonepitmining_a2_65540_0z.dds"},
	{0xE5CA41082E141C61ULL, "vxlt_warehouse_i5_0_0.vlm"},
	{0xE5CA75409B7B24E4ULL, "6118482_m.fxe"},
	{0xE5CA87075149EE07ULL, "mag160_feynriel6020459.cl"},
	{0xE5CA97EA291651CFULL, "6089750_m.fxe"},
	{0xE5CAA351F658282AULL, "973252444.ogg"},
	{0xE5CAA739791B7814ULL, "6172486_m.fxe"},
	{0xE5CABA6810FD2C29ULL, "vxlt_gallowstemplar_2pa_0_0.vlm"},
	{0xE5CAC006F74E119AULL, "746806908.ogg"},
	{0xE5CAECC79C70172EULL, "gen00fl_isabela_init6078163.cl"},
	{0xE5CBA2E8E9ED2A4FULL, "vxlt_0_lowtown_docks_a3_cz1_0_0.vlm"},
	{0xE5CC1B095EF86F6CULL, "hlac_lt_hangedman_-1_13y.dds"},
	{0xE5CC7C62927EF953ULL, "vxlt_warehouse_n_19e_0_0.vlm"},
	{0xE5CCAE25BF2B5D78ULL, "ht_0_hightownexplosion_n_21_0_lc.mmh"},
	{0xE5CCB33D1C214D85ULL, "ts_slavers_arm_table.gda"},
	{0xE5CCEF008C4A1F71ULL, "prop1_s2.psh.fxo"},
	{0xE5CD97B25E0BA366ULL, "vxca_caverns2_980_0_1.vlm"},
	{0xE5CE045847E09052ULL, "vxht_0_hightown_a2_9ay_0_0.vlm"},
	{0xE5CE0DE5EB6A3390ULL, "6059679_m.fxe"},
	{0xE5CE32FD0801BC36ULL, "tbe_dog_act2.ncs"},
	{0xE5CEAC5479D49D1AULL, "46479745.ogg"},
	{0xE5CEB77507B4FFC7ULL, "lgt210_qun_blk_qunari6142309.cl"},
	{0xE5CEBC5BBC1D802AULL, "6075598_m.fxe"},
	{0xE5CEC73686D1E1BFULL, "vxht_chantry_2oy_0_0.vlm"},
	{0xE5CFC68D6B981AF0ULL, "dae000_flemeth6013356.cl"},
	{0xE5CFCC29FC9BB662ULL, "862068196.ogg"},
	{0xE5CFD343F36B9060ULL, "vxht_bartrand_n_6ru_0_0.vlm"},
	{0xE5CFF265D1879CACULL, "301645182.ogg"},
	{0xE5D0E7851E826FACULL, "vxht_mansion2b_30v_0_0.vlm"},
	{0xE5D12DDBFD2FCD00ULL, "dae000cr_grand_cleric6183968.cl"},
	{0xE5D16C6C6071B908ULL, "vxht_mansion2_2bm_0_0.vlm"},
	{0xE5D17956F7A40E5CULL, "hlac_lt_0_lowtown_docks_n_-1_1z.dds"},
	{0xE5D1F34400DDD5F4ULL, "85666782.ogg"},
	{0xE5D23028F81079B1ULL, "lgt000im_ultra_cameo.uti"},
	{0xE5D240B5C1A66290ULL, "6188649_m.fxe"},
	{0xE5D248D789B325EDULL, "vxlt_0_lowtown_n_efh_0_0.vlm"},
	{0xE5D296176B6D6535ULL, "6136000_f.fxe"},
	{0xE5D2C312705A56D9ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_010a_0n.dds"},
	{0xE5D302B52995439DULL, "hla_ht_0_hightown_a3_-65534_0z.dds"},
	{0xE5D306177178D891ULL, "6164835_m.fxe"},
	{0xE5D39573ECF8F49AULL, "vxlt_0_lowtown_n_d7q_0_0.vlm"},
	{0xE5D399A0F663CB45ULL, "vxlt_warehouse_yi_0_0.vlm"},
	{0xE5D3C9BDEF68C848ULL, "vxlt_warehouse_i0_0_0.vlm"},
	{0xE5D3EDE151616E33ULL, "6112223_m.fxe"},
	{0xE5D490FE830693F8ULL, "hlac_ht_0_hightown_-131070_0x.dds"},
	{0xE5D4E2C338FE40A3ULL, "gen00fl_merrill6119642.cl"},
	{0xE5D4F60EB17D9EDDULL, "vxlt_0_lowtown_a3_n_c_gob_0_0.vlm"},
	{0xE5D5090C327F83D8ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef02.phy"},
	{0xE5D525B4315B46CFULL, "933996527.ogg"},
	{0xE5D53048BC641709ULL, "vxw_encounter1_cu_0_0.vlm"},
	{0xE5D538FFC2BD7E09ULL, "vxlt_hangedman_n_ri_0_0.vlm"},
	{0xE5D55360BC29D0B9ULL, "lgt000cr_hate_blood.utc"},
	{0xE5D582EBDC6D26D0ULL, "97130114.ogg"},
	{0xE5D5F37DC2F5E7DEULL, "6209945_f.fxe"},
	{0xE5D648659E915139ULL, "6067774_m.fxe"},
	{0xE5D66AAD738734E8ULL, "lt_0_lowtown_docks_a3_203_0.mmh"},
	{0xE5D676AD7387488BULL, "lt_0_lowtown_docks_a3_203_0.mao"},
	{0xE5D688AD73876712ULL, "lt_0_lowtown_docks_a3_203_0.msh"},
	{0xE5D6BD8CA2B7F7BDULL, "hlac_ca_caverns2prc_5802_63x.dds"},
	{0xE5D795AA4C19D087ULL, "253238233.ogg"},
	{0xE5D7AE0AF0E87E2BULL, "hlac_lt_0_lowtown_a2_1_8y.dds"},
	{0xE5D7DC07FBB75A0AULL, "mag210_arianni6031717.cl"},
	{0xE5D7FBD734C6E6D9ULL, "gen00fl_dog6171159.cl"},
	{0xE5D7FF30844AF86AULL, "6137245_m.fxe"},
	{0xE5D7FF53C49AC007ULL, "vxw_encounter2_ed_0_0.vlm"},
	{0xE5D8141741A15305ULL, "858245043.ogg"},
	{0xE5D83D993B9C9296ULL, "vxca_caverns1_mines_jfv_0_0.vlm"},
	{0xE5D8B2E5200A34E1ULL, "vxlt_warehouse_r4_0_0.vlm"},
	{0xE5D929ECBD219C2FULL, "mag210_arianni.bnk"},
	{0xE5D97ACF4A41158DULL, "vxca_caverns1_e4y_0_0.vlm"},
	{0xE5D9855F9E07727EULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrubble_d.dds"},
	{0xE5DA0E075212ECE0ULL, "mag131_keran6091392.cl"},
	{0xE5DA184604A50AE0ULL, "322526949.ogg"},
	{0xE5DA1A328CB46440ULL, "vxw_bonepit_28l_0_0.vlm"},
	{0xE5DAA64334FB689EULL, "art\\characters\\creatures\\desiredemon\\animation\\choke.evt"},
	{0xE5DB13095657873DULL, "154576351.ogg"},
	{0xE5DBF3D8C1A140DBULL, "vxlt_gallowstemplar_n_5k6_0_0.vlm"},
	{0xE5DC4BECBD240B97ULL, "mag210_arianni.cnv"},
	{0xE5DC741F1D049FBBULL, "isa120_hayder6074981.cl"},
	{0xE5DC863730E9A9DFULL, "11792157.ogg"},
	{0xE5DCA966A2623A6FULL, "crt_emissary.wbd"},
	{0xE5DCAC58906541B7ULL, "175141424.ogg"},
	{0xE5DCE5208AD1DF90ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc102a_0n.dds"},
	{0xE5DCEB7D4EB1AF0DULL, "zz_approval_debug6003828.cl"},
	{0xE5DD0ACC28D8697EULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_masc_0s.dds"},
	{0xE5DD391378C7FA3FULL, "vxht_chantry_2hx_0_0.vlm"},
	{0xE5DD69A72BC0995FULL, "vxht_0_hightown_brv_0_0.vlm"},
	{0xE5DD991AF2D226A1ULL, "863010812.ogg"},
	{0xE5DECD2748F08360ULL, "515819687.ogg"},
	{0xE5DEE29D4EFC72BEULL, "w_bonepitdestroyed_a3_135_0.phy"},
	{0xE5DF5075299EEFE7ULL, "6148637_m.fxe"},
	{0xE5DF55A14BB3F922ULL, "lt_0_lowtown_a3_n_c_42_0.anb"},
	{0xE5DF76F6195AA471ULL, "520301606.ogg"},
	{0xE5E0B2CAEBBA1DB0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple.mao"},
	{0xE5E0B6CAEBBA24C3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple.mmh"},
	{0xE5E0BD4334FFFBC9ULL, "art\\characters\\creatures\\desiredemon\\animation\\choke.gad"},
	{0xE5E0C4CAEBBA3C1DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple.msh"},
	{0xE5E15C0747483CFAULL, "art\\characters\\weapons\\staff\\textures\\w_stf_006a_0t.dds"},
	{0xE5E1E3968FD686BBULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_007a_0s.dds"},
	{0xE5E1ED35ADFC00B6ULL, "6178392_m.fxe"},
	{0xE5E22B2C5F330D04ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.cess.ani"},
	{0xE5E23B9AE15EE0C8ULL, "vxlt_warehouse_r1_0_0.vlm"},
	{0xE5E23E3A04C68873ULL, "vxca_cavern1_haunted_jfg_0_0.vlm"},
	{0xE5E240FE5503B9EBULL, "hlac_dr_slavers_3207_2y.dds"},
	{0xE5E2BA223DF5CE52ULL, "55573692.ogg"},
	{0xE5E33FF037CCB73CULL, "gen00fl_aveline6105438.cl"},
	{0xE5E34D9102ED6EF9ULL, "character_character_p116.psh.fxo"},
	{0xE5E3501E44082925ULL, "6096629_f.fxe"},
	{0xE5E39E82BB6D734FULL, "vxlt_0_lowtown_n_bxg_0_0.vlm"},
	{0xE5E3A225AE5A2F89ULL, "vxw_encounter1_ht_0_0.vlm"},
	{0xE5E3AFBF6CCC93DAULL, "6131099_f.fxe"},
	{0xE5E3E5D375D26FB2ULL, "ave200_aveline_bar6116421.cl"},
	{0xE5E43687DFC5960BULL, "6111381_m.fxe"},
	{0xE5E50820048EB032ULL, "vxlt_0_lowtown_a2_n_e8e_0_0.vlm"},
	{0xE5E516141B86B701ULL, "6040841_m.fxe"},
	{0xE5E530A9A9FA6356ULL, "vxca_cave_1lw_0_1.vlm"},
	{0xE5E589534614B37AULL, "661799651.ogg"},
	{0xE5E59432B765988AULL, "w_bonepit_254_0.mao"},
	{0xE5E59832B7659FD9ULL, "w_bonepit_254_0.mmh"},
	{0xE5E5A232B765B0C7ULL, "w_bonepit_254_0.msh"},
	{0xE5E5B99F455351D3ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdvista01.dura.evt"},
	{0xE5E5C4A079A788D0ULL, "6089804_f.fxe"},
	{0xE5E6113DA0808642ULL, "vxht_keep_n_4hl_0_0.vlm"},
	{0xE5E698C5784DBEA7ULL, "vxca_caverns1_mines_e55_0_0.vlm"},
	{0xE5E69D0A6CFD7F79ULL, "dae000_mother6199750.cl"},
	{0xE5E6ADA5EB0DEE7AULL, "vxw_encounter3_8q_0_0.vlm"},
	{0xE5E736795BFB7E69ULL, "vxlt_0_lowtown_docks_bgf_0_0.vlm"},
	{0xE5E78E433505CA17ULL, "art\\characters\\creatures\\desiredemon\\animation\\choke.ani"},
	{0xE5E843DEA0272836ULL, "llac_lt_gallowscourtyard_n_-131071_3.dds"},
	{0xE5E88C374AB895D6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupb_sa.dds"},
	{0xE5E88D220B41467CULL, "305215514.ogg"},
	{0xE5E89311D3C67A97ULL, "6114051_f.fxe"},
	{0xE5E8A5D31C7039CFULL, "6028144_m.fxe"},
	{0xE5E8ADE203D72D8CULL, "zz_vrc_debug6064277.cl"},
	{0xE5E8F21F1C7A250BULL, "hlac_lt_gallowscourtyard_-196608_0x.dds"},
	{0xE5E9282C5F392766ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.cess.gad"},
	{0xE5E97B157FB2A035ULL, "plt_cod_aow_spellcombo3.plo"},
	{0xE5E996F9BAD9E02AULL, "6160739_m.fxe"},
	{0xE5EA081375E26A9BULL, "6032019_m.fxe"},
	{0xE5EA19A40CC112BEULL, "6042202_m.fxe"},
	{0xE5EA1CD2BB1E2BA7ULL, "zz_vis_debug6050351.cl"},
	{0xE5EA42E63E47FBB3ULL, "776803275.ogg"},
	{0xE5EA86713C3CB4A2ULL, "559686827.ogg"},
	{0xE5EACD4FDEC5E821ULL, "w_bonepitmining_a2_37_0.anb"},
	{0xE5EAF19FCF566490ULL, "6053479_m.fxe"},
	{0xE5EB336317E8BBC1ULL, "6109239_f.fxe"},
	{0xE5EB589BF4BD5978ULL, "993078675.ogg"},
	{0xE5EB858D94A5A60DULL, "vxlt_blackemporium_ry_0_1.vlm"},
	{0xE5EBDB0A8FA254ADULL, "6085415_m.fxe"},
	{0xE5EC1FFE2DD5BFC2ULL, "vxw_bonepitdestroyed_a3_1xe_0_0.vlm"},
	{0xE5EC558A807A64BEULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.impa.evt"},
	{0xE5ECAEBD2E037D00ULL, "hlac_lt_0_lowtown_docks_a3_-196611_0x.dds"},
	{0xE5ECB29F45596414ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdvista01.dura.gad"},
	{0xE5ECC67A6099FDD7ULL, "dre100cr_bodahn.utc"},
	{0xE5ECCFA7C0B1D105ULL, "15974786.ogg"},
	{0xE5ECF340128DEDF7ULL, "ca_cave_2193_lc.mmh"},
	{0xE5ED79030BFA3333ULL, "6140611_m.fxe"},
	{0xE5ED855B49AA747AULL, "722547043.ogg"},
	{0xE5EE03AE850179CDULL, "lt_0_lowtown_30_0.rml"},
	{0xE5EE0DF2C5EC458CULL, "dre000_dougal6129718.cl"},
	{0xE5EE37D555B9586DULL, "gen00fl_aveline6120780.cl"},
	{0xE5EE50048A745293ULL, "w_encounter2_37_0.rml"},
	{0xE5EF232C5F3D8A45ULL, "art\\vfx\\combat\\vfx_glb_cold_frozen_c.cess.evt"},
	{0xE5EF4BE38D5CE03FULL, "vxdr_0_deeproads_cfd_0_0.vlm"},
	{0xE5EF8E0FCE8D7879ULL, "613455416.ogg"},
	{0xE5EF918FF7469886ULL, "vxht_0_hightown_a3_cbg_0_0.vlm"},
	{0xE5EF94837DDD2813ULL, "one100it_thug_note.uti"},
	{0xE5EFCD7BF2FE2742ULL, "6089588_m.fxe"},
	{0xE5EFE2EE0E433F6CULL, "ser221_gascard6133668.cl"},
	{0xE5F034D2CAC03E75ULL, "dae000_brat6156898.cl"},
	{0xE5F0E191F0C30BC6ULL, "llac_dr_0_deeproads_5843_6.dds"},
	{0xE5F16812DCC277DFULL, "6126660_m.fxe"},
	{0xE5F1717A832E6899ULL, "576065305.ogg"},
	{0xE5F1B7CAEBC899E8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr1a_simple.phy"},
	{0xE5F1DD5D49B07BB0ULL, "ht_0_hightownexplosion_n_8_0.phy"},
	{0xE5F243D291FD3A47ULL, "gen00fl_player6207701.cl"},
	{0xE5F247C2D67CFE35ULL, "319115764.ogg"},
	{0xE5F27BF26C5B2C49ULL, "vxht_chantry_n_36r_0_0.vlm"},
	{0xE5F30A6E0712F793ULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc103a_0d.dds"},
	{0xE5F31A5378331287ULL, "dae000_worthy6166574.cl"},
	{0xE5F32C1F827D8239ULL, "6069574_m.fxe"},
	{0xE5F354C929E60E2EULL, "6032133_m.fxe"},
	{0xE5F36C8A8080AAE9ULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.impa.gad"},
	{0xE5F3864EEF19DC9DULL, "follower_banter6155511.cl"},
	{0xE5F3C19F455F9CF2ULL, "art\\vfx\\environment\\level_fx\\birds\\vfx_birdvista01.dura.ani"},
	{0xE5F44A0B45F82D8FULL, "zz_dae_debug6029900.cl"},
	{0xE5F45D980EABF45BULL, "46616907.ogg"},
	{0xE5F4697FF68D9E11ULL, "gen00fl_aveline6137519.cl"},
	{0xE5F4866FE3BB3EBEULL, "959490422.ogg"},
	{0xE5F4A82B33E5111BULL, "455308220.ogg"},
	{0xE5F4E30DA886F107ULL, "vxw_encounter2_kf_0_0.vlm"},
	{0xE5F51F82DD0406FAULL, "vxw_encounter3_5r_0_0.vlm"},
	{0xE5F557048A7A7DEBULL, "w_encounter2_37_0.phy"},
	{0xE5F5FC1F26D39CB7ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nbla.mmh"},
	{0xE5F6001F26D3A3C4ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nbla.mao"},
	{0xE5F6121F26D3C229ULL, "art\\characters\\playercharacter\\humanmale\\hm_cth_nbla.msh"},
	{0xE5F61D7CBF114BFAULL, "vrc221tr_bartrand_final_tlk.utt"},
	{0xE5F638F51209A226ULL, "ht_bartrand_n_mmap.dds"},
	{0xE5F67556EB8203F1ULL, "154677623.ogg"},
	{0xE5F67A744E478737ULL, "6154027_m.fxe"},
	{0xE5F6967B1EE32B11ULL, "vxw_bonepit_1qm_0_0.vlm"},
	{0xE5F69FC664AA9E44ULL, "vxht_keep_n_6f6_0_0.vlm"},
	{0xE5F6BAE71F8E6B5DULL, "317316417.ogg"},
	{0xE5F6DDECE9DED357ULL, "6036950_m.fxe"},
	{0xE5F7B99087C9DB97ULL, "6150962_m.fxe"},
	{0xE5F7E680FE6158CEULL, "6082237_m.fxe"},
	{0xE5F80320191E284AULL, "dre100cr_dead_hurlock.utc"},
	{0xE5F848376BEB2319ULL, "vxlt_gallowstemplar_38p_0_0.vlm"},
	{0xE5F876E68BC694E4ULL, "776717108.ogg"},
	{0xE5F8AE4FDED1F09AULL, "w_bonepitmining_a2_37_0.mmh"},
	{0xE5F8B24FDED1F7D1ULL, "w_bonepitmining_a2_37_0.mao"},
	{0xE5F8C04FDED20FE0ULL, "w_bonepitmining_a2_37_0.msh"},
	{0xE5F925B5B0641327ULL, "lla_ht_0_hightown_a3_65538_1.dds"},
	{0xE5F9C17C53CA8120ULL, "llac_lt_warehouse_n_-65536_2.dds"},
	{0xE5FA05E8CA3D3792ULL, "6093961_m.fxe"},
	{0xE5FA3D8A80867937ULL, "art\\vfx\\combat\\vfx_sg_hurl_aoe_p.impa.ani"},
	{0xE5FA91D83C7B505DULL, "hlac_lt_0_lowtown_a3_n_c_-196609_4z.dds"},
	{0xE5FB16A86E1011F9ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_lgtb.phy"},
	{0xE5FB34323C449B55ULL, "vxlt_0_lowtown_ekf_0_2.vlm"},
	{0xE5FBBD450B3FEEF0ULL, "mcr370_messenger.cnv"},
	{0xE5FC02E8CDCB8F80ULL, "vxca_cavern1_haunted_fn4_0_0.vlm"},
	{0xE5FC6582D1D71666ULL, "tools_light_lg.phy"},
	{0xE5FCA085DAE52BDBULL, "zz_start_qcr6081930.cl"},
	{0xE5FCC20336540288ULL, "vxht_mansion2_2uh_0_0.vlm"},
	{0xE5FCC85753E03F8BULL, "vxht_0_hightown_a2_c45_0_0.vlm"},
	{0xE5FDBAB2073F033FULL, "hlac_lt_0_lowtown_a2_-196609_2z.dds"},
	{0xE5FDE908EBD4C4FEULL, "6142519_m.fxe"},
	{0xE5FE746570B87713ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_meda_0d.dds"},
	{0xE5FE863C137A4B22ULL, "vxlt_gallowstemplar_n_51e_0_0.vlm"},
	{0xE5FEA603D2223E9CULL, "vxlt_gallowsprison_n_6gf_0_0.vlm"},
	{0xE5FEC36692CE9A48ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm02.mmh"},
	{0xE5FEDF450B425E48ULL, "mcr370_messenger.bnk"},
	{0xE5FEE098BF4BCE67ULL, "6209891_m.fxe"},
	{0xE5FEE16692CECD72ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_ant_dm02.msh"},
	{0xE5FF2F44A1728C41ULL, "vxlt_0_lowtown_a2_gvy_0_0.vlm"},
	{0xE5FF32CD52E052A8ULL, "hlac_ht_mansion2b_818_48x.dds"},
	{0xE5FF6C3A4F6CA91DULL, "seb321_leliana6180973.cl"},
	{0xE5FF79EDB3525A8AULL, "vxht_keep_5jx_0_0.vlm"},
	{0xE5FFE244245E2438ULL, "877677411.ogg"},
	{0xE5FFE5E91EB59E19ULL, "6049428_m.fxe"},
	{0xE5FFFDB46CCB4338ULL, "vxht_bartrand_n_3ss_0_0.vlm"},
	{0xE600DA7DCA96C680ULL, "676073641.ogg"},
	{0xE60114FDC7680C7DULL, "vxlt_darktown_5jh_0_0.vlm"},
	{0xE6016299773A9A76ULL, "409448986.ogg"},
	{0xE601C37ED4608ACAULL, "6018667_m.fxe"},
	{0xE601EB2C3C373E11ULL, "vxw_woundedcoast_vz_0_0.vlm"},
	{0xE60239BA9C668590ULL, "llac_dr_0_deeproads_17035_0.dds"},
	{0xE602D478DBF41106ULL, "vxlt_gallowstemplar_n_5r4_0_0.vlm"},
	{0xE603A810457AA675ULL, "vxca_caverns2prc_ae0_0_0.vlm"},
	{0xE6041B3BE40EE113ULL, "vxca_caverns2_7w3_0_0.vlm"},
	{0xE6044759338AD644ULL, "illuminationmask.psh.fxo"},
	{0xE604823870905FFAULL, "582364266.ogg"},
	{0xE604E3D4F0EAABD8ULL, "dae000cr_grand_cleric6147689.cl"},
	{0xE604F0F482DBA86AULL, "dae000_meredith6085242.cl"},
	{0xE605085810D94391ULL, "vxw_bonepit_1zl_0_0.vlm"},
	{0xE6052ECF3D20CF27ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.impa.gad"},
	{0xE6053E4F04DC95A6ULL, "6120783_f.fxe"},
	{0xE6058894785B92C0ULL, "gen00fl_sebastian6179761.cl"},
	{0xE606DB0657494359ULL, "273841671.ogg"},
	{0xE6073C4FEA831B9DULL, "306218407.ogg"},
	{0xE6073D631A495B4BULL, "143117309.ogg"},
	{0xE60760330927623DULL, "vxlt_gallowstemplar_3qw_0_0.vlm"},
	{0xE6076624E399A48DULL, "6116078_m.fxe"},
	{0xE60788918795A2CAULL, "6052957_m.fxe"},
	{0xE60810D652628D37ULL, "gen_im_cft_reg_l1.uti"},
	{0xE6083EE4E71F80BFULL, "6032918_m.fxe"},
	{0xE6094709E6453DDCULL, "dae000_bodhan_letter.cnv"},
	{0xE6095C3389B96EC8ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzshelf.mmh"},
	{0xE609683389B9836BULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzshelf.mao"},
	{0xE6097A3389B9A1F2ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zzshelf.msh"},
	{0xE60980A5E90DF8B0ULL, "vxca_caverns2_7xy_0_0.vlm"},
	{0xE609D99CFE5CF765ULL, "ave221_brennan6110476.cl"},
	{0xE609E3349D69CCE8ULL, "334684790.ogg"},
	{0xE609FC2CECDCB9CEULL, "lt_0_lowtown_docks_n_135_0.msh"},
	{0xE60A0A2CECDCD11FULL, "lt_0_lowtown_docks_n_135_0.mao"},
	{0xE60A162CECDCE5BCULL, "lt_0_lowtown_docks_n_135_0.mmh"},
	{0xE60A3611A709D05DULL, "492749002.ogg"},
	{0xE60A89E25FA777D0ULL, "653489849.ogg"},
	{0xE60ADE31B46190B0ULL, "mcr370_negotiation6080208.cl"},
	{0xE60AFA9C932F64B9ULL, "6042759_diplomatic_solution.cut"},
	{0xE60B198EB342D452ULL, "mag100_cullen6140959.cl"},
	{0xE60B22246C1E4A56ULL, "405803176.ogg"},
	{0xE60B3A3445D6580BULL, "vxht_0_hightown_a2_c34_0_0.vlm"},
	{0xE60B4F815032E6B9ULL, "6125656_m.fxe"},
	{0xE60BA6F7944B4042ULL, "static_static_p5.psh.fxo"},
	{0xE60C23CF3D26DB34ULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.impa.evt"},
	{0xE60C64FA4CF65B22ULL, "600036997.ogg"},
	{0xE60C6909E647AD2CULL, "dae000_bodhan_letter.bnk"},
	{0xE60C74076B584AEFULL, "6133051_m.fxe"},
	{0xE60CA4680F7958ECULL, "hlac_lt_gallowstemplar_f_2384_3y.dds"},
	{0xE60CF6EFB9548221ULL, "hlac_lt_0_lowtown_a2_n_-196608_3y.dds"},
	{0xE60CFA251891A623ULL, "vxlt_gallowstemplar_f_6s1_0_0.vlm"},
	{0xE60D3080F12A7299ULL, "1016490691.ogg"},
	{0xE60D8B8BDC1B0359ULL, "783213292.ogg"},
	{0xE60DA1219368A4C1ULL, "vxlt_0_lowtown_a2_gmx_0_0.vlm"},
	{0xE60DE63FBC192EC3ULL, "6179879_m.fxe"},
	{0xE60E1DA6108B000DULL, "6100244_m.fxe"},
	{0xE60E4A70ED05ECF7ULL, "19313569.ogg"},
	{0xE60EA1E11D792311ULL, "zz_and_debug6178654.cl"},
	{0xE60EDF9B72ED2CE2ULL, "6033717_f.fxe"},
	{0xE60F3AA85F20DA11ULL, "zz_qun_debug6085577.cl"},
	{0xE60F86DAB95E24FDULL, "vxlt_darktown_5og_0_0.vlm"},
	{0xE60F9BCD1D7A9E0FULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hoo_nobc_0n.dds"},
	{0xE60FB65E3AD6D6E6ULL, "6118808_m.fxe"},
	{0xE61002BA3AD0CE2AULL, "zz_followers_banter6159039.cl"},
	{0xE6104EF29276BA40ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sw106a.mao"},
	{0xE6105D092E2D5691ULL, "vxw_woundedcoast_sy_0_0.vlm"},
	{0xE6108C508A7C5445ULL, "vxlt_blackemporium_10k_0_0.vlm"},
	{0xE610B80182FF97E1ULL, "54814287.ogg"},
	{0xE61179D057C6D3F0ULL, "414932746.ogg"},
	{0xE611A549EF827DC5ULL, "vxw_encounter3_8z_0_0.vlm"},
	{0xE6122E4EF6A16ED6ULL, "fx_f1c1.mao"},
	{0xE6124C7CCE43B47EULL, "vxdr_0_deeproads_ad5_0_0.vlm"},
	{0xE61262F754660DE7ULL, "vxlt_0_lowtown_a2_n_c_fab_0_0.vlm"},
	{0xE61264FDC7A24870ULL, "6072096_m.fxe"},
	{0xE6133211C1C21CA7ULL, "clf260_harley6089571.cl"},
	{0xE613DBE856426550ULL, "55365523.ogg"},
	{0xE6146C550176B458ULL, "vxlt_gallowstemplar_f_3iu_0_0.vlm"},
	{0xE61489566EA541BFULL, "vxlt_gallowscourtyard_n_691_0_0.vlm"},
	{0xE614C0EED3ADC215ULL, "tv_dust.mao"},
	{0xE6151675A717771CULL, "w_bonepit_128_0.mao"},
	{0xE6152275A7178BBFULL, "w_bonepit_128_0.mmh"},
	{0xE6152875A7179661ULL, "w_bonepit_128_0.msh"},
	{0xE615524037C9E1B5ULL, "mag131_keran6086261.cl"},
	{0xE615B7D9EE66B651ULL, "ave200_aveline_bar6116183.cl"},
	{0xE615BF481C8E0D83ULL, "dae021pt_brothel.ncs"},
	{0xE615CDEFA4567FBAULL, "gen_im_trs_dem_junk2_03.uti"},
	{0xE61620F495A78F70ULL, "6129327_m.fxe"},
	{0xE6163DA901AF0AD2ULL, "vxht_0_hightown_a2_9xf_0_0.vlm"},
	{0xE6165B3FC70A77C8ULL, "art\\characters\\playercharacter\\elffemale\\rob_enco.msh"},
	{0xE6166D3FC70A9659ULL, "art\\characters\\playercharacter\\elffemale\\rob_enco.mao"},
	{0xE616793FC70AAAF2ULL, "art\\characters\\playercharacter\\elffemale\\rob_enco.mmh"},
	{0xE617008740A05C3FULL, "vxht_chantry_2td_0_0.vlm"},
	{0xE617E9803F260D9CULL, "vxht_mansion_n_4xf_0_0.vlm"},
	{0xE618A0F6C6966F57ULL, "492488547.ogg"},
	{0xE6195678761319D9ULL, "art\\characters\\weapons\\shields\\w_shd_lrdb.msh"},
	{0xE619607876132AC7ULL, "art\\characters\\weapons\\shields\\w_shd_lrdb.mmh"},
	{0xE619647876133114ULL, "art\\characters\\weapons\\shields\\w_shd_lrdb.mao"},
	{0xE619E316E4738D4FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hm_exp_movement.anb"},
	{0xE61A06EA296902C0ULL, "6100308_m.fxe"},
	{0xE61A1BCF3D330B6DULL, "art\\vfx\\combat\\vfx_m_ent_misdirec_lp_c.impa.ani"},
	{0xE61AD759DE6D1CACULL, "dae021_keep_guard_3.utc"},
	{0xE61ADD3377C0ECB3ULL, "vxw_blightlands_62_0_0.vlm"},
	{0xE61AFA2A69973DB9ULL, "6162518_m.fxe"},
	{0xE61B06DE015B1D4FULL, "182512790.ogg"},
	{0xE61B217210CD905BULL, "vxlt_0_lowtown_docks_n_baq_0_0.vlm"},
	{0xE61B89394C9602E0ULL, "mcr371_hostile_aveline6082195.cl"},
	{0xE61B89BDB60E6F9CULL, "vxlt_gallowsprison_n_6md_0_0.vlm"},
	{0xE61C014D5CE25C89ULL, "vxht_mansion_n_5cc_0_0.vlm"},
	{0xE61C7B62F86D6446ULL, "176207722.ogg"},
	{0xE61C820239A65F2AULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls109a_0d.dds"},
	{0xE61C82BFBEA4184FULL, "6101436_m.fxe"},
	{0xE61CC366A2984A45ULL, "crt_emissary.bnk"},
	{0xE61D1B44B611650EULL, "6209326_m.fxe"},
	{0xE61D43FA6263686BULL, "262693573.ogg"},
	{0xE61D9C6FA01425E7ULL, "lt_backalley_n_mmap_ph.dds"},
	{0xE61DAA73E77711ABULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_1.pwk"},
	{0xE61DF8B7AB543D7DULL, "vxlt_darktown_5pf_0_0.vlm"},
	{0xE61E6B9484714B87ULL, "6043594_m.fxe"},
	{0xE61E6F2937B526A8ULL, "seb321_leliana6199284.cl"},
	{0xE61E7BD778F253CEULL, "art\\characters\\playercharacter\\elfmale\\rob_appc.msh"},
	{0xE61E84A79ABFD0C8ULL, "6163750_m.fxe"},
	{0xE61E95D778F27FBCULL, "art\\characters\\playercharacter\\elfmale\\rob_appc.mmh"},
	{0xE61EAF312CD81D56ULL, "vxlt_gallowsprison_n_6n7_0_0.vlm"},
	{0xE61ECB42D36C183AULL, "6151115_m.fxe"},
	{0xE61ED621271FECFAULL, "vxlt_0_lowtown_a2_euq_0_0.vlm"},
	{0xE61F61E7A631060AULL, "vxht_0_hightown_c57_0_0.vlm"},
	{0xE61F6576034F5820ULL, "374120058.ogg"},
	{0xE61FB5C1E3BE693FULL, "516500941.ogg"},
	{0xE6201726E1789645ULL, "vxw_encounter3_5y_0_0.vlm"},
	{0xE620185FC4C9C969ULL, "6096851_m.fxe"},
	{0xE6208148F4E49110ULL, "vxht_brothel_n_5hx_0_0.vlm"},
	{0xE62140906FF44D2EULL, "dae300_cullen6140615.cl"},
	{0xE621827F047313BAULL, "6164850_m.fxe"},
	{0xE6220768A679B884ULL, "202071303.ogg"},
	{0xE62235B20DDDA74AULL, "lgt210_eau_symp_templar6144737.cl"},
	{0xE62237A52CE66A76ULL, "6168142_m.fxe"},
	{0xE6224DAC1ED69258ULL, "dre000_varric6029385.cl"},
	{0xE623E246C8EF98C5ULL, "zz_qcr_debug6056849.cl"},
	{0xE623E7722A6425EEULL, "264725204.ogg"},
	{0xE624AE4FDEF72BB1ULL, "w_bonepitmining_a2_37_0.rml"},
	{0xE624D134353D99EFULL, "vxlt_gallowscourtyard_n_6ee_0_0.vlm"},
	{0xE624F59A1251FC67ULL, "vxw_bonepit_13q_0_0.vlm"},
	{0xE625218E478F40DCULL, "6149480_m.fxe"},
	{0xE6253FFD222B966AULL, "vxlt_0_lowtown_a2_n_eki_0_1.vlm"},
	{0xE6258826EEFCA9DDULL, "492658696.ogg"},
	{0xE6258E13DAF2A0ADULL, "vxht_0_hightown_a3_c7g_0_0.vlm"},
	{0xE62592D92DFCAF79ULL, "vxw_sundermount_23o_0_0.vlm"},
	{0xE626F34D983B1CCDULL, "lt_0_lowtown_docks_a3_11_0_lc.mmh"},
	{0xE6270EE612992577ULL, "6061809_f.fxe"},
	{0xE627390194BCD939ULL, "1023726196.ogg"},
	{0xE6276BF36B245CB8ULL, "6167552_m.fxe"},
	{0xE6276DD7DFF4D69FULL, "gen00fl_varric6104950.cl"},
	{0xE6276EB5C2BAFCA6ULL, "801453010.ogg"},
	{0xE6278994B65750FCULL, "6117304_m.fxe"},
	{0xE6279173B672C87AULL, "vxht_0_hightown_n_b4z_0_0.vlm"},
	{0xE627C7C03D8579A3ULL, "mrl150_merrill6070504.cl"},
	{0xE62843FD61F9BC48ULL, "w_sundermount_25_0.mmh"},
	{0xE6284FFD61F9D0EBULL, "w_sundermount_25_0.mao"},
	{0xE62861FD61F9EF72ULL, "w_sundermount_25_0.msh"},
	{0xE628D1F89C3548DBULL, "336449822.ogg"},
	{0xE6290083D7AC473DULL, "character_character_p138.psh.fxo"},
	{0xE6290DE2C199D142ULL, "6150379_f.fxe"},
	{0xE6296C075B8D5639ULL, "6167618_m.fxe"},
	{0xE62A11358D40F9B8ULL, "6110297_m.fxe"},
	{0xE62B13253A88F378ULL, "gen00fl_isabela6088839.cl"},
	{0xE62B81A15F9BFBFCULL, "docks_scene_dockworkers.cnv"},
	{0xE62BD97C8DABF006ULL, "693776475.ogg"},
	{0xE62C35AC02D62818ULL, "wm_ico_camp.dds"},
	{0xE62C62E216E86865ULL, "llac_lt_gallowstemplar_2394_5.dds"},
	{0xE62C6A949D4A55FDULL, "vxlt_darktown_5ue_0_0.vlm"},
	{0xE62CC6CC51A460DAULL, "vxlt_backalley_n_1f0_0_0.vlm"},
	{0xE62D210E1ECE35D6ULL, "vxlt_gallowsprison_n_6k8_0_0.vlm"},
	{0xE62D3C4348F7846DULL, "729445873.ogg"},
	{0xE62D47FE1916057AULL, "vxlt_0_lowtown_a2_exr_0_0.vlm"},
	{0xE62DC5956E849F19ULL, "ico_mm_enemy.dds"},
	{0xE62DFD0E651C8838ULL, "vxlt_gallowstemplar_n_5cy_0_0.vlm"},
	{0xE62E2A0FB1D65A86ULL, "vxlt_gallowstemplar_n_5c7_0_0.vlm"},
	{0xE62EA3A15F9E6B4CULL, "docks_scene_dockworkers.bnk"},
	{0xE62EC9E08B174A0FULL, "vxw_encounter2_7j_0_0.vlm"},
	{0xE62EF52CECFBC987ULL, "lt_0_lowtown_docks_n_135_0.phy"},
	{0xE62FFECAD27E15A2ULL, "6167019_m.fxe"},
	{0xE6303F4EF6BA8958ULL, "fx_f1c1.dds"},
	{0xE6304A4B2FA60F67ULL, "vxw_woundedcoast_4j_0_0.vlm"},
	{0xE630E993ABB24D10ULL, "mag110_cut_abomination6037105.cl"},
	{0xE630F60C32CDC26DULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef02.msh"},
	{0xE631080C32CDE0F3ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef02.mmh"},
	{0xE6313D822225BCD7ULL, "hlac_w_sundermount_131074_0z.dds"},
	{0xE631500EE562E558ULL, "vxlt_gallowstemplar_f_3ss_0_0.vlm"},
	{0xE63178B046197635ULL, "411572504.ogg"},
	{0xE6327EB9C72E2240ULL, "hla_ht_0_hightown_n_-131072_0z.dds"},
	{0xE6327FEB151A33AAULL, "6128137_m.fxe"},
	{0xE63292B375E1AF3BULL, "hlac_ca_caverns2prc_5802_25x.dds"},
	{0xE632A1C13311429BULL, "zz_and_debug6134090.cl"},
	{0xE632A6A4B9FCCFDFULL, "6041241_f.fxe"},
	{0xE633096288248434ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_player_andersintro.ani"},
	{0xE6331183ACFF5067ULL, "vxlt_0_lowtown_docks_a3_cx3_0_0.vlm"},
	{0xE6333BD22DDB8612ULL, "408963626.ogg"},
	{0xE633499525672039ULL, "character_character_p147.psh.fxo.dx11.assembly"},
	{0xE633DADD6149B73EULL, "824264501.ogg"},
	{0xE633E81A01A4F893ULL, "vxlt_gamlenshouse_n_qp_0_0.vlm"},
	{0xE63404B61FF2C7F9ULL, "vxw_sundermount_24n_0_0.vlm"},
	{0xE635A29EBC481988ULL, "dae000_orsino6085512.cl"},
	{0xE635C25293CB3729ULL, "hlac_lt_0_lowtown_a2_n_131073_2x.dds"},
	{0xE6360E5E00C2CE02ULL, "675704009.ogg"},
	{0xE6365BFF9B3F30ACULL, "plt_dre000pt_extra_funds.plo"},
	{0xE636A45A9759291DULL, "vxlt_undercityserialkiller_1mo_0_0.vlm"},
	{0xE636C571470E88ACULL, "art\\characters\\playercharacter\\humanfemale\\glv_medp.phy"},
	{0xE636C9125D69FF63ULL, "hightown_male_noble2_mob6170898.cl"},
	{0xE63740ACB3BF1710ULL, "s_aoe_wispy.mao"},
	{0xE63756F1BCC01E04ULL, "440009348.ogg"},
	{0xE637B281FD0DDB1FULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_005a_0n.dds"},
	{0xE6385395DAD258CEULL, "fen310_varania6118578.cl"},
	{0xE638E639DB573F03ULL, "vxw_encounter2_camp_1t_0_0.vlm"},
	{0xE6391F074705A7CEULL, "87514470.ogg"},
	{0xE639969D40D9E28BULL, "6041471_m.fxe"},
	{0xE639BAC1881F3D12ULL, "vxdr_0_deeproads_eqn_0_0.vlm"},
	{0xE639E2F505C5EA7FULL, "one200_gold_followers6120137.cl"},
	{0xE63A0662882A9D96ULL, "art\\cinematics\\follower plots\\anders\\act1\\cs_player_andersintro.gad"},
	{0xE63A11E9BAE9ADE9ULL, "hlac_lt_0_lowtown_a2_n_-1_1x.dds"},
	{0xE63A574DACC6F9FBULL, "vxht_0_hightown_a3_c8p_0_0.vlm"},
	{0xE63A7DEDC70CFF90ULL, "6114420_m.fxe"},
	{0xE63AE08FFA509722ULL, "825991916.ogg"},
	{0xE63B1CBC409AA136ULL, "6018713_m.fxe"},
	{0xE63B46D9557EC997ULL, "fex_ruin_arch_lrg_b.msh"},
	{0xE63B5CD9557EEF09ULL, "fex_ruin_arch_lrg_b.mmh"},
	{0xE63B94D7B5098827ULL, "6135985_m.fxe"},
	{0xE63BC1348707B131ULL, "hlac_lt_gallowscourtyard_-1_3x.dds"},
	{0xE63BCA42A674C0AAULL, "gen_im_trs_crc_junk1_02.uti"},
	{0xE63C5021A5C4A2D7ULL, "6096493_m.fxe"},
	{0xE63C5C2264E4361DULL, "6108095_m.fxe"},
	{0xE63CC3E680ADB1DEULL, "423028834.ogg"},
	{0xE63D0B9B89461D25ULL, "6128981_m.fxe"},
	{0xE63D3BBD7D0D628FULL, "vxw_encounter2_2u_0_0.vlm"},
	{0xE63D40FF9B4521E3ULL, "plt_dre000pt_extra_funds.nss"},
	{0xE63E0405A59661C8ULL, "vxca_caverns1_i0x_0_0.vlm"},
	{0xE63E75F22B1E91FCULL, "913515211.ogg"},
	{0xE63E843EC5560F8DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ff_basic.mmh"},
	{0xE63E883EC55616DEULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ff_basic.mao"},
	{0xE63E963EC5562E13ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ff_basic.msh"},
	{0xE63EBC28219C27E7ULL, "vxw_woundedcoast_1e_0_0.vlm"},
	{0xE63ED247835DBC22ULL, "mer210_smith6051410.cl"},
	{0xE63F2B0FBA43D96FULL, "nat380_nathaniel_banter6090046.cl"},
	{0xE63F36441D191BEEULL, "377392593.ogg"},
	{0xE6401ADA624D6531ULL, "mcr370_moira6087502.cl"},
	{0xE640C68550FFCE38ULL, "llac_lt_0_lowtown_docks_a3_-131075_1.dds"},
	{0xE641909D820C3073ULL, "1856193.ogg"},
	{0xE641B4EE1929CAEFULL, "vxlt_gallowscourtyard_n_6og_0_0.vlm"},
	{0xE642177E987BB6CAULL, "vxht_0_hightown_apj_0_0.vlm"},
	{0xE64287F46CCD93FAULL, "vxlt_0_lowtown_n_eoq_0_1.vlm"},
	{0xE6429F528515FD02ULL, "hlac_w_encounter3_-327682_0z.dds"},
	{0xE6432AD67EC37489ULL, "vxca_caverns1_mines_fmh_0_0.vlm"},
	{0xE643321DDD47D5BDULL, "349813465.ogg"},
	{0xE643731A98BAB6D1ULL, "119497587.ogg"},
	{0xE64374D779116387ULL, "art\\characters\\playercharacter\\elfmale\\rob_appc.phy"},
	{0xE644086008EF6967ULL, "98219653.ogg"},
	{0xE6443A6B5E4C3B44ULL, "331226144.ogg"},
	{0xE644664DA67A4FC2ULL, "vxlt_0_lowtown_docks_b0p_0_0.vlm"},
	{0xE644A33994A5FD6CULL, "302329720.ogg"},
	{0xE644F557D86B9A7EULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basket_0.pwk"},
	{0xE644FACF32197EBEULL, "557927472.ogg"},
	{0xE6451DCA81E1D132ULL, "llac_lt_0_lowtown_docks_a3_65533_3.dds"},
	{0xE6452A9E51C86EF5ULL, "vxlt_0_lowtown_a2_n_c_hz7_0_0.vlm"},
	{0xE6454A91A58BAE78ULL, "95700922.ogg"},
	{0xE645903EF9725FCCULL, "vxlt_0_lowtown_a2_n_c_i0a_0_0.vlm"},
	{0xE6459164B6417335ULL, "6080261_m.fxe"},
	{0xE645ADF5FE446564ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax104a_0d.dds"},
	{0xE646422E9A7AC820ULL, "character_character_p203.psh.fxo.dx11"},
	{0xE646523B479DFBF3ULL, "plt_mrl000pt_gift_ring.nss"},
	{0xE646B23AD30F517DULL, "547133796.ogg"},
	{0xE646D2EA98639403ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_enter.ani"},
	{0xE647055087D2FA8CULL, "lt_0_lowtown_docks_a3_118_0.phy"},
	{0xE6473EED282D7BAEULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation3.ani"},
	{0xE6476B89C06F12A0ULL, "vxht_0_hightown_b8s_0_0.vlm"},
	{0xE647D798D8C66939ULL, "vxlt_warehouse_n_15s_0_0.vlm"},
	{0xE648C92F7F931C07ULL, "vxht_brothel_n_6gq_0_0.vlm"},
	{0xE648EFCAB9031810ULL, "6113520_m.fxe"},
	{0xE6490EEF274C1204ULL, "hlac_lt_0_lowtown_docks_-131073_0x.dds"},
	{0xE6495BB9D9973845ULL, "390024591.ogg"},
	{0xE6495C29E1E26E94ULL, "vxlt_gallowstemplar_f_5pd_0_0.vlm"},
	{0xE64987F201AE0DFCULL, "6090812_m.fxe"},
	{0xE649AA86359091DAULL, "vxlt_backalley_n_1l2_0_0.vlm"},
	{0xE649AE8D28376306ULL, "gen00fl_varric6118822.cl"},
	{0xE649BC6D48F2619DULL, "hlac_lt_0_lowtown_docks_-2_1y.dds"},
	{0xE649CDC52B947C45ULL, "vxlt_0_lowtown_a2_euz_0_0.vlm"},
	{0xE649FC880E801B44ULL, "qun230cr_guild_contact6068874.cl"},
	{0xE64A15C139F2C499ULL, "hlac_ca_caverns1_mines_24425_7y.dds"},
	{0xE64A7DCCA7F1A93FULL, "6102799_f.fxe"},
	{0xE64AC90117503FB4ULL, "hlac_lt_0_lowtown_n_65536_1x.dds"},
	{0xE64B9AD85DA2E4C5ULL, "hlac_lt_0_lowtown_n_-3_1x.dds"},
	{0xE64BC8A48BDD96B2ULL, "vxlt_warehouse_a2_0_0.vlm"},
	{0xE64BD5539DA8369AULL, "vxht_chantry_n_1yr_0_0.vlm"},
	{0xE64BF9489F2D8506ULL, "vxdr_0_deeproads_d3q_0_0.vlm"},
	{0xE64C00F69947AF8FULL, "vxw_blightlands_ac_0_0.vlm"},
	{0xE64C5641219CD0B5ULL, "gencr_darkspawn_hurlock_rng.utc"},
	{0xE64C81597B9C79F8ULL, "hlac_lt_0_lowtown_1_2z.dds"},
	{0xE64CD8CF9C51050BULL, "hd_spla.mao"},
	{0xE64E8005FC213C58ULL, "gen00fl_isabela6103925.cl"},
	{0xE64EDC05D17C5F82ULL, "31582706.ogg"},
	{0xE64F653E525E19D3ULL, "zz_varricarrows_tst.cut"},
	{0xE64F7D3EC564774AULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0ff_basic.phy"},
	{0xE64F9D01350BD355ULL, "20511332.ogg"},
	{0xE64FB399F243D024ULL, "77553241.ogg"},
	{0xE650222D1A634083ULL, "vxca_caverns1_j5w_0_0.vlm"},
	{0xE6506ADEF35F5EA5ULL, "air_of_confidence.dds"},
	{0xE65079DDCDB6F968ULL, "194830089.ogg"},
	{0xE65092B136CD9BCAULL, "vxca_caverns2_9fo_0_0.vlm"},
	{0xE6520672A6866AD5ULL, "359073103.ogg"},
	{0xE652279921510622ULL, "6041261_m.fxe"},
	{0xE652B449547EEBAAULL, "6140562_m.fxe"},
	{0xE652D82A98706842ULL, "vxlt_0_lowtown_docks_b7q_0_0.vlm"},
	{0xE652E3FC1F9B13C9ULL, "vxlt_darktown_6c8_0_0.vlm"},
	{0xE653407102CB0282ULL, "868314400.ogg"},
	{0xE6539C7B43BE8775ULL, "vxlt_0_lowtown_a2_n_c_hu6_0_0.vlm"},
	{0xE653B8DD2783FFFFULL, "fx_firetorchfb_add.mao"},
	{0xE653E1C0123AF2AEULL, "lgt000cr_hate_carta.utc"},
	{0xE653EAEA986E47FAULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_enter.evt"},
	{0xE654B899B96228D2ULL, "6113607_m.fxe"},
	{0xE6552465CFC0BABEULL, "gen00fl_varric6087586.cl"},
	{0xE65547FBF2BC600CULL, "lt_0_lowtown_a2_mmi.gff"},
	{0xE6556862432BF0D8ULL, "61745065.ogg"},
	{0xE655E28570DE043CULL, "142748059.ogg"},
	{0xE655F19FAF3F45D1ULL, "art\\characters\\creatures\\stonegolem\\textures\\cn_hed_snga_0i.dds"},
	{0xE6560BD3A6010C7FULL, "6135003_f.fxe"},
	{0xE6562568BA6506E9ULL, "6086906_m.fxe"},
	{0xE6567902C7B4396BULL, "vxlt_ruinedhouse_n_2u_0_0.vlm"},
	{0xE6569D994D7FFC05ULL, "6047524_f.fxe"},
	{0xE6569E7B6C0B6E12ULL, "vxdr_0_deeproads_egh_0_0.vlm"},
	{0xE656BE40DEBC1A58ULL, "zz_followers_banter6158610.cl"},
	{0xE65713B9E3F41E17ULL, "wm_ico_road_roll.dds"},
	{0xE6573B0790B32AFBULL, "vxht_0_hightown_a3_c6r_0_0.vlm"},
	{0xE657726D16B4B415ULL, "hlac_lt_0_lowtown_-2_6y.dds"},
	{0xE6581C632786AA5AULL, "vxlt_backalley_n_1w3_0_0.vlm"},
	{0xE6583FA21D8A94C5ULL, "vxlt_0_lowtown_a2_exy_0_0.vlm"},
	{0xE6585C355719DF06ULL, "6143970_m.fxe"},
	{0xE658B621F8F33AB3ULL, "hm_gen_slaver.mor"},
	{0xE658BF9F95384511ULL, "hlac_lt_0_lowtown_a2_n_131072_5z.dds"},
	{0xE658DCCB2AD962CCULL, "mag330_walter6044812.cl"},
	{0xE6591AC162BE86B9ULL, "hlac_lt_backalley_g_0_11y.dds"},
	{0xE659496D2CD1F603ULL, "dae130ip_at_to_sewers2.utp"},
	{0xE659514285043415ULL, "hlac_ca_caverns2_5802_44z.dds"},
	{0xE65963835AAD2BBCULL, "llac_ht_mansion_n_3369_1.dds"},
	{0xE659648C45F5624CULL, "t1_mub_or1.tnt"},
	{0xE6598A074DC81B08ULL, "23549568.ogg"},
	{0xE659ACA96E47F362ULL, "darktown_male_elf_sta6174560.cl"},
	{0xE659E764F91BC8A7ULL, "6018640_m.fxe"},
	{0xE659E8B15A61BFE6ULL, "326494264.ogg"},
	{0xE65A1F7760F9938FULL, "vxw_encounter2_8w_0_0.vlm"},
	{0xE65A3E5878BE1E1BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_twitch_v2.evt"},
	{0xE65A3F518D6F6CA6ULL, "vxlt_gallowscourtyard_6wi_0_0.vlm"},
	{0xE65A9E14DB997F9CULL, "6121943_m.fxe"},
	{0xE65ACA9596D733C8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spitoon.msh"},
	{0xE65AD65AEE53A5DAULL, "vxw_blightlands_11k_0_0.vlm"},
	{0xE65ADC9596D75259ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spitoon.mao"},
	{0xE65AE89596D766F2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spitoon.mmh"},
	{0xE65AE9EA987464EDULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_pacing_enter.gad"},
	{0xE65AFE69852631FAULL, "vxht_keep_n_6gk_0_0.vlm"},
	{0xE65BF8C0A6F18D1AULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeceltic.phy"},
	{0xE65C0BCCA9D783DBULL, "crack_g05.dds"},
	{0xE65C16995D384489ULL, "zz_fen_debug6165487.cl"},
	{0xE65C17ED283F95E8ULL, "art\\cinematics\\additional animation\\cs_genericfemale_additionalanimation3.gad"},
	{0xE65C32B15D70D84EULL, "vxw_encounter2_28_0_0.vlm"},
	{0xE65D21591C9F59AFULL, "vxdr_primevaltomb_7yo_0_0.vlm"},
	{0xE65D29B3C7CD0C8FULL, "vxlt_0_lowtown_a3_n_c_bz2_0_0.vlm"},
	{0xE65D5C4316812F4EULL, "hlac_lt_0_lowtown_docks_a3_-131073_1z.dds"},
	{0xE65D8BB149109388ULL, "vxlt_gallowscourtyard_7b1_0_0.vlm"},
	{0xE65E1C1BEFDA8815ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls105a_0s.dds"},
	{0xE65EFC7C7D070295ULL, "vxw_sundermount_31t_0_0.vlm"},
	{0xE65F1F7251B4506EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tray_d.dds"},
	{0xE65F3DB0D7874213ULL, "vxlt_gamlenshouse_n_bu_0_0.vlm"},
	{0xE65F421102C4ABFEULL, "63488966.ogg"},
	{0xE65F4CC85A14BD8FULL, "hlac_lt_foundry_n_320_22z.dds"},
	{0xE65FE6A3CE7FC86FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galoflipalco.mao"},
	{0xE65FF2A3CE7FDC0CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galoflipalco.mmh"},
	{0xE65FF8A3CE7FE6FEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_galoflipalco.msh"},
	{0xE6606C72F16C89B3ULL, "6059687_m.fxe"},
	{0xE660D6AF2F4DD546ULL, "6139229_m.fxe"},
	{0xE660EC62061AA1E8ULL, "6118433_m.fxe"},
	{0xE661375878C4315CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_twitch_v2.gad"},
	{0xE661660A1EF6499FULL, "vxht_mansion2_3cd_0_0.vlm"},
	{0xE661844C29D5EA3AULL, "604029078.ogg"},
	{0xE6620E5835B49FF5ULL, "vxlt_0_lowtown_a2_n_c_hp5_0_0.vlm"},
	{0xE6625AFE12749818ULL, "lgt260_way_blood_mage6148372.cl"},
	{0xE6626F7263A45959ULL, "vxlt_backalley_g_1n0_0_2.vlm"},
	{0xE662A8FE0C80393FULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef01.mmh"},
	{0xE662AEFE0C8043E1ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef01.msh"},
	{0xE662C9E2C7C81C06ULL, "723592246.ogg"},
	{0xE6636A1A37872826ULL, "llac_w_encounter2_-65538_0.dds"},
	{0xE664EADFB9AA51EBULL, "vxlt_ruinedhouse_n_5t_0_0.vlm"},
	{0xE6654D6036E39519ULL, "hlac_lt_0_lowtown_a2_-1_2z.dds"},
	{0xE665710180CF1680ULL, "968990108.ogg"},
	{0xE6657B99F8BA99EBULL, "vxca_cavern1_haunted_e8f_0_0.vlm"},
	{0xE6658313AE01482CULL, "hlac_ca_caverns2_5802_63x.dds"},
	{0xE665ADCB6C87DC16ULL, "gen00fl_aveline6139563.cl"},
	{0xE665D336EC100B2DULL, "hlac_ca_cavern1_haunted_24424_25z.dds"},
	{0xE66616F9B730BDAAULL, "6015941_m.fxe"},
	{0xE6664E1E19505684ULL, "6139869_m.fxe"},
	{0xE666C4EA4BFF1987ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.dura.gad"},
	{0xE666DA010E0B133FULL, "470108575.ogg"},
	{0xE667365878C89B5AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_twitch_v2.ani"},
	{0xE66779FA8D7B9D32ULL, "140939130.ogg"},
	{0xE668B12E7F658526ULL, "vxlt_gallowscourtyard_6rj_0_0.vlm"},
	{0xE668E4B07D33E08FULL, "vxw_blightlands_om_0_0.vlm"},
	{0xE668EA07DEE5E934ULL, "hlac_w_encounter3_-131067_0x.dds"},
	{0xE6691B24759EC627ULL, "211423710.ogg"},
	{0xE6692EB8D4D3DF13ULL, "hlac_lt_gallowscourtyard_-65537_2x.dds"},
	{0xE6692EFCC79A3898ULL, "6173148_m.fxe"},
	{0xE6697046771C4A7AULL, "vxht_keep_n_6bd_0_0.vlm"},
	{0xE6697F87BB508D51ULL, "518430879.ogg"},
	{0xE66A4BCE167156B2ULL, "zz_and_debug6134134.cl"},
	{0xE66ACFADE9E1C147ULL, "6089825_f.fxe"},
	{0xE66AF23ADD582240ULL, "6087081_m.fxe"},
	{0xE66B546F00A84164ULL, "71686499.ogg"},
	{0xE66B548824D68400ULL, "113645343.ogg"},
	{0xE66B7584605122E9ULL, "440704808.ogg"},
	{0xE66B792F9689B03EULL, "vxht_mansion2b_28u_0_0.vlm"},
	{0xE66BD356EA2F1FA8ULL, "vxlt_0_lowtown_a2_n_c_hkm_0_0.vlm"},
	{0xE66C2E21A31CE211ULL, "33982383.ogg"},
	{0xE66C997FFE3114B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_bonus_stagger.evt"},
	{0xE66CAB82EA82E801ULL, "hlac_lt_0_lowtown_docks_a3_-65539_3x.dds"},
	{0xE66D048B0A639FA6ULL, "vxht_mansion2_3d8_0_0.vlm"},
	{0xE66D7926453DAB9FULL, "vxht_0_hightown_c3e_0_1.vlm"},
	{0xE66D890910CF8DD5ULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.anb"},
	{0xE66DB9EA4C052594ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.dura.evt"},
	{0xE66DC7AFACA1F875ULL, "486038956.ogg"},
	{0xE66E3C6E65CC6F8DULL, "6023947_m.fxe"},
	{0xE66E7A49A6BE2EEBULL, "qun131_party_comment6117259.cl"},
	{0xE66E94ADEAEAA09CULL, "vxht_mansion_n_4fx_0_0.vlm"},
	{0xE66EDDC7D8AA53E8ULL, "686790071.ogg"},
	{0xE66EF4F98B4B021EULL, "art\\vfx\\cinematic\\vfx_lyrium_cor_c.impa.ani"},
	{0xE66F2316D2D60BBFULL, "w_sundermount_182_0.phy"},
	{0xE66F4798A7DBD06DULL, "181662083.ogg"},
	{0xE66F69009E6F2715ULL, "6175001_f.fxe"},
	{0xE66FD7E710EC621FULL, "vxht_mansion2_3fg_0_0.vlm"},
	{0xE66FE99771F3A853ULL, "ser120_ghyslain6104317.cl"},
	{0xE66FF3F4CE5D2421ULL, "128314662.ogg"},
	{0xE6700DA8BE638F61ULL, "vxht_mansion_n_61b_0_0.vlm"},
	{0xE6706BCE5F318B1DULL, "vxlt_undercityserialkiller_1qk_0_0.vlm"},
	{0xE670DF59F2923B8FULL, "6080996_f.fxe"},
	{0xE670E5D7B1FBE42DULL, "398229860.ogg"},
	{0xE67112367B80A70BULL, "hlac_lt_gallowstemplar_n_2383_1y.dds"},
	{0xE671129EAC60AB01ULL, "art\\levels\\races\\design\\animation\\floortrapactivate.ani"},
	{0xE671417CD649BCBAULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls109a_0t.dds"},
	{0xE6714A414D84D832ULL, "and321st_base_1st_romance.stg"},
	{0xE6715FFF771FB028ULL, "gen00fl_sebastian6178366.cl"},
	{0xE671783FC75754EDULL, "art\\characters\\playercharacter\\elffemale\\rob_enco.phy"},
	{0xE6719337A8125E70ULL, "6064709_m.fxe"},
	{0xE671DA289A20C53BULL, "528938585.ogg"},
	{0xE6721BA1B541C8F6ULL, "art\\characters\\creatures\\mubari\\animation\\com_v1.evt"},
	{0xE6736311377CBEC2ULL, "llac_lt_0_lowtown_docks_n_-2_4.dds"},
	{0xE673907FFE372441ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_bonus_stagger.gad"},
	{0xE673AD13A020AA1EULL, "6089113_m.fxe"},
	{0xE673B0E662F5D741ULL, "399108515.ogg"},
	{0xE673C0CF85D64DE9ULL, "dae200_cullen6140318.cl"},
	{0xE6744F4B637CFE4BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_marketstall_0.pwk"},
	{0xE6745D2131771ECAULL, "127457491.ogg"},
	{0xE674E1847C0BD688ULL, "llac_lt_warehouse_65537_3.dds"},
	{0xE675235C0176C5C5ULL, "vxlt_0_lowtown_a2_efg_0_0.vlm"},
	{0xE6754980B27F2045ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodfull_1.mao"},
	{0xE67585F4B975ADE4ULL, "gen00fl_merrill6106526.cl"},
	{0xE675B2157BDFEB04ULL, "6162717_m.fxe"},
	{0xE675E18EEE1C5D60ULL, "vxht_mansion2b_2v1_0_0.vlm"},
	{0xE6760F87019D64EFULL, "vxdr_0_deeproads_f5x_0_0.vlm"},
	{0xE6762CE182F0DFBCULL, "176932400.ogg"},
	{0xE6771D9EAC6529A3ULL, "art\\levels\\races\\design\\animation\\floortrapactivate.gad"},
	{0xE6772156D85C4E6AULL, "769472186.ogg"},
	{0xE677230B715B9DA6ULL, "vxlt_gallowscourtyard_6yk_0_0.vlm"},
	{0xE6774DC53C2A4F1DULL, "6163621_m.fxe"},
	{0xE677568D6F29F90FULL, "vxw_blightlands_pl_0_0.vlm"},
	{0xE6775842CCC140CBULL, "gen00fl_dog6117188.cl"},
	{0xE6776115328C73ACULL, "vxlt_0_lowtown_a2_n_emx_0_0.vlm"},
	{0xE6778C95E4019B11ULL, "6173174_m.fxe"},
	{0xE6779C714745DA77ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medp.mmh"},
	{0xE677A0714745E084ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medp.mao"},
	{0xE677B2714745FFE9ULL, "art\\characters\\playercharacter\\humanfemale\\glv_medp.msh"},
	{0xE677C841B5B7BA15ULL, "600436693.ogg"},
	{0xE677DFE3765CD91EULL, "6156363_m.fxe"},
	{0xE677F54567215810ULL, "gencr_darkspawn_emissary.utc"},
	{0xE677F5A115461847ULL, "vxht_mansion2_2k5_0_0.vlm"},
	{0xE67812A1B5462481ULL, "art\\characters\\creatures\\mubari\\animation\\com_v1.gad"},
	{0xE67851E534DB45C0ULL, "23673461.ogg"},
	{0xE679913702821C4EULL, "540711354.ogg"},
	{0xE679965E7FB83B48ULL, "vxht_mansion_n_61g_0_0.vlm"},
	{0xE679CA79BFCD59B0ULL, "6001353_m.fxe"},
	{0xE679EB0C887FC8BEULL, "vxht_mansion2b_25v_0_0.vlm"},
	{0xE67A36FAF1D7594BULL, "hla_ht_0_hightown_a2_65538_0x.dds"},
	{0xE67A7CF9468B8AC5ULL, "w_bonepitmining_a2.rml"},
	{0xE67A917FFE3D44EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_damage_bonus_stagger.ani"},
	{0xE67AE4CCA9F1F19DULL, "crack_g05.mao"},
	{0xE67B1C242DED0636ULL, "507019055.ogg"},
	{0xE67BB1EA4C1155CDULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.dura.ani"},
	{0xE67BBD6BFA5312A1ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.cess.ani"},
	{0xE67C581846E2EDCAULL, "473204279.ogg"},
	{0xE67CBB0219FD1FDEULL, "dae021_sister_alema_mob6144850.cl"},
	{0xE67CFE555D83D597ULL, "254072582.ogg"},
	{0xE67D00749907CB51ULL, "6048575_m.fxe"},
	{0xE67D0CF98B5768EFULL, "art\\vfx\\cinematic\\vfx_lyrium_cor_c.impa.evt"},
	{0xE67DABE9CDFF6A76ULL, "334623001.ogg"},
	{0xE67DFA9EAC6B0C88ULL, "art\\levels\\races\\design\\animation\\floortrapactivate.evt"},
	{0xE67E0190BF2342A9ULL, "llac_lt_0_lowtown_a2_n_c_-262144_3.dds"},
	{0xE67E05E2F1A76574ULL, "vxw_woundedcoast_1fn_0_0.vlm"},
	{0xE67E4BADD3A44CDDULL, "hlac_lt_0_lowtown_n_-262143_0x.dds"},
	{0xE67F13A1B54C452FULL, "art\\characters\\creatures\\mubari\\animation\\com_v1.ani"},
	{0xE67F494D6AE3BE19ULL, "486214457.ogg"},
	{0xE67F92FBC09B0B82ULL, "vxlt_0_lowtown_a3_n_c_gx8_0_0.vlm"},
	{0xE67F9CB4988F0786ULL, "438666492.ogg"},
	{0xE6800A1E7F9E6AB9ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_lt_ash_sparks.mmh"},
	{0xE6802416D2E480C4ULL, "w_sundermount_182_0.mmh"},
	{0xE6802816D2E48717ULL, "w_sundermount_182_0.mao"},
	{0xE6803A16D2E4A666ULL, "w_sundermount_182_0.msh"},
	{0xE68040A9CEC030FFULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo12_basic.phy"},
	{0xE680621CF248ED22ULL, "art\\characters\\playercharacter\\humanmale\\boo_dksa.msh"},
	{0xE680641CF248F0D8ULL, "art\\characters\\playercharacter\\humanmale\\boo_dksa.mmh"},
	{0xE680CAFEA4608005ULL, "vxdr_0_deeproads_fk0_0_0.vlm"},
	{0xE681938F24B4454FULL, "vxdr_primevaltomb_2w4_0_0.vlm"},
	{0xE6819575E705FBE7ULL, "dae000ip_crowd_lm4_talk16.utp"},
	{0xE6819D7CEA6DBD4CULL, "hlac_lt_0_lowtown_docks_n_-196611_0z.dds"},
	{0xE6820378AFBCD82EULL, "llac_lt_0_lowtown_a2_n_131070_2.dds"},
	{0xE682931C310E1CA8ULL, "831582939.ogg"},
	{0xE682C86BFA594443ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.cess.gad"},
	{0xE6830DF98B5BD598ULL, "art\\vfx\\cinematic\\vfx_lyrium_cor_c.impa.gad"},
	{0xE683186A6E4D3CBEULL, "hlac_lt_0_lowtown_65535_7x.dds"},
	{0xE6838C027EE167EFULL, "6133626_m.fxe"},
	{0xE68406BACBD69E29ULL, "vxht_0_hightown_a2_b7c_0_0.vlm"},
	{0xE6842A3D8BE81530ULL, "gen00fl_isabela_init6110270.cl"},
	{0xE6850DEDDADE61BBULL, "vxdr_primevaltomb_4zu_0_0.vlm"},
	{0xE6852E1C5E0B5481ULL, "vxca_cavern1_haunted_fn9_0_0.vlm"},
	{0xE685A43DE207820BULL, "vxca_cavern1_haunted_ck5_0_0.vlm"},
	{0xE685C86A6120118FULL, "vxw_blightlands_uo_0_0.vlm"},
	{0xE685D5AEAD6E350BULL, "403400974.ogg"},
	{0xE686A6975B5B4AF2ULL, "vxlt_darktown_47m_0_0.vlm"},
	{0xE686DB74FC885799ULL, "6132788_m.fxe"},
	{0xE686DF069AFA80ABULL, "695692781.ogg"},
	{0xE68714A98E18B497ULL, "vxca_caverns2_8ol_0_0.vlm"},
	{0xE6873D8339A1F975ULL, "41834012.ogg"},
	{0xE68747DE3A51B7C3ULL, "vxca_caverns2_6h6_0_1.vlm"},
	{0xE68761931A5A68DDULL, "zz_otl_debug6060960.cl"},
	{0xE68767DD3C7A34B1ULL, "vrc221_hugin6065248.cl"},
	{0xE68781F1DE84D241ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_staircap.phy"},
	{0xE687DC50880A4C57ULL, "lt_0_lowtown_docks_a3_118_0.mmh"},
	{0xE687E050880A5364ULL, "lt_0_lowtown_docks_a3_118_0.mao"},
	{0xE687EE8B75150764ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxb.msh"},
	{0xE687F250880A71C9ULL, "lt_0_lowtown_docks_a3_118_0.msh"},
	{0xE688008B751525CDULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxb.mao"},
	{0xE688048B75152C06ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxb.mmh"},
	{0xE688450F0B00F33DULL, "653698853.ogg"},
	{0xE68863819163D3F2ULL, "gen00fl_varric6081588.cl"},
	{0xE68876A21B88930AULL, "vxht_brothel_n_5ae_0_0.vlm"},
	{0xE688B87D451F43C8ULL, "vxca_caverns1_k2r_0_2.vlm"},
	{0xE6891E6000FADBE6ULL, "dre000_stroud6036736.cl"},
	{0xE6898DA3B9759AF9ULL, "vxw_blightlands_g3_0_0.vlm"},
	{0xE689A56BFA5F2728ULL, "art\\vfx\\combat\\vfx_str_spit1_aoe_p.cess.evt"},
	{0xE689B28002E37F67ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffidle.gad"},
	{0xE68A4E7697BC38EAULL, "vxdr_slavers_14s_0_0.vlm"},
	{0xE68AC292DD74AFD8ULL, "vxht_0_hightown_a2_a84_0_0.vlm"},
	{0xE68ADF512DB487BBULL, "hlac_lt_0_lowtown_a2_n_c_-65535_2z.dds"},
	{0xE68B45D8DE077D9AULL, "6049675_f.fxe"},
	{0xE68B6B1CF2530463ULL, "art\\characters\\playercharacter\\humanmale\\boo_dksa.phy"},
	{0xE68B8AB3328F2B77ULL, "dae011_badpoet_sta6145181.cl"},
	{0xE68B9B7C3CEACEE0ULL, "vxht_bartrand_n_63f_0_0.vlm"},
	{0xE68B9C2AA31F0087ULL, "dre000_dougal6131167.cl"},
	{0xE68BDF7E338D4178ULL, "vxlt_0_lowtown_a2_n_fgi_0_0.vlm"},
	{0xE68C77BFE39D7DF4ULL, "vxw_woundedcoast_1ao_0_0.vlm"},
	{0xE68C7F7404F7D3C8ULL, "6117927_f.fxe"},
	{0xE68DB30318FF86A4ULL, "vxlt_gallowstemplar_n_4sl_0_0.vlm"},
	{0xE68DED7D60F6C708ULL, "hightown_male_cityguard_mob6170035.cl"},
	{0xE68EB2E308E94291ULL, "dae300_alistair_king6133586.cl"},
	{0xE68ED684E98C8F5EULL, "hlac_lt_0_lowtown_a2_n_c_-131071_2x.dds"},
	{0xE68EF0A873F19EA8ULL, "6080224_m.fxe"},
	{0xE68FA78002E7D874ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffidle.evt"},
	{0xE68FCC31AB308BC0ULL, "lt_0_lowtown_docks_n_103_0.phy"},
	{0xE69083E9293DD09BULL, "vxlt_0_lowtown_a3_n_c_hln_0_0.vlm"},
	{0xE6908C19D5DFE9D8ULL, "6177319_m.fxe"},
	{0xE690E3084A0BAF0EULL, "lt_0_lowtown_docks_a3_n_127_0.mmh"},
	{0xE690E7DDFFBDFC04ULL, "tbe100im_enchanted_shield.uti"},
	{0xE690EF084A0BC3A5ULL, "lt_0_lowtown_docks_a3_n_127_0.mao"},
	{0xE690FD084A0BDBFCULL, "lt_0_lowtown_docks_a3_n_127_0.msh"},
	{0xE6911DCED4D14D1BULL, "hlac_lt_0_lowtown_docks_n_131071_7x.dds"},
	{0xE69141A9CECEA504ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo12_basic.mmh"},
	{0xE69145A9CECEAC57ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo12_basic.mao"},
	{0xE69157A9CECECAA6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo12_basic.msh"},
	{0xE691957A9BB0E914ULL, "vxlt_gallowstemplar_f_5oi_0_0.vlm"},
	{0xE691A282229FBF04ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_0.pwk"},
	{0xE691D9C5873180DAULL, "383325015.ogg"},
	{0xE691FB8B751D71D1ULL, "art\\levels\\races\\ferelden\\props\\lowtown\\prp_portetmxb.phy"},
	{0xE691FDA1FA980644ULL, "hlac_lt_0_lowtown_a3_n_c_0_3x.dds"},
	{0xE6921C32D8EAF674ULL, "6110513_f.fxe"},
	{0xE6927897BDCCB6A9ULL, "vxht_0_hightown_a2_b8b_0_0.vlm"},
	{0xE692BCFA1C855867ULL, "6196685_m.fxe"},
	{0xE693261CCDC9D274ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cupb_d.dds"},
	{0xE693710B3BA2589EULL, "vxlt_backalley_n_1jp_0_0.vlm"},
	{0xE69449C6BB5E068BULL, "vxht_keep_n_70b_0_0.vlm"},
	{0xE69467EA7B90D9C5ULL, "vxht_keep_n_6bo_0_0.vlm"},
	{0xE694D95AF9324947ULL, "vxht_mansion2_2q7_0_0.vlm"},
	{0xE695310362F82DA0ULL, "hlac_lt_0_lowtown_docks_a3_-131076_0x.dds"},
	{0xE695739903856F64ULL, "gen00fl_varric6033561.cl"},
	{0xE6957D30D3B30097ULL, "vxht_0_hightown_bek_0_0.vlm"},
	{0xE6958A0910F0E116ULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.mmh"},
	{0xE695FD5EE210E0B1ULL, "llac_w_encounter2_65541_0.dds"},
	{0xE696516A30C0C0B5ULL, "670457402.ogg"},
	{0xE696CD80A9DE43FCULL, "w_bonepit_9_0_ll.mmh"},
	{0xE696E6050657D017ULL, "1063487064.ogg"},
	{0xE6973FE6DFC221AFULL, "6153916_m.fxe"},
	{0xE697FF80AB6BB379ULL, "vxw_blightlands_h2_0_0.vlm"},
	{0xE698317E78D15360ULL, "dae100_gamlen_letter6146992.cl"},
	{0xE6983C46B3DE1BBEULL, "vxca_cavern1_haunted_gki_0_1.vlm"},
	{0xE69853CC953F5765ULL, "follower_banter6136329.cl"},
	{0xE6989BC8E1CE9370ULL, "hlac_lt_0_lowtown_docks_65534_3z.dds"},
	{0xE69903F1F51E2849ULL, "con_ico_humorous.dds"},
	{0xE69961C6E2972C58ULL, "hlac_ca_caverns1_24425_10y.dds"},
	{0xE699DE9E197CBDFFULL, "401415390.ogg"},
	{0xE69AE0B9CEA13491ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_p.impa.ani"},
	{0xE69AE99CD5939674ULL, "vxw_woundedcoast_1ll_0_0.vlm"},
	{0xE69AFE91F4CA8B83ULL, "698241995.ogg"},
	{0xE69B24C2FB4A20C9ULL, "hlac_ca_caverns1_mines_24427_1y.dds"},
	{0xE69B3DD368FB9D0EULL, "dae000_orsino6085477.cl"},
	{0xE69B5D4B1EA19141ULL, "vxlt_gallowstemplar_f_6sk_0_0.vlm"},
	{0xE69B6BDF1FB02253ULL, "674894706.ogg"},
	{0xE69B6D709AE975F4ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax104a_0t.dds"},
	{0xE69B7037F2253793ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_enter.ani"},
	{0xE69C8343C6C119F5ULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg.phy"},
	{0xE69CB3C594C851EFULL, "6064557_m.fxe"},
	{0xE69CD119B6C55B65ULL, "209135373.ogg"},
	{0xE69CEF9AB765CBB5ULL, "dae000_gamlen6121561.cl"},
	{0xE69CF5058CDB18A9ULL, "vxht_0_hightown_n_ap2_0_0.vlm"},
	{0xE69D5E3A295C1BF1ULL, "mag121_idunna6084262.cl"},
	{0xE69D9F8002F407ADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffidle.ani"},
	{0xE69DFD9F5900F05BULL, "gen_im_acc_rng_act1_02.uti"},
	{0xE69EC49F69994504ULL, "vxw_bonepit_28h_0_0.vlm"},
	{0xE69F99B8EFB9F172ULL, "mcr370_templar26087009.cl"},
	{0xE69FC92EA8F15AF8ULL, "6167109_m.fxe"},
	{0xE6A002463742FCE4ULL, "493793117.ogg"},
	{0xE6A017ED207EE850ULL, "6174220_m.fxe"},
	{0xE6A06839955BD461ULL, "gen00fl_varric6118791.cl"},
	{0xE6A0B43E45A0F9DCULL, "vxw_encounter3_fl_0_1.vlm"},
	{0xE6A11C63958BBDA9ULL, "6149328_m.fxe"},
	{0xE6A11D7F161463B9ULL, "741467237.ogg"},
	{0xE6A127B79F378661ULL, "233874972.ogg"},
	{0xE6A14250654556FEULL, "6155511_m.fxe"},
	{0xE6A1717F1E5A5A08ULL, "vxw_bonepitdestroyed_a3_1s2_0_0.vlm"},
	{0xE6A19C07DC8BB881ULL, "clf221_jalen6070382.cl"},
	{0xE6A1EBB9CEA76633ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_p.impa.gad"},
	{0xE6A2201EA5483185ULL, "6206881_m.fxe"},
	{0xE6A23223A7A66797ULL, "86239690.ogg"},
	{0xE6A232399E019A4FULL, "vxht_playershouse_n_1s4_0_0.vlm"},
	{0xE6A24C0EB17A88B4ULL, "878272129.ogg"},
	{0xE6A267C8A3F9C5E5ULL, "789843428.ogg"},
	{0xE6A309E4668BEF48ULL, "vxht_keep_n_6m7_0_0.vlm"},
	{0xE6A3258E26EF62ACULL, "hlac_dr_slavers_3215_6y.dds"},
	{0xE6A34B37EB2861C7ULL, "vxht_mansion2_2z6_0_0.vlm"},
	{0xE6A38A513F477BF2ULL, "vxlt_darktown_49o_0_0.vlm"},
	{0xE6A3F7B24097E291ULL, "6135958_m.fxe"},
	{0xE6A40920C6346E85ULL, "gentr_party_triggers.utt"},
	{0xE6A42FFA3993D763ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillar.mao"},
	{0xE6A433FA3993DDB0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillar.mmh"},
	{0xE6A441FA3993F5CAULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillar.msh"},
	{0xE6A48CD22FB91767ULL, "114794531.ogg"},
	{0xE6A4928E7146164BULL, "914203334.ogg"},
	{0xE6A4E9867A8BA68EULL, "6017132_m.fxe"},
	{0xE6A50BAEC92E1E73ULL, "llac_w_woundedcoast_-262142_0.dds"},
	{0xE6A5567B5FEDBA35ULL, "6027813.ogg"},
	{0xE6A58FED74B83D9EULL, "hlac_ca_caverns2prc_5802_17x.dds"},
	{0xE6A5913CBF6E97D3ULL, "mrl251ip_bosshorde.ncs"},
	{0xE6A5E4490DBFB631ULL, "vxlt_0_lowtown_a2_n_c_df3_0_0.vlm"},
	{0xE6A6715D9D61CBF9ULL, "vxw_blightlands_m1_0_0.vlm"},
	{0xE6A69C682A2F5131ULL, "mag320_orsino6071112.cl"},
	{0xE6A6C347B63CD142ULL, "hm_walleyed_sam.mor"},
	{0xE6A6CFCDC3C25191ULL, "vxht_chantry_359_0_1.vlm"},
	{0xE6A6D99FAE80C459ULL, "vxlt_shop_ab_0_0.vlm"},
	{0xE6A6E42B68BF9C8FULL, "381801712.ogg"},
	{0xE6A713AAC6E6F9C4ULL, "vfx_vfx_p3.psh.fxo.dx11.assembly"},
	{0xE6A73C54E77A3602ULL, "6043365_f.fxe"},
	{0xE6A7865CFB400346ULL, "380929960.ogg"},
	{0xE6A7F082379B0EF4ULL, "mag101ar_caverns.are"},
	{0xE6A7F30B972D5BE1ULL, "vxht_keep_n_6h1_0_0.vlm"},
	{0xE6A82D263BD9E8D8ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_hlf_hvyo_0d.dds"},
	{0xE6A87F3620D6FFE0ULL, "vxht_bartrand_n_69d_0_0.vlm"},
	{0xE6A8CA6FFEEEFAA2ULL, "237252560.ogg"},
	{0xE6A8E8B9CEAD7F98ULL, "art\\vfx\\combat\\vfx_m_frc_repwave_p.impa.evt"},
	{0xE6A9988CFD0EAE44ULL, "665012543.ogg"},
	{0xE6AAA2DC69350DF8ULL, "dae200_cullen6140274.cl"},
	{0xE6AAA7BBB8CE776AULL, "640897219.ogg"},
	{0xE6AAC951BE813E37ULL, "lgt110_orz_harr_carta6109048.cl"},
	{0xE6AAE9E1C7D7DB92ULL, "lt_0_lowtown_docks_23_0.rml"},
	{0xE6AAEA728DB3C428ULL, "vxlt_hangedman_n_26x_0_0.vlm"},
	{0xE6AAF1073FEF061AULL, "638386781.ogg"},
	{0xE6AB337FC23F485EULL, "mag131_taraneh6094533.cl"},
	{0xE6AB3501C27ECDD6ULL, "follower_banter6148732.cl"},
	{0xE6AB358B040CD4F5ULL, "vxlt_ruinedhouse_n_if_0_0.vlm"},
	{0xE6AB362FC17439DBULL, "209145246.ogg"},
	{0xE6AB6F1FD6E73CF5ULL, "vxht_0_hightown_a3_9zj_0_0.vlm"},
	{0xE6AC4D3B47F4A65CULL, "plt_mrl000pt_gift_ring.plo"},
	{0xE6AC6721DBE4DF85ULL, "6177888_m.fxe"},
	{0xE6AC8D733538D646ULL, "402556985.ogg"},
	{0xE6ACAAD48ECE8653ULL, "hlac_lt_warehouse_1_3x.dds"},
	{0xE6AD539745322BBBULL, "qun211_braggart.bnk"},
	{0xE6AD6FF946B6CE47ULL, "w_bonepitmining_a2.arl"},
	{0xE6AD9A2188C56BE3ULL, "vxlt_gallowstemplar_n_3rs_0_0.vlm"},
	{0xE6ADCA036F851A20ULL, "663555756.ogg"},
	{0xE6ADE02E4FCDF956ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.anb"},
	{0xE6ADFB0E88672BDCULL, "hlac_lt_0_lowtown_65534_4x.dds"},
	{0xE6AEB30131FA111BULL, "6146936_m.fxe"},
	{0xE6AEECDAE8FF14D2ULL, "6119616_m.fxe"},
	{0xE6AF2804B94742C6ULL, "l_env_wal_lowtown_day_dist_oneshots.bnk"},
	{0xE6AF656639F77BCEULL, "hlac_w_encounter2_65537_0z.dds"},
	{0xE6AF6750DEE5C8EBULL, "vxlt_undercitydungeon_1d3_0_0.vlm"},
	{0xE6AF7A4A9DD0C283ULL, "zz_fen_debug6134339.cl"},
	{0xE6B01BF3ADE8E0D3ULL, "6034549_m.fxe"},
	{0xE6B04737F2374E3DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_arm_cross_enter.gad"},
	{0xE6B052F49DD09945ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_curtains_n.dds"},
	{0xE6B054C51F8E899EULL, "vxlt_backalley_n_1pr_0_0.vlm"},
	{0xE6B055974534657BULL, "qun211_braggart.cnv"},
	{0xE6B0A4168FF7B2CFULL, "vxht_playershouse_n_1v7_0_0.vlm"},
	{0xE6B0F95BDED32E31ULL, "vxw_sundermount_23w_0_0.vlm"},
	{0xE6B17BC1588207C8ULL, "vxht_keep_n_6h4_0_0.vlm"},
	{0xE6B1C34E4FB3C4F4ULL, "237855091.ogg"},
	{0xE6B1CEBDF45884F2ULL, "zz_approval_debug6003146.cl"},
	{0xE6B1D1FB5780A410ULL, "dae000_mother6135742.cl"},
	{0xE6B28233AA758B76ULL, "vxlt_0_lowtown_exj_0_0.vlm"},
	{0xE6B2912CD74F8E42ULL, "hlac_ca_caverns1_mines_24421_5x.dds"},
	{0xE6B2A91D50B46D1BULL, "369990742.ogg"},
	{0xE6B2A9FF9A05F5A9ULL, "141115246.ogg"},
	{0xE6B2E22705C8B028ULL, "134683793.ogg"},
	{0xE6B3D3A1B333E6C8ULL, "206941901.ogg"},
	{0xE6B4E33A8F57E479ULL, "vxw_blightlands_v0_0_0.vlm"},
	{0xE6B4E931AB4FD955ULL, "lt_0_lowtown_docks_n_103_0.msh"},
	{0xE6B4EB31AB4FDCCBULL, "lt_0_lowtown_docks_n_103_0.mmh"},
	{0xE6B4F731AB4FF168ULL, "lt_0_lowtown_docks_n_103_0.mao"},
	{0xE6B5055298E0ECB4ULL, "ch1_ds2y.psh.fxo"},
	{0xE6B54B7CA076DCD9ULL, "vxlt_shop_da_0_0.vlm"},
	{0xE6B5BC0D81142247ULL, "6116133_m.fxe"},
	{0xE6B5D06905F16EC2ULL, "6129897_m.fxe"},
	{0xE6B5E491264E9DB6ULL, "hlac_ht_chantry_n_2534_22x.dds"},
	{0xE6B5E795972410EDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_spitoon.phy"},
	{0xE6B5E7991A7A2A03ULL, "hlac_ca_caverns1_24424_19x.dds"},
	{0xE6B61599592C08DFULL, "vxht_0_hightown_b9g_0_0.vlm"},
	{0xE6B664E889237461ULL, "vxht_keep_n_6c0_0_0.vlm"},
	{0xE6B6C8932004D733ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_win.ani"},
	{0xE6B6F921C46BFC6DULL, "6044513_m.fxe"},
	{0xE6B6FC37367162E4ULL, "vxlt_gallowstemplar_51x_0_0.vlm"},
	{0xE6B70E33B1086E4DULL, "6135785_m.fxe"},
	{0xE6B78417F950B5A7ULL, "hlac_w_encounter3_-262140_0x.dds"},
	{0xE6B79F289351D736ULL, "6085391_m.fxe"},
	{0xE6B7CD56B97FC774ULL, "vxw_woundedcoast_1rr_0_0.vlm"},
	{0xE6B7F1C0A73FE043ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeceltic.msh"},
	{0xE6B7FFC0A73FF79DULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeceltic.mmh"},
	{0xE6B803C0A73FFEAEULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzlargeceltic.mao"},
	{0xE6B9DBC56BC14442ULL, "186730425.ogg"},
	{0xE6B9E02F46F07300ULL, "vxht_chantry_n_2as_0_0.vlm"},
	{0xE6BA1AEE74473A5BULL, "59235661.ogg"},
	{0xE6BA8C29F80A5B0EULL, "6128933_m.fxe"},
	{0xE6BAE32E4FD888C1ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_mo_c.mmh"},
	{0xE6BB85E24766E61FULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s1_c.mmh"},
	{0xE6BC2FA0CBF11FEBULL, "234873877.ogg"},
	{0xE6BC789219E33EB3ULL, "vxlt_0_lowtown_a3_n_c_byo_0_0.vlm"},
	{0xE6BDB36485B67C91ULL, "vxw_woundedcoast_7u_0_0.vlm"},
	{0xE6BDD73A8739AF97ULL, "gen_im_npc_arm_hum_boo_cota.uti"},
	{0xE6BE12FC9560F663ULL, "391856764.ogg"},
	{0xE6BEE0FA7D632E59ULL, "gen00fl_anders6119310.cl"},
	{0xE6BF2C86702DF75BULL, "vxlt_hangedman_n_56_0_0.vlm"},
	{0xE6BF6B38D0C946B1ULL, "vxw_sundermount_24v_0_0.vlm"},
	{0xE6BF6CC22115F4F7ULL, "zz_followers_banter6157364.cl"},
	{0xE6BF6DBFADBAA1E6ULL, "vxlt_0_lowtown_docks_c56_0_0.vlm"},
	{0xE6BFCE593BFB666CULL, "vxlt_warehouse_n_d_0_0.vlm"},
	{0xE6BFED9E4A782048ULL, "vxht_keep_n_6c5_0_0.vlm"},
	{0xE6C0CBD3DCA63746ULL, "gen00fl_anders6125819.cl"},
	{0xE6C14F3DECB7356BULL, "hlac_lt_0_lowtown_docks_a3_-196610_6z.dds"},
	{0xE6C166D6834BFDBCULL, "one110_meeran6089475.cl"},
	{0xE6C17D23012EC77AULL, "zz_ave_debug6134021.cl"},
	{0xE6C1D015B3D597C7ULL, "hlac_dr_0_deeproads_5842_5y.dds"},
	{0xE6C2A3EEF56967B1ULL, "hlac_ht_0_hightown_a2_65534_0y.dds"},
	{0xE6C2EE6B18A4A7F5ULL, "zz_dog_debug6127918.cl"},
	{0xE6C30DD8CE1EB5B1ULL, "6176515_m.fxe"},
	{0xE6C3894C02F19C41ULL, "6155538_m.fxe"},
	{0xE6C4190A48FEE553ULL, "887656197.ogg"},
	{0xE6C45411DEE05C47ULL, "6082214_m.fxe"},
	{0xE6C47BB043B37033ULL, "hlac_lt_0_lowtown_a2_-3_3y.dds"},
	{0xE6C4B1FE0CD31284ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef01.phy"},
	{0xE6C4C0932011076AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_win.evt"},
	{0xE6C513F565F50390ULL, "vxw_woundedcoast_fg_0_0.vlm"},
	{0xE6C5208B3BE0CE4EULL, "hlac_lt_darktown_-1_23y.dds"},
	{0xE6C5E5B859DB08CCULL, "6149978_m.fxe"},
	{0xE6C5EF134AB6D7A4ULL, "6060420_m.fxe"},
	{0xE6C63B2FD1BD707CULL, "6087386_m.fxe"},
	{0xE6C65D1BB898D0B0ULL, "525349945.ogg"},
	{0xE6C69D0481DD0312ULL, "370997582.ogg"},
	{0xE6C69E8E766F4B7CULL, "6140209_m.fxe"},
	{0xE6C6A0026E39BAA3ULL, "vxlt_gallowstemplar_n_6t0_0_0.vlm"},
	{0xE6C6CB5B39988B1FULL, "vxlt_0_lowtown_a2_e2y_0_0.vlm"},
	{0xE6C6D78237B594D6ULL, "mag101ar_caverns.ncs"},
	{0xE6C6DA5F8E8F97A7ULL, "llac_ht_0_hightown_n_-131069_0.dds"},
	{0xE6C70A99767F32BAULL, "vxlt_0_lowtown_n_f0b_0_0.vlm"},
	{0xE6C73A606438CE73ULL, "follower_banter6167912.cl"},
	{0xE6C776513BA44E95ULL, "vxlt_0_lowtown_a2_n_c_g60_0_0.vlm"},
	{0xE6C79ACAADF2BB49ULL, "223295547.ogg"},
	{0xE6C7AFADF4EEC9E4ULL, "6183816_f.fxe"},
	{0xE6C7BA00D5A24DEDULL, "llac_lt_darktown_-1_61.dds"},
	{0xE6C7CF32DB650389ULL, "hlac_lt_gallowscourtyard_n_-1_2x.dds"},
	{0xE6C83D8224B20806ULL, "vxht_0_hightown_a3_c9v_0_0.vlm"},
	{0xE6C875B1A654D95EULL, "gen00fl_merrill6115797.cl"},
	{0xE6C8791AAAEC0667ULL, "134442452.ogg"},
	{0xE6C89B045B906543ULL, "vxlt_gallowstemplar_46c_0_0.vlm"},
	{0xE6C8A900FC9996DEULL, "698964842.ogg"},
	{0xE6C8C404E70C1CB0ULL, "follower_banter6148033.cl"},
	{0xE6C9591A2A314D5CULL, "zz_vault_debug6201723.cl"},
	{0xE6C9D3DFB851DBF8ULL, "autobalance.gda"},
	{0xE6CAC85C4B16319CULL, "qun120cr_javaris6042986.cl"},
	{0xE6CAD7DAC9F94C6DULL, "102096553.ogg"},
	{0xE6CB4508399801FEULL, "vxht_0_hightown_a3_c1f_0_0.vlm"},
	{0xE6CB9F932016EDDDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\am_gam_dice_win.gad"},
	{0xE6CBCD192CD1A25EULL, "6140244_f.fxe"},
	{0xE6CBD01D2D8DF2B1ULL, "art\\vfx\\combat\\vfx_r_dwp_front_long_c.anb"},
	{0xE6CBD094CCEB45D2ULL, "6113016_m.fxe"},
	{0xE6CC315FFB4C2FA6ULL, "781586971.ogg"},
	{0xE6CC79794F066B6CULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_templar1.ani"},
	{0xE6CD04674D00940AULL, "236515334.ogg"},
	{0xE6CD95CD7029B42CULL, "6041264_f.fxe"},
	{0xE6CD960C0A2F7315ULL, "6017658_m.fxe"},
	{0xE6CDA1DE588FF732ULL, "20614298.ogg"},
	{0xE6CDE25C00369507ULL, "vxw_blightlands_bb_0_0.vlm"},
	{0xE6CDE491414327C7ULL, "vxlt_hangedman_5k_0_0.vlm"},
	{0xE6CDEF8237BBDC6FULL, "mag101ar_caverns.lst"},
	{0xE6CE5669AECAD8E1ULL, "follower_banter6163437.cl"},
	{0xE6CEE8B3081A003CULL, "6095121_m.fxe"},
	{0xE6CF2254FF059A46ULL, "gen00fl_aveline6106954.cl"},
	{0xE6CF7B07DF986766ULL, "zz_qa_resolved.utp"},
	{0xE6CFA6FE3906C4B8ULL, "1000644564.ogg"},
	{0xE6CFAEB725E57A7CULL, "454081996.ogg"},
	{0xE6D0486087CA4EF8ULL, "6071783_m.fxe"},
	{0xE6D06DD8B5758ECFULL, "hlac_lt_0_lowtown_-131072_2y.dds"},
	{0xE6D0F42999569299ULL, "vxlt_gallowstemplar_3sa_0_0.vlm"},
	{0xE6D13E5E51654B22ULL, "vxlt_0_lowtown_a3_n_c_df9_0_0.vlm"},
	{0xE6D212087191ED92ULL, "follower_banter6170630.cl"},
	{0xE6D286794F0AECAEULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrive_templar1.gad"},
	{0xE6D33E1B899CE8F1ULL, "vxlt_warehouse_n_15k_0_0.vlm"},
	{0xE6D340D59CA5B86AULL, "6137066_f.fxe"},
	{0xE6D385D257EB1C10ULL, "vxw_woundedcoast_cd_0_0.vlm"},
	{0xE6D3B791615497A8ULL, "gen00fl_sebastian6195837.cl"},
	{0xE6D3E2B8F2DC622EULL, "vxdr_0_deeproads_fii_0_0.vlm"},
	{0xE6D3E4400FEA5867ULL, "vxw_bonepitmining_a2_1pv_0_0.vlm"},
	{0xE6D408D4B0CC44A6ULL, "vxw_encounter3_f7_0_0.vlm"},
	{0xE6D4583831199B3BULL, "6171844_m.fxe"},
	{0xE6D4891085E23222ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen.gad"},
	{0xE6D4E3C328045F25ULL, "vxlt_hangedman_n_1pl_0_0.vlm"},
	{0xE6D51C43131DE869ULL, "vxht_keep_n_40s_0_0.vlm"},
	{0xE6D55B06F2C9A6BDULL, "vxdr_primevaltomb_768_0_0.vlm"},
	{0xE6D5942AE7622CC4ULL, "6171032_m.fxe"},
	{0xE6D5B125D983BD71ULL, "6131965_m.fxe"},
	{0xE6D6036BEF32EDC7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_left_hand.gad"},
	{0xE6D6825DC9EE4E38ULL, "vxht_0_hightown_a3_c29_0_0.vlm"},
	{0xE6D6DC30B755071FULL, "vxw_encounter3_dd_0_0.vlm"},
	{0xE6D6E548C471B2A4ULL, "63831797.ogg"},
	{0xE6D6E91F00B61368ULL, "vxlt_0_lowtown_a2_n_c_csj_0_0.vlm"},
	{0xE6D74CFE26F34E76ULL, "vxca_cave_1bh_0_0.vlm"},
	{0xE6D82364CCDCBA24ULL, "llac_lt_0_lowtown_docks_-65537_3.dds"},
	{0xE6D87BFA62026886ULL, "gen_im_cns_pot_attribute_03.uti"},
	{0xE6D8923D60FDAEBEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cd_transition.ani"},
	{0xE6D8AF3B1506AB9BULL, "6053731_m.fxe"},
	{0xE6D8F11D2D98B52AULL, "art\\vfx\\combat\\vfx_r_dwp_front_long_c.mmh"},
	{0xE6D8FC87DFCC5676ULL, "zz_mag_debug6024444.cl"},
	{0xE6D9FA0C6F2EB607ULL, "hlac_lt_house_-1_6y.dds"},
	{0xE6DA7E8D160761F3ULL, "850933399.ogg"},
	{0xE6DB05DE2E500FF6ULL, "6099321_m.fxe"},
	{0xE6DB7A2D0FD265EEULL, "6076457_f.fxe"},
	{0xE6DBAC1085E88C19ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen.evt"},
	{0xE6DBAE4F0D86ACE7ULL, "mag131_keran6086241.cl"},
	{0xE6DC7DCF43E5B98DULL, "6055573_m.fxe"},
	{0xE6DCF86BEF38F9D4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_left_hand.evt"},
	{0xE6DD2FE6705E010DULL, "vxlt_undercityserialkiller_10h_0_0.vlm"},
	{0xE6DD5205580F280DULL, "follower_banter6173348.cl"},
	{0xE6DD53F5CEE42407ULL, "6090882_m.fxe"},
	{0xE6DD570595A07AAAULL, "6218971_m.fxe"},
	{0xE6DD8B1EAA25696EULL, "vxht_0_hightown_a2_c1k_0_0.vlm"},
	{0xE6DDC502F9BCAACBULL, "6148479_m.fxe"},
	{0xE6DE42A7E2520D6BULL, "vxca_caverns2prc_a5n_0_0.vlm"},
	{0xE6DE7E9C264261C4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basicnb.msh"},
	{0xE6DE7F756B1082EFULL, "dae000_gen_turn_in_16178396.cl"},
	{0xE6DE909C2642802DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basicnb.mao"},
	{0xE6DE949C26428766ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basicnb.mmh"},
	{0xE6DEEA29133160AFULL, "434810173.ogg"},
	{0xE6DF2A445F6A3FEDULL, "dae000_merchant_armor_hm6164272.cl"},
	{0xE6DFBED11D97C8D2ULL, "fen311cr_patron_2.utc"},
	{0xE6DFD812918DA091ULL, "6109786_m.fxe"},
	{0xE6DFFCF274674732ULL, "vxlt_0_lowtown_n_eqg_0_1.vlm"},
	{0xE6E029E7F7B66092ULL, "498896370.ogg"},
	{0xE6E02F1FE3A69734ULL, "vxlt_0_lowtown_a2_n_c_gjq_0_0.vlm"},
	{0xE6E052DECFB1297BULL, "gen00fl_dog6148175.cl"},
	{0xE6E053381B29FFBAULL, "gen00fl_fenris6105613.cl"},
	{0xE6E0860197EFDD9EULL, "6153232_m.fxe"},
	{0xE6E15AB612F6E004ULL, "zz_dae_debug6050750.cl"},
	{0xE6E1F7AF49E13490ULL, "vxw_woundedcoast_xe_0_0.vlm"},
	{0xE6E2146999CA248DULL, "lt_gallowscourtyard_50_0_lc.mmh"},
	{0xE6E27AB1A2C25D26ULL, "vxw_encounter3_c8_0_0.vlm"},
	{0xE6E31C2EBC89853AULL, "578147457.ogg"},
	{0xE6E355A019FA77A5ULL, "vxlt_hangedman_n_1mk_0_0.vlm"},
	{0xE6E3748A312F5694ULL, "vxht_bartrand_n_3nl_0_0.vlm"},
	{0xE6E37C9D200BA989ULL, "vxlt_hangedman_14o_0_0.vlm"},
	{0xE6E39643C6FD3C10ULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg.msh"},
	{0xE6E3A443C6FD542AULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg.mmh"},
	{0xE6E3A843C6FD5B61ULL, "art\\characters\\creatures\\gallowsgateguard\\cn_bdy_ggg.mao"},
	{0xE6E43AFA39C9CDCBULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\1ww_pillar.phy"},
	{0xE6E43C7718682141ULL, "6125487_m.fxe"},
	{0xE6E463831F41B04CULL, "lt_warehouse_n_2_0_lc.mmh"},
	{0xE6E486E24789ED5CULL, "art\\vfx\\combat\\vfx_glb_com_brittle_p_s1_c.anb"},
	{0xE6E4A08041097665ULL, "vxca_caverns1_dpt_0_0.vlm"},
	{0xE6E4C9A40B503C2DULL, "6083682_m.fxe"},
	{0xE6E4F43ABBE466B8ULL, "vxht_0_hightown_a3_c76_0_0.vlm"},
	{0xE6E5213C089E3906ULL, "vxht_0_hightown_a3_c7x_0_0.vlm"},
	{0xE6E52352A1103D71ULL, "qun120_arishok6050274.cl"},
	{0xE6E54E0DA94B1F9FULL, "vxw_encounter3_ag_0_0.vlm"},
	{0xE6E5AA3D6108618FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cd_transition.evt"},
	{0xE6E5DE6BDAF54C30ULL, "154596756.ogg"},
	{0xE6E5FF44E27E9137ULL, "581993152.ogg"},
	{0xE6E60BD8A205B622ULL, "280081757.ogg"},
	{0xE6E61BA7E3AC803AULL, "6068487_f.fxe"},
	{0xE6E6365CBC54B862ULL, "vxdr_primevaltomb_4h2_0_0.vlm"},
	{0xE6E6439E7DAFA323ULL, "and121_karl6093890.cl"},
	{0xE6E6763A0B99D768ULL, "6117762_m.fxe"},
	{0xE6E6F0F535066C73ULL, "6051557_m.fxe"},
	{0xE6E6F6017AAD2C39ULL, "vxlt_backalley_n_1y6_0_0.vlm"},
	{0xE6E72C51FB4F4EBEULL, "411692184.ogg"},
	{0xE6E73B9745637B28ULL, "qun211_braggart.stg"},
	{0xE6E7ADC77904C13AULL, "258425419.ogg"},
	{0xE6E7CD1E8CFAFE9DULL, "hlac_w_encounter1_2_0x.dds"},
	{0xE6E8B41085F32408ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\animation\\dsclosedtoopen.ani"},
	{0xE6E8DCF3845CEFE9ULL, "w_bonepit_107_0.phy"},
	{0xE6E9026AF1D1A146ULL, "6053342_f.fxe"},
	{0xE6E915E46A216EE1ULL, "6125273_f.fxe"},
	{0xE6E91926489F19B6ULL, "558318985.ogg"},
	{0xE6E9550684C11A84ULL, "6118522_f.fxe"},
	{0xE6E986F1DED7A57DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_staircap.mao"},
	{0xE6E9890D2550269DULL, "422758333.ogg"},
	{0xE6E98AF1DED7ABB6ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_staircap.mmh"},
	{0xE6E9902F65E993EBULL, "6174308_m.fxe"},
	{0xE6E994F1DED7BCB4ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_staircap.msh"},
	{0xE6E9B1217C11D1A0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galloalcovlo.phy"},
	{0xE6E9D7DB85C00942ULL, "hlac_ca_caverns1_24426_14x.dds"},
	{0xE6E9F06BEF43760DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hands_clasped_gesture_left_hand.ani"},
	{0xE6EA8590115AE68BULL, "372068720.ogg"},
	{0xE6EAED929EC11EDFULL, "vxdr_primevaltomb_6uv_0_0.vlm"},
	{0xE6EB32A36E90782BULL, "6152689_m.fxe"},
	{0xE6EBC5D2C0DA3C87ULL, "6095782_m.fxe"},
	{0xE6EC098451F0F3D3ULL, "6016765_m.fxe"},
	{0xE6EC2F8A4A1F938BULL, "vxlt_0_lowtown_a2_n_c_e8f_0_0.vlm"},
	{0xE6EC56D94EE30E95ULL, "6096016_f.fxe"},
	{0xE6ECAB3D610E8238ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_cd_transition.gad"},
	{0xE6ED0F164CE1BAAFULL, "lt_blackemporium_1055_lc.mmh"},
	{0xE6ED1E01FE9FFB0BULL, "6125819_f.fxe"},
	{0xE6ED248D5164A423ULL, "vxht_0_hightown_a3_bpi_0_0.vlm"},
	{0xE6EDF14454386C70ULL, "hlac_ht_0_hightown_a2_131073_0z.dds"},
	{0xE6EE60B14061D4C4ULL, "gen00fl_fenris6106672.cl"},
	{0xE6EEB095CE753745ULL, "follower_banter6163565.cl"},
	{0xE6EEFE65589CA18CULL, "vxlt_hangedman_1ya_0_0.vlm"},
	{0xE6EF9102A388E417ULL, "llac_w_encounter2_196609_0.dds"},
	{0xE6EF98D7734BA623ULL, "vxlt_0_lowtown_a2_n_hda_0_0.vlm"},
	{0xE6F02A04B97EDD6CULL, "l_env_wal_lowtown_day_dist_oneshots.wbd"},
	{0xE6F0DA96FB1AB39AULL, "llac_lt_0_lowtown_docks_a3_-131075_3.dds"},
	{0xE6F133D9FA73170EULL, "mag360_alain6044642.cl"},
	{0xE6F14F7B29B44B7AULL, "dae000_flemeth6001003.cl"},
	{0xE6F1562E5F43BE6FULL, "ave331_jeven6137297.cl"},
	{0xE6F1D8AB92933E9FULL, "vxht_mansion2_35n_0_0.vlm"},
	{0xE6F1E326A0AF1643ULL, "85993362.ogg"},
	{0xE6F21CC11BB0FFA4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\sidetable_n.dds"},
	{0xE6F256CEBC0BDC56ULL, "6150268_m.fxe"},
	{0xE6F33B80FD4654ADULL, "6086982_m.fxe"},
	{0xE6F3B14880E0B2A2ULL, "346997480.ogg"},
	{0xE6F3FA959BA189E9ULL, "clf260_harley6093822.cl"},
	{0xE6F43E2CD15AE3E9ULL, "vxw_woundedcoast_oe_0_0.vlm"},
	{0xE6F4902A45946A94ULL, "vxlt_gallowstemplar_n_5xd_0_0.vlm"},
	{0xE6F4D8EB11E36032ULL, "water0.psh.fxo"},
	{0xE6F558ABCE51F1CAULL, "hlac_ca_caverns1_24422_1z.dds"},
	{0xE6F55D6847D10D8CULL, "vxlt_gallowsprison_n_5se_0_0.vlm"},
	{0xE6F587CA89D8B7F2ULL, "vxlt_warehouse_n_cm_0_0.vlm"},
	{0xE6F58CB25DFF48B1ULL, "hlac_lt_0_lowtown_a2_n_65535_7x.dds"},
	{0xE6F60A54F30ED383ULL, "art\\characters\\playercharacter\\humanmale\\rob_anda.mao"},
	{0xE6F60E54F30EDAD0ULL, "art\\characters\\playercharacter\\humanmale\\rob_anda.mmh"},
	{0xE6F61C54F30EF2EAULL, "art\\characters\\playercharacter\\humanmale\\rob_anda.msh"},
	{0xE6F6A77753380D25ULL, "vxlt_0_lowtown_eu6_0_0.vlm"},
	{0xE6F70431598064D1ULL, "hlac_lt_0_lowtown_docks_n_-131074_3x.dds"},
	{0xE6F78F2516752DBDULL, "dae000ip_crowd_ff2_talk8.utp"},
	{0xE6F7FBF253232C18ULL, "297281421.ogg"},
	{0xE6F80EA10F805D5AULL, "llac_lt_gallowscourtyard_n_65534_0.dds"},
	{0xE6F833F6DEE725DAULL, "723993285.ogg"},
	{0xE6F91220C67CBE48ULL, "gentr_party_triggers.ncs"},
	{0xE6F91E020B2BD9C9ULL, "634209450.ogg"},
	{0xE6F9383177A3821DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2_tool.phy"},
	{0xE6F95F6F90B7375FULL, "vxdr_primevaltomb_6xq_0_0.vlm"},
	{0xE6F976489232170FULL, "405993100.ogg"},
	{0xE6F9DD0A9098B46DULL, "dae000_gamlen6048466.cl"},
	{0xE6FA690ADDC08ABEULL, "art\\vfx\\combat\\vfx_c_mer_attk2_c.mmh"},
	{0xE6FACC253EF9B139ULL, "976476819.ogg"},
	{0xE6FAE59443303C30ULL, "vxht_playershouse_n_1q1_0_0.vlm"},
	{0xE6FC5522C9B99433ULL, "em_pol.mor"},
	{0xE6FCE0AEAF5CFBB0ULL, "vxlt_warehouse_1bh_0_0.vlm"},
	{0xE6FCF4A1DEB567B6ULL, "784769139.ogg"},
	{0xE6FD33DA7255CF28ULL, "vxlt_ruinedhouse_n_k9_0_0.vlm"},
	{0xE6FD70424A92BA0CULL, "vxlt_hangedman_1df_0_0.vlm"},
	{0xE6FDA09F3E48602DULL, "735625489.ogg"},
	{0xE6FDC14F39623EFDULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisb.mao"},
	{0xE6FDC54F39624536ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisb.mmh"},
	{0xE6FDCF4F39625634ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_buttrisb.msh"},
	{0xE6FE3C650489C73DULL, "484070910.ogg"},
	{0xE6FE878251950343ULL, "vxlt_undercityencounter_z2_0_1.vlm"},
	{0xE6FE8DAC1A0345BFULL, "6123230_m.fxe"},
	{0xE6FEDB692DCD6590ULL, "vxw_woundedcoast_rc_0_0.vlm"},
	{0xE6FEDC839F10C5ABULL, "hlac_w_bonepit_0_12z.dds"},
	{0xE6FEF6E50C308DBEULL, "mer221_sabin6047367.cl"},
	{0xE6FF2355D99096AAULL, "622725233.ogg"},
	{0xE6FFD02FD88AC88AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_maso_0d.dds"},
	{0xE7003959FDE6A8A5ULL, "vxlt_hangedman_n_1gi_0_0.vlm"},
	{0xE7003FA7450AEB63ULL, "zz_approval_debug6003761.cl"},
	{0xE7012CE8455490F9ULL, "6034819_m.fxe"},
	{0xE701C7268CB5692FULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath4.gad"},
	{0xE701C7848F98314FULL, "qcr000_arishok_keep.stg"},
	{0xE7023BFCAA1EAE2CULL, "lgt160_wk_avern_deaddrop6180472.cl"},
	{0xE7026831172FAD72ULL, "6140786_m.fxe"},
	{0xE702B009C350FC69ULL, "vxw_woundedcoast_dd_0_0.vlm"},
	{0xE702FA97E83D659DULL, "350153420.ogg"},
	{0xE7031F1C7C43E489ULL, "vxht_mansion_n_53s_0_0.vlm"},
	{0xE7032AE1F2E38DD7ULL, "210424088.ogg"},
	{0xE7035A9DA2A32414ULL, "m_dae221ar_chantry.bnk"},
	{0xE7038421ED2FE17EULL, "zz_ave_debug6064632.cl"},
	{0xE703CDE82CB86B62ULL, "mag131tr_atk_rand1.utt"},
	{0xE703E5B69DB9DFBBULL, "182808167.ogg"},
	{0xE70451159ECDB8B7ULL, "gen00fl_carver6103012.cl"},
	{0xE70498487B830E1CULL, "follower_banter6163225.cl"},
	{0xE7051954452E25A5ULL, "vxlt_0_lowtown_ep5_0_0.vlm"},
	{0xE7053448876E067BULL, "6139726_f.fxe"},
	{0xE705BE720544EA34ULL, "mcr310cr_circle_mage.utc"},
	{0xE705CC33D2C870ACULL, "6123332_m.fxe"},
	{0xE70657BA04FD45EEULL, "6114862_m.fxe"},
	{0xE70677574D1A29C5ULL, "6124729_m.fxe"},
	{0xE7073AC836B96D13ULL, "mag360_thrask6043529.cl"},
	{0xE7074BAC6E704A93ULL, "llac_lt_0_lowtown_docks_a3_-131074_7.dds"},
	{0xE7077A06FC12170FULL, "6175844_m.fxe"},
	{0xE70788945288E8E2ULL, "mcr310_sibling6076291.cl"},
	{0xE70793998A95509FULL, "6095816_m.fxe"},
	{0xE707ED3038D29D0AULL, "585126962.ogg"},
	{0xE7081E59EE7F6543ULL, "534817212.ogg"},
	{0xE708480ADDCC907DULL, "art\\vfx\\combat\\vfx_c_mer_attk2_c.anb"},
	{0xE708ED0188846308ULL, "dae200_bran6092465.cl"},
	{0xE708F1EC3ECACFF1ULL, "6116359_m.fxe"},
	{0xE708F4FCB6D65C1CULL, "vxca_caverns2prc_4x7_0_0.vlm"},
	{0xE7091EF93B81CFF2ULL, "6053272_f.fxe"},
	{0xE70934B0620B3AC2ULL, "vxw_woundedcoast_yt_0_0.vlm"},
	{0xE709A475907C405BULL, "vxdr_primevaltomb_6nj_0_1.vlm"},
	{0xE70A2D3177B1E289ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2_tool.mao"},
	{0xE70A393177B1F722ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2_tool.mmh"},
	{0xE70A3B3177B1FAD8ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball2_tool.msh"},
	{0xE70A91F4888C80F8ULL, "545005270.ogg"},
	{0xE70AB07A64F1711FULL, "dae011_corff_sta6149630.cl"},
	{0xE70AD93A55530E0BULL, "plt_cod_hst_wilds.plo"},
	{0xE70AED46DA516D05ULL, "6166613_m.fxe"},
	{0xE70C51DC36AD4042ULL, "vxw_bonepit_1qb_0_0.vlm"},
	{0xE70C688A46849586ULL, "6099211_m.fxe"},
	{0xE70C84EDCE136FD5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholebase.phy"},
	{0xE70CEA5858E10CDBULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3.mmh"},
	{0xE70CFF890BF95E3FULL, "6128330_m.fxe"},
	{0xE70D0BC9918FE422ULL, "vxw_encounter1_et_0_0.vlm"},
	{0xE70D0F657B1BDAC8ULL, "37023696.ogg"},
	{0xE70D14A89ECBB3A6ULL, "dae000_arishok6068103.cl"},
	{0xE70D170B020BC273ULL, "vxdr_primevaltomb_50d_0_2.vlm"},
	{0xE70D46E71540F59EULL, "vxlt_0_lowtown_docks_a3_bat_0_0.vlm"},
	{0xE70ECE217C311F35ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galloalcovlo.msh"},
	{0xE70ED0217C3122ABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galloalcovlo.mmh"},
	{0xE70EDC217C313748ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo12_galloalcovlo.mao"},
	{0xE70F063295965F3FULL, "hlac_ht_0_hightown_65534_1z.dds"},
	{0xE70F0FC5D3C62368ULL, "175068234.ogg"},
	{0xE70FBF2D7EA847ADULL, "377260109.ogg"},
	{0xE70FFEE9E665F5F9ULL, "178413791.ogg"},
	{0xE71080126ADA8B60ULL, "483276440.ogg"},
	{0xE710C6A717A45E3AULL, "6018612_m.fxe"},
	{0xE710DA6B856D8EA4ULL, "hlac_ht_0_hightown_a2_65533_3x.dds"},
	{0xE710DD7E9F6D835CULL, "dae100_bran6118294.cl"},
	{0xE7110B3335DA9ED9ULL, "static_static_p68.psh.fxo.dx11"},
	{0xE71121E6B54714E9ULL, "vxw_woundedcoast_ak_0_0.vlm"},
	{0xE71147531E3F3579ULL, "docks_male_human_worker_sta6175998.cl"},
	{0xE711949B738088ABULL, "vxlt_0_lowtown_a2_n_def_0_0.vlm"},
	{0xE711AB228CCD3A6CULL, "6048307_m.fxe"},
	{0xE711EC8B84C71C77ULL, "zz_propt_debug.ncs"},
	{0xE71217B97C49BEA9ULL, "gen00fl_anders6097438.cl"},
	{0xE7135186E1F8EBBCULL, "vxlt_gallowstemplar_3s8_0_0.vlm"},
	{0xE71430402784DB76ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3_tool.msh"},
	{0xE7143A402784EC74ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3_tool.mmh"},
	{0xE7143E402784F287ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3_tool.mao"},
	{0xE714A14544FE47F0ULL, "vxlt_undercityencounter_1bs_0_0.vlm"},
	{0xE714C5F38482030EULL, "w_bonepit_107_0.mmh"},
	{0xE714D1F3848217A5ULL, "w_bonepit_107_0.mao"},
	{0xE714DFF384822FFCULL, "w_bonepit_107_0.msh"},
	{0xE715ECB8EA8C907DULL, "hlac_lt_0_lowtown_a2_n_-65536_5z.dds"},
	{0xE716E4268CC7F6F5ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_fdeath4.ani"},
	{0xE71714985CF06877ULL, "879539.ogg"},
	{0xE7173DECCC46CF87ULL, "vxw_sundermount_2jz_0_0.vlm"},
	{0xE71791D6D61662A0ULL, "vxlt_warehouse_n_14e_0_0.vlm"},
	{0xE717A68D54015342ULL, "vxw_woundedcoast_vu_0_0.vlm"},
	{0xE717A75A8A1BF9A3ULL, "mag321_comtesse_de_launcet6043330.cl"},
	{0xE717B5CC014DEE81ULL, "vxht_chantry_n_2py_0_0.vlm"},
	{0xE717D6A778607C9AULL, "6103410_f.fxe"},
	{0xE7184A2E3C4FC5F0ULL, "mag101_alain6053009.cl"},
	{0xE718AF57C739AD31ULL, "hlac_ca_caverns2_5802_31z.dds"},
	{0xE7191408AC2C4B11ULL, "vxw_bonepitmining_a2_1jf_0_0.vlm"},
	{0xE7194AA1F63CD4D6ULL, "6118066_m.fxe"},
	{0xE7197E18E9A8872AULL, "vxlt_0_lowtown_a2_gwy_0_0.vlm"},
	{0xE719A03F8752E437ULL, "hlac_dr_0_deeproads_5849_5x.dds"},
	{0xE719F91EAAB58FE5ULL, "vxw_bonepit_2ar_0_0.vlm"},
	{0xE719FD14C6E7050BULL, "vxlt_0_lowtown_docks_a3_n_d2p_0_1.vlm"},
	{0xE71A41681DBC1591ULL, "vxlt_backalley_g_1y3_0_0.vlm"},
	{0xE71AC3B928A358C2ULL, "vxw_bonepit_1zc_0_0.vlm"},
	{0xE71B55A30060C4B1ULL, "zz_followers_banter6158859.cl"},
	{0xE71B7066574C9C42ULL, "lt_0_lowtown_a2_n_44_0_lc.mmh"},
	{0xE71B7DA68385FCA2ULL, "vxw_encounter1_ju_0_0.vlm"},
	{0xE71BB794949C0D0BULL, "hlac_lt_0_lowtown_docks_a3_-65539_3z.dds"},
	{0xE71BD27C3DF5BD0EULL, "llac_w_sundermount_-262149_0.dds"},
	{0xE71CE790BA61495FULL, "lgt250_way_dalish_conflict6147156.cl"},
	{0xE71D3BFDF907B894ULL, "vxht_bartrand_n_3zp_0_0.vlm"},
	{0xE71E06612B36F557ULL, "hlac_lt_0_lowtown_a2_n_65536_2x.dds"},
	{0xE71E3A806F497958ULL, "mag291_demon_marethari6166172.cl"},
	{0xE71E4DE4A8AB4675ULL, "72278993.ogg"},
	{0xE71E94E48C775463ULL, "vxlt_0_lowtown_f08_0_0.vlm"},
	{0xE71EA71A52A25C74ULL, "vxlt_darktown_5l7_0_0.vlm"},
	{0xE71EC016E2F27797ULL, "gen00fl_varric6118995.cl"},
	{0xE71F338F58CB35C9ULL, "910237230.ogg"},
	{0xE71F51B93A2F8AE4ULL, "w_bonepitdestroyed_a3_91_0.mmh"},
	{0xE71F55B93A2F9137ULL, "w_bonepitdestroyed_a3_91_0.mao"},
	{0xE71F67B93A2FAF86ULL, "w_bonepitdestroyed_a3_91_0.msh"},
	{0xE71F78D87B717918ULL, "6125815_m.fxe"},
	{0xE71F99B3FA7B6C9CULL, "gencr_street_thug_2h.utc"},
	{0xE72002D660301589ULL, "vxht_mansion_n_59q_0_0.vlm"},
	{0xE72040C5E5B7D1BFULL, "912775114.ogg"},
	{0xE7209F9472387203ULL, "vxht_0_hightown_n_an2_0_0.vlm"},
	{0xE720D20E1AB96F14ULL, "328061237.ogg"},
	{0xE720DD615FBB0172ULL, "vxlt_warehouse_n_ph_0_0.vlm"},
	{0xE72133EA522EBD1EULL, "6057330_m.fxe"},
	{0xE721937BBFAA9326ULL, "pro000cr_hurlock_rng_fake.utc"},
	{0xE721A1939CE4A744ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbl.ani"},
	{0xE721B70E2E11542AULL, "mcr310_argument6127660.cl"},
	{0xE72277EC859BA248ULL, "mag100_cullen6093550.cl"},
	{0xE722E5077BFE227FULL, "83829087.ogg"},
	{0xE72314D5D302AF2EULL, "6094506_m.fxe"},
	{0xE72445725EDF36C6ULL, "6177695_f.fxe"},
	{0xE724541C2A24FF35ULL, "416644132.ogg"},
	{0xE724A5511450854EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_3.evt"},
	{0xE724A847338FA57EULL, "mer140_bonepit_comments6077835.cl"},
	{0xE7254232007172B2ULL, "6129602_m.fxe"},
	{0xE725629D1A57FBCEULL, "hlac_lt_0_lowtown_docks_n_262142_0z.dds"},
	{0xE725AFC9BE3CE807ULL, "vxw_sundermount_2ou_0_0.vlm"},
	{0xE725FFEB63501293ULL, "6103113_m.fxe"},
	{0xE72603B3C80C7B20ULL, "vxlt_warehouse_n_19b_0_0.vlm"},
	{0xE726186A45F76BC2ULL, "vxw_woundedcoast_sv_0_0.vlm"},
	{0xE726D8575E8C422BULL, "615669233.ogg"},
	{0xE727BF529D805829ULL, "990403635.ogg"},
	{0xE727C4313A4A737CULL, "83535576.ogg"},
	{0xE727EFF5DB9E9FAAULL, "vxlt_0_lowtown_a2_gjz_0_0.vlm"},
	{0xE7284BAC337E85F6ULL, "6217156_m.fxe"},
	{0xE7284E0B5D74C026ULL, "604633004.ogg"},
	{0xE7289E939CEAC0A6ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbl.gad"},
	{0xE728C5E87C04D64DULL, "6117287_f.fxe"},
	{0xE72AB8356C95D1DDULL, "62381218.ogg"},
	{0xE72AF479E8B7F44CULL, "6165190_m.fxe"},
	{0xE72B0EBF2AE93A31ULL, "hlac_lt_0_lowtown_docks_n_-196608_0y.dds"},
	{0xE72B939FE628EBA2ULL, "300624001.ogg"},
	{0xE72B9C51145694F9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_3.gad"},
	{0xE72BA705E06DF7F9ULL, "102824981.ogg"},
	{0xE72BE45608A3BA86ULL, "109834067.ogg"},
	{0xE72C00606AE05B5AULL, "989152598.ogg"},
	{0xE72C9B53ED32CFF5ULL, "fxc_puke_shine.dds"},
	{0xE72CF355AAB6E7BFULL, "dre000_dougal6142213.cl"},
	{0xE72D18F7449874F4ULL, "vxlt_darktown_5q4_0_0.vlm"},
	{0xE72D4815E15C9C7BULL, "334942273.ogg"},
	{0xE72DEAB56D679198ULL, "6122915_m.fxe"},
	{0xE72F3615593C562CULL, "gen00fl_isabela6100014.cl"},
	{0xE72F4F3E51B119F2ULL, "vxlt_warehouse_n_wi_0_0.vlm"},
	{0xE72F99939CF0D685ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkbl.evt"},
	{0xE72F9E537E949916ULL, "vxht_mansion2_33q_0_0.vlm"},
	{0xE72FDEE6C8FB1E8AULL, "and000pt_main.ncs"},
	{0xE730C2DCECB93346ULL, "atl_chargen_dxt5_dat.xml"},
	{0xE73112D930EF7AEFULL, "6165225_m.fxe"},
	{0xE73130CB59E060B5ULL, "6060798_f.fxe"},
	{0xE73198FC8AB2D7D9ULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_atlas_jimmy_0d.dds"},
	{0xE731AD51145B1D47ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_3.ani"},
	{0xE731D7E74A5863ECULL, "zz_mag_debug6054603.cl"},
	{0xE7324DEAF33F99B6ULL, "gen_im_trs_gem_coral_03.uti"},
	{0xE73269B60A48ACE5ULL, "77624368.ogg"},
	{0xE7328AA615265859ULL, "6095164_m.fxe"},
	{0xE732B69D7CDA434EULL, "vxdr_0_deeproads_gba_0_0.vlm"},
	{0xE7341911A9BBE9C5ULL, "vxlt_undercitydungeon_1j2_0_1.vlm"},
	{0xE73421A6B0330087ULL, "vxw_sundermount_2pt_0_0.vlm"},
	{0xE73444BE9DD12DECULL, "vxht_mansion2_265_0_0.vlm"},
	{0xE7345C58A8244264ULL, "576083689.ogg"},
	{0xE734F171E5B10BA9ULL, "vxlt_0_lowtown_a3_n_c_cmt_0_0.vlm"},
	{0xE7351B0BCE54EA0AULL, "vxht_0_hightown_a2_c4p_0_0.vlm"},
	{0xE73534CE357C88A1ULL, "mag121_idunna6092364.cl"},
	{0xE73571273D92CF5DULL, "lt_0_lowtown_docks_a3_n_145_0.mao"},
	{0xE73575273D92D596ULL, "lt_0_lowtown_docks_a3_n_145_0.mmh"},
	{0xE7357F273D92E694ULL, "lt_0_lowtown_docks_a3_n_145_0.msh"},
	{0xE735ABF588A19569ULL, "zz_fol_debug6056994.cl"},
	{0xE7361554F344CAEBULL, "art\\characters\\playercharacter\\humanmale\\rob_anda.phy"},
	{0xE7362068CF91342BULL, "61933964.ogg"},
	{0xE7363CFD6C384A32ULL, "gen00fl_aveline6122794.cl"},
	{0xE737679AAE06E568ULL, "vxlt_backalley_g_1er_0_0.vlm"},
	{0xE737765B864AB9F7ULL, "6070264_m.fxe"},
	{0xE737A108AA6D1C0DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_peak.mmh"},
	{0xE737A508AA6D235EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_peak.mao"},
	{0xE737A61DB7B0775FULL, "6119433_m.fxe"},
	{0xE737B308AA6D3A93ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3cp_peak.msh"},
	{0xE738D6CDADBE01E2ULL, "hlac_lt_backalley_g_65537_2z.dds"},
	{0xE73951757737DC37ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_strong_head_shake.gad"},
	{0xE7398C1D28513C87ULL, "6068489_m.fxe"},
	{0xE739D5795D0F1339ULL, "168287302.ogg"},
	{0xE73A2D97B751B577ULL, "6027541_m.fxe"},
	{0xE73AEEADA005A4A0ULL, "vxht_0_hightown_a2_9wy_0_0.vlm"},
	{0xE73B70F73EEA8338ULL, "zz_lgt_3_debug6140430.cl"},
	{0xE73B7EFA297723B1ULL, "182248343.ogg"},
	{0xE73BED68ECF01558ULL, "mag100_thrask6032061.cl"},
	{0xE73C03039904BD6FULL, "hlac_ca_caverns2prc_5802_29x.dds"},
	{0xE73C752584C4C056ULL, "w_encounter2_14_0.rml"},
	{0xE73C777D8B295E69ULL, "vxw_woundedcoast_ph_0_0.vlm"},
	{0xE73C8D50CBBC84C7ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_avaline.gad"},
	{0xE73CF5C718390288ULL, "hlac_lt_0_lowtown_docks_a3_65535_5x.dds"},
	{0xE73D9C55FD309645ULL, "vxht_0_hightown_a2_a70_0_0.vlm"},
	{0xE73DC11B43A73272ULL, "vxlt_warehouse_n_zj_0_0.vlm"},
	{0xE73DE22580E21BEAULL, "hlac_ht_brothel_n_3367_0z.dds"},
	{0xE73EC6A1A14E3169ULL, "6205660_m.fxe"},
	{0xE73EC6F3DB62401FULL, "art\\vfx\\materials\\chunky\\vfx_c_sphere_deform_blend.mao"},
	{0xE73ED1D476ECAC46ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanr.evt"},
	{0xE73F180061C6F2C0ULL, "vxlt_0_lowtown_a2_n_c_gf1_0_0.vlm"},
	{0xE73F6675773C6BC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_strong_head_shake.evt"},
	{0xE73F94A95FCB63E9ULL, "vxw_bonepit_1xb_0_0.vlm"},
	{0xE73F9F221F77EA16ULL, "hlac_lt_0_lowtown_docks_a3_-65537_0x.dds"},
	{0xE73FE490A028C3A1ULL, "mag170_hugh6094561.cl"},
	{0xE740359E36C86DDFULL, "dae000_pc_home_ambients6208298.cl"},
	{0xE7405D1359573B08ULL, "hlac_lt_0_lowtown_1_3y.dds"},
	{0xE740C80C465A000DULL, "llac_lt_undercityserialkiller_-1_5.dds"},
	{0xE740E91337F02C69ULL, "408650639.ogg"},
	{0xE741B2A42FDD89A8ULL, "6136367_m.fxe"},
	{0xE742474B32AEC1BBULL, "zz_ave_debug6207515.cl"},
	{0xE7427FB290A060C4ULL, "tbe100_xenon6190968.cl"},
	{0xE742883246D1ED5FULL, "vxlt_0_lowtown_n_f50_0_0.vlm"},
	{0xE742985964547C43ULL, "425084358.ogg"},
	{0xE74353B72E363C12ULL, "face2_ds2.psh.fxo"},
	{0xE743630C702B54D0ULL, "lt_0_lowtown_docks_n_200_0.mao"},
	{0xE743670C702B5BE3ULL, "lt_0_lowtown_docks_n_200_0.mmh"},
	{0xE7436E2584CAD37AULL, "w_encounter2_14_0.phy"},
	{0xE743750C702B733DULL, "lt_0_lowtown_docks_n_200_0.msh"},
	{0xE7438CE8C04B028AULL, "vxht_0_hightown_a2_c3q_0_0.vlm"},
	{0xE743B29CD1DB6BD1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_gallowsalcov.mmh"},
	{0xE743BE9CD1DB8072ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_gallowsalcov.mao"},
	{0xE743C4C89CECF7D5ULL, "262881017.ogg"},
	{0xE743CC9CD1DB978FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3wo_gallowsalcov.msh"},
	{0xE744749DA2DAE7C6ULL, "m_dae221ar_chantry.wbd"},
	{0xE7449B320D73EE16ULL, "dre000_stroud6036699.cl"},
	{0xE744BB724CBF185CULL, "hlac_lt_undercitydungeon_-1_48y.dds"},
	{0xE745E8387840D277ULL, "6075164_m.fxe"},
	{0xE745E8D476F2F191ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanr.gad"},
	{0xE74600AE1BA31D3CULL, "6095370_m.fxe"},
	{0xE7465E61D8E30732ULL, "vxht_keep_n_6f0_0_0.vlm"},
	{0xE74684ED5C8EBB5FULL, "592569896.ogg"},
	{0xE74719BAF95CA4BFULL, "vxlt_shop_sr_0_0.vlm"},
	{0xE74743B62B527C2FULL, "292751458.ogg"},
	{0xE7475756E84B6EBCULL, "vxht_0_hightown_aqs_0_0.vlm"},
	{0xE74765F2FAD1AD82ULL, "210854705.ogg"},
	{0xE747A86BED8D77A3ULL, "363176920.ogg"},
	{0xE7480CE0701823EFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedflurrish.gad"},
	{0xE748907884504ED4ULL, "6033190_m.fxe"},
	{0xE749777016517325ULL, "278352149.ogg"},
	{0xE74998DAC3DBA336ULL, "243866779.ogg"},
	{0xE749E85B9AA14C71ULL, "333426457.ogg"},
	{0xE74A881547962FEFULL, "519931537.ogg"},
	{0xE74AA9C26DF63375ULL, "6082978_f.fxe"},
	{0xE74AB453ED4BF7FFULL, "fxc_puke_shine.mao"},
	{0xE74AF13C5E9E0696ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_act_c.impa.evt"},
	{0xE74B53599EF3A5DAULL, "170327084.ogg"},
	{0xE74B744BAC5FC95DULL, "184843683.ogg"},
	{0xE74BE37999380555ULL, "6191347_m.fxe"},
	{0xE74BF9A94E37387FULL, "km_scion_2.mor"},
	{0xE74C401E6A74DAFFULL, "hla_ht_0_hightown_65537_0y.dds"},
	{0xE74CE9D476F9123FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_leanr.ani"},
	{0xE74D7AB97C78219AULL, "art\\characters\\weapons\\boulder\\w_bld_ogre.mmh"},
	{0xE74D7E757748D1BDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\body_strong_head_shake.ani"},
	{0xE74D7EB97C7828D1ULL, "art\\characters\\weapons\\boulder\\w_bld_ogre.mao"},
	{0xE74D8CB97C7840E0ULL, "art\\characters\\weapons\\boulder\\w_bld_ogre.msh"},
	{0xE74DF5B882DAFBCDULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowb.msh"},
	{0xE74E03084E515151ULL, "416925617.ogg"},
	{0xE74E03B882DB1300ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowb.mao"},
	{0xE74E07B882DB1A53ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_bowb.mmh"},
	{0xE74E1815CBF4A275ULL, "6047387_m.fxe"},
	{0xE74EA3352767A7EDULL, "6173408_m.fxe"},
	{0xE74EA565666635A6ULL, "hlac_lt_undercityencounter_-1_24z.dds"},
	{0xE74F1628142F0CFBULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2left.msh"},
	{0xE74F16D1347FF408ULL, "zz_followers_banter6158478.cl"},
	{0xE74F21E0701E659CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedflurrish.evt"},
	{0xE74F2828142F2B46ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2left.mao"},
	{0xE74F3428142F3FE5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2left.mmh"},
	{0xE74F5E1468401325ULL, "gen00fl_bethany6102895.cl"},
	{0xE74FB97D94C1CAD6ULL, "6140654_m.fxe"},
	{0xE75048F287ECC95EULL, "hlac_w_woundedcoast_131079_0z.dds"},
	{0xE7504C0DFEF3BC9DULL, "675070989.ogg"},
	{0xE7507A50CBCD0D0DULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_avaline.ani"},
	{0xE7517E1C61C0D528ULL, "hlac_w_encounter2_camp_196611_0x.dds"},
	{0xE751E83C5EA41621ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_act_c.impa.gad"},
	{0xE751E8E82CFABA3BULL, "mag131tr_atk_rand1.ncs"},
	{0xE75256F49B353E43ULL, "llac_w_encounter2_camp_-131069_0.dds"},
	{0xE7525B9484E0D8EDULL, "llac_ht_0_hightown_a2_65533_0.dds"},
	{0xE752EDEA3F516D48ULL, "814523450.ogg"},
	{0xE75397EDCE4F92F0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholebase.msh"},
	{0xE753A5EDCE4FAA0AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholebase.mmh"},
	{0xE753A9EDCE4FB141ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9cc_8xholebase.mao"},
	{0xE753C1BC35DDC96BULL, "vxw_woundedcoast_1cz_0_0.vlm"},
	{0xE753F14CDE2E314BULL, "follower_banter6151287.cl"},
	{0xE75420EF2869AA0FULL, "952947870.ogg"},
	{0xE75489D79B9CA85FULL, "6113233_m.fxe"},
	{0xE7548B2CF07BBAC2ULL, "vxw_bonepit_1no_0_0.vlm"},
	{0xE754D03ECAD91FB2ULL, "vxht_keep_n_6a1_0_0.vlm"},
	{0xE7550B58591EFA10ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall2x3.phy"},
	{0xE755474072F2152FULL, "llac_ca_caverns1_mines_24424_27.dds"},
	{0xE75556CDEB73A9F9ULL, "110537240.ogg"},
	{0xE7558FAA752FD91AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_maso_0t.dds"},
	{0xE755C933DA41873CULL, "vxht_0_hightown_atp_0_0.vlm"},
	{0xE7566F3EC3ED3F0BULL, "gen00fl_anders6111378.cl"},
	{0xE75691332B0248F0ULL, "gen_im_trs_bst_pelt_03.uti"},
	{0xE756D422928EF81FULL, "vxht_0_hightown_n_c72_0_1.vlm"},
	{0xE7575FE83A077C37ULL, "860123724.ogg"},
	{0xE75762B6310562F7ULL, "6106680_m.fxe"},
	{0xE757657276A52A9CULL, "319495021.ogg"},
	{0xE757CCE2482EFAE1ULL, "lgt311_shw_mer_mettin6156533.cl"},
	{0xE75809B3F9EA7320ULL, "517009961.ogg"},
	{0xE75896F42BDE8B96ULL, "vxlt_warehouse_xa_0_0.vlm"},
	{0xE758DCF538FBE39FULL, "vxw_encounter3_9o_0_0.vlm"},
	{0xE758E93C5EAA36CFULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_act_c.impa.ani"},
	{0xE759BD13BBDBA057ULL, "764136484.ogg"},
	{0xE759F65DBBEE9AA1ULL, "6024546_m.fxe"},
	{0xE75A08819DA65E30ULL, "6171595_m.fxe"},
	{0xE75A139B5EC96BB1ULL, "one200dg_solivitus6060351.cl"},
	{0xE75ACA3438F7E10DULL, "art\\levels\\races\\dwarf\\props\\_textures\\p_cocoon_01d.dds"},
	{0xE75B3A2F3E11B987ULL, "dr_0_deeproads_5838_lc.mmh"},
	{0xE75B6D156A425170ULL, "vxlt_0_lowtown_docks_a3_n_api_0_0.vlm"},
	{0xE75B87F35C1B24D4ULL, "490860841.ogg"},
	{0xE75BDEC6D467BB57ULL, "hla_ht_0_hightown_a3_-65534_0x.dds"},
	{0xE75BE27D6A244C4EULL, "6174046_m.fxe"},
	{0xE75C26898B4767C4ULL, "vxw_encounter2_d4_0_0.vlm"},
	{0xE75C29E07028FDB5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\partypicker\\magestaffselectedflurrish.ani"},
	{0xE75C7DA4B3C87A24ULL, "vxht_mansion2_333_0_0.vlm"},
	{0xE75CAD4EC3719007ULL, "6107412_m.fxe"},
	{0xE75D480AB05009F3ULL, "art\\characters\\creatures\\hurlock\\textures\\cn_hed_hlkb_0n.dds"},
	{0xE75D5BFD690C1D24ULL, "6116567_m.fxe"},
	{0xE75D9979566659D5ULL, "vxlt_darktown_5bx_0_0.vlm"},
	{0xE75DA72BBA9F2F89ULL, "vxw_encounter2_camp_rk_0_0.vlm"},
	{0xE75DB82E2E7000DEULL, "mag100st_thrask.stg"},
	{0xE75DE6581248DAD8ULL, "6129103_m.fxe"},
	{0xE75DEB108E9DEFF5ULL, "770006838.ogg"},
	{0xE75E23EF5C373CC1ULL, "485614506.ogg"},
	{0xE75E3F2954F965ACULL, "vxlt_gallowstemplar_f_3gc_0_0.vlm"},
	{0xE75E682E2C83CEB6ULL, "gen00fl_carver6102540.cl"},
	{0xE75E6E1EF5055AF7ULL, "mag110_arianni6015740.cl"},
	{0xE75EFE9C39EE7659ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_table_sa.dds"},
	{0xE75F04C87E9595A1ULL, "6125466_f.fxe"},
	{0xE75F1FF58358805CULL, "w_bonepit_27_0.rml"},
	{0xE75F44A91F2CE63FULL, "vxlt_0_lowtown_a2_d46_0_0.vlm"},
	{0xE75F5C15FD300FA8ULL, "llac_ht_bartrand_n_3370_0.dds"},
	{0xE75FC06F5AD93332ULL, "llac_lt_0_lowtown_n_65538_3.dds"},
	{0xE76009BA99FC7A87ULL, "6077312_f.fxe"},
	{0xE7602DE8AF1D7162ULL, "174541125.ogg"},
	{0xE76030957FAFAF6CULL, "face2_s2.psh.fxo"},
	{0xE760BD43038E2F85ULL, "hlac_ca_cavern1_haunted_24424_5z.dds"},
	{0xE7613091DD906681ULL, "dae000_meredith6154291.cl"},
	{0xE7614B3C2F64F26AULL, "face0.psh.fxo"},
	{0xE761C30981F0BAE2ULL, "6032007_m.fxe"},
	{0xE76228A64D9B0E91ULL, "follower_banter6155883.cl"},
	{0xE76305F5835C3CB1ULL, "w_bonepit_27_0.msh"},
	{0xE76313F5835C54CCULL, "w_bonepit_27_0.mao"},
	{0xE7631FF5835C696FULL, "w_bonepit_27_0.mmh"},
	{0xE7637505663FBE41ULL, "vxlt_gallowstemplar_n_6qi_0_0.vlm"},
	{0xE7646084484B7CA2ULL, "dae000_bartrand6016466.cl"},
	{0xE764A3E663EF1DE2ULL, "framebuffereffects_hslmatrixfade_p0.psh.fxo.dx11.assembly"},
	{0xE764D50CAA4DA859ULL, "vxca_cavern1_haunted_inz_0_0.vlm"},
	{0xE764F3F978B6F8DEULL, "vxdr_primevalentrance_1o0_0_2.vlm"},
	{0xE765134B86BBD352ULL, "6172691_m.fxe"},
	{0xE7652613655D77B8ULL, "llac_lt_0_lowtown_a2_n_c_-131072_5.dds"},
	{0xE7656462303958FCULL, "llac_ca_caverns1_mines_24423_10.dds"},
	{0xE7657047E0389BF5ULL, "368104397.ogg"},
	{0xE7657432683C7FD4ULL, "6039390_m.fxe"},
	{0xE765E87A4C076345ULL, "500628185.ogg"},
	{0xE765E99C2E7D7516ULL, "sib111_mother6098965.cl"},
	{0xE766388CEFC2CF72ULL, "mag360_thrask6043557.cl"},
	{0xE7666705384CABEEULL, "vxca_cavern1_haunted_cnk_0_0.vlm"},
	{0xE76684C890FFEF58ULL, "hlac_lt_undercityencounter_-1_46x.dds"},
	{0xE766F754919018D8ULL, "6278995.ogg"},
	{0xE7674ED22AF1FC1FULL, "vxw_encounter3_2n_0_0.vlm"},
	{0xE767F8F5A16B5124ULL, "eyelash0.psh.fxo"},
	{0xE7682EAD571E9EDAULL, "follower_banter6161030.cl"},
	{0xE76830465685C0C6ULL, "837440440.ogg"},
	{0xE768685CB911BA92ULL, "hlac_dr_slavers_3214_0z.dds"},
	{0xE768A7F2238BB2F5ULL, "6173963_m.fxe"},
	{0xE768D01D7987E7D8ULL, "lowtown_fem_cityguard_mob6164800.cl"},
	{0xE76A0E4721BFB1C4ULL, "vxht_0_hightown_a2_aus_0_0.vlm"},
	{0xE76A0E48237CAE51ULL, "673669529.ogg"},
	{0xE76A374E07779F97ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_e04_0t.dds"},
	{0xE76A40DEF783B19CULL, "zz_dae_debug6029343.cl"},
	{0xE76A48BD4BBB7012ULL, "vxlt_gallowstemplar_f_4tw_0_0.vlm"},
	{0xE76A5EC7F6276C43ULL, "6170825_m.fxe"},
	{0xE76A98667D3D8044ULL, "vxw_encounter2_o5_0_0.vlm"},
	{0xE76AE5142864A02BULL, "6108606_m.fxe"},
	{0xE76AEA4035ADAD69ULL, "vxw_bonepit_1kg_0_0.vlm"},
	{0xE76B778DC8A9A6E7ULL, "6113606_m.fxe"},
	{0xE76B96058223C952ULL, "335811898.ogg"},
	{0xE76C1908AC954809ULL, "vxw_encounter2_camp_mj_0_0.vlm"},
	{0xE76C9B37E73BD94FULL, "186680558.ogg"},
	{0xE76CD9EC36FA0A35ULL, "6069655_m.fxe"},
	{0xE76D14C18B67AFB9ULL, "671806045.ogg"},
	{0xE76D826196493F29ULL, "vrc321ip_spooky7_obj_5.utp"},
	{0xE76E6816F11A0D07ULL, "938787375.ogg"},
	{0xE76EE42FE396D888ULL, "vxht_0_hightown_n_c5b_0_0.vlm"},
	{0xE76F394027D1966FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall6x3_tool.phy"},
	{0xE770900320E08F66ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_2x.dds"},
	{0xE771823DA592C12FULL, "plt_lgt150pt_ntb_wr_ingrdts.plo"},
	{0xE771FCDA65989A77ULL, "vxht_0_hightown_9bi_0_0.vlm"},
	{0xE7731D5311B27289ULL, "hlac_w_sundermount_-262148_0z.dds"},
	{0xE7733990CE29460FULL, "6127267_m.fxe"},
	{0xE7734E10697CE3EEULL, "6134579_f.fxe"},
	{0xE77378127F58D565ULL, "mer210_coterie_contact6079353.cl"},
	{0xE773852878B1EBD2ULL, "6175791_m.fxe"},
	{0xE77389733DF0EAB1ULL, "ts_beasts_prog_claws.gda"},
	{0xE773B576355FCBD4ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_loop.ani"},
	{0xE773F198C7EBE7BDULL, "llac_dr_slavers_3216_9.dds"},
	{0xE7743B43A85CFB6DULL, "741166686.ogg"},
	{0xE77440210F798CB6ULL, "280382933.ogg"},
	{0xE774E31EF9AB3109ULL, "vxca_caverns1_jtj_0_0.vlm"},
	{0xE774EBC5F9C1F107ULL, "1030593471.ogg"},
	{0xE774F942C6C1BEBDULL, "6096738_m.fxe"},
	{0xE7751DFF06000556ULL, "vxlt_0_lowtown_a3_n_c_e19_0_0.vlm"},
	{0xE775527F5D49D41DULL, "6155310_m.fxe"},
	{0xE777280DEFF2CA7FULL, "1048767350.ogg"},
	{0xE777A6B25A8AC606ULL, "vxht_chantry_n_355_0_1.vlm"},
	{0xE777CA097CDD405FULL, "vxlt_warehouse_n_10j_0_0.vlm"},
	{0xE777D359376445ECULL, "hlac_lt_backalley_n_1_7y.dds"},
	{0xE7787E615D6D5085ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firelogs_sa.dds"},
	{0xE7787F5E345DAA57ULL, "520592545.ogg"},
	{0xE778B9D595D0DD4EULL, "art\\vfx\\combat\\vfx_w_jug_tremor_aoe_p.mmh"},
	{0xE778F0454621EA7EULL, "6039947_m.fxe"},
	{0xE7795C1D27A3C5E9ULL, "vxw_bonepit_1lf_0_0.vlm"},
	{0xE77982869EA2023FULL, "hlac_lt_gallowstemplar_n_2383_5y.dds"},
	{0xE779A4551D8B7E1AULL, "103582056.ogg"},
	{0xE779F651F04C7790ULL, "vxca_caverns1_eic_0_0.vlm"},
	{0xE77A0940564BB512ULL, "245212729.ogg"},
	{0xE77A5DFC0FE58EEAULL, "llac_lt_darktown_-1_75.dds"},
	{0xE77A8AE59E8B6089ULL, "vxw_encounter2_camp_hi_0_0.vlm"},
	{0xE77AB2763565E536ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_loop.gad"},
	{0xE77B065753EF41A2ULL, "vxca_caverns1_mines_foa_0_0.vlm"},
	{0xE77B1B9CDBB77E10ULL, "hlac_lt_0_lowtown_docks_131071_0y.dds"},
	{0xE77B2C4047AA4789ULL, "6109709_f.fxe"},
	{0xE77B7CB21ECC5FD4ULL, "6096022_f.fxe"},
	{0xE77B87B357748F41ULL, "vxht_0_hightown_a3_bp3_0_0.vlm"},
	{0xE77BA278CA05A555ULL, "148850188.ogg"},
	{0xE77BA3355DBFC4ADULL, "static_static_p0.psh.fxo.dx11"},
	{0xE77C066F4C5BC9C8ULL, "follower_banter6162110.cl"},
	{0xE77C326B688D7BC9ULL, "6019115_m.fxe"},
	{0xE77CAB23849CACD2ULL, "qun221_viscount6070796.cl"},
	{0xE77CD9463799C94FULL, "vxlt_warehouse_11r_0_0.vlm"},
	{0xE77CF5AD35229E3AULL, "llac_ht_mansion2b_818_35.dds"},
	{0xE77D04B6644E1DCEULL, "w_encounter1_29_0.msh"},
	{0xE77D12B6644E351FULL, "w_encounter1_29_0.mao"},
	{0xE77D1EB6644E49BCULL, "w_encounter1_29_0.mmh"},
	{0xE77D560CD58CF108ULL, "vxht_0_hightown_n_c2c_0_0.vlm"},
	{0xE77DB6FCEBC35944ULL, "mi_cl.mao"},
	{0xE77DC2AC80CD83A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b.gad"},
	{0xE77DCE736C625BE7ULL, "nat380_sibling6067982.cl"},
	{0xE77E07190D4C1D82ULL, "art\\vfx\\combat\\vfx_str_g_pound_l_leg_c.impa.gad"},
	{0xE77E3E18788FF9DBULL, "gen00fl_merrill6129599.cl"},
	{0xE77E71C708594459ULL, "lt_gallowscourtyard_28_0.rml"},
	{0xE77EE3A6F5E71614ULL, "vxht_keep_7ae_0_0.vlm"},
	{0xE77EE8C4C00ABB00ULL, "lowtown_male_beggar_sta6151049.cl"},
	{0xE77EEE25CCC2861AULL, "llac_lt_0_lowtown_docks_n_-131074_8.dds"},
	{0xE77F212B36998DD9ULL, "w_sundermount_93_0.msh"},
	{0xE77F2B2B36999EC7ULL, "w_sundermount_93_0.mmh"},
	{0xE77F2F2B3699A514ULL, "w_sundermount_93_0.mao"},
	{0xE780060FC2C2FF14ULL, "137103575.ogg"},
	{0xE780836FBBF9DB1DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flowerbunch01_0_n.dds"},
	{0xE78085787BAE5868ULL, "twopassblurcocvertical.psh.fxo"},
	{0xE780B57E217709E3ULL, "shatter_01.dds"},
	{0xE780CD76356A7E95ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_action_loop.evt"},
	{0xE78132CA8358C87BULL, "913244471.ogg"},
	{0xE781410A04BD0DCCULL, "996135638.ogg"},
	{0xE7814E7A354B5B82ULL, "vxlt_darktown_4io_0_0.vlm"},
	{0xE781A3B795BDEE6EULL, "6051880_m.fxe"},
	{0xE7826F2AD9D913C5ULL, "hlac_lt_backalley_g_0_16x.dds"},
	{0xE78293938317C99FULL, "vxlt_0_lowtown_a2_eql_0_0.vlm"},
	{0xE782E59DF8148B77ULL, "zz_vault_debug6203261.cl"},
	{0xE7831BEF970571BDULL, "ave200_donnic6110748.cl"},
	{0xE78357C7A7A46179ULL, "6084984_m.fxe"},
	{0xE7835D1C9A15E9C5ULL, "vxlt_undercityencounter_v1_0_0.vlm"},
	{0xE7837AEF34658B62ULL, "6126566_m.fxe"},
	{0xE783B27DCA7381DBULL, "lt_0_lowtown_docks_a3_112_0.msh"},
	{0xE783C47DCA73A026ULL, "lt_0_lowtown_docks_a3_112_0.mao"},
	{0xE783D07DCA73B4C5ULL, "lt_0_lowtown_docks_a3_112_0.mmh"},
	{0xE783D212A0F743F7ULL, "vxht_chantry_n_27j_0_0.vlm"},
	{0xE784189EFC521A02ULL, "llac_ht_0_hightown_a2_131071_0.dds"},
	{0xE78421D0AE772F0CULL, "gen_im_wep_war_one_act3_04.uti"},
	{0xE784B7AC80D38FB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b.evt"},
	{0xE784FA414D30BCE8ULL, "dre000_sandal6028362.cl"},
	{0xE784FE10B03EC36EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowersmall.msh"},
	{0xE785085F73809104ULL, "464737733.ogg"},
	{0xE7850C10B03EDABFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowersmall.mao"},
	{0xE7851810B03EEF5CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowersmall.mmh"},
	{0xE7852A190D527879ULL, "art\\vfx\\combat\\vfx_str_g_pound_l_leg_c.impa.evt"},
	{0xE785D8D595DB9C4DULL, "art\\vfx\\combat\\vfx_w_jug_tremor_aoe_p.anb"},
	{0xE78762D7F6117FD6ULL, "6117675_f.fxe"},
	{0xE787C8CE0C50D12BULL, "6106406_m.fxe"},
	{0xE787CE9AE75B04A7ULL, "lgt261cr_rogue2.utc"},
	{0xE78806DEFDB2BC0AULL, "hlac_ht_mansion2b_818_48z.dds"},
	{0xE7889B010B342A07ULL, "665805405.ogg"},
	{0xE788C31DA056576CULL, "6048116_m.fxe"},
	{0xE788D42AA35586EAULL, "one100cr_nabil.utc"},
	{0xE7890C32843A3D70ULL, "one280_merin6104386.cl"},
	{0xE78965AE0BF40B0CULL, "dae000_hate_enemies6167590.cl"},
	{0xE78A7F5D3D548C2EULL, "art\\characters\\playercharacter\\humanfemale\\arm_maso.msh"},
	{0xE78A80D3DFEA93BCULL, "603567496.ogg"},
	{0xE78A995D3D54B81CULL, "art\\characters\\playercharacter\\humanfemale\\arm_maso.mmh"},
	{0xE78B1408AFB2A10BULL, "vxlt_0_lowtown_a2_n_c_eos_0_0.vlm"},
	{0xE78C368C50462119ULL, "character_character_p57.psh.fxo.dx11"},
	{0xE78C423DDF4A8C97ULL, "mag121_viveka6151873.cl"},
	{0xE78C63C4E5783943ULL, "vxht_0_hightown_n_c11_0_0.vlm"},
	{0xE78C7DE943EB8BEAULL, "6164425_m.fxe"},
	{0xE78D4E94506A9E05ULL, "1043388312.ogg"},
	{0xE78DC8B14BBF9EE4ULL, "boulder_mac_dust.mao"},
	{0xE78E136151E51FDDULL, "410835587.ogg"},
	{0xE78E29AAFD96A084ULL, "6111287_m.fxe"},
	{0xE78EC723108D3F70ULL, "6152307_m.fxe"},
	{0xE78EE21B826E2C2CULL, "249901740.ogg"},
	{0xE78F45BA0122E382ULL, "gen00fl_anders6128346.cl"},
	{0xE78F5F972566B444ULL, "hlac_w_blightlands_-4_0x.dds"},
	{0xE78FA375D229022FULL, "137269708.ogg"},
	{0xE78FD0FFFCB66020ULL, "544041367.ogg"},
	{0xE78FEE020004CFCEULL, "263771479.ogg"},
	{0xE790BE389BC67859ULL, "hlac_lt_undercityencounter_-1_12x.dds"},
	{0xE79253380D86C9A0ULL, "17387598.ogg"},
	{0xE792618E9F2FE1E1ULL, "vxlt_undercitydungeon_17x_0_0.vlm"},
	{0xE792AFAC80DFBFEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_b.ani"},
	{0xE792F954F6B52C14ULL, "hlac_ht_0_hightown_a2_-65537_0z.dds"},
	{0xE79332190D5EC368ULL, "art\\vfx\\combat\\vfx_str_g_pound_l_leg_c.impa.ani"},
	{0xE79363589F17F009ULL, "428431858.ogg"},
	{0xE79405732E3203AFULL, "mag170_thrask6054047.cl"},
	{0xE7944FA150D795EFULL, "hlac_lt_0_lowtown_a2_n_c_-65535_6z.dds"},
	{0xE794A00A34ED2E3EULL, "6115960_m.fxe"},
	{0xE794AD0463BD96FBULL, "vxlt_hangedman_25m_0_0.vlm"},
	{0xE794DB66BE7B8100ULL, "gen00fl_isabela6101358.cl"},
	{0xE7954EDA6DAAC8BAULL, "vxlt_0_lowtown_a2_n_c_i0g_0_0.vlm"},
	{0xE795EDFD531FC9C4ULL, "vxw_encounter2_p8_0_0.vlm"},
	{0xE79613D3A2ECDE39ULL, "tools_sun_dir.mmh"},
	{0xE7961DD3A2ECEF27ULL, "tools_sun_dir.msh"},
	{0xE79631E684640E7CULL, "hlac_lt_0_lowtown_docks_a3_65535_2y.dds"},
	{0xE7963D9E84ACB45CULL, "800832333.ogg"},
	{0xE7968F76BD0C8697ULL, "llac_w_blightlands_-65534_0.dds"},
	{0xE7970D28146CB292ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3pp_arch2left.phy"},
	{0xE79734FA924C6FECULL, "6043016_m.fxe"},
	{0xE797D4BAC54AB6CAULL, "vxdr_primevaltomb_7rg_0_0.vlm"},
	{0xE798069D1CD1C7ACULL, "vxlt_gallowstemplar_f_3sg_0_0.vlm"},
	{0xE79838D8B5ED0AEEULL, "6049236_m.fxe"},
	{0xE798398B4FE242E0ULL, "284527604.ogg"},
	{0xE7984A9D4E23AF97ULL, "101611649.ogg"},
	{0xE7986B6D3B60C041ULL, "vxht_0_hightown_a3_bj1_0_0.vlm"},
	{0xE798B907B6E2BA15ULL, "llac_ht_0_hightown_-131073_0.dds"},
	{0xE79908F1D3CB6015ULL, "6034749_m.fxe"},
	{0xE7993E691BB48A4DULL, "24925779.ogg"},
	{0xE799486670B0078EULL, "zz_followers_banter6159093.cl"},
	{0xE7996A1DC04F1FEFULL, "hlac_w_encounter2_327683_0x.dds"},
	{0xE79A2DE82EF99F13ULL, "vxdr_0_deeproads_esq_0_0.vlm"},
	{0xE79A3579686E4318ULL, "61697414.ogg"},
	{0xE79A7DD06A33FDA3ULL, "dae000cr_commoner_f1.utc"},
	{0xE79ACB5466B82FBBULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh05a.phy"},
	{0xE79B1EB66467470FULL, "w_encounter1_29_0.rml"},
	{0xE79B305FB7C23063ULL, "dae200_bran6187825.cl"},
	{0xE79BE8B7187AA20EULL, "hlac_lt_0_lowtown_docks_n_131072_0y.dds"},
	{0xE79CBDCBE2337EF0ULL, "762223864.ogg"},
	{0xE79CDE50984AE09EULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_c.impa.evt"},
	{0xE79D3A7297A91928ULL, "gen00fl_dog6165733.cl"},
	{0xE79D5EE2AAFD6621ULL, "568712365.ogg"},
	{0xE79DA02080A325E6ULL, "vxlt_backalley_n_1wo_0_0.vlm"},
	{0xE79DCE4067FC87C8ULL, "cod_mgc_rules.utp"},
	{0xE79E6725851772A3ULL, "w_encounter2_14_0.msh"},
	{0xE79E752585178AFDULL, "w_encounter2_14_0.mmh"},
	{0xE79E79258517910EULL, "w_encounter2_14_0.mao"},
	{0xE79E8FC392301CEDULL, "hlac_lt_0_lowtown_a2_131072_0z.dds"},
	{0xE79F1A49D6E08235ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestnew.msh"},
	{0xE79F1C49D6E085ABULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestnew.mmh"},
	{0xE79F2849D6E09A48ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_crestnew.mao"},
	{0xE79F2C21CC26D28CULL, "6150280_m.fxe"},
	{0xE79F750A3E43E792ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls101a_0d.dds"},
	{0xE79F7906B5A05A3BULL, "darktown_fem_fereldan_sta6156751.cl"},
	{0xE79F7E9A7FCCDB5DULL, "llac_lt_0_lowtown_docks_a3_65532_1.dds"},
	{0xE79F7F5CEAE36AFDULL, "hlac_ht_brothel_n_2995_5x.dds"},
	{0xE79FD816E258319BULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.mmh"},
	{0xE79FE4C77593C7FBULL, "isa211_isabela6071613.cl"},
	{0xE7A0131C019FC4F0ULL, "253449904.ogg"},
	{0xE7A01510D619BA6CULL, "hlac_lt_0_lowtown_docks_n_-1_0y.dds"},
	{0xE7A0489FC1983190ULL, "qun110_petrice6150069.cl"},
	{0xE7A066288A6DAEB5ULL, "vxca_caverns2prc_9pv_0_1.vlm"},
	{0xE7A095EE41B6BE94ULL, "6182467_m.fxe"},
	{0xE7A0DBC57195FF7BULL, "183048805.ogg"},
	{0xE7A16C95F2358CB1ULL, "6048966_m.fxe"},
	{0xE7A1784E8D012478ULL, "vxlt_0_lowtown_a2_n_c_ekj_0_2.vlm"},
	{0xE7A1C1963876A45BULL, "vxlt_0_lowtown_fig_0_0.vlm"},
	{0xE7A2B132185BEE9DULL, "6087471_m.fxe"},
	{0xE7A2B7F7F34BCF8DULL, "6097881_m.fxe"},
	{0xE7A2F550984F73C9ULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_c.impa.gad"},
	{0xE7A31EE155B3AF7BULL, "vxlt_hangedman_22l_0_0.vlm"},
	{0xE7A353B2285E0BB7ULL, "21489207.ogg"},
	{0xE7A38BF33288348CULL, "vxht_0_hightown_n_a83_0_0.vlm"},
	{0xE7A3A4C75DF3588BULL, "6148974_m.fxe"},
	{0xE7A47E89782BE8CEULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v1.ani"},
	{0xE7A513D0DA309510ULL, "hlac_lt_0_lowtown_65536_4z.dds"},
	{0xE7A5147A16F35696ULL, "vxw_encounter3_4q_0_0.vlm"},
	{0xE7A584B9F305F8E5ULL, "art\\characters\\playercharacter\\dwarfmale\\mt_bas_dm_lashesm1.mao"},
	{0xE7A59931C0D738D2ULL, "one120_athenril6113280.cl"},
	{0xE7A6163EF51AA050ULL, "gen00fl_fenris6126551.cl"},
	{0xE7A692B5D30DA530ULL, "atl_slideshow_dxt5_dat.xml"},
	{0xE7A6B805C5A384AEULL, "788027572.ogg"},
	{0xE7A6CD76EA137862ULL, "568559673.ogg"},
	{0xE7A770E6AF872D4EULL, "vxlt_warehouse_n_sa_0_0.vlm"},
	{0xE7A81DF312B70602ULL, "gen_im_trs_krk_low_junk2_02.uti"},
	{0xE7A852EDB9DD4C7EULL, "ht_0_hightown_a3_45_0.rml"},
	{0xE7A854234AE83FC2ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_2z.dds"},
	{0xE7A913CF6CFA9B76ULL, "dae000_keep_noblewoman_sta6162052.cl"},
	{0xE7A99DC6F9B3AC9AULL, "dae000_madam_lusine6152196.cl"},
	{0xE7A9A2A128AF7559ULL, "6095335_m.fxe"},
	{0xE7A9B434B05F8888ULL, "vxht_0_hightown_a2_c7w_0_0.vlm"},
	{0xE7A9C65098554217ULL, "art\\vfx\\combat\\vfx_m_sta_atk_nat_c.impa.ani"},
	{0xE7AA60B820CAF72BULL, "gen00fl_merrill6121098.cl"},
	{0xE7AA67CEBB27100BULL, "65712934.ogg"},
	{0xE7AB510998B4C35DULL, "hlac_lt_0_lowtown_65536_12z.dds"},
	{0xE7AC142B36C065FCULL, "w_sundermount_93_0.phy"},
	{0xE7AC31C485568B90ULL, "vxw_woundedcoast_6w_0_0.vlm"},
	{0xE7AC3F3341230422ULL, "mage_revivingpresence.dds"},
	{0xE7ACF916E262F400ULL, "art\\vfx\\combat\\vfx_c_mer_crack_decal_p.anb"},
	{0xE7ACFB8E57FA65E7ULL, "hlac_lt_0_lowtown_a3_n_c_131071_0y.dds"},
	{0xE7AD8FE04815E9DDULL, "501968934.ogg"},
	{0xE7AE6C7A2A300339ULL, "493564866.ogg"},
	{0xE7AEC1BAA43B37D4ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.anb"},
	{0xE7AECBF5396639E6ULL, "vxw_blightlands_148_0_0.vlm"},
	{0xE7AF4044143D085AULL, "6060905_m.fxe"},
	{0xE7AF7C13AB4A2ED8ULL, "6124755_m.fxe"},
	{0xE7AF9CF8D6960D70ULL, "hla_ht_0_hightown_a2_131073_1x.dds"},
	{0xE7AFF896DDC2F5D9ULL, "vxlt_gallowscourtyard_7co_0_0.vlm"},
	{0xE7B0560F96FC3574ULL, "671179742.ogg"},
	{0xE7B0785D3D754FE7ULL, "art\\characters\\playercharacter\\humanfemale\\arm_maso.phy"},
	{0xE7B08BB0A507A8E8ULL, "6048457_m.fxe"},
	{0xE7B0A510A8D4BFA0ULL, "6092465_m.fxe"},
	{0xE7B0E32D8129F0B1ULL, "6107278_m.fxe"},
	{0xE7B15150C7D06CC6ULL, "vxca_caverns1_mines_dq1_0_0.vlm"},
	{0xE7B18D67174FA042ULL, "ss_dalish_m_1.bnk"},
	{0xE7B1A34772DFC52EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_short.msh"},
	{0xE7B1B14772DFDD7FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_short.mao"},
	{0xE7B1BD4772DFF11CULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_short.mmh"},
	{0xE7B1F3C616120FDCULL, "431283735.ogg"},
	{0xE7B218F4C3FB1938ULL, "vxca_cavern1_haunted_j2e_0_1.vlm"},
	{0xE7B22B2B36C4F824ULL, "w_sundermount_93_0.rml"},
	{0xE7B261EB11F0E289ULL, "lgt210_eau_symp_templar6144743.cl"},
	{0xE7B276897838181FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v1.evt"},
	{0xE7B27AADDB67D58DULL, "lla_ht_0_hightown_a3_-131071_0.dds"},
	{0xE7B2B8635709265CULL, "hlac_ca_caverns1_24422_6y.dds"},
	{0xE7B3865708E96F16ULL, "vxw_encounter3_1r_0_0.vlm"},
	{0xE7B3F565A11AF7CEULL, "zz_mrl_debug6109310.cl"},
	{0xE7B426F6ABB6D95FULL, "541629318.ogg"},
	{0xE7B42F93200432C8ULL, "6048505_m.fxe"},
	{0xE7B43BE74C7221B0ULL, "vrc321_party_barks.cnv"},
	{0xE7B44431FF3B4E87ULL, "6119125_m.fxe"},
	{0xE7B4A8D660A34FC8ULL, "48025982.ogg"},
	{0xE7B5DE886C37A7A2ULL, "static_static_p14.psh.fxo.dx11"},
	{0xE7B5E2C3A17D45CEULL, "vxlt_warehouse_n_vb_0_0.vlm"},
	{0xE7B629F2355D0DD9ULL, "art\\levels\\races\\proxy\\props\\_textures\\woodrope_d.dds"},
	{0xE7B66A272C6057EBULL, "llac_lt_0_lowtown_docks_a3_65535_4.dds"},
	{0xE7B6FA9716EEF552ULL, "gen_im_trs_apo_rng_r0.uti"},
	{0xE7B75DE74C749108ULL, "vrc321_party_barks.bnk"},
	{0xE7B772F6973318D1ULL, "swashbklr_challenge.dds"},
	{0xE7B81541516467B7ULL, "hlac_w_sundermount_-196614_0z.dds"},
	{0xE7B82611A255A108ULL, "vxht_0_hightown_a2_c2t_0_0.vlm"},
	{0xE7B8590DAFC456C2ULL, "lt_0_lowtown_docks_a3_n_20_0_lc.mmh"},
	{0xE7B85F9BB654CF56ULL, "6147123_m.fxe"},
	{0xE7B954127809235DULL, "w_bonepit_165_0.mao"},
	{0xE7B95789783E0208ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_kdown_v1.gad"},
	{0xE7B9581278092996ULL, "w_bonepit_165_0.mmh"},
	{0xE7B9621278093A94ULL, "w_bonepit_165_0.msh"},
	{0xE7B9644CD5E1A9C6ULL, "6133032_m.fxe"},
	{0xE7B98C7C2C7642DEULL, "821485820.ogg"},
	{0xE7B9B74593F1D5CBULL, "6060108_m.fxe"},
	{0xE7BA0043E0C5B4F9ULL, "6140190_m.fxe"},
	{0xE7BA34C0A060A26EULL, "73294715.ogg"},
	{0xE7BACD945B519D05ULL, "one101cr_lia.utc"},
	{0xE7BB0A52E4BDD83BULL, "224715329.ogg"},
	{0xE7BB0A7B1089ED80ULL, "vxlt_warehouse_199_0_0.vlm"},
	{0xE7BB4183588F1DF4ULL, "dae000_mother6121035.cl"},
	{0xE7BB8478392A00F7ULL, "vxca_cavern1_haunted_fo8_0_0.vlm"},
	{0xE7BB8D3B9A16C38EULL, "518667351.ogg"},
	{0xE7BBA0BAA4458997ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_t_p.mmh"},
	{0xE7BBD93123F85EE7ULL, "dae110cr_qunari_m_01.utc"},
	{0xE7BC770C17E4E7B0ULL, "vxdr_0_deeproads_ev9_0_0.vlm"},
	{0xE7BCA13DA5D2E114ULL, "plt_lgt150pt_ntb_wr_ingrdts.nss"},
	{0xE7BCFE1A838B9543ULL, "zz_followers_banter6157719.cl"},
	{0xE7BD2D2E5D5ED4CBULL, "aip_raiderwarrior_base.gda"},
	{0xE7BDAE4E2B757307ULL, "443117919.ogg"},
	{0xE7BDF6F25C7C2C8AULL, "vxca_cavern1_haunted_ckp_0_0.vlm"},
	{0xE7BE187BFF710FE4ULL, "dae011_wench_mob6142284.cl"},
	{0xE7BE6A73CFB90E59ULL, "vxlt_gallowscourtyard_7dn_0_0.vlm"},
	{0xE7BE73B5B5FF1AEDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxa_1.mao"},
	{0xE7BE7ECBC97B0590ULL, "dae210_gate_guard6099083.cl"},
	{0xE7BEFD8D96FDC168ULL, "6043557_m.fxe"},
	{0xE7BF35614090B5BCULL, "w_blightlands_37_0.rml"},
	{0xE7BF92D2702ED7F2ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_act_c.mmh"},
	{0xE7C051E6BAB4ACB3ULL, "6034642_m.fxe"},
	{0xE7C09733F4E4746DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masd_0d.dds"},
	{0xE7C15F4BDBC24741ULL, "dae011_drunk1_sta6144142.cl"},
	{0xE7C1A9A5A82CBBDFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_basic.phy"},
	{0xE7C1B33BEE23F531ULL, "6151152_m.fxe"},
	{0xE7C1DE6DD7B4BDDFULL, "follower_banter6169198.cl"},
	{0xE7C2062BBAF08CEDULL, "sb_skydome.mmh"},
	{0xE7C20A2BBAF0933EULL, "sb_skydome.mao"},
	{0xE7C2182BBAF0AB73ULL, "sb_skydome.msh"},
	{0xE7C21B614092BF11ULL, "w_blightlands_37_0.msh"},
	{0xE7C229614092D72CULL, "w_blightlands_37_0.mao"},
	{0xE7C235614092EBCFULL, "w_blightlands_37_0.mmh"},
	{0xE7C238AF771F931FULL, "vxdr_primevalentrance_37a_0_0.vlm"},
	{0xE7C2D5818C990079ULL, "lgt220_shp_mag_guard6149903.cl"},
	{0xE7C30164F93A8653ULL, "hlac_lt_0_lowtown_-196609_2x.dds"},
	{0xE7C31F96B3E7BC19ULL, "703193903.ogg"},
	{0xE7C3A918817E47DDULL, "vxlt_darktown_2r3_0_0.vlm"},
	{0xE7C3A9674B108DDEULL, "237559155.ogg"},
	{0xE7C3B0D258FD22EAULL, "gen_im_trs_crc_junk2_03.uti"},
	{0xE7C406D97F17FF64ULL, "vxht_0_hightown_a2_b14_0_0.vlm"},
	{0xE7C40DDD88B9F3B3ULL, "6167955_m.fxe"},
	{0xE7C4340AF6047974ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentfireball.ani"},
	{0xE7C463ED9DA8E484ULL, "hla_ht_0_hightown_n_-131070_0x.dds"},
	{0xE7C4AAA301469F39ULL, "gen00fl_carver6097427.cl"},
	{0xE7C4AF2630E9C409ULL, "vxlt_0_lowtown_docks_cs3_0_0.vlm"},
	{0xE7C5018E5AD66656ULL, "vxlt_warehouse_n_13v_0_0.vlm"},
	{0xE7C51EB54D000834ULL, "one100dg_lia6014002.cl"},
	{0xE7C5837F04DBE893ULL, "vxdr_0_deeproads_ebr_0_0.vlm"},
	{0xE7C5B15B27B0AD59ULL, "plt_prm000pt_evra_belt.plo"},
	{0xE7C60777F0D36BFCULL, "275540113.ogg"},
	{0xE7C60FEE6C9D08BEULL, "vxlt_0_lowtown_n_ety_0_0.vlm"},
	{0xE7C6263CA77DC44FULL, "gr_sp_03.mao"},
	{0xE7C697EE944BB988ULL, "vxht_0_hightown_a2_c9u_0_0.vlm"},
	{0xE7C6AF01641CA934ULL, "isa121ar_chantry.lst"},
	{0xE7C6B25466DD405AULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh05a.msh"},
	{0xE7C6C45466DD5EA0ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bh05a.mmh"},
	{0xE7C6D178A84AE7D6ULL, "6140023_m.fxe"},
	{0xE7C71AC3EAD252B0ULL, "vxlt_warehouse_18z_0_0.vlm"},
	{0xE7C73BBF0BD99A78ULL, "1018207113.ogg"},
	{0xE7C8F5B756856F66ULL, "vxlt_backalley_n_1xr_0_0.vlm"},
	{0xE7C978905C43BC43ULL, "art\\characters\\creatures\\strider\\animation\\str_crun.ani"},
	{0xE7CA1590D1A1212FULL, "zz_fen_debug6165496.cl"},
	{0xE7CA310AF608DFD6ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentfireball.gad"},
	{0xE7CA5449AEF796E6ULL, "282098227.ogg"},
	{0xE7CB45C1A1CE9BB0ULL, "gen00fl_fenris6107222.cl"},
	{0xE7CBD083F8442054ULL, "hlac_lt_0_lowtown_docks_a3_65532_0z.dds"},
	{0xE7CBE846F2654201ULL, "6077942_m.fxe"},
	{0xE7CCA6CF3BF617F8ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm04.phy"},
	{0xE7CCDC50C1AF26D9ULL, "vxlt_gallowscourtyard_7im_0_0.vlm"},
	{0xE7CD81142D4E9464ULL, "6032070_m.fxe"},
	{0xE7CDA642F07071AEULL, "vxlt_0_lowtown_docks_n_cvm_0_1.vlm"},
	{0xE7CDCF016422FE3DULL, "isa121ar_chantry.ncs"},
	{0xE7CE3EA51A7C7A7EULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hed_orsa_0d.dds"},
	{0xE7CE3F0F5EA6A9F9ULL, "vxlt_foundry_n_zc_0_0.vlm"},
	{0xE7CF21CE834D88D1ULL, "art\\vfx\\combat\\vfx_c_rw_cyclone_c.impa.ani"},
	{0xE7CF2B54F05F6779ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bookgen_sa.dds"},
	{0xE7CF6FC8A5833B11ULL, "6051418_m.fxe"},
	{0xE7CF85F7BFAAFFDEULL, "vxlt_0_lowtown_a2_e24_0_0.vlm"},
	{0xE7CFF51CF06FBF2AULL, "6094512_f.fxe"},
	{0xE7D008CD453C6044ULL, "vxw_blightlands_61_0_0.vlm"},
	{0xE7D0DEBF990BB740ULL, "lla_ht_0_hightown_a2_131070_0.dds"},
	{0xE7D0E5A547BAA96DULL, "hlac_dr_primevaltomb_657_4y.dds"},
	{0xE7D0ED53EC145D84ULL, "mag110_arianni6049713.cl"},
	{0xE7D109903F82ADE4ULL, "art\\vfx\\combat\\vfx_gib_drkspwn_bld_p.impa.ani"},
	{0xE7D14C0AF60F2C35ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spidertalentfireball.evt"},
	{0xE7D1C642E29B38CBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\pp_female_mage.anb"},
	{0xE7D21AF57374605DULL, "vxlt_darktown_2u2_0_0.vlm"},
	{0xE7D247DFC6AB4F08ULL, "vfx_steamloop_10s.wbd"},
	{0xE7D266BEB83918CAULL, "6129480_m.fxe"},
	{0xE7D2693E1DB2BBC3ULL, "vxca_caverns2_79m_0_0.vlm"},
	{0xE7D2AAA5A83B30E4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_basic.mmh"},
	{0xE7D2AEA5A83B3737ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_basic.mao"},
	{0xE7D2C0A5A83B5586ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\3wo_basic.msh"},
	{0xE7D2CF9D89A1794BULL, "87251345.ogg"},
	{0xE7D2FE745C1965E3ULL, "6144192_m.fxe"},
	{0xE7D3194654301239ULL, "781666931.ogg"},
	{0xE7D39D12C222A916ULL, "hlac_lt_0_lowtown_n_65536_1z.dds"},
	{0xE7D3BF6F608FC88EULL, "follower_banter6165478.cl"},
	{0xE7D40358BFD35205ULL, "pro000_guard.bnk"},
	{0xE7D44F63148237B8ULL, "ser121_jethann6104422.cl"},
	{0xE7D52CCE83520773ULL, "art\\vfx\\combat\\vfx_c_rw_cyclone_c.impa.gad"},
	{0xE7D560F7EA25890FULL, "vxlt_darktown_4f3_0_0.vlm"},
	{0xE7D5FD9B4587528FULL, "6099081_m.fxe"},
	{0xE7D6BC85E563A880ULL, "one101_orders_followers6115762.cl"},
	{0xE7D6E558BFD5553DULL, "pro000_guard.cnv"},
	{0xE7D70A3AE66855F8ULL, "6000960_f.fxe"},
	{0xE7D7776717704A60ULL, "ss_dalish_m_1.wbd"},
	{0xE7D790905C50223AULL, "art\\characters\\creatures\\strider\\animation\\str_crun.evt"},
	{0xE7D79C47730088E7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_short.phy"},
	{0xE7D806903F88C746ULL, "art\\vfx\\combat\\vfx_gib_drkspwn_bld_p.impa.gad"},
	{0xE7D82A57D8FC5073ULL, "gen00fl_anders6119099.cl"},
	{0xE7D84C7698B6C0FEULL, "ht_0_hightown_n_53_0_lc.mmh"},
	{0xE7D8DB049579D7E2ULL, "6032616_m.fxe"},
	{0xE7D91153BFA6EC54ULL, "vxw_bonepitmining_a2_264_0_0.vlm"},
	{0xE7D916865CC69C9CULL, "6120849_m.fxe"},
	{0xE7D933E5C44C9860ULL, "hlac_lt_0_lowtown_a2_-131072_2x.dds"},
	{0xE7DA45E7E9890A45ULL, "vxht_brothel_n_62t_0_0.vlm"},
	{0xE7DA77DEFE9ED5BBULL, "lt_0_lowtown_docks_a3_n_195_0.msh"},
	{0xE7DA89DEFE9EF406ULL, "lt_0_lowtown_docks_a3_n_195_0.mao"},
	{0xE7DA95DEFE9F08A5ULL, "lt_0_lowtown_docks_a3_n_195_0.mmh"},
	{0xE7DAB0B145C8F00BULL, "vxlt_gamlenshouse_n_3n_0_0.vlm"},
	{0xE7DB628FE8D65079ULL, "vxlt_gallowstemplar_n_5uw_0_1.vlm"},
	{0xE7DB6783211CC8C4ULL, "631175315.ogg"},
	{0xE7DC29CE835820D8ULL, "art\\vfx\\combat\\vfx_c_rw_cyclone_c.impa.evt"},
	{0xE7DC2BD744A4C09EULL, "6160780_m.fxe"},
	{0xE7DC808FFF6B2E12ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flma_0s.dds"},
	{0xE7DCC8900E8BE817ULL, "vxdr_primevaltomb_6vg_0_0.vlm"},
	{0xE7DD061EC81789AEULL, "hlac_ht_0_hightown_n_65533_0x.dds"},
	{0xE7DD14C0B1BC3E49ULL, "docks_male_elven_worker_sta6175821.cl"},
	{0xE7DD68E68FAF2C7AULL, "vxw_encounter2_kg_0_0.vlm"},
	{0xE7DD8F905C548C2DULL, "art\\characters\\creatures\\strider\\animation\\str_crun.gad"},
	{0xE7DDD77295525B75ULL, "nat380_nathaniel6066497.cl"},
	{0xE7DE3D9A4A9B1993ULL, "6080642_f.fxe"},
	{0xE7DE7927C55E7169ULL, "vxw_bonepit_13o_0_0.vlm"},
	{0xE7DE812C353D38CDULL, "6119453_m.fxe"},
	{0xE7DEE61E0D4F4940ULL, "vxht_chantry_28k_0_0.vlm"},
	{0xE7DEE69C661BE90EULL, "zz_followers_banter6158928.cl"},
	{0xE7DF01903F8EDD25ULL, "art\\vfx\\combat\\vfx_gib_drkspwn_bld_p.impa.evt"},
	{0xE7DF9E47728B8D4EULL, "421101298.ogg"},
	{0xE7E008F5FCB675F9ULL, "vxlt_gallowstemplar_f_5qj_0_0.vlm"},
	{0xE7E0533F01BC679AULL, "752807760.ogg"},
	{0xE7E065A83E1D5024ULL, "vxw_encounter3_10_0_0.vlm"},
	{0xE7E06FD52494195EULL, "vxca_caverns1_iko_0_0.vlm"},
	{0xE7E08CD2656A78DDULL, "vxlt_darktown_2x1_0_0.vlm"},
	{0xE7E0A6E703AD27A3ULL, "6165265_m.fxe"},
	{0xE7E0A95BC9BB150EULL, "147675190.ogg"},
	{0xE7E0D5687662D000ULL, "vxca_caverns2prc_a1m_0_0.vlm"},
	{0xE7E1030A0920ACE8ULL, "hlac_dr_primevaltomb_9621_0x.dds"},
	{0xE7E1210D8B914649ULL, "hlac_lt_0_lowtown_65538_5z.dds"},
	{0xE7E150285EC3C2FCULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\prp_blightrock_drelief.dds"},
	{0xE7E1A0F8E0882D8AULL, "w_blightlands_93_0.mao"},
	{0xE7E1A4F8E08834D9ULL, "w_blightlands_93_0.mmh"},
	{0xE7E1AEF8E08845C7ULL, "w_blightlands_93_0.msh"},
	{0xE7E1CBB1400B0E5BULL, "hlac_lt_0_lowtown_a2_n_131072_5x.dds"},
	{0xE7E1E7BEA4C30F1BULL, "one300_zevran6089088.cl"},
	{0xE7E21F9FDE30AC91ULL, "vxdr_0_deeproads_bxa_0_0.vlm"},
	{0xE7E24FEDBA0EBF22ULL, "ht_0_hightown_a3_45_0.anb"},
	{0xE7E312FF0BAC78FDULL, "vxlt_0_lowtown_a2_n_etr_0_0.vlm"},
	{0xE7E35E5D4C80311BULL, "mag321cr_servant_gossip_2a.utc"},
	{0xE7E3FE7DCEBE83B0ULL, "vxlt_warehouse_12x_0_0.vlm"},
	{0xE7E44462BA7C67FEULL, "6117979_m.fxe"},
	{0xE7E47B6E4BDE6894ULL, "6117810_m.fxe"},
	{0xE7E504CF20F0FD9DULL, "gen00fl_carver6097315.cl"},
	{0xE7E55BE4D06DFDC1ULL, "blood_conj_06_l2.mao"},
	{0xE7E581F1CFF6D48CULL, "vxca_caverns1_jfe_0_0.vlm"},
	{0xE7E5B59E608CB241ULL, "vxca_cavern1_haunted_img_0_0.vlm"},
	{0xE7E600C92413C3D3ULL, "6055113_m.fxe"},
	{0xE7E675621B949FA0ULL, "hlac_ht_mansion2_818_38y.dds"},
	{0xE7E68C01252A9670ULL, "vxht_mansion2b_1ca_0_0.vlm"},
	{0xE7E70DD76C8EF5B5ULL, "65641456.ogg"},
	{0xE7E7441BBAC68D08ULL, "zz_mag_debug6039794.cl"},
	{0xE7E766CE1A61AB78ULL, "dae000_pickpocket_scene6175161.cl"},
	{0xE7E791D270502799ULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_act_c.anb"},
	{0xE7E7AC5B27CD8B6EULL, "plt_prm000pt_evra_belt.nss"},
	{0xE7E8C450BA298DE6ULL, "gen00fl_merrill6121112.cl"},
	{0xE7E90846A7FEBFD1ULL, "mcr310_argument6072314.cl"},
	{0xE7E976F932448B3BULL, "293458822.ogg"},
	{0xE7E97FE2BD064177ULL, "hlac_lt_gallowscourtyard_n_-65536_2x.dds"},
	{0xE7E9CDD8E2B28469ULL, "vxw_woundedcoast_4t_0_0.vlm"},
	{0xE7EA43CE204C609AULL, "6122786_m.fxe"},
	{0xE7EA90A61F210A0EULL, "llac_w_sundermount_-65533_0.dds"},
	{0xE7EBB601643BD107ULL, "isa121ar_chantry.are"},
	{0xE7EBED8302E548CFULL, "zz_dae_debug6047610.cl"},
	{0xE7EC2B1743D78CB9ULL, "hlac_ca_caverns1_mines_24421_10y.dds"},
	{0xE7EC3D35FD316A81ULL, "6101016_m.fxe"},
	{0xE7EC69B1A39730DEULL, "vxlt_0_lowtown_a2_e86_0_0.vlm"},
	{0xE7ECD3080206A271ULL, "vxca_cavern1_haunted_gkk_0_0.vlm"},
	{0xE7ED0197801B61CAULL, "6095090_f.fxe"},
	{0xE7ED88D08C3E4934ULL, "6048439_m.fxe"},
	{0xE7EEAD16138BDF93ULL, "1067968639.ogg"},
	{0xE7EEB9DE7D401766ULL, "art\\vfx\\textures\\vfx_dust_particles.dds"},
	{0xE7EEC5CAD9A10077ULL, "583720042.ogg"},
	{0xE7EFA52BE9DE5C4FULL, "541361674.ogg"},
	{0xE7EFC8D6F00B531CULL, "671576651.ogg"},
	{0xE7EFF64EB9A768DAULL, "vxht_chantry_20y_0_0.vlm"},
	{0xE7F00EDE3EB97430ULL, "276130654.ogg"},
	{0xE7F07CB8FE4B5428ULL, "fen210_hadriana6110564.cl"},
	{0xE7F0B9C3224478DBULL, "mag121_idunna.bnk"},
	{0xE7F0E41DF074F3C7ULL, "llac_lt_undercityserialkiller_-1_7.dds"},
	{0xE7F15F0BB40284A2ULL, "mi_gla_dis.mao"},
	{0xE7F18785A6750A55ULL, "transit1.cut"},
	{0xE7F1BE1989B85296ULL, "hlac_w_encounter3_-131067_0z.dds"},
	{0xE7F1DB0512F667BAULL, "408857107.ogg"},
	{0xE7F202A8D18B6EC0ULL, "hlac_lt_0_lowtown_docks_-65538_2x.dds"},
	{0xE7F22CD4AB7AE298ULL, "6037367_m.fxe"},
	{0xE7F2C437CE32856FULL, "1037218278.ogg"},
	{0xE7F371C8F1485765ULL, "hlac_lt_gallowsprison_n_-1_13z.dds"},
	{0xE7F39D55052194A9ULL, "hf_bethany_h3s3.mor"},
	{0xE7F3BF7871F5F462ULL, "vxlt_0_lowtown_n_byq_0_0.vlm"},
	{0xE7F3DAFF2F923EC8ULL, "ter_grassa_n.dds"},
	{0xE7F3DE64B30CBC59ULL, "vxw_bonepitdestroyed_a3_1pj_0_0.vlm"},
	{0xE7F3E75FF4B1477AULL, "gen_im_trs_dem_junk1_02.uti"},
	{0xE7F43ED4EEDC1DBDULL, "vxlt_darktown_49t_0_0.vlm"},
	{0xE7F48DCF256A95DCULL, "1036526193.ogg"},
	{0xE7F4900C0C786D04ULL, "gen_im_cns_pot_attribute_01.uti"},
	{0xE7F4BBC32248649BULL, "mag121_idunna.cnv"},
	{0xE7F4FDDE1720AEF0ULL, "vxht_mansion2b_1fn_0_0.vlm"},
	{0xE7F512153E303BFCULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.dura.ani"},
	{0xE7F557244FE2A58AULL, "6052113_m.fxe"},
	{0xE7F58376DF12B850ULL, "ser270_templar6165685.cl"},
	{0xE7F595FE0862197DULL, "follower_banter6165361.cl"},
	{0xE7F6FBBCF47F003BULL, "zz_fol_debug6097814.cl"},
	{0xE7F707087F23C68AULL, "24247056.ogg"},
	{0xE7F736853676E974ULL, "hlac_lt_backalley_g_65537_5y.dds"},
	{0xE7F7443157463BC9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transformtospikeball.ani"},
	{0xE7F76A844B063DEEULL, "1033579914.ogg"},
	{0xE7F7ED52E1B3587AULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5whitelyrium.msh"},
	{0xE7F7FB52E1B36FB3ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5whitelyrium.mao"},
	{0xE7F7FF40AB32B014ULL, "mag291_demon_marethari6028933.cl"},
	{0xE7F7FF52E1B376C0ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_5whitelyrium.mmh"},
	{0xE7F81981663294C1ULL, "6118850_m.fxe"},
	{0xE7F889AAB47F06C7ULL, "vxht_0_hightown_n_c42_0_0.vlm"},
	{0xE7F8B7C97B424521ULL, "6013395_m.fxe"},
	{0xE7F9A0DFD14EBB1EULL, "572456133.ogg"},
	{0xE7F9D6F1B05AADA6ULL, "6033549_m.fxe"},
	{0xE7FA605EDCBD44DBULL, "6176197_m.fxe"},
	{0xE7FAE86274635933ULL, "vxlt_darktown_4wh_0_0.vlm"},
	{0xE7FAF71F486A09B3ULL, "496083645.ogg"},
	{0xE7FB3901A68DA5EDULL, "vxca_caverns2_7mo_0_0.vlm"},
	{0xE7FB64E6192969CDULL, "6113253_m.fxe"},
	{0xE7FC09D08F5D5D53ULL, "llac_lt_0_lowtown_n_-4_0.dds"},
	{0xE7FC1A523727E8C2ULL, "beltof_acumen.dds"},
	{0xE7FC1F153E36703EULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.dura.gad"},
	{0xE7FC2FBA44EA505BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_4.evt"},
	{0xE7FD471090BAFB3CULL, "follower_banter6143313.cl"},
	{0xE7FD5F31574AD54BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transformtospikeball.gad"},
	{0xE7FDDF29FA29DE16ULL, "lt_0_lowtown_docks_a3_207_0.msh"},
	{0xE7FDE929FA29EF14ULL, "lt_0_lowtown_docks_a3_207_0.mmh"},
	{0xE7FDED29FA29F627ULL, "lt_0_lowtown_docks_a3_207_0.mao"},
	{0xE7FE259AB47B4AAFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_mora_0s.dds"},
	{0xE7FE326AC934B681ULL, "783219258.ogg"},
	{0xE7FE38397EA75314ULL, "mag291_vincento6027696.cl"},
	{0xE7FE5431D19440FCULL, "vxlt_warehouse_n_149_0_0.vlm"},
	{0xE7FE90BF622C7919ULL, "dae000_unhappy_mage6165334.cl"},
	{0xE7FEA4E92445BAD7ULL, "817060539.ogg"},
	{0xE7FED3EF705D3F20ULL, "6063876_m.fxe"},
	{0xE7FF0D28A558341EULL, "hf_capturedmage.mor"},
	{0xE7FF3232A58F0307ULL, "hlac_ht_0_hightown_n_65535_1x.dds"},
	{0xE7FF56BBEE6E6049ULL, "hlac_lt_0_lowtown_a2_n_c_1_0x.dds"},
	{0xE7FFED4F1B7F9624ULL, "vxht_mansion_n_61k_0_0.vlm"},
	{0xE80058EE74824291ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hm_boo_cas_0d.dds"},
	{0xE8009D975A08522FULL, "6106396_m.fxe"},
	{0xE8010C27FE94A3AEULL, "90091298.ogg"},
	{0xE801700B0512802BULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\8gr_podium_1.mao"},
	{0xE8018972194CB6EEULL, "hlac_lt_0_lowtown_-262144_1x.dds"},
	{0xE801957E6F4DDCB4ULL, "sse_ss_qunari_m_3.gda"},
	{0xE8024B39DA66FAC1ULL, "mag110_arianni6016901.cl"},
	{0xE8025041A502D4D9ULL, "vxw_bonepitdestroyed_a3_1ui_0_0.vlm"},
	{0xE802B51FB51168A5ULL, "vxca_caverns2prc_acr_0_0.vlm"},
	{0xE802EC254C031D26ULL, "vxlt_0_lowtown_eoo_0_0.vlm"},
	{0xE802FA153E3C505DULL, "art\\vfx\\combat\\vfx_m_amg_elemark_c.dura.evt"},
	{0xE803138DF7D78806ULL, "hlac_lt_0_lowtown_a2_-131072_3y.dds"},
	{0xE80328BA44F0629CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_4.gad"},
	{0xE8036FBB0916C770ULL, "vxht_mansion2b_1yo_0_0.vlm"},
	{0xE803C26D3EABD384ULL, "vxlt_0_lowtown_a2_n_c_gf5_0_0.vlm"},
	{0xE8041002B8976DA9ULL, "179357957.ogg"},
	{0xE8045C315750EEB0ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\transformtospikeball.evt"},
	{0xE804FC49E012B5A7ULL, "270061908.ogg"},
	{0xE8054B1BFA22979DULL, "vxlt_gallowstemplar_f_46r_0_0.vlm"},
	{0xE8056211F448F1F6ULL, "vxlt_hangedman_n_266_0_0.vlm"},
	{0xE805780DDB26E122ULL, "6137035_f.fxe"},
	{0xE805E495AE5AEBC4ULL, "mag311_keran6054232.cl"},
	{0xE805ED76F03C8F57ULL, "art\\vfx\\combat\\vfx_gib_qunc_rtlb.anb"},
	{0xE8068B5E5828AD41ULL, "6117950_m.fxe"},
	{0xE806D2880192D743ULL, "gen_im_trs_dwv_junk4_03.uti"},
	{0xE8078328B986C417ULL, "6115785_m.fxe"},
	{0xE808608A9423BBC5ULL, "vxw_encounter2_kl_0_0.vlm"},
	{0xE808BE80E14E7F98ULL, "lowtown_fem_prostitute_sta6161613.cl"},
	{0xE809DEC96549E0A5ULL, "429952172.ogg"},
	{0xE809E1AACA371A82ULL, "ave200_donnic6111993.cl"},
	{0xE809E65519ADB802ULL, "316240583.ogg"},
	{0xE80A27BA44F67F9AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_generic_large_4.ani"},
	{0xE80A7C16E9E10FC3ULL, "6156212_m.fxe"},
	{0xE80AD626320A5C2AULL, "mcr370_cut_epilogue6086868.cl"},
	{0xE80AF360BF158196ULL, "vxht_keep_n_6hz_0_0.vlm"},
	{0xE80B18A667D464E8ULL, "6092292_m.fxe"},
	{0xE80B2F227B1D4BAEULL, "861644692.ogg"},
	{0xE80B5E8CD298BF79ULL, "vxlt_gallowstemplar_f_5bo_0_0.vlm"},
	{0xE80BB55911E105FBULL, "mcr370_assignments6081740.cl"},
	{0xE80C0C35F48A5A6EULL, "fen210_tevinter_ambusher6108306.cl"},
	{0xE80C1CB3C3AF95E4ULL, "6128773_m.fxe"},
	{0xE80C44AD5618F65EULL, "follower_banter6167982.cl"},
	{0xE80C89A2D39F9760ULL, "vxht_0_hightown_a3_aox_0_0.vlm"},
	{0xE80CD665F0B9F903ULL, "art\\characters\\playercharacter\\elffemale\\boo_lgta.mao"},
	{0xE80CDA65F0BA0050ULL, "art\\characters\\playercharacter\\elffemale\\boo_lgta.mmh"},
	{0xE80CDBEB80872AC6ULL, "vxlt_gallowsprison_n_64i_0_0.vlm"},
	{0xE80CE865F0BA186AULL, "art\\characters\\playercharacter\\elffemale\\boo_lgta.msh"},
	{0xE80CEA5008094C5DULL, "698327868.ogg"},
	{0xE80D8F2184DAF2DFULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockcxxl.msh"},
	{0xE80D952184DAFC81ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockcxxl.mmh"},
	{0xE80DA12184DB1122ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockcxxl.mao"},
	{0xE80DA691EB1D719FULL, "vxca_cave_1a0_0_0.vlm"},
	{0xE80E4E42F9E1E89EULL, "art\\vfx\\combat\\vfx_g_spider_fb_act_c.mmh"},
	{0xE80ED40B95CFC258ULL, "6165565_m.fxe"},
	{0xE80ED58A883842A9ULL, "hlac_lt_0_lowtown_a2_n_-65535_3y.dds"},
	{0xE80F312F62E093BDULL, "dae300_bran6143471.cl"},
	{0xE80F61298E064E8AULL, "zz_fen_debug6134376.cl"},
	{0xE80F8F6D55FD8336ULL, "6137205_m.fxe"},
	{0xE810201F442BC387ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_flemeth_female.gad"},
	{0xE8103D3A4B6AB15CULL, "zz_ave_debug6174449.cl"},
	{0xE810412D94FCC97BULL, "730348372.ogg"},
	{0xE8105574B95D036BULL, "1069736557.ogg"},
	{0xE810CAF6D48C40CDULL, "rdr111dg_martin6053568.cl"},
	{0xE810DD09F4C9651CULL, "vxlt_0_lowtown_a3_n_c_eoq_0_0.vlm"},
	{0xE810E09E44AFA07AULL, "static_static_p5.vsh.fxo.dx11.assembly"},
	{0xE810E3558145A3EFULL, "595192488.ogg"},
	{0xE811D5116CD6D5A1ULL, "321930258.ogg"},
	{0xE811E32A70EFA2CAULL, "dae000_junar6147865.cl"},
	{0xE81208BE16D63D5EULL, "vxht_0_hightown_n_be1_0_0.vlm"},
	{0xE8122B6801D6CE59ULL, "vxca_cavern1_haunted_i2f_0_0.vlm"},
	{0xE81276BDE07A8C22ULL, "llac_w_sundermount_-327682_0.dds"},
	{0xE812B7E3D3B0AF7EULL, "vxlt_0_lowtown_a2_f4q_0_0.vlm"},
	{0xE813190F5199F9C2ULL, "vxlt_backalley_g_1w2_0_0.vlm"},
	{0xE81351DE5B5D8C48ULL, "6139630_m.fxe"},
	{0xE813C641F74D5BB8ULL, "vxht_mansion_n_5at_0_0.vlm"},
	{0xE813E770A3443D55ULL, "352003446.ogg"},
	{0xE8141ECC321C3CE8ULL, "l_env_wal_gallows_int_military_night.bnk"},
	{0xE81424F7B88DC86CULL, "vxca_cavern1_haunted_ihy_0_0.vlm"},
	{0xE814A65624D681A7ULL, "dae000ip_crowd_lm2_talk18.utp"},
	{0xE8151FA068206ECBULL, "gen00fl_sebastian6180816.cl"},
	{0xE8153A250F785D1AULL, "llac_lt_0_lowtown_a2_n_c_-131072_7.dds"},
	{0xE81589908C7F6544ULL, "946912010.ogg"},
	{0xE815EE1A97F850C3ULL, "147023855.ogg"},
	{0xE815F505AB7CDC97ULL, "6110885_m.fxe"},
	{0xE8164D68D3C7186DULL, "hlac_ht_0_hightown_1_2y.dds"},
	{0xE81693B0C81CCA1FULL, "llac_lt_shop_-1_3.dds"},
	{0xE816EE4BF17C6FE0ULL, "vxlt_gallowstemplar_4um_0_0.vlm"},
	{0xE81704691ADDB716ULL, "vxlt_0_lowtown_a2_emu_0_0.vlm"},
	{0xE817151F4431CF94ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_flemeth_female.evt"},
	{0xE817419D4E8496B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\left_up.ani"},
	{0xE81775B316546B29ULL, "qcr000st_aveline.stg"},
	{0xE817CA7868779EB5ULL, "6018721_m.fxe"},
	{0xE818854D69575B5EULL, "45703522.ogg"},
	{0xE818E9EA6978D7C6ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.impa.evt"},
	{0xE8196DDFC6E792CAULL, "vfx_steamloop_10s.bnk"},
	{0xE81996F2AD93BE0CULL, "dae030cr_fem_fer_urchin.utc"},
	{0xE819DE3158AD3747ULL, "6171325_m.fxe"},
	{0xE81A1E1BFA097C9EULL, "follower_banter6148761.cl"},
	{0xE81A4DA7A0BDF009ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_3x.dds"},
	{0xE81B3EDBA5F22BF2ULL, "6162272_m.fxe"},
	{0xE81B407D223E4939ULL, "6037010_m.fxe"},
	{0xE81BAF53DE3FD334ULL, "hlac_w_bonepitmining_a2_393218_0x.dds"},
	{0xE81BEBABAFC03DE1ULL, "vxw_encounter2_a6_0_0.vlm"},
	{0xE81C2C06DDAD315EULL, "qun111_ketojan6039150.cl"},
	{0xE81C2D42F9EDEE5DULL, "art\\vfx\\combat\\vfx_g_spider_fb_act_c.anb"},
	{0xE81CB1A703F540E0ULL, "vxht_0_hightown_caa_0_0.vlm"},
	{0xE81CBE682F9B5FE2ULL, "follower_banter6172962.cl"},
	{0xE81D1089476C37F7ULL, "6125881_m.fxe"},
	{0xE81D909426E92A99ULL, "432345017.ogg"},
	{0xE81DA80DF097699CULL, "136563213.ogg"},
	{0xE81DA900E95FB6C8ULL, "hightown_market_guard_sta6169635.cl"},
	{0xE81E1C9D4E8A76FBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\left_up.gad"},
	{0xE81E4119613F953DULL, "follower_banter6156107.cl"},
	{0xE81E5C93D7545F3DULL, "lit_tep_distor.mao"},
	{0xE81F33FB88EF05D9ULL, "vxw_bonepitdestroyed_a3_1co_0_0.vlm"},
	{0xE81FF4C51CAACDDDULL, "329552683.ogg"},
	{0xE82000EA697F1D11ULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.impa.gad"},
	{0xE8201E1C898686B9ULL, "art\\vfx\\materials\\chunky\\vfx_c_potion.mao"},
	{0xE82034F2F8C3DBADULL, "vxdr_primevalentrance_22y_0_0.vlm"},
	{0xE8205374ED02F870ULL, "vxht_mansion2b_1wm_0_0.vlm"},
	{0xE8207760D31375A4ULL, "ht_chantry_n_mmap_ph.dds"},
	{0xE820EF31302DD657ULL, "one110_gustav6173922.cl"},
	{0xE820F3A3B95ABBF8ULL, "dae011_drunk1_sta6149678.cl"},
	{0xE82112FC1C51B5BEULL, "gen00fl_dog6177480.cl"},
	{0xE82182260F3CA487ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masa_0t.dds"},
	{0xE821B09D7C144256ULL, "26904519.ogg"},
	{0xE821CB9EF94FB15CULL, "vxlt_warehouse_19e_0_0.vlm"},
	{0xE8223A927FC714DBULL, "qun160_vashothlead6053925.cl"},
	{0xE822AED54F2D4C29ULL, "art\\vfx\\combat\\vfx_gib_hm_meda_ltr.anb"},
	{0xE822EFF2F43A8273ULL, "6174436_m.fxe"},
	{0xE822F780E65CBA21ULL, "142151573.ogg"},
	{0xE8233C07EFA75289ULL, "dae100cr_brothel_gamlen.utc"},
	{0xE823B71279CD1923ULL, "210587592.ogg"},
	{0xE823E378AA1E0530ULL, "ht_chantry_mmi.gff"},
	{0xE823EE9AA1638492ULL, "vxlt_gallowstemplar_n_4qp_0_0.vlm"},
	{0xE824399D4E8F13E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\left_up.evt"},
	{0xE824AA1D2E878127ULL, "6115590_m.fxe"},
	{0xE824F43F02059A56ULL, "6096319_m.fxe"},
	{0xE824F804CC921B03ULL, "vxca_cave_2bn_0_0.vlm"},
	{0xE8250D1F443DFFCDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_die_flemeth_female.ani"},
	{0xE8254136BC678576ULL, "893106771.ogg"},
	{0xE8254444780FECC5ULL, "vxw_encounter2_qj_0_0.vlm"},
	{0xE825E590DD69739CULL, "gen00fl_dog6141098.cl"},
	{0xE8263DF94A45A2B3ULL, "vxlt_darktown_4dm_0_0.vlm"},
	{0xE82666D3989AF863ULL, "1017136181.ogg"},
	{0xE82691CFD443EA86ULL, "vxw_bonepitdestroyed_a3_25a_0_0.vlm"},
	{0xE826E6C261F7CCC7ULL, "6129351_m.fxe"},
	{0xE82701EA69853DBFULL, "art\\vfx\\creatures\\fxc_arcane_tele_p.impa.ani"},
	{0xE82793CF3C42A28DULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm04.msh"},
	{0xE827A5CF3C42C113ULL, "art\\characters\\playercharacter\\humanmale\\mt_cbt_hm04.mmh"},
	{0xE827B9D03D908B3EULL, "mag371_elsa6041252.cl"},
	{0xE8281289EC78061CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanr.ani"},
	{0xE828289C00B0A024ULL, "pro000_guard6013443.cl"},
	{0xE8282C86D4FD8764ULL, "6059349_m.fxe"},
	{0xE828500E4AA34FC7ULL, "6174025_m.fxe"},
	{0xE828824A65B8A859ULL, "lt_0_lowtown_a2_27_0.anb"},
	{0xE828A2B65AFFE146ULL, "disabled_disarm.cur"},
	{0xE829055B8864E02AULL, "art\\vfx\\combat\\vfx_glb_ele_f_deathblow_c.impa.gad"},
	{0xE8294D6FA6185A94ULL, "vxlt_0_lowtown_docks_n_b92_0_0.vlm"},
	{0xE829826591498101ULL, "vxlt_0_lowtown_a2_n_c_hsp_0_0.vlm"},
	{0xE829B0B897E84472ULL, "6167172_m.fxe"},
	{0xE829BCD5B14DE3EEULL, "hlac_lt_0_lowtown_a2_n_c_-131073_3y.dds"},
	{0xE829C18592ECE167ULL, "vxht_chantry_31r_0_0.vlm"},
	{0xE829F6E734A5F43CULL, "207227848.ogg"},
	{0xE82A3EBCDCD248C9ULL, "plt_cod_cha_varric3.plo"},
	{0xE82A5D88A1B65661ULL, "vxw_encounter2_l5_0_0.vlm"},
	{0xE82ACFFBBC1D36CBULL, "165510348.ogg"},
	{0xE82BB830D3D8FDBCULL, "dae200_alistair_drunk6135189.cl"},
	{0xE82BD626B100F232ULL, "vxlt_gallowstemplar_3or_0_0.vlm"},
	{0xE82C3F66C435D1F7ULL, "vxlt_0_lowtown_h4v_0_0.vlm"},
	{0xE82C88C05F799922ULL, "vxht_mansion_n_54m_0_0.vlm"},
	{0xE82CB738CFE7FBC9ULL, "6042887_m.fxe"},
	{0xE82CCC4ABC846DA9ULL, "6133071_m.fxe"},
	{0xE82DA600AD85C44EULL, "llac_lt_ruinedhouse_n_-1_0.dds"},
	{0xE82DD12F5AFD7F53ULL, "6060438_m.fxe"},
	{0xE82E128C46F7CAD7ULL, "1053828101.ogg"},
	{0xE82EEC77FAC26E5EULL, "vxht_0_hightown_n_bk3_0_0.vlm"},
	{0xE82F1F89EC7E3B5EULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanr.gad"},
	{0xE82F43F513C1C2C0ULL, "qun210cr_guard6087760.cl"},
	{0xE82FE6E29A024464ULL, "6044075_m.fxe"},
	{0xE82FE85B886ACDC1ULL, "art\\vfx\\combat\\vfx_glb_ele_f_deathblow_c.impa.evt"},
	{0xE8301250E3497E7DULL, "hlac_ca_cavern1_haunted_24427_4x.dds"},
	{0xE83051A75BC6EDFBULL, "mag330_walter6045018.cl"},
	{0xE8308E5AB8883FF0ULL, "qun221_petrice6077312.cl"},
	{0xE830D17791BB6AC3ULL, "dae020cr_hf_noble1.utc"},
	{0xE830DD4B79364A07ULL, "258181176.ogg"},
	{0xE83126C34E801C20ULL, "vxca_cavern1_haunted_gje_0_0.vlm"},
	{0xE831724107FA4B96ULL, "6142753_m.fxe"},
	{0xE831A75B16A1E820ULL, "zz_followers_banter6158643.cl"},
	{0xE831BB6422ED1B26ULL, "6137546_m.fxe"},
	{0xE832E7FF81AEA390ULL, "vxw_bonepitmining_a2_1xe_0_0.vlm"},
	{0xE8331D09CEC36C0DULL, "667713865.ogg"},
	{0xE8333B84423C226EULL, "6104153_m.fxe"},
	{0xE83349F2498F86B0ULL, "862469078.ogg"},
	{0xE833E52AF6525513ULL, "mag100_karras6035369.cl"},
	{0xE833E63E630B0248ULL, "vxw_encounter2_l0_0_0.vlm"},
	{0xE833ECBABD4615BBULL, "con_ico_follower.dds"},
	{0xE83404DFCDA83BE3ULL, "w_blightlands_45_0.mao"},
	{0xE83408DFCDA84230ULL, "w_blightlands_45_0.mmh"},
	{0xE83416DFCDA85A4AULL, "w_blightlands_45_0.msh"},
	{0xE834D349972553D9ULL, "vxlt_0_lowtown_gf5_0_0.vlm"},
	{0xE834FF58B6D38FF2ULL, "196254669.ogg"},
	{0xE83503ACC63A0306ULL, "vxw_bonepitdestroyed_a3_20b_0_0.vlm"},
	{0xE83550D61ECA83E4ULL, "mrl351ar_demon_cave.are"},
	{0xE83571DD5E1B7FA4ULL, "vxlt_0_lowtown_a2_ej6_0_0.vlm"},
	{0xE835FA89EC841B7DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\movement\\runleanr.evt"},
	{0xE8367FDF4B636C6DULL, "vxca_cavern1_haunted_ei3_0_0.vlm"},
	{0xE8371B291789E24CULL, "isa211_isabela6202267.cl"},
	{0xE83732E43FF5E3F0ULL, "dae000ip_crowd_mf8_talk8.utp"},
	{0xE837BF4C980E7314ULL, "vxlt_0_lowtown_docks_n_b63_0_0.vlm"},
	{0xE837D2B1F44962A4ULL, "vxht_keep_n_6h8_0_0.vlm"},
	{0xE838105726B32956ULL, "712705680.ogg"},
	{0xE838146394D0D7F8ULL, "116360090.ogg"},
	{0xE8383E7C5FF9742CULL, "hlac_lt_undercityserialkiller_-1_25y.dds"},
	{0xE83847B2B3859264ULL, "vxht_brothel_n_4h5_0_0.vlm"},
	{0xE83910CC323B4092ULL, "l_env_wal_gallows_int_military_night.wbd"},
	{0xE8391CA6F7BC6119ULL, "6025314_m.fxe"},
	{0xE8396E6DFE2A01D4ULL, "6118051_m.fxe"},
	{0xE8399621850039FEULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockcxxl.phy"},
	{0xE839C5C30612AB03ULL, "486375093.ogg"},
	{0xE839F24D04B025B8ULL, "37793298.ogg"},
	{0xE839F5A2EA140359ULL, "vxdr_primevaltomb_8mj_0_0.vlm"},
	{0xE83A26D623FC7D88ULL, "vxht_0_hightown_a2_cam_0_0.vlm"},
	{0xE83A5B04E0DFEC39ULL, "6137677_m.fxe"},
	{0xE83A8BBA99A58EC8ULL, "llac_ht_brothel_n_3365_0.dds"},
	{0xE83B054B64664415ULL, "262491751.ogg"},
	{0xE83B94FACB7DE7F0ULL, "6034295_f.fxe"},
	{0xE83BBEC8A75EA249ULL, "vxht_mansion2b_2vd_0_0.vlm"},
	{0xE83BEF2D32302134ULL, "726790807.ogg"},
	{0xE83C32A0CAABAE9CULL, "vxlt_0_lowtown_a3_n_c_ext_0_0.vlm"},
	{0xE83CCB52D15700B5ULL, "lgt220_mag_tmplr_templar6140348.cl"},
	{0xE83CE05B88754A10ULL, "art\\vfx\\combat\\vfx_glb_ele_f_deathblow_c.impa.ani"},
	{0xE83D0BADF96FDB0DULL, "625074775.ogg"},
	{0xE83D23F24B962A2CULL, "warrior_2.cut"},
	{0xE83D58508F6E7E60ULL, "6108515_m.fxe"},
	{0xE83D61D547749B7AULL, "llac_ht_0_hightown_131071_2.dds"},
	{0xE83D7619A5C1F699ULL, "gen_im_arm_cht_coma_t2.uti"},
	{0xE83ECAA36DB142C0ULL, "6156039_m.fxe"},
	{0xE83EDC9BBD01EABFULL, "vxca_cavern1_haunted_ihv_0_0.vlm"},
	{0xE83F1CA0E19AFC4EULL, "ht_bartrand_n_3364_lc.mmh"},
	{0xE83F6B11BE0B9A27ULL, "fxa_evl_aura.mao"},
	{0xE83FB8ABCC643944ULL, "6048387_m.fxe"},
	{0xE83FF5B76F880F3AULL, "llac_ht_0_hightown_a2_-1_1.dds"},
	{0xE8404AA5004F659EULL, "vxdr_0_deeproads_f4r_0_0.vlm"},
	{0xE8407302D92C4A10ULL, "hlac_lt_gallowstemplar_2384_6y.dds"},
	{0xE840868747CC6A54ULL, "dae000cr_grand_cleric6178496.cl"},
	{0xE8408AF4BEC387C0ULL, "llac_lt_hangedman_n_-1_4.dds"},
	{0xE8409186945B5A9AULL, "533772258.ogg"},
	{0xE840D77B58255462ULL, "609602984.ogg"},
	{0xE840D9E4F1C07446ULL, "dae200cr_viscount6073935.cl"},
	{0xE84101F8307CE388ULL, "vxca_cavern1_haunted_cmt_0_0.vlm"},
	{0xE841185EC3E57C65ULL, "follower_banter6143134.cl"},
	{0xE84159DC73A4BC10ULL, "vxw_bonepitmining_a2_1ub_0_0.vlm"},
	{0xE841B86FCE3FC78BULL, "hlac_lt_0_lowtown_docks_a3_-196609_1y.dds"},
	{0xE841BD852955CA23ULL, "795645766.ogg"},
	{0xE842F2A3D3C40281ULL, "dae220ip_chant_board_ht.ncs"},
	{0xE8430C0B9D6F7FC9ULL, "hlac_lt_darktown_-1_78z.dds"},
	{0xE8432424C79058E5ULL, "art\\characters\\weapons\\spears\\textures\\w_spr_003a_0d.dds"},
	{0xE843A31E09B10773ULL, "6142389_m.fxe"},
	{0xE843C5AD144C6F10ULL, "6030241_m.fxe"},
	{0xE843CA10AF9863C0ULL, "qun160_vashothlead6055579.cl"},
	{0xE8442FCA34526E58ULL, "281112065.ogg"},
	{0xE8443508DB9BB2D2ULL, "639046194.ogg"},
	{0xE84435AEA0C9B281ULL, "vxlt_warehouse_194_0_0.vlm"},
	{0xE844A5F8E0DCB3C6ULL, "w_blightlands_93_0.phy"},
	{0xE845032F5F041B68ULL, "511451698.ogg"},
	{0xE8450D9745A70899ULL, "pro000_athenril6048729.cl"},
	{0xE845117E24B4F3EEULL, "6119651_f.fxe"},
	{0xE8454DAE411D53EAULL, "6118048_m.fxe"},
	{0xE8463C4E396D59BDULL, "vxlt_warehouse_n_cv_0_0.vlm"},
	{0xE846448EE63F7B24ULL, "vxht_keep_n_6c9_0_0.vlm"},
	{0xE84721B66C729430ULL, "ht_0_hightown_26_0.rml"},
	{0xE8478B3B26FEAA3BULL, "vxht_chantry_n_2ba_0_0.vlm"},
	{0xE8478FFBD9B4219CULL, "6125693_m.fxe"},
	{0xE8479AB2680FA8DBULL, "vxw_encounter2_camp_94_0_0.vlm"},
	{0xE847EB0E11EE4BAEULL, "follower_barks6153788.cl"},
	{0xE847F4DDA0E0313AULL, "vxlt_hangedman_n_31j_0_0.vlm"},
	{0xE8485B1CC07046D4ULL, "gen00fl_sebastian6177129.cl"},
	{0xE8487C26E64EA3E3ULL, "492463364.ogg"},
	{0xE849A507340D711DULL, "542932453.ogg"},
	{0xE84A426F880AC95AULL, "vxdr_0_deeproads_cwr_0_0.vlm"},
	{0xE84A5ED893751A54ULL, "dae200cr_mother.utc"},
	{0xE84AC38A920F9214ULL, "mer110_jansen.cnv"},
	{0xE84AC9EFC340F85EULL, "780210891.ogg"},
	{0xE84AF1BB6EC27683ULL, "gui_lightblend_p0.vsh.fxo.dx11"},
	{0xE84BD568F0144B90ULL, "65392263.ogg"},
	{0xE84BF690ABDB08CFULL, "366913025.ogg"},
	{0xE84C0550339953CCULL, "w_encounter2_camp_45_0.phy"},
	{0xE84C1B57D4C79F9AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxa.phy"},
	{0xE84C47D3801DD944ULL, "genip_node_f1.utp"},
	{0xE84CE165F0EFF06BULL, "art\\characters\\playercharacter\\elffemale\\boo_lgta.phy"},
	{0xE84D069F7165CC7CULL, "vxht_0_hightown_a2_9x8_0_0.vlm"},
	{0xE84D37B923048FCBULL, "6098064_m.fxe"},
	{0xE84D4A69BCDC2CF7ULL, "hlac_w_sundermount_-327676_0x.dds"},
	{0xE84D699B158F1FFEULL, "6034327_f.fxe"},
	{0xE84D73C93C9B3250ULL, "6051489_f.fxe"},
	{0xE84D7A395AD88E37ULL, "6005754_m.fxe"},
	{0xE84D975BED2E15DDULL, "gen_im_fla_anders_02.uti"},
	{0xE84E683431C9992DULL, "vxht_0_hightown_n_a6t_0_0.vlm"},
	{0xE84EA58A92134854ULL, "mer110_jansen.bnk"},
	{0xE84F89F75FB748F7ULL, "892267483.ogg"},
	{0xE84FB778186DB74EULL, "ht_0_hightown_a2_35_0_ll.mmh"},
	{0xE84FCBB9659AD490ULL, "vxw_bonepitmining_a2_1rc_0_0.vlm"},
	{0xE850303BB3D746F7ULL, "hlac_lt_gallowstemplar_n_7198_5z.dds"},
	{0xE8506BF9346ACB23ULL, "vxht_brothel_n_5kv_0_0.vlm"},
	{0xE8507B370853C16AULL, "vxw_sundermount_30l_0_0.vlm"},
	{0xE851086B67CBB317ULL, "hirefollower.ncs"},
	{0xE8512FC61D2F29AAULL, "200148326.ogg"},
	{0xE85138E7F1AAC059ULL, "6202733_m.fxe"},
	{0xE8515AC28548CC25ULL, "hla_ht_0_hightown_65538_1y.dds"},
	{0xE8517AC4AE3F5D25ULL, "508604648.ogg"},
	{0xE851B0440C780306ULL, "186610473.ogg"},
	{0xE851B8E49DF4863AULL, "6121048_m.fxe"},
	{0xE85259AD36B481E7ULL, "832796158.ogg"},
	{0xE8527057AADBB4D1ULL, "hlac_dr_primevaltomb_659_2y.dds"},
	{0xE852B4080A990E8FULL, "vxlt_warehouse_n_92_0_0.vlm"},
	{0xE852DC50339F2CD4ULL, "w_encounter2_camp_45_0.rml"},
	{0xE85357D61EE38C86ULL, "mrl351ar_demon_cave.ncs"},
	{0xE85363992F4F9D6DULL, "vxca_cavern1_haunted_ec1_0_0.vlm"},
	{0xE85390B776BAC055ULL, "718227048.ogg"},
	{0xE8539E5BB7F5538BULL, "6174116_f.fxe"},
	{0xE85439FE985331DBULL, "6109205_m.fxe"},
	{0xE8549053E8D6A8ABULL, "6107009_m.fxe"},
	{0xE854909DAEE5ACB5ULL, "hlac_w_bonepitmining_a2_65538_0z.dds"},
	{0xE854DB425B4BF43CULL, "vxdr_primevaltomb_7ls_0_0.vlm"},
	{0xE8551E2BF85E7EF4ULL, "proj_dust.mao"},
	{0xE85583C7028F201EULL, "mag370_meredith6040163.cl"},
	{0xE855D592A8CB4896ULL, "6040575_f.fxe"},
	{0xE8560C8F5A05C15BULL, "vxw_encounter2_camp_47_0_0.vlm"},
	{0xE856D9A22F433750ULL, "art\\characters\\playercharacter\\elffemale\\textures\\bdy_arm_lgta_0d.dds"},
	{0xE85726EE3DF1A092ULL, "zz_area_signage10.utp"},
	{0xE857511A3F793F0FULL, "vxht_0_hightown_a3_ccq_0_0.vlm"},
	{0xE858B44C7A00E1DAULL, "vxdr_0_deeproads_cls_0_0.vlm"},
	{0xE8592C7AAB5D8C22ULL, "6217541_m.fxe"},
	{0xE859D1E137308AF6ULL, "w_sundermount_21_0.msh"},
	{0xE859DBE137309BF4ULL, "w_sundermount_21_0.mmh"},
	{0xE859DFE13730A207ULL, "w_sundermount_21_0.mao"},
	{0xE85A3C0A735AAF60ULL, "6106715_m.fxe"},
	{0xE85A4FD61EE99E7FULL, "mrl351ar_demon_cave.lst"},
	{0xE85AAF3BA696D440ULL, "genip_chest_normal1.utp"},
	{0xE85AF592DD37974FULL, "185627693.ogg"},
	{0xE85AFCEA06CD240EULL, "6094717_m.fxe"},
	{0xE85B1260208A3682ULL, "vxca_caverns1_mines_hvp_0_0.vlm"},
	{0xE85B3C1779AE3309ULL, "gen00fl_bethany6101270.cl"},
	{0xE85B663982C86660ULL, "6169781_m.fxe"},
	{0xE85B77B06C402341ULL, "hlac_w_bonepitmining_a2_196610_0z.dds"},
	{0xE85BD336D63C6CFEULL, "6153687_m.fxe"},
	{0xE85C2C4779DFE2BEULL, "6121205_f.fxe"},
	{0xE85C9020D45E5EECULL, "vxlt_0_lowtown_docks_bua_0_0.vlm"},
	{0xE85C989919BE37D2ULL, "6125750_m.fxe"},
	{0xE85D4277F875EB57ULL, "gen00fl_fenris6126525.cl"},
	{0xE85D6AF08C76F28EULL, "6105851_m.fxe"},
	{0xE85E37D178156816ULL, "921455424.ogg"},
	{0xE85E41A823F62E54ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkright.ani"},
	{0xE85E560289C78AF4ULL, "6208222_m.fxe"},
	{0xE85E74F9127A3030ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon2_n.dds"},
	{0xE85EF5385427721BULL, "vxlt_0_lowtown_docks_a3_n_8rm_0_0.vlm"},
	{0xE85FE1868E38226AULL, "348055744.ogg"},
	{0xE86020E768DC4210ULL, "wm_ico_cave_roll.dds"},
	{0xE86096A45273F8CEULL, "plt_demo_plot_flags.plo"},
	{0xE860A7597668B4FFULL, "vxlt_warehouse_n_5v_0_0.vlm"},
	{0xE860E148AE8A0166ULL, "494773213.ogg"},
	{0xE86190DADAED26B0ULL, "vxlt_undercitydungeon_xq_0_0.vlm"},
	{0xE8619D4CB513B7B5ULL, "vxdr_primevaltomb_77c_0_0.vlm"},
	{0xE861C1ED6719CC83ULL, "699400366.ogg"},
	{0xE862834A65EA2259ULL, "lt_0_lowtown_a2_27_0.rml"},
	{0xE862C8AE6AFFBE53ULL, "gen00fl_carver6108497.cl"},
	{0xE862D4223534E84CULL, "6072881_m.fxe"},
	{0xE863853649582CDAULL, "vxht_chantry_2hq_0_0.vlm"},
	{0xE8639D498967DBE4ULL, "vxht_brothel_n_4w8_0_0.vlm"},
	{0xE863BC3DD63607D0ULL, "hlac_lt_gallowstemplar_n_2388_0x.dds"},
	{0xE863CD222B2BA39EULL, "zz_approval_debug6003678.cl"},
	{0xE8647AEFE3DDBDABULL, "6111993_f.fxe"},
	{0xE864D9C60D4FDEF3ULL, "612208091.ogg"},
	{0xE864E49B57205B43ULL, "vrc221_varric6044223.cl"},
	{0xE8651C32F71CDEEFULL, "vxlt_0_lowtown_a2_d5u_0_0.vlm"},
	{0xE8653EA823FC47B6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkright.gad"},
	{0xE8655F8EEA505C23ULL, "hlac_lt_backalley_g_1_7y.dds"},
	{0xE86624599BF9E4C8ULL, "vxlt_undercitydungeon_1yo_0_0.vlm"},
	{0xE8665FB2E52DF83BULL, "gen_im_trs_qun_sta_nat_r0.uti"},
	{0xE867145F7D40EBC9ULL, "vxht_mansion2b_2ii_0_0.vlm"},
	{0xE868176F01734131ULL, "vxlt_foundry_n_uz_0_0.vlm"},
	{0xE8688AB813E93D97ULL, "313480831.ogg"},
	{0xE86896DD8BCEC242ULL, "vxw_bonepitmining_a2_1vu_0_0.vlm"},
	{0xE868CEC0C1E1F153ULL, "dae000_arishok6068489.cl"},
	{0xE869CA1797C11F21ULL, "vxca_caverns1_mines_ei6_0_1.vlm"},
	{0xE869D9057276B0F1ULL, "vxw_encounter2_camp_qo_0_1.vlm"},
	{0xE869E4BAC29BD2C7ULL, "vxca_caverns2_73s_0_0.vlm"},
	{0xE869F7B6B1C2B4A3ULL, "vxlt_0_lowtown_docks_b6c_0_0.vlm"},
	{0xE86B59A82400E115ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\walkright.evt"},
	{0xE86B6F78AD8CF4B1ULL, "vxlt_gallowstemplar_f_5qr_0_0.vlm"},
	{0xE86BA5043EE77CF5ULL, "mag210st_arianni.stg"},
	{0xE86C0B88FEC5388BULL, "994575125.ogg"},
	{0xE86C81AC3A0448FCULL, "6147460_m.fxe"},
	{0xE86C84E661288E53ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_hed_suca_0d.dds"},
	{0xE86C9312332F8284ULL, "6170585_m.fxe"},
	{0xE86CF0A5C7F66854ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\prp_ccliff02_da.dds"},
	{0xE86D851DDE0E3D5AULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.anb"},
	{0xE86E08DFCDD9BAF3ULL, "w_blightlands_45_0.rml"},
	{0xE86E2825DB24B461ULL, "6146875_m.fxe"},
	{0xE86E5C6FDA1944C9ULL, "vxht_bartrand_n_691_0_0.vlm"},
	{0xE86E8E7340A86EF7ULL, "vrc321ip_spooky7_obj_3.utp"},
	{0xE86EC95B065ABCC0ULL, "llac_lt_gallowstemplar_f_2394_1.dds"},
	{0xE86F1936685ECD7FULL, "vxlt_warehouse_n_8q_0_0.vlm"},
	{0xE86F46F9B2F1583EULL, "485689194.ogg"},
	{0xE86F5FCD5177897DULL, "vxlt_house_gh_0_0.vlm"},
	{0xE86F8F641CF86405ULL, "vxht_mansion2_2e1_0_0.vlm"},
	{0xE86FA55639F0B93FULL, "arm_rev.tnt"},
	{0xE8702BAE803B6863ULL, "792305670.ogg"},
	{0xE87046131C196E11ULL, "w_bonepitdestroyed_a3_232_0.mmh"},
	{0xE87052131C1982B2ULL, "w_bonepitdestroyed_a3_232_0.mao"},
	{0xE87060131C199ACFULL, "w_bonepitdestroyed_a3_232_0.msh"},
	{0xE870A93037CD7657ULL, "hlac_w_encounter1_2_0z.dds"},
	{0xE870B877BA859B92ULL, "vxht_mansion2_3bj_0_0.vlm"},
	{0xE870CC80AEA69302ULL, "follower_banter6136715.cl"},
	{0xE8714D66BA1C0ECDULL, "zz_chargengui.ncs"},
	{0xE87191E50F4FA33DULL, "vxlt_warehouse_n_ps_0_0.vlm"},
	{0xE87194770401F15EULL, "112348959.ogg"},
	{0xE871A42BE75EB244ULL, "vxlt_0_lowtown_a3_n_c_byl_0_0.vlm"},
	{0xE871BEFC3F38253CULL, "vxdr_primevaltomb_7bq_0_0.vlm"},
	{0xE871EF7FEE3ABEB3ULL, "mag311_emile_de_launcet6043868.cl"},
	{0xE872112D8578AEF6ULL, "295728283.ogg"},
	{0xE8722CFDB28970CFULL, "1018169120.ogg"},
	{0xE872841386A54FBBULL, "fen310_fenris6118228.cl"},
	{0xE872FAEA0136CCD8ULL, "438328819.ogg"},
	{0xE87322B66C97D098ULL, "ht_0_hightown_26_0.anb"},
	{0xE87326E8ABE05C49ULL, "mag101_grace6034082.cl"},
	{0xE873413CD267E007ULL, "6162963_m.fxe"},
	{0xE87345C96C2B616FULL, "vxht_mansion2_20x_0_0.vlm"},
	{0xE873790217BD1E37ULL, "vxht_0_hightown_a3_c8d_0_0.vlm"},
	{0xE8737FB2A88E30AFULL, "52342450.ogg"},
	{0xE874273CC2452A09ULL, "vxw_sundermount_30r_0_1.vlm"},
	{0xE87429EF1C1106A6ULL, "hightown_male_noble_gossip6174094.cl"},
	{0xE87487A45284885DULL, "plt_demo_plot_flags.nss"},
	{0xE87525BD58727AE0ULL, "288460251.ogg"},
	{0xE8756FBE192C74C5ULL, "llac_lt_0_lowtown_docks_a3_-131074_1.dds"},
	{0xE87613692D5AB1DBULL, "art\\vfx\\combat\\vfx_glb_nat_imp_ext_c.impa.evt"},
	{0xE87691009790C5B1ULL, "6126899_m.fxe"},
	{0xE87705A65C82E20EULL, "gen_im_cft_rec_bmb_02.uti"},
	{0xE8771ECEAB51602AULL, "zz_curio_test2.lst"},
	{0xE8777682400940E8ULL, "hlac_w_encounter2_-65537_0y.dds"},
	{0xE8781C78A05992BFULL, "gen00fl_isabela6099414.cl"},
	{0xE879AE731E4C27CEULL, "lt_0_lowtown_docks_a3_8_0.anb"},
	{0xE879FF6C50F2B6A3ULL, "6149453_f.fxe"},
	{0xE87AC2165C05DA8AULL, "vxht_0_hightown_a3_ca3_0_1.vlm"},
	{0xE87C01D9A9004FCBULL, "vxlt_warehouse_n_11e_0_0.vlm"},
	{0xE87C30ECC09A848DULL, "gen00fl_fenris6105595.cl"},
	{0xE87C62FB1F7DA5F4ULL, "vxlt_0_lowtown_docks_a3_n_ctk_0_0.vlm"},
	{0xE87D0C692D60C41CULL, "art\\vfx\\combat\\vfx_glb_nat_imp_ext_c.impa.gad"},
	{0xE87D221803B34883ULL, "demo_warrior.utc"},
	{0xE87D37D17D03DC48ULL, "vxht_mansion2b_1kp_0_1.vlm"},
	{0xE87D5CD93D830CA1ULL, "suc_cl.mao"},
	{0xE87DB4867AF57FC3ULL, "285890892.ogg"},
	{0xE87E28A144FAE09CULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsack_0.pwk"},
	{0xE87E52EC29355993ULL, "638387633.ogg"},
	{0xE87EA369B48DB0CCULL, "sound_movement_scuffing.gda"},
	{0xE87F1593989E71C1ULL, "6100276_f.fxe"},
	{0xE87F35C859E6AD09ULL, "6155835_m.fxe"},
	{0xE87F6898E874273FULL, "mag300st_orsino_bestserved.stg"},
	{0xE87F829EAD63F76EULL, "w_bonepit_84_0.mmh"},
	{0xE87F8E9EAD640B05ULL, "w_bonepit_84_0.mao"},
	{0xE87F9C9EAD64235CULL, "w_bonepit_84_0.msh"},
	{0xE87FB3C9DA89A4F8ULL, "qcr221_cut_heros_reward.cnv"},
	{0xE87FF3B45D32C9D1ULL, "976379042.ogg"},
	{0xE88003C20145BBBDULL, "vxlt_warehouse_n_wr_0_0.vlm"},
	{0xE8808664CCF9F16EULL, "hlac_w_sundermount_-327684_0x.dds"},
	{0xE880A4100F770A4BULL, "712949957.ogg"},
	{0xE880A6B2593FC7B9ULL, "543337556.ogg"},
	{0xE880C5D4D72EDE4BULL, "vrc321ip_spooky7_obj_12.utp"},
	{0xE880E40137806421ULL, "52936138.ogg"},
	{0xE880EDFBA246E9A4ULL, "51047045.ogg"},
	{0xE8819411FD6A2CC3ULL, "6156003_m.fxe"},
	{0xE882686BEA9377FBULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8.mmh"},
	{0xE88319394BE89D1AULL, "lt_0_lowtown_docks_a3_234_0.mmh"},
	{0xE8831D394BE8A451ULL, "lt_0_lowtown_docks_a3_234_0.mao"},
	{0xE8832B394BE8BC60ULL, "lt_0_lowtown_docks_a3_234_0.msh"},
	{0xE8832D0967448814ULL, "6087492_m.fxe"},
	{0xE88395C9DA8D5A00ULL, "qcr221_cut_heros_reward.bnk"},
	{0xE8840B692D66E11AULL, "art\\vfx\\combat\\vfx_glb_nat_imp_ext_c.impa.ani"},
	{0xE884539ABD442759ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha15a.mmh"},
	{0xE8845A5F7FB9A84AULL, "6051053_m.fxe"},
	{0xE8845D9ABD443847ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha15a.msh"},
	{0xE8853952CD327A41ULL, "611363322.ogg"},
	{0xE885801B8E5B7093ULL, "spi_02.dds"},
	{0xE885B3D533B663DEULL, "hlac_w_encounter3_-2_0z.dds"},
	{0xE885E4EE5F381B91ULL, "6171266_m.fxe"},
	{0xE8866A755FD06EB4ULL, "811781325.ogg"},
	{0xE886F6B87A96C060ULL, "20439062.ogg"},
	{0xE88701C95FE66D71ULL, "6113947_m.fxe"},
	{0xE88747C4D8D2813FULL, "vxlt_0_lowtown_docks_bun_0_0.vlm"},
	{0xE88777BD643360A9ULL, "6118348_m.fxe"},
	{0xE888593F8FA29BC1ULL, "6089752_m.fxe"},
	{0xE888714942E8CF23ULL, "6142553_f.fxe"},
	{0xE888A196D9CAA0E3ULL, "523508516.ogg"},
	{0xE8895FEDD6EBDDA8ULL, "gen00fl_carver6103691.cl"},
	{0xE889FBA3D4001490ULL, "dae220ip_chant_board_ht.utp"},
	{0xE88A0025F077383FULL, "vxlt_gallowscourtyard_7ac_0_0.vlm"},
	{0xE88A2FE817E17231ULL, "gen00fl_sebastian6197249.cl"},
	{0xE88A538C43914FC8ULL, "vxlt_backalley_g_1ji_0_2.vlm"},
	{0xE88A7907E7595CCDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyc.mmh"},
	{0xE88A7D07E759631EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyc.mao"},
	{0xE88A7FCF86D5C62AULL, "vxlt_gallowsprison_n_64e_0_0.vlm"},
	{0xE88A8B07E7597B53ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyc.msh"},
	{0xE88AD8020BA1FB1FULL, "vxht_mansion_n_6jq_0_0.vlm"},
	{0xE88B1CC0E0109699ULL, "blue.tnt"},
	{0xE88B320B1994D7B8ULL, "vxw_encounter2_camp_6z_0_0.vlm"},
	{0xE88B5F0C664EAA06ULL, "vxw_encounter2_camp_64_0_0.vlm"},
	{0xE88B65C38B2629DCULL, "1064043280.ogg"},
	{0xE88BD521C98E7E20ULL, "937183764.ogg"},
	{0xE88BDC5033CEF197ULL, "w_encounter2_camp_45_0.mmh"},
	{0xE88BE05033CEF8A4ULL, "w_encounter2_camp_45_0.mao"},
	{0xE88BF25033CF1709ULL, "w_encounter2_camp_45_0.msh"},
	{0xE88BF853E230547EULL, "vxlt_undercitydungeon_238_0_0.vlm"},
	{0xE88C250C9F795428ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_min_c.anb"},
	{0xE88C75A0B3AFE2F2ULL, "vxht_mansion2_1tn_0_0.vlm"},
	{0xE88D54FE10CC73C4ULL, "w_blightlands_72_0.msh"},
	{0xE88D66FE10CC922DULL, "w_blightlands_72_0.mao"},
	{0xE88D67CDC2F6CD69ULL, "vxw_bonepitmining_a2_1ph_0_0.vlm"},
	{0xE88D6AFE10CC9966ULL, "w_blightlands_72_0.mmh"},
	{0xE88D7B261EF1ABDCULL, "ser221_party_comments6209859.cl"},
	{0xE88D90C1FE4D27D5ULL, "170518159.ogg"},
	{0xE88DE22B5BE19114ULL, "hlac_lt_darktown_-1_58y.dds"},
	{0xE88E840B2B25BE30ULL, "6105457_m.fxe"},
	{0xE88EA2B62324563CULL, "vxdr_primevaltomb_7xw_0_0.vlm"},
	{0xE88EAB36014FF385ULL, "834530880.ogg"},
	{0xE88EADC772478E72ULL, "ser120_ghyslain6119554.cl"},
	{0xE88EE79D8CE29D13ULL, "75083930.ogg"},
	{0xE88EF53AE889CA5CULL, "lt_0_lowtown_docks_n_125_0.mao"},
	{0xE88F013AE889DEFFULL, "lt_0_lowtown_docks_n_125_0.mmh"},
	{0xE88F073AE889E8A1ULL, "lt_0_lowtown_docks_n_125_0.msh"},
	{0xE88F248F8990A9B4ULL, "6093585_m.fxe"},
	{0xE88FC0DC6EA5B160ULL, "t3_ogr_h02.tnt"},
	{0xE88FD5E81C6570E1ULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_stonetree_1.mao"},
	{0xE88FDAE5EAEACD22ULL, "vxca_caverns1_jle_0_0.vlm"},
	{0xE88FE92DFD4BDB78ULL, "568632149.ogg"},
	{0xE88FED1F35A9FEF3ULL, "hlac_lt_0_lowtown_65538_5x.dds"},
	{0xE8902D826D8423C3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basicnb.mao"},
	{0xE89031826D842A10ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basicnb.mmh"},
	{0xE8903F826D84422AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basicnb.msh"},
	{0xE8905CBBFBA94F37ULL, "vxht_0_hightown_a3_c6f_0_0.vlm"},
	{0xE890677E77347ED9ULL, "vxht_chantry_n_357_0_0.vlm"},
	{0xE890C5A6CB9EC8DFULL, "hf_amb_tranquil.mor"},
	{0xE890FE5BC23AC1B9ULL, "556251856.ogg"},
	{0xE89127C070868B56ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_basic.mao"},
	{0xE89133C070869FF5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_basic.mmh"},
	{0xE89135C07086A36BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\3pp_basic.msh"},
	{0xE891C03DA8DA8E0DULL, "vxca_caverns1_mines_gko_0_0.vlm"},
	{0xE891F35BDA524D65ULL, "vxdr_0_deeproads_cvv_0_0.vlm"},
	{0xE891F53624C2BA70ULL, "177948585.ogg"},
	{0xE89215AE4ACD853FULL, "vxht_keep_5k2_0_0.vlm"},
	{0xE892342E4F043747ULL, "6054543_m.fxe"},
	{0xE8926F2980D1C579ULL, "art\\characters\\weapons\\maces\\w_mce_mc103a.msh"},
	{0xE892792980D1D667ULL, "art\\characters\\weapons\\maces\\w_mce_mc103a.mmh"},
	{0xE892E2EFAF5BB772ULL, "mag100_karras6035317.cl"},
	{0xE8933466E36B4C6CULL, "73529543.ogg"},
	{0xE8937E589C24E571ULL, "646236749.ogg"},
	{0xE893C114CB8B27EDULL, "mag101_grace6034198.cl"},
	{0xE893CACE999AB36CULL, "6120459_f.fxe"},
	{0xE893CCD52CE6256EULL, "mag110_sailor.cnv"},
	{0xE894289EE4CE70E4ULL, "501208514.ogg"},
	{0xE8943D4F26FBB6CAULL, "gen00fl_isabela6102788.cl"},
	{0xE89456CB512E3411ULL, "6174166_m.fxe"},
	{0xE89520CD2BF95466ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyelow.ani"},
	{0xE89582AC8FCAFCA2ULL, "vxw_encounter2_camp_pf_0_0.vlm"},
	{0xE895881DDE2F940DULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_l_c.mmh"},
	{0xE895E99A56297929ULL, "6117248_m.fxe"},
	{0xE895FE5CC00B00E6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wrba_0s.dds"},
	{0xE8960DE3FDB8F016ULL, "hf_shw_mer_sympathizer.mor"},
	{0xE896AE22933F40E2ULL, "vxlt_0_lowtown_a2_n_c_exk_0_0.vlm"},
	{0xE896B2EF22E78179ULL, "plt_cod_ltr_junk_enlarge.nss"},
	{0xE896B2F9409FD067ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_act_c.impa.gad"},
	{0xE896B96F61783695ULL, "6055630_m.fxe"},
	{0xE896C50F836F3E31ULL, "vxlt_gallowstemplar_f_5bw_0_0.vlm"},
	{0xE896D306BCECB82BULL, "633820948.ogg"},
	{0xE89759AE96CA6AF6ULL, "gen00fl_anders6119706.cl"},
	{0xE89761FE10D4DE31ULL, "w_blightlands_72_0.phy"},
	{0xE897EED52CEA47DEULL, "mag110_sailor.bnk"},
	{0xE89862EA195103D1ULL, "6119637_m.fxe"},
	{0xE8986317668C2322ULL, "dialog_pressed.cur"},
	{0xE8988EB67F259E75ULL, "hlac_lt_0_lowtown_a2_n_-131072_6x.dds"},
	{0xE898BFE764C3DFEBULL, "6072842_m.fxe"},
	{0xE89946B50369D6F4ULL, "vxlt_0_lowtown_docks_a3_n_czi_0_0.vlm"},
	{0xE89949DEFD98139FULL, "vxht_mansion_n_6mp_0_0.vlm"},
	{0xE899851ABFAF38E7ULL, "6033565_f.fxe"},
	{0xE899D0E95844C286ULL, "vxw_encounter2_camp_15_0_0.vlm"},
	{0xE89A02681BC31CEAULL, "vxlt_0_lowtown_docks_a3_d23_0_0.vlm"},
	{0xE89A032C19AFC354ULL, "783830443.ogg"},
	{0xE89AA160370A3388ULL, "6209914_f.fxe"},
	{0xE89AD5793F35745AULL, "vxlt_gallowscourtyard_n_698_0_0.vlm"},
	{0xE89B37D7C0A40800ULL, "qun211cr_elf_warrior6072380.cl"},
	{0xE89B4BE4E5DD8AC0ULL, "hlac_ca_caverns1_mines_24424_10y.dds"},
	{0xE89B7207E767C38AULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_gurneyc.phy"},
	{0xE89BCCBDABFCB337ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_water_n.dds"},
	{0xE89BDDEC33982855ULL, "6002175.ogg"},
	{0xE89BE1A263CF26EEULL, "6103242_m.fxe"},
	{0xE89C721FE477E52DULL, "llac_lt_0_lowtown_a3_n_c_-1_4.dds"},
	{0xE89CA7F940A42974ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_act_c.impa.evt"},
	{0xE89CB4822F345990ULL, "gen00fl_merrill6108528.cl"},
	{0xE89CD1DD6206E442ULL, "cth_nb3.tnt"},
	{0xE89D1DCEAB722D8DULL, "zz_curio_test2.are"},
	{0xE89D4CAA11308EDAULL, "vxht_chantry_2tm_0_0.vlm"},
	{0xE89D729A737B03FCULL, "vxlt_backalley_n_1zr_0_0.vlm"},
	{0xE89D99EF22ED7552ULL, "plt_cod_ltr_junk_enlarge.plo"},
	{0xE89DA4B0619A87CCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12ma.msh"},
	{0xE89DAAB0619A91BEULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12ma.mmh"},
	{0xE89DB6B0619AA655ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12ma.mao"},
	{0xE89E42EB49121FACULL, "gen_im_trs_trs_two_2hs_r0.uti"},
	{0xE89EABA4AAEE4C0AULL, "696521825.ogg"},
	{0xE89F53F4F7A4F3B4ULL, "6076531_m.fxe"},
	{0xE89F9880641FB7D0ULL, "vxca_caverns1_mines_i34_0_0.vlm"},
	{0xE8A01E5F6593DF76ULL, "vxht_0_hightown_a3_9ah_0_0.vlm"},
	{0xE8A03ADFC5E1E44BULL, "follower_banter6172709.cl"},
	{0xE8A0CE6EEDDB9192ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0wi_basic_1.mao"},
	{0xE8A0FFAC55E74985ULL, "lt_0_lowtown_a2_n_38_0_lc.mmh"},
	{0xE8A128F5C8855D1EULL, "aip_circlehighmage_base.gda"},
	{0xE8A15E5C28730F9FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v2.gad"},
	{0xE8A32E4FC14CDC6CULL, "vxw_bonepitdestroyed_a3_1vi_0_0.vlm"},
	{0xE8A355D0EED5EC67ULL, "hlac_lt_0_lowtown_n_65538_5y.dds"},
	{0xE8A37751BEC2D364ULL, "6188708_m.fxe"},
	{0xE8A38F78ADE02891ULL, "6031979_m.fxe"},
	{0xE8A3F48981C11522ULL, "vxw_encounter2_camp_kg_0_0.vlm"},
	{0xE8A4486491A0A2AAULL, "vxht_0_hightown_a3_bl0_0_0.vlm"},
	{0xE8A4D0BAF5A34A21ULL, "vxht_mansion_n_53z_0_1.vlm"},
	{0xE8A4DC3B8E94CBDBULL, "553864703.ogg"},
	{0xE8A52F933F9A838EULL, "dae000_meredith6092602.cl"},
	{0xE8A534222CD12BCCULL, "zz_vault_debug6200373.cl"},
	{0xE8A5597036E74E85ULL, "1020954484.ogg"},
	{0xE8A57CD6F957882FULL, "709721733.ogg"},
	{0xE8A615D6ABDAA25BULL, "gen00fl_fenris6112070.cl"},
	{0xE8A65A632FD1CFABULL, "35996805.ogg"},
	{0xE8A6808A8F61E87DULL, "6169919_m.fxe"},
	{0xE8A6EF13D365963AULL, "w_dgr_dg06a_0s.dds"},
	{0xE8A7893C6D492290ULL, "6163488_m.fxe"},
	{0xE8A7B912214E6A90ULL, "781551656.ogg"},
	{0xE8A81457D515F2C3ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxa.msh"},
	{0xE8A82257D5160A1DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxa.mmh"},
	{0xE8A82657D516112EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_woodmeetxa.mao"},
	{0xE8A8535C28791B8CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v2.evt"},
	{0xE8A8DC49CD28E893ULL, "921931204.ogg"},
	{0xE8A907140D3FCDC3ULL, "vxlt_0_lowtown_docks_a3_cz5_0_0.vlm"},
	{0xE8A9609FB42277BBULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpa.msh"},
	{0xE8A97E9FB422AAA5ULL, "art\\levels\\races\\natural\\blightlands\\foliage\\tre_ashstumpa.mmh"},
	{0xE8A98F6AE781FFD2ULL, "666865294.ogg"},
	{0xE8A9D7DF24A88F16ULL, "vxht_keep_n_61e_0_0.vlm"},
	{0xE8AA09CD2C0B8960ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\additive\\eyelow.gad"},
	{0xE8AA66EA868FEDE8ULL, "hlac_ca_caverns2_5802_56x.dds"},
	{0xE8AA98278095C5E8ULL, "230429271.ogg"},
	{0xE8AA9FF940B058ADULL, "art\\vfx\\combat\\vfx_w_gn2_threat_act_c.impa.ani"},
	{0xE8AB0B26C3B4E046ULL, "abi_m_winters_grsp.bnk"},
	{0xE8AB31F16590F968ULL, "gen00fl_aveline6132836.cl"},
	{0xE8AB36229B9FB0E5ULL, "vxht_chantry_2kw_0_0.vlm"},
	{0xE8AB5D520E2DBA2FULL, "6117974_m.fxe"},
	{0xE8AB89D3DE4AFCF4ULL, "327868989.ogg"},
	{0xE8ABA2BD1E5E0BDFULL, "vxlt_darktown_6ho_0_0.vlm"},
	{0xE8ABD4AAAFD60E90ULL, "pro000_colonel_ewald6040581.cl"},
	{0xE8ABE524FF7FD0B2ULL, "vxw_encounter2_d2_0_0.vlm"},
	{0xE8ABFA451A69C313ULL, "6198672_m.fxe"},
	{0xE8AC06FB417B58CDULL, "w_woundedcoast_55_0.mmh"},
	{0xE8AC0AFB417B5F1EULL, "w_woundedcoast_55_0.mao"},
	{0xE8AC18FB417B7753ULL, "w_woundedcoast_55_0.msh"},
	{0xE8AC6957B7802C2BULL, "llac_ht_0_hightown_n_-131073_1.dds"},
	{0xE8ACB1731E778632ULL, "lt_0_lowtown_docks_a3_8_0.rml"},
	{0xE8AD89CA032EB05FULL, "lgt360_mag_orlanna6155310.cl"},
	{0xE8ADCE1425101B48ULL, "273315142.ogg"},
	{0xE8AE22CD988718F3ULL, "hlac_lt_0_lowtown_a2_n_c_1_0z.dds"},
	{0xE8AE4C58163D0099ULL, "6075046_f.fxe"},
	{0xE8AEDBE13778DBF7ULL, "w_sundermount_21_0.rml"},
	{0xE8AEF2296517649AULL, "437743138.ogg"},
	{0xE8AF23C92A807505ULL, "dre000_dougal6129744.cl"},
	{0xE8B043B817947FFFULL, "gen00fl_aveline6115438.cl"},
	{0xE8B10CDC9750C406ULL, "vxca_caverns1_j56_0_0.vlm"},
	{0xE8B10D6C6DA0F51BULL, "w_encounter3_82_0.mmh"},
	{0xE8B1196C6DA109B8ULL, "w_encounter3_82_0.mao"},
	{0xE8B12B6C6DA12805ULL, "w_encounter3_82_0.msh"},
	{0xE8B1546FBE6C3B5BULL, "6044609_m.fxe"},
	{0xE8B1FB78D1E9D725ULL, "art\\vfx\\combat\\vfx_g_spear_att_front_c.anb"},
	{0xE8B29D3575A105FFULL, "follower_banter6172903.cl"},
	{0xE8B2BA418396BB2AULL, "vxht_0_hightown_a3_bk1_0_0.vlm"},
	{0xE8B391DC772B71E2ULL, "vxlt_0_lowtown_a2_n_c_eru_0_0.vlm"},
	{0xE8B3E0208733F338ULL, "one300_nuncio6084356.cl"},
	{0xE8B3E1ADB277FA73ULL, "vxw_bonepitmining_a2_7h_0_0.vlm"},
	{0xE8B3E442936B9152ULL, "400326609.ogg"},
	{0xE8B44124B4C1228DULL, "6114076_m.fxe"},
	{0xE8B462B3CF53C92CULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.dura.ani"},
	{0xE8B5240C9F9C57C3ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_min_c.mmh"},
	{0xE8B53074A08F7D31ULL, "hlac_ht_mansion2b_818_22y.dds"},
	{0xE8B534BDF372520DULL, "plt_lgt000pt_ultra_4_01.nss"},
	{0xE8B546A3FD3D34D1ULL, "6113037_m.fxe"},
	{0xE8B5965E104A4B13ULL, "dae000ip_ambient_sitdrink_s.utp"},
	{0xE8B5B036EF280BCDULL, "6113009_f.fxe"},
	{0xE8B5CDD9667F480DULL, "6129982_m.fxe"},
	{0xE8B606FB4183AD46ULL, "w_woundedcoast_55_0.rml"},
	{0xE8B63B5C288530E5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_defense_mode_attk_v2.ani"},
	{0xE8B66834E75ED07AULL, "vxw_blightlands_bc_0_0.vlm"},
	{0xE8B6928D364ABB79ULL, "vxlt_gallowstemplar_n_5jo_0_0.vlm"},
	{0xE8B6A3C16C59F23FULL, "6173129_m.fxe"},
	{0xE8B747DA427077B0ULL, "6030472_m.fxe"},
	{0xE8B81F8393CE333FULL, "vxht_mansion2b_a1_0_0.vlm"},
	{0xE8B898A16D36F09DULL, "zz_test_restrict2.ncs"},
	{0xE8B97F7CDE99FBFDULL, "ave121ip_captain_door.utp"},
	{0xE8B985149F3301CAULL, "6051641_f.fxe"},
	{0xE8B9A7FF8D95C965ULL, "vxht_chantry_2dv_0_0.vlm"},
	{0xE8B9B70536FA4E00ULL, "6088116_m.fxe"},
	{0xE8B9CCE295F81AB2ULL, "vxht_0_hightown_a2_auu_0_0.vlm"},
	{0xE8BA5701F175E932ULL, "vxw_encounter2_o3_0_0.vlm"},
	{0xE8BB20BB7E44A750ULL, "vxlt_0_lowtown_docks_n_bgh_0_0.vlm"},
	{0xE8BB5668E059CF64ULL, "6061395_m.fxe"},
	{0xE8BB6FB3CF59FE6EULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.dura.gad"},
	{0xE8BBB45D5C0B1296ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1.phy"},
	{0xE8BBBD9841600C1EULL, "vxht_chantry_2c0_0_0.vlm"},
	{0xE8BC11B6050DF59FULL, "plt_cod_hst_mag_elfroot.plo"},
	{0xE8BC3D0BAEEB5D75ULL, "6075666_f.fxe"},
	{0xE8BD48F2B03496E5ULL, "vxdr_0_deeproads_ces_0_0.vlm"},
	{0xE8BDF9F23914C12CULL, "545285063.ogg"},
	{0xE8BE187B8B6AA5B7ULL, "hlac_lt_undercityencounter_-1_38z.dds"},
	{0xE8BE45137F4A5183ULL, "dre000_vampire6034325.cl"},
	{0xE8BE622980F6E99CULL, "art\\characters\\weapons\\maces\\w_mce_mc103a.phy"},
	{0xE8BE6C2B23E30463ULL, "art\\characters\\creatures\\dragon\\animation\\dazed.ani"},
	{0xE8BEA86FEFD70A34ULL, "60927145.ogg"},
	{0xE8BEE9E5D825E64EULL, "w_bonepitdestroyed_a3_163_0.msh"},
	{0xE8BEF0F94FCAAC1DULL, "vxlt_gallowstemplar_n_43q_0_0.vlm"},
	{0xE8BEF7E5D825FD9FULL, "w_bonepitdestroyed_a3_163_0.mao"},
	{0xE8BF03E5D826123CULL, "w_bonepitdestroyed_a3_163_0.mmh"},
	{0xE8BFED246E1A9DBEULL, "gen_im_wep_rog_dua_lore_03.uti"},
	{0xE8C0250B5ECF3865ULL, "mer220_hubert6078296.cl"},
	{0xE8C0441CDF250271ULL, "6081979_f.fxe"},
	{0xE8C05B31C0899526ULL, "dre000_cut_sibling_dies6036329.cl"},
	{0xE8C09C0632E9F2EAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_female.gad"},
	{0xE8C0D84365AD4622ULL, "vxw_encounter2_camp_aa_0_0.vlm"},
	{0xE8C0E8751B8A6A6BULL, "893082584.ogg"},
	{0xE8C0F2C0E03DDE8CULL, "blue.dds"},
	{0xE8C12C1E758CD3AAULL, "vxht_0_hightown_a3_bf2_0_0.vlm"},
	{0xE8C17A2B9681A99CULL, "vxca_cavern1_haunted_j1y_0_0.vlm"},
	{0xE8C1D6E7C3E889C0ULL, "lowtown_fem_alienage1_sta6162537.cl"},
	{0xE8C26AB3CF60144DULL, "art\\vfx\\combat\\vfx_w_avl_avaline2_loop_c.dura.evt"},
	{0xE8C2C90BB2F511FCULL, "gen_im_trs_tmp_shd_met_r0.uti"},
	{0xE8C47F526133682DULL, "w_blightlands_244_0.msh"},
	{0xE8C48D5261338060ULL, "w_blightlands_244_0.mao"},
	{0xE8C49152613386B3ULL, "w_blightlands_244_0.mmh"},
	{0xE8C4F5118397408FULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_01weed.mao"},
	{0xE8C501118397552CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_01weed.mmh"},
	{0xE8C5071183975F1EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_01weed.msh"},
	{0xE8C5B35D5C1365A9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball_1.mmh"},
	{0xE8C62CDC8789ACE2ULL, "dae000_orsino6085468.cl"},
	{0xE8C63D147B8844F2ULL, "vxht_mansion2_1hj_0_0.vlm"},
	{0xE8C6422AEDEA6C6FULL, "vxlt_0_lowtown_docks_n_9r7_0_0.vlm"},
	{0xE8C743B17E767B1DULL, "vxht_keep_6ey_0_0.vlm"},
	{0xE8C77F0632EFDF81ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_female.evt"},
	{0xE8C7A10D45B20949ULL, "hlac_ca_caverns2_5802_40z.dds"},
	{0xE8C7BE816411905FULL, "6164837_m.fxe"},
	{0xE8C83BB42E43DC32ULL, "543545800.ogg"},
	{0xE8C83EBF87EE3332ULL, "vxht_0_hightown_a2_apv_0_0.vlm"},
	{0xE8C888F05D44C8B6ULL, "6118729_m.fxe"},
	{0xE8C8A94C31E78138ULL, "6087643_m.fxe"},
	{0xE8C8B946FD3C3ED3ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mase.phy"},
	{0xE8C8F950F46F9DBCULL, "6171951_m.fxe"},
	{0xE8C9896BEACFB230ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball8.phy"},
	{0xE8C99EDCFE95A4DAULL, "6116724_f.fxe"},
	{0xE8C9AEE9CA450BE9ULL, "art\\vfx\\combat\\vfx_r_var_grenade_aoe_c.impa.ani"},
	{0xE8C9DCCF8F35A764ULL, "tow_glow.mao"},
	{0xE8CA532EEEB53F57ULL, "965684173.ogg"},
	{0xE8CAEC689817849DULL, "qun200_party_comment6182719.cl"},
	{0xE8CAF4C2352EE306ULL, "6185496_f.fxe"},
	{0xE8CB451524F6A6FDULL, "llac_lt_0_lowtown_a2_n_c_196607_0.dds"},
	{0xE8CB775907FBD7A5ULL, "sst_ss_coterie_thug_m.gda"},
	{0xE8CC05549EBB598DULL, "zz_followers_banter6158121.cl"},
	{0xE8CC9CB04E81FE6AULL, "6031816_f.fxe"},
	{0xE8CCB5F0307D7E63ULL, "gen00fl_anders6045372.cl"},
	{0xE8CCC9B89CE7D8AEULL, "hm_carver_h3s1.mor"},
	{0xE8CD2D32370B4575ULL, "vxdr_0_deeproads_cub_0_1.vlm"},
	{0xE8CDE5F3C5C0FEBFULL, "vxw_bonepitdestroyed_a3_1vf_0_0.vlm"},
	{0xE8CE7610708FEE3BULL, "mag160_feynriel6020420.cl"},
	{0xE8CE77C1FAC62FE1ULL, "6018570_m.fxe"},
	{0xE8CE9EE3D7D95D48ULL, "gen00fl_sebastian6188581.cl"},
	{0xE8CEC417E9FA96F9ULL, "docks_male_laborer_mob6175393.cl"},
	{0xE8CF0248B12096F3ULL, "art\\vfx\\creatures\\fxc_ash_ww_p.impa.ani"},
	{0xE8CF543023B8B26BULL, "600299632.ogg"},
	{0xE8D01F764BBB8B6DULL, "art\\characters\\playercharacter\\humanfemale\\hf_bdy_flm_eyesm1.mao"},
	{0xE8D06BD0CF16F13BULL, "mcr371_hostile_fenris6082256.cl"},
	{0xE8D0C9E9CA4B586BULL, "art\\vfx\\combat\\vfx_r_var_grenade_aoe_c.impa.gad"},
	{0xE8D11F7DFA858756ULL, "dae000_foreshadow6179561.cl"},
	{0xE8D138826DBBCD2BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0pp_basicnb.phy"},
	{0xE8D18F43E5A73D19ULL, "293851571.ogg"},
	{0xE8D2232E91325C7CULL, "lt_0_lowtown_docks_n_8_0_ll.mmh"},
	{0xE8D2616B5068F198ULL, "789840381.ogg"},
	{0xE8D2832B23F3D44DULL, "art\\characters\\creatures\\dragon\\animation\\dazed.gad"},
	{0xE8D28F50C5F106A0ULL, "ser121_jethann6106561.cl"},
	{0xE8D292C87F209E21ULL, "688305896.ogg"},
	{0xE8D30BEA3635897BULL, "vxlt_0_lowtown_a3_n_c_elj_0_1.vlm"},
	{0xE8D30BFEBE8FD029ULL, "gen00fl_fenris6126602.cl"},
	{0xE8D31078F8AD25AFULL, "690458069.ogg"},
	{0xE8D394659022867CULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_lightbeam_htc.mmh"},
	{0xE8D41D6CAE7677EDULL, "473062652.ogg"},
	{0xE8D43CA34ACA0B2FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_forcecrush.gad"},
	{0xE8D4FE1B6B9147DEULL, "594891617.ogg"},
	{0xE8D5770632FC0FD0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\spells_and_dive_reaction_female.ani"},
	{0xE8D5CA8535B724CFULL, "vxlt_0_lowtown_docks_a3_n_d4k_0_0.vlm"},
	{0xE8D5E8F882165931ULL, "6167489_m.fxe"},
	{0xE8D61DC8D527DD12ULL, "vxlt_0_lowtown_a2_e1m_0_0.vlm"},
	{0xE8D68A89B10594D2ULL, "116053877.ogg"},
	{0xE8D68C868D2E8AD8ULL, "6152563_m.fxe"},
	{0xE8D6FACD4F3AE136ULL, "6117629_m.fxe"},
	{0xE8D71F34E27F8D08ULL, "lt_0_lowtown_42_0_ll.mmh"},
	{0xE8D77A9A15FB3AD7ULL, "vxlt_0_lowtown_docks_n_ctc_0_0.vlm"},
	{0xE8D7917476949F61ULL, "233040575.ogg"},
	{0xE8D7C6E9CA5171D0ULL, "art\\vfx\\combat\\vfx_r_var_grenade_aoe_c.impa.evt"},
	{0xE8D853E3E4628ADBULL, "zz_supercrit.ncs"},
	{0xE8D8807CDEB4ADBCULL, "ave121ip_captain_door.ncs"},
	{0xE8D88367BE41FC54ULL, "6147172_m.fxe"},
	{0xE8D8920C1A526B68ULL, "hlac_dr_0_deeproads_5841_2z.dds"},
	{0xE8D8B2C23FBA49BFULL, "llac_lt_0_lowtown_65538_1.dds"},
	{0xE8D991EBA56A4A57ULL, "lt_0_lowtown_docks_a3_16_0_ll.mmh"},
	{0xE8DA08601D5C3141ULL, "718304761.ogg"},
	{0xE8DA3C11054E8111ULL, "ca_cavern1_haunted_24425.rml"},
	{0xE8DA522E0C933E8AULL, "hlac_w_encounter2_camp_196611_0z.dds"},
	{0xE8DAABE0706FCE98ULL, "ave200_donnic6110642.cl"},
	{0xE8DAEFB5111FC270ULL, "6026246_m.fxe"},
	{0xE8DAFC78D20CDE66ULL, "art\\vfx\\combat\\vfx_g_spear_att_front_c.mmh"},
	{0xE8DB51A34AD04DDCULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_forcecrush.evt"},
	{0xE8DBD79326BA6E00ULL, "var_placeable.gda"},
	{0xE8DBFA48B12B132AULL, "art\\vfx\\creatures\\fxc_ash_ww_p.impa.evt"},
	{0xE8DC04111415F040ULL, "vxlt_0_lowtown_a2_gvt_0_0.vlm"},
	{0xE8DC559CA875411EULL, "6103308_f.fxe"},
	{0xE8DCD738352E52F6ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2_tool.msh"},
	{0xE8DCE138352E63F4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2_tool.mmh"},
	{0xE8DCE538352E6A07ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2_tool.mao"},
	{0xE8DE4D3DA9BC90EBULL, "vxlt_0_lowtown_a2_n_ep1_0_1.vlm"},
	{0xE8DEE0215A14FED6ULL, "6150040_m.fxe"},
	{0xE8DF36CF13254BA2ULL, "6063760_m.fxe"},
	{0xE8DF6B365A083CEEULL, "nat380_temmerin6067375.cl"},
	{0xE8E003D8C11E8C03ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_tmpa_d.dds"},
	{0xE8E04A770E137C8AULL, "vxlt_0_lowtown_a2_n_c_ey0_0_0.vlm"},
	{0xE8E093BC2EE4C50BULL, "hlac_ca_caverns2_5802_25y.dds"},
	{0xE8E0DBF4B15EE8DDULL, "zz_spy130_darktown_res_16055318.cl"},
	{0xE8E0F48D81BABF9BULL, "vxht_mansion2b_349_0_0.vlm"},
	{0xE8E10416AC507ECBULL, "6113619_m.fxe"},
	{0xE8E151FA71416041ULL, "hlac_lt_0_lowtown_docks_n_-65538_3y.dds"},
	{0xE8E15FD8EBD35FC5ULL, "vxw_blightlands_bh_0_0.vlm"},
	{0xE8E1715A4D8DDB5DULL, "hlac_lt_gallowscourtyard_n_0_0z.dds"},
	{0xE8E1DB51B200F468ULL, "one101_orders_followers6115894.cl"},
	{0xE8E2956DBE0AC29AULL, "art\\vfx\\environment\\vfx_lightbeam_p.mmh"},
	{0xE8E2A4971E27A896ULL, "272707722.ogg"},
	{0xE8E2D948B130F99DULL, "art\\vfx\\creatures\\fxc_ash_ww_p.impa.gad"},
	{0xE8E33BEDF83F043AULL, "394873439.ogg"},
	{0xE8E33D39C82A339BULL, "6108145_m.fxe"},
	{0xE8E38D841CBF90E9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_ctn180l.ani"},
	{0xE8E3DAC9368CFB8EULL, "one110cr_pryce.utc"},
	{0xE8E3E69632F58BBEULL, "llac_lt_0_lowtown_n_-65536_2.dds"},
	{0xE8E4276B6262AC1DULL, "vxht_keep_6kw_0_0.vlm"},
	{0xE8E441F1D02A84BAULL, "673825924.ogg"},
	{0xE8E46C1D1EF95417ULL, "lgt160pt_wk_sophia.ncs"},
	{0xE8E47B67562ED916ULL, "6075802_m.fxe"},
	{0xE8E48501FC3AAB31ULL, "mer320_hubert6053382.cl"},
	{0xE8E48C6B453E0CBFULL, "dae000_bartrand6030601.cl"},
	{0xE8E50234A9B4D0F5ULL, "mag110_samson6049740.cl"},
	{0xE8E5EC7707F15357ULL, "vxlt_0_lowtown_docks_n_cob_0_0.vlm"},
	{0xE8E62E9E52E45912ULL, "vxw_encounter3_ej_0_0.vlm"},
	{0xE8E636A9BD7927CEULL, "hlac_ht_mansion2b_818_44z.dds"},
	{0xE8E647B1B7ED6727ULL, "1463261.ogg"},
	{0xE8E64ABFC48EFE66ULL, "ico_mm_partyarrow.dds"},
	{0xE8E6641DFB6D522DULL, "6143634_m.fxe"},
	{0xE8E6870D70740D96ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12ma.mao"},
	{0xE8E6930D70742235ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12ma.mmh"},
	{0xE8E6950D707425ABULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9uw_tarp12ma.msh"},
	{0xE8E6B76BBDAADB35ULL, "l_lgt331ar_supply_depot.bnk"},
	{0xE8E6C7E1D3C8CD83ULL, "hlac_lt_0_lowtown_docks_-196611_2z.dds"},
	{0xE8E7549ABD98A646ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_har_ha15a.phy"},
	{0xE8E791F7F389EF06ULL, "gen_im_trs_bst_spd_gland_02.uti"},
	{0xE8E7A1B4190D9674ULL, "aip_kirkwallguardarch_base.gda"},
	{0xE8E7A561274BB17AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_loop.gad"},
	{0xE8E7A91EBCE029AFULL, "6077337_m.fxe"},
	{0xE8E7E6829DA92D4BULL, "vxlt_0_lowtown_a3_n_c_er5_0_0.vlm"},
	{0xE8E83027A802790AULL, "llac_ht_bartrand_n_3370_2.dds"},
	{0xE8E8381040EFF2AAULL, "w_bonepitdestroyed_a3_76_0.mao"},
	{0xE8E83C1040EFF9F9ULL, "w_bonepitdestroyed_a3_76_0.mmh"},
	{0xE8E8461040F00AE7ULL, "w_bonepitdestroyed_a3_76_0.msh"},
	{0xE8E8597314C18B1EULL, "6093742_f.fxe"},
	{0xE8E8739693ED2338ULL, "ave200_donnic6110516.cl"},
	{0xE8E8A6AEF4D26137ULL, "vxht_keep_4aq_0_0.vlm"},
	{0xE8E92C080C26CE83ULL, "6115792_m.fxe"},
	{0xE8E959A34ADC98F5ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_forcecrush.ani"},
	{0xE8E981801CBB7F3AULL, "default_miss02.mao"},
	{0xE8E9930F5AA84BD7ULL, "735767297.ogg"},
	{0xE8EA68650A611956ULL, "vxlt_darktown_6jp_0_0.vlm"},
	{0xE8EA75EE060C08C0ULL, "vxlt_0_lowtown_a2_gmu_0_0.vlm"},
	{0xE8EA82B134A5B137ULL, "652938743.ogg"},
	{0xE8EAA8841CC5DD6BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_ctn180l.gad"},
	{0xE8EAC2F4A5E22601ULL, "560622691.ogg"},
	{0xE8EAD332B7D94358ULL, "mer141ip_at_bone_pit_ledge.utp"},
	{0xE8EAF645E68B6D3AULL, "155380044.ogg"},
	{0xE8EB22EAD30D40FCULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_monique.gad"},
	{0xE8EB67A07B1B56ECULL, "vxw_bonepitdestroyed_a3_1ob_0_0.vlm"},
	{0xE8EBC31B984DE8A5ULL, "vxw_woundedcoast_1en_0_0.vlm"},
	{0xE8EBD59CCB7A9A93ULL, "vxlt_gallowscourtyard_n_76q_0_0.vlm"},
	{0xE8EC0D26C3EC7AECULL, "abi_m_winters_grsp.wbd"},
	{0xE8EC3592D503FA22ULL, "nat380_nathaniel6066315.cl"},
	{0xE8EC6EE0A0F35630ULL, "417533629.ogg"},
	{0xE8EE28AFDCD8DB92ULL, "vxlt_gallowstemplar_n_47f_0_0.vlm"},
	{0xE8EE5BD38F796CAEULL, "vxca_caverns1_mines_d60_0_0.vlm"},
	{0xE8EFF358ECCFDB17ULL, "vxlt_0_lowtown_gb7_0_0.vlm"},
	{0xE8F1D81D47883873ULL, "hlac_lt_darktown_-1_78x.dds"},
	{0xE8F1D8483F9EB5B4ULL, "gen00fl_isabela_init6175716.cl"},
	{0xE8F1E0D7A230B41FULL, "pro000_opening_pc6001198.cl"},
	{0xE8F1FA35FE91DE76ULL, "6173911_m.fxe"},
	{0xE8F20FEE28B0ACB5ULL, "vxdr_primevaltomb_7ei_0_0.vlm"},
	{0xE8F221EAD3135DFAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_monique.ani"},
	{0xE8F22F21B6B74E90ULL, "283111633.ogg"},
	{0xE8F272A09CA4DD3BULL, "docks_male_elven_worker_sta6176116.cl"},
	{0xE8F31D070251226DULL, "6125955_m.fxe"},
	{0xE8F32927B1DF2627ULL, "784119913.ogg"},
	{0xE8F36F73556E2B65ULL, "vxht_chantry_2pr_0_0.vlm"},
	{0xE8F3A62ADE3A9BCDULL, "gen00fl_fenris6126718.cl"},
	{0xE8F3D1F02D1BEC37ULL, "180450502.ogg"},
	{0xE8F41987699CBF3DULL, "657947548.ogg"},
	{0xE8F45D2B6BF8C243ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_reaction.evt"},
	{0xE8F55CF162E56557ULL, "6052088_m.fxe"},
	{0xE8F561E3C7180796ULL, "vxlt_undercitydungeon_173_0_0.vlm"},
	{0xE8F6240B0F5400B4ULL, "llac_ht_0_hightown_n_-131074_1.dds"},
	{0xE8F632DBB1433A7BULL, "gen00fl_player6207778.cl"},
	{0xE8F78700F0DDE0C9ULL, "vxht_mansion2b_27w_0_0.vlm"},
	{0xE8F7E808F4002325ULL, "mag121ip_idunna_door.utp"},
	{0xE8F82E6C6DDD3050ULL, "w_encounter3_82_0.phy"},
	{0xE8F85ADCB7D29FC1ULL, "vxlt_0_lowtown_docks_b69_0_0.vlm"},
	{0xE8F89DF6106B7886ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_004a_0n.dds"},
	{0xE8F8C018467738E7ULL, "6091687_m.fxe"},
	{0xE8F8E6D98E463FA8ULL, "t3_gsp_b02.tnt"},
	{0xE8F90C22A86B1A81ULL, "vxlt_hangedman_n_2xs_0_0.vlm"},
	{0xE8F93B8A9BA3483FULL, "vxw_bonepitdestroyed_a3_1ac_0_0.vlm"},
	{0xE8F99D9D24616126ULL, "6153089_m.fxe"},
	{0xE8F9AF1612DEB4E2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_eagleheadsml_sa.dds"},
	{0xE8FA1DE6D62CE84BULL, "6053911_m.fxe"},
	{0xE8FA4779BD70B313ULL, "vxlt_gallowscourtyard_n_79p_0_0.vlm"},
	{0xE8FADFCB41050A9DULL, "388831613.ogg"},
	{0xE8FAFC8B5631D768ULL, "6140234_m.fxe"},
	{0xE8FB362B6BFE9EC4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_reaction.gad"},
	{0xE8FB9A426D821C26ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_idle.ani"},
	{0xE8FBA9C2534C745FULL, "gen_im_trs_drk_junk1_02.uti"},
	{0xE8FBBD5058DF329CULL, "zz_vault_debug6203181.cl"},
	{0xE8FBC061275C8860ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\urgent_walk_loop.ani"},
	{0xE8FCBDBFF4C6270AULL, "art\\characters\\creatures\\mubari\\animation\\pause2relax.gad"},
	{0xE8FCE0A32BA01190ULL, "art\\characters\\playercharacter\\elfmale\\bdy_flsb_1.mao"},
	{0xE8FCFCE972C13891ULL, "6171437_m.fxe"},
	{0xE8FD2E30F1FFAD8AULL, "vxlt_0_lowtown_a2_n_c_es2_0_0.vlm"},
	{0xE8FD350716859197ULL, "6165320_m.fxe"},
	{0xE8FD6BB8997BAF37ULL, "art\\levels\\races\\design\\prp_glitterdust.mmh"},
	{0xE8FD6FB8997BB644ULL, "art\\levels\\races\\design\\prp_glitterdust.mao"},
	{0xE8FD72042F158CB4ULL, "6075168_f.fxe"},
	{0xE8FD81B8997BD4A9ULL, "art\\levels\\races\\design\\prp_glitterdust.msh"},
	{0xE8FD828A6899DDF0ULL, "711836383.ogg"},
	{0xE8FDB47ACAF70979ULL, "6093845_m.fxe"},
	{0xE8FE08613A87E633ULL, "hlac_lt_gallowscourtyard_n_-65537_0y.dds"},
	{0xE8FE8FB877BC7DA9ULL, "6118559_m.fxe"},
	{0xE8FF092CCF804C29ULL, "316166787.ogg"},
	{0xE8FF2090AA8E6D98ULL, "hlac_lt_0_lowtown_docks_a3_n_-1_2z.dds"},
	{0xE8FF62C87DCB1CEEULL, "6208577_f.fxe"},
	{0xE8FF8B64A2588676ULL, "w_woundedcoast_172_0.mao"},
	{0xE8FF8F3669480D24ULL, "hlac_lt_0_lowtown_a3_n_c_65535_3x.dds"},
	{0xE8FF9764A2589A15ULL, "w_woundedcoast_172_0.mmh"},
	{0xE8FF9964A2589D8BULL, "w_woundedcoast_172_0.msh"},
	{0xE8FF9EBCFBC968B7ULL, "gen_im_arm_cht_nobd_t5.uti"},
	{0xE8FFD7DED2C40C08ULL, "6112010_f.fxe"},
	{0xE8FFE708F6F6C757ULL, "273188989.ogg"},
	{0xE8FFED5326A45A46ULL, "vxca_caverns1_hz1_0_0.vlm"},
	{0xE8FFEE840940C698ULL, "6116084_m.fxe"},
	{0xE9000DA10B3DD19AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1b.phy"},
	{0xE9006C12F087F6F6ULL, "6176611_m.fxe"},
	{0xE900F75FAEF7B036ULL, "vxdr_0_deeproads_fi1_0_0.vlm"},
	{0xE90118700B2B0FB2ULL, "115077825.ogg"},
	{0xE901285928744B02ULL, "zz_mcr_debug6080858.cl"},
	{0xE9013F9F65DB7707ULL, "vxlt_gallowscourtyard_n_4wa_0_0.vlm"},
	{0xE901452B6C0323A2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_reaction.ani"},
	{0xE9015CA38044DFD4ULL, "zz_approval_debug6003074.cl"},
	{0xE901DBC936A5FB5CULL, "one110cr_pryce.ncs"},
	{0xE901FCFC9D324DA3ULL, "vxw_sundermount_1zq_0_0.vlm"},
	{0xE90206D5AD35FCB5ULL, "arl_coprsesarmy.gda"},
	{0xE902A449D610DBE2ULL, "hm_qun_hench2.mor"},
	{0xE902A486700EF605ULL, "255371642.ogg"},
	{0xE903A0BFF4CC14A1ULL, "art\\characters\\creatures\\mubari\\animation\\pause2relax.evt"},
	{0xE903A761821B605CULL, "zz_followers_banter6158067.cl"},
	{0xE903D97872B67325ULL, "6089420_m.fxe"},
	{0xE90438A031619337ULL, "mcr370_guardsman6087658.cl"},
	{0xE90481979E81BC3CULL, "follower_banter6136638.cl"},
	{0xE90497491185F64EULL, "vxht_mansion2_1pn_0_0.vlm"},
	{0xE9049D5BD7678F62ULL, "llac_lt_0_lowtown_n_65536_12.dds"},
	{0xE904CBFBF29E2A94ULL, "ts_raiders_02.gda"},
	{0xE904CED38114C877ULL, "335927804.ogg"},
	{0xE9051F4331442452ULL, "740820326.ogg"},
	{0xE905C9B3B5F5CFCDULL, "6132847_m.fxe"},
	{0xE905F813D24C3970ULL, "vxlt_0_lowtown_a2_gc7_0_0.vlm"},
	{0xE905F8DDE2D3F949ULL, "vxht_mansion2b_20v_0_0.vlm"},
	{0xE90602E859BB1956ULL, "6080510_m.fxe"},
	{0xE9063DF05BE71E8EULL, "812786310.ogg"},
	{0xE906994E7A4BCA50ULL, "rock_c_02.dds"},
	{0xE9073D66C62E9930ULL, "6141798_m.fxe"},
	{0xE9077380BC669F55ULL, "vxlt_warehouse_b8_0_0.vlm"},
	{0xE90790D99A77D2DCULL, "6037693_m.fxe"},
	{0xE9079711FBAF95F5ULL, "559500845.ogg"},
	{0xE907E0233256EE1FULL, "6173877_m.fxe"},
	{0xE9080C4848F2A225ULL, "6156470_m.fxe"},
	{0xE9089180AF08F27BULL, "6150275_f.fxe"},
	{0xE908946E259C0D5AULL, "t3_wpn_irn.tnt"},
	{0xE908997636ADE7C4ULL, "195799505.ogg"},
	{0xE908A6D57C3A19A5ULL, "vxw_woundedcoast_1kl_0_0.vlm"},
	{0xE908C939282EE047ULL, "vxca_caverns2_7zd_0_0.vlm"},
	{0xE909EBBC2E62144EULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar5_mcr370_playerrules.ani"},
	{0xE909F6B6D7E9AEBFULL, "mag360_thrask6053519.cl"},
	{0xE90A2EB23A5D0212ULL, "vxht_0_hightown_n_bya_0_0.vlm"},
	{0xE90A68A08B125028ULL, "vxca_caverns2_62k_0_1.vlm"},
	{0xE90A710B1EE019F1ULL, "6049213_m.fxe"},
	{0xE90AFC4862C9CC8DULL, "vxw_bonepitmining_a2_e0_0_0.vlm"},
	{0xE90B51D11F5D64DFULL, "hlac_ht_0_hightown_a2_-65538_0x.dds"},
	{0xE90BA388C7E3B722ULL, "ser270_emeric6187717.cl"},
	{0xE90C07C8CFDCC53BULL, "6016857_m.fxe"},
	{0xE90C9EFB2DBDEC7CULL, "6121213_m.fxe"},
	{0xE90CE3DF84DB66B8ULL, "6161121_m.fxe"},
	{0xE90CEAD4A2D3B3BEULL, "6154282_m.fxe"},
	{0xE90D019569B29629ULL, "6117459_m.fxe"},
	{0xE90D2BD088F43B8FULL, "hlac_w_encounter3_0_11x.dds"},
	{0xE90D5BACFC87DD53ULL, "vxht_bartrand_n_68t_0_0.vlm"},
	{0xE90D84C5B752D86CULL, "lgt310ip_baz_board.ncs"},
	{0xE90D9CAD1ADA80A7ULL, "vxlt_backalley_n_1fq_0_0.vlm"},
	{0xE90DA25C33ED011AULL, "llac_lt_0_lowtown_a2_n_196607_1.dds"},
	{0xE90DFB5C5F753186ULL, "vxht_0_hightown_n_a6k_0_0.vlm"},
	{0xE90E23BDF3BE5E1EULL, "plt_lgt000pt_ultra_4_01.plo"},
	{0xE90E35A2A565237AULL, "vxw_sundermount_2ot_0_0.vlm"},
	{0xE90E421AF83DE9AAULL, "w_bonepit_188_0.mao"},
	{0xE90E461AF83DF0F9ULL, "w_bonepit_188_0.mmh"},
	{0xE90E501AF83E01E7ULL, "w_bonepit_188_0.msh"},
	{0xE90E590DEFA380AEULL, "statecnt_door_default.gda"},
	{0xE90EB18F8E24DEA5ULL, "371912280.ogg"},
	{0xE90ECD3AD3A790A4ULL, "mcr370_negotiation6080271.cl"},
	{0xE90F83426D929E20ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_idle.gad"},
	{0xE90F9D4A30955C7FULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_fire01.dds"},
	{0xE90FE256AC05BEEAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\deathendpose.gad"},
	{0xE9108C2DA1D7C07FULL, "art\\vfx\\textures\\vfx_smoke_puffy_cs_add.dds"},
	{0xE91098BFF4D691F0ULL, "art\\characters\\creatures\\mubari\\animation\\pause2relax.ani"},
	{0xE911050DC1BE878EULL, "6206859_m.fxe"},
	{0xE91112AE81E65282ULL, "6122539_m.fxe"},
	{0xE9116A68480C072EULL, "hlac_lt_darktown_-1_51x.dds"},
	{0xE911939AF2E215FAULL, "6164205_m.fxe"},
	{0xE911B8BD457C7215ULL, "vxlt_0_lowtown_a2_gby_0_0.vlm"},
	{0xE9126B417724575FULL, "mag321_comtesse_de_launcet6062388.cl"},
	{0xE912772D04241A94ULL, "hlac_ca_caverns2_5802_60y.dds"},
	{0xE912898CD8306C8EULL, "gen00fl_fenris6108100.cl"},
	{0xE912974A3317E56AULL, "mag121_idunna6084883.cl"},
	{0xE912BA61107C8733ULL, "dre000_cut2d_ragstoriches6025100.cl"},
	{0xE912BBBFD97D907DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0ww_basic_2.mao"},
	{0xE912DD6BBDD056BFULL, "l_lgt331ar_supply_depot.wbd"},
	{0xE912E0916C81045EULL, "water_water_p1.psh.fxo.dx11.assembly"},
	{0xE9133E71A7AB0EFFULL, "6041563_m.fxe"},
	{0xE9134EF7E54AAED3ULL, "559670175.ogg"},
	{0xE91367170CFF5029ULL, "vxlt_0_lowtown_a2_n_c_gck_0_0.vlm"},
	{0xE9139D49B3109EEEULL, "and330tr_base.ncs"},
	{0xE9146714696CDB06ULL, "382593471.ogg"},
	{0xE91472FF7BDD222DULL, "1060121798.ogg"},
	{0xE914C31790941903ULL, "274169954.ogg"},
	{0xE915C110ADC5E840ULL, "gen00fl_sebastian6189003.cl"},
	{0xE915EFDC8C574B81ULL, "vxlt_hangedman_n_2bq_0_0.vlm"},
	{0xE9163404224E92CFULL, "mag101_grace6034387.cl"},
	{0xE916C4E940FEA330ULL, "gen00fl_anders6125444.cl"},
	{0xE916D68324D5E320ULL, "mag100_thrask6032050.cl"},
	{0xE916DC62BDDC7BD2ULL, "vxht_bartrand_n_68q_0_0.vlm"},
	{0xE916F27195D33E5DULL, "one100dg_nabil6048680.cl"},
	{0xE91727FBB1E9B5BBULL, "760483894.ogg"},
	{0xE917318568FEA22FULL, "6161067_m.fxe"},
	{0xE917B2717DED0603ULL, "218028637.ogg"},
	{0xE917CA8E453F0AACULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_05tree.msh"},
	{0xE917D08E453F149EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_05tree.mmh"},
	{0xE917DC8E453F2935ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_05tree.mao"},
	{0xE917E91E49FDFA7CULL, "vxca_cavern1_haunted_gj9_0_0.vlm"},
	{0xE918AF6283FB0493ULL, "vxca_caverns1_mines_jle_0_0.vlm"},
	{0xE918D444EAEF6944ULL, "hf_qmq_patronf2.mor"},
	{0xE919E342BCB36AD9ULL, "hlac_lt_0_lowtown_a2_n_-3_4y.dds"},
	{0xE919FC67C565F26BULL, "mag110_vincento6016310.cl"},
	{0xE91AD0F5379FB3DEULL, "mag110_arianni6016823.cl"},
	{0xE91B2700A5AB7058ULL, "300515819.ogg"},
	{0xE91B302B7CB0D328ULL, "704887165.ogg"},
	{0xE91BED97E0AC34A8ULL, "vxlt_gallowstemplar_6qg_0_0.vlm"},
	{0xE91BF8B017D1E6AFULL, "hlac_ht_0_hightown_a3_-65536_2y.dds"},
	{0xE91C6BCE4D38F0CEULL, "6053851_m.fxe"},
	{0xE91CA77F975B3BFAULL, "vxw_sundermount_2pu_0_0.vlm"},
	{0xE91CEBFACA946CDEULL, "vxlt_0_lowtown_ehr_0_0.vlm"},
	{0xE91D002FF8735F75ULL, "art\\vfx\\environment\\vfx_trap_radialblade_p.anb"},
	{0xE91D6584FE92F635ULL, "vxdr_primevaltomb_7tv_0_0.vlm"},
	{0xE91DAB812AFD8F87ULL, "4974968.ogg"},
	{0xE91DBF52A4879FA7ULL, "hlac_lt_0_lowtown_n_131071_4x.dds"},
	{0xE91DC77E8BC223BBULL, "6121561_f.fxe"},
	{0xE91DD0AB41998CCCULL, "6082984_f.fxe"},
	{0xE91E3C1B05C82D5EULL, "6077303_m.fxe"},
	{0xE91EBE8752217FF1ULL, "one200_gold_followers6117111.cl"},
	{0xE91EC4BC2E742D88ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\cs_templar5_mcr370_playerrules.gad"},
	{0xE91FC768B974C932ULL, "vxw_blightlands_67_0_0.vlm"},
	{0xE921305CDFCBA302ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ch_basic.msh"},
	{0xE921325CDFCBA6B8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ch_basic.mmh"},
	{0xE9213E5CDFCBBB5BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ch_basic.mao"},
	{0xE921A17A11E47ACDULL, "428051539.ogg"},
	{0xE921AD56558EFC7CULL, "sf_depth.psh.fxo"},
	{0xE921D4F2F6321F15ULL, "gen00fl_aveline6121989.cl"},
	{0xE9224134FC4BE8A4ULL, "vxlt_undercitydungeon_17q_0_0.vlm"},
	{0xE9226F0EEEBE5E3FULL, "vxdr_primevaltomb_8fa_0_0.vlm"},
	{0xE92292AE9193B894ULL, "vxht_mansion_n_5ml_0_0.vlm"},
	{0xE922BE99AC0897D9ULL, "gen_im_trs_dwv_junk4_01.uti"},
	{0xE923AEC74328C82FULL, "art\\vfx\\combat\\vfx_r_ibl_fli_act_c.impa.gad"},
	{0xE923C39C60F2CB98ULL, "one140cr_charade6098611.cl"},
	{0xE923E7434AF48C75ULL, "gen00fl_aveline6114800.cl"},
	{0xE9240F4FBC038C78ULL, "hlac_w_encounter2_4_0z.dds"},
	{0xE92451F0AF6119CFULL, "fen110_fenris6097044.cl"},
	{0xE9245C73C762E49FULL, "rock_re_up.dds"},
	{0xE92465A0BE521BF4ULL, "vxht_mansion2b_3ea_0_0.vlm"},
	{0xE924704098333571ULL, "6116868_f.fxe"},
	{0xE924BD56AC17DBD0ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\deathendpose.ani"},
	{0xE924E008CACE1D3BULL, "vxlt_hangedman_n_24u_0_0.vlm"},
	{0xE9258A8F60264AA5ULL, "vxw_woundedcoast_1qj_0_0.vlm"},
	{0xE925A901136D8C70ULL, "777326273.ogg"},
	{0xE9262E987503F31CULL, "528300165.ogg"},
	{0xE926916CF3B7C949ULL, "6218918_m.fxe"},
	{0xE926AD2E2472DB91ULL, "800787038.ogg"},
	{0xE927CF56E734541AULL, "vxlt_darktown_4lx_0_0.vlm"},
	{0xE92834DA0D4037D7ULL, "vxlt_0_lowtown_a2_n_gmq_0_0.vlm"},
	{0xE9284EA24FDD0ECCULL, "zz_vault_debug6203358.cl"},
	{0xE928C68161873739ULL, "art\\vfx\\creatures\\vfx_c_hd_roar_c.impa.ani"},
	{0xE928ED0319101320ULL, "6103680_m.fxe"},
	{0xE929834591D6B0C6ULL, "vxlt_0_lowtown_a3_n_c_ekg_0_0.vlm"},
	{0xE929B187DC428E6AULL, "dr_primevaltomb_9621_ll.mmh"},
	{0xE92A6A9D901A39FCULL, "dae000_mother6121004.cl"},
	{0xE92A6B8E055BA2A1ULL, "vxht_bartrand_n_67g_0_0.vlm"},
	{0xE92AC3C7432F0ADCULL, "art\\vfx\\combat\\vfx_r_ibl_fli_act_c.impa.evt"},
	{0xE92B0A59678F4EBFULL, "gen00fl_merrill6127614.cl"},
	{0xE92B395CDFD40743ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ch_basic.phy"},
	{0xE92B66655B63A4E3ULL, "1032250855.ogg"},
	{0xE92B766C96AE5403ULL, "hlac_w_blightlands_3_0y.dds"},
	{0xE92B7E98EDBCEF23ULL, "w_bonepitmining_a2_160_0.msh"},
	{0xE92B8838EB032001ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_noise.ani"},
	{0xE92B8C98EDBD077DULL, "w_bonepitmining_a2_160_0.mmh"},
	{0xE92B9098EDBD0D8EULL, "w_bonepitmining_a2_160_0.mao"},
	{0xE92BAB38BCAECABEULL, "llac_lt_gallowstemplar_n_2384_4.dds"},
	{0xE92BB246FD90B048ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mase.mmh"},
	{0xE92BD046FD90E372ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_mase.msh"},
	{0xE92BD761F0890EB5ULL, "vxdr_primevaltomb_7qu_0_0.vlm"},
	{0xE92BF7130348AA1EULL, "ser191_emeric6108792.cl"},
	{0xE92C1309918CC415ULL, "89486682.ogg"},
	{0xE92C1F404CC1508DULL, "w_bonepit_197_0.mmh"},
	{0xE92C23404CC157DEULL, "w_bonepit_197_0.mao"},
	{0xE92C31404CC16F13ULL, "w_bonepit_197_0.msh"},
	{0xE92C34C76725C419ULL, "vxdr_0_deeproads_dq5_0_1.vlm"},
	{0xE92C9B44017BF5D9ULL, "6171914_m.fxe"},
	{0xE92CB3FEAD88BB19ULL, "vxlt_warehouse_n_of_0_0.vlm"},
	{0xE92CE7985DB7AF9DULL, "35248180.ogg"},
	{0xE92CEDA370DD652AULL, "6037080_m.fxe"},
	{0xE92D2CF43D9E1A41ULL, "hlac_w_sundermount_-393213_0z.dds"},
	{0xE92DD0946849C73FULL, "257223534.ogg"},
	{0xE92E023DE117F33BULL, "vxlt_0_lowtown_a2_n_c_htg_0_0.vlm"},
	{0xE92E1BD791C90370ULL, "1030707079.ogg"},
	{0xE92F458E0F291F4AULL, "286491336.ogg"},
	{0xE92FA181618D177BULL, "art\\vfx\\creatures\\vfx_c_hd_roar_c.impa.gad"},
	{0xE92FECDFE7683FCEULL, "vxht_mansion2_1aa_0_0.vlm"},
	{0xE93010D1A3BAF2EDULL, "435603297.ogg"},
	{0xE930632F5960371FULL, "6067119_f.fxe"},
	{0xE930C23C478544B2ULL, "6177909_f.fxe"},
	{0xE9311EDA561EC83BULL, "art\\characters\\weapons\\boulder\\textures\\w_boulder_n.dds"},
	{0xE93128CF0CDC7DB1ULL, "dae011_corff_sta6142435.cl"},
	{0xE93172677DAEDE7DULL, "6013841_m.fxe"},
	{0xE9317EDEB2229DECULL, "vxdr_primevaltomb_8rb_0_0.vlm"},
	{0xE9319338EB079EA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_noise.gad"},
	{0xE931FDA9185469C3ULL, "6103381_f.fxe"},
	{0xE9327176EFBE070FULL, "911579586.ogg"},
	{0xE9329A31E999D9F8ULL, "zz_isa_debug6175472.cl"},
	{0xE932A064D25D368EULL, "art\\vfx\\combat\\vfx_r_gn2_speed_act_c.impa.evt"},
	{0xE932B5874E70AD50ULL, "ht_brothel_n_3370_ll.mmh"},
	{0xE932D36860343092ULL, "424602066.ogg"},
	{0xE932E09FCC3DCE38ULL, "vxca_caverns1_jfq_0_0.vlm"},
	{0xE932E655B5762718ULL, "78041170.ogg"},
	{0xE932F058B77AB55FULL, "gen00fl_anders6119873.cl"},
	{0xE9332140785ECAE9ULL, "6189671_f.fxe"},
	{0xE9338732F52B21B1ULL, "vxlt_gallowstemplar_n_5wp_0_0.vlm"},
	{0xE9339088BA9D2BDCULL, "dae000ip_crowd_em2_talk14.utp"},
	{0xE9345B3572B21405ULL, "vxht_0_hightown_n_ca7_0_0.vlm"},
	{0xE934B38B3B7B548EULL, "vxlt_undercityserialkiller_1d3_0_0.vlm"},
	{0xE934B4CCDE4C2787ULL, "132607086.ogg"},
	{0xE934D3FEDE37BA70ULL, "6188667_m.fxe"},
	{0xE934E5C2302F1511ULL, "qcr000_aveline6054371.cl"},
	{0xE934F270B367F91DULL, "6165443_m.fxe"},
	{0xE93508653B227DB6ULL, "hightown_male_noble1_sta6170799.cl"},
	{0xE93544EF4FC4F1C2ULL, "277090285.ogg"},
	{0xE9354D6E536CBEAFULL, "vxht_bartrand_n_5pc_0_0.vlm"},
	{0xE9357E23097F196CULL, "gen00fl_sebastian6177132.cl"},
	{0xE9358C473CA87E53ULL, "6032824_m.fxe"},
	{0xE935933E47D630FDULL, "dae000ip_crowd_ff7_talk1.utp"},
	{0xE935BE816191B460ULL, "art\\vfx\\creatures\\vfx_c_hd_roar_c.impa.evt"},
	{0xE935D46D93F54CEFULL, "vxht_0_hightown_cf7_0_0.vlm"},
	{0xE9361CED7DE5B27FULL, "6096458_f.fxe"},
	{0xE9364133D92A6C9AULL, "vxlt_darktown_4ky_0_0.vlm"},
	{0xE93787DAA5F48432ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_meda_0t.dds"},
	{0xE937E038357B0DEFULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat8x2_tool.phy"},
	{0xE937E0758A39A474ULL, "gen_im_trs_tmp_hvy_boo_r2.uti"},
	{0xE9386A0BAF764008ULL, "6099213_m.fxe"},
	{0xE9387038EB0D8188ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_overturn_noise.evt"},
	{0xE938B61459F02ABDULL, "hlac_lt_0_lowtown_a2_131070_3y.dds"},
	{0xE938CBC7433B55F5ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_act_c.impa.ani"},
	{0xE938E641F07EA47FULL, "164542412.ogg"},
	{0xE938F243F0BCCA27ULL, "vxlt_backalley_n_1wr_0_0.vlm"},
	{0xE939192D88A1E56EULL, "108186947.ogg"},
	{0xE9396F0FF62DF785ULL, "369095362.ogg"},
	{0xE9399764D2634639ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_act_c.impa.gad"},
	{0xE939D62414AB82CDULL, "vxht_0_hightown_ang_0_0.vlm"},
	{0xE93B25DB9F7ED399ULL, "vxlt_warehouse_n_re_0_0.vlm"},
	{0xE93B4A5AC17EB814ULL, "vxca_caverns1_mines_flz_0_0.vlm"},
	{0xE93B5B48E1F82D02ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.mmh"},
	{0xE93B7AE30DF90ACCULL, "6135249_m.fxe"},
	{0xE93CEFBCCCDBDF30ULL, "vxlt_0_lowtown_docks_a3_n_d1s_0_0.vlm"},
	{0xE93CFE87FAB0DABFULL, "pro000_merchant6051634.cl"},
	{0xE93D317D762D95DBULL, "152089534.ogg"},
	{0xE93DFA6064101D9EULL, "670850804.ogg"},
	{0xE93E507177642067ULL, "6067633_m.fxe"},
	{0xE93E69EF78BFA5F6ULL, "gen_im_trs_krk_drk_junk2_03.uti"},
	{0xE93E828756381040ULL, "hlac_lt_0_lowtown_a2_n_-262144_0z.dds"},
	{0xE93EA07A4BE3AEA4ULL, "555788877.ogg"},
	{0xE93F6A836A3D1933ULL, "dae300_petrice6098233.cl"},
	{0xE93FEE8291C1E0B1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur3_flemeth_real.gad"},
	{0xE940031E2AB198B5ULL, "92296106.ogg"},
	{0xE94015580E710D96ULL, "258609208.ogg"},
	{0xE940A864D2698187ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_act_c.impa.ani"},
	{0xE9411C692A10F51DULL, "71121539.ogg"},
	{0xE9417F1AD4AF9337ULL, "343588066.ogg"},
	{0xE94186CAD78C5742ULL, "mag210_arianni6021239.cl"},
	{0xE941C15B35D997B9ULL, "6087798_m.fxe"},
	{0xE941FC97C6B10144ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curver.mmh"},
	{0xE9420097C6B10797ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curver.mao"},
	{0xE9421297C6B126E6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4sc_curver.msh"},
	{0xE94258E6D4CE7DA8ULL, "hlac_w_encounter2_-65538_0x.dds"},
	{0xE9427FC9231C6352ULL, "dae300cr_solivitus.utc"},
	{0xE94286527EA22A09ULL, "567319077.ogg"},
	{0xE94308D7BEA84987ULL, "1045005228.ogg"},
	{0xE94337789D6085F4ULL, "340205060.ogg"},
	{0xE9435E90053FCACBULL, "follower_banter6148461.cl"},
	{0xE9439B1A80C485B7ULL, "476687681.ogg"},
	{0xE943E5FF109ADC82ULL, "513590926.ogg"},
	{0xE9444212868BA0F7ULL, "one100dg_vanard6020938.cl"},
	{0xE945370367034C07ULL, "hlac_lt_0_lowtown_docks_n_-131075_7y.dds"},
	{0xE94539C5C27B6C46ULL, "vxca_caverns1_k2n_0_0.vlm"},
	{0xE9456F4DC8DF0095ULL, "hlac_w_encounter2_camp_3_0x.dds"},
	{0xE946012FF89665B6ULL, "art\\vfx\\environment\\vfx_trap_radialblade_p.mmh"},
	{0xE946553E267E3D72ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_wesley.gad"},
	{0xE946879671A42D53ULL, "llac_w_encounter2_camp_196610_0.dds"},
	{0xE946ACCB40E62B8BULL, "599259563.ogg"},
	{0xE946CA32B82792F9ULL, "mer141ip_at_bone_pit_ledge.ncs"},
	{0xE946EF8291C8015FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_flemeth_real\\cs_hur3_flemeth_real.ani"},
	{0xE947838C04374735ULL, "858385714.ogg"},
	{0xE9483BC46F794738ULL, "dae000_meredith6166284.cl"},
	{0xE948493D3D711F0EULL, "144808937.ogg"},
	{0xE948B9FC19C89479ULL, "514127021.ogg"},
	{0xE948CCF3580764F6ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_templarbadge_sa.dds"},
	{0xE948F504EA78443EULL, "6115007_m.fxe"},
	{0xE94997B89174EC19ULL, "vxlt_warehouse_n_yd_0_0.vlm"},
	{0xE94A4D611B015A13ULL, "370891696.ogg"},
	{0xE94A79BA455032E6ULL, "vxdr_0_deeproads_f9q_0_0.vlm"},
	{0xE94A8953BB08B765ULL, "932293636.ogg"},
	{0xE94AE3477059BE3CULL, "gen00fl_anders6119384.cl"},
	{0xE94AF1FA5F5AA2E2ULL, "llac_lt_darktown_-1_39.dds"},
	{0xE94B7F907672C390ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_9z.dds"},
	{0xE94BDE5F8B264694ULL, "315105690.ogg"},
	{0xE94BF9ACBD192CF4ULL, "6146892_m.fxe"},
	{0xE94C4A7BC6C4EA2EULL, "6145187_m.fxe"},
	{0xE94CD2FB00A11798ULL, "6135070_player_letcarver.cut"},
	{0xE94D571EED314207ULL, "llac_lt_0_lowtown_65535_4.dds"},
	{0xE94D6BA2B66F21DCULL, "6093668_m.fxe"},
	{0xE94DED8CCD46EEEBULL, "318755620.ogg"},
	{0xE94E6874029F660EULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\mov_p.ani"},
	{0xE94F91BCAC1C180DULL, "llac_dr_dungeon_3217_2.dds"},
	{0xE9504565F5D6567BULL, "170128837.ogg"},
	{0xE95085DD5610D043ULL, "vxdr_primevaltomb_2x1_0_0.vlm"},
	{0xE950BE2254336C5EULL, "mag100_karras6034729.cl"},
	{0xE9521774A9E0447BULL, "6178533_m.fxe"},
	{0xE9522DA2123EE33CULL, "vxlt_warehouse_n_e9_0_0.vlm"},
	{0xE95279CF7070D637ULL, "hlac_lt_undercityserialkiller_-1_30x.dds"},
	{0xE9527B6E9E5F22DEULL, "6041110_f.fxe"},
	{0xE9536EB47716534AULL, "atl_logininfos_dxt5_01.dds"},
	{0xE9542CDB7F3AA706ULL, "character_character_p247.psh.fxo.dx11.assembly"},
	{0xE954367823B7B734ULL, "513030824.ogg"},
	{0xE9545DC1A57E001FULL, "6118298_m.fxe"},
	{0xE9554FA969D57B7DULL, "mcr370_cut_player_survives6171841.cl"},
	{0xE955EA729B8E14CCULL, "mag111im_danzig_letter.uti"},
	{0xE9560C5443DCDB6EULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_015a_0s.dds"},
	{0xE95614F4797C615DULL, "mcr310_confused_guard6077081.cl"},
	{0xE956B68C561E5777ULL, "vxlt_hangedman_n_24a_0_0.vlm"},
	{0xE956DBC0A5CF79C4ULL, "isa310pt_isabelas_trail.ncs"},
	{0xE9571C5954744AD7ULL, "art\\vfx\\models\\vfx_isa_pose3.msh"},
	{0xE957325954747049ULL, "art\\vfx\\models\\vfx_isa_pose3.mmh"},
	{0xE95767864A7E81A1ULL, "art\\characters\\creatures\\spider\\animation\\die_lp_v2.gad"},
	{0xE957727F5AFEEEE7ULL, "290768940.ogg"},
	{0xE9577F318DED696EULL, "vxdr_0_deeproads_evw_0_0.vlm"},
	{0xE957B044082BC805ULL, "vxlt_gallowstemplar_46u_0_0.vlm"},
	{0xE958EA3DC4E50660ULL, "6093551_f.fxe"},
	{0xE9596CFC83707729ULL, "zz_dae_debug6029173.cl"},
	{0xE959EB2DF7B18C3AULL, "6174796_m.fxe"},
	{0xE959F0858001FAA7ULL, "hlac_lt_undercitydungeon_-1_30y.dds"},
	{0xE95A6B89AF0F4574ULL, "6143792_m.fxe"},
	{0xE95AB05A339C6E5DULL, "hlac_lt_darktown_-1_71y.dds"},
	{0xE95AB2E0D6622A62ULL, "vxlt_hangedman_n_217_0_0.vlm"},
	{0xE95AF1F0BAEF5EB3ULL, "hlac_ca_caverns1_mines_24425_11x.dds"},
	{0xE95B403E26907518ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_avalineintro_real\\avalineintro_wesley.ani"},
	{0xE95B4C676B0A99C1ULL, "hlac_ht_mansion_n_2995_4x.dds"},
	{0xE95B607402A9E35FULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\mov_p.evt"},
	{0xE95C06A10B8C24C3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1b.msh"},
	{0xE95C14A10B8C3C1DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1b.mmh"},
	{0xE95C18A10B8C432EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_couldron1b.mao"},
	{0xE95C554E2C891F35ULL, "nat320_delilah6065910.cl"},
	{0xE95D01869F8DA834ULL, "dae010_beggar_m1.utc"},
	{0xE95D61546F9EE600ULL, "6170563_m.fxe"},
	{0xE95E404065832F8DULL, "llac_lt_0_lowtown_a3_n_c_-131071_2.dds"},
	{0xE95E68864A84A24FULL, "art\\characters\\creatures\\spider\\animation\\die_lp_v2.ani"},
	{0xE95E8BC271C7CFCFULL, "vxlt_gallowstemplar_f_455_0_0.vlm"},
	{0xE95EF7BA4806E8C3ULL, "vxdr_primevaltomb_2w0_0_0.vlm"},
	{0xE95F38C71EA0DA96ULL, "vxlt_0_lowtown_a3_n_c_hq0_0_0.vlm"},
	{0xE960A305294F082FULL, "vxht_bartrand_n_5al_0_0.vlm"},
	{0xE960B26F66C9964FULL, "art\\vfx\\combat\\vfx_w_def_def_mode_act_c.mmh"},
	{0xE960C22B2FCF08D4ULL, "hlac_lt_warehouse_65537_1x.dds"},
	{0xE96127D6BF0DFF38ULL, "mrl150_merrill6197585.cl"},
	{0xE961D852A9902EE3ULL, "6118639_f.fxe"},
	{0xE96222E960DAD46EULL, "vxlt_warehouse_184_0_0.vlm"},
	{0xE962356BB585C8FFULL, "224310256.ogg"},
	{0xE962417402AFCD48ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\mov_p.gad"},
	{0xE96288CDA90D871AULL, "hlac_w_sundermount_131069_0y.dds"},
	{0xE962C2737037555FULL, "vxlt_undercityencounter_1bq_0_1.vlm"},
	{0xE96349939D3D0390ULL, "293349981.ogg"},
	{0xE9635A48E2197DA9ULL, "art\\vfx\\combat\\vfx_c_mer_crack_burst_p.anb"},
	{0xE963EBEB735C197FULL, "760290225.ogg"},
	{0xE96447DAC69F13A7ULL, "vxlt_backalley_n_1xo_0_0.vlm"},
	{0xE964958B6A9FC338ULL, "243180244.ogg"},
	{0xE96529A304C54F0EULL, "6060721_m.fxe"},
	{0xE9657EA6E49F634CULL, "209045974.ogg"},
	{0xE9658013605EA577ULL, "gen_im_trs_dal_dua_dag_r2.uti"},
	{0xE96675B4FA735FF1ULL, "25931388.ogg"},
	{0xE9674BA7A6FC9306ULL, "combat_ws_r1_a4.gda"},
	{0xE967F0FF64CBDB5EULL, "6156550_m.fxe"},
	{0xE96813F3EEC2C5DCULL, "432069547.ogg"},
	{0xE96970DDD38A3208ULL, "ts_beasts_spiders_01.gda"},
	{0xE9697AC3C6261B8BULL, "mrl210_merrill6102150.cl"},
	{0xE969D6EA969905D5ULL, "w_bonepitdestroyed_a3_195_0.msh"},
	{0xE969D71BB3F3164AULL, "hlac_dr_primevaltomb_9621_0z.dds"},
	{0xE969D8EA9699094BULL, "w_bonepitdestroyed_a3_195_0.mmh"},
	{0xE969E4EA96991DE8ULL, "w_bonepitdestroyed_a3_195_0.mao"},
	{0xE96A6F57802607E1ULL, "6110418_m.fxe"},
	{0xE96A8E63ACE5D7A3ULL, "zz_arttest.are"},
	{0xE96A9EDB36410D2BULL, "one350cr_terath.utc"},
	{0xE96B5C604EFF1DBEULL, "hlac_ht_0_hightown_a2_1_0z.dds"},
	{0xE96BC3137CBFFE59ULL, "6083986_f.fxe"},
	{0xE96C5173C5908774ULL, "6164041_m.fxe"},
	{0xE96D1C9885209245ULL, "6153103_m.fxe"},
	{0xE96D28481B5099E3ULL, "gen00fl_anders6125799.cl"},
	{0xE96D3AA24162704DULL, "hlac_lt_0_lowtown_n_-262144_3x.dds"},
	{0xE96D741431296DBAULL, "6069578_f.fxe"},
	{0xE96DE1D0240133C0ULL, "76315172.ogg"},
	{0xE96EA849B35D5C8FULL, "and330tr_base.utt"},
	{0xE96EB8C263E9B936ULL, "w_bonepitdestroyed_a3_82_0.msh"},
	{0xE96EC2C263E9CA34ULL, "w_bonepitdestroyed_a3_82_0.mmh"},
	{0xE96EC6C263E9D147ULL, "w_bonepitdestroyed_a3_82_0.mao"},
	{0xE96ED9525A8E8E08ULL, "dae000cr_grand_cleric6147674.cl"},
	{0xE96EF2224A8811C9ULL, "6143177_m.fxe"},
	{0xE96EF59EC8356584ULL, "6056810_m.fxe"},
	{0xE96F8889196E6544ULL, "6109101_m.fxe"},
	{0xE96FAB15D8E1D16EULL, "974007609.ogg"},
	{0xE970A5A414EE1E30ULL, "llac_dr_dungeon_3216_2.dds"},
	{0xE970B3559AB1F230ULL, "gen00fl_anders6121202.cl"},
	{0xE971417B896A311FULL, "6112098_m.fxe"},
	{0xE971471AF8926FE6ULL, "w_bonepit_188_0.phy"},
	{0xE9722CCCB90FCFF7ULL, "vxht_bartrand_n_4ck_0_0.vlm"},
	{0xE9723255B05CB7CEULL, "33368874.ogg"},
	{0xE9724DC5F93AECBDULL, "gen_im_arm_cht_elf_coma_t4.uti"},
	{0xE97253DFC5B1E66EULL, "88816542.ogg"},
	{0xE9727D914796F338ULL, "vxdr_slavers_141_0_0.vlm"},
	{0xE9736699DA1BC189ULL, "isa321_velasco.bnk"},
	{0xE973D6BD8111AF1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_wall_lean_right.anb"},
	{0xE9740BA90B51998BULL, "6132870_m.fxe"},
	{0xE974ACC762A171C1ULL, "815030989.ogg"},
	{0xE974C0D911B138E7ULL, "6178425_m.fxe"},
	{0xE9751917AD113422ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mase_0s.dds"},
	{0xE9751E3E22CB9215ULL, "hlac_lt_0_lowtown_a2_n_c_2_0y.dds"},
	{0xE9752A161CE231E4ULL, "art\\characters\\creatures\\strider\\animation\\walkfrontleft.ani"},
	{0xE97576BF4B573DEEULL, "hlac_lt_0_lowtown_a3_n_c_-3_1z.dds"},
	{0xE975A50EC1449578ULL, "6105000_m.fxe"},
	{0xE975D77CB8AE2446ULL, "dae000_gen_player_turn_in6177853.cl"},
	{0xE9761F854B6818D9ULL, "mcr370cr_boss_meredith.utc"},
	{0xE97654359352D879ULL, "gen00fl_merrill6109634.cl"},
	{0xE9766D1EEFCAC1F3ULL, "hlac_ca_caverns2_5802_40x.dds"},
	{0xE9767F16C50C4194ULL, "vxlt_hangedman_1jp_0_0.vlm"},
	{0xE9768FC3CF9FB929ULL, "hlac_lt_backalley_n_0_13z.dds"},
	{0xE9774899DA1F77C9ULL, "isa321_velasco.cnv"},
	{0xE97832179144F540ULL, "vxlt_gallowstemplar_f_6sf_0_0.vlm"},
	{0xE978A188885D66FCULL, "334135571.ogg"},
	{0xE978E7F706860CC0ULL, "follower_banter6141500.cl"},
	{0xE979B221693237CEULL, "hlac_ht_playershouse_n_855_23x.dds"},
	{0xE979CDA9F8510288ULL, "qcr000_arishok6039309.cl"},
	{0xE97A140296763170ULL, "vxht_keep_30f_0_0.vlm"},
	{0xE97A1AEEF6773FD0ULL, "vxlt_gallowstemplar_n_351_0_0.vlm"},
	{0xE97AEA5139EE3195ULL, "6131969_m.fxe"},
	{0xE97B8E757716AAC5ULL, "6156203_m.fxe"},
	{0xE97BF37CDD49CA16ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.cess.ani"},
	{0xE97C03A38F0D17D5ULL, "427151706.ogg"},
	{0xE97C27161CE84B46ULL, "art\\characters\\creatures\\strider\\animation\\walkfrontleft.gad"},
	{0xE97C4EEA00FC8AB6ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_ice_p.impa.evt"},
	{0xE97CDB3A715D2F93ULL, "t3_dog_o00.tnt"},
	{0xE97D86BF0D3B392FULL, "vxht_bartrand_n_5on_0_0.vlm"},
	{0xE97D930631199B5FULL, "gen_im_wep_war_two_act1_03.uti"},
	{0xE97DD5BB07B24624ULL, "hlac_w_encounter2_camp_-65532_0x.dds"},
	{0xE97E8DBF08067FFBULL, "653463507.ogg"},
	{0xE97F060E433642A4ULL, "matdef_dc_diffuse.dds"},
	{0xE97F06A344C7056EULL, "vxlt_warehouse_126_0_0.vlm"},
	{0xE97F9664C1543DC3ULL, "vxht_playershouse_n_1s0_0_0.vlm"},
	{0xE97FBE197E8FA091ULL, "6094963_m.fxe"},
	{0xE97FE221E5587F68ULL, "dae000_junar6147856.cl"},
	{0xE9809EA9AB05E877ULL, "vxht_bartrand_n_4hj_0_0.vlm"},
	{0xE980A1E9727BDD5EULL, "6087560_m.fxe"},
	{0xE980E4F18C01080CULL, "llac_ca_caverns1_mines_24425_9.dds"},
	{0xE98155329709F34DULL, "885627852.ogg"},
	{0xE982279FF20B7DC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_loop.gad"},
	{0xE982646F388B594FULL, "1040389619.ogg"},
	{0xE982AEF4D33C93E7ULL, "244551776.ogg"},
	{0xE98322161CEE6125ULL, "art\\characters\\creatures\\strider\\animation\\walkfrontleft.evt"},
	{0xE98345EA01029A41ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_ice_p.impa.gad"},
	{0xE9836ECD02CCF666ULL, "fx_streak_dis.mao"},
	{0xE983B1898F10DEE5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grainsack_sa.dds"},
	{0xE9842F36D686380BULL, "ht_0_hightown_58_0.rml"},
	{0xE984544930A8EF17ULL, "lyriumghost_haste.dds"},
	{0xE984C29597C540AEULL, "337046399.ogg"},
	{0xE984C3AC066A6B6BULL, "6087085_f.fxe"},
	{0xE98501EBB7608415ULL, "1054615340.ogg"},
	{0xE98632C4E29D3FD5ULL, "6119642_m.fxe"},
	{0xE98693F007C3E6C6ULL, "vxlt_0_lowtown_docks_b04_0_0.vlm"},
	{0xE9873F6D7AFCBE5FULL, "llac_ht_chantry_2534_26.dds"},
	{0xE988605D2BBB79C2ULL, "512602194.ogg"},
	{0xE988EA1C53DFCD89ULL, "6068906_m.fxe"},
	{0xE988EF6DC3AAC460ULL, "hlac_lt_undercitydungeon_-1_47x.dds"},
	{0xE9897B5B30C7656FULL, "vxht_keep_5gv_0_0.vlm"},
	{0xE989B36F66EC9C8CULL, "art\\vfx\\combat\\vfx_w_def_def_mode_act_c.anb"},
	{0xE989B3F0310A2709ULL, "6064053_f.fxe"},
	{0xE989F3717853B925ULL, "962521323.ogg"},
	{0xE989F559976400DCULL, "6119926_m.fxe"},
	{0xE989FB7CDD561507ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.cess.evt"},
	{0xE98A46EA0108BAEFULL, "art\\vfx\\combat\\vfx_m_amg_estorm_ice_p.impa.ani"},
	{0xE98B512EFE75DF1FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02b.msh"},
	{0xE98B572EFE75E9C1ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02b.mmh"},
	{0xE98B69D25CF9D9A2ULL, "1059934658.ogg"},
	{0xE98C11EDF1F5F88FULL, "character_character_p174.psh.fxo.dx11.assembly"},
	{0xE98C1E9993AE2E8DULL, "gen00fl_merrill6133143.cl"},
	{0xE98C81493A79054DULL, "vxlt_0_lowtown_n_gp8_0_0.vlm"},
	{0xE98CA0B491B54E16ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_portraits_d.dds"},
	{0xE98D1D1D6C669967ULL, "ico_locked_active.dds"},
	{0xE98E0841B34A5643ULL, "vxht_playershouse_n_1v3_0_0.vlm"},
	{0xE98E12854B7C94F7ULL, "mcr370cr_boss_meredith.ncs"},
	{0xE98E68BB14A2A17EULL, "dae000_gamlen6149412.cl"},
	{0xE98E70A71C362AFEULL, "6097558_m.fxe"},
	{0xE98E9008F14BB1A3ULL, "292777667.ogg"},
	{0xE98EE02BFA9E6848ULL, "6057084_m.fxe"},
	{0xE98EEDAF7590EBDDULL, "zz_end_choose_white.cut"},
	{0xE98F8B33C26500BBULL, "dre000_dougal6131156.cl"},
	{0xE9902793991125DAULL, "one200dg_solivitus6048194.cl"},
	{0xE9906C58E624E35DULL, "6113877_m.fxe"},
	{0xE99075D87F6DF062ULL, "6016232_m.fxe"},
	{0xE9907D039008CC14ULL, "vxlt_0_lowtown_a2_eox_0_1.vlm"},
	{0xE990C4107A38469BULL, "6170340_m.fxe"},
	{0xE9911C7CDD5C6C50ULL, "art\\vfx\\combat\\vfx_r_dlt_duelling_loop_c.cess.gad"},
	{0xE99184F743D22BEFULL, "hlac_ca_cavern1_haunted_24423_12y.dds"},
	{0xE9919F3E9FCF3A66ULL, "dae110ip_at_darktown.utp"},
	{0xE99229287423CD1FULL, "art\\characters\\weapons\\staff\\textures\\w_stf_005a_0n.dds"},
	{0xE992623559D33FEEULL, "6098713_m.fxe"},
	{0xE99275694FB37982ULL, "416628818.ogg"},
	{0xE9928830FEBDB9C6ULL, "6119685_m.fxe"},
	{0xE992AB843E33F2D0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic.mao"},
	{0xE992AF843E33F9E3ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic.mmh"},
	{0xE992BD843E34113DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic.msh"},
	{0xE992EB09102F45E9ULL, "hlac_lt_0_lowtown_a2_n_-196608_0x.dds"},
	{0xE99315477021714DULL, "gen00fl_isabela6079021.cl"},
	{0xE993171992B7BDB8ULL, "mag101_grace6034218.cl"},
	{0xE993A94DA24B0FCBULL, "6179986_f.fxe"},
	{0xE9940EA61DCE9DDEULL, "qun160_vashothlead6052561.cl"},
	{0xE9948C86BBD347F5ULL, "dae110ip_store_tailor.utp"},
	{0xE994A6963A0169DEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_femalestatue_d.dds"},
	{0xE994CA855EA63200ULL, "llac_lt_gallowscourtyard_n_-196608_0.dds"},
	{0xE994FB5501340ABDULL, "516702473.ogg"},
	{0xE99505CCF9B9FF46ULL, "vxlt_0_lowtown_docks_b75_0_0.vlm"},
	{0xE9950A7042AE3AD0ULL, "ser270_cut_aftermath6162295.cl"},
	{0xE9957E87A958BAF3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_pot_dp.dds"},
	{0xE995A66F28523A7BULL, "art\\levels\\races\\vfx_models\\_textures\\prp_lava1_sa.dds"},
	{0xE995B9B3400843E9ULL, "6151398_m.fxe"},
	{0xE995FF69F04E90AEULL, "hlac_ca_caverns2_5802_69x.dds"},
	{0xE9962BC3BE5A7A94ULL, "hlac_w_woundedcoast_262148_0x.dds"},
	{0xE996529FF21C6FA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\lean_bar_loop.ani"},
	{0xE9967512CBF023EAULL, "str_bigrock02_n.dds"},
	{0xE9969763AD0B2188ULL, "zz_arttest.lst"},
	{0xE996B08A15B1C11AULL, "vxlt_warehouse_n_8x_0_0.vlm"},
	{0xE996C7C7E1A504CDULL, "168897806.ogg"},
	{0xE997283AEEF78114ULL, "7545065.ogg"},
	{0xE997A097DB593DDCULL, "vxlt_0_lowtown_docks_a3_n_d2b_0_0.vlm"},
	{0xE9987028E9A2A7E7ULL, "qun231_party_comment6077134.cl"},
	{0xE998817FEA26D495ULL, "vxht_0_hightown_a3_b94_0_0.vlm"},
	{0xE998DE6B2328F089ULL, "vxlt_0_lowtown_docks_a3_d1e_0_0.vlm"},
	{0xE998E6AB559EC554ULL, "151140822.ogg"},
	{0xE998FC84CC5E96BEULL, "vxht_chantry_n_28b_0_0.vlm"},
	{0xE999003AE6793C96ULL, "vxlt_0_lowtown_a3_n_c_hm4_0_0.vlm"},
	{0xE9991451B7F62D2AULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_9y.dds"},
	{0xE99956F2F5149255ULL, "gen00fl_sebastian6187774.cl"},
	{0xE999A10794F2E5B3ULL, "art\\characters\\creatures\\dragon\\animation\\com_v1.evt"},
	{0xE999D16E376C007EULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.mmh"},
	{0xE999FE77B3DB3494ULL, "vxw_encounter2_camp_2j_0_0.vlm"},
	{0xE99A48E79E6E16A2ULL, "466741192.ogg"},
	{0xE99AA153ADF9F712ULL, "vxht_0_hightown_n_b7k_0_0.vlm"},
	{0xE99B49E697E911F2ULL, "mrlev_item_acquired.ncs"},
	{0xE99B72D8684F7D24ULL, "fer_ter_dirt_001_norm.dds"},
	{0xE99BA7BB9547EC55ULL, "262887727.ogg"},
	{0xE99BEF65C465CEEDULL, "vxlt_warehouse_15t_0_0.vlm"},
	{0xE99C2A576149E4D5ULL, "vxw_encounter3_73_0_0.vlm"},
	{0xE99CFF28615BF26CULL, "vxht_0_hightown_a2_c55_0_0.vlm"},
	{0xE99D22473714F255ULL, "963849936.ogg"},
	{0xE99D2D8400E30776ULL, "vxw_woundedcoast_13s_0_0.vlm"},
	{0xE99E4D0F3138308CULL, "6052920_m.fxe"},
	{0xE99E611E7B4AD1B2ULL, "133586109.ogg"},
	{0xE99F08A0EA5BF47EULL, "6076674_m.fxe"},
	{0xE99F1E27953B2697ULL, "6144325_m.fxe"},
	{0xE99F84E2C927C964ULL, "6046983_m.fxe"},
	{0xE99FBE6307F29221ULL, "hlac_w_encounter2_camp_131072_0y.dds"},
	{0xE9A078EC02E8E4E0ULL, "6208625_f.fxe"},
	{0xE9A09A0794F8F8F4ULL, "art\\characters\\creatures\\dragon\\animation\\com_v1.gad"},
	{0xE9A0BEB24B5625D8ULL, "gen_im_wep_war_one_act2_07.uti"},
	{0xE9A0E8D33997A39CULL, "506244531.ogg"},
	{0xE9A17C052B992FDDULL, "vxlt_0_lowtown_docks_b82_0_1.vlm"},
	{0xE9A1C82D3B1BA5B2ULL, "vxw_blightlands_c8_0_0.vlm"},
	{0xE9A2135FE29D2B14ULL, "6201269_m.fxe"},
	{0xE9A22FC713F71AD5ULL, "pro000ip_merchant_store.utp"},
	{0xE9A23E8A9DA62F50ULL, "vxlt_0_lowtown_docks_n_b7x_0_0.vlm"},
	{0xE9A25398C3315339ULL, "139439112.ogg"},
	{0xE9A2777C52CE8BF8ULL, "6081361_f.fxe"},
	{0xE9A28061E160A814ULL, "6135739_m.fxe"},
	{0xE9A330D357DA06A7ULL, "6152083_m.fxe"},
	{0xE9A3B0843E426E08ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic.phy"},
	{0xE9A400CF91E1F907ULL, "6067776_m.fxe"},
	{0xE9A463157104F0DEULL, "hlac_ht_0_hightown_a3_-65535_1x.dds"},
	{0xE9A4836589123C96ULL, "hlac_w_bonepitmining_a2_393218_0z.dds"},
	{0xE9A49AC395D9AC92ULL, "vxht_mansion_n_66t_0_0.vlm"},
	{0xE9A4D4047DD0DA49ULL, "llac_w_blightlands_0_2.dds"},
	{0xE9A4E92C0A311698ULL, "6127556_m.fxe"},
	{0xE9A55D48FEA8C06BULL, "155460253.ogg"},
	{0xE9A5A60E062A652EULL, "vxlt_0_lowtown_docks_apg_0_0.vlm"},
	{0xE9A5B54FBE2A6D34ULL, "vxht_0_hightown_a3_a74_0_0.vlm"},
	{0xE9A618F782FAE5A4ULL, "228597700.ogg"},
	{0xE9A629FDDF62AE99ULL, "832226750.ogg"},
	{0xE9A695B39507ADDFULL, "follower_banter6155610.cl"},
	{0xE9A71F9CA724F73BULL, "llac_ht_0_hightown_a2_65537_2.dds"},
	{0xE9A7443201BA3C33ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\bufftwitch.ani"},
	{0xE9A74E26CF5F9198ULL, "vxlt_hangedman_n_1mr_0_0.vlm"},
	{0xE9A76727BCB127B5ULL, "92850685.ogg"},
	{0xE9A7A90794FF30D2ULL, "art\\characters\\creatures\\dragon\\animation\\com_v1.ani"},
	{0xE9A7B06E3778053DULL, "art\\vfx\\combat\\vfx_m_wpn_pois_crust_c.anb"},
	{0xE9A894E9A0B02DB1ULL, "389391945.ogg"},
	{0xE9A8A01D809D952FULL, "hlac_w_bonepit_327685_0x.dds"},
	{0xE9A9B1F2AB98368CULL, "ser191_emeric6116977.cl"},
	{0xE9A9C133A87B85EBULL, "art\\levels\\races\\elf\\props\\_textures\\prp_dream_sa.dds"},
	{0xE9A9EBDF7B3A18ABULL, "vxlt_0_lowtown_docks_cp4_0_0.vlm"},
	{0xE9AA77BAC617A79BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_2.mao"},
	{0xE9AA96E322E386C3ULL, "qcr220_orsino6041834.cl"},
	{0xE9AB6A9E3F5EB0D7ULL, "vxdr_primevalentrance_2oj_0_2.vlm"},
	{0xE9ABEC5043B0D4B6ULL, "6141183_m.fxe"},
	{0xE9AC0070816CADB0ULL, "hlac_w_encounter3_-1_0y.dds"},
	{0xE9AC119C131B432DULL, "llac_lt_0_lowtown_docks_-131073_3.dds"},
	{0xE9AC423675A6CB1BULL, "vxlt_gallowstemplar_654_0_0.vlm"},
	{0xE9AC5F7862E3325FULL, "809028934.ogg"},
	{0xE9AC6968937DC2A3ULL, "ser200st_aveline.stg"},
	{0xE9AD09C0CF5A5CB8ULL, "vxca_cave_27n_0_0.vlm"},
	{0xE9AD4D0E97C76B50ULL, "vxca_caverns1_mines_ckj_0_0.vlm"},
	{0xE9AD4D4AB970384BULL, "hlac_w_sundermount_-196612_0x.dds"},
	{0xE9AD8D8CD10617FCULL, "gen00fl_varric6118738.cl"},
	{0xE9ADE80EEC18ACCFULL, "cod_tomes_tome_3.utp"},
	{0xE9AE619198B21556ULL, "125658465.ogg"},
	{0xE9AE8B1BB4D952E4ULL, "llac_lt_foundry_n_320_16.dds"},
	{0xE9AECA06C39DA5E8ULL, "vxw_bonepitmining_a2_4r_0_0.vlm"},
	{0xE9AEF16718B99E80ULL, "6028917_f.fxe"},
	{0xE9AF205AE2F0E984ULL, "vxlt_0_lowtown_a2_gmq_0_0.vlm"},
	{0xE9AF2D77E70819C5ULL, "6071892_m.fxe"},
	{0xE9B03A0A2D11BE32ULL, "vxw_blightlands_d9_0_0.vlm"},
	{0xE9B0BE03E68C8C29ULL, "vxlt_gallowscourtyard_n_6zr_0_0.vlm"},
	{0xE9B11F9D9EB0C5A6ULL, "vxht_mansion_n_6ni_0_1.vlm"},
	{0xE9B13C24F627ED58ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\combat_movement_varric_crossbow.anb"},
	{0xE9B1F034A69F651FULL, "vxht_brothel_n_6ao_0_0.vlm"},
	{0xE9B1FA846845EA75ULL, "vxlt_0_lowtown_docks_n_bgi_0_0.vlm"},
	{0xE9B2251C06FB2B59ULL, "6174805_m.fxe"},
	{0xE9B2CF15F735B406ULL, "hlac_lt_undercitydungeon_-1_46y.dds"},
	{0xE9B41ADBDA937610ULL, "vxht_chantry_36l_0_0.vlm"},
	{0xE9B42EFB1C056551ULL, "hlac_dr_dungeon_3216_3x.dds"},
	{0xE9B43459476DDCEBULL, "lit_aoe_distor.mao"},
	{0xE9B455BAC61FC2D8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\dr1a_slab_2.pwk"},
	{0xE9B486461A899D4EULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock_n.dds"},
	{0xE9B4D0F206A9AEEFULL, "vxht_keep_5ps_0_0.vlm"},
	{0xE9B4E021304BF815ULL, "vxht_mansion_n_4ve_0_0.vlm"},
	{0xE9B53C3201C66C6AULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\bufftwitch.evt"},
	{0xE9B54C4CFA8AF7AFULL, "llac_lt_gallowscourtyard_n_65535_0.dds"},
	{0xE9B5C003C155AA18ULL, "vxlt_hangedman_n_1js_0_0.vlm"},
	{0xE9B5C97F70A69581ULL, "6172946_m.fxe"},
	{0xE9B5DDBD528C3293ULL, "6217359_m.fxe"},
	{0xE9B6034017EF3717ULL, "6096901_m.fxe"},
	{0xE9B6095BDB04C3A6ULL, "212813157.ogg"},
	{0xE9B73F1D16F671B0ULL, "vxlt_gallowstemplar_f_430_0_0.vlm"},
	{0xE9B7582EFE9B263EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos02b.phy"},
	{0xE9B81E82833D9C78ULL, "6087089_m.fxe"},
	{0xE9B86789406456C6ULL, "dae000_bartrand6206982.cl"},
	{0xE9B899D028EFBC24ULL, "199571220.ogg"},
	{0xE9B8EE5310DC5C39ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi03.mmh"},
	{0xE9B8F85310DC6D27ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi03.msh"},
	{0xE9B941CC5968E161ULL, "108236667.ogg"},
	{0xE9B9ADA9E6E2483AULL, "6083927_f.fxe"},
	{0xE9BA0AA782D33F01ULL, "vxlt_0_lowtown_docks_a3_n_d23_0_0.vlm"},
	{0xE9BA113DE4CF3876ULL, "vxw_woundedcoast_19u_0_0.vlm"},
	{0xE9BA119D8C338ACEULL, "366213906.ogg"},
	{0xE9BA222D041ED0B8ULL, "178067296.ogg"},
	{0xE9BA395646A31DF0ULL, "540602287.ogg"},
	{0xE9BAB413679CE39BULL, "vxlt_gallowstemplar_687_0_0.vlm"},
	{0xE9BABE45F833EF86ULL, "1059957103.ogg"},
	{0xE9BB07E1DF10776AULL, "standard_targeting.cur"},
	{0xE9BB1BC3B7098E66ULL, "vxdr_0_deeproads_cf1_0_0.vlm"},
	{0xE9BB73FF333CCB3AULL, "6126815_f.fxe"},
	{0xE9BB830488A5BC06ULL, "vxdr_0_deeproads_bon_0_1.vlm"},
	{0xE9BB9EAF645B1557ULL, "375793009.ogg"},
	{0xE9BBBEEB89BD83D0ULL, "vxca_caverns1_mines_clk_0_0.vlm"},
	{0xE9BBE7A00A998ABEULL, "1008022735.ogg"},
	{0xE9BC0B952A029EEDULL, "dae000_mother6028524.cl"},
	{0xE9BC1B3201CC52DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\combat\\bufftwitch.gad"},
	{0xE9BC1C504B64932EULL, "813941033.ogg"},
	{0xE9BC3E41DC705AD5ULL, "6080773_m.fxe"},
	{0xE9BC599677F9919BULL, "6206924_m.fxe"},
	{0xE9BD2F5E53E50602ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cartempty_01_0_d.dds"},
	{0xE9BD6AB5A88083A0ULL, "68451954.ogg"},
	{0xE9BDCC77324DE797ULL, "147360576.ogg"},
	{0xE9BDD8DC551A373FULL, "345749546.ogg"},
	{0xE9BDFD1202588430ULL, "zz_vault_debug6178928.cl"},
	{0xE9BEFE2432B53E73ULL, "648105034.ogg"},
	{0xE9C003B5BBAC9D25ULL, "dae000_gen_player_turn_in6177827.cl"},
	{0xE9C0236BED0D5F42ULL, "707772172.ogg"},
	{0xE9C0413FC0F2D134ULL, "mag101_aombanter6052770.cl"},
	{0xE9C0621198957D9FULL, "vxht_brothel_n_6jn_0_0.vlm"},
	{0xE9C0808C079359F1ULL, "123089495.ogg"},
	{0xE9C0AD2A1ACD24F0ULL, "1049226402.ogg"},
	{0xE9C0FA9922424913ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\blendtree\\trooper_movement.anb"},
	{0xE9C153943C3F8428ULL, "vxlt_0_lowtown_a2_n_ewn_0_0.vlm"},
	{0xE9C1C0091AA31C8BULL, "t1_mul_ta2.tnt"},
	{0xE9C212CE8C384795ULL, "vxca_caverns1_mines_do4_0_0.vlm"},
	{0xE9C24F990E3985F2ULL, "6105921_m.fxe"},
	{0xE9C264977D8EEF69ULL, "hlac_ht_0_hightown_-65537_0y.dds"},
	{0xE9C289C7EA16962EULL, "vxlt_0_lowtown_docks_azy_0_0.vlm"},
	{0xE9C292E3F7F17294ULL, "6118738_m.fxe"},
	{0xE9C2E38393C5206AULL, "220917795.ogg"},
	{0xE9C40AE37C117D2AULL, "vxlt_0_lowtown_a3_n_c_eqe_0_0.vlm"},
	{0xE9C43B5C629CAE01ULL, "6175846_m.fxe"},
	{0xE9C476BAE75C16E9ULL, "gen00fl_merrill6106091.cl"},
	{0xE9C56F1D0D4AC0CCULL, "738690632.ogg"},
	{0xE9C59A76769C898AULL, "hlac_lt_0_lowtown_a3_n_c_65536_5y.dds"},
	{0xE9C5B0FA08EC8A30ULL, "vxlt_gallowstemplar_f_481_0_0.vlm"},
	{0xE9C5FA5D34D9F5E7ULL, "vxlt_gallowstemplar_f_477_0_0.vlm"},
	{0xE9C607A1D7E701CBULL, "mag360_samson6042036.cl"},
	{0xE9C63CC6762844DEULL, "docks_scene_dockworkers6176501.cl"},
	{0xE9C64F8C1BD4062DULL, "vxca_cave_2b0_0_0.vlm"},
	{0xE9C722F7BE9D2728ULL, "6134840_m.fxe"},
	{0xE9C75728A398098FULL, "art\\vfx\\textures\\vfx_web_fb.dds"},
	{0xE9C7E5E6CADF7086ULL, "6089809_m.fxe"},
	{0xE9C883FD5F521207ULL, "6057141_f.fxe"},
	{0xE9C916C7BA64D4A7ULL, "6102580_f.fxe"},
	{0xE9C9698F430312D9ULL, "art\\vfx\\textures\\vfx_fenris_glow.dds"},
	{0xE9C9974981C3CB29ULL, "6019252_m.fxe"},
	{0xE9C9A51D3F2C02B4ULL, "6121929_m.fxe"},
	{0xE9C9CF59AC42B8F6ULL, "hlac_ht_bartrand_n_3382_1z.dds"},
	{0xE9C9EDFE01B15C7AULL, "dae300_cullen6135970.cl"},
	{0xE9C9FFCDA9560EE5ULL, "vxht_0_hightown_a3_b9d_0_0.vlm"},
	{0xE9CA155141B9C8B0ULL, "hlac_ht_brothel_n_3370_0x.dds"},
	{0xE9CA198CAF74360FULL, "vxlt_0_lowtown_a2_fig_0_0.vlm"},
	{0xE9CAAA0F2027BC91ULL, "mcr371_scared_mage_f6087710.cl"},
	{0xE9CB6F55986A8A93ULL, "zz_one000_crafter.cnv"},
	{0xE9CC05F86552D5F5ULL, "lt_0_lowtown_52_0_ll.mmh"},
	{0xE9CC17AD0F8AF32CULL, "hlac_ht_0_hightown_a3_65537_2x.dds"},
	{0xE9CC20E71FEA6C3FULL, "427094102.ogg"},
	{0xE9CC47494C3B4C21ULL, "389347753.ogg"},
	{0xE9CCA6FDDB92DCA3ULL, "vxlt_0_lowtown_a2_n_c_fer_0_0.vlm"},
	{0xE9CCE60000C763C2ULL, "vxht_chantry_32r_0_0.vlm"},
	{0xE9CD5B5E92DAC561ULL, "dm_amb_worthy.mor"},
	{0xE9CD892236E11755ULL, "narrative_varric6124295.cl"},
	{0xE9CDA20337420453ULL, "407943998.ogg"},
	{0xE9CDCEB8CEB3BB37ULL, "vxlt_hangedman_25y_0_0.vlm"},
	{0xE9CE08C75E217001ULL, "hlac_ht_mansion2b_818_30z.dds"},
	{0xE9CE0A9DB8C8D111ULL, "w_encounter3_123_0.mmh"},
	{0xE9CE169DB8C8E5B2ULL, "w_encounter3_123_0.mao"},
	{0xE9CE249DB8C8FDCFULL, "w_encounter3_123_0.msh"},
	{0xE9CEFD00E336FE2FULL, "46173132.ogg"},
	{0xE9CF104A77FA1DDAULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_c.impa.gad"},
	{0xE9CF95B8E60D79A4ULL, "491857932.ogg"},
	{0xE9CFC09E310296CDULL, "mer140_bonepit_comments6062371.cl"},
	{0xE9CFC5712E359CA8ULL, "vxlt_0_lowtown_a2_n_exo_0_0.vlm"},
	{0xE9CFF3A3C928E6A6ULL, "279459262.ogg"},
	{0xE9D104C0E9E78B14ULL, "6117638_m.fxe"},
	{0xE9D14614C2620FE6ULL, "230229165.ogg"},
	{0xE9D1C775EEB3AB71ULL, "llac_w_bonepitdestroyed_a3_458756_0.dds"},
	{0xE9D1D20ABC1A756CULL, "vxw_bonepit_1pc_0_0.vlm"},
	{0xE9D1EA6315B8630DULL, "vxca_caverns1_icp_0_0.vlm"},
	{0xE9D1F7D26309D125ULL, "368680868.ogg"},
	{0xE9D21DE5D0E7BECCULL, "6043957_m.fxe"},
	{0xE9D2990E3C842A9DULL, "150148358.ogg"},
	{0xE9D3A096055CF876ULL, "art\\vfx\\creatures\\fxc_footstep_r_c.anb"},
	{0xE9D447CF2EB6C1D4ULL, "1009503024.ogg"},
	{0xE9D4759B3E25A1FBULL, "ts_qunari_01.gda"},
	{0xE9D4907C8426585FULL, "6109784_m.fxe"},
	{0xE9D4C81293991B4BULL, "6087669_f.fxe"},
	{0xE9D4C9D100E183D8ULL, "vxlt_darktown_4lj_0_0.vlm"},
	{0xE9D4D1472DA7DA0DULL, "lgt241ar_tome_cave.lst"},
	{0xE9D5338B34348F67ULL, "gen00fl_fenris6127092.cl"},
	{0xE9D53A3CD5427DF3ULL, "47000617.ogg"},
	{0xE9D5C1897CC9C4A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.ani"},
	{0xE9D6134A780041F1ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_c.impa.evt"},
	{0xE9D66D4C5772D7A4ULL, "gen00fl_carver6104603.cl"},
	{0xE9D71ECE9F2B9855ULL, "6119663_f.fxe"},
	{0xE9D731A46E41F299ULL, "llac_ht_mansion2_818_67.dds"},
	{0xE9D7CA9542712824ULL, "51398512.ogg"},
	{0xE9D8EF0ADCE4025FULL, "6099594_f.fxe"},
	{0xE9D954EE4D2933E3ULL, "812958541.ogg"},
	{0xE9D9F953F64987B1ULL, "hlac_lt_0_lowtown_n_65534_4x.dds"},
	{0xE9DA5E9BF3B6578BULL, "908982825.ogg"},
	{0xE9DA6D3A35FC9B3CULL, "803552245.ogg"},
	{0xE9DACF45F8D4E40FULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\_textures\\til_ruins01_n.dds"},
	{0xE9DAE5DB51989E86ULL, "mag291_ntbanter6056731.cl"},
	{0xE9DB18DACD88F523ULL, "vxlt_0_lowtown_a2_n_c_fhu_0_0.vlm"},
	{0xE9DB699317C11644ULL, "vxlt_darktown_4dn_0_0.vlm"},
	{0xE9DB74EEE407E357ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxa.phy"},
	{0xE9DB7B60ACD12F93ULL, "ranges.gda"},
	{0xE9DC5A467C873C28ULL, "6169859_m.fxe"},
	{0xE9DCCB85AABEB4C9ULL, "vxlt_gallowstemplar_4u8_0_0.vlm"},
	{0xE9DD2A24634484BDULL, "vxht_mansion2_1tu_0_0.vlm"},
	{0xE9DD411C2443E20AULL, "fen110_fenris6096408.cl"},
	{0xE9DD66866D8F77DDULL, "hlac_lt_0_lowtown_-2_5x.dds"},
	{0xE9DD70C32B6D9937ULL, "6064856_m.fxe"},
	{0xE9DD8CCE31D0EAF6ULL, "6147219_m.fxe"},
	{0xE9DDB31A926E709BULL, "hlac_lt_0_lowtown_a2_n_c_131070_1x.dds"},
	{0xE9DE1DB6A03DB0A7ULL, "vxlt_0_lowtown_docks_a3_bsk_0_0.vlm"},
	{0xE9DE5BA683823AD2ULL, "isa111cr_thug_2.utc"},
	{0xE9DEAD17873748A6ULL, "vxw_sundermount_270_0_0.vlm"},
	{0xE9DEDDE669259C03ULL, "6165319_m.fxe"},
	{0xE9DF39A0997ECB23ULL, "vxw_bonepit_17a_0_0.vlm"},
	{0xE9DF455BDBDEF7DDULL, "hlac_ca_caverns2_5802_49y.dds"},
	{0xE9DF9D7B53561613ULL, "vxht_mansion2_3gt_0_0.vlm"},
	{0xE9DFEEB111D6F631ULL, "6111219_m.fxe"},
	{0xE9DFF8111BB8B9D6ULL, "vxlt_darktown_65a_0_0.vlm"},
	{0xE9E0171D142D3A33ULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar5_.ani"},
	{0xE9E090FA63FED575ULL, "6095476_m.fxe"},
	{0xE9E124D2A36B7247ULL, "gen00fl_aveline6015820.cl"},
	{0xE9E1A39605693AE1ULL, "art\\vfx\\creatures\\fxc_footstep_r_c.mmh"},
	{0xE9E1F28FE8BCBEF0ULL, "gen00fl_isabela6090812.cl"},
	{0xE9E26B9016BC809CULL, "113463378.ogg"},
	{0xE9E28EE0BD3785C9ULL, "vxht_0_hightown_ca4_0_0.vlm"},
	{0xE9E2A5A017D9A147ULL, "gen_im_trs_rdr_med_glv_r2`.uti"},
	{0xE9E2A9897CD42617ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.evt"},
	{0xE9E32B4A780AF4C0ULL, "art\\vfx\\combat\\vfx_m_sta_spi_close_c.impa.ani"},
	{0xE9E3369B21A7529FULL, "dae300_bran6143482.cl"},
	{0xE9E39CA2DD035AB8ULL, "vxdr_primevalentrance_2l6_0_0.vlm"},
	{0xE9E3AC3211FD0DFAULL, "llac_w_encounter2_327682_0.dds"},
	{0xE9E4177911ACFD23ULL, "w_encounter2_19_0.rml"},
	{0xE9E4961FB1B9C95DULL, "vxht_0_hightown_8px_0_0.vlm"},
	{0xE9E498ADAAED8350ULL, "hlac_w_bonepit_262150_0x.dds"},
	{0xE9E52F228EBA2807ULL, "vxlt_gallowscourtyard_4xg_0_0.vlm"},
	{0xE9E542710C32995EULL, "montage_subs_demohack6217150.cl"},
	{0xE9E5488572D582DAULL, "vxht_mansion2_263_0_0.vlm"},
	{0xE9E60CF3463582F3ULL, "vxlt_0_lowtown_a2_cmv_0_0.vlm"},
	{0xE9E65E6EEF7B3265ULL, "vxht_mansion_n_4v5_0_0.vlm"},
	{0xE9E6B5857C6C5687ULL, "dae000_bartrand6030652.cl"},
	{0xE9E75040115FD32DULL, "qcr000_arishok6057096.cl"},
	{0xE9E7714CE05EAF60ULL, "ser221_gascard6134247.cl"},
	{0xE9E779C83579A707ULL, "6122754_m.fxe"},
	{0xE9E7D0D5A82CB812ULL, "gen_im_wep_war_two_lore_01.uti"},
	{0xE9E7F279478D2D67ULL, "mer240_jansen_st.stg"},
	{0xE9E84ECA045C1F44ULL, "6144530_f.fxe"},
	{0xE9E87098BC69F528ULL, "vxca_caverns1_mines_i1f_0_0.vlm"},
	{0xE9E87AA118875BE9ULL, "vxht_0_hightown_a3_dw_0_0.vlm"},
	{0xE9E9040C5206203FULL, "hlac_ca_caverns2_5802_21y.dds"},
	{0xE9E91719DC71ED35ULL, "fxe_day_dis.mao"},
	{0xE9E976655D2131F4ULL, "gen00fl_isabela6096960.cl"},
	{0xE9E9778316ACD91DULL, "6119484_m.fxe"},
	{0xE9E9A5103002EF5AULL, "hlac_ht_0_hightown_-131070_0z.dds"},
	{0xE9E9AA897CDA46A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_action_enter.gad"},
	{0xE9EA18E2B24D33B3ULL, "505697809.ogg"},
	{0xE9EA1E7911B1749BULL, "w_encounter2_19_0.phy"},
	{0xE9EA6A70B9F0FBC6ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\blendtree\\doorsimple.anb"},
	{0xE9EA963D4102460EULL, "vxca_cavern1_haunted_j1y_0_2.vlm"},
	{0xE9EB01C25A017F34ULL, "vxht_0_hightown_a3_byq_0_0.vlm"},
	{0xE9EB398FE3C45DAAULL, "m_isa211ar_foundry.wbd"},
	{0xE9EB5D5BE69DC109ULL, "zz_seb_debug6215770.cl"},
	{0xE9EBFEAB23C70376ULL, "6140319_m.fxe"},
	{0xE9EC7B4A37C13FA4ULL, "vxlt_undercitydungeon_1yc_0_0.vlm"},
	{0xE9ED1EF4792D6126ULL, "vxw_sundermount_281_0_0.vlm"},
	{0xE9ED8BAF9DBDFD54ULL, "hlac_ht_mansion2_818_4x.dds"},
	{0xE9EDA9D6862803D0ULL, "vxlt_gallowstemplar_n_3m9_0_0.vlm"},
	{0xE9EDB84522600B41ULL, "sg5_water_cyc.mao"},
	{0xE9EE9BE0DA31CD23ULL, "6068012_m.fxe"},
	{0xE9EECE1CF9A4940DULL, "vxca_caverns1_iin_0_0.vlm"},
	{0xE9EED662B6371632ULL, "6059973_f.fxe"},
	{0xE9EF4A9670CE0D49ULL, "vxht_0_hightown_a3_aq3_0_0.vlm"},
	{0xE9EFB443C1A9D570ULL, "vxdr_0_deeproads_9gn_0_0.vlm"},
	{0xE9EFB5DC92A0BF96ULL, "vxht_brothel_n_6cp_0_0.vlm"},
	{0xE9F08194FF62B03FULL, "566450977.ogg"},
	{0xE9F0BA30427AED36ULL, "lt_0_lowtown_docks_a3_211_0.mao"},
	{0xE9F0C630427B01D5ULL, "lt_0_lowtown_docks_a3_211_0.mmh"},
	{0xE9F0C830427B054BULL, "lt_0_lowtown_docks_a3_211_0.msh"},
	{0xE9F0DF106DBB26B3ULL, "vxlt_gallowscourtyard_3bb_0_0.vlm"},
	{0xE9F0FF0D8FD96FB5ULL, "6110951_m.fxe"},
	{0xE9F142445C4322E6ULL, "vxht_chantry_2q2_0_0.vlm"},
	{0xE9F14700AE1F9299ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_bricks.msh"},
	{0xE9F15100AE1FA387ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_bricks.mmh"},
	{0xE9F15500AE1FAAD4ULL, "art\\levels\\races\\ferelden\\props\\dead\\prp_bricks.mao"},
	{0xE9F174366812895FULL, "6107984_m.fxe"},
	{0xE9F1A92EC38E84A1ULL, "1039021470.ogg"},
	{0xE9F20C605CAA9A5BULL, "63779131.ogg"},
	{0xE9F2AE83BB4F0739ULL, "853874622.ogg"},
	{0xE9F2D04806B8A0AEULL, "art\\vfx\\combat\\vfx_g_appearance_c.dura.evt"},
	{0xE9F4A0DDB8AC6331ULL, "6121007_m.fxe"},
	{0xE9F4BF72D1AC847DULL, "6059387_f.fxe"},
	{0xE9F4EE1D143F50DDULL, "art\\cinematics\\_crit path\\mage crisis\\rulesflees\\playerflees_extratemplar5_.gad"},
	{0xE9F53DE48CA9A50BULL, "vxca_caverns1_mines_jde_0_0.vlm"},
	{0xE9F56AF4B29209E1ULL, "831995282.ogg"},
	{0xE9F56D7DE2A5DBA8ULL, "671074320.ogg"},
	{0xE9F657835472FF42ULL, "1049632171.ogg"},
	{0xE9F6AB9EC4050AF9ULL, "429816298.ogg"},
	{0xE9F6C58A1F4BAACBULL, "168460802.ogg"},
	{0xE9F6FD2B29F9E9C8ULL, "scaleform_text.mao"},
	{0xE9F7162931CCFAFEULL, "isa211cr_walleyedsam.utc"},
	{0xE9F83F5B4EE64CD9ULL, "hlac_ca_cavern1_haunted_24425_8z.dds"},
	{0xE9F8B2B1934B8980ULL, "vxht_mansion2b_30z_0_0.vlm"},
	{0xE9F8CBE831DE7136ULL, "follower_banter6141520.cl"},
	{0xE9F9359DDF0BDAD7ULL, "gen_im_trs_chn_shd_wdn_r0.uti"},
	{0xE9F9517605E2A424ULL, "1057227574.ogg"},
	{0xE9F9C74806BEB059ULL, "art\\vfx\\combat\\vfx_g_appearance_c.dura.gad"},
	{0xE9FA0DDE4730B5BDULL, "vxht_mansion2_1bs_0_0.vlm"},
	{0xE9FAA4F2DEEF4268ULL, "vxlt_hangedman_61_0_0.vlm"},
	{0xE9FB4A8676945787ULL, "fen310_varania6118595.cl"},
	{0xE9FB578FEF38D219ULL, "621842442.ogg"},
	{0xE9FB7A18CFD1C773ULL, "6135008_m.fxe"},
	{0xE9FBBA9B2E200AC6ULL, "dae011_drunk1_sta6149665.cl"},
	{0xE9FBD7F39C44EF7EULL, "409097766.ogg"},
	{0xE9FC408C40F5CB5DULL, "6165146_m.fxe"},
	{0xE9FC82587079D8B0ULL, "gen_im_trs_prf_amu_r0.uti"},
	{0xE9FC89AEC08E97BFULL, "vxw_bonepit_1pl_0_0.vlm"},
	{0xE9FCA80268A12B25ULL, "mcr370_negotiation6078756.cl"},
	{0xE9FD0847625C1AC9ULL, "ser221_party_comment_3.stg"},
	{0xE9FD2DF7850153B3ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.cess.evt"},
	{0xE9FD6DF02F275806ULL, "386136768.ogg"},
	{0xE9FD8BFAF8398CBBULL, "vxlt_hangedman_n_2yb_0_0.vlm"},
	{0xE9FDCC13AFDEB381ULL, "dae000_gamlen6048479.cl"},
	{0xE9FDDCA5DD30AFE7ULL, "vxlt_gallowscourtyard_6wt_0_0.vlm"},
	{0xE9FE12D42BA81EB3ULL, "vxlt_0_lowtown_a2_fe7_0_0.vlm"},
	{0xE9FE3BAAECAAD9BCULL, "6067384_m.fxe"},
	{0xE9FE65EAF0087CA8ULL, "668700816.ogg"},
	{0xE9FE7A7D6B8B4F35ULL, "hlac_ht_mansion2_818_18z.dds"},
	{0xE9FE7CBFB1A317E0ULL, "gen00fl_anders6125934.cl"},
	{0xE9FEA22616D49976ULL, "mcr310_argument6074326.cl"},
	{0xE9FEAC423990EC07ULL, "6088829_m.fxe"},
	{0xE9FEDFC880F02AEFULL, "6180875_m.fxe"},
	{0xE9FEEC3BBDE1AD09ULL, "6073784_m.fxe"},
	{0xE9FF3B04015FE582ULL, "179738412.ogg"},
	{0xE9FF5ABA9F85DA0CULL, "vxlt_undercityencounter_18y_0_0.vlm"},
	{0xE9FF8549F8776FF0ULL, "6106632_m.fxe"},
	{0xE9FFADCE56370BC7ULL, "llac_dr_dungeon_3217_0.dds"},
	{0xE9FFD44D89C46807ULL, "vxdr_0_deeproads_dw8_0_0.vlm"},
	{0xE9FFE3D2890B5392ULL, "lt_0_lowtown_a2_42_0.anb"},
	{0xEA000EFEE21663A9ULL, "seb321_leliana6181813.cl"},
	{0xEA004CBD2C8990E2ULL, "6125462_m.fxe"},
	{0xEA0090E9C7EEDAC9ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivyline.mao"},
	{0xEA009CE9C7EEEF62ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivyline.mmh"},
	{0xEA009EE9C7EEF218ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivyline.msh"},
	{0xEA00D84806C4EBA7ULL, "art\\vfx\\combat\\vfx_g_appearance_c.dura.ani"},
	{0xEA00DA472DCD24CAULL, "lgt241ar_tome_cave.are"},
	{0xEA01F9A31F23FC68ULL, "6080620_m.fxe"},
	{0xEA01FDC45FFB2E5CULL, "w_encounter1_39_0.mao"},
	{0xEA0209C45FFB42FFULL, "w_encounter1_39_0.mmh"},
	{0xEA020FC45FFB4CA1ULL, "w_encounter1_39_0.msh"},
	{0xEA026EDF31C95592ULL, "6171207_m.fxe"},
	{0xEA029694C74112DAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galo2floor.mmh"},
	{0xEA029A94C7411911ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galo2floor.mao"},
	{0xEA029B727376F41EULL, "art\\characters\\creatures\\shriek\\textures\\cn_hed_ems_0s.dds"},
	{0xEA02A894C7413120ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galo2floor.msh"},
	{0xEA03DC22600C23BEULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kita_0t.dds"},
	{0xEA0426F7850766F4ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.cess.gad"},
	{0xEA0437D40E12822AULL, "vxlt_0_lowtown_docks_b08_0_0.vlm"},
	{0xEA04E50379D09245ULL, "character_character_p187.psh.fxo.dx11"},
	{0xEA04E646010536CDULL, "6169094_m.fxe"},
	{0xEA05AF48FE74AE3EULL, "vxlt_hangedman_n_1w6_0_0.vlm"},
	{0xEA05BA7BC5119C2AULL, "mag121_idunna6083943.cl"},
	{0xEA063B8B931572E5ULL, "vxlt_0_lowtown_a2_n_h1x_0_0.vlm"},
	{0xEA0661DE2FB153FEULL, "6126709_f.fxe"},
	{0xEA06E9F8C680AB1DULL, "6088759_m.fxe"},
	{0xEA0724B049B8AAC5ULL, "186232539.ogg"},
	{0xEA0762192FDF3131ULL, "vxht_0_hightown_a3_c15_0_0.vlm"},
	{0xEA07E57C3DEDB034ULL, "vxht_0_hightown_a3_bsw_0_0.vlm"},
	{0xEA08605D43B563F0ULL, "zz_approval_debug6003935.cl"},
	{0xEA08E3264611304CULL, "follower_banter6156081.cl"},
	{0xEA096044F3F3CBBEULL, "565621595.ogg"},
	{0xEA0ABE6CDD808E9FULL, "lt_0_lowtown_46_0_lc.mmh"},
	{0xEA0AC1A6CD296F55ULL, "dae000ip_crowd_ff2_talk13.utp"},
	{0xEA0B1734606E0C94ULL, "llac_lt_0_lowtown_docks_a3_-65537_4.dds"},
	{0xEA0B35F7850D9ED2ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.cess.ani"},
	{0xEA0B6962BB266B60ULL, "qun221_petrice6076742.cl"},
	{0xEA0BA18274D1E57EULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedflurry.evt"},
	{0xEA0BC770EC2D5D33ULL, "vxht_0_hightown_n_8pq_0_0.vlm"},
	{0xEA0C1D9149B0847FULL, "w_bonepitdestroyed_a3_153_0.msh"},
	{0xEA0C239149B08E21ULL, "w_bonepitdestroyed_a3_153_0.mmh"},
	{0xEA0C2F9149B0A2C2ULL, "w_bonepitdestroyed_a3_153_0.mao"},
	{0xEA0C7536367D5DA1ULL, "vxht_0_hightown_a3_a4t_0_0.vlm"},
	{0xEA0C9996768CF096ULL, "vxht_brothel_n_6ir_0_0.vlm"},
	{0xEA0C9C7D2FFF7509ULL, "plt_gen00pt_class_race_gend.plo"},
	{0xEA0CA51B38730385ULL, "131102416.ogg"},
	{0xEA0D0727A01D8385ULL, "vxht_0_hightown_n_c48_0_0.vlm"},
	{0xEA0D347E20FB285DULL, "hlac_lt_blackemporium_1055_21z.dds"},
	{0xEA0D92955E60C323ULL, "60456934.ogg"},
	{0xEA0E16D3C74C753CULL, "6055598_m.fxe"},
	{0xEA0E5C520F9E2347ULL, "llac_lt_0_lowtown_a3_n_c_-131071_0.dds"},
	{0xEA10FC3134202EBDULL, "6145104_m.fxe"},
	{0xEA115EF6B433FC27ULL, "926194004.ogg"},
	{0xEA11B405C55D7BE5ULL, "vxht_mansion_n_4g2_0_0.vlm"},
	{0xEA12A9B100089AAAULL, "vxlt_0_lowtown_docks_b79_0_0.vlm"},
	{0xEA12AEAC933518EAULL, "915802230.ogg"},
	{0xEA12B88274D82AA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedflurry.gad"},
	{0xEA1316B8261B5B6FULL, "6143445_m.fxe"},
	{0xEA13B195146936B4ULL, "dae210ar_docks_nt.lst"},
	{0xEA13C33C58145FC6ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_ltr.anb"},
	{0xEA145DEBFD5583ABULL, "6032474_f.fxe"},
	{0xEA14C6EB38BBEB8AULL, "t3_gsp_b00.tnt"},
	{0xEA150A68D0BA4E8CULL, "6085292_m.fxe"},
	{0xEA1527AA72E2C719ULL, "93194625.ogg"},
	{0xEA155BD5B876C39DULL, "6083659_m.fxe"},
	{0xEA15A03153684E47ULL, "6034930_m.fxe"},
	{0xEA15FF25495FC0E5ULL, "llac_lt_0_lowtown_docks_a3_131070_2.dds"},
	{0xEA164CF8557E1D34ULL, "6101339_f.fxe"},
	{0xEA1657592FE3C8B4ULL, "vxht_0_hightown_a3_bnt_0_0.vlm"},
	{0xEA169A5EAC67E6EBULL, "693866247.ogg"},
	{0xEA16B94F2CB094FFULL, "317811886.ogg"},
	{0xEA16F1982B1CE6BDULL, "vxht_mansion2_1hq_0_0.vlm"},
	{0xEA17057911D8383AULL, "w_encounter2_19_0.msh"},
	{0xEA17137911D85073ULL, "w_encounter2_19_0.mao"},
	{0xEA17177911D85680ULL, "w_encounter2_19_0.mmh"},
	{0xEA17458085F402FAULL, "6115898_m.fxe"},
	{0xEA17A3909F767C69ULL, "vxlt_0_lowtown_docks_a3_n_9nn_0_0.vlm"},
	{0xEA17C1BFADCE3822ULL, "6104276_m.fxe"},
	{0xEA17CD9AD6A1EE96ULL, "6117773_m.fxe"},
	{0xEA17FF3ECB89D4A3ULL, "dae000_denier6162503.cl"},
	{0xEA187409990C3C3EULL, "571503647.ogg"},
	{0xEA18963890A4A522ULL, "76398850.ogg"},
	{0xEA18AB386F534814ULL, "111150452.ogg"},
	{0xEA18E4E854457465ULL, "qun120cr_javaris6172017.cl"},
	{0xEA18F3433782D5FDULL, "vxht_0_hightown_a2_apm_0_0.vlm"},
	{0xEA1911F2B4AB68DFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_kill.gad"},
	{0xEA1914BDEEE8206AULL, "gen00fl_fenris6127033.cl"},
	{0xEA194413CE64D827ULL, "6034822_m.fxe"},
	{0xEA197686080ADFB9ULL, "vxlt_0_lowtown_a2_n_f4m_0_0.vlm"},
	{0xEA19898274DDF9F7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\partypicker\\magestaffselectedflurry.ani"},
	{0xEA19B8C3ADD842F4ULL, "vxca_caverns1_mines_icq_0_1.vlm"},
	{0xEA19D2CF556A84CCULL, "2389992.ogg"},
	{0xEA1A0B5B75E8EFECULL, "vxw_bonepit_1id_0_0.vlm"},
	{0xEA1A2A4A79AC5EC8ULL, "mag130_danzig6018815.cl"},
	{0xEA1AA21A95DA2B3AULL, "hlac_lt_gallowstemplar_n_2383_11y.dds"},
	{0xEA1ABB9213DF568CULL, "w_encounter2_6_0.msh"},
	{0xEA1AC05FC11CE0E7ULL, "vxlt_gallowscourtyard_6yv_0_0.vlm"},
	{0xEA1AC19213DF617EULL, "w_encounter2_6_0.mmh"},
	{0xEA1ACD9213DF7515ULL, "w_encounter2_6_0.mao"},
	{0xEA1AD195146F8BBDULL, "dae210ar_docks_nt.ncs"},
	{0xEA1AEF53112F2826ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hi03.phy"},
	{0xEA1AF15231FA2DB8ULL, "484540883.ogg"},
	{0xEA1B33FFD331568FULL, "vxdr_primevaltomb_7ei_0_2.vlm"},
	{0xEA1B52301B98531EULL, "624023579.ogg"},
	{0xEA1B54A743A9CD88ULL, "412862085.ogg"},
	{0xEA1B6783E0A254DEULL, "944743274.ogg"},
	{0xEA1BB1AAEDD0061DULL, "153205661.ogg"},
	{0xEA1BCA700243D981ULL, "70314482.ogg"},
	{0xEA1C952DC7D4A0A4ULL, "vxht_brothel_n_6c2_0_0.vlm"},
	{0xEA1CDF4213E8616BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.opentoclosed.evt"},
	{0xEA1D29C0B9EA6816ULL, "6136930_m.fxe"},
	{0xEA1E2A3E5FAA64DEULL, "llac_ht_mansion2_818_20.dds"},
	{0xEA1E4D3BCAA89996ULL, "345744661.ogg"},
	{0xEA1E6BFA8757CE81ULL, "680777978.ogg"},
	{0xEA1F13C3EA580CFEULL, "6027975_m.fxe"},
	{0xEA1F707DB9B3B51BULL, "follower_barks6153857.cl"},
	{0xEA1F99BAB56B70B3ULL, "539592680.ogg"},
	{0xEA1FFDB21FAA70D1ULL, "261210807.ogg"},
	{0xEA2006F2B4B174CCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_kill.evt"},
	{0xEA20B14D4F99FA96ULL, "dae000_flemeth6000826.cl"},
	{0xEA20E9EC8AF3B66FULL, "vxht_brothel_n_5al_0_0.vlm"},
	{0xEA217FC36F0F070FULL, "hlac_lt_0_lowtown_a3_n_c_-131073_1y.dds"},
	{0xEA21FB859F29F730ULL, "253365091.ogg"},
	{0xEA22F61CFD0076A0ULL, "combat_st_r5_a1.gda"},
	{0xEA22F84213ECF72CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.opentoclosed.gad"},
	{0xEA2323A73E809CA3ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower3_meredithsintro.ani"},
	{0xEA2349EEE4454FAFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxa.mao"},
	{0xEA2355EEE445644CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxa.mmh"},
	{0xEA235BEEE4456E3EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\9pp_columnringxa.msh"},
	{0xEA23B69F23BC4A20ULL, "one110_meeran6173914.cl"},
	{0xEA23BD8D39F4D251ULL, "hlac_lt_0_lowtown_0_0z.dds"},
	{0xEA23CC7AFBBA214CULL, "6083785_petrice_arrowed.cut"},
	{0xEA23FA7911E2AF9FULL, "w_encounter2_19_0.anb"},
	{0xEA24587FDF1AC399ULL, "llac_w_encounter2_5_0.dds"},
	{0xEA24B4F040F36184ULL, "art\\vfx\\combat\\vfx_gib_arm_masc.mao"},
	{0xEA24DE191AC722A7ULL, "33888573.ogg"},
	{0xEA251A436A2736DDULL, "vxht_0_hightown_n_c3t_0_1.vlm"},
	{0xEA2531BD95ADF0B1ULL, "vxlt_hangedman_n_4w_0_0.vlm"},
	{0xEA2554A9C44FC22FULL, "147106725.ogg"},
	{0xEA257294D13EB7F1ULL, "dae000_madam_lusine6152171.cl"},
	{0xEA2595E7C5C793D7ULL, "follower_banter6161136.cl"},
	{0xEA2599DD63AC7A10ULL, "430047422.ogg"},
	{0xEA25ABD8AABE5B80ULL, "qcr000_arishok6057118.cl"},
	{0xEA269025E743DBC0ULL, "vxlt_gallowscourtyard_72i_0_0.vlm"},
	{0xEA26A37CB965FCCFULL, "vxlt_gallowstemplar_n_477_0_0.vlm"},
	{0xEA26B98182A0C715ULL, "892014320.ogg"},
	{0xEA26BF8E65450C34ULL, "6024525_m.fxe"},
	{0xEA2736D33CAA4ADCULL, "hlac_dr_primevaltomb_9621_7y.dds"},
	{0xEA27B3937994E6DCULL, "329058654.ogg"},
	{0xEA28198835269BE4ULL, "vxlt_backalley_n_1nw_0_1.vlm"},
	{0xEA283C83BBB10DCBULL, "follower_banter6161064.cl"},
	{0xEA28B594C76216E5ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ff44_galo2floor.phy"},
	{0xEA290612835970DEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_enter.ani"},
	{0xEA298072AEF76B51ULL, "vxht_0_hightown_by9_0_0.vlm"},
	{0xEA29817E12BED89CULL, "hlac_dr_slavers_3215_7z.dds"},
	{0xEA29CD96DF4D3D11ULL, "dre000_carver6037615.cl"},
	{0xEA29F74213F3142AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\hightown_gate.opentoclosed.ani"},
	{0xEA2A1CE0A55DB737ULL, "lt_0_lowtown_docks_n_251_0.msh"},
	{0xEA2A2EE0A55DD59AULL, "lt_0_lowtown_docks_n_251_0.mao"},
	{0xEA2A32E0A55DDCA9ULL, "lt_0_lowtown_docks_n_251_0.mmh"},
	{0xEA2A7125C0275AB1ULL, "345092466.ogg"},
	{0xEA2B4B742A06EB3FULL, "978846878.ogg"},
	{0xEA2B5978EB643799ULL, "llac_lt_0_lowtown_docks_n_65535_0.dds"},
	{0xEA2BBA731FE97B9EULL, "dae000_fem_commoner1_mob6174488.cl"},
	{0xEA2BD61D64EF777FULL, "106876823.ogg"},
	{0xEA2C19D8BED409C8ULL, "llac_w_bonepitmining_a2_393216_0.dds"},
	{0xEA2C37DB9CF647A5ULL, "6097427_m.fxe"},
	{0xEA2CEC548FD47D6BULL, "llac_w_encounter1_-65534_0.dds"},
	{0xEA2CEEF2B4BBD625ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\idunna_prostitute_reaction_kill.ani"},
	{0xEA2D0D89BC8D0FB5ULL, "art\\levels\\races\\design\\_textures\\prp_herbs_n.dds"},
	{0xEA2D6854299EF2CBULL, "follower_barks6153347.cl"},
	{0xEA2D80BFF9471603ULL, "dae000ip_crowd_mf7_talk8.utp"},
	{0xEA2DA734DF3BBC14ULL, "hlac_lt_0_lowtown_docks_-4_5z.dds"},
	{0xEA2E2D0E2EFDCB99ULL, "hlac_ca_caverns1_mines_24426_5z.dds"},
	{0xEA2E63D668AB2517ULL, "gen00fl_dog6166801.cl"},
	{0xEA2E9D2C0197950AULL, "255935896.ogg"},
	{0xEA2F0F0A78E7499EULL, "vxht_mansion2_24q_0_0.vlm"},
	{0xEA2F476E38EB26F4ULL, "vxca_caverns2_93n_0_0.vlm"},
	{0xEA2F6CE9AF930021ULL, "dae000ip_crowd_lm1_talk4.utp"},
	{0xEA2F75B5FC0A6028ULL, "x_narrativeroom_0_lc.mmh"},
	{0xEA3015A32D04AE4FULL, "70624376.ogg"},
	{0xEA3051B1FDC2863DULL, "llac_ht_mansion2b_818_21.dds"},
	{0xEA3058EE6A980B38ULL, "6138818_f.fxe"},
	{0xEA305DB4FA29F62AULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.impa.gad"},
	{0xEA32178FE40026C0ULL, "m_isa211ar_foundry.bnk"},
	{0xEA322FBC649372AFULL, "6139663_f.fxe"},
	{0xEA32BF9C9FD9ECC7ULL, "dae000_mother6099193.cl"},
	{0xEA32E6D28936B2CEULL, "lt_0_lowtown_a2_42_0.rml"},
	{0xEA332B5E59C8B786ULL, "6164847_m.fxe"},
	{0xEA3330AB44AE785CULL, "648560436.ogg"},
	{0xEA333AFC3162574AULL, "hlac_ca_caverns2_5802_56z.dds"},
	{0xEA336C153E33911EULL, "856384947.ogg"},
	{0xEA33E9D7A238B7CEULL, "ht_bartrand_n_4152_lc.mmh"},
	{0xEA34D2D4CF56DD8CULL, "vxlt_0_lowtown_docks_a3_d2e_0_0.vlm"},
	{0xEA35025E126CB35BULL, "dae000_merchant_dm_mob6174308.cl"},
	{0xEA352062C50E7DD9ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_hvyn_0s.dds"},
	{0xEA353B6F9E7A85E3ULL, "qun221_petrice6076646.cl"},
	{0xEA358AF341B63DA9ULL, "97418262.ogg"},
	{0xEA35F4724CC0AFCBULL, "353300153.ogg"},
	{0xEA36069555B587B2ULL, "501852579.ogg"},
	{0xEA3616D379511371ULL, "6176119_m.fxe"},
	{0xEA361E12836423AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_enter.evt"},
	{0xEA3740B4FA2FE3C1ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.impa.evt"},
	{0xEA382BC210D33BDDULL, "514165001.ogg"},
	{0xEA383AA73E931F8DULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower3_meredithsintro.gad"},
	{0xEA3887FA0AE087B1ULL, "gen00fl_fenris6110874.cl"},
	{0xEA38B89514885E87ULL, "dae210ar_docks_nt.are"},
	{0xEA38DEFB71B82C3EULL, "hlac_lt_ruinedhouse_n_-1_9y.dds"},
	{0xEA39249149D77D9EULL, "w_bonepitdestroyed_a3_153_0.phy"},
	{0xEA39359A2F5B8EC4ULL, "lgt000cr_kirk_turn_in_4.utc"},
	{0xEA3961513AD58545ULL, "6146879_m.fxe"},
	{0xEA3978E7ABC0D1A4ULL, "vxht_brothel_n_6i0_0_0.vlm"},
	{0xEA39C5B1B9515BF3ULL, "default_grey.dds"},
	{0xEA3A5FA634F418E3ULL, "6048422_f.fxe"},
	{0xEA3A881D7AB929E1ULL, "171556552.ogg"},
	{0xEA3AD369806FDE28ULL, "vxlt_0_lowtown_docks_n_a0n_0_0.vlm"},
	{0xEA3B6EA008C73905ULL, "705021705.ogg"},
	{0xEA3BFA3CB11A7B21ULL, "6042726_m.fxe"},
	{0xEA3C1FF95B65FA61ULL, "hlac_dr_dungeon_3214_5x.dds"},
	{0xEA3C8A85241BDA0AULL, "880300825.ogg"},
	{0xEA3C8C0FCC5FAC85ULL, "hlac_lt_0_lowtown_docks_-65540_2z.dds"},
	{0xEA3C9FBBA2AF877AULL, "657046675.ogg"},
	{0xEA3CEE6EA2E48B19ULL, "tal_tau_out.mao"},
	{0xEA3CF8692A958DE6ULL, "174761091.ogg"},
	{0xEA3D1F12836A4458ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_one_knee_enter.gad"},
	{0xEA3D8915246D0307ULL, "vxlt_0_lowtown_a2_fag_0_0.vlm"},
	{0xEA3E47BE25049BBFULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_viscdesk.phy"},
	{0xEA3E884D1B603EA1ULL, "charhighlight2.vsh.fxo"},
	{0xEA3E89DE039B7817ULL, "vxlt_undercityserialkiller_1ip_0_1.vlm"},
	{0xEA3EDFC6EA24F27BULL, "vxw_bonepit_26m_0_0.vlm"},
	{0xEA3F27A231DF2487ULL, "pro000_merchant6051605.cl"},
	{0xEA3F7B0D8E347868ULL, "6033412_f.fxe"},
	{0xEA3FA0F540AE0CC5ULL, "749321120.ogg"},
	{0xEA3FC38FCB5555EFULL, "lt_warehouse_0_0_lc.mmh"},
	{0xEA403970A160BDFEULL, "hlac_lt_gallowsprison_n_-1_4y.dds"},
	{0xEA4049D1D8E32240ULL, "vxlt_gallowstemplar_n_6qd_0_0.vlm"},
	{0xEA40592B8F2F8A63ULL, "ts_generic_acc_table.gda"},
	{0xEA40AA8921B4B2E4ULL, "wm_ico_deeproads.dds"},
	{0xEA40F638969E0656ULL, "gen00fl_aveline6139422.cl"},
	{0xEA41D31870A40665ULL, "6066618_m.fxe"},
	{0xEA4208A5524C8C00ULL, "gen00fl_anders6111128.cl"},
	{0xEA42984B59B20130ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_04ruinstower.mao"},
	{0xEA429C4B59B20843ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_04ruinstower.mmh"},
	{0xEA42A21A7299757BULL, "hlac_w_encounter2_camp_-65534_0y.dds"},
	{0xEA42A97F16EA234FULL, "vxht_0_hightown_n_by0_0_0.vlm"},
	{0xEA42AA4B59B21F9DULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_04ruinstower.msh"},
	{0xEA42C67AEB963908ULL, "plt_cod_crt_shriek.plo"},
	{0xEA43AB042F5EF0D8ULL, "vxlt_warehouse_n_8j_0_0.vlm"},
	{0xEA43C0CEA963766AULL, "233225676.ogg"},
	{0xEA43F0499685476AULL, "846477396.ogg"},
	{0xEA4438B4FA3A6010ULL, "art\\vfx\\combat\\vfx_m_cre_glyph_tar_lp_c.impa.ani"},
	{0xEA44C2FF7A5D123FULL, "vxw_bonepit_1ik_0_0.vlm"},
	{0xEA44ECA4A604E50DULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexcity.mmh"},
	{0xEA44F0A4A604EC5EULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexcity.mao"},
	{0xEA44FEA4A6050393ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexcity.msh"},
	{0xEA4613347CC67937ULL, "lla_ht_0_hightown_n_131073_1.dds"},
	{0xEA468996A913A00FULL, "vxdr_primevaltomb_7tv_0_2.vlm"},
	{0xEA46A19356A78B8BULL, "and121_karl6116252.cl"},
	{0xEA46E93B675E93B5ULL, "vxca_cave_2i9_0_0.vlm"},
	{0xEA470FD585B4C208ULL, "character_character_p100.psh.fxo.dx11"},
	{0xEA476A78922507B3ULL, "sky_lt_backalley_g.mao"},
	{0xEA4777F432BBA9D1ULL, "6178449_m.fxe"},
	{0xEA47842F0AAFEE4FULL, "hlac_dr_0_deeproads_5842_2x.dds"},
	{0xEA47D34B2D5B2083ULL, "38691862.ogg"},
	{0xEA483A2F6A912A96ULL, "383485751.ogg"},
	{0xEA48D48C27704F07ULL, "qcr000_arishok6109252.cl"},
	{0xEA4917F5CD6FB509ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.cess.ani"},
	{0xEA496101363E6ABAULL, "llac_ca_caverns1_mines_24427_1.dds"},
	{0xEA498ACE6017F540ULL, "art\\effects\\cloudshadows\\w_bonepit\\cloud1.dds"},
	{0xEA49AB68D1B546C1ULL, "vxlt_0_lowtown_a2_n_c_cmq_0_0.vlm"},
	{0xEA49D818ED146BC8ULL, "6160614_m.fxe"},
	{0xEA49EAA2D821DAC0ULL, "vxht_0_hightown_a3_bu1_0_0.vlm"},
	{0xEA4A55542521EAE2ULL, "42235127.ogg"},
	{0xEA4AC957661D3AABULL, "darktown_fem_fer_urch_sta6174783.cl"},
	{0xEA4ACEE0013A1146ULL, "723537229.ogg"},
	{0xEA4AD1383FE7124BULL, "477854628.ogg"},
	{0xEA4ADFB0E4602FBBULL, "hlac_ca_caverns1_24424_17z.dds"},
	{0xEA4B6252C354D1C6ULL, "vxw_woundedcoast_y0_0_0.vlm"},
	{0xEA4B6B44A1B30FF7ULL, "vxw_sundermount_31o_0_1.vlm"},
	{0xEA4BA2051AFEB62CULL, "6052102_m.fxe"},
	{0xEA4BA43577C3CAC9ULL, "789139585.ogg"},
	{0xEA4C3251D7BD7DADULL, "288275368.ogg"},
	{0xEA4C492B417B8119ULL, "6135767_m.fxe"},
	{0xEA4C5E121D5363ADULL, "46023200.ogg"},
	{0xEA4D29426E966734ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_diagonal_1.mao"},
	{0xEA4D51A3DC1B0AFBULL, "vxw_bonepit_21l_0_0.vlm"},
	{0xEA4DE0F20968681FULL, "hlac_lt_gallowstemplar_f_2393_1x.dds"},
	{0xEA4E13764230B4BFULL, "6030202_m.fxe"},
	{0xEA4EE351BFB63E28ULL, "gen00fl_anders6201678.cl"},
	{0xEA4EFFD546236E6AULL, "isa211_isabela6072567.cl"},
	{0xEA4F0AC464F5E62BULL, "vxlt_0_lowtown_a3_n_c_cox_0_0.vlm"},
	{0xEA4F48BE251310C4ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_viscdesk.mmh"},
	{0xEA4F4CBE25131717ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_viscdesk.mao"},
	{0xEA4F5EBE25133666ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_viscdesk.msh"},
	{0xEA4F6A3FB2CFBB7EULL, "170114813.ogg"},
	{0xEA500C53EFA16995ULL, "6172991_m.fxe"},
	{0xEA5029AFAD6B4FB3ULL, "zz_fen_debug6134195.cl"},
	{0xEA5032F5CD76018BULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.cess.gad"},
	{0xEA504FE8C7CD12D7ULL, "lt_0_lowtown_docks_131_0.mmh"},
	{0xEA5053E8C7CD19E4ULL, "lt_0_lowtown_docks_131_0.mao"},
	{0xEA5065E8C7CD3849ULL, "lt_0_lowtown_docks_131_0.msh"},
	{0xEA522B832CAF4C8DULL, "vxlt_0_lowtown_a3_n_c_fj4_0_0.vlm"},
	{0xEA525D20D7855A89ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_03a_0t.dds"},
	{0xEA529B1ADBC0DEF6ULL, "1535668.ogg"},
	{0xEA53900C03B8373DULL, "dae000cr_grand_cleric6124726.cl"},
	{0xEA539D4B59C07D68ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_04ruinstower.phy"},
	{0xEA53A01627E992F3ULL, "llac_w_blightlands_0_0.dds"},
	{0xEA53C7018CA1AF73ULL, "6126530_m.fxe"},
	{0xEA5420C196C8AB48ULL, "157466885.ogg"},
	{0xEA542A2FDABF37BCULL, "617631798.ogg"},
	{0xEA549DB361B11D1CULL, "1061189631.ogg"},
	{0xEA54CD303E926344ULL, "narrative_banter6165539.cl"},
	{0xEA55C6AFFD5A7314ULL, "vxw_encounter2_camp_qw_0_0.vlm"},
	{0xEA55E5A4A6134CCAULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_codexcity.phy"},
	{0xEA5690768EAB4E8FULL, "follower_banter6136679.cl"},
	{0xEA56BE208E7D1ABBULL, "6162272_anders_interjects.cut"},
	{0xEA572FF5CD7C1BF0ULL, "art\\vfx\\combat\\vfx_c_pride_s_start_p.cess.evt"},
	{0xEA575DD7B048743CULL, "6033540_f.fxe"},
	{0xEA5797F30513D316ULL, "vxw_encounter2_6l_0_0.vlm"},
	{0xEA57DD9DCE974457ULL, "vxlt_0_lowtown_docks_n_8pl_0_0.vlm"},
	{0xEA58415B5C040D42ULL, "820528799.ogg"},
	{0xEA5849F5DF0A8448ULL, "6165914_m.fxe"},
	{0xEA585B9CFCAF140EULL, "art\\characters\\weapons\\shields\\textures\\w_shd_kita_0d.dds"},
	{0xEA59290DCF1B529BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_exit.ani"},
	{0xEA597768060B3CE4ULL, "gen00fl_aveline6137563.cl"},
	{0xEA5990A51F7E6DF3ULL, "hm_qsw_warrens6.mor"},
	{0xEA59D42FB54AEA46ULL, "vxw_woundedcoast_v1_0_0.vlm"},
	{0xEA5A13DCA41D7E42ULL, "hlac_lt_gallowstemplar_2391_5x.dds"},
	{0xEA5A3D715AEC6CC3ULL, "vxlt_warehouse_11v_0_0.vlm"},
	{0xEA5A93AF6271A3F4ULL, "vxht_0_hightown_a7q_0_0.vlm"},
	{0xEA5AB4C5506771FEULL, "primitives_primitive_p6.psh.fxo"},
	{0xEA5AD0C9EBFEF540ULL, "vxlt_0_lowtown_a2_n_hd6_0_0.vlm"},
	{0xEA5AD11E8DA23D8BULL, "vxht_keep_640_0_0.vlm"},
	{0xEA5ADD77812031BDULL, "vxht_bartrand_n_5s0_0_0.vlm"},
	{0xEA5AF0A6DF248862ULL, "310772158.ogg"},
	{0xEA5AF172FEEE7E8AULL, "712314725.ogg"},
	{0xEA5B28F46891A40EULL, "6105878_f.fxe"},
	{0xEA5B98921416B219ULL, "w_encounter2_6_0.phy"},
	{0xEA5BE8D6432EA3DCULL, "follower_banter6169744.cl"},
	{0xEA5C80DC282D051DULL, "vxht_mansion_n_6bs_0_0.vlm"},
	{0xEA5C9664352FF714ULL, "zz_followers_banter6158658.cl"},
	{0xEA5D2175BACBFF43ULL, "hlac_lt_0_lowtown_-65537_4y.dds"},
	{0xEA5D6C26667FC7B2ULL, "61007554.ogg"},
	{0xEA5E16A1F3C81D6FULL, "hlac_lt_0_lowtown_a2_-196609_3y.dds"},
	{0xEA5E5048FC3D48A1ULL, "gen00fl_dog6165745.cl"},
	{0xEA5E57861AF0E07CULL, "266372424.ogg"},
	{0xEA5E64C5C426B526ULL, "hlac_w_encounter1_-131069_0z.dds"},
	{0xEA5E9F2D5EF43846ULL, "llac_lt_foundry_n_320_14.dds"},
	{0xEA5EB6D254903765ULL, "6068818_m.fxe"},
	{0xEA5EBD352B16D85CULL, "follower_banter6171082.cl"},
	{0xEA5F02142F2CCE5BULL, "vxca_cave_2e3_0_0.vlm"},
	{0xEA5F706525A6DD61ULL, "237726314.ogg"},
	{0xEA5FAE25C6090C8FULL, "hlac_lt_house_-1_1x.dds"},
	{0xEA5FB242245A3254ULL, "vxht_0_hightown_c1e_0_0.vlm"},
	{0xEA601EE1EE9965AEULL, "vxht_0_hightown_n_c6a_0_0.vlm"},
	{0xEA608F9CB8E751C9ULL, "vxlt_0_lowtown_n_elj_0_0.vlm"},
	{0xEA60A3BCDD769DEBULL, "6094459_m.fxe"},
	{0xEA61661DC524D4CAULL, "hlac_dr_0_deeproads_5841_2x.dds"},
	{0xEA618EF30A22034AULL, "vxht_bartrand_n_5pj_0_0.vlm"},
	{0xEA61956B2898D55AULL, "hlac_lt_0_lowtown_1_2x.dds"},
	{0xEA625D048955580DULL, "vxca_caverns1_i1f_0_0.vlm"},
	{0xEA627D9788A57DF4ULL, "darktown_fem_elf_sta6174512.cl"},
	{0xEA62A192141CE035ULL, "w_encounter2_6_0.rml"},
	{0xEA62DD6BC96FAE79ULL, "hlac_lt_darktown_-1_89z.dds"},
	{0xEA63859B0FCC1C66ULL, "6136358_m.fxe"},
	{0xEA63C67546940E41ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldb.phy"},
	{0xEA63D35F1C3E25B4ULL, "97541503.ogg"},
	{0xEA63F0C3360D6A71ULL, "6013298_m.fxe"},
	{0xEA63F52F808A56DEULL, "vxlt_0_lowtown_a3_n_c_hof_0_0.vlm"},
	{0xEA64388CEF508B94ULL, "vxw_encounter2_camp_lt_0_0.vlm"},
	{0xEA643AB0DEF7C9A6ULL, "statecnt_areatransition.gda"},
	{0xEA64410AC6B87485ULL, "gen00fl_aveline6113672.cl"},
	{0xEA6498C71456D16BULL, "6154910_f.fxe"},
	{0xEA64BA898AC7AFD6ULL, "590376097.ogg"},
	{0xEA65141ECEE12817ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_7z.dds"},
	{0xEA6534291B744911ULL, "hlac_lt_0_lowtown_a2_n_c_131073_0x.dds"},
	{0xEA66310DCF25EA92ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_exit.evt"},
	{0xEA66AF26324C70EBULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.dura.evt"},
	{0xEA66DD56270B9EA1ULL, "235588686.ogg"},
	{0xEA66E2FE32155F20ULL, "vxht_bartrand_n_68s_0_0.vlm"},
	{0xEA67C01D62E979C3ULL, "dae150_harshal6148857.cl"},
	{0xEA67C1C8D7A210BFULL, "754293359.ogg"},
	{0xEA67C5BD92410748ULL, "458809691.ogg"},
	{0xEA68460CA74102C6ULL, "vxw_woundedcoast_s2_0_0.vlm"},
	{0xEA6867BEAA95BAD2ULL, "vxlt_gallowstemplar_4xx_0_0.vlm"},
	{0xEA68A1F298EB18F8ULL, "785231066.ogg"},
	{0xEA68AF4E4CE28543ULL, "vxlt_warehouse_14y_0_0.vlm"},
	{0xEA6942FB7F98560BULL, "vxht_keep_613_0_0.vlm"},
	{0xEA69C88FF49175F1ULL, "art\\characters\\playercharacter\\humanfemale\\hat_isab.mao"},
	{0xEA6A2EBB1D518493ULL, "genip_exploding_barrel.utp"},
	{0xEA6B2F3CB6B714D0ULL, "6108263_m.fxe"},
	{0xEA6B3967B70D49D5ULL, "187638785.ogg"},
	{0xEA6B5F56E36AB1C4ULL, "art\\vfx\\combat\\vfx_r_dwp_critical_strike_c.mmh"},
	{0xEA6BE1F3012D88D9ULL, "1058759963.ogg"},
	{0xEA6C2B1D0FC89677ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04a.phy"},
	{0xEA6CC826325106ACULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.dura.gad"},
	{0xEA6CE3C1B5C8CA6BULL, "vxdr_primevaltomb_8py_0_2.vlm"},
	{0xEA6D500DCF2C3E25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\draw_sword_threaten_exit.gad"},
	{0xEA6DB3269E5AB413ULL, "vxlt_warehouse_sm_0_0.vlm"},
	{0xEA6DFE55FE3A965AULL, "6119713_m.fxe"},
	{0xEA6E90BEE08F7E2EULL, "vxht_0_hightown_n_c3b_0_0.vlm"},
	{0xEA6EE85266D47DFAULL, "6082945_m.fxe"},
	{0xEA6F3FF911D6C0F5ULL, "hlac_lt_backalley_n_0_17z.dds"},
	{0xEA6F610D5A494738ULL, "one300_zevran6076919.cl"},
	{0xEA6FE1B886B1D139ULL, "235797486.ogg"},
	{0xEA70052F2997F55CULL, "mag110tr_arianni.utt"},
	{0xEA707F4DB04D4612ULL, "96371862.ogg"},
	{0xEA70F4B7E07882F5ULL, "vxlt_gallowstemplar_n_350_0_0.vlm"},
	{0xEA71015862BC10A1ULL, "6142048_m.fxe"},
	{0xEA714D3FF86E8A49ULL, "and111cr_shopgirl.utc"},
	{0xEA71CD114E6270FBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_outinr_1.mao"},
	{0xEA71F315015D001FULL, "vxw_sundermount_2ka_0_0.vlm"},
	{0xEA728284D2CFF418ULL, "6174372_m.fxe"},
	{0xEA72A9739F42A3D8ULL, "6067902_m.fxe"},
	{0xEA72AD97BF609C4DULL, "lt_0_lowtown_n_53_0_ll.mmh"},
	{0xEA72F4FEA7B7A99EULL, "vxlt_gallowscourtyard_778_0_0.vlm"},
	{0xEA73191A54940637ULL, "plt_isa100pt_fools.plo"},
	{0xEA73B05E5CEC5356ULL, "w_woundedcoast_146_0.msh"},
	{0xEA73BA5E5CEC6454ULL, "w_woundedcoast_146_0.mmh"},
	{0xEA73BE5E5CEC6B67ULL, "w_woundedcoast_146_0.mao"},
	{0xEA73C726325723AAULL, "art\\vfx\\combat\\vfx_r_asn_pin_stk_lp_c.dura.ani"},
	{0xEA73F21B51D936D5ULL, "vxlt_gallowstemplar_34p_0_0.vlm"},
	{0xEA74A94D6E830264ULL, "6120834_m.fxe"},
	{0xEA74CF0913443896ULL, "vxw_blightlands_fc_0_0.vlm"},
	{0xEA755511083EC2D1ULL, "llac_ht_bartrand_n_3367_1.dds"},
	{0xEA755C6330DED2BCULL, "6133799_m.fxe"},
	{0xEA75ACCDBD08EF4EULL, "arm_pl2.tnt"},
	{0xEA75B7172488332DULL, "170241431.ogg"},
	{0xEA75EDEEDEC9716CULL, "gst_sitting_table_clasped_1.gda"},
	{0xEA7689C8D5BCEF32ULL, "art\\levels\\races\\ferelden\\props\\_textures\\tarp_n.dds"},
	{0xEA769B7F6A61271CULL, "6137632_m.fxe"},
	{0xEA76B19069F1E24DULL, "m_mcr310ar_docks_nt.bnk"},
	{0xEA76FDAAE79E4C70ULL, "444276182.ogg"},
	{0xEA770C03E4CADC9FULL, "hightown_male_noble_gossip6171432.cl"},
	{0xEA775F56395A43E5ULL, "6149193_m.fxe"},
	{0xEA77BCDE8B1AC5CCULL, "vxht_mansion2_2k2_0_0.vlm"},
	{0xEA77F22DDEBA69BFULL, "ave100_donnic6105992.cl"},
	{0xEA78408D1A11BDA9ULL, "gen00fl_anders6125326.cl"},
	{0xEA786BA439055EB6ULL, "llac_lt_0_lowtown_a2_n_c_131070_0.dds"},
	{0xEA789DAE8D90D105ULL, "lgt110_orz_harrowmont6108991.cl"},
	{0xEA792C426813C2C3ULL, "6045968_f.fxe"},
	{0xEA797334A85AD957ULL, "vxlt_gallowstemplar_5cm_0_0.vlm"},
	{0xEA799C30D4FEC276ULL, "hlac_ca_caverns1_24425_6z.dds"},
	{0xEA7A3B0EE8CC585DULL, "vxlt_gallowsprison_n_6f0_0_0.vlm"},
	{0xEA7A4704D91A93ABULL, "hlac_w_bonepit_0_7z.dds"},
	{0xEA7A8AC9FDC94582ULL, "973189764.ogg"},
	{0xEA7AED07169681B4ULL, "lowtown_male_alienage1_sta6162723.cl"},
	{0xEA7AFBF616319BFCULL, "art\\levels\\4cinematics\\ferelden\\_textures\\w_small_dagger_0_d.dds"},
	{0xEA7C3D292F8A587FULL, "and111_lirene6090213.cl"},
	{0xEA7C5F670D2706C9ULL, "lt_0_lowtown_n_53_0.rml"},
	{0xEA7C630F8F6345B1ULL, "884140446.ogg"},
	{0xEA7C69D099ECA596ULL, "9208719.ogg"},
	{0xEA7D029BD28596AEULL, "vxht_0_hightown_n_c8c_0_0.vlm"},
	{0xEA7D14D9083A950BULL, "hlac_ht_mansion2b_818_30x.dds"},
	{0xEA7D173FA836B921ULL, "vxht_0_hightown_a3_bm4_0_0.vlm"},
	{0xEA7D2A2240D1F13BULL, "436920603.ogg"},
	{0xEA7D6E60D80B373BULL, "hlac_lt_0_lowtown_docks_a3_131071_5x.dds"},
	{0xEA7DBC65C9CDC107ULL, "6084695_f.fxe"},
	{0xEA7DF9CAA1EA4476ULL, "dae000cr_pickpocket.utc"},
	{0xEA7E30D0CD32021BULL, "6105537_m.fxe"},
	{0xEA7E8D9A53EC19BAULL, "6039006_m.fxe"},
	{0xEA7F27AD8351BB75ULL, "190385324.ogg"},
	{0xEA7FDF7D2A80788FULL, "vxw_bonepitdestroyed_a3_29h_0_0.vlm"},
	{0xEA7FE0CC188E6E69ULL, "132038431.ogg"},
	{0xEA80353F6E9B3893ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetal_d.dds"},
	{0xEA8049BC5E93B6B3ULL, "hlac_lt_0_lowtown_a2_n_65536_9y.dds"},
	{0xEA805553607A9B45ULL, "llac_w_sundermount_262138_0.dds"},
	{0xEA80AD1DA3C389F9ULL, "6173404_m.fxe"},
	{0xEA80B69FEE3B601DULL, "941986963.ogg"},
	{0xEA80BB16A2064540ULL, "gen00fl_varric6107389.cl"},
	{0xEA80C706E7DB3103ULL, "6144332_m.fxe"},
	{0xEA80E2166E9E8EDAULL, "pro000_colonel_ewald6040550.cl"},
	{0xEA8166DB99ADC21EULL, "vxlt_gallowscourtyard_789_0_0.vlm"},
	{0xEA81DDE5238B4E81ULL, "vxht_mansion2b_30w_0_0.vlm"},
	{0xEA829B684CE1C2B7ULL, "art\\characters\\heraldry\\hld_templars.dds"},
	{0xEA8342C1FF815915ULL, "6055418_m.fxe"},
	{0xEA8445532B0F9E74ULL, "vxlt_gallowstemplar_n_3gc_0_0.vlm"},
	{0xEA84597B0896AE99ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar3_.ani"},
	{0xEA8462D72120D414ULL, "876410165.ogg"},
	{0xEA8491401EEB705CULL, "363706562.ogg"},
	{0xEA849419362920A3ULL, "art\\characters\\weapons\\spears\\w_qui_qspr03a.msh"},
	{0xEA84A219362938FDULL, "art\\characters\\weapons\\spears\\w_qui_qspr03a.mmh"},
	{0xEA84BE170AB0CDACULL, "6103373_f.fxe"},
	{0xEA853E065CD2B2FBULL, "171811044.ogg"},
	{0xEA857D1A75EAA4D7ULL, "w_blightlands_181_0.msh"},
	{0xEA85821C41CD8FD6ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_wine.evt"},
	{0xEA858F1A75EAC33AULL, "w_blightlands_181_0.mao"},
	{0xEA85931A75EACA49ULL, "w_blightlands_181_0.mmh"},
	{0xEA85B5B2FCC5CC38ULL, "vxca_cave_2ba_0_0.vlm"},
	{0xEA85BECA4BDBA8B3ULL, "vxw_encounter1_d2_0_0.vlm"},
	{0xEA85D1332B505C65ULL, "6142093_m.fxe"},
	{0xEA86276A8B72F728ULL, "hlac_lt_warehouse_0_8y.dds"},
	{0xEA86F10DD9B03DE7ULL, "vxca_caverns1_ihc_0_0.vlm"},
	{0xEA8711B9EAF48744ULL, "215694168.ogg"},
	{0xEA87281A54A4C8ACULL, "plt_isa100pt_fools.nss"},
	{0xEA87887F8627234EULL, "art\\characters\\creatures\\spider\\animation\\wkf.evt"},
	{0xEA87A98F8D406DB0ULL, "vxht_keep_6c6_0_0.vlm"},
	{0xEA87F8B737488B12ULL, "6174716_m.fxe"},
	{0xEA88636441BB320EULL, "261413591.ogg"},
	{0xEA886C728D5C0B89ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb107a_0d.dds"},
	{0xEA8937D370CAE072ULL, "6170473_m.fxe"},
	{0xEA895F988DF0D31DULL, "732223596.ogg"},
	{0xEA89A5A54E0A9B96ULL, "6172839_m.fxe"},
	{0xEA89D46A8995B114ULL, "vxlt_0_lowtown_a2_n_c_ekf_0_2.vlm"},
	{0xEA8A1746EA7089A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_p.ani"},
	{0xEA8A96E08246E513ULL, "vxlt_warehouse_yo_0_0.vlm"},
	{0xEA8A9AEA66AED998ULL, "vxca_cavern1_haunted_hm7_0_0.vlm"},
	{0xEA8ABB11AF3EAA71ULL, "hm_vis_hightown3.mor"},
	{0xEA8AC606BB50786FULL, "dae300cr_brothel_gamlen.utc"},
	{0xEA8AE20FE226C75AULL, "6113113_m.fxe"},
	{0xEA8B21ACC065AAF7ULL, "vxlt_darktown_5gm_0_0.vlm"},
	{0xEA8B347B089C8EDBULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\templar3_.gad"},
	{0xEA8BB892B53B3492ULL, "vxht_mansion_n_6fd_0_0.vlm"},
	{0xEA8BD43FC4722F02ULL, "zz_isa_debug6133230.cl"},
	{0xEA8BDF8E49B005BCULL, "mag211_marethari6025145.cl"},
	{0xEA8BE3F0E82BB0E8ULL, "zz_followers_banter6157747.cl"},
	{0xEA8C1012E836ACD4ULL, "hlac_dr_0_deeproads_5849_11x.dds"},
	{0xEA8C791C41D39F61ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_wine.gad"},
	{0xEA8CC24BD3CD3485ULL, "6149367_f.fxe"},
	{0xEA8CE489E0044CCAULL, "vxht_bartrand_n_5ae_0_0.vlm"},
	{0xEA8D345FF6101867ULL, "vxht_0_hightown_n_bq0_0_0.vlm"},
	{0xEA8E42AB8A25DD3DULL, "vxht_keep_4bf_0_0.vlm"},
	{0xEA8E49D3933CA5D3ULL, "vxht_0_hightown_a3_apr_0_0.vlm"},
	{0xEA8E515A1C76910FULL, "vxw_bonepitdestroyed_a3_24k_0_0.vlm"},
	{0xEA8E64C7D35A0800ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingb.phy"},
	{0xEA8E7F7F862D32F9ULL, "art\\characters\\creatures\\spider\\animation\\wkf.gad"},
	{0xEA8EC8BA2841B55CULL, "560538564.ogg"},
	{0xEA8EEC914E5F8984ULL, "hlac_dr_primevaltomb_659_11z.dds"},
	{0xEA8F48CDC4697563ULL, "follower_banter6171168.cl"},
	{0xEA8FCD5C1322CF0CULL, "vxlt_undercityencounter_1jw_0_0.vlm"},
	{0xEA9001600E12334EULL, "137720622.ogg"},
	{0xEA90BA64B55F7500ULL, "vxlt_0_lowtown_a2_n_ehw_0_0.vlm"},
	{0xEA90C456EE29A95AULL, "hlac_ca_caverns2prc_5802_14y.dds"},
	{0xEA91482BD92437C5ULL, "6092172_m.fxe"},
	{0xEA9191D753D5A667ULL, "art\\characters\\creatures\\gallowsslavestatue\\textures\\cn_bdy_css_0n.dds"},
	{0xEA91B2C2F7306996ULL, "vxw_blightlands_le_0_0.vlm"},
	{0xEA91F82041560D81ULL, "6113990_m.fxe"},
	{0xEA9265F6C82D422DULL, "577077307.ogg"},
	{0xEA9274BBB59AAC98ULL, "vxht_brothel_n_57l_0_1.vlm"},
	{0xEA927A1C41D80C0FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_wine.ani"},
	{0xEA9299C741DA655CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_03fern.mao"},
	{0xEA929FBEF6386628ULL, "223217508.ogg"},
	{0xEA92A5C741DA79FFULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_03fern.mmh"},
	{0xEA92ABC741DA83A1ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_03fern.msh"},
	{0xEA92BEF0AC11E981ULL, "6099083_m.fxe"},
	{0xEA937E8B55D0275DULL, "vxht_mansion_n_4y0_0_0.vlm"},
	{0xEA93E1007FDD84D8ULL, "6068507_m.fxe"},
	{0xEA9430A73DD1C133ULL, "vxw_encounter1_i5_0_0.vlm"},
	{0xEA946256E38DBBBBULL, "art\\vfx\\combat\\vfx_r_dwp_critical_strike_c.anb"},
	{0xEA949060E87872B3ULL, "521135240.ogg"},
	{0xEA94907F8631BB47ULL, "art\\characters\\creatures\\spider\\animation\\wkf.ani"},
	{0xEA94A0986F06F6CCULL, "vxht_mansion2_2q0_0_0.vlm"},
	{0xEA9529CB1EF64727ULL, "zz_and_debug6178568.cl"},
	{0xEA952F393CD13A13ULL, "pro000_athenril6049449.cl"},
	{0xEA95B2609EFE4F50ULL, "hlac_lt_darktown_-1_54y.dds"},
	{0xEA95E89F7D0E33C6ULL, "vxca_cavern1_haunted_jtl_0_0.vlm"},
	{0xEA9645490CA0D3DEULL, "6117089_m.fxe"},
	{0xEA96545AE0359BA7ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_medj_0d.dds"},
	{0xEA96AE2D06B267E5ULL, "vxlt_0_lowtown_a2_n_hcn_0_0.vlm"},
	{0xEA96FF46EA7AEB17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_p.evt"},
	{0xEA97466E3BF595D0ULL, "6162667_m.fxe"},
	{0xEA97C463CBF1592FULL, "711524977.ogg"},
	{0xEA97E20151AFAD38ULL, "gen_im_trs_dem_junk2_01.uti"},
	{0xEA97E567030FD3E0ULL, "mag121_viveka6082585.cl"},
	{0xEA98748F7AC87F4BULL, "dae000_meredith6092599.cl"},
	{0xEA98CF9F6D47B30DULL, "hlac_w_encounter2_-131069_0x.dds"},
	{0xEA98DAB939BC894DULL, "gen00fl_anders6125274.cl"},
	{0xEA991648949A2D76ULL, "llac_lt_0_lowtown_-65537_2.dds"},
	{0xEA99185387A77275ULL, "vxlt_0_lowtown_docks_n_b7y_0_0.vlm"},
	{0xEA998F17DB8F7685ULL, "gen_im_wep_war_two_act1_01.uti"},
	{0xEA99BBF2DAA67FE7ULL, "vxdr_0_deeproads_cwb_0_1.vlm"},
	{0xEA9A2962B3B1AF56ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_010a_0s.dds"},
	{0xEA9A3140F74642CBULL, "prop1_ds1y.psh.fxo"},
	{0xEA9B8B6ECF2D458FULL, "dae_rogue_aoe_duration.ncs"},
	{0xEA9BBB5148B6DBF3ULL, "lowtown_fem_prostitute_sta6152364.cl"},
	{0xEA9BC203AD50412BULL, "dm_bodahn_old.mor"},
	{0xEA9C25A184F5C270ULL, "hlac_dr_primevaltomb_661_5z.dds"},
	{0xEA9C7F5D9117138AULL, "6104957_m.fxe"},
	{0xEA9CA3C29297BAE5ULL, "llac_lt_0_lowtown_docks_-196608_0.dds"},
	{0xEA9D48ABD73F499FULL, "vxw_sundermount_2zb_0_0.vlm"},
	{0xEA9E0046EA810BA0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\movement\\mov_p_r90_p.gad"},
	{0xEA9E01705DD6458AULL, "jp_traning_stage.stg"},
	{0xEA9E53F8CC540736ULL, "s_proj_drian.mao"},
	{0xEA9EBB7D1C16CAB9ULL, "vxlt_hangedman_sc_0_0.vlm"},
	{0xEA9EDD57B6E623F2ULL, "383919704.ogg"},
	{0xEA9F2C41A7558D80ULL, "vxlt_0_lowtown_a2_n_emt_0_0.vlm"},
	{0xEA9F531E38FC213DULL, "lowtown_male_lowlife1_sta6165042.cl"},
	{0xEA9F7A670686DE4DULL, "667421964.ogg"},
	{0xEA9FB8E3C561960AULL, "ht_0_hightown_ll.mmh"},
	{0xEA9FD2320E283896ULL, "6148557_m.fxe"},
	{0xEAA050DA61BF7DA6ULL, "one300_murder_followers6086894.cl"},
	{0xEAA0CF1D276CC068ULL, "m_dae221ar_keep.bnk"},
	{0xEAA0E195E6B839A8ULL, "976466631.ogg"},
	{0xEAA19D9DB6C6E58DULL, "cod_ltr_sk3_scribbles2.utp"},
	{0xEAA21491E811E98BULL, "747134959.ogg"},
	{0xEAA2226397884339ULL, "pro000cs_opening_scene3.cut"},
	{0xEAA2385AB18D33C2ULL, "hlac_lt_gallowstemplar_f_7198_4x.dds"},
	{0xEAA23DDE592E571DULL, "305093843.ogg"},
	{0xEAA2A7906A170C77ULL, "m_mcr310ar_docks_nt.wbd"},
	{0xEAA2EF669817B2DFULL, "vxht_mansion2b_2v2_0_0.vlm"},
	{0xEAA3356C20491466ULL, "hlac_lt_0_lowtown_a2_n_c_65534_5x.dds"},
	{0xEAA3CBA9FAF391C9ULL, "796396917.ogg"},
	{0xEAA3FF1063606057ULL, "75081374.ogg"},
	{0xEAA426D781C8AE75ULL, "vxca_caverns1_mines_ckk_0_0.vlm"},
	{0xEAA427F5827DEF55ULL, "2725546.ogg"},
	{0xEAA459B2D4F54668ULL, "vxw_bonepitmining_a2_cc_0_0.vlm"},
	{0xEAA45A9257085E23ULL, "isa211_isabela6069414.cl"},
	{0xEAA45BD15003F3DCULL, "vxht_0_hightown_n_90c_0_0.vlm"},
	{0xEAA48D49712C9EB0ULL, "vxht_keep_6i4_0_0.vlm"},
	{0xEAA4C8CB7E3D22D7ULL, "vxlt_gallowstemplar_5rn_0_0.vlm"},
	{0xEAA4D6AC9D1A0A67ULL, "vxht_keep_6h8_0_0.vlm"},
	{0xEAA4DD08B8E15F71ULL, "vxlt_0_lowtown_a2_n_f4u_0_0.vlm"},
	{0xEAA504456B26F903ULL, "vxht_mansion2b_1ud_0_0.vlm"},
	{0xEAA524386D52DB0AULL, "hlac_lt_0_lowtown_docks_a3_n_65534_2y.dds"},
	{0xEAA546D300CACE61ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_loop.ani"},
	{0xEAA54BFACFFDA623ULL, "128001623.ogg"},
	{0xEAA598CD00579162ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_end.gad"},
	{0xEAA5B097AEFAB52FULL, "vxw_encounter2_camp_p6_0_1.vlm"},
	{0xEAA654EF7F76F464ULL, "ser221_gamlen6131885.cl"},
	{0xEAA6643CC3516BFCULL, "vxlt_gallowscourtyard_n_62c_0_0.vlm"},
	{0xEAA66AE4FE53A183ULL, "dae000_epic_merchant_46179518.cl"},
	{0xEAA6A811F0C2AD0DULL, "hlac_lt_0_lowtown_docks_n_-131074_4y.dds"},
	{0xEAA6E6D2718C97E0ULL, "575539709.ogg"},
	{0xEAA7275B0BD738C1ULL, "llac_w_encounter2_camp_-131070_0.dds"},
	{0xEAA7E0BA53BF1A31ULL, "hlac_w_encounter3_-196608_1z.dds"},
	{0xEAA8420A652C11F2ULL, "one300cr_lia.utc"},
	{0xEAA87724CAA74492ULL, "art\\vfx\\materials\\vfx_dust_add.mao"},
	{0xEAA89C4C99276592ULL, "vxht_mansion_n_6lf_0_0.vlm"},
	{0xEAA8A2DBF8A73DDFULL, "w_encounter3_224_0.msh"},
	{0xEAA8A8DBF8A74781ULL, "w_encounter3_224_0.mmh"},
	{0xEAA8B4DBF8A75C22ULL, "w_encounter3_224_0.mao"},
	{0xEAA952EA2562F199ULL, "qun110_petrice6056801.cl"},
	{0xEAA98198C888427CULL, "hm_hunter_captain.mor"},
	{0xEAA98E7367928074ULL, "6164979_m.fxe"},
	{0xEAA9A5DAEFB4CD36ULL, "vxht_0_hightown_a3_byc_0_0.vlm"},
	{0xEAA9BCEDA485FAA0ULL, "649007796.ogg"},
	{0xEAA9C843C3F07DCAULL, "vxht_bartrand_n_5og_0_0.vlm"},
	{0xEAAA1819D9FC4967ULL, "vxht_0_hightown_n_bg2_0_0.vlm"},
	{0xEAAA29E47F0618C1ULL, "vxlt_backalley_g_1ky_0_0.vlm"},
	{0xEAAA92B744709822ULL, "hlac_lt_0_lowtown_a2_n_c_-1_1y.dds"},
	{0xEAAAD88335B55915ULL, "50188546.ogg"},
	{0xEAAB4F42971285C8ULL, "zz_mcr_debug6071539.cl"},
	{0xEAAB5E060C369CF2ULL, "926186455.ogg"},
	{0xEAAB72C6776C5F14ULL, "868299073.ogg"},
	{0xEAAB737933630030ULL, "620291804.ogg"},
	{0xEAABBBCD005C3859ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_end.evt"},
	{0xEAAC51D300D0FF03ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_action_loop.gad"},
	{0xEAACD4DB881E2BD2ULL, "300655702.ogg"},
	{0xEAACDFC578E38871ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_pmtb_0t.dds"},
	{0xEAADCE87EA8932D3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlm_masf_0n.dds"},
	{0xEAAECD6D01F84C15ULL, "gen_im_trs_krk_grd_junk2_04.uti"},
	{0xEAAF5E670D525EE9ULL, "lt_0_lowtown_n_53_0.anb"},
	{0xEAB050C26AB1F504ULL, "6196732_m.fxe"},
	{0xEAB0A9AF666D704BULL, "6119271_m.fxe"},
	{0xEAB104A5AA3242D4ULL, "zz_qcr_debug6073542.cl"},
	{0xEAB28E12971A0CC3ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.dura.evt"},
	{0xEAB298B473BEC6F5ULL, "vxca_caverns1_mines_clj_0_0.vlm"},
	{0xEAB381C7D3795595ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingb.msh"},
	{0xEAB381C9A57174A2ULL, "389736871.ogg"},
	{0xEAB383C7D379590BULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingb.mmh"},
	{0xEAB38FC7D3796DA8ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_hangingb.mao"},
	{0xEAB3B6CB5274D942ULL, "plt_prm_cod_blood_dragon.plo"},
	{0xEAB3CD4B89D49DF3ULL, "901014473.ogg"},
	{0xEAB40C1D1006176CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04a.mmh"},
	{0xEAB4121D1006215EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear04a.msh"},
	{0xEAB4725775CA05C4ULL, "dae000cr_varric_cuts.utc"},
	{0xEAB475CB36048F6FULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_c.mmh"},
	{0xEAB53E3A242B9CECULL, "946527007.ogg"},
	{0xEAB5B25533EBC9F2ULL, "hlac_lt_gallowstemplar_f_2383_2y.dds"},
	{0xEAB5C5AAAFE255FDULL, "1039765750.ogg"},
	{0xEAB5FF4647D0DC97ULL, "gen00fl_anders6125148.cl"},
	{0xEAB60A7F4BAFD88AULL, "6176725_m.fxe"},
	{0xEAB680B5341BB5F9ULL, "vxlt_hangedman_1ul_0_0.vlm"},
	{0xEAB694FAA8575060ULL, "ave200_donnic6110615.cl"},
	{0xEAB6C32E52AA7487ULL, "vxw_encounter1_ck_0_0.vlm"},
	{0xEAB6CBE55903DB87ULL, "vxlt_hangedman_n_rw_0_0.vlm"},
	{0xEAB6D923D4A2DD5CULL, "6103657_m.fxe"},
	{0xEAB83BC978806A18ULL, "vxca_caverns2prc_adj_0_0.vlm"},
	{0xEAB8A953E6E6245FULL, "vxht_bartrand_n_63e_0_0.vlm"},
	{0xEAB92792938F4529ULL, "955825649.ogg"},
	{0xEAB943B904E423BBULL, "391216599.ogg"},
	{0xEAB964BFE6552FB6ULL, "gen00fl_varric6105424.cl"},
	{0xEAB96712971FE944ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.dura.gad"},
	{0xEAB96CC0726D61A2ULL, "6117671_f.fxe"},
	{0xEAB97E7C4E3A57D2ULL, "733931901.ogg"},
	{0xEAB988F4A80BCF68ULL, "hlac_lt_0_lowtown_docks_n_65533_2y.dds"},
	{0xEAB9C3CD00688348ULL, "art\\characters\\creatures\\desiredemon\\animation\\a_end.ani"},
	{0xEABA31227B0D8B5DULL, "vxlt_gallowstemplar_5qe_0_0.vlm"},
	{0xEABA585FB581C7B3ULL, "hlac_ca_cave_2194_10y.dds"},
	{0xEABACC7E51C45FD9ULL, "mrl150_dalish_guard6068736.cl"},
	{0xEABBF99AAB8C444DULL, "follower_banter6161329.cl"},
	{0xEABCC66A51251E1EULL, "6131654_m.fxe"},
	{0xEABD0C97CE8737A9ULL, "gencr_mercenary_warrior_2h.utc"},
	{0xEABD6C2A4AAE63F9ULL, "vxw_bonepitdestroyed_a3_274_0_0.vlm"},
	{0xEABDA65B2CFF13CEULL, "865417562.ogg"},
	{0xEABE0CC6F2E80074ULL, "vxlt_gallowstemplar_n_3sg_0_0.vlm"},
	{0xEABE5F27AAC900A5ULL, "vxht_store_c1_0_0.vlm"},
	{0xEABE7DA0631CCC49ULL, "6160844_m.fxe"},
	{0xEABF04B97738DDE6ULL, "746774885.ogg"},
	{0xEABF09846BAAF725ULL, "w_bonepitmining_a2_8_0.rml"},
	{0xEABF6B1470897A79ULL, "19792821.ogg"},
	{0xEABF6EC66629E030ULL, "6125698_f.fxe"},
	{0xEABF761297246E22ULL, "art\\vfx\\combat\\vfx_m_cre_haste_lp_c.dura.ani"},
	{0xEABF9A3A002C7DE6ULL, "qun211_braggart6086737.cl"},
	{0xEABFE92483707F84ULL, "60731434.ogg"},
	{0xEAC0AA73C381962FULL, "6028531_m.fxe"},
	{0xEAC0DCE32C7FAE20ULL, "one140cr_charade6134579.cl"},
	{0xEAC14E2FE4CDEBA6ULL, "148762477.ogg"},
	{0xEAC1AC85622953D7ULL, "vxlt_gallowstemplar_5xh_0_0.vlm"},
	{0xEAC1F2136F54D583ULL, "hlac_ht_mansion2b_818_9z.dds"},
	{0xEAC28A6AD3301640ULL, "vxca_cavern1_haunted_imj_0_0.vlm"},
	{0xEAC28D8F95468F07ULL, "vxca_caverns1_mines_e91_0_0.vlm"},
	{0xEAC33C90659DAAB4ULL, "6208451_m.fxe"},
	{0xEAC35991B1305D91ULL, "6050185_m.fxe"},
	{0xEAC3A28D1C7F8CE0ULL, "vxca_caverns2_7w4_0_0.vlm"},
	{0xEAC3AA907C5A2270ULL, "6089497_f.fxe"},
	{0xEAC3DD383560AFDAULL, "6099541_m.fxe"},
	{0xEAC40F021B41074DULL, "fx_firetwister_blend.mao"},
	{0xEAC48A6B599718EFULL, "628894235.ogg"},
	{0xEAC4B38BC297FB71ULL, "vxlt_darktown_2xd_0_1.vlm"},
	{0xEAC5350B44A08D07ULL, "vxw_encounter1_hj_0_0.vlm"},
	{0xEAC538DECD36F18BULL, "art\\characters\\weapons\\axes\\textures\\w_axe_02a_0s.dds"},
	{0xEAC54321D407DD53ULL, "mcr371_angry_mage_m6087734.cl"},
	{0xEAC5447E20130E09ULL, "6115787_m.fxe"},
	{0xEAC5982F09F3225BULL, "6126362_f.fxe"},
	{0xEAC5C11D278BC412ULL, "m_dae221ar_keep.wbd"},
	{0xEAC5CB7546E6E17DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldb.mao"},
	{0xEAC5CF7546E6E7B6ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldb.mmh"},
	{0xEAC5D97546E6F8B4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_scaffoldb.msh"},
	{0xEAC5EFDE0292D7C7ULL, "498516314.ogg"},
	{0xEAC643CFF6DFBA76ULL, "gen_im_trs_gen_sta_fir_r1.uti"},
	{0xEAC6AB288C710CAAULL, "6060740_f.fxe"},
	{0xEAC6ADA66A768298ULL, "vxca_caverns2prc_ack_0_0.vlm"},
	{0xEAC6B5800A5A139EULL, "6101136_m.fxe"},
	{0xEAC6CCB15244D024ULL, "6018584_m.fxe"},
	{0xEAC75B5154EB1C16ULL, "330176824.ogg"},
	{0xEAC75D9166D98BFFULL, "w_woundedcoast_155_0.phy"},
	{0xEAC789C5C5EF43EAULL, "art\\levels\\races\\ferelden\\props\\_textures\\tarp_sa.dds"},
	{0xEAC82F888D1D9EC5ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue_n.dds"},
	{0xEAC913719870A4F3ULL, "6140264_m.fxe"},
	{0xEAC922CC491B9003ULL, "rdr111dg_harbormaster6055079.cl"},
	{0xEAC9B271EE9D1A5FULL, "art\\characters\\weapons\\shields\\textures\\w_shd_dssa_0s.dds"},
	{0xEACB2CF87FA0E70AULL, "hlac_w_encounter2_-65538_0z.dds"},
	{0xEACB6ACE30FE7AC2ULL, "dae000_keep_guard_2_sta6163808.cl"},
	{0xEACBA2338BA8D643ULL, "dae200_bran6191228.cl"},
	{0xEACBB0E5EF80376BULL, "vxlt_gallowstemplar_f_3rc_0_0.vlm"},
	{0xEACBFC2F29E644F9ULL, "mag110tr_arianni.ncs"},
	{0xEACC69E559530BA5ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_003a_0n.dds"},
	{0xEACCFDEA9D5FEC7DULL, "lowtown_male_urchin_sta6161590.cl"},
	{0xEACCFF3511BF4DBCULL, "6060262_m.fxe"},
	{0xEACD3C79E237F825ULL, "vxlt_0_lowtown_docks_a3_n_coc_0_0.vlm"},
	{0xEACD8E563DC00D6FULL, "vxht_0_hightown_a2_c4y_0_0.vlm"},
	{0xEACDABCEB717D54CULL, "vxlt_hangedman_5l_0_0.vlm"},
	{0xEACDF388DE7A571EULL, "vxht_mansion2_2ol_0_0.vlm"},
	{0xEACE5AF17EE54E9AULL, "cod_hst_bonepit.utp"},
	{0xEACF3497773A6EEAULL, "dre000_bodahn6031280.cl"},
	{0xEACF8884110CEA1BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_enter.ani"},
	{0xEACFB95E5D3AC14FULL, "w_woundedcoast_146_0.phy"},
	{0xEAD0715676987CB6ULL, "gen00fl_aveline6139817.cl"},
	{0xEAD07CC61D1C8F1CULL, "gen00fl_anders6125350.cl"},
	{0xEAD0B2A5669B0805ULL, "hlac_lt_gallowscourtyard_131070_0y.dds"},
	{0xEAD0FC47C5262EC0ULL, "vxca_cavern1_haunted_ijk_0_0.vlm"},
	{0xEAD16856A1A3A307ULL, "608556799.ogg"},
	{0xEAD189308FCAD5C6ULL, "zz_approval_debug6003520.cl"},
	{0xEAD1B22FD942FA0EULL, "145154479.ogg"},
	{0xEAD1B9CDFBCE4049ULL, "gen00fl_fenris6106709.cl"},
	{0xEAD1E9E2B6070FD0ULL, "dae300_petrice6097862.cl"},
	{0xEAD213B1C955FE99ULL, "vxht_mansion2_1kd_0_0.vlm"},
	{0xEAD2146A0E75A560ULL, "vxca_caverns2_7l5_0_0.vlm"},
	{0xEAD224345CCA4444ULL, "hlac_lt_0_lowtown_docks_-131074_0y.dds"},
	{0xEAD2B886BC99DE72ULL, "lt_0_lowtown_a2_n_c_46_0_lc.mmh"},
	{0xEAD2D21B21FF9550ULL, "hightown_market_guard_sta6169604.cl"},
	{0xEAD350BAE496FAD9ULL, "688578843.ogg"},
	{0xEAD37978FC001622ULL, "vxlt_darktown_2qm_0_0.vlm"},
	{0xEAD379BF81601FFBULL, "vxlt_backalley_g_1of_0_1.vlm"},
	{0xEAD3A0203F8F1760ULL, "dae021_torelin_mob6150228.cl"},
	{0xEAD3B65B12092689ULL, "6110887_m.fxe"},
	{0xEAD3EB04A3E787FEULL, "392924139.ogg"},
	{0xEAD4B82DB0247293ULL, "vxht_0_hightown_n_9v2_0_0.vlm"},
	{0xEAD4D5341BE952C7ULL, "vxlt_warehouse_11j_0_0.vlm"},
	{0xEAD4EC0EB108029EULL, "941138675.ogg"},
	{0xEAD54700D47BAE26ULL, "gen00fl_fenris6118721.cl"},
	{0xEAD5546F4F4B39CBULL, "334490295.ogg"},
	{0xEAD55601C05E75CBULL, "one100dg_vanard6017495.cl"},
	{0xEAD5BC0A25663DEBULL, "dae011_corff_sta6142378.cl"},
	{0xEAD5F3F0AA1E1B15ULL, "ave200_donnic6061277.cl"},
	{0xEAD602500AAD3B29ULL, "mrl150_merrill6071783.cl"},
	{0xEAD7965DBD1DCD5CULL, "968404124.ogg"},
	{0xEAD79B0CB6B6523AULL, "918917278.ogg"},
	{0xEAD8241BF2E4A0F4ULL, "6079890_f.fxe"},
	{0xEAD82A4FCAE9EBC3ULL, "art\\vfx\\textures\\vfx_smoke_soft_cs.dds"},
	{0xEAD849CDE0832885ULL, "vxht_0_hightown_a2_cf1_0_0.vlm"},
	{0xEAD85E9166E80004ULL, "w_woundedcoast_155_0.mmh"},
	{0xEAD8629166E80757ULL, "w_woundedcoast_155_0.mao"},
	{0xEAD8749166E825A6ULL, "w_woundedcoast_155_0.msh"},
	{0xEAD904DDF909628DULL, "6122125_m.fxe"},
	{0xEAD9276DD8A6A00FULL, "1037521419.ogg"},
	{0xEAD93CB195CAFB9DULL, "lgt120_orz_bln_official6140207.cl"},
	{0xEAD96CCA855255D9ULL, "6086270_m.fxe"},
	{0xEADA47C26AF69327ULL, "mag310cr_progenitor_3.utc"},
	{0xEADAA99B4A1FF5BCULL, "vxht_mansion2_1w9_0_0.vlm"},
	{0xEADAAC54456A0F38ULL, "pro000_refugee_female.cnv"},
	{0xEADABD905E726E6FULL, "623101592.ogg"},
	{0xEADAE58AE7201A88ULL, "lt_0_lowtown_docks_n_188_0.mmh"},
	{0xEADAF18AE7202F2BULL, "lt_0_lowtown_docks_n_188_0.mao"},
	{0xEADB038AE7204DB2ULL, "lt_0_lowtown_docks_n_188_0.msh"},
	{0xEADB8B28CD5E8995ULL, "fen110_fenris6117984.cl"},
	{0xEADC00332FB625EFULL, "vxht_0_hightown_a2_c3x_0_0.vlm"},
	{0xEADC0F02FCE09D1CULL, "243418321.ogg"},
	{0xEADC5BB3AC0B2D0BULL, "vxca_caverns1_mines_j4u_0_0.vlm"},
	{0xEADCAF3373886680ULL, "6050328_m.fxe"},
	{0xEADCC302C1BE3E3AULL, "hlac_lt_0_lowtown_-65536_3z.dds"},
	{0xEADCFA1D609B44CBULL, "31106127.ogg"},
	{0xEADD062C0FCE0875ULL, "qun231ip_chest_a.utp"},
	{0xEADD76CB362795ACULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_c.anb"},
	{0xEADD8E54456C1240ULL, "pro000_refugee_female.bnk"},
	{0xEADD95869E352EF6ULL, "gen_im_arm_cht_nobc_t1.uti"},
	{0xEADD9B4488629833ULL, "hlac_lt_gallowscourtyard_n_-1_2z.dds"},
	{0xEADE437D2B3765F6ULL, "hlac_ht_keep_3743_1y.dds"},
	{0xEADE7955EDFC4E8FULL, "28242003.ogg"},
	{0xEADE7E69FDF1D3F2ULL, "vxlt_hangedman_2wz_0_0.vlm"},
	{0xEADE91196736C545ULL, "461925079.ogg"},
	{0xEADEB38D0A6AF339ULL, "t1_skn_001.tnt"},
	{0xEADF2FA2ED0E4926ULL, "qun160_vashothlead6064591.cl"},
	{0xEADF5205E50DF950ULL, "w_sundermount_80_0.rml"},
	{0xEADF6E29D502CB93ULL, "75807523.ogg"},
	{0xEADF837DFAB828CCULL, "one100dg_lia6049673.cl"},
	{0xEADFBC60C44DF403ULL, "gen00fl_anders6096466.cl"},
	{0xEADFFFCD51160F07ULL, "6120411_m.fxe"},
	{0xEAE0067AA64A6DB2ULL, "gen00fl_aveline6150250.cl"},
	{0xEAE07F28AC40438EULL, "vxw_bonepitmining_a2_bb_0_0.vlm"},
	{0xEAE0C0F2194CE0DAULL, "6093741_m.fxe"},
	{0xEAE0CEC952DB247DULL, "art\\characters\\creatures\\shade\\c_shadeb.phy"},
	{0xEAE0DCCE3E7806FEULL, "art\\characters\\weapons\\longswords\\textures\\w_gsw_022a_0n.dds"},
	{0xEAE165B7150BF2F2ULL, "vxlt_0_lowtown_docks_9do_0_0.vlm"},
	{0xEAE1EB55EDF62EA2ULL, "vxlt_darktown_2tn_0_0.vlm"},
	{0xEAE1F0DE56DAA7BFULL, "6207790_m.fxe"},
	{0xEAE2306352E3B99EULL, "vxw_bonepitdestroyed_a3_178_0_0.vlm"},
	{0xEAE2B2895EEBB309ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml104a_0d.dds"},
	{0xEAE30F0814E0204EULL, "gen00fl_varric6081210.cl"},
	{0xEAE31E79ED48813EULL, "t3_dey_gn3.tnt"},
	{0xEAE32546989FEF98ULL, "gen00fl_dog6167032.cl"},
	{0xEAE347110DDF6B47ULL, "vxlt_warehouse_14e_0_0.vlm"},
	{0xEAE360F569C014C6ULL, "vxca_caverns1_eh2_0_0.vlm"},
	{0xEAE3AF84111DD5A5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\follower_idle_enter.gad"},
	{0xEAE3B777CE75FB7DULL, "vxw_encounter2_49_0_0.vlm"},
	{0xEAE3CEEEAA5CCD42ULL, "349152866.ogg"},
	{0xEAE3EE24778D6E7DULL, "6110123_f.fxe"},
	{0xEAE42BE198036154ULL, "hlac_ht_0_hightown_-131074_2z.dds"},
	{0xEAE45CCA34D2F5D2ULL, "6095831_m.fxe"},
	{0xEAE4C2860D35F0F1ULL, "hlac_lt_backalley_g_0_19y.dds"},
	{0xEAE4C93AD78A5A6FULL, "mrl150_merrill6173793.cl"},
	{0xEAE4DD75A4C3F350ULL, "vxht_0_hightown_amc_0_0.vlm"},
	{0xEAE4E9846BCB904EULL, "w_bonepitmining_a2_8_0.mmh"},
	{0xEAE4F5846BCBA4E5ULL, "w_bonepitmining_a2_8_0.mao"},
	{0xEAE500A88C68BC09ULL, "96630061.ogg"},
	{0xEAE503846BCBBC3CULL, "w_bonepitmining_a2_8_0.msh"},
	{0xEAE533D0DBFD2611ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.dura.ani"},
	{0xEAE551230C91F0F0ULL, "vxca_caverns2prc_9q5_0_0.vlm"},
	{0xEAE58B749AEDC81DULL, "6036695_m.fxe"},
	{0xEAE58C4B295D064CULL, "character_character_p276.psh.fxo.dx11"},
	{0xEAE60BFE0C1A7BEDULL, "dae140ar_bone_pit.lst"},
	{0xEAE628127E5AA91CULL, "6072247_m.fxe"},
	{0xEAE62CB33E095F2FULL, "6027978_m.fxe"},
	{0xEAE656CD7A32B81FULL, "mcr370cs_player_rules.cut"},
	{0xEAE76CA5BCA08E40ULL, "534733191.ogg"},
	{0xEAE80B0460FE1461ULL, "272584393.ogg"},
	{0xEAE842125BCA17C4ULL, "640905910.ogg"},
	{0xEAE9542FA7D628FCULL, "plt_cod_hst_qun_mages.plo"},
	{0xEAE96DE9EA30F017ULL, "hlac_lt_0_lowtown_a3_n_c_-196609_4x.dds"},
	{0xEAE9B8E2FC0C66DCULL, "vxdr_primevalentrance_2q7_0_2.vlm"},
	{0xEAE9E1F739497E3FULL, "character_character_p33.psh.fxo.dx11"},
	{0xEAEA3B1013D27FDEULL, "vxw_sundermount_252_0_0.vlm"},
	{0xEAEA508AAFB5E5C1ULL, "hlac_lt_foundry_n_320_20z.dds"},
	{0xEAEA9EB5E79997DCULL, "6142378_f.fxe"},
	{0xEAEACD909E01458BULL, "vxca_caverns1_mines_j3t_0_0.vlm"},
	{0xEAEAD742C266881EULL, "vxht_mansion2_2un_0_0.vlm"},
	{0xEAEB3A92C428BC84ULL, "vxlt_gallowscourtyard_72m_0_0.vlm"},
	{0xEAEB47A2FDCB91D9ULL, "275887495.ogg"},
	{0xEAEC3ED0DC0357B3ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.dura.gad"},
	{0xEAECB30662EDC632ULL, "vxlt_warehouse_n_td_0_0.vlm"},
	{0xEAECD0228D63B6E9ULL, "229854760.ogg"},
	{0xEAED078FED1B7EB1ULL, "one350_vinell.bnk"},
	{0xEAED23FE0C20C37CULL, "dae140ar_bone_pit.ncs"},
	{0xEAED377679D91586ULL, "llac_lt_0_lowtown_docks_-65537_1.dds"},
	{0xEAED40BCBF185266ULL, "follower_banter6161016.cl"},
	{0xEAEDF8CFC051A159ULL, "6129249_m.fxe"},
	{0xEAEE4250AD6C11FEULL, "48431170.ogg"},
	{0xEAEE425DA4A32A00ULL, "vxht_mansion2b_3gk_0_0.vlm"},
	{0xEAEE614A207350B1ULL, "6171451_m.fxe"},
	{0xEAEEBC190D31AF17ULL, "vxht_bartrand_n_6qg_0_0.vlm"},
	{0xEAEEF1059E365C0EULL, "vxw_bonepitmining_a2_ic_0_0.vlm"},
	{0xEAEF059F33EEF01BULL, "mcr371_hostile_anders6082122.cl"},
	{0xEAEF89032FFBAD03ULL, "496513434.ogg"},
	{0xEAEFDB50C165A9CEULL, "6113571_m.fxe"},
	{0xEAEFE4D4174C32E3ULL, "6071069_m.fxe"},
	{0xEAF061035DF1EB77ULL, "vxlt_0_lowtown_docks_a3_n_d36_0_0.vlm"},
	{0xEAF09AB3BA198BDCULL, "hlac_ca_caverns2_5802_51y.dds"},
	{0xEAF0D7EA8AFBA2D8ULL, "6143705_m.fxe"},
	{0xEAF0E98FED1F34C1ULL, "one350_vinell.cnv"},
	{0xEAF15AB27F024B25ULL, "hlac_lt_0_lowtown_131071_3x.dds"},
	{0xEAF18121071F02CEULL, "346204188.ogg"},
	{0xEAF1C3C952E984E9ULL, "art\\characters\\creatures\\shade\\c_shadeb.mao"},
	{0xEAF1CFC952E99882ULL, "art\\characters\\creatures\\shade\\c_shadeb.mmh"},
	{0xEAF1D1C952E99C38ULL, "art\\characters\\creatures\\shade\\c_shadeb.msh"},
	{0xEAF1DEE71D4B72ACULL, "hlac_ht_mansion2b_818_10y.dds"},
	{0xEAF2D922FDB2DB05ULL, "1037364395.ogg"},
	{0xEAF2DCC3562A2ABAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_1_sylvesta.gad"},
	{0xEAF33BD0DC097118ULL, "art\\vfx\\combat\\vfx_m_wpn_ele_crust_c.dura.evt"},
	{0xEAF33BED7A2FEDAFULL, "ser191_emeric6104795.cl"},
	{0xEAF36F6B3A15A460ULL, "vxlt_gamlenshouse_n_qw_0_0.vlm"},
	{0xEAF37CE61C53098CULL, "vxlt_0_lowtown_docks_a3_n_ars_0_0.vlm"},
	{0xEAF3BFC85EB76591ULL, "gen00fl_isabela6079263.cl"},
	{0xEAF42CFDF5DA9C72ULL, "pro000_athenril6049437.cl"},
	{0xEAF446042AAECBB4ULL, "vxlt_0_lowtown_docks_bgg_0_0.vlm"},
	{0xEAF4C59ADED3186AULL, "vxlt_0_lowtown_a2_gnf_0_0.vlm"},
	{0xEAF4D2A466F068A8ULL, "hlac_lt_0_lowtown_a2_n_65533_1y.dds"},
	{0xEAF4F83B080D813FULL, "451630691.ogg"},
	{0xEAF5300A773AA772ULL, "vxht_0_hightown_a2_bun_0_0.vlm"},
	{0xEAF5396C3733881DULL, "zz_followers_banter6157591.cl"},
	{0xEAF5974688073624ULL, "lt_0_lowtown_docks_a3_n_1_0.anb"},
	{0xEAF5DB71B6A59471ULL, "qun200_party_comment6182720.cl"},
	{0xEAF752C92E7E29F8ULL, "ave200_donnic6061418.cl"},
	{0xEAF7A728C3C529B4ULL, "terrain2.vsh.fxo"},
	{0xEAF81C2670E54387ULL, "one110cr_harimann.utc"},
	{0xEAF8285F4F579087ULL, "tm_glk_b01.tnt"},
	{0xEAF82D800C91F2EEULL, "553102598.ogg"},
	{0xEAF86BDD87E4D9D6ULL, "177071827.ogg"},
	{0xEAF8F3691CDB0131ULL, "character_character_p161.psh.fxo.dx11"},
	{0xEAF97384BED8EAF1ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower1_meredithsintro.ani"},
	{0xEAFAB048957CF680ULL, "855002332.ogg"},
	{0xEAFB30AFBE64BF52ULL, "6072164_m.fxe"},
	{0xEAFB6223E1DE04F2ULL, "vxlt_hangedman_2id_0_0.vlm"},
	{0xEAFB997B21B90652ULL, "vxht_keep_5e9_0_0.vlm"},
	{0xEAFBDC4D8AB5948EULL, "fen110_anso6095023.cl"},
	{0xEAFBEFEB198CB2EEULL, "gen00fl_anders6119333.cl"},
	{0xEAFC03AE463F1D9AULL, "vxlt_0_lowtown_docks_b4e_0_1.vlm"},
	{0xEAFCACEE7C2D4F6AULL, "6123221_f.fxe"},
	{0xEAFCD32712696931ULL, "6174751_m.fxe"},
	{0xEAFCD68212EE833EULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge7.ani"},
	{0xEAFCEF5A04C4E998ULL, "vxlt_undercitydungeon_1c1_0_0.vlm"},
	{0xEAFD26295EF26EA5ULL, "347302280.ogg"},
	{0xEAFD28660CAF0C6EULL, "mrl000_marethari6205606.cl"},
	{0xEAFD64A42543097CULL, "663795761.ogg"},
	{0xEAFD73CAEAEFC3C2ULL, "909666721.ogg"},
	{0xEAFD96F97D14332CULL, "6095859_m.fxe"},
	{0xEAFDA75C7EB93444ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.dura.ani"},
	{0xEAFE59EEAEEE9679ULL, "lgt260_way_blood_mage6148393.cl"},
	{0xEAFF3D7DCD757EA4ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtczz.mmh"},
	{0xEAFF537DCD75A446ULL, "art\\characters\\playercharacter\\humanmale\\glv_lgtczz.msh"},
	{0xEAFF731C94D1D4ECULL, "505503947.ogg"},
	{0xEAFFDE5E3E204F86ULL, "ravenside01.mao"},
	{0xEAFFF95B10D5360EULL, "518676443.ogg"},
	{0xEB0035A1B105E0A9ULL, "hlac_lt_0_lowtown_docks_a3_196608_0z.dds"},
	{0xEB007E84BEDF1B93ULL, "art\\cinematics\\_crit path\\qunari crisis\\hightown\\cs_follower1_meredithsintro.gad"},
	{0xEB01378CC78687C0ULL, "hlac_lt_0_lowtown_docks_131071_1z.dds"},
	{0xEB015EBC7F623FE1ULL, "350562767.ogg"},
	{0xEB0172279633274CULL, "w_blightlands_137_0.msh"},
	{0xEB0178279633313EULL, "w_blightlands_137_0.mmh"},
	{0xEB018427963345D5ULL, "w_blightlands_137_0.mao"},
	{0xEB01D809FC651199ULL, "gen_im_trs_dwv_junk1_02.uti"},
	{0xEB0227BA16A4F7B0ULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlookout_0.pwk"},
	{0xEB0270C7DC17DE25ULL, "gen00fl_sebastian6174022.cl"},
	{0xEB032C71DA227093ULL, "llac_w_sundermount_131072_0.dds"},
	{0xEB044A9306C7AFFEULL, "6093899_m.fxe"},
	{0xEB049A67C2C64EF1ULL, "6098776_m.fxe"},
	{0xEB04A45C7EBF4DA6ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.dura.gad"},
	{0xEB04DA5B1A97D729ULL, "gen00fl_aveline6139180.cl"},
	{0xEB0503477D218AEBULL, "vxca_caverns2_96z_0_0.vlm"},
	{0xEB050C0F1F5770EFULL, "llac_ca_caverns2prc_5802_58.dds"},
	{0xEB05B4FCEB99922AULL, "hlac_lt_0_lowtown_n_-2_1x.dds"},
	{0xEB0612E89307F7AFULL, "70535915.ogg"},
	{0xEB061D8ECF387B85ULL, "330781658.ogg"},
	{0xEB06F7C3563B01A0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\exodus\\standing_neutral_gestures_2_1_sylvesta.ani"},
	{0xEB071E6D1F08A83CULL, "gen_im_acc_amu_lore_02.uti"},
	{0xEB0778E53171609FULL, "one350_terath6087144.cl"},
	{0xEB07969521DD0B88ULL, "w_bonepitmining_a2_91_0.msh"},
	{0xEB07A89521DD2A19ULL, "w_bonepitmining_a2_91_0.mao"},
	{0xEB07B47DA0C2CCBDULL, "436402624.ogg"},
	{0xEB07B49521DD3EB2ULL, "w_bonepitmining_a2_91_0.mmh"},
	{0xEB08793F863BA808ULL, "hlac_lt_darktown_-1_64z.dds"},
	{0xEB08B68301F1F005ULL, "def_cobble_s.dds"},
	{0xEB0A47B6E17615C6ULL, "vxw_bonepit_163_0_0.vlm"},
	{0xEB0B5305E53335B8ULL, "w_sundermount_80_0.anb"},
	{0xEB0B5B1F0D19FCD9ULL, "fen210_fenris6109434.cl"},
	{0xEB0B9F5C7EC56385ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.dura.evt"},
	{0xEB0BD2DE7B51DC78ULL, "vxca_caverns2_7tu_0_0.vlm"},
	{0xEB0C1338CE297B8AULL, "181979900.ogg"},
	{0xEB0C66FC6969F3B8ULL, "6061030_m.fxe"},
	{0xEB0D62A90DA1C318ULL, "6201368_m.fxe"},
	{0xEB0D93E7A962B3DAULL, "vxlt_0_lowtown_docks_bug_0_0.vlm"},
	{0xEB0DEAEF527207D4ULL, "728292440.ogg"},
	{0xEB0E15883F313FEEULL, "hlac_lt_0_lowtown_n_65536_12z.dds"},
	{0xEB0E950FB506BB84ULL, "vxht_0_hightown_a3_bu5_0_0.vlm"},
	{0xEB0F02E1E0402666ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_left.evt"},
	{0xEB0F923AB98384C8ULL, "hlac_ht_brothel_n_3368_0y.dds"},
	{0xEB0F9ABE060FC01BULL, "6037041_m.fxe"},
	{0xEB0FA3B36D45E178ULL, "vxca_caverns1_mines_j2f_0_2.vlm"},
	{0xEB0FF344CD7EA6F3ULL, "vxht_chantry_n_20s_0_0.vlm"},
	{0xEB10111486341565ULL, "follower_banter6170095.cl"},
	{0xEB107D6D8CFF349FULL, "207938716.ogg"},
	{0xEB10B569EA45FBFDULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wo_basic_1.mao"},
	{0xEB10EF8212FF56B8ULL, "art\\cinematics\\_crit path\\prologue\\additional animations\\cs_hurlock_charge7.gad"},
	{0xEB10FC3CD23CCCA6ULL, "96828421.ogg"},
	{0xEB11403136375588ULL, "fex_ruin_wl_b_cn.mmh"},
	{0xEB115E31363788B2ULL, "fex_ruin_wl_b_cn.msh"},
	{0xEB11B107608D387FULL, "205358332.ogg"},
	{0xEB11C7423C19521CULL, "dae000ip_wall_lean_right_s.utp"},
	{0xEB12174B0B2BBDFDULL, "vxlt_0_lowtown_a3_n_c_byq_0_0.vlm"},
	{0xEB124F3CA5E24A45ULL, "6084441_m.fxe"},
	{0xEB12DE9D318EC1ACULL, "gen_im_trs_gen_shd_wdn_r2.uti"},
	{0xEB1314FE0C4178AAULL, "dae140ar_bone_pit.are"},
	{0xEB1326BC80CCE28FULL, "mcr310_argument6072693.cl"},
	{0xEB1354DF6966AAF9ULL, "30858344.ogg"},
	{0xEB1490B8F7205404ULL, "6107688_m.fxe"},
	{0xEB15C360BC903675ULL, "864415081.ogg"},
	{0xEB15F69DDE4459DAULL, "137712573.ogg"},
	{0xEB1611C08CA7127FULL, "6103319_m.fxe"},
	{0xEB1619E1E0466BB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_left.gad"},
	{0xEB16B720FB1354A8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04a.msh"},
	{0xEB16D520FB1387D2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04a.mmh"},
	{0xEB16D7CB618FBB48ULL, "44264118.ogg"},
	{0xEB16EA2D21B15596ULL, "6148746_m.fxe"},
	{0xEB1784645300ABDDULL, "6103397_m.fxe"},
	{0xEB17903345CAAFC9ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_c.impa.ani"},
	{0xEB17D7199E28D627ULL, "ts_kirk_lowtown_prog_junk2.gda"},
	{0xEB17E8A5469EA699ULL, "art\\characters\\creatures\\stonegolem\\animation\\dblw_lp_v1.gad"},
	{0xEB18C9F8D41890FAULL, "638863598.ogg"},
	{0xEB18E486162CE40DULL, "hla_ht_0_hightown_n_-131073_0x.dds"},
	{0xEB196FADA16ABD28ULL, "w_blightlands_61_0.msh"},
	{0xEB1980321BE3B19BULL, "vxlt_gallowstemplar_n_3k4_0_0.vlm"},
	{0xEB1981ADA16ADBB9ULL, "w_blightlands_61_0.mao"},
	{0xEB198DADA16AF052ULL, "w_blightlands_61_0.mmh"},
	{0xEB19B031CAAEE752ULL, "912417895.ogg"},
	{0xEB19E598459F3A01ULL, "vxlt_0_lowtown_a2_n_ehz_0_0.vlm"},
	{0xEB1A2F925E82CBB4ULL, "ico_poison_debilitate.dds"},
	{0xEB1AA6057CC04987ULL, "vxlt_hangedman_215_0_0.vlm"},
	{0xEB1AC6CD571C3A1DULL, "hlac_lt_undercityencounter_-1_18z.dds"},
	{0xEB1B157C694E7963ULL, "821076755.ogg"},
	{0xEB1B3D2157FD58F2ULL, "867038105.ogg"},
	{0xEB1B500CF7BF3CDCULL, "cs_qcr220cs_meredith_arrival.wbd"},
	{0xEB1BBD6EB294F681ULL, "vxlt_0_lowtown_eoe_0_1.vlm"},
	{0xEB1BBFDC8F7B6DA5ULL, "230015000.ogg"},
	{0xEB1C835FC91A6CB6ULL, "gen_im_trs_krk_drk_junk1_02.uti"},
	{0xEB1C9729FD6014F2ULL, "252330519.ogg"},
	{0xEB1CB6519A0366B2ULL, "vxlt_darktown_4yo_0_0.vlm"},
	{0xEB1D1AE1E04C8C5FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_gesture_left.ani"},
	{0xEB1D444C65F17CA2ULL, "zz_followers_banter6157626.cl"},
	{0xEB1DAB3345CF494BULL, "art\\vfx\\combat\\vfx_c_rw_teleport_c.impa.gad"},
	{0xEB1DC5E32E7802ADULL, "vxlt_undercityencounter_1j2_0_0.vlm"},
	{0xEB1E8CECA27F4218ULL, "vxlt_0_lowtown_a3_n_c_hl9_0_0.vlm"},
	{0xEB1EC5020FF7EDE0ULL, "vxlt_gamlenshouse_n_br_0_0.vlm"},
	{0xEB1EF9A546A4E2E7ULL, "art\\characters\\creatures\\stonegolem\\animation\\dblw_lp_v1.ani"},
	{0xEB1F2BC2DECEE8F9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medd_0d.dds"},
	{0xEB1F44779287C6EFULL, "hlac_ht_mansion_n_3381_0x.dds"},
	{0xEB1F7B36C8E3D604ULL, "vxlt_0_lowtown_a2_n_hd2_0_0.vlm"},
	{0xEB2057CFE095C8EAULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_darren.gad"},
	{0xEB20729C45392BD3ULL, "vxht_0_hightown_c7d_0_0.vlm"},
	{0xEB2073C9C32BB3CDULL, "gen00fl_aveline6115414.cl"},
	{0xEB20B4642968E090ULL, "993322822.ogg"},
	{0xEB20FC1502798330ULL, "lt_0_lowtown_a3_n_c_4_0.mao"},
	{0xEB20FD69E4DB9942ULL, "6171557_m.fxe"},
	{0xEB21001502798A43ULL, "lt_0_lowtown_a3_n_c_4_0.mmh"},
	{0xEB2109BFB1942A2BULL, "6108700_f.fxe"},
	{0xEB210E150279A19DULL, "lt_0_lowtown_a3_n_c_4_0.msh"},
	{0xEB2142C30E92F712ULL, "mer221_sabin6080573.cl"},
	{0xEB21702357107E49ULL, "1035771931.ogg"},
	{0xEB2171933AABA845ULL, "6153824_m.fxe"},
	{0xEB21BCCA1ED8A45BULL, "mag320_orsino6071262.cl"},
	{0xEB221AAB620C10D9ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.impa.ani"},
	{0xEB223966DCB92379ULL, "vxw_encounter2_camp_8i_0_0.vlm"},
	{0xEB22E2EC03EE703CULL, "vxht_mansion2_1n2_0_0.vlm"},
	{0xEB239DFF9E732868ULL, "zz_seb_debug6214529.cl"},
	{0xEB23B36AA24AB163ULL, "946433121.ogg"},
	{0xEB23C0681730B0F5ULL, "art\\vfx\\materials\\chunky\\vfx_c_crack_tile_add.mao"},
	{0xEB23CA413662F54EULL, "hlac_lt_0_lowtown_a3_n_c_65536_1y.dds"},
	{0xEB23E7579541060FULL, "vxht_chantry_n_350_0_1.vlm"},
	{0xEB23FD69E401F8C1ULL, "572764839.ogg"},
	{0xEB24A83345D562B0ULL, "art\\vfx\\combat\\vfx_c_rw_teleport_c.impa.evt"},
	{0xEB24BDBC1509D207ULL, "vxca_caverns1_mines_fmv_0_0.vlm"},
	{0xEB24D0253F8E1B99ULL, "614715739.ogg"},
	{0xEB24D1AFAD82FBBAULL, "6111295_f.fxe"},
	{0xEB24F4BD20599FADULL, "llac_lt_warehouse_n_65537_5.dds"},
	{0xEB25098CF562CC51ULL, "6086354_f.fxe"},
	{0xEB257370C05BAE2DULL, "hlac_lt_darktown_-1_41z.dds"},
	{0xEB25804ED70264A9ULL, "hlac_ht_chantry_n_2534_43z.dds"},
	{0xEB259D3ECE1F7F66ULL, "dae000_tomwise6167317.cl"},
	{0xEB25EE7D3269EA06ULL, "hlac_ht_0_hightown_a2_65533_3z.dds"},
	{0xEB25F64144F6C45DULL, "6104297_m.fxe"},
	{0xEB26BCD6321D9E97ULL, "600656592.ogg"},
	{0xEB27EC637EAE7AEAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masc_0s.dds"},
	{0xEB28095FBA7FB83BULL, "ave200_aveline_bar6115016.cl"},
	{0xEB2829945D34C49DULL, "vxlt_undercitydungeon_1fo_0_0.vlm"},
	{0xEB28577537955281ULL, "vxlt_0_lowtown_a2_n_emy_0_0.vlm"},
	{0xEB28B7D77311BC59ULL, "82027763.ogg"},
	{0xEB28E9020B81862DULL, "6125609_m.fxe"},
	{0xEB28F5AB6211F01BULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.impa.gad"},
	{0xEB29E5A65239C987ULL, "344277093.ogg"},
	{0xEB29F352068DCA52ULL, "vxht_0_hightown_c7a_0_0.vlm"},
	{0xEB2A73066D035A36ULL, "6188694_m.fxe"},
	{0xEB2A946A69BF3399ULL, "80114131.ogg"},
	{0xEB2ABC5952E57825ULL, "llac_w_encounter2_camp_65536_0.dds"},
	{0xEB2B1A603844D33FULL, "mcr310_confused_guard6077072.cl"},
	{0xEB2B321071D2D8FCULL, "mcr310st_argument_anders.stg"},
	{0xEB2B42CAC17E5BA2ULL, "gen_im_trs_dem_sta_spi_r2.uti"},
	{0xEB2B437F25E64161ULL, "llac_ht_chantry_2534_28.dds"},
	{0xEB2B800A42A99CACULL, "hlac_lt_0_lowtown_a2_n_c_0_0z.dds"},
	{0xEB2B80A3B831DC2FULL, "vxht_0_hightown_boq_0_0.vlm"},
	{0xEB2B89F9648DA8AFULL, "vxca_caverns2_8pu_0_0.vlm"},
	{0xEB2C0B26A9ACE1ECULL, "vxht_brothel_n_5ly_0_1.vlm"},
	{0xEB2C209B8D59DF93ULL, "vxlt_0_lowtown_a2_efa_0_0.vlm"},
	{0xEB2C26C055E88DDDULL, "6096950_m.fxe"},
	{0xEB2C882AADE1E07BULL, "hlac_w_sundermount_0_2x.dds"},
	{0xEB2C9BC1092EFD0EULL, "hlac_ht_chantry_n_2534_39x.dds"},
	{0xEB2CA0CC894B384CULL, "6034708_m.fxe"},
	{0xEB2E33A7BF40AE4CULL, "hlac_w_bonepitdestroyed_a3_131075_0y.dds"},
	{0xEB2EA35646EE488DULL, "106354411.ogg"},
	{0xEB2EB545C4B1631BULL, "6137688_m.fxe"},
	{0xEB2F6F46D6D1B1C2ULL, "6174657_m.fxe"},
	{0xEB2FA070DC2139EDULL, "141045565.ogg"},
	{0xEB3012AB62184000ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.impa.evt"},
	{0xEB304F88A6A807EBULL, "vxht_0_hightown_a4s_0_0.vlm"},
	{0xEB305CC357507F4CULL, "6167879_m.fxe"},
	{0xEB30657BA441950FULL, "w_bonepit_29_0_lc.mmh"},
	{0xEB30817CCBB6614FULL, "135811066.ogg"},
	{0xEB30AB43CEAF3BF9ULL, "vxw_encounter2_camp_3h_0_0.vlm"},
	{0xEB3101D4C752BFEEULL, "6085676_m.fxe"},
	{0xEB31DB82FB3EDBF4ULL, "vxca_cavern1_haunted_dpl_0_0.vlm"},
	{0xEB31FE43B9894713ULL, "vxdr_primevalentrance_36v_0_0.vlm"},
	{0xEB3201150287FF68ULL, "lt_0_lowtown_a3_n_c_4_0.phy"},
	{0xEB322B7F200DF353ULL, "dae000_bartrand6206884.cl"},
	{0xEB3391AE25D4A330ULL, "6108634_m.fxe"},
	{0xEB33A9C8E59A1DD9ULL, "6118483_m.fxe"},
	{0xEB33AFA7AED66F6FULL, "hlac_lt_0_lowtown_a3_n_c_131071_7x.dds"},
	{0xEB33B3D06BBF6641ULL, "ico_poison_arcane.dds"},
	{0xEB33DD62946003DFULL, "855405648.ogg"},
	{0xEB33EC11C9E97489ULL, "vxdr_0_deeproads_cx0_0_0.vlm"},
	{0xEB340E7BD43326B8ULL, "zz_and330st_base_justice_st.stg"},
	{0xEB344005E5561ACDULL, "w_sundermount_80_0.msh"},
	{0xEB344E05E5563200ULL, "w_sundermount_80_0.mao"},
	{0xEB345205E5563953ULL, "w_sundermount_80_0.mmh"},
	{0xEB3532CFE0A7E5D0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_cheeraggro_1_darren.ani"},
	{0xEB3564158CB76C7BULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portraitc.msh"},
	{0xEB3576158CB78AC6ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portraitc.mao"},
	{0xEB3582158CB79F65ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_portraitc.mmh"},
	{0xEB3690AE3BFFE470ULL, "t3_dey_bl2.tnt"},
	{0xEB369B714F2ADD1DULL, "vxlt_undercitydungeon_1in_0_0.vlm"},
	{0xEB36B646883F0184ULL, "lt_0_lowtown_docks_a3_n_1_0.rml"},
	{0xEB36BA1A5F0ADC61ULL, "vxlt_gallowscourtyard_n_4v4_0_0.vlm"},
	{0xEB3727EFB891DF47ULL, "6187711_m.fxe"},
	{0xEB384E1D12213351ULL, "mag101_alain6032270.cl"},
	{0xEB3860E50553EF61ULL, "dae000_mother6108965.cl"},
	{0xEB38781C885ADE5BULL, "gen00fl_aveline6137698.cl"},
	{0xEB389C1A8ACA4311ULL, "vxlt_gallowscourtyard_64w_0_0.vlm"},
	{0xEB38E97E7F44FD5AULL, "vxlt_0_lowtown_docks_bbb_0_0.vlm"},
	{0xEB38F3B9E29AB06AULL, "llac_lt_0_lowtown_a2_n_c_65536_9.dds"},
	{0xEB3928C1C81AEB29ULL, "mag131_taraneh6086045.cl"},
	{0xEB39AFD0B0B54434ULL, "mer240_jansen6050427.cl"},
	{0xEB39CCC8A46F99C0ULL, "qun111_ketojan6040970.cl"},
	{0xEB3A14BE0028FBF7ULL, "431035072.ogg"},
	{0xEB3A2A6FF7B98E05ULL, "vxca_caverns2_7tl_0_0.vlm"},
	{0xEB3A7A49F5645691ULL, "6140198_m.fxe"},
	{0xEB3AD24178576B22ULL, "dm_amb_merchant_weapons.mor"},
	{0xEB3AF9068A2B355CULL, "145922555.ogg"},
	{0xEB3AFEC80452E8D7ULL, "vxht_0_hightown_b4d_0_0.vlm"},
	{0xEB3B173B6C54D16CULL, "6052748_m.fxe"},
	{0xEB3C0BEC10AA16DDULL, "dae000_adriano6163675.cl"},
	{0xEB3C4AB5D20FB20DULL, "6103567_f.fxe"},
	{0xEB3CA12F3F2E4C24ULL, "vxlt_0_lowtown_a2_n_cmj_0_0.vlm"},
	{0xEB3CF34D157DDB13ULL, "vxht_keep_n_6gf_0_0.vlm"},
	{0xEB3D11D5F79552EBULL, "6044192_f.fxe"},
	{0xEB3D7880277AF6BDULL, "lt_0_lowtown_docks_a3_n_46_0.rml"},
	{0xEB3D8DD40AFED346ULL, "6065843_m.fxe"},
	{0xEB3D9D6274953CE7ULL, "mag291_ntbanter6056839.cl"},
	{0xEB3E0F514A873A29ULL, "vxlt_gallowstemplar_f_6s3_0_0.vlm"},
	{0xEB3E609E128401D1ULL, "431242715.ogg"},
	{0xEB3F2A31F41CB170ULL, "84504089.ogg"},
	{0xEB3F3BD3C6C17F82ULL, "vxht_0_hightown_by6_0_0.vlm"},
	{0xEB3F4171FBBEB90CULL, "19823407.ogg"},
	{0xEB400B5B18301D32ULL, "hlac_lt_hangedman_-1_16x.dds"},
	{0xEB409E59BB3EEB15ULL, "22551059.ogg"},
	{0xEB4107C9A8B40B84ULL, "830208502.ogg"},
	{0xEB4113B0D8012A52ULL, "vxlt_gallowstemplar_45o_0_0.vlm"},
	{0xEB413EB3951C2977ULL, "479370478.ogg"},
	{0xEB415EFADDA00B04ULL, "280618791.ogg"},
	{0xEB42380EE64D6C1DULL, "follower_banter6147417.cl"},
	{0xEB424F27966A82D9ULL, "w_blightlands_137_0.phy"},
	{0xEB425DEEBBDF8D09ULL, "vxdr_0_deeproads_cu7_0_0.vlm"},
	{0xEB427B082B6B21C1ULL, "6092089_f.fxe"},
	{0xEB428CADA18DF3F9ULL, "w_blightlands_61_0.anb"},
	{0xEB42A6C9BFAAADB8ULL, "mcr310_argument6074395.cl"},
	{0xEB42FA049261F8CCULL, "vxlt_gallowstemplar_46x_0_0.vlm"},
	{0xEB43CEAB8F589231ULL, "759012869.ogg"},
	{0xEB43E2F3D3D2F6DCULL, "w_bonepitdestroyed_a3_177_0.mao"},
	{0xEB43EEF3D3D30B7FULL, "w_bonepitdestroyed_a3_177_0.mmh"},
	{0xEB43F4F3D3D31521ULL, "w_bonepitdestroyed_a3_177_0.msh"},
	{0xEB44054C2355AC21ULL, "167392357.ogg"},
	{0xEB440851FE3C3CB1ULL, "vxlt_0_lowtown_a3_n_c_d6f_0_0.vlm"},
	{0xEB4442E3AA7087CBULL, "164208121.ogg"},
	{0xEB44528C56EA34E6ULL, "ht_0_hightownexplosion_n_51_0.rml"},
	{0xEB4492C8A46E6EE8ULL, "6067735_m.fxe"},
	{0xEB4568A4D45EB2D2ULL, "157428935.ogg"},
	{0xEB462700E4967642ULL, "dae000_flemeth6000794.cl"},
	{0xEB46D4D4C0ED7846ULL, "vxlt_backalley_g_1ju_0_0.vlm"},
	{0xEB474E0CF7E47456ULL, "cs_qcr220cs_meredith_arrival.bnk"},
	{0xEB474F20DED0BD06ULL, "6110152_m.fxe"},
	{0xEB479E795698C2EBULL, "6044928_m.fxe"},
	{0xEB485C837CDF1D84ULL, "vxht_0_hightown_a3_ba9_0_0.vlm"},
	{0xEB487E00077AC046ULL, "vxca_caverns1_e59_0_0.vlm"},
	{0xEB48A2199CE29C0BULL, "6085228_m.fxe"},
	{0xEB48B5D890EA0282ULL, "306761083.ogg"},
	{0xEB48DFCE1D6325A4ULL, "zz_mcr_debug6088342.cl"},
	{0xEB48EBB11762A6C7ULL, "hlac_w_encounter2_-131069_0z.dds"},
	{0xEB493D9699CC8E62ULL, "6217376_m.fxe"},
	{0xEB495724F7073C4BULL, "vxw_bonepitmining_a2_245_0_0.vlm"},
	{0xEB49DC25AE6B8959ULL, "188782115.ogg"},
	{0xEB4A3AFD8433E41CULL, "weapon_rendertoshadowmap_p0.vsh.fxo.dx11"},
	{0xEB4A75C43B800D21ULL, "gen00fl_carver6103627.cl"},
	{0xEB4A810C8A4B5557ULL, "992435159.ogg"},
	{0xEB4A94E5B6FEABB5ULL, "mag360_thrask6043897.cl"},
	{0xEB4AC6EAD1D40FF3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_chairxa_0.pwk"},
	{0xEB4B652A0773F393ULL, "vxht_keep_n_6bi_0_0.vlm"},
	{0xEB4BA0F0D523191EULL, "vxlt_gallowscourtyard_7b7_0_0.vlm"},
	{0xEB4C9E9ECFA1AB8BULL, "vxht_mansion2_31n_0_0.vlm"},
	{0xEB4CAD189E814088ULL, "hlac_ht_mansion2_818_28x.dds"},
	{0xEB4CB9689BCCD3B8ULL, "vxlt_undercityencounter_18m_0_0.vlm"},
	{0xEB4D635CC988E840ULL, "6129752_m.fxe"},
	{0xEB4E00FC82904AB7ULL, "hlac_w_sundermount_-393215_0z.dds"},
	{0xEB4E235608FBF7DAULL, "vxht_keep_n_4bf_0_0.vlm"},
	{0xEB4E69D323D3CA9AULL, "593978692.ogg"},
	{0xEB4E92F597B9892FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod.gad"},
	{0xEB4F1C1D125D4715ULL, "vxlt_0_lowtown_docks_bc6_0_0.vlm"},
	{0xEB4F567EE784DC42ULL, "6085179_m.fxe"},
	{0xEB4FAFD0E6E596D3ULL, "hlac_ht_keep_n_3732_2z.dds"},
	{0xEB4FEDEFFD6C9FFEULL, "vxlt_undercityencounter_103_0_0.vlm"},
	{0xEB5000743B63340DULL, "w_encounter3_178_0.mmh"},
	{0xEB5004743B633B5EULL, "w_encounter3_178_0.mao"},
	{0xEB5012743B635293ULL, "w_encounter3_178_0.msh"},
	{0xEB50756809C0D430ULL, "6106434_m.fxe"},
	{0xEB508D82C9864ED9ULL, "6112683_m.fxe"},
	{0xEB509EFB9C3187E5ULL, "vxca_caverns1_hic_0_0.vlm"},
	{0xEB50A900F6197A07ULL, "darktown_fem_fereldan_sta6152805.cl"},
	{0xEB50B38157BD381BULL, "vxlt_0_lowtown_fgi_0_0.vlm"},
	{0xEB50CFCBADD5A589ULL, "vxdr_0_deeproads_cr6_0_0.vlm"},
	{0xEB5111D80DAC14AFULL, "gen00fl_anders6111183.cl"},
	{0xEB5148DCAB6FC3E9ULL, "gen00fl_merrill6130014.cl"},
	{0xEB51D8CAB3E4B91EULL, "dae300_alistair_king6134928.cl"},
	{0xEB525E986F5E3BC6ULL, "77379722.ogg"},
	{0xEB526861397E8AD9ULL, "136213078.ogg"},
	{0xEB52A3A6E059EBCAULL, "qun110_petrice6036420.cl"},
	{0xEB52ABCF34465212ULL, "vxlt_0_lowtown_docks_n_bgk_0_1.vlm"},
	{0xEB52CCF404846527ULL, "vxht_chantry_2e3_0_0.vlm"},
	{0xEB5310FE859AF9CEULL, "6162809_m.fxe"},
	{0xEB53C19D854B822DULL, "w_sundermount_20_0.mmh"},
	{0xEB53C59D854B897EULL, "w_sundermount_20_0.mao"},
	{0xEB53D39D854BA0B3ULL, "w_sundermount_20_0.msh"},
	{0xEB541EE25B495C7FULL, "hlac_lt_0_lowtown_docks_131068_0x.dds"},
	{0xEB5432169651315AULL, "vxw_bonepitdestroyed_a3_1vo_0_0.vlm"},
	{0xEB54500107833DEFULL, "hlac_ca_caverns1_24424_13z.dds"},
	{0xEB5503818D68EDDCULL, "vxht_mansion2b_3g7_0_0.vlm"},
	{0xEB55A79264978F3EULL, "hlac_lt_0_lowtown_65536_0x.dds"},
	{0xEB55A7F597BFCBDCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod.evt"},
	{0xEB5621D9A1D447E6ULL, "vxht_mansion2_29r_0_0.vlm"},
	{0xEB570C5CCDA65BADULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5.mmh"},
	{0xEB571F481B7CA76AULL, "572194338.ogg"},
	{0xEB574F589314BB72ULL, "vxlt_0_lowtown_a2_hdk_0_0.vlm"},
	{0xEB57AA0F36AA14D1ULL, "lt_0_lowtown_a2_n_c_29_0_lc.mmh"},
	{0xEB57EE14FF46ED07ULL, "391935719.ogg"},
	{0xEB580CDFC593E821ULL, "6033689_m.fxe"},
	{0xEB5A3E0F6BFFC974ULL, "dre000_dougal6131005.cl"},
	{0xEB5A3EEE8E52E7DAULL, "steam_03.mao"},
	{0xEB5A7E895106AA49ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flower.mao"},
	{0xEB5A8A895106BEE2ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flower.mmh"},
	{0xEB5A8C895106C198ULL, "art\\levels\\races\\ferelden\\props\\plotspecifc\\prp_flower.msh"},
	{0xEB5A9BAD489D6528ULL, "sw_cl.mao"},
	{0xEB5B3805ECFFF43AULL, "llac_lt_0_lowtown_a2_65534_2.dds"},
	{0xEB5BA98FA7B19D14ULL, "gen00fl_aveline6116742.cl"},
	{0xEB5C1BFA56F4ADD7ULL, "6205709_m.fxe"},
	{0xEB5C654A761A0F57ULL, "462649262.ogg"},
	{0xEB5CC21CE2618A7EULL, "823863105.ogg"},
	{0xEB5CE0654EF4FB10ULL, "vxlt_gallowscourtyard_6og_0_0.vlm"},
	{0xEB5D8497EE16816DULL, "pro000cr_soldier7.utc"},
	{0xEB5DC19D8553D6A6ULL, "w_sundermount_20_0.rml"},
	{0xEB5E254768535DC6ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownenter.ani"},
	{0xEB5E64A5F636953BULL, "vxca_caverns1_mines_imu_0_0.vlm"},
	{0xEB5EDC2491F65286ULL, "566773444.ogg"},
	{0xEB5F61C98AACCFF9ULL, "gen00fl_carver6102660.cl"},
	{0xEB5F66223A82A429ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg101a_0d.dds"},
	{0xEB5FEE3996A9A423ULL, "172911501.ogg"},
	{0xEB60109564D8AC79ULL, "6109687_m.fxe"},
	{0xEB60584974B0C701ULL, "6042227_m.fxe"},
	{0xEB608B424D4A3437ULL, "plt_cod_hst_rivain.plo"},
	{0xEB60F9743B719BCAULL, "w_encounter3_178_0.phy"},
	{0xEB615E2D79B924A1ULL, "mcr371cr_merrill.utc"},
	{0xEB6218C3C54B31DAULL, "368935440.ogg"},
	{0xEB622856216F7E9EULL, "vxht_keep_5iy_0_0.vlm"},
	{0xEB62658B599161AAULL, "hlac_lt_undercityserialkiller_-1_44z.dds"},
	{0xEB62A7CA0AADBE56ULL, "196187673.ogg"},
	{0xEB63125A3EC9FEE0ULL, "llac_lt_0_lowtown_-65537_4.dds"},
	{0xEB635AE118C1BE1EULL, "519490390.ogg"},
	{0xEB63AFF597CC16F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\head_nod.ani"},
	{0xEB641898FCC24F06ULL, "1065148858.ogg"},
	{0xEB6442FA9A2EABDAULL, "scaleform_sf_depth_p0.psh.fxo"},
	{0xEB647238A1402E76ULL, "517815100.ogg"},
	{0xEB64B75FA532CABBULL, "hlac_w_encounter2_camp_-65537_0x.dds"},
	{0xEB65078187EC855BULL, "6106976_m.fxe"},
	{0xEB6528F18C928EE8ULL, "hlac_lt_0_lowtown_a2_n_c_131071_3z.dds"},
	{0xEB65CE9DEAC5C45CULL, "vxlt_gallowstemplar_f_5uv_0_1.vlm"},
	{0xEB667F1BA11AC170ULL, "6092581_f.fxe"},
	{0xEB66CECB59F1BB3CULL, "vxht_bartrand_n_5qs_0_0.vlm"},
	{0xEB674909B5B1BF8FULL, "vxht_0_hightown_n_a2r_0_0.vlm"},
	{0xEB6790013008D955ULL, "vxw_encounter2_g8_0_0.vlm"},
	{0xEB67986C68493410ULL, "251909901.ogg"},
	{0xEB67B5592D6F8515ULL, "qun200_party_comment.bnk"},
	{0xEB67C02B59428020ULL, "hlac_lt_0_lowtown_docks_a3_-131073_5x.dds"},
	{0xEB67F26ECD27CCB4ULL, "hlac_lt_0_lowtown_a2_n_c_-1_8x.dds"},
	{0xEB68055CCDB4C36AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5.phy"},
	{0xEB683F8F9E750020ULL, "gen00fl_fenris6103323.cl"},
	{0xEB6863DE97A8704EULL, "151082234.ogg"},
	{0xEB686A6F2A0B9A42ULL, "6034085_m.fxe"},
	{0xEB686AA49DDAAF29ULL, "hlac_ht_0_hightown_a3_-196608_0x.dds"},
	{0xEB68829C47E899F7ULL, "hlac_lt_0_lowtown_a2_n_131071_0z.dds"},
	{0xEB68E6687BCD999FULL, "qun221_petrice6077458.cl"},
	{0xEB690289CDA3822BULL, "dae011_corff_sta6149612.cl"},
	{0xEB694DB9F4F62668ULL, "gen00fl_aveline6139077.cl"},
	{0xEB69B6BC005B5A3BULL, "6139120_f.fxe"},
	{0xEB69B87DB65F7691ULL, "vxlt_0_lowtown_e2c_0_0.vlm"},
	{0xEB69FBDF61F0771BULL, "fxc_smoke_effect.mao"},
	{0xEB6A2B8C56280533ULL, "mag101_grace6034057.cl"},
	{0xEB6A698A35C0EF6DULL, "vfx_vfx_p4.psh.fxo"},
	{0xEB6AA8DFBF8E2E7AULL, "hlac_ht_chantry_n_2534_29y.dds"},
	{0xEB6B53484652619DULL, "525042092.ogg"},
	{0xEB6BBED7B85DAF51ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_6y.dds"},
	{0xEB6BBF0C794BCE91ULL, "6044532_f.fxe"},
	{0xEB6BD7592D73A86DULL, "qun200_party_comment.cnv"},
	{0xEB6BE0FE8185AA3EULL, "vxlt_hangedman_n_25g_0_0.vlm"},
	{0xEB6C09ABD0D8DE2AULL, "gen_im_arm_hlm_hvy_lore_01.uti"},
	{0xEB6C0D47685F7237ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownenter.evt"},
	{0xEB6C21579F7E48B5ULL, "52290553.ogg"},
	{0xEB6C7AA009CBB213ULL, "hlac_ht_0_hightown_n_65538_0x.dds"},
	{0xEB6C816BC3508CA4ULL, "gen00fl_sebastian6189663.cl"},
	{0xEB6CB941700A51B6ULL, "681528912.ogg"},
	{0xEB6CD682E82CADBBULL, "vxca_caverns1_mines_ijt_0_0.vlm"},
	{0xEB6CF08E81D45F71ULL, "6151075_m.fxe"},
	{0xEB6CF0F2931097FEULL, "cli_army_ds_redcliffe.gda"},
	{0xEB6D27BF8159EC97ULL, "hlac_lt_0_lowtown_n_-262143_0z.dds"},
	{0xEB6DF6C44CB7F395ULL, "458576475.ogg"},
	{0xEB6EA95A64F41F92ULL, "art\\characters\\playercharacter\\humanfemale\\hlm_masc.mao"},
	{0xEB6F6222821F796FULL, "hlac_ht_0_hightown_65534_0y.dds"},
	{0xEB6FB0B343A9EECDULL, "833991525.ogg"},
	{0xEB6FB1E26988FDC5ULL, "hlac_ca_caverns1_24423_10z.dds"},
	{0xEB6FD89B193C610DULL, "599208215.ogg"},
	{0xEB706B21C40C0103ULL, "88743791.ogg"},
	{0xEB7119FD3BAFEC88ULL, "vxlt_gallowstemplar_f_4un_0_0.vlm"},
	{0xEB71C42972A59D53ULL, "gen00fl_anders6116155.cl"},
	{0xEB722928843AFCECULL, "gen00fl_aveline6114854.cl"},
	{0xEB72630FE03C9A91ULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_bigtreasure_sa.dds"},
	{0xEB72D420FB61E4CDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_jaw04a.phy"},
	{0xEB72E1B645B32680ULL, "6116970_m.fxe"},
	{0xEB730096482D3BDEULL, "llac_dr_0_deeproads_5845_4.dds"},
	{0xEB730E47686592C0ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\knockdownenter.gad"},
	{0xEB7316B690A2CBF6ULL, "6172479_m.fxe"},
	{0xEB73D3879680052FULL, "ss_templars_m_1.wbd"},
	{0xEB740E6DB2F38119ULL, "dae000_bartrand6033996.cl"},
	{0xEB74280649031D6DULL, "376586286.ogg"},
	{0xEB74730254FCDE93ULL, "6104884_m.fxe"},
	{0xEB7540A84BE7D3BCULL, "vxht_bartrand_n_5tp_0_0.vlm"},
	{0xEB75B4B61D39B52CULL, "hlac_dr_slavers_3216_1y.dds"},
	{0xEB75E8270C297813ULL, "vrc221_bartrand6065270.cl"},
	{0xEB763E7F2906CC27ULL, "vxdr_0_deeproads_fof_0_0.vlm"},
	{0xEB765E6D01F5AC48ULL, "w_bonepit_71_0.msh"},
	{0xEB76706D01F5CAD9ULL, "w_bonepit_71_0.mao"},
	{0xEB767C6D01F5DF72ULL, "w_bonepit_71_0.mmh"},
	{0xEB76D1E57C891135ULL, "gen00fl_anders6125753.cl"},
	{0xEB776D9134E2EF01ULL, "vxht_mansion2b_3gf_0_0.vlm"},
	{0xEB77750442CB10C9ULL, "6088402_f.fxe"},
	{0xEB7882E96E70CFACULL, "1000326512.ogg"},
	{0xEB793901CD0BF920ULL, "gen00fl_aveline6114014.cl"},
	{0xEB79A99DE9CCB334ULL, "mrl150_merrill6077672.cl"},
	{0xEB79AD37481D93EFULL, "6149505_f.fxe"},
	{0xEB79C7E2CC22DC21ULL, "hlac_lt_0_lowtown_n_-1_5z.dds"},
	{0xEB7A0667B50C84F1ULL, "184465768.ogg"},
	{0xEB7A4D3242ABCB6EULL, "zz_start_mcr6081959.cl"},
	{0xEB7B13848123743BULL, "hlac_lt_0_lowtown_docks_a3_-65539_2y.dds"},
	{0xEB7B8DADA1BDBAF9ULL, "w_blightlands_61_0.rml"},
	{0xEB7BA35874DF91ECULL, "mrl211ip_mirror.utp"},
	{0xEB7C817DC4BCD7F0ULL, "591855236.ogg"},
	{0xEB7CA3F944AE4A69ULL, "follower_barks6169898.cl"},
	{0xEB7CF44F48C4DD79ULL, "6107887_m.fxe"},
	{0xEB7D8AC425DF1DD0ULL, "mag211st_marethari_arianni.stg"},
	{0xEB7DDBC61876FD60ULL, "vxw_sundermount_34k_0_1.vlm"},
	{0xEB7E177BBDDF4690ULL, "vxht_mansion2_3c7_0_0.vlm"},
	{0xEB7F35B18986D854ULL, "511789715.ogg"},
	{0xEB7F87AD6C337ADAULL, "vxw_bonepitdestroyed_a3_1aj_0_0.vlm"},
	{0xEB8094B9C8B106B2ULL, "hlac_lt_gallowstemplar_f_2384_6x.dds"},
	{0xEB80EA7E07D7917CULL, "480738363.ogg"},
	{0xEB80F3D1E09DF339ULL, "193052106.ogg"},
	{0xEB81C55E71514D98ULL, "hlac_lt_0_lowtown_n_-131073_1y.dds"},
	{0xEB823182CB9DD738ULL, "vxlt_0_lowtown_docks_a3_d2q_0_0.vlm"},
	{0xEB8283694D1C3360ULL, "168519491.ogg"},
	{0xEB82981CC553209AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urvb_0d.dds"},
	{0xEB831114C886518CULL, "mcr370_varric6099732.cl"},
	{0xEB832AD74148394EULL, "6204970_m.fxe"},
	{0xEB83618EE7582875ULL, "151124429.ogg"},
	{0xEB8436E75171200CULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_006a_0s.dds"},
	{0xEB845597F156C94BULL, "art\\vfx\\combat\\vfx_w_wns_right_close1_c.impa.ani"},
	{0xEB85B02E89883F82ULL, "zz_end_braveheart.cut"},
	{0xEB85BCDBA26B07F0ULL, "vxht_keep_n_3ir_0_0.vlm"},
	{0xEB85E6E134C12949ULL, "6083502_f.fxe"},
	{0xEB85F2AAB39DB97EULL, "109399254.ogg"},
	{0xEB8658EC54AC7BB2ULL, "llac_lt_0_lowtown_a2_0_3.dds"},
	{0xEB865DBD7359C0BBULL, "vxht_mansion2b_31f_0_0.vlm"},
	{0xEB86DAA268E00E67ULL, "263234540.ogg"},
	{0xEB86E6BDCA9B4B19ULL, "626567928.ogg"},
	{0xEB87251D6DE22853ULL, "100966250.ogg"},
	{0xEB8734D7B014F704ULL, "vxca_cavern1_haunted_imn_0_0.vlm"},
	{0xEB8740D470DF02D3ULL, "187045903.ogg"},
	{0xEB87573EE47722BEULL, "vxht_bartrand_n_4ba_0_0.vlm"},
	{0xEB8787D89622A7FDULL, "vxdr_primevaltomb_4zr_0_1.vlm"},
	{0xEB87E6A3F3F83A70ULL, "42798457.ogg"},
	{0xEB87F837EF057DD5ULL, "hlac_w_woundedcoast_-196607_0z.dds"},
	{0xEB880BC148061CE8ULL, "6122797_m.fxe"},
	{0xEB884EE16C046F0CULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_var.msh"},
	{0xEB8854E16C0479FEULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_var.mmh"},
	{0xEB8860E16C048D95ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_var.mao"},
	{0xEB8895263A2D9300ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunaritrooper_movement.anb"},
	{0xEB88C7A74EB1DE8EULL, "hlac_lt_0_lowtown_a2_-131072_6z.dds"},
	{0xEB88E964BEB00733ULL, "hlac_w_sundermount_-262148_0x.dds"},
	{0xEB890F3543C6EBDFULL, "6174164_m.fxe"},
	{0xEB89485F96DB6E21ULL, "vxlt_0_lowtown_docks_arr_0_0.vlm"},
	{0xEB89B427D80D3991ULL, "dae000_gamlen6145112.cl"},
	{0xEB89C60A13065BF6ULL, "vxlt_0_lowtown_docks_n_btg_0_0.vlm"},
	{0xEB89FC6EF714151AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh_eyesm1.mao"},
	{0xEB8A580B70A5E678ULL, "vxlt_0_lowtown_a2_n_el9_0_2.vlm"},
	{0xEB8A605D26BCDEDFULL, "dae000_elegant6166922.cl"},
	{0xEB8A9B2B4E1232FAULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_f_c.anb"},
	{0xEB8AA362FE42507EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_right_down.ani"},
	{0xEB8AC6566228455BULL, "war_whirlwind.dds"},
	{0xEB8B089840EB2545ULL, "6040764_m.fxe"},
	{0xEB8B340E085712AEULL, "vxlt_0_lowtown_docks_a3_cxf_0_0.vlm"},
	{0xEB8B3D13625CDB9DULL, "vxht_0_hightown_a3_9p4_0_0.vlm"},
	{0xEB8BE407E167ABA1ULL, "llac_ht_bartrand_n_3374_1.dds"},
	{0xEB8BFF8A5D8976D8ULL, "llac_lt_0_lowtown_a2_n_131070_4.dds"},
	{0xEB8C4DA30A6D15E0ULL, "vxw_sundermount_33j_0_1.vlm"},
	{0xEB8C8958AFD55F10ULL, "vxht_mansion2_3f4_0_0.vlm"},
	{0xEB8DE4F01349E212ULL, "1048658754.ogg"},
	{0xEB8E512639518410ULL, "vxlt_gallowscourtyard_n_7h8_0_0.vlm"},
	{0xEB8E8BD944B11CF9ULL, "dae000_merchant_rare_em6177937.cl"},
	{0xEB8F0C47C38299E5ULL, "vxlt_0_lowtown_docks_bfe_0_0.vlm"},
	{0xEB8F86BEAA6CDF9EULL, "vxw_bonepitdestroyed_a3_1s4_0_0.vlm"},
	{0xEB903B298186CF77ULL, "vxca_caverns1_mines_ima_0_0.vlm"},
	{0xEB9049D4E536CCACULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_014a_0t.dds"},
	{0xEB9085007F73D864ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_hoo_chac_d.dds"},
	{0xEB90CA967B709C80ULL, "345709695.ogg"},
	{0xEB90E5C5390E6F20ULL, "gen00fl_varric6076012.cl"},
	{0xEB914A301D895E30ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_isab_0n.dds"},
	{0xEB914B58FC4B0478ULL, "dae000ip_crowd_mf3_talk18.utp"},
	{0xEB916F14010552A8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\prx_black_d.dds"},
	{0xEB91EB37C5A050D5ULL, "617768839.ogg"},
	{0xEB9216A5AB8C12ECULL, "6041300_m.fxe"},
	{0xEB9222D964F89F25ULL, "vxht_mansion2b_3dg_0_0.vlm"},
	{0xEB923D97F162DDC2ULL, "art\\vfx\\combat\\vfx_w_wns_right_close1_c.impa.evt"},
	{0xEB93404C0518C42AULL, "vxw_woundedcoast_72_0_0.vlm"},
	{0xEB937DEDD4F6BBF5ULL, "vxht_0_hightown_avg_0_0.vlm"},
	{0xEB93B0F538AE7FDDULL, "vxdr_primevaltomb_6nm_0_0.vlm"},
	{0xEB93C0800472C4FEULL, "art\\characters\\weapons\\staff\\textures\\w_stf_013a_0n.dds"},
	{0xEB93C725DB47C653ULL, "128008040.ogg"},
	{0xEB94377E01CAA262ULL, "vxca_caverns1_mines_ih7_0_0.vlm"},
	{0xEB94ACD1F75D4446ULL, "dae000_pc_home_ambients6208543.cl"},
	{0xEB94B1DA880B4DCEULL, "vxht_0_hightown_n_a70_0_0.vlm"},
	{0xEB94BEE31C2C99FDULL, "vxlt_darktown_44c_0_0.vlm"},
	{0xEB95728CFBB73359ULL, "vxht_mansion2b_264_0_0.vlm"},
	{0xEB958AB491C57C33ULL, "vxca_caverns1_mines_edz_0_0.vlm"},
	{0xEB95A6B4A20B0F84ULL, "vxca_cavern1_haunted_ijo_0_0.vlm"},
	{0xEB95E32E1DC68DBBULL, "vxlt_gallowstemplar_6sf_0_0.vlm"},
	{0xEB9637702470EE8BULL, "art\\levels\\races\\dwarf\\props\\decor\\prp_potb_0.pwk"},
	{0xEB967E404AD98114ULL, "731543049.ogg"},
	{0xEB968E89F77491AFULL, "758182381.ogg"},
	{0xEB96BFD38F44BAB6ULL, "w_blightlands_84_0.mao"},
	{0xEB96C5E83BD08A70ULL, "zz_dre_debug6056980.cl"},
	{0xEB96CBD38F44CF55ULL, "w_blightlands_84_0.mmh"},
	{0xEB96CDD38F44D2CBULL, "w_blightlands_84_0.msh"},
	{0xEB96DEB27CF3EA2EULL, "vxw_woundedcoast_10_0_0.vlm"},
	{0xEB9715F683D7B6D7ULL, "vrc221_bartrand6043245.cl"},
	{0xEB977BF6766E8164ULL, "dae000_sandal6147783.cl"},
	{0xEB97B9CCA1F1DDEFULL, "633662346.ogg"},
	{0xEB97EC0F49124018ULL, "cli_army_ds_market.gda"},
	{0xEB98179CA9CE47E7ULL, "vxlt_0_lowtown_ets_0_0.vlm"},
	{0xEB982C19BE0EF73BULL, "vxca_caverns1_mines_iyi_0_0.vlm"},
	{0xEB9837E704FC7476ULL, "vxlt_0_lowtown_docks_n_bqh_0_0.vlm"},
	{0xEB983C97F16747D5ULL, "art\\vfx\\combat\\vfx_w_wns_right_close1_c.impa.gad"},
	{0xEB984E5DD5F1250DULL, "vxca_cave_17x_0_0.vlm"},
	{0xEB98BB62FE4EB64FULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_right_down.evt"},
	{0xEB98FFA9D3164EF0ULL, "qcr000_arishok6042475.cl"},
	{0xEB992CBFA5662502ULL, "vxht_0_hightown_a2_b1b_0_0.vlm"},
	{0xEB9A038677AA1013ULL, "6149894_m.fxe"},
	{0xEB9A042AB75660D5ULL, "ser191_emeric6108821.cl"},
	{0xEB9A1BCE7CF91241ULL, "hlac_lt_gallowscourtyard_n_-65536_0x.dds"},
	{0xEB9A98B43FCAE6D1ULL, "ht_0_hightownexplosion_n_49_0.rml"},
	{0xEB9AFAA759E90A9FULL, "dae000_merchant_access_hf6162988.cl"},
	{0xEB9B212F66A575D5ULL, "gen_im_trs_cot_junk2_04.uti"},
	{0xEB9B6A598133A941ULL, "297014486.ogg"},
	{0xEB9BADBB1ED88EE1ULL, "165073751.ogg"},
	{0xEB9BB968502E29A8ULL, "6087698_m.fxe"},
	{0xEB9BCDD28B809BB9ULL, "vxht_0_hightown_c3r_0_0.vlm"},
	{0xEB9BE302E898B565ULL, "vxw_bonepitdestroyed_a3_1qq_0_0.vlm"},
	{0xEB9BEFB574D7BAACULL, "vxlt_warehouse_n_5y_0_0.vlm"},
	{0xEB9C2BBEC838A23FULL, "hlac_lt_0_lowtown_a2_65534_3x.dds"},
	{0xEB9C6B67501FABDAULL, "vxw_bonepitdestroyed_a3_1od_0_0.vlm"},
	{0xEB9CBDBCDE257BF2ULL, "dae200cr_viscount6087161.cl"},
	{0xEB9D3FF0E65D1C3DULL, "hlac_lt_0_lowtown_docks_a3_n_131071_0z.dds"},
	{0xEB9D579450D101C3ULL, "546819561.ogg"},
	{0xEB9D96F4898A30B6ULL, "vxht_mansion_n_5ca_0_1.vlm"},
	{0xEB9FBC62FE54D6F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_quick_rct_die_right_down.gad"},
	{0xEBA01D857A321912ULL, "6112889_f.fxe"},
	{0xEBA029B06A67E0C2ULL, "mcr310_argument6216179.cl"},
	{0xEBA047AE470209A1ULL, "hlac_lt_0_lowtown_a3_n_c_65537_5y.dds"},
	{0xEBA0AE6136074056ULL, "gen_im_trs_drk_two_2ha_r0.uti"},
	{0xEBA179E96DDB6FD2ULL, "vxht_brothel_n_5j0_0_0.vlm"},
	{0xEBA1FB4D5AEA1279ULL, "6116131_m.fxe"},
	{0xEBA20411E85EFFB8ULL, "6027775_m.fxe"},
	{0xEBA215A46A5CB49DULL, "6096305_m.fxe"},
	{0xEBA222D22AA4985DULL, "vxdr_primevaltomb_6ul_0_0.vlm"},
	{0xEBA2415E70C51902ULL, "mag131_keran6121101.cl"},
	{0xEBA2966A5BF268EEULL, "hlac_lt_0_lowtown_docks_-65538_6z.dds"},
	{0xEBA2CC3159DF6343ULL, "hlac_lt_0_lowtown_-65535_3z.dds"},
	{0xEBA3E22B2FE4A1B8ULL, "6157186_m.fxe"},
	{0xEBA411F3B4E7B4D2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\_textures\\til_gallows_int2_da.dds"},
	{0xEBA4189194012804ULL, "vxca_cavern1_haunted_igl_0_0.vlm"},
	{0xEBA43AF8C86353BEULL, "vxht_bartrand_n_4lc_0_0.vlm"},
	{0xEBA46202D9293004ULL, "6125266_m.fxe"},
	{0xEBA5B95416E6A251ULL, "mag371_elsa6041558.cl"},
	{0xEBA66B109E7CE9D3ULL, "vxht_brothel_n_5m4_0_0.vlm"},
	{0xEBA689799BC46067ULL, "vxlt_0_lowtown_eor_0_0.vlm"},
	{0xEBA68CBFF08541EBULL, "gen00fl_merrill6161652.cl"},
	{0xEBA6F20A2AF2C498ULL, "fx_storm_lit.mao"},
	{0xEBA77BBF8E85E7DBULL, "6141586_m.fxe"},
	{0xEBA7D7A1CF20F0C6ULL, "vxw_bonepitmining_a2_27i_0_0.vlm"},
	{0xEBA8756369A02893ULL, "6142994_m.fxe"},
	{0xEBA9279AB103AAE5ULL, "hlac_ca_caverns1_24424_18y.dds"},
	{0xEBA9D6ADCEDAB4CCULL, "vrc221_varric6065960.cl"},
	{0xEBAA1321A8BC380BULL, "gen00fl_aveline6120957.cl"},
	{0xEBAA28105DF5FE50ULL, "vxdr_primevaltomb_7kf_0_0.vlm"},
	{0xEBAA3FAF7D76B439ULL, "vxht_0_hightown_c6q_0_0.vlm"},
	{0xEBAA54DFDA8ECDE5ULL, "vxw_bonepitdestroyed_a3_1zp_0_0.vlm"},
	{0xEBAAB6633879652BULL, "gen00fl_fenris6126969.cl"},
	{0xEBAAC47FED185946ULL, "vxw_bonepitmining_a2_1v1_0_0.vlm"},
	{0xEBAB4339EB7FD47FULL, "vxw_bonepitmining_a2_1wz_0_1.vlm"},
	{0xEBAB736250C8DC6AULL, "207547607.ogg"},
	{0xEBAB9EBFF5AF3B02ULL, "807360712.ogg"},
	{0xEBABBE29456CAEE5ULL, "vxca_cavern1_haunted_i6v_0_0.vlm"},
	{0xEBAC81384CDC1B39ULL, "dae000_arishok6068560.cl"},
	{0xEBAD00803EACD6EEULL, "l_dae_hawke_tavern_name_trigger.wbd"},
	{0xEBAD9FE98D8FA231ULL, "vxw_encounter2_camp_8q_0_0.vlm"},
	{0xEBAE0CF85DEB2D72ULL, "dae000_surprises6162512.cl"},
	{0xEBAE708A3455F267ULL, "vxca_cave_1cv_0_0.vlm"},
	{0xEBAE7116D422B075ULL, "6025329_m.fxe"},
	{0xEBAE98AA1097AC69ULL, "6153657_m.fxe"},
	{0xEBAF081C13C035BCULL, "vxht_bartrand_n_5ht_0_0.vlm"},
	{0xEBAF27AA6DAD1DACULL, "986905670.ogg"},
	{0xEBAF8F79ECD1E619ULL, "64749072.ogg"},
	{0xEBAFAF93216377F3ULL, "vxca_cavern1_haunted_ehn_0_0.vlm"},
	{0xEBB002DD4E18EEEAULL, "art\\characters\\weapons\\promo\\longswords\\w_lsw_025a.mao"},
	{0xEBB006DD4E18F539ULL, "art\\characters\\weapons\\promo\\longswords\\w_lsw_025a.mmh"},
	{0xEBB0097C32B53EC1ULL, "art\\characters\\playercharacter\\elfmale\\em_har_ha1a.phy"},
	{0xEBB010DD4E190627ULL, "art\\characters\\weapons\\promo\\longswords\\w_lsw_025a.msh"},
	{0xEBB050EB8F40A3F4ULL, "vxdr_primevaltomb_73z_0_0.vlm"},
	{0xEBB07FA3502A920BULL, "124611938.ogg"},
	{0xEBB09351B8B64C79ULL, "6082216_m.fxe"},
	{0xEBB094AF1C9AB0DDULL, "vxdr_primevaltomb_6xk_0_0.vlm"},
	{0xEBB0F5D1F1A19729ULL, "6065800_m.fxe"},
	{0xEBB11272784D5170ULL, "vxht_keep_n_3xq_0_0.vlm"},
	{0xEBB14FE6419EA7E9ULL, "6166154_m.fxe"},
	{0xEBB18842AE53FB58ULL, "mag320_orsino6070435.cl"},
	{0xEBB222F58403BFBBULL, "vxca_caverns1_mines_j23_0_0.vlm"},
	{0xEBB27314CC14132CULL, "isa321ip_brothel_door.utp"},
	{0xEBB2D3DFCB1F4884ULL, "6122366_m.fxe"},
	{0xEBB3280B0E53A5C2ULL, "mag110_samson6049734.cl"},
	{0xEBB32EB05909C843ULL, "vxlt_0_lowtown_docks_a3_n_d4o_0_0.vlm"},
	{0xEBB381F7B471BE42ULL, "877659592.ogg"},
	{0xEBB39E2B4E353CADULL, "art\\vfx\\combat\\vfx_m_sta_ele_rng_f_c.mmh"},
	{0xEBB3DA2F41D26A42ULL, "lt_0_lowtown_docks_a3_n_235_0.msh"},
	{0xEBB3DC2F41D26DF8ULL, "lt_0_lowtown_docks_a3_n_235_0.mmh"},
	{0xEBB3E82F41D2819BULL, "lt_0_lowtown_docks_a3_n_235_0.mao"},
	{0xEBB414CDFD9C11FCULL, "268200244.ogg"},
	{0xEBB4210344495F59ULL, "dae100_alistair_drunk6135101.cl"},
	{0xEBB49F831E938F4FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv3.gad"},
	{0xEBB4D61F4AED4433ULL, "904327048.ogg"},
	{0xEBB4ED8796B7C805ULL, "ss_templars_m_1.bnk"},
	{0xEBB5137CAA0EF7E1ULL, "6078529_m.fxe"},
	{0xEBB551C780637E0DULL, "462191857.ogg"},
	{0xEBB5F051B61434BBULL, "6117519_f.fxe"},
	{0xEBB5F37D4B4A2D2FULL, "ser120_ghyslain6104821.cl"},
	{0xEBB6497EC1170946ULL, "vxw_bonepitmining_a2_22j_0_0.vlm"},
	{0xEBB67819E91536C7ULL, "lt_0_lowtown_a2_n_32_0_lc.mmh"},
	{0xEBB695922922AFF7ULL, "571845214.ogg"},
	{0xEBB6C2266530779DULL, "370287721.ogg"},
	{0xEBB773470737CAD8ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_glv_hvyo_0s.dds"},
	{0xEBB7B3E5951FC024ULL, "hla_ht_0_hightown_n_0_0y.dds"},
	{0xEBB7E69C7C06C105ULL, "dae000ip_crowd_mf2_talk16.utp"},
	{0xEBB83440FEA9FAF7ULL, "vxht_mansion2b_31r_0_0.vlm"},
	{0xEBB855B4C19855FBULL, "553036564.ogg"},
	{0xEBB8690E4518FE95ULL, "art\\characters\\weapons\\unique\\meredith\\textures\\w_gsw_014a_0e.dds"},
	{0xEBB86E5A1ECE9E20ULL, "vxlt_gallowstemplar_4xz_0_0.vlm"},
	{0xEBB898827582F3E7ULL, "hlac_w_encounter3_-393213_0x.dds"},
	{0xEBB9365CDF0E71C6ULL, "vxw_bonepitmining_a2_1s2_0_0.vlm"},
	{0xEBB9CFDB43220AB8ULL, "534616719.ogg"},
	{0xEBBA08AB1607F8EFULL, "vxlt_0_lowtown_a2_n_c_fib_0_0.vlm"},
	{0xEBBA099BCDA2C6ACULL, "w_blightlands_119_0.msh"},
	{0xEBBA0F9BCDA2D09EULL, "w_blightlands_119_0.mmh"},
	{0xEBBA1B9BCDA2E535ULL, "w_blightlands_119_0.mao"},
	{0xEBBA1C26150DB428ULL, "mcr310_sibling6076309.cl"},
	{0xEBBAC7990EB39E2DULL, "894254672.ogg"},
	{0xEBBADC55804F291EULL, "vxw_bonepitdestroyed_a3_1b7_0_0.vlm"},
	{0xEBBB3EE43F8EF366ULL, "vxht_mansion2_2dy_0_0.vlm"},
	{0xEBBB534947D09CF6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk2_d.dds"},
	{0xEBBBA899AA35BAACULL, "6054188_m.fxe"},
	{0xEBBBB4831E99D1FCULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv3.evt"},
	{0xEBBBD05D057642FDULL, "llac_lt_0_lowtown_docks_a3_-196610_6.dds"},
	{0xEBBC11C67F85BAB1ULL, "vxw_encounter2_camp_3p_0_0.vlm"},
	{0xEBBC167E08C3D537ULL, "111100898.ogg"},
	{0xEBBC535AE29BC0CCULL, "6179393_m.fxe"},
	{0xEBBCB3E2633B39EDULL, "vxht_mansion2b_329_0_0.vlm"},
	{0xEBBCCA2F134DDC35ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemb_bdy_0d.dds"},
	{0xEBBD472A0C413A47ULL, "770479126.ogg"},
	{0xEBBD8A48ECFD325CULL, "515428975.ogg"},
	{0xEBBDB697C07BBC26ULL, "6202356_m.fxe"},
	{0xEBBDE4DA2EF76342ULL, "vxlt_gallowscourtyard_6zv_0_0.vlm"},
	{0xEBBE274893AB3060ULL, "qun110_petrice6041297.cl"},
	{0xEBBE9A6C4BC8EA1CULL, "hlac_lt_undercityserialkiller_-1_24z.dds"},
	{0xEBBFDA90AD4030D3ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180r.evt"},
	{0xEBC03356C8F87CE8ULL, "vxlt_0_lowtown_a3_n_c_feh_0_0.vlm"},
	{0xEBC043434443840DULL, "vxlt_undercitydungeon_10i_0_0.vlm"},
	{0xEBC064BBB08CDBD5ULL, "plt_isa000pt_gift_ship.nss"},
	{0xEBC0963E83788414ULL, "zz_rdr_debug6206281.cl"},
	{0xEBC0DF37D00A01B7ULL, "mer140_bonepit_comments6077810.cl"},
	{0xEBC17AAFFC1E8147ULL, "qcr000_arishok6059975.cl"},
	{0xEBC202F25FAAA64CULL, "372956440.ogg"},
	{0xEBC214C56800690BULL, "vxw_encounter3_6n_0_0.vlm"},
	{0xEBC25AF6471EBD43ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\standing_cheering_1.ani"},
	{0xEBC32A998BD873E7ULL, "gen00fl_aveline6152574.cl"},
	{0xEBC34AD8C40448B3ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_f_c.impa.ani"},
	{0xEBC39800227A1B43ULL, "art\\characters\\creatures\\mubari\\animation\\com_scratch.evt"},
	{0xEBC3C5A72D64C462ULL, "vxlt_0_lowtown_a2_n_e85_0_0.vlm"},
	{0xEBC3F5CB15FFC7DAULL, "mag110_samson6025541.cl"},
	{0xEBC423F002ABD494ULL, "w_bonepitmining_a2_32_0.rml"},
	{0xEBC478B8C73C13AAULL, "vxlt_backalley_g_1jy_0_0.vlm"},
	{0xEBC4BB5BB30D21C6ULL, "vxw_bonepitmining_a2_29k_0_0.vlm"},
	{0xEBC50F82B6520707ULL, "vxlt_hangedman_14q_0_0.vlm"},
	{0xEBC563D69F84B4F3ULL, "gen00fl_aveline6132842.cl"},
	{0xEBC5BD25C3C3F0CBULL, "145473359.ogg"},
	{0xEBC621E40DC95BAAULL, "vxca_caverns1_e55_0_0.vlm"},
	{0xEBC65AEA502669E8ULL, "w_encounter2_camp_31_0.msh"},
	{0xEBC66CEA50268879ULL, "w_encounter2_camp_31_0.mao"},
	{0xEBC678EA50269C12ULL, "w_encounter2_camp_31_0.mmh"},
	{0xEBC68EF2C4FB53C0ULL, "hlac_lt_gallowstemplar_2384_7z.dds"},
	{0xEBC6D390AD464314ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180r.gad"},
	{0xEBC755302B1EACCDULL, "6207554_m.fxe"},
	{0xEBC77B79FF532EBAULL, "vxht_mansion2_2k4_0_0.vlm"},
	{0xEBC79BA05BE28B3FULL, "170216932.ogg"},
	{0xEBC7A00304F3D611ULL, "mag100_thrask6031751.cl"},
	{0xEBC7F36009227256ULL, "dae000_arishok6067735.cl"},
	{0xEBC8293E11EDCAD5ULL, "67512684.ogg"},
	{0xEBC8862E4D38594CULL, "gen00fl_anders6113089.cl"},
	{0xEBC88B59FE54E676ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_en.evt"},
	{0xEBC8928D57307163ULL, "isa111_isabela6075612.cl"},
	{0xEBC8B6A885CD71DEULL, "sto_rock.mao"},
	{0xEBC8BC831EA46915ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\enterv3.ani"},
	{0xEBC8FD03434518AFULL, "m_dae340ar_bone_pit.wbd"},
	{0xEBC90097C047D7C7ULL, "hla_ht_0_hightown_n_-131073_0z.dds"},
	{0xEBC92BE16C3BCA99ULL, "art\\characters\\playercharacter\\dwarfmale\\bdy_var.phy"},
	{0xEBCA2255CDC991E5ULL, "455927568.ogg"},
	{0xEBCA2AC8B934A1F7ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.cess.gad"},
	{0xEBCA323CD1D8730FULL, "vxca_caverns2_500_0_0.vlm"},
	{0xEBCA7100227FF7C4ULL, "art\\characters\\creatures\\mubari\\animation\\com_scratch.gad"},
	{0xEBCC5EAB0B746EC5ULL, "one130ar_disciple_dungeon.are"},
	{0xEBCC9F5C319F3398ULL, "vxw_woundedcoast_1ku_0_0.vlm"},
	{0xEBCCF29332AB0C59ULL, "llac_dr_slavers_3207_7.dds"},
	{0xEBCD49A726405FE3ULL, "art\\characters\\creatures\\rubblegolem\\textures\\c_rbg_0n.dds"},
	{0xEBCD593F017970B6ULL, "191735325.ogg"},
	{0xEBCDACAC25E097CDULL, "73408232.ogg"},
	{0xEBCDB4FB75E2637AULL, "vxlt_gallowscourtyard_4xf_0_0.vlm"},
	{0xEBCDE290AD4C7BF2ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_ctn180r.ani"},
	{0xEBCE1FE0623A487BULL, "339311274.ogg"},
	{0xEBCE4F1B6840227AULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatebuilding.msh"},
	{0xEBCE5D1B684039B3ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatebuilding.mao"},
	{0xEBCE611B684040C0ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\vista_gatebuilding.mmh"},
	{0xEBCE8259FE594201ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_en.gad"},
	{0xEBCEA05FAD237F53ULL, "522749170.ogg"},
	{0xEBCF698A9D4139E3ULL, "hlac_ca_caverns1_24423_13y.dds"},
	{0xEBCFAE753A5E7482ULL, "vxht_0_hightown_b78_0_0.vlm"},
	{0xEBCFD355E2E71331ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar3_scaredmage.gad"},
	{0xEBCFEF9827D4508DULL, "vxca_cavern1_haunted_hmf_0_0.vlm"},
	{0xEBD00FE99EF91FA2ULL, "6143387_m.fxe"},
	{0xEBD072F6472B233AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\standing_cheering_1.evt"},
	{0xEBD0800022847CA2ULL, "art\\characters\\creatures\\mubari\\animation\\com_scratch.ani"},
	{0xEBD086A259F6818BULL, "vxw_encounter3_3a_0_0.vlm"},
	{0xEBD0F45DBE030CA1ULL, "zz_followers_banter6158118.cl"},
	{0xEBD10DCA0B84E9B1ULL, "llac_lt_darktown_-1_10.dds"},
	{0xEBD13FC8B93AE384ULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.cess.evt"},
	{0xEBD142D8C41078EAULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_f_c.impa.evt"},
	{0xEBD166D472C6E72EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowoff.msh"},
	{0xEBD174D472C6FF7FULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowoff.mao"},
	{0xEBD180D472C7131CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_candlerowoff.mmh"},
	{0xEBD1DDBC24CF7BFFULL, "743853933.ogg"},
	{0xEBD2EA95B9322C2AULL, "vxlt_backalley_g_1wz_0_0.vlm"},
	{0xEBD30B13613A9AF5ULL, "gencr_deepstalker.utc"},
	{0xEBD32110416F1532ULL, "hm_bran.mor"},
	{0xEBD370EC97C204EDULL, "ave331_post_fight6138802.cl"},
	{0xEBD3BFC3CBCAFF44ULL, "6067065_m.fxe"},
	{0xEBD3C17E5C805B3EULL, "6136941_m.fxe"},
	{0xEBD4794D6AC63591ULL, "ser221_party_comments6111562.cl"},
	{0xEBD4B1ECB3E5AF0DULL, "mag110_samson6049814.cl"},
	{0xEBD4DEDFBD14E84EULL, "6034687_m.fxe"},
	{0xEBD52FA92A7603C0ULL, "vxlt_0_lowtown_docks_b64_0_0.vlm"},
	{0xEBD55481E202AD16ULL, "6175661_m.fxe"},
	{0xEBD58359FE5F62AFULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_en.ani"},
	{0xEBD59DFCE174269EULL, "vxdr_primevaltomb_8n4_0_0.vlm"},
	{0xEBD65FAB0B7CC4ABULL, "one130ar_disciple_dungeon.ncs"},
	{0xEBD671F6472F8D2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\standing_cheering_1.gad"},
	{0xEBD686830ED2834BULL, "6044872_m.fxe"},
	{0xEBD6D455E2ED33DFULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_templar3_scaredmage.ani"},
	{0xEBD74FB28129E5A5ULL, "404818628.ogg"},
	{0xEBD7BFE756DD04D9ULL, "sky_lt_foundry_n.mao"},
	{0xEBD7E90F5F4242B1ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0cc_basic.msh"},
	{0xEBD7F70F5F425ACCULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0cc_basic.mao"},
	{0xEBD8030F5F426F6FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0cc_basic.mmh"},
	{0xEBD821D8C4165F5DULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_f_c.impa.gad"},
	{0xEBD8541303458E3FULL, "817164442.ogg"},
	{0xEBD87D32C519D741ULL, "306261645.ogg"},
	{0xEBD87E6D6854CB48ULL, "lgt210_eau_senestra.cnv"},
	{0xEBD88EEEA5BE0ED5ULL, "hlac_w_bonepit_0_8y.dds"},
	{0xEBD8BEBC899D8419ULL, "plt_lgt000pt_civ_bitch.nss"},
	{0xEBD8F6803ED2001CULL, "l_dae_hawke_tavern_name_trigger.bnk"},
	{0xEBD9041F81039FBEULL, "art\\characters\\creatures\\strider\\animation\\ad_dmg.evt"},
	{0xEBD92EF7110E4869ULL, "ave121_brennan.bnk"},
	{0xEBD95F47D6D40F20ULL, "vxht_0_hightown_a3_akt_0_0.vlm"},
	{0xEBD98C0F8E1B42E1ULL, "27691271.ogg"},
	{0xEBD99BA007A7C477ULL, "576162491.ogg"},
	{0xEBD9D94317EB926FULL, "vxlt_0_lowtown_a2_n_c_e24_0_0.vlm"},
	{0xEBDA2EC9F6FDEC8AULL, "791150540.ogg"},
	{0xEBDA57C134E6F130ULL, "ave121tr_approach_captain.utt"},
	{0xEBDAC130D9FF7A57ULL, "6082110_f.fxe"},
	{0xEBDAD7F14562838FULL, "vxht_0_hightown_n_ajz_0_0.vlm"},
	{0xEBDB4311AACA8167ULL, "mcr370_negotiation6080173.cl"},
	{0xEBDB50C70823FFE3ULL, "526150017.ogg"},
	{0xEBDBA06D68573AB0ULL, "lgt210_eau_senestra.bnk"},
	{0xEBDC10F711104AA9ULL, "ave121_brennan.cnv"},
	{0xEBDD05E2266194EBULL, "6138748_m.fxe"},
	{0xEBDD2C0A0F69C8B8ULL, "vxlt_undercityencounter_1jc_0_0.vlm"},
	{0xEBDD5CCBB2E51D76ULL, "110265069.ogg"},
	{0xEBDD5FAB0B82E3F2ULL, "one130ar_disciple_dungeon.lst"},
	{0xEBDDE2DDE2E0E74EULL, "vxlt_0_lowtown_a2_hed_0_0.vlm"},
	{0xEBDE20522C548D02ULL, "vxht_0_hightown_b09_0_0.vlm"},
	{0xEBDE57C8B945977DULL, "art\\vfx\\combat\\vfx_r_dlt_challenge_tag_c.cess.ani"},
	{0xEBDE80E63CF31CC4ULL, "dae000_bartrand6016756.cl"},
	{0xEBDF139EAC902177ULL, "gen_im_arm_boo_med_act1_03.uti"},
	{0xEBDF4AE6459B262CULL, "226949248.ogg"},
	{0xEBDF95E15F19A78AULL, "vxca_cavern1_haunted_k2r_0_1.vlm"},
	{0xEBDF9B7B7774C474ULL, "lowtown_fem_prostitute_sta6152332.cl"},
	{0xEBDFA5BC89A378F2ULL, "plt_lgt000pt_civ_bitch.plo"},
	{0xEBE01B1F8109E5E9ULL, "art\\characters\\creatures\\strider\\animation\\ad_dmg.gad"},
	{0xEBE0BC404A8057FFULL, "6208698_m.fxe"},
	{0xEBE1059685952AFCULL, "268280846.ogg"},
	{0xEBE124735FA40875ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd08a.phy"},
	{0xEBE12622CB5A2D1BULL, "vxlt_0_lowtown_f5c_0_0.vlm"},
	{0xEBE1562523FF258CULL, "mcr370_cut_player_rules6086395.cl"},
	{0xEBE1CE0217B3145BULL, "vxlt_undercityencounter_1p2_0_0.vlm"},
	{0xEBE1CF797428ACD6ULL, "986830213.ogg"},
	{0xEBE1EA1C27332373ULL, "6162978_m.fxe"},
	{0xEBE22BCDE1D135B4ULL, "vxht_keep_n_40r_0_0.vlm"},
	{0xEBE24F2D5DB4AE78ULL, "vxlt_warehouse_1df_0_0.vlm"},
	{0xEBE263BBB0A9C036ULL, "plt_isa000pt_gift_ship.plo"},
	{0xEBE282A58726F2DAULL, "vxdr_primevaltomb_8rd_0_0.vlm"},
	{0xEBE29CFDA284F791ULL, "70240097.ogg"},
	{0xEBE2F02B60A1C095ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.mmh"},
	{0xEBE324908D1713C0ULL, "qcr220cr_stroud.utc"},
	{0xEBE3864BCE03A112ULL, "gen00fl_merrill6150923.cl"},
	{0xEBE3CF397D7603B7ULL, "661944845.ogg"},
	{0xEBE3DFF2E265E4A7ULL, "llac_lt_0_lowtown_docks_n_-1_2.dds"},
	{0xEBE4058C1A19B023ULL, "vxht_chantry_35b_0_0.vlm"},
	{0xEBE4AFDAA6087E68ULL, "ser221ip_chest_1.utp"},
	{0xEBE4B202128A4207ULL, "vxht_mansion_n_53m_0_0.vlm"},
	{0xEBE52B629F4699B5ULL, "vxlt_0_lowtown_docks_a3_n_as1_0_0.vlm"},
	{0xEBE57C3D433FB628ULL, "vxht_0_hightown_ap8_0_0.vlm"},
	{0xEBE5AC9158E600CCULL, "blood_conj_02_l3.mao"},
	{0xEBE5FD23CE87614DULL, "hlac_w_encounter3_-131073_0x.dds"},
	{0xEBE610430736561BULL, "hlac_lt_0_lowtown_docks_n_-131074_3z.dds"},
	{0xEBE6423393751858ULL, "hlac_lt_gallowstemplar_7198_5z.dds"},
	{0xEBE6C24A35A91DEBULL, "gen00fl_anders6104004.cl"},
	{0xEBE6CBB9DCF74DD5ULL, "vxht_brothel_n_6b4_0_0.vlm"},
	{0xEBE6EC1F810FB437ULL, "art\\characters\\creatures\\strider\\animation\\ad_dmg.ani"},
	{0xEBE71FBD436A42B5ULL, "vxht_bartrand_n_5rk_0_0.vlm"},
	{0xEBE730FBA4C360FCULL, "539034028.ogg"},
	{0xEBE74255D026C623ULL, "lt_gallowscourtyard_53_0.rml"},
	{0xEBE77D009E8CBC90ULL, "6071490_f.fxe"},
	{0xEBE7CFB0731BA020ULL, "6082095_m.fxe"},
	{0xEBE7D124C8CA27A0ULL, "vxht_0_hightown_a3_apu_0_0.vlm"},
	{0xEBE7DF871571D449ULL, "vxht_0_hightown_9wx_0_0.vlm"},
	{0xEBE7F0CDB35F4E60ULL, "636502978.ogg"},
	{0xEBE83ABC85C646A4ULL, "isa211_isabela6069442.cl"},
	{0xEBE8B7BD74A7CF58ULL, "dae000_hate_enemies6167572.cl"},
	{0xEBE8F8E7158EF107ULL, "6037012_m.fxe"},
	{0xEBE91375424A8B75ULL, "6174926_m.fxe"},
	{0xEBE91D68E344D942ULL, "follower_banter6167199.cl"},
	{0xEBE970572D105E58ULL, "hlac_lt_gallowstemplar_n_2388_3y.dds"},
	{0xEBE974955A4023BFULL, "vxlt_blackemporium_is_0_0.vlm"},
	{0xEBE98316158B6498ULL, "vxw_woundedcoast_1qs_0_0.vlm"},
	{0xEBE9952BA13B8556ULL, "w_bonepit_94_0.mao"},
	{0xEBE9A12BA13B99F5ULL, "w_bonepit_94_0.mmh"},
	{0xEBE9A32BA13B9D6BULL, "w_bonepit_94_0.msh"},
	{0xEBE9DA45AD066775ULL, "6034231_m.fxe"},
	{0xEBEAA074513ACA68ULL, "6115359_m.fxe"},
	{0xEBEB637AD9875117ULL, "258497200.ogg"},
	{0xEBEB89E2780F0593ULL, "dae000_keep_noble_sta6161517.cl"},
	{0xEBEBE97D744276A3ULL, "hlac_lt_darktown_-1_89x.dds"},
	{0xEBEC54BAD4D6FFCEULL, "vxlt_0_lowtown_a2_hbe_0_0.vlm"},
	{0xEBEC7F60E8E3FAD9ULL, "vrc100pt_n_initial.ncs"},
	{0xEBECD4CEAED73EEAULL, "ringout.mao"},
	{0xEBECEA8BCC539D6CULL, "486800024.ogg"},
	{0xEBED268452FA22ACULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_mark_activate.ani"},
	{0xEBED38CBA5EE4970ULL, "6169666_m.fxe"},
	{0xEBED58DE88305A8CULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.cess.ani"},
	{0xEBEDBBED9AF47723ULL, "6150338_f.fxe"},
	{0xEBEE403AC647125BULL, "hlac_lt_0_lowtown_a2_n_c_131073_0z.dds"},
	{0xEBEE92829D22C154ULL, "783334877.ogg"},
	{0xEBEEFCF2644521C1ULL, "ht_0_hightown_a2_34_0.anb"},
	{0xEBEFF67A8DDD3431ULL, "blood_b.dds"},
	{0xEBF00E493758C78BULL, "vxht_mansion_n_636_0_0.vlm"},
	{0xEBF01D692AC61D60ULL, "630686340.ogg"},
	{0xEBF050AE1C5F2FB4ULL, "280578846.ogg"},
	{0xEBF0AD21A7C0933AULL, "qun131_taskmaster6019207.cl"},
	{0xEBF0ED2B60ADF922ULL, "art\\vfx\\combat\\vfx_m_spi_crh_pris_p.anb"},
	{0xEBF126AB450FD57CULL, "motivations.gda"},
	{0xEBF14E89DDAD6B05ULL, "zz_followers_banter6158226.cl"},
	{0xEBF153EBDE414148ULL, "6018716_f.fxe"},
	{0xEBF166E70ABE1715ULL, "llac_lt_0_lowtown_docks_a3_-196609_0.dds"},
	{0xEBF17B61931B749CULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1webdetail.mao"},
	{0xEBF18761931B893FULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1webdetail.mmh"},
	{0xEBF18D61931B93E1ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1webdetail.msh"},
	{0xEBF1B5536559F05FULL, "6171435_m.fxe"},
	{0xEBF1CAFCE6E248D0ULL, "w_bonepitmining_a2_193_0.mao"},
	{0xEBF1CEFCE6E24FE3ULL, "w_bonepitmining_a2_193_0.mmh"},
	{0xEBF1DCFCE6E2673DULL, "w_bonepitmining_a2_193_0.msh"},
	{0xEBF1E0F4E7654399ULL, "271943768.ogg"},
	{0xEBF280C2A8BE5032ULL, "follower_banter6164688.cl"},
	{0xEBF29928FC0F0782ULL, "one220_yevhen6098937.cl"},
	{0xEBF32103A1FC9B5FULL, "dre000_hireling16038612.cl"},
	{0xEBF3490BB928CF07ULL, "903811436.ogg"},
	{0xEBF366C51A59C0EEULL, "mag121_idunna6092354.cl"},
	{0xEBF38DA2563E0410ULL, "vxw_encounter3_d7_0_0.vlm"},
	{0xEBF3AD133B5E322BULL, "dm_gen_5.mor"},
	{0xEBF40AFE4216B389ULL, "103388547.ogg"},
	{0xEBF43384530057EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_mark_activate.gad"},
	{0xEBF465DE88368FCEULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.cess.gad"},
	{0xEBF503E8CAD0D7DAULL, "gen_im_npc_wep_shd_krkb.uti"},
	{0xEBF57C90B9466B80ULL, "gen00fl_isabela6118385.cl"},
	{0xEBF5919A35605B35ULL, "vxht_bartrand_n_5uj_0_0.vlm"},
	{0xEBF5BBF5F6DF82A6ULL, "427827390.ogg"},
	{0xEBF5BE7AD1075C25ULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.anb"},
	{0xEBF6453E7EFA85DAULL, "506134063.ogg"},
	{0xEBF64E82B49BE870ULL, "6141945_m.fxe"},
	{0xEBF6B2423A03E7D2ULL, "llac_lt_0_lowtown_docks_a3_-1_1.dds"},
	{0xEBF6D6BCCE079989ULL, "6175910_m.fxe"},
	{0xEBF73D1FA328427FULL, "vxht_0_hightown_n_ak9_0_0.vlm"},
	{0xEBF84C229EFC7FF5ULL, "6031131_m.fxe"},
	{0xEBF850E82A53B44CULL, "vxht_0_hightown_n_c45_0_0.vlm"},
	{0xEBF855FECDEA02D5ULL, "vxw_blightlands_99_0_0.vlm"},
	{0xEBF861F3F8F50AC3ULL, "vrc321_varric6084910.cl"},
	{0xEBF8AC9F7A56F2C5ULL, "vxlt_gallowscourtyard_4xm_0_0.vlm"},
	{0xEBF9E92F96188482ULL, "6180434_m.fxe"},
	{0xEBF9FC8F2E7250FBULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_statue_c.impa.ani"},
	{0xEBFA0073C0E2E938ULL, "875430047.ogg"},
	{0xEBFAAC8995E0F001ULL, "hlac_lt_backalley_g_0_7x.dds"},
	{0xEBFAE69BCDDA2239ULL, "w_blightlands_119_0.phy"},
	{0xEBFAFEE15DC64D1FULL, "110959372.ogg"},
	{0xEBFB2E8453066DCDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_mark_activate.evt"},
	{0xEBFB30BEE25D254AULL, "199978823.ogg"},
	{0xEBFB60DE883CA5ADULL, "art\\vfx\\combat\\vfx_m_pri_f_proj_p.cess.evt"},
	{0xEBFBEA216DD0E83AULL, "6059783_m.fxe"},
	{0xEBFC68A30B00BFB2ULL, "363977966.ogg"},
	{0xEBFC6B7225A2A141ULL, "llac_lt_warehouse_1_1.dds"},
	{0xEBFC798530B5C5CDULL, "gen00fl_dog6117429.cl"},
	{0xEBFC93D3EA148CCAULL, "160097988.ogg"},
	{0xEBFC9E153A404A8BULL, "vxht_keep_n_6yd_0_0.vlm"},
	{0xEBFCBB8963C1D864ULL, "and100_anders6095819.cl"},
	{0xEBFD2FAAE4A2B494ULL, "899832832.ogg"},
	{0xEBFD69AF6B1860E4ULL, "llac_ht_keep_n_4519_9.dds"},
	{0xEBFE23F002DD4D57ULL, "w_bonepitmining_a2_32_0.mmh"},
	{0xEBFE27F002DD5464ULL, "w_bonepitmining_a2_32_0.mao"},
	{0xEBFE2996DF7E5686ULL, "gen_im_arm_hlm_elf_comc.uti"},
	{0xEBFE39F002DD72C9ULL, "w_bonepitmining_a2_32_0.msh"},
	{0xEBFFA0290955BBE9ULL, "270029810.ogg"},
	{0xEC0046C13507A3EDULL, "ave121tr_approach_captain.ncs"},
	{0xEC007F658AF74E93ULL, "sky_lt_0_lowtown_a2_n_c.mao"},
	{0xEC0081D10896D674ULL, "mcr370_templar26087641.cl"},
	{0xEC010CFD17FF0F1DULL, "746083809.ogg"},
	{0xEC014DB7A00E3134ULL, "vxw_woundedcoast_od_0_0.vlm"},
	{0xEC0195BBF6767307ULL, "vxht_mansion_n_59o_0_0.vlm"},
	{0xEC01999061FC2202ULL, "vxdr_primevaltomb_3da_0_0.vlm"},
	{0xEC01FF7F48341C90ULL, "vxw_encounter3_a4_0_0.vlm"},
	{0xEC0261CD29884C87ULL, "hlac_dr_dungeon_3217_7y.dds"},
	{0xEC027E8DC7DFCC9DULL, "llac_lt_0_lowtown_n_1_8.dds"},
	{0xEC0284B229331C92ULL, "6035859_m.fxe"},
	{0xEC02AEF37031417CULL, "zz_table_lean_test.cut"},
	{0xEC030F0F2A63FDE9ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_basic.phy"},
	{0xEC0349407CE23021ULL, "930412002.ogg"},
	{0xEC0354032D1946CEULL, "hlac_lt_0_lowtown_docks_a3_n_65534_6y.dds"},
	{0xEC035C85B9041164ULL, "900832938.ogg"},
	{0xEC03AC68D1F1AF88ULL, "gen00fl_merrill6121756.cl"},
	{0xEC03AF73C0E37ED5ULL, "vxht_brothel_n_6h2_0_0.vlm"},
	{0xEC03E1104C6BF9E5ULL, "6141801_m.fxe"},
	{0xEC03F02569B5AA4FULL, "vxlt_undercityencounter_uz_0_0.vlm"},
	{0xEC040377275673B5ULL, "vxht_bartrand_n_5xi_0_0.vlm"},
	{0xEC0567A32E71DDCAULL, "pro000ip_barrier.utp"},
	{0xEC05719D0D8744E1ULL, "dae000_fenarel6111549.cl"},
	{0xEC057D04F2EA9FD4ULL, "6031259_m.fxe"},
	{0xEC05ADFD3623A7DEULL, "vxca_caverns1_fo8_0_0.vlm"},
	{0xEC060C4F9BA9D88DULL, "16399557.ogg"},
	{0xEC06F6420F9A7BECULL, "vxlt_blackemporium_vq_0_0.vlm"},
	{0xEC0772A46601113BULL, "vxlt_0_lowtown_docks_n_ars_0_0.vlm"},
	{0xEC08048F2E7E9BF2ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_statue_c.impa.evt"},
	{0xEC08C6DF7D6C94C4ULL, "6125947_m.fxe"},
	{0xEC09138135A49CC3ULL, "vxht_0_hightown_a3_7lx_0_0.vlm"},
	{0xEC093154FE8E3347ULL, "vxlt_gallowstemplar_f_4y9_0_0.vlm"},
	{0xEC0941EA9AE13225ULL, "one350_vinell6087492.cl"},
	{0xEC098E5EA8D99E73ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_hurt_aveline.evt"},
	{0xEC0998766D968135ULL, "446060823.ogg"},
	{0xEC09B133174DBA70ULL, "6016924_m.fxe"},
	{0xEC09BA4A3775AC99ULL, "journal.gfx"},
	{0xEC09BEFF8745BE28ULL, "and111_lirene6084503.cl"},
	{0xEC0A1703437CDB85ULL, "m_dae340ar_bone_pit.bnk"},
	{0xEC0A22F73E2C838FULL, "vxht_0_hightown_a2_b66_0_0.vlm"},
	{0xEC0A2B06AC0EE9A6ULL, "mcr310_argument6074461.cl"},
	{0xEC0A3EC822DE9ADAULL, "253448655.ogg"},
	{0xEC0AB4B3F5D4D3C7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_inst.gad"},
	{0xEC0C92CC55F249B0ULL, "vxht_keep_n_60d_0_0.vlm"},
	{0xEC0D0091D0106E97ULL, "142466031.ogg"},
	{0xEC0D8D8E06EA226EULL, "zz_lgt_1_debug6114219.cl"},
	{0xEC0E238F2E833B85ULL, "art\\vfx\\cinematic\\prerendered\\vfx_chantry_statue_c.impa.gad"},
	{0xEC0E6214CC625555ULL, "isa321ip_brothel_door.ncs"},
	{0xEC0E78CFEADA4EA8ULL, "dae100_bran6118171.cl"},
	{0xEC0F35895A024481ULL, "vxlt_gallowstemplar_f_3gb_0_0.vlm"},
	{0xEC0F822511565307ULL, "gen00fl_carver6102558.cl"},
	{0xEC0F8A0F6B087F52ULL, "258538924.ogg"},
	{0xEC0FBF94920449B4ULL, "vxw_woundedcoast_de_0_0.vlm"},
	{0xEC0FD97D783008BEULL, "vxlt_darktown_50q_0_0.vlm"},
	{0xEC1017893051213BULL, "gen00fl_bethany6028797.cl"},
	{0xEC103BCCE29EF699ULL, "vxca_caverns1_mines_e7q_0_0.vlm"},
	{0xEC1055D7F0E3898BULL, "583464951.ogg"},
	{0xEC10875EA8DFB0B4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_hurt_aveline.gad"},
	{0xEC110D0C45297665ULL, "vxlt_0_lowtown_docks_b5l_0_0.vlm"},
	{0xEC115FF182B50ACEULL, "def_rock2_n.dds"},
	{0xEC11A9B3F5DADFD4ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_inst.evt"},
	{0xEC11EFA3CDAC3D67ULL, "6072165_m.fxe"},
	{0xEC1207DD4E6BC126ULL, "art\\characters\\weapons\\promo\\longswords\\w_lsw_025a.phy"},
	{0xEC12127C33081836ULL, "art\\characters\\playercharacter\\elfmale\\em_har_ha1a.mmh"},
	{0xEC121A273760CBB2ULL, "follower_banter6171945.cl"},
	{0xEC121C7C33082934ULL, "art\\characters\\playercharacter\\elfmale\\em_har_ha1a.msh"},
	{0xEC122343C30B68B0ULL, "follower_banter6151299.cl"},
	{0xEC122AF7113E061AULL, "ave121_brennan.stg"},
	{0xEC1254DFDFCA2279ULL, "vxca_caverns2prc_8h2_0_0.vlm"},
	{0xEC12CBB56E730FBBULL, "art\\vfx\\combat\\vfx_g_spider_fb_act_c.impa.ani"},
	{0xEC1307D5EE2A43D4ULL, "hlac_lt_0_lowtown_n_65535_4z.dds"},
	{0xEC132A7CC2F2AE1EULL, "6181926_f.fxe"},
	{0xEC13393B2010832CULL, "dre000_bodahn6031273.cl"},
	{0xEC1351CA756132C5ULL, "vxdr_0_deeproads_dw2_0_0.vlm"},
	{0xEC1351DCC4F7C5ECULL, "gen00fl_aveline6112737.cl"},
	{0xEC13C80EBE1EAF43ULL, "zz_approval_debug6133255.cl"},
	{0xEC13D4F4A68EB64EULL, "vxdr_primevalentrance_2q7_0_0.vlm"},
	{0xEC143982DFC56DFCULL, "hlac_lt_blackemporium_1055_17z.dds"},
	{0xEC143FD1034E93EBULL, "261887850.ogg"},
	{0xEC146AD7953FC946ULL, "589557997.ogg"},
	{0xEC14AE68267FD78AULL, "vxca_caverns1_mines_j40_0_0.vlm"},
	{0xEC1536E586C6D18FULL, "art\\characters\\creatures\\spider\\animation\\dazed.gad"},
	{0xEC15AB9A9BB18467ULL, "6129254_m.fxe"},
	{0xEC16657529714802ULL, "25856596.ogg"},
	{0xEC167BE1F50BF70EULL, "341601320.ogg"},
	{0xEC167BF2E7E1D263ULL, "6121932_f.fxe"},
	{0xEC16965EA8E43592ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\wesley_hurt_aveline.ani"},
	{0xEC176A77838BAB91ULL, "593946070.ogg"},
	{0xEC17C4C533A1D0DDULL, "6110748_m.fxe"},
	{0xEC17D4A33BB804BCULL, "hlac_lt_gallowstemplar_f_2388_6z.dds"},
	{0xEC180294FDB1B738ULL, "289214376.ogg"},
	{0xEC18B41692EC10AFULL, "142685096.ogg"},
	{0xEC18E42BB5BB5E4EULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_varric_.ani"},
	{0xEC1981454E3515EEULL, "vxht_0_hightown_n_buq_0_0.vlm"},
	{0xEC1A000385B69B64ULL, "vxht_bartrand_n_67n_0_0.vlm"},
	{0xEC1A3DDD4617DD2FULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.dura.gad"},
	{0xEC1A5EFCD49D206EULL, "vxlt_gallowstemplar_n_45p_0_0.vlm"},
	{0xEC1A8651AA0C7DB2ULL, "one110_meeran6089482.cl"},
	{0xEC1A8A8AB5E06A36ULL, "art\\vfx\\creatures\\fxc_ft_dirt_r_c.anb"},
	{0xEC1ACE31F92CC5BCULL, "835591294.ogg"},
	{0xEC1AD69504331E42ULL, "6094872_m.fxe"},
	{0xEC1AD8F2370643A9ULL, "105349623.ogg"},
	{0xEC1B407DC55DDFB8ULL, "follower_banter6164880.cl"},
	{0xEC1B4BE586CB613CULL, "art\\characters\\creatures\\spider\\animation\\dazed.evt"},
	{0xEC1B6E5C9FBBE520ULL, "6039241_f.fxe"},
	{0xEC1C043331B95973ULL, "6013844_f.fxe"},
	{0xEC1C8EA16EBE37BEULL, "vxw_woundedcoast_1f4_0_0.vlm"},
	{0xEC1D6A6A2B503E3AULL, "vxlt_hangedman_5j_0_0.vlm"},
	{0xEC1D8E7428D78E8DULL, "105752029.ogg"},
	{0xEC1E308419396C75ULL, "hlac_ht_chantry_n_2534_47z.dds"},
	{0xEC1E317183FA6234ULL, "vxw_woundedcoast_aj_0_0.vlm"},
	{0xEC1EA1B3F5E55C0DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_inst.ani"},
	{0xEC1EB39D677D98BFULL, "qun131_taskmaster6059042.cl"},
	{0xEC1EBF7AD12A6366ULL, "art\\vfx\\environment\\vfx_trap_pois_gas_p.mmh"},
	{0xEC1F09D715C2526CULL, "hlac_ca_cave_2194_3z.dds"},
	{0xEC1F0E8C9A4E122BULL, "vxlt_0_lowtown_a2_n_c_gdh_0_0.vlm"},
	{0xEC1F3334268D7081ULL, "307035063.ogg"},
	{0xEC1FC0AE8B76756CULL, "hlac_ca_caverns1_24423_4x.dds"},
	{0xEC1FD089E86778DBULL, "dae000ip_crowd_em1_talk12.utp"},
	{0xEC1FD3B56E7DA7B2ULL, "art\\vfx\\combat\\vfx_g_spider_fb_act_c.impa.evt"},
	{0xEC1FDC4D1751740BULL, "vxht_0_hightown_n_b8k_0_0.vlm"},
	{0xEC20AE03A5413A29ULL, "vxlt_0_lowtown_a2_n_hdc_0_0.vlm"},
	{0xEC2152DD461E1FDCULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.dura.evt"},
	{0xEC21B8FD0017138EULL, "vrc321_spooky_scenes6219438.cl"},
	{0xEC21FA949902A739ULL, "art\\characters\\playercharacter\\humanmale\\rob_appp.msh"},
	{0xEC21FFBA3674E986ULL, "lt_0_lowtown_n_10_0.rml"},
	{0xEC2204949902B827ULL, "art\\characters\\playercharacter\\humanmale\\rob_appp.mmh"},
	{0xEC2208949902BF74ULL, "art\\characters\\playercharacter\\humanmale\\rob_appp.mao"},
	{0xEC2277EA5074F90DULL, "w_encounter2_camp_31_0.phy"},
	{0xEC2296E57C667FCEULL, "6049715_m.fxe"},
	{0xEC229D436D578E4AULL, "538814982.ogg"},
	{0xEC2320451875F00AULL, "vxca_caverns1_mines_j31_0_0.vlm"},
	{0xEC23281A5D890F77ULL, "73655133.ogg"},
	{0xEC23314961277358ULL, "vxca_caverns1_mines_e81_0_0.vlm"},
	{0xEC2437AC8B77C4DEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightb_n.dds"},
	{0xEC24FDF803C2EDF3ULL, "53248076.ogg"},
	{0xEC2526EF7EAD0479ULL, "6125959_m.fxe"},
	{0xEC256D9029F1E767ULL, "hlac_lt_backalley_n_0_3y.dds"},
	{0xEC257B85D56F8C2AULL, "vxw_bonepitmining_a2_27e_0_0.vlm"},
	{0xEC25ED2F5170F382ULL, "art\\vfx\\cinematic\\vfx_meredith_sword_c.mmh"},
	{0xEC2692ED7A60A55CULL, "vxlt_gallowstemplar_f_3s7_0_0.vlm"},
	{0xEC26F2B56E83FB45ULL, "art\\vfx\\combat\\vfx_g_spider_fb_act_c.impa.gad"},
	{0xEC2734553C571B71ULL, "vxht_0_hightown_c3j_0_0.vlm"},
	{0xEC278D8AB5EAF9A1ULL, "art\\vfx\\creatures\\fxc_ft_dirt_r_c.mmh"},
	{0xEC27F322402B2E6EULL, "vxht_0_hightown_n_bpr_0_0.vlm"},
	{0xEC2837735FE02A90ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd08a.msh"},
	{0xEC2845735FE042AAULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd08a.mmh"},
	{0xEC2854BF229117F0ULL, "qun231_varnell.stg"},
	{0xEC285FC1BD89D1BCULL, "w_bonepitdestroyed_a3_34_0.rml"},
	{0xEC2867E15527C4E9ULL, "6166345_m.fxe"},
	{0xEC28DDF264766551ULL, "ht_0_hightown_a2_34_0.rml"},
	{0xEC2917AA49CD98F3ULL, "805772327.ogg"},
	{0xEC2924B8B4A1CA92ULL, "vxw_encounter2_37_0_0.vlm"},
	{0xEC2943E715397547ULL, "6109209_m.fxe"},
	{0xEC2953E586D7AC55ULL, "art\\characters\\creatures\\spider\\animation\\dazed.ani"},
	{0xEC29670210E384BDULL, "seb220ip_to_harimann_estate.utp"},
	{0xEC2978EA507B19B9ULL, "w_encounter2_camp_31_0.rml"},
	{0xEC29A1C29D7B94FEULL, "6136329_m.fxe"},
	{0xEC29B5C8C9DA17D4ULL, "vxlt_0_lowtown_hb3_0_0.vlm"},
	{0xEC2A09EA45B1F5EFULL, "hlac_lt_0_lowtown_n_65538_2x.dds"},
	{0xEC2A188C6DEA202EULL, "vxlt_0_lowtown_docks_a3_c1q_0_0.vlm"},
	{0xEC2A4DC9E6E2E54CULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.dura.ani"},
	{0xEC2A862BB4B3AAE2ULL, "vxlt_warehouse_ab_0_0.vlm"},
	{0xEC2A92608BD1451DULL, "ave331_jeven6137273.cl"},
	{0xEC2AEFC27B97F27FULL, "827534437.ogg"},
	{0xEC2B007E60B4503EULL, "vxw_woundedcoast_1a5_0_0.vlm"},
	{0xEC2B0E223A7D49DEULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_dustmots_lrg.mmh"},
	{0xEC2B45C1BD8BDB11ULL, "w_bonepitdestroyed_a3_34_0.msh"},
	{0xEC2B53C1BD8BF32CULL, "w_bonepitdestroyed_a3_34_0.mao"},
	{0xEC2B540EE95E743AULL, "471280909.ogg"},
	{0xEC2B589728ED1030ULL, "vxht_mansion_n_6k1_0_0.vlm"},
	{0xEC2B5FC1BD8C07CFULL, "w_bonepitdestroyed_a3_34_0.mmh"},
	{0xEC2BEA19D3BDA0E7ULL, "hlac_w_bonepitdestroyed_a3_2_0y.dds"},
	{0xEC2C43A10FA16478ULL, "hlac_ht_brothel_n_3369_2x.dds"},
	{0xEC2C9D93C4EAEA5FULL, "hlac_ht_chantry_2534_17x.dds"},
	{0xEC2CE235852F2B08ULL, "vxlt_gallowstemplar_f_46c_0_0.vlm"},
	{0xEC2D109DEEB03CBFULL, "348944641.ogg"},
	{0xEC2D69E4163EE10FULL, "nat380_nathaniel6071037.cl"},
	{0xEC2DBD2BB5CD7788ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot2_varric_.gad"},
	{0xEC2DC6731A06AE47ULL, "vxlt_0_lowtown_docks_a3_n_d4s_0_0.vlm"},
	{0xEC2E34ECAFC33D7CULL, "543438768.ogg"},
	{0xEC2E5C9DE7CC5AC0ULL, "hlac_lt_gallowstemplar_2394_10y.dds"},
	{0xEC2EADF2B26C3E49ULL, "vxht_keep_n_731_0_0.vlm"},
	{0xEC2EAECD628B77D3ULL, "6018594_m.fxe"},
	{0xEC2EF80F2A89110EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_basic.mmh"},
	{0xEC2F040F2A8925A5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_basic.mao"},
	{0xEC2F0BD94D5A1F3CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medd_0s.dds"},
	{0xEC2F120F2A893DFCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wo_basic.msh"},
	{0xEC2F5ADD462A6AF5ULL, "art\\vfx\\combat\\vfx_w_rev_haste_loop_c.dura.ani"},
	{0xEC3019B94DBB84B3ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_3z.dds"},
	{0xEC30772BC06CD8CFULL, "dae150ip_barrier_pride.ncs"},
	{0xEC30C9C68E45DAF0ULL, "230255205.ogg"},
	{0xEC3129B6B399FCE9ULL, "vxw_encounter3_f4_0_0.vlm"},
	{0xEC315AC9E6E9198EULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.dura.gad"},
	{0xEC319AE3375EC4A7ULL, "gen00fl_merrill6147021.cl"},
	{0xEC3276EC88786901ULL, "gen_im_trs_bst_kidney_01.uti"},
	{0xEC328D78D9FECE31ULL, "45391948.ogg"},
	{0xEC32995817E96D6EULL, "mag310cr_controlled_1.utc"},
	{0xEC32AEFF155BBC61ULL, "one280_iwan6096749.cl"},
	{0xEC331B35B1C54A95ULL, "vxca_cavern1_haunted_j6a_0_0.vlm"},
	{0xEC33779043F1C83BULL, "6076904_f.fxe"},
	{0xEC3394CD63952399ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_9.pwk"},
	{0xEC34FA028DE85AFFULL, "vxht_0_hightown_a2_b8t_0_0.vlm"},
	{0xEC35A6322E4D33F1ULL, "vxht_0_hightown_c6i_0_0.vlm"},
	{0xEC36319EEF2EC461ULL, "6169803_m.fxe"},
	{0xEC365697FE978AE9ULL, "459051504.ogg"},
	{0xEC3665890218AC8BULL, "vxht_keep_n_6ex_0_0.vlm"},
	{0xEC36E4F5DE8519FBULL, "6050099_m.fxe"},
	{0xEC36FC1E951031E7ULL, "6085422_m.fxe"},
	{0xEC371AF7184A2BF2ULL, "vxht_keep_n_5ty_0_0.vlm"},
	{0xEC3730D1FAA7584FULL, "688319482.ogg"},
	{0xEC3755C9E6ED7D6DULL, "art\\vfx\\environment\\vfx_trap_acid_proj_p.dura.evt"},
	{0xEC3772B134D930BDULL, "vxht_0_hightown_a2_bpv_0_0.vlm"},
	{0xEC37D6C04F9BBB59ULL, "hlac_lt_0_lowtown_docks_n_-196610_6z.dds"},
	{0xEC38139F8F71AD7EULL, "6131229_m.fxe"},
	{0xEC3901ABF7FCC70BULL, "art\\levels\\races\\primeval\\props\\_texture\\prp_chestb_sa.dds"},
	{0xEC3905CFEDE9568AULL, "675917313.ogg"},
	{0xEC39503F30140D74ULL, "6076622_f.fxe"},
	{0xEC39725B52AA68BEULL, "vxw_woundedcoast_1l6_0_0.vlm"},
	{0xEC39735B81FEA714ULL, "zz_vrc_debug6200170.cl"},
	{0xEC39C5CFF1FFEA30ULL, "608715773.ogg"},
	{0xEC3A13D746D09467ULL, "vxht_mansion_n_6o6_0_0.vlm"},
	{0xEC3A4201A6147013ULL, "llac_dr_slavers_3214_14.dds"},
	{0xEC3A5A5C25F90393ULL, "vxlt_undercityencounter_v7_0_0.vlm"},
	{0xEC3A61D9BCFDE1CAULL, "gen00fl_isabela_init6077728.cl"},
	{0xEC3AEA6419571C44ULL, "vxw_encounter1_d1_0_0.vlm"},
	{0xEC3B03A974C27234ULL, "mrl150_merrill6071951.cl"},
	{0xEC3B08F40D61CD24ULL, "51296126.ogg"},
	{0xEC3B0E4D323BDB6FULL, "w_woundedcoast_6_0_lc.mmh"},
	{0xEC3B3F2C5DC167C0ULL, "mcr310_sibling6076312.cl"},
	{0xEC3B492A7959DD3BULL, "vxlt_0_lowtown_a2_n_eqo_0_0.vlm"},
	{0xEC3B5DD11FC05AF3ULL, "llac_ht_mansion2_818_58.dds"},
	{0xEC3B678BD2F676B2ULL, "vxht_mansion2_1rx_0_0.vlm"},
	{0xEC3BEFED951A7343ULL, "vxlt_gallowstemplar_f_451_0_0.vlm"},
	{0xEC3C27A1BFB4196AULL, "vxlt_gallowstemplar_n_6st_0_0.vlm"},
	{0xEC3C2B5EB7884A81ULL, "zz_followers_banter6159342.cl"},
	{0xEC3C5835245A09DCULL, "vxlt_gallowstemplar_n_5nk_0_1.vlm"},
	{0xEC3D02F414DCEC2FULL, "vxlt_undercityserialkiller_yj_0_0.vlm"},
	{0xEC3D1FE92FEFE82FULL, "llac_ca_cavern1_haunted_24422_7.dds"},
	{0xEC3E136365A62BD9ULL, "hlac_w_bonepitdestroyed_a3_0_10z.dds"},
	{0xEC3E32573C8898B4ULL, "vxw_bonepit_1wb_0_0.vlm"},
	{0xEC3E6D2B3EA7C9CDULL, "6207745_m.fxe"},
	{0xEC3ED5006DB0183AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobd_0s.dds"},
	{0xEC3F3A1150E58D07ULL, "vxw_encounter2_camp_ru_0_0.vlm"},
	{0xEC3F956EE735206EULL, "vxht_keep_6iz_0_0.vlm"},
	{0xEC3FCC5138385BF2ULL, "hlac_lt_0_lowtown_a2_n_c_65533_0z.dds"},
	{0xEC40869437E77763ULL, "828184770.ogg"},
	{0xEC418D9BD4D20190ULL, "hlac_lt_0_lowtown_docks_-196610_6z.dds"},
	{0xEC424453D4AC6D13ULL, "993968559.ogg"},
	{0xEC425F3FB95BBD2AULL, "vxw_bonepitmining_a2_29g_0_0.vlm"},
	{0xEC427323C2BAE4D1ULL, "6173761_m.fxe"},
	{0xEC42E7294FF24730ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivylineglw.msh"},
	{0xEC42F5294FF25F4AULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivylineglw.mmh"},
	{0xEC42F9294FF26581ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_ivylineglw.mao"},
	{0xEC437FA4B82227C8ULL, "ht_0_hightown_a2_25_0_lc.mmh"},
	{0xEC43855E7B0EB382ULL, "vxlt_0_lowtown_a2_h1t_0_0.vlm"},
	{0xEC43ABB26ABFD77BULL, "vxht_chantry_26n_0_0.vlm"},
	{0xEC43FC239B77C55FULL, "vxlt_gallowstemplar_4un_0_0.vlm"},
	{0xEC4416AC79E5AB9DULL, "422436484.ogg"},
	{0xEC444E6A7FFB08EFULL, "w_encounter3_34_0_ll.mmh"},
	{0xEC44C61FBF3933D7ULL, "hlac_lt_0_lowtown_a2_1_4y.dds"},
	{0xEC44C756ECDAAD17ULL, "lt_0_lowtown_docks_n_26_0_lc.mmh"},
	{0xEC44D6DC24175F6EULL, "vxht_0_hightown_n_bft_0_0.vlm"},
	{0xEC450006C52D8287ULL, "vxlt_0_lowtown_docks_a3_d2s_0_1.vlm"},
	{0xEC456A34EFEA575AULL, "vxlt_0_lowtown_deh_0_0.vlm"},
	{0xEC459EA24A361E6EULL, "gen00fl_aveline6140035.cl"},
	{0xEC45F592DFF9783BULL, "151999243.ogg"},
	{0xEC460213951002DDULL, "6190968_m.fxe"},
	{0xEC460872988DFB92ULL, "vxw_encounter2_91_0_0.vlm"},
	{0xEC460C8FE31EAA44ULL, "6216988_m.fxe"},
	{0xEC46798FAD7A3526ULL, "vxlt_gallowstemplar_f_519_0_0.vlm"},
	{0xEC46888ADCEE8534ULL, "6043536_f.fxe"},
	{0xEC46900CD5AA8268ULL, "gen_im_npc_wep_one_veld.uti"},
	{0xEC46DAA22E037C81ULL, "vxlt_gallowstemplar_n_37j_0_0.vlm"},
	{0xEC474B69DAD4D37CULL, "488578772.ogg"},
	{0xEC47659095A01E42ULL, "dae000_party_night_comment6201621.cl"},
	{0xEC4837BB34950407ULL, "vxca_caverns2prc_8bb_0_0.vlm"},
	{0xEC48540FC957FA76ULL, "308839038.ogg"},
	{0xEC485E63FEA02A6EULL, "894044409.ogg"},
	{0xEC4874CDAE71585EULL, "301641117.ogg"},
	{0xEC48EAD53F6D9B21ULL, "6118568_m.fxe"},
	{0xEC48EC32F7513EBFULL, "33912700.ogg"},
	{0xEC48FCFD21DAA681ULL, "vxlt_gallowstemplar_f_3sf_0_0.vlm"},
	{0xEC4912EC30BA13F1ULL, "mcr370_meredith6136129.cl"},
	{0xEC495C410B4D34C4ULL, "vxw_encounter1_i6_0_0.vlm"},
	{0xEC497F5E5EEABF9AULL, "651874001.ogg"},
	{0xEC49870859DCABB4ULL, "vxw_woundedcoast_pi_0_0.vlm"},
	{0xEC49BF7EADF0965FULL, "vxht_0_hightown_cab_0_0.vlm"},
	{0xEC49D968C4EC8F32ULL, "vxht_mansion2_1uy_0_0.vlm"},
	{0xEC49EBB7C5585D4CULL, "gen00fl_aveline6129043.cl"},
	{0xEC4A10EE74984044ULL, "6105923_m.fxe"},
	{0xEC4AA397B72B7523ULL, "art\\vfx\\combat\\vfx_str_att_back_l_leg_c.anb"},
	{0xEC4C407D734CFA7DULL, "680847257.ogg"},
	{0xEC4CA4342E7EB134ULL, "vxw_bonepit_1xc_0_0.vlm"},
	{0xEC4D762263C8D3A8ULL, "346242610.ogg"},
	{0xEC4D8B9740B78BFFULL, "art\\characters\\playercharacter\\elffemale\\glv_nuda.mmh"},
	{0xEC4D919740B795A1ULL, "art\\characters\\playercharacter\\elffemale\\glv_nuda.msh"},
	{0xEC4DABEE42DBA587ULL, "vxw_encounter2_camp_mt_0_0.vlm"},
	{0xEC4DEC2F51924429ULL, "art\\vfx\\cinematic\\vfx_meredith_sword_c.anb"},
	{0xEC4DED949927CC5CULL, "art\\characters\\playercharacter\\humanmale\\rob_appp.phy"},
	{0xEC4E02EF3660475AULL, "vxht_0_hightown_a2_c53_0_0.vlm"},
	{0xEC4E55730D8B8C0FULL, "llac_lt_0_lowtown_a2_n_131071_6.dds"},
	{0xEC4EA9438C0BE231ULL, "6148154_m.fxe"},
	{0xEC4F5720A257401AULL, "dr_0_deeproads_5845.rml"},
	{0xEC4F642BC0876972ULL, "dae150ip_barrier_pride.utp"},
	{0xEC4FD9A928C51B28ULL, "hlac_ht_keep_3732_4x.dds"},
	{0xEC50082B78667C3FULL, "317412788.ogg"},
	{0xEC501FF7B922272BULL, "art\\vfx\\combat\\vfx_str_spit1_act_c.impa.evt"},
	{0xEC5076457F822F14ULL, "vxca_caverns1_geu_0_0.vlm"},
	{0xEC50C687FCAE03C0ULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.mmh"},
	{0xEC5144CA48264FEFULL, "550678655.ogg"},
	{0xEC51E4EB192078DAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_3.gad"},
	{0xEC51E52DD2B48A94ULL, "1020423905.ogg"},
	{0xEC5206AB57CE436CULL, "593589009.ogg"},
	{0xEC53149C87816A86ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_004a_0n.dds"},
	{0xEC5348B9160D77EEULL, "vxht_0_hightown_n_bau_0_0.vlm"},
	{0xEC5366E0DBDF2ADCULL, "562160668.ogg"},
	{0xEC5481BA6056404DULL, "hlac_lt_darktown_-1_80y.dds"},
	{0xEC5578F3725AF942ULL, "6105415_m.fxe"},
	{0xEC55DBC28A95F462ULL, "vxlt_warehouse_pm_0_0.vlm"},
	{0xEC55F0A6ED15A1CBULL, "6078790_m.fxe"},
	{0xEC5738F7B92870ECULL, "art\\vfx\\combat\\vfx_str_spit1_act_c.impa.gad"},
	{0xEC57C2286EE720CAULL, "vxht_0_hightown_box_0_0.vlm"},
	{0xEC58315B9FE6AEDFULL, "vxht_0_hightown_cdm_0_0.vlm"},
	{0xEC584B45B6E2A7B2ULL, "vxht_mansion2_1xz_0_0.vlm"},
	{0xEC588097B737774CULL, "art\\vfx\\combat\\vfx_str_att_back_l_leg_c.mmh"},
	{0xEC58E7EB19269CF1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_3.evt"},
	{0xEC59ECB9F33E4143ULL, "qun221_viscount6070179.cl"},
	{0xEC5A147278AF6AFEULL, "753682971.ogg"},
	{0xEC5A1881B05F2471ULL, "hlac_lt_0_lowtown_docks_-3_4y.dds"},
	{0xEC5A5A9B2D1DF2C1ULL, "6131200_m.fxe"},
	{0xEC5B5CBF22BC7EA3ULL, "qun231_varnell.bnk"},
	{0xEC5BC0E0BBC3CCD2ULL, "mag311_keran6041409.cl"},
	{0xEC5BDF6E76888722ULL, "vxlt_0_lowtown_a2_n_gf5_0_0.vlm"},
	{0xEC5C027CCA1BD9F8ULL, "301055714.ogg"},
	{0xEC5C1DCB34D1BE07ULL, "vxw_encounter2_camp_hw_0_0.vlm"},
	{0xEC5C305B164E9D4AULL, "marksman_knockbolt.dds"},
	{0xEC5C4BDCB1C6637AULL, "6181903_m.fxe"},
	{0xEC5C823D70FE55D6ULL, "6141648_m.fxe"},
	{0xEC5CC1B4BA0F5BCFULL, "vxca_caverns2prc_1zo_0_0.vlm"},
	{0xEC5CC958948A70C2ULL, "vxlt_gallowscourtyard_63y_0_0.vlm"},
	{0xEC5CF96F6FA95E1EULL, "vxht_0_hightown_n_c5d_0_0.vlm"},
	{0xEC5D01872C30FE58ULL, "6162116_f.fxe"},
	{0xEC5D2F83E1913BD7ULL, "zz_fol_debug6057014.cl"},
	{0xEC5E37F7B92E8DEAULL, "art\\vfx\\combat\\vfx_str_spit1_act_c.impa.ani"},
	{0xEC5E4FA220ECFA7BULL, "6054514_m.fxe"},
	{0xEC5EA987FCBA0FDFULL, "art\\vfx\\combat\\vfx_ha_bd_bfrz_frd_lp_c.anb"},
	{0xEC5F3EBF22C034D3ULL, "qun231_varnell.cnv"},
	{0xEC5F4DAC54FF8F87ULL, "6060478_m.fxe"},
	{0xEC5F5965DF90F33FULL, "976374656.ogg"},
	{0xEC60194BE590F28EULL, "hlac_lt_0_lowtown_a2_-65537_1y.dds"},
	{0xEC60A1CB80187209ULL, "vxlt_gallowsprison_n_6bp_0_0.vlm"},
	{0xEC60CC57AD487B4EULL, "401874916.ogg"},
	{0xEC60E9A64682AF22ULL, "97342252.ogg"},
	{0xEC611EEEA78F937BULL, "character_character_p0.psh.fxo.dx11"},
	{0xEC618FEA5E192CA5ULL, "vxlt_gallowstemplar_n_34s_0_0.vlm"},
	{0xEC61C35EBC6B82AEULL, "vxht_playershouse_n_1uc_0_0.vlm"},
	{0xEC61C5633EC69DF3ULL, "one280im_merin_key.uti"},
	{0xEC61EB043D9DD3E9ULL, "mag160_feynriel6020079.cl"},
	{0xEC621BB6F8D3B362ULL, "dae000_arishok6096335.cl"},
	{0xEC621D57A4B32F42ULL, "vxlt_0_lowtown_a3_n_c_fg8_0_0.vlm"},
	{0xEC6226B478F2B184ULL, "ht_0_hightown_a2_22_0.rml"},
	{0xEC624B9691E0E974ULL, "978600473.ogg"},
	{0xEC6275F7EE0A928EULL, "zz_vault_debug6177611.cl"},
	{0xEC62CEBD71E3D256ULL, "vxht_0_hightown_a2_5jd_0_0.vlm"},
	{0xEC62E116BA2D36B7ULL, "6066515_m.fxe"},
	{0xEC63174CC9F140FCULL, "206253819.ogg"},
	{0xEC63E19A63A4B90DULL, "vxlt_0_lowtown_a3_n_c_hlh_0_0.vlm"},
	{0xEC641782DF0D19A4ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar3_.ani"},
	{0xEC644D9F7C8C0CE2ULL, "vxlt_warehouse_un_0_0.vlm"},
	{0xEC645A260114AB2FULL, "vxdr_primevaltomb_6yr_0_0.vlm"},
	{0xEC6461C8B9002112ULL, "vxw_sundermount_2bk_0_0.vlm"},
	{0xEC64E59285772BD3ULL, "hf_sailor01.mor"},
	{0xEC6501D24AB2A7D4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medd_0s.dds"},
	{0xEC652332349A45A8ULL, "6139528_m.fxe"},
	{0xEC65289591CE81D2ULL, "676890716.ogg"},
	{0xEC65D6B273FD9E65ULL, "dae000_merchant_access_hf.bnk"},
	{0xEC65F3969C13461DULL, "hlac_w_blightlands_-131069_0y.dds"},
	{0xEC65FFEB19314FC0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_3.ani"},
	{0xEC66340560DD394AULL, "vxht_0_hightown_bhy_0_0.vlm"},
	{0xEC66B3E03935F36EULL, "vxht_0_hightown_n_9b7_0_0.vlm"},
	{0xEC67BB78C3404D44ULL, "6069034_m.fxe"},
	{0xEC67E3C0E93E8EA3ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.impa.evt"},
	{0xEC67ED8CADAC89E2ULL, "360724721.ogg"},
	{0xEC68520096EBAB52ULL, "72973226.ogg"},
	{0xEC68AAA21D5E5B67ULL, "6179996_m.fxe"},
	{0xEC68DC8BFE7A0175ULL, "atl_chargen_st_dxt5_00.dds"},
	{0xEC6910D11E184132ULL, "6099880_m.fxe"},
	{0xEC69731AD967B5E0ULL, "21926820.ogg"},
	{0xEC69ABF83870804CULL, "vxca_caverns2_73t_0_0.vlm"},
	{0xEC69B8B27401539DULL, "dae000_merchant_access_hf.cnv"},
	{0xEC69C98076813E2FULL, "740737810.ogg"},
	{0xEC69DD9C72ABB89DULL, "vxlt_0_lowtown_a2_n_csi_0_0.vlm"},
	{0xEC69ED55E0D44BBEULL, "ave200pt_opening.ncs"},
	{0xEC6A04DE5D72E016ULL, "hlac_lt_0_lowtown_n_-196609_1y.dds"},
	{0xEC6A1482DF118006ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr370cs_attack_begins\\max files\\maletemplar3_.gad"},
	{0xEC6A514B687E9FA2ULL, "vxlt_0_lowtown_a2_n_gc6_0_0.vlm"},
	{0xEC6A8FA826C7D687ULL, "vxw_encounter2_camp_cv_0_0.vlm"},
	{0xEC6A95AD79B2C515ULL, "vxw_woundedcoast_h4_0_0.vlm"},
	{0xEC6ACDE05604BC0AULL, "vxca_caverns1_igo_0_0.vlm"},
	{0xEC6AE90DCB0CBD45ULL, "44537327.ogg"},
	{0xEC6B0DDCF550FC6DULL, "llac_lt_0_lowtown_131073_1.dds"},
	{0xEC6B5BAAB951F59AULL, "llac_w_encounter3_-131067_0.dds"},
	{0xEC6B6344AC6D220AULL, "evl_clo.mao"},
	{0xEC6B6ACEB395F6E0ULL, "hlac_dr_primevaltomb_9621_8x.dds"},
	{0xEC6B895E2DC14F7EULL, "ico_knockedback.dds"},
	{0xEC6C38962A8C5815ULL, "pro000cr_refugee2.utc"},
	{0xEC6C5B8630B6AE00ULL, "279489904.ogg"},
	{0xEC6CD87E98B894DAULL, "ip_restrict_attribute_str.ncs"},
	{0xEC6D642A995B06D7ULL, "ser270_templar6149981.cl"},
	{0xEC6D6DB3A9111DBDULL, "ave310_donnic6134726.cl"},
	{0xEC6DAAEDF5B7D848ULL, "332773375.ogg"},
	{0xEC6DD65830F87748ULL, "6085394_f.fxe"},
	{0xEC6EBCC0E9446B24ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.impa.gad"},
	{0xEC6EE287405A80CAULL, "vxlt_gallowstemplar_5av_0_0.vlm"},
	{0xEC6F5247CDB05AFBULL, "932675331.ogg"},
	{0xEC6F97317CC9BF55ULL, "clf221_jalen6070351.cl"},
	{0xEC6FD9AF2D1E9986ULL, "116992529.ogg"},
	{0xEC6FF1F380A3FE0FULL, "vxht_brothel_n_4d6_0_0.vlm"},
	{0xEC70033BBC97A6DEULL, "41264889.ogg"},
	{0xEC706A87CA137B5DULL, "zz_qun_debug6088079.cl"},
	{0xEC707FE7FFDA149EULL, "ser221_gascard6127890.cl"},
	{0xEC70E26AE0228DF2ULL, "vxht_keep_n_5he_0_0.vlm"},
	{0xEC70F0423F9C20EFULL, "204323887.ogg"},
	{0xEC7183C2193E244BULL, "hlac_w_bonepitmining_a2_196610_0x.dds"},
	{0xEC71A4B0080AD230ULL, "character_character_p157.psh.fxo.dx11"},
	{0xEC7261F5740A97E8ULL, "zz_openwhite_cu.cut"},
	{0xEC7268B220299B41ULL, "vxht_chantry_358_0_0.vlm"},
	{0xEC72CB48475204B8ULL, "one280_iwan6096870.cl"},
	{0xEC73286BD9BC6369ULL, "vxlt_gallowscourtyard_66u_0_0.vlm"},
	{0xEC73301FF84714FAULL, "gen00fl_aveline6114849.cl"},
	{0xEC733031E09B55E4ULL, "hlac_lt_undercityserialkiller_-1_14y.dds"},
	{0xEC73A5F309796FF5ULL, "vxht_chantry_n_352_0_1.vlm"},
	{0xEC73CCBAAEF27DA5ULL, "llac_lt_0_lowtown_docks_a3_n_-196610_5.dds"},
	{0xEC7434F8649C1704ULL, "hlac_w_bonepitdestroyed_a3_0_8x.dds"},
	{0xEC7476837128585FULL, "6094961_m.fxe"},
	{0xEC74A301F739C54EULL, "vxdr_primevaltomb_3ox_0_0.vlm"},
	{0xEC75CBC0E94AA302ULL, "art\\vfx\\combat\\vfx_f_mer_ingrasp_tag_lp_c.impa.ani"},
	{0xEC75EC3150BA6023ULL, "vxca_cavern1_haunted_j29_0_2.vlm"},
	{0xEC76149DFBCCF613ULL, "vxca_caverns1_cli_0_0.vlm"},
	{0xEC766EF8A49294C0ULL, "6051590_m.fxe"},
	{0xEC76BB3C9DB7F557ULL, "vxht_0_hightown_a2_9zu_0_0.vlm"},
	{0xEC76FDB9DF7818C9ULL, "155042170.ogg"},
	{0xEC777F7514CB9618ULL, "isa211_isabela6175163.cl"},
	{0xEC779AE7E180EF6BULL, "vxdr_primevaltomb_75w_0_0.vlm"},
	{0xEC77F9CB0460FAB4ULL, "vxw_bonepit_1kf_0_0.vlm"},
	{0xEC78017B551B16B7ULL, "vxw_bonepit_26y_0_0.vlm"},
	{0xEC79078A6BA8DD95ULL, "vxw_woundedcoast_e3_0_0.vlm"},
	{0xEC790BC665ADC594ULL, "hlac_ht_0_hightown_a3_65537_1y.dds"},
	{0xEC793FBD47FAD48AULL, "vxca_caverns1_ijh_0_0.vlm"},
	{0xEC7958CB23C3DB39ULL, "671982653.ogg"},
	{0xEC79839D47CF1C5BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basic.mmh"},
	{0xEC7987220CB13F99ULL, "gen00fl_aveline6114794.cl"},
	{0xEC798F9D47CF30F8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basic.mao"},
	{0xEC79A19D47CF4F45ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basic.msh"},
	{0xEC79F1D593A92644ULL, "lt_0_lowtown_docks_143_0.mmh"},
	{0xEC79F5D593A92C97ULL, "lt_0_lowtown_docks_143_0.mao"},
	{0xEC79F9ED7AC6ADA0ULL, "vxht_0_hightown_c7c_0_0.vlm"},
	{0xEC79FC9C145DD4BBULL, "vxht_chantry_n_2tf_0_1.vlm"},
	{0xEC7A07D593A94BE6ULL, "lt_0_lowtown_docks_143_0.msh"},
	{0xEC7AF05FE300A8DCULL, "character_character_p94.psh.fxo.dx11"},
	{0xEC7B71606A4F58F6ULL, "qun100cr_warrior2.utc"},
	{0xEC7C233A35EDAE8DULL, "plt_cod_hst_parvollen.plo"},
	{0xEC7C44C51D541281ULL, "6164185_m.fxe"},
	{0xEC7CDBFEA9446FECULL, "vxlt_undercityserialkiller_1z3_0_0.vlm"},
	{0xEC7D7DEB3676EDD3ULL, "hlac_lt_0_lowtown_docks_n_65534_4y.dds"},
	{0xEC7DF631C26142E1ULL, "zz_dae_debug6029090.cl"},
	{0xEC7E6D030625BBAAULL, "6134712_m.fxe"},
	{0xEC7E81DB45D50512ULL, "417418964.ogg"},
	{0xEC7ECAE05995B2D0ULL, "hla_ht_0_hightown_-1_0z.dds"},
	{0xEC7ECE94EF173D83ULL, "w_encounter1_58_0.mao"},
	{0xEC7ED294EF1744D0ULL, "w_encounter1_58_0.mmh"},
	{0xEC7EE094EF175CEAULL, "w_encounter1_58_0.msh"},
	{0xEC7EEDE6E5FEB257ULL, "vxlt_backalley_n_1e4_0_0.vlm"},
	{0xEC7F3DF3F3784A23ULL, "127123472.ogg"},
	{0xEC7F4E46C0476FEAULL, "dae200cr_viscount6087378.cl"},
	{0xEC7F73DECE298BC2ULL, "vxlt_0_lowtown_e2l_0_0.vlm"},
	{0xEC805AAA0684060EULL, "vxw_encounter2_p_0_0.vlm"},
	{0xEC8062D366CE35ECULL, "6074524_m.fxe"},
	{0xEC811926885B0852ULL, "725644826.ogg"},
	{0xEC81488490254D27ULL, "war_mightyblow.dds"},
	{0xEC8176A92C49486FULL, "6149241_m.fxe"},
	{0xEC81BD1C9D00F52EULL, "gen_im_trs_gen_one_lsw_r0.uti"},
	{0xEC8213CAA22415FEULL, "235769677.ogg"},
	{0xEC823EA7EC738218ULL, "vxlt_0_lowtown_a2_n_c_f9h_0_1.vlm"},
	{0xEC82C1CB877AD2D6ULL, "vxlt_gallowstemplar_4sq_0_0.vlm"},
	{0xEC82D5D1911ECA5AULL, "vxw_bonepit_1pe_0_0.vlm"},
	{0xEC831BF057B6C6D7ULL, "el_lp_inside_day.mmh"},
	{0xEC83796F1E3BA20DULL, "417793310.ogg"},
	{0xEC83EF63BD7AE166ULL, "vxht_mansion2b_ad_0_0.vlm"},
	{0xEC847CF02F3324F8ULL, "6154734_m.fxe"},
	{0xEC8483D4A446826FULL, "hlac_lt_0_lowtown_docks_a3_-65539_6y.dds"},
	{0xEC84949F76C82588ULL, "892098765.ogg"},
	{0xEC850AD03D4C9460ULL, "vxdr_0_deeproads_ebu_0_0.vlm"},
	{0xEC862D8BDCCFDB96ULL, "6119177_m.fxe"},
	{0xEC866BA7F6571334ULL, "vxw_bonepit_1lg_0_0.vlm"},
	{0xEC86735847112F37ULL, "vxw_bonepit_21x_0_0.vlm"},
	{0xEC8673816F1FDFA2ULL, "one100dg_vanard6173963.cl"},
	{0xEC8697D76CACB7C5ULL, "dre000_dougal6131183.cl"},
	{0xEC86AAB403F152DAULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_loop.gad"},
	{0xEC86E0116D3198BAULL, "mcr310_argument6072276.cl"},
	{0xEC877A4FE0A2630FULL, "508057709.ogg"},
	{0xEC87B19A39F0ED0AULL, "vxca_caverns1_imi_0_0.vlm"},
	{0xEC8868C2056F5F06ULL, "6134965_m.fxe"},
	{0xEC897FC6D5C1D1C9ULL, "vxca_caverns2_7wa_0_0.vlm"},
	{0xEC8985C7F2DAB52DULL, "hlac_lt_0_lowtown_a2_1_5x.dds"},
	{0xEC8A293352B66882ULL, "gen00fl_aveline6139303.cl"},
	{0xEC8A58FE689FE171ULL, "25385011.ogg"},
	{0xEC8A701B2C2E7C46ULL, "vxlt_0_lowtown_docks_n_b82_0_1.vlm"},
	{0xEC8A93426F681D22ULL, "zz_combat_demo_ar.ncs"},
	{0xEC8B6B134B8A5E09ULL, "680599983.ogg"},
	{0xEC8BC37373290FDEULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_013a_0s.dds"},
	{0xEC8BC6412446B1CAULL, "vxlt_gallowstemplar_5wp_0_0.vlm"},
	{0xEC8C1257DB99395BULL, "305963445.ogg"},
	{0xEC8C9730298F25C6ULL, "qun211_braggart6075685.cl"},
	{0xEC8D3388AAB4B3B6ULL, "vxht_0_hightown_n_c69_0_0.vlm"},
	{0xEC8D857ED3E2FFCAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_wmda_0d.dds"},
	{0xEC8DB90BF9C01D50ULL, "6096224_f.fxe"},
	{0xEC8E572A4ACC4596ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf_eyesm1.mao"},
	{0xEC8E719E119ABA97ULL, "406958366.ogg"},
	{0xEC8F5DA0414B63B6ULL, "gen00fl_carver6097260.cl"},
	{0xEC8F8983AA5A08C4ULL, "6162018_m.fxe"},
	{0xEC8F96079266D90CULL, "vxlt_0_lowtown_a2_gn0_0_0.vlm"},
	{0xEC8FB8A856215EFBULL, "6135798_m.fxe"},
	{0xEC902C9A70CE971FULL, "6162512_m.fxe"},
	{0xEC9060118B9DBAAFULL, "s_ballista_dust_l3.mao"},
	{0xEC911A4C4C3A0DA0ULL, "vxlt_gallowstemplar_45m_0_0.vlm"},
	{0xEC914327756C2413ULL, "vxht_brothel_n_62r_0_0.vlm"},
	{0xEC9147AE8314E2DAULL, "vxw_bonepit_1uf_0_0.vlm"},
	{0xEC92128A67A56A32ULL, "6122364_m.fxe"},
	{0xEC9294AC142B61B0ULL, "817558143.ogg"},
	{0xEC931CC09E014A04ULL, "hlac_lt_0_lowtown_docks_131073_1x.dds"},
	{0xEC9335CDF0502A42ULL, "gen00fl_aveline6112713.cl"},
	{0xEC9339A4162AD6CEULL, "1012645418.ogg"},
	{0xEC936334EE436B21ULL, "vxw_woundedcoast_c9_0_0.vlm"},
	{0xEC93A87EE0162321ULL, "6077007_m.fxe"},
	{0xEC93FFC2FB264264ULL, "vxht_0_hightown_a3_b82_0_0.vlm"},
	{0xEC94C5403357DD7DULL, "dae000_mother6035917.cl"},
	{0xEC94C65905AA9B6EULL, "dae000_tomwise6167326.cl"},
	{0xEC94CA07014E359FULL, "vxlt_0_lowtown_docks_n_bf4_0_0.vlm"},
	{0xEC94D13D1CD8B8E8ULL, "gen00fl_isabela6099529.cl"},
	{0xEC956D38803BF298ULL, "isa211_isabela6202281.cl"},
	{0xEC95D686843EB5AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_rdy.ani"},
	{0xEC9601BA51E2CE96ULL, "zz_qcr_debug6055517.cl"},
	{0xEC96AA5109D5278EULL, "6093965_m.fxe"},
	{0xEC9736DC59477FBDULL, "mag291_demon_marethari6180636.cl"},
	{0xEC9773148CDC6450ULL, "matdef_dw_diffuse.dds"},
	{0xEC978FBEB253FA56ULL, "w_encounter3_4_0.rml"},
	{0xEC97C9743C720E1EULL, "vxht_0_hightown_a2_c7q_0_0.vlm"},
	{0xEC97FCDE4E4EE5CDULL, "vxw_bonepitdestroyed_a3_e6_0_0.vlm"},
	{0xEC97FF7A163A3F5CULL, "vxlt_gallowstemplar_n_3o3_0_0.vlm"},
	{0xEC984A32FC9A5F92ULL, "6201371_m.fxe"},
	{0xEC987ACF12A5B33EULL, "6048419_m.fxe"},
	{0xEC98870BB4DAF8EFULL, "6141928_m.fxe"},
	{0xEC98C06C1F7A52FDULL, "vxlt_darktown_4nq_0_0.vlm"},
	{0xEC98E1F81E2494C6ULL, "vxlt_0_lowtown_docks_n_b55_0_1.vlm"},
	{0xEC98E43D4781BA05ULL, "147498284.ogg"},
	{0xEC98FA952BA3DC60ULL, "223440472.ogg"},
	{0xEC98FC9998F62B23ULL, "173827300.ogg"},
	{0xEC99B5D3EF6DACE6ULL, "vxw_bonepitdestroyed_a3_1v3_0_0.vlm"},
	{0xEC99BFB88D30A0ECULL, "vxlt_undercityserialkiller_1t1_0_0.vlm"},
	{0xEC99DA16075AE484ULL, "vxlt_0_lowtown_docks_b60_0_0.vlm"},
	{0xEC9A19E9842B8354ULL, "w_woundedcoast_15_0_ll.mmh"},
	{0xEC9A2D1B466A6317ULL, "hlac_lt_0_lowtown_65536_12x.dds"},
	{0xEC9AC5B4040229C0ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\hands_on_face_loop.ani"},
	{0xEC9AE0C7D6D56275ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowa_n.dds"},
	{0xEC9B043DE60A9C1DULL, "6090192_m.fxe"},
	{0xEC9B2D10361AD6D3ULL, "250841553.ogg"},
	{0xEC9B92057F21E27AULL, "6179391_m.fxe"},
	{0xEC9BD1A0C9EAE357ULL, "vxlt_backalley_n_1k6_0_0.vlm"},
	{0xEC9BE930FD26344EULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magiclecturn_sa.dds"},
	{0xEC9C13FB38E1FF65ULL, "vxw_woundedcoast_hd_0_0.vlm"},
	{0xEC9C773A4BA83E00ULL, "one000_ilen6056859.cl"},
	{0xEC9C9B64096A90B7ULL, "art\\vfx\\combat\\vfx_str_appearance_c.cess.gad"},
	{0xEC9CAC7CAF8E0A87ULL, "6085541_m.fxe"},
	{0xEC9CDC20D97B2C42ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ki_transition.gad"},
	{0xEC9D2DDA3CE326F8ULL, "6081704_m.fxe"},
	{0xEC9D5A66D3359CFDULL, "161627091.ogg"},
	{0xEC9D8EBFEF368995ULL, "rdr111dg_martin6054891.cl"},
	{0xEC9E2C3916AF770BULL, "6174119_m.fxe"},
	{0xEC9E3A3F1F56C412ULL, "xx_brick.dds"},
	{0xEC9E63B93E41324DULL, "6177952_m.fxe"},
	{0xEC9F490739D2AFFAULL, "vxw_encounter1_cj_0_0.vlm"},
	{0xEC9F51BE402C16FAULL, "vxlt_hangedman_n_rv_0_0.vlm"},
	{0xECA044E1A677ADA4ULL, "zz_sneak01.arl"},
	{0xECA071580396101AULL, "870849848.ogg"},
	{0xECA1E58AC63C9A78ULL, "vxdr_0_deeproads_et7_0_0.vlm"},
	{0xECA1F2BB01930ABCULL, "gen00fl_sebastian6195868.cl"},
	{0xECA21883D174AA35ULL, "63101918.ogg"},
	{0xECA2B064096F2044ULL, "art\\vfx\\combat\\vfx_str_appearance_c.cess.evt"},
	{0xECA2C05B2816A84BULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_arm_nud_0n.dds"},
	{0xECA2DE0AEE4B1E70ULL, "501141668.ogg"},
	{0xECA306A343EC9350ULL, "gen00fl_sebastian6177891.cl"},
	{0xECA33BE3F3444E1FULL, "vxlt_0_lowtown_docks_n_bc7_0_0.vlm"},
	{0xECA36AA0CA69BBE7ULL, "dae000_keep_bureaucrat_1_st6165314.cl"},
	{0xECA3858828E16FEBULL, "6167362_m.fxe"},
	{0xECA3A2DF01EEB1D7ULL, "hlac_lt_undercityencounter_-1_18x.dds"},
	{0xECA3CE86844AE5FFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_rdy.evt"},
	{0xECA3FF20D9818639ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ki_transition.evt"},
	{0xECA415EEBAACFC22ULL, "11818641.ogg"},
	{0xECA45D21418B2715ULL, "vxw_woundedcoast_t0_0_0.vlm"},
	{0xECA45F2990E0FC3CULL, "6099716_m.fxe"},
	{0xECA486F7B70AE8A6ULL, "zz_and_debug6134141.cl"},
	{0xECA4A5046B06E099ULL, "llac_lt_0_lowtown_docks_a3_n_-2_4.dds"},
	{0xECA4BC6306B84DB0ULL, "6155856_m.fxe"},
	{0xECA4BE52124B2377ULL, "6016829_f.fxe"},
	{0xECA4C5170DD9A316ULL, "dae011_gossips6149784.cl"},
	{0xECA4CE06C4BFCF5FULL, "447534372.ogg"},
	{0xECA4EE76788FABFAULL, "671498027.ogg"},
	{0xECA540E729A51AFDULL, "408415933.ogg"},
	{0xECA564A7ED513E08ULL, "zz_followers_banter6157332.cl"},
	{0xECA5FF5E869393D5ULL, "zz_dae_debug6036573.cl"},
	{0xECA60D4694F62960ULL, "164434570.ogg"},
	{0xECA63B512E68269EULL, "vxht_0_hightown_a2_c2r_0_0.vlm"},
	{0xECA7023ED4DCCA4DULL, "6174719_m.fxe"},
	{0xECA7059D4E4E6A22ULL, "vxlt_0_lowtown_a2_n_bxz_0_0.vlm"},
	{0xECA716046CAD97CDULL, "vxlt_0_lowtown_docks_a3_n_brq_0_0.vlm"},
	{0xECA768CE4A3BAAC3ULL, "6161359_m.fxe"},
	{0xECA803DC42A95D3DULL, "vxlt_0_lowtown_a2_hdp_0_0.vlm"},
	{0xECA82FBB2ACDA691ULL, "hlac_lt_0_lowtown_a2_131071_6y.dds"},
	{0xECA8475CF6AAB6D0ULL, "dae000_meredith6080993.cl"},
	{0xECA8705CC22E7256ULL, "258424139.ogg"},
	{0xECA87859A8AA6EE5ULL, "6077072_m.fxe"},
	{0xECA8B04705045D6BULL, "lt_0_lowtown_n_37_0.anb"},
	{0xECA8B1E61B8E6A0EULL, "622329369.ogg"},
	{0xECA8DD2CA84BC0C5ULL, "6163610_m.fxe"},
	{0xECA8E751B89C9CF3ULL, "mag310cr_scion_2.utc"},
	{0xECA92264CD799134ULL, "437519898.ogg"},
	{0xECA9696C238B4BDBULL, "gen00fl_fenris6113965.cl"},
	{0xECA9A0A8AC5FF3F3ULL, "zz_explosion.cut"},
	{0xECA9B89B3FFA150BULL, "hlac_lt_backalley_g_0_7z.dds"},
	{0xECAA11C2220740B5ULL, "vxht_mansion2b_1t5_0_0.vlm"},
	{0xECAA85D82AD817E5ULL, "vxw_woundedcoast_ec_0_0.vlm"},
	{0xECAA88B22C64F429ULL, "mag100_thrask6034806.cl"},
	{0xECAAA9FAE5955967ULL, "static_static_p20.psh.fxo"},
	{0xECAAAF868450CFE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_rdy.gad"},
	{0xECAAD3EA48AEA1D3ULL, "851644098.ogg"},
	{0xECAAEDB6BD80DE2BULL, "zz_otl_debug6060588.cl"},
	{0xECAB0A5B5A2991CCULL, "6016804_f.fxe"},
	{0xECAB13687C6ECA7AULL, "vxca_caverns1_mines_e90_0_0.vlm"},
	{0xECAB9FB72ED93F78ULL, "6084404_m.fxe"},
	{0xECABEBC60F2134E5ULL, "hlac_ht_0_hightown_-2_0y.dds"},
	{0xECAC266A417F9D0CULL, "195595529.ogg"},
	{0xECAD1292A490A94CULL, "one100it_gland.uti"},
	{0xECAD93E8E6AA0D8CULL, "100853355.ogg"},
	{0xECADAC3C46C746FCULL, "6079800_m.fxe"},
	{0xECADBAE42BC8C87AULL, "vxw_encounter1_hk_0_0.vlm"},
	{0xECADCB1029C6A89BULL, "6088689_m.fxe"},
	{0xECADEF2B2C51E2C1ULL, "mrl000_marethari6129420.cl"},
	{0xECAEB6562E517C6AULL, "vxw_bonepit_y9_0_0.vlm"},
	{0xECAEBDFBDDC3163EULL, "vxlt_darktown_5kl_0_0.vlm"},
	{0xECAEFC05CA622F5DULL, "755202875.ogg"},
	{0xECAF3390CAE1AD71ULL, "one350_variel6085213.cl"},
	{0xECAF52C34ED3A20BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkleft.ani"},
	{0xECAF921A0515046BULL, "fx_dust_bl_h.dds"},
	{0xECAFB22C029AAE62ULL, "vxlt_blackemporium_13m_0_0.vlm"},
	{0xECB09497410C1844ULL, "art\\characters\\playercharacter\\elffemale\\glv_nuda.phy"},
	{0xECB0C864097B863DULL, "art\\vfx\\combat\\vfx_str_appearance_c.cess.ani"},
	{0xECB10720D98C1E28ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ki_transition.ani"},
	{0xECB12F903AB2E324ULL, "618993851.ogg"},
	{0xECB14EEF49C10862ULL, "lt_0_lowtown_docks_a3_n_176_0.msh"},
	{0xECB150EF49C10B18ULL, "lt_0_lowtown_docks_a3_n_176_0.mmh"},
	{0xECB15CC7EA532028ULL, "6117111_m.fxe"},
	{0xECB15CEF49C11FBBULL, "lt_0_lowtown_docks_a3_n_176_0.mao"},
	{0xECB17D989E360D8DULL, "follower_banter6164047.cl"},
	{0xECB250FFA766E9F4ULL, "vxlt_darktown_428_0_0.vlm"},
	{0xECB251CDA5A58CEBULL, "vxlt_0_lowtown_n_gc7_0_0.vlm"},
	{0xECB30DD2C0717665ULL, "lt_0_lowtown_a2_n_c_28_0.rml"},
	{0xECB3C1E234C6FF82ULL, "463299436.ogg"},
	{0xECB3CB3A89E1E867ULL, "vxlt_0_lowtown_a3_n_c_htb_0_0.vlm"},
	{0xECB40A65DD5CB271ULL, "art\\characters\\creatures\\dragon\\animation\\die_v1.ani"},
	{0xECB47B9613D5120FULL, "vxlt_0_lowtown_a2_h28_0_0.vlm"},
	{0xECB4AD2E205E3F1EULL, "vxht_0_hightown_a2_c9s_0_0.vlm"},
	{0xECB4C28C54236826ULL, "hlac_lt_0_lowtown_a2_n_c_65535_6z.dds"},
	{0xECB4D31366BC2196ULL, "zz_end_warriorshot.cut"},
	{0xECB54C33E3256C87ULL, "zz_followers_banter6157610.cl"},
	{0xECB5C4643001C5D6ULL, "6113805_m.fxe"},
	{0xECB677244227FD17ULL, "dae200_bran6094808.cl"},
	{0xECB6D632E9F8D43EULL, "130309574.ogg"},
	{0xECB70AC2AD3FAA74ULL, "6096927.cut"},
	{0xECB7BCDC7C02AE0DULL, "dae000_bartrand6206965.cl"},
	{0xECB8250DBD235A0CULL, "818941890.ogg"},
	{0xECB83E8073C68B5FULL, "plt_cod_hst_city_elf_cultur.plo"},
	{0xECB875BF630E137AULL, "6177191_m.fxe"},
	{0xECB8F7B51CCE3065ULL, "vxw_woundedcoast_bb_0_0.vlm"},
	{0xECB91BD7D78B71E7ULL, "static_static_p51.psh.fxo"},
	{0xECB9516F1C27A848ULL, "mcr310_argument6075806.cl"},
	{0xECB9819FDA44670DULL, "llac_ht_0_hightown_n_65533_1.dds"},
	{0xECB98E86F2E8B2F0ULL, "560263433.ogg"},
	{0xECB998E526153AEDULL, "vxlt_gallowstemplar_f_376_0_0.vlm"},
	{0xECB9C3A00D2A7577ULL, "6024526_m.fxe"},
	{0xECB9CE3278634A7DULL, "gen_im_wep_war_two_act3_03.uti"},
	{0xECBA0A034C9F7697ULL, "ttx_master.mao"},
	{0xECBAFB1F8B201DF8ULL, "vxlt_warehouse_11y_0_0.vlm"},
	{0xECBB1565DD62E313ULL, "art\\characters\\creatures\\dragon\\animation\\die_v1.gad"},
	{0xECBB1A6604768A70ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grand_1.mao"},
	{0xECBBBED989F4863DULL, "mortals.ncs"},
	{0xECBBDEF4F375C749ULL, "975882198.ogg"},
	{0xECBBEBA888C2DA5FULL, "aip_default.gda"},
	{0xECBC5A3F7A062D42ULL, "nat380_temmerin6067366.cl"},
	{0xECBD2FD8CFB92EBEULL, "vxlt_darktown_5lm_0_0.vlm"},
	{0xECBD3AC34EDFB682ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkleft.evt"},
	{0xECBD5896C8E8E6E4ULL, "llac_ht_bartrand_n_3369_2.dds"},
	{0xECBDDE4BFD2D18C0ULL, "316994967.ogg"},
	{0xECBE061A54340CC6ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_ripple02.dds"},
	{0xECBED994EF4D34EBULL, "w_encounter1_58_0.phy"},
	{0xECBF5DCA2006D593ULL, "vxw_encounter2_kj_0_0.vlm"},
	{0xECBF92760924085DULL, "367395831.ogg"},
	{0xECBFCFF9C3B3B663ULL, "6046947_f.fxe"},
	{0xECBFEF97370E18F5ULL, "6026993_m.fxe"},
	{0xECBFF1D035C17F36ULL, "vxw_woundedcoast_ov_0_0.vlm"},
	{0xECBFF8D3A59A9839ULL, "vxht_0_hightown_a2_c10_0_0.vlm"},
	{0xECC01F9DD7307F1FULL, "vxlt_0_lowtown_docks_n_bu1_0_0.vlm"},
	{0xECC0C2DC995D0274ULL, "vxlt_darktown_499_0_0.vlm"},
	{0xECC0CEF4E0C417B3ULL, "vxht_brothel_n_4bh_0_0.vlm"},
	{0xECC16E8742497C08ULL, "vxlt_0_lowtown_n_h24_0_0.vlm"},
	{0xECC1857C2E375C16ULL, "lt_0_lowtown_a2_35_0_ll.mmh"},
	{0xECC1A49D480D0990ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_45basic.phy"},
	{0xECC1BB5188435277ULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench1_merillintro.gad"},
	{0xECC1C1A381BFFAA0ULL, "6094812_m.fxe"},
	{0xECC1FD21387ECAF8ULL, "222696339.ogg"},
	{0xECC21265DD68FD78ULL, "art\\characters\\creatures\\dragon\\animation\\die_v1.evt"},
	{0xECC25A086C4B3603ULL, "gen00fl_fenris6105856.cl"},
	{0xECC3E05E8BF8C5DBULL, "dre000_carver6037594.cl"},
	{0xECC3E68641C252DFULL, "460898343.ogg"},
	{0xECC438E2CEAD247AULL, "hm_shw_bmage_lead1.mor"},
	{0xECC439C34EE5D395ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spear\\spearwalkleft.gad"},
	{0xECC46355C5CC17A7ULL, "550970614.ogg"},
	{0xECC50B6AC54FF666ULL, "vxw_bonepitdestroyed_a3_1a6_0_0.vlm"},
	{0xECC54CE99B6A1001ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzsacks_0.pwk"},
	{0xECC56F8E1C2ED2B0ULL, "6125187_m.fxe"},
	{0xECC5734504A954D2ULL, "mer210_smith6051397.cl"},
	{0xECC5C6380F57C35CULL, "vxlt_undercityencounter_ws_0_0.vlm"},
	{0xECC5FEC55B4D6DE6ULL, "6052751_m.fxe"},
	{0xECC65151F14D9364ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerc.msh"},
	{0xECC66351F14DB1CDULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerc.mao"},
	{0xECC66751F14DB806ULL, "art\\levels\\races\\ferelden\\props\\armory\\prp_daggerc.mmh"},
	{0xECC66D80C127D488ULL, "500314250.ogg"},
	{0xECC6D82A75B5A18BULL, "vxht_keep_n_63r_0_0.vlm"},
	{0xECC77BA616755468ULL, "vxca_cavern1_haunted_eaj_0_0.vlm"},
	{0xECC83A1EC7EB9BBEULL, "hlac_w_sundermount_-327679_0x.dds"},
	{0xECC8DB66AFAA5901ULL, "vxlt_gallowstemplar_n_3ja_0_0.vlm"},
	{0xECC923D59EE426B4ULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_grimoirea_open_d.dds"},
	{0xECC93ED7F5FA47FFULL, "zz_followers_banter6157562.cl"},
	{0xECC96CFC7D163678ULL, "vxlt_warehouse_14v_0_0.vlm"},
	{0xECCA40AB3E473F45ULL, "vxw_encounter1_ca_0_0.vlm"},
	{0xECCA6E32BCF8D5F0ULL, "6161644_m.fxe"},
	{0xECCAEB1CBDDDA28FULL, "hlac_lt_0_lowtown_docks_n_-131075_0x.dds"},
	{0xECCB0F224AED44DAULL, "vxw_bonepit_1ib_0_0.vlm"},
	{0xECCBF98F246E5CD6ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.cess.ani"},
	{0xECCBFA7FD3824838ULL, "vxht_mansion2b_1ao_0_0.vlm"},
	{0xECCC1C7D6B6ACA73ULL, "645895415.ogg"},
	{0xECCC3F1412CFB3ECULL, "294142307.ogg"},
	{0xECCD57E4FE233A1BULL, "636916781.ogg"},
	{0xECCD868D14B410D8ULL, "938157771.ogg"},
	{0xECCD949BD7C66B79ULL, "hlac_ht_0_hightown_a3_-65536_4y.dds"},
	{0xECCE8EF5AEF5CEE3ULL, "445944548.ogg"},
	{0xECCEC7819669BAA9ULL, "hlac_w_encounter1_65540_0z.dds"},
	{0xECCF2F073522147EULL, "vxht_mansion2b_1y1_0_0.vlm"},
	{0xECCFA0B69C8A7D1CULL, "6173953_m.fxe"},
	{0xECCFC525107ED2A6ULL, "312845832.ogg"},
	{0xECD0D1108D9DF16EULL, "vxht_mansion_n_60n_0_0.vlm"},
	{0xECD0E89B634F47F4ULL, "llac_w_sundermount_262140_0.dds"},
	{0xECD113C222AB04C2ULL, "154209584.ogg"},
	{0xECD17C6A34262200ULL, "6084008_m.fxe"},
	{0xECD1853B5C019B2CULL, "888233417.ogg"},
	{0xECD1A2C27E2AB421ULL, "6166477_m.fxe"},
	{0xECD1D0D5039A9519ULL, "fen110_fenris6117926.cl"},
	{0xECD1FAD4AE4005F7ULL, "6080649_m.fxe"},
	{0xECD219C0A34378C6ULL, "ave200_donnic6061405.cl"},
	{0xECD224310D973255ULL, "vxw_sundermount_2w7_0_0.vlm"},
	{0xECD22541113226ABULL, "gen_im_trs_cot_junk2_02.uti"},
	{0xECD28EF635D8E899ULL, "follower_banter6136354.cl"},
	{0xECD34577847E90E8ULL, "609799497.ogg"},
	{0xECD34E1116F2F39FULL, "396988925.ogg"},
	{0xECD355EE8735EB90ULL, "vxht_0_hightown_n_b93_0_0.vlm"},
	{0xECD411CC59EEA180ULL, "llac_ht_mansion2_818_62.dds"},
	{0xECD43815014DDBDCULL, "vxlt_undercityencounter_rp_0_0.vlm"},
	{0xECD465E190AFA1F3ULL, "vxlt_hangedman_n_26w_0_0.vlm"},
	{0xECD4AD47052992B7ULL, "lt_0_lowtown_n_37_0.rml"},
	{0xECD54D38513A3872ULL, "6051875_m.fxe"},
	{0xECD54F318D186E97ULL, "hlac_lt_0_lowtown_docks_n_196607_2y.dds"},
	{0xECD5A6D28CAC8EF6ULL, "follower_banter6171048.cl"},
	{0xECD5AD3D789350DBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottle_d.dds"},
	{0xECD5B5581E1CEE23ULL, "vxca_caverns1_mines_gkq_0_0.vlm"},
	{0xECD5E851885447FDULL, "art\\cinematics\\_crit path\\followerintros\\cs_hench1_merillintro.ani"},
	{0xECD607D5B11B063BULL, "vxlt_house_ex_0_0.vlm"},
	{0xECD653297DBA5403ULL, "399442985.ogg"},
	{0xECD6B79D3B9C8106ULL, "265231971.ogg"},
	{0xECD6F8DDFD5011B8ULL, "gen00fl_fenris6127123.cl"},
	{0xECD7D7C9DFCCD7E4ULL, "w_blightlands_238_0.msh"},
	{0xECD7E9C9DFCCF64DULL, "w_blightlands_238_0.mao"},
	{0xECD7EDC9DFCCFC86ULL, "w_blightlands_238_0.mmh"},
	{0xECD7F263A6E51FDAULL, "vxlt_darktown_4u2_0_1.vlm"},
	{0xECD800C054C8E765ULL, "6161502_m.fxe"},
	{0xECD86D8604CA56DEULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wo_diagonal_1.mao"},
	{0xECD8B19F770E0D7EULL, "art\\characters\\creatures\\dragonling\\textures\\cn_eye_draa_0s.dds"},
	{0xECD8F2CD930F1569ULL, "6131885_f.fxe"},
	{0xECD9018F2478F4C7ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.cess.evt"},
	{0xECD90C2659B153C3ULL, "llac_lt_0_lowtown_a2_n_-65537_0.dds"},
	{0xECD91FC6D95F034BULL, "gen00fl_aveline6139151.cl"},
	{0xECD92294537A8E9BULL, "gen00fl_fenris6127089.cl"},
	{0xECD959C696E1FDA8ULL, "6140663_m.fxe"},
	{0xECD9793C42154BDFULL, "mag330_walter6045858.cl"},
	{0xECD98F5234B63F7DULL, "334098064.ogg"},
	{0xECD999ADD103AA36ULL, "lgt120_orz_bln_official6140192.cl"},
	{0xECDA1392B3A55FBEULL, "vxlt_darktown_5zo_0_0.vlm"},
	{0xECDA396E7D9D4531ULL, "hlac_dr_primevaltomb_657_3x.dds"},
	{0xECDA4F325E8F4E12ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.dura.gad"},
	{0xECDA6C5CC57860B8ULL, "vxht_mansion2b_1dl_0_0.vlm"},
	{0xECDA946BD821CAD7ULL, "6202083_m.fxe"},
	{0xECDA989F463095C6ULL, "gen00fl_isabela6101543.cl"},
	{0xECDA9BF51415F6B3ULL, "6103320_m.fxe"},
	{0xECDB5AD12CD00972ULL, "vxw_bonepit_20v_0_0.vlm"},
	{0xECDB62041F6EA7B6ULL, "vxw_sundermount_1ym_0_0.vlm"},
	{0xECDB8B32A7B66D73ULL, "6096453_m.fxe"},
	{0xECDC3893100941C4ULL, "6147665_f.fxe"},
	{0xECDC3FBB0B44ADDBULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoork.mmh"},
	{0xECDC418403F30693ULL, "vxw_encounter2_ql_0_0.vlm"},
	{0xECDC4BBB0B44C278ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoork.mao"},
	{0xECDC5DBB0B44E0C5ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\dr1a_lowdoork.msh"},
	{0xECDC662C1ABAA38CULL, "471060644.ogg"},
	{0xECDCA2159155F18BULL, "444537666.ogg"},
	{0xECDCC43A9C0BBBE8ULL, "hlac_lt_0_lowtown_a2_-65535_4z.dds"},
	{0xECDCD58A19ADB036ULL, "vxw_woundedcoast_ax_0_0.vlm"},
	{0xECDCD7A9E580E6E0ULL, "6054737_m.fxe"},
	{0xECDD60B985A24C24ULL, "371538722.ogg"},
	{0xECDDDA37A0A7CA49ULL, "384968173.ogg"},
	{0xECDE0C4F5EE1C00BULL, "6128318-mrlromdipres2.cut"},
	{0xECDE12938E80959CULL, "6176653_m.fxe"},
	{0xECDE2E6408EC384FULL, "605856094.ogg"},
	{0xECDEA70D443DDDE8ULL, "gen00fl_dog6109241.cl"},
	{0xECDEB4FD93D502CBULL, "gen_im_wep_war_two_lore_04.uti"},
	{0xECDF7C179EC0BC01ULL, "441993182.ogg"},
	{0xECDFEF6D7A3FE444ULL, "580509647.ogg"},
	{0xECE0228F247F4B10ULL, "art\\vfx\\combat\\vfx_r_ibl_fli_pose4_p.cess.gad"},
	{0xECE04F399DFDDE1AULL, "gen00fl_bethany6101437.cl"},
	{0xECE068EE178CF608ULL, "vxht_keep_6xk_0_0.vlm"},
	{0xECE0960DFF8D4AD5ULL, "vxw_sundermount_2x6_0_0.vlm"},
	{0xECE0B9BC529785DDULL, "llac_lt_0_lowtown_a2_1_1.dds"},
	{0xECE0DBEC3AF409ABULL, "6037409_m.fxe"},
	{0xECE0EA35BC8E7155ULL, "675291784.ogg"},
	{0xECE10812DD911F85ULL, "gen00fl_varric6107276.cl"},
	{0xECE14DBF1B48AB41ULL, "6143025_m.fxe"},
	{0xECE152325E957249ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.dura.evt"},
	{0xECE1B88C33F1C365ULL, "docks_scene_dockworkers6176489.cl"},
	{0xECE205AF8E5FB354ULL, "vxca_caverns2_79l_0_2.vlm"},
	{0xECE244F970865D5AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_wmda_0t.dds"},
	{0xECE294CA90142A4EULL, "456537569.ogg"},
	{0xECE299A190956884ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundexit.ani"},
	{0xECE2EF2223DFEA9EULL, "str_explo.mao"},
	{0xECE3377ACDD07E86ULL, "667467226.ogg"},
	{0xECE33CADFC5E725AULL, "lt_0_lowtown_docks_a3_n_16_0.anb"},
	{0xECE3FE7E0C84BD54ULL, "lgt110cr_gxa_vgk_slvr_order.utc"},
	{0xECE4EB9B3006A47EULL, "hlac_lt_backalley_n_65537_1x.dds"},
	{0xECE5885A90829E26ULL, "262461650.ogg"},
	{0xECE59CEF49ADC574ULL, "llac_lt_0_lowtown_a3_n_c_65534_4.dds"},
	{0xECE5AA5B546F3508ULL, "vxht_0_hightown_n_bwa_0_0.vlm"},
	{0xECE5AB9B5B090777ULL, "203479280.ogg"},
	{0xECE5C0E088DEB244ULL, "300197985.ogg"},
	{0xECE60CD2C09CCD8DULL, "lt_0_lowtown_a2_n_c_28_0.anb"},
	{0xECE6A477705919F3ULL, "6021341_m.fxe"},
	{0xECE6D285D269AD09ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.impa.ani"},
	{0xECE6F986E049CC27ULL, "757092173.ogg"},
	{0xECE726A2AEFD4F44ULL, "mcr310cr_loyalist_mage_ldr.utc"},
	{0xECE76A77C7527F65ULL, "vxw_bonepit_1yy_0_0.vlm"},
	{0xECE77E55DBCF4110ULL, "vxlt_0_lowtown_a2_eh8_0_0.vlm"},
	{0xECE862B5BCBBF551ULL, "6129278_m.fxe"},
	{0xECE90648CA17E357ULL, "6209183_m.fxe"},
	{0xECE96920C1B2A824ULL, "6096927_f.fxe"},
	{0xECE996A1909B82E6ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundexit.gad"},
	{0xECEA35A0B5FDF485ULL, "6118516_f.fxe"},
	{0xECEA77262152945BULL, "vxht_0_hightown_c9w_0_1.vlm"},
	{0xECEAA48E9BB256BBULL, "dae000_bartrand6030649.cl"},
	{0xECEAB7C6AB52CC11ULL, "266741489.ogg"},
	{0xECEAC39FE2B978E0ULL, "hlac_lt_0_lowtown_a3_n_c_65535_4y.dds"},
	{0xECEAFAFEB080E85FULL, "mcr371_angry_mage_m6087028.cl"},
	{0xECEB621FE2B9D8B2ULL, "212214099.ogg"},
	{0xECEB7FF4B0886E1DULL, "vxca_caverns1_fmu_0_0.vlm"},
	{0xECEB85921777F32DULL, "gen00fl_anders6211957.cl"},
	{0xECEB88B29411BEB4ULL, "vxw_bonepit_13n_0_0.vlm"},
	{0xECEB9F14655EA2FDULL, "373904825.ogg"},
	{0xECEBAC3A55CF327CULL, "53716004.ogg"},
	{0xECEBC3AAF09BF3B2ULL, "hlac_lt_backalley_n_1_2x.dds"},
	{0xECEBCE518BD8B831ULL, "lt_0_lowtown_docks_a3_47_0.rml"},
	{0xECEBFCC0D507A1C3ULL, "vxlt_0_lowtown_a2_e18_0_0.vlm"},
	{0xECEC09B976CA6CAAULL, "mcr370_meredith6136182.cl"},
	{0xECEC12C1190E457EULL, "vxht_mansion2b_1w3_0_0.vlm"},
	{0xECEC3DA94BD9D9E1ULL, "vxw_encounter1_b2_0_0.vlm"},
	{0xECEC592B8ABCFECBULL, "mag320_nobleman_16060262.cl"},
	{0xECED120E59F4273EULL, "6209919_m.fxe"},
	{0xECED2C524414B2C0ULL, "6086002_m.fxe"},
	{0xECEDED85D26FF98BULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.impa.gad"},
	{0xECEDFD033764F84AULL, "hlac_lt_0_lowtown_a2_n_c_131071_3x.dds"},
	{0xECEE3EEF186B0C66ULL, "vxdr_primevaltomb_8f4_0_0.vlm"},
	{0xECEEC6927FC585B1ULL, "6126745_m.fxe"},
	{0xECEEFD9F1E6041D8ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm02.phy"},
	{0xECEF1648E09F0A6EULL, "hlac_lt_0_lowtown_a2_n_-262144_4x.dds"},
	{0xECEF2D03441CE4CEULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.dura.evt"},
	{0xECEF3A325EA186B8ULL, "art\\vfx\\environment\\vfx_pois_gas_barrelgyser.dura.ani"},
	{0xECEF91A1909FE5C5ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\effect\\bloodwoundexit.evt"},
	{0xECEFDC21F3A11A73ULL, "vxht_bartrand_n_4eq_0_0.vlm"},
	{0xECF057E03BBEC928ULL, "vxlt_0_lowtown_a3_n_c_bxo_0_0.vlm"},
	{0xECF0682FD05D95C8ULL, "mrl000_marethari6129351.cl"},
	{0xECF06A4085067919ULL, "lt_gallowstemplar_n_2389.rml"},
	{0xECF0A41857C7A9C1ULL, "94075835.ogg"},
	{0xECF0ADF5982DD7CEULL, "vxlt_0_lowtown_docks_c2k_0_0.vlm"},
	{0xECF0C68077FA3616ULL, "hlac_lt_0_lowtown_a2_n_c_-1_8z.dds"},
	{0xECF0ECB9C6F7E18FULL, "331508890.ogg"},
	{0xECF15EA85092A476ULL, "6162485_m.fxe"},
	{0xECF19F61A3AE3C1EULL, "942607419.ogg"},
	{0xECF235A603B1E402ULL, "6098083_m.fxe"},
	{0xECF2945C8A83525BULL, "dae210_gate_guard6085119.cl"},
	{0xECF2FD640B47D2D6ULL, "art\\levels\\races\\design\\prp_gasbarrel_0.pwk"},
	{0xECF374B9DB738BE4ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_vase.msh"},
	{0xECF386B9DB73AA4DULL, "art\\levels\\races\\elf\\props\\alienage\\prp_vase.mao"},
	{0xECF38AB9DB73B086ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_vase.mmh"},
	{0xECF3B019CDC58BB4ULL, "gen00fl_sebastian6189018.cl"},
	{0xECF41200F642EAFFULL, "vxdr_primevaltomb_39s_0_0.vlm"},
	{0xECF442567ABE886CULL, "lt_warehouse_water_1849.msh"},
	{0xECF448567ABE925EULL, "lt_warehouse_water_1849.mmh"},
	{0xECF454567ABEA6F5ULL, "lt_warehouse_water_1849.mao"},
	{0xECF4547F25F1D283ULL, "vxht_keep_3ij_0_0.vlm"},
	{0xECF4C937A82F6E8CULL, "vxca_caverns2_78t_0_1.vlm"},
	{0xECF4EA85D27613F0ULL, "art\\vfx\\combat\\vfx_spd_web_proj_p.impa.evt"},
	{0xECF4ED081A9A462AULL, "gen_im_wep_war_shd_act0_04.uti"},
	{0xECF5866C1F976CA3ULL, "101864180.ogg"},
	{0xECF5C65F0D2E85C8ULL, "vxw_encounter1_b7_0_0.vlm"},
	{0xECF6000C3FD5F7A9ULL, "hlac_w_encounter1_-65535_1z.dds"},
	{0xECF60A1BDAF4238EULL, "smoke_f.dds"},
	{0xECF624034422F479ULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.dura.gad"},
	{0xECF67AC4880197A5ULL, "dae000_gamlen6102522.cl"},
	{0xECF6A8233A95F249ULL, "551380946.ogg"},
	{0xECF6B16C213851A9ULL, "44572360.ogg"},
	{0xECF75016A96491B8ULL, "vxht_mansion2b_1rj_0_0.vlm"},
	{0xECF7588565101764ULL, "vxca_caverns1_mines_dr6_0_0.vlm"},
	{0xECF775B539218C94ULL, "vxca_cave_5n_0_0.vlm"},
	{0xECF7A5EE4443FDBAULL, "vrc321_party_barks6096302.cl"},
	{0xECF7C48DF97DE805ULL, "vxlt_warehouse_14o_0_0.vlm"},
	{0xECF7FA843ED9597FULL, "dae000_arishok6096350.cl"},
	{0xECF842A8F99AA8B2ULL, "art\\vfx\\combat\\material__2.mao"},
	{0xECF8856C341E6CA5ULL, "llac_ca_caverns2_5802_46.dds"},
	{0xECF88EA43FD2656FULL, "6149036_m.fxe"},
	{0xECF90465B9FCD2ADULL, "dre180ip_lyrium_idol.utp"},
	{0xECF963834C1E0390ULL, "face_face_p18.psh.fxo.dx11"},
	{0xECF981BEB2A6ACA3ULL, "w_encounter3_4_0.msh"},
	{0xECF98FBEB2A6C4FDULL, "w_encounter3_4_0.mmh"},
	{0xECF993BEB2A6CB0EULL, "w_encounter3_4_0.mao"},
	{0xECF9F072974DB05AULL, "941748582.ogg"},
	{0xECF9FA1D454682DBULL, "w_bonepitmining_a2_38_0.msh"},
	{0xECFA0C1D4546A126ULL, "w_bonepitmining_a2_38_0.mao"},
	{0xECFA181D4546B5C5ULL, "w_bonepitmining_a2_38_0.mmh"},
	{0xECFAAF863DCFF261ULL, "vxw_encounter1_g1_0_0.vlm"},
	{0xECFB7B12F588EABCULL, "501673138.ogg"},
	{0xECFBC5EE6B5C8B90ULL, "hlac_ht_0_hightown_a2_65538_0x.dds"},
	{0xECFC0BD2F920C49CULL, "6135177_m.fxe"},
	{0xECFC350344277CC7ULL, "art\\vfx\\combat\\vfx_c_pride_f_fist_c.dura.ani"},
	{0xECFCD66FD23BE6B6ULL, "vxw_bonepit_1wp_0_0.vlm"},
	{0xECFD81B9DB7BF651ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_vase.phy"},
	{0xECFDF887CC909DBCULL, "6040638_f.fxe"},
	{0xECFE0666D9C91433ULL, "vxht_keep_n_5xp_0_0.vlm"},
	{0xECFE323F6B4E4A21ULL, "151366205.ogg"},
	{0xECFE4DA163669DFFULL, "llac_ht_0_hightown_a3_0_1.dds"},
	{0xECFE923C847F4E14ULL, "6173180_m.fxe"},
	{0xECFF06BBAE770F3DULL, "vxlt_0_lowtown_docks_ctf_0_0.vlm"},
	{0xECFF0D1CFC2CCEE7ULL, "886529450.ogg"},
	{0xECFF8849DD63D5BEULL, "6119890_m.fxe"},
	{0xED004905E73C109FULL, "vxht_bartrand_n_33h_0_0.vlm"},
	{0xED0052CD727AA9EAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\blendtree\\c_dsogre_ovr.anb"},
	{0xED00C33B455453CEULL, "dae000_merchant_generic_hm6172665.cl"},
	{0xED00F819FDC70C23ULL, "llac_lt_0_lowtown_a2_n_65538_4.dds"},
	{0xED01C926427F804EULL, "vrc321ip_spooky5_obj_4.utp"},
	{0xED01DD2673734394ULL, "mag130cr_slaver_listener.utc"},
	{0xED01FD5D0047A306ULL, "346638312.ogg"},
	{0xED0210E65A226A22ULL, "hlac_ca_caverns1_24424_20y.dds"},
	{0xED0215C87EF24AB1ULL, "6196786_m.fxe"},
	{0xED02527B582552F5ULL, "306329949.ogg"},
	{0xED028025F29BCBE9ULL, "vxht_0_hightown_n_b09_0_0.vlm"},
	{0xED02C4EB2F6FC59CULL, "mag100_thrask6034870.cl"},
	{0xED02FCEA12F34D7DULL, "6167896_m.fxe"},
	{0xED02FD096D5BFB32ULL, "vxlt_0_lowtown_a2_fe2_0_0.vlm"},
	{0xED032BDE63E884FAULL, "vxlt_hangedman_214_0_0.vlm"},
	{0xED033FB94C480DC2ULL, "6204156_m.fxe"},
	{0xED033FBF2A0901BCULL, "zz_prmpt_debug6198228.cl"},
	{0xED03863195833C18ULL, "zz_vault_debug6200315.cl"},
	{0xED0407A7DCB8A043ULL, "vxlt_gallowstemplar_n_473_0_0.vlm"},
	{0xED04383BFF249E48ULL, "vxw_encounter1_g4_0_0.vlm"},
	{0xED043AAA65F758CDULL, "ss_hadriana.bnk"},
	{0xED044D7500C48D25ULL, "hlac_ca_caverns1_24421_5x.dds"},
	{0xED04B3CBBCE54272ULL, "llac_ca_caverns2prc_5802_33.dds"},
	{0xED058D390D88AB90ULL, "vxlt_0_lowtown_ewa_0_0.vlm"},
	{0xED059268F0D1E14BULL, "107482692.ogg"},
	{0xED05C1F39B5AAA38ULL, "vxht_mansion2b_1uk_0_0.vlm"},
	{0xED05FD35C2E34FF7ULL, "6046979_m.fxe"},
	{0xED07008131C87DEFULL, "6142136_m.fxe"},
	{0xED07546A3689DF4BULL, "vxdr_primevaltomb_4hn_0_1.vlm"},
	{0xED07CCF0350C9A98ULL, "fen310_fenris6118261.cl"},
	{0xED07DE593C6B4077ULL, "vxlt_house_el_0_0.vlm"},
	{0xED083CAD33B5C2EEULL, "159945009.ogg"},
	{0xED08FE27351BA0CAULL, "20960894.ogg"},
	{0xED099B432CD397E4ULL, "577319967.ogg"},
	{0xED0A0C1F951D0BEFULL, "550508068.ogg"},
	{0xED0A3C31CEEEF6F9ULL, "dae000_keep_servant_1_sta6156866.cl"},
	{0xED0A3D1571964748ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef04.msh"},
	{0xED0A4FF3AF6F6D99ULL, "6143498_m.fxe"},
	{0xED0A5B1571967A72ULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef04.mmh"},
	{0xED0A5B4BD3A56FD1ULL, "816951801.ogg"},
	{0xED0ABC3D3469D53BULL, "hlac_dr_dungeon_3216_10x.dds"},
	{0xED0AD3024C26BE8CULL, "tv_wis_sm_dark.mao"},
	{0xED0B3369EDD75B8AULL, "338052829.ogg"},
	{0xED0B74B5795F35C0ULL, "447720440.ogg"},
	{0xED0BBDC51261A4B4ULL, "6151889_m.fxe"},
	{0xED0BC1B44E451A6EULL, "mag291_vincento6023834.cl"},
	{0xED0C1FD89E1C8BF0ULL, "dae000_friend_turn_in6179301.cl"},
	{0xED0C2B9354284ADAULL, "art\\characters\\heraldry\\hld_amell.dds"},
	{0xED0C2E3CEA85F916ULL, "6150545_m.fxe"},
	{0xED0C3EB182D7F07DULL, "vxht_0_hightown_a3_cf4_0_0.vlm"},
	{0xED0C45CE4BA08512ULL, "vxca_caverns1_fly_0_0.vlm"},
	{0xED0C5D43A802CDC6ULL, "529296122.ogg"},
	{0xED0C6DCF43512CBAULL, "al_sebastian_s01.gda"},
	{0xED0C72AB1820F9C6ULL, "6151546_m.fxe"},
	{0xED0C75B66FF4B18EULL, "127658045.ogg"},
	{0xED0CD6317AE878FEULL, "lgt320_shw_ors_secret_thug6156369.cl"},
	{0xED0D4394FC320D7AULL, "vxca_caverns1_mines_fmw_0_0.vlm"},
	{0xED0D93124F9D4CABULL, "6108528_f.fxe"},
	{0xED0DBAADB7901589ULL, "vxlt_shop_d1_0_0.vlm"},
	{0xED0DC7FD32806F70ULL, "lt_backalley_n_2_0_ll.mmh"},
	{0xED0E4EF8A8C133A7ULL, "follower_banter6148726.cl"},
	{0xED0EC9B08B7B28E7ULL, "dae350ar_sundermount.ncs"},
	{0xED101B3FB82AB877ULL, "6042852_m.fxe"},
	{0xED1049F8831827ECULL, "gen00fl_merrill6106452.cl"},
	{0xED1054DED9863EE5ULL, "plt_cod_prc_deathwatchbeetl.plo"},
	{0xED10C62C2CE65E65ULL, "dae000cr_grand_cleric6156770.cl"},
	{0xED1111AFDEC584C5ULL, "vxht_mansion_n_5cw_0_0.vlm"},
	{0xED116D5B2B73C4C0ULL, "48235093.ogg"},
	{0xED11A33BAFD447D9ULL, "gen00fl_aveline6139845.cl"},
	{0xED1212F8DFB66A0EULL, "vxca_cavern1_haunted_egj_0_0.vlm"},
	{0xED12DDB615B8DF11ULL, "430667034.ogg"},
	{0xED1377374A076E79ULL, "mag101_alain6032201.cl"},
	{0xED13802B5F0A26B8ULL, "6155874_m.fxe"},
	{0xED140BD658295003ULL, "vxdr_primevaltomb_55y_0_0.vlm"},
	{0xED14A34816446886ULL, "hlac_lt_0_lowtown_a3_n_c_65535_3z.dds"},
	{0xED15A9B08B811096ULL, "dae350ar_sundermount.lst"},
	{0xED15F443658C6B80ULL, "gen00fl_isabela6099532.cl"},
	{0xED163FADFC89D086ULL, "lt_0_lowtown_docks_a3_n_16_0.rml"},
	{0xED16C9DE1DFA6285ULL, "6079580_m.fxe"},
	{0xED16F5736D5810F9ULL, "6137282_m.fxe"},
	{0xED173363CB4DFC6EULL, "281221279.ogg"},
	{0xED173DECA96A4FD9ULL, "6129800_f.fxe"},
	{0xED185AF8F368870FULL, "559005142.ogg"},
	{0xED18F7C80D9098D6ULL, "vxht_0_hightown_c1w_0_0.vlm"},
	{0xED18FD01E3B11ADFULL, "6083280_m.fxe"},
	{0xED19508C8ED2FB01ULL, "6110609_m.fxe"},
	{0xED19512B893B190AULL, "6107053_m.fxe"},
	{0xED19575B60F49593ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_21basic.phy"},
	{0xED1997EA4A15FD65ULL, "6152588_m.fxe"},
	{0xED1A3904D126D295ULL, "vxca_cavern1_haunted_jfq_0_0.vlm"},
	{0xED1A395BA017A80FULL, "170930708.ogg"},
	{0xED1A4EF37D70FF3DULL, "6208353_m.fxe"},
	{0xED1A859E0750DB72ULL, "llac_lt_0_lowtown_docks_n_-131073_3.dds"},
	{0xED1B0426B0212CDAULL, "w_encounter3_188_0.msh"},
	{0xED1B1226B0214413ULL, "w_encounter3_188_0.mao"},
	{0xED1B1626B0214B20ULL, "w_encounter3_188_0.mmh"},
	{0xED1B95783A40DEB7ULL, "vxdr_dungeon_2hf_0_1.vlm"},
	{0xED1BE19073CA34FAULL, "ach_exp_gallows.dds"},
	{0xED1BF1CD36BA269AULL, "6043390_f.fxe"},
	{0xED1C328F9A71A98AULL, "1008544066.ogg"},
	{0xED1CEFFAA2B895A1ULL, "vxca_caverns1_ec8_0_0.vlm"},
	{0xED1CF73A4567E672ULL, "vxlt_backalley_n_1f8_0_0.vlm"},
	{0xED1D183D20DF9467ULL, "39401745.ogg"},
	{0xED1D8E8EADBC0CADULL, "vxlt_0_lowtown_a2_gnu_0_0.vlm"},
	{0xED1DC932F4F9A69EULL, "llac_lt_0_lowtown_a2_131070_3.dds"},
	{0xED1E2D9B04538FBDULL, "487726858.ogg"},
	{0xED1E5FAF519889A1ULL, "1036243769.ogg"},
	{0xED1ED473507B361AULL, "vxht_0_hightown_n_ak0_0_0.vlm"},
	{0xED1EDDBC8BB9D407ULL, "6054244_m.fxe"},
	{0xED1FB8B08B897E81ULL, "dae350ar_sundermount.are"},
	{0xED200DAA3A3CC6C3ULL, "vxht_0_hightown_n_by4_0_0.vlm"},
	{0xED203FE929C8F388ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_master_n.dds"},
	{0xED20E67793776493ULL, "follower_banter6136723.cl"},
	{0xED20EAF682D8658AULL, "6040666_m.fxe"},
	{0xED2131EB8F2AE165ULL, "vxw_bonepit_1me_0_0.vlm"},
	{0xED2175E268569FCAULL, "dae000cr_grand_cleric6124792.cl"},
	{0xED21CA800D5B2472ULL, "character_character_p41.psh.fxo.dx11"},
	{0xED21F8FA31A8DD55ULL, "vxw_encounter3_fa_0_1.vlm"},
	{0xED21FC3D280A37ABULL, "6088322_m.fxe"},
	{0xED2270F2F174DC90ULL, "vxlt_0_lowtown_emo_0_0.vlm"},
	{0xED22940317EA7BB3ULL, "vxlt_0_lowtown_docks_c4x_0_0.vlm"},
	{0xED230C992D054694ULL, "6111132_f.fxe"},
	{0xED235833E9B5DCC3ULL, "llac_w_encounter2_camp_-2_0.dds"},
	{0xED235F51F4F8CA60ULL, "322004731.ogg"},
	{0xED23A6A675F81561ULL, "mag211_marethari6025803.cl"},
	{0xED24B74F936E0291ULL, "6116150_f.fxe"},
	{0xED24D80CB671D0F9ULL, "774517945.ogg"},
	{0xED2593B64ED7C95CULL, "6178350_m.fxe"},
	{0xED25A1DF7E137A09ULL, "6172748_m.fxe"},
	{0xED26060B02FA220AULL, "mer141st_reveal.stg"},
	{0xED265BB8CE4D7456ULL, "262325985.ogg"},
	{0xED26A14E1D76C42FULL, "and100_anders6095887.cl"},
	{0xED26B25F8B28FDB3ULL, "gen00fl_merrill6107458.cl"},
	{0xED26B2C8121D402EULL, "806748104.ogg"},
	{0xED2739CABAA925B9ULL, "157770662.ogg"},
	{0xED2769A4FF86B156ULL, "vxht_0_hightown_c4x_0_0.vlm"},
	{0xED2793DBC47CC889ULL, "737763525.ogg"},
	{0xED27ED172A5CC181ULL, "prcscr_seb.gda"},
	{0xED280662E0436E66ULL, "vxdr_primevaltomb_8r8_0_0.vlm"},
	{0xED2826BCE08BCAE3ULL, "488657074.ogg"},
	{0xED2839A0D710E680ULL, "llac_lt_0_lowtown_docks_n_65535_5.dds"},
	{0xED283C15B00F031EULL, "vxht_0_hightown_a2_cak_0_0.vlm"},
	{0xED299956A344CB48ULL, "lt_0_lowtown_docks_a3_197_0.mmh"},
	{0xED29A556A344DFEBULL, "lt_0_lowtown_docks_a3_197_0.mao"},
	{0xED29B756A344FE72ULL, "lt_0_lowtown_docks_a3_197_0.msh"},
	{0xED29E5419E799660ULL, "llac_lt_warehouse_65536_2.dds"},
	{0xED2A58038A066E90ULL, "vxca_cave_1ac_0_0.vlm"},
	{0xED2A9811D73F127FULL, "6176027_m.fxe"},
	{0xED2ABDBDB21BB36EULL, "613081933.ogg"},
	{0xED2ACA0A243026B6ULL, "6131399_m.fxe"},
	{0xED2B7277ED65B617ULL, "vxht_0_hightown_a3_ajx_0_0.vlm"},
	{0xED2B847DB6D7B2C0ULL, "fen200pt_bitterpill.ncs"},
	{0xED2BC9FF998E83A5ULL, "vxlt_gallowstemplar_n_3re_0_0.vlm"},
	{0xED2BDEAAA3166EB1ULL, "6140172_m.fxe"},
	{0xED2BF1004FD5A9B9ULL, "vxca_cavern1_haunted_cp2_0_0.vlm"},
	{0xED2C625D732635F1ULL, "989522248.ogg"},
	{0xED2CBC6F4134C2F7ULL, "mag121_eau_banter6107380.cl"},
	{0xED2CC79C7E189DA8ULL, "vxdr_primevaltomb_7y4_0_0.vlm"},
	{0xED2CD4BF5CBBD7E8ULL, "569874470.ogg"},
	{0xED2DC5933EA12BE3ULL, "6161027_m.fxe"},
	{0xED2E44B9E6FED0D0ULL, "6156404_m.fxe"},
	{0xED2E49EBE82F511FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_fast.gad"},
	{0xED2F1737BC8F691EULL, "vxca_cavern1_haunted_cmr_0_0.vlm"},
	{0xED2FC510908B388DULL, "17810639.ogg"},
	{0xED2FCD4BEB6E6057ULL, "235311378.ogg"},
	{0xED2FE405BCC79DA1ULL, "6043673_m.fxe"},
	{0xED30E2CFE36AF510ULL, "vxlt_0_lowtown_ehl_0_0.vlm"},
	{0xED3157A1EB708425ULL, "dae000_fenarel6128123.cl"},
	{0xED31818B51218236ULL, "6121325_m.fxe"},
	{0xED3188D36F0CA7B6ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.dura.evt"},
	{0xED318B55643A97F6ULL, "vxca_caverns2_9xq_0_0.vlm"},
	{0xED325FA5C19B521DULL, "137235286.ogg"},
	{0xED32799146BF167AULL, "185703859.ogg"},
	{0xED3339433C516C94ULL, "832909537.ogg"},
	{0xED335103EA2C1396ULL, "6080635_m.fxe"},
	{0xED335CCE9B0FD6D8ULL, "6125738_m.fxe"},
	{0xED338DC5F3AF33E2ULL, "vxht_keep_n_6aa_0_0.vlm"},
	{0xED33E65DE75EFDF9ULL, "vxlt_undercityencounter_12g_0_1.vlm"},
	{0xED33EF58A195EB62ULL, "dae100_viscount6118522.cl"},
	{0xED3531DA84375C62ULL, "w_bonepit_156_0.msh"},
	{0xED3533DA84375F18ULL, "w_bonepit_156_0.mmh"},
	{0xED353EEBE8355D0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_fast.evt"},
	{0xED353FDA843773BBULL, "w_bonepit_156_0.mao"},
	{0xED35A234200A5057ULL, "6097227_m.fxe"},
	{0xED35D8473D26C46EULL, "6174455_m.fxe"},
	{0xED3627C2ADF88583ULL, "vxw_bonepit_yt_0_0.vlm"},
	{0xED367B6700DC52BDULL, "llac_dr_dungeon_3207_5.dds"},
	{0xED3682FB6D317A19ULL, "6114476_m.fxe"},
	{0xED370284ABE4F701ULL, "457509805.ogg"},
	{0xED37508092EA3BDDULL, "character_character_p6.psh.fxo.dx11.assembly"},
	{0xED3777F692CD3C23ULL, "59176283.ogg"},
	{0xED379863A2589687ULL, "cd_ser221_gascard.wbd"},
	{0xED37DD92CD78815EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_df_transition.evt"},
	{0xED37F04B38545ACCULL, "164491957.ogg"},
	{0xED3800A47BEF18CEULL, "6041258_m.fxe"},
	{0xED3869879F5DFF92ULL, "6056603_m.fxe"},
	{0xED387FD36F12B741ULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.dura.gad"},
	{0xED38832F4E100A8EULL, "hlac_lt_undercitydungeon_-1_43z.dds"},
	{0xED39236168FFE7DAULL, "hlac_w_encounter2_4_0x.dds"},
	{0xED3962E776AB90BEULL, "479280602.ogg"},
	{0xED39B8C8F1607BE3ULL, "weapon_weapon_p2.psh.fxo.dx11.assembly"},
	{0xED39CE8C4AA76CFFULL, "follower_banter6173041.cl"},
	{0xED39DB70DDB8AC5CULL, "6090452_m.fxe"},
	{0xED39FFCA44211000ULL, "409066298.ogg"},
	{0xED3AE937CAFE8303ULL, "dae000_katriela6162664.cl"},
	{0xED3B181D457E4CCEULL, "w_bonepitmining_a2_38_0.rml"},
	{0xED3B3C1717A8CA42ULL, "vxlt_0_lowtown_eqq_0_0.vlm"},
	{0xED3B6F73E7EA7C4BULL, "sib131cr_slave_c.utc"},
	{0xED3C57ED0A571F57ULL, "526559551.ogg"},
	{0xED3C70102784E08FULL, "ss_deeproads.cut"},
	{0xED3DBCF7AEA7C926ULL, "mag210_arianni6024979.cl"},
	{0xED3DCA890C128234ULL, "6142067_m.fxe"},
	{0xED3E0666B96E85F1ULL, "zz_followers_banter6158712.cl"},
	{0xED3E139E5F045CD7ULL, "character_character_p152.psh.fxo.dx11"},
	{0xED3E9548EA57A105ULL, "t3_wdg_tb1.tnt"},
	{0xED3E9F620F406D48ULL, "mrl150_merrill6072370.cl"},
	{0xED3EF492CD7EC689ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_df_transition.gad"},
	{0xED3EFB55773BAD26ULL, "6111234_m.fxe"},
	{0xED3F2350D16D65D6ULL, "lowtown_male_lowlife2_mob6166757.cl"},
	{0xED3F28F7A1EE716AULL, "vxw_bonepit_73_0_0.vlm"},
	{0xED3F3ADA843FBFA3ULL, "w_bonepit_156_0.phy"},
	{0xED3F80D36F18D7EFULL, "art\\vfx\\combat\\vfx_str_scream_tar_lp_c.dura.ani"},
	{0xED3FB60CC249422CULL, "dae210ip_store_tailor.utp"},
	{0xED3FD85BAD0EA1C2ULL, "555359011.ogg"},
	{0xED3FD8C288FEB0C0ULL, "6167220_m.fxe"},
	{0xED3FF3DCA9D4E3F6ULL, "vxht_0_hightown_apv_0_0.vlm"},
	{0xED3FFBF06B28E402ULL, "6065388_m.fxe"},
	{0xED400A03B6BE5A0AULL, "6102941_f.fxe"},
	{0xED40BB78E8ACFCD7ULL, "6020715_m.fxe"},
	{0xED40D81D2C97F79AULL, "6098201_m.fxe"},
	{0xED40D860C12A642EULL, "930267137.ogg"},
	{0xED411393A52B4073ULL, "fx_fadedust.mao"},
	{0xED4114F6D9BD1828ULL, "231411141.ogg"},
	{0xED412C0189C18810ULL, "6109123_m.fxe"},
	{0xED41F47CC9A8D80CULL, "6145219_m.fxe"},
	{0xED41FFE4B8750228ULL, "6147132_m.fxe"},
	{0xED420E29609A6AA8ULL, "6187785_m.fxe"},
	{0xED425249403E184BULL, "w_bonepitdestroyed_a3_10_0_lc.mmh"},
	{0xED4257EA65DBEFCCULL, "vxca_caverns2_7ng_0_0.vlm"},
	{0xED426BE2C8395652ULL, "vxdr_primevaltomb_6py_0_1.vlm"},
	{0xED4326EBE8417265ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_enter_fast.ani"},
	{0xED432EF744649F97ULL, "llac_ht_0_hightown_a2_-131073_0.dds"},
	{0xED433905EA9C1B4BULL, "hlac_w_sundermount_-393213_0x.dds"},
	{0xED44D364CEF2C64CULL, "dre000_varric6207974.cl"},
	{0xED45B1823F9A51B9ULL, "character_character_p22.psh.fxo.dx11"},
	{0xED45C592CD8495D7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_df_transition.ani"},
	{0xED46572CAC6FEB4AULL, "6150866_m.fxe"},
	{0xED46B1F0DEE4D9C2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_table.msh"},
	{0xED46B3F0DEE4DD78ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_table.mmh"},
	{0xED46BFF0DEE4F11BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_table.mao"},
	{0xED475705C35A01EEULL, "llac_lt_undercityencounter_-1_0.dds"},
	{0xED47D9A3BA43D96CULL, "6202285_f.fxe"},
	{0xED47F210F1866163ULL, "6156542_m.fxe"},
	{0xED480F839E8C1D3AULL, "vxht_0_hightown_n_c43_0_0.vlm"},
	{0xED48D4BA33C1DAB9ULL, "vxca_cavern1_haunted_cn0_0_0.vlm"},
	{0xED4940BE4826E311ULL, "6016309_m.fxe"},
	{0xED494FE62527D614ULL, "6018621_m.fxe"},
	{0xED49B50B8E1559D7ULL, "38826956.ogg"},
	{0xED49BEEC8E80D17DULL, "560278986.ogg"},
	{0xED4A719312321A40ULL, "vxw_bonepit_176_0_0.vlm"},
	{0xED4AB64E5F40EFC4ULL, "vxca_caverns1_mines_edy_0_0.vlm"},
	{0xED4AEA9F1EAE806DULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm02.msh"},
	{0xED4AFC9F1EAE9EF3ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm02.mmh"},
	{0xED4B05350CD54ECAULL, "moveaway.ncs"},
	{0xED4B06C8E2172B72ULL, "vxw_blightlands_11c_0_0.vlm"},
	{0xED4B1F069A878FA9ULL, "follower_barks6164417.cl"},
	{0xED4BDCCD5B622260ULL, "51960566.ogg"},
	{0xED4C014BBA8976C9ULL, "sse_ss_danzig.gda"},
	{0xED4C9D0E4014F1DBULL, "mag100_karras6035291.cl"},
	{0xED4CC39546B64F3BULL, "qcr000_arishok6109249.cl"},
	{0xED4CDB4A64ABB464ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_wep_cb.mao"},
	{0xED4DBF6B3B821566ULL, "6115609_m.fxe"},
	{0xED4DE6EC1C2363C3ULL, "681322702.ogg"},
	{0xED4DE80C4C50989FULL, "6016574_m.fxe"},
	{0xED4E6F0F4826C8F6ULL, "vxca_caverns2_9fs_0_0.vlm"},
	{0xED4E824A8C8A9CD4ULL, "dae011_corff_sta6152752.cl"},
	{0xED4EDA3D96BDF24DULL, "6127544_m.fxe"},
	{0xED4F14F8D85CEE14ULL, "6052911_m.fxe"},
	{0xED4F57DD5FF27727ULL, "526201788.ogg"},
	{0xED4F8DDB6792117BULL, "292443940.ogg"},
	{0xED4FC1C709FE2573ULL, "6041138_m.fxe"},
	{0xED500F9868463BD7ULL, "vxca_cavern1_haunted_giw_0_0.vlm"},
	{0xED505160E3F6823CULL, "hlac_lt_0_lowtown_n_131070_4x.dds"},
	{0xED5071C1AA6B1AA8ULL, "6140032_m.fxe"},
	{0xED50BAF0DEED3D03ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_table.phy"},
	{0xED50F90C3F013F7AULL, "dae000_arishok6067797.cl"},
	{0xED51E3FCAD843946ULL, "vxlt_0_lowtown_docks_a3_n_9f5_0_0.vlm"},
	{0xED523E07D7DBA034ULL, "vxw_bonepitmining_a2_1od_0_0.vlm"},
	{0xED5288A87E6AE762ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masf.phy"},
	{0xED528D07FBAD88DAULL, "zz_warriorleap.cut"},
	{0xED5349D3C7D33CC7ULL, "vxlt_gallowstemplar_n_482_0_0.vlm"},
	{0xED53DC75E3863C26ULL, "6079338_m.fxe"},
	{0xED53FCE0B4FC33C9ULL, "hlac_ht_mansion2b_818_35y.dds"},
	{0xED5432884E9278EFULL, "954359015.ogg"},
	{0xED549074FA29E98FULL, "6075854_m.fxe"},
	{0xED54F9475C6B5A9BULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basketbunch.mmh"},
	{0xED5505475C6B6F38ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basketbunch.mao"},
	{0xED5517475C6B8D85ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basketbunch.msh"},
	{0xED55ECF33249AD7AULL, "6075284_m.fxe"},
	{0xED56D52D6015D134ULL, "vxlt_gamlenshouse_n_c5_0_0.vlm"},
	{0xED572D7D8FAB8102ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v2.gad"},
	{0xED57AF365C39FCDFULL, "mcr370_cut_victorious.bnk"},
	{0xED58113983E8C38EULL, "gen_im_trs_rdr_med_boo_r2.uti"},
	{0xED586B18C68F2DABULL, "6050370_m.fxe"},
	{0xED586D8229E25DD9ULL, "6126572_m.fxe"},
	{0xED586E4253AC8CAFULL, "238024190.ogg"},
	{0xED58850BCC7E1FF7ULL, "6013371_m.fxe"},
	{0xED589577517D9EB9ULL, "ser210st_witness.stg"},
	{0xED593C9F78952AF8ULL, "hlac_ht_chantry_n_2534_13z.dds"},
	{0xED5978A5D40D43F2ULL, "vxw_blightlands_14l_0_0.vlm"},
	{0xED5A47BAF04C9D83ULL, "vxlt_0_lowtown_n_f1p_0_0.vlm"},
	{0xED5A91365C3BFFE7ULL, "mcr370_cut_victorious.cnv"},
	{0xED5BDC78FF3C71B7ULL, "vxlt_0_lowtown_docks_a3_aop_0_0.vlm"},
	{0xED5C6E5C33BD4FDBULL, "407194222.ogg"},
	{0xED5C8C0BC9C18DCEULL, "hlac_lt_0_lowtown_docks_196608_0x.dds"},
	{0xED5CE19B5A107E13ULL, "llac_w_sundermount_-131071_0.dds"},
	{0xED5D49C3A0E47422ULL, "142202632.ogg"},
	{0xED5D4C1ECB1E316EULL, "883433308.ogg"},
	{0xED5D90512AA7F29BULL, "hlac_dr_dungeon_3214_4y.dds"},
	{0xED5DDBCA3A5BDE7DULL, "darkring01.mao"},
	{0xED5E3ADA36ED8683ULL, "gen00fl_sebastian6180382.cl"},
	{0xED5E507D8FB1DBF9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v2.evt"},
	{0xED5E56550C474ACFULL, "atl_shared_dxt5_dat.xml"},
	{0xED5E6CD6AABF1BB0ULL, "follower_banter6172713.cl"},
	{0xED5F049848F72256ULL, "6016679_f.fxe"},
	{0xED5F1F8C494B875BULL, "art\\levels\\ztools\\default\\generic\\0ff_window.mmh"},
	{0xED5F2B8C494B9BF8ULL, "art\\levels\\ztools\\default\\generic\\0ff_window.mao"},
	{0xED5F3D8C494BBA45ULL, "art\\levels\\ztools\\default\\generic\\0ff_window.msh"},
	{0xED5F49A86286D20FULL, "ts_circleofmagi_wpn_props.gda"},
	{0xED5F51C65BCA0DDFULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_bdy_hdra_0n.dds"},
	{0xED60A2FAFF366363ULL, "6087572_m.fxe"},
	{0xED60D0D23EA7DBA9ULL, "6131129_m.fxe"},
	{0xED61152709ADD5C1ULL, "hlac_w_encounter1_131071_0y.dds"},
	{0xED619F85912EEBCFULL, "vxw_encounter2_3f_0_0.vlm"},
	{0xED62739FB35520C3ULL, "6075224_f.fxe"},
	{0xED62C665BCAC4B85ULL, "space_si.dds"},
	{0xED62D7FCAA41855DULL, "vxlt_darktown_6hu_0_0.vlm"},
	{0xED62F975A434507CULL, "vxht_0_hightown_n_ccp_0_0.vlm"},
	{0xED636D175C18F81DULL, "pro000_colonel_ewald6039959.cl"},
	{0xED639968DD6F5178ULL, "623503481.ogg"},
	{0xED63B83A6BA51226ULL, "hlac_lt_gallowscourtyard_n_-65538_3x.dds"},
	{0xED63D2DBB347AD8DULL, "hlac_dr_dungeon_3216_4y.dds"},
	{0xED64244A0E3C2D4BULL, "isa300ip_document_chest.utp"},
	{0xED64B94B05698681ULL, "6132813_f.fxe"},
	{0xED6502E4A19EB027ULL, "vxht_mansion2_2ag_0_0.vlm"},
	{0xED655A1571E3246DULL, "art\\characters\\playercharacter\\elffemale\\mt_art_ef04.phy"},
	{0xED663145A80C7E5AULL, "6095026_m.fxe"},
	{0xED6635EDC11DDFB5ULL, "dae000cr_adriano.utc"},
	{0xED664821F448BD20ULL, "6083113_m.fxe"},
	{0xED664C6933B0A571ULL, "6017805_m.fxe"},
	{0xED66B698DE41F5C8ULL, "36238866.ogg"},
	{0xED66EEF9F7E07E64ULL, "vfx_darkring01.dds"},
	{0xED66F0C390965402ULL, "6093526_f.fxe"},
	{0xED67C04B7A375F27ULL, "6118091_f.fxe"},
	{0xED67F417E7626769ULL, "vxca_cave_1ch_0_0.vlm"},
	{0xED68B997E242B603ULL, "vxlt_0_lowtown_n_f4s_0_0.vlm"},
	{0xED68EA655224AC1AULL, "dae000_hate_follower_react6103194.cl"},
	{0xED6907507EE24844ULL, "vxlt_hangedman_2hq_0_0.vlm"},
	{0xED697EBE66F2FDEDULL, "dae000ip_crowd_mf1_talk4.utp"},
	{0xED699DB1845F5AC7ULL, "llac_ht_0_hightown_n_65533_3.dds"},
	{0xED69CA155B2E72C1ULL, "zz_qun_debug6028248.cl"},
	{0xED6A39194D934135ULL, "zz_isa_debug6133983.cl"},
	{0xED6A4760CCEB281DULL, "vxlt_foundry_n_wx_0_0.vlm"},
	{0xED6A47D4B584610BULL, "vxca_cavern1_haunted_hu8_0_0.vlm"},
	{0xED6B940D4D073006ULL, "6174785_m.fxe"},
	{0xED6BA7C1AD5F005CULL, "416481204.ogg"},
	{0xED6BA864AEC63FFBULL, "6152322_m.fxe"},
	{0xED6BD51CBBF85DA3ULL, "vxw_bonepitmining_a2_o0_0_0.vlm"},
	{0xED6C3CFE625195FDULL, "568270421.ogg"},
	{0xED6C587D8FBE26E8ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\h_atk_v2.ani"},
	{0xED6CA4802F8523C7ULL, "vxht_mansion_n_57p_0_1.vlm"},
	{0xED6CE0A4A563D3B1ULL, "932615708.ogg"},
	{0xED6D0743EBB9CC61ULL, "60754504.ogg"},
	{0xED6D6CA87303CC46ULL, "llac_ht_bartrand_n_3369_0.dds"},
	{0xED6DD9D4F5BFC8E2ULL, "llac_w_sundermount_327685_0.dds"},
	{0xED6E2F4E0371BFC6ULL, "mrl351_phantomelves6170340.cl"},
	{0xED6E37E3AEE3445DULL, "505879489.ogg"},
	{0xED6E8ECAB1F50A9DULL, "narrative_varric6172511.cl"},
	{0xED6EFD90FAFC3118ULL, "vxht_mansion2b_34n_0_0.vlm"},
	{0xED6EFE2F31A7F733ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeenter.evt"},
	{0xED6F32C87429E843ULL, "361614989.ogg"},
	{0xED6F481318B04401ULL, "dae000_merchant_access_hf6162959.cl"},
	{0xED6FB3F37C482284ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.anb"},
	{0xED7025D787445889ULL, "mag360_samson6054688.cl"},
	{0xED70F4AE74E44A90ULL, "676781410.ogg"},
	{0xED7144B7900BDAD2ULL, "llac_lt_gallowscourtyard_65536_1.dds"},
	{0xED71587918089ACBULL, "hlac_ht_mansion_n_2995_4z.dds"},
	{0xED7194D572D104E7ULL, "hlac_w_bonepitdestroyed_a3_0_7z.dds"},
	{0xED7197D2D2D88D99ULL, "vxlt_0_lowtown_a2_hfj_0_0.vlm"},
	{0xED71CE3C4FECDA19ULL, "dae000tr_base_.utt"},
	{0xED7310EC6E05D911ULL, "vxlt_0_lowtown_a2_n_c_hp9_0_0.vlm"},
	{0xED7357FAF2D44884ULL, "6164984_m.fxe"},
	{0xED7374C19394C8A7ULL, "vxht_mansion2_2jf_0_0.vlm"},
	{0xED7464C40AD784EFULL, "dae000_meredith6154279.cl"},
	{0xED75137A43113222ULL, "fen110_fenris6100259.cl"},
	{0xED75E75602FB969EULL, "and111_lirene6149840.cl"},
	{0xED75F72F31AE0A74ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeenter.gad"},
	{0xED765F584C8BEA5FULL, "6034813_m.fxe"},
	{0xED76FC7E1D22E9F2ULL, "vxlt_0_lowtown_docks_a3_azr_0_0.vlm"},
	{0xED76FEAC1A7A7DA4ULL, "lt_0_lowtown_a2_43_0_lc.mmh"},
	{0xED776FA87E89D8B5ULL, "art\\characters\\playercharacter\\humanfemale\\arm_masf.mmh"},
	{0xED7771A87E89DC2BULL, "art\\characters\\playercharacter\\humanfemale\\arm_masf.msh"},
	{0xED77869C2D8703A7ULL, "353781396.ogg"},
	{0xED77DCA747B7EBFCULL, "vxlt_gallowstemplar_451_0_0.vlm"},
	{0xED788CD3FFB36440ULL, "vxlt_backalley_g_1nw_0_0.vlm"},
	{0xED78A263A2903EBDULL, "cd_ser221_gascard.bnk"},
	{0xED78D8282CC04C39ULL, "6179503_m.fxe"},
	{0xED79147A1A26FB26ULL, "vxlt_0_lowtown_docks_n_b4w_0_0.vlm"},
	{0xED79448B58B2140EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured.ani"},
	{0xED799CDFB919EFAFULL, "hlac_lt_hangedman_n_-1_27x.dds"},
	{0xED7A607A87F27FC4ULL, "vxlt_0_lowtown_docks_9rc_0_0.vlm"},
	{0xED7AB541761302AFULL, "775780531.ogg"},
	{0xED7ADAB48695C206ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_hugekeg_d.dds"},
	{0xED7AE5838460D23FULL, "hlac_ht_0_hightown_a3_-131070_0z.dds"},
	{0xED7B505B61475408ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_21basic.mmh"},
	{0xED7B5C5B614768ABULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_21basic.mao"},
	{0xED7B6E5B61478732ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_21basic.msh"},
	{0xED7C7042985701FBULL, "llac_w_sundermount_-3_0.dds"},
	{0xED7CA3736653F099ULL, "251642052.ogg"},
	{0xED7CDD45E263E0A2ULL, "dae000_keep_maid_sta6162404.cl"},
	{0xED7CE9D94187151AULL, "zz_otl_debug6060666.cl"},
	{0xED7D062F31B44252ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeenter.ani"},
	{0xED7D13FA4C637FA7ULL, "328780326.ogg"},
	{0xED7DD2F37C549487ULL, "art\\vfx\\combat\\vfx_m_ent_sleep_aoe_p.mmh"},
	{0xED7DD31CD4673140ULL, "fx_smoke_vol_blend.mao"},
	{0xED7DFD6F55AD3708ULL, "820847069.ogg"},
	{0xED7E1E928A89239FULL, "mag170_thrask6020541.cl"},
	{0xED7EA52CB2114949ULL, "hlac_lt_0_lowtown_a2_n_196606_0y.dds"},
	{0xED7EBCA3EC38E0C2ULL, "6042299_m.fxe"},
	{0xED7EBFFDF1A2A7AAULL, "6125317_f.fxe"},
	{0xED7ECAFD48BA483EULL, "hlac_lt_undercityencounter_-1_9y.dds"},
	{0xED7F607D4D8FC631ULL, "78184597.ogg"},
	{0xED7FE3572E35C023ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_humanfemale.mao"},
	{0xED7FE7572E35C770ULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_humanfemale.mmh"},
	{0xED7FF5572E35DE8AULL, "art\\characters\\creatures\\crowds\\humanfemale\\crd_humanfemale.msh"},
	{0xED80117D4ABC7398ULL, "vxlt_undercityserialkiller_1fj_0_1.vlm"},
	{0xED805B57076983D1ULL, "hlac_w_bonepit_196611_0y.dds"},
	{0xED8099B6748E45B9ULL, "vxw_encounter2_rr_0_1.vlm"},
	{0xED80FA45650376EAULL, "6033732_m.fxe"},
	{0xED8113FF9285DADBULL, "6079336_f.fxe"},
	{0xED8153473591191BULL, "6063857_m.fxe"},
	{0xED815BC7124491D4ULL, "253869083.ogg"},
	{0xED81A3DE1D9B04CEULL, "dae000ip_crowd_mf4_talk2.utp"},
	{0xED81F32655B14820ULL, "vxht_brothel_n_588_0_0.vlm"},
	{0xED8256B3EE5ED907ULL, "hlac_lt_0_lowtown_n_-262144_3z.dds"},
	{0xED83B0CF0D8F1B96ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.cess.ani"},
	{0xED846E96C26612FDULL, "hlac_ht_mansion2_818_15y.dds"},
	{0xED854F592322CE52ULL, "868513379.ogg"},
	{0xED85A69779FBB427ULL, "ch2notnt_s1y.psh.fxo"},
	{0xED85F96DE1040545ULL, "vxw_encounter2_camp_pd_0_1.vlm"},
	{0xED863C8B58BC915FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured.evt"},
	{0xED86466EA710C4A6ULL, "1044416813.ogg"},
	{0xED864E74E31A5E1BULL, "gen00fl_aveline6139388.cl"},
	{0xED864EF62CE58CE5ULL, "vxw_bonepit_10n_0_0.vlm"},
	{0xED865FD2EC88D36BULL, "one000_ilen6205636.cl"},
	{0xED86C4F203B3CF41ULL, "ser221_gamlen6131961.cl"},
	{0xED86E504FDB86C72ULL, "6104180_m.fxe"},
	{0xED86F6805A8BA939ULL, "vxca_caverns2_7xl_0_0.vlm"},
	{0xED86F8CCDE59C858ULL, "gen00fl_anders6142494.cl"},
	{0xED86FEB0F1A97CC0ULL, "vxlt_backalley_g_1kt_0_0.vlm"},
	{0xED8726FA5DF1C77AULL, "tbe100ip_shop_codex.utp"},
	{0xED875111355C5343ULL, "436543147.ogg"},
	{0xED8758E61DD24165ULL, "dae000_flemeth6000768.cl"},
	{0xED8777833B4C7506ULL, "526483051.ogg"},
	{0xED8779D364FCC09FULL, "6120419_m.fxe"},
	{0xED879068B6359C7BULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_c.impa.evt"},
	{0xED879B371BCB1C4EULL, "6120076_m.fxe"},
	{0xED87D8D033B173B8ULL, "428996943.ogg"},
	{0xED87F9886BEC686EULL, "898202302.ogg"},
	{0xED88A7DC3836C4C0ULL, "vxca_caverns1_e40_0_0.vlm"},
	{0xED891C2C37B16751ULL, "6116811_f.fxe"},
	{0xED892BB38E24D027ULL, "pro000ip_locked_chest.utp"},
	{0xED8A3B05B095ADFBULL, "6074429_f.fxe"},
	{0xED8A882CB9B43DF9ULL, "lgt000_chanter6147196.cl"},
	{0xED8AA8FF8D8F8532ULL, "vxw_encounter1_d7_0_0.vlm"},
	{0xED8ABC2CFB6080E7ULL, "zz_enable_menu.ncs"},
	{0xED8BDF75E0661D46ULL, "231798619.ogg"},
	{0xED8C8799E6AF917EULL, "zz_followers_banter6158847.cl"},
	{0xED8CC71F9324B60EULL, "762216365.ogg"},
	{0xED8CE1736199B2BDULL, "6103744_m.fxe"},
	{0xED8D1D8B58C27B48ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\exploration\\movement\\mov_rf_injured.gad"},
	{0xED8D3A0CCA5913EDULL, "732198689.ogg"},
	{0xED8E2ACDEA213444ULL, "vxht_0_hightown_a2_bgb_0_0.vlm"},
	{0xED8E2D8F5BB81BBBULL, "vxht_keep_n_621_0_0.vlm"},
	{0xED8E7B8CB6C4BE99ULL, "vxlt_0_lowtown_a2_hhh_0_0.vlm"},
	{0xED8E8968B63BAEBCULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_c.impa.gad"},
	{0xED8EDA65AB9F4272ULL, "zz_ave_debug6207466.cl"},
	{0xED8F54700C9CDDECULL, "fake_im_arm_glv_hvy_champ.uti"},
	{0xED8FB34ECBFE806AULL, "vxca_caverns1_k6n_0_0.vlm"},
	{0xED8FE21357D4360BULL, "519556505.ogg"},
	{0xED8FF1C8F14AE806ULL, "419166168.ogg"},
	{0xED901B8D7FF531C9ULL, "991997672.ogg"},
	{0xED906327195ACAB9ULL, "one280_emrys6104276.cl"},
	{0xED9072E59D0138B0ULL, "vxw_sundermount_298_0_0.vlm"},
	{0xED90E661F147C8A8ULL, "character_character_p223.psh.fxo.dx11.assembly"},
	{0xED910E673C00E89CULL, "gen00fl_dog6167641.cl"},
	{0xED913500979D1218ULL, "vxlt_0_lowtown_docks_a3_n_coz_0_0.vlm"},
	{0xED91522C95F59809ULL, "192313246.ogg"},
	{0xED91AEBDDD9F2810ULL, "6082990_m.fxe"},
	{0xED91B8CF0D9B6687ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.cess.evt"},
	{0xED9225005B24DC0FULL, "vxw_sundermount_34i_0_0.vlm"},
	{0xED9241BA1B301903ULL, "986785895.ogg"},
	{0xED92A4206AA11C2FULL, "1032698846.ogg"},
	{0xED92E9CCB4AEA186ULL, "hlac_w_encounter2_camp_-65532_0z.dds"},
	{0xED934EFD09822746ULL, "ach_seb_loyalty.dds"},
	{0xED9386AC8F7CC2E6ULL, "vxht_0_hightown_a2_c5o_0_0.vlm"},
	{0xED93E759BDD53236ULL, "6111542_m.fxe"},
	{0xED93F572187B72E2ULL, "vxht_0_hightown_a2_627_0_0.vlm"},
	{0xED94C0D31EDBA565ULL, "vxw_bonepit_15m_0_0.vlm"},
	{0xED953247E3F70750ULL, "qun120_arishok_guard6054846.cl"},
	{0xED9563FFBE576DC7ULL, "vxht_keep_6j6_0_0.vlm"},
	{0xED957A9849D14453ULL, "vxlt_gallowstemplar_5uv_0_1.vlm"},
	{0xED958868B641CBBAULL, "art\\vfx\\combat\\vfx_m_sta_phys_close_c.impa.ani"},
	{0xED95AA16B0D7F8E3ULL, "6101450_m.fxe"},
	{0xED95BBE210AC7D39ULL, "6177703_m.fxe"},
	{0xED95DD4396ADBEECULL, "680748054.ogg"},
	{0xED96CD8152A4934FULL, "hlac_lt_0_lowtown_docks_n_-131076_0y.dds"},
	{0xED96EDE1ABD15AE8ULL, "vxca_caverns2prc_abl_0_0.vlm"},
	{0xED9788132E73AE6DULL, "vxlt_gallowstemplar_37a_0_0.vlm"},
	{0xED981042463B9737ULL, "394516519.ogg"},
	{0xED981733B8302D21ULL, "gen00fl_anders6111264.cl"},
	{0xED985D1877B67993ULL, "vxw_blightlands_bn_0_0.vlm"},
	{0xED98D9CF0DA1BDD0ULL, "art\\vfx\\creatures\\fxc_arch_fbp_p.cess.gad"},
	{0xED991ADC7F859DB2ULL, "vxw_encounter1_i0_0_0.vlm"},
	{0xED99216A6A484564ULL, "6149884_m.fxe"},
	{0xED995FCF3AFCEB66ULL, "vxlt_0_lowtown_db5_0_0.vlm"},
	{0xED9A0D53C791C40FULL, "6144840_m.fxe"},
	{0xED9A18D43E3DFA5BULL, "zz_and_debug6178577.cl"},
	{0xED9A2BF66EEA6521ULL, "vxlt_0_lowtown_docks_a3_cop_0_0.vlm"},
	{0xED9A7233C452A11FULL, "hlac_ht_0_hightown_n_-196609_0y.dds"},
	{0xED9A775891AA1AB4ULL, "vxw_bonepitmining_a2_1pi_0_0.vlm"},
	{0xED9A8CEF36D679EDULL, "hlac_ca_caverns1_mines_24426_12x.dds"},
	{0xED9A8DEC748131FDULL, "vxht_mansion2_1ub_0_0.vlm"},
	{0xED9AB26BBA4A378AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft.mao"},
	{0xED9AB66BBA4A3ED9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft.mmh"},
	{0xED9AC06BBA4A4FC7ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft.msh"},
	{0xED9AFF5B6B64C278ULL, "hlac_lt_hangedman_n_-1_13x.dds"},
	{0xED9B37DC6794B667ULL, "247804873.ogg"},
	{0xED9B735F177D25C1ULL, "452915838.ogg"},
	{0xED9BEC4C28B5760BULL, "vxw_encounter2_cz_0_0.vlm"},
	{0xED9C1A475CA795D0ULL, "art\\levels\\races\\elf\\props\\alienage\\prp_basketbunch.phy"},
	{0xED9C24F34D72D123ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fire_c.impa.ani"},
	{0xED9CD72106130B12ULL, "vxlt_gallowstemplar_f_5wg_0_0.vlm"},
	{0xED9CFA2693D81088ULL, "vxht_mansion_n_62r_0_0.vlm"},
	{0xED9D2C25969D2363ULL, "crt_ogre.wbd"},
	{0xED9D84A38760AA5EULL, "vxlt_0_lowtown_docks_n_bfy_0_0.vlm"},
	{0xED9DDDFF48EFA7EAULL, "6039532_m.fxe"},
	{0xED9DE6EB19B73290ULL, "vxca_cave_19k_0_0.vlm"},
	{0xED9EDC2490D49130ULL, "802855699.ogg"},
	{0xED9EF94F93B2726FULL, "vxlt_undercityserialkiller_1fp_0_0.vlm"},
	{0xEDA045457A2FBF5BULL, "zz_mag_debug6054480.cl"},
	{0xEDA096DD4D1AF48FULL, "vxw_sundermount_33h_0_0.vlm"},
	{0xEDA100D8D3E23323ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.dura.evt"},
	{0xEDA11331F5EC5BF0ULL, "388232949.ogg"},
	{0xEDA12309451A1BA8ULL, "895970275.ogg"},
	{0xEDA1258DA180FDF4ULL, "305009258.ogg"},
	{0xEDA19B99011EF383ULL, "hlac_w_sundermount_-65531_0y.dds"},
	{0xEDA257871294BF67ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_dustmots_sml.mmh"},
	{0xEDA2F74D47AD07A0ULL, "6006073_m.fxe"},
	{0xEDA34EEBD051F5A2ULL, "vxht_mansion_n_4xh_0_0.vlm"},
	{0xEDA35A43138F4306ULL, "vxlt_0_lowtown_a2_gzf_0_0.vlm"},
	{0xEDA3DDDA7AF7FE31ULL, "gen00fl_fenris6116559.cl"},
	{0xEDA5212BB0D1A319ULL, "gen00fl_aveline6122092.cl"},
	{0xEDA53A1B549E7540ULL, "6054368_m.fxe"},
	{0xEDA54674D1EF3C33ULL, "6208683_m.fxe"},
	{0xEDA615D6221F6C73ULL, "zz_followers_banter6157791.cl"},
	{0xEDA6C66523214E44ULL, "art\\levels\\ztest\\chad\\w_lbw_sotf.mao"},
	{0xEDA6EF29473A57A2ULL, "vxlt_gallowsprison_n_6mj_0_0.vlm"},
	{0xEDA7408C49897490ULL, "art\\levels\\ztools\\default\\generic\\0ff_window.phy"},
	{0xEDA793475C3E5DE4ULL, "6142984_m.fxe"},
	{0xEDA7D9D8D3E80FA4ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.dura.gad"},
	{0xEDA8186B0C4ED917ULL, "505203884.ogg"},
	{0xEDA8661350D09A45ULL, "llac_dr_slavers_3214_12.dds"},
	{0xEDA961862F26DCB6ULL, "212745739.ogg"},
	{0xEDA97A0EA67F7431ULL, "mag360_samson6044937.cl"},
	{0xEDA9C64C5C8F649DULL, "6121098_m.fxe"},
	{0xEDA9F448575CFB0AULL, "6177670_m.fxe"},
	{0xEDAA2D3502EEE173ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_butcherblock.phy"},
	{0xEDAA3CF34D7F371AULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fire_c.impa.evt"},
	{0xEDAA620EB49528D7ULL, "follower_banter6171911.cl"},
	{0xEDAAAE21943BEFE1ULL, "395386029.ogg"},
	{0xEDAAEBFB26AAFAD8ULL, "6106001_m.fxe"},
	{0xEDAB5CEE03EB4660ULL, "vxlt_gallowscourtyard_n_76v_0_0.vlm"},
	{0xEDABF6807956C2DEULL, "vxlt_0_lowtown_docks_n_bcz_0_0.vlm"},
	{0xEDAC07DDB419722CULL, "w_bonepitdestroyed_a3_13_0.msh"},
	{0xEDAC0DDDB4197C1EULL, "w_bonepitdestroyed_a3_13_0.mmh"},
	{0xEDAC19DDB41990B5ULL, "w_bonepitdestroyed_a3_13_0.mao"},
	{0xEDAC58C80BAD4B10ULL, "vxca_cave_14h_0_0.vlm"},
	{0xEDACDDD211F4439AULL, "433879893.ogg"},
	{0xEDAD7106BFC1426BULL, "sse_ss_merrill.gda"},
	{0xEDAD955B9D7A427AULL, "vxlt_hangedman_14p_0_0.vlm"},
	{0xEDADD9F1E22A4FA9ULL, "1059409799.ogg"},
	{0xEDAE45CD6AFD7D0CULL, "dr_slavers_3216.rml"},
	{0xEDAE6D960B77FAAEULL, "212560480.ogg"},
	{0xEDAED0FE73A46D55ULL, "545643679.ogg"},
	{0xEDAEE8D8D3EE4782ULL, "art\\vfx\\combat\\vfx_m_wpn_nat_crust_c.dura.ani"},
	{0xEDAF861340855051ULL, "793190080.ogg"},
	{0xEDAFEBE547D9F476ULL, "466705967.ogg"},
	{0xEDB007B3700C20A3ULL, "411204151.ogg"},
	{0xEDB0410299565FE2ULL, "qcr220_orsino6061794.cl"},
	{0xEDB075D3272108F5ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_hed_mup.dds"},
	{0xEDB0B6CA7662E81CULL, "vxlt_0_lowtown_n_e1h_0_0.vlm"},
	{0xEDB10AFFF0A71CFBULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock4b.mmh"},
	{0xEDB116FFF0A73098ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock4b.mao"},
	{0xEDB128FFF0A74FE5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_rock4b.msh"},
	{0xEDB13BF34D85540DULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fire_c.impa.gad"},
	{0xEDB13F9A5DD507A6ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.anb"},
	{0xEDB1A91641BAB192ULL, "hlac_lt_0_lowtown_n_-2_6y.dds"},
	{0xEDB206A14BA11156ULL, "vxdr_primevaltomb_7yk_0_1.vlm"},
	{0xEDB228EB631B703FULL, "hlac_lt_backalley_g_0_3z.dds"},
	{0xEDB2F241FD41A89EULL, "qun221_petrice6093253.cl"},
	{0xEDB36AC854EC172DULL, "56021469.ogg"},
	{0xEDB3B12481F4448EULL, "w_encounter3_210_0.msh"},
	{0xEDB3BF2481F45CDFULL, "w_encounter3_210_0.mao"},
	{0xEDB3CB2481F4717CULL, "w_encounter3_210_0.mmh"},
	{0xEDB3DBC1BA3F739AULL, "dre000_bodahn6028167.cl"},
	{0xEDB4717DF3800D55ULL, "follower_banter6161211.cl"},
	{0xEDB471CB00719820ULL, "6171397_m.fxe"},
	{0xEDB511DC5615FB41ULL, "93727760.ogg"},
	{0xEDB58E8A75561062ULL, "vxht_keep_n_64j_0_0.vlm"},
	{0xEDB670294F9B23C2ULL, "6205999_m.fxe"},
	{0xEDB6E52705DF43B2ULL, "630815652.ogg"},
	{0xEDB6F6CBA082B651ULL, "qun211_party_comment6072270.cl"},
	{0xEDB6FED0B5566836ULL, "dae000_bartrand6207000.cl"},
	{0xEDB796D1496F2196ULL, "6208400_m.fxe"},
	{0xEDB799E214B7BC8DULL, "gen00fl_sebastian6174013.cl"},
	{0xEDB7B9DEFEBC67D8ULL, "248317915.ogg"},
	{0xEDB7CA969A2A9996ULL, "91739678.ogg"},
	{0xEDB80EF57939D984ULL, "6166911_m.fxe"},
	{0xEDB8CBAD486FA181ULL, "gen00fl_fenris6114531.cl"},
	{0xEDB94B99BEB73B56ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax105a_0s.dds"},
	{0xEDB9CECAF5E15EE0ULL, "vxlt_gallowscourtyard_n_79w_0_0.vlm"},
	{0xEDB9E527BDA9CAA1ULL, "1062574493.ogg"},
	{0xEDBA3507C6B7A0DAULL, "6109396_m.fxe"},
	{0xEDBA52E0365D8DB0ULL, "vxw_encounter2_5o_0_0.vlm"},
	{0xEDBA723D1B3FC11DULL, "static_static_p0.psh.fxo.dx11.assembly"},
	{0xEDBA949058BCFCAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_enter.ani"},
	{0xEDBAE5C556E2D65BULL, "mag131_keran6086159.cl"},
	{0xEDBB6A3C2ED8ECC1ULL, "491403936.ogg"},
	{0xEDBBA2B42E7120D5ULL, "6095847_m.fxe"},
	{0xEDBBDF398C125DBEULL, "6020992_m.fxe"},
	{0xEDBC71B40F387611ULL, "vxht_0_hightown_a3_9zf_0_0.vlm"},
	{0xEDBCCE19FE0FAC51ULL, "1052013027.ogg"},
	{0xEDBCF5BA43BD7D17ULL, "6115309_m.fxe"},
	{0xEDBD9F147EBABCB0ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookofinquisition_dp.dds"},
	{0xEDBDFB9043E7580CULL, "gen_im_trs_gen_lgt_hlm_r0.uti"},
	{0xEDBE510771FC0921ULL, "hlac_w_sundermount_262142_0y.dds"},
	{0xEDBEA64F2C6E57DCULL, "141890210.ogg"},
	{0xEDBEAD151492AF77ULL, "12764573.ogg"},
	{0xEDBED899F82F480EULL, "vxlt_ruinedhouse_n_2t_0_0.vlm"},
	{0xEDBF429A5DE14A51ULL, "art\\vfx\\combat\\vfx_m_pri_infer_c.mmh"},
	{0xEDBF680DABD5A55BULL, "436818746.ogg"},
	{0xEDBF80E9A385C8DDULL, "zz_test_general6203149.cl"},
	{0xEDBFA102197AAFBBULL, "6179268_m.fxe"},
	{0xEDBFE5316F4329C4ULL, "6096048_f.fxe"},
	{0xEDBFE8A3CAD24461ULL, "follower_banter6161310.cl"},
	{0xEDC02DF1304D6F4CULL, "vxht_playershouse_n_1o7_0_0.vlm"},
	{0xEDC03E669C3F9285ULL, "6115783_m.fxe"},
	{0xEDC0788E709EFBB5ULL, "vxca_caverns2_8ls_0_0.vlm"},
	{0xEDC07B8D11270AA4ULL, "6051323_m.fxe"},
	{0xEDC08327255C7123ULL, "or_db_bl.mao"},
	{0xEDC0C324514CE717ULL, "hm_refugee_2.mor"},
	{0xEDC0EDFE7DDF0551ULL, "6128692_f.fxe"},
	{0xEDC14244A3802AC6ULL, "mrl000ip_at_exit_house.utp"},
	{0xEDC1863669C469CBULL, "vxht_0_hightown_c8e_0_0.vlm"},
	{0xEDC1B72C35E5215DULL, "gen00fl_isabela6106411.cl"},
	{0xEDC213E7C274A5B0ULL, "234285169.ogg"},
	{0xEDC21E2596BC2711ULL, "crt_ogre.bnk"},
	{0xEDC23C17D68E7442ULL, "6043758_f.fxe"},
	{0xEDC24C7AC0EB676CULL, "840800404.ogg"},
	{0xEDC2670D0D70FF3CULL, "vxlt_0_lowtown_a2_hc0_0_0.vlm"},
	{0xEDC345830C616DFEULL, "vxlt_0_lowtown_a2_glh_0_0.vlm"},
	{0xEDC39FD03FCDB117ULL, "vxht_keep_n_6h6_0_1.vlm"},
	{0xEDC3F359390F3836ULL, "vxw_woundedcoast_1h_0_0.vlm"},
	{0xEDC45786E20A5C30ULL, "6112115_f.fxe"},
	{0xEDC4771CE143BFD8ULL, "6170013_m.fxe"},
	{0xEDC4853F52EA3765ULL, "vxca_caverns1_e7h_0_0.vlm"},
	{0xEDC4DF039EEFB21EULL, "hlac_lt_hangedman_n_-1_12y.dds"},
	{0xEDC4EDB4455DA3A8ULL, "939905900.ogg"},
	{0xEDC5B940DBA90811ULL, "75206740.ogg"},
	{0xEDC5DBD088099190ULL, "vxlt_undercitydungeon_12z_0_0.vlm"},
	{0xEDC633E880128637ULL, "hlac_lt_gallowstemplar_f_2394_11x.dds"},
	{0xEDC654DFC9B18EFFULL, "298832461.ogg"},
	{0xEDC681E7E6922BBEULL, "vxlt_0_lowtown_a2_n_e3d_0_0.vlm"},
	{0xEDC741D7A7766FB3ULL, "399723664.ogg"},
	{0xEDC7765345F8DA54ULL, "6114950_m.fxe"},
	{0xEDC7E251A78D44CCULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_dc_transition.ani"},
	{0xEDC81D6826108092ULL, "391801088.ogg"},
	{0xEDC84F27129B114CULL, "gen_im_cft_reg_b.uti"},
	{0xEDC8598EEA3B45E6ULL, "vxw_bonepit_1p9_0_0.vlm"},
	{0xEDC88C9058C92CFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_enter.evt"},
	{0xEDC91704B7A381C4ULL, "follower_barks6153294.cl"},
	{0xEDC94C9794282D2DULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\4so_chantryrb_1.mao"},
	{0xEDC95C8F15098D74ULL, "6039813_m.fxe"},
	{0xEDC9C66B8DAE6723ULL, "sky_dr_slavers.mao"},
	{0xEDC9D102EB9C18FCULL, "vxht_brothel_n_5gi_0_0.vlm"},
	{0xEDC9E611F1C9D2ABULL, "6219270_m.fxe"},
	{0xEDCB1733CC0950ACULL, "656172931.ogg"},
	{0xEDCB4B5B0133AA66ULL, "and100_anders6095905.cl"},
	{0xEDCB4FDA486A092AULL, "218474004.ogg"},
	{0xEDCBE5F02BE8898CULL, "character_character_p269.psh.fxo.dx11"},
	{0xEDCC4E8087A4B2F2ULL, "240922432.ogg"},
	{0xEDCC5F10D41A1B26ULL, "6016947_f.fxe"},
	{0xEDCD4CA23923D3FCULL, "6173908_m.fxe"},
	{0xEDCD656367EA2CEDULL, "llac_lt_0_lowtown_docks_-3_7.dds"},
	{0xEDCD73BC201610D9ULL, "dae000_bartrand6027452.cl"},
	{0xEDCDDB312935A88EULL, "vxca_cavern1_haunted_e8g_0_0.vlm"},
	{0xEDCE6C43C63448A0ULL, "lgt220_shp_mag_guard6149878.cl"},
	{0xEDCECD4A6EB4999DULL, "6134922_m.fxe"},
	{0xEDCEEA6B62951435ULL, "vxca_caverns2_8kr_0_0.vlm"},
	{0xEDCEEF51A793790EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_dc_transition.gad"},
	{0xEDCEFE90A71462E6ULL, "6116374_m.fxe"},
	{0xEDCF3B8350AADA8EULL, "follower_barks6173412.cl"},
	{0xEDCF6D9058CF16E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_action_enter.gad"},
	{0xEDCFDE80707372A8ULL, "one220_yevhen6103114.cl"},
	{0xEDD02142B5651326ULL, "723645041.ogg"},
	{0xEDD0289E826B5773ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13a.phy"},
	{0xEDD04D6BFB437B17ULL, "hlac_lt_gallowscourtyard_n_0_0x.dds"},
	{0xEDD058F1662C4336ULL, "follower_banter6172827.cl"},
	{0xEDD0D8E9FF6717BCULL, "vxlt_0_lowtown_a2_hh1_0_0.vlm"},
	{0xEDD0F24DD55FF374ULL, "527621627.ogg"},
	{0xEDD1945D3B4042D4ULL, "6110877_m.fxe"},
	{0xEDD19B661CB7FCAFULL, "llac_ca_caverns2prc_5802_9.dds"},
	{0xEDD1AFB8A0D8C018ULL, "hlac_lt_0_lowtown_docks_a3_n_-4_7z.dds"},
	{0xEDD1B75FFE57867EULL, "vxlt_0_lowtown_a2_gci_0_0.vlm"},
	{0xEDD1D6DAB0080C17ULL, "6075205_m.fxe"},
	{0xEDD294F48E81BFE9ULL, "llac_lt_0_lowtown_a2_n_-131072_6.dds"},
	{0xEDD2B3181973B867ULL, "800588857.ogg"},
	{0xEDD30BA125BABE9CULL, "gen00fl_anders6125875.cl"},
	{0xEDD346CC51D09B93ULL, "vxw_sundermount_2aw_0_0.vlm"},
	{0xEDD3529EA858379DULL, "mag321ip_corthwaite_store.utp"},
	{0xEDD37B581ED818D2ULL, "hlac_ca_caverns1_24424_3z.dds"},
	{0xEDD38120EF8A1C43ULL, "hlac_dr_primevaltomb_664_4y.dds"},
	{0xEDD41BCEA47F8AA9ULL, "144837271.ogg"},
	{0xEDD42F1713025613ULL, "113817518.ogg"},
	{0xEDD460AB2925FD77ULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em01.phy"},
	{0xEDD4D29990EBAA6EULL, "vxht_mansion_n_6np_0_0.vlm"},
	{0xEDD4EA51A797DCEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_dc_transition.evt"},
	{0xEDD53A0E490C9DA1ULL, "490093994.ogg"},
	{0xEDD59E6C5FD68747ULL, "vxw_encounter2_camp_qy_0_1.vlm"},
	{0xEDD5E83037EEF2D4ULL, "6113850_m.fxe"},
	{0xEDD6410585B6A6D2ULL, "follower_banter6163151.cl"},
	{0xEDD6EAD2CC1D4040ULL, "qcr000_arishok6059659.cl"},
	{0xEDD789F6448DF330ULL, "vxw_blightlands_gf_0_0.vlm"},
	{0xEDD78C233CF8B991ULL, "plt_dae00pt_followers.plo"},
	{0xEDD7C152BB76B12FULL, "1028734987.ogg"},
	{0xEDD7FF572289B3A7ULL, "417591378.ogg"},
	{0xEDD8ADBAD397848AULL, "hlac_ht_keep_3732_4z.dds"},
	{0xEDD95F7BB9DDDB1BULL, "67632461.ogg"},
	{0xEDD98283D5DCDDC7ULL, "6018544_f.fxe"},
	{0xEDD9B768A98A5D43ULL, "dae000_gamlen6139441.cl"},
	{0xEDD9D2F5F33F66D8ULL, "vxdr_slavers_134_0_0.vlm"},
	{0xEDDA492B5038560CULL, "6118165_m.fxe"},
	{0xEDDA6CBCA4A39C37ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_diagonal.phy"},
	{0xEDDA81E070977EAAULL, "vis121_saemus6099698.cl"},
	{0xEDDBF304A9CDC2A4ULL, "552664804.ogg"},
	{0xEDDC5DE883FC408CULL, "vxht_keep_3xr_0_0.vlm"},
	{0xEDDC6A047C7A2852ULL, "vxlt_0_lowtown_a2_n_c_efa_0_0.vlm"},
	{0xEDDC78D5FE8C04ECULL, "6029038_m.fxe"},
	{0xEDDC98782C02974BULL, "gen00fl_fenris_mid_stage.stg"},
	{0xEDDCE088C283957CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_loop.ani"},
	{0xEDDD21B6DB67A506ULL, "vxlt_0_lowtown_a2_gnj_0_0.vlm"},
	{0xEDDD3E63603FA9C3ULL, "358658978.ogg"},
	{0xEDDD457E599C5C6DULL, "vxht_0_hightown_a2_a6k_0_0.vlm"},
	{0xEDDD5488DE1C17C1ULL, "vxw_bonepitmining_a2_et_0_0.vlm"},
	{0xEDDE337DA3CA4D4EULL, "hlac_w_sundermount_327679_0y.dds"},
	{0xEDDF24A382F353BAULL, "gen00fl_aveline6106392.cl"},
	{0xEDDF2DBE7B407AB5ULL, "lt_0_lowtown_docks_n_217_0.msh"},
	{0xEDDF2FBE7B407E2BULL, "lt_0_lowtown_docks_n_217_0.mmh"},
	{0xEDDF3BBE7B4092C8ULL, "lt_0_lowtown_docks_n_217_0.mao"},
	{0xEDDFA4AAE0681BFBULL, "424675128.ogg"},
	{0xEDDFF01B846F438FULL, "732738108.ogg"},
	{0xEDE048309E61623BULL, "6208635_f.fxe"},
	{0xEDE0923CD8BC2E54ULL, "fx_blurdotgui.mao"},
	{0xEDE0BE348C4569ECULL, "leave_03_l2.mao"},
	{0xEDE0F0858195E6D2ULL, "6107435_m.fxe"},
	{0xEDE0FB00AE62045AULL, "181639804.ogg"},
	{0xEDE11D77F2CE674BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\commander\\templarcaptain\\blendtree\\templarcaptain_override.anb"},
	{0xEDE15AF33B10B57CULL, "491690856.ogg"},
	{0xEDE17E8D7F53F2CCULL, "one100dg_vanard6048230.cl"},
	{0xEDE187F4E0859FEEULL, "522377637.ogg"},
	{0xEDE1B8A943C6B413ULL, "vxw_sundermount_2jv_0_0.vlm"},
	{0xEDE1D7CA1E2C7BF8ULL, "hlac_lt_darktown_-1_2x.dds"},
	{0xEDE211CBBD627B8CULL, "6123408_m.fxe"},
	{0xEDE21A6318F67793ULL, "18151739.ogg"},
	{0xEDE2BF8A6BF5C290ULL, "vxlt_undercitydungeon_18x_0_0.vlm"},
	{0xEDE30B2FCBC791E4ULL, "249034343.ogg"},
	{0xEDE3793E72B3836DULL, "6109207_m.fxe"},
	{0xEDE3E65428BA7931ULL, "vxca_caverns2_78v_0_2.vlm"},
	{0xEDE3ED88C289C9BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\conspire_reaction_loop.gad"},
	{0xEDE4161EFD729B8BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medf_0t.dds"},
	{0xEDE49518B72CB6E3ULL, "353051235.ogg"},
	{0xEDE4CE0762B06C3EULL, "301831307.ogg"},
	{0xEDE597E821188F94ULL, "1004879603.ogg"},
	{0xEDE5FA93AFCADFFDULL, "mag110_cut_abomination6037106.cl"},
	{0xEDE5FBD336840BB0ULL, "vxw_blightlands_hg_0_0.vlm"},
	{0xEDE60ECAF20BB774ULL, "145188469.ogg"},
	{0xEDE6B4BCCF8849FCULL, "vxht_brothel_n_5mo_0_0.vlm"},
	{0xEDE6C75035308564ULL, "mag170_thrask6042852.cl"},
	{0xEDE6ED15BA4F42D2ULL, "269264106.ogg"},
	{0xEDE703442A634EDBULL, "vxca_caverns2_8ox_0_0.vlm"},
	{0xEDE7494F8FC0BB0DULL, "76501813.ogg"},
	{0xEDE7515C4D751A03ULL, "474980849.ogg"},
	{0xEDE7C5449F297808ULL, "llac_lt_0_lowtown_a2_131070_5.dds"},
	{0xEDE7D06B840BA5A8ULL, "hm_isa_tevint_mage.mor"},
	{0xEDE7D4DC35CDB8F8ULL, "6064531_f.fxe"},
	{0xEDE838057D090E9BULL, "one300dg_lia6064525.cl"},
	{0xEDE8BB08422E6E8CULL, "6117065_m.fxe"},
	{0xEDE8E0247DF0005EULL, "584771792.ogg"},
	{0xEDE9483FDCBFFBC3ULL, "hlac_dr_dungeon_3214_12x.dds"},
	{0xEDE9865CBABE9A27ULL, "t3_arm_dal.tnt"},
	{0xEDEA1B94DD9EAE7AULL, "vxht_mansion_n_59n_0_0.vlm"},
	{0xEDEA26DAB752CF4FULL, "bld_drian_tu2_a.dds"},
	{0xEDEA6045086864CFULL, "6075642_f.fxe"},
	{0xEDEAB343234A84E5ULL, "6154982_m.fxe"},
	{0xEDEAB64748A49168ULL, "6099367_m.fxe"},
	{0xEDEB0472353982FCULL, "6095346_f.fxe"},
	{0xEDEB356B389D1A46ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.cess.ani"},
	{0xEDEB835BC225AC31ULL, "57990562.ogg"},
	{0xEDEB883C8A207022ULL, "vxht_mansion_n_4cu_0_0.vlm"},
	{0xEDEC02E3568630AAULL, "lgt310ip_drop_zone_4.utp"},
	{0xEDEC7043D1974C6EULL, "qcr000_arishok6060129.cl"},
	{0xEDEC7D6904448817ULL, "vxlt_0_lowtown_docks_aoo_0_0.vlm"},
	{0xEDECA13D41574E94ULL, "6201266_f.fxe"},
	{0xEDECC6A2677331B4ULL, "vxw_bonepitdestroyed_a3_1ws_0_1.vlm"},
	{0xEDECE1825D479D29ULL, "pro000_aveline_intro6109387.cl"},
	{0xEDEF669E8C394729ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\tw.anb"},
	{0xEDF00A010F9A2BAEULL, "vxlt_0_lowtown_docks_a3_n_9r9_0_0.vlm"},
	{0xEDF02A8635BCCC93ULL, "vxw_sundermount_2oy_0_0.vlm"},
	{0xEDF04E7A9AD1F227ULL, "hlac_lt_0_lowtown_n_131072_3x.dds"},
	{0xEDF07F2733584FADULL, "gen_im_arm_boo_lgt_act1_02.uti"},
	{0xEDF083A8AF58940CULL, "6128508_m.fxe"},
	{0xEDF0871A201485B2ULL, "hlac_lt_0_lowtown_65534_0z.dds"},
	{0xEDF09A92003A78A0ULL, "hlac_ht_mansion2b_818_40x.dds"},
	{0xEDF0AF0D41A15CF5ULL, "593810917.ogg"},
	{0xEDF1942EBA0D7602ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_left_walk.gad"},
	{0xEDF1A786AEC5EA84ULL, "zz_dae_debug6029439.cl"},
	{0xEDF1BE8CF959EAF6ULL, "da2terrain_terrain_p0.psh.fxo.dx11"},
	{0xEDF28057823865B3ULL, "art\\vfx\\creatures\\fxc_ft_dirt_r_c.impa.evt"},
	{0xEDF2C5621806F2A1ULL, "vxw_bonepitmining_a2_1x8_0_0.vlm"},
	{0xEDF323B8C79F54E8ULL, "vxlt_warehouse_191_0_0.vlm"},
	{0xEDF3261FFF65DE42ULL, "6109983_m.fxe"},
	{0xEDF3E48581B58B06ULL, "l_mag221ar_store.bnk"},
	{0xEDF3EDDDB456FBD5ULL, "w_bonepitdestroyed_a3_13_0.rml"},
	{0xEDF40199F1BC9D0FULL, "gen00fl_sebastian6195829.cl"},
	{0xEDF46BFE243E5170ULL, "433523521.ogg"},
	{0xEDF46DB0287A2430ULL, "vxw_blightlands_md_0_0.vlm"},
	{0xEDF53DA8D90EFCD5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile_0.pwk"},
	{0xEDF57980E6A3B1D5ULL, "w_blightlands_176_0.msh"},
	{0xEDF57B80E6A3B54BULL, "w_blightlands_176_0.mmh"},
	{0xEDF58780E6A3C9E8ULL, "w_blightlands_176_0.mao"},
	{0xEDF5A80778F9768DULL, "vxlt_0_lowtown_docks_a3_n_cze_0_1.vlm"},
	{0xEDF5EE6199E57FE0ULL, "pro000_varric_cass6208400.cl"},
	{0xEDF6B5C59B378FE5ULL, "6096896_f.fxe"},
	{0xEDF6BBE86F9C154FULL, "sib131_letter_react6127298.cl"},
	{0xEDF6C78FCF4ED38DULL, "qun160cr_rebel_leader.bnk"},
	{0xEDF6DE78F359C544ULL, "momnt_cycle.mao"},
	{0xEDF707F4DDB6D2CDULL, "613608571.ogg"},
	{0xEDF7252D6A58CCF1ULL, "591479626.ogg"},
	{0xEDF75ADAC16726F1ULL, "6129050_m.fxe"},
	{0xEDF764A31D48E5C5ULL, "6078756_m.fxe"},
	{0xEDF82F7EFE270CC5ULL, "vxht_mansion_n_53g_0_0.vlm"},
	{0xEDF869AC6B568678ULL, "gift_values.gda"},
	{0xEDF8B72EBA13D0F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_left_walk.evt"},
	{0xEDF8E22836E51EFCULL, "6205764_m.fxe"},
	{0xEDF91D6B38A92EB7ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.cess.evt"},
	{0xEDF94DBE60665952ULL, "vxlt_0_lowtown_a2_n_c_epc_0_0.vlm"},
	{0xEDF9610D070F7728ULL, "zz_dae_debug6142963.cl"},
	{0xEDF97957823E78F4ULL, "art\\vfx\\creatures\\fxc_ft_dirt_r_c.impa.gad"},
	{0xEDF9D0BB653EDECCULL, "6148432_m.fxe"},
	{0xEDF9FA197C1688A2ULL, "vxht_mansion_n_4fv_0_0.vlm"},
	{0xEDFA0BC343BAB02BULL, "930834088.ogg"},
	{0xEDFA2632DD21467CULL, "6080080_m.fxe"},
	{0xEDFA30821BCB5AE1ULL, "follower_banter6143174.cl"},
	{0xEDFA3842C20848C1ULL, "vxw_bonepitmining_a2_oj_0_0.vlm"},
	{0xEDFAD5FBD960C2DBULL, "ibeam.cur"},
	{0xEDFAE98FCF52F6F5ULL, "qun160cr_rebel_leader.cnv"},
	{0xEDFB07EF9B1DF06CULL, "vxht_keep_676_0_0.vlm"},
	{0xEDFB3706F6494724ULL, "451059246.ogg"},
	{0xEDFB4E0BADF27F67ULL, "474293974.ogg"},
	{0xEDFBA7C7F6BF615AULL, "gen00fl_anders6119427.cl"},
	{0xEDFC9582E02C0845ULL, "mcr310_argument6182282.cl"},
	{0xEDFD78E2317D5B1DULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candle_sa.dds"},
	{0xEDFDA5A616623F93ULL, "mag291_ntbanter6056615.cl"},
	{0xEDFDB76BBA9EBDC6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_dwleft.phy"},
	{0xEDFE29B594140323ULL, "944008824.ogg"},
	{0xEDFE9B91D3BE727EULL, "6142630_m.fxe"},
	{0xEDFE9C6327B2E513ULL, "vxw_sundermount_2px_0_0.vlm"},
	{0xEDFEAE43760316A7ULL, "123884832.ogg"},
	{0xEDFECFCC2BF866C7ULL, "aip_stonegolem_base.gda"},
	{0xEDFED852AE823577ULL, "6072702_f.fxe"},
	{0xEDFEE6494C1C725EULL, "6135169_f.fxe"},
	{0xEDFF0A30631FA39AULL, "vxlt_0_lowtown_a3_n_c_cmk_0_0.vlm"},
	{0xEDFF2ADEF02A4C0AULL, "vxht_keep_n_637_0_0.vlm"},
	{0xEDFF520D466405D3ULL, "6129557_m.fxe"},
	{0xEDFF5B3D49869DAFULL, "vxht_mansion2b_1kk_0_1.vlm"},
	{0xEE001E6B38AF4F40ULL, "art\\vfx\\creatures\\fxc_shale_lit_r_c.cess.gad"},
	{0xEE0063CC5529640AULL, "885285194.ogg"},
	{0xEE0088578244B0D2ULL, "art\\vfx\\creatures\\fxc_ft_dirt_r_c.impa.ani"},
	{0xEE00A47E42DD5079ULL, "vxlt_undercityencounter_10x_0_0.vlm"},
	{0xEE00C2E98FF62F58ULL, "6016450_m.fxe"},
	{0xEE00CB1806407F05ULL, "vxlt_gallowstemplar_f_4o5_0_0.vlm"},
	{0xEE00CEA83A8D95CDULL, "417751396.ogg"},
	{0xEE00DA981D805D21ULL, "261569051.ogg"},
	{0xEE00DF19AB76B7B0ULL, "503992363.ogg"},
	{0xEE012FFECC108B7DULL, "mag311_emile_de_launcet6043758.cl"},
	{0xEE0149537AB0423FULL, "hlac_w_woundedcoast_65537_0z.dds"},
	{0xEE017A1D20367CE2ULL, "gen00fl_isabela6097468.cl"},
	{0xEE0197FCF15BF6C2ULL, "6102883_m.fxe"},
	{0xEE01ED0AFDAC2CE5ULL, "art\\vfx\\creatures\\fxc_drg_bfft_p.mmh"},
	{0xEE01FF1C54B3AE78ULL, "vxlt_gallowscourtyard_n_744_0_0.vlm"},
	{0xEE0233EBB15D8317ULL, "mer110_jansen6034266.cl"},
	{0xEE02792E2D6CEEE8ULL, "611130310.ogg"},
	{0xEE02D07D20E50563ULL, "lgt220ip_shp_mag_clue.ncs"},
	{0xEE02F09902DC2CD6ULL, "630267710.ogg"},
	{0xEE02FBA69A6C5AE7ULL, "783259713.ogg"},
	{0xEE032866B48105DBULL, "6072828_m.fxe"},
	{0xEE037CDAF06663F3ULL, "61516217.ogg"},
	{0xEE03E6FE0E4F7FDBULL, "vxca_caverns2_8qz_0_0.vlm"},
	{0xEE03F375065679E7ULL, "gen00fl_fenris6116593.cl"},
	{0xEE03F5E3569AACD7ULL, "lgt310ip_drop_zone_4.ncs"},
	{0xEE053A276D7238C8ULL, "vxca_caverns2prc_807_0_0.vlm"},
	{0xEE053E7E35E284C4ULL, "one100_party_comments6112676.cl"},
	{0xEE054497AE6025B0ULL, "6054288_f.fxe"},
	{0xEE0572F6C911875BULL, "827117261.ogg"},
	{0xEE06A38BE8477609ULL, "qun110_petrice6018528.cl"},
	{0xEE06BF2EBA201BE8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_forward_left_walk.ani"},
	{0xEE06E9F510E3DD1AULL, "vxlt_0_lowtown_a2_h8w_0_0.vlm"},
	{0xEE0729AEC0D30E09ULL, "vxlt_0_lowtown_a2_bym_0_0.vlm"},
	{0xEE07540528DB377CULL, "6208064_m.fxe"},
	{0xEE075696471CA154ULL, "dae000ip_crowd_em2_talk7.utp"},
	{0xEE07576BB6D1F2C1ULL, "6110227_m.fxe"},
	{0xEE07B37F59DE8A0CULL, "vxht_keep_3iq_0_0.vlm"},
	{0xEE07C03C9FACE86AULL, "ts_apostates_wpn_table.gda"},
	{0xEE07CC43A84DB7B6ULL, "85819919.ogg"},
	{0xEE0800850C9D9035ULL, "hlac_lt_backalley_g_0_8y.dds"},
	{0xEE08651EE1862E66ULL, "6152162_f.fxe"},
	{0xEE08AA1FB3FE6141ULL, "vxw_bonepitmining_a2_ji_0_0.vlm"},
	{0xEE09550B1D6ABF0EULL, "vxlt_undercitydungeon_1a1_0_0.vlm"},
	{0xEE096BEB33DBA084ULL, "410887532.ogg"},
	{0xEE0979CC8D1408ECULL, "vxht_keep_607_0_0.vlm"},
	{0xEE098BD054E87E58ULL, "vxht_0_hightown_a2_b8o_0_0.vlm"},
	{0xEE09F631927D7205ULL, "hlac_lt_warehouse_n_-65535_0x.dds"},
	{0xEE0A50092C27FD3FULL, "zz_one_debug6069472.cl"},
	{0xEE0ACF40E2D2F925ULL, "zz_followers_banter6158613.cl"},
	{0xEE0B1870CF78F235ULL, "one100it_ironwood.uti"},
	{0xEE0B41381E35920EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_woodrockin_2.mao"},
	{0xEE0B565F44CB620DULL, "llac_lt_backalley_n_65537_0.dds"},
	{0xEE0BB77251B76B9EULL, "hlac_lt_darktown_-1_3y.dds"},
	{0xEE0BC22E2106B95DULL, "10779751.ogg"},
	{0xEE0BEF86DF3B77D4ULL, "gen00fl_bethany6109219.cl"},
	{0xEE0C263503419FE8ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_butcherblock.mmh"},
	{0xEE0C32350341B38BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_butcherblock.mao"},
	{0xEE0C44350341D212ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_butcherblock.msh"},
	{0xEE0CB627E3ADD20BULL, "vxht_brothel_n_6eo_0_0.vlm"},
	{0xEE0CE9B38852EF9BULL, "6107179_m.fxe"},
	{0xEE0D02A2A2B3EA7DULL, "6111606_m.fxe"},
	{0xEE0D5578F1736955ULL, "vxdr_primevalentrance_39l_0_1.vlm"},
	{0xEE0DB3C3815C2246ULL, "6171791_m.fxe"},
	{0xEE0DF7703DA538A1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_maso_0t.dds"},
	{0xEE0E01B96D66B74DULL, "6171060_m.fxe"},
	{0xEE0E048AC81AD92BULL, "llac_lt_0_lowtown_a3_n_c_-2_5.dds"},
	{0xEE0E0AB09713B623ULL, "450393926.ogg"},
	{0xEE0E74FAB8D8044EULL, "vis121_saemus6118470.cl"},
	{0xEE0E7FA7035CFE96ULL, "1002469350.ogg"},
	{0xEE0EA50BF1FBB41FULL, "6161294_m.fxe"},
	{0xEE0EED75B5BD9A64ULL, "art\\levels\\races\\design\\prp_gasbarrel_6.pwk"},
	{0xEE0F0A4EF8956419ULL, "6044661_m.fxe"},
	{0xEE0FCE8FE2AA6E8BULL, "vxw_blightlands_7y_0_0.vlm"},
	{0xEE0FEA0AFDB864B2ULL, "art\\vfx\\creatures\\fxc_drg_bfft_p.anb"},
	{0xEE1009D9E3520F42ULL, "6107783_m.fxe"},
	{0xEE1068D24F498F8DULL, "6054035_m.fxe"},
	{0xEE1078AC1CA1DD5FULL, "6201356_m.fxe"},
	{0xEE1094BA5C20621CULL, "isa311_isabela6095330.cl"},
	{0xEE10B806EEBD4C0CULL, "6140137_m.fxe"},
	{0xEE10E092B3D3045AULL, "lgt120pt_bhelen.ncs"},
	{0xEE10E2206D8EEE36ULL, "vxw_bonepitmining_a2_1ov_0_0.vlm"},
	{0xEE11038A991A005FULL, "873632994.ogg"},
	{0xEE117082D6DD9800ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingd_sa.dds"},
	{0xEE11970230CD262DULL, "w_woundedcoast_242_0.msh"},
	{0xEE11A50230CD3E60ULL, "w_woundedcoast_242_0.mao"},
	{0xEE11A90230CD44B3ULL, "w_woundedcoast_242_0.mmh"},
	{0xEE11C7D59A5DE950ULL, "seal_dis.mao"},
	{0xEE123D999C030F70ULL, "vxlt_0_lowtown_a2_n_feg_0_0.vlm"},
	{0xEE125D030B6228F1ULL, "vxca_caverns2_7xt_0_0.vlm"},
	{0xEE1263A8B6F515C8ULL, "vxlt_0_lowtown_n_ghq_0_0.vlm"},
	{0xEE135790E374C384ULL, "one220_yevhen6135762.cl"},
	{0xEE13E0027B2A1CF3ULL, "gen00fl_varric6105759.cl"},
	{0xEE13EE4CC1454036ULL, "792605852.ogg"},
	{0xEE146D1CA4BC3555ULL, "gen00fl_anders6128387.cl"},
	{0xEE14AEC04E9745B2ULL, "qcr221_cut_heroes_reward.stg"},
	{0xEE151338E2133DC5ULL, "vxht_mansion_n_59e_0_0.vlm"},
	{0xEE1517A2E15BFA87ULL, "dae210_gate_guard6085054.cl"},
	{0xEE154D49C9D17342ULL, "one100dg_vanard6054206.cl"},
	{0xEE15DF7D8AFAC323ULL, "vxca_caverns1_icn_0_0.vlm"},
	{0xEE1679C0E109C9EEULL, "6083689_m.fxe"},
	{0xEE17584EE23F1783ULL, "vxca_caverns2prc_978_0_0.vlm"},
	{0xEE177068C83CCBA1ULL, "llac_lt_0_lowtown_docks_n_-131075_6.dds"},
	{0xEE1792865B14B4C3ULL, "gen00fl_fenris6113936.cl"},
	{0xEE17A5213CBE28DDULL, "386370532.ogg"},
	{0xEE17C6E80F60D78EULL, "vxlt_undercitydungeon_1d2_0_0.vlm"},
	{0xEE1831FAFAB68BE8ULL, "6072243_m.fxe"},
	{0xEE18D06D9CF6A0FAULL, "257225779.ogg"},
	{0xEE18F8F99A1FEF0CULL, "vxdr_primevaltomb_50m_0_0.vlm"},
	{0xEE190D565DB90AA4ULL, "llac_ca_caverns1_mines_24426_13.dds"},
	{0xEE19A2F3E2A1A80EULL, "6173326_m.fxe"},
	{0xEE19CC9FECAC8C26ULL, "798737131.ogg"},
	{0xEE1A41C615658A1DULL, "117864921.ogg"},
	{0xEE1BA954A395DE9AULL, "vxdr_primevaltomb_39z_0_0.vlm"},
	{0xEE1BAEE169E1D8FDULL, "802935444.ogg"},
	{0xEE1BAF4998DBF98AULL, "l_dae220ar_hightown_nt.wbd"},
	{0xEE1BCE508ACD5800ULL, "hlac_lt_darktown_-1_55z.dds"},
	{0xEE1BE9BFC57A77ACULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1cashsplnter.phy"},
	{0xEE1C25A073523AC6ULL, "6174491_m.fxe"},
	{0xEE1C41AB29637E6CULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em01.mmh"},
	{0xEE1C43622742F43FULL, "6120009_f.fxe"},
	{0xEE1C47AB2963885EULL, "art\\characters\\playercharacter\\elfmale\\mt_aet_em01.msh"},
	{0xEE1C62931DA40566ULL, "vxca_cavern1_haunted_cjo_0_0.vlm"},
	{0xEE1C8E34D363BA46ULL, "6170789_m.fxe"},
	{0xEE1CD81FD8D26F7FULL, "static_lava_p0.vsh.fxo.dx11"},
	{0xEE1CD9B5C9968802ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_04a_0d.dds"},
	{0xEE1CDAE9FA281FABULL, "6196708_m.fxe"},
	{0xEE1D063FF6D09C6EULL, "zz_dae_debug6029802.cl"},
	{0xEE1D16E8E3F1CC9FULL, "6166394_m.fxe"},
	{0xEE1D6514D6605FB5ULL, "379654052.ogg"},
	{0xEE1D771648ABB45BULL, "241752095.ogg"},
	{0xEE1D9D77EA6E92BDULL, "vxw_bonepit_25n_0_0.vlm"},
	{0xEE1DA471749E9E59ULL, "lt_0_lowtown_docks_a3_n_162_0.msh"},
	{0xEE1DAE71749EAF47ULL, "lt_0_lowtown_docks_a3_n_162_0.mmh"},
	{0xEE1DB271749EB594ULL, "lt_0_lowtown_docks_a3_n_162_0.mao"},
	{0xEE1E9D6F5DBE2944ULL, "vxht_0_hightown_a2_b5h_0_0.vlm"},
	{0xEE1EC0C3BF79E3F1ULL, "vxlt_0_lowtown_docks_a3_cq2_0_0.vlm"},
	{0xEE1ED0095280D926ULL, "lgt100cr_leech.utc"},
	{0xEE1F55EFF2533ECBULL, "6087666_m.fxe"},
	{0xEE1F6EB3100E98F6ULL, "gen00fl_merrill6130068.cl"},
	{0xEE2012A275AF0223ULL, "498425074.ogg"},
	{0xEE20277C6989860AULL, "384806506.ogg"},
	{0xEE20374650710161ULL, "hlac_dr_dungeon_3215_7y.dds"},
	{0xEE208032F06AF28BULL, "cod_hst_llomerynaccord.utp"},
	{0xEE209D33EA002A2EULL, "6062773_f.fxe"},
	{0xEE20C51B36178CB2ULL, "725650496.ogg"},
	{0xEE212DA79A794C78ULL, "vxca_caverns1_mines_eic_0_1.vlm"},
	{0xEE2131E438432DFCULL, "690947915.ogg"},
	{0xEE2141BCA4DF558FULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_diagonal.mao"},
	{0xEE214DBCA4DF6A2CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_diagonal.mmh"},
	{0xEE2153BCA4DF741EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0wo_diagonal.msh"},
	{0xEE218E39C6C75BCDULL, "6035856_m.fxe"},
	{0xEE2194AC87F063F1ULL, "244228010.ogg"},
	{0xEE219B132E09BADAULL, "vxlt_0_lowtown_a2_n_c_f66_0_0.vlm"},
	{0xEE220C09F006F059ULL, "hlac_ht_chantry_2534_0z.dds"},
	{0xEE224D49C8F1F6D9ULL, "vxlt_0_lowtown_n_h8k_0_0.vlm"},
	{0xEE22563F3959D942ULL, "hla_ht_0_hightown_131072_0z.dds"},
	{0xEE22F3C405ADAC2DULL, "mag131_taraneh6088748.cl"},
	{0xEE2371F7103769B8ULL, "6109772_m.fxe"},
	{0xEE239D2F362C5720ULL, "llac_w_bonepitdestroyed_a3_262149_0.dds"},
	{0xEE23DEE11A9067E5ULL, "vxw_bonepitmining_a2_258_0_0.vlm"},
	{0xEE2437B10C9E04A6ULL, "llac_lt_house_-1_4.dds"},
	{0xEE246CB6080A0729ULL, "396951301.ogg"},
	{0xEE24D67875A3191EULL, "vxht_playershouse_n_1nb_0_0.vlm"},
	{0xEE252B120AE8951BULL, "489816913.ogg"},
	{0xEE2589D684A473C5ULL, "836671897.ogg"},
	{0xEE25BF939F9212BFULL, "vxht_keep_679_0_0.vlm"},
	{0xEE261C4FF0384E3AULL, "fen210_orana6109736.cl"},
	{0xEE27049CD37E3A59ULL, "statecnt_trap.gda"},
	{0xEE27BFEA380DC87AULL, "vxw_encounter2_camp_rt_0_0.vlm"},
	{0xEE27D364487258E7ULL, "6135075_m.fxe"},
	{0xEE282B13BD6B0A42ULL, "453334277.ogg"},
	{0xEE286076A708AAFDULL, "hlac_lt_gallowstemplar_f_2394_2z.dds"},
	{0xEE28D214B5AD2480ULL, "6155676_m.fxe"},
	{0xEE28E73F9A2CB80CULL, "734137955.ogg"},
	{0xEE28F90DF876ADD6ULL, "6164201_m.fxe"},
	{0xEE2924BAB69CCBE6ULL, "14297966.ogg"},
	{0xEE29556603092FB9ULL, "gen00fl_fenris6103355.cl"},
	{0xEE295D968E8A90BEULL, "vxlt_0_lowtown_docks_a3_n_cq2_0_0.vlm"},
	{0xEE29924BEB83FB45ULL, "6111189_m.fxe"},
	{0xEE2999E1C79A030BULL, "vxht_brothel_n_6ka_0_0.vlm"},
	{0xEE299A8C8F085254ULL, "563871703.ogg"},
	{0xEE2A19FC350B8CA8ULL, "vxlt_warehouse_n_xz_0_0.vlm"},
	{0xEE2A66059595F4C2ULL, "vxca_caverns2prc_ae3_0_0.vlm"},
	{0xEE2AACA681EBABA3ULL, "6166721_m.fxe"},
	{0xEE2AFC403D923438ULL, "371512418.ogg"},
	{0xEE2C0F54DC64AB3DULL, "vxw_bonepit_20m_0_0.vlm"},
	{0xEE2C35910F77FF5DULL, "dae000cr_commoner_d1.utc"},
	{0xEE2CB9DB08A0056CULL, "vxlt_gallowscourtyard_6ee_0_0.vlm"},
	{0xEE2D18C71302489AULL, "6017129_m.fxe"},
	{0xEE2D3CA3E3E70618ULL, "gen00fl_varric6088862.cl"},
	{0xEE2D5171B343D2E9ULL, "120933919.ogg"},
	{0xEE2DF86DFA94C050ULL, "569982696.ogg"},
	{0xEE2E095218B089C1ULL, "6015689_m.fxe"},
	{0xEE2E1295A1328AAEULL, "164986585.ogg"},
	{0xEE2E92684AEE8479ULL, "291409143.ogg"},
	{0xEE2EB94E708B21A5ULL, "325850325.ogg"},
	{0xEE2F47629AE146C8ULL, "vxlt_0_lowtown_n_gbo_0_0.vlm"},
	{0xEE2F5230B033755DULL, "dae000_porfiria6155308.cl"},
	{0xEE2FFC209971DA4DULL, "cod_tomes_tome_1.utp"},
	{0xEE3003479AAA2812ULL, "qcr220_grey_wardens6060447.cl"},
	{0xEE3097233D44F5F6ULL, "plt_dae00pt_followers.nss"},
	{0xEE30BD941BBD3F7AULL, "vxca_caverns2prc_8bc_0_0.vlm"},
	{0xEE314CE09DE8357CULL, "919531527.ogg"},
	{0xEE319121A58129E5ULL, "130649233.ogg"},
	{0xEE3212F021E9AB85ULL, "t3_nbl_n01.tnt"},
	{0xEE321827056744BCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangbar.mao"},
	{0xEE32219E82BE15E8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13a.mmh"},
	{0xEE3224270567595FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangbar.mmh"},
	{0xEE322A2705676301ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangbar.msh"},
	{0xEE32357EABD7DA2BULL, "6067098_f.fxe"},
	{0xEE323F9E82BE4812ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos13a.msh"},
	{0xEE326B2AF50FD066ULL, "vxht_0_hightown_a2_cbz_0_0.vlm"},
	{0xEE32B4C2C7241E5AULL, "6051615_m.fxe"},
	{0xEE32C3376EE6F423ULL, "vxca_caverns1_iip_0_0.vlm"},
	{0xEE3309CED6F6DC0CULL, "528848443.ogg"},
	{0xEE331228CF27E1A9ULL, "vxw_sundermount_1zs_0_0.vlm"},
	{0xEE333F902D87216EULL, "lt_0_lowtown_docks_a3_n_202_0.mmh"},
	{0xEE334B902D873505ULL, "lt_0_lowtown_docks_a3_n_202_0.mao"},
	{0xEE3359902D874D5CULL, "lt_0_lowtown_docks_a3_n_202_0.msh"},
	{0xEE3362DEE4570044ULL, "332339030.ogg"},
	{0xEE3363F3E5A5455AULL, "hlac_lt_0_lowtown_docks_65534_4y.dds"},
	{0xEE33ADC6B6EEEFB9ULL, "73193077.ogg"},
	{0xEE33C49EC3D23018ULL, "6073889_f.fxe"},
	{0xEE33E68581EB72ACULL, "l_mag221ar_store.wbd"},
	{0xEE34342DFFA2D90BULL, "6127322_m.fxe"},
	{0xEE34B6487F38CF92ULL, "gen00fl_sebastian6188662.cl"},
	{0xEE3533699A65AC48ULL, "102742699.ogg"},
	{0xEE35B6028EC5508AULL, "lowtown_fem_alienage1_sta6162548.cl"},
	{0xEE35C342F6715203ULL, "vxht_keep_n_3zp_0_0.vlm"},
	{0xEE35C84BFDCBA42EULL, "6098807_m.fxe"},
	{0xEE35E1CF89F9854AULL, "dae200_cullen6140319.cl"},
	{0xEE3631C72A03E0FAULL, "vxw_encounter2_camp_mu_0_0.vlm"},
	{0xEE363E621EBBFC09ULL, "llac_ca_caverns1_mines_24423_2.dds"},
	{0xEE3704B4A3059907ULL, "383057494.ogg"},
	{0xEE37823A31548D50ULL, "follower_banter6143344.cl"},
	{0xEE37A52FF32F7B8DULL, "vxlt_0_lowtown_docks_a3_n_c46_0_0.vlm"},
	{0xEE37AD3EF24634E4ULL, "character_character_p159.psh.fxo.dx11.assembly"},
	{0xEE37CF738080A93EULL, "vxlt_0_lowtown_docks_a3_n_cz3_0_0.vlm"},
	{0xEE37F0DF06974173ULL, "pro000_soldier_leader6040814.cl"},
	{0xEE387107CD97BB78ULL, "728874954.ogg"},
	{0xEE39E9A6D30D4907ULL, "vxca_caverns1_gjw_0_0.vlm"},
	{0xEE3A4ED831A028B9ULL, "gen00fl_aveline6114040.cl"},
	{0xEE3ABB87E29F1666ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_crnf.ani"},
	{0xEE3AEC963912D30DULL, "6069938_m.fxe"},
	{0xEE3B0291DF67649EULL, "character_character_p9.psh.fxo.dx11.assembly"},
	{0xEE3B542268F28C71ULL, "pro000_meeran6048786.cl"},
	{0xEE3DE09E97D5C006ULL, "zz_mrl_debug6184191.cl"},
	{0xEE3E2F04D1EB6D10ULL, "vxht_brothel_n_636_0_0.vlm"},
	{0xEE3E4CBEC00DEE26ULL, "dae000ip_ambient_talk_pause.utp"},
	{0xEE3ECBAD307FB716ULL, "6096416_m.fxe"},
	{0xEE3ED849E5B54305ULL, "6165115_m.fxe"},
	{0xEE3F44DDA761A0A5ULL, "llac_lt_0_lowtown_a2_n_c_-65537_4.dds"},
	{0xEE3F5AAAC634409CULL, "6137248_f.fxe"},
	{0xEE3FD2A80F76E139ULL, "vxlt_0_lowtown_a2_n_c_ems_0_0.vlm"},
	{0xEE4017327B045A78ULL, "ca_caverns1_24427_ll.mmh"},
	{0xEE404DEDAA4158FFULL, "follower_banter6136367.cl"},
	{0xEE40FE0D0552EFABULL, "6219021_m.fxe"},
	{0xEE418405C11DFA29ULL, "vxw_sundermount_1ur_0_0.vlm"},
	{0xEE41CE7D869C1FA4ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb102a_0d.dds"},
	{0xEE41CF724F755366ULL, "195579716.ogg"},
	{0xEE41DE936C7A8D39ULL, "6126932_m.fxe"},
	{0xEE42260A5D0C0005ULL, "1034243459.ogg"},
	{0xEE422DA73142B5F8ULL, "6163655_m.fxe"},
	{0xEE4243F005A52B05ULL, "vxlt_0_lowtown_docks_a3_n_d4y_0_0.vlm"},
	{0xEE4269C019C129E2ULL, "6086897_m.fxe"},
	{0xEE428018CB958322ULL, "llac_lt_gallowstemplar_n_2391_6.dds"},
	{0xEE4292FAA09E3C99ULL, "6179028_m.fxe"},
	{0xEE43095F7A8A025AULL, "vxdr_0_deeproads_bhf_0_0.vlm"},
	{0xEE433A15770DA712ULL, "vxht_0_hightown_a3_cb3_0_0.vlm"},
	{0xEE43BD3D5E6C6459ULL, "45710705.ogg"},
	{0xEE447CC1F5CD0822ULL, "116642503.ogg"},
	{0xEE44A3A41BF9F97AULL, "vxw_encounter2_camp_hv_0_0.vlm"},
	{0xEE450499F5AD5F96ULL, "386396337.ogg"},
	{0xEE4548F14646F719ULL, "6143453_f.fxe"},
	{0xEE45660F057B0B1BULL, "6059229_m.fxe"},
	{0xEE459994B20E47B3ULL, "vxca_caverns1_dpr_0_0.vlm"},
	{0xEE4602F08CF36E43ULL, "fxe_fluff.dds"},
	{0xEE466DDB808FD509ULL, "447385211.ogg"},
	{0xEE474FD1F034228BULL, "76769428.ogg"},
	{0xEE4778604AFDB444ULL, "dae000_arishok6070885.cl"},
	{0xEE482F8957D0EE1DULL, "6018803_f.fxe"},
	{0xEE48A387E2AB2AD7ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_crnf.evt"},
	{0xEE48AB3A8E97F057ULL, "vxdr_0_deeproads_f50_0_0.vlm"},
	{0xEE4986C2E1E53D0CULL, "vxw_bonepit_yo_0_0.vlm"},
	{0xEE49ED93885901F1ULL, "dae000_meredith6093183.cl"},
	{0xEE4A0236153D189EULL, "one300cr_nuncio.utc"},
	{0xEE4A207CEA536047ULL, "hlac_ht_bartrand_n_2995_6y.dds"},
	{0xEE4A2D7C5D9C3112ULL, "vxlt_gallowstemplar_f_53k_0_0.vlm"},
	{0xEE4A5377B91C7ECCULL, "6043463_m.fxe"},
	{0xEE4A5469F486D6D4ULL, "art\\vfx\\combat\\vfx_gib_hm_lgta_rtlb.anb"},
	{0xEE4ACECA6E613542ULL, "852144400.ogg"},
	{0xEE4B27882CF70B39ULL, "party_targeted.mao"},
	{0xEE4B2FEB9190B09AULL, "956285847.ogg"},
	{0xEE4BA91DCE8E5D84ULL, "vxlt_backalley_g_1kp_0_0.vlm"},
	{0xEE4BB9630F5B089EULL, "hlac_dr_primevaltomb_661_1x.dds"},
	{0xEE4BE36BD920CB9FULL, "923165197.ogg"},
	{0xEE4BF0CB2080F023ULL, "zz_lgt_1_debug6140406.cl"},
	{0xEE4C47C4E5707690ULL, "hlac_lt_undercityserialkiller_-1_1y.dds"},
	{0xEE4C50E52D2EAC34ULL, "mag160_varian_ilithis6018907.cl"},
	{0xEE4CACD3480D675BULL, "560934541.ogg"},
	{0xEE4CD16B23E4F127ULL, "hlac_lt_warehouse_n_1_3y.dds"},
	{0xEE4D1859564FEC08ULL, "hlac_ht_chantry_n_2534_12y.dds"},
	{0xEE4D42CC527BC444ULL, "vxdr_primevaltomb_3tg_0_0.vlm"},
	{0xEE4D7B778D1B3512ULL, "mer120_hubert6020651.cl"},
	{0xEE4E685CEE7905C5ULL, "6207950_m.fxe"},
	{0xEE4E6F5B5BC7BCCAULL, "1054934224.ogg"},
	{0xEE4EFDA95EA516BEULL, "6104712_f.fxe"},
	{0xEE4F26500CBF1248ULL, "llac_ht_mansion2_818_26.dds"},
	{0xEE4F3DFE73C1FF96ULL, "vxw_encounter1_by_0_0.vlm"},
	{0xEE4F6FE9F749082BULL, "6212121_m.fxe"},
	{0xEE4FA487E2B14B60ULL, "art\\characters\\creatures\\mubari\\animation\\ctn90r_crnf.gad"},
	{0xEE4FC762F7D22490ULL, "hlac_ca_caverns2prc_5802_7y.dds"},
	{0xEE501F934D0E8FDDULL, "vxht_brothel_n_3i4_0_0.vlm"},
	{0xEE5027A8DFF75F7BULL, "and111_refugee_barks6111543.cl"},
	{0xEE504D99CC537C92ULL, "21772461.ogg"},
	{0xEE507BA80FC0D051ULL, "346391263.ogg"},
	{0xEE50B1091C3D373FULL, "6148808_m.fxe"},
	{0xEE50DC4ACCDFB5D5ULL, "vxlt_0_lowtown_eo9_0_1.vlm"},
	{0xEE516371580DFE19ULL, "vxht_keep_n_5zt_0_0.vlm"},
	{0xEE51C4AB35A8B141ULL, "art\\characters\\playercharacter\\qunarimale\\hlm_com.mao"},
	{0xEE51E2B97CBB25E3ULL, "6207813_m.fxe"},
	{0xEE523C3F9218E23CULL, "qun160_vashothlead6053854.cl"},
	{0xEE5315810DF011FAULL, "vxw_encounter2_camp_cw_0_0.vlm"},
	{0xEE5381F16EB55B7EULL, "238793408.ogg"},
	{0xEE53E304A10CAEEAULL, "mag110_arianni6016801.cl"},
	{0xEE5408C7FC4A1FC7ULL, "6120015_m.fxe"},
	{0xEE54CAACD77E83C5ULL, "6101961_f.fxe"},
	{0xEE551CDE8906E222ULL, "6086365_m.fxe"},
	{0xEE552FB14DEDBF99ULL, "6206261_m.fxe"},
	{0xEE555A54EFB3BEBAULL, "6150871_m.fxe"},
	{0xEE5575FF21B62826ULL, "vxht_mansion_n_5b3_0_0.vlm"},
	{0xEE561EB51D67A571ULL, "370565953.ogg"},
	{0xEE566C15C09C3DDCULL, "hlac_lt_0_lowtown_131070_5x.dds"},
	{0xEE571D17808E08D7ULL, "vxdr_0_deeproads_f83_0_0.vlm"},
	{0xEE573B45643A6FD4ULL, "hlac_lt_0_lowtown_65536_8z.dds"},
	{0xEE577384E2DD3FE8ULL, "30649147.ogg"},
	{0xEE57802448980734ULL, "hlac_lt_0_lowtown_docks_-131074_1z.dds"},
	{0xEE57A97C49293A9DULL, "60828937.ogg"},
	{0xEE57F89FD3DB558CULL, "vxw_bonepit_dl_0_0.vlm"},
	{0xEE587FB67EFE3917ULL, "vxlt_0_lowtown_n_e26_0_0.vlm"},
	{0xEE58BC6BAE2C2363ULL, "vxdr_primevalentrance_1qu_0_0.vlm"},
	{0xEE5915B25FB91DA1ULL, "follower_barks6156733.cl"},
	{0xEE5926C0848FDD0AULL, "vxlt_undercitydungeon_14f_0_0.vlm"},
	{0xEE5A1AFAC0847604ULL, "vxlt_backalley_g_1pq_0_0.vlm"},
	{0xEE5A356954CACD06ULL, "728712531.ogg"},
	{0xEE5A4C6FD44D74AAULL, "44735210.ogg"},
	{0xEE5ABF72BC1C0D89ULL, "vxw_bonepitdestroyed_a3_1u9_0_0.vlm"},
	{0xEE5B3BC3312BC013ULL, "mrl150_merrill6077736.cl"},
	{0xEE5B6E7EE9751F6AULL, "llac_lt_0_lowtown_a2_n_65535_4.dds"},
	{0xEE5B9690F6C4C488ULL, "6051486_m.fxe"},
	{0xEE5BB5382031CEC5ULL, "vxca_caverns2prc_8bh_0_0.vlm"},
	{0xEE5BF31C1998A17DULL, "6169595_m.fxe"},
	{0xEE5C1D1E1F088A96ULL, "6177408_f.fxe"},
	{0xEE5CC0BFC5B1C977ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1cashsplnter.mmh"},
	{0xEE5CD6BFC5B1EEE9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_1cashsplnter.msh"},
	{0xEE5D42870557C677ULL, "vxlt_gallowstemplar_3rh_0_0.vlm"},
	{0xEE5D5B37CD280020ULL, "6205975_m.fxe"},
	{0xEE5D8C86D035D747ULL, "170940224.ogg"},
	{0xEE5DAFDB65B81816ULL, "vxw_encounter1_gz_0_0.vlm"},
	{0xEE5DE19E2BE0CFE1ULL, "vxca_caverns2prc_7vs_0_0.vlm"},
	{0xEE5E62A51E8229ACULL, "806178774.ogg"},
	{0xEE5E9DEC3D7B7B1EULL, "vxht_playershouse_n_1rf_0_0.vlm"},
	{0xEE5EECD74956D5F8ULL, "6064706_f.fxe"},
	{0xEE5F047FE54B4B99ULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.mmh"},
	{0xEE5F059C254A766BULL, "qun221_viscount6084709.cl"},
	{0xEE5FE9D0447D2198ULL, "266526247.ogg"},
	{0xEE5FED7FE7FF5871ULL, "str_cra_p.mao"},
	{0xEE5FF4CC83F21EF6ULL, "zz_dre_debug6205401.cl"},
	{0xEE605954F3D8B00EULL, "llac_lt_darktown_-1_2.dds"},
	{0xEE6102A98F9058EEULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.impa.evt"},
	{0xEE6153B59950ABF0ULL, "vxht_mansion2b_2oz_0_0.vlm"},
	{0xEE6172E77611EE1CULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower3_isabellasreturn.gad"},
	{0xEE618792C384DC7FULL, "11867592.ogg"},
	{0xEE61B4A2257F5520ULL, "mer231_brekker6075173.cl"},
	{0xEE628D499917C2A0ULL, "l_dae220ar_hightown_nt.bnk"},
	{0xEE62B6EADA489631ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_maso_0d.dds"},
	{0xEE62CD7AD4E4FAC4ULL, "hlac_ca_caverns1_mines_24424_4z.dds"},
	{0xEE62D7165400B24EULL, "325378101.ogg"},
	{0xEE62EA5369D06428ULL, "763533469.ogg"},
	{0xEE6341B10EB7F03BULL, "vxw_encounter2_b9_0_0.vlm"},
	{0xEE63FD18B199F449ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxb.phy"},
	{0xEE647F6C35C0F139ULL, "6174621_m.fxe"},
	{0xEE64BC94048EC8A3ULL, "142976110.ogg"},
	{0xEE64F0958E9D0DADULL, "vxlt_gallowstemplar_n_6sg_0_0.vlm"},
	{0xEE65984C46DE254AULL, "hlac_lt_0_lowtown_a2_-65535_4x.dds"},
	{0xEE65E5F7FEDEE419ULL, "6016024_m.fxe"},
	{0xEE65FF7CF4C3C289ULL, "rdr111dg_martin6015089.cl"},
	{0xEE6650B8C2C4E914ULL, "208875271.ogg"},
	{0xEE66F9A98F94B499ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.impa.gad"},
	{0xEE670F22446083FCULL, "dae000cr_grand_cleric6181926.cl"},
	{0xEE6716F4C7EE8E49ULL, "zz_bet_debug6123168.cl"},
	{0xEE6729B29622952CULL, "250221598.ogg"},
	{0xEE6734CDF29B316BULL, "6129756_f.fxe"},
	{0xEE673F6BAF373969ULL, "6076291_m.fxe"},
	{0xEE676585C57AC3D3ULL, "hlac_lt_gamlenshouse_n_-1_1z.dds"},
	{0xEE67CAB46B5A1B74ULL, "lgt160_wk_soph_hunter6119756.cl"},
	{0xEE67D202FAF0EB6FULL, "6118261_m.fxe"},
	{0xEE6809C731AFCD85ULL, "hlac_lt_0_lowtown_a3_n_c_196607_1z.dds"},
	{0xEE680F7D18124B0EULL, "mer221_sabin6081192.cl"},
	{0xEE68400D96F63E89ULL, "11175510.ogg"},
	{0xEE686AB9F664AB62ULL, "art\\vfx\\creatures\\fxc_drg_tswipe_c.mmh"},
	{0xEE6871E776180B1AULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower3_isabellasreturn.ani"},
	{0xEE68979D8341CEA2ULL, "6034679_f.fxe"},
	{0xEE68C578086B044AULL, "one300_nuncio.cnv"},
	{0xEE692A1C6E079BBFULL, "hlac_lt_0_lowtown_a3_n_c_1_3x.dds"},
	{0xEE6A10E7CABD4D61ULL, "ser221_gascard6133674.cl"},
	{0xEE6A9A9918F00709ULL, "vxca_cavern1_haunted_i26_0_0.vlm"},
	{0xEE6AB6DF06EAFCF1ULL, "6072152_m.fxe"},
	{0xEE6ACE3606B39C75ULL, "86775912.ogg"},
	{0xEE6B70C6386258C8ULL, "6149220_m.fxe"},
	{0xEE6B88379E0CFA2AULL, "one100dg_lia6176021.cl"},
	{0xEE6BA778086D0772ULL, "one300_nuncio.bnk"},
	{0xEE6BCD18239A7127ULL, "3310695.ogg"},
	{0xEE6BE893A1E3BD42ULL, "vxlt_0_lowtown_a2_n_c_hz4_0_0.vlm"},
	{0xEE6C31539D49B773ULL, "vxw_woundedcoast_132_0_0.vlm"},
	{0xEE6C8428A1256BF0ULL, "gen00fl_anders6136000.cl"},
	{0xEE6D69D00C29318BULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv2.evt"},
	{0xEE6DA563DEE8FE87ULL, "438188247.ogg"},
	{0xEE6E0AA98F9AF0E7ULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_def_c.impa.ani"},
	{0xEE6E11F88544DC35ULL, "6147546_m.fxe"},
	{0xEE6E7100F4802ED6ULL, "llac_lt_0_lowtown_a3_n_c_65534_6.dds"},
	{0xEE6F9AF38D564F26ULL, "vxlt_hangedman_1q3_0_0.vlm"},
	{0xEE6FC11BEABB741AULL, "97529860.ogg"},
	{0xEE7003C33DD9695CULL, "gen00fl_merrill6119643.cl"},
	{0xEE707F56F533D688ULL, "680002329.ogg"},
	{0xEE70B491C06269D4ULL, "6080568_f.fxe"},
	{0xEE714FDF68E0A14DULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc106a_0s.dds"},
	{0xEE71B38E00AE08BBULL, "vxw_encounter2_m8_0_0.vlm"},
	{0xEE7204011778AD59ULL, "589943546.ogg"},
	{0xEE73380FEA31CCCCULL, "648482178.ogg"},
	{0xEE73ADD468FEB156ULL, "vxca_caverns2_74k_0_0.vlm"},
	{0xEE73E8E4ADDCBD23ULL, "vxlt_gallowscourtyard_62k_0_0.vlm"},
	{0xEE73FCC496B6D969ULL, "character_character_p51.psh.fxo"},
	{0xEE740156B134DF22ULL, "llac_w_encounter3_-196604_0.dds"},
	{0xEE740312922B5ABDULL, "560906803.ogg"},
	{0xEE7482D00C2F7B4CULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv2.gad"},
	{0xEE74DD774781811FULL, "hlac_ht_0_hightown_a3_131070_0y.dds"},
	{0xEE7532CBF62827BAULL, "vxca_caverns2_7tg_0_0.vlm"},
	{0xEE757E04DD2A0A54ULL, "mag101tr_apostatea_cower.ncs"},
	{0xEE759BEFDCCDE96FULL, "hlac_ht_mansion_n_3368_0y.dds"},
	{0xEE75A02592185463ULL, "vxdr_primevalentrance_1kw_0_0.vlm"},
	{0xEE763095B4CC19DEULL, "vxlt_0_lowtown_docks_n_b9h_0_0.vlm"},
	{0xEE76BDCA9747CC0EULL, "6113231_f.fxe"},
	{0xEE77EC049960DF10ULL, "vxlt_0_lowtown_a3_n_c_hox_0_0.vlm"},
	{0xEE77EEB446F786CAULL, "752042460.ogg"},
	{0xEE77F6015B29CF05ULL, "6076916_m.fxe"},
	{0xEE77F60E5F16E258ULL, "gen_im_fla_bethany_03.uti"},
	{0xEE78003E493480E0ULL, "431322672.ogg"},
	{0xEE783AD7091AE89CULL, "hlac_ht_mansion2b_818_11z.dds"},
	{0xEE784E3290BF53C5ULL, "vxht_keep_n_511_0_0.vlm"},
	{0xEE789863285E0DC4ULL, "vxdr_primevaltomb_3gj_0_0.vlm"},
	{0xEE7949D7282230CBULL, "685284525.ogg"},
	{0xEE79942FEB56666DULL, "562511603.ogg"},
	{0xEE799A1BD74F4339ULL, "vxlt_0_lowtown_a2_n_c_eqw_0_0.vlm"},
	{0xEE79E2A32A587148ULL, "6142320_m.fxe"},
	{0xEE79E358997E8C37ULL, "837631771.ogg"},
	{0xEE7A5A7093D9D5C2ULL, "vxlt_0_lowtown_a2_n_c_hu5_0_0.vlm"},
	{0xEE7A961023985D4EULL, "419734316.ogg"},
	{0xEE7AF2E120AD09B6ULL, "6060901_m.fxe"},
	{0xEE7B159AB51215F4ULL, "hlac_lt_0_lowtown_docks_-131075_2x.dds"},
	{0xEE7B6A2A18EF6286ULL, "wm_city_day.map"},
	{0xEE7B819203AA1FE1ULL, "6092572_m.fxe"},
	{0xEE7B81D00C35984AULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_attkv2.ani"},
	{0xEE7BA9D6C55799D3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\8gr_vsttowera_1.mao"},
	{0xEE7C75A5137D3A82ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst.gad"},
	{0xEE7CCD66D5B5FBB7ULL, "vxht_chantry_26z_0_0.vlm"},
	{0xEE7D1326DF0A66D6ULL, "6048169_m.fxe"},
	{0xEE7D6C47F9685153ULL, "vxdr_primevaltomb_6qo_0_1.vlm"},
	{0xEE7D77DB2434CECFULL, "672132516.ogg"},
	{0xEE7D7C35550A8CC5ULL, "mag110cr_abomination.utc"},
	{0xEE7DC99B217F9667ULL, "art\\vfx\\environment\\vfx_lyrium_ash.mmh"},
	{0xEE7E376F7D3CDCF0ULL, "vxht_mansion2b_2ex_0_0.vlm"},
	{0xEE7F012BC90E0590ULL, "503510956.ogg"},
	{0xEE7F356FA1659C03ULL, "zz_isa_debug6175244.cl"},
	{0xEE7F9A548FECC9FEULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.mmh"},
	{0xEE7FEC86FB3A927EULL, "lgt110_orz_harrowmont6176029.cl"},
	{0xEE804E821DC6E793ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\basedualwield\\blendtree\\assassin_overlay.anb"},
	{0xEE805607BC62447AULL, "vxlt_0_lowtown_docks_cv3_0_0.vlm"},
	{0xEE80D3F3C5ADC784ULL, "qun221_petrice6098706.cl"},
	{0xEE81027AFA469606ULL, "627153752.ogg"},
	{0xEE8108460A486093ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_har_visa_0d.dds"},
	{0xEE81D295319D6912ULL, "vxlt_gallowstemplar_n_5cc_0_0.vlm"},
	{0xEE81D606297283B9ULL, "19684814.ogg"},
	{0xEE8206D9A8E7C7E6ULL, "dae100_alistair_drunk6135153.cl"},
	{0xEE82948CDC9E92F8ULL, "clf260_harley6069565.cl"},
	{0xEE82AD3517CD93BFULL, "249505484.ogg"},
	{0xEE830210AF5CAA36ULL, "mag160_feynriel6016541.cl"},
	{0xEE8315866E9A2190ULL, "6167112_m.fxe"},
	{0xEE8322B9E1E93499ULL, "vxdr_primevalentrance_1l0_0_0.vlm"},
	{0xEE8398A513839579ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst.evt"},
	{0xEE83AB76D42FA28CULL, "6059781_m.fxe"},
	{0xEE83DD46573F8748ULL, "411362942.ogg"},
	{0xEE841609ABF388A0ULL, "hlac_ht_playershouse_n_855_27z.dds"},
	{0xEE8417F5B3399888ULL, "vxht_mansion_n_6bb_0_0.vlm"},
	{0xEE846CEC0787E36AULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishpile.mao"},
	{0xEE8470EC0787E9B9ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishpile.mmh"},
	{0xEE847AEC0787FAA7ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishpile.msh"},
	{0xEE849D9315B397C0ULL, "6088725_m.fxe"},
	{0xEE84A272A6C2325EULL, "vxlt_0_lowtown_docks_n_b6i_0_0.vlm"},
	{0xEE860A7D728CEAC2ULL, "vxca_caverns2prc_6sz_0_0.vlm"},
	{0xEE86899D80ED4674ULL, "86876368.ogg"},
	{0xEE86D6C5D442EFDBULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_exit.ani"},
	{0xEE86EE97649238F5ULL, "186014850.ogg"},
	{0xEE873A29833DA52DULL, "vxht_keep_3b5_0_0.vlm"},
	{0xEE879E5FE0381AD0ULL, "6062779_f.fxe"},
	{0xEE88017FE56E4B8EULL, "art\\vfx\\environment\\level_fx\\lightbeams\\vfx_duststream.anb"},
	{0xEE884B55F90356E7ULL, "659163506.ogg"},
	{0xEE8857DD91DE4BEBULL, "dae000_flemeth6000753.cl"},
	{0xEE888ED96FF3F9D7ULL, "lt_0_lowtown_docks_a3_231_0.mmh"},
	{0xEE8892D96FF400E4ULL, "lt_0_lowtown_docks_a3_231_0.mao"},
	{0xEE88A4D96FF41F49ULL, "lt_0_lowtown_docks_a3_231_0.msh"},
	{0xEE88ABAB37CB96DCULL, "298338973.ogg"},
	{0xEE88CC4D85CFEE42ULL, "vxlt_0_lowtown_a2_n_c_hp6_0_0.vlm"},
	{0xEE89150D8135E873ULL, "vxw_woundedcoast_194_0_0.vlm"},
	{0xEE891E789A1CF870ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookopen.mao"},
	{0xEE8922789A1CFE83ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookopen.mmh"},
	{0xEE8930789A1D16DDULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookopen.msh"},
	{0xEE893EBF9142CB3DULL, "art\\characters\\weapons\\greatswords\\w_gsw_001a.phy"},
	{0xEE89F5CC268C442EULL, "6179430_m.fxe"},
	{0xEE8A53251C9B972BULL, "gr_01.mao"},
	{0xEE8A8F2C9AEBF924ULL, "vxw_encounter1_g8_0_0.vlm"},
	{0xEE8A9BCC95891C6FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_qunari.gad"},
	{0xEE8AA2AC2855E4E2ULL, "vxw_encounter2_db_0_0.vlm"},
	{0xEE8AAD70C543EEB7ULL, "6124280_m.fxe"},
	{0xEE8AB58BB4B8998BULL, "6174371_m.fxe"},
	{0xEE8AB8FD2211CAA2ULL, "zz_vault_debug6203169.cl"},
	{0xEE8B3F43C7AC1437ULL, "vxht_chantry_23u_0_0.vlm"},
	{0xEE8B7162A6393FE7ULL, "759376738.ogg"},
	{0xEE8BA2BAAF10BF07ULL, "vxlt_darktown_2vm_0_0.vlm"},
	{0xEE8BC335FB7E72CDULL, "hm_refugee_4.mor"},
	{0xEE8CB8BB84723B09ULL, "hlac_lt_0_lowtown_a2_65537_6x.dds"},
	{0xEE8D2696690958D5ULL, "art\\vfx\\materials\\chunky\\vfx_c_crater.mao"},
	{0xEE8D4E644FDFB02EULL, "6107202_m.fxe"},
	{0xEE8D5D04DD3E6381ULL, "mag101tr_apostatea_cower.utt"},
	{0xEE8D79548FF8CEBDULL, "art\\vfx\\combat\\vfx_glb_charm_crust_c.anb"},
	{0xEE8D94D641759EFEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v1.ani"},
	{0xEE8E9BD304DF2904ULL, "6102777_m.fxe"},
	{0xEE8EC92EFCCBCA0DULL, "497992012.ogg"},
	{0xEE8EF187106CB289ULL, "87711797.ogg"},
	{0xEE8EF5004639D79AULL, "6102573_m.fxe"},
	{0xEE8F4FD8409507F3ULL, "6043971_f.fxe"},
	{0xEE8FB19AA6BF3BCDULL, "lt_0_lowtown_docks_a3_n_159_0.mmh"},
	{0xEE8FB59AA6BF421EULL, "lt_0_lowtown_docks_a3_n_159_0.mao"},
	{0xEE8FC39AA6BF5A53ULL, "lt_0_lowtown_docks_a3_n_159_0.msh"},
	{0xEE8FD192E80E67EBULL, "1072921065.ogg"},
	{0xEE8FE38AA672FA13ULL, "6130084_f.fxe"},
	{0xEE9043323B60B664ULL, "zz_one_debug6099935.cl"},
	{0xEE90BAEEF62AC50EULL, "6173155_m.fxe"},
	{0xEE90C8980E2623C0ULL, "821558966.ogg"},
	{0xEE90E37822B57C33ULL, "107634444.ogg"},
	{0xEE90E618B1C0BB6EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxb.mmh"},
	{0xEE90F218B1C0CF05ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxb.mao"},
	{0xEE910018B1C0E75CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0dd_basicxb.msh"},
	{0xEE91573947CC96C1ULL, "170648582.ogg"},
	{0xEE9169B9F687AE09ULL, "art\\vfx\\creatures\\fxc_drg_tswipe_c.anb"},
	{0xEE919496D3DF4D19ULL, "vxdr_primevalentrance_1a7_0_0.vlm"},
	{0xEE91A0A5138FE068ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst.ani"},
	{0xEE91B0CC958F5E1CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_qunari.evt"},
	{0xEE9229E201987797ULL, "vxht_keep_n_7ae_0_0.vlm"},
	{0xEE92493D15AB869AULL, "w_bonepitdestroyed_a3_40_0.mmh"},
	{0xEE924D3D15AB8DD1ULL, "w_bonepitdestroyed_a3_40_0.mao"},
	{0xEE925B3D15ABA5E0ULL, "w_bonepitdestroyed_a3_40_0.msh"},
	{0xEE92C871F980424DULL, "t3_cor_b01.tnt"},
	{0xEE930F7007A9B040ULL, "6083625_m.fxe"},
	{0xEE9326B05CE45390ULL, "cli_army_ds_palace.gda"},
	{0xEE93404065E6E6B9ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_5_ian.gad"},
	{0xEE9359268254A0FCULL, "vxht_0_hightown_a2_c8r_0_0.vlm"},
	{0xEE93EE4F67D080CBULL, "gen00fl_anders6125157.cl"},
	{0xEE942D2705BA32A4ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hangbar.phy"},
	{0xEE94B43CF56C1B4FULL, "6133349_m.fxe"},
	{0xEE94B6980371240AULL, "w_encounter3_167_0.mao"},
	{0xEE94BA9803712B59ULL, "w_encounter3_167_0.mmh"},
	{0xEE94C49803713C47ULL, "w_encounter3_167_0.msh"},
	{0xEE94DEC5D44F3AD2ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_exit.evt"},
	{0xEE9580A3274B4412ULL, "114678171.ogg"},
	{0xEE95B10C66FFAE9FULL, "vxlt_gallowsprison_n_5wf_0_0.vlm"},
	{0xEE95DE90A3254EDFULL, "vxht_keep_n_73g_0_0.vlm"},
	{0xEE95FA291366943CULL, "genip_altar.utp"},
	{0xEE96564ADA9A4EB5ULL, "vxdr_0_deeproads_9p6_0_0.vlm"},
	{0xEE968DF8233F7D34ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb102a_0t.dds"},
	{0xEE970D725C506C1CULL, "hlac_ht_keep_3732_3y.dds"},
	{0xEE973EEADE17E645ULL, "6140971_f.fxe"},
	{0xEE974D79F7CFE450ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyn.mao"},
	{0xEE974DA015BFFB32ULL, "6166913_m.fxe"},
	{0xEE975179F7CFEB63ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyn.mmh"},
	{0xEE975F79F7D002BDULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyn.msh"},
	{0xEE9793A5D23FE44FULL, "80550230.ogg"},
	{0xEE97A2A15C9DD76EULL, "and111_lirene6084387.cl"},
	{0xEE9914891A4BFD62ULL, "vxw_encounter2_oc_0_0.vlm"},
	{0xEE998B875D4638E7ULL, "545974829.ogg"},
	{0xEE9A1497A106D787ULL, "vxlt_darktown_2yl_0_0.vlm"},
	{0xEE9A33BF91512BA9ULL, "art\\characters\\weapons\\greatswords\\w_gsw_001a.mao"},
	{0xEE9A3FBF91514042ULL, "art\\characters\\weapons\\greatswords\\w_gsw_001a.mmh"},
	{0xEE9A41BF915143F8ULL, "art\\characters\\weapons\\greatswords\\w_gsw_001a.msh"},
	{0xEE9A514065ED2207ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_transactionmerchant_5_ian.ani"},
	{0xEE9A82CF6354CA3CULL, "604070517.ogg"},
	{0xEE9A949B5B34F804ULL, "art\\vfx\\combat\\vfx_m_bld_sac_imp_c.mmh"},
	{0xEE9B8D3A544984BEULL, "seb321_leliana6180972.cl"},
	{0xEE9BACD6418204CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v1.evt"},
	{0xEE9BBB4C91ABCD3DULL, "vxw_blightlands_11x_0_0.vlm"},
	{0xEE9BE2320F2E8032ULL, "vxlt_hangedman_2vu_0_0.vlm"},
	{0xEE9BE74ED2843949ULL, "vxlt_darktown_57s_0_0.vlm"},
	{0xEE9BFDC5D4558E65ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\lean_back_wall_arm_cross_exit.gad"},
	{0xEE9C1494582BC724ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml101a_0d.dds"},
	{0xEE9C5D910177AB24ULL, "381581954.ogg"},
	{0xEE9C9F77C6E98F71ULL, "llac_ca_caverns2prc_5802_7.dds"},
	{0xEE9D1F0A41F41CF9ULL, "mag100_karras6034922.cl"},
	{0xEE9D73F5CD5C4830ULL, "vxca_caverns1_fn5_0_0.vlm"},
	{0xEE9DAA6368E97C45ULL, "6139730_m.fxe"},
	{0xEE9DC0806332FA53ULL, "700228901.ogg"},
	{0xEE9E3BEAE8D2AFB6ULL, "322588882.ogg"},
	{0xEE9E4BBA3FA1B02FULL, "zz_followers_banter6159036.cl"},
	{0xEE9E864FBA0A3B2DULL, "character_character_p84.psh.fxo.dx11"},
	{0xEE9EB8CC9599F635ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\dragging_qunari.ani"},
	{0xEE9ECE22A1DC738EULL, "gen00fl_isabela6075231.cl"},
	{0xEE9F24CC8BC6C1E0ULL, "vxht_mansion2_3gs_0_0.vlm"},
	{0xEE9F997E2E7BA69EULL, "464039909.ogg"},
	{0xEE9FB7C75B7F9FCCULL, "300849354.ogg"},
	{0xEEA05970C04B0D41ULL, "mrl310_party_comments6142753.cl"},
	{0xEEA08F30B81BD38CULL, "6053581_m.fxe"},
	{0xEEA1213126F9A58EULL, "dae200_bran6094784.cl"},
	{0xEEA136CCC731CE3DULL, "rdr111cr_martin.utc"},
	{0xEEA1527AB5DBB004ULL, "43734230.ogg"},
	{0xEEA176893F120032ULL, "hlac_lt_gallowstemplar_n_2382_3y.dds"},
	{0xEEA1814CF99FC8C0ULL, "6086925_m.fxe"},
	{0xEEA1875CAF14F259ULL, "w_bonepit_142_0.msh"},
	{0xEEA1915CAF150347ULL, "w_bonepit_142_0.mmh"},
	{0xEEA1955CAF150994ULL, "w_bonepit_142_0.mao"},
	{0xEEA1978AB51AF943ULL, "vxht_mansion_n_6f1_0_0.vlm"},
	{0xEEA1CEEC7D756B29ULL, "6149261_m.fxe"},
	{0xEEA2ADD641882578ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_trip_v1.gad"},
	{0xEEA2B1595DAAE2BBULL, "hlac_lt_backalley_n_0_8x.dds"},
	{0xEEA33B0CC9BB569BULL, "hlac_dr_dungeon_3216_3z.dds"},
	{0xEEA34687952FBD22ULL, "6174561_m.fxe"},
	{0xEEA41F9816BA72C6ULL, "t3_wdg_gn1.tnt"},
	{0xEEA43F5137799AA7ULL, "vxlt_0_lowtown_d06_0_0.vlm"},
	{0xEEA4A6EF30E00BC4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ff_manhole_1.mao"},
	{0xEEA4C023B54E6F44ULL, "vxht_keep_n_5wr_0_0.vlm"},
	{0xEEA4E34C99F7C221ULL, "588153418.ogg"},
	{0xEEA5985552AB1313ULL, "hlac_ht_keep_n_4519_5y.dds"},
	{0xEEA5A47E375446DDULL, "vxht_bartrand_n_6im_0_0.vlm"},
	{0xEEA5B0BB33EA5ED9ULL, "176510302.ogg"},
	{0xEEA5CC962123E8FBULL, "zz_followers_banter6158934.cl"},
	{0xEEA6521EFE97B823ULL, "vxca_caverns1_i1x_0_0.vlm"},
	{0xEEA72AE5336FA41BULL, "954142202.ogg"},
	{0xEEA7389128A04D4AULL, "281114991.ogg"},
	{0xEEA786660C4215E2ULL, "vxw_encounter2_jl_0_0.vlm"},
	{0xEEA78EE0F9FFCF29ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_0.pwk"},
	{0xEEA7CD78C6E4C569ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipgroup.mao"},
	{0xEEA7D978C6E4D902ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipgroup.mmh"},
	{0xEEA7DB78C6E4DCB8ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_turnipgroup.msh"},
	{0xEEA84910AD385619ULL, "6020488_f.fxe"},
	{0xEEA85279F7DE5F88ULL, "art\\characters\\playercharacter\\dwarfmale\\arm_hvyn.phy"},
	{0xEEA8E8B63666D2FCULL, "llac_ca_caverns1_mines_24422_2.dds"},
	{0xEEA967B5E61A0C20ULL, "llac_lt_0_lowtown_a2_n_c_131070_6.dds"},
	{0xEEA983589BC38026ULL, "426828927.ogg"},
	{0xEEA9AE2C0DC56974ULL, "vxht_mansion2_1t8_0_0.vlm"},
	{0xEEA9DA395166F24DULL, "mcr371_hostile_aveline6082196.cl"},
	{0xEEA9E1466A71BDC9ULL, "follower_barks6153428.cl"},
	{0xEEAA592BC47A51C9ULL, "vxlt_darktown_58r_0_0.vlm"},
	{0xEEAAAEE5FA308884ULL, "lt_0_lowtown_docks_242_0.msh"},
	{0xEEAAC0E5FA30A7EDULL, "lt_0_lowtown_docks_242_0.mao"},
	{0xEEAAC4E5FA30AE26ULL, "lt_0_lowtown_docks_242_0.mmh"},
	{0xEEAB7C53E4F538F3ULL, "507359873.ogg"},
	{0xEEAC437C90B55185ULL, "48674705.ogg"},
	{0xEEACBCC2C5D7AEE5ULL, "da2terrain_terrain_tools_p1.psh.fxo.dx11"},
	{0xEEAD47364E727150ULL, "921165736.ogg"},
	{0xEEAD4E2EB94A246BULL, "ser221cr_hostage.utc"},
	{0xEEAE44709DBA00C5ULL, "115348825.ogg"},
	{0xEEAEA91E9B745F34ULL, "vxdr_0_deeproads_ex7_0_0.vlm"},
	{0xEEAF23E2A88AB25CULL, "dae300_alistair_king6134971.cl"},
	{0xEEAF6CA09A575AE1ULL, "vxw_sundermount_32p_0_1.vlm"},
	{0xEEAFC2027617ED44ULL, "vxht_brothel_n_4vo_0_0.vlm"},
	{0xEEAFE9D520151386ULL, "185284686.ogg"},
	{0xEEAFF4DC38D4ECE9ULL, "fen310_fenris6118671.cl"},
	{0xEEB00967A71111C3ULL, "vxht_mansion_n_6i0_0_0.vlm"},
	{0xEEB043ACC7098E4BULL, "464209040.ogg"},
	{0xEEB1CB4E51ED81EEULL, "hlac_lt_0_lowtown_docks_131071_5x.dds"},
	{0xEEB293D2887753B2ULL, "vxht_0_hightown_a2_bty_0_0.vlm"},
	{0xEEB2AE661EF60C12ULL, "664630186.ogg"},
	{0xEEB2EAD35BEEC5F2ULL, "676865565.ogg"},
	{0xEEB2EEEB0AB46924ULL, "389850797.ogg"},
	{0xEEB2F16BA4EF489DULL, "follower_banter6163532.cl"},
	{0xEEB3349B7EEBC2DFULL, "mag121_idunna6092300.cl"},
	{0xEEB3615F314FAC3AULL, "936546478.ogg"},
	{0xEEB39B8E9F223E00ULL, "qun131_warrens_thug6019126.cl"},
	{0xEEB3F504685044B1ULL, "dre000_varric6034645.cl"},
	{0xEEB4165B294A5F5DULL, "vxht_bartrand_n_6rl_0_0.vlm"},
	{0xEEB4241EBD8D1BEDULL, "6188110_m.fxe"},
	{0xEEB4A528669BF1EFULL, "vxca_caverns1_mines_gjf_0_0.vlm"},
	{0xEEB5C2C8C6C9BAA7ULL, "fx_flesh.mao"},
	{0xEEB5E46E3B4C74C7ULL, "ht_0_hightown_a2_43_0.rml"},
	{0xEEB5F044F93E6005ULL, "6165326_m.fxe"},
	{0xEEB6670D6341668FULL, "vxlt_0_lowtown_a2_n_c_gx8_0_0.vlm"},
	{0xEEB6B59371090637ULL, "262114414.ogg"},
	{0xEEB786CD40166264ULL, "tools_sphere.msh"},
	{0xEEB79CCD40168706ULL, "tools_sphere.mmh"},
	{0xEEB7BEEE4831294BULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_stretch_neck_left.evt"},
	{0xEEB7FA08E49ED0DAULL, "vxlt_blackemporium_vw_0_0.vlm"},
	{0xEEB81AAD52B10797ULL, "gen00fl_bethany6102772.cl"},
	{0xEEB849A96CEF6B3AULL, "6178378_m.fxe"},
	{0xEEB87A475FB8AB4EULL, "vxlt_0_lowtown_a2_n_c_gbs_0_0.vlm"},
	{0xEEB8C5EBF31AB132ULL, "vxlt_hangedman_2hw_0_0.vlm"},
	{0xEEB8F2A37C48DCEFULL, "85347354.ogg"},
	{0xEEB92B63C8103BD6ULL, "vxht_0_hightown_9b4_0_0.vlm"},
	{0xEEB96A92F40E8E78ULL, "dm_miner_1.mor"},
	{0xEEB96B6D3CC97B3CULL, "795972928.ogg"},
	{0xEEB9AD8CD910F94FULL, "one300_zevran6076499.cl"},
	{0xEEBA42CB54ECE66AULL, "fx_smokeghost_flipbook.dds"},
	{0xEEBB4284E2F58A3DULL, "150258619.ogg"},
	{0xEEBB7270EEE655C7ULL, "llac_lt_backalley_n_65537_2.dds"},
	{0xEEBB738FEDB5227EULL, "vxdr_0_deeproads_9g0_0_0.vlm"},
	{0xEEBBD49314B07AC0ULL, "59509628.ogg"},
	{0xEEBC0FCA37A71244ULL, "dae000cr_grand_cleric6173659.cl"},
	{0xEEBC1B995C9BC550ULL, "vxht_bartrand_n_5vf_0_0.vlm"},
	{0xEEBC6937472EC428ULL, "zz_followers_banter6158889.cl"},
	{0xEEBCA453C6F562E0ULL, "dae100_bran6058943.cl"},
	{0xEEBD24EB3A33948BULL, "vxw_blightlands_su_0_0.vlm"},
	{0xEEBD6FB4C3E373F2ULL, "hlac_w_woundedcoast_8_0y.dds"},
	{0xEEBE493D15D0C1B1ULL, "w_bonepitdestroyed_a3_40_0.rml"},
	{0xEEBE8B9C6BD77A0BULL, "mag170_hugh6088849.cl"},
	{0xEEBEC06530F9D3D0ULL, "pol.cut"},
	{0xEEBED7EE4837720CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_stretch_neck_left.gad"},
	{0xEEBF18AFE8CAE14EULL, "6055578_m.fxe"},
	{0xEEBF5C28013AFA13ULL, "423113434.ogg"},
	{0xEEBFF3AE9C73BF93ULL, "6119200_m.fxe"},
	{0xEEC105AF7A6D6C32ULL, "vxht_0_hightown_a2_byz_0_0.vlm"},
	{0xEEC191DB43F207A0ULL, "6027457_f.fxe"},
	{0xEEC193CD401ECCD1ULL, "tools_sphere.phy"},
	{0xEEC214AE08FA7684ULL, "6116181_m.fxe"},
	{0xEEC2635332E4D772ULL, "ht_0_hightown_n_44_0_lc.mmh"},
	{0xEEC2979B5B564FFBULL, "art\\vfx\\combat\\vfx_m_bld_sac_imp_c.anb"},
	{0xEEC2CED7E4D06C78ULL, "gen00fl_varric6104951.cl"},
	{0xEEC46A1FF02E46E2ULL, "vxw_encounter2_pn_0_0.vlm"},
	{0xEEC49F1794B3056DULL, "6054871_m.fxe"},
	{0xEEC4CF79F0081CACULL, "772989004.ogg"},
	{0xEEC4D6EE483BDC0AULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\custom\\sword_on_shoulder_twitch_stretch_neck_left.ani"},
	{0xEEC51A70F53D390FULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose4left.anb"},
	{0xEEC553C95785ECF4ULL, "6147802_m.fxe"},
	{0xEEC569B91B64817BULL, "qcr000_arishok6059984.cl"},
	{0xEEC580172A8B8DBCULL, "vxdr_primevalentrance_1li_0_0.vlm"},
	{0xEEC5A0989163E5AEULL, "753319125.ogg"},
	{0xEEC6342C643761FFULL, "6202605_m.fxe"},
	{0xEEC737C8E510C9B2ULL, "vxlt_hangedman_2ep_0_0.vlm"},
	{0xEEC778224ECF2E6CULL, "con_ico_choice.dds"},
	{0xEEC7A1D02BFE20ABULL, "vxlt_shop_j5_0_0.vlm"},
	{0xEEC7AF3F5DC114ADULL, "267042233.ogg"},
	{0xEEC8AD1575549A84ULL, "hlac_lt_darktown_-1_15z.dds"},
	{0xEEC97AAB63418C6DULL, "326327332.ogg"},
	{0xEEC99F328F0DC592ULL, "hlac_w_encounter3_0_7z.dds"},
	{0xEEC9D2BB1E57BB40ULL, "llac_dr_slavers_3207_2.dds"},
	{0xEEC9DF99DBDA64BCULL, "228865526.ogg"},
	{0xEECA1F15FC2AE8B2ULL, "vxlt_0_lowtown_docks_9re_0_0.vlm"},
	{0xEECA64853EBBDEACULL, "122680869.ogg"},
	{0xEECA8D764E91DDD0ULL, "vxht_bartrand_n_5yg_0_0.vlm"},
	{0xEECAA209F23E1AA9ULL, "vxlt_0_lowtown_a2_n_f66_0_1.vlm"},
	{0xEECB2AC43A4A73F3ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4bluelyrium.phy"},
	{0xEECB392AC04D60F1ULL, "vxlt_0_lowtown_a2_n_c_emk_0_0.vlm"},
	{0xEECB96C82C29AD0BULL, "vxw_blightlands_tt_0_0.vlm"},
	{0xEECBB4C7E19794D4ULL, "hlac_ca_caverns1_24423_7y.dds"},
	{0xEECC10724DF87D2DULL, "hairalpha_hair_p0.vsh.fxo.dx11"},
	{0xEECC42981B451797ULL, "hlac_lt_0_lowtown_-2_5z.dds"},
	{0xEECC5EB939339142ULL, "561685214.ogg"},
	{0xEECC7982CD2CBE41ULL, "one100dg_vanard6054290.cl"},
	{0xEECCB9083E6B6D0BULL, "vxw_sundermount_2gi_0_0.vlm"},
	{0xEECD1DF4C52510EBULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_rf.evt"},
	{0xEECD895F15CBDCCDULL, "vxca_caverns2_8gf_0_0.vlm"},
	{0xEECDF5F1A5380822ULL, "lt_0_lowtown_a2_n_c_51_0_lc.mmh"},
	{0xEECE9BD11C0C7D96ULL, "w_bonepitmining_a2_120_0.phy"},
	{0xEECEBE8788A7AFB3ULL, "193324690.ogg"},
	{0xEECF59F9AA0FC961ULL, "pro000_refugee_angry6052168.cl"},
	{0xEECF9B3BF003ADBCULL, "6095068_m.fxe"},
	{0xEECF9DD580795BBAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masg_0s.dds"},
	{0xEECFB5244A86669FULL, "vxw_encounter2_camp_7r_0_0.vlm"},
	{0xEED065805A284F0AULL, "352687458.ogg"},
	{0xEED08B12D7DF902CULL, "6086360_f.fxe"},
	{0xEED12FC5A79034DEULL, "hlac_w_encounter2_196611_0x.dds"},
	{0xEED1B4DFBB74D5F2ULL, "vxht_chantry_27p_0_0.vlm"},
	{0xEED1C18428C4B932ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\textures\\windowglowcombined_da.dds"},
	{0xEED1F38054282342ULL, "vxlt_ruinedhouse_n_ie_0_0.vlm"},
	{0xEED22D1527028B42ULL, "vxht_0_hightown_a3_9zi_0_0.vlm"},
	{0xEED28AD03391C41DULL, "follower_banter6172291.cl"},
	{0xEED2B8D8263638BBULL, "ser270_templar6127493.cl"},
	{0xEED2D8084D428FE3ULL, "vxlt_warehouse_n_128_0_0.vlm"},
	{0xEED3037BF608856FULL, "one200_gold_followers6101132.cl"},
	{0xEED310F486A91DEDULL, "6051771_m.fxe"},
	{0xEED336F4C529A6ACULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_rf.gad"},
	{0xEED3F1F41C81A63CULL, "vxdr_primevalentrance_1an_0_0.vlm"},
	{0xEED44A8E09C50F45ULL, "blood_wave_l2.mao"},
	{0xEED49C7191BE004CULL, "vxlt_warehouse_11m_0_0.vlm"},
	{0xEED4DBBB5DCAFCB8ULL, "6064953_f.fxe"},
	{0xEED4E54ACA3A10CBULL, "6126843_f.fxe"},
	{0xEED522C270018738ULL, "vxdr_0_deeproads_9el_0_0.vlm"},
	{0xEED563314E978569ULL, "619763559.ogg"},
	{0xEED5A9A5D706E232ULL, "vxlt_hangedman_2bq_0_0.vlm"},
	{0xEED5AFD7B763ABCEULL, "189701736.ogg"},
	{0xEED616C4061DA1C0ULL, "aip_raiderwarrior2w_base.gda"},
	{0xEED63EAEE68F07B2ULL, "6075099_m.fxe"},
	{0xEED6798DCF5F229AULL, "93956967.ogg"},
	{0xEED71B4ED70A60ACULL, "324815739.ogg"},
	{0xEED7466A313EECA1ULL, "hlac_ht_0_hightown_a3_0_3x.dds"},
	{0xEED77DEAAF31A2C4ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\walkf.evt"},
	{0xEED795243601A192ULL, "6049636_m.fxe"},
	{0xEED80D8813509D91ULL, "qun230cr_guild_contact6068877.cl"},
	{0xEED880A9B6D315EEULL, "w_bonepitmining_a2_24_0_lc.mmh"},
	{0xEED884D11C14AB37ULL, "w_bonepitmining_a2_120_0.msh"},
	{0xEED88F10803AF6C0ULL, "196873011.ogg"},
	{0xEED896D11C14C99AULL, "w_bonepitmining_a2_120_0.mao"},
	{0xEED89AD11C14D0A9ULL, "w_bonepitmining_a2_120_0.mmh"},
	{0xEED95C6994823100ULL, "6070339_m.fxe"},
	{0xEED96A88943AACD1ULL, "vxht_bartrand_n_677_0_0.vlm"},
	{0xEED98650C12452D6ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_attractionwave.evt"},
	{0xEED98EE8562D40F7ULL, "6044178_f.fxe"},
	{0xEED9F52BB0174D26ULL, "166477535.ogg"},
	{0xEEDA35F4C52FC3AAULL, "art\\characters\\creatures\\mubari\\animation\\rn_p_rf.ani"},
	{0xEEDA5D88961401C0ULL, "vxht_playershouse_n_1qp_0_1.vlm"},
	{0xEEDB2485F7C5B123ULL, "gen00fl_varric6021116.cl"},
	{0xEEDB2E74D84C9E89ULL, "6162283_f.fxe"},
	{0xEEDB3097969F8096ULL, "vxlt_0_lowtown_f16_0_0.vlm"},
	{0xEEDBE4E71B625D1AULL, "gen00fl_sebastian6180002.cl"},
	{0xEEDBF8659BB4913AULL, "261781768.ogg"},
	{0xEEDBFB3C07C1F54DULL, "vxca_caverns2_8be_0_0.vlm"},
	{0xEEDC15FE5A7DEBB9ULL, "962796565.ogg"},
	{0xEEDC783CF1B7F74FULL, "hlac_lt_0_lowtown_docks_n_-131074_7z.dds"},
	{0xEEDDE9695E599D32ULL, "vxht_0_hightown_a2_bgd_0_0.vlm"},
	{0xEEDE87977D26D144ULL, "vxht_keep_n_5kv_0_0.vlm"},
	{0xEEDE993C3673395CULL, "llac_lt_0_lowtown_docks_-131074_5.dds"},
	{0xEEDEFAC1475E1899ULL, "mrl150_merrill6093695.cl"},
	{0xEEDF1E8F58BCC904ULL, "6085344_m.fxe"},
	{0xEEDF617933EC4F95ULL, "hlac_lt_0_lowtown_65537_2z.dds"},
	{0xEEDFE15DC004E10AULL, "vxlt_0_lowtown_a3_n_c_hfe_0_0.vlm"},
	{0xEEE04601807BFC5CULL, "hlac_ca_caverns1_24421_11z.dds"},
	{0xEEE07D50C12A6261ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_attractionwave.gad"},
	{0xEEE0C916A2DD895DULL, "6042040_f.fxe"},
	{0xEEE0EE721B74F165ULL, "gen_im_npc_wep_dua_arib.uti"},
	{0xEEE12252AB10D7FDULL, "6131907_m.fxe"},
	{0xEEE1E97770A30FE7ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.cess.gad"},
	{0xEEE27ACF291DF1D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_extreme_v1.evt"},
	{0xEEE27D9DE5290FA6ULL, "hla_ht_0_hightown_a3_-2_0x.dds"},
	{0xEEE2A39C4B2CB545ULL, "follower_banter6163284.cl"},
	{0xEEE2A696A1447AB1ULL, "165174252.ogg"},
	{0xEEE2AFDD3F1C8BACULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.impa.ani"},
	{0xEEE2B0DD9520390AULL, "vxw_encounter2_n8_0_0.vlm"},
	{0xEEE2D425AAE36709ULL, "gen00fl_varric6093641.cl"},
	{0xEEE30820AEC68F03ULL, "6163762_m.fxe"},
	{0xEEE30E4E83B418CCULL, "vxlt_warehouse_14b_0_0.vlm"},
	{0xEEE3B00B1994EAACULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_pound.ani"},
	{0xEEE3C3CAD49644CFULL, "vxlt_gallowstemplar_f_5z1_0_0.vlm"},
	{0xEEE3FEE20904EF39ULL, "zz_vrc_debug6064274.cl"},
	{0xEEE4E03DF6173A1BULL, "gen00fl_bethany6101540.cl"},
	{0xEEE51A34E3020DCAULL, "6125981_m.fxe"},
	{0xEEE53DCE50D9D1CAULL, "6067671_f.fxe"},
	{0xEEE5EE9F7D925F78ULL, "vxlt_gallowscourtyard_516_0_0.vlm"},
	{0xEEE67E50C12ECF0FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_rct_attractionwave.ani"},
	{0xEEE70C0F71DFD17CULL, "9278314.ogg"},
	{0xEEE76831D333B979ULL, "6000874_m.fxe"},
	{0xEEE7DE7770A768F4ULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.cess.evt"},
	{0xEEE7FBFE21D31BF9ULL, "gen00fl_fenris6115381.cl"},
	{0xEEE8D6AF8E4DB789ULL, "vxht_chantry_n_35g_0_0.vlm"},
	{0xEEE8E92102A7B1FEULL, "517042078.ogg"},
	{0xEEE9665E04DAFF97ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_stand.phy"},
	{0xEEE973CF29240414ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_extreme_v1.gad"},
	{0xEEE9958B0AA65FCAULL, "395584857.ogg"},
	{0xEEE99CC222579E0BULL, "vxw_sundermount_2mk_0_0.vlm"},
	{0xEEE9BCDD3F22C0EEULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.impa.gad"},
	{0xEEE9C8BE620C3024ULL, "zz_fen_debug6215266.cl"},
	{0xEEE9CFB4853BC60DULL, "532063112.ogg"},
	{0xEEE9D4ED1B972FF0ULL, "6087139_m.fxe"},
	{0xEEEA1485E196195EULL, "6131579_m.fxe"},
	{0xEEEA2806A91AB698ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_lightbeam1.dds"},
	{0xEEEA67FD864CB039ULL, "hlac_ht_keep_n_4519_4x.dds"},
	{0xEEEA74410A6B6384ULL, "ave200_donnic6111209.cl"},
	{0xEEEA9581FEDBF69DULL, "6170399_m.fxe"},
	{0xEEEABD0B199B1FEEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_pound.gad"},
	{0xEEEB53176E437698ULL, "llac_lt_undercityencounter_-1_6.dds"},
	{0xEEEB5E5E7F599D94ULL, "100219690.ogg"},
	{0xEEEB960170BC25C3ULL, "qainsf1.arl"},
	{0xEEECD973CC1C6D84ULL, "qun131_taskmaster6059036.cl"},
	{0xEEED73BE42890C7AULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsrock.phy"},
	{0xEEED9D2B727ECFD8ULL, "gen_im_wep_mag_sta_act2_02.uti"},
	{0xEEEE241C52D4775CULL, "6080450_f.fxe"},
	{0xEEEE92D06635F0EDULL, "hlac_w_blightlands_0_9y.dds"},
	{0xEEEE9C207E08BF09ULL, "gencr_slaver_2h.utc"},
	{0xEEEEE278FC292541ULL, "gen00fl_anders6113019.cl"},
	{0xEEEEF97AD8E17BEBULL, "art\\characters\\creatures\\arh_abm_rev\\textures\\cn_bdy_arh_0d.dds"},
	{0xEEEF3865BF2A3046ULL, "6111668_f.fxe"},
	{0xEEF029629DD5A781ULL, "5128207.ogg"},
	{0xEEF0610DABC097E1ULL, "hlac_lt_hangedman_-1_6z.dds"},
	{0xEEF06B88D009ED24ULL, "gen_im_cft_rec_arn_06.uti"},
	{0xEEF07F1ABE87940AULL, "mcr370_meredith6084441.cl"},
	{0xEEF082CF292A3CF2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_spirit_extreme_v1.ani"},
	{0xEEF0B04C6273393CULL, "gen_im_trs_qun_lgt_cht_r0.uti"},
	{0xEEF0B064C67AE25FULL, "vxca_caverns2_7w7_0_0.vlm"},
	{0xEEF0B7DD3F28D6CDULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.impa.evt"},
	{0xEEF0D40EF4CF24B4ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml101a_0t.dds"},
	{0xEEF0E1C535213176ULL, "6060354_m.fxe"},
	{0xEEF11BB132E1DC1CULL, "vxca_cavern1_haunted_fm1_0_0.vlm"},
	{0xEEF122BA8716518AULL, "vxw_encounter2_i9_0_0.vlm"},
	{0xEEF173450304A4E9ULL, "6147678_m.fxe"},
	{0xEEF1B80B19A135CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_pound.evt"},
	{0xEEF1BAF998DA33AFULL, "one100dg_nabil6063821.cl"},
	{0xEEF1BC0784EF78E2ULL, "793692347.ogg"},
	{0xEEF290C4CB40289EULL, "follower_banter6169834.cl"},
	{0xEEF2B29B82D274C5ULL, "916402274.ogg"},
	{0xEEF2EB6F0473BF65ULL, "122147569.ogg"},
	{0xEEF32964C8A6FF90ULL, "vxht_0_hightown_a3_9bc_0_0.vlm"},
	{0xEEF3BB1CA915EA71ULL, "lt_backalley_n_0_0_lc.mmh"},
	{0xEEF40AFF949E2762ULL, "plt_car000pt_main.nss"},
	{0xEEF433B5A0B7E252ULL, "vxlt_undercityserialkiller_1n9_0_0.vlm"},
	{0xEEF5D67770B3982DULL, "art\\vfx\\combat\\vfx_w_off_stagger_loop_c.cess.ani"},
	{0xEEF634DF1731D1E3ULL, "vxht_0_hightown_a2_9i4_0_0.vlm"},
	{0xEEF7488C8043D009ULL, "vxht_chantry_n_32f_0_0.vlm"},
	{0xEEF74A07DC945213ULL, "hlac_lt_0_lowtown_a2_65536_9x.dds"},
	{0xEEF7AF7990DD64E7ULL, "6107434_m.fxe"},
	{0xEEF7CCF58CC4CCCAULL, "zz_fol_debug6056995.cl"},
	{0xEEF7E63FF5071717ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushring.mmh"},
	{0xEEF7EA3FF5071E24ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushring.mao"},
	{0xEEF7FC3FF5073C89ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushring.msh"},
	{0xEEF80B3EEA22EA8AULL, "vxlt_undercitydungeon_1my_0_0.vlm"},
	{0xEEF80E9F144DB68BULL, "vxw_sundermount_2vj_0_0.vlm"},
	{0xEEF8DEF5EBAE264DULL, "vxca_caverns2_8tc_0_0.vlm"},
	{0xEEF9971CB3549FFCULL, "152458337.ogg"},
	{0xEEFA33EB53B4786BULL, "597771753.ogg"},
	{0xEEFAA85072A0FF7EULL, "107090377.ogg"},
	{0xEEFB025A1B1398ADULL, "lt_0_lowtown_docks_229_0.msh"},
	{0xEEFB105A1B13B0E0ULL, "lt_0_lowtown_docks_229_0.mao"},
	{0xEEFB145A1B13B733ULL, "lt_0_lowtown_docks_229_0.mmh"},
	{0xEEFB2D4F1EBFB419ULL, "27031752.ogg"},
	{0xEEFB3394FFA8ECBEULL, "fen110_anso6094961.cl"},
	{0xEEFBEE1CB1C79322ULL, "530894201.ogg"},
	{0xEEFC108FCEB0C817ULL, "hlac_lt_blackemporium_1055_21x.dds"},
	{0xEEFC10A609092B1CULL, "hlac_dr_slavers_3216_0z.dds"},
	{0xEEFC4490FA9D216FULL, "6156357_m.fxe"},
	{0xEEFC4F3E3BDEC7AAULL, "736297237.ogg"},
	{0xEEFDAF123E520A24ULL, "hlac_lt_0_lowtown_a2_n_-65537_6y.dds"},
	{0xEEFDBC2A0D8870A9ULL, "vxlt_0_lowtown_a2_n_c_fep_0_0.vlm"},
	{0xEEFDD97E2A7D91A4ULL, "zz_rogue4.cut"},
	{0xEEFE84EFE24ADD62ULL, "vxw_blightlands_6g_0_0.vlm"},
	{0xEEFEC6538B5CE40CULL, "zz_mag_debug6051363.cl"},
	{0xEEFF534F06CBDB45ULL, "468461796.ogg"},
	{0xEEFF70FF7BF03375ULL, "gen00fl_sebastian6178365.cl"},
	{0xEEFFAF010BE9A4FDULL, "lgt120_orz_bln_official6140175.cl"},
	{0xEEFFBBFC22FD9FE9ULL, "hlac_lt_0_lowtown_n_-1_0y.dds"},
	{0xEF000669DAFF07B6ULL, "hightown_fem_noble2_mob6169489.cl"},
	{0xEF00127C9040FA46ULL, "vxlt_gallowscourtyard_6z2_0_0.vlm"},
	{0xEF001A8F971C9DA4ULL, "ht_0_hightown_51_0_ll.mmh"},
	{0xEF005C9C5CB3E6CBULL, "hlac_lt_foundry_n_320_20x.dds"},
	{0xEF008DC1F050379FULL, "338306460.ogg"},
	{0xEF0097D3E15123B0ULL, "6207528_m.fxe"},
	{0xEF00E20ADFDD09FEULL, "spellfire_red.mao"},
	{0xEF010951A978F6FBULL, "follower_barks6153268.cl"},
	{0xEF0114C8BBDF9281ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredacti.ani"},
	{0xEF015D58D3C006FDULL, "ch1alphem_vl.psh.fxo"},
	{0xEF01AD7F4396DC88ULL, "328476794.ogg"},
	{0xEF01DF712D3E3E79ULL, "6034932_m.fxe"},
	{0xEF01E1AB11910F35ULL, "hlac_lt_0_lowtown_docks_a3_n_-65537_2z.dds"},
	{0xEF024C0AB3ECA841ULL, "vxlt_gallowscourtyard_621_0_0.vlm"},
	{0xEF024D3BD908E99EULL, "200087799.ogg"},
	{0xEF0277411AA4975DULL, "gen00fl_bethany6101227.cl"},
	{0xEF028D29679864F9ULL, "376820869.ogg"},
	{0xEF035D5127A56D31ULL, "llac_lt_undercityencounter_-1_45.dds"},
	{0xEF037E75C78E0EC4ULL, "343175656.ogg"},
	{0xEF03823E5E0A116EULL, "vxlt_0_lowtown_a2_n_c_df0_0_0.vlm"},
	{0xEF0396F4BE463031ULL, "gen00fl_merrill6106525.cl"},
	{0xEF04B21FE4DF7C9DULL, "6117918_m.fxe"},
	{0xEF05009E8825C2F1ULL, "vxlt_0_lowtown_a2_n_c_eqo_0_0.vlm"},
	{0xEF05243608350106ULL, "sse_ss_sebastian.gda"},
	{0xEF055F0F09947E1BULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1egg.msh"},
	{0xEF05710F09949D66ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1egg.mao"},
	{0xEF057D0F0994B105ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1egg.mmh"},
	{0xEF06464229F55E30ULL, "293406992.ogg"},
	{0xEF0686B21FCAFB2BULL, "vxlt_gallowscourtyard_700_0_0.vlm"},
	{0xEF06C690D1D979A5ULL, "lgt351ar_shw_finale_cave.are"},
	{0xEF06F72821D13721ULL, "vxdr_0_deeproads_epv_0_0.vlm"},
	{0xEF071FC8BBE41123ULL, "art\\cinematics\\_crit path\\mage crisis\\meredith fight\\cs_meredith_meredacti.gad"},
	{0xEF07E42EE57FEE89ULL, "6205369_m.fxe"},
	{0xEF080FE8CBD361A4ULL, "vxlt_0_lowtown_f1t_0_0.vlm"},
	{0xEF081810384AB36DULL, "6177013_m.fxe"},
	{0xEF086674D8EE18D4ULL, "775614501.ogg"},
	{0xEF0902FBE1D6927BULL, "vxw_bonepitdestroyed_a3_15c_0_0.vlm"},
	{0xEF098ACB8D84351DULL, "240407859.ogg"},
	{0xEF09ACCD75CB63DBULL, "866434008.ogg"},
	{0xEF09DBCB546E01B1ULL, "gen00fl_anders6125821.cl"},
	{0xEF0B597AD9F7DCADULL, "vxca_caverns2prc_9sr_0_0.vlm"},
	{0xEF0B65D84A685FB6ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_act_c.impa.ani"},
	{0xEF0C137A91ED01DAULL, "lit_aoe_distor_l2.mao"},
	{0xEF0D4F9088C877BBULL, "l_seb221ar_harimann_estate.bnk"},
	{0xEF0D8270A6348BFBULL, "6059672_m.fxe"},
	{0xEF0DCA65D33C69CBULL, "vxca_caverns2_78v_0_0.vlm"},
	{0xEF0F1961AA788E4FULL, "vxlt_gallowstemplar_f_5k2_0_0.vlm"},
	{0xEF0F5ACF6285A78CULL, "vxht_keep_n_3fw_0_0.vlm"},
	{0xEF106FAA8B0334AFULL, "w_bonepit_224_0.mao"},
	{0xEF107BAA8B03494CULL, "w_bonepit_224_0.mmh"},
	{0xEF1081AA8B03533EULL, "w_bonepit_224_0.msh"},
	{0xEF1095B21EE06624ULL, "521040846.ogg"},
	{0xEF10BDE7A5E2C0C1ULL, "vxlt_gallowscourtyard_690_0_0.vlm"},
	{0xEF10C790D1E1CF8BULL, "lgt351ar_shw_finale_cave.ncs"},
	{0xEF1106056C819F4CULL, "6126414_m.fxe"},
	{0xEF11792FFF8E0BBAULL, "6018739_m.fxe"},
	{0xEF11D7B69F44C870ULL, "839724573.ogg"},
	{0xEF11FA9D0394F126ULL, "ave221_brennan6110477.cl"},
	{0xEF122C05B3FEA9B8ULL, "31612379.ogg"},
	{0xEF12C99EE7AAC39BULL, "hlac_lt_0_lowtown_0_0x.dds"},
	{0xEF12D5999ACF9D5BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medf_0d.dds"},
	{0xEF139485E1358FCDULL, "w_bonepit_179_0.mmh"},
	{0xEF139885E135961EULL, "w_bonepit_179_0.mao"},
	{0xEF13A685E135AE53ULL, "w_bonepit_179_0.msh"},
	{0xEF13F72A6D3DDF6BULL, "257544716.ogg"},
	{0xEF143494E8989452ULL, "pro000_refugee_male6052109.cl"},
	{0xEF14A924F344D378ULL, "mcr371_orsino6083351.cl"},
	{0xEF14F88F11C113ABULL, "vxlt_gallowscourtyard_753_0_0.vlm"},
	{0xEF15833D143C152EULL, "gen00fl_varric6079958.cl"},
	{0xEF15849FD03AF43BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gallows.mmh"},
	{0xEF15909FD03B08D8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gallows.mao"},
	{0xEF15A29FD03B2725ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gallows.msh"},
	{0xEF15B405E8C40CB4ULL, "6118547_m.fxe"},
	{0xEF162C74F1122F35ULL, "vxdr_primevaltomb_6wz_0_0.vlm"},
	{0xEF17169421882E18ULL, "vxlt_gallowstemplar_n_39m_0_0.vlm"},
	{0xEF1756AD3296C200ULL, "6163509_m.fxe"},
	{0xEF17676005CBBDEBULL, "dre000_stroud6036731.cl"},
	{0xEF1778F2620639AEULL, "492779811.ogg"},
	{0xEF17C790D1E7EED2ULL, "lgt351ar_shw_finale_cave.lst"},
	{0xEF1860F340EB45C6ULL, "33880855.ogg"},
	{0xEF186D31FBC42501ULL, "static2.vsh.fxo"},
	{0xEF18FDEB0D73C560ULL, "918148409.ogg"},
	{0xEF196DD84A74AAA7ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_act_c.impa.evt"},
	{0xEF1972BFC4487D4CULL, "dae000ip_crowd_lm2_talk6.utp"},
	{0xEF1A3F01FA50B902ULL, "412968327.ogg"},
	{0xEF1A44EA027B674AULL, "6125713_f.fxe"},
	{0xEF1ABEB2627E2C1EULL, "hlac_lt_0_lowtown_docks_65537_0x.dds"},
	{0xEF1ACBBAF9747A56ULL, "238530157.ogg"},
	{0xEF1B76E2B8906804ULL, "mer221_sabin6045002.cl"},
	{0xEF1B8312A9C25EE1ULL, "hlac_ht_mansion2_818_50z.dds"},
	{0xEF1C3429B720F601ULL, "llac_lt_0_lowtown_docks_-65539_0.dds"},
	{0xEF1C3B61E9BF0230ULL, "vxlt_foundry_n_z6_0_0.vlm"},
	{0xEF1C7F9B49CE4D20ULL, "444499284.ogg"},
	{0xEF1CBCCAAED140E2ULL, "425354841.ogg"},
	{0xEF1D08CB67D2933EULL, "vxlt_house_fl_0_0.vlm"},
	{0xEF1D092E2113A1EEULL, "vxw_sundermount_24u_0_0.vlm"},
	{0xEF1D44DFEB309AC7ULL, "6023963_m.fxe"},
	{0xEF1D7AAEE6D1F0ACULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_tunnel.msh"},
	{0xEF1D80AEE6D1FA9EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_tunnel.mmh"},
	{0xEF1D8CAEE6D20F35ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ww_tunnel.mao"},
	{0xEF1D99D923A0B346ULL, "653516066.ogg"},
	{0xEF1DB53C6B8D1AD7ULL, "llac_ht_0_hightown_n_-2_0.dds"},
	{0xEF1E1364B370889FULL, "190144806.ogg"},
	{0xEF1E1CE79031B82EULL, "vxlt_gallowstemplar_f_47b_0_0.vlm"},
	{0xEF1E5A89B851273BULL, "dae_monster_aoe_duration.ncs"},
	{0xEF1E7118151EF2A3ULL, "dae000_flemeth6000960.cl"},
	{0xEF1E97F12A94235BULL, "mag311_nyssa6042387.cl"},
	{0xEF1ECDD3459BF171ULL, "6105386_m.fxe"},
	{0xEF1F14F83C49F445ULL, "dae000_hate_follower_react6167481.cl"},
	{0xEF1F612FAF2325CDULL, "hlac_lt_0_lowtown_a2_n_c_-2_6z.dds"},
	{0xEF1F8ED84A794EF0ULL, "art\\vfx\\combat\\vfx_m_bld_ctrol_act_c.impa.gad"},
	{0xEF2015ED4C324A12ULL, "ts_slavers_acc_props.gda"},
	{0xEF2086F9CFF9FA44ULL, "w_sundermount_65_0_lc.mmh"},
	{0xEF20AFB5F6674FA4ULL, "6135797_m.fxe"},
	{0xEF21637DC923555FULL, "follower_barks6153332.cl"},
	{0xEF21A29BD680D0BAULL, "ser221_gamlen6135003.cl"},
	{0xEF220C1F884F8F6EULL, "507920085.ogg"},
	{0xEF22C3857D2B4230ULL, "vxca_cavern1_haunted_j1w_0_2.vlm"},
	{0xEF22CEE5D51DF9C3ULL, "754029406.ogg"},
	{0xEF235F5A6F5E9C2BULL, "501899412.ogg"},
	{0xEF237424DF1E982AULL, "mind_01.mao"},
	{0xEF239246DC5A9EBDULL, "llac_ht_0_hightown_65537_2.dds"},
	{0xEF23F7FF94C72AE5ULL, "plt_car000pt_main.plo"},
	{0xEF242505FB7BAA46ULL, "one110_meeran6089507.cl"},
	{0xEF2425E2DABA2534ULL, "6117447_m.fxe"},
	{0xEF244D0BCA6239B6ULL, "lgt000cr_merch_turn_in_6.utc"},
	{0xEF2456BC6F8ED3B8ULL, "vxlt_gallowstemplar_n_6s6_0_0.vlm"},
	{0xEF2483BDBC48A606ULL, "vxlt_gallowstemplar_n_6sx_0_0.vlm"},
	{0xEF251A7E6FF4B3C3ULL, "213228122.ogg"},
	{0xEF2536722651AE9CULL, "art\\vfx\\materials\\vfx_rain.mao"},
	{0xEF25514D86C7911BULL, "vxca_caverns1_mines_hu8_0_0.vlm"},
	{0xEF2586295923AA32ULL, "follower_barks6154994.cl"},
	{0xEF263BC79DDCCD69ULL, "hlac_lt_0_lowtown_a3_n_c_65537_6x.dds"},
	{0xEF271CA173B2699FULL, "6116704_m.fxe"},
	{0xEF27F70F48A793B2ULL, "zz_followers_banter6157918.cl"},
	{0xEF281216BC0E197AULL, "art\\characters\\playercharacter\\humanfemale\\bdy_mer.phy"},
	{0xEF2854C9ADC38608ULL, "108373208.ogg"},
	{0xEF286E45B0FF0E1EULL, "464521323.ogg"},
	{0xEF2887948174D9C5ULL, "6076862_m.fxe"},
	{0xEF291677D4A809EDULL, "373125213.ogg"},
	{0xEF293C8E9163CFA2ULL, "6062824_m.fxe"},
	{0xEF29BC9680433A38ULL, "vxht_keep_3xf_0_0.vlm"},
	{0xEF29BD20355AD863ULL, "vxlt_backalley_n_1ex_0_0.vlm"},
	{0xEF29E12F10A8FA2BULL, "vxht_0_hightown_a3_bwh_0_0.vlm"},
	{0xEF29E2B72C93D1A2ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_dg106a_0t.dds"},
	{0xEF2A319F45C8A4ECULL, "352877783.ogg"},
	{0xEF2A97B516313383ULL, "6075003_m.fxe"},
	{0xEF2AF6A207A28667ULL, "vxlt_foundry_n_v9_0_0.vlm"},
	{0xEF2B189845DBF6B3ULL, "gen_im_arm_boo_lgt_act3_02.uti"},
	{0xEF2B7B0B1309BA6EULL, "vxw_sundermount_29v_0_0.vlm"},
	{0xEF2B9FB370DD2544ULL, "117887912.ogg"},
	{0xEF2BA2E14113A700ULL, "667225677.ogg"},
	{0xEF2BB24463E262A2ULL, "dae000_merchant_generic_hm6172676.cl"},
	{0xEF2C11F6E1F4926BULL, "6156245_m.fxe"},
	{0xEF2C24CCA22EC6F6ULL, "6126621_m.fxe"},
	{0xEF2C2C637D501592ULL, "6206136_m.fxe"},
	{0xEF2C6ABB65C941CEULL, "dae000_flemeth6001353.cl"},
	{0xEF2D23C43A9D3268ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4bluelyrium.mmh"},
	{0xEF2D2FC43A9D460BULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4bluelyrium.mao"},
	{0xEF2D3A77D6CB5FE0ULL, "vxlt_warehouse_sj_0_0.vlm"},
	{0xEF2D41C43A9D6492ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_4bluelyrium.msh"},
	{0xEF2D7E21055A7B78ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_torturerack_n.dds"},
	{0xEF2D849DEA9F69ADULL, "llac_lt_gallowsprison_n_-1_2.dds"},
	{0xEF2DC0B835F6CD52ULL, "pro000_merchant6051798.cl"},
	{0xEF2E4EABB3BC39E7ULL, "vxlt_gallowstemplar_f_5r1_0_0.vlm"},
	{0xEF2ED8CFFF966727ULL, "vxht_0_hightown_a2_c5r_0_0.vlm"},
	{0xEF2FC304A93EA962ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qunariarishok.anb"},
	{0xEF3003DBE08C2C94ULL, "plt_cod_hst_kirk_darktown.plo"},
	{0xEF303B5E0516B9EFULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_stand.mao"},
	{0xEF30475E0516CD8CULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_stand.mmh"},
	{0xEF304D5E0516D87EULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_stand.msh"},
	{0xEF305D9B9ED07359ULL, "663184522.ogg"},
	{0xEF30A000EE3A8261ULL, "art\\characters\\creatures\\desiredemon\\animation\\dblw_lp_v1.ani"},
	{0xEF31B9A0F3DFF0F7ULL, "6115121_m.fxe"},
	{0xEF324CBEF7B380A1ULL, "vxdr_0_deeproads_eck_0_0.vlm"},
	{0xEF326CDAE637BC33ULL, "354010060.ogg"},
	{0xEF3324D2FA382AE6ULL, "character_character_p218.psh.fxo.dx11.assembly"},
	{0xEF3355A8FFA56764ULL, "vxht_0_hightown_n_c0q_0_0.vlm"},
	{0xEF3375E0037169F5ULL, "1069570124.ogg"},
	{0xEF33A93D1AAF9CB6ULL, "339365593.ogg"},
	{0xEF33FA4E05745F18ULL, "vxlt_gallowstemplar_n_33k_0_0.vlm"},
	{0xEF343974E0C7ED27ULL, "dae000_foreshadow6179557.cl"},
	{0xEF34A6D4ECEE3B37ULL, "148794781.ogg"},
	{0xEF3505EDA31694ECULL, "539967588.ogg"},
	{0xEF351020CC6A78B1ULL, "llac_ca_caverns2prc_5802_56.dds"},
	{0xEF35356FD8CD1093ULL, "hlac_lt_0_lowtown_n_0_7x.dds"},
	{0xEF354E2E89391E19ULL, "mrl210_merrill6117419.cl"},
	{0xEF358A4F0F2896C8ULL, "6078058_m.fxe"},
	{0xEF3614EC943BE843ULL, "384840839.ogg"},
	{0xEF367582615B4352ULL, "6071082_m.fxe"},
	{0xEF36DD85935DF04EULL, "hlac_ht_keep_3732_8z.dds"},
	{0xEF36DDA9FE1499BEULL, "follower_banter6160752.cl"},
	{0xEF36E0D544C4ABA2ULL, "vxdr_primevaltomb_5be_0_0.vlm"},
	{0xEF3771A9B4BBD4C1ULL, "dae000ip_crowd_em4_talk4.utp"},
	{0xEF37AB00EE40B303ULL, "art\\characters\\creatures\\desiredemon\\animation\\dblw_lp_v1.gad"},
	{0xEF37AE6B8359E822ULL, "6069924_m.fxe"},
	{0xEF385101C4A22E51ULL, "ts_circleofmagi_acc_props.gda"},
	{0xEF38530C029F12ABULL, "vxht_0_hightown_a3_brk_0_0.vlm"},
	{0xEF39416B480CC838ULL, "llac_lt_hangedman_-1_9.dds"},
	{0xEF3956FCAA05AFA4ULL, "6219196_m.fxe"},
	{0xEF39BD7433DF1EB1ULL, "vrc221cr_guards_fake.utc"},
	{0xEF3A747538BEC66EULL, "and100_anders6095974.cl"},
	{0xEF3AA51EF5D8FB31ULL, "6016323_m.fxe"},
	{0xEF3ACF5C3C5225FDULL, "vxlt_0_lowtown_a2_f59_0_0.vlm"},
	{0xEF3AFA83078B82DCULL, "626753137.ogg"},
	{0xEF3B0EB1E7C3D22DULL, "6016315_f.fxe"},
	{0xEF3BCDF98A0500E9ULL, "vxw_sundermount_273_0_0.vlm"},
	{0xEF3BEDA33C235FBBULL, "vxw_encounter2_76_0_0.vlm"},
	{0xEF3BEE36A9430093ULL, "zz_isa_debug6133228.cl"},
	{0xEF3CB72AAAD6825BULL, "vxdr_primevaltomb_6vs_0_0.vlm"},
	{0xEF3CCF0E6AC68737ULL, "265932127.ogg"},
	{0xEF3D066E8FB866FBULL, "vxca_caverns2prc_9py_0_0.vlm"},
	{0xEF3D3CE204FE4706ULL, "6150358_m.fxe"},
	{0xEF3D4AACF18C7FA7ULL, "vxht_0_hightown_a2_c0m_0_0.vlm"},
	{0xEF3D758AF659BC73ULL, "lt_0_lowtown_n_36_0_lc.mmh"},
	{0xEF3D8C3D8FE277C0ULL, "zz_followers_banter6158892.cl"},
	{0xEF3DAE3A71DAAA62ULL, "lt_gallowscourtyard_16_0.rml"},
	{0xEF3E0600662EEBDDULL, "hlac_w_sundermount_196610_0y.dds"},
	{0xEF3EA58AABCD620DULL, "hlac_lt_0_lowtown_a3_n_c_-262144_4z.dds"},
	{0xEF3EACAA2E818F4FULL, "face_face_p11.psh.fxo.dx11.assembly"},
	{0xEF3ECE702D014BEFULL, "gen00fl_anders6096480.cl"},
	{0xEF3F32123F3C3723ULL, "w_encounter3_9_0.rml"},
	{0xEF40227CDF7ED5D9ULL, "hlac_ht_keep_4519_9x.dds"},
	{0xEF4151E260CFC848ULL, "186120701.ogg"},
	{0xEF42ECCE1A1EDD34ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetree.phy"},
	{0xEF42EF72B52ED1D0ULL, "48222413.ogg"},
	{0xEF44246E46DCF4A9ULL, "zz_approval_debug6003716.cl"},
	{0xEF445214A92B1E93ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\1wi_basic_1.mao"},
	{0xEF44909E5C500339ULL, "vxlt_warehouse_18o_0_0.vlm"},
	{0xEF44BFCB27D2DE83ULL, "vxw_woundedcoast_1hr_0_1.vlm"},
	{0xEF44E9F8933AABF2ULL, "6122449_m.fxe"},
	{0xEF44F2E5DA017C1CULL, "vxlt_0_lowtown_h8u_0_0.vlm"},
	{0xEF44F40D5396C0ECULL, "vxlt_0_lowtown_e3a_0_0.vlm"},
	{0xEF4508D769F8C28AULL, "89027293.ogg"},
	{0xEF4543C74B2599F2ULL, "vxht_chantry_2ox_0_0.vlm"},
	{0xEF4552D5D265B5E3ULL, "art\\vfx\\cinematic\\vfx_ship_front_p.dura.ani"},
	{0xEF456906E3A81273ULL, "vxlt_gallowstemplar_6qh_0_0.vlm"},
	{0xEF4581BCDE80876CULL, "510485575.ogg"},
	{0xEF4592FFEE74D83AULL, "218487110.ogg"},
	{0xEF45BE9389B1B28DULL, "6185711_m.fxe"},
	{0xEF462C239E23B716ULL, "6095981_m.fxe"},
	{0xEF4667D1C92D4F00ULL, "vxlt_darktown_4zt_0_0.vlm"},
	{0xEF46A0DA19470963ULL, "vxlt_backalley_n_1kz_0_0.vlm"},
	{0xEF46C8EC7A10F144ULL, "150218427.ogg"},
	{0xEF46E6F2949001D8ULL, "159279401.ogg"},
	{0xEF46EA51C35E19D0ULL, "lgt110_orz_harr_carta6109049.cl"},
	{0xEF4741AB2D214A7FULL, "art\\characters\\creatures\\shriek\\textures\\cn_bdy_ems_0n.dds"},
	{0xEF4745AA39AE9EBAULL, "167578532.ogg"},
	{0xEF47C8D99BFBC824ULL, "6212096_m.fxe"},
	{0xEF47D63770086187ULL, "565328842.ogg"},
	{0xEF47E991583F5A3DULL, "lgt150im_ntb_wr_ingrdnt_2.uti"},
	{0xEF47ECD3E0EDE262ULL, "hla_ht_0_hightown_n_-65537_0z.dds"},
	{0xEF486CBE42D5ABA3ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsrock.msh"},
	{0xEF487ABE42D5C3FDULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsrock.mmh"},
	{0xEF487EBE42D5CA0EULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_02ruinsrock.mao"},
	{0xEF489F1FD2135942ULL, "plt_cod_ltr_lother_barlin.plo"},
	{0xEF48BBCB0440A964ULL, "hlac_lt_warehouse_n_65537_5y.dds"},
	{0xEF48D748EB96D099ULL, "cod_ltr_broker_hunter_0.utp"},
	{0xEF491F0D63AF1E95ULL, "655268342.ogg"},
	{0xEF495459D49F3CA0ULL, "vxlt_0_lowtown_a2_n_c_epa_0_0.vlm"},
	{0xEF498927F7946A8EULL, "6082314_m.fxe"},
	{0xEF49C25ECEB5556AULL, "sib100_father_contact6108967.cl"},
	{0xEF49D93AA7302634ULL, "443769722.ogg"},
	{0xEF4A1E31BAB790E0ULL, "vxlt_warehouse_yh_0_0.vlm"},
	{0xEF4A3FD67BFB1969ULL, "vxw_sundermount_282_0_0.vlm"},
	{0xEF4AB16B1BAB7520ULL, "vxw_sundermount_30g_0_1.vlm"},
	{0xEF4AC37F6234F5DAULL, "vxlt_undercityserialkiller_1t7_0_0.vlm"},
	{0xEF4AF211CCED27A2ULL, "llac_w_sundermount_393211_0.dds"},
	{0xEF4B0D3BBD5A69D1ULL, "6090888_m.fxe"},
	{0xEF4B4387781C5006ULL, "152797164.ogg"},
	{0xEF4B5399E8DFCC78ULL, "6136697_m.fxe"},
	{0xEF4B828B102A89EAULL, "6116056_f.fxe"},
	{0xEF4BF185BA0A25C6ULL, "lgt310_mis_burn_murder26152762.cl"},
	{0xEF4C0E8334B440E0ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9wo_stala1a_1.mao"},
	{0xEF4C1AD14735EBADULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushline.msh"},
	{0xEF4C28D1473603E0ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushline.mao"},
	{0xEF4C2CD147360A33ULL, "art\\levels\\races\\natural\\kirkcave\\foliage\\gra_mushline.mmh"},
	{0xEF4C50B3AE78421EULL, "vxlt_undercityserialkiller_1mx_0_0.vlm"},
	{0xEF4C69BADDCD73ADULL, "zz_dre_debug6025523.cl"},
	{0xEF4C6F9411227FB1ULL, "6049671_m.fxe"},
	{0xEF4CBA6F053B52F8ULL, "vxlt_0_lowtown_a2_e17_0_0.vlm"},
	{0xEF4CC9CE1A26F631ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetree.msh"},
	{0xEF4CD17EAF69D9EBULL, "gen_im_trs_krk_grd_junk2_02.uti"},
	{0xEF4CD7CE1A270E4CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetree.mao"},
	{0xEF4CE3CE1A2722EFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\tre_alienagetree.mmh"},
	{0xEF4D0E6DC8300066ULL, "6113926_m.fxe"},
	{0xEF4D0F8EF1F1B167ULL, "6053744_m.fxe"},
	{0xEF4DE67834A07E5EULL, "6024989_m.fxe"},
	{0xEF4E0FBBA978A02EULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd05a.msh"},
	{0xEF4E29BBA978CC1CULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd05a.mmh"},
	{0xEF4EF74F81DFE1D9ULL, "lgt220_mag_tmplr_templar.bnk"},
	{0xEF4F2A5A2811F04AULL, "fen210_hadriana6110412.cl"},
	{0xEF4F3078DB9FB1A1ULL, "vxdr_0_deeproads_eii_0_0.vlm"},
	{0xEF4F3D7E92CCFF40ULL, "vxht_chantry_355_0_0.vlm"},
	{0xEF4FAA0392769D12ULL, "art\\levels\\races\\natural\\default\\terrainmaterials\\_textures\\def_rock2_d.dds"},
	{0xEF4FBA0AED9C2E9DULL, "llac_ca_caverns2_5802_15.dds"},
	{0xEF4FBBF6E6DEDD99ULL, "dae150_chandan6148651.cl"},
	{0xEF4FE3A9E8B007ADULL, "hlac_ca_cavern1_haunted_24425_1y.dds"},
	{0xEF5020BEC600C8D0ULL, "ht_0_hightownexplosion_n_29_0_lc.mmh"},
	{0xEF503B8AB280D378ULL, "gen00fl_fenris6113887.cl"},
	{0xEF5075C709819F89ULL, "hlac_w_encounter2_camp_327685_0z.dds"},
	{0xEF50842DA82AA23DULL, "6028919_m.fxe"},
	{0xEF50CFC8ADFA7ADCULL, "6174062_m.fxe"},
	{0xEF50F34615782EBBULL, "360742409.ogg"},
	{0xEF51073AEEC703B1ULL, "lt_0_lowtown_docks_a3_n_241_0.mmh"},
	{0xEF51133AEEC71852ULL, "lt_0_lowtown_docks_a3_n_241_0.mao"},
	{0xEF51213AEEC7306FULL, "lt_0_lowtown_docks_a3_n_241_0.msh"},
	{0xEF5152C65F9505BCULL, "6151632_m.fxe"},
	{0xEF5166750F54F251ULL, "dae000_bodhan6129109.cl"},
	{0xEF51769438FDB09CULL, "character_character_p186.psh.fxo.dx11.assembly"},
	{0xEF51AEDCA9985816ULL, "one300_zevran6093505.cl"},
	{0xEF51F81F140AD08FULL, "6048432_m.fxe"},
	{0xEF51FF5490354459ULL, "ser191_emeric6120562.cl"},
	{0xEF520019BD3AF4BDULL, "6107281_m.fxe"},
	{0xEF520218EC9FA291ULL, "lgt110_gxa_vgk_silver_order6117006.cl"},
	{0xEF524E7576850C95ULL, "6171910_m.fxe"},
	{0xEF5257573AE28B10ULL, "hlac_ca_caverns2prc_5802_56y.dds"},
	{0xEF525BA330FB16A3ULL, "vxlt_0_lowtown_epc_0_0.vlm"},
	{0xEF52625B7B7E9FFAULL, "905889760.ogg"},
	{0xEF526AD5D27068DAULL, "art\\vfx\\cinematic\\vfx_ship_front_p.dura.evt"},
	{0xEF52927AF30BEB40ULL, "qun221_viscount6185119.cl"},
	{0xEF52E8FFAF4CD844ULL, "262410447.ogg"},
	{0xEF52FFA13F3D310FULL, "6001094_f.fxe"},
	{0xEF53194F81E40439ULL, "lgt220_mag_tmplr_templar.cnv"},
	{0xEF5331A819C8F703ULL, "vxw_woundedcoast_1ss_0_1.vlm"},
	{0xEF536A4809AA53D1ULL, "vxht_0_hightown_a3_b8k_0_0.vlm"},
	{0xEF542A6C70A5BC36ULL, "and111_lirene6097116.cl"},
	{0xEF55122D562583B8ULL, "vxht_keep_3ie_0_0.vlm"},
	{0xEF5563E3BE6B4FE4ULL, "vxht_keep_n_499_0_0.vlm"},
	{0xEF561A9EAEC034BEULL, "vxht_keep_n_1ab_0_0.vlm"},
	{0xEF56267332553B08ULL, "mag170_thrask6054046.cl"},
	{0xEF5632BC1F414969ULL, "follower_banter6167172.cl"},
	{0xEF56A6865FA5B6B0ULL, "dae000_meredith6092583.cl"},
	{0xEF56BD0F2A7CBB9EULL, "6164047_m.fxe"},
	{0xEF56D58B33EFC2ABULL, "vxht_keep_53f_0_0.vlm"},
	{0xEF576093D0940064ULL, "6162275_merrill_interjects.cut"},
	{0xEF57B0CC380944ECULL, "6104375_m.fxe"},
	{0xEF57C636C6955520ULL, "vxlt_0_lowtown_a2_n_c_eun_0_0.vlm"},
	{0xEF57D96577F3F505ULL, "6144323_m.fxe"},
	{0xEF57F27024770CD0ULL, "dae000cr_grand_cleric6173682.cl"},
	{0xEF588E099E9E6A23ULL, "636446699.ogg"},
	{0xEF58949ABB177463ULL, "6151892_f.fxe"},
	{0xEF58A08E055F60FBULL, "gen00fl_merrill6116148.cl"},
	{0xEF58D241E67A3484ULL, "146789838.ogg"},
	{0xEF595318E59E9874ULL, "6092591_m.fxe"},
	{0xEF5969D5D27685CDULL, "art\\vfx\\cinematic\\vfx_ship_front_p.dura.gad"},
	{0xEF5974D5C16DAED7ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_woodholexa.msh"},
	{0xEF597F18AF508251ULL, "6095788_m.fxe"},
	{0xEF5986D5C16DCD3AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_woodholexa.mao"},
	{0xEF598AD5C16DD449ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\3ww_woodholexa.mmh"},
	{0xEF59A442899E8367ULL, "vxlt_gallowstemplar_f_5c2_0_0.vlm"},
	{0xEF59C1398AC059B4ULL, "mag291_torpor6026969.cl"},
	{0xEF59CC031244BC2CULL, "hlac_dr_slavers_3217_3x.dds"},
	{0xEF5A5699AE818B72ULL, "hlac_dr_0_deeproads_17039_0x.dds"},
	{0xEF5A763C75F93557ULL, "vxht_bartrand_n_64d_0_0.vlm"},
	{0xEF5B273A787B99F6ULL, "lt_0_lowtown_docks_a3_193_0.msh"},
	{0xEF5B273EF933323DULL, "gen00fl_fenris6126552.cl"},
	{0xEF5B313A787BAAF4ULL, "lt_0_lowtown_docks_a3_193_0.mmh"},
	{0xEF5B353A787BB107ULL, "lt_0_lowtown_docks_a3_193_0.mao"},
	{0xEF5C4D24DB6C1C31ULL, "6132780_m.fxe"},
	{0xEF5CA59FD0772F70ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_gallows.phy"},
	{0xEF5CCB4482F7CEF8ULL, "vxw_encounter3_e0_0_0.vlm"},
	{0xEF5D0E8DB798F2C4ULL, "6051877_m.fxe"},
	{0xEF5D1CCF71116CFBULL, "dae000_keep_noblewoman_sta6162055.cl"},
	{0xEF5D5BB8E5A94944ULL, "795418109.ogg"},
	{0xEF5D5ED6BB19B07AULL, "6133094_m.fxe"},
	{0xEF5D6B5503D67C4BULL, "gen00fl_aveline6106953.cl"},
	{0xEF5D8778BA634C86ULL, "6126711_m.fxe"},
	{0xEF5DF58BA1BC844FULL, "hlac_dr_primevalentrance_1725_3y.dds"},
	{0xEF5DF7F9F6D1ADB6ULL, "llac_ht_0_hightown_65533_1.dds"},
	{0xEF5E52D9232DDE55ULL, "vxht_0_hightown_a3_c3c_0_0.vlm"},
	{0xEF5E630D9798DED7ULL, "28184057.ogg"},
	{0xEF5EA6C6FDCC311FULL, "dae000_madam_lusine6152199.cl"},
	{0xEF5EE379ED5973A0ULL, "194840240.ogg"},
	{0xEF5EFEEA1635FA61ULL, "standard_pressed.cur"},
	{0xEF5F82182A3372C9ULL, "514032792.ogg"},
	{0xEF5FB0F951E6016BULL, "6071854_m.fxe"},
	{0xEF5FEA2DE00079B2ULL, "vxht_0_hightown_a2_b0m_0_0.vlm"},
	{0xEF603986058A4B3DULL, "llac_lt_0_lowtown_docks_196607_2.dds"},
	{0xEF6069FC0600E90FULL, "6043532_m.fxe"},
	{0xEF608451B123ABB6ULL, "6067346_m.fxe"},
	{0xEF60DB0347BB091DULL, "pro000_friedrich6050181.cl"},
	{0xEF615A74E0BE729DULL, "6048680_f.fxe"},
	{0xEF617458403C3439ULL, "vxlt_warehouse_12m_0_0.vlm"},
	{0xEF61772991363AFFULL, "gen00fl_anders6125610.cl"},
	{0xEF61C320F7EEC2F3ULL, "art\\vfx\\combat\\vfx_g_spear_att_front_c.impa.ani"},
	{0xEF61C89166251AF5ULL, "gen00fl_sebastian6195834.cl"},
	{0xEF61DC24FBA06C51ULL, "vxht_0_hightown_a3_b7j_0_0.vlm"},
	{0xEF61E3DAF6A88B16ULL, "vxdr_primevalentrance_21o_0_0.vlm"},
	{0xEF6227812F11CAF2ULL, "vxht_chantry_2ez_0_0.vlm"},
	{0xEF629B69CAE3C6DEULL, "gen00fl_merrill6162126.cl"},
	{0xEF62CD890E9C10FFULL, "737495084.ogg"},
	{0xEF62F6FD0870AA98ULL, "hlac_lt_ruinedhouse_n_-1_8x.dds"},
	{0xEF62FBF48308BA46ULL, "303663371.ogg"},
	{0xEF632406737990EBULL, "165115376.ogg"},
	{0xEF6381E70705D82BULL, "dae000_arishok6067930.cl"},
	{0xEF6430B53F2F3F86ULL, "6118116_m.fxe"},
	{0xEF64BE9A6687C04DULL, "zz_approval_debug6003064.cl"},
	{0xEF64DE0D7660B982ULL, "358191722.ogg"},
	{0xEF654F4321DAFD99ULL, "vxw_blightlands_13b_0_0.vlm"},
	{0xEF6586617EF5ACA1ULL, "vxht_chantry_332_0_0.vlm"},
	{0xEF65A56464618929ULL, "one100ip_chest_2.utp"},
	{0xEF65C87451B09872ULL, "vxdr_0_deeproads_cfe_0_0.vlm"},
	{0xEF65DF0C7E81B4DAULL, "53745002.ogg"},
	{0xEF6608C52C6E5CC4ULL, "hlac_dr_0_deeproads_5842_13y.dds"},
	{0xEF6657A79666E8B8ULL, "vxdr_primevaltomb_50y_0_0.vlm"},
	{0xEF6681D81F896AC5ULL, "6165432_m.fxe"},
	{0xEF6684A8E320BB06ULL, "vxdr_primevaltomb_507_0_0.vlm"},
	{0xEF66A32C2F8FC746ULL, "244674016.ogg"},
	{0xEF66A893993E618BULL, "dae_aura.ncs"},
	{0xEF6722350DB6DA6BULL, "vxw_bonepitdestroyed_a3_p3_0_0.vlm"},
	{0xEF6810E2A99A69CEULL, "6113885_m.fxe"},
	{0xEF687AF08C0E75D5ULL, "6033307_f.fxe"},
	{0xEF6885D2D744F749ULL, "w_sundermount_38_0.phy"},
	{0xEF68C469B57CA52EULL, "vxht_mansion2_3dp_0_0.vlm"},
	{0xEF68F8B1A9A8E168ULL, "vxlt_0_lowtown_docks_a3_n_d26_0_0.vlm"},
	{0xEF690AC89ED0B797ULL, "zz_and_debug6178601.cl"},
	{0xEF6951FADC8E30A0ULL, "lt_0_lowtown_docks_n_46_0.rml"},
	{0xEF6A7769B3A7B382ULL, "follower_banter6163436.cl"},
	{0xEF6A8ABD763AA434ULL, "412420872.ogg"},
	{0xEF6A9FA6AD3855F4ULL, "w_encounter2_42_0_lc.mmh"},
	{0xEF6AB7982F58CCAFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\additive\\flemeth_turn_left_180_cloth.gad"},
	{0xEF6AB8F816D80D66ULL, "6093253_m.fxe"},
	{0xEF6AEA1C48F4DAADULL, "730330052.ogg"},
	{0xEF6B14056630B509ULL, "dae000_sabina6161448.cl"},
	{0xEF6B4A77E4633A4CULL, "6061235_f.fxe"},
	{0xEF6B6196CC01190EULL, "533755464.ogg"},
	{0xEF6BF955AC596506ULL, "6123411_m.fxe"},
	{0xEF6BFC2FC95FAA96ULL, "vxlt_hangedman_10p_0_0.vlm"},
	{0xEF6C3E524B481160ULL, "hlac_ht_0_hightown_n_65537_0z.dds"},
	{0xEF6C6EDC8AC2704BULL, "6016506_f.fxe"},
	{0xEF6C901E4E762B1AULL, "ts_request.ncs"},
	{0xEF6CC4B61523F6D5ULL, "vxht_0_hightown_a3_c4b_0_0.vlm"},
	{0xEF6CE4B9478804F7ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.anb"},
	{0xEF6D006260490E15ULL, "gen00fl_isabela6196865.cl"},
	{0xEF6D202D5B153C58ULL, "vxht_brothel_n_4a3_0_1.vlm"},
	{0xEF6D9B6B94D34220ULL, "vxlt_gamlenshouse_n_1w_0_0.vlm"},
	{0xEF6E2E8EEC0514BEULL, "6208471_m.fxe"},
	{0xEF6E5C0AD1F69232ULL, "vxht_0_hightown_a2_b5n_0_0.vlm"},
	{0xEF6E85817D64CF0AULL, "llac_lt_0_lowtown_a2_n_c_-1_8.dds"},
	{0xEF6E88ED02EB03F7ULL, "follower_banter6163185.cl"},
	{0xEF6EBB20F7F93F2AULL, "art\\vfx\\combat\\vfx_g_spear_att_front_c.impa.evt"},
	{0xEF6EE38C7B62AD34ULL, "6078124_m.fxe"},
	{0xEF6F3FAF74AAB767ULL, "fen210_dying_magister6108614.cl"},
	{0xEF6F6F0A86ECDC98ULL, "vxca_caverns2_95t_0_0.vlm"},
	{0xEF6F75D4A04777DEULL, "vxht_mansion_n_5m6_0_0.vlm"},
	{0xEF6F8ED2D74B2545ULL, "w_sundermount_38_0.rml"},
	{0xEF6FBD41CE4E0113ULL, "6106082_m.fxe"},
	{0xEF6FD9C122491CFCULL, "6020423_f.fxe"},
	{0xEF7092835CA81C01ULL, "dae000_mother6121036.cl"},
	{0xEF709515E5B22FB5ULL, "6111592_m.fxe"},
	{0xEF71BD689F0F9880ULL, "vxlt_darktown_4ks_0_0.vlm"},
	{0xEF71C07B47B76CE2ULL, "vxw_encounter2_4r_0_0.vlm"},
	{0xEF7220123F67723AULL, "w_encounter3_9_0.msh"},
	{0xEF722E123F678A73ULL, "w_encounter3_9_0.mao"},
	{0xEF72311EF54777DBULL, "gen00fl_merrill6129018.cl"},
	{0xEF7232123F679080ULL, "w_encounter3_9_0.mmh"},
	{0xEF7232641B5A5FF5ULL, "zz_test_restriction.ncs"},
	{0xEF7295C3B625C91BULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wo_galowint.msh"},
	{0xEF72A7C3B625E866ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wo_galowint.mao"},
	{0xEF72B3C3B625FC05ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wo_galowint.mmh"},
	{0xEF73299398205E4CULL, "hlac_lt_0_lowtown_a2_n_-196609_0z.dds"},
	{0xEF73599BD5C153D7ULL, "selection_highlight.mao"},
	{0xEF73F83E70EBC521ULL, "vxht_chantry_361_0_0.vlm"},
	{0xEF7408BBA99963E7ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_bd05a.phy"},
	{0xEF7461AE33763EBEULL, "art\\vfx\\combat\\vfx_glb_restore_c.impa.ani"},
	{0xEF746AA3E6845B6AULL, "vxlt_backalley_g_1lk_0_0.vlm"},
	{0xEF7485CD7CA84B64ULL, "vxw_woundedcoast_f3_0_0.vlm"},
	{0xEF7495BEFD07BBB6ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_ext_c.impa.evt"},
	{0xEF74CC996910B934ULL, "character_character_p85.psh.fxo.dx11.assembly"},
	{0xEF74D0DD354620D3ULL, "follower_banter6155643.cl"},
	{0xEF74F685D516D386ULL, "vxdr_primevaltomb_558_0_0.vlm"},
	{0xEF7508419CCBE055ULL, "hlac_w_bonepitdestroyed_a3_458754_0z.dds"},
	{0xEF759A20F7FF259DULL, "art\\vfx\\combat\\vfx_g_spear_att_front_c.impa.gad"},
	{0xEF75CCE83EEA610DULL, "follower_banter6167208.cl"},
	{0xEF75F47D996166E8ULL, "gst_armscrossed.gda"},
	{0xEF7608732E3FF63BULL, "vxw_blightlands_o1_0_0.vlm"},
	{0xEF7614480FB5DD31ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_016a_0s.dds"},
	{0xEF76C692AFE18348ULL, "157978603.ogg"},
	{0xEF76CDF94C1D2E02ULL, "mag311_nyssa6041763.cl"},
	{0xEF7722B44F624DEEULL, "hlac_lt_gallowstemplar_2384_3x.dds"},
	{0xEF7735D74DF4762CULL, "6040699_m.fxe"},
	{0xEF77452626A5F5A1ULL, "vfx_chunkyparticle_p0.vsh.fxo"},
	{0xEF77860E0C181B47ULL, "347017366.ogg"},
	{0xEF77B9F8739D0997ULL, "6070822_m.fxe"},
	{0xEF783E08DA2A887DULL, "ave310_brennan6137234.cl"},
	{0xEF788EC63EC0E8E7ULL, "266163496.ogg"},
	{0xEF7906BF61E5D849ULL, "100386252.ogg"},
	{0xEF799D11CDCE11BCULL, "vxlt_backalley_n_1dl_0_0.vlm"},
	{0xEF7A02FC75319601ULL, "lt_darktown_3567.anb"},
	{0xEF7A10F546611494ULL, "seb201_lady_harimann6177400.cl"},
	{0xEF7B8CBEFD0DCB41ULL, "art\\vfx\\combat\\vfx_glb_bld_imp_ext_c.impa.gad"},
	{0xEF7B916AEDD0F61DULL, "gen00fl_sebastian6179872.cl"},
	{0xEF7BB4E2539AC84FULL, "vxht_0_hightown_a3_cbb_0_0.vlm"},
	{0xEF7C1A7E112E82A8ULL, "vxca_caverns1_mines_hic_0_0.vlm"},
	{0xEF7C71B145325711ULL, "gen00fl_fenris6106671.cl"},
	{0xEF7D028FA2DA5D5FULL, "hairalpha_hair_p4.psh.fxo.dx11"},
	{0xEF7D2CD233A1CD16ULL, "vxht_mansion_n_52s_0_0.vlm"},
	{0xEF7D5AC665EB069EULL, "6117284_f.fxe"},
	{0xEF7E0611E5952D75ULL, "nat320_delilah6065895.cl"},
	{0xEF7E15DFE30BED72ULL, "hlac_dr_slavers_3215_3x.dds"},
	{0xEF7E2E7BB26DACFBULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_4x.dds"},
	{0xEF7E4A06D7332654ULL, "vxlt_0_lowtown_docks_a3_n_d3y_0_0.vlm"},
	{0xEF7EA3A110A68A18ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_roundtable_sa.dds"},
	{0xEF7EF0B96FA05113ULL, "6162606_m.fxe"},
	{0xEF7F068C3F11FACFULL, "mcr370_meredith6147457.cl"},
	{0xEF7F2D6A9A4D596DULL, "360812960.ogg"},
	{0xEF7F40F31D76FCA3ULL, "dae000_madam_lusine6152083.cl"},
	{0xEF7F9FE7B00DB38DULL, "vxlt_gallowstemplar_652_0_0.vlm"},
	{0xEF7FD4982F6B5A75ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\additive\\flemeth_turn_left_180_cloth.ani"},
	{0xEF7FE33BD00CB245ULL, "6168985_m.fxe"},
	{0xEF7FE5957A92C4F9ULL, "vxht_mansion2b_1yz_0_0.vlm"},
	{0xEF8032A9AE4FE490ULL, "vxw_sundermount_2f5_0_0.vlm"},
	{0xEF803CD9FF453FF3ULL, "mag360_alain6044645.cl"},
	{0xEF80636F02D6295AULL, "follower_banter6164011.cl"},
	{0xEF8085B463F1B9E2ULL, "vxw_blightlands_ch_0_0.vlm"},
	{0xEF80D0B2ADAEA704ULL, "153095177.ogg"},
	{0xEF80D196B2AEA3ABULL, "dae000_quintus6154870.cl"},
	{0xEF812F5ADBA554FAULL, "387084420.ogg"},
	{0xEF8175B38EB28CB1ULL, "weapon_weapon_p4.psh.fxo.dx11"},
	{0xEF8179AE3380F18FULL, "art\\vfx\\combat\\vfx_glb_restore_c.impa.evt"},
	{0xEF818B4D5B6B0107ULL, "gen00fl_aveline6139869.cl"},
	{0xEF826B2117643830ULL, "437785570.ogg"},
	{0xEF828DBEFD13EBEFULL, "art\\vfx\\combat\\vfx_glb_bld_imp_ext_c.impa.ani"},
	{0xEF8298E770963948ULL, "hlac_w_sundermount_-65535_0z.dds"},
	{0xEF830B16BC5AB8A3ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_mer.msh"},
	{0xEF831916BC5AD0FDULL, "art\\characters\\playercharacter\\humanfemale\\bdy_mer.mmh"},
	{0xEF831D16BC5AD70EULL, "art\\characters\\playercharacter\\humanfemale\\bdy_mer.mao"},
	{0xEF83D115A5668653ULL, "831596212.ogg"},
	{0xEF8410CF15DFB0BCULL, "lgt110_orz_harrowmont6172614.cl"},
	{0xEF847A5020360EBBULL, "vxw_blightlands_p0_0_0.vlm"},
	{0xEF84919B9BA0BA76ULL, "vxlt_warehouse_n_x4_0_0.vlm"},
	{0xEF85C0F77484474FULL, "and110_refugee6111603.cl"},
	{0xEF85FFBEBB320B70ULL, "dae_monster_singletarget.ncs"},
	{0xEF865237E57D02DDULL, "vxw_encounter3_fi_0_1.vlm"},
	{0xEF86BC113672893EULL, "lt_hangedman_n_mmap_ph.dds"},
	{0xEF870ABF45E2301EULL, "6135662_f.fxe"},
	{0xEF870E6290D59CD4ULL, "zz_dog_debug6127922.cl"},
	{0xEF8710E14ED019E0ULL, "vxlt_blackemporium_10j_0_0.vlm"},
	{0xEF871D0B597C378FULL, "follower_banter6141853.cl"},
	{0xEF8738EE6132EA24ULL, "art\\characters\\playercharacter\\humanfemale\\mt_bas_hf_lashesm1.mao"},
	{0xEF8791A9B3DF96D7ULL, "6182464_m.fxe"},
	{0xEF87B07434212A7FULL, "vrc221cr_guards_fake.ncs"},
	{0xEF87CE57BEDCF96AULL, "6074495_f.fxe"},
	{0xEF8815E68062A30FULL, "hlac_dr_dungeon_3217_0x.dds"},
	{0xEF887AAE33871238ULL, "art\\vfx\\combat\\vfx_glb_restore_c.impa.gad"},
	{0xEF88A6747E5FC3DDULL, "57333835.ogg"},
	{0xEF88D580E60CB11EULL, "177339956.ogg"},
	{0xEF88E1ADE0B63B7BULL, "vxdr_primevaltomb_84g_0_0.vlm"},
	{0xEF890DEA89F9F93CULL, "art\\characters\\playercharacter\\humanmale\\hlh_medb.mao"},
	{0xEF8919EA89FA0DDFULL, "art\\characters\\playercharacter\\humanmale\\hlh_medb.mmh"},
	{0xEF891FEA89FA1781ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medb.msh"},
	{0xEF898821B019BA08ULL, "vxlt_undercitydungeon_1zz_0_0.vlm"},
	{0xEF89DE696207A8B9ULL, "6042233_f.fxe"},
	{0xEF8A269FF6DAFBDBULL, "cd_mag320_orsino.bnk"},
	{0xEF8A48702B6D3036ULL, "llac_dr_0_deeproads_5849_8.dds"},
	{0xEF8AA27B9A1C6F38ULL, "vxht_0_hightown_n_4c7_0_0.vlm"},
	{0xEF8B05E732F8F9F2ULL, "vxlt_gallowscourtyard_n_690_0_0.vlm"},
	{0xEF8BA2E656004C73ULL, "vxlt_0_lowtown_docks_n_aop_0_0.vlm"},
	{0xEF8C2DB3EBBE1D06ULL, "6147004_m.fxe"},
	{0xEF8C6550EE7AC4BFULL, "ht_0_hightown_a2_37_0_lc.mmh"},
	{0xEF8C687AB7CCBDB7ULL, "llac_lt_undercitydungeon_-1_6.dds"},
	{0xEF8CBE0162F941E6ULL, "vxca_cavern1_haunted_gkq_0_1.vlm"},
	{0xEF8D7D1804F41472ULL, "vxht_chantry_2te_0_0.vlm"},
	{0xEF8E08E1791F216DULL, "6162301_m.fxe"},
	{0xEF8E74A95D3308CFULL, "6147197_m.fxe"},
	{0xEF8E80BA773318D5ULL, "6164592_m.fxe"},
	{0xEF8EA486A045FD10ULL, "vxw_sundermount_2k2_0_0.vlm"},
	{0xEF8ECB6A80B9B798ULL, "art\\characters\\playercharacter\\humanfemale\\cth_coma.phy"},
	{0xEF8EF79155E7D262ULL, "vxw_blightlands_di_0_0.vlm"},
	{0xEF8EF7CB384840CFULL, "vxlt_gallowstemplar_n_5r1_0_0.vlm"},
	{0xEF8F2DED76F64844ULL, "6153348_m.fxe"},
	{0xEF8F9A982B66487DULL, "follower_banter6161246.cl"},
	{0xEF90441A0415F8EAULL, "ca_caverns1_mines_24425_lc.mmh"},
	{0xEF9047B9851E8E1CULL, "24113159.ogg"},
	{0xEF9047CD33668A80ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_c.mmh"},
	{0xEF907EA2FD2284D0ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtsml.msh"},
	{0xEF90856FD8364121ULL, "6027622_m.fxe"},
	{0xEF908CA2FD229CEAULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtsml.mmh"},
	{0xEF9090A2FD22A321ULL, "art\\levels\\races\\vfx_models\\prp_lgtbemhtsml.mao"},
	{0xEF913E3260E91CA6ULL, "llac_ht_0_hightown_a3_-65537_1.dds"},
	{0xEF916C8B44BAB6E9ULL, "789603798.ogg"},
	{0xEF9197FE949B89C5ULL, "885857879.ogg"},
	{0xEF919D690BF20C70ULL, "71399166.ogg"},
	{0xEF91E15E1BCCBC31ULL, "776134287.ogg"},
	{0xEF91EB446C218BE7ULL, "fen210_fenris6108725.cl"},
	{0xEF92133845467FA6ULL, "650466959.ogg"},
	{0xEF9219027E8A6BCFULL, "ts_darkspawn_prog_junk1.gda"},
	{0xEF92C435626C4322ULL, "vxw_bonepit_8f_0_0.vlm"},
	{0xEF92C68BEE3CE09CULL, "dae000cr_grand_cleric6164466.cl"},
	{0xEF92E86E5BB7E2C2ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_beta_0t.dds"},
	{0xEF92EC2D122C273BULL, "vxw_blightlands_u3_0_0.vlm"},
	{0xEF941323248D12CBULL, "hlac_lt_0_lowtown_-131071_1y.dds"},
	{0xEF941B4764A12AD8ULL, "fen210_orana6109727.cl"},
	{0xEF948E059DE55577ULL, "art\\characters\\weapons\\dagger\\w_dgr_106.mao"},
	{0xEF948F46D8AED984ULL, "hlac_lt_hangedman_-1_10x.dds"},
	{0xEF94B815038796EBULL, "6153332_m.fxe"},
	{0xEF95312BFDDCCDB3ULL, "pls_hilight.dds"},
	{0xEF956ED2D76BBE6EULL, "w_sundermount_38_0.mmh"},
	{0xEF957AD2D76BD205ULL, "w_sundermount_38_0.mao"},
	{0xEF9588D2D76BEA5CULL, "w_sundermount_38_0.msh"},
	{0xEF95E1B947AB04F8ULL, "art\\vfx\\combat\\vfx_q_ari_charge_c.mmh"},
	{0xEF969BB7A5CC5BE0ULL, "vxw_encounter3_5x_0_0.vlm"},
	{0xEF96A98D572870E7ULL, "hlac_ca_caverns1_mines_24424_20y.dds"},
	{0xEF9712722AE60906ULL, "vxw_bonepit_y5_0_0.vlm"},
	{0xEF97B3AF3EBA825EULL, "vxlt_0_lowtown_f2g_0_0.vlm"},
	{0xEF98A4C396D66DB0ULL, "ser120_ghyslain6104208.cl"},
	{0xEF9941FADDCA065EULL, "vxw_encounter3_y_0_0.vlm"},
	{0xEF999DDF2F48F119ULL, "6125848_f.fxe"},
	{0xEF99D31A30134BEDULL, "llac_ht_mansion2_818_7.dds"},
	{0xEF9A35CC1C6ADD94ULL, "llac_lt_0_lowtown_docks_a3_-4_2.dds"},
	{0xEF9AD626935392C6ULL, "6164555_m.fxe"},
	{0xEF9BA98CAD250006ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf.ani"},
	{0xEF9C198D5C289E5BULL, "213162837.ogg"},
	{0xEF9C20AA1B18292DULL, "6064955_m.fxe"},
	{0xEF9C401099255E21ULL, "6075445_m.fxe"},
	{0xEF9C4E1F5312F43FULL, "vxca_caverns1_i94_0_0.vlm"},
	{0xEF9C7ABE6B1539EDULL, "6167001_m.fxe"},
	{0xEF9C80571C2813C9ULL, "228900701.ogg"},
	{0xEF9CCC1189364674ULL, "6139533_m.fxe"},
	{0xEF9CD8464123A1D3ULL, "6041400_m.fxe"},
	{0xEF9D0CBFB8E27139ULL, "203504407.ogg"},
	{0xEF9D1663923C1590ULL, "vxw_sundermount_2l3_0_0.vlm"},
	{0xEF9D2ACD3370E39FULL, "art\\vfx\\combat\\vfx_m_sph_healaura_c.anb"},
	{0xEF9D2BE602CAF0C7ULL, "273934955.ogg"},
	{0xEF9D696E47DDEAE2ULL, "vxw_blightlands_ij_0_0.vlm"},
	{0xEF9D8A2BD20603C7ULL, "823708573.ogg"},
	{0xEF9D905C0F89B038ULL, "393368620.ogg"},
	{0xEF9D9FCA68EC60C4ULL, "6156048_m.fxe"},
	{0xEF9DCADCEFF3F7CEULL, "752355722.ogg"},
	{0xEF9DF030526BF038ULL, "lla_ht_0_hightown_2_0.dds"},
	{0xEF9E13ED75ADF549ULL, "vxca_caverns1_mines_jl7_0_0.vlm"},
	{0xEF9F117A1256DB7AULL, "6113908_m.fxe"},
	{0xEF9F15DDAF5E60ECULL, "6013264_m.fxe"},
	{0xEF9F75158FDD6C49ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxa.phy"},
	{0xEF9FDB64528A94E4ULL, "vxw_woundedcoast_u6_0_0.vlm"},
	{0xEFA074CC9A775126ULL, "follower_banter6136328.cl"},
	{0xEFA0E9A8D58F3C6BULL, "vxw_bonepitdestroyed_a3_l7_0_0.vlm"},
	{0xEFA0F04849A319D4ULL, "hlac_lt_undercityencounter_-1_37x.dds"},
	{0xEFA127B10B6243DCULL, "blood_dot.mao"},
	{0xEFA15358A5809832ULL, "6116183_m.fxe"},
	{0xEFA1852C56B38F80ULL, "llac_lt_0_lowtown_docks_a3_n_-2_1.dds"},
	{0xEFA24BA9E37F482FULL, "vxdr_0_deeproads_fid_0_0.vlm"},
	{0xEFA29CC4754AC2C9ULL, "6173088_m.fxe"},
	{0xEFA36D031E4FA2FCULL, "222468299.ogg"},
	{0xEFA39437DAFE4832ULL, "llac_lt_0_lowtown_docks_n_-196609_1.dds"},
	{0xEFA3C1A838084522ULL, "gen_im_npc_ammo_quiver_qunc.uti"},
	{0xEFA3E2525ED6002EULL, "6162711_m.fxe"},
	{0xEFA3FCA7F541E948ULL, "llac_dr_0_deeproads_5845_2.dds"},
	{0xEFA40515DCD328F1ULL, "6166580_f.fxe"},
	{0xEFA428BB2898EE88ULL, "vxlt_ruinedhouse_n_h0_0_0.vlm"},
	{0xEFA4453C0B489647ULL, "6115474_f.fxe"},
	{0xEFA460AB36B8D5E3ULL, "6061785_m.fxe"},
	{0xEFA4B5C91A0D9F97ULL, "gen00fl_bethany6098048.cl"},
	{0xEFA4C04B62119972ULL, "dae000_meredith6166303.cl"},
	{0xEFA4F2A8A3B05B3CULL, "vxlt_backalley_n_1uc_0_0.vlm"},
	{0xEFA5554AA0CA8CE5ULL, "6027675_m.fxe"},
	{0xEFA5574DD0224F38ULL, "vxw_bonepit_dx_0_0.vlm"},
	{0xEFA5BB169E685037ULL, "hlac_lt_0_lowtown_docks_n_-3_3y.dds"},
	{0xEFA6117A69CD06F8ULL, "dae011_corff_sta6149631.cl"},
	{0xEFA619E3CEA6E8F6ULL, "176820284.ogg"},
	{0xEFA69FC59FAC248EULL, "vxlt_backalley_g_1nx_0_1.vlm"},
	{0xEFA6C0F893403875ULL, "ch2alph.psh.fxo"},
	{0xEFA6CEFDE209014BULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.mmh"},
	{0xEFA743D17297B399ULL, "mag360_alain6044658.cl"},
	{0xEFA7456316C78251ULL, "6068483_m.fxe"},
	{0xEFA85F538F0EE081ULL, "gen00fl_sebastian6188119.cl"},
	{0xEFA8611677F1D61FULL, "vxw_encounter2_camp_lc_0_0.vlm"},
	{0xEFA8A8850855CAB7ULL, "6115752_m.fxe"},
	{0xEFA96EE52D9ECCFEULL, "375890963.ogg"},
	{0xEFA9918CAD311577ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf.evt"},
	{0xEFA9A8EC1E797370ULL, "801216888.ogg"},
	{0xEFAA0AF0022C1F09ULL, "45254302.ogg"},
	{0xEFAAB45277EB65A2ULL, "vxlt_0_lowtown_a3_n_c_erx_0_0.vlm"},
	{0xEFAAE3C71E8E2133ULL, "506099248.ogg"},
	{0xEFABDB4B39D40362ULL, "vxw_blightlands_rk_0_0.vlm"},
	{0xEFAC20EF4052C178ULL, "vxht_0_hightown_n_bj6_0_0.vlm"},
	{0xEFAC21A980CDF0EFULL, "hlac_lt_backalley_n_0_4x.dds"},
	{0xEFACBD0421D79189ULL, "binkalpha.psh.fxo"},
	{0xEFACC698F34A7E66ULL, "follower_banter6153970.cl"},
	{0xEFAD21B295EAFAEFULL, "vxca_caverns1_mines_j49_0_0.vlm"},
	{0xEFAD4C9D5D2FF25BULL, "6143399_m.fxe"},
	{0xEFAD5EA0EA4FE07AULL, "vxw_bonepit_224_0_0.vlm"},
	{0xEFAD6853231C101AULL, "docks_male_human_worker_sta6175999.cl"},
	{0xEFAD72C7B180E064ULL, "dae000ip_crowd_lm1_talk16.utp"},
	{0xEFADF76FC3A8000FULL, "545373808.ogg"},
	{0xEFAE3217AE5CBD6AULL, "vxlt_backalley_g_1xg_0_0.vlm"},
	{0xEFAE36DC6ED90580ULL, "gen_im_trs_dem_sta_spi_r0.uti"},
	{0xEFAE981FD26A0009ULL, "plt_cod_ltr_lother_barlin.nss"},
	{0xEFAEE4BD6157A5A1ULL, "6169375_m.fxe"},
	{0xEFAEE883A3F65B6FULL, "t3_cor_b03.tnt"},
	{0xEFAEF8F7AC06BF9CULL, "vxht_chantry_n_2dt_0_0.vlm"},
	{0xEFAF928CAD358200ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_rf.gad"},
	{0xEFB027E029F7134BULL, "hlac_lt_gallowscourtyard_n_-65536_0z.dds"},
	{0xEFB17B092C58495DULL, "6174584_m.fxe"},
	{0xEFB1E078FDF22960ULL, "vxlt_0_lowtown_a2_erz_0_0.vlm"},
	{0xEFB2A224013EB17CULL, "284681906.ogg"},
	{0xEFB2A5C4424F8BDDULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken4xa.phy"},
	{0xEFB2F562469D0875ULL, "vxht_bartrand_n_5vg_0_0.vlm"},
	{0xEFB3105E7A804BE9ULL, "gen00fl_isabela6015823.cl"},
	{0xEFB31C9FD232962CULL, "6069921_f.fxe"},
	{0xEFB3C92AC21867B8ULL, "vxw_bonepit_oy_0_0.vlm"},
	{0xEFB3E3FC7562D991ULL, "lt_darktown_3567.rml"},
	{0xEFB3F62C0ED23A06ULL, "vxw_bonepit_o7_0_0.vlm"},
	{0xEFB40D05069E19D8ULL, "470735178.ogg"},
	{0xEFB412D28541BE91ULL, "6122851_m.fxe"},
	{0xEFB4400F7A195268ULL, "hlac_ht_chantry_2534_23x.dds"},
	{0xEFB4AFFDE21509B0ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.anb"},
	{0xEFB5478EC7DE0201ULL, "382727994.ogg"},
	{0xEFB5AA09AA45ED77ULL, "gen00fl_anders6119275.cl"},
	{0xEFB6066C8075CCC2ULL, "626538475.ogg"},
	{0xEFB69DFD7BCD45CEULL, "6064677_m.fxe"},
	{0xEFB6B858B499D309ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_grinder.phy"},
	{0xEFB6D2F369E7EE9FULL, "vxw_encounter2_camp_gb_0_0.vlm"},
	{0xEFB6DC84B021FB47ULL, "hlac_w_bonepit_131074_0y.dds"},
	{0xEFB6F3CB3D10F11FULL, "6087792_m.fxe"},
	{0xEFB76988C0C4AAF9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_loop.ani"},
	{0xEFB76E830CF2ECA4ULL, "vxlt_0_lowtown_docks_n_b5g_0_1.vlm"},
	{0xEFB7B4B68DBDA9A7ULL, "vxca_cavern1_haunted_cjr_0_0.vlm"},
	{0xEFB8272A10395007ULL, "lowtown_male_drunk_mob6151165.cl"},
	{0xEFB86544937D03B3ULL, "w_encounter2_camp_42_0_ll.mmh"},
	{0xEFB8CD2D8AE35EF3ULL, "6099395_m.fxe"},
	{0xEFB8F964D3209D3FULL, "gen00fl_merrill6150542.cl"},
	{0xEFB9262F69E17E22ULL, "vxlt_0_lowtown_a3_n_c_euy_0_0.vlm"},
	{0xEFB94890BF3E9138ULL, "lgt250_way_dalish_conflict6147157.cl"},
	{0xEFB98CC3B661BCB2ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\1wo_galowint.phy"},
	{0xEFB9B34CE9CB8BCBULL, "vxca_cavern1_haunted_fmr_0_0.vlm"},
	{0xEFB9D7FFF02D78D5ULL, "narrative_varric6172476.cl"},
	{0xEFB9EDB917437E43ULL, "llac_dr_primevalentrance_1726_7.dds"},
	{0xEFB9FA1D76284690ULL, "vxw_sundermount_2z1_0_0.vlm"},
	{0xEFBA4D620E2A8A4FULL, "vxlt_gallowstemplar_n_5c2_0_0.vlm"},
	{0xEFBAAA82D6AECB8EULL, "405711057.ogg"},
	{0xEFBB7AC8E4CA23C7ULL, "769032730.ogg"},
	{0xEFBB9463A35A59A0ULL, "121347770.ogg"},
	{0xEFBBF316A679B586ULL, "gen_im_npc_wep_dua_gena.uti"},
	{0xEFBC008167B839D0ULL, "lgt300cr_hanker.utc"},
	{0xEFBC7A79A58FB686ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv2.ani"},
	{0xEFBD16310623F284ULL, "ico_mace_5.dds"},
	{0xEFBDCF80E5671265ULL, "lowtown_fem_prostitute_sta6161610.cl"},
	{0xEFBE1BA1762C66A6ULL, "250393345.ogg"},
	{0xEFBE4488C0CA8A3BULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_loop.gad"},
	{0xEFBE5D39F542EAB2ULL, "6085346_m.fxe"},
	{0xEFBEA38CB9E82A6DULL, "6041279_m.fxe"},
	{0xEFBEA9EC74C62A42ULL, "vxw_bonepit_1wu_0_1.vlm"},
	{0xEFBEB34EE66D26DCULL, "190200080.ogg"},
	{0xEFBEFFEC560C0CD5ULL, "6072314_m.fxe"},
	{0xEFBF60F155DEEDCEULL, "lgt310_mag_melson_guards6153019.cl"},
	{0xEFBF7B18993E8F8BULL, "ch1emarmtnt.psh.fxo"},
	{0xEFBFD78215CEE8F5ULL, "hurlock_emissary_off.utc"},
	{0xEFBFE284E0F69AC6ULL, "597239269.ogg"},
	{0xEFC001054E0542EBULL, "787162273.ogg"},
	{0xEFC06FA1F54073E0ULL, "plt_lgt000pt_ultra_3_05.nss"},
	{0xEFC09C1BA681B71AULL, "vxlt_0_lowtown_a2_n_c_gmz_0_0.vlm"},
	{0xEFC0B974A5AF5A14ULL, "l_mag101ar_caverns.wbd"},
	{0xEFC109865AD95FFEULL, "vxht_keep_n_372_0_0.vlm"},
	{0xEFC12F809EBF1751ULL, "qcr220_grey_wardens6060433.cl"},
	{0xEFC1673F389320F5ULL, "vxht_bartrand_n_5yf_0_0.vlm"},
	{0xEFC1F836D55D5802ULL, "6083832_m.fxe"},
	{0xEFC2FD6571912AC6ULL, "ser221_party_comments6122851.cl"},
	{0xEFC3037274CF7C29ULL, "245006633.ogg"},
	{0xEFC38949456367EAULL, "276104826.ogg"},
	{0xEFC3998C147B1D17ULL, "vxlt_0_lowtown_docks_a3_cvm_0_0.vlm"},
	{0xEFC39AC4425DEC49ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken4xa.mao"},
	{0xEFC39B404A780EF8ULL, "6051598_m.fxe"},
	{0xEFC39EF2A475CC9FULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\dr2a_simple_2.pwk"},
	{0xEFC3A6C4425E00E2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken4xa.mmh"},
	{0xEFC3A8C4425E0398ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_broken4xa.msh"},
	{0xEFC3B49829EBF165ULL, "6169340_m.fxe"},
	{0xEFC3BAE1E9A0D480ULL, "vxlt_warehouse_n_5u_0_0.vlm"},
	{0xEFC3EEAA3FDF0D1BULL, "mag210_arianni6031699.cl"},
	{0xEFC48988DB0AD023ULL, "6087367_m.fxe"},
	{0xEFC4D12D26824AC0ULL, "nat380_sibling6207565.cl"},
	{0xEFC4F59D67E5A10AULL, "397036611.ogg"},
	{0xEFC544D05BDE071FULL, "vxw_encounter2_camp_bm_0_0.vlm"},
	{0xEFC553693A3917F1ULL, "vxdr_0_deeproads_ev4_0_0.vlm"},
	{0xEFC56188C0D0DA20ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\pose\\follower_dm_loop.evt"},
	{0xEFC5625C06B8416FULL, "6202182_m.fxe"},
	{0xEFC56F7E72B53188ULL, "6015064_m.fxe"},
	{0xEFC5A1FF83C8604CULL, "lowtown_argument6169363.cl"},
	{0xEFC5A5BD252746DBULL, "l_dae211ar_tavern.bnk"},
	{0xEFC5C1A4006251DEULL, "22680781.ogg"},
	{0xEFC5F52B46C67C35ULL, "6041222_m.fxe"},
	{0xEFC6BF9175B94BDBULL, "art\\characters\\playercharacter\\elfmale\\bdy_flsb.mmh"},
	{0xEFC6CB9175B96078ULL, "art\\characters\\playercharacter\\elfmale\\bdy_flsb.mao"},
	{0xEFC6DD9175B97EC5ULL, "art\\characters\\playercharacter\\elfmale\\bdy_flsb.msh"},
	{0xEFC6EFDD4CC1DF24ULL, "vxht_mansion_n_587_0_0.vlm"},
	{0xEFC7033B61FD1A63ULL, "ch1alphmultnt_vl.psh.fxo"},
	{0xEFC7189F2F7D2552ULL, "6001287_m.fxe"},
	{0xEFC7A3FBF952E53AULL, "art\\vfx\\models\\vfx_rockarmor_lh.mmh"},
	{0xEFC7B5FBF9530380ULL, "art\\vfx\\models\\vfx_rockarmor_lh.msh"},
	{0xEFC7B8F63EABFBBAULL, "hlac_lt_0_lowtown_docks_n_65533_3x.dds"},
	{0xEFC7C6E6179968AFULL, "aip_rockwraith_act_1.gda"},
	{0xEFC7CDB2E2EBD9F0ULL, "ht_0_hightownexplosion_n_0_0.mao"},
	{0xEFC7D1B2E2EBE003ULL, "ht_0_hightownexplosion_n_0_0.mmh"},
	{0xEFC7D40C859F3FDDULL, "ht_bartrand_n_3382.rml"},
	{0xEFC7DFB2E2EBF85DULL, "ht_0_hightownexplosion_n_0_0.msh"},
	{0xEFC870FF304517E9ULL, "6024541_f.fxe"},
	{0xEFC90DA87F3B1D60ULL, "vxht_0_hightown_n_b4p_0_0.vlm"},
	{0xEFC92929A675BEC0ULL, "art\\vfx\\materials\\vfx_slash_fb_add.mao"},
	{0xEFC942510CD7FABAULL, "art\\cinematics\\props\\prp_bookinquisb.mao"},
	{0xEFC9779B75DBC489ULL, "mer240_jansen6050327.cl"},
	{0xEFC9E467FCCE5745ULL, "vxw_bonepit_270_0_0.vlm"},
	{0xEFCA611AC828D6D5ULL, "6176740_m.fxe"},
	{0xEFCA6279A59BCBF7ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv2.evt"},
	{0xEFCA6B4369C13D0AULL, "307254153.ogg"},
	{0xEFCA7F725BBD2445ULL, "dae000cr_grand_cleric6187934.cl"},
	{0xEFCAECA307A37506ULL, "vxca_caverns2prc_9sm_0_0.vlm"},
	{0xEFCB112B44D77D69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\twoverlay.anb"},
	{0xEFCB11C2B20372F4ULL, "6108088_f.fxe"},
	{0xEFCB403B613A1B0BULL, "llac_lt_0_lowtown_docks_-65539_2.dds"},
	{0xEFCB5D6235E529E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_death_magic.evt"},
	{0xEFCB6F95586EEC94ULL, "6034357_m.fxe"},
	{0xEFCB7729BE0740D7ULL, "6117813_m.fxe"},
	{0xEFCB827D5EC5F80FULL, "art\\vfx\\combat\\vfx_c_rw_aoeburst_c.impa.gad"},
	{0xEFCBC295B75CB4DBULL, "6075766_m.fxe"},
	{0xEFCBC8774543D6A1ULL, "6163175_m.fxe"},
	{0xEFCC409AFDA1F5E9ULL, "vxw_sundermount_2e9_0_0.vlm"},
	{0xEFCC5E159004336EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxa.mmh"},
	{0xEFCC6A1590044705ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxa.mao"},
	{0xEFCC781590045F5CULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_timberpostxa.msh"},
	{0xEFCCECE064AFC9A0ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat_5x_5_0.pwk"},
	{0xEFCD3EE955821F90ULL, "lt_gallowscourtyard_n_77_0.rml"},
	{0xEFCDABA2B71B498BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_loop_2.ani"},
	{0xEFCDB5C6CA171876ULL, "character_character_p272.psh.fxo.dx11.assembly"},
	{0xEFCDBD4E652974A7ULL, "vxht_0_hightown_a2_cbg_0_0.vlm"},
	{0xEFCE15DD273C4FA8ULL, "6126602_f.fxe"},
	{0xEFCE2D2B8872723BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_exit.evt"},
	{0xEFCEC432E1DE5A60ULL, "vxlt_0_lowtown_a2_exx_0_0.vlm"},
	{0xEFCF02982D431C1AULL, "hlac_lt_0_lowtown_a2_n_c_-131071_1y.dds"},
	{0xEFCF1C2A68568E7BULL, "hlac_ht_0_hightown_n_-131073_1x.dds"},
	{0xEFCF6C4567A5A416ULL, "143505048.ogg"},
	{0xEFCF858A23DDA2DCULL, "w_blightlands_190_0.mao"},
	{0xEFCF918A23DDB77FULL, "w_blightlands_190_0.mmh"},
	{0xEFCF978A23DDC121ULL, "w_blightlands_190_0.msh"},
	{0xEFCFA6FB62D08C7AULL, "6041503_f.fxe"},
	{0xEFD0126052E9A333ULL, "abi_r_stun_gren.bnk"},
	{0xEFD06379A5A03880ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderattkfrontv2.gad"},
	{0xEFD06A13C7537082ULL, "6086732_m.fxe"},
	{0xEFD0CE9AB3F7EB77ULL, "6142064_m.fxe"},
	{0xEFD0E975E3979AEDULL, "90670188.ogg"},
	{0xEFD151F2BE049C3BULL, "6105748_m.fxe"},
	{0xEFD1977D5ECA87BCULL, "art\\vfx\\combat\\vfx_c_rw_aoeburst_c.impa.evt"},
	{0xEFD1A3A0228F832EULL, "vxht_keep_6yj_0_0.vlm"},
	{0xEFD20B6906713597ULL, "vxlt_0_lowtown_docks_a3_cyl_0_0.vlm"},
	{0xEFD2149FF71892D9ULL, "cd_mag320_orsino.wbd"},
	{0xEFD22CBEDB96ED00ULL, "vxlt_warehouse_n_8r_0_0.vlm"},
	{0xEFD22F66BB903CE4ULL, "ch1_d.psh.fxo"},
	{0xEFD2379962CEC46AULL, "zz_fen_debug6165486.cl"},
	{0xEFD2746235EB6F31ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_death_magic.gad"},
	{0xEFD29C658F0E0AF0ULL, "hlac_lt_hangedman_-1_22y.dds"},
	{0xEFD2EBF6D8AF782EULL, "rdr111dg_martin6053569.cl"},
	{0xEFD3808600432155ULL, "vxht_keep_4hp_0_0.vlm"},
	{0xEFD4262B8876D27CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_exit.gad"},
	{0xEFD4C519DDF172FEULL, "6018836_m.fxe"},
	{0xEFD4D11DF46F6BBDULL, "185527867.ogg"},
	{0xEFD50A87292AC636ULL, "dae000_player_generic_find6179054.cl"},
	{0xEFD518A5892FC304ULL, "hlac_lt_gallowstemplar_f_2384_0x.dds"},
	{0xEFD5402F1837826BULL, "scaleform_sf_main_p1.psh.fxo"},
	{0xEFD58A7C21733DD2ULL, "6004187_m.fxe"},
	{0xEFD68A5E114D1621ULL, "mage_forcecrush.dds"},
	{0xEFD740A06C43A624ULL, "gen00fl_sebastian6180817.cl"},
	{0xEFD75BC0E34FD50DULL, "6126499_m.fxe"},
	{0xEFD7A4A57DD428E5ULL, "gen00fl_anders6125906.cl"},
	{0xEFD7D84D7C044215ULL, "6174801_m.fxe"},
	{0xEFD7F0AFCFD7FD38ULL, "6174192_m.fxe"},
	{0xEFD7F61FC5D7F7AEULL, "839577821.ogg"},
	{0xEFD86487CD882F0FULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheadb_d.dds"},
	{0xEFD893E5D0BC0E0AULL, "vxw_blightlands_e6_0_0.vlm"},
	{0xEFD8BF7D4867D334ULL, "1744897.ogg"},
	{0xEFD8D2B2E2FA5528ULL, "ht_0_hightownexplosion_n_0_0.phy"},
	{0xEFD8DC60F806A82BULL, "purchaseconfirm.gfx"},
	{0xEFD969F59D6655ADULL, "6103404_f.fxe"},
	{0xEFD9756235F18FDFULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_death_magic.ani"},
	{0xEFD982AC176E35E7ULL, "vxlt_gallowstemplar_n_5z1_0_0.vlm"},
	{0xEFD9B22706C52C0FULL, "zz_qun_debug6028224.cl"},
	{0xEFDA4EE46338A44DULL, "vxca_caverns2prc_4o3_0_0.vlm"},
	{0xEFDA600E138D1DC3ULL, "llac_ht_brothel_n_3381_1.dds"},
	{0xEFDAB277EF980E69ULL, "vxw_sundermount_2n8_0_0.vlm"},
	{0xEFDB01FB2317151CULL, "408433172.ogg"},
	{0xEFDB1F2E5E29C1F9ULL, "w_woundedcoast_57_0.msh"},
	{0xEFDB252B887CEF7AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_kd_exit.ani"},
	{0xEFDB292E5E29D2E7ULL, "w_woundedcoast_57_0.mmh"},
	{0xEFDB2D2E5E29D934ULL, "w_woundedcoast_57_0.mao"},
	{0xEFDC02D041A7C9E3ULL, "mag371_elsa6041255.cl"},
	{0xEFDC0CE95B5CF39AULL, "980151086.ogg"},
	{0xEFDC4209A7F811BEULL, "vxca_caverns2_945_0_0.vlm"},
	{0xEFDC66060E7F877EULL, "vxca_caverns2prc_9fn_0_0.vlm"},
	{0xEFDC7DD1BAF320F8ULL, "dre000_dougal6129758.cl"},
	{0xEFDC910C2FEE2E87ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_cupwood_01_0_sa.dds"},
	{0xEFDC969E3BA5B5B8ULL, "dae000_pc_home_ambients6208299.cl"},
	{0xEFDCC84D625A9102ULL, "169843320.ogg"},
	{0xEFDD51B0EBA8A986ULL, "114965051.ogg"},
	{0xEFDDD25E6BCC784CULL, "em_terath.mor"},
	{0xEFDDDDD5B8A1FB59ULL, "6120835_m.fxe"},
	{0xEFDE3A904DDB62AEULL, "vxw_encounter3_cp_0_0.vlm"},
	{0xEFDEBD670001DD00ULL, "hlac_lt_undercityencounter_-1_23y.dds"},
	{0xEFDF0497FCF9BA6DULL, "hlac_lt_warehouse_n_0_5z.dds"},
	{0xEFDF06B54C6AD739ULL, "vxlt_undercitydungeon_xd_0_0.vlm"},
	{0xEFDF83D9E648EB8BULL, "stone_rock.mao"},
	{0xEFDF9F7D5ED6D2D5ULL, "art\\vfx\\combat\\vfx_c_rw_aoeburst_c.impa.ani"},
	{0xEFDFBF2177744ACBULL, "945004142.ogg"},
	{0xEFDFF6815586FA31ULL, "vxht_mansion2b_1hm_0_0.vlm"},
	{0xEFE0117A597A4EC0ULL, "263420801.ogg"},
	{0xEFE0157D14859BAEULL, "vxht_keep_6bk_0_0.vlm"},
	{0xEFE021C9C8249884ULL, "141888963.ogg"},
	{0xEFE05435F36A7096ULL, "hlac_lt_0_lowtown_docks_-131074_1x.dds"},
	{0xEFE07CF729E37C30ULL, "6165345_m.fxe"},
	{0xEFE122C833C53EA4ULL, "qun200_party_comment6098547.cl"},
	{0xEFE207A4845AB54FULL, "271186884.ogg"},
	{0xEFE25471E7A4953EULL, "llac_lt_warehouse_n_1_1.dds"},
	{0xEFE292A2B72D7B15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_loop_2.gad"},
	{0xEFE2A158B4BEE72EULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_grinder.mmh"},
	{0xEFE2AD58B4BEFBC5ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_grinder.mao"},
	{0xEFE2BB58B4BF131CULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_grinder.msh"},
	{0xEFE2DA19EABF82B0ULL, "vxht_0_hightown_a3_cah_0_0.vlm"},
	{0xEFE318675C419DEDULL, "hlac_ca_caverns2prc_5802_47z.dds"},
	{0xEFE3DA3152AC99ACULL, "389773180.ogg"},
	{0xEFE3E88D95BA4FE9ULL, "gen00fl_carver6097399.cl"},
	{0xEFE40369D2A1A4ECULL, "vxht_mansion_n_3yj_0_0.vlm"},
	{0xEFE45957E075DDCFULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastrock_n.dds"},
	{0xEFE487F22E97B385ULL, "gen_im_wep_rog_dua_act2_02.uti"},
	{0xEFE4F1AC94A081C9ULL, "6092125_m.fxe"},
	{0xEFE5207449FCDC4EULL, "static_static_p98.psh.fxo.dx11"},
	{0xEFE556EA5167DAAEULL, "vxht_0_hightown_n_bqe_0_0.vlm"},
	{0xEFE58FEE67289016ULL, "vxht_brothel_n_2k4_0_0.vlm"},
	{0xEFE59A52CCD94594ULL, "vxw_encounter1_jb_0_1.vlm"},
	{0xEFE5AF4CCAE1FB2CULL, "6043823_m.fxe"},
	{0xEFE5B4F54134B0B9ULL, "gen00fl_merrill6121258.cl"},
	{0xEFE5D08244D8D293ULL, "6044825_m.fxe"},
	{0xEFE65D70E183AE6FULL, "vxca_caverns1_iyd_0_0.vlm"},
	{0xEFE68CA076E680BFULL, "eye_eye_p0.vsh.fxo"},
	{0xEFE6E19DC28EF3DEULL, "vxw_encounter2_camp_20_0_0.vlm"},
	{0xEFE705C2C2B2268AULL, "vxw_blightlands_n7_0_0.vlm"},
	{0xEFE7202672D7A00CULL, "154132167.ogg"},
	{0xEFE72540691E7F07ULL, "vxlt_gallowstemplar_n_6s4_0_1.vlm"},
	{0xEFE778862ED2516AULL, "6080176_m.fxe"},
	{0xEFE7DB691C9743D3ULL, "142049511.ogg"},
	{0xEFE8381F4CFEF5ECULL, "580879364.ogg"},
	{0xEFE853E9062B26ECULL, "6176316_m.fxe"},
	{0xEFE8A8154760CEE7ULL, "plt_cod_hst_vael_family.plo"},
	{0xEFE91B4026529C30ULL, "6128064_f.fxe"},
	{0xEFE9B39753D305CBULL, "atl_worldmap_dxt1_00.dds"},
	{0xEFEA6CE1226E298BULL, "hlac_ht_chantry_2534_36y.dds"},
	{0xEFEA7ABED209D18DULL, "hlac_lt_0_lowtown_docks_n_-3_2x.dds"},
	{0xEFEAA7E1268BA839ULL, "6026966_m.fxe"},
	{0xEFEAB86A8107F62DULL, "art\\characters\\playercharacter\\humanfemale\\cth_coma.msh"},
	{0xEFEAC66A81080E60ULL, "art\\characters\\playercharacter\\humanfemale\\cth_coma.mao"},
	{0xEFEACA6A810814B3ULL, "art\\characters\\playercharacter\\humanfemale\\cth_coma.mmh"},
	{0xEFEB22EA8A4CE724ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medb.phy"},
	{0xEFEB808F93D0FFABULL, "l_qun221ar_chantry.wbd"},
	{0xEFEBA7ECC5CA8B60ULL, "vxlt_0_lowtown_a2_eff_0_0.vlm"},
	{0xEFEBC1BC58E3A20BULL, "art\\characters\\weapons\\staff\\textures\\w_stf_009a_0n.dds"},
	{0xEFECA774A5D476AEULL, "l_mag101ar_caverns.bnk"},
	{0xEFECB840062FBB0DULL, "follower_banter6141936.cl"},
	{0xEFECC2FC1E3BE04EULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnl180walkf.ani"},
	{0xEFECECC3987350C4ULL, "vis121_saemus6101283.cl"},
	{0xEFED0106BEF0460DULL, "rog_combobreaker.dds"},
	{0xEFED2D4786C3E77DULL, "hlac_lt_0_lowtown_a2_65536_10z.dds"},
	{0xEFED7761C637E9D5ULL, "vxht_keep_736_0_0.vlm"},
	{0xEFED83E376CA1FC7ULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetnew_d.dds"},
	{0xEFEDABA759F228EFULL, "follower_barks6153822.cl"},
	{0xEFEDB8E1F4497062ULL, "follower_banter6171066.cl"},
	{0xEFEDEB91160E81E0ULL, "6117417_f.fxe"},
	{0xEFEE035FC00AB37DULL, "vxdr_primevalentrance_236_0_0.vlm"},
	{0xEFEE204CC0CC4280ULL, "gen00fl_aveline6063550.cl"},
	{0xEFEE29C6A315F16CULL, "6095118_m.fxe"},
	{0xEFEE3B597427F6B4ULL, "6048750_m.fxe"},
	{0xEFEE835FB6749DAFULL, "llac_lt_0_lowtown_docks_a3_n_65535_3.dds"},
	{0xEFEEA374CE60BB7EULL, "hlac_lt_0_lowtown_a3_n_c_-65536_5z.dds"},
	{0xEFEEC19C55E855C7ULL, "hlac_lt_0_lowtown_a3_n_c_-262144_4x.dds"},
	{0xEFEED29665DC6C12ULL, "427685767.ogg"},
	{0xEFEF25E9AC04C102ULL, "vxht_bartrand_n_678_0_0.vlm"},
	{0xEFEFBD0DFD38774CULL, "qun221_petrice6098751.cl"},
	{0xEFEFC980ACFCCC74ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_act_c.impa.ani"},
	{0xEFF04D27C6B7B2E8ULL, "395762989.ogg"},
	{0xEFF185035922F078ULL, "292422999.ogg"},
	{0xEFF2215D5F38FBBEULL, "hlac_lt_0_lowtown_docks_-196610_2x.dds"},
	{0xEFF22A7C55DEF1B9ULL, "6046542_m.fxe"},
	{0xEFF24598C2B7A1A6ULL, "6076032_m.fxe"},
	{0xEFF3026DE773B1FCULL, "ser120_ghyslain6104806.cl"},
	{0xEFF3638986969A49ULL, "6097225_m.fxe"},
	{0xEFF386ED138538FCULL, "6174876_f.fxe"},
	{0xEFF3AD0B0DD86A45ULL, "hlac_ht_chantry_n_2534_50z.dds"},
	{0xEFF3C3F517F4C09BULL, "vxlt_0_lowtown_a2_n_db3_0_0.vlm"},
	{0xEFF3EA3832856ACEULL, "458958780.ogg"},
	{0xEFF46FF41B44BAE2ULL, "641864924.ogg"},
	{0xEFF47EBF2986D832ULL, "vxht_keep_n_5wt_0_0.vlm"},
	{0xEFF49BD3250D7059ULL, "one100dg_kelder6055447.cl"},
	{0xEFF501B9F0A7FCABULL, "bld_mac_clu2.mao"},
	{0xEFF5355DC8D20232ULL, "473357470.ogg"},
	{0xEFF5779FB4A83F0AULL, "vxw_blightlands_s0_0_0.vlm"},
	{0xEFF5BA099982D02BULL, "vxht_keep_5ds_0_0.vlm"},
	{0xEFF5C680AD0132D6ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_act_c.impa.gad"},
	{0xEFF5DC41A4AEA2E3ULL, "hlac_lt_0_lowtown_a2_n_c_196606_0y.dds"},
	{0xEFF640265214CAA8ULL, "6075394_m.fxe"},
	{0xEFF64F410844B8FCULL, "vxlt_0_lowtown_a2_n_c_el9_0_0.vlm"},
	{0xEFF6A9749AACA6C3ULL, "hlac_lt_0_lowtown_a2_n_c_131072_0y.dds"},
	{0xEFF715C497DF5CFDULL, "431250446.ogg"},
	{0xEFF7F15099423F36ULL, "42172776.ogg"},
	{0xEFF83FFC37390943ULL, "vxw_sundermount_217_0_0.vlm"},
	{0xEFF8640C7E96F0E2ULL, "6179764_m.fxe"},
	{0xEFF87C16D9541C46ULL, "vxlt_0_lowtown_had_0_0.vlm"},
	{0xEFF889A71684BD14ULL, "art\\characters\\creatures\\ragedemon\\animation\\crn_rdy_rf.ani"},
	{0xEFF88DF18EAE8314ULL, "llac_ht_brothel_n_2995_4.dds"},
	{0xEFF9BB496A3B1199ULL, "zz_getsetplotflag.ncs"},
	{0xEFF9DF5C9D59B1E9ULL, "character_character_p13.psh.fxo.dx11.assembly"},
	{0xEFFA8338F9120AB7ULL, "ht_bartrand_n_mmap_ph.dds"},
	{0xEFFABAFC1E480F9FULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnl180walkf.evt"},
	{0xEFFAC52E7F89BB44ULL, "vxca_caverns1_dpq_0_0.vlm"},
	{0xEFFAD30C85CA97F5ULL, "ht_bartrand_n_3382.anb"},
	{0xEFFB19357B83CA07ULL, "hlac_w_encounter3_-131073_0z.dds"},
	{0xEFFB2E1C643DC9CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v2.evt"},
	{0xEFFB3E5884605B3FULL, "vxlt_hangedman_n_1vt_0_0.vlm"},
	{0xEFFB9D81CEB23809ULL, "vxw_bonepitmining_a2_2ap_0_0.vlm"},
	{0xEFFBC5246415C168ULL, "1050033390.ogg"},
	{0xEFFC7A9E4DEE86E0ULL, "vxht_keep_n_6ga_0_0.vlm"},
	{0xEFFC9AD18C8F0EAAULL, "687266179.ogg"},
	{0xEFFCE180AD077F35ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_act_c.impa.evt"},
	{0xEFFCF936F871CCAEULL, "vxht_keep_6hm_0_0.vlm"},
	{0xEFFD0AAEA611F931ULL, "fx_fadecloud_trans.dds"},
	{0xEFFD18E9FE739071ULL, "164276139.ogg"},
	{0xEFFD68A1F57415A3ULL, "plt_lgt000pt_ultra_3_05.plo"},
	{0xEFFDA705007C9E54ULL, "hlac_lt_0_lowtown_a2_n_c_-131072_5z.dds"},
	{0xEFFDF66CD6B1E7F9ULL, "vxlt_0_lowtown_a2_n_bz7_0_0.vlm"},
	{0xEFFE1D91019AF55EULL, "vxca_caverns1_j41_0_0.vlm"},
	{0xEFFE49994C47491AULL, "lgt250_way_dalish_conflict6147146.cl"},
	{0xEFFEB6334AAFDA0EULL, "vxlt_0_lowtown_docks_a3_n_c2k_0_0.vlm"},
	{0xEFFEBBF64C409749ULL, "lgt000cr_baz_den_rev2.utc"},
	{0xEFFF809DEA505632ULL, "vxht_brothel_n_4vi_0_0.vlm"},
	{0xEFFF86A7168AD776ULL, "art\\characters\\creatures\\ragedemon\\animation\\crn_rdy_rf.gad"},
	{0xEFFFA94880B06349ULL, "follower_banter6163226.cl"},
	{0xEFFFEA340C3DD6B6ULL, "w_bonepitdestroyed_a3_65_0.mao"},
	{0xEFFFF6340C3DEB55ULL, "w_bonepitdestroyed_a3_65_0.mmh"},
	{0xEFFFF8340C3DEECBULL, "w_bonepitdestroyed_a3_65_0.msh"},
	{0xEFFFFAF16FE65D2FULL, "character_character_p48.psh.fxo"},
	{0xF0006E01500003CBULL, "vxlt_0_lowtown_a2_n_c_e1m_0_0.vlm"},
	{0xF0009C5947D50A39ULL, "6041442_m.fxe"},
	{0xF000ACDB2A1DFC38ULL, "t1_skn_q01.tnt"},
	{0xF000B9130069A2B4ULL, "vxht_mansion2_78_0_0.vlm"},
	{0xF001022421B2670DULL, "61074238.ogg"},
	{0xF0018861A238DCBCULL, "926249149.ogg"},
	{0xF0019BFC1E4DF988ULL, "art\\characters\\creatures\\stonegolem\\animation\\pauseturnl180walkf.gad"},
	{0xF001B95D9C410B11ULL, "mag110_arianni6020496.cl"},
	{0xF0021AA4CFBCFF58ULL, "6054359_m.fxe"},
	{0xF0023AA435540BAEULL, "vxht_0_hightown_n_bgg_0_0.vlm"},
	{0xF002471C6444128CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v2.gad"},
	{0xF002F09C1B7CF0B2ULL, "vxht_keep_n_5xu_0_0.vlm"},
	{0xF003E97CA69E578AULL, "vxw_blightlands_t1_0_0.vlm"},
	{0xF003E9AC5FE47F56ULL, "hlac_lt_0_lowtown_docks_-131075_2z.dds"},
	{0xF00412A5BBCCCCABULL, "vxdr_primevaltomb_847_0_0.vlm"},
	{0xF004466B142454B8ULL, "llac_ht_mansion2_818_13.dds"},
	{0xF004D842ED507F67ULL, "vxlt_gallowstemplar_n_5k2_0_0.vlm"},
	{0xF004E0BF317B7A33ULL, "6208356_f.fxe"},
	{0xF004F3B1B4E602A8ULL, "vxht_bartrand_n_5p8_0_0.vlm"},
	{0xF0054E8193E2E9C8ULL, "6173805_m.fxe"},
	{0xF005E63A354B7333ULL, "hlac_lt_undercityencounter_-1_13y.dds"},
	{0xF0063204596FA567ULL, "dae000cr_gossip_old_m1.utc"},
	{0xF0075DD07CCAC30EULL, "follower_banter6169157.cl"},
	{0xF0077CA1D3DBED6DULL, "351458913.ogg"},
	{0xF007EFB3E0C56D35ULL, "6171929_m.fxe"},
	{0xF008461C64487C8AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stggr_v2.ani"},
	{0xF0086BE076CD5561ULL, "6096338_m.fxe"},
	{0xF0095F2B1CE3ADA2ULL, "vxca_caverns2_4y2_0_1.vlm"},
	{0xF009D3FB8D7A2F31ULL, "6166983_m.fxe"},
	{0xF00A290511D9463AULL, "6021667_m.fxe"},
	{0xF00A3C02EA52CEA9ULL, "289371762.ogg"},
	{0xF00AC15E44FDE833ULL, "arc_mit_07.mao"},
	{0xF00AEC7B3FE49F60ULL, "vxht_keep_n_6bn_0_0.vlm"},
	{0xF00AFF2BB7B94DEFULL, "851332632.ogg"},
	{0xF00B4FB044D31A04ULL, "142009439.ogg"},
	{0xF00B7834FEA710EDULL, "hf_keran.mor"},
	{0xF00B8594941B1995ULL, "6017809_m.fxe"},
	{0xF00C2768167E518CULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_glv_avea_0s.dds"},
	{0xF00C52FB1FB4B584ULL, "art\\levels\\races\\proxy\\props\\_textures\\logpile_d.dds"},
	{0xF00C847023138E64ULL, "vxca_caverns1_ec1_0_0.vlm"},
	{0xF00D50ED27E4072BULL, "hlac_dr_dungeon_3217_11z.dds"},
	{0xF00D8A8190EB2BCBULL, "vxlt_0_lowtown_h86_0_0.vlm"},
	{0xF00D93BD2564DDD9ULL, "l_dae211ar_tavern.wbd"},
	{0xF00E29020A2C8AAAULL, "vxlt_gallowscourtyard_640_0_0.vlm"},
	{0xF00E292E5E552C44ULL, "w_woundedcoast_57_0.rml"},
	{0xF00E4E4595F52569ULL, "317086219.ogg"},
	{0xF00E5443F066E79BULL, "gen00fl_aveline6114091.cl"},
	{0xF00E6CCE61DC75AFULL, "character_character_p39.psh.fxo"},
	{0xF00E877FDB1E9339ULL, "follower_banter6155766.cl"},
	{0xF00EA65A641726C4ULL, "6104182_m.fxe"},
	{0xF00EC70CD3445D95ULL, "515352149.ogg"},
	{0xF00EE09175F73910ULL, "art\\characters\\playercharacter\\elfmale\\bdy_flsb.phy"},
	{0xF00F4DBD03A2F65BULL, "6093660_f.fxe"},
	{0xF00FAF428CDBAB08ULL, "796753098.ogg"},
	{0xF00FD6C2B0990FEDULL, "6034183_m.fxe"},
	{0xF0101C7C0ADC7868ULL, "6176657_m.fxe"},
	{0xF01023072C99461EULL, "862284022.ogg"},
	{0xF010628F93EFE979ULL, "l_qun221ar_chantry.bnk"},
	{0xF010988CD4E923E8ULL, "6079050_f.fxe"},
	{0xF010C02CC2A2B902ULL, "140161188.ogg"},
	{0xF010DE342EFCF7C5ULL, "6091933_m.fxe"},
	{0xF010EC52C0E45A79ULL, "6068270_m.fxe"},
	{0xF01151CA60B48081ULL, "6075856_m.fxe"},
	{0xF011B307B765F7EFULL, "vxca_caverns1_ija_0_0.vlm"},
	{0xF011BA332A020EFDULL, "vxht_0_hightown_a2_bya_0_0.vlm"},
	{0xF011EC07BF65D2F3ULL, "vxca_caverns1_mines_i1i_0_0.vlm"},
	{0xF011F83AE20A1CEAULL, "follower_banter6169478.cl"},
	{0xF01257BE61A7BB90ULL, "follower_banter6171107.cl"},
	{0xF01269EEEE1C4326ULL, "586446164.ogg"},
	{0xF0127783D2A0A24BULL, "llac_lt_warehouse_1_3.dds"},
	{0xF0127F3A2EF0E7D2ULL, "mag121_idunna6084263.cl"},
	{0xF012ADDC09E94913ULL, "vxdr_primevaltomb_2wq_0_1.vlm"},
	{0xF012BB38B43BB5ACULL, "vxlt_0_lowtown_a3_n_c_cmm_0_0.vlm"},
	{0xF012BC40DE96D566ULL, "6134525_m.fxe"},
	{0xF0133E01C808E6DBULL, "follower_banter6148735.cl"},
	{0xF013E7EB6FB1E004ULL, "vxht_chantry_351_0_0.vlm"},
	{0xF014C248657A4A46ULL, "6118202_m.fxe"},
	{0xF01546B428361A35ULL, "6039019_f.fxe"},
	{0xF015C81D37CDAA15ULL, "194226481.ogg"},
	{0xF015EDF72A9643AFULL, "161904547.ogg"},
	{0xF0161734F383242EULL, "128247279.ogg"},
	{0xF016196059ABA7B7ULL, "vxlt_gallowstemplar_f_6qg_0_0.vlm"},
	{0xF01730605325D891ULL, "abi_r_stun_gren.wbd"},
	{0xF017EA8F97B121C2ULL, "hlac_ht_mansion2_818_66x.dds"},
	{0xF017EC4C94A56C7DULL, "vxlt_hangedman_2ek_0_0.vlm"},
	{0xF0186DC49A90FAAEULL, "shieldgreen_test.tnt"},
	{0xF01912E7CFC59FF7ULL, "198544983.ogg"},
	{0xF0195F187D6B2274ULL, "gen00fl_merrill6129598.cl"},
	{0xF01A2903BB1E801EULL, "38897229.ogg"},
	{0xF01A670A4DC6C8F4ULL, "146124514.ogg"},
	{0xF01A7EA3D717DD21ULL, "qun221_petrice6082812.cl"},
	{0xF01B72CEEEFB2BDAULL, "750353964.ogg"},
	{0xF01C1BAEA62CC6DBULL, "fx_fadecloud_trans.mao"},
	{0xF01D1EA7CBB579B6ULL, "vxlt_0_lowtown_docks_a3_n_9ra_0_0.vlm"},
	{0xF01D39A8A584D1FEULL, "zz_isa_debug6175236.cl"},
	{0xF01D57022494B748ULL, "gen00fl_merrill6121334.cl"},
	{0xF01D8FC300B5E1E0ULL, "6127985_m.fxe"},
	{0xF01DC2161B978A7CULL, "329281812.ogg"},
	{0xF01DDB1382B433EAULL, "art\\levels\\races\\proxy\\props\\storage\\prp_zzbarrel_0.pwk"},
	{0xF01E420721396833ULL, "vxdr_0_deeproads_ceu_0_0.vlm"},
	{0xF01E489FA28F286DULL, "6031283_m.fxe"},
	{0xF01EC8847D4EF83CULL, "vxlt_0_lowtown_docks_a3_bfg_0_0.vlm"},
	{0xF01ED5F121107283ULL, "390414200.ogg"},
	{0xF01F501120F31045ULL, "6096833_m.fxe"},
	{0xF01FBA6EBDDF45C4ULL, "one140cr_charade6128523.cl"},
	{0xF0204A6B5317B07DULL, "w_bonepitdestroyed_a3_121_0.phy"},
	{0xF0206B61C7A42940ULL, "6188595_m.fxe"},
	{0xF02090835118ACAFULL, "198327748.ogg"},
	{0xF020C520B4BF7074ULL, "llac_lt_0_lowtown_n_-2_1.dds"},
	{0xF0216E0B97DBCB1CULL, "225188024.ogg"},
	{0xF022222B13D3EF29ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_3.pwk"},
	{0xF02269636B0977BDULL, "vxht_chantry_27k_0_0.vlm"},
	{0xF0228DAB6A340EC2ULL, "gen00fl_aveline6140026.cl"},
	{0xF022C272E3606DD4ULL, "gen_im_trs_drk_two_2ha_r2.uti"},
	{0xF022F5795C94F17DULL, "320171418.ogg"},
	{0xF022F597D7FD805DULL, "hlac_lt_blackemporium_1055_1z.dds"},
	{0xF0233425577062C6ULL, "6117302_m.fxe"},
	{0xF023B506E613CF37ULL, "vxdr_primevaltomb_6nm_0_2.vlm"},
	{0xF0245B0D05F6693AULL, "vxlt_warehouse_11k_0_0.vlm"},
	{0xF025245E38FF708CULL, "vxht_brothel_n_34f_0_0.vlm"},
	{0xF025DFFFF359F64EULL, "dae010_cityguard_m1.utc"},
	{0xF0265E29869B84FDULL, "vxlt_hangedman_2bj_0_0.vlm"},
	{0xF0266A0C0BDB13D8ULL, "art\\levels\\races\\dwarf\\props\\spider\\fern.mao"},
	{0xF0268FFF21EE3A1AULL, "6171513_m.fxe"},
	{0xF02695151C4B7915ULL, "6016687_f.fxe"},
	{0xF0271C89B966098FULL, "6126581_m.fxe"},
	{0xF02760AD922D6BA4ULL, "zz_add_all_rogue.ncs"},
	{0xF0278088EF53A437ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\128x128.dds"},
	{0xF0279DF40525B8B1ULL, "746253639.ogg"},
	{0xF0281472F2649109ULL, "vxht_brothel_n_3nm_0_0.vlm"},
	{0xF028EBDC0F7B0996ULL, "vxca_caverns2prc_8zg_0_0.vlm"},
	{0xF02912F0F090B333ULL, "vxht_chantry_2as_0_0.vlm"},
	{0xF0293DC9CA093686ULL, "272844267.ogg"},
	{0xF029682A06FFBF64ULL, "vxca_caverns1_ei7_0_0.vlm"},
	{0xF0298B4A4B413301ULL, "sound_weapon_actions.gda"},
	{0xF029A99491118954ULL, "bkl_ciga.mao"},
	{0xF029FFB4C2D05520ULL, "gen00fl_aveline6139745.cl"},
	{0xF02A138E99E0A7A2ULL, "23400966.ogg"},
	{0xF02A21BF3C4365B3ULL, "6054972_m.fxe"},
	{0xF02A31E47AD98235ULL, "hlac_w_encounter1_65535_0z.dds"},
	{0xF02A74ABEDD9CEDBULL, "87842736.ogg"},
	{0xF02A87155E6A17D9ULL, "gen00fl_isabela6100017.cl"},
	{0xF02AA35949A9B4D6ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blightrubble_darelief.dds"},
	{0xF02B03EB26D584C6ULL, "zz_followers_banter6159660.cl"},
	{0xF02B6D448841E188ULL, "vxdr_0_deeproads_6qv_0_1.vlm"},
	{0xF02B9EC7BB01F83FULL, "vxht_mansion_n_3sg_0_0.vlm"},
	{0xF02D00A1CAAAE323ULL, "gen00fl_fenris6115372.cl"},
	{0xF02D348920BE90CFULL, "6017762_m.fxe"},
	{0xF02D8DB08FDB5BCCULL, "vxlt_0_lowtown_docks_a3_n_d4t_0_0.vlm"},
	{0xF02DC732B7420206ULL, "vxw_woundedcoast_1dw_0_1.vlm"},
	{0xF02E40A29854A325ULL, "371071543.ogg"},
	{0xF02E63CD5B053CD8ULL, "ave331_jeven6152183.cl"},
	{0xF02E83B842E72CD3ULL, "6060311_m.fxe"},
	{0xF02EA14DA6267553ULL, "vxca_caverns1_mines_jfw_0_0.vlm"},
	{0xF02EEE009C728708ULL, "llac_lt_gallowscourtyard_65539_0.dds"},
	{0xF02F6E616A6CB0E6ULL, "6081204_f.fxe"},
	{0xF02F817D615B322BULL, "vxht_keep_5xw_0_0.vlm"},
	{0xF02FBE1F8F11B253ULL, "6050082_m.fxe"},
	{0xF02FE1F279C3AAA9ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st02a_0d.dds"},
	{0xF0300E3D537E368CULL, "mag121_idunna6084900.cl"},
	{0xF03103726EEC9FCCULL, "99800079.ogg"},
	{0xF0311650C406155FULL, "zz_ave_debug6064132.cl"},
	{0xF0313F6B532610E9ULL, "w_bonepitdestroyed_a3_121_0.mao"},
	{0xF03149D6418BE779ULL, "794098336.ogg"},
	{0xF0314B6B53262482ULL, "w_bonepitdestroyed_a3_121_0.mmh"},
	{0xF0314D6B53262838ULL, "w_bonepitdestroyed_a3_121_0.msh"},
	{0xF031569A185CDEA0ULL, "181424640.ogg"},
	{0xF03261CC8CBF4944ULL, "plt_cod_ltr_sk3_scribbles2.plo"},
	{0xF032A5402A41CAD0ULL, "1060702066.ogg"},
	{0xF032ADCAC0D4DC44ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.cess.ani"},
	{0xF032C898FC154673ULL, "vxlt_gallowstemplar_f_34m_0_0.vlm"},
	{0xF032CCE9F7EC81BAULL, "vxlt_warehouse_14d_0_0.vlm"},
	{0xF033418123999EF0ULL, "and230tr_worldmap.ncs"},
	{0xF0338BE39EEFFB08ULL, "zz_qun_debug6085866.cl"},
	{0xF033F9B2F2CF5527ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k.gad"},
	{0xF0345F7553750B2BULL, "hlac_lt_0_lowtown_a2_n_c_-65535_4x.dds"},
	{0xF03505CC4C38BD3FULL, "vxlt_hangedman_n_1bp_0_0.vlm"},
	{0xF03593FF4A86B667ULL, "vxlt_0_lowtown_docks_n_bdz_0_0.vlm"},
	{0xF03656F45B850A87ULL, "699746202.ogg"},
	{0xF0366086CC590CA8ULL, "6182720_m.fxe"},
	{0xF0366CAEA56694F1ULL, "fighting60a.mao"},
	{0xF0366F6431A4330FULL, "vxlt_0_lowtown_a2_gjs_0_0.vlm"},
	{0xF036B1B603695DF5ULL, "character_character_p48.psh.fxo.dx11"},
	{0xF036FADDF1E07178ULL, "6139775_m.fxe"},
	{0xF037421EEAC743A6ULL, "296178430.ogg"},
	{0xF0377FAB21EA19DBULL, "zz_vrc_debug6200243.cl"},
	{0xF0378A12E3E2C26DULL, "hlac_w_woundedcoast_-196605_0x.dds"},
	{0xF037EB6174D5536FULL, "vxlt_warehouse_q3_0_0.vlm"},
	{0xF03815E711510E9CULL, "860229041.ogg"},
	{0xF0382203677B13D2ULL, "vxca_caverns1_mines_jfr_0_0.vlm"},
	{0xF039AACAC0DAF5A6ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.cess.gad"},
	{0xF03A737CE4A212E0ULL, "6129785_m.fxe"},
	{0xF03AA9AC46FFE66CULL, "llac_lt_gallowstemplar_2395_1.dds"},
	{0xF03AEEB2F2D56134ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k.evt"},
	{0xF03B9A06BA7D35D4ULL, "darktown_fem_fereldan_sta6156750.cl"},
	{0xF03BF0973283190BULL, "hlac_lt_0_lowtown_docks_65533_4y.dds"},
	{0xF03C1819A017D179ULL, "6056723_m.fxe"},
	{0xF03C5A342DF8DAD0ULL, "zz_one_debug6078376.cl"},
	{0xF03D5F3CCD20C714ULL, "vxht_0_hightown_b90_0_0.vlm"},
	{0xF03DA621D0C38B88ULL, "llac_ca_caverns2_5802_72.dds"},
	{0xF03DBB5797044EB3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hugekeg.phy"},
	{0xF03DF37319759DE9ULL, "one200dg_solivitus6066515.cl"},
	{0xF03EB086577D2020ULL, "and230_mage_girl.cnv"},
	{0xF03F03B0B4818F43ULL, "vxw_encounter2_3b_0_0.vlm"},
	{0xF03F1A8D2640B519ULL, "6116074_m.fxe"},
	{0xF03FA2B255EA0280ULL, "1027598190.ogg"},
	{0xF03FB4B75E1352FEULL, "t1_mub_br2.tnt"},
	{0xF04067C301D6E396ULL, "fen200cr_tev_hunter_1.utc"},
	{0xF040A5CAC0E10B85ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.cess.evt"},
	{0xF0410FDAB516475FULL, "vxlt_0_lowtown_docks_a3_9rc_0_0.vlm"},
	{0xF041401F98BE0B52ULL, "llac_lt_0_lowtown_docks_a3_-2_6.dds"},
	{0xF04144DF85819FD3ULL, "w_encounter2_ll.mmh"},
	{0xF04157B2BC9DBFDAULL, "hlac_ht_brothel_n_3369_2z.dds"},
	{0xF0419286577F2358ULL, "and230_mage_girl.bnk"},
	{0xF0420066BA3DB22AULL, "gencr_gate_slave_r3.utc"},
	{0xF0424A7EEB7B3D19ULL, "324052234.ogg"},
	{0xF043AC9B9FCA57A2ULL, "one300_nuncio6076782.cl"},
	{0xF04405DC3C7CCEE7ULL, "vxlt_0_lowtown_docks_n_bau_0_0.vlm"},
	{0xF04410CF25438A16ULL, "zz_followers_banter6158946.cl"},
	{0xF04462B59D73D97FULL, "vxlt_0_lowtown_a2_gv7_0_0.vlm"},
	{0xF044D21137C7A03FULL, "785690929.ogg"},
	{0xF044DB11A43A399BULL, "vxlt_0_lowtown_docks_n_cov_0_0.vlm"},
	{0xF04527D8C931EA90ULL, "6140202_f.fxe"},
	{0xF04535AA7996F67CULL, "6095455_m.fxe"},
	{0xF0457064617A51A7ULL, "6170498_m.fxe"},
	{0xF046116178091DEAULL, "lla_ht_0_hightown_-65536_0.dds"},
	{0xF04643AA235555D6ULL, "120847783.ogg"},
	{0xF0465D3E66CB6BEFULL, "vxlt_warehouse_z2_0_0.vlm"},
	{0xF046F75341773CA3ULL, "670993861.ogg"},
	{0xF0474245773EEE53ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4_tool.phy"},
	{0xF0474EF3789CBE55ULL, "6095753_m.fxe"},
	{0xF047763F7E368286ULL, "gen00fl_sebastian6180415.cl"},
	{0xF047A8E76D009833ULL, "l_mag321ar_store.bnk"},
	{0xF04843DEBC7D0374ULL, "650098000.ogg"},
	{0xF048E6B2F2E1916DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_k.ani"},
	{0xF04940C405F2A87CULL, "62480374.ogg"},
	{0xF049DA48195F15F1ULL, "vxlt_0_lowtown_docks_a3_95h_0_0.vlm"},
	{0xF04A2128EB0C81F7ULL, "vxlt_darktown_48e_0_0.vlm"},
	{0xF04B7AFF2E7651E6ULL, "hlac_ht_0_hightown_n_131071_1x.dds"},
	{0xF04C096C2A382604ULL, "llac_ht_mansion2b_818_59.dds"},
	{0xF04C297C04419231ULL, "dae011_wench_mob6142287.cl"},
	{0xF04C3AFC8D0E25CBULL, "vxht_chantry_n_2s2_0_0.vlm"},
	{0xF04C913AC7C32027ULL, "ht_chantry_2534.rml"},
	{0xF04CCD61D1437068ULL, "685443424.ogg"},
	{0xF04CDF9727AD91A2ULL, "vxca_cavern1_haunted_j1w_0_0.vlm"},
	{0xF04CE61579E09981ULL, "vxlt_warehouse_n_5x_0_0.vlm"},
	{0xF04D199BF21318B4ULL, "6176639_m.fxe"},
	{0xF04D21D9D5486AF1ULL, "hlac_w_encounter3_-65536_0z.dds"},
	{0xF04D2D026EB47D28ULL, "vxht_bartrand_n_5o5_0_0.vlm"},
	{0xF04D450AF5812F57ULL, "6089691_m.fxe"},
	{0xF04D6A1F2C9391B8ULL, "mag211_marethari6021350.cl"},
	{0xF04D9048CB06442CULL, "361709222.ogg"},
	{0xF04E5BEF3D52EAC9ULL, "825846426.ogg"},
	{0xF04E923696425037ULL, "zz_vault_debug6203386.cl"},
	{0xF04ECBCEA60EC6B3ULL, "583253681.ogg"},
	{0xF04F60FC8B23D743ULL, "dae000_bartrand6206879.cl"},
	{0xF050213CA5D78AEAULL, "vxht_mansion2b_2p9_0_0.vlm"},
	{0xF050370592A49C02ULL, "qcr221ip_throne_door.utp"},
	{0xF050732367943F16ULL, "6120822_f.fxe"},
	{0xF050C44538F14B72ULL, "pro000_meeran6048844.cl"},
	{0xF051199D45D3C5E9ULL, "329216596.ogg"},
	{0xF05146C70918CA18ULL, "6141844_m.fxe"},
	{0xF051776424DCDBAAULL, "gencr_rage_demon.utc"},
	{0xF051CC200709FBC8ULL, "6118200_m.fxe"},
	{0xF0525BCF75950B26ULL, "93237660.ogg"},
	{0xF052AC2E06E6DE20ULL, "vxlt_undercityserialkiller_1c8_0_0.vlm"},
	{0xF0536281D209779AULL, "6116078_f.fxe"},
	{0xF05364B34190A663ULL, "hlac_ht_chantry_n_2534_53y.dds"},
	{0xF053769D7892ED65ULL, "6135997_f.fxe"},
	{0xF0538C1E6830E704ULL, "225333253.ogg"},
	{0xF053F84B152993CFULL, "hlac_dr_dungeon_3214_0y.dds"},
	{0xF05460D4810637CFULL, "vxht_0_hightown_a3_c9s_0_0.vlm"},
	{0xF054650A9EB06AF3ULL, "802841749.ogg"},
	{0xF0554879489EE952ULL, "740884777.ogg"},
	{0xF05605156CAB5F36ULL, "hlac_lt_0_lowtown_docks_a3_-262144_2y.dds"},
	{0xF05664B40B9AFC61ULL, "gen_im_trs_prf_sta_ele_r1.uti"},
	{0xF056C3ED299FAE1CULL, "gen_im_arm_glv_hvy_act2_01.uti"},
	{0xF056FA81331B5EA8ULL, "hlac_lt_0_lowtown_docks_n_65535_3x.dds"},
	{0xF05798ACC2C7CD3EULL, "923935535.ogg"},
	{0xF057FBEEAD15F14DULL, "876648268.ogg"},
	{0xF0589EDCBE5232A1ULL, "vxlt_gallowscourtyard_n_75v_0_0.vlm"},
	{0xF058AC18819135E3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_rob_maga_0s.dds"},
	{0xF058B9681DDF2885ULL, "1020851637.ogg"},
	{0xF05929EBD1D82412ULL, "vxht_mansion_n_5dm_0_0.vlm"},
	{0xF059FAD090D9742EULL, "515219706.ogg"},
	{0xF05A019C0FCEA961ULL, "6099982_m.fxe"},
	{0xF05AB6BC526C1C5BULL, "165490952.ogg"},
	{0xF05B769F379323BFULL, "61601577.ogg"},
	{0xF05B8E8CFE2B3687ULL, "gen00fl_anders6103782.cl"},
	{0xF05BC02A33724A58ULL, "6147895_m.fxe"},
	{0xF05BE76A986DC043ULL, "vxw_encounter2_9d_0_0.vlm"},
	{0xF05C135ADBA6C62DULL, "qcr000_varric6208261.cl"},
	{0xF05C42654BA4F109ULL, "hlac_lt_0_lowtown_docks_a3_-131072_0y.dds"},
	{0xF05C4C2784412876ULL, "398105235.ogg"},
	{0xF05D23FFDED15912ULL, "409138843.ogg"},
	{0xF05D70BCCE3A89BBULL, "6167212_m.fxe"},
	{0xF05E4CF66B7DC8B7ULL, "follower_banter6170325.cl"},
	{0xF05EAFCB595F578AULL, "6055109_m.fxe"},
	{0xF05ECB13BE4D0E47ULL, "365068104.ogg"},
	{0xF05ED91D4E2192A9ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_hunter_fenrintro.gad"},
	{0xF05EF6819175DB1AULL, "lgt220_shp_mag_guard6149902.cl"},
	{0xF05F0142E2665A02ULL, "ht_keep_4519.rml"},
	{0xF05F06AC933A0D3EULL, "vxlt_undercityencounter_zz_0_0.vlm"},
	{0xF05F2F3CC7C2721EULL, "vxlt_gallowstemplar_f_4uh_0_0.vlm"},
	{0xF05F69A06CA07810ULL, "one100dg_elren6014253.cl"},
	{0xF0600B9101D03082ULL, "pro000_athenril6048733.cl"},
	{0xF0601CA831E86E5BULL, "6116989_f.fxe"},
	{0xF0603DFCF9001448ULL, "6117300_m.fxe"},
	{0xF060866983C30B9EULL, "6071172_m.fxe"},
	{0xF060FFE832900DA1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maga_0s.dds"},
	{0xF06189C302DDFF88ULL, "cd_and121_karl.bnk"},
	{0xF061DBE6BA3CF309ULL, "vxht_brothel_n_3zq_0_0.vlm"},
	{0xF0627BA8D29DFE0CULL, "sib100_father_contact6110090.cl"},
	{0xF06296B2E25FC364ULL, "llac_w_bonepitdestroyed_a3_0_13.dds"},
	{0xF062D2B172FC504FULL, "vxht_0_hightown_a3_c2r_0_0.vlm"},
	{0xF062EA8A8DC242BBULL, "character_character_p20.psh.fxo.dx11"},
	{0xF062ECF1F3FE6151ULL, "6067687_f.fxe"},
	{0xF062F729A097897AULL, "vxlt_0_lowtown_docks_a3_d1z_0_0.vlm"},
	{0xF0633795A204BED5ULL, "vxw_encounter1_k9_0_0.vlm"},
	{0xF063A043C86484C7ULL, "gen00fl_anders6125621.cl"},
	{0xF0645DAF4FA7318CULL, "50226083.ogg"},
	{0xF064D386B2A53C10ULL, "vxlt_0_lowtown_a2_n_c_bxv_0_0.vlm"},
	{0xF06575C75580D8A3ULL, "ht_keep_mmap_ph.dds"},
	{0xF065AA1D4E2761F7ULL, "art\\cinematics\\follower plots\\fenris\\act 1\\fen110cs_fenris_intro\\cs_hunter_fenrintro.ani"},
	{0xF066C9D3CE852BC9ULL, "vxlt_gallowscourtyard_n_7i2_0_0.vlm"},
	{0xF066FAE905190116ULL, "469372827.ogg"},
	{0xF0677AAF53483664ULL, "6107323_m.fxe"},
	{0xF0679BC8C3CE3C92ULL, "vxht_mansion_n_5an_0_0.vlm"},
	{0xF068AF223EFEFC6DULL, "gen_im_cns_cft_pot_07.uti"},
	{0xF068D4FBA372651EULL, "vxht_bartrand_n_29v_0_0.vlm"},
	{0xF06925A921746802ULL, "vxht_0_hightown_a3_b8d_0_0.vlm"},
	{0xF06940CFC6ED2A70ULL, "vxlt_0_lowtown_n_eoz_0_0.vlm"},
	{0xF0694F8EA84DE674ULL, "6175322_m.fxe"},
	{0xF069654EBBDABC7AULL, "lowtown_guard_chase6169515.cl"},
	{0xF06A29C849C34B1BULL, "6001358_m.fxe"},
	{0xF06BFF88C75BF667ULL, "6066388_m.fxe"},
	{0xF06C0AFE6904E046ULL, "vxlt_0_lowtown_h9l_0_0.vlm"},
	{0xF06C293AA07F7916ULL, "98485721.ogg"},
	{0xF06C5F6B961D0039ULL, "6180510_m.fxe"},
	{0xF06D4FFC23066DC5ULL, "6133318_m.fxe"},
	{0xF06D8ADF538DA5B4ULL, "dae100_bran6118153.cl"},
	{0xF06D9695C02EE6D8ULL, "vxlt_0_lowtown_a2_n_c_gmh_0_0.vlm"},
	{0xF06DC145491259E8ULL, "764510825.ogg"},
	{0xF06DFD04671E3B1AULL, "6034041_m.fxe"},
	{0xF06E0F41110EB723ULL, "vxlt_0_lowtown_a3_n_c_hpr_0_0.vlm"},
	{0xF06E31B0FD992172ULL, "950919953.ogg"},
	{0xF06E72C73C0957A0ULL, "gen00fl_aveline6132861.cl"},
	{0xF06EF7DC86FC48C6ULL, "vxlt_0_lowtown_ev4_0_0.vlm"},
	{0xF06F2F26449B9E98ULL, "vxlt_0_lowtown_f54_0_0.vlm"},
	{0xF06F36DED2FA649AULL, "mrl150_merrill6093845.cl"},
	{0xF06F63FFE6DF135DULL, "hlac_lt_0_lowtown_docks_-65537_4z.dds"},
	{0xF06F739B67151057ULL, "327423716.ogg"},
	{0xF06F76CC8CF31A8BULL, "plt_cod_ltr_sk3_scribbles2.nss"},
	{0xF06FDCCCDBA2AE22ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_boo_roba_0s.dds"},
	{0xF06FE5A479F89FD9ULL, "914715710.ogg"},
	{0xF07013750F3F357EULL, "llac_ht_chantry_n_2534_11.dds"},
	{0xF07073260E54FF74ULL, "116857439.ogg"},
	{0xF0714F57012587F9ULL, "hlac_lt_gallowscourtyard_n_-1_4z.dds"},
	{0xF0722ACB068939EAULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_noba_0d.dds"},
	{0xF0736382D7654066ULL, "vxlt_undercityserialkiller_1zi_0_0.vlm"},
	{0xF073740F311F7963ULL, "hlac_ht_keep_n_4519_4z.dds"},
	{0xF07427CA2B5BCFCCULL, "6113025_m.fxe"},
	{0xF074AECF27AC1182ULL, "hlac_lt_0_lowtown_docks_a3_-131074_4x.dds"},
	{0xF074B7AB88264D97ULL, "lt_0_lowtown_a3_n_c_51_0_lc.mmh"},
	{0xF074DF2DFA14DCEDULL, "art\\characters\\weapons\\staff\\w_stf_st105a.mmh"},
	{0xF074F12DFA14FB73ULL, "art\\characters\\weapons\\staff\\w_stf_st105a.msh"},
	{0xF075241430AD4594ULL, "vxdr_primevaltomb_54i_0_0.vlm"},
	{0xF07614DF1CD9F3A3ULL, "vxca_cavern1_haunted_e59_0_0.vlm"},
	{0xF0761FFADFA2B7AAULL, "vxlt_0_lowtown_hah_0_0.vlm"},
	{0xF07699EEAD4A6841ULL, "487005669.ogg"},
	{0xF07705E4BE275D3DULL, "6167492_m.fxe"},
	{0xF0773C690EDBA651ULL, "vxht_0_hightown_a3_au7_0_0.vlm"},
	{0xF0775A37F4962D73ULL, "50866042.ogg"},
	{0xF07761645EEFA2A2ULL, "llac_lt_undercityserialkiller_-1_40.dds"},
	{0xF0777BF4469C5D93ULL, "6067112_m.fxe"},
	{0xF0785FB4BDC42D26ULL, "149562629.ogg"},
	{0xF078B197B793059BULL, "vxlt_0_lowtown_a2_n_gpb_0_0.vlm"},
	{0xF0797ED604244D86ULL, "llac_ca_caverns1_24424_18.dds"},
	{0xF07A3A55540AF912ULL, "vxw_bonepitdestroyed_a3_255_0_0.vlm"},
	{0xF07B4643D114D98AULL, "art\\vfx\\materials\\vfx_dust_blend.mao"},
	{0xF07B568EA5D524A1ULL, "llac_ht_playershouse_n_855_31.dds"},
	{0xF07BD10F63130A78ULL, "property_functions.gda"},
	{0xF07BF30D46386782ULL, "art\\vfx\\combat\\vfx_d_invulner_exit_c.mmh"},
	{0xF07C898669EF872DULL, "vxlt_hangedman_n_306_0_0.vlm"},
	{0xF07C8F8F2958DAD3ULL, "126277763.ogg"},
	{0xF07D649280BCD0F6ULL, "llac_ht_0_hightown_a2_-65534_1.dds"},
	{0xF07DB91D69710587ULL, "6109443_m.fxe"},
	{0xF07E6456D961249AULL, "darktown_fem_lowlife_mob6174142.cl"},
	{0xF07EA4F107FD10A3ULL, "780695310.ogg"},
	{0xF07ECD7A370FCE72ULL, "hlac_lt_undercityencounter_-1_28z.dds"},
	{0xF07ED78A0BE1B692ULL, "qun110_petrice6042144.cl"},
	{0xF07EF172AA884EBFULL, "6162251_m.fxe"},
	{0xF07FCE6BFDDC46D1ULL, "vxlt_0_lowtown_docks_b5a_0_1.vlm"},
	{0xF0801A6A5A9870BDULL, "6075799_m.fxe"},
	{0xF08046C078BBC52BULL, "hlac_lt_gallowstemplar_2382_1y.dds"},
	{0xF08105D10DE0FB64ULL, "6177575_m.fxe"},
	{0xF08105E37D71C488ULL, "539812213.ogg"},
	{0xF081097421C52793ULL, "6031747_f.fxe"},
	{0xF08125CF25CDD93EULL, "gen00fl_carver6097314.cl"},
	{0xF0812653249E3D89ULL, "6070107_m.fxe"},
	{0xF081388123DBB0FDULL, "and230tr_worldmap.utt"},
	{0xF0817F5B168F908DULL, "821576246.ogg"},
	{0xF081DC479CD9A8BBULL, "6086221_m.fxe"},
	{0xF081EA4E204C8DA1ULL, "347113680.ogg"},
	{0xF082257D484B0647ULL, "789990353.ogg"},
	{0xF08258818B67B5D5ULL, "mag291_vincento6027398.cl"},
	{0xF0829A70882F12FAULL, "vxlt_darktown_2ym_0_0.vlm"},
	{0xF0830B8E5C4C4947ULL, "6064654_m.fxe"},
	{0xF083E3891CBEAA97ULL, "197285234.ogg"},
	{0xF084B6AE1121CCB9ULL, "dae000_hate_enemies6167593.cl"},
	{0xF084CE68E461A923ULL, "vxdr_primevaltomb_50g_0_1.vlm"},
	{0xF0855DE9954E43E8ULL, "6028859_m.fxe"},
	{0xF086DB31A615BF33ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highbed_d.dds"},
	{0xF08778B04253362AULL, "776402134.ogg"},
	{0xF0877BC302FEB732ULL, "cd_and121_karl.wbd"},
	{0xF087F2BA387141A8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shielda_n.dds"},
	{0xF08823A194D1F87AULL, "267491638.ogg"},
	{0xF088AC3246011192ULL, "vxw_bonepitdestroyed_a3_206_0_0.vlm"},
	{0xF088EEB84CF11146ULL, "vxlt_0_lowtown_h3n_0_0.vlm"},
	{0xF089C06BEF455DE4ULL, "vxlt_0_lowtown_docks_a3_coy_0_0.vlm"},
	{0xF08A584959751F29ULL, "83307897.ogg"},
	{0xF08A7A4FA41B17D8ULL, "vxlt_0_lowtown_a2_n_c_gcf_0_0.vlm"},
	{0xF08AFF84FA76E1A0ULL, "vxht_bartrand_n_70b_0_0.vlm"},
	{0xF08B27612B62F0BDULL, "vxdr_0_deeproads_cwa_0_0.vlm"},
	{0xF08B2A112675E00FULL, "llac_lt_gallowstemplar_n_2392_0.dds"},
	{0xF08B4A49C9FD8135ULL, "6016657_m.fxe"},
	{0xF08B9688A05AB81CULL, "6145138_m.fxe"},
	{0xF08BB50313142281ULL, "hlac_w_sundermount_-6_0y.dds"},
	{0xF08BFFAEED26751CULL, "vxlt_0_lowtown_a2_ekl_0_1.vlm"},
	{0xF08C2AFA5B671E07ULL, "6102543_m.fxe"},
	{0xF08C60A076D68F7EULL, "lgt310_shw_ors_informant.cnv"},
	{0xF08C95BC3F392B6FULL, "6126390_m.fxe"},
	{0xF08CA3AB440D6414ULL, "mer231_brekker6075168.cl"},
	{0xF08D0A54BB505FE8ULL, "6095640_m.fxe"},
	{0xF08D289AF3AF212BULL, "vxht_brothel_n_5fy_0_0.vlm"},
	{0xF08D4F8D4142B9CCULL, "6080754_m.fxe"},
	{0xF08D6E82DA2859CFULL, "hlac_ca_cavern1_haunted_24423_9y.dds"},
	{0xF08D943AC7FABC3BULL, "ht_chantry_2534.anb"},
	{0xF08DEECDA50C18C5ULL, "vxlt_0_lowtown_docks_a3_d1q_0_0.vlm"},
	{0xF08E257874EF1837ULL, "452009478.ogg"},
	{0xF08E4943F4F5E529ULL, "417677526.ogg"},
	{0xF08EC6E76D3CCD91ULL, "l_mag321ar_store.wbd"},
	{0xF08F1A608473EB41ULL, "zz_dae_debug6047595.cl"},
	{0xF08F42A076D8918EULL, "lgt310_shw_ors_informant.bnk"},
	{0xF08F43B894897F11ULL, "hlac_dr_0_deeproads_5845_8y.dds"},
	{0xF08F5126E5199219ULL, "vxlt_gallowstemplar_f_6s2_0_1.vlm"},
	{0xF08FFAF08C7CC239ULL, "506000842.ogg"},
	{0xF090229DAF89CEC6ULL, "53881249.ogg"},
	{0xF090329A21647C12ULL, "vxlt_hangedman_1yo_0_0.vlm"},
	{0xF090BF122E2A2121ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_108a_0t.dds"},
	{0xF090DA8CB5A78B11ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavebase.mmh"},
	{0xF090E68CB5A79FB2ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavebase.mao"},
	{0xF090F48CB5A7B7CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavebase.msh"},
	{0xF09133C46AA46B8CULL, "9136054.ogg"},
	{0xF09190BB3A65608AULL, "718645137.ogg"},
	{0xF0927C002BC674CEULL, "qcr220_orsino6041785.cl"},
	{0xF0937BE6CC121FA7ULL, "6121565_f.fxe"},
	{0xF093DDD22CE4D9C3ULL, "6065936_m.fxe"},
	{0xF0941D5910A2BDE5ULL, "1040391814.ogg"},
	{0xF0943D66CE71108EULL, "genip_pp_gli_standing.utp"},
	{0xF094A6AEB6EC1D4BULL, "6103782_m.fxe"},
	{0xF0956B14DD74B1D0ULL, "vxht_0_hightown_bgj_0_0.vlm"},
	{0xF0959DB5802FCD3BULL, "vxlt_warehouse_n_sp_0_0.vlm"},
	{0xF09640A298287971ULL, "135369167.ogg"},
	{0xF096431D762A2246ULL, "118955214.ogg"},
	{0xF096CE794285BCD5ULL, "qun221_viscount6092421.cl"},
	{0xF096D0828A7103E0ULL, "xx_extpatternfloor.dds"},
	{0xF096D5AFF6FF968AULL, "static_static_p94.psh.fxo.dx11"},
	{0xF0974000867C7B14ULL, "vxw_bonepit_zj_0_0.vlm"},
	{0xF09869CC3D3AFEB0ULL, "466826691.ogg"},
	{0xF098C73086C11BB3ULL, "1033115548.ogg"},
	{0xF098FE42E297CDFEULL, "ht_keep_4519.anb"},
	{0xF099344E5A35299AULL, "6206860_m.fxe"},
	{0xF09935D442342C1DULL, "678558358.ogg"},
	{0xF09A0B916556752BULL, "594065413.ogg"},
	{0xF09A1535F95ACF53ULL, "fen210_tevinter_ambusher6108301.cl"},
	{0xF09A96D07C24C547ULL, "hlac_lt_0_lowtown_docks_n_-3_2z.dds"},
	{0xF09AB1BF670954BAULL, "dae000_unhappy_mage6165335.cl"},
	{0xF09AD96F37A9418BULL, "6154548_m.fxe"},
	{0xF09BA3C91A7040CCULL, "hlac_w_bonepitdestroyed_a3_393218_0x.dds"},
	{0xF09CBA121462E0A1ULL, "6136018_m.fxe"},
	{0xF09CDE7E979AE36DULL, "462128628.ogg"},
	{0xF09DAA2992D7308FULL, "zz_fen_debug6134379.cl"},
	{0xF09DACEBC398F97EULL, "hlac_w_woundedcoast_65540_0y.dds"},
	{0xF09E2A0592E6A7DFULL, "qcr221ip_throne_door.ncs"},
	{0xF09F03CAF742959EULL, "622211552.ogg"},
	{0xF09F4A3668CFF76BULL, "crt_strider.wbd"},
	{0xF09F527FB593114BULL, "vxlt_0_lowtown_a2_n_c_exf_0_0.vlm"},
	{0xF09F83E08403F1BCULL, "139269091.ogg"},
	{0xF0A006517E21441BULL, "6000717_m.fxe"},
	{0xF0A030C8E73D6C6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\enter.evt"},
	{0xF0A0A64864E6AC2AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_beta_n.dds"},
	{0xF0A0B4579758C028ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hugekeg.mmh"},
	{0xF0A0C0579758D4CBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hugekeg.mao"},
	{0xF0A0D2579758F352ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hugekeg.msh"},
	{0xF0A122220794CC6CULL, "6085919_m.fxe"},
	{0xF0A18DD4824B81C2ULL, "6154219_m.fxe"},
	{0xF0A23C734562FC99ULL, "184711568.ogg"},
	{0xF0A26C3D5905E307ULL, "916263529.ogg"},
	{0xF0A2E0AFF411427BULL, "hlac_lt_0_lowtown_a2_n_1_11z.dds"},
	{0xF0A381F74474A985ULL, "zz_lgt_3_debug6140433.cl"},
	{0xF0A3E6477BC696D0ULL, "vxca_caverns2_8wo_0_0.vlm"},
	{0xF0A3F20D4659B829ULL, "art\\vfx\\combat\\vfx_d_invulner_exit_c.anb"},
	{0xF0A40F927225E5BBULL, "vxlt_warehouse_n_vs_0_0.vlm"},
	{0xF0A4FC97365045EEULL, "character_character_p199.psh.fxo.dx11"},
	{0xF0A4FD222F7956BEULL, "818096284.ogg"},
	{0xF0A52E2EC4CBD675ULL, "6122020_m.fxe"},
	{0xF0A57DDE6DA5FDB7ULL, "1026973595.ogg"},
	{0xF0A5AD1AC0DFD3DFULL, "539929981.ogg"},
	{0xF0A5F34EFF89E31FULL, "6099485_m.fxe"},
	{0xF0A6234F98EEDBEAULL, "vxw_sundermount_341_0_1.vlm"},
	{0xF0A6294843B601C1ULL, "gen00fl_isabela_init6175715.cl"},
	{0xF0A62EAA2BF3734CULL, "925670858.ogg"},
	{0xF0A63AAA3BC45AA3ULL, "gen00fl_isabela6096767.cl"},
	{0xF0A649C8E742022CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\enter.gad"},
	{0xF0A732732C0559D6ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_avea_s.dds"},
	{0xF0A774DD86723411ULL, "vxca_caverns1_mines_i39_0_0.vlm"},
	{0xF0A77CF2CD7BA180ULL, "591489156.ogg"},
	{0xF0A7D09C2B4BBECCULL, "hlac_lt_darktown_-1_24z.dds"},
	{0xF0A7F4CBDEE13ED8ULL, "ave121_brennan6105894.cl"},
	{0xF0A843591B381365ULL, "953480875.ogg"},
	{0xF0A8E04232E7664CULL, "dae000_mother6001874.cl"},
	{0xF0A96FC93AC2F87AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelierb.msh"},
	{0xF0A97DC93AC30FB3ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelierb.mao"},
	{0xF0A981C93AC316C0ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_chanadelierb.mmh"},
	{0xF0A9D01401CADE4FULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.cess.gad"},
	{0xF0A9EF1594ACCCFCULL, "gen00fl_anders6125282.cl"},
	{0xF0AA36B2E82E62D5ULL, "vxlt_hangedman_27i_0_0.vlm"},
	{0xF0AA3B4577935FC8ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4_tool.mmh"},
	{0xF0AA47457793746BULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4_tool.mao"},
	{0xF0AA522F67BBBC5EULL, "dae300_bran6143470.cl"},
	{0xF0AA5945779392F2ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_flat2x4_tool.msh"},
	{0xF0AAC1B50087B263ULL, "zz_ave_debug6059903.cl"},
	{0xF0AAFE9ECAE4E6C4ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_filecabinet_d.dds"},
	{0xF0AB200B8A5EFB8DULL, "95233838.ogg"},
	{0xF0AB7747B3F5EA88ULL, "vxdr_0_deeproads_cuh_0_2.vlm"},
	{0xF0AB7936FA60DA35ULL, "6139441_f.fxe"},
	{0xF0ABAD9D3647F114ULL, "6144591_m.fxe"},
	{0xF0AC0BB67022455AULL, "vxht_keep_670_0_0.vlm"},
	{0xF0AC1BD87AABDB0EULL, "6129417_m.fxe"},
	{0xF0AD48C8E7481F2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\enter.ani"},
	{0xF0AE0A4670A088F9ULL, "llac_lt_hangedman_n_-1_29.dds"},
	{0xF0AE21FB914201AEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagchantrya_n.dds"},
	{0xF0AEE1BB5EF1A86CULL, "hlac_ca_caverns1_24421_10y.dds"},
	{0xF0AFA9DF063A0C93ULL, "vxlt_gallowscourtyard_4xk_0_0.vlm"},
	{0xF0B0B13AC580924BULL, "vxlt_0_lowtown_a2_n_c_h21_0_0.vlm"},
	{0xF0B0E51401D120FCULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.cess.evt"},
	{0xF0B21398B6301075ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala0a.phy"},
	{0xF0B218A7E747E9DFULL, "vxdr_0_deeproads_ebv_0_0.vlm"},
	{0xF0B230C63BD00D44ULL, "lt_0_lowtown_docks_a3_243_0.mmh"},
	{0xF0B234C63BD01397ULL, "lt_0_lowtown_docks_a3_243_0.mao"},
	{0xF0B23C887BE2363EULL, "vxlt_hangedman_a_0_0.vlm"},
	{0xF0B246C63BD032E6ULL, "lt_0_lowtown_docks_a3_243_0.msh"},
	{0xF0B36F64E79AF301ULL, "6170755_m.fxe"},
	{0xF0B420E6C1845BE0ULL, "vxht_0_hightown_a2_a71_0_0.vlm"},
	{0xF0B45A7CBE3F3A1AULL, "llac_ht_mansion2_818_11.dds"},
	{0xF0B460E11ECA196BULL, "6081707_m.fxe"},
	{0xF0B49CE03F4805FDULL, "vxca_caverns1_mines_j24_0_1.vlm"},
	{0xF0B4DB0E3EA4C716ULL, "llac_lt_0_lowtown_docks_a3_-131076_0.dds"},
	{0xF0B5358880B3DB5DULL, "vxlt_0_lowtown_eo1_0_1.vlm"},
	{0xF0B537FAC6B2B074ULL, "llac_lt_0_lowtown_a3_n_c_-65535_4.dds"},
	{0xF0B58F89D8120483ULL, "hlac_ht_chantry_n_2534_30z.dds"},
	{0xF0B5A398C53EEBC6ULL, "ef_nat_were_dalish_grd.mor"},
	{0xF0B6370AA01B1915ULL, "6046472_m.fxe"},
	{0xF0B65B1F0D9FCCDAULL, "363665810.ogg"},
	{0xF0B6A0C35C538CE2ULL, "art\\vfx\\combat\\vfx_glb_ele_n_deathblow_c.impa.gad"},
	{0xF0B6BFFC1FE64711ULL, "hlac_ht_mansion2_818_42y.dds"},
	{0xF0B72F23C54F553BULL, "vxlt_darktown_4qj_0_0.vlm"},
	{0xF0B7399C0582D671ULL, "pro000_guard6013440.cl"},
	{0xF0B743875FC783EEULL, "vxlt_foundry_n_zx_0_0.vlm"},
	{0xF0B8551EC5EA3886ULL, "dae000_arishok6075211.cl"},
	{0xF0B87CFD15178F94ULL, "6103580_f.fxe"},
	{0xF0B88C1F5D3E6BDAULL, "208197308.ogg"},
	{0xF0B939BBA4A6A5DDULL, "6033414_m.fxe"},
	{0xF0B96C91F47CFF16ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.cess.ani"},
	{0xF0B97C8A713640AAULL, "6085235_f.fxe"},
	{0xF0B9D56FEF171A95ULL, "arc_mit_05.mao"},
	{0xF0BA1F7A283E0994ULL, "6141691_m.fxe"},
	{0xF0BA295550BF7BC0ULL, "vxw_bonepitmining_a2_ey_0_0.vlm"},
	{0xF0BA5BFBEC2BA483ULL, "nat380_temmerin6120936.cl"},
	{0xF0BA65973E598FB9ULL, "vxlt_0_lowtown_docks_a3_n_d1f_0_0.vlm"},
	{0xF0BA72372390CDB1ULL, "1053981008.ogg"},
	{0xF0BA7D9362185DDAULL, "vxht_keep_601_0_0.vlm"},
	{0xF0BAA1D9DCD4E580ULL, "vrc221_varric6043933.cl"},
	{0xF0BADD640CBCC386ULL, "510728007.ogg"},
	{0xF0BAF19D8B781983ULL, "6133002_m.fxe"},
	{0xF0BAFD50A0DE2A50ULL, "vxlt_gallowstemplar_5bu_0_0.vlm"},
	{0xF0BB7721A7CE4D46ULL, "6101229_f.fxe"},
	{0xF0BB9B664AB9D2E4ULL, "vxlt_gamlenshouse_n_ac_0_0.vlm"},
	{0xF0BC3639997F424BULL, "vxlt_0_lowtown_a2_n_c_erx_0_0.vlm"},
	{0xF0BC37FB12336808ULL, "6039017_m.fxe"},
	{0xF0BCC3C35C5833D9ULL, "art\\vfx\\combat\\vfx_glb_ele_n_deathblow_c.impa.evt"},
	{0xF0BCFE8925FD54BEULL, "109628976.ogg"},
	{0xF0BD5B9284A23D74ULL, "qun160_vashothlead6053924.cl"},
	{0xF0BD94F51892B22DULL, "qun210cr_guard6087763.cl"},
	{0xF0BDECAC11A024AAULL, "vxht_0_hightown_a2_ds_0_0.vlm"},
	{0xF0BDED1401DBB815ULL, "art\\vfx\\combat\\vfx_w_off_accuracy_loop_c.cess.ani"},
	{0xF0BE172929510A11ULL, "vxca_cavern1_haunted_djf_0_1.vlm"},
	{0xF0BE9F9083BD0505ULL, "hlac_lt_backalley_g_1_4z.dds"},
	{0xF0BEE088513C234DULL, "hlac_lt_gallowstemplar_2388_1y.dds"},
	{0xF0BEF1A7CEA3217EULL, "547760340.ogg"},
	{0xF0BF2096C2B2CE89ULL, "6163777_m.fxe"},
	{0xF0BFB5434B63F4CEULL, "257760986.ogg"},
	{0xF0BFF85B1B72D78DULL, "zz_followers_banter6158640.cl"},
	{0xF0C04CE5EE2FDEFDULL, "6076440_m.fxe"},
	{0xF0C0BF4DFB2588B8ULL, "820045452.ogg"},
	{0xF0C11D35A8402461ULL, "vxlt_darktown_2w8_0_0.vlm"},
	{0xF0C1E5C072D68246ULL, "6087380_m.fxe"},
	{0xF0C1EDCE1B241F74ULL, "art\\characters\\weapons\\dagger\\w_dger_110a.phy"},
	{0xF0C1F9326BD48742ULL, "950705323.ogg"},
	{0xF0C28A9ED6AB8494ULL, "188630633.ogg"},
	{0xF0C295975C5EC494ULL, "vxw_bonepit_ki_0_0.vlm"},
	{0xF0C2C046726E03D5ULL, "vxca_caverns1_mines_fn5_0_0.vlm"},
	{0xF0C2D2BE10C031EBULL, "6084407_m.fxe"},
	{0xF0C2FD18A0972391ULL, "vxlt_0_lowtown_docks_n_bge_0_0.vlm"},
	{0xF0C340513B0C1DAEULL, "w_bonepitdestroyed_a3_138_0.msh"},
	{0xF0C34E513B0C35FFULL, "w_bonepitdestroyed_a3_138_0.mao"},
	{0xF0C35A513B0C499CULL, "w_bonepitdestroyed_a3_138_0.mmh"},
	{0xF0C43BB44E9575BEULL, "hairalpha_rendertoshadowmap_p0.vsh.fxo"},
	{0xF0C46FBE5EF07234ULL, "vxlt_gallowscourtyard_6wk_0_0.vlm"},
	{0xF0C4776EAFAA8E37ULL, "vxlt_gallowscourtyard_70t_0_0.vlm"},
	{0xF0C4A8E792113195ULL, "6041572_m.fxe"},
	{0xF0C52C3668F09339ULL, "crt_strider.bnk"},
	{0xF0C533F52F543ABDULL, "543949502.ogg"},
	{0xF0C5B56451BD9C6EULL, "vxlt_foundry_n_uy_0_0.vlm"},
	{0xF0C60CEFF33DFF68ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgertes.msh"},
	{0xF0C61EEFF33E1DF9ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgertes.mao"},
	{0xF0C62AEFF33E3192ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgertes.mmh"},
	{0xF0C69C414DDDDFBEULL, "vxht_0_hightown_a2_aow_0_0.vlm"},
	{0xF0C6A79ED4675723ULL, "6218022_m.fxe"},
	{0xF0C6ADEEC5C95A75ULL, "hlac_ht_0_hightown_a2_-1_0z.dds"},
	{0xF0C774390B800777ULL, "vxlt_warehouse_n_sd_0_0.vlm"},
	{0xF0C77491F4894A07ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.cess.evt"},
	{0xF0C7CD330D7D89D6ULL, "vxht_bartrand_n_4h7_0_0.vlm"},
	{0xF0C7EA45A32E4A7AULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_noba_0t.dds"},
	{0xF0C85BFA1E2067D6ULL, "hlac_lt_0_lowtown_docks_n_-4_3y.dds"},
	{0xF0C90D6DFDD74FEEULL, "vxlt_gallowstemplar_f_5qq_0_0.vlm"},
	{0xF0C916A6E8CE72DDULL, "art\\vfx\\combat\\vfx_gib_hlm_masc_1.mao"},
	{0xF0C96F2D92D442D0ULL, "vxlt_gallowstemplar_5or_0_0.vlm"},
	{0xF0C9970F0DA5EEAEULL, "vxdr_primevalentrance_32s_0_0.vlm"},
	{0xF0C9A1BC85486199ULL, "ch2alpharmtnt.psh.fxo"},
	{0xF0C9D1D5BAC6AB94ULL, "vxlt_0_lowtown_n_f5g_0_0.vlm"},
	{0xF0C9F7E90A210AB0ULL, "vxht_0_hightown_a3_c1x_0_0.vlm"},
	{0xF0CA46275C3E7DE3ULL, "1044451721.ogg"},
	{0xF0CA58C7DC812DF6ULL, "lt_0_lowtown_docks_a3_n_mmap_ph.dds"},
	{0xF0CACBC35C647EC8ULL, "art\\vfx\\combat\\vfx_glb_ele_n_deathblow_c.impa.ani"},
	{0xF0CAFDF559DCC932ULL, "gen_im_cut_invis_staff.uti"},
	{0xF0CB283AC3D6F2EAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose5left.anb"},
	{0xF0CB5BE658B14619ULL, "6105759_m.fxe"},
	{0xF0CB708D8BC3DA62ULL, "vxlt_warehouse_n_t4_0_0.vlm"},
	{0xF0CBCACE1B2C3871ULL, "art\\characters\\weapons\\dagger\\w_dger_110a.msh"},
	{0xF0CBE4CE1B2C642FULL, "art\\characters\\weapons\\dagger\\w_dger_110a.mmh"},
	{0xF0CC222A9A5BB4E4ULL, "vxdr_primevaltomb_2hr_0_0.vlm"},
	{0xF0CC2BFCBDDBE1D6ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_chanterboard.dura.ani"},
	{0xF0CCA7EAB658290EULL, "qun221_petrice6083432.cl"},
	{0xF0CCBD569E3777A7ULL, "79980356.ogg"},
	{0xF0CD1A39642619D9ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01b.mmh"},
	{0xF0CD243964262AC7ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01b.msh"},
	{0xF0CDEC86A36E924DULL, "441619985.ogg"},
	{0xF0CE66899992124CULL, "vxht_bartrand_n_2b5_0_0.vlm"},
	{0xF0CE725FF0C83882ULL, "vxdr_primevaltomb_7eh_0_2.vlm"},
	{0xF0CE8D2E9A348CCAULL, "vxdr_0_deeproads_fim_0_0.vlm"},
	{0xF0CE9591F48FA150ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_p.cess.gad"},
	{0xF0CED856E63A5822ULL, "291003414.ogg"},
	{0xF0CF28AF1031888AULL, "hlac_w_encounter2_camp_327678_0y.dds"},
	{0xF0CF3288A54D13D0ULL, "vxht_0_hightown_bsv_0_0.vlm"},
	{0xF0D00432215E73A3ULL, "6097326_m.fxe"},
	{0xF0D0A7EE29585BA9ULL, "hlac_lt_0_lowtown_docks_a3_n_65532_1y.dds"},
	{0xF0D12DB6CA73CE12ULL, "335733598.ogg"},
	{0xF0D14D5879830955ULL, "43857167.ogg"},
	{0xF0D1F0E67C64F8D9ULL, "art\\levels\\ztools\\helpers\\character\\prp_humansit_1.mao"},
	{0xF0D2514E07FF4B0CULL, "dae000cr_grand_cleric6147704.cl"},
	{0xF0D2E19B50E68AB4ULL, "vxlt_gallowscourtyard_6rh_0_0.vlm"},
	{0xF0D3199C88A9F49CULL, "vxht_bartrand_n_4pe_0_0.vlm"},
	{0xF0D346210BA96AF6ULL, "vxht_mansion2_1v0_0_0.vlm"},
	{0xF0D3C7A8C7E6EC50ULL, "art\\characters\\creatures\\highdragon\\animation\\c_highdragon.fly.evt"},
	{0xF0D3F03806984975ULL, "llac_lt_0_lowtown_a2_n_-65537_6.dds"},
	{0xF0D3F45B8D1E7D34ULL, "fen210_dying_magister6108589.cl"},
	{0xF0D4135C6FF860F1ULL, "6038679_m.fxe"},
	{0xF0D470A6C56FB1ADULL, "l_thi121ar_mansion.bnk"},
	{0xF0D4E91D0F49CEF3ULL, "329393828.ogg"},
	{0xF0D5305886861C52ULL, "283383447.ogg"},
	{0xF0D58ACEE8636C24ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medf.msh"},
	{0xF0D5A0CEE86391C6ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medf.mmh"},
	{0xF0D5B704B595DA46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_2.ani"},
	{0xF0D5DEAB5172ED57ULL, "6200881_m.fxe"},
	{0xF0D611103B0A23C6ULL, "vxht_chantry_n_2ck_0_0.vlm"},
	{0xF0D633F7AC2173C9ULL, "vxlt_warehouse_e9_0_0.vlm"},
	{0xF0D67F2D596AA647ULL, "vxw_encounter2_9x_0_0.vlm"},
	{0xF0D6AEFEC67F5DDFULL, "hlac_w_bonepit_0_3z.dds"},
	{0xF0D704BFFEE0BB4FULL, "777274317.ogg"},
	{0xF0D70D0F34ABACC0ULL, "vxw_bonepitmining_a2_og_0_0.vlm"},
	{0xF0D71BD6633DDCA7ULL, "6153589_m.fxe"},
	{0xF0D76225781B498EULL, "vxlt_0_lowtown_docks_a3_n_cyz_0_0.vlm"},
	{0xF0D792A9257E9403ULL, "vxht_mansion2b_2pt_0_0.vlm"},
	{0xF0D7A65BEEB017A6ULL, "vxca_caverns2prc_8h0_0_1.vlm"},
	{0xF0D8072850EFBA04ULL, "gen_im_npc_wep_dua_genc.uti"},
	{0xF0D810873BD7578AULL, "485765928.ogg"},
	{0xF0D815974E890D38ULL, "6152277_m.fxe"},
	{0xF0D8961C986EA657ULL, "llac_ca_caverns2_5802_17.dds"},
	{0xF0D8ACC9CCE1D087ULL, "zz_dae_debug6029113.cl"},
	{0xF0D8F19C96E15524ULL, "vxlt_0_lowtown_f6p_0_1.vlm"},
	{0xF0D933FCBDE679C7ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_chanterboard.dura.evt"},
	{0xF0D9C978BEB50925ULL, "hlac_dr_primevaltomb_664_7z.dds"},
	{0xF0DA620F84AB0CD1ULL, "hlac_ht_0_hightown_a2_-131069_0y.dds"},
	{0xF0DA6E30882A096DULL, "104886688.ogg"},
	{0xF0DAC51A8187B9DDULL, "521411571.ogg"},
	{0xF0DB38E85E9AE088ULL, "vxw_woundedcoast_157_0_0.vlm"},
	{0xF0DB458C720F51CEULL, "6164281_m.fxe"},
	{0xF0DB613B8932A4A0ULL, "w_encounter2_12_0.phy"},
	{0xF0DC3BE199A7A484ULL, "hlac_lt_hangedman_n_-1_4z.dds"},
	{0xF0DC5CF446A5D37DULL, "mag131_keran6088308.cl"},
	{0xF0DCBDA062F10392ULL, "231370125.ogg"},
	{0xF0DD1118F10C7EA6ULL, "hlac_w_sundermount_-262145_0z.dds"},
	{0xF0DD163CCFBEF7C2ULL, "qun221_petrice6083818.cl"},
	{0xF0DD8F80309CB6CCULL, "711458071.ogg"},
	{0xF0DDBDA1DDA45A5AULL, "vxlt_gallowscourtyard_6ec_0_0.vlm"},
	{0xF0DDC079C25426EEULL, "vxca_cave_29w_0_1.vlm"},
	{0xF0DE5B9395B83F2DULL, "6125285_m.fxe"},
	{0xF0DEC278EC1BB230ULL, "dae200cr_viscount6087390.cl"},
	{0xF0DF96FFF4529E95ULL, "6033771_m.fxe"},
	{0xF0E054FCBDECD010ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_chanterboard.dura.gad"},
	{0xF0E056E8F69D7A3DULL, "hlac_lt_0_lowtown_docks_n_-2_0y.dds"},
	{0xF0E05BAB532675E3ULL, "257798236.ogg"},
	{0xF0E097CEE86D8991ULL, "art\\characters\\playercharacter\\humanfemale\\hlh_medf.phy"},
	{0xF0E1537842DCA334ULL, "vxlt_gallowscourtyard_6yi_0_0.vlm"},
	{0xF0E15781BF8C662EULL, "ser191_emeric6104745.cl"},
	{0xF0E1B7FDFD9F8376ULL, "vxht_mansion2_1y1_0_0.vlm"},
	{0xF0E259D79861EB64ULL, "vxw_bonepitmining_a2_1x1_0_0.vlm"},
	{0xF0E26F8663DF8BF6ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockdxxl.msh"},
	{0xF0E2798663DF9CF4ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockdxxl.mmh"},
	{0xF0E27D8663DFA307ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_kcrockdxxl.mao"},
	{0xF0E2803B8938F858ULL, "w_encounter2_12_0.rml"},
	{0xF0E298562701E166ULL, "6104306_m.fxe"},
	{0xF0E2F566516D5902ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_csand01_n.dds"},
	{0xF0E3039468027421ULL, "hlac_lt_undercitydungeon_-1_17x.dds"},
	{0xF0E39E6C89991010ULL, "zz_isa_debug6133966.cl"},
	{0xF0E39F04B5A1EEB7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_2.evt"},
	{0xF0E41D580EF5C4F0ULL, "vxw_bonepitmining_a2_n8_0_0.vlm"},
	{0xF0E579D7CF3F2962ULL, "vxw_sundermount_305_0_1.vlm"},
	{0xF0E57EEC26A1C540ULL, "vxw_bonepitmining_a2_jd_0_0.vlm"},
	{0xF0E584BAE1BD6253ULL, "llac_lt_gallowstemplar_2388_0.dds"},
	{0xF0E5DEC5E6D1FC00ULL, "cod_hst_thedas_geogrphy.utp"},
	{0xF0E652E776C073D0ULL, "vxlt_gallowstemplar_5up_0_0.vlm"},
	{0xF0E6E148F9A09069ULL, "549708215.ogg"},
	{0xF0E74E86BA32085FULL, "dae000_bodhan6129125.cl"},
	{0xF0E76E0DF26030FCULL, "docks_male_human_worker_sta6176035.cl"},
	{0xF0E8EA2A977E6B9FULL, "lgt110_gxa_vgk_silver_order6117026.cl"},
	{0xF0E939513B2CE167ULL, "w_bonepitdestroyed_a3_138_0.phy"},
	{0xF0E97A89CE82D461ULL, "6080571_m.fxe"},
	{0xF0E9AAC55090F908ULL, "vxw_woundedcoast_104_0_0.vlm"},
	{0xF0E9D63FFEA41657ULL, "vxht_0_hightown_a3_aie_0_0.vlm"},
	{0xF0E9EEC49CEE7196ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_i.evt"},
	{0xF0EA765BE9FB2FCDULL, "destroy.cur"},
	{0xF0EAA004B5A80F40ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\escort_away_action_2.gad"},
	{0xF0EAC648E25836DFULL, "6104538_m.fxe"},
	{0xF0EB2B421C8C0CEFULL, "gen00fl_aveline6114640.cl"},
	{0xF0EBC08CEB7B2303ULL, "w_bonepitmining_a2_146_0.msh"},
	{0xF0EBC604091C187EULL, "gen00fl_varric6076216.cl"},
	{0xF0EBCE8CEB7B3B5DULL, "w_bonepitmining_a2_146_0.mmh"},
	{0xF0EBD28CEB7B426EULL, "w_bonepitmining_a2_146_0.mao"},
	{0xF0EC1642893944D0ULL, "vxht_0_hightown_byt_0_0.vlm"},
	{0xF0EC9BC08D4AE42AULL, "vxlt_0_lowtown_ev8_0_0.vlm"},
	{0xF0ED58E39D3BF11AULL, "6060718_m.fxe"},
	{0xF0ED6975169A3A2AULL, "6152183_m.fxe"},
	{0xF0EDCD1EEF699E93ULL, "6105814_m.fxe"},
	{0xF0EE3A4E758974D4ULL, "vxlt_warehouse_va_0_0.vlm"},
	{0xF0EE4E321406177CULL, "259920728.ogg"},
	{0xF0EE832D35F9D406ULL, "6076263_m.fxe"},
	{0xF0EEAAC8003B0466ULL, "dre000_bethany6037342.cl"},
	{0xF0EEE710AB6AFC94ULL, "hlac_w_woundedcoast_0_1x.dds"},
	{0xF0EEF633EBE11400ULL, "6178666.ogg"},
	{0xF0EF402A0F484C80ULL, "6133629_m.fxe"},
	{0xF0EF61CA4AA3E8A5ULL, "vxca_caverns1_mines_huv_0_0.vlm"},
	{0xF0EF9097ABEAE8A5ULL, "hlac_dr_dungeon_3214_13y.dds"},
	{0xF0F0344316C862A7ULL, "hlac_lt_0_lowtown_a2_-65536_2z.dds"},
	{0xF0F0CBB48A5803E4ULL, "vxw_bonepitmining_a2_1u6_0_0.vlm"},
	{0xF0F0E5C49CF48121ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_i.gad"},
	{0xF0F1047B1D0CCB67ULL, "hlac_lt_0_lowtown_docks_n_-65536_0x.dds"},
	{0xF0F15299FB3DA98CULL, "6111463_m.fxe"},
	{0xF0F19FCA5A369C5CULL, "98199790.ogg"},
	{0xF0F24FD681C0C7ECULL, "hlac_ca_caverns1_24426_1x.dds"},
	{0xF0F25CFD2EBF9AA9ULL, "gen_im_fla_aveline_03.uti"},
	{0xF0F277AD9E656D6FULL, "dae300_alistair_king6133827.cl"},
	{0xF0F2CEC5426FD452ULL, "ht_0_hightown_a2_60_0.rml"},
	{0xF0F2DD31458F84BFULL, "hlac_ht_chantry_2534_32y.dds"},
	{0xF0F36CD9B71F79A8ULL, "hlac_lt_0_lowtown_a3_n_c_65534_4z.dds"},
	{0xF0F3704E3DA9CA5EULL, "55766227.ogg"},
	{0xF0F437731BB12E71ULL, "hlac_ca_cavern1_haunted_24425_6x.dds"},
	{0xF0F46304D3B9996EULL, "vxlt_gallowstemplar_f_5bt_0_0.vlm"},
	{0xF0F51B0AE6A571CEULL, "gen00fl_dog6167690.cl"},
	{0xF0F5AA392D9E9E25ULL, "364115342.ogg"},
	{0xF0F5F7E598B3995AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile0a_1.mao"},
	{0xF0F604EB645CD3A4ULL, "443092400.ogg"},
	{0xF0F6AFABDA49F11BULL, "llac_ht_chantry_2534_39.dds"},
	{0xF0F729FC44D9372BULL, "358155594.ogg"},
	{0xF0F75733FEFDABD3ULL, "vxlt_gallowscourtyard_n_7gr_0_0.vlm"},
	{0xF0F77BBE5E6679D5ULL, "follower_banter6149172.cl"},
	{0xF0F7E6C49CFAA1CFULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_i.ani"},
	{0xF0F8581095E85069ULL, "follower_banter6143310.cl"},
	{0xF0F8664C8DE383BFULL, "dae030cr_male_fereldan.utc"},
	{0xF0F89792B01FAEFDULL, "455838624.ogg"},
	{0xF0F89E167EC01AA7ULL, "llac_w_sundermount_131069_0.dds"},
	{0xF0F8B9AAFA0261EFULL, "vxht_keep_n_6j5_0_0.vlm"},
	{0xF0F92698B66C3290ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala0a.msh"},
	{0xF0F93498B66C4AAAULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala0a.mmh"},
	{0xF0F93898B66C51E1ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala0a.mao"},
	{0xF0F9ACECD568BD4CULL, "m_mcr370ar_gallows_nt.bnk"},
	{0xF0F9C7F6C6AA8202ULL, "vxdr_primevaltomb_7tw_0_2.vlm"},
	{0xF0F9DE4C0CC89D10ULL, "llac_w_encounter1_-131067_0.dds"},
	{0xF0F9E1AD5E69AF60ULL, "vxlt_undercityencounter_v0_0_0.vlm"},
	{0xF0F9FB22922098E7ULL, "475159297.ogg"},
	{0xF0FB0D9D7F40FCAAULL, "vxlt_0_lowtown_eq9_0_0.vlm"},
	{0xF0FB56F9FD85EE17ULL, "430880925.ogg"},
	{0xF0FBF160397E559EULL, "vxlt_gallowstemplar_n_4xw_0_0.vlm"},
	{0xF0FBFFD5BA8E950AULL, "6116148_f.fxe"},
	{0xF0FC2377E19AA666ULL, "6176134_m.fxe"},
	{0xF0FC27C8683EE87DULL, "vxdr_primevaltomb_3iy_0_0.vlm"},
	{0xF0FC5FC4F5A97D84ULL, "587944539.ogg"},
	{0xF0FCEDD5958C1D68ULL, "47350896.ogg"},
	{0xF0FD24E4AD8F5428ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoireao.msh"},
	{0xF0FD36E4AD8F72B9ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoireao.mao"},
	{0xF0FD42E4AD8F8752ULL, "art\\levels\\races\\mage\\prop\\artifacts\\prp_grimoireao.mmh"},
	{0xF0FE22102A6E2B35ULL, "6076616_f.fxe"},
	{0xF0FE6A862FA6B7A0ULL, "art\\vfx\\materials\\vfx_smoke_whispy.mao"},
	{0xF0FE74E45E4D4982ULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_rock_002_sa.dds"},
	{0xF0FE9BB7E18BB476ULL, "vxht_mansion2_1g3_0_0.vlm"},
	{0xF0FEE1F640FB0A36ULL, "lt_gallowscourtyard_34_0.rml"},
	{0xF0FF68747F1DCFBBULL, "vxlt_darktown_4ha_0_0.vlm"},
	{0xF0FF6FB1132A63ECULL, "llac_ht_chantry_2534_10.dds"},
	{0xF100533187384AF9ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb2.mmh"},
	{0xF1005D3187385BE7ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_fereldenb2.msh"},
	{0xF10066A6C594DBD7ULL, "l_thi121ar_mansion.wbd"},
	{0xF1007E3B8951F235ULL, "w_encounter2_12_0.msh"},
	{0xF100803B8951F5ABULL, "w_encounter2_12_0.mmh"},
	{0xF1008C3B89520A48ULL, "w_encounter2_12_0.mao"},
	{0xF100B5C3991E3EB1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_crouch_squat.anb"},
	{0xF100F4BE39D40FDDULL, "vxht_0_hightown_cax_0_0.vlm"},
	{0xF10172293F39C3DDULL, "gen00fl_varric6099456.cl"},
	{0xF10272EF20E1E8ABULL, "hightown_male_noble_gossip6174093.cl"},
	{0xF102735E324AF25BULL, "6177309_m.fxe"},
	{0xF103BF2A9BA1A212ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v3.gad"},
	{0xF10436E585C6C0B5ULL, "6105764_m.fxe"},
	{0xF104B5621717D292ULL, "vxca_caverns1_j5b_0_0.vlm"},
	{0xF104E3D3D55C1031ULL, "vrc221_bartrand6065298.cl"},
	{0xF1051513C334CA39ULL, "6066908_m.fxe"},
	{0xF1055D55222C3DA3ULL, "261407937.ogg"},
	{0xF106318CBC4033B5ULL, "461901819.ogg"},
	{0xF106E6705E3D17D3ULL, "6148927_m.fxe"},
	{0xF107AE3C5500DAA7ULL, "dae021tr_alema.ncs"},
	{0xF107E0144708099DULL, "324420374.ogg"},
	{0xF109148B2C817C3BULL, "hlac_lt_backalley_g_0_6y.dds"},
	{0xF1092BC5AD5E6250ULL, "6066343_m.fxe"},
	{0xF109A2E1276B5B7BULL, "380297064.ogg"},
	{0xF10A487B8C2EB802ULL, "710463618.ogg"},
	{0xF10AC22A9BA7C649ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v3.evt"},
	{0xF10AC927C5F98DDDULL, "dae211ar_gamlens_house.ncs"},
	{0xF10AED8299BC22D8ULL, "vxlt_gallowscourtyard_n_6pa_0_0.vlm"},
	{0xF10B138F65865E99ULL, "dae000_arishok6082559.cl"},
	{0xF10B18968CFA7039ULL, "hlac_w_encounter2_camp_5_0z.dds"},
	{0xF10B9DC0B698367AULL, "qun210cr_random_thug6090297.cl"},
	{0xF10BD3548EC4A531ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.impa.ani"},
	{0xF10D2D9A0C036455ULL, "484784930.ogg"},
	{0xF10D447157A1607FULL, "framebuffereffects_vsmblur_p0.psh.fxo"},
	{0xF10DDFCDAB7D9462ULL, "549827038.ogg"},
	{0xF10DE907504ECD3DULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medd_d.dds"},
	{0xF10EC216B85EFBB2ULL, "6072090_m.fxe"},
	{0xF10F427A29050F04ULL, "6107294_f.fxe"},
	{0xF10F59957FCDA8B2ULL, "140663253.ogg"},
	{0xF10F669B2BCA285DULL, "vxht_0_hightown_cdw_0_0.vlm"},
	{0xF10FD87C471774AFULL, "6142122_f.fxe"},
	{0xF10FEE0FFA840085ULL, "257529733.ogg"},
	{0xF110048FE5837ECCULL, "and230ip_at_dungeon.utp"},
	{0xF1100F339DA889BEULL, "349371526.ogg"},
	{0xF11024929BA2B5A3ULL, "6160792_m.fxe"},
	{0xF110D3FA54A8D515ULL, "vxlt_gallowscourtyard_6l3_0_0.vlm"},
	{0xF110E475CABA30F6ULL, "117399659.ogg"},
	{0xF1111631FEDEF936ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04a.msh"},
	{0xF1112031FEDF0A34ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04a.mmh"},
	{0xF111304D0F9003B3ULL, "vxlt_gallowsprison_n_6c1_0_0.vlm"},
	{0xF111DB1533BB22F1ULL, "6059340_m.fxe"},
	{0xF1124BFEEC1933D5ULL, "694888084.ogg"},
	{0xF1126895D1502DF4ULL, "vxht_0_hightown_a2_bu8_0_0.vlm"},
	{0xF112AAC48C7AA8C6ULL, "llac_w_bonepitdestroyed_a3_0_11.dds"},
	{0xF112DE548ECAD6D3ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.impa.gad"},
	{0xF11307BE92F50B04ULL, "454629567.ogg"},
	{0xF1130C33E561075BULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.impa.ani"},
	{0xF113170E3B950188ULL, "vxca_caverns1_fmd_0_0.vlm"},
	{0xF1132ECE1C15FEF2ULL, "hlac_lt_undercityserialkiller_-1_20x.dds"},
	{0xF113BE3D724C1D3EULL, "gen_im_cns_cft_wrn_02.uti"},
	{0xF113D14B4D637CCCULL, "77958587.ogg"},
	{0xF113D9B41D3ACC31ULL, "6116113_m.fxe"},
	{0xF11418670ACD924DULL, "llac_lt_gallowscourtyard_n_-65537_1.dds"},
	{0xF114A32BFD96E249ULL, "hm_ave_donnic.mor"},
	{0xF115C753A8806682ULL, "219880103.ogg"},
	{0xF11616AD8797B3AAULL, "384748347.ogg"},
	{0xF116224E39C26EC0ULL, "dae211ar_tavern_nt.are"},
	{0xF1162B4284EA628EULL, "6171412_m.fxe"},
	{0xF11659369A13C432ULL, "dae000_fenarel6128272.cl"},
	{0xF11721A7B8C05EBEULL, "llac_ht_keep_n_3732_16.dds"},
	{0xF117750855ABAE3DULL, "6015740_m.fxe"},
	{0xF118368B6C71B6BFULL, "gen_im_trs_cot_blt_r2.uti"},
	{0xF118AA2A9BB3DAB8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_kdwn_v3.ani"},
	{0xF118DB548ECF3D38ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_proj_p.impa.evt"},
	{0xF118E967944386E4ULL, "914848419.ogg"},
	{0xF119194E39C49542ULL, "dae211ar_tavern_nt.ncs"},
	{0xF119209E4EE2902FULL, "fen110_fenris6118125.cl"},
	{0xF1198CB39FEA3515ULL, "plt_lgt000pt_ultra_3_03.plo"},
	{0xF11A22E648BA1A51ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_ex.ani"},
	{0xF11A42CDC633BE36ULL, "ico_mm_followerlocal.dds"},
	{0xF11A60274A75387CULL, "6133741_f.fxe"},
	{0xF11A893F649577DCULL, "190649322.ogg"},
	{0xF11ABBED033867F6ULL, "673514181.ogg"},
	{0xF11AF7751141B09EULL, "vxlt_gallowstemplar_f_46e_0_0.vlm"},
	{0xF11BA85EB69BDB54ULL, "19088264.ogg"},
	{0xF11BAE492983509FULL, "plt_lgt000pt_civ_wild.nss"},
	{0xF11C8ACF1E788B7CULL, "vxlt_undercityserialkiller_1if_0_0.vlm"},
	{0xF11CA23431FE701DULL, "vxlt_0_lowtown_n_gll_0_0.vlm"},
	{0xF11CB1CE69E83E67ULL, "vxlt_0_lowtown_docks_n_b4j_0_0.vlm"},
	{0xF11D8B9CDA5933A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bf_transition.gad"},
	{0xF11DACED8C0E348CULL, "vxlt_backalley_g_1i2_0_0.vlm"},
	{0xF11DDF3DCACAF648ULL, "vxw_bonepitdestroyed_a3_mo_0_0.vlm"},
	{0xF11E019ADF5C9639ULL, "vxht_keep_633_0_0.vlm"},
	{0xF11E51040C2D5A24ULL, "6041390_f.fxe"},
	{0xF11EC0F9523F3EF6ULL, "468931465.ogg"},
	{0xF11EF6A50DE5C976ULL, "hlac_lt_gallowstemplar_2383_7y.dds"},
	{0xF11FCE824CE3CE20ULL, "hlac_lt_0_lowtown_docks_n_196608_0y.dds"},
	{0xF1202DE648BE98F3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_ex.gad"},
	{0xF120314E39CADCD3ULL, "dae211ar_tavern_nt.lst"},
	{0xF12055248398094DULL, "6071095_f.fxe"},
	{0xF12096790D50C55FULL, "119404758.ogg"},
	{0xF120DF5B50B1F27DULL, "vxlt_0_lowtown_a2_n_ekj_0_1.vlm"},
	{0xF121054651CB79FBULL, "139709170.ogg"},
	{0xF121068005EA9D9BULL, "vxdr_primevaltomb_6z6_0_1.vlm"},
	{0xF1211433E56D5252ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.impa.evt"},
	{0xF121DBB8652266CDULL, "llac_lt_gallowstemplar_f_2391_5.dds"},
	{0xF12229EFF38C8E8DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_01hgertes.phy"},
	{0xF122407AC894E87EULL, "vxlt_hangedman_n_327_0_0.vlm"},
	{0xF12247AF3E7F5DBFULL, "161033722.ogg"},
	{0xF122726755CB6ED8ULL, "follower_banter6143128.cl"},
	{0xF1229BE125FBEB17ULL, "dae000_arishok6081307.cl"},
	{0xF1238841124E34D8ULL, "255379791.ogg"},
	{0xF124809CDA5F3FB4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bf_transition.evt"},
	{0xF1249BF1E17C00BFULL, "dae000_keep_noble_sta6161531.cl"},
	{0xF125039A2CB56E55ULL, "244889138.ogg"},
	{0xF1258BE23695C833ULL, "588224385.ogg"},
	{0xF12629857F32A3FAULL, "6103228_m.fxe"},
	{0xF1268848DB816BE4ULL, "vxht_playershouse_n_1pv_0_0.vlm"},
	{0xF126A95F8EC9EFDDULL, "419760116.ogg"},
	{0xF126BD04F4151830ULL, "842750459.ogg"},
	{0xF12703763E677ED4ULL, "455751287.ogg"},
	{0xF1272AE648C4B258ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_ability_impale_ex.evt"},
	{0xF12798F68316EC5DULL, "llac_lt_0_lowtown_a2_n_c_1_8.dds"},
	{0xF127B0B854508FBCULL, "371778658.ogg"},
	{0xF1283333E573A5E5ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_m_p.impa.gad"},
	{0xF128579280DE7867ULL, "285225753.ogg"},
	{0xF1299A19CF3EDDE8ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_eye_sw.mao"},
	{0xF12A16E51489F280ULL, "523640442.ogg"},
	{0xF12AE550DDF8DA72ULL, "art\\vfx\\creatures\\vfx_c_drag_bfft_p.impa.gad"},
	{0xF12AFCAC106EA3FCULL, "vxlt_undercityserialkiller_1ng_0_0.vlm"},
	{0xF12B2FA4D74362A1ULL, "vxlt_gallowscourtyard_6j9_0_0.vlm"},
	{0xF12B38CB790E7479ULL, "hlac_lt_hangedman_n_-1_25x.dds"},
	{0xF12C51FB4F57EEE1ULL, "406323452.ogg"},
	{0xF12CED8B8B4C8F96ULL, "6103912_m.fxe"},
	{0xF12D5DC3731896B1ULL, "6126852_f.fxe"},
	{0xF12D7D6A2AE295FDULL, "hlac_w_bonepitmining_a2_262144_0x.dds"},
	{0xF12D8171CF8B3610ULL, "501602758.ogg"},
	{0xF12D88E58D58CA0DULL, "94071686.ogg"},
	{0xF12DA0100FD5F332ULL, "vxca_caverns1_dnu_0_0.vlm"},
	{0xF12DB7B438950615ULL, "vxlt_gallowscourtyard_6v1_0_0.vlm"},
	{0xF12DD6079CF0C926ULL, "mag211_marethari6025323.cl"},
	{0xF12E0392DC2977D3ULL, "hlac_lt_gallowstemplar_n_2394_0z.dds"},
	{0xF12E146BC6F88A0FULL, "6121922_m.fxe"},
	{0xF12E6900DEEADB5BULL, "6134721_f.fxe"},
	{0xF12EA19E3925BC0EULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_crnf.ani"},
	{0xF12EC9062B459CB7ULL, "6033096_m.fxe"},
	{0xF12FA1492993E3B0ULL, "plt_lgt000pt_civ_wild.plo"},
	{0xF1301B39647A98C6ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_hed01b.phy"},
	{0xF13055BD870402A2ULL, "6156060_m.fxe"},
	{0xF130E3956CFD1EB1ULL, "6085338_m.fxe"},
	{0xF130EEBFA84F41F0ULL, "448819503.ogg"},
	{0xF1314168C8E27562ULL, "l_one351ar_zevran_cave.wbd"},
	{0xF131543DE1E38DD1ULL, "ht_mansion_n.arl"},
	{0xF131A62C934FBAF3ULL, "art\\characters\\creatures\\corpse\\textures\\cn_hed_cor_0s.dds"},
	{0xF131C029ED4BCED4ULL, "570963246.ogg"},
	{0xF131E850DDFEFDA9ULL, "art\\vfx\\creatures\\vfx_c_drag_bfft_p.impa.evt"},
	{0xF132789CDA6B6FEDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_bf_transition.ani"},
	{0xF132830F841A390DULL, "vxlt_0_lowtown_fhn_0_0.vlm"},
	{0xF13312D47743CCBCULL, "611818489.ogg"},
	{0xF1331AA10A9481DCULL, "vxht_0_hightown_a3_c5x_0_0.vlm"},
	{0xF13370A0A6368292ULL, "vxdr_primevalentrance_222_0_0.vlm"},
	{0xF133BBA8F6CC8BADULL, "6035135_m.fxe"},
	{0xF133F159314AEE10ULL, "323036955.ogg"},
	{0xF133F394F8EEE8BDULL, "dae011_gossips6143022.cl"},
	{0xF134A09B844BAAE8ULL, "6202206_m.fxe"},
	{0xF134E06F0681C8FFULL, "pro000_friedrich6050172.cl"},
	{0xF134F46C824D57C2ULL, "1045544148.ogg"},
	{0xF13552EA5B7C41ABULL, "hlac_lt_0_lowtown_a3_n_c_131070_4z.dds"},
	{0xF1359BEF343DA1E0ULL, "vxca_caverns1_cln_0_0.vlm"},
	{0xF135F23FFCC9C99BULL, "6042833_f.fxe"},
	{0xF1361260A23C615BULL, "6054998_m.fxe"},
	{0xF136140B36A80196ULL, "6039419_m.fxe"},
	{0xF13682BAE3F5C07AULL, "6119627_m.fxe"},
	{0xF13698BCB6DA8D06ULL, "6018528_m.fxe"},
	{0xF1377692EA44A12FULL, "mer240_jansen6050336.cl"},
	{0xF1377BF8534EBB5EULL, "dae011_corff_sta6163835.cl"},
	{0xF137978D0ABF714AULL, "vxlt_0_lowtown_a2_db5_0_0.vlm"},
	{0xF137A90A4BBE51AAULL, "vxca_cavern1_haunted_edw_0_0.vlm"},
	{0xF137CE922E5012C6ULL, "lla_ht_0_hightown_a2_0_0.dds"},
	{0xF138196E007DC75BULL, "222783508.ogg"},
	{0xF13821E56747818BULL, "gen00fl_sebastian6177814.cl"},
	{0xF1383BE3CB66AE06ULL, "605829589.ogg"},
	{0xF139056A841AD9A3ULL, "vxw_woundedcoast_1e8_0_0.vlm"},
	{0xF139A181C9397B21ULL, "vxlt_gallowscourtyard_6q8_0_0.vlm"},
	{0xF139DAFE54344F7AULL, "1062488656.ogg"},
	{0xF13AA03A15CFEE6AULL, "714724012.ogg"},
	{0xF13AB6ECD5A0657EULL, "m_mcr370ar_gallows_nt.wbd"},
	{0xF13AD93A48F30FF4ULL, "310429481.ogg"},
	{0xF13B270AF5D58289ULL, "dre000_sibling_has_blight6036501.cl"},
	{0xF13B8F7854F55BD6ULL, "sky_ht_mansion2b.mao"},
	{0xF13B999E3930395FULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_crnf.evt"},
	{0xF13BD39F1E1EB4A6ULL, "ef_cora.mor"},
	{0xF13CADF5CB72BB6BULL, "zz_hawke_cu_ve2.cut"},
	{0xF13CB50CA10641BDULL, "lt_0_lowtown_docks_a3_n_251_0.mao"},
	{0xF13CB90CA10648F6ULL, "lt_0_lowtown_docks_a3_n_251_0.mmh"},
	{0xF13CC30CA10659F4ULL, "lt_0_lowtown_docks_a3_n_251_0.msh"},
	{0xF13D142124EBBBCAULL, "hlac_ht_chantry_2534_23z.dds"},
	{0xF13D652538D882E3ULL, "256580847.ogg"},
	{0xF13DABF17CB32359ULL, "gen00fl_sebastian6174071.cl"},
	{0xF13DC315349E237DULL, "vxlt_0_lowtown_a2_n_eqh_0_1.vlm"},
	{0xF13DF4DE543795A1ULL, "vxht_mansion2_3cj_0_0.vlm"},
	{0xF13DFE52FD777845ULL, "6205642_m.fxe"},
	{0xF13E323CBE49EC9FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\gt2a_lowtowngate_1.pwk"},
	{0xF13EB050CCAC77CAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop_lifted.gad"},
	{0xF13EB28998E49C5BULL, "vxlt_darktown_589_0_1.vlm"},
	{0xF13FC45102D55968ULL, "vxlt_undercityserialkiller_1j6_0_1.vlm"},
	{0xF13FD050DE0B1218ULL, "art\\vfx\\creatures\\vfx_c_drag_bfft_p.impa.ani"},
	{0xF1404796B11AB004ULL, "hlac_lt_backalley_n_1_4x.dds"},
	{0xF140D069B2D1A1FFULL, "6180444_m.fxe"},
	{0xF140FA2FBA94F971ULL, "gen00fl_isabela_init6127653.cl"},
	{0xF1410A72BC4CA95CULL, "49578443.ogg"},
	{0xF1415CB516D72917ULL, "vxlt_0_lowtown_docks_a3_9zs_0_0.vlm"},
	{0xF14211FEBD3BFEF8ULL, "vxlt_undercityencounter_uq_0_0.vlm"},
	{0xF14222EB97B8CD78ULL, "49176548.ogg"},
	{0xF14234B1165ECF87ULL, "vxlt_gallowsprison_n_6bn_0_0.vlm"},
	{0xF1425FE55CB7018BULL, "6206379_m.fxe"},
	{0xF1427A9E39362348ULL, "art\\characters\\creatures\\mubari\\animation\\ctn180r_crnf.gad"},
	{0xF142EE332E91E40CULL, "25649206.ogg"},
	{0xF1430967650133C3ULL, "6119683_f.fxe"},
	{0xF1430C39E3A42D22ULL, "vxht_brothel_n_3fb_0_0.vlm"},
	{0xF1430C78649CE114ULL, "gencr_gate_slave_r5.utc"},
	{0xF143305AF57F5408ULL, "open_inventory.cur"},
	{0xF143B1BB18BB9223ULL, "dae000_gamlen6149415.cl"},
	{0xF14401CC8300A591ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_020a_d.dds"},
	{0xF14445E570C6D982ULL, "vxlt_0_lowtown_a2_n_hic_0_0.vlm"},
	{0xF14460E755A4743FULL, "hlac_lt_0_lowtown_docks_65533_0y.dds"},
	{0xF144843D943279DBULL, "6051698_m.fxe"},
	{0xF1448C6446F07C9CULL, "mag170_thrask6177341.cl"},
	{0xF1456E0BD20CAC8BULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ia_transition.evt"},
	{0xF1459350CCB26561ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop_lifted.evt"},
	{0xF145D3BC5289C45EULL, "hm_icarver_h1s2.mor"},
	{0xF146497AC99C0899ULL, "vxdr_0_deeproads_ctt_0_0.vlm"},
	{0xF1467203ECCA109FULL, "583244951.ogg"},
	{0xF14687FED3EBBAD9ULL, "vxw_woundedcoast_1du_0_0.vlm"},
	{0xF14721A7EE133852ULL, "234528398.ogg"},
	{0xF1474807629EBE4FULL, "ave331_post_fight6138831.cl"},
	{0xF1479910097FCA83ULL, "218977490.ogg"},
	{0xF14822CFFCA0B036ULL, "6114017_m.fxe"},
	{0xF149028461F07E0CULL, "vxlt_backalley_g_1x1_0_0.vlm"},
	{0xF1495EE069149FAEULL, "qun220cr_javaris6071499.cl"},
	{0xF149C610B9C0E601ULL, "636525334.ogg"},
	{0xF14A1B7046C6CCBDULL, "ser270_cut_aftermath6162296.cl"},
	{0xF14A83C9F3C22432ULL, "vxca_caverns1_dpw_0_0.vlm"},
	{0xF14AC032E1B602C6ULL, "mag101_aombanter6052615.cl"},
	{0xF14AFA25243C652FULL, "vxht_0_hightown_a3_by6_0_0.vlm"},
	{0xF14B0F784FED6613ULL, "234478920.ogg"},
	{0xF14B2A56CF2A9777ULL, "hlac_w_sundermount_327684_0x.dds"},
	{0xF14C66BB462DAE21ULL, "vxht_mansion2_3fi_0_0.vlm"},
	{0xF14C870BD212F64CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ia_transition.gad"},
	{0xF14D82CA396EF872ULL, "6051581_m.fxe"},
	{0xF14DB143F49A24C9ULL, "hlac_lt_0_lowtown_65537_6z.dds"},
	{0xF14DDC2E3C67371DULL, "6101863_m.fxe"},
	{0xF14E2E88CFA9D321ULL, "vxlt_gallowscourtyard_n_7he_0_0.vlm"},
	{0xF14E3E29FCDF8BF7ULL, "llac_w_bonepitdestroyed_a3_0_9.dds"},
	{0xF14E5E2657E117D3ULL, "101809273.ogg"},
	{0xF14E611E62D6B687ULL, "vxht_0_hightown_5tc_0_0.vlm"},
	{0xF14E67ACFBF44DD4ULL, "40586308.ogg"},
	{0xF14EFE357A7E4DD8ULL, "follower_banter6172902.cl"},
	{0xF14F1776E620DB82ULL, "800354338.ogg"},
	{0xF14F3F9997D319EEULL, "gen00fl_merrill6133142.cl"},
	{0xF14F407BBA8A262CULL, "art\\levels\\races\\proxy\\props\\_textures\\bigboat_n.dds"},
	{0xF14F46644F39A1F9ULL, "vxdr_0_deeproads_9iu_0_0.vlm"},
	{0xF15027E7C8B9792DULL, "lowtown_fem_alienage1_sta6162534.cl"},
	{0xF150352199E4F8FDULL, "vxht_brothel_n_4vr_0_0.vlm"},
	{0xF150CA78ADDCCFE0ULL, "vxht_brothel_n_62u_0_0.vlm"},
	{0xF150F3849A518EA0ULL, "15200257.ogg"},
	{0xF150F81D55CC7437ULL, "6104299_m.fxe"},
	{0xF151736B23823229ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_end.ani"},
	{0xF1521E401392C76CULL, "ico_potion_spindleweed.dds"},
	{0xF15235C0B7510A2AULL, "docks_male_elven_worker_sta6175820.cl"},
	{0xF15260E7993D0A25ULL, "gen00fl_aveline6114026.cl"},
	{0xF15306BED2B70CF7ULL, "134705773.ogg"},
	{0xF1536674C7E1F17AULL, "6113027_m.fxe"},
	{0xF153860BD219134AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\pose\\pose_ia_transition.ani"},
	{0xF1538B50CCBE94B0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop_lifted.ani"},
	{0xF153A51FCB11937DULL, "vxht_keep_n_5xn_0_0.vlm"},
	{0xF153ADF0A477FCABULL, "168496533.ogg"},
	{0xF1540F34410D760EULL, "vxht_mansion2b_3ca_0_0.vlm"},
	{0xF1545090A3ABA36FULL, "6218912_m.fxe"},
	{0xF15461E3F4469DFDULL, "762233882.ogg"},
	{0xF1548C3165738556ULL, "6130135_f.fxe"},
	{0xF154D63DEF0B8C57ULL, "art\\cinematics\\props\\w_dgr_106a.mao"},
	{0xF154F9DBC5E1D359ULL, "vxw_woundedcoast_1ot_0_0.vlm"},
	{0xF1552CC1CA0EA74BULL, "t3_arm_pl3.tnt"},
	{0xF155688D0DF1457FULL, "ser270_templar6164085.cl"},
	{0xF15638A4214D2F31ULL, "greygrid.dds"},
	{0xF15658FC51E2EE2FULL, "rdr111dg_harbormaster6016140.cl"},
	{0xF156C4EB589A23ECULL, "vxht_keep_n_8au_0_0.vlm"},
	{0xF1577F76068B0ED6ULL, "zz_followers_banter6157693.cl"},
	{0xF1577FB3A01F80F2ULL, "plt_lgt000pt_ultra_3_03.nss"},
	{0xF1583ED37A6259AEULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedem_bdy_0s.dds"},
	{0xF158701D856D78F6ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.dura.evt"},
	{0xF1588E6B23887EABULL, "art\\characters\\creatures\\desiredemon\\animation\\d_end.gad"},
	{0xF1590ECE753EB7D5ULL, "dae020pt_amb_hightown.ncs"},
	{0xF15916EE4899CA9CULL, "6060709_m.fxe"},
	{0xF1595CFA5AA33495ULL, "6118487_m.fxe"},
	{0xF159DB93A852B23DULL, "6131878_f.fxe"},
	{0xF15A53A8F7CA63DFULL, "369218853.ogg"},
	{0xF15A74A3A1402D9FULL, "745215456.ogg"},
	{0xF15A9128EDC5DF40ULL, "qun231_party_comment6077135.cl"},
	{0xF15B674AA9A84F33ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_enter.ani"},
	{0xF15BEAA2A888583CULL, "vxlt_backalley_n_1u2_0_1.vlm"},
	{0xF15C7BDB2C9C3301ULL, "zz_fol_debug6098093.cl"},
	{0xF15C7D18D24CF9F9ULL, "vxlt_0_lowtown_n_fhu_0_0.vlm"},
	{0xF15CA80AE6C18A68ULL, "vxlt_undercityserialkiller_1d4_0_1.vlm"},
	{0xF15D59C8689AC7F0ULL, "6180973_m.fxe"},
	{0xF15DD5839638DC11ULL, "hlac_lt_darktown_-1_87x.dds"},
	{0xF15DE4F8DCB3315CULL, "174020792.ogg"},
	{0xF15DF4EC11085AD9ULL, "688653755.ogg"},
	{0xF15E671D8571D481ULL, "art\\vfx\\creatures\\fxc_arch_vc_p.dura.gad"},
	{0xF15E76DC966F6C5CULL, "vxca_caverns1_gkf_0_1.vlm"},
	{0xF15E89B9EDC1D614ULL, "156750563.ogg"},
	{0xF15E8B6B238CE510ULL, "art\\characters\\creatures\\desiredemon\\animation\\d_end.evt"},
	{0xF15ECA8AA7BFDE8FULL, "gen00fl_anders6113035.cl"},
	{0xF15EFABC62952F31ULL, "6198133_m.fxe"},
	{0xF15F5ECBDBE52306ULL, "ts_kirk_guards_prog_junk1.gda"},
	{0xF15F9F58D91E4594ULL, "vxw_encounter2_camp_pq_0_1.vlm"},
	{0xF15FC176DBF3B231ULL, "gen00fl_isabela6201248.cl"},
	{0xF15FEF5B7D29F2ECULL, "vxlt_undercitydungeon_1n2_0_0.vlm"},
	{0xF15FEFF3C7905E22ULL, "vxht_brothel_n_3hl_0_0.vlm"},
	{0xF16040DBC0C6155FULL, "follower_banter6173333.cl"},
	{0xF1605040B0BD43A1ULL, "mag291_demon_marethari6028930.cl"},
	{0xF16059D680AE2B8FULL, "cs_act1_intro.wbd"},
	{0xF1605B5928952141ULL, "vxca_caverns2prc_a6u_0_0.vlm"},
	{0xF160EFEC9FEFD996ULL, "6149802_m.fxe"},
	{0xF1612144734C4F5CULL, "6201621_m.fxe"},
	{0xF161AE07A45B710FULL, "llac_lt_0_lowtown_docks_a3_65534_1.dds"},
	{0xF162CF3455C6D1F0ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindows.msh"},
	{0xF162DD3455C6E90AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindows.mmh"},
	{0xF162E13455C6F041ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\9ww_warmwindows.mao"},
	{0xF16316628CA8B2C5ULL, "278441849.ogg"},
	{0xF163B3C65FDD074AULL, "vxlt_0_lowtown_a2_eov_0_1.vlm"},
	{0xF1641A10802AD846ULL, "art\\characters\\playercharacter\\children\\mt_bas_kf_facem1.mao"},
	{0xF164CF4985274062ULL, "552823838.ogg"},
	{0xF165681D8577F52FULL, "art\\vfx\\creatures\\fxc_arch_vc_p.dura.ani"},
	{0xF16628B702E8A63BULL, "42981991.ogg"},
	{0xF166B162E52440D5ULL, "9704571.ogg"},
	{0xF1670D689CF4603EULL, "qun200_party_comment6182718.cl"},
	{0xF16778C4404FAC1FULL, "tbe100_xenon6190983.cl"},
	{0xF16857C56678873CULL, "mag120_macha6091373.cl"},
	{0xF168894E4FD02484ULL, "hlac_lt_0_lowtown_docks_n_-1_3x.dds"},
	{0xF1695E520CC33F66ULL, "vxlt_gallowstemplar_f_4xw_0_0.vlm"},
	{0xF169F18D155B2AEBULL, "603154075.ogg"},
	{0xF16A18C9F5CE7BBEULL, "map_pin_images.gda"},
	{0xF16A2C318F3A5117ULL, "vxdr_0_deeproads_euk_0_0.vlm"},
	{0xF16A50E909809240ULL, "166592711.ogg"},
	{0xF16A9710756CC9D4ULL, "mag160_feynriel6020421.cl"},
	{0xF16AA5E4EE4982F0ULL, "gen00fl_aveline6122763.cl"},
	{0xF16AAD2FE0B18FB7ULL, "w_blightlands_224_0.msh"},
	{0xF16ABF2FE0B1AE1AULL, "w_blightlands_224_0.mao"},
	{0xF16AC32FE0B1B529ULL, "w_blightlands_224_0.mmh"},
	{0xF16AE517EED7719AULL, "docks_male_laborer_mob6175392.cl"},
	{0xF16BF07CC167DD52ULL, "6143458_f.fxe"},
	{0xF16BF38FE5D1C0F5ULL, "and230ip_at_dungeon.ncs"},
	{0xF16C86C976A0FE49ULL, "zz_dre_debug6092015.cl"},
	{0xF16C93A7F6DF8BC6ULL, "vxca_caverns2prc_aew_0_0.vlm"},
	{0xF16D0C4491C90145ULL, "768783358.ogg"},
	{0xF16D1F31FF2D672FULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk04a.phy"},
	{0xF16D69E693B52548ULL, "6199337_m.fxe"},
	{0xF16D74365ED8B1D3ULL, "nat380_temmerin6067372.cl"},
	{0xF16DF91DBC950C7DULL, "hlac_dr_0_deeproads_5839_1y.dds"},
	{0xF16E10766C10F96CULL, "565278308.ogg"},
	{0xF16E19DA5CAB4B39ULL, "485233857.ogg"},
	{0xF16E50B12591865AULL, "hlac_ht_chantry_n_2534_13x.dds"},
	{0xF16E814E7348A7DCULL, "hlac_ht_0_hightown_a2_65535_2y.dds"},
	{0xF16F04619C86F8A0ULL, "fen210_orana6109756.cl"},
	{0xF16F9BA9FF7E78A0ULL, "6084325_m.fxe"},
	{0xF16FEC51B6D176B5ULL, "one101_orders_followers6115897.cl"},
	{0xF1703E4AA9BA65DDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\partypicker_female_mage_enter.gad"},
	{0xF170420D9DEE9252ULL, "vxlt_warehouse_cx_0_0.vlm"},
	{0xF170699039A3B085ULL, "vxlt_undercityencounter_uh_0_0.vlm"},
	{0xF17094DA98F50AA1ULL, "230087704.ogg"},
	{0xF1718A73592CEE15ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fire_c.anb"},
	{0xF17239279112CBCDULL, "367613492.ogg"},
	{0xF172557765638637ULL, "ht_0_hightown_a2_54_0.rml"},
	{0xF17279EC362FB951ULL, "vxht_brothel_n_5bw_0_0.vlm"},
	{0xF172830C35466A38ULL, "vxht_brothel_n_34r_0_0.vlm"},
	{0xF1728A3F5B2B6903ULL, "lt_0_lowtown_docks_a3_27_0_ll.mmh"},
	{0xF172AD1A36624D33ULL, "vxht_mansion_n_4g4_0_0.vlm"},
	{0xF172D5E1C9E33C89ULL, "dae000_denier6162682.cl"},
	{0xF172EADAAAE94562ULL, "673188463.ogg"},
	{0xF17335FCCA3D1E7CULL, "vxlt_undercityserialkiller_1wx_0_0.vlm"},
	{0xF1733B72F523C366ULL, "6176303_m.fxe"},
	{0xF173FA0756E16C8FULL, "hlac_lt_0_lowtown_a2_n_c_-262144_2y.dds"},
	{0xF17419FDBA5A0F33ULL, "42366033.ogg"},
	{0xF1744EC23B54F246ULL, "vxlt_0_lowtown_a2_n_c_hnl_0_0.vlm"},
	{0xF1748DAD5BA37B83ULL, "follower_banter6167985.cl"},
	{0xF174923FC50A1D41ULL, "mag101_aombanter6052773.cl"},
	{0xF1750B99F2242B7CULL, "6118438_f.fxe"},
	{0xF1752D7EB7139CE9ULL, "326244769.ogg"},
	{0xF1757E37E9E6E071ULL, "738834149.ogg"},
	{0xF17594998CED2BF7ULL, "92500894.ogg"},
	{0xF175C0F22BE11684ULL, "42365365.ogg"},
	{0xF175C2456A765C5AULL, "6117287_m.fxe"},
	{0xF175E33C2E87BB00ULL, "560143500.ogg"},
	{0xF175F268399B275EULL, "304132375.ogg"},
	{0xF1762822FAAE8B37ULL, "vxca_caverns2prc_9pm_0_0.vlm"},
	{0xF1767D319C1AB30BULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v2.evt"},
	{0xF176ED4797C75B0DULL, "s_imp_shot.mao"},
	{0xF17785D1E58795B9ULL, "art\\vfx\\materials\\vfx_link_smo_fb.mao"},
	{0xF1780992DF770765ULL, "6169429_m.fxe"},
	{0xF17832BC0932FAFDULL, "vxlt_0_lowtown_docks_c4f_0_0.vlm"},
	{0xF1784F68C91E8F18ULL, "l_one351ar_zevran_cave.bnk"},
	{0xF1785BE7674ECF87ULL, "vxht_mansion_n_4wh_0_0.vlm"},
	{0xF1798E29C67B1345ULL, "458980751.ogg"},
	{0xF179E9FE5CB3AC44ULL, "19085986.ogg"},
	{0xF179F7C067C213EFULL, "dae150_chandan6148781.cl"},
	{0xF17A14CFBF30F25AULL, "692914292.ogg"},
	{0xF17A42C8B0802BBFULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinswall.msh"},
	{0xF17A48C8B0803661ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinswall.mmh"},
	{0xF17A54C8B0804A02ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinswall.mao"},
	{0xF17B5B062FD2F9C5ULL, "722491260.ogg"},
	{0xF17C0AA1E3BE2543ULL, "6071252_m.fxe"},
	{0xF17C19BBDBB07165ULL, "vxlt_gallowscourtyard_6xo_0_0.vlm"},
	{0xF17C82D4D46CA570ULL, "342256022.ogg"},
	{0xF17C836F2BCACFB0ULL, "6046834_f.fxe"},
	{0xF17CDB8690534B55ULL, "273723098.ogg"},
	{0xF17CFB8672D48AC2ULL, "zz_approval_debug6003832.cl"},
	{0xF17D01BE10E820CFULL, "980554615.ogg"},
	{0xF17D65CD32560D6AULL, "6176075_m.fxe"},
	{0xF17D96319C20FCCCULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v2.gad"},
	{0xF17EB3EA8FE4AAD2ULL, "vxlt_warehouse_dy_0_0.vlm"},
	{0xF17EE51B58778160ULL, "vxw_encounter2_km_0_0.vlm"},
	{0xF17EFAB6A0F3DCFDULL, "vxht_keep_n_5kk_0_0.vlm"},
	{0xF17F15D6AE50852CULL, "6034546_m.fxe"},
	{0xF17F37B600E24820ULL, "6020491_m.fxe"},
	{0xF17F723E4A6D85DAULL, "6135661_m.fxe"},
	{0xF17FDC685D6E3DB2ULL, "vxw_bonepit_2aq_0_0.vlm"},
	{0xF18037F6A86FB0B4ULL, "vxlt_gallowscourtyard_66t_0_0.vlm"},
	{0xF1804F729BC41CD9ULL, "vxw_woundedcoast_1py_0_0.vlm"},
	{0xF180F4E9273C82B8ULL, "vxht_brothel_n_33s_0_0.vlm"},
	{0xF1813DEACA2A5913ULL, "vxw_encounter1_cg_0_0.vlm"},
	{0xF1815EAE48A37C86ULL, "367969293.ogg"},
	{0xF1817C8F5D0E463FULL, "vxht_keep_n_8az_0_0.vlm"},
	{0xF181ADEDD870EFCCULL, "895613354.ogg"},
	{0xF181D1660C04AEB2ULL, "vxdr_0_deeproads_gfx_0_1.vlm"},
	{0xF181F05FB8C8ADB1ULL, "vxht_keep_30k_0_0.vlm"},
	{0xF182475BBFAD612FULL, "gen00fl_aveline6139911.cl"},
	{0xF1826033773D3536ULL, "hlac_lt_undercityserialkiller_-1_15x.dds"},
	{0xF1827BCDE212B938ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_tomec_sa.dds"},
	{0xF182A6746EF0DC50ULL, "w_sundermount_64_0_ll.mmh"},
	{0xF182FFF96DB6CEE5ULL, "629169020.ogg"},
	{0xF1831C9133D8641DULL, "6048324_m.fxe"},
	{0xF183282A013D4C97ULL, "dae311ar_tavern_nt.ncs"},
	{0xF18374D5306A4851ULL, "hm_gen_templar.mor"},
	{0xF183923BBEAEB16DULL, "hlac_w_bonepitdestroyed_a3_196611_0y.dds"},
	{0xF183AD4F5B6F7720ULL, "crt_sw_cyc_g.mao"},
	{0xF183CA1A32D466C3ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazier.msh"},
	{0xF183D81A32D47E1DULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazier.mmh"},
	{0xF183DC1A32D4852EULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_brazier.mao"},
	{0xF183FDE75AE0AC1EULL, "113870549.ogg"},
	{0xF18495319C2719CAULL, "art\\characters\\creatures\\dragon\\animation\\p_tw_v2.ani"},
	{0xF184C198EC14C72FULL, "vxht_0_hightown_a3_be2_0_0.vlm"},
	{0xF18551D4F7373921ULL, "art\\characters\\creatures\\stonegolem\\animation\\dieloop.ani"},
	{0xF1863BA606C9665EULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_couldron_d.dds"},
	{0xF18697BAEB7F4E4AULL, "gen00fl_merrill6106090.cl"},
	{0xF186B62698BEAF12ULL, "vxca_caverns1_jfk_0_0.vlm"},
	{0xF186B960423602CCULL, "vxht_keep_6a0_0_0.vlm"},
	{0xF1875D449B742F28ULL, "6168342_m.fxe"},
	{0xF188289EDBB3F720ULL, "vxca_caverns1_mines_jfp_0_0.vlm"},
	{0xF188A772A9D9D9C3ULL, "wm_ico_hightown.dds"},
	{0xF188D309AB31DC7DULL, "531221014.ogg"},
	{0xF18A282A01436BE6ULL, "dae311ar_tavern_nt.lst"},
	{0xF18A34800270C2FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v2.gad"},
	{0xF18AB3A4B5D62E87ULL, "lgt120_orz_bln_official6140186.cl"},
	{0xF18ADA0965FB12FEULL, "6118057_m.fxe"},
	{0xF18B1114F75D1394ULL, "6118132_m.fxe"},
	{0xF18B443AE5B792D5ULL, "shock_projectile_f_c.dds"},
	{0xF18C44DDC775F4F5ULL, "vxht_0_hightown_bgk_0_0.vlm"},
	{0xF18C53C20C5C2951ULL, "948280789.ogg"},
	{0xF18C5CD4F73D6AC3ULL, "art\\characters\\creatures\\stonegolem\\animation\\dieloop.gad"},
	{0xF18C6239C50AFD56ULL, "6054937_m.fxe"},
	{0xF18C6F69DF2D9C05ULL, "lt_0_lowtown_docks_a3_3_0_ll.mmh"},
	{0xF18C807A54349EE4ULL, "6167512_m.fxe"},
	{0xF18C84FB99166030ULL, "hlac_ht_keep_n_4519_13y.dds"},
	{0xF18CA8A699DA19EEULL, "tbe100im_weight_boots.uti"},
	{0xF18CF7CA26A5BA82ULL, "vxht_0_hightown_a3_au8_0_0.vlm"},
	{0xF18DAE9F032CD9EDULL, "hf_gencr_city_guard_archer.mor"},
	{0xF18DBFEEE3718363ULL, "265053824.ogg"},
	{0xF18F3F9276B27885ULL, "493920567.ogg"},
	{0xF18F5DA61A1BEA51ULL, "vxht_brothel_n_5hu_0_0.vlm"},
	{0xF18F61A781D01E01ULL, "6078064_m.fxe"},
	{0xF18FAFC7BC207193ULL, "vxw_encounter1_hf_0_0.vlm"},
	{0xF1900CAD2D8886AAULL, "hlac_lt_0_lowtown_a2_n_-65536_0z.dds"},
	{0xF190AD2C3A0DF0C0ULL, "character_character_p9.psh.fxo.dx11"},
	{0xF190F026D6B1142DULL, "341138977.ogg"},
	{0xF191252845757CEEULL, "vxlt_gallowstemplar_n_5ws_0_0.vlm"},
	{0xF19133800276DFFAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\quick_deathpose_v2.ani"},
	{0xF1916BECEBFC9FCDULL, "1049762329.ogg"},
	{0xF19176EE438190DCULL, "newmatobj3.mao"},
	{0xF191AB2416E4044AULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_conj_c.anb"},
	{0xF191DE7A57F7BB88ULL, "vxlt_foundry_n_wi_0_0.vlm"},
	{0xF191E19E3525CE2EULL, "mer140_bonepit_comments6062370.cl"},
	{0xF191EE7BAA24EF43ULL, "llac_lt_0_lowtown_docks_n_131072_0.dds"},
	{0xF19217CB0B45B5DCULL, "dae000_mother6004095.cl"},
	{0xF1923661A22856AEULL, "vxht_mansion_n_6oc_0_0.vlm"},
	{0xF19253DBB6201514ULL, "gen00fl_player6207779.cl"},
	{0xF1930BDCDE9ABC37ULL, "vxca_caverns2prc_9zo_0_0.vlm"},
	{0xF19328B7FE755792ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_standing_left.anb"},
	{0xF193593D232FD67DULL, "vxdr_slavers_1jj_0_0.vlm"},
	{0xF193751C3F5F122DULL, "qun160_vashothlead6053897.cl"},
	{0xF1939658133CB852ULL, "follower_banter6155529.cl"},
	{0xF193CD22DDDB5CE0ULL, "6095658_f.fxe"},
	{0xF194372A014BD9F1ULL, "dae311ar_tavern_nt.are"},
	{0xF194C161ED4FA4D1ULL, "water_water_p4.psh.fxo"},
	{0xF194E7547BAA6A1DULL, "qun200_party_comment6090452.cl"},
	{0xF1952B3D342C1B4CULL, "vxht_keep_6j1_0_0.vlm"},
	{0xF1956F6E22D53168ULL, "llac_lt_0_lowtown_docks_n_-3_7.dds"},
	{0xF195AADBE3AC6727ULL, "258404590.ogg"},
	{0xF19607D7B2831D96ULL, "6085741_f.fxe"},
	{0xF1968345E45EB301ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_loop.gad"},
	{0xF196DB80CEF54794ULL, "vxht_keep_n_3yf_0_0.vlm"},
	{0xF1970241314E8249ULL, "vxht_keep_n_6v7_0_0.vlm"},
	{0xF197A3FA9EAFE63FULL, "888588453.ogg"},
	{0xF197CA1854493887ULL, "vxdr_primevaltomb_4zs_0_2.vlm"},
	{0xF1981F20F43EBA38ULL, "zz_followers_banter6157934.cl"},
	{0xF19821A88EA4E53BULL, "follower_banter6163996.cl"},
	{0xF1982B92FD4655E0ULL, "6106944_m.fxe"},
	{0xF198A47F55ABA790ULL, "hlac_ht_keep_n_3736_3z.dds"},
	{0xF198CFE9ADD4F0A7ULL, "vxlt_blackemporium_v6_0_0.vlm"},
	{0xF198D3F4247080ADULL, "gen00fl_carver6103689.cl"},
	{0xF198FF3DDCDB598FULL, "vxlt_gallowsprison_n_64l_0_0.vlm"},
	{0xF1992333BC7A9D2FULL, "6164411_m.fxe"},
	{0xF1997EACA2B1BD86ULL, "666009919.ogg"},
	{0xF199A306B43CF7CFULL, "hlac_dr_dungeon_3216_7z.dds"},
	{0xF199AE6D06FCC0E7ULL, "6088803_m.fxe"},
	{0xF199D2AC0CAC9121ULL, "670897297.ogg"},
	{0xF19A1EC80F0AB2EBULL, "qun120cr_javaris6042976.cl"},
	{0xF19A7786A57DBB51ULL, "vxlt_0_lowtown_docks_b4c_0_0.vlm"},
	{0xF19A8B73594FF556ULL, "art\\vfx\\combat\\vfx_m_sta_atk_close_fire_c.mmh"},
	{0xF19AD416B70115D6ULL, "6051637_m.fxe"},
	{0xF19B59C769DCCDD1ULL, "360255354.ogg"},
	{0xF19BB77C11908D84ULL, "vxw_bonepitmining_a2_oc_0_0.vlm"},
	{0xF19BC8D53C63B260ULL, "vxw_encounter2_qk_0_0.vlm"},
	{0xF19CB3C0A6F041E3ULL, "6143001_m.fxe"},
	{0xF19CCB7736719CD7ULL, "6044923_f.fxe"},
	{0xF19CE4F7F4E8F497ULL, "759152288.ogg"},
	{0xF19D0BFA7AA522D1ULL, "hlac_lt_0_lowtown_docks_a3_n_-131075_0y.dds"},
	{0xF19D1BB08C5BE1B4ULL, "vxlt_gallowscourtyard_68z_0_0.vlm"},
	{0xF19D53B1C41F4B9CULL, "vxht_bartrand_n_46w_0_0.vlm"},
	{0xF19D6D49B18BFDAAULL, "mag221ar_store.ncs"},
	{0xF19D8445E464D3AFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_loop.ani"},
	{0xF19D907C9D1B0F75ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_medc_0s.dds"},
	{0xF19DB42F9934C413ULL, "debuffertacticpreset.gda"},
	{0xF19DDE3AD879C6F1ULL, "mcr370_negotiation6080272.cl"},
	{0xF19DE74C00B4F57DULL, "zz_followers_banter6159162.cl"},
	{0xF19E9A46F5BB7BE3ULL, "docks_male_human_worker_sta6176041.cl"},
	{0xF19EAAEB03F30096ULL, "vxw_bonepitdestroyed_a3_82_0_0.vlm"},
	{0xF19EEBD69E7F65DDULL, "follower_banter6151335.cl"},
	{0xF19EF1E719C7F751ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_newlyriumc.mmh"},
	{0xF19EFDE719C80BF2ULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_newlyriumc.mao"},
	{0xF19EFE867CD1715CULL, "vxlt_0_lowtown_docks_a3_d0b_0_1.vlm"},
	{0xF19F0BE719C8230FULL, "art\\levels\\races\\dwarf\\props\\lyrium\\prp_newlyriumc.msh"},
	{0xF19F81EC4E0D44AFULL, "vxht_mansion2_2ld_0_0.vlm"},
	{0xF1A04C84BA501842ULL, "vxca_caverns1_mines_i94_0_0.vlm"},
	{0xF1A059A0363E6ED0ULL, "mcr370_guardsman6087659.cl"},
	{0xF1A073D680E43C65ULL, "cs_act1_intro.bnk"},
	{0xF1A088CBA47255D3ULL, "vxlt_warehouse_wa_0_0.vlm"},
	{0xF1A0A2A13B2DF002ULL, "hlac_ht_bartrand_n_3369_2x.dds"},
	{0xF1A0B0FB8640E148ULL, "gen00fl_sebastian6187768.cl"},
	{0xF1A0D683ED697F86ULL, "68434243.ogg"},
	{0xF1A111E9DC65E91EULL, "vxlt_0_lowtown_n_h71_0_0.vlm"},
	{0xF1A1537A9A2C6332ULL, "hlac_lt_0_lowtown_docks_n_131070_2y.dds"},
	{0xF1A15995606EB9A6ULL, "612145250.ogg"},
	{0xF1A2210A699A48CAULL, "256516005.ogg"},
	{0xF1A2314A43C1302EULL, "hlac_ht_keep_n_3743_0y.dds"},
	{0xF1A26C89AA29EB33ULL, "6015294_m.fxe"},
	{0xF1A27131645242D9ULL, "199463038.ogg"},
	{0xF1A309148E646B1CULL, "752159094.ogg"},
	{0xF1A311C036546F14ULL, "lt_0_lowtown_a3_n_c_46_0_ll.mmh"},
	{0xF1A35A2F6ED9D721ULL, "vxlt_0_lowtown_a2_e2w_0_0.vlm"},
	{0xF1A366118E27240FULL, "6080641_m.fxe"},
	{0xF1A3B17E3D311907ULL, "vxht_mansion_n_4de_0_0.vlm"},
	{0xF1A41210B296D7ADULL, "gen00fl_sebastian6189000.cl"},
	{0xF1A45137FD595944ULL, "277508854.ogg"},
	{0xF1A466B5F6C19727ULL, "mag101_grace6034022.cl"},
	{0xF1A4741BEA8EFBB7ULL, "49059543.ogg"},
	{0xF1A47B484A23ACADULL, "135692197.ogg"},
	{0xF1A48F855F574CA6ULL, "dae000ip_crowd_lm3_talk9.utp"},
	{0xF1A4BD3D6DCBA777ULL, "vxht_0_hightown_a3_cd8_0_0.vlm"},
	{0xF1A4BDA41D048F8CULL, "dae000_bartrand6206140.cl"},
	{0xF1A51357146B7908ULL, "6076618_m.fxe"},
	{0xF1A5278329A6D28DULL, "mag100_thrask6032053.cl"},
	{0xF1A55F3BF2223389ULL, "933001229.ogg"},
	{0xF1A59220A9712FF2ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyo_0d.dds"},
	{0xF1A5A51DE30F82D5ULL, "zz_dae_debug6075037.cl"},
	{0xF1A5F34EFC9EDA76ULL, "dae221_viscount_chantry6154263.cl"},
	{0xF1A611F713CAFD0DULL, "vxlt_0_lowtown_f0f_0_0.vlm"},
	{0xF1A6270C76138DB8ULL, "mag121_viveka6156626.cl"},
	{0xF1A62B5FA74BC429ULL, "hlac_dr_dungeon_3215_4x.dds"},
	{0xF1A649C8B0A572DEULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_03ruinswall.phy"},
	{0xF1A66CB407412808ULL, "6128513_m.fxe"},
	{0xF1A7C59244B7E699ULL, "vxht_chantry_2rx_0_0.vlm"},
	{0xF1A853BCD4B7E659ULL, "156548176.ogg"},
	{0xF1A88E274120F59AULL, "190319249.ogg"},
	{0xF1A8B81790A975F7ULL, "6118544_m.fxe"},
	{0xF1A8CA56BBCDAD66ULL, "849381733.ogg"},
	{0xF1A92F921E1F4814ULL, "hlac_lt_0_lowtown_docks_a3_-65538_3z.dds"},
	{0xF1A9618DDE805FF4ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_act_c.anb"},
	{0xF1A999325F91D9D6ULL, "llac_lt_0_lowtown_n_-2_3.dds"},
	{0xF1AA098165C6F452ULL, "vxlt_warehouse_wd_0_0.vlm"},
	{0xF1AA1AE7BFF8AB91ULL, "vxlt_0_lowtown_docks_n_b7u_0_0.vlm"},
	{0xF1AA2FF0F7816E1BULL, "art\\vfx\\models\\vfx_rockarmor_ra.msh"},
	{0xF1AA4DF0F781A105ULL, "art\\vfx\\models\\vfx_rockarmor_ra.mmh"},
	{0xF1AA5FA5F3F97498ULL, "6163371_f.fxe"},
	{0xF1AA8B94CC0E2876ULL, "vxlt_gallowscourtyard_3t9_0_0.vlm"},
	{0xF1AAAE627C64D7E1ULL, "6131352_m.fxe"},
	{0xF1AB47BC2829B0CBULL, "lt_foundry_n_320.anb"},
	{0xF1ABCF33D0B31843ULL, "w_bonepit_46_0.rml"},
	{0xF1ACD5B0123BB0A3ULL, "6110910_f.fxe"},
	{0xF1AD6E2E1A91BEE5ULL, "vxlt_0_lowtown_n_h1t_0_0.vlm"},
	{0xF1ADAC32D004DEB6ULL, "623731178.ogg"},
	{0xF1ADEB62ACFEF9EDULL, "al_warrior_t01.gda"},
	{0xF1AF75F701EC7411ULL, "162091396.ogg"},
	{0xF1AF83C6CE5C019EULL, "vxlt_0_lowtown_n_h22_0_0.vlm"},
	{0xF1AFE102C576F242ULL, "injury.ncs"},
	{0xF1AFEF96C286ED37ULL, "vxca_caverns2prc_9li_0_0.vlm"},
	{0xF1B010237C8967C0ULL, "clf260_harley6069716.cl"},
	{0xF1B0136D18611DDAULL, "hlac_lt_hangedman_n_-1_13z.dds"},
	{0xF1B026DF87075153ULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_wkf.ani"},
	{0xF1B04FB9607D5884ULL, "hlac_lt_gallowstemplar_n_2384_2z.dds"},
	{0xF1B0592E678E5840ULL, "6071079_m.fxe"},
	{0xF1B06021B19FFB7EULL, "6029062_f.fxe"},
	{0xF1B08972AFCB1D12ULL, "follower_banter6163270.cl"},
	{0xF1B0EAAD3A40A94EULL, "413079723.ogg"},
	{0xF1B0F982DC8D9ECBULL, "lt_0_lowtown_28_0_lc.mmh"},
	{0xF1B0F9FC0F823C16ULL, "gen00fl_varric6093613.cl"},
	{0xF1B111BD955FBAC3ULL, "6153970_m.fxe"},
	{0xF1B114B0FA700311ULL, "195279181.ogg"},
	{0xF1B12D18D27170E0ULL, "llac_lt_0_lowtown_docks_n_-262144_3.dds"},
	{0xF1B172A8F3CD7996ULL, "6042093_m.fxe"},
	{0xF1B17B78A0373290ULL, "vxlt_0_lowtown_docks_n_bfg_0_0.vlm"},
	{0xF1B1D7198ADF6D75ULL, "vxlt_gallowstemplar_5bt_0_0.vlm"},
	{0xF1B26B5A85108CFFULL, "284872044.ogg"},
	{0xF1B2981240410699ULL, "511514066.ogg"},
	{0xF1B2B52230981A4FULL, "vxw_bonepitdestroyed_a3_25d_0_0.vlm"},
	{0xF1B2F30B2A065181ULL, "344236052.ogg"},
	{0xF1B327FCEACD62F2ULL, "gen00fl_isabela6211932.cl"},
	{0xF1B362F27B5DD44BULL, "dae000_gamlen6026190.cl"},
	{0xF1B36B04EC965321ULL, "vxw_encounter3_dj_0_0.vlm"},
	{0xF1B3B86B041F5718ULL, "dae310cr_human_m_01.utc"},
	{0xF1B3FF0CDA05CE2EULL, "llac_w_encounter1_5_0.dds"},
	{0xF1B47E4DDEC08FD2ULL, "794176167.ogg"},
	{0xF1B483D405C1158DULL, "vxlt_0_lowtown_f5e_0_0.vlm"},
	{0xF1B5296BBA19E791ULL, "vxca_caverns1_mines_ckg_0_0.vlm"},
	{0xF1B5592C8155A2F2ULL, "vxca_cavern1_haunted_ihw_0_0.vlm"},
	{0xF1B581E31115569BULL, "vxht_0_hightown_a2_7u2_0_0.vlm"},
	{0xF1B5AEB0EC47A024ULL, "741708209.ogg"},
	{0xF1B6D0791280A8C0ULL, "gen00fl_anders6127437.cl"},
	{0xF1B729F4829F8E77ULL, "6117444_m.fxe"},
	{0xF1B7408DDE8C64B7ULL, "art\\vfx\\combat\\vfx_r_dlt_teleport_act_c.mmh"},
	{0xF1B76ADC892ECE3BULL, "s_imp_spit.mao"},
	{0xF1B78CBC82B07D73ULL, "gen00fl_aveline6131898.cl"},
	{0xF1B82F3554AFC788ULL, "6147006_m.fxe"},
	{0xF1B87463B0FECBD4ULL, "vxlt_warehouse_8s_0_0.vlm"},
	{0xF1B87B5E57BD0CD2ULL, "vxlt_warehouse_xe_0_0.vlm"},
	{0xF1B8945F5DA8D613ULL, "vfx_chunkyparticle_p0.vsh.fxo.dx11.assembly"},
	{0xF1B89980DA612D37ULL, "6047518_m.fxe"},
	{0xF1B8C296F9F9DE10ULL, "dae300_alistair_king6134892.cl"},
	{0xF1B8F97AD4EC7AB8ULL, "186570689.ogg"},
	{0xF1B9B21C20DE0F69ULL, "follower_banter6149235.cl"},
	{0xF1BA1751995980ADULL, "vxlt_backalley_g_1vv_0_0.vlm"},
	{0xF1BA9C23FDEA794DULL, "vxlt_0_lowtown_docks_a3_bcy_0_0.vlm"},
	{0xF1BAAE2417070E3DULL, "art\\vfx\\combat\\vfx_m_pri_chain_lit_conj_c.mmh"},
	{0xF1BADF0339927067ULL, "dr_primevaltomb_probe_8k3.dds"},
	{0xF1BAF312D5216516ULL, "vxw_sundermount_37i_0_0.vlm"},
	{0xF1BB3DBD925F9209ULL, "hlac_lt_0_lowtown_docks_n_-65538_0x.dds"},
	{0xF1BB9E41B03979B0ULL, "mag311_huon6172430.cl"},
	{0xF1BBF7E25A332940ULL, "472812754.ogg"},
	{0xF1BC25683D8332D9ULL, "art\\levels\\races\\vfx_models\\_textures\\prp_firelogs_d.dds"},
	{0xF1BC8812F5236CC6ULL, "vxlt_0_lowtown_a2_n_c_huq_0_0.vlm"},
	{0xF1BC89CDE541EF17ULL, "mcr370_assignments.bnk"},
	{0xF1BCAF575CD6790FULL, "258313106.ogg"},
	{0xF1BD09A04B23A848ULL, "plt_mrl300pt_n_newplace.plo"},
	{0xF1BD1EDF8711CD8AULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_wkf.evt"},
	{0xF1BD303E9E3C4942ULL, "vxca_caverns1_mines_i36_0_0.vlm"},
	{0xF1BD30F876628F2AULL, "6094013_f.fxe"},
	{0xF1BDABE54F370DFEULL, "lt_undercitydungeon_mmap.dds"},
	{0xF1BDF13AB2182E99ULL, "6099195_m.fxe"},
	{0xF1BE6173B47D05B7ULL, "vxca_caverns2prc_9kh_0_0.vlm"},
	{0xF1BF90BDB13A875CULL, "641244827.ogg"},
	{0xF1BFED55922D4B10ULL, "vxlt_0_lowtown_docks_n_bcd_0_0.vlm"},
	{0xF1BFF7037102E8C9ULL, "vxlt_0_lowtown_a2_ei2_0_0.vlm"},
	{0xF1BFF80B64AA70ABULL, "288166231.ogg"},
	{0xF1C0434D7C725E71ULL, "481117766.ogg"},
	{0xF1C048F67CD585F5ULL, "vxlt_gallowstemplar_5os_0_0.vlm"},
	{0xF1C06BCDE545A52FULL, "mcr370_assignments.cnv"},
	{0xF1C0B88326E195A4ULL, "bkl_star2.mao"},
	{0xF1C0E3CE9B295957ULL, "dae000_gen_turn_in_36178453.cl"},
	{0xF1C1609C3AC13EF7ULL, "gen00fl_aveline6006077.cl"},
	{0xF1C1874988E8378DULL, "210767125.ogg"},
	{0xF1C1A4CEF6CAE894ULL, "615815749.ogg"},
	{0xF1C1CB9F17AEF814ULL, "215389886.ogg"},
	{0xF1C25438518A5844ULL, "289570292.ogg"},
	{0xF1C257D80730CBC9ULL, "vxht_keep_n_6g4_0_0.vlm"},
	{0xF1C2C2914F8B5FF2ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_killdagger_d.dds"},
	{0xF1C3D6768B37CB71ULL, "hlac_w_bonepitmining_a2_458758_0z.dds"},
	{0xF1C3FDDF8717B4FDULL, "art\\characters\\creatures\\mubari\\animation\\p_tn90r_wkf.gad"},
	{0xF1C47918E2A99D30ULL, "and100_anders6095987.cl"},
	{0xF1C48333147563AEULL, "70962679.ogg"},
	{0xF1C49B531694E1A8ULL, "tbe100_xenon_golem6189745.cl"},
	{0xF1C4FBC3FCBEC344ULL, "vxdr_0_deeproads_chq_0_0.vlm"},
	{0xF1C52CE5CAE515C8ULL, "zz_dre_debug6054627.cl"},
	{0xF1C548AD3A8AEB43ULL, "6110360_m.fxe"},
	{0xF1C60C518F4E56F5ULL, "vxht_0_hightown_bsw_0_0.vlm"},
	{0xF1C69E5EE6531A1CULL, "em_radical.mor"},
	{0xF1C7103BE37D9DE8ULL, "ca_caverns2_water_12194.mmh"},
	{0xF1C71C3BE37DB18BULL, "ca_caverns2_water_12194.mao"},
	{0xF1C72E3BE37DD012ULL, "ca_caverns2_water_12194.msh"},
	{0xF1C768908A2AC4C1ULL, "vxw_woundedcoast_1eb_0_0.vlm"},
	{0xF1C7A2160087E7FCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\set_lowtowndoor_sa.dds"},
	{0xF1C806EBA071CD95ULL, "fen310_fenris6118657.cl"},
	{0xF1C80EEF6AA89E93ULL, "vxht_mansion_n_5cq_0_0.vlm"},
	{0xF1C95F4CF1953CC2ULL, "qcr000_arishok6060132.cl"},
	{0xF1C964EFC7177D96ULL, "vxw_sundermount_32j_0_0.vlm"},
	{0xF1C9651962F96FA0ULL, "337582444.ogg"},
	{0xF1C99508C93C5621ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_welcome.ani"},
	{0xF1C9A1FC2C30F14AULL, "follower_banter6171035.cl"},
	{0xF1C9B3165F1CFD48ULL, "177748532.ogg"},
	{0xF1C9BBFE16CBD4D9ULL, "408242140.ogg"},
	{0xF1CA08F0F762677FULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_d05_0t.dds"},
	{0xF1CA42798FBAEC90ULL, "fen110_party_comment6095222.cl"},
	{0xF1CA466A85A1BC43ULL, "585498099.ogg"},
	{0xF1CA9B43D017FE57ULL, "685423559.ogg"},
	{0xF1CAC9DAAE2148B0ULL, "dae000_arishok6081669.cl"},
	{0xF1CAD3B1E2518609ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_generalwood_sa.dds"},
	{0xF1CAF9EFE7198546ULL, "vxlt_0_lowtown_a2_n_c_hpr_0_0.vlm"},
	{0xF1CB0F94F6ACC80FULL, "pro000_meeran6050099.cl"},
	{0xF1CB7CFDBC480A10ULL, "w_encounter2_camp_16_0.mao"},
	{0xF1CB80FDBC481123ULL, "w_encounter2_camp_16_0.mmh"},
	{0xF1CB8A3C3926E1A4ULL, "vxw_bonepitdestroyed_a3_8p_0_0.vlm"},
	{0xF1CB8EFDBC48297DULL, "w_encounter2_camp_16_0.msh"},
	{0xF1CC6B6A5F02ADE1ULL, "72708809.ogg"},
	{0xF1CD2954B4F1A2C9ULL, "vxdr_dungeon_23h_0_0.vlm"},
	{0xF1CD2ABBFF6F13FEULL, "577437502.ogg"},
	{0xF1CD707877229AFAULL, "vxca_cave_nm_0_0.vlm"},
	{0xF1CE5F3284236390ULL, "vxlt_0_lowtown_docks_n_bxe_0_0.vlm"},
	{0xF1CE7E2DA39F7B9EULL, "vxht_keep_6xm_0_0.vlm"},
	{0xF1CE8A6C3E22921CULL, "lgt160cr_wk_avern_messenger.utc"},
	{0xF1CF2AC2AC139E9CULL, "vxlt_darktown_5jm_0_0.vlm"},
	{0xF1CF46CE06D0506BULL, "575542739.ogg"},
	{0xF1CF95CE004D2597ULL, "llac_lt_0_lowtown_a2_1_3.dds"},
	{0xF1CFC484869BBB74ULL, "622887168.ogg"},
	{0xF1CFCCADCC160E44ULL, "vxht_chantry_2ds_0_0.vlm"},
	{0xF1CFD496412DB6CFULL, "hlac_ht_bartrand_n_2995_1x.dds"},
	{0xF1CFDAA5F024A328ULL, "vxca_caverns2_8c3_0_0.vlm"},
	{0xF1CFED44B7A0ACECULL, "6082741_m.fxe"},
	{0xF1D064619F963098ULL, "335727009.ogg"},
	{0xF1D089B7A808C459ULL, "hlac_lt_0_lowtown_a2_n_-1_7z.dds"},
	{0xF1D0A008C94287C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_welcome.gad"},
	{0xF1D0B469049001FBULL, "hlac_ht_mansion_n_2995_5y.dds"},
	{0xF1D0CF5F4499F2F5ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1junk.msh"},
	{0xF1D0D15F4499F66BULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1junk.mmh"},
	{0xF1D0DD5F449A0A08ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_burn1junk.mao"},
	{0xF1D11EC925F86B9AULL, "6041412_m.fxe"},
	{0xF1D16D9FE421737AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdc.mmh"},
	{0xF1D1719FE42179B1ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdc.mao"},
	{0xF1D17F9FE42191C0ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdc.msh"},
	{0xF1D193500E19D3A8ULL, "586582428.ogg"},
	{0xF1D1A7910ECD0E1DULL, "vxca_cave_23c_0_0.vlm"},
	{0xF1D1BEC6C809ABCEULL, "vxht_0_hightown_a3_blt_0_0.vlm"},
	{0xF1D1E5976F9E6024ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_c.anb"},
	{0xF1D24009834306F9ULL, "815941072.ogg"},
	{0xF1D2531CE7FCBC26ULL, "vxlt_0_lowtown_docks_a3_d09_0_0.vlm"},
	{0xF1D263B4D13F8040ULL, "hlac_lt_undercityserialkiller_-1_0z.dds"},
	{0xF1D2C34C235B2B4DULL, "lt_0_lowtown_docks_a3_n_193_0.msh"},
	{0xF1D2D14C235B4280ULL, "lt_0_lowtown_docks_a3_n_193_0.mao"},
	{0xF1D2D54C235B49D3ULL, "lt_0_lowtown_docks_a3_n_193_0.mmh"},
	{0xF1D3BF12000D3115ULL, "vxlt_0_lowtown_docks_a3_cyw_0_1.vlm"},
	{0xF1D3BF9AE081BA9EULL, "vxht_0_hightown_n_bwg_0_0.vlm"},
	{0xF1D3C5AC99E5C932ULL, "mag311_keran6041715.cl"},
	{0xF1D3D0590DDFF614ULL, "ach_comp_enchant.dds"},
	{0xF1D49D5266E9C51EULL, "vxlt_gallowstemplar_n_43f_0_0.vlm"},
	{0xF1D4BAE48027724FULL, "gentr_party_banter.ncs"},
	{0xF1D524ECD087306FULL, "100166089.ogg"},
	{0xF1D52797D53179D8ULL, "vxlt_gallowscourtyard_n_66s_0_0.vlm"},
	{0xF1D592DC7398AF0DULL, "433397503.ogg"},
	{0xF1D5E352E3A12E40ULL, "hlac_ca_caverns2prc_5802_6z.dds"},
	{0xF1D61748DE8D3A08ULL, "dae300_vashothlead6062773.cl"},
	{0xF1D6AEE5024A1593ULL, "vxlt_hangedman_246_0_0.vlm"},
	{0xF1D6FB0B7D45B1ADULL, "vxlt_backalley_g_1xt_0_0.vlm"},
	{0xF1D7119D9DCC3F78ULL, "llac_ht_mansion_n_3364_0.dds"},
	{0xF1D744BC284EE617ULL, "lt_foundry_n_320.rml"},
	{0xF1D7652871B21A7FULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_act_c.mmh"},
	{0xF1D76BF7A4365F9FULL, "llac_w_bonepitdestroyed_a3_65538_0.dds"},
	{0xF1D778B77720147DULL, "335055817.ogg"},
	{0xF1D77D08C9486AA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\talk_welcome.evt"},
	{0xF1D7972DF81BE23DULL, "ts_dalish_acc_props.gda"},
	{0xF1D7E605F230D71FULL, "vxlt_darktown_2ri_0_0.vlm"},
	{0xF1D8115E396E9521ULL, "658776646.ogg"},
	{0xF1D83F6449663D4CULL, "514817666.ogg"},
	{0xF1D85FA0A5E00DECULL, "static_static_p27.psh.fxo"},
	{0xF1D861CF44A3BF9BULL, "vxlt_gallowstemplar_5cy_0_0.vlm"},
	{0xF1D8B879B86138C2ULL, "vxlt_0_lowtown_docks_n_bgj_0_0.vlm"},
	{0xF1D8F3CC8C22BAC5ULL, "88723944.ogg"},
	{0xF1D93AA7DBC4EAE7ULL, "402524413.ogg"},
	{0xF1DA0BF3832E4EEEULL, "6206675_m.fxe"},
	{0xF1DA37A78A01E6FEULL, "vxca_cavern1_haunted_jgq_0_0.vlm"},
	{0xF1DA523EEE894314ULL, "hlac_lt_0_lowtown_docks_a3_n_-3_2x.dds"},
	{0xF1DAACE7001020B1ULL, "vxdr_0_deeproads_9im_0_0.vlm"},
	{0xF1DAE0005B9BC537ULL, "hlac_lt_0_lowtown_docks_a3_n_-196609_1x.dds"},
	{0xF1DBBFE3F24DAEBFULL, "ico_mm_partymember.dds"},
	{0xF1DBD631AC63D327ULL, "6072716_m.fxe"},
	{0xF1DC81FDBC568648ULL, "w_encounter2_camp_16_0.phy"},
	{0xF1DD2CB060C1B6F5ULL, "vxlt_gallowstemplar_5uq_0_0.vlm"},
	{0xF1DD9C9F9E09B71CULL, "vxlt_darktown_5ob_0_0.vlm"},
	{0xF1DDE7CF4E94B45DULL, "w_encounter3_223_0.mao"},
	{0xF1DDEBCF4E94BA96ULL, "w_encounter3_223_0.mmh"},
	{0xF1DDF5CF4E94CB94ULL, "w_encounter3_223_0.msh"},
	{0xF1DE5DEE08DBC0DDULL, "593464154.ogg"},
	{0xF1DEAB87B2DF5998ULL, "vxht_brothel_n_221_0_0.vlm"},
	{0xF1DEBD33D0DE535AULL, "w_bonepit_46_0.msh"},
	{0xF1DEC3948A7C90F7ULL, "6087275_m.fxe"},
	{0xF1DECB33D0DE6A93ULL, "w_bonepit_46_0.mao"},
	{0xF1DECF33D0DE71A0ULL, "w_bonepit_46_0.mmh"},
	{0xF1DF005CFF44ACB7ULL, "vxht_keep_5u0_0_0.vlm"},
	{0xF1DF06A04B408907ULL, "plt_mrl300pt_n_newplace.nss"},
	{0xF1DF0CBEE72D33BCULL, "dae000_mother6136949.cl"},
	{0xF1DFCE92DDEDC80AULL, "hlac_lt_0_lowtown_docks_n_65535_3z.dds"},
	{0xF1E004976FAAD227ULL, "art\\vfx\\combat\\vfx_m_frc_attwave_c.mmh"},
	{0xF1E06224B4710C87ULL, "6094734_m.fxe"},
	{0xF1E18C84A35F7407ULL, "isa211_isabela6202154.cl"},
	{0xF1E19A548D718967ULL, "hlac_w_sundermount_327676_0y.dds"},
	{0xF1E1EFEE61078DD5ULL, "zz_vault_debug6177773.cl"},
	{0xF1E215C8AEAD454AULL, "ctplt_fir_si.mao"},
	{0xF1E272A3DAE786F7ULL, "hlac_lt_0_lowtown_n_-262144_2y.dds"},
	{0xF1E2F00B733A87F5ULL, "vxht_0_hightown_byu_0_0.vlm"},
	{0xF1E380FDBC5CA340ULL, "w_encounter2_camp_16_0.rml"},
	{0xF1E404D0F744CE28ULL, "vxlt_gallowstemplar_2oq_0_0.vlm"},
	{0xF1E479A42E80E335ULL, "plt_cod_hst_city_elf.plo"},
	{0xF1E4A35F6B7B7C36ULL, "128788853.ogg"},
	{0xF1E4EE66377707E4ULL, "6044817_m.fxe"},
	{0xF1E5BA6E55E14085ULL, "6176289_m.fxe"},
	{0xF1E5D8D7A10B83FCULL, "6031307_m.fxe"},
	{0xF1E5DCF5C60514A1ULL, "vxht_0_hightown_n_ber_0_0.vlm"},
	{0xF1E657E2E426EF9FULL, "vxlt_darktown_2uh_0_0.vlm"},
	{0xF1E6CB476D7730FFULL, "163911037.ogg"},
	{0xF1E6D3AC3699D81BULL, "vxlt_gallowstemplar_5lx_0_0.vlm"},
	{0xF1E6E76C4158E8B2ULL, "991301887.ogg"},
	{0xF1E6F7DD20C2A121ULL, "vxlt_gallowscourtyard_654_0_0.vlm"},
	{0xF1E70AB6485FB1EBULL, "458751555.ogg"},
	{0xF1E742EECC434CA4ULL, "plt_lgt111pt_alistair_drunk.plo"},
	{0xF1E785F736A4FD5AULL, "vxlt_0_lowtown_e8f_0_0.vlm"},
	{0xF1E847E8B175282AULL, "mag101_grace6034083.cl"},
	{0xF1E86A172C17FC9DULL, "6017427_m.fxe"},
	{0xF1E92D2355D74869ULL, "700275000.ogg"},
	{0xF1EA92484001A75DULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtc.phy"},
	{0xF1EAF1A9596B4B8CULL, "vxht_mansion2b_2po_0_0.vlm"},
	{0xF1EB0CD2249AF9B7ULL, "vxht_0_hightown_a3_c3x_0_1.vlm"},
	{0xF1EB42EC680F9490ULL, "vxlt_0_lowtown_docks_n_brc_0_0.vlm"},
	{0xF1EBAE78A03FD755ULL, "vxw_encounter2_camp_j7_0_0.vlm"},
	{0xF1EC1FD357754136ULL, "hlac_ht_0_hightown_65535_2y.dds"},
	{0xF1EC239505AD487CULL, "6118669_f.fxe"},
	{0xF1EC7D8CCDCB31C3ULL, "vxlt_gallowstemplar_f_5k6_0_0.vlm"},
	{0xF1EDF5DE655548FFULL, "static_static_p22.psh.fxo.dx11"},
	{0xF1EE0D2B070FD3F7ULL, "107427460.ogg"},
	{0xF1EE6334AEEEEAF6ULL, "mag110_samson6049741.cl"},
	{0xF1EEA280ABF5DCCEULL, "vxht_0_hightown_a3_bfv_0_0.vlm"},
	{0xF1EF178CFD932B69ULL, "vxlt_0_lowtown_docks_n_bdd_0_0.vlm"},
	{0xF1EF33F006C97D56ULL, "vxlt_darktown_5m2_0_0.vlm"},
	{0xF1EF38788C89B693ULL, "vxca_caverns1_mines_fmz_0_0.vlm"},
	{0xF1EF38DF4D11442FULL, "l_env_wal_lowtowndocks_military_night.wbd"},
	{0xF1EF4F8372C7F550ULL, "hlac_lt_backalley_n_65537_5z.dds"},
	{0xF1EF6810C2FF1394ULL, "vxht_chantry_n_33p_0_0.vlm"},
	{0xF1EF89542DC22A24ULL, "follower_barks6153346.cl"},
	{0xF1F0079CE4B1150EULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_keeplogo.msh"},
	{0xF1F0159CE4B12D5FULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_keeplogo.mao"},
	{0xF1F0219CE4B141FCULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_keeplogo.mmh"},
	{0xF1F05AEEFF6559FAULL, "588631153.ogg"},
	{0xF1F0A3D74A5E1CBEULL, "mag291_torpor6021690.cl"},
	{0xF1F0F3F651586C3CULL, "vxht_0_hightown_n_7u2_0_0.vlm"},
	{0xF1F1851F1102E2F0ULL, "vxlt_0_lowtown_a3_n_c_fe0_0_0.vlm"},
	{0xF1F2A83EB9438896ULL, "art\\characters\\creatures\\pridedemon\\textures\\c_pridedemb_hed_0s.dds"},
	{0xF1F30DB11874E73FULL, "ave200_donnic6111987.cl"},
	{0xF1F31A716648F871ULL, "hlac_ht_0_hightown_a3_-196610_0y.dds"},
	{0xF1F3381A82364140ULL, "hlac_lt_0_lowtown_-196608_1y.dds"},
	{0xF1F35B45A20A268FULL, "6150706_m.fxe"},
	{0xF1F36486408AE813ULL, "vxht_mansion_n_5rr_0_0.vlm"},
	{0xF1F3A4415CAC9062ULL, "382224469.ogg"},
	{0xF1F3A4505BC46469ULL, "hlac_ht_keep_3731_2z.dds"},
	{0xF1F3A7E4804202F6ULL, "gentr_party_banter.utt"},
	{0xF1F47DC845DF2A6AULL, "6176224_m.fxe"},
	{0xF1F4851659A114E4ULL, "vxlt_0_lowtown_a2_n_c_ep4_0_1.vlm"},
	{0xF1F493171CC15BB1ULL, "mer110_jansen6032868.cl"},
	{0xF1F4C9BFD61D081FULL, "vxlt_darktown_2xk_0_0.vlm"},
	{0xF1F548404A864B08ULL, "vrc321tr_door_shift_2.utt"},
	{0xF1F5822ED8441B2BULL, "306984618.ogg"},
	{0xF1F5D7E9E22BDAECULL, "vxw_woundedcoast_1bz_0_0.vlm"},
	{0xF1F5F7D4289B15DAULL, "vxlt_0_lowtown_e3g_0_0.vlm"},
	{0xF1F622B24212EC5AULL, "vxht_mansion_n_32y_0_0.vlm"},
	{0xF1F6289563A76BF8ULL, "263105164.ogg"},
	{0xF1F6358977CDD358ULL, "lgt320_shw_ors_black_op6156325.cl"},
	{0xF1F677EE84DB2768ULL, "qun211cr_braggart_ally_m.utc"},
	{0xF1F68C9FE440C385ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockdc.phy"},
	{0xF1F6B3C3427D6FA8ULL, "lt_gallowstemplar_f_2393.rml"},
	{0xF1F6CE505E0C87C9ULL, "zz_vault_debug6203182.cl"},
	{0xF1F6EBDBCB28D75AULL, "hlac_lt_darktown_-1_2z.dds"},
	{0xF1F75440EEA33684ULL, "445113569.ogg"},
	{0xF1F78E2A8D7A26C3ULL, "423766050.ogg"},
	{0xF1F790A0E3FC51B1ULL, "vxdr_0_deeproads_9gc_0_0.vlm"},
	{0xF1F7EB0748B048E8ULL, "vxlt_gallowstemplar_f_3sc_0_0.vlm"},
	{0xF1F8405AA5E59A05ULL, "hairalpha.mat.dx11"},
	{0xF1F8759BF3D3214BULL, "art\\vfx\\combat\\vfx_c_debili_cast_aoe_c.impa.ani"},
	{0xF1F8C722043CBDBDULL, "lt_0_lowtown_docks_n_33_0_lc.mmh"},
	{0xF1F92DF91ADABD47ULL, "qun211_braggart6075750.cl"},
	{0xF1F9B06CA4647102ULL, "161936806.ogg"},
	{0xF1F9BAF01353F3F6ULL, "gen_im_trs_cot_lgt_cht_r0.uti"},
	{0xF1FA20559235EFD5ULL, "vxw_encounter2_camp_e6_0_0.vlm"},
	{0xF1FA348C4761811DULL, "6048175_m.fxe"},
	{0xF1FA3EECEE11346AULL, "vxw_bonepitmining_a2_8d_0_0.vlm"},
	{0xF1FA424B9D50C85FULL, "and100_anders6095920.cl"},
	{0xF1FA4E0CA0D68355ULL, "vxdr_slavers_15k_0_0.vlm"},
	{0xF1FA5B68ABF85023ULL, "hlac_lt_gallowscourtyard_n_-1_4x.dds"},
	{0xF1FA73D8ADA2F05CULL, "hlac_ht_chantry_2534_24y.dds"},
	{0xF1FAC0884FFF2A1AULL, "vxca_cave_1ah_0_1.vlm"},
	{0xF1FB782E817D3953ULL, "vxw_encounter3_4p_0_0.vlm"},
	{0xF1FB8748401007C9ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtc.mao"},
	{0xF1FB88E87A907E21ULL, "139170383.ogg"},
	{0xF1FB934840101C62ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtc.mmh"},
	{0xF1FB954840101F18ULL, "art\\characters\\playercharacter\\humanfemale\\boo_lgtc.msh"},
	{0xF1FBEC6B49820924ULL, "hlac_ca_caverns2_5802_7z.dds"},
	{0xF1FC10F170A1D0A5ULL, "zz_vrc_debug6064292.cl"},
	{0xF1FC2CA1B7BB74F3ULL, "6135101_m.fxe"},
	{0xF1FC77A930273024ULL, "6119324_m.fxe"},
	{0xF1FC7D559EB2DD73ULL, "hlac_lt_0_lowtown_65537_6x.dds"},
	{0xF1FC9179FC2FDF31ULL, "hlac_ca_caverns2prc_5802_64y.dds"},
	{0xF1FD7BABCD8137B3ULL, "152628700.ogg"},
	{0xF1FD8969EF8943E9ULL, "vxlt_0_lowtown_docks_n_bak_0_0.vlm"},
	{0xF1FDA30C41B47AC1ULL, "6208299_f.fxe"},
	{0xF1FDC86AA6B645EEULL, "qun131_taskmaster6039722.cl"},
	{0xF1FDE5C512D408E2ULL, "931288412.ogg"},
	{0xF1FDE8D519B6942DULL, "343923258.ogg"},
	{0xF1FDEA525F60C455ULL, "dae000cr_grand_cleric6147677.cl"},
	{0xF1FEB846DEAAD136ULL, "vxht_keep_6y2_0_0.vlm"},
	{0xF1FEB8618748B489ULL, "zz_followers_banter6158064.cl"},
	{0xF1FED382B684A86AULL, "65029086.ogg"},
	{0xF1FF35D1A8ABA75DULL, "6128642_m.fxe"},
	{0xF1FF9F00401E46E7ULL, "gen00fl_aveline6132815.cl"},
	{0xF1FFF33EE87CA78FULL, "gen00fl_anders6119396.cl"},
	{0xF2000150BB34D966ULL, "vxht_keep_60m_0_0.vlm"},
	{0xF2003EB5E90C1828ULL, "128359000.ogg"},
	{0xF200662871D520BCULL, "art\\vfx\\combat\\vfx_ha_hd_detonate_act_c.anb"},
	{0xF2007CB6B4CAF072ULL, "vxlt_gallowscourtyard_7ab_0_0.vlm"},
	{0xF2007E85527C586FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfalll.mao"},
	{0xF2008A85527C6C0CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfalll.mmh"},
	{0xF2009085527C76FEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfalll.msh"},
	{0xF200BD5835A8047EULL, "llac_lt_0_lowtown_131071_0.dds"},
	{0xF2013EA09AFA3F82ULL, "6207665_m.fxe"},
	{0xF201465FD78BCCEFULL, "vxca_cavern1_haunted_hih_0_0.vlm"},
	{0xF2016BF1E210812EULL, "hlac_w_bonepit_65542_0x.dds"},
	{0xF20285B84AFE3686ULL, "vxca_cavern1_haunted_jld_0_0.vlm"},
	{0xF202A0C6F6F75010ULL, "hlac_ht_chantry_n_2534_24x.dds"},
	{0xF202F7EA0E5259CAULL, "314054964.ogg"},
	{0xF2033D7E7BD7D048ULL, "vxht_mansion2b_1ge_0_0.vlm"},
	{0xF20387692FE7DF7CULL, "plt_genpt_personality.plo"},
	{0xF204207CBD7F064BULL, "dae000_gen_player_turn_in6177854.cl"},
	{0xF204C64EFE4C2FBAULL, "608988447.ogg"},
	{0xF205CA64F2B06F27ULL, "180036432.ogg"},
	{0xF206429E79A8A0E6ULL, "dae000ip_crowd_mf5_talk7.utp"},
	{0xF2065D9BF3DF35C2ULL, "art\\vfx\\combat\\vfx_c_debili_cast_aoe_c.impa.evt"},
	{0xF206B0ADB2122C65ULL, "6124673_m.fxe"},
	{0xF206BD33C5B33F70ULL, "7621677.ogg"},
	{0xF206C01C68678E94ULL, "hlac_dr_slavers_3217_0y.dds"},
	{0xF20738F70B56FC2DULL, "follower_banter6141503.cl"},
	{0xF2083C918B5DDA62ULL, "6128945_f.fxe"},
	{0xF2084B8498DA3897ULL, "6099654_m.fxe"},
	{0xF208923F0F213D72ULL, "vxht_keep_5k3_0_0.vlm"},
	{0xF2090C66F966BE6EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunklo.msh"},
	{0xF2091A66F966D5BFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunklo.mao"},
	{0xF2092666F966EA5CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\0ww23_cellchunklo.mmh"},
	{0xF209EA0B737351D3ULL, "vxw_encounter3_1s_0_0.vlm"},
	{0xF20A502AA1B2EAC4ULL, "hlac_lt_gallowstemplar_2394_4z.dds"},
	{0xF20AA33E9B59CC2DULL, "hlac_ht_0_hightown_65536_4y.dds"},
	{0xF20B464377658265ULL, "prty_aveline_armor_2.dds"},
	{0xF20B4F8D6AF51990ULL, "mag170_thrask6054099.cl"},
	{0xF20B60A241B2439BULL, "325770672.ogg"},
	{0xF20B687CD5C6BC30ULL, "vxw_bonepitdestroyed_a3_27a_0_0.vlm"},
	{0xF20B7121073FAE75ULL, "zz_fen_debug6134364.cl"},
	{0xF20C5C9BF3E39FD5ULL, "art\\vfx\\combat\\vfx_c_debili_cast_aoe_c.impa.gad"},
	{0xF20D2A23D0A0E9B6ULL, "vxht_keep_6b3_0_0.vlm"},
	{0xF20D4E57AC4D9C40ULL, "ave100_ambush_1.cnv"},
	{0xF20DA80D37AFF389ULL, "6078675_m.fxe"},
	{0xF20DC736FEF075BDULL, "142274933.ogg"},
	{0xF20E4ACF6D44FEBDULL, "vxlt_shop_sh_0_0.vlm"},
	{0xF20EB5A6477EE4A7ULL, "vxlt_undercityserialkiller_1zt_0_0.vlm"},
	{0xF20ED660258FBEAAULL, "vxlt_0_lowtown_a2_n_c_hdb_0_0.vlm"},
	{0xF20EF7AC463B93F3ULL, "vxht_0_hightown_ap7_0_0.vlm"},
	{0xF20F00B35B7DA02BULL, "vxlt_undercityserialkiller_1it_0_1.vlm"},
	{0xF20FF0E042B5A334ULL, "lt_0_lowtown_docks_a3_6_0.rml"},
	{0xF2103057AC4F9F78ULL, "ave100_ambush_1.bnk"},
	{0xF2110CA9D41A5DF3ULL, "art\\vfx\\combat\\vfx_c_ogr_rock_conj_c.anb"},
	{0xF2111161171495B1ULL, "gen00fl_fenris6113871.cl"},
	{0xF2114EBB5DB71C87ULL, "vxlt_gallowscourtyard_740_0_0.vlm"},
	{0xF2116AC4A439B830ULL, "vxlt_0_lowtown_a2_f4o_0_0.vlm"},
	{0xF211D1EB9E625B08ULL, "hlac_w_bonepitdestroyed_a3_327682_0z.dds"},
	{0xF2124AAFB18E87ECULL, "zz_fen_debug6134194.cl"},
	{0xF2127535982FB31AULL, "gen00fl_merrill6109635.cl"},
	{0xF21284DB4FA48A6FULL, "hlac_lt_backalley_g_0_2y.dds"},
	{0xF212BBA3C6180BECULL, "vxw_woundedcoast_1hx_0_0.vlm"},
	{0xF2131DF204AC66D9ULL, "dae000ip_ambient_talk.utp"},
	{0xF213701CE2AA7939ULL, "follower_banter6141622.cl"},
	{0xF213945E129D9785ULL, "6013328_m.fxe"},
	{0xF2148CFFE39B3C6EULL, "hlac_lt_undercityserialkiller_-1_48z.dds"},
	{0xF214B17BD551A4A0ULL, "mcr310_sibling6076244.cl"},
	{0xF214B7B50411D9DEULL, "129356639.ogg"},
	{0xF214C2D263A61011ULL, "hlac_w_woundedcoast_131080_0y.dds"},
	{0xF214EA74257AA57BULL, "828824483.ogg"},
	{0xF214F450CF986532ULL, "655626804.ogg"},
	{0xF215BF8AF88D4C2CULL, "6111254_m.fxe"},
	{0xF215F8A7201BD8DEULL, "llac_lt_0_lowtown_a2_65536_0.dds"},
	{0xF217825EC2A2D9CFULL, "vxw_encounter2_camp_61_0_0.vlm"},
	{0xF217CE978CBE7C01ULL, "darktown_fem_elf_sta6174511.cl"},
	{0xF218C534D1DFCF24ULL, "vxlt_undercityencounter_17l_0_0.vlm"},
	{0xF218EEFF4EB209D5ULL, "vxlt_0_lowtown_docks_cos_0_0.vlm"},
	{0xF218F98F11656CDBULL, "296783219.ogg"},
	{0xF2195B6314136124ULL, "6113524_m.fxe"},
	{0xF219C1D7466286DDULL, "246100451.ogg"},
	{0xF21A96D1C33F8D22ULL, "vxw_bonepitmining_a2_23y_0_0.vlm"},
	{0xF21B3D1ED53CE9EBULL, "18085629.ogg"},
	{0xF21B505F307763A8ULL, "vxht_0_hightown_a2_bkz_0_0.vlm"},
	{0xF21B6DD24F84CC6AULL, "357741476.ogg"},
	{0xF21BB7B0A0F97E0CULL, "character_character_p249.psh.fxo.dx11.assembly"},
	{0xF21BBC26CEAD56AEULL, "569867740.ogg"},
	{0xF21BF1208CBE1985ULL, "one300_nuncio6084355.cl"},
	{0xF21C1FA605EC7063ULL, "getinventory.ncs"},
	{0xF21C3A323D7E7ED1ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.dura.ani"},
	{0xF21C477E191CC3ADULL, "vxlt_backalley_g_207_0_0.vlm"},
	{0xF21D22D1ED3BE369ULL, "llac_lt_0_lowtown_a2_n_-131071_1.dds"},
	{0xF21DF45C4C343C6DULL, "vxca_caverns1_mines_hm6_0_0.vlm"},
	{0xF21E9A38AD7B8A0FULL, "894301266.ogg"},
	{0xF21ED011CDD086F1ULL, "651037524.ogg"},
	{0xF21F212538741FBCULL, "lt_0_lowtown_docks_n_240_0.mao"},
	{0xF21F29A9D426CC5CULL, "art\\vfx\\combat\\vfx_c_ogr_rock_conj_c.mmh"},
	{0xF21F2D253874345FULL, "lt_0_lowtown_docks_n_240_0.mmh"},
	{0xF21F332538743E01ULL, "lt_0_lowtown_docks_n_240_0.msh"},
	{0xF21FA43AA8940910ULL, "vxlt_gallowstemplar_n_522_0_0.vlm"},
	{0xF21FD7CC934EA675ULL, "dae000_arishok6067806.cl"},
	{0xF21FDDFE22675AA3ULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_p.anb"},
	{0xF22032CEA7D19BBDULL, "60871092.ogg"},
	{0xF2208F8DE69FFD3FULL, "vxw_woundedcoast_1bu_0_0.vlm"},
	{0xF220F60D05C703B9ULL, "tell_03.dds"},
	{0xF2212D80B80E246CULL, "vxw_woundedcoast_1sy_0_0.vlm"},
	{0xF2222ADE88C85B3EULL, "gen00fl_aveline6113996.cl"},
	{0xF2222CAAF4CDDD6AULL, "16026971.ogg"},
	{0xF22245323D82FD73ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.dura.gad"},
	{0xF222DD05C7218B05ULL, "267550505.ogg"},
	{0xF22420C602FAF4CEULL, "275219736.ogg"},
	{0xF22457EECC771EEBULL, "plt_lgt111pt_alistair_drunk.nss"},
	{0xF224926E657E6384ULL, "512144551.ogg"},
	{0xF2249C3570B2F3ECULL, "llac_lt_blackemporium_1055_13.dds"},
	{0xF2257B9D95478E29ULL, "dae000_mother6121007.cl"},
	{0xF2257E7599364451ULL, "vxw_woundedcoast_ik_0_0.vlm"},
	{0xF227010165AB89E4ULL, "water_water_p1.psh.fxo"},
	{0xF2274AF37617342BULL, "vxw_woundedcoast_5m_0_0.vlm"},
	{0xF2281FFDEFB7F635ULL, "vxht_bartrand_n_6pz_0_0.vlm"},
	{0xF2287687B792AE5DULL, "550536081.ogg"},
	{0xF2288EAA22B48F8DULL, "59519552.ogg"},
	{0xF22942323D8916D8ULL, "art\\vfx\\combat\\vfx_m_pri_stonefist_p.dura.evt"},
	{0xF2294DF9FD9EC56EULL, "6138815_m.fxe"},
	{0xF229B37BE7FD1C3BULL, "w_bonepitmining_a2_175_0.mmh"},
	{0xF229BF7BE7FD30D8ULL, "w_bonepitmining_a2_175_0.mao"},
	{0xF229D17BE7FD4F25ULL, "w_bonepitmining_a2_175_0.msh"},
	{0xF22A0DDDB48D1AB6ULL, "vxht_keep_6h5_0_0.vlm"},
	{0xF22AE61496CDC005ULL, "dae000_meredith6085391.cl"},
	{0xF22BDFEA08F6D8DFULL, "6120565_m.fxe"},
	{0xF22C550ED023E462ULL, "hlac_ht_keep_n_4519_18z.dds"},
	{0xF22C8B8BAA06F83BULL, "vxht_mansion2_1ds_0_0.vlm"},
	{0xF22C9F713EDC6461ULL, "vxlt_0_lowtown_a2_n_gnr_0_0.vlm"},
	{0xF22D3A79DF845B49ULL, "688767324.ogg"},
	{0xF22D9818388D2B26ULL, "llac_w_bonepitmining_a2_0_11.dds"},
	{0xF22DA5CDF34D42D9ULL, "vxw_woundedcoast_14e_0_0.vlm"},
	{0xF22DBAFE22735CCCULL, "art\\vfx\\combat\\vfx_w_gn2_knbk_aoe_p.mmh"},
	{0xF22DC675C66BB717ULL, "685688279.ogg"},
	{0xF22E05EE3A446D9DULL, "zz_followers_banter6157959.cl"},
	{0xF22E21680AB56606ULL, "llac_ca_caverns1_mines_24426_11.dds"},
	{0xF22E39681616024BULL, "838325617.ogg"},
	{0xF22E80729FD201F6ULL, "6122074_m.fxe"},
	{0xF22E8472273E9A61ULL, "714683103.ogg"},
	{0xF22EE85AA32F2586ULL, "follower_banter6155907.cl"},
	{0xF22F016AD89615BFULL, "vxw_woundedcoast_1mt_0_0.vlm"},
	{0xF22F326BD5DC8AC3ULL, "llac_lt_0_lowtown_a2_n_c_-2_2.dds"},
	{0xF22F3E071AADCDC1ULL, "lowtown_male_alienage1_sta6162720.cl"},
	{0xF23052DF4D490705ULL, "l_env_wal_lowtowndocks_military_night.bnk"},
	{0xF23054FAB3090D44ULL, "117900828.ogg"},
	{0xF230A5DB6CF10752ULL, "art\\characters\\creatures\\spider\\textures\\c_poispidra_0d.dds"},
	{0xF230B9EA54E49767ULL, "6206973_m.fxe"},
	{0xF23104E6D6B946F4ULL, "808045256.ogg"},
	{0xF23140D6F7BB0F5EULL, "277304201.ogg"},
	{0xF23159E234C59F04ULL, "ss_bethany.bnk"},
	{0xF2318E9C4E0DE313ULL, "ts_apostates_01.gda"},
	{0xF231DAFBE0517CEFULL, "vxca_caverns1_ck4_0_0.vlm"},
	{0xF231EF373F1A97CAULL, "vxht_0_hightown_a3_a7z_0_0.vlm"},
	{0xF231F5ADADCBA0A2ULL, "mer140ip_dragon_treasure.utp"},
	{0xF2320C178D2B38BFULL, "299981818.ogg"},
	{0xF23217AD49376360ULL, "em_gen_1.mor"},
	{0xF232792260E857AFULL, "llac_lt_undercityserialkiller_-1_27.dds"},
	{0xF2330611441A3C60ULL, "129705435.ogg"},
	{0xF2339C48A11FC02FULL, "hlac_lt_undercitydungeon_-1_37x.dds"},
	{0xF233F0528B2C5CD1ULL, "vxw_woundedcoast_fj_0_0.vlm"},
	{0xF234107D1FFB460EULL, "vxlt_gallowstemplar_f_3rb_0_0.vlm"},
	{0xF23421E874B8B519ULL, "259545917.ogg"},
	{0xF2344E181578FE27ULL, "llac_lt_ruinedhouse_n_-1_5.dds"},
	{0xF234519F124F114CULL, "vxca_cave_224_0_1.vlm"},
	{0xF23478AF221C987EULL, "vxlt_backalley_g_1et_0_0.vlm"},
	{0xF2353ADD6F819D3FULL, "6062115_m.fxe"},
	{0xF2356BC96E3D2FA1ULL, "vxw_encounter3_9a_0_0.vlm"},
	{0xF2357C0DF295105CULL, "mag210_arianni6024546.cl"},
	{0xF235BCD0680D4CABULL, "vxw_woundedcoast_2l_0_0.vlm"},
	{0xF235D2B9329E3AD5ULL, "vxlt_0_lowtown_docks_cuq_0_0.vlm"},
	{0xF235F482E1A0F91FULL, "6125250_m.fxe"},
	{0xF236187E1703A7FBULL, "w_encounter3_194_0.mmh"},
	{0xF2361F9C1094DCF1ULL, "mag360_thrask6043916.cl"},
	{0xF236247E1703BB98ULL, "w_encounter3_194_0.mao"},
	{0xF236282700311048ULL, "vxlt_0_lowtown_a2_n_gnw_0_0.vlm"},
	{0xF236367E1703DAE5ULL, "w_encounter3_194_0.msh"},
	{0xF236B5D59CD90CC7ULL, "hf_lirene.mor"},
	{0xF236E5A61BCE9524ULL, "gen_im_trs_crc_lgt_hlm_r0.uti"},
	{0xF236EDCC88CC745AULL, "hlac_dr_primevaltomb_661_2y.dds"},
	{0xF2378F8CC0698CD6ULL, "vxlt_0_lowtown_docks_a3_n_d3k_0_0.vlm"},
	{0xF237CD6240365662ULL, "mcr310_argument6182470.cl"},
	{0xF238157AE2D9EB72ULL, "mag110_samson6018596.cl"},
	{0xF2383A808A464DACULL, "494133086.ogg"},
	{0xF23860E0180925FEULL, "6140074_m.fxe"},
	{0xF238AA010E01F2B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\humanoid_local_qunari_grovel.anb"},
	{0xF239085DD81F92D8ULL, "character_character_p36.psh.fxo.dx11"},
	{0xF2397517226B3FB8ULL, "791272598.ogg"},
	{0xF23A13DA22BF32C5ULL, "6146992_m.fxe"},
	{0xF23B248FA0713817ULL, "6165931_m.fxe"},
	{0xF23BEC82893C9CB3ULL, "98659789.ogg"},
	{0xF23BFDC4AA2FAB4DULL, "zz_followers_banter6157249.cl"},
	{0xF23C8A451723B40CULL, "6131595_m.fxe"},
	{0xF23D7347CA8C2E3FULL, "vxw_woundedcoast_1hw_0_0.vlm"},
	{0xF23DC0A603907A13ULL, "285903073.ogg"},
	{0xF23E0C0C718519D6ULL, "llac_lt_0_lowtown_a3_n_c_-65535_6.dds"},
	{0xF23E1269FC2FAC80ULL, "gen00fl_aveline6137484.cl"},
	{0xF23E53C3FCBF7D72ULL, "vxht_keep_n_4bn_0_0.vlm"},
	{0xF23E92DD2447889FULL, "vxlt_undercityencounter_1cw_0_0.vlm"},
	{0xF23EA0857FF1D86FULL, "670911052.ogg"},
	{0xF23EBCA9AAF521A0ULL, "gen00fl_anders6121181.cl"},
	{0xF23F24E7351A18A6ULL, "ser120_ghyslain6105487.cl"},
	{0xF23F7B396E537854ULL, "vxht_keep_n_67e_0_0.vlm"},
	{0xF23F7B9ED71E4AAAULL, "vxlt_hangedman_n_20l_0_0.vlm"},
	{0xF24019C3C6689E96ULL, "6087289_m.fxe"},
	{0xF24075C02C1BE681ULL, "gen_im_npc_elf_arm_boo_lgte.uti"},
	{0xF2409F7EBBA5C1FFULL, "vxca_caverns2prc_9v3_0_0.vlm"},
	{0xF2425EE8DCE22B26ULL, "6029403_m.fxe"},
	{0xF242622F7D227551ULL, "vxw_woundedcoast_ci_0_0.vlm"},
	{0xF242B373612C16BCULL, "558115234.ogg"},
	{0xF242E9848C314714ULL, "llac_dr_dungeon_3215_1.dds"},
	{0xF24396A6F55C79EBULL, "6143258_m.fxe"},
	{0xF24421D31FF7343EULL, "character_character_p256.psh.fxo.dx11"},
	{0xF244390C37F22D91ULL, "6072725_m.fxe"},
	{0xF24462848A14B6F2ULL, "vxw_bonepitdestroyed_a3_1vg_0_0.vlm"},
	{0xF244A6B61EB028E6ULL, "dae300_alistair_king6133778.cl"},
	{0xF24558128B54D0F8ULL, "hlac_ht_0_hightown_131073_0x.dds"},
	{0xF2456FBD4AAC54B8ULL, "seb221_ruxton_harimann6198351.cl"},
	{0xF245C5D22517DEE3ULL, "599820345.ogg"},
	{0xF245F44775871369ULL, "gen00fl_anders6119387.cl"},
	{0xF245FC51C942DDA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_enter.gad"},
	{0xF2462A1EFE15F528ULL, "6149909_m.fxe"},
	{0xF24648504724CBD7ULL, "79348912.ogg"},
	{0xF2466EC877199AF5ULL, "gen00fl_varric6105418.cl"},
	{0xF246D352CC6D368EULL, "948617337.ogg"},
	{0xF246DAC9F970313BULL, "static_motionvectors_p0.vsh.fxo.dx11.assembly"},
	{0xF247127D2B33D12BULL, "qun231_varnell6078137.cl"},
	{0xF2471F474B7470A7ULL, "hlac_lt_warehouse_1_0y.dds"},
	{0xF247EE0F99DEF115ULL, "vxht_chantry_n_20e_0_0.vlm"},
	{0xF2487779EDC5D3E6ULL, "6027550_f.fxe"},
	{0xF2487BA8C268E7C4ULL, "116837903.ogg"},
	{0xF249D572B22DB136ULL, "w_encounter3_170_0.mao"},
	{0xF249E172B22DC5D5ULL, "w_encounter3_170_0.mmh"},
	{0xF249E372B22DC94BULL, "w_encounter3_170_0.msh"},
	{0xF24A21F1D438F9EFULL, "508044576.ogg"},
	{0xF24A388607551C36ULL, "333292815.ogg"},
	{0xF24A69DA45A2619CULL, "230969805.ogg"},
	{0xF24B2BC8AF990C60ULL, "907517104.ogg"},
	{0xF24B66AF27FE7850ULL, "6127893_m.fxe"},
	{0xF24B8EAF7F030AB4ULL, "807537028.ogg"},
	{0xF24BE524BC8246BFULL, "vxw_woundedcoast_1sv_0_0.vlm"},
	{0xF24BE70C7D4E6E73ULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower2_isabellasreturn.ani"},
	{0xF24C0A73CBB990B3ULL, "llac_ca_caverns1_mines_24423_0.dds"},
	{0xF24C565AB2F9B203ULL, "w_blightlands_131_0.mao"},
	{0xF24C5A5AB2F9B950ULL, "w_blightlands_131_0.mmh"},
	{0xF24C685AB2F9D16AULL, "w_blightlands_131_0.msh"},
	{0xF24CF151C948E9B4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_enter.evt"},
	{0xF24D08472CBF8B75ULL, "genip_cin_cross_bolt.utp"},
	{0xF24D40D6615A61BAULL, "6177857_f.fxe"},
	{0xF24E02F36D1D6ACDULL, "6171895_m.fxe"},
	{0xF24E24E5FD8DD0C8ULL, "vxdr_primevaltomb_4c8_0_1.vlm"},
	{0xF24E89645256B9D5ULL, "vxdr_primevalentrance_2ed_0_0.vlm"},
	{0xF24EAAB3A97C495BULL, "vxht_bartrand_n_6qs_0_0.vlm"},
	{0xF24EC9F369147334ULL, "character_character_p153.psh.fxo.dx11"},
	{0xF24ED22DA3A927D1ULL, "6041129_m.fxe"},
	{0xF24F35484FA0E4D7ULL, "vxca_caverns1_ei9_0_1.vlm"},
	{0xF24FF032436119F7ULL, "vxht_mansion2_1ad_0_0.vlm"},
	{0xF250D40C6F188DD1ULL, "vxw_woundedcoast_xh_0_0.vlm"},
	{0xF250D9BE247DBC3EULL, "vxlt_darktown_42b_0_0.vlm"},
	{0xF250E62B9992CD19ULL, "hlac_w_blightlands_0_0z.dds"},
	{0xF250FCDC055C2B24ULL, "vxlt_darktown_2w1_0_0.vlm"},
	{0xF251046DFC6680D6ULL, "ser221_gamlen6134826.cl"},
	{0xF25137404AD48D35ULL, "vrc321tr_door_shift_2.ncs"},
	{0xF2518FE28E946553ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn.ani"},
	{0xF251980845985E94ULL, "llac_lt_0_lowtown_a3_n_c_-131073_1.dds"},
	{0xF2521A16766F2E28ULL, "vxlt_0_lowtown_a2_hgx_0_0.vlm"},
	{0xF25266817BF5E2A4ULL, "595249561.ogg"},
	{0xF252A742BBADD22CULL, "qun131_taskmaster6019249.cl"},
	{0xF252A8CC8C77D885ULL, "llac_lt_gallowstemplar_2388_6.dds"},
	{0xF25455B452F7158CULL, "gen00flst_varric_actonea.stg"},
	{0xF2549BFBF00C0DA8ULL, "6142809_m.fxe"},
	{0xF25516F6C696912DULL, "lt_0_lowtown_docks_a3_192_0.mmh"},
	{0xF2551AF6C696987EULL, "lt_0_lowtown_docks_a3_192_0.mao"},
	{0xF2551BAC29359CB4ULL, "273793416.ogg"},
	{0xF25528F6C696AFB3ULL, "lt_0_lowtown_docks_a3_192_0.msh"},
	{0xF25558E582FAFC36ULL, "173397321.ogg"},
	{0xF255B2789EF26AB7ULL, "6087806_m.fxe"},
	{0xF255FABE2BBADFECULL, "hlac_ca_caverns2prc_5802_8x.dds"},
	{0xF25602E1BE1C5354ULL, "follower_banter6141881.cl"},
	{0xF256492A9D0D656EULL, "gen00fl_anders6111254.cl"},
	{0xF2566BDA4B5E2B51ULL, "6162510_m.fxe"},
	{0xF2568B13AFC9CA33ULL, "193047656.ogg"},
	{0xF2573B292BDC90DDULL, "6087295_m.fxe"},
	{0xF257533F949FCC9BULL, "vxht_0_hightown_a2_4kg_0_0.vlm"},
	{0xF257E469B0272560ULL, "vxw_blightlands_bi_0_0.vlm"},
	{0xF258A62F9419CB6DULL, "6071917_m.fxe"},
	{0xF258B5080C9787B8ULL, "lgt160_ntb_elf_conflict6163396.cl"},
	{0xF258C09A758F13CBULL, "6043891_m.fxe"},
	{0xF258C7FCC98B3852ULL, "gen_im_trs_gen_lgt_cht_r0.uti"},
	{0xF258FDAA676CBA78ULL, "vxw_sundermount_218_0_0.vlm"},
	{0xF259190D76ED6BC3ULL, "vxht_0_hightown_a3_cbf_0_0.vlm"},
	{0xF2592E2D1D3C8F79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_lean_forward.gad"},
	{0xF25942637732CBF9ULL, "vxw_bonepit_lm_0_0.vlm"},
	{0xF259F7E2A75D809EULL, "character_character_p65.psh.fxo.dx11.assembly"},
	{0xF259F886FCE45FBFULL, "follower_banter6136705.cl"},
	{0xF25A0ADB69EEC16AULL, "911336346.ogg"},
	{0xF25A5B56D724974EULL, "6053280_m.fxe"},
	{0xF25AAC4E1DC45F87ULL, "621809526.ogg"},
	{0xF25AE951C95519EDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\grab_reaction_enter.ani"},
	{0xF25BE3DEECCC7090ULL, "ser121_jethann6108798.cl"},
	{0xF25C14A5F3258D64ULL, "6004071_m.fxe"},
	{0xF25C4847079AE2D9ULL, "6037029_m.fxe"},
	{0xF25C4B821C3C6EE0ULL, "6119293_f.fxe"},
	{0xF25CCFBF271AFA37ULL, "262892125.ogg"},
	{0xF25CFE846651AE01ULL, "553674474.ogg"},
	{0xF25D1C909B7261DBULL, "vxht_bartrand_n_6zr_0_0.vlm"},
	{0xF25DD199B1E6D2D5ULL, "zz_one_debug.cnv"},
	{0xF25DDC514CDA132CULL, "lowtown_fem_prostitute_sta6152365.cl"},
	{0xF25E87E28E9EE18AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn.evt"},
	{0xF25EA2B5002B9948ULL, "vxca_caverns2_78p_0_1.vlm"},
	{0xF25EE08945E82CD6ULL, "6070180_m.fxe"},
	{0xF25F3F2D1D4117C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_lean_forward.ani"},
	{0xF25F45E9610EA651ULL, "vxw_woundedcoast_uo_0_0.vlm"},
	{0xF25F4B9B1673D4BEULL, "vxlt_darktown_49c_0_0.vlm"},
	{0xF25FBE0C7D5ED11DULL, "art\\cinematics\\_crit path\\qunari crisis\\isabellasreturn\\cs_follower2_isabellasreturn.gad"},
	{0xF25FF63A73E861B6ULL, "404913311.ogg"},
	{0xF25FF7179E9F190EULL, "485821374.ogg"},
	{0xF2601651C796237BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromvulnerable.evt"},
	{0xF260177BD34798CFULL, "dae000ip_crowd_mf1_merchant.utp"},
	{0xF2608A651C498177ULL, "w_sundermount_235_0.msh"},
	{0xF2609C651C499FDAULL, "w_sundermount_235_0.mao"},
	{0xF260A0651C49A6E9ULL, "w_sundermount_235_0.mmh"},
	{0xF26105BC3D090C06ULL, "qun211_party_comment6072253.cl"},
	{0xF26196738F3152D3ULL, "41166894.ogg"},
	{0xF261D0E75DB79686ULL, "142825480.ogg"},
	{0xF261EE0E664CE302ULL, "6077268_m.fxe"},
	{0xF262A5B1A866CFE2ULL, "141997324.ogg"},
	{0xF262BFB89B76DD70ULL, "one101dg_smuggler.cnv"},
	{0xF262C7F2681E6003ULL, "vxlt_gallowstemplar_n_2iw_0_0.vlm"},
	{0xF262CA3B27BBF69AULL, "6060321_m.fxe"},
	{0xF2632FA6FDE0DBE8ULL, "dae200_bran.cnv"},
	{0xF263502D3CD8F734ULL, "501111500.ogg"},
	{0xF26350B3C1E3DB81ULL, "6209930_m.fxe"},
	{0xF263BCEE34BCEB43ULL, "llac_w_bonepitdestroyed_a3_131077_0.dds"},
	{0xF2645B9C9264B063ULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.dura.ani"},
	{0xF2647EDEF15F5235ULL, "6116207_m.fxe"},
	{0xF26483E4D0A42987ULL, "one100dg_kelder6055427.cl"},
	{0xF26505171BDC24EDULL, "vxht_0_hightown_a3_c8v_0_0.vlm"},
	{0xF265107D52ACAB4DULL, "gen00fl_isabela6135592.cl"},
	{0xF2653509261E7558ULL, "227586150.ogg"},
	{0xF26566E28EA4C8FDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps_no_headturn.gad"},
	{0xF2657E4C5C450DF1ULL, "gen00fl_carver6104600.cl"},
	{0xF2659D2204D732AFULL, "vxlt_0_lowtown_docks_akk_0_0.vlm"},
	{0xF265A5D20545BE59ULL, "zz_ave_debug6175056.cl"},
	{0xF265DE5D102E6B22ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_boxa_d.dds"},
	{0xF265E1B89B794CC8ULL, "one101dg_smuggler.bnk"},
	{0xF265E607D855C3FAULL, "dre000_varric6035193.cl"},
	{0xF2660695AECC6596ULL, "120070275.ogg"},
	{0xF26608A59C97BFE5ULL, "6165663_m.fxe"},
	{0xF26651A6FDE34B50ULL, "dae200_bran.bnk"},
	{0xF2668E3A3B229B6AULL, "hlac_lt_0_lowtown_docks_n_1_0x.dds"},
	{0xF26708FD0C49AA3BULL, "226056147.ogg"},
	{0xF2670F51C79C35BCULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromvulnerable.gad"},
	{0xF26724C5380948A0ULL, "199777703.ogg"},
	{0xF267620F42CB26F5ULL, "hlac_w_sundermount_-65534_0x.dds"},
	{0xF267CE6263119EA9ULL, "486319431.ogg"},
	{0xF267D499B1EF2C32ULL, "zz_one_debug.ncs"},
	{0xF2682C3CDB1D5AC6ULL, "6147466_m.fxe"},
	{0xF26878AEF1D82120ULL, "w_bonepitmining_a2_166_0.phy"},
	{0xF26881D1C2B8D717ULL, "vxlt_darktown_5b2_0_0.vlm"},
	{0xF2692EDB5669808BULL, "mag291_ntbanter6056736.cl"},
	{0xF2699705B693C259ULL, "6156468_f.fxe"},
	{0xF2699F7F4F99EFD0ULL, "273281719.ogg"},
	{0xF269E30EA7EB2465ULL, "107284247.ogg"},
	{0xF269E873FA29D21FULL, "vxlt_undercityencounter_1rp_0_0.vlm"},
	{0xF269EBE5D0F9A67FULL, "6177845_f.fxe"},
	{0xF26A299E8ED6E9BCULL, "gen_im_arm_cht_lgt_start.uti"},
	{0xF26A8BC9BC2477CAULL, "1013764952.ogg"},
	{0xF26B1FA059A198ACULL, "6041704_m.fxe"},
	{0xF26B8A1C2914C40FULL, "fen110_fenris6096407.cl"},
	{0xF26BE2B2FFF881CFULL, "lt_0_lowtown_docks_n_30_0.anb"},
	{0xF26BEB6D82C3A3CFULL, "vxdr_primevalentrance_22c_0_0.vlm"},
	{0xF26C5BF65B9AE443ULL, "vxlt_gallowstemplar_n_5r5_0_0.vlm"},
	{0xF26C9E45583B407DULL, "gen_im_arm_hlm_hvy_act1_01.uti"},
	{0xF26CB24B129543EFULL, "vxw_encounter2_c6_0_0.vlm"},
	{0xF26D1C921D8422B4ULL, "vxlt_0_lowtown_eom_0_0.vlm"},
	{0xF26D3C1DD3B347D2ULL, "6037035_m.fxe"},
	{0xF26D4CF192CC9F0DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wmda_0t.dds"},
	{0xF26D5D9936F66124ULL, "6106650_m.fxe"},
	{0xF26D681D856799D3ULL, "vxlt_0_lowtown_a2_emt_0_0.vlm"},
	{0xF26E0E51C7A252BAULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\appearfromvulnerable.ani"},
	{0xF26E44CB86DE2A15ULL, "follower_banter6167150.cl"},
	{0xF26E762987B0320BULL, "docks_male_human_worker_sta6175963.cl"},
	{0xF26F3F26DC99B3FFULL, "llac_ht_keep_n_3732_8.dds"},
	{0xF26FF87269B83236ULL, "vxlt_0_lowtown_e0r_0_0.vlm"},
	{0xF270751EC9FCE2A5ULL, "682694499.ogg"},
	{0xF2707F1C93BEE767ULL, "vxw_encounter2_camp_90_0_0.vlm"},
	{0xF2708C8482F2E320ULL, "6125851_m.fxe"},
	{0xF2708C9DDD9EEDDCULL, "vxlt_hangedman_20q_0_0.vlm"},
	{0xF2716BA0F65311EAULL, "6095219_m.fxe"},
	{0xF271739C926F635AULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.dura.evt"},
	{0xF2718A8FDE1A8F7AULL, "678520808.ogg"},
	{0xF271AF7EA5F37A1BULL, "6083305_m.fxe"},
	{0xF272182EBA49D680ULL, "307241948.ogg"},
	{0xF272480525730076ULL, "vxdr_dungeon_1m2_0_0.vlm"},
	{0xF272BEED78ED300AULL, "6076087_m.fxe"},
	{0xF272E621DDDD1649ULL, "ss_raider_m_2.bnk"},
	{0xF2731D6EEE24506DULL, "6067815_m.fxe"},
	{0xF2740EFEF6CD4B2FULL, "vxlt_0_lowtown_docks_apj_0_0.vlm"},
	{0xF274526D45A9D3EDULL, "523421707.ogg"},
	{0xF27503A2B615B4B4ULL, "vxca_cave_1ci_0_0.vlm"},
	{0xF27588EA01092FB8ULL, "gen00fl_aveline6139644.cl"},
	{0xF275E967FE22FDB0ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_eluviancrack_n.dds"},
	{0xF27615052FAAAAEDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_rust_d.dds"},
	{0xF2765527029B5AA9ULL, "hlac_lt_hangedman_-1_3y.dds"},
	{0xF2765FE11C49A7FFULL, "qun231_varnell6099590.cl"},
	{0xF2767C9B6AC31807ULL, "ser270_emeric6113351.cl"},
	{0xF27699D8393A92E4ULL, "6067106_m.fxe"},
	{0xF276BEAD4EA7792CULL, "art\\characters\\playercharacter\\elfmale\\rob_encn.msh"},
	{0xF276C4AD4EA7831EULL, "art\\characters\\playercharacter\\elfmale\\rob_encn.mmh"},
	{0xF276D0AD4EA797B5ULL, "art\\characters\\playercharacter\\elfmale\\rob_encn.mao"},
	{0xF277996060D9C381ULL, "lgt110_gxa_vgk_silver_order6116984.cl"},
	{0xF277C92F1C3DA289ULL, "325730450.ogg"},
	{0xF277C970D67FFB68ULL, "vxlt_gallowstemplar_n_3jd_0_0.vlm"},
	{0xF277EF8BE0C2B609ULL, "6033630_m.fxe"},
	{0xF278729C9275804DULL, "art\\vfx\\combat\\vfx_glb_paralyze_crust_c.dura.gad"},
	{0xF27941561580868AULL, "990530383.ogg"},
	{0xF2798D57D5C00CECULL, "gen00fl_dog6165764.cl"},
	{0xF27A00B36F302943ULL, "art\\characters\\creatures\\desiredemon\\animation\\talent_dance.ani"},
	{0xF27A81E0500BFECDULL, "6044940_m.fxe"},
	{0xF27A9ADA868EFA07ULL, "gen00fl_anders6128305.cl"},
	{0xF27AB026B6C52A8BULL, "vxw_encounter1_ey_0_0.vlm"},
	{0xF27AE934AA1A7B3DULL, "679237238.ogg"},
	{0xF27AF1A9E5077BE1ULL, "34517966.ogg"},
	{0xF27B0248F8C612EAULL, "682550707.ogg"},
	{0xF27B0AFEDA978D2DULL, "dae000_bartrand6031118.cl"},
	{0xF27B184F5F82407AULL, "zz_dae_debug6029860.cl"},
	{0xF27B1F90D91B4C02ULL, "plt_cod_bks_lite_tombs_last.plo"},
	{0xF27B2428048B5C6FULL, "vxw_encounter2_n1_0_0.vlm"},
	{0xF27B7165338AB2C2ULL, "718716123.ogg"},
	{0xF27BBE5831A987B4ULL, "mag131_keran6086239.cl"},
	{0xF27C0B30525B4CEAULL, "6151630_m.fxe"},
	{0xF27C40EB61F1E30AULL, "hlac_lt_0_lowtown_a3_n_c_65534_4x.dds"},
	{0xF27C4DECE97AD8F2ULL, "6054379_m.fxe"},
	{0xF27CC52F6776431DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medf_0t.dds"},
	{0xF27CCB09459D32BFULL, "defaultstate.sta"},
	{0xF27D18EBAF7B88C1ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\lighthit.ani"},
	{0xF27D620E7C3203C2ULL, "follower_banter6173352.cl"},
	{0xF27D8FC59A992F6EULL, "qun111_ketojan6041034.cl"},
	{0xF27E5493BF6A95FDULL, "vxca_caverns1_dnn_0_0.vlm"},
	{0xF27ED5E94688C99DULL, "gen00fl_anders6125447.cl"},
	{0xF27EF0F985B4FFE7ULL, "vxw_encounter2_camp_43_0_0.vlm"},
	{0xF27F02DB035C712EULL, "ts_chantry_01.gda"},
	{0xF27F0AAA8849D3F5ULL, "plt_bet100pt_birthright.plo"},
	{0xF27F2A20F33B4F09ULL, "416141686.ogg"},
	{0xF27F2A6B6BBC9AFCULL, "708947221.ogg"},
	{0xF27F36BB7F7C3CDBULL, "gen00fl_isabela6079562.cl"},
	{0xF27F76456994DA2DULL, "l_mer341ar_caverns.wbd"},
	{0xF27FC9AFBC6F0DE8ULL, "950072281.ogg"},
	{0xF27FD551678FD845ULL, "hlac_lt_0_lowtown_docks_n_-65537_1z.dds"},
	{0xF2805B54FE8D5EA9ULL, "vxca_caverns2_7o8_0_1.vlm"},
	{0xF28099798154A75FULL, "lgt110_orz_harr_carta.bnk"},
	{0xF280DBC108575893ULL, "mag291ip_lever.utp"},
	{0xF280E8072B7011E5ULL, "hlac_dr_0_deeproads_5845_7x.dds"},
	{0xF2810E9A3430F884ULL, "hlac_ht_0_hightown_n_-65538_1y.dds"},
	{0xF2811DDF04AE0D70ULL, "gen00fl_aveline6101764.cl"},
	{0xF281E8D0FFC855EDULL, "vxht_0_hightown_a3_c6t_0_0.vlm"},
	{0xF281FAC0DBA8D251ULL, "360134776.ogg"},
	{0xF282CFB2E5F84BEEULL, "vxlt_hangedman_n_4t_0_0.vlm"},
	{0xF282DADB409F7E39ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isa.msh"},
	{0xF282E4DB409F8F27ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isa.mmh"},
	{0xF282F0BB0BDD420AULL, "zz_mcr_debug6078871.cl"},
	{0xF28313EC4FF39127ULL, "6054752_m.fxe"},
	{0xF2831CA54C43C331ULL, "clf260_harley6088769.cl"},
	{0xF2831F2A32881F1AULL, "6116675_m.fxe"},
	{0xF28324B63EE3FB37ULL, "hlac_lt_undercitydungeon_-1_27y.dds"},
	{0xF2834DC7C099B124ULL, "6071479_m.fxe"},
	{0xF2837B798156AA67ULL, "lgt110_orz_harr_carta.cnv"},
	{0xF283BF1CF1D191F0ULL, "vxca_caverns2_7qr_0_0.vlm"},
	{0xF28423EBAF81BA63ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\lighthit.gad"},
	{0xF284F64D7F58B553ULL, "6147799_m.fxe"},
	{0xF285822254AA152BULL, "vxca_caverns1_e8k_0_0.vlm"},
	{0xF28622B744765BCAULL, "297415235.ogg"},
	{0xF2867F4AC9261E4BULL, "187544701.ogg"},
	{0xF28688E9204C35CDULL, "gen00fl_merrill6106433.cl"},
	{0xF286C68521661CEFULL, "vxlt_0_lowtown_a3_n_c_hco_0_0.vlm"},
	{0xF286E591C2412ED9ULL, "21460316.ogg"},
	{0xF286E93F8BA14368ULL, "349202055.ogg"},
	{0xF287103C04971C4EULL, "vxlt_gallowscourtyard_7fk_0_0.vlm"},
	{0xF287557CFA80C471ULL, "vxht_0_hightown_n_bu2_0_0.vlm"},
	{0xF287E31B2618273DULL, "hightown_market_guard_sta6169607.cl"},
	{0xF28818B36F3C8F3AULL, "art\\characters\\creatures\\desiredemon\\animation\\talent_dance.evt"},
	{0xF288333480E8CEF1ULL, "vxlt_gallowstemplar_n_453_0_0.vlm"},
	{0xF2884F695660D639ULL, "6084760_m.fxe"},
	{0xF2886FFD6CBCB949ULL, "vxht_0_hightown_a2_a2k_0_0.vlm"},
	{0xF28885F5029468DEULL, "6118080_m.fxe"},
	{0xF2889DB85F3184A3ULL, "lt_0_lowtown_docks_a3_175_0.mao"},
	{0xF288A1B85F318BF0ULL, "lt_0_lowtown_docks_a3_175_0.mmh"},
	{0xF288AFB85F31A30AULL, "lt_0_lowtown_docks_a3_175_0.msh"},
	{0xF28901BDAE21E334ULL, "155859474.ogg"},
	{0xF2892203A8BB430BULL, "vxw_encounter1_jx_0_0.vlm"},
	{0xF289387F24938AC9ULL, "6117811_m.fxe"},
	{0xF28997C2855E2BCFULL, "gen00fl_aveline6139069.cl"},
	{0xF289C59139BF7501ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.cess.ani"},
	{0xF28A445FEB859977ULL, "mag331ar_evelinas_hideout.ncs"},
	{0xF28A4859549204AEULL, "hlac_dr_slavers_3217_9x.dds"},
	{0xF28A639F63181AB6ULL, "cd_mer340_cowan.wbd"},
	{0xF28B00EBAF879D48ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\lighthit.evt"},
	{0xF28B0F661FEA961EULL, "vxht_mansion_n_62t_0_0.vlm"},
	{0xF28B16C85B172C7CULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_spear1_low_arishokresponse.ani"},
	{0xF28B43F1AF8D6114ULL, "lt_0_lowtown_docks_a3_42_0.rml"},
	{0xF28B550427E35EA8ULL, "mag101_grace6034386.cl"},
	{0xF28B6A94F811F34EULL, "vxca_caverns2prc_9n6_0_0.vlm"},
	{0xF28B9DAFF3310A21ULL, "and000im_gft_amulet.uti"},
	{0xF28BA1C46C0799A7ULL, "408446401.ogg"},
	{0xF28BE1FE80BCB66EULL, "vxlt_undercityencounter_13k_0_1.vlm"},
	{0xF28C53B43D9F2507ULL, "one140cr_charade6110305.cl"},
	{0xF28C8F258D0537D8ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgertes.phy"},
	{0xF28CCDBFB674074DULL, "gen00fl_anders6125937.cl"},
	{0xF28D5E048C849F40ULL, "vxlt_hangedman_n_1pm_0_0.vlm"},
	{0xF28D95AEF1F76EB5ULL, "w_bonepitmining_a2_166_0.msh"},
	{0xF28D97AEF1F7722BULL, "w_bonepitmining_a2_166_0.mmh"},
	{0xF28D9CFD5A3AAD2BULL, "940371111.ogg"},
	{0xF28DA3AEF1F786C8ULL, "w_bonepitmining_a2_166_0.mao"},
	{0xF28DCC7536E9171FULL, "935555014.ogg"},
	{0xF28E17B36F40F92DULL, "art\\characters\\creatures\\desiredemon\\animation\\talent_dance.gad"},
	{0xF28E253A6060DCCCULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb14a.msh"},
	{0xF28E2B3A6060E6BEULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb14a.mmh"},
	{0xF28E7F65B67F7F1BULL, "6161148_m.fxe"},
	{0xF28EFBFCFAD94507ULL, "vxw_bonepitdestroyed_a3_193_0_0.vlm"},
	{0xF28FB9ED13EC61CFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_action.gad"},
	{0xF28FD09139C3F3A3ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.cess.gad"},
	{0xF290445FEB8A04C6ULL, "mag331ar_evelinas_hideout.lst"},
	{0xF2907E89598E0DEFULL, "vxlt_undercitydungeon_1mp_0_0.vlm"},
	{0xF290D9E3EA66218AULL, "mrl000_marethari6069611.cl"},
	{0xF290F2B8DAB97C2FULL, "vxlt_0_lowtown_docks_azt_0_0.vlm"},
	{0xF2916D5BA5AA0258ULL, "lt_0_lowtown_docks_2_0_lc.mmh"},
	{0xF2916D6B5531B2DBULL, "716558906.ogg"},
	{0xF29185C941E9A9A7ULL, "6051852_m.fxe"},
	{0xF29223C85B1D60BEULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_spear1_low_arishokresponse.gad"},
	{0xF29230F9E3C7AA70ULL, "vxca_caverns2_7vs_0_0.vlm"},
	{0xF292EDB0907F90E0ULL, "684296366.ogg"},
	{0xF29333511D8E3AA1ULL, "vxht_0_hightown_n_b9n_0_0.vlm"},
	{0xF29373932F3E6CC8ULL, "6128131_m.fxe"},
	{0xF293BF56645F215AULL, "344525300.ogg"},
	{0xF293C0CCF629E37DULL, "hlac_lt_0_lowtown_-3_7y.dds"},
	{0xF293DF8757B64BD2ULL, "one200_gold_followers6117110.cl"},
	{0xF29413B8655FED35ULL, "ave200_donnic_patrol_2.stg"},
	{0xF294176D80FF0AE6ULL, "s_proj_cen.mao"},
	{0xF2941F63BEDDDACAULL, "vxht_0_hightown_a3_bs9_0_0.vlm"},
	{0xF29424A837CD3963ULL, "hlac_w_encounter2_camp_5_0x.dds"},
	{0xF29425C52BE09A3EULL, "299317825.ogg"},
	{0xF2942D11AFD9D3F8ULL, "hlac_lt_gallowstemplar_f_2391_0z.dds"},
	{0xF294EB5D2E3FDDA5ULL, "6104203_f.fxe"},
	{0xF2950730A7A5F9DAULL, "vxht_mansion_n_3yl_0_0.vlm"},
	{0xF295B249CCA6CE36ULL, "6135269_m.fxe"},
	{0xF295C759EC76DCF1ULL, "vxht_0_hightown_n_bp1_0_0.vlm"},
	{0xF2968E33AAF14A06ULL, "vxlt_backalley_g_1xk_0_0.vlm"},
	{0xF296AD9139C9D688ULL, "art\\vfx\\combat\\vfx_w_rev_incense_loop_c.cess.evt"},
	{0xF296B57F2ADDFB8AULL, "86206921.ogg"},
	{0xF296CEED13F2A47CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_action.evt"},
	{0xF2970A139F645909ULL, "6133017_m.fxe"},
	{0xF297ADF8D2B1F215ULL, "vfx_chunkyparticle_p4.vsh.fxo"},
	{0xF297B18D317D2DC3ULL, "vxlt_gallowstemplar_n_5c6_0_0.vlm"},
	{0xF297D27AF2754CE2ULL, "vxlt_gallowscourtyard_4wc_0_0.vlm"},
	{0xF2989DD504250057ULL, "gen00fl_fenris6126868.cl"},
	{0xF2989F932690F400ULL, "gen00fl_anders6114484.cl"},
	{0xF298AF1FD723168EULL, "vxht_0_hightown_a4r_0_0.vlm"},
	{0xF298BFB3001F3473ULL, "lt_0_lowtown_docks_n_30_0.rml"},
	{0xF2993A66A7317CB1ULL, "gen00flst_prehaunted.stg"},
	{0xF299ED62A09BCF1BULL, "vxht_0_hightown_b4p_0_0.vlm"},
	{0xF29A05BC04A64CC6ULL, "206391786.ogg"},
	{0xF29A1EAEBBC9614FULL, "vxht_0_hightown_n_a6n_0_0.vlm"},
	{0xF29A6F270A00D107ULL, "gen00fl_isabela6100077.cl"},
	{0xF29A83ACFE3B1EC6ULL, "vxlt_hangedman_n_2eo_0_0.vlm"},
	{0xF29B0D1925E12E1EULL, "6125748_f.fxe"},
	{0xF29B384DA356C6FDULL, "vxca_caverns1_dpl_0_0.vlm"},
	{0xF29B535FEB9425D1ULL, "mag331ar_evelinas_hideout.are"},
	{0xF29BD4AD3E3FA880ULL, "6137033_f.fxe"},
	{0xF29C6ADA59398FAEULL, "vxht_brothel_n_6fw_0_0.vlm"},
	{0xF29D4101CC4DE1A5ULL, "vxht_0_hightown_c73_0_1.vlm"},
	{0xF29D492D50EFF778ULL, "mrl000_marethari6129218.cl"},
	{0xF29E2E4A66A6AF9CULL, "hlac_lt_0_lowtown_docks_n_65535_4y.dds"},
	{0xF29E35B43138AFC7ULL, "6077297_f.fxe"},
	{0xF29ED72357AD3E87ULL, "vxca_cavern1_haunted_in4_0_0.vlm"},
	{0xF29EDB94E982CBF5ULL, "hlac_dr_dungeon_3215_8x.dds"},
	{0xF2A02A09CABA98D7ULL, "vxht_0_hightown_a2_c41_0_0.vlm"},
	{0xF2A0344A0A563654ULL, "561522579.ogg"},
	{0xF2A0A2D6D5BDC2F0ULL, "vxca_caverns2_7kp_0_0.vlm"},
	{0xF2A0B14BA5FF1E07ULL, "6177859_m.fxe"},
	{0xF2A130F35A771F8EULL, "aip_pridedemon_fire.gda"},
	{0xF2A13FFA7C375FE0ULL, "vxw_sundermount_2jq_0_0.vlm"},
	{0xF2A1772C009D21A9ULL, "hlac_ht_mansion2_818_55y.dds"},
	{0xF2A1BE4ED2FEBB50ULL, "6109747_m.fxe"},
	{0xF2A1D58A944888C8ULL, "llac_ca_caverns1_24421_8.dds"},
	{0xF2A284CD316FCFB3ULL, "hlac_lt_0_lowtown_a2_65537_6z.dds"},
	{0xF2A2CFACB1E0533AULL, "vxlt_gallowstemplar_n_483_0_0.vlm"},
	{0xF2A2E911D95E4009ULL, "mag291_vincento6027725.cl"},
	{0xF2A33EF1DD7E51C5ULL, "lt_0_lowtown_53_0_lc.mmh"},
	{0xF2A3790D999C125AULL, "vxht_mansion_n_3vm_0_0.vlm"},
	{0xF2A3D20E1628DD7BULL, "lt_0_lowtown_a2_n_c_13_0.rml"},
	{0xF2A3D6ED13FD3B95ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\knife_stab_v1_action.ani"},
	{0xF2A4362E66925124ULL, "vxw_bonepitdestroyed_a3_mc_0_0.vlm"},
	{0xF2A480820D47F511ULL, "vxht_0_hightown_avk_0_0.vlm"},
	{0xF2A502F2B0C5F839ULL, "ser191_emeric6116974.cl"},
	{0xF2A524DABC61418BULL, "art\\levels\\races\\design\\animation\\floortrapready.ani"},
	{0xF2A56E83282C9E3BULL, "llac_ht_keep_n_4519_11.dds"},
	{0xF2A5DBD06C012E59ULL, "559590471.ogg"},
	{0xF2A62FB629B774F6ULL, "6085936_f.fxe"},
	{0xF2A6B59FA116F195ULL, "vxca_caverns2prc_55d_0_0.vlm"},
	{0xF2A6C49B6668F7DBULL, "380048264.ogg"},
	{0xF2A76D241110DFDFULL, "601234100.ogg"},
	{0xF2A825371BA225B9ULL, "1032934120.ogg"},
	{0xF2A8E2FEDF72C698ULL, "hlac_lt_undercityencounter_-1_8x.dds"},
	{0xF2A964A5D1DF0244ULL, "328807880.ogg"},
	{0xF2AAA2E43EC8E66FULL, "dae011_corff_sta6140574.cl"},
	{0xF2AAC74F276588C2ULL, "6030283_m.fxe"},
	{0xF2AAD90E162F07D3ULL, "lt_0_lowtown_a2_n_c_13_0.phy"},
	{0xF2AAFB477AAF1798ULL, "hlac_lt_0_lowtown_docks_a3_65534_6z.dds"},
	{0xF2AB00C89A831FF5ULL, "1242602.ogg"},
	{0xF2AB5C25DFE40A29ULL, "vxw_bonepitmining_a2_j1_0_0.vlm"},
	{0xF2AB8208CE079DBCULL, "lt_0_lowtown_docks_a3_n_9_0.rml"},
	{0xF2ABC300A41458A8ULL, "vxht_0_hightown_a2_b9p_0_0.vlm"},
	{0xF2ACBDDD6BFE9028ULL, "ht_0_hightown_n_42_0_lc.mmh"},
	{0xF2AD39609E56F7A1ULL, "vxdr_primevalentrance_30n_0_0.vlm"},
	{0xF2AD490049A35707ULL, "vxca_cavern1_haunted_ik7_0_0.vlm"},
	{0xF2AD69A69380CEF1ULL, "vxht_keep_6i9_0_0.vlm"},
	{0xF2AE4EF4164BFD0EULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobc_0n.dds"},
	{0xF2AE882F20A6B238ULL, "fx_grass_b.mao"},
	{0xF2AE9BE6BCB0B157ULL, "vxht_0_hightown_a2_c30_0_0.vlm"},
	{0xF2AECDDB40C4A35CULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_isa.phy"},
	{0xF2AF21B8B7B84D40ULL, "vxlt_0_lowtown_eu7_0_0.vlm"},
	{0xF2AF2CBE8A0A2693ULL, "vxht_mansion_n_53a_0_0.vlm"},
	{0xF2B04CC475036D85ULL, "dae000_meredith6166287.cl"},
	{0xF2B0F20913EA4C76ULL, "65303102.ogg"},
	{0xF2B14407417ECA48ULL, "240142748.ogg"},
	{0xF2B16A8B6F7E1079ULL, "395584238.ogg"},
	{0xF2B2079F288D398DULL, "one110_meeran6173917.cl"},
	{0xF2B209F11B942B7AULL, "796943312.ogg"},
	{0xF2B27E22F53939BBULL, "34178295.ogg"},
	{0xF2B29E3A85837D64ULL, "docks_scene_dockworkers6176200.cl"},
	{0xF2B3C65D786A3961ULL, "237029776.ogg"},
	{0xF2B444DD23E13CA3ULL, "hlac_lt_hangedman_n_-1_25z.dds"},
	{0xF2B4E94EEF2886FEULL, "vxlt_undercityserialkiller_24h_0_0.vlm"},
	{0xF2B5499F7D751DA5ULL, "6068159_m.fxe"},
	{0xF2B580C61ED45843ULL, "vxdr_0_deeproads_d3p_0_0.vlm"},
	{0xF2B5825A085260BCULL, "seb221_flora_harimann6177723.cl"},
	{0xF2B5960C659F1A06ULL, "567868084.ogg"},
	{0xF2B5C1514D27EEFBULL, "6166765_m.fxe"},
	{0xF2B5E7A9197DE2E7ULL, "hlac_lt_darktown_-1_30x.dds"},
	{0xF2B6899F633D9624ULL, "cd_mer340_cowan.bnk"},
	{0xF2B6F92C0BE793AAULL, "dae000ip_crowd_lm3_talk16.utp"},
	{0xF2B761FDF22637DFULL, "vxlt_0_lowtown_n_eox_0_1.vlm"},
	{0xF2B79BAD4EDED4B9ULL, "art\\characters\\playercharacter\\elfmale\\rob_encn.phy"},
	{0xF2B8882C195ED8E7ULL, "6016707_m.fxe"},
	{0xF2B896F4A11EBB2EULL, "llac_lt_0_lowtown_a2_n_-262144_2.dds"},
	{0xF2B8B39B6266E8C0ULL, "vxlt_hangedman_n_1gh_0_0.vlm"},
	{0xF2B8C5602E2E961DULL, "vxca_cave_2cs_0_0.vlm"},
	{0xF2B94F1188200ADEULL, "6219381_m.fxe"},
	{0xF2B9C421DE18DE9BULL, "ss_raider_m_2.wbd"},
	{0xF2BA026FFE812734ULL, "ser121_jethann6104567.cl"},
	{0xF2BA0BDABC737315ULL, "art\\levels\\races\\design\\animation\\floortrapready.gad"},
	{0xF2BA44E31708DFBEULL, "mcr370_meredith6136137.cl"},
	{0xF2BAF745E9C800CDULL, "lt_0_lowtown_docks_155_0.mmh"},
	{0xF2BAFB45E9C8071EULL, "lt_0_lowtown_docks_155_0.mao"},
	{0xF2BB0945E9C81F53ULL, "lt_0_lowtown_docks_155_0.msh"},
	{0xF2BB4F0ED8DEDB9EULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationaoe.evt"},
	{0xF2BB51EF1CC8EFDAULL, "vxlt_hangedman_n_1by_0_0.vlm"},
	{0xF2BB821DA2555C5BULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_splash.mao"},
	{0xF2BB917DBE9090B4ULL, "follower_barks6153856.cl"},
	{0xF2BBA0B3838F559BULL, "lt_gallowscourtyard_n_27_0.rml"},
	{0xF2BBE76AC99F733CULL, "522404782.ogg"},
	{0xF2BBFDAA887D6BD2ULL, "plt_bet100pt_birthright.nss"},
	{0xF2BC3FC8541FEA27ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.anb"},
	{0xF2BC564981837173ULL, "605437547.ogg"},
	{0xF2BC6CED292FBBAFULL, "art\\vfx\\combat\\vfx_w_jug_smash_act_c.mmh"},
	{0xF2BC96DCC0EC5AD7ULL, "701974603.ogg"},
	{0xF2BCA518C8A17178ULL, "vxw_encounter2_9k_0_0.vlm"},
	{0xF2BCBAD55B6705BCULL, "plt_lgt000pt_ultra_4_04.nss"},
	{0xF2BCDC4BCBCD0322ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_006a_0s.dds"},
	{0xF2BD519412CC70CAULL, "vxht_mansion2_2cl_0_0.vlm"},
	{0xF2BD9395A9AE65C0ULL, "vxlt_0_lowtown_ep4_0_0.vlm"},
	{0xF2BF2A4D43E47C7FULL, "hlac_lt_0_lowtown_a3_n_c_-1_6x.dds"},
	{0xF2BF6AC61AA2C9E8ULL, "vxht_chantry_n_1yp_0_0.vlm"},
	{0xF2BF8A67F2B64F1EULL, "347331191.ogg"},
	{0xF2BFD648D7C2C0C2ULL, "vxlt_0_lowtown_docks_n_b7z_0_0.vlm"},
	{0xF2C048A9122775A0ULL, "vxlt_warehouse_cz_0_0.vlm"},
	{0xF2C050B99F94D7E5ULL, "6087479_m.fxe"},
	{0xF2C06E4469CB951BULL, "el_lr_new_day.mmh"},
	{0xF2C0804569CC8257ULL, "l_mer341ar_caverns.bnk"},
	{0xF2C11A0EE5D00531ULL, "qun111_ketojan6040638.cl"},
	{0xF2C1660ED8E36EC9ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationaoe.gad"},
	{0xF2C17C3F20252173ULL, "mer240_jansen6050394.cl"},
	{0xF2C3E8EF55ADBD93ULL, "581086492.ogg"},
	{0xF2C42E9AF2189F89ULL, "mcr310_argument6074495.cl"},
	{0xF2C46BCA5322AFC6ULL, "vxht_bartrand_n_47j_0_0.vlm"},
	{0xF2C511F57BC0B7BDULL, "vxht_mansion2_2a5_0_0.vlm"},
	{0xF2C555B3FC39967CULL, "dae000_madam_lusine6151393.cl"},
	{0xF2C566CE2E07CFB8ULL, "dae000_flemeth6000730.cl"},
	{0xF2C7295BBAF3F939ULL, "382831394.ogg"},
	{0xF2C78CBC28865EC2ULL, "6207562_m.fxe"},
	{0xF2C7E45E72857C8DULL, "hlac_lt_undercitydungeon_-1_26x.dds"},
	{0xF2C80C6AD8BB6533ULL, "docks_male_elven_worker_sta6175791.cl"},
	{0xF2C8370ED8E93D17ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\devastationaoe.ani"},
	{0xF2C9038774554699ULL, "vxlt_0_lowtown_a2_n_h2e_0_0.vlm"},
	{0xF2C9E4F2C856470BULL, "6120749_m.fxe"},
	{0xF2CA2B119AD4979DULL, "hlac_ht_keep_4519_3z.dds"},
	{0xF2CAB29A5C31CC1EULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_ex.evt"},
	{0xF2CAE4CCD1E3FCC2ULL, "vxca_caverns1_mines_ckh_0_0.vlm"},
	{0xF2CAF609D9E0E2CEULL, "vxw_bonepitdestroyed_a3_1wh_0_0.vlm"},
	{0xF2CB0FBD4287977BULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfr.evt"},
	{0xF2CB2978543388B2ULL, "fen110cr_hunter_1.utc"},
	{0xF2CB5E3666781D88ULL, "llac_lt_0_lowtown_n_196607_1.dds"},
	{0xF2CB690C30498DA3ULL, "llac_w_encounter2_camp_65534_0.dds"},
	{0xF2CBC37104C2894AULL, "vxht_mansion2_2lm_0_0.vlm"},
	{0xF2CBD75A5D91FB09ULL, "llac_lt_0_lowtown_a2_-1_0.dds"},
	{0xF2CBF045E9D6678AULL, "lt_0_lowtown_docks_155_0.phy"},
	{0xF2CC10786DF65793ULL, "vxht_mansion_n_59c_0_0.vlm"},
	{0xF2CC32A435C2175FULL, "vxht_mansion2_3gp_0_0.vlm"},
	{0xF2CC992F20BFCCBAULL, "fx_grass_b.dds"},
	{0xF2CD40085D9C2C0FULL, "6136117_m.fxe"},
	{0xF2CD83D55EE74851ULL, "6162701_m.fxe"},
	{0xF2CDCABAC92DBE8CULL, "6055613_m.fxe"},
	{0xF2CE18F0611CF205ULL, "vxw_encounter2_3t_0_0.vlm"},
	{0xF2CE2C14845FDAB2ULL, "267858219.ogg"},
	{0xF2CE53919FC3F459ULL, "6103995_m.fxe"},
	{0xF2CE83DC0F974A6CULL, "vxw_woundedcoast_17u_0_0.vlm"},
	{0xF2CEF1262C86C010ULL, "757005079.ogg"},
	{0xF2CF023A60983859ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb14a.phy"},
	{0xF2CF174C7C7D33D0ULL, "621247712.ogg"},
	{0xF2CF8ECDB44F3EF1ULL, "vxht_0_hightown_n_bl5_0_0.vlm"},
	{0xF2D0235E1747DBF4ULL, "dae000_merchant_dm_mob6174309.cl"},
	{0xF2D04DBCBD5ECDBCULL, "vxca_caverns1_mines_k2u_0_0.vlm"},
	{0xF2D051723AFC6558ULL, "hlac_lt_gallowsprison_n_-1_5x.dds"},
	{0xF2D090EC0D02E07DULL, "vxw_bonepit_2aj_0_0.vlm"},
	{0xF2D0C99A5C365F49ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_ex.gad"},
	{0xF2D10DE66124FA03ULL, "vxlt_0_lowtown_n_ewu_0_0.vlm"},
	{0xF2D138693B287B59ULL, "6037589_f.fxe"},
	{0xF2D16B846DEECBA9ULL, "6027345_m.fxe"},
	{0xF2D1A840130AF3FBULL, "538075542.ogg"},
	{0xF2D1D78FB88F5986ULL, "gen00fl_isabela6099347.cl"},
	{0xF2D208BD428DA9BCULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfr.gad"},
	{0xF2D28A51D217108FULL, "vxlt_0_lowtown_a3_n_c_eql_0_0.vlm"},
	{0xF2D28D5FCB1C72DEULL, "w_sundermount_28_0.rml"},
	{0xF2D2A565B6429BBEULL, "6099329_m.fxe"},
	{0xF2D2B098E79D8101ULL, "682935419.ogg"},
	{0xF2D310A519E85E30ULL, "vxdr_primevalentrance_27k_0_0.vlm"},
	{0xF2D313A0237A5C49ULL, "vxdr_primevaltomb_w2_0_0.vlm"},
	{0xF2D33229FE95A14DULL, "6105607_m.fxe"},
	{0xF2D33DE57BDEC61CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wmda_0n.dds"},
	{0xF2D37CA406ACEF07ULL, "969268777.ogg"},
	{0xF2D3B0D037EA51C7ULL, "vxht_0_hightown_a3_cbz_0_0.vlm"},
	{0xF2D3DC7895D71321ULL, "vxlt_0_lowtown_er3_0_0.vlm"},
	{0xF2D3F4EB76F3EFA2ULL, "hlac_ht_0_hightown_a3_65538_0y.dds"},
	{0xF2D43ECDD7D80D7BULL, "vxlt_undercityserialkiller_186_0_0.vlm"},
	{0xF2D4443F6D34C31CULL, "vxca_caverns1_mines_ihy_0_0.vlm"},
	{0xF2D4514039644339ULL, "253881134.ogg"},
	{0xF2D47E33D82A6DFCULL, "vxht_keep_62y_0_0.vlm"},
	{0xF2D4D3EE6FD4D305ULL, "ht_0_hightown_a2_45_0_ll.mmh"},
	{0xF2D6355C1CF4B369ULL, "vxlt_0_lowtown_docks_n_b4t_0_0.vlm"},
	{0xF2D656BF1345BFDCULL, "6114459_m.fxe"},
	{0xF2D65D92A887BA7EULL, "blood_conj_04_l3.mao"},
	{0xF2D74ECD866E6D78ULL, "6205687_m.fxe"},
	{0xF2D75445CEDEBFB3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small_2.ani"},
	{0xF2D765A0B9F6A7D6ULL, "86244260.ogg"},
	{0xF2D78308CE2CDA0CULL, "lt_0_lowtown_docks_a3_n_9_0.anb"},
	{0xF2D79A9A5C3C2D97ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_ex.ani"},
	{0xF2D7F971BEE895C1ULL, "299553994.ogg"},
	{0xF2D8092B8528D309ULL, "94227742.ogg"},
	{0xF2D833A2B05866E5ULL, "203259916.ogg"},
	{0xF2D8734726B194C0ULL, "hlac_ca_caverns2prc_5802_57z.dds"},
	{0xF2D8762F44BD31E4ULL, "w_blightlands_29_0.msh"},
	{0xF2D884E4A8BA55BDULL, "vxlt_shop_5z_0_0.vlm"},
	{0xF2D8882F44BD504DULL, "w_blightlands_29_0.mao"},
	{0xF2D88C2F44BD5686ULL, "w_blightlands_29_0.mmh"},
	{0xF2D907BD4293C6BAULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfr.ani"},
	{0xF2D90ACABD8A2B08ULL, "507886985.ogg"},
	{0xF2D9109D0B476F6FULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_strip_d.dds"},
	{0xF2D956A9C3DA1542ULL, "vxca_caverns1_mines_cli_0_0.vlm"},
	{0xF2D967E6CBD6FB4EULL, "vxw_bonepitdestroyed_a3_1xi_0_0.vlm"},
	{0xF2D97F6B0791773FULL, "43219231.ogg"},
	{0xF2D9E58E53BE09B4ULL, "gen_im_npc_elf_arm_boo_dalb.uti"},
	{0xF2DA32C3F8F167EDULL, "vxht_keep_4ac_0_0.vlm"},
	{0xF2DA3BCEF2459E17ULL, "vxca_caverns1_mines_jgt_0_0.vlm"},
	{0xF2DA481E2F616114ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium2_c.impa.ani"},
	{0xF2DA559CC7867E9BULL, "6088072_m.fxe"},
	{0xF2DA685D98D1F3A8ULL, "vxdr_primevaltomb_3hw_0_0.vlm"},
	{0xF2DA6CF632504E2EULL, "vxw_sundermount_27x_0_0.vlm"},
	{0xF2DAFE6EE91396F8ULL, "vxw_sundermount_2b1_0_0.vlm"},
	{0xF2DC9EAB36E75A3EULL, "vxlt_warehouse_n_pd_0_0.vlm"},
	{0xF2DCC56E91BA0CD9ULL, "6104895_m.fxe"},
	{0xF2DCC95CD218BA0CULL, "hlac_lt_0_lowtown_a2_-65537_6z.dds"},
	{0xF2DCF5B9018D62ECULL, "vxw_woundedcoast_12j_0_0.vlm"},
	{0xF2DDD62868C583C1ULL, "genip_control_rod.utp"},
	{0xF2DDFCE5C80E5C7EULL, "vxlt_0_lowtown_a2_n_fe9_0_0.vlm"},
	{0xF2DE47F5D5205711ULL, "vxht_0_hightown_ajo_0_0.vlm"},
	{0xF2DEE59E34B95E73ULL, "llac_w_woundedcoast_262145_0.dds"},
	{0xF2DF1CBFE2FB60DBULL, "508932783.ogg"},
	{0xF2DF4AC60074BC4FULL, "ave200_donnic6110733.cl"},
	{0xF2DF8CF99B481463ULL, "hlac_lt_0_lowtown_docks_n_-65537_2y.dds"},
	{0xF2DFC09A1DC9918FULL, "hlac_lt_0_lowtown_n_65533_2y.dds"},
	{0xF2E0247BE7D653C4ULL, "vxht_0_hightown_a3_b9a_0_0.vlm"},
	{0xF2E0948ECC8A3C6FULL, "vxw_blightlands_75_0_0.vlm"},
	{0xF2E0F930A7FC8564ULL, "vxlt_undercityserialkiller_1io_0_1.vlm"},
	{0xF2E1451E2F677B76ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium2_c.impa.gad"},
	{0xF2E16244862EA27FULL, "582835501.ogg"},
	{0xF2E16ECF6D49888AULL, "vxlt_gallowscourtyard_4v8_0_0.vlm"},
	{0xF2E1BA9096320F62ULL, "vxlt_0_lowtown_a2_feb_0_0.vlm"},
	{0xF2E1D27608EAE9F5ULL, "mag160_feynriel6020398.cl"},
	{0xF2E1E986CB276695ULL, "vxlt_0_lowtown_h6r_0_0.vlm"},
	{0xF2E257D380729208ULL, "568572078.ogg"},
	{0xF2E2C673E5911272ULL, "vxlt_0_lowtown_n_eo9_0_1.vlm"},
	{0xF2E2EDD6A2576F84ULL, "lt_0_lowtown_docks_a3_142_0.msh"},
	{0xF2E2FFD6A2578EEDULL, "lt_0_lowtown_docks_a3_142_0.mao"},
	{0xF2E303D6A2579526ULL, "lt_0_lowtown_docks_a3_142_0.mmh"},
	{0xF2E33E5F202C735DULL, "962043416.ogg"},
	{0xF2E390E31323E652ULL, "6060493_m.fxe"},
	{0xF2E4A8E628094AB1ULL, "vxw_bonepit_lu_0_0.vlm"},
	{0xF2E4E0B8AD1038BEULL, "lt_0_lowtown_n_18_0.rml"},
	{0xF2E4E9D8A864843AULL, "vxht_keep_6j7_0_0.vlm"},
	{0xF2E531F322D30592ULL, "start.dds"},
	{0xF2E53CC85442EA28ULL, "art\\vfx\\environment\\vfx_pois_gas_area_sml.mmh"},
	{0xF2E54C45CEEAEFEAULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small_2.evt"},
	{0xF2E555B6443D13B8ULL, "140764172.ogg"},
	{0xF2E560576A99D409ULL, "llac_lt_0_lowtown_docks_131071_0.dds"},
	{0xF2E56DED2952C2ECULL, "art\\vfx\\combat\\vfx_w_jug_smash_act_c.anb"},
	{0xF2E5AB1847B2C158ULL, "lt_gallowscourtyard_58_0.rml"},
	{0xF2E5C0AA786485F8ULL, "6208187_m.fxe"},
	{0xF2E5C6AE28EBBF16ULL, "312249773.ogg"},
	{0xF2E5D93ADE1222C9ULL, "6101764_m.fxe"},
	{0xF2E61E17262F1CB9ULL, "vxlt_warehouse_5y_0_0.vlm"},
	{0xF2E6D28DFF71FE95ULL, "vxht_chantry_n_2yj_0_0.vlm"},
	{0xF2E6E195410BA55BULL, "hlac_lt_darktown_-1_87z.dds"},
	{0xF2E6EEAEB4E2B41CULL, "vxlt_0_lowtown_a2_n_e2e_0_0.vlm"},
	{0xF2E788680ADBBF31ULL, "gen00fl_aveline6137560.cl"},
	{0xF2E7AA5DEDC3C215ULL, "vxca_cavern1_haunted_ehx_0_0.vlm"},
	{0xF2E8527800C8D01AULL, "vxht_0_hightown_a3_bpt_0_0.vlm"},
	{0xF2E8601E2F6DC7D5ULL, "art\\vfx\\combat\\vfx_w_wns_front_medium2_c.impa.evt"},
	{0xF2E87C258D53766DULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgertes.msh"},
	{0xF2E88A258D538DA0ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgertes.mao"},
	{0xF2E88E258D5394F3ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_02hgertes.mmh"},
	{0xF2E8BC9457196B87ULL, "vxlt_0_lowtown_a2_bys_0_0.vlm"},
	{0xF2E8C4D750C4AEA5ULL, "character_character_p143.psh.fxo.dx11"},
	{0xF2E8DC349264B53CULL, "6135627_m.fxe"},
	{0xF2E965EE30F24881ULL, "926318563.ogg"},
	{0xF2E98C2F44CBCA4DULL, "w_blightlands_29_0.rml"},
	{0xF2E9BB1EA2B8903EULL, "6031812_f.fxe"},
	{0xF2E9F463EB7A4303ULL, "6099665_m.fxe"},
	{0xF2EA545F76B217BFULL, "zz_mcr_debug6080862.cl"},
	{0xF2EA675B57C7775BULL, "6054051_m.fxe"},
	{0xF2EAFCAA45092305ULL, "vxw_encounter2_9r_0_0.vlm"},
	{0xF2EAFF1BD26BBFE5ULL, "vxlt_0_lowtown_exm_0_0.vlm"},
	{0xF2EB108828DD72BEULL, "vxlt_warehouse_n_we_0_0.vlm"},
	{0xF2EB64B5183C1FB4ULL, "hlac_dr_0_deeproads_5842_12z.dds"},
	{0xF2EB9E3FE809BF20ULL, "vxlt_warehouse_ry_0_0.vlm"},
	{0xF2EBBBB4E705B50DULL, "gen00fl_aveline6122131.cl"},
	{0xF2EBCA146DDD299EULL, "6075976_m.fxe"},
	{0xF2EC2B45CEF0D65DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_weightshift_small_2.gad"},
	{0xF2ED1D0070F3C045ULL, "6071243_m.fxe"},
	{0xF2ED2D802EE46D31ULL, "vxca_caverns2prc_9q8_0_0.vlm"},
	{0xF2EDC6E53E138A48ULL, "lt_gallowscourtyard_47_0.rml"},
	{0xF2EDD81A84A9F132ULL, "gen00fl_anders6214955.cl"},
	{0xF2EDF1A045112B03ULL, "vxlt_0_lowtown_n_emw_0_0.vlm"},
	{0xF2EDFF6C305FD214ULL, "235103426.ogg"},
	{0xF2EE00D0B4CC9041ULL, "mer240_jansen6050424.cl"},
	{0xF2EEA85EF57F3FDBULL, "w_blightlands_165_0.mmh"},
	{0xF2EEB45EF57F5478ULL, "w_blightlands_165_0.mao"},
	{0xF2EEC153EE836FEBULL, "81435466.ogg"},
	{0xF2EEC22284F6D1B0ULL, "6154312_m.fxe"},
	{0xF2EEC65EF57F72C5ULL, "w_blightlands_165_0.msh"},
	{0xF2EEDD34DDE5F89CULL, "6116642_m.fxe"},
	{0xF2EFB05EBF1A91BEULL, "279617855.ogg"},
	{0xF2EFBDE64C20D2AAULL, "224024871.ogg"},
	{0xF2EFDDB8B22E23D5ULL, "vxdr_slavers_1bz_0_0.vlm"},
	{0xF2F0055C34876B33ULL, "vxlt_0_lowtown_docks_bfc_0_0.vlm"},
	{0xF2F0208BF47FB799ULL, "vxca_caverns1_mines_dpm_0_0.vlm"},
	{0xF2F05D447766AD88ULL, "30197530.ogg"},
	{0xF2F0678C51A3013DULL, "vxht_mansion2_2p2_0_0.vlm"},
	{0xF2F0BC095631B2BBULL, "tm_war_d05.tnt"},
	{0xF2F1011E90D1E94CULL, "gen00fl_anders6125888.cl"},
	{0xF2F107F279C45FAFULL, "6147179_m.fxe"},
	{0xF2F1739B067D8BAEULL, "qcr220_meredith6054756.cl"},
	{0xF2F1BF633CAF29F6ULL, "6018670_m.fxe"},
	{0xF2F29124AA1D07E7ULL, "6105502_m.fxe"},
	{0xF2F333450E58CAF9ULL, "vxlt_0_lowtown_a3_n_c_ei2_0_0.vlm"},
	{0xF2F3F2C8A1DADD97ULL, "6096736_m.fxe"},
	{0xF2F413CFEA96CBC3ULL, "6115558_m.fxe"},
	{0xF2F52B83C369B984ULL, "169894025.ogg"},
	{0xF2F5446AF1681715ULL, "vxht_chantry_n_2ti_0_0.vlm"},
	{0xF2F5608BA6D8CC9CULL, "vxlt_0_lowtown_a2_n_e7z_0_0.vlm"},
	{0xF2F5BEFE6414C687ULL, "6095345_f.fxe"},
	{0xF2F5FA74FF90BC84ULL, "hlac_lt_gallowstemplar_n_2382_1y.dds"},
	{0xF2F61F44BB69617DULL, "llac_ht_chantry_n_2534_38.dds"},
	{0xF2F63E988AE6CB27ULL, "vxlt_gallowstemplar_f_4sk_0_0.vlm"},
	{0xF2F64BA0AFC32C4EULL, "vxw_bonepitdestroyed_a3_1fk_0_0.vlm"},
	{0xF2F6C454F2BEE89AULL, "vxht_0_hightown_a3_bou_0_0.vlm"},
	{0xF2F7037B822D54CAULL, "gen_im_trs_bst_claws_04.uti"},
	{0xF2F77DCF92360652ULL, "96469866.ogg"},
	{0xF2F781127F60E87CULL, "vxlt_gallowstemplar_n_4st_0_0.vlm"},
	{0xF2F7D6A58D0AFB32ULL, "vxdr_0_deeproads_cvu_0_0.vlm"},
	{0xF2F8196432EFAA06ULL, "852964576.ogg"},
	{0xF2F862EC13DE5B55ULL, "vxht_keep_n_5iy_0_0.vlm"},
	{0xF2F879F3BEC44D6FULL, "ofthingsnotlost.dds"},
	{0xF2F87A04E50420B0ULL, "art\\levels\\ztest\\chad\\w_gsw_sotf.mao"},
	{0xF2F8CD6E5E7FBA70ULL, "vxca_cavern1_haunted_ea2_0_0.vlm"},
	{0xF2F93B80140B6CBFULL, "vxw_woundedcoast_17z_0_0.vlm"},
	{0xF2F95B641973515BULL, "vxht_mansion2_3gm_0_1.vlm"},
	{0xF2F982651AD38B3EULL, "vxlt_warehouse_n_zf_0_0.vlm"},
	{0xF2F9D972E57993ECULL, "vxw_woundedcoast_18h_0_0.vlm"},
	{0xF2FA101CD9FFD7A0ULL, "vxlt_warehouse_wf_0_0.vlm"},
	{0xF2FAB7F9FCE03570ULL, "263054943.ogg"},
	{0xF2FAEF9A3DD53F71ULL, "dae000_gamlen6147222.cl"},
	{0xF2FB576C24F8007AULL, "hlac_lt_warehouse_0_9x.dds"},
	{0xF2FBAECECA38C75FULL, "lt_0_lowtown_docks_n_148_0.phy"},
	{0xF2FC010F01056F5BULL, "vxht_bartrand_n_65o_0_0.vlm"},
	{0xF2FC0C6C3311A5F8ULL, "lgt220cr_mag_apost_apost3.utc"},
	{0xF2FCF1CC69D793A5ULL, "hlac_lt_0_lowtown_a2_n_c_131072_3z.dds"},
	{0xF2FCF2BC7CFE005FULL, "hlac_ht_0_hightown_a3_131073_0x.dds"},
	{0xF2FD5B9559B78613ULL, "249802208.ogg"},
	{0xF2FD7D340B19EA71ULL, "llac_lt_undercityserialkiller_-1_29.dds"},
	{0xF2FD99CA76A0C32DULL, "254341248.ogg"},
	{0xF2FD9CE363D7FD29ULL, "one100ip_at_ruin_ext.utp"},
	{0xF2FE334A9852F319ULL, "vxht_mansion2_219_0_0.vlm"},
	{0xF2FEBC2EE165C616ULL, "t3_wpn_yew.tnt"},
	{0xF2FEEBE45DBC546FULL, "173555751.ogg"},
	{0xF2FF815FCB434C56ULL, "w_sundermount_28_0.mao"},
	{0xF2FF8D5FCB4360F5ULL, "w_sundermount_28_0.mmh"},
	{0xF2FF8F5FCB43646BULL, "w_sundermount_28_0.msh"},
	{0xF2FFB6B4677D455CULL, "6121756_m.fxe"},
	{0xF2FFBE6278B87440ULL, "mag291_ntbanter6056838.cl"},
	{0xF3001F1A817E5C10ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockds.msh"},
	{0xF3002D1A817E742AULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockds.mmh"},
	{0xF300311A817E7B61ULL, "art\\levels\\races\\natural\\kirkcave\\props\\prp_caverockds.mao"},
	{0xF3003B938F3271D3ULL, "vxht_keep_n_61d_0_0.vlm"},
	{0xF3009B98179E2509ULL, "460325591.ogg"},
	{0xF30140D60E350246ULL, "6218975_m.fxe"},
	{0xF3018DCADCC1197AULL, "qun100_mage_party_comment6019122.cl"},
	{0xF301CA3654EF1B1AULL, "6129982_f.fxe"},
	{0xF301DC29F93501B8ULL, "qun110_petrice6018655.cl"},
	{0xF302051D6E8345ADULL, "ht_0_hightown_a3_mmap.dds"},
	{0xF30232B613E70579ULL, "vxht_keep_n_3ig_0_0.vlm"},
	{0xF30255F2B4673E59ULL, "6149885_m.fxe"},
	{0xF302AC93DD9991DAULL, "6113009_m.fxe"},
	{0xF302F3A019527D6AULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masb_0d.dds"},
	{0xF3030A9B373917B2ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ferelden_jimbrn_sa.dds"},
	{0xF303201FF495BAA8ULL, "vxht_brothel_n_4fs_0_0.vlm"},
	{0xF30330A0AB0B540EULL, "586736342.ogg"},
	{0xF303DA9867D9DF4DULL, "6206965_f.fxe"},
	{0xF303E5D21DF00ED2ULL, "6116767_m.fxe"},
	{0xF3040CB774A74922ULL, "framebuffereffects_hslmatrix_p0.psh.fxo"},
	{0xF304BD7DA1B944CEULL, "vxw_bonepitdestroyed_a3_1kl_0_0.vlm"},
	{0xF30502C3B486A191ULL, "art\\levels\\atmosphere\\sky\\_textures\\cld_puffylow_n.dds"},
	{0xF3054126FA039D53ULL, "6140652_f.fxe"},
	{0xF30581696105D378ULL, "446853491.ogg"},
	{0xF305ED5F719129DAULL, "pro000_meeran6048813.cl"},
	{0xF306D81DCF667403ULL, "6093465_m.fxe"},
	{0xF3070780C71E6567ULL, "weapon_vfx_p0.psh.fxo"},
	{0xF307414E9E478C59ULL, "zz_isa_debug6133213.cl"},
	{0xF307730C92A6D3B0ULL, "6081322_m.fxe"},
	{0xF307AD5D0601853FULL, "vxw_woundedcoast_12e_0_0.vlm"},
	{0xF307CC542E848957ULL, "plt_cod_ltr_forcarver.plo"},
	{0xF30820234FB5E58BULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wo_basic_2.mao"},
	{0xF3082C17E95427D2ULL, "52215052.ogg"},
	{0xF30881F9CBF5F020ULL, "vxlt_warehouse_xg_0_0.vlm"},
	{0xF30903D23B634122ULL, "442125727.ogg"},
	{0xF30949D6616AF48BULL, "289086972.ogg"},
	{0xF30968F0A3FC9D1DULL, "art\\characters\\playercharacter\\humanmale\\hlf_masd.phy"},
	{0xF3097896206072B2ULL, "vxht_0_hightown_n_7ix_0_0.vlm"},
	{0xF309B4CDC8657193ULL, "vxw_encounter2_camp_ry_0_0.vlm"},
	{0xF309BC4950871052ULL, "vxht_keep_n_61a_0_0.vlm"},
	{0xF309CB4D4027659CULL, "6126663_m.fxe"},
	{0xF30A68E9D428EE08ULL, "626276535.ogg"},
	{0xF30B4DF44B939F33ULL, "vxw_bonepitdestroyed_a3_1zv_0_0.vlm"},
	{0xF30B6FB7470EE31DULL, "gen00fl_aveline6139629.cl"},
	{0xF30B8373F46CF584ULL, "6065889_m.fxe"},
	{0xF30BB413B370F422ULL, "271721615.ogg"},
	{0xF30BB4C2489C7D66ULL, "dae000_bartrand6206915.cl"},
	{0xF30BC8A9AE6FCCEBULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_ctn180l.ani"},
	{0xF30C8A5FCB4DE582ULL, "w_sundermount_28_0.anb"},
	{0xF30CAE9EE29B8024ULL, "6133033_f.fxe"},
	{0xF30CAFCECA473C64ULL, "lt_0_lowtown_docks_n_148_0.mmh"},
	{0xF30CB3CECA4742B7ULL, "lt_0_lowtown_docks_n_148_0.mao"},
	{0xF30CB73DB6718033ULL, "vxca_cavern1_haunted_i6p_0_0.vlm"},
	{0xF30CC5CECA476106ULL, "lt_0_lowtown_docks_n_148_0.msh"},
	{0xF30D6F199BB91944ULL, "llac_dr_primevaltomb_664_0.dds"},
	{0xF30DD20E16837948ULL, "lt_0_lowtown_a2_n_c_13_0.mmh"},
	{0xF30DD3156C622FB7ULL, "114978782.ogg"},
	{0xF30DDE0E16838DEBULL, "lt_0_lowtown_a2_n_c_13_0.mao"},
	{0xF30DF00E1683AC72ULL, "lt_0_lowtown_a2_n_c_13_0.msh"},
	{0xF30E6F9BF2D3E42CULL, "gen00fl_dog6141142.cl"},
	{0xF30E94DAFAEA2DEFULL, "614894111.ogg"},
	{0xF30EECE328FDB527ULL, "hlac_lt_backalley_n_65536_5z.dds"},
	{0xF30FB9708D39D1D1ULL, "vxw_bonepit_11m_0_0.vlm"},
	{0xF30FBC2C2FDDE4ABULL, "495508396.ogg"},
	{0xF30FC9E64EE982E5ULL, "vxca_cave_1bo_0_0.vlm"},
	{0xF30FE2B0F61D8A54ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_aoe_p.anb"},
	{0xF30FE9491FD4D0CDULL, "6138720_m.fxe"},
	{0xF3101BC45BE5F812ULL, "vxca_caverns2_8fc_0_1.vlm"},
	{0xF3108D3A4367D027ULL, "hlac_w_sundermount_196609_0y.dds"},
	{0xF310C478660B749BULL, "6050200_m.fxe"},
	{0xF3110FA9C0F50B33ULL, "671390113.ogg"},
	{0xF3111E970E241248ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\varric\\dm_cth_vara_0d.dds"},
	{0xF3112617282910FEULL, "follower_banter6160671.cl"},
	{0xF311612F1101380DULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cartempty.mmh"},
	{0xF311652F11013F5EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cartempty.mao"},
	{0xF311732F11015693ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cartempty.msh"},
	{0xF3118C1847D7C710ULL, "lt_gallowscourtyard_58_0.anb"},
	{0xF311A639C54C9119ULL, "vxlt_shop_ev_0_0.vlm"},
	{0xF312B4D23CD7E761ULL, "gen_im_npc_dwf_arm_cht_meda.uti"},
	{0xF312B7426418158FULL, "vxlt_0_lowtown_docks_b01_0_0.vlm"},
	{0xF313532DE395FE98ULL, "ave100_donnic6105993.cl"},
	{0xF313A80ED6AB199AULL, "vxht_0_hightown_a3_baw_0_0.vlm"},
	{0xF3140E45564A22CDULL, "dae000_bartrand6027571.cl"},
	{0xF314CBD55BB198E7ULL, "plt_lgt000pt_ultra_4_04.plo"},
	{0xF316BE8144E3E938ULL, "zz_vault_debug6203318.cl"},
	{0xF316CEC2A2F6D542ULL, "6139382_m.fxe"},
	{0xF317637E58BCB6C2ULL, "and100_anders6095945.cl"},
	{0xF3176AC261F2EAA8ULL, "tbe100_xenon6189676.cl"},
	{0xF317C96D985DE48CULL, "sse_ss_varian_ilithis.gda"},
	{0xF31826AABA5B8A13ULL, "vxw_encounter2_camp_mx_0_0.vlm"},
	{0xF31838F3FAB0CA25ULL, "6040852_m.fxe"},
	{0xF31861EF5BB7D958ULL, "316218867.ogg"},
	{0xF3186F4D6077DF41ULL, "vxlt_0_lowtown_docks_a3_n_c1q_0_0.vlm"},
	{0xF318CF3A1050AEECULL, "dre000_dougal6131148.cl"},
	{0xF3193449EDC1804DULL, "646844139.ogg"},
	{0xF319531E570CD238ULL, "hlac_lt_0_lowtown_docks_0_6z.dds"},
	{0xF319C41A1B8E2C86ULL, "hf_ave1_dbandit_leader.mor"},
	{0xF319F4F0ECFC3335ULL, "zz_followers_banter6157744.cl"},
	{0xF31A00031D4175E0ULL, "237722144.ogg"},
	{0xF31A151DD81910AFULL, "936601181.ogg"},
	{0xF31A69F0A40B1222ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masd.mmh"},
	{0xF31A6BF0A40B15D8ULL, "art\\characters\\playercharacter\\humanmale\\hlf_masd.msh"},
	{0xF31AEFCDB3955FA4ULL, "406793487.ogg"},
	{0xF31B0E21AA8DE4D1ULL, "vxw_woundedcoast_6z_0_0.vlm"},
	{0xF31B1384DC251FACULL, "llac_lt_0_lowtown_docks_a3_n_-196611_1.dds"},
	{0xF31B258A0AF56463ULL, "966370206.ogg"},
	{0xF31BDB542E954BCCULL, "plt_cod_ltr_forcarver.nss"},
	{0xF31C17ECF2C35D02ULL, "vxht_keep_n_42n_0_0.vlm"},
	{0xF31C5A4B8C140B54ULL, "524820737.ogg"},
	{0xF31C9391440C22D2ULL, "159695197.ogg"},
	{0xF31CAAC8A88BA1CBULL, "dae000_gen_turn_in_26178487.cl"},
	{0xF31CC10EF598683DULL, "mcr370_samson6087534.cl"},
	{0xF31CC1B0F627DC17ULL, "art\\vfx\\combat\\vfx_r_gn1_sgrenade_aoe_p.mmh"},
	{0xF31D5DA6AE047A6FULL, "hlac_ca_caverns1_mines_24424_27x.dds"},
	{0xF31DB1834C8CE457ULL, "art\\characters\\weapons\\axes\\w_axe_ax105.phy"},
	{0xF31DDE7C908536B4ULL, "nat380_sibling6067998.cl"},
	{0xF31E437FCDA79ACAULL, "llac_lt_0_lowtown_docks_n_-3_5.dds"},
	{0xF31EEB7C7B40E166ULL, "18105119.ogg"},
	{0xF31F1B4BE04EBDDBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_enter.evt"},
	{0xF31F8B49404E7732ULL, "vxht_chantry_2du_0_0.vlm"},
	{0xF31F936C857F8C9AULL, "dae100_viscount6103302.cl"},
	{0xF31FAFA9AE804B75ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_ctn180l.gad"},
	{0xF31FE316470CEBDEULL, "gen00fl_varric6093622.cl"},
	{0xF320192D4278AF74ULL, "524230811.ogg"},
	{0xF3203BF7E6E803E4ULL, "llac_w_woundedcoast_65536_0.dds"},
	{0xF320DB6A6FA1F433ULL, "plt_seb120pt_duty_sunder.plo"},
	{0xF321273D92C45782ULL, "dae000_gamlen6124145.cl"},
	{0xF32167568F6F9033ULL, "86710568.ogg"},
	{0xF3217476BB8DBEFCULL, "6075985_m.fxe"},
	{0xF3220194EAF46AF2ULL, "vxht_brothel_n_502_0_0.vlm"},
	{0xF3225A2F110F9FCAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cartempty.phy"},
	{0xF322B32461119D5DULL, "706529244.ogg"},
	{0xF322CA88788FC652ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_flmb_0n.dds"},
	{0xF322E0321D2E3C2DULL, "hlac_ht_keep_n_3732_16y.dds"},
	{0xF323220F54E3AF86ULL, "6018700_m.fxe"},
	{0xF3232C3C62ED44B2ULL, "vxdr_0_deeproads_cep_0_0.vlm"},
	{0xF323C291174FF391ULL, "follower_banter6155595.cl"},
	{0xF3240290367C037FULL, "vxht_brothel_n_5kb_0_0.vlm"},
	{0xF3246AF058ABF13CULL, "bur_hole_rock.mao"},
	{0xF3249116E9EDB63FULL, "vxw_woundedcoast_18g_0_0.vlm"},
	{0xF325FC4906D7758AULL, "48289063.ogg"},
	{0xF3260D051B9343A9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_deadloop_wesley.ani"},
	{0xF326144BE054D01CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_enter.gad"},
	{0xF326667291164C10ULL, "gen00fl_aveline6132771.cl"},
	{0xF3269642D2B3D71EULL, "6111164_m.fxe"},
	{0xF326C2B8B0975165ULL, "vxht_chantry_n_2t9_0_0.vlm"},
	{0xF326D30F38DFAAF6ULL, "llac_lt_0_lowtown_-131073_1.dds"},
	{0xF32766DF1D6981B8ULL, "mag170_hugh6080729.cl"},
	{0xF32857C2843D6C3BULL, "vxdr_primevaltomb_7e5_0_2.vlm"},
	{0xF32890229FAD75DDULL, "6119691_m.fxe"},
	{0xF3289DF8EEC953A2ULL, "8535711.ogg"},
	{0xF328E84AE3670248ULL, "74402179.ogg"},
	{0xF32905F058E1DB57ULL, "vxca_cavern1_haunted_ck1_0_0.vlm"},
	{0xF3293637241372BFULL, "6016853_m.fxe"},
	{0xF3293EEAD2DB9181ULL, "art\\levels\\ztools\\helpers\\character\\prp_humanstand_1.mao"},
	{0xF3297FFE9C83FD51ULL, "vxw_woundedcoast_3y_0_0.vlm"},
	{0xF32CD2EF7B7EF5A7ULL, "6098334_m.fxe"},
	{0xF32D134BE05AED1AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_bench_enter.ani"},
	{0xF32D28051B99902BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\wesley_deadloop_wesley.gad"},
	{0xF32DB741930FFA97ULL, "vis100pt_offer.ncs"},
	{0xF32DFD2632448FB2ULL, "vxht_chantry_2av_0_0.vlm"},
	{0xF32E157A2472F282ULL, "6209853_f.fxe"},
	{0xF32E33C2BA6AA350ULL, "llac_lt_house_-1_2.dds"},
	{0xF32E3834D42EF942ULL, "hlac_ht_chantry_n_2534_4z.dds"},
	{0xF32E4818431CDE2DULL, "w_bonepitdestroyed_a3_74_0.msh"},
	{0xF32E5618431CF660ULL, "w_bonepitdestroyed_a3_74_0.mao"},
	{0xF32E5A18431CFCB3ULL, "w_bonepitdestroyed_a3_74_0.mmh"},
	{0xF32E87D2DF3C4BAFULL, "vxht_0_hightown_n_c6l_0_0.vlm"},
	{0xF32E99DA66905FABULL, "one300_murder_followers6086893.cl"},
	{0xF32ECCC54A449BEAULL, "84594396.ogg"},
	{0xF32EDAB63C16EC66ULL, "212106413.ogg"},
	{0xF32F316BD2E5D266ULL, "61581344.ogg"},
	{0xF32FBEDCDCCDD712ULL, "437039123.ogg"},
	{0xF32FFBF7E28C0757ULL, "282676192.ogg"},
	{0xF3301888DFD13E47ULL, "vxdr_0_deeproads_d3l_0_0.vlm"},
	{0xF330433B7C6F042DULL, "6094769_m.fxe"},
	{0xF330F24207A7C364ULL, "ttx_wall_a_1.msh"},
	{0xF331084207A7E806ULL, "ttx_wall_a_1.mmh"},
	{0xF3312C952F976ACEULL, "dae000_mother6028525.cl"},
	{0xF3316E6FBE66B95FULL, "6125503_f.fxe"},
	{0xF3317315053F7DD4ULL, "vxw_encounter2_mc_0_0.vlm"},
	{0xF3317E74A2BAC64EULL, "lt_0_lowtown_docks_n_101_0.phy"},
	{0xF331921CE0B4DFE0ULL, "6153263_m.fxe"},
	{0xF331C4FFA458DB43ULL, "vxht_0_hightown_a3_c9w_0_0.vlm"},
	{0xF331D22E0D06B92CULL, "6120359_m.fxe"},
	{0xF3329DD50FC7A2E9ULL, "6079658_m.fxe"},
	{0xF332A2E3A1C825FAULL, "gen_im_trs_prf_junk2_01.uti"},
	{0xF33326D49261496EULL, "gen00fl_merrill6127222.cl"},
	{0xF333414AC33A7FADULL, "lt_0_lowtown_docks_a3_129_0.msh"},
	{0xF3334F4AC33A97E0ULL, "lt_0_lowtown_docks_a3_129_0.mao"},
	{0xF333534AC33A9E33ULL, "lt_0_lowtown_docks_a3_129_0.mmh"},
	{0xF33358AD9CFEC4E2ULL, "6042488_m.fxe"},
	{0xF33379AA188DD768ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\exploration\\pauseturnr180runf.evt"},
	{0xF333C631B319207AULL, "dr_dungeon_3218_ll.mmh"},
	{0xF3341C07942F2E87ULL, "mag160_feynriel6016577.cl"},
	{0xF3347D415C1F8E87ULL, "hlac_lt_0_lowtown_a2_n_c_-2_6x.dds"},
	{0xF334DF04A9F7D38AULL, "141608747.ogg"},
	{0xF334F9221A496FE0ULL, "hlac_lt_0_lowtown_n_1_5z.dds"},
	{0xF3350A649E47BB13ULL, "vxw_encounter2_camp_cz_0_0.vlm"},
	{0xF3351096A511D72EULL, "docks_male_human_worker_sta.cnv"},
	{0xF33566D0E8B2E7B1ULL, "vxca_caverns2prc_9h3_0_0.vlm"},
	{0xF335C658C445DF49ULL, "vxht_0_hightown_a2_avw_0_0.vlm"},
	{0xF335F7A667616E9AULL, "54316926.ogg"},
	{0xF33718AD648E479CULL, "zz_vault_debug6203202.cl"},
	{0xF33736D19F4CD3E5ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.anb"},
	{0xF337AB7932385FD7ULL, "vxht_keep_n_70f_0_0.vlm"},
	{0xF3381A48D0F794B0ULL, "555643484.ogg"},
	{0xF3383296A514469EULL, "docks_male_human_worker_sta.bnk"},
	{0xF3387D58A26C1786ULL, "vxht_mansion2b_2p5_0_0.vlm"},
	{0xF338BAEAE6D79D69ULL, "qun110_petrice6042246.cl"},
	{0xF339167CC4FE27E0ULL, "art\\characters\\heraldry\\hld_dwarven_guild.dds"},
	{0xF339970B6904DB27ULL, "zz_lgt_2_debug6206585.cl"},
	{0xF3399FAFD5D5CB6FULL, "6074044_m.fxe"},
	{0xF339E67138E33411ULL, "character_character_p60.psh.fxo"},
	{0xF33A2419AEC99098ULL, "667730232.ogg"},
	{0xF33A2FC59F7D6CF4ULL, "vxlt_backalley_n_1fn_0_0.vlm"},
	{0xF33A49ABD5E2C52AULL, "42197298.ogg"},
	{0xF33A741E3DD749DEULL, "lowtown_male_lowlife1_sta6165043.cl"},
	{0xF33B02F3F995ACEDULL, "hlac_lt_0_lowtown_a2_n_-131072_5y.dds"},
	{0xF33B1165124CC139ULL, "6132861_m.fxe"},
	{0xF33B39D6B1005882ULL, "vxw_woundedcoast_id_0_0.vlm"},
	{0xF33C016E4290AA99ULL, "hlac_lt_0_lowtown_-131072_4y.dds"},
	{0xF33C238504D48930ULL, "zz_followers_banter6159114.cl"},
	{0xF33C62C409454C03ULL, "ser200_aveline6201490.cl"},
	{0xF33CE793BC6E1CA8ULL, "vxht_brothel_n_4zo_0_0.vlm"},
	{0xF33CF9AFD132642FULL, "vxht_0_hightown_n_c3o_0_0.vlm"},
	{0xF33D01F3D13D1A43ULL, "w_bonepitmining_a2_204_0.mao"},
	{0xF33D05F3D13D2090ULL, "w_bonepitmining_a2_204_0.mmh"},
	{0xF33D13F3D13D38AAULL, "w_bonepitmining_a2_204_0.msh"},
	{0xF33DA4274C79CE7BULL, "6163394_f.fxe"},
	{0xF33DD248FA2B4571ULL, "6203554_m.fxe"},
	{0xF33E3719006C8251ULL, "545254323.ogg"},
	{0xF33E4AADAFC4B3C8ULL, "6218977_m.fxe"},
	{0xF33E4DFDB9B21484ULL, "6111235_f.fxe"},
	{0xF33EC4F948D181CAULL, "follower_barks6169899.cl"},
	{0xF33ECD663FA28A79ULL, "6076685_f.fxe"},
	{0xF33F177345524CC9ULL, "art\\characters\\creatures\\ragedemon\\textures\\c_ragedmon_0d.dds"},
	{0xF33F273D8B2CF9A7ULL, "hlac_lt_0_lowtown_n_-131072_6x.dds"},
	{0xF33FBC6C46D8C3CAULL, "6164397_m.fxe"},
	{0xF34013ECDC868E49ULL, "686530720.ogg"},
	{0xF34030D02BD280C1ULL, "703250565.ogg"},
	{0xF34036DC964EF3C3ULL, "vxht_0_hightown_a3_c2v_0_0.vlm"},
	{0xF3404807D6200FF9ULL, "6120017_m.fxe"},
	{0xF340CADCC053FA47ULL, "475065965.ogg"},
	{0xF3411B6B931E73ADULL, "722515577.ogg"},
	{0xF3412B17D565666FULL, "6131606_m.fxe"},
	{0xF34143C686341B11ULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_golem_orsinosdefiance.gad"},
	{0xF3420FAE59714B2AULL, "vxht_bartrand_n_47f_0_0.vlm"},
	{0xF342193EC7235319ULL, "vxlt_gallowscourtyard_78r_0_0.vlm"},
	{0xF342D8DD5AA3CBD6ULL, "497772068.ogg"},
	{0xF342E41A684B9C47ULL, "qun120_arishok_guard6060798.cl"},
	{0xF342EF399F5C0540ULL, "hlac_ht_0_hightown_a2_-2_0z.dds"},
	{0xF3432D911354CF41ULL, "6092180_f.fxe"},
	{0xF343973B4C8525D3ULL, "follower_banter6136688.cl"},
	{0xF3440758655F3117ULL, "mag291_torpor6116234.cl"},
	{0xF3440F7457DC7517ULL, "al_warrior_t03.gda"},
	{0xF3443835B63BF7C9ULL, "vxht_0_hightown_a2_aqv_0_0.vlm"},
	{0xF3444BCEBD4408A3ULL, "w_encounter3_79_0.msh"},
	{0xF34459CEBD4420FDULL, "w_encounter3_79_0.mmh"},
	{0xF3445DCEBD44270EULL, "w_encounter3_79_0.mao"},
	{0xF3445DF9558B6ED7ULL, "llac_w_bonepitmining_a2_0_9.dds"},
	{0xF34490790A506462ULL, "392703383.ogg"},
	{0xF344C25511B9C649ULL, "vxw_encounter2_k8_0_0.vlm"},
	{0xF344ED4F9762E023ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheadb_sa.dds"},
	{0xF3451568136C0133ULL, "vxw_bonepitdestroyed_a3_1nz_0_0.vlm"},
	{0xF345F7987C9FA5D5ULL, "vxw_bonepit_29j_0_0.vlm"},
	{0xF346400FFBCDB860ULL, "453651071.ogg"},
	{0xF346BFD594776040ULL, "llac_w_sundermount_-262150_0.dds"},
	{0xF346C23447A85DB8ULL, "vxht_mansion2b_2mx_0_0.vlm"},
	{0xF346F4A9BCF05729ULL, "vxca_caverns1_icl_0_0.vlm"},
	{0xF346F9D1BED05BCBULL, "zz_and_debug6178678.cl"},
	{0xF347076FB22DD9B3ULL, "qun111_ketojan6040001.cl"},
	{0xF3482C2C44C79B34ULL, "ser221_quentin6215886.cl"},
	{0xF34844C6863A3BBFULL, "art\\cinematics\\_crit path\\mage crisis\\mcr371cs_orsino_defiance\\cs_golem_orsinosdefiance.ani"},
	{0xF3485CC1A3543A40ULL, "6002781_m.fxe"},
	{0xF348CE368037C975ULL, "670338552.ogg"},
	{0xF349ABB3A2F67102ULL, "vxw_woundedcoast_fe_0_0.vlm"},
	{0xF34AD4BBD8C08BCEULL, "vxdr_primevaltomb_8vl_0_0.vlm"},
	{0xF34B0F3B1B853A26ULL, "453246.ogg"},
	{0xF34B340DD584A83FULL, "docks_male_overseer_mob6176624.cl"},
	{0xF34B6B8CC3287CAFULL, "vxht_0_hightown_n_c8n_0_0.vlm"},
	{0xF34BC09C371A11B1ULL, "6031998_f.fxe"},
	{0xF34BC59831A1EA53ULL, "vxlt_undercitydungeon_172_0_0.vlm"},
	{0xF34BEA6EA92D9483ULL, "vxw_bonepitmining_a2_f7_0_0.vlm"},
	{0xF34C445C9E0812EFULL, "6168114_m.fxe"},
	{0xF34C5ED53044249BULL, "rdr111dg_martin6054581.cl"},
	{0xF34D4E7F7D06A041ULL, "cd_qcr221_diplomatic_sol.wbd"},
	{0xF34D6FDBC6CAD6E4ULL, "mag101_alain6032138.cl"},
	{0xF34DEFB9B33E7C51ULL, "vxca_caverns1_e6g_0_0.vlm"},
	{0xF34E02E3EE942C8EULL, "6219064_m.fxe"},
	{0xF34E04824989346AULL, "33400381.ogg"},
	{0xF34E78D94CBADB3DULL, "vxlt_0_lowtown_docks_but_0_0.vlm"},
	{0xF34E9AB59525137BULL, "6045018_f.fxe"},
	{0xF34EA8B988450C43ULL, "vxht_0_hightown_a3_c7y_0_0.vlm"},
	{0xF34EDFA9E7A5BC1CULL, "llac_ht_brothel_n_3370_2.dds"},
	{0xF34F5ABC40AC5096ULL, "6104882_f.fxe"},
	{0xF34FA64AF56737CDULL, "6122833_m.fxe"},
	{0xF34FE4F40EC2CFFDULL, "zz_lgt_2_debug6140452.cl"},
	{0xF35046FFCC215C61ULL, "fxa_evl_center.mao"},
	{0xF3507AB8AC3483E1ULL, "144422295.ogg"},
	{0xF3509EE15738E5EFULL, "mrl150_merrill6093961.cl"},
	{0xF350C26A6FCAEDD0ULL, "plt_seb120pt_duty_sunder.nss"},
	{0xF350EF6425BAEFA2ULL, "6125567_m.fxe"},
	{0xF351A1C9D98DE249ULL, "521470941.ogg"},
	{0xF352093253BE0F2EULL, "fx_snow_wsip.dds"},
	{0xF35220464CADB830ULL, "gen00fl_anders6125149.cl"},
	{0xF35242F9DA7236BAULL, "art\\characters\\weapons\\maces\\textures\\w_mce_mc105a_0s.dds"},
	{0xF352B452A62E3399ULL, "6104262_m.fxe"},
	{0xF352C610A46755DAULL, "6132847_f.fxe"},
	{0xF3536B239F14D650ULL, "hlac_ht_0_hightown_a3_65532_0y.dds"},
	{0xF3537C157E354313ULL, "ch2notnt_dy.psh.fxo"},
	{0xF3539CDB558265C8ULL, "vxht_0_hightown_5i3_0_0.vlm"},
	{0xF354684BC2DCE13CULL, "mer210_coterie_contact6036773.cl"},
	{0xF354FF7FB70ABB20ULL, "vxw_encounter3_4w_0_0.vlm"},
	{0xF355464DF2F688D2ULL, "vxlt_undercitydungeon_177_0_0.vlm"},
	{0xF355B22E01FB9A45ULL, "vxlt_gallowsprison_n_6bd_0_0.vlm"},
	{0xF355B95636A505F1ULL, "6125476_m.fxe"},
	{0xF355DBF1C361A49DULL, "zz_o_11_gallows_nt_v1.cut"},
	{0xF3563869F4E000FEULL, "art\\vfx\\cinematic\\vfx_chantry_int_exp_p.mmh"},
	{0xF3565228266EF00EULL, "68170749.ogg"},
	{0xF356CB9A4148AF66ULL, "hlac_lt_0_lowtown_docks_a3_n_65536_0y.dds"},
	{0xF356F9D6485BF809ULL, "follower_banter6169747.cl"},
	{0xF357137F83699DF4ULL, "vxlt_backalley_n_1ll_0_0.vlm"},
	{0xF3574D6D1B41795CULL, "mer240_jansen6059688.cl"},
	{0xF3577CA41BC108CFULL, "828797048.ogg"},
	{0xF35792030B4F0092ULL, "6077230_f.fxe"},
	{0xF357AD55A5EDD027ULL, "169284649.ogg"},
	{0xF3581D9094EC8982ULL, "vxw_woundedcoast_cf_0_0.vlm"},
	{0xF3584182C49AE087ULL, "6105421_m.fxe"},
	{0xF3588AB26ECB466CULL, "hlac_ca_caverns2prc_5802_59x.dds"},
	{0xF358B31AB5F78DFAULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masb_0t.dds"},
	{0xF3594F7BA24F3562ULL, "zz_dae_debug6029525.cl"},
	{0xF35A3B6BD43463F0ULL, "vxlt_warehouse_1cg_0_0.vlm"},
	{0xF35A484F80C6B55FULL, "vxlt_warehouse_si_0_0.vlm"},
	{0xF35B8786CB36CF09ULL, "648118144.ogg"},
	{0xF35B89103BBE2DF5ULL, "mcr371cr_harvester1.utc"},
	{0xF35BBBF05ED56E91ULL, "432640475.ogg"},
	{0xF35C46644D20F28DULL, "llac_lt_0_lowtown_a3_n_c_131072_2.dds"},
	{0xF35CC067731577FFULL, "6084494_m.fxe"},
	{0xF35CEEEC417F52B9ULL, "280633256.ogg"},
	{0xF35D36195723265CULL, "hlac_lt_0_lowtown_a3_n_c_65533_2y.dds"},
	{0xF35D72456DF53897ULL, "6065824_m.fxe"},
	{0xF35D75D7B5E070A7ULL, "hlac_lt_0_lowtown_-196609_1y.dds"},
	{0xF35DA2B6D926C932ULL, "ave200_donnic6110693.cl"},
	{0xF35E68BB118028AFULL, "229862158.ogg"},
	{0xF35E7075F7954F77ULL, "gen00fl_aveline6139942.cl"},
	{0xF35E9B9F998090DDULL, "6080702_m.fxe"},
	{0xF35EB85D2FD9033CULL, "6134726_f.fxe"},
	{0xF35EE567F885723CULL, "310256218.ogg"},
	{0xF35F108B98CD6DB5ULL, "vxca_caverns1_mines_k6t_0_0.vlm"},
	{0xF35F3AD5AA51E08CULL, "hlac_ht_playershouse_n_855_35y.dds"},
	{0xF35F6D1C4084BE96ULL, "6135144_m.fxe"},
	{0xF35FD230949BB7CBULL, "zz_approval_debug6003527.cl"},
	{0xF3600263FAECC641ULL, "qcr000_varric6208258.cl"},
	{0xF36037D19F6FDA26ULL, "art\\vfx\\combat\\vfx_ha_bd_gthrow_proj_p.mmh"},
	{0xF3607A4E5B5F9884ULL, "vxlt_hangedman_n_1mn_0_0.vlm"},
	{0xF360C174E6603C66ULL, "159476475.ogg"},
	{0xF360D092853CD539ULL, "1035023111.ogg"},
	{0xF360E1C04E770AFAULL, "vxht_mansion_n_4wi_0_0.vlm"},
	{0xF3612EC978A9AD4DULL, "vxdr_0_deeproads_fbs_0_0.vlm"},
	{0xF3616249DD7957B4ULL, "gen_im_wep_war_two_act3_06.uti"},
	{0xF3617E5BEC3440EAULL, "zz_seb_debug6215771.cl"},
	{0xF361E56596E03F0EULL, "239248932.ogg"},
	{0xF36286A0CAC6BEEBULL, "6136694_m.fxe"},
	{0xF363715CA900D3A0ULL, "vxw_encounter3_1t_0_0.vlm"},
	{0xF3637D0690F5ECF1ULL, "hlac_w_blightlands_131074_1y.dds"},
	{0xF363809837A1E858ULL, "vxlt_gallowscourtyard_721_0_0.vlm"},
	{0xF363D863A0DC8829ULL, "vxca_caverns1_iir_0_0.vlm"},
	{0xF3641769F4EC05BDULL, "art\\vfx\\cinematic\\vfx_chantry_int_exp_p.anb"},
	{0xF3645596801A927BULL, "541306488.ogg"},
	{0xF364EEE5CA33ACDDULL, "97213827.ogg"},
	{0xF364F2A2D6875552ULL, "vxca_cavern1_haunted_hl2_0_0.vlm"},
	{0xF36541BB50410059ULL, "hlac_lt_house_-1_3x.dds"},
	{0xF365855C755FB674ULL, "vxlt_backalley_n_1wm_0_0.vlm"},
	{0xF36592834CCA654CULL, "art\\characters\\weapons\\axes\\w_axe_ax105.mmh"},
	{0xF36598834CCA6F3EULL, "art\\characters\\weapons\\axes\\w_axe_ax105.msh"},
	{0xF3659942B26ADA25ULL, "mag320_orsino6070436.cl"},
	{0xF36604804BD80361ULL, "hlac_lt_0_lowtown_a2_n_65535_1z.dds"},
	{0xF36641D8B67F3433ULL, "hlac_w_encounter2_camp_327685_0x.dds"},
	{0xF3664568CC9D0443ULL, "llac_ht_0_hightown_n_0_1.dds"},
	{0xF36646A78E6392FCULL, "6093560_f.fxe"},
	{0xF3668F6D86E2A202ULL, "vxw_woundedcoast_xg_0_0.vlm"},
	{0xF36691CE7967CD7FULL, "418847247.ogg"},
	{0xF36765888B7A4614ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\basedualwield\\blendtree\\assassin_movement.anb"},
	{0xF3679FCA93D889DEULL, "hlac_ca_caverns1_mines_24424_12y.dds"},
	{0xF367B1B299F81EC8ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06a.msh"},
	{0xF367B99111F6C3ADULL, "pro000_opening6207385.cl"},
	{0xF367CFB299F851F2ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06a.mmh"},
	{0xF3681DFF7CBFA266ULL, "vxca_caverns1_i9a_0_0.vlm"},
	{0xF3683F543C2CB1F4ULL, "vxw_blightlands_115_0_0.vlm"},
	{0xF368BD275F937C85ULL, "22776599.ogg"},
	{0xF368CE288D19C583ULL, "vxw_bonepitmining_a2_h9_0_0.vlm"},
	{0xF3691EEAA23CC71EULL, "lowtown_male_urchin_sta6161591.cl"},
	{0xF3696686B6659962ULL, "vxw_encounter1_dg_0_0.vlm"},
	{0xF36A02A6007571FAULL, "hlac_lt_0_lowtown_a2_n_65533_0x.dds"},
	{0xF36A056AAA32355DULL, "art\\characters\\creatures\\dragonling\\textures\\c_dragling_0n.dds"},
	{0xF36AC7C755C3766FULL, "gen00fl_fenris6110810.cl"},
	{0xF36ADF97A5ED432EULL, "vxw_sundermount_2er_0_0.vlm"},
	{0xF36B18D03A07470AULL, "ht_0_hightownexplosion_n_11_0.mao"},
	{0xF36B1CD03A074E59ULL, "ht_0_hightownexplosion_n_11_0.mmh"},
	{0xF36B26D03A075F47ULL, "ht_0_hightownexplosion_n_11_0.msh"},
	{0xF36CB0B533B2FA24ULL, "vxht_mansion2b_1ik_0_0.vlm"},
	{0xF36CD6C9A014C2E5ULL, "6133109_m.fxe"},
	{0xF36D24EE4E884434ULL, "vrc221_varric6044172.cl"},
	{0xF36DBC50043202ADULL, "hlac_lt_0_lowtown_a2_n_c_65536_8z.dds"},
	{0xF36DC56DEA4D7599ULL, "llac_lt_warehouse_n_0_3.dds"},
	{0xF36E1D45ED008D24ULL, "219775337.ogg"},
	{0xF36EB3C0397448B4ULL, "6117448_f.fxe"},
	{0xF36F715D493DD75DULL, "zz_approval_debug6003936.cl"},
	{0xF36F7AE376BC6069ULL, "hairalpha_hair_p2.psh.fxo.dx11"},
	{0xF37005D6EAD559C4ULL, "qun100ar_plot_start.are"},
	{0xF370743B85C96F8FULL, "qun160_party_comment_2.bnk"},
	{0xF370AA492D6C8EBAULL, "1022898482.ogg"},
	{0xF370D9F5CCF482A9ULL, "gen00fl_dog6166483.cl"},
	{0xF3717D74A2F0A811ULL, "lt_0_lowtown_docks_n_101_0.mmh"},
	{0xF3718974A2F0BCB2ULL, "lt_0_lowtown_docks_n_101_0.mao"},
	{0xF3719774A2F0D4CFULL, "lt_0_lowtown_docks_n_101_0.msh"},
	{0xF3720620B62E47B3ULL, "vxlt_0_lowtown_docks_b5j_0_0.vlm"},
	{0xF3722D353F4C1E1EULL, "vxht_mansion_n_6bd_0_0.vlm"},
	{0xF3725E8664ECA5DAULL, "clf260_fell_orden6070483.cl"},
	{0xF372707F7D25F643ULL, "cd_qcr221_diplomatic_sol.bnk"},
	{0xF3728B9610884B76ULL, "vxca_caverns2_8zb_0_0.vlm"},
	{0xF373063F84E24FF2ULL, "vxw_bonepit_1pm_0_0.vlm"},
	{0xF37319E8606C800AULL, "248350004.ogg"},
	{0xF373CC25949D2E04ULL, "vxlt_0_lowtown_eu3_0_0.vlm"},
	{0xF374068646D40F22ULL, "hlac_lt_0_lowtown_docks_a3_196607_0x.dds"},
	{0xF374963B85CD92F7ULL, "qun160_party_comment_2.cnv"},
	{0xF374E6D3A6B63EB1ULL, "543079533.ogg"},
	{0xF374FEF8E74E105BULL, "64229606.ogg"},
	{0xF3752280F3E8CAB8ULL, "gen00fl_aveline6132649.cl"},
	{0xF375447190C148F3ULL, "803530206.ogg"},
	{0xF375AA5F898F6A5EULL, "441705669.ogg"},
	{0xF3767DD75FA60B33ULL, "gen00fl_isabela6102457.cl"},
	{0xF376A64994769F26ULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_slam.ani"},
	{0xF376BA62CF8FF11EULL, "365902146.ogg"},
	{0xF37711B7A4752A36ULL, "vxht_mansion_n_6p6_0_0.vlm"},
	{0xF3772C0964B2E65FULL, "vxlt_warehouse_yk_0_0.vlm"},
	{0xF377809FDA3FFD03ULL, "art\\vfx\\combat\\vfx_gib_qunw_rblt.anb"},
	{0xF377924FC979577EULL, "6080561_m.fxe"},
	{0xF37800701CA07047ULL, "hlac_lt_undercitydungeon_-1_26z.dds"},
	{0xF37829279AC0D148ULL, "160838538.ogg"},
	{0xF3783B43ADE4A32BULL, "vxw_bonepitmining_a2_15m_0_0.vlm"},
	{0xF3783B6FE49521ADULL, "147736850.ogg"},
	{0xF3786E197C4A161AULL, "vxlt_hangedman_31y_0_0.vlm"},
	{0xF379517497E35BAEULL, "vxw_sundermount_2ns_0_0.vlm"},
	{0xF379636C3644A940ULL, "371773497.ogg"},
	{0xF379B05889D0AB2DULL, "hlac_ht_chantry_2534_9y.dds"},
	{0xF37AB7B041676219ULL, "vrc221_varric6043945.cl"},
	{0xF37AFDCD31792BD2ULL, "6116536_m.fxe"},
	{0xF37B5A93FAC902B9ULL, "861723912.ogg"},
	{0xF37BAC8E9C003E8DULL, "gen00fl_isabela_init6175658.cl"},
	{0xF37BAF1AC8247467ULL, "46463798.ogg"},
	{0xF37BDD60C92ACFBCULL, "gen00fl_anders6096467.cl"},
	{0xF37C27975CADC85BULL, "lt_0_lowtown_docks_27_0_lc.mmh"},
	{0xF37C51938A2AEEA2ULL, "art\\vfx\\combat\\vfx_w_tmp_clean_aoe_p.impa.gad"},
	{0xF37D5A4BB5F2F8B0ULL, "llac_lt_0_lowtown_a3_n_c_131073_2.dds"},
	{0xF37D64B94B9EC3AAULL, "t3_bro_b01.tnt"},
	{0xF37D9AA75795F783ULL, "gen00fl_merrill6145140.cl"},
	{0xF37DDDFB2CC21E92ULL, "vxlt_gallowstemplar_5us_0_1.vlm"},
	{0xF37E0CF3D174C3ABULL, "w_bonepitmining_a2_204_0.phy"},
	{0xF37E12835C95DE4DULL, "vxdr_0_deeproads_fhq_0_0.vlm"},
	{0xF37ED79BE05BC654ULL, "str_slime_s.dds"},
	{0xF37F4A4811EFF8C1ULL, "6027705_m.fxe"},
	{0xF37F85B1E44A9DF0ULL, "6086172_m.fxe"},
	{0xF37FAB119BE2DF2CULL, "vxlt_0_lowtown_a2_gv8_0_0.vlm"},
	{0xF37FE1DD613423F4ULL, "gen_im_arm_cht_elf_coma_t1.uti"},
	{0xF3809D195660102CULL, "plt_cod_mgc_lyrium.plo"},
	{0xF380D3A6975E72EBULL, "6143980_m.fxe"},
	{0xF380F33E10A1B048ULL, "and111ip_donation_box.utp"},
	{0xF381000566E4F6F7ULL, "s_po_w_smo.mao"},
	{0xF3816E70DCD080F1ULL, "599368687.ogg"},
	{0xF3823E0286934684ULL, "vxlt_0_lowtown_ep0_0_0.vlm"},
	{0xF38241478DDAE21FULL, "one300_murder_followers6089900.cl"},
	{0xF382BCB3600F21A1ULL, "704453583.ogg"},
	{0xF38310140DA56B0EULL, "307657022.ogg"},
	{0xF38373276ACED302ULL, "vxw_woundedcoast_ra_0_0.vlm"},
	{0xF38374938A314899ULL, "art\\vfx\\combat\\vfx_w_tmp_clean_aoe_p.impa.evt"},
	{0xF3838E4994810097ULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_slam.evt"},
	{0xF3842727C987AE7EULL, "seb200im_allure_staff.uti"},
	{0xF38439DB8902A9F6ULL, "vxht_0_hightown_a2_bl5_0_0.vlm"},
	{0xF3848E789AB90CC0ULL, "90810299.ogg"},
	{0xF3849796E65B79ABULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0sc_basic_4.mao"},
	{0xF384DF47928E2278ULL, "6068451_m.fxe"},
	{0xF385039FF0B75065ULL, "vxlt_hangedman_n_16v_0_0.vlm"},
	{0xF38533D5146E0001ULL, "hlac_dr_0_deeproads_5849_7x.dds"},
	{0xF3854B998FFBAB40ULL, "gen00fl_aveline6152575.cl"},
	{0xF3872D8E4EEA3977ULL, "lt_0_lowtown_a2_46_0_lc.mmh"},
	{0xF38790DD116A0B97ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile.phy"},
	{0xF387A37D66B67035ULL, "vxlt_0_lowtown_docks_a3_n_cyd_0_1.vlm"},
	{0xF387F044D0AA9DA0ULL, "430509948.ogg"},
	{0xF387F514DECB49AFULL, "550800852.ogg"},
	{0xF38808CB8D949508ULL, "hlac_lt_gallowstemplar_f_2391_1y.dds"},
	{0xF38858724305165FULL, "6129359_m.fxe"},
	{0xF3885DA4D7CA8E2EULL, "lgt200cr_tome_horror.utc"},
	{0xF3886DC8A4A84681ULL, "gen_im_arm_cht_hvy_act1_03.uti"},
	{0xF388C3AE06998502ULL, "6127015_m.fxe"},
	{0xF3894F031B7D3004ULL, "zz_followers_banter6159639.cl"},
	{0xF389946F179F2B24ULL, "vxht_mansion2b_1gi_0_0.vlm"},
	{0xF38998A2F28BA13CULL, "579487286.ogg"},
	{0xF38999F6B8521121ULL, "hlac_lt_0_lowtown_a2_n_65536_0x.dds"},
	{0xF389A29F359FDEF6ULL, "vxlt_gallowstemplar_5wl_0_0.vlm"},
	{0xF389BA838400F3E5ULL, "6139709_m.fxe"},
	{0xF38A08A6C76FFA39ULL, "vxht_keep_n_60q_0_0.vlm"},
	{0xF38A0E2415732F0BULL, "hlac_w_bonepitdestroyed_a3_393216_0y.dds"},
	{0xF38A8F4994872120ULL, "art\\characters\\creatures\\stonegolem\\animation\\tal_slam.gad"},
	{0xF38B450F02863904ULL, "hlac_dr_0_deeproads_17038_5z.dds"},
	{0xF38BA8DA2C5420CDULL, "6016719_f.fxe"},
	{0xF38BB91A28711B07ULL, "vxlt_0_lowtown_ghu_0_0.vlm"},
	{0xF38BBA691FCA9F84ULL, "14780956.ogg"},
	{0xF38C27E3748F6A5FULL, "6137288_m.fxe"},
	{0xF38C37572459547AULL, "vxht_mansion_n_4dd_0_0.vlm"},
	{0xF38C7B4C50348DE7ULL, "574725528.ogg"},
	{0xF38C7FD884AF5C92ULL, "vxw_sundermount_304_0_0.vlm"},
	{0xF38CE19739586527ULL, "vxht_keep_67q_0_0.vlm"},
	{0xF38CED5346A17B09ULL, "531558622.ogg"},
	{0xF38D03AF2B06B26DULL, "79204557.ogg"},
	{0xF38DEAEA2413626FULL, "vxw_blightlands_s9_0_0.vlm"},
	{0xF38E0CD6EAEE6366ULL, "qun100ar_plot_start.ncs"},
	{0xF38E1CEE8DD8F7ACULL, "vxlt_0_lowtown_a2_gm9_0_0.vlm"},
	{0xF38E5CBB7D49BE2EULL, "dae000_orsino6085428.cl"},
	{0xF38EDE318012F35AULL, "zz_followers_banter6158634.cl"},
	{0xF38F0847AC541F3DULL, "vxht_keep_n_4bu_0_0.vlm"},
	{0xF38F0F95C25A24B8ULL, "836625343.ogg"},
	{0xF38F4F3B688FFB14ULL, "6201548_m.fxe"},
	{0xF38F913CC0F9ECD7ULL, "vxlt_0_lowtown_a2_n_c_e8b_0_0.vlm"},
	{0xF38F9FC666C48A16ULL, "vxlt_0_lowtown_a2_n_gn9_0_0.vlm"},
	{0xF38FA9BEC2D8F948ULL, "vxlt_gallowsprison_n_6n9_0_0.vlm"},
	{0xF3902DF7601C3D25ULL, "llac_lt_0_lowtown_docks_a3_n_-65539_2.dds"},
	{0xF3907C938A3BE088ULL, "art\\vfx\\combat\\vfx_w_tmp_clean_aoe_p.impa.ani"},
	{0xF390A3DB20100321ULL, "6137585_m.fxe"},
	{0xF390B78D1D2BA6A3ULL, "6139930_m.fxe"},
	{0xF390DAF34B41FFF4ULL, "vxlt_backalley_n_1xp_0_0.vlm"},
	{0xF3917021902E29A4ULL, "vxca_caverns1_mines_icr_0_2.vlm"},
	{0xF391FE952624293AULL, "one000_ilen6205709.cl"},
	{0xF39286E3DF267FE3ULL, "zz_followers_banter6159069.cl"},
	{0xF3930587A9FF8A3BULL, "vxlt_0_lowtown_a2_f66_0_0.vlm"},
	{0xF393B936E6D453C1ULL, "318370915.ogg"},
	{0xF393FAD6A5FA8F77ULL, "llac_lt_0_lowtown_a2_n_1_0.dds"},
	{0xF394362CD914418EULL, "hlac_lt_0_lowtown_a2_n_c_-196609_1y.dds"},
	{0xF39492E5F37A65E1ULL, "vxlt_gallowsprison_n_6k3_0_0.vlm"},
	{0xF394EC7B7B8DC281ULL, "lowtown_fem_prostitute_sta6152331.cl"},
	{0xF39504D6EAF4745FULL, "qun100ar_plot_start.lst"},
	{0xF3950BEB1341EEECULL, "6086284_m.fxe"},
	{0xF3950F59A1A4F50CULL, "vxw_bonepitmining_a2_82_0_0.vlm"},
	{0xF3954CF21A75FFD8ULL, "vxht_0_hightown_a3_bpf_0_0.vlm"},
	{0xF395BA109855F114ULL, "309791805.ogg"},
	{0xF396019824A1B52AULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_dssba_0n.dds"},
	{0xF396352E7BCF8CAEULL, "vxw_sundermount_2tu_0_0.vlm"},
	{0xF396B0CF635F1F25ULL, "dae000_mother6004126.cl"},
	{0xF396D2A41B2EFD03ULL, "148668129.ogg"},
	{0xF397032FC039A15EULL, "hlac_lt_darktown_-1_48z.dds"},
	{0xF3979A4D6EEB20F2ULL, "ser221_party_comments6111563.cl"},
	{0xF397AA047A763FA0ULL, "atl_startmenu2_dxt1_00.dds"},
	{0xF397D2731710046DULL, "hightown_fem_courtesan_mob6168919.cl"},
	{0xF397D2ECB80A9A6EULL, "mag110_samson6049815.cl"},
	{0xF397FA2A115254C9ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb107a_0n.dds"},
	{0xF398064C099543A4ULL, "vxht_mansion2b_1zn_0_0.vlm"},
	{0xF3980D8B01207A80ULL, "vxht_0_hightown_a2_b8w_0_0.vlm"},
	{0xF398147C2795F776ULL, "vxlt_gallowstemplar_5pm_0_0.vlm"},
	{0xF3984810A432CE66ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_3y.dds"},
	{0xF398760B2EC1E090ULL, "112443456.ogg"},
	{0xF398E8AF705306A2ULL, "ht_0_hightown_a3_34_0.rml"},
	{0xF399458504FE2827ULL, "78413523.ogg"},
	{0xF3995633E1B8639BULL, "574571.ogg"},
	{0xF399758674F1CAA8ULL, "6052923_f.fxe"},
	{0xF399F36ABB150794ULL, "fen210_orana6109767.cl"},
	{0xF39A9394D6D383D2ULL, "dae000_madam_lusine6152170.cl"},
	{0xF39AC708BF214E91ULL, "lt_ruinedhouse_n_probe_6l.dds"},
	{0xF39AD181D21EDF97ULL, "218671040.ogg"},
	{0xF39B046D82D7A595ULL, "6128776_m.fxe"},
	{0xF39B36B96CF07A11ULL, "hlac_ca_cavern1_haunted_24425_13z.dds"},
	{0xF39B479362D69DC0ULL, "art\\vfx\\combat\\vfx_impact_p.mmh"},
	{0xF39B53742B4E7DA7ULL, "vxht_keep_60p_0_0.vlm"},
	{0xF39B751CCD82A6BBULL, "vxw_bonepitdestroyed_a3_247_0_0.vlm"},
	{0xF39B8411B370F495ULL, "6093536_f.fxe"},
	{0xF39C5CC716097AEFULL, "vxw_blightlands_t8_0_0.vlm"},
	{0xF39C9D499AD56609ULL, "gencr_raider_2w_f.utc"},
	{0xF39CA110B49BEE43ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.cess.evt"},
	{0xF39D1D64E841E47BULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothe.phy"},
	{0xF39D5BC55455EE3FULL, "1042438494.ogg"},
	{0xF39E1B9BB4CF11C8ULL, "vxlt_gallowsprison_n_6k6_0_0.vlm"},
	{0xF39E1E91DDC9D05BULL, "vis121_saemus6099987.cl"},
	{0xF39E43EFD7E1E754ULL, "hlac_ca_caverns1_24426_2y.dds"},
	{0xF39F1D6F499AF1CCULL, "6017848_m.fxe"},
	{0xF39F5FF8841F9E47ULL, "sst_ss_arcanehorror_1.gda"},
	{0xF39F733528D0D4DBULL, "6127137_m.fxe"},
	{0xF39FF7069E4D2287ULL, "577295660.ogg"},
	{0xF3A07057B089FE3AULL, "mrl310_merrill6120411.cl"},
	{0xF3A08EEF684DB492ULL, "vxlt_undercityencounter_18w_0_0.vlm"},
	{0xF3A1C45CA6E5FCD1ULL, "vxw_bonepitmining_a2_1xh_0_0.vlm"},
	{0xF3A211DDA9F54FC9ULL, "6207983_m.fxe"},
	{0xF3A2724C08801725ULL, "vxlt_gallowstemplar_5c7_0_0.vlm"},
	{0xF3A297CF2A3BA166ULL, "hlac_ht_0_hightown_a2_0_1z.dds"},
	{0xF3A2A4583C7CD70AULL, "gen00fl_anders6096458.cl"},
	{0xF3A32F06825C9567ULL, "55758014.ogg"},
	{0xF3A37A10B4A1CAC4ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.cess.gad"},
	{0xF3A3B99CF7929431ULL, "620821264.ogg"},
	{0xF3A48C1B571A99DCULL, "art\\levels\\races\\design\\_textures\\prp_gasbarrel_d.dds"},
	{0xF3A490D949A1BC8BULL, "478541613.ogg"},
	{0xF3A4A70B6DC5A52EULL, "vxw_sundermount_2yv_0_0.vlm"},
	{0xF3A4E1857EF7848AULL, "hlac_lt_0_lowtown_n_65536_0y.dds"},
	{0xF3A4E6E40670A9E4ULL, "572321965.ogg"},
	{0xF3A4FE10DFD25129ULL, "vxlt_gallowscourtyard_62i_0_0.vlm"},
	{0xF3A5B4012633ED71ULL, "llac_lt_0_lowtown_a2_n_65537_5.dds"},
	{0xF3A6B315D9061AD9ULL, "w_bonepitmining_a2_mmi.gff"},
	{0xF3A771DF7118B742ULL, "6165387_m.fxe"},
	{0xF3A77FD45EBFDB56ULL, "vxlt_warehouse_vs_0_0.vlm"},
	{0xF3A784EA8C0DA689ULL, "675454424.ogg"},
	{0xF3A789A45469002CULL, "hlac_lt_0_lowtown_a2_n_c_131074_0x.dds"},
	{0xF3A807E8F464C94AULL, "6130965_m.fxe"},
	{0xF3A84B8C939A0593ULL, "ser221_gascard6166102.cl"},
	{0xF3A84D9237799606ULL, "32389597.ogg"},
	{0xF3A86120289BC8EAULL, "218110430.ogg"},
	{0xF3A8FF389C281F5BULL, "gen00fl_aveline6139425.cl"},
	{0xF3A92A9362E2A9DFULL, "art\\vfx\\combat\\vfx_impact_p.anb"},
	{0xF3A98910B4A64FA2ULL, "art\\vfx\\combat\\vfx_m_arc_modal_loop_c.cess.ani"},
	{0xF3A9F9FD44DF58EDULL, "hlac_lt_0_lowtown_a2_n_-65535_2z.dds"},
	{0xF3AAECC441B19B73ULL, "vxlt_darktown_40q_0_0.vlm"},
	{0xF3AB0E2349F72AE8ULL, "qun211_braggart6075706.cl"},
	{0xF3ABD9504AB215ECULL, "sb03_back.dds"},
	{0xF3AC6EBE35B814A4ULL, "571682566.ogg"},
	{0xF3AC8342748A48C1ULL, "198444252.ogg"},
	{0xF3AC91B5DD045BF6ULL, "vxlt_0_lowtown_a2_hg6_0_0.vlm"},
	{0xF3ACD2DB4532ACC4ULL, "vxlt_0_lowtown_a2_n_hck_0_0.vlm"},
	{0xF3ACDF32B1954ED1ULL, "406190153.ogg"},
	{0xF3AD38CFF6A08E90ULL, "vxw_encounter3_g_0_0.vlm"},
	{0xF3ADA8FA10755392ULL, "gen00fl_fenris6110875.cl"},
	{0xF3ADF3DE43B136F1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\aveline\\aveline_intimidate_en.evt"},
	{0xF3AE0CCDC36E16EDULL, "426368129.ogg"},
	{0xF3AE18CACE991742ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb01a.msh"},
	{0xF3AE1ACACE991AF8ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb01a.mmh"},
	{0xF3AE7EF94A24B7C0ULL, "zz_followers_banter6157884.cl"},
	{0xF3AE82D2E07A610AULL, "vxlt_0_lowtown_ffo_0_0.vlm"},
	{0xF3AEDE129EA9A9DCULL, "6044642_m.fxe"},
	{0xF3AEF707CE5AF2F1ULL, "hlac_ht_0_hightown_65536_3x.dds"},
	{0xF3AF6F253FA95E23ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_wk_p_rf.ani"},
	{0xF3AF9BC4B396C817ULL, "vxht_mansion2_3ef_0_0.vlm"},
	{0xF3AFEE9F934522A3ULL, "art\\characters\\creatures\\mubari\\animation\\com_growl.ani"},
	{0xF3B004244A8F84AFULL, "6043136_m.fxe"},
	{0xF3B089D96A0BBEC4ULL, "ico_money.dds"},
	{0xF3B0D297FE417920ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglehead_sa.dds"},
	{0xF3B154A3724F6319ULL, "vxw_bonepitdestroyed_a3_18r_0_0.vlm"},
	{0xF3B1769FD76696E1ULL, "vxlt_gallowsprison_n_6q1_0_0.vlm"},
	{0xF3B20E44E743F241ULL, "llac_lt_0_lowtown_a2_-3_6.dds"},
	{0xF3B28540AE531D0BULL, "mag210_arianni6024529.cl"},
	{0xF3B296E12483E069ULL, "vxht_mansion_n_5b0_0_0.vlm"},
	{0xF3B33F5A704BFE63ULL, "vxht_0_hightown_cf3_0_0.vlm"},
	{0xF3B41EB8F8CB6F5FULL, "vxlt_blackemporium_10i_0_0.vlm"},
	{0xF3B448A23773F160ULL, "pro000_merchant6051604.cl"},
	{0xF3B4D1CAC42A73F5ULL, "llac_dr_primevalentrance_1726_1.dds"},
	{0xF3B4D2573BAEC697ULL, "vfx0_addv.vsh.fxo"},
	{0xF3B4EA05ED8174A4ULL, "vxht_mansion2b_1pl_0_0.vlm"},
	{0xF3B4F60A04089836ULL, "6176035_m.fxe"},
	{0xF3B57B11D8201DFCULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_c.impa.ani"},
	{0xF3B59C7B892ABD53ULL, "vxlt_0_lowtown_a3_n_c_hq1_0_0.vlm"},
	{0xF3B5D2D10CAFD481ULL, "mcr370_templar26087642.cl"},
	{0xF3B6A2DA9DFDE23BULL, "6117270_f.fxe"},
	{0xF3B77C7764BAF23CULL, "6104758_f.fxe"},
	{0xF3B7AC3EA59AE129ULL, "385233443.ogg"},
	{0xF3B7C4706546E9FBULL, "6072679_m.fxe"},
	{0xF3B7C4A2E25C23B9ULL, "6103555_m.fxe"},
	{0xF3B821CACEA17A83ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb01a.phy"},
	{0xF3B92E585DF0EB84ULL, "dae200_bran6142984.cl"},
	{0xF3B9AB4F51DFECBBULL, "872690296.ogg"},
	{0xF3BA3E46ED0BFF26ULL, "6178356_m.fxe"},
	{0xF3BA74C6032AE9D6ULL, "313735545.ogg"},
	{0xF3BB3F903EB0CA72ULL, "vxw_bonepit_1ij_0_0.vlm"},
	{0xF3BB53B028338455ULL, "6162734_m.fxe"},
	{0xF3BB9666E96447ACULL, "cod_hst_vael_family.utp"},
	{0xF3BBE31CCBD060F6ULL, "w_bonepitmining_a2_205_0.phy"},
	{0xF3BC8811D826523EULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_c.impa.gad"},
	{0xF3BCC09BCB3DD4DFULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb101a.mmh"},
	{0xF3BCC69BCB3DDE81ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb101a.msh"},
	{0xF3BCFE98EBE9EAD4ULL, "6163787_m.fxe"},
	{0xF3BD5D1920F774BAULL, "6162295_player_bethany.cut"},
	{0xF3BD82BD70B51FD9ULL, "zz_qcr_debug6054933.cl"},
	{0xF3BDE41DB6C33CF8ULL, "gen00fl_merrill6161510.cl"},
	{0xF3BDF13DAE3422B4ULL, "6094799_m.fxe"},
	{0xF3BE069F9351889AULL, "art\\characters\\creatures\\mubari\\animation\\com_growl.evt"},
	{0xF3BE274A75BC0D34ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolegibsgolem_demonhead_mindblast.ani"},
	{0xF3BE5652D284F827ULL, "vxlt_gallowstemplar_n_4ue_0_0.vlm"},
	{0xF3BECE9C592A5415ULL, "6044070_m.fxe"},
	{0xF3BECEF0E5B9725BULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hpath_n.dds"},
	{0xF3BEE9622C18298BULL, "hlac_w_blightlands_393215_0y.dds"},
	{0xF3BEF6CAB04FDA6AULL, "art\\levels\\races\\elf\\props\\plotspecifc\\prp_eluviancrack_0.pwk"},
	{0xF3BF1D314A7F5BD2ULL, "vxlt_0_lowtown_a3_n_c_hq4_0_0.vlm"},
	{0xF3BF273BFF93CA82ULL, "471101705.ogg"},
	{0xF3BF497904E93203ULL, "hlac_ht_mansion2b_818_52x.dds"},
	{0xF3BF5ACDD2D2C1EDULL, "vxlt_hangedman_25k_0_0.vlm"},
	{0xF3BF8AEB540A6FB6ULL, "229364602.ogg"},
	{0xF3C021D4C287BE3EULL, "6102813_f.fxe"},
	{0xF3C061318A18BB8DULL, "hlac_lt_0_lowtown_a2_-2_1z.dds"},
	{0xF3C06C9F0F8BA23DULL, "vxw_bonepitdestroyed_a3_1ay_0_0.vlm"},
	{0xF3C0C749D67A6BA0ULL, "422048008.ogg"},
	{0xF3C1DD4DA1B6865DULL, "vxlt_0_lowtown_a2_n_c_e5w_0_0.vlm"},
	{0xF3C1FD2600AA0956ULL, "dae000_mother6035927.cl"},
	{0xF3C2916C629C28BCULL, "vxht_mansion2_2by_0_0.vlm"},
	{0xF3C2AB5D5DDF80BDULL, "321897984.ogg"},
	{0xF3C2C3D3316414D1ULL, "751059929.ogg"},
	{0xF3C2CEB29A44FBEDULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_ear06a.phy"},
	{0xF3C2D617E7209B62ULL, "vxw_bonepitdestroyed_a3_26p_0_0.vlm"},
	{0xF3C31C9E49717768ULL, "269398254.ogg"},
	{0xF3C36311D82C325DULL, "art\\vfx\\combat\\vfx_m_sta_nat_rng_c.impa.evt"},
	{0xF3C37AAB1AF29881ULL, "935604345.ogg"},
	{0xF3C4068BD90A608CULL, "vxlt_hangedman_4s_0_0.vlm"},
	{0xF3C434A1E012E1C1ULL, "6084411_f.fxe"},
	{0xF3C486253FBBE10DULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_wk_p_rf.gad"},
	{0xF3C5059F9357A58DULL, "art\\characters\\creatures\\mubari\\animation\\com_growl.gad"},
	{0xF3C514B78FF3FABBULL, "6112170_f.fxe"},
	{0xF3C51BB1D20A30EBULL, "71341829.ogg"},
	{0xF3C5244A75C22696ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolegibsgolem_demonhead_mindblast.gad"},
	{0xF3C55078CE17EFB7ULL, "6081390_m.fxe"},
	{0xF3C5C5A8B8010541ULL, "vxlt_0_lowtown_docks_a3_n_cum_0_0.vlm"},
	{0xF3C5CC1CCBD88D97ULL, "w_bonepitmining_a2_205_0.msh"},
	{0xF3C5DE1CCBD8ACFAULL, "w_bonepitmining_a2_205_0.mao"},
	{0xF3C5E21CCBD8B309ULL, "w_bonepitmining_a2_205_0.mmh"},
	{0xF3C600F48325FEA6ULL, "789074627.ogg"},
	{0xF3C6408176041BF4ULL, "llac_ca_cavern1_haunted_24421_7.dds"},
	{0xF3C6EF6EBAF51BA0ULL, "vxlt_0_lowtown_a2_ems_0_0.vlm"},
	{0xF3C6F73AAA64F7AFULL, "22115729.ogg"},
	{0xF3C703A7042D4C31ULL, "656343569.ogg"},
	{0xF3C71CD03A55ACB2ULL, "ht_0_hightownexplosion_n_11_0.rml"},
	{0xF3C71F4E9D49BAE6ULL, "vxht_0_hightown_c3p_0_1.vlm"},
	{0xF3C7CE9E482B973EULL, "vrc321_varric6082696.cl"},
	{0xF3C7D40AB48DA4E6ULL, "581723592.ogg"},
	{0xF3C87D1AB4405174ULL, "zz_dae_debug6029923.cl"},
	{0xF3C87E90B15613A2ULL, "6106238_m.fxe"},
	{0xF3C90464E866F41AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothe.msh"},
	{0xF3C91264E8670C53ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothe.mao"},
	{0xF3C91664E8671360ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_paintbrothe.mmh"},
	{0xF3C9292AF7F745DCULL, "vxht_keep_n_55p_0_0.vlm"},
	{0xF3C9407C4941F247ULL, "vxht_0_hightown_a3_c7e_0_0.vlm"},
	{0xF3C967300127B79AULL, "hlac_lt_0_lowtown_docks_0_6x.dds"},
	{0xF3C9C58D1A299CD5ULL, "6167434_m.fxe"},
	{0xF3C9CC2B00EB16DCULL, "6034663_m.fxe"},
	{0xF3C9F0A8A822CC2BULL, "6097902_m.fxe"},
	{0xF3CA4E2E948EB111ULL, "ico_key.dds"},
	{0xF3CAF7DEF2E5B5F9ULL, "505438279.ogg"},
	{0xF3CBDBC64FE7942CULL, "48948636.ogg"},
	{0xF3CC3F4A75C873F5ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolegibsgolem_demonhead_mindblast.evt"},
	{0xF3CC6A044FB894CBULL, "vxlt_0_lowtown_a2_n_e88_0_0.vlm"},
	{0xF3CCB1769441CE68ULL, "follower_banter6136678.cl"},
	{0xF3CCCF38D967FD72ULL, "hlac_ht_mansion2b_818_15x.dds"},
	{0xF3CD4990B98B4795ULL, "vxlt_0_lowtown_a2_n_c_hkt_0_0.vlm"},
	{0xF3CDC073056D64FFULL, "hlac_ht_mansion_n_2995_0z.dds"},
	{0xF3CDC7E2DE6260A5ULL, "vxlt_gallowstemplar_5r4_0_0.vlm"},
	{0xF3CE1DD03A5BCD46ULL, "ht_0_hightownexplosion_n_11_0.phy"},
	{0xF3CE30E99CD6B053ULL, "pro00fl_carver_stunt.utc"},
	{0xF3CE41034D2E175DULL, "gen00fl_varric6019885.cl"},
	{0xF3CE65DD11A5C5EFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile.mao"},
	{0xF3CE71DD11A5D98CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile.mmh"},
	{0xF3CE77DD11A5E47EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\trap_floortile.msh"},
	{0xF3CE86017BECE226ULL, "6090454_m.fxe"},
	{0xF3CEDE7C0181BABDULL, "vxw_bonepitdestroyed_a3_1jx_0_0.vlm"},
	{0xF3D01D55CFD37A1FULL, "236346782.ogg"},
	{0xF3D05E8E1190CC52ULL, "vxw_encounter2_fx_0_0.vlm"},
	{0xF3D0AE4547E5AC36ULL, "245208124.ogg"},
	{0xF3D147F4D916B3E2ULL, "vxw_bonepitdestroyed_a3_21q_0_0.vlm"},
	{0xF3D1E5AF7082C69EULL, "ht_0_hightown_a3_34_0.anb"},
	{0xF3D2CCF017676676ULL, "llac_lt_0_lowtown_a2_131071_2.dds"},
	{0xF3D3B670CDFF6C4BULL, "vxca_caverns1_mines_j5f_0_0.vlm"},
	{0xF3D4FD6166559224ULL, "qun120cr_javaris6043981.cl"},
	{0xF3D53FB5473D08D3ULL, "6106648_m.fxe"},
	{0xF3D581BEB06340F9ULL, "6069871_m.fxe"},
	{0xF3D593939F4385FFULL, "vxw_sundermount_1ze_0_0.vlm"},
	{0xF3D59BA7F47AE803ULL, "6096747_m.fxe"},
	{0xF3D5A9D73A00DB9BULL, "vxht_0_hightown_a2_9za_0_0.vlm"},
	{0xF3D5DB19E3BA1D86ULL, "firewhispy.mao"},
	{0xF3D60709981EB812ULL, "vxlt_0_lowtown_docks_a3_d2k_0_0.vlm"},
	{0xF3D60E31227E7FC0ULL, "zz_e_roguevshighdragon.cut"},
	{0xF3D61AB1A4DF6676ULL, "one100dg_kelder6055594.cl"},
	{0xF3D6384F6C915AEAULL, "104235405.ogg"},
	{0xF3D68517540DDB96ULL, "6135333_m.fxe"},
	{0xF3D6A8D099A08343ULL, "str_cyc.mao"},
	{0xF3D7009DF9DE1B09ULL, "401762386.ogg"},
	{0xF3D74CCE01DC75B2ULL, "6026240_m.fxe"},
	{0xF3D7674B308D4C29ULL, "vxca_caverns1_i1z_0_0.vlm"},
	{0xF3D785A51FA9008FULL, "vxw_woundedcoast_y5_0_0.vlm"},
	{0xF3D7F0CE22355BF1ULL, "zz_mcr_debug6088341.cl"},
	{0xF3D898F9D989E19FULL, "654173450.ogg"},
	{0xF3D8DB72168A2D87ULL, "vxht_chantry_n_32x_0_0.vlm"},
	{0xF3D9454B93F06C19ULL, "45233819.ogg"},
	{0xF3D95EC9CF49A470ULL, "151786835.ogg"},
	{0xF3D9AFF81FB8E14BULL, "atl_itemupgrad_dxt1_00.dds"},
	{0xF3D9D5DD464DE3EDULL, "6027392_m.fxe"},
	{0xF3DA787BAD7FFEFAULL, "w_blightlands_156_0.phy"},
	{0xF3DA8103F04350B1ULL, "708592855.ogg"},
	{0xF3DACE7BF20F2C6BULL, "follower_banter6170116.cl"},
	{0xF3DACF756689B62CULL, "6122462_f.fxe"},
	{0xF3DAF5B935311001ULL, "214855828.ogg"},
	{0xF3DAFC0C417D3E8EULL, "zz_qun_debug6028270.cl"},
	{0xF3DB08EB586ED2ECULL, "vxdr_0_deeproads_a25_0_0.vlm"},
	{0xF3DBAAD3E46D4E61ULL, "gen00fl_sebastian6174880.cl"},
	{0xF3DBAB8E2D3721F3ULL, "dm_stumps.mor"},
	{0xF3DBD76DF02875EAULL, "425186413.ogg"},
	{0xF3DCC9E80F715A80ULL, "mag311_nyssa6042399.cl"},
	{0xF3DD48C12640D31FULL, "762135911.ogg"},
	{0xF3DD5058F377D33DULL, "vxw_bonepitdestroyed_a3_1ow_0_0.vlm"},
	{0xF3DE0AB732272A5BULL, "vxht_0_hightown_a3_a6p_0_0.vlm"},
	{0xF3DE590EEB2A47BEULL, "follower_banter6147416.cl"},
	{0xF3DE80FF4DA80A2EULL, "6196734_m.fxe"},
	{0xF3DEBA26BC0E61BAULL, "art\\characters\\playercharacter\\humanmale\\hlf_medi.mmh"},
	{0xF3DECC26BC0E8000ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medi.msh"},
	{0xF3DED06B0386E4D2ULL, "vxw_encounter2_ay_0_0.vlm"},
	{0xF3DFA258C1409433ULL, "vxlt_0_lowtown_dow_0_0.vlm"},
	{0xF3DFAE9426B2A1A5ULL, "544107635.ogg"},
	{0xF3DFB9FC083ED64BULL, "mag100_cullen6082342.cl"},
	{0xF3E029E4F47620F0ULL, "532969530.ogg"},
	{0xF3E114E88C8F36F2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_dismiss.gad"},
	{0xF3E121439AD1D830ULL, "genip_cin_wine.utp"},
	{0xF3E1C6DE2EAF4F30ULL, "hlac_ca_cavern1_haunted_24424_28x.dds"},
	{0xF3E36ED6F06458AFULL, "611936791.ogg"},
	{0xF3E3A44028C65A3FULL, "716125638.ogg"},
	{0xF3E40899AAADA2F9ULL, "597528728.ogg"},
	{0xF3E42DAA6863CD6FULL, "191898210.ogg"},
	{0xF3E46CE69511B18CULL, "vxlt_0_lowtown_n_ewn_0_0.vlm"},
	{0xF3E4BE29043371E1ULL, "vxlt_0_lowtown_n_fib_0_0.vlm"},
	{0xF3E4DD9DDFB84885ULL, "45065348.ogg"},
	{0xF3E53C1A9DB4F9E4ULL, "6208899_m.fxe"},
	{0xF3E5622FA511B636ULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.cess.evt"},
	{0xF3E59FF14DE59C6BULL, "mer320_hubert6089686.cl"},
	{0xF3E5AE6BC7DF78DBULL, "hlac_ht_0_hightown_a3_0_2y.dds"},
	{0xF3E5F782119F190FULL, "vxw_woundedcoast_v4_0_0.vlm"},
	{0xF3E817E88C955A29ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_dismiss.evt"},
	{0xF3E82DC9B8509C81ULL, "6143523_f.fxe"},
	{0xF3E85ED44254C51DULL, "306676854.ogg"},
	{0xF3E8FE8FB1744E32ULL, "718884680.ogg"},
	{0xF3E9B4449C5D869FULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_zbm_0d.dds"},
	{0xF3E9FB1A308381F8ULL, "333464689.ogg"},
	{0xF3EA0E516FE61D2AULL, "6175166_m.fxe"},
	{0xF3EA3D54CD4AEE51ULL, "66081088.ogg"},
	{0xF3EB0CCB7808E56CULL, "10938926.ogg"},
	{0xF3EBB5CC6906B37FULL, "6175798_m.fxe"},
	{0xF3EC090C024D136CULL, "hf_ave3_insurgent3.mor"},
	{0xF3EC1CF4B9167DE3ULL, "66211795.ogg"},
	{0xF3EC4D5196800676ULL, "6151353_m.fxe"},
	{0xF3EC592FA517C5C1ULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.cess.gad"},
	{0xF3EC834B459F95C3ULL, "vxht_mansion2_2wl_0_1.vlm"},
	{0xF3ECAAC1CDA7BDE7ULL, "129852882.ogg"},
	{0xF3ED28702F604CDDULL, "6175716_m.fxe"},
	{0xF3ED4247F57CFD52ULL, "vxw_encounter2_lz_0_0.vlm"},
	{0xF3ED56045D35DD42ULL, "421452604.ogg"},
	{0xF3ED602D99E3975FULL, "vxlt_0_lowtown_a2_eux_0_0.vlm"},
	{0xF3ED74845CC81E6AULL, "ss_frozen.cut"},
	{0xF3EE46DEC5881DC8ULL, "vxlt_0_lowtown_n_fig_0_0.vlm"},
	{0xF3EE7867BA91576BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\powerattack.ani"},
	{0xF3EED35FE284B4F9ULL, "con_ico_aggressive.dds"},
	{0xF3EF0B5796DBF751ULL, "vxdr_0_deeproads_b4d_0_1.vlm"},
	{0xF3EF0F17F712A2C7ULL, "130953004.ogg"},
	{0xF3EF51E5642CFF8FULL, "hlac_lt_darktown_-1_83z.dds"},
	{0xF3EF63EF51031F53ULL, "vxlt_0_lowtown_a3_n_c_hm5_0_0.vlm"},
	{0xF3EFBF586EC7B7D4ULL, "vxw_encounter2_du_0_1.vlm"},
	{0xF3F02FD692EA8FD8ULL, "6124023_m.fxe"},
	{0xF3F0FA8C272E559BULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb106a_0s.dds"},
	{0xF3F1B85BB62FA1E7ULL, "vxht_0_hightown_a2_8pn_0_0.vlm"},
	{0xF3F2AD6A345A8E7EULL, "dae000_mother6135357.cl"},
	{0xF3F2FDB3441964BBULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_appr.ani"},
	{0xF3F35A2FA51DE66FULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.cess.ani"},
	{0xF3F400262032519BULL, "dre000_varric_rockwraith.stg"},
	{0xF3F45FD50A20B32BULL, "hlac_ca_caverns1_24424_11x.dds"},
	{0xF3F4695F0395318FULL, "vxw_woundedcoast_s7_0_0.vlm"},
	{0xF3F4C208F8272A78ULL, "6050917_m.fxe"},
	{0xF3F4E188D2AAA513ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_lying_on_ground.anb"},
	{0xF3F56EE21C7F8C48ULL, "168713464.ogg"},
	{0xF3F5758D9629F744ULL, "vxlt_shop_6z_0_0.vlm"},
	{0xF3F5FFE88CA16E98ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_dismiss.ani"},
	{0xF3F6908FA345EF8DULL, "gen00fl_fenris6103320.cl"},
	{0xF3F696FBE8C82207ULL, "vxca_cavern1_haunted_j7d_0_0.vlm"},
	{0xF3F7995F71B07BCDULL, "235793846.ogg"},
	{0xF3F978A91676D558ULL, "hlac_w_blightlands_65534_0x.dds"},
	{0xF3F98B317E574BB7ULL, "88063626.ogg"},
	{0xF3F999C4D4AE7BCEULL, "vxw_bonepit_1qn_0_0.vlm"},
	{0xF3F9A18194805B50ULL, "6083023_m.fxe"},
	{0xF3F9BF56CFF9DB92ULL, "6020398_m.fxe"},
	{0xF3F9D49444DF57FAULL, "vxlt_gallowscourtyard_741_0_0.vlm"},
	{0xF3F9F48E4FD26E40ULL, "6153634_m.fxe"},
	{0xF3FA01486E210C42ULL, "hlac_lt_0_lowtown_a2_n_c_-131073_0x.dds"},
	{0xF3FA81A9D6B28B7EULL, "qun221_petrice6083788.cl"},
	{0xF3FA84887C83E6D7ULL, "552876091.ogg"},
	{0xF3FAD3D9B0326173ULL, "6102558_m.fxe"},
	{0xF3FB1BE58ACB4CA5ULL, "6108792_m.fxe"},
	{0xF3FB6067BA9BB8E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\powerattack.evt"},
	{0xF3FB7367C6DF8AF9ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_sb101a_0n.dds"},
	{0xF3FB926BC822C2F1ULL, "gen00fl_sebastian6189660.cl"},
	{0xF3FB981C2D65DB9EULL, "t1_mub_pu1.tnt"},
	{0xF3FBAD2A40B4D4D4ULL, "vxw_encounter2_7m_0_0.vlm"},
	{0xF3FC01259B66BB46ULL, "6110674_f.fxe"},
	{0xF3FC08DA206B6213ULL, "6036482_m.fxe"},
	{0xF3FC3C2839295575ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupc.phy"},
	{0xF3FC43B9984C1E00ULL, "z_dae_ai_test.ncs"},
	{0xF3FC8FBA4824FAEBULL, "qun231_varnell6077154.cl"},
	{0xF3FCB07BF137A49CULL, "zz_isa_debug6133948.cl"},
	{0xF3FCF5FF0BD22F1FULL, "zz_dae_debug6029074.cl"},
	{0xF3FD0D0FA9E6D0B3ULL, "vxw_woundedcoast_hb_0_0.vlm"},
	{0xF3FD5326760DE64EULL, "lla_ht_0_hightown_n_-65537_0.dds"},
	{0xF3FE2DD471823374ULL, "6150269_f.fxe"},
	{0xF3FE3B30CD7BF04CULL, "184490291.ogg"},
	{0xF3FE4AF3F0D0C47CULL, "mcr371_hostile_aveline6132487.cl"},
	{0xF3FE87CE4B9E9AF1ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_006a_0s.dds"},
	{0xF3FE8E102D4DE502ULL, "vxca_caverns1_eia_0_0.vlm"},
	{0xF3FF7ADD238E22D8ULL, "vxca_cavern1_haunted_im2_0_0.vlm"},
	{0xF3FFBFFE68C1D084ULL, "6108436_m.fxe"},
	{0xF3FFECE4ABA40AC2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_left.gad"},
	{0xF40005B34423FCB2ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_appr.evt"},
	{0xF400498EB938819BULL, "llac_w_sundermount_262147_0.dds"},
	{0xF40051FF81560B86ULL, "isa211_isabela6202218.cl"},
	{0xF40085495385FD27ULL, "gen00fl_fenris6108051.cl"},
	{0xF400A54C18148FD3ULL, "vxw_encounter2_ra_0_0.vlm"},
	{0xF40150A078FDE28CULL, "vxlt_0_lowtown_n_eml_0_0.vlm"},
	{0xF401563DD6C49556ULL, "vxlt_blackemporium_15s_0_0.vlm"},
	{0xF401D55A2F7C7601ULL, "319404390.ogg"},
	{0xF4024CB2678C2059ULL, "hlac_w_bonepitdestroyed_a3_327685_0z.dds"},
	{0xF4025F67BAA1D5F5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basestaff\\combat\\powerattack.gad"},
	{0xF4027533DEE084C2ULL, "vxht_0_hightown_ama_0_0.vlm"},
	{0xF403081945D6BCDDULL, "vxlt_gamlenshouse_n_bk_0_0.vlm"},
	{0xF4032709B500430BULL, "vxw_encounter2_camp_pk_0_0.vlm"},
	{0xF403D926BC2DB2C5ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medi.phy"},
	{0xF403EF669B15CF36ULL, "hlac_ca_cavern1_haunted_24421_4y.dds"},
	{0xF403F5B0A3DFB823ULL, "w_blightlands_56_0.rml"},
	{0xF404D6E4573A0CB7ULL, "14612061.ogg"},
	{0xF404EE75F2028ECEULL, "vxw_woundedcoast_vy_0_0.vlm"},
	{0xF404F3DE6C7FCF6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_crouch_pray.anb"},
	{0xF405BF687D669FA4ULL, "6037345_m.fxe"},
	{0xF4062426D90DF7ADULL, "vxw_bonepitmining_a2_6y_0_0.vlm"},
	{0xF4064A65D6489861ULL, "97142165.ogg"},
	{0xF4064C5850EBAE06ULL, "zz_followers_banter6159567.cl"},
	{0xF406562BF0BB1414ULL, "6175014_m.fxe"},
	{0xF4068064D7A1B678ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\blendtree\\trooper_overlay.anb"},
	{0xF406E6480D1E16E3ULL, "vxht_keep_5gr_0_0.vlm"},
	{0xF4070FE4ABAA64B9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_left.evt"},
	{0xF40724B3442A5045ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_appr.gad"},
	{0xF4072C8AE6F88CE3ULL, "art\\vfx\\combat\\vfx_glb_ele_e_minor_c.impa.evt"},
	{0xF4076F9C08ED6605ULL, "540021108.ogg"},
	{0xF4078A01D1DCE88DULL, "gen00fl_aveline6114017.cl"},
	{0xF407A3A0AC0EE099ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_action.ani"},
	{0xF407B766A5DACF79ULL, "283684921.ogg"},
	{0xF407C8B22D9E78DAULL, "vxht_keep_n_8as_0_0.vlm"},
	{0xF4080BA1C6A4944EULL, "vxw_bonepit_1zo_0_0.vlm"},
	{0xF408467136D5707AULL, "vxlt_gallowscourtyard_792_0_0.vlm"},
	{0xF408E2C63B9D2904ULL, "743030533.ogg"},
	{0xF40975AC4E0FB162ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mfollower2_scaredmage.gad"},
	{0xF409B29E925BE176ULL, "cli600cr_hurlock_melee.utc"},
	{0xF409E69920B0C3E8ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bloomdesk2_sa.dds"},
	{0xF40A1100B35B00ACULL, "hlac_lt_0_lowtown_1_4x.dds"},
	{0xF40A175C316244E3ULL, "mag101_grace6035357.cl"},
	{0xF40A1F0732AAED54ULL, "vxw_encounter2_2r_0_0.vlm"},
	{0xF40A2601D9692E52ULL, "vxw_encounter2_rd_0_0.vlm"},
	{0xF40AA2A55E1E8AD6ULL, "673886556.ogg"},
	{0xF40B2A98A9744EC8ULL, "vxlt_0_lowtown_n_fge_0_0.vlm"},
	{0xF40B2C056A4A379AULL, "art\\characters\\creatures\\stonegolem\\animation\\p_wkf.gad"},
	{0xF40B3DDEDDD9FD6EULL, "vxw_encounter2_camp_8r_0_0.vlm"},
	{0xF40BE270FE9870E6ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_newmetal_sa.dds"},
	{0xF40C2C098DA5DE94ULL, "6096616_m.fxe"},
	{0xF40C434F24EE3BF0ULL, "gen_im_arm_cht_nobc.uti"},
	{0xF40C5EFB6CF729BDULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmred.phy"},
	{0xF40C60AC5D0FEC15ULL, "vxlt_hangedman_n_26a_0_0.vlm"},
	{0xF40C6275F73BE647ULL, "llac_lt_0_lowtown_a3_n_c_131072_0.dds"},
	{0xF40C910879C0BE02ULL, "hlac_lt_gallowstemplar_f_2391_8x.dds"},
	{0xF40CADE1D1374869ULL, "6141104_m.fxe"},
	{0xF40CEEDC42D4CFD5ULL, "vxlt_0_lowtown_docks_a3_aps_0_0.vlm"},
	{0xF40D143D136F77F4ULL, "vxht_brothel_n_4cu_0_1.vlm"},
	{0xF40DCAF618417873ULL, "fen110_fenris6117968.cl"},
	{0xF40DECBA15843B58ULL, "vxca_cavern1_haunted_ij3_0_0.vlm"},
	{0xF40E058AE6FE6964ULL, "art\\vfx\\combat\\vfx_glb_ele_e_minor_c.impa.gad"},
	{0xF40E2617DAAC63A2ULL, "132815666.ogg"},
	{0xF40E7EA0AC14C0DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_action.gad"},
	{0xF40E904D5FABD43AULL, "art\\cinematics\\_crit path\\followerintros\\cs_wench_isa_intro.gad"},
	{0xF40EE0A637566501ULL, "vxdr_primevalentrance_16n_0_0.vlm"},
	{0xF40EF1393EB51775ULL, "82487848.ogg"},
	{0xF40F237B0C3BC515ULL, "737552439.ogg"},
	{0xF40F5C3A1D921925ULL, "464756566.ogg"},
	{0xF40F77A936A2F925ULL, "84381157.ogg"},
	{0xF40F83D0167E7A7BULL, "hlac_lt_gallowscourtyard_n_-65536_1y.dds"},
	{0xF40F84C7D7D89901ULL, "226576867.ogg"},
	{0xF40FA482816F279FULL, "609406796.ogg"},
	{0xF4105D1FB8144205ULL, "clf260_harley6069661.cl"},
	{0xF4106A1FFD853115ULL, "follower_banter6143376.cl"},
	{0xF4107454E23E2BCCULL, "dae150_radha6148930.cl"},
	{0xF411774BC860CCD5ULL, "llac_lt_0_lowtown_docks_a3_-65537_1.dds"},
	{0xF41198E6A6F65B8BULL, "vxw_encounter2_camp_kj_0_0.vlm"},
	{0xF411D368FCDE6BD1ULL, "755664088.ogg"},
	{0xF411F35437E72605ULL, "636769853.ogg"},
	{0xF4122F056A505BB1ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_wkf.evt"},
	{0xF4128F6FD4D8B551ULL, "follower_banter6141752.cl"},
	{0xF41295C5E39729E0ULL, "6120954_f.fxe"},
	{0xF4132FCECB4E200DULL, "573307860.ogg"},
	{0xF4136052E3F8A74EULL, "vxw_woundedcoast_sz_0_0.vlm"},
	{0xF413EE5B558B9E94ULL, "sib131im_key_to_cellar.uti"},
	{0xF41417E4ABB4FCA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\right_hand_gesture_left.ani"},
	{0xF4146B9C900F1A01ULL, "qcr000_arishok6060201.cl"},
	{0xF4150D9FE9F1C1C8ULL, "zz_seb_debug6189579.cl"},
	{0xF415148AE704A142ULL, "art\\vfx\\combat\\vfx_glb_ele_e_minor_c.impa.ani"},
	{0xF4154F202E0CD202ULL, "fx_lightning01.mao"},
	{0xF4156F29784679F1ULL, "vxht_keep_n_60i_0_0.vlm"},
	{0xF4159BA0AC1B10C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\fake_meredith_death_action.evt"},
	{0xF415CA6D967EAAD3ULL, "plt_one200pt_herb_tattoo.plo"},
	{0xF416CD20EB5FF31AULL, "w_sundermount_213_0.phy"},
	{0xF416D8011E27DA4DULL, "art\\levels\\races\\natural\\blightlands\\props\\_textures\\prp_blightrock_n.dds"},
	{0xF4178430C03CAE5CULL, "6049735_m.fxe"},
	{0xF417BCF9AAEAB8DCULL, "vxht_0_hightown_n_aux_0_0.vlm"},
	{0xF418007DDD8773DAULL, "lgt220cr_mag_apost_apost1.utc"},
	{0xF41844BA546F2688ULL, "vxca_cavern1_haunted_gko_0_1.vlm"},
	{0xF418DD2347DE06DAULL, "6091699_m.fxe"},
	{0xF418E4404EE53A54ULL, "vxw_blightlands_ad_0_0.vlm"},
	{0xF419ABEC53DE381DULL, "vxw_bonepitdestroyed_a3_nk_0_0.vlm"},
	{0xF419AFBBCFD015EEULL, "vxw_encounter2_camp_3s_0_0.vlm"},
	{0xF419F0C98DD301B3ULL, "vxw_woundedcoast_bd_0_0.vlm"},
	{0xF41AAC058CCE40FEULL, "mag360_samson6044925.cl"},
	{0xF41AB164D6209B8EULL, "vxlt_0_lowtown_n_gc6_0_0.vlm"},
	{0xF41B2D25559726FEULL, "809173895.ogg"},
	{0xF41B9831687C2ADEULL, "hlac_ca_caverns2_5802_70z.dds"},
	{0xF41C72CCE96C0573ULL, "555260714.ogg"},
	{0xF41CC96AC998654CULL, "mcr310_argument6075796.cl"},
	{0xF41D53FB6D058A29ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmred.mao"},
	{0xF41D5FFB6D059EC2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmred.mmh"},
	{0xF41D61FB6D05A278ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_cratesmred.msh"},
	{0xF41DA0AC4E20A348ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\cs_mfollower2_scaredmage.ani"},
	{0xF41E059E82E1719BULL, "mag360_alain6065046.cl"},
	{0xF41F36C53DDF5E8DULL, "gen00fl_varric6076011.cl"},
	{0xF41F621ED874B962ULL, "6030467_m.fxe"},
	{0xF41FF0AB0BB6FCCBULL, "m_dre180ar_keep_fake.wbd"},
	{0xF4200AC398EC740BULL, "vxw_encounter2_camp_fm_0_0.vlm"},
	{0xF42047056A5CC180ULL, "art\\characters\\creatures\\stonegolem\\animation\\p_wkf.ani"},
	{0xF42117614AD183D0ULL, "hlac_w_sundermount_393213_0z.dds"},
	{0xF42272D1DEFA5A2DULL, "hlac_ca_cavern1_haunted_24426_6y.dds"},
	{0xF4227B48CE0AC9EEULL, "6137619_f.fxe"},
	{0xF422AB4D5FBCAB20ULL, "art\\cinematics\\_crit path\\followerintros\\cs_wench_isa_intro.ani"},
	{0xF422DE8E29B2F20BULL, "172593864.ogg"},
	{0xF422F5D1FC2E264BULL, "dae000_pc_home_ambients6208544.cl"},
	{0xF42307E0BCFA28ADULL, "vxw_bonepitmining_a2_8w_0_0.vlm"},
	{0xF423E4573215E6E8ULL, "follower_banter6156218.cl"},
	{0xF4242169DFCB303FULL, "140822639.ogg"},
	{0xF4246DE4FA3AD1E9ULL, "6108925_m.fxe"},
	{0xF42475810FA9BAF0ULL, "vxlt_warehouse_19i_0_0.vlm"},
	{0xF4247E8744FD68D7ULL, "6087081_f.fxe"},
	{0xF4249A1C91A924EDULL, "hlac_w_bonepitdestroyed_a3_458756_0x.dds"},
	{0xF4250AE0DE11F718ULL, "6150697_m.fxe"},
	{0xF42511EB4BF58597ULL, "vxw_encounter3_bf_0_0.vlm"},
	{0xF425F61FFF0882CBULL, "hlac_ht_0_hightown_n_65535_2y.dds"},
	{0xF42614F4F7161D14ULL, "zz_isa_debug6175883.cl"},
	{0xF426424B3379F17CULL, "6135753_m.fxe"},
	{0xF42650A9DEA8B8F9ULL, "6106978_m.fxe"},
	{0xF4265B6355D3B0EEULL, "57721048.ogg"},
	{0xF426782E6D6BD93FULL, "107619162.ogg"},
	{0xF42680A1E1B8ADC6ULL, "gen00fl_fenris6106612.cl"},
	{0xF42698DFD8C28B48ULL, "770798680.ogg"},
	{0xF426AB1BC4F765BBULL, "gen00fl_anders6119482.cl"},
	{0xF426B5B9E562EF00ULL, "gen_im_npc_ammo_quiver_quna.uti"},
	{0xF426F07E73500006ULL, "vxlt_gallowstemplar_f_2j7_0_0.vlm"},
	{0xF426FA583F7131DDULL, "vxlt_0_lowtown_a2_n_c_ehp_0_0.vlm"},
	{0xF42702C116971E54ULL, "vxw_encounter2_8p_0_0.vlm"},
	{0xF42727A161BBF95FULL, "vxlt_0_lowtown_a2_eid_0_0.vlm"},
	{0xF427561D40DB52D4ULL, "vxw_blightlands_je_0_0.vlm"},
	{0xF427D04A9E126691ULL, "gen00fl_isabela6100134.cl"},
	{0xF42805CC9F31B655ULL, "57255308.ogg"},
	{0xF4281BB7E05A0E2BULL, "vxlt_undercityserialkiller_1sw_0_0.vlm"},
	{0xF4284D63F499C509ULL, "519427636.ogg"},
	{0xF42935827558B677ULL, "docks_male_sailor_mob6176672.cl"},
	{0xF42935F7499473E4ULL, "vxlt_gallowscourtyard_6q1_0_0.vlm"},
	{0xF429446640FC1D15ULL, "vxlt_hangedman_n_28o_0_0.vlm"},
	{0xF4298875F7E9DC5FULL, "vxht_keep_n_6gb_0_0.vlm"},
	{0xF42A0CFD7D6A9754ULL, "vxlt_0_lowtown_a3_n_c_bz5_0_0.vlm"},
	{0xF42A51C40681BDB1ULL, "gen00fl_varric6133864.cl"},
	{0xF42AD3B673A60E06ULL, "70923867.ogg"},
	{0xF42BA627024FE839ULL, "233339084.ogg"},
	{0xF42C1A837E22B422ULL, "338227337.ogg"},
	{0xF42C779B64684F6BULL, "267721718.ogg"},
	{0xF42C8015348B6E36ULL, "684435826.ogg"},
	{0xF42CC1DAD6BD3D9BULL, "399885412.ogg"},
	{0xF42D26128D678B58ULL, "6107082_m.fxe"},
	{0xF42D35432977F129ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.cess.ani"},
	{0xF42D5A38B5EFA517ULL, "zz_fol_debug6103799.cl"},
	{0xF42D8953D4928E64ULL, "vxht_mansion2_3cc_0_0.vlm"},
	{0xF42E75BB268CFA57ULL, "vxlt_0_lowtown_a2_n_c_eow_0_0.vlm"},
	{0xF42E86E0ABDA4689ULL, "23309156.ogg"},
	{0xF42EAB993CF5B1D6ULL, "415313478.ogg"},
	{0xF42F51215E0EB364ULL, "vxlt_gallowscourtyard_n_7eo_0_0.vlm"},
	{0xF42FE3175C0EBCB2ULL, "vxht_0_hightown_a3_b9g_0_0.vlm"},
	{0xF42FEB2ACD22C493ULL, "pro000_varric.bnk"},
	{0xF4302FAAAD79E8BCULL, "rdr111dg_assistant6016633.cl"},
	{0xF4304AB1615D8C99ULL, "vxca_caverns1_hvn_0_0.vlm"},
	{0xF4306055D1223F13ULL, "609184915.ogg"},
	{0xF430C982C257F902ULL, "vxw_woundedcoast_6u_0_0.vlm"},
	{0xF430D254DF399103ULL, "crafting_recipes.gda"},
	{0xF4310190DBEAA992ULL, "vxlt_undercityencounter_1jy_0_0.vlm"},
	{0xF4310363ABBD5933ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_ian.ani"},
	{0xF4314C9640DB2819ULL, "vxht_keep_5l1_0_0.vlm"},
	{0xF43157E5100AD516ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfish.phy"},
	{0xF4317D2FBE0400EFULL, "309224742.ogg"},
	{0xF4319C89353DA497ULL, "vxht_mansion2_30y_0_0.vlm"},
	{0xF4325AC1F78792B7ULL, "720914765.ogg"},
	{0xF432BACF31520073ULL, "330024270.ogg"},
	{0xF4330D2ACD2534E3ULL, "pro000_varric.cnv"},
	{0xF43358010758AF9CULL, "dr_primevaltomb_663_lc.mmh"},
	{0xF43398BA25673BD5ULL, "hlac_ht_keep_3732_12x.dds"},
	{0xF433E726A1007FC7ULL, "488780110.ogg"},
	{0xF43416000EF6DA23ULL, "zz_dae_debug6049964.cl"},
	{0xF4343FE6BE87250BULL, "hlac_dr_0_deeproads_5849_7z.dds"},
	{0xF43449C7A5654D11ULL, "vxlt_0_lowtown_a3_n_c_gxw_0_0.vlm"},
	{0xF4345043297E3DABULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.cess.gad"},
	{0xF434EB06D3270969ULL, "vxlt_backalley_n_1y7_0_1.vlm"},
	{0xF435717BADCC9E23ULL, "w_blightlands_156_0.msh"},
	{0xF4357629CA1DAE07ULL, "nat380_party_banter6066173.cl"},
	{0xF4357F7BADCCB67DULL, "w_blightlands_156_0.mmh"},
	{0xF435837BADCCBC8EULL, "w_blightlands_156_0.mao"},
	{0xF435C7FA32D16B54ULL, "vxw_blightlands_oj_0_0.vlm"},
	{0xF436098F1BD7A42BULL, "rdr370ar_finale.lst"},
	{0xF4365BA75EC4A078ULL, "dae000_merchant_access_hf6162989.cl"},
	{0xF4367C1B27EBD41BULL, "character_character_p286.psh.fxo.dx11.assembly"},
	{0xF436A8194B3C8E94ULL, "6206185_m.fxe"},
	{0xF436D48371BF32B3ULL, "vxw_woundedcoast_tf_0_0.vlm"},
	{0xF436E3B0A40AF33AULL, "w_blightlands_56_0.msh"},
	{0xF436EDC72000470CULL, "29826823.ogg"},
	{0xF436F1B0A40B0B73ULL, "w_blightlands_56_0.mao"},
	{0xF436F5B0A40B1180ULL, "w_blightlands_56_0.mmh"},
	{0xF436FC4181EEE5E8ULL, "6206875_m.fxe"},
	{0xF4378914270E6133ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.inactivetoactive.ani"},
	{0xF43796E574AEF4E1ULL, "322318217.ogg"},
	{0xF437D0DC36269B1BULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.dura.evt"},
	{0xF437F19AA8C73C37ULL, "388300269.ogg"},
	{0xF437FA52E9DFF4DFULL, "vxht_keep_n_6bm_0_0.vlm"},
	{0xF438F9F7458576D3ULL, "one120_athenril6088907.cl"},
	{0xF4393FEC78EBD4ACULL, "hlac_ca_caverns2prc_5802_16y.dds"},
	{0xF43A4D432982A410ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_lp_c.cess.evt"},
	{0xF43A6421D99CA90CULL, "mcr371_angry_mage_m6087735.cl"},
	{0xF43AC308AA3C3419ULL, "6098611_m.fxe"},
	{0xF43B0405EB85F92AULL, "lt_0_lowtown_docks_10_0_ll.mmh"},
	{0xF43B40E5101302B7ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfish.msh"},
	{0xF43B52E51013211AULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfish.mao"},
	{0xF43B56E510132829ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfish.mmh"},
	{0xF43BAEB2EC03ACB7ULL, "hlac_ht_bartrand_n_3366_0z.dds"},
	{0xF43C1F535F02ADE7ULL, "hlac_lt_0_lowtown_-131073_2z.dds"},
	{0xF43C2B603EB78579ULL, "sib131_slaver_chest.utp"},
	{0xF43C68D2725A950DULL, "75075872.ogg"},
	{0xF43CF2BC17FB154BULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_follower_.ani"},
	{0xF43D30259B388540ULL, "vxca_caverns2prc_a6x_0_0.vlm"},
	{0xF43D4AB223F5B76FULL, "7475660.ogg"},
	{0xF43D4DCF9E9625FDULL, "gen00fl_fenris6127018.cl"},
	{0xF43D64B418A6ED40ULL, "864736132.ogg"},
	{0xF43DDEFA1C3F2665ULL, "vxlt_hangedman_n_261_0_0.vlm"},
	{0xF43DE0EF236A7D4BULL, "vrc321_party_barks6095788.cl"},
	{0xF43EC9DC362CAE5CULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.dura.gad"},
	{0xF43EF5745417560AULL, "vxlt_0_lowtown_f4a_0_0.vlm"},
	{0xF43F1B24C0954303ULL, "6065815_m.fxe"},
	{0xF43F3B5FB44E1182ULL, "vxw_woundedcoast_3v_0_0.vlm"},
	{0xF43F84850E3E4832ULL, "art\\vfx\\combat\\vfx_r_dwp_rightb_close1_c.mmh"},
	{0xF4400A7486041C4BULL, "6047232.cut"},
	{0xF4400E662733BD17ULL, "vxht_mansion2_37x_0_0.vlm"},
	{0xF44090B19F00D20BULL, "6170645_m.fxe"},
	{0xF4409534A3654985ULL, "971597700.ogg"},
	{0xF440B8F6EDA3A223ULL, "static_rendertodepth_p0.psh.fxo"},
	{0xF440D09461726530ULL, "llac_w_encounter2_327679_0.dds"},
	{0xF441593AF395EBF0ULL, "vxlt_warehouse_13w_0_0.vlm"},
	{0xF4417A0E8C93F641ULL, "6155736_m.fxe"},
	{0xF4417D270542837DULL, "189940275.ogg"},
	{0xF441B46F67541A69ULL, "659792003.ogg"},
	{0xF441D3158E7CF64EULL, "vxw_bonepit_1nc_0_0.vlm"},
	{0xF441E08BE8D4234AULL, "ser221_gascard6165654.cl"},
	{0xF4424B9E07FDC2CDULL, "1030557402.ogg"},
	{0xF44275BA391D1090ULL, "dre000_mother6037119.cl"},
	{0xF4434C2C833C43FCULL, "6060536_m.fxe"},
	{0xF4434F2839657790ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupc.msh"},
	{0xF4435D2839658FAAULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupc.mmh"},
	{0xF4436128396596E1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_blrockgroupc.mao"},
	{0xF4437593881CBAACULL, "6125733_f.fxe"},
	{0xF443787D7EB7B544ULL, "art\\vfx\\models\\vfx_tube_taper.mmh"},
	{0xF4437C7D7EB7BB97ULL, "art\\vfx\\models\\vfx_tube_taper.mao"},
	{0xF4438E7D7EB7DAE6ULL, "art\\vfx\\models\\vfx_tube_taper.msh"},
	{0xF44439D724C783D4ULL, "vxw_blightlands_pk_0_0.vlm"},
	{0xF44466A04563E83BULL, "gen00fl_carver6097267.cl"},
	{0xF4448696C1D2D2DEULL, "445885747.ogg"},
	{0xF444BBB24F35DA60ULL, "787780456.ogg"},
	{0xF444C8DC3631185AULL, "art\\vfx\\combat\\vfx_r_ibl_loop_modal_c.dura.ani"},
	{0xF445297F0DAF1F85ULL, "hlac_lt_darktown_-1_88y.dds"},
	{0xF44534FD29E65891ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.impa.ani"},
	{0xF445466063B54B33ULL, "vxw_woundedcoast_qi_0_0.vlm"},
	{0xF4458114271A916AULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.inactivetoactive.evt"},
	{0xF445B16D96A7A470ULL, "plt_one200pt_herb_tattoo.nss"},
	{0xF445D2AB0BD79899ULL, "m_dre180ar_keep_fake.bnk"},
	{0xF445DA63ABCF6FDDULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_obnoxiousdrunk_1_ian.gad"},
	{0xF4461990FD2AA253ULL, "vxht_brothel_n_6cq_0_0.vlm"},
	{0xF4464E22A52F2A31ULL, "mer141_jansen6031830.cl"},
	{0xF4467C6594FF4ABCULL, "art\\levels\\races\\natural\\kirkcave\\props\\_textures\\prp_bb_d.dds"},
	{0xF446812BDB839190ULL, "hlac_ca_caverns1_mines_24421_4y.dds"},
	{0xF446D070BB524D49ULL, "6201268_m.fxe"},
	{0xF446F3A4D3087DCFULL, "hlac_w_encounter2_camp_-65533_0x.dds"},
	{0xF447237B2CE7C225ULL, "vxca_caverns2_814_0_0.vlm"},
	{0xF44836EA8ADC3BBEULL, "vxht_mansion2_1tb_0_0.vlm"},
	{0xF44838C9F4150623ULL, "follower_banter6156057.cl"},
	{0xF448DD8458B84E34ULL, "vxw_sundermount_272_0_0.vlm"},
	{0xF44911FF336E2E97ULL, "vxht_0_hightown_n_bxo_0_0.vlm"},
	{0xF449922B3A8C969EULL, "dae011_conspiracy_mob6141234.cl"},
	{0xF44B2B220703E02FULL, "vxlt_0_lowtown_docks_a3_cuh_0_0.vlm"},
	{0xF44B40A2F8BDFD1BULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml105a_0s.dds"},
	{0xF44B6C978AA1CD5DULL, "vxlt_gallowstemplar_f_5us_0_1.vlm"},
	{0xF44B99E3E3CEF71AULL, "vxht_brothel_n_5kk_0_0.vlm"},
	{0xF44C22ED4D868196ULL, "lt_0_lowtown_docks_n_185_0.mao"},
	{0xF44C2EED4D869635ULL, "lt_0_lowtown_docks_n_185_0.mmh"},
	{0xF44C30ED4D8699ABULL, "lt_0_lowtown_docks_n_185_0.msh"},
	{0xF44C3FFD29EC8A33ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.impa.gad"},
	{0xF44C41B8A233E032ULL, "vxht_mansion_n_4v6_0_0.vlm"},
	{0xF44C52CD3D5A588BULL, "vxca_cave_226_0_0.vlm"},
	{0xF44C6014272077DDULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\animation\\gallowint_trap.inactivetoactive.gad"},
	{0xF44C69432AA685CEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\8gr_podium.msh"},
	{0xF44C77432AA69D1FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\8gr_podium.mao"},
	{0xF44C83432AA6B1BCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\8gr_podium.mmh"},
	{0xF44C9B47231187C1ULL, "6136209_m.fxe"},
	{0xF44CF1733923BD10ULL, "697004363.ogg"},
	{0xF44EB6F54C3E2A78ULL, "gen_im_trs_prf_junk2_03.uti"},
	{0xF44EC7D3D917D5D1ULL, "vxdr_primevaltomb_7eg_0_2.vlm"},
	{0xF44F9A46BE7F40D2ULL, "vxht_brothel_n_6ct_0_0.vlm"},
	{0xF45009581B15651CULL, "6119430_m.fxe"},
	{0xF4500E42BAF3AC4BULL, "vxlt_warehouse_oa_0_0.vlm"},
	{0xF4502AB2696624BFULL, "318334333.ogg"},
	{0xF45035E97CD4A5B8ULL, "fx_smokebillow01.dds"},
	{0xF4503F4B1929F971ULL, "370162310.ogg"},
	{0xF450A507D9563FF2ULL, "hlac_dr_slavers_3216_4x.dds"},
	{0xF450D9BC180B93D5ULL, "art\\cinematics\\_crit path\\1_deep roads expedition\\dre000_cut_varric_intro\\cs_varricintro_follower_.gad"},
	{0xF4514409831DDEDEULL, "gen00fl_merrill6106968.cl"},
	{0xF452A25D6ECC6299ULL, "6097316_f.fxe"},
	{0xF452A99B1C36E5ACULL, "llac_lt_0_lowtown_docks_65535_2.dds"},
	{0xF452ABB416BD9C54ULL, "vxw_blightlands_uh_0_0.vlm"},
	{0xF453072345A70F57ULL, "hlac_ht_keep_4519_3x.dds"},
	{0xF4532A140517A02FULL, "6115983_m.fxe"},
	{0xF4533CFD29F2A398ULL, "art\\vfx\\combat\\vfx_m_sph_healaura_crust_c.impa.evt"},
	{0xF453956B38076A73ULL, "vxlt_warehouse_n_xu_0_0.vlm"},
	{0xF45442BDE333AD01ULL, "dae200_cullen6140336.cl"},
	{0xF454A619F3059D5BULL, "zz_isa_debug6133170.cl"},
	{0xF455A4EB52E47DBAULL, "dae000_mother6108959.cl"},
	{0xF455B97E986F3279ULL, "art\\characters\\weapons\\mauls\\textures\\w_mal_ml102a_0n.dds"},
	{0xF4561E230BF69EDFULL, "405709637.ogg"},
	{0xF456F092DBE9FC67ULL, "6074981_m.fxe"},
	{0xF4572589697B0A80ULL, "vxdr_primevaltomb_39p_0_0.vlm"},
	{0xF4574F614AAE66B4ULL, "vxw_sundermount_283_0_0.vlm"},
	{0xF457EBCC9AE288B6ULL, "hlac_lt_warehouse_n_65537_4x.dds"},
	{0xF4599CFEF8F9F8AFULL, "vxlt_0_lowtown_docks_a3_cxk_0_0.vlm"},
	{0xF4599EF37087EF1AULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_rubblelarge.msh"},
	{0xF459AC4CFDB38052ULL, "6117785_f.fxe"},
	{0xF459ACF370880753ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_rubblelarge.mao"},
	{0xF459B0F370880E60ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ff_rubblelarge.mmh"},
	{0xF459C2E4C4BFF3A0ULL, "vxht_keep_n_61c_0_0.vlm"},
	{0xF459D46F73E859DDULL, "hlac_ht_0_hightown_a3_65536_2z.dds"},
	{0xF45AB254BB8ECB7EULL, "815274636.ogg"},
	{0xF45AC95795F1904AULL, "vxht_0_hightown_n_c6e_0_0.vlm"},
	{0xF45AC9DA5BC677B9ULL, "294800589.ogg"},
	{0xF45B33219F5A8A50ULL, "vxlt_0_lowtown_docks_a3_b9b_0_0.vlm"},
	{0xF45B3A31842D665AULL, "6119212_m.fxe"},
	{0xF45B50131DB04695ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_002a_0t.dds"},
	{0xF45B5942EFDD4595ULL, "t1_stb_bln.tnt"},
	{0xF45B8CB1D9C7130EULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.impa.evt"},
	{0xF45BC790CE817527ULL, "prm000im_blood_chest.uti"},
	{0xF45C223906F57B6DULL, "372171334.ogg"},
	{0xF45C9FF420DC5844ULL, "one100dg_nabil6033146.cl"},
	{0xF45CA85D6A2A049EULL, "527545275.ogg"},
	{0xF45CCC70C990B3D3ULL, "pro000_captain_wright6004306.cl"},
	{0xF45CD6B13EC71394ULL, "vxht_bartrand_n_690_0_0.vlm"},
	{0xF45CDCA5C35589CBULL, "vxlt_0_lowtown_a2_n_en2_0_0.vlm"},
	{0xF45CDE0FD789DC66ULL, "llac_lt_0_lowtown_a2_n_-2_1.dds"},
	{0xF45D585CFAC348BDULL, "dae000_meredith6080990.cl"},
	{0xF45D752D9933050EULL, "24699947.ogg"},
	{0xF45DA1E90AACA04CULL, "gen00fl_anders6119469.cl"},
	{0xF45DBEDD39856A94ULL, "w_bonepitdestroyed_a3_45_0.rml"},
	{0xF45DEB9F90620BF1ULL, "6170360_m.fxe"},
	{0xF45E2DEABEADF544ULL, "hlac_lt_gallowstemplar_f_2383_0y.dds"},
	{0xF45E3882B2856D5AULL, "art\\vfx\\textures\\vfx_branch_normal.dds"},
	{0xF45F20958871672DULL, "455911309.ogg"},
	{0xF45FACC5EDED7AE0ULL, "6084516_m.fxe"},
	{0xF45FC6183000E135ULL, "vxht_mansion2_2ko_0_0.vlm"},
	{0xF4600664A2EA3D7EULL, "9886874.ogg"},
	{0xF4601E0D2402500EULL, "hlac_w_sundermount_196606_0y.dds"},
	{0xF4609E63DFD8F49EULL, "vxca_cave_14g_0_1.vlm"},
	{0xF46173EFF31E6EC5ULL, "166726214.ogg"},
	{0xF461A06C91F2B6A6ULL, "6131640_f.fxe"},
	{0xF4621A8F1BFCFC48ULL, "rdr370ar_finale.are"},
	{0xF46283B1D9CD22B9ULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.impa.gad"},
	{0xF462F7DF4249759FULL, "536989944.ogg"},
	{0xF462FD4AE116D353ULL, "vxht_brothel_n_6is_0_0.vlm"},
	{0xF46314E1E0F83264ULL, "121957658.ogg"},
	{0xF46325A4D0956FCEULL, "vxca_caverns1_egg_0_0.vlm"},
	{0xF463AB1ACB089082ULL, "vxca_caverns1_e6h_0_0.vlm"},
	{0xF464B26844CAAF81ULL, "gen00fl_dog6115067.cl"},
	{0xF4651AA46EC86CBEULL, "vxht_mansion2_1bd_0_0.vlm"},
	{0xF4653C51217854CDULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamlarge.msh"},
	{0xF46543F6FE4F1E4EULL, "6179315_m.fxe"},
	{0xF4654A5121786C00ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamlarge.mao"},
	{0xF4654E5121787353ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_beamlarge.mmh"},
	{0xF465A158C00EBCF1ULL, "zz_prmpt_debug6200977.cl"},
	{0xF4660CA82B4A0F12ULL, "120820375.ogg"},
	{0xF466827B6E383D79ULL, "hlac_w_sundermount_327680_0z.dds"},
	{0xF4669513A2E2640CULL, "zz_ave_debug6134005.cl"},
	{0xF4674FF4D0A694C9ULL, "6118739_m.fxe"},
	{0xF467C947F4DBEB90ULL, "6154749_m.fxe"},
	{0xF467CB137941EC47ULL, "281270575.ogg"},
	{0xF468335F3AA72356ULL, "art\\characters\\creatures\\mubari\\animation\\wkb.ani"},
	{0xF46883850E614BD9ULL, "art\\vfx\\combat\\vfx_r_dwp_rightb_close1_c.anb"},
	{0xF468CB4D24D83B4DULL, "w_bonepitmining_a2_52_0.mmh"},
	{0xF468CF4D24D8419EULL, "w_bonepitmining_a2_52_0.mao"},
	{0xF468DD4D24D859D3ULL, "w_bonepitmining_a2_52_0.msh"},
	{0xF469298ABB036B3EULL, "68955991.ogg"},
	{0xF4693B3487E7A8CAULL, "vxht_0_hightown_n_c3f_0_0.vlm"},
	{0xF46979CDEB016F8BULL, "6141847_m.fxe"},
	{0xF46986E6EBEF3331ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkback.ani"},
	{0xF46994B1D9D35E07ULL, "art\\vfx\\cinematic\\vfx_fire_vertical_p.impa.ani"},
	{0xF469AC0E76237EDAULL, "6114512_m.fxe"},
	{0xF46A907FF6313391ULL, "382659660.ogg"},
	{0xF46AD4BF196233C2ULL, "982179037.ogg"},
	{0xF46B230EE33C4CF8ULL, "hlac_w_woundedcoast_327683_0x.dds"},
	{0xF46B4C014D4A7861ULL, "m_var221ar_mansion_fake.wbd"},
	{0xF46B96D7000A1837ULL, "fx_stormcloud_small.mao"},
	{0xF46BABB358F11A31ULL, "vxht_mansion_n_53j_0_1.vlm"},
	{0xF46BF5B1DADB28C8ULL, "w_woundedcoast_236_0.msh"},
	{0xF46BFE71E25B53D1ULL, "vxw_bonepitmining_a2_16z_0_0.vlm"},
	{0xF46C07B1DADB4759ULL, "w_woundedcoast_236_0.mao"},
	{0xF46C13B1DADB5BF2ULL, "w_woundedcoast_236_0.mmh"},
	{0xF46C14CF768E7DA0ULL, "904901720.ogg"},
	{0xF46C42E0F1967B74ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\blendtree\\dm_exp_movement.anb"},
	{0xF46C4A5A06514A26ULL, "vxca_caverns2_8oa_0_0.vlm"},
	{0xF46C5B1DC43C8DA3ULL, "8059669.ogg"},
	{0xF46C672206388487ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_right.gad"},
	{0xF46C7E00A26B71D2ULL, "vxht_brothel_n_6iv_0_0.vlm"},
	{0xF46C91E641C39F11ULL, "dae000_bartrand6016755.cl"},
	{0xF46CB7527F1194CDULL, "88823794.ogg"},
	{0xF46CF1FC9EDFDD4BULL, "vxlt_warehouse_uc_0_0.vlm"},
	{0xF46D126464E8274EULL, "41801756.ogg"},
	{0xF46D2FD0D471BD87ULL, "zz_one000_barber6179406.cl"},
	{0xF46D36FC37A614A2ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medd.phy"},
	{0xF46DC84E56D57B41ULL, "vxht_mansion2b_3dk_0_0.vlm"},
	{0xF46E19B1CC9516ECULL, "dae000_arishok6067856.cl"},
	{0xF46E5D42C2552E1FULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_6z.dds"},
	{0xF46E63B3056700A2ULL, "dae000cr_katriela.utc"},
	{0xF46E757FDFC46E1DULL, "hlac_w_bonepitdestroyed_a3_131076_0y.dds"},
	{0xF46ED7281351CFD3ULL, "hlac_lt_0_lowtown_a2_65535_8z.dds"},
	{0xF46EDAA0F09AFC6BULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_left.ani"},
	{0xF46F118D5226B4CAULL, "14378070.ogg"},
	{0xF46F4997C67A2F96ULL, "210539163.ogg"},
	{0xF46F4F11C9229439ULL, "577261904.ogg"},
	{0xF46F57AB5C32DAE2ULL, "6141789_m.fxe"},
	{0xF4707D433433AF47ULL, "hlac_lt_0_lowtown_a2_-2_1x.dds"},
	{0xF47091E6EBF564D3ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkback.gad"},
	{0xF47094887665C7FCULL, "6102801_m.fxe"},
	{0xF470BCDE89018906ULL, "925107240.ogg"},
	{0xF470C5479126B6EAULL, "vxlt_0_lowtown_a3_n_c_f4u_0_0.vlm"},
	{0xF470DEC4CCBA0F15ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_torchwall_n.dds"},
	{0xF4714BBCAE813FD3ULL, "w_bonepit_202_0.phy"},
	{0xF47162432AC59587ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\8gr_podium.phy"},
	{0xF47167F104807104ULL, "vxca_caverns2prc_adv_0_0.vlm"},
	{0xF471B9E99A353256ULL, "vxlt_warehouse_8a_0_0.vlm"},
	{0xF471D5C4249F5079ULL, "12877220.ogg"},
	{0xF471DAE35491A2C7ULL, "6045858_f.fxe"},
	{0xF471F09BA33AE11BULL, "6059892_m.fxe"},
	{0xF472999669ABC501ULL, "90057206.ogg"},
	{0xF472C620EBAE4643ULL, "w_sundermount_213_0.msh"},
	{0xF472CB4D24E08FC6ULL, "w_bonepitmining_a2_52_0.rml"},
	{0xF472D420EBAE5D9DULL, "w_sundermount_213_0.mmh"},
	{0xF472D820EBAE64AEULL, "w_sundermount_213_0.mao"},
	{0xF4734E861440A5A6ULL, "16933567.ogg"},
	{0xF4735C22063E9094ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_right.evt"},
	{0xF473BDD7EF8EF831ULL, "6034914_m.fxe"},
	{0xF473CDB3189731E0ULL, "vxht_bartrand_n_3qu_0_0.vlm"},
	{0xF473E0C01281E77DULL, "vxdr_0_deeproads_cek_0_0.vlm"},
	{0xF473FF6658B9ED27ULL, "vxht_keep_6ga_0_0.vlm"},
	{0xF47438C6C30E8916ULL, "hlac_dr_0_deeproads_5842_12x.dds"},
	{0xF4746FDAC36AB4B5ULL, "6044049_m.fxe"},
	{0xF4748E9BBE139E68ULL, "zz_vrc_debug6200220.cl"},
	{0xF4748FB84D39A3B3ULL, "6119011_m.fxe"},
	{0xF474A0A5CD0C350FULL, "222561872.ogg"},
	{0xF474F56659951CBDULL, "6116997_m.fxe"},
	{0xF474FEF270FA9229ULL, "vxlt_0_lowtown_eub_0_0.vlm"},
	{0xF4753B5F3AB1BB47ULL, "art\\characters\\creatures\\mubari\\animation\\wkb.evt"},
	{0xF4754BB2F12B9137ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_arms_crossed.anb"},
	{0xF475643B8C5B9920ULL, "6062989_m.fxe"},
	{0xF475C79A310D7420ULL, "6161608_m.fxe"},
	{0xF4768EE6EBF9CB38ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\dual\\dualwalkback.evt"},
	{0xF476DB8436394906ULL, "qun221_viscount6084737.cl"},
	{0xF4774BBC8A987CF1ULL, "isa211_isabela6069441.cl"},
	{0xF47781D5E201807AULL, "vxw_bonepitdestroyed_a3_192_0_0.vlm"},
	{0xF477AD1179DDC14AULL, "vxht_0_hightown_n_c8g_0_0.vlm"},
	{0xF478A31195942AB0ULL, "216497560.ogg"},
	{0xF478E90BDA77DCB1ULL, "hlac_w_bonepitmining_a2_327685_0z.dds"},
	{0xF4792B3E3CF888B0ULL, "vxca_cave_2a0_0_0.vlm"},
	{0xF4793F1355B7FED3ULL, "112304086.ogg"},
	{0xF47AE56993FE9A52ULL, "vxlt_0_lowtown_a3_n_c_h29_0_0.vlm"},
	{0xF47B51365CB71F5AULL, "6069425_m.fxe"},
	{0xF47BC2A0F0A55DE2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_left.evt"},
	{0xF47BD8E8363E8CFEULL, "hlac_ca_cavern1_haunted_24422_4x.dds"},
	{0xF47C1DF5E82ABF25ULL, "522287616.ogg"},
	{0xF47C5C5F3AB81190ULL, "art\\characters\\creatures\\mubari\\animation\\wkb.gad"},
	{0xF47C93C043C54C44ULL, "vxlt_gallowscourtyard_7ex_0_0.vlm"},
	{0xF47CB6361ADBA0E2ULL, "vxca_caverns2_5a6_0_0.vlm"},
	{0xF47CFC208053F357ULL, "mag330cr_cricket.utc"},
	{0xF47D1385903D3214ULL, "6104586_m.fxe"},
	{0xF47D863D5274589FULL, "vxht_0_hightown_bc5_0_0.vlm"},
	{0xF47EB1A9E1D9327DULL, "vxht_chantry_2am_0_0.vlm"},
	{0xF47F5888A145AD91ULL, "ht_0_hightown_n_12_0_lc.mmh"},
	{0xF47F7027CB858223ULL, "6045854_f.fxe"},
	{0xF47FEE47E6CCE25BULL, "vxlt_gallowstemplar_n_2j6_0_0.vlm"},
	{0xF4800D2BE9EA98A1ULL, "vxht_brothel_n_6db_0_0.vlm"},
	{0xF48058BB5351AD6AULL, "t3_rob_pl1.tnt"},
	{0xF480E1A788B17113ULL, "vxw_bonepit_276_0_0.vlm"},
	{0xF4815422064AC0CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_right.ani"},
	{0xF48169604A91A8EFULL, "96034173.ogg"},
	{0xF481A6B903C7A401ULL, "mcr370st_assignments.stg"},
	{0xF481FE5E52B49DBEULL, "vxht_mansion2_1hf_0_0.vlm"},
	{0xF48210BF2C611749ULL, "w_bonepitdestroyed_a3_239_0.mao"},
	{0xF4821CBF2C612BE2ULL, "w_bonepitdestroyed_a3_239_0.mmh"},
	{0xF4821EBF2C612E98ULL, "w_bonepitdestroyed_a3_239_0.msh"},
	{0xF4824F6F7F11E002ULL, "vxlt_gamlenshouse_n_n4_0_0.vlm"},
	{0xF4828627EE7BD7D8ULL, "6053372_m.fxe"},
	{0xF482A367319625BBULL, "hlac_lt_0_lowtown_a2_n_c_65538_0z.dds"},
	{0xF482C1A0F0AB7AF5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_hvy_dmg_f_left.gad"},
	{0xF482D8F1D81D483AULL, "gen_im_trs_gen_rng_r0.uti"},
	{0xF482F8AC368CD38EULL, "818396111.ogg"},
	{0xF483192C62FA48AEULL, "vxlt_foundry_n_vl_0_0.vlm"},
	{0xF48370CF62F0AAA9ULL, "vxlt_0_lowtown_epa_0_0.vlm"},
	{0xF4838F81FA4DF441ULL, "vxlt_gallowstemplar_f_31x_0_0.vlm"},
	{0xF483D6187E51B1A0ULL, "6069889_m.fxe"},
	{0xF4843F1299430C20ULL, "608653787.ogg"},
	{0xF48491C1106F8750ULL, "w_blightlands_46_0.mao"},
	{0xF48495C1106F8E63ULL, "w_blightlands_46_0.mmh"},
	{0xF484A3C1106FA5BDULL, "w_blightlands_46_0.msh"},
	{0xF48530BBD6BB55FEULL, "llac_ca_cavern1_haunted_24424_8.dds"},
	{0xF485648E1FD5FFC6ULL, "6016795_m.fxe"},
	{0xF4859784E9858469ULL, "seb000im_leliana_gloves.uti"},
	{0xF485E9234831F3E9ULL, "622652954.ogg"},
	{0xF4860D8C024D8673ULL, "gen_im_arm_glv_hvy_act3_01.uti"},
	{0xF48671360F13FC2EULL, "vxlt_gallowstemplar_f_5rd_0_0.vlm"},
	{0xF4869619E1CF6722ULL, "6130131_f.fxe"},
	{0xF486E0372352C042ULL, "and230st_base_milk.stg"},
	{0xF48711F020355772ULL, "vxw_bonepitmining_a2_1lk_0_0.vlm"},
	{0xF4871ADB3D4A38CCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p.gad"},
	{0xF4875CFC3ED579FAULL, "vxca_cavern1_haunted_in5_0_0.vlm"},
	{0xF4875D3DB855D8BCULL, "vxht_0_hightown_n_cbg_0_0.vlm"},
	{0xF4876B40194C315FULL, "437272635.ogg"},
	{0xF4879D1B2EEEA130ULL, "vxca_cave_2j1_0_0.vlm"},
	{0xF487E67E5ADC0CE7ULL, "vxca_cave_2e7_0_0.vlm"},
	{0xF488167104258D2CULL, "fen210_orana6109778.cl"},
	{0xF4883EE87F844BCDULL, "125099012.ogg"},
	{0xF4887CBD1334FBA0ULL, "165714965.ogg"},
	{0xF4892E13EA3D7B26ULL, "vxca_caverns2_8qc_0_0.vlm"},
	{0xF4895F2B6538ECDDULL, "179552383.ogg"},
	{0xF4896745F4B79602ULL, "hlac_ca_caverns2_5802_38z.dds"},
	{0xF489D6CAF1B11DB8ULL, "129331274.ogg"},
	{0xF489D7F63DAD837AULL, "6074382_m.fxe"},
	{0xF48A00959FBAD299ULL, "vxw_bonepitdestroyed_a3_1gc_0_0.vlm"},
	{0xF48A226E43850550ULL, "6119385_f.fxe"},
	{0xF48B611289B0505BULL, "vxht_0_hightown_a3_ajl_0_0.vlm"},
	{0xF48BD971D6DFFF4FULL, "6113871_m.fxe"},
	{0xF48BF81A446A711FULL, "vxht_0_hightown_bl4_0_0.vlm"},
	{0xF48C444FBBB16AFBULL, "hlac_lt_0_lowtown_a2_n_c_196607_0z.dds"},
	{0xF48C9EA1157C54B8ULL, "fer_ter_rock_001_norm.dds"},
	{0xF48CD7F9E32CAEBCULL, "6123307_m.fxe"},
	{0xF48CE9D5E993E884ULL, "t3_wdg_w12.tnt"},
	{0xF48D2386D3CF4AFDULL, "vxht_chantry_2zl_0_0.vlm"},
	{0xF48D74FDB237C40FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_grandcrackl_1.mao"},
	{0xF48E19DB3D5055CAULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\p.ani"},
	{0xF48EA2ADD51C2EBBULL, "lgt120_orz_bln_official6140195.cl"},
	{0xF48EC98BA11EC535ULL, "ht_0_hightown_62_0_lc.mmh"},
	{0xF48EFD41E996BAC9ULL, "654889177.ogg"},
	{0xF48FBAE7BBE69B2EULL, "dae000ip_crowd_ff8_talk1.utp"},
	{0xF4906E014D69CE63ULL, "m_var221ar_mansion_fake.bnk"},
	{0xF4908ED83BD31CD9ULL, "llac_lt_0_lowtown_docks_n_-2_3.dds"},
	{0xF490914B8AD0C4CFULL, "gen00fl_dog6107703.cl"},
	{0xF4914E6518C5646BULL, "art\\vfx\\creatures\\fxc_ash_proj_p.cess.ani"},
	{0xF491A0602D2B6740ULL, "hlac_lt_0_lowtown_a2_n_-65536_8x.dds"},
	{0xF4923DFAB4AB741EULL, "vxlt_ruinedhouse_n_h6_0_0.vlm"},
	{0xF49298F48C817A9AULL, "6029379_m.fxe"},
	{0xF49299690337C943ULL, "plt_cod_mgc_mageunderground.plo"},
	{0xF492B7DD33A23029ULL, "6182303_m.fxe"},
	{0xF492B97E025CC9B3ULL, "124435541.ogg"},
	{0xF49311FC37C6A556ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medd.mao"},
	{0xF49319A973131483ULL, "vxlt_0_lowtown_a3_n_c_fi3_0_0.vlm"},
	{0xF4931DFC37C6B9F5ULL, "art\\characters\\playercharacter\\humanmale\\hlh_medd.mmh"},
	{0xF4931FFC37C6BD6BULL, "art\\characters\\playercharacter\\humanmale\\hlh_medd.msh"},
	{0xF49341D74686DA0BULL, "hlac_w_blightlands_-65533_0y.dds"},
	{0xF493750F62DE726CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall4x3_0.pwk"},
	{0xF493A7BE3F1FF79BULL, "hlac_w_blightlands_-2_0y.dds"},
	{0xF4944AF040977B1DULL, "444474631.ogg"},
	{0xF494E832A81AD0E1ULL, "6110861_m.fxe"},
	{0xF494F3C89CF482F7ULL, "vxdr_primevalentrance_289_0_0.vlm"},
	{0xF49545F56D91BCBBULL, "vxht_0_hightown_a2_bes_0_0.vlm"},
	{0xF495984946B119C7ULL, "294938763.ogg"},
	{0xF495C1279A61AD08ULL, "6062776_m.fxe"},
	{0xF495CED930CB927AULL, "vxca_cavern1_haunted_ik6_0_0.vlm"},
	{0xF495D6FB98749115ULL, "hlac_dr_0_deeproads_5840_2x.dds"},
	{0xF495E3F5E672F3AAULL, "vxca_caverns1_mines_icn_0_0.vlm"},
	{0xF4966B093DCA448FULL, "vxw_bonepit_166_0_0.vlm"},
	{0xF4967A19EBFDD662ULL, "6030636_m.fxe"},
	{0xF496BF19100CC644ULL, "vxlt_0_lowtown_docks_bcc_0_0.vlm"},
	{0xF496BF30AC1BCE31ULL, "6201606_m.fxe"},
	{0xF496D4B0346396F1ULL, "812011723.ogg"},
	{0xF49733069FDAAC37ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.impa.gad"},
	{0xF4974E0D61B93124ULL, "gen_im_arm_hlm_noba.uti"},
	{0xF497969B2FD7C726ULL, "vxht_0_hightown_bkh_0_0.vlm"},
	{0xF497BEDD39B6E357ULL, "w_bonepitdestroyed_a3_45_0.mmh"},
	{0xF497C2DD39B6EA64ULL, "w_bonepitdestroyed_a3_45_0.mao"},
	{0xF497CFFF8C164175ULL, "and111_lirene6084500.cl"},
	{0xF497D4DD39B708C9ULL, "w_bonepitdestroyed_a3_45_0.msh"},
	{0xF49849D32FA39BFAULL, "6079082_m.fxe"},
	{0xF4984D25551A5664ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_wax_n.dds"},
	{0xF498727291B0EB19ULL, "vxw_bonepitdestroyed_a3_1hb_0_0.vlm"},
	{0xF4987406B0DFCBABULL, "mcr310_argument6074466.cl"},
	{0xF4991A5FF315E9CAULL, "959069527.ogg"},
	{0xF499232F12E16A10ULL, "mrl000ip_at_exit_home.ncs"},
	{0xF4993F8EED73891DULL, "vxca_cavern1_haunted_k2q_0_1.vlm"},
	{0xF499B9E96AE36578ULL, "6100237_f.fxe"},
	{0xF49A69F73660899FULL, "vxht_0_hightown_bi7_0_0.vlm"},
	{0xF49AD0842EBAA921ULL, "default_specularmask.dds"},
	{0xF49B8916BA05F9DBULL, "vxht_0_hightown_cbg_0_0.vlm"},
	{0xF49C95C110841F80ULL, "w_blightlands_46_0.rml"},
	{0xF49CBF91391C2131ULL, "6185078_m.fxe"},
	{0xF49CC901C19737E0ULL, "llac_lt_0_lowtown_a2_131071_4.dds"},
	{0xF49CD369E39D8C23ULL, "39039056.ogg"},
	{0xF49D48069FDF3BC4ULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.impa.evt"},
	{0xF49D9F28DDE7A3A9ULL, "zz_followers_banter6159456.cl"},
	{0xF49DA6BCE95E96CFULL, "6098597_f.fxe"},
	{0xF49DD598F0D89492ULL, "vxlt_gallowscourtyard_n_5z2_0_0.vlm"},
	{0xF49E09DC6C8791D6ULL, "381033367.ogg"},
	{0xF49E366518CFC5E2ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.cess.evt"},
	{0xF49E54D9674A67E3ULL, "vxht_brothel_n_5ah_0_0.vlm"},
	{0xF49E8B2A45E0D643ULL, "vxw_bonepitdestroyed_a3_20c_0_0.vlm"},
	{0xF49EB68A10D67D1EULL, "art\\characters\\creatures\\mubari\\animation\\com_howl.ani"},
	{0xF49EDA4F2E9E632DULL, "dae200_saemus6091703.cl"},
	{0xF49F2FD0053ECE96ULL, "25920253.ogg"},
	{0xF49FB7E0DA2A0A3FULL, "6177319_f.fxe"},
	{0xF4A03D94DDA414FFULL, "vxlt_0_lowtown_a2_n_c_fef_0_0.vlm"},
	{0xF4A0548946DCDBA9ULL, "vxlt_0_lowtown_efg_0_0.vlm"},
	{0xF4A0D596D94D77E4ULL, "287306279.ogg"},
	{0xF4A0EA2925CA2D9EULL, "6089485_m.fxe"},
	{0xF4A0F00E1ADEC0DAULL, "lowtown_fem_cityguard_mob6164824.cl"},
	{0xF4A27979E6760FC5ULL, "vxw_bonepitdestroyed_a3_199_0_0.vlm"},
	{0xF4A2CF4624B0D277ULL, "art\\characters\\weapons\\quivers\\w_qui_qu02a.mmh"},
	{0xF4A2D34624B0D884ULL, "art\\characters\\weapons\\quivers\\w_qui_qu02a.mao"},
	{0xF4A2E54624B0F7E9ULL, "art\\characters\\weapons\\quivers\\w_qui_qu02a.msh"},
	{0xF4A2EDFE75CC3226ULL, "cod_cha_divine_justina.utp"},
	{0xF4A2FF7A6CF14BE4ULL, "vxw_encounter3_dc_0_0.vlm"},
	{0xF4A33F51F82C0FA0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_partycall_n.dds"},
	{0xF4A354EFF3002D2EULL, "vxlt_gallowstemplar_f_5xf_0_0.vlm"},
	{0xF4A37D8F42482C33ULL, "vxca_caverns1_edg_0_0.vlm"},
	{0xF4A387E51CD5494CULL, "6072117_m.fxe"},
	{0xF4A3AE63CBB3D413ULL, "274476399.ogg"},
	{0xF4A3DE87D61AE7DFULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_rob_appb_0n.dds"},
	{0xF4A433048C57C588ULL, "6069676_m.fxe"},
	{0xF4A43D5CFCCE28C0ULL, "vxw_woundedcoast_1eo_0_0.vlm"},
	{0xF4A461FA47A5A081ULL, "6160761_m.fxe"},
	{0xF4A510AD9A1A0048ULL, "gen_im_trs_crc_sta_phy_r1.uti"},
	{0xF4A5356518D5E2F5ULL, "art\\vfx\\creatures\\fxc_ash_proj_p.cess.gad"},
	{0xF4A54CE9672F6C20ULL, "vxlt_undercitydungeon_175_0_0.vlm"},
	{0xF4A5742F1CE525B6ULL, "409339136.ogg"},
	{0xF4A58A064B259AFBULL, "vxht_mansion2b_3gu_0_1.vlm"},
	{0xF4A5B2DEF440CB7BULL, "vxca_caverns1_mines_gi3_0_0.vlm"},
	{0xF4A693299FF23472ULL, "6188587_f.fxe"},
	{0xF4A712AFCD9C6217ULL, "vxw_encounter3_7y_0_0.vlm"},
	{0xF4A7317FCD0EDB51ULL, "649015398.ogg"},
	{0xF4A743DBEADA904CULL, "6129284_m.fxe"},
	{0xF4A74C7F25A2BF0CULL, "dae000_bartrand6206885.cl"},
	{0xF4A7BB78C682B69FULL, "vxht_mansion_n_4yj_0_0.vlm"},
	{0xF4A857373ECC2185ULL, "198409611.ogg"},
	{0xF4A87F7150870099ULL, "vxlt_0_lowtown_docks_b9l_0_0.vlm"},
	{0xF4AA0740B7BB7BFDULL, "vxht_chantry_2pj_0_0.vlm"},
	{0xF4AA0F72C78CDCB8ULL, "backgroundmovie.gfx"},
	{0xF4AA25F01DE20663ULL, "dae021tr_scene_starts.ncs"},
	{0xF4AA38CA1D13C7FEULL, "mag170_thrask6016024.cl"},
	{0xF4AA8D425B023E33ULL, "6053919_m.fxe"},
	{0xF4AA9CBC682C9081ULL, "vxca_caverns1_gk7_0_1.vlm"},
	{0xF4AAF98BF6E5CE64ULL, "vxlt_gallowstemplar_n_46w_0_0.vlm"},
	{0xF4AB4A90661BCD75ULL, "542700054.ogg"},
	{0xF4AB60069FEBA1BDULL, "art\\vfx\\combat\\vfx_c_mer_stair_barrier_p.impa.ani"},
	{0xF4ABCBF8A9FD01C1ULL, "dre000_varric.bnk"},
	{0xF4ABF584519DE04BULL, "109866656.ogg"},
	{0xF4AC2A67B787F18BULL, "pet_small.mao"},
	{0xF4ACCE8A10E2E3EFULL, "art\\characters\\creatures\\mubari\\animation\\com_howl.evt"},
	{0xF4ACEB45A97925C6ULL, "vxlt_0_lowtown_a2_n_ejg_0_0.vlm"},
	{0xF4AD43B3C9CA127CULL, "6187921_m.fxe"},
	{0xF4AD8C27857368F8ULL, "vxht_bartrand_n_3i5_0_0.vlm"},
	{0xF4AE979E024C392FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf_r.gad"},
	{0xF4AEAF71CF9A2D7FULL, "vxlt_0_lowtown_a2_n_c_fha_0_0.vlm"},
	{0xF4AEEDF8A9FF7131ULL, "dre000_varric.cnv"},
	{0xF4AF618AA9A9FFE9ULL, "vxca_cavern1_haunted_gkr_0_1.vlm"},
	{0xF4AF74DE74E68019ULL, "6053287_m.fxe"},
	{0xF4AF9D6AD10B3EA1ULL, "vxw_sundermount_20r_0_0.vlm"},
	{0xF4AFAD34A9E140F2ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.mmh"},
	{0xF4B0A9DE1D5DABDAULL, "117732670.ogg"},
	{0xF4B0C61062FED127ULL, "art\\cinematics\\deathblows\\highdragon\\cs_db_drag_dragwarr.gad"},
	{0xF4B0C72C6E9A7141ULL, "w_encounter3_54_0.rml"},
	{0xF4B0F016FEF721BAULL, "6102871_f.fxe"},
	{0xF4B10F81AFB6546CULL, "gen_im_trs_dal_med_hlm_r2.uti"},
	{0xF4B1C6CCE4F645AEULL, "vxlt_gallowstemplar_f_5cg_0_0.vlm"},
	{0xF4B1F6BF7EDF3F14ULL, "hlac_lt_gallowscourtyard_-65538_4x.dds"},
	{0xF4B2CF8A10E75098ULL, "art\\characters\\creatures\\mubari\\animation\\com_howl.gad"},
	{0xF4B2CFB046162CD0ULL, "qun160_vashothlead6064665.cl"},
	{0xF4B3242A7D9EE227ULL, "6102121_m.fxe"},
	{0xF4B331E2956C998FULL, "121322602.ogg"},
	{0xF4B370FA91F6DB44ULL, "vxca_cavern1_haunted_i9r_0_0.vlm"},
	{0xF4B4045F3ADC0529ULL, "340884644.ogg"},
	{0xF4B404CAD58117FDULL, "zz_gregtest.cut"},
	{0xF4B409DF802778EAULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_lgta_0n.dds"},
	{0xF4B40D4EE7751C0CULL, "vxw_encounter2_pa_0_1.vlm"},
	{0xF4B426502DE6D1BAULL, "character_character_p140.psh.fxo.dx11.assembly"},
	{0xF4B4498F88138AAAULL, "lgt311_shw_mer_mettin6156482.cl"},
	{0xF4B50CF55A3398AFULL, "vxca_caverns1_e90_0_0.vlm"},
	{0xF4B59216318402E6ULL, "mag291_vincento6027654.cl"},
	{0xF4B6B2E8EC095CCDULL, "dae000_pc_home_ambients6208209.cl"},
	{0xF4B6C51036F30D8BULL, "darktown_male_lowlife_mob6176177.cl"},
	{0xF4B721C80B2800D7ULL, "lt_gallowstemplar_f_2383.rml"},
	{0xF4B77FBDBEB01102ULL, "vxw_bonepitmining_a2_1xg_0_0.vlm"},
	{0xF4B7A8ACB289EF0BULL, "6170432_m.fxe"},
	{0xF4B7AA6FB21D7E58ULL, "zz_demoarmor.ncs"},
	{0xF4B7DF395ABD6B69ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10b.phy"},
	{0xF4B8428036FEAA11ULL, "dae000_orsino6085498.cl"},
	{0xF4B8D177B2CF6687ULL, "llac_ht_mansion_n_3381_0.dds"},
	{0xF4B962B0EB5322CCULL, "vxw_encounter2_3y_0_0.vlm"},
	{0xF4B9B267B6AF1ADDULL, "312547665.ogg"},
	{0xF4B9D138F6B459BBULL, "728603786.ogg"},
	{0xF4B9D6BBA7211CF8ULL, "vxht_0_hightown_a3_cbi_0_0.vlm"},
	{0xF4BA2956BC030C62ULL, "lgt260_eau_tmp_asst_harbor6143637.cl"},
	{0xF4BA38B7CE83F47CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_medj_0s.dds"},
	{0xF4BADFAE1C54555FULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1.mmh"},
	{0xF4BAEBA168E8F109ULL, "6076343_m.fxe"},
	{0xF4BB489C2D6F12B5ULL, "717704418.ogg"},
	{0xF4BB9EBB9CE7F5DEULL, "w_sundermount_79_0_ll.mmh"},
	{0xF4BBCBDC9A1C1E58ULL, "mrl150ip_return_to_camp.ncs"},
	{0xF4BBF135D39AC487ULL, "gen00fl_isabela_init6076474.cl"},
	{0xF4BC1BAC868F1085ULL, "t1_mue_pu1.tnt"},
	{0xF4BCB25BF18D13ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_varian_loop.ani"},
	{0xF4BD391F9DE8FCF4ULL, "llac_lt_0_lowtown_docks_n_-65539_4.dds"},
	{0xF4BE0F47C3015721ULL, "vxw_sundermount_25q_0_0.vlm"},
	{0xF4BE44099131D114ULL, "hla_ht_0_hightown_65535_0x.dds"},
	{0xF4BE5D2BE3D64752ULL, "gen00fl_varric6033664.cl"},
	{0xF4BEBA445EC6071EULL, "mcr370_meredith6084416.cl"},
	{0xF4BF468ADF8C0982ULL, "ser191_emeric6104756.cl"},
	{0xF4BFCAD849CD55C7ULL, "vxca_caverns1_k2q_0_2.vlm"},
	{0xF4C022D0376A13ABULL, "6127403_m.fxe"},
	{0xF4C0798E190F12C4ULL, "vxw_bonepitdestroyed_a3_1zu_0_0.vlm"},
	{0xF4C0A5E475836F92ULL, "lit_bt_01.mao"},
	{0xF4C0B5894F9B314FULL, "vxlt_0_lowtown_a3_n_c_i18_0_0.vlm"},
	{0xF4C0CBB9B99A608CULL, "hlac_w_encounter3_-327680_0z.dds"},
	{0xF4C12116E0BA59C0ULL, "vxw_woundedcoast_1km_0_0.vlm"},
	{0xF4C1568A53782E89ULL, "vxht_0_hightown_9zo_0_0.vlm"},
	{0xF4C15F1EF2A37009ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\sitting_chair_pose_a.gad"},
	{0xF4C1CDB83B51C258ULL, "sib100_father_contact6110072.cl"},
	{0xF4C22D64153D3D11ULL, "llac_w_bonepitmining_a2_327683_0.dds"},
	{0xF4C250C758162109ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.gad"},
	{0xF4C2A6434D900892ULL, "199186893.ogg"},
	{0xF4C2EC3205BA10A6ULL, "vxht_0_hightown_bzk_0_0.vlm"},
	{0xF4C350C2C9EA183AULL, "453113290.ogg"},
	{0xF4C3B198B9449481ULL, "vxlt_0_lowtown_a3_n_c_cml_0_0.vlm"},
	{0xF4C3B49E025EC6F5ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_cmov_wf_r.ani"},
	{0xF4C3BAC33476F1BDULL, "vxw_bonepit_1pv_0_0.vlm"},
	{0xF4C41F3B721C1B15ULL, "ts_kirk_alienage_01.gda"},
	{0xF4C49BACDCAD95B1ULL, "dae000_bartrand6201362.cl"},
	{0xF4C4FAA5613C7DCFULL, "vxw_sundermount_30e_0_0.vlm"},
	{0xF4C5128816DAD310ULL, "vxw_encounter2_camp_l0_0_0.vlm"},
	{0xF4C560E26BFDF243ULL, "zz_lgt320ip_mis_resp_chest.utp"},
	{0xF4C5B31063110D6DULL, "art\\cinematics\\deathblows\\highdragon\\cs_db_drag_dragwarr.ani"},
	{0xF4C5B74F06B99006ULL, "vxw_bonepitmining_a2_6f_0_0.vlm"},
	{0xF4C5BF8E0C42D31FULL, "vxht_0_hightown_bx0_0_0.vlm"},
	{0xF4C5CA00FBAE46E0ULL, "darktown_fem_fereldan_sta6152804.cl"},
	{0xF4C5CFA9DA5AE0C1ULL, "art\\characters\\playercharacter\\shared\\heads\\textures\\mt_tat_e02_0t.dds"},
	{0xF4C5EE3A22E86CD2ULL, "vxw_encounter2_1i_0_0.vlm"},
	{0xF4C629C4A7AEC142ULL, "6127301_m.fxe"},
	{0xF4C633C2974C6531ULL, "vxw_encounter2_camp_rr_0_1.vlm"},
	{0xF4C6C86616E4974BULL, "vxlt_blackemporium_16c_0_0.vlm"},
	{0xF4C733FB502F62F1ULL, "hlac_ht_keep_n_3732_11x.dds"},
	{0xF4C762277CFEAF86ULL, "514632519.ogg"},
	{0xF4C8301EF2A93F57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\sitting_chair_pose_a.ani"},
	{0xF4C8682201EF9D24ULL, "and100_anders6095994.cl"},
	{0xF4C87C0FB2904274ULL, "101131906.ogg"},
	{0xF4C904EF4A1214C3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_lp.ani"},
	{0xF4C9215D9BB59EA0ULL, "48356776.ogg"},
	{0xF4C921C7581BF057ULL, "art\\characters\\playercharacter\\elffemale\\animation\\cindes\\dialogue\\sitting_table_clasped_2_loop.ani"},
	{0xF4C9868B9C90D390ULL, "792907541.ogg"},
	{0xF4C9D0CCCCA85326ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_p.evt"},
	{0xF4CA71B622545545ULL, "112212170.ogg"},
	{0xF4CBB9DC0C18A4A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_kick_feet.ani"},
	{0xF4CC0FACB389873BULL, "vxlt_0_lowtown_a2_heu_0_0.vlm"},
	{0xF4CC9CAC18F9DF1FULL, "character_character_p255.psh.fxo.dx11.assembly"},
	{0xF4CCDE37EBE78996ULL, "zz_followers_banter6158018.cl"},
	{0xF4CD06F24375D551ULL, "6060472_m.fxe"},
	{0xF4CD47196EE0FDB9ULL, "276228044.ogg"},
	{0xF4CD9D654FF873C0ULL, "6133294_m.fxe"},
	{0xF4CE633118B7F085ULL, "856684316.ogg"},
	{0xF4CEAA86C8E276AEULL, "vxlt_gallowstemplar_f_5iy_0_0.vlm"},
	{0xF4CEB8E3CD6748A4ULL, "dae000_arishok6081676.cl"},
	{0xF4CF850B84B3BF8FULL, "6198492_m.fxe"},
	{0xF4CFE7CCCCACE671ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_p.gad"},
	{0xF4CFFB0D69F29016ULL, "lla_ht_0_hightown_131073_1.dds"},
	{0xF4D084FC9F43D18DULL, "lt_0_lowtown_docks_n_194_0.msh"},
	{0xF4D092FC9F43E9C0ULL, "lt_0_lowtown_docks_n_194_0.mao"},
	{0xF4D096FC9F43F013ULL, "lt_0_lowtown_docks_n_194_0.mmh"},
	{0xF4D115DB26E7524FULL, "llac_lt_0_lowtown_docks_n_-131074_1.dds"},
	{0xF4D1426D3E41A6FAULL, "vxlt_0_lowtown_a2_byr_0_0.vlm"},
	{0xF4D19067F43B05E2ULL, "188655713.ogg"},
	{0xF4D1995BF19F4535ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\knifepoint_actions_varian_loop.gad"},
	{0xF4D1E2D86318E667ULL, "12830829.ogg"},
	{0xF4D22C415A774559ULL, "vxlt_gallowstemplar_357_0_0.vlm"},
	{0xF4D2923FBC1F452EULL, "6125260_m.fxe"},
	{0xF4D2C423BD730FC6ULL, "327779881.ogg"},
	{0xF4D2C4DC0C1ED643ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_kick_feet.gad"},
	{0xF4D2F8271D41A70FULL, "vxlt_backalley_g_1jp_0_0.vlm"},
	{0xF4D300A1AFB3FFADULL, "killparty.ncs"},
	{0xF4D30267A1BEFA1EULL, "vxlt_0_lowtown_a3_n_c_h1x_0_0.vlm"},
	{0xF4D321C210F2692AULL, "6125145_m.fxe"},
	{0xF4D322BF781E2C85ULL, "dae131ip_at_to_darktown2.utp"},
	{0xF4D368415674256AULL, "vxlt_0_lowtown_h7n_0_0.vlm"},
	{0xF4D37326187DE50AULL, "6051810_f.fxe"},
	{0xF4D3846508D0EB90ULL, "vxw_encounter2_camp_g1_0_0.vlm"},
	{0xF4D395287E9299F5ULL, "594842715.ogg"},
	{0xF4D3A7A24F069353ULL, "gen00fl_aveline6140032.cl"},
	{0xF4D3E2DC9A30D751ULL, "mrl150ip_return_to_camp.utp"},
	{0xF4D41769B6913AACULL, "hla_ht_0_hightown_a3_65538_0x.dds"},
	{0xF4D444BCAED5B148ULL, "w_bonepit_202_0.mmh"},
	{0xF4D450BCAED5C5EBULL, "w_bonepit_202_0.mao"},
	{0xF4D462BCAED5E472ULL, "w_bonepit_202_0.msh"},
	{0xF4D46377A29C4202ULL, "vxw_bonepitmining_a2_1ra_0_0.vlm"},
	{0xF4D4AF43F1EC2AF0ULL, "281718205.ogg"},
	{0xF4D4B3733F22B9D2ULL, "vxw_blightlands_fg_0_0.vlm"},
	{0xF4D4C93D0A12FF7FULL, "6125759_m.fxe"},
	{0xF4D5122F0F50AB24ULL, "200232913.ogg"},
	{0xF4D570EFB2A9783BULL, "dae000_madam_lusine6151572.cl"},
	{0xF4D5FF1847BC359FULL, "llac_w_bonepitdestroyed_a3_327681_0.dds"},
	{0xF4D659A48B47DD67ULL, "vxw_bonepitdestroyed_a3_7b_0_0.vlm"},
	{0xF4D6E8CCCCB3061FULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\exploration\\movement\\mov_p_r90_p.ani"},
	{0xF4D6F5D4F1150D23ULL, "vxlt_0_lowtown_docks_n_b0j_0_0.vlm"},
	{0xF4D73F62083AD801ULL, "6080434_m.fxe"},
	{0xF4D74498D3EE0536ULL, "6181921_m.fxe"},
	{0xF4D77D46AF3135C7ULL, "6177305_m.fxe"},
	{0xF4D78DDA5EB0FADCULL, "6111580_m.fxe"},
	{0xF4D7AC34AA029099ULL, "art\\vfx\\combat\\vfx_c_pride_pois_p.anb"},
	{0xF4D7F9C3505CB808ULL, "mag100_cullen6095130.cl"},
	{0xF4D81163C3BE6057ULL, "vrc321ip_spooky_8_owner.utp"},
	{0xF4D8312A33EE9313ULL, "277335451.ogg"},
	{0xF4D89B478D8A13E5ULL, "vxht_0_hightown_a3_9ao_0_0.vlm"},
	{0xF4D95025CC554334ULL, "vxw_sundermount_2e8_0_0.vlm"},
	{0xF4D996F76BDE8342ULL, "zz_fen_debug6134390.cl"},
	{0xF4D9A1DC0C24B928ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_f_kick_feet.evt"},
	{0xF4DA8189A57F9FBBULL, "vxlt_0_lowtown_a2_hbt_0_0.vlm"},
	{0xF4DB14E98AB3D993ULL, "dae000_arishok6075375.cl"},
	{0xF4DB5DE20BA6C297ULL, "zz_followers_banter6159678.cl"},
	{0xF4DC2D2AD0E17A80ULL, "103901279.ogg"},
	{0xF4DC59EA27C2D6F7ULL, "759812401.ogg"},
	{0xF4DC8FD6DAAC3561ULL, "vxlt_0_lowtown_a3_n_c_e12_0_0.vlm"},
	{0xF4DC913AB4200BBDULL, "mag291_ntbanter6116214.cl"},
	{0xF4DD1BEF4A22E4ADULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\ambient\\am_grdstand_lp.gad"},
	{0xF4DD306764F89587ULL, "vxw_bonepitmining_a2_2an_0_0.vlm"},
	{0xF4DE04D0C4A68AC0ULL, "vxw_woundedcoast_1qk_0_0.vlm"},
	{0xF4DE283DB38EDF97ULL, "6120972_m.fxe"},
	{0xF4DEFEA90547F232ULL, "6108434_m.fxe"},
	{0xF4E044E4B76DE5A1ULL, "42867451.ogg"},
	{0xF4E0BC651D4C3CC6ULL, "lt_0_lowtown_a2_n_c_30_0.rml"},
	{0xF4E16A040F37BF8FULL, "vxlt_backalley_g_1ws_0_0.vlm"},
	{0xF4E1740D81F4C685ULL, "639336663.ogg"},
	{0xF4E1BD351E4C36A9ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.dura.ani"},
	{0xF4E1F627AAD8D79EULL, "hlac_lt_gallowstemplar_f_2383_9x.dds"},
	{0xF4E21DBDAF85CE2FULL, "art\\vfx\\combat\\vfx_gib_hm_meda_rtlb.anb"},
	{0xF4E2775746C11C52ULL, "274240124.ogg"},
	{0xF4E27C34D6D21AB9ULL, "w_woundedcoast_145_0.msh"},
	{0xF4E28634D6D22BA7ULL, "w_woundedcoast_145_0.mmh"},
	{0xF4E28A34D6D232F4ULL, "w_woundedcoast_145_0.mao"},
	{0xF4E29B08EAA5C106ULL, "vxw_bonepitmining_a2_8h_0_0.vlm"},
	{0xF4E2A2163AE7ED7EULL, "728073694.ogg"},
	{0xF4E2B984F23F595BULL, "motionblur.mao"},
	{0xF4E39F2C8E82078CULL, "follower_banter6172241.cl"},
	{0xF4E3C8395AE27E8EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10b.mmh"},
	{0xF4E3E2395AE2AB7CULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos10b.msh"},
	{0xF4E432FE414695C0ULL, "gen00fl_varric6076089.cl"},
	{0xF4E48E05558590C6ULL, "44111096.ogg"},
	{0xF4E4CD946CD1ABEAULL, "339457874.ogg"},
	{0xF4E51539897D4AC8ULL, "12546653.ogg"},
	{0xF4E52721534EF2E9ULL, "lowtown_male_alienage1_sta6162751.cl"},
	{0xF4E5882D6A636DCBULL, "vxw_encounter2_p2_0_1.vlm"},
	{0xF4E58DF7F6A82657ULL, "6173183_m.fxe"},
	{0xF4E5E87FDA8ADD2FULL, "6015936_f.fxe"},
	{0xF4E5F50A537D6469ULL, "ser270_templar6127477.cl"},
	{0xF4E6188C9C00E148ULL, "vxlt_0_lowtown_a3_n_c_e17_0_0.vlm"},
	{0xF4E72A51F11320D0ULL, "226348357.ogg"},
	{0xF4E7480A9DEA5E33ULL, "vxw_bonepit_10h_0_0.vlm"},
	{0xF4E7C202BE4B5BB4ULL, "vxw_sundermount_2n9_0_0.vlm"},
	{0xF4E80598605A9F8CULL, "w_woundedcoast_12_0.msh"},
	{0xF4E80B98605AAA7EULL, "w_woundedcoast_12_0.mmh"},
	{0xF4E8167761BA4C53ULL, "vxw_blightlands_ld_0_0.vlm"},
	{0xF4E81798605ABE15ULL, "w_woundedcoast_12_0.mao"},
	{0xF4E82E4D2388CE85ULL, "vxht_0_hightown_a3_cbp_0_0.vlm"},
	{0xF4E83CDB87B474AAULL, "6114462_m.fxe"},
	{0xF4E8D8351E52832BULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.dura.gad"},
	{0xF4E90B53D2D761E7ULL, "344569046.ogg"},
	{0xF4EA46F5CB5E9B23ULL, "vxca_caverns2_908_0_0.vlm"},
	{0xF4EA740815076989ULL, "6160633_m.fxe"},
	{0xF4EA80997DA75948ULL, "tbe100_mirror.cnv"},
	{0xF4EA8525C868E4BBULL, "493012938.ogg"},
	{0xF4EAC8CE0576A369ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_ric_0d.dds"},
	{0xF4EB24664E8E9581ULL, "vxlt_0_lowtown_docks_a3_d03_0_1.vlm"},
	{0xF4EB3DD434F4E9E8ULL, "plt_vrc300pt_n_initial.nss"},
	{0xF4EB5CFB7502E47DULL, "vxlt_0_lowtown_a2_go2_0_0.vlm"},
	{0xF4EBCF24EEF15C44ULL, "vxw_bonepitdestroyed_a3_1ex_0_0.vlm"},
	{0xF4EC419F0537FD4AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\blendtree\\aveline_overlay.anb"},
	{0xF4ECA314BE131EBBULL, "vxht_brothel_n_42p_0_0.vlm"},
	{0xF4ECEFC2E7459B53ULL, "289072768.ogg"},
	{0xF4ED41F5F86F7FADULL, "vxw_bonepitmining_a2_fi_0_0.vlm"},
	{0xF4ED42FB18476150ULL, "45007994.ogg"},
	{0xF4EDA2997DA9C8B0ULL, "tbe100_mirror.bnk"},
	{0xF4EDC330430A4D9BULL, "art\\vfx\\combat\\vfx_w_rev_devour_act_c.impa.evt"},
	{0xF4EDE8C657C986B2ULL, "hlac_lt_0_lowtown_a2_n_c_131071_4y.dds"},
	{0xF4EE343BDFD37C5BULL, "vxlt_0_lowtown_n_ex4_0_0.vlm"},
	{0xF4EED5351E56E990ULL, "art\\vfx\\combat\\vfx_m_cre_haste_c.dura.evt"},
	{0xF4EF35FB94ABD28CULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.cess.ani"},
	{0xF4EF5D557B0765FFULL, "vxlt_backalley_g_1k7_0_0.vlm"},
	{0xF4EFE13C0774D9A6ULL, "712686509.ogg"},
	{0xF4F041D32F9C719CULL, "965673820.ogg"},
	{0xF4F0985012F4F99DULL, "vxht_mansion2b_1mb_0_0.vlm"},
	{0xF4F0A9A4A624A8E8ULL, "76098857.ogg"},
	{0xF4F146D434F964FBULL, "plt_vrc300pt_n_initial.plo"},
	{0xF4F18578B9A93806ULL, "zz_vault_debug6203305.cl"},
	{0xF4F1972D230EEAD2ULL, "vxw_blightlands_la_0_0.vlm"},
	{0xF4F1B92C6ED1F030ULL, "w_encounter3_54_0.msh"},
	{0xF4F1C72C6ED2084AULL, "w_encounter3_54_0.mmh"},
	{0xF4F1CB2C6ED20E81ULL, "w_encounter3_54_0.mao"},
	{0xF4F2223F114C0B84ULL, "art\\characters\\weapons\\shortbows\\textures\\w_sbw_001a_0s.dds"},
	{0xF4F255252A4C3E64ULL, "vxht_0_hightown_n_bky_0_0.vlm"},
	{0xF4F259A0B223B136ULL, "hlac_ht_mansion_n_3369_2y.dds"},
	{0xF4F3029BB3877987ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.impa.gad"},
	{0xF4F37E7B7C717415ULL, "vxlt_hangedman_n_2fq_0_0.vlm"},
	{0xF4F382ADBA89DC0EULL, "lt_0_lowtown_docks_n_207_0.msh"},
	{0xF4F382C37CD1A7CCULL, "496876659.ogg"},
	{0xF4F390ADBA89F45FULL, "lt_0_lowtown_docks_n_207_0.mao"},
	{0xF4F39CADBA8A08FCULL, "lt_0_lowtown_docks_n_207_0.mmh"},
	{0xF4F3A853534DC453ULL, "hlac_lt_0_lowtown_n_-65537_2x.dds"},
	{0xF4F3D003A2E16C25ULL, "vxdr_primevalentrance_1ka_0_0.vlm"},
	{0xF4F3F2FDA47A2996ULL, "vxht_mansion_n_4zv_0_0.vlm"},
	{0xF4F4731F4781C814ULL, "vxht_mansion2b_20w_0_0.vlm"},
	{0xF4F47F0E2B66CBD6ULL, "gen00fl_fenris6126621.cl"},
	{0xF4F4BC30431060DCULL, "art\\vfx\\combat\\vfx_w_rev_devour_act_c.impa.gad"},
	{0xF4F4E689E5F57D43ULL, "vxw_encounter2_camp_65_0_0.vlm"},
	{0xF4F52EC785F0059CULL, "906787357.ogg"},
	{0xF4F54F380AAC305BULL, "25336301.ogg"},
	{0xF4F5666377954907ULL, "hlac_ca_caverns1_mines_24425_12y.dds"},
	{0xF4F5B1BD3455A326ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_action.evt"},
	{0xF4F5B9E78FE076B3ULL, "vxw_bonepit_15k_0_0.vlm"},
	{0xF4F5DB61E0F2751CULL, "zz_followers_banter6159420.cl"},
	{0xF4F5E67ECC2F17C6ULL, "731228144.ogg"},
	{0xF4F5FCD811B280A0ULL, "6127512_m.fxe"},
	{0xF4F642FB94B207CEULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.cess.gad"},
	{0xF4F6819C4E1FA90DULL, "6129018_m.fxe"},
	{0xF4F6B0215A8AB230ULL, "63360462.ogg"},
	{0xF4F750F26E71CF72ULL, "face_face_p16.psh.fxo.dx11"},
	{0xF4F77101F73AA23BULL, "884218949.ogg"},
	{0xF4F7933B1B628B96ULL, "vxlt_gallowsprison_n_6kx_0_0.vlm"},
	{0xF4F83F1579BDCD6EULL, "929360973.ogg"},
	{0xF4F85A6015DC242CULL, "plt_dae00pt_rumors.plo"},
	{0xF4F86F3B61B0DEF8ULL, "vxlt_gallowstemplar_n_5c9_0_0.vlm"},
	{0xF4F8945DFD7C26D8ULL, "vxht_brothel_n_5ky_0_0.vlm"},
	{0xF4F8C0BF5789DED6ULL, "vxht_keep_n_67w_0_0.vlm"},
	{0xF4F8E68F8A65B6E7ULL, "676992587.ogg"},
	{0xF4F9075D11220C50ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_1x.dds"},
	{0xF4F94801457CFE4DULL, "vxht_mansion2_28z_0_0.vlm"},
	{0xF4F94F71B1BD943EULL, "vxht_keep_4bq_0_0.vlm"},
	{0xF4F966B1AB83BEE1ULL, "tree_04.dds"},
	{0xF4F9F79BB38D8594ULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.impa.evt"},
	{0xF4FA4101E0E774C4ULL, "vxw_bonepitdestroyed_a3_1ny_0_0.vlm"},
	{0xF4FA6F2145FA85AAULL, "708190230.ogg"},
	{0xF4FA9722FC1649D8ULL, "vxht_mansion_n_700_0_1.vlm"},
	{0xF4FABB304314CADAULL, "art\\vfx\\combat\\vfx_w_rev_devour_act_c.impa.ani"},
	{0xF4FB00160B610868ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandlright.msh"},
	{0xF4FB12160B6126F9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandlright.mao"},
	{0xF4FB151B7EE0752BULL, "vxht_0_hightown_3ub_0_0.vlm"},
	{0xF4FB1E160B613A92ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_grandlright.mmh"},
	{0xF4FBB2CC7A394E2DULL, "631712117.ogg"},
	{0xF4FBB3D2EA65982DULL, "vxw_bonepitmining_a2_mh_0_0.vlm"},
	{0xF4FBC8BD345A3671ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_action.gad"},
	{0xF4FBE138CFF819B5ULL, "vxlt_0_lowtown_docks_n_b52_0_1.vlm"},
	{0xF4FBE262E9342332ULL, "6106634_m.fxe"},
	{0xF4FC038051E4E22CULL, "vxht_0_hightown_a3_bpr_0_0.vlm"},
	{0xF4FCA618D1C994DBULL, "vxlt_0_lowtown_n_eu7_0_0.vlm"},
	{0xF4FCED4ABC35EF67ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_enter.gad"},
	{0xF4FD3DFB94B81DADULL, "art\\vfx\\combat\\vfx_c_ogr_charge_c.cess.evt"},
	{0xF4FD85760D73D278ULL, "6071071_m.fxe"},
	{0xF4FD87BC7E4CE7F7ULL, "6166591_f.fxe"},
	{0xF4FD8FD7BD2C04B0ULL, "llac_lt_0_lowtown_docks_a3_n_-65537_1.dds"},
	{0xF4FDA07C824FCB19ULL, "298065618.ogg"},
	{0xF4FDE6303ED9C177ULL, "vxlt_0_lowtown_a2_hea_0_0.vlm"},
	{0xF4FDF078C18D4B4EULL, "hightown_male_noble_gossip6171383.cl"},
	{0xF4FE1E7A622F7A9CULL, "llac_lt_0_lowtown_-65536_1.dds"},
	{0xF4FE675FD3941FB6ULL, "6051611_m.fxe"},
	{0xF4FE8DE1377B5C3EULL, "83812156.ogg"},
	{0xF4FEA98032653FD0ULL, "vxlt_backalley_n_1xe_0_1.vlm"},
	{0xF4FEB78EFACFC198ULL, "6109048_m.fxe"},
	{0xF4FEBED10990D78BULL, "371519759.ogg"},
	{0xF4FF0B26D18F5525ULL, "6077973_m.fxe"},
	{0xF4FF2AF01E2A4EFAULL, "dae021tr_scene_starts.utt"},
	{0xF4FF3F6015E21563ULL, "plt_dae00pt_rumors.nss"},
	{0xF4FFAE6FBE33F37DULL, "897006663.ogg"},
	{0xF4FFB5F7647A59BFULL, "325900747.ogg"},
	{0xF500090A15050352ULL, "vxw_blightlands_qb_0_0.vlm"},
	{0xF501016C58E8C1E7ULL, "llac_lt_darktown_-1_7.dds"},
	{0xF50165DE17AAE4BAULL, "character_character_p137.psh.fxo.dx11"},
	{0xF501812BB7F891DFULL, "llac_lt_0_lowtown_docks_n_-4_7.dds"},
	{0xF501F0586E678C95ULL, "vxlt_hangedman_n_2mp_0_0.vlm"},
	{0xF502B63CEF63C124ULL, "6113084_f.fxe"},
	{0xF502C9BD3460561FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\rogue_dualwield_persuade_action.ani"},
	{0xF502E24ABC3A4874ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_enter.evt"},
	{0xF50311503EBDF04EULL, "hlac_lt_0_lowtown_n_65536_4y.dds"},
	{0xF5032FD8D69B565CULL, "vxca_caverns2_854_0_1.vlm"},
	{0xF5035866D7EB95C3ULL, "vxw_encounter2_camp_14_0_0.vlm"},
	{0xF50497EB5E7949A0ULL, "lt_gallowstemplar_n_2393.rml"},
	{0xF504D8C6B8DA4F93ULL, "hlac_w_encounter2_327685_0z.dds"},
	{0xF504DB0D1DC96B0FULL, "lt_0_lowtown_a2_25_0.anb"},
	{0xF5050688C448CCE4ULL, "rog_staggersneakattack.dds"},
	{0xF5051C81BFAF1843ULL, "zz_ave_debug6056766.cl"},
	{0xF50591AF26A9301BULL, "6084160_m.fxe"},
	{0xF505AE4605362EE6ULL, "115817315.ogg"},
	{0xF505B333EC6ED663ULL, "gen_im_cns_cft_pot_01.uti"},
	{0xF505C844DC488D5BULL, "6068513_m.fxe"},
	{0xF5067A5225CDE5CCULL, "hlac_lt_0_lowtown_a3_n_c_65536_6z.dds"},
	{0xF506968ED11EF025ULL, "6132771_m.fxe"},
	{0xF506C639F8118A94ULL, "179344981.ogg"},
	{0xF506F9F10374869EULL, "523551018.ogg"},
	{0xF507329C497FF756ULL, "vxht_keep_n_62x_0_0.vlm"},
	{0xF507446BA13FDEB5ULL, "545326970.ogg"},
	{0xF5078DCD8B863CBAULL, "6095141_m.fxe"},
	{0xF507AA230E9C58D6ULL, "6126412_m.fxe"},
	{0xF507AE2FA1CE6035ULL, "vxht_0_hightown_n_cag_0_0.vlm"},
	{0xF507EF9BB399B5CDULL, "art\\vfx\\cinematic\\vfx_ground_rumble_p.impa.ani"},
	{0xF50822EFBD1CC4C3ULL, "6169904_m.fxe"},
	{0xF508DF70F216757FULL, "6083802_m.fxe"},
	{0xF5090FF70B043C9AULL, "lt_0_lowtown_docks_97_0.msh"},
	{0xF5091DF70B0454D3ULL, "lt_0_lowtown_docks_97_0.mao"},
	{0xF50921F70B045BE0ULL, "lt_0_lowtown_docks_97_0.mmh"},
	{0xF5096FE015E5A882ULL, "hlac_lt_gallowstemplar_n_2392_0y.dds"},
	{0xF5098FDB0D46A909ULL, "vxht_0_hightown_9ar_0_0.vlm"},
	{0xF50A755D43DAFAACULL, "vxht_0_hightown_a3_bos_0_0.vlm"},
	{0xF50B783AD53D8AACULL, "ico_recipe.dds"},
	{0xF50B93D4F070229AULL, "6132853_m.fxe"},
	{0xF50BF413EE456C3DULL, "vxw_bonepit_1iq_0_0.vlm"},
	{0xF50C37B92D16C0A9ULL, "6070135_f.fxe"},
	{0xF50D61808B9A1518ULL, "6131186_m.fxe"},
	{0xF50D9746C51851EFULL, "dae200cr_viscount6087377.cl"},
	{0xF50DF94261BE6D5CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_buzzheada_d.dds"},
	{0xF50E7AE706FB1BD2ULL, "vxw_blightlands_zc_0_0.vlm"},
	{0xF50EC02B9A1F2EA7ULL, "6054846_m.fxe"},
	{0xF50EF612A67E39C8ULL, "44231978.ogg"},
	{0xF50F67FF8940BE2BULL, "109296316.ogg"},
	{0xF50F901E8B63CD34ULL, "6160997_m.fxe"},
	{0xF50FA86056FB39CEULL, "hlac_lt_0_lowtown_a3_n_c_-4_0x.dds"},
	{0xF50FCDC60F5AC128ULL, "vxlt_0_lowtown_a2_n_c_gjm_0_0.vlm"},
	{0xF510604B4BBD63E2ULL, "and121_karl6093879.cl"},
	{0xF510B038B759B9D4ULL, "vfx_rendertoshadowmap_p2.psh.fxo.dx11.assembly"},
	{0xF510DA4ABC4677ADULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\wall_lean_left_enter.ani"},
	{0xF5112A36AD82FB22ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_act_c.impa.gad"},
	{0xF51137BFFF1C8CAAULL, "gen00fl_anders6142539.cl"},
	{0xF511D1A985B32017ULL, "hlac_lt_blackemporium_1055_1x.dds"},
	{0xF51201C97A48057BULL, "513197522.ogg"},
	{0xF512201B7AD41058ULL, "vxlt_0_lowtown_docks_b6l_0_0.vlm"},
	{0xF512ADDC79838429ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\major.ani"},
	{0xF512C4671A0DDA76ULL, "tbe100ip_store_magic_good.ncs"},
	{0xF512E72B5371407AULL, "37985455.ogg"},
	{0xF513272DAD779662ULL, "vxdr_primevaltomb_50b_0_1.vlm"},
	{0xF51357C06AC66F22ULL, "dae220ar_hightown_nt.are"},
	{0xF5139F49789B68F0ULL, "273794152.ogg"},
	{0xF514596670229812ULL, "dae000_bodhan6146819.cl"},
	{0xF5147B16865E0118ULL, "fen210_fenris6109406.cl"},
	{0xF5149114ECADE249ULL, "251419717.ogg"},
	{0xF5149931BC7DE38BULL, "qun131_party_comment6117260.cl"},
	{0xF5149D89AFE3D808ULL, "171609530.ogg"},
	{0xF5153EC4E20B1200ULL, "one110_tree_minder6120976.cl"},
	{0xF5160E42B7EC71CCULL, "lt_0_lowtown_docks_n_16_0_lc.mmh"},
	{0xF5171097C80C9778ULL, "6001003_m.fxe"},
	{0xF517801FF4128E14ULL, "lit_conj_elet_v.mao"},
	{0xF517F68473F3F0E5ULL, "131923453.ogg"},
	{0xF518350B35A27110ULL, "1035433201.ogg"},
	{0xF5184D36AD895519ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_act_c.impa.evt"},
	{0xF518668A839E2997ULL, "6168033_m.fxe"},
	{0xF518A5EF494D9BFCULL, "ser270_templar6112553.cl"},
	{0xF519C8DC7989D0ABULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\major.gad"},
	{0xF51A945A16F79743ULL, "vxca_caverns1_j57_0_0.vlm"},
	{0xF51AA9961F0CD929ULL, "6075235_f.fxe"},
	{0xF51AB2EC50E9AF7FULL, "vxlt_backalley_g_1z0_0_0.vlm"},
	{0xF51AC5AD8196CEE5ULL, "mcr371_hostile_fenris6082214.cl"},
	{0xF51AE0302E6007CBULL, "qun211_braggart6075682.cl"},
	{0xF51AE0BB0B251977ULL, "hlac_lt_0_lowtown_a2_65533_0x.dds"},
	{0xF51B218E96210096ULL, "256445172.ogg"},
	{0xF51B651348A55289ULL, "vxht_chantry_n_26w_0_0.vlm"},
	{0xF51B69F87910FB39ULL, "284063291.ogg"},
	{0xF51B9FC43B021A53ULL, "mag101_decimus6037865.cl"},
	{0xF51BBAB4B4697DA5ULL, "584451983.ogg"},
	{0xF51BDE00054AAB24ULL, "6080735_f.fxe"},
	{0xF51C3231A67817A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_up_side.ani"},
	{0xF51C6A7810595BF2ULL, "hlac_w_sundermount_-393214_0x.dds"},
	{0xF51CE7D2CC3292C1ULL, "hlac_ht_0_hightown_a2_131074_0z.dds"},
	{0xF51CE8AE1CA72EA4ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_box1x1.phy"},
	{0xF51D44B1D387664EULL, "vxlt_backalley_g_1y0_0_0.vlm"},
	{0xF51D761BE850DDA6ULL, "671086086.ogg"},
	{0xF51DECD6385EF1ADULL, "dae100_bran6118169.cl"},
	{0xF51DEDD0E4841C8DULL, "and100_anders6114229.cl"},
	{0xF51E3DB04D458FEEULL, "6085535_m.fxe"},
	{0xF51E4EC06AD067F5ULL, "dae220ar_hightown_nt.lst"},
	{0xF51E57F88D59E9A0ULL, "737200058.ogg"},
	{0xF51E8EE11BCD3C21ULL, "6144107_m.fxe"},
	{0xF51EEEE9AE3009FBULL, "vxw_bonepitmining_a2_gp_0_0.vlm"},
	{0xF51F0710A6F39DA4ULL, "hlac_dr_0_deeproads_17035_1y.dds"},
	{0xF51FBD3866A365D3ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medg.phy"},
	{0xF51FC5DC798E3710ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\major.evt"},
	{0xF51FD006A0032912ULL, "vxht_mansion2b_1au_0_0.vlm"},
	{0xF51FE209DB92546CULL, "hlac_lt_0_lowtown_a2_65536_6z.dds"},
	{0xF5200DA2EFC58C8DULL, "820400010.ogg"},
	{0xF5201AD73FBED6FCULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfood.mao"},
	{0xF52026D73FBEEA9FULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfood.mmh"},
	{0xF5202CD73FBEF541ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfood.msh"},
	{0xF5202E530D56F21FULL, "6102763_m.fxe"},
	{0xF52034338216900AULL, "6095479_m.fxe"},
	{0xF520604542729997ULL, "gen_im_npc_elf_arm_cht_dald.uti"},
	{0xF520652985C9AFA0ULL, "vxw_encounter2_fz_0_0.vlm"},
	{0xF520CD89ABC77330ULL, "6108577_m.fxe"},
	{0xF520DA5985BB5B5DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2a.mao"},
	{0xF520DE5985BB6196ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2a.mmh"},
	{0xF520E85985BB7294ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9cc_stala2a.msh"},
	{0xF5214A9BFA845C33ULL, "6084487_m.fxe"},
	{0xF5218E83F4F79F61ULL, "gen00fl_aveline6053305.cl"},
	{0xF521A3FFFA27A1C7ULL, "hlac_ca_cavern1_haunted_24424_17y.dds"},
	{0xF521BB33C1F1B3C1ULL, "963126653.ogg"},
	{0xF522471F731B7BE5ULL, "vxlt_0_lowtown_docks_a3_n_cy4_0_1.vlm"},
	{0xF52274005FE101C1ULL, "vxlt_gallowstemplar_f_3nu_0_0.vlm"},
	{0xF5229391F8CEBBE7ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.cess.gad"},
	{0xF522BDCB25FC1975ULL, "440337099.ogg"},
	{0xF5233D31A67E4943ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\arm_l_up_side.gad"},
	{0xF52388D46088F5FEULL, "co_cloud.mao"},
	{0xF5240D27D33B94D0ULL, "w_bonepitmining_a2_174_0.msh"},
	{0xF5241B27D33BACEAULL, "w_bonepitmining_a2_174_0.mmh"},
	{0xF5241F27D33BB321ULL, "w_bonepitmining_a2_174_0.mao"},
	{0xF52442CB17C3435BULL, "hlac_ca_cavern1_haunted_24425_13x.dds"},
	{0xF52446C06AD4C5E4ULL, "dae220ar_hightown_nt.ncs"},
	{0xF524669999C5C3F5ULL, "dae000ip_crowd_mf8_talk1.utp"},
	{0xF52489804DB99465ULL, "6088751_m.fxe"},
	{0xF524B34699AA68ECULL, "gen00fl_bethany6102874.cl"},
	{0xF525003EEE982864ULL, "vxw_encounter3_9h_0_0.vlm"},
	{0xF52553A753E1C383ULL, "vxca_caverns2_7mq_0_0.vlm"},
	{0xF5255536AD93ED08ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_act_c.impa.ani"},
	{0xF5257E53C3EF08B3ULL, "vxca_caverns1_e7n_0_0.vlm"},
	{0xF525EC726B24775BULL, "pro000st_aveline_intro.stg"},
	{0xF526546865FD2789ULL, "6172641_m.fxe"},
	{0xF526861DA1759535ULL, "703913118.ogg"},
	{0xF526A049D1CFE911ULL, "6179006_m.fxe"},
	{0xF526E9693699B16DULL, "vxlt_gallowstemplar_6s1_0_1.vlm"},
	{0xF526EF85086504DFULL, "vxlt_undercityencounter_v3_0_0.vlm"},
	{0xF526F3DE3CE9E5EBULL, "dre000_varric6035129.cl"},
	{0xF52702FDCC3631F5ULL, "6174337_m.fxe"},
	{0xF52731BA83A9BB44ULL, "vxlt_0_lowtown_docks_b5i_0_0.vlm"},
	{0xF527C20F22530355ULL, "6094008_m.fxe"},
	{0xF52830563CEE74DEULL, "213639635.ogg"},
	{0xF5283482F5381C46ULL, "356345640.ogg"},
	{0xF5288891F8D314F4ULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.cess.evt"},
	{0xF528E73E890317A1ULL, "zz_rdr_debug6206282.cl"},
	{0xF5293478B7C9FB07ULL, "llac_lt_gallowscourtyard_n_-65537_3.dds"},
	{0xF52943CAB6D05555ULL, "vxw_encounter2_camp_4x_0_1.vlm"},
	{0xF52A5B3126B15A65ULL, "hlac_ht_keep_3733_0z.dds"},
	{0xF52ABEEB55EAEC03ULL, "gen00fl_anders6107481.cl"},
	{0xF52AF9E6147E3A09ULL, "l_dae111ar_gamlens_house_nt.bnk"},
	{0xF52B2591948CF453ULL, "gen00fl_fenris6126378.cl"},
	{0xF52B95B9967FB177ULL, "6163191_m.fxe"},
	{0xF52BACA482468701ULL, "609266475.ogg"},
	{0xF52BBFF8EFE8F9C3ULL, "hlac_lt_0_lowtown_2_1y.dds"},
	{0xF52BC0D89A11E9D9ULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_lp.ani"},
	{0xF52C237FB418E32EULL, "ave310_brennan6136874.cl"},
	{0xF52C36C5D6EF13B0ULL, "6075053_m.fxe"},
	{0xF52C7664DDAB1527ULL, "hlac_dr_primevalentrance_1726_12x.dds"},
	{0xF52CB4C7223C351CULL, "character_character_p243.psh.fxo.dx11"},
	{0xF52CEB753B597177ULL, "vxht_brothel_n_47g_0_0.vlm"},
	{0xF52D00BE0DC354A1ULL, "6141207_m.fxe"},
	{0xF52D60C6A026227BULL, "vxw_bonepitmining_a2_bs_0_0.vlm"},
	{0xF52DCC5F3F4B7C3DULL, "vxht_0_hightown_a2_cb5_0_0.vlm"},
	{0xF52E41E391F94192ULL, "vxht_mansion2b_1dv_0_0.vlm"},
	{0xF52E845FE099F352ULL, "63151350.ogg"},
	{0xF52E9D1A0FBCEC30ULL, "163451781.ogg"},
	{0xF52EC77ADAE75CE3ULL, "vxht_brothel_n_57r_0_0.vlm"},
	{0xF52F76449B90E931ULL, "128334776.ogg"},
	{0xF52FEB9860982935ULL, "w_woundedcoast_12_0.rml"},
	{0xF530902CAC201824ULL, "6076367_f.fxe"},
	{0xF530B80D1DEE6AB3ULL, "lt_0_lowtown_a2_25_0.rml"},
	{0xF5316ECD260DEB97ULL, "follower_banter6160841.cl"},
	{0xF531CB949158AB38ULL, "vxlt_0_lowtown_n_ewz_0_0.vlm"},
	{0xF531E66325F8A344ULL, "vxdr_primevaltomb_8xx_0_0.vlm"},
	{0xF531F2D9459F999FULL, "zz_otl_debug6060669.cl"},
	{0xF53285D9FEF5D582ULL, "6123315_f.fxe"},
	{0xF5329BD89A17C91BULL, "art\\characters\\creatures\\pridedemon\\animation\\tal_grb_lp.gad"},
	{0xF532AB50B2227B61ULL, "440511453.ogg"},
	{0xF532E9371466FC83ULL, "nat380_nathaniel_banter6067314.cl"},
	{0xF532FB5D3FAFACE5ULL, "6083651_m.fxe"},
	{0xF533155D2B376D45ULL, "vxw_bonepit_4q_0_0.vlm"},
	{0xF533255094444C02ULL, "gen00fl_isabela6079019.cl"},
	{0xF533969F2EA5784AULL, "splash_alpha.mao"},
	{0xF533FAF42B751F8FULL, "vxw_bonepitmining_a2_27l_0_0.vlm"},
	{0xF5348148F2C9E3B7ULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd02a.phy"},
	{0xF534AD4C2F808462ULL, "120788134.ogg"},
	{0xF5356D3457851FA4ULL, "llac_dr_primevalentrance_1724_2.dds"},
	{0xF535FC1BF0D74248ULL, "vxlt_warehouse_n_13x_0_0.vlm"},
	{0xF53610B44A85C657ULL, "6052912_m.fxe"},
	{0xF5363AE5F0C14843ULL, "lt_darktown_mmap.dds"},
	{0xF5365EC0659CF2F1ULL, "gen00fl_aveline6139610.cl"},
	{0xF5368091F8DF442DULL, "art\\vfx\\combat\\vfx_r_gn2_accu_loop_c.cess.ani"},
	{0xF536873FE3EA1738ULL, "6169536_m.fxe"},
	{0xF536B76941B0F9E7ULL, "6127685_m.fxe"},
	{0xF536E7D2496C880FULL, "vxw_bonepitmining_a2_1v4_0_0.vlm"},
	{0xF536F7CA121ECF87ULL, "llac_lt_gallowstemplar_f_2391_7.dds"},
	{0xF537C0E536A80D8EULL, "sse_ss_hadriana.gda"},
	{0xF537C8BAA7DB3E39ULL, "vxw_encounter2_5z_0_0.vlm"},
	{0xF537CF671A2D0883ULL, "tbe100ip_store_magic_good.utp"},
	{0xF5389FEE27E2ED90ULL, "vxlt_0_lowtown_eox_0_1.vlm"},
	{0xF538F13891A8185DULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls104a_0n.dds"},
	{0xF5390106A1BDD9AAULL, "vxlt_0_lowtown_a2_n_eod_0_0.vlm"},
	{0xF539177DF1944921ULL, "96555384.ogg"},
	{0xF5392421EFED1CC2ULL, "hlac_lt_0_lowtown_n_65537_4z.dds"},
	{0xF53975219D75A494ULL, "6171829_m.fxe"},
	{0xF53998E8E0162DCEULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.mmh"},
	{0xF53A03B83D889D2DULL, "6050202_m.fxe"},
	{0xF53A67BE71B054CAULL, "lt_0_lowtown_docks_a3_n_120_0.mao"},
	{0xF53A6BBE71B05B19ULL, "lt_0_lowtown_docks_a3_n_120_0.mmh"},
	{0xF53A75BE71B06C07ULL, "lt_0_lowtown_docks_a3_n_120_0.msh"},
	{0xF53AF5E9E3A4F4CEULL, "6086354_m.fxe"},
	{0xF53B09CCE270A725ULL, "gen00fl_anders6142497.cl"},
	{0xF53C32231603044BULL, "vxht_keep_n_6al_0_0.vlm"},
	{0xF53CB036E1A8803FULL, "hlac_dr_0_deeproads_5849_3z.dds"},
	{0xF53CBF5E3100E762ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_highwardrobe.msh"},
	{0xF53CC15E3100EA18ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_highwardrobe.mmh"},
	{0xF53CCD5E3100FEBBULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_highwardrobe.mao"},
	{0xF53D043B3E164FD7ULL, "6182068_m.fxe"},
	{0xF53D1F9AE477F0D3ULL, "ico_2h_sword_4.dds"},
	{0xF53D48E369B5E0A0ULL, "vxw_encounter2_lx_0_0.vlm"},
	{0xF53D80B451E9FCC0ULL, "hlac_w_encounter1_-65533_0z.dds"},
	{0xF53DAA85DB282665ULL, "qun220cr_javaris6071955.cl"},
	{0xF53DB1438FB3A430ULL, "6106777_m.fxe"},
	{0xF53DCEF7DE14E951ULL, "hlac_lt_0_lowtown_a2_n_c_1_8x.dds"},
	{0xF53DD600ECF5AF4BULL, "6031982_m.fxe"},
	{0xF53F1D86A9B74943ULL, "ts_kirk_lowtown_01.gda"},
	{0xF53F97B2A0C84C9CULL, "t3_shd_hst.tnt"},
	{0xF54035A15F64317CULL, "art\\vfx\\textures\\vfx_glyph_damage.dds"},
	{0xF5406A72D0089606ULL, "vxlt_0_lowtown_n_ep5_0_0.vlm"},
	{0xF54071A24685973CULL, "w_encounter2_27_0.mao"},
	{0xF5407DA24685ABDFULL, "w_encounter2_27_0.mmh"},
	{0xF54083A24685B581ULL, "w_encounter2_27_0.msh"},
	{0xF54242A6DBE79BEDULL, "6122143_m.fxe"},
	{0xF54264222975DD51ULL, "739284606.ogg"},
	{0xF54329309CD6E36DULL, "vxca_caverns1_mines_j2d_0_1.vlm"},
	{0xF543549477909B02ULL, "6125383_m.fxe"},
	{0xF5436ECC50835087ULL, "vxht_brothel_n_3zo_0_0.vlm"},
	{0xF5437A7B344DFC36ULL, "follower_banter6167964.cl"},
	{0xF54380821E7523E9ULL, "vxlt_0_lowtown_a3_n_c_gmp_0_0.vlm"},
	{0xF5438403B5BC1A11ULL, "6177206_m.fxe"},
	{0xF543BEB44A8ACC9EULL, "6092597_m.fxe"},
	{0xF543E28C918BA03CULL, "170952163.ogg"},
	{0xF543F49392341725ULL, "1033024848.ogg"},
	{0xF5443CB7CA846BF9ULL, "gen00fl_aveline6129040.cl"},
	{0xF544A31035E6B0CEULL, "follower_banter6166963.cl"},
	{0xF544DA8DEDBF0F80ULL, "hlac_lt_0_lowtown_a2_n_c_65535_1y.dds"},
	{0xF544F34831D31268ULL, "gen00fl_anders6215657.cl"},
	{0xF544F64C33BCB9A6ULL, "vxca_caverns2_80p_0_0.vlm"},
	{0xF54523E54496390EULL, "mag100_cullen6093798.cl"},
	{0xF5454ECC0C27A9CAULL, "dae000_gamlen6124361.cl"},
	{0xF54559AF3B62A08FULL, "vxw_bonepitmining_a2_1s7_0_0.vlm"},
	{0xF54565F013A853C4ULL, "6098594_f.fxe"},
	{0xF545BB32082A6642ULL, "hlac_lt_0_lowtown_a3_n_c_-3_0y.dds"},
	{0xF545DD1F97E61A96ULL, "6162694_m.fxe"},
	{0xF5468B69E8F97A20ULL, "mcr310cr_scared_mage.utc"},
	{0xF546B7E8E020ECCDULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_pen_c.anb"},
	{0xF546C85E31094AA3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_highwardrobe.phy"},
	{0xF546DFCE2DEF3858ULL, "608145415.ogg"},
	{0xF547E095FCA19751ULL, "mag211_marethari6021274.cl"},
	{0xF5489A867B881253ULL, "6000681_m.fxe"},
	{0xF548A171D546260EULL, "536532437.ogg"},
	{0xF548BC184243BFDCULL, "562074976.ogg"},
	{0xF548E07333997F77ULL, "zz_followers_banter6158088.cl"},
	{0xF549944568C5EDB5ULL, "6122114_m.fxe"},
	{0xF54A0528BD2F6242ULL, "vxlt_0_lowtown_a3_n_c_gxx_0_0.vlm"},
	{0xF54A4347E80F993DULL, "qun120_arishok_guard6054845.cl"},
	{0xF54A5498433C54B4ULL, "vxlt_0_lowtown_eqf_0_1.vlm"},
	{0xF54B10CF6E0C5C47ULL, "815473524.ogg"},
	{0xF54B259D75E57292ULL, "vxht_mansion2b_1rp_0_0.vlm"},
	{0xF54B4DE0F1005E1FULL, "115693778.ogg"},
	{0xF54BD6A5AAEF04C2ULL, "6043260_f.fxe"},
	{0xF54C59453B7969F6ULL, "6082156_m.fxe"},
	{0xF54C6FE4C50C99D1ULL, "6135081_f.fxe"},
	{0xF54C971D550099E2ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.impa.gad"},
	{0xF54D2808C4E2F8BEULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1.mmh"},
	{0xF54D5F60D55860CDULL, "gen00fl_dog6167066.cl"},
	{0xF54DB543B7AEBBFCULL, "6119409_m.fxe"},
	{0xF54E037DB466D3CFULL, "vxlt_blackemporium_0_0_0.vlm"},
	{0xF54E1DBE7AE7A590ULL, "vxlt_0_lowtown_a3_n_c_gvt_0_0.vlm"},
	{0xF54E81E9369E4C21ULL, "vxw_sundermount_2ko_0_0.vlm"},
	{0xF54E88F056B2292EULL, "vxlt_gallowstemplar_n_5pf_0_0.vlm"},
	{0xF54EAF4E7544DC38ULL, "vxlt_0_lowtown_n_emx_0_0.vlm"},
	{0xF54EB6DD38B40E64ULL, "vxlt_0_lowtown_a2_eqk_0_0.vlm"},
	{0xF54F149320D68556ULL, "llac_ca_cavern1_haunted_24421_5.dds"},
	{0xF54F32572EDA9C96ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_008a_0s.dds"},
	{0xF54FA24DDB3FA264ULL, "w_blightlands_106_0.msh"},
	{0xF54FB44DDB3FC0CDULL, "w_blightlands_106_0.mao"},
	{0xF54FB84DDB3FC706ULL, "w_blightlands_106_0.mmh"},
	{0xF550619101108105ULL, "6106888_m.fxe"},
	{0xF5506A1A282BBA99ULL, "vxw_woundedcoast_nj_0_0.vlm"},
	{0xF550A2CAEC660A87ULL, "llac_w_encounter2_camp_262146_0.dds"},
	{0xF550DEAE0F61508FULL, "vxw_bonepitmining_a2_29n_0_0.vlm"},
	{0xF550FD1B2670BA71ULL, "zz_vault_debug6178952.cl"},
	{0xF55166A6E3A8320EULL, "6140198_f.fxe"},
	{0xF5517CDAE165489DULL, "lt_0_lowtown_a3_n_c_35_0_lc.mmh"},
	{0xF551A1AADFA375ACULL, "genip_inkwell.utp"},
	{0xF55274A5D918FD13ULL, "lgt000cr_dwf_turn_in_5.utc"},
	{0xF552AD705725B29BULL, "6115936_m.fxe"},
	{0xF552BA1D550540D9ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.impa.evt"},
	{0xF552BCD7F63CD0B0ULL, "vxht_0_hightown_du_0_0.vlm"},
	{0xF552CDFE09EA94EBULL, "dae000ip_crowd_mf3_talk4.utp"},
	{0xF553117E349B70A9ULL, "6136722_m.fxe"},
	{0xF55352E5D1361503ULL, "77754344.ogg"},
	{0xF5535C49FD03BF25ULL, "394835319.ogg"},
	{0xF5537B3F896336B7ULL, "hlac_ca_caverns2prc_5802_21x.dds"},
	{0xF5544EFC89DC3316ULL, "6167594_m.fxe"},
	{0xF55465F7568D6337ULL, "895292934.ogg"},
	{0xF554AD8FE18331C6ULL, "vxlt_0_lowtown_docks_n_bs2_0_0.vlm"},
	{0xF554E09A82FE5F76ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_arm_wic_0d.dds"},
	{0xF554ECA8AD8C6261ULL, "hlac_dr_slavers_3207_8y.dds"},
	{0xF5552B3C09276957ULL, "6110175_m.fxe"},
	{0xF55538825FC81C2FULL, "lt_house_mmi.gff"},
	{0xF555C8F18AB68FCAULL, "hightown_market_guard_sta6169651.cl"},
	{0xF555DEC629CFFDCEULL, "vxw_bonepitmining_a2_1vy_0_0.vlm"},
	{0xF555E4C085AA0AAAULL, "vxlt_0_lowtown_a2_n_euf_0_0.vlm"},
	{0xF5561D772690A393ULL, "vxlt_darktown_2va_0_0.vlm"},
	{0xF55676DFAB4AC366ULL, "vxlt_undercitydungeon_1nh_0_0.vlm"},
	{0xF55770BEF68824BAULL, "mrl210_merrill6197793.cl"},
	{0xF557A77EAACEA719ULL, "6035359_m.fxe"},
	{0xF5580398A20A5ABEULL, "qun111_ketojan6040781.cl"},
	{0xF55863BACB5BFFD3ULL, "hlac_ht_keep_4519_18z.dds"},
	{0xF558B621AD4AAC3FULL, "qun131_taskmaster6019208.cl"},
	{0xF55920CD156ADB90ULL, "231915573.ogg"},
	{0xF5592E9128D86261ULL, "lt_gallowscourtyard_46_0.rml"},
	{0xF55959F5B7297F4CULL, "6065248_f.fxe"},
	{0xF5595CE86D9B27D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v1.gad"},
	{0xF559973C89DE1259ULL, "6107111_m.fxe"},
	{0xF559977A67DB8B12ULL, "vxht_mansion2b_1uq_0_0.vlm"},
	{0xF559AF4DDB480CD1ULL, "w_blightlands_106_0.phy"},
	{0xF559BC32EE5CB357ULL, "256418355.ogg"},
	{0xF55A2C9D4DA211A0ULL, "vxw_encounter2_rf_0_0.vlm"},
	{0xF55A756CC99F28D3ULL, "hlac_lt_0_lowtown_a2_n_65534_0x.dds"},
	{0xF55ACB222D6F8276ULL, "6085056_m.fxe"},
	{0xF55B302E0946A8C5ULL, "dre000_varric6034630.cl"},
	{0xF55B55419D0D5CB0ULL, "zz_vault_debug6177796.cl"},
	{0xF55B636B17DFFE56ULL, "528226843.ogg"},
	{0xF55BE521BE1FAB3DULL, "270427463.ogg"},
	{0xF55C33315B05E25BULL, "lgt250cr_way_dalish_hunter.utc"},
	{0xF55CD0CDFBBC8633ULL, "zz_posttest.arl"},
	{0xF55CF3C6289464A1ULL, "vxw_sundermount_2ln_0_0.vlm"},
	{0xF55D4E2CB3F4C706ULL, "vxlt_0_lowtown_n_ef7_0_0.vlm"},
	{0xF55DCD71D7CB12ECULL, "hlac_lt_0_lowtown_n_0_8z.dds"},
	{0xF55E2830A10D158AULL, "397076336.ogg"},
	{0xF55E40119494B317ULL, "hlac_lt_0_lowtown_docks_-65537_4x.dds"},
	{0xF55E8A01B3681231ULL, "hlac_lt_0_lowtown_a2_131070_2z.dds"},
	{0xF55F4306A2E33A08ULL, "gen_im_trs_bst_liver_02.uti"},
	{0xF55F5FE86D9F9809ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v1.evt"},
	{0xF55FF1261E93DBF2ULL, "vxlt_blackemporium_ir_0_0.vlm"},
	{0xF5603E14ED07FFBAULL, "6167958_m.fxe"},
	{0xF56062C0A8145ACBULL, "ave200_donnic6061402.cl"},
	{0xF560C21D55118BC8ULL, "art\\vfx\\creatures\\fxc_rage_crust_c.impa.ani"},
	{0xF560FABE95A603E9ULL, "31839308.ogg"},
	{0xF5611F4D491BCB1BULL, "6110636_m.fxe"},
	{0xF56125E6DF606A52ULL, "6036560_m.fxe"},
	{0xF5612C910CEB4BACULL, "mag370_meredith6064578.cl"},
	{0xF56188123DD2B444ULL, "56234969.ogg"},
	{0xF561D0B06B4D1E26ULL, "vxht_0_hightown_b5v_0_0.vlm"},
	{0xF561D5F56CDF0FE4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_1.gad"},
	{0xF5621A2B60F16FEDULL, "lt_0_lowtown_a2_12_0.rml"},
	{0xF5624B91272066E1ULL, "hlac_lt_0_lowtown_a2_196607_0y.dds"},
	{0xF562BF2EE766AB5BULL, "6143715_m.fxe"},
	{0xF562C18BD230F104ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\mov_p_injured.ani"},
	{0xF562D36B0ADBD60EULL, "w_sundermount_mmap.dds"},
	{0xF56371ADA801BC39ULL, "vxw_blightlands_hr_0_0.vlm"},
	{0xF563B021291DAEF2ULL, "hlac_ht_mansion_n_3373_2z.dds"},
	{0xF563F68B0E1344AFULL, "332121867.ogg"},
	{0xF5646953AF0FD0AEULL, "dae000_arishok6072399.cl"},
	{0xF56473DC61CB1AD1ULL, "6116517_m.fxe"},
	{0xF564822109F08351ULL, "6156760_m.fxe"},
	{0xF56487614336D71CULL, "cod_hst_vehnadahl.utp"},
	{0xF56542FF0EFA227EULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_en.evt"},
	{0xF56558FAF724F841ULL, "vxlt_0_lowtown_docks_n_b00_0_0.vlm"},
	{0xF5656788B03DD970ULL, "ser270_emeric6201576.cl"},
	{0xF5661BA33AC1AD67ULL, "lt_0_lowtown_docks_n_31_0_lc.mmh"},
	{0xF566E1DC92B4ED5EULL, "hlac_ht_0_hightown_a3_-196607_0z.dds"},
	{0xF566EDB45BDDB799ULL, "vxht_0_hightown_a2_c29_0_0.vlm"},
	{0xF567C0B01C5B493AULL, "6147574_m.fxe"},
	{0xF567EB7950F73EACULL, "aip_kirkwallguard_base.gda"},
	{0xF568E4F56CE547C2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\cindes\\dialogue\\guard_stairs_loop_1.ani"},
	{0xF568E7C9D9B30984ULL, "vxw_woundedcoast_1ek_0_0.vlm"},
	{0xF56911097B8AB344ULL, "vrc221_varric6048385.cl"},
	{0xF5694DE7732B635DULL, "vxdr_0_deeproads_ebl_0_0.vlm"},
	{0xF569BB2C20237386ULL, "deathscreen.gfx"},
	{0xF569BE8BD2370B66ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\mov_p_injured.gad"},
	{0xF56AE46289AF078EULL, "art\\vfx\\combat\\vfx_m_cre_haste_act_c.impa.evt"},
	{0xF56B016672F82FADULL, "fex_ruin_spt_c.msh"},
	{0xF56B1298EE30106DULL, "262501605.ogg"},
	{0xF56B136672F84E33ULL, "fex_ruin_spt_c.mmh"},
	{0xF56B5ABC7F88A36BULL, "lgt120_orz_bln_traitor_lead6140235.cl"},
	{0xF56B65A31A8A7D21ULL, "vxw_sundermount_2qm_0_0.vlm"},
	{0xF56B80E0EBB61371ULL, "6146793_f.fxe"},
	{0xF56BF2AA2FC9A17AULL, "6095862_m.fxe"},
	{0xF56C58E1E6DDCA03ULL, "163134058.ogg"},
	{0xF56C59FF0F0067A9ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_en.gad"},
	{0xF56C6995C3DC8A4DULL, "llac_lt_gallowstemplar_f_2394_8.dds"},
	{0xF56C920BC2A7A14EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clapping.ani"},
	{0xF56C9B6E262670A5ULL, "vxlt_0_lowtown_a2_n_c_ep1_0_1.vlm"},
	{0xF56C9C9FEF3435A9ULL, "6168174_m.fxe"},
	{0xF56CCA797194BA32ULL, "hlac_w_bonepitmining_a2_196614_0y.dds"},
	{0xF56CEB8B491C9A54ULL, "112090100.ogg"},
	{0xF56D47E86DABAD78ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_attk_med_rng_v1.ani"},
	{0xF56D649FF5C3F42DULL, "weapon_vfx_p0.psh.fxo.dx11.assembly"},
	{0xF56D999761A8C60DULL, "6129209_m.fxe"},
	{0xF56DAE4D4C8AE02EULL, "sb_fade_base.mao"},
	{0xF56DBB53E5C6601EULL, "art\\vfx\\cinematic\\vfx_merrill_reflect_c.impa.ani"},
	{0xF56DF1D5087770BAULL, "fen110_fenris6117927.cl"},
	{0xF56DF8555F12434BULL, "lt_gallowscourtyard_n_30_0.rml"},
	{0xF56E0754CD1985C4ULL, "183401085.ogg"},
	{0xF56E69F3024D03E0ULL, "seb221st_harimann_brett.stg"},
	{0xF56EAFF63AB5C43AULL, "follower_banter6136355.cl"},
	{0xF56EBD8969753C6DULL, "qun160_vashothlead6053934.cl"},
	{0xF56EE36DB098EDB3ULL, "159199544.ogg"},
	{0xF56F096D1F36B894ULL, "6005292_m.fxe"},
	{0xF56F31303EAAB860ULL, "vxlt_gallowscourtyard_4xl_0_0.vlm"},
	{0xF56F4ACE36223979ULL, "6064656_m.fxe"},
	{0xF56FCCFEE906AE0AULL, "vxca_caverns2prc_8s0_0_0.vlm"},
	{0xF57007DD2C6BCD43ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.cess.evt"},
	{0xF571078CD769B2FFULL, "vxca_caverns2_814_0_2.vlm"},
	{0xF571556B576F717DULL, "80098077.ogg"},
	{0xF571676043D3FF82ULL, "llac_lt_0_lowtown_65536_2.dds"},
	{0xF571DB6289B51739ULL, "art\\vfx\\combat\\vfx_m_cre_haste_act_c.impa.gad"},
	{0xF571E38A99F7D4B9ULL, "vxw_blightlands_mq_0_0.vlm"},
	{0xF572C1D572B110F2ULL, "411067708.ogg"},
	{0xF572D7E614BBB65BULL, "l_dae111ar_gamlens_house_nt.wbd"},
	{0xF572E9745CEF3A85ULL, "6033634_m.fxe"},
	{0xF5732AFF0F0636F7ULL, "art\\characters\\creatures\\mubari\\animation\\dg_f_r_5p_en.ani"},
	{0xF573E6B790485AB6ULL, "6105998_m.fxe"},
	{0xF573EE396D63E10CULL, "6047485_m.fxe"},
	{0xF574091DAD1E14F2ULL, "hf_ave3_insurgent1.mor"},
	{0xF5740BBA10BA482BULL, "6000898_m.fxe"},
	{0xF575ADF466C57876ULL, "6167250_f.fxe"},
	{0xF575D2F6C17803FCULL, "zz_approval_debug6003092.cl"},
	{0xF576E0DD2C71A9C4ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.cess.gad"},
	{0xF576F6CA0C64A6FDULL, "vxlt_gallowstemplar_n_651_0_0.vlm"},
	{0xF57782A8B872D7FDULL, "6051577_m.fxe"},
	{0xF57782D9DF1C04C3ULL, "vxht_0_hightown_n_a6j_0_0.vlm"},
	{0xF5781B145825F484ULL, "lt_ruinedhouse_n_mmap_ph.dds"},
	{0xF578EC6289BB5287ULL, "art\\vfx\\combat\\vfx_m_cre_haste_act_c.impa.ani"},
	{0xF578F325C5AEE88CULL, "hlac_ht_0_hightown_a2_-131074_0z.dds"},
	{0xF5790C4E855B74A3ULL, "vxlt_gallowstemplar_n_3r3_0_0.vlm"},
	{0xF57917CBA4A5EDB2ULL, "qun211_party_comment6072271.cl"},
	{0xF579304F653C471FULL, "follower_banter6143110.cl"},
	{0xF5798E9A0C8DCE44ULL, "6100032_m.fxe"},
	{0xF579B35DF44F869EULL, "ts_useabi.ncs"},
	{0xF579B6028EA19F12ULL, "hla_ht_0_hightown_-2_0y.dds"},
	{0xF579D7800C8095A1ULL, "vxw_sundermount_2zl_0_0.vlm"},
	{0xF579D82BA8BA1E1DULL, "hlac_lt_0_lowtown_a2_n_131072_1z.dds"},
	{0xF579DE872C9472AEULL, "vxlt_gallowstemplar_n_5ay_0_0.vlm"},
	{0xF57A307362825366ULL, "6118462_m.fxe"},
	{0xF57A8A0BC2B3D09FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clapping.evt"},
	{0xF57AF8B82FDD9799ULL, "blood_conj_06a.mao"},
	{0xF57B4B2345CDBFE9ULL, "art\\characters\\playercharacter\\humanmale\\mt_bas_hm_scalpm1.mao"},
	{0xF57B6248F305B1ACULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd02a.mmh"},
	{0xF57B6848F305BB9EULL, "art\\characters\\playercharacter\\elfmale\\em_har_hd02a.msh"},
	{0xF57B8BC54F8BEF5EULL, "436126.ogg"},
	{0xF57BD353E5D2C6EFULL, "art\\vfx\\cinematic\\vfx_merrill_reflect_c.impa.evt"},
	{0xF57CA9AA3A324D1BULL, "6084731_m.fxe"},
	{0xF57CEFDD2C762EA2ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_aoe_p.cess.ani"},
	{0xF57D1F6A0DE2CC51ULL, "and230_mage_girl6135822.cl"},
	{0xF57D53D2A5346909ULL, "6147209_m.fxe"},
	{0xF57DBCAB281851F9ULL, "6063756_m.fxe"},
	{0xF57DD85057D8255FULL, "vxht_brothel_n_62v_0_0.vlm"},
	{0xF57DEBC5F46881C6ULL, "mag160_feynriel6037373.cl"},
	{0xF57DFCDD5A56B49CULL, "gen00fl_varric6152695.cl"},
	{0xF57F594257C6FF69ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_reaction_left_shoulder.ani"},
	{0xF57F87A2E49B9F5AULL, "vxw_woundedcoast_17s_0_0.vlm"},
	{0xF58012D14C0621D2ULL, "dimscreen.mao"},
	{0xF5803AE6D347A84AULL, "zz_refugeemine.cut"},
	{0xF58055678BEDED39ULL, "vxw_blightlands_vp_0_0.vlm"},
	{0xF5807D75FD7063FFULL, "hlac_lt_0_lowtown_a2_131073_1x.dds"},
	{0xF580B712DF46DAD6ULL, "m_mag100ar_eau_road.wbd"},
	{0xF58102F430A654A1ULL, "6176573_m.fxe"},
	{0xF5816B0BC2B9BA88ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_clapping.gad"},
	{0xF5816CB1536E9543ULL, "docks_male_elven_worker_sta6175845.cl"},
	{0xF58184478D1608D1ULL, "hlac_lt_darktown_-1_70z.dds"},
	{0xF581ACCE4CED246AULL, "6033659_m.fxe"},
	{0xF581D453E5D73398ULL, "art\\vfx\\cinematic\\vfx_merrill_reflect_c.impa.gad"},
	{0xF581F820A3BD09B9ULL, "6131892_f.fxe"},
	{0xF58266A4A0BB1617ULL, "ht_0_hightownexplosion_n_36_0.rml"},
	{0xF582AB69678FAA0EULL, "6044532_m.fxe"},
	{0xF582B63866F7D748ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medg.mmh"},
	{0xF582C23866F7EBEBULL, "art\\characters\\playercharacter\\humanmale\\hlf_medg.mao"},
	{0xF582C7C69DE93364ULL, "vxht_0_hightown_n_b9g_0_0.vlm"},
	{0xF582D43866F80A72ULL, "art\\characters\\playercharacter\\humanmale\\hlf_medg.msh"},
	{0xF582F7765BA67A6EULL, "hlac_lt_undercityserialkiller_-1_4x.dds"},
	{0xF58300FFABC61EF4ULL, "vxdr_0_deeproads_cf3_0_0.vlm"},
	{0xF58324AB479A82DCULL, "vxdr_primevaltomb_52j_0_1.vlm"},
	{0xF5832FD7401377E4ULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfood.phy"},
	{0xF583381F74C12568ULL, "384754797.ogg"},
	{0xF583AAF1BFB01A93ULL, "613995471.ogg"},
	{0xF583C8C05A7D37B0ULL, "614525854.ogg"},
	{0xF583D347EBE10A45ULL, "vxca_caverns2prc_907_0_0.vlm"},
	{0xF584A581B7CF030AULL, "vxca_cave_22s_0_0.vlm"},
	{0xF584A5C03668FAA6ULL, "gen00fl_aveline6137648.cl"},
	{0xF584D3B50AFDDAC4ULL, "plt_cod_crt_hurlock.plo"},
	{0xF58502EB421DDA08ULL, "833215680.ogg"},
	{0xF58561F16A44C7BBULL, "follower_banter6172820.cl"},
	{0xF5858BE85A1C11A9ULL, "art\\vfx\\combat\\vfx_gib_qun_rtl.ani"},
	{0xF5859643A8B5E911ULL, "gencr_stone_golem.utc"},
	{0xF585CB83BD9F3A84ULL, "vxw_woundedcoast_1ki_0_0.vlm"},
	{0xF585DE83AFAF3D4AULL, "vxlt_0_lowtown_docks_a3_cxb_0_0.vlm"},
	{0xF5863477613C86F7ULL, "vxdr_primevaltomb_3jb_0_0.vlm"},
	{0xF586412C3E2EA477ULL, "6142468_m.fxe"},
	{0xF586744257CD4BEBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_reaction_left_shoulder.gad"},
	{0xF58677146E08286EULL, "hlac_ca_caverns2prc_5802_2x.dds"},
	{0xF586C44BE4A7F953ULL, "321291498.ogg"},
	{0xF5871BF8C3649EC1ULL, "vxca_caverns2_95a_0_0.vlm"},
	{0xF58737ACEE227A9EULL, "6139535_m.fxe"},
	{0xF58751D97D050F41ULL, "645279325.ogg"},
	{0xF5875BFEB55E3038ULL, "mcr371_angry_mage_m6087029.cl"},
	{0xF587E7705C061503ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_low.mao"},
	{0xF587EB705C061C50ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_low.mmh"},
	{0xF587F9705C06346AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_low.msh"},
	{0xF5881C6C3521DE8AULL, "art\\characters\\creatures\\spider\\animation\\rn_lnr.gad"},
	{0xF588307CF1265AE6ULL, "llac_lt_0_lowtown_docks_a3_-262143_0.dds"},
	{0xF5884712C07F851AULL, "6103567_m.fxe"},
	{0xF588C0B4401D5DA3ULL, "902933867.ogg"},
	{0xF58955580D7CD503ULL, "hlac_lt_0_lowtown_a2_n_1_4x.dds"},
	{0xF5897C0A98AFE580ULL, "859945392.ogg"},
	{0xF58A28374908848DULL, "6064111_m.fxe"},
	{0xF58A30CBE18059E2ULL, "vxdr_primevaltomb_3i2_0_0.vlm"},
	{0xF58A318DF0041C99ULL, "vxw_woundedcoast_zn_0_0.vlm"},
	{0xF58AD5057E77171FULL, "qun211cr_braggart.utc"},
	{0xF58ADD65CD5F5071ULL, "vxw_sundermount_246_0_0.vlm"},
	{0xF58ADE51DE46DE75ULL, "723410987.ogg"},
	{0xF58BC5AF972A8189ULL, "6140309_m.fxe"},
	{0xF58C0F4462AD4A7CULL, "vxlt_undercitydungeon_17i_0_0.vlm"},
	{0xF58C13B754B933EFULL, "hlac_lt_0_lowtown_a2_n_c_65538_4z.dds"},
	{0xF58CA6E85A225E2BULL, "art\\vfx\\combat\\vfx_gib_qun_rtl.gad"},
	{0xF58D19496550AEBCULL, "vxht_brothel_n_59k_0_0.vlm"},
	{0xF58D2CB5823FDD5CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlava.mao"},
	{0xF58D38B5823FF1FFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlava.mmh"},
	{0xF58D39783AB0B407ULL, "638580977.ogg"},
	{0xF58D3EB5823FFBA1ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\9ff_cavernlava.msh"},
	{0xF58D58756C974685ULL, "242756898.ogg"},
	{0xF58D714257D36550ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_c_reaction_left_shoulder.evt"},
	{0xF58DC069444409EFULL, "41235925.ogg"},
	{0xF58DE14B8B019E48ULL, "6037086_m.fxe"},
	{0xF58DF97FD691B7DAULL, "vxw_woundedcoast_12l_0_0.vlm"},
	{0xF58DFF08C51A4A59ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1.phy"},
	{0xF58E36746C29B849ULL, "one280_iwan6096929.cl"},
	{0xF58E5BF92CBBD7E4ULL, "vxdr_primevalentrance_37f_0_0.vlm"},
	{0xF58E9B2608A53A25ULL, "6169515_m.fxe"},
	{0xF58EA9AA1B9932DCULL, "vxdr_primevaltomb_6v1_0_1.vlm"},
	{0xF58EB3740D7F0CF9ULL, "vxlt_0_lowtown_a2_n_fer_0_0.vlm"},
	{0xF58EBACA2AA87758ULL, "6108308_m.fxe"},
	{0xF58ECC3B4A24C8B3ULL, "dae000_merchant_generic_hm6172662.cl"},
	{0xF58EF3C006F3DD55ULL, "fxe_ripple_t2.mao"},
	{0xF58EFF6C3527CC21ULL, "art\\characters\\creatures\\spider\\animation\\rn_lnr.evt"},
	{0xF58F281294AF72E1ULL, "1028552200.ogg"},
	{0xF59102DCFB4D5592ULL, "hlac_lt_0_lowtown_docks_a3_n_-262144_1y.dds"},
	{0xF5910A594FF35709ULL, "118771903.ogg"},
	{0xF5918189453BA9A4ULL, "vxlt_0_lowtown_n_fgi_0_0.vlm"},
	{0xF5918928244C14CAULL, "vxw_woundedcoast_14_0_0.vlm"},
	{0xF591F4DAD6E2858EULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowbutress.msh"},
	{0xF59202DAD6E29DDFULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowbutress.mao"},
	{0xF5920EDAD6E2B27CULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowbutress.mmh"},
	{0xF592139807062DD9ULL, "6059241_m.fxe"},
	{0xF5926ED88BE6C3F3ULL, "6171538_m.fxe"},
	{0xF59282ADD48C8C78ULL, "vxht_0_hightown_a3_c9x_0_0.vlm"},
	{0xF59287D08B2AA035ULL, "hlac_dr_0_deeproads_5849_8y.dds"},
	{0xF592C77770CE13BCULL, "vxw_woundedcoast_n3_0_0.vlm"},
	{0xF592F8EA5714E0FEULL, "vxw_bonepitmining_a2_dd_0_0.vlm"},
	{0xF592FBEEF931A1F6ULL, "437803517.ogg"},
	{0xF5930E8F9ADA57C1ULL, "hlac_lt_0_lowtown_n_-262144_4y.dds"},
	{0xF593F26E9F1F232BULL, "29102638.ogg"},
	{0xF594A766169575ECULL, "art\\vfx\\materials\\chunky\\vfx_tube.mao"},
	{0xF59579711D80BA3CULL, "hlac_lt_0_lowtown_-262144_0y.dds"},
	{0xF5965C25B12B1CD9ULL, "112573031.ogg"},
	{0xF596DB3BAB7C592CULL, "6038230_m.fxe"},
	{0xF596F02DF28288E4ULL, "vxlt_0_lowtown_a2_ehl_0_0.vlm"},
	{0xF597142D466E8466ULL, "6112662_m.fxe"},
	{0xF59789C3143BECE1ULL, "llac_lt_darktown_-1_72.dds"},
	{0xF59817F6162EB5EFULL, "784589776.ogg"},
	{0xF5983BB9D5141F7FULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_basic.msh"},
	{0xF59841B9D5142921ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_basic.mmh"},
	{0xF5984DB9D5143DC2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_basic.mao"},
	{0xF598A1F15C148CE6ULL, "mag100_cullen6140991.cl"},
	{0xF598F821584997B2ULL, "hlac_lt_0_lowtown_131070_1z.dds"},
	{0xF5994F42BF5568F1ULL, "vxw_sundermount_295_0_0.vlm"},
	{0xF59A4A79F1B34770ULL, "vxlt_0_lowtown_docks_a3_n_d2n_0_0.vlm"},
	{0xF59A606667CBE631ULL, "zz_approval_debug6003929.cl"},
	{0xF59AB8910BE32842ULL, "411892414.ogg"},
	{0xF59AD934B289F83CULL, "vxht_0_hightown_a2_c7c_0_0.vlm"},
	{0xF59AFB9339BFB309ULL, "259176173.ogg"},
	{0xF59B25A5F9B5674BULL, "271004957.ogg"},
	{0xF59B37F35A9457B4ULL, "205592946.ogg"},
	{0xF59BF76C35324970ULL, "art\\characters\\creatures\\spider\\animation\\rn_lnr.ani"},
	{0xF59BFB1E4CFBC20BULL, "llac_w_encounter3_-65536_0.dds"},
	{0xF59C00586E7C8650ULL, "vxht_bartrand_n_67z_0_0.vlm"},
	{0xF59C43D5DE4F81ECULL, "dae100_viscount6102013.cl"},
	{0xF59C6CBE72032706ULL, "lt_0_lowtown_docks_a3_n_120_0.phy"},
	{0xF59C76B9E99107A8ULL, "mer340cr_cowan.utc"},
	{0xF59C9A6FCD8BCCF8ULL, "6004112_m.fxe"},
	{0xF59DD323B66B26B7ULL, "vxlt_undercityserialkiller_1j4_0_0.vlm"},
	{0xF59E451BD468A8C6ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_lb105a_0s.dds"},
	{0xF59EDD171A782563ULL, "follower_banter6141453.cl"},
	{0xF59EFBF2A656B60DULL, "dae000_gamlen6048439.cl"},
	{0xF59F547AAF692D56ULL, "dae000_bartrand6205983.cl"},
	{0xF59FA0E232B82420ULL, "vxht_brothel_n_6cv_0_0.vlm"},
	{0xF59FC0B0D0A256E1ULL, "hlac_lt_0_lowtown_65536_10x.dds"},
	{0xF5A029FDC1E8A087ULL, "142863029.ogg"},
	{0xF5A0CE373555E6B2ULL, "357215720.ogg"},
	{0xF5A0F48AC682A4F8ULL, "vxht_0_hightown_a3_c2y_0_0.vlm"},
	{0xF5A1032DD81BCF62ULL, "fen110_fenris6101951.cl"},
	{0xF5A16AC7490AF97EULL, "vxw_bonepitmining_a2_ce_0_0.vlm"},
	{0xF5A1BC0FE030CE04ULL, "938309.ogg"},
	{0xF5A1C45B87893AA1ULL, "zz_vrc_debug6200173.cl"},
	{0xF5A1DBDDAF3E4524ULL, "dae000_gen_turn_in_46178440.cl"},
	{0xF5A1FD947918A61CULL, "qun110_petrice6018534.cl"},
	{0xF5A244419FC313D4ULL, "vxlt_darktown_4f4_0_0.vlm"},
	{0xF5A249F0A8101AF3ULL, "vxht_mansion2_1vq_0_0.vlm"},
	{0xF5A286A246D88524ULL, "w_encounter2_27_0.phy"},
	{0xF5A2AF3DA18B6B84ULL, "vxw_woundedcoast_1qo_0_0.vlm"},
	{0xF5A2E0789EF46BA5ULL, "mag131_keran6094503.cl"},
	{0xF5A2E1B9E20EE79BULL, "1010762615.ogg"},
	{0xF5A3C0840C07C825ULL, "art\\vfx\\combat\\vfx_gib_qunw_rtl.anb"},
	{0xF5A3E4252283C933ULL, "art\\levels\\races\\proxy\\props\\rugs_rich\\prp_zzsmallred_0.pwk"},
	{0xF5A402ABAC94F8B2ULL, "575029396.ogg"},
	{0xF5A42C2FBD4DEBBFULL, "6206560_m.fxe"},
	{0xF5A4711067CF67D6ULL, "vxlt_gallowstemplar_3r5_0_0.vlm"},
	{0xF5A47747F2D8F399ULL, "539289935.ogg"},
	{0xF5A4D28F9FC5A9AAULL, "6018649_m.fxe"},
	{0xF5A5394F371DB8BCULL, "229399482.ogg"},
	{0xF5A5B5152B961511ULL, "hlac_dr_primevalentrance_1726_7z.dds"},
	{0xF5A5BB99334984B8ULL, "art\\characters\\creatures\\pridedemon\\animation\\blendtree\\pridedemon_movement.anb"},
	{0xF5A6E41968033EAEULL, "gen00fl_aveline6116678.cl"},
	{0xF5A71975296F9C6CULL, "vxca_caverns2_97z_0_0.vlm"},
	{0xF5A7CC21990D01A3ULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.anb"},
	{0xF5A7ECFE4623F025ULL, "vxlt_backalley_n_1en_0_0.vlm"},
	{0xF5A83E728D0E8074ULL, "vxlt_gallowscourtyard_n_69f_0_0.vlm"},
	{0xF5A8957F17898B77ULL, "442117837.ogg"},
	{0xF5A8C1942B51A47DULL, "dre000_sibling_has_blight6036484.cl"},
	{0xF5A8C9E794099566ULL, "t3_wdg_w10.tnt"},
	{0xF5A9444CF17B4EC5ULL, "731400868.ogg"},
	{0xF5A97DA246DE94ACULL, "w_encounter2_27_0.rml"},
	{0xF5A99F2AD60E4E8DULL, "358259319.ogg"},
	{0xF5A9C32A7E78C7C1ULL, "art\\characters\\creatures\\strider\\textures\\c_stridera_0n.dds"},
	{0xF5A9DC6A2EA77515ULL, "6095004_f.fxe"},
	{0xF5AA0ACE79642D21ULL, "ht_0_hightown_12_0_ll.mmh"},
	{0xF5AA4F708F9204CDULL, "mag311_emile_de_launcet6043804.cl"},
	{0xF5AA97A252FAC1AAULL, "6052979_m.fxe"},
	{0xF5AAB5A35F099AF4ULL, "vxht_chantry_2t3_0_0.vlm"},
	{0xF5AADD39BA7DE8DAULL, "vxw_woundedcoast_18n_0_0.vlm"},
	{0xF5AB4D62ACF221CEULL, "lt_0_lowtown_docks_a3_2_0_ll.mmh"},
	{0xF5ABF2F30272D6DFULL, "vxw_encounter2_kn_0_0.vlm"},
	{0xF5AC4500A8613F37ULL, "vxlt_undercityserialkiller_1o7_0_0.vlm"},
	{0xF5AC550D72163778ULL, "hlac_w_encounter1_-2_0z.dds"},
	{0xF5AC88AA135FFB1DULL, "6052767_m.fxe"},
	{0xF5ACA81CD96145B8ULL, "vxw_bonepitmining_a2_f8_0_0.vlm"},
	{0xF5ACC43BB4AF707AULL, "gen00fl_aveline6139844.cl"},
	{0xF5ACDD12DF6C5644ULL, "m_mag100ar_eau_road.bnk"},
	{0xF5ACF5934B5AAAE7ULL, "84448395.ogg"},
	{0xF5AD24B2E8E7FD28ULL, "vxlt_gallowscourtyard_n_6qb_0_0.vlm"},
	{0xF5AD6F70609A6A75ULL, "hlac_lt_0_lowtown_docks_a3_n_65535_3z.dds"},
	{0xF5ADEF87E61185B7ULL, "gen00fl_fenris_base_act3_1.stg"},
	{0xF5AE05139D2E7CC3ULL, "distor.dds"},
	{0xF5AE92F614D427E7ULL, "follower_banter6156149.cl"},
	{0xF5AEE7B5FF7D0E85ULL, "6139613_m.fxe"},
	{0xF5AF556EE9336E63ULL, "6108934_m.fxe"},
	{0xF5AF6667B878BD78ULL, "vxht_0_hightown_a3_c7v_0_0.vlm"},
	{0xF5AF77903A4E9ABFULL, "284060436.ogg"},
	{0xF5AF8789BF5CED43ULL, "zz_mjarvis_test2.cut"},
	{0xF5B00F9267E92A50ULL, "gen00fl_fenris6115351.cl"},
	{0xF5B067F5AF8348E7ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_conj_c.impa.gad"},
	{0xF5B06DEAC7938592ULL, "leap_cyc.mao"},
	{0xF5B0EA9367DACBCBULL, "llac_w_encounter2_camp_327685_0.dds"},
	{0xF5B17457CBFD2A07ULL, "444436451.ogg"},
	{0xF5B1A889B8EAC122ULL, "vxdr_primevaltomb_4hr_0_0.vlm"},
	{0xF5B1EBDC0B309574ULL, "629490387.ogg"},
	{0xF5B21B3716AA50FFULL, "138436363.ogg"},
	{0xF5B22BBC62E76EACULL, "260196062.ogg"},
	{0xF5B238A977624444ULL, "vxw_woundedcoast_ha_0_0.vlm"},
	{0xF5B247FC6C54C866ULL, "vxca_caverns2_7xn_0_1.vlm"},
	{0xF5B2EF6E0BBC507AULL, "vxht_chantry_n_35x_0_0.vlm"},
	{0xF5B33488815E4A3EULL, "hlac_ht_chantry_n_2534_20z.dds"},
	{0xF5B347D08C322A22ULL, "dadefaultsub.sub.xml"},
	{0xF5B35B7AED20F602ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comd_0n.dds"},
	{0xF5B385E9551599A8ULL, "155968381.ogg"},
	{0xF5B38C6394CE6526ULL, "vxdr_primevaltomb_6si_0_0.vlm"},
	{0xF5B3D7EE330D5CD2ULL, "vxdr_primevaltomb_2li_0_1.vlm"},
	{0xF5B44190546F295EULL, "6121574_m.fxe"},
	{0xF5B474A05657172DULL, "hlac_w_bonepitdestroyed_a3_196614_0z.dds"},
	{0xF5B4F688FAF8CFBEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_electr.ani"},
	{0xF5B580B0EB8A34CCULL, "gen00fl_anders6128353.cl"},
	{0xF5B59AAC1FA03D10ULL, "879926245.ogg"},
	{0xF5B5A921991903CCULL, "art\\vfx\\combat\\vfx_m_arc_shield_mod_c.mmh"},
	{0xF5B615453FCCBD87ULL, "crt_rage_demon.bnk"},
	{0xF5B65CF5AF87A1F4ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_conj_c.impa.evt"},
	{0xF5B661C791B5D1B3ULL, "art\\characters\\heraldry\\hld_viscounts_keep.dds"},
	{0xF5B67941B2C5CDA6ULL, "vxdr_primevaltomb_521_0_0.vlm"},
	{0xF5B6DD29F98875BBULL, "vxht_brothel_n_4xn_0_0.vlm"},
	{0xF5B70D3100629FEAULL, "vxca_caverns2prc_aag_0_0.vlm"},
	{0xF5B7242EC10F8B2CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_action.ani"},
	{0xF5B77B5563D62817ULL, "316591938.ogg"},
	{0xF5B78F54F5C00F12ULL, "295279783.ogg"},
	{0xF5B7BCEE9676293CULL, "vxht_0_hightown_a2_c9a_0_0.vlm"},
	{0xF5B7EDDAD7034947ULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\3ww_galowbutress.phy"},
	{0xF5B8655E5D09A41BULL, "w_bonepit_87_0.msh"},
	{0xF5B86A4873266BABULL, "llac_dr_slavers_3215_0.dds"},
	{0xF5B8775E5D09C366ULL, "w_bonepit_87_0.mao"},
	{0xF5B8835E5D09D705ULL, "w_bonepit_87_0.mmh"},
	{0xF5B9278050FFB374ULL, "vxht_chantry_2q0_0_0.vlm"},
	{0xF5B9D2965FB35081ULL, "vxlt_hangedman_n_a_0_0.vlm"},
	{0xF5BA26D1353AF8C4ULL, "mag291_demon_marethari6029138.cl"},
	{0xF5BA3936B30E4DCDULL, "429315870.ogg"},
	{0xF5BA8628F00517BAULL, "6085922_m.fxe"},
	{0xF5BB1164289190FBULL, "dae000_gen_turn_in_56178420.cl"},
	{0xF5BB19F9CB575E38ULL, "vxw_bonepitmining_a2_m9_0_0.vlm"},
	{0xF5BB281DAE62E5B2ULL, "vxlt_0_lowtown_docks_brz_0_0.vlm"},
	{0xF5BB46FB18113086ULL, "vxw_bonepitmining_a2_mw_0_0.vlm"},
	{0xF5BBAE9F59DB9B2AULL, "602990516.ogg"},
	{0xF5BC0BA0D129237CULL, "ser121_jethann6104615.cl"},
	{0xF5BC849C16A45520ULL, "vxht_brothel_n_6it_0_0.vlm"},
	{0xF5BCEA224A04CE3AULL, "w_encounter3_74_0.msh"},
	{0xF5BCF8224A04E673ULL, "w_encounter3_74_0.mao"},
	{0xF5BCFC224A04EC80ULL, "w_encounter3_74_0.mmh"},
	{0xF5BD2523CC19EECFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\set_kirkwallgates_n.dds"},
	{0xF5BD334B99AAB320ULL, "291270896.ogg"},
	{0xF5BDFB0C38C5AF16ULL, "899948641.ogg"},
	{0xF5BE0D64FD8BF201ULL, "hlac_lt_undercityserialkiller_-1_32x.dds"},
	{0xF5BE312EC115C06EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_action.gad"},
	{0xF5BE32E797A9E0CBULL, "vxht_keep_n_64g_0_0.vlm"},
	{0xF5BE3C81D9B69883ULL, "character_character_p60.psh.fxo.dx11.assembly"},
	{0xF5BE4AA9AC26A8CAULL, "w_bonepit_100_0.mao"},
	{0xF5BE4EA9AC26AF19ULL, "w_bonepit_100_0.mmh"},
	{0xF5BE58A9AC26C007ULL, "w_bonepit_100_0.msh"},
	{0xF5BEA7DAAF944F3FULL, "6124755_f.fxe"},
	{0xF5BEB04D6E0ACD9CULL, "471578691.ogg"},
	{0xF5BEC84B44D8066AULL, "vxlt_0_lowtown_a2_n_c_hjl_0_0.vlm"},
	{0xF5BF4EA79E7AB141ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic8x.phy"},
	{0xF5BF564C5BD6E5FCULL, "gen00fl_aveline6139936.cl"},
	{0xF5C02092712D6BB4ULL, "zz_dae_debug6047633.cl"},
	{0xF5C041868B8387F5ULL, "seb221_ruxton_harimann6177221.cl"},
	{0xF5C0AA8669585CC4ULL, "vxw_woundedcoast_ef_0_0.vlm"},
	{0xF5C0DA3F50F43E05ULL, "vxht_0_hightown_a3_c9a_0_0.vlm"},
	{0xF5C0DEEEEE113F0DULL, "663751856.ogg"},
	{0xF5C1614AFDB268FAULL, "vxht_chantry_n_32y_0_0.vlm"},
	{0xF5C19F58C4E127BBULL, "vxlt_0_lowtown_a2_h2d_0_0.vlm"},
	{0xF5C2FF3CDF1FB340ULL, "vrc321_spooky_scenes6216684.cl"},
	{0xF5C454F5AF93D12DULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_conj_c.impa.ani"},
	{0xF5C45D425D3B17A5ULL, "706147248.ogg"},
	{0xF5C4D0B82A102125ULL, "vxlt_backalley_n_1kl_0_0.vlm"},
	{0xF5C4D19C8D4D401FULL, "vxdr_primevaltomb_6x1_0_0.vlm"},
	{0xF5C5238A2879416FULL, "80689390.ogg"},
	{0xF5C52C2EC11BD64DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_back_action.evt"},
	{0xF5C542B9D53B189EULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\0wo_basic.phy"},
	{0xF5C56CC9AFBB6491ULL, "106219653.ogg"},
	{0xF5C5A5F0E9EA6F64ULL, "follower_banter6161124.cl"},
	{0xF5C5D8E1D17A9FBAULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaopen.mmh"},
	{0xF5C5DCE1D17AA6F1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaopen.mao"},
	{0xF5C5E794DB902830ULL, "mcr310_argument6072722.cl"},
	{0xF5C5EAE1D17ABE00ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaopen.msh"},
	{0xF5C6C69B11FB80F1ULL, "hlac_ca_cavern1_haunted_24426_1x.dds"},
	{0xF5C7162D8C7E8419ULL, "vxw_bonepit_1jd_0_0.vlm"},
	{0xF5C721A6525BA90DULL, "vxlt_0_lowtown_a2_n_db5_0_0.vlm"},
	{0xF5C74C4A4B4D41C6ULL, "qun221_viscount6095591.cl"},
	{0xF5C7A158A81EABEFULL, "6083598_m.fxe"},
	{0xF5C7DD6F460FEA90ULL, "mag121_eau_banter6107381.cl"},
	{0xF5C7F2705C3C0C6BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3pp_low.phy"},
	{0xF5C822F23EB1A1AFULL, "238195795.ogg"},
	{0xF5C833F23302A0A3ULL, "80178180.ogg"},
	{0xF5C83451871EBF11ULL, "hlac_lt_gallowstemplar_2388_6x.dds"},
	{0xF5C8CCEBFC639880ULL, "dae000_gen_turn_in_46198684.cl"},
	{0xF5C9083D90FFF9F0ULL, "hlac_w_encounter2_131070_0x.dds"},
	{0xF5C90F88FB09A338ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\r_electr.gad"},
	{0xF5C91DD367EB73D7ULL, "274588331.ogg"},
	{0xF5C93F51637333E2ULL, "art\\vfx\\combat\\vfx_r_asn_regen_imp_c.impa.gad"},
	{0xF5C98BD6BD4D76B8ULL, "vxw_bonepitmining_a2_h6_0_0.vlm"},
	{0xF5C9B8D80A074906ULL, "vxw_bonepitmining_a2_hx_0_0.vlm"},
	{0xF5C9D62D576DB1A9ULL, "zz_dog_debug6127860.cl"},
	{0xF5CA3E108441EAFCULL, "6123233_m.fxe"},
	{0xF5CA496AFF3FA944ULL, "6096869_m.fxe"},
	{0xF5CA992144AFFE55ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1.phy"},
	{0xF5CB45C70F6C8394ULL, "vxlt_gallowstemplar_4u9_0_0.vlm"},
	{0xF5CBC8E7E88852F2ULL, "283763922.ogg"},
	{0xF5CC8EEB38A675BCULL, "vxw_woundedcoast_z7_0_0.vlm"},
	{0xF5CCB5F26621B176ULL, "vrc221_bartrand6065310.cl"},
	{0xF5CD4A823553510DULL, "410736587.ogg"},
	{0xF5CD8CF4136D032CULL, "mag170_thrask6016927.cl"},
	{0xF5CD9F877DF26473ULL, "vxht_mansion2_1gr_0_0.vlm"},
	{0xF5CDA4A98E11E6DDULL, "pro000_cut2d_voyage6004229.cl"},
	{0xF5CDA800C392FDC1ULL, "vxlt_gallowstemplar_n_3fu_0_0.vlm"},
	{0xF5CEA9AB77DF1031ULL, "6048294_m.fxe"},
	{0xF5CEE049C90219DDULL, "w_encounter3_128_0.phy"},
	{0xF5CF1C635B4E7544ULL, "vxw_woundedcoast_bg_0_0.vlm"},
	{0xF5CF25F7AD25F373ULL, "6162439_m.fxe"},
	{0xF5CF61E5AE4ABAD9ULL, "508048369.ogg"},
	{0xF5CF62516377DAD9ULL, "art\\vfx\\combat\\vfx_r_asn_regen_imp_c.impa.evt"},
	{0xF5CFAC170D487B3CULL, "vxw_bonepit_1v9_0_0.vlm"},
	{0xF5CFF4ACD889D64AULL, "follower_banter6154012.cl"},
	{0xF5D007D0CC630287ULL, "w_bonepitmining_a2_44_0.rml"},
	{0xF5D07EB248159659ULL, "hlac_lt_0_lowtown_docks_n_-131075_2x.dds"},
	{0xF5D0FAADAE829EFAULL, "133655548.ogg"},
	{0xF5D1092221E55494ULL, "vxht_0_hightown_ca5_0_0.vlm"},
	{0xF5D109CF3F5D26B3ULL, "hlac_lt_0_lowtown_docks_n_-65538_0z.dds"},
	{0xF5D1FEA175DDFE49ULL, "gen00fl_fenris6112115.cl"},
	{0xF5D2C29C95EF3001ULL, "gen_im_trs_gem_ruby_03.uti"},
	{0xF5D342951C0639A5ULL, "vxlt_backalley_n_1vk_0_0.vlm"},
	{0xF5D3640F47C08E21ULL, "56160304.ogg"},
	{0xF5D413AEDB6B49EEULL, "hlac_ht_keep_n_3740_0x.dds"},
	{0xF5D500D4ECF74503ULL, "249845371.ogg"},
	{0xF5D52FFC9005D531ULL, "fxe_water_form_b.mao"},
	{0xF5D6218996C4893DULL, "6119505_m.fxe"},
	{0xF5D67C5A2CFFD50BULL, "l_mag331ar_evelinas_hideout.wbd"},
	{0xF5D728AC71D28F5EULL, "vxlt_gallowstemplar_5qr_0_0.vlm"},
	{0xF5D7C004DCA37E9EULL, "vxw_woundedcoast_102_0_0.vlm"},
	{0xF5D7C673CFC9F93DULL, "vxw_bonepitmining_a2_1lp_0_0.vlm"},
	{0xF5D829F57E8B4AA7ULL, "6162336_m.fxe"},
	{0xF5D8C99B4AC4616CULL, "6060015_m.fxe"},
	{0xF5D8D62D321299F3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_flagviscount_d.dds"},
	{0xF5D99E032BE8C448ULL, "hf_thug02.mor"},
	{0xF5D9C9DBA0B9B664ULL, "vxca_cave_1a7_0_0.vlm"},
	{0xF5D9D1F66CCE9CB3ULL, "106747585.ogg"},
	{0xF5DA4B52C248762AULL, "980191073.ogg"},
	{0xF5DA96759FD2762FULL, "hlac_lt_0_lowtown_docks_a3_n_-131074_8x.dds"},
	{0xF5DA9C05BD7CEE1FULL, "69679523.ogg"},
	{0xF5DADD328B9BAEC5ULL, "246970702.ogg"},
	{0xF5DB4DBB6D1E42F0ULL, "art\\vfx\\environment\\vfx_lt_ash_sparks.mmh"},
	{0xF5DC456C80E6F968ULL, "llac_ht_0_hightown_a3_-131074_1.dds"},
	{0xF5DD06C5241349BBULL, "6166813_m.fxe"},
	{0xF5DD1B8869D528B1ULL, "6043394_m.fxe"},
	{0xF5DD6A51638425C8ULL, "art\\vfx\\combat\\vfx_r_asn_regen_imp_c.impa.ani"},
	{0xF5DDAECC76000690ULL, "7007309.ogg"},
	{0xF5DDBDF934E06F05ULL, "vxht_0_hightown_a3_c7o_0_0.vlm"},
	{0xF5DDC4D7D57BDC14ULL, "vxht_0_hightown_a3_aq2_0_0.vlm"},
	{0xF5DDEC7AC8375BABULL, "lgt110tr_coa_meet_warn.ncs"},
	{0xF5DE2022312DE0DBULL, "fex_bd_d_chm.msh"},
	{0xF5DE3E22312E13C5ULL, "fex_bd_d_chm.mmh"},
	{0xF5DE4214D650E707ULL, "static_static_p66.psh.fxo.dx11"},
	{0xF5DE777B8B9EEE3EULL, "vxlt_hangedman_26y_0_0.vlm"},
	{0xF5DE7EC184B33C12ULL, "858257515.ogg"},
	{0xF5DEBA5D4B8EB89BULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyo.phy"},
	{0xF5DF137241929DB0ULL, "vxlt_undercityencounter_1b3_0_0.vlm"},
	{0xF5DF19A0D95AE4CAULL, "art\\vfx\\combat\\vfx_d_qun_attk_v1_c.impa.gad"},
	{0xF5DFC1A38E533231ULL, "6143490_m.fxe"},
	{0xF5DFD549C9107A49ULL, "w_encounter3_128_0.mao"},
	{0xF5DFE149C9108EE2ULL, "w_encounter3_128_0.mmh"},
	{0xF5DFE349C9109198ULL, "w_encounter3_128_0.msh"},
	{0xF5E02B36803E092DULL, "6143480_f.fxe"},
	{0xF5E054A7FA1E3CE4ULL, "vxlt_0_lowtown_eh8_0_0.vlm"},
	{0xF5E05A21CF39CFB5ULL, "vxht_0_hightown_n_c4h_0_0.vlm"},
	{0xF5E0665366E22371ULL, "vxw_blightlands_gk_0_0.vlm"},
	{0xF5E0D41C429B0496ULL, "461156492.ogg"},
	{0xF5E126945A606360ULL, "212332726.ogg"},
	{0xF5E20B453FF1E7BDULL, "crt_rage_demon.wbd"},
	{0xF5E24DD057719CB3ULL, "143247936.ogg"},
	{0xF5E26FCBFF4412C9ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_nud_0d.dds"},
	{0xF5E2D3D11A446D8FULL, "art\\vfx\\combat\\vfx_w_wns_front_close1_c.mmh"},
	{0xF5E2DFA39C22555AULL, "84514608.ogg"},
	{0xF5E2E8E9E7FC1777ULL, "hlac_dr_primevalentrance_1724_3z.dds"},
	{0xF5E2F7EFF584D947ULL, "6167304_m.fxe"},
	{0xF5E457E8D09E8293ULL, "6150280_f.fxe"},
	{0xF5E4F0F7C64A73BDULL, "rdr111dg_martin6014873.cl"},
	{0xF5E506ECEB2C9630ULL, "175147054.ogg"},
	{0xF5E522E993FBFD4EULL, "vxdr_0_deeproads_bxb_0_0.vlm"},
	{0xF5E53834611444F9ULL, "hlac_lt_undercitydungeon_-1_35x.dds"},
	{0xF5E582B69B05B068ULL, "ave310_party_comment6134847.cl"},
	{0xF5E58A6696B1079DULL, "follower_banter6169350.cl"},
	{0xF5E5D74681A0F933ULL, "75484687.ogg"},
	{0xF5E5FCA0D960D261ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v1_c.impa.evt"},
	{0xF5E604141C84374CULL, "hlac_lt_0_lowtown_docks_a3_n_65534_1z.dds"},
	{0xF5E60C747039A137ULL, "vxlt_undercityserialkiller_1s3_0_0.vlm"},
	{0xF5E67DB484452F32ULL, "vxlt_0_lowtown_docks_bce_0_0.vlm"},
	{0xF5E6DEEE3F3B4D51ULL, "dre000_sibling_has_blight6036479.cl"},
	{0xF5E707D0CC75E1A4ULL, "w_bonepitmining_a2_44_0.mmh"},
	{0xF5E70A7E894F4072ULL, "llac_lt_gallowscourtyard_n_-131072_0.dds"},
	{0xF5E70BD0CC75E8F7ULL, "w_bonepitmining_a2_44_0.mao"},
	{0xF5E71DD0CC760746ULL, "w_bonepitmining_a2_44_0.msh"},
	{0xF5E7E343A3CCA3EAULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackright.gad"},
	{0xF5E8C9BE3B0BD4D6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_diagonal.mao"},
	{0xF5E8D5BE3B0BE975ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_diagonal.mmh"},
	{0xF5E8D7BE3B0BECEBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_diagonal.msh"},
	{0xF5E95F0DEBAC160CULL, "237349121.ogg"},
	{0xF5EA7292F3A04164ULL, "hlac_lt_0_lowtown_a2_n_c_-1_2z.dds"},
	{0xF5EA9FC1ECFE1C47ULL, "vrc221_varric6043969.cl"},
	{0xF5EAB89AEE75E456ULL, "vxw_encounter2_mq_0_0.vlm"},
	{0xF5EB8E13DD65E18EULL, "166595327.ogg"},
	{0xF5EBC30CA08D16F9ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_medg_0s.dds"},
	{0xF5EBD72D229C9C29ULL, "coins_d.dds"},
	{0xF5EC001D3F3AA644ULL, "vxw_woundedcoast_te_0_0.vlm"},
	{0xF5EC51786D38A1E4ULL, "68084552.ogg"},
	{0xF5EC56A304AB090FULL, "dae000_hate_enemies6167623.cl"},
	{0xF5EC58EF0226F845ULL, "vxht_chantry_n_32r_0_0.vlm"},
	{0xF5EC933A356D1F3DULL, "follower_banter6143347.cl"},
	{0xF5ECD119869EF217ULL, "llac_lt_0_lowtown_docks_n_-196608_1.dds"},
	{0xF5ED112385AF0AF7ULL, "mag291_vincento6027709.cl"},
	{0xF5ED7A17E770A2F1ULL, "sib131_letter_react6127322.cl"},
	{0xF5ED8EB1C1E675FAULL, "vxca_cavern1_haunted_j2f_0_0.vlm"},
	{0xF5EDC1C390D000CAULL, "402210289.ogg"},
	{0xF5EE84B3C2CA2DAFULL, "598226555.ogg"},
	{0xF5EE9F7726602B15ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basicnb.msh"},
	{0xF5EEA17726602E8BULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basicnb.mmh"},
	{0xF5EEA4A41A8CB550ULL, "mag101tr_aom_decimus.utt"},
	{0xF5EEAD7726604328ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\1ww_basicnb.mao"},
	{0xF5EEC643A3D29081ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackright.evt"},
	{0xF5EECBFEC12FE835ULL, "vxht_0_hightown_n_c1w_0_0.vlm"},
	{0xF5EED83058D83BF1ULL, "vxw_blightlands_hj_0_0.vlm"},
	{0xF5EEE91172BBB1BFULL, "mcr310_argument6072279.cl"},
	{0xF5EF02DA8C6EFF49ULL, "213224825.ogg"},
	{0xF5EF317A2272EF66ULL, "mrl150_merrill6150731.cl"},
	{0xF5EF7AF8558FF85FULL, "663148924.ogg"},
	{0xF5EF8E8D190B05A3ULL, "hlac_w_sundermount_327680_0x.dds"},
	{0xF5F16B95FEA3005BULL, "hlac_lt_undercitydungeon_-1_16y.dds"},
	{0xF5F16CD519AE5841ULL, "726537674.ogg"},
	{0xF5F21A9CA018EAC7ULL, "vxht_0_hightown_n_a6v_0_0.vlm"},
	{0xF5F26BC7F89B4E35ULL, "621142782.ogg"},
	{0xF5F28E260CC56CE5ULL, "vxw_woundedcoast_19r_0_0.vlm"},
	{0xF5F29B43349806CAULL, "6171926_m.fxe"},
	{0xF5F2D09C9574EB74ULL, "467518511.ogg"},
	{0xF5F343719866BE22ULL, "6129004_m.fxe"},
	{0xF5F3F4A0D96D01B0ULL, "art\\vfx\\combat\\vfx_d_qun_attk_v1_c.impa.ani"},
	{0xF5F4DFE32E0CF0C1ULL, "llac_lt_gallowstemplar_n_2394_4.dds"},
	{0xF5F4EF91763B47B2ULL, "vxlt_0_lowtown_docks_bff_0_0.vlm"},
	{0xF5F57AE50E24809CULL, "vxht_0_hightown_a2_aki_0_0.vlm"},
	{0xF5F5D7665AAB1F84ULL, "hlac_lt_0_lowtown_docks_n_131070_0y.dds"},
	{0xF5F5E627BA361D7CULL, "vxlt_0_lowtown_a3_n_c_hqj_0_0.vlm"},
	{0xF5F5FF6CCA1041DBULL, "vxw_encounter2_qu_0_1.vlm"},
	{0xF5F667B6F4C8028BULL, "6092365_m.fxe"},
	{0xF5F6DDEC077CC67EULL, "682965901.ogg"},
	{0xF5F702C893565192ULL, "vxlt_0_lowtown_a2_gz2_0_0.vlm"},
	{0xF5F7D0EC477D6F4CULL, "qun221_viscount_st.stg"},
	{0xF5F87D6B3380231AULL, "art\\characters\\weapons\\unique\\meredith\\textures\\w_gsw_014a_0n.dds"},
	{0xF5F8DDFD53DF8AB0ULL, "vxw_sundermount_33z_0_1.vlm"},
	{0xF5F92A77E06BFCD6ULL, "vxw_encounter2_hr_0_0.vlm"},
	{0xF5F9320B28A3FBA2ULL, "vxht_0_hightown_9xb_0_0.vlm"},
	{0xF5F93F474418430CULL, "381391768.ogg"},
	{0xF5F9712523E7D8CBULL, "conversation.gfx"},
	{0xF5F990D8507218D5ULL, "432958794.ogg"},
	{0xF5FA3C5EAC6EDCB5ULL, "868690440.ogg"},
	{0xF5FA5BDA3BCA623CULL, "gen00fl_sebastian6180383.cl"},
	{0xF5FA71FA3130BEC4ULL, "vxw_woundedcoast_qj_0_0.vlm"},
	{0xF5FB4153E59B7B20ULL, "hlac_w_blightlands_-131072_0x.dds"},
	{0xF5FB5E5A2D1EBED9ULL, "l_mag331ar_evelinas_hideout.bnk"},
	{0xF5FBA03E803F1899ULL, "llac_ht_keep_3732_12.dds"},
	{0xF5FBF17AC8506192ULL, "lgt110tr_coa_meet_warn.utt"},
	{0xF5FC18506C66E300ULL, "vxlt_0_lowtown_a2_n_c_gmp_0_0.vlm"},
	{0xF5FC75FA125233BCULL, "6121895_m.fxe"},
	{0xF5FCBE43A3DEC0D0ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\walkbackright.ani"},
	{0xF5FCF4E59B8BCAA4ULL, "scaleform_sf_solid_p0.psh.fxo"},
	{0xF5FD4F3F5C6EE2B1ULL, "lt_gallowscourtyard_n_18_0.rml"},
	{0xF5FD57C31A617457ULL, "356229635.ogg"},
	{0xF5FE650168D10307ULL, "6113277_m.fxe"},
	{0xF5FE88ED02ACD066ULL, "468230607.ogg"},
	{0xF5FE8BCD1315F13BULL, "art\\characters\\creatures\\arh_abm_rev\\c_arcanehorror.phy"},
	{0xF5FF1D82B4593324ULL, "fen200ip_blood_sacrafice.ncs"},
	{0xF5FF48A90F298D46ULL, "6217123_m.fxe"},
	{0xF5FF52C3901F5344ULL, "770402978.ogg"},
	{0xF5FFBB88D881921FULL, "6105429_m.fxe"},
	{0xF5FFDCEC8F4FEDCEULL, "6105046_m.fxe"},
	{0xF600A1E53145F841ULL, "mag160_varian_ilithis6018904.cl"},
	{0xF600E4B8D0B99034ULL, "leap_cl.mao"},
	{0xF6012C4DC7A7871EULL, "vxca_caverns1_fmb_0_0.vlm"},
	{0xF601CC9CD9CF9DDCULL, "dre000_varric6208016.cl"},
	{0xF601FE4214D2035AULL, "dae000_gamlen6145122.cl"},
	{0xF6024097B5FC9422ULL, "645459780.ogg"},
	{0xF60265E8DB2ACE74ULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_act_c.mmh"},
	{0xF602F668310E5BBCULL, "5813074.ogg"},
	{0xF602FF31707B1C39ULL, "6101503_f.fxe"},
	{0xF603279A8F016625ULL, "gen00fl_aveline6139711.cl"},
	{0xF6032D74D9A7B1FFULL, "vxlt_undercityserialkiller_15m_0_0.vlm"},
	{0xF603733F92EEEDCFULL, "llac_lt_0_lowtown_-3_2.dds"},
	{0xF60391072DAB1215ULL, "6120831_m.fxe"},
	{0xF6043A73C7091251ULL, "hlac_lt_0_lowtown_-2_2y.dds"},
	{0xF60494885939EBC6ULL, "713801059.ogg"},
	{0xF604B2E86A107EB6ULL, "162113354.ogg"},
	{0xF60556FC883460C6ULL, "6142928_m.fxe"},
	{0xF605CF6BAB4403AAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v3.gad"},
	{0xF606002BBAAED907ULL, "6033705_f.fxe"},
	{0xF606194935589251ULL, "hlac_ca_caverns2_5802_48z.dds"},
	{0xF60658D60E40D28DULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_ars_0d.dds"},
	{0xF606E37933BC76E9ULL, "zz_ave_debug6056759.cl"},
	{0xF606FC62CFB41AC2ULL, "pro000_meeran6050070.cl"},
	{0xF607019BFEA680EAULL, "vxlt_0_lowtown_a2_n_c_hqi_0_0.vlm"},
	{0xF6070FABB28F261DULL, "752348709.ogg"},
	{0xF6075E0263CB26D9ULL, "6202290_m.fxe"},
	{0xF6076F748B6B5FC1ULL, "vxlt_gallowstemplar_n_3ri_0_0.vlm"},
	{0xF60782F8A52B37A0ULL, "vxw_encounter2_6j_0_0.vlm"},
	{0xF607CD46E5C2E918ULL, "6042287_f.fxe"},
	{0xF60895CB48DBA8BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ready.evt"},
	{0xF6090DBAABE774D2ULL, "6103691_m.fxe"},
	{0xF6091B99EC72B62DULL, "vxlt_0_lowtown_a3_n_c_ffp_0_0.vlm"},
	{0xF6094A95A2AF6AF8ULL, "196466075.ogg"},
	{0xF609C6E5ECAB098DULL, "vxlt_0_lowtown_docks_n_ctq_0_0.vlm"},
	{0xF609CC02BA2C0159ULL, "snow_01_l3.mao"},
	{0xF60A45FC92AC5468ULL, "vxlt_warehouse_n_8z_0_0.vlm"},
	{0xF60AD4D11A65C1CCULL, "art\\vfx\\combat\\vfx_w_wns_front_close1_c.anb"},
	{0xF60AF9315794D3F7ULL, "6131134_m.fxe"},
	{0xF60B3E82D966B50CULL, "mrl250cr_spider_wasp.utc"},
	{0xF60BA15D4BB57C3AULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyo.msh"},
	{0xF60BAF5D4BB59473ULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyo.mao"},
	{0xF60BB2044E9A990BULL, "zz_one_debug6073989.cl"},
	{0xF60BB35D4BB59A80ULL, "art\\characters\\playercharacter\\elfmale\\bdy_hvyo.mmh"},
	{0xF60BBBEA3CC46CF1ULL, "vxw_blightlands_vh_0_0.vlm"},
	{0xF60C74CDB4C81830ULL, "6105695_m.fxe"},
	{0xF60CB26BAB49F141ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v3.evt"},
	{0xF60E6761A0554F63ULL, "ave331_jeven6152177.cl"},
	{0xF60E8ECB48E008FCULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ready.gad"},
	{0xF60EC4F0D059DC97ULL, "hlac_ca_caverns1_mines_24426_13y.dds"},
	{0xF60F1599EC01E24EULL, "6062833_m.fxe"},
	{0xF60F3FA1F245BF63ULL, "ser270_emeric6149327.cl"},
	{0xF60F4459F7C76436ULL, "l_rdr361ar_cave.wbd"},
	{0xF60F574EA8558F5CULL, "756859897.ogg"},
	{0xF60F885E2FF3117EULL, "728929782.ogg"},
	{0xF60F8FB2BC76EAAFULL, "6034152_m.fxe"},
	{0xF60FBF934A117DACULL, "hlac_w_bonepitmining_a2_0_2z.dds"},
	{0xF60FC464EB53526BULL, "685808844.ogg"},
	{0xF61038298D477BB2ULL, "vxw_bonepitdestroyed_a3_1zs_0_0.vlm"},
	{0xF61040B69C71F35CULL, "710095954.ogg"},
	{0xF6108DEEFAA7FB12ULL, "character_character_p159.psh.fxo.dx11"},
	{0xF611007CF67E6486ULL, "hlac_ca_caverns2_5802_7x.dds"},
	{0xF611922813FB4821ULL, "hlac_w_sundermount_393214_0z.dds"},
	{0xF611A172F8255CB2ULL, "vxca_cavern1_haunted_i6u_0_0.vlm"},
	{0xF611BA2144EC388AULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_ball1.mmh"},
	{0xF6122A409C728E2EULL, "and111ar_andersshop.ncs"},
	{0xF6122F679D1C5682ULL, "654399501.ogg"},
	{0xF61393A41AABB40DULL, "mag101tr_aom_decimus.ncs"},
	{0xF613949E852AFC6AULL, "6135625_m.fxe"},
	{0xF613D34E21A0FA42ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave6statue.msh"},
	{0xF613D54E21A0FDF8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave6statue.mmh"},
	{0xF613E14E21A1119BULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slave6statue.mao"},
	{0xF613FDCC8808F07BULL, "zz_dre_debug6205406.cl"},
	{0xF61401C0A37890E0ULL, "6084219_m.fxe"},
	{0xF614359B0E1886C8ULL, "97639664.ogg"},
	{0xF61497E05E1F9726ULL, "vxdr_0_deeproads_9iw_0_1.vlm"},
	{0xF614A7DF1DFFE363ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.down.ani"},
	{0xF614D765E53AC1B3ULL, "334953323.ogg"},
	{0xF6154C27068904F1ULL, "dre000_stroud6036746.cl"},
	{0xF6158DCB48E625FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basedualwield\\combat\\ready.ani"},
	{0xF615D022F503C928ULL, "6034325_f.fxe"},
	{0xF615F4D597215020ULL, "vxw_encounter2_1k_0_0.vlm"},
	{0xF61621D5F6D24AA3ULL, "983213560.ogg"},
	{0xF61621F304DEF0DDULL, "hlac_lt_0_lowtown_docks_n_131071_3z.dds"},
	{0xF6162C6954FC8E2BULL, "dae000_sandal6148202.cl"},
	{0xF6165FE52957606AULL, "hlac_lt_0_lowtown_docks_a3_n_131071_5z.dds"},
	{0xF6179D5628F9515AULL, "vxca_caverns1_mines_k2r_0_1.vlm"},
	{0xF617E567C716F5BCULL, "vxw_bonepit_1o2_0_0.vlm"},
	{0xF61838C2DEA1220DULL, "vxlt_0_lowtown_docks_n_cop_0_0.vlm"},
	{0xF6185AD7264423C0ULL, "gen00fl_aveline6137707.cl"},
	{0xF6188655C65B0545ULL, "vxca_cavern1_haunted_j2m_0_0.vlm"},
	{0xF6189A336BB53A51ULL, "zz_one_debug6058980.cl"},
	{0xF6189DD46EE7A19AULL, "vxca_caverns1_e4z_0_0.vlm"},
	{0xF618EA2F8EA1671AULL, "6173164_m.fxe"},
	{0xF6192064FCA18559ULL, "hlac_ht_keep_3735_0x.dds"},
	{0xF61922409C789F07ULL, "and111ar_andersshop.lst"},
	{0xF619914A755DAD4EULL, "gen00fl_sebastian6188107.cl"},
	{0xF619AA6BAB546D90ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\bow\\archer_range_attack_v3.ani"},
	{0xF61A08CFF428D9C6ULL, "dae000_hate_enemies.cnv"},
	{0xF61A7E609C08FFEBULL, "store_lgt_epic3_3.utm"},
	{0xF61A83DC7782F2AEULL, "823936016.ogg"},
	{0xF61B01731C2D9DF4ULL, "gen00fl_aveline6137475.cl"},
	{0xF61B482DFA0CBFE6ULL, "225842867.ogg"},
	{0xF61B4B329EB4974CULL, "vxlt_foundry_n_x0_0_0.vlm"},
	{0xF61B5BAA91C48034ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.mmh"},
	{0xF61B8B2760F6E56AULL, "art\\characters\\playercharacter\\shared\\textures\\cn_stn_0d.dds"},
	{0xF61BD94BCA6BFBD1ULL, "one101st_lia.stg"},
	{0xF61CA985945A70A9ULL, "949973552.ogg"},
	{0xF61D1BB46F731981ULL, "lgt160_wk_soph_hunter6119755.cl"},
	{0xF61DB40E76DE056EULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipe.evt"},
	{0xF61E0ACFF42CC586ULL, "dae000_hate_enemies.bnk"},
	{0xF61E289467A5EC27ULL, "vxlt_hangedman_n_173_0_0.vlm"},
	{0xF61E3C1A883FF2C0ULL, "6040577_f.fxe"},
	{0xF61F89DEAF55EB44ULL, "854608588.ogg"},
	{0xF6206FEBE22FD468ULL, "6137508_m.fxe"},
	{0xF620E29A1BC4E3CDULL, "vxlt_hangedman_n_1vo_0_1.vlm"},
	{0xF62153A79ECD847DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic8x.mao"},
	{0xF62157A79ECD8AB6ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic8x.mmh"},
	{0xF62161A79ECD9BB4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic8x.msh"},
	{0xF621635A35B1E04FULL, "6104291_m.fxe"},
	{0xF621738D1929496CULL, "vxlt_0_lowtown_docks_a3_n_cy9_0_1.vlm"},
	{0xF6217A0E1A859356ULL, "6055425_f.fxe"},
	{0xF621E1CBE5694135ULL, "330585803.ogg"},
	{0xF6233A6A8C690C91ULL, "em_amb_darktown.mor"},
	{0xF6233DCE4CEE6D09ULL, "191893640.ogg"},
	{0xF623AB0E76E26119ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipe.gad"},
	{0xF624BA0EB35B9D20ULL, "vxw_blightlands_fe_0_0.vlm"},
	{0xF624D24725D27B78ULL, "6106320_m.fxe"},
	{0xF624F0E961F3F205ULL, "s_aoe_cr2.mao"},
	{0xF624F88F8836C613ULL, "41330240.ogg"},
	{0xF6251C471A1AA9FBULL, "488355336.ogg"},
	{0xF62527FAC147F673ULL, "310847796.ogg"},
	{0xF625C41A3D828B71ULL, "w_blightlands_51_0_lc.mmh"},
	{0xF62605064377BCBAULL, "vxw_encounter2_9y_0_0.vlm"},
	{0xF6266B2D8E39D491ULL, "w_bonepitdestroyed_a3_197_0.mmh"},
	{0xF6267148374ACA11ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\combat_face_additive.anb"},
	{0xF626772D8E39E932ULL, "w_bonepitdestroyed_a3_197_0.mao"},
	{0xF626852D8E3A014FULL, "w_bonepitdestroyed_a3_197_0.msh"},
	{0xF62692D61747E6CAULL, "102212080.ogg"},
	{0xF626A92CBE91189AULL, "lgt000_chanter6147197.cl"},
	{0xF626B67A88EA7DF8ULL, "llac_lt_0_lowtown_a2_n_-196609_0.dds"},
	{0xF626D9627E014DF5ULL, "6128858_f.fxe"},
	{0xF626F85B5466FFA4ULL, "llac_lt_0_lowtown_n_-65537_1.dds"},
	{0xF6273AF3DE110C04ULL, "6163462_m.fxe"},
	{0xF627B8FF1E25B1F5ULL, "mag311_huon6042572.cl"},
	{0xF627CDEF168985D4ULL, "6126425_m.fxe"},
	{0xF62821F71AE467F6ULL, "834537231.ogg"},
	{0xF62865F017C4DF34ULL, "6099200_m.fxe"},
	{0xF62884C973E8A3ABULL, "mag121cr_idunna.utc"},
	{0xF628BEDF1E10B34DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.down.gad"},
	{0xF628CAA1773F5D71ULL, "442367881.ogg"},
	{0xF62968558FB6F605ULL, "dr_dungeon.arl"},
	{0xF62A167D82C38075ULL, "llac_lt_0_lowtown_a2_n_c_-2_4.dds"},
	{0xF62A23409C87143CULL, "and111ar_andersshop.are"},
	{0xF62A68E8DB4C252BULL, "art\\vfx\\combat\\vfx_r_gn4_smo_g_act_c.anb"},
	{0xF62A72CD133B01DAULL, "art\\characters\\creatures\\arh_abm_rev\\c_arcanehorror.msh"},
	{0xF62A84CD133B2020ULL, "art\\characters\\creatures\\arh_abm_rev\\c_arcanehorror.mmh"},
	{0xF62A8C9EC7E794D5ULL, "art\\levels\\ztools\\default\\generic\\1wi_basic.phy"},
	{0xF62AA598311A4C2AULL, "and230tr_dissent_fight.utt"},
	{0xF62ABC0E76E89D67ULL, "art\\characters\\creatures\\highdragon\\animation\\talent\\swipe.ani"},
	{0xF62B4BD62626ED3CULL, "vxht_0_hightown_a2_cay_0_0.vlm"},
	{0xF62BBF29EF9D3B8CULL, "hlac_w_encounter2_camp_196612_0y.dds"},
	{0xF62C1176E1F78761ULL, "vxca_cavern1_haunted_j47_0_0.vlm"},
	{0xF62C6B24DC3EA181ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_mera_0d.dds"},
	{0xF62C84271E37A65AULL, "one280_emrys6104277.cl"},
	{0xF62C94B41EB96404ULL, "6050430_m.fxe"},
	{0xF62D35BA5F04FDA8ULL, "vxw_bonepitdestroyed_a3_1wo_0_1.vlm"},
	{0xF62E2A295CDA8B46ULL, "tbe100im_invisible_gloves.uti"},
	{0xF62E5EE722A47630ULL, "458175585.ogg"},
	{0xF62E6F20B7E1E073ULL, "vxht_bartrand_n_5p7_0_0.vlm"},
	{0xF62E959067C51CF0ULL, "120226193.ogg"},
	{0xF62F7D7358B9055DULL, "552286346.ogg"},
	{0xF62FC9FEE6EE5276ULL, "lgt000cr_epic_merchant_1.utc"},
	{0xF62FF88A93D63AF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_enter.gad"},
	{0xF62FFC97CF4332E2ULL, "lgt110_gxa_vgk_silver_order.cnv"},
	{0xF6306C8498A32D2BULL, "148352619.ogg"},
	{0xF630C79B8E79D3F6ULL, "131854583.ogg"},
	{0xF630E1A8E9AA86FAULL, "zz_isabellareveal.cut"},
	{0xF630F05C9BBE27ECULL, "6128746_m.fxe"},
	{0xF630F4464C745088ULL, "gen00fl_dog6109234.cl"},
	{0xF631DD4A58E3568AULL, "610722774.ogg"},
	{0xF6322D1BCBBB2EA9ULL, "hlac_lt_gallowscourtyard_-1_0y.dds"},
	{0xF6332BEBA551B5A0ULL, "vxw_blightlands_kb_0_0.vlm"},
	{0xF633350CC8D004A2ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablewht.msh"},
	{0xF633370CC8D00858ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablewht.mmh"},
	{0xF6333833BD0B55C2ULL, "gen00fl_anders6111265.cl"},
	{0xF633430CC8D01CFBULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablewht.mao"},
	{0xF6334EE550982D46ULL, "darktown_male_fereldan_sta6175136.cl"},
	{0xF633D713E4CCF0BCULL, "6126992_f.fxe"},
	{0xF633DE97CF46E81AULL, "lgt110_gxa_vgk_silver_order.bnk"},
	{0xF633E5C1656CBC97ULL, "173493303.ogg"},
	{0xF633F2191570CEEEULL, "lt_0_lowtown_docks_a3_n_109_0.mmh"},
	{0xF633FE191570E285ULL, "lt_0_lowtown_docks_a3_n_109_0.mao"},
	{0xF6340C191570FADCULL, "lt_0_lowtown_docks_a3_n_109_0.msh"},
	{0xF6347130C254FFADULL, "vxlt_0_lowtown_a3_n_c_fim_0_0.vlm"},
	{0xF63539D4431922F4ULL, "zz_and_debug6178576.cl"},
	{0xF6359560308F0B9BULL, "gen00fl_merrill6105658.cl"},
	{0xF635974A1B9D6B05ULL, "pro000_colonel_ewald6039937.cl"},
	{0xF6359A2CA34C3348ULL, "vxca_cavern1_haunted_j42_0_0.vlm"},
	{0xF635DB132804F143ULL, "6026199_m.fxe"},
	{0xF6361AA2F10E6B5EULL, "514395267.ogg"},
	{0xF63686EB45D347F8ULL, "58248883.ogg"},
	{0xF6370D8A93DC7C84ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_enter.evt"},
	{0xF63771D9C003E156ULL, "672697544.ogg"},
	{0xF63796F49AAAB984ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm04.msh"},
	{0xF6379DB8EDB51588ULL, "391886069.ogg"},
	{0xF637ACF49AAADF26ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm04.mmh"},
	{0xF63859FB253E2943ULL, "vxht_0_hightown_a2_cbj_0_1.vlm"},
	{0xF638B29A1794E382ULL, "vxht_chantry_36n_0_0.vlm"},
	{0xF638B2A787CE7B0CULL, "vxca_caverns2_7mj_0_0.vlm"},
	{0xF638EAC25DDF6402ULL, "gen00fl_anders6125266.cl"},
	{0xF6390DF582097C6AULL, "vxca_caverns2prc_a4x_0_0.vlm"},
	{0xF63A615434D69C13ULL, "161598960.ogg"},
	{0xF63AFF9220464EC9ULL, "698550099.ogg"},
	{0xF63B13C9DC671D8DULL, "art\\vfx\\models\\vfx_vines01.mmh"},
	{0xF63B17C9DC6724DEULL, "art\\vfx\\models\\vfx_vines01.mao"},
	{0xF63B25C9DC673C13ULL, "art\\vfx\\models\\vfx_vines01.msh"},
	{0xF63B66457F0AE7F4ULL, "zz_mag_debug6054481.cl"},
	{0xF63B6A59F7ECDFA4ULL, "l_rdr361ar_cave.bnk"},
	{0xF63B88F0FA429AC2ULL, "hlac_lt_undercitydungeon_-1_3z.dds"},
	{0xF63BB0A8313971E9ULL, "6043163_m.fxe"},
	{0xF63C0AB6708DDD05ULL, "617747344.ogg"},
	{0xF63C16E44AFA82C5ULL, "mag311_nyssa6042406.cl"},
	{0xF63C6B2E00172844ULL, "vxdr_primevalentrance_2zm_0_1.vlm"},
	{0xF63D0360F8F70195ULL, "6067749_m.fxe"},
	{0xF63D38DBBE4433BDULL, "lgt100cr_gang_1_lt.utc"},
	{0xF63D9AD4B6C83258ULL, "w_encounter3_2_0.rml"},
	{0xF63DA9641C2CBC56ULL, "176474339.ogg"},
	{0xF63DF8E8C4CEAF87ULL, "lt_0_lowtown_a2_n_c_22_0.rml"},
	{0xF63E3E0CC8DA1BE3ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_hextablewht.phy"},
	{0xF63E6E95B8246E41ULL, "w_bonepitdestroyed_a3_154_0.phy"},
	{0xF63F09DE02DA3805ULL, "gen00fl_fenris6127120.cl"},
	{0xF63F2464D41ED4F9ULL, "dre000_varric6207977.cl"},
	{0xF63F7881F8B4E09CULL, "gen00fl_aveline6132753.cl"},
	{0xF63F902C02945B16ULL, "vxlt_0_lowtown_a3_n_c_e1y_0_0.vlm"},
	{0xF63FBB67BAA70D68ULL, "vxlt_undercitydungeon_1d5_0_1.vlm"},
	{0xF6404DE6F98309A4ULL, "w_bonepitmining_a2_251_0.msh"},
	{0xF6405FE6F983280DULL, "w_bonepitmining_a2_251_0.mao"},
	{0xF64063E6F9832F46ULL, "w_bonepitmining_a2_251_0.mmh"},
	{0xF6406F297488F5A4ULL, "411682073.ogg"},
	{0xF64118363D460EC6ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_cth_comb_0s.dds"},
	{0xF641721C8AA0E546ULL, "594317131.ogg"},
	{0xF6419DC89747CE20ULL, "vxw_blightlands_lc_0_0.vlm"},
	{0xF641A3F49AB324F1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hm04.phy"},
	{0xF6420FD530C1082AULL, "ch2armtnt_s1y.psh.fxo"},
	{0xF6422FEE45FD4F40ULL, "735748932.ogg"},
	{0xF64237A7D3FBB14CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bunkhigh_d.dds"},
	{0xF6428098312E9F13ULL, "and230tr_dissent_fight.ncs"},
	{0xF642FA23DDF38262ULL, "gen00fl_varric6063870.cl"},
	{0xF64307DD9ADF923DULL, "vxca_caverns2_78r_0_2.vlm"},
	{0xF643F04EF80DF6E3ULL, "follower_banter6155517.cl"},
	{0xF6440EC6823F44F7ULL, "w_bonepitdestroyed_a3_92_0.mmh"},
	{0xF64412C6823F4B04ULL, "w_bonepitdestroyed_a3_92_0.mao"},
	{0xF64424C6823F6A69ULL, "w_bonepitdestroyed_a3_92_0.msh"},
	{0xF644258A93E7307DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\thinking_enter.ani"},
	{0xF6443058616BF185ULL, "164698571.ogg"},
	{0xF6445EAA91E78AEBULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_imbz_c.anb"},
	{0xF6446D915F2A9695ULL, "mag311_keran6054389.cl"},
	{0xF6455922CA8E7FBFULL, "998167827.ogg"},
	{0xF645A4A773A410A6ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_runf.evt"},
	{0xF6460D3148BB6656ULL, "llac_lt_0_lowtown_docks_n_-65539_6.dds"},
	{0xF64612E75C0DA817ULL, "vxca_caverns1_cjs_0_0.vlm"},
	{0xF646830EB9720470ULL, "follower_banter6171910.cl"},
	{0xF646861552D5D7A5ULL, "190228393.ogg"},
	{0xF646F4B8CD308648ULL, "llac_lt_0_lowtown_a2_65536_6.dds"},
	{0xF646FA1EC755D79EULL, "482272832.ogg"},
	{0xF6470EF53DF563F2ULL, "vxlt_blackemporium_9b_0_0.vlm"},
	{0xF6473E52B49832DFULL, "isa211_isabela6072595.cl"},
	{0xF6478C1C8CC2CFB8ULL, "135202526.ogg"},
	{0xF647A9D3FB2587BCULL, "vxca_caverns1_ea7_0_0.vlm"},
	{0xF647C8938E0D12F4ULL, "gen00fl_isabela6089263.cl"},
	{0xF647DD8AD258AA32ULL, "vxdr_primevaltomb_8lr_0_0.vlm"},
	{0xF648450FA9D9DF6EULL, "mer120_hubert6059408.cl"},
	{0xF6484BA163F83C36ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dmg_v1.ani"},
	{0xF648AC9D72CF7A9AULL, "mag100_cullen6148623.cl"},
	{0xF648BB7E058CDAA4ULL, "6072676_m.fxe"},
	{0xF648C4E020B9705BULL, "follower_banter6170479.cl"},
	{0xF648D503CC6E3FE1ULL, "ave200_donnic6110609.cl"},
	{0xF64906BBFE0187E4ULL, "18800896.ogg"},
	{0xF6499EF4E5282C27ULL, "650139896.ogg"},
	{0xF649B7FF6E58F775ULL, "llac_lt_0_lowtown_docks_n_-196611_1.dds"},
	{0xF649D815F32B5AFBULL, "vxht_mansion_n_34n_0_0.vlm"},
	{0xF649FC3D4222EB4BULL, "567419823.ogg"},
	{0xF64A7121E4224261ULL, "hlac_lt_0_lowtown_n_65534_2z.dds"},
	{0xF64AA116A495948CULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicerchair.msh"},
	{0xF64AA4D8D9DAB33CULL, "6150486_m.fxe"},
	{0xF64AA716A4959F7EULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicerchair.mmh"},
	{0xF64AB316A495B315ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicerchair.mao"},
	{0xF64BBBA773A8A3F1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_runf.gad"},
	{0xF64BD9E99BFEFD8DULL, "dae000_mother6004117.cl"},
	{0xF64BE1227567F6E8ULL, "art\\characters\\playercharacter\\elfmale\\em_arm_lgte.msh"},
	{0xF64BFF2275682912ULL, "art\\characters\\playercharacter\\elfmale\\em_arm_lgte.mmh"},
	{0xF64BFFACBCF6ED34ULL, "651181453.ogg"},
	{0xF64C085A2E0299D2ULL, "llac_lt_0_lowtown_docks_n_65536_0.dds"},
	{0xF64C0F48F1F9EE0DULL, "vxlt_0_lowtown_a2_n_gpd_0_0.vlm"},
	{0xF64C250DC4E149CCULL, "hlac_lt_0_lowtown_a3_n_c_65538_5y.dds"},
	{0xF64C4C58D7B6F731ULL, "vxht_mansion2b_2yq_0_0.vlm"},
	{0xF64C7FC9FE486478ULL, "hlac_w_woundedcoast_-65535_0x.dds"},
	{0xF64CFA12592B49F4ULL, "157513895.ogg"},
	{0xF64D41B68F6B5453ULL, "t3_dog_ow1.tnt"},
	{0xF64E4B5FF9C90C97ULL, "6149205_m.fxe"},
	{0xF64EA4269276BA3AULL, "190298982.ogg"},
	{0xF64EB07A18DD274DULL, "414024116.ogg"},
	{0xF64F349AE5C66F81ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_act_c.anb"},
	{0xF64F3718F4911D60ULL, "vxw_encounter1_ha_0_0.vlm"},
	{0xF64F45486E57241AULL, "845407217.ogg"},
	{0xF64F6EBE49C655F8ULL, "303159672.ogg"},
	{0xF64F882C95A505C4ULL, "hlac_lt_gallowstemplar_n_7198_0z.dds"},
	{0xF64FD63A5860C363ULL, "seb321_leliana6180975.cl"},
	{0xF64FF48013BABF0FULL, "pro000_colonel_ewald6047383.cl"},
	{0xF6503DA21652261DULL, "803721037.ogg"},
	{0xF650464FE06086C2ULL, "807210015.ogg"},
	{0xF6506A167D3A9543ULL, "6087774_m.fxe"},
	{0xF6508444C057F0DEULL, "vxca_caverns2_82r_0_0.vlm"},
	{0xF6511B4C2B44CED9ULL, "952052037.ogg"},
	{0xF65243B72B1756D5ULL, "6116562_m.fxe"},
	{0xF652BCA773AEC39FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\ari_runf.ani"},
	{0xF652F32B6FD548DEULL, "6075132_m.fxe"},
	{0xF653686FCF03AFE8ULL, "mrl310_merrill6154555.cl"},
	{0xF653BB2A0D3A1A9EULL, "6175324_m.fxe"},
	{0xF653E072A118D3ACULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantry.phy"},
	{0xF654171C0DFC7FFFULL, "6067130_m.fxe"},
	{0xF65506DF1696EDADULL, "dae000ip_crowd_lm1_talk13.utp"},
	{0xF6552BF35B57429AULL, "310692573.ogg"},
	{0xF6556C3C4204BFACULL, "374901532.ogg"},
	{0xF655C4C571BD96F4ULL, "115411939.ogg"},
	{0xF655C5386F210E47ULL, "6165579_m.fxe"},
	{0xF6560E505F6AEA6FULL, "6208001_m.fxe"},
	{0xF65653A164048727ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dmg_v1.evt"},
	{0xF6567542C4207217ULL, "follower_banter6164733.cl"},
	{0xF656A513610AA477ULL, "gen00fl_varric6033692.cl"},
	{0xF656B097A9EE2E74ULL, "art\\characters\\heraldry\\hld_coterie.dds"},
	{0xF656FB4BC13E3FE4ULL, "vxht_mansion_n_6jv_0_0.vlm"},
	{0xF6571FD1398E26D1ULL, "isa311_isabela6091838.cl"},
	{0xF657442FF44243E3ULL, "vxlt_0_lowtown_a2_n_c_e20_0_0.vlm"},
	{0xF657553442599F70ULL, "w_encounter2_camp_mmap_ph.dds"},
	{0xF6575A74654E8E6CULL, "565642091.ogg"},
	{0xF6578B8EC0572F0CULL, "hlac_ht_keep_3734_1y.dds"},
	{0xF657D24D41477229ULL, "gen00fl_isabela6079296.cl"},
	{0xF658939D6B8B93DEULL, "218791237.ogg"},
	{0xF658C74C039BDD41ULL, "6155393_f.fxe"},
	{0xF6590F887481E125ULL, "vxlt_undercityserialkiller_1ij_0_1.vlm"},
	{0xF659325A14D10032ULL, "vxht_0_hightown_a2_31g_0_0.vlm"},
	{0xF659584A2ADA99DCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_basic.mao"},
	{0xF659644A2ADAAE7FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_basic.mmh"},
	{0xF6596A4A2ADAB821ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_basic.msh"},
	{0xF659945B06048C6BULL, "and100_anders6095902.cl"},
	{0xF65A3940CFE27D10ULL, "zz_ave_debug6065022.cl"},
	{0xF65A4902B4BC4239ULL, "ts_category.gda"},
	{0xF65A6C2438512C5AULL, "hlac_ht_0_hightown_131073_0z.dds"},
	{0xF65A71F4A23F70A9ULL, "6137497_m.fxe"},
	{0xF65A91B41D127213ULL, "ach_kill_arishok.dds"},
	{0xF65AA18431675178ULL, "724716960.ogg"},
	{0xF65B43C26E264DEDULL, "6109319_m.fxe"},
	{0xF65B98D4B6E12C35ULL, "w_encounter3_2_0.msh"},
	{0xF65B9AD4B6E12FABULL, "w_encounter3_2_0.mmh"},
	{0xF65BA1E9A862A47EULL, "zz_test_general6203148.cl"},
	{0xF65BA6D4B6E14448ULL, "w_encounter3_2_0.mao"},
	{0xF65C74A164092B70ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\dmg_v1.gad"},
	{0xF65CC247FDB98F71ULL, "hlac_dr_primevaltomb_658_6z.dds"},
	{0xF65D159AE5D278FAULL, "art\\vfx\\combat\\vfx_r_gn2_speed_act_c.mmh"},
	{0xF65E4483557D0373ULL, "follower_barks6173415.cl"},
	{0xF65E74C92F0A50C8ULL, "t3_wdg_bb2.tnt"},
	{0xF65F371796AEECB6ULL, "6129340_f.fxe"},
	{0xF65FEF55C1569E44ULL, "399632613.ogg"},
	{0xF660192EA40A4163ULL, "89413545.ogg"},
	{0xF6601ECCEB2CCC5DULL, "gen00fl_merrill6121456.cl"},
	{0xF6607B91AD660B41ULL, "147294769.ogg"},
	{0xF660A20DA008167FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_gallows.phy"},
	{0xF6617F961FC3478FULL, "785931907.ogg"},
	{0xF661CD853A5F5D2FULL, "w_bonepitdestroyed_a3_246_0.mao"},
	{0xF661D9853A5F71CCULL, "w_bonepitdestroyed_a3_246_0.mmh"},
	{0xF661DF853A5F7BBEULL, "w_bonepitdestroyed_a3_246_0.msh"},
	{0xF6624055E5082C62ULL, "153275593.ogg"},
	{0xF66275D85F7046D0ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_meda_0t.dds"},
	{0xF662C9036C649515ULL, "vxca_caverns1_mines_iao_0_0.vlm"},
	{0xF6631E53B9F7DB90ULL, "art\\characters\\creatures\\stonegolem\\c_stonegolem_head.mao"},
	{0xF6632CE0E17DB0F6ULL, "ser191_emeric6108774.cl"},
	{0xF663808DC08E32F6ULL, "w_encounter2_camp_21_0.rml"},
	{0xF664A411CAB7F76DULL, "clf260_harley6089573.cl"},
	{0xF664C3F893F14696ULL, "6074368_m.fxe"},
	{0xF664D50586FAFB40ULL, "6061795_m.fxe"},
	{0xF6651296EAC9A827ULL, "hlac_lt_undercityserialkiller_-1_29z.dds"},
	{0xF6656487F7A4A05DULL, "6139711_f.fxe"},
	{0xF6656D28B3345864ULL, "vxht_mansion_n_6mw_0_0.vlm"},
	{0xF665840F6742B3B2ULL, "hlac_lt_0_lowtown_a2_-65535_3y.dds"},
	{0xF6658E87EF5BB81EULL, "llac_lt_gallowscourtyard_n_-2_2.dds"},
	{0xF66612179C476306ULL, "art\\vfx\\cinematic\\vfx_bld_dir_dsk_p.impa.evt"},
	{0xF6665C8DE6A19C19ULL, "vxht_chantry_35a_0_1.vlm"},
	{0xF666CF739FEC0E63ULL, "ht_0_hightown_n_33_0.rml"},
	{0xF6684C2CE2B40A84ULL, "follower_banter6143296.cl"},
	{0xF66894BC24F1397AULL, "dae000_bartrand6027453.cl"},
	{0xF668DDC2B333E6CCULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.dura.ani"},
	{0xF668F302D5E61F0DULL, "vxlt_0_lowtown_a2_n_gjb_0_0.vlm"},
	{0xF668F93B75E4A2D6ULL, "art\\vfx\\combat\\vfx_itm_fel_bomb_aoe_p.impa.ani"},
	{0xF6690A9EA2EF237DULL, "hlac_lt_gallowscourtyard_n_65536_1x.dds"},
	{0xF6693012BBA32831ULL, "vxht_mansion2b_2ow_0_0.vlm"},
	{0xF66943D4E36BC4DAULL, "6052171_m.fxe"},
	{0xF6696E4355E7684AULL, "6053315_f.fxe"},
	{0xF669798DC092921AULL, "w_encounter2_camp_21_0.phy"},
	{0xF6698D90593A1D6DULL, "vxw_bonepit_26k_0_0.vlm"},
	{0xF6699CCC6DD929CBULL, "6056853_m.fxe"},
	{0xF66A20D38D6658BAULL, "w_blightlands_8_0.rml"},
	{0xF66A55600C3DBE32ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st03a_0n.dds"},
	{0xF66A612A7F8EE0D9ULL, "vxdr_primevaltomb_74b_0_0.vlm"},
	{0xF66A9B54B23AA260ULL, "436040465.ogg"},
	{0xF66AA595D5FA2480ULL, "llac_lt_warehouse_n_0_6.dds"},
	{0xF66B84430F402132ULL, "67187285.ogg"},
	{0xF66BDF5F0A5BCBB8ULL, "fer_grass.mao"},
	{0xF66BE5ED202612CDULL, "574021699.ogg"},
	{0xF66C328EDA197751ULL, "6099487_m.fxe"},
	{0xF66D1976A7A5170BULL, "hlac_lt_undercityserialkiller_-1_32z.dds"},
	{0xF66D29179C4DA951ULL, "art\\vfx\\cinematic\\vfx_bld_dir_dsk_p.impa.gad"},
	{0xF66D481C6427BAF3ULL, "hlac_dr_0_deeproads_5842_1y.dds"},
	{0xF66D576A97B920BEULL, "llac_lt_0_lowtown_docks_n_-65540_2.dds"},
	{0xF66E32E6EB90BD7FULL, "6032261_f.fxe"},
	{0xF66E38AE0A07E60CULL, "vxlt_0_lowtown_h78_0_0.vlm"},
	{0xF66E61F936FD19EEULL, "prop1_ds2y.psh.fxo"},
	{0xF66E8F2148FA8B9AULL, "mag111ip_ws_chest.utp"},
	{0xF66EB6B856E8E09BULL, "dae221ip_home_amb_fenris_wn.utp"},
	{0xF66EC20FB46062C9ULL, "dae000ip_crowd_mf3_talk6.utp"},
	{0xF66EDAC840C167E2ULL, "6081258_m.fxe"},
	{0xF66EFCA545C83A61ULL, "6043048_m.fxe"},
	{0xF66F57E2109ED2ACULL, "gen00fl_aveline6015803.cl"},
	{0xF66F7D7DDE934046ULL, "dae000_gamlen6139613.cl"},
	{0xF66FB34DC7A2EE97ULL, "vxht_brothel_n_61g_0_0.vlm"},
	{0xF66FB7A51FE9E088ULL, "962333187.ogg"},
	{0xF66FBDAA43D18A37ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\0dd_basic_2.mao"},
	{0xF66FD910DC0C6436ULL, "6031827_f.fxe"},
	{0xF66FEAC2B33A1B0EULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.dura.gad"},
	{0xF67025CFCA005D38ULL, "one300_zevran6093486.cl"},
	{0xF670283EE6AE3488ULL, "6180592_f.fxe"},
	{0xF670463E798B5AD3ULL, "303241129.ogg"},
	{0xF67121EA8EBC890DULL, "seb321_leliana6191269.cl"},
	{0xF6715E34969A8BFFULL, "6166291_m.fxe"},
	{0xF6719F9EC823B7F0ULL, "art\\levels\\ztools\\default\\generic\\1wi_basic.msh"},
	{0xF671A30DA0168A84ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_gallows.mmh"},
	{0xF671A70DA01691D7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_gallows.mao"},
	{0xF671AD9EC823CF0AULL, "art\\levels\\ztools\\default\\generic\\1wi_basic.mmh"},
	{0xF671B19EC823D641ULL, "art\\levels\\ztools\\default\\generic\\1wi_basic.mao"},
	{0xF671B90DA016B026ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1pp_gallows.msh"},
	{0xF67278599C5B5697ULL, "vxw_encounter2_gr_0_0.vlm"},
	{0xF6732A179C5216FFULL, "art\\vfx\\cinematic\\vfx_bld_dir_dsk_p.impa.ani"},
	{0xF67335D585E75F16ULL, "6079068_m.fxe"},
	{0xF67369E44D0819B1ULL, "pro000cr_athenril.utc"},
	{0xF673D84DBB60CBF8ULL, "242572338.ogg"},
	{0xF673E9DE1AE9AB51ULL, "llac_lt_darktown_-1_65.dds"},
	{0xF673F9852CA4418CULL, "vxca_cavern1_haunted_ei6_0_0.vlm"},
	{0xF6745E789C066DF9ULL, "pro000_merchant6051650.cl"},
	{0xF6745F51F7DD484FULL, "vrc321tr_door_shift_4.ncs"},
	{0xF674E9D7F73A3B00ULL, "958527433.ogg"},
	{0xF6758E9C468EE574ULL, "gen00fl_anders6114475.cl"},
	{0xF675D1432CE094F9ULL, "vxlt_0_lowtown_a2_n_f5b_0_0.vlm"},
	{0xF675E5C2B33E7EEDULL, "art\\vfx\\combat\\vfx_c_2nd_buff_loop_c.dura.evt"},
	{0xF676013B75EF3AC7ULL, "art\\vfx\\combat\\vfx_itm_fel_bomb_aoe_p.impa.evt"},
	{0xF6767350410AB265ULL, "ave310_brennan6136776.cl"},
	{0xF6768CDE2424A38BULL, "gen00fl_fenris6126877.cl"},
	{0xF676D51FAD0E5A8EULL, "248831875.ogg"},
	{0xF676D9690F4940C0ULL, "vxht_store_c0_0_0.vlm"},
	{0xF676EE1862609FF0ULL, "hlac_lt_0_lowtown_a2_n_-2_1x.dds"},
	{0xF676F055F7E22432ULL, "vxlt_0_lowtown_docks_b5o_0_0.vlm"},
	{0xF677E1A86A5F3E6BULL, "isa211_isabela6070113.cl"},
	{0xF677E4DAEEA7E867ULL, "6036561_m.fxe"},
	{0xF677FF6D4B3035EDULL, "vxw_bonepit_21j_0_0.vlm"},
	{0xF6780D9B83E3202BULL, "6065895_m.fxe"},
	{0xF67873C513659D87ULL, "86304741.ogg"},
	{0xF67891E355553DEEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basic.msh"},
	{0xF6789FE35555553FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basic.mao"},
	{0xF678ABE3555569DCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basic.mmh"},
	{0xF679354E4A525FC5ULL, "6169911_m.fxe"},
	{0xF679A9B07470B57DULL, "222980307.ogg"},
	{0xF679C0628107AAFDULL, "201960205.ogg"},
	{0xF679D05C58E32422ULL, "170575610.ogg"},
	{0xF67A1E5198019693ULL, "dae021_sister_samea_sta6142139.cl"},
	{0xF67ABA55AA8A4CE2ULL, "damagertacticspreset.gda"},
	{0xF67AC981CE40DA39ULL, "gen_im_trs_tmp_med_boo_r0.uti"},
	{0xF67BE55DF96538DDULL, "6074356_m.fxe"},
	{0xF67CFF80BABF9A18ULL, "6133874_m.fxe"},
	{0xF67D027264D1A13CULL, "vxw_bonepit_125_0_0.vlm"},
	{0xF67D223B75F59110ULL, "art\\vfx\\combat\\vfx_itm_fel_bomb_aoe_p.impa.gad"},
	{0xF67D9C360240A731ULL, "6162806_m.fxe"},
	{0xF67DF1B6202162ABULL, "vxca_caverns1_iob_0_0.vlm"},
	{0xF67E33C6D7CFAC57ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3sail.mmh"},
	{0xF67E37C6D7CFB364ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3sail.mao"},
	{0xF67E49C6D7CFD1C9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9uw_awning3sail.msh"},
	{0xF67EB15B3FC2A800ULL, "6144814_m.fxe"},
	{0xF67EE6F8C9DB5335ULL, "hlac_lt_backalley_g_0_18z.dds"},
	{0xF67EF73C01D59B94ULL, "166168893.ogg"},
	{0xF67F2774856CE4BFULL, "vxca_caverns2_9sy_0_0.vlm"},
	{0xF67FA5C3310B6683ULL, "dae000_flemeth6090068.cl"},
	{0xF681300FC3CC7F33ULL, "6069442_m.fxe"},
	{0xF68137647830F696ULL, "vxlt_0_lowtown_a2_bxq_0_0.vlm"},
	{0xF681515E93350E37ULL, "965022777.ogg"},
	{0xF6818CA12A98C962ULL, "gen_im_wep_rog_arc_act2_07.uti"},
	{0xF68245975BB1B69CULL, "hlac_lt_0_lowtown_docks_n_65536_0x.dds"},
	{0xF6824B0558FFFEC5ULL, "6141456_m.fxe"},
	{0xF682674FAC7C69E3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_en.ani"},
	{0xF68293EEDCB6116EULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.cess.evt"},
	{0xF68313C13C056A41ULL, "zz_and_debug6134092.cl"},
	{0xF6832027251CB772ULL, "147059873.ogg"},
	{0xF683358CD1C28824ULL, "sky.psh.fxo"},
	{0xF683413F969A50D8ULL, "vxlt_0_lowtown_a2_n_c_fhz_0_0.vlm"},
	{0xF68345F3A93B549DULL, "556479070.ogg"},
	{0xF6834635B736192EULL, "vxlt_0_lowtown_docks_n_b10_0_0.vlm"},
	{0xF6837EA14BCA3239ULL, "lowtown_fem_alien_elf_2_mob6166333.cl"},
	{0xF6838FD7F7F45D22ULL, "art\\characters\\weapons\\mauls\\w_mal_ml105a.msh"},
	{0xF68391D7F7F460D8ULL, "art\\characters\\weapons\\mauls\\w_mal_ml105a.mmh"},
	{0xF683EC129D8E1DF0ULL, "vxlt_darktown_58f_0_1.vlm"},
	{0xF6845DFFBC19500CULL, "6059688_m.fxe"},
	{0xF684F2415A227ADFULL, "vxw_blightlands_bj_0_0.vlm"},
	{0xF68511DD8B214258ULL, "vxht_0_hightown_n_c3u_0_1.vlm"},
	{0xF6854B7AC8CF7719ULL, "follower_barks6173408.cl"},
	{0xF6857FC66603CA86ULL, "6144290_m.fxe"},
	{0xF685B6BB46B83840ULL, "vxlt_0_lowtown_docks_a3_n_cob_0_0.vlm"},
	{0xF68613CC9F8F5931ULL, "vxht_mansion2b_2eu_0_0.vlm"},
	{0xF686FE5099FC9EA9ULL, "vxw_woundedcoast_1k8_0_0.vlm"},
	{0xF6879640A3194A60ULL, "vxht_mansion_n_5cv_0_0.vlm"},
	{0xF687CB812C67407DULL, "223916244.ogg"},
	{0xF688290DFB57A605ULL, "mrl150_merrill6091891.cl"},
	{0xF6882A4FEBC2B4AEULL, "6041208_f.fxe"},
	{0xF6883253748D83B2ULL, "lgt111pt_alistair_drunk.ncs"},
	{0xF6888AEEDCBA6D19ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.cess.gad"},
	{0xF688AF67B87F6DB8ULL, "character_character_p18.psh.fxo.dx11"},
	{0xF6890708B1188EEEULL, "hlac_ca_caverns2prc_5802_53x.dds"},
	{0xF6891648BE05A851ULL, "6093687_m.fxe"},
	{0xF6897E3A7A4F301DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothicnb.mao"},
	{0xF689816ABF894115ULL, "297967591.ogg"},
	{0xF689823A7A4F3756ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothicnb.mmh"},
	{0xF6898C3A7A4F4854ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothicnb.msh"},
	{0xF6899F18E9C85E12ULL, "6109051_m.fxe"},
	{0xF689AD75FE7CE369ULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_002a_0t.dds"},
	{0xF689B8F7840506E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_action.ani"},
	{0xF68A573AEB5B515DULL, "6079056_m.fxe"},
	{0xF68A66C39A8243BAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\lt_set_tarp_dp.dds"},
	{0xF68A67F660EA1A7EULL, "785633655.ogg"},
	{0xF68A8137178EB36EULL, "hlac_w_bonepit_65541_0y.dds"},
	{0xF68AFEDEFD8C308AULL, "zz_vault_debug6177754.cl"},
	{0xF68B69882D194FDAULL, "6207745_f.fxe"},
	{0xF68B7E16A4CCF019ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_vnicerchair.phy"},
	{0xF68B8491DDC84020ULL, "vxlt_backalley_g_1vn_0_1.vlm"},
	{0xF68BD50877E71857ULL, "43838081.ogg"},
	{0xF68CE4A04111BF09ULL, "sw_blood_dot.mao"},
	{0xF68E98D7F7FE7463ULL, "art\\characters\\weapons\\mauls\\w_mal_ml105a.phy"},
	{0xF68F09379D13E03AULL, "6162360_m.fxe"},
	{0xF68F100B6CDD8C0DULL, "fen210_fenris6108752.cl"},
	{0xF68F73BB576ABC16ULL, "vxlt_warehouse_n_136_0_0.vlm"},
	{0xF68F7F4FAC871CDAULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_en.evt"},
	{0xF68F9BEEDCC0A967ULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mem_c.cess.ani"},
	{0xF68FFCE48ACC2C4FULL, "228771554.ogg"},
	{0xF6904B227EF6E190ULL, "30355813.ogg"},
	{0xF69059C4D10F0453ULL, "757924203.ogg"},
	{0xF690ADC6E58256CCULL, "415865256.ogg"},
	{0xF690DD3F1090728CULL, "vxca_cavern1_haunted_ec4_0_0.vlm"},
	{0xF690E2C78071A8CAULL, "art\\levels\\4cinematics\\ferelden\\prp_bookofinquisition.mao"},
	{0xF691277EB1383E75ULL, "6043282_m.fxe"},
	{0xF691735C19DA5965ULL, "hlac_lt_darktown_-1_38z.dds"},
	{0xF6934C51F7F7D8F6ULL, "vrc321tr_door_shift_4.utt"},
	{0xF6934F7E3AB30711ULL, "one100_party_comments6112675.cl"},
	{0xF6936608ED10F0EAULL, "vxlt_darktown_530_0_0.vlm"},
	{0xF693EF5B3F31A44BULL, "static_static_p26.psh.fxo.dx11"},
	{0xF693FFF035F2FA46ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_one_chance.evt"},
	{0xF694612F7F90CBEBULL, "905986341.ogg"},
	{0xF694B772A1502577ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantry.mmh"},
	{0xF694BB72A1502B84ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantry.mao"},
	{0xF694CD72A1504AE9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\8gr_chantry.msh"},
	{0xF694F6C6D0324374ULL, "187668257.ogg"},
	{0xF695247BFA2BBCC8ULL, "one200_gold_followers6101133.cl"},
	{0xF695691E75A74EA1ULL, "vxlt_gallowstemplar_f_5ar_0_0.vlm"},
	{0xF69636363ABAC160ULL, "vxlt_hangedman_241_0_0.vlm"},
	{0xF69644DD2FC60B57ULL, "6134946_m.fxe"},
	{0xF69676BA9D2A5E21ULL, "sky_ht_mansion_n.mao"},
	{0xF6967E4FAC8D39CDULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_en.gad"},
	{0xF69684DADC5D2565ULL, "vxlt_0_lowtown_docks_n_b5b_0_1.vlm"},
	{0xF696B29602875F40ULL, "507195591.ogg"},
	{0xF696B59893E65E30ULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v1_c.mmh"},
	{0xF6970777990DE44CULL, "follower_banter6136722.cl"},
	{0xF697A0F784111B57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_action.evt"},
	{0xF6980178DF62B06BULL, "character_character_p242.psh.fxo.dx11.assembly"},
	{0xF69812738E76746FULL, "dae000_keep_guard_1_sta6163796.cl"},
	{0xF6981B316C957699ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_en.ani"},
	{0xF698685452EB260CULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfl.ani"},
	{0xF698999F999F28DFULL, "6187719_m.fxe"},
	{0xF6996F3D45F41F00ULL, "vxlt_gallowscourtyard_n_6py_0_0.vlm"},
	{0xF699B0D77A0956D2ULL, "6173076_m.fxe"},
	{0xF69A86A49DBB26C6ULL, "hlac_lt_0_lowtown_a2_n_c_-1_2x.dds"},
	{0xF69A8DAE2263421AULL, "1049435756.ogg"},
	{0xF69AA8317A5FCBD3ULL, "llac_ht_0_hightown_a3_131071_2.dds"},
	{0xF69ABB29C600CAFEULL, "539662283.ogg"},
	{0xF69ACB26989A8533ULL, "364969682.ogg"},
	{0xF69B16F035F93F91ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_one_chance.gad"},
	{0xF69B2E88177588F2ULL, "qun230cr_guild_contact6068876.cl"},
	{0xF69C7DFABDA87933ULL, "vis121_saemus6118477.cl"},
	{0xF69D1064A5697282ULL, "follower_banter6169294.cl"},
	{0xF69D28C0A6FBA4D9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_winup.msh"},
	{0xF69D32C0A6FBB5C7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_winup.mmh"},
	{0xF69D36C0A6FBBC14ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_winup.mao"},
	{0xF69D43D116BFD197ULL, "vxht_0_hightown_a2_an1_0_0.vlm"},
	{0xF69D8AE355744DA7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\0ww_basic.phy"},
	{0xF69DE5984960D496ULL, "vxlt_warehouse_n_187_0_0.vlm"},
	{0xF69E016282CC4E83ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_stocks_sa.dds"},
	{0xF69E257A4635F443ULL, "dae000_arishok6067743.cl"},
	{0xF69E2E896D71844DULL, "515404745.ogg"},
	{0xF69E6E677B2D3A4CULL, "390347471.ogg"},
	{0xF69E8BCC67309A9CULL, "art\\characters\\playercharacter\\humanmale\\cth_nobb.mao"},
	{0xF69E97CC6730AF3FULL, "art\\characters\\playercharacter\\humanmale\\cth_nobb.mmh"},
	{0xF69E9DCC6730B9E1ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobb.msh"},
	{0xF69EA1F784173BE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\hold_at_knifepoint_action.gad"},
	{0xF69ECE1C4DA6AC97ULL, "6042694_m.fxe"},
	{0xF69EF6316C9B56DBULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_en.gad"},
	{0xF69F48C30C30F1BBULL, "vxw_sundermount_345_0_0.vlm"},
	{0xF69F607DF996862AULL, "6151344_m.fxe"},
	{0xF69F755452F15B4EULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfl.gad"},
	{0xF6A0111EC2CFF865ULL, "640578681.ogg"},
	{0xF6A01B2CE39F1FEAULL, "vxw_woundedcoast_1gs_0_0.vlm"},
	{0xF6A07395B877417DULL, "w_bonepitdestroyed_a3_154_0.mao"},
	{0xF6A07795B87747B6ULL, "w_bonepitdestroyed_a3_154_0.mmh"},
	{0xF6A08195B87758B4ULL, "w_bonepitdestroyed_a3_154_0.msh"},
	{0xF6A08FC1519A591AULL, "t3_dps_b01.tnt"},
	{0xF6A0C60F6C050CF1ULL, "zz_pridedemon.cut"},
	{0xF6A0F41CA032E5ECULL, "107633688.ogg"},
	{0xF6A18837D5CA5DAEULL, "6164982_m.fxe"},
	{0xF6A1CB4AE4C94316ULL, "zz_mer_debug6017108.cl"},
	{0xF6A217F035FF603FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_one_chance.ani"},
	{0xF6A280603800CE57ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed2.gad"},
	{0xF6A2C90574619B0BULL, "vxw_bonepitdestroyed_a3_j1_0_0.vlm"},
	{0xF6A32FC0E7A5AF55ULL, "6116543_f.fxe"},
	{0xF6A3675C221D0C7FULL, "art\\vfx\\combat\\vfx_c_pride_cast_f_c.impa.gad"},
	{0xF6A37715D16A8659ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\dw.anb"},
	{0xF6A38AB4DB370A97ULL, "pin_li.mao"},
	{0xF6A3A945C681399BULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight1a.mmh"},
	{0xF6A3B545C6814E38ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight1a.mao"},
	{0xF6A3C745C6816C85ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_scraplight1a.msh"},
	{0xF6A3D89893F1244FULL, "art\\vfx\\combat\\vfx_m_sta_rng_atk_v1_c.anb"},
	{0xF6A5B10931039218ULL, "zz_one_debug6069473.cl"},
	{0xF6A5BBF77B52C141ULL, "vxht_chantry_n_13w_0_0.vlm"},
	{0xF6A5CA00EE2417BCULL, "6046846_m.fxe"},
	{0xF6A613316CA1A6C0ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\custom\\am_weep_en.evt"},
	{0xF6A65052EFB07871ULL, "vxca_caverns1_fn8_0_0.vlm"},
	{0xF6A6705452F7712DULL, "art\\characters\\creatures\\hurlock\\animation\\sword_shield\\hurlock_sw_walkfl.evt"},
	{0xF6A7B273A02373DFULL, "ht_0_hightown_n_33_0.anb"},
	{0xF6A7FE2275B6860DULL, "art\\characters\\playercharacter\\elfmale\\em_arm_lgte.phy"},
	{0xF6A8A5B715FC6AEAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamp.mao"},
	{0xF6A8A9B715FC7139ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamp.mmh"},
	{0xF6A8B3B715FC8227ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_lamp.msh"},
	{0xF6A941235CD62F5AULL, "hlac_lt_gallowstemplar_f_2391_0x.dds"},
	{0xF6A95C5C2221656CULL, "art\\vfx\\combat\\vfx_c_pride_cast_f_c.impa.evt"},
	{0xF6A99560380710E4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed2.evt"},
	{0xF6A9A0868557A381ULL, "6217393_m.fxe"},
	{0xF6AA0B044F19E786ULL, "342163173.ogg"},
	{0xF6AA43AC8F326705ULL, "ico_2h_sword_2.dds"},
	{0xF6AA7C9ACA7328C1ULL, "hlac_lt_0_lowtown_a2_n_-131073_0y.dds"},
	{0xF6AAE5815D5BEA44ULL, "vxw_bonepit_15h_0_0.vlm"},
	{0xF6AAF5268184CAC5ULL, "403870135.ogg"},
	{0xF6AB48C112B0EBB7ULL, "fx_fireflamefb03_blend.mao"},
	{0xF6AB67F6D89B7055ULL, "hlac_lt_0_lowtown_131072_4x.dds"},
	{0xF6ABC21818B61557ULL, "hlac_lt_gallowsprison_n_-1_14y.dds"},
	{0xF6ABF4A4B8B4CB28ULL, "6109387_m.fxe"},
	{0xF6AC137929E72506ULL, "6164472_m.fxe"},
	{0xF6AC800ABC136984ULL, "125581820.ogg"},
	{0xF6ACDD0D6A1A179BULL, "686900628.ogg"},
	{0xF6AD40E3781FD2C2ULL, "708296606.ogg"},
	{0xF6AD68DFC268EF7AULL, "456770210.ogg"},
	{0xF6ADBA9FFE270A3BULL, "vxw_sundermount_334_0_0.vlm"},
	{0xF6ADC4D663014A38ULL, "398007485.ogg"},
	{0xF6AE1616DD159B5AULL, "686130402.ogg"},
	{0xF6AE3C6029DF9EA1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.ani"},
	{0xF6AF1A8470E14A19ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_glv_wrba_0s.dds"},
	{0xF6AF3E8A66DCC781ULL, "607751494.ogg"},
	{0xF6AF7C9503C02B63ULL, "fen110_anso6094966.cl"},
	{0xF6AFCC4751EF4219ULL, "vxw_bonepitmining_a2_1ki_0_0.vlm"},
	{0xF6AFE0AF1C05DC85ULL, "hlac_ht_0_hightown_a3_-65536_3x.dds"},
	{0xF6B02607E73D5526ULL, "art\\characters\\creatures\\spider\\animation\\p_wkf.ani"},
	{0xF6B02AC32E95E64BULL, "vxw_encounter2_ja_0_0.vlm"},
	{0xF6B02DD31F1596B6ULL, "vxlt_foundry_n_v4_0_0.vlm"},
	{0xF6B0A5A9CE287DBDULL, "vxlt_blackemporium_ii_0_0.vlm"},
	{0xF6B0F1544E67003EULL, "vxht_0_hightown_a2_bub_0_0.vlm"},
	{0xF6B100D38DA22571ULL, "w_blightlands_8_0.mmh"},
	{0xF6B10CD38DA23912ULL, "w_blightlands_8_0.mao"},
	{0xF6B113569695881CULL, "6135959_m.fxe"},
	{0xF6B11AD38DA2512FULL, "w_blightlands_8_0.msh"},
	{0xF6B1AE6003B55BFCULL, "vxlt_0_lowtown_a3_n_c_h2g_0_0.vlm"},
	{0xF6B2753F6F755228ULL, "hlac_lt_0_lowtown_n_-196609_2z.dds"},
	{0xF6B286F6442DA8CDULL, "pro000_cut2d_thefirstyear6015282.cl"},
	{0xF6B2ABCA8FB9DD48ULL, "toolset_groups.gda"},
	{0xF6B2BC17807A48EDULL, "6129040_f.fxe"},
	{0xF6B2D0922159FE16ULL, "vxht_0_hightown_a2_9h2_0_0.vlm"},
	{0xF6B317C497CB0A19ULL, "hlac_dr_0_deeproads_5842_0x.dds"},
	{0xF6B32203E4E19F11ULL, "6171994_m.fxe"},
	{0xF6B33578F44AF40AULL, "zz_followers_banter6159372.cl"},
	{0xF6B36E5C59449018ULL, "6165626_m.fxe"},
	{0xF6B39DCDC85F9A94ULL, "hlac_lt_0_lowtown_docks_a3_65535_1x.dds"},
	{0xF6B3DBF99CFD6B08ULL, "one100dg_nabil6063820.cl"},
	{0xF6B3EC840AF18A3AULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_barge_anim.dura.gad"},
	{0xF6B43BDDE01A303CULL, "6041946_m.fxe"},
	{0xF6B48BF98B22CE76ULL, "plotaudio_base.gda"},
	{0xF6B4E911028D5F5EULL, "763347004.ogg"},
	{0xF6B546C45055B4F6ULL, "gen00fl_fenris6126395.cl"},
	{0xF6B5476029E5D043ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.gad"},
	{0xF6B561D6EA5463D6ULL, "254365076.ogg"},
	{0xF6B588755F26F01DULL, "6116364_m.fxe"},
	{0xF6B5B59F2FCDB822ULL, "hlac_lt_0_lowtown_a2_n_-196609_4x.dds"},
	{0xF6B5D34A92137D61ULL, "dae011_corff_sta6152751.cl"},
	{0xF6B6E2845C683646ULL, "vxht_0_hightown_a2_6pm_0_0.vlm"},
	{0xF6B7051A097DD314ULL, "hlac_lt_0_lowtown_1_7y.dds"},
	{0xF6B7445C222D79C5ULL, "art\\vfx\\combat\\vfx_c_pride_cast_f_c.impa.ani"},
	{0xF6B7AD60381376DDULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\animation\\4co_lowtown_gate.opentoclosed2.ani"},
	{0xF6B7B6FFCD471E05ULL, "6019043_m.fxe"},
	{0xF6B7D5044FE722EEULL, "w_bonepit_129_0.mmh"},
	{0xF6B7E1044FE73685ULL, "w_bonepit_129_0.mao"},
	{0xF6B7EF044FE74EDCULL, "w_bonepit_129_0.msh"},
	{0xF6B7F552FC14EA20ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_meda_0d.dds"},
	{0xF6B9020C448B587FULL, "dae000_arishok6067798.cl"},
	{0xF6B964D712421B9BULL, "vxht_bartrand_n_64p_0_0.vlm"},
	{0xF6B9B8C4A47A8627ULL, "zz_slate2_tracking.cut"},
	{0xF6BA0DF80B0C5E2FULL, "hlac_w_encounter3_-131072_0y.dds"},
	{0xF6BA257984CD89CAULL, "213794516.ogg"},
	{0xF6BAB9CD6E221E54ULL, "6151626_f.fxe"},
	{0xF6BAEF840AF7AE51ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_barge_anim.dura.evt"},
	{0xF6BB003DA156CB87ULL, "814030137.ogg"},
	{0xF6BB8F803B46AD6FULL, "qun221_petrice6083754.cl"},
	{0xF6BBAC1DF591F5C2ULL, "836110537.ogg"},
	{0xF6BBD74048A01DB8ULL, "vxca_caverns2_77l_0_0.vlm"},
	{0xF6BC246029EBB328ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\arm_around_shoulder_reaction_exit.evt"},
	{0xF6BC2E0C014E731CULL, "291706692.ogg"},
	{0xF6BC530C8C9E9D24ULL, "vxlt_warehouse_n_13t_0_0.vlm"},
	{0xF6BC6D4A2B2F3AC4ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1wo_basic.phy"},
	{0xF6BD0E07E747B697ULL, "art\\characters\\creatures\\spider\\animation\\p_wkf.evt"},
	{0xF6BD446B3D0E2FFEULL, "vxlt_0_lowtown_a3_n_c_fen_0_0.vlm"},
	{0xF6BD5A742D9CB222ULL, "6179975_m.fxe"},
	{0xF6BEA7556D9B6A85ULL, "w_woundedcoast_30_0.anb"},
	{0xF6BF31A56E8BFD22ULL, "vxlt_0_lowtown_n_by1_0_0.vlm"},
	{0xF6BFDBF5343C9CDFULL, "vxw_sundermount_2as_0_0.vlm"},
	{0xF6BFED9F2D0BFCF5ULL, "w_encounter3_91_0.msh"},
	{0xF6BFEF9F2D0C006BULL, "w_encounter3_91_0.mmh"},
	{0xF6BFFB9F2D0C1408ULL, "w_encounter3_91_0.mao"},
	{0xF6C00DF4884D5412ULL, "6177267_m.fxe"},
	{0xF6C083D2E74E8E82ULL, "gen00fl_aveline6115406.cl"},
	{0xF6C0BD1CAF4E5891ULL, "vxdr_primevaltomb_2vw_0_2.vlm"},
	{0xF6C19071F57BC6F5ULL, "299403602.ogg"},
	{0xF6C1C14D439F2F98ULL, "dr_0_deeproads_5839_lc.mmh"},
	{0xF6C2073A3F6104DEULL, "vxdr_primevaltomb_543_0_0.vlm"},
	{0xF6C241DCFC6CA331ULL, "364539043.ogg"},
	{0xF6C364AF03DAC2F0ULL, "6153300_m.fxe"},
	{0xF6C3B27EE16F9172ULL, "622788683.ogg"},
	{0xF6C3CF8789804B10ULL, "rdr110cr_guard_warehouse2.utc"},
	{0xF6C3F1D5526AB1CDULL, "gen00fl_sebastian6178332.cl"},
	{0xF6C3F7F67EEE4191ULL, "vxht_bartrand_n_5vk_0_0.vlm"},
	{0xF6C40F07E74DD720ULL, "art\\characters\\creatures\\spider\\animation\\p_wkf.gad"},
	{0xF6C4DF1FFD6E9687ULL, "vxca_caverns1_mines_j25_0_2.vlm"},
	{0xF6C5728DC0E0E543ULL, "w_encounter2_camp_21_0.msh"},
	{0xF6C5766607E60B5AULL, "gen00fl_fenris6103354.cl"},
	{0xF6C5808DC0E0FC9DULL, "w_encounter2_camp_21_0.mmh"},
	{0xF6C5848DC0E103AEULL, "w_encounter2_camp_21_0.mao"},
	{0xF6C584905E657D82ULL, "gen00fl_aveline6139080.cl"},
	{0xF6C59271C78575F3ULL, "dae000_male_commoner1_mob6174468.cl"},
	{0xF6C5B6404C20D0FAULL, "vxw_bonepitmining_a2_2ao_0_0.vlm"},
	{0xF6C6008AF605C56DULL, "14138343.ogg"},
	{0xF6C6016380AD5B5CULL, "984303045.ogg"},
	{0xF6C607EEF0786B0EULL, "zz_followers_banter6159714.cl"},
	{0xF6C60A0B71B9EC83ULL, "vxlt_warehouse_15j_0_0.vlm"},
	{0xF6C60DE2FE043DCDULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_sandpile2a_2.mao"},
	{0xF6C664CEB5DFF4C6ULL, "mer110cr_mine_looter.utc"},
	{0xF6C696B8612E2202ULL, "ico_gloves_medium_3.dds"},
	{0xF6C6EA62196623F3ULL, "6105436_m.fxe"},
	{0xF6C70B5804036C99ULL, "hlac_w_sundermount_-65532_0z.dds"},
	{0xF6C70E040928B332ULL, "563274238.ogg"},
	{0xF6C7E315974A0B3AULL, "252052899.ogg"},
	{0xF6C7F562644DB9ECULL, "vxht_0_hightown_anb_0_0.vlm"},
	{0xF6C807840B026120ULL, "art\\vfx\\environment\\level_fx\\ambient\\vfx_barge_anim.dura.ani"},
	{0xF6C81A9C9A179F14ULL, "hlac_ht_0_hightown_n_-131071_0z.dds"},
	{0xF6C87AFC49A0092BULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfr.ani"},
	{0xF6C8AEC21A1E4E86ULL, "vxlt_0_lowtown_a3_n_c_ffo_0_0.vlm"},
	{0xF6C946EA26A062D9ULL, "945009499.ogg"},
	{0xF6C9CFC2016F5A66ULL, "6048486_m.fxe"},
	{0xF6CA1BC0A7227CFCULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\1ww12_winup.phy"},
	{0xF6CA50C28D9FB9E1ULL, "pro000st_aveline_gamlen.stg"},
	{0xF6CA5E51B7330232ULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_bookinquisb_d.dds"},
	{0xF6CAC4E97E94B5A4ULL, "vxlt_warehouse_n_18u_0_0.vlm"},
	{0xF6CAECF2A21549EBULL, "hlac_lt_0_lowtown_docks_a3_n_-65539_2z.dds"},
	{0xF6CB459BD17BFAC3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_rght_180.evt"},
	{0xF6CBFD7B0D841085ULL, "dae310ar_lowtown.are"},
	{0xF6CC299E9CA6A20BULL, "zz_mrl_debug6184196.cl"},
	{0xF6CC4486680B7BD9ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_pillar.msh"},
	{0xF6CC4E86680B8CC7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_pillar.mmh"},
	{0xF6CC5286680B9314ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_pillar.mao"},
	{0xF6CC907C81932F7DULL, "hlac_ht_chantry_n_2534_9z.dds"},
	{0xF6CCB19F0AE8897BULL, "6113911_f.fxe"},
	{0xF6CD0E7D1282174BULL, "vxw_encounter2_pc_0_0.vlm"},
	{0xF6CD264B72F4076FULL, "6208276_m.fxe"},
	{0xF6CD3625FEF77D4FULL, "zz_test_training.stg"},
	{0xF6CD7BFF4A3C4781ULL, "hlac_lt_0_lowtown_docks_n_-131074_5z.dds"},
	{0xF6CD7D09E3CD386CULL, "bld_dri_small.mao"},
	{0xF6CD7FF6FB17D5DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_rhand.evt"},
	{0xF6CD8915AF2B13BCULL, "w_bonepitmining_a2_153_0.mao"},
	{0xF6CD9515AF2B285FULL, "w_bonepitmining_a2_153_0.mmh"},
	{0xF6CD9B15AF2B3201ULL, "w_bonepitmining_a2_153_0.msh"},
	{0xF6CD9C2D21FFA758ULL, "897722881.ogg"},
	{0xF6CDDB16CD0D11D7ULL, "dm_whiskey.mor"},
	{0xF6CDDD40C1A127A8ULL, "132801089.ogg"},
	{0xF6CE118B4B5465DBULL, "6118763_f.fxe"},
	{0xF6CE4DD22632B55FULL, "vxw_sundermount_2jr_0_0.vlm"},
	{0xF6CED8D30E5321C6ULL, "166103302.ogg"},
	{0xF6CEE7510A8D8A46ULL, "6032876_m.fxe"},
	{0xF6CF1F616B72CEB6ULL, "w_encounter1_21_0.msh"},
	{0xF6CF29616B72DFB4ULL, "w_encounter1_21_0.mmh"},
	{0xF6CF2D616B72E6C7ULL, "w_encounter1_21_0.mao"},
	{0xF6CF65E82864BF04ULL, "hlac_w_encounter1_-65531_0x.dds"},
	{0xF6CF81CC3E31E90EULL, "vxw_bonepitdestroyed_a3_p2_0_0.vlm"},
	{0xF6CF90C596C4632BULL, "hlac_lt_0_lowtown_docks_a3_n_0_4y.dds"},
	{0xF6CF957217E24F8FULL, "vxlt_0_lowtown_docks_a3_n_9a1_0_0.vlm"},
	{0xF6D0767CB317889AULL, "gen_im_trs_qun_two_2hs_r1.uti"},
	{0xF6D086CD8BDAE89FULL, "stoney_n.dds"},
	{0xF6D0A902BDAE021EULL, "6096311_m.fxe"},
	{0xF6D0C6AA3F86F3D7ULL, "vxlt_gallowstemplar_n_46y_0_1.vlm"},
	{0xF6D0ECC36F0AC0BEULL, "805108938.ogg"},
	{0xF6D1151947B63509ULL, "w_bonepitdestroyed_a3_125_0.phy"},
	{0xF6D1225C5A674226ULL, "vxdr_primevaltomb_53o_0_1.vlm"},
	{0xF6D12DB882BDBD7FULL, "6173705_m.fxe"},
	{0xF6D1BF36BC6F3D0CULL, "isa200cr_tevinter_highmage.utc"},
	{0xF6D1D68BF5D0DB70ULL, "6156000_m.fxe"},
	{0xF6D21E9BD181D744ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_rght_180.gad"},
	{0xF6D229C6381BA0D6ULL, "vxw_blightlands_av_0_0.vlm"},
	{0xF6D2B6C0CFE61138ULL, "llac_w_bonepit_327682_0.dds"},
	{0xF6D2DE82B5F27336ULL, "w_sundermount_62_0_ll.mmh"},
	{0xF6D300C51D66E285ULL, "192252336.ogg"},
	{0xF6D37D3D2FAE3366ULL, "vxw_woundedcoast_120_0_0.vlm"},
	{0xF6D3B2A30FBD5800ULL, "6128316_f.fxe"},
	{0xF6D478F6FB1DE81CULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_rhand.gad"},
	{0xF6D47BEBE293AA17ULL, "hm_bonnylem.mor"},
	{0xF6D51B9B17C4FEE2ULL, "vxw_sundermount_31w_0_0.vlm"},
	{0xF6D52A3DB15499BCULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhanggreen.mao"},
	{0xF6D5363DB154AE5FULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhanggreen.mmh"},
	{0xF6D53C3DB154B801ULL, "art\\levels\\races\\proxy\\props\\rugs_poor\\prp_zzhanggreen.msh"},
	{0xF6D53FDF0608FC7BULL, "vxlt_hangedman_n_1xj_0_0.vlm"},
	{0xF6D5BF3C44AD024DULL, "llac_lt_0_lowtown_n_65534_4.dds"},
	{0xF6D5FE7B0D8C676BULL, "dae310ar_lowtown.ncs"},
	{0xF6D5FE944D406A39ULL, "818269063.ogg"},
	{0xF6D63C4FDBC5B56AULL, "dae000ip_crowd_mf4_talk13.utp"},
	{0xF6D662FC49AC1DA2ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfr.evt"},
	{0xF6D66B1719F2EA06ULL, "6043739_m.fxe"},
	{0xF6D6CD91265BD481ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.impa.ani"},
	{0xF6D82D9BD1865C22ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_attk_trn_rght_180.ani"},
	{0xF6D89C15583A054DULL, "combat_dw_r2_combo.gda"},
	{0xF6D8C27F7F11E098ULL, "vfx_sparkae_w.dds"},
	{0xF6D8CB9323503523ULL, "zz_otl_debug6060966.cl"},
	{0xF6DA9BF071787ADBULL, "vrc321cr_servant_ghost.ncs"},
	{0xF6DAA9ED916A7C44ULL, "vxca_caverns1_e7m_0_0.vlm"},
	{0xF6DB21DE27D18B99ULL, "vxw_bonepitmining_a2_1zn_0_0.vlm"},
	{0xF6DB2C5D00DB1F91ULL, "859410316.ogg"},
	{0xF6DB5216F9414000ULL, "6170761_m.fxe"},
	{0xF6DB77F6FB24051AULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\rogue\\dual weapons\\dw_right_down_rhand.ani"},
	{0xF6DB7A7DD3454ACAULL, "gen00fl_fenris6113908.cl"},
	{0xF6DB951D30E552ECULL, "hla_ht_0_hightown_n_-65535_0z.dds"},
	{0xF6DB99E1798AB1D9ULL, "dae021_brothel_onenter6196711.cl"},
	{0xF6DBDC14498A92B6ULL, "ser211cr_mother.utc"},
	{0xF6DCAEEDAF1EA0D8ULL, "follower_banter6136366.cl"},
	{0xF6DCD89126605323ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.impa.gad"},
	{0xF6DCF7E18DE5AAFEULL, "mcr370_assignments6176755.cl"},
	{0xF6DCFE7B0D9285B2ULL, "dae310ar_lowtown.lst"},
	{0xF6DD0B5C40C5E302ULL, "6053305_m.fxe"},
	{0xF6DD61FC49B23AB5ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_cwkfr.gad"},
	{0xF6DD623A34FF77C7ULL, "gen00fl_aveline6114837.cl"},
	{0xF6DD8403DDF7A008ULL, "hlac_ht_0_hightown_a2_65536_0z.dds"},
	{0xF6DDD7853C3D333EULL, "195283802.ogg"},
	{0xF6DDF0120F3FA4D9ULL, "vxdr_primevaltomb_7lj_0_0.vlm"},
	{0xF6DE220D847117E6ULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.cess.evt"},
	{0xF6DE386EF3E1B278ULL, "vxht_mansion_n_5a4_0_0.vlm"},
	{0xF6DEAC295EBF9E55ULL, "57169768.ogg"},
	{0xF6DEC89985EDAD70ULL, "6105986_f.fxe"},
	{0xF6DF425188006457ULL, "follower_banter6155979.cl"},
	{0xF6E02461AAA618C4ULL, "844252702.ogg"},
	{0xF6E02D0FEFD7501CULL, "vxlt_0_lowtown_a3_n_c_byt_0_0.vlm"},
	{0xF6E04846FFCD84D6ULL, "vxw_encounter2_8b_0_0.vlm"},
	{0xF6E08374034B235CULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_spitoon_d.dds"},
	{0xF6E085F4F5C89996ULL, "6148221_m.fxe"},
	{0xF6E096EAB3318E82ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv1.gad"},
	{0xF6E1A36C0A8F8FB3ULL, "llac_lt_0_lowtown_a2_-1_2.dds"},
	{0xF6E2D6F7C1B2FCF5ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4x.phy"},
	{0xF6E2DEFE0C62B7ADULL, "hlac_lt_0_lowtown_a2_-262144_1y.dds"},
	{0xF6E3984E0BD17837ULL, "143778586.ogg"},
	{0xF6E3B59126663608ULL, "art\\vfx\\combat\\vfx_gui_pp_selected_p.impa.evt"},
	{0xF6E415E18BBEDB1CULL, "hlac_w_bonepitmining_a2_65540_0x.dds"},
	{0xF6E4232695303A5EULL, "vxw_bonepit_z0_0_0.vlm"},
	{0xF6E43B2976CBC3BFULL, "vxht_0_hightown_akb_0_0.vlm"},
	{0xF6E48CE3A4AEB60EULL, "hlac_lt_0_lowtown_n_1_1x.dds"},
	{0xF6E521EE976DD5F5ULL, "gen00fl_isabela6096826.cl"},
	{0xF6E5390D84775D31ULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.cess.gad"},
	{0xF6E54A94AC980F74ULL, "dog_summon.dds"},
	{0xF6E56F2B7622A5D9ULL, "hlac_w_blightlands_393213_0y.dds"},
	{0xF6E692556DBC9924ULL, "w_woundedcoast_30_0.msh"},
	{0xF6E6A4556DBCB78DULL, "w_woundedcoast_30_0.mao"},
	{0xF6E6A8556DBCBEC6ULL, "w_woundedcoast_30_0.mmh"},
	{0xF6E6FCF1AD9EFA02ULL, "vxlt_backalley_n_1wz_0_1.vlm"},
	{0xF6E7B9EAB337E979ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv1.evt"},
	{0xF6E993BB19C7A419ULL, "vxw_bonepitmining_a2_1wm_0_0.vlm"},
	{0xF6EA0F9338A33AC0ULL, "abi_one.gda"},
	{0xF6EAB8C8163F623EULL, "vxht_0_hightown_a2_bif_0_0.vlm"},
	{0xF6EB19B0AFEC1086ULL, "723534071.ogg"},
	{0xF6EB26AA022BD863ULL, "follower_banter6160755.cl"},
	{0xF6EB429F221CA613ULL, "483950051.ogg"},
	{0xF6EBF899A2B2C629ULL, "hlac_lt_0_lowtown_a2_n_65535_4y.dds"},
	{0xF6EBFAA075A80E7FULL, "vxlt_0_lowtown_docks_a3_n_9bt_0_0.vlm"},
	{0xF6EC21A823CD49C7ULL, "qun110_petrice_intro.stg"},
	{0xF6EC3A0D847D7DDFULL, "art\\vfx\\combat\\vfx_c_debili_proj_p.cess.ani"},
	{0xF6EC48A8E4D43A14ULL, "and111_refugee_barks6111542.cl"},
	{0xF6EC67D371F0FD2CULL, "dae000_orsino6085471.cl"},
	{0xF6EC6FAD5C34D56EULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_wobble.ani"},
	{0xF6ED6D611169D811ULL, "41023665.ogg"},
	{0xF6EE144FF23A6B0AULL, "6025326_m.fxe"},
	{0xF6EE4553A0A915FCULL, "vxw_encounter2_66_0_0.vlm"},
	{0xF6EE498CFDCC6609ULL, "vxlt_hangedman_n_27z_0_0.vlm"},
	{0xF6EE81D468555FEBULL, "6205684_m.fxe"},
	{0xF6EE9FEA14B28550ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\_textures\\ben_boosterprops_n.dds"},
	{0xF6EEBAC7A1F5E8EAULL, "water_water_p3.psh.fxo"},
	{0xF6EEC119E1985E71ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_oldmetalb_n.dds"},
	{0xF6EFC6621B7B9945ULL, "gen00fl_anders6183797.cl"},
	{0xF6EFCAA8527BD128ULL, "hlac_lt_0_lowtown_docks_a3_n_131072_0z.dds"},
	{0xF6EFE184D44E5EA2ULL, "vxlt_undercityencounter_uj_0_1.vlm"},
	{0xF6F005A22A50448DULL, "mer231_brekker6075170.cl"},
	{0xF6F0302E5BBB0C38ULL, "hlac_lt_warehouse_-65535_1y.dds"},
	{0xF6F060F7139A6466ULL, "vxw_woundedcoast_182_0_0.vlm"},
	{0xF6F081C04C5796ADULL, "hlac_lt_gallowsprison_n_-1_15x.dds"},
	{0xF6F0ADE450956045ULL, "vxw_bonepitmining_a2_2ad_0_0.vlm"},
	{0xF6F0CE1DE21319CCULL, "ht_0_hightown_a2_51_0_ll.mmh"},
	{0xF6F0D6FB841E6EF8ULL, "vxw_bonepitdestroyed_a3_25o_0_0.vlm"},
	{0xF6F1804AA9B0DAE3ULL, "wm_ico_backalley.dds"},
	{0xF6F1C9031B6467A7ULL, "vxlt_undercitydungeon_1nu_0_0.vlm"},
	{0xF6F1FE0228E1B897ULL, "lgt200_eau_tome6189899.cl"},
	{0xF6F225CDEAEE5269ULL, "6147591_m.fxe"},
	{0xF6F29503872652DEULL, "vxw_bonepit_e1_0_0.vlm"},
	{0xF6F2FA2656FE6C07ULL, "vxw_woundedcoast_n_0_0.vlm"},
	{0xF6F30FD89E80F6ADULL, "wm_ico_genericpin_roll.dds"},
	{0xF6F32219A6DC206EULL, "600211570.ogg"},
	{0xF6F333025FE753B9ULL, "259498725.ogg"},
	{0xF6F3363691EEE7CDULL, "lt_0_lowtown_docks_a3_255_0.mmh"},
	{0xF6F33A3691EEEE1EULL, "lt_0_lowtown_docks_a3_255_0.mao"},
	{0xF6F3483691EF0653ULL, "lt_0_lowtown_docks_a3_255_0.msh"},
	{0xF6F3DCFB64D6B1CBULL, "395391364.ogg"},
	{0xF6F43D5E1710AA75ULL, "zz_followers_banter6157645.cl"},
	{0xF6F55CB1D55AA666ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_wng_hdra_0n.dds"},
	{0xF6F5C1EAB3443468ULL, "art\\characters\\creatures\\highdragon\\animation\\reaction\\deathblowmountv1.ani"},
	{0xF6F63A0AB34C8B24ULL, "110509811.ogg"},
	{0xF6F7906BBCCA7604ULL, "da2terrain_terrain_tools_p2.psh.fxo.dx11"},
	{0xF6F7A8556DCB318DULL, "w_woundedcoast_30_0.rml"},
	{0xF6F7D3B669037310ULL, "6144057_f.fxe"},
	{0xF6F81936AF31796AULL, "art\\characters\\creatures\\crowds\\humanmale\\textures\\prp_lowtownjimmy_s.dds"},
	{0xF6F8DEFA87F6F83DULL, "vxlt_blackemporium_vd_0_0.vlm"},
	{0xF6F8E79DB98641BCULL, "174910833.ogg"},
	{0xF6F93786683253FCULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0ww_pillar.phy"},
	{0xF6F9DBE621C45A8FULL, "hlac_lt_undercitydungeon_-1_12y.dds"},
	{0xF6FAB08395C759DEULL, "hlac_lt_gallowstemplar_2394_12y.dds"},
	{0xF6FAD8CB3E3634DBULL, "795202763.ogg"},
	{0xF6FB163A40EC7FF9ULL, "vxlt_hangedman_rp_0_0.vlm"},
	{0xF6FB2C69179768B3ULL, "llac_lt_0_lowtown_docks_131071_2.dds"},
	{0xF6FB2EFBDE10176EULL, "814108895.ogg"},
	{0xF6FB65B5D50CB851ULL, "vxht_mansion2_3f9_0_0.vlm"},
	{0xF6FBCE89DE709FB9ULL, "hlac_lt_0_lowtown_a2_n_c_1_6z.dds"},
	{0xF6FBFC2720B42BA3ULL, "gen_im_arm_cht_visa.uti"},
	{0xF6FC2717CC4678B4ULL, "vxlt_0_lowtown_docks_n_bde_0_0.vlm"},
	{0xF6FC2EF21717CE8FULL, "llac_w_sundermount_-327676_0.dds"},
	{0xF6FC5A536FF259CFULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_hv_dmg.gad"},
	{0xF6FC6039751E4694ULL, "vxlt_0_lowtown_n_e0w_0_0.vlm"},
	{0xF6FCB730929F2E7CULL, "vxw_encounter2_17_0_0.vlm"},
	{0xF6FCBB69EFC27E89ULL, "vxlt_hangedman_n_22y_0_0.vlm"},
	{0xF6FCFD70F7016435ULL, "vxlt_0_lowtown_docks_a3_cwt_0_0.vlm"},
	{0xF6FCFE1947DB492EULL, "w_bonepitdestroyed_a3_125_0.mmh"},
	{0xF6FD0A1947DB5DC5ULL, "w_bonepitdestroyed_a3_125_0.mao"},
	{0xF6FD181947DB751CULL, "w_bonepitdestroyed_a3_125_0.msh"},
	{0xF6FD2D835E88DD51ULL, "vxlt_gallowscourtyard_n_7h5_0_0.vlm"},
	{0xF6FD33AB6CE3AEFCULL, "52728893.ogg"},
	{0xF6FD7F5D0DFDEA56ULL, "vxw_blightlands_py_0_0.vlm"},
	{0xF6FE32D1047C96D7ULL, "vxht_keep_644_0_0.vlm"},
	{0xF6FE5CEB6F848EBEULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medc_0s.dds"},
	{0xF6FE68086591F887ULL, "138024416.ogg"},
	{0xF6FEF97EDA5F54CEULL, "6170899_m.fxe"},
	{0xF6FF16AE1B8C0EDAULL, "qun110_petrice6037687.cl"},
	{0xF6FF48D876148778ULL, "vxw_bonepitdestroyed_a3_20l_0_0.vlm"},
	{0xF6FF88880D7E33B3ULL, "vxlt_blackemporium_8x_0_0.vlm"},
	{0xF70022E5FA6CCDFCULL, "6104429_m.fxe"},
	{0xF70025EF3B3FE397ULL, "vxht_brothel_n_6oy_0_0.vlm"},
	{0xF70048AD5C453BA8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\ambient\\dialogue\\drunk_wobble.gad"},
	{0xF70097AADCE46AE9ULL, "6140491_m.fxe"},
	{0xF700A0CC67838884ULL, "art\\characters\\playercharacter\\humanmale\\cth_nobb.phy"},
	{0xF700AD1F663AD30BULL, "zz_ev_placeable.ncs"},
	{0xF700C015BE0A57D1ULL, "194601947.ogg"},
	{0xF70110A638CFAE16ULL, "w_bonepitdestroyed_a3_113_0.mao"},
	{0xF7011CA638CFC2B5ULL, "w_bonepitdestroyed_a3_113_0.mmh"},
	{0xF7011EA638CFC62BULL, "w_bonepitdestroyed_a3_113_0.msh"},
	{0xF701F744EB9AAD32ULL, "vxw_woundedcoast_hg_0_0.vlm"},
	{0xF701F7B196A14B8AULL, "6082202_m.fxe"},
	{0xF701FC5427955C52ULL, "6109424_m.fxe"},
	{0xF7026AFB53592368ULL, "cheese_sa.dds"},
	{0xF702B43D538C95D7ULL, "hlac_lt_0_lowtown_a2_n_131072_1x.dds"},
	{0xF702EF2638C8C037ULL, "6101270_m.fxe"},
	{0xF7033E573A3C8017ULL, "vxw_blightlands_9s_0_0.vlm"},
	{0xF7036F536FF89C7CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_hv_dmg.evt"},
	{0xF704272B389F55C5ULL, "6013849_m.fxe"},
	{0xF7058F7653628E11ULL, "997795173.ogg"},
	{0xF705D2A0840C7FF6ULL, "w_bonepit_206_0.msh"},
	{0xF705DCA0840C90F4ULL, "w_bonepit_206_0.mmh"},
	{0xF705E0A0840C9707ULL, "w_bonepit_206_0.mao"},
	{0xF70624A680C0A48BULL, "mcr310_argument6075137.cl"},
	{0xF707294D049EADCBULL, "vxw_blightlands_6j_0_0.vlm"},
	{0xF70909465ED3CBC3ULL, "6070906_f.fxe"},
	{0xF7094940060A4B0EULL, "vxw_bonepitdestroyed_a3_l6_0_0.vlm"},
	{0xF709D770CB77771CULL, "6107987_m.fxe"},
	{0xF709EE61EB7904CCULL, "6090874_m.fxe"},
	{0xF709F36389B86AA7ULL, "957164869.ogg"},
	{0xF70A4EAC5852E1CDULL, "vxht_0_hightown_n_c7i_0_0.vlm"},
	{0xF70A6D9032AA4F2EULL, "vxlt_hangedman_n_su_0_0.vlm"},
	{0xF70B3109248A0628ULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_1y.dds"},
	{0xF70B6806669387F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1.ani"},
	{0xF70B7C69AED97B7CULL, "vxw_blightlands_f9_0_0.vlm"},
	{0xF70B7D9CB7C586C9ULL, "lt_0_lowtown_docks_n_234_0.mao"},
	{0xF70B899CB7C59B62ULL, "lt_0_lowtown_docks_n_234_0.mmh"},
	{0xF70B8B9CB7C59E18ULL, "lt_0_lowtown_docks_n_234_0.msh"},
	{0xF70B9E580ED025AEULL, "one300_zevran6076988.cl"},
	{0xF70BA5BA614DB749ULL, "isa311_isabela6095333.cl"},
	{0xF70BE9687A64A025ULL, "6101207_m.fxe"},
	{0xF70BF139FFF402D6ULL, "vxw_blightlands_uz_0_0.vlm"},
	{0xF70C8D089DA76A1DULL, "zz_dae_debug6029317.cl"},
	{0xF70CA4ADF672AF57ULL, "vxht_keep_617_0_0.vlm"},
	{0xF70CD1108585EC9FULL, "root_vengeance.dds"},
	{0xF70D02BE18470C46ULL, "217998503.ogg"},
	{0xF70D455B27A43295ULL, "6095902_m.fxe"},
	{0xF70D7D00FF9DCE06ULL, "572884974.ogg"},
	{0xF70DB2BC1708507EULL, "w_sundermount_11_0.rml"},
	{0xF70DCB06E3DFBD00ULL, "257851057.ogg"},
	{0xF70E08A92E91F7D8ULL, "6088141_m.fxe"},
	{0xF70E4767045FD1EBULL, "205155193.ogg"},
	{0xF70E97CC2D35FC17ULL, "vxht_brothel_n_6hx_0_0.vlm"},
	{0xF70F1E59BD439D85ULL, "6043872_f.fxe"},
	{0xF70F78BD6B1283DEULL, "vxw_bonepit_k3_0_0.vlm"},
	{0xF70F7F2851E1B16AULL, "196845287.ogg"},
	{0xF70F9C3419020111ULL, "296663380.ogg"},
	{0xF7100CC124D30DF9ULL, "gen_im_trs_gem_tigerseye_04.uti"},
	{0xF7104612A0A3220AULL, "vxw_sundermount_353_0_0.vlm"},
	{0xF710775370033395ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_hv_dmg.ani"},
	{0xF710BBC3B058F0C3ULL, "6081596_m.fxe"},
	{0xF71142CB267A253DULL, "mcr370_meredith6136169.cl"},
	{0xF71159531996139AULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hgrass2_n.dds"},
	{0xF7117B3175060510ULL, "vxlt_darktown_2x8_0_0.vlm"},
	{0xF711FFB8C57C01E5ULL, "vxlt_0_lowtown_docks_n_aq4_0_0.vlm"},
	{0xF7121B6BD59FEEE2ULL, "947593410.ogg"},
	{0xF7122295005055E1ULL, "6115875_m.fxe"},
	{0xF7122415EC197D0BULL, "1060023591.ogg"},
	{0xF712590D421EE5F8ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_framea.phy"},
	{0xF712847F4C5D4209ULL, "803871392.ogg"},
	{0xF712A72292646871ULL, "6054237_m.fxe"},
	{0xF712B7FDA63A6E78ULL, "llac_w_sundermount_196605_0.dds"},
	{0xF71310A26A41143EULL, "vxlt_0_lowtown_a2_hdg_0_0.vlm"},
	{0xF7132FC82EC08C56ULL, "685903700.ogg"},
	{0xF71457F119A59395ULL, "vxlt_0_lowtown_a2_n_erh_0_0.vlm"},
	{0xF7152318F0C48DF2ULL, "lgt110_gxa_vgk_silver_order6117007.cl"},
	{0xF7153F9AEDD7E13FULL, "darktown_male_sickly_sta6175326.cl"},
	{0xF715480D59041A6DULL, "6086268_f.fxe"},
	{0xF715B1FF1F340AD1ULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_angler.gad"},
	{0xF71689232C6A74C3ULL, "fex_bd_a_bs_wl.mmh"},
	{0xF71697232C6A8C1DULL, "fex_bd_a_bs_wl.msh"},
	{0xF71729BD690321F6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_swordc_n.dds"},
	{0xF717775865F3F2CAULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_dsgaa_0n.dds"},
	{0xF7186006669E042AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1.evt"},
	{0xF718603EDD6F1D4CULL, "6095774_m.fxe"},
	{0xF718719973E79AD9ULL, "lt_0_lowtown_docks_33_0.anb"},
	{0xF7189818E69EBDBCULL, "gen00fl_aveline6053603.cl"},
	{0xF71A471A5CA6DD9CULL, "702480625.ogg"},
	{0xF71A59A37332B6ACULL, "25427802.ogg"},
	{0xF71AC1682E0BD574ULL, "6072889_m.fxe"},
	{0xF71B19EF9E9677AFULL, "506000097.ogg"},
	{0xF71BB2FF1F38787FULL, "art\\characters\\creatures\\hurlock\\animation\\additive\\hurlock_jaw_angler.ani"},
	{0xF71BF3F7F1B18C02ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.dura.gad"},
	{0xF71C01E6D47C2E48ULL, "392576989.ogg"},
	{0xF71C7A8620881058ULL, "6083041_m.fxe"},
	{0xF71C866B9B1720FEULL, "vxlt_0_lowtown_n_eou_0_1.vlm"},
	{0xF71D2C726DB70DE0ULL, "522646997.ogg"},
	{0xF71E475C743E41C6ULL, "711640535.ogg"},
	{0xF71ED3001595E4FBULL, "75802230.ogg"},
	{0xF71EDAFECF86DE32ULL, "vxw_woundedcoast_ba_0_0.vlm"},
	{0xF71F2E8D00862085ULL, "vxw_bonepitdestroyed_a3_25v_0_0.vlm"},
	{0xF71F302B61FD3E1DULL, "6083487_f.fxe"},
	{0xF71F36E4B1C41C9EULL, "mcr310cr_apostate.utc"},
	{0xF71F3F0666A3EA9DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_medium_trim1.gad"},
	{0xF71FB8EC12CAE423ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammerb.mao"},
	{0xF71FBB1EE9E92A5CULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_boo_lgte_0n.dds"},
	{0xF71FBCEC12CAEB70ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammerb.mmh"},
	{0xF71FCAEC12CB028AULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammerb.msh"},
	{0xF72020A95A250685ULL, "art\\vfx\\combat\\vfx_glb_crust_c.anb"},
	{0xF7209B8A196D9734ULL, "art\\characters\\weapons\\longswords\\w_lsw_011a.phy"},
	{0xF7211DDDE6EB6821ULL, "lt_0_lowtown_n_35_0_ll.mmh"},
	{0xF7214451306A7117ULL, "vxw_sundermount_26n_0_0.vlm"},
	{0xF722C9CE0B9BAC15ULL, "vxlt_0_lowtown_a2_n_eww_0_0.vlm"},
	{0xF722FFF8A31153FEULL, "6110871_m.fxe"},
	{0xF72316F7F1B7E6F9ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.dura.evt"},
	{0xF723B2FA356422F8ULL, "vxlt_hangedman_1c3_0_0.vlm"},
	{0xF723F4DC95C0F65AULL, "150242350.ogg"},
	{0xF72429616BBB1FB7ULL, "w_encounter1_21_0.rml"},
	{0xF7248F6D74544221ULL, "lgt160cr_wk_soph_hunter.utc"},
	{0xF7251DC6F60E1DACULL, "870621626.ogg"},
	{0xF7251EBD59AA459FULL, "vxht_chantry_n_36d_0_0.vlm"},
	{0xF72552C30AA725ACULL, "vxht_0_hightown_a2_3xr_0_0.vlm"},
	{0xF72552F968E14779ULL, "hlac_lt_gallowscourtyard_-2_2z.dds"},
	{0xF7256CF6480621C5ULL, "hlac_ca_cavern1_haunted_24423_13x.dds"},
	{0xF725F94569163F4AULL, "mrl000_marethari6129450.cl"},
	{0xF7262962C90E1F59ULL, "vxdr_primevaltomb_7um_0_0.vlm"},
	{0xF7265417214EBA5CULL, "379610945.ogg"},
	{0xF7267D2E4502B6AFULL, "hlac_lt_warehouse_n_1_4x.dds"},
	{0xF726A38744AA0F89ULL, "vxht_bartrand_n_4nv_0_0.vlm"},
	{0xF727092A6EA79373ULL, "6026955_m.fxe"},
	{0xF7271570763E33BBULL, "6128979_f.fxe"},
	{0xF72746A84DED3042ULL, "hlac_lt_0_lowtown_a2_n_65537_6z.dds"},
	{0xF72766990E8AE3A0ULL, "hlac_ht_0_hightown_n_-131074_0y.dds"},
	{0xF7285825E5372B08ULL, "6110972_m.fxe"},
	{0xF7288B55ADFBD129ULL, "dae000_gen_turn_in_16198705.cl"},
	{0xF728B6AE6062460AULL, "6174929_m.fxe"},
	{0xF728BD314B9E87CFULL, "114240168.ogg"},
	{0xF7292687E8C00294ULL, "zz_mag_debug6024442.cl"},
	{0xF72980AC3BDD9747ULL, "715327053.ogg"},
	{0xF729A1333666AB8FULL, "art\\vfx\\textures\\vfx_smoke_puffy_cs.dds"},
	{0xF729A7032864F781ULL, "ss_hawkecu1.cut"},
	{0xF729DF5D8463BD11ULL, "6100266_m.fxe"},
	{0xF729E9F7C1EF1F10ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4x.msh"},
	{0xF729F7F7C1EF372AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4x.mmh"},
	{0xF729FBF7C1EF3E61ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0cc_basic4x.mao"},
	{0xF72A069910835EBCULL, "follower_banter6143098.cl"},
	{0xF72A124326947DD0ULL, "vxht_keep_588_0_0.vlm"},
	{0xF72A2264A422D1ABULL, "454627100.ogg"},
	{0xF72A6571C502DCD3ULL, "dae100_bran6118336.cl"},
	{0xF72A788A1975B031ULL, "art\\characters\\weapons\\longswords\\w_lsw_011a.msh"},
	{0xF72A85B58A407378ULL, "w_bonepitdestroyed_a3_106_0.phy"},
	{0xF72A868A1975C84CULL, "art\\characters\\weapons\\longswords\\w_lsw_011a.mao"},
	{0xF72A928A1975DCEFULL, "art\\characters\\weapons\\longswords\\w_lsw_011a.mmh"},
	{0xF72AC4C886FAA59AULL, "vxlt_undercityserialkiller_15d_0_0.vlm"},
	{0xF72AE573DCF178E1ULL, "423451693.ogg"},
	{0xF72B9FE29F7F7271ULL, "vxca_cavern1_haunted_j1z_0_2.vlm"},
	{0xF72BFE6E5E9C27A9ULL, "37931693.ogg"},
	{0xF72C4F15B2CC031EULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv2.evt"},
	{0xF72C57F5F8D84E24ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\swagger_pose.gad"},
	{0xF72C73C7F9F482E0ULL, "6155535_m.fxe"},
	{0xF72C799B3F5BB502ULL, "gen_im_trs_bks_risekirk.uti"},
	{0xF72CC4E88763BF6FULL, "6201395_f.fxe"},
	{0xF72D20D09AC360CCULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandcrackr_1.mao"},
	{0xF72E2C17E3F1C25BULL, "6125579_m.fxe"},
	{0xF72E7C225E0BAA7EULL, "6041514_f.fxe"},
	{0xF72E9FF1AE889310ULL, "6136688_m.fxe"},
	{0xF72EB17574043C95ULL, "dae000_gen_turn_in_16178394.cl"},
	{0xF72EC126D668DE5BULL, "hlac_dr_primevalentrance_1726_7x.dds"},
	{0xF72F7D956FE372CFULL, "vxlt_0_lowtown_a2_gzc_0_0.vlm"},
	{0xF72F9E15AF7E01A4ULL, "w_bonepitmining_a2_153_0.phy"},
	{0xF72FDFC655A5EBCCULL, "vxdr_0_deeproads_d3k_0_0.vlm"},
	{0xF7300E3991D13E01ULL, "6160875_m.fxe"},
	{0xF730415194D86307ULL, "lt_0_lowtown_a2_21_0_ll.mmh"},
	{0xF730EAF037C57745ULL, "mcr371_alain6087140.cl"},
	{0xF7311EF7F1C431E8ULL, "art\\vfx\\combat\\vfx_d_arc_immune_cast_c.dura.ani"},
	{0xF7313BAAFD91C495ULL, "vxlt_0_lowtown_a2_n_exv_0_0.vlm"},
	{0xF73160AF78CDEEC0ULL, "fen210_dying_magister6108615.cl"},
	{0xF731C68E21FAE07FULL, "llac_w_bonepitdestroyed_a3_65541_0.dds"},
	{0xF731FA03A2BBCFDBULL, "vxlt_0_lowtown_docks_a3_cyx_0_0.vlm"},
	{0xF7323383BF611B2EULL, "art\\vfx\\textures\\vfx_fire_ring_fb.dds"},
	{0xF7326615B2D09649ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv2.gad"},
	{0xF73366F5F8DE8602ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\swagger_pose.ani"},
	{0xF733F2D149D1CD5CULL, "6217510_m.fxe"},
	{0xF734136B098F45F1ULL, "71617489.ogg"},
	{0xF734E7740FCAAA63ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\additive.evt"},
	{0xF73536B4016DBE41ULL, "870749811.ogg"},
	{0xF7357227BF0BEB76ULL, "art\\characters\\playercharacter\\humanmale\\arm_masb.msh"},
	{0xF7357C27BF0BFC74ULL, "art\\characters\\playercharacter\\humanmale\\arm_masb.mmh"},
	{0xF7358027BF0C0287ULL, "art\\characters\\playercharacter\\humanmale\\arm_masb.mao"},
	{0xF735F3AC2B7AB4D6ULL, "85702381.ogg"},
	{0xF736A61153C87EDBULL, "character_character_p149.psh.fxo.dx11.assembly"},
	{0xF736CB57BE97C143ULL, "6043508_f.fxe"},
	{0xF73702AB40ABB52FULL, "6058284_m.fxe"},
	{0xF7371408A5F7463EULL, "vxht_keep_n_57x_0_0.vlm"},
	{0xF7376461E6DEB565ULL, "vxlt_0_lowtown_a2_n_ei9_0_0.vlm"},
	{0xF738190553D07E3DULL, "522475769.ogg"},
	{0xF73844EACC7C4BE5ULL, "347826650.ogg"},
	{0xF738D300FCB823B5ULL, "vxca_caverns2_7c9_0_0.vlm"},
	{0xF738E53DB9407D10ULL, "dae000ip_crowd_lm3_talk14.utp"},
	{0xF7393715B2D66497ULL, "art\\characters\\creatures\\highdragon\\animation\\combat\\frontleftlegattackv2.ani"},
	{0xF739A6BC172D76F6ULL, "w_sundermount_11_0.mao"},
	{0xF739B2BC172D8A95ULL, "w_sundermount_11_0.mmh"},
	{0xF739B4BC172D8E0BULL, "w_sundermount_11_0.msh"},
	{0xF739D238E91D7FD0ULL, "vxht_0_hightown_a3_auz_0_0.vlm"},
	{0xF739F705E4801952ULL, "195519408.ogg"},
	{0xF73A5E633D32A7A2ULL, "vxht_chantry_n_2dz_0_0.vlm"},
	{0xF73AABB2E4B99B5FULL, "dae000_bartrand6206930.cl"},
	{0xF73B120C130CEC5CULL, "hlac_ht_0_hightown_-131073_0y.dds"},
	{0xF73B165ADFD515FBULL, "lt_0_lowtown_docks_a3_n_238_0.msh"},
	{0xF73B285ADFD53446ULL, "lt_0_lowtown_docks_a3_n_238_0.mao"},
	{0xF73B345ADFD548E5ULL, "lt_0_lowtown_docks_a3_n_238_0.mmh"},
	{0xF73B44DF81FFF540ULL, "qun100pt_blackform.ncs"},
	{0xF73B87FA05A4B55DULL, "233535882.ogg"},
	{0xF73B9DFF929A4FBFULL, "322498171.ogg"},
	{0xF73BA3AC81B1DE0EULL, "6116150_m.fxe"},
	{0xF73BBEB8B3730F32ULL, "vxw_woundedcoast_tc_0_0.vlm"},
	{0xF73BC0740FD086E4ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\additive.gad"},
	{0xF73BD585B920C953ULL, "hlac_lt_0_lowtown_a3_n_c_0_8z.dds"},
	{0xF73C2895641C78DEULL, "617072351.ogg"},
	{0xF73C7A70C527E8E2ULL, "mrl310_party_comments6142752.cl"},
	{0xF73C91C733CBEC52ULL, "mrl000_marethari6069729.cl"},
	{0xF73D00B2363D273EULL, "vxdr_slavers_11x_0_0.vlm"},
	{0xF73D0954CCA62C57ULL, "278233660.ogg"},
	{0xF73DBBDB1124C337ULL, "hlac_ht_0_hightown_a2_131072_0z.dds"},
	{0xF73DD2EE507E114AULL, "6191274_m.fxe"},
	{0xF73E661C3E5F2F83ULL, "6142312_m.fxe"},
	{0xF73E801683C75681ULL, "6165775_m.fxe"},
	{0xF73F937800CE9B87ULL, "vxlt_0_lowtown_a2_n_elj_0_2.vlm"},
	{0xF740924C593DD684ULL, "493910550.ogg"},
	{0xF7409D1ED9594AA1ULL, "vxlt_gallowstemplar_n_5ir_0_0.vlm"},
	{0xF7426847010ABE42ULL, "vxca_caverns1_mines_hl2_0_0.vlm"},
	{0xF742CF740FD6BEC2ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\elementalreacts\\physical\\additive.ani"},
	{0xF7439C687DE88F83ULL, "vxlt_undercitydungeon_ye_0_0.vlm"},
	{0xF743EF1AD3A1F532ULL, "241538720.ogg"},
	{0xF7440B3C49F64AAFULL, "6062090_m.fxe"},
	{0xF744397E3C878A63ULL, "one200_gold_followers6117108.cl"},
	{0xF744DA4EEEBEAB30ULL, "610423734.ogg"},
	{0xF744DC0B134EA8F8ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robc_0t.dds"},
	{0xF744ED789C902541ULL, "lgt300cr_gang_3_ht.utc"},
	{0xF745100DDABD7564ULL, "fen311cr_danarius.utc"},
	{0xF74516CCF16C2A55ULL, "vxlt_shop_rs_0_0.vlm"},
	{0xF745454A8C626757ULL, "dis_cu_alph.dds"},
	{0xF74546F440BD9941ULL, "6110858_m.fxe"},
	{0xF74553829C1916F1ULL, "vxlt_gallowstemplar_f_5t1_0_0.vlm"},
	{0xF745748832A1CDAFULL, "6055584_m.fxe"},
	{0xF745D63ED8D4CDE5ULL, "vxlt_0_lowtown_a2_n_er8_0_0.vlm"},
	{0xF74614F98869BC1DULL, "dae000_sandal_st.stg"},
	{0xF74671A4A7AB14C2ULL, "ave200_donnic6128948.cl"},
	{0xF7468F118F588151ULL, "58600815.ogg"},
	{0xF746C3A700CF68D3ULL, "hlac_lt_0_lowtown_a2_n_c_-3_5z.dds"},
	{0xF746DEAC860D6496ULL, "6095888_m.fxe"},
	{0xF746F61D4FE5846FULL, "vxlt_0_lowtown_fff_0_0.vlm"},
	{0xF747065AFE549827ULL, "6105744_m.fxe"},
	{0xF7473AE5F49925E1ULL, "650348618.ogg"},
	{0xF74821A95A465AC6ULL, "art\\vfx\\combat\\vfx_glb_crust_c.mmh"},
	{0xF7482C4768B809A5ULL, "fen210_orana6109724.cl"},
	{0xF7488A803DBF5ED5ULL, "689277684.ogg"},
	{0xF748BFE087A2AC3CULL, "dae351ip_at_sundermount_b.utp"},
	{0xF748FB8BB70A242FULL, "215126475.ogg"},
	{0xF74955B51F13A9F1ULL, "art\\levels\\races\\natural\\blightlands\\terrainmats\\ter_blackdust_n.dds"},
	{0xF74994AB72833965ULL, "566738994.ogg"},
	{0xF74A3095A56927B2ULL, "vxw_woundedcoast_ql_0_0.vlm"},
	{0xF74A477899AEB9EFULL, "rdr111ip_martin_door.utp"},
	{0xF74A6128239D1904ULL, "vxlt_0_lowtown_docks_a3_n_cof_0_0.vlm"},
	{0xF74A9D1AC978DF40ULL, "vxht_mansion2b_3df_0_0.vlm"},
	{0xF74AAB860ED36D54ULL, "731974898.ogg"},
	{0xF74B7625A0A4F86FULL, "vxw_encounter1_e5_0_0.vlm"},
	{0xF74B7EAF997E8ED2ULL, "lgt320_mis_resp_swordhunter6150331.cl"},
	{0xF74BF6D2FF85BDDEULL, "vxlt_warehouse_n_10g_0_0.vlm"},
	{0xF74CCE2C27BCB2F4ULL, "6043067_m.fxe"},
	{0xF74D4D3F97463769ULL, "qun160_vashothlead6053857.cl"},
	{0xF74D4DC17A0B2D32ULL, "vxw_bonepit_1yz_0_0.vlm"},
	{0xF74D8839B5B3B893ULL, "6135934_m.fxe"},
	{0xF74DBB6CB53E50FFULL, "w_bonepitmining_a2_22_0_ll.mmh"},
	{0xF74E1346F743F63BULL, "6054028_m.fxe"},
	{0xF74E2B3B8AB2E308ULL, "moduleoverrides.gda"},
	{0xF74E58E87BB92CDFULL, "art\\characters\\weapons\\zzdao\\textures\\w_lsw_ls01a_0d.dds"},
	{0xF74EE70B0E487BCBULL, "186542377.ogg"},
	{0xF74EF869F8592F8EULL, "6016148_m.fxe"},
	{0xF74F14D76506878FULL, "6144066_m.fxe"},
	{0xF74FA84458121B92ULL, "500061965.ogg"},
	{0xF74FBE7D57476650ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masb.mao"},
	{0xF74FC27D57476D63ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masb.mmh"},
	{0xF74FD07D574784BDULL, "art\\characters\\playercharacter\\humanmale\\bdy_masb.msh"},
	{0xF74FF4B43F956493ULL, "6129672_m.fxe"},
	{0xF7500BD62BC8CD5BULL, "gst_female_neutral_2.gda"},
	{0xF7502F84D82C485DULL, "llac_lt_warehouse_-65536_2.dds"},
	{0xF7506EA9D4E0E02FULL, "vxca_caverns1_mines_gh8_0_0.vlm"},
	{0xF750B438F31E53E8ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_twrc_0s.dds"},
	{0xF751406331F1885BULL, "hlac_lt_gallowstemplar_2388_6z.dds"},
	{0xF75188392B2B173FULL, "w_encounter3_251_0.msh"},
	{0xF7518E392B2B21E1ULL, "w_encounter3_251_0.mmh"},
	{0xF7519A392B2B3582ULL, "w_encounter3_251_0.mao"},
	{0xF751D12EB761F53CULL, "vxht_0_hightown_a2_c72_0_1.vlm"},
	{0xF751DD9DB3F6608CULL, "vxca_caverns2prc_aa1_0_0.vlm"},
	{0xF7527299741914D9ULL, "lt_0_lowtown_docks_33_0.rml"},
	{0xF753131251B4F029ULL, "768349730.ogg"},
	{0xF7540C447044C340ULL, "fen210_fenris6108724.cl"},
	{0xF7543BAD578DE330ULL, "gen00fl_bethany6102773.cl"},
	{0xF75502557E136B0BULL, "gen00fl_isabela6085688.cl"},
	{0xF7558315A7C3263DULL, "gen00fl_carver6103010.cl"},
	{0xF755DBC5E25B0261ULL, "vxlt_darktown_60i_0_0.vlm"},
	{0xF756EEFEA12C0950ULL, "w_bonepitmining_a2_18_0.rml"},
	{0xF7579AD55BBD3967ULL, "hlac_lt_gallowscourtyard_-65537_1y.dds"},
	{0xF757B1247471F442ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avea.phy"},
	{0xF757C3F1E829FB80ULL, "zz_approval_debug6133715.cl"},
	{0xF7580C59EAA4DA19ULL, "vxca_caverns1_mines_gkr_0_1.vlm"},
	{0xF758109E19262E37ULL, "vxw_bonepitmining_a2_gd_0_0.vlm"},
	{0xF758927F30544B9CULL, "ave300ar_militia_rally.ncs"},
	{0xF75899B17596BEC9ULL, "6065952_m.fxe"},
	{0xF7591729BF290C82ULL, "869901081.ogg"},
	{0xF7597395084E3B13ULL, "hlac_lt_0_lowtown_a3_n_c_131071_4y.dds"},
	{0xF7597B7B6AC2DE28ULL, "6177180_m.fxe"},
	{0xF759A9FFF1A474ECULL, "73377991.ogg"},
	{0xF759E802929B10EFULL, "vxw_encounter1_j4_0_0.vlm"},
	{0xF75A0C1A0B86BAE2ULL, "hlac_lt_0_lowtown_a2_-131072_7y.dds"},
	{0xF75A3372EA35C525ULL, "vxlt_gallowscourtyard_n_6ys_0_0.vlm"},
	{0xF75ADDECAF86E4FDULL, "vxht_keep_n_4na_0_0.vlm"},
	{0xF75AEE89016B180AULL, "522441710.ogg"},
	{0xF75B0A931E6D18BDULL, "460957800.ogg"},
	{0xF75B547899BD429AULL, "rdr111ip_martin_door.ncs"},
	{0xF75C8F84CCE58082ULL, "dae111ar_gamlens_house.ncs"},
	{0xF75CD1AF4F8C2B8BULL, "vxht_mansion2b_1k6_0_0.vlm"},
	{0xF75D3209C0D1A619ULL, "6162020_m.fxe"},
	{0xF75D9710D6543E49ULL, "lowtown_fem_prostitute_sta6174627.cl"},
	{0xF75DFE8F2E85F104ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_masd_0n.dds"},
	{0xF75E2AA91AE02EF7ULL, "6118948_m.fxe"},
	{0xF75E3102FDDEE610ULL, "429905115.ogg"},
	{0xF75E3164E4328B15ULL, "6095923_f.fxe"},
	{0xF75EB63D6FFCCB5EULL, "vxlt_0_lowtown_g93_0_0.vlm"},
	{0xF75F0B37257811E4ULL, "600035976.ogg"},
	{0xF75F4B49D6C07B00ULL, "733944536.ogg"},
	{0xF75F647BA3AFAE48ULL, "vxlt_darktown_60l_0_0.vlm"},
	{0xF75F8231DCB21AC0ULL, "face0_dy.psh.fxo"},
	{0xF75FC02A1520DC7FULL, "693415164.ogg"},
	{0xF75FF16A4CE1CB94ULL, "6076208_f.fxe"},
	{0xF75FF9A04F43A2B9ULL, "1037106762.ogg"},
	{0xF75FFF7BB83D533CULL, "vxlt_0_lowtown_a2_n_c_f5i_0_0.vlm"},
	{0xF760C37D5755E188ULL, "art\\characters\\playercharacter\\humanmale\\bdy_masb.phy"},
	{0xF760C3EC13028D8BULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammerb.phy"},
	{0xF760CEC8DC653142ULL, "art\\characters\\playercharacter\\qunarimale\\har_ars.mao"},
	{0xF760CF5F990F1B11ULL, "986165783.ogg"},
	{0xF760E70DE98A0EACULL, "one300_zevran6070896.cl"},
	{0xF7610F488C6612D9ULL, "505097175.ogg"},
	{0xF7612511224620B2ULL, "6142199_m.fxe"},
	{0xF7613DE7655278D5ULL, "t3_wdg_db1.tnt"},
	{0xF76142E55E3CF617ULL, "281719902.ogg"},
	{0xF7618AB920415C14ULL, "qcr000_arishok6059985.cl"},
	{0xF761B4D9FE1F8A1FULL, "853055489.ogg"},
	{0xF7623B2D8540941CULL, "lowtown_male_lowlife1_sta6164950.cl"},
	{0xF7624E4089371957ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.impa.gad"},
	{0xF762C2CAF8A00D0DULL, "192991260.ogg"},
	{0xF7637377E43FD6DDULL, "mrl150_merrill6173805.cl"},
	{0xF763F197D20A6240ULL, "6036728_m.fxe"},
	{0xF7645106D8062A26ULL, "6033163_m.fxe"},
	{0xF7658136730F4733ULL, "gen00fl_isabela6100059.cl"},
	{0xF7659FF2A1A234F4ULL, "ch2armtnt_ds2.psh.fxo"},
	{0xF76628C29E533F1FULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_lgta_0s.dds"},
	{0xF76676288BC2BB50ULL, "plt_cod_hst_mag_ambrosia.plo"},
	{0xF766D0FB11AD42BFULL, "220062362.ogg"},
	{0xF766DAD7B756C411ULL, "zz_ave_debug6064097.cl"},
	{0xF7671C5B21D8E6B4ULL, "250258600.ogg"},
	{0xF767D9103589A5EEULL, "6107170_m.fxe"},
	{0xF7689FDF1A4E3B35ULL, "vxlt_gamlenshouse_n_mz_0_0.vlm"},
	{0xF768D3C91781E564ULL, "vxht_0_hightown_a3_9v3_0_0.vlm"},
	{0xF769450937BBD4CFULL, "vxlt_0_lowtown_a2_gno_0_0.vlm"},
	{0xF7696340893D5BE4ULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.impa.evt"},
	{0xF7697DF6671CC062ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.mmh"},
	{0xF76999030DD6CDB1ULL, "vxlt_hangedman_lj_0_0.vlm"},
	{0xF769CECD0B087273ULL, "6148084_m.fxe"},
	{0xF76A01F48A2357C7ULL, "vxca_cave_2b2_0_0.vlm"},
	{0xF76A1381D60EC63EULL, "vxlt_0_lowtown_docks_ctq_0_0.vlm"},
	{0xF76A35BBE816B715ULL, "l_dae320ar_hightown_nt.bnk"},
	{0xF76AF6BA8E618E71ULL, "lt_hangedman_n_probe_3b1.dds"},
	{0xF76B09CE12ED003AULL, "vxlt_foundry_n_x6_0_0.vlm"},
	{0xF76B3E20A3A21C9DULL, "vxht_keep_3fy_0_0.vlm"},
	{0xF76B5AB86C4B4186ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightreappear.evt"},
	{0xF76B629E6BE29DB2ULL, "hlac_ht_0_hightown_a2_65537_1x.dds"},
	{0xF76B7AF9AEECA402ULL, "pro000_refugee_angry6052169.cl"},
	{0xF76BA3E6B2C7BE99ULL, "6167320_m.fxe"},
	{0xF76D460D426B708DULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_framea.msh"},
	{0xF76D4A0D2E2C8328ULL, "vxht_0_hightown_a3_b5y_0_0.vlm"},
	{0xF76D540D426B88C0ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_framea.mao"},
	{0xF76D580D426B8F13ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_framea.mmh"},
	{0xF76D91BC8B377F08ULL, "710086969.ogg"},
	{0xF76E44460BE70D23ULL, "hlac_lt_undercitydungeon_-1_35z.dds"},
	{0xF76E79DAA02D7007ULL, "6155886_m.fxe"},
	{0xF76E92027654AC51ULL, "836984056.ogg"},
	{0xF76EB14916D14301ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgte_0t.dds"},
	{0xF76F9547D531622AULL, "6115599_m.fxe"},
	{0xF77054E120A5F7FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_activate.gad"},
	{0xF7706C63C74E73D5ULL, "vxlt_shop_cp_0_0.vlm"},
	{0xF770A1019B152A97ULL, "806806335.ogg"},
	{0xF770B39224368471ULL, "6152799_m.fxe"},
	{0xF770C294EE723E0CULL, "dre000_bodahn.cnv"},
	{0xF770DBDC6D6B75D1ULL, "vxw_encounter3_a9_0_0.vlm"},
	{0xF77104803F0F1CB9ULL, "262153400.ogg"},
	{0xF7713FC8A1D35470ULL, "vxht_0_hightown_n_9zq_0_0.vlm"},
	{0xF77187EC7FD37C7CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_z_neg.ani"},
	{0xF771AA51EEBD2E25ULL, "vxdr_0_deeproads_f5j_0_0.vlm"},
	{0xF771B4E454981351ULL, "vxca_caverns1_ins_0_0.vlm"},
	{0xF771D0F6ED3D0251ULL, "6071037_f.fxe"},
	{0xF7722DAF967B7CF8ULL, "lgt311_shw_mer_mettin6156552.cl"},
	{0xF77271B86C5187D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightreappear.gad"},
	{0xF772B33E6820D952ULL, "gen00fl_aveline6139259.cl"},
	{0xF772B4A8C22F6280ULL, "6142139_m.fxe"},
	{0xF772DADD18C59142ULL, "hlac_lt_0_lowtown_docks_-65538_7y.dds"},
	{0xF773EA7E010D542FULL, "dae000_gen_turn_in_16178389.cl"},
	{0xF77414A46FFBF948ULL, "299340508.ogg"},
	{0xF774330A03459D2FULL, "hm_dalianshaw.mor"},
	{0xF7746BF86997F256ULL, "vxlt_darktown_466_0_0.vlm"},
	{0xF77483FB815533D7ULL, "vxlt_0_lowtown_a2_eqe_0_1.vlm"},
	{0xF77486BAE5DF96F2ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_crumble_1.mao"},
	{0xF774A494EE75F43CULL, "dre000_bodahn.bnk"},
	{0xF774B0330B5149B9ULL, "art\\characters\\creatures\\spider\\textures\\c_giaspidra_0t.dds"},
	{0xF77589C6CAC6723FULL, "hlac_lt_undercityserialkiller_-1_36z.dds"},
	{0xF77657E120AA6811ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_activate.evt"},
	{0xF7777B408949C1DDULL, "art\\vfx\\combat\\vfx_w_rev_aura_pain_loop_c.impa.ani"},
	{0xF777B00BEAA9D981ULL, "gen00fl_aveline6115490.cl"},
	{0xF7780BA2FF527535ULL, "mag100_karras6053194.cl"},
	{0xF77872B86C55F57FULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\boss\\meredith\\combat\\fightreappear.ani"},
	{0xF77894EC7FD9B0BEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\weapon_z_neg.gad"},
	{0xF778B4BAFC608686ULL, "genip_cin_viscount_decap.utp"},
	{0xF77947B34EE351B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_second_wind.ani"},
	{0xF77A2549E0D7FE9CULL, "22017851.ogg"},
	{0xF77A4783CA6C239FULL, "6066902_m.fxe"},
	{0xF77A73409F0BFD01ULL, "mag330_walter6045962.cl"},
	{0xF77A754A6FE104B3ULL, "6126753_f.fxe"},
	{0xF77AB3C39E1D8137ULL, "dot_01.mao"},
	{0xF77B1011D633B58FULL, "follower_banter6141552.cl"},
	{0xF77B1BC14C3AF43AULL, "mrl150_merrill6093694.cl"},
	{0xF77B68683AF25B24ULL, "192451622.ogg"},
	{0xF77B7121976CC403ULL, "hlac_lt_shop_-1_8y.dds"},
	{0xF77C5735E94DA141ULL, "8981931.ogg"},
	{0xF77C7E2E2C5DE544ULL, "llac_ca_caverns1_24426_7.dds"},
	{0xF77C8810F4556C8BULL, "hlac_lt_0_lowtown_docks_n_-131074_5x.dds"},
	{0xF77C8D9336BDAD7EULL, "hlac_lt_darktown_-1_68x.dds"},
	{0xF77C98247490E595ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avea.mmh"},
	{0xF77C9A247490E90BULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_avea.msh"},
	{0xF77CC48BEDDDF5EAULL, "qun110_petrice6018529.cl"},
	{0xF77CC6FFEA18635FULL, "dre000_sibling_has_blight6036455.cl"},
	{0xF77D5970C8DCA19EULL, "mcr310_sibling6076398.cl"},
	{0xF77D6555B3AB0F65ULL, "dre000_dougal6142211.cl"},
	{0xF77EFA1EA1761306ULL, "gen_im_trs_gen_sta_cld_r1.uti"},
	{0xF77F040973C17AA2ULL, "vxlt_0_lowtown_a2_n_c_er5_0_0.vlm"},
	{0xF77FCAB828859A21ULL, "vxlt_0_lowtown_docks_n_bc9_0_0.vlm"},
	{0xF77FF1CC1C8A2CC0ULL, "character_rendertodepth_p2.vsh.fxo"},
	{0xF78009D9A9209D66ULL, "585333870.ogg"},
	{0xF7801FE20DE1CADAULL, "zz_vrc_debug6064275.cl"},
	{0xF780E5A229D8DEDAULL, "lt_0_lowtown_docks_a3_n_174_0.mmh"},
	{0xF780E9A229D8E511ULL, "lt_0_lowtown_docks_a3_n_174_0.mao"},
	{0xF780F7A229D8FD20ULL, "lt_0_lowtown_docks_a3_n_174_0.msh"},
	{0xF78231869D609857ULL, "gen00fl_isabela6099339.cl"},
	{0xF782985F9D9979A2ULL, "hlac_lt_0_lowtown_docks_a3_-131073_0y.dds"},
	{0xF782E17ED611FDA1ULL, "mag101_alain6052993.cl"},
	{0xF7833B8FA3725E46ULL, "6093635_m.fxe"},
	{0xF7835080A2E24EB2ULL, "qcr220_grey_wardens6060432.cl"},
	{0xF78367E3C1A590DDULL, "vxht_mansion2_3gj_0_0.vlm"},
	{0xF783A2C32BCCD96BULL, "398035050.ogg"},
	{0xF783EB28C47DC439ULL, "158366675.ogg"},
	{0xF783EDC38B1855C1ULL, "18195715.ogg"},
	{0xF78415D339DE10B0ULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef05.mmh"},
	{0xF78423D339DE28CAULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef05.msh"},
	{0xF7846FE120B6CEE0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\assassin\\templarseeker\\ass_activate.ani"},
	{0xF7851B9E6473BB85ULL, "follower_banter6163558.cl"},
	{0xF78543C5E1033927ULL, "character_character_p28.psh.fxo.dx11.assembly"},
	{0xF78572B58A8CFE0DULL, "w_bonepitdestroyed_a3_106_0.msh"},
	{0xF7857D0CD5ACCB87ULL, "6103557_m.fxe"},
	{0xF78580B58A8D1640ULL, "w_bonepitdestroyed_a3_106_0.mao"},
	{0xF78584B58A8D1C93ULL, "w_bonepitdestroyed_a3_106_0.mmh"},
	{0xF785ABCDD02B6953ULL, "plt_zz_mcrpt_debug.nss"},
	{0xF785B782FB6E1664ULL, "vxht_0_hightown_a3_9h1_0_0.vlm"},
	{0xF7862721E57FEBC1ULL, "gen_im_wep_mag_sta_lore_01.uti"},
	{0xF786452DD65A8D6EULL, "mrl000_marethari6069862.cl"},
	{0xF7864A3C20C7CC61ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_basic.phy"},
	{0xF78677AD29178DC1ULL, "hlac_w_encounter2_camp_-131069_0z.dds"},
	{0xF7867CBCF1C2FEB1ULL, "vxlt_hangedman_rh_0_0.vlm"},
	{0xF786BBB618EEAD88ULL, "zz_dre_debug6056966.cl"},
	{0xF7872752570CD68FULL, "108763251.ogg"},
	{0xF7874FB34EEF9CA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_second_wind.evt"},
	{0xF787A704B7F31710ULL, "cd_one300_zevran.bnk"},
	{0xF788347980A5DC75ULL, "6079154_f.fxe"},
	{0xF78834A42E6A381BULL, "vxlt_0_lowtown_a2_d06_0_0.vlm"},
	{0xF788372644F091A1ULL, "follower_banter6149312.cl"},
	{0xF7883ABB7C388448ULL, "mag370_meredith6064862.cl"},
	{0xF7888F2BD9829B3DULL, "6104565_m.fxe"},
	{0xF7888F7C4E537934ULL, "llac_ca_caverns2prc_5802_40.dds"},
	{0xF788EE0E55D1E800ULL, "gen00fl_sebastian6178344.cl"},
	{0xF789117EC510F71EULL, "gen_im_wep_war_one_start.uti"},
	{0xF7892A1186458A4BULL, "vxw_blightlands_ce_0_0.vlm"},
	{0xF78940DF8232F20DULL, "hlac_ca_caverns1_mines_24427_0z.dds"},
	{0xF789AF40973098F1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_caress_face_turn_away.ani"},
	{0xF789FE25D9A52C91ULL, "6023969_m.fxe"},
	{0xF78A13872D434ABBULL, "dae000_player_generic_find6179053.cl"},
	{0xF78A25F038D7010FULL, "25432546.ogg"},
	{0xF78A383F9E49CBD9ULL, "art\\levels\\ztools\\default\\generic\\9ff_rock.msh"},
	{0xF78A423F9E49DCC7ULL, "art\\levels\\ztools\\default\\generic\\9ff_rock.mmh"},
	{0xF78A463F9E49E314ULL, "art\\levels\\ztools\\default\\generic\\9ff_rock.mao"},
	{0xF78A844615A1CC9BULL, "1061360200.ogg"},
	{0xF78AB405FCE2A283ULL, "hlac_lt_0_lowtown_-1_6z.dds"},
	{0xF78AEFB3D563B654ULL, "6113072_m.fxe"},
	{0xF78B4963393C038FULL, "vxca_caverns2_91r_0_1.vlm"},
	{0xF78B832BA200BA4AULL, "qun111_ketojan6040161.cl"},
	{0xF78B97C9A6670360ULL, "vxht_mansion_n_59d_0_0.vlm"},
	{0xF78C6E026A245B34ULL, "zz_followers_banter6157875.cl"},
	{0xF78C855B4E31C264ULL, "vxht_brothel_n_6ji_0_0.vlm"},
	{0xF78CB95D21C86939ULL, "llac_ht_0_hightown_a2_-3_0.dds"},
	{0xF78CBE1BEDB96A59ULL, "231398164.ogg"},
	{0xF78D6F00277FCBC0ULL, "llac_lt_0_lowtown_docks_n_-2_6.dds"},
	{0xF78D70B34EF440F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\shield and weapon\\sw_second_wind.gad"},
	{0xF78D7378F168E89EULL, "418720507.ogg"},
	{0xF78DAFF3DD2AE0BDULL, "6198480_f.fxe"},
	{0xF78DD8FBA36CA2CDULL, "hlac_w_encounter1_65537_0x.dds"},
	{0xF78E254FF5C2673FULL, "fen210_orana6109739.cl"},
	{0xF78EC080F80759F3ULL, "zz_vrc_debug6200214.cl"},
	{0xF78F4F22F1E4F70CULL, "6095143_m.fxe"},
	{0xF78F91566DFCAFE7ULL, "6075689_m.fxe"},
	{0xF79024F64B1FB0E3ULL, "tal_01.dds"},
	{0xF7907B27BF58A66FULL, "art\\characters\\playercharacter\\humanmale\\arm_masb.phy"},
	{0xF790B54C8851CB03ULL, "zz_seb_debug6215754.cl"},
	{0xF790BA409736C993ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_caress_face_turn_away.gad"},
	{0xF79167A68E298878ULL, "6078432_m.fxe"},
	{0xF79195FD26419C74ULL, "vxht_mansion2_2ax_0_0.vlm"},
	{0xF791BAE448E10BDFULL, "hlac_lt_0_lowtown_a3_n_c_131070_0z.dds"},
	{0xF791CAB95A2A8360ULL, "vxht_brothel_n_70b_0_0.vlm"},
	{0xF791FD96A3DFF093ULL, "284271380.ogg"},
	{0xF7927CF6673FC309ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_lit_p.anb"},
	{0xF792EB75558023D6ULL, "hlac_w_sundermount_196611_0y.dds"},
	{0xF793D0416022CF20ULL, "gen00fl_isabela6135638.cl"},
	{0xF79439B5C830C6CFULL, "ico_stamina_1.dds"},
	{0xF79443372BA6DE87ULL, "dae000cr_grand_cleric6182016.cl"},
	{0xF79467D3172ECF26ULL, "follower_banter6136449.cl"},
	{0xF794FACBD905C388ULL, "plt_ave200pt_night_terrors.plo"},
	{0xF794FC27F24DAD55ULL, "art\\vfx\\creatures\\vfx_c_hd_tswipe_c.anb"},
	{0xF7958418694A21D1ULL, "vxht_chantry_n_2t4_0_1.vlm"},
	{0xF795A0982C64764DULL, "vxca_caverns2_4ap_0_0.vlm"},
	{0xF795B148826AF2ECULL, "vxlt_0_lowtown_docks_n_b5o_0_1.vlm"},
	{0xF795F6D7AFA1AB51ULL, "vxlt_gallowscourtyard_65d_0_0.vlm"},
	{0xF7965BBBE83C329FULL, "l_dae320ar_hightown_nt.wbd"},
	{0xF7967A76A9D84A1DULL, "vxdr_primevaltomb_6p3_0_0.vlm"},
	{0xF7967EC9BC2D50BCULL, "358754866.ogg"},
	{0xF79691C62AECC25EULL, "265829635.ogg"},
	{0xF79742E5019710B2ULL, "6077192_m.fxe"},
	{0xF797770D1FB95D2DULL, "6081433_m.fxe"},
	{0xF7979BEE783BA2CBULL, "vxw_blightlands_dd_0_0.vlm"},
	{0xF797B740973CE3F8ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_caress_face_turn_away.evt"},
	{0xF79814C40B42780EULL, "mag131_taraneh6088749.cl"},
	{0xF79844F4E9600CA2ULL, "gen00fl_anders6125730.cl"},
	{0xF7986B81C8C04ACFULL, "dae000_bartrand6064055.cl"},
	{0xF79963F7E0329094ULL, "ser221_gamlen6131636.cl"},
	{0xF799F3B775D583A8ULL, "6105989_m.fxe"},
	{0xF79A2000D841F3E6ULL, "6038654_m.fxe"},
	{0xF79A2D3F76B6565FULL, "primitives_primitive_p1.psh.fxo.dx11.assembly"},
	{0xF79A8D2AFD33D8F3ULL, "hm_gencr_coterie_thug.mor"},
	{0xF79AB42104B4807DULL, "vxw_bonepitdestroyed_a3_1nd_0_0.vlm"},
	{0xF79AFCC9963DEBECULL, "vxca_caverns2prc_7o9_0_0.vlm"},
	{0xF79B9AB95DEF9373ULL, "gen00fl_merrill6130052.cl"},
	{0xF79C18F62588C2B2ULL, "one140cr_charade6137068.cl"},
	{0xF79C3C2463E5CAF2ULL, "vxht_keep_678_0_0.vlm"},
	{0xF79CA31AD9818A90ULL, "dae200_bran6097000.cl"},
	{0xF79CD5BE37B35B28ULL, "cs_pro000cs_opening_mage.wbd"},
	{0xF79D24595E5049BCULL, "hlac_lt_0_lowtown_65535_7z.dds"},
	{0xF79D39649805F6F2ULL, "6039980_f.fxe"},
	{0xF79D80B2D50E9740ULL, "gen_im_wep_rog_arc_act2_05.uti"},
	{0xF79DEB685A39DA10ULL, "and100_anders6095859.cl"},
	{0xF79DF635E35C360DULL, "hlac_w_bonepit_458758_0z.dds"},
	{0xF79E2F191CB0A53EULL, "6142487_m.fxe"},
	{0xF79E36AF289A3474ULL, "art\\characters\\playercharacter\\humanmale\\arm_urva.phy"},
	{0xF79E3F34EEA7E682ULL, "6162544_m.fxe"},
	{0xF79E433D3BEFD839ULL, "hlac_lt_0_lowtown_a3_n_c_131071_5x.dds"},
	{0xF79EB799875FF706ULL, "one100_solivitus6048169.cl"},
	{0xF79F082E90459087ULL, "6173345_m.fxe"},
	{0xF79F088925F38FCEULL, "6116811_m.fxe"},
	{0xF79F756B208BF3E3ULL, "6197246_m.fxe"},
	{0xF79FC3ECB7C96477ULL, "mrl150_merrill_pt4.stg"},
	{0xF7A00111644378DFULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_loop.gad"},
	{0xF7A007DA1837B4F4ULL, "vxht_mansion2_2jy_0_0.vlm"},
	{0xF7A03150324657A6ULL, "gen_im_npc_arm_hum_cht_apoa.uti"},
	{0xF7A03C7313E65D07ULL, "zz_qcr_debug.cnv"},
	{0xF7A0439D0865D32BULL, "ave221_brennan6110470.cl"},
	{0xF7A044A2843B9A90ULL, "cry_val_custic.mao"},
	{0xF7A0612DEC1ACC57ULL, "487802851.ogg"},
	{0xF7A088D22DC45DFEULL, "dre000_mother6037122.cl"},
	{0xF7A0DF68C92535D5ULL, "vxlt_gallowscourtyard_76t_0_0.vlm"},
	{0xF7A102ABC1BDB31FULL, "hlac_lt_0_lowtown_a2_n_c_65537_7z.dds"},
	{0xF7A1572B591366F6ULL, "vxht_0_hightown_n_7m0_0_0.vlm"},
	{0xF7A1E82984027AFEULL, "vxlt_0_lowtown_h5p_0_0.vlm"},
	{0xF7A25ED0848F2143ULL, "vxw_sundermount_30a_0_0.vlm"},
	{0xF7A29B3CDF5A4764ULL, "vxht_0_hightown_a3_9b7_0_0.vlm"},
	{0xF7A2C868D00AB39DULL, "483592547.ogg"},
	{0xF7A34515B2114D97ULL, "vxw_sundermount_2ca_0_0.vlm"},
	{0xF7A45EDB4372356DULL, "art\\vfx\\combat\\vfx_r_gn3_embu_disappear_p.mmh"},
	{0xF7A4D45B6FDF0849ULL, "6115564_f.fxe"},
	{0xF7A4DA62D7D72DA5ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\dr1a_stonedefault_3.pwk"},
	{0xF7A4EC539BCE629DULL, "vxdr_primevaltomb_6w2_0_0.vlm"},
	{0xF7A55F3857105F65ULL, "vxlt_0_lowtown_docks_n_bbb_0_0.vlm"},
	{0xF7A5BB446F8CE65BULL, "6106200_m.fxe"},
	{0xF7A5BE2DACB71886ULL, "229885585.ogg"},
	{0xF7A5FFF4C6B3F28EULL, "287748330.ogg"},
	{0xF7A60DCB6A31BB4BULL, "vxw_blightlands_ig_0_0.vlm"},
	{0xF7A6652486E6C109ULL, "193949581.ogg"},
	{0xF7A66F3EA58FADFCULL, "vxlt_0_lowtown_a2_n_cmr_0_0.vlm"},
	{0xF7A6A8D8EB70F74CULL, "qun210ip_loot_b.utp"},
	{0xF7A6D130B84D12FEULL, "vxht_chantry_n_2zx_0_0.vlm"},
	{0xF7A6F611644984CCULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_loop.evt"},
	{0xF7A7045A11EFE798ULL, "800471699.ogg"},
	{0xF7A813AF28A24D71ULL, "art\\characters\\playercharacter\\humanmale\\arm_urva.msh"},
	{0xF7A821AF28A2648CULL, "art\\characters\\playercharacter\\humanmale\\arm_urva.mao"},
	{0xF7A82DAF28A2792FULL, "art\\characters\\playercharacter\\humanmale\\arm_urva.mmh"},
	{0xF7A842758AD5A1CAULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\blendtree\\gallowsslavestatue_pose6.anb"},
	{0xF7A84ACCA50D29BDULL, "6024090_m.fxe"},
	{0xF7A867C483FC08B2ULL, "vxca_caverns1_edb_0_0.vlm"},
	{0xF7A8C4ABA6C0FFCBULL, "vxlt_0_lowtown_a3_n_c_by7_0_0.vlm"},
	{0xF7A8D27CA1805280ULL, "6077136_m.fxe"},
	{0xF7A9062985BABE1DULL, "lt_0_lowtown_water_21225.mao"},
	{0xF7A90A2985BAC556ULL, "lt_0_lowtown_water_21225.mmh"},
	{0xF7A9142985BAD654ULL, "lt_0_lowtown_water_21225.msh"},
	{0xF7A9159CE2FAAD29ULL, "art\\characters\\creatures\\hurlock\\c_uni_hrlk_bdy_cb.mao"},
	{0xF7A94BEF202E267FULL, "vxca_caverns1_mines_gke_0_0.vlm"},
	{0xF7A9CC746A9468E2ULL, "6119448_f.fxe"},
	{0xF7AA34E2ADB80689ULL, "dae300_alistair_king6134972.cl"},
	{0xF7AA677313EEFB74ULL, "zz_qcr_debug.ncs"},
	{0xF7AA96BD814B732AULL, "dae021_chantry_initiate_mob6145196.cl"},
	{0xF7AAAE0155DBE372ULL, "vxht_keep_609_0_0.vlm"},
	{0xF7AAF0E7C98B333DULL, "6166994_m.fxe"},
	{0xF7ABDA9257828F9FULL, "6102877_m.fxe"},
	{0xF7ABDCFEA1742ACDULL, "w_bonepitmining_a2_18_0.msh"},
	{0xF7ABEAFEA1744200ULL, "w_bonepitmining_a2_18_0.mao"},
	{0xF7ABEEFEA1744953ULL, "w_bonepitmining_a2_18_0.mmh"},
	{0xF7AC14140C04F77AULL, "6137695_f.fxe"},
	{0xF7AC285F2D1099ADULL, "gen_im_trs_gen_lgt_glv_r1.uti"},
	{0xF7ACEC5B17DBAACCULL, "hlac_w_sundermount_65531_0y.dds"},
	{0xF7ACF4509F4033A3ULL, "4351798.ogg"},
	{0xF7AD3608C3E3BD3FULL, "weapon_weapon_p0.psh.fxo"},
	{0xF7AD528061129986ULL, "vxlt_0_lowtown_h2a_0_0.vlm"},
	{0xF7AD660998C0ABE6ULL, "6123653_m.fxe"},
	{0xF7AD877E0ED36E44ULL, "w_bonepitdestroyed_a3_211_0.mmh"},
	{0xF7AD8904B813B3DAULL, "cd_one300_zevran.wbd"},
	{0xF7AD8B7E0ED37497ULL, "w_bonepitdestroyed_a3_211_0.mao"},
	{0xF7AD9D7E0ED393E6ULL, "w_bonepitdestroyed_a3_211_0.msh"},
	{0xF7ADA3BD435C4FE4ULL, "vxw_sundermount_25x_0_0.vlm"},
	{0xF7AE05929AB9EC53ULL, "llac_ht_brothel_n_3371_0.dds"},
	{0xF7AE6F300D4F329FULL, "6108524_f.fxe"},
	{0xF7AF6D3987E34F8BULL, "712306385.ogg"},
	{0xF7AF922E35792ED4ULL, "gen00fl_carver6102546.cl"},
	{0xF7AFCCA75E156048ULL, "follower_barks6153823.cl"},
	{0xF7B0188948A5F127ULL, "55933482.ogg"},
	{0xF7B0438C77CBD609ULL, "fen110_anso6094959.cl"},
	{0xF7B04AA631D85384ULL, "lt_0_lowtown_docks_a3_n_208_0.mmh"},
	{0xF7B04EA631D85AD7ULL, "lt_0_lowtown_docks_a3_n_208_0.mao"},
	{0xF7B060A631D87926ULL, "lt_0_lowtown_docks_a3_n_208_0.msh"},
	{0xF7B0A0910F2FA76DULL, "lt_0_lowtown_a2_n_22_0.rml"},
	{0xF7B0E8CA91A49D3CULL, "358915922.ogg"},
	{0xF7B0EDB55330A121ULL, "mag170_recruit01.bnk"},
	{0xF7B118A6876A82A0ULL, "follower_banter6173013.cl"},
	{0xF7B1C00B7C53ECF5ULL, "hlac_ht_0_hightown_a2_65534_1z.dds"},
	{0xF7B21B6F3C078B5FULL, "llac_lt_0_lowtown_a2_n_c_65536_0.dds"},
	{0xF7B21CC230DEF5EEULL, "786046016.ogg"},
	{0xF7B22C26A0E3C53EULL, "gen00fl_isabela_init6127649.cl"},
	{0xF7B27E304851E878ULL, "6161262_m.fxe"},
	{0xF7B2857F0C4C51F5ULL, "6109104_m.fxe"},
	{0xF7B28FA2AF48DBF2ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_018a_0n.dds"},
	{0xF7B2EE1731C87F35ULL, "vxdr_primevaltomb_2xw_0_0.vlm"},
	{0xF7B3116BE396D729ULL, "hlac_ht_0_hightown_-131071_0z.dds"},
	{0xF7B34BBCFC9A5CBAULL, "6106606_f.fxe"},
	{0xF7B35299505FCD9FULL, "lgt250_way_dalish_conflict6147149.cl"},
	{0xF7B36B521EC69169ULL, "lgt150_nat_were_supporter6111006.cl"},
	{0xF7B3C23B79639773ULL, "vxdr_primevalentrance_2zn_0_1.vlm"},
	{0xF7B3DE116453E625ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_heart_kill_kneeling_reaction_loop.ani"},
	{0xF7B44F7B69A98279ULL, "art\\vfx\\materials\\vfx_lightbeam.mao"},
	{0xF7B4512F599F5EEEULL, "hlac_lt_0_lowtown_a2_-4_0x.dds"},
	{0xF7B45FCB17C8D372ULL, "383844378.ogg"},
	{0xF7B47B0A563A15A2ULL, "mrl310_party_comments6185814.cl"},
	{0xF7B47FA85C27D3CBULL, "vxw_blightlands_rf_0_0.vlm"},
	{0xF7B4AEDDD7F7E044ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_3.ani"},
	{0xF7B4DA9206F56F16ULL, "vxlt_darktown_6j0_0_0.vlm"},
	{0xF7B50152F5CAECA9ULL, "6126467_f.fxe"},
	{0xF7B50FB55334C391ULL, "mag170_recruit01.cnv"},
	{0xF7B541BA15739C48ULL, "rdr111dg_assistant6016655.cl"},
	{0xF7B54A7E7903D0B3ULL, "hlac_ht_keep_4519_8y.dds"},
	{0xF7B56EFA983D3141ULL, "6165142_m.fxe"},
	{0xF7B6326C3526BDD1ULL, "w_encounter3_246_0.mmh"},
	{0xF7B63E6C3526D272ULL, "w_encounter3_246_0.mao"},
	{0xF7B64C6C3526E98FULL, "w_encounter3_246_0.msh"},
	{0xF7B6F7CBD922A547ULL, "plt_ave200pt_night_terrors.nss"},
	{0xF7B7178590B477FEULL, "hlac_lt_0_lowtown_a3_n_c_65535_6y.dds"},
	{0xF7B72B3F9E70A3FCULL, "art\\levels\\ztools\\default\\generic\\9ff_rock.phy"},
	{0xF7B757BAC65CCF01ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_action.ani"},
	{0xF7B78B0AE6BF3FC3ULL, "llac_w_sundermount_-262148_0.dds"},
	{0xF7B7C88A58468618ULL, "vxdr_primevalentrance_1kx_0_0.vlm"},
	{0xF7B8D1159CAD80CAULL, "vxw_encounter1_jl_0_1.vlm"},
	{0xF7B9678DDB5BA1D7ULL, "qcr000_cut2d_thechampion6025127.cl"},
	{0xF7B98459FF219FCCULL, "gen_im_wep_rog_dua_act0_01.uti"},
	{0xF7B9C7A090915659ULL, "hlac_w_blightlands_-65535_0y.dds"},
	{0xF7B9FE572433D553ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_exit.ani"},
	{0xF7BA258FD4E8F7F5ULL, "mcr371_hostile_anders6082102.cl"},
	{0xF7BA6D711662535BULL, "6154012_m.fxe"},
	{0xF7BA9DE8366A9231ULL, "vxlt_gallowstemplar_n_2xr_0_0.vlm"},
	{0xF7BACB3061F9C6B6ULL, "vxlt_0_lowtown_docks_n_bdw_0_0.vlm"},
	{0xF7BB7FB6ACF91024ULL, "284243108.ogg"},
	{0xF7BB99DFD7C55477ULL, "vxlt_0_lowtown_a2_n_c_glt_0_0.vlm"},
	{0xF7BBABDDD7FDF9A6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_3.gad"},
	{0xF7BBBDB68EDC0929ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb03a.phy"},
	{0xF7BBC45D5308B206ULL, "vxlt_0_lowtown_h7b_0_0.vlm"},
	{0xF7BCCC0FE7D53E2EULL, "6102771_m.fxe"},
	{0xF7BCFD27F26F0096ULL, "art\\vfx\\creatures\\vfx_c_hd_tswipe_c.mmh"},
	{0xF7BD2F5313C8C41BULL, "6124239_m.fxe"},
	{0xF7BD62BAC6614DA3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_action.gad"},
	{0xF7BDF9C0478DF689ULL, "mrl150_merrill6070506.cl"},
	{0xF7BE38000AEA148AULL, "vxw_bonepitdestroyed_a3_dz_0_0.vlm"},
	{0xF7BE3AE53CFF97F6ULL, "vxht_0_hightown_n_7g2_0_0.vlm"},
	{0xF7BE45252215389DULL, "6050384_f.fxe"},
	{0xF7BEB49F7E763BF5ULL, "6147681_m.fxe"},
	{0xF7BF679ED4BD0AC3ULL, "hlac_ht_keep_n_3732_4x.dds"},
	{0xF7BF9BDC64BB4A9CULL, "hlac_ca_caverns2prc_5802_17z.dds"},
	{0xF7C0500A125B6735ULL, "36643595.ogg"},
	{0xF7C06A9B1FF0EFC6ULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_lgtb_d.dds"},
	{0xF7C0E31453ED5F23ULL, "vis121_saemus6101181.cl"},
	{0xF7C0E767DCDF28EEULL, "vxlt_0_lowtown_docks_a3_n_apg_0_0.vlm"},
	{0xF7C0F00D3A4800F8ULL, "6166362_m.fxe"},
	{0xF7C15FF423BE97B5ULL, "vxdr_primevaltomb_2wv_0_0.vlm"},
	{0xF7C17C750A703045ULL, "6087761_m.fxe"},
	{0xF7C199B9A557119FULL, "llac_lt_backalley_n_1_5.dds"},
	{0xF7C1E268DDF583C9ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v1.ani"},
	{0xF7C241CB2EC0DB82ULL, "mag360_alain6044664.cl"},
	{0xF7C26B361C15D8AAULL, "zz_one000_barber6179399.cl"},
	{0xF7C293064EEB59D8ULL, "llac_lt_0_lowtown_a2_n_-262144_4.dds"},
	{0xF7C29AF4675315D7ULL, "vxdr_primevalentrance_30i_0_1.vlm"},
	{0xF7C2A6DDD8040F85ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\leftarm_generic_small_3.evt"},
	{0xF7C2B5BC83AB4AD0ULL, "llac_ht_brothel_n_3364_1.dds"},
	{0xF7C2F1854E1DEC4BULL, "vxw_blightlands_wy_0_0.vlm"},
	{0xF7C370C3B374A091ULL, "art\\characters\\playercharacter\\elfmale\\textures\\bdy_arm_lgte_0d.dds"},
	{0xF7C388046CADB811ULL, "hlac_lt_0_lowtown_docks_-2_0z.dds"},
	{0xF7C3B4EA9C3943FEULL, "vxht_chantry_n_2pz_0_0.vlm"},
	{0xF7C41CD33A1400CBULL, "art\\characters\\playercharacter\\elffemale\\mt_aet_ef05.phy"},
	{0xF7C43FBAC6673088ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_post_throat_slit_action.evt"},
	{0xF7C443E93C2E6208ULL, "vxlt_undercityencounter_13l_0_0.vlm"},
	{0xF7C512ACD069A41DULL, "553109048.ogg"},
	{0xF7C56A09A8FA7E8BULL, "455371093.ogg"},
	{0xF7C64E8DADE8F710ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1c.msh"},
	{0xF7C65B08B8D5DB48ULL, "hm_qsw_shady.mor"},
	{0xF7C65C8DADE90F2AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1c.mmh"},
	{0xF7C6608DADE91661ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookflat1c.mao"},
	{0xF7C71F75A6C26E4DULL, "vxht_0_hightown_a2_cds_0_0.vlm"},
	{0xF7C768BE65C04E7DULL, "follower_banter6171104.cl"},
	{0xF7C7A4D9A776C6FDULL, "vxlt_0_lowtown_docks_b5t_0_0.vlm"},
	{0xF7C7FD68DDFA1D4BULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v1.gad"},
	{0xF7C80B998CAD9E97ULL, "314443356.ogg"},
	{0xF7C80D82439E93E6ULL, "6024551_f.fxe"},
	{0xF7C818695833F4E6ULL, "llac_ht_playershouse_n_855_3.dds"},
	{0xF7C81A569441F34BULL, "llac_lt_0_lowtown_a2_-3_4.dds"},
	{0xF7C89EC59E4BDB7DULL, "marksman_staminaregen.dds"},
	{0xF7C8A1A6EF16A9AEULL, "vxw_encounter2_camp_9e_0_0.vlm"},
	{0xF7C907B9BF169FCCULL, "vxht_0_hightown_a3_c7b_0_0.vlm"},
	{0xF7C93D0D53EFDF36ULL, "vxlt_0_lowtown_docks_n_bax_0_0.vlm"},
	{0xF7C988E8B4E69459ULL, "963438510.ogg"},
	{0xF7CA893E07381F63ULL, "character_character_p259.psh.fxo.dx11.assembly"},
	{0xF7CB28D693F73759ULL, "151021380.ogg"},
	{0xF7CB6F9E1FC913E8ULL, "zz_prmpt_debug6198268.cl"},
	{0xF7CB9A9055CEC57AULL, "152467214.ogg"},
	{0xF7CBC3CCF8EE342CULL, "666409320.ogg"},
	{0xF7CC698CAB8DB9F6ULL, "vxdr_0_deeproads_fiq_0_0.vlm"},
	{0xF7CC851A9315B1E8ULL, "hlac_ht_0_hightown_a3_65533_1y.dds"},
	{0xF7CC9D1BC7529B40ULL, "hlac_ca_caverns1_mines_24421_5z.dds"},
	{0xF7CCBCBEA5F9196AULL, "zz_dre_debug6056955.cl"},
	{0xF7CD149EC652B6D6ULL, "dae000_orsino6085510.cl"},
	{0xF7CD56C25E03787EULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.dura.ani"},
	{0xF7CD6B61D4B4FEA6ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.anb"},
	{0xF7CDB447B9B2AA66ULL, "dae000_arishok6068507.cl"},
	{0xF7CDD557244438FDULL, "art\\characters\\playercharacter\\humanfemale\\animation\\ambient\\dialogue\\lean_back_wall_hand_on_hip_exit.gad"},
	{0xF7CE7A8DB5A45625ULL, "129300781.ogg"},
	{0xF7CE840E2737329AULL, "art\\characters\\weapons\\dagger\\w_dgr_apdc.mao"},
	{0xF7CE9AAC87F39717ULL, "vxw_sundermount_2rb_0_0.vlm"},
	{0xF7CEFA68DE0036B0ULL, "art\\characters\\creatures\\stonegolem\\animation\\atk_v1.evt"},
	{0xF7CF4D80BA2DFC75ULL, "lowtown_fem_lowlife_mob6169189.cl"},
	{0xF7CFC6FD9EE20E21ULL, "314492415.ogg"},
	{0xF7CFD9CC8BBBDB7CULL, "llac_lt_0_lowtown_n_65537_6.dds"},
	{0xF7CFDCD12786937CULL, "931244744.ogg"},
	{0xF7D01373C59F3007ULL, "300692863.ogg"},
	{0xF7D03B414BC90D0CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_firebeam.mmh"},
	{0xF7D05BE8253E9CEFULL, "6173082_m.fxe"},
	{0xF7D0D20823F9EF03ULL, "86050802.ogg"},
	{0xF7D0DBEBD7DC86AAULL, "249434011.ogg"},
	{0xF7D15CC336C23FCCULL, "mrl150_merrill6077737.cl"},
	{0xF7D16F2E8E15F9BAULL, "mrl210_merrill6117418.cl"},
	{0xF7D1E50CD9B71DA9ULL, "6128667_f.fxe"},
	{0xF7D239C8E4859460ULL, "vxca_caverns2prc_8bi_0_0.vlm"},
	{0xF7D2F8C6CD61CFE8ULL, "lt_0_lowtown_docks_n_166_0.mmh"},
	{0xF7D304C6CD61E38BULL, "lt_0_lowtown_docks_n_166_0.mao"},
	{0xF7D316C6CD620212ULL, "lt_0_lowtown_docks_n_166_0.msh"},
	{0xF7D3700B598FDCB1ULL, "w_sundermount_41_0_ll.mmh"},
	{0xF7D39528A6ADDF5DULL, "gen00fl_anders6136003.cl"},
	{0xF7D49D97DF9A2701ULL, "6144979_f.fxe"},
	{0xF7D4AA1A79DFDEE5ULL, "383258969.ogg"},
	{0xF7D51F1E1BB60B74ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_enter.gad"},
	{0xF7D5915298B886CDULL, "vxht_0_hightown_a2_ccr_0_0.vlm"},
	{0xF7D592380D4D9331ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_hammer_1.mao"},
	{0xF7D60573C3597F5AULL, "hm_sailor01.mor"},
	{0xF7D659C87A1FA314ULL, "vxht_0_hightown_9wy_0_0.vlm"},
	{0xF7D6DA5B6C212DCEULL, "6128692_m.fxe"},
	{0xF7D7152A093BEC0AULL, "dae000_serendipity6161362.cl"},
	{0xF7D744CD46730F79ULL, "vxlt_0_lowtown_n_erh_0_0.vlm"},
	{0xF7D78B37BF861F8CULL, "dae000_bodahn_deeproads.stg"},
	{0xF7D7A7472E5A48CAULL, "6206931_f.fxe"},
	{0xF7D7BEC8995985B9ULL, "205692627.ogg"},
	{0xF7D7BF42A0A7D558ULL, "vxlt_undercityserialkiller_15v_0_0.vlm"},
	{0xF7D811320545C861ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_bowc_d.dds"},
	{0xF7D840880F4FB5F8ULL, "vxht_0_hightown_n_a6e_0_0.vlm"},
	{0xF7D911C5A81691B0ULL, "6077019_m.fxe"},
	{0xF7D92D208C0B37E5ULL, "6071294_m.fxe"},
	{0xF7D933DF506D3F54ULL, "vxw_bonepitdestroyed_a3_2ag_0_0.vlm"},
	{0xF7D96444865AECFCULL, "hlac_lt_gallowstemplar_n_2388_4x.dds"},
	{0xF7D9690BA5A6A40CULL, "vxlt_warehouse_15q_0_0.vlm"},
	{0xF7D96F9FD86496C6ULL, "153709081.ogg"},
	{0xF7D986B661D41FD3ULL, "6061824_m.fxe"},
	{0xF7D9B35796B856C2ULL, "vxht_bartrand_n_5vd_0_0.vlm"},
	{0xF7D9BEB45C06B41CULL, "face_face_p9.psh.fxo"},
	{0xF7DA3DD0AC7E34F4ULL, "ts_message.ncs"},
	{0xF7DA4A1AC60D9B94ULL, "6135277_f.fxe"},
	{0xF7DA559CE179E1EFULL, "vxlt_0_lowtown_cw0_0_0.vlm"},
	{0xF7DA6CF9BBA57141ULL, "6076890_f.fxe"},
	{0xF7DAC1F4EBD731DEULL, "689729795.ogg"},
	{0xF7DB6E61D4C14151ULL, "art\\vfx\\combat\\vfx_r_shd_tdecay_loop_c.mmh"},
	{0xF7DB6EC25E0FDE4FULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.dura.evt"},
	{0xF7DBDF89706770ACULL, "hlac_lt_0_lowtown_docks_65534_6y.dds"},
	{0xF7DBFA86C018DE0DULL, "llac_lt_0_lowtown_docks_a3_n_-131072_0.dds"},
	{0xF7DC10CDA8DCCAC0ULL, "388796489.ogg"},
	{0xF7DC2E1E1BBC4352ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_2_enter.ani"},
	{0xF7DD178870E5B190ULL, "863158582.ogg"},
	{0xF7DD413ACB3F0059ULL, "490753739.ogg"},
	{0xF7DDEFA79098DBE6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_13.ani"},
	{0xF7DE2F2A196099FAULL, "225947157.ogg"},
	{0xF7DE58E917E32CECULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.cess.ani"},
	{0xF7DF1178564486DFULL, "art\\characters\\playercharacter\\humanmale\\boo_masa.msh"},
	{0xF7DF177856449081ULL, "art\\characters\\playercharacter\\humanmale\\boo_masa.mmh"},
	{0xF7DF17D24F45E0F3ULL, "vxdr_primevalentrance_2kq_0_1.vlm"},
	{0xF7DF23785644A522ULL, "art\\characters\\playercharacter\\humanmale\\boo_masa.mao"},
	{0xF7DFBB9882774FD5ULL, "vxca_caverns1_gkk_0_1.vlm"},
	{0xF7DFBEA281AD402DULL, "6040584_m.fxe"},
	{0xF7DFD765A3CC58B0ULL, "zz_dae_debug6029479.cl"},
	{0xF7E147EA2BEDA7B7ULL, "221416328.ogg"},
	{0xF7E14C403CF617CBULL, "696690928.ogg"},
	{0xF7E26FC25E15FEF8ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_c.dura.gad"},
	{0xF7E274CD5F1C1BA5ULL, "ave331_jeven6152180.cl"},
	{0xF7E344E6EBA800B4ULL, "vxlt_0_lowtown_docks_n_b4u_0_0.vlm"},
	{0xF7E34D13E51C2EDEULL, "6133753_m.fxe"},
	{0xF7E34D2B98155F56ULL, "vxlt_warehouse_n_b6_0_0.vlm"},
	{0xF7E389DAAA7BB834ULL, "hlac_lt_gallowstemplar_f_2383_1z.dds"},
	{0xF7E38C2D57E80BCCULL, "6084245_f.fxe"},
	{0xF7E465E917E7AE2EULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.cess.gad"},
	{0xF7E46A63FD806C3EULL, "vxht_keep_n_5st_0_0.vlm"},
	{0xF7E46AC3FD1C39D1ULL, "vxw_encounter3_91_0_0.vlm"},
	{0xF7E4FBBE37F151EAULL, "cs_pro000cs_opening_mage.bnk"},
	{0xF7E5377337277155ULL, "mag170_thrask6054045.cl"},
	{0xF7E568669E690624ULL, "hlac_lt_0_lowtown_196607_1y.dds"},
	{0xF7E5B6AA386927F9ULL, "vxlt_0_lowtown_n_eoo_0_0.vlm"},
	{0xF7E5CAE7F933D760ULL, "991404678.ogg"},
	{0xF7E5DB12D3817746ULL, "zz_qun_debug6087152.cl"},
	{0xF7E662AD6B60875BULL, "10723581.ogg"},
	{0xF7E718DC3968C9FAULL, "and111_lirene6084549.cl"},
	{0xF7E73188CDD8A442ULL, "drg_dust_dot.mao"},
	{0xF7E7A6B68F011D4EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb03a.mmh"},
	{0xF7E7C0B68F01493CULL, "art\\characters\\playercharacter\\humanmale\\hm_har_hb03a.msh"},
	{0xF7E7E122A3FCA265ULL, "llac_w_woundedcoast_-196601_0.dds"},
	{0xF7E8253488AE6F42ULL, "vxht_bartrand_n_5ye_0_0.vlm"},
	{0xF7E889BDAAE6DA1FULL, "6187957_m.fxe"},
	{0xF7E8B82A17546440ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_basic.phy"},
	{0xF7E8CF1473AE6B51ULL, "single_star.dds"},
	{0xF7E90BD89AB78100ULL, "docks_male_overseer_mob6176522.cl"},
	{0xF7E91EC162935BD5ULL, "166269436.ogg"},
	{0xF7E94F3C211C519DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_basic.mao"},
	{0xF7E9533C211C58D6ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_basic.mmh"},
	{0xF7E95D3C211C69D4ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\3ww_basic.msh"},
	{0xF7E9F7CDCE09332CULL, "vxlt_0_lowtown_a2_n_c_fhn_0_0.vlm"},
	{0xF7E9FD97422C434CULL, "vxca_caverns2prc_4x6_0_1.vlm"},
	{0xF7EA5CA7C2F918A2ULL, "hlac_lt_darktown_-1_20x.dds"},
	{0xF7EA6C2D680E35DDULL, "vxw_encounter3_gk_0_0.vlm"},
	{0xF7EAA80CCA4A2072ULL, "297766913.ogg"},
	{0xF7EABAF7AAAD1792ULL, "vxw_bonepit_ya_0_0.vlm"},
	{0xF7EAF69B1E925E08ULL, "299648789.ogg"},
	{0xF7EB60E917EDC40DULL, "art\\vfx\\combat\\vfx_w_gn1_wcry_loop_acc_c.cess.evt"},
	{0xF7EB7F4474348520ULL, "6036979_m.fxe"},
	{0xF7EBA6CDD08213BCULL, "plt_zz_mcrpt_debug.plo"},
	{0xF7EBB6770D13B2F1ULL, "vxht_0_hightown_a3_cae_0_0.vlm"},
	{0xF7EBD7A790A4F057ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_13.evt"},
	{0xF7EBEDA176378309ULL, "follower_banter6160748.cl"},
	{0xF7EC47B124C6C252ULL, "6080474_f.fxe"},
	{0xF7ECE9DBCA1B3755ULL, "148569046.ogg"},
	{0xF7ED19D6EEAE9A2AULL, "572618160.ogg"},
	{0xF7ED7E25A6784A72ULL, "430725788.ogg"},
	{0xF7EDF734B50258D3ULL, "6174300_m.fxe"},
	{0xF7EDFE25CD38B5A9ULL, "hairoverdraw.psh.fxo"},
	{0xF7EE03DBA8518190ULL, "hlac_lt_gallowstemplar_2391_4y.dds"},
	{0xF7EEA643838E307DULL, "lgtb_bandit_progression.gda"},
	{0xF7EF5E1AFEEC8C7FULL, "281618652.ogg"},
	{0xF7F00051467C04D7ULL, "vxht_mansion2_31j_0_0.vlm"},
	{0xF7F0551CF0E438B5ULL, "6082609_m.fxe"},
	{0xF7F05EDCBF9ADEE1ULL, "barr_rock.mao"},
	{0xF7F0707B8F020535ULL, "vxht_playershouse_n_1vu_0_0.vlm"},
	{0xF7F10CD0F80A094FULL, "267944397.ogg"},
	{0xF7F115183CB3ED76ULL, "lt_gallowstemplar_f_7198.rml"},
	{0xF7F161D5601E9933ULL, "gen00fl_aveline6120786.cl"},
	{0xF7F1FEA0F7FC1345ULL, "hlac_lt_0_lowtown_a2_-196609_2x.dds"},
	{0xF7F22861863B75D0ULL, "dae000ip_crowd_mf4_talk11.utp"},
	{0xF7F2D8A790AB10E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\po_head_13.gad"},
	{0xF7F2DCA0EF125251ULL, "vxw_encounter3_20_0_0.vlm"},
	{0xF7F2FB27080C7AF4ULL, "6201571_m.fxe"},
	{0xF7F34366F1B47262ULL, "hf_mother_h2s3.mor"},
	{0xF7F3BB21E4B36BC9ULL, "hlac_ca_caverns1_mines_24423_13z.dds"},
	{0xF7F3D209B1121049ULL, "6052910_m.fxe"},
	{0xF7F3DFCD832A0685ULL, "6165937_m.fxe"},
	{0xF7F43CF46D0486B5ULL, "hlac_lt_0_lowtown_docks_-3_5z.dds"},
	{0xF7F46D43DD8343B1ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_ls109a_0s.dds"},
	{0xF7F48D825957C557ULL, "plt_lgt210pt_ingredients.nss"},
	{0xF7F52C3E50545C72ULL, "ico_locked_passive.dds"},
	{0xF7F57A3F156766EAULL, "1070042619.ogg"},
	{0xF7F5B5D29C62F58DULL, "gen00fl_player6207703.cl"},
	{0xF7F5CBD511808526ULL, "6201627_m.fxe"},
	{0xF7F5D0C37E3A5846ULL, "6080086_m.fxe"},
	{0xF7F5D8C95B8DBE75ULL, "345923843.ogg"},
	{0xF7F66B008FE462C8ULL, "bee_up2.dds"},
	{0xF7F6B97D9D49A512ULL, "llac_lt_0_lowtown_n_-1_0.dds"},
	{0xF7F79EE979F53EFAULL, "6095776_m.fxe"},
	{0xF7F7E260E124105CULL, "art\\levels\\races\\mage\\prop\\_textures\\prp_magiclecturn_d.dds"},
	{0xF7F868EE2CFEAC43ULL, "w_sundermount_106_0.mao"},
	{0xF7F86CEE2CFEB290ULL, "w_sundermount_106_0.mmh"},
	{0xF7F87AEE2CFECAAAULL, "w_sundermount_106_0.msh"},
	{0xF7F87C0B505F2FB5ULL, "zz_dae_debug6029902.cl"},
	{0xF7F8AD6F26B90A70ULL, "w_bonepit_79_0.msh"},
	{0xF7F8BB6F26B9218AULL, "w_bonepit_79_0.mmh"},
	{0xF7F8BF6F26B928C1ULL, "w_bonepit_79_0.mao"},
	{0xF7F91F68F640A1E6ULL, "mag100_thrask6032063.cl"},
	{0xF7FA07C702A7C6F8ULL, "dae000_madam_lusine6152198.cl"},
	{0xF7FA6802AEE77672ULL, "follower_banter6141674.cl"},
	{0xF7FA701D515AA386ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_enter.evt"},
	{0xF7FB93E6FD7E286BULL, "611719578.ogg"},
	{0xF7FBCD10BC32D2C5ULL, "art\\vfx\\textures\\vfx_flare01.dds"},
	{0xF7FC29BB1405A4D2ULL, "572294995.ogg"},
	{0xF7FC33C2BC308D51ULL, "6125252_m.fxe"},
	{0xF7FC356CB9C2D810ULL, "hlac_ht_mansion2b_818_60z.dds"},
	{0xF7FC4E8EC3B84C3DULL, "6178320_m.fxe"},
	{0xF7FCC154120EDC2EULL, "vxca_caverns1_j3x_0_0.vlm"},
	{0xF7FD0B720537E29CULL, "vxlt_warehouse_n_qe_0_0.vlm"},
	{0xF7FD10117C6D9DC1ULL, "mcr310cs_chantry_explosion.cut"},
	{0xF7FD3930BAEBD395ULL, "lowtown_male_alienage1_sta6162777.cl"},
	{0xF7FD3A5DF6AE888AULL, "341463304.ogg"},
	{0xF7FD6CCBA5154F7FULL, "739979582.ogg"},
	{0xF7FDA7B14BC4CBD2ULL, "6000723_m.fxe"},
	{0xF7FDCAE0EDFBF010ULL, "6130054_f.fxe"},
	{0xF7FE53DC7AD7B99CULL, "mag170st_hugh.stg"},
	{0xF7FE65F8D43AE6F4ULL, "6110792_m.fxe"},
	{0xF7FEE557DA0485A3ULL, "w_bonepitmining_a2_93_0.mao"},
	{0xF7FEE957DA048CF0ULL, "w_bonepitmining_a2_93_0.mmh"},
	{0xF7FEF757DA04A40AULL, "w_bonepitmining_a2_93_0.msh"},
	{0xF7FF0927F4FFA5CBULL, "cry_val03.mao"},
	{0xF7FF49F3186A005EULL, "ch2notnt_d.psh.fxo"},
	{0xF7FF6C719E8FE9AAULL, "6082127_m.fxe"},
	{0xF7FF80828C1CF2E3ULL, "vxlt_undercityserialkiller_1lr_0_0.vlm"},
	{0xF7FFA4BE2A2D5E7AULL, "w_bonepitmining_a2_42_0.mmh"},
	{0xF7FFA8BE2A2D64B1ULL, "w_bonepitmining_a2_42_0.mao"},
	{0xF7FFB6BE2A2D7CC0ULL, "w_bonepitmining_a2_42_0.msh"},
	{0xF80004D9120EF8DEULL, "533183968.ogg"},
	{0xF80030F1DBC261BFULL, "vxdr_primevaltomb_7sm_0_0.vlm"},
	{0xF8003384C15D40DDULL, "6088749_f.fxe"},
	{0xF8004ADA11B91659ULL, "22449358.ogg"},
	{0xF8006A4F5EFB5EABULL, "hlac_ht_mansion2b_818_1z.dds"},
	{0xF8009DB000235033ULL, "228615209.ogg"},
	{0xF800D13D41026C6CULL, "570691664.ogg"},
	{0xF801871D5160E9D1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_enter.gad"},
	{0xF801A42558E8BBE1ULL, "hlac_lt_blackemporium_1055_23x.dds"},
	{0xF802382CE1E7BAB4ULL, "923937852.ogg"},
	{0xF802AF39FAE0B402ULL, "vxht_mansion2_3c5_0_0.vlm"},
	{0xF803A6896A90E904ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy.ani"},
	{0xF803E3CA75C31640ULL, "hlac_ca_caverns1_24426_14z.dds"},
	{0xF8041F65AA0EA4ECULL, "zz_mrl_debug6109316.cl"},
	{0xF8045CE97D14A62FULL, "6060272_m.fxe"},
	{0xF8047707845CD902ULL, "vxlt_gallowscourtyard_n_7e9_0_0.vlm"},
	{0xF8049A1B4752A0BDULL, "984491727.ogg"},
	{0xF804C88D644F32DAULL, "w_bonepit_154_0.mmh"},
	{0xF804CC8D644F3911ULL, "w_bonepit_154_0.mao"},
	{0xF804DA8D644F5120ULL, "w_bonepit_154_0.msh"},
	{0xF805A40CB6DD2873ULL, "lgt251ip_at_sundermount.utp"},
	{0xF805F1BBA929EEFFULL, "mag100_cullen6093744.cl"},
	{0xF806600271EAF7B3ULL, "6078798_m.fxe"},
	{0xF80671CB007090A5ULL, "vxdr_primevalentrance_fe_0_0.vlm"},
	{0xF8068CF74E26F2F5ULL, "1915610.ogg"},
	{0xF806A275BBDD49C9ULL, "vxw_encounter1_c5_0_0.vlm"},
	{0xF8071F5116F5C6F4ULL, "hlac_lt_gallowstemplar_f_2382_2x.dds"},
	{0xF807881D5165577FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\controlling_golem_enter.ani"},
	{0xF80789CE41BEEB25ULL, "vxlt_gallowscourtyard_n_65g_0_0.vlm"},
	{0xF8078C764AA40E54ULL, "6163788_f.fxe"},
	{0xF807BCEBEFABFCCEULL, "6153377_m.fxe"},
	{0xF807C5AED9B26046ULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_rugredlong_d.dds"},
	{0xF80875EBB1689D1BULL, "prm000im_blood_boots.uti"},
	{0xF808DDF063BB25C6ULL, "face_face_p10.psh.fxo.dx11.assembly"},
	{0xF8095A2C8DA44559ULL, "follower_banter6149309.cl"},
	{0xF80A9FDFB98F2478ULL, "gen00fl_anders6125545.cl"},
	{0xF80AA3896A970366ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy.gad"},
	{0xF80B18785669CDFEULL, "art\\characters\\playercharacter\\humanmale\\boo_masa.phy"},
	{0xF80B76EDB55E378AULL, "follower_banter6168033.cl"},
	{0xF80B7D4EF72DFB1CULL, "vxlt_warehouse_n_tz_0_0.vlm"},
	{0xF80B96AA03FD6454ULL, "plt_cod_itm_puzzleringfox.plo"},
	{0xF80BBA37EC1526FBULL, "vxlt_ruinedhouse_n_2e_0_0.vlm"},
	{0xF80BF6EC855710ABULL, "39316397.ogg"},
	{0xF80BFE555ACC3E3BULL, "art\\levels\\races\\dwarf\\props\\_textures\\prp_lyriumblue_sa.dds"},
	{0xF80C7EADF16DBBCAULL, "gen00fl_anders6114459.cl"},
	{0xF80CAF6D8DDEB413ULL, "vxlt_gallowsprison_n_6bb_0_0.vlm"},
	{0xF80D521EFA22A074ULL, "gen00fl_merrill6129019.cl"},
	{0xF80D6B3CE3B743F1ULL, "vxlt_gallowstemplar_n_5j7_0_0.vlm"},
	{0xF80DD52A1773B1D5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_basic.msh"},
	{0xF80DD72A1773B54BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_basic.mmh"},
	{0xF80DE32A1773C9E8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0cc_basic.mao"},
	{0xF80E208A22FF7372ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkb.gad"},
	{0xF80E3761F971BCDCULL, "vxlt_0_lowtown_a2_gvh_0_0.vlm"},
	{0xF80E4589AB103986ULL, "6059691_m.fxe"},
	{0xF80EA561B353595DULL, "6083649_f.fxe"},
	{0xF80F4787A65DC004ULL, "vxht_mansion2b_3db_0_0.vlm"},
	{0xF80FB0CBD270731EULL, "dae210_gate_guard6099081.cl"},
	{0xF80FFA8FCC9B1C6DULL, "6165577_m.fxe"},
	{0xF810023D87DF3D77ULL, "232032775.ogg"},
	{0xF81012E5993B7EA0ULL, "llac_ht_keep_4519_8.dds"},
	{0xF810234B95746B12ULL, "vxca_caverns2prc_9t8_0_0.vlm"},
	{0xF8109E896A9B6645ULL, "art\\characters\\creatures\\desiredemon\\animation\\rdy.evt"},
	{0xF810BD944274E517ULL, "351974204.ogg"},
	{0xF810D596B8D6E7DFULL, "nat380_sibling.bnk"},
	{0xF810F9DADD95A64EULL, "gentr_generic.utt"},
	{0xF8112116ECD6CC82ULL, "vxht_mansion2_3f6_0_0.vlm"},
	{0xF811DF94A2CDC299ULL, "6106818_f.fxe"},
	{0xF811F93035E49C92ULL, "t3_bow_ibk.tnt"},
	{0xF8121EB8D9F04C84ULL, "dae000cr_chandan.utc"},
	{0xF812A83164FC25BDULL, "325657458.ogg"},
	{0xF812F00DED610704ULL, "6081810_m.fxe"},
	{0xF81353F3D6FFD527ULL, "vxlt_0_lowtown_docks_b6n_0_1.vlm"},
	{0xF813B796B8D8EAE7ULL, "nat380_sibling.cnv"},
	{0xF8143CA41379080CULL, "625427590.ogg"},
	{0xF81442D0761DA109ULL, "vxlt_darktown_5co_0_0.vlm"},
	{0xF814A9FA7383B08BULL, "vxlt_0_lowtown_a2_n_bxw_0_0.vlm"},
	{0xF815006D1370C198ULL, "vxlt_warehouse_n_127_0_0.vlm"},
	{0xF8151452ADD36249ULL, "vxw_encounter1_h4_0_0.vlm"},
	{0xF815238A230596A9ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkb.evt"},
	{0xF8155B6C7C5B6556ULL, "vxht_chantry_13w_0_0.vlm"},
	{0xF815CE24FD678AE6ULL, "540206887.ogg"},
	{0xF816066CEC9478FAULL, "107726445.ogg"},
	{0xF8169EDD253DA343ULL, "572905889.ogg"},
	{0xF816F4843F6584E1ULL, "vxlt_0_lowtown_a2_n_ffo_0_0.vlm"},
	{0xF817000596D4FA08ULL, "346331336.ogg"},
	{0xF817B26AF2ADD1BEULL, "gen00fl_sebastian6179873.cl"},
	{0xF818CC82D57D3612ULL, "gen00fl_fenris6128014.cl"},
	{0xF818DE07E9B9BA7FULL, "vxdr_primevaltomb_7ej_0_1.vlm"},
	{0xF818E322552CA82FULL, "vxlt_gallowscourtyard_59u_0_0.vlm"},
	{0xF818F8F1F3C5F757ULL, "vxlt_warehouse_zz_0_0.vlm"},
	{0xF81957BB80C05816ULL, "325937920.ogg"},
	{0xF8196B944A083F63ULL, "vxca_cavern1_haunted_i1k_0_2.vlm"},
	{0xF8196F140DDEDC76ULL, "ht_0_hightown_n_54_0.rml"},
	{0xF8198C8E3C573735ULL, "hlac_w_bonepit_65539_0y.dds"},
	{0xF819FDC1070293BFULL, "6143287_m.fxe"},
	{0xF81A2ADBB1D3BF75ULL, "hlac_lt_0_lowtown_docks_n_-196608_1z.dds"},
	{0xF81A36B44C2D5589ULL, "hlac_ht_0_hightown_a3_-65534_1x.dds"},
	{0xF81A49172CC30D8AULL, "one300dg_lia6064548.cl"},
	{0xF81A79C28C3E69F6ULL, "6043252_m.fxe"},
	{0xF81A9F0E524944FFULL, "6054129_m.fxe"},
	{0xF81AA06885B5A928ULL, "vxht_0_hightown_a3_bym_0_0.vlm"},
	{0xF81ABCCF1D1B877BULL, "vxca_cavern1_haunted_e8v_0_0.vlm"},
	{0xF81AED7B9C53A799ULL, "6016846_m.fxe"},
	{0xF81B224C8ABE381BULL, "art\\vfx\\creatures\\fxc_succubus_c.mmh"},
	{0xF81B4790DA968ED5ULL, "zz_fen_debug6165494.cl"},
	{0xF81B58B3C468A0DFULL, "6179004_m.fxe"},
	{0xF81B6B975D33C6A4ULL, "w_bonepit_215_0.mmh"},
	{0xF81B6F975D33CDF7ULL, "w_bonepit_215_0.mao"},
	{0xF81B81975D33EC46ULL, "w_bonepit_215_0.msh"},
	{0xF81BF21CEC36C473ULL, "vxlt_gallowstemplar_2ja_0_0.vlm"},
	{0xF81CA93EEB67D55CULL, "vxlt_0_lowtown_a2_gmi_0_0.vlm"},
	{0xF81DA6275CA8FA7DULL, "6043541_m.fxe"},
	{0xF81DA75478D2E641ULL, "342544270.ogg"},
	{0xF81DB9B3F5EC6637ULL, "lgt260_eau_tmp_flowergirl.bnk"},
	{0xF81DD119B1E8F55AULL, "art\\characters\\creatures\\mubari\\animation\\howl.gad"},
	{0xF81DD15022672F90ULL, "vxlt_0_lowtown_docks_n_cqs_0_0.vlm"},
	{0xF81DF3499D57B7B9ULL, "898800943.ogg"},
	{0xF81E12FD3CA87AF0ULL, "art\\characters\\playercharacter\\humanmale\\hoo_roba.mao"},
	{0xF81E16FD3CA88103ULL, "art\\characters\\playercharacter\\humanmale\\hoo_roba.mmh"},
	{0xF81E24FD3CA8995DULL, "art\\characters\\playercharacter\\humanmale\\hoo_roba.msh"},
	{0xF81E9528876A8392ULL, "vxca_caverns2prc_9s9_0_0.vlm"},
	{0xF81E9CE04A576A9CULL, "vxlt_darktown_4ws_0_0.vlm"},
	{0xF81EE309940FF885ULL, "mcr370_cut_attack_begins.bnk"},
	{0xF81EFB4A6FD74803ULL, "vxlt_0_lowtown_docks_a3_n_d0c_0_0.vlm"},
	{0xF81F522D56418AFFULL, "ser221_gascard6149260.cl"},
	{0xF81FC87F4734BDECULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_mera_0n.dds"},
	{0xF81FEF6BFEA0475DULL, "vxlt_warehouse_tr_0_0.vlm"},
	{0xF820C04E1C432F91ULL, "vxdr_primevalentrance_1eb_0_1.vlm"},
	{0xF82136230F130F1BULL, "zz_dae_debug6131010.cl"},
	{0xF8219BB3F5F01C4FULL, "lgt260_eau_tmp_flowergirl.cnv"},
	{0xF821A3C9689C07C4ULL, "543897499.ogg"},
	{0xF821C5099411FBBDULL, "mcr370_cut_attack_begins.cnv"},
	{0xF82305EF605715ABULL, "292922353.ogg"},
	{0xF8230B8A2311AB18ULL, "art\\characters\\creatures\\desiredemon\\animation\\cwkb.ani"},
	{0xF8244488730DB3CFULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.cess.gad"},
	{0xF82451240A4BB1A7ULL, "hlac_lt_0_lowtown_a2_n_c_-196608_2z.dds"},
	{0xF824817FFB90B37FULL, "art\\levels\\4cinematics\\ferelden\\w_small_dagger_0.msh"},
	{0xF824877FFB90BD21ULL, "art\\levels\\4cinematics\\ferelden\\w_small_dagger_0.mmh"},
	{0xF824937FFB90D1C2ULL, "art\\levels\\4cinematics\\ferelden\\w_small_dagger_0.mao"},
	{0xF824A2DBC9B0964DULL, "w_encounter1_2_0.mmh"},
	{0xF824A6DBC9B09C9EULL, "w_encounter1_2_0.mao"},
	{0xF824B4DBC9B0B4D3ULL, "w_encounter1_2_0.msh"},
	{0xF824C94E514CCF75ULL, "6053254_m.fxe"},
	{0xF824D419B1EF1971ULL, "art\\characters\\creatures\\mubari\\animation\\howl.evt"},
	{0xF8250216E3F39A6FULL, "hf_idunna.mor"},
	{0xF82522F3558B2CC5ULL, "hlac_lt_0_lowtown_a2_n_c_1_7y.dds"},
	{0xF825665C25675EE8ULL, "243819700.ogg"},
	{0xF8256A907A5B41CCULL, "dae300_cullen6140613.cl"},
	{0xF825DBBA1F8E04BEULL, "vxlt_0_lowtown_docks_c5n_0_0.vlm"},
	{0xF826250DD2F83322ULL, "vxlt_hangedman_n_21w_0_0.vlm"},
	{0xF82660B89DCB6F08ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\0cc_basic_1.mao"},
	{0xF826AD6812365B7BULL, "6171888_m.fxe"},
	{0xF826CFD5E2AE46F3ULL, "6128495_m.fxe"},
	{0xF826F1FB801A764DULL, "vxlt_0_lowtown_docks_a3_cte_0_0.vlm"},
	{0xF8271A749CE384F0ULL, "vxlt_gallowstemplar_n_377_0_0.vlm"},
	{0xF8271B155387B21AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08a.phy"},
	{0xF8272D2528D2F2E7ULL, "6102525_m.fxe"},
	{0xF827C8B98EE3120BULL, "mag110_samson6053547.cl"},
	{0xF82835F84183A106ULL, "dae000_hate_follower_react6167480.cl"},
	{0xF828434C8AC8FA80ULL, "art\\vfx\\creatures\\fxc_succubus_c.anb"},
	{0xF8295505EE5AC70DULL, "hlac_ht_0_hightown_a2_-3_0x.dds"},
	{0xF8297002741ADA29ULL, "mer120_hubert6070233.cl"},
	{0xF829932771CAE513ULL, "vxlt_gallowsprison_n_6hd_0_0.vlm"},
	{0xF8299E9F6308303EULL, "6121127_m.fxe"},
	{0xF82A12183CE3AD2AULL, "lt_gallowstemplar_f_7198.anb"},
	{0xF82A688905A2E532ULL, "vxca_caverns1_e7k_0_0.vlm"},
	{0xF82AA5B179766BACULL, "vxlt_backalley_g_1k8_0_0.vlm"},
	{0xF82ADF790CFC3065ULL, "vxdr_primevaltomb_758_0_1.vlm"},
	{0xF82ADFD7EA5A92C5ULL, "gen00fl_varric6104952.cl"},
	{0xF82AE78D647036E5ULL, "w_bonepit_154_0.phy"},
	{0xF82AFA5F8140D403ULL, "195076100.ogg"},
	{0xF82B59887313F67CULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.cess.evt"},
	{0xF82B67A8D7B97743ULL, "vxlt_blackemporium_4_0_0.vlm"},
	{0xF82B7C01CF774BBDULL, "216372428.ogg"},
	{0xF82BA4BE2A529891ULL, "w_bonepitmining_a2_42_0.rml"},
	{0xF82CA3B4B74EE6BEULL, "vxht_keep_n_5hi_0_0.vlm"},
	{0xF82D1C62F1F4CB2FULL, "hlac_ht_mansion_n_2995_1y.dds"},
	{0xF82D8470AB7AAA65ULL, "6149470_m.fxe"},
	{0xF82E23F5A2BA0944ULL, "6113784_m.fxe"},
	{0xF82E3675A0A4B18CULL, "vxlt_darktown_53f_0_0.vlm"},
	{0xF82E6148F0965FDDULL, "vxlt_warehouse_yq_0_0.vlm"},
	{0xF82E67A3008124ACULL, "450201392.ogg"},
	{0xF82E7132C37CE757ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wi_edge_2.mao"},
	{0xF82EA2DBC9B8EAC6ULL, "w_encounter1_2_0.rml"},
	{0xF82F17FD3CB6F628ULL, "art\\characters\\playercharacter\\humanmale\\hoo_roba.phy"},
	{0xF82F66C6E447A13CULL, "hlac_w_bonepitmining_a2_196609_0y.dds"},
	{0xF82FCA8FCD8F752DULL, "vxlt_0_lowtown_a2_n_c_hqz_0_0.vlm"},
	{0xF82FF4510E7C14BFULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_drain_1.mao"},
	{0xF830757D6B287AB5ULL, "6179895_m.fxe"},
	{0xF830877D0404AE9BULL, "rdr110dg_guard_warehouse6016704.cl"},
	{0xF830A171A0EBBE01ULL, "vxlt_0_lowtown_a2_gv9_0_0.vlm"},
	{0xF831268A5A09D209ULL, "vxlt_darktown_5im_0_0.vlm"},
	{0xF8315DBEA265CEE2ULL, "hlac_lt_0_lowtown_a2_-65537_2x.dds"},
	{0xF831EC19B1F9CC40ULL, "art\\characters\\creatures\\mubari\\animation\\howl.ani"},
	{0xF8326CEE2D302B53ULL, "w_sundermount_106_0.rml"},
	{0xF832766F94A83217ULL, "qun230cr_guild_contact6068840.cl"},
	{0xF832B101607A7CD9ULL, "722962438.ogg"},
	{0xF832B296C4B6D7EFULL, "323290068.ogg"},
	{0xF8330D666B47D3BEULL, "vxdr_primevalentrance_1o2_0_0.vlm"},
	{0xF834262B688A7186ULL, "6133376_f.fxe"},
	{0xF834A6B9C870AF97ULL, "6118225_m.fxe"},
	{0xF8356ECF29E517E3ULL, "gen00fl_carver6097313.cl"},
	{0xF8359BDBC9BEFD0AULL, "w_encounter1_2_0.phy"},
	{0xF835FDF0BDE0D076ULL, "6128765_f.fxe"},
	{0xF8360463AD0709F2ULL, "73050573.ogg"},
	{0xF8362859B5C13710ULL, "hlac_lt_gallowstemplar_f_7198_5y.dds"},
	{0xF83638DFAA32BF79ULL, "hlac_lt_0_lowtown_a2_65537_0z.dds"},
	{0xF83659F1417D28F3ULL, "vxht_0_hightown_a2_bfr_0_0.vlm"},
	{0xF837D2D0CBDF2608ULL, "vxlt_undercityencounter_1kd_0_0.vlm"},
	{0xF837D7CDE19027E1ULL, "gen_im_trs_apo_sta_cld_r2.uti"},
	{0xF837F840AF6F521EULL, "llac_lt_0_lowtown_n_0_6.dds"},
	{0xF838296EED98FB0AULL, "lt_0_lowtown_docks_n_145_0.mao"},
	{0xF8382D6EED990259ULL, "lt_0_lowtown_docks_n_145_0.mmh"},
	{0xF838376EED991347ULL, "lt_0_lowtown_docks_n_145_0.msh"},
	{0xF8385D0E6797F20EULL, "hlac_lt_gallowstemplar_n_2395_0x.dds"},
	{0xF8386188731E8D95ULL, "art\\vfx\\combat\\vfx_m_pri_petrify_loop_c.cess.ani"},
	{0xF8395CF12C4DE5C7ULL, "hlac_ca_caverns1_mines_24427_0x.dds"},
	{0xF83973EE2D3655ABULL, "w_sundermount_106_0.phy"},
	{0xF839F86C8B88DC28ULL, "vxlt_0_lowtown_a2_n_hgs_0_0.vlm"},
	{0xF83A25A077BD161FULL, "llac_ca_caverns2_5802_35.dds"},
	{0xF83AEB39E882BF9FULL, "hlac_lt_0_lowtown_n_131071_3y.dds"},
	{0xF83B040604CCB69EULL, "dae121st_viscount_1_alt.stg"},
	{0xF83B49511A47BB8AULL, "hlac_lt_0_lowtown_n_-196609_2x.dds"},
	{0xF83BD27EB147BCC8ULL, "hlac_lt_0_lowtown_a2_n_-262144_2x.dds"},
	{0xF83C129587B07E63ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\0wi_diagonal_1.mao"},
	{0xF83CAA8516BFE4DFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_lying_injured.anb"},
	{0xF83CD6475D640EE0ULL, "800271723.ogg"},
	{0xF83CD9ADB5117055ULL, "vxca_cavern1_haunted_iiw_0_0.vlm"},
	{0xF83D1A3BDB722B6AULL, "6128014_m.fxe"},
	{0xF83D22ACC4065F7CULL, "lt_0_lowtown_docks_a3_31_0_ll.mmh"},
	{0xF83D3DE7BFFA9C8BULL, "6150694_m.fxe"},
	{0xF83D53F77D4E8934ULL, "hlac_ht_playershouse_n_855_7x.dds"},
	{0xF83D64FB8CD4AFE2ULL, "6173380_m.fxe"},
	{0xF83E018BACDE017CULL, "l_mer241ar_caverns.wbd"},
	{0xF83E0CC7E732947BULL, "vxlt_0_lowtown_a3_n_c_hr4_0_0.vlm"},
	{0xF83E165EB6F944E4ULL, "vxw_sundermount_2kf_0_0.vlm"},
	{0xF83E3BA6B68EB6E1ULL, "643349626.ogg"},
	{0xF83EED95519ABF4DULL, "vxlt_0_lowtown_n_bz7_0_0.vlm"},
	{0xF83F1DDEC514DB75ULL, "vxht_0_hightown_a3_arz_0_0.vlm"},
	{0xF83F708259977F38ULL, "plt_lgt210pt_ingredients.plo"},
	{0xF83F98674BFFEA89ULL, "vxlt_darktown_5rl_0_0.vlm"},
	{0xF8400FA373CEE644ULL, "6139149_f.fxe"},
	{0xF8403F68763537F6ULL, "hlac_dr_0_deeproads_17035_0x.dds"},
	{0xF840E70840BF1808ULL, "598825111.ogg"},
	{0xF840F28A798FE360ULL, "ss_warrior_groundstab.cut"},
	{0xF84106141A3AC3A1ULL, "fx_mfx_dmg.dds"},
	{0xF8419B4E35E9ABA2ULL, "vxht_bartrand_n_4fi_0_0.vlm"},
	{0xF841F0C209521364ULL, "hlac_ca_caverns1_24424_23z.dds"},
	{0xF84289293038CDC7ULL, "672444198.ogg"},
	{0xF842D6A3DECEB413ULL, "vxht_mansion_n_4wd_0_0.vlm"},
	{0xF842F1AC666244AFULL, "6137489_f.fxe"},
	{0xF842FF6C5157341FULL, "14396026.ogg"},
	{0xF843041F71E457DCULL, "ave331cr_jeven.utc"},
	{0xF843542DE774BCD5ULL, "hlac_lt_0_lowtown_a2_131071_7z.dds"},
	{0xF84364D177748F3AULL, "mag360_alain6044659.cl"},
	{0xF843748118ECF130ULL, "6043806_f.fxe"},
	{0xF843E1065A20FD56ULL, "268075990.ogg"},
	{0xF844339EBF9C03FFULL, "vxdr_primevaltomb_7tu_0_1.vlm"},
	{0xF8443B661C54777DULL, "6148723_m.fxe"},
	{0xF84456BD2F7848B1ULL, "hlac_lt_0_lowtown_a2_n_-131072_2x.dds"},
	{0xF84563B84423006FULL, "914154090.ogg"},
	{0xF8461265F2FDD0FBULL, "vxca_cavern1_haunted_e7k_0_0.vlm"},
	{0xF84684973B1BBE19ULL, "6100341_m.fxe"},
	{0xF84728B679A21EB7ULL, "gen_im_trs_dem_lgt_cht_r1.uti"},
	{0xF84766122297D49BULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.impa.evt"},
	{0xF8476DC37B9EBB39ULL, "vxht_0_hightown_a3_c1m_0_0.vlm"},
	{0xF8478A33C90E0157ULL, "zz_mer_debug6062705.cl"},
	{0xF848A1FD29BF4291ULL, "llac_lt_0_lowtown_a3_n_c_-1_0.dds"},
	{0xF848B3500DECAA7AULL, "385791388.ogg"},
	{0xF849F123C66A60E0ULL, "ss_spellcombo.cut"},
	{0xF84A2BBD24B14072ULL, "vis121_saemus6101560.cl"},
	{0xF84B4D4F119BD2A7ULL, "follower_banter6166461.cl"},
	{0xF84BD49FB23771F4ULL, "vxht_mansion2b_af_0_0.vlm"},
	{0xF84BDC6EA472EF86ULL, "6043322_m.fxe"},
	{0xF84BEA9FE0E157C6ULL, "gen00fl_varric6105792.cl"},
	{0xF84C38867789B9A0ULL, "730835842.ogg"},
	{0xF84C883BA8EF5D64ULL, "vxw_sundermount_2lg_0_0.vlm"},
	{0xF84CC2C6E01E0354ULL, "vxw_bonepitdestroyed_a3_29o_0_0.vlm"},
	{0xF84E12478D7729C7ULL, "hlac_w_bonepit_458758_0x.dds"},
	{0xF84E4FD9FDA86F7AULL, "gen_im_trs_gen_coinbag_02.uti"},
	{0xF84E5F12229DE7DCULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.impa.gad"},
	{0xF84EBC55E6B968C7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\0wi_basic_2.mao"},
	{0xF84FCDD79F29C193ULL, "hlac_lt_0_lowtown_n_65538_1y.dds"},
	{0xF84FD86C7302E039ULL, "hf_broth_fake_beth.mor"},
	{0xF84FF43FA5D81CD6ULL, "719355592.ogg"},
	{0xF8500D2B27DFC422ULL, "vxht_bartrand_n_4kj_0_0.vlm"},
	{0xF8504C3D0BAF3E80ULL, "6195870_m.fxe"},
	{0xF8504EB413A86540ULL, "734986304.ogg"},
	{0xF850A5EE7A02B70FULL, "133968805.ogg"},
	{0xF850BA175F3A448EULL, "hlac_w_sundermount_-196605_0y.dds"},
	{0xF850CB09AF211510ULL, "gen00fl_anders6119274.cl"},
	{0xF8510C71B8853B0EULL, "862112176.ogg"},
	{0xF851F375F45DC3D0ULL, "vxht_brothel_n_5h8_0_0.vlm"},
	{0xF852A6078EBF9F88ULL, "vxht_0_hightown_c3n_0_1.vlm"},
	{0xF852E2BC9DC10EE0ULL, "hlac_lt_backalley_g_65537_2x.dds"},
	{0xF8533DAB256959F3ULL, "vxht_0_hightown_a2_blt_0_0.vlm"},
	{0xF853650C5EFC582DULL, "vxca_caverns2prc_a9l_0_0.vlm"},
	{0xF85389C330D71A55ULL, "vxlt_0_lowtown_a3_n_c_e89_0_0.vlm"},
	{0xF853CD1024286CBFULL, "hlac_lt_gallowscourtyard_131069_0y.dds"},
	{0xF853F07404B54A19ULL, "6086189_f.fxe"},
	{0xF8545E1222A251DAULL, "art\\vfx\\combat\\vfx_m_arc_b_proj_p.impa.ani"},
	{0xF8550C01E9FF048FULL, "6130087_m.fxe"},
	{0xF8572E32F4716337ULL, "780975296.ogg"},
	{0xF8574EA356E32E2DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\pose_urgent_run.anb"},
	{0xF857649F00D2A239ULL, "art\\vfx\\cinematic\\vfx_bld_dir_p.mmh"},
	{0xF857CCE5A02589C2ULL, "dae000_bartrand6206955.cl"},
	{0xF857F32514D657C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\animation\\blendtree\\floortrap.anb"},
	{0xF8582393949C7D22ULL, "vis121_saemus6101810.cl"},
	{0xF8583DA0A08F5CE1ULL, "941902462.ogg"},
	{0xF85958CD5B8EBE9FULL, "482957653.ogg"},
	{0xF859B0B35512E28FULL, "hlac_lt_gallowstemplar_2388_2z.dds"},
	{0xF859BD6798FDA155ULL, "vxca_cavern1_haunted_icu_0_0.vlm"},
	{0xF859C03025073A00ULL, "6033598_m.fxe"},
	{0xF859C3EE95D23701ULL, "6103946_m.fxe"},
	{0xF859F5790AC5F3C1ULL, "6172821_m.fxe"},
	{0xF85A650EF8A54BDEULL, "6133542_m.fxe"},
	{0xF85A749184C4DC34ULL, "gen_im_acc_amu_act3_10.uti"},
	{0xF85ADA9523B61F57ULL, "dae000_orsino6081373.cl"},
	{0xF85AF081CB1EC57BULL, "vxlt_0_lowtown_a3_n_c_hd6_0_0.vlm"},
	{0xF85B60774120575FULL, "6188116_m.fxe"},
	{0xF85B8C4F56D503F4ULL, "vxht_keep_n_4b8_0_0.vlm"},
	{0xF85BA20112C720CFULL, "6117762_f.fxe"},
	{0xF85C7C212FEC1B89ULL, "vxlt_darktown_5xr_0_0.vlm"},
	{0xF85D690E4ED26E80ULL, "mag210_arianni6031702.cl"},
	{0xF85DB1E0F1F245EFULL, "hlac_ht_mansion2b_818_58z.dds"},
	{0xF85E5238AD8C680EULL, "alphamagicplate01.mao"},
	{0xF85EB751C82AA44EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaclosed.mmh"},
	{0xF85EC351C82AB8E5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaclosed.mao"},
	{0xF85ED151C82AD03CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_tomeaclosed.msh"},
	{0xF85F0006B36D0A4BULL, "539371865.ogg"},
	{0xF85F02F0F5484523ULL, "gen00fl_fenris6110843.cl"},
	{0xF85F24FBE5CAA2E4ULL, "707475277.ogg"},
	{0xF85F5D1B90E84E6AULL, "vxht_mansion2b_32j_0_0.vlm"},
	{0xF85FA1D4F9F3B507ULL, "225176130.ogg"},
	{0xF8608099679FA66FULL, "zz_fen_debug6165489.cl"},
	{0xF8610F0CCF6B05A9ULL, "422371619.ogg"},
	{0xF8619500F9412DBCULL, "431757694.ogg"},
	{0xF8619B0FE42A3D74ULL, "vxw_bonepitdestroyed_a3_1v1_0_0.vlm"},
	{0xF86243BE609F9DFBULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfalll_1.mao"},
	{0xF8627A50D99A6679ULL, "mcr310_confused_guard6077052.cl"},
	{0xF862A0A874BF5028ULL, "vxlt_backalley_g_1w9_0_1.vlm"},
	{0xF86316BFFC521F6FULL, "6114293_f.fxe"},
	{0xF86326E03F6BF448ULL, "88999292.ogg"},
	{0xF86366E01DC97931ULL, "resistances.gda"},
	{0xF8636D5ED8020ADEULL, "692076578.ogg"},
	{0xF86422DFDC876E7BULL, "6119658_f.fxe"},
	{0xF8643E72B567EBC6ULL, "vxca_cavern1_haunted_j28_0_2.vlm"},
	{0xF864C8641D9F90B1ULL, "character_character_p57.psh.fxo.dx11.assembly"},
	{0xF864EBA83AEF55ABULL, "l_lgt311ar_gang_hideout.wbd"},
	{0xF865212BAB625BCFULL, "235968466.ogg"},
	{0xF86526A604843E4CULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1d.msh"},
	{0xF8652CA60484483EULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1d.mmh"},
	{0xF86538A604845CD5ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1d.mao"},
	{0xF8654EACF21C177DULL, "51446591.ogg"},
	{0xF8657677DEDAC257ULL, "vxw_encounter3_6j_0_0.vlm"},
	{0xF86579BF159BE2EDULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_overlay.anb"},
	{0xF865A8947ED92D9BULL, "vxht_0_hightown_a3_ass_0_0.vlm"},
	{0xF865ECC4E4DB2827ULL, "130820816.ogg"},
	{0xF8660D3DB163E915ULL, "vxca_caverns1_mines_cn4_0_0.vlm"},
	{0xF86615C786ECE676ULL, "zz_pro_debug6059758.cl"},
	{0xF866D6310425E51CULL, "vxlt_darktown_4nt_0_0.vlm"},
	{0xF8676192C0FBD7ADULL, "vxlt_warehouse_154_0_0.vlm"},
	{0xF86767CE007CCEA9ULL, "gen_im_wep_rog_arc_act3_03.uti"},
	{0xF86798B1BC389F82ULL, "gen00fl_bethany6007744.cl"},
	{0xF867BAAEFB66736DULL, "39030350.ogg"},
	{0xF867D7351F4E582CULL, "mag210_arianni6030459.cl"},
	{0xF867F4F578873421ULL, "llac_w_sundermount_-5_0.dds"},
	{0xF8691F30014A0850ULL, "gen00fl_sebastian6188655.cl"},
	{0xF869324EC938FB84ULL, "6020925_m.fxe"},
	{0xF8696BF58CDB8E64ULL, "vxw_sundermount_2ze_0_0.vlm"},
	{0xF869BF8D5E295EE8ULL, "testspritearmy.dds"},
	{0xF86A854CE5F31EC9ULL, "vxlt_hangedman_21k_0_0.vlm"},
	{0xF86AFF8BAD04EBF6ULL, "l_mer241ar_caverns.bnk"},
	{0xF86B474FC092FB85ULL, "vxdr_slavers_14h_0_0.vlm"},
	{0xF86B7864A8B36184ULL, "hlac_lt_gallowstemplar_n_2394_12z.dds"},
	{0xF86BE584FB890A18ULL, "vxlt_backalley_n_1ew_0_0.vlm"},
	{0xF86C6CF1E9B5125DULL, "vxw_encounter3_8b_0_0.vlm"},
	{0xF86D27431449C8D2ULL, "669684399.ogg"},
	{0xF86DFEA12C71EDC2ULL, "gen00fl_fenris6126481.cl"},
	{0xF86E04E109FA9E8DULL, "nat380_nathaniel6066721.cl"},
	{0xF86E2C3AB4B0FD93ULL, "vxht_mansion_n_4di_0_0.vlm"},
	{0xF86E3FC10886CBEDULL, "mag320_orsino6071259.cl"},
	{0xF86ED3CCD831626CULL, "llac_lt_0_lowtown_docks_a3_131072_1.dds"},
	{0xF86EF1A96B3BA222ULL, "hlac_ca_caverns2_5802_18x.dds"},
	{0xF86F521508A8F80FULL, "vxlt_gallowscourtyard_6p5_0_0.vlm"},
	{0xF86F8F827F1B80FDULL, "vxw_woundedcoast_bz_0_0.vlm"},
	{0xF86F9FC2F865F5C2ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeexit.gad"},
	{0xF86FE842061D1A51ULL, "515398741.ogg"},
	{0xF87015BBB5B5038CULL, "115830570.ogg"},
	{0xF87033C8389774F1ULL, "qun200_party_comment6098544.cl"},
	{0xF8704917278BD242ULL, "w_sundermount_107_0.rml"},
	{0xF8711323B26AD3C6ULL, "llac_w_woundedcoast_327683_0.dds"},
	{0xF8711D2CE93FA6B1ULL, "vxlt_0_lowtown_n_eow_0_0.vlm"},
	{0xF8713DE2A72B3910ULL, "mrl000_marethari6129287.cl"},
	{0xF872AC9288BB3B81ULL, "qun160_vashothlead6053927.cl"},
	{0xF872E511DEE4E3FCULL, "583524021.ogg"},
	{0xF8731C86F0DA16B2ULL, "377736689.ogg"},
	{0xF8734F60D23093AEULL, "gen_im_arm_cht_nobe.uti"},
	{0xF8737714C8BAAB77ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_keep.phy"},
	{0xF873A1BB02D6CA66ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\heavydamage.ani"},
	{0xF873E854D0D0DAD7ULL, "vxw_encounter3_3e_0_0.vlm"},
	{0xF873F8DD9367FF3AULL, "6041517_m.fxe"},
	{0xF8745429D8A1CDDEULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.cess.evt"},
	{0xF874699632C004D1ULL, "vxlt_0_lowtown_ewl_0_0.vlm"},
	{0xF874C3555CF42BD5ULL, "hlac_ca_caverns2prc_5802_34x.dds"},
	{0xF874CD9D9004D75AULL, "mag291_ntbanter6056609.cl"},
	{0xF87509C613A79477ULL, "131542825.ogg"},
	{0xF8753667B466BADDULL, "isa311ar_castillon_base.lst"},
	{0xF8754B8CCCD732F9ULL, "sys_comboeffects.ncs"},
	{0xF8757C647798128DULL, "243414816.ogg"},
	{0xF87587DF0BC95B1BULL, "153967095.ogg"},
	{0xF87603FB367426BBULL, "llac_lt_undercityserialkiller_-1_3.dds"},
	{0xF876175A72530D51ULL, "1024005603.ogg"},
	{0xF876AD6D56BE081EULL, "520241996.ogg"},
	{0xF876C2C2F86C4FB9ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeexit.evt"},
	{0xF8774A172791F2F6ULL, "w_sundermount_107_0.phy"},
	{0xF8782560933F7182ULL, "vxw_encounter3_d5_0_0.vlm"},
	{0xF87825C7FD8DDB33ULL, "6103648_f.fxe"},
	{0xF8789C3F0DB65DC7ULL, "326030454.ogg"},
	{0xF8789EB9FD4E3A8AULL, "hlac_lt_0_lowtown_docks_a3_n_131072_0x.dds"},
	{0xF8792B3993A32B8EULL, "lt_0_lowtown_docks_n_230_0.mmh"},
	{0xF879373993A34025ULL, "lt_0_lowtown_docks_n_230_0.mao"},
	{0xF879453993A3587CULL, "lt_0_lowtown_docks_n_230_0.msh"},
	{0xF879C5D576A5015DULL, "vxlt_blackemporium_17r_0_0.vlm"},
	{0xF879DF119913506DULL, "zz_dae_debug6029919.cl"},
	{0xF87A01BB58E16A83ULL, "6106994_m.fxe"},
	{0xF87A55A7B9CE27C1ULL, "hlac_lt_darktown_-1_81z.dds"},
	{0xF87A55EB39F56C1FULL, "mer340_cowan6053779.cl"},
	{0xF87ADECEDBAB2ADDULL, "vxw_encounter3_5a_0_0.vlm"},
	{0xF87B0151D4DC588DULL, "6146912_m.fxe"},
	{0xF87B6B29D8A81309ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.cess.gad"},
	{0xF87B7D449066F5AFULL, "hlac_lt_warehouse_n_65536_0z.dds"},
	{0xF87B9585796B0BF3ULL, "vxlt_0_lowtown_a2_hgw_0_0.vlm"},
	{0xF87C2A3A9DA44D1FULL, "dr_primevaltomb_663_ll.mmh"},
	{0xF87C2E67B46CCBCCULL, "isa311ar_castillon_base.ncs"},
	{0xF87CC264D14BFE86ULL, "hlac_lt_backalley_g_65537_3y.dds"},
	{0xF87D1EAE3C5483F3ULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_c.anb"},
	{0xF87D2436A7BF1D52ULL, "500680513.ogg"},
	{0xF87D3F9F326BAA67ULL, "qun221_viscount6070144.cl"},
	{0xF87DA3491DD8AC98ULL, "dae000_meredith6187898.cl"},
	{0xF87DC00CC5B47E05ULL, "802092592.ogg"},
	{0xF87EBB7667FC62D8ULL, "hlac_lt_0_lowtown_a3_n_c_-65536_2y.dds"},
	{0xF87ECC00034D4942ULL, "hlac_ht_0_hightown_0_3z.dds"},
	{0xF87ED3EE17E93CA8ULL, "6059782_f.fxe"},
	{0xF87F619F00F3EF2EULL, "art\\vfx\\cinematic\\vfx_bld_dir_p.anb"},
	{0xF87F846258AB8128ULL, "vxlt_backalley_g_1y7_0_1.vlm"},
	{0xF87FB7F2C5BDF20BULL, "vxlt_0_lowtown_docks_n_d0t_0_0.vlm"},
	{0xF87FC0BA70A010ACULL, "vxlt_undercityserialkiller_259_0_0.vlm"},
	{0xF88037209038A85AULL, "hlac_w_woundedcoast_327683_0z.dds"},
	{0xF88085EDB7D87FE6ULL, "gen00fl_aveline6132786.cl"},
	{0xF880AFE000705B2FULL, "543714250.ogg"},
	{0xF881277DD5C7999FULL, "6214927_m.fxe"},
	{0xF8813317279A1F97ULL, "w_sundermount_107_0.msh"},
	{0xF8814517279A3EFAULL, "w_sundermount_107_0.mao"},
	{0xF8814917279A4509ULL, "w_sundermount_107_0.mmh"},
	{0xF88189BB02E2DED7ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\heavydamage.evt"},
	{0xF8818CD2E99A313BULL, "278710901.ogg"},
	{0xF881AD0CCB2ED491ULL, "691423288.ogg"},
	{0xF881CE223D8E8BC5ULL, "6081587_m.fxe"},
	{0xF882306D0CDE0CBAULL, "hm_danzig.mor"},
	{0xF8823C29D8ADE257ULL, "art\\vfx\\environment\\vfx_blt_forcefield_p.cess.ani"},
	{0xF8826819583FCC61ULL, "gen_im_trs_gen_med_hlm_r0.uti"},
	{0xF883141553D60543ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08a.msh"},
	{0xF883221553D61C9DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye08a.mmh"},
	{0xF88326E6AA5B45D1ULL, "6173057_m.fxe"},
	{0xF8832E0820A9CEEDULL, "739193764.ogg"},
	{0xF8833E72BF4CDD1FULL, "lt_0_lowtown_n_mmi.gff"},
	{0xF883CAC2F876E7A8ULL, "art\\characters\\creatures\\darkspawnogre\\animation\\combat\\chargeexit.ani"},
	{0xF884DA9B894368E3ULL, "hlac_lt_0_lowtown_a2_n_c_1_6x.dds"},
	{0xF885163770651A79ULL, "zz_dae_debug6029818.cl"},
	{0xF88596E366D8DBF8ULL, "vxht_0_hightown_n_ary_0_0.vlm"},
	{0xF885BA199BB64B9CULL, "198329907.ogg"},
	{0xF885CDC75F9FCFEEULL, "vxht_mansion2b_1wc_0_0.vlm"},
	{0xF8864934C7EA8CC3ULL, "llac_w_sundermount_-65532_0.dds"},
	{0xF8866AE63E07B2F2ULL, "genip_cin_nice_chair.utp"},
	{0xF8866BE52C65B914ULL, "vxca_caverns2_7la_0_0.vlm"},
	{0xF8867EDCF1AFC2CCULL, "6001008_f.fxe"},
	{0xF8868415669122F4ULL, "845934478.ogg"},
	{0xF886973D85358A02ULL, "vxw_encounter3_a6_0_0.vlm"},
	{0xF886B962F15617A7ULL, "188441975.ogg"},
	{0xF886C2B036D5D528ULL, "art\\vfx\\models\\vfx_dragon.mmh"},
	{0xF886E0B036D60852ULL, "art\\vfx\\models\\vfx_dragon.msh"},
	{0xF886E8F32782A61AULL, "6069938_f.fxe"},
	{0xF88829C562FB6EBFULL, "6129724_m.fxe"},
	{0xF8884230B66A8BC5ULL, "vxlt_blackemporium_z8_0_0.vlm"},
	{0xF8885C3D964F173EULL, "gen00fl_isabela_init6110272.cl"},
	{0xF8888ABB02E8FF60ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\arcanehorror\\heavydamage.gad"},
	{0xF8888B61C6299236ULL, "dae351ar_mount_caverns.are"},
	{0xF888AA3E120649BEULL, "vxw_bonepit_25y_0_0.vlm"},
	{0xF888C93EDF753B18ULL, "vxlt_backalley_n_1ku_0_0.vlm"},
	{0xF889276E78A91E06ULL, "6113588_m.fxe"},
	{0xF88927E7EC688E3CULL, "vxlt_undercityserialkiller_1qn_0_0.vlm"},
	{0xF88954B6C7A36FEEULL, "785880219.ogg"},
	{0xF889CDA83B0E3F79ULL, "l_lgt311ar_gang_hideout.bnk"},
	{0xF889E60DA80B91B6ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzdeadbody.mao"},
	{0xF889F1AF9BD65ED6ULL, "gen_im_trs_gen_dua_dag_r0.uti"},
	{0xF889F20DA80BA655ULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzdeadbody.mmh"},
	{0xF889F40DA80BA9CBULL, "art\\levels\\races\\proxy\\props\\corpses\\prp_zzdeadbody.msh"},
	{0xF88A161E36BB2E38ULL, "art\\levels\\races\\proxy\\_textures\\nodiffuse.dds"},
	{0xF88A5B6343DA62F2ULL, "6089770_f.fxe"},
	{0xF88B220B7D8C3D4AULL, "vxca_caverns1_mines_ge2_0_0.vlm"},
	{0xF88B3BAE3C60F25CULL, "art\\vfx\\combat\\vfx_m_sta_spi_range_c.mmh"},
	{0xF88CA92F5F4DCB83ULL, "326565860.ogg"},
	{0xF88CE84FFEFF65E0ULL, "6082680_m.fxe"},
	{0xF88D0BBE0E18E879ULL, "6075768_m.fxe"},
	{0xF88D1D14CEF4E263ULL, "w_bonepitmining_a2_223_0.msh"},
	{0xF88D2B14CEF4F9BDULL, "w_bonepitmining_a2_223_0.mmh"},
	{0xF88D2F14CEF500CEULL, "w_bonepitmining_a2_223_0.mao"},
	{0xF88D56D27C63B30BULL, "character_character_p46.psh.fxo.dx11"},
	{0xF88DA97C0A9A96D9ULL, "hlac_dr_dungeon_3217_2x.dds"},
	{0xF88DC64FAA758101ULL, "6149856_m.fxe"},
	{0xF88DFA05282F1F77ULL, "w_encounter3_169_0.mmh"},
	{0xF88DFE05282F2584ULL, "w_encounter3_169_0.mao"},
	{0xF88E1005282F44E9ULL, "w_encounter3_169_0.msh"},
	{0xF88EDEEABA2CE211ULL, "vxca_caverns2_8ml_0_0.vlm"},
	{0xF88F059F1C5AD6F9ULL, "280374328.ogg"},
	{0xF88F3504BAF7DF72ULL, "hlac_lt_0_lowtown_a3_n_c_65534_3y.dds"},
	{0xF88FB03C256F0CF3ULL, "92885461.ogg"},
	{0xF88FB437BBB5D2A7ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_urva_0d.dds"},
	{0xF88FD27AEA2C20ADULL, "mag311_nyssa6042235.cl"},
	{0xF8900208BA582D71ULL, "vxlt_darktown_52u_0_1.vlm"},
	{0xF890A82CFF2EB08EULL, "6217695_m.fxe"},
	{0xF89143E4FAB48DA4ULL, "dae200cr_viscount6073933.cl"},
	{0xF89160C01E369748ULL, "6095679_m.fxe"},
	{0xF891696C8A5A82F3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_interact_low.ani"},
	{0xF891920404C9A368ULL, "zz_one_debug6074008.cl"},
	{0xF89219190B5CB8DAULL, "601648098.ogg"},
	{0xF8923A6B2D10704DULL, "6163363_f.fxe"},
	{0xF89266F807335961ULL, "270834603.ogg"},
	{0xF8930092B535C11CULL, "6171267_m.fxe"},
	{0xF893187110932136ULL, "hlac_lt_0_lowtown_a3_n_c_-65537_4z.dds"},
	{0xF89325702D99CE67ULL, "hlac_lt_shop_-1_3x.dds"},
	{0xF8936FFBAE3320D7ULL, "vxht_mansion_n_632_0_0.vlm"},
	{0xF8939BD6AB0E8C3BULL, "hlac_dr_0_deeproads_5849_6y.dds"},
	{0xF893FDA74B21E959ULL, "1041223017.ogg"},
	{0xF8940486D751F9F3ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_fireplace.dura.evt"},
	{0xF8943039299A3AEEULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_1.ani"},
	{0xF89438863B014A0AULL, "vxlt_0_lowtown_docks_n_cum_0_0.vlm"},
	{0xF8949D7FD2CB5EB9ULL, "hlac_lt_0_lowtown_n_65538_0x.dds"},
	{0xF895A49F82A0767FULL, "6125815_f.fxe"},
	{0xF895C55877A13C72ULL, "6077358_m.fxe"},
	{0xF8969F3336D7D88CULL, "plt_cod_ltr_sk_shipment.plo"},
	{0xF89719199BE2F531ULL, "572436662.ogg"},
	{0xF8971C1B03FC623EULL, "vxw_bonepit_20z_0_0.vlm"},
	{0xF8973B1BD16B5398ULL, "vxlt_backalley_n_1vr_0_0.vlm"},
	{0xF89771A74F151B49ULL, "zz_approval_debug6003763.cl"},
	{0xF89801F84EF29D5BULL, "hlac_lt_0_lowtown_a2_n_65536_1y.dds"},
	{0xF898165A4F405944ULL, "6071219_m.fxe"},
	{0xF898367F5782AF19ULL, "vxca_caverns1_icm_0_1.vlm"},
	{0xF898AF6C8DB1A2FDULL, "zz_isa_debug6133965.cl"},
	{0xF8996CD2465C9423ULL, "t3_har_bln.tnt"},
	{0xF8998ABC0908ADC2ULL, "t3_fat_f03.tnt"},
	{0xF89995D622BE4591ULL, "lt_0_lowtown_a2_n_c_24_0.rml"},
	{0xF89A1B59D40940DBULL, "6201478_f.fxe"},
	{0xF89A75734F7A4940ULL, "257695389.ogg"},
	{0xF89AFD86D7580C34ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_fireplace.dura.gad"},
	{0xF89B09EE8B9D49B4ULL, "hlac_ht_0_hightown_-1_0y.dds"},
	{0xF89B4573757019E6ULL, "mrl310_merrill6202142.cl"},
	{0xF89B5A2CF0F57E60ULL, "6085780_m.fxe"},
	{0xF89C52124CA81569ULL, "6027532_m.fxe"},
	{0xF89C899EB4C02403ULL, "6033540_m.fxe"},
	{0xF89CA43CBA370F86ULL, "6214791_m.fxe"},
	{0xF89D4771FFC72E11ULL, "vxht_0_hightown_bgg_0_0.vlm"},
	{0xF89D76B579330988ULL, "vxht_bartrand_n_5r2_0_0.vlm"},
	{0xF89D843336DDC9C3ULL, "plt_cod_ltr_sk_shipment.nss"},
	{0xF89DBEA2D7AF642CULL, "lgt000ip_ultra_hm_remains.ncs"},
	{0xF89E31A44C88302FULL, "vxw_encounter3_fr_0_0.vlm"},
	{0xF89E38965634B356ULL, "docks_fem_sailor_mob6176427.cl"},
	{0xF89E5A831D61049AULL, "6171409_m.fxe"},
	{0xF89E616C8A64FF2AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_interact_low.evt"},
	{0xF89E9FA67BD32470ULL, "hlac_lt_warehouse_0_0y.dds"},
	{0xF89FBF2D08A24E51ULL, "vxlt_0_lowtown_eha_0_0.vlm"},
	{0xF89FDBA8EEFCC0CCULL, "6162977_m.fxe"},
	{0xF8A0211E395D5637ULL, "6019092_m.fxe"},
	{0xF8A05F01928F0056ULL, "dae200_bran6092467.cl"},
	{0xF8A0DB8AC292875FULL, "vxht_bartrand_n_3qv_0_0.vlm"},
	{0xF8A10C86D75C9112ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_fireplace.dura.ani"},
	{0xF8A122580289C892ULL, "hlac_lt_0_lowtown_docks_65535_0z.dds"},
	{0xF8A1283929A4B73FULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_1.evt"},
	{0xF8A13F67B48C04BAULL, "isa311ar_castillon_base.are"},
	{0xF8A13F8FEC00B6C4ULL, "vxht_chantry_30n_0_0.vlm"},
	{0xF8A16E9C84D1A2ABULL, "295032385.ogg"},
	{0xF8A1726FA72BD99CULL, "6174167_m.fxe"},
	{0xF8A1E8FF815B4EB6ULL, "vxlt_0_lowtown_docks_n_b4g_0_0.vlm"},
	{0xF8A1F5B9F8A9137BULL, "vxca_caverns2prc_a0q_0_0.vlm"},
	{0xF8A2806AEA04A12AULL, "hlac_lt_gallowstemplar_f_2394_7z.dds"},
	{0xF8A287338ECC2473ULL, "hlac_ca_caverns1_mines_24423_13x.dds"},
	{0xF8A30CC78427A53FULL, "6068507_f.fxe"},
	{0xF8A3325C368226D6ULL, "14918287.ogg"},
	{0xF8A3CEE9AB17E92BULL, "vxlt_ruinedhouse_n_87_0_0.vlm"},
	{0xF8A4167C74968EFFULL, "6122915_f.fxe"},
	{0xF8A4956D25C710DDULL, "vxlt_0_lowtown_a2_eub_0_0.vlm"},
	{0xF8A4DF708B2693D6ULL, "17052803.ogg"},
	{0xF8A524B4F1160367ULL, "398918863.ogg"},
	{0xF8A5406C8A6AE59DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\crouch_squat_interact_low.gad"},
	{0xF8A54A36FD3E812CULL, "6090434_m.fxe"},
	{0xF8A59936C48577F0ULL, "vxw_encounter3_g3_0_1.vlm"},
	{0xF8A5A1B2DE4BE253ULL, "vxw_encounter1_bx_0_0.vlm"},
	{0xF8A603A604BB99D9ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bazaar1d.phy"},
	{0xF8A649C453C5C74CULL, "183305574.ogg"},
	{0xF8A758195101F84FULL, "hlac_ca_cavern1_haunted_24424_10x.dds"},
	{0xF8A75A933F89217BULL, "372809790.ogg"},
	{0xF8A7A350475A14D2ULL, "58618431.ogg"},
	{0xF8A7E06F9F1A9B43ULL, "w_bonepit_11_0_ll.mmh"},
	{0xF8A8092FF53FD752ULL, "6171544_m.fxe"},
	{0xF8A8093929AAA128ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\custom\\fenris_interrogation_reaction_head_smash_1.gad"},
	{0xF8A80CB30D973269ULL, "6075135_m.fxe"},
	{0xF8A87602484CC338ULL, "vxca_cavern1_haunted_e92_0_0.vlm"},
	{0xF8A8AF68DFD741A3ULL, "mag371_elsa6041192.cl"},
	{0xF8A8B19C2AF82479ULL, "gencr_gate_slave_l2.utc"},
	{0xF8AA0940C6F3A2F2ULL, "348535837.ogg"},
	{0xF8AA656876A0C1DCULL, "vxlt_gallowstemplar_f_6sz_0_0.vlm"},
	{0xF8AB12BBD36EB710ULL, "w_sundermount_207_0.msh"},
	{0xF8AB20BBD36ECF2AULL, "w_sundermount_207_0.mmh"},
	{0xF8AB24BBD36ED661ULL, "w_sundermount_207_0.mao"},
	{0xF8AB7D6FC2BB11DAULL, "l_one130ar_booker_dungeon.bnk"},
	{0xF8AB939CABD3FA35ULL, "625956871.ogg"},
	{0xF8AB9C5BE542FAA9ULL, "hlac_ht_mansion2b_818_27z.dds"},
	{0xF8ABA3C858398684ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wk_p_rf.ani"},
	{0xF8ABC2A49E191311ULL, "vxca_caverns2_8wb_0_0.vlm"},
	{0xF8ABCF79F3B19098ULL, "21174311.ogg"},
	{0xF8ABE8926B292208ULL, "vxht_bartrand_n_5u3_0_0.vlm"},
	{0xF8ABED6B28141EA2ULL, "lt_0_lowtown_docks_a3_n_213_0.msh"},
	{0xF8ABEF6B28142258ULL, "lt_0_lowtown_docks_a3_n_213_0.mmh"},
	{0xF8ABFB6B281436FBULL, "lt_0_lowtown_docks_a3_n_213_0.mao"},
	{0xF8ACABD6A44336FAULL, "follower_banter6172874.cl"},
	{0xF8ACDD24A9389194ULL, "zz_ave_debug6064069.cl"},
	{0xF8AD1944C1D5B5EFULL, "6089905_m.fxe"},
	{0xF8AD8164F660DD23ULL, "vxw_bonepitdestroyed_a3_9t_0_0.vlm"},
	{0xF8AD9261C648BA01ULL, "dae351ar_mount_caverns.lst"},
	{0xF8ADCB6EC2AFC811ULL, "one140cr_charade6128520.cl"},
	{0xF8AE448129D2AFE3ULL, "42869204.ogg"},
	{0xF8AE5F0B13B40FA3ULL, "hlac_lt_gallowscourtyard_-2_2x.dds"},
	{0xF8AE7E6E052881AFULL, "vxht_keep_6bf_0_0.vlm"},
	{0xF8AE8BAF670E962EULL, "vxca_cavern1_haunted_k6k_0_0.vlm"},
	{0xF8AEE6E615CDDE28ULL, "112352374.ogg"},
	{0xF8AFBF760E7831AEULL, "vxw_encounter2_camp_iu_0_0.vlm"},
	{0xF8AFE5876C0B0486ULL, "dae000_fem_priest_mob6174189.cl"},
	{0xF8B052BBA85203BCULL, "392860822.ogg"},
	{0xF8B059B54C544FA1ULL, "hlac_lt_0_lowtown_a2_1_2y.dds"},
	{0xF8B08F042B4D8212ULL, "vxht_keep_n_3fy_0_0.vlm"},
	{0xF8B0ABF666274A57ULL, "isa211_isabela6202200.cl"},
	{0xF8B118F3C3389B88ULL, "mag320_orsino6071234.cl"},
	{0xF8B13BE90631EAA1ULL, "747060128.ogg"},
	{0xF8B1DEA55C15D390ULL, "329817839.ogg"},
	{0xF8B21142EC7816BCULL, "6082598_m.fxe"},
	{0xF8B2266E9870C9B1ULL, "6120079_m.fxe"},
	{0xF8B2897A9B72B2E5ULL, "hlac_ht_mansion2b_818_53y.dds"},
	{0xF8B28D4193BA99C3ULL, "fadeclouds.vsh.fxo"},
	{0xF8B2A0C8583FA0E6ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wk_p_rf.gad"},
	{0xF8B3074A17BD295DULL, "vxlt_0_lowtown_a2_exa_0_0.vlm"},
	{0xF8B386BB7079647CULL, "71073497.ogg"},
	{0xF8B3DF6167949B60ULL, "6165521_m.fxe"},
	{0xF8B5A2858A585200ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_dw_hammer_0_n.dds"},
	{0xF8B5B5FCB558FA38ULL, "dae000_gamlen6162116.cl"},
	{0xF8B6242B0D279CF4ULL, "vxdr_primevaltomb_8f6_0_0.vlm"},
	{0xF8B6A64F35BCECD7ULL, "zz_followers_banter6159519.cl"},
	{0xF8B6E39597ADCFE9ULL, "349988361.ogg"},
	{0xF8B711274B46FCEFULL, "6110445_m.fxe"},
	{0xF8B76EC506F09A77ULL, "144410182.ogg"},
	{0xF8B7A437B84FEE32ULL, "dae000_orana6149073.cl"},
	{0xF8B84C8AB80AF9C5ULL, "gen00fl_fenris6113884.cl"},
	{0xF8B84F84BFC75CDFULL, "vxlt_gallowscourtyard_n_4wi_0_0.vlm"},
	{0xF8B88A68A5840813ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_basic.phy"},
	{0xF8B89BC8584403C5ULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_wk_p_rf.evt"},
	{0xF8B932BACD4D74F8ULL, "309584874.ogg"},
	{0xF8B98E9DF11412E9ULL, "735133128.ogg"},
	{0xF8B99AA07893EFC7ULL, "6067930_m.fxe"},
	{0xF8BA5A6F5D1F3A88ULL, "vxht_bartrand_n_5x0_0_0.vlm"},
	{0xF8BB4C14C8F817CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_keep.mao"},
	{0xF8BB5814C8F82C6CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_keep.mmh"},
	{0xF8BB5E14C8F8365EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\0rl_keep.msh"},
	{0xF8BB6A01BFAA4877ULL, "6099203_m.fxe"},
	{0xF8BC26835D49C2E4ULL, "plt_lgt000pt_ultra_5_03.nss"},
	{0xF8BC59574914AE9DULL, "vxlt_0_lowtown_a2_n_c_i0t_0_0.vlm"},
	{0xF8BCA5AB36A41BB5ULL, "vxlt_warehouse_110_0_0.vlm"},
	{0xF8BD1E773B3AAD67ULL, "6155688_m.fxe"},
	{0xF8BD3C7EA7B744A5ULL, "356474387.ogg"},
	{0xF8BD69EA1C1F1BA0ULL, "qcr220_orsino6061763.cl"},
	{0xF8BD7128CBDA6BD5ULL, "vxw_bonepitdestroyed_a3_1rx_0_0.vlm"},
	{0xF8BE5110B44C571FULL, "849951872.ogg"},
	{0xF8BE6D88C06255D7ULL, "29586326.ogg"},
	{0xF8BE729DCB0F4B05ULL, "mrl000_marethari6142117.cl"},
	{0xF8BEA0BAE34C2D81ULL, "182416104.ogg"},
	{0xF8BEE064B41775C2ULL, "cloth1.vsh.fxo"},
	{0xF8BEE74AFB23FA15ULL, "ser200_aveline6201539.cl"},
	{0xF8BEF0FD34BFF133ULL, "vxlt_0_lowtown_a2_ch9_0_0.vlm"},
	{0xF8BF06603DC5572DULL, "822192405.ogg"},
	{0xF8BF57BF5DEE22C7ULL, "6059423_m.fxe"},
	{0xF8BFA66BF3372DB4ULL, "vxht_mansion_n_5b1_0_0.vlm"},
	{0xF8BFA82A7CEFE131ULL, "40195208.ogg"},
	{0xF8BFB73781423AA1ULL, "vxht_0_hightown_7g2_0_0.vlm"},
	{0xF8BFF2FF88F46EF9ULL, "lowtown_argument6169360.cl"},
	{0xF8C0625BA2D0E6F1ULL, "lgt210cr_shp_kill_qunari3.utc"},
	{0xF8C0831FDE6E2F3CULL, "6085498_m.fxe"},
	{0xF8C0B3554ACF0433ULL, "6119500_f.fxe"},
	{0xF8C0BD2B5CD781F9ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_walkway.msh"},
	{0xF8C0C707015B21C3ULL, "art\\levels\\races\\ferelden\\props\\_textures\\plc_cupwood_01_0_d.dds"},
	{0xF8C0C72B5CD792E7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_walkway.mmh"},
	{0xF8C0C9D3E631813DULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_har_isab_0d.dds"},
	{0xF8C0CB2B5CD79934ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_walkway.mao"},
	{0xF8C131089FE44F1BULL, "hlac_lt_0_lowtown_65536_11y.dds"},
	{0xF8C1396DC3BD2F56ULL, "llac_ht_mansion2b_818_48.dds"},
	{0xF8C158C5D62A2361ULL, "lt_0_lowtown_5_0.rml"},
	{0xF8C16BD0B5FC6623ULL, "vxca_caverns2_7gb_0_1.vlm"},
	{0xF8C1F5FA24131098ULL, "controllerlayout.gfx"},
	{0xF8C1F9AD799D48AEULL, "29209226.ogg"},
	{0xF8C260123AEB4CC0ULL, "hlac_lt_0_lowtown_a3_n_c_-131072_2x.dds"},
	{0xF8C26C5FC23D0AA0ULL, "tow_distor.mao"},
	{0xF8C2D9ADC330A691ULL, "vxlt_gallowstemplar_5bx_0_0.vlm"},
	{0xF8C327BB55599CB6ULL, "store_dae370cr_armor.utm"},
	{0xF8C33CC8CF126280ULL, "711087237.ogg"},
	{0xF8C33DFF8E2BCD57ULL, "vxht_0_hightown_n_b8o_0_0.vlm"},
	{0xF8C4320F2945BB56ULL, "772990853.ogg"},
	{0xF8C51B5B5D546D70ULL, "420567763.ogg"},
	{0xF8C5369B62BD3121ULL, "hlac_ht_0_hightown_2_1y.dds"},
	{0xF8C55454207CEC89ULL, "380648978.ogg"},
	{0xF8C56578BD76DB8DULL, "485076988.ogg"},
	{0xF8C765D3E3C84DE5ULL, "6121090_m.fxe"},
	{0xF8C76DBACA084766ULL, "351299177.ogg"},
	{0xF8C80932ABB484A4ULL, "dre000_varric6207992.cl"},
	{0xF8C80F0C4E5A160AULL, "vxlt_0_lowtown_a2_n_gza_0_0.vlm"},
	{0xF8C8534899FCEE9FULL, "vxht_bartrand_n_6r6_0_0.vlm"},
	{0xF8C8DE847592D0DEULL, "6136938_m.fxe"},
	{0xF8C938F937E3123DULL, "static_static_p20.psh.fxo.dx11"},
	{0xF8C949E75463003AULL, "zz_mag_debug6054601.cl"},
	{0xF8CA71AE8BDB6E8BULL, "6071945_f.fxe"},
	{0xF8CAB9FD8A70ADB4ULL, "6095525_m.fxe"},
	{0xF8CB3DB3B3F5306BULL, "w_bonepitdestroyed_a3_36_0_lc.mmh"},
	{0xF8CB5E70D277E421ULL, "ser200_aveline6201478.cl"},
	{0xF8CB66345B483A62ULL, "llac_lt_hangedman_-1_18.dds"},
	{0xF8CB7CA6CB17EA41ULL, "hlac_lt_0_lowtown_a2_65533_2x.dds"},
	{0xF8CB8BF772A206EFULL, "vxca_caverns2prc_7nu_0_0.vlm"},
	{0xF8CBFC6C3CA2359BULL, "primitives_primitive_p5.vsh.fxo"},
	{0xF8CCA2F11D757EA2ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchhang.phy"},
	{0xF8CD3562F8FFDB32ULL, "one120_athenril6088294.cl"},
	{0xF8CD410165499FFCULL, "one140cr_charade6137628.cl"},
	{0xF8CD7750C8E35D38ULL, "zz_ave_debug6064133.cl"},
	{0xF8CE06BFA47FE7FDULL, "hlac_lt_0_lowtown_n_-3_2y.dds"},
	{0xF8CF9E0F6136DDE1ULL, "6027477_m.fxe"},
	{0xF8CFDA424794C830ULL, "6097008_m.fxe"},
	{0xF8D0A39525F6F1AFULL, "vxht_0_hightown_n_bgj_0_0.vlm"},
	{0xF8D0E342CD811A1AULL, "vxca_caverns1_mines_gkl_0_0.vlm"},
	{0xF8D16293EF77F0EFULL, "271499490.ogg"},
	{0xF8D167A0EC672F2BULL, "vxlt_undercitydungeon_1ap_0_0.vlm"},
	{0xF8D1976FC2DC0D78ULL, "l_one130ar_booker_dungeon.wbd"},
	{0xF8D21BB64CD9BE6EULL, "lt_0_lowtown_a3_n_c_3_0.msh"},
	{0xF8D229B64CD9D5BFULL, "lt_0_lowtown_a3_n_c_3_0.mao"},
	{0xF8D235B64CD9EA5CULL, "lt_0_lowtown_a3_n_c_3_0.mmh"},
	{0xF8D2B4905792E4EBULL, "288556658.ogg"},
	{0xF8D2D4462C753AF1ULL, "vxht_0_hightown_a3_c1u_0_0.vlm"},
	{0xF8D2F3D8E214F914ULL, "6164456_m.fxe"},
	{0xF8D32EEBEBF5C21FULL, "947002887.ogg"},
	{0xF8D3B525A7C338DFULL, "mag100_cullen6082315.cl"},
	{0xF8D3BD3CAAA4C4FDULL, "hm_zevran.mor"},
	{0xF8D3F1C9856A44A6ULL, "vxht_bartrand_n_6qj_0_0.vlm"},
	{0xF8D465848B9AA456ULL, "lgt160cr_wk_avern_archer.utc"},
	{0xF8D482BE7E6ADAB7ULL, "gen00fl_dog6166838.cl"},
	{0xF8D4C848BCF738AAULL, "character_character_p195.psh.fxo.dx11"},
	{0xF8D53463CD4D53B1ULL, "dae200_viscount.stg"},
	{0xF8D581842C34D67EULL, "vxlt_0_lowtown_a2_n_c_hss_0_0.vlm"},
	{0xF8D5F237ACA6FC86ULL, "t3_wer_h02.tnt"},
	{0xF8D65FDC065F7405ULL, "467599571.ogg"},
	{0xF8D680E940502E8AULL, "vxlt_0_lowtown_a2_n_gwb_0_0.vlm"},
	{0xF8D6DA98C689EB97ULL, "gen00fl_fenris6106606.cl"},
	{0xF8D6F61DCF2EB6FCULL, "hlac_ht_0_hightown_n_65533_1y.dds"},
	{0xF8D8137B6FB26E5EULL, "fen310_varania6118628.cl"},
	{0xF8D82FF15CE65D78ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic.phy"},
	{0xF8D8A81ECF5C5C16ULL, "llac_lt_backalley_n_0_9.dds"},
	{0xF8D8B8841C943F3EULL, "vxlt_backalley_n_1f4_0_0.vlm"},
	{0xF8D9071ADB95969BULL, "729175743.ogg"},
	{0xF8D971179875BAC7ULL, "hlac_ht_0_hightown_a2_-3_0z.dds"},
	{0xF8D9CB605ED754B4ULL, "vxlt_hangedman_1q1_0_0.vlm"},
	{0xF8DA029C2E1C180BULL, "197010946.ogg"},
	{0xF8DA54766783829CULL, "vxht_0_hightown_a3_cf1_0_0.vlm"},
	{0xF8DB7303BD005CF7ULL, "plt_lgt111_gxa_vgk_nbls_fnd.nss"},
	{0xF8DB7FFF3E26A77AULL, "vxw_woundedcoast_o_0_0.vlm"},
	{0xF8DB8A0E9BD24FABULL, "490534475.ogg"},
	{0xF8DC666155FF6305ULL, "vxdr_primevalentrance_2lo_0_0.vlm"},
	{0xF8DD05BDF6A66611ULL, "952371538.ogg"},
	{0xF8DD12D04CB542B3ULL, "vxlt_gallowscourtyard_6xi_0_0.vlm"},
	{0xF8DD24E9FE75D6D3ULL, "hlac_w_sundermount_5_0y.dds"},
	{0xF8DD55CB679C54C5ULL, "460662522.ogg"},
	{0xF8DD5BC0600126C5ULL, "6115879_m.fxe"},
	{0xF8DD6CB6030AA8BFULL, "471863242.ogg"},
	{0xF8DE27835D66AADFULL, "plt_lgt000pt_ultra_5_03.plo"},
	{0xF8DE4E6585B1C33CULL, "6135707_f.fxe"},
	{0xF8DE709E28608E23ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_enter.evt"},
	{0xF8DEB4989F7E4D84ULL, "vxlt_gallowstemplar_n_34v_0_0.vlm"},
	{0xF8DF24AF9C3E2A41ULL, "6104293_m.fxe"},
	{0xF8DFB8E203C00DFDULL, "mcr371_idunna6087278.cl"},
	{0xF8E04FCECD115E56ULL, "gen_im_trs_bks_enchanter.uti"},
	{0xF8E09DA775BBC9B3ULL, "6217096_m.fxe"},
	{0xF8E0BDB583CD55DDULL, "vxht_keep_n_6gx_0_0.vlm"},
	{0xF8E0EBDB0944F506ULL, "vxlt_0_lowtown_a2_n_c_hvd_0_0.vlm"},
	{0xF8E108DBE7565D9FULL, "dae000cr_grand_cleric6173672.cl"},
	{0xF8E109E445366BD1ULL, "310612022.ogg"},
	{0xF8E12A974E343F2CULL, "6082002_m.fxe"},
	{0xF8E135056BC734EAULL, "dae000_sabina6161449.cl"},
	{0xF8E140E75F0CF3D2ULL, "llac_lt_0_lowtown_docks_a3_n_-131074_1.dds"},
	{0xF8E1648330D02CBEULL, "6041074_m.fxe"},
	{0xF8E200352FBAB3C9ULL, "23864949.ogg"},
	{0xF8E2428E79BCA61CULL, "522864708.ogg"},
	{0xF8E287D0A12EA284ULL, "6166568_m.fxe"},
	{0xF8E29C59C22118C7ULL, "6156054_m.fxe"},
	{0xF8E416395A4C60DCULL, "6103345_m.fxe"},
	{0xF8E44440FE0EC3C0ULL, "vxlt_gallowstemplar_n_3rd_0_0.vlm"},
	{0xF8E4A04B33F70234ULL, "gen_im_trs_slv_sta_cld_r2.uti"},
	{0xF8E4F2C63246470AULL, "vxlt_0_lowtown_a2_n_glc_0_0.vlm"},
	{0xF8E5499E28666AA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_enter.gad"},
	{0xF8E5555251E11A33ULL, "gen_im_cft_reg_a2.uti"},
	{0xF8E5787C2D30122AULL, "mer341_jansen6082978.cl"},
	{0xF8E5C8D59A61FEF2ULL, "6146797_m.fxe"},
	{0xF8E5D0AA03B41762ULL, "enemytrap_reticle.dds"},
	{0xF8E5FD26FF2ED2F5ULL, "6162275_m.fxe"},
	{0xF8E646BEB40F4A21ULL, "592203998.ogg"},
	{0xF8E6532035D671F2ULL, "vxht_mansion_n_32q_0_0.vlm"},
	{0xF8E6BBA4FBA67594ULL, "plt_cod_hst_free_marches.plo"},
	{0xF8E6C79591B2F50DULL, "6068268_f.fxe"},
	{0xF8E6DA47D830FAECULL, "6152227_f.fxe"},
	{0xF8E6EACEBD605EB6ULL, "fake_mage_female.utc"},
	{0xF8E6F1C3BDC19D48ULL, "mer341_jansen.cnv"},
	{0xF8E71D0A66A1DBC0ULL, "dae000_elegant6166893.cl"},
	{0xF8E72B6B359A3EFEULL, "1006441914.ogg"},
	{0xF8E7A30D80193325ULL, "922592026.ogg"},
	{0xF8E893A0837DFCE7ULL, "vxca_caverns2_80m_0_0.vlm"},
	{0xF8E95F803EDF9B1BULL, "vxht_0_hightown_c5b_0_0.vlm"},
	{0xF8E976148D0BFB7FULL, "vxca_caverns1_icz_0_0.vlm"},
	{0xF8E99C704D28BDACULL, "hlac_lt_0_lowtown_docks_-131074_8y.dds"},
	{0xF8E9C2D7E3FA78E9ULL, "vxht_0_hightown_c3s_0_1.vlm"},
	{0xF8EA039D857CC955ULL, "6125205_m.fxe"},
	{0xF8EA0E0E026485F9ULL, "qun221_petrice6098752.cl"},
	{0xF8EA13C3BDC40CB0ULL, "mer341_jansen.bnk"},
	{0xF8EAB3369B1F2BD0ULL, "zz_vault_debug6203387.cl"},
	{0xF8EAEC0C9A8E7CA2ULL, "bodyheal2.dds"},
	{0xF8EB08CD2B8C5C14ULL, "6217468_m.fxe"},
	{0xF8EB81FC9000B2FCULL, "dae000_bartrand6206878.cl"},
	{0xF8EBDDA2D7F1BA2DULL, "lgt000ip_ultra_hm_remains.utp"},
	{0xF8EBF9B531484131ULL, "vxw_bonepitmining_a2_n5_0_0.vlm"},
	{0xF8EBFA204E900E55ULL, "follower_banter6169978.cl"},
	{0xF8EC589E286CA282ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\twist_look_left_enter.ani"},
	{0xF8ECB02B5CFCA61CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_walkway.phy"},
	{0xF8ECCACBA8FE479BULL, "art\\vfx\\cinematic\\vfx_ogre_footsteps_p.impa.ani"},
	{0xF8ED7E8EC4202A29ULL, "6208252_m.fxe"},
	{0xF8EDC7B9A416F087ULL, "6077849_m.fxe"},
	{0xF8EE569CD197A628ULL, "231056984.ogg"},
	{0xF8EE711E42124E01ULL, "300280112.ogg"},
	{0xF8EE788B7FDDF879ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop.ani"},
	{0xF8EE8EB7DBB9D61DULL, "6040819_f.fxe"},
	{0xF8EF176A315FEB39ULL, "hlac_lt_0_lowtown_docks_a3_-65539_0y.dds"},
	{0xF8EF2F9275C36E5DULL, "vxht_keep_n_6bw_0_0.vlm"},
	{0xF8EF303733461548ULL, "gen00fl_aveline6107065.cl"},
	{0xF8EF3A3A16FE5E80ULL, "art\\characters\\weapons\\longbows\\w_lbw_lb102.mao"},
	{0xF8EF53BABFE0D8EFULL, "136173790.ogg"},
	{0xF8EFF5613B8EBD07ULL, "vxlt_0_lowtown_a2_cif_0_0.vlm"},
	{0xF8F034CD3BB507FEULL, "vxlt_warehouse_197_0_0.vlm"},
	{0xF8F035B64CF2E7AFULL, "lt_0_lowtown_a3_n_c_3_0.rml"},
	{0xF8F036689A7B8FECULL, "6166872_m.fxe"},
	{0xF8F138FD5C08D1E6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_5.ani"},
	{0xF8F1E5A4801F4C4CULL, "vxlt_0_lowtown_a3_n_c_eop_0_1.vlm"},
	{0xF8F204D3B36CF8C6ULL, "hlac_ca_caverns1_24424_23x.dds"},
	{0xF8F270329487C870ULL, "6125583_m.fxe"},
	{0xF8F27DF11D960F56ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchhang.mao"},
	{0xF8F289F11D9623F5ULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchhang.mmh"},
	{0xF8F28BF11D96276BULL, "art\\levels\\races\\dwarf\\props\\lights\\prp_torchhang.msh"},
	{0xF8F2A7FB3F74EACFULL, "vxdr_primevaltomb_502_0_0.vlm"},
	{0xF8F3061DC48CDC7FULL, "static_static_p4.vsh.fxo.dx11"},
	{0xF8F32C3BEE4AA224ULL, "seb201_lady_harimann6177354.cl"},
	{0xF8F3AD85291A0C0BULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.impa.evt"},
	{0xF8F46B4EEBADDB66ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyo_0d.dds"},
	{0xF8F46F03F124EB75ULL, "6167175_m.fxe"},
	{0xF8F4D0172E1F93A4ULL, "lt_0_lowtown_docks_219_0.msh"},
	{0xF8F4E2172E1FB20DULL, "lt_0_lowtown_docks_219_0.mao"},
	{0xF8F4E6172E1FB946ULL, "lt_0_lowtown_docks_219_0.mmh"},
	{0xF8F4F36A50DA3170ULL, "mcr310_cut_chantry_blows_up6087563.cl"},
	{0xF8F5538B7FE3D7BBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop.gad"},
	{0xF8F57306A51B66B5ULL, "w_bonepitmining_a2_24_0_ll.mmh"},
	{0xF8F59C3E0080703EULL, "vxlt_backalley_n_1l6_0_0.vlm"},
	{0xF8F65993488FE3ACULL, "dae000_meredith6092604.cl"},
	{0xF8F687D6B4CEC901ULL, "gen00fl_fenris6112072.cl"},
	{0xF8F6AC4D60FFCDE0ULL, "gen00fl_aveline6139868.cl"},
	{0xF8F714B64CF8CE27ULL, "lt_0_lowtown_a3_n_c_3_0.phy"},
	{0xF8F7153BC910C61AULL, "lgt000im_ultra_pants.uti"},
	{0xF8F768D0DA1E3BCEULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.dura.ani"},
	{0xF8F7AFCC88836D0BULL, "vxlt_gallowsprison_n_6ge_0_0.vlm"},
	{0xF8F7CF3194EA1C92ULL, "572334727.ogg"},
	{0xF8F7D88B569F3F96ULL, "hm_gencr_raider_reaver.mor"},
	{0xF8F810EB639E18A0ULL, "dae000ip_crowd_em2_talk18.utp"},
	{0xF8F8693D7DAC64E5ULL, "art\\characters\\creatures\\darkspawnogre\\textures\\cn_hed_ogra_0n.dds"},
	{0xF8F86F1A8BF1BF8DULL, "310934140.ogg"},
	{0xF8F880666BEBC780ULL, "llac_ht_keep_3732_17.dds"},
	{0xF8F91C483390AB7DULL, "vxca_caverns1_edy_0_0.vlm"},
	{0xF8F999B62F915CAAULL, "dae300_alistair_king6133815.cl"},
	{0xF8F9B4A3267BE74BULL, "6108754_m.fxe"},
	{0xF8F9D2CBA908DF92ULL, "art\\vfx\\cinematic\\vfx_ogre_footsteps_p.impa.evt"},
	{0xF8F9D5DCF014F0C4ULL, "146365924.ogg"},
	{0xF8F9E44049DD63D1ULL, "6102549_m.fxe"},
	{0xF8FA42E56B6AB9E4ULL, "gen00fl_sebastian6177815.cl"},
	{0xF8FA77795AE4FCC5ULL, "181676356.ogg"},
	{0xF8FA79C4E223036DULL, "hlac_lt_0_lowtown_-65537_1x.dds"},
	{0xF8FA79E8DBA85F67ULL, "w_sundermount_32_0.rml"},
	{0xF8FAC685292055CCULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.impa.gad"},
	{0xF8FB86A3AD6D27EBULL, "vxht_keep_5hg_0_0.vlm"},
	{0xF8FC708B7FEA27A0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\cuddle_lying_male_loop.evt"},
	{0xF8FC8EEFA459713DULL, "6149799_m.fxe"},
	{0xF8FCCC8508ACEC98ULL, "6071949_m.fxe"},
	{0xF8FCD236E7DB9B3BULL, "pro000_captain_wright6005381.cl"},
	{0xF8FD070A9AFDF533ULL, "dae000_gamlen6048460.cl"},
	{0xF8FD282447E2A20CULL, "w_sundermount_171_0.msh"},
	{0xF8FD2E2447E2ACFEULL, "w_sundermount_171_0.mmh"},
	{0xF8FD2F9B647580C9ULL, "lgt220cr_shp_mag_guard.utc"},
	{0xF8FD3A2447E2C095ULL, "w_sundermount_171_0.mao"},
	{0xF8FD74682082FE16ULL, "561151984.ogg"},
	{0xF8FD7DDBFF4F1EBDULL, "760122111.ogg"},
	{0xF8FEC4685A648435ULL, "fen110_anso6095013.cl"},
	{0xF8FECB8DB2CDDF0BULL, "vxht_keep_370_0_0.vlm"},
	{0xF8FEFFB2ECF655E2ULL, "pro000_captain_wright6004385.cl"},
	{0xF8FF11440F06E453ULL, "vxw_bonepitdestroyed_a3_28k_0_0.vlm"},
	{0xF90003B18F45C33FULL, "6143705_f.fxe"},
	{0xF90021743C5C2531ULL, "llac_ht_mansion2b_818_32.dds"},
	{0xF900F1CBA90F3325ULL, "art\\vfx\\cinematic\\vfx_ogre_footsteps_p.impa.gad"},
	{0xF9017BF9FD322D98ULL, "art\\toolset\\characters\\playercharacter\\humanfemale\\mt_art_hf00.phy"},
	{0xF901A4529A9C4E84ULL, "atl_newcontent_dxt5_00.dds"},
	{0xF901C585292672CAULL, "art\\vfx\\combat\\vfx_m_wpn_fire_crust_c.impa.ani"},
	{0xF901F452599F2388ULL, "llac_lt_0_lowtown_n_0_0.dds"},
	{0xF9023418A0CE0BE7ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_a_c.anb"},
	{0xF9024332E8C59B29ULL, "gen00fl_varric6201279.cl"},
	{0xF902483C20CDEAC8ULL, "6107389_m.fxe"},
	{0xF90266CA8B97ABC4ULL, "follower_banter6136472.cl"},
	{0xF902D9116F0514D6ULL, "vxht_mansion2_2ic_0_1.vlm"},
	{0xF902E871F0A4A9FFULL, "1034023282.ogg"},
	{0xF9034A2186536F45ULL, "hlac_ht_0_hightown_65534_1x.dds"},
	{0xF903605A73241F62ULL, "vxlt_0_lowtown_a2_n_byu_0_0.vlm"},
	{0xF903868AF5FC46D4ULL, "p_cl.mao"},
	{0xF903E3CC9BCDD173ULL, "vxdr_primevaltomb_3hx_0_0.vlm"},
	{0xF903F8DF5F082B96ULL, "355346985.ogg"},
	{0xF9040E1AF27688BEULL, "vxlt_backalley_n_1w7_0_0.vlm"},
	{0xF90490D498264CC2ULL, "dae000ip_crowd_ff5_talk15.utp"},
	{0xF904ED03857675B8ULL, "gen_im_trs_gen_rng_r2.uti"},
	{0xF90533550DDDB277ULL, "dae000_arishok6068456.cl"},
	{0xF90560D0DA2A6B1FULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.dura.evt"},
	{0xF90621FD5C1B06E0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_5.gad"},
	{0xF9063C73656FFC83ULL, "612560535.ogg"},
	{0xF90659CE70F82C7FULL, "vxca_caverns1_iid_0_0.vlm"},
	{0xF9065DFC65600937ULL, "vxw_encounter2_u_0_0.vlm"},
	{0xF906B0BA3350FF1EULL, "da2terrain_terrain_p2.psh.fxo.dx11.assembly"},
	{0xF906BC8BA02DACE1ULL, "6034723_m.fxe"},
	{0xF906C812B943843EULL, "vxw_blightlands_11o_0_0.vlm"},
	{0xF906CA99745D84E5ULL, "follower_banter6141781.cl"},
	{0xF906D83446AD3D04ULL, "vxlt_0_lowtown_docks_n_b09_0_0.vlm"},
	{0xF906F7E76F5CD2CAULL, "fxa_cen_aoe.dds"},
	{0xF9074B27C1A3A765ULL, "640608113.ogg"},
	{0xF90891F9D05B82D2ULL, "vxw_bonepitdestroyed_a3_28n_0_0.vlm"},
	{0xF90897F1D077C0A9ULL, "dae000_arishok6067666.cl"},
	{0xF90902F15A3EB6FBULL, "vxw_woundedcoast_1cj_0_0.vlm"},
	{0xF90969CB4B862C49ULL, "vxlt_hangedman_2xp_0_0.vlm"},
	{0xF909C7A20AE4E6E2ULL, "hlac_lt_undercitydungeon_-1_42y.dds"},
	{0xF909F96CC83B7AF7ULL, "llac_ht_mansion2b_818_3.dds"},
	{0xF90A76F0212D8381ULL, "160124696.ogg"},
	{0xF90A78826E32F3E1ULL, "6167892_m.fxe"},
	{0xF90B4C08AB82D144ULL, "w_encounter3_140_0.mmh"},
	{0xF90B5008AB82D797ULL, "w_encounter3_140_0.mao"},
	{0xF90B6208AB82F6E6ULL, "w_encounter3_140_0.msh"},
	{0xF90BB04257B94479ULL, "740423931.ogg"},
	{0xF90BB7212E2FE848ULL, "hlac_lt_ruinedhouse_n_-1_2z.dds"},
	{0xF90BC4C834D3C443ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjre_lp.evt"},
	{0xF90C41D0DA305508ULL, "art\\vfx\\combat\\vfx_d_invulner_act_c.dura.gad"},
	{0xF90CAA8C1A984F2DULL, "hlac_lt_gallowscourtyard_65534_3y.dds"},
	{0xF90CC3483EEE8BD1ULL, "vxca_cave_19f_0_0.vlm"},
	{0xF90CE5ABC108E968ULL, "llac_lt_undercityencounter_-1_26.dds"},
	{0xF90D2F12DE5006EFULL, "zz_lgt_1_debug6188684.cl"},
	{0xF90DC3AC0ABABA22ULL, "6059314_m.fxe"},
	{0xF90DDE6A61638B2CULL, "660637109.ogg"},
	{0xF90E03820C90875FULL, "305628562.ogg"},
	{0xF90E0FC1F946F5F9ULL, "tau_star.dds"},
	{0xF90E227A6F572D45ULL, "dae011_corff_sta6149632.cl"},
	{0xF90E72583745C3ADULL, "ca_cavern1_haunted_mmap.dds"},
	{0xF90E81D7949FD876ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho05.phy"},
	{0xF90EFE8956767E66ULL, "art\\vfx\\combat\\vfx_glb_ele_f_minor_c.impa.evt"},
	{0xF90F5DD6460D8FCBULL, "follower_banter6160838.cl"},
	{0xF9107A5482BB2429ULL, "vxht_mansion2b_3d3_0_0.vlm"},
	{0xF910F348F5873C0BULL, "vxht_mansion_n_4xe_0_0.vlm"},
	{0xF911402653E288ACULL, "l_lgt231ar_tome_final.bnk"},
	{0xF911638DF925E296ULL, "llac_ca_caverns2prc_5802_42.dds"},
	{0xF9119799E4D1D124ULL, "6109127_m.fxe"},
	{0xF911F077529C2735ULL, "6147707_f.fxe"},
	{0xF911F6C6537E15AEULL, "94447903.ogg"},
	{0xF9120E4F1EBB1CB7ULL, "6065675_m.fxe"},
	{0xF91279E8DBBCF084ULL, "w_sundermount_32_0.mmh"},
	{0xF9127DE8DBBCF7D7ULL, "w_sundermount_32_0.mao"},
	{0xF9128FE8DBBD1626ULL, "w_sundermount_32_0.msh"},
	{0xF912903546B82886ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_highlightb_sa.dds"},
	{0xF9129DC834D9A0C4ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjre_lp.gad"},
	{0xF912C469996D487FULL, "w_blightlands_5_0.rml"},
	{0xF912CFCC841565F0ULL, "vxca_caverns2prc_1ye_0_0.vlm"},
	{0xF912F8348CA7F086ULL, "vxca_caverns2prc_a9s_0_0.vlm"},
	{0xF9134B42464960CCULL, "l_mcr371ar_gallows_templar_nt.bnk"},
	{0xF913F5F75754CB1FULL, "6128385_m.fxe"},
	{0xF913F84992F6B69EULL, "hlac_ht_0_hightown_a3_-1_2z.dds"},
	{0xF9141A8B78627E71ULL, "qun160ip_vashothlead_dummy.utp"},
	{0xF9148E171FF4AD46ULL, "6033559_f.fxe"},
	{0xF91493866C6F9E0BULL, "vxlt_gallowsprison_n_6mg_0_0.vlm"},
	{0xF914B0C9A694F053ULL, "180764553.ogg"},
	{0xF9155CCFD4B7E380ULL, "vxht_0_hightown_a3_buq_0_0.vlm"},
	{0xF915618D63237997ULL, "vxht_mansion2b_1gg_0_1.vlm"},
	{0xF91581E1C6FC9997ULL, "vxlt_gallowstemplar_f_5ji_0_0.vlm"},
	{0xF915847EC1555590ULL, "vxht_0_hightown_8pq_0_0.vlm"},
	{0xF9161589567CC3B1ULL, "art\\vfx\\combat\\vfx_glb_ele_f_minor_c.impa.gad"},
	{0xF9166848B57531ECULL, "6122735_m.fxe"},
	{0xF916B2264A67E9EFULL, "vxca_caverns2prc_8t8_0_0.vlm"},
	{0xF916DB75C55BEAC5ULL, "446317121.ogg"},
	{0xF91774CE4C34CF7BULL, "vxw_woundedcoast_1nm_0_0.vlm"},
	{0xF917FB6C18CC09B6ULL, "lowtown_fem_prostitute_sta6152352.cl"},
	{0xF9182F699E018D08ULL, "vxht_mansion2b_1ms_0_0.vlm"},
	{0xF9186AD794A80517ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho05.msh"},
	{0xF91880D794A82A89ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ho05.mmh"},
	{0xF918ACC834DE25A2ULL, "art\\characters\\creatures\\stonegolem\\animation\\ua_acnjre_lp.ani"},
	{0xF918C655334F08BAULL, "vxht_0_hightown_a3_c7d_0_0.vlm"},
	{0xF918C75D3789E2CAULL, "vxlt_gallowstemplar_f_47f_0_0.vlm"},
	{0xF91936FB13B4F40CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railoc.phy"},
	{0xF91A4AF40EA07471ULL, "hlac_ht_brothel_n_2995_2y.dds"},
	{0xF91A8368A5D6C688ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_basic.mmh"},
	{0xF91A8F68A5D6DB2BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_basic.mao"},
	{0xF91AA168A5D6F9B2ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\1wo_basic.msh"},
	{0xF91B6C415739A636ULL, "dre000_sandal6028360.cl"},
	{0xF91B983BBD259237ULL, "525106071.ogg"},
	{0xF91BD4DC73F8E106ULL, "dae200_cullen6140276.cl"},
	{0xF91CA0A1BBA6A0FDULL, "plt_seb000pt_endstage.plo"},
	{0xF91CA87060321E0FULL, "vxlt_gallowscourtyard_649_0_0.vlm"},
	{0xF91CC8734DE9D262ULL, "6109705_m.fxe"},
	{0xF91CE85073FC0BEBULL, "qun131_taskmaster6039714.cl"},
	{0xF91D16895682E45FULL, "art\\vfx\\combat\\vfx_glb_ele_f_minor_c.impa.ani"},
	{0xF91D18CC273A266FULL, "vxca_cave_22z_0_0.vlm"},
	{0xF91D2F54A7B0B9D3ULL, "zz_followers_banter6158127.cl"},
	{0xF91D934FE6780626ULL, "783420798.ogg"},
	{0xF91DABCACE3B97EBULL, "vxht_0_hightown_n_ccs_0_0.vlm"},
	{0xF91DB0C82162812CULL, "731772807.ogg"},
	{0xF91EA29380FC81BDULL, "vxlt_gallowscourtyard_6ep_0_0.vlm"},
	{0xF91F5C194BA2A5BDULL, "qcr000_arishok6059326.cl"},
	{0xF91FF95D6569C69FULL, "239957194.ogg"},
	{0xF920D7AE15FEA85AULL, "dae000_hate_enemies6167592.cl"},
	{0xF920E80011393B30ULL, "vxht_mansion_n_43g_0_0.vlm"},
	{0xF9217B478D6FD4C7ULL, "ser221_gascard6149251.cl"},
	{0xF921CDD8040982F3ULL, "555809018.ogg"},
	{0xF9223DDF9B6DBB03ULL, "220422559.ogg"},
	{0xF9225E1415017E8AULL, "zz_followers_banter6157859.cl"},
	{0xF9225F68AB3A78D4ULL, "6094539_m.fxe"},
	{0xF9228E5114BA0F5AULL, "w_blightlands_71_0.phy"},
	{0xF922938A47DB4267ULL, "t1_tat_dbr.tnt"},
	{0xF923A3A1BBACC42AULL, "plt_seb000pt_endstage.nss"},
	{0xF923CB52228C7D3FULL, "art\\vfx\\combat\\vfx_r_brd_ssong_act_c.impa.gad"},
	{0xF923D4BBE7AA6265ULL, "6138789_f.fxe"},
	{0xF9243087AFD39CE1ULL, "art\\vfx\\materials\\vfx_d_heatnoise.mao"},
	{0xF924B9CB4BACE5C7ULL, "gen00fl_merrill6147018.cl"},
	{0xF925072B6E0FD6ABULL, "hlac_w_encounter3_-393215_0z.dds"},
	{0xF92574F29924B2A7ULL, "vxlt_foundry_n_tc_0_0.vlm"},
	{0xF925E6AB3E2AE7FBULL, "vxw_woundedcoast_1il_0_0.vlm"},
	{0xF9265603BD4016D8ULL, "plt_lgt111_gxa_vgk_nbls_fnd.plo"},
	{0xF9267C5B268AAAACULL, "107514017.ogg"},
	{0xF92691756DCE18C4ULL, "hlac_dr_slavers_3215_5x.dds"},
	{0xF926C7B9A1ED9DB8ULL, "vxlt_warehouse_15e_0_0.vlm"},
	{0xF926CB7F66D3B5FEULL, "vxlt_0_lowtown_a3_n_c_f66_0_1.vlm"},
	{0xF92704C78D5A2AF5ULL, "141999431.ogg"},
	{0xF9271C8A00DFB6ACULL, "art\\characters\\playercharacter\\humanmale\\boo_masc.phy"},
	{0xF9272E6405187AEAULL, "6121897_m.fxe"},
	{0xF9274F4EE6C2A163ULL, "hlac_lt_0_lowtown_a3_n_c_131071_5z.dds"},
	{0xF927CE2A2934DE09ULL, "vxca_caverns1_mines_inz_0_0.vlm"},
	{0xF928754C4418C683ULL, "follower_banter6155838.cl"},
	{0xF9289B3212ECE384ULL, "vxca_caverns2_95h_0_0.vlm"},
	{0xF92933076F8A714CULL, "vxlt_0_lowtown_goj_0_1.vlm"},
	{0xF9294CB52AAEBE62ULL, "static_billboard_p0.psh.fxo.dx11"},
	{0xF929573A7256D647ULL, "6146819_m.fxe"},
	{0xF92A3118A0EF58E8ULL, "art\\vfx\\combat\\vfx_w_wns_shield_defen_a_c.mmh"},
	{0xF92AC0522292892CULL, "art\\vfx\\combat\\vfx_r_brd_ssong_act_c.impa.evt"},
	{0xF92B46E93FD2EEFFULL, "6035924_m.fxe"},
	{0xF92B49F46AE7147DULL, "760993124.ogg"},
	{0xF92B7387AC81961BULL, "sicesheet.dds"},
	{0xF92BC9855A2453BAULL, "gen_im_arm_boo_lgt_set_high.uti"},
	{0xF92C8120AC822B1AULL, "vxht_mansion_n_500_0_1.vlm"},
	{0xF92CA1811F1CC683ULL, "hlac_ht_chantry_2534_14y.dds"},
	{0xF92D324295F681F2ULL, "6101473_m.fxe"},
	{0xF92D482A15CE1CE0ULL, "lowtown_male_drunk_mob6151164.cl"},
	{0xF92D83729BAF789EULL, "144725946.ogg"},
	{0xF92DA01473363D16ULL, "6103839_m.fxe"},
	{0xF92E3EA8EE2DCAEAULL, "6094520_m.fxe"},
	{0xF92E439EA9F389EBULL, "6001000_m.fxe"},
	{0xF92EB5196AE949AAULL, "zz_ave_debug6064629.cl"},
	{0xF92EEC7F71DAF53CULL, "art\\vfx\\combat\\vfx_glb_ele_p_extreme_c.impa.ani"},
	{0xF92F7157FF910171ULL, "6053049_m.fxe"},
	{0xF92FCEC4D0F959F8ULL, "isa311_castillon6092100.cl"},
	{0xF930374C2BD23711ULL, "t1_mue_pi1.tnt"},
	{0xF931944C01296353ULL, "vxlt_0_lowtown_f17_0_0.vlm"},
	{0xF931A7A17930514FULL, "vxw_bonepit_d3_0_0.vlm"},
	{0xF9325A4AA3C3FF8CULL, "vxlt_0_lowtown_docks_a3_n_d0x_0_0.vlm"},
	{0xF932AC89E8AA41CCULL, "llac_w_sundermount_131075_0.dds"},
	{0xF932F3E4793093CBULL, "hlac_ht_0_hightown_a2_131074_0x.dds"},
	{0xF9331CF15D32E80DULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic.msh"},
	{0xF9332AF15D330040ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic.mao"},
	{0xF9332EF15D330693ULL, "art\\levels\\races\\proxy\\tilesets\\setext\\8co_basic\\4so_basic.mmh"},
	{0xF933DD830C4CF184ULL, "ser270_emeric6113170.cl"},
	{0xF93429549E2B0ACFULL, "6043557_f.fxe"},
	{0xF9342C536B4F4A54ULL, "vxht_0_hightown_n_a2u_0_0.vlm"},
	{0xF93480E2DA2F59E8ULL, "6104414_m.fxe"},
	{0xF934FE6B4E323465ULL, "dae000_bartrand6030603.cl"},
	{0xF9354539FCB25CDFULL, "6162307_m.fxe"},
	{0xF93565080B0EAA47ULL, "abi_w_ss_front_clear.bnk"},
	{0xF935AF0E16A31315ULL, "6085518_m.fxe"},
	{0xF935F97F71E12A7EULL, "art\\vfx\\combat\\vfx_glb_ele_p_extreme_c.impa.gad"},
	{0xF9367B011B666D95ULL, "6142981_m.fxe"},
	{0xF93732522A3ECCE1ULL, "gen00fl_anders6131323.cl"},
	{0xF93741D020C8401FULL, "pro000cr_wesley.utc"},
	{0xF9376687C54E2951ULL, "211253582.ogg"},
	{0xF937C6AC53771945ULL, "806519433.ogg"},
	{0xF9388F21EBC7A52AULL, "one110_ambush_st.stg"},
	{0xF938A852229E9D85ULL, "art\\vfx\\combat\\vfx_r_brd_ssong_act_c.impa.ani"},
	{0xF93A45FE00A37767ULL, "vxdr_primevaltomb_6iz_0_0.vlm"},
	{0xF93A8B1A34F5216AULL, "zz_vault_debug6201721.cl"},
	{0xF93B119BC61E0DEFULL, "ts_demonic_prog_junk1.gda"},
	{0xF93B1501C27E01D2ULL, "vxlt_0_lowtown_f12_0_0.vlm"},
	{0xF93BEB5ED519FA4AULL, "ave200_aveline_bar6116177.cl"},
	{0xF93C11F1652C5596ULL, "6104939_m.fxe"},
	{0xF93C7E3FEEF5B785ULL, "dae000_male_templar_mob.bnk"},
	{0xF93CD47F71E7099DULL, "art\\vfx\\combat\\vfx_glb_ele_p_extreme_c.impa.evt"},
	{0xF93CE5DC5006C4BAULL, "700054237.ogg"},
	{0xF93CE6C38F32DD7DULL, "vxw_bonepit_10v_0_0.vlm"},
	{0xF93E05244819FD99ULL, "w_sundermount_171_0.phy"},
	{0xF93E4D7CF7FF0663ULL, "vxht_0_hightown_a3_cbw_0_1.vlm"},
	{0xF93E75D0D816B7ACULL, "vxlt_undercityserialkiller_15b_0_0.vlm"},
	{0xF93EA2A450F8B0FBULL, "mag210_arianni6057321.cl"},
	{0xF93F3A204509F174ULL, "vxht_brothel_n_50d_0_0.vlm"},
	{0xF93F603FEEF7BABDULL, "dae000_male_templar_mob.cnv"},
	{0xF94068F8C1221CE0ULL, "vxw_bonepit_271_0_0.vlm"},
	{0xF9408B46D72A4521ULL, "vxlt_hangedman_1tw_0_0.vlm"},
	{0xF9408C7431C48615ULL, "hlac_dr_primevaltomb_659_3z.dds"},
	{0xF940A4111C5E8E26ULL, "llac_lt_blackemporium_1055_24.dds"},
	{0xF940D93FC6291ECAULL, "615601300.ogg"},
	{0xF941DBD7B5E2EA91ULL, "1065112573.ogg"},
	{0xF9426CB218484304ULL, "hlac_lt_0_lowtown_a2_n_c_131071_6y.dds"},
	{0xF9427F4350571AECULL, "zz_followers_banter6157633.cl"},
	{0xF9429E305D4562D4ULL, "vxht_0_hightown_n_a7z_0_0.vlm"},
	{0xF942F13BF735CBD8ULL, "follower_banter6172263.cl"},
	{0xF9432A0A9F370351ULL, "vxlt_undercitydungeon_15v_0_0.vlm"},
	{0xF943B720CC0A31AAULL, "168174351.ogg"},
	{0xF943E5B3F73D37ADULL, "vxlt_backalley_g_1jh_0_2.vlm"},
	{0xF943F76220FE5AF2ULL, "574695536.ogg"},
	{0xF94428889DF58345ULL, "404702338.ogg"},
	{0xF9445D6F96A0C6FDULL, "qun110_petrice6036858.cl"},
	{0xF944C179D8AA2F5DULL, "6103825_m.fxe"},
	{0xF944C2B92171A8CAULL, "dae000_bartrand6016138.cl"},
	{0xF9450B891BB6E252ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_enter.gad"},
	{0xF945182A8304D8E1ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masc_0t.dds"},
	{0xF945B94B3D147DE1ULL, "zz_ave_debug6207517.cl"},
	{0xF945CC882A2AFC39ULL, "6107696_m.fxe"},
	{0xF94601B5B426F801ULL, "vxlt_0_lowtown_a2_d6f_0_0.vlm"},
	{0xF9467AEAB049D66AULL, "vxht_mansion2b_3bz_0_0.vlm"},
	{0xF947075F8CE2F456ULL, "pro000_opening6178527.cl"},
	{0xF9475BDDAE0015D2ULL, "zz_dog_debug6204272.cl"},
	{0xF9489F1A3B069638ULL, "qcr000_arishok6060152.cl"},
	{0xF948B081219DF997ULL, "hlac_ht_0_hightown_a3_65536_2x.dds"},
	{0xF948CD3EDE5F06F5ULL, "vxht_brothel_n_5h9_0_0.vlm"},
	{0xF948EA7662D8E789ULL, "vxlt_0_lowtown_a2_n_e1u_0_0.vlm"},
	{0xF9491DBAAA832964ULL, "126503887.ogg"},
	{0xF94978B5011411C0ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\_textures\\lt_floorbasec_n.dds"},
	{0xF94A566F960EBD4EULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_control.evt"},
	{0xF94A814604817B06ULL, "llac_dr_primevalentrance_1724_0.dds"},
	{0xF94AB13C6B75FFE1ULL, "hlac_dr_primevaltomb_657_5z.dds"},
	{0xF94B0E891BBB5289ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_enter.evt"},
	{0xF94B49085BDA7734ULL, "prm_col_quest_given.ncs"},
	{0xF94B58A08128F5FDULL, "vxw_bonepit_15u_0_0.vlm"},
	{0xF94B93117EE5618AULL, "6102825_m.fxe"},
	{0xF94BAD8C430A9B3AULL, "6037371_f.fxe"},
	{0xF94C0C8BF4305FF5ULL, "fen110_fenris6100236.cl"},
	{0xF94C198524FF981AULL, "hlac_w_woundedcoast_327686_0y.dds"},
	{0xF94C652D0E07F569ULL, "vxca_caverns2_79m_0_2.vlm"},
	{0xF94C76AE40E721AEULL, "736571399.ogg"},
	{0xF94C94161780815BULL, "hlac_lt_0_lowtown_docks_-2_0x.dds"},
	{0xF94CA61DA851EDE4ULL, "6135924_m.fxe"},
	{0xF94D59717B3D5D0BULL, "6177279_m.fxe"},
	{0xF94DB4CA4340D73EULL, "and100_anders6096009.cl"},
	{0xF94DB9D186CAD346ULL, "72487565.ogg"},
	{0xF94E70A2E7D3F2D9ULL, "6208898_m.fxe"},
	{0xF94F6A853BED9898ULL, "2904870.ogg"},
	{0xF94F86D2DEAD80D5ULL, "196739682.ogg"},
	{0xF94FD3270C5BABE8ULL, "zz_qun_debug6028225.cl"},
	{0xF95000BE8480888AULL, "vxlt_0_lowtown_docks_n_c4b_0_0.vlm"},
	{0xF9502A26D9B01573ULL, "6119464_m.fxe"},
	{0xF9511B8D74CE5232ULL, "character_character_p104.psh.fxo.dx11.assembly"},
	{0xF9514D6F9614CCF9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_control.gad"},
	{0xF951773BA618CEEAULL, "depth_speedtree_debug.vsh.fxo"},
	{0xF9519BE7912D1BD1ULL, "vxlt_undercitydungeon_18u_0_0.vlm"},
	{0xF951C875205A89D1ULL, "564985598.ogg"},
	{0xF952448AE6CBBD20ULL, "vxlt_0_lowtown_n_ep2_0_1.vlm"},
	{0xF9524A26541A30DEULL, "l_lgt231ar_tome_final.wbd"},
	{0xF952BB6DA971C415ULL, "948015631.ogg"},
	{0xF95383F696E17522ULL, "6055421_f.fxe"},
	{0xF9539086A2159952ULL, "578055295.ogg"},
	{0xF954019B88E18DA0ULL, "one280_iwan6096896.cl"},
	{0xF9545542468108FEULL, "l_mcr371ar_gallows_templar_nt.wbd"},
	{0xF954A72482E2CB6AULL, "617813426.ogg"},
	{0xF9556A80CB18D12FULL, "vxlt_0_lowtown_a2_n_c_eny_0_1.vlm"},
	{0xF956138E08D32BF9ULL, "qun131_taskmaster6059062.cl"},
	{0xF9564865D568EA09ULL, "llac_lt_0_lowtown_n_-196608_2.dds"},
	{0xF95668AAE17EA083ULL, "mag291_demon_marethari6029038.cl"},
	{0xF956B3C371212736ULL, "art\\characters\\creatures\\dragon\\animation\\ev_v1.ani"},
	{0xF956E2D657B535B2ULL, "6027827_m.fxe"},
	{0xF9575E6F96195547ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_blood_control.ani"},
	{0xF957A7F3136CC849ULL, "vxht_brothel_n_707_0_0.vlm"},
	{0xF9581E398C53468CULL, "535023543.ogg"},
	{0xF9585F04A68A6AC2ULL, "6055623_f.fxe"},
	{0xF95885F08A1451F7ULL, "vxlt_0_lowtown_a2_gp9_0_0.vlm"},
	{0xF958C376CC9FA85CULL, "vxlt_gallowscourtyard_72u_0_0.vlm"},
	{0xF958DDF376213486ULL, "498044971.ogg"},
	{0xF958F6891BC767F8ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\ranged\\ranged_enter.ani"},
	{0xF959A5251BDD107DULL, "6140240_m.fxe"},
	{0xF959AF418CD527A0ULL, "dae210cr_qunari_m_01.utc"},
	{0xF959CD32D699A530ULL, "vxdr_primevaltomb_538_0_0.vlm"},
	{0xF95A0DFB13EC45D7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railoc.mmh"},
	{0xF95A11FB13EC4CE4ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railoc.mao"},
	{0xF95A23FB13EC6B49ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_railoc.msh"},
	{0xF95A31D46736AB9CULL, "vxht_mansion2_1rf_0_0.vlm"},
	{0xF95A4AB4213D1950ULL, "6155643_m.fxe"},
	{0xF95A551AD496388DULL, "vxlt_warehouse_n_12f_0_0.vlm"},
	{0xF95A825C35066E63ULL, "183093157.ogg"},
	{0xF95A902E3D3F86A2ULL, "vxlt_darktown_5jc_0_0.vlm"},
	{0xF95A99AFB2A4E9FBULL, "457972254.ogg"},
	{0xF95B6BE884395374ULL, "vxht_0_hightown_a2_c5m_0_0.vlm"},
	{0xF95B80FA997199A8ULL, "851337067.ogg"},
	{0xF95C8AEE930C5299ULL, "hlac_ht_keep_3736_1z.dds"},
	{0xF95CC6FDD32D9047ULL, "art\\characters\\creatures\\dragon\\animation\\ddie.gad"},
	{0xF95CC906A571A896ULL, "vxw_bonepitmining_a2_26f_0_0.vlm"},
	{0xF95CDB24D8863701ULL, "220567799.ogg"},
	{0xF95D008952F2F86EULL, "vxlt_0_lowtown_a2_gl8_0_0.vlm"},
	{0xF95D68F9FD806C2DULL, "art\\toolset\\characters\\playercharacter\\humanfemale\\mt_art_hf00.msh"},
	{0xF95D7AF9FD808AB3ULL, "art\\toolset\\characters\\playercharacter\\humanfemale\\mt_art_hf00.mmh"},
	{0xF95DB3A8FECEEF2EULL, "6072094_m.fxe"},
	{0xF95E2BA0F7E8E138ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_vnicechair_n.dds"},
	{0xF95E4C9B9342F45DULL, "6173817_m.fxe"},
	{0xF95E687CB2E43F97ULL, "appearance_large.gda"},
	{0xF95EA9921F97CA91ULL, "643353720.ogg"},
	{0xF95EBE92D6419590ULL, "6066862_m.fxe"},
	{0xF95ED29F85BB2F93ULL, "llac_lt_0_lowtown_65538_5.dds"},
	{0xF95FBEEF849CE84CULL, "180188424.ogg"},
	{0xF95FE56D0A7CFAE0ULL, "619858085.ogg"},
	{0xF95FEAA64CC2A97FULL, "6072345_m.fxe"},
	{0xF9601C57E5C699D1ULL, "6098974_m.fxe"},
	{0xF96040196FD56BE5ULL, "941360608.ogg"},
	{0xF96068BA57D00788ULL, "vxht_mansion2b_1tl_0_0.vlm"},
	{0xF960B86999AF556FULL, "w_blightlands_5_0.mao"},
	{0xF960C46999AF690CULL, "w_blightlands_5_0.mmh"},
	{0xF960CA6999AF73FEULL, "w_blightlands_5_0.msh"},
	{0xF960F998AB9C924EULL, "6090888_f.fxe"},
	{0xF9615B080B33D47DULL, "abi_w_ss_front_clear.wbd"},
	{0xF96221BAD166A5BBULL, "hlac_ca_cave_2194_4x.dds"},
	{0xF9623EFD3DDA5C3CULL, "vxlt_gallowstemplar_f_4uz_0_0.vlm"},
	{0xF9626B0EB9BA6BC7ULL, "838934338.ogg"},
	{0xF96277754E9C5EDDULL, "6209978_m.fxe"},
	{0xF962AB2A78E50990ULL, "6123315_player_concerned.cut"},
	{0xF962B73EC27C145DULL, "hlac_lt_0_lowtown_a2_n_c_131070_5z.dds"},
	{0xF962F81B600EB216ULL, "53184050.ogg"},
	{0xF9636EDF5CD4CFCAULL, "vxht_mansion2_3dt_0_0.vlm"},
	{0xF963B38CE125CCACULL, "gen00fl_fenris6108106.cl"},
	{0xF963BBFDD3339C54ULL, "art\\characters\\creatures\\dragon\\animation\\ddie.evt"},
	{0xF963BEC02677A99AULL, "400633855.ogg"},
	{0xF9642E9128626F2FULL, "zz_ave_debug6056780.cl"},
	{0xF964BBC3712D7227ULL, "art\\characters\\creatures\\dragon\\animation\\ev_v1.evt"},
	{0xF9659B94D685C0E7ULL, "vxdr_primevaltomb_6vw_0_0.vlm"},
	{0xF966B46F53B374ACULL, "6085375_m.fxe"},
	{0xF966F7CD7C0A6A77ULL, "vxlt_0_lowtown_a2_gw8_0_0.vlm"},
	{0xF9670203A78F44DEULL, "isa120_hayder6067137.cl"},
	{0xF9675400D65BFF11ULL, "6087185_m.fxe"},
	{0xF96787B5FAE68180ULL, "mag101_grace6034023.cl"},
	{0xF967DA8D29B658C4ULL, "hlac_lt_0_lowtown_a2_n_131071_5z.dds"},
	{0xF967F38A01170877ULL, "art\\characters\\playercharacter\\humanmale\\boo_masc.mmh"},
	{0xF967F78A01170E84ULL, "art\\characters\\playercharacter\\humanmale\\boo_masc.mao"},
	{0xF968098A01172DE9ULL, "art\\characters\\playercharacter\\humanmale\\boo_masc.msh"},
	{0xF9680D9973C683F0ULL, "hlac_w_sundermount_196608_0y.dds"},
	{0xF968320FF0A3BB28ULL, "6033477_m.fxe"},
	{0xF968589BB2A9DFFFULL, "6037367_f.fxe"},
	{0xF96866238EBEE395ULL, "hlac_dr_0_deeproads_5843_5x.dds"},
	{0xF968A3B1592CC41CULL, "vxht_mansion2_1ug_0_0.vlm"},
	{0xF969020B2F359F22ULL, "vxlt_darktown_5ol_0_0.vlm"},
	{0xF9691A7FE73FCD0CULL, "fen310_fenris6118683.cl"},
	{0xF96951DFCF2D2313ULL, "vxca_cavern1_haunted_in8_0_0.vlm"},
	{0xF96999F9C0304B67ULL, "vxlt_undercityserialkiller_wx_0_0.vlm"},
	{0xF969DDC5762F6BF4ULL, "vxht_0_hightown_a2_c0r_0_0.vlm"},
	{0xF96A40BFE359FB93ULL, "vxht_0_hightown_n_4n3_0_0.vlm"},
	{0xF96ADCC371321670ULL, "art\\characters\\creatures\\dragon\\animation\\ev_v1.gad"},
	{0xF96AF421FC4E3D62ULL, "vxlt_0_lowtown_a2_ghx_0_0.vlm"},
	{0xF96B1C29D39DCE1EULL, "distor_cyl_alp.dds"},
	{0xF96B726644E910EEULL, "vxlt_0_lowtown_a2_gc9_0_0.vlm"},
	{0xF96B8123D4E093DAULL, "15272221.ogg"},
	{0xF96B8DB158A3E4B6ULL, "one300cr_solivitus.ncs"},
	{0xF96C147B031D6990ULL, "llac_ht_playershouse_n_855_5.dds"},
	{0xF96C5DEC554E2196ULL, "hlac_lt_gallowstemplar_f_2383_1x.dds"},
	{0xF96CA014400A16ACULL, "vxdr_primevaltomb_7e6_0_2.vlm"},
	{0xF96CAAF963B117E2ULL, "435938565.ogg"},
	{0xF96D020AFF85727CULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizeb.mao"},
	{0xF96D0B50AAD60FE8ULL, "dae000_arishok6067711.cl"},
	{0xF96D0E0AFF85861FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizeb.mmh"},
	{0xF96D140AFF8590C1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_slavefreizeb.msh"},
	{0xF96D6A8A22E39874ULL, "hlac_ht_0_hightown_n_-65538_0z.dds"},
	{0xF96E23B830F366FDULL, "cst_s_star.mao"},
	{0xF96E28D0AE230BBEULL, "642855079.ogg"},
	{0xF96EBACE9D111D73ULL, "lgt231ip_at_darktown.utp"},
	{0xF96ED4DA20D4C660ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_hatch_2.mao"},
	{0xF96EE094594B2AA5ULL, "vxht_mansion2_20j_0_0.vlm"},
	{0xF96EFF5A7C4A171DULL, "vxlt_hangedman_n_31y_0_0.vlm"},
	{0xF9703F376CFF38ABULL, "654660565.ogg"},
	{0xF970B3FDD33E188DULL, "art\\characters\\creatures\\dragon\\animation\\ddie.ani"},
	{0xF970F2F8BD82E9FBULL, "6129144_m.fxe"},
	{0xF970F6C173A781D5ULL, "6173022_m.fxe"},
	{0xF97158829C8BA166ULL, "art\\characters\\creatures\\gibgolem\\textures\\cn_bdy_gib_0n.dds"},
	{0xF97159B411CCBA76ULL, "vrc321ip_spooky6_obj_3.utp"},
	{0xF9723FE09C40FAA1ULL, "vxlt_gallowsprison_n_5wh_0_0.vlm"},
	{0xF973F7C801AC22C9ULL, "238822132.ogg"},
	{0xF97418C7D55CA701ULL, "991849218.ogg"},
	{0xF97474C997FAFA20ULL, "554900535.ogg"},
	{0xF97489DF333808F9ULL, "6018807_f.fxe"},
	{0xF974FFF2E5677F20ULL, "27283394.ogg"},
	{0xF9758C834074BBB4ULL, "544507191.ogg"},
	{0xF975A9BC2DD349AAULL, "6149987_m.fxe"},
	{0xF976144DD106B200ULL, "mcr371ar_gallows_templar_nt.are"},
	{0xF9763A623C64DF2FULL, "gen00fl_aveline6139490.cl"},
	{0xF9767A0DEECFB5CDULL, "6143230_m.fxe"},
	{0xF976843F0ED99613ULL, "6170591_m.fxe"},
	{0xF9769BAE32DC44BCULL, "gen00fl_anders6125369.cl"},
	{0xF976B36BABFE84C8ULL, "54161087.ogg"},
	{0xF976E161899B05ABULL, "6169658_m.fxe"},
	{0xF9773AF6E8762D38ULL, "6093497_m.fxe"},
	{0xF977791F845E0834ULL, "227778799.ogg"},
	{0xF9777D0DF3A5DA20ULL, "vxlt_0_lowtown_a2_fgz_0_0.vlm"},
	{0xF977C3271F236DA2ULL, "6146902_m.fxe"},
	{0xF97829A60927CC22ULL, "lt_0_lowtown_docks_a3_n_137_0.phy"},
	{0xF97835815637A162ULL, "293209892.ogg"},
	{0xF978561D06298B16ULL, "si_cl.mao"},
	{0xF97870C4D1CB1865ULL, "vxht_0_hightown_apq_0_0.vlm"},
	{0xF97875DA666D06C6ULL, "370950110.ogg"},
	{0xF978A442D121B038ULL, "6042983_m.fxe"},
	{0xF97925D837AE806AULL, "vxlt_gallowstemplar_f_39g_0_0.vlm"},
	{0xF9792C91F46D8CFDULL, "llac_lt_gallowscourtyard_n_-1_2.dds"},
	{0xF979ACC0895DD996ULL, "vxw_bonepitmining_a2_28h_0_0.vlm"},
	{0xF979C89F0E3A1BB4ULL, "gen00fl_varric6093578.cl"},
	{0xF979F6123E552C3FULL, "character_character_p0.psh.fxo.dx11.assembly"},
	{0xF979F7D39DBAE4F9ULL, "237897610.ogg"},
	{0xF97A0B4DD10A8B82ULL, "mcr371ar_gallows_templar_nt.ncs"},
	{0xF97A865D85955B1EULL, "vxht_0_hightown_au5_0_0.vlm"},
	{0xF97ABC9A9F97C7EEULL, "and230_mage_girl6061854.cl"},
	{0xF97B1D0CB53787FDULL, "vxca_caverns1_e7p_0_0.vlm"},
	{0xF97B2E5E8471A44EULL, "475593805.ogg"},
	{0xF97BBD0C93BD8AE2ULL, "6085276_m.fxe"},
	{0xF97BD798156C5D80ULL, "161060534.ogg"},
	{0xF97BE09B4131DB7CULL, "395224665.ogg"},
	{0xF97C035A3A56B5A1ULL, "hor_smo_bri.mao"},
	{0xF97C5C9078D53D68ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_mag_robc_0d.dds"},
	{0xF97CB4653260B976ULL, "301959830.ogg"},
	{0xF97CF3C8050BE66BULL, "dre000_bethany6037345.cl"},
	{0xF97DFA4B3A6E5F0FULL, "mag291_torpor6026942.cl"},
	{0xF97E875115086183ULL, "w_blightlands_71_0.msh"},
	{0xF97E9551150879DDULL, "w_blightlands_71_0.mmh"},
	{0xF97E9951150880EEULL, "w_blightlands_71_0.mao"},
	{0xF97EA676B00D446BULL, "6061791_f.fxe"},
	{0xF97EE9F5FF096893ULL, "one280_iwan6096750.cl"},
	{0xF980081ED0AE7D86ULL, "48351106.ogg"},
	{0xF980422DB6219522ULL, "gen00fl_carver.cnv"},
	{0xF980446D6F01AE4FULL, "6092292_f.fxe"},
	{0xF9809D69779E145BULL, "656164962.ogg"},
	{0xF981234DD110D313ULL, "mcr371ar_gallows_templar_nt.lst"},
	{0xF9814824DCF36AA6ULL, "vxht_bartrand_n_65v_0_0.vlm"},
	{0xF981907A7753FF62ULL, "133670602.ogg"},
	{0xF981A724BD7BD8E5ULL, "70758341.ogg"},
	{0xF9820F7EAA31576AULL, "dae100_bran6118296.cl"},
	{0xF982EDE0F374C724ULL, "83676688.ogg"},
	{0xF983242DB623985AULL, "gen00fl_carver.bnk"},
	{0xF98338191BA7DCAEULL, "hlac_ht_0_hightown_n_-65536_2x.dds"},
	{0xF983A45298308775ULL, "gen00fl_isabela6088102.cl"},
	{0xF983DB6929D24EA4ULL, "6053277_m.fxe"},
	{0xF9844B2A9C5A0178ULL, "lgt110_gxa_vgk_silver_order6117027.cl"},
	{0xF98465411F28B611ULL, "vxht_0_hightown_b7w_0_0.vlm"},
	{0xF984835815A3606BULL, "vxht_brothel_n_5gj_0_0.vlm"},
	{0xF9849F6935A84B0FULL, "vxlt_0_lowtown_haa_0_0.vlm"},
	{0xF984E9164103E435ULL, "485737807.ogg"},
	{0xF984EC182A23CFF0ULL, "6113934_f.fxe"},
	{0xF9851147BCE560F5ULL, "6044917_m.fxe"},
	{0xF985876B3D18F51CULL, "vxht_mansion2_1ce_0_0.vlm"},
	{0xF9859964934BCE96ULL, "vxdr_primevaltomb_7zq_0_2.vlm"},
	{0xF985D29F3F740D75ULL, "760507495.ogg"},
	{0xF985DA081C6E1C47ULL, "lowtown_male_urchin_sta6164212.cl"},
	{0xF98613F3DD58EF36ULL, "hlac_dr_slavers_3214_1y.dds"},
	{0xF986350411198622ULL, "6143602_m.fxe"},
	{0xF9863A86F03CEEF4ULL, "93334509.ogg"},
	{0xF9863CD078FFDDD0ULL, "hlac_ca_cavern1_haunted_24422_0z.dds"},
	{0xF986EBF7701CFD65ULL, "vxca_caverns2_9ft_0_0.vlm"},
	{0xF987D7DBE03A6E62ULL, "vxlt_0_lowtown_a2_gbz_0_0.vlm"},
	{0xF987D8087869C971ULL, "6202298_m.fxe"},
	{0xF987E5BE60A0587DULL, "vxlt_backalley_g_1u3_0_0.vlm"},
	{0xF987EEB90855F176ULL, "fen210_hadriana6110566.cl"},
	{0xF9880D48137CD470ULL, "6163664_m.fxe"},
	{0xF988135F2E0A99C4ULL, "gen00fl_varric6079800.cl"},
	{0xF9884A479379DA38ULL, "vxw_woundedcoast_1g1_0_0.vlm"},
	{0xF98861458023E79FULL, "dae000cr_grand_cleric6147716.cl"},
	{0xF988DE5747446CB9ULL, "vxw_bonepitdestroyed_a3_27t_0_0.vlm"},
	{0xF989197451664E19ULL, "409210393.ogg"},
	{0xF9895D8259C665B0ULL, "297367912.ogg"},
	{0xF98A07008EE1C3B8ULL, "vxlt_darktown_54s_0_0.vlm"},
	{0xF98A24F74A8A1902ULL, "vxlt_0_lowtown_ewc_0_0.vlm"},
	{0xF98A6311B116529CULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_m_p.impa.ani"},
	{0xF98AA6C90CC14801ULL, "hlac_dr_dungeon_3216_5z.dds"},
	{0xF98AB0EE50B107D3ULL, "zz_fen_debug6134388.cl"},
	{0xF98C316EA6DF188BULL, "375198121.ogg"},
	{0xF98D52520EE1A0D1ULL, "1068643690.ogg"},
	{0xF98DEE4993294946ULL, "vxlt_darktown_4ik_0_0.vlm"},
	{0xF98E2B8E19444A16ULL, "165749842.ogg"},
	{0xF98E9EA32CA04472ULL, "vxlt_hangedman_n_1hw_0_0.vlm"},
	{0xF98EE09F15B768B9ULL, "vxlt_0_lowtown_a2_f4z_0_0.vlm"},
	{0xF98FB09CA7E519EDULL, "ht_0_hightown_probe_9x4.dds"},
	{0xF98FC5CD646350A8ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax107a_0s.dds"},
	{0xF98FF0DF2BADB64EULL, "709880497.ogg"},
	{0xF98FF362C1C83056ULL, "hlac_lt_gallowstemplar_f_2382_2z.dds"},
	{0xF990231D95829624ULL, "6142323_f.fxe"},
	{0xF99073CEA9D1C0E7ULL, "zz_qcr_debug6215493.cl"},
	{0xF9917011B11C87DEULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_m_p.impa.gad"},
	{0xF991BCFA8BBE09B6ULL, "hairdebug.psh.fxo"},
	{0xF99298FB78D5BB76ULL, "zz_one_debug6074015.cl"},
	{0xF992A1DCBDFA09FCULL, "6112111_f.fxe"},
	{0xF992F535079978EBULL, "vxht_brothel_n_5jm_0_0.vlm"},
	{0xF9937B316DA5E99EULL, "557139956.ogg"},
	{0xF993DEB2ED4CBFCFULL, "art\\characters\\weapons\\staff\\textures\\w_stf_014a_0t.dds"},
	{0xF993F9482F0F0D9CULL, "vxht_mansion2_1fz_0_0.vlm"},
	{0xF9947F24DAC100A3ULL, "6071164_m.fxe"},
	{0xF994A2553FE1A474ULL, "art\\levels\\races\\proxy\\props\\_textures\\cart_d.dds"},
	{0xF994F855D639D333ULL, "84846999.ogg"},
	{0xF99545FC4FC58AB2ULL, "6087116_f.fxe"},
	{0xF995604B54CC4C17ULL, "vxlt_gallowstemplar_n_5ql_0_0.vlm"},
	{0xF996081818F89BAEULL, "hlac_ca_caverns1_24425_3y.dds"},
	{0xF99624A6E87EC4B9ULL, "llac_ca_caverns1_mines_24425_0.dds"},
	{0xF996286EE000DEC2ULL, "vxca_caverns2_8fb_0_0.vlm"},
	{0xF996579B529670FDULL, "vxlt_backalley_g_1r2_0_0.vlm"},
	{0xF996906B450A7084ULL, "zz_one_debug6099940.cl"},
	{0xF996B4964D58BB0EULL, "hlac_ca_caverns2_5802_8y.dds"},
	{0xF996E3CE640742A6ULL, "vxht_mansion2b_2k0_0_0.vlm"},
	{0xF9971A3E9EE45623ULL, "513108377.ogg"},
	{0xF9971EBBA5D8BB56ULL, "vxht_mansion2_1a9_0_0.vlm"},
	{0xF9977F93FE2EB8FEULL, "mag370_meredith6046873.cl"},
	{0xF997B7D01D279E73ULL, "1046206743.ogg"},
	{0xF998164B3E926D37ULL, "vxca_cavern1_haunted_ego_0_0.vlm"},
	{0xF99838D96079945BULL, "6114451_f.fxe"},
	{0xF9984B11B12267FDULL, "art\\vfx\\combat\\vfx_m_pri_fb_imp_m_p.impa.evt"},
	{0xF999F50BFAC395C3ULL, "qun110_petrice6042208.cl"},
	{0xF99AD7A51FA9E971ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_hlm_masc_0d.dds"},
	{0xF99ADA206287B3E3ULL, "3020420.ogg"},
	{0xF99B869EB37A6156ULL, "zz_one_debug6051528.cl"},
	{0xF99C5E1F44B5F068ULL, "6116072-mrl_pc_wanderf2.cut"},
	{0xF99C93339D3D41DCULL, "ht_0_hightown_n_12_0.anb"},
	{0xF99C9854710DC0F3ULL, "7750978.ogg"},
	{0xF99CD0434DC58CF3ULL, "85227513.ogg"},
	{0xF99D0C26CE319A6FULL, "hlac_dr_0_deeproads_5849_2y.dds"},
	{0xF99DB1279B9DA8DBULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_sluce.phy"},
	{0xF99DFF1BBE50FC6FULL, "mcr370_meredith6085874.cl"},
	{0xF99E04A609485CD6ULL, "lt_0_lowtown_docks_a3_n_137_0.mao"},
	{0xF99E10A609487175ULL, "lt_0_lowtown_docks_a3_n_137_0.mmh"},
	{0xF99E12A6094874EBULL, "lt_0_lowtown_docks_a3_n_137_0.msh"},
	{0xF99E56ECEB8CBEB7ULL, "hlac_lt_0_lowtown_docks_a3_131071_1x.dds"},
	{0xF99E9756E1120E83ULL, "store.gfx"},
	{0xF99EB0A7D0EDEEF4ULL, "vxw_bonepit_1u8_0_0.vlm"},
	{0xF99EDD8342025EA7ULL, "152546789.ogg"},
	{0xF99EFDB951353DF4ULL, "6020554_m.fxe"},
	{0xF99F2048A0B27186ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_disappea.evt"},
	{0xF99F70C149003E96ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\blendtree\\qm_exp_idle.anb"},
	{0xF99F93B0BFE86753ULL, "llac_w_bonepitmining_a2_393220_0.dds"},
	{0xF99FEAB70F564CB8ULL, "34030242.ogg"},
	{0xF9A060CD12936F53ULL, "hlac_lt_hangedman_n_-1_24y.dds"},
	{0xF9A0689894790B7BULL, "628199734.ogg"},
	{0xF9A072487213F78FULL, "hlac_lt_0_lowtown_a2_n_65536_5y.dds"},
	{0xF9A095E3845D9DD5ULL, "llac_lt_0_lowtown_docks_a3_-4_7.dds"},
	{0xF9A0A186D57783FBULL, "6162071_m.fxe"},
	{0xF9A0CBCF86378076ULL, "w_encounter3_218_0.msh"},
	{0xF9A0D5CF86379174ULL, "w_encounter3_218_0.mmh"},
	{0xF9A0D9CF86379787ULL, "w_encounter3_218_0.mao"},
	{0xF9A16711F98F916BULL, "vxht_brothel_n_5ml_0_0.vlm"},
	{0xF9A176CD2DAD4893ULL, "ht_bartrand_n_3366.anb"},
	{0xF9A1E14DD9EE7864ULL, "vxlt_gallowstemplar_f_5oy_0_0.vlm"},
	{0xF9A1EA9E3140D00EULL, "w_bonepitmining_a2_194_0.msh"},
	{0xF9A1F89E3140E85FULL, "w_bonepitmining_a2_194_0.mao"},
	{0xF9A2049E3140FCFCULL, "w_bonepitmining_a2_194_0.mmh"},
	{0xF9A2282DB63E4FEFULL, "gen00fl_carver.utc"},
	{0xF9A228C1B4D3C57EULL, "259309992.ogg"},
	{0xF9A26701DAB0E183ULL, "193568542.ogg"},
	{0xF9A2860AF72CE4F9ULL, "577699907.ogg"},
	{0xF9A2E96FE9BE8C07ULL, "vxht_0_hightown_9zq_0_0.vlm"},
	{0xF9A34819AC6D1EBFULL, "gen00fl_sebastian6180513.cl"},
	{0xF9A383774384C1B9ULL, "hm_friedrich.mor"},
	{0xF9A44C91277A0275ULL, "hlac_ht_mansion2b_818_23z.dds"},
	{0xF9A45A16AE1C40B8ULL, "vxlt_undercitydungeon_yj_0_0.vlm"},
	{0xF9A48E943F48D3A4ULL, "6043984_m.fxe"},
	{0xF9A4AE1071472FFDULL, "412706223.ogg"},
	{0xF9A4ECC613138B3EULL, "6074491_f.fxe"},
	{0xF9A4FE032A9710F4ULL, "llac_w_encounter2_camp_-65538_0.dds"},
	{0xF9A50B4143566FFEULL, "vxlt_hangedman_n_2xp_0_0.vlm"},
	{0xF9A54D263346646CULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef02.msh"},
	{0xF9A5532633466E5EULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef02.mmh"},
	{0xF9A555AB55FD5B26ULL, "vxht_mansion2b_2d1_0_0.vlm"},
	{0xF9A55B99BD7CA6EBULL, "6001211_m.fxe"},
	{0xF9A5AD26E9CBB776ULL, "llac_ht_keep_n_3743_1.dds"},
	{0xF9A63748A0B8B7D1ULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_disappea.gad"},
	{0xF9A65C844A1FCA65ULL, "dae000_foreshadow.bnk"},
	{0xF9A6711C6DC2FAD6ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage1.ani"},
	{0xF9A6D42A9C818BA4ULL, "vxlt_0_lowtown_a2_n_fff_0_0.vlm"},
	{0xF9A708B12E764A02ULL, "vxlt_0_lowtown_emm_0_0.vlm"},
	{0xF9A72A00063CA962ULL, "hlac_w_woundedcoast_-131065_0z.dds"},
	{0xF9A73EB6024B2430ULL, "vxw_bonepit_l8_0_0.vlm"},
	{0xF9A77EF482722DA3ULL, "mcr310_confused_guard6077087.cl"},
	{0xF9A7C1993F9B8E5DULL, "6060098_m.fxe"},
	{0xF9A7DC1E07782535ULL, "6180420_m.fxe"},
	{0xF9A8397EF5ECB625ULL, "framebuffereffects_fadedistortion_p0.psh.fxo.dx11"},
	{0xF9A8A98BE3536E10ULL, "vxht_0_hightown_blg_0_0.vlm"},
	{0xF9A907887467C40DULL, "vxht_0_hightown_a2_9zg_0_0.vlm"},
	{0xF9A9C980C7B6A397ULL, "zz_followers_banter6158259.cl"},
	{0xF9AA2D149646DC9EULL, "174548799.ogg"},
	{0xF9AA3E844A237F9DULL, "dae000_foreshadow.cnv"},
	{0xF9AA4179B259377AULL, "hm_noble_gen_1.mor"},
	{0xF9AA799C0530FE36ULL, "1022652370.ogg"},
	{0xF9AADDE8070687F1ULL, "hlac_ca_cavern1_haunted_24427_3y.dds"},
	{0xF9AB7B70F0874A7EULL, "757402915.ogg"},
	{0xF9AB825D108C7572ULL, "vxlt_hangedman_n_1bq_0_0.vlm"},
	{0xF9AC3848A0BD257FULL, "art\\characters\\creatures\\gibgolem\\animation\\head\\gibsgolem_demonhead_disappea.ani"},
	{0xF9AC8586CD2A6E6AULL, "138358266.ogg"},
	{0xF9ACF9E056EDC101ULL, "vxht_0_hightown_a3_bps_0_0.vlm"},
	{0xF9AD00EDF6BFD651ULL, "vxlt_0_lowtown_a3_n_c_hsq_0_0.vlm"},
	{0xF9AD08B04F588AC9ULL, "987857111.ogg"},
	{0xF9AD53AD31C3C011ULL, "6061335_f.fxe"},
	{0xF9ADA142C1194CFCULL, "6207793_m.fxe"},
	{0xF9ADA25F00086D4CULL, "mer320_hubert6065788.cl"},
	{0xF9AEEFC1C707DC02ULL, "vxca_cave_1aa_0_0.vlm"},
	{0xF9AF1363C76D9C7BULL, "6167371_m.fxe"},
	{0xF9AF213C2014E03AULL, "877239576.ogg"},
	{0xF9AFC350092646B0ULL, "vxca_caverns2prc_9qu_0_0.vlm"},
	{0xF9B01D59672FDA54ULL, "vxca_caverns2prc_8oj_0_0.vlm"},
	{0xF9B0A847FC6AE9FDULL, "face2_d.psh.fxo"},
	{0xF9B1C74591E5D230ULL, "209441080.ogg"},
	{0xF9B21E47E2F82888ULL, "hlac_lt_darktown_-1_57x.dds"},
	{0xF9B2960E21161D8DULL, "150114229.ogg"},
	{0xF9B2D4B506F8B35FULL, "273909577.ogg"},
	{0xF9B302D317914342ULL, "vxht_0_hightown_bgh_0_0.vlm"},
	{0xF9B3791C6DCD92C7ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage1.evt"},
	{0xF9B3C78847F373A6ULL, "vxht_mansion2b_2a2_0_0.vlm"},
	{0xF9B3E6F3DAB8E3E7ULL, "6139299_m.fxe"},
	{0xF9B434CA7655846BULL, "hlac_w_sundermount_-262150_0y.dds"},
	{0xF9B4A920672795D1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_loop.gad"},
	{0xF9B4E937ED1DDB30ULL, "800356598.ogg"},
	{0xF9B546078E77A424ULL, "vxlt_0_lowtown_a2_n_fag_0_0.vlm"},
	{0xF9B54C433A5DFCB7ULL, "102843908.ogg"},
	{0xF9B57B1F321DB805ULL, "mag211_marethari6021353.cl"},
	{0xF9B5803596CBA53EULL, "starline2.dds"},
	{0xF9B63F7A0AF890ADULL, "806742790.ogg"},
	{0xF9B65DB53412ED8DULL, "gen00fl_aveline6106379.cl"},
	{0xF9B6B222B65B2368ULL, "6173171_m.fxe"},
	{0xF9B6D4B522B4D0B3ULL, "672555461.ogg"},
	{0xF9B6E432E3B04DDFULL, "vxdr_0_deeproads_dw0_0_0.vlm"},
	{0xF9B6FD89F7BDE865ULL, "6039962_m.fxe"},
	{0xF9B7509DF438EE22ULL, "art\\vfx\\creatures\\fxc_withfn_c.dura.gad"},
	{0xF9B8478F61326DD3ULL, "vxdr_primevalentrance_21n_0_0.vlm"},
	{0xF9B8DD1F4C4636D8ULL, "hightown_male_noble2_mob6170932.cl"},
	{0xF9B9C8E43742D0F3ULL, "445098876.ogg"},
	{0xF9BA3625C55C3F3CULL, "118556345.ogg"},
	{0xF9BA9A1C6DD3E910ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage3ball\\runtransitiontostage1.gad"},
	{0xF9BAAA20672C037FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_loop.ani"},
	{0xF9BB2045B344F956ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.cess.evt"},
	{0xF9BB59B9CDA1D272ULL, "atl_startmenu_dxt5_00.dds"},
	{0xF9BB72CAE8B5EED1ULL, "vxlt_0_lowtown_a3_n_c_hxp_0_0.vlm"},
	{0xF9BC504078DA221BULL, "6042399_m.fxe"},
	{0xF9BC9AF817EDDE72ULL, "6127125_m.fxe"},
	{0xF9BD1498B72B79AEULL, "belt_evra.dds"},
	{0xF9BDCEB05E2C0768ULL, "vxdr_primevalentrance_16k_0_0.vlm"},
	{0xF9BE037D00959FA5ULL, "hlac_lt_0_lowtown_docks_a3_-196610_2x.dds"},
	{0xF9BE1253AD8E47AFULL, "6119873_m.fxe"},
	{0xF9BE739DF43F4819ULL, "art\\vfx\\creatures\\fxc_withfn_c.dura.evt"},
	{0xF9BEBCE12AD8C6B0ULL, "dae000_arishok6081306.cl"},
	{0xF9BEBF4A7EB0DE14ULL, "gen00fl_isabela6135641.cl"},
	{0xF9BEDA7F86408128ULL, "lt_0_lowtown_docks_a3_n_110_0.mmh"},
	{0xF9BEE1835C559643ULL, "6136467_m.fxe"},
	{0xF9BEE67F864095CBULL, "lt_0_lowtown_docks_a3_n_110_0.mao"},
	{0xF9BEF87F8640B452ULL, "lt_0_lowtown_docks_a3_n_110_0.msh"},
	{0xF9BF44F1550587A3ULL, "hlac_lt_0_lowtown_a2_65537_0x.dds"},
	{0xF9BFC8E2ACBD4F0CULL, "downsample.mao"},
	{0xF9C0D2214D28EC58ULL, "6036719_m.fxe"},
	{0xF9C10F11AB81AA8FULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_loop_c.impa.gad"},
	{0xF9C1691F1F1292DAULL, "hlac_w_encounter1_-2_0x.dds"},
	{0xF9C1870077F00114ULL, "sw_dot.mao"},
	{0xF9C21745B34B08E1ULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.cess.gad"},
	{0xF9C25FE433E98D53ULL, "vxlt_hangedman_10q_0_0.vlm"},
	{0xF9C3044A6E8CF0CFULL, "537369637.ogg"},
	{0xF9C33A69BF4D5B46ULL, "253383299.ogg"},
	{0xF9C3ED858D53380DULL, "6066355_m.fxe"},
	{0xF9C4481CCB9C32AFULL, "124345587.ogg"},
	{0xF9C46CEA656BB390ULL, "vxht_mansion_n_4y9_0_0.vlm"},
	{0xF9C51267BB4E8873ULL, "177723266.ogg"},
	{0xF9C53D6F12E48188ULL, "6125171_f.fxe"},
	{0xF9C578954F780EB2ULL, "llac_lt_0_lowtown_a3_n_c_1_2.dds"},
	{0xF9C584F8262263DFULL, "6031285_m.fxe"},
	{0xF9C5E62CF592FB96ULL, "gen_im_acc_rng_act3_12.uti"},
	{0xF9C628092820F296ULL, "hlac_ht_playershouse_n_855_7z.dds"},
	{0xF9C653D6E5F47A33ULL, "6016331_f.fxe"},
	{0xF9C703C6DA96C511ULL, "art\\vfx\\textures\\vfx_tear_cs.dds"},
	{0xF9C7081CAABD9B5DULL, "252775605.ogg"},
	{0xF9C71375C27FEDBBULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_avgdesk_n.dds"},
	{0xF9C72411AB863A3CULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_loop_c.impa.evt"},
	{0xF9C7B607F8A0BCB0ULL, "hlac_ht_0_hightown_a3_65535_2z.dds"},
	{0xF9C81845B34F758FULL, "art\\vfx\\combat\\vfx_m_ent_hexdamage_c.cess.ani"},
	{0xF9C81EE44512EE89ULL, "592321258.ogg"},
	{0xF9C8F2B9ED13C0FCULL, "llac_lt_0_lowtown_n_-196609_2.dds"},
	{0xF9C96B3C091F0269ULL, "vxca_caverns2_8ob_0_0.vlm"},
	{0xF9C998279BC2B97AULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_sluce.msh"},
	{0xF9C99FCF161BB676ULL, "vxht_mansion2_2lq_0_0.vlm"},
	{0xF9C9A6279BC2D0B3ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_sluce.mao"},
	{0xF9C9AA279BC2D7C0ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_sluce.mmh"},
	{0xF9C9E4A7DAAC0751ULL, "vxlt_0_lowtown_a3_n_c_hew_0_0.vlm"},
	{0xF9CAFF2F73859A0EULL, "6122851_f.fxe"},
	{0xF9CAFF4D8B4BFCEBULL, "909003774.ogg"},
	{0xF9CB3217A719B924ULL, "6180951_m.fxe"},
	{0xF9CB7B9DF449E008ULL, "art\\vfx\\creatures\\fxc_withfn_c.dura.ani"},
	{0xF9CBE099F53E2BD2ULL, "vxlt_hangedman_10t_0_0.vlm"},
	{0xF9CCCCF4D116720EULL, "840363673.ogg"},
	{0xF9CD7B564B42B8C0ULL, "587448.ogg"},
	{0xF9CD8D292192CB9AULL, "isa211_isabela6202265.cl"},
	{0xF9CD93CD2DD2B54FULL, "ht_bartrand_n_3366.rml"},
	{0xF9CDA24E0D2D0CB9ULL, "dae000cr_grand_cleric6147707.cl"},
	{0xF9CDC373F58048FCULL, "397997431.ogg"},
	{0xF9CEDA3E7A2515C3ULL, "gen00fl_varric6103555.cl"},
	{0xF9D097123A911DD4ULL, "gen00fl_merrill6126308.cl"},
	{0xF9D09B3BAE76D9CCULL, "6067708_m.fxe"},
	{0xF9D0AF8D03C00360ULL, "gen00fl_anders6103783.cl"},
	{0xF9D1360CA6242C67ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_relaxed_loop.gad"},
	{0xF9D154065B83443CULL, "02_bloodimpact_2.mao"},
	{0xF9D1AE0ED4920BDBULL, "llac_lt_0_lowtown_a3_n_c_-1_2.dds"},
	{0xF9D1F1C8B0C9CCA1ULL, "vxca_caverns1_j4r_0_0.vlm"},
	{0xF9D308B7F46120E8ULL, "990026064.ogg"},
	{0xF9D376A83E16A54AULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_wf_injured.gad"},
	{0xF9D376FC333228FBULL, "llac_lt_0_lowtown_65535_0.dds"},
	{0xF9D390869E2BAFD3ULL, "vxht_mansion_n_52r_0_0.vlm"},
	{0xF9D390F7469E2CF5ULL, "hlac_dr_slavers_3207_7x.dds"},
	{0xF9D3CD21B8E95019ULL, "6060142_m.fxe"},
	{0xF9D40E7732DBF4BDULL, "vxht_keep_6g3_0_0.vlm"},
	{0xF9D4DAE9015E9CA5ULL, "character_character_p204.psh.fxo.dx11.assembly"},
	{0xF9D52C11AB928555ULL, "art\\vfx\\combat\\vfx_w_def_dam_mode_loop_c.impa.ani"},
	{0xF9D55B603DFBBFA1ULL, "w_dgr_dg06a_0d.dds"},
	{0xF9D620F448A9FD11ULL, "vxdr_primevaltomb_6qd_0_0.vlm"},
	{0xF9D673DE8CDF99E3ULL, "gen00fl_aveline6113991.cl"},
	{0xF9D67CFCABC9686FULL, "dae011_corff_sta6164635.cl"},
	{0xF9D6FB1C7F3CD05CULL, "dae320st_merch_weapon.utm"},
	{0xF9D720380FD5D6A4ULL, "lt_0_lowtown_docks_a3_16_0_lc.mmh"},
	{0xF9D811AC0811CEF6ULL, "vxht_mansion2_2ir_0_0.vlm"},
	{0xF9D85684CCA21FD1ULL, "vxlt_0_lowtown_a3_n_c_hbv_0_0.vlm"},
	{0xF9D8950EDAFABBC2ULL, "vxlt_gallowstemplar_5bw_0_0.vlm"},
	{0xF9D8CCF1818E4BFAULL, "gen00fl_sebastian6174070.cl"},
	{0xF9D8E505F5338D4AULL, "vxw_encounter3_ct_0_0.vlm"},
	{0xF9D954D3E697DB20ULL, "bld_dri_tu.mao"},
	{0xF9D9661E351BD0EFULL, "hairalpha_hair_p0.psh.fxo.dx11.assembly"},
	{0xF9D9BE0DDC602ACAULL, "vxca_caverns1_ih2_0_0.vlm"},
	{0xF9DA0F38F40B547BULL, "547140424.ogg"},
	{0xF9DA124E21EC7D8DULL, "626457227.ogg"},
	{0xF9DA6682B1736561ULL, "38239408.ogg"},
	{0xF9DA72DF0245D2AAULL, "zz_dae_debug6029341.cl"},
	{0xF9DABFF2BBDDC64AULL, "vxht_keep_6bo_0_0.vlm"},
	{0xF9DB0980D7F1BD32ULL, "propalpha1_d.psh.fxo"},
	{0xF9DB18E6DF089030ULL, "vxca_caverns2prc_9fp_0_0.vlm"},
	{0xF9DB225EDDBD80C7ULL, "6154711_m.fxe"},
	{0xF9DB2CEEB80AA1D9ULL, "6161160_m.fxe"},
	{0xF9DBFB8C3537955EULL, "vrc321st_spooky_scene_7.stg"},
	{0xF9DC91B495DA805BULL, "art\\characters\\weapons\\longswords\\w_lsw_015a.mmh"},
	{0xF9DC92339D73246CULL, "ht_0_hightown_n_12_0.rml"},
	{0xF9DC9DB495DA94F8ULL, "art\\characters\\weapons\\longswords\\w_lsw_015a.mao"},
	{0xF9DCAFB495DAB345ULL, "art\\characters\\weapons\\longswords\\w_lsw_015a.msh"},
	{0xF9DD48999CA38ED3ULL, "gen00fl_merrill6133145.cl"},
	{0xF9DD55A35464C464ULL, "lt_0_lowtown_docks_n_26_0_ll.mmh"},
	{0xF9DDA3F184CFBE50ULL, "vxw_woundedcoast_67_0_0.vlm"},
	{0xF9DDB60881FBCECFULL, "fen110_anso.bnk"},
	{0xF9DE6E66D069C57AULL, "cd_fen310_varania.bnk"},
	{0xF9DEBC2557B5CB92ULL, "em_gen_act1.mor"},
	{0xF9DED21605938918ULL, "llac_dr_primevalentrance_1726_8.dds"},
	{0xF9DEF447D2E0776AULL, "6041944_m.fxe"},
	{0xF9DFA512E9E18104ULL, "dae000ip_crowd_em4_talk16.utp"},
	{0xF9DFFEB12340112BULL, "6149083_m.fxe"},
	{0xF9E00908B633CF2EULL, "vxca_cave_2eb_0_0.vlm"},
	{0xF9E09187D0B7683AULL, "w_bonepitmining_a2_4_0.mmh"},
	{0xF9E09587D0B76F71ULL, "w_bonepitmining_a2_4_0.mao"},
	{0xF9E0A0B36ADCFD46ULL, "891176334.ogg"},
	{0xF9E0A387D0B78680ULL, "w_bonepitmining_a2_4_0.msh"},
	{0xF9E0D80881FE3E37ULL, "fen110_anso.cnv"},
	{0xF9E0E6EAAD2BAA12ULL, "mcr370_assignments6176746.cl"},
	{0xF9E15143544577FBULL, "gen00fl_aveline6114806.cl"},
	{0xF9E17BB61934756EULL, "672612477.ogg"},
	{0xF9E23FC3C3185474ULL, "1043537542.ogg"},
	{0xF9E25664EE435C74ULL, "hlac_lt_gallowstemplar_n_2382_0z.dds"},
	{0xF9E270FFAB2BD010ULL, "vxht_0_hightown_bxc_0_0.vlm"},
	{0xF9E33F4779151D93ULL, "gen00fl_isabela6079027.cl"},
	{0xF9E3439447A6FCC8ULL, "7903913.ogg"},
	{0xF9E34E7CCA832D2DULL, "816627041.ogg"},
	{0xF9E3A3BE55F26FAAULL, "hlac_w_encounter2_camp_1_0y.dds"},
	{0xF9E44505D9631417ULL, "zz_qun_debug6028893.cl"},
	{0xF9E48AEFB94D9D84ULL, "6072874_m.fxe"},
	{0xF9E4BC03442B7F8BULL, "6143221_m.fxe"},
	{0xF9E5230CA634B4ADULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_relaxed_loop.ani"},
	{0xF9E52A26337C0CF9ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef02.phy"},
	{0xF9E53C704BA1F55EULL, "ser270_cut_aftermath6162297.cl"},
	{0xF9E6071BCDA8AE7BULL, "mrl000_marethari6129404.cl"},
	{0xF9E68388FA07E776ULL, "vxht_mansion2_2rs_0_0.vlm"},
	{0xF9E686DF513E2CA4ULL, "754035112.ogg"},
	{0xF9E7EAF9206CF5C1ULL, "vxca_caverns1_mines_hkx_0_0.vlm"},
	{0xF9E84A0950276C74ULL, "6167871_m.fxe"},
	{0xF9E851A83E28C230ULL, "art\\characters\\playercharacter\\elfmale\\animation\\exploration\\movement\\mov_wf_injured.ani"},
	{0xF9E88808430BC25AULL, "vxlt_0_lowtown_n_gj5_0_0.vlm"},
	{0xF9E89F29E1B46F7DULL, "t3_rob_kwl.tnt"},
	{0xF9E8E228F296CEADULL, "qun231_party_comment6077136.cl"},
	{0xF9E92DDB4F7D3ABBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v3.evt"},
	{0xF9E9A201BA31D178ULL, "804735777.ogg"},
	{0xF9EA0E3528B835C7ULL, "vxht_mansion2b_1u8_0_0.vlm"},
	{0xF9EA0F79067669F9ULL, "lt_0_lowtown_docks_157_0.msh"},
	{0xF9EA129CB71A8713ULL, "vxw_bonepit_4w_0_0.vlm"},
	{0xF9EA197906767AE7ULL, "lt_0_lowtown_docks_157_0.mmh"},
	{0xF9EA1D7906768134ULL, "lt_0_lowtown_docks_157_0.mao"},
	{0xF9EA200BD6BA75EDULL, "6053150_f.fxe"},
	{0xF9EA445DBE4387A0ULL, "6139959_m.fxe"},
	{0xF9EAD63796191CFDULL, "rog_burstshot.dds"},
	{0xF9EB1F818AA91790ULL, "vxlt_gallowstemplar_5ql_0_0.vlm"},
	{0xF9EB50971DA09B7DULL, "6121951_m.fxe"},
	{0xF9EC0C30B3C34553ULL, "matdef_dw_specular.dds"},
	{0xF9EC5023035B8FBDULL, "253461077.ogg"},
	{0xF9ECC37C3D0F9E54ULL, "mag130_danzig6037290.cl"},
	{0xF9ECCA46DF69A542ULL, "vxht_0_hightown_bst_0_0.vlm"},
	{0xF9ED945E1E2B20D0ULL, "42189790.ogg"},
	{0xF9EF1F830D4A97F5ULL, "zz_dae_debug6047612.cl"},
	{0xF9EF26DB4F819AFCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v3.gad"},
	{0xF9EF430E1BE0AA38ULL, "371084234.ogg"},
	{0xF9EFE39B0DE56BD9ULL, "vxht_bartrand_n_5qj_0_0.vlm"},
	{0xF9EFF820C23B8C05ULL, "gen_im_npc_arm_hum_hlm_mrcb.uti"},
	{0xF9F05E296205FE0AULL, "641980226.ogg"},
	{0xF9F0897A348AF438ULL, "340958827.ogg"},
	{0xF9F0F23116C825BDULL, "vxht_keep_6m1_0_0.vlm"},
	{0xF9F16851BBB5D3C4ULL, "seb200im_twilight_belt.uti"},
	{0xF9F2FFEF74183DF4ULL, "6031400_m.fxe"},
	{0xF9F40A4D0F2CA441ULL, "art\\effects\\cloudshadows\\w_encounter3\\cloud1.dds"},
	{0xF9F41DD0A76162AFULL, "6080160_m.fxe"},
	{0xF9F4A54FF6764B5EULL, "6119847_m.fxe"},
	{0xF9F4AF2330C80964ULL, "fx_dust07.dds"},
	{0xF9F4BDBBB88BE2DCULL, "vxca_cavern1_haunted_imv_0_0.vlm"},
	{0xF9F4D8097A19F462ULL, "vxlt_gallowsprison_n_6cd_0_0.vlm"},
	{0xF9F51DD021CF46D9ULL, "ht_mansion_n_2995.anb"},
	{0xF9F566BCFCD59A22ULL, "342843312.ogg"},
	{0xF9F5B9F203162541ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02a.phy"},
	{0xF9F5DDB261857402ULL, "fx_floor_tile.dds"},
	{0xF9F5E61D33427ED2ULL, "6108997_m.fxe"},
	{0xF9F625DB4F87B7FAULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\die_v3.ani"},
	{0xF9F649E04643C734ULL, "gen00fl_aveline6137714.cl"},
	{0xF9F6853239DA28D8ULL, "941745095.ogg"},
	{0xF9F6BB1354CF1301ULL, "6197709_m.fxe"},
	{0xF9F7431D3D6B7C82ULL, "294824144.ogg"},
	{0xF9F76A6EEFA75F77ULL, "vxlt_0_lowtown_a2_g92_0_0.vlm"},
	{0xF9F7A3AC9FC9F74AULL, "vxht_keep_6hi_0_0.vlm"},
	{0xF9F7B877378BBAE1ULL, "158839014.ogg"},
	{0xF9F7CF0D8B5E6EA0ULL, "vxlt_gallowscourtyard_43m_0_0.vlm"},
	{0xF9F82E50C5590F0FULL, "vxlt_0_lowtown_h9i_0_0.vlm"},
	{0xF9F84DA330FE5718ULL, "644935979.ogg"},
	{0xF9F88668ED2A9647ULL, "vxlt_gallowscourtyard_n_50i_0_0.vlm"},
	{0xF9F89CDB31790EA2ULL, "zz_fol_debug6098092.cl"},
	{0xF9F8E60EB1150BD2ULL, "787505910.ogg"},
	{0xF9F8FFFE12C7C743ULL, "follower_banter6165367.cl"},
	{0xF9F957BCCD88CCFCULL, "art\\levels\\ztools\\default\\generic\\9ff_rockpile.mao"},
	{0xF9F963BCCD88E09FULL, "art\\levels\\ztools\\default\\generic\\9ff_rockpile.mmh"},
	{0xF9F969BCCD88EB41ULL, "art\\levels\\ztools\\default\\generic\\9ff_rockpile.msh"},
	{0xF9F9A890067872D4ULL, "lt_0_lowtown_docks_a3_n_153_0.phy"},
	{0xF9F9C707459100E6ULL, "hlac_lt_0_lowtown_n_-65536_6z.dds"},
	{0xF9FA5D6156FA2E11ULL, "hlac_w_blightlands_131070_0z.dds"},
	{0xF9FA6D4354047916ULL, "lt_0_lowtown_docks_n_162_0.msh"},
	{0xF9FA6DBCFEE4BC61ULL, "zz_fol_debug6097816.cl"},
	{0xF9FA774354048A14ULL, "lt_0_lowtown_docks_n_162_0.mmh"},
	{0xF9FA7B4354049127ULL, "lt_0_lowtown_docks_n_162_0.mao"},
	{0xF9FB6A9DCC97236DULL, "744844446.ogg"},
	{0xF9FC5F2ED364A33BULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.impa.ani"},
	{0xF9FCA231B93AF678ULL, "hlac_lt_0_lowtown_a2_n_-2_2y.dds"},
	{0xF9FCBB3C8FE91035ULL, "combat_smaoe_r1_a4.gda"},
	{0xF9FCC25C40A1E6A1ULL, "genip_barrel_static.utp"},
	{0xF9FCE519DCAC364AULL, "vxht_0_hightown_n_bgc_0_0.vlm"},
	{0xF9FD925BA17DCEEBULL, "hlac_lt_0_lowtown_a2_n_c_65537_3z.dds"},
	{0xF9FDF28522F898BCULL, "llac_lt_0_lowtown_n_-65536_0.dds"},
	{0xF9FE2E1640B58FA8ULL, "vxht_0_hightown_n_c67_0_0.vlm"},
	{0xF9FE5577FFDB8459ULL, "vxht_bartrand_n_5ti_0_0.vlm"},
	{0xF9FE878422963448ULL, "hlac_lt_0_lowtown_docks_131071_3x.dds"},
	{0xF9FEA1601C446B02ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_shelf.msh"},
	{0xF9FEA3601C446EB8ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_shelf.mmh"},
	{0xF9FEAF601C44835BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_shelf.mao"},
	{0xF9FECB08BC941EF8ULL, "134225744.ogg"},
	{0xF9FEEE08F6A14355ULL, "llac_ht_brothel_n_2995_1.dds"},
	{0xF9FFC25CD227772DULL, "w_bonepit_207_0.mmh"},
	{0xF9FFC5CC05A76235ULL, "6114484_f.fxe"},
	{0xF9FFC65CD2277E7EULL, "w_bonepit_207_0.mao"},
	{0xF9FFD45CD22795B3ULL, "w_bonepit_207_0.msh"},
	{0xFA0034F7634D8F79ULL, "vxca_caverns2_7qg_0_0.vlm"},
	{0xFA00B2B18C69B4B6ULL, "llac_ht_mansion_n_2995_3.dds"},
	{0xFA00F9694CC5A0EEULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_countthrone.phy"},
	{0xFA01229173595DD2ULL, "103691480.ogg"},
	{0xFA019EC897518D39ULL, "zz_followers_banter6157542.cl"},
	{0xFA01C46C8F445149ULL, "vfx_rendertoshadowmap_p1.vsh.fxo.dx11"},
	{0xFA02C443E01BF0B0ULL, "718363107.ogg"},
	{0xFA02C4CFA0EAFCB5ULL, "lgt310_shw_ors_informant6156346.cl"},
	{0xFA031C7EB4C2D278ULL, "vxw_sundermount_30n_0_0.vlm"},
	{0xFA032F98AA81FB5CULL, "vxca_cavern1_haunted_ijw_0_0.vlm"},
	{0xFA038866D0890D18ULL, "cd_fen310_varania.wbd"},
	{0xFA03E04C901DF4A4ULL, "255891019.ogg"},
	{0xFA04735E695FDE60ULL, "vxlt_0_lowtown_docks_a3_d1p_0_0.vlm"},
	{0xFA04859006823ED1ULL, "lt_0_lowtown_docks_a3_n_153_0.msh"},
	{0xFA049390068256ECULL, "lt_0_lowtown_docks_a3_n_153_0.mao"},
	{0xFA049F9006826A8FULL, "lt_0_lowtown_docks_a3_n_153_0.mmh"},
	{0xFA04D605409A58EFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps.gad"},
	{0xFA04DB1220B616FDULL, "hlac_lt_backalley_g_0_15y.dds"},
	{0xFA0528D096CEE1FAULL, "art\\characters\\weapons\\shields\\w_shd_002a.msh"},
	{0xFA0536D096CEF933ULL, "art\\characters\\weapons\\shields\\w_shd_002a.mao"},
	{0xFA053AD096CF0040ULL, "art\\characters\\weapons\\shields\\w_shd_002a.mmh"},
	{0xFA05606BC490A426ULL, "6071162_f.fxe"},
	{0xFA0567EE62BC4732ULL, "413361222.ogg"},
	{0xFA0625F53CEEFF60ULL, "34388503.ogg"},
	{0xFA069BA0E8601DD6ULL, "follower_banter6163652.cl"},
	{0xFA06D9D12E2C7D4FULL, "hlac_lt_0_lowtown_a3_n_c_-196609_3y.dds"},
	{0xFA07CB617C18B92EULL, "eye_eye_p1.psh.fxo.dx11.assembly"},
	{0xFA080B01230AA129ULL, "vxlt_0_lowtown_docks_n_b0h_0_0.vlm"},
	{0xFA0835098614DFE5ULL, "6170391_m.fxe"},
	{0xFA08934590BF3C65ULL, "6034374_m.fxe"},
	{0xFA08AA601C4CCF43ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_shelf.phy"},
	{0xFA08D18949587024ULL, "dae000_bartrand6206984.cl"},
	{0xFA09672ED36F3B32ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.impa.evt"},
	{0xFA09AE00C355D642ULL, "vxht_0_hightown_byv_0_0.vlm"},
	{0xFA0A201C6D1E4B58ULL, "103460202.ogg"},
	{0xFA0AEBF4DB0AF1CBULL, "llac_lt_gallowstemplar_n_2394_6.dds"},
	{0xFA0B3B80FE51508DULL, "hla_ht_0_hightown_a2_0_0y.dds"},
	{0xFA0B4070CA16D549ULL, "307433973.ogg"},
	{0xFA0B4652AD2965DFULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_orsi.msh"},
	{0xFA0B4C52AD296F81ULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_orsi.mmh"},
	{0xFA0B5A674F23347CULL, "hlac_lt_darktown_-1_50y.dds"},
	{0xFA0B72C644917222ULL, "hlac_ht_mansion2b_818_45y.dds"},
	{0xFA0B76DC6C455495ULL, "49412081.ogg"},
	{0xFA0B922D9977E4AAULL, "vxlt_darktown_4ig_0_0.vlm"},
	{0xFA0BC025D8DD1EEFULL, "349020393.ogg"},
	{0xFA0BC0C0D80B48D0ULL, "vxlt_0_lowtown_docks_a3_com_0_0.vlm"},
	{0xFA0BD7DC0418DF69ULL, "vxlt_gallowstemplar_5mm_0_0.vlm"},
	{0xFA0BDC81CFE8F3E1ULL, "6078180_f.fxe"},
	{0xFA0BEB0540A09A9CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps.evt"},
	{0xFA0BFB6F40D58FCFULL, "175926221.ogg"},
	{0xFA0C64DE4F373BC2ULL, "6067085_m.fxe"},
	{0xFA0C6A0F04B0AEF9ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\enter_walk_male.gad"},
	{0xFA0C9187D0DCA351ULL, "w_bonepitmining_a2_4_0.rml"},
	{0xFA0C9FF332ABA828ULL, "vxht_0_hightown_n_c34_0_0.vlm"},
	{0xFA0D10B88CE28612ULL, "658713622.ogg"},
	{0xFA0D2AB4A15FB1CCULL, "ico_gloves_light_4.dds"},
	{0xFA0DC699C2FC255DULL, "hlac_lt_0_lowtown_131071_0y.dds"},
	{0xFA0DECB2619A1FA0ULL, "fx_floor_tile.mao"},
	{0xFA0EA09E84E4D0E2ULL, "vxht_0_hightown_a3_b97_0_0.vlm"},
	{0xFA0F55FB24F3BAA5ULL, "658842701.ogg"},
	{0xFA10161649924851ULL, "35610840.ogg"},
	{0xFA10862ED3758EC5ULL, "art\\vfx\\combat\\vfx_m_wpn_ice_crust_c.impa.gad"},
	{0xFA10B33FC9E6F8E2ULL, "mag101_aombanter6052772.cl"},
	{0xFA110D683A5EEF1AULL, "vxca_caverns1_ics_0_0.vlm"},
	{0xFA11A1759C7813DCULL, "vxca_cavern1_haunted_igt_0_0.vlm"},
	{0xFA11E33BB4C533ADULL, "high_dragon_heads.gda"},
	{0xFA123B34C7974BDEULL, "ser120_ghyslain6118399.cl"},
	{0xFA124C60FD073A5FULL, "dae000cr_brothel_orana.utc"},
	{0xFA127B0F04B53747ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\enter_walk_male.ani"},
	{0xFA13BE0BBB356441ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v7.ani"},
	{0xFA13D2C0543C4518ULL, "gen00fl_anders6128335.cl"},
	{0xFA13DC6E47C18697ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_knockback.gad"},
	{0xFA14278F7CD38E7EULL, "140245477.ogg"},
	{0xFA1444D9F81AEC71ULL, "vxw_bonepitdestroyed_a3_27l_0_0.vlm"},
	{0xFA1461FDEF7651C5ULL, "art\\vfx\\materials\\vfx_d_forcevert_blend.mao"},
	{0xFA147003F0766F96ULL, "rdr111st_martin_nt.stg"},
	{0xFA147A69765A415AULL, "rdr111dg_martin6054551.cl"},
	{0xFA14DA574722648DULL, "60738780.ogg"},
	{0xFA15A4C85F014F60ULL, "vxlt_darktown_2vf_0_0.vlm"},
	{0xFA15A5B6A8E83A80ULL, "6019110_m.fxe"},
	{0xFA15B8D93F338EE1ULL, "follower_banter6147628.cl"},
	{0xFA1739A1E0DC6F71ULL, "mag360_samson6042034.cl"},
	{0xFA1744B2D7372AB7ULL, "gen00fl_anders6125856.cl"},
	{0xFA178EC19B7BCAFAULL, "129466581.ogg"},
	{0xFA18F30540AB32B5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_5_steps.ani"},
	{0xFA18FEE6095D3A88ULL, "llac_lt_0_lowtown_a2_n_-65534_0.dds"},
	{0xFA1906463E6EC64FULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_003a_0s.dds"},
	{0xFA191802EE129DC9ULL, "113183330.ogg"},
	{0xFA194E68A8899A02ULL, "lt_0_lowtown_docks_n_153_0.msh"},
	{0xFA195068A8899DB8ULL, "lt_0_lowtown_docks_n_153_0.mmh"},
	{0xFA195C68A889B25BULL, "lt_0_lowtown_docks_n_153_0.mao"},
	{0xFA1999558E75CF4AULL, "6038621_m.fxe"},
	{0xFA19ED1AD538BCD0ULL, "35142876.ogg"},
	{0xFA1AC90BBB3B95E3ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v7.gad"},
	{0xFA1AF16E47C7C924ULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_knockback.evt"},
	{0xFA1B11D024A1C0A8ULL, "vxht_0_hightown_n_c85_0_0.vlm"},
	{0xFA1B1442019F56CFULL, "gen00fl_merrill6108675.cl"},
	{0xFA1B9F61F1B56E03ULL, "gen_im_trs_chn_andraste_03.uti"},
	{0xFA1BC5FA9EEE98E6ULL, "6125457_m.fxe"},
	{0xFA1BC768D65CF43EULL, "hlac_ht_keep_n_3735_0x.dds"},
	{0xFA1DA9917D9E27E3ULL, "hlac_lt_0_lowtown_n_65538_0z.dds"},
	{0xFA1E073587599ABCULL, "vxlt_gallowstemplar_f_46w_0_0.vlm"},
	{0xFA1E2BF99F9ADF18ULL, "vxlt_gallowscourtyard_n_6yj_0_0.vlm"},
	{0xFA1E5BC87BF7A879ULL, "817801921.ogg"},
	{0xFA1E97C776FC74DFULL, "6100264_m.fxe"},
	{0xFA1EA94AC76A9C60ULL, "llac_lt_0_lowtown_n_65536_10.dds"},
	{0xFA200D140887C8E9ULL, "vxht_0_hightown_bzh_0_0.vlm"},
	{0xFA2070F1B0B36C63ULL, "hlac_w_sundermount_327683_0y.dds"},
	{0xFA20D1EDE0F67AF6ULL, "gen00fl_carver6103693.cl"},
	{0xFA21A60BBB4178C8ULL, "art\\characters\\playercharacter\\elffemale\\animation\\exploration\\idle\\twitch_v7.evt"},
	{0xFA21AB196BA54303ULL, "follower_banter6156101.cl"},
	{0xFA21BF966525A2E5ULL, "da2terrain_terrain_p5.psh.fxo.dx11"},
	{0xFA21E79143BEB0D6ULL, "w_bonepit_117_0.mao"},
	{0xFA21F39143BEC575ULL, "w_bonepit_117_0.mmh"},
	{0xFA21F59143BEC8EBULL, "w_bonepit_117_0.msh"},
	{0xFA222955928ED328ULL, "615763389.ogg"},
	{0xFA225E663BF1A021ULL, "245353999.ogg"},
	{0xFA227E4F65EEC6D7ULL, "hdistortion.dds"},
	{0xFA230FDD19644A2EULL, "99761723.ogg"},
	{0xFA235768A891FE43ULL, "lt_0_lowtown_docks_n_153_0.phy"},
	{0xFA236AFDA0E08391ULL, "one350_vinell6106368.cl"},
	{0xFA23A0787449D380ULL, "89590319.ogg"},
	{0xFA23E680A0E8BE8DULL, "hlac_w_encounter1_-65532_0x.dds"},
	{0xFA23EAC9890C22F6ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_boo_hvyo_0t.dds"},
	{0xFA2416A550F767E0ULL, "vxlt_darktown_2yg_0_0.vlm"},
	{0xFA244A95EF7E96A0ULL, "6142092_m.fxe"},
	{0xFA246480358A30E6ULL, "6120810_m.fxe"},
	{0xFA2482600A99B663ULL, "6070224_m.fxe"},
	{0xFA24B2B496186D90ULL, "art\\characters\\weapons\\longswords\\w_lsw_015a.phy"},
	{0xFA24FD92E8FA64BAULL, "vxht_0_hightown_a3_c3i_0_1.vlm"},
	{0xFA25788411087BC1ULL, "232370284.ogg"},
	{0xFA25A58B3D28B62DULL, "gen00fl_fenris6127090.cl"},
	{0xFA25B44C734344FAULL, "654450370.ogg"},
	{0xFA25ED7FB8583DF3ULL, "w_bonepitdestroyed_a3_134_0.phy"},
	{0xFA2681D9ECC5E61EULL, "hlac_lt_0_lowtown_a2_n_-2_5z.dds"},
	{0xFA271C9E667E0F32ULL, "426456034.ogg"},
	{0xFA2725BB34FBAC27ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cooker_dp.dds"},
	{0xFA2757E4FB8FD65CULL, "vxca_caverns2prc_9xb_0_0.vlm"},
	{0xFA278AA122DAFDBDULL, "6174052_m.fxe"},
	{0xFA279CF90540793FULL, "art\\levels\\ztools\\default\\generic\\1ww_basic.phy"},
	{0xFA28096E47D27C1DULL, "art\\characters\\creatures\\gibgolem\\animation\\body\\gibsgolem_knockback.ani"},
	{0xFA2869B7CE0A6F0FULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.dura.gad"},
	{0xFA29489833236CC2ULL, "6069285_m.fxe"},
	{0xFA299873B9BCD5F9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ledgedropland.ani"},
	{0xFA29A2D4F7DD3EEEULL, "zz_ave_debug6207686.cl"},
	{0xFA29CDFEEF4A47A9ULL, "vxca_caverns2_959_0_0.vlm"},
	{0xFA2A1ADADA9F1D31ULL, "w_dgr_dg06a_0t.dds"},
	{0xFA2A7A07D817CD49ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_saint_dp.dds"},
	{0xFA2AFF3FCB09F462ULL, "vxht_mansion_n_4vf_0_0.vlm"},
	{0xFA2B311A7738DA81ULL, "zz_weightshift_test.cut"},
	{0xFA2B46AE5625CCCCULL, "6088831_m.fxe"},
	{0xFA2BB86C3454C610ULL, "art\\characters\\playercharacter\\qunarimale\\hed_ars.mao"},
	{0xFA2BE5BD585595B5ULL, "913750329.ogg"},
	{0xFA2CDB098BCADA9BULL, "hlac_lt_0_lowtown_a2_n_c_1_8z.dds"},
	{0xFA2CF120AD8A333AULL, "6113135_m.fxe"},
	{0xFA2D04E53EB11732ULL, "vxdr_0_deeproads_cft_0_1.vlm"},
	{0xFA2DF1221E4B201AULL, "vxca_caverns1_iim_0_0.vlm"},
	{0xFA2E7C2EE69F9FAEULL, "vxw_bonepitdestroyed_a3_77_0_0.vlm"},
	{0xFA2E7EB7CE0EFEBCULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.dura.evt"},
	{0xFA2F0C2678C9A2D1ULL, "vxht_brothel_n_6d2_0_0.vlm"},
	{0xFA2F1ED02200C0D9ULL, "ht_mansion_n_2995.rml"},
	{0xFA307373B9C2B53BULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ledgedropland.gad"},
	{0xFA307BA6439D5AC4ULL, "art\\cinematics\\follower plots\\sebastian\\cs_letter_sebastianintro.gad"},
	{0xFA30EC5DCBC25B28ULL, "vxw_woundedcoast_ox_0_0.vlm"},
	{0xFA30F08648AEB882ULL, "vxca_cave_14j_0_0.vlm"},
	{0xFA310854808745BEULL, "qun200_party_comment6090453.cl"},
	{0xFA3121D6A7247DC5ULL, "barr_glow.mao"},
	{0xFA317410312A8405ULL, "vxw_sundermount_30w_0_0.vlm"},
	{0xFA31BAAC08E12FD9ULL, "vxht_keep_4i7_0_0.vlm"},
	{0xFA329EF2452DB37CULL, "542969452.ogg"},
	{0xFA32B5F51E277E0EULL, "qun210cr_guard6087762.cl"},
	{0xFA32BDBAA77B0AB0ULL, "vxlt_0_lowtown_a3_n_c_eff_0_0.vlm"},
	{0xFA33CE85ABAC5E32ULL, "vxlt_0_lowtown_docks_a3_n_9do_0_0.vlm"},
	{0xFA33D3D673B6C84DULL, "gen00fl_aveline6004158.cl"},
	{0xFA33F3D58D8F164FULL, "19006997.ogg"},
	{0xFA340884BAEB437AULL, "w_bonepit_213_0.mmh"},
	{0xFA340C84BAEB49B1ULL, "w_bonepit_213_0.mao"},
	{0xFA341A84BAEB61C0ULL, "w_bonepit_213_0.msh"},
	{0xFA343B5D2A5ACB7DULL, "one280_merin6116760.cl"},
	{0xFA34FA780E172EF4ULL, "6181909_m.fxe"},
	{0xFA35183E7C25E5BDULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_standb.phy"},
	{0xFA353542BC4FCEECULL, "vxlt_0_lowtown_docks_a3_n_brt_0_0.vlm"},
	{0xFA35A2D8A4224818ULL, "vxdr_0_deeproads_f5s_0_0.vlm"},
	{0xFA35E0206E93249EULL, "art\\levels\\4cinematics\\ferelden\\_textures\\prp_moneypoucha_d.dds"},
	{0xFA35E52A5DEDE420ULL, "ave200_donnic6060699.cl"},
	{0xFA36244A9DE7DC74ULL, "hlac_ca_caverns1_24422_5x.dds"},
	{0xFA362E0D74C39885ULL, "ch1notnt_s1.psh.fxo"},
	{0xFA366F35C5CDC160ULL, "llac_lt_0_lowtown_docks_a3_n_-3_8.dds"},
	{0xFA368AA643A1DFA2ULL, "art\\cinematics\\follower plots\\sebastian\\cs_letter_sebastianintro.ani"},
	{0xFA374D52AD4EACFEULL, "art\\characters\\playercharacter\\elfmale\\mt_cbt_orsi.phy"},
	{0xFA377CD890598555ULL, "270319588.ogg"},
	{0xFA379073B9C90520ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\ledgedropland.evt"},
	{0xFA37EA1C60B20958ULL, "6036508_m.fxe"},
	{0xFA381CEBC7B3E659ULL, "vxht_bartrand_n_5hm_0_0.vlm"},
	{0xFA3828FBB0CD9830ULL, "gen_im_ammo_quiver_05.uti"},
	{0xFA384ADC505A7BEEULL, "vxdr_0_deeproads_9in_0_0.vlm"},
	{0xFA389DF9054EEE44ULL, "art\\levels\\ztools\\default\\generic\\1ww_basic.mmh"},
	{0xFA389F2BEB7B7DF8ULL, "274202366.ogg"},
	{0xFA38A1F9054EF497ULL, "art\\levels\\ztools\\default\\generic\\1ww_basic.mao"},
	{0xFA38B3F9054F13E6ULL, "art\\levels\\ztools\\default\\generic\\1ww_basic.msh"},
	{0xFA38DB2BE6AB1F60ULL, "dae000_merchant_generic_hm6172641.cl"},
	{0xFA3905C44BC423BCULL, "vxlt_gallowscourtyard_6b0_0_0.vlm"},
	{0xFA390E7A74665FDBULL, "vxht_mansion2_6t_0_0.vlm"},
	{0xFA39501CBC04CE93ULL, "vxlt_0_lowtown_gfw_0_0.vlm"},
	{0xFA39B2DAB6DA6D0BULL, "hlac_dr_dungeon_3216_5x.dds"},
	{0xFA3A314C5C20C2EAULL, "vxlt_gallowstemplar_372_0_0.vlm"},
	{0xFA3A7758A7CB5065ULL, "104251769.ogg"},
	{0xFA3ACD521B4F5423ULL, "95021402.ogg"},
	{0xFA3B0A9F01650AF8ULL, "6150143_m.fxe"},
	{0xFA3B6677F4605E93ULL, "506347215.ogg"},
	{0xFA3BA62236DA151AULL, "zz_vault_debug6200371.cl"},
	{0xFA3BCA8C35E6D95AULL, "llac_lt_0_lowtown_a2_n_-196609_2.dds"},
	{0xFA3C0C6D01635B06ULL, "llac_lt_0_lowtown_n_-65537_3.dds"},
	{0xFA3C1A1E790CB504ULL, "310048295.ogg"},
	{0xFA3C86B7CE1B49D5ULL, "art\\vfx\\combat\\vfx_gui_ground_ring_l_p.dura.ani"},
	{0xFA3CA566F3F36264ULL, "179417449.ogg"},
	{0xFA3CC699FEE40E35ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinstower.phy"},
	{0xFA3CD53D364188D4ULL, "vxht_brothel_n_4d1_0_0.vlm"},
	{0xFA3D50FD5959A325ULL, "6113672_f.fxe"},
	{0xFA3D686314A5FC14ULL, "393604786.ogg"},
	{0xFA3D7E67454579AEULL, "556264914.ogg"},
	{0xFA3E4AC06F4E732BULL, "vxca_caverns1_mines_k2t_0_2.vlm"},
	{0xFA3E4EDBEB2A8215ULL, "sky_sky2_p2.psh.fxo"},
	{0xFA3E6CEC3E3A4BB4ULL, "6018618_m.fxe"},
	{0xFA3E796E95FC44FDULL, "6128844_m.fxe"},
	{0xFA3EA3714BAA5BE4ULL, "vxlt_gallowstemplar_n_5lx_0_0.vlm"},
	{0xFA3EE244B16E83F3ULL, "lt_darktown_mmi.gff"},
	{0xFA3F605BFCC469E8ULL, "313121243.ogg"},
	{0xFA3F8C5716296695ULL, "hlac_lt_0_lowtown_-65535_4y.dds"},
	{0xFA3F8CA6EBFF9E8DULL, "gen_im_trs_slv_hvy_boo_r1.uti"},
	{0xFA40183C54C8EE41ULL, "6036425_f.fxe"},
	{0xFA40280C9EF803B5ULL, "dae200_saemus6091699.cl"},
	{0xFA40B23939FBC4EEULL, "gen_im_trs_gen_med_cht_r1.uti"},
	{0xFA413D69325D3C16ULL, "narrative_varric.cnv"},
	{0xFA41F8694CFD35B1ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_countthrone.mmh"},
	{0xFA4204694CFD4A52ULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_countthrone.mao"},
	{0xFA4212694CFD626FULL, "art\\levels\\races\\ferelden\\props\\keep\\prp_countthrone.msh"},
	{0xFA42BD6AC0B6D528ULL, "w_bonepit_130_0.mmh"},
	{0xFA42C96AC0B6E9CBULL, "w_bonepit_130_0.mao"},
	{0xFA42CA7263390FBCULL, "63449794.ogg"},
	{0xFA42DB6AC0B70852ULL, "w_bonepit_130_0.msh"},
	{0xFA4414D6364B755DULL, "vxca_caverns1_gkc_0_1.vlm"},
	{0xFA4484E509144482ULL, "llac_ht_keep_3732_0.dds"},
	{0xFA44C95552AD10E1ULL, "718046256.ogg"},
	{0xFA451E54E73D4FF8ULL, "zz_one_debug6076183.cl"},
	{0xFA453F69326128F6ULL, "narrative_varric.bnk"},
	{0xFA454F1377C2E5CAULL, "follower_banter6148758.cl"},
	{0xFA458190D36DD3EEULL, "vxlt_0_lowtown_n_fhn_0_0.vlm"},
	{0xFA45B727ACB1D7BCULL, "llac_lt_gallowstemplar_n_2384_6.dds"},
	{0xFA45E821B1006C90ULL, "mag121_idunna6084232.cl"},
	{0xFA45FE810D8AA138ULL, "6096775_m.fxe"},
	{0xFA460D3E7C344629ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_standb.mao"},
	{0xFA46193E7C345AC2ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_standb.mmh"},
	{0xFA461B3E7C345E78ULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_standb.msh"},
	{0xFA467662759DD551ULL, "1039222547.ogg"},
	{0xFA46A9345B0E7843ULL, "vxlt_0_lowtown_a2_gnk_0_0.vlm"},
	{0xFA46D378F1A5D89DULL, "dae200cr_viscount6087393.cl"},
	{0xFA47DA5A856841BAULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zztent.mmh"},
	{0xFA47DE5A856848F1ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zztent.mao"},
	{0xFA47EC5A85686000ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zztent.msh"},
	{0xFA48D29B242DECE3ULL, "863096786.ogg"},
	{0xFA493A8C2C1FAA0FULL, "6149560_m.fxe"},
	{0xFA49F5E83AD3D154ULL, "follower_banter6141526.cl"},
	{0xFA4A21540F8BC545ULL, "eye_eye_p0.vsh.fxo.dx11"},
	{0xFA4AAE866AA26BA7ULL, "20333888.ogg"},
	{0xFA4ABB7614F73416ULL, "928024322.ogg"},
	{0xFA4AE9265328F136ULL, "228386647.ogg"},
	{0xFA4B58E840C196E8ULL, "753764516.ogg"},
	{0xFA4BEFE05CB5D3D1ULL, "vxht_brothel_n_6j0_0_0.vlm"},
	{0xFA4BF83043709133ULL, "vxlt_0_lowtown_exk_0_0.vlm"},
	{0xFA4C230E5CE867B3ULL, "vxht_mansion_n_3uu_0_0.vlm"},
	{0xFA4C2A876568668DULL, "389255357.ogg"},
	{0xFA4C4E118AB2085FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hlf_meda_0d.dds"},
	{0xFA4C575DD6485033ULL, "t1_mue_pi3.tnt"},
	{0xFA4CA3AFC8322DF8ULL, "vxlt_0_lowtown_docks_a3_d21_0_0.vlm"},
	{0xFA4CC5E4F6D1DBA8ULL, "vxlt_blackemporium_4j_0_1.vlm"},
	{0xFA4CEFF69DA658A3ULL, "1042128016.ogg"},
	{0xFA4CF53690223475ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_barge_dp.dds"},
	{0xFA4D91618CE1CA70ULL, "lgt210_magu_selby6142017.cl"},
	{0xFA4DBC0061ED69A3ULL, "vxlt_0_lowtown_docks_n_cto_0_0.vlm"},
	{0xFA4DBFDB9185E7D7ULL, "vxw_encounter2_ny_0_0.vlm"},
	{0xFA4DD017AFAE8C28ULL, "vxw_woundedcoast_av_0_0.vlm"},
	{0xFA4DF8B3ED9059A9ULL, "mag100_cullen6095115.cl"},
	{0xFA4E0B7BAC5BE86BULL, "vxht_0_hightown_a2_c8q_0_0.vlm"},
	{0xFA4E60DA879D29F6ULL, "vxca_caverns1_hm8_0_0.vlm"},
	{0xFA4E6CA317D4CC74ULL, "140165780.ogg"},
	{0xFA4E79ECEB969033ULL, "72829638.ogg"},
	{0xFA4E7E5B735A81E2ULL, "mer320_hubert6079336.cl"},
	{0xFA4EB04A2DD30615ULL, "art\\levels\\ztools\\default\\generic\\3oo_basic.phy"},
	{0xFA4ECC261FC84594ULL, "mcr310_argument6074320.cl"},
	{0xFA4F31A74F16E1EEULL, "502716322.ogg"},
	{0xFA4FA39235ADFBA4ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.cess.ani"},
	{0xFA4FEC1F35844B57ULL, "dae000_orana6149023.cl"},
	{0xFA501FAA6E54715BULL, "ps3umbradebug.vsh.fxo"},
	{0xFA50264FB1D555D0ULL, "mag320_orsino6071142.cl"},
	{0xFA50900CED287134ULL, "vxlt_0_lowtown_a3_n_c_gmq_0_0.vlm"},
	{0xFA50B4ACFFA41A3FULL, "vxht_0_hightown_a3_5wg_0_0.vlm"},
	{0xFA50CD128CFE703AULL, "33324539.ogg"},
	{0xFA512E7DEEF6E54AULL, "546183278.ogg"},
	{0xFA51CE1A0B1AF115ULL, "vxlt_0_lowtown_a3_n_c_bxv_0_0.vlm"},
	{0xFA51CE311E62931EULL, "vxca_cavern1_haunted_d60_0_0.vlm"},
	{0xFA520A38611B528CULL, "sst_ss_profanemob_1.gda"},
	{0xFA5246BA5FA70F6BULL, "gen00fl_merrill6127214.cl"},
	{0xFA5265241C51D85CULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_smelt.mao"},
	{0xFA5271241C51ECFFULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_smelt.mmh"},
	{0xFA5277241C51F6A1ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_smelt.msh"},
	{0xFA52EAC7BB49F46AULL, "mag371_elsa6041249.cl"},
	{0xFA5306B2E715D9F6ULL, "vxlt_0_lowtown_docks_bv2_0_0.vlm"},
	{0xFA533DE5CFB59815ULL, "zz_dre_debug6054624.cl"},
	{0xFA534305248A5902ULL, "zz_mag_debug6044312.cl"},
	{0xFA537018A1170B49ULL, "hlac_lt_0_lowtown_docks_n_-131074_0y.dds"},
	{0xFA53C285A5683A0FULL, "vxht_chantry_n_2ho_0_0.vlm"},
	{0xFA53DDA480EC8508ULL, "536315588.ogg"},
	{0xFA53DE838811ABB7ULL, "vxca_cavern1_haunted_e8b_0_0.vlm"},
	{0xFA54007DADB1E3FFULL, "22205259.ogg"},
	{0xFA545D18CE38F398ULL, "dre100cr_sandal.utc"},
	{0xFA547DDB0A9B6CB4ULL, "902004902.ogg"},
	{0xFA547F026199B81FULL, "lgt110_gxa_coa_spy6117407.cl"},
	{0xFA55254EDE02CB62ULL, "isa311_party_comments6117818.cl"},
	{0xFA554A5566CA4969ULL, "6139154_m.fxe"},
	{0xFA55771389548269ULL, "dre000_vampire6034327.cl"},
	{0xFA55873F80CB2825ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_003a_d.dds"},
	{0xFA55A09235B26206ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.cess.gad"},
	{0xFA55CC8A3913CA07ULL, "follower_banter6163475.cl"},
	{0xFA561A316CCDAA09ULL, "6106985_m.fxe"},
	{0xFA5630093EF56B3FULL, "6006254_m.fxe"},
	{0xFA5633D69FF0FF93ULL, "vxlt_0_lowtown_ghy_0_0.vlm"},
	{0xFA5691104AC92C4CULL, "vxdr_slavers_14e_0_0.vlm"},
	{0xFA57051FB2F0BCD4ULL, "6112133_f.fxe"},
	{0xFA57B5764F8685CAULL, "103729835.ogg"},
	{0xFA57C2F20368FEB6ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02a.mmh"},
	{0xFA57CCF203690FB4ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_chk02a.msh"},
	{0xFA582AB676D3E324ULL, "gen00fl_bethany6102768.cl"},
	{0xFA58989544946620ULL, "vxw_bonepitdestroyed_a3_28l_0_0.vlm"},
	{0xFA59101434A2BBA0ULL, "6173868_f.fxe"},
	{0xFA596A730CB8EF97ULL, "6126354_m.fxe"},
	{0xFA5ACE83F09AEF9BULL, "zz_monsters.cut"},
	{0xFA5C21486FB22DABULL, "llac_ht_mansion_n_3373_3.dds"},
	{0xFA5C2DDD53E38223ULL, "vxlt_0_lowtown_docks_n_con_0_0.vlm"},
	{0xFA5C2EEB71E8681DULL, "327290109.ogg"},
	{0xFA5C31B8837C0057ULL, "vxw_encounter2_ix_0_0.vlm"},
	{0xFA5C945C7517B850ULL, "vxdr_primevaltomb_2gg_0_0.vlm"},
	{0xFA5C9B9235B878E5ULL, "art\\vfx\\combat\\vfx_r_arc_pin_shot_palyz_c.cess.evt"},
	{0xFA5CDA1B719BBCD3ULL, "1044995069.ogg"},
	{0xFA5D04CEA00634F0ULL, "dae000_gen_turn_in_36178452.cl"},
	{0xFA5E13517D5D5430ULL, "vxlt_0_lowtown_a3_n_c_ewe_0_0.vlm"},
	{0xFA5EF3D243F37C07ULL, "6042837_f.fxe"},
	{0xFA5FA9AAFE60AB22ULL, "vxca_caverns1_mines_ihw_0_0.vlm"},
	{0xFA5FD3691EF990A2ULL, "6153997_m.fxe"},
	{0xFA5FF6526C6672EEULL, "zz_vrc_debug6200169.cl"},
	{0xFA606C3F4641F221ULL, "dre000_sandal6031378.cl"},
	{0xFA609F5292CB3CAFULL, "vxht_mansion_n_6on_0_0.vlm"},
	{0xFA60A9DF9E7A1111ULL, "lit_aoe_01.mao"},
	{0xFA60ED50B6EA7A64ULL, "vxlt_0_lowtown_docks_n_bf3_0_0.vlm"},
	{0xFA6100B71385D6B3ULL, "tactics.gfx"},
	{0xFA612F75CC49F84CULL, "6016589_f.fxe"},
	{0xFA6141EAC4299F7EULL, "follower_banter6143210.cl"},
	{0xFA615E755E276ED7ULL, "t3_wdg_gw3.tnt"},
	{0xFA617F53F5D14B97ULL, "gen_im_npc_wep_one_swdb.uti"},
	{0xFA61857E8460B01EULL, "561089667.ogg"},
	{0xFA6193DBAB44BFDAULL, "hlac_w_woundedcoast_-65535_0z.dds"},
	{0xFA623462975E528FULL, "vxht_chantry_n_2cn_0_0.vlm"},
	{0xFA624C824455656BULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v1.evt"},
	{0xFA6266E0F8B50D05ULL, "vxlt_undercityencounter_122_0_0.vlm"},
	{0xFA627CA3A8AE4763ULL, "6163274_m.fxe"},
	{0xFA628CBF2DB25BE9ULL, "hlac_lt_0_lowtown_docks_-3_1z.dds"},
	{0xFA631A3CE6889957ULL, "vxlt_undercitydungeon_1m8_0_0.vlm"},
	{0xFA6328328E56505DULL, "vxw_encounter2_kt_0_0.vlm"},
	{0xFA634AF39961C388ULL, "llac_w_encounter2_camp_-1_0.dds"},
	{0xFA6351CD0E1525A6ULL, "vxht_0_hightown_a3_a6i_0_0.vlm"},
	{0xFA63860D273AD5AAULL, "961811877.ogg"},
	{0xFA63AFEDF2F68573ULL, "lgt200cr_gang_2_lt.utc"},
	{0xFA644AF16AAAAB14ULL, "6048206_m.fxe"},
	{0xFA64727DAF833FD1ULL, "vxdr_primevaltomb_6pu_0_2.vlm"},
	{0xFA64D32DC55B948FULL, "6015830_m.fxe"},
	{0xFA6570C1A59E5C4CULL, "vxht_0_hightown_n_ani_0_0.vlm"},
	{0xFA65CB60AD3877ABULL, "qun120cr_javaris6042833.cl"},
	{0xFA65D93B76B5CE47ULL, "hlac_ca_caverns2_5802_43y.dds"},
	{0xFA65F75714AB352DULL, "follower_banter6173142.cl"},
	{0xFA6609C61DBE947EULL, "vxlt_0_lowtown_docks_a3_n_d20_0_0.vlm"},
	{0xFA675748A8F73002ULL, "vxlt_undercitydungeon_18z_0_0.vlm"},
	{0xFA6865824459FB2CULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v1.gad"},
	{0xFA68CDF521184BF1ULL, "vxdr_primevaltomb_91c_0_0.vlm"},
	{0xFA69555870D42B17ULL, "gen00fl_aveline6139793.cl"},
	{0xFA6A1A28EB198E44ULL, "6083988_m.fxe"},
	{0xFA6ABDD3EDCE4115ULL, "447253383.ogg"},
	{0xFA6AD010EE1E9147ULL, "459932201.ogg"},
	{0xFA6B56C1745BE4CFULL, "266255799.ogg"},
	{0xFA6B920AF489B3EFULL, "hlac_ca_cave_2194_8x.dds"},
	{0xFA6BAF923092FA0CULL, "vxdr_primevalentrance_23r_0_1.vlm"},
	{0xFA6C2D4E307B1EBFULL, "6162116_m.fxe"},
	{0xFA6C59651222760CULL, "vxca_caverns2_5a9_0_1.vlm"},
	{0xFA6C64A2BD1CA694ULL, "189481472.ogg"},
	{0xFA6CC3E442EC1DC8ULL, "dae011_corff_sta6140575.cl"},
	{0xFA6CF95A858792C5ULL, "art\\levels\\races\\proxy\\props\\camp\\prp_zztent.phy"},
	{0xFA6D45111376D505ULL, "dae320ip_at_player_home_nt.utp"},
	{0xFA6DBB49DA587B29ULL, "875613836.ogg"},
	{0xFA6DCE534629B668ULL, "699591290.ogg"},
	{0xFA6DF125D6C7FD28ULL, "6076478_m.fxe"},
	{0xFA6DFD372E7FE140ULL, "gen00fl_isabela6097905.cl"},
	{0xFA6E09DB85711F75ULL, "lt_0_lowtown_a3_n_c_30_0.rml"},
	{0xFA6EA948C4BAA023ULL, "315492247.ogg"},
	{0xFA6EC129FC6EF4D0ULL, "vxht_playershouse_n_1lf_0_0.vlm"},
	{0xFA6EECE959EAF802ULL, "one100dg_vanard6059704.cl"},
	{0xFA6EFCB6842830ECULL, "vxlt_0_lowtown_docks_a3_n_bfh_0_0.vlm"},
	{0xFA6F16CF5D40E91AULL, "vxlt_gallowscourtyard_62v_0_0.vlm"},
	{0xFA6F5F2DA8E092E4ULL, "vxlt_0_lowtown_docks_n_bc0_0_0.vlm"},
	{0xFA6F64824460182AULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_comm_v1.ani"},
	{0xFA6F684EFDC84812ULL, "lt_0_lowtown_a3_n_c_probe_gpy.dds"},
	{0xFA6F96BF2C171149ULL, "hlac_w_woundedcoast_196611_0x.dds"},
	{0xFA7048B71ED6A836ULL, "llac_w_blightlands_-65539_0.dds"},
	{0xFA7090593ACBFA1BULL, "hlac_lt_darktown_-1_70x.dds"},
	{0xFA710FF024CE9A5DULL, "vxht_0_hightown_a2_al9_0_0.vlm"},
	{0xFA714035755BD93DULL, "594927568.ogg"},
	{0xFA71419AC63BDB1DULL, "115858735.ogg"},
	{0xFA718283701B99F1ULL, "6109412_m.fxe"},
	{0xFA71CA8E979048F3ULL, "496109082.ogg"},
	{0xFA71EE9A71566FFCULL, "279943718.ogg"},
	{0xFA72111FD305DA13ULL, "594303948.ogg"},
	{0xFA7291BE8F73AA40ULL, "hm_drunk1.mor"},
	{0xFA7291C28DEAC41FULL, "491486380.ogg"},
	{0xFA72A76440D07635ULL, "vxlt_0_lowtown_docks_n_cw1_0_1.vlm"},
	{0xFA73E1570AA39E6AULL, "76412606.ogg"},
	{0xFA740A58C424C122ULL, "6110387_m.fxe"},
	{0xFA7459B54D462C3FULL, "hlac_lt_0_lowtown_docks_a3_n_-131073_3x.dds"},
	{0xFA7467ED43F571F1ULL, "548120250.ogg"},
	{0xFA74D4FD48CAD919ULL, "139656522.ogg"},
	{0xFA752DF01429D41BULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba04a.phy"},
	{0xFA753B9A7E7CFB9DULL, "zz_dre_debug6205398.cl"},
	{0xFA767E0DBB3D09A1ULL, "vxw_encounter2_camp_qn_0_0.vlm"},
	{0xFA769CB0FE19EAD4ULL, "vxht_brothel_n_4x5_0_0.vlm"},
	{0xFA77558076C2F56FULL, "ht_chantry_2534_ll.mmh"},
	{0xFA7764F202A726FBULL, "454762768.ogg"},
	{0xFA77DA237BADAB23ULL, "680509643.ogg"},
	{0xFA784D51BC09DDB6ULL, "one101_orders_followers6115896.cl"},
	{0xFA787BED5CD64209ULL, "6164574_m.fxe"},
	{0xFA787C2FD6E65A9BULL, "6127533_m.fxe"},
	{0xFA7893EF26786784ULL, "hightown_male_noble_gossip6174092.cl"},
	{0xFA78D94EE24089CEULL, "6067687_m.fxe"},
	{0xFA7925AE8590D5A8ULL, "vxw_woundedcoast_pu_0_0.vlm"},
	{0xFA793C5F16F91CA7ULL, "hlac_w_bonepit_393216_0y.dds"},
	{0xFA7960791CFC4CE2ULL, "gen00flst_beth_carv_final.stg"},
	{0xFA7979B5F51878D3ULL, "6109115_m.fxe"},
	{0xFA798D9CE1196059ULL, "hlac_lt_0_lowtown_a2_n_c_-262144_4y.dds"},
	{0xFA79986C069FC81EULL, "vxdr_primevaltomb_7o4_0_0.vlm"},
	{0xFA7B1DF289742C66ULL, "6166867_m.fxe"},
	{0xFA7B3231B12C6CD5ULL, "dae000cr_grand_cleric6181903.cl"},
	{0xFA7B6AC27B16FFBAULL, "691198332.ogg"},
	{0xFA7B849F74E1A730ULL, "vxht_keep_n_3x1_0_0.vlm"},
	{0xFA7C03F045DA1384ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgta.msh"},
	{0xFA7C15F045DA32EDULL, "art\\characters\\playercharacter\\elffemale\\arm_lgta.mao"},
	{0xFA7C19F045DA3926ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgta.mmh"},
	{0xFA7C44F09E8C72DFULL, "vxw_encounter1_hb_0_0.vlm"},
	{0xFA7C9559A40655E5ULL, "hlac_lt_0_lowtown_a2_n_1_5y.dds"},
	{0xFA7CB3F33E78C4A5ULL, "166258072.ogg"},
	{0xFA7D0B335CCF99EAULL, "784237207.ogg"},
	{0xFA7D1349823EFBDEULL, "802733652.ogg"},
	{0xFA7D3183EBB3934CULL, "6106193_m.fxe"},
	{0xFA7D3B4743B07B2DULL, "art\\characters\\heraldry\\hld_lord_harimann.dds"},
	{0xFA7D8B7B40EEC6D4ULL, "w_bonepit_173_0.phy"},
	{0xFA7DA7200F5CD62AULL, "sst_ss_coterie_thug_f.gda"},
	{0xFA7E16D258C61A73ULL, "hla_ht_0_hightown_-131071_0y.dds"},
	{0xFA7E2E3396FC2202ULL, "art\\characters\\creatures\\desiredemon\\textures\\cn_bdy_suca_0e.dds"},
	{0xFA7E42F18727F293ULL, "6037401_m.fxe"},
	{0xFA7E892F4C6D71ADULL, "isa211_isabela6071923.cl"},
	{0xFA7E9A5432946071ULL, "follower_barks6153345.cl"},
	{0xFA7F214636624014ULL, "hlac_lt_darktown_-1_60z.dds"},
	{0xFA7FA752A5C3381AULL, "lt_0_lowtown_docks_n_177_0.mmh"},
	{0xFA7FAB52A5C33F51ULL, "lt_0_lowtown_docks_n_177_0.mao"},
	{0xFA7FB8F8A00AF938ULL, "vxlt_0_lowtown_docks_a3_n_d0l_0_0.vlm"},
	{0xFA7FB952A5C35760ULL, "lt_0_lowtown_docks_n_177_0.msh"},
	{0xFA7FBBF18420EB8DULL, "vxlt_darktown_6kl_0_0.vlm"},
	{0xFA7FEA0753214620ULL, "tablewood_04.mao"},
	{0xFA800BEC7242815DULL, "vxw_encounter2_qr_0_0.vlm"},
	{0xFA804C895CC317B3ULL, "952161360.ogg"},
	{0xFA8131A47A71EC78ULL, "hlac_lt_undercityencounter_-1_20x.dds"},
	{0xFA817FC81C8182BCULL, "257302131.ogg"},
	{0xFA81E4EFC98760D9ULL, "6043319_dragon_reveal.cut"},
	{0xFA825BE43EF60A8DULL, "vxca_cavern1_haunted_gie_0_0.vlm"},
	{0xFA82B3F047343BCCULL, "642749482.ogg"},
	{0xFA830BE9CB79FA43ULL, "602837302.ogg"},
	{0xFA8409DCFD691356ULL, "6135806_f.fxe"},
	{0xFA840C02DDA1D6F7ULL, "829490764.ogg"},
	{0xFA841DD110456FFDULL, "242326527.ogg"},
	{0xFA845DB069E13D81ULL, "mer341_earl6054241.cl"},
	{0xFA849C801CAD7A5DULL, "vxlt_0_lowtown_docks_a3_d0g_0_1.vlm"},
	{0xFA84A2BA96988A9FULL, "403623939.ogg"},
	{0xFA84BEC876F1F4E6ULL, "vxca_caverns2_9g0_0_0.vlm"},
	{0xFA84CDAB9C363EBBULL, "llac_dr_dungeon_3217_4.dds"},
	{0xFA84D999FF21E450ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinstower.msh"},
	{0xFA84E799FF21FC6AULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinstower.mmh"},
	{0xFA84EB99FF2202A1ULL, "art\\levels\\races\\natural\\kirkcoastal\\props\\prp_03ruinstower.mao"},
	{0xFA84EFEAAD332221ULL, "vxw_encounter2_camp_lm_0_0.vlm"},
	{0xFA85445332B35D37ULL, "746802299.ogg"},
	{0xFA85684F3F022ABCULL, "hlac_ht_0_hightown_a2_1_0x.dds"},
	{0xFA859AD87625B76EULL, "6034055_f.fxe"},
	{0xFA85D2955AD2DA31ULL, "1026509.ogg"},
	{0xFA86054BF323C362ULL, "1030163688.ogg"},
	{0xFA860E06A2CDBB7BULL, "art\\levels\\races\\dwarf\\props\\paragon\\prp_stand_1.mao"},
	{0xFA8610F045E27EF1ULL, "art\\characters\\playercharacter\\elffemale\\arm_lgta.phy"},
	{0xFA863F3E3AB0A49FULL, "zz_fol_debug6096541.cl"},
	{0xFA865782D4B8D8FEULL, "508789193.ogg"},
	{0xFA86C567AF8EEBC3ULL, "756055739.ogg"},
	{0xFA871618A9B42300ULL, "vxht_brothel_n_5ka_0_0.vlm"},
	{0xFA8783F37C1BC471ULL, "vxw_encounter2_camp_30_0_0.vlm"},
	{0xFA879A8E406EEDA5ULL, "44155349.ogg"},
	{0xFA87A8BDA84E6577ULL, "6065280_m.fxe"},
	{0xFA87E67FB8AAFC68ULL, "w_bonepitdestroyed_a3_134_0.mmh"},
	{0xFA87E767ACB12A33ULL, "hlac_w_encounter3_0_12y.dds"},
	{0xFA87ECCD440638B0ULL, "ht_0_hightown_a2_62_0.rml"},
	{0xFA87F27FB8AB100BULL, "w_bonepitdestroyed_a3_134_0.mao"},
	{0xFA88047FB8AB2E92ULL, "w_bonepitdestroyed_a3_134_0.msh"},
	{0xFA884E00C1F504A8ULL, "835644121.ogg"},
	{0xFA885C2B344E528AULL, "follower_banter6155877.cl"},
	{0xFA88687B40F892D1ULL, "w_bonepit_173_0.msh"},
	{0xFA88767B40F8AAECULL, "w_bonepit_173_0.mao"},
	{0xFA88827B40F8BE8FULL, "w_bonepit_173_0.mmh"},
	{0xFA88D0463B598AD2ULL, "vxca_caverns2prc_8zc_0_0.vlm"},
	{0xFA8A40B4A0E68D10ULL, "294671050.ogg"},
	{0xFA8B1B9D36B6E520ULL, "vxlt_0_lowtown_f10_0_0.vlm"},
	{0xFA8CD61D698EA1DEULL, "vxht_bartrand_n_63h_0_0.vlm"},
	{0xFA8D0DE467762CF5ULL, "6105926_m.fxe"},
	{0xFA8D7BC87194A548ULL, "6181924_f.fxe"},
	{0xFA8DF8E5F6400C49ULL, "ico_shield_1.dds"},
	{0xFA8E3CC2E7A54522ULL, "62973236.ogg"},
	{0xFA8E6EB11D507C18ULL, "ave200_donnic6111986.cl"},
	{0xFA8EDB7CD432E161ULL, "vxht_0_hightown_c4s_0_0.vlm"},
	{0xFA8EFF9008500DAAULL, "llac_lt_foundry_n_320_18.dds"},
	{0xFA8F3913D57CB3D7ULL, "vxw_sundermount_2zk_0_1.vlm"},
	{0xFA8F585D7C46C237ULL, "vxlt_undercitydungeon_1a4_0_0.vlm"},
	{0xFA8FEF68F765BDFCULL, "hlac_ca_caverns2_5802_68y.dds"},
	{0xFA9050CAFCC3CF7DULL, "6047729_m.fxe"},
	{0xFA90BC07C86FD53CULL, "6128367_f.fxe"},
	{0xFA9109E62AF27E83ULL, "vxlt_gallowstemplar_f_3k4_0_0.vlm"},
	{0xFA91124B08438C73ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ff_stala2a_1.mao"},
	{0xFA911257F2A2DACDULL, "gen00fl_anders6113059.cl"},
	{0xFA91333ACB19EAF0ULL, "6136666_m.fxe"},
	{0xFA9159F29AF6FF78ULL, "gen00fl_merrill6128755.cl"},
	{0xFA917F3F2591C651ULL, "6171529_m.fxe"},
	{0xFA91B5ADE08583F6ULL, "365082826.ogg"},
	{0xFA91DEF800341BD6ULL, "vxht_0_hightown_n_c3k_0_1.vlm"},
	{0xFA91FB0B8123BEF6ULL, "mcr370_negotiation.cnv"},
	{0xFA923E6A1A30B644ULL, "vxlt_gallowscourtyard_6xj_0_0.vlm"},
	{0xFA92A8B157F032B4ULL, "ter_bigrock01_s.dds"},
	{0xFA9361C79F293AA1ULL, "vxw_encounter2_camp_gl_0_0.vlm"},
	{0xFA9388030FDD4725ULL, "vxht_keep_5jc_0_0.vlm"},
	{0xFA93B12D5B7BE02AULL, "llac_lt_0_lowtown_a2_196607_0.dds"},
	{0xFA94351CAAC186B3ULL, "6127036_m.fxe"},
	{0xFA94E4519C13F993ULL, "vxht_chantry_n_35u_0_0.vlm"},
	{0xFA94F4D51C161749ULL, "hightown_male_noble_gossip6171337.cl"},
	{0xFA952141512A2DE2ULL, "gen00fl_anders6111074.cl"},
	{0xFA9532879CA83F8BULL, "6133877_m.fxe"},
	{0xFA95B9BD6BDB52CAULL, "llac_lt_undercityencounter_-1_24.dds"},
	{0xFA95C34A2E0F2930ULL, "art\\levels\\ztools\\default\\generic\\3oo_basic.msh"},
	{0xFA95D14A2E0F414AULL, "art\\levels\\ztools\\default\\generic\\3oo_basic.mmh"},
	{0xFA95D54A2E0F4781ULL, "art\\levels\\ztools\\default\\generic\\3oo_basic.mao"},
	{0xFA95FD0B8127AAD6ULL, "mcr370_negotiation.bnk"},
	{0xFA965BD03CF01A50ULL, "vxdr_primevaltomb_2sc_0_0.vlm"},
	{0xFA967C25EA8BF91EULL, "vxdr_primevaltomb_7e6_0_0.vlm"},
	{0xFA96A34BA22E1038ULL, "and100_anders6095921.cl"},
	{0xFA977AE9C842DB6BULL, "hlac_lt_0_lowtown_n_131071_7y.dds"},
	{0xFA97853E88F68F25ULL, "6119601_f.fxe"},
	{0xFA981FFE8FE0B345ULL, "6070576_m.fxe"},
	{0xFA98643295878D48ULL, "vxht_0_hightown_c4v_0_0.vlm"},
	{0xFA9887131695491FULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.cess.gad"},
	{0xFA988E6F73670C89ULL, "6141575_m.fxe"},
	{0xFA98C06606E57930ULL, "6035295_m.fxe"},
	{0xFA98C5FE613B6611ULL, "character_character_p79.psh.fxo.dx11"},
	{0xFA98D36D29282094ULL, "vxlt_darktown_58s_0_0.vlm"},
	{0xFA99200529C3196DULL, "mag100_karras6035288.cl"},
	{0xFA993759229CCE85ULL, "6096518_m.fxe"},
	{0xFA99B45A889C533FULL, "vxlt_0_lowtown_docks_a3_n_bfg_0_0.vlm"},
	{0xFA9B3132F3B62F34ULL, "586178519.ogg"},
	{0xFA9BEC6B8F10C954ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_loop.ani"},
	{0xFA9C713B22BBE898ULL, "22059387.ogg"},
	{0xFA9CF93A9E6C7048ULL, "hlac_ca_caverns1_24423_8x.dds"},
	{0xFA9D2F4578DBDE76ULL, "hlac_ca_caverns1_24424_2y.dds"},
	{0xFA9D3FF6DEC3EB52ULL, "929860257.ogg"},
	{0xFA9D7A0A7C1C96CFULL, "w_woundedcoast_24_0_lc.mmh"},
	{0xFA9D8FD185B2C1E3ULL, "vxca_caverns1_imd_0_0.vlm"},
	{0xFA9DABF042F1E789ULL, "l_one130ar_disciple_dungeon.bnk"},
	{0xFA9DB4C939E82DA4ULL, "bee_ball.dds"},
	{0xFA9DCDBA7676CD20ULL, "vrc321_varric6087091.cl"},
	{0xFA9E582770D22235ULL, "vxlt_gallowstemplar_f_45w_0_0.vlm"},
	{0xFA9E5BD19FF34AD7ULL, "llac_lt_0_lowtown_a2_-65537_5.dds"},
	{0xFA9EBFE6BB27C5F1ULL, "vxca_caverns1_j77_0_0.vlm"},
	{0xFA9EC53F50C10F98ULL, "isa000im_gft_talisman.uti"},
	{0xFA9EDDC360A65537ULL, "gen00fl_aveline6115276.cl"},
	{0xFA9F38356D76BE4CULL, "vxht_0_hightown_n_arm_0_0.vlm"},
	{0xFA9F69D54F1787C8ULL, "isa211_isabela6072569.cl"},
	{0xFA9F7159E4194F09ULL, "vxlt_house_dm_0_0.vlm"},
	{0xFA9F7C13169B550CULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.cess.evt"},
	{0xFA9F84D32DEFECE8ULL, "6093627_f.fxe"},
	{0xFA9FC4F858D94083ULL, "dae011_corff_sta6163832.cl"},
	{0xFAA07245381541F2ULL, "mag311_keran6041690.cl"},
	{0xFAA0A1D6270C0C80ULL, "vxlt_0_lowtown_docks_b6t_0_0.vlm"},
	{0xFAA107107952B440ULL, "vxlt_0_lowtown_a3_n_c_gj0_0_0.vlm"},
	{0xFAA13AE0AA194D59ULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_ian.ani"},
	{0xFAA14FD2A46DCE66ULL, "6122746_m.fxe"},
	{0xFAA19C01ABE834A9ULL, "26398933.ogg"},
	{0xFAA1D3A4911F5321ULL, "vxw_encounter2_camp_bc_0_0.vlm"},
	{0xFAA214F0145097BAULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba04a.msh"},
	{0xFAA218FCD81DB8E9ULL, "scaleform.mat.dx11"},
	{0xFAA21ACABA8887D6ULL, "446890799.ogg"},
	{0xFAA21F4B0A23C9C5ULL, "vxdr_primevaltomb_4hp_0_1.vlm"},
	{0xFAA226F01450B600ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_brd_ba04a.mmh"},
	{0xFAA291F0BA29CA78ULL, "aip_dalisharcher_base.gda"},
	{0xFAA2D63A3C8C965FULL, "dae300_bran6143086.cl"},
	{0xFAA2E96B8F16E2B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_loop.gad"},
	{0xFAA301D84CC3837AULL, "genip_cin_seb_paper.utp"},
	{0xFAA307DBB4C61ABBULL, "vxht_0_hightown_a3_bzc_0_0.vlm"},
	{0xFAA3562E8E0A1213ULL, "vxht_chantry_n_32t_0_0.vlm"},
	{0xFAA45A6137B11844ULL, "t3_wpn_rst.tnt"},
	{0xFAA471D8096AAF75ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallobigbird.msh"},
	{0xFAA473D8096AB2EBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallobigbird.mmh"},
	{0xFAA47FD8096AC688ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\9ww_gallobigbird.mao"},
	{0xFAA4853269F9F20BULL, "vxdr_primevaltomb_5bh_0_0.vlm"},
	{0xFAA4C334DAE2EEC6ULL, "fx_dust05.dds"},
	{0xFAA4DD1A1FCAA5FAULL, "character_character_p61.psh.fxo.dx11.assembly"},
	{0xFAA541ADED0A7822ULL, "w_bonepit_228_0.msh"},
	{0xFAA543ADED0A7BD8ULL, "w_bonepit_228_0.mmh"},
	{0xFAA54FADED0A907BULL, "w_bonepit_228_0.mao"},
	{0xFAA559E7B7440B8EULL, "plt_mag320pt_otl_evelina.plo"},
	{0xFAA5A6D5864597EAULL, "6120919_m.fxe"},
	{0xFAA667865C935C6CULL, "97861573.ogg"},
	{0xFAA6D3A9B98E7750ULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_act_c.mmh"},
	{0xFAA78D562B522D29ULL, "vxlt_0_lowtown_fhr_0_0.vlm"},
	{0xFAA7A7FEF9B69259ULL, "art\\vfx\\textures\\level_fx\\vfx_foam01.dds"},
	{0xFAA815E0AA1F2C9BULL, "art\\characters\\creatures\\crowds\\humanmale\\animation\\kirkwall_and_lowtown\\humanmale_ambient_beat_threaten_1_ian.gad"},
	{0xFAA8889DC44756D0ULL, "vxht_playershouse_n_1pb_0_0.vlm"},
	{0xFAA9046B8F1B7C15ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\bless_loop.evt"},
	{0xFAA94B2670235674ULL, "303073161.ogg"},
	{0xFAA9B9D74D7AD2DEULL, "vxht_bartrand_n_69j_0_0.vlm"},
	{0xFAAA217AB75108A9ULL, "vxlt_gallowstemplar_n_3r1_0_0.vlm"},
	{0xFAAA3A778F9036C7ULL, "268938829.ogg"},
	{0xFAAA5B88E6AD5E6DULL, "art\\levels\\races\\elf\\props\\_textures\\prp_hangingc_sa.dds"},
	{0xFAAB114CADFCDC1EULL, "hlac_lt_undercityencounter_-1_21y.dds"},
	{0xFAAB20C25E3F99C5ULL, "art\\characters\\weapons\\boulder\\textures\\w_boulder_sa.dds"},
	{0xFAABB76B0F4D2808ULL, "vxht_keep_36x_0_0.vlm"},
	{0xFAAC3C176032F337ULL, "vxlt_undercitydungeon_1o6_0_0.vlm"},
	{0xFAAC557BAD3A3E82ULL, "6156734_m.fxe"},
	{0xFAAC5EB37376C118ULL, "dae011_corff_sta6149644.cl"},
	{0xFAACDE609830901DULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_mede_0s.dds"},
	{0xFAACE8221D13654FULL, "6174582_m.fxe"},
	{0xFAAD21B06D91C5A0ULL, "vxdr_primevaltomb_90m_0_0.vlm"},
	{0xFAAD391D418D39C7ULL, "isa211_isabela6202059.cl"},
	{0xFAAD59310961E96DULL, "333070928.ogg"},
	{0xFAAD641316A76A65ULL, "art\\vfx\\cinematic\\vfx_dragon_foot_dust_c.cess.ani"},
	{0xFAAD6B2207F8B06FULL, "62765881.ogg"},
	{0xFAAE118AB0FABEAFULL, "hlac_lt_0_lowtown_n_131073_2x.dds"},
	{0xFAAE197B7B5A3C6BULL, "133675352.ogg"},
	{0xFAAE566D84F2A23AULL, "6081683_m.fxe"},
	{0xFAAEAA763F7B442CULL, "154358752.ogg"},
	{0xFAAF44D52CD36B0EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_right.evt"},
	{0xFAAF7E9696673569ULL, "6133545_m.fxe"},
	{0xFAAFDA0F4D44BD7FULL, "vxca_cave_17r_0_0.vlm"},
	{0xFAB017DA5A882B57ULL, "vxht_brothel_n_6ek_0_0.vlm"},
	{0xFAB06F4562DFF692ULL, "vxlt_backalley_g_1l3_0_0.vlm"},
	{0xFAB080B40477C0ADULL, "hlac_ht_0_hightown_-65537_1z.dds"},
	{0xFAB0E3A3E298925EULL, "vxht_mansion2b_2b3_0_0.vlm"},
	{0xFAB1B147FC320776ULL, "55686121.ogg"},
	{0xFAB1BEAE33F56302ULL, "zz_followers_banter6159291.cl"},
	{0xFAB1DA8FA34EBD01ULL, "882794678.ogg"},
	{0xFAB1FBCA0D961E85ULL, "lgt360_mag_orlanna6155312.cl"},
	{0xFAB20DC3DE448091ULL, "w_blightlands_174_0.mmh"},
	{0xFAB219C3DE449532ULL, "w_blightlands_174_0.mao"},
	{0xFAB227C3DE44AD4FULL, "w_blightlands_174_0.msh"},
	{0xFAB24D5E547FB3C3ULL, "110231645.ogg"},
	{0xFAB258FD8125AE44ULL, "distor_c1_alp.dds"},
	{0xFAB2F4321C2F47ABULL, "6106914_m.fxe"},
	{0xFAB2FBC6FC335986ULL, "dae000_mother6103037.cl"},
	{0xFAB2FEB832063182ULL, "6100131_m.fxe"},
	{0xFAB31C2EC48E188BULL, "6085682_m.fxe"},
	{0xFAB3F6A9B9993D6FULL, "art\\vfx\\combat\\vfx_r_gn4_stealth_act_c.anb"},
	{0xFAB3FD9006529CBDULL, "vxlt_hangedman_n_1yo_0_0.vlm"},
	{0xFAB490D3172D248CULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grass.mmh"},
	{0xFAB49326BAFF208CULL, "llac_ca_cavern1_haunted_24425_3.dds"},
	{0xFAB496D3172D2F7EULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\gra_grass.msh"},
	{0xFAB4A4184D149FDFULL, "vxht_mansion_n_5cu_0_0.vlm"},
	{0xFAB54A330B6EB0AAULL, "gen00fl_dog6115127.cl"},
	{0xFAB5564E512C04E3ULL, "vxlt_warehouse_q7_0_0.vlm"},
	{0xFAB57A241CA67944ULL, "art\\levels\\races\\ferelden\\props\\foundry\\prp_smelt.phy"},
	{0xFAB5BBC939FCCBF2ULL, "bee_ball.mao"},
	{0xFAB61B2FC029C552ULL, "gen00fl_isabela_init6127652.cl"},
	{0xFAB63BD52CD97AB9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_right.gad"},
	{0xFAB6AC28A86533F5ULL, "w_encounter2_camp_39_0.phy"},
	{0xFAB6C8AF49551384ULL, "6031168_f.fxe"},
	{0xFAB6FA7AB63D6F50ULL, "vxht_playershouse_n_1wc_0_0.vlm"},
	{0xFAB6FEA35ED8344EULL, "llac_lt_hangedman_n_-1_22.dds"},
	{0xFAB70E5465627B5DULL, "vxht_brothel_n_6gc_0_0.vlm"},
	{0xFAB7D324252B398FULL, "dae300_alistair_king6133782.cl"},
	{0xFAB7E2120365B7A3ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1b.mao"},
	{0xFAB7E6120365BEF0ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1b.mmh"},
	{0xFAB7F4120365D60AULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_bookclose1b.msh"},
	{0xFAB800753864CFFCULL, "442814115.ogg"},
	{0xFAB84B7437EF6434ULL, "vxht_0_hightown_a2_8pq_0_0.vlm"},
	{0xFAB84B7CABA2DE1DULL, "6167988_m.fxe"},
	{0xFAB87D2C8D7A6C9FULL, "dae000_serendipity6161404.cl"},
	{0xFAB88029559139A0ULL, "mrl150_dalish_guard6068335.cl"},
	{0xFAB8CAC2351A7843ULL, "6121213_f.fxe"},
	{0xFAB910F6EE0726A1ULL, "478155647.ogg"},
	{0xFAB936884A091345ULL, "6169186_m.fxe"},
	{0xFABA07FB27C17004ULL, "hlac_lt_0_lowtown_a2_-65535_1y.dds"},
	{0xFABA3B077A184B56ULL, "hlac_lt_0_lowtown_65534_1y.dds"},
	{0xFABA4AE7B7564E1DULL, "plt_mag320pt_otl_evelina.nss"},
	{0xFABA8A0ECDE17F68ULL, "mag291_torpor6025999.cl"},
	{0xFABAD542E4D4CCACULL, "6129370_m.fxe"},
	{0xFABBAA63A6D82A44ULL, "hlac_lt_undercityserialkiller_-1_26x.dds"},
	{0xFABC1CBBB455EA35ULL, "6027786_f.fxe"},
	{0xFABC75B0979FF68CULL, "415237741.ogg"},
	{0xFABC7A40118A5ABBULL, "w_blightlands_68_0.mmh"},
	{0xFABC8640118A6F58ULL, "w_blightlands_68_0.mao"},
	{0xFABC9840118A8DA5ULL, "w_blightlands_68_0.msh"},
	{0xFABCB3958AFBE3D9ULL, "hlac_lt_0_lowtown_a3_n_c_65537_0z.dds"},
	{0xFABD0F427115C693ULL, "dae200cr_viscount6087247.cl"},
	{0xFABD4CD52CDFB607ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\left_hand_point_right.ani"},
	{0xFABDC314F237A2C0ULL, "tm_hwp_h01.tnt"},
	{0xFABDCD28A86B8A21ULL, "w_encounter2_camp_39_0.rml"},
	{0xFABDFA72AAE23CE4ULL, "vxht_bartrand_n_34n_0_0.vlm"},
	{0xFABE0C04579C64C9ULL, "81293916.ogg"},
	{0xFABE31B21DD18C69ULL, "campaign_base_it-it.cre"},
	{0xFABE3BF61D72D7ACULL, "vxlt_undercityserialkiller_1ig_0_1.vlm"},
	{0xFABF180F28149513ULL, "786608347.ogg"},
	{0xFABF216726C26654ULL, "zz_jh.ncs"},
	{0xFABFFEC5FD9ECFA6ULL, "vxht_mansion2b_2eo_0_1.vlm"},
	{0xFAC024ABCF6006E4ULL, "hlac_lt_0_lowtown_n_-2_4y.dds"},
	{0xFAC055652240F124ULL, "1029887818.ogg"},
	{0xFAC0A89399FF8C70ULL, "llac_lt_0_lowtown_a3_n_c_65536_6.dds"},
	{0xFAC0C5779F78CB9CULL, "287923456.ogg"},
	{0xFAC13C94CBEE3FAAULL, "140854925.ogg"},
	{0xFAC140F71C0B5E47ULL, "vxlt_0_lowtown_a2_gnw_0_0.vlm"},
	{0xFAC1CC511BD97164ULL, "face_rendertodepth_p2.vsh.fxo.dx11"},
	{0xFAC25FDFBA994DBBULL, "vxlt_0_lowtown_a2_n_hhm_0_0.vlm"},
	{0xFAC26F6CF848B53DULL, "vxlt_hangedman_n_1vn_0_0.vlm"},
	{0xFAC2AB4177FC3300ULL, "hlac_w_encounter3_-327682_0x.dds"},
	{0xFAC2C6C30EECDA82ULL, "vxht_brothel_n_634_0_0.vlm"},
	{0xFAC3440DE4B616DFULL, "vxlt_hangedman_242_0_0.vlm"},
	{0xFAC3C82B43221D63ULL, "vxlt_warehouse_z6_0_0.vlm"},
	{0xFAC3DEEA200B0048ULL, "hlac_lt_gallowstemplar_2384_5x.dds"},
	{0xFAC44184B4FE194AULL, "vxdr_primevalentrance_32w_0_0.vlm"},
	{0xFAC53A59AA87BAA8ULL, "764000001.ogg"},
	{0xFAC5AFF6A3E7F7A5ULL, "hlac_ht_keep_n_3732_5y.dds"},
	{0xFAC64217F1F80A75ULL, "577319360.ogg"},
	{0xFAC6633C699EA116ULL, "205987340.ogg"},
	{0xFAC6C2C7675747F2ULL, "984131202.ogg"},
	{0xFAC76FF18080220BULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockback.ani"},
	{0xFAC7F5423DEF660DULL, "vxlt_darktown_6rg_0_0.vlm"},
	{0xFAC8799FE558446DULL, "2525298.ogg"},
	{0xFAC8AA8D2305D70FULL, "gen00fl_anders6125328.cl"},
	{0xFAC8C049B72C3FD4ULL, "447871533.ogg"},
	{0xFAC8FCA3C8119E7FULL, "83779047.ogg"},
	{0xFAC97302E95CFE0AULL, "193629862.ogg"},
	{0xFAC978AC3C353AFBULL, "82202098.ogg"},
	{0xFAC9D9585D6F879AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01a.mmh"},
	{0xFAC9E3FB3377D170ULL, "zz_one_debug6073993.cl"},
	{0xFAC9EB2A2099AE3FULL, "vxht_bartrand_n_5wf_0_0.vlm"},
	{0xFAC9EB585D6FA6E0ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01a.msh"},
	{0xFAC9FC7248F618D3ULL, "6082984_m.fxe"},
	{0xFACA7336A5819DC4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archbase.msh"},
	{0xFACA8536A581BC2DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archbase.mao"},
	{0xFACA8936A581C366ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archbase.mmh"},
	{0xFACAC57321F7AA79ULL, "6200029_m.fxe"},
	{0xFACBD7C928A11FFFULL, "gen_im_trs_krk_hgh_junk3_01.uti"},
	{0xFACBF8742ADBCD16ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.dura.ani"},
	{0xFACC44AC594BED80ULL, "6160755_m.fxe"},
	{0xFACC6C4F9CD85564ULL, "vxht_bartrand_n_33o_0_0.vlm"},
	{0xFACC7566942C7E3BULL, "hlac_lt_undercityserialkiller_-1_33y.dds"},
	{0xFACCA59171A91F37ULL, "vxca_caverns2prc_a5j_0_0.vlm"},
	{0xFACCF6B1307759C5ULL, "llac_lt_0_lowtown_65538_3.dds"},
	{0xFACCFB943E745C57ULL, "vxht_brothel_n_6ke_0_0.vlm"},
	{0xFACD71881301D8A9ULL, "6089925_m.fxe"},
	{0xFACDA76E8FDD52DBULL, "6147885_m.fxe"},
	{0xFACE12FB9315C1AEULL, "vxlt_0_lowtown_docks_a3_d0q_0_0.vlm"},
	{0xFACE32732ED96D14ULL, "lgt220_mag_apost_apost3.cnv"},
	{0xFACF552FEA6A12B9ULL, "vxca_caverns2prc_8gs_0_0.vlm"},
	{0xFACF7DB38343C117ULL, "6028917_m.fxe"},
	{0xFACF8395523E1F13ULL, "vxca_cavern1_haunted_j2k_0_0.vlm"},
	{0xFACFDDE024DEE2E2ULL, "6129471_m.fxe"},
	{0xFACFEAF8A0A49229ULL, "127645523.ogg"},
	{0xFAD079C07E77DB9CULL, "780171862.ogg"},
	{0xFAD10CDCE451F0E5ULL, "442163894.ogg"},
	{0xFAD1249A65A0FD11ULL, "llac_lt_warehouse_n_65537_1.dds"},
	{0xFAD13D905C4A4C2FULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.cess.gad"},
	{0xFAD197DD8353F111ULL, "6068880_f.fxe"},
	{0xFAD1B6E4F3D1F65DULL, "gen00fl_aveline6122760.cl"},
	{0xFAD214732EDD2354ULL, "lgt220_mag_apost_apost3.bnk"},
	{0xFAD26367352D6074ULL, "6067285_m.fxe"},
	{0xFAD2E0D3B6A41651ULL, "art\\vfx\\combat\\vfx_g_spear_att_back_c.impa.ani"},
	{0xFAD311313A402490ULL, "vxdr_primevaltomb_6ue_0_0.vlm"},
	{0xFAD33F995DEB4BAAULL, "ave200_donnic6111711.cl"},
	{0xFAD3C26180D5DAFEULL, "75857228.ogg"},
	{0xFAD3DDD0D97CAD61ULL, "mcr371_hostile_fenris6082254.cl"},
	{0xFAD402924B03B247ULL, "hlac_w_encounter1_-65532_0z.dds"},
	{0xFAD406716E64E044ULL, "17851556.ogg"},
	{0xFAD48036A58A0831ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_archbase.phy"},
	{0xFAD489FC7B1FB459ULL, "6050079_m.fxe"},
	{0xFAD4DE5F401654F7ULL, "vxht_0_hightown_a3_bzw_0_0.vlm"},
	{0xFAD4EB85B3058669ULL, "vxca_caverns2_7xm_0_1.vlm"},
	{0xFAD5056993B90F17ULL, "aip_streetthug_base.gda"},
	{0xFAD52042AADFC3A7ULL, "art\\characters\\weapons\\arrows\\textures\\w_arw_001_sa.dds"},
	{0xFAD557F1808C3682ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockback.evt"},
	{0xFAD572E36BD0EE5FULL, "dae000cr_grand_cleric6148094.cl"},
	{0xFAD5E45F2B197C4EULL, "gen00fl_dog6115075.cl"},
	{0xFAD66E087BAC5D7FULL, "611038574.ogg"},
	{0xFAD6B3305DAA381DULL, "6095300_m.fxe"},
	{0xFAD6CF5B5699A4BCULL, "vxht_0_hightown_n_bwu_0_0.vlm"},
	{0xFAD6EFB168C196E6ULL, "191601101.ogg"},
	{0xFAD71D53B35237C3ULL, "hlac_lt_gamlenshouse_n_-1_7x.dds"},
	{0xFAD731B44B534F5BULL, "static_static_p35.psh.fxo.dx11"},
	{0xFAD78A32F95550E6ULL, "42333540.ogg"},
	{0xFAD7F63C0C316ED5ULL, "6178539_m.fxe"},
	{0xFAD80E2D0F0E568AULL, "nat380_nathaniel6068344.cl"},
	{0xFAD8191A04F73B79ULL, "hlac_lt_warehouse_n_1_6x.dds"},
	{0xFAD852905C508EDCULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.cess.evt"},
	{0xFAD85D07128FC6BFULL, "vxht_bartrand_n_5za_0_0.vlm"},
	{0xFAD8A39F9EB77640ULL, "lgt000pt_chanter.ncs"},
	{0xFAD8EBD3B6A894F3ULL, "art\\vfx\\combat\\vfx_g_spear_att_back_c.impa.gad"},
	{0xFAD8F3A9385C2269ULL, "vxdr_primevaltomb_2w0_0_2.vlm"},
	{0xFAD9486583D6E6E2ULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_cth_comb_0t.dds"},
	{0xFAD9D9DEE67991A3ULL, "6163601_m.fxe"},
	{0xFAD9E2551E473A93ULL, "art\\characters\\weapons\\greatswords\\w_gsw_007a.phy"},
	{0xFADA00742AE81807ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.dura.evt"},
	{0xFADA317643EF454BULL, "6207400_m.fxe"},
	{0xFADA63605C50D675ULL, "guitypes.gda"},
	{0xFADADF6503B0BB7BULL, "6087099_m.fxe"},
	{0xFADB0CACB8695612ULL, "vxht_brothel_n_5e8_0_1.vlm"},
	{0xFADBA2BBF04AFC0DULL, "6162561_m.fxe"},
	{0xFADBED023C3C8E71ULL, "vxdr_primevalentrance_27f_0_0.vlm"},
	{0xFADC56F180925395ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\knockback.gad"},
	{0xFADD4C8A1816C898ULL, "hlac_ht_chantry_n_2534_27y.dds"},
	{0xFADD7E4E5F98986FULL, "1925927.ogg"},
	{0xFADE12408570C6F5ULL, "vxdr_primevaltomb_7bm_0_1.vlm"},
	{0xFADE8E9E50A70F62ULL, "art\\vfx\\models\\vfx_lyrium_shard.msh"},
	{0xFADE909E50A71218ULL, "art\\vfx\\models\\vfx_lyrium_shard.mmh"},
	{0xFADECBC8ACAED924ULL, "dae000_gen_turn_in_26178486.cl"},
	{0xFADEDA3D5DEB140AULL, "6172088_m.fxe"},
	{0xFADF16E9FFB52DB2ULL, "lt_gallowsprison_n.arl"},
	{0xFADFA008F2761B82ULL, "vxlt_0_lowtown_a3_n_c_hex_0_0.vlm"},
	{0xFADFAE82ABE96732ULL, "180577345.ogg"},
	{0xFADFE8D3B6AEAE58ULL, "art\\vfx\\combat\\vfx_g_spear_att_back_c.impa.evt"},
	{0xFAE05D2D8A3C274AULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v3.gad"},
	{0xFAE0D5DD9F783B4CULL, "w_sundermount_254_0.msh"},
	{0xFAE0DBDD9F78453EULL, "w_sundermount_254_0.mmh"},
	{0xFAE0E7DD9F7859D5ULL, "w_sundermount_254_0.mao"},
	{0xFAE121742AEE6F50ULL, "art\\vfx\\combat\\vfx_w_gn2_threat_loop_ho_c.dura.gad"},
	{0xFAE1288EEC89B544ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\8co_basic\\1wo_diagonal_1.mao"},
	{0xFAE12934F872F331ULL, "6064678_m.fxe"},
	{0xFAE134DAC1B17B17ULL, "233334243.ogg"},
	{0xFAE14D8B033569DBULL, "llac_lt_0_lowtown_docks_-131073_5.dds"},
	{0xFAE1830E2C363D10ULL, "vxdr_primevaltomb_6xb_0_0.vlm"},
	{0xFAE1A4E5DAD80D79ULL, "vxca_caverns2prc_5x7_0_0.vlm"},
	{0xFAE1FA3EA0090014ULL, "gen_im_acc_rng_act3_10.uti"},
	{0xFAE218A9953DD0DAULL, "genip_pp_cav_standing.utp"},
	{0xFAE218F42A54352AULL, "6164292_m.fxe"},
	{0xFAE21C5181B114BAULL, "follower_banter6163861.cl"},
	{0xFAE282AF6515CC29ULL, "6149557_m.fxe"},
	{0xFAE28516ABE42761ULL, "gen_im_trs_dem_sta_fir_r2.uti"},
	{0xFAE2C1F773A0FFB0ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemon.msh"},
	{0xFAE2CFF773A117CAULL, "art\\characters\\creatures\\pridedemon\\c_pridedemon.mmh"},
	{0xFAE2D3F773A11E01ULL, "art\\characters\\creatures\\pridedemon\\c_pridedemon.mao"},
	{0xFAE377E457ABC1E4ULL, "385012661.ogg"},
	{0xFAE38C91CFD0E46DULL, "6146817_m.fxe"},
	{0xFAE467478B4766F4ULL, "378686486.ogg"},
	{0xFAE467A57A5BE985ULL, "vxlt_undercityencounter_1g9_0_0.vlm"},
	{0xFAE4944B1E9D2C2CULL, "vxca_caverns1_mines_gkj_0_0.vlm"},
	{0xFAE4E101FC9E90FAULL, "6125668_m.fxe"},
	{0xFAE56757A3283E3CULL, "497261832.ogg"},
	{0xFAE589F0432F63DBULL, "l_one130ar_disciple_dungeon.wbd"},
	{0xFAE5B9124E3FED64ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16man.mmh"},
	{0xFAE5BD124E3FF3B7ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16man.mao"},
	{0xFAE5CF124E401206ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp16man.msh"},
	{0xFAE65A905C5CD9F5ULL, "art\\vfx\\combat\\vfx_m_pri_cone_cold_c.cess.ani"},
	{0xFAE66D92E7E68651ULL, "6102761_m.fxe"},
	{0xFAE6AC18353506FDULL, "159313663.ogg"},
	{0xFAE6AC5BC2D96F83ULL, "gen_im_cns_cft_bmb_03.uti"},
	{0xFAE6F5040BA2795AULL, "dre000_varric6027781.cl"},
	{0xFAE734846E50390FULL, "516573033.ogg"},
	{0xFAE7402D8A4214E1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v3.evt"},
	{0xFAE858F33AAD2A94ULL, "qcr220_orsino6061778.cl"},
	{0xFAE8FFA9961D0F35ULL, "vxht_mansion2_3by_0_0.vlm"},
	{0xFAE904B942B03593ULL, "gen00fl_anders6125272.cl"},
	{0xFAE9C6EB819F45D6ULL, "6149280_m.fxe"},
	{0xFAEA310CCF095FC4ULL, "357497326.ogg"},
	{0xFAEA4029369DFAF4ULL, "635354770.ogg"},
	{0xFAEA9F459861BB13ULL, "zz_car_debug6122476.cl"},
	{0xFAEADB4DF1A96B07ULL, "llac_lt_0_lowtown_n_65534_6.dds"},
	{0xFAEAFCCD5F490672ULL, "6061787_f.fxe"},
	{0xFAEBACC69CA0F7DFULL, "vxlt_hangedman_1zw_0_0.vlm"},
	{0xFAEBDC5560741242ULL, "vxdr_primevaltomb_6qk_0_0.vlm"},
	{0xFAEC594574BC73E5ULL, "combat_lc_r1_a2.gda"},
	{0xFAEC612FAD1C5C7AULL, "gen00fl_varric6033710.cl"},
	{0xFAEC7111FD522102ULL, "6083429_m.fxe"},
	{0xFAECF1FE6E4DE47AULL, "6041135_m.fxe"},
	{0xFAED475AD92E991FULL, "vxlt_gallowstemplar_f_26f_0_0.vlm"},
	{0xFAED7460A94E2742ULL, "6180683_m.fxe"},
	{0xFAEE0B763359506AULL, "hlac_ht_bartrand_n_3382_0y.dds"},
	{0xFAEE1ED827A4673BULL, "vxlt_shop_1x_0_0.vlm"},
	{0xFAEEF6ED44273D0BULL, "110199628.ogg"},
	{0xFAEEF8585D8ED8A5ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_lip01a.phy"},
	{0xFAEF788D12FD6F2CULL, "6110382_f.fxe"},
	{0xFAEFD31F6FFC1355ULL, "79046113.ogg"},
	{0xFAEFD508BC3AC949ULL, "lgt351tr_shw_mer_mettin.ncs"},
	{0xFAF009E6002AF9B5ULL, "170349366.ogg"},
	{0xFAF07377A7C7045FULL, "mrl351_phantomelves6170332.cl"},
	{0xFAF0BA3696EAE178ULL, "6175087_m.fxe"},
	{0xFAF0EAE60A70831BULL, "art\\levels\\ztest\\chad\\textures\\prp_orlaischair_d.dds"},
	{0xFAF16005E146B709ULL, "hla_ht_0_hightown_-131072_0z.dds"},
	{0xFAF2235954CA88CEULL, "gen00fl_isabela6135661.cl"},
	{0xFAF2EDAA7F83BA46ULL, "plt_cod_itm_valdasine.plo"},
	{0xFAF352DEEE94A97AULL, "6122768_m.fxe"},
	{0xFAF3D8345422837BULL, "6088414_m.fxe"},
	{0xFAF514E55DD7D95FULL, "445118897.ogg"},
	{0xFAF5255AE30E839BULL, "hlac_ca_caverns2_5802_48x.dds"},
	{0xFAF5382D8A4E4430ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v3.ani"},
	{0xFAF5598701BFF498ULL, "follower_banter6136704.cl"},
	{0xFAF6089C388BFDD1ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\powerattack.ani"},
	{0xFAF6342752FAA7BEULL, "vxdr_0_deeproads_cwv_0_0.vlm"},
	{0xFAF63E9BCDB601D6ULL, "hlac_ht_0_hightown_n_-65538_0x.dds"},
	{0xFAF677E8599AB396ULL, "hlac_lt_0_lowtown_a2_n_c_65533_1y.dds"},
	{0xFAF6B563E3D942BAULL, "gen_im_arm_cht_nobc_t5.uti"},
	{0xFAF7128B0756100CULL, "vxlt_gallowstemplar_f_373_0_0.vlm"},
	{0xFAF71919D9BEBE99ULL, "fen310_fenris6118231.cl"},
	{0xFAF7671F4244C564ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_05ruinstower.msh"},
	{0xFAF7791F4244E3CDULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_05ruinstower.mao"},
	{0xFAF77D1F4244EA06ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_05ruinstower.mmh"},
	{0xFAF838C873955E56ULL, "6142380_m.fxe"},
	{0xFAF83DF0D39C9AEBULL, "6206982_m.fxe"},
	{0xFAF8A8A0658D6C54ULL, "hlac_ca_caverns1_mines_24421_8y.dds"},
	{0xFAF8B6FA5333BA8FULL, "vxca_caverns2prc_aer_0_0.vlm"},
	{0xFAF9A200DB08595DULL, "425592947.ogg"},
	{0xFAFAE1C977122C20ULL, "follower_banter6171051.cl"},
	{0xFAFB0A9C0EC22CDCULL, "ts_beasts_dragons_03.gda"},
	{0xFAFB470AC5DEC371ULL, "d15_b_sp_color.dds"},
	{0xFAFB9A5E21FC8A64ULL, "vxht_0_hightown_n_5tc_0_0.vlm"},
	{0xFAFBB07B4D076ADAULL, "783931364.ogg"},
	{0xFAFBB4E155449BA4ULL, "w_sundermount_157_0.msh"},
	{0xFAFBB68B17128DA5ULL, "141091977.ogg"},
	{0xFAFBC6E15544BA0DULL, "w_sundermount_157_0.mao"},
	{0xFAFBCAE15544C146ULL, "w_sundermount_157_0.mmh"},
	{0xFAFBFE804EE1277EULL, "and100_anders6095862.cl"},
	{0xFAFC0862D59F7A7FULL, "18732271.ogg"},
	{0xFAFC139C38907C73ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\powerattack.gad"},
	{0xFAFC83C2D6624C82ULL, "vxlt_0_lowtown_a3_n_c_hoz_0_0.vlm"},
	{0xFAFC9AB43DF4ED76ULL, "49249372.ogg"},
	{0xFAFD4C98597C0736ULL, "866164894.ogg"},
	{0xFAFDA54FB4A4490AULL, "6147787_m.fxe"},
	{0xFAFDBF28A8A15610ULL, "w_encounter2_camp_39_0.msh"},
	{0xFAFDCD28A8A16E2AULL, "w_encounter2_camp_39_0.mmh"},
	{0xFAFDD128A8A17561ULL, "w_encounter2_camp_39_0.mao"},
	{0xFAFDDCDAE2EF34B9ULL, "vxdr_primevalentrance_2sj_0_0.vlm"},
	{0xFAFDF4E7DFC96412ULL, "620415589.ogg"},
	{0xFAFE0C8937CFA4E9ULL, "674940830.ogg"},
	{0xFAFE95C01018C1E5ULL, "hlac_ht_0_hightown_a2_-65538_0z.dds"},
	{0xFAFED35FDE546C4FULL, "vxht_brothel_n_349_0_0.vlm"},
	{0xFAFEDA9CE661A83BULL, "root_war_reaver.dds"},
	{0xFAFFE4DBE2805AFFULL, "698095808.ogg"},
	{0xFAFFF8FB566C6F36ULL, "6151058_m.fxe"},
	{0xFB003020F9C8E085ULL, "zz_followers_banter6157937.cl"},
	{0xFB00BAF6AA46E48BULL, "mag121_idunna6084116.cl"},
	{0xFB00BE08BC49152CULL, "lgt351tr_shw_mer_mettin.utt"},
	{0xFB0105112251F2BEULL, "vxht_chantry_2ot_0_0.vlm"},
	{0xFB014E37DDF22C57ULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef04.phy"},
	{0xFB01741F424D2FD1ULL, "art\\levels\\races\\natural\\blightlands\\props\\prp_05ruinstower.phy"},
	{0xFB02364B73A83326ULL, "vxht_0_hightown_a3_amt_0_0.vlm"},
	{0xFB0239CD68A2F0AAULL, "vxca_caverns1_mines_hlz_0_0.vlm"},
	{0xFB024A1146189BFBULL, "6083514_m.fxe"},
	{0xFB02562498FBFA6EULL, "6001889_m.fxe"},
	{0xFB029A89C20C8BF5ULL, "vxlt_0_lowtown_docks_a3_col_0_0.vlm"},
	{0xFB02ACDC236D6D63ULL, "vxca_caverns1_i2m_0_0.vlm"},
	{0xFB02D29754FECAC6ULL, "316605360.ogg"},
	{0xFB03109C389695D8ULL, "art\\characters\\playercharacter\\elfmale\\animation\\trooperarcher\\powerattack.evt"},
	{0xFB03373D5001B8F9ULL, "hlac_lt_0_lowtown_a3_n_c_2_0z.dds"},
	{0xFB041E73515DC75DULL, "one300_murder_followers6086770.cl"},
	{0xFB04A60444F0C03EULL, "vxdr_0_deeproads_clw_0_0.vlm"},
	{0xFB04FE04B2949097ULL, "hlac_lt_0_lowtown_docks_n_131071_3x.dds"},
	{0xFB05BBDB47DC01E9ULL, "mattscreen_1.cut"},
	{0xFB0646327A6461C9ULL, "vxw_bonepit_27d_0_0.vlm"},
	{0xFB064B5264992B56ULL, "dae000cr_grand_cleric6147676.cl"},
	{0xFB066CBC5254D4CAULL, "ave200_aveline_bar6115514.cl"},
	{0xFB06A0BE6DAAA076ULL, "hlac_lt_0_lowtown_196607_0x.dds"},
	{0xFB06C1E1554EB911ULL, "w_sundermount_157_0.phy"},
	{0xFB06CA00B62C2DB3ULL, "w_bonepitdestroyed_a3_18_0_lc.mmh"},
	{0xFB06F6D70596A780ULL, "follower_banter6170463.cl"},
	{0xFB0706FAB14B772EULL, "ave200_donnic6110617.cl"},
	{0xFB07376DC7568F8FULL, "476505079.ogg"},
	{0xFB08961C44F3DE0EULL, "qun160_vashothlead6053896.cl"},
	{0xFB08ABB6625A00D4ULL, "qun120_arishok_guard6060607.cl"},
	{0xFB091AD1C2F39324ULL, "zz_and_debug6178679.cl"},
	{0xFB096D0572367CE7ULL, "vxlt_darktown_6ob_0_0.vlm"},
	{0xFB098438AB6CA823ULL, "vxlt_0_lowtown_docks_n_c3z_0_0.vlm"},
	{0xFB09E6D2AB2A938BULL, "vxlt_0_lowtown_a3_n_c_eor_0_0.vlm"},
	{0xFB0A0CC8C8DE0630ULL, "lit_aoe_05_l2.mao"},
	{0xFB0B59267256C1B1ULL, "dae010cr_overseer.utc"},
	{0xFB0BB37F7F0C25D8ULL, "t1_mul_bk1.tnt"},
	{0xFB0C1E50782937BBULL, "zz_anderscu.cut"},
	{0xFB0C70B04F8F7B53ULL, "6043994_m.fxe"},
	{0xFB0C9A396DAAEFCCULL, "6120409_m.fxe"},
	{0xFB0CA5114289FCE7ULL, "6108996_m.fxe"},
	{0xFB0CA9A6E95C8016ULL, "597629246.ogg"},
	{0xFB0D0CD87B4A3826ULL, "369523253.ogg"},
	{0xFB0D453CD04A84CFULL, "vxht_brothel_n_338_0_0.vlm"},
	{0xFB0D9BAFB6BC4899ULL, "zz_fen_debug6134197.cl"},
	{0xFB0DB10DE510D3F3ULL, "6173352_m.fxe"},
	{0xFB0DD853EE7B259EULL, "135964132.ogg"},
	{0xFB0E6A91947563D1ULL, "6177853_m.fxe"},
	{0xFB0E7BE52A8B3481ULL, "dae000_oldmangossip.bnk"},
	{0xFB0E94F93E9F9B88ULL, "vxdr_primevaltomb_2u8_0_1.vlm"},
	{0xFB0EE273DC72C55CULL, "hlac_lt_0_lowtown_a2_n_c_-131071_2z.dds"},
	{0xFB0F3FC8149F7EC4ULL, "qun120cr_javaris6042977.cl"},
	{0xFB0F4BC912E92A89ULL, "6112027_m.fxe"},
	{0xFB0F7A2A9AD9CF3DULL, "mag360_thrask6043651.cl"},
	{0xFB0F8A5F4FE1625BULL, "vxht_mansion2_3er_0_0.vlm"},
	{0xFB0F9549FCFBACAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_1_step.ani"},
	{0xFB1050D42ED3C79DULL, "6040553_m.fxe"},
	{0xFB105D7772878FC0ULL, "art\\characters\\weapons\\spears\\w_qui_qspr.mao"},
	{0xFB105E37BAB5191EULL, "vxht_mansion_n_504_0_1.vlm"},
	{0xFB10A828659E4BA6ULL, "vxht_0_hightown_a3_aju_0_0.vlm"},
	{0xFB10AD45979C1D69ULL, "vxdr_primevaltomb_6sj_0_0.vlm"},
	{0xFB10CC5B3022D9E8ULL, "hlac_lt_0_lowtown_docks_n_-4_0z.dds"},
	{0xFB11280BB83C0CE6ULL, "vxlt_0_lowtown_n_gpg_0_0.vlm"},
	{0xFB11475FBF45C4F5ULL, "342268570.ogg"},
	{0xFB119DE52A8DA4F1ULL, "dae000_oldmangossip.cnv"},
	{0xFB12247ADA6828BFULL, "vxht_bartrand_n_5nm_0_0.vlm"},
	{0xFB12B8FDD0A2C3CCULL, "6101465_f.fxe"},
	{0xFB133A9707038830ULL, "art\\vfx\\textures\\level_fx\\vfx_lvl_flare04.dds"},
	{0xFB1345EC2389CBDCULL, "74651083.ogg"},
	{0xFB136D34512043BAULL, "vxdr_primevalentrance_22z_0_0.vlm"},
	{0xFB138170EFC9544AULL, "w_bonepitmining_a2_49_0.mao"},
	{0xFB138570EFC95A99ULL, "w_bonepitmining_a2_49_0.mmh"},
	{0xFB138F70EFC96B87ULL, "w_bonepitmining_a2_49_0.msh"},
	{0xFB13BA6065B69F22ULL, "lgt110_gxa_vgk_silver_order6116985.cl"},
	{0xFB1430B6291D80BBULL, "zz_ser_debug6091929.cl"},
	{0xFB157470FA1290A6ULL, "lgt110_urn_barker6132999.cl"},
	{0xFB158F7E456B5DB8ULL, "6064045_f.fxe"},
	{0xFB15DCF1630D3CC2ULL, "follower_banter6167212.cl"},
	{0xFB1671109661D61FULL, "6127400_f.fxe"},
	{0xFB16CC46B053B1ACULL, "hlac_ht_chantry_n_2534_15x.dds"},
	{0xFB17298660213FA2ULL, "vxlt_0_lowtown_docks_n_azv_0_0.vlm"},
	{0xFB1791CA3994CFF5ULL, "vxlt_gamlenshouse_n_kh_0_0.vlm"},
	{0xFB17C7E25A7760C4ULL, "art\\vfx\\combat\\vfx_r_arc_scloud_act_c.mmh"},
	{0xFB17E8B5881515CCULL, "575265121.ogg"},
	{0xFB17FA489A8B229BULL, "one300dg_solivitus6048309.cl"},
	{0xFB18158126FA9911ULL, "mag100_cullen6082296.cl"},
	{0xFB184DD59EE8D770ULL, "follower_banter6164493.cl"},
	{0xFB1858AF9D20AC0BULL, "vxlt_0_lowtown_a3_n_c_eru_0_0.vlm"},
	{0xFB1989224A80E38AULL, "hlac_ca_caverns1_24424_28y.dds"},
	{0xFB1A523F375A5A43ULL, "llac_ht_playershouse_n_855_24.dds"},
	{0xFB1A9BA5E4EB9AAEULL, "vxlt_undercityencounter_1on_0_0.vlm"},
	{0xFB1B2758BFC5E8A8ULL, "vxw_bonepit_13r_0_0.vlm"},
	{0xFB1B41FE0552608BULL, "ave200_donnic6110496.cl"},
	{0xFB1BE73569770F20ULL, "vxlt_hangedman_10v_0_0.vlm"},
	{0xFB1C114976EE18D8ULL, "vxlt_gallowscourtyard_62d_0_0.vlm"},
	{0xFB1CB008C71494ECULL, "6053564_f.fxe"},
	{0xFB1D2E67CFCB4091ULL, "mag110_vincento6016312.cl"},
	{0xFB1D360211E0BBEAULL, "6105842_m.fxe"},
	{0xFB1D8D49FD07DCFFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_1_step.evt"},
	{0xFB1E617738A20A71ULL, "gen_im_trs_krk_low_junk1_04.uti"},
	{0xFB1E62D0D62FC9F3ULL, "hlac_w_woundedcoast_196611_0z.dds"},
	{0xFB1E88C2EEB4597AULL, "806112835.ogg"},
	{0xFB1EA578ECFF001FULL, "693994843.ogg"},
	{0xFB1F30B893518DE3ULL, "llac_ca_caverns1_mines_24425_2.dds"},
	{0xFB1F50479C0DB4EBULL, "vxht_playershouse_n_18u_0_0.vlm"},
	{0xFB1F8A15B3BBBC34ULL, "6116494_f.fxe"},
	{0xFB2056980B17B8C5ULL, "vxlt_0_lowtown_a2_csj_0_0.vlm"},
	{0xFB205F7E63B76F8CULL, "art\\characters\\creatures\\mubari\\animation\\blendtree\\mabari_explore.anb"},
	{0xFB221BE2BEFACA5EULL, "dae300_petrice6097860.cl"},
	{0xFB229050A4EE4D59ULL, "795951830.ogg"},
	{0xFB22B089F2E23115ULL, "qun120_arishok6052908.cl"},
	{0xFB230C28EBDF7F0DULL, "6129685_m.fxe"},
	{0xFB23147D282AB41BULL, "hlac_lt_blackemporium_1055_22y.dds"},
	{0xFB23B410379E2C87ULL, "zz_isa_debug6133996.cl"},
	{0xFB2443C36C01287BULL, "vxlt_0_lowtown_docks_a3_n_cyk_0_0.vlm"},
	{0xFB246E49FD0DC6E8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\walk_back_1_step.gad"},
	{0xFB24AD5E2C427983ULL, "37858059.ogg"},
	{0xFB25290166E7043AULL, "mcr310_confused_guard6077105.cl"},
	{0xFB2566152EF97898ULL, "6037399_f.fxe"},
	{0xFB2571565C058F5EULL, "hlac_ht_chantry_2534_45z.dds"},
	{0xFB2671DD441F88C1ULL, "vxca_caverns2prc_7qk_0_1.vlm"},
	{0xFB27A163D090E771ULL, "isa321_velasco6117638.cl"},
	{0xFB27E25430D24C1FULL, "lt_undercitydungeon_mmap_ph.dds"},
	{0xFB27F93F81BB2733ULL, "6137038_f.fxe"},
	{0xFB290D82D6E1B32EULL, "vxlt_undercityencounter_1do_0_0.vlm"},
	{0xFB2943425A733CA8ULL, "vxlt_0_lowtown_docks_a3_cx0_0_0.vlm"},
	{0xFB299DFCF4DC83E1ULL, "163083211.ogg"},
	{0xFB29A6B19EBF15E3ULL, "lgt120_orz_bln_official6140201.cl"},
	{0xFB2A753FD90FF679ULL, "6170586.ogg"},
	{0xFB2A832668E43158ULL, "vxlt_gallowscourtyard_69e_0_0.vlm"},
	{0xFB2ADA2F308E7F72ULL, "331847343.ogg"},
	{0xFB2BCA698E4C855AULL, "181264284.ogg"},
	{0xFB2C101C1BF4F239ULL, "vxht_0_hightown_n_aqm_0_0.vlm"},
	{0xFB2C778B9158B79CULL, "llac_w_blightlands_65532_0.dds"},
	{0xFB2C8A18E833C39DULL, "and100_anders6095984.cl"},
	{0xFB2CC97FFD876C29ULL, "6040715_m.fxe"},
	{0xFB2D17D7D3B931A0ULL, "vxht_mansion_n_52u_0_0.vlm"},
	{0xFB2D7640400BAF9AULL, "hlac_w_woundedcoast_-65530_0y.dds"},
	{0xFB2DC4DD851334CCULL, "hlac_w_woundedcoast_0_12x.dds"},
	{0xFB2DDD03BE76A3FEULL, "vxht_mansion2b_30t_0_0.vlm"},
	{0xFB2E01EFF5A74D8FULL, "580945653.ogg"},
	{0xFB2E0BC59C283DE6ULL, "vxlt_0_lowtown_n_gji_0_0.vlm"},
	{0xFB2E2B33C274AAF1ULL, "zz_ave_debug6066169.cl"},
	{0xFB2E7B4B609421F3ULL, "llac_lt_0_lowtown_docks_a3_-131073_4.dds"},
	{0xFB2EF1D61CD00CFAULL, "static_static_p12.vsh.fxo.dx11"},
	{0xFB2FFB9B1AD309BEULL, "vxdr_0_deeproads_ccj_0_0.vlm"},
	{0xFB30F4232CAE681EULL, "6167474_f.fxe"},
	{0xFB31F7275F4AE338ULL, "6143442_m.fxe"},
	{0xFB31F806FC39B084ULL, "398102852.ogg"},
	{0xFB3221DA1870CED0ULL, "6101215_m.fxe"},
	{0xFB3287654F02FA22ULL, "6001265_m.fxe"},
	{0xFB32A2D16F26DEB7ULL, "llac_ht_0_hightown_n_-65537_0.dds"},
	{0xFB3307149C05BAC6ULL, "dae000_meredith6085390.cl"},
	{0xFB33E49160855725ULL, "951219873.ogg"},
	{0xFB33F73B4FB3DCB9ULL, "6143128_m.fxe"},
	{0xFB340B3684E87262ULL, "6042387_m.fxe"},
	{0xFB34855298176947ULL, "6039273_f.fxe"},
	{0xFB34EDE6CF101428ULL, "vxw_woundedcoast_1f_0_0.vlm"},
	{0xFB364CFB484B3DF3ULL, "gen00fl_isabela_init6077415.cl"},
	{0xFB3696445A9F7A08ULL, "llac_lt_0_lowtown_n_-131071_2.dds"},
	{0xFB376258C1E22385ULL, "gen00fl_anders6119871.cl"},
	{0xFB378B739C2B2E99ULL, "gen_im_trs_chn_andraste_01.uti"},
	{0xFB37C8D0D85118EBULL, "329621404.ogg"},
	{0xFB38218C40E4E1BFULL, "vxht_0_hightown_n_cbi_0_1.vlm"},
	{0xFB38A71C127BA580ULL, "vxw_sundermount_1zf_0_0.vlm"},
	{0xFB38C4750CE84C93ULL, "hlac_dr_primevaltomb_659_16x.dds"},
	{0xFB38C5840C01536FULL, "6089699_m.fxe"},
	{0xFB39184345FC5FDBULL, "6030497_m.fxe"},
	{0xFB392411596BE39CULL, "6147099_m.fxe"},
	{0xFB392AA48DF67619ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.cess.ani"},
	{0xFB3933CF92B977B8ULL, "hla_ht_0_hightown_n_131070_0y.dds"},
	{0xFB39350D5B2CC7EEULL, "gen_im_ammo_quiver_03.uti"},
	{0xFB39405F04FE41F4ULL, "llac_lt_0_lowtown_docks_-2_3.dds"},
	{0xFB39493512569EF3ULL, "vxlt_0_lowtown_a2_n_c_gjb_0_0.vlm"},
	{0xFB39EB4ED1C21C46ULL, "lt_0_lowtown_docks_a3_22_0.rml"},
	{0xFB39F112BDD1D657ULL, "276404774.ogg"},
	{0xFB3A89BC58BE08B6ULL, "vxlt_0_lowtown_a3_n_c_gzf_0_0.vlm"},
	{0xFB3AE001E791216AULL, "331303306.ogg"},
	{0xFB3AF7F14ECED97CULL, "56030233.ogg"},
	{0xFB3BBBAE9BF60160ULL, "770236395.ogg"},
	{0xFB3BDB551E99F908ULL, "art\\characters\\weapons\\greatswords\\w_gsw_007a.mmh"},
	{0xFB3BE7551E9A0DABULL, "art\\characters\\weapons\\greatswords\\w_gsw_007a.mao"},
	{0xFB3BEFDC25C7060CULL, "6156308_f.fxe"},
	{0xFB3BF9551E9A2C32ULL, "art\\characters\\weapons\\greatswords\\w_gsw_007a.msh"},
	{0xFB3C1F18CF529F29ULL, "llac_lt_0_lowtown_docks_65532_0.dds"},
	{0xFB3C2780E605F45DULL, "vxw_blightlands_bp_0_0.vlm"},
	{0xFB3C4EE0B06CBC7EULL, "vxht_mansion2b_37u_0_0.vlm"},
	{0xFB3C7DA28E1E5666ULL, "vxlt_0_lowtown_n_goj_0_0.vlm"},
	{0xFB3C8E2F8183F322ULL, "vxlt_0_lowtown_a2_n_ety_0_0.vlm"},
	{0xFB3C91CBD3D710E1ULL, "249048815.ogg"},
	{0xFB3CC4F7A058C4F3ULL, "ico_shield_3.dds"},
	{0xFB3CF1607D7133C3ULL, "6133459_m.fxe"},
	{0xFB3D8B1F2D9DD61FULL, "darktown_fem_lowlife_mob6176161.cl"},
	{0xFB3DC9DDD21FCC8DULL, "1020466921.ogg"},
	{0xFB3DF01DD51755D8ULL, "isa311_isabela6095631.cl"},
	{0xFB3DF60B06B0353DULL, "vxht_mansion_n_6s0_0_0.vlm"},
	{0xFB3DFEC1B7D70C79ULL, "vxlt_0_lowtown_a2_n_c_hl5_0_0.vlm"},
	{0xFB3E20340BBAE2B6ULL, "81187784.ogg"},
	{0xFB3F0F29F28DF53FULL, "w_bonepitmining_a2_234_0.msh"},
	{0xFB3F1529F28DFFE1ULL, "w_bonepitmining_a2_234_0.mmh"},
	{0xFB3F2129F28E1382ULL, "w_bonepitmining_a2_234_0.mao"},
	{0xFB3F3013E9A8BB35ULL, "ser270_emeric6111463.cl"},
	{0xFB3F779F3CE317C1ULL, "mcr371_hostile_anders6082120.cl"},
	{0xFB4005A48DFC565BULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.cess.gad"},
	{0xFB40170C9909B494ULL, "6095110_m.fxe"},
	{0xFB40AF3A8A53FFB1ULL, "docks_scene_dockworkers6176203.cl"},
	{0xFB40CAE25A9A6ABBULL, "art\\vfx\\combat\\vfx_r_arc_scloud_act_c.anb"},
	{0xFB41268C47BB7BC5ULL, "plt_cod_aow_letters.plo"},
	{0xFB413094FC4347E8ULL, "pro000_meeran6050098.cl"},
	{0xFB41592E47F7A390ULL, "gen00fl_aveline6121919.cl"},
	{0xFB41955A223D793BULL, "vxlt_hangedman_32a_0_0.vlm"},
	{0xFB41A5D8131FEF7FULL, "vxlt_0_lowtown_a3_n_c_hpf_0_0.vlm"},
	{0xFB41B2AF0FCDC27DULL, "vxht_chantry_30s_0_0.vlm"},
	{0xFB41C8E32CEDB6A9ULL, "qcr220_orsino6041836.cl"},
	{0xFB42232AD9FBC83FULL, "hlac_dr_dungeon_3216_9x.dds"},
	{0xFB4246DB7853A5EEULL, "vxw_bonepit_lv_0_0.vlm"},
	{0xFB4288959F59F73FULL, "6125738_f.fxe"},
	{0xFB42B24ADAEAB8F7ULL, "640884412.ogg"},
	{0xFB42F4E42CB349A8ULL, "hm_templar_officer01.mor"},
	{0xFB42FDED3799BCAEULL, "hlac_ht_0_hightown_a2_-131073_1z.dds"},
	{0xFB43244ED1084D83ULL, "6040638_m.fxe"},
	{0xFB43DA602F85E61DULL, "6070126_f.fxe"},
	{0xFB441F780BC51AFDULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_3roothanging.mao"},
	{0xFB4423780BC52136ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_3roothanging.mmh"},
	{0xFB442D5BB4DF3058ULL, "6108784_m.fxe"},
	{0xFB442D780BC53234ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\tre_3roothanging.msh"},
	{0xFB443D10488566B4ULL, "ter_default_d.dds"},
	{0xFB44414DAF0F241EULL, "792448480.ogg"},
	{0xFB45A36C4027A263ULL, "zz_followers_banter6157597.cl"},
	{0xFB45C33451E328F1ULL, "6004078_m.fxe"},
	{0xFB465ED99472CFBBULL, "oc_rock_04.mao"},
	{0xFB466B7905790713ULL, "follower_banter6143143.cl"},
	{0xFB4722A48E02A640ULL, "art\\vfx\\combat\\vfx_m_sta_rng_proj_spi_p.cess.evt"},
	{0xFB472CD81ACC4280ULL, "770483627.ogg"},
	{0xFB47B9378D7CDB06ULL, "vxht_mansion2b_32f_0_0.vlm"},
	{0xFB47D30ED6F2A716ULL, "vxw_bonepitmining_a2_5o_0_0.vlm"},
	{0xFB47FDA753F5A320ULL, "llac_dr_primevaltomb_658_9.dds"},
	{0xFB4876D31951EC75ULL, "art\\vfx\\combat\\vfx_w_wns_flurry_c.anb"},
	{0xFB48CC3D8DB7C8ADULL, "567059047.ogg"},
	{0xFB492F37DE2FAD4CULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef04.mmh"},
	{0xFB493537DE2FB73EULL, "art\\characters\\playercharacter\\elffemale\\mt_cbt_ef04.msh"},
	{0xFB493E61DC206D3EULL, "vxht_chantry_2ti_0_0.vlm"},
	{0xFB49751FEC94E2C2ULL, "707384766.ogg"},
	{0xFB498319D6BB8CEEULL, "vxlt_0_lowtown_n_f6p_0_0.vlm"},
	{0xFB4A60B93748289EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_action.evt"},
	{0xFB4A698A6E8BE2CCULL, "114833691.ogg"},
	{0xFB4ABB25D0493526ULL, "6163607_f.fxe"},
	{0xFB4B38CB80FAD879ULL, "root_branch2.mao"},
	{0xFB4B6AFB21B32B01ULL, "6147238_m.fxe"},
	{0xFB4CE2EB0A60E6FFULL, "gen00fl_sebastian6174730.cl"},
	{0xFB4D0EC22839DD35ULL, "6053926_m.fxe"},
	{0xFB4D20BA4509459CULL, "art\\cinematics\\deathblows\\highdragon\\c_pc_swp.gad"},
	{0xFB4DCCCD32085CFCULL, "vxw_bonepitdestroyed_a3_231_0_0.vlm"},
	{0xFB4E128D90C5DC00ULL, "6036957_m.fxe"},
	{0xFB4F75F86EBB5669ULL, "gen00fl_dog6167255.cl"},
	{0xFB4FA6128C8BD10BULL, "vxlt_0_lowtown_gjk_0_0.vlm"},
	{0xFB4FCA2F1FBCB219ULL, "545487794.ogg"},
	{0xFB4FD37EB993323EULL, "hlac_lt_0_lowtown_-65538_0y.dds"},
	{0xFB506AAE0EFD1349ULL, "6144590_m.fxe"},
	{0xFB5077B9374CBBC9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_action.gad"},
	{0xFB509682B72AF3B7ULL, "vxw_woundedcoast_1ny_0_0.vlm"},
	{0xFB50B3EE3FFA2083ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basic.mao"},
	{0xFB50B7EE3FFA27D0ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basic.mmh"},
	{0xFB50C5EE3FFA3FEAULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basic.msh"},
	{0xFB512BABBCBF228FULL, "6147132_f.fxe"},
	{0xFB512E0FBFDA89A1ULL, "6076079_m.fxe"},
	{0xFB51C8C23C8F5295ULL, "vxlt_warehouse_n_dg_0_0.vlm"},
	{0xFB51FDC89FB5CF6DULL, "vxca_cavern1_haunted_ioi_0_0.vlm"},
	{0xFB5214C0BF17D557ULL, "vxw_blightlands_nv_0_0.vlm"},
	{0xFB52757BF0E1B285ULL, "gencr_darkspawn_hurlock_shd.utc"},
	{0xFB52952CFA794CA2ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.mmh"},
	{0xFB52993C16AFBDEBULL, "6161030_m.fxe"},
	{0xFB530B16C3A95D19ULL, "mag160_feynriel6020439.cl"},
	{0xFB541FBA450F629AULL, "art\\cinematics\\deathblows\\highdragon\\c_pc_swp.ani"},
	{0xFB54392F189C3193ULL, "6185119_m.fxe"},
	{0xFB54A0C63377080CULL, "6087378_f.fxe"},
	{0xFB55183F24F28AAEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_reaction.evt"},
	{0xFB552740FEC6B2FDULL, "vxlt_blackemporium_8f_0_0.vlm"},
	{0xFB559B950DD3A228ULL, "dae000_mother6121017.cl"},
	{0xFB55A93DEFED846FULL, "6087499_m.fxe"},
	{0xFB55C34B24A83610ULL, "tools_way_pt01.mmh"},
	{0xFB55D14B24A84E2AULL, "tools_way_pt01.msh"},
	{0xFB5711364A993A41ULL, "nat380_sibling6207579.cl"},
	{0xFB5712E54F5B9D12ULL, "hlac_lt_0_lowtown_docks_a3_131073_0x.dds"},
	{0xFB571D42436DB40FULL, "vxlt_0_lowtown_a2_n_c_hpw_0_0.vlm"},
	{0xFB57280A8102134AULL, "ser191_emeric6108789.cl"},
	{0xFB5739E4B1F1072EULL, "615187127.ogg"},
	{0xFB5748B937528A17ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_exchange_action.ani"},
	{0xFB57FCEA817FABFCULL, "gen00fl_anders6212029.cl"},
	{0xFB585BECBD5A8256ULL, "vxlt_0_lowtown_cms_0_0.vlm"},
	{0xFB587B2D5DE2F144ULL, "hlac_w_encounter1_-65536_0y.dds"},
	{0xFB58A5B5A74ABD76ULL, "vxht_0_hightown_n_c6y_0_0.vlm"},
	{0xFB58B9A9743B2943ULL, "mcr370_cut_player_survives6171847.cl"},
	{0xFB592D515207D4D9ULL, "lowtown_fem_prostitute_sta6152366.cl"},
	{0xFB592F3CE5F19444ULL, "ts_templars_prog_junk1.gda"},
	{0xFB5967AC3C348338ULL, "541219448.ogg"},
	{0xFB59A663E52D8876ULL, "dae000_orsino6154300.cl"},
	{0xFB59D51A615D64C3ULL, "6139259_m.fxe"},
	{0xFB5A8F8B234B7EFEULL, "172984533.ogg"},
	{0xFB5B466B0595CDBDULL, "6140328_f.fxe"},
	{0xFB5C0F3F24F89A59ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_reaction.gad"},
	{0xFB5C63ED9279A8F2ULL, "6088675_m.fxe"},
	{0xFB5C6ACF6E36FA33ULL, "mer141_jansen6017762.cl"},
	{0xFB5C9E2690757159ULL, "dre000_sandal6028394.cl"},
	{0xFB5CDF4BBEE056B2ULL, "203722033.ogg"},
	{0xFB5D0A7534825DDCULL, "vxlt_0_lowtown_n_ent_0_1.vlm"},
	{0xFB5D12611D1A8E0DULL, "6170328_m.fxe"},
	{0xFB5D2554EEFB7D9EULL, "6110808_m.fxe"},
	{0xFB5D33F36CDAD362ULL, "one140cr_charade6134292.cl"},
	{0xFB5DC0E3F59D862FULL, "407255237.ogg"},
	{0xFB5E75D02E7BCCB5ULL, "vxw_encounter3_ey_0_0.vlm"},
	{0xFB5F084A10C266D4ULL, "vxw_sundermount_34n_0_0.vlm"},
	{0xFB5F28CFCED11B5DULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_tool.phy"},
	{0xFB5F75067212E80DULL, "gen_im_trs_dal_amu_r0.uti"},
	{0xFB5FAE111C612397ULL, "vxw_encounter1_fp_0_0.vlm"},
	{0xFB5FEDA8185C4610ULL, "vxlt_warehouse_n_10y_0_0.vlm"},
	{0xFB5FF6C4F02E5EF0ULL, "6133144_m.fxe"},
	{0xFB604923E0718323ULL, "ttx_window_wall_1.msh"},
	{0xFB605723E0719B7DULL, "ttx_window_wall_1.mmh"},
	{0xFB606FA591ABE7EDULL, "vxca_cavern1_haunted_idh_0_0.vlm"},
	{0xFB60869DB10DEDD7ULL, "vxw_blightlands_sq_0_0.vlm"},
	{0xFB60C4F5A89ED9A6ULL, "588980938.ogg"},
	{0xFB610BC57FB59687ULL, "99063624.ogg"},
	{0xFB614DBC3CA7558AULL, "gen00fl_sebastian6179990.cl"},
	{0xFB61DE7675AD3D11ULL, "qun221_petrice6185128.cl"},
	{0xFB62085E458A25A2ULL, "fx_crescent02_blend.mao"},
	{0xFB62D4F69EBA152EULL, "vxlt_undercityencounter_1pc_0_0.vlm"},
	{0xFB62F5EFC89F9C9BULL, "llac_lt_darktown_-1_67.dds"},
	{0xFB630E029E533808ULL, "6027788_m.fxe"},
	{0xFB63203F24FED5A7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\stabbed_front_reaction.ani"},
	{0xFB63C373AC77A25AULL, "ss_tanner_deep_roads_2.cut"},
	{0xFB642E80E5FBD02DULL, "vxca_caverns1_mines_i2z_0_0.vlm"},
	{0xFB64BDE1B8201679ULL, "hlac_lt_0_lowtown_docks_n_-65538_6z.dds"},
	{0xFB657FEB5362FD96ULL, "lt_0_lowtown_a2_44_0_ll.mmh"},
	{0xFB65922A320844B5ULL, "vxht_0_hightown_n_bst_0_0.vlm"},
	{0xFB65B17D56379504ULL, "vxlt_0_lowtown_a3_n_c_gj4_0_0.vlm"},
	{0xFB65B8A51354EF99ULL, "vxht_mansion_n_3ou_0_0.vlm"},
	{0xFB6670B0CBD48210ULL, "vxdr_primevaltomb_5f3_0_0.vlm"},
	{0xFB66DFA19BDB02D9ULL, "235632350.ogg"},
	{0xFB66F5C3414DD67FULL, "and110_refugee6216100.cl"},
	{0xFB67120239FD9FCEULL, "6133883_m.fxe"},
	{0xFB67192821D07615ULL, "6123367_f.fxe"},
	{0xFB67A25931DB114EULL, "vxlt_0_lowtown_a2_n_c_hu9_0_0.vlm"},
	{0xFB680C9A723420A4ULL, "6118203_f.fxe"},
	{0xFB680ED164BB319CULL, "pro000_friedrich6050205.cl"},
	{0xFB6825EA0362B54FULL, "gen00fl_aveline_act3_intro.stg"},
	{0xFB6859280DACFAEAULL, "hla_ht_0_hightown_a3_-65537_0x.dds"},
	{0xFB68E48285DFE899ULL, "hlac_lt_0_lowtown_docks_-262144_3z.dds"},
	{0xFB6925BD02B1341EULL, "6016569_f.fxe"},
	{0xFB69C931FFF8054CULL, "vxca_cave_2b5_0_0.vlm"},
	{0xFB6A1A1598FC0250ULL, "vxdr_0_deeproads_ewj_0_0.vlm"},
	{0xFB6AD3CB49DC75AFULL, "vxht_brothel_n_6fz_0_0.vlm"},
	{0xFB6B2A769915C5D6ULL, "hlac_lt_gallowstemplar_n_2382_0x.dds"},
	{0xFB6B7F4C41E10CA5ULL, "825802772.ogg"},
	{0xFB6BBEBBE84AB562ULL, "6205887_m.fxe"},
	{0xFB6C1CE62B22D8FDULL, "6096473_m.fxe"},
	{0xFB6C224FDD2CCE1BULL, "lt_0_lowtown_docks_a3_n_201_0.msh"},
	{0xFB6C344FDD2CED66ULL, "lt_0_lowtown_docks_a3_n_201_0.mao"},
	{0xFB6C404FDD2D0105ULL, "lt_0_lowtown_docks_a3_n_201_0.mmh"},
	{0xFB6D518172098F5AULL, "6071095_m.fxe"},
	{0xFB6D6DEA17D77FC5ULL, "840618965.ogg"},
	{0xFB6D7A2702B87F54ULL, "vxw_sundermount_33o_0_0.vlm"},
	{0xFB6D8A2B00C0B944ULL, "6146795_m.fxe"},
	{0xFB6DB32C15D80AD9ULL, "vxlt_backalley_n_1yg_0_1.vlm"},
	{0xFB6DDED15E22553BULL, "vxca_caverns1_hkw_0_0.vlm"},
	{0xFB6E1E925A8121BAULL, "6044051_m.fxe"},
	{0xFB6E1FEE0E573C17ULL, "vxw_encounter1_ks_0_0.vlm"},
	{0xFB6E817B878F98D1ULL, "420333390.ogg"},
	{0xFB6EC299E71C1E35ULL, "vxht_keep_6am_0_0.vlm"},
	{0xFB6EC9E26C02EF21ULL, "2949379.ogg"},
	{0xFB6EF87AA3040657ULL, "vxw_blightlands_tp_0_0.vlm"},
	{0xFB6F063BAA0EA577ULL, "lgt320_mis_resp_swordhunter6174169.cl"},
	{0xFB6F1BAA6F93E688ULL, "302420267.ogg"},
	{0xFB6F43DE7F32CAADULL, "vxht_mansion2b_3bi_0_0.vlm"},
	{0xFB6F8570F017B9F2ULL, "w_bonepitmining_a2_49_0.rml"},
	{0xFB6F967C63B0C348ULL, "6162084_f.fxe"},
	{0xFB6FC79057315C23ULL, "genip_pp_hti_sustained.utp"},
	{0xFB6FE513F8C44B9CULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.dura.ani"},
	{0xFB70189B0FEC16DBULL, "6097616_m.fxe"},
	{0xFB701ABAB545C657ULL, "vxw_sundermount_2gm_0_0.vlm"},
	{0xFB701DCFCEDF7BC9ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_tool.mao"},
	{0xFB7029CFCEDF9062ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_tool.mmh"},
	{0xFB702BCFCEDF9318ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_wall8x3_tool.msh"},
	{0xFB7129AE73992AFFULL, "674348904.ogg"},
	{0xFB7174876B79E738ULL, "gui_guipixelshader_p0.psh.fxo"},
	{0xFB7177D31974F2B6ULL, "art\\vfx\\combat\\vfx_w_wns_flurry_c.mmh"},
	{0xFB7188AF446180C8ULL, "vxdr_primevaltomb_6q3_0_2.vlm"},
	{0xFB71C125433ECAC4ULL, "6119165_f.fxe"},
	{0xFB720AE3D8B969CBULL, "vxw_encounter1_dj_0_0.vlm"},
	{0xFB724A73C503B6DBULL, "vxlt_warehouse_bf_0_0.vlm"},
	{0xFB7310A8DA2F00F3ULL, "vxlt_0_lowtown_a2_n_c_gvf_0_0.vlm"},
	{0xFB739B25C1B61E9CULL, "6102953_m.fxe"},
	{0xFB73BE25659F24EEULL, "110830821.ogg"},
	{0xFB741FCB94208B40ULL, "hlac_lt_gallowstemplar_2391_5z.dds"},
	{0xFB743D24B4EC6345ULL, "836210373.ogg"},
	{0xFB74A640B2765564ULL, "mag210_arianni6024528.cl"},
	{0xFB74A7EEB2747A74ULL, "vxw_woundedcoast_xp_0_1.vlm"},
	{0xFB74DE57DAEDCD99ULL, "gen00fl_dog6165767.cl"},
	{0xFB7584AB601E6AE4ULL, "218157415.ogg"},
	{0xFB762497D6591363ULL, "vxht_keep_n_6j1_0_0.vlm"},
	{0xFB76B7974229A221ULL, "hlac_ca_cavern1_haunted_24425_0z.dds"},
	{0xFB76F213F8CA80DEULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.dura.gad"},
	{0xFB771134C020165DULL, "vxw_sundermount_2ai_0_0.vlm"},
	{0xFB7733BA1334EB78ULL, "812689871.ogg"},
	{0xFB779BA123A947CEULL, "32681424.ogg"},
	{0xFB77C6357FAF6E79ULL, "vxlt_0_lowtown_a2_n_c_hx9_0_0.vlm"},
	{0xFB77EDCED63D355FULL, "6138808_f.fxe"},
	{0xFB782B6A82A9E303ULL, "6140556_m.fxe"},
	{0xFB788BF28AF21AD0ULL, "vxdr_0_deeproads_epk_0_0.vlm"},
	{0xFB7953BB55B8EA9BULL, "47976857.ogg"},
	{0xFB79BC27D2851508ULL, "6198711_f.fxe"},
	{0xFB79BE0D714FF97EULL, "mcr370_varric6208279.cl"},
	{0xFB7A1349B94309A4ULL, "vxlt_gallowstemplar_4ss_0_0.vlm"},
	{0xFB7A310DC3DF4A7DULL, "hlac_lt_0_lowtown_docks_-3_8y.dds"},
	{0xFB7A835C6BB81B65ULL, "llac_ht_chantry_2534_24.dds"},
	{0xFB7AB33348841591ULL, "ser221_gamlen6133079.cl"},
	{0xFB7AC0153C6E403BULL, "6143045_m.fxe"},
	{0xFB7B0D8C280F092FULL, "art\\characters\\playercharacter\\elfmale\\textures\\em_hlf_meda_0t.dds"},
	{0xFB7B37206A34E471ULL, "lt_0_lowtown_docks_n_187_0.msh"},
	{0xFB7B45206A34FB8CULL, "lt_0_lowtown_docks_n_187_0.mao"},
	{0xFB7B51206A35102FULL, "lt_0_lowtown_docks_n_187_0.mmh"},
	{0xFB7B6D4BDAF8517FULL, "vxlt_0_lowtown_a3_n_c_hlb_0_0.vlm"},
	{0xFB7B8C604138FAE5ULL, "mcr310_confused_guard6077070.cl"},
	{0xFB7B942CFA9C5049ULL, "art\\vfx\\combat\\vfx_r_arc_rain_arr_proj_p.anb"},
	{0xFB7BAC93D561D9F7ULL, "467642972.ogg"},
	{0xFB7BCB8AFB8F6E33ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magb.phy"},
	{0xFB7C179E4C42D5E3ULL, "vrc321_varric6082691.cl"},
	{0xFB7CBC5B76ADDC70ULL, "m_dae350ar_sundermount.bnk"},
	{0xFB7D1E5912719C15ULL, "vxlt_warehouse_n_ud_0_0.vlm"},
	{0xFB7D30DD51B9956EULL, "hlac_ca_caverns1_mines_24421_1x.dds"},
	{0xFB7D3476D91236B5ULL, "vxht_keep_6jl_0_0.vlm"},
	{0xFB7DCD13F8D060FDULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t2_c.dura.evt"},
	{0xFB7E0C54160B6814ULL, "ht_0_hightownexplosion_n_26_0.rml"},
	{0xFB7E4492160DA9ACULL, "dae300_vashothlead6062809.cl"},
	{0xFB7E6E96CCF6AD8EULL, "dae000_madam_lusine6152075.cl"},
	{0xFB7E878AE1D161EBULL, "373267191.ogg"},
	{0xFB7E8C97A73BDED7ULL, "vxw_sundermount_2hl_0_0.vlm"},
	{0xFB7FEFD4C8F3D8EBULL, "qun110_petrice6018728.cl"},
	{0xFB7FFF631E134E91ULL, "108331474.ogg"},
	{0xFB8006C973E8BDD7ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.dura.gad"},
	{0xFB807CC0CAAF824BULL, "vxw_encounter1_im_0_0.vlm"},
	{0xFB81DCAE038FCA28ULL, "vxw_bonepitmining_a2_1ox_0_0.vlm"},
	{0xFB8255E85D82B81FULL, "6122901_m.fxe"},
	{0xFB826387ECACDE45ULL, "llac_lt_0_lowtown_docks_196608_1.dds"},
	{0xFB8264435508AF8DULL, "427542140.ogg"},
	{0xFB8275E415F475B5ULL, "vxht_0_hightown_n_byr_0_0.vlm"},
	{0xFB82E5783266F6A6ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_market_stall_0_n.dds"},
	{0xFB82FCB2EBDA7A29ULL, "484917598.ogg"},
	{0xFB832E9D370C9DAAULL, "llac_lt_undercitydungeon_-1_21.dds"},
	{0xFB83697301CCF75BULL, "hlac_w_blightlands_131070_0x.dds"},
	{0xFB83CD6A0AE72092ULL, "llac_w_bonepitmining_a2_262147_0.dds"},
	{0xFB842CC490FEA16CULL, "hlac_lt_gallowstemplar_n_2391_4z.dds"},
	{0xFB850085E8401FDBULL, "vxw_encounter3_br_0_0.vlm"},
	{0xFB8546A4A05AB4BEULL, "47330184.ogg"},
	{0xFB8632CEEE1C4C9EULL, "60596851.ogg"},
	{0xFB86E37723BDFC6FULL, "703761043.ogg"},
	{0xFB871BC973EF0064ULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.dura.evt"},
	{0xFB872043CE208813ULL, "6145219_f.fxe"},
	{0xFB8771636ACA6360ULL, "6146841_m.fxe"},
	{0xFB8798D3686B9CE0ULL, "qun221_viscount6085783.cl"},
	{0xFB886BDCC21E6B47ULL, "642296554.ogg"},
	{0xFB892535B1F1327EULL, "441731626.ogg"},
	{0xFB89435D93C5B471ULL, "vxdr_primevalentrance_2sr_0_0.vlm"},
	{0xFB89DF28CCEE69FFULL, "vxlt_0_lowtown_a3_n_c_him_0_0.vlm"},
	{0xFB89E861B5DF1400ULL, "934579322.ogg"},
	{0xFB8A0B17AAE51EAFULL, "6176041_m.fxe"},
	{0xFB8A6D0C6BEAA165ULL, "6139402_m.fxe"},
	{0xFB8A73F0A774EDE9ULL, "6172836_m.fxe"},
	{0xFB8A9CED648EAB77ULL, "294188643.ogg"},
	{0xFB8A9EE69584BB73ULL, "6090341_m.fxe"},
	{0xFB8ABA1333D8F70CULL, "419434467.ogg"},
	{0xFB8ACDB714FD08ACULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_rf.ani"},
	{0xFB8B63FD325788F6ULL, "vxw_woundedcoast_o6_0_0.vlm"},
	{0xFB8B6F48D0E6C77AULL, "vxht_0_hightown_9zp_0_0.vlm"},
	{0xFB8B90360467B495ULL, "vxlt_warehouse_n_xc_0_0.vlm"},
	{0xFB8BB6F91E9C39D9ULL, "llac_lt_0_lowtown_a2_n_c_65533_3.dds"},
	{0xFB8BDD4C243D2564ULL, "311936115.ogg"},
	{0xFB8C23B2E40098E3ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\_textures\\prp_spruce2_dp.dds"},
	{0xFB8C25DE22332612ULL, "vxht_mansion2b_2pa_0_0.vlm"},
	{0xFB8C61B368984F4CULL, "llac_w_sundermount_-131076_0.dds"},
	{0xFB8C63BAE153C527ULL, "6122134_m.fxe"},
	{0xFB8CC8492A5296C2ULL, "follower_banter6141928.cl"},
	{0xFB8CE0D74980814AULL, "vxht_mansion_n_6og_0_0.vlm"},
	{0xFB8CFE749931F757ULL, "vxw_sundermount_2mo_0_0.vlm"},
	{0xFB8D3DDC1CC4385AULL, "6107031_m.fxe"},
	{0xFB8D53F646CB8F42ULL, "25394920.ogg"},
	{0xFB8D6455C2D5E09BULL, "vxlt_0_lowtown_a2_n_c_eoc_0_0.vlm"},
	{0xFB8DBF8DA4FEF53DULL, "hlac_ca_caverns2_5802_35z.dds"},
	{0xFB8DE403F24536DBULL, "vxca_caverns1_mines_ei4_0_1.vlm"},
	{0xFB8E084C5B4FB464ULL, "400559874.ogg"},
	{0xFB8E0B3B3BEF7943ULL, "vxlt_gallowstemplar_n_6sy_0_0.vlm"},
	{0xFB8E1F269602803CULL, "vxht_mansion_n_62f_0_0.vlm"},
	{0xFB8E63D0A5BCB86CULL, "412716890.ogg"},
	{0xFB8EAF2661FC8A04ULL, "6143367_m.fxe"},
	{0xFB8FC00049D23165ULL, "208355318.ogg"},
	{0xFB9034C493364513ULL, "art\\vfx\\cinematic\\vfx_blood_pool_grow_p.impa.ani"},
	{0xFB90BEEE403017EBULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0wi_basic.phy"},
	{0xFB90C1C8E611ADDCULL, "war_shielddefense.dds"},
	{0xFB91061BFFC20A06ULL, "vxlt_gallowstemplar_6dj_0_0.vlm"},
	{0xFB913FBFE30D4DE1ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_enter.ani"},
	{0xFB91844DC45EBC66ULL, "vxca_caverns2prc_7oc_0_0.vlm"},
	{0xFB918B23D1883EDBULL, "vxht_mansion2_30m_0_0.vlm"},
	{0xFB91C203221CEEE6ULL, "dae000_reset_companions.cnv"},
	{0xFB91DAB715033DEEULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_rf.gad"},
	{0xFB936688BB6B01B1ULL, "ser221_gascard6122441.cl"},
	{0xFB93E40B690FA3FCULL, "vxdr_primevaltomb_901_0_0.vlm"},
	{0xFB93F4EEA40C475DULL, "vxw_sundermount_2og_0_0.vlm"},
	{0xFB9402E396F34C45ULL, "mag160_feynriel6033376.cl"},
	{0xFB94120272ABBAEBULL, "mcr370_negotiation6078750.cl"},
	{0xFB944C70652134A7ULL, "mag100_karras6052951.cl"},
	{0xFB94735C049E791AULL, "mer320_hubert6089964.cl"},
	{0xFB9476A9FFB25DDFULL, "399148935.ogg"},
	{0xFB94912B33B9DBC0ULL, "6206880_m.fxe"},
	{0xFB94FD60FFE0E2A7ULL, "6118507_f.fxe"},
	{0xFB95193CB3463338ULL, "vxlt_0_lowtown_docks_a3_n_cp3_0_0.vlm"},
	{0xFB9533C973FB665DULL, "art\\vfx\\combat\\vfx_w_jug_smash_loop_c.dura.ani"},
	{0xFB953FCC77C049E2ULL, "vxca_caverns1_hi0_0_0.vlm"},
	{0xFB9541BFB7FDA21FULL, "6077841_m.fxe"},
	{0xFB955C1B5DE4024AULL, "mag320_mage_conspirator6040770.cl"},
	{0xFB95C4032220DAA6ULL, "dae000_reset_companions.bnk"},
	{0xFB95DBF47E3D7F37ULL, "ter_cracked01_d.dds"},
	{0xFB9604CE97F3BEC0ULL, "125731700.ogg"},
	{0xFB96A775D68246F1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_carver.ani"},
	{0xFB96CA4B24DFD92BULL, "tools_way_pt01.phy"},
	{0xFB96D846BE11C939ULL, "6180969_m.fxe"},
	{0xFB970C1EAAF3B572ULL, "vxw_woundedcoast_1bt_0_0.vlm"},
	{0xFB9722386EA50CD8ULL, "hlac_w_sundermount_1_0z.dds"},
	{0xFB97547DB7F23CA8ULL, "lop.cut"},
	{0xFB97E3804186A479ULL, "w_bonepitmining_a2_50_0.msh"},
	{0xFB97ED804186B567ULL, "w_bonepitmining_a2_50_0.mmh"},
	{0xFB97F1804186BBB4ULL, "w_bonepitmining_a2_50_0.mao"},
	{0xFB97F7FC12A8BA48ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_7y.dds"},
	{0xFB98186EE7926E20ULL, "vxca_caverns2_7m7_0_1.vlm"},
	{0xFB981D323EE4F211ULL, "ca_cavern1_haunted_mmi.gff"},
	{0xFB984ABFE3137E83ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_enter.gad"},
	{0xFB984FDDB282DF35ULL, "static_static_p28.psh.fxo.dx11"},
	{0xFB986BA5A1BB50E3ULL, "382779454.ogg"},
	{0xFB9878A62835FEBCULL, "qun160_vashothlead6052567.cl"},
	{0xFB98B63211C1A3D0ULL, "gen00fl_fenris6105880.cl"},
	{0xFB98D5B7150953CDULL, "art\\characters\\creatures\\mubari\\animation\\wk_p_rf.evt"},
	{0xFB98E1C53FCF4FACULL, "282258902.ogg"},
	{0xFB9948C21F8B9CC0ULL, "hlac_lt_0_lowtown_docks_n_0_0y.dds"},
	{0xFB99D5DA244DA176ULL, "vxw_woundedcoast_d7_0_0.vlm"},
	{0xFB99EB21A82D32BDULL, "181811986.ogg"},
	{0xFB99FDC24D6D5F6BULL, "dae000_bartrand6206912.cl"},
	{0xFB9A4E712D76D06CULL, "fen110_fenris6100266.cl"},
	{0xFB9A640A30FD1BAFULL, "vxw_sundermount_22j_0_0.vlm"},
	{0xFB9A97BB14293E92ULL, "vxht_mansion2b_2mb_0_0.vlm"},
	{0xFB9AE1D332354A0FULL, "6041097_m.fxe"},
	{0xFB9B65C73FE57761ULL, "gen00fl_varric6088829.cl"},
	{0xFB9C43E81F94D0CFULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2open.mao"},
	{0xFB9C4FE81F94E56CULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2open.mmh"},
	{0xFB9C55E81F94EF5EULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ww_curt2open.msh"},
	{0xFB9C7232DFA3B4EDULL, "startmenu.gfx"},
	{0xFB9C83BED6158ECBULL, "hlac_w_encounter2_camp_-131069_0x.dds"},
	{0xFB9CA0F7EBF1D04AULL, "6063820_f.fxe"},
	{0xFB9CB417F1770F18ULL, "6061328_m.fxe"},
	{0xFB9D275B6B6D4A4FULL, "1029890912.ogg"},
	{0xFB9DB275D6887793ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_carver.gad"},
	{0xFB9DD0400730D15FULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_1.gad"},
	{0xFB9DD84D5C1559EAULL, "hlac_ca_caverns1_24425_7y.dds"},
	{0xFB9DDA40937D4EBBULL, "vxlt_0_lowtown_docks_ber_0_0.vlm"},
	{0xFB9E2CC49342754AULL, "art\\vfx\\cinematic\\vfx_blood_pool_grow_p.impa.evt"},
	{0xFB9E72144C801AA7ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_grogbottlecin_sa.dds"},
	{0xFB9ED80E8D43F7B3ULL, "vxlt_hangedman_n_267_0_0.vlm"},
	{0xFB9ED82E6A5689C0ULL, "6177889_m.fxe"},
	{0xFB9EDF86B58C8214ULL, "6136874_f.fxe"},
	{0xFB9F27BFE3196268ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\dialogue\\leaning_back_low_enter.evt"},
	{0xFB9F500B85F06F4EULL, "dae000ip_crowd_mf1_talk18.utp"},
	{0xFB9F652F65A808AEULL, "llac_lt_0_lowtown_docks_a3_n_65533_4.dds"},
	{0xFB9FDB3987DEA5F2ULL, "975986330.ogg"},
	{0xFBA0F0D234F354FBULL, "vxht_mansion2b_3cp_0_0.vlm"},
	{0xFBA1083D70267DF7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\pp_male_rogue.anb"},
	{0xFBA1688D62A280B1ULL, "sib131_slaver_guard6165291.cl"},
	{0xFBA19E5B76CCC53AULL, "m_dae350ar_sundermount.wbd"},
	{0xFBA1A261FD9E52FFULL, "144839543.ogg"},
	{0xFBA266CB96025FDDULL, "vxw_sundermount_2pf_0_0.vlm"},
	{0xFBA271FFC6DC35C0ULL, "zz_followers_banter6157725.cl"},
	{0xFBA29C45A5FFE718ULL, "art\\vfx\\materials\\gui\\vfx_gui_target_ret_3.mao"},
	{0xFBA2F50D50690B87ULL, "hlac_w_encounter1_65537_0z.dds"},
	{0xFBA33B626A19A7FDULL, "6067727_m.fxe"},
	{0xFBA344E4616F60C8ULL, "llac_w_blightlands_-65533_0.dds"},
	{0xFBA359DFB6B53790ULL, "hlac_lt_0_lowtown_a2_0_5z.dds"},
	{0xFBA39C933900A356ULL, "89754984.ogg"},
	{0xFBA3CC31F45DAD06ULL, "zz_isa_debug6175470.cl"},
	{0xFBA3F3EDCB34EAE8ULL, "219314754.ogg"},
	{0xFBA45B2D48EB3F77ULL, "vxlt_0_lowtown_docks_bte_0_0.vlm"},
	{0xFBA4AF75D68E91F8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\brother_collapse_help_up_carver.evt"},
	{0xFBA4C5400736DD4CULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_1.evt"},
	{0xFBA50BC493485BBDULL, "art\\vfx\\cinematic\\vfx_blood_pool_grow_p.impa.gad"},
	{0xFBA544419A8E989FULL, "223298511.ogg"},
	{0xFBA57547A3FDD6B0ULL, "lgt320_mis_resp_swordhunter6150272.cl"},
	{0xFBA57BC266C36DF5ULL, "tbe100_xenon6189675.cl"},
	{0xFBA57DFB9CE9CDF2ULL, "vxw_woundedcoast_1mu_0_0.vlm"},
	{0xFBA619DF0ABDE585ULL, "hlac_lt_0_lowtown_docks_a3_n_65533_4y.dds"},
	{0xFBA7138E608FF677ULL, "vxlt_gallowscourtyard_7fn_0_0.vlm"},
	{0xFBA7738D7112BE14ULL, "gen00fl_bethany6101493.cl"},
	{0xFBA7AB23DC787A13ULL, "vxw_bonepitmining_a2_2ab_0_0.vlm"},
	{0xFBA7B40A089497A4ULL, "39418982.ogg"},
	{0xFBA847B71643B9F6ULL, "vxw_woundedcoast_a8_0_0.vlm"},
	{0xFBA89AAD70107AEBULL, "6078431_m.fxe"},
	{0xFBA8BB27471AE8C7ULL, "mag101_aombanter6052744.cl"},
	{0xFBA8D5E722F3342FULL, "vxw_sundermount_27u_0_0.vlm"},
	{0xFBA8F08D2E6E1A5DULL, "w_bonepitmining_a2_185_0.phy"},
	{0xFBA95E764043F666ULL, "pro000_athenril6048768.cl"},
	{0xFBA9A656BD8D354DULL, "288048106.ogg"},
	{0xFBAA1C6DB359F205ULL, "402939840.ogg"},
	{0xFBAA4EC41BFA2AB1ULL, "ave200_aveline_bar6116406.cl"},
	{0xFBAA54629D17E178ULL, "6143019_m.fxe"},
	{0xFBAA7295DE4BB348ULL, "zz_mag_debug6039835.cl"},
	{0xFBAB0F36633CF9B9ULL, "6143379_m.fxe"},
	{0xFBAB4E60A7AB9570ULL, "638924662.ogg"},
	{0xFBAC13CAC5701E6CULL, "132694517.ogg"},
	{0xFBAC4EAB9857EDC3ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_bdy_lgta_0n.dds"},
	{0xFBAC851053BA268EULL, "mcr371cr_aveline.utc"},
	{0xFBAC88C83893C256ULL, "art\\levels\\races\\proxy\\props\\_textures\\p_armoire_01_n.dds"},
	{0xFBACA37BCD985E56ULL, "dae250ip_at_scaverns_a.utp"},
	{0xFBACBBADE24D24A1ULL, "6126723_m.fxe"},
	{0xFBACC92E4429D23EULL, "hlac_ht_mansion2b_818_64x.dds"},
	{0xFBAD49CDE8FE8622ULL, "lowtown_male_beggar_sta6151054.cl"},
	{0xFBAD682E88D80742ULL, "138143879.ogg"},
	{0xFBAD78926B87C900ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_arm_medj_0n.dds"},
	{0xFBAD814AD9B024A0ULL, "vxca_caverns2prc_8kb_0_0.vlm"},
	{0xFBADD443A03EB275ULL, "vxht_playershouse_n_1wb_0_0.vlm"},
	{0xFBAE284E3CF56682ULL, "370639774.ogg"},
	{0xFBAE57DB2B3C9919ULL, "576785113.ogg"},
	{0xFBAEB931E05603C8ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_rob_encn_0s.dds"},
	{0xFBAEDC3EF2725B49ULL, "hlac_dr_primevaltomb_659_7z.dds"},
	{0xFBAF2C5A9C761366ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_sledge_n.dds"},
	{0xFBAFB8F4129AE108ULL, "vxht_keep_3hf_0_0.vlm"},
	{0xFBB01A5D2CCA75A3ULL, "626091872.ogg"},
	{0xFBB09C4A83B54816ULL, "mag130_danzig6018817.cl"},
	{0xFBB0FE27C36AC356ULL, "sse_ss_pride_demon.gda"},
	{0xFBB155F12BF1E58FULL, "6163696_m.fxe"},
	{0xFBB1AD4007413EA5ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\cassandra\\cassandra_1.ani"},
	{0xFBB1D79F582F4AC3ULL, "default_cubemap.dds"},
	{0xFBB1E40FD5CD1000ULL, "6136121_m.fxe"},
	{0xFBB2322D36897797ULL, "vxlt_hangedman_1vk_0_0.vlm"},
	{0xFBB2475E3F5C6956ULL, "vxlt_0_lowtown_docks_9k6_0_0.vlm"},
	{0xFBB2DB2774DB32B2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_walk.gad"},
	{0xFBB2F55FE455DF31ULL, "6119772_m.fxe"},
	{0xFBB2FC9E41027CF4ULL, "hlac_lt_0_lowtown_docks_a3_-196608_0z.dds"},
	{0xFBB3321959BD445FULL, "6097869_m.fxe"},
	{0xFBB3666B96230623ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06a.mmh"},
	{0xFBB3746B96231E7DULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06a.msh"},
	{0xFBB38FC7A4CDD4D8ULL, "zz_swordogre.cut"},
	{0xFBB3B459E2E3197DULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_basic.mao"},
	{0xFBB3B859E2E31FB6ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_basic.mmh"},
	{0xFBB3C259E2E330B4ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wi_basic.msh"},
	{0xFBB3D7C27A065F63ULL, "isa111_isabela6067130.cl"},
	{0xFBB3EFD88EDFE672ULL, "vxw_woundedcoast_1hv_0_0.vlm"},
	{0xFBB4778C002A106EULL, "hlac_ca_caverns1_24426_10x.dds"},
	{0xFBB4A6D49561C1CFULL, "6033412_m.fxe"},
	{0xFBB4BE1F61933B57ULL, "mag101_alain.bnk"},
	{0xFBB4D923672EA8EBULL, "479276686.ogg"},
	{0xFBB4F1B85B055EC7ULL, "9229940.ogg"},
	{0xFBB4F397FC553C44ULL, "6101471_m.fxe"},
	{0xFBB5B902388129C6ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\set_visdoortrim_n.dds"},
	{0xFBB5CE0DE21CEA01ULL, "665112291.ogg"},
	{0xFBB6B2F9D1477115ULL, "mcr310_argument6074521.cl"},
	{0xFBB6F7FC1B30D0FEULL, "6095228_m.fxe"},
	{0xFBB747C414E94CAFULL, "vxw_sundermount_28t_0_0.vlm"},
	{0xFBB77276C1AEE15BULL, "vxht_0_hightown_n_bcz_0_0.vlm"},
	{0xFBB77B74F8156F92ULL, "vxht_mansion2b_2cl_0_0.vlm"},
	{0xFBB7994254B93228ULL, "vxht_playershouse_n_1j8_0_0.vlm"},
	{0xFBB79EC80BFB3475ULL, "89743811.ogg"},
	{0xFBB7A01F61953E6FULL, "mag101_alain.cnv"},
	{0xFBB8384FB7350B0DULL, "follower_banter6172956.cl"},
	{0xFBB83FEE2B64249FULL, "371992387.ogg"},
	{0xFBB8A5A1506258DFULL, "vxht_mansion_n_59g_0_0.vlm"},
	{0xFBB8AAE2D0E8A401ULL, "lowtown_argument6169438.cl"},
	{0xFBB918C87664B26BULL, "319373764.ogg"},
	{0xFBB922D8AA8982BBULL, "vxw_bonepitdestroyed_a3_1wy_0_0.vlm"},
	{0xFBB999B2925C09AAULL, "dae000_gamlen6149409.cl"},
	{0xFBB9B18FA76ADAEEULL, "gen00fl_fenris6103321.cl"},
	{0xFBB9DAD507F51BCDULL, "vxlt_gallowstemplar_f_43i_0_0.vlm"},
	{0xFBB9DE2774E156E9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_walk.evt"},
	{0xFBB9E58D2E7C7AC9ULL, "w_bonepitmining_a2_185_0.mao"},
	{0xFBB9EE6E49CFFCF1ULL, "gen00fl_dog6117766.cl"},
	{0xFBB9F18D2E7C8F62ULL, "w_bonepitmining_a2_185_0.mmh"},
	{0xFBB9F38D2E7C9218ULL, "w_bonepitmining_a2_185_0.msh"},
	{0xFBBA2C2B888DBA16ULL, "79003878.ogg"},
	{0xFBBA83F44CE2D80BULL, "vxca_cavern1_haunted_j1z_0_0.vlm"},
	{0xFBBAC9D6670A83C9ULL, "art\\vfx\\combat\\vfx_m_ent_horrow_imp_c.mmh"},
	{0xFBBB2D8AD4433D21ULL, "6123423_m.fxe"},
	{0xFBBB3FF5195107AEULL, "wm_ico_bonepit_roll.dds"},
	{0xFBBB622A2E67841BULL, "t3_dog_or2.tnt"},
	{0xFBBB87E6FD996EABULL, "66920646.ogg"},
	{0xFBBBF327CBA63D20ULL, "vxca_caverns2prc_8lc_0_0.vlm"},
	{0xFBBBFDCE711CB45EULL, "montage_subs_demohack6217099.cl"},
	{0xFBBC0C27D41C0F55ULL, "6126940_f.fxe"},
	{0xFBBC134D2ED3E812ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_smoke01.mao"},
	{0xFBBC1924A4B6459CULL, "vxdr_primevalentrance_233_0_0.vlm"},
	{0xFBBC4449A19040C0ULL, "hlac_lt_gallowstemplar_f_7198_4z.dds"},
	{0xFBBC67768FD1C0ACULL, "6004160_f.fxe"},
	{0xFBBC6B58EA61877CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_loop.ani"},
	{0xFBBCA62416637ED2ULL, "6148980_m.fxe"},
	{0xFBBD0E83C396EEE5ULL, "vxlt_warehouse_n_x3_0_0.vlm"},
	{0xFBBDC71A8D45133FULL, "102609427.ogg"},
	{0xFBBDC977E5395D4BULL, "vxht_0_hightown_a2_2es_0_0.vlm"},
	{0xFBBDCF156511E979ULL, "vxlt_warehouse_13b_0_0.vlm"},
	{0xFBBDEFA378204C4CULL, "456264229.ogg"},
	{0xFBBE9DA81D04E57CULL, "vxlt_undercitydungeon_22y_0_0.vlm"},
	{0xFBBEC9CE47055593ULL, "82787634.ogg"},
	{0xFBBEF85C48BA45D6ULL, "hlac_ca_caverns1_24422_5z.dds"},
	{0xFBBF3A57AA7DBC5BULL, "719440091.ogg"},
	{0xFBBF3B68F2295ECEULL, "vxlt_ruinedhouse_n_ii_0_0.vlm"},
	{0xFBBF74FDC503C6CEULL, "vxht_0_hightown_a3_9ze_0_0.vlm"},
	{0xFBC0155F7E18F792ULL, "hlac_lt_0_lowtown_a3_n_c_65538_0x.dds"},
	{0xFBC08D2461176D9DULL, "519044356.ogg"},
	{0xFBC0C1F88488A846ULL, "6041049_m.fxe"},
	{0xFBC0C3A5220E34C9ULL, "dae000ip_crowd_mf4_talk9.utp"},
	{0xFBC1661625844073ULL, "278117708.ogg"},
	{0xFBC185F50F2A61BBULL, "vxht_keep_n_4hq_0_0.vlm"},
	{0xFBC1D24FA2BB6A81ULL, "vxw_sundermount_1zk_0_0.vlm"},
	{0xFBC21465D19CAB97ULL, "779762135.ogg"},
	{0xFBC22AD7F45B7979ULL, "static_static_p11.psh.fxo.dx11"},
	{0xFBC24DAE801208A6ULL, "hlac_w_bonepitmining_a2_458753_0x.dds"},
	{0xFBC261B580D5FEF2ULL, "vxw_woundedcoast_1sw_0_0.vlm"},
	{0xFBC2F3F1880A1D31ULL, "6129986_m.fxe"},
	{0xFBC2FC59DC521C6BULL, "close.cur"},
	{0xFBC370E2EA7E1CBCULL, "zz_followers_banter6158412.cl"},
	{0xFBC37858EA67BBBEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\threaten_reaction_loop.gad"},
	{0xFBC38BF5C28CB862ULL, "w_bonepitdestroyed_a3_14_0.rml"},
	{0xFBC420053E333E19ULL, "vxca_cave_1dm_0_0.vlm"},
	{0xFBC4569D659DDCF0ULL, "mcr310_argument6075129.cl"},
	{0xFBC4676B96317B48ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_eye06a.phy"},
	{0xFBC4C43E4C7E5085ULL, "vxht_mansion2b_1a6_0_0.vlm"},
	{0xFBC4D35CED36D994ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.impa.ani"},
	{0xFBC51FDB0BC41AE9ULL, "729326101.ogg"},
	{0xFBC54DB8BA37AD00ULL, "193663790.ogg"},
	{0xFBC5DA01F9638246ULL, "310019916.ogg"},
	{0xFBC5E4B6E08A8F7EULL, "vxlt_0_lowtown_a2_n_ehy_0_0.vlm"},
	{0xFBC5F05AC3559C98ULL, "hlac_lt_darktown_-1_28z.dds"},
	{0xFBC656B8097DA96CULL, "vxlt_0_lowtown_docks_n_c08_0_0.vlm"},
	{0xFBC6B4364C2828FCULL, "ht_0_hightown_n_20_0_lc.mmh"},
	{0xFBC6C4CA3F572562ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_to_back.gad"},
	{0xFBC6C62774EBB858ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\enter_walk.ani"},
	{0xFBC6F0D3C2C75EABULL, "zz_followers_banter6158574.cl"},
	{0xFBC70C0B8B1808C4ULL, "llac_ca_caverns1_24425_0.dds"},
	{0xFBC713D12AA76BD8ULL, "mag210_arianni6021220.cl"},
	{0xFBC7360EA04DAD49ULL, "follower_banter6160649.cl"},
	{0xFBC75579C533F35FULL, "6167677_f.fxe"},
	{0xFBC79ABA08E75BF5ULL, "896063902.ogg"},
	{0xFBC79E2ED5D57EE2ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_dsaxa_0d.dds"},
	{0xFBC816D669912F5DULL, "6134910_f.fxe"},
	{0xFBC820D34110002CULL, "759909005.ogg"},
	{0xFBC871C83AEC1B27ULL, "art\\characters\\creatures\\dragon\\animation\\rst_lp.gad"},
	{0xFBC8CD31F7939C8EULL, "hlac_w_bonepit_393219_0y.dds"},
	{0xFBC8DAFDC74F03A2ULL, "vxlt_house_gc_0_0.vlm"},
	{0xFBC90764EABE5AA7ULL, "dae000_merchant_generic_hm6172635.cl"},
	{0xFBC99C4EA33F1B89ULL, "6103748_m.fxe"},
	{0xFBC99F67C63955A1ULL, "6128523_m.fxe"},
	{0xFBCA032A24E44121ULL, "965495878.ogg"},
	{0xFBCA15FEBD5E44A8ULL, "vxw_bonepitmining_a2_1pu_0_0.vlm"},
	{0xFBCA1B9DF2C0DAE2ULL, "mrl150_merrill6077670.cl"},
	{0xFBCAA6C6808FA5BCULL, "docks_scene_dockworkers6176507.cl"},
	{0xFBCAED8041B20EC4ULL, "w_bonepitmining_a2_50_0.rml"},
	{0xFBCAFCAF7D2DC7F1ULL, "art\\characters\\weapons\\staff\\w_stf_sf02a.mmh"},
	{0xFBCB08AF7D2DDB92ULL, "art\\characters\\weapons\\staff\\w_stf_sf02a.mao"},
	{0xFBCB16AF7D2DF3AFULL, "art\\characters\\weapons\\staff\\w_stf_sf02a.msh"},
	{0xFBCB1F3852D2E7FFULL, "art\\characters\\weapons\\staff\\textures\\w_stf_014a_0d.dds"},
	{0xFBCB87E27D13E05BULL, "24042427.ogg"},
	{0xFBCB9963763708A9ULL, "hlac_ca_caverns1_mines_24426_8y.dds"},
	{0xFBCBD05CED3CF3F6ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.impa.gad"},
	{0xFBCC127D9833DF00ULL, "w_sundermount_138_0.phy"},
	{0xFBCC366C86FCA9B9ULL, "131945097.ogg"},
	{0xFBCCE7CA3F5BCC59ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_to_back.evt"},
	{0xFBCCFB98456A9AB5ULL, "6211926_m.fxe"},
	{0xFBCD27DDBFA92108ULL, "dae300_cullen6135792.cl"},
	{0xFBCD55D2292F29DCULL, "502036356.ogg"},
	{0xFBCD838407921F89ULL, "art\\levels\\atmosphere\\sky\\sb03_front.dds"},
	{0xFBCDE6DAB6F9DF4EULL, "vxht_0_hightown_a3_9wf_0_0.vlm"},
	{0xFBCE9B1FD1F554EAULL, "plt_rdr100pt_keepers.plo"},
	{0xFBCF15E71A75A897ULL, "vxlt_hangedman_1xe_0_0.vlm"},
	{0xFBCF77E12D88FD9BULL, "338913059.ogg"},
	{0xFBCF8DF34AB2465DULL, "6126323_m.fxe"},
	{0xFBCFD10159E929EFULL, "6207989_m.fxe"},
	{0xFBCFDC729412C387ULL, "vxlt_hangedman_n_27d_0_0.vlm"},
	{0xFBCFEBF61C64AFACULL, "fen110_fenris6117969.cl"},
	{0xFBD054E17AD2FF5CULL, "hlac_w_woundedcoast_131079_0x.dds"},
	{0xFBD127D9DEAE8C87ULL, "art\\vfx\\combat\\vfx_w_2nd_charge_c.impa.gad"},
	{0xFBD13C37912D0DBBULL, "vxlt_0_lowtown_a3_n_c_en1_0_0.vlm"},
	{0xFBD1CC06D8399F05ULL, "gen00fl_carver6108886.cl"},
	{0xFBD1E596DA45BC93ULL, "hlac_lt_backalley_n_0_7y.dds"},
	{0xFBD2EB5CED434055ULL, "art\\vfx\\combat\\vfx_glb_com_stagger_s1_c.impa.evt"},
	{0xFBD3118B502E7D46ULL, "6034001_m.fxe"},
	{0xFBD3317D983A32B8ULL, "w_sundermount_138_0.rml"},
	{0xFBD35A6318AE2C4FULL, "6101278_m.fxe"},
	{0xFBD3A8998AB541FAULL, "vxht_0_hightown_9am_0_0.vlm"},
	{0xFBD3C37CADE97775ULL, "119178569.ogg"},
	{0xFBD3F293A054CF7DULL, "vxw_bonepitmining_a2_76_0_0.vlm"},
	{0xFBD45693D280A7FEULL, "vxlt_0_lowtown_a2_n_emz_0_0.vlm"},
	{0xFBD475F5C29B05B7ULL, "w_bonepitdestroyed_a3_14_0.msh"},
	{0xFBD487F5C29B241AULL, "w_bonepitdestroyed_a3_14_0.mao"},
	{0xFBD48BF5C29B2B29ULL, "w_bonepitdestroyed_a3_14_0.mmh"},
	{0xFBD501B90FB4838CULL, "vxlt_gallowstemplar_37d_0_0.vlm"},
	{0xFBD516194B50ACF6ULL, "6125992_m.fxe"},
	{0xFBD52E1F1CB95ACAULL, "hlac_lt_0_lowtown_docks_-131075_3y.dds"},
	{0xFBD597EE0132CCFBULL, "hlac_lt_0_lowtown_a2_n_c_0_7x.dds"},
	{0xFBD5B3FAE17A0A17ULL, "6080814_m.fxe"},
	{0xFBD5DC8AAAB24DA3ULL, "ch2armtnt_ds1.psh.fxo"},
	{0xFBD5E5B6B74F8C6FULL, "hlac_lt_undercityserialkiller_-1_37y.dds"},
	{0xFBD64D3778C4069FULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfirelrg.dura.gad"},
	{0xFBD6700AF73C1399ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stunned.ani"},
	{0xFBD6AEA354D8DDF0ULL, "6208557_f.fxe"},
	{0xFBD6BA79C0C808F6ULL, "vxw_bonepit_1x1_0_0.vlm"},
	{0xFBD6F0940C1AD416ULL, "6036484_f.fxe"},
	{0xFBD707170DA827F9ULL, "hlac_w_sundermount_-131073_0x.dds"},
	{0xFBD714A23DED30D1ULL, "fex_bd_a_md_wl.mmh"},
	{0xFBD72EA23DED5C8FULL, "fex_bd_a_md_wl.msh"},
	{0xFBD73408E479BEC9ULL, "40879397.ogg"},
	{0xFBD79AFE1CD5EF25ULL, "vxht_0_hightown_a2_c3j_0_0.vlm"},
	{0xFBD81CD9DEB49894ULL, "art\\vfx\\combat\\vfx_w_2nd_charge_c.impa.evt"},
	{0xFBD823141C2EF36BULL, "6166361_m.fxe"},
	{0xFBD83F47AA1E82B0ULL, "ht_0_hightown_n_36_0_ll.mmh"},
	{0xFBD8427E1BBB3D7DULL, "6125681_m.fxe"},
	{0xFBD885092488B61CULL, "hlac_lt_0_lowtown_docks_-1_1y.dds"},
	{0xFBD8EA885A8AEFB0ULL, "vxca_cavern1_haunted_jgo_0_0.vlm"},
	{0xFBD98391662A1A75ULL, "6062809_m.fxe"},
	{0xFBD98F06649ADC57ULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3.phy"},
	{0xFBD9B530E5C777D3ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjut.phy"},
	{0xFBDA5418F721206CULL, "707891352.ogg"},
	{0xFBDA69BCE9C5FB33ULL, "hlac_lt_house_-1_2y.dds"},
	{0xFBDA7680F11656B8ULL, "6127537_f.fxe"},
	{0xFBDA98C3B13A4210ULL, "549809296.ogg"},
	{0xFBDA9CC314D2C0FFULL, "one220_yevhen6135789.cl"},
	{0xFBDA9F4C61D9D9D2ULL, "gen00fl_carver6104601.cl"},
	{0xFBDAEFCA3F681748ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\sitting_forward_to_back.ani"},
	{0xFBDAFB2AE10AC034ULL, "6033965_m.fxe"},
	{0xFBDB67067408AF20ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_post.phy"},
	{0xFBDB820218439449ULL, "vxlt_darktown_2v3_0_0.vlm"},
	{0xFBDC0893FF60FC5CULL, "163003801.ogg"},
	{0xFBDC2251FF6C2456ULL, "vxdr_primevalentrance_28d_0_0.vlm"},
	{0xFBDC279234459AB5ULL, "6135198_m.fxe"},
	{0xFBDC94D170800137ULL, "649011550.ogg"},
	{0xFBDD423778CA128CULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfirelrg.dura.evt"},
	{0xFBDD4B0AF741F3DBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stunned.gad"},
	{0xFBDD5EC83AFE576DULL, "art\\characters\\creatures\\dragon\\animation\\rst_lp.ani"},
	{0xFBDE17C3A0262D8AULL, "663789959.ogg"},
	{0xFBDE7E0354BDA6C3ULL, "vxdr_primevaltomb_559_0_0.vlm"},
	{0xFBDE8D1D9D65BCACULL, "6187918_m.fxe"},
	{0xFBDEC48AFBE3DFA8ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magb.mmh"},
	{0xFBDED08AFBE3F44BULL, "art\\characters\\playercharacter\\humanfemale\\glv_magb.mao"},
	{0xFBDEE28AFBE412D2ULL, "art\\characters\\playercharacter\\humanfemale\\glv_magb.msh"},
	{0xFBDF1D8D9464A95DULL, "304676508.ogg"},
	{0xFBDF49A51953B3C5ULL, "lgt110cr_arl_templar_accost.utc"},
	{0xFBDF4C0F040FF5DEULL, "hlac_lt_0_lowtown_a2_n_c_65536_1z.dds"},
	{0xFBDF6962138E9FACULL, "hlac_dr_primevaltomb_661_0y.dds"},
	{0xFBDFA2F8AF8017C5ULL, "318099010.ogg"},
	{0xFBDFCFF80EADA11CULL, "vxht_0_hightown_a2_byd_0_0.vlm"},
	{0xFBE031DF55C2E5E3ULL, "6139127_m.fxe"},
	{0xFBE03B9E59A63D77ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_3rockpile.msh"},
	{0xFBE04D9E59A65BDAULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_3rockpile.mao"},
	{0xFBE0519E59A662E9ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\prp_3rockpile.mmh"},
	{0xFBE0537A8D6A90FDULL, "6041114_m.fxe"},
	{0xFBE0D541496281D8ULL, "118815347.ogg"},
	{0xFBE152B826DE3B76ULL, "vxca_caverns1_j21_0_2.vlm"},
	{0xFBE1BB4C552966B3ULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.anb"},
	{0xFBE1C3C9CA453CE4ULL, "6172899_m.fxe"},
	{0xFBE1FD058E691F32ULL, "vxlt_gallowscourtyard_6xl_0_0.vlm"},
	{0xFBE24BDD9D536D9CULL, "vxlt_gallowstemplar_2iw_0_0.vlm"},
	{0xFBE28FDDFF892731ULL, "128717933.ogg"},
	{0xFBE3135BB15EACA9ULL, "6090448_m.fxe"},
	{0xFBE369F50A24EE96ULL, "vxht_0_hightown_c17_0_0.vlm"},
	{0xFBE3C6D6672D83FEULL, "art\\vfx\\combat\\vfx_m_ent_horrow_imp_c.anb"},
	{0xFBE424E55A73E2D7ULL, "328145154.ogg"},
	{0xFBE4680AF74843C0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\rct_stunned.evt"},
	{0xFBE4786F806BCC3BULL, "vxw_bonepitdestroyed_a3_1fz_0_0.vlm"},
	{0xFBE4B0531949EA6AULL, "165247414.ogg"},
	{0xFBE4B5E80DD89E57ULL, "6091715_m.fxe"},
	{0xFBE4EA9BA8848377ULL, "vxht_keep_n_4mj_0_0.vlm"},
	{0xFBE4F25A31DC55D0ULL, "follower_banter6163207.cl"},
	{0xFBE5248FF3220DFEULL, "gen00fl_isabela6090810.cl"},
	{0xFBE543E30198EA0CULL, "6057357_f.fxe"},
	{0xFBE5539866C784BFULL, "w_woundedcoast_95_0.msh"},
	{0xFBE5599866C78F61ULL, "w_woundedcoast_95_0.mmh"},
	{0xFBE55E4160744315ULL, "hlac_w_woundedcoast_196612_0y.dds"},
	{0xFBE5659866C7A302ULL, "w_woundedcoast_95_0.mao"},
	{0xFBE5D570FDD05D2FULL, "llac_lt_0_lowtown_docks_-65538_7.dds"},
	{0xFBE602117950FE9CULL, "vxlt_hangedman_2en_0_0.vlm"},
	{0xFBE614D9DEC0C8CDULL, "art\\vfx\\combat\\vfx_w_2nd_charge_c.impa.ani"},
	{0xFBE649D206770E58ULL, "vxw_encounter3_fh_0_1.vlm"},
	{0xFBE6709700490A6FULL, "70234566.ogg"},
	{0xFBE6A89B2C0D0EC5ULL, "dae300_bran6143480.cl"},
	{0xFBE6D66E2C8829A2ULL, "6167693_m.fxe"},
	{0xFBE6E918115F9E95ULL, "vxlt_darktown_4ju_0_0.vlm"},
	{0xFBE7230C66914C68ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_tool.mmh"},
	{0xFBE72F0C6691600BULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_tool.mao"},
	{0xFBE7410C66917E92ULL, "art\\levels\\races\\proxy\\props\\4tools\\vis_spike_tool.msh"},
	{0xFBE77E4E156587FBULL, "347175404.ogg"},
	{0xFBE82EDF0A3680DDULL, "gen00fl_aveline6101767.cl"},
	{0xFBE83C462DB2FB86ULL, "501809962.ogg"},
	{0xFBE86B187B40F2B3ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_lp.evt"},
	{0xFBE87FD535210034ULL, "rdr111dg_martin6054580.cl"},
	{0xFBE8BBDDAF8E2841ULL, "502505150.ogg"},
	{0xFBE8EA153B9AEE5DULL, "character_character_p275.psh.fxo.dx11.assembly"},
	{0xFBE900CD97B35BCFULL, "543929836.ogg"},
	{0xFBE938885C8839FCULL, "666685797.ogg"},
	{0xFBE98FF9F7C4024EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupb.mmh"},
	{0xFBE99BF9F7C416E5ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupb.mao"},
	{0xFBE9A9F9F7C42E3CULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_cupb.msh"},
	{0xFBE9AC711699FA3CULL, "montage_subs_demohack6217156.cl"},
	{0xFBE9E0D1A9B91C64ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd05.msh"},
	{0xFBE9F3DF0A39ACC9ULL, "vxlt_darktown_2y2_0_0.vlm"},
	{0xFBE9F6D1A9B94106ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd05.mmh"},
	{0xFBEA1CFD99EB77E2ULL, "6083472_m.fxe"},
	{0xFBEA7C2D95A1D0F9ULL, "58658443.ogg"},
	{0xFBEAD5D3B67FB992ULL, "6084503_m.fxe"},
	{0xFBEAD982FE5A9FF3ULL, "gen00fl_merrill_act1_pt1.stg"},
	{0xFBEAF95C35D9BCF7ULL, "vxw_bonepitdestroyed_a3_1wm_0_0.vlm"},
	{0xFBEB2A3778D627E5ULL, "art\\vfx\\environment\\level_fx\\fire\\vfx_campfirelrg.dura.ani"},
	{0xFBEB52B2F7043C07ULL, "vxht_bartrand_n_4dn_0_0.vlm"},
	{0xFBEB7A401BC6C7F3ULL, "qcr000_arishok6057090.cl"},
	{0xFBEC122B081659D8ULL, "vxlt_0_lowtown_euo_0_0.vlm"},
	{0xFBEC1588ACC44723ULL, "6112960_m.fxe"},
	{0xFBEC71AD2E9E775DULL, "6128123_m.fxe"},
	{0xFBEC9D9D0B603623ULL, "blood_imp_10.mao"},
	{0xFBED11CEFAAC8F65ULL, "vxlt_darktown_4e8_0_0.vlm"},
	{0xFBED51E762E3EC57ULL, "6175017_m.fxe"},
	{0xFBED6A64C5E73954ULL, "596439141.ogg"},
	{0xFBED6FB4B09BA73FULL, "dae000_keep_noblewoman_sta6162068.cl"},
	{0xFBEDCF9BE76DCE6AULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv2.gad"},
	{0xFBEE2463BDDFC4FFULL, "6098833_m.fxe"},
	{0xFBEED84C5534221CULL, "art\\vfx\\combat\\vfx_m_cre_heroaura_loop_c.mmh"},
	{0xFBEF0DFBC113601AULL, "lt_0_lowtown_docks_n_21_0_lc.mmh"},
	{0xFBEF64187B4705F4ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_lp.gad"},
	{0xFBEFA1717F9E84A3ULL, "t3_wdg_tn3.tnt"},
	{0xFBEFB49AF44D0823ULL, "vxlt_gallowstemplar_5ca_0_0.vlm"},
	{0xFBF02BFEE02C590EULL, "dae000_bartrand6031119.cl"},
	{0xFBF035A6BC3B5564ULL, "6175999_m.fxe"},
	{0xFBF10E5C0DF1CBBFULL, "vxlt_0_lowtown_docks_n_c07_0_0.vlm"},
	{0xFBF12F7D98532C95ULL, "w_sundermount_138_0.msh"},
	{0xFBF1317D9853300BULL, "w_sundermount_138_0.mmh"},
	{0xFBF13D7D985344A8ULL, "w_sundermount_138_0.mao"},
	{0xFBF15B0B44C06AB3ULL, "vxht_0_hightown_a3_9xb_0_0.vlm"},
	{0xFBF1DBD1FC1B0716ULL, "vxht_0_hightown_c48_0_0.vlm"},
	{0xFBF309D0E91CDA15ULL, "6179770_m.fxe"},
	{0xFBF30F67FD888F94ULL, "6085911_m.fxe"},
	{0xFBF35C789A7A9BF7ULL, "vxht_keep_n_4he_0_0.vlm"},
	{0xFBF3EDD1A9C186D1ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hd05.phy"},
	{0xFBF4499435DF8AC5ULL, "620548583.ogg"},
	{0xFBF46C30CF96BC28ULL, "6113032_m.fxe"},
	{0xFBF4B29BE773BB01ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv2.evt"},
	{0xFBF4D587AD05D509ULL, "vxlt_gallowstemplar_3sq_0_0.vlm"},
	{0xFBF50E9F6E7CE353ULL, "vxw_bonepitdestroyed_a3_83_0_0.vlm"},
	{0xFBF549CF3727076CULL, "659294410.ogg"},
	{0xFBF62C5853AE847CULL, "832871241.ogg"},
	{0xFBF673187B4D3DD2ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\mage\\mage_ability_immunity_lp.ani"},
	{0xFBF684AE7B2EC873ULL, "art\\vfx\\combat\\vfx_glb_ele_p_major_c.anb"},
	{0xFBF693E850939460ULL, "follower_banter6172858.cl"},
	{0xFBF6BDE08F423EDDULL, "gen00fl_sebastian6188700.cl"},
	{0xFBF6F8560FF78FD4ULL, "character_character_p10.psh.fxo"},
	{0xFBF782A31022A503ULL, "art\\vfx\\creatures\\fxc_lotf_c.impa.evt"},
	{0xFBF7AE57CB28BAE4ULL, "gen00fl_carver6103097.cl"},
	{0xFBF7B7CC2D743138ULL, "hlac_ca_caverns1_mines_24424_13x.dds"},
	{0xFBF7F60CDF49FF0BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_masc_0s.dds"},
	{0xFBF829BBA1FA8C59ULL, "741146558.ogg"},
	{0xFBF8BF9839E7A364ULL, "lt_0_lowtown_a2_n_c_32_0.anb"},
	{0xFBF8DDF29A3362E2ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_basic.msh"},
	{0xFBF8DFF29A336598ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_basic.mmh"},
	{0xFBF8EBF29A337A3BULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_basic.mao"},
	{0xFBF93AE6EC1A7AEBULL, "vxlt_0_lowtown_a2_n_c_eu1_0_0.vlm"},
	{0xFBF9C6FC8D853271ULL, "510775730.ogg"},
	{0xFBFA0C61280E75AEULL, "vxlt_0_lowtown_docks_n_cs3_0_0.vlm"},
	{0xFBFAAB0AD0C1514BULL, "gen00fl_aveline6113674.cl"},
	{0xFBFB2AC99ABFA676ULL, "static_static_p29.psh.fxo"},
	{0xFBFB4BCE96118BF6ULL, "hslmatrix.mao"},
	{0xFBFBB7B6E235B5B4ULL, "404335801.ogg"},
	{0xFBFBE3B6467BC956ULL, "vxht_0_hightown_n_a7h_0_0.vlm"},
	{0xFBFC52CEA8439013ULL, "6032473_m.fxe"},
	{0xFBFCA94C9AC1AEDFULL, "llac_lt_0_lowtown_a2_n_c_65535_7.dds"},
	{0xFBFD1D630CE1F7D3ULL, "dae000ip_crowd_ff8_talk13.utp"},
	{0xFBFD3B7FDE128917ULL, "6050328_f.fxe"},
	{0xFBFD8F48AB893DCBULL, "ser221_gascard6148858.cl"},
	{0xFBFDA41FD21CD4E1ULL, "plt_rdr100pt_keepers.nss"},
	{0xFBFDC1460776463BULL, "901949256.ogg"},
	{0xFBFE342E7C0D16DDULL, "6147410_m.fxe"},
	{0xFBFE36F83E035ED0ULL, "6085835_f.fxe"},
	{0xFBFE5BA310288184ULL, "art\\vfx\\creatures\\fxc_lotf_c.impa.gad"},
	{0xFBFE8BB21456B285ULL, "vxht_mansion2b_1u2_0_0.vlm"},
	{0xFBFE8F552FD181D2ULL, "vxw_bonepitdestroyed_a3_86_0_0.vlm"},
	{0xFBFEBC8680FA13CDULL, "fen310_varania6118597.cl"},
	{0xFBFF47A08E80E9DFULL, "vxca_caverns2prc_8bj_0_0.vlm"},
	{0xFBFFFA47B21D032EULL, "520235564.ogg"},
	{0xFC002E3BD521B802ULL, "hla_ht_0_hightown_n_-131074_0z.dds"},
	{0xFC0084067427FCB5ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_post.msh"},
	{0xFC0086067428002BULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_post.mmh"},
	{0xFC009206742814C8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\4so_post.mao"},
	{0xFC0099EFAA8CDC43ULL, "vxw_bonepit_y4_0_0.vlm"},
	{0xFC009FB076E17410ULL, "211142126.ogg"},
	{0xFC0123CA10EC04C4ULL, "vxlt_0_lowtown_exh_0_0.vlm"},
	{0xFC014EA82A63DB44ULL, "vxht_mansion_n_3uv_0_0.vlm"},
	{0xFC019F96D3D44D23ULL, "665550365.ogg"},
	{0xFC01DCBF034B8804ULL, "cen_smk_01.mao"},
	{0xFC026E7195AB9327ULL, "dae000_bartrand6205975.cl"},
	{0xFC02767E2826550EULL, "hlac_lt_backalley_g_65537_6z.dds"},
	{0xFC02986D6D4F3019ULL, "gen_im_cns_cft_bmb_01.uti"},
	{0xFC02A455DC0F1B9AULL, "art\\vfx\\combat\\vfx_w_wns_front_close3_c.anb"},
	{0xFC02AA9BE77FEB50ULL, "art\\characters\\creatures\\corpse\\animation\\combat\\meleeattackv2.ani"},
	{0xFC02CE26C3CB7203ULL, "6218927_m.fxe"},
	{0xFC02E6F29A3BC623ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0wi_basic.phy"},
	{0xFC03511D50DC3480ULL, "vxlt_0_lowtown_a2_n_c_fee_0_0.vlm"},
	{0xFC03B04CEC875F94ULL, "gen00fl_anders6125659.cl"},
	{0xFC04248DEF385000ULL, "llac_ht_0_hightown_a2_131071_2.dds"},
	{0xFC048BBEC07ED150ULL, "follower_banter6172748.cl"},
	{0xFC04A1AE7B3B36DCULL, "art\\vfx\\combat\\vfx_glb_ele_p_major_c.mmh"},
	{0xFC04A511ACAFBEDBULL, "vxht_0_hightown_n_9p0_0_0.vlm"},
	{0xFC04CA67EB515202ULL, "vxw_sundermount_2f7_0_0.vlm"},
	{0xFC04DF3AC4495644ULL, "187572956.ogg"},
	{0xFC053D4BC2AB3123ULL, "6087451_m.fxe"},
	{0xFC056AA3102EBA62ULL, "art\\vfx\\creatures\\fxc_lotf_c.impa.ani"},
	{0xFC05CDCC235948CFULL, "vxlt_backalley_g_1vl_0_0.vlm"},
	{0xFC0605EC7A0D452CULL, "221496907.ogg"},
	{0xFC060D0F45803C91ULL, "gen00fl_anders6116166.cl"},
	{0xFC06D96887F77F13ULL, "540055484.ogg"},
	{0xFC06F8F05B59062FULL, "6095074_m.fxe"},
	{0xFC07014A69E6FC5BULL, "vxw_encounter3_7m_0_0.vlm"},
	{0xFC076F6B80D10C03ULL, "447665955.ogg"},
	{0xFC07EDA61E55F4ADULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef.mmh"},
	{0xFC07FEE5C04F973BULL, "mag360_thrask6043891.cl"},
	{0xFC07FFA61E561333ULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef.msh"},
	{0xFC082E941BF701F0ULL, "mag311_keran6041740.cl"},
	{0xFC089AC6101F594FULL, "6048511_m.fxe"},
	{0xFC08F2544C23A868ULL, "6103610_f.fxe"},
	{0xFC0920A6F6169CBFULL, "gen00fl_dog6172088.cl"},
	{0xFC098C9679AD5225ULL, "407818630.ogg"},
	{0xFC09BCF34055C6F9ULL, "6152571_m.fxe"},
	{0xFC0A1B2CD203EA7CULL, "art\\vfx\\combat\\vfx_gib_cora_rblt.gad"},
	{0xFC0A65F00C12DDC0ULL, "336503473.ogg"},
	{0xFC0B3E579938C0E8ULL, "gui_guivertexshader_p0.vsh.fxo"},
	{0xFC0B72CA722E76D8ULL, "gen00fl_dog6117804.cl"},
	{0xFC0B8CF9A2498844ULL, "llac_w_bonepit_458753_0.dds"},
	{0xFC0CDA25D9395A9BULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_boo_wic_0s.dds"},
	{0xFC0E71466FFEE401ULL, "hlac_lt_0_lowtown_docks_131072_0y.dds"},
	{0xFC0EFC6BDA0F8A28ULL, "zz_vault_debug6203286.cl"},
	{0xFC0F07AE9799ADCBULL, "lgt110_orz_harrowmont6108997.cl"},
	{0xFC0F66DFDFC83849ULL, "6042199_m.fxe"},
	{0xFC0FCE4C2A250E7CULL, "6140354_f.fxe"},
	{0xFC102F6143D32E1CULL, "439521958.ogg"},
	{0xFC1049463F58CFA4ULL, "vxht_0_hightown_c1u_0_0.vlm"},
	{0xFC10A96FC4A446DFULL, "hlac_w_bonepitdestroyed_a3_3_0z.dds"},
	{0xFC111A2CD20A0771ULL, "art\\vfx\\combat\\vfx_gib_cora_rblt.anb"},
	{0xFC111A2CD20A077AULL, "art\\vfx\\combat\\vfx_gib_cora_rblt.ani"},
	{0xFC111E8FA493AFEDULL, "gen_im_trs_krk_shd_met_r2.uti"},
	{0xFC111EB1463FEB30ULL, "743992163.ogg"},
	{0xFC114EAC2B836C01ULL, "hlac_lt_0_lowtown_docks_n_-3_5y.dds"},
	{0xFC1156C73FAB47D3ULL, "vxw_sundermount_37h_0_0.vlm"},
	{0xFC117BD2E2CD2113ULL, "art\\characters\\weapons\\shields\\textures\\w_shd_007a_n.dds"},
	{0xFC121137D8FF6F3DULL, "mer140_bonepit_comments6077812.cl"},
	{0xFC12C73E2831F2D5ULL, "vxlt_blackemporium_yq_0_0.vlm"},
	{0xFC133C44DD476A82ULL, "vxw_sundermount_2k0_0_0.vlm"},
	{0xFC135177CD4A6CB2ULL, "vxlt_0_lowtown_n_h1w_0_0.vlm"},
	{0xFC138F8D496FB98DULL, "gen00fl_merrill6108251.cl"},
	{0xFC139C9994CC9AADULL, "gen00fl_aveline6152576.cl"},
	{0xFC13F884CBE159DDULL, "218472359.ogg"},
	{0xFC1406B8E7DC5128ULL, "seb221_flora_harimann6177697.cl"},
	{0xFC141D39A1FC012DULL, "cry_val01.mao"},
	{0xFC1443447D123318ULL, "6084211_m.fxe"},
	{0xFC1501A7821B2D52ULL, "qcr270_fake_meredith6066325.cl"},
	{0xFC15787ACE77F385ULL, "6142303_m.fxe"},
	{0xFC164C939984E8ACULL, "vxca_caverns1_ixt_0_0.vlm"},
	{0xFC164EF30BBC0677ULL, "vxw_bonepitdestroyed_a3_1fn_0_0.vlm"},
	{0xFC16505C4CB447A4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlh_medb_0s.dds"},
	{0xFC16B53F0DE759B9ULL, "hlac_lt_backalley_n_0_6x.dds"},
	{0xFC17B210C5F0589DULL, "hlac_dr_dungeon_3217_8z.dds"},
	{0xFC17B8787B1BA7E9ULL, "6148749_m.fxe"},
	{0xFC1811AFBD99E334ULL, "pro000st_friedrich.stg"},
	{0xFC189E5247899A42ULL, "ch1_vl.psh.fxo"},
	{0xFC18E6A61E645C6AULL, "art\\characters\\playercharacter\\elffemale\\mt_bas_ef.phy"},
	{0xFC1924D6035B4829ULL, "58489887.ogg"},
	{0xFC1937C8994A94F9ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeexit.ani"},
	{0xFC19B73E0DB412F0ULL, "256566196.ogg"},
	{0xFC19BBA75C72D33CULL, "gen00fl_merrill6145141.cl"},
	{0xFC19C779FD1610EFULL, "hm_ser_gascard.mor"},
	{0xFC19F41137396DEBULL, "plt_mrl250pt_varterral.plo"},
	{0xFC1ABF243ADD318CULL, "vxht_0_hightown_a2_a6n_0_0.vlm"},
	{0xFC1ACA308D39150DULL, "mag371_elsa6041451.cl"},
	{0xFC1AD77D00FFE652ULL, "vxw_sundermount_37m_0_0.vlm"},
	{0xFC1B0A31CA2F51A3ULL, "vxlt_gallowstemplar_5rb_0_0.vlm"},
	{0xFC1C61883C78E3FAULL, "533902673.ogg"},
	{0xFC1C7A216481969EULL, "llac_w_encounter3_0_7.dds"},
	{0xFC1CBDD0BB60B4CCULL, "vxw_sundermount_30z_0_0.vlm"},
	{0xFC1D2E81D10E3443ULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.anb"},
	{0xFC1D4564757A93E3ULL, "dae000_arishok6068438.cl"},
	{0xFC1DA24792B677F8ULL, "one300_murder_followers6089901.cl"},
	{0xFC1DDDA2E93FD34AULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslba_0d.dds"},
	{0xFC1DEADEDF26E517ULL, "vxw_encounter2_camp_5l_0_0.vlm"},
	{0xFC1DEC48309913DEULL, "vxht_0_hightown_cao_0_0.vlm"},
	{0xFC1E8D3FE3B44748ULL, "6031303_m.fxe"},
	{0xFC1EBB23314EE824ULL, "vxht_0_hightown_c4z_0_0.vlm"},
	{0xFC1EE752BC7E9345ULL, "vxht_chantry_n_21b_0_0.vlm"},
	{0xFC1FC57D6A7E1AACULL, "66870504.ogg"},
	{0xFC1FC8A431A16053ULL, "vxw_sundermount_32k_0_0.vlm"},
	{0xFC1FDB0515242B3BULL, "vxlt_0_lowtown_docks_b0m_0_0.vlm"},
	{0xFC2012C89950743BULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeexit.gad"},
	{0xFC20B95C44BE675DULL, "hlac_ca_caverns1_24421_6y.dds"},
	{0xFC21700664D85D4CULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_door1_6x2_3.mmh"},
	{0xFC21AE21CF3D8302ULL, "vxw_sundermount_2l1_0_0.vlm"},
	{0xFC222D8F4B1C6CABULL, "follower_barks6153628.cl"},
	{0xFC226DDD4E05C8F9ULL, "6083541_m.fxe"},
	{0xFC22DEC1D25ED927ULL, "fort_crt_cyc.mao"},
	{0xFC230CAB2A887898ULL, "war_threaten.dds"},
	{0xFC233D46DA50032DULL, "128891844.ogg"},
	{0xFC23655AC8DCAD22ULL, "6108306_m.fxe"},
	{0xFC236CAFCFCE447EULL, "vxht_mansion2b_3ge_0_0.vlm"},
	{0xFC23779C60DB5A71ULL, "art\\characters\\playercharacter\\dwarfmale\\dm_bdy_meda.mao"},
	{0xFC23C1A913A76886ULL, "vxca_caverns1_mines_i2e_0_0.vlm"},
	{0xFC23E01C1AD2D5C3ULL, "hm_vincento.mor"},
	{0xFC240D51870CEE1AULL, "598497934.ogg"},
	{0xFC2439BD138771F0ULL, "zz_followers_banter6158355.cl"},
	{0xFC2455E7F4EB3367ULL, "161366484.ogg"},
	{0xFC24673E1CBA2C18ULL, "vxht_mansion2_27r_0_0.vlm"},
	{0xFC246DE3DD953F47ULL, "718680027.ogg"},
	{0xFC24B1EF68186C35ULL, "lgt310ip_at_bandit_alley.utp"},
	{0xFC24B9E7ECE0F443ULL, "vxlt_0_lowtown_n_f10_0_0.vlm"},
	{0xFC24BE708B7B012CULL, "vxca_caverns1_icu_0_0.vlm"},
	{0xFC24C0CFFDB21EF7ULL, "vxw_bonepitdestroyed_a3_1ki_0_0.vlm"},
	{0xFC24CEADDB49B116ULL, "gen_im_npc_hum_arm_hlm_chnb.uti"},
	{0xFC24DAC39034D7AFULL, "t3_wpn_drb.tnt"},
	{0xFC24DB4D5A00EF34ULL, "dae000_flemeth6000820.cl"},
	{0xFC24FBFB9DC5199FULL, "vxca_caverns2_a9s_0_0.vlm"},
	{0xFC251C91325BF63FULL, "pro000_colonel_ewald6040621.cl"},
	{0xFC2572BA8CC0D500ULL, "dae121ip_at_cht_hightown.utp"},
	{0xFC25C6D9093BFB44ULL, "dae000cr_grand_cleric6144095.cl"},
	{0xFC25DB146C65A01FULL, "395299929.ogg"},
	{0xFC2604F684488769ULL, "lt_gallowscourtyard_77_0.rml"},
	{0xFC261B87E61FC99EULL, "art\\characters\\weapons\\spears\\textures\\w_spr_002a_0n.dds"},
	{0xFC261C4AA05BD132ULL, "dre000_varric6207947.cl"},
	{0xFC26A6FF5A8F4D8FULL, "vxlt_hangedman_n_2ej_0_0.vlm"},
	{0xFC26B4CD776C96B3ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot.ani"},
	{0xFC272FC89956C420ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage1light\\chargeexit.evt"},
	{0xFC27498829803AAAULL, "hlac_lt_0_lowtown_docks_a3_-262144_1x.dds"},
	{0xFC279D8C5F78FE19ULL, "mag101_grace6034055.cl"},
	{0xFC27A0C536B6060EULL, "fx_c_fresnel_in_add.mao"},
	{0xFC288DB11E7A325EULL, "dae000cr_worthy.utc"},
	{0xFC28BAE55A7B54A0ULL, "lgt100cr_gang_2_docks.utc"},
	{0xFC292A301A3915AFULL, "dre000_bodahn6031364.cl"},
	{0xFC294959F2F5FED2ULL, "vxw_sundermount_32n_0_0.vlm"},
	{0xFC2A682B51309555ULL, "hlac_dr_0_deeproads_5845_9z.dds"},
	{0xFC2A8FA04DB94CF1ULL, "894152428.ogg"},
	{0xFC2AA7FDE42F9B5EULL, "vxht_0_hightown_a3_aqh_0_0.vlm"},
	{0xFC2B0B81D11A366CULL, "art\\vfx\\combat\\vfx_gui_mov_target_ret_p.mmh"},
	{0xFC2B93BC5888FA24ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\originals\\attackv1.gad"},
	{0xFC2BA755DC32264DULL, "art\\vfx\\combat\\vfx_w_wns_front_close3_c.mmh"},
	{0xFC2BB511B3BEA82DULL, "mcr370_negotiation6080171.cl"},
	{0xFC2BBECFA8EADECFULL, "6033603_m.fxe"},
	{0xFC2C51DE31FC3731ULL, "vxlt_warehouse_19d_0_0.vlm"},
	{0xFC2C7060723223DBULL, "vxca_caverns2prc_8bw_0_1.vlm"},
	{0xFC2C8B7DAF022461ULL, "one280st_golem.stg"},
	{0xFC2C8D502BC2DDB6ULL, "6016427_m.fxe"},
	{0xFC2D108C175BEDE2ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs05a_0t.dds"},
	{0xFC2D97AE396BF96FULL, "vxht_mansion2b_1kz_0_0.vlm"},
	{0xFC2DE6CF46A61E93ULL, "6117065_f.fxe"},
	{0xFC2E4CE2071A43BBULL, "vxlt_0_lowtown_docks_b7l_0_0.vlm"},
	{0xFC2ED706ACDE6306ULL, "vxht_mansion2b_3bv_0_0.vlm"},
	{0xFC2F3492C329F52FULL, "llac_lt_0_lowtown_docks_131068_0.dds"},
	{0xFC2F927D3B5EA564ULL, "40581446.ogg"},
	{0xFC3002C5F92AEA33ULL, "hlac_ht_0_hightown_a3_-65534_1z.dds"},
	{0xFC30426004CED9D5ULL, "hlac_w_bonepitmining_a2_327681_0z.dds"},
	{0xFC30B7E8B402E25BULL, "vxlt_0_lowtown_a2_ewj_0_0.vlm"},
	{0xFC30D1DA8725353DULL, "seb321_leliana.bnk"},
	{0xFC30FD55D8B48640ULL, "6023883_m.fxe"},
	{0xFC310D7B806A9E22ULL, "lowtown_fem_prostitute_sta6152330.cl"},
	{0xFC31679074346D1DULL, "gen00fl_fenris6105564.cl"},
	{0xFC3179977A4DB40CULL, "vxca_caverns2_6zf_0_0.vlm"},
	{0xFC320C96443614DBULL, "one110_tree_minder6133739.cl"},
	{0xFC3229AD69BB9CC9ULL, "zz_vault_debug6203201.cl"},
	{0xFC3265FEEAF4D29DULL, "149742990.ogg"},
	{0xFC32A2BC588F3202ULL, "art\\characters\\creatures\\rubblegolem\\animation\\stage2heavy\\originals\\attackv1.ani"},
	{0xFC32D91B0EB04498ULL, "vxht_mansion2_20s_0_0.vlm"},
	{0xFC3332ACEFA83777ULL, "vxw_bonepitdestroyed_a3_1lh_0_0.vlm"},
	{0xFC33373AB5158E60ULL, "vxlt_0_lowtown_a2_n_eoo_0_1.vlm"},
	{0xFC3355DFE02A8484ULL, "art\\levels\\races\\proxy\\props\\4tools\\_textures\\debug_d.dds"},
	{0xFC338BE03AF2257CULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_mag_robb_0n.dds"},
	{0xFC33AE86EE006D07ULL, "6095994_f.fxe"},
	{0xFC33B3DA87273845ULL, "seb321_leliana.cnv"},
	{0xFC33F45475C0F173ULL, "hlac_w_encounter2_0_0y.dds"},
	{0xFC34ACCD7778C6EAULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot.evt"},
	{0xFC357077760A27B9ULL, "hlac_w_encounter3_-196609_0z.dds"},
	{0xFC35AF6FB3D04413ULL, "6128508_f.fxe"},
	{0xFC35EE33C8EEF237ULL, "llac_ht_mansion2_818_54.dds"},
	{0xFC35FC04FC3E78F4ULL, "6074515_m.fxe"},
	{0xFC361A6E521A0C63ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.presseddown.ani"},
	{0xFC364DCEA32F1369ULL, "hlac_lt_0_lowtown_a2_1_1x.dds"},
	{0xFC374D34A54E7089ULL, "vxht_0_hightown_at9_0_0.vlm"},
	{0xFC37568A51AA3D09ULL, "vxca_caverns2_9a5_0_0.vlm"},
	{0xFC37915A9075D2DFULL, "374521227.ogg"},
	{0xFC37EDEBAE1B82A3ULL, "vxlt_gallowstemplar_5xd_0_0.vlm"},
	{0xFC37F76F6A942262ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden5.mmh"},
	{0xFC37F96F6A942518ULL, "art\\characters\\creatures\\crowds\\humanmale\\crd_ferelden5.msh"},
	{0xFC38DE983A1DBBC4ULL, "lt_0_lowtown_a2_n_c_32_0.rml"},
	{0xFC391F234EC7AD35ULL, "qun211_braggart6075705.cl"},
	{0xFC392F3F1B44024CULL, "466914240.ogg"},
	{0xFC3945D90C461F0DULL, "vxlt_0_lowtown_docks_a3_8pz_0_0.vlm"},
	{0xFC395E6EE3830F47ULL, "vxht_keep_3i6_0_0.vlm"},
	{0xFC39868997EE941BULL, "40835997.ogg"},
	{0xFC39B65135D81EB9ULL, "6106703_m.fxe"},
	{0xFC39C48AA37361D2ULL, "120576292.ogg"},
	{0xFC3A2B27666E6B52ULL, "hlac_ht_playershouse_n_855_30x.dds"},
	{0xFC3A5A62A79BF70EULL, "vxlt_0_lowtown_docks_a3_bqt_0_0.vlm"},
	{0xFC3A8C926110DB09ULL, "isa211_isabela6069416.cl"},
	{0xFC3AE0804545D5B9ULL, "6169797_m.fxe"},
	{0xFC3B056966E6C5F2ULL, "6036739_m.fxe"},
	{0xFC3B2446FE95913EULL, "ser121_jethann6164721.cl"},
	{0xFC3B3F0376B4EE2CULL, "crafting.gfx"},
	{0xFC3B4A41620DB705ULL, "zz_ser_debug6078579.cl"},
	{0xFC3B56B143D5FF8BULL, "49258551.ogg"},
	{0xFC3B617271408C79ULL, "vxlt_undercitydungeon_wy_0_0.vlm"},
	{0xFC3B8BCD777EAD5DULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_pinning_shot.gad"},
	{0xFC3B94AF262E84B0ULL, "rdr111dg_martin6015060.cl"},
	{0xFC3BD5FD11BE975CULL, "cod_mgc_chantry_founded.utp"},
	{0xFC3C2276451AC0E8ULL, "6171432_m.fxe"},
	{0xFC3CAE30E61BE948ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjut.mmh"},
	{0xFC3CBA30E61BFDEBULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjut.mao"},
	{0xFC3CBE8A91253C36ULL, "6041262_f.fxe"},
	{0xFC3CCC30E61C1C72ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\1ww_gallowjut.msh"},
	{0xFC3CCFF94EF5A72DULL, "zz_followers_banter6157887.cl"},
	{0xFC3D7805DCD4E648ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\til_hightownint_concrete_sa.dds"},
	{0xFC3E0679C207A7F9ULL, "dae000ip_crowd_mf2_talk12.utp"},
	{0xFC3E4002C2084E49ULL, "6117033_m.fxe"},
	{0xFC3E76E80C58E699ULL, "follower_banter6141899.cl"},
	{0xFC3E91DBB329B402ULL, "vxw_sundermount_2z3_0_0.vlm"},
	{0xFC3E948C1DCF0FEEULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_bdy_hvyo_s.dds"},
	{0xFC3EA3BAEA7EC84EULL, "961106720.ogg"},
	{0xFC3EF78FCB10F9BBULL, "gen_im_trs_bst_spd_fang_03.uti"},
	{0xFC3F0B9685E1E817ULL, "llac_lt_warehouse_-65536_0.dds"},
	{0xFC3F82775ED71C77ULL, "vxdr_primevaltomb_8el_0_0.vlm"},
	{0xFC4010FC4958EF1BULL, "vxlt_undercityserialkiller_142_0_0.vlm"},
	{0xFC40483CF6A91E96ULL, "gen_im_trs_rdr_med_hlm_r1.uti"},
	{0xFC407B744E1792B0ULL, "gen_im_wep_war_shd_act2_04.uti"},
	{0xFC409CBF803D4AEFULL, "23907253.ogg"},
	{0xFC40E3F49CDC6472ULL, "489221471.ogg"},
	{0xFC41152A6A4BEC64ULL, "qun100cr_qunari_warrior1.utc"},
	{0xFC411A1964E7AB5EULL, "60373603.ogg"},
	{0xFC41F80A7C7F4C79ULL, "408891568.ogg"},
	{0xFC420234C81FE4ADULL, "365161322.ogg"},
	{0xFC420D8D0297F086ULL, "45004852.ogg"},
	{0xFC4299A23C44E0CDULL, "pro000_merchant6051607.cl"},
	{0xFC42A45DACEC9DDEULL, "w_woundedcoast_45_0_lc.mmh"},
	{0xFC42ACF19E3430EDULL, "dae000_keep_servant_1_sta.bnk"},
	{0xFC43326E5224BF5AULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.presseddown.evt"},
	{0xFC4398AF1028FF17ULL, "vxca_caverns1_j2n_0_0.vlm"},
	{0xFC439D528C4B4291ULL, "gen_im_trs_gen_lgt_boo_r2.uti"},
	{0xFC44688C8727E1C6ULL, "649566334.ogg"},
	{0xFC449E3A5056DC06ULL, "vxlt_gallowstemplar_f_37i_0_0.vlm"},
	{0xFC44FCDF4781D9D1ULL, "face_face_p2.psh.fxo.dx11"},
	{0xFC4508D4B8458AAEULL, "atl_loginviewt_dxt5_dat.xml"},
	{0xFC452A6754483FE0ULL, "1062662461.ogg"},
	{0xFC45B1F38ABA7E70ULL, "6165406_m.fxe"},
	{0xFC45BB4D5621853CULL, "hlac_lt_0_lowtown_docks_131071_4y.dds"},
	{0xFC45CEF19E36A055ULL, "dae000_keep_servant_1_sta.cnv"},
	{0xFC460BE291E1C49CULL, "843641833.ogg"},
	{0xFC467D806F01AF3BULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst_lp.ani"},
	{0xFC46D238413CE214ULL, "vxlt_gallowscourtyard_n_7f2_0_0.vlm"},
	{0xFC46EAABC146CE1BULL, "art\\vfx\\combat\\vfx_c_rw_crust_c.dura.evt"},
	{0xFC47144D049F10ADULL, "6041325-pc_choice_three.cut"},
	{0xFC473F5862C16ED1ULL, "dae200_bran6142987.cl"},
	{0xFC481C9CFBA338D8ULL, "6077947_m.fxe"},
	{0xFC48A2898CD70FCFULL, "6039403_m.fxe"},
	{0xFC49065DBD7B3FA2ULL, "hlac_lt_0_lowtown_a3_n_c_65536_2x.dds"},
	{0xFC49348A79CA16C0ULL, "6083368_m.fxe"},
	{0xFC49359815E86831ULL, "vxlt_warehouse_13z_0_0.vlm"},
	{0xFC49499BF4240755ULL, "llac_dr_dungeon_3215_4.dds"},
	{0xFC49733D725FC327ULL, "182015473.ogg"},
	{0xFC4A316E522ADC4DULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\animation\\4co_trap_pressurestone.presseddown.gad"},
	{0xFC4A855C6B982618ULL, "hlac_ca_caverns2_5802_10z.dds"},
	{0xFC4A85A68AFCF883ULL, "vxlt_0_lowtown_a2_n_c_ded_0_0.vlm"},
	{0xFC4ABEBD5166C467ULL, "zz_mer_debug6017859.cl"},
	{0xFC4AD0D26A9310B6ULL, "67047426.ogg"},
	{0xFC4AF566AB73DA34ULL, "vxlt_0_lowtown_docks_a3_d58_0_0.vlm"},
	{0xFC4AF7DB78274DA5ULL, "ca_caverns1_mines_24421_ll.mmh"},
	{0xFC4B25595042859AULL, "vxca_caverns2prc_ac9_0_0.vlm"},
	{0xFC4B7A443C3A9EF0ULL, "gen_im_npc_arm_hum_cht_mrca.uti"},
	{0xFC4B825955C454F8ULL, "6179875_f.fxe"},
	{0xFC4C057EA5B187D3ULL, "6044920_m.fxe"},
	{0xFC4CB21B9F11B417ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masc_0n.dds"},
	{0xFC4CD883B39990B9ULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressureresetup.ani"},
	{0xFC4DE3ABC14CE15CULL, "art\\vfx\\combat\\vfx_c_rw_crust_c.dura.gad"},
	{0xFC4DEB52EFB3CF74ULL, "512559658.ogg"},
	{0xFC4E52DDF210A83BULL, "154323476.ogg"},
	{0xFC4E7300859B6E93ULL, "65571074.ogg"},
	{0xFC4EDCF602785886ULL, "art\\characters\\creatures\\spider\\animation\\kd_enter.evt"},
	{0xFC4F3A2AAFFD20C8ULL, "647474697.ogg"},
	{0xFC4F9E2C2D362194ULL, "llac_lt_0_lowtown_a3_n_c_131070_1.dds"},
	{0xFC4FC46E9CE62915ULL, "281959014.ogg"},
	{0xFC500F49D691D230ULL, "vxlt_backalley_g_1oh_0_0.vlm"},
	{0xFC50341C153153D7ULL, "zz_mer_debug6016971.cl"},
	{0xFC50420BEA895B10ULL, "vxht_chantry_n_334_0_0.vlm"},
	{0xFC5045C1D85FE1C8ULL, "hlac_lt_0_lowtown_docks_a3_-131075_5z.dds"},
	{0xFC5077B9217D08ECULL, "rdr111dg_martin.cnv"},
	{0xFC50D80E08225420ULL, "vxlt_0_lowtown_docks_d0t_0_0.vlm"},
	{0xFC51154807687B68ULL, "vxlt_0_lowtown_a2_n_c_gcv_0_0.vlm"},
	{0xFC514C32F1D6B7C8ULL, "art\\characters\\creatures\\dragonling\\c_dragonling.msh"},
	{0xFC515E32F1D6D659ULL, "art\\characters\\creatures\\dragonling\\c_dragonling.mao"},
	{0xFC516A32F1D6EAF2ULL, "art\\characters\\creatures\\dragonling\\c_dragonling.mmh"},
	{0xFC51A32305320448ULL, "hlac_lt_undercityencounter_-1_2z.dds"},
	{0xFC51A3D36286D752ULL, "and100_anders6096012.cl"},
	{0xFC51B188A0746577ULL, "vxlt_0_lowtown_docks_b0y_0_0.vlm"},
	{0xFC51B74EB93485F8ULL, "gen_im_trs_krk_drk_junk1_04.uti"},
	{0xFC51E46EDC8D5D55ULL, "hlac_ht_mansion2_818_43z.dds"},
	{0xFC51F3D1118CB022ULL, "mcr370_templar26087643.cl"},
	{0xFC520507B9071925ULL, "vxca_cavern1_haunted_ckk_0_0.vlm"},
	{0xFC523095F82A02FAULL, "vxlt_hangedman_n_1q4_0_0.vlm"},
	{0xFC5270DDEE3F8580ULL, "vxlt_backalley_g_1k4_0_0.vlm"},
	{0xFC52A585A21E42C9ULL, "mag100_cullen6140965.cl"},
	{0xFC52C31BBB7E7C62ULL, "6088301_m.fxe"},
	{0xFC52CFB623271512ULL, "93743181.ogg"},
	{0xFC5359B9217F0B1CULL, "rdr111dg_martin.bnk"},
	{0xFC53641908A623C1ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopening_rog.gad"},
	{0xFC5382404E831E24ULL, "art\\characters\\weapons\\greataxes\\w_axe_ax017a.msh"},
	{0xFC5385806F0C4732ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst_lp.evt"},
	{0xFC5394404E833C8DULL, "art\\characters\\weapons\\greataxes\\w_axe_ax017a.mao"},
	{0xFC5398404E8343C6ULL, "art\\characters\\weapons\\greataxes\\w_axe_ax017a.mmh"},
	{0xFC53B383B39F70FBULL, "art\\levels\\races\\proxy\\props\\pressure\\animation\\pressureresetup.gad"},
	{0xFC53E2ABC1514B5AULL, "art\\vfx\\combat\\vfx_c_rw_crust_c.dura.ani"},
	{0xFC542500B255E725ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.mmh"},
	{0xFC54599649733957ULL, "art\\characters\\playercharacter\\humanmale\\rob_tmpa_1.mao"},
	{0xFC5472E69F9A2D67ULL, "one140cr_charade6137614.cl"},
	{0xFC54A377CF585F67ULL, "6201297_m.fxe"},
	{0xFC54BF5409EEBF39ULL, "mag320_orsino6071079.cl"},
	{0xFC55A4C5451D02B6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_nervous.ani"},
	{0xFC55F3F6027E9ED1ULL, "art\\characters\\creatures\\spider\\animation\\kd_enter.gad"},
	{0xFC568E77B6895E9AULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeprooflrg.mmh"},
	{0xFC569277B68965D1ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeprooflrg.mao"},
	{0xFC56A077B6897DE0ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeprooflrg.msh"},
	{0xFC56E76925A4F66DULL, "6138824_f.fxe"},
	{0xFC577ADD52F62C73ULL, "hightown_fem_noble2_mob6169532.cl"},
	{0xFC58A3BD7590487AULL, "zz_qcr_debug6054932.cl"},
	{0xFC592E3A9B332FA3ULL, "hlac_lt_0_lowtown_a3_n_c_-1_5y.dds"},
	{0xFC59B6E159577B32ULL, "6180975_m.fxe"},
	{0xFC59DAF43F468095ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_7z.dds"},
	{0xFC5A30AC1073C65BULL, "llac_lt_warehouse_n_65537_3.dds"},
	{0xFC5A3BFC0292D4C0ULL, "follower_banter6171547.cl"},
	{0xFC5A5BA766DF0264ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_007a_0n.dds"},
	{0xFC5A651908AC446FULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening_pc\\hurl2_pcopening_rog.ani"},
	{0xFC5AA4806F129AC5ULL, "art\\characters\\creatures\\pridedemon\\animation\\d_cst_lp.gad"},
	{0xFC5ABAAAA42DA5C2ULL, "6051890_f.fxe"},
	{0xFC5B50FD97680411ULL, "art\\characters\\weapons\\arrows\\textures\\w_arw_001_n.dds"},
	{0xFC5BA89D656E13ECULL, "vxca_caverns2_8gc_0_0.vlm"},
	{0xFC5BF4F602830C7FULL, "art\\characters\\creatures\\spider\\animation\\kd_enter.ani"},
	{0xFC5C09649E657B04ULL, "6139911_m.fxe"},
	{0xFC5C363C7323B568ULL, "116271550.ogg"},
	{0xFC5C4513DA37A80CULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v1.ani"},
	{0xFC5D788B905D3F05ULL, "qalevelext.arl"},
	{0xFC5D7AE0394537F9ULL, "669850867.ogg"},
	{0xFC5DA8509C8B13F3ULL, "hlac_dr_primevaltomb_659_7x.dds"},
	{0xFC5DCF997E1EC85EULL, "hlac_ht_0_hightown_n_131070_0x.dds"},
	{0xFC5E8F404E8D3B91ULL, "art\\characters\\weapons\\greataxes\\w_axe_ax017a.phy"},
	{0xFC5EE337FA768DE4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maso_0d.dds"},
	{0xFC5F1D41098B6F0AULL, "vxlt_0_lowtown_a3_n_c_eu1_0_0.vlm"},
	{0xFC5FBD0F38740724ULL, "llac_lt_0_lowtown_docks_0_6.dds"},
	{0xFC602365926A7DF7ULL, "vxlt_0_lowtown_docks_b7x_0_0.vlm"},
	{0xFC60576BE655956AULL, "115222942.ogg"},
	{0xFC6076E4AAFD31A5ULL, "vxca_cavern1_haunted_clj_0_0.vlm"},
	{0xFC60A84190E62F5DULL, "6177664_m.fxe"},
	{0xFC60F3DE82FD4575ULL, "vxdr_0_deeproads_ewk_0_0.vlm"},
	{0xFC6100C0E43674D6ULL, "vxht_0_hightown_n_ajo_0_0.vlm"},
	{0xFC612200B2606CF2ULL, "art\\vfx\\combat\\vfx_ha_hd_bdrain_tar_lp_c.anb"},
	{0xFC61252BAFCA03A3ULL, "hlac_lt_warehouse_n_1_6z.dds"},
	{0xFC613970C68BD622ULL, "384550003.ogg"},
	{0xFC616BB15B569A8DULL, "mag310cr_progenitor_1.utc"},
	{0xFC6181F434430D06ULL, "vxlt_gallowstemplar_f_39k_0_0.vlm"},
	{0xFC61851C65745FE3ULL, "gen_im_trs_cot_junk3_01.uti"},
	{0xFC61B683AF8E8ACBULL, "6127230_m.fxe"},
	{0xFC61D4561C9BF364ULL, "vxlt_gallowsprison_n_5wa_0_0.vlm"},
	{0xFC623149E6966DF7ULL, "316900833.ogg"},
	{0xFC6244A27B60D7ACULL, "6167696_f.fxe"},
	{0xFC635213DA3DDD4EULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v1.gad"},
	{0xFC636834C6CA520BULL, "296952578.ogg"},
	{0xFC63ACC545294DA7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_nervous.evt"},
	{0xFC63B82AC6608086ULL, "152858090.ogg"},
	{0xFC649E1F6C7272AFULL, "vxw_sundermount_2tx_0_0.vlm"},
	{0xFC651284D9B828B6ULL, "lgt220_mag_apost_apost1.cnv"},
	{0xFC65D37E47090732ULL, "follower_banter6160990.cl"},
	{0xFC65E8C7FDCEB7DFULL, "6206930_f.fxe"},
	{0xFC65F0F4D243EE4EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha03.phy"},
	{0xFC663848172A2F66ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_app_aoe_p.impa.ani"},
	{0xFC67474F88D2D7D7ULL, "695343793.ogg"},
	{0xFC68122923A46F97ULL, "760634502.ogg"},
	{0xFC681484D9BA6196ULL, "lgt220_mag_apost_apost1.bnk"},
	{0xFC685B0EAFDF3B70ULL, "206154091.ogg"},
	{0xFC68718C066F6158ULL, "6101013_m.fxe"},
	{0xFC68791654BC3712ULL, "6185382_f.fxe"},
	{0xFC68ECE038FA525BULL, "vxdr_slavers_2hh_0_1.vlm"},
	{0xFC698EDB4C26D02EULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hm_overlay.anb"},
	{0xFC699464F7C0EF66ULL, "dae000ip_crowd_em1_talk6.utp"},
	{0xFC69CDC5452DF1F0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\dg_tk_nervous.gad"},
	{0xFC6A050C464F6773ULL, "zz_qun_debug6028277.cl"},
	{0xFC6A124139A1F85CULL, "vxlt_0_lowtown_a2_n_c_fe9_0_0.vlm"},
	{0xFC6A4D13DA43F32DULL, "art\\characters\\creatures\\ashwraith\\animation\\ras_atk_v1.evt"},
	{0xFC6A6203520AF2FEULL, "gen00fl_varric6019884.cl"},
	{0xFC6ABD965828BA76ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_viscounthead_arishokaddress.ani"},
	{0xFC6AD80E34B965F7ULL, "vxdr_primevaltomb_8vi_0_0.vlm"},
	{0xFC6B2FAB68E8EF0AULL, "6110564_f.fxe"},
	{0xFC6B3ED4F0722DFBULL, "6071843_m.fxe"},
	{0xFC6B9BA523DBA2B9ULL, "llac_ht_bartrand_n_3381_1.dds"},
	{0xFC6BADC72B900D76ULL, "433278971.ogg"},
	{0xFC6BCC60E5FF71EDULL, "hlac_lt_0_lowtown_docks_a3_n_65532_0z.dds"},
	{0xFC6BD83199886291ULL, "288780796.ogg"},
	{0xFC6C583739A053CEULL, "6162510_f.fxe"},
	{0xFC6CE0311209275BULL, "vxht_keep_n_6id_0_0.vlm"},
	{0xFC6D6FAB79FF742FULL, "317842763.ogg"},
	{0xFC6D80AC7391BD83ULL, "hlac_lt_0_lowtown_n_1_8z.dds"},
	{0xFC6DF901EB54AC68ULL, "vxlt_0_lowtown_a2_n_c_gyt_0_0.vlm"},
	{0xFC6E4B85A33417BEULL, "6096892_m.fxe"},
	{0xFC6E8316E6903941ULL, "6059152_f.fxe"},
	{0xFC6ED1468D80A15EULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.mmh"},
	{0xFC6EF433AFBB51EFULL, "vxlt_0_lowtown_a2_n_gwk_0_0.vlm"},
	{0xFC6F2F7E1277C637ULL, "dae000_flemeth6013331.cl"},
	{0xFC6F65BB74F35DF5ULL, "vxdr_0_deeproads_epj_0_0.vlm"},
	{0xFC6FB381549A9C3CULL, "hlac_lt_0_lowtown_65536_0z.dds"},
	{0xFC70109A436A0B52ULL, "dae000_gamlen6147223.cl"},
	{0xFC70451EEA7ACB08ULL, "282833002.ogg"},
	{0xFC70D97016E1206DULL, "hlac_dr_primevaltomb_664_1x.dds"},
	{0xFC70EC623449A6E3ULL, "isa211_isabela6083209.cl"},
	{0xFC70F1886DDA1D96ULL, "6173317_m.fxe"},
	{0xFC7147115E4F757CULL, "6102908_m.fxe"},
	{0xFC71CBD9AE5D983EULL, "dae000cr_loyal_mage_amb.utc"},
	{0xFC728ED0766238A0ULL, "vis121_saemus6071966.cl"},
	{0xFC729D1D85E330DAULL, "art\\characters\\weapons\\longbows\\textures\\w_lbw_dslba_0t.dds"},
	{0xFC72F3D3C55246EEULL, "hlac_lt_0_lowtown_a3_n_c_-131072_6z.dds"},
	{0xFC7398E16A8930A6ULL, "6116011_m.fxe"},
	{0xFC73B8B5782BB7FDULL, "lgt160cr_wk_avern_bandit.utc"},
	{0xFC73BBAEF5BC33A1ULL, "ht_0_hightown_n_26_0.rml"},
	{0xFC742048173643D7ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_app_aoe_p.impa.evt"},
	{0xFC74D71C5625A6CBULL, "352531837.ogg"},
	{0xFC7545699F2E5BBDULL, "6128318_m.fxe"},
	{0xFC75DEB83F68A125ULL, "sib100_father_contact6110071.cl"},
	{0xFC75F57E04C1121AULL, "one100dg_lia6049671.cl"},
	{0xFC75F9F7954E466EULL, "681142448.ogg"},
	{0xFC765C5FF43F9EC6ULL, "art\\characters\\creatures\\qunari\\textures\\qm_bdy_com_0n.dds"},
	{0xFC765DB262B96127ULL, "gen_im_trs_apo_junk2_02.uti"},
	{0xFC76B0B81A0C4679ULL, "hlac_w_bonepitmining_a2_131075_0x.dds"},
	{0xFC77448ABD81DD69ULL, "6047527_m.fxe"},
	{0xFC77565D851EC59DULL, "6031272_m.fxe"},
	{0xFC775F9279D10D26ULL, "vxht_keep_n_7ah_0_0.vlm"},
	{0xFC778F1DD52B87E7ULL, "vxht_bartrand_n_6qw_0_0.vlm"},
	{0xFC77C3624022AD3BULL, "w_bonepit_mmap.dds"},
	{0xFC7804AF967DAE64ULL, "6056876_m.fxe"},
	{0xFC783328B69FB87FULL, "900615341.ogg"},
	{0xFC789A45EDCB7D61ULL, "llac_lt_0_lowtown_a2_n_65535_3.dds"},
	{0xFC78D3E8F02C942EULL, "dae000_pc_home_ambients6208208.cl"},
	{0xFC792424E09C18A2ULL, "dae300_cullen6062155.cl"},
	{0xFC7949EB26AF7E77ULL, "vxdr_primevaltomb_8qh_0_0.vlm"},
	{0xFC794CC9DD836D69ULL, "601632135.ogg"},
	{0xFC79D13381093602ULL, "t1_skn_d06.tnt"},
	{0xFC7A9903F3170E92ULL, "vxlt_gallowstemplar_f_2jc_0_0.vlm"},
	{0xFC7B2148173C6460ULL, "art\\vfx\\combat\\vfx_r_shd_decoy_app_aoe_p.impa.gad"},
	{0xFC7B606DE226384CULL, "vxht_mansion2b_24j_0_1.vlm"},
	{0xFC7B65E67F6B8A76ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v2.evt"},
	{0xFC7B9C111C91DDE2ULL, "891106001.ogg"},
	{0xFC7BAD77B6A8AFA5ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9cc_keeprooflrg.phy"},
	{0xFC7BAE579EBA76AFULL, "gen00fl_merrill6151000.cl"},
	{0xFC7BB0468D8AF31DULL, "art\\vfx\\combat\\vfx_m_spi_frc_charge_c.anb"},
	{0xFC7BD36A8C48FF4DULL, "vxht_keep_6fp_0_0.vlm"},
	{0xFC7C503260891AF0ULL, "303418603.ogg"},
	{0xFC7E10EB6EE2232FULL, "6095225_m.fxe"},
	{0xFC7E29EDCA553D72ULL, "vxw_woundedcoast_12d_0_0.vlm"},
	{0xFC7E2B7898A78817ULL, "follower_banter6155562.cl"},
	{0xFC7E5B1D973CBC84ULL, "260474223.ogg"},
	{0xFC7E7F870944002EULL, "mer110_jansen6032616.cl"},
	{0xFC7EE6965839A8B0ULL, "art\\cinematics\\_crit path\\qunari crisis\\arishok address\\cs_viscounthead_arishokaddress.gad"},
	{0xFC7F0B11378E94B8ULL, "plt_mrl250pt_varterral.nss"},
	{0xFC7F650F2CA6D6A3ULL, "6207941_m.fxe"},
	{0xFC7FB5FD080C0F2AULL, "105478632.ogg"},
	{0xFC80F21B13E124FBULL, "vxlt_gallowstemplar_f_3rs_0_0.vlm"},
	{0xFC8130118E585A70ULL, "523526829.ogg"},
	{0xFC815CE67F6FE601ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v2.gad"},
	{0xFC81C16E0C64B7FBULL, "follower_banter6136662.cl"},
	{0xFC82902479451728ULL, "364832167.ogg"},
	{0xFC82B0E277D56066ULL, "536275833.ogg"},
	{0xFC82B60114F8E42CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1_reverse.ani"},
	{0xFC833157EE4DB270ULL, "plt_cod_cha_alistair.plo"},
	{0xFC837410468D19DEULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_paperlamp_d.dds"},
	{0xFC83A4A6DD1FD6D6ULL, "6173865_m.fxe"},
	{0xFC83AED2A75A20BCULL, "vxht_mansion_n_6ew_0_0.vlm"},
	{0xFC8401F4FC06CB52ULL, "vxlt_darktown_2ww_0_0.vlm"},
	{0xFC842A6C8B6C3DA5ULL, "dae000cr_grand_cleric6178468.cl"},
	{0xFC843A434740D231ULL, "630530547.ogg"},
	{0xFC84B0C05B9FF54BULL, "vxdr_primevalentrance_23p_0_0.vlm"},
	{0xFC84EC000AED3A20ULL, "6166403_m.fxe"},
	{0xFC851C39ABD41ED7ULL, "324583332.ogg"},
	{0xFC851DDECFF5ACE4ULL, "hlac_w_encounter2_-1_0x.dds"},
	{0xFC8530BB92AB2EB2ULL, "774117582.ogg"},
	{0xFC85A1BCE8BDA306ULL, "6041566_m.fxe"},
	{0xFC8623356D35429EULL, "453595335.ogg"},
	{0xFC87335E05474BC8ULL, "6034212_m.fxe"},
	{0xFC8736246598705DULL, "dae000_arishok6082043.cl"},
	{0xFC873B6583B48FF9ULL, "gen00fl_aveline6116716.cl"},
	{0xFC875685308AA486ULL, "lt_0_lowtown_n_27_0_ll.mmh"},
	{0xFC880CA6FE52DADDULL, "follower_banter6170231.cl"},
	{0xFC885D4D9CD3EAC2ULL, "6092342_m.fxe"},
	{0xFC885DE67F7606AFULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\vrc_bayonet_atk_v2.ani"},
	{0xFC88904570E81790ULL, "art\\levels\\ztools\\helpers\\texels\\_textures\\1024x2048m.dds"},
	{0xFC88C546A71A39F1ULL, "6042490_m.fxe"},
	{0xFC890AE0E50D2712ULL, "vxlt_gallowstemplar_f_2ol_0_0.vlm"},
	{0xFC897855D8FAE5EBULL, "art\\levels\\races\\proxy\\props\\_textures\\carpetnew_sa.dds"},
	{0xFC89C30114FF196EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1_reverse.gad"},
	{0xFC89C9113BB41C7BULL, "lowtown_male_urchin_sta6164209.cl"},
	{0xFC89D5145D0C16EDULL, "llac_lt_gallowsprison_n_-1_11.dds"},
	{0xFC8A0BC349160F29ULL, "dae000_pc_home_ambients6208692.cl"},
	{0xFC8A7AB104CD415BULL, "vxlt_0_lowtown_docks_cs8_0_1.vlm"},
	{0xFC8AEAF01062B0D5ULL, "780878870.ogg"},
	{0xFC8BDE50230FBEF2ULL, "hlac_w_sundermount_-2_2x.dds"},
	{0xFC8C434EFAD48123ULL, "hlac_lt_0_lowtown_a3_n_c_2_0x.dds"},
	{0xFC8C7C50E11BF981ULL, "vxlt_0_lowtown_a2_n_c_feh_0_0.vlm"},
	{0xFC8CC2884CA9040DULL, "vxht_bartrand_n_64v_0_0.vlm"},
	{0xFC8CDB637D7CF295ULL, "6094563_m.fxe"},
	{0xFC8D7B273A52C75EULL, "vxlt_0_lowtown_n_es0_0_0.vlm"},
	{0xFC8DBF2083723782ULL, "dae000ip_crowd_lm4_talk2.utp"},
	{0xFC8E2ACCCE8CE1BFULL, "57441424.ogg"},
	{0xFC8EBB59D22ECD6BULL, "art\\levels\\races\\design\\_textures\\plc_metaltrp010_0_sa.dds"},
	{0xFC8FA96F90EBE804ULL, "117413126.ogg"},
	{0xFC8FBEF30DE4FA9BULL, "6024575_m.fxe"},
	{0xFC904D103A88AD35ULL, "137866521.ogg"},
	{0xFC904FA5FB53434AULL, "464963255.ogg"},
	{0xFC9061B966CC28CBULL, "hlac_lt_darktown_-1_81x.dds"},
	{0xFC906CA9B50EAE5CULL, "gen00fl_aveline6104954.cl"},
	{0xFC90BE0115052F4DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1_reverse.evt"},
	{0xFC90F0DA1F119A67ULL, "6109991_m.fxe"},
	{0xFC913A3D3DC5613EULL, "vxht_mansion2_267_0_0.vlm"},
	{0xFC91D9C144C0B428ULL, "6208558_m.fxe"},
	{0xFC93229D4D960DCBULL, "vxdr_primevalentrance_26s_0_0.vlm"},
	{0xFC94A281BB110147ULL, "clf260_harley6080091.cl"},
	{0xFC94B24951A24133ULL, "t1_stb_org.tnt"},
	{0xFC95A53AF73D6448ULL, "6031112_m.fxe"},
	{0xFC961D0A1AB83EBBULL, "one300dg_solivitus6060303.cl"},
	{0xFC96CF2A8ECA0342ULL, "6097042_m.fxe"},
	{0xFC97015031AE8AAAULL, "fen310_party_comment6116853.cl"},
	{0xFC9727D279448F07ULL, "lt_0_lowtown_docks_a3_34_0.rml"},
	{0xFC97B7295B676E28ULL, "hlac_lt_hangedman_n_-1_19z.dds"},
	{0xFC996F5F2595FAA6ULL, "w_bonepitdestroyed_a3_30_0_ll.mmh"},
	{0xFC99A06CDAF5434AULL, "hlac_lt_0_lowtown_docks_n_-4_0x.dds"},
	{0xFC99A18FF2B98BFFULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc3a.msh"},
	{0xFC99A78FF2B995A1ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc3a.mmh"},
	{0xFC9A74FC5D95FD77ULL, "hlac_ht_mansion2_818_23z.dds"},
	{0xFC9AB55EF76F0A61ULL, "6099018_m.fxe"},
	{0xFC9ABB8BD45277ADULL, "w_encounter3_219_0.mmh"},
	{0xFC9ABF8BD4527EFEULL, "w_encounter3_219_0.mao"},
	{0xFC9ACD8BD4529633ULL, "w_encounter3_219_0.msh"},
	{0xFC9B23DA77E8D039ULL, "vxw_sundermount_36u_0_0.vlm"},
	{0xFC9B284E2F379544ULL, "vxlt_hangedman_n_16s_0_0.vlm"},
	{0xFC9B4457EEF5C76AULL, "target_creature_invalid.cur"},
	{0xFC9B6D5C47F469F3ULL, "vxdr_0_deeproads_fi0_0_0.vlm"},
	{0xFC9B7AB66AF177DBULL, "w_sundermount_7_0.rml"},
	{0xFC9B90A8E208AAE8ULL, "786262461.ogg"},
	{0xFC9BC3D3AE4E8D21ULL, "follower_banter6164555.cl"},
	{0xFC9C2C7575BDAE55ULL, "ch1alphemarmtnt_vl.psh.fxo"},
	{0xFC9C9D21E21D4D3CULL, "gen00fl_merrill6129882.cl"},
	{0xFC9D1F7B8AA4A6B2ULL, "hm_lgt_conrad.mor"},
	{0xFC9D4B91A2125D31ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_hugekeg_0.pwk"},
	{0xFC9DFC03F99A8F7DULL, "6113845_f.fxe"},
	{0xFC9E9BA26F812409ULL, "hlac_lt_0_lowtown_a2_131071_3z.dds"},
	{0xFC9EE5560AC85D15ULL, "6119230_f.fxe"},
	{0xFC9EE8F2AB4CA300ULL, "412531064.ogg"},
	{0xFC9FA6B16C011F0CULL, "68299686.ogg"},
	{0xFC9FF769E1578D0FULL, "mag131_keran6086212.cl"},
	{0xFCA055D5B461F098ULL, "lt_hangedman_probe_2rf.dds"},
	{0xFCA096A813D58574ULL, "6085840_f.fxe"},
	{0xFCA098C8285463D9ULL, "58706234.ogg"},
	{0xFCA0BA886F1662D0ULL, "ca_caverns1_mmi.gff"},
	{0xFCA1364E2DAE3666ULL, "6127194_m.fxe"},
	{0xFCA182A66DDFEFCCULL, "hla_ht_0_hightown_131070_0z.dds"},
	{0xFCA3049F884B451AULL, "vxlt_0_lowtown_a3_n_c_hlk_0_0.vlm"},
	{0xFCA3359939FBEBC3ULL, "374520145.ogg"},
	{0xFCA3DFB92DAD613FULL, "282769230.ogg"},
	{0xFCA434850A5EAF9DULL, "zz_followers_banter6159117.cl"},
	{0xFCA4FF90A271EAACULL, "vxlt_0_lowtown_a3_n_c_fj1_0_0.vlm"},
	{0xFCA5EFF4D279D011ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha03.mmh"},
	{0xFCA5F164827F188DULL, "6088135_m.fxe"},
	{0xFCA609F4D279FCCFULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha03.msh"},
	{0xFCA6409921F2D50CULL, "6114037_m.fxe"},
	{0xFCA64E2400574B2FULL, "hlac_lt_0_lowtown_docks_a3_-65540_1x.dds"},
	{0xFCA69AAEF5E75531ULL, "ht_0_hightown_n_26_0.anb"},
	{0xFCA6A33D54540982ULL, "w_encounter2_camp_32_0.rml"},
	{0xFCA71892685235D3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\quick_enter.evt"},
	{0xFCA73C9DDCED45E8ULL, "dae000_mother6136909.cl"},
	{0xFCA73D87FE418E09ULL, "hlac_w_encounter2_3_0z.dds"},
	{0xFCA74CFD67FF8CA1ULL, "322174055.ogg"},
	{0xFCA766E28B422978ULL, "vxlt_0_lowtown_a2_gnd_0_0.vlm"},
	{0xFCA7BCF11AD4164FULL, "vxht_mansion2b_2z6_0_0.vlm"},
	{0xFCA7CBDDD78F169AULL, "hlac_ca_caverns1_mines_24424_13z.dds"},
	{0xFCA8160A57A09BCAULL, "ser270_templar6127476.cl"},
	{0xFCA83AAE69CEC4CDULL, "6180508_f.fxe"},
	{0xFCA8D289D0EAC8C7ULL, "6048844_f.fxe"},
	{0xFCA96D57D03F3CC1ULL, "vxlt_gallowstemplar_5r8_0_0.vlm"},
	{0xFCA9BD29B16678F0ULL, "hlac_w_blightlands_262146_0z.dds"},
	{0xFCAA34B119341908ULL, "dae000_bartrand6027583.cl"},
	{0xFCAA4A4AF218C3E4ULL, "vxw_encounter2_camp_2z_0_0.vlm"},
	{0xFCAA5EE11E3EF85EULL, "vxlt_0_lowtown_n_ei2_0_0.vlm"},
	{0xFCAA914B97728437ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_5y.dds"},
	{0xFCAA964A21F732ADULL, "84974111.ogg"},
	{0xFCAAA2BA838900F4ULL, "hlac_dr_primevaltomb_9621_4x.dds"},
	{0xFCAAE1F408DCB43BULL, "vxw_woundedcoast_mu_0_0.vlm"},
	{0xFCAB0826474F1D26ULL, "vxlt_0_lowtown_n_eom_0_1.vlm"},
	{0xFCAB1710EA1FAFB8ULL, "lit_bt_distor.mao"},
	{0xFCAB62D6BD2ADC9EULL, "737975246.ogg"},
	{0xFCAC18B221F4D0A4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv1.ani"},
	{0xFCAC81C2CEAC5404ULL, "299164430.ogg"},
	{0xFCADA43D545A2A36ULL, "w_encounter2_camp_32_0.phy"},
	{0xFCADDFA04BFAC86BULL, "496592150.ogg"},
	{0xFCADF63CB1B3BFFCULL, "6187868_m.fxe"},
	{0xFCAE119268584814ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\quick_enter.gad"},
	{0xFCAEB7CD9409DBFBULL, "6156021_m.fxe"},
	{0xFCAED05389703E27ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\prp_scaffoldb_1.mao"},
	{0xFCAEE7BD1CCC2A47ULL, "hlac_lt_0_lowtown_docks_n_-65535_0y.dds"},
	{0xFCAF0BE348472E47ULL, "vxca_caverns2prc_9sp_0_0.vlm"},
	{0xFCAF2B8615D7310EULL, "6109226_m.fxe"},
	{0xFCAF45E8BCA82E63ULL, "vxca_caverns1_ck0_0_0.vlm"},
	{0xFCAF5AFC3F4DA596ULL, "w_bonepitdestroyed_a3_130_0.msh"},
	{0xFCAF64FC3F4DB694ULL, "w_bonepitdestroyed_a3_130_0.mmh"},
	{0xFCAF68FC3F4DBDA7ULL, "w_bonepitdestroyed_a3_130_0.mao"},
	{0xFCAFA0895938DB9CULL, "611058100.ogg"},
	{0xFCAFC991AC62F039ULL, "6044962_m.fxe"},
	{0xFCAFE523ACE6B429ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_right.ani"},
	{0xFCB064BC1378E579ULL, "294046077.ogg"},
	{0xFCB0D982A801EFF1ULL, "llac_lt_0_lowtown_docks_-65538_9.dds"},
	{0xFCB1767C7A415D9AULL, "vxlt_0_lowtown_a3_n_c_hid_0_0.vlm"},
	{0xFCB180FA4754CD4BULL, "vxlt_0_lowtown_docks_9r8_0_0.vlm"},
	{0xFCB1B54D23C6CAA1ULL, "vxlt_0_lowtown_docks_a3_d6h_0_0.vlm"},
	{0xFCB1E1087758895DULL, "hlac_w_bonepitmining_a2_3_0x.dds"},
	{0xFCB1FFF72E6F56D1ULL, "180333911.ogg"},
	{0xFCB20DA718AC5FD6ULL, "36769951.ogg"},
	{0xFCB215B221F93706ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv1.gad"},
	{0xFCB243A71DB0A0DAULL, "6167600_m.fxe"},
	{0xFCB25928C840E79EULL, "vxlt_undercityencounter_13j_0_0.vlm"},
	{0xFCB2B5B82EFCF9E3ULL, "mcr370_guardsman6087665.cl"},
	{0xFCB2DE5CCF55DDA9ULL, "tools_way_pt03.phy"},
	{0xFCB32843082FBA52ULL, "6154615_m.fxe"},
	{0xFCB39FF7F6FFF2ACULL, "zz_vault_debug6177617.cl"},
	{0xFCB46074A7F121F4ULL, "art\\characters\\creatures\\ragedemon\\animation\\dblw_dmg_v1.ani"},
	{0xFCB463417475310DULL, "6083035_m.fxe"},
	{0xFCB474DA27722321ULL, "vxlt_darktown_2rg_0_0.vlm"},
	{0xFCB47A064CC6281FULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_102a_0t.dds"},
	{0xFCB47F5B31813365ULL, "68714171.ogg"},
	{0xFCB4A2B2971B9E74ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_boo_maso_0t.dds"},
	{0xFCB4A3E7B7CA26DEULL, "6095251_m.fxe"},
	{0xFCB4BE4DF5DAFCCFULL, "hlac_ht_0_hightown_65533_3y.dds"},
	{0xFCB4C9A1E6898FCBULL, "gen00fl_fenris6106615.cl"},
	{0xFCB52092685E80F2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\quick_enter.ani"},
	{0xFCB56346D3324455ULL, "6051326_m.fxe"},
	{0xFCB5A3760327C225ULL, "vxlt_0_lowtown_a2_n_c_fip_0_0.vlm"},
	{0xFCB70023ACED00ABULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_right.gad"},
	{0xFCB78D3D546257D7ULL, "w_encounter2_camp_32_0.msh"},
	{0xFCB79F3D5462763AULL, "w_encounter2_camp_32_0.mao"},
	{0xFCB7A33D54627D49ULL, "w_encounter2_camp_32_0.mmh"},
	{0xFCB806A4338EE333ULL, "gen00fl_fenris6114528.cl"},
	{0xFCB83FB16A8FC1F1ULL, "one280_iwan6097276.cl"},
	{0xFCB89054211F2AA5ULL, "vxlt_0_lowtown_a2_n_c_e88_0_0.vlm"},
	{0xFCB8B0B867E60BFAULL, "117513769.ogg"},
	{0xFCB8B24E9334D412ULL, "mag100_cullen6157168.cl"},
	{0xFCB8B83B5219F18CULL, "follower_banter6136689.cl"},
	{0xFCB910B221FF4DE5ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\basebow\\combat\\rangedv1.evt"},
	{0xFCB93201D705495CULL, "vxlt_backalley_g_1kh_0_0.vlm"},
	{0xFCB94344DC48F4E7ULL, "401509969.ogg"},
	{0xFCB987CD74306E3AULL, "vxca_cave_2b3_0_0.vlm"},
	{0xFCB9BE8FAB9BB84FULL, "6067735_f.fxe"},
	{0xFCBA036C6C5E72CBULL, "6152128_m.fxe"},
	{0xFCBA22C5FC6F858DULL, "6169991_m.fxe"},
	{0xFCBA447CA182787BULL, "vxca_caverns2_96j_0_0.vlm"},
	{0xFCBA51894923B307ULL, "vxw_woundedcoast_1p7_0_0.vlm"},
	{0xFCBA5D74A7F58856ULL, "art\\characters\\creatures\\ragedemon\\animation\\dblw_dmg_v1.gad"},
	{0xFCBADC4B9103A04DULL, "vxht_0_hightown_n_bab_0_0.vlm"},
	{0xFCBB1162E30AD94DULL, "704010380.ogg"},
	{0xFCBB34A7F4ED99FEULL, "vxlt_gallowstemplar_f_3ga_0_0.vlm"},
	{0xFCBB8290F105EFF2ULL, "6107185_m.fxe"},
	{0xFCBCFD23ACF16710ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_injured_punched_right.evt"},
	{0xFCBD835499F52835ULL, "hlac_ht_0_hightown_n_-2_0z.dds"},
	{0xFCBDC37DE8CD64F7ULL, "zz_mcr_debug6071545.cl"},
	{0xFCBE41F6D6EADD0EULL, "106462980.ogg"},
	{0xFCBEE3A650E928E8ULL, "6094966_m.fxe"},
	{0xFCBEEC9B12E6979FULL, "zz_followers_banter6157294.cl"},
	{0xFCBF7B1D4F6EA12EULL, "405689392.ogg"},
	{0xFCBF83F3D6BB6330ULL, "vxw_bonepitmining_a2_21e_0_0.vlm"},
	{0xFCBF8BE0A244AF99ULL, "hlac_dr_dungeon_3214_6y.dds"},
	{0xFCBF8C003031405EULL, "hlac_lt_0_lowtown_-196608_3y.dds"},
	{0xFCBFAF43D42895E6ULL, "gen00fl_aveline6113606.cl"},
	{0xFCBFB8ACD7F9DD74ULL, "6168434_m.fxe"},
	{0xFCC0174C125BA391ULL, "928157085.ogg"},
	{0xFCC0D3B07EE9E3D0ULL, "6054580_m.fxe"},
	{0xFCC0EF34233605CCULL, "6120360_m.fxe"},
	{0xFCC13F89129C01CCULL, "vxlt_undercityencounter_1j7_0_0.vlm"},
	{0xFCC17874A7FBD4B5ULL, "art\\characters\\creatures\\ragedemon\\animation\\dblw_dmg_v1.evt"},
	{0xFCC21470AFD0AB56ULL, "llac_lt_0_lowtown_docks_-2_1.dds"},
	{0xFCC23389C4CAE297ULL, "ave121_brennan6105815.cl"},
	{0xFCC234C39528F0B4ULL, "342447383.ogg"},
	{0xFCC243A4A30CB2C1ULL, "558633708.ogg"},
	{0xFCC2827576981C3BULL, "357666125.ogg"},
	{0xFCC2E6B719683BA1ULL, "vxlt_darktown_2uf_0_0.vlm"},
	{0xFCC30DAA69E08949ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1d.mao"},
	{0xFCC3152C08716576ULL, "6129398_m.fxe"},
	{0xFCC319AA69E09DE2ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1d.mmh"},
	{0xFCC31BAA69E0A098ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_papers1d.msh"},
	{0xFCC456827A33DE10ULL, "docks_male_sailor_mob6176673.cl"},
	{0xFCC479194923E312ULL, "dae000cr_tranquil_em.utc"},
	{0xFCC4A0AAFEC0474FULL, "vxht_mansion2b_2p0_0_0.vlm"},
	{0xFCC4AE3910CF4294ULL, "vxlt_warehouse_e8_0_0.vlm"},
	{0xFCC4DCFC96965941ULL, "6111659_m.fxe"},
	{0xFCC4F43BCD7EE159ULL, "410381682.ogg"},
	{0xFCC54B4DA98381CBULL, "vxht_0_hightown_a2_bg9_0_0.vlm"},
	{0xFCC5997391CFDB97ULL, "vxca_caverns1_jga_0_0.vlm"},
	{0xFCC6148D799C7D5EULL, "mag121_idunna6083969.cl"},
	{0xFCC6A58EE7B2604AULL, "vxw_sundermount_22c_0_0.vlm"},
	{0xFCC6A88FF2E0851EULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_hc3a.phy"},
	{0xFCC7093828B031D4ULL, "vxlt_gallowscourtyard_n_5dz_0_0.vlm"},
	{0xFCC70D27C1F58717ULL, "llac_ht_0_hightown_a3_196607_0.dds"},
	{0xFCC7C5ADECC8E53BULL, "vxw_woundedcoast_gw_0_0.vlm"},
	{0xFCC80E78B9E1173CULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_archdiv.mao"},
	{0xFCC81A78B9E12BDFULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_archdiv.mmh"},
	{0xFCC82078B9E13581ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_archdiv.msh"},
	{0xFCC87B563D349E9BULL, "247786904.ogg"},
	{0xFCC944C99C255980ULL, "fen300cr_danarius.ncs"},
	{0xFCC9A9D05DF65D70ULL, "4566515.ogg"},
	{0xFCCA334FAF569642ULL, "731399983.ogg"},
	{0xFCCA33A727731650ULL, "vxlt_gallowscourtyard_n_75k_0_0.vlm"},
	{0xFCCC09FC922D2FC6ULL, "art\\vfx\\combat\\vfx_gib_hm_robc_rtl.anb"},
	{0xFCCC16AB3F735939ULL, "vxlt_undercityencounter_108_0_0.vlm"},
	{0xFCCC36BEC64F5FE0ULL, "vxlt_gallowsprison_n_6be_0_0.vlm"},
	{0xFCCC900A92D2242CULL, "6201643_m.fxe"},
	{0xFCCD6E217379008AULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_window.mao"},
	{0xFCCD7221737907D9ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_window.mmh"},
	{0xFCCD7C21737918C7ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\3ww_window.msh"},
	{0xFCCDFBF99CCBD73FULL, "603432447.ogg"},
	{0xFCCE3C2193D2C682ULL, "w_bonepitdestroyed_a3_141_0.msh"},
	{0xFCCE3E2193D2CA38ULL, "w_bonepitdestroyed_a3_141_0.mmh"},
	{0xFCCE4A2193D2DEDBULL, "w_bonepitdestroyed_a3_141_0.mao"},
	{0xFCCE938E6C9E58A9ULL, "904404439.ogg"},
	{0xFCCF1C6947E5E790ULL, "hlac_lt_undercityserialkiller_-1_13x.dds"},
	{0xFCCF8B8A69695BFAULL, "dae021_chantry_initiate_mob6147587.cl"},
	{0xFCCFB16604921A4CULL, "vxlt_undercityencounter_1g4_0_0.vlm"},
	{0xFCD00AC9E5009309ULL, "hlac_ca_caverns2prc_5802_38x.dds"},
	{0xFCD08483C5126775ULL, "follower_banter6167958.cl"},
	{0xFCD0A555C250C2A9ULL, "hlac_dr_primevaltomb_657_0y.dds"},
	{0xFCD158940B5E5421ULL, "vxlt_darktown_2xe_0_0.vlm"},
	{0xFCD1C042D9D5E010ULL, "plt_cod_hst_raiders.plo"},
	{0xFCD24AF2965B8183ULL, "6070484_m.fxe"},
	{0xFCD2E8365197EFF4ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masc.phy"},
	{0xFCD36918B51B7F2AULL, "art\\characters\\creatures\\spider\\animation\\appr.gad"},
	{0xFCD3715163D72CCFULL, "839542312.ogg"},
	{0xFCD4ADD1B41F159EULL, "6164209_m.fxe"},
	{0xFCD4D7C350029EACULL, "llac_lt_0_lowtown_docks_a3_131068_0.dds"},
	{0xFCD4E20AF6E15174ULL, "6191368_m.fxe"},
	{0xFCD5514767B9976CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothic.msh"},
	{0xFCD5574767B9A15EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothic.mmh"},
	{0xFCD5634767B9B5F5ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0cc_gothic.mao"},
	{0xFCD5E262D1AB1673ULL, "6121954_f.fxe"},
	{0xFCD6386B4763A6CBULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_enter.ani"},
	{0xFCD63BCFBFA86B1BULL, "w_woundedcoast_26_0.mmh"},
	{0xFCD64022C49F0301ULL, "worldvault.gda"},
	{0xFCD647CFBFA87FB8ULL, "w_woundedcoast_26_0.mao"},
	{0xFCD659CFBFA89E05ULL, "w_woundedcoast_26_0.msh"},
	{0xFCD6BA31C0A11BE4ULL, "qun131_party_comment6117261.cl"},
	{0xFCD73CD4DE0FE6D0ULL, "vxca_caverns1_mines_gf1_0_0.vlm"},
	{0xFCD73D24487FD2EBULL, "vxlt_undercityserialkiller_1ie_0_0.vlm"},
	{0xFCD75347D8C39B44ULL, "6025984_m.fxe"},
	{0xFCD7B63964783EF8ULL, "6131555_m.fxe"},
	{0xFCD8452193DB2AC3ULL, "w_bonepitdestroyed_a3_141_0.phy"},
	{0xFCD85A51DD6D27EAULL, "915607697.ogg"},
	{0xFCD8BA23CB3BA280ULL, "270112685.ogg"},
	{0xFCD930F6E7F09F70ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyo.mao"},
	{0xFCD934F6E7F0A583ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyo.mmh"},
	{0xFCD942F6E7F0BDDDULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyo.msh"},
	{0xFCD96ECFA373009EULL, "gen00fl_fenris6127019.cl"},
	{0xFCD9759A02DEC23FULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1f.msh"},
	{0xFCD97B9A02DECCE1ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1f.mmh"},
	{0xFCD9879A02DEE082ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\prp_dooriron1f.mao"},
	{0xFCD9EFCD62ED89A7ULL, "hlac_w_blightlands_65538_1x.dds"},
	{0xFCDA44BB74706515ULL, "vxht_mansion2_1vg_0_0.vlm"},
	{0xFCDA4B0A07D5D4D2ULL, "6084745_m.fxe"},
	{0xFCDA4C18B5216CC1ULL, "art\\characters\\creatures\\spider\\animation\\appr.evt"},
	{0xFCDAB51C29B5DE3AULL, "hlac_ht_0_hightown_a3_65533_0x.dds"},
	{0xFCDB9C117E7F4A81ULL, "vxlt_backalley_g_1k9_0_0.vlm"},
	{0xFCDBE90EA8F2078DULL, "6037069_m.fxe"},
	{0xFCDC0DFB58E32D76ULL, "6120115_m.fxe"},
	{0xFCDC56C5A9D4C213ULL, "mrl000cr_frostaura.ncs"},
	{0xFCDCC53651A008F1ULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masc.msh"},
	{0xFCDCD33651A0200CULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masc.mao"},
	{0xFCDCDF3651A034AFULL, "art\\characters\\playercharacter\\humanfemale\\bdy_masc.mmh"},
	{0xFCDD412142A25F63ULL, "w_encounter2_camp_59_0.mao"},
	{0xFCDD452142A265B0ULL, "w_encounter2_camp_59_0.mmh"},
	{0xFCDD532142A27DCAULL, "w_encounter2_camp_59_0.msh"},
	{0xFCDD961A415A16DDULL, "6094677_m.fxe"},
	{0xFCDDB0F0A77512D3ULL, "6067071_m.fxe"},
	{0xFCDDBF30B57A924FULL, "6126444_m.fxe"},
	{0xFCDDDE23BE8F1BAAULL, "llac_ht_keep_n_4519_7.dds"},
	{0xFCDE09489F928B3EULL, "dae000_mother6099235.cl"},
	{0xFCDE63CCF4397B31ULL, "llac_lt_0_lowtown_a2_65534_5.dds"},
	{0xFCDE8F980E2F27A8ULL, "hlac_lt_darktown_-1_8z.dds"},
	{0xFCDEABCF1F98610FULL, "ave310_brennan6136222.cl"},
	{0xFCDEC75CCF7AF1CEULL, "tools_way_pt03.mmh"},
	{0xFCDEE15CCF7B1DBCULL, "tools_way_pt03.msh"},
	{0xFCDEE88D6B7EC6D8ULL, "741356909.ogg"},
	{0xFCDEF957E8942B49ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basic.phy"},
	{0xFCDF35AB3BF5EE1DULL, "6174830_m.fxe"},
	{0xFCDF4FA70BC6D932ULL, "vxht_chantry_n_206_0_0.vlm"},
	{0xFCE087A04A40C3D4ULL, "gen00fl_carver6097266.cl"},
	{0xFCE09F2AE8536737ULL, "dae000_gamlen6139539.cl"},
	{0xFCE0B4CC1470C162ULL, "vxw_woundedcoast_h7_0_0.vlm"},
	{0xFCE0BCCF88519A03ULL, "6075784_m.fxe"},
	{0xFCE163B25B46126BULL, "hlac_w_sundermount_-131072_0y.dds"},
	{0xFCE18464E2AC784FULL, "vxht_mansion2b_2f2_0_0.vlm"},
	{0xFCE188B87815EB4BULL, "hlac_lt_0_lowtown_a2_65533_2z.dds"},
	{0xFCE1B689D87710CEULL, "art\\levels\\races\\elf\\props\\dalish\\prp_aravell.phy"},
	{0xFCE23356495DBF74ULL, "hlac_lt_0_lowtown_docks_n_131070_1z.dds"},
	{0xFCE238462F16BCCEULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\blendtree\\abomination_overlay.anb"},
	{0xFCE24CCF58BFA984ULL, "dae000_arishok6075345.cl"},
	{0xFCE29A435948A957ULL, "qun210cr_guard6023947.cl"},
	{0xFCE2F0BF606054DDULL, "443403043.ogg"},
	{0xFCE35CCFBFB32D80ULL, "w_woundedcoast_26_0.anb"},
	{0xFCE399A2F775AA04ULL, "qun160_vashothlead6064597.cl"},
	{0xFCE3F070D4214208ULL, "vxlt_0_lowtown_a2_n_c_i0e_0_0.vlm"},
	{0xFCE47071D5A4BBB1ULL, "hlac_lt_0_lowtown_a2_n_c_65536_11x.dds"},
	{0xFCE4A967D0B5163BULL, "vxw_woundedcoast_yi_0_0.vlm"},
	{0xFCE4B475AAB40AC1ULL, "hlac_lt_gallowstemplar_2388_0z.dds"},
	{0xFCE5024FD5453C19ULL, "131439530.ogg"},
	{0xFCE5266F0B6103FFULL, "6162534_m.fxe"},
	{0xFCE5425AEB8A4BF9ULL, "155345273.ogg"},
	{0xFCE5CAD61F3DA990ULL, "508906769.ogg"},
	{0xFCE60EB94AD2DB4FULL, "335754772.ogg"},
	{0xFCE62924E17E2197ULL, "dae000_arishok6072891.cl"},
	{0xFCE6458CEFBE39F4ULL, "84450313.ogg"},
	{0xFCE65B7C806C5BACULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v4.ani"},
	{0xFCE676330EB16808ULL, "llac_w_encounter3_0_1.dds"},
	{0xFCE74418B52BE910ULL, "art\\characters\\creatures\\spider\\animation\\appr.ani"},
	{0xFCE799085FDC8D62ULL, "6185136_m.fxe"},
	{0xFCE7C0A25A88573DULL, "vxw_woundedcoast_1bo_0_0.vlm"},
	{0xFCE7ECF7376DE4E3ULL, "6165333_m.fxe"},
	{0xFCE8030C2C0026EFULL, "dae000ip_crowd_lm2_talk9.utp"},
	{0xFCE819CAD1355775ULL, "hlac_lt_0_lowtown_docks_n_65534_2y.dds"},
	{0xFCE824F41A9C3C17ULL, "vxw_encounter2_camp_ob_0_0.vlm"},
	{0xFCE87408F3EBABEFULL, "vxlt_0_lowtown_docks_n_c4k_0_0.vlm"},
	{0xFCE88104420017B2ULL, "650720066.ogg"},
	{0xFCE943606B729899ULL, "6110453_f.fxe"},
	{0xFCE96B44EFA6A982ULL, "vxlt_0_lowtown_a2_n_c_bxt_0_0.vlm"},
	{0xFCE9F1423992E426ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_2.evt"},
	{0xFCEA1F6B47742555ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_bench_forward_3_enter.gad"},
	{0xFCEA35F6E7FF1AA8ULL, "art\\characters\\playercharacter\\elfmale\\arm_hvyo.phy"},
	{0xFCEB6CFEC55ACA68ULL, "lt_0_lowtown_docks_202_0.msh"},
	{0xFCEB7EFEC55AE8F9ULL, "lt_0_lowtown_docks_202_0.mao"},
	{0xFCEB8AFEC55AFC92ULL, "lt_0_lowtown_docks_202_0.mmh"},
	{0xFCEC22CD996B2B53ULL, "6062371_m.fxe"},
	{0xFCEC34EEB7C28B64ULL, "6072698_m.fxe"},
	{0xFCEC5458971A4DFCULL, "dae000ip_crowd_ff1_talk17.utp"},
	{0xFCEC71095281975BULL, "genev_tutorial.ncs"},
	{0xFCED687C807290EEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v4.gad"},
	{0xFCEDC9F362F2DEA3ULL, "hlac_lt_0_lowtown_docks_n_-65538_6x.dds"},
	{0xFCEDDF611E4E0516ULL, "6090304_m.fxe"},
	{0xFCEE4D8C9DEB8884ULL, "288040617.ogg"},
	{0xFCEF26A90666D9E2ULL, "vxw_woundedcoast_e0_0_0.vlm"},
	{0xFCEF45DB3840D921ULL, "335126166.ogg"},
	{0xFCEF50F3CB14058DULL, "hlac_lt_0_lowtown_docks_a3_n_-2_4x.dds"},
	{0xFCEF6C0778F31BC7ULL, "rdr111dg_assistant6055109.cl"},
	{0xFCEF78A7860533DEULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed.ani"},
	{0xFCEFBD26B53BB8E7ULL, "llac_w_bonepitdestroyed_a3_196614_0.dds"},
	{0xFCEFC719F7E0C5F4ULL, "zz_isa_debug6133171.cl"},
	{0xFCF0084239977771ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_2.gad"},
	{0xFCF063BDE81088A2ULL, "dae200_cullen6140337.cl"},
	{0xFCF079E3B6F1B671ULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishcutting.msh"},
	{0xFCF087E3B6F1CD8CULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishcutting.mao"},
	{0xFCF093E3B6F1E22FULL, "art\\levels\\races\\ferelden\\props\\food\\prp_fishcutting.mmh"},
	{0xFCF0AABCC97E0044ULL, "follower_banter6161010.cl"},
	{0xFCF0E0EB9AC393FAULL, "1052914422.ogg"},
	{0xFCF1D3533597E52CULL, "gen00fl_anders6125642.cl"},
	{0xFCF22E0F08F44F12ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_3.gad"},
	{0xFCF2A0A07ECA1253ULL, "244705393.ogg"},
	{0xFCF2D9C0C8EED1FEULL, "vxlt_gallowstemplar_n_37i_0_0.vlm"},
	{0xFCF31B44C2AB2EBBULL, "vxw_woundedcoast_vh_0_0.vlm"},
	{0xFCF4637C8078A6CDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_attk_v4.evt"},
	{0xFCF4FC1BBCC5FBFEULL, "vxlt_gallowstemplar_f_3se_0_0.vlm"},
	{0xFCF53582E6ADAAABULL, "937503770.ogg"},
	{0xFCF5B885239E21FFULL, "vxca_caverns2_8kq_0_1.vlm"},
	{0xFCF5D5B60C63FE87ULL, "vxlt_darktown_5cq_0_0.vlm"},
	{0xFCF6327F4C7E6FBDULL, "vxw_woundedcoast_1mn_0_0.vlm"},
	{0xFCF6F2CB22CC2F92ULL, "dae300_alistair_king6133550.cl"},
	{0xFCF70942399D971FULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\rightarm_point_forward_2.ani"},
	{0xFCF7768BA10DBA96ULL, "mcr310_party_banter6087779.cl"},
	{0xFCF832A34E05B924ULL, "180614859.ogg"},
	{0xFCF88F14E0E20DE3ULL, "vxw_encounter2_n5_0_0.vlm"},
	{0xFCF892C39DB84E37ULL, "353090823.ogg"},
	{0xFCF9083820F173AFULL, "llac_lt_0_lowtown_a2_n_-196608_2.dds"},
	{0xFCF9578207B32EB9ULL, "mag110_vincento6016323.cl"},
	{0xFCF99C3178191F77ULL, "vxw_woundedcoast_gc_0_0.vlm"},
	{0xFCF9DB6C2C7F7281ULL, "gen00fl_fenris6113967.cl"},
	{0xFCFA07952BAE423FULL, "one000_ilen6205706.cl"},
	{0xFCFBBC2CC36411B6ULL, "6166287_m.fxe"},
	{0xFCFC1AAD07F0636FULL, "follower_banter6151360.cl"},
	{0xFCFC4496C5B29597ULL, "gen00fl_varric6105786.cl"},
	{0xFCFC90A7860FE6AFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed.evt"},
	{0xFCFCC811C98A4979ULL, "6027715_f.fxe"},
	{0xFCFCCD936D2B1F27ULL, "gen00fl_anders6125333.cl"},
	{0xFCFD18970B469EE4ULL, "vxht_mansion_n_57n_0_0.vlm"},
	{0xFCFD29E73A0D8DC8ULL, "6105493_m.fxe"},
	{0xFCFE0A6D84CA8380ULL, "vxlt_0_lowtown_a2_epa_0_1.vlm"},
	{0xFCFE802C8C2423F0ULL, "6096821_m.fxe"},
	{0xFCFF3F69C5AE2772ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3ww_broken_1.mao"},
	{0xFCFF681BDE5A62D4ULL, "6033316_m.fxe"},
	{0xFCFF779943EEA6E3ULL, "vxlt_undercitydungeon_1ju_0_0.vlm"},
	{0xFCFF945EB620DF79ULL, "133076586.ogg"},
	{0xFCFFBFB9FEFF0FB6ULL, "gen00fl_aveline6139075.cl"},
	{0xFD001A0F2A80E21CULL, "lgt160_wk_soph_sophia6119769.cl"},
	{0xFD00927F8D93BE56ULL, "6090872_m.fxe"},
	{0xFD013B4FF93EA497ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hn_glv_lgta_0d.dds"},
	{0xFD013F6205351170ULL, "21383863.ogg"},
	{0xFD018D21B4A1473BULL, "vxw_woundedcoast_sk_0_0.vlm"},
	{0xFD021138F557BCD6ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_exit.ani"},
	{0xFD02446AC09F9B21ULL, "fen210_orana6109764.cl"},
	{0xFD029F9BC2E420B5ULL, "zz_vrc_debug6200223.cl"},
	{0xFD02A798A72B6DD3ULL, "follower_banter6164041.cl"},
	{0xFD02F340CBDBC901ULL, "6031746_m.fxe"},
	{0xFD03254F3A63F55FULL, "14854483.ogg"},
	{0xFD03805B888A7C54ULL, "263867747.ogg"},
	{0xFD0391A786160758ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\animation\\dr1a_stonedefault.opentoclosed.gad"},
	{0xFD039A8AB87A6535ULL, "6119246_m.fxe"},
	{0xFD043171FDBCCE63ULL, "vxlt_0_lowtown_a3_n_c_hck_0_0.vlm"},
	{0xFD04476586ED54ADULL, "6088319_m.fxe"},
	{0xFD047AB66B4A60A8ULL, "w_sundermount_7_0.mmh"},
	{0xFD0486B66B4A754BULL, "w_sundermount_7_0.mao"},
	{0xFD0498B66B4A93D2ULL, "w_sundermount_7_0.msh"},
	{0xFD04A45C3E74883DULL, "vxw_woundedcoast_1hm_0_0.vlm"},
	{0xFD050D55136F61CFULL, "hlac_lt_0_lowtown_65538_1x.dds"},
	{0xFD054C2C62E8270EULL, "231702740.ogg"},
	{0xFD0579DE7FE32B90ULL, "gen_im_trs_slv_lgt_glv_r1.uti"},
	{0xFD05BFC4269C99DCULL, "hlac_w_encounter3_-2_0x.dds"},
	{0xFD062AD6501561B1ULL, "hlac_lt_0_lowtown_docks_n_-2_1z.dds"},
	{0xFD064C7B876E83FAULL, "6148382_m.fxe"},
	{0xFD067ACDC4323ADBULL, "6068438_m.fxe"},
	{0xFD07190F090687B8ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\varric\\varric_3.ani"},
	{0xFD07372053521CDFULL, "ser270_templar6164148.cl"},
	{0xFD07A3E6A9EA7223ULL, "follower_banter6166674.cl"},
	{0xFD07D5A055D8F2B3ULL, "155360740.ogg"},
	{0xFD07E4B5ED4C9F78ULL, "pou_cl.mao"},
	{0xFD0808AEA57A3D9AULL, "207198746.ogg"},
	{0xFD085CABF6860656ULL, "6126539_m.fxe"},
	{0xFD08EBF309B139DDULL, "genip_trap_gas.utp"},
	{0xFD099B288E45993AULL, "gen00fl_aveline6114856.cl"},
	{0xFD09C6B02D6169F4ULL, "vxht_playershouse_n_1p7_0_1.vlm"},
	{0xFD0A019600C4CEB2ULL, "mag211_marethari6021275.cl"},
	{0xFD0AC554E76A3A79ULL, "dae150_radha6148933.cl"},
	{0xFD0B12B35396513CULL, "hlac_lt_gallowstemplar_2384_2y.dds"},
	{0xFD0B296E6119D837ULL, "qcr220_grey_wardens6135944.cl"},
	{0xFD0BC309339F9F65ULL, "vxht_mansion2_1v7_0_0.vlm"},
	{0xFD0BE257E8BAF26EULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basic.mmh"},
	{0xFD0BEE57E8BB0605ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basic.mao"},
	{0xFD0BFC57E8BB1E5CULL, "art\\levels\\races\\proxy\\tilesets\\setint\\benstestset\\0ww_basic.msh"},
	{0xFD0C05C1492F8C54ULL, "gen_im_trs_gen_dua_dag_r2.uti"},
	{0xFD0C19CABD26A19FULL, "vxca_caverns2_aic_0_0.vlm"},
	{0xFD0C2AE480E7BCC5ULL, "follower_barks6153756.cl"},
	{0xFD0C920CA86D7022ULL, "6067679_f.fxe"},
	{0xFD0CB074BB62D7D5ULL, "dark_fog_g.mao"},
	{0xFD0CFA2D93BF8719ULL, "vxlt_undercitydungeon_1i2_0_0.vlm"},
	{0xFD0D7A61DECD121FULL, "hlac_lt_0_lowtown_n_131072_4y.dds"},
	{0xFD0DE97635E4BF63ULL, "vxlt_undercitydungeon_1mt_0_0.vlm"},
	{0xFD0E6A3453BAFBFEULL, "6163781_m.fxe"},
	{0xFD0E8C975B2A2B90ULL, "489962006.ogg"},
	{0xFD0F045C7F89D6D6ULL, "6095772_m.fxe"},
	{0xFD0F05715E61D4D9ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb107a.mmh"},
	{0xFD0F0F715E61E5C7ULL, "art\\characters\\weapons\\shortbows\\w_sbw_sb107a.msh"},
	{0xFD0F1938F56254C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_exit.evt"},
	{0xFD0F59E6FAAE4E9EULL, "isa211_isabela6072545.cl"},
	{0xFD0FC13202D5B041ULL, "403801809.ogg"},
	{0xFD0FDE1A67A2B8E4ULL, "6000981_m.fxe"},
	{0xFD104B299995ABF3ULL, "zz_mag_debug6024520.cl"},
	{0xFD1063F23279BAF9ULL, "die_cl_f.mao"},
	{0xFD107ADAF051D902ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.mmh"},
	{0xFD109AE2639C0BC9ULL, "hlac_dr_dungeon_3216_0y.dds"},
	{0xFD10CCEC74CA0FEAULL, "479481702.ogg"},
	{0xFD11AC2B18965DD6ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.cess.ani"},
	{0xFD121495443FE1CFULL, "fx_dis_noise01.mao"},
	{0xFD125F035B87766BULL, "static_lava_p0.vsh.fxo.dx11.assembly"},
	{0xFD1271A182C8C0B5ULL, "vxht_mansion_n_6iv_0_0.vlm"},
	{0xFD12A443F4835D5AULL, "llac_ht_0_hightown_n_65534_1.dds"},
	{0xFD12B96FF0502F87ULL, "vxlt_darktown_5is_0_0.vlm"},
	{0xFD1384F21D7BA407ULL, "6107694_m.fxe"},
	{0xFD13B276B059632BULL, "vxlt_0_lowtown_docks_n_b5m_0_0.vlm"},
	{0xFD13D115F794A953ULL, "dre000_dougal6129809.cl"},
	{0xFD14303E35085F72ULL, "hlac_ca_caverns2prc_5802_13x.dds"},
	{0xFD14525E94ECEC81ULL, "dre000_carver6037596.cl"},
	{0xFD14DD94A8DCC659ULL, "gen00fl_merrill_act1_pt3.stg"},
	{0xFD152D22A1ADE6D8ULL, "vxlt_0_lowtown_eo0_0_1.vlm"},
	{0xFD155B7A966D8A63ULL, "6172805_m.fxe"},
	{0xFD159E048E1B7B8CULL, "766210318.ogg"},
	{0xFD162FBE6BBDB7C7ULL, "zz_pro_debug6059740.cl"},
	{0xFD163A38F568AB10ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\custom\\qunari_sit_back_throne_exit.gad"},
	{0xFD16473D520F016EULL, "hlac_lt_gallowstemplar_f_2383_10z.dds"},
	{0xFD16E2F300E8AF47ULL, "6166202_m.fxe"},
	{0xFD16EBC3A04C009DULL, "bm_leak_drop.dds"},
	{0xFD17182C7A07DFC1ULL, "art\\characters\\creatures\\dragon\\animation\\kd_exit.ani"},
	{0xFD17452142D3DE73ULL, "w_encounter2_camp_59_0.rml"},
	{0xFD1758BC8B33E4B5ULL, "mcr310_argument6072691.cl"},
	{0xFD17763365FB931EULL, "vxlt_undercityencounter_1na_0_0.vlm"},
	{0xFD18082E528E3230ULL, "6115276_m.fxe"},
	{0xFD1816CA660106EDULL, "6108002_m.fxe"},
	{0xFD18B45C97BB9A3FULL, "mer220_hubert6096319.cl"},
	{0xFD19024E31BC0A43ULL, "777594794.ogg"},
	{0xFD193174B757FC51ULL, "dae000ip_crowd_ff8_talk11.utp"},
	{0xFD199D1730A755F7ULL, "vxw_bonepit_1qk_0_0.vlm"},
	{0xFD1A7C3FDC492362ULL, "vxw_woundedcoast_t3_0_0.vlm"},
	{0xFD1AE9C53F43A55AULL, "6162777_m.fxe"},
	{0xFD1BD0C24CF6B24BULL, "631093197.ogg"},
	{0xFD1BEDBF3F290C5BULL, "vxlt_0_lowtown_byk_0_0.vlm"},
	{0xFD1C02837FDD7012ULL, "hlac_w_sundermount_393210_0y.dds"},
	{0xFD1CF5D9D2817CBFULL, "hlac_lt_0_lowtown_a3_n_c_-262144_3y.dds"},
	{0xFD1D1E29552AEAE4ULL, "gen00fl_aveline6110490.cl"},
	{0xFD1D1F075CA6D0C6ULL, "6075983_m.fxe"},
	{0xFD1D2EF2F55DB289ULL, "w_encounter2_38_0_lc.mmh"},
	{0xFD1D3CDCC2BF2189ULL, "art\\characters\\playercharacter\\humanfemale\\rob_chaa.phy"},
	{0xFD1D4C2142D855CBULL, "w_encounter2_camp_59_0.phy"},
	{0xFD1D55A96407C9A8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winbrick.mmh"},
	{0xFD1D598D2C682E56ULL, "848275440.ogg"},
	{0xFD1D61A96407DE4BULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winbrick.mao"},
	{0xFD1D73A96407FCD2ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9ww_winbrick.msh"},
	{0xFD1D8B5B13108FEAULL, "vxlt_0_lowtown_eqh_0_1.vlm"},
	{0xFD1DA5C121BCF952ULL, "hlac_ht_mansion_n_3374_1z.dds"},
	{0xFD1E0CDC2324DA13ULL, "hlac_w_bonepitdestroyed_a3_0_3z.dds"},
	{0xFD1E232C7A0E1163ULL, "art\\characters\\creatures\\dragon\\animation\\kd_exit.gad"},
	{0xFD1E3D43E5EB4F58ULL, "vxca_cavern1_haunted_e56_0_0.vlm"},
	{0xFD1E4AE67E908B1CULL, "vxht_brothel_n_4vw_0_0.vlm"},
	{0xFD1EB42B18A0F5C7ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.cess.evt"},
	{0xFD1EEAAA68651043ULL, "63816872.ogg"},
	{0xFD1EEE652E15CA27ULL, "w_encounter2_camp_8_0.rml"},
	{0xFD1F3CC4E4F48D09ULL, "6040621_m.fxe"},
	{0xFD1F41AEFB2B90E0ULL, "vxht_mansion_n_4aa_0_0.vlm"},
	{0xFD1F5DB7B195FBC5ULL, "ave310_donnic6135897.cl"},
	{0xFD1F8564BB3FF7DEULL, "mag130_danzig6018826.cl"},
	{0xFD1FEC74DEFADA27ULL, "follower_banter6161208.cl"},
	{0xFD20360CCD85C757ULL, "gen00fl_anders6112964.cl"},
	{0xFD2057D5C56E8BD1ULL, "vxlt_0_lowtown_docks_n_bfj_0_0.vlm"},
	{0xFD20E37E74BED935ULL, "vxht_mansion_n_6lu_0_0.vlm"},
	{0xFD217AC97BFCAEC9ULL, "snow_04_f.dds"},
	{0xFD21879C26F4AFEFULL, "lt_0_lowtown_docks_n_41_0.rml"},
	{0xFD21B589D8ACF291ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_aravell.mmh"},
	{0xFD21BAE4AFBD259CULL, "vxht_keep_n_5xk_0_0.vlm"},
	{0xFD21C189D8AD0732ULL, "art\\levels\\races\\elf\\props\\dalish\\prp_aravell.mao"},
	{0xFD21CF89D8AD1F4FULL, "art\\levels\\races\\elf\\props\\dalish\\prp_aravell.msh"},
	{0xFD21DF0ED7A9F5E7ULL, "6082198_f.fxe"},
	{0xFD2222400C24C32CULL, "mag291cr_vincento.utc"},
	{0xFD223273434D2CCAULL, "6173667_m.fxe"},
	{0xFD227B391568F8F4ULL, "art\\vfx\\textures\\vfx_rain.dds"},
	{0xFD227E0C2E3EDF95ULL, "vxht_mansion2_1mb_0_0.vlm"},
	{0xFD22CCE9BD49C133ULL, "425575076.ogg"},
	{0xFD22D1B10F89B774ULL, "vxlt_gallowstemplar_f_4st_0_0.vlm"},
	{0xFD2355028FE9B5CCULL, "6116505_m.fxe"},
	{0xFD23D757BB048717ULL, "vxlt_0_lowtown_docks_co9_0_0.vlm"},
	{0xFD23E162010BE907ULL, "14794543.ogg"},
	{0xFD242E98AAC22C23ULL, "vxdr_primevalentrance_1q5_0_0.vlm"},
	{0xFD243BCFBFEA8CA8ULL, "w_woundedcoast_26_0.rml"},
	{0xFD24E5792CB4ADE7ULL, "vxht_bartrand_n_65k_0_0.vlm"},
	{0xFD25002C7A13F448ULL, "art\\characters\\creatures\\dragon\\animation\\kd_exit.evt"},
	{0xFD2502431FA0CDD9ULL, "532522563.ogg"},
	{0xFD25392C88577BD0ULL, "ts_kirk_hightown_02.gda"},
	{0xFD25A960E3695A98ULL, "459590233.ogg"},
	{0xFD25D52B18A74C10ULL, "art\\vfx\\combat\\vfx_c_pride_p_shield_c.cess.gad"},
	{0xFD25E81057F1AB9EULL, "vxlt_undercityencounter_1kb_0_0.vlm"},
	{0xFD25ED652E1BE73FULL, "w_encounter2_camp_8_0.phy"},
	{0xFD2605FB3BF99828ULL, "6151049_m.fxe"},
	{0xFD26438129604E0EULL, "6109055_m.fxe"},
	{0xFD268494FDE5ADB6ULL, "6055075_m.fxe"},
	{0xFD26BA089B11F854ULL, "cs_pro000cs_innblood_beth.wbd"},
	{0xFD2752A62EF872F0ULL, "vxht_0_hightown_a3_c0w_0_0.vlm"},
	{0xFD27932A0A5A1F8EULL, "w_bonepitmining_a2_181_0.mmh"},
	{0xFD279BA1E06FAB92ULL, "6061331_m.fxe"},
	{0xFD279F2A0A5A3425ULL, "w_bonepitmining_a2_181_0.mao"},
	{0xFD27AD2A0A5A4C7CULL, "w_bonepitmining_a2_181_0.msh"},
	{0xFD280EF4229D6E77ULL, "vxw_bonepit_1zj_0_0.vlm"},
	{0xFD285FA425A0F880ULL, "dae000_mother6136912.cl"},
	{0xFD28901FB4319122ULL, "hlac_lt_0_lowtown_docks_a3_n_65534_5x.dds"},
	{0xFD28B426650D6E31ULL, "428308666.ogg"},
	{0xFD29073ED8852D44ULL, "dae000_tomwise6167311.cl"},
	{0xFD2A2378BA340524ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0ww_archdiv.phy"},
	{0xFD2A39FC1802AF9AULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cust.gad"},
	{0xFD2B29954753A779ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16man.msh"},
	{0xFD2B33954753B867ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16man.mmh"},
	{0xFD2B37954753BEB4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp16man.mao"},
	{0xFD2BCB2334D6296EULL, "680427256.ogg"},
	{0xFD2BE69D51AB6BE9ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh.phy"},
	{0xFD2BF8BF84292D5FULL, "hlac_lt_0_lowtown_docks_a3_n_0_0y.dds"},
	{0xFD2C296405BE7D27ULL, "llac_dr_0_deeproads_5849_10.dds"},
	{0xFD2C5F78662CFA0DULL, "325181346.ogg"},
	{0xFD2C67B2B558E22AULL, "art\\levels\\races\\natural\\proxy\\props\\_textures\\small_sa.dds"},
	{0xFD2D0F07D6467AA4ULL, "lt_0_lowtown_docks_a3_119_0.msh"},
	{0xFD2D2107D646990DULL, "lt_0_lowtown_docks_a3_119_0.mao"},
	{0xFD2D2507D646A046ULL, "lt_0_lowtown_docks_a3_119_0.mmh"},
	{0xFD2DB38BED21A960ULL, "vxht_mansion_n_4dn_0_0.vlm"},
	{0xFD2E57CAB662388DULL, "llac_lt_0_lowtown_-2_1.dds"},
	{0xFD2E6D96E8BC569BULL, "241832319.ogg"},
	{0xFD2EC9B2B764A451ULL, "vxlt_0_lowtown_docks_n_bci_0_0.vlm"},
	{0xFD2F566ED83E83A4ULL, "art\\characters\\weapons\\dagger\\w_dger_108a.msh"},
	{0xFD2F57048AB8F2A8ULL, "175091700.ogg"},
	{0xFD2F59BAA438A5D2ULL, "565128579.ogg"},
	{0xFD2F640B3F56F390ULL, "6183976_m.fxe"},
	{0xFD2F6C6ED83EA946ULL, "art\\characters\\weapons\\dagger\\w_dger_108a.mmh"},
	{0xFD2FFFE45F25963BULL, "vxw_bonepit_1nr_0_0.vlm"},
	{0xFD3056C23ED5BF23ULL, "gen00fl_anders6125838.cl"},
	{0xFD30A400329CD02DULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_door_1.mao"},
	{0xFD30B71A75A8EF81ULL, "vxlt_undercityserialkiller_258_0_0.vlm"},
	{0xFD30D4742F4196AAULL, "6167150_m.fxe"},
	{0xFD3116304CF75489ULL, "241070896.ogg"},
	{0xFD311D83EAE91EF9ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzbigboat.msh"},
	{0xFD312783EAE92FE7ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzbigboat.mmh"},
	{0xFD31294DFABAB6E1ULL, "dae000_hate_enemies6167496.cl"},
	{0xFD312B83EAE93634ULL, "art\\levels\\races\\proxy\\props\\vehicles\\prp_zzbigboat.mao"},
	{0xFD313CFC1808D3B1ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cust.evt"},
	{0xFD31438E017FCFF4ULL, "vxlt_gallowstemplar_f_4xu_0_0.vlm"},
	{0xFD31BD803F7315C4ULL, "668990392.ogg"},
	{0xFD3217D7FC0F713AULL, "vxdr_primevalentrance_2oh_0_1.vlm"},
	{0xFD3229EA7293C044ULL, "274427927.ogg"},
	{0xFD32467608BAD474ULL, "carpetnew_n.dds"},
	{0xFD32B1893DFDC1FDULL, "vxlt_gallowscourtyard_6xw_0_0.vlm"},
	{0xFD33585D7F9A89EDULL, "follower_banter6156266.cl"},
	{0xFD339E82BCFAEAF2ULL, "character_character_p76.psh.fxo.dx11.assembly"},
	{0xFD33F22FC3CAF189ULL, "412427335.ogg"},
	{0xFD34241240B2696FULL, "6147856_m.fxe"},
	{0xFD34950B910C649AULL, "w_bonepitdestroyed_a3_129_0.mmh"},
	{0xFD34990B910C6BD1ULL, "w_bonepitdestroyed_a3_129_0.mao"},
	{0xFD34A70B910C83E0ULL, "w_bonepitdestroyed_a3_129_0.msh"},
	{0xFD3531F07A109246ULL, "hlac_w_encounter2_-1_0z.dds"},
	{0xFD35791B43B2C296ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_spi_p.anb"},
	{0xFD357CAFB88A9FD9ULL, "character_character_p169.psh.fxo.dx11"},
	{0xFD35A85DCF42A3D2ULL, "zz_maxhealth.ncs"},
	{0xFD35C48320EE8B70ULL, "vxht_0_hightown_a3_c5t_0_0.vlm"},
	{0xFD35C48838AB7BCFULL, "6089093_m.fxe"},
	{0xFD36A38099DE0D33ULL, "vxw_woundedcoast_13r_0_0.vlm"},
	{0xFD36E34782CAAC4FULL, "654353663.ogg"},
	{0xFD36EE652E2A5C44ULL, "w_encounter2_camp_8_0.mmh"},
	{0xFD36F2652E2A6297ULL, "w_encounter2_camp_8_0.mao"},
	{0xFD3704652E2A81E6ULL, "w_encounter2_camp_8_0.msh"},
	{0xFD3718A00981E8E5ULL, "vxht_mansion2_1g4_0_0.vlm"},
	{0xFD3775E58450DAD9ULL, "600151846.ogg"},
	{0xFD3783C58CCB2FCCULL, "493572774.ogg"},
	{0xFD37AF63E2338E12ULL, "qun221_petrice6077329.cl"},
	{0xFD380E4165EB8282ULL, "636314461.ogg"},
	{0xFD381C07D6509811ULL, "lt_0_lowtown_docks_a3_119_0.phy"},
	{0xFD383B7D8626DB89ULL, "llac_ca_caverns2_5802_2.dds"},
	{0xFD3848EAE70266C7ULL, "14746409.ogg"},
	{0xFD384F5F8D120B45ULL, "681911814.ogg"},
	{0xFD3864A3089419CDULL, "vxht_mansion_n_60y_0_0.vlm"},
	{0xFD3879DAF07329A9ULL, "art\\vfx\\combat\\vfx_r_asn_mdeath_tag_loop_c.anb"},
	{0xFD38A6580B04C54EULL, "58570890.ogg"},
	{0xFD3982B2D06F38E3ULL, "6168529_m.fxe"},
	{0xFD39AAC2902365F7ULL, "rmtst_probe_3679.mtx"},
	{0xFD39CFFC9B006B9BULL, "ser221_gascard6164254.cl"},
	{0xFD39D18F323FFC4EULL, "hlac_lt_0_lowtown_docks_n_-196611_0x.dds"},
	{0xFD39F5FBE3E55BE6ULL, "sib131cr_slaver_b.utc"},
	{0xFD3A636ED848A111ULL, "art\\characters\\weapons\\dagger\\w_dger_108a.phy"},
	{0xFD3A97F86989397BULL, "vxlt_gallowstemplar_n_3qr_0_0.vlm"},
	{0xFD3B7D324C4C79CFULL, "llac_lt_hangedman_-1_2.dds"},
	{0xFD3BD0AFEBD4E656ULL, "hlac_lt_0_lowtown_docks_a3_-196608_0x.dds"},
	{0xFD3C89A4A7CA8116ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_inquisitionsoldier2_.ani"},
	{0xFD3D96A6F38E7187ULL, "mag110_arianni6036798.cl"},
	{0xFD3DF6E51F99A615ULL, "572949278.ogg"},
	{0xFD3E0331ADE63C62ULL, "mrl000_marethari6129182.cl"},
	{0xFD3ED4EA42EEC1AEULL, "art\\characters\\playercharacter\\humanmale\\bdy_medp.msh"},
	{0xFD3EE2EA42EED9FFULL, "art\\characters\\playercharacter\\humanmale\\bdy_medp.mao"},
	{0xFD3EEEEA42EEED9CULL, "art\\characters\\playercharacter\\humanmale\\bdy_medp.mmh"},
	{0xFD3EFBDFC2A24092ULL, "follower_banter6151305.cl"},
	{0xFD3F128B6C66D6C7ULL, "dae000ip_crowd_mf2_talk14.utp"},
	{0xFD3F406A120603B2ULL, "and230_mage_girl6135823.cl"},
	{0xFD3F54FC18153980ULL, "art\\vfx\\combat\\vfx_c_rw_charge_c.cust.ani"},
	{0xFD3F64DE8977CC55ULL, "gencr_city_guard_commander.utc"},
	{0xFD4044D8A834C327ULL, "w_encounter2_59_0.rml"},
	{0xFD405507280C333AULL, "zz_curio_test.lst"},
	{0xFD40A0E1C32E9EEBULL, "460629837.ogg"},
	{0xFD40E5ADEA6BC7BCULL, "vxlt_undercitydungeon_1fj_0_0.vlm"},
	{0xFD40F4AA97695D89ULL, "6110582_m.fxe"},
	{0xFD4109336FD879FAULL, "nat320_delilah6065946.cl"},
	{0xFD4112528EAE5D23ULL, "vxdr_primevalentrance_1k7_0_0.vlm"},
	{0xFD418785F876C26EULL, "gen_im_wep_war_shd_act2_02.uti"},
	{0xFD422B7E3806BC12ULL, "6155955_m.fxe"},
	{0xFD42502CA0A05741ULL, "clf260_harley6069724.cl"},
	{0xFD425A522A13051EULL, "6116719_f.fxe"},
	{0xFD4295EF32A881EFULL, "6140956_m.fxe"},
	{0xFD437C1B43BF0501ULL, "art\\vfx\\combat\\vfx_m_amg_estorm_spi_p.mmh"},
	{0xFD4476BF1AE472F4ULL, "mag101_grace6034038.cl"},
	{0xFD44B0A0F415E6A9ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperright.mao"},
	{0xFD44BCA0F415FB42ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperright.mmh"},
	{0xFD44BEA0F415FEF8ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\8bi_2upperright.msh"},
	{0xFD4565C41AF2198DULL, "follower_barks6153918.cl"},
	{0xFD45C4DF829FB23EULL, "vxca_cavern1_haunted_iyx_0_0.vlm"},
	{0xFD45ECAE59060870ULL, "mer341_jansen6082990.cl"},
	{0xFD464880EDDE8D1CULL, "vxlt_0_lowtown_docks_c4c_0_0.vlm"},
	{0xFD465DF0B4E09B1BULL, "ave200_donnic6061271.cl"},
	{0xFD46A30FB0C4CBD6ULL, "vxdr_0_deeproads_err_0_1.vlm"},
	{0xFD46B9E1FA17B2BDULL, "6179593_f.fxe"},
	{0xFD47048CFD512A65ULL, "6215184_m.fxe"},
	{0xFD47084A4E7DDD6FULL, "127124147.ogg"},
	{0xFD4740F6BA5B5EA2ULL, "6105561_m.fxe"},
	{0xFD4743D8A83AE03FULL, "w_encounter2_59_0.phy"},
	{0xFD474C122B2A5E4BULL, "vxht_0_hightown_a2_b00_0_0.vlm"},
	{0xFD4762D14C04B885ULL, "hlac_lt_0_lowtown_docks_131068_0z.dds"},
	{0xFD47E5378087394AULL, "w_bonepit_218_0.mao"},
	{0xFD47E93780873F99ULL, "w_bonepit_218_0.mmh"},
	{0xFD47F33780875087ULL, "w_bonepit_218_0.msh"},
	{0xFD48757EA6C975FFULL, "llac_ht_0_hightown_a3_65536_2.dds"},
	{0xFD4876692912B189ULL, "llac_w_bonepit_196612_0.dds"},
	{0xFD487B91C827E5CFULL, "249756785.ogg"},
	{0xFD488DC4E32CB362ULL, "dae000_flemeth6000989.cl"},
	{0xFD48A99C79D132B0ULL, "vxht_mansion2b_2e8_0_0.vlm"},
	{0xFD48D2C2B5481120ULL, "305778875.ogg"},
	{0xFD48E574C481E5ADULL, "gen_im_cns_tom_attribute_03.uti"},
	{0xFD4925DCC2E435AEULL, "art\\characters\\playercharacter\\humanfemale\\rob_chaa.mmh"},
	{0xFD4931DCC2E44A45ULL, "art\\characters\\playercharacter\\humanfemale\\rob_chaa.mao"},
	{0xFD493FDCC2E4619CULL, "art\\characters\\playercharacter\\humanfemale\\rob_chaa.msh"},
	{0xFD4943ECA38EB0D8ULL, "follower_banter6155664.cl"},
	{0xFD49584DBBD5EC7BULL, "hla_ht_0_hightown_n_65534_0y.dds"},
	{0xFD4A06A79C37278BULL, "6153268_m.fxe"},
	{0xFD4A1495084585C7ULL, "89990385.ogg"},
	{0xFD4AD3F52FA416AEULL, "6106630_m.fxe"},
	{0xFD4AEE8A7F97830AULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.gad"},
	{0xFD4B5CCB6ECF0D88ULL, "gen00fl_isabela6101457.cl"},
	{0xFD4BDBA3249F6D46ULL, "vxht_0_hightown_by2_0_0.vlm"},
	{0xFD4C7530195EB670ULL, "llac_ht_mansion2b_818_64.dds"},
	{0xFD4CCC2474DF8EA6ULL, "dae000_arishok6074968.cl"},
	{0xFD4CE3487B521B9EULL, "vxlt_0_lowtown_docks_a3_n_bn2_0_0.vlm"},
	{0xFD4D0ABBD54EC0F1ULL, "557086330.ogg"},
	{0xFD4D6E117B48EB5EULL, "129006961.ogg"},
	{0xFD4DFA0A27E4B9E0ULL, "vxlt_shop_k9_0_0.vlm"},
	{0xFD4E33E6956686B2ULL, "6068853_m.fxe"},
	{0xFD4E95F0A40A6520ULL, "vxw_bonepitdestroyed_a3_84_0_0.vlm"},
	{0xFD4E9A8235FCE39EULL, "vxca_caverns1_icu_0_2.vlm"},
	{0xFD4F578ADC61E03CULL, "vxlt_undercitydungeon_1ik_0_0.vlm"},
	{0xFD50374CF129E0DAULL, "vxht_bartrand_n_5wo_0_0.vlm"},
	{0xFD50C511A4BF3159ULL, "art\\characters\\playercharacter\\humanfemale\\boo_bda.mmh"},
	{0xFD50CF11A4BF4247ULL, "art\\characters\\playercharacter\\humanfemale\\boo_bda.msh"},
	{0xFD50D6CA9998F872ULL, "242789260.ogg"},
	{0xFD50D6F5564B655AULL, "6206965_m.fxe"},
	{0xFD50E26585246DB2ULL, "vxlt_0_lowtown_exn_0_0.vlm"},
	{0xFD5117D00E55305DULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0ww_grandfalll_3.mao"},
	{0xFD51807342EFE90EULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastgravel_da.dds"},
	{0xFD51821D539D2EAFULL, "vxlt_foundry_n_va_0_0.vlm"},
	{0xFD51B2A4A7DD2350ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_opening\\introshot3_inquisitionsoldier2_.gad"},
	{0xFD51D18A7F9D70A1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.evt"},
	{0xFD51F645EB4E4559ULL, "323147463.ogg"},
	{0xFD5345EA395C4BA7ULL, "mrl000_marethari6069609.cl"},
	{0xFD53873A7DCA3E33ULL, "vxw_woundedcoast_19t_0_0.vlm"},
	{0xFD53A8089B38C7EEULL, "cs_pro000cs_innblood_beth.bnk"},
	{0xFD53AB7A3B4544CEULL, "lgt000cr_hate_raider.utc"},
	{0xFD53E35A740645CAULL, "zz_followers_banter6158397.cl"},
	{0xFD53F0307FC059ACULL, "6147796_m.fxe"},
	{0xFD5403A85ABF6B44ULL, "vxlt_hangedman_n_264_0_0.vlm"},
	{0xFD54DA26FFB6E22FULL, "vxlt_gallowstemplar_f_5ri_0_0.vlm"},
	{0xFD5502067C92FFD9ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha01.msh"},
	{0xFD550C067C9310C7ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha01.mmh"},
	{0xFD551C3C9161E6EBULL, "6089527_m.fxe"},
	{0xFD55629EDBE6224CULL, "w_woundedcoast_100_0.msh"},
	{0xFD55689EDBE62C3EULL, "w_woundedcoast_100_0.mmh"},
	{0xFD55749EDBE640D5ULL, "w_woundedcoast_100_0.mao"},
	{0xFD560528915F90D7ULL, "6005291_m.fxe"},
	{0xFD5657FDFB5029FBULL, "540467225.ogg"},
	{0xFD566950BD4DCE36ULL, "157034080.ogg"},
	{0xFD56EDC9940ADFADULL, "character_character_p172.psh.fxo.dx11.assembly"},
	{0xFD578089C5426118ULL, "vxht_keep_5jz_0_0.vlm"},
	{0xFD57C0AB70BFD47CULL, "lt_foundry_n_320_lc.mmh"},
	{0xFD57CF9D51D07F0EULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh.mmh"},
	{0xFD57E99D51D0ABFCULL, "art\\characters\\playercharacter\\shared\\heads\\mt_bas_uh.msh"},
	{0xFD57F0EBA3778B00ULL, "6160841_m.fxe"},
	{0xFD5844D8A8495544ULL, "w_encounter2_59_0.mmh"},
	{0xFD5848D8A8495B97ULL, "w_encounter2_59_0.mao"},
	{0xFD585AD8A8497AE6ULL, "w_encounter2_59_0.msh"},
	{0xFD59873323D1592CULL, "hlac_lt_0_lowtown_docks_n_131068_0x.dds"},
	{0xFD59B354EB1D968BULL, "644505531.ogg"},
	{0xFD59BB935859CA3BULL, "follower_banter6163484.cl"},
	{0xFD5A1F498D46EDD1ULL, "vxlt_0_lowtown_docks_n_brn_0_0.vlm"},
	{0xFD5A28DEFE368248ULL, "6075981_m.fxe"},
	{0xFD5AC0CD3FE1105CULL, "lgt261ip_at_to_coast.utp"},
	{0xFD5AD73ADA150F2DULL, "gen00fl_aveline6114088.cl"},
	{0xFD5AD90E2DDFC437ULL, "t1_mul_wh2.tnt"},
	{0xFD5AED11CF1F50A7ULL, "418576640.ogg"},
	{0xFD5AF83F3F98FE70ULL, "hlac_lt_warehouse_n_0_8x.dds"},
	{0xFD5B6EBB7BEDC8FAULL, "lgt110_gxa_coa_spy6117527.cl"},
	{0xFD5C0A9A803001EAULL, "art\\vfx\\combat\\vfx_g_dual_att_back_c.anb"},
	{0xFD5C3E6A1EEA20FFULL, "6162765_m.fxe"},
	{0xFD5C4418921B0CBAULL, "6109378_m.fxe"},
	{0xFD5C713CD590E283ULL, "hlac_ht_0_hightown_n_0_1x.dds"},
	{0xFD5C8B1DEABC33CEULL, "ht_0_hightown_a2_52_0_lc.mmh"},
	{0xFD5C9D9CABD7354BULL, "131385646.ogg"},
	{0xFD5CB7A6DCED0570ULL, "243736953.ogg"},
	{0xFD5D12A96AA671DEULL, "llac_lt_0_lowtown_a2_n_c_0_6.dds"},
	{0xFD5D5123B6B88B13ULL, "87726836.ogg"},
	{0xFD5DC967CE57F8BCULL, "vxlt_undercitydungeon_1lh_0_0.vlm"},
	{0xFD5DC9E3D2397EF1ULL, "dae000_arishok6081675.cl"},
	{0xFD5DFF7BA8E0EDE3ULL, "vxw_blightlands_71_0_0.vlm"},
	{0xFD5EC98A7FA7EDF0ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_3.ani"},
	{0xFD5EF39F7452563AULL, "dae000_party_night_comment6201643.cl"},
	{0xFD5F2CE9D8CF8042ULL, "6156239_m.fxe"},
	{0xFD601328B87AF023ULL, "hlac_w_sundermount_-131073_0z.dds"},
	{0xFD6071D0FB863F9FULL, "560426736.ogg"},
	{0xFD6083568D52F2E4ULL, "mcr371_hostile_merrill6082147.cl"},
	{0xFD60BE95C39D6B14ULL, "seb221_flora_harimann6177183.cl"},
	{0xFD60FB52FB5ED301ULL, "100972486.ogg"},
	{0xFD612B023D2BC3EAULL, "936572330.ogg"},
	{0xFD61BE3A5650C576ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_crumble.msh"},
	{0xFD61C83A5650D674ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_crumble.mmh"},
	{0xFD61CC3A5650DC87ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_crumble.mao"},
	{0xFD61D667EA75D077ULL, "vxw_bonepit_1nf_0_0.vlm"},
	{0xFD61E613A81025B9ULL, "zz_ave_debug6134006.cl"},
	{0xFD624A16AFBF282CULL, "1074671.ogg"},
	{0xFD637815E31A42EEULL, "gen00fl_isabela6085760.cl"},
	{0xFD642571D479CC66ULL, "6087035_f.fxe"},
	{0xFD6460CEDB2EB271ULL, "433060867.ogg"},
	{0xFD64CDEA430F8567ULL, "art\\characters\\playercharacter\\humanmale\\bdy_medp.phy"},
	{0xFD65383617780898ULL, "ht_0_hightown_n_38_0.rml"},
	{0xFD6540F02AFB60F5ULL, "gen00fl_aveline6120771.cl"},
	{0xFD655407282B4DBDULL, "zz_curio_test.are"},
	{0xFD655E3EC962A749ULL, "vxdr_0_deeproads_cxp_0_0.vlm"},
	{0xFD6581CD6926EBA8ULL, "hlac_dr_0_deeproads_5838_3y.dds"},
	{0xFD658A0D4D456DCAULL, "vxw_sundermount_2ev_0_0.vlm"},
	{0xFD65BAC99122C893ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_001a_0n.dds"},
	{0xFD6612833B980264ULL, "vxw_encounter2_camp_qg_0_0.vlm"},
	{0xFD6640FD4CDFD180ULL, "vxlt_undercityserialkiller_15n_0_0.vlm"},
	{0xFD66837C2DF0CFD9ULL, "6013356_f.fxe"},
	{0xFD67A5E1534145C6ULL, "lgt110_gxa_coa_spy6117448.cl"},
	{0xFD6898E9C14059D5ULL, "art\\characters\\playercharacter\\elfmale\\rob_appa.phy"},
	{0xFD69313714E7500EULL, "art\\characters\\weapons\\longbows\\w_lbw_lw105a.msh"},
	{0xFD693C6562F92379ULL, "207516901.ogg"},
	{0xFD694B3714E77CFCULL, "art\\characters\\weapons\\longbows\\w_lbw_lw105a.mmh"},
	{0xFD696AB1D1C2D799ULL, "dae000_arishok6067855.cl"},
	{0xFD6ADE187538C9A0ULL, "vxw_sundermount_37o_0_0.vlm"},
	{0xFD6AFF529AC6E824ULL, "vxca_caverns2_74i_0_0.vlm"},
	{0xFD6B8974CDCCFB5CULL, "783602185.ogg"},
	{0xFD6B99D050E803D1ULL, "llac_ca_cavern1_haunted_24423_13.dds"},
	{0xFD6BD672A3F6AC24ULL, "6087096_m.fxe"},
	{0xFD6C147782667EB3ULL, "llac_lt_0_lowtown_n_-196608_0.dds"},
	{0xFD6C1AEB70A7F09FULL, "hlac_w_bonepitmining_a2_262145_0y.dds"},
	{0xFD6C20D9DD090806ULL, "836676752.ogg"},
	{0xFD6C391596236E03ULL, "lowtown_male_lowlife1_sta6164968.cl"},
	{0xFD6C3B44C04E113CULL, "vxlt_undercitydungeon_1wi_0_0.vlm"},
	{0xFD6C71589AD70663ULL, "vxw_blightlands_80_0_0.vlm"},
	{0xFD6CA99C8FB03459ULL, "hlac_lt_0_lowtown_n_-65534_0y.dds"},
	{0xFD6CF857563049D7ULL, "follower_banter6141703.cl"},
	{0xFD6D10D34EB449A8ULL, "6069416_m.fxe"},
	{0xFD6D414CD84641A1ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobd_0d.dds"},
	{0xFD6D76D06B1D725FULL, "vxw_bonepit_272_0_0.vlm"},
	{0xFD6D80331672FB46ULL, "341089036.ogg"},
	{0xFD6DB34E2C3F9092ULL, "gen00fl_bethany6098920.cl"},
	{0xFD6E197320238338ULL, "character_character_p113.psh.fxo.dx11"},
	{0xFD6E27B93FFBE8E2ULL, "vxca_cavern1_haunted_ei9_0_1.vlm"},
	{0xFD6E41A1B96AA6D6ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.dura.evt"},
	{0xFD6E8741F3EE46DAULL, "623425149.ogg"},
	{0xFD6EBD3E6739D105ULL, "421201797.ogg"},
	{0xFD6EDCAD2CDA57CAULL, "141150581.ogg"},
	{0xFD6EE833811A02B6ULL, "706073386.ogg"},
	{0xFD6F2F1E5174E0DBULL, "mrl000_marethari6142490.cl"},
	{0xFD6F5098D06DE17CULL, "art\\levels\\races\\elf\\props\\alienage\\prp_tikia_0.pwk"},
	{0xFD6F74CEA0C39BFDULL, "llac_ht_keep_n_3735_1.dds"},
	{0xFD6FD79FBBBB5025ULL, "llac_lt_shop_-1_1.dds"},
	{0xFD70DF9FD5E186C2ULL, "art\\characters\\weapons\\axes\\textures\\w_axe_ax101a_0s.dds"},
	{0xFD71296E9BDD420BULL, "6087158_m.fxe"},
	{0xFD7150863F735992ULL, "6122166_m.fxe"},
	{0xFD7155E15FFE709BULL, "dre000_sibling_has_blight6036554.cl"},
	{0xFD7174EAA13E7B10ULL, "6001227_m.fxe"},
	{0xFD71FFF6754E17EBULL, "6087727_m.fxe"},
	{0xFD721CE72F2DF554ULL, "mag311_keran.cnv"},
	{0xFD726B91B1E7D020ULL, "w_blightlands_150_0.phy"},
	{0xFD72BCF7F7CB6719ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v1.ani"},
	{0xFD731319B5E89B2FULL, "cod_bks_chantry_hist.ncs"},
	{0xFD73D61FF69F1795ULL, "6044812_m.fxe"},
	{0xFD73F3BD1613F08DULL, "llac_lt_0_lowtown_docks_a3_0_2.dds"},
	{0xFD73FBEA3F3B864AULL, "vxw_sundermount_2nw_0_0.vlm"},
	{0xFD74069B00F841F4ULL, "hlac_ca_caverns2_5802_52x.dds"},
	{0xFD7484602D8E1AE4ULL, "vxw_encounter2_camp_ld_0_0.vlm"},
	{0xFD74ABFE2A7250ACULL, "666599861.ogg"},
	{0xFD74DA854A95AE7EULL, "vxlt_gallowstemplar_n_3jb_0_0.vlm"},
	{0xFD74ED1F28A557F0ULL, "vxlt_0_lowtown_a2_n_df0_0_0.vlm"},
	{0xFD74FEE72F2FF794ULL, "mag311_keran.bnk"},
	{0xFD7530B767C3190EULL, "mage_deathsyphon.dds"},
	{0xFD7538A1B970B661ULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.dura.gad"},
	{0xFD754ACE642A28C4ULL, "hlac_ht_mansion2b_818_13x.dds"},
	{0xFD75A991CDC4173EULL, "236815373.ogg"},
	{0xFD767C1F5FC810B9ULL, "6125915_m.fxe"},
	{0xFD76B0DBFE6356FCULL, "ser210_witness6168123.cl"},
	{0xFD776883B8210C98ULL, "vxht_0_hightown_c55_0_0.vlm"},
	{0xFD77ED542D4A612EULL, "ht_0_hightownexplosion_n_7_0.msh"},
	{0xFD77FB542D4A797FULL, "ht_0_hightownexplosion_n_7_0.mao"},
	{0xFD7807542D4A8D1CULL, "ht_0_hightownexplosion_n_7_0.mmh"},
	{0xFD781935A414A045ULL, "6028379_m.fxe"},
	{0xFD78386DBD5A75BFULL, "dae000_bartrand6033998.cl"},
	{0xFD78AFDB9EE2881FULL, "30990864.ogg"},
	{0xFD78C1299FCE056FULL, "255058019.ogg"},
	{0xFD78E0FD9631BA55ULL, "6129308_m.fxe"},
	{0xFD794FF5672EE220ULL, "vxw_sundermount_32l_0_0.vlm"},
	{0xFD799540AD318A51ULL, "llac_lt_0_lowtown_docks_-3_3.dds"},
	{0xFD7997F7F7D1475BULL, "art\\characters\\creatures\\dragon\\animation\\atk_v1.gad"},
	{0xFD79A41C91489E9BULL, "324920241.ogg"},
	{0xFD79DBA8333CD0D9ULL, "6098826_m.fxe"},
	{0xFD7A13B3FEBC4EB4ULL, "154321107.ogg"},
	{0xFD7A75650CD66CABULL, "6084542_m.fxe"},
	{0xFD7A891EBD3F2C19ULL, "6106079_m.fxe"},
	{0xFD7AAD21B24429BCULL, "vxlt_undercitydungeon_1zn_0_0.vlm"},
	{0xFD7B39A1B975230FULL, "art\\vfx\\combat\\vfx_w_avl_avaline1_loop_c.dura.ani"},
	{0xFD7B6D9FED3DD869ULL, "407785123.ogg"},
	{0xFD7CA80A7720593BULL, "vxca_caverns1_ehk_0_0.vlm"},
	{0xFD7CEE6F4A287D7DULL, "mag121_eau_banter6107382.cl"},
	{0xFD7D0F970F168761ULL, "l_dre180ar_deep_roads.bnk"},
	{0xFD7D165CCA4C20B2ULL, "gen00fl_fenris6126409.cl"},
	{0xFD7E0BF38CEE4CFFULL, "character_character_p20.psh.fxo.dx11.assembly"},
	{0xFD7E34257159E6B2ULL, "vxht_chantry_n_2y9_0_0.vlm"},
	{0xFD7E524546131DEAULL, "3782684.ogg"},
	{0xFD7EB5FC13DA117DULL, "ach_seb_memento.dds"},
	{0xFD7ECD96416DC7FFULL, "933284802.ogg"},
	{0xFD7F42ED55D69099ULL, "lt_0_lowtown_docks_a3_n_221_0.msh"},
	{0xFD7F4CED55D6A187ULL, "lt_0_lowtown_docks_a3_n_221_0.mmh"},
	{0xFD7F50ED55D6A8D4ULL, "lt_0_lowtown_docks_a3_n_221_0.mao"},
	{0xFD7FCE78BE7A1A0BULL, "zz_vault_debug6203302.cl"},
	{0xFD7FEE7C5F0CC24FULL, "gen_im_trs_gen_sta_phy_r2.uti"},
	{0xFD802FBDD5992BAFULL, "vxlt_gallowstemplar_f_5cj_0_0.vlm"},
	{0xFD80B4F7F7D79740ULL, "art\\characters\\creatures\\dragon\\animation\\atk_v1.evt"},
	{0xFD80BEF26A3882FBULL, "vrc221_bartrand6065317.cl"},
	{0xFD80DB61D5C1674CULL, "hlac_lt_foundry_n_320_18x.dds"},
	{0xFD810D5367AA30EBULL, "hlac_w_bonepitmining_a2_131074_0y.dds"},
	{0xFD812639FD0487B5ULL, "307485837.ogg"},
	{0xFD81AE86BD2E42DAULL, "plt_cod_crt_corpse.plo"},
	{0xFD81B4060913D50FULL, "6076734_m.fxe"},
	{0xFD81F5067CB9D7FCULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_ha01.phy"},
	{0xFD826DC731319ECAULL, "vxw_sundermount_2sp_0_0.vlm"},
	{0xFD827982CE87EC6EULL, "67905067.ogg"},
	{0xFD82F63D1F843364ULL, "vxw_encounter2_camp_ge_0_0.vlm"},
	{0xFD832AED23D31531ULL, "796475432.ogg"},
	{0xFD83C18A76180B8FULL, "vfx_protect01.dds"},
	{0xFD842B9E2350AAE1ULL, "genip_vial_of_blood.utp"},
	{0xFD850D9A80530BDDULL, "art\\vfx\\combat\\vfx_g_dual_att_back_c.mmh"},
	{0xFD857C8295FC5114ULL, "vxht_keep_n_6v6_0_0.vlm"},
	{0xFD85B9D93A1820C6ULL, "lgt320_mis_resp_swordhunter6150366.cl"},
	{0xFD85BC74AF5548E4ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0dd_basic.msh"},
	{0xFD85CE74AF55674DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0dd_basic.mao"},
	{0xFD85D274AF556D86ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0dd_basic.mmh"},
	{0xFD85D487A0C37A4CULL, "abi_m_glyph_paralysis.bnk"},
	{0xFD8600F0A3558E44ULL, "6057109_m.fxe"},
	{0xFD865B424DE43A97ULL, "lt_0_lowtown_docks_a3_n_171_0.mmh"},
	{0xFD865F424DE441A4ULL, "lt_0_lowtown_docks_a3_n_171_0.mao"},
	{0xFD8671424DE46009ULL, "lt_0_lowtown_docks_a3_n_171_0.msh"},
	{0xFD8694A737A969ADULL, "matdef_df_diffuse.dds"},
	{0xFD86C9F12F9B09ECULL, "22459664.ogg"},
	{0xFD87249C594E8B3EULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_103a_sa.dds"},
	{0xFD874AFF672BE282ULL, "137843866.ogg"},
	{0xFD877FAA588C560FULL, "vxca_caverns1_mines_iie_0_0.vlm"},
	{0xFD87B889E7B647CBULL, "zz_followers_banter6158220.cl"},
	{0xFD88D603655208D2ULL, "w_woundedcoast_1_0_lc.mmh"},
	{0xFD88F30DC0A31CA9ULL, "gen00fl_varric6093638.cl"},
	{0xFD891F1B6DFDE3D6ULL, "176232869.ogg"},
	{0xFD893173663F6576ULL, "6079693_m.fxe"},
	{0xFD8956062E044B5AULL, "629060522.ogg"},
	{0xFD89D9FB45702BC3ULL, "llac_w_bonepit_0_9.dds"},
	{0xFD89F1890D466E25ULL, "6144873_m.fxe"},
	{0xFD8A34C4A3B3B82FULL, "zz_followers_banter6157984.cl"},
	{0xFD8A97437E72C961ULL, "mrl150_merrill6093769.cl"},
	{0xFD8B7BB218E6BA7AULL, "6116730_m.fxe"},
	{0xFD8B8BF505458F1EULL, "vxht_bartrand_n_5r4_0_0.vlm"},
	{0xFD8BAF969393A452ULL, "6004402_m.fxe"},
	{0xFD8BD5F67FDC0BD8ULL, "6129598_m.fxe"},
	{0xFD8CEDE8AB806AFBULL, "zz_followers_banter6159708.cl"},
	{0xFD8CF27404D82A4CULL, "479018377.ogg"},
	{0xFD8DF4634D91C729ULL, "6134889_m.fxe"},
	{0xFD8E3E086C62E899ULL, "6121181_f.fxe"},
	{0xFD8EB6B9199BE586ULL, "vxlt_gallowstemplar_n_3dt_0_0.vlm"},
	{0xFD8F2A7016D3C56DULL, "6120454_f.fxe"},
	{0xFD8F6CFDDC388C08ULL, "character_character_p272.psh.fxo.dx11"},
	{0xFD8F7E8108EA43ACULL, "gui_guipixelshader_p4.psh.fxo"},
	{0xFD8F8C72158F025AULL, "524922330.ogg"},
	{0xFD8F93F402FF3AA9ULL, "vrc321_varric6084912.cl"},
	{0xFD8F9A1628B6266DULL, "316358515.ogg"},
	{0xFD8FC974AF5DB351ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\0dd_basic.phy"},
	{0xFD90515F37D6B30EULL, "6054894_m.fxe"},
	{0xFD9058236BE4CEE2ULL, "vxca_cavern1_haunted_jfr_0_0.vlm"},
	{0xFD9066C2BC5F5E1FULL, "art\\characters\\playercharacter\\humanmale\\animation\\blendtree\\hm.anb"},
	{0xFD906FADB8F5B245ULL, "hlac_lt_0_lowtown_a2_65534_3z.dds"},
	{0xFD90932372D979B4ULL, "hlac_lt_0_lowtown_docks_a3_n_-196607_0z.dds"},
	{0xFD90A29BF264DEF8ULL, "592336178.ogg"},
	{0xFD90DFA42327B74AULL, "vxw_sundermount_2tq_0_0.vlm"},
	{0xFD90F1C429064042ULL, "6036335_m.fxe"},
	{0xFD91681A117A4BE4ULL, "vxw_encounter2_camp_bj_0_0.vlm"},
	{0xFD9179DB2E89F019ULL, "660702507.ogg"},
	{0xFD91DEEF97DDCD57ULL, "dae011_corff_sta6142477.cl"},
	{0xFD93030A9D396006ULL, "226514244.ogg"},
	{0xFD9309BE90E18EE6ULL, "470216262.ogg"},
	{0xFD935A6A2D949FE7ULL, "hlac_lt_gallowstemplar_n_2393_1x.dds"},
	{0xFD938287FFA3B5F3ULL, "587418192.ogg"},
	{0xFD938C4CAEB5E3C6ULL, "genip_bones_pile_1.utp"},
	{0xFD9412F84FCFA492ULL, "vxlt_0_lowtown_n_epa_0_0.vlm"},
	{0xFD948B9AEB2DFEA1ULL, "dae000_oldmangossip6175757.cl"},
	{0xFD94F7CB1813782AULL, "6171841_m.fxe"},
	{0xFD95144A34B9B2DBULL, "llac_w_encounter3_-327676_0.dds"},
	{0xFD963F9EDC1D7DD9ULL, "w_woundedcoast_100_0.phy"},
	{0xFD96A2119C03E2B0ULL, "6165648_m.fxe"},
	{0xFD96B8029493A091ULL, "6088420_f.fxe"},
	{0xFD96B9BCA1F8E0BBULL, "77913224.ogg"},
	{0xFD96C57C312860F3ULL, "vxht_0_hightown_a3_b5v_0_0.vlm"},
	{0xFD9785333EFC374EULL, "sst_ss_city_guard_m_1.gda"},
	{0xFD978891B2071DB5ULL, "w_blightlands_150_0.msh"},
	{0xFD978A91B207212BULL, "w_blightlands_150_0.mmh"},
	{0xFD979691B20735C8ULL, "w_blightlands_150_0.mao"},
	{0xFD97B3E579BDE494ULL, "111175163.ogg"},
	{0xFD97E839437164E5ULL, "vxht_bartrand_n_5pq_0_0.vlm"},
	{0xFD98E493B12BB2C2ULL, "222613446.ogg"},
	{0xFD99514511642C9AULL, "6094461_m.fxe"},
	{0xFD99F7E84635D97EULL, "zz_dre_debug6056982.cl"},
	{0xFD99FDD1F73BA79EULL, "vxht_bartrand_n_5u5_0_0.vlm"},
	{0xFD9B9646CFDB1181ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man1_.gad"},
	{0xFD9BC00D80524DACULL, "hlac_w_woundedcoast_196610_0z.dds"},
	{0xFD9C31A9DD7B9DFEULL, "qcr000_arishok6042477.cl"},
	{0xFD9C6E4159635F23ULL, "307850231.ogg"},
	{0xFD9CAAC6A8147820ULL, "6067703_f.fxe"},
	{0xFD9CBA97D697E503ULL, "llac_lt_gallowscourtyard_131070_1.dds"},
	{0xFD9D1377B9855CAFULL, "vxlt_gallowstemplar_f_5it_0_0.vlm"},
	{0xFD9D4A40A06D55AEULL, "6051609_f.fxe"},
	{0xFD9D73B563FAAF1FULL, "914953411.ogg"},
	{0xFD9DAE444635C53DULL, "738038.ogg"},
	{0xFD9DBB60CA00EBC3ULL, "hlac_lt_gallowstemplar_n_2394_6x.dds"},
	{0xFD9DE6542D6B24E7ULL, "ht_0_hightownexplosion_n_7_0.phy"},
	{0xFD9E0747770050A2ULL, "249761108.ogg"},
	{0xFD9E39A7CDA3830EULL, "art\\vfx\\combat\\vfx_gui_target_ret_p.dura.evt"},
	{0xFD9E4A3549D898A3ULL, "vxlt_warehouse_n_12x_0_0.vlm"},
	{0xFD9E6D6273D63280ULL, "6136709_m.fxe"},
	{0xFD9E7504EBC075CDULL, "652881243.ogg"},
	{0xFD9F5181151DCFCAULL, "vxw_sundermount_2yr_0_0.vlm"},
	{0xFD9F6D05752EF947ULL, "hlac_lt_0_lowtown_docks_a3_n_-2_4z.dds"},
	{0xFD9F80C003ED6EAFULL, "gen00fl_anders6142536.cl"},
	{0xFD9FC150B8BA22E3ULL, "hlac_lt_backalley_n_0_6z.dds"},
	{0xFDA08E2270C2D057ULL, "hlac_dr_dungeon_3217_8x.dds"},
	{0xFDA091782FC87067ULL, "hlac_ht_0_hightown_n_65536_0z.dds"},
	{0xFDA091A2A7C1633CULL, "vxca_cave_1ng_0_1.vlm"},
	{0xFDA093EC30C8E06EULL, "seb201_lady_harimann6177419.cl"},
	{0xFDA0AD5E45AB31E1ULL, "llac_lt_0_lowtown_a2_n_c_65535_9.dds"},
	{0xFDA0B7423166EE7AULL, "gen00fl_aveline6139325.cl"},
	{0xFDA1E8B085132018ULL, "one280_merin6104409.cl"},
	{0xFDA1F6625D50A3C6ULL, "6180584_m.fxe"},
	{0xFDA29746CFE1322FULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarriveestablishing_man1_.ani"},
	{0xFDA29B7994CE79C9ULL, "qun131cr_warrens_thug2.utc"},
	{0xFDA29D7F2AD080B9ULL, "dae000_bartrand6206886.cl"},
	{0xFDA38ECBEED65A3BULL, "vxw_bonepit_16j_0_0.vlm"},
	{0xFDA3DCB590CF0291ULL, "vxht_bartrand_n_67w_0_0.vlm"},
	{0xFDA3FF2C0942FD5EULL, "art\\vfx\\combat\\vfx_gib_hm_eye.mao"},
	{0xFDA41A9105F1F413ULL, "238426866.ogg"},
	{0xFDA436DE0BE3E01AULL, "285482542.ogg"},
	{0xFDA530A7CDA992B9ULL, "art\\vfx\\combat\\vfx_gui_target_ret_p.dura.gad"},
	{0xFDA59A052B63FDF4ULL, "6076463_m.fxe"},
	{0xFDA5A4B4BC253F3CULL, "130278368.ogg"},
	{0xFDA5D9E05269DFF9ULL, "6067932_m.fxe"},
	{0xFDA5E78310CF02E3ULL, "551131348.ogg"},
	{0xFDA5ECFA14754682ULL, "270942453.ogg"},
	{0xFDA6DECB6D5AFFC5ULL, "552513891.ogg"},
	{0xFDA72CE6CD761896ULL, "585865133.ogg"},
	{0xFDA773302D48FE6AULL, "vxca_cave_18g_0_1.vlm"},
	{0xFDA773CF38863519ULL, "27040918.ogg"},
	{0xFDA7942B5B673346ULL, "6076748_m.fxe"},
	{0xFDA86FAEE931C01EULL, "vxht_bartrand_n_5x6_0_0.vlm"},
	{0xFDA879A3022E475AULL, "249301357.ogg"},
	{0xFDA91732AA2935B5ULL, "vxht_brothel_n_60z_0_0.vlm"},
	{0xFDA99B4DF5AB71A9ULL, "6041491_m.fxe"},
	{0xFDA9BEBFFAEA9011ULL, "gen00fl_merrill6161650.cl"},
	{0xFDAA9729CFB27AE0ULL, "nat380_party_banter6066172.cl"},
	{0xFDAAE8951EE1AD4DULL, "31097810.ogg"},
	{0xFDAAFA0EC828DF29ULL, "zz_approval_debug6133257.cl"},
	{0xFDABDA8E9BE52510ULL, "hlac_ht_mansion2_818_67y.dds"},
	{0xFDAC41A7CDAFCE07ULL, "art\\vfx\\combat\\vfx_gui_target_ret_p.dura.ani"},
	{0xFDAD266178A5ED92ULL, "follower_banter6171585.cl"},
	{0xFDAD4814CB2AC95FULL, "zz_followers_banter6159612.cl"},
	{0xFDAD9228E5B72CCAULL, "6095974_m.fxe"},
	{0xFDAE16611B4B9C66ULL, "vxlt_undercityencounter_11x_0_0.vlm"},
	{0xFDAE543E3B7BAE64ULL, "llac_lt_0_lowtown_a2_n_c_-196608_3.dds"},
	{0xFDAE7144D5B0B794ULL, "vxlt_0_lowtown_a2_ei3_0_0.vlm"},
	{0xFDAEAF666CC4F699ULL, "llac_ht_chantry_2534_19.dds"},
	{0xFDAF89A8BD587950ULL, "vxht_mansion2_3cw_0_0.vlm"},
	{0xFDAFABE9C17C7CF0ULL, "art\\characters\\playercharacter\\elfmale\\rob_appa.msh"},
	{0xFDAFB9E9C17C940AULL, "art\\characters\\playercharacter\\elfmale\\rob_appa.mmh"},
	{0xFDAFC036EFC97840ULL, "w_sundermount_215_0.phy"},
	{0xFDAFF4FAA6384FBEULL, "gen_im_trs_tmp_shd_met_r2.uti"},
	{0xFDB0D2196BDE9A94ULL, "vxht_keep_n_6g5_0_0.vlm"},
	{0xFDB1D58E9CC92EE1ULL, "164383587.ogg"},
	{0xFDB2091793D6EF04ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\unequipped_weapons_toward.gad"},
	{0xFDB291BEB0D21741ULL, "vxdr_primevalentrance_1qo_0_0.vlm"},
	{0xFDB29F435EFB03E6ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_005a_0d.dds"},
	{0xFDB334CBB8A0EF2BULL, "vxlt_0_lowtown_fec_0_0.vlm"},
	{0xFDB376E59207BE24ULL, "85250691.ogg"},
	{0xFDB3C611A513B046ULL, "art\\characters\\playercharacter\\humanfemale\\boo_bda.phy"},
	{0xFDB3E5D630257315ULL, "vxlt_hangedman_1vp_0_1.vlm"},
	{0xFDB3FADA42CBA144ULL, "dm_gen_2.mor"},
	{0xFDB43C0EEC8CC600ULL, "sst_ss_revenant_1.gda"},
	{0xFDB4AC56022C89FFULL, "6049808_m.fxe"},
	{0xFDB5376165BE74FEULL, "hlac_lt_0_lowtown_a2_n_-65537_5z.dds"},
	{0xFDB58F224B602157ULL, "arch_blood.mao"},
	{0xFDB625F8E28DFB6DULL, "mcr370cs_player_flees.cut"},
	{0xFDB6D2946ACEC9EAULL, "art\\levels\\races\\ferelden\\kirkgallows\\int\\4co_gallowsint\\dr1a_door_2.pwk"},
	{0xFDB6D3EBFCC8D350ULL, "w_blightlands_69_0.msh"},
	{0xFDB6E1EBFCC8EB6AULL, "w_blightlands_69_0.mmh"},
	{0xFDB6E5EBFCC8F1A1ULL, "w_blightlands_69_0.mao"},
	{0xFDB7389B3D98DD55ULL, "772473111.ogg"},
	{0xFDB741197EA9CB26ULL, "vxlt_0_lowtown_docks_a3_cya_0_0.vlm"},
	{0xFDB83EBEAE918D92ULL, "6035316_m.fxe"},
	{0xFDB84D92DDF2B52BULL, "gen_im_trs_gen_hvy_hlm_r1.uti"},
	{0xFDB8624B7B3AFEFAULL, "vxlt_hangedman_n_27e_0_0.vlm"},
	{0xFDB88ADB78D6A28BULL, "gen_im_trs_bst_kidney_03.uti"},
	{0xFDB8CBD6BE6DDA3BULL, "crt_harvester_head.bnk"},
	{0xFDB9181793DD27E2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\unequipped_weapons_toward.ani"},
	{0xFDB957FFF885675FULL, "ca_cavern1_haunted_24427.rml"},
	{0xFDB977A0B94EF3ABULL, "warninglights.gfx"},
	{0xFDBA067DCCC6D901ULL, "315282383.ogg"},
	{0xFDBB1E294B32665CULL, "843792913.ogg"},
	{0xFDBB83F8DEDEAE3DULL, "hlac_lt_gallowscourtyard_n_-131072_4y.dds"},
	{0xFDBB8CEB5F97F14BULL, "vxht_brothel_n_4g3_0_0.vlm"},
	{0xFDBB960EDA5C306FULL, "hm_raider_4.mor"},
	{0xFDBBF43D601CE211ULL, "240495654.ogg"},
	{0xFDBC655EB92435F2ULL, "ser191_emeric6120422.cl"},
	{0xFDBCC3BD40D64455ULL, "6050093_m.fxe"},
	{0xFDBCC73A569D806FULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\1ww_crumble.phy"},
	{0xFDBD6446144A1CAAULL, "hlac_ht_0_hightown_65535_1x.dds"},
	{0xFDBD82406E976614ULL, "6016696_m.fxe"},
	{0xFDBDB08DD3756ED6ULL, "6088791_f.fxe"},
	{0xFDBDCD90D2EAA4D1ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.dura.ani"},
	{0xFDBDED7243B64322ULL, "53399245.ogg"},
	{0xFDBE020CB622E18BULL, "sse_ss_qunari_mage.gda"},
	{0xFDBE5DC7B6C21B2DULL, "llac_ca_caverns1_24424_4.dds"},
	{0xFDBE6FC08F04F479ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_03bonetrack.msh"},
	{0xFDBE7783AA70ADCCULL, "72186448.ogg"},
	{0xFDBE79C08F050567ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_03bonetrack.mmh"},
	{0xFDBE7C8920DCF0E3ULL, "hlac_w_encounter3_-196609_0x.dds"},
	{0xFDBE7DC08F050BB4ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_03bonetrack.mao"},
	{0xFDBE8B6694766340ULL, "vxlt_undercityencounter_1br_0_1.vlm"},
	{0xFDBF115608E1177EULL, "628793803.ogg"},
	{0xFDBF4149BEFF39F4ULL, "mag200_feynriel6217345.cl"},
	{0xFDBF93DDE15A97ADULL, "static1_punch_ds1y.psh.fxo"},
	{0xFDBFB1828ADE8794ULL, "6174813_m.fxe"},
	{0xFDBFC35338CAB6D0ULL, "vxlt_gallowscourtyard_n_7hx_0_0.vlm"},
	{0xFDC0B8319ED3C8A6ULL, "zz_vault_debug6200317.cl"},
	{0xFDC10D7011745975ULL, "vxlt_gallowscourtyard_n_75j_0_0.vlm"},
	{0xFDC12236A1536676ULL, "378099701.ogg"},
	{0xFDC200C774E99F31ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_nobd_0t.dds"},
	{0xFDC2460A87D0E0DEULL, "dae000cr_grand_cleric6148167.cl"},
	{0xFDC27E281247BEE0ULL, "docks_male_overseer_mob6176614.cl"},
	{0xFDC2D6124BE48F6CULL, "llac_ht_mansion2_818_31.dds"},
	{0xFDC3ABE712EB753EULL, "6134608_m.fxe"},
	{0xFDC3C06006768414ULL, "hlac_ht_chantry_n_2534_16y.dds"},
	{0xFDC3CD95B7F8ED30ULL, "vxlt_0_lowtown_a2_n_f57_0_0.vlm"},
	{0xFDC3D890D2EF2373ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.dura.gad"},
	{0xFDC3E673F2EE19E5ULL, "732556584.ogg"},
	{0xFDC41D970F52A113ULL, "l_dre180ar_deep_roads.wbd"},
	{0xFDC4843869E9D7A1ULL, "424887279.ogg"},
	{0xFDC4D06FAECE5F89ULL, "vxlt_darktown_41p_0_0.vlm"},
	{0xFDC4E6F46187D1A9ULL, "565380275.ogg"},
	{0xFDC50D27E142980CULL, "6172785_m.fxe"},
	{0xFDC580706D042A5DULL, "hlac_lt_0_lowtown_-65535_7x.dds"},
	{0xFDC5FAB04F511EE2ULL, "zz_bowing.cut"},
	{0xFDC652FBB11422D9ULL, "139881151.ogg"},
	{0xFDC68CB1CC5E6B9AULL, "97343264.ogg"},
	{0xFDC6A87F925A43ACULL, "372196268.ogg"},
	{0xFDC6DE87A0FB227EULL, "abi_m_glyph_paralysis.wbd"},
	{0xFDC70CAD6A292419ULL, "follower_banter6170415.cl"},
	{0xFDC765CFCC96D140ULL, "vxlt_0_lowtown_docks_n_co9_0_0.vlm"},
	{0xFDC79F3E0EA79614ULL, "llac_ht_chantry_2534_43.dds"},
	{0xFDC7AC98D7EB4A5EULL, "6115326_m.fxe"},
	{0xFDC7DEC210CC38C6ULL, "6060361_m.fxe"},
	{0xFDC7E68CAFF169F4ULL, "vxht_keep_60o_0_0.vlm"},
	{0xFDC86B1DE56352BCULL, "592510271.ogg"},
	{0xFDC9BA1BF2D38331ULL, "342454621.ogg"},
	{0xFDCA42441C0C3B69ULL, "mer320_hubert6052587.cl"},
	{0xFDCA672D8824F0CCULL, "w_bonepit_151_0.phy"},
	{0xFDCA71B28032D1C2ULL, "gen00fl_aveline6128406.cl"},
	{0xFDCAD590D2F53CD8ULL, "art\\vfx\\combat\\vfx_a_stealth_loop_c.dura.evt"},
	{0xFDCB0C157FFA9CF1ULL, "vxlt_undercityencounter_1h8_0_0.vlm"},
	{0xFDCB1D601387B3F4ULL, "dae000_arishok6067733.cl"},
	{0xFDCB357DC34912B8ULL, "6163408_m.fxe"},
	{0xFDCB36533B6B95D5ULL, "6174019_f.fxe"},
	{0xFDCB3CA81B024070ULL, "520910896.ogg"},
	{0xFDCB5333A8204229ULL, "6111112_m.fxe"},
	{0xFDCBDDF03375407FULL, "vxht_mansion2b_345_0_0.vlm"},
	{0xFDCC5D412E35684CULL, "237331706.ogg"},
	{0xFDCCADDDA63E990EULL, "w_sundermount_88_0_ll.mmh"},
	{0xFDCCBB76EB4B2ED3ULL, "6141876_m.fxe"},
	{0xFDCD1EEB76CB61BFULL, "hightown_fem_cityguard_mob.bnk"},
	{0xFDCD6F8F3502AE70ULL, "art\\characters\\creatures\\highdragon\\animation\\blendtree\\highdragon_pcoverride.anb"},
	{0xFDCE1066F1EC2393ULL, "dae000_pc_home_ambients6208647.cl"},
	{0xFDCE169DC81129F5ULL, "vxca_caverns1_mines_gf0_0_0.vlm"},
	{0xFDCE82EA7B02378DULL, "art\\vfx\\environment\\vfx_smoke_billow_l_p.mmh"},
	{0xFDCEA389B189ED34ULL, "558821227.ogg"},
	{0xFDCEE278B43BC4F9ULL, "hlac_lt_0_lowtown_docks_-262143_0z.dds"},
	{0xFDCF757894BE4841ULL, "vxdr_primevalentrance_1km_0_0.vlm"},
	{0xFDCFADFC8113B276ULL, "panning1_v.cur"},
	{0xFDCFDCDB7DC3AC43ULL, "w_blightlands_203_0.mao"},
	{0xFDCFE0DB7DC3B290ULL, "w_blightlands_203_0.mmh"},
	{0xFDCFEEDB7DC3CAAAULL, "w_blightlands_203_0.msh"},
	{0xFDD000EB76CD64C7ULL, "hightown_fem_cityguard_mob.cnv"},
	{0xFDD04B25BD844969ULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl05.phy"},
	{0xFDD0ED67B3A9108AULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_exit.gad"},
	{0xFDD109727D9D0F35ULL, "6167548_m.fxe"},
	{0xFDD1BB38E427B956ULL, "240375324.ogg"},
	{0xFDD21F45DB508ACAULL, "art\\characters\\playercharacter\\children\\animation\\blendtree\\child_overlay.anb"},
	{0xFDD266F438A5F0DBULL, "vxw_encounter2_gf_0_0.vlm"},
	{0xFDD29FC4AB1AE2AEULL, "and111_lirene6085106.cl"},
	{0xFDD39D8C59D67DEFULL, "dae000_mother6002005.cl"},
	{0xFDD472E695532537ULL, "524120427.ogg"},
	{0xFDD4BFEFC6F326FEULL, "803987281.ogg"},
	{0xFDD4DD36EFE8C5D5ULL, "w_sundermount_215_0.msh"},
	{0xFDD4DF36EFE8C94BULL, "w_sundermount_215_0.mmh"},
	{0xFDD4EB36EFE8DDE8ULL, "w_sundermount_215_0.mao"},
	{0xFDD5C42B12C1FD6DULL, "vxlt_0_lowtown_a2_n_c_fj1_0_0.vlm"},
	{0xFDD5C6DF7DE0C55DULL, "vxht_mansion2_1s0_0_0.vlm"},
	{0xFDD5C9735B4414EFULL, "vxht_0_hightown_c3u_0_1.vlm"},
	{0xFDD650A3D9C9A4ECULL, "275503178.ogg"},
	{0xFDD66C7870C2EEA9ULL, "6148024_m.fxe"},
	{0xFDD6838BE9C4FB69ULL, "750719576.ogg"},
	{0xFDD68C1BE019723FULL, "dae110ip_store4.utp"},
	{0xFDD6E4409420F455ULL, "592343306.ogg"},
	{0xFDD786A8645992B4ULL, "caustic_blend.mao"},
	{0xFDD791826D461DC2ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.cess.gad"},
	{0xFDD7D067B3AEFE21ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_exit.evt"},
	{0xFDD89E697BC7528AULL, "6070477_f.fxe"},
	{0xFDD8A5371AC5A6DAULL, "6016719_m.fxe"},
	{0xFDD9678F97E9F2B4ULL, "6148111_m.fxe"},
	{0xFDD9884D2D64F7ACULL, "6188630_m.fxe"},
	{0xFDD9A432DE6F5482ULL, "6163327_m.fxe"},
	{0xFDD9A8C4ED4EE1F0ULL, "lt_0_lowtown_docks_n_27_0_lc.mmh"},
	{0xFDD9F7AE2F496FD7ULL, "6188595_f.fxe"},
	{0xFDD9F941B9414ABCULL, "812989236.ogg"},
	{0xFDDA12869FE97937ULL, "vxht_mansion2b_3cd_0_0.vlm"},
	{0xFDDA5FB74731029EULL, "6115894_m.fxe"},
	{0xFDDB4725BCE73B8DULL, "6153411_m.fxe"},
	{0xFDDBB22CC2A99D1FULL, "lgt000_chanter6147198.cl"},
	{0xFDDC0918F48D2593ULL, "dae300_cullen6140599.cl"},
	{0xFDDC6F67BA3B1BFAULL, "hlac_dr_dungeon_3207_7y.dds"},
	{0xFDDCF199EFA3AADCULL, "zz_followers_banter6158841.cl"},
	{0xFDDD2C97B0AE11F3ULL, "57228333.ogg"},
	{0xFDDD41A42B86700FULL, "6126602_m.fxe"},
	{0xFDDD65AB827D24A5ULL, "hlac_lt_gamlenshouse_n_-1_6y.dds"},
	{0xFDDDB4FEEE92ABEAULL, "and120tr_anders.utt"},
	{0xFDDE38C22216157FULL, "562798635.ogg"},
	{0xFDDE73DC607D2C47ULL, "llac_lt_0_lowtown_-2_3.dds"},
	{0xFDDEB373C8ED7BF9ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_aoe_p.impa.ani"},
	{0xFDDEB4826D4C77B9ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.cess.evt"},
	{0xFDDEC46EDFB16203ULL, "994492208.ogg"},
	{0xFDDEE69CD7B7B189ULL, "6218954_m.fxe"},
	{0xFDDF21E19C8B4101ULL, "6094790_m.fxe"},
	{0xFDDF661DC9E265C7ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3_b.gad"},
	{0xFDE0715784B59D96ULL, "face0_bssrdf_2.psh.fxo"},
	{0xFDE07D3E2D789998ULL, "vxht_brothel_n_5ag_0_0.vlm"},
	{0xFDE0BA00977AA777ULL, "vxw_woundedcoast_7s_0_0.vlm"},
	{0xFDE0DADA47342C7EULL, "78872974.ogg"},
	{0xFDE12DCF4D2C3558ULL, "vxw_woundedcoast_1e7_0_0.vlm"},
	{0xFDE136D1038F1C83ULL, "6136733_m.fxe"},
	{0xFDE243A488F82D02ULL, "126934270.ogg"},
	{0xFDE2460DB7EEA8BCULL, "hlac_w_sundermount_-327679_0z.dds"},
	{0xFDE359EF7FAF8E45ULL, "vxlt_hangedman_n_27n_0_0.vlm"},
	{0xFDE38F2D667612ECULL, "hlac_w_encounter3_-131068_0y.dds"},
	{0xFDE3B694E75DBDBDULL, "sky_ht_keep_n.mao"},
	{0xFDE3B82013701623ULL, "6147816_m.fxe"},
	{0xFDE3C14C3EA64454ULL, "251222554.ogg"},
	{0xFDE42542F8D86344ULL, "6088677_m.fxe"},
	{0xFDE49738CD0941A1ULL, "6070364_f.fxe"},
	{0xFDE4C867B3B97B70ULL, "art\\characters\\playercharacter\\elfmale\\animation\\cindes\\dialogue\\fenris_turn_left_90_exit.ani"},
	{0xFDE4C8D9B0DA289CULL, "934164535.ogg"},
	{0xFDE4D172A06E64C2ULL, "hlac_lt_undercityserialkiller_-1_49y.dds"},
	{0xFDE504BA38E5EB90ULL, "vxht_bartrand_n_6re_0_0.vlm"},
	{0xFDE56447ECEAC1DEULL, "qun120_arishok_guard6054844.cl"},
	{0xFDE58E73C8F35B3BULL, "art\\vfx\\combat\\vfx_ha_bd_slam_aoe_p.impa.gad"},
	{0xFDE5FA4A69440CBCULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_swordshield.ani"},
	{0xFDE6037AD14E6CD3ULL, "vxlt_0_lowtown_a2_n_gn8_0_0.vlm"},
	{0xFDE65086C2F8DB0AULL, "6037325_m.fxe"},
	{0xFDE65B1DC9E871D4ULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3_b.evt"},
	{0xFDE6EB6651F14336ULL, "6095800_m.fxe"},
	{0xFDE74C1F005CEB2BULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.cess.ani"},
	{0xFDE769DBCE0F113EULL, "hlac_ht_brothel_n_3372_0x.dds"},
	{0xFDE84C1909E70D54ULL, "713026270.ogg"},
	{0xFDE87CCAE4EDE4B3ULL, "6053326_m.fxe"},
	{0xFDE8846391DF91B7ULL, "vxht_mansion2b_3fg_0_0.vlm"},
	{0xFDE8D19439271B1EULL, "6110794_m.fxe"},
	{0xFDE8D28877A5DF09ULL, "gen_im_arm_cht_lgt_act2_04.uti"},
	{0xFDE9070194B0D469ULL, "cd_qun231_varnell.bnk"},
	{0xFDE9EBBADFBB3477ULL, "and100_anders6096024.cl"},
	{0xFDE9F44A400E6C11ULL, "gen00fl_anders6104006.cl"},
	{0xFDEA28DC9636C8BFULL, "6061349_m.fxe"},
	{0xFDEAE96B5727E474ULL, "vxlt_gallowstemplar_n_4uz_0_0.vlm"},
	{0xFDEAFCEE62868D58ULL, "ser270_emeric6113220.cl"},
	{0xFDEB1168EC5AD229ULL, "771253761.ogg"},
	{0xFDEB8D54CD9024CFULL, "hlac_ca_caverns2_5802_44x.dds"},
	{0xFDEB9A7157665A88ULL, "6074469_m.fxe"},
	{0xFDEBBC826D570FA8ULL, "art\\vfx\\combat\\vfx_r_gn2_speed_loop_c.cess.ani"},
	{0xFDEBE063F33555E2ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_2.gad"},
	{0xFDEC70F5E22F1496ULL, "6098813_f.fxe"},
	{0xFDECAB73C8F9AB20ULL, "art\\vfx\\combat\\vfx_ha_bd_slam_aoe_p.impa.evt"},
	{0xFDED074A694A41FEULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_swordshield.gad"},
	{0xFDED0F1A5909E32AULL, "dm_qun_javaris.mor"},
	{0xFDED12B2A207D702ULL, "513557714.ogg"},
	{0xFDED27EC343CFC47ULL, "qun221_petrice6185221.cl"},
	{0xFDEE012535ADF4E7ULL, "gen00fl_anders6135989.cl"},
	{0xFDEE69076FE1DFE7ULL, "zz_approval_debug6003682.cl"},
	{0xFDEE8CEE64F248CDULL, "w_bonepitdestroyed_a3_203_0.mmh"},
	{0xFDEE90EE64F24F1EULL, "w_bonepitdestroyed_a3_203_0.mao"},
	{0xFDEE9EEE64F26753ULL, "w_bonepitdestroyed_a3_203_0.msh"},
	{0xFDEEB7195375EC3EULL, "mcr310_party_banter6200743.cl"},
	{0xFDEEBFF5CC0A58C4ULL, "vxht_0_hightown_a3_9aj_0_0.vlm"},
	{0xFDEEC8E96C232A78ULL, "vxlt_gallowstemplar_n_6sv_0_0.vlm"},
	{0xFDEEEA969114F54CULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v2.ani"},
	{0xFDEEFFC5578DDC9FULL, "dae000_oldmangossip6174983.cl"},
	{0xFDEF680B772F371DULL, "534600537.ogg"},
	{0xFDEF6C30DD1F9681ULL, "vxlt_undercityserialkiller_15c_0_0.vlm"},
	{0xFDF02755ECADEE63ULL, "hlac_lt_0_lowtown_a2_65534_0y.dds"},
	{0xFDF088CBC0EA119AULL, "plt_ave000pt_endstage.nss"},
	{0xFDF0A0CD181E0141ULL, "ave200_aveline_bar6116200.cl"},
	{0xFDF0EC61E61FCA49ULL, "zz_followers_banter6159423.cl"},
	{0xFDF17DA29DD9FFE8ULL, "mag110_arianni.cnv"},
	{0xFDF1D350C2D5193FULL, "vxdr_primevaltomb_2vz_0_1.vlm"},
	{0xFDF20363F339FCD9ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_2.evt"},
	{0xFDF208B69EB04497ULL, "6084008_f.fxe"},
	{0xFDF2971FEACE7BC4ULL, "6083777_m.fxe"},
	{0xFDF30FFFBD67634FULL, "art\\vfx\\cinematic\\vfx_meredith_smokey_c.mmh"},
	{0xFDF3531DC9F2EE0DULL, "art\\characters\\creatures\\gallowsslavestatue\\animation\\alive_dropv3_b.ani"},
	{0xFDF3B2517538C6D9ULL, "hlac_ht_0_hightown_a3_0_0y.dds"},
	{0xFDF3B8B1E6209BBDULL, "345478035.ogg"},
	{0xFDF3E24A6950211DULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\reactions\\generic exits\\genericexit_swordshield.evt"},
	{0xFDF407C5C7B88DABULL, "hlac_ht_chantry_n_2534_38z.dds"},
	{0xFDF41721B2264118ULL, "qun131_taskmaster6019209.cl"},
	{0xFDF42F111ECE8585ULL, "658790163.ogg"},
	{0xFDF49FA29DDC6F50ULL, "mag110_arianni.bnk"},
	{0xFDF4FA8221AD2EEAULL, "hlac_dr_slavers_3216_12y.dds"},
	{0xFDF52F4D31F0E123ULL, "vxlt_backalley_n_1e8_0_0.vlm"},
	{0xFDF5341F0068FFA2ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.cess.evt"},
	{0xFDF54C43154DE886ULL, "1066449105.ogg"},
	{0xFDF58FFEEEA6FED3ULL, "and120tr_anders.ncs"},
	{0xFDF598839AA58373ULL, "hlac_ht_bartrand_n_2995_2y.dds"},
	{0xFDF5CD8B6BDC4365ULL, "dirty_stuff2.mao"},
	{0xFDF5F734A7F68064ULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_arm_mrla_0n.dds"},
	{0xFDF5F796911B298EULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v2.gad"},
	{0xFDF67D5987EB1D17ULL, "304672454.ogg"},
	{0xFDF7638680696FBEULL, "6151173_m.fxe"},
	{0xFDF7B43C4D1126AAULL, "6106628_f.fxe"},
	{0xFDF7D2CC381DFA1EULL, "vxht_mansion2b_1jt_0_0.vlm"},
	{0xFDF82D2779C23B86ULL, "ser221_gamlen.cnv"},
	{0xFDF8BC9C0479485CULL, "301949213.ogg"},
	{0xFDF8C591A0949816ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_ready.evt"},
	{0xFDF9209D64186B44ULL, "hlac_dr_slavers_3216_2x.dds"},
	{0xFDF97BF02AD86E96ULL, "488361051.ogg"},
	{0xFDF98B33EA4A7EF9ULL, "ht_0_hightown_8_0.msh"},
	{0xFDF99533EA4A8FE7ULL, "ht_0_hightown_8_0.mmh"},
	{0xFDF99933EA4A9634ULL, "ht_0_hightown_8_0.mao"},
	{0xFDFA255E0F72A10FULL, "6128402_m.fxe"},
	{0xFDFA834067018AD0ULL, "vxca_caverns1_iz2_0_0.vlm"},
	{0xFDFAC1F2BE27E183ULL, "6168161_m.fxe"},
	{0xFDFAC6111E80A117ULL, "vxdr_slavers_1b0_0_0.vlm"},
	{0xFDFB2F2779C47546ULL, "ser221_gamlen.bnk"},
	{0xFDFBD1A026838086ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\_textures\\set_hightownint_window_d.dds"},
	{0xFDFBD24440AA233BULL, "vxw_bonepitmining_a2_1mu_0_0.vlm"},
	{0xFDFBDCD891C65FBFULL, "449971777.ogg"},
	{0xFDFBF296911F8D6DULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\arishock\\attk_v2.evt"},
	{0xFDFC331F006F1CB5ULL, "art\\vfx\\combat\\vfx_glb_fire_imp_dbl_c.cess.gad"},
	{0xFDFC3425BDA95C8EULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl05.mmh"},
	{0xFDFC4E25BDA9897CULL, "art\\characters\\playercharacter\\humanmale\\hm_brd_hl05.msh"},
	{0xFDFC78D906EEA109ULL, "one100dg_elren.bnk"},
	{0xFDFC84CAE9E712FFULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_diningtable.msh"},
	{0xFDFC8ACAE9E71CA1ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_diningtable.mmh"},
	{0xFDFC96CAE9E73142ULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_diningtable.mao"},
	{0xFDFDC1011929C84FULL, "character_character_p51.psh.fxo.dx11"},
	{0xFDFE10FF9ABACBF1ULL, "947691036.ogg"},
	{0xFDFE118931186658ULL, "vxw_woundedcoast_1k5_0_0.vlm"},
	{0xFDFE67B072E16E2FULL, "art\\levels\\races\\proxy\\tilesets\\setext\\4co_basic\\1wi_basic_1.mao"},
	{0xFDFE67C1665DEE01ULL, "art\\vfx\\materials\\vfx_blood_sh_si.mao"},
	{0xFDFED43BE03C964DULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_boo_hvyn_0d.dds"},
	{0xFDFEE92EA7FA33EFULL, "zz_warrior3.cut"},
	{0xFDFF85EE6500AF8AULL, "w_bonepitdestroyed_a3_203_0.phy"},
	{0xFDFFB9D6BEA9BE39ULL, "crt_harvester_head.wbd"},
	{0xFDFFBC91A09AA7A1ULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_ready.gad"},
	{0xFDFFBCC9C4DF0EA3ULL, "hlac_w_bonepitmining_a2_131075_0z.dds"},
	{0xFDFFC6228BB4561DULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\_textures\\til_lowtown_main_d.dds"},
	{0xFE000B63F34647C8ULL, "art\\characters\\playercharacter\\dwarfmale\\animation\\cindes\\varric\\varric_2.ani"},
	{0xFE0023CE61410FBEULL, "vxw_bonepitdestroyed_a3_1vk_0_0.vlm"},
	{0xFE005AD906F25749ULL, "one100dg_elren.cnv"},
	{0xFE006256E3E8A3D1ULL, "333456407.ogg"},
	{0xFE008552FA18D2F9ULL, "6088669_m.fxe"},
	{0xFE00AE8A2F2DDD43ULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_2.ani"},
	{0xFE00E19C29B69CD3ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.evt"},
	{0xFE00E1A0E6D902D3ULL, "539316220.ogg"},
	{0xFE0108FCDCC49444ULL, "6086877_m.fxe"},
	{0xFE0170B919CF0B07ULL, "t3_wer_b01.tnt"},
	{0xFE02B1E181C1C259ULL, "vxlt_backalley_n_1du_0_0.vlm"},
	{0xFE02B5C0F783AD67ULL, "774501935.ogg"},
	{0xFE0350B478220C57ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_incense.gad"},
	{0xFE03F545C81F494FULL, "570914841.ogg"},
	{0xFE04282AE6DE8014ULL, "dae200_bran6094799.cl"},
	{0xFE0445C1C3338DF8ULL, "dre000_bodahn6028169.cl"},
	{0xFE0493EDF4C70554ULL, "one100dg_kelder6055417.cl"},
	{0xFE04B9C2E85B1A2AULL, "sib131ip_door_vault.utp"},
	{0xFE04D956850EA390ULL, "963355292.ogg"},
	{0xFE04E74ECFF92AD3ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\caveinterior\\9ww_sandpile2a_2.mao"},
	{0xFE04E899B25BEA48ULL, "6060363_m.fxe"},
	{0xFE0556ED8D4228E1ULL, "vxlt_hangedman_n_247_0_0.vlm"},
	{0xFE05B24B906590A8ULL, "gen00fl_dog6107702.cl"},
	{0xFE05B60F0A726AA7ULL, "6095957_m.fxe"},
	{0xFE0644A92A14129EULL, "vxht_mansion2b_1mu_0_0.vlm"},
	{0xFE0682321864064FULL, "llac_ht_mansion2_818_18.dds"},
	{0xFE06BD91A0A0C84FULL, "art\\characters\\playercharacter\\qunarimale\\animation\\combat\\trooperbase\\troop_ready.ani"},
	{0xFE06EF215DD668E7ULL, "vxht_0_hightown_a3_a6t_0_0.vlm"},
	{0xFE074E01AC6842FDULL, "6137602_m.fxe"},
	{0xFE07C851E19D1EAAULL, "ss_hawkecu_clean.cut"},
	{0xFE07DA9C29BCAF14ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.gad"},
	{0xFE07E529AFA1049EULL, "llac_lt_0_lowtown_docks_n_131071_6.dds"},
	{0xFE0828D0BE4BC854ULL, "dae000_bartrand6207006.cl"},
	{0xFE087D5DCEC93CA4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_slow_zone_loop.gad"},
	{0xFE08A734850DA394ULL, "vxlt_0_lowtown_a3_n_c_ekg_0_2.vlm"},
	{0xFE0923015CBE8C55ULL, "vxw_bonepitmining_a2_pg_0_0.vlm"},
	{0xFE092FF5FA6469B3ULL, "6043899_m.fxe"},
	{0xFE09610FBF7B3B87ULL, "vxca_caverns1_mines_ind_0_0.vlm"},
	{0xFE0A182E79CCBDF2ULL, "hlac_ht_bartrand_n_3367_0x.dds"},
	{0xFE0A3E2D885A8E97ULL, "w_bonepit_151_0.mmh"},
	{0xFE0A422D885A95A4ULL, "w_bonepit_151_0.mao"},
	{0xFE0A542D885AB409ULL, "w_bonepit_151_0.msh"},
	{0xFE0A65B478284EE4ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_incense.evt"},
	{0xFE0AC15CF405656CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_up.ani"},
	{0xFE0AC8FABDA4C8E2ULL, "171667158.ogg"},
	{0xFE0B837C5432DE3AULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.resetup.gad"},
	{0xFE0C450C99A33034ULL, "gut_sh_red.mao"},
	{0xFE0CBDD24A3754D7ULL, "zz_isa_debug6175284.cl"},
	{0xFE0D45F7EFB064E4ULL, "lt_0_lowtown_a3_n_c_51_0_ll.mmh"},
	{0xFE0D50F981659CCEULL, "785630041.ogg"},
	{0xFE0D9B0EDAA321F8ULL, "6110983_m.fxe"},
	{0xFE0DD48B6BF0E0A3ULL, "dirty_stuff2.dds"},
	{0xFE0DDBED8C194BF5ULL, "521206546.ogg"},
	{0xFE0DF0345BC6EF49ULL, "337573540.ogg"},
	{0xFE0EC68A2F3A433AULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_2.evt"},
	{0xFE0EDFA34E96D4C8ULL, "vxlt_hangedman_n_242_0_0.vlm"},
	{0xFE0EE99C29C2E7F2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_clasped_1_gesture_1.ani"},
	{0xFE0EF72FEC0EEB79ULL, "6083769_m.fxe"},
	{0xFE0F2E0F52992C6CULL, "154803717.ogg"},
	{0xFE0F8C5DCECF7482ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_slow_zone_loop.ani"},
	{0xFE0FC3F09E2C72ADULL, "distortionskin.vsh.fxo"},
	{0xFE0FD1497DA619CEULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bunklow.msh"},
	{0xFE0FDF497DA6311FULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bunklow.mao"},
	{0xFE0FEB497DA645BCULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bunklow.mmh"},
	{0xFE0FFEF4BE15093EULL, "vxca_cavern1_haunted_i7v_0_0.vlm"},
	{0xFE105D7E871F1F4BULL, "473989063.ogg"},
	{0xFE10A2A5E992579AULL, "3203352.ogg"},
	{0xFE10CE5CF409E6AEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_up.gad"},
	{0xFE10FE2486D46ABAULL, "vxlt_undercityencounter_1j1_0_0.vlm"},
	{0xFE116B9C6A6D2C7FULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha09a.msh"},
	{0xFE11719C6A6D3621ULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha09a.mmh"},
	{0xFE11773776585FEEULL, "448657835.ogg"},
	{0xFE117999BF4270EDULL, "hlac_lt_hangedman_-1_2z.dds"},
	{0xFE1190CADA0EB891ULL, "6129119_m.fxe"},
	{0xFE12633B6C6FDCF7ULL, "6121041_m.fxe"},
	{0xFE1274C221CE295CULL, "6069957_m.fxe"},
	{0xFE12867C54390251ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.resetup.evt"},
	{0xFE12C905E22DA747ULL, "lt_0_lowtown_a3_n_c_37_0_lc.mmh"},
	{0xFE13B6EF4E7AF029ULL, "ser270_templar6112550.cl"},
	{0xFE13C8CA7F384161ULL, "vxlt_hangedman_n_236_0_0.vlm"},
	{0xFE13F3D8DEB0C0ACULL, "follower_banter6171073.cl"},
	{0xFE142EF1A44BCBE9ULL, "vxht_bartrand_n_6kk_0_0.vlm"},
	{0xFE14C58A2F3EAD2DULL, "art\\characters\\playercharacter\\humanfemale\\animation\\cindes\\custom\\cassandra_book_gesture_2.gad"},
	{0xFE14D02CC6350B04ULL, "llac_ht_keep_4519_15.dds"},
	{0xFE14E6CC2F3FE0B1ULL, "lowtown_male_alienage1_sta6161633.cl"},
	{0xFE150394C145BE43ULL, "168461885.ogg"},
	{0xFE15285C832F1F9CULL, "vxw_sundermount_319_0_0.vlm"},
	{0xFE15706ED48D5115ULL, "llac_lt_gallowscourtyard_n_-65538_2.dds"},
	{0xFE15976FB8728955ULL, "6135792_m.fxe"},
	{0xFE1682C5BA643ADDULL, "vxlt_gallowstemplar_n_53h_0_0.vlm"},
	{0xFE16921746136644ULL, "hlac_ht_0_hightown_-65536_3z.dds"},
	{0xFE16F36F100B957BULL, "59870332.ogg"},
	{0xFE171A7376BA13F6ULL, "65683299.ogg"},
	{0xFE174506B6894C1FULL, "vxlt_gamlenshouse_n_b1_0_0.vlm"},
	{0xFE17653F0FEB13E1ULL, "hlac_lt_blackemporium_1055_3x.dds"},
	{0xFE17C95CF40FFC8DULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\elbow_l_up.evt"},
	{0xFE1816AA8EDB622FULL, "fxc_bm_leak.mao"},
	{0xFE187D7DB8F62AC7ULL, "6076886_m.fxe"},
	{0xFE187DB47834B4DDULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\warrior\\twohanded\\tw_incense.ani"},
	{0xFE18AF2AF233AB36ULL, "vxlt_0_lowtown_docks_n_cxj_0_0.vlm"},
	{0xFE18CFEF58CF2F3AULL, "showeffects.ncs"},
	{0xFE18FE0AAAC9D560ULL, "vxw_bonepitdestroyed_a3_198_0_0.vlm"},
	{0xFE1940D13DB15E32ULL, "isa311_isabela6091839.cl"},
	{0xFE19BC67D1C8C8B9ULL, "33569119.ogg"},
	{0xFE1A72742A060392ULL, "1047463267.ogg"},
	{0xFE1AAFAB8F630800ULL, "vxht_chantry_35u_0_0.vlm"},
	{0xFE1AF54315049758ULL, "vxw_woundedcoast_1q3_0_0.vlm"},
	{0xFE1B15AD57A915BAULL, "llac_lt_undercityencounter_-1_11.dds"},
	{0xFE1B1BCABB744D8FULL, "284418343.ogg"},
	{0xFE1B1E0A0C7F3CF0ULL, "vxca_caverns1_mines_e8y_0_0.vlm"},
	{0xFE1B440063C2169BULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16b.phy"},
	{0xFE1BE67A5A1F8941ULL, "isa111_chantry_st.stg"},
	{0xFE1C10FFBD8A698CULL, "art\\vfx\\cinematic\\vfx_meredith_smokey_c.anb"},
	{0xFE1C3CB98BD16449ULL, "w_bonepit_16_0_lc.mmh"},
	{0xFE1CF1E147550F2FULL, "vxlt_gallowstemplar_n_5pk_0_0.vlm"},
	{0xFE1D207AE88ADC05ULL, "vxlt_gallowstemplar_n_6so_0_0.vlm"},
	{0xFE1D87B5C3D7C879ULL, "hlac_lt_gallowscourtyard_n_-65536_3y.dds"},
	{0xFE1D8DB1584B70FCULL, "docks_male_elven_worker_sta6175844.cl"},
	{0xFE1E009DC14068A4ULL, "dae000_madam_lusine6151708.cl"},
	{0xFE1F2D1A721FF297ULL, "6217144_m.fxe"},
	{0xFE1F9E7C5443B520ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\animation\\4co_lowtown_int_trap.resetup.ani"},
	{0xFE1FE182B5A64765ULL, "hlac_ht_chantry_2534_17z.dds"},
	{0xFE2004469ED82999ULL, "gen00fl_bethany6102877.cl"},
	{0xFE20658359A03AACULL, "follower_barks6173414.cl"},
	{0xFE208B3B0639D78AULL, "hlac_lt_hangedman_n_-1_19x.dds"},
	{0xFE20B64B9580EB54ULL, "6090845_m.fxe"},
	{0xFE20EF58773436A8ULL, "mag210_arianni6030491.cl"},
	{0xFE2182F16F21A354ULL, "follower_banter6172821.cl"},
	{0xFE22354784003325ULL, "dm_emrys.mor"},
	{0xFE226E16D9D72C0BULL, "6124715_f.fxe"},
	{0xFE2271EC83CE5771ULL, "one300cr_crow_1.utc"},
	{0xFE232D7605CC0728ULL, "mcr310_argument6182300.cl"},
	{0xFE23C705CEFCA428ULL, "ht_0_hightown_n_18_0_ll.mmh"},
	{0xFE240FCEC02EE447ULL, "llac_lt_0_lowtown_docks_a3_0_0.dds"},
	{0xFE2451F7C650337BULL, "w_bonepitmining_a2_210_0.mmh"},
	{0xFE245DB4B5B25F14ULL, "vxca_caverns1_k2n_0_2.vlm"},
	{0xFE245DF7C6504718ULL, "w_bonepitmining_a2_210_0.mao"},
	{0xFE246F0463EB26DCULL, "hlac_ht_0_hightown_a3_-65535_1z.dds"},
	{0xFE246FF7C6506665ULL, "w_bonepitmining_a2_210_0.msh"},
	{0xFE24F4A2AC5A535DULL, "vxlt_gallowstemplar_n_56g_0_0.vlm"},
	{0xFE24FFCF55B736D0ULL, "vxw_encounter3_dw_0_0.vlm"},
	{0xFE2503670336B777ULL, "w_encounter2_camp_5_0_lc.mmh"},
	{0xFE257E33EA6FA31CULL, "ht_0_hightown_8_0.phy"},
	{0xFE2582A868C09634ULL, "6133254_m.fxe"},
	{0xFE258B9AC8773B26ULL, "8277464.ogg"},
	{0xFE25D8D73CE3D450ULL, "vxca_caverns1_ik1_0_0.vlm"},
	{0xFE25F108BDD83C35ULL, "llac_w_bonepitmining_a2_196612_0.dds"},
	{0xFE260344644EDA43ULL, "mcr370_meredith6084411.cl"},
	{0xFE26F78D4D4FA586ULL, "6110843_f.fxe"},
	{0xFE2717E4606C3D29ULL, "mag320_orsino6071211.cl"},
	{0xFE275EA056516416ULL, "art\\characters\\weapons\\longswords\\w_lsw_017a.phy"},
	{0xFE27A31956E42CBBULL, "zz_isa_debug6133985.cl"},
	{0xFE27F472CFEC5837ULL, "lgtcr_dalish_assassin.utc"},
	{0xFE27F938D2549B5AULL, "art\\levels\\races\\natural\\highlands\\props\\_textures\\prp_hcliff2_sa.dds"},
	{0xFE28192D6766C198ULL, "191432998.ogg"},
	{0xFE28AE1B4B3A8399ULL, "primitives_primitive_p3.psh.fxo.dx11"},
	{0xFE29429AAA213936ULL, "ht_0_hightownexplosion_n_12_0_lc.mmh"},
	{0xFE2977F5A3B0E103ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_c.impa.evt"},
	{0xFE2981FF84B9F1B4ULL, "art\\levels\\races\\natural\\highlands\\terrainmats\\ter_hpath_da.dds"},
	{0xFE298BCAEA0E0C1EULL, "art\\levels\\races\\ferelden\\props\\middleclass\\prp_diningtable.phy"},
	{0xFE2A385256AD06E2ULL, "6050174_f.fxe"},
	{0xFE2A7D5BF91F4507ULL, "vxlt_0_lowtown_a2_n_e1k_0_0.vlm"},
	{0xFE2AB2EF7B536CD5ULL, "gen00fl_anders6119246.cl"},
	{0xFE2ADF9C4AA7E381ULL, "gen00fl_anders6114476.cl"},
	{0xFE2B79653723593EULL, "vxw_bonepitdestroyed_a3_1an_0_0.vlm"},
	{0xFE2CBF0027D72BE3ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_180_c.impa.evt"},
	{0xFE2CDC1D05E1DCFBULL, "vxdr_primevaltomb_75g_0_0.vlm"},
	{0xFE2CF9E0DFABB950ULL, "vxlt_gallowstemplar_n_46c_0_0.vlm"},
	{0xFE2D38A18E08874DULL, "6110577_m.fxe"},
	{0xFE2DE13B90183C71ULL, "6180521_m.fxe"},
	{0xFE2DE6AAA3F9A729ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztorch.phy"},
	{0xFE2DF44F19EBBA4DULL, "vxht_mansion_n_6kh_0_0.vlm"},
	{0xFE2E615481B66C50ULL, "hlac_w_sundermount_-131069_0z.dds"},
	{0xFE2E752D23F1C5C0ULL, "786535738.ogg"},
	{0xFE2E88B1C025247FULL, "vxdr_primevaltomb_6px_0_1.vlm"},
	{0xFE2F2828877703D4ULL, "6095745_m.fxe"},
	{0xFE2FE50194EC9CBBULL, "cd_qun231_varnell.wbd"},
	{0xFE3024D9557E10F1ULL, "w_bonepitdestroyed_a3_111_0.msh"},
	{0xFE3032D9557E280CULL, "w_bonepitdestroyed_a3_111_0.mao"},
	{0xFE303ED9557E3CAFULL, "w_bonepitdestroyed_a3_111_0.mmh"},
	{0xFE3050F5A3B6BD84ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_c.impa.gad"},
	{0xFE3065A52CE27F19ULL, "222926224.ogg"},
	{0xFE30C9D492FB3F78ULL, "vxdr_primevaltomb_509_0_0.vlm"},
	{0xFE3112AB8837FCE9ULL, "vxht_bartrand_n_6qi_0_0.vlm"},
	{0xFE31359B8691003AULL, "6173612_m.fxe"},
	{0xFE3147A0565991B7ULL, "art\\characters\\weapons\\longswords\\w_lsw_017a.msh"},
	{0xFE3159A05659B01AULL, "art\\characters\\weapons\\longswords\\w_lsw_017a.mao"},
	{0xFE315DA05659B729ULL, "art\\characters\\weapons\\longswords\\w_lsw_017a.mmh"},
	{0xFE318136135B33DFULL, "vxlt_0_lowtown_docks_a3_d1s_0_0.vlm"},
	{0xFE318DA53E7A2A1BULL, "927201334.ogg"},
	{0xFE3194D67AE3FFD3ULL, "zz_app_debug.ncs"},
	{0xFE31AEC3397E6220ULL, "llac_ht_mansion_n_2995_5.dds"},
	{0xFE31D2B3BBE6EBD0ULL, "vxlt_0_lowtown_a2_e8h_0_0.vlm"},
	{0xFE31FDD579E48C58ULL, "6172943_m.fxe"},
	{0xFE3371AC47AD4F50ULL, "vxw_encounter3_at_0_0.vlm"},
	{0xFE3376CC2E5B6A56ULL, "hlac_dr_primevaltomb_9621_4z.dds"},
	{0xFE33980027DD0864ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_180_c.impa.gad"},
	{0xFE33CF270FBBC663ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf02.mmh"},
	{0xFE33DD270FBBDDBDULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf02.msh"},
	{0xFE345FB74F651463ULL, "art\\characters\\playercharacter\\humanmale\\textures\\bdy_arm_medp_n.dds"},
	{0xFE34789832A0D5D5ULL, "vxw_bonepitmining_a2_ad_0_0.vlm"},
	{0xFE349AF80928CA59ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier2_arishokresponse.gad"},
	{0xFE34A32C4F810866ULL, "follower_banter6141601.cl"},
	{0xFE34CA497DC52987ULL, "art\\levels\\races\\ferelden\\props\\lowerclass\\prp_bunklow.phy"},
	{0xFE35153EFDC19A58ULL, "dae000ip_crowd_ff4_talk3.utp"},
	{0xFE3546C4ADB68FD0ULL, "mer120_hubert6070528.cl"},
	{0xFE3592E4D61FDC36ULL, "vxlt_0_lowtown_docks_n_cnl_0_0.vlm"},
	{0xFE361336DD38A002ULL, "vxlt_0_lowtown_docks_n_bfe_0_0.vlm"},
	{0xFE366293C763D25CULL, "569371577.ogg"},
	{0xFE36AC4DB3AAE271ULL, "hf_ave3_guard.mor"},
	{0xFE375FF5A3BCF662ULL, "art\\vfx\\combat\\vfx_m_sta_fire_close_c.impa.ani"},
	{0xFE37A7B9617A8034ULL, "hf_gracious_gillian.mor"},
	{0xFE37D521E4FCB39CULL, "llac_lt_0_lowtown_a2_131070_1.dds"},
	{0xFE385C55AD616C41ULL, "849660075.ogg"},
	{0xFE3864758B2B9CEFULL, "vxht_mansion_n_4f7_0_0.vlm"},
	{0xFE38B0EFB60C9C0FULL, "168932705.ogg"},
	{0xFE38C7C0A37E65C8ULL, "vxht_bartrand_n_4h9_0_0.vlm"},
	{0xFE38CD629073B337ULL, "86143531.ogg"},
	{0xFE390645B940160DULL, "vxht_0_hightown_a3_asu_0_0.vlm"},
	{0xFE397CF20EAA63BBULL, "art\\characters\\weapons\\staff\\textures\\w_stf_st03a_0s.dds"},
	{0xFE399E07C79AA766ULL, "art\\vfx\\combat\\vfx_d_teleport_com_p.impa.evt"},
	{0xFE39B66B695E39FDULL, "6096944_f.fxe"},
	{0xFE3AA70027E34042ULL, "art\\vfx\\combat\\vfx_w_2nd_atk_r_180_c.impa.ani"},
	{0xFE3AB7F5D94A987DULL, "334031696.ogg"},
	{0xFE3AD84276987FFAULL, "hlac_dr_slavers_3214_8z.dds"},
	{0xFE3B313555EB886BULL, "30771683.ogg"},
	{0xFE3B368D675B49CCULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.impa.ani"},
	{0xFE3B678BEEDB3BD7ULL, "642200467.ogg"},
	{0xFE3B682BCE48D299ULL, "hlac_ht_bartrand_n_2995_3x.dds"},
	{0xFE3BABF8092F05A7ULL, "art\\cinematics\\_crit path\\qunari crisis\\quanari compound\\cs_soldier2_arishokresponse.ani"},
	{0xFE3C662C0BE1D2CDULL, "vxht_mansion_n_6ng_0_0.vlm"},
	{0xFE3C9B9EBD4437EDULL, "hlac_ht_mansion2_818_54z.dds"},
	{0xFE3CA7138BF31A84ULL, "566788776.ogg"},
	{0xFE3CE55B8C646342ULL, "zz_vrc_debug6200172.cl"},
	{0xFE3D0884F031168CULL, "hlac_ca_cavern1_haunted_24421_11y.dds"},
	{0xFE3D13C84E602689ULL, "ave331cr_insurgent_f3.utc"},
	{0xFE3D14124DEA5BCAULL, "vrc321_varric6154753.cl"},
	{0xFE3D32A8444D75A2ULL, "vxw_encounter2_42_0_0.vlm"},
	{0xFE3D48042F804077ULL, "vxlt_backalley_g_1y5_0_0.vlm"},
	{0xFE3D5C87ABF3B80BULL, "6117959_f.fxe"},
	{0xFE3D8C635E337B0DULL, "6059995_m.fxe"},
	{0xFE3E729C6A94259EULL, "art\\characters\\playercharacter\\humanmale\\hm_har_ha09a.phy"},
	{0xFE3ED28A42882DBAULL, "gen00fl_aveline6053339.cl"},
	{0xFE3EDE66500C2C90ULL, "vxlt_0_lowtown_docks_a3_8rm_0_0.vlm"},
	{0xFE3F4B5AF501C70AULL, "l_pro000ar_gallow_court.wbd"},
	{0xFE3F82C653784432ULL, "6129182_f.fxe"},
	{0xFE3F8975E52A2E00ULL, "lgt260_eau_tmp_flowergirl6143865.cl"},
	{0xFE3FB9A9C1EE049FULL, "585843657.ogg"},
	{0xFE3FD3DF933A6E1EULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_adjust.evt"},
	{0xFE3FD6F7B532A1DDULL, "gen00fl_varric6105380.cl"},
	{0xFE4074661A07961EULL, "6085849_m.fxe"},
	{0xFE408CEB7995F6BEULL, "mcr371ip_to_courtyard.utp"},
	{0xFE40B507C7A0ECB1ULL, "art\\vfx\\combat\\vfx_d_teleport_com_p.impa.gad"},
	{0xFE40B62ADED5BF2EULL, "vxca_caverns2_808_0_0.vlm"},
	{0xFE40C157478104FEULL, "lowtown_male_lowlife1_sta6165036.cl"},
	{0xFE40C59E67E553F4ULL, "353865356.ogg"},
	{0xFE40C97B0CBA1C70ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_ingots_n.dds"},
	{0xFE40E0F57454C2CFULL, "6113097_m.fxe"},
	{0xFE411EB6DC81353BULL, "vxw_bonepitmining_a2_270_0_0.vlm"},
	{0xFE41543D73F0634DULL, "843504984.ogg"},
	{0xFE418382EA39654AULL, "653729732.ogg"},
	{0xFE4190430CDB1229ULL, "vxlt_hangedman_1vt_0_1.vlm"},
	{0xFE41B187DE61C89DULL, "345918517.ogg"},
	{0xFE41B624700EDFC9ULL, "700789376.ogg"},
	{0xFE420715CD7AAC15ULL, "gen00fl_sebastian6188601.cl"},
	{0xFE4218FD3BE5050DULL, "hlac_ht_mansion2b_818_34z.dds"},
	{0xFE42237FCF41E509ULL, "vxlt_warehouse_n_hw_0_0.vlm"},
	{0xFE42438D67617E0EULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.impa.gad"},
	{0xFE4266624CACC734ULL, "6139454_m.fxe"},
	{0xFE427F8F5FA77BBCULL, "zz_approval_debug6003759.cl"},
	{0xFE42813BEBCB29A4ULL, "llac_dr_slavers_3216_4.dds"},
	{0xFE42BBDCA0A17C6CULL, "lt_0_lowtown_docks_16_0_ll.mmh"},
	{0xFE42EA752496EE55ULL, "vxw_bonepitmining_a2_lc_0_0.vlm"},
	{0xFE440B94FA789DBBULL, "vxw_bonepitmining_a2_1vh_0_0.vlm"},
	{0xFE441BD788B186F3ULL, "vxht_0_hightown_a3_byb_0_0.vlm"},
	{0xFE442ACCE7AA54E6ULL, "gen00fl_anders6142496.cl"},
	{0xFE44CED2CB69EF31ULL, "871128093.ogg"},
	{0xFE44D0270FCA3A88ULL, "art\\characters\\playercharacter\\humanfemale\\mt_ant_hf02.phy"},
	{0xFE4562FF0D93EF46ULL, "vxht_brothel_n_4wj_0_0.vlm"},
	{0xFE45A6B3E19A9133ULL, "hlac_lt_0_lowtown_docks_n_-131075_3y.dds"},
	{0xFE45CB85E0602026ULL, "qun220cr_javaris6071954.cl"},
	{0xFE45EADF933F0149ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_adjust.gad"},
	{0xFE45F4301E677C1DULL, "413788488.ogg"},
	{0xFE462A2C9381EF42ULL, "6061241_m.fxe"},
	{0xFE468D9F1D4BA31BULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_coastsand_da.dds"},
	{0xFE47536CA8B44D68ULL, "art\\characters\\weapons\\staff\\textures\\w_stf_011a_0n.dds"},
	{0xFE478074F7C36D0FULL, "vxlt_0_lowtown_a2_n_eom_0_0.vlm"},
	{0xFE47B607C7A70D5FULL, "art\\vfx\\combat\\vfx_d_teleport_com_p.impa.ani"},
	{0xFE47F76EEC24434CULL, "6093695_m.fxe"},
	{0xFE482B0063E8DA3AULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16b.msh"},
	{0xFE483D0063E8F880ULL, "art\\characters\\playercharacter\\shared\\heads\\mt_amt_nos16b.mmh"},
	{0xFE483E8D6765E1EDULL, "art\\vfx\\combat\\vfx_r_dwp_combo_loop_c.impa.evt"},
	{0xFE4847781D3758AFULL, "vxlt_gallowstemplar_n_5at_0_0.vlm"},
	{0xFE491BFDDF15EE01ULL, "cod_tomes_tome_5.utp"},
	{0xFE49CE37F4DC770FULL, "988752530.ogg"},
	{0xFE4A0DB4EDEAF58CULL, "art\\levels\\ztools\\default\\generic\\9ff_box.phy"},
	{0xFE4A248094CDFEA9ULL, "33832879.ogg"},
	{0xFE4AA022CA2B2CD0ULL, "llac_lt_blackemporium_1055_22.dds"},
	{0xFE4AF0FCD0324DB8ULL, "art\\levels\\races\\proxy\\tilesets\\setint\\4co_basic\\1wo_diagonal_1.mao"},
	{0xFE4B396A5C963345ULL, "6103022_f.fxe"},
	{0xFE4B3E5A73989DFDULL, "hlac_lt_0_lowtown_docks_n_-65539_3z.dds"},
	{0xFE4B712826B41632ULL, "844349552.ogg"},
	{0xFE4B9B9CFDFB11EDULL, "character_character_p284.psh.fxo.dx11"},
	{0xFE4BFE405029938DULL, "6056895_m.fxe"},
	{0xFE4CBBDF9344CF97ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\sitting_table_adjust.ani"},
	{0xFE4D016C0A8A989DULL, "dre000_cut_bartrandbetrayal6033815.cl"},
	{0xFE4D83A4A7F43F0AULL, "mrl310_merrill6120486.cl"},
	{0xFE4E5619D1E7EBD6ULL, "73823474.ogg"},
	{0xFE4E7DF9E3E28D1EULL, "vxht_mansion2b_1tj_0_0.vlm"},
	{0xFE4ED2FAF9FEEABDULL, "6204968_m.fxe"},
	{0xFE4F8A6422A51A7BULL, "919765026.ogg"},
	{0xFE502EFC3BC76221ULL, "gen00fl_varric6105744.cl"},
	{0xFE50955CC137FD89ULL, "vxlt_warehouse_n_ov_0_0.vlm"},
	{0xFE5135E9904A594CULL, "dae000_arishok6075374.cl"},
	{0xFE513B0149827996ULL, "vxht_chantry_n_2vd_0_0.vlm"},
	{0xFE51408EF0396929ULL, "fen110_fenris6118101.cl"},
	{0xFE518BFDBCD8C6D9ULL, "pro000cr_barrier_disciple.utc"},
	{0xFE5193507031B417ULL, "hlac_lt_0_lowtown_a2_n_c_131070_5x.dds"},
	{0xFE519BD88AA2F0ACULL, "60056577.ogg"},
	{0xFE51A07370B6A7FCULL, "vxw_sundermount_373_0_0.vlm"},
	{0xFE51A2ACE594655BULL, "vxht_brothel_n_6bj_0_0.vlm"},
	{0xFE51A35056F51375ULL, "6104187_m.fxe"},
	{0xFE522956BA8C03AFULL, "llac_lt_gallowstemplar_f_7198_3.dds"},
	{0xFE5238B2E5648EBCULL, "6135960_m.fxe"},
	{0xFE523DA5E34BFD6BULL, "6150140_m.fxe"},
	{0xFE524C051254D8C0ULL, "6067758_f.fxe"},
	{0xFE5254AC6E295F32ULL, "378851176.ogg"},
	{0xFE52DF2C86834721ULL, "6136375_m.fxe"},
	{0xFE535831C7C1F8E5ULL, "ser221_gascard6175526.cl"},
	{0xFE53B6F0EEBAF1B1ULL, "follower_banter6161127.cl"},
	{0xFE540E709D9B5AD4ULL, "hlac_ht_0_hightown_-131072_1y.dds"},
	{0xFE545181E5A8AB81ULL, "l_dae221ar_chantry.wbd"},
	{0xFE54B2902CB7601FULL, "452903174.ogg"},
	{0xFE54CB754DCCBB6EULL, "199409958.ogg"},
	{0xFE5534917C6F750AULL, "ico_armor_heavy_2.dds"},
	{0xFE55A5CBC13F430DULL, "plt_ave000pt_endstage.plo"},
	{0xFE55DD23330706A7ULL, "ht_0_hightown_n_28_0_ll.mmh"},
	{0xFE55FCFFD45CF1A8ULL, "hlac_w_encounter2_393215_0y.dds"},
	{0xFE560DC5D685DB6CULL, "508993312.ogg"},
	{0xFE563886337769DBULL, "333775670.ogg"},
	{0xFE566EAB881F7EFBULL, "art\\levels\\atmosphere\\sky\\sb_day_base.mao"},
	{0xFE56707E3ED7F272ULL, "one100_party_comments6112674.cl"},
	{0xFE56C353F3C8EDE3ULL, "6208632_m.fxe"},
	{0xFE56D130D4B2DDAFULL, "ht_0_hightown_29_0_ll.mmh"},
	{0xFE56D6FCC6FDAED4ULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_merrill.gad"},
	{0xFE57362261861DD3ULL, "6016804_m.fxe"},
	{0xFE577CF469B060C7ULL, "264826388.ogg"},
	{0xFE57C95150188C10ULL, "llac_ht_0_hightown_n_65535_2.dds"},
	{0xFE57FAA57595C4D7ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_arm_medp_0d.dds"},
	{0xFE58A704785F1F97ULL, "w_blightlands_202_0.msh"},
	{0xFE58B904785F3EFAULL, "w_blightlands_202_0.mao"},
	{0xFE58BD04785F4509ULL, "w_blightlands_202_0.mmh"},
	{0xFE598C08765566E9ULL, "gen00fl_fenris6105854.cl"},
	{0xFE59CFAAA41EBB4EULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztorch.mmh"},
	{0xFE59D8034E23D2C3ULL, "6079800_f.fxe"},
	{0xFE59DBAAA41ECFE5ULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztorch.mao"},
	{0xFE59E9AAA41EE73CULL, "art\\levels\\races\\proxy\\props\\furniture\\prp_zztorch.msh"},
	{0xFE5A1B0E8626130BULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamfast.dura.evt"},
	{0xFE5A33739299ABC8ULL, "dae000_keep_guard_1_sta6163797.cl"},
	{0xFE5A459927981272ULL, "6110260_m.fxe"},
	{0xFE5A6035FA993FD3ULL, "hlac_lt_undercitydungeon_-1_34y.dds"},
	{0xFE5AA83612F0B835ULL, "vxw_sundermount_21q_0_0.vlm"},
	{0xFE5AB13556A536B1ULL, "1466196.ogg"},
	{0xFE5AC209351C24E5ULL, "zz_one_debug6069470.cl"},
	{0xFE5ADC3C21351566ULL, "826207513.ogg"},
	{0xFE5B1A01443B86A6ULL, "651434553.ogg"},
	{0xFE5B5BA85CDB6E4DULL, "gen00fl_sebastian6189912.cl"},
	{0xFE5BE701CAC8A963ULL, "6129482_f.fxe"},
	{0xFE5C122E34AB5821ULL, "6068632.ogg"},
	{0xFE5C9FC65C31280DULL, "dae000_keep_noblewoman_sta6162044.cl"},
	{0xFE5CD2494C6028DFULL, "hlac_ht_mansion2b_818_5z.dds"},
	{0xFE5D03603ABDFCC2ULL, "w_encounter3_236_0.msh"},
	{0xFE5D05603ABE0078ULL, "w_encounter3_236_0.mmh"},
	{0xFE5D11603ABE141BULL, "w_encounter3_236_0.mao"},
	{0xFE5D88B39548B9C1ULL, "32149549.ogg"},
	{0xFE5DE5FCC703E6B2ULL, "art\\cinematics\\act3\\romances\\merrill\\rom000cs_merrill_merrill.ani"},
	{0xFE5E05FA3FA681CFULL, "docks_scene_dockworkers6176472.cl"},
	{0xFE5E09E32D1D1BEEULL, "sst_ss_coterie_assassin_m.gda"},
	{0xFE5E2C4133D2CAD7ULL, "1003853737.ogg"},
	{0xFE5E684FE59693C6ULL, "llac_lt_0_lowtown_a2_n_c_-196608_1.dds"},
	{0xFE5E9EFAC1CBB16CULL, "vis121_saemus6118476.cl"},
	{0xFE5EA522187EEC99ULL, "316532522.ogg"},
	{0xFE5F0739B32E1609ULL, "vxlt_warehouse_n_ru_0_0.vlm"},
	{0xFE5F5C576F4ED2F5ULL, "ttx_roof_arch_1.msh"},
	{0xFE5F5E576F4ED66BULL, "ttx_roof_arch_1.mmh"},
	{0xFE5FAD7670EF4379ULL, "hlac_lt_0_lowtown_n_131073_0x.dds"},
	{0xFE5FCBD797B2138FULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.impa.gad"},
	{0xFE60125062ACC07CULL, "vxw_sundermount_320_0_0.vlm"},
	{0xFE606183EBF68ADDULL, "zz_fol_debug6057016.cl"},
	{0xFE6062FCB4AE3B8BULL, "follower_banter6156039.cl"},
	{0xFE60BA581A94E09DULL, "6118476_m.fxe"},
	{0xFE60C63F4972F190ULL, "103091871.ogg"},
	{0xFE611CFDA4E34DAEULL, "darktown_fem_fereldan_sta6156749.cl"},
	{0xFE61340E862C5CCCULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamfast.dura.gad"},
	{0xFE61458136CF1D98ULL, "6118855_m.fxe"},
	{0xFE632CF1D6871A5BULL, "vxw_blightlands_9g_0_0.vlm"},
	{0xFE63826DACD301C1ULL, "al_warrior_s02.gda"},
	{0xFE6388C308AE5CEFULL, "art\\characters\\creatures\\strider\\animation\\groundstomploop.gad"},
	{0xFE64642EDBAF9E0FULL, "vxlt_0_lowtown_a2_n_euo_0_0.vlm"},
	{0xFE65E0D797B6A33CULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.impa.evt"},
	{0xFE673F9D50851D1CULL, "233887445.ogg"},
	{0xFE6763A9B901910AULL, "hlac_lt_darktown_-1_8x.dds"},
	{0xFE67AF074F1792E5ULL, "6136340_m.fxe"},
	{0xFE683066545EF149ULL, "147178726.ogg"},
	{0xFE68330E863279CAULL, "art\\vfx\\environment\\level_fx\\smoke\\vfx_steamfast.dura.ani"},
	{0xFE683B7C5B14FEB9ULL, "vxht_mansion2b_1w8_0_0.vlm"},
	{0xFE6847377D7681DFULL, "709609798.ogg"},
	{0xFE6864059081162AULL, "vxht_0_hightown_b01_0_0.vlm"},
	{0xFE687FB1F61F29E2ULL, "vxw_bonepitmining_a2_25k_0_0.vlm"},
	{0xFE694E128802591CULL, "vxca_caverns1_dpi_0_0.vlm"},
	{0xFE694FC4546C867BULL, "gen00fl_fenris6126392.cl"},
	{0xFE6A0B5A25CDCC5EULL, "vxht_keep_n_73j_0_0.vlm"},
	{0xFE6A9DC308B49E9CULL, "art\\characters\\creatures\\strider\\animation\\groundstomploop.evt"},
	{0xFE6B0767F43028D6ULL, "hlac_lt_0_lowtown_docks_n_131070_1x.dds"},
	{0xFE6B079DE29B3F1CULL, "vrc321st_spooky_scene_5.stg"},
	{0xFE6B23991530078EULL, "286231000.ogg"},
	{0xFE6B5C4E5652403DULL, "6034023_f.fxe"},
	{0xFE6B6D92EE6F41BBULL, "hla_ht_0_hightown_n_65537_0x.dds"},
	{0xFE6B927B81033822ULL, "vxlt_0_lowtown_a3_n_c_byz_0_0.vlm"},
	{0xFE6BB9038B6EEEEAULL, "hlac_w_encounter2_camp_4_0z.dds"},
	{0xFE6BC7BF001DDC4CULL, "hlac_lt_0_lowtown_a2_n_0_8x.dds"},
	{0xFE6C0D35CAA509C3ULL, "vxca_cavern1_haunted_jle_0_0.vlm"},
	{0xFE6C2B8D62C85305ULL, "vxw_woundedcoast_h5_0_1.vlm"},
	{0xFE6C319955BAE45FULL, "394668065.ogg"},
	{0xFE6D2CF94D8FD397ULL, "t1_stb_rbr.tnt"},
	{0xFE6D405FA7D90E39ULL, "pro000_friedrich6050152.cl"},
	{0xFE6D58E6A74B6704ULL, "mag120_macha6080667.cl"},
	{0xFE6D7916A5242E89ULL, "vxlt_warehouse_n_yt_0_0.vlm"},
	{0xFE6EBE5B46F9F0CFULL, "vxw_bonepitmining_a2_8m_0_0.vlm"},
	{0xFE6F6C311A53F9BDULL, "1041854710.ogg"},
	{0xFE71D60C5FE79446ULL, "995592236.ogg"},
	{0xFE72BEA20709FBE0ULL, "vxlt_0_lowtown_a2_byx_0_0.vlm"},
	{0xFE72C91CA697C4BAULL, "hlac_ca_caverns1_mines_24422_0z.dds"},
	{0xFE7393B69FD632B5ULL, "ave310_party_comment6134844.cl"},
	{0xFE73C5811BA3D5F5ULL, "6102568_f.fxe"},
	{0xFE73CEB89A17922FULL, "234844930.ogg"},
	{0xFE73E8D797C2EE55ULL, "art\\vfx\\creatures\\fxc_bm_tnt_app_c.impa.ani"},
	{0xFE748A99A1E48B4AULL, "229806005.ogg"},
	{0xFE748B262E5E6488ULL, "6067331_m.fxe"},
	{0xFE74DDEDD7B032E5ULL, "fen110_anso6170068.cl"},
	{0xFE7561EF6426197EULL, "138133999.ogg"},
	{0xFE759E0F47BB24AEULL, "vxht_keep_3zj_0_0.vlm"},
	{0xFE75BB7704066B27ULL, "hlac_lt_0_lowtown_docks_a3_-65537_5z.dds"},
	{0xFE75C9B331C176C5ULL, "6047400_m.fxe"},
	{0xFE75DE65D3E7AF70ULL, "character_character_p133.psh.fxo.dx11.assembly"},
	{0xFE769E7B89489314ULL, "dae321ar_brothel_nt.are"},
	{0xFE76B69C8583D1A2ULL, "1057218181.ogg"},
	{0xFE76EA3ED16A8814ULL, "vxht_0_hightown_a2_a2j_0_0.vlm"},
	{0xFE76F22E61109D26ULL, "sky_ht_chantry.mao"},
	{0xFE7712268150A1BDULL, "vxlt_foundry_n_y2_0_0.vlm"},
	{0xFE77A5C308BF36B5ULL, "art\\characters\\creatures\\strider\\animation\\groundstomploop.ani"},
	{0xFE77E0327AD81DBBULL, "terrain1.vsh.fxo"},
	{0xFE78BCCDA659B4E2ULL, "6214661_m.fxe"},
	{0xFE79469F8E39FA95ULL, "vxdr_dungeon_1me_0_0.vlm"},
	{0xFE79AD1741AD947AULL, "hlac_ht_mansion_n_3366_0z.dds"},
	{0xFE7A6B0422EDCE14ULL, "dre000_varric6036965.cl"},
	{0xFE7A7381E5C9B483ULL, "l_dae221ar_chantry.bnk"},
	{0xFE7B5B2C767785E0ULL, "6149754_m.fxe"},
	{0xFE7BE2CAD5075132ULL, "6086174_m.fxe"},
	{0xFE7BF168E1C90FD7ULL, "llac_lt_0_lowtown_a2_n_-131072_8.dds"},
	{0xFE7C50D1282A0F90ULL, "ser221_gascard6122833.cl"},
	{0xFE7CA92497515B0AULL, "dae000ip_crowd_em4_talk10.utp"},
	{0xFE7CEE34118EF52CULL, "vxlt_0_lowtown_a3_n_c_hpi_0_0.vlm"},
	{0xFE7CF843BB76AEDBULL, "vxht_brothel_n_6uo_0_0.vlm"},
	{0xFE7D17A20BD4007BULL, "llac_ht_0_hightown_n_196607_0.dds"},
	{0xFE7D1870DCBDF339ULL, "hlac_lt_backalley_g_0_4y.dds"},
	{0xFE7D9E11FE81119DULL, "6112676_m.fxe"},
	{0xFE7DC3A20A52734AULL, "vxlt_foundry_n_vh_0_0.vlm"},
	{0xFE7DF4D68FB328A5ULL, "6142116_m.fxe"},
	{0xFE7E632537CADB32ULL, "dae000ip_crowd_ff1_talk8.utp"},
	{0xFE7E649D097078EAULL, "984653315.ogg"},
	{0xFE7EB96C2824ABF0ULL, "art\\vfx\\combat\\vfx_v_bld_cast_c.mmh"},
	{0xFE7EC6A5E3994D42ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.mmh"},
	{0xFE7F88556CD15274ULL, "6181906_f.fxe"},
	{0xFE7F97F939463670ULL, "354002833.ogg"},
	{0xFE8132EBCCB50B5BULL, "vxw_sundermount_26z_0_0.vlm"},
	{0xFE8346B94F102CF5ULL, "zz_followers_banter6158442.cl"},
	{0xFE83717A1E45C71DULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_004a_0n.dds"},
	{0xFE8435C7AE8D60FFULL, "756255966.ogg"},
	{0xFE855A6D0FA6DB52ULL, "vxlt_0_lowtown_a2_gyk_0_0.vlm"},
	{0xFE85B7ABFD6084F9ULL, "t3_crd_lt2.tnt"},
	{0xFE85BF55286024A2ULL, "vxw_blightlands_w4_0_0.vlm"},
	{0xFE85D458AC5DBC3FULL, "vxlt_warehouse_n_155_0_0.vlm"},
	{0xFE85EA1104F22832ULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\0wi_grand_1.mao"},
	{0xFE86295AF53D9020ULL, "l_pro000ar_gallow_court.bnk"},
	{0xFE864D9442920448ULL, "lt_0_lowtown_docks_n_33_0.anb"},
	{0xFE869D2475519B12ULL, "833092029.ogg"},
	{0xFE8704E5F1583A20ULL, "502749144.ogg"},
	{0xFE87526E7659ECC5ULL, "vxlt_darktown_56t_0_0.vlm"},
	{0xFE87934005A0BBE3ULL, "vxw_sundermount_30p_0_1.vlm"},
	{0xFE8832238A8A3290ULL, "mag291_vincento6027708.cl"},
	{0xFE898BEE0345CC69ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_wallssoft_n.dds"},
	{0xFE899D0E5A61B947ULL, "6153758_m.fxe"},
	{0xFE89A4514C4EC40AULL, "910503860.ogg"},
	{0xFE89C49434F09D94ULL, "art\\vfx\\materials\\vfx_d_speed_blend.mao"},
	{0xFE8A008B1822F719ULL, "vxw_woundedcoast_15z_0_0.vlm"},
	{0xFE8A10BEAD671AB6ULL, "6097870_f.fxe"},
	{0xFE8AB261C4332897ULL, "980853877.ogg"},
	{0xFE8AC8999D8966C2ULL, "m_dae211ar_tavern.bnk"},
	{0xFE8AE4B4EE224757ULL, "art\\levels\\ztools\\default\\generic\\9ff_box.mmh"},
	{0xFE8AE8B4EE224E64ULL, "art\\levels\\ztools\\default\\generic\\9ff_box.mao"},
	{0xFE8AFAB4EE226CC9ULL, "art\\levels\\ztools\\default\\generic\\9ff_box.msh"},
	{0xFE8B0E92D17A4CD6ULL, "vrc321_party_barks6095758.cl"},
	{0xFE8B83B657EA4FF9ULL, "528486277.ogg"},
	{0xFE8BD9445D1A81A6ULL, "gen00fl_varric6076443.cl"},
	{0xFE8C223D8EDC2325ULL, "vxlt_undercityserialkiller_1ac_0_0.vlm"},
	{0xFE8C28F325CD1A60ULL, "998629962.ogg"},
	{0xFE8CDC6C2831240FULL, "art\\vfx\\combat\\vfx_v_bld_cast_c.anb"},
	{0xFE8D21A7E044F098ULL, "vxht_0_hightown_a2_c4r_0_0.vlm"},
	{0xFE8D3E2885F74B3EULL, "gen_im_npc_wep_one_axea.uti"},
	{0xFE8D5189AC7D46C4ULL, "art\\characters\\creatures\\gibgolem\\uni_har.msh"},
	{0xFE8D6789AC7D6C66ULL, "art\\characters\\creatures\\gibgolem\\uni_har.mmh"},
	{0xFE8E486696268033ULL, "hlac_ht_keep_3735_1y.dds"},
	{0xFE8E4A9EA0C9DA64ULL, "zz_mrl_debug6184197.cl"},
	{0xFE8F30890231B56BULL, "vxht_keep_62z_0_0.vlm"},
	{0xFE8F4BE3DFDA3CC1ULL, "6039610_m.fxe"},
	{0xFE8F5135A3C53AECULL, "28571670.ogg"},
	{0xFE90439F17047770ULL, "llac_lt_0_lowtown_a3_n_c_-65537_6.dds"},
	{0xFE906324D341F913ULL, "mer220_hubert6050047.cl"},
	{0xFE90B5233740CEB9ULL, "llac_ht_mansion2b_818_14.dds"},
	{0xFE90BFEDB3357FA5ULL, "follower_banter6136365.cl"},
	{0xFE90FE2FD281BFEDULL, "w_bonepitmining_a2_61_0.msh"},
	{0xFE910C2FD281D720ULL, "w_bonepitmining_a2_61_0.mao"},
	{0xFE91102FD281DE73ULL, "w_bonepitmining_a2_61_0.mmh"},
	{0xFE91444D15D1193DULL, "bder_proj_01.mao"},
	{0xFE914EE9B3B7A035ULL, "art\\characters\\creatures\\flemethdragon\\textures\\cn_fdra_bdy_0n.dds"},
	{0xFE918FF611670BB6ULL, "llac_lt_gallowstemplar_n_2391_2.dds"},
	{0xFE9240E191FE9BA3ULL, "mcr370_assignments6176752.cl"},
	{0xFE925B15906F1190ULL, "vxht_bartrand_n_66i_0_0.vlm"},
	{0xFE930E06E327BB0CULL, "371477027.ogg"},
	{0xFE93112B5F2009BAULL, "240864671.ogg"},
	{0xFE93619C0D317F8BULL, "vxht_mansion2b_24h_0_0.vlm"},
	{0xFE9401C3C203AAEDULL, "hlac_lt_0_lowtown_a2_n_c_131072_5x.dds"},
	{0xFE9407926CE81BC5ULL, "vxw_woundedcoast_1d1_0_0.vlm"},
	{0xFE95829C1F1BC668ULL, "lgt220cr_shp_mag_bofort.utc"},
	{0xFE95857B896319F6ULL, "dae321ar_brothel_nt.ncs"},
	{0xFE95BB0A4A465C66ULL, "vxht_bartrand_n_5w3_0_0.vlm"},
	{0xFE9606AF6A7FE31FULL, "648576999.ogg"},
	{0xFE9645B6276E47DAULL, "hlac_lt_undercityencounter_-1_20z.dds"},
	{0xFE96C14661A4A3A3ULL, "6076147_m.fxe"},
	{0xFE975E89AC85B131ULL, "art\\characters\\creatures\\gibgolem\\uni_har.phy"},
	{0xFE97815C38B0DA7CULL, "6105876_f.fxe"},
	{0xFE97B14C3E00412AULL, "vxht_brothel_n_4hk_0_0.vlm"},
	{0xFE97CDC317B8675DULL, "hlac_lt_0_lowtown_0_4z.dds"},
	{0xFE97F059100FDC0CULL, "dae000_tomwise6167320.cl"},
	{0xFE980D4D1765DF71ULL, "dae000cr_dalish_apprentice1.utc"},
	{0xFE980EEB4C57D1C7ULL, "6156458_m.fxe"},
	{0xFE982E8B50D3B2C4ULL, "6149379_f.fxe"},
	{0xFE9872680A190F99ULL, "vxw_woundedcoast_10y_0_0.vlm"},
	{0xFE98995F2079ACCCULL, "ss_crazy_m_1.bnk"},
	{0xFE98B3EFC5475D9BULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mth.mmh"},
	{0xFE98D1EFC5479085ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mth.msh"},
	{0xFE99054919E5C720ULL, "270393217.ogg"},
	{0xFE9967BECC949850ULL, "vxca_caverns1_i39_0_0.vlm"},
	{0xFE9A1FCBDC61B58EULL, "and110_refugee6111598.cl"},
	{0xFE9A2F2360E8EA3EULL, "6162266_f.fxe"},
	{0xFE9A5032F65E6193ULL, "330712573.ogg"},
	{0xFE9A93FB48B249D3ULL, "vxw_sundermount_2zw_0_1.vlm"},
	{0xFE9ADEBAD27BDEE4ULL, "scaleform_sf_main_p0.vsh.fxo.dx11"},
	{0xFE9B02F04FB7A64AULL, "552932945.ogg"},
	{0xFE9B9384D23B0918ULL, "vxht_0_hightown_a2_c3s_0_0.vlm"},
	{0xFE9B9BB0CA8F2098ULL, "6085788_m.fxe"},
	{0xFE9B9D7B8967AD8FULL, "dae321ar_brothel_nt.lst"},
	{0xFE9BA52C4C11F611ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_flat.gad"},
	{0xFE9BCAE7930DC287ULL, "vxlt_0_lowtown_docks_a3_n_cx8_0_0.vlm"},
	{0xFE9BDEA0AFE142EEULL, "vxdr_primevaltomb_6nh_0_2.vlm"},
	{0xFE9BEA8E09829DB5ULL, "gen_im_trs_gen_sta_phy_r0.uti"},
	{0xFE9C15309671CBC7ULL, "248820982.ogg"},
	{0xFE9C75823EF142F1ULL, "544003478.ogg"},
	{0xFE9C98D84EC2AA55ULL, "6187960_m.fxe"},
	{0xFE9D083AE91F1B06ULL, "337408779.ogg"},
	{0xFE9DEC0E95D11FDBULL, "6066331_m.fxe"},
	{0xFE9DFF659A5857CAULL, "vxlt_gallowstemplar_f_5xb_0_0.vlm"},
	{0xFE9E0A6625BECEB0ULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm05.mmh"},
	{0xFE9E186625BEE6CAULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm05.msh"},
	{0xFE9E362D52909ADCULL, "llac_ht_mansion2_818_22.dds"},
	{0xFE9EA6B0F1AE729CULL, "vxw_bonepit_2ao_0_0.vlm"},
	{0xFE9EBABA7C21ED72ULL, "vfx_vfx_p3.psh.fxo.dx11"},
	{0xFE9EC7546204AA0AULL, "436438110.ogg"},
	{0xFE9EF48B98753990ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_cenotaph_n.dds"},
	{0xFE9F6C463C9EF0FEULL, "gen00fl_aveline6121922.cl"},
	{0xFE9F918D11CB28EDULL, "qun200_party_comment6095266.cl"},
	{0xFEA02B50325B5C54ULL, "hlac_lt_gallowstemplar_2391_8y.dds"},
	{0xFEA04A4346CB0423ULL, "art\\vfx\\combat\\vfx_gib_qunw_ltrb.anb"},
	{0xFEA085D57187BD49ULL, "vxlt_gallowsprison_n_6o7_0_0.vlm"},
	{0xFEA0870375CB99BDULL, "6034671_m.fxe"},
	{0xFEA0CCF282652A10ULL, "vxht_bartrand_n_63v_0_0.vlm"},
	{0xFEA150286BA4D0AAULL, "6179046_f.fxe"},
	{0xFEA1D378FF27980BULL, "vxht_mansion2b_21k_0_0.vlm"},
	{0xFEA222C16370788CULL, "one350_variel6071051.cl"},
	{0xFEA2A62C4C1816BFULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\ambient\\compressed\\sitting_table_right_hand_flat.ani"},
	{0xFEA2D762304BEE7AULL, "vxw_woundedcoast_1p6_0_0.vlm"},
	{0xFEA339BB95BC1781ULL, "zz_qun_debug6049744.cl"},
	{0xFEA348A8815A06FBULL, "56700988.ogg"},
	{0xFEA384E31D520FE7ULL, "1058936872.ogg"},
	{0xFEA3C243CEF7D432ULL, "vxlt_hangedman_n_262_0_0.vlm"},
	{0xFEA3DADF1FA2CD01ULL, "vxht_chantry_35x_0_0.vlm"},
	{0xFEA40CC910AE6ADAULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_nicechair2_d.dds"},
	{0xFEA42CE73C3C74E6ULL, "vxht_bartrand_n_5z4_0_0.vlm"},
	{0xFEA4399847A4AD2CULL, "6106647_m.fxe"},
	{0xFEA43BF8E5957682ULL, "vxlt_gallowstemplar_n_520_0_0.vlm"},
	{0xFEA442FA43D0E0BDULL, "6167900_m.fxe"},
	{0xFEA4A4EA7098C9CAULL, "art\\vfx\\cinematic\\vfx_blood_b_splatters_p.anb"},
	{0xFEA4AD6E97978902ULL, "rune_weapon_lyrium.dds"},
	{0xFEA4B669FAC3CD94ULL, "m_qcr221ar_keep.bnk"},
	{0xFEA4CDF2D549B864ULL, "vxca_cave_2fq_0_0.vlm"},
	{0xFEA4E44C239D7EA6ULL, "vxlt_gamlenshouse_n_2t_0_0.vlm"},
	{0xFEA5120809BD3B26ULL, "art\\cinematics\\_crit path\\mage crisis\\lowtown\\textures\\windowglowcombined_e.dds"},
	{0xFEA62E61F55D79C4ULL, "art\\characters\\playercharacter\\humanmale\\textures\\arm_bdaa_0e.dds"},
	{0xFEA66371A759BC75ULL, "vxht_mansion2_3cv_0_0.vlm"},
	{0xFEA6ABD72B15132DULL, "gen00fl_aveline6137704.cl"},
	{0xFEA6FB0A0EE7EDA0ULL, "demo_rockwraith_warrior.cut"},
	{0xFEA70BE708DDC641ULL, "6070811_m.fxe"},
	{0xFEA73EDE92FE9959ULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody3_.ani"},
	{0xFEA79A543469F95CULL, "12269758.ogg"},
	{0xFEA7C5A5E3BC50E9ULL, "art\\vfx\\combat\\vfx_m_pri_infer_aoe2_p.anb"},
	{0xFEA7D824950ED76AULL, "ht_0_hightown_a2_water_15346.mao"},
	{0xFEA7DC24950EDDB9ULL, "ht_0_hightown_a2_water_15346.mmh"},
	{0xFEA7E624950EEEA7ULL, "ht_0_hightown_a2_water_15346.msh"},
	{0xFEA7FABE68F7FA6FULL, "17991484.ogg"},
	{0xFEA82B3F30D5E175ULL, "vxlt_gallowscourtyard_n_7ez_0_0.vlm"},
	{0xFEA8C80F5C67DC94ULL, "gen00fl_isabela6098193.cl"},
	{0xFEA8C91D3DAEC4ACULL, "gen_im_trs_apo_sta_ele_r2.uti"},
	{0xFEA8D271F9EACC5EULL, "6042794_m.fxe"},
	{0xFEA8EE446AE32446ULL, "6187927_m.fxe"},
	{0xFEA8F0F7C1724984ULL, "zz_and_debug6134147.cl"},
	{0xFEA979C1C175C3B3ULL, "ave200_donnic6059892.cl"},
	{0xFEA9E7B52B8597CCULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_large.msh"},
	{0xFEA9EDB52B85A1BEULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_large.mmh"},
	{0xFEA9EEB23E4D295BULL, "293347104.ogg"},
	{0xFEA9F9B52B85B655ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_large.mao"},
	{0xFEAA3C720FB3C5A4ULL, "6021118_f.fxe"},
	{0xFEAADE1E7B2CD8D0ULL, "128297650.ogg"},
	{0xFEAAF6B543C6A2D0ULL, "gen_im_arm_boo_lgt_champion.uti"},
	{0xFEAB0D13F823CC17ULL, "zz_fol_debug6066137.cl"},
	{0xFEAB284F8E27C731ULL, "6095087_m.fxe"},
	{0xFEAB9D674CF37B0FULL, "hf_bethany_h3s1.mor"},
	{0xFEAC0178A42E1966ULL, "mag131_keran6094502.cl"},
	{0xFEAC71428C4E704AULL, "vxlt_gallowstemplar_f_5cc_0_0.vlm"},
	{0xFEACB528CF25AACFULL, "6165966_m.fxe"},
	{0xFEADCE6AB2A5D0B1ULL, "6139983_m.fxe"},
	{0xFEAE19DE9304789BULL, "art\\cinematics\\_crit path\\mage crisis\\gallows arrival\\gallowsarrivedeadbody3_.gad"},
	{0xFEAE6368AA47C109ULL, "6070495_m.fxe"},
	{0xFEAEE21088B079B1ULL, "vxht_0_hightown_n_90b_0_0.vlm"},
	{0xFEAF28FEB5E1BC79ULL, "plt_lgt000pt_baz_cache.plo"},
	{0xFEAF511FFCF857C2ULL, "cd_mag360_thrask.wbd"},
	{0xFEAF6850C18B1AC4ULL, "441202463.ogg"},
	{0xFEAF885E592BBF1CULL, "genip_bones_pile_3.utp"},
	{0xFEAFD1B0F0B64379ULL, "gen00fl_anders6128350.cl"},
	{0xFEAFDEAFC1BE14A6ULL, "185285818.ogg"},
	{0xFEB01C8790E346F2ULL, "6004217_m.fxe"},
	{0xFEB04DFD4AA66898ULL, "191136962.ogg"},
	{0xFEB0B2999DAA10E0ULL, "m_dae211ar_tavern.wbd"},
	{0xFEB19632ADC341F4ULL, "79216102.ogg"},
	{0xFEB202606FFE8181ULL, "954871766.ogg"},
	{0xFEB21C0670F3658AULL, "vxw_bonepitmining_a2_240_0_0.vlm"},
	{0xFEB239205DDED32EULL, "vxdr_primevaltomb_52y_0_0.vlm"},
	{0xFEB2F0EDC8327126ULL, "atl_quickbar_dxt5_dat.xml"},
	{0xFEB3499CF1582AD7ULL, "vxlt_0_lowtown_a2_fhx_0_0.vlm"},
	{0xFEB4B67082958664ULL, "zz_mer_debug6028948.cl"},
	{0xFEB574505672B50EULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.impa.ani"},
	{0xFEB58758966E9773ULL, "vxht_mansion2_2r2_0_0.vlm"},
	{0xFEB649DAED8DD8F8ULL, "555256965.ogg"},
	{0xFEB6843FFF7C9B98ULL, "art\\levels\\races\\natural\\proxy\\terrainmats\\ter_bigrock01_sa.dds"},
	{0xFEB69D1C22CBF9F5ULL, "vxlt_gallowscourtyard_n_7hy_0_0.vlm"},
	{0xFEB6B5A7D967572CULL, "vxlt_0_lowtown_a3_n_c_hlm_0_0.vlm"},
	{0xFEB6C38243C8C757ULL, "vxht_keep_70i_0_0.vlm"},
	{0xFEB6D8FFC7B6347BULL, "hlac_w_sundermount_327682_0z.dds"},
	{0xFEB6DF569B2D229DULL, "mrl000_marethari.bnk"},
	{0xFEB71CA0D65677A9ULL, "ser121_jethann6104616.cl"},
	{0xFEB737EE7D83D763ULL, "6134458_m.fxe"},
	{0xFEB77471771B28A3ULL, "6064896_m.fxe"},
	{0xFEB7BACDAB04A3C8ULL, "6107705_m.fxe"},
	{0xFEB80D5BEB707A9AULL, "dae000_bartrand6030624.cl"},
	{0xFEB834874E275F7CULL, "6111749_f.fxe"},
	{0xFEB85C2B19F76AEEULL, "vxht_mansion2b_2rs_0_0.vlm"},
	{0xFEB87BDAD0F5021FULL, "6171048_m.fxe"},
	{0xFEB8DE78DEC87D8CULL, "951239746.ogg"},
	{0xFEB8E3BFD702E181ULL, "lt_0_lowtown_docks_a3_32_0.anb"},
	{0xFEB8FE32E562232DULL, "6024577_m.fxe"},
	{0xFEB93CB4744FF522ULL, "lgt160_wk_soph_hunter6119754.cl"},
	{0xFEB9423587AEC362ULL, "6106564_m.fxe"},
	{0xFEB99816A89917C2ULL, "vxht_bartrand_n_67x_0_0.vlm"},
	{0xFEB9F06C5313CE83ULL, "dae000_gamlen6102708.cl"},
	{0xFEBA1175A43AFF3DULL, "ht_0_hightown_30_0.anb"},
	{0xFEBA13CF7ED74937ULL, "vxlt_gallowstemplar_f_3rg_0_0.vlm"},
	{0xFEBA5C9B43E75CBFULL, "hlac_ht_keep_n_4519_8z.dds"},
	{0xFEBA9C038A11E1F9ULL, "vxca_caverns1_mines_i08_0_0.vlm"},
	{0xFEBAA21686EAD217ULL, "716104107.ogg"},
	{0xFEBAC1569B30D8A5ULL, "mrl000_marethari.cnv"},
	{0xFEBB87F493FAF6A6ULL, "270011375.ogg"},
	{0xFEBBCD374E1488B7ULL, "556124433.ogg"},
	{0xFEBBF41815E92DA8ULL, "llac_lt_0_lowtown_docks_a3_131071_2.dds"},
	{0xFEBBF93915B1A768ULL, "hlac_ht_keep_n_4519_10x.dds"},
	{0xFEBC2C1BE24DE228ULL, "vxlt_gallowstemplar_f_474_0_0.vlm"},
	{0xFEBC6E8AF2FEF803ULL, "205833712.ogg"},
	{0xFEBC92ED5B896519ULL, "zz_pro_cutscenes6005833.cl"},
	{0xFEBC955128FB003AULL, "6095967_m.fxe"},
	{0xFEBC9CD4908DE334ULL, "211567750.ogg"},
	{0xFEBCBD317946838DULL, "vxht_0_hightown_c5d_0_0.vlm"},
	{0xFEBD2758D67CBE21ULL, "follower_banter6172949.cl"},
	{0xFEBD4662E5D7BB25ULL, "352013744.ogg"},
	{0xFEBDB0AC66515B10ULL, "vxht_bartrand_n_69t_0_0.vlm"},
	{0xFEBE5D95B8811447ULL, "6177188_f.fxe"},
	{0xFEBE7AA318A5417DULL, "vxlt_0_lowtown_ens_0_0.vlm"},
	{0xFEBFAA516F78EEB3ULL, "6088872_m.fxe"},
	{0xFEC00546514686D5ULL, "gen00fl_dog6109237.cl"},
	{0xFEC04417F8A41B1BULL, "vxlt_gallowscourtyard_n_7ap_0_0.vlm"},
	{0xFEC0DED52A7B16FEULL, "60887159.ogg"},
	{0xFEC1083491E00BCCULL, "vxlt_0_lowtown_a2_gnp_0_0.vlm"},
	{0xFEC11236D1C33D21ULL, "970681729.ogg"},
	{0xFEC18DE188282B14ULL, "vxdr_primevaltomb_w3_0_0.vlm"},
	{0xFEC1B0A4BA46D614ULL, "204401070.ogg"},
	{0xFEC2002254E3ED13ULL, "6166905_m.fxe"},
	{0xFEC26C50567D325FULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.impa.evt"},
	{0xFEC2FD22F1BABFECULL, "6202264_m.fxe"},
	{0xFEC306E313E28AAAULL, "vxht_brothel_n_4wf_0_0.vlm"},
	{0xFEC31754E2556FF9ULL, "867358857.ogg"},
	{0xFEC36641A297831DULL, "zz_vault_debug6177795.cl"},
	{0xFEC385B750665C00ULL, "vxlt_undercitydungeon_1j3_0_1.vlm"},
	{0xFEC3896033E5AB05ULL, "vxca_caverns2prc_9sz_0_0.vlm"},
	{0xFEC39E5B8F0A7812ULL, "follower_banter6156182.cl"},
	{0xFEC3B10BF9634BA3ULL, "vxlt_darktown_6kr_0_0.vlm"},
	{0xFEC3C7AF3793F227ULL, "6054024_m.fxe"},
	{0xFEC436461FAD0AF0ULL, "781277165.ogg"},
	{0xFEC46D80E5B53867ULL, "vxht_0_hightown_cbc_0_0.vlm"},
	{0xFEC4C8ED99BB3BE2ULL, "hlac_lt_0_lowtown_docks_n_65533_1z.dds"},
	{0xFEC4CDC3CF51181CULL, "223289356.ogg"},
	{0xFEC50F98DF12AF69ULL, "ss_city_guard_f_1.wbd"},
	{0xFEC52784CB5D6FACULL, "vxlt_0_lowtown_a3_n_c_hib_0_0.vlm"},
	{0xFEC5E64E69114123ULL, "6063996_m.fxe"},
	{0xFEC600625662744EULL, "vxlt_0_lowtown_docks_n_bgf_0_0.vlm"},
	{0xFEC61E1B2C058652ULL, "zz_vault_debug6178953.cl"},
	{0xFEC646702C039E02ULL, "mag121_viveka6082598.cl"},
	{0xFEC683095B2D05B1ULL, "lgt000cr_mage_turn_in_6.utc"},
	{0xFEC6BEF687C711DDULL, "1046792732.ogg"},
	{0xFEC6CE080BED836EULL, "vxht_mansion2b_2ot_0_0.vlm"},
	{0xFEC71E096C1F3C97ULL, "vxlt_0_lowtown_a3_n_c_i0g_0_0.vlm"},
	{0xFEC7392971CAE37EULL, "6033668_m.fxe"},
	{0xFEC74C9442C99900ULL, "lt_0_lowtown_docks_n_33_0.rml"},
	{0xFEC7619EB79370F9ULL, "6154713_m.fxe"},
	{0xFEC7DE1945D7E789ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.gad"},
	{0xFEC82ABEFFF455C9ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_3z.dds"},
	{0xFEC83B2ED081A4C9ULL, "vxht_mansion_n_4w6_0_0.vlm"},
	{0xFEC8889B78A14724ULL, "6116844_m.fxe"},
	{0xFEC8B84A3F8D6E8EULL, "182665442.ogg"},
	{0xFEC8DDF4189AC4D9ULL, "mag170_thrask6016924.cl"},
	{0xFEC94D5056831C48ULL, "art\\vfx\\combat\\vfx_w_gn1_battle_loop_t3_c.impa.gad"},
	{0xFEC96F94B8B6F1E2ULL, "262911276.ogg"},
	{0xFEC97354A592D82AULL, "6105907_m.fxe"},
	{0xFEC98794413C7361ULL, "415536801.ogg"},
	{0xFEC9ADFA9F844A6FULL, "6172276_m.fxe"},
	{0xFECA0E0B478CE467ULL, "130348587.ogg"},
	{0xFECA60F2790FEAC5ULL, "6208700_m.fxe"},
	{0xFECA8F703A67012CULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend01_da.dds"},
	{0xFECA9D6512113F48ULL, "hlac_lt_0_lowtown_a2_n_c_65534_0y.dds"},
	{0xFECB1890E77F037EULL, "305375935.ogg"},
	{0xFECB27E2F2DF9513ULL, "hlac_ht_0_hightown_a3_65534_0x.dds"},
	{0xFECB4CD81FF6568AULL, "dae200_cullen6140306.cl"},
	{0xFECB524F77405BF9ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderjump.ani"},
	{0xFECBDD4D1B42A85EULL, "706423123.ogg"},
	{0xFECBF2DA2F74F3A3ULL, "6177184_f.fxe"},
	{0xFECC8DC52C5CC5CFULL, "vxlt_0_lowtown_n_ep0_0_0.vlm"},
	{0xFECCEC800A9B59FDULL, "vxlt_0_lowtown_eir_0_0.vlm"},
	{0xFECD558B09E4089DULL, "one110cr_gustav.ncs"},
	{0xFECD5E4211352262ULL, "6102949_f.fxe"},
	{0xFECDA7EA70BBD3BDULL, "art\\vfx\\cinematic\\vfx_blood_b_splatters_p.mmh"},
	{0xFECE1C2E616F0733ULL, "6083972_m.fxe"},
	{0xFECE8F3B19F14F32ULL, "6167356_m.fxe"},
	{0xFECEAF1945DDB6D7ULL, "art\\vfx\\combat\\vfx_gib_hlk_p.justlegs.ani"},
	{0xFECF337E680684C2ULL, "6150673_m.fxe"},
	{0xFECF9FB06C1DCEF0ULL, "hlac_lt_foundry_n_320_17y.dds"},
	{0xFECFBA22EB43D6D5ULL, "372444012.ogg"},
	{0xFECFC0790F404CFDULL, "vxca_cavern1_haunted_eh0_0_0.vlm"},
	{0xFED03519FAD2C88FULL, "dae200_bran6144561.cl"},
	{0xFED11B8F347CF240ULL, "227922268.ogg"},
	{0xFED123FEB5FE998EULL, "plt_lgt000pt_baz_cache.nss"},
	{0xFED17D0AE719181EULL, "648989097.ogg"},
	{0xFED1E13B59D0D608ULL, "llac_lt_0_lowtown_docks_n_131071_0.dds"},
	{0xFED2044DCEB23923ULL, "6150272_m.fxe"},
	{0xFED22D4F77463B3BULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderjump.gad"},
	{0xFED23AC99D585C54ULL, "416094440.ogg"},
	{0xFED266F62C86BAF9ULL, "hlac_lt_warehouse_1_5x.dds"},
	{0xFED2F8CB749D45BBULL, "141409857.ogg"},
	{0xFED355DC3C8B4815ULL, "follower_banter6170316.cl"},
	{0xFED359910F6989EAULL, "6103721_m.fxe"},
	{0xFED3BD4715F2968AULL, "6060992_m.fxe"},
	{0xFED47F9DECA88745ULL, "wm_ico_house_roll.dds"},
	{0xFED50ADA224D7284ULL, "ser221_gascard6165436.cl"},
	{0xFED51A732381AF2FULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_rdy.gad"},
	{0xFED51C8F4C7080A1ULL, "seb221_flora_harimann6177661.cl"},
	{0xFED563C39F67B689ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mcn.mao"},
	{0xFED56FC39F67CB22ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mcn.mmh"},
	{0xFED571C39F67CED8ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_elftarp12mcn.msh"},
	{0xFED61EEB4FCE6022ULL, "hlac_w_encounter3_-65537_0z.dds"},
	{0xFED67AC6D7D24FB4ULL, "vxca_caverns2_8oc_0_0.vlm"},
	{0xFED6C12D6B56DD36ULL, "6071908_f.fxe"},
	{0xFED70078973E3CC0ULL, "zz_lgt_2_debug6206607.cl"},
	{0xFED727AA90CFD19AULL, "6016157_m.fxe"},
	{0xFED7A8ECFF8ED982ULL, "6155670_m.fxe"},
	{0xFED88519C0FDED1AULL, "6120559_m.fxe"},
	{0xFED8FC4B390FF41AULL, "6105613_m.fxe"},
	{0xFED91C3B8AA0F37EULL, "6148654_m.fxe"},
	{0xFED947C4E04A4913ULL, "1294706.ogg"},
	{0xFED94A4F774C8B20ULL, "art\\characters\\creatures\\gallowsgateguard\\animation\\spider\\spiderjump.evt"},
	{0xFED958D16768AEABULL, "58686604.ogg"},
	{0xFED98E1C1F449C0CULL, "846492117.ogg"},
	{0xFED9909412648F6DULL, "hm_qsw_warrens4.mor"},
	{0xFED9A35F20B154FEULL, "ss_crazy_m_1.wbd"},
	{0xFED9B47B2FE6C711ULL, "llac_lt_gallowsprison_n_-1_6.dds"},
	{0xFEDA1690022B4D5CULL, "1005002019.ogg"},
	{0xFEDA23AC2E25504FULL, "6172610_m.fxe"},
	{0xFEDA353731E45F51ULL, "6128387_m.fxe"},
	{0xFEDA407BDC5D7DB5ULL, "llac_lt_0_lowtown_a3_n_c_0_8.dds"},
	{0xFEDB963027D2C80AULL, "159686315.ogg"},
	{0xFEDBBC82818D590AULL, "gen00fl_dog6115035.cl"},
	{0xFEDC2F732387F1DCULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_rdy.evt"},
	{0xFEDCAB9672F761ABULL, "hlac_lt_darktown_-1_33y.dds"},
	{0xFEDD0989EF60B36AULL, "6125303_m.fxe"},
	{0xFEDD57014396986DULL, "454998361.ogg"},
	{0xFEDD8DA0BE276422ULL, "645654301.ogg"},
	{0xFEDDBFF83A18CF4BULL, "vxw_sundermount_31z_0_0.vlm"},
	{0xFEDDD98DAB8C61DEULL, "6142032_m.fxe"},
	{0xFEDE116625F4BECBULL, "art\\characters\\playercharacter\\humanmale\\mt_ant_hm05.phy"},
	{0xFEDE5B8EF38C39FAULL, "vxlt_darktown_5cp_0_0.vlm"},
	{0xFEDE6906DFC122E9ULL, "vxht_bartrand_n_65u_0_0.vlm"},
	{0xFEDE80BC33717D77ULL, "yellow25.dds"},
	{0xFEDE82AB34B628D4ULL, "vxlt_0_lowtown_a2_gon_0_1.vlm"},
	{0xFEDEAC00BF2E6C9BULL, "vxca_caverns1_mines_j4e_0_0.vlm"},
	{0xFEDEE73667E37A91ULL, "336368473.ogg"},
	{0xFEDF54778F4E877FULL, "269517117.ogg"},
	{0xFEDFC673B466A1E6ULL, "324208298.ogg"},
	{0xFEDFD4EFC58398D0ULL, "art\\characters\\playercharacter\\humanfemale\\hf_har_mth.phy"},
	{0xFEDFD9CFDA0E4139ULL, "6116395_m.fxe"},
	{0xFEE07BBC851D6F44ULL, "lgt120_orz_bln_traitor_lead6140234.cl"},
	{0xFEE12AEBEBA1686EULL, "366291786.ogg"},
	{0xFEE2048BA907F0B2ULL, "6124739_m.fxe"},
	{0xFEE21EBDFC594B76ULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_005a_0t.dds"},
	{0xFEE2446DED74EAD7ULL, "vxlt_0_lowtown_a3_n_c_df0_0_0.vlm"},
	{0xFEE2FD48521425C6ULL, "one280_iwan6096872.cl"},
	{0xFEE3002C9331D55BULL, "636382063.ogg"},
	{0xFEE342F105861149ULL, "638264125.ogg"},
	{0xFEE3B1C1EFD9B46EULL, "vxht_mansion2b_2ev_0_0.vlm"},
	{0xFEE4204C254B8870ULL, "6065935_m.fxe"},
	{0xFEE431B728D349F9ULL, "404039298.ogg"},
	{0xFEE5160CA784A871ULL, "art\\characters\\playercharacter\\humanfemale\\textures\\hf_cth_noba_0s.dds"},
	{0xFEE5204F70C5EE71ULL, "gen00fl_anders6125155.cl"},
	{0xFEE5C1612E28E0CCULL, "hlac_lt_0_lowtown_n_65536_3z.dds"},
	{0xFEE5D069FAFB9146ULL, "m_qcr221ar_keep.wbd"},
	{0xFEE5EF3942BBA7CEULL, "457865568.ogg"},
	{0xFEE5F81C0C72D5C8ULL, "6187925_m.fxe"},
	{0xFEE64D5F4A02543DULL, "814793564.ogg"},
	{0xFEE68EEDE4A343EDULL, "630965754.ogg"},
	{0xFEE76F68935F4711ULL, "311733973.ogg"},
	{0xFEE93385E8B62E22ULL, "zz_start_qcr6081933.cl"},
	{0xFEE94BBAF3FED0B5ULL, "ttx_steps_a_1.msh"},
	{0xFEE94DBAF3FED42BULL, "ttx_steps_a_1.mmh"},
	{0xFEE9B4A4E9221B6DULL, "art\\characters\\playercharacter\\elffemale\\textures\\ef_cth_comf_0s.dds"},
	{0xFEE9C8E715406E16ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1a.msh"},
	{0xFEE9D2E715407F14ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1a.mmh"},
	{0xFEE9D6E715408627ULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1a.mao"},
	{0xFEEA254427C01DA6ULL, "1015002429.ogg"},
	{0xFEEA377323943CF5ULL, "art\\characters\\creatures\\arh_abm_rev\\animation\\a_rdy.ani"},
	{0xFEEAC4B52BBCF359ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\1pp_large.phy"},
	{0xFEEAE198DF33304BULL, "ss_city_guard_f_1.bnk"},
	{0xFEEB6DE701E077BBULL, "6078350_m.fxe"},
	{0xFEEBCF1A675980C4ULL, "vxw_woundedcoast_13q_0_0.vlm"},
	{0xFEEBD03FFE2C140AULL, "24133810.ogg"},
	{0xFEEC4089775A8934ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_shieldc_d.dds"},
	{0xFEEC573B6151E401ULL, "6118942_m.fxe"},
	{0xFEEC7F735467FE5DULL, "w_woundedcoast_17_0_lc.mmh"},
	{0xFEEC960F22722D0CULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\blendtree\\common_gestures.anb"},
	{0xFEECC166E98773EBULL, "vxdr_primevaltomb_6pw_0_2.vlm"},
	{0xFEECDCC678A33612ULL, "mrl251ip_treasureblocker.utp"},
	{0xFEECED097E85F72BULL, "art\\vfx\\combat\\vfx_a_stealth_aoe_p.impa.evt"},
	{0xFEED1CB86D4D45FAULL, "fen110cr_hunter.utc"},
	{0xFEEDAE897BE9680FULL, "gen_im_arm_glv_med_act2_03.uti"},
	{0xFEEDB5A3134F51F9ULL, "vxht_0_hightown_n_9zd_0_0.vlm"},
	{0xFEEE22F8C5021169ULL, "6151137_f.fxe"},
	{0xFEEE5CB965869941ULL, "136721274.ogg"},
	{0xFEEEB84BCBFF1701ULL, "hlac_lt_undercitydungeon_-1_21y.dds"},
	{0xFEEF8C6C26438D0DULL, "6065230-strongandhandsome.cut"},
	{0xFEEFBE8F244F9D7DULL, "6208606_f.fxe"},
	{0xFEEFD6853C494154ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_dead.gad"},
	{0xFEEFF822AAD1D3ACULL, "gen00fl_isabela6075237.cl"},
	{0xFEF01BA4A4FE74EEULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofplain.msh"},
	{0xFEF029A4A4FE8C3FULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofplain.mao"},
	{0xFEF035A4A4FEA0DCULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofplain.mmh"},
	{0xFEF0849E3206CDA5ULL, "lit_aoe_01_l3.mao"},
	{0xFEF0DA6A61A4015AULL, "910446644.ogg"},
	{0xFEF12426E7A418F0ULL, "vxlt_foundry_n_11h_0_0.vlm"},
	{0xFEF13118D17D2F9CULL, "vxlt_0_lowtown_a2_n_c_gnu_0_0.vlm"},
	{0xFEF1444876D3579AULL, "6077790_m.fxe"},
	{0xFEF191905421FA05ULL, "llac_ht_0_hightown_a3_0_3.dds"},
	{0xFEF1BB064366A4D9ULL, "6109733_m.fxe"},
	{0xFEF1D3AF035D038BULL, "dr_0_deeproads_mmap.dds"},
	{0xFEF2350EE5FFF8C7ULL, "hlac_ht_mansion2b_818_34x.dds"},
	{0xFEF251C8EC7FD268ULL, "w_sundermount_148_0.msh"},
	{0xFEF263C8EC7FF0F9ULL, "w_sundermount_148_0.mao"},
	{0xFEF26FC8EC800492ULL, "w_sundermount_148_0.mmh"},
	{0xFEF299FF877403E0ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_dummy_dp.dds"},
	{0xFEF2C4BFD7342511ULL, "lt_0_lowtown_docks_a3_32_0.rml"},
	{0xFEF2E4C10EA4D59FULL, "w_encounter3_196_0.msh"},
	{0xFEF2EAC10EA4E041ULL, "w_encounter3_196_0.mmh"},
	{0xFEF2F6C10EA4F4E2ULL, "w_encounter3_196_0.mao"},
	{0xFEF3116D6D883EC6ULL, "art\\characters\\creatures\\dragon\\animation\\ctn180l.evt"},
	{0xFEF3B77740BADCB6ULL, "art\\characters\\playercharacter\\qunarimale\\mt_bas_qm.msh"},
	{0xFEF3C17740BAEDB4ULL, "art\\characters\\playercharacter\\qunarimale\\mt_bas_qm.mmh"},
	{0xFEF3C20CDF831188ULL, "gen_im_trs_gen_junk3_02.uti"},
	{0xFEF3CD49E45306B1ULL, "art\\vfx\\creatures\\fxc_shockwave_p.anb"},
	{0xFEF404B066D4FAE9ULL, "vxht_mansion2b_2k3_0_0.vlm"},
	{0xFEF406097E8C40ECULL, "art\\vfx\\combat\\vfx_a_stealth_aoe_p.impa.gad"},
	{0xFEF40E648AA49C28ULL, "vxlt_gallowscourtyard_6ww_0_0.vlm"},
	{0xFEF4123620BEBF0AULL, "6116046_m.fxe"},
	{0xFEF41275A46C7935ULL, "ht_0_hightown_30_0.rml"},
	{0xFEF43C7CFFD3F018ULL, "397092167.ogg"},
	{0xFEF4C1A0BEC0024CULL, "593655687.ogg"},
	{0xFEF4DB5A04159697ULL, "ser121_jethann6104438.cl"},
	{0xFEF4EEF0847E0974ULL, "vxlt_0_lowtown_e88_0_0.vlm"},
	{0xFEF516D57D980847ULL, "vxca_cavern1_haunted_jkx_0_0.vlm"},
	{0xFEF57D39DCBDC7D3ULL, "ss_aveline.bnk"},
	{0xFEF5B80F367B5398ULL, "6156567_m.fxe"},
	{0xFEF5F7ECB26A145DULL, "ts_chantry_acc_props.gda"},
	{0xFEF63861860BB658ULL, "hm_amb_darktown_sickly.mor"},
	{0xFEF6454086744732ULL, "vxht_0_hightown_a3_9xg_0_0.vlm"},
	{0xFEF65F1FFD347278ULL, "cd_mag360_thrask.bnk"},
	{0xFEF6E5853C4F7932ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\two_handed_persuade_dead.ani"},
	{0xFEF6F5E308A7478CULL, "lt_0_lowtown_docks_112_0.msh"},
	{0xFEF6FBE308A7527EULL, "lt_0_lowtown_docks_112_0.mmh"},
	{0xFEF707E308A76615ULL, "lt_0_lowtown_docks_112_0.mao"},
	{0xFEF71E7C3F2F810EULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxb.phy"},
	{0xFEF74F4B2E7C78BDULL, "mcr370_templar16086969.cl"},
	{0xFEF780D7463B1064ULL, "gen00fl_anders6097239.cl"},
	{0xFEF9F4A7015025BEULL, "vxht_0_hightown_a2_c57_0_0.vlm"},
	{0xFEFA286D6D8E8411ULL, "art\\characters\\creatures\\dragon\\animation\\ctn180l.gad"},
	{0xFEFA3265DB27AC15ULL, "vxht_0_hightown_a2_7ix_0_0.vlm"},
	{0xFEFA5ECA85EE998EULL, "dae010_commoner_m1.utc"},
	{0xFEFA7AC26DB158FFULL, "hlac_w_sundermount_131076_0x.dds"},
	{0xFEFA87274E57D018ULL, "zz_cbev.ncs"},
	{0xFEFA985AFBD9E366ULL, "254845710.ogg"},
	{0xFEFB05097E925DEAULL, "art\\vfx\\combat\\vfx_a_stealth_aoe_p.impa.ani"},
	{0xFEFB07C2BFFC79B8ULL, "t3_hlk_b03.tnt"},
	{0xFEFB394110B0BFEFULL, "vxlt_warehouse_n_1df_0_0.vlm"},
	{0xFEFB3F48D7786AFAULL, "vxlt_darktown_5ir_0_0.vlm"},
	{0xFEFB760883A13140ULL, "6108696_m.fxe"},
	{0xFEFB98736BDD334DULL, "181947731.ogg"},
	{0xFEFC020F6675DD08ULL, "art\\levels\\races\\dwarf\\common\\int\\_textures\\til_cavernlava_da.dds"},
	{0xFEFC11EE3CB40B77ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkb.msh"},
	{0xFEFC23EE3CB429DAULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkb.mao"},
	{0xFEFC27EE3CB430E9ULL, "art\\levels\\races\\ferelden\\props\\mining\\prp_junkb.mmh"},
	{0xFEFC3FCCF009A7FEULL, "gen00fl_merrill6121457.cl"},
	{0xFEFC4612F4A4C83AULL, "gen00fl_aveline6132926.cl"},
	{0xFEFC876C7D1D8FA4ULL, "vxlt_0_lowtown_a2_n_c_ety_0_0.vlm"},
	{0xFEFCABB0990465E9ULL, "hlac_ca_cavern1_haunted_24425_5y.dds"},
	{0xFEFCDAACABCAAB56ULL, "hlac_ca_caverns2_5802_52z.dds"},
	{0xFEFD021637A14DAEULL, "zz_start_dre6080524.cl"},
	{0xFEFD1F87E5A05E3CULL, "ave200_donnic6111069.cl"},
	{0xFEFD422CC9BB37ECULL, "art\\levels\\races\\proxy\\props\\foodstuffs\\prp_zzbarrelfish_0.pwk"},
	{0xFEFD503F37401C13ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v1.evt"},
	{0xFEFDA1794D9B4F35ULL, "hlac_w_blightlands_131074_0z.dds"},
	{0xFEFDD5492EA5989CULL, "hlac_w_sundermount_4_0y.dds"},
	{0xFEFDDD11F237B24CULL, "119568208.ogg"},
	{0xFEFEEC9746ED386DULL, "vxlt_undercityserialkiller_180_0_0.vlm"},
	{0xFEFEF15261BD0B77ULL, "mcr310_cut_chantry_blows_up6087539.cl"},
	{0xFEFF33FDAF9CCAB3ULL, "vxht_0_hightown_n_2xc_0_0.vlm"},
	{0xFEFF3A2DE6231C71ULL, "hlac_ht_mansion2_818_14z.dds"},
	{0xFEFF73C6D4253B20ULL, "438144239.ogg"},
	{0xFEFFA3ADD822A4D5ULL, "678335032.ogg"},
	{0xFEFFC7D61E3AD64EULL, "vxlt_0_lowtown_docks_n_bsz_0_0.vlm"},
	{0xFF0015B12A981D43ULL, "vxlt_backalley_g_1xj_0_0.vlm"},
	{0xFF00399594108ECEULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pound_fist.evt"},
	{0xFF004F83CE5A71D6ULL, "zz_vault_debug6203253.cl"},
	{0xFF0093EB13BA0CD3ULL, "art\\levels\\ztools\\helpers\\character\\prp_dwarfstand_2.mao"},
	{0xFF00EE49E45DC92AULL, "art\\vfx\\creatures\\fxc_shockwave_p.mmh"},
	{0xFF01129D58A18822ULL, "hlac_lt_0_lowtown_a2_n_65533_2z.dds"},
	{0xFF01296D6D94A4BFULL, "art\\characters\\creatures\\dragon\\animation\\ctn180l.ani"},
	{0xFF01C229E2108361ULL, "vrc321tr_spooky_scenes_4.ncs"},
	{0xFF01EE70A7BC73FFULL, "480049509.ogg"},
	{0xFF01FAF180657777ULL, "6039937_m.fxe"},
	{0xFF02768D58CB1369ULL, "vxht_mansion2b_2d2_0_0.vlm"},
	{0xFF028CC491B8C0BFULL, "57561933.ogg"},
	{0xFF02A3689F96704EULL, "dae000_sandal6090027.cl"},
	{0xFF02A5A70E180A34ULL, "hlac_w_bonepitdestroyed_a3_327686_0z.dds"},
	{0xFF0333B04E5BE7DBULL, "vxw_sundermount_2cu_0_0.vlm"},
	{0xFF03493F37447C54ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v1.gad"},
	{0xFF0360CD767421F4ULL, "vxlt_0_lowtown_e39_0_0.vlm"},
	{0xFF03B8C49A6C4AD2ULL, "6142788_m.fxe"},
	{0xFF03CDA2258F86CCULL, "follower_banter6026243.cl"},
	{0xFF043C0E3DBF713CULL, "vxca_caverns1_fmp_0_0.vlm"},
	{0xFF04EB13310B26D5ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_junke_sa.dds"},
	{0xFF050FB1D27DD91BULL, "6162291_m.fxe"},
	{0xFF05245E7735C449ULL, "vxlt_0_lowtown_docks_b9m_0_1.vlm"},
	{0xFF052DE5CD25D4A5ULL, "621455749.ogg"},
	{0xFF05AF09B156F9BFULL, "hlac_lt_gallowscourtyard_n_-65535_2y.dds"},
	{0xFF0663DEF36359C1ULL, "6039401_m.fxe"},
	{0xFF067C171EC939DAULL, "131222391.ogg"},
	{0xFF06EE46726BC09BULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1.ani"},
	{0xFF072C178E444BB3ULL, "hla_ht_0_hightown_-131072_0x.dds"},
	{0xFF07309594169E79ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pound_fist.gad"},
	{0xFF07969F439C07DAULL, "147112866.ogg"},
	{0xFF07ED3D78D1D3EBULL, "6175084_m.fxe"},
	{0xFF0826A467370C6AULL, "dae011_gossips6143001.cl"},
	{0xFF0941ECA614E2E0ULL, "6050198_f.fxe"},
	{0xFF09B125C96E837AULL, "vxlt_darktown_5rs_0_0.vlm"},
	{0xFF09E2506051882BULL, "art\\characters\\playercharacter\\dwarfmale\\textures\\dm_arm_hvyo_0s.dds"},
	{0xFF0A583F374AB432ULL, "art\\characters\\playercharacter\\humanmale\\animation\\exploration\\idle\\am_idle_tw_v1.ani"},
	{0xFF0BBAE624A3DD22ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_meda_0n.dds"},
	{0xFF0BC917FA11A808ULL, "665857052.ogg"},
	{0xFF0C5F7793664600ULL, "vxca_caverns1_mines_gkf_0_0.vlm"},
	{0xFF0C6AD5A1742BF3ULL, "distortionwpntrl.vsh.fxo"},
	{0xFF0C7CD38357F8E6ULL, "6178886_m.fxe"},
	{0xFF0C9055BF90BD86ULL, "w_bonepitdestroyed_a3_17_0_lc.mmh"},
	{0xFF0CA55E936DBDD3ULL, "hlac_ca_caverns1_mines_24424_3x.dds"},
	{0xFF0CAC643A7268DAULL, "vxca_caverns2_8ge_0_0.vlm"},
	{0xFF0CEC77CDCFDD38ULL, "6033590_m.fxe"},
	{0xFF0D10630ECA723DULL, "sky_lt_0_lowtown_a2_n.mao"},
	{0xFF0D4195941B26C7ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\2arm_pound_fist.ani"},
	{0xFF0D6BDC91F67DB6ULL, "water_water_p3.psh.fxo.dx11"},
	{0xFF0DDAA2924938E3ULL, "6051051_f.fxe"},
	{0xFF0DF8BFF9F9099BULL, "rdr111dg_martin6054897.cl"},
	{0xFF0E1C807B866AAEULL, "vxw_bonepit_hz_0_0.vlm"},
	{0xFF0E68F9442B5598ULL, "6087557_m.fxe"},
	{0xFF0E985C65363E4DULL, "and100_anders6111688.cl"},
	{0xFF0E9C6BE7D7DE2FULL, "vxlt_hangedman_n_5j_0_0.vlm"},
	{0xFF0EB639A5CFCA18ULL, "gen00fl_aveline6139316.cl"},
	{0xFF0ED557DE2F665EULL, "864653569.ogg"},
	{0xFF0EFCF9B013E037ULL, "zz_lgt_1_debug6136991.cl"},
	{0xFF0F074233165230ULL, "gen00fl_fenris6126928.cl"},
	{0xFF103434ED697CB5ULL, "hlac_ht_keep_n_4519_3y.dds"},
	{0xFF10C2D151BBA03AULL, "6174518_m.fxe"},
	{0xFF10E86A4AC12BE9ULL, "vxht_mansion2b_2a1_0_0.vlm"},
	{0xFF10F21E6E90CD28ULL, "vxlt_gallowscourtyard_6yu_0_0.vlm"},
	{0xFF1106FA74645015ULL, "596543708.ogg"},
	{0xFF11A58D4052005BULL, "vxw_sundermount_2dt_0_0.vlm"},
	{0xFF11E80EDECECFD6ULL, "597888379.ogg"},
	{0xFF13F64672765892ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1.evt"},
	{0xFF143E81C8B95AFBULL, "vxlt_0_lowtown_a3_n_c_gc8_0_0.vlm"},
	{0xFF14468B0A1FF277ULL, "one110cr_gustav.utc"},
	{0xFF14C92C060CED65ULL, "vxlt_gallowstemplar_53g_0_0.vlm"},
	{0xFF14EA56912FE497ULL, "vxlt_0_lowtown_docks_a3_d3a_0_0.vlm"},
	{0xFF1514A4A51D84A7ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\2ca_roofplain.phy"},
	{0xFF153650E4414FF5ULL, "llac_ht_playershouse_n_855_22.dds"},
	{0xFF15373D0AFC4A59ULL, "286400940.ogg"},
	{0xFF15ADDDEA9386D8ULL, "347470591.ogg"},
	{0xFF15FD8E6E6C3318ULL, "vxca_cavern1_haunted_ckr_0_0.vlm"},
	{0xFF16695E915613DBULL, "zz_dae_debug6036575.cl"},
	{0xFF16856EF9C1E286ULL, "522043468.ogg"},
	{0xFF16CEC1E412F45BULL, "128964588.ogg"},
	{0xFF1718A60B82BAE7ULL, "vxlt_0_lowtown_n_ex0_0_0.vlm"},
	{0xFF172FC491D5606BULL, "6166616_f.fxe"},
	{0xFF17946D622EC2E2ULL, "vxdr_0_deeproads_bp6_0_0.vlm"},
	{0xFF17B7AB56500D0AULL, "qun221_viscount6084683.cl"},
	{0xFF18F082DB4873D0ULL, "vxlt_0_lowtown_a2_ehx_0_0.vlm"},
	{0xFF19257105A25EF4ULL, "565177120.ogg"},
	{0xFF1926BDEA2EE721ULL, "6178395_m.fxe"},
	{0xFF1967351DABE316ULL, "hlac_lt_0_lowtown_docks_a3_n_-196607_0x.dds"},
	{0xFF19AF464B9F772BULL, "6075704_m.fxe"},
	{0xFF19C7533DD9BCF0ULL, "vxht_0_hightown_a2_7g3_0_0.vlm"},
	{0xFF1A816354BF1723ULL, "hlac_lt_0_lowtown_a2_131073_2y.dds"},
	{0xFF1B098C0432F039ULL, "207725844.ogg"},
	{0xFF1B1546727CAC25ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\pose\\pose_a_weightshift_small_trim1.gad"},
	{0xFF1B1E412C68815AULL, "vxca_caverns2_8bf_0_0.vlm"},
	{0xFF1BB84EB49CD2E8ULL, "6048394_m.fxe"},
	{0xFF1BD9983FB879F5ULL, "vxw_encounter3_dv_0_0.vlm"},
	{0xFF1C152D1BD7E7D1ULL, "vxw_encounter3_j_0_0.vlm"},
	{0xFF1C761ABEE9CB7CULL, "vxw_bonepitdestroyed_a3_hx_0_0.vlm"},
	{0xFF1C961460E5051BULL, "6075058_m.fxe"},
	{0xFF1CF2B944E82E80ULL, "6130137_f.fxe"},
	{0xFF1D0DD3143FF142ULL, "313087273.ogg"},
	{0xFF1D3105FC9A2802ULL, "vxlt_0_lowtown_docks_n_b6u_0_0.vlm"},
	{0xFF1D795AFA4874DBULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.dura.evt"},
	{0xFF1E2C222D7987E6ULL, "hlac_ht_chantry_n_2534_5y.dds"},
	{0xFF1EB28CFA719210ULL, "6161024_m.fxe"},
	{0xFF1EF63C02496851ULL, "framebuffereffects_illuminationmask_p0.vsh.fxo.dx11.assembly"},
	{0xFF1F0969199F8F9DULL, "fen110_fenris6118066.cl"},
	{0xFF1F88DB8293C297ULL, "vxca_caverns1_dr0_0_0.vlm"},
	{0xFF1F8D23BCD0E182ULL, "6075248_m.fxe"},
	{0xFF20176A324818DBULL, "vxw_sundermount_2iw_0_0.vlm"},
	{0xFF20C89EA6590533ULL, "6162754_m.fxe"},
	{0xFF216C7ACDAC52BAULL, "follower_barks6173409.cl"},
	{0xFF233014DBFDE688ULL, "118676319.ogg"},
	{0xFF23357BFEFC3F15ULL, "one200_gold_followers6101130.cl"},
	{0xFF23DCE08A190C69ULL, "dae300_petrice6098086.cl"},
	{0xFF23DF08DC497F40ULL, "hlac_lt_0_lowtown_n_-65536_1y.dds"},
	{0xFF23E610B9E85823ULL, "982314941.ogg"},
	{0xFF2441224DB10F8AULL, "dae000cr_grand_cleric6181924.cl"},
	{0xFF24725AFA4E871CULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.dura.gad"},
	{0xFF24739ACC42044DULL, "dre000st_varric_enttomb.stg"},
	{0xFF24B38D1E87856EULL, "hlac_lt_gallowstemplar_2391_1x.dds"},
	{0xFF2508533879D3DDULL, "vxw_encounter1_cy_0_0.vlm"},
	{0xFF256418E4C9D88DULL, "6114088_f.fxe"},
	{0xFF258A82FD78D367ULL, "vxlt_0_lowtown_n_eu3_0_0.vlm"},
	{0xFF25CF347C01B78EULL, "art\\vfx\\combat\\vfx_ha_ar_armoff_act_c.impa.ani"},
	{0xFF260FCEC8DD1A2AULL, "hlac_w_blightlands_-65539_0z.dds"},
	{0xFF2636ECDBECFA45ULL, "vxlt_darktown_5iy_0_0.vlm"},
	{0xFF26A1F15DA893E5ULL, "6054832_f.fxe"},
	{0xFF279609A801FE9EULL, "1064557330.ogg"},
	{0xFF27CD4A3379A2FCULL, "vxca_caverns2prc_936_0_1.vlm"},
	{0xFF2802CF970D6D1BULL, "vrc300pt_n_initial.ncs"},
	{0xFF28AC7CA5E82EF5ULL, "vxlt_0_lowtown_a2_e8i_0_0.vlm"},
	{0xFF294022C4A35A12ULL, "mag110cr_guardsman2.utc"},
	{0xFF29B09DCEA47BD5ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.anb"},
	{0xFF2A2435023EE7BAULL, "6179885_m.fxe"},
	{0xFF2A4B7531AE9275ULL, "vxw_encounter3_au_0_0.vlm"},
	{0xFF2A9B0817605942ULL, "t3_arm_kwl.tnt"},
	{0xFF2AAC1D0E0D34C1ULL, "208100071.ogg"},
	{0xFF2B715AFA54A41AULL, "art\\vfx\\combat\\vfx_w_def_ele_def_loop_c.dura.ani"},
	{0xFF2B805C1AB0E337ULL, "vxlt_gallowstemplar_n_3nk_0_0.vlm"},
	{0xFF2BDB6861EF8D5BULL, "llac_lt_0_lowtown_docks_-131076_0.dds"},
	{0xFF2BF3AF6D262A7AULL, "qun210cr_guard6072133.cl"},
	{0xFF2C2A89C3224C91ULL, "6054057_m.fxe"},
	{0xFF2C2D73AC661332ULL, "hlac_ht_bartrand_n_3364_0y.dds"},
	{0xFF2CBC9620AB4080ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugstair.phy"},
	{0xFF2D291B73AF0FA4ULL, "lt_warehouse_probe_1fa.dds"},
	{0xFF2D5778530110E6ULL, "vxht_mansion2b_1xv_0_0.vlm"},
	{0xFF2DAE3974D58DBDULL, "llac_lt_0_lowtown_docks_a3_65535_2.dds"},
	{0xFF2DCEAD3AE6AAD5ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_noba.phy"},
	{0xFF2E437002E0B542ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_run.gad"},
	{0xFF2E614B76F717EFULL, "lt_gallowstemplar_mmi.gff"},
	{0xFF2E8947243E315BULL, "vxw_sundermount_2rv_0_0.vlm"},
	{0xFF2EC7895C0EA8BEULL, "vxw_bonepit_1pa_0_0.vlm"},
	{0xFF2EFA84054EFAF2ULL, "llac_w_encounter2_camp_0_0.dds"},
	{0xFF2F15B87F7E55A7ULL, "mag100_thrask6034813.cl"},
	{0xFF2F27D8C0D0EB20ULL, "mag291_demon_marethari6028872.cl"},
	{0xFF2F4FBD84792BF7ULL, "113850832.ogg"},
	{0xFF2F7BEA6F6746C4ULL, "one280_iwan6097200.cl"},
	{0xFF2F90F7C586599CULL, "6116138_m.fxe"},
	{0xFF2FEB15C0FBF121ULL, "730047976.ogg"},
	{0xFF3018ED860BB309ULL, "anim_placeables.gda"},
	{0xFF3029FA64191928ULL, "hlac_ht_0_hightown_a3_-3_0x.dds"},
	{0xFF312C436723676AULL, "dae000_arishok6068477.cl"},
	{0xFF31DC48EF08AD10ULL, "59791883.ogg"},
	{0xFF31EEA048E6FA11ULL, "vxht_keep_585_0_0.vlm"},
	{0xFF3276511E84E166ULL, "art\\levels\\races\\elf\\props\\_textures\\prp_post_n.dds"},
	{0xFF32C7347C0C34DFULL, "art\\vfx\\combat\\vfx_ha_ar_armoff_act_c.impa.evt"},
	{0xFF32E204D328119DULL, "6119681_f.fxe"},
	{0xFF333C9676679514ULL, "vxw_encounter2_k9_0_0.vlm"},
	{0xFF337A302A6FEC5DULL, "vxw_encounter1_hx_0_0.vlm"},
	{0xFF33FC5FEF6EEBE7ULL, "vxlt_0_lowtown_n_en2_0_0.vlm"},
	{0xFF34280A75E8432EULL, "fen310_fenris6118252.cl"},
	{0xFF34A8C9CDE312C5ULL, "vxlt_darktown_5rx_0_0.vlm"},
	{0xFF355CD07302AEA0ULL, "6020774_m.fxe"},
	{0xFF35667002E70F39ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_run.evt"},
	{0xFF356713DD32161CULL, "245319754.ogg"},
	{0xFF35791158722616ULL, "hlac_lt_undercityencounter_-1_29y.dds"},
	{0xFF359AED4B57DFB1ULL, "6177369_m.fxe"},
	{0xFF360B7E05308B88ULL, "dae000_gen_turn_in_16178388.cl"},
	{0xFF364AA857597802ULL, "llac_ca_cavern1_haunted_24424_11.dds"},
	{0xFF367439F49C1325ULL, "dae011_conspiracy_mob6149572.cl"},
	{0xFF367B0C247B4815ULL, "w_bonepitmining_a2_113_0.msh"},
	{0xFF367D0C247B4B8BULL, "w_bonepitmining_a2_113_0.mmh"},
	{0xFF36890C247B6028ULL, "w_bonepitmining_a2_113_0.mao"},
	{0xFF36B166E8B17C82ULL, "vxca_caverns1_mines_cod_0_0.vlm"},
	{0xFF37603C77C21AD6ULL, "nat380_nathaniel6068362.cl"},
	{0xFF376BE69DCC59A9ULL, "llac_lt_0_lowtown_docks_a3_-2_1.dds"},
	{0xFF37E0C41E3136D1ULL, "366025847.ogg"},
	{0xFF381D7C3F6716D1ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxb.mmh"},
	{0xFF38297C3F672B72ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxb.mao"},
	{0xFF38377C3F67428FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\0ff_tabletopxb.msh"},
	{0xFF38ABF2CAF72A23ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs106a_0s.dds"},
	{0xFF38E21218239AB7ULL, "6048476_m.fxe"},
	{0xFF3988F2D40DB45FULL, "dre000_dougal6129715.cl"},
	{0xFF39940E12C54C3AULL, "vxht_mansion2b_1u9_0_0.vlm"},
	{0xFF39A8347C121EC8ULL, "art\\vfx\\combat\\vfx_ha_ar_armoff_act_c.impa.gad"},
	{0xFF39B26CD9E209E4ULL, "6154503_m.fxe"},
	{0xFF3A04B1E6F0E60CULL, "hlac_lt_warehouse_65536_0x.dds"},
	{0xFF3A5F66D8D7BC46ULL, "mer120_hubert6077167.cl"},
	{0xFF3A8981FDE235C9ULL, "gen00fl_aveline6132750.cl"},
	{0xFF3A8AAC4905D76EULL, "gen00fl_sebastian6177295.cl"},
	{0xFF3B4D64C8FB36D7ULL, "6136878_m.fxe"},
	{0xFF3B685344F2ADC5ULL, "474001887.ogg"},
	{0xFF3BC95544F72966ULL, "vxht_mansion2b_1sw_0_0.vlm"},
	{0xFF3C622E295CE640ULL, "hla_ht_0_hightown_131072_0x.dds"},
	{0xFF3D39664E04C13EULL, "vxw_bonepit_1ub_0_0.vlm"},
	{0xFF3D74DF2E1E4279ULL, "6032192_m.fxe"},
	{0xFF3E099D5EA2FBEDULL, "genip_cin_buzzard.utp"},
	{0xFF3E2D8EA2AF0214ULL, "6060235_m.fxe"},
	{0xFF3F2CDAB7094849ULL, "vxlt_shop_dq_0_0.vlm"},
	{0xFF3F6E3E4E682739ULL, "vxlt_0_lowtown_a2_glc_0_0.vlm"},
	{0xFF3F99F889EA0969ULL, "6096419_m.fxe"},
	{0xFF3FD81FB895C24CULL, "dae000_gamlen6005776.cl"},
	{0xFF402F7FA3CE2E3BULL, "hlac_lt_gallowstemplar_f_2394_12y.dds"},
	{0xFF408FF98579C06DULL, "23547808.ogg"},
	{0xFF40924E34E588BBULL, "llac_lt_backalley_n_65537_6.dds"},
	{0xFF40ADE048DDB7C2ULL, "illuminationmaskmult.psh.fxo"},
	{0xFF40D3374DDA6BA2ULL, "llac_ht_chantry_n_2534_20.dds"},
	{0xFF40FA1258ECD2BBULL, "117520339.ogg"},
	{0xFF4139CF4716B42EULL, "6086448_m.fxe"},
	{0xFF4162CDF95896CAULL, "6137692_m.fxe"},
	{0xFF4179FABC2C019FULL, "646177914.ogg"},
	{0xFF41A4258E50CB9DULL, "llac_lt_foundry_n_320_0.dds"},
	{0xFF41AE960B6FFC11ULL, "vxht_0_hightown_a3_auw_0_0.vlm"},
	{0xFF422C55C7F4882EULL, "vxht_mansion_n_53i_0_1.vlm"},
	{0xFF425496786D9178ULL, "vxca_caverns1_k6l_0_0.vlm"},
	{0xFF426E7002F1A728ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\dialogue\\exit_left_run.ani"},
	{0xFF4298B55F2127A8ULL, "llac_lt_0_lowtown_docks_-196609_0.dds"},
	{0xFF429BE0DF50D75EULL, "hlac_lt_0_lowtown_docks_a3_-3_7z.dds"},
	{0xFF42F28E7BA0D777ULL, "dae000_porfiria6155265.cl"},
	{0xFF42FF2537FB2995ULL, "vxht_0_hightown_a2_bwg_0_0.vlm"},
	{0xFF431AA6BFD92B45ULL, "vxlt_darktown_5wg_0_0.vlm"},
	{0xFF432194077C4613ULL, "hlac_ca_caverns1_mines_24424_24y.dds"},
	{0xFF4410CCA4ECAABEULL, "567934763.ogg"},
	{0xFF4537E450342F86ULL, "mag311_nyssa6042407.cl"},
	{0xFF45BBCECC885E7BULL, "hlac_lt_gallowstemplar_n_2394_8z.dds"},
	{0xFF45D1E7158EDC0FULL, "art\\levels\\races\\qunari\\props\\military\\prp_qunarifort1a.phy"},
	{0xFF45FA77C35A8D56ULL, "art\\vfx\\combat\\vfx_blank_c.anb"},
	{0xFF4602F7755AA951ULL, "llac_ht_mansion2_818_3.dds"},
	{0xFF46B40CBF47AC32ULL, "zz_hurlockcharge.cut"},
	{0xFF47245923C19D21ULL, "300877291.ogg"},
	{0xFF473859A632C837ULL, "ch2armtnt_d.psh.fxo"},
	{0xFF476A4D7465604FULL, "vxw_bonepitmining_a2_mr_0_0.vlm"},
	{0xFF4798FBDCD37374ULL, "dae000_arishok6079572.cl"},
	{0xFF479CEC031783A9ULL, "6119898_f.fxe"},
	{0xFF484DA9F23776F2ULL, "vxca_caverns2_97t_0_0.vlm"},
	{0xFF4885E060CCD5F5ULL, "areadata_seb.gda"},
	{0xFF48CB29E24C9574ULL, "vrc321tr_spooky_scenes_4.utt"},
	{0xFF4905C8554AF452ULL, "vxlt_gallowsprison_n_6ns_0_0.vlm"},
	{0xFF490E45C53E06DEULL, "follower_banter6149241.cl"},
	{0xFF49217850E734CDULL, "93776065.ogg"},
	{0xFF4927D61AE1D013ULL, "hlac_w_encounter3_-393214_0y.dds"},
	{0xFF4968848B6540A7ULL, "vxlt_undercityencounter_11e_0_0.vlm"},
	{0xFF49B41F76A7D6A1ULL, "hlac_ht_keep_n_3732_17z.dds"},
	{0xFF49E821DCBC2EF7ULL, "hlac_w_woundedcoast_-65529_0z.dds"},
	{0xFF49F75BFE2366ECULL, "art\\vfx\\combat\\vfx_r_gn2_accu_act_c.impa.ani"},
	{0xFF4A302B7C772453ULL, "163338979.ogg"},
	{0xFF4A56B20F040C53ULL, "vxht_mansion_n_4zw_0_0.vlm"},
	{0xFF4ACEC701C6CF6BULL, "6023887_f.fxe"},
	{0xFF4B39AEFDEE5EC6ULL, "gen00fl_aveline6099719.cl"},
	{0xFF4B60B679FDD7DFULL, "pose_generation.gda"},
	{0xFF4C1E7B0D037D9BULL, "6163568_m.fxe"},
	{0xFF4C4C365C56C69AULL, "490842506.ogg"},
	{0xFF4C4EFC0FC65456ULL, "lt_0_lowtown_docks_a3_n_114_0.msh"},
	{0xFF4C517901A1ABCDULL, "w_encounter3_134_0.mmh"},
	{0xFF4C557901A1B21EULL, "w_encounter3_134_0.mao"},
	{0xFF4C58FC0FC66554ULL, "lt_0_lowtown_docks_a3_n_114_0.mmh"},
	{0xFF4C5CFC0FC66C67ULL, "lt_0_lowtown_docks_a3_n_114_0.mao"},
	{0xFF4C637901A1CA53ULL, "w_encounter3_134_0.msh"},
	{0xFF4D71B32A8EE9E1ULL, "aura2.mao"},
	{0xFF4DAD36D9B4D3FFULL, "llac_lt_0_lowtown_a2_n_c_-1_1.dds"},
	{0xFF4DD66A63FB6BD4ULL, "6202112_m.fxe"},
	{0xFF4DE01B405E3FB9ULL, "vxlt_0_lowtown_a2_gcb_0_0.vlm"},
	{0xFF4E4B8919273B75ULL, "vxht_0_hightown_n_ben_0_0.vlm"},
	{0xFF4E6EC8ECCE618DULL, "w_sundermount_148_0.phy"},
	{0xFF4F158A68EF5A02ULL, "6163738_m.fxe"},
	{0xFF4F61E20772E6A6ULL, "vxdr_0_deeproads_euf_0_0.vlm"},
	{0xFF4FB17AE9CE5B5BULL, "564218263.ogg"},
	{0xFF4FC07741094AAFULL, "art\\characters\\playercharacter\\qunarimale\\mt_bas_qm.phy"},
	{0xFF50045BFE27E82EULL, "art\\vfx\\combat\\vfx_r_gn2_accu_act_c.impa.gad"},
	{0xFF505163191A6AA6ULL, "6043278_m.fxe"},
	{0xFF508A61F81D5C94ULL, "fx_m1c4.dds"},
	{0xFF50AF9F893A5D59ULL, "6085262_m.fxe"},
	{0xFF5160BC3D7B8371ULL, "hlac_w_bonepitdestroyed_a3_131072_0y.dds"},
	{0xFF51A60E9170B850ULL, "455570110.ogg"},
	{0xFF51AE1F538518EFULL, "542331193.ogg"},
	{0xFF51B19DCEC5CF16ULL, "art\\vfx\\combat\\vfx_ha_bd_bdrain_proj_c.mmh"},
	{0xFF51D99620CA8E15ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugstair.msh"},
	{0xFF51DB9620CA918BULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugstair.mmh"},
	{0xFF51E79620CAA628ULL, "art\\levels\\races\\ferelden\\kirkhigh\\int\\4co_hightownint\\9ff_rugstair.mao"},
	{0xFF51EB6665412876ULL, "391726727.ogg"},
	{0xFF51F44A96F05802ULL, "dae011_corff_sta6152750.cl"},
	{0xFF524FAF945C08D5ULL, "hlac_ht_chantry_n_2534_37y.dds"},
	{0xFF52ABDCBE749FFDULL, "337396598.ogg"},
	{0xFF52BC25D4F3D949ULL, "vxlt_gallowscourtyard_73o_0_0.vlm"},
	{0xFF52FD77C3651CC1ULL, "art\\vfx\\combat\\vfx_blank_c.mmh"},
	{0xFF5323397255062BULL, "ts_chantry_wpn_props.gda"},
	{0xFF533D14740D7779ULL, "432749707.ogg"},
	{0xFF53B4672D21B812ULL, "721531760.ogg"},
	{0xFF53BB0EE65EFF79ULL, "w_encounter1_0_0.msh"},
	{0xFF53C39269378238ULL, "183264460.ogg"},
	{0xFF53C50EE65F1067ULL, "w_encounter1_0_0.mmh"},
	{0xFF53C90EE65F16B4ULL, "w_encounter1_0_0.mao"},
	{0xFF53E70045399A0EULL, "follower_banter6172801.cl"},
	{0xFF53EDA1411C31BEULL, "hlac_lt_0_lowtown_a2_0_1x.dds"},
	{0xFF546BA2D484A560ULL, "vxht_chantry_n_35r_0_0.vlm"},
	{0xFF547060147AE1FFULL, "mag110ip_store_arthuris.utp"},
	{0xFF54AF0F44CEA3F4ULL, "6013255_m.fxe"},
	{0xFF54BCE7B7122BAEULL, "ss_midclass_thugs_f_1.wbd"},
	{0xFF556912145BDDBEULL, "sse_ss_circlemages_f_1.gda"},
	{0xFF559974DA8F6481ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_brother_victim.ani"},
	{0xFF55D1D231DD4EA3ULL, "dre000_mother6037125.cl"},
	{0xFF56FF5BFE2DFE0DULL, "art\\vfx\\combat\\vfx_r_gn2_accu_act_c.impa.evt"},
	{0xFF577B39DE623E0FULL, "207503336.ogg"},
	{0xFF57BF9EA4B5CDADULL, "6013386_m.fxe"},
	{0xFF57EE8A5F0E8D23ULL, "hlac_lt_0_lowtown_docks_-262143_0x.dds"},
	{0xFF58C73036642794ULL, "919343476.ogg"},
	{0xFF596888F97EBE13ULL, "407793979.ogg"},
	{0xFF5A95BC484EAA47ULL, "one101ar_ggt_cavern.ncs"},
	{0xFF5AEC8D6D005740ULL, "vxlt_gallowstemplar_5c6_0_0.vlm"},
	{0xFF5B412EB5990E4BULL, "hlac_lt_0_lowtown_a2_n_c_-2_5y.dds"},
	{0xFF5B674DB7FB60E0ULL, "art\\vfx\\materials\\level_fx\\vfx_lvl_flys.mao"},
	{0xFF5BA474DA93E323ULL, "art\\cinematics\\_crit path\\prologue\\pro000cs_victim\\cs_brother_victim.gad"},
	{0xFF5BE7EA7DC36368ULL, "vxht_0_hightown_a3_bpv_0_0.vlm"},
	{0xFF5CA67695631495ULL, "vxca_caverns2_8p7_0_0.vlm"},
	{0xFF5D342AC9D2475DULL, "6060414_m.fxe"},
	{0xFF5E6E5F91F7C7C0ULL, "6085743_f.fxe"},
	{0xFF5EDB81DE05346CULL, "833886776.ogg"},
	{0xFF5F15442AE9F9C2ULL, "618219519.ogg"},
	{0xFF5F688C69D9D0CAULL, "298843855.ogg"},
	{0xFF5F89C0C61573B1ULL, "vxht_mansion2b_3g6_0_0.vlm"},
	{0xFF5FA3F658A519F9ULL, "6107199_m.fxe"},
	{0xFF6054368AA37EA9ULL, "6005830_f.fxe"},
	{0xFF6100234AA7C122ULL, "w_encounter1_ll.mmh"},
	{0xFF6175BC485492F6ULL, "one101ar_ggt_cavern.lst"},
	{0xFF620466C814DE12ULL, "bd_rock.mao"},
	{0xFF622DFAEE1A2BA0ULL, "6169341_m.fxe"},
	{0xFF625C41B34EC24EULL, "6135081_m.fxe"},
	{0xFF6277A30A4188E8ULL, "dae000_hate_enemies6167622.cl"},
	{0xFF62CF2D7EB00565ULL, "vxca_caverns2_8cz_0_0.vlm"},
	{0xFF62DD7FC67ABDE0ULL, "vxht_chantry_n_32s_0_0.vlm"},
	{0xFF6355BDE1049857ULL, "dae000_mother6103049.cl"},
	{0xFF6377F3AD0061FEULL, "503428345.ogg"},
	{0xFF64A5342F32A464ULL, "448193021.ogg"},
	{0xFF65005DF28B0025ULL, "6033577_m.fxe"},
	{0xFF657A9F95E91DE2ULL, "hlac_lt_0_lowtown_a2_n_c_-196609_2x.dds"},
	{0xFF65C551FF5E159CULL, "hlac_dr_primevaltomb_661_1z.dds"},
	{0xFF65DBF6D11A5B26ULL, "806805065.ogg"},
	{0xFF65E2879278E454ULL, "vxlt_gallowsprison_n_64k_0_0.vlm"},
	{0xFF664E372BC3F8A4ULL, "gen00fl_carver6122539.cl"},
	{0xFF6675D43875060AULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfbase.mao"},
	{0xFF6679D438750D59ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfbase.mmh"},
	{0xFF6683D438751E47ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfbase.msh"},
	{0xFF674E6B7850FF5FULL, "medbush_dp.dds"},
	{0xFF67803E01313A2CULL, "246151544.ogg"},
	{0xFF679161F83047C2ULL, "fx_m1c4.mao"},
	{0xFF6912DCC871706CULL, "555761968.ogg"},
	{0xFF6975AF9E18C51BULL, "vxlt_warehouse_qo_0_0.vlm"},
	{0xFF69EA2945D6957BULL, "seb321_leliana6199281.cl"},
	{0xFF6A0A11BE8D2DEBULL, "hlac_lt_0_lowtown_n_131072_0y.dds"},
	{0xFF6A14477E69F3E4ULL, "6069183_f.fxe"},
	{0xFF6A3C3764C143BBULL, "6167480_f.fxe"},
	{0xFF6A59C76FB97BE8ULL, "vxht_0_hightown_a3_bow_0_0.vlm"},
	{0xFF6B65D8076B1998ULL, "231206716.ogg"},
	{0xFF6B84BC485D00E1ULL, "one101ar_ggt_cavern.are"},
	{0xFF6BD957D8F8195DULL, "vxht_mansion_n_5co_0_0.vlm"},
	{0xFF6C41911D050755ULL, "vxht_0_hightown_9h1_0_0.vlm"},
	{0xFF6C5D249F2DA670ULL, "6129272_f.fxe"},
	{0xFF6CC9720C8FEF24ULL, "w_bonepitdestroyed_a3_237_0.mmh"},
	{0xFF6CCD720C8FF677ULL, "w_bonepitdestroyed_a3_237_0.mao"},
	{0xFF6CD53AB931B722ULL, "6129065_f.fxe"},
	{0xFF6CD80B2BF73AEEULL, "hlac_lt_gallowstemplar_f_7198_0x.dds"},
	{0xFF6CDF720C9014C6ULL, "w_bonepitdestroyed_a3_237_0.msh"},
	{0xFF6CE897FB21B4EBULL, "122630762.ogg"},
	{0xFF6D7FF6705753F3ULL, "character_character_p215.psh.fxo.dx11.assembly"},
	{0xFF6D95082EC04CEDULL, "ser121_jethann6105772.cl"},
	{0xFF6DECB7BBAEBA07ULL, "441571027.ogg"},
	{0xFF6DFAA98DF76B32ULL, "mrl310_merrill6144189.cl"},
	{0xFF6E49DF1D879FF6ULL, "lt_gallowscourtyard_77_0_lc.mmh"},
	{0xFF6EB3B3FA323619ULL, "6112910_f.fxe"},
	{0xFF6ED5C926207FE0ULL, "vxca_caverns2_1z9_0_0.vlm"},
	{0xFF6F18F7F1E7DA1FULL, "vxdr_0_deeproads_ey7_0_0.vlm"},
	{0xFF6F2972751DE245ULL, "hlac_ht_0_hightown_a3_-131070_0x.dds"},
	{0xFF6F2A1DD72B0C8EULL, "6069810_m.fxe"},
	{0xFF6F9FDFB8E00A49ULL, "vxlt_gallowscourtyard_79m_0_0.vlm"},
	{0xFF70AC27F4D54305ULL, "vxca_caverns1_k6u_0_0.vlm"},
	{0xFF7150EA2EA1777EULL, "hlac_lt_undercityencounter_-1_40x.dds"},
	{0xFF7192A654BBFCDBULL, "211134036.ogg"},
	{0xFF721A5B6EFD182AULL, "zz_followers_banter6159141.cl"},
	{0xFF7284F933DD5D93ULL, "hlac_lt_darktown_-1_12x.dds"},
	{0xFF72F6EB3A31E6AAULL, "mag100_thrask6034872.cl"},
	{0xFF732FF6A73CAF48ULL, "art\\levels\\races\\natural\\kirkcave\\setint\\_textures\\til_lava_n.dds"},
	{0xFF738F66F12B82D7ULL, "one100ar_ironwood.ncs"},
	{0xFF742F55B8E44B72ULL, "866508100.ogg"},
	{0xFF74B4A13611BEAFULL, "vxw_encounter2_os_0_1.vlm"},
	{0xFF74E1AD3B22CDF0ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_noba.msh"},
	{0xFF74EFAD3B22E50AULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_noba.mmh"},
	{0xFF74F3AD3B22EC41ULL, "art\\characters\\playercharacter\\humanfemale\\hf_cth_noba.mao"},
	{0xFF74F441B18DEC28ULL, "vxht_0_hightown_n_bv2_0_0.vlm"},
	{0xFF75713A527BB433ULL, "lt_0_lowtown_20_0_lc.mmh"},
	{0xFF7572EB62078238ULL, "zz_fen_debug6165673.cl"},
	{0xFF757AD85391F062ULL, "6206385_m.fxe"},
	{0xFF765B1310D38F8FULL, "6164709_m.fxe"},
	{0xFF767D2CEACF2FADULL, "gen_im_wep_rog_dua_act1_07.uti"},
	{0xFF76AC7E8695EF13ULL, "dae100_gamlen_letter6146997.cl"},
	{0xFF76F5D88F126DD5ULL, "116139060.ogg"},
	{0xFF76F6D0AA0D0E73ULL, "hlac_lt_0_lowtown_a2_n_c_-65536_3x.dds"},
	{0xFF76F8CB6BA1D5A1ULL, "vxht_0_hightown_bi9_0_0.vlm"},
	{0xFF7700DA15DD233EULL, "vxw_bonepit_1if_0_0.vlm"},
	{0xFF77D5E2E35C1DBBULL, "art\\characters\\weapons\\dagger\\w_dgr_102.mao"},
	{0xFF77E78C900EDD9BULL, "vxlt_warehouse_zn_0_0.vlm"},
	{0xFF7897301ACCEA19ULL, "w_bonepit_238_0.msh"},
	{0xFF789D24647195B6ULL, "dae000_tomwise6167267.cl"},
	{0xFF78A1301ACCFB07ULL, "w_bonepit_238_0.mmh"},
	{0xFF78A5301ACD0254ULL, "w_bonepit_238_0.mao"},
	{0xFF78C6A99114A4A8ULL, "6029060_f.fxe"},
	{0xFF78D389089D4FA7ULL, "6127102_m.fxe"},
	{0xFF78FBA4800C9285ULL, "101302587.ogg"},
	{0xFF7908F6ED092254ULL, "111267011.ogg"},
	{0xFF7980DACCF37915ULL, "vxht_0_hightown_bu1_0_0.vlm"},
	{0xFF7A1E4D41CBF774ULL, "art\\characters\\playercharacter\\humanmale\\hlh_vael.phy"},
	{0xFF7A8F66F131A126ULL, "one100ar_ironwood.lst"},
	{0xFF7AAD73EB950347ULL, "vxlt_backalley_g_1xv_0_0.vlm"},
	{0xFF7AC66393EE5325ULL, "darktown_male_lowlife_mob6174407.cl"},
	{0xFF7AEA3628F736ECULL, "vxlt_0_lowtown_a2_elj_0_2.vlm"},
	{0xFF7B07B1401305B8ULL, "6170941_m.fxe"},
	{0xFF7BA9DA3BD55CF7ULL, "gen00fl_anders6142547.cl"},
	{0xFF7BCDCAE6561751ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast_mini.ani"},
	{0xFF7BDB31DE77B066ULL, "hlac_ca_cavern1_haunted_24424_24z.dds"},
	{0xFF7C94BB6EBCEEECULL, "dae000_flemeth6001355.cl"},
	{0xFF7C9C9F06A8454EULL, "w_encounter2_camp_11_0.msh"},
	{0xFF7CAA9F06A85C9FULL, "w_encounter2_camp_11_0.mao"},
	{0xFF7CB69F06A8713CULL, "w_encounter2_camp_11_0.mmh"},
	{0xFF7D3A0592579C6FULL, "vxlt_warehouse_n_11i_0_0.vlm"},
	{0xFF7D663C28A2FE2EULL, "1035743205.ogg"},
	{0xFF7D9513D04BAB19ULL, "581569025.ogg"},
	{0xFF7DBA6CF0F25F00ULL, "zz_followers_banter6157664.cl"},
	{0xFF7DBB1C5962E441ULL, "art\\levels\\races\\ferelden\\props\\_textures\\prp_candle_d.dds"},
	{0xFF7DE91DF402ACEDULL, "plt_one100pt_herb_ironwood.nss"},
	{0xFF7E94504642AC26ULL, "ave310_brennan6136777.cl"},
	{0xFF7EA9CD0EBFE945ULL, "art\\levels\\races\\proxy\\props\\_textures\\tables_d.dds"},
	{0xFF7EC5DE6F170209ULL, "546660709.ogg"},
	{0xFF7ED12C93E2FDA9ULL, "vxlt_0_lowtown_docks_n_ctm_0_0.vlm"},
	{0xFF7F22E6C50BA462ULL, "397533365.ogg"},
	{0xFF7FAAB411D993F3ULL, "6171079_m.fxe"},
	{0xFF7FAE0EE684239CULL, "w_encounter1_0_0.phy"},
	{0xFF7FAF1C16A5FE18ULL, "456342679.ogg"},
	{0xFF8039F5EBDA3A0AULL, "one000_ilen6057371.cl"},
	{0xFF80610ABE0F0F7AULL, "hlac_lt_0_lowtown_docks_a3_n_-196610_0x.dds"},
	{0xFF80662E28D0A011ULL, "hf_terrie.mor"},
	{0xFF8160D321E3CECDULL, "gen00fl_sebastian6179599.cl"},
	{0xFF817512F8A08BEEULL, "472843981.ogg"},
	{0xFF8198B095146C28ULL, "6173764_m.fxe"},
	{0xFF81B2E7B73908DCULL, "ss_midclass_thugs_f_1.bnk"},
	{0xFF81B734B1B5817BULL, "art\\characters\\weapons\\longswords\\textures\\w_lsw_018a_0s.dds"},
	{0xFF81D8CAE65A95F3ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast_mini.gad"},
	{0xFF8270D707CAD6DEULL, "6070801_m.fxe"},
	{0xFF82EF4FE9207C65ULL, "vxht_0_hightown_a2_bx6_0_0.vlm"},
	{0xFF831F2742F368B5ULL, "6217087_m.fxe"},
	{0xFF8370A78488DB42ULL, "ht_chantry.arl"},
	{0xFF83FB4D41D41071ULL, "art\\characters\\playercharacter\\humanmale\\hlh_vael.msh"},
	{0xFF84094D41D4278CULL, "art\\characters\\playercharacter\\humanmale\\hlh_vael.mao"},
	{0xFF84154D41D43C2FULL, "art\\characters\\playercharacter\\humanmale\\hlh_vael.mmh"},
	{0xFF84493F19C28404ULL, "816900608.ogg"},
	{0xFF844A4CFBC402A1ULL, "vxht_mansion_n_4yd_0_0.vlm"},
	{0xFF84685DEB915A89ULL, "hlac_lt_0_lowtown_a2_n_-131073_3x.dds"},
	{0xFF849CC09C09CA4DULL, "vxw_sundermount_24b_0_0.vlm"},
	{0xFF849E66F13A0F31ULL, "one100ar_ironwood.are"},
	{0xFF85211C3E26B9C3ULL, "vxht_mansion2b_2z2_0_0.vlm"},
	{0xFF859B824D558480ULL, "vxw_bonepitmining_a2_e9_0_0.vlm"},
	{0xFF85E8A63FD3FFA5ULL, "715948722.ogg"},
	{0xFF85FE4C4498405DULL, "vxlt_hangedman_1po_0_0.vlm"},
	{0xFF86422442E2A0C0ULL, "vxlt_gallowstemplar_5r5_0_0.vlm"},
	{0xFF86C50EE68A69C4ULL, "w_encounter1_0_0.rml"},
	{0xFF86D1AA9E1ABB74ULL, "follower_banter6169843.cl"},
	{0xFF873D8153A5ACE8ULL, "vxht_0_hightown_a3_bau_0_0.vlm"},
	{0xFF876974C46E5A5BULL, "741400621.ogg"},
	{0xFF8771FFF0E68247ULL, "art\\characters\\creatures\\gallowsgateguard\\textures\\cn_wep_ggg_0d.dds"},
	{0xFF877C741C32A7F8ULL, "hlac_lt_0_lowtown_65533_1z.dds"},
	{0xFF87F13C3B1DE339ULL, "hlac_ca_caverns1_mines_24424_25x.dds"},
	{0xFF8817487BED0436ULL, "392442.ogg"},
	{0xFF8870DD9663CAA2ULL, "vxlt_warehouse_n_qk_0_0.vlm"},
	{0xFF88B80EA31A67F5ULL, "6121141_f.fxe"},
	{0xFF88BD11BCE44A5DULL, "vxht_mansion_n_5im_0_0.vlm"},
	{0xFF88D5CAE660AF58ULL, "art\\characters\\creatures\\shriek\\animation\\combat\\devestationcast_mini.evt"},
	{0xFF88F1D2D0B5F4AAULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01fern.mao"},
	{0xFF88F5D2D0B5FBF9ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01fern.mmh"},
	{0xFF88FFD2D0B60CE7ULL, "art\\levels\\races\\natural\\kirkcoastal\\foliage\\bsh_01fern.msh"},
	{0xFF896954B621721EULL, "6020552_m.fxe"},
	{0xFF89ABF45474A1F2ULL, "584854701.ogg"},
	{0xFF89CE1B228D9206ULL, "904647772.ogg"},
	{0xFF8A36F954F46891ULL, "6152579_m.fxe"},
	{0xFF8AA267EED4B0CFULL, "146326265.ogg"},
	{0xFF8AFC93944B6BBCULL, "dae021cr_sister_samea.utc"},
	{0xFF8B5480B6C0B1C1ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_down.ani"},
	{0xFF8B6D0933D4823FULL, "vrc221cr_crazed_warrior.utc"},
	{0xFF8BABE2844DB4EFULL, "vxlt_warehouse_n_16h_0_0.vlm"},
	{0xFF8BDFA1843927D7ULL, "vxw_encounter2_camp_qi_0_1.vlm"},
	{0xFF8C15823A5B145EULL, "960719918.ogg"},
	{0xFF8C37D1E83EB2B1ULL, "6198681_m.fxe"},
	{0xFF8C72BAF933623DULL, "art\\levels\\races\\proxy\\props\\woodenstuff\\prp_zzlogpile_0.pwk"},
	{0xFF8D1600067D84BAULL, "671237743.ogg"},
	{0xFF8D430985D91629ULL, "vxlt_0_lowtown_docks_n_col_0_0.vlm"},
	{0xFF8D79515F720EB7ULL, "6209976_m.fxe"},
	{0xFF8DBC8A4D7508B4ULL, "6148858_m.fxe"},
	{0xFF8DD3912C6567FAULL, "t1_mul_bk3.tnt"},
	{0xFF8E192A3D1535A1ULL, "vxlt_gallowstemplar_5k3_0_0.vlm"},
	{0xFF8E24C454924EE5ULL, "vxca_caverns2_8pg_0_0.vlm"},
	{0xFF8E356D80FBE9DDULL, "1042665302.ogg"},
	{0xFF8E7A85377CA8BEULL, "dae000ip_crowd_em2_talk5.utp"},
	{0xFF8E9C5E164A7345ULL, "544057651.ogg"},
	{0xFF8F0AE68AAECAE0ULL, "vxca_cavern1_haunted_j2l_0_0.vlm"},
	{0xFF8F1FBCC9304283ULL, "vxht_0_hightown_a3_c8h_0_0.vlm"},
	{0xFF8F434152BE9BD2ULL, "humanmale.fxa"},
	{0xFF8F6F79A3CFEC2BULL, "vxlt_0_lowtown_a3_n_c_gch_0_0.vlm"},
	{0xFF8F799859E61525ULL, "24237544.ogg"},
	{0xFF8FC3C0F3586BAFULL, "705119259.ogg"},
	{0xFF904187B695DB75ULL, "vxca_caverns1_i38_0_0.vlm"},
	{0xFF905B8F082432AAULL, "6170491_m.fxe"},
	{0xFF90D30D111A1476ULL, "zz_dae_debug6142961.cl"},
	{0xFF9100C94E8A2A58ULL, "ave121_jeven6105715.cl"},
	{0xFF9158779E39F2F9ULL, "follower_banter6136721.cl"},
	{0xFF91ABCB279DE9B0ULL, "vxw_bonepitmining_a2_dz_0_0.vlm"},
	{0xFF925F80B6C6E363ULL, "art\\characters\\playercharacter\\humanmale\\animation\\cindes\\additive\\head_down.gad"},
	{0xFF92BA16F556B67FULL, "llac_w_bonepitmining_a2_327681_0.dds"},
	{0xFF9302E0FAD9315DULL, "vxca_caverns1_mines_fmd_0_0.vlm"},
	{0xFF9362F411C8B1B9ULL, "vxlt_gallowstemplar_f_6dj_0_0.vlm"},
	{0xFF9375F513F3BEE6ULL, "one200_gold_followers6120134.cl"},
	{0xFF93FF2BAD0E199DULL, "99701278.ogg"},
	{0xFF9424039DA7CD60ULL, "423149089.ogg"},
	{0xFF948E6E4FD10E0FULL, "zz_approval_debug6003718.cl"},
	{0xFF94D477E1666E1DULL, "688319798.ogg"},
	{0xFF95395D192824B4ULL, "606722629.ogg"},
	{0xFF958AAB23A60B01ULL, "vxca_caverns1_mines_gkk_0_0.vlm"},
	{0xFF96547965CFFCECULL, "hlac_w_bonepitmining_a2_393217_0y.dds"},
	{0xFF96C430C4616BF8ULL, "6144263_m.fxe"},
	{0xFF96E2BA8859E322ULL, "vxlt_warehouse_n_tt_0_0.vlm"},
	{0xFF970CA0C485D6D0ULL, "17348186.ogg"},
	{0xFF9730393C960EC3ULL, "6095346_m.fxe"},
	{0xFF9742C5825E2FC9ULL, "hlac_lt_0_lowtown_docks_n_-3_6x.dds"},
	{0xFF97C054960EBCB7ULL, "6176719_m.fxe"},
	{0xFF97DB31A8178A9EULL, "6027452_m.fxe"},
	{0xFF98BF3C768F91DBULL, "hlac_w_woundedcoast_131077_0y.dds"},
	{0xFF997CA608056F8FULL, "gen00fl_fenris6127063.cl"},
	{0xFF99C6C9AFD129F6ULL, "6036758_f.fxe"},
	{0xFF9A60754273CBDAULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mb.msh"},
	{0xFF9A6E754273E313ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mb.mao"},
	{0xFF9A72754273EA20ULL, "art\\levels\\races\\ferelden\\kirklow\\ext\\4co_lowtown\\9uw_tarp12mb.mmh"},
	{0xFF9AA75FA0E6EACDULL, "6070442_m.fxe"},
	{0xFF9AB69F06C16E8FULL, "w_encounter2_camp_11_0.rml"},
	{0xFF9B84CCFF84FD6BULL, "vxlt_warehouse_w9_0_0.vlm"},
	{0xFF9C0B1568F301D3ULL, "hlac_w_encounter3_-262147_0y.dds"},
	{0xFF9C0C8F7CE70450ULL, "6180847_m.fxe"},
	{0xFF9C327942D8505FULL, "109845840.ogg"},
	{0xFF9C5B85C2FE3F24ULL, "lgt310_mis_burn_murder26152764.cl"},
	{0xFF9C8B072F0B4E21ULL, "vxlt_gallowstemplar_5t2_0_0.vlm"},
	{0xFF9CB465064FD32EULL, "vxw_encounter2_camp_rq_0_1.vlm"},
	{0xFF9D7D1FA0DCCFBDULL, "238375854.ogg"},
	{0xFF9DA82FA7A6EBA5ULL, "905222951.ogg"},
	{0xFF9DD79FD7138376ULL, "vxdr_0_deeproads_crt_0_1.vlm"},
	{0xFF9DF0AF315C79CFULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleftcrn.mao"},
	{0xFF9DFCAF315C8E6CULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleftcrn.mmh"},
	{0xFF9E02AF315C985EULL, "art\\levels\\races\\dwarf\\common\\int\\deeproads\\3wi_grandleftcrn.msh"},
	{0xFF9F3C5F2186E0EDULL, "272794475.ogg"},
	{0xFF9F694524C775BFULL, "vxca_cavern1_haunted_gk5_0_0.vlm"},
	{0xFF9F71D1F486E25EULL, "gen00fl_sebastian6180671.cl"},
	{0xFF9FB70AB5490AAAULL, "ser211ar_dungeon.lst"},
	{0xFF9FFC2F19B38CFFULL, "lt_0_lowtown_docks_a3_n_103_0.phy"},
	{0xFFA0051F67605BD7ULL, "vxht_0_hightown_4kg_0_0.vlm"},
	{0xFFA005349FC70EC5ULL, "gen00fl_dog6146962.cl"},
	{0xFFA049D8877035A8ULL, "vxht_0_hightown_n_bg1_0_0.vlm"},
	{0xFFA0584152CD32E7ULL, "humanmale.mor"},
	{0xFFA06E3B61237D72ULL, "gen_im_trs_rdr_hvy_glv_r2.uti"},
	{0xFFA07DD17483BBD3ULL, "196623129.ogg"},
	{0xFFA08445E34A45A1ULL, "hlac_ht_chantry_2534_8z.dds"},
	{0xFFA0955277A8D8F4ULL, "gen00fl_fenris6108029.cl"},
	{0xFFA0A190B7E40596ULL, "sse_ss_qunari_m_1.gda"},
	{0xFFA0F546F199F2B0ULL, "art\\characters\\weapons\\greatswords\\textures\\w_gsw_gs101a_0s.dds"},
	{0xFFA1059F12F42A9FULL, "gen00fl_merrill6150982.cl"},
	{0xFFA149C2A421C364ULL, "vxlt_0_lowtown_docks_a3_d6a_0_0.vlm"},
	{0xFFA168608BA6D6F7ULL, "hlac_w_sundermount_0_6x.dds"},
	{0xFFA1843F208809B5ULL, "676943925.ogg"},
	{0xFFA19208D64A8F99ULL, "6072196_m.fxe"},
	{0xFFA1959F06C75507ULL, "w_encounter2_camp_11_0.phy"},
	{0xFFA22145A3469EF5ULL, "gencr_mercenary_assassin.utc"},
	{0xFFA225450B8D45FEULL, "w_sundermount_89_0_lc.mmh"},
	{0xFFA27029F57F05F4ULL, "vxht_mansion_n_3y2_0_0.vlm"},
	{0xFFA2E7E362A46A91ULL, "6083549_m.fxe"},
	{0xFFA320010668C397ULL, "6093616_m.fxe"},
	{0xFFA371EE5F598A8EULL, "168132076.ogg"},
	{0xFFA379B60038E01AULL, "vxdr_primevaltomb_6un_0_1.vlm"},
	{0xFFA390BC5F7BB636ULL, "vxca_caverns2_8jr_0_0.vlm"},
	{0xFFA3AFADE468D208ULL, "llac_w_bonepitdestroyed_a3_131074_0.dds"},
	{0xFFA4094FA7CD1BE4ULL, "vxlt_darktown_2rn_0_0.vlm"},
	{0xFFA486738C98E721ULL, "6090268_m.fxe"},
	{0xFFA4A8026B8B3B10ULL, "hlac_lt_warehouse_n_65537_2x.dds"},
	{0xFFA5C7B8FA2FB4C2ULL, "988392597.ogg"},
	{0xFFA602E8BB700140ULL, "dae000cr_grand_cleric6124769.cl"},
	{0xFFA64C8B768A0AB4ULL, "hightown_fem_cityguard_mob6168530.cl"},
	{0xFFA66CDA53D88045ULL, "hlac_lt_backalley_g_0_3x.dds"},
	{0xFFA6CC34746A9FE5ULL, "fx_black_blast.dds"},
	{0xFFA724E0FD06EE25ULL, "hlac_lt_gallowstemplar_f_2393_1z.dds"},
	{0xFFA736699877B640ULL, "564998658.ogg"},
	{0xFFA7A2C6D20CCA4AULL, "6171104_m.fxe"},
	{0xFFA7D924EA1635E2ULL, "6132725_m.fxe"},
	{0xFFA82E68053EA130ULL, "6089102_m.fxe"},
	{0xFFA88A1A566A68C8ULL, "vxlt_0_lowtown_docks_a3_n_d3e_0_0.vlm"},
	{0xFFA8B8490DA03CADULL, "mag170_thrask6054012.cl"},
	{0xFFA8F24C69411717ULL, "gen00fl_varric6033627.cl"},
	{0xFFA927974446241BULL, "art\\characters\\playercharacter\\humanfemale\\arm_wic.mmh"},
	{0xFFA93397444638B8ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wic.mao"},
	{0xFFA9459744465705ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wic.msh"},
	{0xFFA94AC995516A61ULL, "and130ip_anders_door.utp"},
	{0xFFA95C159A5012CAULL, "65209019.ogg"},
	{0xFFA97948F6E448A1ULL, "l_dae270ar_gallows.wbd"},
	{0xFFA9F065CB012D7FULL, "6073889_m.fxe"},
	{0xFFA9F6A9F17B15EBULL, "vxlt_warehouse_x8_0_0.vlm"},
	{0xFFAA271033CEF0ECULL, "mer221_sabin6065703.cl"},
	{0xFFAACD0462DDC40AULL, "mag291_vincento6166232.cl"},
	{0xFFAAD51770C0DECDULL, "996021840.ogg"},
	{0xFFAAE7BDC1D211C0ULL, "hlac_ht_0_hightown_a3_2_1y.dds"},
	{0xFFAB5279AF888DD1ULL, "vxlt_hangedman_172_0_0.vlm"},
	{0xFFABBFC22EE68A1EULL, "zz_followers_banter6157367.cl"},
	{0xFFABC8BC1DFC5303ULL, "81318585.ogg"},
	{0xFFAC0376AD1C7383ULL, "vxht_0_hightown_a3_c6j_0_0.vlm"},
	{0xFFACBD0CF6C89446ULL, "gen_im_trs_gem_diamond_02.uti"},
	{0xFFAD1E3175C3FC4EULL, "vxlt_0_lowtown_docks_n_b7v_0_0.vlm"},
	{0xFFAD2A5552A8CA75ULL, "framebuffereffects_range_dof_p0.psh.fxo.dx11.assembly"},
	{0xFFADA49196926BFDULL, "vxca_caverns2prc_af1_0_0.vlm"},
	{0xFFADE72B4768C7ACULL, "character_character_p250.psh.fxo.dx11"},
	{0xFFAE44E6BF02C5E5ULL, "vxht_0_hightown_a2_bk3_0_0.vlm"},
	{0xFFAE4E4C219273B0ULL, "92998131.ogg"},
	{0xFFAE6C5F45A86166ULL, "vxlt_0_lowtown_docks_n_c5a_0_0.vlm"},
	{0xFFAE858F291FEC27ULL, "vxlt_undercityencounter_1lb_0_0.vlm"},
	{0xFFAE8F850B8A1AB0ULL, "vxw_bonepitmining_a2_nx_0_0.vlm"},
	{0xFFAEC486651EE4A9ULL, "gen00fl_fenris6113934.cl"},
	{0xFFAF26B9053DBF34ULL, "vxca_caverns2_80r_0_0.vlm"},
	{0xFFAF6EFC30E18B14ULL, "625759576.ogg"},
	{0xFFB026CFEFFDD431ULL, "157244576.ogg"},
	{0xFFB033AAA499BC39ULL, "llac_dr_0_deeproads_5842_17.dds"},
	{0xFFB04BF00F5F0A85ULL, "190011269.ogg"},
	{0xFFB05C44A80B581DULL, "hlac_dr_0_deeproads_5845_4y.dds"},
	{0xFFB098041E4971A4ULL, "515925232.ogg"},
	{0xFFB0A09AF2B37618ULL, "darktown_male_sickly_sta6175327.cl"},
	{0xFFB0AFD2C9A8D24EULL, "zz_vis_debug6050352.cl"},
	{0xFFB0E206E7751E74ULL, "vxht_mansion_n_3v3_0_0.vlm"},
	{0xFFB0EC04CBD53222ULL, "vxw_sundermount_30u_0_1.vlm"},
	{0xFFB0FD2F19C20104ULL, "lt_0_lowtown_docks_a3_n_103_0.mmh"},
	{0xFFB1012F19C20857ULL, "lt_0_lowtown_docks_a3_n_103_0.mao"},
	{0xFFB1132F19C226A6ULL, "lt_0_lowtown_docks_a3_n_103_0.msh"},
	{0xFFB155E40FE908ACULL, "llac_ca_caverns2_5802_65.dds"},
	{0xFFB1D7CB12FFAFC8ULL, "gen_im_cns_flu_isabela_02.uti"},
	{0xFFB1D9CC422777DAULL, "6116529_m.fxe"},
	{0xFFB1E7BC1265C0E9ULL, "lgt331ar_shw_bloodmage.lst"},
	{0xFFB1E810EA4CE05AULL, "6156770_m.fxe"},
	{0xFFB1E9299A2A61A5ULL, "gen00fl_anders6125612.cl"},
	{0xFFB1F943A5160236ULL, "vxht_0_hightown_bnw_0_0.vlm"},
	{0xFFB202995171CEB6ULL, "vxca_caverns2_8qs_0_0.vlm"},
	{0xFFB22762C5FBE9ADULL, "435706015.ogg"},
	{0xFFB22AC66533F91DULL, "vxlt_0_lowtown_n_bxm_0_0.vlm"},
	{0xFFB27B2C99C33464ULL, "vxlt_darktown_2uo_0_0.vlm"},
	{0xFFB2B056CA8635FEULL, "472348448.ogg"},
	{0xFFB2C9AF21E31937ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1cocoon.msh"},
	{0xFFB2DBAF21E3379AULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1cocoon.mao"},
	{0xFFB2DFAF21E33EA9ULL, "art\\levels\\races\\dwarf\\props\\spider\\prp_1cocoon.mmh"},
	{0xFFB322EF7BEE58C7ULL, "dae000_bartrand6031134.cl"},
	{0xFFB38F6BAD7F16FDULL, "llac_ht_0_hightown_a3_65537_1.dds"},
	{0xFFB3905045CB5615ULL, "vxlt_0_lowtown_a2_hga_0_0.vlm"},
	{0xFFB3C9DB5FA74D58ULL, "vxht_store_ch_0_0.vlm"},
	{0xFFB479FAD4CD1547ULL, "art\\characters\\weapons\\greataxes\\textures\\w_gax_ga102a_0s.dds"},
	{0xFFB4E87CC5683370ULL, "vxlt_hangedman_n_m0_0_1.vlm"},
	{0xFFB4E89A6F7B6C93ULL, "zz_approval_debug6003062.cl"},
	{0xFFB5385D8FDE1B35ULL, "6114008_f.fxe"},
	{0xFFB593220926DA80ULL, "371383700.ogg"},
	{0xFFB69AF504413E5AULL, "genlock_alpha_sbow_mix.utc"},
	{0xFFB6CC3D705A2D9BULL, "vxca_caverns1_i25_0_0.vlm"},
	{0xFFB82CB56FE5384EULL, "vxca_caverns1_mines_ckd_0_0.vlm"},
	{0xFFB8AD5386BC062EULL, "dae000_worthy6166577.cl"},
	{0xFFB8B32589CD019DULL, "457053193.ogg"},
	{0xFFB8BFBC126B9B20ULL, "lgt331ar_shw_bloodmage.ncs"},
	{0xFFB8D6D2A087E0EEULL, "gen00fl_player6207702.cl"},
	{0xFFB92102B57E3E9DULL, "vxca_caverns2_7nr_0_0.vlm"},
	{0xFFB96EC112F77F21ULL, "vxlt_gallowstemplar_5z0_0_0.vlm"},
	{0xFFBA00FE491E7931ULL, "393870441.ogg"},
	{0xFFBA40749513C33AULL, "mer221_sabin6081189.cl"},
	{0xFFBA4159CC740282ULL, "6031661_f.fxe"},
	{0xFFBA740A86709F35ULL, "978429224.ogg"},
	{0xFFBA8C0D967A43F0ULL, "hlac_lt_0_lowtown_a2_n_0_7y.dds"},
	{0xFFBAE6A87881E420ULL, "6027007_m.fxe"},
	{0xFFBBBA33FA4D93ABULL, "art\\levels\\races\\proxy\\props\\4tools\\inv_capsule2x1_0.pwk"},
	{0xFFBBDB1993DBDE23ULL, "571720034.ogg"},
	{0xFFBBE5DD22A8A86EULL, "zz_followers_banter6158289.cl"},
	{0xFFBBFD41F26D8544ULL, "6143060_m.fxe"},
	{0xFFBC2D9C152CDF09ULL, "vxht_mansion2_1wp_0_0.vlm"},
	{0xFFBC51A6369AF92EULL, "737924914.ogg"},
	{0xFFBCB6C3B0F8DE65ULL, "vxht_0_hightown_a2_bl2_0_0.vlm"},
	{0xFFBCDE3C379E79E6ULL, "vxlt_0_lowtown_docks_n_c0b_0_0.vlm"},
	{0xFFBCEE98275C37CDULL, "vxw_encounter2_56_0_0.vlm"},
	{0xFFBDD1B4AA25A9EDULL, "6118784_m.fxe"},
	{0xFFBEC2ECD4B43F9AULL, "vxlt_gallowstemplar_651_0_0.vlm"},
	{0xFFBEDFF33631EDFAULL, "72269949.ogg"},
	{0xFFBEE89005FF1BC3ULL, "vxht_mansion2b_2f6_0_0.vlm"},
	{0xFFBEF53B2A552067ULL, "one280_merin6120921.cl"},
	{0xFFBF0E4C289C7436ULL, "gen00fl_aveline6116616.cl"},
	{0xFFBF1EB13F45C0A4ULL, "vxht_bartrand_n_4h5_0_0.vlm"},
	{0xFFBF3A28902852BAULL, "art\\levels\\races\\elf\\props\\dalish\\prp_postc.phy"},
	{0xFFBF46CE62023A42ULL, "gen00fl_varric6076029.cl"},
	{0xFFC0183CCDDD3A1FULL, "vxlt_darktown_52z_0_0.vlm"},
	{0xFFC072AF10E2DDE9ULL, "vxht_0_hightown_a3_a6j_0_0.vlm"},
	{0xFFC0A6D19F44C4CBULL, "6101541_f.fxe"},
	{0xFFC0BDC85A2BF98BULL, "hlac_ht_0_hightown_n_2_0y.dds"},
	{0xFFC0D669D1CF96F2ULL, "zz_followers_banter6159219.cl"},
	{0xFFC0ED098BB94CE4ULL, "vxlt_darktown_2xl_0_0.vlm"},
	{0xFFC0F2212A4F226FULL, "vxca_caverns2_7xk_0_1.vlm"},
	{0xFFC12638E4188D87ULL, "331788838.ogg"},
	{0xFFC1A370DBB5CC0CULL, "lt_0_lowtown_a3_n_c_water_21225.msh"},
	{0xFFC1A970DBB5D6FEULL, "lt_0_lowtown_a3_n_c_water_21225.mmh"},
	{0xFFC1B570DBB5EA95ULL, "lt_0_lowtown_a3_n_c_water_21225.mao"},
	{0xFFC2E0BC1274278EULL, "lgt331ar_shw_bloodmage.are"},
	{0xFFC3AA3A81D433B5ULL, "6113108_f.fxe"},
	{0xFFC3BBAE15802F70ULL, "vxht_0_hightown_n_akj_0_0.vlm"},
	{0xFFC501BC7855290EULL, "vxht_keep_57k_0_0.vlm"},
	{0xFFC50A338D1CF845ULL, "t3_wdg_dr1.tnt"},
	{0xFFC5906218341044ULL, "6137635_f.fxe"},
	{0xFFC5B60AB569D80DULL, "ser211ar_dungeon.are"},
	{0xFFC7A31B81E095DCULL, "6071831_m.fxe"},
	{0xFFC7D9B993B0B5CBULL, "art\\vfx\\combat\\vfx_r_asn_assassin_act_c.mmh"},
	{0xFFC8700ED78AC94CULL, "524633605.ogg"},
	{0xFFC936ED6BA84E2BULL, "vxlt_0_lowtown_a3_n_c_gwl_0_0.vlm"},
	{0xFFC96EF0F75BF07DULL, "dae000_fenarel6127250.cl"},
	{0xFFC97AD438C98C46ULL, "art\\levels\\races\\ferelden\\props\\higherclass\\prp_shelfbase.phy"},
	{0xFFC982A0995CB4F4ULL, "6208543_m.fxe"},
	{0xFFCA98FAF7FCE25AULL, "6087780_m.fxe"},
	{0xFFCAF4F90BC0C8D1ULL, "vxht_mansion_n_506_0_0.vlm"},
	{0xFFCB944F1BC94775ULL, "812522542.ogg"},
	{0xFFCB9AF2D571EA5BULL, "6088898_m.fxe"},
	{0xFFCBB6BDB94A9026ULL, "vxlt_darktown_516_0_0.vlm"},
	{0xFFCC011D3CCE23B4ULL, "vxht_0_hightown_b5t_0_0.vlm"},
	{0xFFCC3E93CFD9BBD3ULL, "6081967_m.fxe"},
	{0xFFCC43919C1BC26DULL, "6117684_m.fxe"},
	{0xFFCC6F352A5B0059ULL, "vxdr_0_deeproads_f4y_0_0.vlm"},
	{0xFFCD1C8F27B5E0AAULL, "134360870.ogg"},
	{0xFFCD989A454A94ECULL, "art\\characters\\playercharacter\\humanmale\\arm_lgte.phy"},
	{0xFFCDFD3F56B8C6FBULL, "620606285.ogg"},
	{0xFFCE14DC25A50E6CULL, "6139075_f.fxe"},
	{0xFFCE94C987A9FC3CULL, "183105051.ogg"},
	{0xFFCE9B48F7039EA3ULL, "l_dae270ar_gallows.bnk"},
	{0xFFCEB55B0B995844ULL, "and100_anders6095903.cl"},
	{0xFFCED9953328994FULL, "mag371_elsa6041750.cl"},
	{0xFFCEDCFD89023336ULL, "vxht_0_hightown_bty_0_0.vlm"},
	{0xFFCF54AD6601E636ULL, "hlac_lt_0_lowtown_n_-2_5x.dds"},
	{0xFFCFCA1396606E12ULL, "art\\characters\\weapons\\dagger\\textures\\w_dgr_107a_sa.dds"},
	{0xFFD031E74A3CA856ULL, "w_bonepit_134_0.msh"},
	{0xFFD03BE74A3CB954ULL, "w_bonepit_134_0.mmh"},
	{0xFFD03FE74A3CC067ULL, "w_bonepit_134_0.mao"},
	{0xFFD05B2E237482D9ULL, "6149103_f.fxe"},
	{0xFFD091644F8C2BE5ULL, "vxlt_gallowstemplar_5pj_0_0.vlm"},
	{0xFFD0E8E81021762EULL, "vxlt_0_lowtown_gbr_0_0.vlm"},
	{0xFFD1E50BEBDDBAC0ULL, "949558886.ogg"},
	{0xFFD22D8B077647F0ULL, "vxht_0_hightown_n_alk_0_0.vlm"},
	{0xFFD2B5A4B513A09EULL, "6167877_m.fxe"},
	{0xFFD2C37550713E37ULL, "6060096_m.fxe"},
	{0xFFD31D44915ABF6FULL, "vxdr_0_deeproads_cfv_0_2.vlm"},
	{0xFFD44C25DD5B645DULL, "6060605_m.fxe"},
	{0xFFD45A7E17511546ULL, "6128335_m.fxe"},
	{0xFFD4C30C311FB81DULL, "mag311_keran6062061.cl"},
	{0xFFD4D97E2017C2A7ULL, "vxw_bonepitdestroyed_a3_kv_0_0.vlm"},
	{0xFFD4E9115A234ABCULL, "llac_w_woundedcoast_131073_0.dds"},
	{0xFFD51BBE3589CABEULL, "vxdr_primevaltomb_4zv_0_2.vlm"},
	{0xFFD533D957EAAAAAULL, "hlac_ca_cavern1_haunted_24421_7x.dds"},
	{0xFFD5BAB993BCBE30ULL, "art\\vfx\\combat\\vfx_r_asn_assassin_act_c.anb"},
	{0xFFD5D81DF44D05FEULL, "plt_one100pt_herb_ironwood.plo"},
	{0xFFD65E0E1CF70477ULL, "hlac_w_bonepit_196610_0x.dds"},
	{0xFFD675911F97E93EULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\_textures\\prp_eaglehead_d.dds"},
	{0xFFD67EC49A145BFCULL, "vxht_mansion2b_1is_0_0.vlm"},
	{0xFFD6A6C1E66BFB41ULL, "276600789.ogg"},
	{0xFFD6EBE3A24653C5ULL, "amb_scene_sleeping_snore.wbd"},
	{0xFFD70CB2E996E338ULL, "dae000_bartrand6206931.cl"},
	{0xFFD7A6BFD42A57BBULL, "195064381.ogg"},
	{0xFFD7D9B4D635C8FAULL, "377286779.ogg"},
	{0xFFD840FA585B8B41ULL, "t1_skn_tst.tnt.tnt"},
	{0xFFD89A4283EE6366ULL, "6059264_m.fxe"},
	{0xFFD8FBA0AA042AD0ULL, "vxht_mansion_n_6jb_0_0.vlm"},
	{0xFFD91CE7B8B487FFULL, "genip_exploding_crate_sm.utp"},
	{0xFFD93E750384DE8EULL, "519154623.ogg"},
	{0xFFD992897DE25B0FULL, "vxlt_gallowstemplar_n_3mz_0_0.vlm"},
	{0xFFD9B01E2828B927ULL, "shelf_d.dds"},
	{0xFFD9CE71DF87EE8FULL, "follower_banter6173171.cl"},
	{0xFFDA5B2B8AADD530ULL, "con_ico_no.dds"},
	{0xFFDAF59431FB3EF6ULL, "6104175_m.fxe"},
	{0xFFDB3C691188D36CULL, "sky_sky2_p1.vsh.fxo.dx11"},
	{0xFFDC70B91953DE45ULL, "828552087.ogg"},
	{0xFFDC8F0788D5F187ULL, "llac_dr_slavers_3207_9.dds"},
	{0xFFDCFBF6B1C96B1FULL, "vxlt_darktown_58d_0_0.vlm"},
	{0xFFDD15987F211683ULL, "vxdr_0_deeproads_dvs_0_0.vlm"},
	{0xFFDD9F42849B3E51ULL, "w_bonepitmining_a2_58_0.msh"},
	{0xFFDDAD42849B566CULL, "w_bonepitmining_a2_58_0.mao"},
	{0xFFDDB942849B6A0FULL, "w_bonepitmining_a2_58_0.mmh"},
	{0xFFDEDB44640C1A49ULL, "vxw_bonepitdestroyed_a3_19m_0_0.vlm"},
	{0xFFDF24AAD2DA914BULL, "20023952.ogg"},
	{0xFFDF3D15D1AF7ADBULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_cone_of_cold.ani"},
	{0xFFDFA9B7E75DA297ULL, "6119910_m.fxe"},
	{0xFFE05A7E4164651CULL, "one200_gold_followers6117109.cl"},
	{0xFFE066D3DD8EA521ULL, "6082296_m.fxe"},
	{0xFFE13663942BAA11ULL, "6127131_m.fxe"},
	{0xFFE1E313E25F7965ULL, "vxca_caverns2_559_0_0.vlm"},
	{0xFFE25D447515B2ADULL, "fen210_fenris6108727.cl"},
	{0xFFE2A0D34CE0A44AULL, "6177831_f.fxe"},
	{0xFFE376917649C160ULL, "6163826_m.fxe"},
	{0xFFE3ACAD8E5FBB59ULL, "mer320_hubert6052976.cl"},
	{0xFFE41AB2420D59EEULL, "vxlt_hangedman_rk_0_0.vlm"},
	{0xFFE44E6561B5845DULL, "123982493.ogg"},
	{0xFFE49FB3331623E1ULL, "6202799_m.fxe"},
	{0xFFE4CAD8D7FD437FULL, "6153444_m.fxe"},
	{0xFFE50E9E04FA9065ULL, "vxlt_0_lowtown_a2_hg1_0_0.vlm"},
	{0xFFE575463F2DE2E0ULL, "zz_followers_banter6159312.cl"},
	{0xFFE5968947669170ULL, "s_ballista_02_l2.mao"},
	{0xFFE5B674EE4F780EULL, "mag360_thrask6046983.cl"},
	{0xFFE5D91A7AC93C5CULL, "110413929.ogg"},
	{0xFFE74ECB682F1EF8ULL, "hlac_lt_0_lowtown_a2_n_c_65535_2x.dds"},
	{0xFFE76D7D9BFA4350ULL, "vxht_mansion_n_6mc_0_0.vlm"},
	{0xFFE7B8D3771EBED9ULL, "dae000_orsino6085472.cl"},
	{0xFFE7C790FF27C271ULL, "one100_solivitus6048151.cl"},
	{0xFFE82D4C2C033FF0ULL, "hlac_lt_foundry_n_320_4y.dds"},
	{0xFFE83E464954AB40ULL, "610226036.ogg"},
	{0xFFE8B9881BC20BA3ULL, "hlac_lt_0_lowtown_n_131073_0z.dds"},
	{0xFFE8BD8696F2BAF9ULL, "hlac_lt_0_lowtown_-196609_4x.dds"},
	{0xFFE91866EA688F10ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerc.msh"},
	{0xFFE91EAE870079EFULL, "vxlt_0_lowtown_a3_n_c_er9_0_0.vlm"},
	{0xFFE92666EA68A72AULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerc.mmh"},
	{0xFFE92A66EA68AE61ULL, "art\\levels\\races\\ferelden\\kirkhigh\\ext\\4co_hightown\\vista_gtowerc.mao"},
	{0xFFE93DF391221596ULL, "zz_beggar_test.cnv"},
	{0xFFE96493BA308DC3ULL, "mag110_cut_abomination6037100.cl"},
	{0xFFE9B6074BEF946DULL, "6082559_m.fxe"},
	{0xFFE9C17939727B9EULL, "hlac_lt_0_lowtown_docks_n_-262143_0x.dds"},
	{0xFFE9E9B993064A64ULL, "mag110_samson6053546.cl"},
	{0xFFEA56D5D7D0F6AEULL, "hlac_lt_0_lowtown_docks_a3_-131074_0x.dds"},
	{0xFFEBEC52BD0C9997ULL, "6072495_m.fxe"},
	{0xFFEC5CA6361B485CULL, "vxw_bonepitmining_a2_ee_0_0.vlm"},
	{0xFFECB3A2A48D5BC8ULL, "mag291_vincento6027357.cl"},
	{0xFFED2E022629F2F0ULL, "hlac_w_blightlands_0_12x.dds"},
	{0xFFED3E6E75E34738ULL, "gen00fl_fenris6126465.cl"},
	{0xFFED4515D1BBC5D2ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_cone_of_cold.evt"},
	{0xFFEDD26CF5F79C64ULL, "vxlt_undercityencounter_1ir_0_0.vlm"},
	{0xFFEDD818A9E6712DULL, "hlac_lt_0_lowtown_n_-1_1z.dds"},
	{0xFFEE0F05295C07E9ULL, "345218390.ogg"},
	{0xFFEE2EB8890F2A69ULL, "647413615.ogg"},
	{0xFFEE69E8881DA1C3ULL, "133437778.ogg"},
	{0xFFEED8B0CF84BDA1ULL, "6085396_m.fxe"},
	{0xFFEF010AF3A298AEULL, "llac_w_woundedcoast_-65535_0.dds"},
	{0xFFEF13B12E4D8315ULL, "art\\levels\\races\\natural\\kirkcoastal\\terrainmats\\ter_ccliffblend03_darelief.dds"},
	{0xFFEF153D647D96B3ULL, "115156581.ogg"},
	{0xFFEF23BD32A9F757ULL, "dre000_varric6035169.cl"},
	{0xFFF0489744825F50ULL, "art\\characters\\playercharacter\\humanfemale\\arm_wic.phy"},
	{0xFFF064E5475B87E2ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls08a.msh"},
	{0xFFF066E5475B8A98ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls08a.mmh"},
	{0xFFF072E5475B9F3BULL, "art\\characters\\weapons\\longswords\\w_lsw_ls08a.mao"},
	{0xFFF0C21A815A53BCULL, "6094005_m.fxe"},
	{0xFFF1E574EB6934DDULL, "6114295_f.fxe"},
	{0xFFF1E6E8A6A376FAULL, "vxca_caverns1_mines_ine_0_0.vlm"},
	{0xFFF2957C57331632ULL, "dae_monster_modal.ncs"},
	{0xFFF2A024BCF0A3BDULL, "161104967.ogg"},
	{0xFFF30ABE4260D2ECULL, "hlac_ht_0_hightown_a3_-65535_0y.dds"},
	{0xFFF30C853B1371DCULL, "llac_dr_0_deeproads_5845_6.dds"},
	{0xFFF316EC7B89E7D6ULL, "vxw_encounter3_fv_0_1.vlm"},
	{0xFFF33CB5C9F35BFFULL, "6156744_m.fxe"},
	{0xFFF3627E7E008CFCULL, "vxht_mansion2b_1gq_0_0.vlm"},
	{0xFFF39CB76558CCCEULL, "6059303_m.fxe"},
	{0xFFF444914E26AB34ULL, "art\\vfx\\combat\\vfx_r_stun_grena_proj_p.anb"},
	{0xFFF46415D1C21965ULL, "art\\characters\\playercharacter\\humanmale\\animation\\combat\\mage\\mw_cone_of_cold.gad"},
	{0xFFF5020CC0ECADE1ULL, "qun120cr_javaris6043292.cl"},
	{0xFFF55B367789B76BULL, "vxlt_blackemporium_149_0_0.vlm"},
	{0xFFF573AE56CA7050ULL, "fxc_puke_ani_add.mao"},
	{0xFFF58D426B6D5245ULL, "hlac_w_woundedcoast_65537_0x.dds"},
	{0xFFF5C34A11F12EDEULL, "mrl000_marethari6129362.cl"},
	{0xFFF5F50FDD054109ULL, "vxht_mansion2_1kt_0_0.vlm"},
	{0xFFF67F3EC0790AD3ULL, "6043463_f.fxe"},
	{0xFFF6A27329025D47ULL, "dae000_quintus6154939.cl"},
	{0xFFF6C4223B4D26D7ULL, "vxlt_undercitydungeon_217_0_0.vlm"},
	{0xFFF713EBFF2FB438ULL, "hlac_lt_0_lowtown_docks_a3_n_-65538_6z.dds"},
	{0xFFF7214278B63ED8ULL, "vxlt_gallowscourtyard_n_3vz_0_0.vlm"},
	{0xFFF73B3D733CE2EDULL, "llac_ca_caverns1_24424_22.dds"},
	{0xFFF7664684979C71ULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\.mmh"},
	{0xFFF780468497C82FULL, "art\\levels\\races\\ferelden\\kirklow\\int\\4co_lowtown\\.msh"},
	{0xFFF7CBA9F2A3E2C1ULL, "dae000_keep_bureaucrat_1_st6165309.cl"},
	{0xFFF827E43DE5ACEDULL, "6085459_m.fxe"},
	{0xFFF859322DBAEB01ULL, "lt_0_lowtown_docks_a3_28_0_lc.mmh"},
	{0xFFF9051D6B4C8030ULL, "413895994.ogg"},
	{0xFFF9CA81B26023ABULL, "495869545.ogg"},
	{0xFFF9E8B5FF0C5397ULL, "6070339_f.fxe"},
	{0xFFFA35337849F73FULL, "6166817_m.fxe"},
	{0xFFFA512D9925D3B4ULL, "6093822_m.fxe"},
	{0xFFFA6D0DDA497EAEULL, "hm_lgt_shadyman.mor"},
	{0xFFFA6DE54763EB23ULL, "art\\characters\\weapons\\longswords\\w_lsw_ls08a.phy"},
	{0xFFFA7AF947CF5D21ULL, "6133674_f.fxe"},
	{0xFFFB091C0E14CC37ULL, "hf_refugee_1.mor"},
	{0xFFFBA1E470B06175ULL, "vxht_0_hightown_n_b9r_0_0.vlm"},
	{0xFFFBAA123CC7C725ULL, "6112045_m.fxe"},
	{0xFFFBB96059F8EA69ULL, "mag160_feynriel6107440.cl"},
	{0xFFFBCF1A7D3B81FFULL, "art\\vfx\\combat\\vfx_r_brd_psong_loop_mdl_c.mmh"},
	{0xFFFC2F4AA2DE177BULL, "6114640_f.fxe"},
	{0xFFFD2847B2D62C49ULL, "vxca_cavern1_haunted_iki_0_0.vlm"},
	{0xFFFE0A21EB0F8521ULL, "art\\characters\\playercharacter\\humanmale\\textures\\hm_hlf_masg_0d.dds"},
	{0xFFFE51483B636935ULL, "hlac_lt_gallowscourtyard_131071_2x.dds"},
	{0xFFFE53C99599B928ULL, "and130ip_anders_door.ncs"},
	{0xFFFE773A7284DCA9ULL, "art\\characters\\playercharacter\\humanmale\\animation\\archetypes\\trooper\\baseswordshield\\combat\\meleeattackv1.ani"},
	{0xFFFF0116505D6676ULL, "art\\characters\\playercharacter\\humanfemale\\glv_ave.mao"},
	{0xFFFF2E47FCE86941ULL, "gen_im_trs_gen_arc_lbw_r2.uti"},
	{0xFFFF6F7BABFC4E59ULL, "hlac_lt_undercitydungeon_-1_14y.dds"},
	{0xFFFF86C06EB714E1ULL, "vxlt_house_ek_0_1.vlm"},
	{0xFFFF98BF8A333012ULL, "hlac_lt_undercityserialkiller_-1_19x.dds"},
	{0xFFFF9DBC91680240ULL, "vxca_caverns2_815_0_0.vlm"}
};

const char *findDragonAgeFile(uint64 hash) {
	const DragonAgeFileHash *file = Common::binarySearch(kDragonAgeFilesHashes, ARRAYSIZE(kDragonAgeFilesHashes), hash);
	if (!file)
		return 0;

	return file->value;
}
